/*
Module : AADynamicalTime.cpp
Purpose: Implementation for the algorithms which provides for conversion between Universal Time (both UT1 and UTC) and Terrestrial Time (TT) aka Terrestrial Dynamical Time (TDT) 
         aka Ephemeris Time (ET)
Created: PJN / 29-12-2003
History: PJN / 01-02-2005 1. Fixed a problem with the declaration of the variable "Index" in the function 
                          CAADynamicalTime::DeltaT. Thanks to Mika Heiskanen for reporting this problem.
         PJN / 26-01-2007 1. Update to fit in with new layout of CAADate class
         PJN / 28-01-2007 1. Further updates to fit in with new layout of CAADate class
         PJN / 08-05-2011 1. Fixed a compilation issue on GCC where size_t was undefined in various methods. Thanks to 
                          Carsten A. Arnholm and Andrew Hammond for reporting this bug.
         PJN / 01-05-2012 1. Updated CAADynamicalTime::DeltaT to use new polynomical expressions from Espenak & Meeus 2006.
                          References used: http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html and 
                          http://www.staff.science.uu.nl/~gent0113/deltat/deltat_old.htm (Espenak & Meeus 2006 section). 
                          Thanks to Murphy Chesney for prompting this update.
         PJN / 02-05-2012 1. To further improve the accuracy of the CAADynamicalTime::DeltaT method, the code now uses a 
                          lookup table between the dates of 1 February 1973 to 1 April 2012 (for observed values) and predicted 
                          values from April 2012 till April 2015. These values are as provided by IERS Rapid
                          Service/Prediction Center at http://maia.usno.navy.mil/ser7/deltat.data and 
                          http://maia.usno.navy.mil/ser7/deltat.preds. This lookup table will of course need to be kept up to 
                          date as IERS update this information. As currently coded there is a single discontinuity of c. one second 
                          in early April 2015. At this point http://maia.usno.navy.mil/ser7/deltat.preds indicates an error value 
                          for DeltaT of about 0.9 seconds anyway.
                          2. A new CAADynamicalTime::CumulativeLeapSeconds has been provided. This method takes as input the Julian
                          Day value and returns the cumulative total of Leap seconds which have been applied to this point. For more
                          information about leap seconds please see http://en.wikipedia.org/wiki/Leap_second. Using this method you 
                          can now implement code which converts from Terrestial Time to Coordinated Universal time as follows:
                          
                          double TerrestialTime = some calculation using AA+ algorithms(JD);
                          double DeltaT = CAADynamicalTime::DeltaT(JD);
                          double UniversalTime1 = TerrestialTime - DeltaT/86400.0; //The time of the event using the UT1 time scale
                          double TerrestialAtomicTime = TerrestialTime - (32.184/86400.0); //The time of the event using the TAI time scale
                          double CumulativeLeapSeconds = CAADynamicalTime::CumulativeLeapSeconds(JD);
                          double CoordinatedUniversalTime = (DeltaT - CumulativeLeapSeconds - 32.184)/86400.0 + UniversalTime1; //The time of the event using the UTC time scale
         PJN / 13-10-2012 1. Fixed a typo in the spelling of Coefficient throughout the AADynamicalTime.cpp module
         PJN / 04-08-2013 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st April 2013
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to 1st January 2023
         PJN / 28-10-2013 1. Addition of a TT2UTC method which converts from TT to UTC.
                          2. Addition of a UTC2TT method which converts from UTC to TT.
                          3. Addition of a TT2TAI method which converts from TT to TAI.
                          4. Addition of a TAI2TT method which converts from TAI to TT.
                          5. Addition of a TT2UT1 method which converts from TT to UT1.
                          6. Addition of a UT12TT method which converts from UT1 to TT.
                          7. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st September 2013
                          8. Addition of a UT1MinusUTC method which returns UT1 - UTC.
         PJN / 12-11-2014 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st October 2014
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to 1st January 2024
         PJN / 15-02-2015 1. Updated copyright details.
                          2. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st January 2015
                          3. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to 1st January 2024
                          4. Updated the CumulativeLeapSeconds table from http://maia.usno.navy.mil/ser7/tai-utc.dat to 1st July 2015
         PJN / 05-07-2015 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st April 2015
         PJN / 01-09-2015 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st July 2015
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to 1st January 2025
         PJN / 31-12-2015 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st October 2015
                          2. Verified the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds are up to date
         PJN / 09-03-2016 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st February 2016
                          2. Verified the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds are up to date
         PJN / 07-07-2016 1. Updated the CAADynamicalTime::CumulativeLeapSeconds method as taken from 
                          http://maia.usno.navy.mil/ser7/tai-utc.dat to include the leap second which will occur on 1 January 2017.
                          2. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st July 2016
                          3. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to July 2026
         PJN / 05-02-2017 1. Updated copyright details.
                          2. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st January 2017
                          3. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to Jan 2026
         PJN / 18-02-2017 1. Fixed a transcription error on the DeltaT value for 1 May 2015. The correct value is 67.8012 instead 
                          of 67.8011. Thanks to Luigi Candurro for reporting this error.
                          2. Fixed a number of transcription errors on the predicted DeltaT values from 2019.75 to 2025.75. Thanks 
                          to Luigi Candurro for reporting these errors.
                          3. CAADynamicalTime::TT2UTC is now implemented as TT2UT1 for date ranges prior to 1 January 1961 and 500
                          days after the last leap second (which is currently 1 January 2017). Also CAADynamicalTime::UTC2TT is 
                          now implemented as UT12TT for date ranges prior to 1 January 1961 and 500 days after the last leap 
                          second. These changes address problems where these two methods would end up using a constant offset
                          between UTC and TT for dates away from the current epoch. This problem was discovered while calculating 
                          rise, transit and set times for the Moon in B.C.E years. Thanks to Luigi Candurro for prompting this 
                          update.
         PJN / 19-02-2017 1. Fixed a bug in the CAADynamicalTime::UTC2TT & CAADynamicalTime::TT2UTC methods where the code would
                          incorrectly use BaseMJD instead of JD when determining if the date is in the valid range of UTC. Thanks
                          to Luigi Candurro for reporting this bug.
         PJN / 30-07-2017 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st June 2017
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to Jan 2026
         PJN / 20-01-2018 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st January 2018
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to Jan 2026
         PJN / 22-09-2018 1. Updated the observed DeltaT values from http://toshi.nofs.navy.mil/ser7/deltat.data to 1st April 2018
         PJN / 03-01-2019 1. Updated copyright details.
                          2. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st October 2018
         PJN / 03-01-2019 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st January 2019
         PJN / 08-06-2019 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st April 2019
                          2. Updated the predicted DeltaT values from http://maia.usno.navy.mil/ser7/deltat.preds to October 2027
         PJN / 23-06-2019 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st June 2019
         PJN / 18-08-2019 1. Fixed some further compiler warnings when using VC 2019 Preview v16.3.0 Preview 2.0
         PJN / 28-09-2019 1. Updated the observed DeltaT values from http://maia.usno.navy.mil/ser7/deltat.data to 1st September 2019
         PJN / 22-04-2020 1. Updated the observed DeltaT values from ftp://cddis.gsfc.nasa.gov/pub/products/iers/deltat.data to 1st February 2020
                          2. Reworked C arrays to use std::array
                          3. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 22-11-2020 1. Updated the observed DeltaT values from https://datacenter.iers.org/eop.php to 1st November 2020. The IERS web site
                          is the definitive source for these values as the USNO websites which I have previously used are offline at the moment for
                          modernization efforts (See https://www.usno.navy.mil/ for the details). To obtain values for DeltaT to add to the 
                          g_DeltaTValues table below, the value "UT1-UTC" was taken from Bulletin A and the formula: 
                          DeltaT = AccumulatedLeapsSeconds + 32.184 - (UT1-UTC) was used. The "AccumulatedLeapsSeconds" for 2020 is currently 37 
                          seconds. If you plot the values for DeltaT for 2020 you will see that around June 2020 the value for DeltaT has started to
                          decrease instead of the long term increasing trend for DeltaT. If this trend continues we may see the first negative leap 
                          second occur in the next few years.
         PJN / 23-11-2020 1. Reworked CAADynamicalTime::TT2UTC & CAADynamicalTime::UTC2TT methods to use std::array::size instead of sizeof.
         PJN / 04-12-2020 1. Updated the observed DeltaT values from https://datacenter.iers.org/eop.php to 1st December 2020.
         PJN / 09-01-2021 1. Updated the observed DeltaT values from https://datacenter.iers.org/eop.php to 1st January 2021.
         PJN / 05-02-2021 1. Updated the observed DeltaT values from https://datacenter.iers.org/eop.php to 1st February 2021.
         PJN / 07-03-2021 1. Updated the observed DeltaT values from https://datacenter.iers.org/eop.php to 1st March 2021.
         PJN / 12-03-2021 1. Reworked CAADynamicalTime::DeltaT to use std::upper_bound to speed up logic to find correct entry in the g_DeltaTValues
                          lookup table.
                          2. Reworked CAADynamicalTime::CumulativeLeapSeconds to use std::upper_bound to speed up logic to find correct entry in the 
                          g_LeapSecondCoefficients lookup table.
                          3. Updated the g_DeltaTValues lookup table to use values from the year 1657 to 1972.50 to use data from 
                          https://cddis.nasa.gov/archive/products/iers/historic_deltat.data which provides a granularity of 0.5 of a year. Also 
                          updated the same array to use values from 2 January 1973 to 19 March 2022 from 
                          https://cddis.nasa.gov/archive/products/iers/finals2000A.all which provides a granularity of 1 day. Also updated the same
                          array to use values from the year 2022.25 to 2027.75 from https://cddis.nasa.gov/archive/products/iers/deltat.preds which
                          provides a granularity of 0.25 of a year. All of this was implemented using a program which machine generates the lookup 
                          table using custom code. This new lookup table now includes > 18k elements and provides 1 - 1 precision with all the data
                          released by IERS. The machine generated code will also make it easier to update this data as new data is published by 
                          IERS going forward. Using this new lookup array we now have a discontinuity of c. 3.7 seconds in January 1657 and a 
                          discontinuity of c. 2.5 seconds in October 2027 for the CAADynamicalTime::DeltaT method. With these updates AA+ is now 
                          29.7 Megabytes and 416 thousand lines of C++ source code.
                          4. Implemented a CAADynamicalTime::SetUserDefinedDeltaT method which allows a user provided DeltaT function to be 
                          provided. By default the CAADynamicalTime::DeltaT function uses its own lookup tables and polynomials to calculate 
                          DeltaT, but if CAADynamicalTime::SetUserDefinedDeltaT has been called to provide a custom DeltaT function then the 
                          callback function provided to this method will be used instead.
         PJN / 02-04-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 1 April 2021 and predicted values to 
                          9 April 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 20-04-2021 1. Added a preprocessor directive "AAPLUS_DELTAT_NO_HIGH_PRECISION" which if defined will only use the polynomial 
                          equations instead of lookup tables from IERS for calculating DeltaT in the method CAADynamicalTime::DeltaT. By defining
                          this value you reduce the memory requirements needed for the lookup table "g_DeltaTValues" by 300 KB.
         PJN / 01-05-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 29 April 2021 and predicted values to
                          7 May 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 05-06-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 3 June 2021 and predicted values to
                          11 June 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 08-07-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 8 July 2021 and predicted values to
                          16 July 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 09-08-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 5 August 2021 and predicted values to
                          13 August 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 12-09-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 9 September 2021 and predicted values to
                          17 September 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 23-10-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 21 October 2021 and predicted values to
                          29 October 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 06-11-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 4 November 2021 and predicted values to
                          12 November 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 22-12-2021 1. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 16 December 2021 and predicted values to
                          24 December 2022 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.
         PJN / 24-01-2022 1. Updated copyright details.
                          2. Updated the g_DeltaTValues lookup table to use observed DeltaT values to 20 January 2022 and predicted values to
                          28 January 2023 from https://cddis.nasa.gov/archive/products/iers/finals2000A.all.

Copyright (c) 2003 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


////////////////////////////////// Includes ///////////////////////////////////

#include "stdafx.h"
#include "AADynamicalTime.h"
#include "AADate.h"
#include <cassert>
#include <cstddef>
#include <array>
#include <algorithm>
using namespace std;


////////////////////////////////// Macros / Defines ///////////////////////////

#ifdef _MSC_VER
#pragma warning(disable : 26446 26482)
#endif //#ifdef _MSC_VER


#ifndef AAPLUS_DELTAT_NO_HIGH_PRECISION

struct DeltaTValue
{
  double JD;
  double DeltaT;
};

constexpr array<DeltaTValue, 18940> g_DeltaTValues
{ {
//Values from https://cddis.nasa.gov/archive/products/iers/historic_deltat.data
  { 2326267.50, 44.000    }, //1657.00
  { 2326450.00, 43.000    }, //1657.50
  { 2326632.50, 43.000    }, //1658.00
  { 2326815.00, 41.000    }, //1658.50
  { 2326997.50, 40.000    }, //1659.00
  { 2327180.00, 39.000    }, //1659.50
  { 2327362.50, 38.000    }, //1660.00
  { 2327545.50, 37.000    }, //1660.50
  { 2327728.50, 37.000    }, //1661.00
  { 2327911.00, 36.000    }, //1661.50
  { 2328093.50, 36.000    }, //1662.00
  { 2328276.00, 36.000    }, //1662.50
  { 2328458.50, 37.000    }, //1663.00
  { 2328641.00, 37.000    }, //1663.50
  { 2328823.50, 38.000    }, //1664.00
  { 2329006.50, 37.000    }, //1664.50
  { 2329189.50, 36.000    }, //1665.00
  { 2329372.00, 36.000    }, //1665.50
  { 2329554.50, 35.000    }, //1666.00
  { 2329737.00, 35.000    }, //1666.50
  { 2329919.50, 34.000    }, //1667.00
  { 2330102.00, 33.000    }, //1667.50
  { 2330284.50, 33.000    }, //1668.00
  { 2330467.50, 32.000    }, //1668.50
  { 2330650.50, 32.000    }, //1669.00
  { 2330833.00, 31.000    }, //1669.50
  { 2331015.50, 31.000    }, //1670.00
  { 2331198.00, 30.000    }, //1670.50
  { 2331380.50, 30.000    }, //1671.00
  { 2331563.00, 29.000    }, //1671.50
  { 2331745.50, 29.000    }, //1672.00
  { 2331928.50, 29.000    }, //1672.50
  { 2332111.50, 29.000    }, //1673.00
  { 2332294.00, 29.000    }, //1673.50
  { 2332476.50, 28.000    }, //1674.00
  { 2332659.00, 28.000    }, //1674.50
  { 2332841.50, 27.000    }, //1675.00
  { 2333024.00, 27.000    }, //1675.50
  { 2333206.50, 26.000    }, //1676.00
  { 2333389.50, 26.000    }, //1676.50
  { 2333572.50, 25.000    }, //1677.00
  { 2333755.00, 25.000    }, //1677.50
  { 2333937.50, 25.000    }, //1678.00
  { 2334120.00, 26.000    }, //1678.50
  { 2334302.50, 26.000    }, //1679.00
  { 2334485.00, 26.000    }, //1679.50
  { 2334667.50, 26.000    }, //1680.00
  { 2334850.50, 25.000    }, //1680.50
  { 2335033.50, 25.000    }, //1681.00
  { 2335216.00, 25.000    }, //1681.50
  { 2335398.50, 24.000    }, //1682.00
  { 2335581.00, 24.000    }, //1682.50
  { 2335763.50, 24.000    }, //1683.00
  { 2335946.00, 24.000    }, //1683.50
  { 2336128.50, 24.000    }, //1684.00
  { 2336311.50, 24.000    }, //1684.50
  { 2336494.50, 24.000    }, //1685.00
  { 2336677.00, 24.000    }, //1685.50
  { 2336859.50, 24.000    }, //1686.00
  { 2337042.00, 24.000    }, //1686.50
  { 2337224.50, 23.000    }, //1687.00
  { 2337407.00, 23.000    }, //1687.50
  { 2337589.50, 23.000    }, //1688.00
  { 2337772.50, 23.000    }, //1688.50
  { 2337955.50, 22.000    }, //1689.00
  { 2338138.00, 22.000    }, //1689.50
  { 2338320.50, 22.000    }, //1690.00
  { 2338503.00, 22.000    }, //1690.50
  { 2338685.50, 22.000    }, //1691.00
  { 2338868.00, 21.000    }, //1691.50
  { 2339050.50, 21.000    }, //1692.00
  { 2339233.50, 21.000    }, //1692.50
  { 2339416.50, 21.000    }, //1693.00
  { 2339599.00, 21.000    }, //1693.50
  { 2339781.50, 21.000    }, //1694.00
  { 2339964.00, 21.000    }, //1694.50
  { 2340146.50, 21.000    }, //1695.00
  { 2340329.00, 20.000    }, //1695.50
  { 2340511.50, 20.000    }, //1696.00
  { 2340694.50, 20.000    }, //1696.50
  { 2340877.50, 20.000    }, //1697.00
  { 2341060.00, 20.000    }, //1697.50
  { 2341242.50, 20.000    }, //1698.00
  { 2341425.00, 20.000    }, //1698.50
  { 2341607.50, 20.000    }, //1699.00
  { 2341790.00, 20.000    }, //1699.50
  { 2341972.50, 21.000    }, //1700.00
  { 2342155.00, 21.000    }, //1700.50
  { 2342337.50, 21.000    }, //1701.00
  { 2342520.00, 20.000    }, //1701.50
  { 2342702.50, 20.000    }, //1702.00
  { 2342885.00, 20.000    }, //1702.50
  { 2343067.50, 20.000    }, //1703.00
  { 2343250.00, 20.000    }, //1703.50
  { 2343432.50, 19.000    }, //1704.00
  { 2343615.50, 19.000    }, //1704.50
  { 2343798.50, 19.000    }, //1705.00
  { 2343981.00, 19.000    }, //1705.50
  { 2344163.50, 19.000    }, //1706.00
  { 2344346.00, 20.000    }, //1706.50
  { 2344528.50, 20.000    }, //1707.00
  { 2344711.00, 20.000    }, //1707.50
  { 2344893.50, 20.000    }, //1708.00
  { 2345076.50, 19.000    }, //1708.50
  { 2345259.50, 20.000    }, //1709.00
  { 2345442.00, 20.000    }, //1709.50
  { 2345624.50, 20.000    }, //1710.00
  { 2345807.00, 20.000    }, //1710.50
  { 2345989.50, 20.000    }, //1711.00
  { 2346172.00, 20.000    }, //1711.50
  { 2346354.50, 21.000    }, //1712.00
  { 2346537.50, 21.000    }, //1712.50
  { 2346720.50, 21.000    }, //1713.00
  { 2346903.00, 21.000    }, //1713.50
  { 2347085.50, 21.000    }, //1714.00
  { 2347268.00, 21.000    }, //1714.50
  { 2347450.50, 21.000    }, //1715.00
  { 2347633.00, 21.000    }, //1715.50
  { 2347815.50, 21.000    }, //1716.00
  { 2347998.50, 21.000    }, //1716.50
  { 2348181.50, 21.000    }, //1717.00
  { 2348364.00, 21.000    }, //1717.50
  { 2348546.50, 21.000    }, //1718.00
  { 2348729.00, 21.000    }, //1718.50
  { 2348911.50, 21.000    }, //1719.00
  { 2349094.00, 21.000    }, //1719.50
  { 2349276.50, 21.100    }, //1720.00
  { 2349459.50, 21.000    }, //1720.50
  { 2349642.50, 21.000    }, //1721.00
  { 2349825.00, 21.000    }, //1721.50
  { 2350007.50, 20.900    }, //1722.00
  { 2350190.00, 20.800    }, //1722.50
  { 2350372.50, 20.700    }, //1723.00
  { 2350555.00, 20.600    }, //1723.50
  { 2350737.50, 20.400    }, //1724.00
  { 2350920.50, 20.200    }, //1724.50
  { 2351103.50, 20.000    }, //1725.00
  { 2351286.00, 19.700    }, //1725.50
  { 2351468.50, 19.400    }, //1726.00
  { 2351651.00, 19.100    }, //1726.50
  { 2351833.50, 18.700    }, //1727.00
  { 2352016.00, 18.300    }, //1727.50
  { 2352198.50, 17.800    }, //1728.00
  { 2352381.50, 17.400    }, //1728.50
  { 2352564.50, 17.000    }, //1729.00
  { 2352747.00, 16.800    }, //1729.50
  { 2352929.50, 16.600    }, //1730.00
  { 2353112.00, 16.400    }, //1730.50
  { 2353294.50, 16.100    }, //1731.00
  { 2353477.00, 15.900    }, //1731.50
  { 2353659.50, 15.700    }, //1732.00
  { 2353842.50, 15.500    }, //1732.50
  { 2354025.50, 15.300    }, //1733.00
  { 2354208.00, 15.000    }, //1733.50
  { 2354390.50, 14.700    }, //1734.00
  { 2354573.00, 14.500    }, //1734.50
  { 2354755.50, 14.300    }, //1735.00
  { 2354938.00, 14.200    }, //1735.50
  { 2355120.50, 14.100    }, //1736.00
  { 2355303.50, 14.100    }, //1736.50
  { 2355486.50, 14.100    }, //1737.00
  { 2355669.00, 13.900    }, //1737.50
  { 2355851.50, 13.700    }, //1738.00
  { 2356034.00, 13.600    }, //1738.50
  { 2356216.50, 13.500    }, //1739.00
  { 2356399.00, 13.500    }, //1739.50
  { 2356581.50, 13.500    }, //1740.00
  { 2356764.50, 13.500    }, //1740.50
  { 2356947.50, 13.400    }, //1741.00
  { 2357130.00, 13.400    }, //1741.50
  { 2357312.50, 13.400    }, //1742.00
  { 2357495.00, 13.400    }, //1742.50
  { 2357677.50, 13.300    }, //1743.00
  { 2357860.00, 13.300    }, //1743.50
  { 2358042.50, 13.200    }, //1744.00
  { 2358225.50, 13.200    }, //1744.50
  { 2358408.50, 13.200    }, //1745.00
  { 2358591.00, 13.100    }, //1745.50
  { 2358773.50, 13.100    }, //1746.00
  { 2358956.00, 13.100    }, //1746.50
  { 2359138.50, 13.000    }, //1747.00
  { 2359321.00, 13.200    }, //1747.50
  { 2359503.50, 13.300    }, //1748.00
  { 2359686.50, 13.400    }, //1748.50
  { 2359869.50, 13.500    }, //1749.00
  { 2360052.00, 13.600    }, //1749.50
  { 2360234.50, 13.700    }, //1750.00
  { 2360417.00, 13.800    }, //1750.50
  { 2360599.50, 13.900    }, //1751.00
  { 2360782.00, 14.000    }, //1751.50
  { 2360964.50, 14.000    }, //1752.00
  { 2361147.50, 14.100    }, //1752.50
  { 2361330.50, 14.100    }, //1753.00
  { 2361513.00, 14.100    }, //1753.50
  { 2361695.50, 14.100    }, //1754.00
  { 2361878.00, 14.200    }, //1754.50
  { 2362060.50, 14.300    }, //1755.00
  { 2362243.00, 14.400    }, //1755.50
  { 2362425.50, 14.400    }, //1756.00
  { 2362608.50, 14.500    }, //1756.50
  { 2362791.50, 14.600    }, //1757.00
  { 2362974.00, 14.600    }, //1757.50
  { 2363156.50, 14.700    }, //1758.00
  { 2363339.00, 14.700    }, //1758.50
  { 2363521.50, 14.700    }, //1759.00
  { 2363704.00, 14.800    }, //1759.50
  { 2363886.50, 14.800    }, //1760.00
  { 2364069.50, 14.900    }, //1760.50
  { 2364252.50, 14.900    }, //1761.00
  { 2364435.00, 15.000    }, //1761.50
  { 2364617.50, 15.000    }, //1762.00
  { 2364800.00, 15.100    }, //1762.50
  { 2364982.50, 15.200    }, //1763.00
  { 2365165.00, 15.300    }, //1763.50
  { 2365347.50, 15.400    }, //1764.00
  { 2365530.50, 15.500    }, //1764.50
  { 2365713.50, 15.600    }, //1765.00
  { 2365896.00, 15.600    }, //1765.50
  { 2366078.50, 15.600    }, //1766.00
  { 2366261.00, 15.800    }, //1766.50
  { 2366443.50, 15.900    }, //1767.00
  { 2366626.00, 15.900    }, //1767.50
  { 2366808.50, 15.900    }, //1768.00
  { 2366991.50, 15.800    }, //1768.50
  { 2367174.50, 15.700    }, //1769.00
  { 2367357.00, 15.800    }, //1769.50
  { 2367539.50, 15.700    }, //1770.00
  { 2367722.00, 15.700    }, //1770.50
  { 2367904.50, 15.700    }, //1771.00
  { 2368087.00, 15.800    }, //1771.50
  { 2368269.50, 15.900    }, //1772.00
  { 2368452.50, 16.100    }, //1772.50
  { 2368635.50, 16.100    }, //1773.00
  { 2368818.00, 16.000    }, //1773.50
  { 2369000.50, 15.900    }, //1774.00
  { 2369183.00, 15.900    }, //1774.50
  { 2369365.50, 15.700    }, //1775.00
  { 2369548.00, 15.400    }, //1775.50
  { 2369730.50, 15.300    }, //1776.00
  { 2369913.50, 15.400    }, //1776.50
  { 2370096.50, 15.500    }, //1777.00
  { 2370279.00, 15.600    }, //1777.50
  { 2370461.50, 15.600    }, //1778.00
  { 2370644.00, 15.600    }, //1778.50
  { 2370826.50, 15.600    }, //1779.00
  { 2371009.00, 15.600    }, //1779.50
  { 2371191.50, 15.600    }, //1780.00
  { 2371374.50, 15.600    }, //1780.50
  { 2371557.50, 15.500    }, //1781.00
  { 2371740.00, 15.500    }, //1781.50
  { 2371922.50, 15.400    }, //1782.00
  { 2372105.00, 15.300    }, //1782.50
  { 2372287.50, 15.200    }, //1783.00
  { 2372470.00, 15.100    }, //1783.50
  { 2372652.50, 14.900    }, //1784.00
  { 2372835.50, 14.800    }, //1784.50
  { 2373018.50, 14.600    }, //1785.00
  { 2373201.00, 14.400    }, //1785.50
  { 2373383.50, 14.300    }, //1786.00
  { 2373566.00, 14.200    }, //1786.50
  { 2373748.50, 14.100    }, //1787.00
  { 2373931.00, 14.200    }, //1787.50
  { 2374113.50, 14.200    }, //1788.00
  { 2374296.50, 13.900    }, //1788.50
  { 2374479.50, 13.700    }, //1789.00
  { 2374662.00, 13.500    }, //1789.50
  { 2374844.50, 13.300    }, //1790.00
  { 2375027.00, 13.100    }, //1790.50
  { 2375209.50, 13.000    }, //1791.00
  { 2375392.00, 13.200    }, //1791.50
  { 2375574.50, 13.200    }, //1792.00
  { 2375757.50, 13.100    }, //1792.50
  { 2375940.50, 13.100    }, //1793.00
  { 2376123.00, 13.200    }, //1793.50
  { 2376305.50, 13.300    }, //1794.00
  { 2376488.00, 13.500    }, //1794.50
  { 2376670.50, 13.500    }, //1795.00
  { 2376853.00, 13.400    }, //1795.50
  { 2377035.50, 13.200    }, //1796.00
  { 2377218.50, 13.200    }, //1796.50
  { 2377401.50, 13.100    }, //1797.00
  { 2377584.00, 13.100    }, //1797.50
  { 2377766.50, 13.000    }, //1798.00
  { 2377949.00, 12.800    }, //1798.50
  { 2378131.50, 12.600    }, //1799.00
  { 2378314.00, 12.700    }, //1799.50
  { 2378496.50, 12.600    }, //1800.00
  { 2378679.00, 12.300    }, //1800.50
  { 2378861.50, 12.000    }, //1801.00
  { 2379044.00, 11.900    }, //1801.50
  { 2379226.50, 11.800    }, //1802.00
  { 2379409.00, 11.600    }, //1802.50
  { 2379591.50, 11.400    }, //1803.00
  { 2379774.00, 11.200    }, //1803.50
  { 2379956.50, 11.100    }, //1804.00
  { 2380139.50, 11.100    }, //1804.50
  { 2380322.50, 11.100    }, //1805.00
  { 2380505.00, 11.100    }, //1805.50
  { 2380687.50, 11.100    }, //1806.00
  { 2380870.00, 11.200    }, //1806.50
  { 2381052.50, 11.100    }, //1807.00
  { 2381235.00, 11.100    }, //1807.50
  { 2381417.50, 11.200    }, //1808.00
  { 2381600.50, 11.400    }, //1808.50
  { 2381783.50, 11.500    }, //1809.00
  { 2381966.00, 11.300    }, //1809.50
  { 2382148.50, 11.200    }, //1810.00
  { 2382331.00, 11.400    }, //1810.50
  { 2382513.50, 11.700    }, //1811.00
  { 2382696.00, 11.900    }, //1811.50
  { 2382878.50, 11.900    }, //1812.00
  { 2383061.50, 11.900    }, //1812.50
  { 2383244.50, 11.800    }, //1813.00
  { 2383427.00, 11.700    }, //1813.50
  { 2383609.50, 11.800    }, //1814.00
  { 2383792.00, 11.800    }, //1814.50
  { 2383974.50, 11.800    }, //1815.00
  { 2384157.00, 11.700    }, //1815.50
  { 2384339.50, 11.600    }, //1816.00
  { 2384522.50, 11.600    }, //1816.50
  { 2384705.50, 11.500    }, //1817.00
  { 2384888.00, 11.500    }, //1817.50
  { 2385070.50, 11.400    }, //1818.00
  { 2385253.00, 11.400    }, //1818.50
  { 2385435.50, 11.300    }, //1819.00
  { 2385618.00, 11.300    }, //1819.50
  { 2385800.50, 11.130    }, //1820.00
  { 2385983.50, 11.160    }, //1820.50
  { 2386166.50, 10.940    }, //1821.00
  { 2386349.00, 10.720    }, //1821.50
  { 2386531.50, 10.290    }, //1822.00
  { 2386714.00, 10.040    }, //1822.50
  { 2386896.50, 9.940    }, //1823.00
  { 2387079.00, 9.910    }, //1823.50
  { 2387261.50, 9.880    }, //1824.00
  { 2387444.50, 9.860    }, //1824.50
  { 2387627.50, 9.720    }, //1825.00
  { 2387810.00, 9.670    }, //1825.50
  { 2387992.50, 9.660    }, //1826.00
  { 2388175.00, 9.640    }, //1826.50
  { 2388357.50, 9.510    }, //1827.00
  { 2388540.00, 9.400    }, //1827.50
  { 2388722.50, 9.210    }, //1828.00
  { 2388905.50, 9.000    }, //1828.50
  { 2389088.50, 8.600    }, //1829.00
  { 2389271.00, 8.290    }, //1829.50
  { 2389453.50, 7.950    }, //1830.00
  { 2389636.00, 7.730    }, //1830.50
  { 2389818.50, 7.590    }, //1831.00
  { 2390001.00, 7.490    }, //1831.50
  { 2390183.50, 7.360    }, //1832.00
  { 2390366.50, 7.260    }, //1832.50
  { 2390549.50, 7.100    }, //1833.00
  { 2390732.00, 7.000    }, //1833.50
  { 2390914.50, 6.890    }, //1834.00
  { 2391097.00, 6.820    }, //1834.50
  { 2391279.50, 6.730    }, //1835.00
  { 2391462.00, 6.640    }, //1835.50
  { 2391644.50, 6.390    }, //1836.00
  { 2391827.50, 6.280    }, //1836.50
  { 2392010.50, 6.250    }, //1837.00
  { 2392193.00, 6.270    }, //1837.50
  { 2392375.50, 6.250    }, //1838.00
  { 2392558.00, 6.270    }, //1838.50
  { 2392740.50, 6.220    }, //1839.00
  { 2392923.00, 6.240    }, //1839.50
  { 2393105.50, 6.220    }, //1840.00
  { 2393288.50, 6.270    }, //1840.50
  { 2393471.50, 6.300    }, //1841.00
  { 2393654.00, 6.360    }, //1841.50
  { 2393836.50, 6.350    }, //1842.00
  { 2394019.00, 6.370    }, //1842.50
  { 2394201.50, 6.320    }, //1843.00
  { 2394384.00, 6.330    }, //1843.50
  { 2394566.50, 6.330    }, //1844.00
  { 2394749.50, 6.370    }, //1844.50
  { 2394932.50, 6.370    }, //1845.00
  { 2395115.00, 6.410    }, //1845.50
  { 2395297.50, 6.400    }, //1846.00
  { 2395480.00, 6.440    }, //1846.50
  { 2395662.50, 6.460    }, //1847.00
  { 2395845.00, 6.510    }, //1847.50
  { 2396027.50, 6.480    }, //1848.00
  { 2396210.50, 6.510    }, //1848.50
  { 2396393.50, 6.530    }, //1849.00
  { 2396576.00, 6.580    }, //1849.50
  { 2396758.50, 6.550    }, //1850.00
  { 2396941.00, 6.610    }, //1850.50
  { 2397123.50, 6.690    }, //1851.00
  { 2397306.00, 6.800    }, //1851.50
  { 2397488.50, 6.840    }, //1852.00
  { 2397671.50, 6.940    }, //1852.50
  { 2397854.50, 7.030    }, //1853.00
  { 2398037.00, 7.130    }, //1853.50
  { 2398219.50, 7.150    }, //1854.00
  { 2398402.00, 7.220    }, //1854.50
  { 2398584.50, 7.260    }, //1855.00
  { 2398767.00, 7.300    }, //1855.50
  { 2398949.50, 7.230    }, //1856.00
  { 2399132.50, 7.220    }, //1856.50
  { 2399315.50, 7.210    }, //1857.00
  { 2399498.00, 7.200    }, //1857.50
  { 2399680.50, 6.990    }, //1858.00
  { 2399863.00, 6.980    }, //1858.50
  { 2400045.50, 7.190    }, //1859.00
  { 2400228.00, 7.360    }, //1859.50
  { 2400410.50, 7.350    }, //1860.00
  { 2400593.50, 7.390    }, //1860.50
  { 2400776.50, 7.410    }, //1861.00
  { 2400959.00, 7.450    }, //1861.50
  { 2401141.50, 7.360    }, //1862.00
  { 2401324.00, 7.180    }, //1862.50
  { 2401506.50, 6.950    }, //1863.00
  { 2401689.00, 6.720    }, //1863.50
  { 2401871.50, 6.450    }, //1864.00
  { 2402054.50, 6.240    }, //1864.50
  { 2402237.50, 5.920    }, //1865.00
  { 2402420.00, 5.590    }, //1865.50
  { 2402602.50, 5.150    }, //1866.00
  { 2402785.00, 4.670    }, //1866.50
  { 2402967.50, 4.110    }, //1867.00
  { 2403150.00, 3.520    }, //1867.50
  { 2403332.50, 2.940    }, //1868.00
  { 2403515.50, 2.470    }, //1868.50
  { 2403698.50, 1.970    }, //1869.00
  { 2403881.00, 1.520    }, //1869.50
  { 2404063.50, 1.040    }, //1870.00
  { 2404246.00, 0.600    }, //1870.50
  { 2404428.50, 0.110    }, //1871.00
  { 2404611.00, -0.340    }, //1871.50
  { 2404793.50, -0.820    }, //1872.00
  { 2404976.50, -1.250    }, //1872.50
  { 2405159.50, -1.700    }, //1873.00
  { 2405342.00, -2.080    }, //1873.50
  { 2405524.50, -2.480    }, //1874.00
  { 2405707.00, -2.820    }, //1874.50
  { 2405889.50, -3.190    }, //1875.00
  { 2406072.00, -3.500    }, //1875.50
  { 2406254.50, -3.840    }, //1876.00
  { 2406437.50, -4.140    }, //1876.50
  { 2406620.50, -4.430    }, //1877.00
  { 2406803.00, -4.590    }, //1877.50
  { 2406985.50, -4.790    }, //1878.00
  { 2407168.00, -4.920    }, //1878.50
  { 2407350.50, -5.090    }, //1879.00
  { 2407533.00, -5.240    }, //1879.50
  { 2407715.50, -5.360    }, //1880.00
  { 2407898.50, -5.340    }, //1880.50
  { 2408081.50, -5.370    }, //1881.00
  { 2408264.00, -5.320    }, //1881.50
  { 2408446.50, -5.340    }, //1882.00
  { 2408629.00, -5.330    }, //1882.50
  { 2408811.50, -5.400    }, //1883.00
  { 2408994.00, -5.470    }, //1883.50
  { 2409176.50, -5.580    }, //1884.00
  { 2409359.50, -5.660    }, //1884.50
  { 2409542.50, -5.740    }, //1885.00
  { 2409725.00, -5.680    }, //1885.50
  { 2409907.50, -5.690    }, //1886.00
  { 2410090.00, -5.650    }, //1886.50
  { 2410272.50, -5.670    }, //1887.00
  { 2410455.00, -5.680    }, //1887.50
  { 2410637.50, -5.730    }, //1888.00
  { 2410820.50, -5.720    }, //1888.50
  { 2411003.50, -5.780    }, //1889.00
  { 2411186.00, -5.790    }, //1889.50
  { 2411368.50, -5.860    }, //1890.00
  { 2411551.00, -5.890    }, //1890.50
  { 2411733.50, -6.010    }, //1891.00
  { 2411916.00, -6.130    }, //1891.50
  { 2412098.50, -6.280    }, //1892.00
  { 2412281.50, -6.410    }, //1892.50
  { 2412464.50, -6.530    }, //1893.00
  { 2412647.00, -6.490    }, //1893.50
  { 2412829.50, -6.500    }, //1894.00
  { 2413012.00, -6.450    }, //1894.50
  { 2413194.50, -6.410    }, //1895.00
  { 2413377.00, -6.260    }, //1895.50
  { 2413559.50, -6.110    }, //1896.00
  { 2413742.50, -5.900    }, //1896.50
  { 2413925.50, -5.630    }, //1897.00
  { 2414108.00, -5.130    }, //1897.50
  { 2414290.50, -4.680    }, //1898.00
  { 2414473.00, -4.190    }, //1898.50
  { 2414655.50, -3.720    }, //1899.00
  { 2414838.00, -3.210    }, //1899.50
  { 2415020.50, -2.700    }, //1900.00
  { 2415203.00, -2.090    }, //1900.50
  { 2415385.50, -1.480    }, //1901.00
  { 2415568.00, -0.750    }, //1901.50
  { 2415750.50, -0.080    }, //1902.00
  { 2415933.00, 0.620    }, //1902.50
  { 2416115.50, 1.260    }, //1903.00
  { 2416298.00, 1.950    }, //1903.50
  { 2416480.50, 2.590    }, //1904.00
  { 2416663.50, 3.280    }, //1904.50
  { 2416846.50, 3.920    }, //1905.00
  { 2417029.00, 4.610    }, //1905.50
  { 2417211.50, 5.200    }, //1906.00
  { 2417394.00, 5.730    }, //1906.50
  { 2417576.50, 6.290    }, //1907.00
  { 2417759.00, 7.000    }, //1907.50
  { 2417941.50, 7.680    }, //1908.00
  { 2418124.50, 8.450    }, //1908.50
  { 2418307.50, 9.130    }, //1909.00
  { 2418490.00, 9.780    }, //1909.50
  { 2418672.50, 10.380    }, //1910.00
  { 2418855.00, 10.990    }, //1910.50
  { 2419037.50, 11.640    }, //1911.00
  { 2419220.00, 12.470    }, //1911.50
  { 2419402.50, 13.230    }, //1912.00
  { 2419585.50, 14.000    }, //1912.50
  { 2419768.50, 14.690    }, //1913.00
  { 2419951.00, 15.380    }, //1913.50
  { 2420133.50, 16.000    }, //1914.00
  { 2420316.00, 16.640    }, //1914.50
  { 2420498.50, 17.190    }, //1915.00
  { 2420681.00, 17.720    }, //1915.50
  { 2420863.50, 18.190    }, //1916.00
  { 2421046.50, 18.670    }, //1916.50
  { 2421229.50, 19.130    }, //1917.00
  { 2421412.00, 19.690    }, //1917.50
  { 2421594.50, 20.140    }, //1918.00
  { 2421777.00, 20.540    }, //1918.50
  { 2421959.50, 20.860    }, //1919.00
  { 2422142.00, 21.140    }, //1919.50
  { 2422324.50, 21.410    }, //1920.00
  { 2422507.50, 21.780    }, //1920.50
  { 2422690.50, 22.060    }, //1921.00
  { 2422873.00, 22.300    }, //1921.50
  { 2423055.50, 22.510    }, //1922.00
  { 2423238.00, 22.790    }, //1922.50
  { 2423420.50, 23.010    }, //1923.00
  { 2423603.00, 23.290    }, //1923.50
  { 2423785.50, 23.460    }, //1924.00
  { 2423968.50, 23.550    }, //1924.50
  { 2424151.50, 23.630    }, //1925.00
  { 2424334.00, 23.800    }, //1925.50
  { 2424516.50, 23.950    }, //1926.00
  { 2424699.00, 24.250    }, //1926.50
  { 2424881.50, 24.390    }, //1927.00
  { 2425064.00, 24.420    }, //1927.50
  { 2425246.50, 24.340    }, //1928.00
  { 2425429.50, 24.220    }, //1928.50
  { 2425612.50, 24.100    }, //1929.00
  { 2425795.00, 24.080    }, //1929.50
  { 2425977.50, 24.020    }, //1930.00
  { 2426160.00, 24.040    }, //1930.50
  { 2426342.50, 23.980    }, //1931.00
  { 2426525.00, 23.910    }, //1931.50
  { 2426707.50, 23.890    }, //1932.00
  { 2426890.50, 23.950    }, //1932.50
  { 2427073.50, 23.930    }, //1933.00
  { 2427256.00, 23.920    }, //1933.50
  { 2427438.50, 23.880    }, //1934.00
  { 2427621.00, 23.940    }, //1934.50
  { 2427803.50, 23.910    }, //1935.00
  { 2427986.00, 23.820    }, //1935.50
  { 2428168.50, 23.760    }, //1936.00
  { 2428351.50, 23.870    }, //1936.50
  { 2428534.50, 23.910    }, //1937.00
  { 2428717.00, 23.950    }, //1937.50
  { 2428899.50, 23.960    }, //1938.00
  { 2429082.00, 24.000    }, //1938.50
  { 2429264.50, 24.040    }, //1939.00
  { 2429447.00, 24.200    }, //1939.50
  { 2429629.50, 24.350    }, //1940.00
  { 2429812.50, 24.610    }, //1940.50
  { 2429995.50, 24.820    }, //1941.00
  { 2430178.00, 25.090    }, //1941.50
  { 2430360.50, 25.300    }, //1942.00
  { 2430543.00, 25.560    }, //1942.50
  { 2430725.50, 25.770    }, //1943.00
  { 2430908.00, 26.050    }, //1943.50
  { 2431090.50, 26.270    }, //1944.00
  { 2431273.50, 26.540    }, //1944.50
  { 2431456.50, 26.760    }, //1945.00
  { 2431639.00, 27.040    }, //1945.50
  { 2431821.50, 27.270    }, //1946.00
  { 2432004.00, 27.550    }, //1946.50
  { 2432186.50, 27.770    }, //1947.00
  { 2432369.00, 28.030    }, //1947.50
  { 2432551.50, 28.250    }, //1948.00
  { 2432734.50, 28.500    }, //1948.50
  { 2432917.50, 28.700    }, //1949.00
  { 2433100.00, 28.950    }, //1949.50
  { 2433282.50, 29.150    }, //1950.00
  { 2433465.00, 29.380    }, //1950.50
  { 2433647.50, 29.570    }, //1951.00
  { 2433830.00, 29.800    }, //1951.50
  { 2434012.50, 29.970    }, //1952.00
  { 2434195.50, 30.190    }, //1952.50
  { 2434378.50, 30.360    }, //1953.00
  { 2434561.00, 30.570    }, //1953.50
  { 2434743.50, 30.720    }, //1954.00
  { 2434926.00, 30.930    }, //1954.50
  { 2435108.50, 31.070    }, //1955.00
  { 2435291.00, 31.240    }, //1955.50
  { 2435473.50, 31.349    }, //1956.00
  { 2435656.50, 31.516    }, //1956.50
  { 2435839.50, 31.677    }, //1957.00
  { 2436022.00, 31.923    }, //1957.50
  { 2436204.50, 32.166    }, //1958.00
  { 2436387.00, 32.449    }, //1958.50
  { 2436569.50, 32.671    }, //1959.00
  { 2436752.00, 32.919    }, //1959.50
  { 2436934.50, 33.150    }, //1960.00
  { 2437117.50, 33.397    }, //1960.50
  { 2437300.50, 33.584    }, //1961.00
  { 2437483.00, 33.804    }, //1961.50
  { 2437665.50, 33.992    }, //1962.00
  { 2437848.00, 34.240    }, //1962.50
  { 2438030.50, 34.466    }, //1963.00
  { 2438213.00, 34.731    }, //1963.50
  { 2438395.50, 35.030    }, //1964.00
  { 2438578.50, 35.400    }, //1964.50
  { 2438761.50, 35.738    }, //1965.00
  { 2438944.00, 36.147    }, //1965.50
  { 2439126.50, 36.546    }, //1966.00
  { 2439309.00, 36.995    }, //1966.50
  { 2439491.50, 37.429    }, //1967.00
  { 2439674.00, 37.879    }, //1967.50
  { 2439856.50, 38.291    }, //1968.00
  { 2440039.50, 38.753    }, //1968.50
  { 2440222.50, 39.204    }, //1969.00
  { 2440405.00, 39.707    }, //1969.50
  { 2440587.50, 40.182    }, //1970.00
  { 2440770.00, 40.706    }, //1970.50
  { 2440952.50, 41.170    }, //1971.00
  { 2441135.00, 41.686    }, //1971.50
  { 2441317.50, 42.227    }, //1972.00
  { 2441500.50, 42.825    }, //1972.50

//Values from https://cddis.nasa.gov/archive/products/iers/finals2000A.all
  { 2441684.50, 43.375582 }, //02 January 1973, UT1-UTC= 0.8084178, Accumulated Leap Seconds=12
  { 2441685.50, 43.378384 }, //03 January 1973, UT1-UTC= 0.8056163, Accumulated Leap Seconds=12
  { 2441686.50, 43.381210 }, //04 January 1973, UT1-UTC= 0.8027895, Accumulated Leap Seconds=12
  { 2441687.50, 43.384127 }, //05 January 1973, UT1-UTC= 0.7998729, Accumulated Leap Seconds=12
  { 2441688.50, 43.387186 }, //06 January 1973, UT1-UTC= 0.7968144, Accumulated Leap Seconds=12
  { 2441689.50, 43.390418 }, //07 January 1973, UT1-UTC= 0.7935822, Accumulated Leap Seconds=12
  { 2441690.50, 43.393829 }, //08 January 1973, UT1-UTC= 0.7901710, Accumulated Leap Seconds=12
  { 2441691.50, 43.397395 }, //09 January 1973, UT1-UTC= 0.7866053, Accumulated Leap Seconds=12
  { 2441692.50, 43.401061 }, //10 January 1973, UT1-UTC= 0.7829389, Accumulated Leap Seconds=12
  { 2441693.50, 43.404751 }, //11 January 1973, UT1-UTC= 0.7792487, Accumulated Leap Seconds=12
  { 2441694.50, 43.408378 }, //12 January 1973, UT1-UTC= 0.7756218, Accumulated Leap Seconds=12
  { 2441695.50, 43.411864 }, //13 January 1973, UT1-UTC= 0.7721359, Accumulated Leap Seconds=12
  { 2441696.50, 43.415162 }, //14 January 1973, UT1-UTC= 0.7688378, Accumulated Leap Seconds=12
  { 2441697.50, 43.418276 }, //15 January 1973, UT1-UTC= 0.7657235, Accumulated Leap Seconds=12
  { 2441698.50, 43.421267 }, //16 January 1973, UT1-UTC= 0.7627325, Accumulated Leap Seconds=12
  { 2441699.50, 43.424239 }, //17 January 1973, UT1-UTC= 0.7597608, Accumulated Leap Seconds=12
  { 2441700.50, 43.427307 }, //18 January 1973, UT1-UTC= 0.7566931, Accumulated Leap Seconds=12
  { 2441701.50, 43.430556 }, //19 January 1973, UT1-UTC= 0.7534441, Accumulated Leap Seconds=12
  { 2441702.50, 43.434012 }, //20 January 1973, UT1-UTC= 0.7499880, Accumulated Leap Seconds=12
  { 2441703.50, 43.437635 }, //21 January 1973, UT1-UTC= 0.7463651, Accumulated Leap Seconds=12
  { 2441704.50, 43.441338 }, //22 January 1973, UT1-UTC= 0.7426621, Accumulated Leap Seconds=12
  { 2441705.50, 43.445021 }, //23 January 1973, UT1-UTC= 0.7389789, Accumulated Leap Seconds=12
  { 2441706.50, 43.448600 }, //24 January 1973, UT1-UTC= 0.7353998, Accumulated Leap Seconds=12
  { 2441707.50, 43.452022 }, //25 January 1973, UT1-UTC= 0.7319778, Accumulated Leap Seconds=12
  { 2441708.50, 43.455265 }, //26 January 1973, UT1-UTC= 0.7287348, Accumulated Leap Seconds=12
  { 2441709.50, 43.458333 }, //27 January 1973, UT1-UTC= 0.7256670, Accumulated Leap Seconds=12
  { 2441710.50, 43.461250 }, //28 January 1973, UT1-UTC= 0.7227502, Accumulated Leap Seconds=12
  { 2441711.50, 43.464057 }, //29 January 1973, UT1-UTC= 0.7199427, Accumulated Leap Seconds=12
  { 2441712.50, 43.466813 }, //30 January 1973, UT1-UTC= 0.7171871, Accumulated Leap Seconds=12
  { 2441713.50, 43.469583 }, //31 January 1973, UT1-UTC= 0.7144166, Accumulated Leap Seconds=12
  { 2441714.50, 43.472437 }, //01 February 1973, UT1-UTC= 0.7115628, Accumulated Leap Seconds=12
  { 2441715.50, 43.475434 }, //02 February 1973, UT1-UTC= 0.7085656, Accumulated Leap Seconds=12
  { 2441716.50, 43.478617 }, //03 February 1973, UT1-UTC= 0.7053831, Accumulated Leap Seconds=12
  { 2441717.50, 43.482000 }, //04 February 1973, UT1-UTC= 0.7020005, Accumulated Leap Seconds=12
  { 2441718.50, 43.485563 }, //05 February 1973, UT1-UTC= 0.6984370, Accumulated Leap Seconds=12
  { 2441719.50, 43.489253 }, //06 February 1973, UT1-UTC= 0.6947473, Accumulated Leap Seconds=12
  { 2441720.50, 43.492985 }, //07 February 1973, UT1-UTC= 0.6910149, Accumulated Leap Seconds=12
  { 2441721.50, 43.496665 }, //08 February 1973, UT1-UTC= 0.6873348, Accumulated Leap Seconds=12
  { 2441722.50, 43.500209 }, //09 February 1973, UT1-UTC= 0.6837911, Accumulated Leap Seconds=12
  { 2441723.50, 43.503565 }, //10 February 1973, UT1-UTC= 0.6804349, Accumulated Leap Seconds=12
  { 2441724.50, 43.506731 }, //11 February 1973, UT1-UTC= 0.6772686, Accumulated Leap Seconds=12
  { 2441725.50, 43.509757 }, //12 February 1973, UT1-UTC= 0.6742427, Accumulated Leap Seconds=12
  { 2441726.50, 43.512733 }, //13 February 1973, UT1-UTC= 0.6712667, Accumulated Leap Seconds=12
  { 2441727.50, 43.515767 }, //14 February 1973, UT1-UTC= 0.6682328, Accumulated Leap Seconds=12
  { 2441728.50, 43.518953 }, //15 February 1973, UT1-UTC= 0.6650466, Accumulated Leap Seconds=12
  { 2441729.50, 43.522343 }, //16 February 1973, UT1-UTC= 0.6616571, Accumulated Leap Seconds=12
  { 2441730.50, 43.525928 }, //17 February 1973, UT1-UTC= 0.6580719, Accumulated Leap Seconds=12
  { 2441731.50, 43.529646 }, //18 February 1973, UT1-UTC= 0.6543535, Accumulated Leap Seconds=12
  { 2441732.50, 43.533403 }, //19 February 1973, UT1-UTC= 0.6505970, Accumulated Leap Seconds=12
  { 2441733.50, 43.537100 }, //20 February 1973, UT1-UTC= 0.6468995, Accumulated Leap Seconds=12
  { 2441734.50, 43.540664 }, //21 February 1973, UT1-UTC= 0.6433358, Accumulated Leap Seconds=12
  { 2441735.50, 43.544053 }, //22 February 1973, UT1-UTC= 0.6399465, Accumulated Leap Seconds=12
  { 2441736.50, 43.547261 }, //23 February 1973, UT1-UTC= 0.6367394, Accumulated Leap Seconds=12
  { 2441737.50, 43.550304 }, //24 February 1973, UT1-UTC= 0.6336961, Accumulated Leap Seconds=12
  { 2441738.50, 43.553222 }, //25 February 1973, UT1-UTC= 0.6307782, Accumulated Leap Seconds=12
  { 2441739.50, 43.556068 }, //26 February 1973, UT1-UTC= 0.6279321, Accumulated Leap Seconds=12
  { 2441740.50, 43.558907 }, //27 February 1973, UT1-UTC= 0.6250933, Accumulated Leap Seconds=12
  { 2441741.50, 43.561807 }, //28 February 1973, UT1-UTC= 0.6221928, Accumulated Leap Seconds=12
  { 2441742.50, 43.564835 }, //01 March 1973, UT1-UTC= 0.6191649, Accumulated Leap Seconds=12
  { 2441743.50, 43.568043 }, //02 March 1973, UT1-UTC= 0.6159565, Accumulated Leap Seconds=12
  { 2441744.50, 43.571464 }, //03 March 1973, UT1-UTC= 0.6125364, Accumulated Leap Seconds=12
  { 2441745.50, 43.575094 }, //04 March 1973, UT1-UTC= 0.6089055, Accumulated Leap Seconds=12
  { 2441746.50, 43.578896 }, //05 March 1973, UT1-UTC= 0.6051039, Accumulated Leap Seconds=12
  { 2441747.50, 43.582790 }, //06 March 1973, UT1-UTC= 0.6012100, Accumulated Leap Seconds=12
  { 2441748.50, 43.586673 }, //07 March 1973, UT1-UTC= 0.5973269, Accumulated Leap Seconds=12
  { 2441749.50, 43.590443 }, //08 March 1973, UT1-UTC= 0.5935565, Accumulated Leap Seconds=12
  { 2441750.50, 43.594031 }, //09 March 1973, UT1-UTC= 0.5899692, Accumulated Leap Seconds=12
  { 2441751.50, 43.597418 }, //10 March 1973, UT1-UTC= 0.5865821, Accumulated Leap Seconds=12
  { 2441752.50, 43.600646 }, //11 March 1973, UT1-UTC= 0.5833541, Accumulated Leap Seconds=12
  { 2441753.50, 43.603801 }, //12 March 1973, UT1-UTC= 0.5801992, Accumulated Leap Seconds=12
  { 2441754.50, 43.606988 }, //13 March 1973, UT1-UTC= 0.5770120, Accumulated Leap Seconds=12
  { 2441755.50, 43.610303 }, //14 March 1973, UT1-UTC= 0.5736970, Accumulated Leap Seconds=12
  { 2441756.50, 43.613807 }, //15 March 1973, UT1-UTC= 0.5701931, Accumulated Leap Seconds=12
  { 2441757.50, 43.617511 }, //16 March 1973, UT1-UTC= 0.5664886, Accumulated Leap Seconds=12
  { 2441758.50, 43.621378 }, //17 March 1973, UT1-UTC= 0.5626220, Accumulated Leap Seconds=12
  { 2441759.50, 43.625330 }, //18 March 1973, UT1-UTC= 0.5586697, Accumulated Leap Seconds=12
  { 2441760.50, 43.629275 }, //19 March 1973, UT1-UTC= 0.5547245, Accumulated Leap Seconds=12
  { 2441761.50, 43.633128 }, //20 March 1973, UT1-UTC= 0.5508716, Accumulated Leap Seconds=12
  { 2441762.50, 43.636829 }, //21 March 1973, UT1-UTC= 0.5471712, Accumulated Leap Seconds=12
  { 2441763.50, 43.640350 }, //22 March 1973, UT1-UTC= 0.5436500, Accumulated Leap Seconds=12
  { 2441764.50, 43.643696 }, //23 March 1973, UT1-UTC= 0.5403039, Accumulated Leap Seconds=12
  { 2441765.50, 43.646897 }, //24 March 1973, UT1-UTC= 0.5371033, Accumulated Leap Seconds=12
  { 2441766.50, 43.649999 }, //25 March 1973, UT1-UTC= 0.5340013, Accumulated Leap Seconds=12
  { 2441767.50, 43.653060 }, //26 March 1973, UT1-UTC= 0.5309397, Accumulated Leap Seconds=12
  { 2441768.50, 43.656145 }, //27 March 1973, UT1-UTC= 0.5278547, Accumulated Leap Seconds=12
  { 2441769.50, 43.659316 }, //28 March 1973, UT1-UTC= 0.5246836, Accumulated Leap Seconds=12
  { 2441770.50, 43.662628 }, //29 March 1973, UT1-UTC= 0.5213715, Accumulated Leap Seconds=12
  { 2441771.50, 43.666122 }, //30 March 1973, UT1-UTC= 0.5178783, Accumulated Leap Seconds=12
  { 2441772.50, 43.669813 }, //31 March 1973, UT1-UTC= 0.5141868, Accumulated Leap Seconds=12
  { 2441773.50, 43.673686 }, //01 April 1973, UT1-UTC= 0.5103137, Accumulated Leap Seconds=12
  { 2441774.50, 43.677684 }, //02 April 1973, UT1-UTC= 0.5063163, Accumulated Leap Seconds=12
  { 2441775.50, 43.681711 }, //03 April 1973, UT1-UTC= 0.5022885, Accumulated Leap Seconds=12
  { 2441776.50, 43.685659 }, //04 April 1973, UT1-UTC= 0.4983412, Accumulated Leap Seconds=12
  { 2441777.50, 43.689430 }, //05 April 1973, UT1-UTC= 0.4945702, Accumulated Leap Seconds=12
  { 2441778.50, 43.692977 }, //06 April 1973, UT1-UTC= 0.4910225, Accumulated Leap Seconds=12
  { 2441779.50, 43.696320 }, //07 April 1973, UT1-UTC= 0.4876799, Accumulated Leap Seconds=12
  { 2441780.50, 43.699535 }, //08 April 1973, UT1-UTC= 0.4844647, Accumulated Leap Seconds=12
  { 2441781.50, 43.702730 }, //09 April 1973, UT1-UTC= 0.4812696, Accumulated Leap Seconds=12
  { 2441782.50, 43.706007 }, //10 April 1973, UT1-UTC= 0.4779933, Accumulated Leap Seconds=12
  { 2441783.50, 43.709431 }, //11 April 1973, UT1-UTC= 0.4745689, Accumulated Leap Seconds=12
  { 2441784.50, 43.713022 }, //12 April 1973, UT1-UTC= 0.4709778, Accumulated Leap Seconds=12
  { 2441785.50, 43.716753 }, //13 April 1973, UT1-UTC= 0.4672468, Accumulated Leap Seconds=12
  { 2441786.50, 43.720564 }, //14 April 1973, UT1-UTC= 0.4634359, Accumulated Leap Seconds=12
  { 2441787.50, 43.724378 }, //15 April 1973, UT1-UTC= 0.4596221, Accumulated Leap Seconds=12
  { 2441788.50, 43.728118 }, //16 April 1973, UT1-UTC= 0.4558821, Accumulated Leap Seconds=12
  { 2441789.50, 43.731722 }, //17 April 1973, UT1-UTC= 0.4522776, Accumulated Leap Seconds=12
  { 2441790.50, 43.735155 }, //18 April 1973, UT1-UTC= 0.4488449, Accumulated Leap Seconds=12
  { 2441791.50, 43.738409 }, //19 April 1973, UT1-UTC= 0.4455907, Accumulated Leap Seconds=12
  { 2441792.50, 43.741506 }, //20 April 1973, UT1-UTC= 0.4424939, Accumulated Leap Seconds=12
  { 2441793.50, 43.744488 }, //21 April 1973, UT1-UTC= 0.4395122, Accumulated Leap Seconds=12
  { 2441794.50, 43.747410 }, //22 April 1973, UT1-UTC= 0.4365897, Accumulated Leap Seconds=12
  { 2441795.50, 43.750335 }, //23 April 1973, UT1-UTC= 0.4336648, Accumulated Leap Seconds=12
  { 2441796.50, 43.753323 }, //24 April 1973, UT1-UTC= 0.4306768, Accumulated Leap Seconds=12
  { 2441797.50, 43.756427 }, //25 April 1973, UT1-UTC= 0.4275726, Accumulated Leap Seconds=12
  { 2441798.50, 43.759689 }, //26 April 1973, UT1-UTC= 0.4243110, Accumulated Leap Seconds=12
  { 2441799.50, 43.763132 }, //27 April 1973, UT1-UTC= 0.4208678, Accumulated Leap Seconds=12
  { 2441800.50, 43.766757 }, //28 April 1973, UT1-UTC= 0.4172431, Accumulated Leap Seconds=12
  { 2441801.50, 43.770530 }, //29 April 1973, UT1-UTC= 0.4134697, Accumulated Leap Seconds=12
  { 2441802.50, 43.774383 }, //30 April 1973, UT1-UTC= 0.4096171, Accumulated Leap Seconds=12
  { 2441803.50, 43.778216 }, //01 May 1973, UT1-UTC= 0.4057844, Accumulated Leap Seconds=12
  { 2441804.50, 43.781924 }, //02 May 1973, UT1-UTC= 0.4020762, Accumulated Leap Seconds=12
  { 2441805.50, 43.785431 }, //03 May 1973, UT1-UTC= 0.3985685, Accumulated Leap Seconds=12
  { 2441806.50, 43.788722 }, //04 May 1973, UT1-UTC= 0.3952777, Accumulated Leap Seconds=12
  { 2441807.50, 43.791849 }, //05 May 1973, UT1-UTC= 0.3921511, Accumulated Leap Seconds=12
  { 2441808.50, 43.794914 }, //06 May 1973, UT1-UTC= 0.3890860, Accumulated Leap Seconds=12
  { 2441809.50, 43.798030 }, //07 May 1973, UT1-UTC= 0.3859701, Accumulated Leap Seconds=12
  { 2441810.50, 43.801278 }, //08 May 1973, UT1-UTC= 0.3827218, Accumulated Leap Seconds=12
  { 2441811.50, 43.804686 }, //09 May 1973, UT1-UTC= 0.3793140, Accumulated Leap Seconds=12
  { 2441812.50, 43.808226 }, //10 May 1973, UT1-UTC= 0.3757735, Accumulated Leap Seconds=12
  { 2441813.50, 43.811837 }, //11 May 1973, UT1-UTC= 0.3721634, Accumulated Leap Seconds=12
  { 2441814.50, 43.815440 }, //12 May 1973, UT1-UTC= 0.3685599, Accumulated Leap Seconds=12
  { 2441815.50, 43.818965 }, //13 May 1973, UT1-UTC= 0.3650348, Accumulated Leap Seconds=12
  { 2441816.50, 43.822355 }, //14 May 1973, UT1-UTC= 0.3616450, Accumulated Leap Seconds=12
  { 2441817.50, 43.825573 }, //15 May 1973, UT1-UTC= 0.3584266, Accumulated Leap Seconds=12
  { 2441818.50, 43.828608 }, //16 May 1973, UT1-UTC= 0.3553922, Accumulated Leap Seconds=12
  { 2441819.50, 43.831470 }, //17 May 1973, UT1-UTC= 0.3525297, Accumulated Leap Seconds=12
  { 2441820.50, 43.834196 }, //18 May 1973, UT1-UTC= 0.3498042, Accumulated Leap Seconds=12
  { 2441821.50, 43.836836 }, //19 May 1973, UT1-UTC= 0.3471638, Accumulated Leap Seconds=12
  { 2441822.50, 43.839452 }, //20 May 1973, UT1-UTC= 0.3445481, Accumulated Leap Seconds=12
  { 2441823.50, 43.842103 }, //21 May 1973, UT1-UTC= 0.3418967, Accumulated Leap Seconds=12
  { 2441824.50, 43.844843 }, //22 May 1973, UT1-UTC= 0.3391570, Accumulated Leap Seconds=12
  { 2441825.50, 43.847711 }, //23 May 1973, UT1-UTC= 0.3362893, Accumulated Leap Seconds=12
  { 2441826.50, 43.850730 }, //24 May 1973, UT1-UTC= 0.3332699, Accumulated Leap Seconds=12
  { 2441827.50, 43.853906 }, //25 May 1973, UT1-UTC= 0.3300944, Accumulated Leap Seconds=12
  { 2441828.50, 43.857218 }, //26 May 1973, UT1-UTC= 0.3267817, Accumulated Leap Seconds=12
  { 2441829.50, 43.860621 }, //27 May 1973, UT1-UTC= 0.3233789, Accumulated Leap Seconds=12
  { 2441830.50, 43.864039 }, //28 May 1973, UT1-UTC= 0.3199611, Accumulated Leap Seconds=12
  { 2441831.50, 43.867380 }, //29 May 1973, UT1-UTC= 0.3166201, Accumulated Leap Seconds=12
  { 2441832.50, 43.870560 }, //30 May 1973, UT1-UTC= 0.3134398, Accumulated Leap Seconds=12
  { 2441833.50, 43.873536 }, //31 May 1973, UT1-UTC= 0.3104640, Accumulated Leap Seconds=12
  { 2441834.50, 43.876327 }, //01 June 1973, UT1-UTC= 0.3076727, Accumulated Leap Seconds=12
  { 2441835.50, 43.879016 }, //02 June 1973, UT1-UTC= 0.3049837, Accumulated Leap Seconds=12
  { 2441836.50, 43.881719 }, //03 June 1973, UT1-UTC= 0.3022809, Accumulated Leap Seconds=12
  { 2441837.50, 43.884539 }, //04 June 1973, UT1-UTC= 0.2994605, Accumulated Leap Seconds=12
  { 2441838.50, 43.887529 }, //05 June 1973, UT1-UTC= 0.2964705, Accumulated Leap Seconds=12
  { 2441839.50, 43.890675 }, //06 June 1973, UT1-UTC= 0.2933252, Accumulated Leap Seconds=12
  { 2441840.50, 43.893909 }, //07 June 1973, UT1-UTC= 0.2900913, Accumulated Leap Seconds=12
  { 2441841.50, 43.897142 }, //08 June 1973, UT1-UTC= 0.2868582, Accumulated Leap Seconds=12
  { 2441842.50, 43.900291 }, //09 June 1973, UT1-UTC= 0.2837086, Accumulated Leap Seconds=12
  { 2441843.50, 43.903298 }, //10 June 1973, UT1-UTC= 0.2807024, Accumulated Leap Seconds=12
  { 2441844.50, 43.906127 }, //11 June 1973, UT1-UTC= 0.2778733, Accumulated Leap Seconds=12
  { 2441845.50, 43.908768 }, //12 June 1973, UT1-UTC= 0.2752317, Accumulated Leap Seconds=12
  { 2441846.50, 43.911233 }, //13 June 1973, UT1-UTC= 0.2727668, Accumulated Leap Seconds=12
  { 2441847.50, 43.913552 }, //14 June 1973, UT1-UTC= 0.2704483, Accumulated Leap Seconds=12
  { 2441848.50, 43.915771 }, //15 June 1973, UT1-UTC= 0.2682287, Accumulated Leap Seconds=12
  { 2441849.50, 43.917951 }, //16 June 1973, UT1-UTC= 0.2660491, Accumulated Leap Seconds=12
  { 2441850.50, 43.920152 }, //17 June 1973, UT1-UTC= 0.2638477, Accumulated Leap Seconds=12
  { 2441851.50, 43.922430 }, //18 June 1973, UT1-UTC= 0.2615698, Accumulated Leap Seconds=12
  { 2441852.50, 43.924826 }, //19 June 1973, UT1-UTC= 0.2591741, Accumulated Leap Seconds=12
  { 2441853.50, 43.927363 }, //20 June 1973, UT1-UTC= 0.2566369, Accumulated Leap Seconds=12
  { 2441854.50, 43.930045 }, //21 June 1973, UT1-UTC= 0.2539551, Accumulated Leap Seconds=12
  { 2441855.50, 43.932854 }, //22 June 1973, UT1-UTC= 0.2511461, Accumulated Leap Seconds=12
  { 2441856.50, 43.935751 }, //23 June 1973, UT1-UTC= 0.2482491, Accumulated Leap Seconds=12
  { 2441857.50, 43.938677 }, //24 June 1973, UT1-UTC= 0.2453232, Accumulated Leap Seconds=12
  { 2441858.50, 43.941557 }, //25 June 1973, UT1-UTC= 0.2424429, Accumulated Leap Seconds=12
  { 2441859.50, 43.944317 }, //26 June 1973, UT1-UTC= 0.2396828, Accumulated Leap Seconds=12
  { 2441860.50, 43.946906 }, //27 June 1973, UT1-UTC= 0.2370944, Accumulated Leap Seconds=12
  { 2441861.50, 43.949318 }, //28 June 1973, UT1-UTC= 0.2346823, Accumulated Leap Seconds=12
  { 2441862.50, 43.951608 }, //29 June 1973, UT1-UTC= 0.2323922, Accumulated Leap Seconds=12
  { 2441863.50, 43.953878 }, //30 June 1973, UT1-UTC= 0.2301216, Accumulated Leap Seconds=12
  { 2441864.50, 43.956244 }, //01 July 1973, UT1-UTC= 0.2277557, Accumulated Leap Seconds=12
  { 2441865.50, 43.958788 }, //02 July 1973, UT1-UTC= 0.2252119, Accumulated Leap Seconds=12
  { 2441866.50, 43.961527 }, //03 July 1973, UT1-UTC= 0.2224729, Accumulated Leap Seconds=12
  { 2441867.50, 43.964410 }, //04 July 1973, UT1-UTC= 0.2195902, Accumulated Leap Seconds=12
  { 2441868.50, 43.967339 }, //05 July 1973, UT1-UTC= 0.2166607, Accumulated Leap Seconds=12
  { 2441869.50, 43.970211 }, //06 July 1973, UT1-UTC= 0.2137885, Accumulated Leap Seconds=12
  { 2441870.50, 43.972945 }, //07 July 1973, UT1-UTC= 0.2110549, Accumulated Leap Seconds=12
  { 2441871.50, 43.975495 }, //08 July 1973, UT1-UTC= 0.2085047, Accumulated Leap Seconds=12
  { 2441872.50, 43.977851 }, //09 July 1973, UT1-UTC= 0.2061493, Accumulated Leap Seconds=12
  { 2441873.50, 43.980025 }, //10 July 1973, UT1-UTC= 0.2039750, Accumulated Leap Seconds=12
  { 2441874.50, 43.982050 }, //11 July 1973, UT1-UTC= 0.2019501, Accumulated Leap Seconds=12
  { 2441875.50, 43.983971 }, //12 July 1973, UT1-UTC= 0.2000285, Accumulated Leap Seconds=12
  { 2441876.50, 43.985848 }, //13 July 1973, UT1-UTC= 0.1981521, Accumulated Leap Seconds=12
  { 2441877.50, 43.987742 }, //14 July 1973, UT1-UTC= 0.1962575, Accumulated Leap Seconds=12
  { 2441878.50, 43.989716 }, //15 July 1973, UT1-UTC= 0.1942844, Accumulated Leap Seconds=12
  { 2441879.50, 43.991815 }, //16 July 1973, UT1-UTC= 0.1921851, Accumulated Leap Seconds=12
  { 2441880.50, 43.994069 }, //17 July 1973, UT1-UTC= 0.1899311, Accumulated Leap Seconds=12
  { 2441881.50, 43.996483 }, //18 July 1973, UT1-UTC= 0.1875170, Accumulated Leap Seconds=12
  { 2441882.50, 43.999039 }, //19 July 1973, UT1-UTC= 0.1849610, Accumulated Leap Seconds=12
  { 2441883.50, 44.001696 }, //20 July 1973, UT1-UTC= 0.1823037, Accumulated Leap Seconds=12
  { 2441884.50, 44.004396 }, //21 July 1973, UT1-UTC= 0.1796035, Accumulated Leap Seconds=12
  { 2441885.50, 44.007071 }, //22 July 1973, UT1-UTC= 0.1769293, Accumulated Leap Seconds=12
  { 2441886.50, 44.009652 }, //23 July 1973, UT1-UTC= 0.1743483, Accumulated Leap Seconds=12
  { 2441887.50, 44.012090 }, //24 July 1973, UT1-UTC= 0.1719096, Accumulated Leap Seconds=12
  { 2441888.50, 44.014373 }, //25 July 1973, UT1-UTC= 0.1696270, Accumulated Leap Seconds=12
  { 2441889.50, 44.016534 }, //26 July 1973, UT1-UTC= 0.1674656, Accumulated Leap Seconds=12
  { 2441890.50, 44.018657 }, //27 July 1973, UT1-UTC= 0.1653434, Accumulated Leap Seconds=12
  { 2441891.50, 44.020849 }, //28 July 1973, UT1-UTC= 0.1631514, Accumulated Leap Seconds=12
  { 2441892.50, 44.023210 }, //29 July 1973, UT1-UTC= 0.1607898, Accumulated Leap Seconds=12
  { 2441893.50, 44.025793 }, //30 July 1973, UT1-UTC= 0.1582066, Accumulated Leap Seconds=12
  { 2441894.50, 44.028582 }, //31 July 1973, UT1-UTC= 0.1554181, Accumulated Leap Seconds=12
  { 2441895.50, 44.031496 }, //01 August 1973, UT1-UTC= 0.1525044, Accumulated Leap Seconds=12
  { 2441896.50, 44.034422 }, //02 August 1973, UT1-UTC= 0.1495778, Accumulated Leap Seconds=12
  { 2441897.50, 44.037255 }, //03 August 1973, UT1-UTC= 0.1467448, Accumulated Leap Seconds=12
  { 2441898.50, 44.039922 }, //04 August 1973, UT1-UTC= 0.1440777, Accumulated Leap Seconds=12
  { 2441899.50, 44.042395 }, //05 August 1973, UT1-UTC= 0.1416050, Accumulated Leap Seconds=12
  { 2441900.50, 44.044681 }, //06 August 1973, UT1-UTC= 0.1393192, Accumulated Leap Seconds=12
  { 2441901.50, 44.046812 }, //07 August 1973, UT1-UTC= 0.1371881, Accumulated Leap Seconds=12
  { 2441902.50, 44.048835 }, //08 August 1973, UT1-UTC= 0.1351648, Accumulated Leap Seconds=12
  { 2441903.50, 44.050808 }, //09 August 1973, UT1-UTC= 0.1331923, Accumulated Leap Seconds=12
  { 2441904.50, 44.052793 }, //10 August 1973, UT1-UTC= 0.1312073, Accumulated Leap Seconds=12
  { 2441905.50, 44.054853 }, //11 August 1973, UT1-UTC= 0.1291470, Accumulated Leap Seconds=12
  { 2441906.50, 44.057043 }, //12 August 1973, UT1-UTC= 0.1269571, Accumulated Leap Seconds=12
  { 2441907.50, 44.059400 }, //13 August 1973, UT1-UTC= 0.1246002, Accumulated Leap Seconds=12
  { 2441908.50, 44.061938 }, //14 August 1973, UT1-UTC= 0.1220622, Accumulated Leap Seconds=12
  { 2441909.50, 44.064645 }, //15 August 1973, UT1-UTC= 0.1193548, Accumulated Leap Seconds=12
  { 2441910.50, 44.067484 }, //16 August 1973, UT1-UTC= 0.1165163, Accumulated Leap Seconds=12
  { 2441911.50, 44.070393 }, //17 August 1973, UT1-UTC= 0.1136071, Accumulated Leap Seconds=12
  { 2441912.50, 44.073299 }, //18 August 1973, UT1-UTC= 0.1107006, Accumulated Leap Seconds=12
  { 2441913.50, 44.076132 }, //19 August 1973, UT1-UTC= 0.1078680, Accumulated Leap Seconds=12
  { 2441914.50, 44.078839 }, //20 August 1973, UT1-UTC= 0.1051613, Accumulated Leap Seconds=12
  { 2441915.50, 44.081403 }, //21 August 1973, UT1-UTC= 0.1025971, Accumulated Leap Seconds=12
  { 2441916.50, 44.083852 }, //22 August 1973, UT1-UTC= 0.1001482, Accumulated Leap Seconds=12
  { 2441917.50, 44.086255 }, //23 August 1973, UT1-UTC= 0.0977447, Accumulated Leap Seconds=12
  { 2441918.50, 44.088710 }, //24 August 1973, UT1-UTC= 0.0952896, Accumulated Leap Seconds=12
  { 2441919.50, 44.091315 }, //25 August 1973, UT1-UTC= 0.0926851, Accumulated Leap Seconds=12
  { 2441920.50, 44.094137 }, //26 August 1973, UT1-UTC= 0.0898626, Accumulated Leap Seconds=12
  { 2441921.50, 44.097192 }, //27 August 1973, UT1-UTC= 0.0868080, Accumulated Leap Seconds=12
  { 2441922.50, 44.100430 }, //28 August 1973, UT1-UTC= 0.0835702, Accumulated Leap Seconds=12
  { 2441923.50, 44.103753 }, //29 August 1973, UT1-UTC= 0.0802474, Accumulated Leap Seconds=12
  { 2441924.50, 44.107045 }, //30 August 1973, UT1-UTC= 0.0769554, Accumulated Leap Seconds=12
  { 2441925.50, 44.110206 }, //31 August 1973, UT1-UTC= 0.0737935, Accumulated Leap Seconds=12
  { 2441926.50, 44.113179 }, //01 September 1973, UT1-UTC= 0.0708212, Accumulated Leap Seconds=12
  { 2441927.50, 44.115947 }, //02 September 1973, UT1-UTC= 0.0680534, Accumulated Leap Seconds=12
  { 2441928.50, 44.118530 }, //03 September 1973, UT1-UTC= 0.0654699, Accumulated Leap Seconds=12
  { 2441929.50, 44.120972 }, //04 September 1973, UT1-UTC= 0.0630280, Accumulated Leap Seconds=12
  { 2441930.50, 44.123327 }, //05 September 1973, UT1-UTC= 0.0606731, Accumulated Leap Seconds=12
  { 2441931.50, 44.125656 }, //06 September 1973, UT1-UTC= 0.0583445, Accumulated Leap Seconds=12
  { 2441932.50, 44.128020 }, //07 September 1973, UT1-UTC= 0.0559804, Accumulated Leap Seconds=12
  { 2441933.50, 44.130476 }, //08 September 1973, UT1-UTC= 0.0535242, Accumulated Leap Seconds=12
  { 2441934.50, 44.133068 }, //09 September 1973, UT1-UTC= 0.0509322, Accumulated Leap Seconds=12
  { 2441935.50, 44.135821 }, //10 September 1973, UT1-UTC= 0.0481788, Accumulated Leap Seconds=12
  { 2441936.50, 44.138738 }, //11 September 1973, UT1-UTC= 0.0452619, Accumulated Leap Seconds=12
  { 2441937.50, 44.141793 }, //12 September 1973, UT1-UTC= 0.0422068, Accumulated Leap Seconds=12
  { 2441938.50, 44.144934 }, //13 September 1973, UT1-UTC= 0.0390659, Accumulated Leap Seconds=12
  { 2441939.50, 44.148087 }, //14 September 1973, UT1-UTC= 0.0359132, Accumulated Leap Seconds=12
  { 2441940.50, 44.151171 }, //15 September 1973, UT1-UTC= 0.0328289, Accumulated Leap Seconds=12
  { 2441941.50, 44.154121 }, //16 September 1973, UT1-UTC= 0.0298786, Accumulated Leap Seconds=12
  { 2441942.50, 44.156907 }, //17 September 1973, UT1-UTC= 0.0270925, Accumulated Leap Seconds=12
  { 2441943.50, 44.159547 }, //18 September 1973, UT1-UTC= 0.0244530, Accumulated Leap Seconds=12
  { 2441944.50, 44.162102 }, //19 September 1973, UT1-UTC= 0.0218975, Accumulated Leap Seconds=12
  { 2441945.50, 44.164665 }, //20 September 1973, UT1-UTC= 0.0193345, Accumulated Leap Seconds=12
  { 2441946.50, 44.167333 }, //21 September 1973, UT1-UTC= 0.0166675, Accumulated Leap Seconds=12
  { 2441947.50, 44.170178 }, //22 September 1973, UT1-UTC= 0.0138216, Accumulated Leap Seconds=12
  { 2441948.50, 44.173236 }, //23 September 1973, UT1-UTC= 0.0107642, Accumulated Leap Seconds=12
  { 2441949.50, 44.176485 }, //24 September 1973, UT1-UTC= 0.0075152, Accumulated Leap Seconds=12
  { 2441950.50, 44.179857 }, //25 September 1973, UT1-UTC= 0.0041426, Accumulated Leap Seconds=12
  { 2441951.50, 44.183256 }, //26 September 1973, UT1-UTC= 0.0007438, Accumulated Leap Seconds=12
  { 2441952.50, 44.186582 }, //27 September 1973, UT1-UTC=-0.0025816, Accumulated Leap Seconds=12
  { 2441953.50, 44.189758 }, //28 September 1973, UT1-UTC=-0.0057578, Accumulated Leap Seconds=12
  { 2441954.50, 44.192748 }, //29 September 1973, UT1-UTC=-0.0087481, Accumulated Leap Seconds=12
  { 2441955.50, 44.195557 }, //30 September 1973, UT1-UTC=-0.0115567, Accumulated Leap Seconds=12
  { 2441956.50, 44.198219 }, //01 October 1973, UT1-UTC=-0.0142187, Accumulated Leap Seconds=12
  { 2441957.50, 44.200788 }, //02 October 1973, UT1-UTC=-0.0167878, Accumulated Leap Seconds=12
  { 2441958.50, 44.203326 }, //03 October 1973, UT1-UTC=-0.0193259, Accumulated Leap Seconds=12
  { 2441959.50, 44.205896 }, //04 October 1973, UT1-UTC=-0.0218957, Accumulated Leap Seconds=12
  { 2441960.50, 44.208555 }, //05 October 1973, UT1-UTC=-0.0245553, Accumulated Leap Seconds=12
  { 2441961.50, 44.211352 }, //06 October 1973, UT1-UTC=-0.0273522, Accumulated Leap Seconds=12
  { 2441962.50, 44.214318 }, //07 October 1973, UT1-UTC=-0.0303185, Accumulated Leap Seconds=12
  { 2441963.50, 44.217465 }, //08 October 1973, UT1-UTC=-0.0334655, Accumulated Leap Seconds=12
  { 2441964.50, 44.220780 }, //09 October 1973, UT1-UTC=-0.0367801, Accumulated Leap Seconds=12
  { 2441965.50, 44.224223 }, //10 October 1973, UT1-UTC=-0.0402230, Accumulated Leap Seconds=12
  { 2441966.50, 44.227729 }, //11 October 1973, UT1-UTC=-0.0437289, Accumulated Leap Seconds=12
  { 2441967.50, 44.231215 }, //12 October 1973, UT1-UTC=-0.0472151, Accumulated Leap Seconds=12
  { 2441968.50, 44.234601 }, //13 October 1973, UT1-UTC=-0.0506007, Accumulated Leap Seconds=12
  { 2441969.50, 44.237831 }, //14 October 1973, UT1-UTC=-0.0538311, Accumulated Leap Seconds=12
  { 2441970.50, 44.240899 }, //15 October 1973, UT1-UTC=-0.0568993, Accumulated Leap Seconds=12
  { 2441971.50, 44.243852 }, //16 October 1973, UT1-UTC=-0.0598522, Accumulated Leap Seconds=12
  { 2441972.50, 44.246777 }, //17 October 1973, UT1-UTC=-0.0627770, Accumulated Leap Seconds=12
  { 2441973.50, 44.249772 }, //18 October 1973, UT1-UTC=-0.0657721, Accumulated Leap Seconds=12
  { 2441974.50, 44.252916 }, //19 October 1973, UT1-UTC=-0.0689162, Accumulated Leap Seconds=12
  { 2441975.50, 44.256247 }, //20 October 1973, UT1-UTC=-0.0722474, Accumulated Leap Seconds=12
  { 2441976.50, 44.259755 }, //21 October 1973, UT1-UTC=-0.0757549, Accumulated Leap Seconds=12
  { 2441977.50, 44.263385 }, //22 October 1973, UT1-UTC=-0.0793851, Accumulated Leap Seconds=12
  { 2441978.50, 44.267057 }, //23 October 1973, UT1-UTC=-0.0830574, Accumulated Leap Seconds=12
  { 2441979.50, 44.270685 }, //24 October 1973, UT1-UTC=-0.0866852, Accumulated Leap Seconds=12
  { 2441980.50, 44.274194 }, //25 October 1973, UT1-UTC=-0.0901937, Accumulated Leap Seconds=12
  { 2441981.50, 44.277536 }, //26 October 1973, UT1-UTC=-0.0935357, Accumulated Leap Seconds=12
  { 2441982.50, 44.280699 }, //27 October 1973, UT1-UTC=-0.0966987, Accumulated Leap Seconds=12
  { 2441983.50, 44.283703 }, //28 October 1973, UT1-UTC=-0.0997029, Accumulated Leap Seconds=12
  { 2441984.50, 44.286593 }, //29 October 1973, UT1-UTC=-0.1025932, Accumulated Leap Seconds=12
  { 2441985.50, 44.289428 }, //30 October 1973, UT1-UTC=-0.1054279, Accumulated Leap Seconds=12
  { 2441986.50, 44.292269 }, //31 October 1973, UT1-UTC=-0.1082691, Accumulated Leap Seconds=12
  { 2441987.50, 44.295175 }, //01 November 1973, UT1-UTC=-0.1111747, Accumulated Leap Seconds=12
  { 2441988.50, 44.298193 }, //02 November 1973, UT1-UTC=-0.1141927, Accumulated Leap Seconds=12
  { 2441989.50, 44.301357 }, //03 November 1973, UT1-UTC=-0.1173573, Accumulated Leap Seconds=12
  { 2441990.50, 44.304685 }, //04 November 1973, UT1-UTC=-0.1206852, Accumulated Leap Seconds=12
  { 2441991.50, 44.308174 }, //05 November 1973, UT1-UTC=-0.1241736, Accumulated Leap Seconds=12
  { 2441992.50, 44.311798 }, //06 November 1973, UT1-UTC=-0.1277982, Accumulated Leap Seconds=12
  { 2441993.50, 44.315511 }, //07 November 1973, UT1-UTC=-0.1315113, Accumulated Leap Seconds=12
  { 2441994.50, 44.319244 }, //08 November 1973, UT1-UTC=-0.1352436, Accumulated Leap Seconds=12
  { 2441995.50, 44.322914 }, //09 November 1973, UT1-UTC=-0.1389145, Accumulated Leap Seconds=12
  { 2441996.50, 44.326454 }, //10 November 1973, UT1-UTC=-0.1424540, Accumulated Leap Seconds=12
  { 2441997.50, 44.329827 }, //11 November 1973, UT1-UTC=-0.1458272, Accumulated Leap Seconds=12
  { 2441998.50, 44.333054 }, //12 November 1973, UT1-UTC=-0.1490537, Accumulated Leap Seconds=12
  { 2441999.50, 44.336205 }, //13 November 1973, UT1-UTC=-0.1522051, Accumulated Leap Seconds=12
  { 2442000.50, 44.339381 }, //14 November 1973, UT1-UTC=-0.1553812, Accumulated Leap Seconds=12
  { 2442001.50, 44.342672 }, //15 November 1973, UT1-UTC=-0.1586718, Accumulated Leap Seconds=12
  { 2442002.50, 44.346124 }, //16 November 1973, UT1-UTC=-0.1621245, Accumulated Leap Seconds=12
  { 2442003.50, 44.349731 }, //17 November 1973, UT1-UTC=-0.1657309, Accumulated Leap Seconds=12
  { 2442004.50, 44.353436 }, //18 November 1973, UT1-UTC=-0.1694358, Accumulated Leap Seconds=12
  { 2442005.50, 44.357158 }, //19 November 1973, UT1-UTC=-0.1731585, Accumulated Leap Seconds=12
  { 2442006.50, 44.360816 }, //20 November 1973, UT1-UTC=-0.1768161, Accumulated Leap Seconds=12
  { 2442007.50, 44.364341 }, //21 November 1973, UT1-UTC=-0.1803414, Accumulated Leap Seconds=12
  { 2442008.50, 44.367691 }, //22 November 1973, UT1-UTC=-0.1836911, Accumulated Leap Seconds=12
  { 2442009.50, 44.370850 }, //23 November 1973, UT1-UTC=-0.1868497, Accumulated Leap Seconds=12
  { 2442010.50, 44.373829 }, //24 November 1973, UT1-UTC=-0.1898289, Accumulated Leap Seconds=12
  { 2442011.50, 44.376664 }, //25 November 1973, UT1-UTC=-0.1926643, Accumulated Leap Seconds=12
  { 2442012.50, 44.379408 }, //26 November 1973, UT1-UTC=-0.1954080, Accumulated Leap Seconds=12
  { 2442013.50, 44.382120 }, //27 November 1973, UT1-UTC=-0.1981199, Accumulated Leap Seconds=12
  { 2442014.50, 44.384858 }, //28 November 1973, UT1-UTC=-0.2008582, Accumulated Leap Seconds=12
  { 2442015.50, 44.387672 }, //29 November 1973, UT1-UTC=-0.2036719, Accumulated Leap Seconds=12
  { 2442016.50, 44.390596 }, //30 November 1973, UT1-UTC=-0.2065957, Accumulated Leap Seconds=12
  { 2442017.50, 44.393647 }, //01 December 1973, UT1-UTC=-0.2096471, Accumulated Leap Seconds=12
  { 2442018.50, 44.396826 }, //02 December 1973, UT1-UTC=-0.2128261, Accumulated Leap Seconds=12
  { 2442019.50, 44.400115 }, //03 December 1973, UT1-UTC=-0.2161153, Accumulated Leap Seconds=12
  { 2442020.50, 44.403480 }, //04 December 1973, UT1-UTC=-0.2194796, Accumulated Leap Seconds=12
  { 2442021.50, 44.406867 }, //05 December 1973, UT1-UTC=-0.2228670, Accumulated Leap Seconds=12
  { 2442022.50, 44.410212 }, //06 December 1973, UT1-UTC=-0.2262122, Accumulated Leap Seconds=12
  { 2442023.50, 44.413450 }, //07 December 1973, UT1-UTC=-0.2294497, Accumulated Leap Seconds=12
  { 2442024.50, 44.416534 }, //08 December 1973, UT1-UTC=-0.2325339, Accumulated Leap Seconds=12
  { 2442025.50, 44.419462 }, //09 December 1973, UT1-UTC=-0.2354623, Accumulated Leap Seconds=12
  { 2442026.50, 44.422287 }, //10 December 1973, UT1-UTC=-0.2382872, Accumulated Leap Seconds=12
  { 2442027.50, 44.425104 }, //11 December 1973, UT1-UTC=-0.2411036, Accumulated Leap Seconds=12
  { 2442028.50, 44.428015 }, //12 December 1973, UT1-UTC=-0.2440146, Accumulated Leap Seconds=12
  { 2442029.50, 44.431091 }, //13 December 1973, UT1-UTC=-0.2470908, Accumulated Leap Seconds=12
  { 2442030.50, 44.434342 }, //14 December 1973, UT1-UTC=-0.2503419, Accumulated Leap Seconds=12
  { 2442031.50, 44.437714 }, //15 December 1973, UT1-UTC=-0.2537141, Accumulated Leap Seconds=12
  { 2442032.50, 44.441114 }, //16 December 1973, UT1-UTC=-0.2571138, Accumulated Leap Seconds=12
  { 2442033.50, 44.444442 }, //17 December 1973, UT1-UTC=-0.2604418, Accumulated Leap Seconds=12
  { 2442034.50, 44.447621 }, //18 December 1973, UT1-UTC=-0.2636212, Accumulated Leap Seconds=12
  { 2442035.50, 44.450608 }, //19 December 1973, UT1-UTC=-0.2666082, Accumulated Leap Seconds=12
  { 2442036.50, 44.453392 }, //20 December 1973, UT1-UTC=-0.2693916, Accumulated Leap Seconds=12
  { 2442037.50, 44.455986 }, //21 December 1973, UT1-UTC=-0.2719860, Accumulated Leap Seconds=12
  { 2442038.50, 44.458426 }, //22 December 1973, UT1-UTC=-0.2744256, Accumulated Leap Seconds=12
  { 2442039.50, 44.460759 }, //23 December 1973, UT1-UTC=-0.2767591, Accumulated Leap Seconds=12
  { 2442040.50, 44.463044 }, //24 December 1973, UT1-UTC=-0.2790441, Accumulated Leap Seconds=12
  { 2442041.50, 44.465340 }, //25 December 1973, UT1-UTC=-0.2813401, Accumulated Leap Seconds=12
  { 2442042.50, 44.467699 }, //26 December 1973, UT1-UTC=-0.2836993, Accumulated Leap Seconds=12
  { 2442043.50, 44.470160 }, //27 December 1973, UT1-UTC=-0.2861602, Accumulated Leap Seconds=12
  { 2442044.50, 44.472743 }, //28 December 1973, UT1-UTC=-0.2887427, Accumulated Leap Seconds=12
  { 2442045.50, 44.475448 }, //29 December 1973, UT1-UTC=-0.2914476, Accumulated Leap Seconds=12
  { 2442046.50, 44.478258 }, //30 December 1973, UT1-UTC=-0.2942581, Accumulated Leap Seconds=12
  { 2442047.50, 44.481142 }, //31 December 1973, UT1-UTC=-0.2971424, Accumulated Leap Seconds=12
  { 2442048.50, 44.484056 }, //01 January 1974, UT1-UTC= 0.6999438, Accumulated Leap Seconds=13
  { 2442049.50, 44.486946 }, //02 January 1974, UT1-UTC= 0.6970539, Accumulated Leap Seconds=13
  { 2442050.50, 44.489756 }, //03 January 1974, UT1-UTC= 0.6942436, Accumulated Leap Seconds=13
  { 2442051.50, 44.492442 }, //04 January 1974, UT1-UTC= 0.6915576, Accumulated Leap Seconds=13
  { 2442052.50, 44.494988 }, //05 January 1974, UT1-UTC= 0.6890115, Accumulated Leap Seconds=13
  { 2442053.50, 44.497424 }, //06 January 1974, UT1-UTC= 0.6865758, Accumulated Leap Seconds=13
  { 2442054.50, 44.499826 }, //07 January 1974, UT1-UTC= 0.6841739, Accumulated Leap Seconds=13
  { 2442055.50, 44.502299 }, //08 January 1974, UT1-UTC= 0.6817011, Accumulated Leap Seconds=13
  { 2442056.50, 44.504937 }, //09 January 1974, UT1-UTC= 0.6790627, Accumulated Leap Seconds=13
  { 2442057.50, 44.507787 }, //10 January 1974, UT1-UTC= 0.6762134, Accumulated Leap Seconds=13
  { 2442058.50, 44.510821 }, //11 January 1974, UT1-UTC= 0.6731788, Accumulated Leap Seconds=13
  { 2442059.50, 44.513953 }, //12 January 1974, UT1-UTC= 0.6700467, Accumulated Leap Seconds=13
  { 2442060.50, 44.517066 }, //13 January 1974, UT1-UTC= 0.6669336, Accumulated Leap Seconds=13
  { 2442061.50, 44.520055 }, //14 January 1974, UT1-UTC= 0.6639450, Accumulated Leap Seconds=13
  { 2442062.50, 44.522853 }, //15 January 1974, UT1-UTC= 0.6611472, Accumulated Leap Seconds=13
  { 2442063.50, 44.525439 }, //16 January 1974, UT1-UTC= 0.6585612, Accumulated Leap Seconds=13
  { 2442064.50, 44.527828 }, //17 January 1974, UT1-UTC= 0.6561715, Accumulated Leap Seconds=13
  { 2442065.50, 44.530060 }, //18 January 1974, UT1-UTC= 0.6539399, Accumulated Leap Seconds=13
  { 2442066.50, 44.532184 }, //19 January 1974, UT1-UTC= 0.6518159, Accumulated Leap Seconds=13
  { 2442067.50, 44.534258 }, //20 January 1974, UT1-UTC= 0.6497421, Accumulated Leap Seconds=13
  { 2442068.50, 44.536341 }, //21 January 1974, UT1-UTC= 0.6476585, Accumulated Leap Seconds=13
  { 2442069.50, 44.538491 }, //22 January 1974, UT1-UTC= 0.6455090, Accumulated Leap Seconds=13
  { 2442070.50, 44.540751 }, //23 January 1974, UT1-UTC= 0.6432493, Accumulated Leap Seconds=13
  { 2442071.50, 44.543147 }, //24 January 1974, UT1-UTC= 0.6408530, Accumulated Leap Seconds=13
  { 2442072.50, 44.545685 }, //25 January 1974, UT1-UTC= 0.6383150, Accumulated Leap Seconds=13
  { 2442073.50, 44.548350 }, //26 January 1974, UT1-UTC= 0.6356505, Accumulated Leap Seconds=13
  { 2442074.50, 44.551108 }, //27 January 1974, UT1-UTC= 0.6328924, Accumulated Leap Seconds=13
  { 2442075.50, 44.553914 }, //28 January 1974, UT1-UTC= 0.6300862, Accumulated Leap Seconds=13
  { 2442076.50, 44.556715 }, //29 January 1974, UT1-UTC= 0.6272847, Accumulated Leap Seconds=13
  { 2442077.50, 44.559460 }, //30 January 1974, UT1-UTC= 0.6245402, Accumulated Leap Seconds=13
  { 2442078.50, 44.562105 }, //31 January 1974, UT1-UTC= 0.6218946, Accumulated Leap Seconds=13
  { 2442079.50, 44.564633 }, //01 February 1974, UT1-UTC= 0.6193665, Accumulated Leap Seconds=13
  { 2442080.50, 44.567061 }, //02 February 1974, UT1-UTC= 0.6169393, Accumulated Leap Seconds=13
  { 2442081.50, 44.569446 }, //03 February 1974, UT1-UTC= 0.6145544, Accumulated Leap Seconds=13
  { 2442082.50, 44.571881 }, //04 February 1974, UT1-UTC= 0.6121192, Accumulated Leap Seconds=13
  { 2442083.50, 44.574468 }, //05 February 1974, UT1-UTC= 0.6095322, Accumulated Leap Seconds=13
  { 2442084.50, 44.577281 }, //06 February 1974, UT1-UTC= 0.6067194, Accumulated Leap Seconds=13
  { 2442085.50, 44.580334 }, //07 February 1974, UT1-UTC= 0.6036662, Accumulated Leap Seconds=13
  { 2442086.50, 44.583571 }, //08 February 1974, UT1-UTC= 0.6004295, Accumulated Leap Seconds=13
  { 2442087.50, 44.586880 }, //09 February 1974, UT1-UTC= 0.5971202, Accumulated Leap Seconds=13
  { 2442088.50, 44.590136 }, //10 February 1974, UT1-UTC= 0.5938641, Accumulated Leap Seconds=13
  { 2442089.50, 44.593237 }, //11 February 1974, UT1-UTC= 0.5907626, Accumulated Leap Seconds=13
  { 2442090.50, 44.596132 }, //12 February 1974, UT1-UTC= 0.5878679, Accumulated Leap Seconds=13
  { 2442091.50, 44.598818 }, //13 February 1974, UT1-UTC= 0.5851819, Accumulated Leap Seconds=13
  { 2442092.50, 44.601330 }, //14 February 1974, UT1-UTC= 0.5826700, Accumulated Leap Seconds=13
  { 2442093.50, 44.603720 }, //15 February 1974, UT1-UTC= 0.5802797, Accumulated Leap Seconds=13
  { 2442094.50, 44.606049 }, //16 February 1974, UT1-UTC= 0.5779513, Accumulated Leap Seconds=13
  { 2442095.50, 44.608376 }, //17 February 1974, UT1-UTC= 0.5756244, Accumulated Leap Seconds=13
  { 2442096.50, 44.610759 }, //18 February 1974, UT1-UTC= 0.5732413, Accumulated Leap Seconds=13
  { 2442097.50, 44.613247 }, //19 February 1974, UT1-UTC= 0.5707527, Accumulated Leap Seconds=13
  { 2442098.50, 44.615876 }, //20 February 1974, UT1-UTC= 0.5681244, Accumulated Leap Seconds=13
  { 2442099.50, 44.618657 }, //21 February 1974, UT1-UTC= 0.5653429, Accumulated Leap Seconds=13
  { 2442100.50, 44.621583 }, //22 February 1974, UT1-UTC= 0.5624169, Accumulated Leap Seconds=13
  { 2442101.50, 44.624624 }, //23 February 1974, UT1-UTC= 0.5593762, Accumulated Leap Seconds=13
  { 2442102.50, 44.627733 }, //24 February 1974, UT1-UTC= 0.5562673, Accumulated Leap Seconds=13
  { 2442103.50, 44.630853 }, //25 February 1974, UT1-UTC= 0.5531474, Accumulated Leap Seconds=13
  { 2442104.50, 44.633925 }, //26 February 1974, UT1-UTC= 0.5500745, Accumulated Leap Seconds=13
  { 2442105.50, 44.636903 }, //27 February 1974, UT1-UTC= 0.5470965, Accumulated Leap Seconds=13
  { 2442106.50, 44.639762 }, //28 February 1974, UT1-UTC= 0.5442380, Accumulated Leap Seconds=13
  { 2442107.50, 44.642510 }, //01 March 1974, UT1-UTC= 0.5414901, Accumulated Leap Seconds=13
  { 2442108.50, 44.645193 }, //02 March 1974, UT1-UTC= 0.5388066, Accumulated Leap Seconds=13
  { 2442109.50, 44.647891 }, //03 March 1974, UT1-UTC= 0.5361088, Accumulated Leap Seconds=13
  { 2442110.50, 44.650698 }, //04 March 1974, UT1-UTC= 0.5333018, Accumulated Leap Seconds=13
  { 2442111.50, 44.653698 }, //05 March 1974, UT1-UTC= 0.5303016, Accumulated Leap Seconds=13
  { 2442112.50, 44.656936 }, //06 March 1974, UT1-UTC= 0.5270642, Accumulated Leap Seconds=13
  { 2442113.50, 44.660393 }, //07 March 1974, UT1-UTC= 0.5236066, Accumulated Leap Seconds=13
  { 2442114.50, 44.663993 }, //08 March 1974, UT1-UTC= 0.5200074, Accumulated Leap Seconds=13
  { 2442115.50, 44.667616 }, //09 March 1974, UT1-UTC= 0.5163838, Accumulated Leap Seconds=13
  { 2442116.50, 44.671145 }, //10 March 1974, UT1-UTC= 0.5128546, Accumulated Leap Seconds=13
  { 2442117.50, 44.674495 }, //11 March 1974, UT1-UTC= 0.5095052, Accumulated Leap Seconds=13
  { 2442118.50, 44.677631 }, //12 March 1974, UT1-UTC= 0.5063690, Accumulated Leap Seconds=13
  { 2442119.50, 44.680569 }, //13 March 1974, UT1-UTC= 0.5034312, Accumulated Leap Seconds=13
  { 2442120.50, 44.683355 }, //14 March 1974, UT1-UTC= 0.5006447, Accumulated Leap Seconds=13
  { 2442121.50, 44.686051 }, //15 March 1974, UT1-UTC= 0.4979486, Accumulated Leap Seconds=13
  { 2442122.50, 44.688720 }, //16 March 1974, UT1-UTC= 0.4952804, Accumulated Leap Seconds=13
  { 2442123.50, 44.691419 }, //17 March 1974, UT1-UTC= 0.4925811, Accumulated Leap Seconds=13
  { 2442124.50, 44.694201 }, //18 March 1974, UT1-UTC= 0.4897991, Accumulated Leap Seconds=13
  { 2442125.50, 44.697105 }, //19 March 1974, UT1-UTC= 0.4868946, Accumulated Leap Seconds=13
  { 2442126.50, 44.700155 }, //20 March 1974, UT1-UTC= 0.4838452, Accumulated Leap Seconds=13
  { 2442127.50, 44.703350 }, //21 March 1974, UT1-UTC= 0.4806495, Accumulated Leap Seconds=13
  { 2442128.50, 44.706672 }, //22 March 1974, UT1-UTC= 0.4773279, Accumulated Leap Seconds=13
  { 2442129.50, 44.710079 }, //23 March 1974, UT1-UTC= 0.4739205, Accumulated Leap Seconds=13
  { 2442130.50, 44.713517 }, //24 March 1974, UT1-UTC= 0.4704827, Accumulated Leap Seconds=13
  { 2442131.50, 44.716923 }, //25 March 1974, UT1-UTC= 0.4670769, Accumulated Leap Seconds=13
  { 2442132.50, 44.720239 }, //26 March 1974, UT1-UTC= 0.4637605, Accumulated Leap Seconds=13
  { 2442133.50, 44.723430 }, //27 March 1974, UT1-UTC= 0.4605702, Accumulated Leap Seconds=13
  { 2442134.50, 44.726491 }, //28 March 1974, UT1-UTC= 0.4575087, Accumulated Leap Seconds=13
  { 2442135.50, 44.729460 }, //29 March 1974, UT1-UTC= 0.4545396, Accumulated Leap Seconds=13
  { 2442136.50, 44.732407 }, //30 March 1974, UT1-UTC= 0.4515928, Accumulated Leap Seconds=13
  { 2442137.50, 44.735419 }, //31 March 1974, UT1-UTC= 0.4485812, Accumulated Leap Seconds=13
  { 2442138.50, 44.738577 }, //01 April 1974, UT1-UTC= 0.4454233, Accumulated Leap Seconds=13
  { 2442139.50, 44.741933 }, //02 April 1974, UT1-UTC= 0.4420667, Accumulated Leap Seconds=13
  { 2442140.50, 44.745495 }, //03 April 1974, UT1-UTC= 0.4385054, Accumulated Leap Seconds=13
  { 2442141.50, 44.749214 }, //04 April 1974, UT1-UTC= 0.4347858, Accumulated Leap Seconds=13
  { 2442142.50, 44.753003 }, //05 April 1974, UT1-UTC= 0.4309973, Accumulated Leap Seconds=13
  { 2442143.50, 44.756752 }, //06 April 1974, UT1-UTC= 0.4272484, Accumulated Leap Seconds=13
  { 2442144.50, 44.760364 }, //07 April 1974, UT1-UTC= 0.4236357, Accumulated Leap Seconds=13
  { 2442145.50, 44.763781 }, //08 April 1974, UT1-UTC= 0.4202190, Accumulated Leap Seconds=13
  { 2442146.50, 44.766990 }, //09 April 1974, UT1-UTC= 0.4170100, Accumulated Leap Seconds=13
  { 2442147.50, 44.770021 }, //10 April 1974, UT1-UTC= 0.4139793, Accumulated Leap Seconds=13
  { 2442148.50, 44.772928 }, //11 April 1974, UT1-UTC= 0.4110721, Accumulated Leap Seconds=13
  { 2442149.50, 44.775775 }, //12 April 1974, UT1-UTC= 0.4082249, Accumulated Leap Seconds=13
  { 2442150.50, 44.778623 }, //13 April 1974, UT1-UTC= 0.4053770, Accumulated Leap Seconds=13
  { 2442151.50, 44.781524 }, //14 April 1974, UT1-UTC= 0.4024755, Accumulated Leap Seconds=13
  { 2442152.50, 44.784521 }, //15 April 1974, UT1-UTC= 0.3994786, Accumulated Leap Seconds=13
  { 2442153.50, 44.787641 }, //16 April 1974, UT1-UTC= 0.3963591, Accumulated Leap Seconds=13
  { 2442154.50, 44.790892 }, //17 April 1974, UT1-UTC= 0.3931082, Accumulated Leap Seconds=13
  { 2442155.50, 44.794264 }, //18 April 1974, UT1-UTC= 0.3897363, Accumulated Leap Seconds=13
  { 2442156.50, 44.797727 }, //19 April 1974, UT1-UTC= 0.3862732, Accumulated Leap Seconds=13
  { 2442157.50, 44.801234 }, //20 April 1974, UT1-UTC= 0.3827656, Accumulated Leap Seconds=13
  { 2442158.50, 44.804728 }, //21 April 1974, UT1-UTC= 0.3792718, Accumulated Leap Seconds=13
  { 2442159.50, 44.808147 }, //22 April 1974, UT1-UTC= 0.3758526, Accumulated Leap Seconds=13
  { 2442160.50, 44.811443 }, //23 April 1974, UT1-UTC= 0.3725566, Accumulated Leap Seconds=13
  { 2442161.50, 44.814597 }, //24 April 1974, UT1-UTC= 0.3694030, Accumulated Leap Seconds=13
  { 2442162.50, 44.817630 }, //25 April 1974, UT1-UTC= 0.3663697, Accumulated Leap Seconds=13
  { 2442163.50, 44.820605 }, //26 April 1974, UT1-UTC= 0.3633949, Accumulated Leap Seconds=13
  { 2442164.50, 44.823606 }, //27 April 1974, UT1-UTC= 0.3603935, Accumulated Leap Seconds=13
  { 2442165.50, 44.826716 }, //28 April 1974, UT1-UTC= 0.3572837, Accumulated Leap Seconds=13
  { 2442166.50, 44.829988 }, //29 April 1974, UT1-UTC= 0.3540118, Accumulated Leap Seconds=13
  { 2442167.50, 44.833433 }, //30 April 1974, UT1-UTC= 0.3505675, Accumulated Leap Seconds=13
  { 2442168.50, 44.837013 }, //01 May 1974, UT1-UTC= 0.3469865, Accumulated Leap Seconds=13
  { 2442169.50, 44.840660 }, //02 May 1974, UT1-UTC= 0.3433401, Accumulated Leap Seconds=13
  { 2442170.50, 44.844282 }, //03 May 1974, UT1-UTC= 0.3397183, Accumulated Leap Seconds=13
  { 2442171.50, 44.847791 }, //04 May 1974, UT1-UTC= 0.3362085, Accumulated Leap Seconds=13
  { 2442172.50, 44.851125 }, //05 May 1974, UT1-UTC= 0.3328749, Accumulated Leap Seconds=13
  { 2442173.50, 44.854254 }, //06 May 1974, UT1-UTC= 0.3297456, Accumulated Leap Seconds=13
  { 2442174.50, 44.857190 }, //07 May 1974, UT1-UTC= 0.3268099, Accumulated Leap Seconds=13
  { 2442175.50, 44.859974 }, //08 May 1974, UT1-UTC= 0.3240255, Accumulated Leap Seconds=13
  { 2442176.50, 44.862666 }, //09 May 1974, UT1-UTC= 0.3213335, Accumulated Leap Seconds=13
  { 2442177.50, 44.865328 }, //10 May 1974, UT1-UTC= 0.3186717, Accumulated Leap Seconds=13
  { 2442178.50, 44.868015 }, //11 May 1974, UT1-UTC= 0.3159848, Accumulated Leap Seconds=13
  { 2442179.50, 44.870771 }, //12 May 1974, UT1-UTC= 0.3132294, Accumulated Leap Seconds=13
  { 2442180.50, 44.873623 }, //13 May 1974, UT1-UTC= 0.3103768, Accumulated Leap Seconds=13
  { 2442181.50, 44.876585 }, //14 May 1974, UT1-UTC= 0.3074146, Accumulated Leap Seconds=13
  { 2442182.50, 44.879652 }, //15 May 1974, UT1-UTC= 0.3043480, Accumulated Leap Seconds=13
  { 2442183.50, 44.882801 }, //16 May 1974, UT1-UTC= 0.3011990, Accumulated Leap Seconds=13
  { 2442184.50, 44.885996 }, //17 May 1974, UT1-UTC= 0.2980043, Accumulated Leap Seconds=13
  { 2442185.50, 44.889187 }, //18 May 1974, UT1-UTC= 0.2948130, Accumulated Leap Seconds=13
  { 2442186.50, 44.892319 }, //19 May 1974, UT1-UTC= 0.2916807, Accumulated Leap Seconds=13
  { 2442187.50, 44.895341 }, //20 May 1974, UT1-UTC= 0.2886593, Accumulated Leap Seconds=13
  { 2442188.50, 44.898219 }, //21 May 1974, UT1-UTC= 0.2857812, Accumulated Leap Seconds=13
  { 2442189.50, 44.900957 }, //22 May 1974, UT1-UTC= 0.2830427, Accumulated Leap Seconds=13
  { 2442190.50, 44.903604 }, //23 May 1974, UT1-UTC= 0.2803960, Accumulated Leap Seconds=13
  { 2442191.50, 44.906240 }, //24 May 1974, UT1-UTC= 0.2777595, Accumulated Leap Seconds=13
  { 2442192.50, 44.908956 }, //25 May 1974, UT1-UTC= 0.2750436, Accumulated Leap Seconds=13
  { 2442193.50, 44.911816 }, //26 May 1974, UT1-UTC= 0.2721837, Accumulated Leap Seconds=13
  { 2442194.50, 44.914837 }, //27 May 1974, UT1-UTC= 0.2691634, Accumulated Leap Seconds=13
  { 2442195.50, 44.917981 }, //28 May 1974, UT1-UTC= 0.2660190, Accumulated Leap Seconds=13
  { 2442196.50, 44.921175 }, //29 May 1974, UT1-UTC= 0.2628251, Accumulated Leap Seconds=13
  { 2442197.50, 44.924329 }, //30 May 1974, UT1-UTC= 0.2596711, Accumulated Leap Seconds=13
  { 2442198.50, 44.927361 }, //31 May 1974, UT1-UTC= 0.2566387, Accumulated Leap Seconds=13
  { 2442199.50, 44.930214 }, //01 June 1974, UT1-UTC= 0.2537862, Accumulated Leap Seconds=13
  { 2442200.50, 44.932859 }, //02 June 1974, UT1-UTC= 0.2511410, Accumulated Leap Seconds=13
  { 2442201.50, 44.935301 }, //03 June 1974, UT1-UTC= 0.2486987, Accumulated Leap Seconds=13
  { 2442202.50, 44.937573 }, //04 June 1974, UT1-UTC= 0.2464268, Accumulated Leap Seconds=13
  { 2442203.50, 44.939727 }, //05 June 1974, UT1-UTC= 0.2442733, Accumulated Leap Seconds=13
  { 2442204.50, 44.941822 }, //06 June 1974, UT1-UTC= 0.2421779, Accumulated Leap Seconds=13
  { 2442205.50, 44.943917 }, //07 June 1974, UT1-UTC= 0.2400825, Accumulated Leap Seconds=13
  { 2442206.50, 44.946060 }, //08 June 1974, UT1-UTC= 0.2379399, Accumulated Leap Seconds=13
  { 2442207.50, 44.948282 }, //09 June 1974, UT1-UTC= 0.2357183, Accumulated Leap Seconds=13
  { 2442208.50, 44.950597 }, //10 June 1974, UT1-UTC= 0.2334032, Accumulated Leap Seconds=13
  { 2442209.50, 44.953002 }, //11 June 1974, UT1-UTC= 0.2309975, Accumulated Leap Seconds=13
  { 2442210.50, 44.955480 }, //12 June 1974, UT1-UTC= 0.2285200, Accumulated Leap Seconds=13
  { 2442211.50, 44.957998 }, //13 June 1974, UT1-UTC= 0.2260020, Accumulated Leap Seconds=13
  { 2442212.50, 44.960516 }, //14 June 1974, UT1-UTC= 0.2234839, Accumulated Leap Seconds=13
  { 2442213.50, 44.962988 }, //15 June 1974, UT1-UTC= 0.2210118, Accumulated Leap Seconds=13
  { 2442214.50, 44.965368 }, //16 June 1974, UT1-UTC= 0.2186316, Accumulated Leap Seconds=13
  { 2442215.50, 44.967621 }, //17 June 1974, UT1-UTC= 0.2163787, Accumulated Leap Seconds=13
  { 2442216.50, 44.969738 }, //18 June 1974, UT1-UTC= 0.2142621, Accumulated Leap Seconds=13
  { 2442217.50, 44.971749 }, //19 June 1974, UT1-UTC= 0.2122513, Accumulated Leap Seconds=13
  { 2442218.50, 44.973725 }, //20 June 1974, UT1-UTC= 0.2102747, Accumulated Leap Seconds=13
  { 2442219.50, 44.975763 }, //21 June 1974, UT1-UTC= 0.2082374, Accumulated Leap Seconds=13
  { 2442220.50, 44.977946 }, //22 June 1974, UT1-UTC= 0.2060541, Accumulated Leap Seconds=13
  { 2442221.50, 44.980315 }, //23 June 1974, UT1-UTC= 0.2036846, Accumulated Leap Seconds=13
  { 2442222.50, 44.982848 }, //24 June 1974, UT1-UTC= 0.2011517, Accumulated Leap Seconds=13
  { 2442223.50, 44.985467 }, //25 June 1974, UT1-UTC= 0.1985326, Accumulated Leap Seconds=13
  { 2442224.50, 44.988069 }, //26 June 1974, UT1-UTC= 0.1959314, Accumulated Leap Seconds=13
  { 2442225.50, 44.990555 }, //27 June 1974, UT1-UTC= 0.1934446, Accumulated Leap Seconds=13
  { 2442226.50, 44.992863 }, //28 June 1974, UT1-UTC= 0.1911372, Accumulated Leap Seconds=13
  { 2442227.50, 44.994965 }, //29 June 1974, UT1-UTC= 0.1890351, Accumulated Leap Seconds=13
  { 2442228.50, 44.996871 }, //30 June 1974, UT1-UTC= 0.1871290, Accumulated Leap Seconds=13
  { 2442229.50, 44.998615 }, //01 July 1974, UT1-UTC= 0.1853854, Accumulated Leap Seconds=13
  { 2442230.50, 45.000246 }, //02 July 1974, UT1-UTC= 0.1837541, Accumulated Leap Seconds=13
  { 2442231.50, 45.001822 }, //03 July 1974, UT1-UTC= 0.1821775, Accumulated Leap Seconds=13
  { 2442232.50, 45.003403 }, //04 July 1974, UT1-UTC= 0.1805972, Accumulated Leap Seconds=13
  { 2442233.50, 45.005038 }, //05 July 1974, UT1-UTC= 0.1789623, Accumulated Leap Seconds=13
  { 2442234.50, 45.006764 }, //06 July 1974, UT1-UTC= 0.1772358, Accumulated Leap Seconds=13
  { 2442235.50, 45.008601 }, //07 July 1974, UT1-UTC= 0.1753992, Accumulated Leap Seconds=13
  { 2442236.50, 45.010547 }, //08 July 1974, UT1-UTC= 0.1734528, Accumulated Leap Seconds=13
  { 2442237.50, 45.012586 }, //09 July 1974, UT1-UTC= 0.1714139, Accumulated Leap Seconds=13
  { 2442238.50, 45.014687 }, //10 July 1974, UT1-UTC= 0.1693126, Accumulated Leap Seconds=13
  { 2442239.50, 45.016812 }, //11 July 1974, UT1-UTC= 0.1671878, Accumulated Leap Seconds=13
  { 2442240.50, 45.018918 }, //12 July 1974, UT1-UTC= 0.1650821, Accumulated Leap Seconds=13
  { 2442241.50, 45.020963 }, //13 July 1974, UT1-UTC= 0.1630370, Accumulated Leap Seconds=13
  { 2442242.50, 45.022914 }, //14 July 1974, UT1-UTC= 0.1610859, Accumulated Leap Seconds=13
  { 2442243.50, 45.024756 }, //15 July 1974, UT1-UTC= 0.1592442, Accumulated Leap Seconds=13
  { 2442244.50, 45.026504 }, //16 July 1974, UT1-UTC= 0.1574964, Accumulated Leap Seconds=13
  { 2442245.50, 45.028212 }, //17 July 1974, UT1-UTC= 0.1557884, Accumulated Leap Seconds=13
  { 2442246.50, 45.029967 }, //18 July 1974, UT1-UTC= 0.1540328, Accumulated Leap Seconds=13
  { 2442247.50, 45.031867 }, //19 July 1974, UT1-UTC= 0.1521334, Accumulated Leap Seconds=13
  { 2442248.50, 45.033979 }, //20 July 1974, UT1-UTC= 0.1500210, Accumulated Leap Seconds=13
  { 2442249.50, 45.036314 }, //21 July 1974, UT1-UTC= 0.1476858, Accumulated Leap Seconds=13
  { 2442250.50, 45.038813 }, //22 July 1974, UT1-UTC= 0.1451868, Accumulated Leap Seconds=13
  { 2442251.50, 45.041366 }, //23 July 1974, UT1-UTC= 0.1426340, Accumulated Leap Seconds=13
  { 2442252.50, 45.043851 }, //24 July 1974, UT1-UTC= 0.1401491, Accumulated Leap Seconds=13
  { 2442253.50, 45.046173 }, //25 July 1974, UT1-UTC= 0.1378273, Accumulated Leap Seconds=13
  { 2442254.50, 45.048286 }, //26 July 1974, UT1-UTC= 0.1357144, Accumulated Leap Seconds=13
  { 2442255.50, 45.050193 }, //27 July 1974, UT1-UTC= 0.1338074, Accumulated Leap Seconds=13
  { 2442256.50, 45.051931 }, //28 July 1974, UT1-UTC= 0.1320688, Accumulated Leap Seconds=13
  { 2442257.50, 45.053557 }, //29 July 1974, UT1-UTC= 0.1304431, Accumulated Leap Seconds=13
  { 2442258.50, 45.055130 }, //30 July 1974, UT1-UTC= 0.1288695, Accumulated Leap Seconds=13
  { 2442259.50, 45.056712 }, //31 July 1974, UT1-UTC= 0.1272882, Accumulated Leap Seconds=13
  { 2442260.50, 45.058355 }, //01 August 1974, UT1-UTC= 0.1256451, Accumulated Leap Seconds=13
  { 2442261.50, 45.060102 }, //02 August 1974, UT1-UTC= 0.1238975, Accumulated Leap Seconds=13
  { 2442262.50, 45.061981 }, //03 August 1974, UT1-UTC= 0.1220194, Accumulated Leap Seconds=13
  { 2442263.50, 45.063995 }, //04 August 1974, UT1-UTC= 0.1200048, Accumulated Leap Seconds=13
  { 2442264.50, 45.066133 }, //05 August 1974, UT1-UTC= 0.1178673, Accumulated Leap Seconds=13
  { 2442265.50, 45.068363 }, //06 August 1974, UT1-UTC= 0.1156365, Accumulated Leap Seconds=13
  { 2442266.50, 45.070648 }, //07 August 1974, UT1-UTC= 0.1133517, Accumulated Leap Seconds=13
  { 2442267.50, 45.072943 }, //08 August 1974, UT1-UTC= 0.1110567, Accumulated Leap Seconds=13
  { 2442268.50, 45.075207 }, //09 August 1974, UT1-UTC= 0.1087935, Accumulated Leap Seconds=13
  { 2442269.50, 45.077404 }, //10 August 1974, UT1-UTC= 0.1065964, Accumulated Leap Seconds=13
  { 2442270.50, 45.079516 }, //11 August 1974, UT1-UTC= 0.1044843, Accumulated Leap Seconds=13
  { 2442271.50, 45.081549 }, //12 August 1974, UT1-UTC= 0.1024513, Accumulated Leap Seconds=13
  { 2442272.50, 45.083540 }, //13 August 1974, UT1-UTC= 0.1004595, Accumulated Leap Seconds=13
  { 2442273.50, 45.085562 }, //14 August 1974, UT1-UTC= 0.0984382, Accumulated Leap Seconds=13
  { 2442274.50, 45.087703 }, //15 August 1974, UT1-UTC= 0.0962965, Accumulated Leap Seconds=13
  { 2442275.50, 45.090050 }, //16 August 1974, UT1-UTC= 0.0939503, Accumulated Leap Seconds=13
  { 2442276.50, 45.092644 }, //17 August 1974, UT1-UTC= 0.0913562, Accumulated Leap Seconds=13
  { 2442277.50, 45.095464 }, //18 August 1974, UT1-UTC= 0.0885363, Accumulated Leap Seconds=13
  { 2442278.50, 45.098421 }, //19 August 1974, UT1-UTC= 0.0855791, Accumulated Leap Seconds=13
  { 2442279.50, 45.101387 }, //20 August 1974, UT1-UTC= 0.0826134, Accumulated Leap Seconds=13
  { 2442280.50, 45.104235 }, //21 August 1974, UT1-UTC= 0.0797654, Accumulated Leap Seconds=13
  { 2442281.50, 45.106879 }, //22 August 1974, UT1-UTC= 0.0771206, Accumulated Leap Seconds=13
  { 2442282.50, 45.109294 }, //23 August 1974, UT1-UTC= 0.0747058, Accumulated Leap Seconds=13
  { 2442283.50, 45.111503 }, //24 August 1974, UT1-UTC= 0.0724966, Accumulated Leap Seconds=13
  { 2442284.50, 45.113563 }, //25 August 1974, UT1-UTC= 0.0704374, Accumulated Leap Seconds=13
  { 2442285.50, 45.115537 }, //26 August 1974, UT1-UTC= 0.0684627, Accumulated Leap Seconds=13
  { 2442286.50, 45.117490 }, //27 August 1974, UT1-UTC= 0.0665101, Accumulated Leap Seconds=13
  { 2442287.50, 45.119475 }, //28 August 1974, UT1-UTC= 0.0645249, Accumulated Leap Seconds=13
  { 2442288.50, 45.121537 }, //29 August 1974, UT1-UTC= 0.0624629, Accumulated Leap Seconds=13
  { 2442289.50, 45.123706 }, //30 August 1974, UT1-UTC= 0.0602943, Accumulated Leap Seconds=13
  { 2442290.50, 45.125992 }, //31 August 1974, UT1-UTC= 0.0580075, Accumulated Leap Seconds=13
  { 2442291.50, 45.128391 }, //01 September 1974, UT1-UTC= 0.0556089, Accumulated Leap Seconds=13
  { 2442292.50, 45.130877 }, //02 September 1974, UT1-UTC= 0.0531226, Accumulated Leap Seconds=13
  { 2442293.50, 45.133415 }, //03 September 1974, UT1-UTC= 0.0505851, Accumulated Leap Seconds=13
  { 2442294.50, 45.135960 }, //04 September 1974, UT1-UTC= 0.0480399, Accumulated Leap Seconds=13
  { 2442295.50, 45.138470 }, //05 September 1974, UT1-UTC= 0.0455299, Accumulated Leap Seconds=13
  { 2442296.50, 45.140909 }, //06 September 1974, UT1-UTC= 0.0430908, Accumulated Leap Seconds=13
  { 2442297.50, 45.143257 }, //07 September 1974, UT1-UTC= 0.0407426, Accumulated Leap Seconds=13
  { 2442298.50, 45.145517 }, //08 September 1974, UT1-UTC= 0.0384828, Accumulated Leap Seconds=13
  { 2442299.50, 45.147720 }, //09 September 1974, UT1-UTC= 0.0362799, Accumulated Leap Seconds=13
  { 2442300.50, 45.149926 }, //10 September 1974, UT1-UTC= 0.0340742, Accumulated Leap Seconds=13
  { 2442301.50, 45.152215 }, //11 September 1974, UT1-UTC= 0.0317851, Accumulated Leap Seconds=13
  { 2442302.50, 45.154672 }, //12 September 1974, UT1-UTC= 0.0293283, Accumulated Leap Seconds=13
  { 2442303.50, 45.157359 }, //13 September 1974, UT1-UTC= 0.0266415, Accumulated Leap Seconds=13
  { 2442304.50, 45.160289 }, //14 September 1974, UT1-UTC= 0.0237106, Accumulated Leap Seconds=13
  { 2442305.50, 45.163416 }, //15 September 1974, UT1-UTC= 0.0205839, Accumulated Leap Seconds=13
  { 2442306.50, 45.166635 }, //16 September 1974, UT1-UTC= 0.0173645, Accumulated Leap Seconds=13
  { 2442307.50, 45.169820 }, //17 September 1974, UT1-UTC= 0.0141804, Accumulated Leap Seconds=13
  { 2442308.50, 45.172856 }, //18 September 1974, UT1-UTC= 0.0111442, Accumulated Leap Seconds=13
  { 2442309.50, 45.175679 }, //19 September 1974, UT1-UTC= 0.0083205, Accumulated Leap Seconds=13
  { 2442310.50, 45.178285 }, //20 September 1974, UT1-UTC= 0.0057149, Accumulated Leap Seconds=13
  { 2442311.50, 45.180715 }, //21 September 1974, UT1-UTC= 0.0032855, Accumulated Leap Seconds=13
  { 2442312.50, 45.183034 }, //22 September 1974, UT1-UTC= 0.0009657, Accumulated Leap Seconds=13
  { 2442313.50, 45.185314 }, //23 September 1974, UT1-UTC=-0.0013138, Accumulated Leap Seconds=13
  { 2442314.50, 45.187614 }, //24 September 1974, UT1-UTC=-0.0036138, Accumulated Leap Seconds=13
  { 2442315.50, 45.189982 }, //25 September 1974, UT1-UTC=-0.0059825, Accumulated Leap Seconds=13
  { 2442316.50, 45.192454 }, //26 September 1974, UT1-UTC=-0.0084542, Accumulated Leap Seconds=13
  { 2442317.50, 45.195047 }, //27 September 1974, UT1-UTC=-0.0110472, Accumulated Leap Seconds=13
  { 2442318.50, 45.197762 }, //28 September 1974, UT1-UTC=-0.0137616, Accumulated Leap Seconds=13
  { 2442319.50, 45.200580 }, //29 September 1974, UT1-UTC=-0.0165796, Accumulated Leap Seconds=13
  { 2442320.50, 45.203468 }, //30 September 1974, UT1-UTC=-0.0194678, Accumulated Leap Seconds=13
  { 2442321.50, 45.206383 }, //01 October 1974, UT1-UTC=-0.0223835, Accumulated Leap Seconds=13
  { 2442322.50, 45.209280 }, //02 October 1974, UT1-UTC=-0.0252801, Accumulated Leap Seconds=13
  { 2442323.50, 45.212115 }, //03 October 1974, UT1-UTC=-0.0281155, Accumulated Leap Seconds=13
  { 2442324.50, 45.214861 }, //04 October 1974, UT1-UTC=-0.0308605, Accumulated Leap Seconds=13
  { 2442325.50, 45.217508 }, //05 October 1974, UT1-UTC=-0.0335083, Accumulated Leap Seconds=13
  { 2442326.50, 45.220081 }, //06 October 1974, UT1-UTC=-0.0360812, Accumulated Leap Seconds=13
  { 2442327.50, 45.222630 }, //07 October 1974, UT1-UTC=-0.0386305, Accumulated Leap Seconds=13
  { 2442328.50, 45.225229 }, //08 October 1974, UT1-UTC=-0.0412288, Accumulated Leap Seconds=13
  { 2442329.50, 45.227954 }, //09 October 1974, UT1-UTC=-0.0439543, Accumulated Leap Seconds=13
  { 2442330.50, 45.230870 }, //10 October 1974, UT1-UTC=-0.0468704, Accumulated Leap Seconds=13
  { 2442331.50, 45.234006 }, //11 October 1974, UT1-UTC=-0.0500056, Accumulated Leap Seconds=13
  { 2442332.50, 45.237340 }, //12 October 1974, UT1-UTC=-0.0533405, Accumulated Leap Seconds=13
  { 2442333.50, 45.240806 }, //13 October 1974, UT1-UTC=-0.0568059, Accumulated Leap Seconds=13
  { 2442334.50, 45.244298 }, //14 October 1974, UT1-UTC=-0.0602980, Accumulated Leap Seconds=13
  { 2442335.50, 45.247706 }, //15 October 1974, UT1-UTC=-0.0637056, Accumulated Leap Seconds=13
  { 2442336.50, 45.250943 }, //16 October 1974, UT1-UTC=-0.0669432, Accumulated Leap Seconds=13
  { 2442337.50, 45.253974 }, //17 October 1974, UT1-UTC=-0.0699736, Accumulated Leap Seconds=13
  { 2442338.50, 45.256811 }, //18 October 1974, UT1-UTC=-0.0728115, Accumulated Leap Seconds=13
  { 2442339.50, 45.259510 }, //19 October 1974, UT1-UTC=-0.0755102, Accumulated Leap Seconds=13
  { 2442340.50, 45.262139 }, //20 October 1974, UT1-UTC=-0.0781390, Accumulated Leap Seconds=13
  { 2442341.50, 45.264764 }, //21 October 1974, UT1-UTC=-0.0807645, Accumulated Leap Seconds=13
  { 2442342.50, 45.267440 }, //22 October 1974, UT1-UTC=-0.0834400, Accumulated Leap Seconds=13
  { 2442343.50, 45.270204 }, //23 October 1974, UT1-UTC=-0.0862039, Accumulated Leap Seconds=13
  { 2442344.50, 45.273078 }, //24 October 1974, UT1-UTC=-0.0890785, Accumulated Leap Seconds=13
  { 2442345.50, 45.276070 }, //25 October 1974, UT1-UTC=-0.0920700, Accumulated Leap Seconds=13
  { 2442346.50, 45.279167 }, //26 October 1974, UT1-UTC=-0.0951674, Accumulated Leap Seconds=13
  { 2442347.50, 45.282344 }, //27 October 1974, UT1-UTC=-0.0983442, Accumulated Leap Seconds=13
  { 2442348.50, 45.285562 }, //28 October 1974, UT1-UTC=-0.1015618, Accumulated Leap Seconds=13
  { 2442349.50, 45.288776 }, //29 October 1974, UT1-UTC=-0.1047759, Accumulated Leap Seconds=13
  { 2442350.50, 45.291942 }, //30 October 1974, UT1-UTC=-0.1079424, Accumulated Leap Seconds=13
  { 2442351.50, 45.295025 }, //31 October 1974, UT1-UTC=-0.1110253, Accumulated Leap Seconds=13
  { 2442352.50, 45.298007 }, //01 November 1974, UT1-UTC=-0.1140068, Accumulated Leap Seconds=13
  { 2442353.50, 45.300897 }, //02 November 1974, UT1-UTC=-0.1168966, Accumulated Leap Seconds=13
  { 2442354.50, 45.303736 }, //03 November 1974, UT1-UTC=-0.1197360, Accumulated Leap Seconds=13
  { 2442355.50, 45.306592 }, //04 November 1974, UT1-UTC=-0.1225924, Accumulated Leap Seconds=13
  { 2442356.50, 45.309541 }, //05 November 1974, UT1-UTC=-0.1255415, Accumulated Leap Seconds=13
  { 2442357.50, 45.312646 }, //06 November 1974, UT1-UTC=-0.1286461, Accumulated Leap Seconds=13
  { 2442358.50, 45.315936 }, //07 November 1974, UT1-UTC=-0.1319364, Accumulated Leap Seconds=13
  { 2442359.50, 45.319400 }, //08 November 1974, UT1-UTC=-0.1354003, Accumulated Leap Seconds=13
  { 2442360.50, 45.322984 }, //09 November 1974, UT1-UTC=-0.1389844, Accumulated Leap Seconds=13
  { 2442361.50, 45.326605 }, //10 November 1974, UT1-UTC=-0.1426054, Accumulated Leap Seconds=13
  { 2442362.50, 45.330169 }, //11 November 1974, UT1-UTC=-0.1461694, Accumulated Leap Seconds=13
  { 2442363.50, 45.333595 }, //12 November 1974, UT1-UTC=-0.1495947, Accumulated Leap Seconds=13
  { 2442364.50, 45.336831 }, //13 November 1974, UT1-UTC=-0.1528314, Accumulated Leap Seconds=13
  { 2442365.50, 45.339872 }, //14 November 1974, UT1-UTC=-0.1558721, Accumulated Leap Seconds=13
  { 2442366.50, 45.342750 }, //15 November 1974, UT1-UTC=-0.1587500, Accumulated Leap Seconds=13
  { 2442367.50, 45.345524 }, //16 November 1974, UT1-UTC=-0.1615237, Accumulated Leap Seconds=13
  { 2442368.50, 45.348259 }, //17 November 1974, UT1-UTC=-0.1642589, Accumulated Leap Seconds=13
  { 2442369.50, 45.351013 }, //18 November 1974, UT1-UTC=-0.1670133, Accumulated Leap Seconds=13
  { 2442370.50, 45.353829 }, //19 November 1974, UT1-UTC=-0.1698291, Accumulated Leap Seconds=13
  { 2442371.50, 45.356731 }, //20 November 1974, UT1-UTC=-0.1727309, Accumulated Leap Seconds=13
  { 2442372.50, 45.359727 }, //21 November 1974, UT1-UTC=-0.1757270, Accumulated Leap Seconds=13
  { 2442373.50, 45.362810 }, //22 November 1974, UT1-UTC=-0.1788103, Accumulated Leap Seconds=13
  { 2442374.50, 45.365959 }, //23 November 1974, UT1-UTC=-0.1819591, Accumulated Leap Seconds=13
  { 2442375.50, 45.369141 }, //24 November 1974, UT1-UTC=-0.1851408, Accumulated Leap Seconds=13
  { 2442376.50, 45.372316 }, //25 November 1974, UT1-UTC=-0.1883156, Accumulated Leap Seconds=13
  { 2442377.50, 45.375442 }, //26 November 1974, UT1-UTC=-0.1914423, Accumulated Leap Seconds=13
  { 2442378.50, 45.378484 }, //27 November 1974, UT1-UTC=-0.1944842, Accumulated Leap Seconds=13
  { 2442379.50, 45.381417 }, //28 November 1974, UT1-UTC=-0.1974171, Accumulated Leap Seconds=13
  { 2442380.50, 45.384239 }, //29 November 1974, UT1-UTC=-0.2002394, Accumulated Leap Seconds=13
  { 2442381.50, 45.386981 }, //30 November 1974, UT1-UTC=-0.2029809, Accumulated Leap Seconds=13
  { 2442382.50, 45.389702 }, //01 December 1974, UT1-UTC=-0.2057017, Accumulated Leap Seconds=13
  { 2442383.50, 45.392480 }, //02 December 1974, UT1-UTC=-0.2084797, Accumulated Leap Seconds=13
  { 2442384.50, 45.395386 }, //03 December 1974, UT1-UTC=-0.2113859, Accumulated Leap Seconds=13
  { 2442385.50, 45.398459 }, //04 December 1974, UT1-UTC=-0.2144586, Accumulated Leap Seconds=13
  { 2442386.50, 45.401689 }, //05 December 1974, UT1-UTC=-0.2176890, Accumulated Leap Seconds=13
  { 2442387.50, 45.405023 }, //06 December 1974, UT1-UTC=-0.2210227, Accumulated Leap Seconds=13
  { 2442388.50, 45.408377 }, //07 December 1974, UT1-UTC=-0.2243769, Accumulated Leap Seconds=13
  { 2442389.50, 45.411663 }, //08 December 1974, UT1-UTC=-0.2276635, Accumulated Leap Seconds=13
  { 2442390.50, 45.414809 }, //09 December 1974, UT1-UTC=-0.2308093, Accumulated Leap Seconds=13
  { 2442391.50, 45.417770 }, //10 December 1974, UT1-UTC=-0.2337703, Accumulated Leap Seconds=13
  { 2442392.50, 45.420537 }, //11 December 1974, UT1-UTC=-0.2365371, Accumulated Leap Seconds=13
  { 2442393.50, 45.423134 }, //12 December 1974, UT1-UTC=-0.2391339, Accumulated Leap Seconds=13
  { 2442394.50, 45.425610 }, //13 December 1974, UT1-UTC=-0.2416100, Accumulated Leap Seconds=13
  { 2442395.50, 45.428027 }, //14 December 1974, UT1-UTC=-0.2440273, Accumulated Leap Seconds=13
  { 2442396.50, 45.430446 }, //15 December 1974, UT1-UTC=-0.2464465, Accumulated Leap Seconds=13
  { 2442397.50, 45.432916 }, //16 December 1974, UT1-UTC=-0.2489155, Accumulated Leap Seconds=13
  { 2442398.50, 45.435465 }, //17 December 1974, UT1-UTC=-0.2514648, Accumulated Leap Seconds=13
  { 2442399.50, 45.438106 }, //18 December 1974, UT1-UTC=-0.2541065, Accumulated Leap Seconds=13
  { 2442400.50, 45.440836 }, //19 December 1974, UT1-UTC=-0.2568362, Accumulated Leap Seconds=13
  { 2442401.50, 45.443636 }, //20 December 1974, UT1-UTC=-0.2596358, Accumulated Leap Seconds=13
  { 2442402.50, 45.446477 }, //21 December 1974, UT1-UTC=-0.2624767, Accumulated Leap Seconds=13
  { 2442403.50, 45.449324 }, //22 December 1974, UT1-UTC=-0.2653240, Accumulated Leap Seconds=13
  { 2442404.50, 45.452141 }, //23 December 1974, UT1-UTC=-0.2681412, Accumulated Leap Seconds=13
  { 2442405.50, 45.454895 }, //24 December 1974, UT1-UTC=-0.2708955, Accumulated Leap Seconds=13
  { 2442406.50, 45.457563 }, //25 December 1974, UT1-UTC=-0.2735629, Accumulated Leap Seconds=13
  { 2442407.50, 45.460136 }, //26 December 1974, UT1-UTC=-0.2761359, Accumulated Leap Seconds=13
  { 2442408.50, 45.462633 }, //27 December 1974, UT1-UTC=-0.2786327, Accumulated Leap Seconds=13
  { 2442409.50, 45.465103 }, //28 December 1974, UT1-UTC=-0.2811029, Accumulated Leap Seconds=13
  { 2442410.50, 45.467622 }, //29 December 1974, UT1-UTC=-0.2836216, Accumulated Leap Seconds=13
  { 2442411.50, 45.470270 }, //30 December 1974, UT1-UTC=-0.2862702, Accumulated Leap Seconds=13
  { 2442412.50, 45.473106 }, //31 December 1974, UT1-UTC=-0.2891063, Accumulated Leap Seconds=13
  { 2442413.50, 45.476138 }, //01 January 1975, UT1-UTC= 0.7078620, Accumulated Leap Seconds=14
  { 2442414.50, 45.479316 }, //02 January 1975, UT1-UTC= 0.7046843, Accumulated Leap Seconds=14
  { 2442415.50, 45.482546 }, //03 January 1975, UT1-UTC= 0.7014537, Accumulated Leap Seconds=14
  { 2442416.50, 45.485725 }, //04 January 1975, UT1-UTC= 0.6982754, Accumulated Leap Seconds=14
  { 2442417.50, 45.488764 }, //05 January 1975, UT1-UTC= 0.6952360, Accumulated Leap Seconds=14
  { 2442418.50, 45.491615 }, //06 January 1975, UT1-UTC= 0.6923847, Accumulated Leap Seconds=14
  { 2442419.50, 45.494271 }, //07 January 1975, UT1-UTC= 0.6897288, Accumulated Leap Seconds=14
  { 2442420.50, 45.496758 }, //08 January 1975, UT1-UTC= 0.6872419, Accumulated Leap Seconds=14
  { 2442421.50, 45.499125 }, //09 January 1975, UT1-UTC= 0.6848751, Accumulated Leap Seconds=14
  { 2442422.50, 45.501431 }, //10 January 1975, UT1-UTC= 0.6825688, Accumulated Leap Seconds=14
  { 2442423.50, 45.503737 }, //11 January 1975, UT1-UTC= 0.6802626, Accumulated Leap Seconds=14
  { 2442424.50, 45.506095 }, //12 January 1975, UT1-UTC= 0.6779046, Accumulated Leap Seconds=14
  { 2442425.50, 45.508542 }, //13 January 1975, UT1-UTC= 0.6754580, Accumulated Leap Seconds=14
  { 2442426.50, 45.511095 }, //14 January 1975, UT1-UTC= 0.6729050, Accumulated Leap Seconds=14
  { 2442427.50, 45.513754 }, //15 January 1975, UT1-UTC= 0.6702462, Accumulated Leap Seconds=14
  { 2442428.50, 45.516502 }, //16 January 1975, UT1-UTC= 0.6674983, Accumulated Leap Seconds=14
  { 2442429.50, 45.519311 }, //17 January 1975, UT1-UTC= 0.6646893, Accumulated Leap Seconds=14
  { 2442430.50, 45.522146 }, //18 January 1975, UT1-UTC= 0.6618535, Accumulated Leap Seconds=14
  { 2442431.50, 45.524973 }, //19 January 1975, UT1-UTC= 0.6590266, Accumulated Leap Seconds=14
  { 2442432.50, 45.527759 }, //20 January 1975, UT1-UTC= 0.6562407, Accumulated Leap Seconds=14
  { 2442433.50, 45.530480 }, //21 January 1975, UT1-UTC= 0.6535197, Accumulated Leap Seconds=14
  { 2442434.50, 45.533126 }, //22 January 1975, UT1-UTC= 0.6508739, Accumulated Leap Seconds=14
  { 2442435.50, 45.535706 }, //23 January 1975, UT1-UTC= 0.6482936, Accumulated Leap Seconds=14
  { 2442436.50, 45.538258 }, //24 January 1975, UT1-UTC= 0.6457422, Accumulated Leap Seconds=14
  { 2442437.50, 45.540845 }, //25 January 1975, UT1-UTC= 0.6431545, Accumulated Leap Seconds=14
  { 2442438.50, 45.543552 }, //26 January 1975, UT1-UTC= 0.6404476, Accumulated Leap Seconds=14
  { 2442439.50, 45.546455 }, //27 January 1975, UT1-UTC= 0.6375449, Accumulated Leap Seconds=14
  { 2442440.50, 45.549591 }, //28 January 1975, UT1-UTC= 0.6344086, Accumulated Leap Seconds=14
  { 2442441.50, 45.552938 }, //29 January 1975, UT1-UTC= 0.6310622, Accumulated Leap Seconds=14
  { 2442442.50, 45.556410 }, //30 January 1975, UT1-UTC= 0.6275904, Accumulated Leap Seconds=14
  { 2442443.50, 45.559886 }, //31 January 1975, UT1-UTC= 0.6241144, Accumulated Leap Seconds=14
  { 2442444.50, 45.563249 }, //01 February 1975, UT1-UTC= 0.6207515, Accumulated Leap Seconds=14
  { 2442445.50, 45.566419 }, //02 February 1975, UT1-UTC= 0.6175812, Accumulated Leap Seconds=14
  { 2442446.50, 45.569371 }, //03 February 1975, UT1-UTC= 0.6146292, Accumulated Leap Seconds=14
  { 2442447.50, 45.572127 }, //04 February 1975, UT1-UTC= 0.6118733, Accumulated Leap Seconds=14
  { 2442448.50, 45.574738 }, //05 February 1975, UT1-UTC= 0.6092617, Accumulated Leap Seconds=14
  { 2442449.50, 45.577269 }, //06 February 1975, UT1-UTC= 0.6067312, Accumulated Leap Seconds=14
  { 2442450.50, 45.579780 }, //07 February 1975, UT1-UTC= 0.6042198, Accumulated Leap Seconds=14
  { 2442451.50, 45.582326 }, //08 February 1975, UT1-UTC= 0.6016743, Accumulated Leap Seconds=14
  { 2442452.50, 45.584946 }, //09 February 1975, UT1-UTC= 0.5990538, Accumulated Leap Seconds=14
  { 2442453.50, 45.587666 }, //10 February 1975, UT1-UTC= 0.5963343, Accumulated Leap Seconds=14
  { 2442454.50, 45.590489 }, //11 February 1975, UT1-UTC= 0.5935105, Accumulated Leap Seconds=14
  { 2442455.50, 45.593405 }, //12 February 1975, UT1-UTC= 0.5905952, Accumulated Leap Seconds=14
  { 2442456.50, 45.596384 }, //13 February 1975, UT1-UTC= 0.5876157, Accumulated Leap Seconds=14
  { 2442457.50, 45.599392 }, //14 February 1975, UT1-UTC= 0.5846077, Accumulated Leap Seconds=14
  { 2442458.50, 45.602390 }, //15 February 1975, UT1-UTC= 0.5816096, Accumulated Leap Seconds=14
  { 2442459.50, 45.605343 }, //16 February 1975, UT1-UTC= 0.5786568, Accumulated Leap Seconds=14
  { 2442460.50, 45.608223 }, //17 February 1975, UT1-UTC= 0.5757765, Accumulated Leap Seconds=14
  { 2442461.50, 45.611016 }, //18 February 1975, UT1-UTC= 0.5729835, Accumulated Leap Seconds=14
  { 2442462.50, 45.613726 }, //19 February 1975, UT1-UTC= 0.5702742, Accumulated Leap Seconds=14
  { 2442463.50, 45.616378 }, //20 February 1975, UT1-UTC= 0.5676221, Accumulated Leap Seconds=14
  { 2442464.50, 45.619026 }, //21 February 1975, UT1-UTC= 0.5649742, Accumulated Leap Seconds=14
  { 2442465.50, 45.621745 }, //22 February 1975, UT1-UTC= 0.5622551, Accumulated Leap Seconds=14
  { 2442466.50, 45.624618 }, //23 February 1975, UT1-UTC= 0.5593821, Accumulated Leap Seconds=14
  { 2442467.50, 45.627708 }, //24 February 1975, UT1-UTC= 0.5562918, Accumulated Leap Seconds=14
  { 2442468.50, 45.631031 }, //25 February 1975, UT1-UTC= 0.5529687, Accumulated Leap Seconds=14
  { 2442469.50, 45.634538 }, //26 February 1975, UT1-UTC= 0.5494621, Accumulated Leap Seconds=14
  { 2442470.50, 45.638122 }, //27 February 1975, UT1-UTC= 0.5458778, Accumulated Leap Seconds=14
  { 2442471.50, 45.641654 }, //28 February 1975, UT1-UTC= 0.5423464, Accumulated Leap Seconds=14
  { 2442472.50, 45.645019 }, //01 March 1975, UT1-UTC= 0.5389811, Accumulated Leap Seconds=14
  { 2442473.50, 45.648156 }, //02 March 1975, UT1-UTC= 0.5358444, Accumulated Leap Seconds=14
  { 2442474.50, 45.651062 }, //03 March 1975, UT1-UTC= 0.5329378, Accumulated Leap Seconds=14
  { 2442475.50, 45.653784 }, //04 March 1975, UT1-UTC= 0.5302157, Accumulated Leap Seconds=14
  { 2442476.50, 45.656390 }, //05 March 1975, UT1-UTC= 0.5276097, Accumulated Leap Seconds=14
  { 2442477.50, 45.658950 }, //06 March 1975, UT1-UTC= 0.5250500, Accumulated Leap Seconds=14
  { 2442478.50, 45.661523 }, //07 March 1975, UT1-UTC= 0.5224774, Accumulated Leap Seconds=14
  { 2442479.50, 45.664153 }, //08 March 1975, UT1-UTC= 0.5198474, Accumulated Leap Seconds=14
  { 2442480.50, 45.666870 }, //09 March 1975, UT1-UTC= 0.5171301, Accumulated Leap Seconds=14
  { 2442481.50, 45.669688 }, //10 March 1975, UT1-UTC= 0.5143123, Accumulated Leap Seconds=14
  { 2442482.50, 45.672602 }, //11 March 1975, UT1-UTC= 0.5113982, Accumulated Leap Seconds=14
  { 2442483.50, 45.675591 }, //12 March 1975, UT1-UTC= 0.5084088, Accumulated Leap Seconds=14
  { 2442484.50, 45.678623 }, //13 March 1975, UT1-UTC= 0.5053768, Accumulated Leap Seconds=14
  { 2442485.50, 45.681660 }, //14 March 1975, UT1-UTC= 0.5023401, Accumulated Leap Seconds=14
  { 2442486.50, 45.684664 }, //15 March 1975, UT1-UTC= 0.4993360, Accumulated Leap Seconds=14
  { 2442487.50, 45.687605 }, //16 March 1975, UT1-UTC= 0.4963947, Accumulated Leap Seconds=14
  { 2442488.50, 45.690465 }, //17 March 1975, UT1-UTC= 0.4935350, Accumulated Leap Seconds=14
  { 2442489.50, 45.693241 }, //18 March 1975, UT1-UTC= 0.4907588, Accumulated Leap Seconds=14
  { 2442490.50, 45.695953 }, //19 March 1975, UT1-UTC= 0.4880470, Accumulated Leap Seconds=14
  { 2442491.50, 45.698643 }, //20 March 1975, UT1-UTC= 0.4853569, Accumulated Leap Seconds=14
  { 2442492.50, 45.701376 }, //21 March 1975, UT1-UTC= 0.4826241, Accumulated Leap Seconds=14
  { 2442493.50, 45.704228 }, //22 March 1975, UT1-UTC= 0.4797722, Accumulated Leap Seconds=14
  { 2442494.50, 45.707270 }, //23 March 1975, UT1-UTC= 0.4767299, Accumulated Leap Seconds=14
  { 2442495.50, 45.710545 }, //24 March 1975, UT1-UTC= 0.4734549, Accumulated Leap Seconds=14
  { 2442496.50, 45.714043 }, //25 March 1975, UT1-UTC= 0.4699565, Accumulated Leap Seconds=14
  { 2442497.50, 45.717697 }, //26 March 1975, UT1-UTC= 0.4663029, Accumulated Leap Seconds=14
  { 2442498.50, 45.721392 }, //27 March 1975, UT1-UTC= 0.4626084, Accumulated Leap Seconds=14
  { 2442499.50, 45.725001 }, //28 March 1975, UT1-UTC= 0.4589992, Accumulated Leap Seconds=14
  { 2442500.50, 45.728427 }, //29 March 1975, UT1-UTC= 0.4555732, Accumulated Leap Seconds=14
  { 2442501.50, 45.731627 }, //30 March 1975, UT1-UTC= 0.4523729, Accumulated Leap Seconds=14
  { 2442502.50, 45.734619 }, //31 March 1975, UT1-UTC= 0.4493813, Accumulated Leap Seconds=14
  { 2442503.50, 45.737459 }, //01 April 1975, UT1-UTC= 0.4465407, Accumulated Leap Seconds=14
  { 2442504.50, 45.740221 }, //02 April 1975, UT1-UTC= 0.4437788, Accumulated Leap Seconds=14
  { 2442505.50, 45.742970 }, //03 April 1975, UT1-UTC= 0.4410300, Accumulated Leap Seconds=14
  { 2442506.50, 45.745755 }, //04 April 1975, UT1-UTC= 0.4382447, Accumulated Leap Seconds=14
  { 2442507.50, 45.748609 }, //05 April 1975, UT1-UTC= 0.4353906, Accumulated Leap Seconds=14
  { 2442508.50, 45.751549 }, //06 April 1975, UT1-UTC= 0.4324509, Accumulated Leap Seconds=14
  { 2442509.50, 45.754576 }, //07 April 1975, UT1-UTC= 0.4294241, Accumulated Leap Seconds=14
  { 2442510.50, 45.757676 }, //08 April 1975, UT1-UTC= 0.4263244, Accumulated Leap Seconds=14
  { 2442511.50, 45.760821 }, //09 April 1975, UT1-UTC= 0.4231794, Accumulated Leap Seconds=14
  { 2442512.50, 45.763975 }, //10 April 1975, UT1-UTC= 0.4200251, Accumulated Leap Seconds=14
  { 2442513.50, 45.767101 }, //11 April 1975, UT1-UTC= 0.4168989, Accumulated Leap Seconds=14
  { 2442514.50, 45.770166 }, //12 April 1975, UT1-UTC= 0.4138340, Accumulated Leap Seconds=14
  { 2442515.50, 45.773147 }, //13 April 1975, UT1-UTC= 0.4108533, Accumulated Leap Seconds=14
  { 2442516.50, 45.776035 }, //14 April 1975, UT1-UTC= 0.4079649, Accumulated Leap Seconds=14
  { 2442517.50, 45.778843 }, //15 April 1975, UT1-UTC= 0.4051570, Accumulated Leap Seconds=14
  { 2442518.50, 45.781605 }, //16 April 1975, UT1-UTC= 0.4023952, Accumulated Leap Seconds=14
  { 2442519.50, 45.784375 }, //17 April 1975, UT1-UTC= 0.3996246, Accumulated Leap Seconds=14
  { 2442520.50, 45.787222 }, //18 April 1975, UT1-UTC= 0.3967781, Accumulated Leap Seconds=14
  { 2442521.50, 45.790210 }, //19 April 1975, UT1-UTC= 0.3937899, Accumulated Leap Seconds=14
  { 2442522.50, 45.793387 }, //20 April 1975, UT1-UTC= 0.3906134, Accumulated Leap Seconds=14
  { 2442523.50, 45.796762 }, //21 April 1975, UT1-UTC= 0.3872379, Accumulated Leap Seconds=14
  { 2442524.50, 45.800300 }, //22 April 1975, UT1-UTC= 0.3836997, Accumulated Leap Seconds=14
  { 2442525.50, 45.803921 }, //23 April 1975, UT1-UTC= 0.3800794, Accumulated Leap Seconds=14
  { 2442526.50, 45.807516 }, //24 April 1975, UT1-UTC= 0.3764845, Accumulated Leap Seconds=14
  { 2442527.50, 45.810981 }, //25 April 1975, UT1-UTC= 0.3730187, Accumulated Leap Seconds=14
  { 2442528.50, 45.814249 }, //26 April 1975, UT1-UTC= 0.3697510, Accumulated Leap Seconds=14
  { 2442529.50, 45.817302 }, //27 April 1975, UT1-UTC= 0.3666976, Accumulated Leap Seconds=14
  { 2442530.50, 45.820176 }, //28 April 1975, UT1-UTC= 0.3638240, Accumulated Leap Seconds=14
  { 2442531.50, 45.822934 }, //29 April 1975, UT1-UTC= 0.3610656, Accumulated Leap Seconds=14
  { 2442532.50, 45.825648 }, //30 April 1975, UT1-UTC= 0.3583524, Accumulated Leap Seconds=14
  { 2442533.50, 45.828372 }, //01 May 1975, UT1-UTC= 0.3556279, Accumulated Leap Seconds=14
  { 2442534.50, 45.831144 }, //02 May 1975, UT1-UTC= 0.3528558, Accumulated Leap Seconds=14
  { 2442535.50, 45.833982 }, //03 May 1975, UT1-UTC= 0.3500183, Accumulated Leap Seconds=14
  { 2442536.50, 45.836887 }, //04 May 1975, UT1-UTC= 0.3471131, Accumulated Leap Seconds=14
  { 2442537.50, 45.839849 }, //05 May 1975, UT1-UTC= 0.3441513, Accumulated Leap Seconds=14
  { 2442538.50, 45.842844 }, //06 May 1975, UT1-UTC= 0.3411564, Accumulated Leap Seconds=14
  { 2442539.50, 45.845839 }, //07 May 1975, UT1-UTC= 0.3381611, Accumulated Leap Seconds=14
  { 2442540.50, 45.848799 }, //08 May 1975, UT1-UTC= 0.3352008, Accumulated Leap Seconds=14
  { 2442541.50, 45.851691 }, //09 May 1975, UT1-UTC= 0.3323086, Accumulated Leap Seconds=14
  { 2442542.50, 45.854491 }, //10 May 1975, UT1-UTC= 0.3295094, Accumulated Leap Seconds=14
  { 2442543.50, 45.857185 }, //11 May 1975, UT1-UTC= 0.3268149, Accumulated Leap Seconds=14
  { 2442544.50, 45.859781 }, //12 May 1975, UT1-UTC= 0.3242189, Accumulated Leap Seconds=14
  { 2442545.50, 45.862307 }, //13 May 1975, UT1-UTC= 0.3216932, Accumulated Leap Seconds=14
  { 2442546.50, 45.864812 }, //14 May 1975, UT1-UTC= 0.3191880, Accumulated Leap Seconds=14
  { 2442547.50, 45.867361 }, //15 May 1975, UT1-UTC= 0.3166393, Accumulated Leap Seconds=14
  { 2442548.50, 45.870016 }, //16 May 1975, UT1-UTC= 0.3139843, Accumulated Leap Seconds=14
  { 2442549.50, 45.872822 }, //17 May 1975, UT1-UTC= 0.3111779, Accumulated Leap Seconds=14
  { 2442550.50, 45.875793 }, //18 May 1975, UT1-UTC= 0.3082074, Accumulated Leap Seconds=14
  { 2442551.50, 45.878901 }, //19 May 1975, UT1-UTC= 0.3050988, Accumulated Leap Seconds=14
  { 2442552.50, 45.882086 }, //20 May 1975, UT1-UTC= 0.3019142, Accumulated Leap Seconds=14
  { 2442553.50, 45.885261 }, //21 May 1975, UT1-UTC= 0.2987388, Accumulated Leap Seconds=14
  { 2442554.50, 45.888338 }, //22 May 1975, UT1-UTC= 0.2956617, Accumulated Leap Seconds=14
  { 2442555.50, 45.891247 }, //23 May 1975, UT1-UTC= 0.2927525, Accumulated Leap Seconds=14
  { 2442556.50, 45.893957 }, //24 May 1975, UT1-UTC= 0.2900434, Accumulated Leap Seconds=14
  { 2442557.50, 45.896479 }, //25 May 1975, UT1-UTC= 0.2875213, Accumulated Leap Seconds=14
  { 2442558.50, 45.898864 }, //26 May 1975, UT1-UTC= 0.2851360, Accumulated Leap Seconds=14
  { 2442559.50, 45.901180 }, //27 May 1975, UT1-UTC= 0.2828200, Accumulated Leap Seconds=14
  { 2442560.50, 45.903491 }, //28 May 1975, UT1-UTC= 0.2805093, Accumulated Leap Seconds=14
  { 2442561.50, 45.905842 }, //29 May 1975, UT1-UTC= 0.2781577, Accumulated Leap Seconds=14
  { 2442562.50, 45.908259 }, //30 May 1975, UT1-UTC= 0.2757407, Accumulated Leap Seconds=14
  { 2442563.50, 45.910747 }, //31 May 1975, UT1-UTC= 0.2732527, Accumulated Leap Seconds=14
  { 2442564.50, 45.913298 }, //01 June 1975, UT1-UTC= 0.2707024, Accumulated Leap Seconds=14
  { 2442565.50, 45.915890 }, //02 June 1975, UT1-UTC= 0.2681096, Accumulated Leap Seconds=14
  { 2442566.50, 45.918497 }, //03 June 1975, UT1-UTC= 0.2655034, Accumulated Leap Seconds=14
  { 2442567.50, 45.921081 }, //04 June 1975, UT1-UTC= 0.2629185, Accumulated Leap Seconds=14
  { 2442568.50, 45.923611 }, //05 June 1975, UT1-UTC= 0.2603892, Accumulated Leap Seconds=14
  { 2442569.50, 45.926057 }, //06 June 1975, UT1-UTC= 0.2579426, Accumulated Leap Seconds=14
  { 2442570.50, 45.928406 }, //07 June 1975, UT1-UTC= 0.2555943, Accumulated Leap Seconds=14
  { 2442571.50, 45.930656 }, //08 June 1975, UT1-UTC= 0.2533436, Accumulated Leap Seconds=14
  { 2442572.50, 45.932830 }, //09 June 1975, UT1-UTC= 0.2511698, Accumulated Leap Seconds=14
  { 2442573.50, 45.934971 }, //10 June 1975, UT1-UTC= 0.2490292, Accumulated Leap Seconds=14
  { 2442574.50, 45.937140 }, //11 June 1975, UT1-UTC= 0.2468600, Accumulated Leap Seconds=14
  { 2442575.50, 45.939404 }, //12 June 1975, UT1-UTC= 0.2445960, Accumulated Leap Seconds=14
  { 2442576.50, 45.941813 }, //13 June 1975, UT1-UTC= 0.2421869, Accumulated Leap Seconds=14
  { 2442577.50, 45.944383 }, //14 June 1975, UT1-UTC= 0.2396174, Accumulated Leap Seconds=14
  { 2442578.50, 45.947085 }, //15 June 1975, UT1-UTC= 0.2369150, Accumulated Leap Seconds=14
  { 2442579.50, 45.949855 }, //16 June 1975, UT1-UTC= 0.2341448, Accumulated Leap Seconds=14
  { 2442580.50, 45.952608 }, //17 June 1975, UT1-UTC= 0.2313915, Accumulated Leap Seconds=14
  { 2442581.50, 45.955261 }, //18 June 1975, UT1-UTC= 0.2287388, Accumulated Leap Seconds=14
  { 2442582.50, 45.957751 }, //19 June 1975, UT1-UTC= 0.2262494, Accumulated Leap Seconds=14
  { 2442583.50, 45.960046 }, //20 June 1975, UT1-UTC= 0.2239539, Accumulated Leap Seconds=14
  { 2442584.50, 45.962153 }, //21 June 1975, UT1-UTC= 0.2218465, Accumulated Leap Seconds=14
  { 2442585.50, 45.964111 }, //22 June 1975, UT1-UTC= 0.2198887, Accumulated Leap Seconds=14
  { 2442586.50, 45.965979 }, //23 June 1975, UT1-UTC= 0.2180206, Accumulated Leap Seconds=14
  { 2442587.50, 45.967821 }, //24 June 1975, UT1-UTC= 0.2161786, Accumulated Leap Seconds=14
  { 2442588.50, 45.969690 }, //25 June 1975, UT1-UTC= 0.2143102, Accumulated Leap Seconds=14
  { 2442589.50, 45.971617 }, //26 June 1975, UT1-UTC= 0.2123831, Accumulated Leap Seconds=14
  { 2442590.50, 45.973613 }, //27 June 1975, UT1-UTC= 0.2103870, Accumulated Leap Seconds=14
  { 2442591.50, 45.975670 }, //28 June 1975, UT1-UTC= 0.2083295, Accumulated Leap Seconds=14
  { 2442592.50, 45.977769 }, //29 June 1975, UT1-UTC= 0.2062314, Accumulated Leap Seconds=14
  { 2442593.50, 45.979879 }, //30 June 1975, UT1-UTC= 0.2041206, Accumulated Leap Seconds=14
  { 2442594.50, 45.981970 }, //01 July 1975, UT1-UTC= 0.2020295, Accumulated Leap Seconds=14
  { 2442595.50, 45.984010 }, //02 July 1975, UT1-UTC= 0.1999898, Accumulated Leap Seconds=14
  { 2442596.50, 45.985972 }, //03 July 1975, UT1-UTC= 0.1980282, Accumulated Leap Seconds=14
  { 2442597.50, 45.987839 }, //04 July 1975, UT1-UTC= 0.1961611, Accumulated Leap Seconds=14
  { 2442598.50, 45.989610 }, //05 July 1975, UT1-UTC= 0.1943904, Accumulated Leap Seconds=14
  { 2442599.50, 45.991300 }, //06 July 1975, UT1-UTC= 0.1926999, Accumulated Leap Seconds=14
  { 2442600.50, 45.992948 }, //07 July 1975, UT1-UTC= 0.1910515, Accumulated Leap Seconds=14
  { 2442601.50, 45.994614 }, //08 July 1975, UT1-UTC= 0.1893860, Accumulated Leap Seconds=14
  { 2442602.50, 45.996369 }, //09 July 1975, UT1-UTC= 0.1876314, Accumulated Leap Seconds=14
  { 2442603.50, 45.998278 }, //10 July 1975, UT1-UTC= 0.1857224, Accumulated Leap Seconds=14
  { 2442604.50, 46.000375 }, //11 July 1975, UT1-UTC= 0.1836248, Accumulated Leap Seconds=14
  { 2442605.50, 46.002647 }, //12 July 1975, UT1-UTC= 0.1813529, Accumulated Leap Seconds=14
  { 2442606.50, 46.005029 }, //13 July 1975, UT1-UTC= 0.1789710, Accumulated Leap Seconds=14
  { 2442607.50, 46.007425 }, //14 July 1975, UT1-UTC= 0.1765746, Accumulated Leap Seconds=14
  { 2442608.50, 46.009739 }, //15 July 1975, UT1-UTC= 0.1742613, Accumulated Leap Seconds=14
  { 2442609.50, 46.011896 }, //16 July 1975, UT1-UTC= 0.1721038, Accumulated Leap Seconds=14
  { 2442610.50, 46.013865 }, //17 July 1975, UT1-UTC= 0.1701352, Accumulated Leap Seconds=14
  { 2442611.50, 46.015652 }, //18 July 1975, UT1-UTC= 0.1683480, Accumulated Leap Seconds=14
  { 2442612.50, 46.017297 }, //19 July 1975, UT1-UTC= 0.1667032, Accumulated Leap Seconds=14
  { 2442613.50, 46.018857 }, //20 July 1975, UT1-UTC= 0.1651432, Accumulated Leap Seconds=14
  { 2442614.50, 46.020395 }, //21 July 1975, UT1-UTC= 0.1636046, Accumulated Leap Seconds=14
  { 2442615.50, 46.021969 }, //22 July 1975, UT1-UTC= 0.1620306, Accumulated Leap Seconds=14
  { 2442616.50, 46.023620 }, //23 July 1975, UT1-UTC= 0.1603799, Accumulated Leap Seconds=14
  { 2442617.50, 46.025368 }, //24 July 1975, UT1-UTC= 0.1586320, Accumulated Leap Seconds=14
  { 2442618.50, 46.027213 }, //25 July 1975, UT1-UTC= 0.1567872, Accumulated Leap Seconds=14
  { 2442619.50, 46.029138 }, //26 July 1975, UT1-UTC= 0.1548623, Accumulated Leap Seconds=14
  { 2442620.50, 46.031115 }, //27 July 1975, UT1-UTC= 0.1528845, Accumulated Leap Seconds=14
  { 2442621.50, 46.033114 }, //28 July 1975, UT1-UTC= 0.1508862, Accumulated Leap Seconds=14
  { 2442622.50, 46.035100 }, //29 July 1975, UT1-UTC= 0.1488999, Accumulated Leap Seconds=14
  { 2442623.50, 46.037046 }, //30 July 1975, UT1-UTC= 0.1469542, Accumulated Leap Seconds=14
  { 2442624.50, 46.038931 }, //31 July 1975, UT1-UTC= 0.1450691, Accumulated Leap Seconds=14
  { 2442625.50, 46.040748 }, //01 August 1975, UT1-UTC= 0.1432516, Accumulated Leap Seconds=14
  { 2442626.50, 46.042507 }, //02 August 1975, UT1-UTC= 0.1414926, Accumulated Leap Seconds=14
  { 2442627.50, 46.044236 }, //03 August 1975, UT1-UTC= 0.1397635, Accumulated Leap Seconds=14
  { 2442628.50, 46.045986 }, //04 August 1975, UT1-UTC= 0.1380145, Accumulated Leap Seconds=14
  { 2442629.50, 46.047822 }, //05 August 1975, UT1-UTC= 0.1361780, Accumulated Leap Seconds=14
  { 2442630.50, 46.049819 }, //06 August 1975, UT1-UTC= 0.1341813, Accumulated Leap Seconds=14
  { 2442631.50, 46.052031 }, //07 August 1975, UT1-UTC= 0.1319693, Accumulated Leap Seconds=14
  { 2442632.50, 46.054470 }, //08 August 1975, UT1-UTC= 0.1295300, Accumulated Leap Seconds=14
  { 2442633.50, 46.057091 }, //09 August 1975, UT1-UTC= 0.1269087, Accumulated Leap Seconds=14
  { 2442634.50, 46.059799 }, //10 August 1975, UT1-UTC= 0.1242014, Accumulated Leap Seconds=14
  { 2442635.50, 46.062474 }, //11 August 1975, UT1-UTC= 0.1215261, Accumulated Leap Seconds=14
  { 2442636.50, 46.065014 }, //12 August 1975, UT1-UTC= 0.1189857, Accumulated Leap Seconds=14
  { 2442637.50, 46.067361 }, //13 August 1975, UT1-UTC= 0.1166393, Accumulated Leap Seconds=14
  { 2442638.50, 46.069507 }, //14 August 1975, UT1-UTC= 0.1144927, Accumulated Leap Seconds=14
  { 2442639.50, 46.071491 }, //15 August 1975, UT1-UTC= 0.1125086, Accumulated Leap Seconds=14
  { 2442640.50, 46.073374 }, //16 August 1975, UT1-UTC= 0.1106264, Accumulated Leap Seconds=14
  { 2442641.50, 46.075220 }, //17 August 1975, UT1-UTC= 0.1087804, Accumulated Leap Seconds=14
  { 2442642.50, 46.077087 }, //18 August 1975, UT1-UTC= 0.1069126, Accumulated Leap Seconds=14
  { 2442643.50, 46.079021 }, //19 August 1975, UT1-UTC= 0.1049793, Accumulated Leap Seconds=14
  { 2442644.50, 46.081045 }, //20 August 1975, UT1-UTC= 0.1029550, Accumulated Leap Seconds=14
  { 2442645.50, 46.083167 }, //21 August 1975, UT1-UTC= 0.1008334, Accumulated Leap Seconds=14
  { 2442646.50, 46.085373 }, //22 August 1975, UT1-UTC= 0.0986271, Accumulated Leap Seconds=14
  { 2442647.50, 46.087638 }, //23 August 1975, UT1-UTC= 0.0963623, Accumulated Leap Seconds=14
  { 2442648.50, 46.089927 }, //24 August 1975, UT1-UTC= 0.0940731, Accumulated Leap Seconds=14
  { 2442649.50, 46.092205 }, //25 August 1975, UT1-UTC= 0.0917951, Accumulated Leap Seconds=14
  { 2442650.50, 46.094441 }, //26 August 1975, UT1-UTC= 0.0895594, Accumulated Leap Seconds=14
  { 2442651.50, 46.096611 }, //27 August 1975, UT1-UTC= 0.0873887, Accumulated Leap Seconds=14
  { 2442652.50, 46.098707 }, //28 August 1975, UT1-UTC= 0.0852931, Accumulated Leap Seconds=14
  { 2442653.50, 46.100733 }, //29 August 1975, UT1-UTC= 0.0832673, Accumulated Leap Seconds=14
  { 2442654.50, 46.102711 }, //30 August 1975, UT1-UTC= 0.0812891, Accumulated Leap Seconds=14
  { 2442655.50, 46.104683 }, //31 August 1975, UT1-UTC= 0.0793170, Accumulated Leap Seconds=14
  { 2442656.50, 46.106708 }, //01 September 1975, UT1-UTC= 0.0772916, Accumulated Leap Seconds=14
  { 2442657.50, 46.108859 }, //02 September 1975, UT1-UTC= 0.0751413, Accumulated Leap Seconds=14
  { 2442658.50, 46.111202 }, //03 September 1975, UT1-UTC= 0.0727981, Accumulated Leap Seconds=14
  { 2442659.50, 46.113779 }, //04 September 1975, UT1-UTC= 0.0702207, Accumulated Leap Seconds=14
  { 2442660.50, 46.116582 }, //05 September 1975, UT1-UTC= 0.0674183, Accumulated Leap Seconds=14
  { 2442661.50, 46.119542 }, //06 September 1975, UT1-UTC= 0.0644583, Accumulated Leap Seconds=14
  { 2442662.50, 46.122546 }, //07 September 1975, UT1-UTC= 0.0614537, Accumulated Leap Seconds=14
  { 2442663.50, 46.125472 }, //08 September 1975, UT1-UTC= 0.0585280, Accumulated Leap Seconds=14
  { 2442664.50, 46.128224 }, //09 September 1975, UT1-UTC= 0.0557757, Accumulated Leap Seconds=14
  { 2442665.50, 46.130764 }, //10 September 1975, UT1-UTC= 0.0532356, Accumulated Leap Seconds=14
  { 2442666.50, 46.133112 }, //11 September 1975, UT1-UTC= 0.0508883, Accumulated Leap Seconds=14
  { 2442667.50, 46.135326 }, //12 September 1975, UT1-UTC= 0.0486743, Accumulated Leap Seconds=14
  { 2442668.50, 46.137480 }, //13 September 1975, UT1-UTC= 0.0465203, Accumulated Leap Seconds=14
  { 2442669.50, 46.139640 }, //14 September 1975, UT1-UTC= 0.0443600, Accumulated Leap Seconds=14
  { 2442670.50, 46.141856 }, //15 September 1975, UT1-UTC= 0.0421439, Accumulated Leap Seconds=14
  { 2442671.50, 46.144158 }, //16 September 1975, UT1-UTC= 0.0398415, Accumulated Leap Seconds=14
  { 2442672.50, 46.146560 }, //17 September 1975, UT1-UTC= 0.0374400, Accumulated Leap Seconds=14
  { 2442673.50, 46.149056 }, //18 September 1975, UT1-UTC= 0.0349439, Accumulated Leap Seconds=14
  { 2442674.50, 46.151628 }, //19 September 1975, UT1-UTC= 0.0323719, Accumulated Leap Seconds=14
  { 2442675.50, 46.154246 }, //20 September 1975, UT1-UTC= 0.0297541, Accumulated Leap Seconds=14
  { 2442676.50, 46.156875 }, //21 September 1975, UT1-UTC= 0.0271247, Accumulated Leap Seconds=14
  { 2442677.50, 46.159484 }, //22 September 1975, UT1-UTC= 0.0245156, Accumulated Leap Seconds=14
  { 2442678.50, 46.162049 }, //23 September 1975, UT1-UTC= 0.0219512, Accumulated Leap Seconds=14
  { 2442679.50, 46.164556 }, //24 September 1975, UT1-UTC= 0.0194443, Accumulated Leap Seconds=14
  { 2442680.50, 46.167006 }, //25 September 1975, UT1-UTC= 0.0169938, Accumulated Leap Seconds=14
  { 2442681.50, 46.169417 }, //26 September 1975, UT1-UTC= 0.0145829, Accumulated Leap Seconds=14
  { 2442682.50, 46.171822 }, //27 September 1975, UT1-UTC= 0.0121777, Accumulated Leap Seconds=14
  { 2442683.50, 46.174272 }, //28 September 1975, UT1-UTC= 0.0097282, Accumulated Leap Seconds=14
  { 2442684.50, 46.176828 }, //29 September 1975, UT1-UTC= 0.0071724, Accumulated Leap Seconds=14
  { 2442685.50, 46.179554 }, //30 September 1975, UT1-UTC= 0.0044456, Accumulated Leap Seconds=14
  { 2442686.50, 46.182504 }, //01 October 1975, UT1-UTC= 0.0014959, Accumulated Leap Seconds=14
  { 2442687.50, 46.185695 }, //02 October 1975, UT1-UTC=-0.0016954, Accumulated Leap Seconds=14
  { 2442688.50, 46.189097 }, //03 October 1975, UT1-UTC=-0.0050966, Accumulated Leap Seconds=14
  { 2442689.50, 46.192624 }, //04 October 1975, UT1-UTC=-0.0086242, Accumulated Leap Seconds=14
  { 2442690.50, 46.196161 }, //05 October 1975, UT1-UTC=-0.0121610, Accumulated Leap Seconds=14
  { 2442691.50, 46.199591 }, //06 October 1975, UT1-UTC=-0.0155912, Accumulated Leap Seconds=14
  { 2442692.50, 46.202837 }, //07 October 1975, UT1-UTC=-0.0188373, Accumulated Leap Seconds=14
  { 2442693.50, 46.205881 }, //08 October 1975, UT1-UTC=-0.0218807, Accumulated Leap Seconds=14
  { 2442694.50, 46.208758 }, //09 October 1975, UT1-UTC=-0.0247580, Accumulated Leap Seconds=14
  { 2442695.50, 46.211538 }, //10 October 1975, UT1-UTC=-0.0275378, Accumulated Leap Seconds=14
  { 2442696.50, 46.214293 }, //11 October 1975, UT1-UTC=-0.0302926, Accumulated Leap Seconds=14
  { 2442697.50, 46.217079 }, //12 October 1975, UT1-UTC=-0.0330788, Accumulated Leap Seconds=14
  { 2442698.50, 46.219930 }, //13 October 1975, UT1-UTC=-0.0359304, Accumulated Leap Seconds=14
  { 2442699.50, 46.222862 }, //14 October 1975, UT1-UTC=-0.0388617, Accumulated Leap Seconds=14
  { 2442700.50, 46.225872 }, //15 October 1975, UT1-UTC=-0.0418719, Accumulated Leap Seconds=14
  { 2442701.50, 46.228947 }, //16 October 1975, UT1-UTC=-0.0449471, Accumulated Leap Seconds=14
  { 2442702.50, 46.232062 }, //17 October 1975, UT1-UTC=-0.0480620, Accumulated Leap Seconds=14
  { 2442703.50, 46.235184 }, //18 October 1975, UT1-UTC=-0.0511839, Accumulated Leap Seconds=14
  { 2442704.50, 46.238279 }, //19 October 1975, UT1-UTC=-0.0542794, Accumulated Leap Seconds=14
  { 2442705.50, 46.241321 }, //20 October 1975, UT1-UTC=-0.0573205, Accumulated Leap Seconds=14
  { 2442706.50, 46.244290 }, //21 October 1975, UT1-UTC=-0.0602901, Accumulated Leap Seconds=14
  { 2442707.50, 46.247184 }, //22 October 1975, UT1-UTC=-0.0631842, Accumulated Leap Seconds=14
  { 2442708.50, 46.250014 }, //23 October 1975, UT1-UTC=-0.0660136, Accumulated Leap Seconds=14
  { 2442709.50, 46.252805 }, //24 October 1975, UT1-UTC=-0.0688051, Accumulated Leap Seconds=14
  { 2442710.50, 46.255601 }, //25 October 1975, UT1-UTC=-0.0716014, Accumulated Leap Seconds=14
  { 2442711.50, 46.258457 }, //26 October 1975, UT1-UTC=-0.0744569, Accumulated Leap Seconds=14
  { 2442712.50, 46.261431 }, //27 October 1975, UT1-UTC=-0.0774306, Accumulated Leap Seconds=14
  { 2442713.50, 46.264574 }, //28 October 1975, UT1-UTC=-0.0805737, Accumulated Leap Seconds=14
  { 2442714.50, 46.267916 }, //29 October 1975, UT1-UTC=-0.0839158, Accumulated Leap Seconds=14
  { 2442715.50, 46.271450 }, //30 October 1975, UT1-UTC=-0.0874502, Accumulated Leap Seconds=14
  { 2442716.50, 46.275126 }, //31 October 1975, UT1-UTC=-0.0911263, Accumulated Leap Seconds=14
  { 2442717.50, 46.278856 }, //01 November 1975, UT1-UTC=-0.0948561, Accumulated Leap Seconds=14
  { 2442718.50, 46.282534 }, //02 November 1975, UT1-UTC=-0.0985343, Accumulated Leap Seconds=14
  { 2442719.50, 46.286070 }, //03 November 1975, UT1-UTC=-0.1020696, Accumulated Leap Seconds=14
  { 2442720.50, 46.289412 }, //04 November 1975, UT1-UTC=-0.1054123, Accumulated Leap Seconds=14
  { 2442721.50, 46.292567 }, //05 November 1975, UT1-UTC=-0.1085673, Accumulated Leap Seconds=14
  { 2442722.50, 46.295585 }, //06 November 1975, UT1-UTC=-0.1115847, Accumulated Leap Seconds=14
  { 2442723.50, 46.298535 }, //07 November 1975, UT1-UTC=-0.1145351, Accumulated Leap Seconds=14
  { 2442724.50, 46.301483 }, //08 November 1975, UT1-UTC=-0.1174833, Accumulated Leap Seconds=14
  { 2442725.50, 46.304472 }, //09 November 1975, UT1-UTC=-0.1204725, Accumulated Leap Seconds=14
  { 2442726.50, 46.307521 }, //10 November 1975, UT1-UTC=-0.1235210, Accumulated Leap Seconds=14
  { 2442727.50, 46.310628 }, //11 November 1975, UT1-UTC=-0.1266285, Accumulated Leap Seconds=14
  { 2442728.50, 46.313782 }, //12 November 1975, UT1-UTC=-0.1297818, Accumulated Leap Seconds=14
  { 2442729.50, 46.316958 }, //13 November 1975, UT1-UTC=-0.1329585, Accumulated Leap Seconds=14
  { 2442730.50, 46.320129 }, //14 November 1975, UT1-UTC=-0.1361291, Accumulated Leap Seconds=14
  { 2442731.50, 46.323261 }, //15 November 1975, UT1-UTC=-0.1392609, Accumulated Leap Seconds=14
  { 2442732.50, 46.326325 }, //16 November 1975, UT1-UTC=-0.1423246, Accumulated Leap Seconds=14
  { 2442733.50, 46.329299 }, //17 November 1975, UT1-UTC=-0.1452995, Accumulated Leap Seconds=14
  { 2442734.50, 46.332178 }, //18 November 1975, UT1-UTC=-0.1481779, Accumulated Leap Seconds=14
  { 2442735.50, 46.334966 }, //19 November 1975, UT1-UTC=-0.1509664, Accumulated Leap Seconds=14
  { 2442736.50, 46.337687 }, //20 November 1975, UT1-UTC=-0.1536873, Accumulated Leap Seconds=14
  { 2442737.50, 46.340378 }, //21 November 1975, UT1-UTC=-0.1563780, Accumulated Leap Seconds=14
  { 2442738.50, 46.343088 }, //22 November 1975, UT1-UTC=-0.1590882, Accumulated Leap Seconds=14
  { 2442739.50, 46.345873 }, //23 November 1975, UT1-UTC=-0.1618727, Accumulated Leap Seconds=14
  { 2442740.50, 46.348779 }, //24 November 1975, UT1-UTC=-0.1647790, Accumulated Leap Seconds=14
  { 2442741.50, 46.351835 }, //25 November 1975, UT1-UTC=-0.1678355, Accumulated Leap Seconds=14
  { 2442742.50, 46.355041 }, //26 November 1975, UT1-UTC=-0.1710407, Accumulated Leap Seconds=14
  { 2442743.50, 46.358360 }, //27 November 1975, UT1-UTC=-0.1743596, Accumulated Leap Seconds=14
  { 2442744.50, 46.361727 }, //28 November 1975, UT1-UTC=-0.1777271, Accumulated Leap Seconds=14
  { 2442745.50, 46.365061 }, //29 November 1975, UT1-UTC=-0.1810612, Accumulated Leap Seconds=14
  { 2442746.50, 46.368282 }, //30 November 1975, UT1-UTC=-0.1842822, Accumulated Leap Seconds=14
  { 2442747.50, 46.371335 }, //01 December 1975, UT1-UTC=-0.1873351, Accumulated Leap Seconds=14
  { 2442748.50, 46.374206 }, //02 December 1975, UT1-UTC=-0.1902060, Accumulated Leap Seconds=14
  { 2442749.50, 46.376924 }, //03 December 1975, UT1-UTC=-0.1929245, Accumulated Leap Seconds=14
  { 2442750.50, 46.379551 }, //04 December 1975, UT1-UTC=-0.1955510, Accumulated Leap Seconds=14
  { 2442751.50, 46.382154 }, //05 December 1975, UT1-UTC=-0.1981536, Accumulated Leap Seconds=14
  { 2442752.50, 46.384787 }, //06 December 1975, UT1-UTC=-0.2007866, Accumulated Leap Seconds=14
  { 2442753.50, 46.387479 }, //07 December 1975, UT1-UTC=-0.2034792, Accumulated Leap Seconds=14
  { 2442754.50, 46.390236 }, //08 December 1975, UT1-UTC=-0.2062362, Accumulated Leap Seconds=14
  { 2442755.50, 46.393046 }, //09 December 1975, UT1-UTC=-0.2090458, Accumulated Leap Seconds=14
  { 2442756.50, 46.395886 }, //10 December 1975, UT1-UTC=-0.2118865, Accumulated Leap Seconds=14
  { 2442757.50, 46.398732 }, //11 December 1975, UT1-UTC=-0.2147317, Accumulated Leap Seconds=14
  { 2442758.50, 46.401552 }, //12 December 1975, UT1-UTC=-0.2175525, Accumulated Leap Seconds=14
  { 2442759.50, 46.404321 }, //13 December 1975, UT1-UTC=-0.2203215, Accumulated Leap Seconds=14
  { 2442760.50, 46.407018 }, //14 December 1975, UT1-UTC=-0.2230177, Accumulated Leap Seconds=14
  { 2442761.50, 46.409631 }, //15 December 1975, UT1-UTC=-0.2256313, Accumulated Leap Seconds=14
  { 2442762.50, 46.412167 }, //16 December 1975, UT1-UTC=-0.2281667, Accumulated Leap Seconds=14
  { 2442763.50, 46.414643 }, //17 December 1975, UT1-UTC=-0.2306432, Accumulated Leap Seconds=14
  { 2442764.50, 46.417096 }, //18 December 1975, UT1-UTC=-0.2330956, Accumulated Leap Seconds=14
  { 2442765.50, 46.419572 }, //19 December 1975, UT1-UTC=-0.2355722, Accumulated Leap Seconds=14
  { 2442766.50, 46.422129 }, //20 December 1975, UT1-UTC=-0.2381288, Accumulated Leap Seconds=14
  { 2442767.50, 46.424817 }, //21 December 1975, UT1-UTC=-0.2408168, Accumulated Leap Seconds=14
  { 2442768.50, 46.427668 }, //22 December 1975, UT1-UTC=-0.2436677, Accumulated Leap Seconds=14
  { 2442769.50, 46.430680 }, //23 December 1975, UT1-UTC=-0.2466803, Accumulated Leap Seconds=14
  { 2442770.50, 46.433817 }, //24 December 1975, UT1-UTC=-0.2498172, Accumulated Leap Seconds=14
  { 2442771.50, 46.437012 }, //25 December 1975, UT1-UTC=-0.2530125, Accumulated Leap Seconds=14
  { 2442772.50, 46.440188 }, //26 December 1975, UT1-UTC=-0.2561882, Accumulated Leap Seconds=14
  { 2442773.50, 46.443272 }, //27 December 1975, UT1-UTC=-0.2592722, Accumulated Leap Seconds=14
  { 2442774.50, 46.446214 }, //28 December 1975, UT1-UTC=-0.2622145, Accumulated Leap Seconds=14
  { 2442775.50, 46.448998 }, //29 December 1975, UT1-UTC=-0.2649982, Accumulated Leap Seconds=14
  { 2442776.50, 46.451642 }, //30 December 1975, UT1-UTC=-0.2676425, Accumulated Leap Seconds=14
  { 2442777.50, 46.454196 }, //31 December 1975, UT1-UTC=-0.2701957, Accumulated Leap Seconds=14
  { 2442778.50, 46.456721 }, //01 January 1976, UT1-UTC= 0.7272793, Accumulated Leap Seconds=15
  { 2442779.50, 46.459276 }, //02 January 1976, UT1-UTC= 0.7247235, Accumulated Leap Seconds=15
  { 2442780.50, 46.461902 }, //03 January 1976, UT1-UTC= 0.7220977, Accumulated Leap Seconds=15
  { 2442781.50, 46.464612 }, //04 January 1976, UT1-UTC= 0.7193884, Accumulated Leap Seconds=15
  { 2442782.50, 46.467395 }, //05 January 1976, UT1-UTC= 0.7166048, Accumulated Leap Seconds=15
  { 2442783.50, 46.470230 }, //06 January 1976, UT1-UTC= 0.7137704, Accumulated Leap Seconds=15
  { 2442784.50, 46.473085 }, //07 January 1976, UT1-UTC= 0.7109152, Accumulated Leap Seconds=15
  { 2442785.50, 46.475930 }, //08 January 1976, UT1-UTC= 0.7080704, Accumulated Leap Seconds=15
  { 2442786.50, 46.478735 }, //09 January 1976, UT1-UTC= 0.7052651, Accumulated Leap Seconds=15
  { 2442787.50, 46.481477 }, //10 January 1976, UT1-UTC= 0.7025227, Accumulated Leap Seconds=15
  { 2442788.50, 46.484144 }, //11 January 1976, UT1-UTC= 0.6998561, Accumulated Leap Seconds=15
  { 2442789.50, 46.486735 }, //12 January 1976, UT1-UTC= 0.6972650, Accumulated Leap Seconds=15
  { 2442790.50, 46.489267 }, //13 January 1976, UT1-UTC= 0.6947333, Accumulated Leap Seconds=15
  { 2442791.50, 46.491770 }, //14 January 1976, UT1-UTC= 0.6922296, Accumulated Leap Seconds=15
  { 2442792.50, 46.494293 }, //15 January 1976, UT1-UTC= 0.6897067, Accumulated Leap Seconds=15
  { 2442793.50, 46.496894 }, //16 January 1976, UT1-UTC= 0.6871063, Accumulated Leap Seconds=15
  { 2442794.50, 46.499632 }, //17 January 1976, UT1-UTC= 0.6843682, Accumulated Leap Seconds=15
  { 2442795.50, 46.502553 }, //18 January 1976, UT1-UTC= 0.6814470, Accumulated Leap Seconds=15
  { 2442796.50, 46.505670 }, //19 January 1976, UT1-UTC= 0.6783304, Accumulated Leap Seconds=15
  { 2442797.50, 46.508949 }, //20 January 1976, UT1-UTC= 0.6750506, Accumulated Leap Seconds=15
  { 2442798.50, 46.512320 }, //21 January 1976, UT1-UTC= 0.6716801, Accumulated Leap Seconds=15
  { 2442799.50, 46.515689 }, //22 January 1976, UT1-UTC= 0.6683108, Accumulated Leap Seconds=15
  { 2442800.50, 46.518971 }, //23 January 1976, UT1-UTC= 0.6650293, Accumulated Leap Seconds=15
  { 2442801.50, 46.522106 }, //24 January 1976, UT1-UTC= 0.6618941, Accumulated Leap Seconds=15
  { 2442802.50, 46.525075 }, //25 January 1976, UT1-UTC= 0.6589248, Accumulated Leap Seconds=15
  { 2442803.50, 46.527896 }, //26 January 1976, UT1-UTC= 0.6561040, Accumulated Leap Seconds=15
  { 2442804.50, 46.530614 }, //27 January 1976, UT1-UTC= 0.6533864, Accumulated Leap Seconds=15
  { 2442805.50, 46.533288 }, //28 January 1976, UT1-UTC= 0.6507125, Accumulated Leap Seconds=15
  { 2442806.50, 46.535977 }, //29 January 1976, UT1-UTC= 0.6480234, Accumulated Leap Seconds=15
  { 2442807.50, 46.538727 }, //30 January 1976, UT1-UTC= 0.6452727, Accumulated Leap Seconds=15
  { 2442808.50, 46.541564 }, //31 January 1976, UT1-UTC= 0.6424362, Accumulated Leap Seconds=15
  { 2442809.50, 46.544486 }, //01 February 1976, UT1-UTC= 0.6395140, Accumulated Leap Seconds=15
  { 2442810.50, 46.547474 }, //02 February 1976, UT1-UTC= 0.6365256, Accumulated Leap Seconds=15
  { 2442811.50, 46.550497 }, //03 February 1976, UT1-UTC= 0.6335025, Accumulated Leap Seconds=15
  { 2442812.50, 46.553520 }, //04 February 1976, UT1-UTC= 0.6304796, Accumulated Leap Seconds=15
  { 2442813.50, 46.556510 }, //05 February 1976, UT1-UTC= 0.6274900, Accumulated Leap Seconds=15
  { 2442814.50, 46.559439 }, //06 February 1976, UT1-UTC= 0.6245606, Accumulated Leap Seconds=15
  { 2442815.50, 46.562291 }, //07 February 1976, UT1-UTC= 0.6217091, Accumulated Leap Seconds=15
  { 2442816.50, 46.565059 }, //08 February 1976, UT1-UTC= 0.6189407, Accumulated Leap Seconds=15
  { 2442817.50, 46.567754 }, //09 February 1976, UT1-UTC= 0.6162458, Accumulated Leap Seconds=15
  { 2442818.50, 46.570401 }, //10 February 1976, UT1-UTC= 0.6135993, Accumulated Leap Seconds=15
  { 2442819.50, 46.573039 }, //11 February 1976, UT1-UTC= 0.6109607, Accumulated Leap Seconds=15
  { 2442820.50, 46.575724 }, //12 February 1976, UT1-UTC= 0.6082763, Accumulated Leap Seconds=15
  { 2442821.50, 46.578516 }, //13 February 1976, UT1-UTC= 0.6054842, Accumulated Leap Seconds=15
  { 2442822.50, 46.581474 }, //14 February 1976, UT1-UTC= 0.6025261, Accumulated Leap Seconds=15
  { 2442823.50, 46.584633 }, //15 February 1976, UT1-UTC= 0.5993671, Accumulated Leap Seconds=15
  { 2442824.50, 46.587985 }, //16 February 1976, UT1-UTC= 0.5960145, Accumulated Leap Seconds=15
  { 2442825.50, 46.591475 }, //17 February 1976, UT1-UTC= 0.5925249, Accumulated Leap Seconds=15
  { 2442826.50, 46.595007 }, //18 February 1976, UT1-UTC= 0.5889930, Accumulated Leap Seconds=15
  { 2442827.50, 46.598476 }, //19 February 1976, UT1-UTC= 0.5855239, Accumulated Leap Seconds=15
  { 2442828.50, 46.601799 }, //20 February 1976, UT1-UTC= 0.5822010, Accumulated Leap Seconds=15
  { 2442829.50, 46.604937 }, //21 February 1976, UT1-UTC= 0.5790631, Accumulated Leap Seconds=15
  { 2442830.50, 46.607899 }, //22 February 1976, UT1-UTC= 0.5761012, Accumulated Leap Seconds=15
  { 2442831.50, 46.610731 }, //23 February 1976, UT1-UTC= 0.5732693, Accumulated Leap Seconds=15
  { 2442832.50, 46.613496 }, //24 February 1976, UT1-UTC= 0.5705043, Accumulated Leap Seconds=15
  { 2442833.50, 46.616257 }, //25 February 1976, UT1-UTC= 0.5677434, Accumulated Leap Seconds=15
  { 2442834.50, 46.619064 }, //26 February 1976, UT1-UTC= 0.5649360, Accumulated Leap Seconds=15
  { 2442835.50, 46.621950 }, //27 February 1976, UT1-UTC= 0.5620501, Accumulated Leap Seconds=15
  { 2442836.50, 46.624925 }, //28 February 1976, UT1-UTC= 0.5590749, Accumulated Leap Seconds=15
  { 2442837.50, 46.627980 }, //29 February 1976, UT1-UTC= 0.5560198, Accumulated Leap Seconds=15
  { 2442838.50, 46.631090 }, //01 March 1976, UT1-UTC= 0.5529101, Accumulated Leap Seconds=15
  { 2442839.50, 46.634220 }, //02 March 1976, UT1-UTC= 0.5497798, Accumulated Leap Seconds=15
  { 2442840.50, 46.637336 }, //03 March 1976, UT1-UTC= 0.5466638, Accumulated Leap Seconds=15
  { 2442841.50, 46.640408 }, //04 March 1976, UT1-UTC= 0.5435922, Accumulated Leap Seconds=15
  { 2442842.50, 46.643414 }, //05 March 1976, UT1-UTC= 0.5405861, Accumulated Leap Seconds=15
  { 2442843.50, 46.646345 }, //06 March 1976, UT1-UTC= 0.5376552, Accumulated Leap Seconds=15
  { 2442844.50, 46.649204 }, //07 March 1976, UT1-UTC= 0.5347955, Accumulated Leap Seconds=15
  { 2442845.50, 46.652012 }, //08 March 1976, UT1-UTC= 0.5319882, Accumulated Leap Seconds=15
  { 2442846.50, 46.654800 }, //09 March 1976, UT1-UTC= 0.5292000, Accumulated Leap Seconds=15
  { 2442847.50, 46.657617 }, //10 March 1976, UT1-UTC= 0.5263832, Accumulated Leap Seconds=15
  { 2442848.50, 46.660520 }, //11 March 1976, UT1-UTC= 0.5234797, Accumulated Leap Seconds=15
  { 2442849.50, 46.663573 }, //12 March 1976, UT1-UTC= 0.5204274, Accumulated Leap Seconds=15
  { 2442850.50, 46.666824 }, //13 March 1976, UT1-UTC= 0.5171755, Accumulated Leap Seconds=15
  { 2442851.50, 46.670297 }, //14 March 1976, UT1-UTC= 0.5137031, Accumulated Leap Seconds=15
  { 2442852.50, 46.673964 }, //15 March 1976, UT1-UTC= 0.5100364, Accumulated Leap Seconds=15
  { 2442853.50, 46.677748 }, //16 March 1976, UT1-UTC= 0.5062517, Accumulated Leap Seconds=15
  { 2442854.50, 46.681541 }, //17 March 1976, UT1-UTC= 0.5024587, Accumulated Leap Seconds=15
  { 2442855.50, 46.685233 }, //18 March 1976, UT1-UTC= 0.4987669, Accumulated Leap Seconds=15
  { 2442856.50, 46.688749 }, //19 March 1976, UT1-UTC= 0.4952510, Accumulated Leap Seconds=15
  { 2442857.50, 46.692069 }, //20 March 1976, UT1-UTC= 0.4919312, Accumulated Leap Seconds=15
  { 2442858.50, 46.695224 }, //21 March 1976, UT1-UTC= 0.4887763, Accumulated Leap Seconds=15
  { 2442859.50, 46.698276 }, //22 March 1976, UT1-UTC= 0.4857239, Accumulated Leap Seconds=15
  { 2442860.50, 46.701294 }, //23 March 1976, UT1-UTC= 0.4827058, Accumulated Leap Seconds=15
  { 2442861.50, 46.704333 }, //24 March 1976, UT1-UTC= 0.4796666, Accumulated Leap Seconds=15
  { 2442862.50, 46.707430 }, //25 March 1976, UT1-UTC= 0.4765704, Accumulated Leap Seconds=15
  { 2442863.50, 46.710598 }, //26 March 1976, UT1-UTC= 0.4734018, Accumulated Leap Seconds=15
  { 2442864.50, 46.713837 }, //27 March 1976, UT1-UTC= 0.4701631, Accumulated Leap Seconds=15
  { 2442865.50, 46.717129 }, //28 March 1976, UT1-UTC= 0.4668712, Accumulated Leap Seconds=15
  { 2442866.50, 46.720445 }, //29 March 1976, UT1-UTC= 0.4635546, Accumulated Leap Seconds=15
  { 2442867.50, 46.723753 }, //30 March 1976, UT1-UTC= 0.4602468, Accumulated Leap Seconds=15
  { 2442868.50, 46.727021 }, //31 March 1976, UT1-UTC= 0.4569794, Accumulated Leap Seconds=15
  { 2442869.50, 46.730223 }, //01 April 1976, UT1-UTC= 0.4537769, Accumulated Leap Seconds=15
  { 2442870.50, 46.733348 }, //02 April 1976, UT1-UTC= 0.4506523, Accumulated Leap Seconds=15
  { 2442871.50, 46.736394 }, //03 April 1976, UT1-UTC= 0.4476063, Accumulated Leap Seconds=15
  { 2442872.50, 46.739374 }, //04 April 1976, UT1-UTC= 0.4446260, Accumulated Leap Seconds=15
  { 2442873.50, 46.742315 }, //05 April 1976, UT1-UTC= 0.4416849, Accumulated Leap Seconds=15
  { 2442874.50, 46.745256 }, //06 April 1976, UT1-UTC= 0.4387441, Accumulated Leap Seconds=15
  { 2442875.50, 46.748247 }, //07 April 1976, UT1-UTC= 0.4357534, Accumulated Leap Seconds=15
  { 2442876.50, 46.751343 }, //08 April 1976, UT1-UTC= 0.4326566, Accumulated Leap Seconds=15
  { 2442877.50, 46.754600 }, //09 April 1976, UT1-UTC= 0.4294003, Accumulated Leap Seconds=15
  { 2442878.50, 46.758052 }, //10 April 1976, UT1-UTC= 0.4259475, Accumulated Leap Seconds=15
  { 2442879.50, 46.761705 }, //11 April 1976, UT1-UTC= 0.4222945, Accumulated Leap Seconds=15
  { 2442880.50, 46.765517 }, //12 April 1976, UT1-UTC= 0.4184833, Accumulated Leap Seconds=15
  { 2442881.50, 46.769401 }, //13 April 1976, UT1-UTC= 0.4145990, Accumulated Leap Seconds=15
  { 2442882.50, 46.773250 }, //14 April 1976, UT1-UTC= 0.4107499, Accumulated Leap Seconds=15
  { 2442883.50, 46.776966 }, //15 April 1976, UT1-UTC= 0.4070336, Accumulated Leap Seconds=15
  { 2442884.50, 46.780494 }, //16 April 1976, UT1-UTC= 0.4035056, Accumulated Leap Seconds=15
  { 2442885.50, 46.783835 }, //17 April 1976, UT1-UTC= 0.4001653, Accumulated Leap Seconds=15
  { 2442886.50, 46.787035 }, //18 April 1976, UT1-UTC= 0.3969651, Accumulated Leap Seconds=15
  { 2442887.50, 46.790164 }, //19 April 1976, UT1-UTC= 0.3938361, Accumulated Leap Seconds=15
  { 2442888.50, 46.793285 }, //20 April 1976, UT1-UTC= 0.3907146, Accumulated Leap Seconds=15
  { 2442889.50, 46.796442 }, //21 April 1976, UT1-UTC= 0.3875584, Accumulated Leap Seconds=15
  { 2442890.50, 46.799650 }, //22 April 1976, UT1-UTC= 0.3843495, Accumulated Leap Seconds=15
  { 2442891.50, 46.802912 }, //23 April 1976, UT1-UTC= 0.3810883, Accumulated Leap Seconds=15
  { 2442892.50, 46.806211 }, //24 April 1976, UT1-UTC= 0.3777885, Accumulated Leap Seconds=15
  { 2442893.50, 46.809527 }, //25 April 1976, UT1-UTC= 0.3744727, Accumulated Leap Seconds=15
  { 2442894.50, 46.812830 }, //26 April 1976, UT1-UTC= 0.3711696, Accumulated Leap Seconds=15
  { 2442895.50, 46.816091 }, //27 April 1976, UT1-UTC= 0.3679092, Accumulated Leap Seconds=15
  { 2442896.50, 46.819283 }, //28 April 1976, UT1-UTC= 0.3647171, Accumulated Leap Seconds=15
  { 2442897.50, 46.822391 }, //29 April 1976, UT1-UTC= 0.3616088, Accumulated Leap Seconds=15
  { 2442898.50, 46.825413 }, //30 April 1976, UT1-UTC= 0.3585869, Accumulated Leap Seconds=15
  { 2442899.50, 46.828359 }, //01 May 1976, UT1-UTC= 0.3556412, Accumulated Leap Seconds=15
  { 2442900.50, 46.831251 }, //02 May 1976, UT1-UTC= 0.3527487, Accumulated Leap Seconds=15
  { 2442901.50, 46.834125 }, //03 May 1976, UT1-UTC= 0.3498753, Accumulated Leap Seconds=15
  { 2442902.50, 46.837023 }, //04 May 1976, UT1-UTC= 0.3469771, Accumulated Leap Seconds=15
  { 2442903.50, 46.839995 }, //05 May 1976, UT1-UTC= 0.3440048, Accumulated Leap Seconds=15
  { 2442904.50, 46.843090 }, //06 May 1976, UT1-UTC= 0.3409099, Accumulated Leap Seconds=15
  { 2442905.50, 46.846345 }, //07 May 1976, UT1-UTC= 0.3376548, Accumulated Leap Seconds=15
  { 2442906.50, 46.849776 }, //08 May 1976, UT1-UTC= 0.3342237, Accumulated Leap Seconds=15
  { 2442907.50, 46.853366 }, //09 May 1976, UT1-UTC= 0.3306338, Accumulated Leap Seconds=15
  { 2442908.50, 46.857060 }, //10 May 1976, UT1-UTC= 0.3269395, Accumulated Leap Seconds=15
  { 2442909.50, 46.860775 }, //11 May 1976, UT1-UTC= 0.3232252, Accumulated Leap Seconds=15
  { 2442910.50, 46.864416 }, //12 May 1976, UT1-UTC= 0.3195839, Accumulated Leap Seconds=15
  { 2442911.50, 46.867912 }, //13 May 1976, UT1-UTC= 0.3160883, Accumulated Leap Seconds=15
  { 2442912.50, 46.871233 }, //14 May 1976, UT1-UTC= 0.3127673, Accumulated Leap Seconds=15
  { 2442913.50, 46.874401 }, //15 May 1976, UT1-UTC= 0.3095987, Accumulated Leap Seconds=15
  { 2442914.50, 46.877476 }, //16 May 1976, UT1-UTC= 0.3065239, Accumulated Leap Seconds=15
  { 2442915.50, 46.880525 }, //17 May 1976, UT1-UTC= 0.3034746, Accumulated Leap Seconds=15
  { 2442916.50, 46.883602 }, //18 May 1976, UT1-UTC= 0.3003976, Accumulated Leap Seconds=15
  { 2442917.50, 46.886733 }, //19 May 1976, UT1-UTC= 0.2972669, Accumulated Leap Seconds=15
  { 2442918.50, 46.889918 }, //20 May 1976, UT1-UTC= 0.2940822, Accumulated Leap Seconds=15
  { 2442919.50, 46.893138 }, //21 May 1976, UT1-UTC= 0.2908622, Accumulated Leap Seconds=15
  { 2442920.50, 46.896365 }, //22 May 1976, UT1-UTC= 0.2876346, Accumulated Leap Seconds=15
  { 2442921.50, 46.899569 }, //23 May 1976, UT1-UTC= 0.2844308, Accumulated Leap Seconds=15
  { 2442922.50, 46.902717 }, //24 May 1976, UT1-UTC= 0.2812826, Accumulated Leap Seconds=15
  { 2442923.50, 46.905782 }, //25 May 1976, UT1-UTC= 0.2782184, Accumulated Leap Seconds=15
  { 2442924.50, 46.908741 }, //26 May 1976, UT1-UTC= 0.2752588, Accumulated Leap Seconds=15
  { 2442925.50, 46.911588 }, //27 May 1976, UT1-UTC= 0.2724120, Accumulated Leap Seconds=15
  { 2442926.50, 46.914327 }, //28 May 1976, UT1-UTC= 0.2696726, Accumulated Leap Seconds=15
  { 2442927.50, 46.916978 }, //29 May 1976, UT1-UTC= 0.2670220, Accumulated Leap Seconds=15
  { 2442928.50, 46.919570 }, //30 May 1976, UT1-UTC= 0.2644299, Accumulated Leap Seconds=15
  { 2442929.50, 46.922144 }, //31 May 1976, UT1-UTC= 0.2618564, Accumulated Leap Seconds=15
  { 2442930.50, 46.924744 }, //01 June 1976, UT1-UTC= 0.2592557, Accumulated Leap Seconds=15
  { 2442931.50, 46.927417 }, //02 June 1976, UT1-UTC= 0.2565833, Accumulated Leap Seconds=15
  { 2442932.50, 46.930195 }, //03 June 1976, UT1-UTC= 0.2538051, Accumulated Leap Seconds=15
  { 2442933.50, 46.933093 }, //04 June 1976, UT1-UTC= 0.2509074, Accumulated Leap Seconds=15
  { 2442934.50, 46.936097 }, //05 June 1976, UT1-UTC= 0.2479026, Accumulated Leap Seconds=15
  { 2442935.50, 46.939168 }, //06 June 1976, UT1-UTC= 0.2448319, Accumulated Leap Seconds=15
  { 2442936.50, 46.942240 }, //07 June 1976, UT1-UTC= 0.2417603, Accumulated Leap Seconds=15
  { 2442937.50, 46.945236 }, //08 June 1976, UT1-UTC= 0.2387643, Accumulated Leap Seconds=15
  { 2442938.50, 46.948089 }, //09 June 1976, UT1-UTC= 0.2359114, Accumulated Leap Seconds=15
  { 2442939.50, 46.950759 }, //10 June 1976, UT1-UTC= 0.2332408, Accumulated Leap Seconds=15
  { 2442940.50, 46.953250 }, //11 June 1976, UT1-UTC= 0.2307498, Accumulated Leap Seconds=15
  { 2442941.50, 46.955604 }, //12 June 1976, UT1-UTC= 0.2283957, Accumulated Leap Seconds=15
  { 2442942.50, 46.957887 }, //13 June 1976, UT1-UTC= 0.2261128, Accumulated Leap Seconds=15
  { 2442943.50, 46.960162 }, //14 June 1976, UT1-UTC= 0.2238375, Accumulated Leap Seconds=15
  { 2442944.50, 46.962471 }, //15 June 1976, UT1-UTC= 0.2215288, Accumulated Leap Seconds=15
  { 2442945.50, 46.964825 }, //16 June 1976, UT1-UTC= 0.2191751, Accumulated Leap Seconds=15
  { 2442946.50, 46.967211 }, //17 June 1976, UT1-UTC= 0.2167889, Accumulated Leap Seconds=15
  { 2442947.50, 46.969605 }, //18 June 1976, UT1-UTC= 0.2143953, Accumulated Leap Seconds=15
  { 2442948.50, 46.971977 }, //19 June 1976, UT1-UTC= 0.2120227, Accumulated Leap Seconds=15
  { 2442949.50, 46.974302 }, //20 June 1976, UT1-UTC= 0.2096983, Accumulated Leap Seconds=15
  { 2442950.50, 46.976554 }, //21 June 1976, UT1-UTC= 0.2074462, Accumulated Leap Seconds=15
  { 2442951.50, 46.978716 }, //22 June 1976, UT1-UTC= 0.2052841, Accumulated Leap Seconds=15
  { 2442952.50, 46.980781 }, //23 June 1976, UT1-UTC= 0.2032185, Accumulated Leap Seconds=15
  { 2442953.50, 46.982757 }, //24 June 1976, UT1-UTC= 0.2012426, Accumulated Leap Seconds=15
  { 2442954.50, 46.984665 }, //25 June 1976, UT1-UTC= 0.1993349, Accumulated Leap Seconds=15
  { 2442955.50, 46.986539 }, //26 June 1976, UT1-UTC= 0.1974614, Accumulated Leap Seconds=15
  { 2442956.50, 46.988422 }, //27 June 1976, UT1-UTC= 0.1955782, Accumulated Leap Seconds=15
  { 2442957.50, 46.990364 }, //28 June 1976, UT1-UTC= 0.1936356, Accumulated Leap Seconds=15
  { 2442958.50, 46.992416 }, //29 June 1976, UT1-UTC= 0.1915842, Accumulated Leap Seconds=15
  { 2442959.50, 46.994614 }, //30 June 1976, UT1-UTC= 0.1893860, Accumulated Leap Seconds=15
  { 2442960.50, 46.996976 }, //01 July 1976, UT1-UTC= 0.1870243, Accumulated Leap Seconds=15
  { 2442961.50, 46.999488 }, //02 July 1976, UT1-UTC= 0.1845115, Accumulated Leap Seconds=15
  { 2442962.50, 47.002111 }, //03 July 1976, UT1-UTC= 0.1818895, Accumulated Leap Seconds=15
  { 2442963.50, 47.004779 }, //04 July 1976, UT1-UTC= 0.1792212, Accumulated Leap Seconds=15
  { 2442964.50, 47.007423 }, //05 July 1976, UT1-UTC= 0.1765765, Accumulated Leap Seconds=15
  { 2442965.50, 47.009984 }, //06 July 1976, UT1-UTC= 0.1740161, Accumulated Leap Seconds=15
  { 2442966.50, 47.012422 }, //07 July 1976, UT1-UTC= 0.1715776, Accumulated Leap Seconds=15
  { 2442967.50, 47.014734 }, //08 July 1976, UT1-UTC= 0.1692657, Accumulated Leap Seconds=15
  { 2442968.50, 47.016949 }, //09 July 1976, UT1-UTC= 0.1670512, Accumulated Leap Seconds=15
  { 2442969.50, 47.019120 }, //10 July 1976, UT1-UTC= 0.1648797, Accumulated Leap Seconds=15
  { 2442970.50, 47.021310 }, //11 July 1976, UT1-UTC= 0.1626904, Accumulated Leap Seconds=15
  { 2442971.50, 47.023565 }, //12 July 1976, UT1-UTC= 0.1604353, Accumulated Leap Seconds=15
  { 2442972.50, 47.025907 }, //13 July 1976, UT1-UTC= 0.1580926, Accumulated Leap Seconds=15
  { 2442973.50, 47.028330 }, //14 July 1976, UT1-UTC= 0.1556695, Accumulated Leap Seconds=15
  { 2442974.50, 47.030806 }, //15 July 1976, UT1-UTC= 0.1531943, Accumulated Leap Seconds=15
  { 2442975.50, 47.033296 }, //16 July 1976, UT1-UTC= 0.1507039, Accumulated Leap Seconds=15
  { 2442976.50, 47.035765 }, //17 July 1976, UT1-UTC= 0.1482349, Accumulated Leap Seconds=15
  { 2442977.50, 47.038182 }, //18 July 1976, UT1-UTC= 0.1458182, Accumulated Leap Seconds=15
  { 2442978.50, 47.040523 }, //19 July 1976, UT1-UTC= 0.1434774, Accumulated Leap Seconds=15
  { 2442979.50, 47.042774 }, //20 July 1976, UT1-UTC= 0.1412263, Accumulated Leap Seconds=15
  { 2442980.50, 47.044934 }, //21 July 1976, UT1-UTC= 0.1390659, Accumulated Leap Seconds=15
  { 2442981.50, 47.047017 }, //22 July 1976, UT1-UTC= 0.1369828, Accumulated Leap Seconds=15
  { 2442982.50, 47.049051 }, //23 July 1976, UT1-UTC= 0.1349489, Accumulated Leap Seconds=15
  { 2442983.50, 47.051076 }, //24 July 1976, UT1-UTC= 0.1329242, Accumulated Leap Seconds=15
  { 2442984.50, 47.053140 }, //25 July 1976, UT1-UTC= 0.1308603, Accumulated Leap Seconds=15
  { 2442985.50, 47.055294 }, //26 July 1976, UT1-UTC= 0.1287061, Accumulated Leap Seconds=15
  { 2442986.50, 47.057584 }, //27 July 1976, UT1-UTC= 0.1264160, Accumulated Leap Seconds=15
  { 2442987.50, 47.060037 }, //28 July 1976, UT1-UTC= 0.1239628, Accumulated Leap Seconds=15
  { 2442988.50, 47.062650 }, //29 July 1976, UT1-UTC= 0.1213501, Accumulated Leap Seconds=15
  { 2442989.50, 47.065382 }, //30 July 1976, UT1-UTC= 0.1186180, Accumulated Leap Seconds=15
  { 2442990.50, 47.068164 }, //31 July 1976, UT1-UTC= 0.1158360, Accumulated Leap Seconds=15
  { 2442991.50, 47.070915 }, //01 August 1976, UT1-UTC= 0.1130852, Accumulated Leap Seconds=15
  { 2442992.50, 47.073564 }, //02 August 1976, UT1-UTC= 0.1104361, Accumulated Leap Seconds=15
  { 2442993.50, 47.076069 }, //03 August 1976, UT1-UTC= 0.1079306, Accumulated Leap Seconds=15
  { 2442994.50, 47.078425 }, //04 August 1976, UT1-UTC= 0.1055750, Accumulated Leap Seconds=15
  { 2442995.50, 47.080658 }, //05 August 1976, UT1-UTC= 0.1033424, Accumulated Leap Seconds=15
  { 2442996.50, 47.082817 }, //06 August 1976, UT1-UTC= 0.1011825, Accumulated Leap Seconds=15
  { 2442997.50, 47.084963 }, //07 August 1976, UT1-UTC= 0.0990371, Accumulated Leap Seconds=15
  { 2442998.50, 47.087145 }, //08 August 1976, UT1-UTC= 0.0968547, Accumulated Leap Seconds=15
  { 2442999.50, 47.089397 }, //09 August 1976, UT1-UTC= 0.0946034, Accumulated Leap Seconds=15
  { 2443000.50, 47.091723 }, //10 August 1976, UT1-UTC= 0.0922767, Accumulated Leap Seconds=15
  { 2443001.50, 47.094109 }, //11 August 1976, UT1-UTC= 0.0898910, Accumulated Leap Seconds=15
  { 2443002.50, 47.096521 }, //12 August 1976, UT1-UTC= 0.0874785, Accumulated Leap Seconds=15
  { 2443003.50, 47.098923 }, //13 August 1976, UT1-UTC= 0.0850769, Accumulated Leap Seconds=15
  { 2443004.50, 47.101280 }, //14 August 1976, UT1-UTC= 0.0827199, Accumulated Leap Seconds=15
  { 2443005.50, 47.103568 }, //15 August 1976, UT1-UTC= 0.0804324, Accumulated Leap Seconds=15
  { 2443006.50, 47.105772 }, //16 August 1976, UT1-UTC= 0.0782283, Accumulated Leap Seconds=15
  { 2443007.50, 47.107890 }, //17 August 1976, UT1-UTC= 0.0761102, Accumulated Leap Seconds=15
  { 2443008.50, 47.109933 }, //18 August 1976, UT1-UTC= 0.0740671, Accumulated Leap Seconds=15
  { 2443009.50, 47.111927 }, //19 August 1976, UT1-UTC= 0.0720730, Accumulated Leap Seconds=15
  { 2443010.50, 47.113911 }, //20 August 1976, UT1-UTC= 0.0700887, Accumulated Leap Seconds=15
  { 2443011.50, 47.115934 }, //21 August 1976, UT1-UTC= 0.0680656, Accumulated Leap Seconds=15
  { 2443012.50, 47.118050 }, //22 August 1976, UT1-UTC= 0.0659503, Accumulated Leap Seconds=15
  { 2443013.50, 47.120310 }, //23 August 1976, UT1-UTC= 0.0636903, Accumulated Leap Seconds=15
  { 2443014.50, 47.122756 }, //24 August 1976, UT1-UTC= 0.0612441, Accumulated Leap Seconds=15
  { 2443015.50, 47.125404 }, //25 August 1976, UT1-UTC= 0.0585955, Accumulated Leap Seconds=15
  { 2443016.50, 47.128234 }, //26 August 1976, UT1-UTC= 0.0557664, Accumulated Leap Seconds=15
  { 2443017.50, 47.131181 }, //27 August 1976, UT1-UTC= 0.0528187, Accumulated Leap Seconds=15
  { 2443018.50, 47.134159 }, //28 August 1976, UT1-UTC= 0.0498411, Accumulated Leap Seconds=15
  { 2443019.50, 47.137076 }, //29 August 1976, UT1-UTC= 0.0469235, Accumulated Leap Seconds=15
  { 2443020.50, 47.139871 }, //30 August 1976, UT1-UTC= 0.0441290, Accumulated Leap Seconds=15
  { 2443021.50, 47.142522 }, //31 August 1976, UT1-UTC= 0.0414781, Accumulated Leap Seconds=15
  { 2443022.50, 47.145052 }, //01 September 1976, UT1-UTC= 0.0389485, Accumulated Leap Seconds=15
  { 2443023.50, 47.147512 }, //02 September 1976, UT1-UTC= 0.0364884, Accumulated Leap Seconds=15
  { 2443024.50, 47.149965 }, //03 September 1976, UT1-UTC= 0.0340349, Accumulated Leap Seconds=15
  { 2443025.50, 47.152469 }, //04 September 1976, UT1-UTC= 0.0315307, Accumulated Leap Seconds=15
  { 2443026.50, 47.155065 }, //05 September 1976, UT1-UTC= 0.0289351, Accumulated Leap Seconds=15
  { 2443027.50, 47.157770 }, //06 September 1976, UT1-UTC= 0.0262297, Accumulated Leap Seconds=15
  { 2443028.50, 47.160580 }, //07 September 1976, UT1-UTC= 0.0234196, Accumulated Leap Seconds=15
  { 2443029.50, 47.163470 }, //08 September 1976, UT1-UTC= 0.0205296, Accumulated Leap Seconds=15
  { 2443030.50, 47.166404 }, //09 September 1976, UT1-UTC= 0.0175962, Accumulated Leap Seconds=15
  { 2443031.50, 47.169341 }, //10 September 1976, UT1-UTC= 0.0146585, Accumulated Leap Seconds=15
  { 2443032.50, 47.172250 }, //11 September 1976, UT1-UTC= 0.0117497, Accumulated Leap Seconds=15
  { 2443033.50, 47.175108 }, //12 September 1976, UT1-UTC= 0.0088917, Accumulated Leap Seconds=15
  { 2443034.50, 47.177906 }, //13 September 1976, UT1-UTC= 0.0060937, Accumulated Leap Seconds=15
  { 2443035.50, 47.180648 }, //14 September 1976, UT1-UTC= 0.0033520, Accumulated Leap Seconds=15
  { 2443036.50, 47.183351 }, //15 September 1976, UT1-UTC= 0.0006492, Accumulated Leap Seconds=15
  { 2443037.50, 47.186045 }, //16 September 1976, UT1-UTC=-0.0020446, Accumulated Leap Seconds=15
  { 2443038.50, 47.188770 }, //17 September 1976, UT1-UTC=-0.0047701, Accumulated Leap Seconds=15
  { 2443039.50, 47.191575 }, //18 September 1976, UT1-UTC=-0.0075751, Accumulated Leap Seconds=15
  { 2443040.50, 47.194510 }, //19 September 1976, UT1-UTC=-0.0105104, Accumulated Leap Seconds=15
  { 2443041.50, 47.197623 }, //20 September 1976, UT1-UTC=-0.0136229, Accumulated Leap Seconds=15
  { 2443042.50, 47.200944 }, //21 September 1976, UT1-UTC=-0.0169440, Accumulated Leap Seconds=15
  { 2443043.50, 47.204476 }, //22 September 1976, UT1-UTC=-0.0204755, Accumulated Leap Seconds=15
  { 2443044.50, 47.208178 }, //23 September 1976, UT1-UTC=-0.0241782, Accumulated Leap Seconds=15
  { 2443045.50, 47.211974 }, //24 September 1976, UT1-UTC=-0.0279738, Accumulated Leap Seconds=15
  { 2443046.50, 47.215763 }, //25 September 1976, UT1-UTC=-0.0317629, Accumulated Leap Seconds=15
  { 2443047.50, 47.219455 }, //26 September 1976, UT1-UTC=-0.0354553, Accumulated Leap Seconds=15
  { 2443048.50, 47.222998 }, //27 September 1976, UT1-UTC=-0.0389984, Accumulated Leap Seconds=15
  { 2443049.50, 47.226391 }, //28 September 1976, UT1-UTC=-0.0423905, Accumulated Leap Seconds=15
  { 2443050.50, 47.229673 }, //29 September 1976, UT1-UTC=-0.0456728, Accumulated Leap Seconds=15
  { 2443051.50, 47.232907 }, //30 September 1976, UT1-UTC=-0.0489075, Accumulated Leap Seconds=15
  { 2443052.50, 47.236154 }, //01 October 1976, UT1-UTC=-0.0521542, Accumulated Leap Seconds=15
  { 2443053.50, 47.239455 }, //02 October 1976, UT1-UTC=-0.0554553, Accumulated Leap Seconds=15
  { 2443054.50, 47.242830 }, //03 October 1976, UT1-UTC=-0.0588305, Accumulated Leap Seconds=15
  { 2443055.50, 47.246278 }, //04 October 1976, UT1-UTC=-0.0622778, Accumulated Leap Seconds=15
  { 2443056.50, 47.249778 }, //05 October 1976, UT1-UTC=-0.0657781, Accumulated Leap Seconds=15
  { 2443057.50, 47.253300 }, //06 October 1976, UT1-UTC=-0.0692999, Accumulated Leap Seconds=15
  { 2443058.50, 47.256806 }, //07 October 1976, UT1-UTC=-0.0728058, Accumulated Leap Seconds=15
  { 2443059.50, 47.260262 }, //08 October 1976, UT1-UTC=-0.0762616, Accumulated Leap Seconds=15
  { 2443060.50, 47.263642 }, //09 October 1976, UT1-UTC=-0.0796421, Accumulated Leap Seconds=15
  { 2443061.50, 47.266933 }, //10 October 1976, UT1-UTC=-0.0829332, Accumulated Leap Seconds=15
  { 2443062.50, 47.270135 }, //11 October 1976, UT1-UTC=-0.0861346, Accumulated Leap Seconds=15
  { 2443063.50, 47.273259 }, //12 October 1976, UT1-UTC=-0.0892595, Accumulated Leap Seconds=15
  { 2443064.50, 47.276333 }, //13 October 1976, UT1-UTC=-0.0923330, Accumulated Leap Seconds=15
  { 2443065.50, 47.279389 }, //14 October 1976, UT1-UTC=-0.0953893, Accumulated Leap Seconds=15
  { 2443066.50, 47.282470 }, //15 October 1976, UT1-UTC=-0.0984704, Accumulated Leap Seconds=15
  { 2443067.50, 47.285622 }, //16 October 1976, UT1-UTC=-0.1016220, Accumulated Leap Seconds=15
  { 2443068.50, 47.288890 }, //17 October 1976, UT1-UTC=-0.1048897, Accumulated Leap Seconds=15
  { 2443069.50, 47.292311 }, //18 October 1976, UT1-UTC=-0.1083106, Accumulated Leap Seconds=15
  { 2443070.50, 47.295904 }, //19 October 1976, UT1-UTC=-0.1119040, Accumulated Leap Seconds=15
  { 2443071.50, 47.299659 }, //20 October 1976, UT1-UTC=-0.1156586, Accumulated Leap Seconds=15
  { 2443072.50, 47.303525 }, //21 October 1976, UT1-UTC=-0.1195251, Accumulated Leap Seconds=15
  { 2443073.50, 47.307421 }, //22 October 1976, UT1-UTC=-0.1234210, Accumulated Leap Seconds=15
  { 2443074.50, 47.311252 }, //23 October 1976, UT1-UTC=-0.1272516, Accumulated Leap Seconds=15
  { 2443075.50, 47.314940 }, //24 October 1976, UT1-UTC=-0.1309398, Accumulated Leap Seconds=15
  { 2443076.50, 47.318452 }, //25 October 1976, UT1-UTC=-0.1344522, Accumulated Leap Seconds=15
  { 2443077.50, 47.321805 }, //26 October 1976, UT1-UTC=-0.1378055, Accumulated Leap Seconds=15
  { 2443078.50, 47.325053 }, //27 October 1976, UT1-UTC=-0.1410532, Accumulated Leap Seconds=15
  { 2443079.50, 47.328259 }, //28 October 1976, UT1-UTC=-0.1442595, Accumulated Leap Seconds=15
  { 2443080.50, 47.331475 }, //29 October 1976, UT1-UTC=-0.1474754, Accumulated Leap Seconds=15
  { 2443081.50, 47.334726 }, //30 October 1976, UT1-UTC=-0.1507262, Accumulated Leap Seconds=15
  { 2443082.50, 47.338014 }, //31 October 1976, UT1-UTC=-0.1540137, Accumulated Leap Seconds=15
  { 2443083.50, 47.341324 }, //01 November 1976, UT1-UTC=-0.1573241, Accumulated Leap Seconds=15
  { 2443084.50, 47.344634 }, //02 November 1976, UT1-UTC=-0.1606338, Accumulated Leap Seconds=15
  { 2443085.50, 47.347912 }, //03 November 1976, UT1-UTC=-0.1639125, Accumulated Leap Seconds=15
  { 2443086.50, 47.351129 }, //04 November 1976, UT1-UTC=-0.1671292, Accumulated Leap Seconds=15
  { 2443087.50, 47.354257 }, //05 November 1976, UT1-UTC=-0.1702568, Accumulated Leap Seconds=15
  { 2443088.50, 47.357277 }, //06 November 1976, UT1-UTC=-0.1732772, Accumulated Leap Seconds=15
  { 2443089.50, 47.360185 }, //07 November 1976, UT1-UTC=-0.1761852, Accumulated Leap Seconds=15
  { 2443090.50, 47.362990 }, //08 November 1976, UT1-UTC=-0.1789899, Accumulated Leap Seconds=15
  { 2443091.50, 47.365714 }, //09 November 1976, UT1-UTC=-0.1817139, Accumulated Leap Seconds=15
  { 2443092.50, 47.368390 }, //10 November 1976, UT1-UTC=-0.1843896, Accumulated Leap Seconds=15
  { 2443093.50, 47.371056 }, //11 November 1976, UT1-UTC=-0.1870564, Accumulated Leap Seconds=15
  { 2443094.50, 47.373758 }, //12 November 1976, UT1-UTC=-0.1897580, Accumulated Leap Seconds=15
  { 2443095.50, 47.376538 }, //13 November 1976, UT1-UTC=-0.1925376, Accumulated Leap Seconds=15
  { 2443096.50, 47.379433 }, //14 November 1976, UT1-UTC=-0.1954333, Accumulated Leap Seconds=15
  { 2443097.50, 47.382470 }, //15 November 1976, UT1-UTC=-0.1984697, Accumulated Leap Seconds=15
  { 2443098.50, 47.385650 }, //16 November 1976, UT1-UTC=-0.2016502, Accumulated Leap Seconds=15
  { 2443099.50, 47.388949 }, //17 November 1976, UT1-UTC=-0.2049494, Accumulated Leap Seconds=15
  { 2443100.50, 47.392314 }, //18 November 1976, UT1-UTC=-0.2083144, Accumulated Leap Seconds=15
  { 2443101.50, 47.395673 }, //19 November 1976, UT1-UTC=-0.2116730, Accumulated Leap Seconds=15
  { 2443102.50, 47.398952 }, //20 November 1976, UT1-UTC=-0.2149518, Accumulated Leap Seconds=15
  { 2443103.50, 47.402100 }, //21 November 1976, UT1-UTC=-0.2180999, Accumulated Leap Seconds=15
  { 2443104.50, 47.405109 }, //22 November 1976, UT1-UTC=-0.2211095, Accumulated Leap Seconds=15
  { 2443105.50, 47.408017 }, //23 November 1976, UT1-UTC=-0.2240172, Accumulated Leap Seconds=15
  { 2443106.50, 47.410885 }, //24 November 1976, UT1-UTC=-0.2268848, Accumulated Leap Seconds=15
  { 2443107.50, 47.413772 }, //25 November 1976, UT1-UTC=-0.2297720, Accumulated Leap Seconds=15
  { 2443108.50, 47.416716 }, //26 November 1976, UT1-UTC=-0.2327159, Accumulated Leap Seconds=15
  { 2443109.50, 47.419724 }, //27 November 1976, UT1-UTC=-0.2357238, Accumulated Leap Seconds=15
  { 2443110.50, 47.422779 }, //28 November 1976, UT1-UTC=-0.2387789, Accumulated Leap Seconds=15
  { 2443111.50, 47.425853 }, //29 November 1976, UT1-UTC=-0.2418531, Accumulated Leap Seconds=15
  { 2443112.50, 47.428915 }, //30 November 1976, UT1-UTC=-0.2449153, Accumulated Leap Seconds=15
  { 2443113.50, 47.431936 }, //01 December 1976, UT1-UTC=-0.2479364, Accumulated Leap Seconds=15
  { 2443114.50, 47.434892 }, //02 December 1976, UT1-UTC=-0.2508917, Accumulated Leap Seconds=15
  { 2443115.50, 47.437763 }, //03 December 1976, UT1-UTC=-0.2537633, Accumulated Leap Seconds=15
  { 2443116.50, 47.440544 }, //04 December 1976, UT1-UTC=-0.2565442, Accumulated Leap Seconds=15
  { 2443117.50, 47.443241 }, //05 December 1976, UT1-UTC=-0.2592407, Accumulated Leap Seconds=15
  { 2443118.50, 47.445873 }, //06 December 1976, UT1-UTC=-0.2618727, Accumulated Leap Seconds=15
  { 2443119.50, 47.448470 }, //07 December 1976, UT1-UTC=-0.2644700, Accumulated Leap Seconds=15
  { 2443120.50, 47.451069 }, //08 December 1976, UT1-UTC=-0.2670691, Accumulated Leap Seconds=15
  { 2443121.50, 47.453709 }, //09 December 1976, UT1-UTC=-0.2697089, Accumulated Leap Seconds=15
  { 2443122.50, 47.456427 }, //10 December 1976, UT1-UTC=-0.2724274, Accumulated Leap Seconds=15
  { 2443123.50, 47.459257 }, //11 December 1976, UT1-UTC=-0.2752571, Accumulated Leap Seconds=15
  { 2443124.50, 47.462219 }, //12 December 1976, UT1-UTC=-0.2782186, Accumulated Leap Seconds=15
  { 2443125.50, 47.465315 }, //13 December 1976, UT1-UTC=-0.2813151, Accumulated Leap Seconds=15
  { 2443126.50, 47.468528 }, //14 December 1976, UT1-UTC=-0.2845277, Accumulated Leap Seconds=15
  { 2443127.50, 47.471815 }, //15 December 1976, UT1-UTC=-0.2878148, Accumulated Leap Seconds=15
  { 2443128.50, 47.475116 }, //16 December 1976, UT1-UTC=-0.2911165, Accumulated Leap Seconds=15
  { 2443129.50, 47.478367 }, //17 December 1976, UT1-UTC=-0.2943671, Accumulated Leap Seconds=15
  { 2443130.50, 47.481512 }, //18 December 1976, UT1-UTC=-0.2975122, Accumulated Leap Seconds=15
  { 2443131.50, 47.484526 }, //19 December 1976, UT1-UTC=-0.3005265, Accumulated Leap Seconds=15
  { 2443132.50, 47.487423 }, //20 December 1976, UT1-UTC=-0.3034234, Accumulated Leap Seconds=15
  { 2443133.50, 47.490250 }, //21 December 1976, UT1-UTC=-0.3062498, Accumulated Leap Seconds=15
  { 2443134.50, 47.493065 }, //22 December 1976, UT1-UTC=-0.3090653, Accumulated Leap Seconds=15
  { 2443135.50, 47.495918 }, //23 December 1976, UT1-UTC=-0.3119181, Accumulated Leap Seconds=15
  { 2443136.50, 47.498829 }, //24 December 1976, UT1-UTC=-0.3148291, Accumulated Leap Seconds=15
  { 2443137.50, 47.501789 }, //25 December 1976, UT1-UTC=-0.3177886, Accumulated Leap Seconds=15
  { 2443138.50, 47.504766 }, //26 December 1976, UT1-UTC=-0.3207663, Accumulated Leap Seconds=15
  { 2443139.50, 47.507726 }, //27 December 1976, UT1-UTC=-0.3237258, Accumulated Leap Seconds=15
  { 2443140.50, 47.510634 }, //28 December 1976, UT1-UTC=-0.3266345, Accumulated Leap Seconds=15
  { 2443141.50, 47.513467 }, //29 December 1976, UT1-UTC=-0.3294667, Accumulated Leap Seconds=15
  { 2443142.50, 47.516205 }, //30 December 1976, UT1-UTC=-0.3322049, Accumulated Leap Seconds=15
  { 2443143.50, 47.518842 }, //31 December 1976, UT1-UTC=-0.3348417, Accumulated Leap Seconds=15
  { 2443144.50, 47.521381 }, //01 January 1977, UT1-UTC= 0.6626185, Accumulated Leap Seconds=16
  { 2443145.50, 47.523842 }, //02 January 1977, UT1-UTC= 0.6601581, Accumulated Leap Seconds=16
  { 2443146.50, 47.526252 }, //03 January 1977, UT1-UTC= 0.6577476, Accumulated Leap Seconds=16
  { 2443147.50, 47.528651 }, //04 January 1977, UT1-UTC= 0.6553493, Accumulated Leap Seconds=16
  { 2443148.50, 47.531078 }, //05 January 1977, UT1-UTC= 0.6529220, Accumulated Leap Seconds=16
  { 2443149.50, 47.533576 }, //06 January 1977, UT1-UTC= 0.6504243, Accumulated Leap Seconds=16
  { 2443150.50, 47.536180 }, //07 January 1977, UT1-UTC= 0.6478203, Accumulated Leap Seconds=16
  { 2443151.50, 47.538914 }, //08 January 1977, UT1-UTC= 0.6450864, Accumulated Leap Seconds=16
  { 2443152.50, 47.541781 }, //09 January 1977, UT1-UTC= 0.6422189, Accumulated Leap Seconds=16
  { 2443153.50, 47.544762 }, //10 January 1977, UT1-UTC= 0.6392375, Accumulated Leap Seconds=16
  { 2443154.50, 47.547816 }, //11 January 1977, UT1-UTC= 0.6361842, Accumulated Leap Seconds=16
  { 2443155.50, 47.550885 }, //12 January 1977, UT1-UTC= 0.6331152, Accumulated Leap Seconds=16
  { 2443156.50, 47.553912 }, //13 January 1977, UT1-UTC= 0.6300883, Accumulated Leap Seconds=16
  { 2443157.50, 47.556849 }, //14 January 1977, UT1-UTC= 0.6271505, Accumulated Leap Seconds=16
  { 2443158.50, 47.559675 }, //15 January 1977, UT1-UTC= 0.6243251, Accumulated Leap Seconds=16
  { 2443159.50, 47.562395 }, //16 January 1977, UT1-UTC= 0.6216051, Accumulated Leap Seconds=16
  { 2443160.50, 47.565047 }, //17 January 1977, UT1-UTC= 0.6189530, Accumulated Leap Seconds=16
  { 2443161.50, 47.567687 }, //18 January 1977, UT1-UTC= 0.6163126, Accumulated Leap Seconds=16
  { 2443162.50, 47.570372 }, //19 January 1977, UT1-UTC= 0.6136280, Accumulated Leap Seconds=16
  { 2443163.50, 47.573136 }, //20 January 1977, UT1-UTC= 0.6108637, Accumulated Leap Seconds=16
  { 2443164.50, 47.575984 }, //21 January 1977, UT1-UTC= 0.6080158, Accumulated Leap Seconds=16
  { 2443165.50, 47.578890 }, //22 January 1977, UT1-UTC= 0.6051100, Accumulated Leap Seconds=16
  { 2443166.50, 47.581811 }, //23 January 1977, UT1-UTC= 0.6021892, Accumulated Leap Seconds=16
  { 2443167.50, 47.584702 }, //24 January 1977, UT1-UTC= 0.5992982, Accumulated Leap Seconds=16
  { 2443168.50, 47.587526 }, //25 January 1977, UT1-UTC= 0.5964735, Accumulated Leap Seconds=16
  { 2443169.50, 47.590261 }, //26 January 1977, UT1-UTC= 0.5937394, Accumulated Leap Seconds=16
  { 2443170.50, 47.592892 }, //27 January 1977, UT1-UTC= 0.5911085, Accumulated Leap Seconds=16
  { 2443171.50, 47.595419 }, //28 January 1977, UT1-UTC= 0.5885810, Accumulated Leap Seconds=16
  { 2443172.50, 47.597856 }, //29 January 1977, UT1-UTC= 0.5861440, Accumulated Leap Seconds=16
  { 2443173.50, 47.600228 }, //30 January 1977, UT1-UTC= 0.5837720, Accumulated Leap Seconds=16
  { 2443174.50, 47.602572 }, //31 January 1977, UT1-UTC= 0.5814283, Accumulated Leap Seconds=16
  { 2443175.50, 47.604931 }, //01 February 1977, UT1-UTC= 0.5790687, Accumulated Leap Seconds=16
  { 2443176.50, 47.607353 }, //02 February 1977, UT1-UTC= 0.5766468, Accumulated Leap Seconds=16
  { 2443177.50, 47.609881 }, //03 February 1977, UT1-UTC= 0.5741194, Accumulated Leap Seconds=16
  { 2443178.50, 47.612547 }, //04 February 1977, UT1-UTC= 0.5714534, Accumulated Leap Seconds=16
  { 2443179.50, 47.615365 }, //05 February 1977, UT1-UTC= 0.5686346, Accumulated Leap Seconds=16
  { 2443180.50, 47.618324 }, //06 February 1977, UT1-UTC= 0.5656757, Accumulated Leap Seconds=16
  { 2443181.50, 47.621382 }, //07 February 1977, UT1-UTC= 0.5626180, Accumulated Leap Seconds=16
  { 2443182.50, 47.624476 }, //08 February 1977, UT1-UTC= 0.5595238, Accumulated Leap Seconds=16
  { 2443183.50, 47.627540 }, //09 February 1977, UT1-UTC= 0.5564602, Accumulated Leap Seconds=16
  { 2443184.50, 47.630519 }, //10 February 1977, UT1-UTC= 0.5534807, Accumulated Leap Seconds=16
  { 2443185.50, 47.633388 }, //11 February 1977, UT1-UTC= 0.5506118, Accumulated Leap Seconds=16
  { 2443186.50, 47.636152 }, //12 February 1977, UT1-UTC= 0.5478478, Accumulated Leap Seconds=16
  { 2443187.50, 47.638846 }, //13 February 1977, UT1-UTC= 0.5451542, Accumulated Leap Seconds=16
  { 2443188.50, 47.641523 }, //14 February 1977, UT1-UTC= 0.5424775, Accumulated Leap Seconds=16
  { 2443189.50, 47.644240 }, //15 February 1977, UT1-UTC= 0.5397602, Accumulated Leap Seconds=16
  { 2443190.50, 47.647043 }, //16 February 1977, UT1-UTC= 0.5369573, Accumulated Leap Seconds=16
  { 2443191.50, 47.649950 }, //17 February 1977, UT1-UTC= 0.5340500, Accumulated Leap Seconds=16
  { 2443192.50, 47.652951 }, //18 February 1977, UT1-UTC= 0.5310492, Accumulated Leap Seconds=16
  { 2443193.50, 47.656009 }, //19 February 1977, UT1-UTC= 0.5279908, Accumulated Leap Seconds=16
  { 2443194.50, 47.659077 }, //20 February 1977, UT1-UTC= 0.5249227, Accumulated Leap Seconds=16
  { 2443195.50, 47.662109 }, //21 February 1977, UT1-UTC= 0.5218909, Accumulated Leap Seconds=16
  { 2443196.50, 47.665070 }, //22 February 1977, UT1-UTC= 0.5189301, Accumulated Leap Seconds=16
  { 2443197.50, 47.667939 }, //23 February 1977, UT1-UTC= 0.5160611, Accumulated Leap Seconds=16
  { 2443198.50, 47.670709 }, //24 February 1977, UT1-UTC= 0.5132906, Accumulated Leap Seconds=16
  { 2443199.50, 47.673388 }, //25 February 1977, UT1-UTC= 0.5106119, Accumulated Leap Seconds=16
  { 2443200.50, 47.675994 }, //26 February 1977, UT1-UTC= 0.5080055, Accumulated Leap Seconds=16
  { 2443201.50, 47.678560 }, //27 February 1977, UT1-UTC= 0.5054395, Accumulated Leap Seconds=16
  { 2443202.50, 47.681126 }, //28 February 1977, UT1-UTC= 0.5028736, Accumulated Leap Seconds=16
  { 2443203.50, 47.683739 }, //01 March 1977, UT1-UTC= 0.5002612, Accumulated Leap Seconds=16
  { 2443204.50, 47.686444 }, //02 March 1977, UT1-UTC= 0.4975561, Accumulated Leap Seconds=16
  { 2443205.50, 47.689282 }, //03 March 1977, UT1-UTC= 0.4947180, Accumulated Leap Seconds=16
  { 2443206.50, 47.692279 }, //04 March 1977, UT1-UTC= 0.4917208, Accumulated Leap Seconds=16
  { 2443207.50, 47.695437 }, //05 March 1977, UT1-UTC= 0.4885627, Accumulated Leap Seconds=16
  { 2443208.50, 47.698726 }, //06 March 1977, UT1-UTC= 0.4852740, Accumulated Leap Seconds=16
  { 2443209.50, 47.702084 }, //07 March 1977, UT1-UTC= 0.4819160, Accumulated Leap Seconds=16
  { 2443210.50, 47.705432 }, //08 March 1977, UT1-UTC= 0.4785676, Accumulated Leap Seconds=16
  { 2443211.50, 47.708698 }, //09 March 1977, UT1-UTC= 0.4753023, Accumulated Leap Seconds=16
  { 2443212.50, 47.711833 }, //10 March 1977, UT1-UTC= 0.4721665, Accumulated Leap Seconds=16
  { 2443213.50, 47.714832 }, //11 March 1977, UT1-UTC= 0.4691678, Accumulated Leap Seconds=16
  { 2443214.50, 47.717723 }, //12 March 1977, UT1-UTC= 0.4662772, Accumulated Leap Seconds=16
  { 2443215.50, 47.720559 }, //13 March 1977, UT1-UTC= 0.4634414, Accumulated Leap Seconds=16
  { 2443216.50, 47.723399 }, //14 March 1977, UT1-UTC= 0.4606005, Accumulated Leap Seconds=16
  { 2443217.50, 47.726297 }, //15 March 1977, UT1-UTC= 0.4577033, Accumulated Leap Seconds=16
  { 2443218.50, 47.729281 }, //16 March 1977, UT1-UTC= 0.4547186, Accumulated Leap Seconds=16
  { 2443219.50, 47.732360 }, //17 March 1977, UT1-UTC= 0.4516395, Accumulated Leap Seconds=16
  { 2443220.50, 47.735516 }, //18 March 1977, UT1-UTC= 0.4484838, Accumulated Leap Seconds=16
  { 2443221.50, 47.738713 }, //19 March 1977, UT1-UTC= 0.4452868, Accumulated Leap Seconds=16
  { 2443222.50, 47.741910 }, //20 March 1977, UT1-UTC= 0.4420902, Accumulated Leap Seconds=16
  { 2443223.50, 47.745068 }, //21 March 1977, UT1-UTC= 0.4389315, Accumulated Leap Seconds=16
  { 2443224.50, 47.748163 }, //22 March 1977, UT1-UTC= 0.4358371, Accumulated Leap Seconds=16
  { 2443225.50, 47.751179 }, //23 March 1977, UT1-UTC= 0.4328211, Accumulated Leap Seconds=16
  { 2443226.50, 47.754116 }, //24 March 1977, UT1-UTC= 0.4298835, Accumulated Leap Seconds=16
  { 2443227.50, 47.756989 }, //25 March 1977, UT1-UTC= 0.4270109, Accumulated Leap Seconds=16
  { 2443228.50, 47.759822 }, //26 March 1977, UT1-UTC= 0.4241782, Accumulated Leap Seconds=16
  { 2443229.50, 47.762649 }, //27 March 1977, UT1-UTC= 0.4213506, Accumulated Leap Seconds=16
  { 2443230.50, 47.765513 }, //28 March 1977, UT1-UTC= 0.4184867, Accumulated Leap Seconds=16
  { 2443231.50, 47.768457 }, //29 March 1977, UT1-UTC= 0.4155428, Accumulated Leap Seconds=16
  { 2443232.50, 47.771522 }, //30 March 1977, UT1-UTC= 0.4124775, Accumulated Leap Seconds=16
  { 2443233.50, 47.774744 }, //31 March 1977, UT1-UTC= 0.4092564, Accumulated Leap Seconds=16
  { 2443234.50, 47.778138 }, //01 April 1977, UT1-UTC= 0.4058619, Accumulated Leap Seconds=16
  { 2443235.50, 47.781697 }, //02 April 1977, UT1-UTC= 0.4023032, Accumulated Leap Seconds=16
  { 2443236.50, 47.785377 }, //03 April 1977, UT1-UTC= 0.3986232, Accumulated Leap Seconds=16
  { 2443237.50, 47.789106 }, //04 April 1977, UT1-UTC= 0.3948942, Accumulated Leap Seconds=16
  { 2443238.50, 47.792799 }, //05 April 1977, UT1-UTC= 0.3912009, Accumulated Leap Seconds=16
  { 2443239.50, 47.796386 }, //06 April 1977, UT1-UTC= 0.3876139, Accumulated Leap Seconds=16
  { 2443240.50, 47.799833 }, //07 April 1977, UT1-UTC= 0.3841665, Accumulated Leap Seconds=16
  { 2443241.50, 47.803152 }, //08 April 1977, UT1-UTC= 0.3808477, Accumulated Leap Seconds=16
  { 2443242.50, 47.806387 }, //09 April 1977, UT1-UTC= 0.3776128, Accumulated Leap Seconds=16
  { 2443243.50, 47.809596 }, //10 April 1977, UT1-UTC= 0.3744038, Accumulated Leap Seconds=16
  { 2443244.50, 47.812830 }, //11 April 1977, UT1-UTC= 0.3711700, Accumulated Leap Seconds=16
  { 2443245.50, 47.816120 }, //12 April 1977, UT1-UTC= 0.3678802, Accumulated Leap Seconds=16
  { 2443246.50, 47.819473 }, //13 April 1977, UT1-UTC= 0.3645267, Accumulated Leap Seconds=16
  { 2443247.50, 47.822877 }, //14 April 1977, UT1-UTC= 0.3611225, Accumulated Leap Seconds=16
  { 2443248.50, 47.826304 }, //15 April 1977, UT1-UTC= 0.3576961, Accumulated Leap Seconds=16
  { 2443249.50, 47.829716 }, //16 April 1977, UT1-UTC= 0.3542844, Accumulated Leap Seconds=16
  { 2443250.50, 47.833075 }, //17 April 1977, UT1-UTC= 0.3509252, Accumulated Leap Seconds=16
  { 2443251.50, 47.836351 }, //18 April 1977, UT1-UTC= 0.3476494, Accumulated Leap Seconds=16
  { 2443252.50, 47.839524 }, //19 April 1977, UT1-UTC= 0.3444757, Accumulated Leap Seconds=16
  { 2443253.50, 47.842592 }, //20 April 1977, UT1-UTC= 0.3414078, Accumulated Leap Seconds=16
  { 2443254.50, 47.845564 }, //21 April 1977, UT1-UTC= 0.3384355, Accumulated Leap Seconds=16
  { 2443255.50, 47.848465 }, //22 April 1977, UT1-UTC= 0.3355354, Accumulated Leap Seconds=16
  { 2443256.50, 47.851326 }, //23 April 1977, UT1-UTC= 0.3326741, Accumulated Leap Seconds=16
  { 2443257.50, 47.854188 }, //24 April 1977, UT1-UTC= 0.3298116, Accumulated Leap Seconds=16
  { 2443258.50, 47.857095 }, //25 April 1977, UT1-UTC= 0.3269046, Accumulated Leap Seconds=16
  { 2443259.50, 47.860088 }, //26 April 1977, UT1-UTC= 0.3239115, Accumulated Leap Seconds=16
  { 2443260.50, 47.863204 }, //27 April 1977, UT1-UTC= 0.3207960, Accumulated Leap Seconds=16
  { 2443261.50, 47.866467 }, //28 April 1977, UT1-UTC= 0.3175325, Accumulated Leap Seconds=16
  { 2443262.50, 47.869886 }, //29 April 1977, UT1-UTC= 0.3141144, Accumulated Leap Seconds=16
  { 2443263.50, 47.873438 }, //30 April 1977, UT1-UTC= 0.3105615, Accumulated Leap Seconds=16
  { 2443264.50, 47.877076 }, //01 May 1977, UT1-UTC= 0.3069244, Accumulated Leap Seconds=16
  { 2443265.50, 47.880721 }, //02 May 1977, UT1-UTC= 0.3032788, Accumulated Leap Seconds=16
  { 2443266.50, 47.884295 }, //03 May 1977, UT1-UTC= 0.2997048, Accumulated Leap Seconds=16
  { 2443267.50, 47.887739 }, //04 May 1977, UT1-UTC= 0.2962607, Accumulated Leap Seconds=16
  { 2443268.50, 47.891038 }, //05 May 1977, UT1-UTC= 0.2929619, Accumulated Leap Seconds=16
  { 2443269.50, 47.894220 }, //06 May 1977, UT1-UTC= 0.2897801, Accumulated Leap Seconds=16
  { 2443270.50, 47.897340 }, //07 May 1977, UT1-UTC= 0.2866596, Accumulated Leap Seconds=16
  { 2443271.50, 47.900457 }, //08 May 1977, UT1-UTC= 0.2835434, Accumulated Leap Seconds=16
  { 2443272.50, 47.903606 }, //09 May 1977, UT1-UTC= 0.2803938, Accumulated Leap Seconds=16
  { 2443273.50, 47.906799 }, //10 May 1977, UT1-UTC= 0.2772006, Accumulated Leap Seconds=16
  { 2443274.50, 47.910022 }, //11 May 1977, UT1-UTC= 0.2739783, Accumulated Leap Seconds=16
  { 2443275.50, 47.913244 }, //12 May 1977, UT1-UTC= 0.2707555, Accumulated Leap Seconds=16
  { 2443276.50, 47.916433 }, //13 May 1977, UT1-UTC= 0.2675670, Accumulated Leap Seconds=16
  { 2443277.50, 47.919553 }, //14 May 1977, UT1-UTC= 0.2644468, Accumulated Leap Seconds=16
  { 2443278.50, 47.922576 }, //15 May 1977, UT1-UTC= 0.2614238, Accumulated Leap Seconds=16
  { 2443279.50, 47.925482 }, //16 May 1977, UT1-UTC= 0.2585176, Accumulated Leap Seconds=16
  { 2443280.50, 47.928265 }, //17 May 1977, UT1-UTC= 0.2557354, Accumulated Leap Seconds=16
  { 2443281.50, 47.930929 }, //18 May 1977, UT1-UTC= 0.2530705, Accumulated Leap Seconds=16
  { 2443282.50, 47.933496 }, //19 May 1977, UT1-UTC= 0.2505039, Accumulated Leap Seconds=16
  { 2443283.50, 47.935993 }, //20 May 1977, UT1-UTC= 0.2480069, Accumulated Leap Seconds=16
  { 2443284.50, 47.938455 }, //21 May 1977, UT1-UTC= 0.2455448, Accumulated Leap Seconds=16
  { 2443285.50, 47.940920 }, //22 May 1977, UT1-UTC= 0.2430797, Accumulated Leap Seconds=16
  { 2443286.50, 47.943425 }, //23 May 1977, UT1-UTC= 0.2405745, Accumulated Leap Seconds=16
  { 2443287.50, 47.946004 }, //24 May 1977, UT1-UTC= 0.2379961, Accumulated Leap Seconds=16
  { 2443288.50, 47.948680 }, //25 May 1977, UT1-UTC= 0.2353201, Accumulated Leap Seconds=16
  { 2443289.50, 47.951465 }, //26 May 1977, UT1-UTC= 0.2325353, Accumulated Leap Seconds=16
  { 2443290.50, 47.954351 }, //27 May 1977, UT1-UTC= 0.2296489, Accumulated Leap Seconds=16
  { 2443291.50, 47.957309 }, //28 May 1977, UT1-UTC= 0.2266911, Accumulated Leap Seconds=16
  { 2443292.50, 47.960285 }, //29 May 1977, UT1-UTC= 0.2237150, Accumulated Leap Seconds=16
  { 2443293.50, 47.963213 }, //30 May 1977, UT1-UTC= 0.2207873, Accumulated Leap Seconds=16
  { 2443294.50, 47.966032 }, //31 May 1977, UT1-UTC= 0.2179679, Accumulated Leap Seconds=16
  { 2443295.50, 47.968710 }, //01 June 1977, UT1-UTC= 0.2152896, Accumulated Leap Seconds=16
  { 2443296.50, 47.971255 }, //02 June 1977, UT1-UTC= 0.2127446, Accumulated Leap Seconds=16
  { 2443297.50, 47.973712 }, //03 June 1977, UT1-UTC= 0.2102882, Accumulated Leap Seconds=16
  { 2443298.50, 47.976141 }, //04 June 1977, UT1-UTC= 0.2078594, Accumulated Leap Seconds=16
  { 2443299.50, 47.978593 }, //05 June 1977, UT1-UTC= 0.2054067, Accumulated Leap Seconds=16
  { 2443300.50, 47.981093 }, //06 June 1977, UT1-UTC= 0.2029074, Accumulated Leap Seconds=16
  { 2443301.50, 47.983630 }, //07 June 1977, UT1-UTC= 0.2003697, Accumulated Leap Seconds=16
  { 2443302.50, 47.986177 }, //08 June 1977, UT1-UTC= 0.1978232, Accumulated Leap Seconds=16
  { 2443303.50, 47.988694 }, //09 June 1977, UT1-UTC= 0.1953057, Accumulated Leap Seconds=16
  { 2443304.50, 47.991147 }, //10 June 1977, UT1-UTC= 0.1928528, Accumulated Leap Seconds=16
  { 2443305.50, 47.993508 }, //11 June 1977, UT1-UTC= 0.1904918, Accumulated Leap Seconds=16
  { 2443306.50, 47.995760 }, //12 June 1977, UT1-UTC= 0.1882404, Accumulated Leap Seconds=16
  { 2443307.50, 47.997894 }, //13 June 1977, UT1-UTC= 0.1861060, Accumulated Leap Seconds=16
  { 2443308.50, 47.999917 }, //14 June 1977, UT1-UTC= 0.1840834, Accumulated Leap Seconds=16
  { 2443309.50, 48.001845 }, //15 June 1977, UT1-UTC= 0.1821549, Accumulated Leap Seconds=16
  { 2443310.50, 48.003706 }, //16 June 1977, UT1-UTC= 0.1802937, Accumulated Leap Seconds=16
  { 2443311.50, 48.005534 }, //17 June 1977, UT1-UTC= 0.1784663, Accumulated Leap Seconds=16
  { 2443312.50, 48.007363 }, //18 June 1977, UT1-UTC= 0.1766367, Accumulated Leap Seconds=16
  { 2443313.50, 48.009230 }, //19 June 1977, UT1-UTC= 0.1747697, Accumulated Leap Seconds=16
  { 2443314.50, 48.011166 }, //20 June 1977, UT1-UTC= 0.1728340, Accumulated Leap Seconds=16
  { 2443315.50, 48.013194 }, //21 June 1977, UT1-UTC= 0.1708059, Accumulated Leap Seconds=16
  { 2443316.50, 48.015326 }, //22 June 1977, UT1-UTC= 0.1686738, Accumulated Leap Seconds=16
  { 2443317.50, 48.017557 }, //23 June 1977, UT1-UTC= 0.1664431, Accumulated Leap Seconds=16
  { 2443318.50, 48.019863 }, //24 June 1977, UT1-UTC= 0.1641366, Accumulated Leap Seconds=16
  { 2443319.50, 48.022205 }, //25 June 1977, UT1-UTC= 0.1617951, Accumulated Leap Seconds=16
  { 2443320.50, 48.024529 }, //26 June 1977, UT1-UTC= 0.1594713, Accumulated Leap Seconds=16
  { 2443321.50, 48.026782 }, //27 June 1977, UT1-UTC= 0.1572181, Accumulated Leap Seconds=16
  { 2443322.50, 48.028928 }, //28 June 1977, UT1-UTC= 0.1550723, Accumulated Leap Seconds=16
  { 2443323.50, 48.030960 }, //29 June 1977, UT1-UTC= 0.1530401, Accumulated Leap Seconds=16
  { 2443324.50, 48.032908 }, //30 June 1977, UT1-UTC= 0.1510924, Accumulated Leap Seconds=16
  { 2443325.50, 48.034826 }, //01 July 1977, UT1-UTC= 0.1491743, Accumulated Leap Seconds=16
  { 2443326.50, 48.036773 }, //02 July 1977, UT1-UTC= 0.1472272, Accumulated Leap Seconds=16
  { 2443327.50, 48.038786 }, //03 July 1977, UT1-UTC= 0.1452138, Accumulated Leap Seconds=16
  { 2443328.50, 48.040868 }, //04 July 1977, UT1-UTC= 0.1431315, Accumulated Leap Seconds=16
  { 2443329.50, 48.042990 }, //05 July 1977, UT1-UTC= 0.1410102, Accumulated Leap Seconds=16
  { 2443330.50, 48.045103 }, //06 July 1977, UT1-UTC= 0.1388972, Accumulated Leap Seconds=16
  { 2443331.50, 48.047162 }, //07 July 1977, UT1-UTC= 0.1368378, Accumulated Leap Seconds=16
  { 2443332.50, 48.049134 }, //08 July 1977, UT1-UTC= 0.1348656, Accumulated Leap Seconds=16
  { 2443333.50, 48.051000 }, //09 July 1977, UT1-UTC= 0.1330001, Accumulated Leap Seconds=16
  { 2443334.50, 48.052751 }, //10 July 1977, UT1-UTC= 0.1312487, Accumulated Leap Seconds=16
  { 2443335.50, 48.054393 }, //11 July 1977, UT1-UTC= 0.1296067, Accumulated Leap Seconds=16
  { 2443336.50, 48.055943 }, //12 July 1977, UT1-UTC= 0.1280574, Accumulated Leap Seconds=16
  { 2443337.50, 48.057428 }, //13 July 1977, UT1-UTC= 0.1265717, Accumulated Leap Seconds=16
  { 2443338.50, 48.058888 }, //14 July 1977, UT1-UTC= 0.1251116, Accumulated Leap Seconds=16
  { 2443339.50, 48.060365 }, //15 July 1977, UT1-UTC= 0.1236351, Accumulated Leap Seconds=16
  { 2443340.50, 48.061899 }, //16 July 1977, UT1-UTC= 0.1221012, Accumulated Leap Seconds=16
  { 2443341.50, 48.063527 }, //17 July 1977, UT1-UTC= 0.1204734, Accumulated Leap Seconds=16
  { 2443342.50, 48.065276 }, //18 July 1977, UT1-UTC= 0.1187243, Accumulated Leap Seconds=16
  { 2443343.50, 48.067160 }, //19 July 1977, UT1-UTC= 0.1168399, Accumulated Leap Seconds=16
  { 2443344.50, 48.069176 }, //20 July 1977, UT1-UTC= 0.1148243, Accumulated Leap Seconds=16
  { 2443345.50, 48.071299 }, //21 July 1977, UT1-UTC= 0.1127012, Accumulated Leap Seconds=16
  { 2443346.50, 48.073488 }, //22 July 1977, UT1-UTC= 0.1105123, Accumulated Leap Seconds=16
  { 2443347.50, 48.075692 }, //23 July 1977, UT1-UTC= 0.1083077, Accumulated Leap Seconds=16
  { 2443348.50, 48.077864 }, //24 July 1977, UT1-UTC= 0.1061359, Accumulated Leap Seconds=16
  { 2443349.50, 48.079968 }, //25 July 1977, UT1-UTC= 0.1040324, Accumulated Leap Seconds=16
  { 2443350.50, 48.081990 }, //26 July 1977, UT1-UTC= 0.1020104, Accumulated Leap Seconds=16
  { 2443351.50, 48.083945 }, //27 July 1977, UT1-UTC= 0.1000546, Accumulated Leap Seconds=16
  { 2443352.50, 48.085876 }, //28 July 1977, UT1-UTC= 0.0981241, Accumulated Leap Seconds=16
  { 2443353.50, 48.087835 }, //29 July 1977, UT1-UTC= 0.0961651, Accumulated Leap Seconds=16
  { 2443354.50, 48.089868 }, //30 July 1977, UT1-UTC= 0.0941317, Accumulated Leap Seconds=16
  { 2443355.50, 48.091996 }, //31 July 1977, UT1-UTC= 0.0920042, Accumulated Leap Seconds=16
  { 2443356.50, 48.094202 }, //01 August 1977, UT1-UTC= 0.0897979, Accumulated Leap Seconds=16
  { 2443357.50, 48.096443 }, //02 August 1977, UT1-UTC= 0.0875573, Accumulated Leap Seconds=16
  { 2443358.50, 48.098663 }, //03 August 1977, UT1-UTC= 0.0853374, Accumulated Leap Seconds=16
  { 2443359.50, 48.100814 }, //04 August 1977, UT1-UTC= 0.0831856, Accumulated Leap Seconds=16
  { 2443360.50, 48.102867 }, //05 August 1977, UT1-UTC= 0.0811332, Accumulated Leap Seconds=16
  { 2443361.50, 48.104806 }, //06 August 1977, UT1-UTC= 0.0791938, Accumulated Leap Seconds=16
  { 2443362.50, 48.106634 }, //07 August 1977, UT1-UTC= 0.0773662, Accumulated Leap Seconds=16
  { 2443363.50, 48.108364 }, //08 August 1977, UT1-UTC= 0.0756364, Accumulated Leap Seconds=16
  { 2443364.50, 48.110022 }, //09 August 1977, UT1-UTC= 0.0739781, Accumulated Leap Seconds=16
  { 2443365.50, 48.111646 }, //10 August 1977, UT1-UTC= 0.0723543, Accumulated Leap Seconds=16
  { 2443366.50, 48.113278 }, //11 August 1977, UT1-UTC= 0.0707221, Accumulated Leap Seconds=16
  { 2443367.50, 48.114961 }, //12 August 1977, UT1-UTC= 0.0690390, Accumulated Leap Seconds=16
  { 2443368.50, 48.116733 }, //13 August 1977, UT1-UTC= 0.0672667, Accumulated Leap Seconds=16
  { 2443369.50, 48.118625 }, //14 August 1977, UT1-UTC= 0.0653746, Accumulated Leap Seconds=16
  { 2443370.50, 48.120656 }, //15 August 1977, UT1-UTC= 0.0633441, Accumulated Leap Seconds=16
  { 2443371.50, 48.122826 }, //16 August 1977, UT1-UTC= 0.0611742, Accumulated Leap Seconds=16
  { 2443372.50, 48.125115 }, //17 August 1977, UT1-UTC= 0.0588849, Accumulated Leap Seconds=16
  { 2443373.50, 48.127482 }, //18 August 1977, UT1-UTC= 0.0565178, Accumulated Leap Seconds=16
  { 2443374.50, 48.129873 }, //19 August 1977, UT1-UTC= 0.0541272, Accumulated Leap Seconds=16
  { 2443375.50, 48.132234 }, //20 August 1977, UT1-UTC= 0.0517664, Accumulated Leap Seconds=16
  { 2443376.50, 48.134527 }, //21 August 1977, UT1-UTC= 0.0494732, Accumulated Leap Seconds=16
  { 2443377.50, 48.136740 }, //22 August 1977, UT1-UTC= 0.0472599, Accumulated Leap Seconds=16
  { 2443378.50, 48.138890 }, //23 August 1977, UT1-UTC= 0.0451103, Accumulated Leap Seconds=16
  { 2443379.50, 48.141016 }, //24 August 1977, UT1-UTC= 0.0429836, Accumulated Leap Seconds=16
  { 2443380.50, 48.143175 }, //25 August 1977, UT1-UTC= 0.0408246, Accumulated Leap Seconds=16
  { 2443381.50, 48.145421 }, //26 August 1977, UT1-UTC= 0.0385791, Accumulated Leap Seconds=16
  { 2443382.50, 48.147790 }, //27 August 1977, UT1-UTC= 0.0362104, Accumulated Leap Seconds=16
  { 2443383.50, 48.150288 }, //28 August 1977, UT1-UTC= 0.0337119, Accumulated Leap Seconds=16
  { 2443384.50, 48.152891 }, //29 August 1977, UT1-UTC= 0.0311093, Accumulated Leap Seconds=16
  { 2443385.50, 48.155549 }, //30 August 1977, UT1-UTC= 0.0284512, Accumulated Leap Seconds=16
  { 2443386.50, 48.158208 }, //31 August 1977, UT1-UTC= 0.0257924, Accumulated Leap Seconds=16
  { 2443387.50, 48.160820 }, //01 September 1977, UT1-UTC= 0.0231795, Accumulated Leap Seconds=16
  { 2443388.50, 48.163360 }, //02 September 1977, UT1-UTC= 0.0206401, Accumulated Leap Seconds=16
  { 2443389.50, 48.165817 }, //03 September 1977, UT1-UTC= 0.0181826, Accumulated Leap Seconds=16
  { 2443390.50, 48.168201 }, //04 September 1977, UT1-UTC= 0.0157992, Accumulated Leap Seconds=16
  { 2443391.50, 48.170530 }, //05 September 1977, UT1-UTC= 0.0134699, Accumulated Leap Seconds=16
  { 2443392.50, 48.172835 }, //06 September 1977, UT1-UTC= 0.0111649, Accumulated Leap Seconds=16
  { 2443393.50, 48.175153 }, //07 September 1977, UT1-UTC= 0.0088469, Accumulated Leap Seconds=16
  { 2443394.50, 48.177525 }, //08 September 1977, UT1-UTC= 0.0064745, Accumulated Leap Seconds=16
  { 2443395.50, 48.179992 }, //09 September 1977, UT1-UTC= 0.0040079, Accumulated Leap Seconds=16
  { 2443396.50, 48.182587 }, //10 September 1977, UT1-UTC= 0.0014130, Accumulated Leap Seconds=16
  { 2443397.50, 48.185334 }, //11 September 1977, UT1-UTC=-0.0013344, Accumulated Leap Seconds=16
  { 2443398.50, 48.188244 }, //12 September 1977, UT1-UTC=-0.0042437, Accumulated Leap Seconds=16
  { 2443399.50, 48.191302 }, //13 September 1977, UT1-UTC=-0.0073023, Accumulated Leap Seconds=16
  { 2443400.50, 48.194472 }, //14 September 1977, UT1-UTC=-0.0104725, Accumulated Leap Seconds=16
  { 2443401.50, 48.197695 }, //15 September 1977, UT1-UTC=-0.0136949, Accumulated Leap Seconds=16
  { 2443402.50, 48.200901 }, //16 September 1977, UT1-UTC=-0.0169013, Accumulated Leap Seconds=16
  { 2443403.50, 48.204034 }, //17 September 1977, UT1-UTC=-0.0200341, Accumulated Leap Seconds=16
  { 2443404.50, 48.207063 }, //18 September 1977, UT1-UTC=-0.0230626, Accumulated Leap Seconds=16
  { 2443405.50, 48.209991 }, //19 September 1977, UT1-UTC=-0.0259907, Accumulated Leap Seconds=16
  { 2443406.50, 48.212852 }, //20 September 1977, UT1-UTC=-0.0288518, Accumulated Leap Seconds=16
  { 2443407.50, 48.215697 }, //21 September 1977, UT1-UTC=-0.0316966, Accumulated Leap Seconds=16
  { 2443408.50, 48.218578 }, //22 September 1977, UT1-UTC=-0.0345779, Accumulated Leap Seconds=16
  { 2443409.50, 48.221536 }, //23 September 1977, UT1-UTC=-0.0375362, Accumulated Leap Seconds=16
  { 2443410.50, 48.224588 }, //24 September 1977, UT1-UTC=-0.0405881, Accumulated Leap Seconds=16
  { 2443411.50, 48.227723 }, //25 September 1977, UT1-UTC=-0.0437228, Accumulated Leap Seconds=16
  { 2443412.50, 48.230905 }, //26 September 1977, UT1-UTC=-0.0469046, Accumulated Leap Seconds=16
  { 2443413.50, 48.234084 }, //27 September 1977, UT1-UTC=-0.0500839, Accumulated Leap Seconds=16
  { 2443414.50, 48.237211 }, //28 September 1977, UT1-UTC=-0.0532112, Accumulated Leap Seconds=16
  { 2443415.50, 48.240249 }, //29 September 1977, UT1-UTC=-0.0562493, Accumulated Leap Seconds=16
  { 2443416.50, 48.243180 }, //30 September 1977, UT1-UTC=-0.0591797, Accumulated Leap Seconds=16
  { 2443417.50, 48.246003 }, //01 October 1977, UT1-UTC=-0.0620028, Accumulated Leap Seconds=16
  { 2443418.50, 48.248734 }, //02 October 1977, UT1-UTC=-0.0647341, Accumulated Leap Seconds=16
  { 2443419.50, 48.251402 }, //03 October 1977, UT1-UTC=-0.0674021, Accumulated Leap Seconds=16
  { 2443420.50, 48.254045 }, //04 October 1977, UT1-UTC=-0.0700452, Accumulated Leap Seconds=16
  { 2443421.50, 48.256706 }, //05 October 1977, UT1-UTC=-0.0727063, Accumulated Leap Seconds=16
  { 2443422.50, 48.259430 }, //06 October 1977, UT1-UTC=-0.0754298, Accumulated Leap Seconds=16
  { 2443423.50, 48.262256 }, //07 October 1977, UT1-UTC=-0.0782563, Accumulated Leap Seconds=16
  { 2443424.50, 48.265219 }, //08 October 1977, UT1-UTC=-0.0812191, Accumulated Leap Seconds=16
  { 2443425.50, 48.268341 }, //09 October 1977, UT1-UTC=-0.0843411, Accumulated Leap Seconds=16
  { 2443426.50, 48.271629 }, //10 October 1977, UT1-UTC=-0.0876288, Accumulated Leap Seconds=16
  { 2443427.50, 48.275063 }, //11 October 1977, UT1-UTC=-0.0910627, Accumulated Leap Seconds=16
  { 2443428.50, 48.278596 }, //12 October 1977, UT1-UTC=-0.0945959, Accumulated Leap Seconds=16
  { 2443429.50, 48.282162 }, //13 October 1977, UT1-UTC=-0.0981618, Accumulated Leap Seconds=16
  { 2443430.50, 48.285690 }, //14 October 1977, UT1-UTC=-0.1016902, Accumulated Leap Seconds=16
  { 2443431.50, 48.289129 }, //15 October 1977, UT1-UTC=-0.1051289, Accumulated Leap Seconds=16
  { 2443432.50, 48.292460 }, //16 October 1977, UT1-UTC=-0.1084605, Accumulated Leap Seconds=16
  { 2443433.50, 48.295705 }, //17 October 1977, UT1-UTC=-0.1117055, Accumulated Leap Seconds=16
  { 2443434.50, 48.298910 }, //18 October 1977, UT1-UTC=-0.1149104, Accumulated Leap Seconds=16
  { 2443435.50, 48.302128 }, //19 October 1977, UT1-UTC=-0.1181282, Accumulated Leap Seconds=16
  { 2443436.50, 48.305401 }, //20 October 1977, UT1-UTC=-0.1214013, Accumulated Leap Seconds=16
  { 2443437.50, 48.308749 }, //21 October 1977, UT1-UTC=-0.1247493, Accumulated Leap Seconds=16
  { 2443438.50, 48.312167 }, //22 October 1977, UT1-UTC=-0.1281672, Accumulated Leap Seconds=16
  { 2443439.50, 48.315629 }, //23 October 1977, UT1-UTC=-0.1316292, Accumulated Leap Seconds=16
  { 2443440.50, 48.319096 }, //24 October 1977, UT1-UTC=-0.1350957, Accumulated Leap Seconds=16
  { 2443441.50, 48.322522 }, //25 October 1977, UT1-UTC=-0.1385223, Accumulated Leap Seconds=16
  { 2443442.50, 48.325869 }, //26 October 1977, UT1-UTC=-0.1418689, Accumulated Leap Seconds=16
  { 2443443.50, 48.329108 }, //27 October 1977, UT1-UTC=-0.1451081, Accumulated Leap Seconds=16
  { 2443444.50, 48.332230 }, //28 October 1977, UT1-UTC=-0.1482296, Accumulated Leap Seconds=16
  { 2443445.50, 48.335240 }, //29 October 1977, UT1-UTC=-0.1512403, Accumulated Leap Seconds=16
  { 2443446.50, 48.338160 }, //30 October 1977, UT1-UTC=-0.1541601, Accumulated Leap Seconds=16
  { 2443447.50, 48.341019 }, //31 October 1977, UT1-UTC=-0.1570190, Accumulated Leap Seconds=16
  { 2443448.50, 48.343853 }, //01 November 1977, UT1-UTC=-0.1598529, Accumulated Leap Seconds=16
  { 2443449.50, 48.346701 }, //02 November 1977, UT1-UTC=-0.1627007, Accumulated Leap Seconds=16
  { 2443450.50, 48.349600 }, //03 November 1977, UT1-UTC=-0.1656005, Accumulated Leap Seconds=16
  { 2443451.50, 48.352585 }, //04 November 1977, UT1-UTC=-0.1685848, Accumulated Leap Seconds=16
  { 2443452.50, 48.355677 }, //05 November 1977, UT1-UTC=-0.1716771, Accumulated Leap Seconds=16
  { 2443453.50, 48.358889 }, //06 November 1977, UT1-UTC=-0.1748889, Accumulated Leap Seconds=16
  { 2443454.50, 48.362216 }, //07 November 1977, UT1-UTC=-0.1782163, Accumulated Leap Seconds=16
  { 2443455.50, 48.365635 }, //08 November 1977, UT1-UTC=-0.1816348, Accumulated Leap Seconds=16
  { 2443456.50, 48.369098 }, //09 November 1977, UT1-UTC=-0.1850976, Accumulated Leap Seconds=16
  { 2443457.50, 48.372543 }, //10 November 1977, UT1-UTC=-0.1885430, Accumulated Leap Seconds=16
  { 2443458.50, 48.375912 }, //11 November 1977, UT1-UTC=-0.1919116, Accumulated Leap Seconds=16
  { 2443459.50, 48.379168 }, //12 November 1977, UT1-UTC=-0.1951680, Accumulated Leap Seconds=16
  { 2443460.50, 48.382314 }, //13 November 1977, UT1-UTC=-0.1983145, Accumulated Leap Seconds=16
  { 2443461.50, 48.385389 }, //14 November 1977, UT1-UTC=-0.2013892, Accumulated Leap Seconds=16
  { 2443462.50, 48.388448 }, //15 November 1977, UT1-UTC=-0.2044481, Accumulated Leap Seconds=16
  { 2443463.50, 48.391541 }, //16 November 1977, UT1-UTC=-0.2075413, Accumulated Leap Seconds=16
  { 2443464.50, 48.394695 }, //17 November 1977, UT1-UTC=-0.2106952, Accumulated Leap Seconds=16
  { 2443465.50, 48.397907 }, //18 November 1977, UT1-UTC=-0.2139069, Accumulated Leap Seconds=16
  { 2443466.50, 48.401151 }, //19 November 1977, UT1-UTC=-0.2171508, Accumulated Leap Seconds=16
  { 2443467.50, 48.404390 }, //20 November 1977, UT1-UTC=-0.2203904, Accumulated Leap Seconds=16
  { 2443468.50, 48.407587 }, //21 November 1977, UT1-UTC=-0.2235871, Accumulated Leap Seconds=16
  { 2443469.50, 48.410708 }, //22 November 1977, UT1-UTC=-0.2267076, Accumulated Leap Seconds=16
  { 2443470.50, 48.413728 }, //23 November 1977, UT1-UTC=-0.2297285, Accumulated Leap Seconds=16
  { 2443471.50, 48.416639 }, //24 November 1977, UT1-UTC=-0.2326390, Accumulated Leap Seconds=16
  { 2443472.50, 48.419443 }, //25 November 1977, UT1-UTC=-0.2354430, Accumulated Leap Seconds=16
  { 2443473.50, 48.422158 }, //26 November 1977, UT1-UTC=-0.2381579, Accumulated Leap Seconds=16
  { 2443474.50, 48.424811 }, //27 November 1977, UT1-UTC=-0.2408114, Accumulated Leap Seconds=16
  { 2443475.50, 48.427438 }, //28 November 1977, UT1-UTC=-0.2434384, Accumulated Leap Seconds=16
  { 2443476.50, 48.430078 }, //29 November 1977, UT1-UTC=-0.2460776, Accumulated Leap Seconds=16
  { 2443477.50, 48.432767 }, //30 November 1977, UT1-UTC=-0.2487671, Accumulated Leap Seconds=16
  { 2443478.50, 48.435540 }, //01 December 1977, UT1-UTC=-0.2515405, Accumulated Leap Seconds=16
  { 2443479.50, 48.438424 }, //02 December 1977, UT1-UTC=-0.2544239, Accumulated Leap Seconds=16
  { 2443480.50, 48.441433 }, //03 December 1977, UT1-UTC=-0.2574326, Accumulated Leap Seconds=16
  { 2443481.50, 48.444569 }, //04 December 1977, UT1-UTC=-0.2605691, Accumulated Leap Seconds=16
  { 2443482.50, 48.447821 }, //05 December 1977, UT1-UTC=-0.2638210, Accumulated Leap Seconds=16
  { 2443483.50, 48.451159 }, //06 December 1977, UT1-UTC=-0.2671588, Accumulated Leap Seconds=16
  { 2443484.50, 48.454537 }, //07 December 1977, UT1-UTC=-0.2705366, Accumulated Leap Seconds=16
  { 2443485.50, 48.457901 }, //08 December 1977, UT1-UTC=-0.2739012, Accumulated Leap Seconds=16
  { 2443486.50, 48.461209 }, //09 December 1977, UT1-UTC=-0.2772087, Accumulated Leap Seconds=16
  { 2443487.50, 48.464444 }, //10 December 1977, UT1-UTC=-0.2804441, Accumulated Leap Seconds=16
  { 2443488.50, 48.467629 }, //11 December 1977, UT1-UTC=-0.2836291, Accumulated Leap Seconds=16
  { 2443489.50, 48.470813 }, //12 December 1977, UT1-UTC=-0.2868129, Accumulated Leap Seconds=16
  { 2443490.50, 48.474048 }, //13 December 1977, UT1-UTC=-0.2900484, Accumulated Leap Seconds=16
  { 2443491.50, 48.477370 }, //14 December 1977, UT1-UTC=-0.2933703, Accumulated Leap Seconds=16
  { 2443492.50, 48.480781 }, //15 December 1977, UT1-UTC=-0.2967807, Accumulated Leap Seconds=16
  { 2443493.50, 48.484251 }, //16 December 1977, UT1-UTC=-0.3002507, Accumulated Leap Seconds=16
  { 2443494.50, 48.487734 }, //17 December 1977, UT1-UTC=-0.3037341, Accumulated Leap Seconds=16
  { 2443495.50, 48.491183 }, //18 December 1977, UT1-UTC=-0.3071833, Accumulated Leap Seconds=16
  { 2443496.50, 48.494560 }, //19 December 1977, UT1-UTC=-0.3105604, Accumulated Leap Seconds=16
  { 2443497.50, 48.497841 }, //20 December 1977, UT1-UTC=-0.3138409, Accumulated Leap Seconds=16
  { 2443498.50, 48.501014 }, //21 December 1977, UT1-UTC=-0.3170136, Accumulated Leap Seconds=16
  { 2443499.50, 48.504080 }, //22 December 1977, UT1-UTC=-0.3200801, Accumulated Leap Seconds=16
  { 2443500.50, 48.507054 }, //23 December 1977, UT1-UTC=-0.3230544, Accumulated Leap Seconds=16
  { 2443501.50, 48.509961 }, //24 December 1977, UT1-UTC=-0.3259612, Accumulated Leap Seconds=16
  { 2443502.50, 48.512832 }, //25 December 1977, UT1-UTC=-0.3288325, Accumulated Leap Seconds=16
  { 2443503.50, 48.515704 }, //26 December 1977, UT1-UTC=-0.3317040, Accumulated Leap Seconds=16
  { 2443504.50, 48.518610 }, //27 December 1977, UT1-UTC=-0.3346102, Accumulated Leap Seconds=16
  { 2443505.50, 48.521582 }, //28 December 1977, UT1-UTC=-0.3375816, Accumulated Leap Seconds=16
  { 2443506.50, 48.524641 }, //29 December 1977, UT1-UTC=-0.3406413, Accumulated Leap Seconds=16
  { 2443507.50, 48.527802 }, //30 December 1977, UT1-UTC=-0.3438023, Accumulated Leap Seconds=16
  { 2443508.50, 48.531065 }, //31 December 1977, UT1-UTC=-0.3470650, Accumulated Leap Seconds=16
  { 2443509.50, 48.534416 }, //01 January 1978, UT1-UTC= 0.6495837, Accumulated Leap Seconds=17
  { 2443510.50, 48.537830 }, //02 January 1978, UT1-UTC= 0.6461695, Accumulated Leap Seconds=17
  { 2443511.50, 48.541271 }, //03 January 1978, UT1-UTC= 0.6427289, Accumulated Leap Seconds=17
  { 2443512.50, 48.544695 }, //04 January 1978, UT1-UTC= 0.6393046, Accumulated Leap Seconds=17
  { 2443513.50, 48.548064 }, //05 January 1978, UT1-UTC= 0.6359360, Accumulated Leap Seconds=17
  { 2443514.50, 48.551353 }, //06 January 1978, UT1-UTC= 0.6326465, Accumulated Leap Seconds=17
  { 2443515.50, 48.554568 }, //07 January 1978, UT1-UTC= 0.6294315, Accumulated Leap Seconds=17
  { 2443516.50, 48.557744 }, //08 January 1978, UT1-UTC= 0.6262563, Accumulated Leap Seconds=17
  { 2443517.50, 48.560932 }, //09 January 1978, UT1-UTC= 0.6230682, Accumulated Leap Seconds=17
  { 2443518.50, 48.564180 }, //10 January 1978, UT1-UTC= 0.6198196, Accumulated Leap Seconds=17
  { 2443519.50, 48.567511 }, //11 January 1978, UT1-UTC= 0.6164887, Accumulated Leap Seconds=17
  { 2443520.50, 48.570909 }, //12 January 1978, UT1-UTC= 0.6130910, Accumulated Leap Seconds=17
  { 2443521.50, 48.574328 }, //13 January 1978, UT1-UTC= 0.6096721, Accumulated Leap Seconds=17
  { 2443522.50, 48.577710 }, //14 January 1978, UT1-UTC= 0.6062895, Accumulated Leap Seconds=17
  { 2443523.50, 48.581007 }, //15 January 1978, UT1-UTC= 0.6029926, Accumulated Leap Seconds=17
  { 2443524.50, 48.584187 }, //16 January 1978, UT1-UTC= 0.5998129, Accumulated Leap Seconds=17
  { 2443525.50, 48.587238 }, //17 January 1978, UT1-UTC= 0.5967623, Accumulated Leap Seconds=17
  { 2443526.50, 48.590164 }, //18 January 1978, UT1-UTC= 0.5938363, Accumulated Leap Seconds=17
  { 2443527.50, 48.592982 }, //19 January 1978, UT1-UTC= 0.5910181, Accumulated Leap Seconds=17
  { 2443528.50, 48.595720 }, //20 January 1978, UT1-UTC= 0.5882799, Accumulated Leap Seconds=17
  { 2443529.50, 48.598414 }, //21 January 1978, UT1-UTC= 0.5855859, Accumulated Leap Seconds=17
  { 2443530.50, 48.601104 }, //22 January 1978, UT1-UTC= 0.5828958, Accumulated Leap Seconds=17
  { 2443531.50, 48.603830 }, //23 January 1978, UT1-UTC= 0.5801696, Accumulated Leap Seconds=17
  { 2443532.50, 48.606628 }, //24 January 1978, UT1-UTC= 0.5773717, Accumulated Leap Seconds=17
  { 2443533.50, 48.609526 }, //25 January 1978, UT1-UTC= 0.5744735, Accumulated Leap Seconds=17
  { 2443534.50, 48.612543 }, //26 January 1978, UT1-UTC= 0.5714572, Accumulated Leap Seconds=17
  { 2443535.50, 48.615682 }, //27 January 1978, UT1-UTC= 0.5683182, Accumulated Leap Seconds=17
  { 2443536.50, 48.618933 }, //28 January 1978, UT1-UTC= 0.5650671, Accumulated Leap Seconds=17
  { 2443537.50, 48.622271 }, //29 January 1978, UT1-UTC= 0.5617291, Accumulated Leap Seconds=17
  { 2443538.50, 48.625661 }, //30 January 1978, UT1-UTC= 0.5583390, Accumulated Leap Seconds=17
  { 2443539.50, 48.629065 }, //31 January 1978, UT1-UTC= 0.5549349, Accumulated Leap Seconds=17
  { 2443540.50, 48.632451 }, //01 February 1978, UT1-UTC= 0.5515494, Accumulated Leap Seconds=17
  { 2443541.50, 48.635797 }, //02 February 1978, UT1-UTC= 0.5482029, Accumulated Leap Seconds=17
  { 2443542.50, 48.639105 }, //03 February 1978, UT1-UTC= 0.5448952, Accumulated Leap Seconds=17
  { 2443543.50, 48.642399 }, //04 February 1978, UT1-UTC= 0.5416005, Accumulated Leap Seconds=17
  { 2443544.50, 48.645728 }, //05 February 1978, UT1-UTC= 0.5382717, Accumulated Leap Seconds=17
  { 2443545.50, 48.649144 }, //06 February 1978, UT1-UTC= 0.5348558, Accumulated Leap Seconds=17
  { 2443546.50, 48.652684 }, //07 February 1978, UT1-UTC= 0.5313156, Accumulated Leap Seconds=17
  { 2443547.50, 48.656352 }, //08 February 1978, UT1-UTC= 0.5276480, Accumulated Leap Seconds=17
  { 2443548.50, 48.660112 }, //09 February 1978, UT1-UTC= 0.5238881, Accumulated Leap Seconds=17
  { 2443549.50, 48.663902 }, //10 February 1978, UT1-UTC= 0.5200979, Accumulated Leap Seconds=17
  { 2443550.50, 48.667655 }, //11 February 1978, UT1-UTC= 0.5163453, Accumulated Leap Seconds=17
  { 2443551.50, 48.671316 }, //12 February 1978, UT1-UTC= 0.5126843, Accumulated Leap Seconds=17
  { 2443552.50, 48.674855 }, //13 February 1978, UT1-UTC= 0.5091454, Accumulated Leap Seconds=17
  { 2443553.50, 48.678265 }, //14 February 1978, UT1-UTC= 0.5057354, Accumulated Leap Seconds=17
  { 2443554.50, 48.681557 }, //15 February 1978, UT1-UTC= 0.5024432, Accumulated Leap Seconds=17
  { 2443555.50, 48.684754 }, //16 February 1978, UT1-UTC= 0.4992455, Accumulated Leap Seconds=17
  { 2443556.50, 48.687891 }, //17 February 1978, UT1-UTC= 0.4961093, Accumulated Leap Seconds=17
  { 2443557.50, 48.691005 }, //18 February 1978, UT1-UTC= 0.4929954, Accumulated Leap Seconds=17
  { 2443558.50, 48.694137 }, //19 February 1978, UT1-UTC= 0.4898629, Accumulated Leap Seconds=17
  { 2443559.50, 48.697327 }, //20 February 1978, UT1-UTC= 0.4866734, Accumulated Leap Seconds=17
  { 2443560.50, 48.700604 }, //21 February 1978, UT1-UTC= 0.4833960, Accumulated Leap Seconds=17
  { 2443561.50, 48.703990 }, //22 February 1978, UT1-UTC= 0.4800104, Accumulated Leap Seconds=17
  { 2443562.50, 48.707489 }, //23 February 1978, UT1-UTC= 0.4765106, Accumulated Leap Seconds=17
  { 2443563.50, 48.711093 }, //24 February 1978, UT1-UTC= 0.4729070, Accumulated Leap Seconds=17
  { 2443564.50, 48.714772 }, //25 February 1978, UT1-UTC= 0.4692277, Accumulated Leap Seconds=17
  { 2443565.50, 48.718484 }, //26 February 1978, UT1-UTC= 0.4655161, Accumulated Leap Seconds=17
  { 2443566.50, 48.722179 }, //27 February 1978, UT1-UTC= 0.4618208, Accumulated Leap Seconds=17
  { 2443567.50, 48.725817 }, //28 February 1978, UT1-UTC= 0.4581833, Accumulated Leap Seconds=17
  { 2443568.50, 48.729372 }, //01 March 1978, UT1-UTC= 0.4546282, Accumulated Leap Seconds=17
  { 2443569.50, 48.732844 }, //02 March 1978, UT1-UTC= 0.4511561, Accumulated Leap Seconds=17
  { 2443570.50, 48.736258 }, //03 March 1978, UT1-UTC= 0.4477425, Accumulated Leap Seconds=17
  { 2443571.50, 48.739658 }, //04 March 1978, UT1-UTC= 0.4443420, Accumulated Leap Seconds=17
  { 2443572.50, 48.743101 }, //05 March 1978, UT1-UTC= 0.4408993, Accumulated Leap Seconds=17
  { 2443573.50, 48.746634 }, //06 March 1978, UT1-UTC= 0.4373656, Accumulated Leap Seconds=17
  { 2443574.50, 48.750285 }, //07 March 1978, UT1-UTC= 0.4337146, Accumulated Leap Seconds=17
  { 2443575.50, 48.754044 }, //08 March 1978, UT1-UTC= 0.4299562, Accumulated Leap Seconds=17
  { 2443576.50, 48.757865 }, //09 March 1978, UT1-UTC= 0.4261347, Accumulated Leap Seconds=17
  { 2443577.50, 48.761684 }, //10 March 1978, UT1-UTC= 0.4223156, Accumulated Leap Seconds=17
  { 2443578.50, 48.765436 }, //11 March 1978, UT1-UTC= 0.4185641, Accumulated Leap Seconds=17
  { 2443579.50, 48.769072 }, //12 March 1978, UT1-UTC= 0.4149282, Accumulated Leap Seconds=17
  { 2443580.50, 48.772569 }, //13 March 1978, UT1-UTC= 0.4114311, Accumulated Leap Seconds=17
  { 2443581.50, 48.775927 }, //14 March 1978, UT1-UTC= 0.4080726, Accumulated Leap Seconds=17
  { 2443582.50, 48.779164 }, //15 March 1978, UT1-UTC= 0.4048358, Accumulated Leap Seconds=17
  { 2443583.50, 48.782308 }, //16 March 1978, UT1-UTC= 0.4016919, Accumulated Leap Seconds=17
  { 2443584.50, 48.785395 }, //17 March 1978, UT1-UTC= 0.3986047, Accumulated Leap Seconds=17
  { 2443585.50, 48.788467 }, //18 March 1978, UT1-UTC= 0.3955330, Accumulated Leap Seconds=17
  { 2443586.50, 48.791567 }, //19 March 1978, UT1-UTC= 0.3924333, Accumulated Leap Seconds=17
  { 2443587.50, 48.794733 }, //20 March 1978, UT1-UTC= 0.3892666, Accumulated Leap Seconds=17
  { 2443588.50, 48.797997 }, //21 March 1978, UT1-UTC= 0.3860030, Accumulated Leap Seconds=17
  { 2443589.50, 48.801375 }, //22 March 1978, UT1-UTC= 0.3826246, Accumulated Leap Seconds=17
  { 2443590.50, 48.804872 }, //23 March 1978, UT1-UTC= 0.3791281, Accumulated Leap Seconds=17
  { 2443591.50, 48.808471 }, //24 March 1978, UT1-UTC= 0.3755294, Accumulated Leap Seconds=17
  { 2443592.50, 48.812135 }, //25 March 1978, UT1-UTC= 0.3718651, Accumulated Leap Seconds=17
  { 2443593.50, 48.815813 }, //26 March 1978, UT1-UTC= 0.3681869, Accumulated Leap Seconds=17
  { 2443594.50, 48.819451 }, //27 March 1978, UT1-UTC= 0.3645485, Accumulated Leap Seconds=17
  { 2443595.50, 48.823009 }, //28 March 1978, UT1-UTC= 0.3609908, Accumulated Leap Seconds=17
  { 2443596.50, 48.826472 }, //29 March 1978, UT1-UTC= 0.3575284, Accumulated Leap Seconds=17
  { 2443597.50, 48.829853 }, //30 March 1978, UT1-UTC= 0.3541465, Accumulated Leap Seconds=17
  { 2443598.50, 48.833193 }, //31 March 1978, UT1-UTC= 0.3508067, Accumulated Leap Seconds=17
  { 2443599.50, 48.836541 }, //01 April 1978, UT1-UTC= 0.3474586, Accumulated Leap Seconds=17
  { 2443600.50, 48.839946 }, //02 April 1978, UT1-UTC= 0.3440540, Accumulated Leap Seconds=17
  { 2443601.50, 48.843440 }, //03 April 1978, UT1-UTC= 0.3405596, Accumulated Leap Seconds=17
  { 2443602.50, 48.847031 }, //04 April 1978, UT1-UTC= 0.3369688, Accumulated Leap Seconds=17
  { 2443603.50, 48.850695 }, //05 April 1978, UT1-UTC= 0.3333051, Accumulated Leap Seconds=17
  { 2443604.50, 48.854382 }, //06 April 1978, UT1-UTC= 0.3296181, Accumulated Leap Seconds=17
  { 2443605.50, 48.858031 }, //07 April 1978, UT1-UTC= 0.3259694, Accumulated Leap Seconds=17
  { 2443606.50, 48.861585 }, //08 April 1978, UT1-UTC= 0.3224153, Accumulated Leap Seconds=17
  { 2443607.50, 48.865008 }, //09 April 1978, UT1-UTC= 0.3189922, Accumulated Leap Seconds=17
  { 2443608.50, 48.868288 }, //10 April 1978, UT1-UTC= 0.3157120, Accumulated Leap Seconds=17
  { 2443609.50, 48.871434 }, //11 April 1978, UT1-UTC= 0.3125655, Accumulated Leap Seconds=17
  { 2443610.50, 48.874472 }, //12 April 1978, UT1-UTC= 0.3095284, Accumulated Leap Seconds=17
  { 2443611.50, 48.877433 }, //13 April 1978, UT1-UTC= 0.3065667, Accumulated Leap Seconds=17
  { 2443612.50, 48.880360 }, //14 April 1978, UT1-UTC= 0.3036402, Accumulated Leap Seconds=17
  { 2443613.50, 48.883295 }, //15 April 1978, UT1-UTC= 0.3007051, Accumulated Leap Seconds=17
  { 2443614.50, 48.886280 }, //16 April 1978, UT1-UTC= 0.2977200, Accumulated Leap Seconds=17
  { 2443615.50, 48.889348 }, //17 April 1978, UT1-UTC= 0.2946518, Accumulated Leap Seconds=17
  { 2443616.50, 48.892520 }, //18 April 1978, UT1-UTC= 0.2914803, Accumulated Leap Seconds=17
  { 2443617.50, 48.895802 }, //19 April 1978, UT1-UTC= 0.2881979, Accumulated Leap Seconds=17
  { 2443618.50, 48.899188 }, //20 April 1978, UT1-UTC= 0.2848121, Accumulated Leap Seconds=17
  { 2443619.50, 48.902652 }, //21 April 1978, UT1-UTC= 0.2813478, Accumulated Leap Seconds=17
  { 2443620.50, 48.906151 }, //22 April 1978, UT1-UTC= 0.2778487, Accumulated Leap Seconds=17
  { 2443621.50, 48.909630 }, //23 April 1978, UT1-UTC= 0.2743702, Accumulated Leap Seconds=17
  { 2443622.50, 48.913034 }, //24 April 1978, UT1-UTC= 0.2709660, Accumulated Leap Seconds=17
  { 2443623.50, 48.916332 }, //25 April 1978, UT1-UTC= 0.2676683, Accumulated Leap Seconds=17
  { 2443624.50, 48.919523 }, //26 April 1978, UT1-UTC= 0.2644766, Accumulated Leap Seconds=17
  { 2443625.50, 48.922642 }, //27 April 1978, UT1-UTC= 0.2613583, Accumulated Leap Seconds=17
  { 2443626.50, 48.925738 }, //28 April 1978, UT1-UTC= 0.2582621, Accumulated Leap Seconds=17
  { 2443627.50, 48.928864 }, //29 April 1978, UT1-UTC= 0.2551362, Accumulated Leap Seconds=17
  { 2443628.50, 48.932055 }, //30 April 1978, UT1-UTC= 0.2519451, Accumulated Leap Seconds=17
  { 2443629.50, 48.935323 }, //01 May 1978, UT1-UTC= 0.2486768, Accumulated Leap Seconds=17
  { 2443630.50, 48.938655 }, //02 May 1978, UT1-UTC= 0.2453449, Accumulated Leap Seconds=17
  { 2443631.50, 48.942016 }, //03 May 1978, UT1-UTC= 0.2419836, Accumulated Leap Seconds=17
  { 2443632.50, 48.945361 }, //04 May 1978, UT1-UTC= 0.2386393, Accumulated Leap Seconds=17
  { 2443633.50, 48.948641 }, //05 May 1978, UT1-UTC= 0.2353591, Accumulated Leap Seconds=17
  { 2443634.50, 48.951820 }, //06 May 1978, UT1-UTC= 0.2321799, Accumulated Leap Seconds=17
  { 2443635.50, 48.954880 }, //07 May 1978, UT1-UTC= 0.2291195, Accumulated Leap Seconds=17
  { 2443636.50, 48.957824 }, //08 May 1978, UT1-UTC= 0.2261756, Accumulated Leap Seconds=17
  { 2443637.50, 48.960672 }, //09 May 1978, UT1-UTC= 0.2233282, Accumulated Leap Seconds=17
  { 2443638.50, 48.963454 }, //10 May 1978, UT1-UTC= 0.2205458, Accumulated Leap Seconds=17
  { 2443639.50, 48.966207 }, //11 May 1978, UT1-UTC= 0.2177930, Accumulated Leap Seconds=17
  { 2443640.50, 48.968968 }, //12 May 1978, UT1-UTC= 0.2150318, Accumulated Leap Seconds=17
  { 2443641.50, 48.971775 }, //13 May 1978, UT1-UTC= 0.2122252, Accumulated Leap Seconds=17
  { 2443642.50, 48.974659 }, //14 May 1978, UT1-UTC= 0.2093413, Accumulated Leap Seconds=17
  { 2443643.50, 48.977642 }, //15 May 1978, UT1-UTC= 0.2063578, Accumulated Leap Seconds=17
  { 2443644.50, 48.980736 }, //16 May 1978, UT1-UTC= 0.2032638, Accumulated Leap Seconds=17
  { 2443645.50, 48.983939 }, //17 May 1978, UT1-UTC= 0.2000605, Accumulated Leap Seconds=17
  { 2443646.50, 48.987238 }, //18 May 1978, UT1-UTC= 0.1967624, Accumulated Leap Seconds=17
  { 2443647.50, 48.990601 }, //19 May 1978, UT1-UTC= 0.1933993, Accumulated Leap Seconds=17
  { 2443648.50, 48.993984 }, //20 May 1978, UT1-UTC= 0.1900161, Accumulated Leap Seconds=17
  { 2443649.50, 48.997335 }, //21 May 1978, UT1-UTC= 0.1866651, Accumulated Leap Seconds=17
  { 2443650.50, 49.000609 }, //22 May 1978, UT1-UTC= 0.1833913, Accumulated Leap Seconds=17
  { 2443651.50, 49.003785 }, //23 May 1978, UT1-UTC= 0.1802152, Accumulated Leap Seconds=17
  { 2443652.50, 49.006877 }, //24 May 1978, UT1-UTC= 0.1771234, Accumulated Leap Seconds=17
  { 2443653.50, 49.009926 }, //25 May 1978, UT1-UTC= 0.1740744, Accumulated Leap Seconds=17
  { 2443654.50, 49.012983 }, //26 May 1978, UT1-UTC= 0.1710172, Accumulated Leap Seconds=17
  { 2443655.50, 49.016087 }, //27 May 1978, UT1-UTC= 0.1679135, Accumulated Leap Seconds=17
  { 2443656.50, 49.019247 }, //28 May 1978, UT1-UTC= 0.1647525, Accumulated Leap Seconds=17
  { 2443657.50, 49.022448 }, //29 May 1978, UT1-UTC= 0.1615517, Accumulated Leap Seconds=17
  { 2443658.50, 49.025651 }, //30 May 1978, UT1-UTC= 0.1583492, Accumulated Leap Seconds=17
  { 2443659.50, 49.028809 }, //31 May 1978, UT1-UTC= 0.1551910, Accumulated Leap Seconds=17
  { 2443660.50, 49.031878 }, //01 June 1978, UT1-UTC= 0.1521219, Accumulated Leap Seconds=17
  { 2443661.50, 49.034820 }, //02 June 1978, UT1-UTC= 0.1491796, Accumulated Leap Seconds=17
  { 2443662.50, 49.037612 }, //03 June 1978, UT1-UTC= 0.1463881, Accumulated Leap Seconds=17
  { 2443663.50, 49.040245 }, //04 June 1978, UT1-UTC= 0.1437546, Accumulated Leap Seconds=17
  { 2443664.50, 49.042730 }, //05 June 1978, UT1-UTC= 0.1412696, Accumulated Leap Seconds=17
  { 2443665.50, 49.045089 }, //06 June 1978, UT1-UTC= 0.1389108, Accumulated Leap Seconds=17
  { 2443666.50, 49.047352 }, //07 June 1978, UT1-UTC= 0.1366481, Accumulated Leap Seconds=17
  { 2443667.50, 49.049553 }, //08 June 1978, UT1-UTC= 0.1344472, Accumulated Leap Seconds=17
  { 2443668.50, 49.051730 }, //09 June 1978, UT1-UTC= 0.1322703, Accumulated Leap Seconds=17
  { 2443669.50, 49.053917 }, //10 June 1978, UT1-UTC= 0.1300826, Accumulated Leap Seconds=17
  { 2443670.50, 49.056143 }, //11 June 1978, UT1-UTC= 0.1278566, Accumulated Leap Seconds=17
  { 2443671.50, 49.058425 }, //12 June 1978, UT1-UTC= 0.1255754, Accumulated Leap Seconds=17
  { 2443672.50, 49.060766 }, //13 June 1978, UT1-UTC= 0.1232336, Accumulated Leap Seconds=17
  { 2443673.50, 49.063163 }, //14 June 1978, UT1-UTC= 0.1208370, Accumulated Leap Seconds=17
  { 2443674.50, 49.065597 }, //15 June 1978, UT1-UTC= 0.1184027, Accumulated Leap Seconds=17
  { 2443675.50, 49.068041 }, //16 June 1978, UT1-UTC= 0.1159589, Accumulated Leap Seconds=17
  { 2443676.50, 49.070457 }, //17 June 1978, UT1-UTC= 0.1135425, Accumulated Leap Seconds=17
  { 2443677.50, 49.072809 }, //18 June 1978, UT1-UTC= 0.1111914, Accumulated Leap Seconds=17
  { 2443678.50, 49.075070 }, //19 June 1978, UT1-UTC= 0.1089300, Accumulated Leap Seconds=17
  { 2443679.50, 49.077244 }, //20 June 1978, UT1-UTC= 0.1067559, Accumulated Leap Seconds=17
  { 2443680.50, 49.079365 }, //21 June 1978, UT1-UTC= 0.1046347, Accumulated Leap Seconds=17
  { 2443681.50, 49.081488 }, //22 June 1978, UT1-UTC= 0.1025116, Accumulated Leap Seconds=17
  { 2443682.50, 49.083666 }, //23 June 1978, UT1-UTC= 0.1003337, Accumulated Leap Seconds=17
  { 2443683.50, 49.085926 }, //24 June 1978, UT1-UTC= 0.0980735, Accumulated Leap Seconds=17
  { 2443684.50, 49.088260 }, //25 June 1978, UT1-UTC= 0.0957399, Accumulated Leap Seconds=17
  { 2443685.50, 49.090627 }, //26 June 1978, UT1-UTC= 0.0933732, Accumulated Leap Seconds=17
  { 2443686.50, 49.092971 }, //27 June 1978, UT1-UTC= 0.0910285, Accumulated Leap Seconds=17
  { 2443687.50, 49.095242 }, //28 June 1978, UT1-UTC= 0.0887581, Accumulated Leap Seconds=17
  { 2443688.50, 49.097400 }, //29 June 1978, UT1-UTC= 0.0865997, Accumulated Leap Seconds=17
  { 2443689.50, 49.099427 }, //30 June 1978, UT1-UTC= 0.0845729, Accumulated Leap Seconds=17
  { 2443690.50, 49.101320 }, //01 July 1978, UT1-UTC= 0.0826795, Accumulated Leap Seconds=17
  { 2443691.50, 49.103094 }, //02 July 1978, UT1-UTC= 0.0809063, Accumulated Leap Seconds=17
  { 2443692.50, 49.104772 }, //03 July 1978, UT1-UTC= 0.0792280, Accumulated Leap Seconds=17
  { 2443693.50, 49.106389 }, //04 July 1978, UT1-UTC= 0.0776105, Accumulated Leap Seconds=17
  { 2443694.50, 49.107984 }, //05 July 1978, UT1-UTC= 0.0760160, Accumulated Leap Seconds=17
  { 2443695.50, 49.109593 }, //06 July 1978, UT1-UTC= 0.0744070, Accumulated Leap Seconds=17
  { 2443696.50, 49.111250 }, //07 July 1978, UT1-UTC= 0.0727501, Accumulated Leap Seconds=17
  { 2443697.50, 49.112980 }, //08 July 1978, UT1-UTC= 0.0710202, Accumulated Leap Seconds=17
  { 2443698.50, 49.114798 }, //09 July 1978, UT1-UTC= 0.0692022, Accumulated Leap Seconds=17
  { 2443699.50, 49.116707 }, //10 July 1978, UT1-UTC= 0.0672930, Accumulated Leap Seconds=17
  { 2443700.50, 49.118698 }, //11 July 1978, UT1-UTC= 0.0653022, Accumulated Leap Seconds=17
  { 2443701.50, 49.120750 }, //12 July 1978, UT1-UTC= 0.0632501, Accumulated Leap Seconds=17
  { 2443702.50, 49.122834 }, //13 July 1978, UT1-UTC= 0.0611657, Accumulated Leap Seconds=17
  { 2443703.50, 49.124918 }, //14 July 1978, UT1-UTC= 0.0590820, Accumulated Leap Seconds=17
  { 2443704.50, 49.126967 }, //15 July 1978, UT1-UTC= 0.0570327, Accumulated Leap Seconds=17
  { 2443705.50, 49.128956 }, //16 July 1978, UT1-UTC= 0.0550440, Accumulated Leap Seconds=17
  { 2443706.50, 49.130876 }, //17 July 1978, UT1-UTC= 0.0531236, Accumulated Leap Seconds=17
  { 2443707.50, 49.132748 }, //18 July 1978, UT1-UTC= 0.0512523, Accumulated Leap Seconds=17
  { 2443708.50, 49.134615 }, //19 July 1978, UT1-UTC= 0.0493852, Accumulated Leap Seconds=17
  { 2443709.50, 49.136533 }, //20 July 1978, UT1-UTC= 0.0474668, Accumulated Leap Seconds=17
  { 2443710.50, 49.138545 }, //21 July 1978, UT1-UTC= 0.0454547, Accumulated Leap Seconds=17
  { 2443711.50, 49.140660 }, //22 July 1978, UT1-UTC= 0.0433395, Accumulated Leap Seconds=17
  { 2443712.50, 49.142847 }, //23 July 1978, UT1-UTC= 0.0411526, Accumulated Leap Seconds=17
  { 2443713.50, 49.145046 }, //24 July 1978, UT1-UTC= 0.0389542, Accumulated Leap Seconds=17
  { 2443714.50, 49.147189 }, //25 July 1978, UT1-UTC= 0.0368111, Accumulated Leap Seconds=17
  { 2443715.50, 49.149225 }, //26 July 1978, UT1-UTC= 0.0347747, Accumulated Leap Seconds=17
  { 2443716.50, 49.151128 }, //27 July 1978, UT1-UTC= 0.0328718, Accumulated Leap Seconds=17
  { 2443717.50, 49.152895 }, //28 July 1978, UT1-UTC= 0.0311050, Accumulated Leap Seconds=17
  { 2443718.50, 49.154542 }, //29 July 1978, UT1-UTC= 0.0294582, Accumulated Leap Seconds=17
  { 2443719.50, 49.156097 }, //30 July 1978, UT1-UTC= 0.0279029, Accumulated Leap Seconds=17
  { 2443720.50, 49.157598 }, //31 July 1978, UT1-UTC= 0.0264023, Accumulated Leap Seconds=17
  { 2443721.50, 49.159084 }, //01 August 1978, UT1-UTC= 0.0249156, Accumulated Leap Seconds=17
  { 2443722.50, 49.160598 }, //02 August 1978, UT1-UTC= 0.0234018, Accumulated Leap Seconds=17
  { 2443723.50, 49.162176 }, //03 August 1978, UT1-UTC= 0.0218238, Accumulated Leap Seconds=17
  { 2443724.50, 49.163847 }, //04 August 1978, UT1-UTC= 0.0201525, Accumulated Leap Seconds=17
  { 2443725.50, 49.165632 }, //05 August 1978, UT1-UTC= 0.0183683, Accumulated Leap Seconds=17
  { 2443726.50, 49.167537 }, //06 August 1978, UT1-UTC= 0.0164629, Accumulated Leap Seconds=17
  { 2443727.50, 49.169559 }, //07 August 1978, UT1-UTC= 0.0144413, Accumulated Leap Seconds=17
  { 2443728.50, 49.171679 }, //08 August 1978, UT1-UTC= 0.0123208, Accumulated Leap Seconds=17
  { 2443729.50, 49.173871 }, //09 August 1978, UT1-UTC= 0.0101289, Accumulated Leap Seconds=17
  { 2443730.50, 49.176102 }, //10 August 1978, UT1-UTC= 0.0078977, Accumulated Leap Seconds=17
  { 2443731.50, 49.178342 }, //11 August 1978, UT1-UTC= 0.0056578, Accumulated Leap Seconds=17
  { 2443732.50, 49.180567 }, //12 August 1978, UT1-UTC= 0.0034327, Accumulated Leap Seconds=17
  { 2443733.50, 49.182769 }, //13 August 1978, UT1-UTC= 0.0012305, Accumulated Leap Seconds=17
  { 2443734.50, 49.184962 }, //14 August 1978, UT1-UTC=-0.0009617, Accumulated Leap Seconds=17
  { 2443735.50, 49.187180 }, //15 August 1978, UT1-UTC=-0.0031802, Accumulated Leap Seconds=17
  { 2443736.50, 49.189477 }, //16 August 1978, UT1-UTC=-0.0054772, Accumulated Leap Seconds=17
  { 2443737.50, 49.191904 }, //17 August 1978, UT1-UTC=-0.0079042, Accumulated Leap Seconds=17
  { 2443738.50, 49.194489 }, //18 August 1978, UT1-UTC=-0.0104891, Accumulated Leap Seconds=17
  { 2443739.50, 49.197220 }, //19 August 1978, UT1-UTC=-0.0132203, Accumulated Leap Seconds=17
  { 2443740.50, 49.200046 }, //20 August 1978, UT1-UTC=-0.0160459, Accumulated Leap Seconds=17
  { 2443741.50, 49.202889 }, //21 August 1978, UT1-UTC=-0.0188891, Accumulated Leap Seconds=17
  { 2443742.50, 49.205673 }, //22 August 1978, UT1-UTC=-0.0216734, Accumulated Leap Seconds=17
  { 2443743.50, 49.208343 }, //23 August 1978, UT1-UTC=-0.0243430, Accumulated Leap Seconds=17
  { 2443744.50, 49.210871 }, //24 August 1978, UT1-UTC=-0.0268715, Accumulated Leap Seconds=17
  { 2443745.50, 49.213260 }, //25 August 1978, UT1-UTC=-0.0292597, Accumulated Leap Seconds=17
  { 2443746.50, 49.215528 }, //26 August 1978, UT1-UTC=-0.0315276, Accumulated Leap Seconds=17
  { 2443747.50, 49.217706 }, //27 August 1978, UT1-UTC=-0.0337064, Accumulated Leap Seconds=17
  { 2443748.50, 49.219834 }, //28 August 1978, UT1-UTC=-0.0358339, Accumulated Leap Seconds=17
  { 2443749.50, 49.221949 }, //29 August 1978, UT1-UTC=-0.0379492, Accumulated Leap Seconds=17
  { 2443750.50, 49.224089 }, //30 August 1978, UT1-UTC=-0.0400888, Accumulated Leap Seconds=17
  { 2443751.50, 49.226283 }, //31 August 1978, UT1-UTC=-0.0422832, Accumulated Leap Seconds=17
  { 2443752.50, 49.228553 }, //01 September 1978, UT1-UTC=-0.0445534, Accumulated Leap Seconds=17
  { 2443753.50, 49.230909 }, //02 September 1978, UT1-UTC=-0.0469087, Accumulated Leap Seconds=17
  { 2443754.50, 49.233346 }, //03 September 1978, UT1-UTC=-0.0493461, Accumulated Leap Seconds=17
  { 2443755.50, 49.235851 }, //04 September 1978, UT1-UTC=-0.0518512, Accumulated Leap Seconds=17
  { 2443756.50, 49.238399 }, //05 September 1978, UT1-UTC=-0.0543988, Accumulated Leap Seconds=17
  { 2443757.50, 49.240955 }, //06 September 1978, UT1-UTC=-0.0569553, Accumulated Leap Seconds=17
  { 2443758.50, 49.243486 }, //07 September 1978, UT1-UTC=-0.0594862, Accumulated Leap Seconds=17
  { 2443759.50, 49.245966 }, //08 September 1978, UT1-UTC=-0.0619664, Accumulated Leap Seconds=17
  { 2443760.50, 49.248387 }, //09 September 1978, UT1-UTC=-0.0643867, Accumulated Leap Seconds=17
  { 2443761.50, 49.250759 }, //10 September 1978, UT1-UTC=-0.0667587, Accumulated Leap Seconds=17
  { 2443762.50, 49.253116 }, //11 September 1978, UT1-UTC=-0.0691163, Accumulated Leap Seconds=17
  { 2443763.50, 49.255510 }, //12 September 1978, UT1-UTC=-0.0715105, Accumulated Leap Seconds=17
  { 2443764.50, 49.257997 }, //13 September 1978, UT1-UTC=-0.0739971, Accumulated Leap Seconds=17
  { 2443765.50, 49.260620 }, //14 September 1978, UT1-UTC=-0.0766204, Accumulated Leap Seconds=17
  { 2443766.50, 49.263396 }, //15 September 1978, UT1-UTC=-0.0793961, Accumulated Leap Seconds=17
  { 2443767.50, 49.266301 }, //16 September 1978, UT1-UTC=-0.0823006, Accumulated Leap Seconds=17
  { 2443768.50, 49.269275 }, //17 September 1978, UT1-UTC=-0.0852752, Accumulated Leap Seconds=17
  { 2443769.50, 49.272244 }, //18 September 1978, UT1-UTC=-0.0882444, Accumulated Leap Seconds=17
  { 2443770.50, 49.275140 }, //19 September 1978, UT1-UTC=-0.0911401, Accumulated Leap Seconds=17
  { 2443771.50, 49.277920 }, //20 September 1978, UT1-UTC=-0.0939198, Accumulated Leap Seconds=17
  { 2443772.50, 49.280573 }, //21 September 1978, UT1-UTC=-0.0965733, Accumulated Leap Seconds=17
  { 2443773.50, 49.283117 }, //22 September 1978, UT1-UTC=-0.0991171, Accumulated Leap Seconds=17
  { 2443774.50, 49.285584 }, //23 September 1978, UT1-UTC=-0.1015837, Accumulated Leap Seconds=17
  { 2443775.50, 49.288013 }, //24 September 1978, UT1-UTC=-0.1040132, Accumulated Leap Seconds=17
  { 2443776.50, 49.290449 }, //25 September 1978, UT1-UTC=-0.1064490, Accumulated Leap Seconds=17
  { 2443777.50, 49.292933 }, //26 September 1978, UT1-UTC=-0.1089332, Accumulated Leap Seconds=17
  { 2443778.50, 49.295502 }, //27 September 1978, UT1-UTC=-0.1115022, Accumulated Leap Seconds=17
  { 2443779.50, 49.298182 }, //28 September 1978, UT1-UTC=-0.1141821, Accumulated Leap Seconds=17
  { 2443780.50, 49.300987 }, //29 September 1978, UT1-UTC=-0.1169875, Accumulated Leap Seconds=17
  { 2443781.50, 49.303920 }, //30 September 1978, UT1-UTC=-0.1199202, Accumulated Leap Seconds=17
  { 2443782.50, 49.306968 }, //01 October 1978, UT1-UTC=-0.1229683, Accumulated Leap Seconds=17
  { 2443783.50, 49.310106 }, //02 October 1978, UT1-UTC=-0.1261057, Accumulated Leap Seconds=17
  { 2443784.50, 49.313294 }, //03 October 1978, UT1-UTC=-0.1292935, Accumulated Leap Seconds=17
  { 2443785.50, 49.316486 }, //04 October 1978, UT1-UTC=-0.1324859, Accumulated Leap Seconds=17
  { 2443786.50, 49.319641 }, //05 October 1978, UT1-UTC=-0.1356407, Accumulated Leap Seconds=17
  { 2443787.50, 49.322731 }, //06 October 1978, UT1-UTC=-0.1387315, Accumulated Leap Seconds=17
  { 2443788.50, 49.325757 }, //07 October 1978, UT1-UTC=-0.1417569, Accumulated Leap Seconds=17
  { 2443789.50, 49.328741 }, //08 October 1978, UT1-UTC=-0.1447412, Accumulated Leap Seconds=17
  { 2443790.50, 49.331727 }, //09 October 1978, UT1-UTC=-0.1477273, Accumulated Leap Seconds=17
  { 2443791.50, 49.334766 }, //10 October 1978, UT1-UTC=-0.1507659, Accumulated Leap Seconds=17
  { 2443792.50, 49.337901 }, //11 October 1978, UT1-UTC=-0.1539010, Accumulated Leap Seconds=17
  { 2443793.50, 49.341156 }, //12 October 1978, UT1-UTC=-0.1571565, Accumulated Leap Seconds=17
  { 2443794.50, 49.344525 }, //13 October 1978, UT1-UTC=-0.1605252, Accumulated Leap Seconds=17
  { 2443795.50, 49.347967 }, //14 October 1978, UT1-UTC=-0.1639674, Accumulated Leap Seconds=17
  { 2443796.50, 49.351420 }, //15 October 1978, UT1-UTC=-0.1674204, Accumulated Leap Seconds=17
  { 2443797.50, 49.354814 }, //16 October 1978, UT1-UTC=-0.1708141, Accumulated Leap Seconds=17
  { 2443798.50, 49.358091 }, //17 October 1978, UT1-UTC=-0.1740909, Accumulated Leap Seconds=17
  { 2443799.50, 49.361217 }, //18 October 1978, UT1-UTC=-0.1772174, Accumulated Leap Seconds=17
  { 2443800.50, 49.364189 }, //19 October 1978, UT1-UTC=-0.1801891, Accumulated Leap Seconds=17
  { 2443801.50, 49.367025 }, //20 October 1978, UT1-UTC=-0.1830246, Accumulated Leap Seconds=17
  { 2443802.50, 49.369759 }, //21 October 1978, UT1-UTC=-0.1857591, Accumulated Leap Seconds=17
  { 2443803.50, 49.372434 }, //22 October 1978, UT1-UTC=-0.1884340, Accumulated Leap Seconds=17
  { 2443804.50, 49.375091 }, //23 October 1978, UT1-UTC=-0.1910914, Accumulated Leap Seconds=17
  { 2443805.50, 49.377771 }, //24 October 1978, UT1-UTC=-0.1937706, Accumulated Leap Seconds=17
  { 2443806.50, 49.380505 }, //25 October 1978, UT1-UTC=-0.1965053, Accumulated Leap Seconds=17
  { 2443807.50, 49.383321 }, //26 October 1978, UT1-UTC=-0.1993213, Accumulated Leap Seconds=17
  { 2443808.50, 49.386232 }, //27 October 1978, UT1-UTC=-0.2022321, Accumulated Leap Seconds=17
  { 2443809.50, 49.389239 }, //28 October 1978, UT1-UTC=-0.2052389, Accumulated Leap Seconds=17
  { 2443810.50, 49.392330 }, //29 October 1978, UT1-UTC=-0.2083301, Accumulated Leap Seconds=17
  { 2443811.50, 49.395479 }, //30 October 1978, UT1-UTC=-0.2114792, Accumulated Leap Seconds=17
  { 2443812.50, 49.398648 }, //31 October 1978, UT1-UTC=-0.2146480, Accumulated Leap Seconds=17
  { 2443813.50, 49.401794 }, //01 November 1978, UT1-UTC=-0.2177939, Accumulated Leap Seconds=17
  { 2443814.50, 49.404882 }, //02 November 1978, UT1-UTC=-0.2208821, Accumulated Leap Seconds=17
  { 2443815.50, 49.407898 }, //03 November 1978, UT1-UTC=-0.2238985, Accumulated Leap Seconds=17
  { 2443816.50, 49.410858 }, //04 November 1978, UT1-UTC=-0.2268581, Accumulated Leap Seconds=17
  { 2443817.50, 49.413801 }, //05 November 1978, UT1-UTC=-0.2298014, Accumulated Leap Seconds=17
  { 2443818.50, 49.416781 }, //06 November 1978, UT1-UTC=-0.2327809, Accumulated Leap Seconds=17
  { 2443819.50, 49.419844 }, //07 November 1978, UT1-UTC=-0.2358440, Accumulated Leap Seconds=17
  { 2443820.50, 49.423018 }, //08 November 1978, UT1-UTC=-0.2390181, Accumulated Leap Seconds=17
  { 2443821.50, 49.426304 }, //09 November 1978, UT1-UTC=-0.2423041, Accumulated Leap Seconds=17
  { 2443822.50, 49.429675 }, //10 November 1978, UT1-UTC=-0.2456753, Accumulated Leap Seconds=17
  { 2443823.50, 49.433084 }, //11 November 1978, UT1-UTC=-0.2490841, Accumulated Leap Seconds=17
  { 2443824.50, 49.436473 }, //12 November 1978, UT1-UTC=-0.2524725, Accumulated Leap Seconds=17
  { 2443825.50, 49.439786 }, //13 November 1978, UT1-UTC=-0.2557861, Accumulated Leap Seconds=17
  { 2443826.50, 49.442986 }, //14 November 1978, UT1-UTC=-0.2589864, Accumulated Leap Seconds=17
  { 2443827.50, 49.446058 }, //15 November 1978, UT1-UTC=-0.2620583, Accumulated Leap Seconds=17
  { 2443828.50, 49.449011 }, //16 November 1978, UT1-UTC=-0.2650110, Accumulated Leap Seconds=17
  { 2443829.50, 49.451871 }, //17 November 1978, UT1-UTC=-0.2678712, Accumulated Leap Seconds=17
  { 2443830.50, 49.454675 }, //18 November 1978, UT1-UTC=-0.2706751, Accumulated Leap Seconds=17
  { 2443831.50, 49.457462 }, //19 November 1978, UT1-UTC=-0.2734620, Accumulated Leap Seconds=17
  { 2443832.50, 49.460269 }, //20 November 1978, UT1-UTC=-0.2762690, Accumulated Leap Seconds=17
  { 2443833.50, 49.463128 }, //21 November 1978, UT1-UTC=-0.2791284, Accumulated Leap Seconds=17
  { 2443834.50, 49.466064 }, //22 November 1978, UT1-UTC=-0.2820636, Accumulated Leap Seconds=17
  { 2443835.50, 49.469087 }, //23 November 1978, UT1-UTC=-0.2850868, Accumulated Leap Seconds=17
  { 2443836.50, 49.472198 }, //24 November 1978, UT1-UTC=-0.2881979, Accumulated Leap Seconds=17
  { 2443837.50, 49.475386 }, //25 November 1978, UT1-UTC=-0.2913857, Accumulated Leap Seconds=17
  { 2443838.50, 49.478629 }, //26 November 1978, UT1-UTC=-0.2946290, Accumulated Leap Seconds=17
  { 2443839.50, 49.481897 }, //27 November 1978, UT1-UTC=-0.2978967, Accumulated Leap Seconds=17
  { 2443840.50, 49.485150 }, //28 November 1978, UT1-UTC=-0.3011505, Accumulated Leap Seconds=17
  { 2443841.50, 49.488353 }, //29 November 1978, UT1-UTC=-0.3043527, Accumulated Leap Seconds=17
  { 2443842.50, 49.491478 }, //30 November 1978, UT1-UTC=-0.3074782, Accumulated Leap Seconds=17
  { 2443843.50, 49.494526 }, //01 December 1978, UT1-UTC=-0.3105263, Accumulated Leap Seconds=17
  { 2443844.50, 49.497526 }, //02 December 1978, UT1-UTC=-0.3135263, Accumulated Leap Seconds=17
  { 2443845.50, 49.500529 }, //03 December 1978, UT1-UTC=-0.3165286, Accumulated Leap Seconds=17
  { 2443846.50, 49.503585 }, //04 December 1978, UT1-UTC=-0.3195849, Accumulated Leap Seconds=17
  { 2443847.50, 49.506729 }, //05 December 1978, UT1-UTC=-0.3227288, Accumulated Leap Seconds=17
  { 2443848.50, 49.509963 }, //06 December 1978, UT1-UTC=-0.3259629, Accumulated Leap Seconds=17
  { 2443849.50, 49.513260 }, //07 December 1978, UT1-UTC=-0.3292599, Accumulated Leap Seconds=17
  { 2443850.50, 49.516572 }, //08 December 1978, UT1-UTC=-0.3325722, Accumulated Leap Seconds=17
  { 2443851.50, 49.519846 }, //09 December 1978, UT1-UTC=-0.3358459, Accumulated Leap Seconds=17
  { 2443852.50, 49.523032 }, //10 December 1978, UT1-UTC=-0.3390324, Accumulated Leap Seconds=17
  { 2443853.50, 49.526098 }, //11 December 1978, UT1-UTC=-0.3420976, Accumulated Leap Seconds=17
  { 2443854.50, 49.529027 }, //12 December 1978, UT1-UTC=-0.3450266, Accumulated Leap Seconds=17
  { 2443855.50, 49.531825 }, //13 December 1978, UT1-UTC=-0.3478253, Accumulated Leap Seconds=17
  { 2443856.50, 49.534516 }, //14 December 1978, UT1-UTC=-0.3505161, Accumulated Leap Seconds=17
  { 2443857.50, 49.537132 }, //15 December 1978, UT1-UTC=-0.3531320, Accumulated Leap Seconds=17
  { 2443858.50, 49.539710 }, //16 December 1978, UT1-UTC=-0.3557104, Accumulated Leap Seconds=17
  { 2443859.50, 49.542288 }, //17 December 1978, UT1-UTC=-0.3582879, Accumulated Leap Seconds=17
  { 2443860.50, 49.544897 }, //18 December 1978, UT1-UTC=-0.3608971, Accumulated Leap Seconds=17
  { 2443861.50, 49.547564 }, //19 December 1978, UT1-UTC=-0.3635638, Accumulated Leap Seconds=17
  { 2443862.50, 49.550306 }, //20 December 1978, UT1-UTC=-0.3663063, Accumulated Leap Seconds=17
  { 2443863.50, 49.553132 }, //21 December 1978, UT1-UTC=-0.3691321, Accumulated Leap Seconds=17
  { 2443864.50, 49.556037 }, //22 December 1978, UT1-UTC=-0.3720375, Accumulated Leap Seconds=17
  { 2443865.50, 49.559010 }, //23 December 1978, UT1-UTC=-0.3750097, Accumulated Leap Seconds=17
  { 2443866.50, 49.562029 }, //24 December 1978, UT1-UTC=-0.3780288, Accumulated Leap Seconds=17
  { 2443867.50, 49.565069 }, //25 December 1978, UT1-UTC=-0.3810693, Accumulated Leap Seconds=17
  { 2443868.50, 49.568103 }, //26 December 1978, UT1-UTC=-0.3841026, Accumulated Leap Seconds=17
  { 2443869.50, 49.571105 }, //27 December 1978, UT1-UTC=-0.3871046, Accumulated Leap Seconds=17
  { 2443870.50, 49.574067 }, //28 December 1978, UT1-UTC=-0.3900665, Accumulated Leap Seconds=17
  { 2443871.50, 49.577006 }, //29 December 1978, UT1-UTC=-0.3930058, Accumulated Leap Seconds=17
  { 2443872.50, 49.579966 }, //30 December 1978, UT1-UTC=-0.3959657, Accumulated Leap Seconds=17
  { 2443873.50, 49.583000 }, //31 December 1978, UT1-UTC=-0.3989996, Accumulated Leap Seconds=17
  { 2443874.50, 49.586149 }, //01 January 1979, UT1-UTC= 0.5978505, Accumulated Leap Seconds=18
  { 2443875.50, 49.589426 }, //02 January 1979, UT1-UTC= 0.5945735, Accumulated Leap Seconds=18
  { 2443876.50, 49.592803 }, //03 January 1979, UT1-UTC= 0.5911970, Accumulated Leap Seconds=18
  { 2443877.50, 49.596222 }, //04 January 1979, UT1-UTC= 0.5877782, Accumulated Leap Seconds=18
  { 2443878.50, 49.599615 }, //05 January 1979, UT1-UTC= 0.5843848, Accumulated Leap Seconds=18
  { 2443879.50, 49.602924 }, //06 January 1979, UT1-UTC= 0.5810764, Accumulated Leap Seconds=18
  { 2443880.50, 49.606108 }, //07 January 1979, UT1-UTC= 0.5778920, Accumulated Leap Seconds=18
  { 2443881.50, 49.609152 }, //08 January 1979, UT1-UTC= 0.5748479, Accumulated Leap Seconds=18
  { 2443882.50, 49.612061 }, //09 January 1979, UT1-UTC= 0.5719391, Accumulated Leap Seconds=18
  { 2443883.50, 49.614856 }, //10 January 1979, UT1-UTC= 0.5691435, Accumulated Leap Seconds=18
  { 2443884.50, 49.617572 }, //11 January 1979, UT1-UTC= 0.5664281, Accumulated Leap Seconds=18
  { 2443885.50, 49.620245 }, //12 January 1979, UT1-UTC= 0.5637548, Accumulated Leap Seconds=18
  { 2443886.50, 49.622914 }, //13 January 1979, UT1-UTC= 0.5610860, Accumulated Leap Seconds=18
  { 2443887.50, 49.625611 }, //14 January 1979, UT1-UTC= 0.5583886, Accumulated Leap Seconds=18
  { 2443888.50, 49.628364 }, //15 January 1979, UT1-UTC= 0.5556363, Accumulated Leap Seconds=18
  { 2443889.50, 49.631188 }, //16 January 1979, UT1-UTC= 0.5528117, Accumulated Leap Seconds=18
  { 2443890.50, 49.634092 }, //17 January 1979, UT1-UTC= 0.5499081, Accumulated Leap Seconds=18
  { 2443891.50, 49.637070 }, //18 January 1979, UT1-UTC= 0.5469302, Accumulated Leap Seconds=18
  { 2443892.50, 49.640107 }, //19 January 1979, UT1-UTC= 0.5438931, Accumulated Leap Seconds=18
  { 2443893.50, 49.643181 }, //20 January 1979, UT1-UTC= 0.5408193, Accumulated Leap Seconds=18
  { 2443894.50, 49.646265 }, //21 January 1979, UT1-UTC= 0.5377346, Accumulated Leap Seconds=18
  { 2443895.50, 49.649335 }, //22 January 1979, UT1-UTC= 0.5346649, Accumulated Leap Seconds=18
  { 2443896.50, 49.652367 }, //23 January 1979, UT1-UTC= 0.5316326, Accumulated Leap Seconds=18
  { 2443897.50, 49.655350 }, //24 January 1979, UT1-UTC= 0.5286500, Accumulated Leap Seconds=18
  { 2443898.50, 49.658289 }, //25 January 1979, UT1-UTC= 0.5257105, Accumulated Leap Seconds=18
  { 2443899.50, 49.661217 }, //26 January 1979, UT1-UTC= 0.5227828, Accumulated Leap Seconds=18
  { 2443900.50, 49.664185 }, //27 January 1979, UT1-UTC= 0.5198154, Accumulated Leap Seconds=18
  { 2443901.50, 49.667246 }, //28 January 1979, UT1-UTC= 0.5167535, Accumulated Leap Seconds=18
  { 2443902.50, 49.670436 }, //29 January 1979, UT1-UTC= 0.5135643, Accumulated Leap Seconds=18
  { 2443903.50, 49.673744 }, //30 January 1979, UT1-UTC= 0.5102556, Accumulated Leap Seconds=18
  { 2443904.50, 49.677122 }, //31 January 1979, UT1-UTC= 0.5068781, Accumulated Leap Seconds=18
  { 2443905.50, 49.680491 }, //01 February 1979, UT1-UTC= 0.5035093, Accumulated Leap Seconds=18
  { 2443906.50, 49.683772 }, //02 February 1979, UT1-UTC= 0.5002283, Accumulated Leap Seconds=18
  { 2443907.50, 49.686907 }, //03 February 1979, UT1-UTC= 0.4970926, Accumulated Leap Seconds=18
  { 2443908.50, 49.689872 }, //04 February 1979, UT1-UTC= 0.4941278, Accumulated Leap Seconds=18
  { 2443909.50, 49.692670 }, //05 February 1979, UT1-UTC= 0.4913303, Accumulated Leap Seconds=18
  { 2443910.50, 49.695324 }, //06 February 1979, UT1-UTC= 0.4886756, Accumulated Leap Seconds=18
  { 2443911.50, 49.697873 }, //07 February 1979, UT1-UTC= 0.4861267, Accumulated Leap Seconds=18
  { 2443912.50, 49.700359 }, //08 February 1979, UT1-UTC= 0.4836415, Accumulated Leap Seconds=18
  { 2443913.50, 49.702822 }, //09 February 1979, UT1-UTC= 0.4811778, Accumulated Leap Seconds=18
  { 2443914.50, 49.705302 }, //10 February 1979, UT1-UTC= 0.4786975, Accumulated Leap Seconds=18
  { 2443915.50, 49.707830 }, //11 February 1979, UT1-UTC= 0.4761702, Accumulated Leap Seconds=18
  { 2443916.50, 49.710425 }, //12 February 1979, UT1-UTC= 0.4735748, Accumulated Leap Seconds=18
  { 2443917.50, 49.713099 }, //13 February 1979, UT1-UTC= 0.4709008, Accumulated Leap Seconds=18
  { 2443918.50, 49.715851 }, //14 February 1979, UT1-UTC= 0.4681491, Accumulated Leap Seconds=18
  { 2443919.50, 49.718668 }, //15 February 1979, UT1-UTC= 0.4653319, Accumulated Leap Seconds=18
  { 2443920.50, 49.721529 }, //16 February 1979, UT1-UTC= 0.4624714, Accumulated Leap Seconds=18
  { 2443921.50, 49.724405 }, //17 February 1979, UT1-UTC= 0.4595950, Accumulated Leap Seconds=18
  { 2443922.50, 49.727270 }, //18 February 1979, UT1-UTC= 0.4567298, Accumulated Leap Seconds=18
  { 2443923.50, 49.730103 }, //19 February 1979, UT1-UTC= 0.4538972, Accumulated Leap Seconds=18
  { 2443924.50, 49.732893 }, //20 February 1979, UT1-UTC= 0.4511072, Accumulated Leap Seconds=18
  { 2443925.50, 49.735645 }, //21 February 1979, UT1-UTC= 0.4483545, Accumulated Leap Seconds=18
  { 2443926.50, 49.738386 }, //22 February 1979, UT1-UTC= 0.4456141, Accumulated Leap Seconds=18
  { 2443927.50, 49.741160 }, //23 February 1979, UT1-UTC= 0.4428400, Accumulated Leap Seconds=18
  { 2443928.50, 49.744025 }, //24 February 1979, UT1-UTC= 0.4399747, Accumulated Leap Seconds=18
  { 2443929.50, 49.747032 }, //25 February 1979, UT1-UTC= 0.4369679, Accumulated Leap Seconds=18
  { 2443930.50, 49.750200 }, //26 February 1979, UT1-UTC= 0.4337999, Accumulated Leap Seconds=18
  { 2443931.50, 49.753504 }, //27 February 1979, UT1-UTC= 0.4304964, Accumulated Leap Seconds=18
  { 2443932.50, 49.756875 }, //28 February 1979, UT1-UTC= 0.4271250, Accumulated Leap Seconds=18
  { 2443933.50, 49.760226 }, //01 March 1979, UT1-UTC= 0.4237736, Accumulated Leap Seconds=18
  { 2443934.50, 49.763477 }, //02 March 1979, UT1-UTC= 0.4205228, Accumulated Leap Seconds=18
  { 2443935.50, 49.766576 }, //03 March 1979, UT1-UTC= 0.4174241, Accumulated Leap Seconds=18
  { 2443936.50, 49.769508 }, //04 March 1979, UT1-UTC= 0.4144918, Accumulated Leap Seconds=18
  { 2443937.50, 49.772291 }, //05 March 1979, UT1-UTC= 0.4117092, Accumulated Leap Seconds=18
  { 2443938.50, 49.774960 }, //06 March 1979, UT1-UTC= 0.4090396, Accumulated Leap Seconds=18
  { 2443939.50, 49.777561 }, //07 March 1979, UT1-UTC= 0.4064385, Accumulated Leap Seconds=18
  { 2443940.50, 49.780140 }, //08 March 1979, UT1-UTC= 0.4038600, Accumulated Leap Seconds=18
  { 2443941.50, 49.782739 }, //09 March 1979, UT1-UTC= 0.4012614, Accumulated Leap Seconds=18
  { 2443942.50, 49.785394 }, //10 March 1979, UT1-UTC= 0.3986063, Accumulated Leap Seconds=18
  { 2443943.50, 49.788132 }, //11 March 1979, UT1-UTC= 0.3958675, Accumulated Leap Seconds=18
  { 2443944.50, 49.790971 }, //12 March 1979, UT1-UTC= 0.3930293, Accumulated Leap Seconds=18
  { 2443945.50, 49.793912 }, //13 March 1979, UT1-UTC= 0.3900882, Accumulated Leap Seconds=18
  { 2443946.50, 49.796948 }, //14 March 1979, UT1-UTC= 0.3870522, Accumulated Leap Seconds=18
  { 2443947.50, 49.800059 }, //15 March 1979, UT1-UTC= 0.3839407, Accumulated Leap Seconds=18
  { 2443948.50, 49.803217 }, //16 March 1979, UT1-UTC= 0.3807828, Accumulated Leap Seconds=18
  { 2443949.50, 49.806386 }, //17 March 1979, UT1-UTC= 0.3776143, Accumulated Leap Seconds=18
  { 2443950.50, 49.809532 }, //18 March 1979, UT1-UTC= 0.3744683, Accumulated Leap Seconds=18
  { 2443951.50, 49.812634 }, //19 March 1979, UT1-UTC= 0.3713660, Accumulated Leap Seconds=18
  { 2443952.50, 49.815690 }, //20 March 1979, UT1-UTC= 0.3683102, Accumulated Leap Seconds=18
  { 2443953.50, 49.818717 }, //21 March 1979, UT1-UTC= 0.3652833, Accumulated Leap Seconds=18
  { 2443954.50, 49.821752 }, //22 March 1979, UT1-UTC= 0.3622478, Accumulated Leap Seconds=18
  { 2443955.50, 49.824848 }, //23 March 1979, UT1-UTC= 0.3591522, Accumulated Leap Seconds=18
  { 2443956.50, 49.828057 }, //24 March 1979, UT1-UTC= 0.3559428, Accumulated Leap Seconds=18
  { 2443957.50, 49.831416 }, //25 March 1979, UT1-UTC= 0.3525835, Accumulated Leap Seconds=18
  { 2443958.50, 49.834927 }, //26 March 1979, UT1-UTC= 0.3490732, Accumulated Leap Seconds=18
  { 2443959.50, 49.838545 }, //27 March 1979, UT1-UTC= 0.3454547, Accumulated Leap Seconds=18
  { 2443960.50, 49.842193 }, //28 March 1979, UT1-UTC= 0.3418067, Accumulated Leap Seconds=18
  { 2443961.50, 49.845778 }, //29 March 1979, UT1-UTC= 0.3382216, Accumulated Leap Seconds=18
  { 2443962.50, 49.849226 }, //30 March 1979, UT1-UTC= 0.3347743, Accumulated Leap Seconds=18
  { 2443963.50, 49.852494 }, //31 March 1979, UT1-UTC= 0.3315058, Accumulated Leap Seconds=18
  { 2443964.50, 49.855580 }, //01 April 1979, UT1-UTC= 0.3284195, Accumulated Leap Seconds=18
  { 2443965.50, 49.858511 }, //02 April 1979, UT1-UTC= 0.3254888, Accumulated Leap Seconds=18
  { 2443966.50, 49.861329 }, //03 April 1979, UT1-UTC= 0.3226711, Accumulated Leap Seconds=18
  { 2443967.50, 49.864082 }, //04 April 1979, UT1-UTC= 0.3199182, Accumulated Leap Seconds=18
  { 2443968.50, 49.866817 }, //05 April 1979, UT1-UTC= 0.3171834, Accumulated Leap Seconds=18
  { 2443969.50, 49.869575 }, //06 April 1979, UT1-UTC= 0.3144246, Accumulated Leap Seconds=18
  { 2443970.50, 49.872392 }, //07 April 1979, UT1-UTC= 0.3116080, Accumulated Leap Seconds=18
  { 2443971.50, 49.875288 }, //08 April 1979, UT1-UTC= 0.3087115, Accumulated Leap Seconds=18
  { 2443972.50, 49.878274 }, //09 April 1979, UT1-UTC= 0.3057255, Accumulated Leap Seconds=18
  { 2443973.50, 49.881347 }, //10 April 1979, UT1-UTC= 0.3026528, Accumulated Leap Seconds=18
  { 2443974.50, 49.884493 }, //11 April 1979, UT1-UTC= 0.2995074, Accumulated Leap Seconds=18
  { 2443975.50, 49.887686 }, //12 April 1979, UT1-UTC= 0.2963140, Accumulated Leap Seconds=18
  { 2443976.50, 49.890894 }, //13 April 1979, UT1-UTC= 0.2931056, Accumulated Leap Seconds=18
  { 2443977.50, 49.894084 }, //14 April 1979, UT1-UTC= 0.2899163, Accumulated Leap Seconds=18
  { 2443978.50, 49.897228 }, //15 April 1979, UT1-UTC= 0.2867723, Accumulated Leap Seconds=18
  { 2443979.50, 49.900317 }, //16 April 1979, UT1-UTC= 0.2836831, Accumulated Leap Seconds=18
  { 2443980.50, 49.903364 }, //17 April 1979, UT1-UTC= 0.2806363, Accumulated Leap Seconds=18
  { 2443981.50, 49.906402 }, //18 April 1979, UT1-UTC= 0.2775983, Accumulated Leap Seconds=18
  { 2443982.50, 49.909478 }, //19 April 1979, UT1-UTC= 0.2745222, Accumulated Leap Seconds=18
  { 2443983.50, 49.912642 }, //20 April 1979, UT1-UTC= 0.2713580, Accumulated Leap Seconds=18
  { 2443984.50, 49.915933 }, //21 April 1979, UT1-UTC= 0.2680666, Accumulated Leap Seconds=18
  { 2443985.50, 49.919367 }, //22 April 1979, UT1-UTC= 0.2646328, Accumulated Leap Seconds=18
  { 2443986.50, 49.922924 }, //23 April 1979, UT1-UTC= 0.2610759, Accumulated Leap Seconds=18
  { 2443987.50, 49.926550 }, //24 April 1979, UT1-UTC= 0.2574504, Accumulated Leap Seconds=18
  { 2443988.50, 49.930165 }, //25 April 1979, UT1-UTC= 0.2538348, Accumulated Leap Seconds=18
  { 2443989.50, 49.933689 }, //26 April 1979, UT1-UTC= 0.2503110, Accumulated Leap Seconds=18
  { 2443990.50, 49.937059 }, //27 April 1979, UT1-UTC= 0.2469414, Accumulated Leap Seconds=18
  { 2443991.50, 49.940247 }, //28 April 1979, UT1-UTC= 0.2437528, Accumulated Leap Seconds=18
  { 2443992.50, 49.943263 }, //29 April 1979, UT1-UTC= 0.2407366, Accumulated Leap Seconds=18
  { 2443993.50, 49.946140 }, //30 April 1979, UT1-UTC= 0.2378598, Accumulated Leap Seconds=18
  { 2443994.50, 49.948922 }, //01 May 1979, UT1-UTC= 0.2350776, Accumulated Leap Seconds=18
  { 2443995.50, 49.951656 }, //02 May 1979, UT1-UTC= 0.2323435, Accumulated Leap Seconds=18
  { 2443996.50, 49.954385 }, //03 May 1979, UT1-UTC= 0.2296149, Accumulated Leap Seconds=18
  { 2443997.50, 49.957143 }, //04 May 1979, UT1-UTC= 0.2268568, Accumulated Leap Seconds=18
  { 2443998.50, 49.959956 }, //05 May 1979, UT1-UTC= 0.2240443, Accumulated Leap Seconds=18
  { 2443999.50, 49.962834 }, //06 May 1979, UT1-UTC= 0.2211655, Accumulated Leap Seconds=18
  { 2444000.50, 49.965779 }, //07 May 1979, UT1-UTC= 0.2182209, Accumulated Leap Seconds=18
  { 2444001.50, 49.968778 }, //08 May 1979, UT1-UTC= 0.2152219, Accumulated Leap Seconds=18
  { 2444002.50, 49.971810 }, //09 May 1979, UT1-UTC= 0.2121896, Accumulated Leap Seconds=18
  { 2444003.50, 49.974847 }, //10 May 1979, UT1-UTC= 0.2091528, Accumulated Leap Seconds=18
  { 2444004.50, 49.977856 }, //11 May 1979, UT1-UTC= 0.2061444, Accumulated Leap Seconds=18
  { 2444005.50, 49.980804 }, //12 May 1979, UT1-UTC= 0.2031958, Accumulated Leap Seconds=18
  { 2444006.50, 49.983674 }, //13 May 1979, UT1-UTC= 0.2003263, Accumulated Leap Seconds=18
  { 2444007.50, 49.986467 }, //14 May 1979, UT1-UTC= 0.1975334, Accumulated Leap Seconds=18
  { 2444008.50, 49.989210 }, //15 May 1979, UT1-UTC= 0.1947898, Accumulated Leap Seconds=18
  { 2444009.50, 49.991950 }, //16 May 1979, UT1-UTC= 0.1920500, Accumulated Leap Seconds=18
  { 2444010.50, 49.994738 }, //17 May 1979, UT1-UTC= 0.1892624, Accumulated Leap Seconds=18
  { 2444011.50, 49.997615 }, //18 May 1979, UT1-UTC= 0.1863851, Accumulated Leap Seconds=18
  { 2444012.50, 50.000600 }, //19 May 1979, UT1-UTC= 0.1833995, Accumulated Leap Seconds=18
  { 2444013.50, 50.003684 }, //20 May 1979, UT1-UTC= 0.1803163, Accumulated Leap Seconds=18
  { 2444014.50, 50.006825 }, //21 May 1979, UT1-UTC= 0.1771746, Accumulated Leap Seconds=18
  { 2444015.50, 50.009964 }, //22 May 1979, UT1-UTC= 0.1740361, Accumulated Leap Seconds=18
  { 2444016.50, 50.013029 }, //23 May 1979, UT1-UTC= 0.1709710, Accumulated Leap Seconds=18
  { 2444017.50, 50.015959 }, //24 May 1979, UT1-UTC= 0.1680411, Accumulated Leap Seconds=18
  { 2444018.50, 50.018715 }, //25 May 1979, UT1-UTC= 0.1652848, Accumulated Leap Seconds=18
  { 2444019.50, 50.021290 }, //26 May 1979, UT1-UTC= 0.1627103, Accumulated Leap Seconds=18
  { 2444020.50, 50.023701 }, //27 May 1979, UT1-UTC= 0.1602988, Accumulated Leap Seconds=18
  { 2444021.50, 50.025986 }, //28 May 1979, UT1-UTC= 0.1580143, Accumulated Leap Seconds=18
  { 2444022.50, 50.028186 }, //29 May 1979, UT1-UTC= 0.1558142, Accumulated Leap Seconds=18
  { 2444023.50, 50.030344 }, //30 May 1979, UT1-UTC= 0.1536557, Accumulated Leap Seconds=18
  { 2444024.50, 50.032499 }, //31 May 1979, UT1-UTC= 0.1515013, Accumulated Leap Seconds=18
  { 2444025.50, 50.034678 }, //01 June 1979, UT1-UTC= 0.1493223, Accumulated Leap Seconds=18
  { 2444026.50, 50.036900 }, //02 June 1979, UT1-UTC= 0.1471003, Accumulated Leap Seconds=18
  { 2444027.50, 50.039172 }, //03 June 1979, UT1-UTC= 0.1448280, Accumulated Leap Seconds=18
  { 2444028.50, 50.041490 }, //04 June 1979, UT1-UTC= 0.1425098, Accumulated Leap Seconds=18
  { 2444029.50, 50.043841 }, //05 June 1979, UT1-UTC= 0.1401591, Accumulated Leap Seconds=18
  { 2444030.50, 50.046204 }, //06 June 1979, UT1-UTC= 0.1377958, Accumulated Leap Seconds=18
  { 2444031.50, 50.048556 }, //07 June 1979, UT1-UTC= 0.1354436, Accumulated Leap Seconds=18
  { 2444032.50, 50.050872 }, //08 June 1979, UT1-UTC= 0.1331280, Accumulated Leap Seconds=18
  { 2444033.50, 50.053130 }, //09 June 1979, UT1-UTC= 0.1308701, Accumulated Leap Seconds=18
  { 2444034.50, 50.055323 }, //10 June 1979, UT1-UTC= 0.1286773, Accumulated Leap Seconds=18
  { 2444035.50, 50.057467 }, //11 June 1979, UT1-UTC= 0.1265333, Accumulated Leap Seconds=18
  { 2444036.50, 50.059604 }, //12 June 1979, UT1-UTC= 0.1243962, Accumulated Leap Seconds=18
  { 2444037.50, 50.061789 }, //13 June 1979, UT1-UTC= 0.1222107, Accumulated Leap Seconds=18
  { 2444038.50, 50.064073 }, //14 June 1979, UT1-UTC= 0.1199266, Accumulated Leap Seconds=18
  { 2444039.50, 50.066482 }, //15 June 1979, UT1-UTC= 0.1175181, Accumulated Leap Seconds=18
  { 2444040.50, 50.069007 }, //16 June 1979, UT1-UTC= 0.1149927, Accumulated Leap Seconds=18
  { 2444041.50, 50.071610 }, //17 June 1979, UT1-UTC= 0.1123900, Accumulated Leap Seconds=18
  { 2444042.50, 50.074231 }, //18 June 1979, UT1-UTC= 0.1097687, Accumulated Leap Seconds=18
  { 2444043.50, 50.076808 }, //19 June 1979, UT1-UTC= 0.1071922, Accumulated Leap Seconds=18
  { 2444044.50, 50.079285 }, //20 June 1979, UT1-UTC= 0.1047145, Accumulated Leap Seconds=18
  { 2444045.50, 50.081629 }, //21 June 1979, UT1-UTC= 0.1023711, Accumulated Leap Seconds=18
  { 2444046.50, 50.083828 }, //22 June 1979, UT1-UTC= 0.1001723, Accumulated Leap Seconds=18
  { 2444047.50, 50.085896 }, //23 June 1979, UT1-UTC= 0.0981042, Accumulated Leap Seconds=18
  { 2444048.50, 50.087865 }, //24 June 1979, UT1-UTC= 0.0961347, Accumulated Leap Seconds=18
  { 2444049.50, 50.089777 }, //25 June 1979, UT1-UTC= 0.0942227, Accumulated Leap Seconds=18
  { 2444050.50, 50.091673 }, //26 June 1979, UT1-UTC= 0.0923267, Accumulated Leap Seconds=18
  { 2444051.50, 50.093590 }, //27 June 1979, UT1-UTC= 0.0904105, Accumulated Leap Seconds=18
  { 2444052.50, 50.095554 }, //28 June 1979, UT1-UTC= 0.0884461, Accumulated Leap Seconds=18
  { 2444053.50, 50.097584 }, //29 June 1979, UT1-UTC= 0.0864157, Accumulated Leap Seconds=18
  { 2444054.50, 50.099686 }, //30 June 1979, UT1-UTC= 0.0843137, Accumulated Leap Seconds=18
  { 2444055.50, 50.101853 }, //01 July 1979, UT1-UTC= 0.0821469, Accumulated Leap Seconds=18
  { 2444056.50, 50.104067 }, //02 July 1979, UT1-UTC= 0.0799326, Accumulated Leap Seconds=18
  { 2444057.50, 50.106306 }, //03 July 1979, UT1-UTC= 0.0776942, Accumulated Leap Seconds=18
  { 2444058.50, 50.108542 }, //04 July 1979, UT1-UTC= 0.0754580, Accumulated Leap Seconds=18
  { 2444059.50, 50.110750 }, //05 July 1979, UT1-UTC= 0.0732501, Accumulated Leap Seconds=18
  { 2444060.50, 50.112906 }, //06 July 1979, UT1-UTC= 0.0710943, Accumulated Leap Seconds=18
  { 2444061.50, 50.114995 }, //07 July 1979, UT1-UTC= 0.0690048, Accumulated Leap Seconds=18
  { 2444062.50, 50.117022 }, //08 July 1979, UT1-UTC= 0.0669781, Accumulated Leap Seconds=18
  { 2444063.50, 50.119014 }, //09 July 1979, UT1-UTC= 0.0649857, Accumulated Leap Seconds=18
  { 2444064.50, 50.121022 }, //10 July 1979, UT1-UTC= 0.0629778, Accumulated Leap Seconds=18
  { 2444065.50, 50.123101 }, //11 July 1979, UT1-UTC= 0.0608988, Accumulated Leap Seconds=18
  { 2444066.50, 50.125288 }, //12 July 1979, UT1-UTC= 0.0587116, Accumulated Leap Seconds=18
  { 2444067.50, 50.127585 }, //13 July 1979, UT1-UTC= 0.0564154, Accumulated Leap Seconds=18
  { 2444068.50, 50.129950 }, //14 July 1979, UT1-UTC= 0.0540501, Accumulated Leap Seconds=18
  { 2444069.50, 50.132318 }, //15 July 1979, UT1-UTC= 0.0516816, Accumulated Leap Seconds=18
  { 2444070.50, 50.134617 }, //16 July 1979, UT1-UTC= 0.0493827, Accumulated Leap Seconds=18
  { 2444071.50, 50.136787 }, //17 July 1979, UT1-UTC= 0.0472132, Accumulated Leap Seconds=18
  { 2444072.50, 50.138792 }, //18 July 1979, UT1-UTC= 0.0452079, Accumulated Leap Seconds=18
  { 2444073.50, 50.140626 }, //19 July 1979, UT1-UTC= 0.0433739, Accumulated Leap Seconds=18
  { 2444074.50, 50.142306 }, //20 July 1979, UT1-UTC= 0.0416942, Accumulated Leap Seconds=18
  { 2444075.50, 50.143866 }, //21 July 1979, UT1-UTC= 0.0401338, Accumulated Leap Seconds=18
  { 2444076.50, 50.145352 }, //22 July 1979, UT1-UTC= 0.0386482, Accumulated Leap Seconds=18
  { 2444077.50, 50.146808 }, //23 July 1979, UT1-UTC= 0.0371918, Accumulated Leap Seconds=18
  { 2444078.50, 50.148276 }, //24 July 1979, UT1-UTC= 0.0357240, Accumulated Leap Seconds=18
  { 2444079.50, 50.149787 }, //25 July 1979, UT1-UTC= 0.0342126, Accumulated Leap Seconds=18
  { 2444080.50, 50.151365 }, //26 July 1979, UT1-UTC= 0.0326352, Accumulated Leap Seconds=18
  { 2444081.50, 50.153020 }, //27 July 1979, UT1-UTC= 0.0309802, Accumulated Leap Seconds=18
  { 2444082.50, 50.154753 }, //28 July 1979, UT1-UTC= 0.0292468, Accumulated Leap Seconds=18
  { 2444083.50, 50.156555 }, //29 July 1979, UT1-UTC= 0.0274452, Accumulated Leap Seconds=18
  { 2444084.50, 50.158406 }, //30 July 1979, UT1-UTC= 0.0255938, Accumulated Leap Seconds=18
  { 2444085.50, 50.160286 }, //31 July 1979, UT1-UTC= 0.0237142, Accumulated Leap Seconds=18
  { 2444086.50, 50.162172 }, //01 August 1979, UT1-UTC= 0.0218277, Accumulated Leap Seconds=18
  { 2444087.50, 50.164049 }, //02 August 1979, UT1-UTC= 0.0199512, Accumulated Leap Seconds=18
  { 2444088.50, 50.165906 }, //03 August 1979, UT1-UTC= 0.0180939, Accumulated Leap Seconds=18
  { 2444089.50, 50.167747 }, //04 August 1979, UT1-UTC= 0.0162528, Accumulated Leap Seconds=18
  { 2444090.50, 50.169593 }, //05 August 1979, UT1-UTC= 0.0144074, Accumulated Leap Seconds=18
  { 2444091.50, 50.171486 }, //06 August 1979, UT1-UTC= 0.0125144, Accumulated Leap Seconds=18
  { 2444092.50, 50.173484 }, //07 August 1979, UT1-UTC= 0.0105159, Accumulated Leap Seconds=18
  { 2444093.50, 50.175641 }, //08 August 1979, UT1-UTC= 0.0083593, Accumulated Leap Seconds=18
  { 2444094.50, 50.177980 }, //09 August 1979, UT1-UTC= 0.0060201, Accumulated Leap Seconds=18
  { 2444095.50, 50.180481 }, //10 August 1979, UT1-UTC= 0.0035194, Accumulated Leap Seconds=18
  { 2444096.50, 50.183076 }, //11 August 1979, UT1-UTC= 0.0009244, Accumulated Leap Seconds=18
  { 2444097.50, 50.185672 }, //12 August 1979, UT1-UTC=-0.0016716, Accumulated Leap Seconds=18
  { 2444098.50, 50.188180 }, //13 August 1979, UT1-UTC=-0.0041799, Accumulated Leap Seconds=18
  { 2444099.50, 50.190539 }, //14 August 1979, UT1-UTC=-0.0065394, Accumulated Leap Seconds=18
  { 2444100.50, 50.192726 }, //15 August 1979, UT1-UTC=-0.0087259, Accumulated Leap Seconds=18
  { 2444101.50, 50.194748 }, //16 August 1979, UT1-UTC=-0.0107483, Accumulated Leap Seconds=18
  { 2444102.50, 50.196640 }, //17 August 1979, UT1-UTC=-0.0126402, Accumulated Leap Seconds=18
  { 2444103.50, 50.198447 }, //18 August 1979, UT1-UTC=-0.0144473, Accumulated Leap Seconds=18
  { 2444104.50, 50.200216 }, //19 August 1979, UT1-UTC=-0.0162163, Accumulated Leap Seconds=18
  { 2444105.50, 50.201989 }, //20 August 1979, UT1-UTC=-0.0179894, Accumulated Leap Seconds=18
  { 2444106.50, 50.203799 }, //21 August 1979, UT1-UTC=-0.0197992, Accumulated Leap Seconds=18
  { 2444107.50, 50.205668 }, //22 August 1979, UT1-UTC=-0.0216679, Accumulated Leap Seconds=18
  { 2444108.50, 50.207606 }, //23 August 1979, UT1-UTC=-0.0236058, Accumulated Leap Seconds=18
  { 2444109.50, 50.209612 }, //24 August 1979, UT1-UTC=-0.0256121, Accumulated Leap Seconds=18
  { 2444110.50, 50.211675 }, //25 August 1979, UT1-UTC=-0.0276754, Accumulated Leap Seconds=18
  { 2444111.50, 50.213775 }, //26 August 1979, UT1-UTC=-0.0297751, Accumulated Leap Seconds=18
  { 2444112.50, 50.215885 }, //27 August 1979, UT1-UTC=-0.0318847, Accumulated Leap Seconds=18
  { 2444113.50, 50.217979 }, //28 August 1979, UT1-UTC=-0.0339794, Accumulated Leap Seconds=18
  { 2444114.50, 50.220041 }, //29 August 1979, UT1-UTC=-0.0360406, Accumulated Leap Seconds=18
  { 2444115.50, 50.222059 }, //30 August 1979, UT1-UTC=-0.0380593, Accumulated Leap Seconds=18
  { 2444116.50, 50.224039 }, //31 August 1979, UT1-UTC=-0.0400395, Accumulated Leap Seconds=18
  { 2444117.50, 50.226001 }, //01 September 1979, UT1-UTC=-0.0420014, Accumulated Leap Seconds=18
  { 2444118.50, 50.227984 }, //02 September 1979, UT1-UTC=-0.0439842, Accumulated Leap Seconds=18
  { 2444119.50, 50.230045 }, //03 September 1979, UT1-UTC=-0.0460449, Accumulated Leap Seconds=18
  { 2444120.50, 50.232246 }, //04 September 1979, UT1-UTC=-0.0482458, Accumulated Leap Seconds=18
  { 2444121.50, 50.234635 }, //05 September 1979, UT1-UTC=-0.0506350, Accumulated Leap Seconds=18
  { 2444122.50, 50.237222 }, //06 September 1979, UT1-UTC=-0.0532225, Accumulated Leap Seconds=18
  { 2444123.50, 50.239967 }, //07 September 1979, UT1-UTC=-0.0559672, Accumulated Leap Seconds=18
  { 2444124.50, 50.242785 }, //08 September 1979, UT1-UTC=-0.0587853, Accumulated Leap Seconds=18
  { 2444125.50, 50.245575 }, //09 September 1979, UT1-UTC=-0.0615751, Accumulated Leap Seconds=18
  { 2444126.50, 50.248249 }, //10 September 1979, UT1-UTC=-0.0642492, Accumulated Leap Seconds=18
  { 2444127.50, 50.250758 }, //11 September 1979, UT1-UTC=-0.0667581, Accumulated Leap Seconds=18
  { 2444128.50, 50.253096 }, //12 September 1979, UT1-UTC=-0.0690955, Accumulated Leap Seconds=18
  { 2444129.50, 50.255289 }, //13 September 1979, UT1-UTC=-0.0712894, Accumulated Leap Seconds=18
  { 2444130.50, 50.257387 }, //14 September 1979, UT1-UTC=-0.0733870, Accumulated Leap Seconds=18
  { 2444131.50, 50.259440 }, //15 September 1979, UT1-UTC=-0.0754396, Accumulated Leap Seconds=18
  { 2444132.50, 50.261493 }, //16 September 1979, UT1-UTC=-0.0774934, Accumulated Leap Seconds=18
  { 2444133.50, 50.263585 }, //17 September 1979, UT1-UTC=-0.0795852, Accumulated Leap Seconds=18
  { 2444134.50, 50.265740 }, //18 September 1979, UT1-UTC=-0.0817400, Accumulated Leap Seconds=18
  { 2444135.50, 50.267971 }, //19 September 1979, UT1-UTC=-0.0839707, Accumulated Leap Seconds=18
  { 2444136.50, 50.270280 }, //20 September 1979, UT1-UTC=-0.0862795, Accumulated Leap Seconds=18
  { 2444137.50, 50.272658 }, //21 September 1979, UT1-UTC=-0.0886578, Accumulated Leap Seconds=18
  { 2444138.50, 50.275088 }, //22 September 1979, UT1-UTC=-0.0910883, Accumulated Leap Seconds=18
  { 2444139.50, 50.277547 }, //23 September 1979, UT1-UTC=-0.0935467, Accumulated Leap Seconds=18
  { 2444140.50, 50.280005 }, //24 September 1979, UT1-UTC=-0.0960055, Accumulated Leap Seconds=18
  { 2444141.50, 50.282439 }, //25 September 1979, UT1-UTC=-0.0984395, Accumulated Leap Seconds=18
  { 2444142.50, 50.284834 }, //26 September 1979, UT1-UTC=-0.1008341, Accumulated Leap Seconds=18
  { 2444143.50, 50.287189 }, //27 September 1979, UT1-UTC=-0.1031889, Accumulated Leap Seconds=18
  { 2444144.50, 50.289521 }, //28 September 1979, UT1-UTC=-0.1055210, Accumulated Leap Seconds=18
  { 2444145.50, 50.291864 }, //29 September 1979, UT1-UTC=-0.1078645, Accumulated Leap Seconds=18
  { 2444146.50, 50.294268 }, //30 September 1979, UT1-UTC=-0.1102684, Accumulated Leap Seconds=18
  { 2444147.50, 50.296791 }, //01 October 1979, UT1-UTC=-0.1127905, Accumulated Leap Seconds=18
  { 2444148.50, 50.299485 }, //02 October 1979, UT1-UTC=-0.1154848, Accumulated Leap Seconds=18
  { 2444149.50, 50.302382 }, //03 October 1979, UT1-UTC=-0.1183824, Accumulated Leap Seconds=18
  { 2444150.50, 50.305474 }, //04 October 1979, UT1-UTC=-0.1214737, Accumulated Leap Seconds=18
  { 2444151.50, 50.308702 }, //05 October 1979, UT1-UTC=-0.1247019, Accumulated Leap Seconds=18
  { 2444152.50, 50.311975 }, //06 October 1979, UT1-UTC=-0.1279750, Accumulated Leap Seconds=18
  { 2444153.50, 50.315193 }, //07 October 1979, UT1-UTC=-0.1311935, Accumulated Leap Seconds=18
  { 2444154.50, 50.318280 }, //08 October 1979, UT1-UTC=-0.1342802, Accumulated Leap Seconds=18
  { 2444155.50, 50.321200 }, //09 October 1979, UT1-UTC=-0.1372001, Accumulated Leap Seconds=18
  { 2444156.50, 50.323961 }, //10 October 1979, UT1-UTC=-0.1399606, Accumulated Leap Seconds=18
  { 2444157.50, 50.326599 }, //11 October 1979, UT1-UTC=-0.1425989, Accumulated Leap Seconds=18
  { 2444158.50, 50.329166 }, //12 October 1979, UT1-UTC=-0.1451656, Accumulated Leap Seconds=18
  { 2444159.50, 50.331711 }, //13 October 1979, UT1-UTC=-0.1477108, Accumulated Leap Seconds=18
  { 2444160.50, 50.334277 }, //14 October 1979, UT1-UTC=-0.1502768, Accumulated Leap Seconds=18
  { 2444161.50, 50.336894 }, //15 October 1979, UT1-UTC=-0.1528938, Accumulated Leap Seconds=18
  { 2444162.50, 50.339579 }, //16 October 1979, UT1-UTC=-0.1555792, Accumulated Leap Seconds=18
  { 2444163.50, 50.342336 }, //17 October 1979, UT1-UTC=-0.1583361, Accumulated Leap Seconds=18
  { 2444164.50, 50.345155 }, //18 October 1979, UT1-UTC=-0.1611553, Accumulated Leap Seconds=18
  { 2444165.50, 50.348016 }, //19 October 1979, UT1-UTC=-0.1640163, Accumulated Leap Seconds=18
  { 2444166.50, 50.350891 }, //20 October 1979, UT1-UTC=-0.1668910, Accumulated Leap Seconds=18
  { 2444167.50, 50.353748 }, //21 October 1979, UT1-UTC=-0.1697481, Accumulated Leap Seconds=18
  { 2444168.50, 50.356555 }, //22 October 1979, UT1-UTC=-0.1725555, Accumulated Leap Seconds=18
  { 2444169.50, 50.359286 }, //23 October 1979, UT1-UTC=-0.1752861, Accumulated Leap Seconds=18
  { 2444170.50, 50.361925 }, //24 October 1979, UT1-UTC=-0.1779250, Accumulated Leap Seconds=18
  { 2444171.50, 50.364476 }, //25 October 1979, UT1-UTC=-0.1804765, Accumulated Leap Seconds=18
  { 2444172.50, 50.366966 }, //26 October 1979, UT1-UTC=-0.1829659, Accumulated Leap Seconds=18
  { 2444173.50, 50.369436 }, //27 October 1979, UT1-UTC=-0.1854364, Accumulated Leap Seconds=18
  { 2444174.50, 50.371943 }, //28 October 1979, UT1-UTC=-0.1879431, Accumulated Leap Seconds=18
  { 2444175.50, 50.374540 }, //29 October 1979, UT1-UTC=-0.1905405, Accumulated Leap Seconds=18
  { 2444176.50, 50.377268 }, //30 October 1979, UT1-UTC=-0.1932684, Accumulated Leap Seconds=18
  { 2444177.50, 50.380139 }, //31 October 1979, UT1-UTC=-0.1961386, Accumulated Leap Seconds=18
  { 2444178.50, 50.383125 }, //01 November 1979, UT1-UTC=-0.1991254, Accumulated Leap Seconds=18
  { 2444179.50, 50.386166 }, //02 November 1979, UT1-UTC=-0.2021660, Accumulated Leap Seconds=18
  { 2444180.50, 50.389175 }, //03 November 1979, UT1-UTC=-0.2051749, Accumulated Leap Seconds=18
  { 2444181.50, 50.392069 }, //04 November 1979, UT1-UTC=-0.2080693, Accumulated Leap Seconds=18
  { 2444182.50, 50.394793 }, //05 November 1979, UT1-UTC=-0.2107935, Accumulated Leap Seconds=18
  { 2444183.50, 50.397333 }, //06 November 1979, UT1-UTC=-0.2133327, Accumulated Leap Seconds=18
  { 2444184.50, 50.399710 }, //07 November 1979, UT1-UTC=-0.2157097, Accumulated Leap Seconds=18
  { 2444185.50, 50.401972 }, //08 November 1979, UT1-UTC=-0.2179716, Accumulated Leap Seconds=18
  { 2444186.50, 50.404173 }, //09 November 1979, UT1-UTC=-0.2201728, Accumulated Leap Seconds=18
  { 2444187.50, 50.406363 }, //10 November 1979, UT1-UTC=-0.2223635, Accumulated Leap Seconds=18
  { 2444188.50, 50.408584 }, //11 November 1979, UT1-UTC=-0.2245837, Accumulated Leap Seconds=18
  { 2444189.50, 50.410861 }, //12 November 1979, UT1-UTC=-0.2268610, Accumulated Leap Seconds=18
  { 2444190.50, 50.413210 }, //13 November 1979, UT1-UTC=-0.2292099, Accumulated Leap Seconds=18
  { 2444191.50, 50.415631 }, //14 November 1979, UT1-UTC=-0.2316314, Accumulated Leap Seconds=18
  { 2444192.50, 50.418115 }, //15 November 1979, UT1-UTC=-0.2341150, Accumulated Leap Seconds=18
  { 2444193.50, 50.420642 }, //16 November 1979, UT1-UTC=-0.2366418, Accumulated Leap Seconds=18
  { 2444194.50, 50.423186 }, //17 November 1979, UT1-UTC=-0.2391864, Accumulated Leap Seconds=18
  { 2444195.50, 50.425721 }, //18 November 1979, UT1-UTC=-0.2417206, Accumulated Leap Seconds=18
  { 2444196.50, 50.428217 }, //19 November 1979, UT1-UTC=-0.2442172, Accumulated Leap Seconds=18
  { 2444197.50, 50.430656 }, //20 November 1979, UT1-UTC=-0.2466563, Accumulated Leap Seconds=18
  { 2444198.50, 50.433033 }, //21 November 1979, UT1-UTC=-0.2490333, Accumulated Leap Seconds=18
  { 2444199.50, 50.435366 }, //22 November 1979, UT1-UTC=-0.2513657, Accumulated Leap Seconds=18
  { 2444200.50, 50.437693 }, //23 November 1979, UT1-UTC=-0.2536929, Accumulated Leap Seconds=18
  { 2444201.50, 50.440068 }, //24 November 1979, UT1-UTC=-0.2560683, Accumulated Leap Seconds=18
  { 2444202.50, 50.442546 }, //25 November 1979, UT1-UTC=-0.2585460, Accumulated Leap Seconds=18
  { 2444203.50, 50.445166 }, //26 November 1979, UT1-UTC=-0.2611662, Accumulated Leap Seconds=18
  { 2444204.50, 50.447944 }, //27 November 1979, UT1-UTC=-0.2639437, Accumulated Leap Seconds=18
  { 2444205.50, 50.450861 }, //28 November 1979, UT1-UTC=-0.2668609, Accumulated Leap Seconds=18
  { 2444206.50, 50.453870 }, //29 November 1979, UT1-UTC=-0.2698698, Accumulated Leap Seconds=18
  { 2444207.50, 50.456901 }, //30 November 1979, UT1-UTC=-0.2729006, Accumulated Leap Seconds=18
  { 2444208.50, 50.459877 }, //01 December 1979, UT1-UTC=-0.2758772, Accumulated Leap Seconds=18
  { 2444209.50, 50.462737 }, //02 December 1979, UT1-UTC=-0.2787369, Accumulated Leap Seconds=18
  { 2444210.50, 50.465445 }, //03 December 1979, UT1-UTC=-0.2814455, Accumulated Leap Seconds=18
  { 2444211.50, 50.468004 }, //04 December 1979, UT1-UTC=-0.2840045, Accumulated Leap Seconds=18
  { 2444212.50, 50.470444 }, //05 December 1979, UT1-UTC=-0.2864441, Accumulated Leap Seconds=18
  { 2444213.50, 50.472811 }, //06 December 1979, UT1-UTC=-0.2888107, Accumulated Leap Seconds=18
  { 2444214.50, 50.475152 }, //07 December 1979, UT1-UTC=-0.2911522, Accumulated Leap Seconds=18
  { 2444215.50, 50.477509 }, //08 December 1979, UT1-UTC=-0.2935087, Accumulated Leap Seconds=18
  { 2444216.50, 50.479910 }, //09 December 1979, UT1-UTC=-0.2959096, Accumulated Leap Seconds=18
  { 2444217.50, 50.482372 }, //10 December 1979, UT1-UTC=-0.2983721, Accumulated Leap Seconds=18
  { 2444218.50, 50.484901 }, //11 December 1979, UT1-UTC=-0.3009008, Accumulated Leap Seconds=18
  { 2444219.50, 50.487487 }, //12 December 1979, UT1-UTC=-0.3034874, Accumulated Leap Seconds=18
  { 2444220.50, 50.490114 }, //13 December 1979, UT1-UTC=-0.3061136, Accumulated Leap Seconds=18
  { 2444221.50, 50.492755 }, //14 December 1979, UT1-UTC=-0.3087547, Accumulated Leap Seconds=18
  { 2444222.50, 50.495383 }, //15 December 1979, UT1-UTC=-0.3113834, Accumulated Leap Seconds=18
  { 2444223.50, 50.497973 }, //16 December 1979, UT1-UTC=-0.3139733, Accumulated Leap Seconds=18
  { 2444224.50, 50.500502 }, //17 December 1979, UT1-UTC=-0.3165021, Accumulated Leap Seconds=18
  { 2444225.50, 50.502957 }, //18 December 1979, UT1-UTC=-0.3189573, Accumulated Leap Seconds=18
  { 2444226.50, 50.505343 }, //19 December 1979, UT1-UTC=-0.3213431, Accumulated Leap Seconds=18
  { 2444227.50, 50.507687 }, //20 December 1979, UT1-UTC=-0.3236869, Accumulated Leap Seconds=18
  { 2444228.50, 50.510036 }, //21 December 1979, UT1-UTC=-0.3260360, Accumulated Leap Seconds=18
  { 2444229.50, 50.512446 }, //22 December 1979, UT1-UTC=-0.3284461, Accumulated Leap Seconds=18
  { 2444230.50, 50.514963 }, //23 December 1979, UT1-UTC=-0.3309628, Accumulated Leap Seconds=18
  { 2444231.50, 50.517606 }, //24 December 1979, UT1-UTC=-0.3336061, Accumulated Leap Seconds=18
  { 2444232.50, 50.520362 }, //25 December 1979, UT1-UTC=-0.3363620, Accumulated Leap Seconds=18
  { 2444233.50, 50.523187 }, //26 December 1979, UT1-UTC=-0.3391870, Accumulated Leap Seconds=18
  { 2444234.50, 50.526020 }, //27 December 1979, UT1-UTC=-0.3420202, Accumulated Leap Seconds=18
  { 2444235.50, 50.528799 }, //28 December 1979, UT1-UTC=-0.3447989, Accumulated Leap Seconds=18
  { 2444236.50, 50.531472 }, //29 December 1979, UT1-UTC=-0.3474720, Accumulated Leap Seconds=18
  { 2444237.50, 50.534010 }, //30 December 1979, UT1-UTC=-0.3500104, Accumulated Leap Seconds=18
  { 2444238.50, 50.536413 }, //31 December 1979, UT1-UTC=-0.3524133, Accumulated Leap Seconds=18
  { 2444239.50, 50.538707 }, //01 January 1980, UT1-UTC= 0.6452932, Accumulated Leap Seconds=19
  { 2444240.50, 50.540934 }, //02 January 1980, UT1-UTC= 0.6430658, Accumulated Leap Seconds=19
  { 2444241.50, 50.543143 }, //03 January 1980, UT1-UTC= 0.6408565, Accumulated Leap Seconds=19
  { 2444242.50, 50.545377 }, //04 January 1980, UT1-UTC= 0.6386226, Accumulated Leap Seconds=19
  { 2444243.50, 50.547667 }, //05 January 1980, UT1-UTC= 0.6363331, Accumulated Leap Seconds=19
  { 2444244.50, 50.550031 }, //06 January 1980, UT1-UTC= 0.6339692, Accumulated Leap Seconds=19
  { 2444245.50, 50.552475 }, //07 January 1980, UT1-UTC= 0.6315248, Accumulated Leap Seconds=19
  { 2444246.50, 50.554994 }, //08 January 1980, UT1-UTC= 0.6290058, Accumulated Leap Seconds=19
  { 2444247.50, 50.557571 }, //09 January 1980, UT1-UTC= 0.6264286, Accumulated Leap Seconds=19
  { 2444248.50, 50.560183 }, //10 January 1980, UT1-UTC= 0.6238168, Accumulated Leap Seconds=19
  { 2444249.50, 50.562804 }, //11 January 1980, UT1-UTC= 0.6211960, Accumulated Leap Seconds=19
  { 2444250.50, 50.565411 }, //12 January 1980, UT1-UTC= 0.6185894, Accumulated Leap Seconds=19
  { 2444251.50, 50.567984 }, //13 January 1980, UT1-UTC= 0.6160163, Accumulated Leap Seconds=19
  { 2444252.50, 50.570511 }, //14 January 1980, UT1-UTC= 0.6134892, Accumulated Leap Seconds=19
  { 2444253.50, 50.572992 }, //15 January 1980, UT1-UTC= 0.6110084, Accumulated Leap Seconds=19
  { 2444254.50, 50.575444 }, //16 January 1980, UT1-UTC= 0.6085557, Accumulated Leap Seconds=19
  { 2444255.50, 50.577909 }, //17 January 1980, UT1-UTC= 0.6060913, Accumulated Leap Seconds=19
  { 2444256.50, 50.580440 }, //18 January 1980, UT1-UTC= 0.6035604, Accumulated Leap Seconds=19
  { 2444257.50, 50.583089 }, //19 January 1980, UT1-UTC= 0.6009113, Accumulated Leap Seconds=19
  { 2444258.50, 50.585883 }, //20 January 1980, UT1-UTC= 0.5981171, Accumulated Leap Seconds=19
  { 2444259.50, 50.588808 }, //21 January 1980, UT1-UTC= 0.5951917, Accumulated Leap Seconds=19
  { 2444260.50, 50.591810 }, //22 January 1980, UT1-UTC= 0.5921897, Accumulated Leap Seconds=19
  { 2444261.50, 50.594809 }, //23 January 1980, UT1-UTC= 0.5891907, Accumulated Leap Seconds=19
  { 2444262.50, 50.597724 }, //24 January 1980, UT1-UTC= 0.5862760, Accumulated Leap Seconds=19
  { 2444263.50, 50.600491 }, //25 January 1980, UT1-UTC= 0.5835088, Accumulated Leap Seconds=19
  { 2444264.50, 50.603077 }, //26 January 1980, UT1-UTC= 0.5809226, Accumulated Leap Seconds=19
  { 2444265.50, 50.605481 }, //27 January 1980, UT1-UTC= 0.5785193, Accumulated Leap Seconds=19
  { 2444266.50, 50.607726 }, //28 January 1980, UT1-UTC= 0.5762735, Accumulated Leap Seconds=19
  { 2444267.50, 50.609858 }, //29 January 1980, UT1-UTC= 0.5741419, Accumulated Leap Seconds=19
  { 2444268.50, 50.611926 }, //30 January 1980, UT1-UTC= 0.5720739, Accumulated Leap Seconds=19
  { 2444269.50, 50.613977 }, //31 January 1980, UT1-UTC= 0.5700229, Accumulated Leap Seconds=19
  { 2444270.50, 50.616048 }, //01 February 1980, UT1-UTC= 0.5679516, Accumulated Leap Seconds=19
  { 2444271.50, 50.618163 }, //02 February 1980, UT1-UTC= 0.5658367, Accumulated Leap Seconds=19
  { 2444272.50, 50.620333 }, //03 February 1980, UT1-UTC= 0.5636671, Accumulated Leap Seconds=19
  { 2444273.50, 50.622556 }, //04 February 1980, UT1-UTC= 0.5614440, Accumulated Leap Seconds=19
  { 2444274.50, 50.624822 }, //05 February 1980, UT1-UTC= 0.5591784, Accumulated Leap Seconds=19
  { 2444275.50, 50.627111 }, //06 February 1980, UT1-UTC= 0.5568893, Accumulated Leap Seconds=19
  { 2444276.50, 50.629401 }, //07 February 1980, UT1-UTC= 0.5545991, Accumulated Leap Seconds=19
  { 2444277.50, 50.631671 }, //08 February 1980, UT1-UTC= 0.5523286, Accumulated Leap Seconds=19
  { 2444278.50, 50.633907 }, //09 February 1980, UT1-UTC= 0.5500926, Accumulated Leap Seconds=19
  { 2444279.50, 50.636103 }, //10 February 1980, UT1-UTC= 0.5478967, Accumulated Leap Seconds=19
  { 2444280.50, 50.638265 }, //11 February 1980, UT1-UTC= 0.5457353, Accumulated Leap Seconds=19
  { 2444281.50, 50.640412 }, //12 February 1980, UT1-UTC= 0.5435882, Accumulated Leap Seconds=19
  { 2444282.50, 50.642584 }, //13 February 1980, UT1-UTC= 0.5414157, Accumulated Leap Seconds=19
  { 2444283.50, 50.644842 }, //14 February 1980, UT1-UTC= 0.5391578, Accumulated Leap Seconds=19
  { 2444284.50, 50.647253 }, //15 February 1980, UT1-UTC= 0.5367465, Accumulated Leap Seconds=19
  { 2444285.50, 50.649868 }, //16 February 1980, UT1-UTC= 0.5341320, Accumulated Leap Seconds=19
  { 2444286.50, 50.652696 }, //17 February 1980, UT1-UTC= 0.5313039, Accumulated Leap Seconds=19
  { 2444287.50, 50.655698 }, //18 February 1980, UT1-UTC= 0.5283020, Accumulated Leap Seconds=19
  { 2444288.50, 50.658790 }, //19 February 1980, UT1-UTC= 0.5252095, Accumulated Leap Seconds=19
  { 2444289.50, 50.661871 }, //20 February 1980, UT1-UTC= 0.5221292, Accumulated Leap Seconds=19
  { 2444290.50, 50.664847 }, //21 February 1980, UT1-UTC= 0.5191531, Accumulated Leap Seconds=19
  { 2444291.50, 50.667660 }, //22 February 1980, UT1-UTC= 0.5163397, Accumulated Leap Seconds=19
  { 2444292.50, 50.670294 }, //23 February 1980, UT1-UTC= 0.5137060, Accumulated Leap Seconds=19
  { 2444293.50, 50.672767 }, //24 February 1980, UT1-UTC= 0.5112333, Accumulated Leap Seconds=19
  { 2444294.50, 50.675118 }, //25 February 1980, UT1-UTC= 0.5088817, Accumulated Leap Seconds=19
  { 2444295.50, 50.677398 }, //26 February 1980, UT1-UTC= 0.5066019, Accumulated Leap Seconds=19
  { 2444296.50, 50.679654 }, //27 February 1980, UT1-UTC= 0.5043462, Accumulated Leap Seconds=19
  { 2444297.50, 50.681924 }, //28 February 1980, UT1-UTC= 0.5020762, Accumulated Leap Seconds=19
  { 2444298.50, 50.684233 }, //29 February 1980, UT1-UTC= 0.4997665, Accumulated Leap Seconds=19
  { 2444299.50, 50.686594 }, //01 March 1980, UT1-UTC= 0.4974059, Accumulated Leap Seconds=19
  { 2444300.50, 50.689005 }, //02 March 1980, UT1-UTC= 0.4949949, Accumulated Leap Seconds=19
  { 2444301.50, 50.691456 }, //03 March 1980, UT1-UTC= 0.4925444, Accumulated Leap Seconds=19
  { 2444302.50, 50.693925 }, //04 March 1980, UT1-UTC= 0.4900745, Accumulated Leap Seconds=19
  { 2444303.50, 50.696389 }, //05 March 1980, UT1-UTC= 0.4876114, Accumulated Leap Seconds=19
  { 2444304.50, 50.698817 }, //06 March 1980, UT1-UTC= 0.4851831, Accumulated Leap Seconds=19
  { 2444305.50, 50.701187 }, //07 March 1980, UT1-UTC= 0.4828125, Accumulated Leap Seconds=19
  { 2444306.50, 50.703488 }, //08 March 1980, UT1-UTC= 0.4805115, Accumulated Leap Seconds=19
  { 2444307.50, 50.705722 }, //09 March 1980, UT1-UTC= 0.4782783, Accumulated Leap Seconds=19
  { 2444308.50, 50.707903 }, //10 March 1980, UT1-UTC= 0.4760968, Accumulated Leap Seconds=19
  { 2444309.50, 50.710066 }, //11 March 1980, UT1-UTC= 0.4739335, Accumulated Leap Seconds=19
  { 2444310.50, 50.712264 }, //12 March 1980, UT1-UTC= 0.4717359, Accumulated Leap Seconds=19
  { 2444311.50, 50.714563 }, //13 March 1980, UT1-UTC= 0.4694373, Accumulated Leap Seconds=19
  { 2444312.50, 50.717029 }, //14 March 1980, UT1-UTC= 0.4669708, Accumulated Leap Seconds=19
  { 2444313.50, 50.719708 }, //15 March 1980, UT1-UTC= 0.4642916, Accumulated Leap Seconds=19
  { 2444314.50, 50.722601 }, //16 March 1980, UT1-UTC= 0.4613989, Accumulated Leap Seconds=19
  { 2444315.50, 50.725653 }, //17 March 1980, UT1-UTC= 0.4583468, Accumulated Leap Seconds=19
  { 2444316.50, 50.728769 }, //18 March 1980, UT1-UTC= 0.4552311, Accumulated Leap Seconds=19
  { 2444317.50, 50.731841 }, //19 March 1980, UT1-UTC= 0.4521590, Accumulated Leap Seconds=19
  { 2444318.50, 50.734784 }, //20 March 1980, UT1-UTC= 0.4492160, Accumulated Leap Seconds=19
  { 2444319.50, 50.737556 }, //21 March 1980, UT1-UTC= 0.4464437, Accumulated Leap Seconds=19
  { 2444320.50, 50.740163 }, //22 March 1980, UT1-UTC= 0.4438365, Accumulated Leap Seconds=19
  { 2444321.50, 50.742645 }, //23 March 1980, UT1-UTC= 0.4413547, Accumulated Leap Seconds=19
  { 2444322.50, 50.745057 }, //24 March 1980, UT1-UTC= 0.4389426, Accumulated Leap Seconds=19
  { 2444323.50, 50.747456 }, //25 March 1980, UT1-UTC= 0.4365445, Accumulated Leap Seconds=19
  { 2444324.50, 50.749885 }, //26 March 1980, UT1-UTC= 0.4341151, Accumulated Leap Seconds=19
  { 2444325.50, 50.752376 }, //27 March 1980, UT1-UTC= 0.4316235, Accumulated Leap Seconds=19
  { 2444326.50, 50.754946 }, //28 March 1980, UT1-UTC= 0.4290541, Accumulated Leap Seconds=19
  { 2444327.50, 50.757593 }, //29 March 1980, UT1-UTC= 0.4264066, Accumulated Leap Seconds=19
  { 2444328.50, 50.760307 }, //30 March 1980, UT1-UTC= 0.4236932, Accumulated Leap Seconds=19
  { 2444329.50, 50.763064 }, //31 March 1980, UT1-UTC= 0.4209361, Accumulated Leap Seconds=19
  { 2444330.50, 50.765836 }, //01 April 1980, UT1-UTC= 0.4181638, Accumulated Leap Seconds=19
  { 2444331.50, 50.768592 }, //02 April 1980, UT1-UTC= 0.4154084, Accumulated Leap Seconds=19
  { 2444332.50, 50.771299 }, //03 April 1980, UT1-UTC= 0.4127006, Accumulated Leap Seconds=19
  { 2444333.50, 50.773937 }, //04 April 1980, UT1-UTC= 0.4100631, Accumulated Leap Seconds=19
  { 2444334.50, 50.776494 }, //05 April 1980, UT1-UTC= 0.4075055, Accumulated Leap Seconds=19
  { 2444335.50, 50.778979 }, //06 April 1980, UT1-UTC= 0.4050213, Accumulated Leap Seconds=19
  { 2444336.50, 50.781414 }, //07 April 1980, UT1-UTC= 0.4025860, Accumulated Leap Seconds=19
  { 2444337.50, 50.783841 }, //08 April 1980, UT1-UTC= 0.4001589, Accumulated Leap Seconds=19
  { 2444338.50, 50.786314 }, //09 April 1980, UT1-UTC= 0.3976860, Accumulated Leap Seconds=19
  { 2444339.50, 50.788894 }, //10 April 1980, UT1-UTC= 0.3951058, Accumulated Leap Seconds=19
  { 2444340.50, 50.791636 }, //11 April 1980, UT1-UTC= 0.3923640, Accumulated Leap Seconds=19
  { 2444341.50, 50.794566 }, //12 April 1980, UT1-UTC= 0.3894342, Accumulated Leap Seconds=19
  { 2444342.50, 50.797664 }, //13 April 1980, UT1-UTC= 0.3863356, Accumulated Leap Seconds=19
  { 2444343.50, 50.800863 }, //14 April 1980, UT1-UTC= 0.3831366, Accumulated Leap Seconds=19
  { 2444344.50, 50.804061 }, //15 April 1980, UT1-UTC= 0.3799394, Accumulated Leap Seconds=19
  { 2444345.50, 50.807152 }, //16 April 1980, UT1-UTC= 0.3768477, Accumulated Leap Seconds=19
  { 2444346.50, 50.810066 }, //17 April 1980, UT1-UTC= 0.3739345, Accumulated Leap Seconds=19
  { 2444347.50, 50.812777 }, //18 April 1980, UT1-UTC= 0.3712231, Accumulated Leap Seconds=19
  { 2444348.50, 50.815311 }, //19 April 1980, UT1-UTC= 0.3686894, Accumulated Leap Seconds=19
  { 2444349.50, 50.817721 }, //20 April 1980, UT1-UTC= 0.3662792, Accumulated Leap Seconds=19
  { 2444350.50, 50.820070 }, //21 April 1980, UT1-UTC= 0.3639295, Accumulated Leap Seconds=19
  { 2444351.50, 50.822415 }, //22 April 1980, UT1-UTC= 0.3615854, Accumulated Leap Seconds=19
  { 2444352.50, 50.824794 }, //23 April 1980, UT1-UTC= 0.3592064, Accumulated Leap Seconds=19
  { 2444353.50, 50.827232 }, //24 April 1980, UT1-UTC= 0.3567678, Accumulated Leap Seconds=19
  { 2444354.50, 50.829740 }, //25 April 1980, UT1-UTC= 0.3542602, Accumulated Leap Seconds=19
  { 2444355.50, 50.832312 }, //26 April 1980, UT1-UTC= 0.3516879, Accumulated Leap Seconds=19
  { 2444356.50, 50.834933 }, //27 April 1980, UT1-UTC= 0.3490669, Accumulated Leap Seconds=19
  { 2444357.50, 50.837578 }, //28 April 1980, UT1-UTC= 0.3464219, Accumulated Leap Seconds=19
  { 2444358.50, 50.840218 }, //29 April 1980, UT1-UTC= 0.3437821, Accumulated Leap Seconds=19
  { 2444359.50, 50.842823 }, //30 April 1980, UT1-UTC= 0.3411765, Accumulated Leap Seconds=19
  { 2444360.50, 50.845370 }, //01 May 1980, UT1-UTC= 0.3386302, Accumulated Leap Seconds=19
  { 2444361.50, 50.847841 }, //02 May 1980, UT1-UTC= 0.3361585, Accumulated Leap Seconds=19
  { 2444362.50, 50.850238 }, //03 May 1980, UT1-UTC= 0.3337620, Accumulated Leap Seconds=19
  { 2444363.50, 50.852577 }, //04 May 1980, UT1-UTC= 0.3314229, Accumulated Leap Seconds=19
  { 2444364.50, 50.854894 }, //05 May 1980, UT1-UTC= 0.3291059, Accumulated Leap Seconds=19
  { 2444365.50, 50.857238 }, //06 May 1980, UT1-UTC= 0.3267622, Accumulated Leap Seconds=19
  { 2444366.50, 50.859663 }, //07 May 1980, UT1-UTC= 0.3243374, Accumulated Leap Seconds=19
  { 2444367.50, 50.862218 }, //08 May 1980, UT1-UTC= 0.3217821, Accumulated Leap Seconds=19
  { 2444368.50, 50.864935 }, //09 May 1980, UT1-UTC= 0.3190645, Accumulated Leap Seconds=19
  { 2444369.50, 50.867815 }, //10 May 1980, UT1-UTC= 0.3161854, Accumulated Leap Seconds=19
  { 2444370.50, 50.870813 }, //11 May 1980, UT1-UTC= 0.3131870, Accumulated Leap Seconds=19
  { 2444371.50, 50.873851 }, //12 May 1980, UT1-UTC= 0.3101489, Accumulated Leap Seconds=19
  { 2444372.50, 50.876831 }, //13 May 1980, UT1-UTC= 0.3071692, Accumulated Leap Seconds=19
  { 2444373.50, 50.879664 }, //14 May 1980, UT1-UTC= 0.3043362, Accumulated Leap Seconds=19
  { 2444374.50, 50.882298 }, //15 May 1980, UT1-UTC= 0.3017021, Accumulated Leap Seconds=19
  { 2444375.50, 50.884729 }, //16 May 1980, UT1-UTC= 0.2992708, Accumulated Leap Seconds=19
  { 2444376.50, 50.886995 }, //17 May 1980, UT1-UTC= 0.2970052, Accumulated Leap Seconds=19
  { 2444377.50, 50.889153 }, //18 May 1980, UT1-UTC= 0.2948472, Accumulated Leap Seconds=19
  { 2444378.50, 50.891262 }, //19 May 1980, UT1-UTC= 0.2927380, Accumulated Leap Seconds=19
  { 2444379.50, 50.893369 }, //20 May 1980, UT1-UTC= 0.2906307, Accumulated Leap Seconds=19
  { 2444380.50, 50.895505 }, //21 May 1980, UT1-UTC= 0.2884949, Accumulated Leap Seconds=19
  { 2444381.50, 50.897683 }, //22 May 1980, UT1-UTC= 0.2863167, Accumulated Leap Seconds=19
  { 2444382.50, 50.899904 }, //23 May 1980, UT1-UTC= 0.2840964, Accumulated Leap Seconds=19
  { 2444383.50, 50.902153 }, //24 May 1980, UT1-UTC= 0.2818466, Accumulated Leap Seconds=19
  { 2444384.50, 50.904411 }, //25 May 1980, UT1-UTC= 0.2795889, Accumulated Leap Seconds=19
  { 2444385.50, 50.906650 }, //26 May 1980, UT1-UTC= 0.2773499, Accumulated Leap Seconds=19
  { 2444386.50, 50.908844 }, //27 May 1980, UT1-UTC= 0.2751562, Accumulated Leap Seconds=19
  { 2444387.50, 50.910969 }, //28 May 1980, UT1-UTC= 0.2730310, Accumulated Leap Seconds=19
  { 2444388.50, 50.913010 }, //29 May 1980, UT1-UTC= 0.2709900, Accumulated Leap Seconds=19
  { 2444389.50, 50.914962 }, //30 May 1980, UT1-UTC= 0.2690376, Accumulated Leap Seconds=19
  { 2444390.50, 50.916839 }, //31 May 1980, UT1-UTC= 0.2671610, Accumulated Leap Seconds=19
  { 2444391.50, 50.918672 }, //01 June 1980, UT1-UTC= 0.2653280, Accumulated Leap Seconds=19
  { 2444392.50, 50.920510 }, //02 June 1980, UT1-UTC= 0.2634902, Accumulated Leap Seconds=19
  { 2444393.50, 50.922408 }, //03 June 1980, UT1-UTC= 0.2615918, Accumulated Leap Seconds=19
  { 2444394.50, 50.924417 }, //04 June 1980, UT1-UTC= 0.2595828, Accumulated Leap Seconds=19
  { 2444395.50, 50.926569 }, //05 June 1980, UT1-UTC= 0.2574306, Accumulated Leap Seconds=19
  { 2444396.50, 50.928870 }, //06 June 1980, UT1-UTC= 0.2551297, Accumulated Leap Seconds=19
  { 2444397.50, 50.931294 }, //07 June 1980, UT1-UTC= 0.2527062, Accumulated Leap Seconds=19
  { 2444398.50, 50.933783 }, //08 June 1980, UT1-UTC= 0.2502166, Accumulated Leap Seconds=19
  { 2444399.50, 50.936262 }, //09 June 1980, UT1-UTC= 0.2477381, Accumulated Leap Seconds=19
  { 2444400.50, 50.938650 }, //10 June 1980, UT1-UTC= 0.2453499, Accumulated Leap Seconds=19
  { 2444401.50, 50.940889 }, //11 June 1980, UT1-UTC= 0.2431111, Accumulated Leap Seconds=19
  { 2444402.50, 50.942955 }, //12 June 1980, UT1-UTC= 0.2410453, Accumulated Leap Seconds=19
  { 2444403.50, 50.944864 }, //13 June 1980, UT1-UTC= 0.2391361, Accumulated Leap Seconds=19
  { 2444404.50, 50.946662 }, //14 June 1980, UT1-UTC= 0.2373381, Accumulated Leap Seconds=19
  { 2444405.50, 50.948405 }, //15 June 1980, UT1-UTC= 0.2355946, Accumulated Leap Seconds=19
  { 2444406.50, 50.950145 }, //16 June 1980, UT1-UTC= 0.2338555, Accumulated Leap Seconds=19
  { 2444407.50, 50.951913 }, //17 June 1980, UT1-UTC= 0.2320865, Accumulated Leap Seconds=19
  { 2444408.50, 50.953730 }, //18 June 1980, UT1-UTC= 0.2302704, Accumulated Leap Seconds=19
  { 2444409.50, 50.955595 }, //19 June 1980, UT1-UTC= 0.2284052, Accumulated Leap Seconds=19
  { 2444410.50, 50.957498 }, //20 June 1980, UT1-UTC= 0.2265019, Accumulated Leap Seconds=19
  { 2444411.50, 50.959419 }, //21 June 1980, UT1-UTC= 0.2245812, Accumulated Leap Seconds=19
  { 2444412.50, 50.961329 }, //22 June 1980, UT1-UTC= 0.2226707, Accumulated Leap Seconds=19
  { 2444413.50, 50.963202 }, //23 June 1980, UT1-UTC= 0.2207984, Accumulated Leap Seconds=19
  { 2444414.50, 50.965012 }, //24 June 1980, UT1-UTC= 0.2189880, Accumulated Leap Seconds=19
  { 2444415.50, 50.966743 }, //25 June 1980, UT1-UTC= 0.2172565, Accumulated Leap Seconds=19
  { 2444416.50, 50.968389 }, //26 June 1980, UT1-UTC= 0.2156113, Accumulated Leap Seconds=19
  { 2444417.50, 50.969953 }, //27 June 1980, UT1-UTC= 0.2140471, Accumulated Leap Seconds=19
  { 2444418.50, 50.971459 }, //28 June 1980, UT1-UTC= 0.2125405, Accumulated Leap Seconds=19
  { 2444419.50, 50.972951 }, //29 June 1980, UT1-UTC= 0.2110491, Accumulated Leap Seconds=19
  { 2444420.50, 50.974484 }, //30 June 1980, UT1-UTC= 0.2095162, Accumulated Leap Seconds=19
  { 2444421.50, 50.976115 }, //01 July 1980, UT1-UTC= 0.2078852, Accumulated Leap Seconds=19
  { 2444422.50, 50.977883 }, //02 July 1980, UT1-UTC= 0.2061174, Accumulated Leap Seconds=19
  { 2444423.50, 50.979795 }, //03 July 1980, UT1-UTC= 0.2042045, Accumulated Leap Seconds=19
  { 2444424.50, 50.981827 }, //04 July 1980, UT1-UTC= 0.2021725, Accumulated Leap Seconds=19
  { 2444425.50, 50.983924 }, //05 July 1980, UT1-UTC= 0.2000763, Accumulated Leap Seconds=19
  { 2444426.50, 50.986012 }, //06 July 1980, UT1-UTC= 0.1979878, Accumulated Leap Seconds=19
  { 2444427.50, 50.988022 }, //07 July 1980, UT1-UTC= 0.1959782, Accumulated Leap Seconds=19
  { 2444428.50, 50.989897 }, //08 July 1980, UT1-UTC= 0.1941025, Accumulated Leap Seconds=19
  { 2444429.50, 50.991612 }, //09 July 1980, UT1-UTC= 0.1923876, Accumulated Leap Seconds=19
  { 2444430.50, 50.993173 }, //10 July 1980, UT1-UTC= 0.1908268, Accumulated Leap Seconds=19
  { 2444431.50, 50.994617 }, //11 July 1980, UT1-UTC= 0.1893832, Accumulated Leap Seconds=19
  { 2444432.50, 50.995998 }, //12 July 1980, UT1-UTC= 0.1880024, Accumulated Leap Seconds=19
  { 2444433.50, 50.997370 }, //13 July 1980, UT1-UTC= 0.1866302, Accumulated Leap Seconds=19
  { 2444434.50, 50.998775 }, //14 July 1980, UT1-UTC= 0.1852252, Accumulated Leap Seconds=19
  { 2444435.50, 51.000237 }, //15 July 1980, UT1-UTC= 0.1837628, Accumulated Leap Seconds=19
  { 2444436.50, 51.001766 }, //16 July 1980, UT1-UTC= 0.1822344, Accumulated Leap Seconds=19
  { 2444437.50, 51.003355 }, //17 July 1980, UT1-UTC= 0.1806451, Accumulated Leap Seconds=19
  { 2444438.50, 51.004989 }, //18 July 1980, UT1-UTC= 0.1790106, Accumulated Leap Seconds=19
  { 2444439.50, 51.006645 }, //19 July 1980, UT1-UTC= 0.1773545, Accumulated Leap Seconds=19
  { 2444440.50, 51.008297 }, //20 July 1980, UT1-UTC= 0.1757029, Accumulated Leap Seconds=19
  { 2444441.50, 51.009919 }, //21 July 1980, UT1-UTC= 0.1740805, Accumulated Leap Seconds=19
  { 2444442.50, 51.011493 }, //22 July 1980, UT1-UTC= 0.1725069, Accumulated Leap Seconds=19
  { 2444443.50, 51.013007 }, //23 July 1980, UT1-UTC= 0.1709927, Accumulated Leap Seconds=19
  { 2444444.50, 51.014463 }, //24 July 1980, UT1-UTC= 0.1695374, Accumulated Leap Seconds=19
  { 2444445.50, 51.015874 }, //25 July 1980, UT1-UTC= 0.1681259, Accumulated Leap Seconds=19
  { 2444446.50, 51.017274 }, //26 July 1980, UT1-UTC= 0.1667259, Accumulated Leap Seconds=19
  { 2444447.50, 51.018712 }, //27 July 1980, UT1-UTC= 0.1652882, Accumulated Leap Seconds=19
  { 2444448.50, 51.020244 }, //28 July 1980, UT1-UTC= 0.1637564, Accumulated Leap Seconds=19
  { 2444449.50, 51.021918 }, //29 July 1980, UT1-UTC= 0.1620816, Accumulated Leap Seconds=19
  { 2444450.50, 51.023757 }, //30 July 1980, UT1-UTC= 0.1602432, Accumulated Leap Seconds=19
  { 2444451.50, 51.025738 }, //31 July 1980, UT1-UTC= 0.1582620, Accumulated Leap Seconds=19
  { 2444452.50, 51.027802 }, //01 August 1980, UT1-UTC= 0.1561983, Accumulated Leap Seconds=19
  { 2444453.50, 51.029866 }, //02 August 1980, UT1-UTC= 0.1541336, Accumulated Leap Seconds=19
  { 2444454.50, 51.031852 }, //03 August 1980, UT1-UTC= 0.1521480, Accumulated Leap Seconds=19
  { 2444455.50, 51.033700 }, //04 August 1980, UT1-UTC= 0.1503001, Accumulated Leap Seconds=19
  { 2444456.50, 51.035384 }, //05 August 1980, UT1-UTC= 0.1486155, Accumulated Leap Seconds=19
  { 2444457.50, 51.036915 }, //06 August 1980, UT1-UTC= 0.1470852, Accumulated Leap Seconds=19
  { 2444458.50, 51.038328 }, //07 August 1980, UT1-UTC= 0.1456720, Accumulated Leap Seconds=19
  { 2444459.50, 51.039679 }, //08 August 1980, UT1-UTC= 0.1443215, Accumulated Leap Seconds=19
  { 2444460.50, 51.041023 }, //09 August 1980, UT1-UTC= 0.1429769, Accumulated Leap Seconds=19
  { 2444461.50, 51.042409 }, //10 August 1980, UT1-UTC= 0.1415911, Accumulated Leap Seconds=19
  { 2444462.50, 51.043866 }, //11 August 1980, UT1-UTC= 0.1401338, Accumulated Leap Seconds=19
  { 2444463.50, 51.045407 }, //12 August 1980, UT1-UTC= 0.1385931, Accumulated Leap Seconds=19
  { 2444464.50, 51.047028 }, //13 August 1980, UT1-UTC= 0.1369724, Accumulated Leap Seconds=19
  { 2444465.50, 51.048713 }, //14 August 1980, UT1-UTC= 0.1352866, Accumulated Leap Seconds=19
  { 2444466.50, 51.050441 }, //15 August 1980, UT1-UTC= 0.1335591, Accumulated Leap Seconds=19
  { 2444467.50, 51.052183 }, //16 August 1980, UT1-UTC= 0.1318170, Accumulated Leap Seconds=19
  { 2444468.50, 51.053913 }, //17 August 1980, UT1-UTC= 0.1300871, Accumulated Leap Seconds=19
  { 2444469.50, 51.055610 }, //18 August 1980, UT1-UTC= 0.1283896, Accumulated Leap Seconds=19
  { 2444470.50, 51.057266 }, //19 August 1980, UT1-UTC= 0.1267338, Accumulated Leap Seconds=19
  { 2444471.50, 51.058883 }, //20 August 1980, UT1-UTC= 0.1251170, Accumulated Leap Seconds=19
  { 2444472.50, 51.060476 }, //21 August 1980, UT1-UTC= 0.1235244, Accumulated Leap Seconds=19
  { 2444473.50, 51.062074 }, //22 August 1980, UT1-UTC= 0.1219262, Accumulated Leap Seconds=19
  { 2444474.50, 51.063724 }, //23 August 1980, UT1-UTC= 0.1202764, Accumulated Leap Seconds=19
  { 2444475.50, 51.065485 }, //24 August 1980, UT1-UTC= 0.1185148, Accumulated Leap Seconds=19
  { 2444476.50, 51.067420 }, //25 August 1980, UT1-UTC= 0.1165802, Accumulated Leap Seconds=19
  { 2444477.50, 51.069568 }, //26 August 1980, UT1-UTC= 0.1144316, Accumulated Leap Seconds=19
  { 2444478.50, 51.071930 }, //27 August 1980, UT1-UTC= 0.1120698, Accumulated Leap Seconds=19
  { 2444479.50, 51.074454 }, //28 August 1980, UT1-UTC= 0.1095462, Accumulated Leap Seconds=19
  { 2444480.50, 51.077047 }, //29 August 1980, UT1-UTC= 0.1069525, Accumulated Leap Seconds=19
  { 2444481.50, 51.079607 }, //30 August 1980, UT1-UTC= 0.1043928, Accumulated Leap Seconds=19
  { 2444482.50, 51.082047 }, //31 August 1980, UT1-UTC= 0.1019525, Accumulated Leap Seconds=19
  { 2444483.50, 51.084323 }, //01 September 1980, UT1-UTC= 0.0996770, Accumulated Leap Seconds=19
  { 2444484.50, 51.086434 }, //02 September 1980, UT1-UTC= 0.0975662, Accumulated Leap Seconds=19
  { 2444485.50, 51.088416 }, //03 September 1980, UT1-UTC= 0.0955840, Accumulated Leap Seconds=19
  { 2444486.50, 51.090325 }, //04 September 1980, UT1-UTC= 0.0936746, Accumulated Leap Seconds=19
  { 2444487.50, 51.092222 }, //05 September 1980, UT1-UTC= 0.0917784, Accumulated Leap Seconds=19
  { 2444488.50, 51.094155 }, //06 September 1980, UT1-UTC= 0.0898453, Accumulated Leap Seconds=19
  { 2444489.50, 51.096158 }, //07 September 1980, UT1-UTC= 0.0878418, Accumulated Leap Seconds=19
  { 2444490.50, 51.098247 }, //08 September 1980, UT1-UTC= 0.0857530, Accumulated Leap Seconds=19
  { 2444491.50, 51.100418 }, //09 September 1980, UT1-UTC= 0.0835816, Accumulated Leap Seconds=19
  { 2444492.50, 51.102656 }, //10 September 1980, UT1-UTC= 0.0813438, Accumulated Leap Seconds=19
  { 2444493.50, 51.104935 }, //11 September 1980, UT1-UTC= 0.0790647, Accumulated Leap Seconds=19
  { 2444494.50, 51.107225 }, //12 September 1980, UT1-UTC= 0.0767746, Accumulated Leap Seconds=19
  { 2444495.50, 51.109496 }, //13 September 1980, UT1-UTC= 0.0745041, Accumulated Leap Seconds=19
  { 2444496.50, 51.111720 }, //14 September 1980, UT1-UTC= 0.0722798, Accumulated Leap Seconds=19
  { 2444497.50, 51.113882 }, //15 September 1980, UT1-UTC= 0.0701176, Accumulated Leap Seconds=19
  { 2444498.50, 51.115981 }, //16 September 1980, UT1-UTC= 0.0680194, Accumulated Leap Seconds=19
  { 2444499.50, 51.118027 }, //17 September 1980, UT1-UTC= 0.0659725, Accumulated Leap Seconds=19
  { 2444500.50, 51.120050 }, //18 September 1980, UT1-UTC= 0.0639502, Accumulated Leap Seconds=19
  { 2444501.50, 51.122089 }, //19 September 1980, UT1-UTC= 0.0619105, Accumulated Leap Seconds=19
  { 2444502.50, 51.124203 }, //20 September 1980, UT1-UTC= 0.0597970, Accumulated Leap Seconds=19
  { 2444503.50, 51.126456 }, //21 September 1980, UT1-UTC= 0.0575444, Accumulated Leap Seconds=19
  { 2444504.50, 51.128907 }, //22 September 1980, UT1-UTC= 0.0550927, Accumulated Leap Seconds=19
  { 2444505.50, 51.131590 }, //23 September 1980, UT1-UTC= 0.0524103, Accumulated Leap Seconds=19
  { 2444506.50, 51.134486 }, //24 September 1980, UT1-UTC= 0.0495143, Accumulated Leap Seconds=19
  { 2444507.50, 51.137524 }, //25 September 1980, UT1-UTC= 0.0464761, Accumulated Leap Seconds=19
  { 2444508.50, 51.140597 }, //26 September 1980, UT1-UTC= 0.0434028, Accumulated Leap Seconds=19
  { 2444509.50, 51.143596 }, //27 September 1980, UT1-UTC= 0.0404041, Accumulated Leap Seconds=19
  { 2444510.50, 51.146439 }, //28 September 1980, UT1-UTC= 0.0375614, Accumulated Leap Seconds=19
  { 2444511.50, 51.149078 }, //29 September 1980, UT1-UTC= 0.0349219, Accumulated Leap Seconds=19
  { 2444512.50, 51.151513 }, //30 September 1980, UT1-UTC= 0.0324871, Accumulated Leap Seconds=19
  { 2444513.50, 51.153793 }, //01 October 1980, UT1-UTC= 0.0302072, Accumulated Leap Seconds=19
  { 2444514.50, 51.155990 }, //02 October 1980, UT1-UTC= 0.0280100, Accumulated Leap Seconds=19
  { 2444515.50, 51.158173 }, //03 October 1980, UT1-UTC= 0.0258271, Accumulated Leap Seconds=19
  { 2444516.50, 51.160392 }, //04 October 1980, UT1-UTC= 0.0236077, Accumulated Leap Seconds=19
  { 2444517.50, 51.162676 }, //05 October 1980, UT1-UTC= 0.0213242, Accumulated Leap Seconds=19
  { 2444518.50, 51.165031 }, //06 October 1980, UT1-UTC= 0.0189688, Accumulated Leap Seconds=19
  { 2444519.50, 51.167451 }, //07 October 1980, UT1-UTC= 0.0165489, Accumulated Leap Seconds=19
  { 2444520.50, 51.169916 }, //08 October 1980, UT1-UTC= 0.0140836, Accumulated Leap Seconds=19
  { 2444521.50, 51.172401 }, //09 October 1980, UT1-UTC= 0.0115985, Accumulated Leap Seconds=19
  { 2444522.50, 51.174877 }, //10 October 1980, UT1-UTC= 0.0091226, Accumulated Leap Seconds=19
  { 2444523.50, 51.177317 }, //11 October 1980, UT1-UTC= 0.0066830, Accumulated Leap Seconds=19
  { 2444524.50, 51.179700 }, //12 October 1980, UT1-UTC= 0.0043002, Accumulated Leap Seconds=19
  { 2444525.50, 51.182016 }, //13 October 1980, UT1-UTC= 0.0019836, Accumulated Leap Seconds=19
  { 2444526.50, 51.184273 }, //14 October 1980, UT1-UTC=-0.0002728, Accumulated Leap Seconds=19
  { 2444527.50, 51.186490 }, //15 October 1980, UT1-UTC=-0.0024905, Accumulated Leap Seconds=19
  { 2444528.50, 51.188704 }, //16 October 1980, UT1-UTC=-0.0047041, Accumulated Leap Seconds=19
  { 2444529.50, 51.190960 }, //17 October 1980, UT1-UTC=-0.0069601, Accumulated Leap Seconds=19
  { 2444530.50, 51.193313 }, //18 October 1980, UT1-UTC=-0.0093133, Accumulated Leap Seconds=19
  { 2444531.50, 51.195818 }, //19 October 1980, UT1-UTC=-0.0118184, Accumulated Leap Seconds=19
  { 2444532.50, 51.198527 }, //20 October 1980, UT1-UTC=-0.0145267, Accumulated Leap Seconds=19
  { 2444533.50, 51.201457 }, //21 October 1980, UT1-UTC=-0.0174568, Accumulated Leap Seconds=19
  { 2444534.50, 51.204579 }, //22 October 1980, UT1-UTC=-0.0205794, Accumulated Leap Seconds=19
  { 2444535.50, 51.207807 }, //23 October 1980, UT1-UTC=-0.0238069, Accumulated Leap Seconds=19
  { 2444536.50, 51.211017 }, //24 October 1980, UT1-UTC=-0.0270174, Accumulated Leap Seconds=19
  { 2444537.50, 51.214101 }, //25 October 1980, UT1-UTC=-0.0301012, Accumulated Leap Seconds=19
  { 2444538.50, 51.216991 }, //26 October 1980, UT1-UTC=-0.0329912, Accumulated Leap Seconds=19
  { 2444539.50, 51.219677 }, //27 October 1980, UT1-UTC=-0.0356769, Accumulated Leap Seconds=19
  { 2444540.50, 51.222198 }, //28 October 1980, UT1-UTC=-0.0381975, Accumulated Leap Seconds=19
  { 2444541.50, 51.224618 }, //29 October 1980, UT1-UTC=-0.0406183, Accumulated Leap Seconds=19
  { 2444542.50, 51.227006 }, //30 October 1980, UT1-UTC=-0.0430060, Accumulated Leap Seconds=19
  { 2444543.50, 51.229412 }, //31 October 1980, UT1-UTC=-0.0454116, Accumulated Leap Seconds=19
  { 2444544.50, 51.231865 }, //01 November 1980, UT1-UTC=-0.0478645, Accumulated Leap Seconds=19
  { 2444545.50, 51.234374 }, //02 November 1980, UT1-UTC=-0.0503741, Accumulated Leap Seconds=19
  { 2444546.50, 51.236933 }, //03 November 1980, UT1-UTC=-0.0529330, Accumulated Leap Seconds=19
  { 2444547.50, 51.239522 }, //04 November 1980, UT1-UTC=-0.0555216, Accumulated Leap Seconds=19
  { 2444548.50, 51.242113 }, //05 November 1980, UT1-UTC=-0.0581126, Accumulated Leap Seconds=19
  { 2444549.50, 51.244675 }, //06 November 1980, UT1-UTC=-0.0606753, Accumulated Leap Seconds=19
  { 2444550.50, 51.247180 }, //07 November 1980, UT1-UTC=-0.0631800, Accumulated Leap Seconds=19
  { 2444551.50, 51.249603 }, //08 November 1980, UT1-UTC=-0.0656032, Accumulated Leap Seconds=19
  { 2444552.50, 51.251932 }, //09 November 1980, UT1-UTC=-0.0679318, Accumulated Leap Seconds=19
  { 2444553.50, 51.254166 }, //10 November 1980, UT1-UTC=-0.0701665, Accumulated Leap Seconds=19
  { 2444554.50, 51.256324 }, //11 November 1980, UT1-UTC=-0.0723244, Accumulated Leap Seconds=19
  { 2444555.50, 51.258439 }, //12 November 1980, UT1-UTC=-0.0744394, Accumulated Leap Seconds=19
  { 2444556.50, 51.260559 }, //13 November 1980, UT1-UTC=-0.0765590, Accumulated Leap Seconds=19
  { 2444557.50, 51.262738 }, //14 November 1980, UT1-UTC=-0.0787384, Accumulated Leap Seconds=19
  { 2444558.50, 51.265034 }, //15 November 1980, UT1-UTC=-0.0810339, Accumulated Leap Seconds=19
  { 2444559.50, 51.267493 }, //16 November 1980, UT1-UTC=-0.0834934, Accumulated Leap Seconds=19
  { 2444560.50, 51.270144 }, //17 November 1980, UT1-UTC=-0.0861438, Accumulated Leap Seconds=19
  { 2444561.50, 51.272977 }, //18 November 1980, UT1-UTC=-0.0889769, Accumulated Leap Seconds=19
  { 2444562.50, 51.275941 }, //19 November 1980, UT1-UTC=-0.0919406, Accumulated Leap Seconds=19
  { 2444563.50, 51.278947 }, //20 November 1980, UT1-UTC=-0.0949470, Accumulated Leap Seconds=19
  { 2444564.50, 51.281893 }, //21 November 1980, UT1-UTC=-0.0978933, Accumulated Leap Seconds=19
  { 2444565.50, 51.284695 }, //22 November 1980, UT1-UTC=-0.1006952, Accumulated Leap Seconds=19
  { 2444566.50, 51.287313 }, //23 November 1980, UT1-UTC=-0.1033130, Accumulated Leap Seconds=19
  { 2444567.50, 51.289759 }, //24 November 1980, UT1-UTC=-0.1057592, Accumulated Leap Seconds=19
  { 2444568.50, 51.292086 }, //25 November 1980, UT1-UTC=-0.1080864, Accumulated Leap Seconds=19
  { 2444569.50, 51.294362 }, //26 November 1980, UT1-UTC=-0.1103624, Accumulated Leap Seconds=19
  { 2444570.50, 51.296646 }, //27 November 1980, UT1-UTC=-0.1126465, Accumulated Leap Seconds=19
  { 2444571.50, 51.298977 }, //28 November 1980, UT1-UTC=-0.1149769, Accumulated Leap Seconds=19
  { 2444572.50, 51.301369 }, //29 November 1980, UT1-UTC=-0.1173687, Accumulated Leap Seconds=19
  { 2444573.50, 51.303818 }, //30 November 1980, UT1-UTC=-0.1198180, Accumulated Leap Seconds=19
  { 2444574.50, 51.306308 }, //01 December 1980, UT1-UTC=-0.1223080, Accumulated Leap Seconds=19
  { 2444575.50, 51.308813 }, //02 December 1980, UT1-UTC=-0.1248129, Accumulated Leap Seconds=19
  { 2444576.50, 51.311303 }, //03 December 1980, UT1-UTC=-0.1273031, Accumulated Leap Seconds=19
  { 2444577.50, 51.313750 }, //04 December 1980, UT1-UTC=-0.1297503, Accumulated Leap Seconds=19
  { 2444578.50, 51.316131 }, //05 December 1980, UT1-UTC=-0.1321314, Accumulated Leap Seconds=19
  { 2444579.50, 51.318431 }, //06 December 1980, UT1-UTC=-0.1344309, Accumulated Leap Seconds=19
  { 2444580.50, 51.320645 }, //07 December 1980, UT1-UTC=-0.1366445, Accumulated Leap Seconds=19
  { 2444581.50, 51.322782 }, //08 December 1980, UT1-UTC=-0.1387825, Accumulated Leap Seconds=19
  { 2444582.50, 51.324871 }, //09 December 1980, UT1-UTC=-0.1408711, Accumulated Leap Seconds=19
  { 2444583.50, 51.326951 }, //10 December 1980, UT1-UTC=-0.1429511, Accumulated Leap Seconds=19
  { 2444584.50, 51.329074 }, //11 December 1980, UT1-UTC=-0.1450737, Accumulated Leap Seconds=19
  { 2444585.50, 51.331292 }, //12 December 1980, UT1-UTC=-0.1472916, Accumulated Leap Seconds=19
  { 2444586.50, 51.333648 }, //13 December 1980, UT1-UTC=-0.1496479, Accumulated Leap Seconds=19
  { 2444587.50, 51.336167 }, //14 December 1980, UT1-UTC=-0.1521666, Accumulated Leap Seconds=19
  { 2444588.50, 51.338844 }, //15 December 1980, UT1-UTC=-0.1548444, Accumulated Leap Seconds=19
  { 2444589.50, 51.341646 }, //16 December 1980, UT1-UTC=-0.1576463, Accumulated Leap Seconds=19
  { 2444590.50, 51.344507 }, //17 December 1980, UT1-UTC=-0.1605071, Accumulated Leap Seconds=19
  { 2444591.50, 51.347344 }, //18 December 1980, UT1-UTC=-0.1633439, Accumulated Leap Seconds=19
  { 2444592.50, 51.350077 }, //19 December 1980, UT1-UTC=-0.1660767, Accumulated Leap Seconds=19
  { 2444593.50, 51.352653 }, //20 December 1980, UT1-UTC=-0.1686533, Accumulated Leap Seconds=19
  { 2444594.50, 51.355064 }, //21 December 1980, UT1-UTC=-0.1710645, Accumulated Leap Seconds=19
  { 2444595.50, 51.357344 }, //22 December 1980, UT1-UTC=-0.1733442, Accumulated Leap Seconds=19
  { 2444596.50, 51.359553 }, //23 December 1980, UT1-UTC=-0.1755530, Accumulated Leap Seconds=19
  { 2444597.50, 51.361755 }, //24 December 1980, UT1-UTC=-0.1777547, Accumulated Leap Seconds=19
  { 2444598.50, 51.363997 }, //25 December 1980, UT1-UTC=-0.1799971, Accumulated Leap Seconds=19
  { 2444599.50, 51.366304 }, //26 December 1980, UT1-UTC=-0.1823038, Accumulated Leap Seconds=19
  { 2444600.50, 51.368676 }, //27 December 1980, UT1-UTC=-0.1846761, Accumulated Leap Seconds=19
  { 2444601.50, 51.371100 }, //28 December 1980, UT1-UTC=-0.1870997, Accumulated Leap Seconds=19
  { 2444602.50, 51.373551 }, //29 December 1980, UT1-UTC=-0.1895509, Accumulated Leap Seconds=19
  { 2444603.50, 51.376001 }, //30 December 1980, UT1-UTC=-0.1920009, Accumulated Leap Seconds=19
  { 2444604.50, 51.378421 }, //31 December 1980, UT1-UTC=-0.1944207, Accumulated Leap Seconds=19
  { 2444605.50, 51.380785 }, //01 January 1981, UT1-UTC=-0.1967849, Accumulated Leap Seconds=19
  { 2444606.50, 51.383077 }, //02 January 1981, UT1-UTC=-0.1990768, Accumulated Leap Seconds=19
  { 2444607.50, 51.385290 }, //03 January 1981, UT1-UTC=-0.2012902, Accumulated Leap Seconds=19
  { 2444608.50, 51.387432 }, //04 January 1981, UT1-UTC=-0.2034317, Accumulated Leap Seconds=19
  { 2444609.50, 51.389522 }, //05 January 1981, UT1-UTC=-0.2055222, Accumulated Leap Seconds=19
  { 2444610.50, 51.391598 }, //06 January 1981, UT1-UTC=-0.2075984, Accumulated Leap Seconds=19
  { 2444611.50, 51.393711 }, //07 January 1981, UT1-UTC=-0.2097110, Accumulated Leap Seconds=19
  { 2444612.50, 51.395916 }, //08 January 1981, UT1-UTC=-0.2119158, Accumulated Leap Seconds=19
  { 2444613.50, 51.398261 }, //09 January 1981, UT1-UTC=-0.2142612, Accumulated Leap Seconds=19
  { 2444614.50, 51.400773 }, //10 January 1981, UT1-UTC=-0.2167729, Accumulated Leap Seconds=19
  { 2444615.50, 51.403445 }, //11 January 1981, UT1-UTC=-0.2194455, Accumulated Leap Seconds=19
  { 2444616.50, 51.406240 }, //12 January 1981, UT1-UTC=-0.2222400, Accumulated Leap Seconds=19
  { 2444617.50, 51.409092 }, //13 January 1981, UT1-UTC=-0.2250916, Accumulated Leap Seconds=19
  { 2444618.50, 51.411924 }, //14 January 1981, UT1-UTC=-0.2279239, Accumulated Leap Seconds=19
  { 2444619.50, 51.414665 }, //15 January 1981, UT1-UTC=-0.2306646, Accumulated Leap Seconds=19
  { 2444620.50, 51.417262 }, //16 January 1981, UT1-UTC=-0.2332625, Accumulated Leap Seconds=19
  { 2444621.50, 51.419700 }, //17 January 1981, UT1-UTC=-0.2356997, Accumulated Leap Seconds=19
  { 2444622.50, 51.421996 }, //18 January 1981, UT1-UTC=-0.2379958, Accumulated Leap Seconds=19
  { 2444623.50, 51.424200 }, //19 January 1981, UT1-UTC=-0.2401999, Accumulated Leap Seconds=19
  { 2444624.50, 51.426372 }, //20 January 1981, UT1-UTC=-0.2423722, Accumulated Leap Seconds=19
  { 2444625.50, 51.428564 }, //21 January 1981, UT1-UTC=-0.2445643, Accumulated Leap Seconds=19
  { 2444626.50, 51.430806 }, //22 January 1981, UT1-UTC=-0.2468059, Accumulated Leap Seconds=19
  { 2444627.50, 51.433102 }, //23 January 1981, UT1-UTC=-0.2491021, Accumulated Leap Seconds=19
  { 2444628.50, 51.435438 }, //24 January 1981, UT1-UTC=-0.2514384, Accumulated Leap Seconds=19
  { 2444629.50, 51.437790 }, //25 January 1981, UT1-UTC=-0.2537904, Accumulated Leap Seconds=19
  { 2444630.50, 51.440129 }, //26 January 1981, UT1-UTC=-0.2561285, Accumulated Leap Seconds=19
  { 2444631.50, 51.442422 }, //27 January 1981, UT1-UTC=-0.2584219, Accumulated Leap Seconds=19
  { 2444632.50, 51.444643 }, //28 January 1981, UT1-UTC=-0.2606435, Accumulated Leap Seconds=19
  { 2444633.50, 51.446775 }, //29 January 1981, UT1-UTC=-0.2627747, Accumulated Leap Seconds=19
  { 2444634.50, 51.448809 }, //30 January 1981, UT1-UTC=-0.2648091, Accumulated Leap Seconds=19
  { 2444635.50, 51.450754 }, //31 January 1981, UT1-UTC=-0.2667537, Accumulated Leap Seconds=19
  { 2444636.50, 51.452629 }, //01 February 1981, UT1-UTC=-0.2686292, Accumulated Leap Seconds=19
  { 2444637.50, 51.454470 }, //02 February 1981, UT1-UTC=-0.2704700, Accumulated Leap Seconds=19
  { 2444638.50, 51.456325 }, //03 February 1981, UT1-UTC=-0.2723251, Accumulated Leap Seconds=19
  { 2444639.50, 51.458254 }, //04 February 1981, UT1-UTC=-0.2742544, Accumulated Leap Seconds=19
  { 2444640.50, 51.460317 }, //05 February 1981, UT1-UTC=-0.2763172, Accumulated Leap Seconds=19
  { 2444641.50, 51.462554 }, //06 February 1981, UT1-UTC=-0.2785542, Accumulated Leap Seconds=19
  { 2444642.50, 51.464970 }, //07 February 1981, UT1-UTC=-0.2809702, Accumulated Leap Seconds=19
  { 2444643.50, 51.467528 }, //08 February 1981, UT1-UTC=-0.2835277, Accumulated Leap Seconds=19
  { 2444644.50, 51.470155 }, //09 February 1981, UT1-UTC=-0.2861550, Accumulated Leap Seconds=19
  { 2444645.50, 51.472765 }, //10 February 1981, UT1-UTC=-0.2887649, Accumulated Leap Seconds=19
  { 2444646.50, 51.475278 }, //11 February 1981, UT1-UTC=-0.2912784, Accumulated Leap Seconds=19
  { 2444647.50, 51.477644 }, //12 February 1981, UT1-UTC=-0.2936444, Accumulated Leap Seconds=19
  { 2444648.50, 51.479850 }, //13 February 1981, UT1-UTC=-0.2958499, Accumulated Leap Seconds=19
  { 2444649.50, 51.481919 }, //14 February 1981, UT1-UTC=-0.2979186, Accumulated Leap Seconds=19
  { 2444650.50, 51.483903 }, //15 February 1981, UT1-UTC=-0.2999027, Accumulated Leap Seconds=19
  { 2444651.50, 51.485868 }, //16 February 1981, UT1-UTC=-0.3018679, Accumulated Leap Seconds=19
  { 2444652.50, 51.487876 }, //17 February 1981, UT1-UTC=-0.3038764, Accumulated Leap Seconds=19
  { 2444653.50, 51.489972 }, //18 February 1981, UT1-UTC=-0.3059724, Accumulated Leap Seconds=19
  { 2444654.50, 51.492175 }, //19 February 1981, UT1-UTC=-0.3081755, Accumulated Leap Seconds=19
  { 2444655.50, 51.494482 }, //20 February 1981, UT1-UTC=-0.3104817, Accumulated Leap Seconds=19
  { 2444656.50, 51.496871 }, //21 February 1981, UT1-UTC=-0.3128708, Accumulated Leap Seconds=19
  { 2444657.50, 51.499314 }, //22 February 1981, UT1-UTC=-0.3153137, Accumulated Leap Seconds=19
  { 2444658.50, 51.501779 }, //23 February 1981, UT1-UTC=-0.3177786, Accumulated Leap Seconds=19
  { 2444659.50, 51.504236 }, //24 February 1981, UT1-UTC=-0.3202357, Accumulated Leap Seconds=19
  { 2444660.50, 51.506662 }, //25 February 1981, UT1-UTC=-0.3226623, Accumulated Leap Seconds=19
  { 2444661.50, 51.509047 }, //26 February 1981, UT1-UTC=-0.3250466, Accumulated Leap Seconds=19
  { 2444662.50, 51.511391 }, //27 February 1981, UT1-UTC=-0.3273911, Accumulated Leap Seconds=19
  { 2444663.50, 51.513712 }, //28 February 1981, UT1-UTC=-0.3297123, Accumulated Leap Seconds=19
  { 2444664.50, 51.516039 }, //01 March 1981, UT1-UTC=-0.3320394, Accumulated Leap Seconds=19
  { 2444665.50, 51.518412 }, //02 March 1981, UT1-UTC=-0.3344120, Accumulated Leap Seconds=19
  { 2444666.50, 51.520881 }, //03 March 1981, UT1-UTC=-0.3368807, Accumulated Leap Seconds=19
  { 2444667.50, 51.523503 }, //04 March 1981, UT1-UTC=-0.3395035, Accumulated Leap Seconds=19
  { 2444668.50, 51.526331 }, //05 March 1981, UT1-UTC=-0.3423313, Accumulated Leap Seconds=19
  { 2444669.50, 51.529386 }, //06 March 1981, UT1-UTC=-0.3453865, Accumulated Leap Seconds=19
  { 2444670.50, 51.532646 }, //07 March 1981, UT1-UTC=-0.3486457, Accumulated Leap Seconds=19
  { 2444671.50, 51.536037 }, //08 March 1981, UT1-UTC=-0.3520370, Accumulated Leap Seconds=19
  { 2444672.50, 51.539455 }, //09 March 1981, UT1-UTC=-0.3554548, Accumulated Leap Seconds=19
  { 2444673.50, 51.542790 }, //10 March 1981, UT1-UTC=-0.3587897, Accumulated Leap Seconds=19
  { 2444674.50, 51.545959 }, //11 March 1981, UT1-UTC=-0.3619587, Accumulated Leap Seconds=19
  { 2444675.50, 51.548925 }, //12 March 1981, UT1-UTC=-0.3649249, Accumulated Leap Seconds=19
  { 2444676.50, 51.551699 }, //13 March 1981, UT1-UTC=-0.3676989, Accumulated Leap Seconds=19
  { 2444677.50, 51.554327 }, //14 March 1981, UT1-UTC=-0.3703271, Accumulated Leap Seconds=19
  { 2444678.50, 51.556873 }, //15 March 1981, UT1-UTC=-0.3728734, Accumulated Leap Seconds=19
  { 2444679.50, 51.559400 }, //16 March 1981, UT1-UTC=-0.3754002, Accumulated Leap Seconds=19
  { 2444680.50, 51.561955 }, //17 March 1981, UT1-UTC=-0.3779546, Accumulated Leap Seconds=19
  { 2444681.50, 51.564561 }, //18 March 1981, UT1-UTC=-0.3805608, Accumulated Leap Seconds=19
  { 2444682.50, 51.567221 }, //19 March 1981, UT1-UTC=-0.3832206, Accumulated Leap Seconds=19
  { 2444683.50, 51.569917 }, //20 March 1981, UT1-UTC=-0.3859169, Accumulated Leap Seconds=19
  { 2444684.50, 51.572621 }, //21 March 1981, UT1-UTC=-0.3886210, Accumulated Leap Seconds=19
  { 2444685.50, 51.575299 }, //22 March 1981, UT1-UTC=-0.3912986, Accumulated Leap Seconds=19
  { 2444686.50, 51.577916 }, //23 March 1981, UT1-UTC=-0.3939164, Accumulated Leap Seconds=19
  { 2444687.50, 51.580446 }, //24 March 1981, UT1-UTC=-0.3964461, Accumulated Leap Seconds=19
  { 2444688.50, 51.582870 }, //25 March 1981, UT1-UTC=-0.3988704, Accumulated Leap Seconds=19
  { 2444689.50, 51.585186 }, //26 March 1981, UT1-UTC=-0.4011864, Accumulated Leap Seconds=19
  { 2444690.50, 51.587408 }, //27 March 1981, UT1-UTC=-0.4034080, Accumulated Leap Seconds=19
  { 2444691.50, 51.589565 }, //28 March 1981, UT1-UTC=-0.4055650, Accumulated Leap Seconds=19
  { 2444692.50, 51.591700 }, //29 March 1981, UT1-UTC=-0.4077000, Accumulated Leap Seconds=19
  { 2444693.50, 51.593867 }, //30 March 1981, UT1-UTC=-0.4098667, Accumulated Leap Seconds=19
  { 2444694.50, 51.596128 }, //31 March 1981, UT1-UTC=-0.4121276, Accumulated Leap Seconds=19
  { 2444695.50, 51.598548 }, //01 April 1981, UT1-UTC=-0.4145479, Accumulated Leap Seconds=19
  { 2444696.50, 51.601179 }, //02 April 1981, UT1-UTC=-0.4171794, Accumulated Leap Seconds=19
  { 2444697.50, 51.604040 }, //03 April 1981, UT1-UTC=-0.4200401, Accumulated Leap Seconds=19
  { 2444698.50, 51.607097 }, //04 April 1981, UT1-UTC=-0.4230968, Accumulated Leap Seconds=19
  { 2444699.50, 51.610265 }, //05 April 1981, UT1-UTC=-0.4262651, Accumulated Leap Seconds=19
  { 2444700.50, 51.613432 }, //06 April 1981, UT1-UTC=-0.4294318, Accumulated Leap Seconds=19
  { 2444701.50, 51.616492 }, //07 April 1981, UT1-UTC=-0.4324918, Accumulated Leap Seconds=19
  { 2444702.50, 51.619381 }, //08 April 1981, UT1-UTC=-0.4353808, Accumulated Leap Seconds=19
  { 2444703.50, 51.622091 }, //09 April 1981, UT1-UTC=-0.4380908, Accumulated Leap Seconds=19
  { 2444704.50, 51.624664 }, //10 April 1981, UT1-UTC=-0.4406638, Accumulated Leap Seconds=19
  { 2444705.50, 51.627170 }, //11 April 1981, UT1-UTC=-0.4431695, Accumulated Leap Seconds=19
  { 2444706.50, 51.629680 }, //12 April 1981, UT1-UTC=-0.4456805, Accumulated Leap Seconds=19
  { 2444707.50, 51.632254 }, //13 April 1981, UT1-UTC=-0.4482538, Accumulated Leap Seconds=19
  { 2444708.50, 51.634921 }, //14 April 1981, UT1-UTC=-0.4509213, Accumulated Leap Seconds=19
  { 2444709.50, 51.637690 }, //15 April 1981, UT1-UTC=-0.4536905, Accumulated Leap Seconds=19
  { 2444710.50, 51.640548 }, //16 April 1981, UT1-UTC=-0.4565482, Accumulated Leap Seconds=19
  { 2444711.50, 51.643466 }, //17 April 1981, UT1-UTC=-0.4594661, Accumulated Leap Seconds=19
  { 2444712.50, 51.646407 }, //18 April 1981, UT1-UTC=-0.4624072, Accumulated Leap Seconds=19
  { 2444713.50, 51.649331 }, //19 April 1981, UT1-UTC=-0.4653310, Accumulated Leap Seconds=19
  { 2444714.50, 51.652198 }, //20 April 1981, UT1-UTC=-0.4681983, Accumulated Leap Seconds=19
  { 2444715.50, 51.654978 }, //21 April 1981, UT1-UTC=-0.4709777, Accumulated Leap Seconds=19
  { 2444716.50, 51.657651 }, //22 April 1981, UT1-UTC=-0.4736507, Accumulated Leap Seconds=19
  { 2444717.50, 51.660215 }, //23 April 1981, UT1-UTC=-0.4762150, Accumulated Leap Seconds=19
  { 2444718.50, 51.662687 }, //24 April 1981, UT1-UTC=-0.4786868, Accumulated Leap Seconds=19
  { 2444719.50, 51.665097 }, //25 April 1981, UT1-UTC=-0.4810974, Accumulated Leap Seconds=19
  { 2444720.50, 51.667490 }, //26 April 1981, UT1-UTC=-0.4834904, Accumulated Leap Seconds=19
  { 2444721.50, 51.669918 }, //27 April 1981, UT1-UTC=-0.4859176, Accumulated Leap Seconds=19
  { 2444722.50, 51.672434 }, //28 April 1981, UT1-UTC=-0.4884342, Accumulated Leap Seconds=19
  { 2444723.50, 51.675092 }, //29 April 1981, UT1-UTC=-0.4910916, Accumulated Leap Seconds=19
  { 2444724.50, 51.677923 }, //30 April 1981, UT1-UTC=-0.4939229, Accumulated Leap Seconds=19
  { 2444725.50, 51.680924 }, //01 May 1981, UT1-UTC=-0.4969240, Accumulated Leap Seconds=19
  { 2444726.50, 51.684042 }, //02 May 1981, UT1-UTC=-0.5000422, Accumulated Leap Seconds=19
  { 2444727.50, 51.687181 }, //03 May 1981, UT1-UTC=-0.5031810, Accumulated Leap Seconds=19
  { 2444728.50, 51.690226 }, //04 May 1981, UT1-UTC=-0.5062258, Accumulated Leap Seconds=19
  { 2444729.50, 51.693082 }, //05 May 1981, UT1-UTC=-0.5090820, Accumulated Leap Seconds=19
  { 2444730.50, 51.695706 }, //06 May 1981, UT1-UTC=-0.5117062, Accumulated Leap Seconds=19
  { 2444731.50, 51.698115 }, //07 May 1981, UT1-UTC=-0.5141149, Accumulated Leap Seconds=19
  { 2444732.50, 51.700371 }, //08 May 1981, UT1-UTC=-0.5163715, Accumulated Leap Seconds=19
  { 2444733.50, 51.702557 }, //09 May 1981, UT1-UTC=-0.5185568, Accumulated Leap Seconds=19
  { 2444734.50, 51.704743 }, //10 May 1981, UT1-UTC=-0.5207429, Accumulated Leap Seconds=19
  { 2444735.50, 51.706977 }, //11 May 1981, UT1-UTC=-0.5229772, Accumulated Leap Seconds=19
  { 2444736.50, 51.709279 }, //12 May 1981, UT1-UTC=-0.5252793, Accumulated Leap Seconds=19
  { 2444737.50, 51.711646 }, //13 May 1981, UT1-UTC=-0.5276464, Accumulated Leap Seconds=19
  { 2444738.50, 51.714060 }, //14 May 1981, UT1-UTC=-0.5300601, Accumulated Leap Seconds=19
  { 2444739.50, 51.716493 }, //15 May 1981, UT1-UTC=-0.5324934, Accumulated Leap Seconds=19
  { 2444740.50, 51.718915 }, //16 May 1981, UT1-UTC=-0.5349155, Accumulated Leap Seconds=19
  { 2444741.50, 51.721296 }, //17 May 1981, UT1-UTC=-0.5372959, Accumulated Leap Seconds=19
  { 2444742.50, 51.723609 }, //18 May 1981, UT1-UTC=-0.5396090, Accumulated Leap Seconds=19
  { 2444743.50, 51.725839 }, //19 May 1981, UT1-UTC=-0.5418392, Accumulated Leap Seconds=19
  { 2444744.50, 51.727984 }, //20 May 1981, UT1-UTC=-0.5439842, Accumulated Leap Seconds=19
  { 2444745.50, 51.730057 }, //21 May 1981, UT1-UTC=-0.5460571, Accumulated Leap Seconds=19
  { 2444746.50, 51.732087 }, //22 May 1981, UT1-UTC=-0.5480870, Accumulated Leap Seconds=19
  { 2444747.50, 51.734115 }, //23 May 1981, UT1-UTC=-0.5501150, Accumulated Leap Seconds=19
  { 2444748.50, 51.736189 }, //24 May 1981, UT1-UTC=-0.5521892, Accumulated Leap Seconds=19
  { 2444749.50, 51.738360 }, //25 May 1981, UT1-UTC=-0.5543596, Accumulated Leap Seconds=19
  { 2444750.50, 51.740672 }, //26 May 1981, UT1-UTC=-0.5566719, Accumulated Leap Seconds=19
  { 2444751.50, 51.743160 }, //27 May 1981, UT1-UTC=-0.5591598, Accumulated Leap Seconds=19
  { 2444752.50, 51.745833 }, //28 May 1981, UT1-UTC=-0.5618327, Accumulated Leap Seconds=19
  { 2444753.50, 51.748663 }, //29 May 1981, UT1-UTC=-0.5646635, Accumulated Leap Seconds=19
  { 2444754.50, 51.751583 }, //30 May 1981, UT1-UTC=-0.5675835, Accumulated Leap Seconds=19
  { 2444755.50, 51.754492 }, //31 May 1981, UT1-UTC=-0.5704924, Accumulated Leap Seconds=19
  { 2444756.50, 51.757285 }, //01 June 1981, UT1-UTC=-0.5732854, Accumulated Leap Seconds=19
  { 2444757.50, 51.759888 }, //02 June 1981, UT1-UTC=-0.5758879, Accumulated Leap Seconds=19
  { 2444758.50, 51.762279 }, //03 June 1981, UT1-UTC=-0.5782792, Accumulated Leap Seconds=19
  { 2444759.50, 51.764494 }, //04 June 1981, UT1-UTC=-0.5804940, Accumulated Leap Seconds=19
  { 2444760.50, 51.766602 }, //05 June 1981, UT1-UTC=-0.5826020, Accumulated Leap Seconds=19
  { 2444761.50, 51.768678 }, //06 June 1981, UT1-UTC=-0.5846779, Accumulated Leap Seconds=19
  { 2444762.50, 51.770776 }, //07 June 1981, UT1-UTC=-0.5867761, Accumulated Leap Seconds=19
  { 2444763.50, 51.772920 }, //08 June 1981, UT1-UTC=-0.5889200, Accumulated Leap Seconds=19
  { 2444764.50, 51.775105 }, //09 June 1981, UT1-UTC=-0.5911049, Accumulated Leap Seconds=19
  { 2444765.50, 51.777307 }, //10 June 1981, UT1-UTC=-0.5933073, Accumulated Leap Seconds=19
  { 2444766.50, 51.779495 }, //11 June 1981, UT1-UTC=-0.5954946, Accumulated Leap Seconds=19
  { 2444767.50, 51.781631 }, //12 June 1981, UT1-UTC=-0.5976311, Accumulated Leap Seconds=19
  { 2444768.50, 51.783683 }, //13 June 1981, UT1-UTC=-0.5996832, Accumulated Leap Seconds=19
  { 2444769.50, 51.785623 }, //14 June 1981, UT1-UTC=-0.6016228, Accumulated Leap Seconds=19
  { 2444770.50, 51.787431 }, //15 June 1981, UT1-UTC=-0.6034315, Accumulated Leap Seconds=19
  { 2444771.50, 51.789104 }, //16 June 1981, UT1-UTC=-0.6051043, Accumulated Leap Seconds=19
  { 2444772.50, 51.790651 }, //17 June 1981, UT1-UTC=-0.6066515, Accumulated Leap Seconds=19
  { 2444773.50, 51.792099 }, //18 June 1981, UT1-UTC=-0.6080991, Accumulated Leap Seconds=19
  { 2444774.50, 51.793487 }, //19 June 1981, UT1-UTC=-0.6094870, Accumulated Leap Seconds=19
  { 2444775.50, 51.794865 }, //20 June 1981, UT1-UTC=-0.6108651, Accumulated Leap Seconds=19
  { 2444776.50, 51.796286 }, //21 June 1981, UT1-UTC=-0.6122857, Accumulated Leap Seconds=19
  { 2444777.50, 51.797796 }, //22 June 1981, UT1-UTC=-0.6137956, Accumulated Leap Seconds=19
  { 2444778.50, 51.799427 }, //23 June 1981, UT1-UTC=-0.6154270, Accumulated Leap Seconds=19
  { 2444779.50, 51.801191 }, //24 June 1981, UT1-UTC=-0.6171909, Accumulated Leap Seconds=19
  { 2444780.50, 51.803071 }, //25 June 1981, UT1-UTC=-0.6190711, Accumulated Leap Seconds=19
  { 2444781.50, 51.805021 }, //26 June 1981, UT1-UTC=-0.6210211, Accumulated Leap Seconds=19
  { 2444782.50, 51.806967 }, //27 June 1981, UT1-UTC=-0.6229673, Accumulated Leap Seconds=19
  { 2444783.50, 51.808825 }, //28 June 1981, UT1-UTC=-0.6248250, Accumulated Leap Seconds=19
  { 2444784.50, 51.810521 }, //29 June 1981, UT1-UTC=-0.6265211, Accumulated Leap Seconds=19
  { 2444785.50, 51.812019 }, //30 June 1981, UT1-UTC=-0.6280192, Accumulated Leap Seconds=19
  { 2444786.50, 51.813333 }, //01 July 1981, UT1-UTC= 0.3706665, Accumulated Leap Seconds=20
  { 2444787.50, 51.814522 }, //02 July 1981, UT1-UTC= 0.3694777, Accumulated Leap Seconds=20
  { 2444788.50, 51.815664 }, //03 July 1981, UT1-UTC= 0.3683358, Accumulated Leap Seconds=20
  { 2444789.50, 51.816828 }, //04 July 1981, UT1-UTC= 0.3671721, Accumulated Leap Seconds=20
  { 2444790.50, 51.818053 }, //05 July 1981, UT1-UTC= 0.3659465, Accumulated Leap Seconds=20
  { 2444791.50, 51.819347 }, //06 July 1981, UT1-UTC= 0.3646531, Accumulated Leap Seconds=20
  { 2444792.50, 51.820688 }, //07 July 1981, UT1-UTC= 0.3633115, Accumulated Leap Seconds=20
  { 2444793.50, 51.822045 }, //08 July 1981, UT1-UTC= 0.3619550, Accumulated Leap Seconds=20
  { 2444794.50, 51.823379 }, //09 July 1981, UT1-UTC= 0.3606205, Accumulated Leap Seconds=20
  { 2444795.50, 51.824658 }, //10 July 1981, UT1-UTC= 0.3593416, Accumulated Leap Seconds=20
  { 2444796.50, 51.825855 }, //11 July 1981, UT1-UTC= 0.3581453, Accumulated Leap Seconds=20
  { 2444797.50, 51.826952 }, //12 July 1981, UT1-UTC= 0.3570477, Accumulated Leap Seconds=20
  { 2444798.50, 51.827948 }, //13 July 1981, UT1-UTC= 0.3560519, Accumulated Leap Seconds=20
  { 2444799.50, 51.828853 }, //14 July 1981, UT1-UTC= 0.3551467, Accumulated Leap Seconds=20
  { 2444800.50, 51.829694 }, //15 July 1981, UT1-UTC= 0.3543064, Accumulated Leap Seconds=20
  { 2444801.50, 51.830509 }, //16 July 1981, UT1-UTC= 0.3534912, Accumulated Leap Seconds=20
  { 2444802.50, 51.831350 }, //17 July 1981, UT1-UTC= 0.3526498, Accumulated Leap Seconds=20
  { 2444803.50, 51.832276 }, //18 July 1981, UT1-UTC= 0.3517243, Accumulated Leap Seconds=20
  { 2444804.50, 51.833340 }, //19 July 1981, UT1-UTC= 0.3506603, Accumulated Leap Seconds=20
  { 2444805.50, 51.834581 }, //20 July 1981, UT1-UTC= 0.3494185, Accumulated Leap Seconds=20
  { 2444806.50, 51.836015 }, //21 July 1981, UT1-UTC= 0.3479847, Accumulated Leap Seconds=20
  { 2444807.50, 51.837624 }, //22 July 1981, UT1-UTC= 0.3463761, Accumulated Leap Seconds=20
  { 2444808.50, 51.839360 }, //23 July 1981, UT1-UTC= 0.3446398, Accumulated Leap Seconds=20
  { 2444809.50, 51.841156 }, //24 July 1981, UT1-UTC= 0.3428438, Accumulated Leap Seconds=20
  { 2444810.50, 51.842936 }, //25 July 1981, UT1-UTC= 0.3410644, Accumulated Leap Seconds=20
  { 2444811.50, 51.844631 }, //26 July 1981, UT1-UTC= 0.3393691, Accumulated Leap Seconds=20
  { 2444812.50, 51.846202 }, //27 July 1981, UT1-UTC= 0.3377979, Accumulated Leap Seconds=20
  { 2444813.50, 51.847649 }, //28 July 1981, UT1-UTC= 0.3363514, Accumulated Leap Seconds=20
  { 2444814.50, 51.849010 }, //29 July 1981, UT1-UTC= 0.3349899, Accumulated Leap Seconds=20
  { 2444815.50, 51.850353 }, //30 July 1981, UT1-UTC= 0.3336469, Accumulated Leap Seconds=20
  { 2444816.50, 51.851746 }, //31 July 1981, UT1-UTC= 0.3322535, Accumulated Leap Seconds=20
  { 2444817.50, 51.853238 }, //01 August 1981, UT1-UTC= 0.3307615, Accumulated Leap Seconds=20
  { 2444818.50, 51.854844 }, //02 August 1981, UT1-UTC= 0.3291561, Accumulated Leap Seconds=20
  { 2444819.50, 51.856546 }, //03 August 1981, UT1-UTC= 0.3274544, Accumulated Leap Seconds=20
  { 2444820.50, 51.858306 }, //04 August 1981, UT1-UTC= 0.3256944, Accumulated Leap Seconds=20
  { 2444821.50, 51.860078 }, //05 August 1981, UT1-UTC= 0.3239221, Accumulated Leap Seconds=20
  { 2444822.50, 51.861818 }, //06 August 1981, UT1-UTC= 0.3221817, Accumulated Leap Seconds=20
  { 2444823.50, 51.863490 }, //07 August 1981, UT1-UTC= 0.3205100, Accumulated Leap Seconds=20
  { 2444824.50, 51.865067 }, //08 August 1981, UT1-UTC= 0.3189326, Accumulated Leap Seconds=20
  { 2444825.50, 51.866539 }, //09 August 1981, UT1-UTC= 0.3174612, Accumulated Leap Seconds=20
  { 2444826.50, 51.867908 }, //10 August 1981, UT1-UTC= 0.3160922, Accumulated Leap Seconds=20
  { 2444827.50, 51.869193 }, //11 August 1981, UT1-UTC= 0.3148068, Accumulated Leap Seconds=20
  { 2444828.50, 51.870427 }, //12 August 1981, UT1-UTC= 0.3135734, Accumulated Leap Seconds=20
  { 2444829.50, 51.871651 }, //13 August 1981, UT1-UTC= 0.3123490, Accumulated Leap Seconds=20
  { 2444830.50, 51.872918 }, //14 August 1981, UT1-UTC= 0.3110819, Accumulated Leap Seconds=20
  { 2444831.50, 51.874283 }, //15 August 1981, UT1-UTC= 0.3097165, Accumulated Leap Seconds=20
  { 2444832.50, 51.875794 }, //16 August 1981, UT1-UTC= 0.3082058, Accumulated Leap Seconds=20
  { 2444833.50, 51.877473 }, //17 August 1981, UT1-UTC= 0.3065269, Accumulated Leap Seconds=20
  { 2444834.50, 51.879307 }, //18 August 1981, UT1-UTC= 0.3046927, Accumulated Leap Seconds=20
  { 2444835.50, 51.881247 }, //19 August 1981, UT1-UTC= 0.3027534, Accumulated Leap Seconds=20
  { 2444836.50, 51.883214 }, //20 August 1981, UT1-UTC= 0.3007863, Accumulated Leap Seconds=20
  { 2444837.50, 51.885124 }, //21 August 1981, UT1-UTC= 0.2988761, Accumulated Leap Seconds=20
  { 2444838.50, 51.886907 }, //22 August 1981, UT1-UTC= 0.2970932, Accumulated Leap Seconds=20
  { 2444839.50, 51.888523 }, //23 August 1981, UT1-UTC= 0.2954768, Accumulated Leap Seconds=20
  { 2444840.50, 51.889974 }, //24 August 1981, UT1-UTC= 0.2940264, Accumulated Leap Seconds=20
  { 2444841.50, 51.891296 }, //25 August 1981, UT1-UTC= 0.2927041, Accumulated Leap Seconds=20
  { 2444842.50, 51.892555 }, //26 August 1981, UT1-UTC= 0.2914453, Accumulated Leap Seconds=20
  { 2444843.50, 51.893823 }, //27 August 1981, UT1-UTC= 0.2901771, Accumulated Leap Seconds=20
  { 2444844.50, 51.895162 }, //28 August 1981, UT1-UTC= 0.2888384, Accumulated Leap Seconds=20
  { 2444845.50, 51.896604 }, //29 August 1981, UT1-UTC= 0.2873960, Accumulated Leap Seconds=20
  { 2444846.50, 51.898151 }, //30 August 1981, UT1-UTC= 0.2858492, Accumulated Leap Seconds=20
  { 2444847.50, 51.899775 }, //31 August 1981, UT1-UTC= 0.2842247, Accumulated Leap Seconds=20
  { 2444848.50, 51.901436 }, //01 September 1981, UT1-UTC= 0.2825642, Accumulated Leap Seconds=20
  { 2444849.50, 51.903088 }, //02 September 1981, UT1-UTC= 0.2809121, Accumulated Leap Seconds=20
  { 2444850.50, 51.904693 }, //03 September 1981, UT1-UTC= 0.2793066, Accumulated Leap Seconds=20
  { 2444851.50, 51.906225 }, //04 September 1981, UT1-UTC= 0.2777746, Accumulated Leap Seconds=20
  { 2444852.50, 51.907671 }, //05 September 1981, UT1-UTC= 0.2763288, Accumulated Leap Seconds=20
  { 2444853.50, 51.909035 }, //06 September 1981, UT1-UTC= 0.2749651, Accumulated Leap Seconds=20
  { 2444854.50, 51.910337 }, //07 September 1981, UT1-UTC= 0.2736626, Accumulated Leap Seconds=20
  { 2444855.50, 51.911613 }, //08 September 1981, UT1-UTC= 0.2723866, Accumulated Leap Seconds=20
  { 2444856.50, 51.912909 }, //09 September 1981, UT1-UTC= 0.2710914, Accumulated Leap Seconds=20
  { 2444857.50, 51.914277 }, //10 September 1981, UT1-UTC= 0.2697232, Accumulated Leap Seconds=20
  { 2444858.50, 51.915777 }, //11 September 1981, UT1-UTC= 0.2682232, Accumulated Leap Seconds=20
  { 2444859.50, 51.917465 }, //12 September 1981, UT1-UTC= 0.2665346, Accumulated Leap Seconds=20
  { 2444860.50, 51.919384 }, //13 September 1981, UT1-UTC= 0.2646162, Accumulated Leap Seconds=20
  { 2444861.50, 51.921540 }, //14 September 1981, UT1-UTC= 0.2624596, Accumulated Leap Seconds=20
  { 2444862.50, 51.923899 }, //15 September 1981, UT1-UTC= 0.2601009, Accumulated Leap Seconds=20
  { 2444863.50, 51.926381 }, //16 September 1981, UT1-UTC= 0.2576186, Accumulated Leap Seconds=20
  { 2444864.50, 51.928885 }, //17 September 1981, UT1-UTC= 0.2551147, Accumulated Leap Seconds=20
  { 2444865.50, 51.931315 }, //18 September 1981, UT1-UTC= 0.2526846, Accumulated Leap Seconds=20
  { 2444866.50, 51.933610 }, //19 September 1981, UT1-UTC= 0.2503899, Accumulated Leap Seconds=20
  { 2444867.50, 51.935755 }, //20 September 1981, UT1-UTC= 0.2482445, Accumulated Leap Seconds=20
  { 2444868.50, 51.937783 }, //21 September 1981, UT1-UTC= 0.2462167, Accumulated Leap Seconds=20
  { 2444869.50, 51.939756 }, //22 September 1981, UT1-UTC= 0.2442442, Accumulated Leap Seconds=20
  { 2444870.50, 51.941745 }, //23 September 1981, UT1-UTC= 0.2422552, Accumulated Leap Seconds=20
  { 2444871.50, 51.943811 }, //24 September 1981, UT1-UTC= 0.2401886, Accumulated Leap Seconds=20
  { 2444872.50, 51.945992 }, //25 September 1981, UT1-UTC= 0.2380076, Accumulated Leap Seconds=20
  { 2444873.50, 51.948294 }, //26 September 1981, UT1-UTC= 0.2357062, Accumulated Leap Seconds=20
  { 2444874.50, 51.950693 }, //27 September 1981, UT1-UTC= 0.2333073, Accumulated Leap Seconds=20
  { 2444875.50, 51.953146 }, //28 September 1981, UT1-UTC= 0.2308535, Accumulated Leap Seconds=20
  { 2444876.50, 51.955604 }, //29 September 1981, UT1-UTC= 0.2283959, Accumulated Leap Seconds=20
  { 2444877.50, 51.958016 }, //30 September 1981, UT1-UTC= 0.2259836, Accumulated Leap Seconds=20
  { 2444878.50, 51.960343 }, //01 October 1981, UT1-UTC= 0.2236567, Accumulated Leap Seconds=20
  { 2444879.50, 51.962560 }, //02 October 1981, UT1-UTC= 0.2214404, Accumulated Leap Seconds=20
  { 2444880.50, 51.964657 }, //03 October 1981, UT1-UTC= 0.2193426, Accumulated Leap Seconds=20
  { 2444881.50, 51.966648 }, //04 October 1981, UT1-UTC= 0.2173522, Accumulated Leap Seconds=20
  { 2444882.50, 51.968561 }, //05 October 1981, UT1-UTC= 0.2154394, Accumulated Leap Seconds=20
  { 2444883.50, 51.970439 }, //06 October 1981, UT1-UTC= 0.2135611, Accumulated Leap Seconds=20
  { 2444884.50, 51.972334 }, //07 October 1981, UT1-UTC= 0.2116664, Accumulated Leap Seconds=20
  { 2444885.50, 51.974301 }, //08 October 1981, UT1-UTC= 0.2096990, Accumulated Leap Seconds=20
  { 2444886.50, 51.976399 }, //09 October 1981, UT1-UTC= 0.2076010, Accumulated Leap Seconds=20
  { 2444887.50, 51.978680 }, //10 October 1981, UT1-UTC= 0.2053203, Accumulated Leap Seconds=20
  { 2444888.50, 51.981175 }, //11 October 1981, UT1-UTC= 0.2028253, Accumulated Leap Seconds=20
  { 2444889.50, 51.983878 }, //12 October 1981, UT1-UTC= 0.2001219, Accumulated Leap Seconds=20
  { 2444890.50, 51.986735 }, //13 October 1981, UT1-UTC= 0.1972647, Accumulated Leap Seconds=20
  { 2444891.50, 51.989649 }, //14 October 1981, UT1-UTC= 0.1943507, Accumulated Leap Seconds=20
  { 2444892.50, 51.992507 }, //15 October 1981, UT1-UTC= 0.1914929, Accumulated Leap Seconds=20
  { 2444893.50, 51.995217 }, //16 October 1981, UT1-UTC= 0.1887834, Accumulated Leap Seconds=20
  { 2444894.50, 51.997735 }, //17 October 1981, UT1-UTC= 0.1862652, Accumulated Leap Seconds=20
  { 2444895.50, 52.000077 }, //18 October 1981, UT1-UTC= 0.1839228, Accumulated Leap Seconds=20
  { 2444896.50, 52.002304 }, //19 October 1981, UT1-UTC= 0.1816955, Accumulated Leap Seconds=20
  { 2444897.50, 52.004496 }, //20 October 1981, UT1-UTC= 0.1795035, Accumulated Leap Seconds=20
  { 2444898.50, 52.006726 }, //21 October 1981, UT1-UTC= 0.1772743, Accumulated Leap Seconds=20
  { 2444899.50, 52.009040 }, //22 October 1981, UT1-UTC= 0.1749598, Accumulated Leap Seconds=20
  { 2444900.50, 52.011457 }, //23 October 1981, UT1-UTC= 0.1725434, Accumulated Leap Seconds=20
  { 2444901.50, 52.013963 }, //24 October 1981, UT1-UTC= 0.1700374, Accumulated Leap Seconds=20
  { 2444902.50, 52.016525 }, //25 October 1981, UT1-UTC= 0.1674746, Accumulated Leap Seconds=20
  { 2444903.50, 52.019100 }, //26 October 1981, UT1-UTC= 0.1648997, Accumulated Leap Seconds=20
  { 2444904.50, 52.021641 }, //27 October 1981, UT1-UTC= 0.1623594, Accumulated Leap Seconds=20
  { 2444905.50, 52.024104 }, //28 October 1981, UT1-UTC= 0.1598955, Accumulated Leap Seconds=20
  { 2444906.50, 52.026461 }, //29 October 1981, UT1-UTC= 0.1575393, Accumulated Leap Seconds=20
  { 2444907.50, 52.028692 }, //30 October 1981, UT1-UTC= 0.1553075, Accumulated Leap Seconds=20
  { 2444908.50, 52.030801 }, //31 October 1981, UT1-UTC= 0.1531987, Accumulated Leap Seconds=20
  { 2444909.50, 52.032807 }, //01 November 1981, UT1-UTC= 0.1511928, Accumulated Leap Seconds=20
  { 2444910.50, 52.034747 }, //02 November 1981, UT1-UTC= 0.1492531, Accumulated Leap Seconds=20
  { 2444911.50, 52.036668 }, //03 November 1981, UT1-UTC= 0.1473323, Accumulated Leap Seconds=20
  { 2444912.50, 52.038622 }, //04 November 1981, UT1-UTC= 0.1453780, Accumulated Leap Seconds=20
  { 2444913.50, 52.040663 }, //05 November 1981, UT1-UTC= 0.1433366, Accumulated Leap Seconds=20
  { 2444914.50, 52.042843 }, //06 November 1981, UT1-UTC= 0.1411574, Accumulated Leap Seconds=20
  { 2444915.50, 52.045200 }, //07 November 1981, UT1-UTC= 0.1387998, Accumulated Leap Seconds=20
  { 2444916.50, 52.047753 }, //08 November 1981, UT1-UTC= 0.1362468, Accumulated Leap Seconds=20
  { 2444917.50, 52.050479 }, //09 November 1981, UT1-UTC= 0.1335205, Accumulated Leap Seconds=20
  { 2444918.50, 52.053310 }, //10 November 1981, UT1-UTC= 0.1306898, Accumulated Leap Seconds=20
  { 2444919.50, 52.056142 }, //11 November 1981, UT1-UTC= 0.1278584, Accumulated Leap Seconds=20
  { 2444920.50, 52.058864 }, //12 November 1981, UT1-UTC= 0.1251355, Accumulated Leap Seconds=20
  { 2444921.50, 52.061401 }, //13 November 1981, UT1-UTC= 0.1225986, Accumulated Leap Seconds=20
  { 2444922.50, 52.063733 }, //14 November 1981, UT1-UTC= 0.1202668, Accumulated Leap Seconds=20
  { 2444923.50, 52.065900 }, //15 November 1981, UT1-UTC= 0.1181004, Accumulated Leap Seconds=20
  { 2444924.50, 52.067978 }, //16 November 1981, UT1-UTC= 0.1160223, Accumulated Leap Seconds=20
  { 2444925.50, 52.070049 }, //17 November 1981, UT1-UTC= 0.1139505, Accumulated Leap Seconds=20
  { 2444926.50, 52.072175 }, //18 November 1981, UT1-UTC= 0.1118252, Accumulated Leap Seconds=20
  { 2444927.50, 52.074378 }, //19 November 1981, UT1-UTC= 0.1096218, Accumulated Leap Seconds=20
  { 2444928.50, 52.076652 }, //20 November 1981, UT1-UTC= 0.1073481, Accumulated Leap Seconds=20
  { 2444929.50, 52.078967 }, //21 November 1981, UT1-UTC= 0.1050334, Accumulated Leap Seconds=20
  { 2444930.50, 52.081284 }, //22 November 1981, UT1-UTC= 0.1027164, Accumulated Leap Seconds=20
  { 2444931.50, 52.083562 }, //23 November 1981, UT1-UTC= 0.1004375, Accumulated Leap Seconds=20
  { 2444932.50, 52.085766 }, //24 November 1981, UT1-UTC= 0.0982337, Accumulated Leap Seconds=20
  { 2444933.50, 52.087867 }, //25 November 1981, UT1-UTC= 0.0961332, Accumulated Leap Seconds=20
  { 2444934.50, 52.089848 }, //26 November 1981, UT1-UTC= 0.0941523, Accumulated Leap Seconds=20
  { 2444935.50, 52.091708 }, //27 November 1981, UT1-UTC= 0.0922918, Accumulated Leap Seconds=20
  { 2444936.50, 52.093464 }, //28 November 1981, UT1-UTC= 0.0905357, Accumulated Leap Seconds=20
  { 2444937.50, 52.095149 }, //29 November 1981, UT1-UTC= 0.0888511, Accumulated Leap Seconds=20
  { 2444938.50, 52.096808 }, //30 November 1981, UT1-UTC= 0.0871918, Accumulated Leap Seconds=20
  { 2444939.50, 52.098496 }, //01 December 1981, UT1-UTC= 0.0855043, Accumulated Leap Seconds=20
  { 2444940.50, 52.100265 }, //02 December 1981, UT1-UTC= 0.0837345, Accumulated Leap Seconds=20
  { 2444941.50, 52.102167 }, //03 December 1981, UT1-UTC= 0.0818334, Accumulated Leap Seconds=20
  { 2444942.50, 52.104239 }, //04 December 1981, UT1-UTC= 0.0797612, Accumulated Leap Seconds=20
  { 2444943.50, 52.106506 }, //05 December 1981, UT1-UTC= 0.0774939, Accumulated Leap Seconds=20
  { 2444944.50, 52.108966 }, //06 December 1981, UT1-UTC= 0.0750336, Accumulated Leap Seconds=20
  { 2444945.50, 52.111582 }, //07 December 1981, UT1-UTC= 0.0724176, Accumulated Leap Seconds=20
  { 2444946.50, 52.114279 }, //08 December 1981, UT1-UTC= 0.0697211, Accumulated Leap Seconds=20
  { 2444947.50, 52.116959 }, //09 December 1981, UT1-UTC= 0.0670413, Accumulated Leap Seconds=20
  { 2444948.50, 52.119530 }, //10 December 1981, UT1-UTC= 0.0644701, Accumulated Leap Seconds=20
  { 2444949.50, 52.121940 }, //11 December 1981, UT1-UTC= 0.0620603, Accumulated Leap Seconds=20
  { 2444950.50, 52.124195 }, //12 December 1981, UT1-UTC= 0.0598052, Accumulated Leap Seconds=20
  { 2444951.50, 52.126354 }, //13 December 1981, UT1-UTC= 0.0576465, Accumulated Leap Seconds=20
  { 2444952.50, 52.128498 }, //14 December 1981, UT1-UTC= 0.0555017, Accumulated Leap Seconds=20
  { 2444953.50, 52.130701 }, //15 December 1981, UT1-UTC= 0.0532986, Accumulated Leap Seconds=20
  { 2444954.50, 52.133000 }, //16 December 1981, UT1-UTC= 0.0509998, Accumulated Leap Seconds=20
  { 2444955.50, 52.135392 }, //17 December 1981, UT1-UTC= 0.0486079, Accumulated Leap Seconds=20
  { 2444956.50, 52.137844 }, //18 December 1981, UT1-UTC= 0.0461555, Accumulated Leap Seconds=20
  { 2444957.50, 52.140311 }, //19 December 1981, UT1-UTC= 0.0436892, Accumulated Leap Seconds=20
  { 2444958.50, 52.142744 }, //20 December 1981, UT1-UTC= 0.0412561, Accumulated Leap Seconds=20
  { 2444959.50, 52.145103 }, //21 December 1981, UT1-UTC= 0.0388965, Accumulated Leap Seconds=20
  { 2444960.50, 52.147360 }, //22 December 1981, UT1-UTC= 0.0366400, Accumulated Leap Seconds=20
  { 2444961.50, 52.149496 }, //23 December 1981, UT1-UTC= 0.0345044, Accumulated Leap Seconds=20
  { 2444962.50, 52.151506 }, //24 December 1981, UT1-UTC= 0.0324942, Accumulated Leap Seconds=20
  { 2444963.50, 52.153402 }, //25 December 1981, UT1-UTC= 0.0305981, Accumulated Leap Seconds=20
  { 2444964.50, 52.155210 }, //26 December 1981, UT1-UTC= 0.0287895, Accumulated Leap Seconds=20
  { 2444965.50, 52.156972 }, //27 December 1981, UT1-UTC= 0.0270275, Accumulated Leap Seconds=20
  { 2444966.50, 52.158738 }, //28 December 1981, UT1-UTC= 0.0252621, Accumulated Leap Seconds=20
  { 2444967.50, 52.160560 }, //29 December 1981, UT1-UTC= 0.0234396, Accumulated Leap Seconds=20
  { 2444968.50, 52.162488 }, //30 December 1981, UT1-UTC= 0.0215121, Accumulated Leap Seconds=20
  { 2444969.50, 52.164556 }, //31 December 1981, UT1-UTC= 0.0194441, Accumulated Leap Seconds=20
  { 2444970.50, 52.166783 }, //01 January 1982, UT1-UTC= 0.0172174, Accumulated Leap Seconds=20
  { 2444971.50, 52.169165 }, //02 January 1982, UT1-UTC= 0.0148352, Accumulated Leap Seconds=20
  { 2444972.50, 52.171674 }, //03 January 1982, UT1-UTC= 0.0123255, Accumulated Leap Seconds=20
  { 2444973.50, 52.174256 }, //04 January 1982, UT1-UTC= 0.0097435, Accumulated Leap Seconds=20
  { 2444974.50, 52.176834 }, //05 January 1982, UT1-UTC= 0.0071659, Accumulated Leap Seconds=20
  { 2444975.50, 52.179326 }, //06 January 1982, UT1-UTC= 0.0046742, Accumulated Leap Seconds=20
  { 2444976.50, 52.181670 }, //07 January 1982, UT1-UTC= 0.0023299, Accumulated Leap Seconds=20
  { 2444977.50, 52.183850 }, //08 January 1982, UT1-UTC= 0.0001498, Accumulated Leap Seconds=20
  { 2444978.50, 52.185902 }, //09 January 1982, UT1-UTC=-0.0019018, Accumulated Leap Seconds=20
  { 2444979.50, 52.187899 }, //10 January 1982, UT1-UTC=-0.0038991, Accumulated Leap Seconds=20
  { 2444980.50, 52.189925 }, //11 January 1982, UT1-UTC=-0.0059248, Accumulated Leap Seconds=20
  { 2444981.50, 52.192037 }, //12 January 1982, UT1-UTC=-0.0080372, Accumulated Leap Seconds=20
  { 2444982.50, 52.194252 }, //13 January 1982, UT1-UTC=-0.0102519, Accumulated Leap Seconds=20
  { 2444983.50, 52.196544 }, //14 January 1982, UT1-UTC=-0.0125440, Accumulated Leap Seconds=20
  { 2444984.50, 52.198864 }, //15 January 1982, UT1-UTC=-0.0148636, Accumulated Leap Seconds=20
  { 2444985.50, 52.201155 }, //16 January 1982, UT1-UTC=-0.0171553, Accumulated Leap Seconds=20
  { 2444986.50, 52.203371 }, //17 January 1982, UT1-UTC=-0.0193709, Accumulated Leap Seconds=20
  { 2444987.50, 52.205476 }, //18 January 1982, UT1-UTC=-0.0214756, Accumulated Leap Seconds=20
  { 2444988.50, 52.207449 }, //19 January 1982, UT1-UTC=-0.0234491, Accumulated Leap Seconds=20
  { 2444989.50, 52.209286 }, //20 January 1982, UT1-UTC=-0.0252864, Accumulated Leap Seconds=20
  { 2444990.50, 52.210997 }, //21 January 1982, UT1-UTC=-0.0269973, Accumulated Leap Seconds=20
  { 2444991.50, 52.212606 }, //22 January 1982, UT1-UTC=-0.0286065, Accumulated Leap Seconds=20
  { 2444992.50, 52.214152 }, //23 January 1982, UT1-UTC=-0.0301516, Accumulated Leap Seconds=20
  { 2444993.50, 52.215681 }, //24 January 1982, UT1-UTC=-0.0316810, Accumulated Leap Seconds=20
  { 2444994.50, 52.217250 }, //25 January 1982, UT1-UTC=-0.0332497, Accumulated Leap Seconds=20
  { 2444995.50, 52.218911 }, //26 January 1982, UT1-UTC=-0.0349115, Accumulated Leap Seconds=20
  { 2444996.50, 52.220710 }, //27 January 1982, UT1-UTC=-0.0367096, Accumulated Leap Seconds=20
  { 2444997.50, 52.222667 }, //28 January 1982, UT1-UTC=-0.0386668, Accumulated Leap Seconds=20
  { 2444998.50, 52.224780 }, //29 January 1982, UT1-UTC=-0.0407796, Accumulated Leap Seconds=20
  { 2444999.50, 52.227018 }, //30 January 1982, UT1-UTC=-0.0430177, Accumulated Leap Seconds=20
  { 2445000.50, 52.229328 }, //31 January 1982, UT1-UTC=-0.0453276, Accumulated Leap Seconds=20
  { 2445001.50, 52.231642 }, //01 February 1982, UT1-UTC=-0.0476418, Accumulated Leap Seconds=20
  { 2445002.50, 52.233892 }, //02 February 1982, UT1-UTC=-0.0498919, Accumulated Leap Seconds=20
  { 2445003.50, 52.236025 }, //03 February 1982, UT1-UTC=-0.0520251, Accumulated Leap Seconds=20
  { 2445004.50, 52.238022 }, //04 February 1982, UT1-UTC=-0.0540217, Accumulated Leap Seconds=20
  { 2445005.50, 52.239906 }, //05 February 1982, UT1-UTC=-0.0559059, Accumulated Leap Seconds=20
  { 2445006.50, 52.241742 }, //06 February 1982, UT1-UTC=-0.0577419, Accumulated Leap Seconds=20
  { 2445007.50, 52.243613 }, //07 February 1982, UT1-UTC=-0.0596135, Accumulated Leap Seconds=20
  { 2445008.50, 52.245595 }, //08 February 1982, UT1-UTC=-0.0615946, Accumulated Leap Seconds=20
  { 2445009.50, 52.247724 }, //09 February 1982, UT1-UTC=-0.0637236, Accumulated Leap Seconds=20
  { 2445010.50, 52.249994 }, //10 February 1982, UT1-UTC=-0.0659939, Accumulated Leap Seconds=20
  { 2445011.50, 52.252362 }, //11 February 1982, UT1-UTC=-0.0683618, Accumulated Leap Seconds=20
  { 2445012.50, 52.254766 }, //12 February 1982, UT1-UTC=-0.0707658, Accumulated Leap Seconds=20
  { 2445013.50, 52.257146 }, //13 February 1982, UT1-UTC=-0.0731459, Accumulated Leap Seconds=20
  { 2445014.50, 52.259455 }, //14 February 1982, UT1-UTC=-0.0754554, Accumulated Leap Seconds=20
  { 2445015.50, 52.261665 }, //15 February 1982, UT1-UTC=-0.0776655, Accumulated Leap Seconds=20
  { 2445016.50, 52.263765 }, //16 February 1982, UT1-UTC=-0.0797650, Accumulated Leap Seconds=20
  { 2445017.50, 52.265761 }, //17 February 1982, UT1-UTC=-0.0817608, Accumulated Leap Seconds=20
  { 2445018.50, 52.267676 }, //18 February 1982, UT1-UTC=-0.0836763, Accumulated Leap Seconds=20
  { 2445019.50, 52.269549 }, //19 February 1982, UT1-UTC=-0.0855492, Accumulated Leap Seconds=20
  { 2445020.50, 52.271428 }, //20 February 1982, UT1-UTC=-0.0874278, Accumulated Leap Seconds=20
  { 2445021.50, 52.273367 }, //21 February 1982, UT1-UTC=-0.0893667, Accumulated Leap Seconds=20
  { 2445022.50, 52.275421 }, //22 February 1982, UT1-UTC=-0.0914207, Accumulated Leap Seconds=20
  { 2445023.50, 52.277637 }, //23 February 1982, UT1-UTC=-0.0936370, Accumulated Leap Seconds=20
  { 2445024.50, 52.280044 }, //24 February 1982, UT1-UTC=-0.0960440, Accumulated Leap Seconds=20
  { 2445025.50, 52.282641 }, //25 February 1982, UT1-UTC=-0.0986412, Accumulated Leap Seconds=20
  { 2445026.50, 52.285393 }, //26 February 1982, UT1-UTC=-0.1013933, Accumulated Leap Seconds=20
  { 2445027.50, 52.288234 }, //27 February 1982, UT1-UTC=-0.1042338, Accumulated Leap Seconds=20
  { 2445028.50, 52.291077 }, //28 February 1982, UT1-UTC=-0.1070769, Accumulated Leap Seconds=20
  { 2445029.50, 52.293838 }, //01 March 1982, UT1-UTC=-0.1098376, Accumulated Leap Seconds=20
  { 2445030.50, 52.296453 }, //02 March 1982, UT1-UTC=-0.1124530, Accumulated Leap Seconds=20
  { 2445031.50, 52.298897 }, //03 March 1982, UT1-UTC=-0.1148974, Accumulated Leap Seconds=20
  { 2445032.50, 52.301187 }, //04 March 1982, UT1-UTC=-0.1171872, Accumulated Leap Seconds=20
  { 2445033.50, 52.303376 }, //05 March 1982, UT1-UTC=-0.1193761, Accumulated Leap Seconds=20
  { 2445034.50, 52.305541 }, //06 March 1982, UT1-UTC=-0.1215411, Accumulated Leap Seconds=20
  { 2445035.50, 52.307761 }, //07 March 1982, UT1-UTC=-0.1237611, Accumulated Leap Seconds=20
  { 2445036.50, 52.310091 }, //08 March 1982, UT1-UTC=-0.1260908, Accumulated Leap Seconds=20
  { 2445037.50, 52.312545 }, //09 March 1982, UT1-UTC=-0.1285454, Accumulated Leap Seconds=20
  { 2445038.50, 52.315098 }, //10 March 1982, UT1-UTC=-0.1310981, Accumulated Leap Seconds=20
  { 2445039.50, 52.317693 }, //11 March 1982, UT1-UTC=-0.1336935, Accumulated Leap Seconds=20
  { 2445040.50, 52.320266 }, //12 March 1982, UT1-UTC=-0.1362662, Accumulated Leap Seconds=20
  { 2445041.50, 52.322759 }, //13 March 1982, UT1-UTC=-0.1387587, Accumulated Leap Seconds=20
  { 2445042.50, 52.325131 }, //14 March 1982, UT1-UTC=-0.1411307, Accumulated Leap Seconds=20
  { 2445043.50, 52.327363 }, //15 March 1982, UT1-UTC=-0.1433630, Accumulated Leap Seconds=20
  { 2445044.50, 52.329458 }, //16 March 1982, UT1-UTC=-0.1454582, Accumulated Leap Seconds=20
  { 2445045.50, 52.331439 }, //17 March 1982, UT1-UTC=-0.1474387, Accumulated Leap Seconds=20
  { 2445046.50, 52.333344 }, //18 March 1982, UT1-UTC=-0.1493436, Accumulated Leap Seconds=20
  { 2445047.50, 52.335224 }, //19 March 1982, UT1-UTC=-0.1512237, Accumulated Leap Seconds=20
  { 2445048.50, 52.337135 }, //20 March 1982, UT1-UTC=-0.1531353, Accumulated Leap Seconds=20
  { 2445049.50, 52.339136 }, //21 March 1982, UT1-UTC=-0.1551361, Accumulated Leap Seconds=20
  { 2445050.50, 52.341281 }, //22 March 1982, UT1-UTC=-0.1572808, Accumulated Leap Seconds=20
  { 2445051.50, 52.343614 }, //23 March 1982, UT1-UTC=-0.1596140, Accumulated Leap Seconds=20
  { 2445052.50, 52.346158 }, //24 March 1982, UT1-UTC=-0.1621584, Accumulated Leap Seconds=20
  { 2445053.50, 52.348902 }, //25 March 1982, UT1-UTC=-0.1649019, Accumulated Leap Seconds=20
  { 2445054.50, 52.351791 }, //26 March 1982, UT1-UTC=-0.1677915, Accumulated Leap Seconds=20
  { 2445055.50, 52.354741 }, //27 March 1982, UT1-UTC=-0.1707415, Accumulated Leap Seconds=20
  { 2445056.50, 52.357655 }, //28 March 1982, UT1-UTC=-0.1736550, Accumulated Leap Seconds=20
  { 2445057.50, 52.360453 }, //29 March 1982, UT1-UTC=-0.1764527, Accumulated Leap Seconds=20
  { 2445058.50, 52.363095 }, //30 March 1982, UT1-UTC=-0.1790950, Accumulated Leap Seconds=20
  { 2445059.50, 52.365590 }, //31 March 1982, UT1-UTC=-0.1815903, Accumulated Leap Seconds=20
  { 2445060.50, 52.367990 }, //01 April 1982, UT1-UTC=-0.1839897, Accumulated Leap Seconds=20
  { 2445061.50, 52.370369 }, //02 April 1982, UT1-UTC=-0.1863689, Accumulated Leap Seconds=20
  { 2445062.50, 52.372806 }, //03 April 1982, UT1-UTC=-0.1888058, Accumulated Leap Seconds=20
  { 2445063.50, 52.375359 }, //04 April 1982, UT1-UTC=-0.1913588, Accumulated Leap Seconds=20
  { 2445064.50, 52.378053 }, //05 April 1982, UT1-UTC=-0.1940531, Accumulated Leap Seconds=20
  { 2445065.50, 52.380875 }, //06 April 1982, UT1-UTC=-0.1968752, Accumulated Leap Seconds=20
  { 2445066.50, 52.383779 }, //07 April 1982, UT1-UTC=-0.1997791, Accumulated Leap Seconds=20
  { 2445067.50, 52.386700 }, //08 April 1982, UT1-UTC=-0.2027005, Accumulated Leap Seconds=20
  { 2445068.50, 52.389574 }, //09 April 1982, UT1-UTC=-0.2055739, Accumulated Leap Seconds=20
  { 2445069.50, 52.392345 }, //10 April 1982, UT1-UTC=-0.2083455, Accumulated Leap Seconds=20
  { 2445070.50, 52.394981 }, //11 April 1982, UT1-UTC=-0.2109809, Accumulated Leap Seconds=20
  { 2445071.50, 52.397467 }, //12 April 1982, UT1-UTC=-0.2134670, Accumulated Leap Seconds=20
  { 2445072.50, 52.399812 }, //13 April 1982, UT1-UTC=-0.2158123, Accumulated Leap Seconds=20
  { 2445073.50, 52.402046 }, //14 April 1982, UT1-UTC=-0.2180457, Accumulated Leap Seconds=20
  { 2445074.50, 52.404213 }, //15 April 1982, UT1-UTC=-0.2202126, Accumulated Leap Seconds=20
  { 2445075.50, 52.406368 }, //16 April 1982, UT1-UTC=-0.2223681, Accumulated Leap Seconds=20
  { 2445076.50, 52.408570 }, //17 April 1982, UT1-UTC=-0.2245697, Accumulated Leap Seconds=20
  { 2445077.50, 52.410874 }, //18 April 1982, UT1-UTC=-0.2268738, Accumulated Leap Seconds=20
  { 2445078.50, 52.413331 }, //19 April 1982, UT1-UTC=-0.2293305, Accumulated Leap Seconds=20
  { 2445079.50, 52.415977 }, //20 April 1982, UT1-UTC=-0.2319768, Accumulated Leap Seconds=20
  { 2445080.50, 52.418824 }, //21 April 1982, UT1-UTC=-0.2348241, Accumulated Leap Seconds=20
  { 2445081.50, 52.421845 }, //22 April 1982, UT1-UTC=-0.2378448, Accumulated Leap Seconds=20
  { 2445082.50, 52.424969 }, //23 April 1982, UT1-UTC=-0.2409687, Accumulated Leap Seconds=20
  { 2445083.50, 52.428095 }, //24 April 1982, UT1-UTC=-0.2440945, Accumulated Leap Seconds=20
  { 2445084.50, 52.431119 }, //25 April 1982, UT1-UTC=-0.2471189, Accumulated Leap Seconds=20
  { 2445085.50, 52.433970 }, //26 April 1982, UT1-UTC=-0.2499699, Accumulated Leap Seconds=20
  { 2445086.50, 52.436630 }, //27 April 1982, UT1-UTC=-0.2526301, Accumulated Leap Seconds=20
  { 2445087.50, 52.439138 }, //28 April 1982, UT1-UTC=-0.2551382, Accumulated Leap Seconds=20
  { 2445088.50, 52.441570 }, //29 April 1982, UT1-UTC=-0.2575703, Accumulated Leap Seconds=20
  { 2445089.50, 52.444010 }, //30 April 1982, UT1-UTC=-0.2600099, Accumulated Leap Seconds=20
  { 2445090.50, 52.446522 }, //01 May 1982, UT1-UTC=-0.2625221, Accumulated Leap Seconds=20
  { 2445091.50, 52.449138 }, //02 May 1982, UT1-UTC=-0.2651383, Accumulated Leap Seconds=20
  { 2445092.50, 52.451853 }, //03 May 1982, UT1-UTC=-0.2678529, Accumulated Leap Seconds=20
  { 2445093.50, 52.454629 }, //04 May 1982, UT1-UTC=-0.2706294, Accumulated Leap Seconds=20
  { 2445094.50, 52.457412 }, //05 May 1982, UT1-UTC=-0.2734116, Accumulated Leap Seconds=20
  { 2445095.50, 52.460137 }, //06 May 1982, UT1-UTC=-0.2761374, Accumulated Leap Seconds=20
  { 2445096.50, 52.462751 }, //07 May 1982, UT1-UTC=-0.2787506, Accumulated Leap Seconds=20
  { 2445097.50, 52.465209 }, //08 May 1982, UT1-UTC=-0.2812091, Accumulated Leap Seconds=20
  { 2445098.50, 52.467492 }, //09 May 1982, UT1-UTC=-0.2834916, Accumulated Leap Seconds=20
  { 2445099.50, 52.469598 }, //10 May 1982, UT1-UTC=-0.2855978, Accumulated Leap Seconds=20
  { 2445100.50, 52.471548 }, //11 May 1982, UT1-UTC=-0.2875484, Accumulated Leap Seconds=20
  { 2445101.50, 52.473383 }, //12 May 1982, UT1-UTC=-0.2893827, Accumulated Leap Seconds=20
  { 2445102.50, 52.475155 }, //13 May 1982, UT1-UTC=-0.2911547, Accumulated Leap Seconds=20
  { 2445103.50, 52.476923 }, //14 May 1982, UT1-UTC=-0.2929234, Accumulated Leap Seconds=20
  { 2445104.50, 52.478747 }, //15 May 1982, UT1-UTC=-0.2947468, Accumulated Leap Seconds=20
  { 2445105.50, 52.480677 }, //16 May 1982, UT1-UTC=-0.2966775, Accumulated Leap Seconds=20
  { 2445106.50, 52.482758 }, //17 May 1982, UT1-UTC=-0.2987583, Accumulated Leap Seconds=20
  { 2445107.50, 52.485016 }, //18 May 1982, UT1-UTC=-0.3010159, Accumulated Leap Seconds=20
  { 2445108.50, 52.487450 }, //19 May 1982, UT1-UTC=-0.3034504, Accumulated Leap Seconds=20
  { 2445109.50, 52.490024 }, //20 May 1982, UT1-UTC=-0.3060244, Accumulated Leap Seconds=20
  { 2445110.50, 52.492661 }, //21 May 1982, UT1-UTC=-0.3086614, Accumulated Leap Seconds=20
  { 2445111.50, 52.495262 }, //22 May 1982, UT1-UTC=-0.3112621, Accumulated Leap Seconds=20
  { 2445112.50, 52.497737 }, //23 May 1982, UT1-UTC=-0.3137372, Accumulated Leap Seconds=20
  { 2445113.50, 52.500040 }, //24 May 1982, UT1-UTC=-0.3160402, Accumulated Leap Seconds=20
  { 2445114.50, 52.502184 }, //25 May 1982, UT1-UTC=-0.3181844, Accumulated Leap Seconds=20
  { 2445115.50, 52.504236 }, //26 May 1982, UT1-UTC=-0.3202362, Accumulated Leap Seconds=20
  { 2445116.50, 52.506286 }, //27 May 1982, UT1-UTC=-0.3222861, Accumulated Leap Seconds=20
  { 2445117.50, 52.508413 }, //28 May 1982, UT1-UTC=-0.3244133, Accumulated Leap Seconds=20
  { 2445118.50, 52.510661 }, //29 May 1982, UT1-UTC=-0.3266607, Accumulated Leap Seconds=20
  { 2445119.50, 52.513027 }, //30 May 1982, UT1-UTC=-0.3290274, Accumulated Leap Seconds=20
  { 2445120.50, 52.515477 }, //31 May 1982, UT1-UTC=-0.3314775, Accumulated Leap Seconds=20
  { 2445121.50, 52.517955 }, //01 June 1982, UT1-UTC=-0.3339552, Accumulated Leap Seconds=20
  { 2445122.50, 52.520401 }, //02 June 1982, UT1-UTC=-0.3364009, Accumulated Leap Seconds=20
  { 2445123.50, 52.522760 }, //03 June 1982, UT1-UTC=-0.3387603, Accumulated Leap Seconds=20
  { 2445124.50, 52.524992 }, //04 June 1982, UT1-UTC=-0.3409918, Accumulated Leap Seconds=20
  { 2445125.50, 52.527070 }, //05 June 1982, UT1-UTC=-0.3430701, Accumulated Leap Seconds=20
  { 2445126.50, 52.528988 }, //06 June 1982, UT1-UTC=-0.3449881, Accumulated Leap Seconds=20
  { 2445127.50, 52.530758 }, //07 June 1982, UT1-UTC=-0.3467581, Accumulated Leap Seconds=20
  { 2445128.50, 52.532411 }, //08 June 1982, UT1-UTC=-0.3484110, Accumulated Leap Seconds=20
  { 2445129.50, 52.533992 }, //09 June 1982, UT1-UTC=-0.3499924, Accumulated Leap Seconds=20
  { 2445130.50, 52.535556 }, //10 June 1982, UT1-UTC=-0.3515560, Accumulated Leap Seconds=20
  { 2445131.50, 52.537156 }, //11 June 1982, UT1-UTC=-0.3531558, Accumulated Leap Seconds=20
  { 2445132.50, 52.538839 }, //12 June 1982, UT1-UTC=-0.3548392, Accumulated Leap Seconds=20
  { 2445133.50, 52.540642 }, //13 June 1982, UT1-UTC=-0.3566418, Accumulated Leap Seconds=20
  { 2445134.50, 52.542585 }, //14 June 1982, UT1-UTC=-0.3585851, Accumulated Leap Seconds=20
  { 2445135.50, 52.544672 }, //15 June 1982, UT1-UTC=-0.3606721, Accumulated Leap Seconds=20
  { 2445136.50, 52.546881 }, //16 June 1982, UT1-UTC=-0.3628807, Accumulated Leap Seconds=20
  { 2445137.50, 52.549157 }, //17 June 1982, UT1-UTC=-0.3651570, Accumulated Leap Seconds=20
  { 2445138.50, 52.551422 }, //18 June 1982, UT1-UTC=-0.3674218, Accumulated Leap Seconds=20
  { 2445139.50, 52.553597 }, //19 June 1982, UT1-UTC=-0.3695967, Accumulated Leap Seconds=20
  { 2445140.50, 52.555622 }, //20 June 1982, UT1-UTC=-0.3716219, Accumulated Leap Seconds=20
  { 2445141.50, 52.557470 }, //21 June 1982, UT1-UTC=-0.3734702, Accumulated Leap Seconds=20
  { 2445142.50, 52.559167 }, //22 June 1982, UT1-UTC=-0.3751668, Accumulated Leap Seconds=20
  { 2445143.50, 52.560789 }, //23 June 1982, UT1-UTC=-0.3767885, Accumulated Leap Seconds=20
  { 2445144.50, 52.562429 }, //24 June 1982, UT1-UTC=-0.3784291, Accumulated Leap Seconds=20
  { 2445145.50, 52.564156 }, //25 June 1982, UT1-UTC=-0.3801559, Accumulated Leap Seconds=20
  { 2445146.50, 52.565987 }, //26 June 1982, UT1-UTC=-0.3819872, Accumulated Leap Seconds=20
  { 2445147.50, 52.567895 }, //27 June 1982, UT1-UTC=-0.3838948, Accumulated Leap Seconds=20
  { 2445148.50, 52.569822 }, //28 June 1982, UT1-UTC=-0.3858222, Accumulated Leap Seconds=20
  { 2445149.50, 52.571707 }, //29 June 1982, UT1-UTC=-0.3877067, Accumulated Leap Seconds=20
  { 2445150.50, 52.573494 }, //30 June 1982, UT1-UTC=-0.3894945, Accumulated Leap Seconds=20
  { 2445151.50, 52.575148 }, //01 July 1982, UT1-UTC= 0.6088515, Accumulated Leap Seconds=21
  { 2445152.50, 52.576648 }, //02 July 1982, UT1-UTC= 0.6073515, Accumulated Leap Seconds=21
  { 2445153.50, 52.577991 }, //03 July 1982, UT1-UTC= 0.6060085, Accumulated Leap Seconds=21
  { 2445154.50, 52.579191 }, //04 July 1982, UT1-UTC= 0.6048090, Accumulated Leap Seconds=21
  { 2445155.50, 52.580276 }, //05 July 1982, UT1-UTC= 0.6037236, Accumulated Leap Seconds=21
  { 2445156.50, 52.581291 }, //06 July 1982, UT1-UTC= 0.6027089, Accumulated Leap Seconds=21
  { 2445157.50, 52.582288 }, //07 July 1982, UT1-UTC= 0.6017119, Accumulated Leap Seconds=21
  { 2445158.50, 52.583323 }, //08 July 1982, UT1-UTC= 0.6006769, Accumulated Leap Seconds=21
  { 2445159.50, 52.584446 }, //09 July 1982, UT1-UTC= 0.5995535, Accumulated Leap Seconds=21
  { 2445160.50, 52.585696 }, //10 July 1982, UT1-UTC= 0.5983043, Accumulated Leap Seconds=21
  { 2445161.50, 52.587091 }, //11 July 1982, UT1-UTC= 0.5969092, Accumulated Leap Seconds=21
  { 2445162.50, 52.588632 }, //12 July 1982, UT1-UTC= 0.5953676, Accumulated Leap Seconds=21
  { 2445163.50, 52.590301 }, //13 July 1982, UT1-UTC= 0.5936992, Accumulated Leap Seconds=21
  { 2445164.50, 52.592054 }, //14 July 1982, UT1-UTC= 0.5919461, Accumulated Leap Seconds=21
  { 2445165.50, 52.593829 }, //15 July 1982, UT1-UTC= 0.5901712, Accumulated Leap Seconds=21
  { 2445166.50, 52.595550 }, //16 July 1982, UT1-UTC= 0.5884504, Accumulated Leap Seconds=21
  { 2445167.50, 52.597147 }, //17 July 1982, UT1-UTC= 0.5868530, Accumulated Leap Seconds=21
  { 2445168.50, 52.598584 }, //18 July 1982, UT1-UTC= 0.5854160, Accumulated Leap Seconds=21
  { 2445169.50, 52.599876 }, //19 July 1982, UT1-UTC= 0.5841239, Accumulated Leap Seconds=21
  { 2445170.50, 52.601091 }, //20 July 1982, UT1-UTC= 0.5829089, Accumulated Leap Seconds=21
  { 2445171.50, 52.602325 }, //21 July 1982, UT1-UTC= 0.5816746, Accumulated Leap Seconds=21
  { 2445172.50, 52.603665 }, //22 July 1982, UT1-UTC= 0.5803349, Accumulated Leap Seconds=21
  { 2445173.50, 52.605153 }, //23 July 1982, UT1-UTC= 0.5788473, Accumulated Leap Seconds=21
  { 2445174.50, 52.606775 }, //24 July 1982, UT1-UTC= 0.5772252, Accumulated Leap Seconds=21
  { 2445175.50, 52.608473 }, //25 July 1982, UT1-UTC= 0.5755268, Accumulated Leap Seconds=21
  { 2445176.50, 52.610170 }, //26 July 1982, UT1-UTC= 0.5738299, Accumulated Leap Seconds=21
  { 2445177.50, 52.611793 }, //27 July 1982, UT1-UTC= 0.5722068, Accumulated Leap Seconds=21
  { 2445178.50, 52.613291 }, //28 July 1982, UT1-UTC= 0.5707089, Accumulated Leap Seconds=21
  { 2445179.50, 52.614637 }, //29 July 1982, UT1-UTC= 0.5693632, Accumulated Leap Seconds=21
  { 2445180.50, 52.615825 }, //30 July 1982, UT1-UTC= 0.5681748, Accumulated Leap Seconds=21
  { 2445181.50, 52.616870 }, //31 July 1982, UT1-UTC= 0.5671299, Accumulated Leap Seconds=21
  { 2445182.50, 52.617801 }, //01 August 1982, UT1-UTC= 0.5661988, Accumulated Leap Seconds=21
  { 2445183.50, 52.618661 }, //02 August 1982, UT1-UTC= 0.5653387, Accumulated Leap Seconds=21
  { 2445184.50, 52.619502 }, //03 August 1982, UT1-UTC= 0.5644976, Accumulated Leap Seconds=21
  { 2445185.50, 52.620381 }, //04 August 1982, UT1-UTC= 0.5636194, Accumulated Leap Seconds=21
  { 2445186.50, 52.621351 }, //05 August 1982, UT1-UTC= 0.5626487, Accumulated Leap Seconds=21
  { 2445187.50, 52.622460 }, //06 August 1982, UT1-UTC= 0.5615404, Accumulated Leap Seconds=21
  { 2445188.50, 52.623733 }, //07 August 1982, UT1-UTC= 0.5602669, Accumulated Leap Seconds=21
  { 2445189.50, 52.625177 }, //08 August 1982, UT1-UTC= 0.5588232, Accumulated Leap Seconds=21
  { 2445190.50, 52.626771 }, //09 August 1982, UT1-UTC= 0.5572292, Accumulated Leap Seconds=21
  { 2445191.50, 52.628473 }, //10 August 1982, UT1-UTC= 0.5555270, Accumulated Leap Seconds=21
  { 2445192.50, 52.630225 }, //11 August 1982, UT1-UTC= 0.5537746, Accumulated Leap Seconds=21
  { 2445193.50, 52.631963 }, //12 August 1982, UT1-UTC= 0.5520367, Accumulated Leap Seconds=21
  { 2445194.50, 52.633627 }, //13 August 1982, UT1-UTC= 0.5503727, Accumulated Leap Seconds=21
  { 2445195.50, 52.635181 }, //14 August 1982, UT1-UTC= 0.5488185, Accumulated Leap Seconds=21
  { 2445196.50, 52.636629 }, //15 August 1982, UT1-UTC= 0.5473707, Accumulated Leap Seconds=21
  { 2445197.50, 52.638021 }, //16 August 1982, UT1-UTC= 0.5459791, Accumulated Leap Seconds=21
  { 2445198.50, 52.639442 }, //17 August 1982, UT1-UTC= 0.5445579, Accumulated Leap Seconds=21
  { 2445199.50, 52.640987 }, //18 August 1982, UT1-UTC= 0.5430131, Accumulated Leap Seconds=21
  { 2445200.50, 52.642721 }, //19 August 1982, UT1-UTC= 0.5412787, Accumulated Leap Seconds=21
  { 2445201.50, 52.644659 }, //20 August 1982, UT1-UTC= 0.5393408, Accumulated Leap Seconds=21
  { 2445202.50, 52.646759 }, //21 August 1982, UT1-UTC= 0.5372407, Accumulated Leap Seconds=21
  { 2445203.50, 52.648942 }, //22 August 1982, UT1-UTC= 0.5350577, Accumulated Leap Seconds=21
  { 2445204.50, 52.651120 }, //23 August 1982, UT1-UTC= 0.5328798, Accumulated Leap Seconds=21
  { 2445205.50, 52.653221 }, //24 August 1982, UT1-UTC= 0.5307793, Accumulated Leap Seconds=21
  { 2445206.50, 52.655200 }, //25 August 1982, UT1-UTC= 0.5288002, Accumulated Leap Seconds=21
  { 2445207.50, 52.657042 }, //26 August 1982, UT1-UTC= 0.5269575, Accumulated Leap Seconds=21
  { 2445208.50, 52.658758 }, //27 August 1982, UT1-UTC= 0.5252420, Accumulated Leap Seconds=21
  { 2445209.50, 52.660375 }, //28 August 1982, UT1-UTC= 0.5236252, Accumulated Leap Seconds=21
  { 2445210.50, 52.661936 }, //29 August 1982, UT1-UTC= 0.5220644, Accumulated Leap Seconds=21
  { 2445211.50, 52.663492 }, //30 August 1982, UT1-UTC= 0.5205077, Accumulated Leap Seconds=21
  { 2445212.50, 52.665101 }, //31 August 1982, UT1-UTC= 0.5188991, Accumulated Leap Seconds=21
  { 2445213.50, 52.666816 }, //01 September 1982, UT1-UTC= 0.5171840, Accumulated Leap Seconds=21
  { 2445214.50, 52.668685 }, //02 September 1982, UT1-UTC= 0.5153154, Accumulated Leap Seconds=21
  { 2445215.50, 52.670739 }, //03 September 1982, UT1-UTC= 0.5132613, Accumulated Leap Seconds=21
  { 2445216.50, 52.672988 }, //04 September 1982, UT1-UTC= 0.5110124, Accumulated Leap Seconds=21
  { 2445217.50, 52.675411 }, //05 September 1982, UT1-UTC= 0.5085886, Accumulated Leap Seconds=21
  { 2445218.50, 52.677960 }, //06 September 1982, UT1-UTC= 0.5060400, Accumulated Leap Seconds=21
  { 2445219.50, 52.680561 }, //07 September 1982, UT1-UTC= 0.5034385, Accumulated Leap Seconds=21
  { 2445220.50, 52.683137 }, //08 September 1982, UT1-UTC= 0.5008627, Accumulated Leap Seconds=21
  { 2445221.50, 52.685619 }, //09 September 1982, UT1-UTC= 0.4983814, Accumulated Leap Seconds=21
  { 2445222.50, 52.687962 }, //10 September 1982, UT1-UTC= 0.4960381, Accumulated Leap Seconds=21
  { 2445223.50, 52.690162 }, //11 September 1982, UT1-UTC= 0.4938382, Accumulated Leap Seconds=21
  { 2445224.50, 52.692253 }, //12 September 1982, UT1-UTC= 0.4917474, Accumulated Leap Seconds=21
  { 2445225.50, 52.694302 }, //13 September 1982, UT1-UTC= 0.4896977, Accumulated Leap Seconds=21
  { 2445226.50, 52.696393 }, //14 September 1982, UT1-UTC= 0.4876067, Accumulated Leap Seconds=21
  { 2445227.50, 52.698597 }, //15 September 1982, UT1-UTC= 0.4854026, Accumulated Leap Seconds=21
  { 2445228.50, 52.700959 }, //16 September 1982, UT1-UTC= 0.4830409, Accumulated Leap Seconds=21
  { 2445229.50, 52.703473 }, //17 September 1982, UT1-UTC= 0.4805269, Accumulated Leap Seconds=21
  { 2445230.50, 52.706084 }, //18 September 1982, UT1-UTC= 0.4779157, Accumulated Leap Seconds=21
  { 2445231.50, 52.708710 }, //19 September 1982, UT1-UTC= 0.4752903, Accumulated Leap Seconds=21
  { 2445232.50, 52.711267 }, //20 September 1982, UT1-UTC= 0.4727328, Accumulated Leap Seconds=21
  { 2445233.50, 52.713694 }, //21 September 1982, UT1-UTC= 0.4703058, Accumulated Leap Seconds=21
  { 2445234.50, 52.715959 }, //22 September 1982, UT1-UTC= 0.4680406, Accumulated Leap Seconds=21
  { 2445235.50, 52.718062 }, //23 September 1982, UT1-UTC= 0.4659380, Accumulated Leap Seconds=21
  { 2445236.50, 52.720025 }, //24 September 1982, UT1-UTC= 0.4639747, Accumulated Leap Seconds=21
  { 2445237.50, 52.721891 }, //25 September 1982, UT1-UTC= 0.4621090, Accumulated Leap Seconds=21
  { 2445238.50, 52.723712 }, //26 September 1982, UT1-UTC= 0.4602881, Accumulated Leap Seconds=21
  { 2445239.50, 52.725545 }, //27 September 1982, UT1-UTC= 0.4584548, Accumulated Leap Seconds=21
  { 2445240.50, 52.727448 }, //28 September 1982, UT1-UTC= 0.4565524, Accumulated Leap Seconds=21
  { 2445241.50, 52.729469 }, //29 September 1982, UT1-UTC= 0.4545311, Accumulated Leap Seconds=21
  { 2445242.50, 52.731647 }, //30 September 1982, UT1-UTC= 0.4523526, Accumulated Leap Seconds=21
  { 2445243.50, 52.734004 }, //01 October 1982, UT1-UTC= 0.4499964, Accumulated Leap Seconds=21
  { 2445244.50, 52.736532 }, //02 October 1982, UT1-UTC= 0.4474676, Accumulated Leap Seconds=21
  { 2445245.50, 52.739197 }, //03 October 1982, UT1-UTC= 0.4448034, Accumulated Leap Seconds=21
  { 2445246.50, 52.741927 }, //04 October 1982, UT1-UTC= 0.4420726, Accumulated Leap Seconds=21
  { 2445247.50, 52.744638 }, //05 October 1982, UT1-UTC= 0.4393619, Accumulated Leap Seconds=21
  { 2445248.50, 52.747245 }, //06 October 1982, UT1-UTC= 0.4367550, Accumulated Leap Seconds=21
  { 2445249.50, 52.749692 }, //07 October 1982, UT1-UTC= 0.4343079, Accumulated Leap Seconds=21
  { 2445250.50, 52.751968 }, //08 October 1982, UT1-UTC= 0.4320322, Accumulated Leap Seconds=21
  { 2445251.50, 52.754108 }, //09 October 1982, UT1-UTC= 0.4298920, Accumulated Leap Seconds=21
  { 2445252.50, 52.756185 }, //10 October 1982, UT1-UTC= 0.4278149, Accumulated Leap Seconds=21
  { 2445253.50, 52.758287 }, //11 October 1982, UT1-UTC= 0.4257127, Accumulated Leap Seconds=21
  { 2445254.50, 52.760495 }, //12 October 1982, UT1-UTC= 0.4235055, Accumulated Leap Seconds=21
  { 2445255.50, 52.762856 }, //13 October 1982, UT1-UTC= 0.4211443, Accumulated Leap Seconds=21
  { 2445256.50, 52.765376 }, //14 October 1982, UT1-UTC= 0.4186240, Accumulated Leap Seconds=21
  { 2445257.50, 52.768015 }, //15 October 1982, UT1-UTC= 0.4159852, Accumulated Leap Seconds=21
  { 2445258.50, 52.770698 }, //16 October 1982, UT1-UTC= 0.4133022, Accumulated Leap Seconds=21
  { 2445259.50, 52.773339 }, //17 October 1982, UT1-UTC= 0.4106613, Accumulated Leap Seconds=21
  { 2445260.50, 52.775861 }, //18 October 1982, UT1-UTC= 0.4081392, Accumulated Leap Seconds=21
  { 2445261.50, 52.778212 }, //19 October 1982, UT1-UTC= 0.4057877, Accumulated Leap Seconds=21
  { 2445262.50, 52.780373 }, //20 October 1982, UT1-UTC= 0.4036267, Accumulated Leap Seconds=21
  { 2445263.50, 52.782356 }, //21 October 1982, UT1-UTC= 0.4016441, Accumulated Leap Seconds=21
  { 2445264.50, 52.784197 }, //22 October 1982, UT1-UTC= 0.3998034, Accumulated Leap Seconds=21
  { 2445265.50, 52.785948 }, //23 October 1982, UT1-UTC= 0.3980519, Accumulated Leap Seconds=21
  { 2445266.50, 52.787672 }, //24 October 1982, UT1-UTC= 0.3963275, Accumulated Leap Seconds=21
  { 2445267.50, 52.789433 }, //25 October 1982, UT1-UTC= 0.3945673, Accumulated Leap Seconds=21
  { 2445268.50, 52.791285 }, //26 October 1982, UT1-UTC= 0.3927147, Accumulated Leap Seconds=21
  { 2445269.50, 52.793276 }, //27 October 1982, UT1-UTC= 0.3907243, Accumulated Leap Seconds=21
  { 2445270.50, 52.795435 }, //28 October 1982, UT1-UTC= 0.3885650, Accumulated Leap Seconds=21
  { 2445271.50, 52.797775 }, //29 October 1982, UT1-UTC= 0.3862246, Accumulated Leap Seconds=21
  { 2445272.50, 52.800283 }, //30 October 1982, UT1-UTC= 0.3837172, Accumulated Leap Seconds=21
  { 2445273.50, 52.802910 }, //31 October 1982, UT1-UTC= 0.3810899, Accumulated Leap Seconds=21
  { 2445274.50, 52.805579 }, //01 November 1982, UT1-UTC= 0.3784208, Accumulated Leap Seconds=21
  { 2445275.50, 52.808198 }, //02 November 1982, UT1-UTC= 0.3758020, Accumulated Leap Seconds=21
  { 2445276.50, 52.810689 }, //03 November 1982, UT1-UTC= 0.3733113, Accumulated Leap Seconds=21
  { 2445277.50, 52.813016 }, //04 November 1982, UT1-UTC= 0.3709844, Accumulated Leap Seconds=21
  { 2445278.50, 52.815199 }, //05 November 1982, UT1-UTC= 0.3688011, Accumulated Leap Seconds=21
  { 2445279.50, 52.817307 }, //06 November 1982, UT1-UTC= 0.3666925, Accumulated Leap Seconds=21
  { 2445280.50, 52.819435 }, //07 November 1982, UT1-UTC= 0.3645653, Accumulated Leap Seconds=21
  { 2445281.50, 52.821667 }, //08 November 1982, UT1-UTC= 0.3623329, Accumulated Leap Seconds=21
  { 2445282.50, 52.824061 }, //09 November 1982, UT1-UTC= 0.3599387, Accumulated Leap Seconds=21
  { 2445283.50, 52.826631 }, //10 November 1982, UT1-UTC= 0.3573693, Accumulated Leap Seconds=21
  { 2445284.50, 52.829345 }, //11 November 1982, UT1-UTC= 0.3546546, Accumulated Leap Seconds=21
  { 2445285.50, 52.832144 }, //12 November 1982, UT1-UTC= 0.3518560, Accumulated Leap Seconds=21
  { 2445286.50, 52.834950 }, //13 November 1982, UT1-UTC= 0.3490498, Accumulated Leap Seconds=21
  { 2445287.50, 52.837690 }, //14 November 1982, UT1-UTC= 0.3463100, Accumulated Leap Seconds=21
  { 2445288.50, 52.840306 }, //15 November 1982, UT1-UTC= 0.3436937, Accumulated Leap Seconds=21
  { 2445289.50, 52.842767 }, //16 November 1982, UT1-UTC= 0.3412326, Accumulated Leap Seconds=21
  { 2445290.50, 52.845069 }, //17 November 1982, UT1-UTC= 0.3389307, Accumulated Leap Seconds=21
  { 2445291.50, 52.847232 }, //18 November 1982, UT1-UTC= 0.3367679, Accumulated Leap Seconds=21
  { 2445292.50, 52.849296 }, //19 November 1982, UT1-UTC= 0.3347043, Accumulated Leap Seconds=21
  { 2445293.50, 52.851314 }, //20 November 1982, UT1-UTC= 0.3326864, Accumulated Leap Seconds=21
  { 2445294.50, 52.853344 }, //21 November 1982, UT1-UTC= 0.3306559, Accumulated Leap Seconds=21
  { 2445295.50, 52.855443 }, //22 November 1982, UT1-UTC= 0.3285572, Accumulated Leap Seconds=21
  { 2445296.50, 52.857655 }, //23 November 1982, UT1-UTC= 0.3263445, Accumulated Leap Seconds=21
  { 2445297.50, 52.860014 }, //24 November 1982, UT1-UTC= 0.3239858, Accumulated Leap Seconds=21
  { 2445298.50, 52.862535 }, //25 November 1982, UT1-UTC= 0.3214646, Accumulated Leap Seconds=21
  { 2445299.50, 52.865217 }, //26 November 1982, UT1-UTC= 0.3187832, Accumulated Leap Seconds=21
  { 2445300.50, 52.868032 }, //27 November 1982, UT1-UTC= 0.3159685, Accumulated Leap Seconds=21
  { 2445301.50, 52.870922 }, //28 November 1982, UT1-UTC= 0.3130776, Accumulated Leap Seconds=21
  { 2445302.50, 52.873806 }, //29 November 1982, UT1-UTC= 0.3101936, Accumulated Leap Seconds=21
  { 2445303.50, 52.876594 }, //30 November 1982, UT1-UTC= 0.3074063, Accumulated Leap Seconds=21
  { 2445304.50, 52.879219 }, //01 December 1982, UT1-UTC= 0.3047811, Accumulated Leap Seconds=21
  { 2445305.50, 52.881669 }, //02 December 1982, UT1-UTC= 0.3023308, Accumulated Leap Seconds=21
  { 2445306.50, 52.883993 }, //03 December 1982, UT1-UTC= 0.3000071, Accumulated Leap Seconds=21
  { 2445307.50, 52.886281 }, //04 December 1982, UT1-UTC= 0.2977185, Accumulated Leap Seconds=21
  { 2445308.50, 52.888635 }, //05 December 1982, UT1-UTC= 0.2953651, Accumulated Leap Seconds=21
  { 2445309.50, 52.891125 }, //06 December 1982, UT1-UTC= 0.2928746, Accumulated Leap Seconds=21
  { 2445310.50, 52.893775 }, //07 December 1982, UT1-UTC= 0.2902250, Accumulated Leap Seconds=21
  { 2445311.50, 52.896557 }, //08 December 1982, UT1-UTC= 0.2874426, Accumulated Leap Seconds=21
  { 2445312.50, 52.899412 }, //09 December 1982, UT1-UTC= 0.2845876, Accumulated Leap Seconds=21
  { 2445313.50, 52.902267 }, //10 December 1982, UT1-UTC= 0.2817327, Accumulated Leap Seconds=21
  { 2445314.50, 52.905054 }, //11 December 1982, UT1-UTC= 0.2789457, Accumulated Leap Seconds=21
  { 2445315.50, 52.907722 }, //12 December 1982, UT1-UTC= 0.2762783, Accumulated Leap Seconds=21
  { 2445316.50, 52.910240 }, //13 December 1982, UT1-UTC= 0.2737600, Accumulated Leap Seconds=21
  { 2445317.50, 52.912603 }, //14 December 1982, UT1-UTC= 0.2713973, Accumulated Leap Seconds=21
  { 2445318.50, 52.914826 }, //15 December 1982, UT1-UTC= 0.2691743, Accumulated Leap Seconds=21
  { 2445319.50, 52.916944 }, //16 December 1982, UT1-UTC= 0.2670564, Accumulated Leap Seconds=21
  { 2445320.50, 52.919005 }, //17 December 1982, UT1-UTC= 0.2649953, Accumulated Leap Seconds=21
  { 2445321.50, 52.921062 }, //18 December 1982, UT1-UTC= 0.2629377, Accumulated Leap Seconds=21
  { 2445322.50, 52.923169 }, //19 December 1982, UT1-UTC= 0.2608307, Accumulated Leap Seconds=21
  { 2445323.50, 52.925372 }, //20 December 1982, UT1-UTC= 0.2586284, Accumulated Leap Seconds=21
  { 2445324.50, 52.927703 }, //21 December 1982, UT1-UTC= 0.2562967, Accumulated Leap Seconds=21
  { 2445325.50, 52.930182 }, //22 December 1982, UT1-UTC= 0.2538178, Accumulated Leap Seconds=21
  { 2445326.50, 52.932809 }, //23 December 1982, UT1-UTC= 0.2511908, Accumulated Leap Seconds=21
  { 2445327.50, 52.935567 }, //24 December 1982, UT1-UTC= 0.2484329, Accumulated Leap Seconds=21
  { 2445328.50, 52.938417 }, //25 December 1982, UT1-UTC= 0.2455832, Accumulated Leap Seconds=21
  { 2445329.50, 52.941296 }, //26 December 1982, UT1-UTC= 0.2427042, Accumulated Leap Seconds=21
  { 2445330.50, 52.944125 }, //27 December 1982, UT1-UTC= 0.2398746, Accumulated Leap Seconds=21
  { 2445331.50, 52.946832 }, //28 December 1982, UT1-UTC= 0.2371677, Accumulated Leap Seconds=21
  { 2445332.50, 52.949378 }, //29 December 1982, UT1-UTC= 0.2346222, Accumulated Leap Seconds=21
  { 2445333.50, 52.951780 }, //30 December 1982, UT1-UTC= 0.2322199, Accumulated Leap Seconds=21
  { 2445334.50, 52.954114 }, //31 December 1982, UT1-UTC= 0.2298861, Accumulated Leap Seconds=21
  { 2445335.50, 52.956484 }, //01 January 1983, UT1-UTC= 0.2275162, Accumulated Leap Seconds=21
  { 2445336.50, 52.958983 }, //02 January 1983, UT1-UTC= 0.2250174, Accumulated Leap Seconds=21
  { 2445337.50, 52.961655 }, //03 January 1983, UT1-UTC= 0.2223447, Accumulated Leap Seconds=21
  { 2445338.50, 52.964487 }, //04 January 1983, UT1-UTC= 0.2195132, Accumulated Leap Seconds=21
  { 2445339.50, 52.967414 }, //05 January 1983, UT1-UTC= 0.2165864, Accumulated Leap Seconds=21
  { 2445340.50, 52.970352 }, //06 January 1983, UT1-UTC= 0.2136481, Accumulated Leap Seconds=21
  { 2445341.50, 52.973224 }, //07 January 1983, UT1-UTC= 0.2107756, Accumulated Leap Seconds=21
  { 2445342.50, 52.975976 }, //08 January 1983, UT1-UTC= 0.2080235, Accumulated Leap Seconds=21
  { 2445343.50, 52.978582 }, //09 January 1983, UT1-UTC= 0.2054182, Accumulated Leap Seconds=21
  { 2445344.50, 52.981040 }, //10 January 1983, UT1-UTC= 0.2029598, Accumulated Leap Seconds=21
  { 2445345.50, 52.983372 }, //11 January 1983, UT1-UTC= 0.2006275, Accumulated Leap Seconds=21
  { 2445346.50, 52.985616 }, //12 January 1983, UT1-UTC= 0.1983841, Accumulated Leap Seconds=21
  { 2445347.50, 52.987819 }, //13 January 1983, UT1-UTC= 0.1961809, Accumulated Leap Seconds=21
  { 2445348.50, 52.990039 }, //14 January 1983, UT1-UTC= 0.1939612, Accumulated Leap Seconds=21
  { 2445349.50, 52.992333 }, //15 January 1983, UT1-UTC= 0.1916668, Accumulated Leap Seconds=21
  { 2445350.50, 52.994754 }, //16 January 1983, UT1-UTC= 0.1892461, Accumulated Leap Seconds=21
  { 2445351.50, 52.997340 }, //17 January 1983, UT1-UTC= 0.1866602, Accumulated Leap Seconds=21
  { 2445352.50, 53.000111 }, //18 January 1983, UT1-UTC= 0.1838891, Accumulated Leap Seconds=21
  { 2445353.50, 53.003067 }, //19 January 1983, UT1-UTC= 0.1809333, Accumulated Leap Seconds=21
  { 2445354.50, 53.006187 }, //20 January 1983, UT1-UTC= 0.1778134, Accumulated Leap Seconds=21
  { 2445355.50, 53.009432 }, //21 January 1983, UT1-UTC= 0.1745679, Accumulated Leap Seconds=21
  { 2445356.50, 53.012748 }, //22 January 1983, UT1-UTC= 0.1712518, Accumulated Leap Seconds=21
  { 2445357.50, 53.016067 }, //23 January 1983, UT1-UTC= 0.1679328, Accumulated Leap Seconds=21
  { 2445358.50, 53.019319 }, //24 January 1983, UT1-UTC= 0.1646808, Accumulated Leap Seconds=21
  { 2445359.50, 53.022452 }, //25 January 1983, UT1-UTC= 0.1615482, Accumulated Leap Seconds=21
  { 2445360.50, 53.025455 }, //26 January 1983, UT1-UTC= 0.1585450, Accumulated Leap Seconds=21
  { 2445361.50, 53.028375 }, //27 January 1983, UT1-UTC= 0.1556253, Accumulated Leap Seconds=21
  { 2445362.50, 53.031301 }, //28 January 1983, UT1-UTC= 0.1526985, Accumulated Leap Seconds=21
  { 2445363.50, 53.034338 }, //29 January 1983, UT1-UTC= 0.1496624, Accumulated Leap Seconds=21
  { 2445364.50, 53.037554 }, //30 January 1983, UT1-UTC= 0.1464464, Accumulated Leap Seconds=21
  { 2445365.50, 53.040959 }, //31 January 1983, UT1-UTC= 0.1430413, Accumulated Leap Seconds=21
  { 2445366.50, 53.044497 }, //01 February 1983, UT1-UTC= 0.1395029, Accumulated Leap Seconds=21
  { 2445367.50, 53.048071 }, //02 February 1983, UT1-UTC= 0.1359290, Accumulated Leap Seconds=21
  { 2445368.50, 53.051576 }, //03 February 1983, UT1-UTC= 0.1324243, Accumulated Leap Seconds=21
  { 2445369.50, 53.054928 }, //04 February 1983, UT1-UTC= 0.1290716, Accumulated Leap Seconds=21
  { 2445370.50, 53.058082 }, //05 February 1983, UT1-UTC= 0.1259177, Accumulated Leap Seconds=21
  { 2445371.50, 53.061025 }, //06 February 1983, UT1-UTC= 0.1229750, Accumulated Leap Seconds=21
  { 2445372.50, 53.063771 }, //07 February 1983, UT1-UTC= 0.1202285, Accumulated Leap Seconds=21
  { 2445373.50, 53.066356 }, //08 February 1983, UT1-UTC= 0.1176439, Accumulated Leap Seconds=21
  { 2445374.50, 53.068826 }, //09 February 1983, UT1-UTC= 0.1151739, Accumulated Leap Seconds=21
  { 2445375.50, 53.071238 }, //10 February 1983, UT1-UTC= 0.1127618, Accumulated Leap Seconds=21
  { 2445376.50, 53.073652 }, //11 February 1983, UT1-UTC= 0.1103476, Accumulated Leap Seconds=21
  { 2445377.50, 53.076126 }, //12 February 1983, UT1-UTC= 0.1078737, Accumulated Leap Seconds=21
  { 2445378.50, 53.078706 }, //13 February 1983, UT1-UTC= 0.1052937, Accumulated Leap Seconds=21
  { 2445379.50, 53.081421 }, //14 February 1983, UT1-UTC= 0.1025785, Accumulated Leap Seconds=21
  { 2445380.50, 53.084280 }, //15 February 1983, UT1-UTC= 0.0997205, Accumulated Leap Seconds=21
  { 2445381.50, 53.087264 }, //16 February 1983, UT1-UTC= 0.0967356, Accumulated Leap Seconds=21
  { 2445382.50, 53.090341 }, //17 February 1983, UT1-UTC= 0.0936592, Accumulated Leap Seconds=21
  { 2445383.50, 53.093459 }, //18 February 1983, UT1-UTC= 0.0905413, Accumulated Leap Seconds=21
  { 2445384.50, 53.096562 }, //19 February 1983, UT1-UTC= 0.0874380, Accumulated Leap Seconds=21
  { 2445385.50, 53.099596 }, //20 February 1983, UT1-UTC= 0.0844038, Accumulated Leap Seconds=21
  { 2445386.50, 53.102521 }, //21 February 1983, UT1-UTC= 0.0814787, Accumulated Leap Seconds=21
  { 2445387.50, 53.105328 }, //22 February 1983, UT1-UTC= 0.0786717, Accumulated Leap Seconds=21
  { 2445388.50, 53.108052 }, //23 February 1983, UT1-UTC= 0.0759478, Accumulated Leap Seconds=21
  { 2445389.50, 53.110771 }, //24 February 1983, UT1-UTC= 0.0732289, Accumulated Leap Seconds=21
  { 2445390.50, 53.113584 }, //25 February 1983, UT1-UTC= 0.0704155, Accumulated Leap Seconds=21
  { 2445391.50, 53.116582 }, //26 February 1983, UT1-UTC= 0.0674177, Accumulated Leap Seconds=21
  { 2445392.50, 53.119807 }, //27 February 1983, UT1-UTC= 0.0641933, Accumulated Leap Seconds=21
  { 2445393.50, 53.123231 }, //28 February 1983, UT1-UTC= 0.0607686, Accumulated Leap Seconds=21
  { 2445394.50, 53.126769 }, //01 March 1983, UT1-UTC= 0.0572310, Accumulated Leap Seconds=21
  { 2445395.50, 53.130302 }, //02 March 1983, UT1-UTC= 0.0536983, Accumulated Leap Seconds=21
  { 2445396.50, 53.133720 }, //03 March 1983, UT1-UTC= 0.0502796, Accumulated Leap Seconds=21
  { 2445397.50, 53.136950 }, //04 March 1983, UT1-UTC= 0.0470495, Accumulated Leap Seconds=21
  { 2445398.50, 53.139962 }, //05 March 1983, UT1-UTC= 0.0440379, Accumulated Leap Seconds=21
  { 2445399.50, 53.142764 }, //06 March 1983, UT1-UTC= 0.0412361, Accumulated Leap Seconds=21
  { 2445400.50, 53.145393 }, //07 March 1983, UT1-UTC= 0.0386074, Accumulated Leap Seconds=21
  { 2445401.50, 53.147901 }, //08 March 1983, UT1-UTC= 0.0360988, Accumulated Leap Seconds=21
  { 2445402.50, 53.150351 }, //09 March 1983, UT1-UTC= 0.0336493, Accumulated Leap Seconds=21
  { 2445403.50, 53.152804 }, //10 March 1983, UT1-UTC= 0.0311964, Accumulated Leap Seconds=21
  { 2445404.50, 53.155319 }, //11 March 1983, UT1-UTC= 0.0286811, Accumulated Leap Seconds=21
  { 2445405.50, 53.157946 }, //12 March 1983, UT1-UTC= 0.0260544, Accumulated Leap Seconds=21
  { 2445406.50, 53.160717 }, //13 March 1983, UT1-UTC= 0.0232825, Accumulated Leap Seconds=21
  { 2445407.50, 53.163647 }, //14 March 1983, UT1-UTC= 0.0203530, Accumulated Leap Seconds=21
  { 2445408.50, 53.166721 }, //15 March 1983, UT1-UTC= 0.0172785, Accumulated Leap Seconds=21
  { 2445409.50, 53.169903 }, //16 March 1983, UT1-UTC= 0.0140966, Accumulated Leap Seconds=21
  { 2445410.50, 53.173134 }, //17 March 1983, UT1-UTC= 0.0108659, Accumulated Leap Seconds=21
  { 2445411.50, 53.176344 }, //18 March 1983, UT1-UTC= 0.0076555, Accumulated Leap Seconds=21
  { 2445412.50, 53.179469 }, //19 March 1983, UT1-UTC= 0.0045306, Accumulated Leap Seconds=21
  { 2445413.50, 53.182463 }, //20 March 1983, UT1-UTC= 0.0015371, Accumulated Leap Seconds=21
  { 2445414.50, 53.185311 }, //21 March 1983, UT1-UTC=-0.0013109, Accumulated Leap Seconds=21
  { 2445415.50, 53.188039 }, //22 March 1983, UT1-UTC=-0.0040393, Accumulated Leap Seconds=21
  { 2445416.50, 53.190714 }, //23 March 1983, UT1-UTC=-0.0067144, Accumulated Leap Seconds=21
  { 2445417.50, 53.193431 }, //24 March 1983, UT1-UTC=-0.0094311, Accumulated Leap Seconds=21
  { 2445418.50, 53.196288 }, //25 March 1983, UT1-UTC=-0.0122879, Accumulated Leap Seconds=21
  { 2445419.50, 53.199355 }, //26 March 1983, UT1-UTC=-0.0153552, Accumulated Leap Seconds=21
  { 2445420.50, 53.202648 }, //27 March 1983, UT1-UTC=-0.0186478, Accumulated Leap Seconds=21
  { 2445421.50, 53.206115 }, //28 March 1983, UT1-UTC=-0.0221153, Accumulated Leap Seconds=21
  { 2445422.50, 53.209658 }, //29 March 1983, UT1-UTC=-0.0256580, Accumulated Leap Seconds=21
  { 2445423.50, 53.213159 }, //30 March 1983, UT1-UTC=-0.0291587, Accumulated Leap Seconds=21
  { 2445424.50, 53.216518 }, //31 March 1983, UT1-UTC=-0.0325178, Accumulated Leap Seconds=21
  { 2445425.50, 53.219675 }, //01 April 1983, UT1-UTC=-0.0356749, Accumulated Leap Seconds=21
  { 2445426.50, 53.222615 }, //02 April 1983, UT1-UTC=-0.0386148, Accumulated Leap Seconds=21
  { 2445427.50, 53.225359 }, //03 April 1983, UT1-UTC=-0.0413595, Accumulated Leap Seconds=21
  { 2445428.50, 53.227955 }, //04 April 1983, UT1-UTC=-0.0439548, Accumulated Leap Seconds=21
  { 2445429.50, 53.230459 }, //05 April 1983, UT1-UTC=-0.0464595, Accumulated Leap Seconds=21
  { 2445430.50, 53.232936 }, //06 April 1983, UT1-UTC=-0.0489363, Accumulated Leap Seconds=21
  { 2445431.50, 53.235445 }, //07 April 1983, UT1-UTC=-0.0514449, Accumulated Leap Seconds=21
  { 2445432.50, 53.238037 }, //08 April 1983, UT1-UTC=-0.0540366, Accumulated Leap Seconds=21
  { 2445433.50, 53.240748 }, //09 April 1983, UT1-UTC=-0.0567483, Accumulated Leap Seconds=21
  { 2445434.50, 53.243601 }, //10 April 1983, UT1-UTC=-0.0596006, Accumulated Leap Seconds=21
  { 2445435.50, 53.246593 }, //11 April 1983, UT1-UTC=-0.0625928, Accumulated Leap Seconds=21
  { 2445436.50, 53.249699 }, //12 April 1983, UT1-UTC=-0.0656991, Accumulated Leap Seconds=21
  { 2445437.50, 53.252868 }, //13 April 1983, UT1-UTC=-0.0688677, Accumulated Leap Seconds=21
  { 2445438.50, 53.256027 }, //14 April 1983, UT1-UTC=-0.0720272, Accumulated Leap Seconds=21
  { 2445439.50, 53.259100 }, //15 April 1983, UT1-UTC=-0.0751004, Accumulated Leap Seconds=21
  { 2445440.50, 53.262024 }, //16 April 1983, UT1-UTC=-0.0780245, Accumulated Leap Seconds=21
  { 2445441.50, 53.264771 }, //17 April 1983, UT1-UTC=-0.0807706, Accumulated Leap Seconds=21
  { 2445442.50, 53.267355 }, //18 April 1983, UT1-UTC=-0.0833549, Accumulated Leap Seconds=21
  { 2445443.50, 53.269835 }, //19 April 1983, UT1-UTC=-0.0858353, Accumulated Leap Seconds=21
  { 2445444.50, 53.272298 }, //20 April 1983, UT1-UTC=-0.0882984, Accumulated Leap Seconds=21
  { 2445445.50, 53.274837 }, //21 April 1983, UT1-UTC=-0.0908370, Accumulated Leap Seconds=21
  { 2445446.50, 53.277524 }, //22 April 1983, UT1-UTC=-0.0935240, Accumulated Leap Seconds=21
  { 2445447.50, 53.280391 }, //23 April 1983, UT1-UTC=-0.0963907, Accumulated Leap Seconds=21
  { 2445448.50, 53.283414 }, //24 April 1983, UT1-UTC=-0.0994141, Accumulated Leap Seconds=21
  { 2445449.50, 53.286522 }, //25 April 1983, UT1-UTC=-0.1025219, Accumulated Leap Seconds=21
  { 2445450.50, 53.289614 }, //26 April 1983, UT1-UTC=-0.1056141, Accumulated Leap Seconds=21
  { 2445451.50, 53.292589 }, //27 April 1983, UT1-UTC=-0.1085892, Accumulated Leap Seconds=21
  { 2445452.50, 53.295372 }, //28 April 1983, UT1-UTC=-0.1113719, Accumulated Leap Seconds=21
  { 2445453.50, 53.297927 }, //29 April 1983, UT1-UTC=-0.1139272, Accumulated Leap Seconds=21
  { 2445454.50, 53.300261 }, //30 April 1983, UT1-UTC=-0.1162615, Accumulated Leap Seconds=21
  { 2445455.50, 53.302414 }, //01 May 1983, UT1-UTC=-0.1184139, Accumulated Leap Seconds=21
  { 2445456.50, 53.304444 }, //02 May 1983, UT1-UTC=-0.1204438, Accumulated Leap Seconds=21
  { 2445457.50, 53.306420 }, //03 May 1983, UT1-UTC=-0.1224196, Accumulated Leap Seconds=21
  { 2445458.50, 53.308409 }, //04 May 1983, UT1-UTC=-0.1244092, Accumulated Leap Seconds=21
  { 2445459.50, 53.310472 }, //05 May 1983, UT1-UTC=-0.1264725, Accumulated Leap Seconds=21
  { 2445460.50, 53.312656 }, //06 May 1983, UT1-UTC=-0.1286562, Accumulated Leap Seconds=21
  { 2445461.50, 53.314990 }, //07 May 1983, UT1-UTC=-0.1309902, Accumulated Leap Seconds=21
  { 2445462.50, 53.317483 }, //08 May 1983, UT1-UTC=-0.1334833, Accumulated Leap Seconds=21
  { 2445463.50, 53.320122 }, //09 May 1983, UT1-UTC=-0.1361219, Accumulated Leap Seconds=21
  { 2445464.50, 53.322868 }, //10 May 1983, UT1-UTC=-0.1388681, Accumulated Leap Seconds=21
  { 2445465.50, 53.325660 }, //11 May 1983, UT1-UTC=-0.1416598, Accumulated Leap Seconds=21
  { 2445466.50, 53.328417 }, //12 May 1983, UT1-UTC=-0.1444174, Accumulated Leap Seconds=21
  { 2445467.50, 53.331063 }, //13 May 1983, UT1-UTC=-0.1470626, Accumulated Leap Seconds=21
  { 2445468.50, 53.333543 }, //14 May 1983, UT1-UTC=-0.1495435, Accumulated Leap Seconds=21
  { 2445469.50, 53.335856 }, //15 May 1983, UT1-UTC=-0.1518561, Accumulated Leap Seconds=21
  { 2445470.50, 53.338048 }, //16 May 1983, UT1-UTC=-0.1540477, Accumulated Leap Seconds=21
  { 2445471.50, 53.340204 }, //17 May 1983, UT1-UTC=-0.1562041, Accumulated Leap Seconds=21
  { 2445472.50, 53.342422 }, //18 May 1983, UT1-UTC=-0.1584216, Accumulated Leap Seconds=21
  { 2445473.50, 53.344777 }, //19 May 1983, UT1-UTC=-0.1607772, Accumulated Leap Seconds=21
  { 2445474.50, 53.347307 }, //20 May 1983, UT1-UTC=-0.1633072, Accumulated Leap Seconds=21
  { 2445475.50, 53.349998 }, //21 May 1983, UT1-UTC=-0.1659979, Accumulated Leap Seconds=21
  { 2445476.50, 53.352792 }, //22 May 1983, UT1-UTC=-0.1687922, Accumulated Leap Seconds=21
  { 2445477.50, 53.355604 }, //23 May 1983, UT1-UTC=-0.1716045, Accumulated Leap Seconds=21
  { 2445478.50, 53.358343 }, //24 May 1983, UT1-UTC=-0.1743432, Accumulated Leap Seconds=21
  { 2445479.50, 53.360930 }, //25 May 1983, UT1-UTC=-0.1769304, Accumulated Leap Seconds=21
  { 2445480.50, 53.363318 }, //26 May 1983, UT1-UTC=-0.1793179, Accumulated Leap Seconds=21
  { 2445481.50, 53.365494 }, //27 May 1983, UT1-UTC=-0.1814945, Accumulated Leap Seconds=21
  { 2445482.50, 53.367483 }, //28 May 1983, UT1-UTC=-0.1834831, Accumulated Leap Seconds=21
  { 2445483.50, 53.369331 }, //29 May 1983, UT1-UTC=-0.1853312, Accumulated Leap Seconds=21
  { 2445484.50, 53.371101 }, //30 May 1983, UT1-UTC=-0.1871007, Accumulated Leap Seconds=21
  { 2445485.50, 53.372858 }, //31 May 1983, UT1-UTC=-0.1888580, Accumulated Leap Seconds=21
  { 2445486.50, 53.374664 }, //01 June 1983, UT1-UTC=-0.1906645, Accumulated Leap Seconds=21
  { 2445487.50, 53.376569 }, //02 June 1983, UT1-UTC=-0.1925689, Accumulated Leap Seconds=21
  { 2445488.50, 53.378602 }, //03 June 1983, UT1-UTC=-0.1946017, Accumulated Leap Seconds=21
  { 2445489.50, 53.380775 }, //04 June 1983, UT1-UTC=-0.1967747, Accumulated Leap Seconds=21
  { 2445490.50, 53.383082 }, //05 June 1983, UT1-UTC=-0.1990816, Accumulated Leap Seconds=21
  { 2445491.50, 53.385497 }, //06 June 1983, UT1-UTC=-0.2014974, Accumulated Leap Seconds=21
  { 2445492.50, 53.387977 }, //07 June 1983, UT1-UTC=-0.2039770, Accumulated Leap Seconds=21
  { 2445493.50, 53.390455 }, //08 June 1983, UT1-UTC=-0.2064550, Accumulated Leap Seconds=21
  { 2445494.50, 53.392855 }, //09 June 1983, UT1-UTC=-0.2088550, Accumulated Leap Seconds=21
  { 2445495.50, 53.395110 }, //10 June 1983, UT1-UTC=-0.2111105, Accumulated Leap Seconds=21
  { 2445496.50, 53.397192 }, //11 June 1983, UT1-UTC=-0.2131924, Accumulated Leap Seconds=21
  { 2445497.50, 53.399127 }, //12 June 1983, UT1-UTC=-0.2151267, Accumulated Leap Seconds=21
  { 2445498.50, 53.400992 }, //13 June 1983, UT1-UTC=-0.2169918, Accumulated Leap Seconds=21
  { 2445499.50, 53.402893 }, //14 June 1983, UT1-UTC=-0.2188927, Accumulated Leap Seconds=21
  { 2445500.50, 53.404922 }, //15 June 1983, UT1-UTC=-0.2209224, Accumulated Leap Seconds=21
  { 2445501.50, 53.407129 }, //16 June 1983, UT1-UTC=-0.2231291, Accumulated Leap Seconds=21
  { 2445502.50, 53.409502 }, //17 June 1983, UT1-UTC=-0.2255023, Accumulated Leap Seconds=21
  { 2445503.50, 53.411979 }, //18 June 1983, UT1-UTC=-0.2279791, Accumulated Leap Seconds=21
  { 2445504.50, 53.414467 }, //19 June 1983, UT1-UTC=-0.2304674, Accumulated Leap Seconds=21
  { 2445505.50, 53.416871 }, //20 June 1983, UT1-UTC=-0.2328710, Accumulated Leap Seconds=21
  { 2445506.50, 53.419110 }, //21 June 1983, UT1-UTC=-0.2351104, Accumulated Leap Seconds=21
  { 2445507.50, 53.421134 }, //22 June 1983, UT1-UTC=-0.2371338, Accumulated Leap Seconds=21
  { 2445508.50, 53.422922 }, //23 June 1983, UT1-UTC=-0.2389222, Accumulated Leap Seconds=21
  { 2445509.50, 53.424489 }, //24 June 1983, UT1-UTC=-0.2404892, Accumulated Leap Seconds=21
  { 2445510.50, 53.425874 }, //25 June 1983, UT1-UTC=-0.2418739, Accumulated Leap Seconds=21
  { 2445511.50, 53.427132 }, //26 June 1983, UT1-UTC=-0.2431316, Accumulated Leap Seconds=21
  { 2445512.50, 53.428326 }, //27 June 1983, UT1-UTC=-0.2443256, Accumulated Leap Seconds=21
  { 2445513.50, 53.429519 }, //28 June 1983, UT1-UTC=-0.2455186, Accumulated Leap Seconds=21
  { 2445514.50, 53.430763 }, //29 June 1983, UT1-UTC=-0.2467632, Accumulated Leap Seconds=21
  { 2445515.50, 53.432097 }, //30 June 1983, UT1-UTC=-0.2480974, Accumulated Leap Seconds=21
  { 2445516.50, 53.433540 }, //01 July 1983, UT1-UTC= 0.7504601, Accumulated Leap Seconds=22
  { 2445517.50, 53.435088 }, //02 July 1983, UT1-UTC= 0.7489115, Accumulated Leap Seconds=22
  { 2445518.50, 53.436722 }, //03 July 1983, UT1-UTC= 0.7472782, Accumulated Leap Seconds=22
  { 2445519.50, 53.438403 }, //04 July 1983, UT1-UTC= 0.7455971, Accumulated Leap Seconds=22
  { 2445520.50, 53.440080 }, //05 July 1983, UT1-UTC= 0.7439198, Accumulated Leap Seconds=22
  { 2445521.50, 53.441690 }, //06 July 1983, UT1-UTC= 0.7423097, Accumulated Leap Seconds=22
  { 2445522.50, 53.443169 }, //07 July 1983, UT1-UTC= 0.7408313, Accumulated Leap Seconds=22
  { 2445523.50, 53.444474 }, //08 July 1983, UT1-UTC= 0.7395256, Accumulated Leap Seconds=22
  { 2445524.50, 53.445612 }, //09 July 1983, UT1-UTC= 0.7383876, Accumulated Leap Seconds=22
  { 2445525.50, 53.446645 }, //10 July 1983, UT1-UTC= 0.7373551, Accumulated Leap Seconds=22
  { 2445526.50, 53.447679 }, //11 July 1983, UT1-UTC= 0.7363211, Accumulated Leap Seconds=22
  { 2445527.50, 53.448831 }, //12 July 1983, UT1-UTC= 0.7351691, Accumulated Leap Seconds=22
  { 2445528.50, 53.450183 }, //13 July 1983, UT1-UTC= 0.7338170, Accumulated Leap Seconds=22
  { 2445529.50, 53.451755 }, //14 July 1983, UT1-UTC= 0.7322452, Accumulated Leap Seconds=22
  { 2445530.50, 53.453500 }, //15 July 1983, UT1-UTC= 0.7304995, Accumulated Leap Seconds=22
  { 2445531.50, 53.455331 }, //16 July 1983, UT1-UTC= 0.7286694, Accumulated Leap Seconds=22
  { 2445532.50, 53.457143 }, //17 July 1983, UT1-UTC= 0.7268571, Accumulated Leap Seconds=22
  { 2445533.50, 53.458851 }, //18 July 1983, UT1-UTC= 0.7251489, Accumulated Leap Seconds=22
  { 2445534.50, 53.460405 }, //19 July 1983, UT1-UTC= 0.7235954, Accumulated Leap Seconds=22
  { 2445535.50, 53.461789 }, //20 July 1983, UT1-UTC= 0.7222106, Accumulated Leap Seconds=22
  { 2445536.50, 53.463017 }, //21 July 1983, UT1-UTC= 0.7209826, Accumulated Leap Seconds=22
  { 2445537.50, 53.464121 }, //22 July 1983, UT1-UTC= 0.7198788, Accumulated Leap Seconds=22
  { 2445538.50, 53.465147 }, //23 July 1983, UT1-UTC= 0.7188526, Accumulated Leap Seconds=22
  { 2445539.50, 53.466153 }, //24 July 1983, UT1-UTC= 0.7178470, Accumulated Leap Seconds=22
  { 2445540.50, 53.467196 }, //25 July 1983, UT1-UTC= 0.7168039, Accumulated Leap Seconds=22
  { 2445541.50, 53.468329 }, //26 July 1983, UT1-UTC= 0.7156714, Accumulated Leap Seconds=22
  { 2445542.50, 53.469589 }, //27 July 1983, UT1-UTC= 0.7144113, Accumulated Leap Seconds=22
  { 2445543.50, 53.470995 }, //28 July 1983, UT1-UTC= 0.7130053, Accumulated Leap Seconds=22
  { 2445544.50, 53.472545 }, //29 July 1983, UT1-UTC= 0.7114550, Accumulated Leap Seconds=22
  { 2445545.50, 53.474221 }, //30 July 1983, UT1-UTC= 0.7097792, Accumulated Leap Seconds=22
  { 2445546.50, 53.475988 }, //31 July 1983, UT1-UTC= 0.7080116, Accumulated Leap Seconds=22
  { 2445547.50, 53.477801 }, //01 August 1983, UT1-UTC= 0.7061985, Accumulated Leap Seconds=22
  { 2445548.50, 53.479605 }, //02 August 1983, UT1-UTC= 0.7043952, Accumulated Leap Seconds=22
  { 2445549.50, 53.481341 }, //03 August 1983, UT1-UTC= 0.7026587, Accumulated Leap Seconds=22
  { 2445550.50, 53.482964 }, //04 August 1983, UT1-UTC= 0.7010361, Accumulated Leap Seconds=22
  { 2445551.50, 53.484454 }, //05 August 1983, UT1-UTC= 0.6995459, Accumulated Leap Seconds=22
  { 2445552.50, 53.485841 }, //06 August 1983, UT1-UTC= 0.6981587, Accumulated Leap Seconds=22
  { 2445553.50, 53.487207 }, //07 August 1983, UT1-UTC= 0.6967933, Accumulated Leap Seconds=22
  { 2445554.50, 53.488663 }, //08 August 1983, UT1-UTC= 0.6953371, Accumulated Leap Seconds=22
  { 2445555.50, 53.490312 }, //09 August 1983, UT1-UTC= 0.6936880, Accumulated Leap Seconds=22
  { 2445556.50, 53.492203 }, //10 August 1983, UT1-UTC= 0.6917972, Accumulated Leap Seconds=22
  { 2445557.50, 53.494309 }, //11 August 1983, UT1-UTC= 0.6896908, Accumulated Leap Seconds=22
  { 2445558.50, 53.496539 }, //12 August 1983, UT1-UTC= 0.6874608, Accumulated Leap Seconds=22
  { 2445559.50, 53.498768 }, //13 August 1983, UT1-UTC= 0.6852322, Accumulated Leap Seconds=22
  { 2445560.50, 53.500879 }, //14 August 1983, UT1-UTC= 0.6831210, Accumulated Leap Seconds=22
  { 2445561.50, 53.502797 }, //15 August 1983, UT1-UTC= 0.6812028, Accumulated Leap Seconds=22
  { 2445562.50, 53.504497 }, //16 August 1983, UT1-UTC= 0.6795028, Accumulated Leap Seconds=22
  { 2445563.50, 53.505996 }, //17 August 1983, UT1-UTC= 0.6780038, Accumulated Leap Seconds=22
  { 2445564.50, 53.507338 }, //18 August 1983, UT1-UTC= 0.6766624, Accumulated Leap Seconds=22
  { 2445565.50, 53.508579 }, //19 August 1983, UT1-UTC= 0.6754213, Accumulated Leap Seconds=22
  { 2445566.50, 53.509782 }, //20 August 1983, UT1-UTC= 0.6742180, Accumulated Leap Seconds=22
  { 2445567.50, 53.511009 }, //21 August 1983, UT1-UTC= 0.6729913, Accumulated Leap Seconds=22
  { 2445568.50, 53.512314 }, //22 August 1983, UT1-UTC= 0.6716863, Accumulated Leap Seconds=22
  { 2445569.50, 53.513739 }, //23 August 1983, UT1-UTC= 0.6702610, Accumulated Leap Seconds=22
  { 2445570.50, 53.515308 }, //24 August 1983, UT1-UTC= 0.6686917, Accumulated Leap Seconds=22
  { 2445571.50, 53.517024 }, //25 August 1983, UT1-UTC= 0.6669759, Accumulated Leap Seconds=22
  { 2445572.50, 53.518868 }, //26 August 1983, UT1-UTC= 0.6651322, Accumulated Leap Seconds=22
  { 2445573.50, 53.520802 }, //27 August 1983, UT1-UTC= 0.6631980, Accumulated Leap Seconds=22
  { 2445574.50, 53.522775 }, //28 August 1983, UT1-UTC= 0.6612246, Accumulated Leap Seconds=22
  { 2445575.50, 53.524730 }, //29 August 1983, UT1-UTC= 0.6592703, Accumulated Leap Seconds=22
  { 2445576.50, 53.526609 }, //30 August 1983, UT1-UTC= 0.6573905, Accumulated Leap Seconds=22
  { 2445577.50, 53.528371 }, //31 August 1983, UT1-UTC= 0.6556291, Accumulated Leap Seconds=22
  { 2445578.50, 53.529994 }, //01 September 1983, UT1-UTC= 0.6540063, Accumulated Leap Seconds=22
  { 2445579.50, 53.531495 }, //02 September 1983, UT1-UTC= 0.6525049, Accumulated Leap Seconds=22
  { 2445580.50, 53.532937 }, //03 September 1983, UT1-UTC= 0.6510632, Accumulated Leap Seconds=22
  { 2445581.50, 53.534420 }, //04 September 1983, UT1-UTC= 0.6495800, Accumulated Leap Seconds=22
  { 2445582.50, 53.536057 }, //05 September 1983, UT1-UTC= 0.6479430, Accumulated Leap Seconds=22
  { 2445583.50, 53.537930 }, //06 September 1983, UT1-UTC= 0.6460696, Accumulated Leap Seconds=22
  { 2445584.50, 53.540057 }, //07 September 1983, UT1-UTC= 0.6439426, Accumulated Leap Seconds=22
  { 2445585.50, 53.542381 }, //08 September 1983, UT1-UTC= 0.6416186, Accumulated Leap Seconds=22
  { 2445586.50, 53.544788 }, //09 September 1983, UT1-UTC= 0.6392122, Accumulated Leap Seconds=22
  { 2445587.50, 53.547141 }, //10 September 1983, UT1-UTC= 0.6368585, Accumulated Leap Seconds=22
  { 2445588.50, 53.549333 }, //11 September 1983, UT1-UTC= 0.6346673, Accumulated Leap Seconds=22
  { 2445589.50, 53.551300 }, //12 September 1983, UT1-UTC= 0.6326998, Accumulated Leap Seconds=22
  { 2445590.50, 53.553039 }, //13 September 1983, UT1-UTC= 0.6309608, Accumulated Leap Seconds=22
  { 2445591.50, 53.554588 }, //14 September 1983, UT1-UTC= 0.6294124, Accumulated Leap Seconds=22
  { 2445592.50, 53.556006 }, //15 September 1983, UT1-UTC= 0.6279940, Accumulated Leap Seconds=22
  { 2445593.50, 53.557362 }, //16 September 1983, UT1-UTC= 0.6266384, Accumulated Leap Seconds=22
  { 2445594.50, 53.558719 }, //17 September 1983, UT1-UTC= 0.6252807, Accumulated Leap Seconds=22
  { 2445595.50, 53.560138 }, //18 September 1983, UT1-UTC= 0.6238623, Accumulated Leap Seconds=22
  { 2445596.50, 53.561665 }, //19 September 1983, UT1-UTC= 0.6223348, Accumulated Leap Seconds=22
  { 2445597.50, 53.563333 }, //20 September 1983, UT1-UTC= 0.6206672, Accumulated Leap Seconds=22
  { 2445598.50, 53.565152 }, //21 September 1983, UT1-UTC= 0.6188483, Accumulated Leap Seconds=22
  { 2445599.50, 53.567111 }, //22 September 1983, UT1-UTC= 0.6168886, Accumulated Leap Seconds=22
  { 2445600.50, 53.569180 }, //23 September 1983, UT1-UTC= 0.6148200, Accumulated Leap Seconds=22
  { 2445601.50, 53.571307 }, //24 September 1983, UT1-UTC= 0.6126926, Accumulated Leap Seconds=22
  { 2445602.50, 53.573435 }, //25 September 1983, UT1-UTC= 0.6105654, Accumulated Leap Seconds=22
  { 2445603.50, 53.575505 }, //26 September 1983, UT1-UTC= 0.6084951, Accumulated Leap Seconds=22
  { 2445604.50, 53.577476 }, //27 September 1983, UT1-UTC= 0.6065239, Accumulated Leap Seconds=22
  { 2445605.50, 53.579331 }, //28 September 1983, UT1-UTC= 0.6046690, Accumulated Leap Seconds=22
  { 2445606.50, 53.581086 }, //29 September 1983, UT1-UTC= 0.6029143, Accumulated Leap Seconds=22
  { 2445607.50, 53.582793 }, //30 September 1983, UT1-UTC= 0.6012065, Accumulated Leap Seconds=22
  { 2445608.50, 53.584539 }, //01 October 1983, UT1-UTC= 0.5994608, Accumulated Leap Seconds=22
  { 2445609.50, 53.586425 }, //02 October 1983, UT1-UTC= 0.5975750, Accumulated Leap Seconds=22
  { 2445610.50, 53.588542 }, //03 October 1983, UT1-UTC= 0.5954578, Accumulated Leap Seconds=22
  { 2445611.50, 53.590936 }, //04 October 1983, UT1-UTC= 0.5930642, Accumulated Leap Seconds=22
  { 2445612.50, 53.593581 }, //05 October 1983, UT1-UTC= 0.5904186, Accumulated Leap Seconds=22
  { 2445613.50, 53.596385 }, //06 October 1983, UT1-UTC= 0.5876153, Accumulated Leap Seconds=22
  { 2445614.50, 53.599207 }, //07 October 1983, UT1-UTC= 0.5847934, Accumulated Leap Seconds=22
  { 2445615.50, 53.601907 }, //08 October 1983, UT1-UTC= 0.5820926, Accumulated Leap Seconds=22
  { 2445616.50, 53.604387 }, //09 October 1983, UT1-UTC= 0.5796131, Accumulated Leap Seconds=22
  { 2445617.50, 53.606605 }, //10 October 1983, UT1-UTC= 0.5773946, Accumulated Leap Seconds=22
  { 2445618.50, 53.608579 }, //11 October 1983, UT1-UTC= 0.5754209, Accumulated Leap Seconds=22
  { 2445619.50, 53.610362 }, //12 October 1983, UT1-UTC= 0.5736376, Accumulated Leap Seconds=22
  { 2445620.50, 53.612028 }, //13 October 1983, UT1-UTC= 0.5719722, Accumulated Leap Seconds=22
  { 2445621.50, 53.613649 }, //14 October 1983, UT1-UTC= 0.5703506, Accumulated Leap Seconds=22
  { 2445622.50, 53.615293 }, //15 October 1983, UT1-UTC= 0.5687067, Accumulated Leap Seconds=22
  { 2445623.50, 53.617015 }, //16 October 1983, UT1-UTC= 0.5669849, Accumulated Leap Seconds=22
  { 2445624.50, 53.618854 }, //17 October 1983, UT1-UTC= 0.5651460, Accumulated Leap Seconds=22
  { 2445625.50, 53.620829 }, //18 October 1983, UT1-UTC= 0.5631707, Accumulated Leap Seconds=22
  { 2445626.50, 53.622942 }, //19 October 1983, UT1-UTC= 0.5610581, Accumulated Leap Seconds=22
  { 2445627.50, 53.625173 }, //20 October 1983, UT1-UTC= 0.5588267, Accumulated Leap Seconds=22
  { 2445628.50, 53.627485 }, //21 October 1983, UT1-UTC= 0.5565148, Accumulated Leap Seconds=22
  { 2445629.50, 53.629823 }, //22 October 1983, UT1-UTC= 0.5541771, Accumulated Leap Seconds=22
  { 2445630.50, 53.632126 }, //23 October 1983, UT1-UTC= 0.5518736, Accumulated Leap Seconds=22
  { 2445631.50, 53.634344 }, //24 October 1983, UT1-UTC= 0.5496557, Accumulated Leap Seconds=22
  { 2445632.50, 53.636449 }, //25 October 1983, UT1-UTC= 0.5475509, Accumulated Leap Seconds=22
  { 2445633.50, 53.638451 }, //26 October 1983, UT1-UTC= 0.5455492, Accumulated Leap Seconds=22
  { 2445634.50, 53.640401 }, //27 October 1983, UT1-UTC= 0.5435992, Accumulated Leap Seconds=22
  { 2445635.50, 53.642380 }, //28 October 1983, UT1-UTC= 0.5416204, Accumulated Leap Seconds=22
  { 2445636.50, 53.644483 }, //29 October 1983, UT1-UTC= 0.5395172, Accumulated Leap Seconds=22
  { 2445637.50, 53.646802 }, //30 October 1983, UT1-UTC= 0.5371975, Accumulated Leap Seconds=22
  { 2445638.50, 53.649395 }, //31 October 1983, UT1-UTC= 0.5346047, Accumulated Leap Seconds=22
  { 2445639.50, 53.652263 }, //01 November 1983, UT1-UTC= 0.5317372, Accumulated Leap Seconds=22
  { 2445640.50, 53.655345 }, //02 November 1983, UT1-UTC= 0.5286549, Accumulated Leap Seconds=22
  { 2445641.50, 53.658530 }, //03 November 1983, UT1-UTC= 0.5254704, Accumulated Leap Seconds=22
  { 2445642.50, 53.661681 }, //04 November 1983, UT1-UTC= 0.5223186, Accumulated Leap Seconds=22
  { 2445643.50, 53.664681 }, //05 November 1983, UT1-UTC= 0.5193194, Accumulated Leap Seconds=22
  { 2445644.50, 53.667451 }, //06 November 1983, UT1-UTC= 0.5165493, Accumulated Leap Seconds=22
  { 2445645.50, 53.669971 }, //07 November 1983, UT1-UTC= 0.5140290, Accumulated Leap Seconds=22
  { 2445646.50, 53.672269 }, //08 November 1983, UT1-UTC= 0.5117311, Accumulated Leap Seconds=22
  { 2445647.50, 53.674405 }, //09 November 1983, UT1-UTC= 0.5095951, Accumulated Leap Seconds=22
  { 2445648.50, 53.676456 }, //10 November 1983, UT1-UTC= 0.5075441, Accumulated Leap Seconds=22
  { 2445649.50, 53.678496 }, //11 November 1983, UT1-UTC= 0.5055044, Accumulated Leap Seconds=22
  { 2445650.50, 53.680587 }, //12 November 1983, UT1-UTC= 0.5034134, Accumulated Leap Seconds=22
  { 2445651.50, 53.682774 }, //13 November 1983, UT1-UTC= 0.5012261, Accumulated Leap Seconds=22
  { 2445652.50, 53.685082 }, //14 November 1983, UT1-UTC= 0.4989183, Accumulated Leap Seconds=22
  { 2445653.50, 53.687514 }, //15 November 1983, UT1-UTC= 0.4964856, Accumulated Leap Seconds=22
  { 2445654.50, 53.690056 }, //16 November 1983, UT1-UTC= 0.4939437, Accumulated Leap Seconds=22
  { 2445655.50, 53.692673 }, //17 November 1983, UT1-UTC= 0.4913271, Accumulated Leap Seconds=22
  { 2445656.50, 53.695311 }, //18 November 1983, UT1-UTC= 0.4886891, Accumulated Leap Seconds=22
  { 2445657.50, 53.697906 }, //19 November 1983, UT1-UTC= 0.4860939, Accumulated Leap Seconds=22
  { 2445658.50, 53.700392 }, //20 November 1983, UT1-UTC= 0.4836084, Accumulated Leap Seconds=22
  { 2445659.50, 53.702716 }, //21 November 1983, UT1-UTC= 0.4812844, Accumulated Leap Seconds=22
  { 2445660.50, 53.704862 }, //22 November 1983, UT1-UTC= 0.4791380, Accumulated Leap Seconds=22
  { 2445661.50, 53.706863 }, //23 November 1983, UT1-UTC= 0.4771374, Accumulated Leap Seconds=22
  { 2445662.50, 53.708793 }, //24 November 1983, UT1-UTC= 0.4752071, Accumulated Leap Seconds=22
  { 2445663.50, 53.710753 }, //25 November 1983, UT1-UTC= 0.4732469, Accumulated Leap Seconds=22
  { 2445664.50, 53.712838 }, //26 November 1983, UT1-UTC= 0.4711618, Accumulated Leap Seconds=22
  { 2445665.50, 53.715111 }, //27 November 1983, UT1-UTC= 0.4688893, Accumulated Leap Seconds=22
  { 2445666.50, 53.717583 }, //28 November 1983, UT1-UTC= 0.4664169, Accumulated Leap Seconds=22
  { 2445667.50, 53.720214 }, //29 November 1983, UT1-UTC= 0.4637860, Accumulated Leap Seconds=22
  { 2445668.50, 53.722917 }, //30 November 1983, UT1-UTC= 0.4610830, Accumulated Leap Seconds=22
  { 2445669.50, 53.725584 }, //01 December 1983, UT1-UTC= 0.4584156, Accumulated Leap Seconds=22
  { 2445670.50, 53.728114 }, //02 December 1983, UT1-UTC= 0.4558864, Accumulated Leap Seconds=22
  { 2445671.50, 53.730430 }, //03 December 1983, UT1-UTC= 0.4535700, Accumulated Leap Seconds=22
  { 2445672.50, 53.732502 }, //04 December 1983, UT1-UTC= 0.4514976, Accumulated Leap Seconds=22
  { 2445673.50, 53.734346 }, //05 December 1983, UT1-UTC= 0.4496539, Accumulated Leap Seconds=22
  { 2445674.50, 53.736014 }, //06 December 1983, UT1-UTC= 0.4479860, Accumulated Leap Seconds=22
  { 2445675.50, 53.737579 }, //07 December 1983, UT1-UTC= 0.4464213, Accumulated Leap Seconds=22
  { 2445676.50, 53.739116 }, //08 December 1983, UT1-UTC= 0.4448837, Accumulated Leap Seconds=22
  { 2445677.50, 53.740696 }, //09 December 1983, UT1-UTC= 0.4433035, Accumulated Leap Seconds=22
  { 2445678.50, 53.742373 }, //10 December 1983, UT1-UTC= 0.4416268, Accumulated Leap Seconds=22
  { 2445679.50, 53.744179 }, //11 December 1983, UT1-UTC= 0.4398205, Accumulated Leap Seconds=22
  { 2445680.50, 53.746126 }, //12 December 1983, UT1-UTC= 0.4378741, Accumulated Leap Seconds=22
  { 2445681.50, 53.748202 }, //13 December 1983, UT1-UTC= 0.4357983, Accumulated Leap Seconds=22
  { 2445682.50, 53.750378 }, //14 December 1983, UT1-UTC= 0.4336221, Accumulated Leap Seconds=22
  { 2445683.50, 53.752612 }, //15 December 1983, UT1-UTC= 0.4313877, Accumulated Leap Seconds=22
  { 2445684.50, 53.754852 }, //16 December 1983, UT1-UTC= 0.4291483, Accumulated Leap Seconds=22
  { 2445685.50, 53.757035 }, //17 December 1983, UT1-UTC= 0.4269650, Accumulated Leap Seconds=22
  { 2445686.50, 53.759103 }, //18 December 1983, UT1-UTC= 0.4248974, Accumulated Leap Seconds=22
  { 2445687.50, 53.761019 }, //19 December 1983, UT1-UTC= 0.4229814, Accumulated Leap Seconds=22
  { 2445688.50, 53.762790 }, //20 December 1983, UT1-UTC= 0.4212099, Accumulated Leap Seconds=22
  { 2445689.50, 53.764477 }, //21 December 1983, UT1-UTC= 0.4195234, Accumulated Leap Seconds=22
  { 2445690.50, 53.766177 }, //22 December 1983, UT1-UTC= 0.4178229, Accumulated Leap Seconds=22
  { 2445691.50, 53.767998 }, //23 December 1983, UT1-UTC= 0.4160021, Accumulated Leap Seconds=22
  { 2445692.50, 53.770018 }, //24 December 1983, UT1-UTC= 0.4139816, Accumulated Leap Seconds=22
  { 2445693.50, 53.772263 }, //25 December 1983, UT1-UTC= 0.4117373, Accumulated Leap Seconds=22
  { 2445694.50, 53.774692 }, //26 December 1983, UT1-UTC= 0.4093081, Accumulated Leap Seconds=22
  { 2445695.50, 53.777220 }, //27 December 1983, UT1-UTC= 0.4067800, Accumulated Leap Seconds=22
  { 2445696.50, 53.779741 }, //28 December 1983, UT1-UTC= 0.4042590, Accumulated Leap Seconds=22
  { 2445697.50, 53.782156 }, //29 December 1983, UT1-UTC= 0.4018441, Accumulated Leap Seconds=22
  { 2445698.50, 53.784394 }, //30 December 1983, UT1-UTC= 0.3996064, Accumulated Leap Seconds=22
  { 2445699.50, 53.786420 }, //31 December 1983, UT1-UTC= 0.3975795, Accumulated Leap Seconds=22
  { 2445700.50, 53.788242 }, //01 January 1984, UT1-UTC= 0.3957582, Accumulated Leap Seconds=22
  { 2445701.50, 53.789893 }, //02 January 1984, UT1-UTC= 0.3941071, Accumulated Leap Seconds=22
  { 2445702.50, 53.791430 }, //03 January 1984, UT1-UTC= 0.3925702, Accumulated Leap Seconds=22
  { 2445703.50, 53.792916 }, //04 January 1984, UT1-UTC= 0.3910842, Accumulated Leap Seconds=22
  { 2445704.50, 53.794411 }, //05 January 1984, UT1-UTC= 0.3895893, Accumulated Leap Seconds=22
  { 2445705.50, 53.795963 }, //06 January 1984, UT1-UTC= 0.3880369, Accumulated Leap Seconds=22
  { 2445706.50, 53.797606 }, //07 January 1984, UT1-UTC= 0.3863944, Accumulated Leap Seconds=22
  { 2445707.50, 53.799351 }, //08 January 1984, UT1-UTC= 0.3846493, Accumulated Leap Seconds=22
  { 2445708.50, 53.801189 }, //09 January 1984, UT1-UTC= 0.3828108, Accumulated Leap Seconds=22
  { 2445709.50, 53.803093 }, //10 January 1984, UT1-UTC= 0.3809067, Accumulated Leap Seconds=22
  { 2445710.50, 53.805023 }, //11 January 1984, UT1-UTC= 0.3789766, Accumulated Leap Seconds=22
  { 2445711.50, 53.806930 }, //12 January 1984, UT1-UTC= 0.3770702, Accumulated Leap Seconds=22
  { 2445712.50, 53.808757 }, //13 January 1984, UT1-UTC= 0.3752434, Accumulated Leap Seconds=22
  { 2445713.50, 53.810453 }, //14 January 1984, UT1-UTC= 0.3735466, Accumulated Leap Seconds=22
  { 2445714.50, 53.811984 }, //15 January 1984, UT1-UTC= 0.3720162, Accumulated Leap Seconds=22
  { 2445715.50, 53.813343 }, //16 January 1984, UT1-UTC= 0.3706567, Accumulated Leap Seconds=22
  { 2445716.50, 53.814576 }, //17 January 1984, UT1-UTC= 0.3694237, Accumulated Leap Seconds=22
  { 2445717.50, 53.815777 }, //18 January 1984, UT1-UTC= 0.3682231, Accumulated Leap Seconds=22
  { 2445718.50, 53.817065 }, //19 January 1984, UT1-UTC= 0.3669347, Accumulated Leap Seconds=22
  { 2445719.50, 53.818545 }, //20 January 1984, UT1-UTC= 0.3654554, Accumulated Leap Seconds=22
  { 2445720.50, 53.820263 }, //21 January 1984, UT1-UTC= 0.3637371, Accumulated Leap Seconds=22
  { 2445721.50, 53.822194 }, //22 January 1984, UT1-UTC= 0.3618062, Accumulated Leap Seconds=22
  { 2445722.50, 53.824245 }, //23 January 1984, UT1-UTC= 0.3597547, Accumulated Leap Seconds=22
  { 2445723.50, 53.826293 }, //24 January 1984, UT1-UTC= 0.3577070, Accumulated Leap Seconds=22
  { 2445724.50, 53.828221 }, //25 January 1984, UT1-UTC= 0.3557785, Accumulated Leap Seconds=22
  { 2445725.50, 53.829951 }, //26 January 1984, UT1-UTC= 0.3540485, Accumulated Leap Seconds=22
  { 2445726.50, 53.831448 }, //27 January 1984, UT1-UTC= 0.3525517, Accumulated Leap Seconds=22
  { 2445727.50, 53.832722 }, //28 January 1984, UT1-UTC= 0.3512783, Accumulated Leap Seconds=22
  { 2445728.50, 53.833815 }, //29 January 1984, UT1-UTC= 0.3501851, Accumulated Leap Seconds=22
  { 2445729.50, 53.834789 }, //30 January 1984, UT1-UTC= 0.3492107, Accumulated Leap Seconds=22
  { 2445730.50, 53.835714 }, //31 January 1984, UT1-UTC= 0.3482856, Accumulated Leap Seconds=22
  { 2445731.50, 53.836662 }, //01 February 1984, UT1-UTC= 0.3473375, Accumulated Leap Seconds=22
  { 2445732.50, 53.837697 }, //02 February 1984, UT1-UTC= 0.3463029, Accumulated Leap Seconds=22
  { 2445733.50, 53.838865 }, //03 February 1984, UT1-UTC= 0.3451347, Accumulated Leap Seconds=22
  { 2445734.50, 53.840193 }, //04 February 1984, UT1-UTC= 0.3438073, Accumulated Leap Seconds=22
  { 2445735.50, 53.841682 }, //05 February 1984, UT1-UTC= 0.3423177, Accumulated Leap Seconds=22
  { 2445736.50, 53.843310 }, //06 February 1984, UT1-UTC= 0.3406896, Accumulated Leap Seconds=22
  { 2445737.50, 53.845031 }, //07 February 1984, UT1-UTC= 0.3389685, Accumulated Leap Seconds=22
  { 2445738.50, 53.846792 }, //08 February 1984, UT1-UTC= 0.3372082, Accumulated Leap Seconds=22
  { 2445739.50, 53.848536 }, //09 February 1984, UT1-UTC= 0.3354641, Accumulated Leap Seconds=22
  { 2445740.50, 53.850209 }, //10 February 1984, UT1-UTC= 0.3337909, Accumulated Leap Seconds=22
  { 2445741.50, 53.851773 }, //11 February 1984, UT1-UTC= 0.3322269, Accumulated Leap Seconds=22
  { 2445742.50, 53.853213 }, //12 February 1984, UT1-UTC= 0.3307868, Accumulated Leap Seconds=22
  { 2445743.50, 53.854549 }, //13 February 1984, UT1-UTC= 0.3294511, Accumulated Leap Seconds=22
  { 2445744.50, 53.855846 }, //14 February 1984, UT1-UTC= 0.3281535, Accumulated Leap Seconds=22
  { 2445745.50, 53.857214 }, //15 February 1984, UT1-UTC= 0.3267865, Accumulated Leap Seconds=22
  { 2445746.50, 53.858771 }, //16 February 1984, UT1-UTC= 0.3252288, Accumulated Leap Seconds=22
  { 2445747.50, 53.860608 }, //17 February 1984, UT1-UTC= 0.3233925, Accumulated Leap Seconds=22
  { 2445748.50, 53.862738 }, //18 February 1984, UT1-UTC= 0.3212624, Accumulated Leap Seconds=22
  { 2445749.50, 53.865089 }, //19 February 1984, UT1-UTC= 0.3189109, Accumulated Leap Seconds=22
  { 2445750.50, 53.867525 }, //20 February 1984, UT1-UTC= 0.3164745, Accumulated Leap Seconds=22
  { 2445751.50, 53.869892 }, //21 February 1984, UT1-UTC= 0.3141078, Accumulated Leap Seconds=22
  { 2445752.50, 53.872060 }, //22 February 1984, UT1-UTC= 0.3119404, Accumulated Leap Seconds=22
  { 2445753.50, 53.873954 }, //23 February 1984, UT1-UTC= 0.3100456, Accumulated Leap Seconds=22
  { 2445754.50, 53.875565 }, //24 February 1984, UT1-UTC= 0.3084352, Accumulated Leap Seconds=22
  { 2445755.50, 53.876926 }, //25 February 1984, UT1-UTC= 0.3070736, Accumulated Leap Seconds=22
  { 2445756.50, 53.878116 }, //26 February 1984, UT1-UTC= 0.3058842, Accumulated Leap Seconds=22
  { 2445757.50, 53.879227 }, //27 February 1984, UT1-UTC= 0.3047733, Accumulated Leap Seconds=22
  { 2445758.50, 53.880353 }, //28 February 1984, UT1-UTC= 0.3036466, Accumulated Leap Seconds=22
  { 2445759.50, 53.881579 }, //29 February 1984, UT1-UTC= 0.3024211, Accumulated Leap Seconds=22
  { 2445760.50, 53.882966 }, //01 March 1984, UT1-UTC= 0.3010335, Accumulated Leap Seconds=22
  { 2445761.50, 53.884541 }, //02 March 1984, UT1-UTC= 0.2994587, Accumulated Leap Seconds=22
  { 2445762.50, 53.886299 }, //03 March 1984, UT1-UTC= 0.2977010, Accumulated Leap Seconds=22
  { 2445763.50, 53.888213 }, //04 March 1984, UT1-UTC= 0.2957867, Accumulated Leap Seconds=22
  { 2445764.50, 53.890242 }, //05 March 1984, UT1-UTC= 0.2937584, Accumulated Leap Seconds=22
  { 2445765.50, 53.892329 }, //06 March 1984, UT1-UTC= 0.2916714, Accumulated Leap Seconds=22
  { 2445766.50, 53.894417 }, //07 March 1984, UT1-UTC= 0.2895828, Accumulated Leap Seconds=22
  { 2445767.50, 53.896451 }, //08 March 1984, UT1-UTC= 0.2875494, Accumulated Leap Seconds=22
  { 2445768.50, 53.898380 }, //09 March 1984, UT1-UTC= 0.2856198, Accumulated Leap Seconds=22
  { 2445769.50, 53.900177 }, //10 March 1984, UT1-UTC= 0.2838229, Accumulated Leap Seconds=22
  { 2445770.50, 53.901844 }, //11 March 1984, UT1-UTC= 0.2821564, Accumulated Leap Seconds=22
  { 2445771.50, 53.903429 }, //12 March 1984, UT1-UTC= 0.2805711, Accumulated Leap Seconds=22
  { 2445772.50, 53.905025 }, //13 March 1984, UT1-UTC= 0.2789747, Accumulated Leap Seconds=22
  { 2445773.50, 53.906752 }, //14 March 1984, UT1-UTC= 0.2772484, Accumulated Leap Seconds=22
  { 2445774.50, 53.908719 }, //15 March 1984, UT1-UTC= 0.2752812, Accumulated Leap Seconds=22
  { 2445775.50, 53.910987 }, //16 March 1984, UT1-UTC= 0.2730125, Accumulated Leap Seconds=22
  { 2445776.50, 53.913537 }, //17 March 1984, UT1-UTC= 0.2704625, Accumulated Leap Seconds=22
  { 2445777.50, 53.916267 }, //18 March 1984, UT1-UTC= 0.2677332, Accumulated Leap Seconds=22
  { 2445778.50, 53.919024 }, //19 March 1984, UT1-UTC= 0.2649764, Accumulated Leap Seconds=22
  { 2445779.50, 53.921658 }, //20 March 1984, UT1-UTC= 0.2623416, Accumulated Leap Seconds=22
  { 2445780.50, 53.924070 }, //21 March 1984, UT1-UTC= 0.2599304, Accumulated Leap Seconds=22
  { 2445781.50, 53.926218 }, //22 March 1984, UT1-UTC= 0.2577817, Accumulated Leap Seconds=22
  { 2445782.50, 53.928124 }, //23 March 1984, UT1-UTC= 0.2558761, Accumulated Leap Seconds=22
  { 2445783.50, 53.929845 }, //24 March 1984, UT1-UTC= 0.2541553, Accumulated Leap Seconds=22
  { 2445784.50, 53.931453 }, //25 March 1984, UT1-UTC= 0.2525470, Accumulated Leap Seconds=22
  { 2445785.50, 53.933022 }, //26 March 1984, UT1-UTC= 0.2509782, Accumulated Leap Seconds=22
  { 2445786.50, 53.934623 }, //27 March 1984, UT1-UTC= 0.2493771, Accumulated Leap Seconds=22
  { 2445787.50, 53.936315 }, //28 March 1984, UT1-UTC= 0.2476848, Accumulated Leap Seconds=22
  { 2445788.50, 53.938132 }, //29 March 1984, UT1-UTC= 0.2458675, Accumulated Leap Seconds=22
  { 2445789.50, 53.940082 }, //30 March 1984, UT1-UTC= 0.2439182, Accumulated Leap Seconds=22
  { 2445790.50, 53.942147 }, //31 March 1984, UT1-UTC= 0.2418529, Accumulated Leap Seconds=22
  { 2445791.50, 53.944290 }, //01 April 1984, UT1-UTC= 0.2397096, Accumulated Leap Seconds=22
  { 2445792.50, 53.946457 }, //02 April 1984, UT1-UTC= 0.2375427, Accumulated Leap Seconds=22
  { 2445793.50, 53.948586 }, //03 April 1984, UT1-UTC= 0.2354138, Accumulated Leap Seconds=22
  { 2445794.50, 53.950618 }, //04 April 1984, UT1-UTC= 0.2333819, Accumulated Leap Seconds=22
  { 2445795.50, 53.952509 }, //05 April 1984, UT1-UTC= 0.2314908, Accumulated Leap Seconds=22
  { 2445796.50, 53.954240 }, //06 April 1984, UT1-UTC= 0.2297600, Accumulated Leap Seconds=22
  { 2445797.50, 53.955824 }, //07 April 1984, UT1-UTC= 0.2281758, Accumulated Leap Seconds=22
  { 2445798.50, 53.957311 }, //08 April 1984, UT1-UTC= 0.2266885, Accumulated Leap Seconds=22
  { 2445799.50, 53.958786 }, //09 April 1984, UT1-UTC= 0.2252140, Accumulated Leap Seconds=22
  { 2445800.50, 53.960356 }, //10 April 1984, UT1-UTC= 0.2236441, Accumulated Leap Seconds=22
  { 2445801.50, 53.962132 }, //11 April 1984, UT1-UTC= 0.2218677, Accumulated Leap Seconds=22
  { 2445802.50, 53.964199 }, //12 April 1984, UT1-UTC= 0.2198005, Accumulated Leap Seconds=22
  { 2445803.50, 53.966579 }, //13 April 1984, UT1-UTC= 0.2174211, Accumulated Leap Seconds=22
  { 2445804.50, 53.969210 }, //14 April 1984, UT1-UTC= 0.2147900, Accumulated Leap Seconds=22
  { 2445805.50, 53.971961 }, //15 April 1984, UT1-UTC= 0.2120386, Accumulated Leap Seconds=22
  { 2445806.50, 53.974669 }, //16 April 1984, UT1-UTC= 0.2093306, Accumulated Leap Seconds=22
  { 2445807.50, 53.977192 }, //17 April 1984, UT1-UTC= 0.2068083, Accumulated Leap Seconds=22
  { 2445808.50, 53.979450 }, //18 April 1984, UT1-UTC= 0.2045502, Accumulated Leap Seconds=22
  { 2445809.50, 53.981442 }, //19 April 1984, UT1-UTC= 0.2025581, Accumulated Leap Seconds=22
  { 2445810.50, 53.983226 }, //20 April 1984, UT1-UTC= 0.2007738, Accumulated Leap Seconds=22
  { 2445811.50, 53.984882 }, //21 April 1984, UT1-UTC= 0.1991176, Accumulated Leap Seconds=22
  { 2445812.50, 53.986494 }, //22 April 1984, UT1-UTC= 0.1975056, Accumulated Leap Seconds=22
  { 2445813.50, 53.988133 }, //23 April 1984, UT1-UTC= 0.1958665, Accumulated Leap Seconds=22
  { 2445814.50, 53.989850 }, //24 April 1984, UT1-UTC= 0.1941500, Accumulated Leap Seconds=22
  { 2445815.50, 53.991673 }, //25 April 1984, UT1-UTC= 0.1923267, Accumulated Leap Seconds=22
  { 2445816.50, 53.993615 }, //26 April 1984, UT1-UTC= 0.1903852, Accumulated Leap Seconds=22
  { 2445817.50, 53.995667 }, //27 April 1984, UT1-UTC= 0.1883326, Accumulated Leap Seconds=22
  { 2445818.50, 53.997805 }, //28 April 1984, UT1-UTC= 0.1861949, Accumulated Leap Seconds=22
  { 2445819.50, 53.999986 }, //29 April 1984, UT1-UTC= 0.1840142, Accumulated Leap Seconds=22
  { 2445820.50, 54.002155 }, //30 April 1984, UT1-UTC= 0.1818451, Accumulated Leap Seconds=22
  { 2445821.50, 54.004248 }, //01 May 1984, UT1-UTC= 0.1797522, Accumulated Leap Seconds=22
  { 2445822.50, 54.006206 }, //02 May 1984, UT1-UTC= 0.1777936, Accumulated Leap Seconds=22
  { 2445823.50, 54.007993 }, //03 May 1984, UT1-UTC= 0.1760074, Accumulated Leap Seconds=22
  { 2445824.50, 54.009599 }, //04 May 1984, UT1-UTC= 0.1744012, Accumulated Leap Seconds=22
  { 2445825.50, 54.011056 }, //05 May 1984, UT1-UTC= 0.1729440, Accumulated Leap Seconds=22
  { 2445826.50, 54.012434 }, //06 May 1984, UT1-UTC= 0.1715659, Accumulated Leap Seconds=22
  { 2445827.50, 54.013832 }, //07 May 1984, UT1-UTC= 0.1701682, Accumulated Leap Seconds=22
  { 2445828.50, 54.015355 }, //08 May 1984, UT1-UTC= 0.1686447, Accumulated Leap Seconds=22
  { 2445829.50, 54.017093 }, //09 May 1984, UT1-UTC= 0.1669073, Accumulated Leap Seconds=22
  { 2445830.50, 54.019088 }, //10 May 1984, UT1-UTC= 0.1649122, Accumulated Leap Seconds=22
  { 2445831.50, 54.021316 }, //11 May 1984, UT1-UTC= 0.1626841, Accumulated Leap Seconds=22
  { 2445832.50, 54.023684 }, //12 May 1984, UT1-UTC= 0.1603162, Accumulated Leap Seconds=22
  { 2445833.50, 54.026054 }, //13 May 1984, UT1-UTC= 0.1579464, Accumulated Leap Seconds=22
  { 2445834.50, 54.028280 }, //14 May 1984, UT1-UTC= 0.1557197, Accumulated Leap Seconds=22
  { 2445835.50, 54.030255 }, //15 May 1984, UT1-UTC= 0.1537447, Accumulated Leap Seconds=22
  { 2445836.50, 54.031937 }, //16 May 1984, UT1-UTC= 0.1520625, Accumulated Leap Seconds=22
  { 2445837.50, 54.033355 }, //17 May 1984, UT1-UTC= 0.1506448, Accumulated Leap Seconds=22
  { 2445838.50, 54.034585 }, //18 May 1984, UT1-UTC= 0.1494152, Accumulated Leap Seconds=22
  { 2445839.50, 54.035720 }, //19 May 1984, UT1-UTC= 0.1482799, Accumulated Leap Seconds=22
  { 2445840.50, 54.036848 }, //20 May 1984, UT1-UTC= 0.1471519, Accumulated Leap Seconds=22
  { 2445841.50, 54.038036 }, //21 May 1984, UT1-UTC= 0.1459637, Accumulated Leap Seconds=22
  { 2445842.50, 54.039326 }, //22 May 1984, UT1-UTC= 0.1446738, Accumulated Leap Seconds=22
  { 2445843.50, 54.040733 }, //23 May 1984, UT1-UTC= 0.1432668, Accumulated Leap Seconds=22
  { 2445844.50, 54.042249 }, //24 May 1984, UT1-UTC= 0.1417508, Accumulated Leap Seconds=22
  { 2445845.50, 54.043845 }, //25 May 1984, UT1-UTC= 0.1401545, Accumulated Leap Seconds=22
  { 2445846.50, 54.045482 }, //26 May 1984, UT1-UTC= 0.1385176, Accumulated Leap Seconds=22
  { 2445847.50, 54.047112 }, //27 May 1984, UT1-UTC= 0.1368879, Accumulated Leap Seconds=22
  { 2445848.50, 54.048683 }, //28 May 1984, UT1-UTC= 0.1353168, Accumulated Leap Seconds=22
  { 2445849.50, 54.050147 }, //29 May 1984, UT1-UTC= 0.1338526, Accumulated Leap Seconds=22
  { 2445850.50, 54.051467 }, //30 May 1984, UT1-UTC= 0.1325325, Accumulated Leap Seconds=22
  { 2445851.50, 54.052624 }, //31 May 1984, UT1-UTC= 0.1313760, Accumulated Leap Seconds=22
  { 2445852.50, 54.053634 }, //01 June 1984, UT1-UTC= 0.1303658, Accumulated Leap Seconds=22
  { 2445853.50, 54.054557 }, //02 June 1984, UT1-UTC= 0.1294432, Accumulated Leap Seconds=22
  { 2445854.50, 54.055479 }, //03 June 1984, UT1-UTC= 0.1285205, Accumulated Leap Seconds=22
  { 2445855.50, 54.056498 }, //04 June 1984, UT1-UTC= 0.1275022, Accumulated Leap Seconds=22
  { 2445856.50, 54.057692 }, //05 June 1984, UT1-UTC= 0.1263082, Accumulated Leap Seconds=22
  { 2445857.50, 54.059101 }, //06 June 1984, UT1-UTC= 0.1248990, Accumulated Leap Seconds=22
  { 2445858.50, 54.060710 }, //07 June 1984, UT1-UTC= 0.1232903, Accumulated Leap Seconds=22
  { 2445859.50, 54.062446 }, //08 June 1984, UT1-UTC= 0.1215544, Accumulated Leap Seconds=22
  { 2445860.50, 54.064195 }, //09 June 1984, UT1-UTC= 0.1198048, Accumulated Leap Seconds=22
  { 2445861.50, 54.065837 }, //10 June 1984, UT1-UTC= 0.1181633, Accumulated Leap Seconds=22
  { 2445862.50, 54.067272 }, //11 June 1984, UT1-UTC= 0.1167280, Accumulated Leap Seconds=22
  { 2445863.50, 54.068451 }, //12 June 1984, UT1-UTC= 0.1155489, Accumulated Leap Seconds=22
  { 2445864.50, 54.069384 }, //13 June 1984, UT1-UTC= 0.1146165, Accumulated Leap Seconds=22
  { 2445865.50, 54.070129 }, //14 June 1984, UT1-UTC= 0.1138714, Accumulated Leap Seconds=22
  { 2445866.50, 54.070770 }, //15 June 1984, UT1-UTC= 0.1132296, Accumulated Leap Seconds=22
  { 2445867.50, 54.071396 }, //16 June 1984, UT1-UTC= 0.1126038, Accumulated Leap Seconds=22
  { 2445868.50, 54.072080 }, //17 June 1984, UT1-UTC= 0.1119198, Accumulated Leap Seconds=22
  { 2445869.50, 54.072874 }, //18 June 1984, UT1-UTC= 0.1111256, Accumulated Leap Seconds=22
  { 2445870.50, 54.073806 }, //19 June 1984, UT1-UTC= 0.1101944, Accumulated Leap Seconds=22
  { 2445871.50, 54.074875 }, //20 June 1984, UT1-UTC= 0.1091254, Accumulated Leap Seconds=22
  { 2445872.50, 54.076062 }, //21 June 1984, UT1-UTC= 0.1079382, Accumulated Leap Seconds=22
  { 2445873.50, 54.077331 }, //22 June 1984, UT1-UTC= 0.1066690, Accumulated Leap Seconds=22
  { 2445874.50, 54.078636 }, //23 June 1984, UT1-UTC= 0.1053636, Accumulated Leap Seconds=22
  { 2445875.50, 54.079927 }, //24 June 1984, UT1-UTC= 0.1040729, Accumulated Leap Seconds=22
  { 2445876.50, 54.081144 }, //25 June 1984, UT1-UTC= 0.1028558, Accumulated Leap Seconds=22
  { 2445877.50, 54.082232 }, //26 June 1984, UT1-UTC= 0.1017684, Accumulated Leap Seconds=22
  { 2445878.50, 54.083151 }, //27 June 1984, UT1-UTC= 0.1008493, Accumulated Leap Seconds=22
  { 2445879.50, 54.083895 }, //28 June 1984, UT1-UTC= 0.1001049, Accumulated Leap Seconds=22
  { 2445880.50, 54.084502 }, //29 June 1984, UT1-UTC= 0.0994975, Accumulated Leap Seconds=22
  { 2445881.50, 54.085052 }, //30 June 1984, UT1-UTC= 0.0989482, Accumulated Leap Seconds=22
  { 2445882.50, 54.085644 }, //01 July 1984, UT1-UTC= 0.0983560, Accumulated Leap Seconds=22
  { 2445883.50, 54.086369 }, //02 July 1984, UT1-UTC= 0.0976307, Accumulated Leap Seconds=22
  { 2445884.50, 54.087274 }, //03 July 1984, UT1-UTC= 0.0967263, Accumulated Leap Seconds=22
  { 2445885.50, 54.088342 }, //04 July 1984, UT1-UTC= 0.0956576, Accumulated Leap Seconds=22
  { 2445886.50, 54.089515 }, //05 July 1984, UT1-UTC= 0.0944846, Accumulated Leap Seconds=22
  { 2445887.50, 54.090700 }, //06 July 1984, UT1-UTC= 0.0933000, Accumulated Leap Seconds=22
  { 2445888.50, 54.091795 }, //07 July 1984, UT1-UTC= 0.0922050, Accumulated Leap Seconds=22
  { 2445889.50, 54.092719 }, //08 July 1984, UT1-UTC= 0.0912813, Accumulated Leap Seconds=22
  { 2445890.50, 54.093426 }, //09 July 1984, UT1-UTC= 0.0905740, Accumulated Leap Seconds=22
  { 2445891.50, 54.093912 }, //10 July 1984, UT1-UTC= 0.0900881, Accumulated Leap Seconds=22
  { 2445892.50, 54.094215 }, //11 July 1984, UT1-UTC= 0.0897854, Accumulated Leap Seconds=22
  { 2445893.50, 54.094402 }, //12 July 1984, UT1-UTC= 0.0895977, Accumulated Leap Seconds=22
  { 2445894.50, 54.094554 }, //13 July 1984, UT1-UTC= 0.0894462, Accumulated Leap Seconds=22
  { 2445895.50, 54.094743 }, //14 July 1984, UT1-UTC= 0.0892569, Accumulated Leap Seconds=22
  { 2445896.50, 54.095032 }, //15 July 1984, UT1-UTC= 0.0889679, Accumulated Leap Seconds=22
  { 2445897.50, 54.095460 }, //16 July 1984, UT1-UTC= 0.0885400, Accumulated Leap Seconds=22
  { 2445898.50, 54.096039 }, //17 July 1984, UT1-UTC= 0.0879610, Accumulated Leap Seconds=22
  { 2445899.50, 54.096758 }, //18 July 1984, UT1-UTC= 0.0872415, Accumulated Leap Seconds=22
  { 2445900.50, 54.097590 }, //19 July 1984, UT1-UTC= 0.0864095, Accumulated Leap Seconds=22
  { 2445901.50, 54.098489 }, //20 July 1984, UT1-UTC= 0.0855113, Accumulated Leap Seconds=22
  { 2445902.50, 54.099398 }, //21 July 1984, UT1-UTC= 0.0846016, Accumulated Leap Seconds=22
  { 2445903.50, 54.100267 }, //22 July 1984, UT1-UTC= 0.0837331, Accumulated Leap Seconds=22
  { 2445904.50, 54.101050 }, //23 July 1984, UT1-UTC= 0.0829500, Accumulated Leap Seconds=22
  { 2445905.50, 54.101722 }, //24 July 1984, UT1-UTC= 0.0822783, Accumulated Leap Seconds=22
  { 2445906.50, 54.102277 }, //25 July 1984, UT1-UTC= 0.0817233, Accumulated Leap Seconds=22
  { 2445907.50, 54.102744 }, //26 July 1984, UT1-UTC= 0.0812556, Accumulated Leap Seconds=22
  { 2445908.50, 54.103198 }, //27 July 1984, UT1-UTC= 0.0808025, Accumulated Leap Seconds=22
  { 2445909.50, 54.103744 }, //28 July 1984, UT1-UTC= 0.0802564, Accumulated Leap Seconds=22
  { 2445910.50, 54.104496 }, //29 July 1984, UT1-UTC= 0.0795038, Accumulated Leap Seconds=22
  { 2445911.50, 54.105534 }, //30 July 1984, UT1-UTC= 0.0784655, Accumulated Leap Seconds=22
  { 2445912.50, 54.106866 }, //31 July 1984, UT1-UTC= 0.0771344, Accumulated Leap Seconds=22
  { 2445913.50, 54.108412 }, //01 August 1984, UT1-UTC= 0.0755878, Accumulated Leap Seconds=22
  { 2445914.50, 54.110041 }, //02 August 1984, UT1-UTC= 0.0739594, Accumulated Leap Seconds=22
  { 2445915.50, 54.111600 }, //03 August 1984, UT1-UTC= 0.0723995, Accumulated Leap Seconds=22
  { 2445916.50, 54.112967 }, //04 August 1984, UT1-UTC= 0.0710330, Accumulated Leap Seconds=22
  { 2445917.50, 54.114068 }, //05 August 1984, UT1-UTC= 0.0699322, Accumulated Leap Seconds=22
  { 2445918.50, 54.114893 }, //06 August 1984, UT1-UTC= 0.0691069, Accumulated Leap Seconds=22
  { 2445919.50, 54.115496 }, //07 August 1984, UT1-UTC= 0.0685041, Accumulated Leap Seconds=22
  { 2445920.50, 54.115967 }, //08 August 1984, UT1-UTC= 0.0680327, Accumulated Leap Seconds=22
  { 2445921.50, 54.116406 }, //09 August 1984, UT1-UTC= 0.0675941, Accumulated Leap Seconds=22
  { 2445922.50, 54.116904 }, //10 August 1984, UT1-UTC= 0.0670961, Accumulated Leap Seconds=22
  { 2445923.50, 54.117536 }, //11 August 1984, UT1-UTC= 0.0664639, Accumulated Leap Seconds=22
  { 2445924.50, 54.118354 }, //12 August 1984, UT1-UTC= 0.0656464, Accumulated Leap Seconds=22
  { 2445925.50, 54.119378 }, //13 August 1984, UT1-UTC= 0.0646221, Accumulated Leap Seconds=22
  { 2445926.50, 54.120594 }, //14 August 1984, UT1-UTC= 0.0634058, Accumulated Leap Seconds=22
  { 2445927.50, 54.121963 }, //15 August 1984, UT1-UTC= 0.0620372, Accumulated Leap Seconds=22
  { 2445928.50, 54.123428 }, //16 August 1984, UT1-UTC= 0.0605719, Accumulated Leap Seconds=22
  { 2445929.50, 54.124924 }, //17 August 1984, UT1-UTC= 0.0590760, Accumulated Leap Seconds=22
  { 2445930.50, 54.126385 }, //18 August 1984, UT1-UTC= 0.0576148, Accumulated Leap Seconds=22
  { 2445931.50, 54.127755 }, //19 August 1984, UT1-UTC= 0.0562448, Accumulated Leap Seconds=22
  { 2445932.50, 54.128991 }, //20 August 1984, UT1-UTC= 0.0550085, Accumulated Leap Seconds=22
  { 2445933.50, 54.130070 }, //21 August 1984, UT1-UTC= 0.0539296, Accumulated Leap Seconds=22
  { 2445934.50, 54.130995 }, //22 August 1984, UT1-UTC= 0.0530050, Accumulated Leap Seconds=22
  { 2445935.50, 54.131810 }, //23 August 1984, UT1-UTC= 0.0521899, Accumulated Leap Seconds=22
  { 2445936.50, 54.132623 }, //24 August 1984, UT1-UTC= 0.0513765, Accumulated Leap Seconds=22
  { 2445937.50, 54.133576 }, //25 August 1984, UT1-UTC= 0.0504239, Accumulated Leap Seconds=22
  { 2445938.50, 54.134795 }, //26 August 1984, UT1-UTC= 0.0492045, Accumulated Leap Seconds=22
  { 2445939.50, 54.136349 }, //27 August 1984, UT1-UTC= 0.0476506, Accumulated Leap Seconds=22
  { 2445940.50, 54.138219 }, //28 August 1984, UT1-UTC= 0.0457809, Accumulated Leap Seconds=22
  { 2445941.50, 54.140298 }, //29 August 1984, UT1-UTC= 0.0437022, Accumulated Leap Seconds=22
  { 2445942.50, 54.142429 }, //30 August 1984, UT1-UTC= 0.0415708, Accumulated Leap Seconds=22
  { 2445943.50, 54.144463 }, //31 August 1984, UT1-UTC= 0.0395373, Accumulated Leap Seconds=22
  { 2445944.50, 54.146294 }, //01 September 1984, UT1-UTC= 0.0377058, Accumulated Leap Seconds=22
  { 2445945.50, 54.147871 }, //02 September 1984, UT1-UTC= 0.0361291, Accumulated Leap Seconds=22
  { 2445946.50, 54.149194 }, //03 September 1984, UT1-UTC= 0.0348056, Accumulated Leap Seconds=22
  { 2445947.50, 54.150311 }, //04 September 1984, UT1-UTC= 0.0336891, Accumulated Leap Seconds=22
  { 2445948.50, 54.151305 }, //05 September 1984, UT1-UTC= 0.0326946, Accumulated Leap Seconds=22
  { 2445949.50, 54.152262 }, //06 September 1984, UT1-UTC= 0.0317380, Accumulated Leap Seconds=22
  { 2445950.50, 54.153250 }, //07 September 1984, UT1-UTC= 0.0307500, Accumulated Leap Seconds=22
  { 2445951.50, 54.154326 }, //08 September 1984, UT1-UTC= 0.0296741, Accumulated Leap Seconds=22
  { 2445952.50, 54.155527 }, //09 September 1984, UT1-UTC= 0.0284734, Accumulated Leap Seconds=22
  { 2445953.50, 54.156865 }, //10 September 1984, UT1-UTC= 0.0271351, Accumulated Leap Seconds=22
  { 2445954.50, 54.158331 }, //11 September 1984, UT1-UTC= 0.0256687, Accumulated Leap Seconds=22
  { 2445955.50, 54.159895 }, //12 September 1984, UT1-UTC= 0.0241046, Accumulated Leap Seconds=22
  { 2445956.50, 54.161504 }, //13 September 1984, UT1-UTC= 0.0224955, Accumulated Leap Seconds=22
  { 2445957.50, 54.163100 }, //14 September 1984, UT1-UTC= 0.0208997, Accumulated Leap Seconds=22
  { 2445958.50, 54.164633 }, //15 September 1984, UT1-UTC= 0.0193671, Accumulated Leap Seconds=22
  { 2445959.50, 54.166066 }, //16 September 1984, UT1-UTC= 0.0179335, Accumulated Leap Seconds=22
  { 2445960.50, 54.167383 }, //17 September 1984, UT1-UTC= 0.0166165, Accumulated Leap Seconds=22
  { 2445961.50, 54.168586 }, //18 September 1984, UT1-UTC= 0.0154137, Accumulated Leap Seconds=22
  { 2445962.50, 54.169706 }, //19 September 1984, UT1-UTC= 0.0142939, Accumulated Leap Seconds=22
  { 2445963.50, 54.170809 }, //20 September 1984, UT1-UTC= 0.0131906, Accumulated Leap Seconds=22
  { 2445964.50, 54.171993 }, //21 September 1984, UT1-UTC= 0.0120066, Accumulated Leap Seconds=22
  { 2445965.50, 54.173371 }, //22 September 1984, UT1-UTC= 0.0106287, Accumulated Leap Seconds=22
  { 2445966.50, 54.175041 }, //23 September 1984, UT1-UTC= 0.0089587, Accumulated Leap Seconds=22
  { 2445967.50, 54.177038 }, //24 September 1984, UT1-UTC= 0.0069624, Accumulated Leap Seconds=22
  { 2445968.50, 54.179305 }, //25 September 1984, UT1-UTC= 0.0046949, Accumulated Leap Seconds=22
  { 2445969.50, 54.181713 }, //26 September 1984, UT1-UTC= 0.0022872, Accumulated Leap Seconds=22
  { 2445970.50, 54.184097 }, //27 September 1984, UT1-UTC=-0.0000970, Accumulated Leap Seconds=22
  { 2445971.50, 54.186305 }, //28 September 1984, UT1-UTC=-0.0023051, Accumulated Leap Seconds=22
  { 2445972.50, 54.188250 }, //29 September 1984, UT1-UTC=-0.0042499, Accumulated Leap Seconds=22
  { 2445973.50, 54.189928 }, //30 September 1984, UT1-UTC=-0.0059279, Accumulated Leap Seconds=22
  { 2445974.50, 54.191399 }, //01 October 1984, UT1-UTC=-0.0073988, Accumulated Leap Seconds=22
  { 2445975.50, 54.192753 }, //02 October 1984, UT1-UTC=-0.0087535, Accumulated Leap Seconds=22
  { 2445976.50, 54.194087 }, //03 October 1984, UT1-UTC=-0.0100868, Accumulated Leap Seconds=22
  { 2445977.50, 54.195477 }, //04 October 1984, UT1-UTC=-0.0114774, Accumulated Leap Seconds=22
  { 2445978.50, 54.196977 }, //05 October 1984, UT1-UTC=-0.0129767, Accumulated Leap Seconds=22
  { 2445979.50, 54.198609 }, //06 October 1984, UT1-UTC=-0.0146090, Accumulated Leap Seconds=22
  { 2445980.50, 54.200373 }, //07 October 1984, UT1-UTC=-0.0163734, Accumulated Leap Seconds=22
  { 2445981.50, 54.202248 }, //08 October 1984, UT1-UTC=-0.0182483, Accumulated Leap Seconds=22
  { 2445982.50, 54.204194 }, //09 October 1984, UT1-UTC=-0.0201941, Accumulated Leap Seconds=22
  { 2445983.50, 54.206161 }, //10 October 1984, UT1-UTC=-0.0221614, Accumulated Leap Seconds=22
  { 2445984.50, 54.208097 }, //11 October 1984, UT1-UTC=-0.0240968, Accumulated Leap Seconds=22
  { 2445985.50, 54.209950 }, //12 October 1984, UT1-UTC=-0.0259501, Accumulated Leap Seconds=22
  { 2445986.50, 54.211675 }, //13 October 1984, UT1-UTC=-0.0276754, Accumulated Leap Seconds=22
  { 2445987.50, 54.213245 }, //14 October 1984, UT1-UTC=-0.0292454, Accumulated Leap Seconds=22
  { 2445988.50, 54.214660 }, //15 October 1984, UT1-UTC=-0.0306600, Accumulated Leap Seconds=22
  { 2445989.50, 54.215953 }, //16 October 1984, UT1-UTC=-0.0319526, Accumulated Leap Seconds=22
  { 2445990.50, 54.217187 }, //17 October 1984, UT1-UTC=-0.0331875, Accumulated Leap Seconds=22
  { 2445991.50, 54.218460 }, //18 October 1984, UT1-UTC=-0.0344605, Accumulated Leap Seconds=22
  { 2445992.50, 54.219887 }, //19 October 1984, UT1-UTC=-0.0358872, Accumulated Leap Seconds=22
  { 2445993.50, 54.221579 }, //20 October 1984, UT1-UTC=-0.0375788, Accumulated Leap Seconds=22
  { 2445994.50, 54.223604 }, //21 October 1984, UT1-UTC=-0.0396037, Accumulated Leap Seconds=22
  { 2445995.50, 54.225953 }, //22 October 1984, UT1-UTC=-0.0419531, Accumulated Leap Seconds=22
  { 2445996.50, 54.228522 }, //23 October 1984, UT1-UTC=-0.0445216, Accumulated Leap Seconds=22
  { 2445997.50, 54.231139 }, //24 October 1984, UT1-UTC=-0.0471394, Accumulated Leap Seconds=22
  { 2445998.50, 54.233629 }, //25 October 1984, UT1-UTC=-0.0496295, Accumulated Leap Seconds=22
  { 2445999.50, 54.235862 }, //26 October 1984, UT1-UTC=-0.0518618, Accumulated Leap Seconds=22
  { 2446000.50, 54.237787 }, //27 October 1984, UT1-UTC=-0.0537872, Accumulated Leap Seconds=22
  { 2446001.50, 54.239439 }, //28 October 1984, UT1-UTC=-0.0554395, Accumulated Leap Seconds=22
  { 2446002.50, 54.240909 }, //29 October 1984, UT1-UTC=-0.0569086, Accumulated Leap Seconds=22
  { 2446003.50, 54.242301 }, //30 October 1984, UT1-UTC=-0.0583007, Accumulated Leap Seconds=22
  { 2446004.50, 54.243710 }, //31 October 1984, UT1-UTC=-0.0597100, Accumulated Leap Seconds=22
  { 2446005.50, 54.245202 }, //01 November 1984, UT1-UTC=-0.0612023, Accumulated Leap Seconds=22
  { 2446006.50, 54.246813 }, //02 November 1984, UT1-UTC=-0.0628134, Accumulated Leap Seconds=22
  { 2446007.50, 54.248552 }, //03 November 1984, UT1-UTC=-0.0645525, Accumulated Leap Seconds=22
  { 2446008.50, 54.250406 }, //04 November 1984, UT1-UTC=-0.0664059, Accumulated Leap Seconds=22
  { 2446009.50, 54.252343 }, //05 November 1984, UT1-UTC=-0.0683428, Accumulated Leap Seconds=22
  { 2446010.50, 54.254318 }, //06 November 1984, UT1-UTC=-0.0703182, Accumulated Leap Seconds=22
  { 2446011.50, 54.256269 }, //07 November 1984, UT1-UTC=-0.0722694, Accumulated Leap Seconds=22
  { 2446012.50, 54.258131 }, //08 November 1984, UT1-UTC=-0.0741305, Accumulated Leap Seconds=22
  { 2446013.50, 54.259845 }, //09 November 1984, UT1-UTC=-0.0758451, Accumulated Leap Seconds=22
  { 2446014.50, 54.261375 }, //10 November 1984, UT1-UTC=-0.0773755, Accumulated Leap Seconds=22
  { 2446015.50, 54.262714 }, //11 November 1984, UT1-UTC=-0.0787138, Accumulated Leap Seconds=22
  { 2446016.50, 54.263897 }, //12 November 1984, UT1-UTC=-0.0798973, Accumulated Leap Seconds=22
  { 2446017.50, 54.265000 }, //13 November 1984, UT1-UTC=-0.0809999, Accumulated Leap Seconds=22
  { 2446018.50, 54.266118 }, //14 November 1984, UT1-UTC=-0.0821181, Accumulated Leap Seconds=22
  { 2446019.50, 54.267356 }, //15 November 1984, UT1-UTC=-0.0833560, Accumulated Leap Seconds=22
  { 2446020.50, 54.268811 }, //16 November 1984, UT1-UTC=-0.0848107, Accumulated Leap Seconds=22
  { 2446021.50, 54.270553 }, //17 November 1984, UT1-UTC=-0.0865525, Accumulated Leap Seconds=22
  { 2446022.50, 54.272595 }, //18 November 1984, UT1-UTC=-0.0885955, Accumulated Leap Seconds=22
  { 2446023.50, 54.274886 }, //19 November 1984, UT1-UTC=-0.0908858, Accumulated Leap Seconds=22
  { 2446024.50, 54.277308 }, //20 November 1984, UT1-UTC=-0.0933085, Accumulated Leap Seconds=22
  { 2446025.50, 54.279712 }, //21 November 1984, UT1-UTC=-0.0957120, Accumulated Leap Seconds=22
  { 2446026.50, 54.281949 }, //22 November 1984, UT1-UTC=-0.0979486, Accumulated Leap Seconds=22
  { 2446027.50, 54.283924 }, //23 November 1984, UT1-UTC=-0.0999235, Accumulated Leap Seconds=22
  { 2446028.50, 54.285617 }, //24 November 1984, UT1-UTC=-0.1016168, Accumulated Leap Seconds=22
  { 2446029.50, 54.287078 }, //25 November 1984, UT1-UTC=-0.1030785, Accumulated Leap Seconds=22
  { 2446030.50, 54.288400 }, //26 November 1984, UT1-UTC=-0.1043996, Accumulated Leap Seconds=22
  { 2446031.50, 54.289687 }, //27 November 1984, UT1-UTC=-0.1056874, Accumulated Leap Seconds=22
  { 2446032.50, 54.291030 }, //28 November 1984, UT1-UTC=-0.1070300, Accumulated Leap Seconds=22
  { 2446033.50, 54.292480 }, //29 November 1984, UT1-UTC=-0.1084801, Accumulated Leap Seconds=22
  { 2446034.50, 54.294059 }, //30 November 1984, UT1-UTC=-0.1100592, Accumulated Leap Seconds=22
  { 2446035.50, 54.295762 }, //01 December 1984, UT1-UTC=-0.1117622, Accumulated Leap Seconds=22
  { 2446036.50, 54.297560 }, //02 December 1984, UT1-UTC=-0.1135605, Accumulated Leap Seconds=22
  { 2446037.50, 54.299411 }, //03 December 1984, UT1-UTC=-0.1154113, Accumulated Leap Seconds=22
  { 2446038.50, 54.301264 }, //04 December 1984, UT1-UTC=-0.1172637, Accumulated Leap Seconds=22
  { 2446039.50, 54.303066 }, //05 December 1984, UT1-UTC=-0.1190658, Accumulated Leap Seconds=22
  { 2446040.50, 54.304772 }, //06 December 1984, UT1-UTC=-0.1207718, Accumulated Leap Seconds=22
  { 2446041.50, 54.306342 }, //07 December 1984, UT1-UTC=-0.1223422, Accumulated Leap Seconds=22
  { 2446042.50, 54.307756 }, //08 December 1984, UT1-UTC=-0.1237556, Accumulated Leap Seconds=22
  { 2446043.50, 54.309023 }, //09 December 1984, UT1-UTC=-0.1250232, Accumulated Leap Seconds=22
  { 2446044.50, 54.310190 }, //10 December 1984, UT1-UTC=-0.1261900, Accumulated Leap Seconds=22
  { 2446045.50, 54.311330 }, //11 December 1984, UT1-UTC=-0.1273298, Accumulated Leap Seconds=22
  { 2446046.50, 54.312531 }, //12 December 1984, UT1-UTC=-0.1285315, Accumulated Leap Seconds=22
  { 2446047.50, 54.313880 }, //13 December 1984, UT1-UTC=-0.1298801, Accumulated Leap Seconds=22
  { 2446048.50, 54.315431 }, //14 December 1984, UT1-UTC=-0.1314314, Accumulated Leap Seconds=22
  { 2446049.50, 54.317196 }, //15 December 1984, UT1-UTC=-0.1331958, Accumulated Leap Seconds=22
  { 2446050.50, 54.319130 }, //16 December 1984, UT1-UTC=-0.1351298, Accumulated Leap Seconds=22
  { 2446051.50, 54.321141 }, //17 December 1984, UT1-UTC=-0.1371409, Accumulated Leap Seconds=22
  { 2446052.50, 54.323108 }, //18 December 1984, UT1-UTC=-0.1391081, Accumulated Leap Seconds=22
  { 2446053.50, 54.324911 }, //19 December 1984, UT1-UTC=-0.1409111, Accumulated Leap Seconds=22
  { 2446054.50, 54.326466 }, //20 December 1984, UT1-UTC=-0.1424658, Accumulated Leap Seconds=22
  { 2446055.50, 54.327750 }, //21 December 1984, UT1-UTC=-0.1437496, Accumulated Leap Seconds=22
  { 2446056.50, 54.328806 }, //22 December 1984, UT1-UTC=-0.1448064, Accumulated Leap Seconds=22
  { 2446057.50, 54.329726 }, //23 December 1984, UT1-UTC=-0.1457265, Accumulated Leap Seconds=22
  { 2446058.50, 54.330619 }, //24 December 1984, UT1-UTC=-0.1466187, Accumulated Leap Seconds=22
  { 2446059.50, 54.331590 }, //25 December 1984, UT1-UTC=-0.1475905, Accumulated Leap Seconds=22
  { 2446060.50, 54.332718 }, //26 December 1984, UT1-UTC=-0.1487180, Accumulated Leap Seconds=22
  { 2446061.50, 54.334034 }, //27 December 1984, UT1-UTC=-0.1500341, Accumulated Leap Seconds=22
  { 2446062.50, 54.335539 }, //28 December 1984, UT1-UTC=-0.1515390, Accumulated Leap Seconds=22
  { 2446063.50, 54.337208 }, //29 December 1984, UT1-UTC=-0.1532078, Accumulated Leap Seconds=22
  { 2446064.50, 54.338996 }, //30 December 1984, UT1-UTC=-0.1549960, Accumulated Leap Seconds=22
  { 2446065.50, 54.340848 }, //31 December 1984, UT1-UTC=-0.1568485, Accumulated Leap Seconds=22
  { 2446066.50, 54.342702 }, //01 January 1985, UT1-UTC=-0.1587024, Accumulated Leap Seconds=22
  { 2446067.50, 54.344499 }, //02 January 1985, UT1-UTC=-0.1604985, Accumulated Leap Seconds=22
  { 2446068.50, 54.346190 }, //03 January 1985, UT1-UTC=-0.1621898, Accumulated Leap Seconds=22
  { 2446069.50, 54.347747 }, //04 January 1985, UT1-UTC=-0.1637475, Accumulated Leap Seconds=22
  { 2446070.50, 54.349169 }, //05 January 1985, UT1-UTC=-0.1651692, Accumulated Leap Seconds=22
  { 2446071.50, 54.350489 }, //06 January 1985, UT1-UTC=-0.1664894, Accumulated Leap Seconds=22
  { 2446072.50, 54.351779 }, //07 January 1985, UT1-UTC=-0.1677786, Accumulated Leap Seconds=22
  { 2446073.50, 54.353131 }, //08 January 1985, UT1-UTC=-0.1691310, Accumulated Leap Seconds=22
  { 2446074.50, 54.354642 }, //09 January 1985, UT1-UTC=-0.1706416, Accumulated Leap Seconds=22
  { 2446075.50, 54.356378 }, //10 January 1985, UT1-UTC=-0.1723776, Accumulated Leap Seconds=22
  { 2446076.50, 54.358358 }, //11 January 1985, UT1-UTC=-0.1743582, Accumulated Leap Seconds=22
  { 2446077.50, 54.360541 }, //12 January 1985, UT1-UTC=-0.1765408, Accumulated Leap Seconds=22
  { 2446078.50, 54.362828 }, //13 January 1985, UT1-UTC=-0.1788282, Accumulated Leap Seconds=22
  { 2446079.50, 54.365096 }, //14 January 1985, UT1-UTC=-0.1810959, Accumulated Leap Seconds=22
  { 2446080.50, 54.367224 }, //15 January 1985, UT1-UTC=-0.1832243, Accumulated Leap Seconds=22
  { 2446081.50, 54.369122 }, //16 January 1985, UT1-UTC=-0.1851221, Accumulated Leap Seconds=22
  { 2446082.50, 54.370749 }, //17 January 1985, UT1-UTC=-0.1867494, Accumulated Leap Seconds=22
  { 2446083.50, 54.372125 }, //18 January 1985, UT1-UTC=-0.1881254, Accumulated Leap Seconds=22
  { 2446084.50, 54.373317 }, //19 January 1985, UT1-UTC=-0.1893172, Accumulated Leap Seconds=22
  { 2446085.50, 54.374420 }, //20 January 1985, UT1-UTC=-0.1904198, Accumulated Leap Seconds=22
  { 2446086.50, 54.375532 }, //21 January 1985, UT1-UTC=-0.1915325, Accumulated Leap Seconds=22
  { 2446087.50, 54.376732 }, //22 January 1985, UT1-UTC=-0.1927319, Accumulated Leap Seconds=22
  { 2446088.50, 54.378059 }, //23 January 1985, UT1-UTC=-0.1940590, Accumulated Leap Seconds=22
  { 2446089.50, 54.379518 }, //24 January 1985, UT1-UTC=-0.1955181, Accumulated Leap Seconds=22
  { 2446090.50, 54.381084 }, //25 January 1985, UT1-UTC=-0.1970843, Accumulated Leap Seconds=22
  { 2446091.50, 54.382714 }, //26 January 1985, UT1-UTC=-0.1987141, Accumulated Leap Seconds=22
  { 2446092.50, 54.384354 }, //27 January 1985, UT1-UTC=-0.2003543, Accumulated Leap Seconds=22
  { 2446093.50, 54.385950 }, //28 January 1985, UT1-UTC=-0.2019500, Accumulated Leap Seconds=22
  { 2446094.50, 54.387452 }, //29 January 1985, UT1-UTC=-0.2034525, Accumulated Leap Seconds=22
  { 2446095.50, 54.388825 }, //30 January 1985, UT1-UTC=-0.2048253, Accumulated Leap Seconds=22
  { 2446096.50, 54.390039 }, //31 January 1985, UT1-UTC=-0.2060393, Accumulated Leap Seconds=22
  { 2446097.50, 54.391086 }, //01 February 1985, UT1-UTC=-0.2070864, Accumulated Leap Seconds=22
  { 2446098.50, 54.391988 }, //02 February 1985, UT1-UTC=-0.2079876, Accumulated Leap Seconds=22
  { 2446099.50, 54.392799 }, //03 February 1985, UT1-UTC=-0.2087992, Accumulated Leap Seconds=22
  { 2446100.50, 54.393614 }, //04 February 1985, UT1-UTC=-0.2096142, Accumulated Leap Seconds=22
  { 2446101.50, 54.394551 }, //05 February 1985, UT1-UTC=-0.2105515, Accumulated Leap Seconds=22
  { 2446102.50, 54.395717 }, //06 February 1985, UT1-UTC=-0.2117173, Accumulated Leap Seconds=22
  { 2446103.50, 54.397169 }, //07 February 1985, UT1-UTC=-0.2131689, Accumulated Leap Seconds=22
  { 2446104.50, 54.398890 }, //08 February 1985, UT1-UTC=-0.2148902, Accumulated Leap Seconds=22
  { 2446105.50, 54.400793 }, //09 February 1985, UT1-UTC=-0.2167934, Accumulated Leap Seconds=22
  { 2446106.50, 54.402743 }, //10 February 1985, UT1-UTC=-0.2187426, Accumulated Leap Seconds=22
  { 2446107.50, 54.404600 }, //11 February 1985, UT1-UTC=-0.2206003, Accumulated Leap Seconds=22
  { 2446108.50, 54.406269 }, //12 February 1985, UT1-UTC=-0.2222693, Accumulated Leap Seconds=22
  { 2446109.50, 54.407713 }, //13 February 1985, UT1-UTC=-0.2237131, Accumulated Leap Seconds=22
  { 2446110.50, 54.408958 }, //14 February 1985, UT1-UTC=-0.2249583, Accumulated Leap Seconds=22
  { 2446111.50, 54.410078 }, //15 February 1985, UT1-UTC=-0.2260776, Accumulated Leap Seconds=22
  { 2446112.50, 54.411166 }, //16 February 1985, UT1-UTC=-0.2271657, Accumulated Leap Seconds=22
  { 2446113.50, 54.412317 }, //17 February 1985, UT1-UTC=-0.2283172, Accumulated Leap Seconds=22
  { 2446114.50, 54.413607 }, //18 February 1985, UT1-UTC=-0.2296066, Accumulated Leap Seconds=22
  { 2446115.50, 54.415077 }, //19 February 1985, UT1-UTC=-0.2310769, Accumulated Leap Seconds=22
  { 2446116.50, 54.416733 }, //20 February 1985, UT1-UTC=-0.2327332, Accumulated Leap Seconds=22
  { 2446117.50, 54.418546 }, //21 February 1985, UT1-UTC=-0.2345458, Accumulated Leap Seconds=22
  { 2446118.50, 54.420463 }, //22 February 1985, UT1-UTC=-0.2364631, Accumulated Leap Seconds=22
  { 2446119.50, 54.422419 }, //23 February 1985, UT1-UTC=-0.2384191, Accumulated Leap Seconds=22
  { 2446120.50, 54.424345 }, //24 February 1985, UT1-UTC=-0.2403446, Accumulated Leap Seconds=22
  { 2446121.50, 54.426179 }, //25 February 1985, UT1-UTC=-0.2421794, Accumulated Leap Seconds=22
  { 2446122.50, 54.427877 }, //26 February 1985, UT1-UTC=-0.2438770, Accumulated Leap Seconds=22
  { 2446123.50, 54.429410 }, //27 February 1985, UT1-UTC=-0.2454103, Accumulated Leap Seconds=22
  { 2446124.50, 54.430774 }, //28 February 1985, UT1-UTC=-0.2467737, Accumulated Leap Seconds=22
  { 2446125.50, 54.431988 }, //01 March 1985, UT1-UTC=-0.2479877, Accumulated Leap Seconds=22
  { 2446126.50, 54.433110 }, //02 March 1985, UT1-UTC=-0.2491099, Accumulated Leap Seconds=22
  { 2446127.50, 54.434233 }, //03 March 1985, UT1-UTC=-0.2502332, Accumulated Leap Seconds=22
  { 2446128.50, 54.435476 }, //04 March 1985, UT1-UTC=-0.2514761, Accumulated Leap Seconds=22
  { 2446129.50, 54.436962 }, //05 March 1985, UT1-UTC=-0.2529621, Accumulated Leap Seconds=22
  { 2446130.50, 54.438784 }, //06 March 1985, UT1-UTC=-0.2547844, Accumulated Leap Seconds=22
  { 2446131.50, 54.440973 }, //07 March 1985, UT1-UTC=-0.2569729, Accumulated Leap Seconds=22
  { 2446132.50, 54.443455 }, //08 March 1985, UT1-UTC=-0.2594551, Accumulated Leap Seconds=22
  { 2446133.50, 54.446082 }, //09 March 1985, UT1-UTC=-0.2620822, Accumulated Leap Seconds=22
  { 2446134.50, 54.448679 }, //10 March 1985, UT1-UTC=-0.2646794, Accumulated Leap Seconds=22
  { 2446135.50, 54.451102 }, //11 March 1985, UT1-UTC=-0.2671017, Accumulated Leap Seconds=22
  { 2446136.50, 54.453270 }, //12 March 1985, UT1-UTC=-0.2692698, Accumulated Leap Seconds=22
  { 2446137.50, 54.455181 }, //13 March 1985, UT1-UTC=-0.2711812, Accumulated Leap Seconds=22
  { 2446138.50, 54.456898 }, //14 March 1985, UT1-UTC=-0.2728985, Accumulated Leap Seconds=22
  { 2446139.50, 54.458518 }, //15 March 1985, UT1-UTC=-0.2745181, Accumulated Leap Seconds=22
  { 2446140.50, 54.460138 }, //16 March 1985, UT1-UTC=-0.2761381, Accumulated Leap Seconds=22
  { 2446141.50, 54.461839 }, //17 March 1985, UT1-UTC=-0.2778392, Accumulated Leap Seconds=22
  { 2446142.50, 54.463672 }, //18 March 1985, UT1-UTC=-0.2796716, Accumulated Leap Seconds=22
  { 2446143.50, 54.465652 }, //19 March 1985, UT1-UTC=-0.2816519, Accumulated Leap Seconds=22
  { 2446144.50, 54.467765 }, //20 March 1985, UT1-UTC=-0.2837655, Accumulated Leap Seconds=22
  { 2446145.50, 54.469972 }, //21 March 1985, UT1-UTC=-0.2859721, Accumulated Leap Seconds=22
  { 2446146.50, 54.472218 }, //22 March 1985, UT1-UTC=-0.2882180, Accumulated Leap Seconds=22
  { 2446147.50, 54.474445 }, //23 March 1985, UT1-UTC=-0.2904451, Accumulated Leap Seconds=22
  { 2446148.50, 54.476600 }, //24 March 1985, UT1-UTC=-0.2925996, Accumulated Leap Seconds=22
  { 2446149.50, 54.478639 }, //25 March 1985, UT1-UTC=-0.2946393, Accumulated Leap Seconds=22
  { 2446150.50, 54.480539 }, //26 March 1985, UT1-UTC=-0.2965388, Accumulated Leap Seconds=22
  { 2446151.50, 54.482284 }, //27 March 1985, UT1-UTC=-0.2982838, Accumulated Leap Seconds=22
  { 2446152.50, 54.483880 }, //28 March 1985, UT1-UTC=-0.2998803, Accumulated Leap Seconds=22
  { 2446153.50, 54.485359 }, //29 March 1985, UT1-UTC=-0.3013587, Accumulated Leap Seconds=22
  { 2446154.50, 54.486774 }, //30 March 1985, UT1-UTC=-0.3027741, Accumulated Leap Seconds=22
  { 2446155.50, 54.488208 }, //31 March 1985, UT1-UTC=-0.3042083, Accumulated Leap Seconds=22
  { 2446156.50, 54.489770 }, //01 April 1985, UT1-UTC=-0.3057699, Accumulated Leap Seconds=22
  { 2446157.50, 54.491560 }, //02 April 1985, UT1-UTC=-0.3075603, Accumulated Leap Seconds=22
  { 2446158.50, 54.493640 }, //03 April 1985, UT1-UTC=-0.3096395, Accumulated Leap Seconds=22
  { 2446159.50, 54.495992 }, //04 April 1985, UT1-UTC=-0.3119925, Accumulated Leap Seconds=22
  { 2446160.50, 54.498521 }, //05 April 1985, UT1-UTC=-0.3145206, Accumulated Leap Seconds=22
  { 2446161.50, 54.501058 }, //06 April 1985, UT1-UTC=-0.3170582, Accumulated Leap Seconds=22
  { 2446162.50, 54.503431 }, //07 April 1985, UT1-UTC=-0.3194308, Accumulated Leap Seconds=22
  { 2446163.50, 54.505518 }, //08 April 1985, UT1-UTC=-0.3215179, Accumulated Leap Seconds=22
  { 2446164.50, 54.507288 }, //09 April 1985, UT1-UTC=-0.3232883, Accumulated Leap Seconds=22
  { 2446165.50, 54.508799 }, //10 April 1985, UT1-UTC=-0.3247995, Accumulated Leap Seconds=22
  { 2446166.50, 54.510167 }, //11 April 1985, UT1-UTC=-0.3261666, Accumulated Leap Seconds=22
  { 2446167.50, 54.511518 }, //12 April 1985, UT1-UTC=-0.3275176, Accumulated Leap Seconds=22
  { 2446168.50, 54.512960 }, //13 April 1985, UT1-UTC=-0.3289597, Accumulated Leap Seconds=22
  { 2446169.50, 54.514561 }, //14 April 1985, UT1-UTC=-0.3305615, Accumulated Leap Seconds=22
  { 2446170.50, 54.516352 }, //15 April 1985, UT1-UTC=-0.3323518, Accumulated Leap Seconds=22
  { 2446171.50, 54.518322 }, //16 April 1985, UT1-UTC=-0.3343216, Accumulated Leap Seconds=22
  { 2446172.50, 54.520434 }, //17 April 1985, UT1-UTC=-0.3364336, Accumulated Leap Seconds=22
  { 2446173.50, 54.522632 }, //18 April 1985, UT1-UTC=-0.3386320, Accumulated Leap Seconds=22
  { 2446174.50, 54.524850 }, //19 April 1985, UT1-UTC=-0.3408504, Accumulated Leap Seconds=22
  { 2446175.50, 54.527023 }, //20 April 1985, UT1-UTC=-0.3430228, Accumulated Leap Seconds=22
  { 2446176.50, 54.529089 }, //21 April 1985, UT1-UTC=-0.3450892, Accumulated Leap Seconds=22
  { 2446177.50, 54.531002 }, //22 April 1985, UT1-UTC=-0.3470025, Accumulated Leap Seconds=22
  { 2446178.50, 54.532740 }, //23 April 1985, UT1-UTC=-0.3487397, Accumulated Leap Seconds=22
  { 2446179.50, 54.534305 }, //24 April 1985, UT1-UTC=-0.3503052, Accumulated Leap Seconds=22
  { 2446180.50, 54.535732 }, //25 April 1985, UT1-UTC=-0.3517320, Accumulated Leap Seconds=22
  { 2446181.50, 54.537082 }, //26 April 1985, UT1-UTC=-0.3530818, Accumulated Leap Seconds=22
  { 2446182.50, 54.538435 }, //27 April 1985, UT1-UTC=-0.3544347, Accumulated Leap Seconds=22
  { 2446183.50, 54.539885 }, //28 April 1985, UT1-UTC=-0.3558846, Accumulated Leap Seconds=22
  { 2446184.50, 54.541526 }, //29 April 1985, UT1-UTC=-0.3575265, Accumulated Leap Seconds=22
  { 2446185.50, 54.543434 }, //30 April 1985, UT1-UTC=-0.3594340, Accumulated Leap Seconds=22
  { 2446186.50, 54.545636 }, //01 May 1985, UT1-UTC=-0.3616360, Accumulated Leap Seconds=22
  { 2446187.50, 54.548084 }, //02 May 1985, UT1-UTC=-0.3640839, Accumulated Leap Seconds=22
  { 2446188.50, 54.550653 }, //03 May 1985, UT1-UTC=-0.3666535, Accumulated Leap Seconds=22
  { 2446189.50, 54.553181 }, //04 May 1985, UT1-UTC=-0.3691808, Accumulated Leap Seconds=22
  { 2446190.50, 54.555513 }, //05 May 1985, UT1-UTC=-0.3715135, Accumulated Leap Seconds=22
  { 2446191.50, 54.557551 }, //06 May 1985, UT1-UTC=-0.3735509, Accumulated Leap Seconds=22
  { 2446192.50, 54.559281 }, //07 May 1985, UT1-UTC=-0.3752815, Accumulated Leap Seconds=22
  { 2446193.50, 54.560775 }, //08 May 1985, UT1-UTC=-0.3767753, Accumulated Leap Seconds=22
  { 2446194.50, 54.562145 }, //09 May 1985, UT1-UTC=-0.3781449, Accumulated Leap Seconds=22
  { 2446195.50, 54.563505 }, //10 May 1985, UT1-UTC=-0.3795045, Accumulated Leap Seconds=22
  { 2446196.50, 54.564927 }, //11 May 1985, UT1-UTC=-0.3809268, Accumulated Leap Seconds=22
  { 2446197.50, 54.566457 }, //12 May 1985, UT1-UTC=-0.3824566, Accumulated Leap Seconds=22
  { 2446198.50, 54.568105 }, //13 May 1985, UT1-UTC=-0.3841048, Accumulated Leap Seconds=22
  { 2446199.50, 54.569855 }, //14 May 1985, UT1-UTC=-0.3858549, Accumulated Leap Seconds=22
  { 2446200.50, 54.571671 }, //15 May 1985, UT1-UTC=-0.3876707, Accumulated Leap Seconds=22
  { 2446201.50, 54.573494 }, //16 May 1985, UT1-UTC=-0.3894939, Accumulated Leap Seconds=22
  { 2446202.50, 54.575259 }, //17 May 1985, UT1-UTC=-0.3912588, Accumulated Leap Seconds=22
  { 2446203.50, 54.576911 }, //18 May 1985, UT1-UTC=-0.3929113, Accumulated Leap Seconds=22
  { 2446204.50, 54.578410 }, //19 May 1985, UT1-UTC=-0.3944104, Accumulated Leap Seconds=22
  { 2446205.50, 54.579736 }, //20 May 1985, UT1-UTC=-0.3957362, Accumulated Leap Seconds=22
  { 2446206.50, 54.580893 }, //21 May 1985, UT1-UTC=-0.3968926, Accumulated Leap Seconds=22
  { 2446207.50, 54.581908 }, //22 May 1985, UT1-UTC=-0.3979085, Accumulated Leap Seconds=22
  { 2446208.50, 54.582838 }, //23 May 1985, UT1-UTC=-0.3988384, Accumulated Leap Seconds=22
  { 2446209.50, 54.583758 }, //24 May 1985, UT1-UTC=-0.3997580, Accumulated Leap Seconds=22
  { 2446210.50, 54.584756 }, //25 May 1985, UT1-UTC=-0.4007565, Accumulated Leap Seconds=22
  { 2446211.50, 54.585931 }, //26 May 1985, UT1-UTC=-0.4019311, Accumulated Leap Seconds=22
  { 2446212.50, 54.587363 }, //27 May 1985, UT1-UTC=-0.4033630, Accumulated Leap Seconds=22
  { 2446213.50, 54.589096 }, //28 May 1985, UT1-UTC=-0.4050957, Accumulated Leap Seconds=22
  { 2446214.50, 54.591113 }, //29 May 1985, UT1-UTC=-0.4071132, Accumulated Leap Seconds=22
  { 2446215.50, 54.593330 }, //30 May 1985, UT1-UTC=-0.4093302, Accumulated Leap Seconds=22
  { 2446216.50, 54.595596 }, //31 May 1985, UT1-UTC=-0.4115962, Accumulated Leap Seconds=22
  { 2446217.50, 54.597741 }, //01 June 1985, UT1-UTC=-0.4137410, Accumulated Leap Seconds=22
  { 2446218.50, 54.599632 }, //02 June 1985, UT1-UTC=-0.4156320, Accumulated Leap Seconds=22
  { 2446219.50, 54.601212 }, //03 June 1985, UT1-UTC=-0.4172123, Accumulated Leap Seconds=22
  { 2446220.50, 54.602513 }, //04 June 1985, UT1-UTC=-0.4185133, Accumulated Leap Seconds=22
  { 2446221.50, 54.603641 }, //05 June 1985, UT1-UTC=-0.4196413, Accumulated Leap Seconds=22
  { 2446222.50, 54.604725 }, //06 June 1985, UT1-UTC=-0.4207246, Accumulated Leap Seconds=22
  { 2446223.50, 54.605867 }, //07 June 1985, UT1-UTC=-0.4218666, Accumulated Leap Seconds=22
  { 2446224.50, 54.607123 }, //08 June 1985, UT1-UTC=-0.4231233, Accumulated Leap Seconds=22
  { 2446225.50, 54.608503 }, //09 June 1985, UT1-UTC=-0.4245033, Accumulated Leap Seconds=22
  { 2446226.50, 54.609984 }, //10 June 1985, UT1-UTC=-0.4259840, Accumulated Leap Seconds=22
  { 2446227.50, 54.611521 }, //11 June 1985, UT1-UTC=-0.4275212, Accumulated Leap Seconds=22
  { 2446228.50, 54.613061 }, //12 June 1985, UT1-UTC=-0.4290608, Accumulated Leap Seconds=22
  { 2446229.50, 54.614546 }, //13 June 1985, UT1-UTC=-0.4305462, Accumulated Leap Seconds=22
  { 2446230.50, 54.615925 }, //14 June 1985, UT1-UTC=-0.4319251, Accumulated Leap Seconds=22
  { 2446231.50, 54.617157 }, //15 June 1985, UT1-UTC=-0.4331574, Accumulated Leap Seconds=22
  { 2446232.50, 54.618224 }, //16 June 1985, UT1-UTC=-0.4342243, Accumulated Leap Seconds=22
  { 2446233.50, 54.619132 }, //17 June 1985, UT1-UTC=-0.4351318, Accumulated Leap Seconds=22
  { 2446234.50, 54.619913 }, //18 June 1985, UT1-UTC=-0.4359130, Accumulated Leap Seconds=22
  { 2446235.50, 54.620626 }, //19 June 1985, UT1-UTC=-0.4366264, Accumulated Leap Seconds=22
  { 2446236.50, 54.621314 }, //20 June 1985, UT1-UTC=-0.4373143, Accumulated Leap Seconds=22
  { 2446237.50, 54.622027 }, //21 June 1985, UT1-UTC=-0.4380268, Accumulated Leap Seconds=22
  { 2446238.50, 54.622867 }, //22 June 1985, UT1-UTC=-0.4388666, Accumulated Leap Seconds=22
  { 2446239.50, 54.623931 }, //23 June 1985, UT1-UTC=-0.4399306, Accumulated Leap Seconds=22
  { 2446240.50, 54.625274 }, //24 June 1985, UT1-UTC=-0.4412745, Accumulated Leap Seconds=22
  { 2446241.50, 54.626873 }, //25 June 1985, UT1-UTC=-0.4428733, Accumulated Leap Seconds=22
  { 2446242.50, 54.628637 }, //26 June 1985, UT1-UTC=-0.4446372, Accumulated Leap Seconds=22
  { 2446243.50, 54.630436 }, //27 June 1985, UT1-UTC=-0.4464363, Accumulated Leap Seconds=22
  { 2446244.50, 54.632124 }, //28 June 1985, UT1-UTC=-0.4481239, Accumulated Leap Seconds=22
  { 2446245.50, 54.633570 }, //29 June 1985, UT1-UTC=-0.4495697, Accumulated Leap Seconds=22
  { 2446246.50, 54.634695 }, //30 June 1985, UT1-UTC=-0.4506950, Accumulated Leap Seconds=22
  { 2446247.50, 54.635496 }, //01 July 1985, UT1-UTC= 0.5485038, Accumulated Leap Seconds=23
  { 2446248.50, 54.636043 }, //02 July 1985, UT1-UTC= 0.5479567, Accumulated Leap Seconds=23
  { 2446249.50, 54.636448 }, //03 July 1985, UT1-UTC= 0.5475517, Accumulated Leap Seconds=23
  { 2446250.50, 54.636829 }, //04 July 1985, UT1-UTC= 0.5471715, Accumulated Leap Seconds=23
  { 2446251.50, 54.637276 }, //05 July 1985, UT1-UTC= 0.5467238, Accumulated Leap Seconds=23
  { 2446252.50, 54.637838 }, //06 July 1985, UT1-UTC= 0.5461623, Accumulated Leap Seconds=23
  { 2446253.50, 54.638511 }, //07 July 1985, UT1-UTC= 0.5454889, Accumulated Leap Seconds=23
  { 2446254.50, 54.639260 }, //08 July 1985, UT1-UTC= 0.5447398, Accumulated Leap Seconds=23
  { 2446255.50, 54.640032 }, //09 July 1985, UT1-UTC= 0.5439675, Accumulated Leap Seconds=23
  { 2446256.50, 54.640780 }, //10 July 1985, UT1-UTC= 0.5432201, Accumulated Leap Seconds=23
  { 2446257.50, 54.641459 }, //11 July 1985, UT1-UTC= 0.5425412, Accumulated Leap Seconds=23
  { 2446258.50, 54.642034 }, //12 July 1985, UT1-UTC= 0.5419664, Accumulated Leap Seconds=23
  { 2446259.50, 54.642480 }, //13 July 1985, UT1-UTC= 0.5415204, Accumulated Leap Seconds=23
  { 2446260.50, 54.642795 }, //14 July 1985, UT1-UTC= 0.5412047, Accumulated Leap Seconds=23
  { 2446261.50, 54.642999 }, //15 July 1985, UT1-UTC= 0.5410011, Accumulated Leap Seconds=23
  { 2446262.50, 54.643126 }, //16 July 1985, UT1-UTC= 0.5408741, Accumulated Leap Seconds=23
  { 2446263.50, 54.643235 }, //17 July 1985, UT1-UTC= 0.5407652, Accumulated Leap Seconds=23
  { 2446264.50, 54.643405 }, //18 July 1985, UT1-UTC= 0.5405949, Accumulated Leap Seconds=23
  { 2446265.50, 54.643725 }, //19 July 1985, UT1-UTC= 0.5402746, Accumulated Leap Seconds=23
  { 2446266.50, 54.644279 }, //20 July 1985, UT1-UTC= 0.5397209, Accumulated Leap Seconds=23
  { 2446267.50, 54.645110 }, //21 July 1985, UT1-UTC= 0.5388898, Accumulated Leap Seconds=23
  { 2446268.50, 54.646198 }, //22 July 1985, UT1-UTC= 0.5378020, Accumulated Leap Seconds=23
  { 2446269.50, 54.647453 }, //23 July 1985, UT1-UTC= 0.5365473, Accumulated Leap Seconds=23
  { 2446270.50, 54.648748 }, //24 July 1985, UT1-UTC= 0.5352523, Accumulated Leap Seconds=23
  { 2446271.50, 54.649949 }, //25 July 1985, UT1-UTC= 0.5340505, Accumulated Leap Seconds=23
  { 2446272.50, 54.650947 }, //26 July 1985, UT1-UTC= 0.5330528, Accumulated Leap Seconds=23
  { 2446273.50, 54.651678 }, //27 July 1985, UT1-UTC= 0.5323222, Accumulated Leap Seconds=23
  { 2446274.50, 54.652145 }, //28 July 1985, UT1-UTC= 0.5318552, Accumulated Leap Seconds=23
  { 2446275.50, 54.652418 }, //29 July 1985, UT1-UTC= 0.5315821, Accumulated Leap Seconds=23
  { 2446276.50, 54.652615 }, //30 July 1985, UT1-UTC= 0.5313851, Accumulated Leap Seconds=23
  { 2446277.50, 54.652856 }, //31 July 1985, UT1-UTC= 0.5311439, Accumulated Leap Seconds=23
  { 2446278.50, 54.653235 }, //01 August 1985, UT1-UTC= 0.5307648, Accumulated Leap Seconds=23
  { 2446279.50, 54.653803 }, //02 August 1985, UT1-UTC= 0.5301969, Accumulated Leap Seconds=23
  { 2446280.50, 54.654562 }, //03 August 1985, UT1-UTC= 0.5294384, Accumulated Leap Seconds=23
  { 2446281.50, 54.655468 }, //04 August 1985, UT1-UTC= 0.5285317, Accumulated Leap Seconds=23
  { 2446282.50, 54.656455 }, //05 August 1985, UT1-UTC= 0.5275450, Accumulated Leap Seconds=23
  { 2446283.50, 54.657447 }, //06 August 1985, UT1-UTC= 0.5265533, Accumulated Leap Seconds=23
  { 2446284.50, 54.658372 }, //07 August 1985, UT1-UTC= 0.5256277, Accumulated Leap Seconds=23
  { 2446285.50, 54.659174 }, //08 August 1985, UT1-UTC= 0.5248259, Accumulated Leap Seconds=23
  { 2446286.50, 54.659817 }, //09 August 1985, UT1-UTC= 0.5241828, Accumulated Leap Seconds=23
  { 2446287.50, 54.660288 }, //10 August 1985, UT1-UTC= 0.5237117, Accumulated Leap Seconds=23
  { 2446288.50, 54.660597 }, //11 August 1985, UT1-UTC= 0.5234032, Accumulated Leap Seconds=23
  { 2446289.50, 54.660777 }, //12 August 1985, UT1-UTC= 0.5232225, Accumulated Leap Seconds=23
  { 2446290.50, 54.660904 }, //13 August 1985, UT1-UTC= 0.5230958, Accumulated Leap Seconds=23
  { 2446291.50, 54.661068 }, //14 August 1985, UT1-UTC= 0.5229319, Accumulated Leap Seconds=23
  { 2446292.50, 54.661372 }, //15 August 1985, UT1-UTC= 0.5226284, Accumulated Leap Seconds=23
  { 2446293.50, 54.661914 }, //16 August 1985, UT1-UTC= 0.5220862, Accumulated Leap Seconds=23
  { 2446294.50, 54.662761 }, //17 August 1985, UT1-UTC= 0.5212385, Accumulated Leap Seconds=23
  { 2446295.50, 54.663922 }, //18 August 1985, UT1-UTC= 0.5200779, Accumulated Leap Seconds=23
  { 2446296.50, 54.665326 }, //19 August 1985, UT1-UTC= 0.5186741, Accumulated Leap Seconds=23
  { 2446297.50, 54.666843 }, //20 August 1985, UT1-UTC= 0.5171573, Accumulated Leap Seconds=23
  { 2446298.50, 54.668319 }, //21 August 1985, UT1-UTC= 0.5156806, Accumulated Leap Seconds=23
  { 2446299.50, 54.669624 }, //22 August 1985, UT1-UTC= 0.5143765, Accumulated Leap Seconds=23
  { 2446300.50, 54.670677 }, //23 August 1985, UT1-UTC= 0.5133225, Accumulated Leap Seconds=23
  { 2446301.50, 54.671465 }, //24 August 1985, UT1-UTC= 0.5125351, Accumulated Leap Seconds=23
  { 2446302.50, 54.672030 }, //25 August 1985, UT1-UTC= 0.5119700, Accumulated Leap Seconds=23
  { 2446303.50, 54.672468 }, //26 August 1985, UT1-UTC= 0.5115321, Accumulated Leap Seconds=23
  { 2446304.50, 54.672908 }, //27 August 1985, UT1-UTC= 0.5110924, Accumulated Leap Seconds=23
  { 2446305.50, 54.673465 }, //28 August 1985, UT1-UTC= 0.5105346, Accumulated Leap Seconds=23
  { 2446306.50, 54.674211 }, //29 August 1985, UT1-UTC= 0.5097895, Accumulated Leap Seconds=23
  { 2446307.50, 54.675163 }, //30 August 1985, UT1-UTC= 0.5088366, Accumulated Leap Seconds=23
  { 2446308.50, 54.676307 }, //31 August 1985, UT1-UTC= 0.5076929, Accumulated Leap Seconds=23
  { 2446309.50, 54.677594 }, //01 September 1985, UT1-UTC= 0.5064060, Accumulated Leap Seconds=23
  { 2446310.50, 54.678958 }, //02 September 1985, UT1-UTC= 0.5050416, Accumulated Leap Seconds=23
  { 2446311.50, 54.680320 }, //03 September 1985, UT1-UTC= 0.5036798, Accumulated Leap Seconds=23
  { 2446312.50, 54.681604 }, //04 September 1985, UT1-UTC= 0.5023956, Accumulated Leap Seconds=23
  { 2446313.50, 54.682759 }, //05 September 1985, UT1-UTC= 0.5012414, Accumulated Leap Seconds=23
  { 2446314.50, 54.683754 }, //06 September 1985, UT1-UTC= 0.5002459, Accumulated Leap Seconds=23
  { 2446315.50, 54.684583 }, //07 September 1985, UT1-UTC= 0.4994173, Accumulated Leap Seconds=23
  { 2446316.50, 54.685270 }, //08 September 1985, UT1-UTC= 0.4987303, Accumulated Leap Seconds=23
  { 2446317.50, 54.685868 }, //09 September 1985, UT1-UTC= 0.4981323, Accumulated Leap Seconds=23
  { 2446318.50, 54.686452 }, //10 September 1985, UT1-UTC= 0.4975475, Accumulated Leap Seconds=23
  { 2446319.50, 54.687121 }, //11 September 1985, UT1-UTC= 0.4968790, Accumulated Leap Seconds=23
  { 2446320.50, 54.687980 }, //12 September 1985, UT1-UTC= 0.4960204, Accumulated Leap Seconds=23
  { 2446321.50, 54.689126 }, //13 September 1985, UT1-UTC= 0.4948742, Accumulated Leap Seconds=23
  { 2446322.50, 54.690610 }, //14 September 1985, UT1-UTC= 0.4933905, Accumulated Leap Seconds=23
  { 2446323.50, 54.692405 }, //15 September 1985, UT1-UTC= 0.4915952, Accumulated Leap Seconds=23
  { 2446324.50, 54.694404 }, //16 September 1985, UT1-UTC= 0.4895964, Accumulated Leap Seconds=23
  { 2446325.50, 54.696442 }, //17 September 1985, UT1-UTC= 0.4875584, Accumulated Leap Seconds=23
  { 2446326.50, 54.698351 }, //18 September 1985, UT1-UTC= 0.4856490, Accumulated Leap Seconds=23
  { 2446327.50, 54.700013 }, //19 September 1985, UT1-UTC= 0.4839873, Accumulated Leap Seconds=23
  { 2446328.50, 54.701389 }, //20 September 1985, UT1-UTC= 0.4826109, Accumulated Leap Seconds=23
  { 2446329.50, 54.702522 }, //21 September 1985, UT1-UTC= 0.4814778, Accumulated Leap Seconds=23
  { 2446330.50, 54.703511 }, //22 September 1985, UT1-UTC= 0.4804895, Accumulated Leap Seconds=23
  { 2446331.50, 54.704482 }, //23 September 1985, UT1-UTC= 0.4795176, Accumulated Leap Seconds=23
  { 2446332.50, 54.705553 }, //24 September 1985, UT1-UTC= 0.4784473, Accumulated Leap Seconds=23
  { 2446333.50, 54.706799 }, //25 September 1985, UT1-UTC= 0.4772010, Accumulated Leap Seconds=23
  { 2446334.50, 54.708255 }, //26 September 1985, UT1-UTC= 0.4757446, Accumulated Leap Seconds=23
  { 2446335.50, 54.709914 }, //27 September 1985, UT1-UTC= 0.4740863, Accumulated Leap Seconds=23
  { 2446336.50, 54.711728 }, //28 September 1985, UT1-UTC= 0.4722721, Accumulated Leap Seconds=23
  { 2446337.50, 54.713627 }, //29 September 1985, UT1-UTC= 0.4703729, Accumulated Leap Seconds=23
  { 2446338.50, 54.715531 }, //30 September 1985, UT1-UTC= 0.4684689, Accumulated Leap Seconds=23
  { 2446339.50, 54.717364 }, //01 October 1985, UT1-UTC= 0.4666357, Accumulated Leap Seconds=23
  { 2446340.50, 54.719069 }, //02 October 1985, UT1-UTC= 0.4649305, Accumulated Leap Seconds=23
  { 2446341.50, 54.720622 }, //03 October 1985, UT1-UTC= 0.4633780, Accumulated Leap Seconds=23
  { 2446342.50, 54.722022 }, //04 October 1985, UT1-UTC= 0.4619783, Accumulated Leap Seconds=23
  { 2446343.50, 54.723295 }, //05 October 1985, UT1-UTC= 0.4607050, Accumulated Leap Seconds=23
  { 2446344.50, 54.724489 }, //06 October 1985, UT1-UTC= 0.4595111, Accumulated Leap Seconds=23
  { 2446345.50, 54.725666 }, //07 October 1985, UT1-UTC= 0.4583344, Accumulated Leap Seconds=23
  { 2446346.50, 54.726910 }, //08 October 1985, UT1-UTC= 0.4570897, Accumulated Leap Seconds=23
  { 2446347.50, 54.728319 }, //09 October 1985, UT1-UTC= 0.4556811, Accumulated Leap Seconds=23
  { 2446348.50, 54.729980 }, //10 October 1985, UT1-UTC= 0.4540203, Accumulated Leap Seconds=23
  { 2446349.50, 54.731951 }, //11 October 1985, UT1-UTC= 0.4520493, Accumulated Leap Seconds=23
  { 2446350.50, 54.734233 }, //12 October 1985, UT1-UTC= 0.4497670, Accumulated Leap Seconds=23
  { 2446351.50, 54.736754 }, //13 October 1985, UT1-UTC= 0.4472462, Accumulated Leap Seconds=23
  { 2446352.50, 54.739366 }, //14 October 1985, UT1-UTC= 0.4446337, Accumulated Leap Seconds=23
  { 2446353.50, 54.741890 }, //15 October 1985, UT1-UTC= 0.4421096, Accumulated Leap Seconds=23
  { 2446354.50, 54.744173 }, //16 October 1985, UT1-UTC= 0.4398265, Accumulated Leap Seconds=23
  { 2446355.50, 54.746142 }, //17 October 1985, UT1-UTC= 0.4378582, Accumulated Leap Seconds=23
  { 2446356.50, 54.747819 }, //18 October 1985, UT1-UTC= 0.4361807, Accumulated Leap Seconds=23
  { 2446357.50, 54.749307 }, //19 October 1985, UT1-UTC= 0.4346926, Accumulated Leap Seconds=23
  { 2446358.50, 54.750745 }, //20 October 1985, UT1-UTC= 0.4332550, Accumulated Leap Seconds=23
  { 2446359.50, 54.752257 }, //21 October 1985, UT1-UTC= 0.4317426, Accumulated Leap Seconds=23
  { 2446360.50, 54.753928 }, //22 October 1985, UT1-UTC= 0.4300719, Accumulated Leap Seconds=23
  { 2446361.50, 54.755792 }, //23 October 1985, UT1-UTC= 0.4282081, Accumulated Leap Seconds=23
  { 2446362.50, 54.757839 }, //24 October 1985, UT1-UTC= 0.4261608, Accumulated Leap Seconds=23
  { 2446363.50, 54.760031 }, //25 October 1985, UT1-UTC= 0.4239694, Accumulated Leap Seconds=23
  { 2446364.50, 54.762308 }, //26 October 1985, UT1-UTC= 0.4216922, Accumulated Leap Seconds=23
  { 2446365.50, 54.764602 }, //27 October 1985, UT1-UTC= 0.4193983, Accumulated Leap Seconds=23
  { 2446366.50, 54.766835 }, //28 October 1985, UT1-UTC= 0.4171647, Accumulated Leap Seconds=23
  { 2446367.50, 54.768939 }, //29 October 1985, UT1-UTC= 0.4150614, Accumulated Leap Seconds=23
  { 2446368.50, 54.770860 }, //30 October 1985, UT1-UTC= 0.4131396, Accumulated Leap Seconds=23
  { 2446369.50, 54.772576 }, //31 October 1985, UT1-UTC= 0.4114242, Accumulated Leap Seconds=23
  { 2446370.50, 54.774096 }, //01 November 1985, UT1-UTC= 0.4099043, Accumulated Leap Seconds=23
  { 2446371.50, 54.775460 }, //02 November 1985, UT1-UTC= 0.4085396, Accumulated Leap Seconds=23
  { 2446372.50, 54.776729 }, //03 November 1985, UT1-UTC= 0.4072712, Accumulated Leap Seconds=23
  { 2446373.50, 54.777973 }, //04 November 1985, UT1-UTC= 0.4060266, Accumulated Leap Seconds=23
  { 2446374.50, 54.779275 }, //05 November 1985, UT1-UTC= 0.4047250, Accumulated Leap Seconds=23
  { 2446375.50, 54.780717 }, //06 November 1985, UT1-UTC= 0.4032825, Accumulated Leap Seconds=23
  { 2446376.50, 54.782383 }, //07 November 1985, UT1-UTC= 0.4016172, Accumulated Leap Seconds=23
  { 2446377.50, 54.784321 }, //08 November 1985, UT1-UTC= 0.3996793, Accumulated Leap Seconds=23
  { 2446378.50, 54.786521 }, //09 November 1985, UT1-UTC= 0.3974790, Accumulated Leap Seconds=23
  { 2446379.50, 54.788900 }, //10 November 1985, UT1-UTC= 0.3950995, Accumulated Leap Seconds=23
  { 2446380.50, 54.791318 }, //11 November 1985, UT1-UTC= 0.3926824, Accumulated Leap Seconds=23
  { 2446381.50, 54.793605 }, //12 November 1985, UT1-UTC= 0.3903947, Accumulated Leap Seconds=23
  { 2446382.50, 54.795632 }, //13 November 1985, UT1-UTC= 0.3883684, Accumulated Leap Seconds=23
  { 2446383.50, 54.797354 }, //14 November 1985, UT1-UTC= 0.3866464, Accumulated Leap Seconds=23
  { 2446384.50, 54.798826 }, //15 November 1985, UT1-UTC= 0.3851739, Accumulated Leap Seconds=23
  { 2446385.50, 54.800177 }, //16 November 1985, UT1-UTC= 0.3838234, Accumulated Leap Seconds=23
  { 2446386.50, 54.801552 }, //17 November 1985, UT1-UTC= 0.3824478, Accumulated Leap Seconds=23
  { 2446387.50, 54.803068 }, //18 November 1985, UT1-UTC= 0.3809319, Accumulated Leap Seconds=23
  { 2446388.50, 54.804782 }, //19 November 1985, UT1-UTC= 0.3792178, Accumulated Leap Seconds=23
  { 2446389.50, 54.806695 }, //20 November 1985, UT1-UTC= 0.3773048, Accumulated Leap Seconds=23
  { 2446390.50, 54.808764 }, //21 November 1985, UT1-UTC= 0.3752355, Accumulated Leap Seconds=23
  { 2446391.50, 54.810905 }, //22 November 1985, UT1-UTC= 0.3730954, Accumulated Leap Seconds=23
  { 2446392.50, 54.813019 }, //23 November 1985, UT1-UTC= 0.3709805, Accumulated Leap Seconds=23
  { 2446393.50, 54.815050 }, //24 November 1985, UT1-UTC= 0.3689505, Accumulated Leap Seconds=23
  { 2446394.50, 54.816950 }, //25 November 1985, UT1-UTC= 0.3670495, Accumulated Leap Seconds=23
  { 2446395.50, 54.818693 }, //26 November 1985, UT1-UTC= 0.3653072, Accumulated Leap Seconds=23
  { 2446396.50, 54.820261 }, //27 November 1985, UT1-UTC= 0.3637393, Accumulated Leap Seconds=23
  { 2446397.50, 54.821662 }, //28 November 1985, UT1-UTC= 0.3623383, Accumulated Leap Seconds=23
  { 2446398.50, 54.822929 }, //29 November 1985, UT1-UTC= 0.3610709, Accumulated Leap Seconds=23
  { 2446399.50, 54.824118 }, //30 November 1985, UT1-UTC= 0.3598815, Accumulated Leap Seconds=23
  { 2446400.50, 54.825302 }, //01 December 1985, UT1-UTC= 0.3586977, Accumulated Leap Seconds=23
  { 2446401.50, 54.826558 }, //02 December 1985, UT1-UTC= 0.3574421, Accumulated Leap Seconds=23
  { 2446402.50, 54.827958 }, //03 December 1985, UT1-UTC= 0.3560418, Accumulated Leap Seconds=23
  { 2446403.50, 54.829566 }, //04 December 1985, UT1-UTC= 0.3544335, Accumulated Leap Seconds=23
  { 2446404.50, 54.831425 }, //05 December 1985, UT1-UTC= 0.3525745, Accumulated Leap Seconds=23
  { 2446405.50, 54.833538 }, //06 December 1985, UT1-UTC= 0.3504624, Accumulated Leap Seconds=23
  { 2446406.50, 54.835843 }, //07 December 1985, UT1-UTC= 0.3481569, Accumulated Leap Seconds=23
  { 2446407.50, 54.838221 }, //08 December 1985, UT1-UTC= 0.3457790, Accumulated Leap Seconds=23
  { 2446408.50, 54.840519 }, //09 December 1985, UT1-UTC= 0.3434813, Accumulated Leap Seconds=23
  { 2446409.50, 54.842597 }, //10 December 1985, UT1-UTC= 0.3414028, Accumulated Leap Seconds=23
  { 2446410.50, 54.844379 }, //11 December 1985, UT1-UTC= 0.3396208, Accumulated Leap Seconds=23
  { 2446411.50, 54.845871 }, //12 December 1985, UT1-UTC= 0.3381291, Accumulated Leap Seconds=23
  { 2446412.50, 54.847159 }, //13 December 1985, UT1-UTC= 0.3368413, Accumulated Leap Seconds=23
  { 2446413.50, 54.848375 }, //14 December 1985, UT1-UTC= 0.3356254, Accumulated Leap Seconds=23
  { 2446414.50, 54.849646 }, //15 December 1985, UT1-UTC= 0.3343543, Accumulated Leap Seconds=23
  { 2446415.50, 54.851054 }, //16 December 1985, UT1-UTC= 0.3329457, Accumulated Leap Seconds=23
  { 2446416.50, 54.852625 }, //17 December 1985, UT1-UTC= 0.3313753, Accumulated Leap Seconds=23
  { 2446417.50, 54.854328 }, //18 December 1985, UT1-UTC= 0.3296716, Accumulated Leap Seconds=23
  { 2446418.50, 54.856106 }, //19 December 1985, UT1-UTC= 0.3278943, Accumulated Leap Seconds=23
  { 2446419.50, 54.857888 }, //20 December 1985, UT1-UTC= 0.3261116, Accumulated Leap Seconds=23
  { 2446420.50, 54.859616 }, //21 December 1985, UT1-UTC= 0.3243840, Accumulated Leap Seconds=23
  { 2446421.50, 54.861231 }, //22 December 1985, UT1-UTC= 0.3227687, Accumulated Leap Seconds=23
  { 2446422.50, 54.862687 }, //23 December 1985, UT1-UTC= 0.3213131, Accumulated Leap Seconds=23
  { 2446423.50, 54.863956 }, //24 December 1985, UT1-UTC= 0.3200444, Accumulated Leap Seconds=23
  { 2446424.50, 54.865034 }, //25 December 1985, UT1-UTC= 0.3189663, Accumulated Leap Seconds=23
  { 2446425.50, 54.865942 }, //26 December 1985, UT1-UTC= 0.3180578, Accumulated Leap Seconds=23
  { 2446426.50, 54.866727 }, //27 December 1985, UT1-UTC= 0.3172731, Accumulated Leap Seconds=23
  { 2446427.50, 54.867454 }, //28 December 1985, UT1-UTC= 0.3165460, Accumulated Leap Seconds=23
  { 2446428.50, 54.868201 }, //29 December 1985, UT1-UTC= 0.3157992, Accumulated Leap Seconds=23
  { 2446429.50, 54.869043 }, //30 December 1985, UT1-UTC= 0.3149569, Accumulated Leap Seconds=23
  { 2446430.50, 54.870044 }, //31 December 1985, UT1-UTC= 0.3139556, Accumulated Leap Seconds=23
  { 2446431.50, 54.871251 }, //01 January 1986, UT1-UTC= 0.3127488, Accumulated Leap Seconds=23
  { 2446432.50, 54.872673 }, //02 January 1986, UT1-UTC= 0.3113274, Accumulated Leap Seconds=23
  { 2446433.50, 54.874272 }, //03 January 1986, UT1-UTC= 0.3097276, Accumulated Leap Seconds=23
  { 2446434.50, 54.875972 }, //04 January 1986, UT1-UTC= 0.3080284, Accumulated Leap Seconds=23
  { 2446435.50, 54.877661 }, //05 January 1986, UT1-UTC= 0.3063385, Accumulated Leap Seconds=23
  { 2446436.50, 54.879225 }, //06 January 1986, UT1-UTC= 0.3047745, Accumulated Leap Seconds=23
  { 2446437.50, 54.880575 }, //07 January 1986, UT1-UTC= 0.3034249, Accumulated Leap Seconds=23
  { 2446438.50, 54.881689 }, //08 January 1986, UT1-UTC= 0.3023108, Accumulated Leap Seconds=23
  { 2446439.50, 54.882629 }, //09 January 1986, UT1-UTC= 0.3013713, Accumulated Leap Seconds=23
  { 2446440.50, 54.883514 }, //10 January 1986, UT1-UTC= 0.3004860, Accumulated Leap Seconds=23
  { 2446441.50, 54.884480 }, //11 January 1986, UT1-UTC= 0.2995195, Accumulated Leap Seconds=23
  { 2446442.50, 54.885637 }, //12 January 1986, UT1-UTC= 0.2983631, Accumulated Leap Seconds=23
  { 2446443.50, 54.887033 }, //13 January 1986, UT1-UTC= 0.2969669, Accumulated Leap Seconds=23
  { 2446444.50, 54.888654 }, //14 January 1986, UT1-UTC= 0.2953462, Accumulated Leap Seconds=23
  { 2446445.50, 54.890434 }, //15 January 1986, UT1-UTC= 0.2935662, Accumulated Leap Seconds=23
  { 2446446.50, 54.892272 }, //16 January 1986, UT1-UTC= 0.2917281, Accumulated Leap Seconds=23
  { 2446447.50, 54.894066 }, //17 January 1986, UT1-UTC= 0.2899339, Accumulated Leap Seconds=23
  { 2446448.50, 54.895748 }, //18 January 1986, UT1-UTC= 0.2882517, Accumulated Leap Seconds=23
  { 2446449.50, 54.897275 }, //19 January 1986, UT1-UTC= 0.2867252, Accumulated Leap Seconds=23
  { 2446450.50, 54.898625 }, //20 January 1986, UT1-UTC= 0.2853748, Accumulated Leap Seconds=23
  { 2446451.50, 54.899811 }, //21 January 1986, UT1-UTC= 0.2841888, Accumulated Leap Seconds=23
  { 2446452.50, 54.900861 }, //22 January 1986, UT1-UTC= 0.2831394, Accumulated Leap Seconds=23
  { 2446453.50, 54.901820 }, //23 January 1986, UT1-UTC= 0.2821802, Accumulated Leap Seconds=23
  { 2446454.50, 54.902755 }, //24 January 1986, UT1-UTC= 0.2812455, Accumulated Leap Seconds=23
  { 2446455.50, 54.903743 }, //25 January 1986, UT1-UTC= 0.2802567, Accumulated Leap Seconds=23
  { 2446456.50, 54.904871 }, //26 January 1986, UT1-UTC= 0.2791292, Accumulated Leap Seconds=23
  { 2446457.50, 54.906212 }, //27 January 1986, UT1-UTC= 0.2777884, Accumulated Leap Seconds=23
  { 2446458.50, 54.907812 }, //28 January 1986, UT1-UTC= 0.2761882, Accumulated Leap Seconds=23
  { 2446459.50, 54.909674 }, //29 January 1986, UT1-UTC= 0.2743264, Accumulated Leap Seconds=23
  { 2446460.50, 54.911752 }, //30 January 1986, UT1-UTC= 0.2722484, Accumulated Leap Seconds=23
  { 2446461.50, 54.913951 }, //31 January 1986, UT1-UTC= 0.2700491, Accumulated Leap Seconds=23
  { 2446462.50, 54.916146 }, //01 February 1986, UT1-UTC= 0.2678540, Accumulated Leap Seconds=23
  { 2446463.50, 54.918214 }, //02 February 1986, UT1-UTC= 0.2657865, Accumulated Leap Seconds=23
  { 2446464.50, 54.920060 }, //03 February 1986, UT1-UTC= 0.2639402, Accumulated Leap Seconds=23
  { 2446465.50, 54.921648 }, //04 February 1986, UT1-UTC= 0.2623517, Accumulated Leap Seconds=23
  { 2446466.50, 54.923012 }, //05 February 1986, UT1-UTC= 0.2609880, Accumulated Leap Seconds=23
  { 2446467.50, 54.924243 }, //06 February 1986, UT1-UTC= 0.2597569, Accumulated Leap Seconds=23
  { 2446468.50, 54.925464 }, //07 February 1986, UT1-UTC= 0.2585362, Accumulated Leap Seconds=23
  { 2446469.50, 54.926792 }, //08 February 1986, UT1-UTC= 0.2572082, Accumulated Leap Seconds=23
  { 2446470.50, 54.928303 }, //09 February 1986, UT1-UTC= 0.2556970, Accumulated Leap Seconds=23
  { 2446471.50, 54.930014 }, //10 February 1986, UT1-UTC= 0.2539859, Accumulated Leap Seconds=23
  { 2446472.50, 54.931884 }, //11 February 1986, UT1-UTC= 0.2521160, Accumulated Leap Seconds=23
  { 2446473.50, 54.933837 }, //12 February 1986, UT1-UTC= 0.2501631, Accumulated Leap Seconds=23
  { 2446474.50, 54.935788 }, //13 February 1986, UT1-UTC= 0.2482124, Accumulated Leap Seconds=23
  { 2446475.50, 54.937658 }, //14 February 1986, UT1-UTC= 0.2463422, Accumulated Leap Seconds=23
  { 2446476.50, 54.939381 }, //15 February 1986, UT1-UTC= 0.2446187, Accumulated Leap Seconds=23
  { 2446477.50, 54.940911 }, //16 February 1986, UT1-UTC= 0.2430887, Accumulated Leap Seconds=23
  { 2446478.50, 54.942225 }, //17 February 1986, UT1-UTC= 0.2417748, Accumulated Leap Seconds=23
  { 2446479.50, 54.943328 }, //18 February 1986, UT1-UTC= 0.2406717, Accumulated Leap Seconds=23
  { 2446480.50, 54.944254 }, //19 February 1986, UT1-UTC= 0.2397458, Accumulated Leap Seconds=23
  { 2446481.50, 54.945072 }, //20 February 1986, UT1-UTC= 0.2389280, Accumulated Leap Seconds=23
  { 2446482.50, 54.945869 }, //21 February 1986, UT1-UTC= 0.2381308, Accumulated Leap Seconds=23
  { 2446483.50, 54.946740 }, //22 February 1986, UT1-UTC= 0.2372596, Accumulated Leap Seconds=23
  { 2446484.50, 54.947773 }, //23 February 1986, UT1-UTC= 0.2362265, Accumulated Leap Seconds=23
  { 2446485.50, 54.949037 }, //24 February 1986, UT1-UTC= 0.2349627, Accumulated Leap Seconds=23
  { 2446486.50, 54.950564 }, //25 February 1986, UT1-UTC= 0.2334362, Accumulated Leap Seconds=23
  { 2446487.50, 54.952328 }, //26 February 1986, UT1-UTC= 0.2316716, Accumulated Leap Seconds=23
  { 2446488.50, 54.954247 }, //27 February 1986, UT1-UTC= 0.2297533, Accumulated Leap Seconds=23
  { 2446489.50, 54.956197 }, //28 February 1986, UT1-UTC= 0.2278028, Accumulated Leap Seconds=23
  { 2446490.50, 54.958053 }, //01 March 1986, UT1-UTC= 0.2259465, Accumulated Leap Seconds=23
  { 2446491.50, 54.959708 }, //02 March 1986, UT1-UTC= 0.2242917, Accumulated Leap Seconds=23
  { 2446492.50, 54.961109 }, //03 March 1986, UT1-UTC= 0.2228908, Accumulated Leap Seconds=23
  { 2446493.50, 54.962280 }, //04 March 1986, UT1-UTC= 0.2217202, Accumulated Leap Seconds=23
  { 2446494.50, 54.963301 }, //05 March 1986, UT1-UTC= 0.2206991, Accumulated Leap Seconds=23
  { 2446495.50, 54.964284 }, //06 March 1986, UT1-UTC= 0.2197162, Accumulated Leap Seconds=23
  { 2446496.50, 54.965348 }, //07 March 1986, UT1-UTC= 0.2186525, Accumulated Leap Seconds=23
  { 2446497.50, 54.966580 }, //08 March 1986, UT1-UTC= 0.2174198, Accumulated Leap Seconds=23
  { 2446498.50, 54.968018 }, //09 March 1986, UT1-UTC= 0.2159823, Accumulated Leap Seconds=23
  { 2446499.50, 54.969637 }, //10 March 1986, UT1-UTC= 0.2143628, Accumulated Leap Seconds=23
  { 2446500.50, 54.971368 }, //11 March 1986, UT1-UTC= 0.2126321, Accumulated Leap Seconds=23
  { 2446501.50, 54.973107 }, //12 March 1986, UT1-UTC= 0.2108925, Accumulated Leap Seconds=23
  { 2446502.50, 54.974750 }, //13 March 1986, UT1-UTC= 0.2092503, Accumulated Leap Seconds=23
  { 2446503.50, 54.976207 }, //14 March 1986, UT1-UTC= 0.2077933, Accumulated Leap Seconds=23
  { 2446504.50, 54.977429 }, //15 March 1986, UT1-UTC= 0.2065707, Accumulated Leap Seconds=23
  { 2446505.50, 54.978423 }, //16 March 1986, UT1-UTC= 0.2055769, Accumulated Leap Seconds=23
  { 2446506.50, 54.979215 }, //17 March 1986, UT1-UTC= 0.2047848, Accumulated Leap Seconds=23
  { 2446507.50, 54.979850 }, //18 March 1986, UT1-UTC= 0.2041497, Accumulated Leap Seconds=23
  { 2446508.50, 54.980395 }, //19 March 1986, UT1-UTC= 0.2036051, Accumulated Leap Seconds=23
  { 2446509.50, 54.980931 }, //20 March 1986, UT1-UTC= 0.2030694, Accumulated Leap Seconds=23
  { 2446510.50, 54.981552 }, //21 March 1986, UT1-UTC= 0.2024476, Accumulated Leap Seconds=23
  { 2446511.50, 54.982367 }, //22 March 1986, UT1-UTC= 0.2016333, Accumulated Leap Seconds=23
  { 2446512.50, 54.983471 }, //23 March 1986, UT1-UTC= 0.2005292, Accumulated Leap Seconds=23
  { 2446513.50, 54.984928 }, //24 March 1986, UT1-UTC= 0.1990719, Accumulated Leap Seconds=23
  { 2446514.50, 54.986742 }, //25 March 1986, UT1-UTC= 0.1972580, Accumulated Leap Seconds=23
  { 2446515.50, 54.988846 }, //26 March 1986, UT1-UTC= 0.1951541, Accumulated Leap Seconds=23
  { 2446516.50, 54.991097 }, //27 March 1986, UT1-UTC= 0.1929030, Accumulated Leap Seconds=23
  { 2446517.50, 54.993313 }, //28 March 1986, UT1-UTC= 0.1906872, Accumulated Leap Seconds=23
  { 2446518.50, 54.995333 }, //29 March 1986, UT1-UTC= 0.1886672, Accumulated Leap Seconds=23
  { 2446519.50, 54.997062 }, //30 March 1986, UT1-UTC= 0.1869378, Accumulated Leap Seconds=23
  { 2446520.50, 54.998496 }, //31 March 1986, UT1-UTC= 0.1855035, Accumulated Leap Seconds=23
  { 2446521.50, 54.999719 }, //01 April 1986, UT1-UTC= 0.1842814, Accumulated Leap Seconds=23
  { 2446522.50, 55.000853 }, //02 April 1986, UT1-UTC= 0.1831467, Accumulated Leap Seconds=23
  { 2446523.50, 55.002032 }, //03 April 1986, UT1-UTC= 0.1819681, Accumulated Leap Seconds=23
  { 2446524.50, 55.003365 }, //04 April 1986, UT1-UTC= 0.1806353, Accumulated Leap Seconds=23
  { 2446525.50, 55.004914 }, //05 April 1986, UT1-UTC= 0.1790861, Accumulated Leap Seconds=23
  { 2446526.50, 55.006680 }, //06 April 1986, UT1-UTC= 0.1773201, Accumulated Leap Seconds=23
  { 2446527.50, 55.008609 }, //07 April 1986, UT1-UTC= 0.1753909, Accumulated Leap Seconds=23
  { 2446528.50, 55.010619 }, //08 April 1986, UT1-UTC= 0.1733811, Accumulated Leap Seconds=23
  { 2446529.50, 55.012615 }, //09 April 1986, UT1-UTC= 0.1713846, Accumulated Leap Seconds=23
  { 2446530.50, 55.014513 }, //10 April 1986, UT1-UTC= 0.1694872, Accumulated Leap Seconds=23
  { 2446531.50, 55.016256 }, //11 April 1986, UT1-UTC= 0.1677442, Accumulated Leap Seconds=23
  { 2446532.50, 55.017820 }, //12 April 1986, UT1-UTC= 0.1661804, Accumulated Leap Seconds=23
  { 2446533.50, 55.019210 }, //13 April 1986, UT1-UTC= 0.1647902, Accumulated Leap Seconds=23
  { 2446534.50, 55.020459 }, //14 April 1986, UT1-UTC= 0.1635413, Accumulated Leap Seconds=23
  { 2446535.50, 55.021620 }, //15 April 1986, UT1-UTC= 0.1623801, Accumulated Leap Seconds=23
  { 2446536.50, 55.022755 }, //16 April 1986, UT1-UTC= 0.1612452, Accumulated Leap Seconds=23
  { 2446537.50, 55.023928 }, //17 April 1986, UT1-UTC= 0.1600723, Accumulated Leap Seconds=23
  { 2446538.50, 55.025204 }, //18 April 1986, UT1-UTC= 0.1587962, Accumulated Leap Seconds=23
  { 2446539.50, 55.026644 }, //19 April 1986, UT1-UTC= 0.1573557, Accumulated Leap Seconds=23
  { 2446540.50, 55.028298 }, //20 April 1986, UT1-UTC= 0.1557019, Accumulated Leap Seconds=23
  { 2446541.50, 55.030198 }, //21 April 1986, UT1-UTC= 0.1538015, Accumulated Leap Seconds=23
  { 2446542.50, 55.032328 }, //22 April 1986, UT1-UTC= 0.1516723, Accumulated Leap Seconds=23
  { 2446543.50, 55.034599 }, //23 April 1986, UT1-UTC= 0.1494010, Accumulated Leap Seconds=23
  { 2446544.50, 55.036868 }, //24 April 1986, UT1-UTC= 0.1471320, Accumulated Leap Seconds=23
  { 2446545.50, 55.038979 }, //25 April 1986, UT1-UTC= 0.1450208, Accumulated Leap Seconds=23
  { 2446546.50, 55.040821 }, //26 April 1986, UT1-UTC= 0.1431791, Accumulated Leap Seconds=23
  { 2446547.50, 55.042366 }, //27 April 1986, UT1-UTC= 0.1416339, Accumulated Leap Seconds=23
  { 2446548.50, 55.043680 }, //28 April 1986, UT1-UTC= 0.1403200, Accumulated Leap Seconds=23
  { 2446549.50, 55.044893 }, //29 April 1986, UT1-UTC= 0.1391068, Accumulated Leap Seconds=23
  { 2446550.50, 55.046152 }, //30 April 1986, UT1-UTC= 0.1378475, Accumulated Leap Seconds=23
  { 2446551.50, 55.047571 }, //01 May 1986, UT1-UTC= 0.1364290, Accumulated Leap Seconds=23
  { 2446552.50, 55.049204 }, //02 May 1986, UT1-UTC= 0.1347961, Accumulated Leap Seconds=23
  { 2446553.50, 55.051048 }, //03 May 1986, UT1-UTC= 0.1329516, Accumulated Leap Seconds=23
  { 2446554.50, 55.053057 }, //04 May 1986, UT1-UTC= 0.1309427, Accumulated Leap Seconds=23
  { 2446555.50, 55.055155 }, //05 May 1986, UT1-UTC= 0.1288453, Accumulated Leap Seconds=23
  { 2446556.50, 55.057254 }, //06 May 1986, UT1-UTC= 0.1267458, Accumulated Leap Seconds=23
  { 2446557.50, 55.059274 }, //07 May 1986, UT1-UTC= 0.1247264, Accumulated Leap Seconds=23
  { 2446558.50, 55.061146 }, //08 May 1986, UT1-UTC= 0.1228541, Accumulated Leap Seconds=23
  { 2446559.50, 55.062830 }, //09 May 1986, UT1-UTC= 0.1211703, Accumulated Leap Seconds=23
  { 2446560.50, 55.064312 }, //10 May 1986, UT1-UTC= 0.1196884, Accumulated Leap Seconds=23
  { 2446561.50, 55.065598 }, //11 May 1986, UT1-UTC= 0.1184023, Accumulated Leap Seconds=23
  { 2446562.50, 55.066719 }, //12 May 1986, UT1-UTC= 0.1172812, Accumulated Leap Seconds=23
  { 2446563.50, 55.067726 }, //13 May 1986, UT1-UTC= 0.1162741, Accumulated Leap Seconds=23
  { 2446564.50, 55.068682 }, //14 May 1986, UT1-UTC= 0.1153183, Accumulated Leap Seconds=23
  { 2446565.50, 55.069647 }, //15 May 1986, UT1-UTC= 0.1143529, Accumulated Leap Seconds=23
  { 2446566.50, 55.070678 }, //16 May 1986, UT1-UTC= 0.1133222, Accumulated Leap Seconds=23
  { 2446567.50, 55.071858 }, //17 May 1986, UT1-UTC= 0.1121424, Accumulated Leap Seconds=23
  { 2446568.50, 55.073255 }, //18 May 1986, UT1-UTC= 0.1107449, Accumulated Leap Seconds=23
  { 2446569.50, 55.074886 }, //19 May 1986, UT1-UTC= 0.1091144, Accumulated Leap Seconds=23
  { 2446570.50, 55.076697 }, //20 May 1986, UT1-UTC= 0.1073034, Accumulated Leap Seconds=23
  { 2446571.50, 55.078575 }, //21 May 1986, UT1-UTC= 0.1054253, Accumulated Leap Seconds=23
  { 2446572.50, 55.080369 }, //22 May 1986, UT1-UTC= 0.1036311, Accumulated Leap Seconds=23
  { 2446573.50, 55.081939 }, //23 May 1986, UT1-UTC= 0.1020607, Accumulated Leap Seconds=23
  { 2446574.50, 55.083206 }, //24 May 1986, UT1-UTC= 0.1007935, Accumulated Leap Seconds=23
  { 2446575.50, 55.084185 }, //25 May 1986, UT1-UTC= 0.0998149, Accumulated Leap Seconds=23
  { 2446576.50, 55.084978 }, //26 May 1986, UT1-UTC= 0.0990217, Accumulated Leap Seconds=23
  { 2446577.50, 55.085728 }, //27 May 1986, UT1-UTC= 0.0982724, Accumulated Leap Seconds=23
  { 2446578.50, 55.086558 }, //28 May 1986, UT1-UTC= 0.0974415, Accumulated Leap Seconds=23
  { 2446579.50, 55.087539 }, //29 May 1986, UT1-UTC= 0.0964606, Accumulated Leap Seconds=23
  { 2446580.50, 55.088669 }, //30 May 1986, UT1-UTC= 0.0953312, Accumulated Leap Seconds=23
  { 2446581.50, 55.089905 }, //31 May 1986, UT1-UTC= 0.0940954, Accumulated Leap Seconds=23
  { 2446582.50, 55.091178 }, //01 June 1986, UT1-UTC= 0.0928222, Accumulated Leap Seconds=23
  { 2446583.50, 55.092411 }, //02 June 1986, UT1-UTC= 0.0915886, Accumulated Leap Seconds=23
  { 2446584.50, 55.093536 }, //03 June 1986, UT1-UTC= 0.0904643, Accumulated Leap Seconds=23
  { 2446585.50, 55.094497 }, //04 June 1986, UT1-UTC= 0.0895028, Accumulated Leap Seconds=23
  { 2446586.50, 55.095259 }, //05 June 1986, UT1-UTC= 0.0887405, Accumulated Leap Seconds=23
  { 2446587.50, 55.095808 }, //06 June 1986, UT1-UTC= 0.0881923, Accumulated Leap Seconds=23
  { 2446588.50, 55.096152 }, //07 June 1986, UT1-UTC= 0.0878477, Accumulated Leap Seconds=23
  { 2446589.50, 55.096331 }, //08 June 1986, UT1-UTC= 0.0876685, Accumulated Leap Seconds=23
  { 2446590.50, 55.096405 }, //09 June 1986, UT1-UTC= 0.0875946, Accumulated Leap Seconds=23
  { 2446591.50, 55.096443 }, //10 June 1986, UT1-UTC= 0.0875566, Accumulated Leap Seconds=23
  { 2446592.50, 55.096519 }, //11 June 1986, UT1-UTC= 0.0874811, Accumulated Leap Seconds=23
  { 2446593.50, 55.096706 }, //12 June 1986, UT1-UTC= 0.0872937, Accumulated Leap Seconds=23
  { 2446594.50, 55.097070 }, //13 June 1986, UT1-UTC= 0.0869304, Accumulated Leap Seconds=23
  { 2446595.50, 55.097655 }, //14 June 1986, UT1-UTC= 0.0863451, Accumulated Leap Seconds=23
  { 2446596.50, 55.098477 }, //15 June 1986, UT1-UTC= 0.0855227, Accumulated Leap Seconds=23
  { 2446597.50, 55.099519 }, //16 June 1986, UT1-UTC= 0.0844807, Accumulated Leap Seconds=23
  { 2446598.50, 55.100718 }, //17 June 1986, UT1-UTC= 0.0832817, Accumulated Leap Seconds=23
  { 2446599.50, 55.101964 }, //18 June 1986, UT1-UTC= 0.0820359, Accumulated Leap Seconds=23
  { 2446600.50, 55.103129 }, //19 June 1986, UT1-UTC= 0.0808705, Accumulated Leap Seconds=23
  { 2446601.50, 55.104096 }, //20 June 1986, UT1-UTC= 0.0799042, Accumulated Leap Seconds=23
  { 2446602.50, 55.104818 }, //21 June 1986, UT1-UTC= 0.0791816, Accumulated Leap Seconds=23
  { 2446603.50, 55.105344 }, //22 June 1986, UT1-UTC= 0.0786557, Accumulated Leap Seconds=23
  { 2446604.50, 55.105787 }, //23 June 1986, UT1-UTC= 0.0782125, Accumulated Leap Seconds=23
  { 2446605.50, 55.106284 }, //24 June 1986, UT1-UTC= 0.0777163, Accumulated Leap Seconds=23
  { 2446606.50, 55.106945 }, //25 June 1986, UT1-UTC= 0.0770549, Accumulated Leap Seconds=23
  { 2446607.50, 55.107815 }, //26 June 1986, UT1-UTC= 0.0761849, Accumulated Leap Seconds=23
  { 2446608.50, 55.108864 }, //27 June 1986, UT1-UTC= 0.0751358, Accumulated Leap Seconds=23
  { 2446609.50, 55.110017 }, //28 June 1986, UT1-UTC= 0.0739825, Accumulated Leap Seconds=23
  { 2446610.50, 55.111184 }, //29 June 1986, UT1-UTC= 0.0728159, Accumulated Leap Seconds=23
  { 2446611.50, 55.112278 }, //30 June 1986, UT1-UTC= 0.0717222, Accumulated Leap Seconds=23
  { 2446612.50, 55.113239 }, //01 July 1986, UT1-UTC= 0.0707614, Accumulated Leap Seconds=23
  { 2446613.50, 55.114029 }, //02 July 1986, UT1-UTC= 0.0699709, Accumulated Leap Seconds=23
  { 2446614.50, 55.114633 }, //03 July 1986, UT1-UTC= 0.0693670, Accumulated Leap Seconds=23
  { 2446615.50, 55.115059 }, //04 July 1986, UT1-UTC= 0.0689412, Accumulated Leap Seconds=23
  { 2446616.50, 55.115337 }, //05 July 1986, UT1-UTC= 0.0686625, Accumulated Leap Seconds=23
  { 2446617.50, 55.115521 }, //06 July 1986, UT1-UTC= 0.0684785, Accumulated Leap Seconds=23
  { 2446618.50, 55.115681 }, //07 July 1986, UT1-UTC= 0.0683187, Accumulated Leap Seconds=23
  { 2446619.50, 55.115899 }, //08 July 1986, UT1-UTC= 0.0681008, Accumulated Leap Seconds=23
  { 2446620.50, 55.116250 }, //09 July 1986, UT1-UTC= 0.0677496, Accumulated Leap Seconds=23
  { 2446621.50, 55.116786 }, //10 July 1986, UT1-UTC= 0.0672144, Accumulated Leap Seconds=23
  { 2446622.50, 55.117529 }, //11 July 1986, UT1-UTC= 0.0664710, Accumulated Leap Seconds=23
  { 2446623.50, 55.118473 }, //12 July 1986, UT1-UTC= 0.0655274, Accumulated Leap Seconds=23
  { 2446624.50, 55.119567 }, //13 July 1986, UT1-UTC= 0.0644334, Accumulated Leap Seconds=23
  { 2446625.50, 55.120728 }, //14 July 1986, UT1-UTC= 0.0632724, Accumulated Leap Seconds=23
  { 2446626.50, 55.121861 }, //15 July 1986, UT1-UTC= 0.0621389, Accumulated Leap Seconds=23
  { 2446627.50, 55.122863 }, //16 July 1986, UT1-UTC= 0.0611374, Accumulated Leap Seconds=23
  { 2446628.50, 55.123646 }, //17 July 1986, UT1-UTC= 0.0603540, Accumulated Leap Seconds=23
  { 2446629.50, 55.124175 }, //18 July 1986, UT1-UTC= 0.0598251, Accumulated Leap Seconds=23
  { 2446630.50, 55.124486 }, //19 July 1986, UT1-UTC= 0.0595136, Accumulated Leap Seconds=23
  { 2446631.50, 55.124693 }, //20 July 1986, UT1-UTC= 0.0593065, Accumulated Leap Seconds=23
  { 2446632.50, 55.124945 }, //21 July 1986, UT1-UTC= 0.0590548, Accumulated Leap Seconds=23
  { 2446633.50, 55.125377 }, //22 July 1986, UT1-UTC= 0.0586225, Accumulated Leap Seconds=23
  { 2446634.50, 55.126065 }, //23 July 1986, UT1-UTC= 0.0579349, Accumulated Leap Seconds=23
  { 2446635.50, 55.127001 }, //24 July 1986, UT1-UTC= 0.0569992, Accumulated Leap Seconds=23
  { 2446636.50, 55.128102 }, //25 July 1986, UT1-UTC= 0.0558980, Accumulated Leap Seconds=23
  { 2446637.50, 55.129247 }, //26 July 1986, UT1-UTC= 0.0547533, Accumulated Leap Seconds=23
  { 2446638.50, 55.130316 }, //27 July 1986, UT1-UTC= 0.0536835, Accumulated Leap Seconds=23
  { 2446639.50, 55.131222 }, //28 July 1986, UT1-UTC= 0.0527776, Accumulated Leap Seconds=23
  { 2446640.50, 55.131922 }, //29 July 1986, UT1-UTC= 0.0520778, Accumulated Leap Seconds=23
  { 2446641.50, 55.132403 }, //30 July 1986, UT1-UTC= 0.0515974, Accumulated Leap Seconds=23
  { 2446642.50, 55.132674 }, //31 July 1986, UT1-UTC= 0.0513261, Accumulated Leap Seconds=23
  { 2446643.50, 55.132774 }, //01 August 1986, UT1-UTC= 0.0512260, Accumulated Leap Seconds=23
  { 2446644.50, 55.132766 }, //02 August 1986, UT1-UTC= 0.0512339, Accumulated Leap Seconds=23
  { 2446645.50, 55.132726 }, //03 August 1986, UT1-UTC= 0.0512737, Accumulated Leap Seconds=23
  { 2446646.50, 55.132727 }, //04 August 1986, UT1-UTC= 0.0512729, Accumulated Leap Seconds=23
  { 2446647.50, 55.132866 }, //05 August 1986, UT1-UTC= 0.0511340, Accumulated Leap Seconds=23
  { 2446648.50, 55.133229 }, //06 August 1986, UT1-UTC= 0.0507705, Accumulated Leap Seconds=23
  { 2446649.50, 55.133872 }, //07 August 1986, UT1-UTC= 0.0501277, Accumulated Leap Seconds=23
  { 2446650.50, 55.134807 }, //08 August 1986, UT1-UTC= 0.0491930, Accumulated Leap Seconds=23
  { 2446651.50, 55.135984 }, //09 August 1986, UT1-UTC= 0.0480159, Accumulated Leap Seconds=23
  { 2446652.50, 55.137304 }, //10 August 1986, UT1-UTC= 0.0466961, Accumulated Leap Seconds=23
  { 2446653.50, 55.138641 }, //11 August 1986, UT1-UTC= 0.0453587, Accumulated Leap Seconds=23
  { 2446654.50, 55.139869 }, //12 August 1986, UT1-UTC= 0.0441308, Accumulated Leap Seconds=23
  { 2446655.50, 55.140888 }, //13 August 1986, UT1-UTC= 0.0431124, Accumulated Leap Seconds=23
  { 2446656.50, 55.141644 }, //14 August 1986, UT1-UTC= 0.0423556, Accumulated Leap Seconds=23
  { 2446657.50, 55.142152 }, //15 August 1986, UT1-UTC= 0.0418480, Accumulated Leap Seconds=23
  { 2446658.50, 55.142494 }, //16 August 1986, UT1-UTC= 0.0415055, Accumulated Leap Seconds=23
  { 2446659.50, 55.142803 }, //17 August 1986, UT1-UTC= 0.0411966, Accumulated Leap Seconds=23
  { 2446660.50, 55.143219 }, //18 August 1986, UT1-UTC= 0.0407812, Accumulated Leap Seconds=23
  { 2446661.50, 55.143848 }, //19 August 1986, UT1-UTC= 0.0401517, Accumulated Leap Seconds=23
  { 2446662.50, 55.144732 }, //20 August 1986, UT1-UTC= 0.0392677, Accumulated Leap Seconds=23
  { 2446663.50, 55.145834 }, //21 August 1986, UT1-UTC= 0.0381660, Accumulated Leap Seconds=23
  { 2446664.50, 55.147058 }, //22 August 1986, UT1-UTC= 0.0369417, Accumulated Leap Seconds=23
  { 2446665.50, 55.148287 }, //23 August 1986, UT1-UTC= 0.0357128, Accumulated Leap Seconds=23
  { 2446666.50, 55.149418 }, //24 August 1986, UT1-UTC= 0.0345818, Accumulated Leap Seconds=23
  { 2446667.50, 55.150382 }, //25 August 1986, UT1-UTC= 0.0336178, Accumulated Leap Seconds=23
  { 2446668.50, 55.151149 }, //26 August 1986, UT1-UTC= 0.0328505, Accumulated Leap Seconds=23
  { 2446669.50, 55.151724 }, //27 August 1986, UT1-UTC= 0.0322762, Accumulated Leap Seconds=23
  { 2446670.50, 55.152134 }, //28 August 1986, UT1-UTC= 0.0318657, Accumulated Leap Seconds=23
  { 2446671.50, 55.152428 }, //29 August 1986, UT1-UTC= 0.0315723, Accumulated Leap Seconds=23
  { 2446672.50, 55.152665 }, //30 August 1986, UT1-UTC= 0.0313351, Accumulated Leap Seconds=23
  { 2446673.50, 55.152914 }, //31 August 1986, UT1-UTC= 0.0310858, Accumulated Leap Seconds=23
  { 2446674.50, 55.153242 }, //01 September 1986, UT1-UTC= 0.0307577, Accumulated Leap Seconds=23
  { 2446675.50, 55.153712 }, //02 September 1986, UT1-UTC= 0.0302882, Accumulated Leap Seconds=23
  { 2446676.50, 55.154391 }, //03 September 1986, UT1-UTC= 0.0296094, Accumulated Leap Seconds=23
  { 2446677.50, 55.155322 }, //04 September 1986, UT1-UTC= 0.0286777, Accumulated Leap Seconds=23
  { 2446678.50, 55.156506 }, //05 September 1986, UT1-UTC= 0.0274936, Accumulated Leap Seconds=23
  { 2446679.50, 55.157893 }, //06 September 1986, UT1-UTC= 0.0261070, Accumulated Leap Seconds=23
  { 2446680.50, 55.159391 }, //07 September 1986, UT1-UTC= 0.0246089, Accumulated Leap Seconds=23
  { 2446681.50, 55.160880 }, //08 September 1986, UT1-UTC= 0.0231196, Accumulated Leap Seconds=23
  { 2446682.50, 55.162247 }, //09 September 1986, UT1-UTC= 0.0217530, Accumulated Leap Seconds=23
  { 2446683.50, 55.163414 }, //10 September 1986, UT1-UTC= 0.0205860, Accumulated Leap Seconds=23
  { 2446684.50, 55.164359 }, //11 September 1986, UT1-UTC= 0.0196405, Accumulated Leap Seconds=23
  { 2446685.50, 55.165127 }, //12 September 1986, UT1-UTC= 0.0188731, Accumulated Leap Seconds=23
  { 2446686.50, 55.165829 }, //13 September 1986, UT1-UTC= 0.0181707, Accumulated Leap Seconds=23
  { 2446687.50, 55.166606 }, //14 September 1986, UT1-UTC= 0.0173944, Accumulated Leap Seconds=23
  { 2446688.50, 55.167579 }, //15 September 1986, UT1-UTC= 0.0164212, Accumulated Leap Seconds=23
  { 2446689.50, 55.168826 }, //16 September 1986, UT1-UTC= 0.0151742, Accumulated Leap Seconds=23
  { 2446690.50, 55.170356 }, //17 September 1986, UT1-UTC= 0.0136440, Accumulated Leap Seconds=23
  { 2446691.50, 55.172104 }, //18 September 1986, UT1-UTC= 0.0118957, Accumulated Leap Seconds=23
  { 2446692.50, 55.173960 }, //19 September 1986, UT1-UTC= 0.0100397, Accumulated Leap Seconds=23
  { 2446693.50, 55.175804 }, //20 September 1986, UT1-UTC= 0.0081958, Accumulated Leap Seconds=23
  { 2446694.50, 55.177534 }, //21 September 1986, UT1-UTC= 0.0064660, Accumulated Leap Seconds=23
  { 2446695.50, 55.179082 }, //22 September 1986, UT1-UTC= 0.0049175, Accumulated Leap Seconds=23
  { 2446696.50, 55.180427 }, //23 September 1986, UT1-UTC= 0.0035730, Accumulated Leap Seconds=23
  { 2446697.50, 55.181584 }, //24 September 1986, UT1-UTC= 0.0024156, Accumulated Leap Seconds=23
  { 2446698.50, 55.182600 }, //25 September 1986, UT1-UTC= 0.0014001, Accumulated Leap Seconds=23
  { 2446699.50, 55.183541 }, //26 September 1986, UT1-UTC= 0.0004592, Accumulated Leap Seconds=23
  { 2446700.50, 55.184487 }, //27 September 1986, UT1-UTC=-0.0004874, Accumulated Leap Seconds=23
  { 2446701.50, 55.185520 }, //28 September 1986, UT1-UTC=-0.0015203, Accumulated Leap Seconds=23
  { 2446702.50, 55.186714 }, //29 September 1986, UT1-UTC=-0.0027139, Accumulated Leap Seconds=23
  { 2446703.50, 55.188129 }, //30 September 1986, UT1-UTC=-0.0041289, Accumulated Leap Seconds=23
  { 2446704.50, 55.189800 }, //01 October 1986, UT1-UTC=-0.0058003, Accumulated Leap Seconds=23
  { 2446705.50, 55.191726 }, //02 October 1986, UT1-UTC=-0.0077263, Accumulated Leap Seconds=23
  { 2446706.50, 55.193861 }, //03 October 1986, UT1-UTC=-0.0098608, Accumulated Leap Seconds=23
  { 2446707.50, 55.196105 }, //04 October 1986, UT1-UTC=-0.0121055, Accumulated Leap Seconds=23
  { 2446708.50, 55.198324 }, //05 October 1986, UT1-UTC=-0.0143239, Accumulated Leap Seconds=23
  { 2446709.50, 55.200380 }, //06 October 1986, UT1-UTC=-0.0163802, Accumulated Leap Seconds=23
  { 2446710.50, 55.202182 }, //07 October 1986, UT1-UTC=-0.0181820, Accumulated Leap Seconds=23
  { 2446711.50, 55.203706 }, //08 October 1986, UT1-UTC=-0.0197064, Accumulated Leap Seconds=23
  { 2446712.50, 55.205004 }, //09 October 1986, UT1-UTC=-0.0210044, Accumulated Leap Seconds=23
  { 2446713.50, 55.206185 }, //10 October 1986, UT1-UTC=-0.0221855, Accumulated Leap Seconds=23
  { 2446714.50, 55.207379 }, //11 October 1986, UT1-UTC=-0.0233788, Accumulated Leap Seconds=23
  { 2446715.50, 55.208693 }, //12 October 1986, UT1-UTC=-0.0246927, Accumulated Leap Seconds=23
  { 2446716.50, 55.210203 }, //13 October 1986, UT1-UTC=-0.0262035, Accumulated Leap Seconds=23
  { 2446717.50, 55.211931 }, //14 October 1986, UT1-UTC=-0.0279313, Accumulated Leap Seconds=23
  { 2446718.50, 55.213840 }, //15 October 1986, UT1-UTC=-0.0298396, Accumulated Leap Seconds=23
  { 2446719.50, 55.215852 }, //16 October 1986, UT1-UTC=-0.0318521, Accumulated Leap Seconds=23
  { 2446720.50, 55.217876 }, //17 October 1986, UT1-UTC=-0.0338758, Accumulated Leap Seconds=23
  { 2446721.50, 55.219813 }, //18 October 1986, UT1-UTC=-0.0358129, Accumulated Leap Seconds=23
  { 2446722.50, 55.221596 }, //19 October 1986, UT1-UTC=-0.0375965, Accumulated Leap Seconds=23
  { 2446723.50, 55.223196 }, //20 October 1986, UT1-UTC=-0.0391958, Accumulated Leap Seconds=23
  { 2446724.50, 55.224615 }, //21 October 1986, UT1-UTC=-0.0406149, Accumulated Leap Seconds=23
  { 2446725.50, 55.225887 }, //22 October 1986, UT1-UTC=-0.0418875, Accumulated Leap Seconds=23
  { 2446726.50, 55.227067 }, //23 October 1986, UT1-UTC=-0.0430674, Accumulated Leap Seconds=23
  { 2446727.50, 55.228221 }, //24 October 1986, UT1-UTC=-0.0442210, Accumulated Leap Seconds=23
  { 2446728.50, 55.229414 }, //25 October 1986, UT1-UTC=-0.0454145, Accumulated Leap Seconds=23
  { 2446729.50, 55.230691 }, //26 October 1986, UT1-UTC=-0.0466908, Accumulated Leap Seconds=23
  { 2446730.50, 55.232084 }, //27 October 1986, UT1-UTC=-0.0480845, Accumulated Leap Seconds=23
  { 2446731.50, 55.233644 }, //28 October 1986, UT1-UTC=-0.0496441, Accumulated Leap Seconds=23
  { 2446732.50, 55.235403 }, //29 October 1986, UT1-UTC=-0.0514033, Accumulated Leap Seconds=23
  { 2446733.50, 55.237355 }, //30 October 1986, UT1-UTC=-0.0533552, Accumulated Leap Seconds=23
  { 2446734.50, 55.239442 }, //31 October 1986, UT1-UTC=-0.0554416, Accumulated Leap Seconds=23
  { 2446735.50, 55.241553 }, //01 November 1986, UT1-UTC=-0.0575531, Accumulated Leap Seconds=23
  { 2446736.50, 55.243547 }, //02 November 1986, UT1-UTC=-0.0595465, Accumulated Leap Seconds=23
  { 2446737.50, 55.245294 }, //03 November 1986, UT1-UTC=-0.0612940, Accumulated Leap Seconds=23
  { 2446738.50, 55.246732 }, //04 November 1986, UT1-UTC=-0.0627318, Accumulated Leap Seconds=23
  { 2446739.50, 55.247884 }, //05 November 1986, UT1-UTC=-0.0638844, Accumulated Leap Seconds=23
  { 2446740.50, 55.248855 }, //06 November 1986, UT1-UTC=-0.0648547, Accumulated Leap Seconds=23
  { 2446741.50, 55.249789 }, //07 November 1986, UT1-UTC=-0.0657888, Accumulated Leap Seconds=23
  { 2446742.50, 55.250828 }, //08 November 1986, UT1-UTC=-0.0668277, Accumulated Leap Seconds=23
  { 2446743.50, 55.252064 }, //09 November 1986, UT1-UTC=-0.0680639, Accumulated Leap Seconds=23
  { 2446744.50, 55.253532 }, //10 November 1986, UT1-UTC=-0.0695324, Accumulated Leap Seconds=23
  { 2446745.50, 55.255207 }, //11 November 1986, UT1-UTC=-0.0712069, Accumulated Leap Seconds=23
  { 2446746.50, 55.257015 }, //12 November 1986, UT1-UTC=-0.0730151, Accumulated Leap Seconds=23
  { 2446747.50, 55.258864 }, //13 November 1986, UT1-UTC=-0.0748640, Accumulated Leap Seconds=23
  { 2446748.50, 55.260657 }, //14 November 1986, UT1-UTC=-0.0766574, Accumulated Leap Seconds=23
  { 2446749.50, 55.262314 }, //15 November 1986, UT1-UTC=-0.0783136, Accumulated Leap Seconds=23
  { 2446750.50, 55.263779 }, //16 November 1986, UT1-UTC=-0.0797790, Accumulated Leap Seconds=23
  { 2446751.50, 55.265036 }, //17 November 1986, UT1-UTC=-0.0810361, Accumulated Leap Seconds=23
  { 2446752.50, 55.266101 }, //18 November 1986, UT1-UTC=-0.0821015, Accumulated Leap Seconds=23
  { 2446753.50, 55.267021 }, //19 November 1986, UT1-UTC=-0.0830210, Accumulated Leap Seconds=23
  { 2446754.50, 55.267861 }, //20 November 1986, UT1-UTC=-0.0838608, Accumulated Leap Seconds=23
  { 2446755.50, 55.268696 }, //21 November 1986, UT1-UTC=-0.0846963, Accumulated Leap Seconds=23
  { 2446756.50, 55.269604 }, //22 November 1986, UT1-UTC=-0.0856042, Accumulated Leap Seconds=23
  { 2446757.50, 55.270651 }, //23 November 1986, UT1-UTC=-0.0866508, Accumulated Leap Seconds=23
  { 2446758.50, 55.271885 }, //24 November 1986, UT1-UTC=-0.0878854, Accumulated Leap Seconds=23
  { 2446759.50, 55.273336 }, //25 November 1986, UT1-UTC=-0.0893359, Accumulated Leap Seconds=23
  { 2446760.50, 55.275002 }, //26 November 1986, UT1-UTC=-0.0910023, Accumulated Leap Seconds=23
  { 2446761.50, 55.276844 }, //27 November 1986, UT1-UTC=-0.0928436, Accumulated Leap Seconds=23
  { 2446762.50, 55.278772 }, //28 November 1986, UT1-UTC=-0.0947719, Accumulated Leap Seconds=23
  { 2446763.50, 55.280659 }, //29 November 1986, UT1-UTC=-0.0966592, Accumulated Leap Seconds=23
  { 2446764.50, 55.282368 }, //30 November 1986, UT1-UTC=-0.0983685, Accumulated Leap Seconds=23
  { 2446765.50, 55.283803 }, //01 December 1986, UT1-UTC=-0.0998030, Accumulated Leap Seconds=23
  { 2446766.50, 55.284951 }, //02 December 1986, UT1-UTC=-0.1009513, Accumulated Leap Seconds=23
  { 2446767.50, 55.285894 }, //03 December 1986, UT1-UTC=-0.1018935, Accumulated Leap Seconds=23
  { 2446768.50, 55.286774 }, //04 December 1986, UT1-UTC=-0.1027738, Accumulated Leap Seconds=23
  { 2446769.50, 55.287747 }, //05 December 1986, UT1-UTC=-0.1037469, Accumulated Leap Seconds=23
  { 2446770.50, 55.288921 }, //06 December 1986, UT1-UTC=-0.1049210, Accumulated Leap Seconds=23
  { 2446771.50, 55.290323 }, //07 December 1986, UT1-UTC=-0.1063235, Accumulated Leap Seconds=23
  { 2446772.50, 55.291917 }, //08 December 1986, UT1-UTC=-0.1079170, Accumulated Leap Seconds=23
  { 2446773.50, 55.293618 }, //09 December 1986, UT1-UTC=-0.1096183, Accumulated Leap Seconds=23
  { 2446774.50, 55.295327 }, //10 December 1986, UT1-UTC=-0.1113269, Accumulated Leap Seconds=23
  { 2446775.50, 55.296951 }, //11 December 1986, UT1-UTC=-0.1129513, Accumulated Leap Seconds=23
  { 2446776.50, 55.298420 }, //12 December 1986, UT1-UTC=-0.1144196, Accumulated Leap Seconds=23
  { 2446777.50, 55.299685 }, //13 December 1986, UT1-UTC=-0.1156850, Accumulated Leap Seconds=23
  { 2446778.50, 55.300728 }, //14 December 1986, UT1-UTC=-0.1167276, Accumulated Leap Seconds=23
  { 2446779.50, 55.301556 }, //15 December 1986, UT1-UTC=-0.1175559, Accumulated Leap Seconds=23
  { 2446780.50, 55.302214 }, //16 December 1986, UT1-UTC=-0.1182141, Accumulated Leap Seconds=23
  { 2446781.50, 55.302784 }, //17 December 1986, UT1-UTC=-0.1187843, Accumulated Leap Seconds=23
  { 2446782.50, 55.303362 }, //18 December 1986, UT1-UTC=-0.1193623, Accumulated Leap Seconds=23
  { 2446783.50, 55.304042 }, //19 December 1986, UT1-UTC=-0.1200421, Accumulated Leap Seconds=23
  { 2446784.50, 55.304901 }, //20 December 1986, UT1-UTC=-0.1209007, Accumulated Leap Seconds=23
  { 2446785.50, 55.305991 }, //21 December 1986, UT1-UTC=-0.1219906, Accumulated Leap Seconds=23
  { 2446786.50, 55.307326 }, //22 December 1986, UT1-UTC=-0.1233260, Accumulated Leap Seconds=23
  { 2446787.50, 55.308890 }, //23 December 1986, UT1-UTC=-0.1248903, Accumulated Leap Seconds=23
  { 2446788.50, 55.310637 }, //24 December 1986, UT1-UTC=-0.1266374, Accumulated Leap Seconds=23
  { 2446789.50, 55.312495 }, //25 December 1986, UT1-UTC=-0.1284950, Accumulated Leap Seconds=23
  { 2446790.50, 55.314368 }, //26 December 1986, UT1-UTC=-0.1303680, Accumulated Leap Seconds=23
  { 2446791.50, 55.316144 }, //27 December 1986, UT1-UTC=-0.1321440, Accumulated Leap Seconds=23
  { 2446792.50, 55.317724 }, //28 December 1986, UT1-UTC=-0.1337238, Accumulated Leap Seconds=23
  { 2446793.50, 55.319058 }, //29 December 1986, UT1-UTC=-0.1350576, Accumulated Leap Seconds=23
  { 2446794.50, 55.320172 }, //30 December 1986, UT1-UTC=-0.1361719, Accumulated Leap Seconds=23
  { 2446795.50, 55.321172 }, //31 December 1986, UT1-UTC=-0.1371719, Accumulated Leap Seconds=23
  { 2446796.50, 55.322210 }, //01 January 1987, UT1-UTC=-0.1382105, Accumulated Leap Seconds=23
  { 2446797.50, 55.323427 }, //02 January 1987, UT1-UTC=-0.1394269, Accumulated Leap Seconds=23
  { 2446798.50, 55.324893 }, //03 January 1987, UT1-UTC=-0.1408929, Accumulated Leap Seconds=23
  { 2446799.50, 55.326592 }, //04 January 1987, UT1-UTC=-0.1425920, Accumulated Leap Seconds=23
  { 2446800.50, 55.328436 }, //05 January 1987, UT1-UTC=-0.1444363, Accumulated Leap Seconds=23
  { 2446801.50, 55.330304 }, //06 January 1987, UT1-UTC=-0.1463044, Accumulated Leap Seconds=23
  { 2446802.50, 55.332079 }, //07 January 1987, UT1-UTC=-0.1480795, Accumulated Leap Seconds=23
  { 2446803.50, 55.333676 }, //08 January 1987, UT1-UTC=-0.1496760, Accumulated Leap Seconds=23
  { 2446804.50, 55.335045 }, //09 January 1987, UT1-UTC=-0.1510451, Accumulated Leap Seconds=23
  { 2446805.50, 55.336172 }, //10 January 1987, UT1-UTC=-0.1521722, Accumulated Leap Seconds=23
  { 2446806.50, 55.337077 }, //11 January 1987, UT1-UTC=-0.1530771, Accumulated Leap Seconds=23
  { 2446807.50, 55.337805 }, //12 January 1987, UT1-UTC=-0.1538053, Accumulated Leap Seconds=23
  { 2446808.50, 55.338423 }, //13 January 1987, UT1-UTC=-0.1544225, Accumulated Leap Seconds=23
  { 2446809.50, 55.339009 }, //14 January 1987, UT1-UTC=-0.1550095, Accumulated Leap Seconds=23
  { 2446810.50, 55.339650 }, //15 January 1987, UT1-UTC=-0.1556502, Accumulated Leap Seconds=23
  { 2446811.50, 55.340415 }, //16 January 1987, UT1-UTC=-0.1564149, Accumulated Leap Seconds=23
  { 2446812.50, 55.341361 }, //17 January 1987, UT1-UTC=-0.1573613, Accumulated Leap Seconds=23
  { 2446813.50, 55.342521 }, //18 January 1987, UT1-UTC=-0.1585208, Accumulated Leap Seconds=23
  { 2446814.50, 55.343894 }, //19 January 1987, UT1-UTC=-0.1598938, Accumulated Leap Seconds=23
  { 2446815.50, 55.345449 }, //20 January 1987, UT1-UTC=-0.1614490, Accumulated Leap Seconds=23
  { 2446816.50, 55.347120 }, //21 January 1987, UT1-UTC=-0.1631200, Accumulated Leap Seconds=23
  { 2446817.50, 55.348813 }, //22 January 1987, UT1-UTC=-0.1648127, Accumulated Leap Seconds=23
  { 2446818.50, 55.350425 }, //23 January 1987, UT1-UTC=-0.1664250, Accumulated Leap Seconds=23
  { 2446819.50, 55.351862 }, //24 January 1987, UT1-UTC=-0.1678617, Accumulated Leap Seconds=23
  { 2446820.50, 55.353063 }, //25 January 1987, UT1-UTC=-0.1690630, Accumulated Leap Seconds=23
  { 2446821.50, 55.354035 }, //26 January 1987, UT1-UTC=-0.1700355, Accumulated Leap Seconds=23
  { 2446822.50, 55.354862 }, //27 January 1987, UT1-UTC=-0.1708621, Accumulated Leap Seconds=23
  { 2446823.50, 55.355685 }, //28 January 1987, UT1-UTC=-0.1716852, Accumulated Leap Seconds=23
  { 2446824.50, 55.356662 }, //29 January 1987, UT1-UTC=-0.1726620, Accumulated Leap Seconds=23
  { 2446825.50, 55.357909 }, //30 January 1987, UT1-UTC=-0.1739095, Accumulated Leap Seconds=23
  { 2446826.50, 55.359464 }, //31 January 1987, UT1-UTC=-0.1754637, Accumulated Leap Seconds=23
  { 2446827.50, 55.361268 }, //01 February 1987, UT1-UTC=-0.1772676, Accumulated Leap Seconds=23
  { 2446828.50, 55.363203 }, //02 February 1987, UT1-UTC=-0.1792026, Accumulated Leap Seconds=23
  { 2446829.50, 55.365135 }, //03 February 1987, UT1-UTC=-0.1811346, Accumulated Leap Seconds=23
  { 2446830.50, 55.366954 }, //04 February 1987, UT1-UTC=-0.1829543, Accumulated Leap Seconds=23
  { 2446831.50, 55.368591 }, //05 February 1987, UT1-UTC=-0.1845907, Accumulated Leap Seconds=23
  { 2446832.50, 55.370016 }, //06 February 1987, UT1-UTC=-0.1860159, Accumulated Leap Seconds=23
  { 2446833.50, 55.371241 }, //07 February 1987, UT1-UTC=-0.1872412, Accumulated Leap Seconds=23
  { 2446834.50, 55.372304 }, //08 February 1987, UT1-UTC=-0.1883039, Accumulated Leap Seconds=23
  { 2446835.50, 55.373257 }, //09 February 1987, UT1-UTC=-0.1892573, Accumulated Leap Seconds=23
  { 2446836.50, 55.374168 }, //10 February 1987, UT1-UTC=-0.1901681, Accumulated Leap Seconds=23
  { 2446837.50, 55.375117 }, //11 February 1987, UT1-UTC=-0.1911175, Accumulated Leap Seconds=23
  { 2446838.50, 55.376183 }, //12 February 1987, UT1-UTC=-0.1921833, Accumulated Leap Seconds=23
  { 2446839.50, 55.377425 }, //13 February 1987, UT1-UTC=-0.1934247, Accumulated Leap Seconds=23
  { 2446840.50, 55.378876 }, //14 February 1987, UT1-UTC=-0.1948763, Accumulated Leap Seconds=23
  { 2446841.50, 55.380549 }, //15 February 1987, UT1-UTC=-0.1965486, Accumulated Leap Seconds=23
  { 2446842.50, 55.382416 }, //16 February 1987, UT1-UTC=-0.1984156, Accumulated Leap Seconds=23
  { 2446843.50, 55.384414 }, //17 February 1987, UT1-UTC=-0.2004144, Accumulated Leap Seconds=23
  { 2446844.50, 55.386455 }, //18 February 1987, UT1-UTC=-0.2024549, Accumulated Leap Seconds=23
  { 2446845.50, 55.388435 }, //19 February 1987, UT1-UTC=-0.2044354, Accumulated Leap Seconds=23
  { 2446846.50, 55.390269 }, //20 February 1987, UT1-UTC=-0.2062694, Accumulated Leap Seconds=23
  { 2446847.50, 55.391907 }, //21 February 1987, UT1-UTC=-0.2079068, Accumulated Leap Seconds=23
  { 2446848.50, 55.393351 }, //22 February 1987, UT1-UTC=-0.2093507, Accumulated Leap Seconds=23
  { 2446849.50, 55.394668 }, //23 February 1987, UT1-UTC=-0.2106682, Accumulated Leap Seconds=23
  { 2446850.50, 55.395982 }, //24 February 1987, UT1-UTC=-0.2119817, Accumulated Leap Seconds=23
  { 2446851.50, 55.397440 }, //25 February 1987, UT1-UTC=-0.2134399, Accumulated Leap Seconds=23
  { 2446852.50, 55.399174 }, //26 February 1987, UT1-UTC=-0.2151744, Accumulated Leap Seconds=23
  { 2446853.50, 55.401251 }, //27 February 1987, UT1-UTC=-0.2172510, Accumulated Leap Seconds=23
  { 2446854.50, 55.403648 }, //28 February 1987, UT1-UTC=-0.2196479, Accumulated Leap Seconds=23
  { 2446855.50, 55.406262 }, //01 March 1987, UT1-UTC=-0.2222620, Accumulated Leap Seconds=23
  { 2446856.50, 55.408936 }, //02 March 1987, UT1-UTC=-0.2249359, Accumulated Leap Seconds=23
  { 2446857.50, 55.411510 }, //03 March 1987, UT1-UTC=-0.2275097, Accumulated Leap Seconds=23
  { 2446858.50, 55.413862 }, //04 March 1987, UT1-UTC=-0.2298622, Accumulated Leap Seconds=23
  { 2446859.50, 55.415925 }, //05 March 1987, UT1-UTC=-0.2319251, Accumulated Leap Seconds=23
  { 2446860.50, 55.417681 }, //06 March 1987, UT1-UTC=-0.2336810, Accumulated Leap Seconds=23
  { 2446861.50, 55.419165 }, //07 March 1987, UT1-UTC=-0.2351653, Accumulated Leap Seconds=23
  { 2446862.50, 55.420445 }, //08 March 1987, UT1-UTC=-0.2364452, Accumulated Leap Seconds=23
  { 2446863.50, 55.421603 }, //09 March 1987, UT1-UTC=-0.2376034, Accumulated Leap Seconds=23
  { 2446864.50, 55.422731 }, //10 March 1987, UT1-UTC=-0.2387312, Accumulated Leap Seconds=23
  { 2446865.50, 55.423921 }, //11 March 1987, UT1-UTC=-0.2399209, Accumulated Leap Seconds=23
  { 2446866.50, 55.425255 }, //12 March 1987, UT1-UTC=-0.2412554, Accumulated Leap Seconds=23
  { 2446867.50, 55.426790 }, //13 March 1987, UT1-UTC=-0.2427898, Accumulated Leap Seconds=23
  { 2446868.50, 55.428548 }, //14 March 1987, UT1-UTC=-0.2445485, Accumulated Leap Seconds=23
  { 2446869.50, 55.430519 }, //15 March 1987, UT1-UTC=-0.2465190, Accumulated Leap Seconds=23
  { 2446870.50, 55.432649 }, //16 March 1987, UT1-UTC=-0.2486492, Accumulated Leap Seconds=23
  { 2446871.50, 55.434854 }, //17 March 1987, UT1-UTC=-0.2508542, Accumulated Leap Seconds=23
  { 2446872.50, 55.437030 }, //18 March 1987, UT1-UTC=-0.2530304, Accumulated Leap Seconds=23
  { 2446873.50, 55.439078 }, //19 March 1987, UT1-UTC=-0.2550779, Accumulated Leap Seconds=23
  { 2446874.50, 55.440929 }, //20 March 1987, UT1-UTC=-0.2569290, Accumulated Leap Seconds=23
  { 2446875.50, 55.442567 }, //21 March 1987, UT1-UTC=-0.2585674, Accumulated Leap Seconds=23
  { 2446876.50, 55.444030 }, //22 March 1987, UT1-UTC=-0.2600297, Accumulated Leap Seconds=23
  { 2446877.50, 55.445405 }, //23 March 1987, UT1-UTC=-0.2614050, Accumulated Leap Seconds=23
  { 2446878.50, 55.446813 }, //24 March 1987, UT1-UTC=-0.2628126, Accumulated Leap Seconds=23
  { 2446879.50, 55.448368 }, //25 March 1987, UT1-UTC=-0.2643678, Accumulated Leap Seconds=23
  { 2446880.50, 55.450148 }, //26 March 1987, UT1-UTC=-0.2661481, Accumulated Leap Seconds=23
  { 2446881.50, 55.452172 }, //27 March 1987, UT1-UTC=-0.2681724, Accumulated Leap Seconds=23
  { 2446882.50, 55.454387 }, //28 March 1987, UT1-UTC=-0.2703867, Accumulated Leap Seconds=23
  { 2446883.50, 55.456682 }, //29 March 1987, UT1-UTC=-0.2726823, Accumulated Leap Seconds=23
  { 2446884.50, 55.458930 }, //30 March 1987, UT1-UTC=-0.2749299, Accumulated Leap Seconds=23
  { 2446885.50, 55.461016 }, //31 March 1987, UT1-UTC=-0.2770162, Accumulated Leap Seconds=23
  { 2446886.50, 55.462872 }, //01 April 1987, UT1-UTC=-0.2788719, Accumulated Leap Seconds=23
  { 2446887.50, 55.464480 }, //02 April 1987, UT1-UTC=-0.2804804, Accumulated Leap Seconds=23
  { 2446888.50, 55.465870 }, //03 April 1987, UT1-UTC=-0.2818702, Accumulated Leap Seconds=23
  { 2446889.50, 55.467101 }, //04 April 1987, UT1-UTC=-0.2831008, Accumulated Leap Seconds=23
  { 2446890.50, 55.468247 }, //05 April 1987, UT1-UTC=-0.2842471, Accumulated Leap Seconds=23
  { 2446891.50, 55.469384 }, //06 April 1987, UT1-UTC=-0.2853842, Accumulated Leap Seconds=23
  { 2446892.50, 55.470584 }, //07 April 1987, UT1-UTC=-0.2865842, Accumulated Leap Seconds=23
  { 2446893.50, 55.471909 }, //08 April 1987, UT1-UTC=-0.2879088, Accumulated Leap Seconds=23
  { 2446894.50, 55.473403 }, //09 April 1987, UT1-UTC=-0.2894035, Accumulated Leap Seconds=23
  { 2446895.50, 55.475091 }, //10 April 1987, UT1-UTC=-0.2910912, Accumulated Leap Seconds=23
  { 2446896.50, 55.476970 }, //11 April 1987, UT1-UTC=-0.2929698, Accumulated Leap Seconds=23
  { 2446897.50, 55.479007 }, //12 April 1987, UT1-UTC=-0.2950066, Accumulated Leap Seconds=23
  { 2446898.50, 55.481131 }, //13 April 1987, UT1-UTC=-0.2971309, Accumulated Leap Seconds=23
  { 2446899.50, 55.483241 }, //14 April 1987, UT1-UTC=-0.2992408, Accumulated Leap Seconds=23
  { 2446900.50, 55.485227 }, //15 April 1987, UT1-UTC=-0.3012266, Accumulated Leap Seconds=23
  { 2446901.50, 55.487001 }, //16 April 1987, UT1-UTC=-0.3030008, Accumulated Leap Seconds=23
  { 2446902.50, 55.488519 }, //17 April 1987, UT1-UTC=-0.3045195, Accumulated Leap Seconds=23
  { 2446903.50, 55.489809 }, //18 April 1987, UT1-UTC=-0.3058088, Accumulated Leap Seconds=23
  { 2446904.50, 55.490960 }, //19 April 1987, UT1-UTC=-0.3069597, Accumulated Leap Seconds=23
  { 2446905.50, 55.492100 }, //20 April 1987, UT1-UTC=-0.3081004, Accumulated Leap Seconds=23
  { 2446906.50, 55.493360 }, //21 April 1987, UT1-UTC=-0.3093603, Accumulated Leap Seconds=23
  { 2446907.50, 55.494829 }, //22 April 1987, UT1-UTC=-0.3108288, Accumulated Leap Seconds=23
  { 2446908.50, 55.496539 }, //23 April 1987, UT1-UTC=-0.3125389, Accumulated Leap Seconds=23
  { 2446909.50, 55.498483 }, //24 April 1987, UT1-UTC=-0.3144828, Accumulated Leap Seconds=23
  { 2446910.50, 55.500600 }, //25 April 1987, UT1-UTC=-0.3165997, Accumulated Leap Seconds=23
  { 2446911.50, 55.502771 }, //26 April 1987, UT1-UTC=-0.3187708, Accumulated Leap Seconds=23
  { 2446912.50, 55.504866 }, //27 April 1987, UT1-UTC=-0.3208656, Accumulated Leap Seconds=23
  { 2446913.50, 55.506780 }, //28 April 1987, UT1-UTC=-0.3227795, Accumulated Leap Seconds=23
  { 2446914.50, 55.508454 }, //29 April 1987, UT1-UTC=-0.3244536, Accumulated Leap Seconds=23
  { 2446915.50, 55.509878 }, //30 April 1987, UT1-UTC=-0.3258783, Accumulated Leap Seconds=23
  { 2446916.50, 55.511091 }, //01 May 1987, UT1-UTC=-0.3270909, Accumulated Leap Seconds=23
  { 2446917.50, 55.512154 }, //02 May 1987, UT1-UTC=-0.3281539, Accumulated Leap Seconds=23
  { 2446918.50, 55.513140 }, //03 May 1987, UT1-UTC=-0.3291403, Accumulated Leap Seconds=23
  { 2446919.50, 55.514119 }, //04 May 1987, UT1-UTC=-0.3301188, Accumulated Leap Seconds=23
  { 2446920.50, 55.515151 }, //05 May 1987, UT1-UTC=-0.3311506, Accumulated Leap Seconds=23
  { 2446921.50, 55.516296 }, //06 May 1987, UT1-UTC=-0.3322962, Accumulated Leap Seconds=23
  { 2446922.50, 55.517596 }, //07 May 1987, UT1-UTC=-0.3335959, Accumulated Leap Seconds=23
  { 2446923.50, 55.519070 }, //08 May 1987, UT1-UTC=-0.3350698, Accumulated Leap Seconds=23
  { 2446924.50, 55.520727 }, //09 May 1987, UT1-UTC=-0.3367271, Accumulated Leap Seconds=23
  { 2446925.50, 55.522540 }, //10 May 1987, UT1-UTC=-0.3385405, Accumulated Leap Seconds=23
  { 2446926.50, 55.524424 }, //11 May 1987, UT1-UTC=-0.3404244, Accumulated Leap Seconds=23
  { 2446927.50, 55.526258 }, //12 May 1987, UT1-UTC=-0.3422580, Accumulated Leap Seconds=23
  { 2446928.50, 55.527920 }, //13 May 1987, UT1-UTC=-0.3439203, Accumulated Leap Seconds=23
  { 2446929.50, 55.529331 }, //14 May 1987, UT1-UTC=-0.3453306, Accumulated Leap Seconds=23
  { 2446930.50, 55.530479 }, //15 May 1987, UT1-UTC=-0.3464794, Accumulated Leap Seconds=23
  { 2446931.50, 55.531443 }, //16 May 1987, UT1-UTC=-0.3474435, Accumulated Leap Seconds=23
  { 2446932.50, 55.532360 }, //17 May 1987, UT1-UTC=-0.3483597, Accumulated Leap Seconds=23
  { 2446933.50, 55.533374 }, //18 May 1987, UT1-UTC=-0.3493744, Accumulated Leap Seconds=23
  { 2446934.50, 55.534597 }, //19 May 1987, UT1-UTC=-0.3505967, Accumulated Leap Seconds=23
  { 2446935.50, 55.536070 }, //20 May 1987, UT1-UTC=-0.3520703, Accumulated Leap Seconds=23
  { 2446936.50, 55.537771 }, //21 May 1987, UT1-UTC=-0.3537712, Accumulated Leap Seconds=23
  { 2446937.50, 55.539623 }, //22 May 1987, UT1-UTC=-0.3556234, Accumulated Leap Seconds=23
  { 2446938.50, 55.541524 }, //23 May 1987, UT1-UTC=-0.3575245, Accumulated Leap Seconds=23
  { 2446939.50, 55.543375 }, //24 May 1987, UT1-UTC=-0.3593754, Accumulated Leap Seconds=23
  { 2446940.50, 55.545099 }, //25 May 1987, UT1-UTC=-0.3610990, Accumulated Leap Seconds=23
  { 2446941.50, 55.546635 }, //26 May 1987, UT1-UTC=-0.3626348, Accumulated Leap Seconds=23
  { 2446942.50, 55.547953 }, //27 May 1987, UT1-UTC=-0.3639528, Accumulated Leap Seconds=23
  { 2446943.50, 55.549060 }, //28 May 1987, UT1-UTC=-0.3650600, Accumulated Leap Seconds=23
  { 2446944.50, 55.549993 }, //29 May 1987, UT1-UTC=-0.3659926, Accumulated Leap Seconds=23
  { 2446945.50, 55.550807 }, //30 May 1987, UT1-UTC=-0.3668065, Accumulated Leap Seconds=23
  { 2446946.50, 55.551578 }, //31 May 1987, UT1-UTC=-0.3675780, Accumulated Leap Seconds=23
  { 2446947.50, 55.552378 }, //01 June 1987, UT1-UTC=-0.3683777, Accumulated Leap Seconds=23
  { 2446948.50, 55.553259 }, //02 June 1987, UT1-UTC=-0.3692588, Accumulated Leap Seconds=23
  { 2446949.50, 55.554253 }, //03 June 1987, UT1-UTC=-0.3702526, Accumulated Leap Seconds=23
  { 2446950.50, 55.555371 }, //04 June 1987, UT1-UTC=-0.3713708, Accumulated Leap Seconds=23
  { 2446951.50, 55.556629 }, //05 June 1987, UT1-UTC=-0.3726286, Accumulated Leap Seconds=23
  { 2446952.50, 55.558026 }, //06 June 1987, UT1-UTC=-0.3740264, Accumulated Leap Seconds=23
  { 2446953.50, 55.559534 }, //07 June 1987, UT1-UTC=-0.3755337, Accumulated Leap Seconds=23
  { 2446954.50, 55.561073 }, //08 June 1987, UT1-UTC=-0.3770730, Accumulated Leap Seconds=23
  { 2446955.50, 55.562524 }, //09 June 1987, UT1-UTC=-0.3785244, Accumulated Leap Seconds=23
  { 2446956.50, 55.563782 }, //10 June 1987, UT1-UTC=-0.3797816, Accumulated Leap Seconds=23
  { 2446957.50, 55.564794 }, //11 June 1987, UT1-UTC=-0.3807935, Accumulated Leap Seconds=23
  { 2446958.50, 55.565590 }, //12 June 1987, UT1-UTC=-0.3815903, Accumulated Leap Seconds=23
  { 2446959.50, 55.566279 }, //13 June 1987, UT1-UTC=-0.3822791, Accumulated Leap Seconds=23
  { 2446960.50, 55.567005 }, //14 June 1987, UT1-UTC=-0.3830048, Accumulated Leap Seconds=23
  { 2446961.50, 55.567900 }, //15 June 1987, UT1-UTC=-0.3839004, Accumulated Leap Seconds=23
  { 2446962.50, 55.569036 }, //16 June 1987, UT1-UTC=-0.3850359, Accumulated Leap Seconds=23
  { 2446963.50, 55.570398 }, //17 June 1987, UT1-UTC=-0.3863981, Accumulated Leap Seconds=23
  { 2446964.50, 55.571905 }, //18 June 1987, UT1-UTC=-0.3879049, Accumulated Leap Seconds=23
  { 2446965.50, 55.573443 }, //19 June 1987, UT1-UTC=-0.3894426, Accumulated Leap Seconds=23
  { 2446966.50, 55.574900 }, //20 June 1987, UT1-UTC=-0.3908996, Accumulated Leap Seconds=23
  { 2446967.50, 55.576182 }, //21 June 1987, UT1-UTC=-0.3921822, Accumulated Leap Seconds=23
  { 2446968.50, 55.577234 }, //22 June 1987, UT1-UTC=-0.3932339, Accumulated Leap Seconds=23
  { 2446969.50, 55.578036 }, //23 June 1987, UT1-UTC=-0.3940360, Accumulated Leap Seconds=23
  { 2446970.50, 55.578603 }, //24 June 1987, UT1-UTC=-0.3946032, Accumulated Leap Seconds=23
  { 2446971.50, 55.578975 }, //25 June 1987, UT1-UTC=-0.3949750, Accumulated Leap Seconds=23
  { 2446972.50, 55.579212 }, //26 June 1987, UT1-UTC=-0.3952123, Accumulated Leap Seconds=23
  { 2446973.50, 55.579403 }, //27 June 1987, UT1-UTC=-0.3954025, Accumulated Leap Seconds=23
  { 2446974.50, 55.579634 }, //28 June 1987, UT1-UTC=-0.3956345, Accumulated Leap Seconds=23
  { 2446975.50, 55.579983 }, //29 June 1987, UT1-UTC=-0.3959826, Accumulated Leap Seconds=23
  { 2446976.50, 55.580494 }, //30 June 1987, UT1-UTC=-0.3964936, Accumulated Leap Seconds=23
  { 2446977.50, 55.581188 }, //01 July 1987, UT1-UTC=-0.3971877, Accumulated Leap Seconds=23
  { 2446978.50, 55.582057 }, //02 July 1987, UT1-UTC=-0.3980574, Accumulated Leap Seconds=23
  { 2446979.50, 55.583066 }, //03 July 1987, UT1-UTC=-0.3990660, Accumulated Leap Seconds=23
  { 2446980.50, 55.584153 }, //04 July 1987, UT1-UTC=-0.4001529, Accumulated Leap Seconds=23
  { 2446981.50, 55.585238 }, //05 July 1987, UT1-UTC=-0.4012378, Accumulated Leap Seconds=23
  { 2446982.50, 55.586227 }, //06 July 1987, UT1-UTC=-0.4022275, Accumulated Leap Seconds=23
  { 2446983.50, 55.587038 }, //07 July 1987, UT1-UTC=-0.4030377, Accumulated Leap Seconds=23
  { 2446984.50, 55.587621 }, //08 July 1987, UT1-UTC=-0.4036215, Accumulated Leap Seconds=23
  { 2446985.50, 55.587992 }, //09 July 1987, UT1-UTC=-0.4039925, Accumulated Leap Seconds=23
  { 2446986.50, 55.588230 }, //10 July 1987, UT1-UTC=-0.4042298, Accumulated Leap Seconds=23
  { 2446987.50, 55.588470 }, //11 July 1987, UT1-UTC=-0.4044705, Accumulated Leap Seconds=23
  { 2446988.50, 55.588867 }, //12 July 1987, UT1-UTC=-0.4048669, Accumulated Leap Seconds=23
  { 2446989.50, 55.589533 }, //13 July 1987, UT1-UTC=-0.4055332, Accumulated Leap Seconds=23
  { 2446990.50, 55.590493 }, //14 July 1987, UT1-UTC=-0.4064930, Accumulated Leap Seconds=23
  { 2446991.50, 55.591661 }, //15 July 1987, UT1-UTC=-0.4076606, Accumulated Leap Seconds=23
  { 2446992.50, 55.592892 }, //16 July 1987, UT1-UTC=-0.4088921, Accumulated Leap Seconds=23
  { 2446993.50, 55.594043 }, //17 July 1987, UT1-UTC=-0.4100431, Accumulated Leap Seconds=23
  { 2446994.50, 55.595008 }, //18 July 1987, UT1-UTC=-0.4110080, Accumulated Leap Seconds=23
  { 2446995.50, 55.595738 }, //19 July 1987, UT1-UTC=-0.4117376, Accumulated Leap Seconds=23
  { 2446996.50, 55.596217 }, //20 July 1987, UT1-UTC=-0.4122167, Accumulated Leap Seconds=23
  { 2446997.50, 55.596467 }, //21 July 1987, UT1-UTC=-0.4124667, Accumulated Leap Seconds=23
  { 2446998.50, 55.596538 }, //22 July 1987, UT1-UTC=-0.4125380, Accumulated Leap Seconds=23
  { 2446999.50, 55.596504 }, //23 July 1987, UT1-UTC=-0.4125039, Accumulated Leap Seconds=23
  { 2447000.50, 55.596445 }, //24 July 1987, UT1-UTC=-0.4124448, Accumulated Leap Seconds=23
  { 2447001.50, 55.596440 }, //25 July 1987, UT1-UTC=-0.4124401, Accumulated Leap Seconds=23
  { 2447002.50, 55.596553 }, //26 July 1987, UT1-UTC=-0.4125534, Accumulated Leap Seconds=23
  { 2447003.50, 55.596827 }, //27 July 1987, UT1-UTC=-0.4128271, Accumulated Leap Seconds=23
  { 2447004.50, 55.597282 }, //28 July 1987, UT1-UTC=-0.4132823, Accumulated Leap Seconds=23
  { 2447005.50, 55.597916 }, //29 July 1987, UT1-UTC=-0.4139161, Accumulated Leap Seconds=23
  { 2447006.50, 55.598694 }, //30 July 1987, UT1-UTC=-0.4146942, Accumulated Leap Seconds=23
  { 2447007.50, 55.599559 }, //31 July 1987, UT1-UTC=-0.4155590, Accumulated Leap Seconds=23
  { 2447008.50, 55.600437 }, //01 August 1987, UT1-UTC=-0.4164372, Accumulated Leap Seconds=23
  { 2447009.50, 55.601251 }, //02 August 1987, UT1-UTC=-0.4172506, Accumulated Leap Seconds=23
  { 2447010.50, 55.601930 }, //03 August 1987, UT1-UTC=-0.4179303, Accumulated Leap Seconds=23
  { 2447011.50, 55.602424 }, //04 August 1987, UT1-UTC=-0.4184242, Accumulated Leap Seconds=23
  { 2447012.50, 55.602722 }, //05 August 1987, UT1-UTC=-0.4187224, Accumulated Leap Seconds=23
  { 2447013.50, 55.602880 }, //06 August 1987, UT1-UTC=-0.4188798, Accumulated Leap Seconds=23
  { 2447014.50, 55.603019 }, //07 August 1987, UT1-UTC=-0.4190189, Accumulated Leap Seconds=23
  { 2447015.50, 55.603302 }, //08 August 1987, UT1-UTC=-0.4193021, Accumulated Leap Seconds=23
  { 2447016.50, 55.603856 }, //09 August 1987, UT1-UTC=-0.4198559, Accumulated Leap Seconds=23
  { 2447017.50, 55.604730 }, //10 August 1987, UT1-UTC=-0.4207305, Accumulated Leap Seconds=23
  { 2447018.50, 55.605884 }, //11 August 1987, UT1-UTC=-0.4218842, Accumulated Leap Seconds=23
  { 2447019.50, 55.607205 }, //12 August 1987, UT1-UTC=-0.4232046, Accumulated Leap Seconds=23
  { 2447020.50, 55.608542 }, //13 August 1987, UT1-UTC=-0.4245424, Accumulated Leap Seconds=23
  { 2447021.50, 55.609754 }, //14 August 1987, UT1-UTC=-0.4257538, Accumulated Leap Seconds=23
  { 2447022.50, 55.610746 }, //15 August 1987, UT1-UTC=-0.4267460, Accumulated Leap Seconds=23
  { 2447023.50, 55.611493 }, //16 August 1987, UT1-UTC=-0.4274927, Accumulated Leap Seconds=23
  { 2447024.50, 55.612023 }, //17 August 1987, UT1-UTC=-0.4280233, Accumulated Leap Seconds=23
  { 2447025.50, 55.612403 }, //18 August 1987, UT1-UTC=-0.4284028, Accumulated Leap Seconds=23
  { 2447026.50, 55.612706 }, //19 August 1987, UT1-UTC=-0.4287061, Accumulated Leap Seconds=23
  { 2447027.50, 55.613011 }, //20 August 1987, UT1-UTC=-0.4290106, Accumulated Leap Seconds=23
  { 2447028.50, 55.613394 }, //21 August 1987, UT1-UTC=-0.4293936, Accumulated Leap Seconds=23
  { 2447029.50, 55.613929 }, //22 August 1987, UT1-UTC=-0.4299291, Accumulated Leap Seconds=23
  { 2447030.50, 55.614677 }, //23 August 1987, UT1-UTC=-0.4306769, Accumulated Leap Seconds=23
  { 2447031.50, 55.615663 }, //24 August 1987, UT1-UTC=-0.4316628, Accumulated Leap Seconds=23
  { 2447032.50, 55.616869 }, //25 August 1987, UT1-UTC=-0.4328692, Accumulated Leap Seconds=23
  { 2447033.50, 55.618254 }, //26 August 1987, UT1-UTC=-0.4342537, Accumulated Leap Seconds=23
  { 2447034.50, 55.619753 }, //27 August 1987, UT1-UTC=-0.4357534, Accumulated Leap Seconds=23
  { 2447035.50, 55.621292 }, //28 August 1987, UT1-UTC=-0.4372917, Accumulated Leap Seconds=23
  { 2447036.50, 55.622785 }, //29 August 1987, UT1-UTC=-0.4387851, Accumulated Leap Seconds=23
  { 2447037.50, 55.624140 }, //30 August 1987, UT1-UTC=-0.4401396, Accumulated Leap Seconds=23
  { 2447038.50, 55.625287 }, //31 August 1987, UT1-UTC=-0.4412866, Accumulated Leap Seconds=23
  { 2447039.50, 55.626202 }, //01 September 1987, UT1-UTC=-0.4422020, Accumulated Leap Seconds=23
  { 2447040.50, 55.626922 }, //02 September 1987, UT1-UTC=-0.4429225, Accumulated Leap Seconds=23
  { 2447041.50, 55.627551 }, //03 September 1987, UT1-UTC=-0.4435509, Accumulated Leap Seconds=23
  { 2447042.50, 55.628234 }, //04 September 1987, UT1-UTC=-0.4442340, Accumulated Leap Seconds=23
  { 2447043.50, 55.629124 }, //05 September 1987, UT1-UTC=-0.4451240, Accumulated Leap Seconds=23
  { 2447044.50, 55.630334 }, //06 September 1987, UT1-UTC=-0.4463339, Accumulated Leap Seconds=23
  { 2447045.50, 55.631896 }, //07 September 1987, UT1-UTC=-0.4478957, Accumulated Leap Seconds=23
  { 2447046.50, 55.633729 }, //08 September 1987, UT1-UTC=-0.4497292, Accumulated Leap Seconds=23
  { 2447047.50, 55.635681 }, //09 September 1987, UT1-UTC=-0.4516813, Accumulated Leap Seconds=23
  { 2447048.50, 55.637581 }, //10 September 1987, UT1-UTC=-0.4535815, Accumulated Leap Seconds=23
  { 2447049.50, 55.639292 }, //11 September 1987, UT1-UTC=-0.4552923, Accumulated Leap Seconds=23
  { 2447050.50, 55.640741 }, //12 September 1987, UT1-UTC=-0.4567412, Accumulated Leap Seconds=23
  { 2447051.50, 55.641916 }, //13 September 1987, UT1-UTC=-0.4579157, Accumulated Leap Seconds=23
  { 2447052.50, 55.642852 }, //14 September 1987, UT1-UTC=-0.4588525, Accumulated Leap Seconds=23
  { 2447053.50, 55.643623 }, //15 September 1987, UT1-UTC=-0.4596229, Accumulated Leap Seconds=23
  { 2447054.50, 55.644319 }, //16 September 1987, UT1-UTC=-0.4603190, Accumulated Leap Seconds=23
  { 2447055.50, 55.645044 }, //17 September 1987, UT1-UTC=-0.4610445, Accumulated Leap Seconds=23
  { 2447056.50, 55.645892 }, //18 September 1987, UT1-UTC=-0.4618919, Accumulated Leap Seconds=23
  { 2447057.50, 55.646925 }, //19 September 1987, UT1-UTC=-0.4629249, Accumulated Leap Seconds=23
  { 2447058.50, 55.648184 }, //20 September 1987, UT1-UTC=-0.4641842, Accumulated Leap Seconds=23
  { 2447059.50, 55.649686 }, //21 September 1987, UT1-UTC=-0.4656859, Accumulated Leap Seconds=23
  { 2447060.50, 55.651419 }, //22 September 1987, UT1-UTC=-0.4674193, Accumulated Leap Seconds=23
  { 2447061.50, 55.653326 }, //23 September 1987, UT1-UTC=-0.4693262, Accumulated Leap Seconds=23
  { 2447062.50, 55.655317 }, //24 September 1987, UT1-UTC=-0.4713171, Accumulated Leap Seconds=23
  { 2447063.50, 55.657290 }, //25 September 1987, UT1-UTC=-0.4732897, Accumulated Leap Seconds=23
  { 2447064.50, 55.659147 }, //26 September 1987, UT1-UTC=-0.4751469, Accumulated Leap Seconds=23
  { 2447065.50, 55.660812 }, //27 September 1987, UT1-UTC=-0.4768124, Accumulated Leap Seconds=23
  { 2447066.50, 55.662245 }, //28 September 1987, UT1-UTC=-0.4782455, Accumulated Leap Seconds=23
  { 2447067.50, 55.663464 }, //29 September 1987, UT1-UTC=-0.4794643, Accumulated Leap Seconds=23
  { 2447068.50, 55.664549 }, //30 September 1987, UT1-UTC=-0.4805488, Accumulated Leap Seconds=23
  { 2447069.50, 55.665627 }, //01 October 1987, UT1-UTC=-0.4816271, Accumulated Leap Seconds=23
  { 2447070.50, 55.666847 }, //02 October 1987, UT1-UTC=-0.4828474, Accumulated Leap Seconds=23
  { 2447071.50, 55.668330 }, //03 October 1987, UT1-UTC=-0.4843299, Accumulated Leap Seconds=23
  { 2447072.50, 55.670133 }, //04 October 1987, UT1-UTC=-0.4861326, Accumulated Leap Seconds=23
  { 2447073.50, 55.672234 }, //05 October 1987, UT1-UTC=-0.4882336, Accumulated Leap Seconds=23
  { 2447074.50, 55.674534 }, //06 October 1987, UT1-UTC=-0.4905342, Accumulated Leap Seconds=23
  { 2447075.50, 55.676887 }, //07 October 1987, UT1-UTC=-0.4928867, Accumulated Leap Seconds=23
  { 2447076.50, 55.679130 }, //08 October 1987, UT1-UTC=-0.4951299, Accumulated Leap Seconds=23
  { 2447077.50, 55.681141 }, //09 October 1987, UT1-UTC=-0.4971413, Accumulated Leap Seconds=23
  { 2447078.50, 55.682866 }, //10 October 1987, UT1-UTC=-0.4988659, Accumulated Leap Seconds=23
  { 2447079.50, 55.684313 }, //11 October 1987, UT1-UTC=-0.5003134, Accumulated Leap Seconds=23
  { 2447080.50, 55.685541 }, //12 October 1987, UT1-UTC=-0.5015413, Accumulated Leap Seconds=23
  { 2447081.50, 55.686639 }, //13 October 1987, UT1-UTC=-0.5026389, Accumulated Leap Seconds=23
  { 2447082.50, 55.687702 }, //14 October 1987, UT1-UTC=-0.5037024, Accumulated Leap Seconds=23
  { 2447083.50, 55.688820 }, //15 October 1987, UT1-UTC=-0.5048198, Accumulated Leap Seconds=23
  { 2447084.50, 55.690063 }, //16 October 1987, UT1-UTC=-0.5060628, Accumulated Leap Seconds=23
  { 2447085.50, 55.691483 }, //17 October 1987, UT1-UTC=-0.5074828, Accumulated Leap Seconds=23
  { 2447086.50, 55.693100 }, //18 October 1987, UT1-UTC=-0.5090996, Accumulated Leap Seconds=23
  { 2447087.50, 55.694901 }, //19 October 1987, UT1-UTC=-0.5109007, Accumulated Leap Seconds=23
  { 2447088.50, 55.696849 }, //20 October 1987, UT1-UTC=-0.5128490, Accumulated Leap Seconds=23
  { 2447089.50, 55.698883 }, //21 October 1987, UT1-UTC=-0.5148832, Accumulated Leap Seconds=23
  { 2447090.50, 55.700926 }, //22 October 1987, UT1-UTC=-0.5169265, Accumulated Leap Seconds=23
  { 2447091.50, 55.702900 }, //23 October 1987, UT1-UTC=-0.5189004, Accumulated Leap Seconds=23
  { 2447092.50, 55.704729 }, //24 October 1987, UT1-UTC=-0.5207295, Accumulated Leap Seconds=23
  { 2447093.50, 55.706353 }, //25 October 1987, UT1-UTC=-0.5223526, Accumulated Leap Seconds=23
  { 2447094.50, 55.707760 }, //26 October 1987, UT1-UTC=-0.5237596, Accumulated Leap Seconds=23
  { 2447095.50, 55.709005 }, //27 October 1987, UT1-UTC=-0.5250048, Accumulated Leap Seconds=23
  { 2447096.50, 55.710206 }, //28 October 1987, UT1-UTC=-0.5262058, Accumulated Leap Seconds=23
  { 2447097.50, 55.711500 }, //29 October 1987, UT1-UTC=-0.5274997, Accumulated Leap Seconds=23
  { 2447098.50, 55.713001 }, //30 October 1987, UT1-UTC=-0.5290006, Accumulated Leap Seconds=23
  { 2447099.50, 55.714769 }, //31 October 1987, UT1-UTC=-0.5307692, Accumulated Leap Seconds=23
  { 2447100.50, 55.716800 }, //01 November 1987, UT1-UTC=-0.5327999, Accumulated Leap Seconds=23
  { 2447101.50, 55.719025 }, //02 November 1987, UT1-UTC=-0.5350249, Accumulated Leap Seconds=23
  { 2447102.50, 55.721327 }, //03 November 1987, UT1-UTC=-0.5373266, Accumulated Leap Seconds=23
  { 2447103.50, 55.723571 }, //04 November 1987, UT1-UTC=-0.5395710, Accumulated Leap Seconds=23
  { 2447104.50, 55.725646 }, //05 November 1987, UT1-UTC=-0.5416456, Accumulated Leap Seconds=23
  { 2447105.50, 55.727484 }, //06 November 1987, UT1-UTC=-0.5434840, Accumulated Leap Seconds=23
  { 2447106.50, 55.729065 }, //07 November 1987, UT1-UTC=-0.5450654, Accumulated Leap Seconds=23
  { 2447107.50, 55.730430 }, //08 November 1987, UT1-UTC=-0.5464298, Accumulated Leap Seconds=23
  { 2447108.50, 55.731658 }, //09 November 1987, UT1-UTC=-0.5476583, Accumulated Leap Seconds=23
  { 2447109.50, 55.732848 }, //10 November 1987, UT1-UTC=-0.5488479, Accumulated Leap Seconds=23
  { 2447110.50, 55.734089 }, //11 November 1987, UT1-UTC=-0.5500894, Accumulated Leap Seconds=23
  { 2447111.50, 55.735457 }, //12 November 1987, UT1-UTC=-0.5514568, Accumulated Leap Seconds=23
  { 2447112.50, 55.736997 }, //13 November 1987, UT1-UTC=-0.5529967, Accumulated Leap Seconds=23
  { 2447113.50, 55.738723 }, //14 November 1987, UT1-UTC=-0.5547231, Accumulated Leap Seconds=23
  { 2447114.50, 55.740618 }, //15 November 1987, UT1-UTC=-0.5566183, Accumulated Leap Seconds=23
  { 2447115.50, 55.742642 }, //16 November 1987, UT1-UTC=-0.5586416, Accumulated Leap Seconds=23
  { 2447116.50, 55.744747 }, //17 November 1987, UT1-UTC=-0.5607468, Accumulated Leap Seconds=23
  { 2447117.50, 55.746872 }, //18 November 1987, UT1-UTC=-0.5628720, Accumulated Leap Seconds=23
  { 2447118.50, 55.748939 }, //19 November 1987, UT1-UTC=-0.5649386, Accumulated Leap Seconds=23
  { 2447119.50, 55.750866 }, //20 November 1987, UT1-UTC=-0.5668659, Accumulated Leap Seconds=23
  { 2447120.50, 55.752594 }, //21 November 1987, UT1-UTC=-0.5685939, Accumulated Leap Seconds=23
  { 2447121.50, 55.754104 }, //22 November 1987, UT1-UTC=-0.5701038, Accumulated Leap Seconds=23
  { 2447122.50, 55.755438 }, //23 November 1987, UT1-UTC=-0.5714376, Accumulated Leap Seconds=23
  { 2447123.50, 55.756694 }, //24 November 1987, UT1-UTC=-0.5726937, Accumulated Leap Seconds=23
  { 2447124.50, 55.757998 }, //25 November 1987, UT1-UTC=-0.5739980, Accumulated Leap Seconds=23
  { 2447125.50, 55.759468 }, //26 November 1987, UT1-UTC=-0.5754680, Accumulated Leap Seconds=23
  { 2447126.50, 55.761184 }, //27 November 1987, UT1-UTC=-0.5771837, Accumulated Leap Seconds=23
  { 2447127.50, 55.763155 }, //28 November 1987, UT1-UTC=-0.5791552, Accumulated Leap Seconds=23
  { 2447128.50, 55.765324 }, //29 November 1987, UT1-UTC=-0.5813241, Accumulated Leap Seconds=23
  { 2447129.50, 55.767585 }, //30 November 1987, UT1-UTC=-0.5835847, Accumulated Leap Seconds=23
  { 2447130.50, 55.769810 }, //01 December 1987, UT1-UTC=-0.5858097, Accumulated Leap Seconds=23
  { 2447131.50, 55.771882 }, //02 December 1987, UT1-UTC=-0.5878816, Accumulated Leap Seconds=23
  { 2447132.50, 55.773723 }, //03 December 1987, UT1-UTC=-0.5897234, Accumulated Leap Seconds=23
  { 2447133.50, 55.775307 }, //04 December 1987, UT1-UTC=-0.5913070, Accumulated Leap Seconds=23
  { 2447134.50, 55.776651 }, //05 December 1987, UT1-UTC=-0.5926508, Accumulated Leap Seconds=23
  { 2447135.50, 55.777810 }, //06 December 1987, UT1-UTC=-0.5938102, Accumulated Leap Seconds=23
  { 2447136.50, 55.778868 }, //07 December 1987, UT1-UTC=-0.5948685, Accumulated Leap Seconds=23
  { 2447137.50, 55.779912 }, //08 December 1987, UT1-UTC=-0.5959118, Accumulated Leap Seconds=23
  { 2447138.50, 55.781013 }, //09 December 1987, UT1-UTC=-0.5970130, Accumulated Leap Seconds=23
  { 2447139.50, 55.782230 }, //10 December 1987, UT1-UTC=-0.5982298, Accumulated Leap Seconds=23
  { 2447140.50, 55.783602 }, //11 December 1987, UT1-UTC=-0.5996019, Accumulated Leap Seconds=23
  { 2447141.50, 55.785141 }, //12 December 1987, UT1-UTC=-0.6011407, Accumulated Leap Seconds=23
  { 2447142.50, 55.786834 }, //13 December 1987, UT1-UTC=-0.6028336, Accumulated Leap Seconds=23
  { 2447143.50, 55.788645 }, //14 December 1987, UT1-UTC=-0.6046452, Accumulated Leap Seconds=23
  { 2447144.50, 55.790520 }, //15 December 1987, UT1-UTC=-0.6065196, Accumulated Leap Seconds=23
  { 2447145.50, 55.792384 }, //16 December 1987, UT1-UTC=-0.6083840, Accumulated Leap Seconds=23
  { 2447146.50, 55.794161 }, //17 December 1987, UT1-UTC=-0.6101611, Accumulated Leap Seconds=23
  { 2447147.50, 55.795782 }, //18 December 1987, UT1-UTC=-0.6117822, Accumulated Leap Seconds=23
  { 2447148.50, 55.797212 }, //19 December 1987, UT1-UTC=-0.6132118, Accumulated Leap Seconds=23
  { 2447149.50, 55.798472 }, //20 December 1987, UT1-UTC=-0.6144719, Accumulated Leap Seconds=23
  { 2447150.50, 55.799651 }, //21 December 1987, UT1-UTC=-0.6156514, Accumulated Leap Seconds=23
  { 2447151.50, 55.800885 }, //22 December 1987, UT1-UTC=-0.6168848, Accumulated Leap Seconds=23
  { 2447152.50, 55.802305 }, //23 December 1987, UT1-UTC=-0.6183055, Accumulated Leap Seconds=23
  { 2447153.50, 55.804001 }, //24 December 1987, UT1-UTC=-0.6200012, Accumulated Leap Seconds=23
  { 2447154.50, 55.805985 }, //25 December 1987, UT1-UTC=-0.6219847, Accumulated Leap Seconds=23
  { 2447155.50, 55.808188 }, //26 December 1987, UT1-UTC=-0.6241884, Accumulated Leap Seconds=23
  { 2447156.50, 55.810488 }, //27 December 1987, UT1-UTC=-0.6264885, Accumulated Leap Seconds=23
  { 2447157.50, 55.812746 }, //28 December 1987, UT1-UTC=-0.6287460, Accumulated Leap Seconds=23
  { 2447158.50, 55.814843 }, //29 December 1987, UT1-UTC=-0.6308434, Accumulated Leap Seconds=23
  { 2447159.50, 55.816704 }, //30 December 1987, UT1-UTC=-0.6327042, Accumulated Leap Seconds=23
  { 2447160.50, 55.818301 }, //31 December 1987, UT1-UTC=-0.6343015, Accumulated Leap Seconds=23
  { 2447161.50, 55.819661 }, //01 January 1988, UT1-UTC= 0.3643391, Accumulated Leap Seconds=24
  { 2447162.50, 55.820843 }, //02 January 1988, UT1-UTC= 0.3631572, Accumulated Leap Seconds=24
  { 2447163.50, 55.821926 }, //03 January 1988, UT1-UTC= 0.3620737, Accumulated Leap Seconds=24
  { 2447164.50, 55.822995 }, //04 January 1988, UT1-UTC= 0.3610047, Accumulated Leap Seconds=24
  { 2447165.50, 55.824121 }, //05 January 1988, UT1-UTC= 0.3598785, Accumulated Leap Seconds=24
  { 2447166.50, 55.825360 }, //06 January 1988, UT1-UTC= 0.3586397, Accumulated Leap Seconds=24
  { 2447167.50, 55.826750 }, //07 January 1988, UT1-UTC= 0.3572503, Accumulated Leap Seconds=24
  { 2447168.50, 55.828302 }, //08 January 1988, UT1-UTC= 0.3556980, Accumulated Leap Seconds=24
  { 2447169.50, 55.830005 }, //09 January 1988, UT1-UTC= 0.3539945, Accumulated Leap Seconds=24
  { 2447170.50, 55.831830 }, //10 January 1988, UT1-UTC= 0.3521697, Accumulated Leap Seconds=24
  { 2447171.50, 55.833719 }, //11 January 1988, UT1-UTC= 0.3502810, Accumulated Leap Seconds=24
  { 2447172.50, 55.835597 }, //12 January 1988, UT1-UTC= 0.3484034, Accumulated Leap Seconds=24
  { 2447173.50, 55.837392 }, //13 January 1988, UT1-UTC= 0.3466079, Accumulated Leap Seconds=24
  { 2447174.50, 55.839042 }, //14 January 1988, UT1-UTC= 0.3449580, Accumulated Leap Seconds=24
  { 2447175.50, 55.840504 }, //15 January 1988, UT1-UTC= 0.3434961, Accumulated Leap Seconds=24
  { 2447176.50, 55.841767 }, //16 January 1988, UT1-UTC= 0.3422334, Accumulated Leap Seconds=24
  { 2447177.50, 55.842874 }, //17 January 1988, UT1-UTC= 0.3411257, Accumulated Leap Seconds=24
  { 2447178.50, 55.843930 }, //18 January 1988, UT1-UTC= 0.3400704, Accumulated Leap Seconds=24
  { 2447179.50, 55.845065 }, //19 January 1988, UT1-UTC= 0.3389347, Accumulated Leap Seconds=24
  { 2447180.50, 55.846392 }, //20 January 1988, UT1-UTC= 0.3376077, Accumulated Leap Seconds=24
  { 2447181.50, 55.847964 }, //21 January 1988, UT1-UTC= 0.3360362, Accumulated Leap Seconds=24
  { 2447182.50, 55.849749 }, //22 January 1988, UT1-UTC= 0.3342511, Accumulated Leap Seconds=24
  { 2447183.50, 55.851635 }, //23 January 1988, UT1-UTC= 0.3323647, Accumulated Leap Seconds=24
  { 2447184.50, 55.853475 }, //24 January 1988, UT1-UTC= 0.3305253, Accumulated Leap Seconds=24
  { 2447185.50, 55.855141 }, //25 January 1988, UT1-UTC= 0.3288589, Accumulated Leap Seconds=24
  { 2447186.50, 55.856555 }, //26 January 1988, UT1-UTC= 0.3274454, Accumulated Leap Seconds=24
  { 2447187.50, 55.857696 }, //27 January 1988, UT1-UTC= 0.3263037, Accumulated Leap Seconds=24
  { 2447188.50, 55.858605 }, //28 January 1988, UT1-UTC= 0.3253945, Accumulated Leap Seconds=24
  { 2447189.50, 55.859356 }, //29 January 1988, UT1-UTC= 0.3246443, Accumulated Leap Seconds=24
  { 2447190.50, 55.860031 }, //30 January 1988, UT1-UTC= 0.3239692, Accumulated Leap Seconds=24
  { 2447191.50, 55.860719 }, //31 January 1988, UT1-UTC= 0.3232813, Accumulated Leap Seconds=24
  { 2447192.50, 55.861503 }, //01 February 1988, UT1-UTC= 0.3224972, Accumulated Leap Seconds=24
  { 2447193.50, 55.862451 }, //02 February 1988, UT1-UTC= 0.3215485, Accumulated Leap Seconds=24
  { 2447194.50, 55.863610 }, //03 February 1988, UT1-UTC= 0.3203896, Accumulated Leap Seconds=24
  { 2447195.50, 55.864998 }, //04 February 1988, UT1-UTC= 0.3190024, Accumulated Leap Seconds=24
  { 2447196.50, 55.866601 }, //05 February 1988, UT1-UTC= 0.3173992, Accumulated Leap Seconds=24
  { 2447197.50, 55.868383 }, //06 February 1988, UT1-UTC= 0.3156167, Accumulated Leap Seconds=24
  { 2447198.50, 55.870290 }, //07 February 1988, UT1-UTC= 0.3137102, Accumulated Leap Seconds=24
  { 2447199.50, 55.872254 }, //08 February 1988, UT1-UTC= 0.3117456, Accumulated Leap Seconds=24
  { 2447200.50, 55.874206 }, //09 February 1988, UT1-UTC= 0.3097938, Accumulated Leap Seconds=24
  { 2447201.50, 55.876071 }, //10 February 1988, UT1-UTC= 0.3079291, Accumulated Leap Seconds=24
  { 2447202.50, 55.877787 }, //11 February 1988, UT1-UTC= 0.3062131, Accumulated Leap Seconds=24
  { 2447203.50, 55.879325 }, //12 February 1988, UT1-UTC= 0.3046751, Accumulated Leap Seconds=24
  { 2447204.50, 55.880708 }, //13 February 1988, UT1-UTC= 0.3032924, Accumulated Leap Seconds=24
  { 2447205.50, 55.882026 }, //14 February 1988, UT1-UTC= 0.3019738, Accumulated Leap Seconds=24
  { 2447206.50, 55.883429 }, //15 February 1988, UT1-UTC= 0.3005708, Accumulated Leap Seconds=24
  { 2447207.50, 55.885076 }, //16 February 1988, UT1-UTC= 0.2989235, Accumulated Leap Seconds=24
  { 2447208.50, 55.887079 }, //17 February 1988, UT1-UTC= 0.2969211, Accumulated Leap Seconds=24
  { 2447209.50, 55.889453 }, //18 February 1988, UT1-UTC= 0.2945468, Accumulated Leap Seconds=24
  { 2447210.50, 55.892110 }, //19 February 1988, UT1-UTC= 0.2918898, Accumulated Leap Seconds=24
  { 2447211.50, 55.894879 }, //20 February 1988, UT1-UTC= 0.2891212, Accumulated Leap Seconds=24
  { 2447212.50, 55.897567 }, //21 February 1988, UT1-UTC= 0.2864326, Accumulated Leap Seconds=24
  { 2447213.50, 55.900024 }, //22 February 1988, UT1-UTC= 0.2839763, Accumulated Leap Seconds=24
  { 2447214.50, 55.902164 }, //23 February 1988, UT1-UTC= 0.2818358, Accumulated Leap Seconds=24
  { 2447215.50, 55.903983 }, //24 February 1988, UT1-UTC= 0.2800168, Accumulated Leap Seconds=24
  { 2447216.50, 55.905547 }, //25 February 1988, UT1-UTC= 0.2784534, Accumulated Leap Seconds=24
  { 2447217.50, 55.906955 }, //26 February 1988, UT1-UTC= 0.2770448, Accumulated Leap Seconds=24
  { 2447218.50, 55.908315 }, //27 February 1988, UT1-UTC= 0.2756852, Accumulated Leap Seconds=24
  { 2447219.50, 55.909725 }, //28 February 1988, UT1-UTC= 0.2742752, Accumulated Leap Seconds=24
  { 2447220.50, 55.911267 }, //29 February 1988, UT1-UTC= 0.2727334, Accumulated Leap Seconds=24
  { 2447221.50, 55.912988 }, //01 March 1988, UT1-UTC= 0.2710117, Accumulated Leap Seconds=24
  { 2447222.50, 55.914910 }, //02 March 1988, UT1-UTC= 0.2690901, Accumulated Leap Seconds=24
  { 2447223.50, 55.917021 }, //03 March 1988, UT1-UTC= 0.2669788, Accumulated Leap Seconds=24
  { 2447224.50, 55.919282 }, //04 March 1988, UT1-UTC= 0.2647180, Accumulated Leap Seconds=24
  { 2447225.50, 55.921627 }, //05 March 1988, UT1-UTC= 0.2623726, Accumulated Leap Seconds=24
  { 2447226.50, 55.923976 }, //06 March 1988, UT1-UTC= 0.2600238, Accumulated Leap Seconds=24
  { 2447227.50, 55.926242 }, //07 March 1988, UT1-UTC= 0.2577578, Accumulated Leap Seconds=24
  { 2447228.50, 55.928344 }, //08 March 1988, UT1-UTC= 0.2556557, Accumulated Leap Seconds=24
  { 2447229.50, 55.930219 }, //09 March 1988, UT1-UTC= 0.2537812, Accumulated Leap Seconds=24
  { 2447230.50, 55.931830 }, //10 March 1988, UT1-UTC= 0.2521702, Accumulated Leap Seconds=24
  { 2447231.50, 55.933178 }, //11 March 1988, UT1-UTC= 0.2508219, Accumulated Leap Seconds=24
  { 2447232.50, 55.934331 }, //12 March 1988, UT1-UTC= 0.2496692, Accumulated Leap Seconds=24
  { 2447233.50, 55.935412 }, //13 March 1988, UT1-UTC= 0.2485880, Accumulated Leap Seconds=24
  { 2447234.50, 55.936572 }, //14 March 1988, UT1-UTC= 0.2474275, Accumulated Leap Seconds=24
  { 2447235.50, 55.937950 }, //15 March 1988, UT1-UTC= 0.2460500, Accumulated Leap Seconds=24
  { 2447236.50, 55.939630 }, //16 March 1988, UT1-UTC= 0.2443702, Accumulated Leap Seconds=24
  { 2447237.50, 55.941605 }, //17 March 1988, UT1-UTC= 0.2423948, Accumulated Leap Seconds=24
  { 2447238.50, 55.943777 }, //18 March 1988, UT1-UTC= 0.2402232, Accumulated Leap Seconds=24
  { 2447239.50, 55.945989 }, //19 March 1988, UT1-UTC= 0.2380114, Accumulated Leap Seconds=24
  { 2447240.50, 55.948078 }, //20 March 1988, UT1-UTC= 0.2359220, Accumulated Leap Seconds=24
  { 2447241.50, 55.949923 }, //21 March 1988, UT1-UTC= 0.2340768, Accumulated Leap Seconds=24
  { 2447242.50, 55.951479 }, //22 March 1988, UT1-UTC= 0.2325209, Accumulated Leap Seconds=24
  { 2447243.50, 55.952781 }, //23 March 1988, UT1-UTC= 0.2312189, Accumulated Leap Seconds=24
  { 2447244.50, 55.953916 }, //24 March 1988, UT1-UTC= 0.2300844, Accumulated Leap Seconds=24
  { 2447245.50, 55.954983 }, //25 March 1988, UT1-UTC= 0.2290168, Accumulated Leap Seconds=24
  { 2447246.50, 55.956086 }, //26 March 1988, UT1-UTC= 0.2279140, Accumulated Leap Seconds=24
  { 2447247.50, 55.957308 }, //27 March 1988, UT1-UTC= 0.2266917, Accumulated Leap Seconds=24
  { 2447248.50, 55.958711 }, //28 March 1988, UT1-UTC= 0.2252886, Accumulated Leap Seconds=24
  { 2447249.50, 55.960327 }, //29 March 1988, UT1-UTC= 0.2236735, Accumulated Leap Seconds=24
  { 2447250.50, 55.962156 }, //30 March 1988, UT1-UTC= 0.2218443, Accumulated Leap Seconds=24
  { 2447251.50, 55.964178 }, //31 March 1988, UT1-UTC= 0.2198219, Accumulated Leap Seconds=24
  { 2447252.50, 55.966347 }, //01 April 1988, UT1-UTC= 0.2176526, Accumulated Leap Seconds=24
  { 2447253.50, 55.968597 }, //02 April 1988, UT1-UTC= 0.2154028, Accumulated Leap Seconds=24
  { 2447254.50, 55.970850 }, //03 April 1988, UT1-UTC= 0.2131499, Accumulated Leap Seconds=24
  { 2447255.50, 55.973024 }, //04 April 1988, UT1-UTC= 0.2109755, Accumulated Leap Seconds=24
  { 2447256.50, 55.975051 }, //05 April 1988, UT1-UTC= 0.2089488, Accumulated Leap Seconds=24
  { 2447257.50, 55.976893 }, //06 April 1988, UT1-UTC= 0.2071069, Accumulated Leap Seconds=24
  { 2447258.50, 55.978557 }, //07 April 1988, UT1-UTC= 0.2054428, Accumulated Leap Seconds=24
  { 2447259.50, 55.980098 }, //08 April 1988, UT1-UTC= 0.2039017, Accumulated Leap Seconds=24
  { 2447260.50, 55.981612 }, //09 April 1988, UT1-UTC= 0.2023878, Accumulated Leap Seconds=24
  { 2447261.50, 55.983217 }, //10 April 1988, UT1-UTC= 0.2007825, Accumulated Leap Seconds=24
  { 2447262.50, 55.985030 }, //11 April 1988, UT1-UTC= 0.1989702, Accumulated Leap Seconds=24
  { 2447263.50, 55.987136 }, //12 April 1988, UT1-UTC= 0.1968642, Accumulated Leap Seconds=24
  { 2447264.50, 55.989551 }, //13 April 1988, UT1-UTC= 0.1944488, Accumulated Leap Seconds=24
  { 2447265.50, 55.992205 }, //14 April 1988, UT1-UTC= 0.1917951, Accumulated Leap Seconds=24
  { 2447266.50, 55.994948 }, //15 April 1988, UT1-UTC= 0.1890522, Accumulated Leap Seconds=24
  { 2447267.50, 55.997594 }, //16 April 1988, UT1-UTC= 0.1864062, Accumulated Leap Seconds=24
  { 2447268.50, 55.999982 }, //17 April 1988, UT1-UTC= 0.1840175, Accumulated Leap Seconds=24
  { 2447269.50, 56.002022 }, //18 April 1988, UT1-UTC= 0.1819779, Accumulated Leap Seconds=24
  { 2447270.50, 56.003703 }, //19 April 1988, UT1-UTC= 0.1802974, Accumulated Leap Seconds=24
  { 2447271.50, 56.005086 }, //20 April 1988, UT1-UTC= 0.1789135, Accumulated Leap Seconds=24
  { 2447272.50, 56.006280 }, //21 April 1988, UT1-UTC= 0.1777201, Accumulated Leap Seconds=24
  { 2447273.50, 56.007393 }, //22 April 1988, UT1-UTC= 0.1766070, Accumulated Leap Seconds=24
  { 2447274.50, 56.008522 }, //23 April 1988, UT1-UTC= 0.1754784, Accumulated Leap Seconds=24
  { 2447275.50, 56.009740 }, //24 April 1988, UT1-UTC= 0.1742598, Accumulated Leap Seconds=24
  { 2447276.50, 56.011101 }, //25 April 1988, UT1-UTC= 0.1728986, Accumulated Leap Seconds=24
  { 2447277.50, 56.012632 }, //26 April 1988, UT1-UTC= 0.1713676, Accumulated Leap Seconds=24
  { 2447278.50, 56.014335 }, //27 April 1988, UT1-UTC= 0.1696654, Accumulated Leap Seconds=24
  { 2447279.50, 56.016184 }, //28 April 1988, UT1-UTC= 0.1678156, Accumulated Leap Seconds=24
  { 2447280.50, 56.018133 }, //29 April 1988, UT1-UTC= 0.1658668, Accumulated Leap Seconds=24
  { 2447281.50, 56.020104 }, //30 April 1988, UT1-UTC= 0.1638957, Accumulated Leap Seconds=24
  { 2447282.50, 56.022010 }, //01 May 1988, UT1-UTC= 0.1619898, Accumulated Leap Seconds=24
  { 2447283.50, 56.023772 }, //02 May 1988, UT1-UTC= 0.1602277, Accumulated Leap Seconds=24
  { 2447284.50, 56.025340 }, //03 May 1988, UT1-UTC= 0.1586604, Accumulated Leap Seconds=24
  { 2447285.50, 56.026707 }, //04 May 1988, UT1-UTC= 0.1572926, Accumulated Leap Seconds=24
  { 2447286.50, 56.027926 }, //05 May 1988, UT1-UTC= 0.1560735, Accumulated Leap Seconds=24
  { 2447287.50, 56.029092 }, //06 May 1988, UT1-UTC= 0.1549079, Accumulated Leap Seconds=24
  { 2447288.50, 56.030325 }, //07 May 1988, UT1-UTC= 0.1536752, Accumulated Leap Seconds=24
  { 2447289.50, 56.031740 }, //08 May 1988, UT1-UTC= 0.1522597, Accumulated Leap Seconds=24
  { 2447290.50, 56.033419 }, //09 May 1988, UT1-UTC= 0.1505810, Accumulated Leap Seconds=24
  { 2447291.50, 56.035383 }, //10 May 1988, UT1-UTC= 0.1486168, Accumulated Leap Seconds=24
  { 2447292.50, 56.037585 }, //11 May 1988, UT1-UTC= 0.1464149, Accumulated Leap Seconds=24
  { 2447293.50, 56.039916 }, //12 May 1988, UT1-UTC= 0.1440838, Accumulated Leap Seconds=24
  { 2447294.50, 56.042236 }, //13 May 1988, UT1-UTC= 0.1417637, Accumulated Leap Seconds=24
  { 2447295.50, 56.044411 }, //14 May 1988, UT1-UTC= 0.1395894, Accumulated Leap Seconds=24
  { 2447296.50, 56.046336 }, //15 May 1988, UT1-UTC= 0.1376640, Accumulated Leap Seconds=24
  { 2447297.50, 56.047971 }, //16 May 1988, UT1-UTC= 0.1360288, Accumulated Leap Seconds=24
  { 2447298.50, 56.049343 }, //17 May 1988, UT1-UTC= 0.1346574, Accumulated Leap Seconds=24
  { 2447299.50, 56.050521 }, //18 May 1988, UT1-UTC= 0.1334793, Accumulated Leap Seconds=24
  { 2447300.50, 56.051586 }, //19 May 1988, UT1-UTC= 0.1324141, Accumulated Leap Seconds=24
  { 2447301.50, 56.052625 }, //20 May 1988, UT1-UTC= 0.1313745, Accumulated Leap Seconds=24
  { 2447302.50, 56.053720 }, //21 May 1988, UT1-UTC= 0.1302795, Accumulated Leap Seconds=24
  { 2447303.50, 56.054928 }, //22 May 1988, UT1-UTC= 0.1290715, Accumulated Leap Seconds=24
  { 2447304.50, 56.056280 }, //23 May 1988, UT1-UTC= 0.1277202, Accumulated Leap Seconds=24
  { 2447305.50, 56.057783 }, //24 May 1988, UT1-UTC= 0.1262165, Accumulated Leap Seconds=24
  { 2447306.50, 56.059421 }, //25 May 1988, UT1-UTC= 0.1245788, Accumulated Leap Seconds=24
  { 2447307.50, 56.061148 }, //26 May 1988, UT1-UTC= 0.1228523, Accumulated Leap Seconds=24
  { 2447308.50, 56.062904 }, //27 May 1988, UT1-UTC= 0.1210956, Accumulated Leap Seconds=24
  { 2447309.50, 56.064627 }, //28 May 1988, UT1-UTC= 0.1193732, Accumulated Leap Seconds=24
  { 2447310.50, 56.066250 }, //29 May 1988, UT1-UTC= 0.1177502, Accumulated Leap Seconds=24
  { 2447311.50, 56.067704 }, //30 May 1988, UT1-UTC= 0.1162955, Accumulated Leap Seconds=24
  { 2447312.50, 56.068951 }, //31 May 1988, UT1-UTC= 0.1150485, Accumulated Leap Seconds=24
  { 2447313.50, 56.070001 }, //01 June 1988, UT1-UTC= 0.1139985, Accumulated Leap Seconds=24
  { 2447314.50, 56.070917 }, //02 June 1988, UT1-UTC= 0.1130834, Accumulated Leap Seconds=24
  { 2447315.50, 56.071798 }, //03 June 1988, UT1-UTC= 0.1122017, Accumulated Leap Seconds=24
  { 2447316.50, 56.072777 }, //04 June 1988, UT1-UTC= 0.1112229, Accumulated Leap Seconds=24
  { 2447317.50, 56.073957 }, //05 June 1988, UT1-UTC= 0.1100429, Accumulated Leap Seconds=24
  { 2447318.50, 56.075378 }, //06 June 1988, UT1-UTC= 0.1086222, Accumulated Leap Seconds=24
  { 2447319.50, 56.077005 }, //07 June 1988, UT1-UTC= 0.1069952, Accumulated Leap Seconds=24
  { 2447320.50, 56.078732 }, //08 June 1988, UT1-UTC= 0.1052676, Accumulated Leap Seconds=24
  { 2447321.50, 56.080424 }, //09 June 1988, UT1-UTC= 0.1035758, Accumulated Leap Seconds=24
  { 2447322.50, 56.081954 }, //10 June 1988, UT1-UTC= 0.1020464, Accumulated Leap Seconds=24
  { 2447323.50, 56.083230 }, //11 June 1988, UT1-UTC= 0.1007698, Accumulated Leap Seconds=24
  { 2447324.50, 56.084219 }, //12 June 1988, UT1-UTC= 0.0997809, Accumulated Leap Seconds=24
  { 2447325.50, 56.084943 }, //13 June 1988, UT1-UTC= 0.0990566, Accumulated Leap Seconds=24
  { 2447326.50, 56.085469 }, //14 June 1988, UT1-UTC= 0.0985306, Accumulated Leap Seconds=24
  { 2447327.50, 56.085887 }, //15 June 1988, UT1-UTC= 0.0981126, Accumulated Leap Seconds=24
  { 2447328.50, 56.086290 }, //16 June 1988, UT1-UTC= 0.0977104, Accumulated Leap Seconds=24
  { 2447329.50, 56.086751 }, //17 June 1988, UT1-UTC= 0.0972491, Accumulated Leap Seconds=24
  { 2447330.50, 56.087316 }, //18 June 1988, UT1-UTC= 0.0966836, Accumulated Leap Seconds=24
  { 2447331.50, 56.087993 }, //19 June 1988, UT1-UTC= 0.0960067, Accumulated Leap Seconds=24
  { 2447332.50, 56.088770 }, //20 June 1988, UT1-UTC= 0.0952299, Accumulated Leap Seconds=24
  { 2447333.50, 56.089629 }, //21 June 1988, UT1-UTC= 0.0943709, Accumulated Leap Seconds=24
  { 2447334.50, 56.090538 }, //22 June 1988, UT1-UTC= 0.0934619, Accumulated Leap Seconds=24
  { 2447335.50, 56.091449 }, //23 June 1988, UT1-UTC= 0.0925512, Accumulated Leap Seconds=24
  { 2447336.50, 56.092297 }, //24 June 1988, UT1-UTC= 0.0917026, Accumulated Leap Seconds=24
  { 2447337.50, 56.093019 }, //25 June 1988, UT1-UTC= 0.0909808, Accumulated Leap Seconds=24
  { 2447338.50, 56.093559 }, //26 June 1988, UT1-UTC= 0.0904415, Accumulated Leap Seconds=24
  { 2447339.50, 56.093879 }, //27 June 1988, UT1-UTC= 0.0901208, Accumulated Leap Seconds=24
  { 2447340.50, 56.093982 }, //28 June 1988, UT1-UTC= 0.0900180, Accumulated Leap Seconds=24
  { 2447341.50, 56.093934 }, //29 June 1988, UT1-UTC= 0.0900655, Accumulated Leap Seconds=24
  { 2447342.50, 56.093860 }, //30 June 1988, UT1-UTC= 0.0901400, Accumulated Leap Seconds=24
  { 2447343.50, 56.093903 }, //01 July 1988, UT1-UTC= 0.0900965, Accumulated Leap Seconds=24
  { 2447344.50, 56.094187 }, //02 July 1988, UT1-UTC= 0.0898133, Accumulated Leap Seconds=24
  { 2447345.50, 56.094765 }, //03 July 1988, UT1-UTC= 0.0892348, Accumulated Leap Seconds=24
  { 2447346.50, 56.095612 }, //04 July 1988, UT1-UTC= 0.0883882, Accumulated Leap Seconds=24
  { 2447347.50, 56.096623 }, //05 July 1988, UT1-UTC= 0.0873765, Accumulated Leap Seconds=24
  { 2447348.50, 56.097653 }, //06 July 1988, UT1-UTC= 0.0863472, Accumulated Leap Seconds=24
  { 2447349.50, 56.098563 }, //07 July 1988, UT1-UTC= 0.0854367, Accumulated Leap Seconds=24
  { 2447350.50, 56.099255 }, //08 July 1988, UT1-UTC= 0.0847450, Accumulated Leap Seconds=24
  { 2447351.50, 56.099683 }, //09 July 1988, UT1-UTC= 0.0843173, Accumulated Leap Seconds=24
  { 2447352.50, 56.099865 }, //10 July 1988, UT1-UTC= 0.0841351, Accumulated Leap Seconds=24
  { 2447353.50, 56.099872 }, //11 July 1988, UT1-UTC= 0.0841282, Accumulated Leap Seconds=24
  { 2447354.50, 56.099794 }, //12 July 1988, UT1-UTC= 0.0842057, Accumulated Leap Seconds=24
  { 2447355.50, 56.099730 }, //13 July 1988, UT1-UTC= 0.0842702, Accumulated Leap Seconds=24
  { 2447356.50, 56.099768 }, //14 July 1988, UT1-UTC= 0.0842317, Accumulated Leap Seconds=24
  { 2447357.50, 56.099968 }, //15 July 1988, UT1-UTC= 0.0840321, Accumulated Leap Seconds=24
  { 2447358.50, 56.100349 }, //16 July 1988, UT1-UTC= 0.0836510, Accumulated Leap Seconds=24
  { 2447359.50, 56.100903 }, //17 July 1988, UT1-UTC= 0.0830968, Accumulated Leap Seconds=24
  { 2447360.50, 56.101599 }, //18 July 1988, UT1-UTC= 0.0824008, Accumulated Leap Seconds=24
  { 2447361.50, 56.102396 }, //19 July 1988, UT1-UTC= 0.0816045, Accumulated Leap Seconds=24
  { 2447362.50, 56.103238 }, //20 July 1988, UT1-UTC= 0.0807624, Accumulated Leap Seconds=24
  { 2447363.50, 56.104065 }, //21 July 1988, UT1-UTC= 0.0799349, Accumulated Leap Seconds=24
  { 2447364.50, 56.104824 }, //22 July 1988, UT1-UTC= 0.0791762, Accumulated Leap Seconds=24
  { 2447365.50, 56.105472 }, //23 July 1988, UT1-UTC= 0.0785285, Accumulated Leap Seconds=24
  { 2447366.50, 56.105970 }, //24 July 1988, UT1-UTC= 0.0780301, Accumulated Leap Seconds=24
  { 2447367.50, 56.106306 }, //25 July 1988, UT1-UTC= 0.0776942, Accumulated Leap Seconds=24
  { 2447368.50, 56.106507 }, //26 July 1988, UT1-UTC= 0.0774927, Accumulated Leap Seconds=24
  { 2447369.50, 56.106661 }, //27 July 1988, UT1-UTC= 0.0773393, Accumulated Leap Seconds=24
  { 2447370.50, 56.106900 }, //28 July 1988, UT1-UTC= 0.0770997, Accumulated Leap Seconds=24
  { 2447371.50, 56.107364 }, //29 July 1988, UT1-UTC= 0.0766355, Accumulated Leap Seconds=24
  { 2447372.50, 56.108138 }, //30 July 1988, UT1-UTC= 0.0758623, Accumulated Leap Seconds=24
  { 2447373.50, 56.109208 }, //31 July 1988, UT1-UTC= 0.0747922, Accumulated Leap Seconds=24
  { 2447374.50, 56.110463 }, //01 August 1988, UT1-UTC= 0.0735370, Accumulated Leap Seconds=24
  { 2447375.50, 56.111749 }, //02 August 1988, UT1-UTC= 0.0722505, Accumulated Leap Seconds=24
  { 2447376.50, 56.112912 }, //03 August 1988, UT1-UTC= 0.0710878, Accumulated Leap Seconds=24
  { 2447377.50, 56.113837 }, //04 August 1988, UT1-UTC= 0.0701627, Accumulated Leap Seconds=24
  { 2447378.50, 56.114482 }, //05 August 1988, UT1-UTC= 0.0695180, Accumulated Leap Seconds=24
  { 2447379.50, 56.114869 }, //06 August 1988, UT1-UTC= 0.0691308, Accumulated Leap Seconds=24
  { 2447380.50, 56.115060 }, //07 August 1988, UT1-UTC= 0.0689399, Accumulated Leap Seconds=24
  { 2447381.50, 56.115115 }, //08 August 1988, UT1-UTC= 0.0688849, Accumulated Leap Seconds=24
  { 2447382.50, 56.115128 }, //09 August 1988, UT1-UTC= 0.0688718, Accumulated Leap Seconds=24
  { 2447383.50, 56.115195 }, //10 August 1988, UT1-UTC= 0.0688053, Accumulated Leap Seconds=24
  { 2447384.50, 56.115390 }, //11 August 1988, UT1-UTC= 0.0686103, Accumulated Leap Seconds=24
  { 2447385.50, 56.115762 }, //12 August 1988, UT1-UTC= 0.0682377, Accumulated Leap Seconds=24
  { 2447386.50, 56.116327 }, //13 August 1988, UT1-UTC= 0.0676731, Accumulated Leap Seconds=24
  { 2447387.50, 56.117065 }, //14 August 1988, UT1-UTC= 0.0669348, Accumulated Leap Seconds=24
  { 2447388.50, 56.117934 }, //15 August 1988, UT1-UTC= 0.0660663, Accumulated Leap Seconds=24
  { 2447389.50, 56.118873 }, //16 August 1988, UT1-UTC= 0.0651271, Accumulated Leap Seconds=24
  { 2447390.50, 56.119818 }, //17 August 1988, UT1-UTC= 0.0641818, Accumulated Leap Seconds=24
  { 2447391.50, 56.120702 }, //18 August 1988, UT1-UTC= 0.0632978, Accumulated Leap Seconds=24
  { 2447392.50, 56.121464 }, //19 August 1988, UT1-UTC= 0.0625359, Accumulated Leap Seconds=24
  { 2447393.50, 56.122062 }, //20 August 1988, UT1-UTC= 0.0619383, Accumulated Leap Seconds=24
  { 2447394.50, 56.122480 }, //21 August 1988, UT1-UTC= 0.0615202, Accumulated Leap Seconds=24
  { 2447395.50, 56.122737 }, //22 August 1988, UT1-UTC= 0.0612633, Accumulated Leap Seconds=24
  { 2447396.50, 56.122900 }, //23 August 1988, UT1-UTC= 0.0611001, Accumulated Leap Seconds=24
  { 2447397.50, 56.123086 }, //24 August 1988, UT1-UTC= 0.0609137, Accumulated Leap Seconds=24
  { 2447398.50, 56.123437 }, //25 August 1988, UT1-UTC= 0.0605631, Accumulated Leap Seconds=24
  { 2447399.50, 56.124074 }, //26 August 1988, UT1-UTC= 0.0599264, Accumulated Leap Seconds=24
  { 2447400.50, 56.125051 }, //27 August 1988, UT1-UTC= 0.0589488, Accumulated Leap Seconds=24
  { 2447401.50, 56.126329 }, //28 August 1988, UT1-UTC= 0.0576714, Accumulated Leap Seconds=24
  { 2447402.50, 56.127770 }, //29 August 1988, UT1-UTC= 0.0562298, Accumulated Leap Seconds=24
  { 2447403.50, 56.129191 }, //30 August 1988, UT1-UTC= 0.0548094, Accumulated Leap Seconds=24
  { 2447404.50, 56.130422 }, //31 August 1988, UT1-UTC= 0.0535782, Accumulated Leap Seconds=24
  { 2447405.50, 56.131374 }, //01 September 1988, UT1-UTC= 0.0526264, Accumulated Leap Seconds=24
  { 2447406.50, 56.132039 }, //02 September 1988, UT1-UTC= 0.0519614, Accumulated Leap Seconds=24
  { 2447407.50, 56.132475 }, //03 September 1988, UT1-UTC= 0.0515249, Accumulated Leap Seconds=24
  { 2447408.50, 56.132781 }, //04 September 1988, UT1-UTC= 0.0512189, Accumulated Leap Seconds=24
  { 2447409.50, 56.133056 }, //05 September 1988, UT1-UTC= 0.0509438, Accumulated Leap Seconds=24
  { 2447410.50, 56.133383 }, //06 September 1988, UT1-UTC= 0.0506166, Accumulated Leap Seconds=24
  { 2447411.50, 56.133826 }, //07 September 1988, UT1-UTC= 0.0501735, Accumulated Leap Seconds=24
  { 2447412.50, 56.134424 }, //08 September 1988, UT1-UTC= 0.0495759, Accumulated Leap Seconds=24
  { 2447413.50, 56.135188 }, //09 September 1988, UT1-UTC= 0.0488117, Accumulated Leap Seconds=24
  { 2447414.50, 56.136107 }, //10 September 1988, UT1-UTC= 0.0478926, Accumulated Leap Seconds=24
  { 2447415.50, 56.137148 }, //11 September 1988, UT1-UTC= 0.0468522, Accumulated Leap Seconds=24
  { 2447416.50, 56.138260 }, //12 September 1988, UT1-UTC= 0.0457403, Accumulated Leap Seconds=24
  { 2447417.50, 56.139388 }, //13 September 1988, UT1-UTC= 0.0446119, Accumulated Leap Seconds=24
  { 2447418.50, 56.140479 }, //14 September 1988, UT1-UTC= 0.0435205, Accumulated Leap Seconds=24
  { 2447419.50, 56.141487 }, //15 September 1988, UT1-UTC= 0.0425126, Accumulated Leap Seconds=24
  { 2447420.50, 56.142379 }, //16 September 1988, UT1-UTC= 0.0416208, Accumulated Leap Seconds=24
  { 2447421.50, 56.143143 }, //17 September 1988, UT1-UTC= 0.0408569, Accumulated Leap Seconds=24
  { 2447422.50, 56.143795 }, //18 September 1988, UT1-UTC= 0.0402052, Accumulated Leap Seconds=24
  { 2447423.50, 56.144385 }, //19 September 1988, UT1-UTC= 0.0396149, Accumulated Leap Seconds=24
  { 2447424.50, 56.145003 }, //20 September 1988, UT1-UTC= 0.0389966, Accumulated Leap Seconds=24
  { 2447425.50, 56.145766 }, //21 September 1988, UT1-UTC= 0.0382336, Accumulated Leap Seconds=24
  { 2447426.50, 56.146796 }, //22 September 1988, UT1-UTC= 0.0372042, Accumulated Leap Seconds=24
  { 2447427.50, 56.148170 }, //23 September 1988, UT1-UTC= 0.0358299, Accumulated Leap Seconds=24
  { 2447428.50, 56.149893 }, //24 September 1988, UT1-UTC= 0.0341065, Accumulated Leap Seconds=24
  { 2447429.50, 56.151881 }, //25 September 1988, UT1-UTC= 0.0321194, Accumulated Leap Seconds=24
  { 2447430.50, 56.153970 }, //26 September 1988, UT1-UTC= 0.0300301, Accumulated Leap Seconds=24
  { 2447431.50, 56.155964 }, //27 September 1988, UT1-UTC= 0.0280358, Accumulated Leap Seconds=24
  { 2447432.50, 56.157701 }, //28 September 1988, UT1-UTC= 0.0262990, Accumulated Leap Seconds=24
  { 2447433.50, 56.159106 }, //29 September 1988, UT1-UTC= 0.0248943, Accumulated Leap Seconds=24
  { 2447434.50, 56.160202 }, //30 September 1988, UT1-UTC= 0.0237979, Accumulated Leap Seconds=24
  { 2447435.50, 56.161084 }, //01 October 1988, UT1-UTC= 0.0229161, Accumulated Leap Seconds=24
  { 2447436.50, 56.161871 }, //02 October 1988, UT1-UTC= 0.0221288, Accumulated Leap Seconds=24
  { 2447437.50, 56.162679 }, //03 October 1988, UT1-UTC= 0.0213207, Accumulated Leap Seconds=24
  { 2447438.50, 56.163604 }, //04 October 1988, UT1-UTC= 0.0203962, Accumulated Leap Seconds=24
  { 2447439.50, 56.164709 }, //05 October 1988, UT1-UTC= 0.0192910, Accumulated Leap Seconds=24
  { 2447440.50, 56.166030 }, //06 October 1988, UT1-UTC= 0.0179704, Accumulated Leap Seconds=24
  { 2447441.50, 56.167561 }, //07 October 1988, UT1-UTC= 0.0164394, Accumulated Leap Seconds=24
  { 2447442.50, 56.169267 }, //08 October 1988, UT1-UTC= 0.0147330, Accumulated Leap Seconds=24
  { 2447443.50, 56.171096 }, //09 October 1988, UT1-UTC= 0.0129039, Accumulated Leap Seconds=24
  { 2447444.50, 56.172980 }, //10 October 1988, UT1-UTC= 0.0110195, Accumulated Leap Seconds=24
  { 2447445.50, 56.174847 }, //11 October 1988, UT1-UTC= 0.0091531, Accumulated Leap Seconds=24
  { 2447446.50, 56.176621 }, //12 October 1988, UT1-UTC= 0.0073790, Accumulated Leap Seconds=24
  { 2447447.50, 56.178242 }, //13 October 1988, UT1-UTC= 0.0057583, Accumulated Leap Seconds=24
  { 2447448.50, 56.179676 }, //14 October 1988, UT1-UTC= 0.0043242, Accumulated Leap Seconds=24
  { 2447449.50, 56.180925 }, //15 October 1988, UT1-UTC= 0.0030754, Accumulated Leap Seconds=24
  { 2447450.50, 56.182030 }, //16 October 1988, UT1-UTC= 0.0019703, Accumulated Leap Seconds=24
  { 2447451.50, 56.183075 }, //17 October 1988, UT1-UTC= 0.0009247, Accumulated Leap Seconds=24
  { 2447452.50, 56.184175 }, //18 October 1988, UT1-UTC=-0.0001748, Accumulated Leap Seconds=24
  { 2447453.50, 56.185447 }, //19 October 1988, UT1-UTC=-0.0014468, Accumulated Leap Seconds=24
  { 2447454.50, 56.186990 }, //20 October 1988, UT1-UTC=-0.0029904, Accumulated Leap Seconds=24
  { 2447455.50, 56.188857 }, //21 October 1988, UT1-UTC=-0.0048569, Accumulated Leap Seconds=24
  { 2447456.50, 56.191010 }, //22 October 1988, UT1-UTC=-0.0070096, Accumulated Leap Seconds=24
  { 2447457.50, 56.193326 }, //23 October 1988, UT1-UTC=-0.0093265, Accumulated Leap Seconds=24
  { 2447458.50, 56.195634 }, //24 October 1988, UT1-UTC=-0.0116338, Accumulated Leap Seconds=24
  { 2447459.50, 56.197755 }, //25 October 1988, UT1-UTC=-0.0137554, Accumulated Leap Seconds=24
  { 2447460.50, 56.199572 }, //26 October 1988, UT1-UTC=-0.0155717, Accumulated Leap Seconds=24
  { 2447461.50, 56.201059 }, //27 October 1988, UT1-UTC=-0.0170590, Accumulated Leap Seconds=24
  { 2447462.50, 56.202280 }, //28 October 1988, UT1-UTC=-0.0182804, Accumulated Leap Seconds=24
  { 2447463.50, 56.203353 }, //29 October 1988, UT1-UTC=-0.0193532, Accumulated Leap Seconds=24
  { 2447464.50, 56.204406 }, //30 October 1988, UT1-UTC=-0.0204058, Accumulated Leap Seconds=24
  { 2447465.50, 56.205545 }, //31 October 1988, UT1-UTC=-0.0215455, Accumulated Leap Seconds=24
  { 2447466.50, 56.206843 }, //01 November 1988, UT1-UTC=-0.0228432, Accumulated Leap Seconds=24
  { 2447467.50, 56.208331 }, //02 November 1988, UT1-UTC=-0.0243307, Accumulated Leap Seconds=24
  { 2447468.50, 56.210010 }, //03 November 1988, UT1-UTC=-0.0260098, Accumulated Leap Seconds=24
  { 2447469.50, 56.211859 }, //04 November 1988, UT1-UTC=-0.0278594, Accumulated Leap Seconds=24
  { 2447470.50, 56.213838 }, //05 November 1988, UT1-UTC=-0.0298382, Accumulated Leap Seconds=24
  { 2447471.50, 56.215880 }, //06 November 1988, UT1-UTC=-0.0318800, Accumulated Leap Seconds=24
  { 2447472.50, 56.217901 }, //07 November 1988, UT1-UTC=-0.0339009, Accumulated Leap Seconds=24
  { 2447473.50, 56.219823 }, //08 November 1988, UT1-UTC=-0.0358233, Accumulated Leap Seconds=24
  { 2447474.50, 56.221600 }, //09 November 1988, UT1-UTC=-0.0376000, Accumulated Leap Seconds=24
  { 2447475.50, 56.223209 }, //10 November 1988, UT1-UTC=-0.0392095, Accumulated Leap Seconds=24
  { 2447476.50, 56.224646 }, //11 November 1988, UT1-UTC=-0.0406457, Accumulated Leap Seconds=24
  { 2447477.50, 56.225932 }, //12 November 1988, UT1-UTC=-0.0419323, Accumulated Leap Seconds=24
  { 2447478.50, 56.227146 }, //13 November 1988, UT1-UTC=-0.0431459, Accumulated Leap Seconds=24
  { 2447479.50, 56.228395 }, //14 November 1988, UT1-UTC=-0.0443952, Accumulated Leap Seconds=24
  { 2447480.50, 56.229796 }, //15 November 1988, UT1-UTC=-0.0457959, Accumulated Leap Seconds=24
  { 2447481.50, 56.231443 }, //16 November 1988, UT1-UTC=-0.0474426, Accumulated Leap Seconds=24
  { 2447482.50, 56.233383 }, //17 November 1988, UT1-UTC=-0.0493834, Accumulated Leap Seconds=24
  { 2447483.50, 56.235600 }, //18 November 1988, UT1-UTC=-0.0516004, Accumulated Leap Seconds=24
  { 2447484.50, 56.238005 }, //19 November 1988, UT1-UTC=-0.0540051, Accumulated Leap Seconds=24
  { 2447485.50, 56.240458 }, //20 November 1988, UT1-UTC=-0.0564582, Accumulated Leap Seconds=24
  { 2447486.50, 56.242803 }, //21 November 1988, UT1-UTC=-0.0588033, Accumulated Leap Seconds=24
  { 2447487.50, 56.244912 }, //22 November 1988, UT1-UTC=-0.0609119, Accumulated Leap Seconds=24
  { 2447488.50, 56.246717 }, //23 November 1988, UT1-UTC=-0.0627171, Accumulated Leap Seconds=24
  { 2447489.50, 56.248230 }, //24 November 1988, UT1-UTC=-0.0642305, Accumulated Leap Seconds=24
  { 2447490.50, 56.249531 }, //25 November 1988, UT1-UTC=-0.0655306, Accumulated Leap Seconds=24
  { 2447491.50, 56.250736 }, //26 November 1988, UT1-UTC=-0.0667364, Accumulated Leap Seconds=24
  { 2447492.50, 56.251965 }, //27 November 1988, UT1-UTC=-0.0679649, Accumulated Leap Seconds=24
  { 2447493.50, 56.253302 }, //28 November 1988, UT1-UTC=-0.0693018, Accumulated Leap Seconds=24
  { 2447494.50, 56.254792 }, //29 November 1988, UT1-UTC=-0.0707922, Accumulated Leap Seconds=24
  { 2447495.50, 56.256446 }, //30 November 1988, UT1-UTC=-0.0724465, Accumulated Leap Seconds=24
  { 2447496.50, 56.258250 }, //01 December 1988, UT1-UTC=-0.0742503, Accumulated Leap Seconds=24
  { 2447497.50, 56.260160 }, //02 December 1988, UT1-UTC=-0.0761600, Accumulated Leap Seconds=24
  { 2447498.50, 56.262113 }, //03 December 1988, UT1-UTC=-0.0781131, Accumulated Leap Seconds=24
  { 2447499.50, 56.264037 }, //04 December 1988, UT1-UTC=-0.0800369, Accumulated Leap Seconds=24
  { 2447500.50, 56.265860 }, //05 December 1988, UT1-UTC=-0.0818602, Accumulated Leap Seconds=24
  { 2447501.50, 56.267529 }, //06 December 1988, UT1-UTC=-0.0835287, Accumulated Leap Seconds=24
  { 2447502.50, 56.269007 }, //07 December 1988, UT1-UTC=-0.0850071, Accumulated Leap Seconds=24
  { 2447503.50, 56.270288 }, //08 December 1988, UT1-UTC=-0.0862881, Accumulated Leap Seconds=24
  { 2447504.50, 56.271402 }, //09 December 1988, UT1-UTC=-0.0874021, Accumulated Leap Seconds=24
  { 2447505.50, 56.272417 }, //10 December 1988, UT1-UTC=-0.0884172, Accumulated Leap Seconds=24
  { 2447506.50, 56.273414 }, //11 December 1988, UT1-UTC=-0.0894141, Accumulated Leap Seconds=24
  { 2447507.50, 56.274482 }, //12 December 1988, UT1-UTC=-0.0904824, Accumulated Leap Seconds=24
  { 2447508.50, 56.275722 }, //13 December 1988, UT1-UTC=-0.0917216, Accumulated Leap Seconds=24
  { 2447509.50, 56.277191 }, //14 December 1988, UT1-UTC=-0.0931914, Accumulated Leap Seconds=24
  { 2447510.50, 56.278889 }, //15 December 1988, UT1-UTC=-0.0948885, Accumulated Leap Seconds=24
  { 2447511.50, 56.280744 }, //16 December 1988, UT1-UTC=-0.0967439, Accumulated Leap Seconds=24
  { 2447512.50, 56.282633 }, //17 December 1988, UT1-UTC=-0.0986328, Accumulated Leap Seconds=24
  { 2447513.50, 56.284416 }, //18 December 1988, UT1-UTC=-0.1004165, Accumulated Leap Seconds=24
  { 2447514.50, 56.285978 }, //19 December 1988, UT1-UTC=-0.1019779, Accumulated Leap Seconds=24
  { 2447515.50, 56.287249 }, //20 December 1988, UT1-UTC=-0.1032494, Accumulated Leap Seconds=24
  { 2447516.50, 56.288236 }, //21 December 1988, UT1-UTC=-0.1042362, Accumulated Leap Seconds=24
  { 2447517.50, 56.289004 }, //22 December 1988, UT1-UTC=-0.1050037, Accumulated Leap Seconds=24
  { 2447518.50, 56.289655 }, //23 December 1988, UT1-UTC=-0.1056551, Accumulated Leap Seconds=24
  { 2447519.50, 56.290302 }, //24 December 1988, UT1-UTC=-0.1063018, Accumulated Leap Seconds=24
  { 2447520.50, 56.291038 }, //25 December 1988, UT1-UTC=-0.1070381, Accumulated Leap Seconds=24
  { 2447521.50, 56.291926 }, //26 December 1988, UT1-UTC=-0.1079260, Accumulated Leap Seconds=24
  { 2447522.50, 56.292989 }, //27 December 1988, UT1-UTC=-0.1089889, Accumulated Leap Seconds=24
  { 2447523.50, 56.294220 }, //28 December 1988, UT1-UTC=-0.1102202, Accumulated Leap Seconds=24
  { 2447524.50, 56.295593 }, //29 December 1988, UT1-UTC=-0.1115932, Accumulated Leap Seconds=24
  { 2447525.50, 56.297064 }, //30 December 1988, UT1-UTC=-0.1130642, Accumulated Leap Seconds=24
  { 2447526.50, 56.298568 }, //31 December 1988, UT1-UTC=-0.1145681, Accumulated Leap Seconds=24
  { 2447527.50, 56.300035 }, //01 January 1989, UT1-UTC=-0.1160349, Accumulated Leap Seconds=24
  { 2447528.50, 56.301405 }, //02 January 1989, UT1-UTC=-0.1174047, Accumulated Leap Seconds=24
  { 2447529.50, 56.302635 }, //03 January 1989, UT1-UTC=-0.1186354, Accumulated Leap Seconds=24
  { 2447530.50, 56.303706 }, //04 January 1989, UT1-UTC=-0.1197061, Accumulated Leap Seconds=24
  { 2447531.50, 56.304618 }, //05 January 1989, UT1-UTC=-0.1206183, Accumulated Leap Seconds=24
  { 2447532.50, 56.305410 }, //06 January 1989, UT1-UTC=-0.1214101, Accumulated Leap Seconds=24
  { 2447533.50, 56.306161 }, //07 January 1989, UT1-UTC=-0.1221614, Accumulated Leap Seconds=24
  { 2447534.50, 56.306980 }, //08 January 1989, UT1-UTC=-0.1229796, Accumulated Leap Seconds=24
  { 2447535.50, 56.307971 }, //09 January 1989, UT1-UTC=-0.1239712, Accumulated Leap Seconds=24
  { 2447536.50, 56.309222 }, //10 January 1989, UT1-UTC=-0.1252219, Accumulated Leap Seconds=24
  { 2447537.50, 56.310751 }, //11 January 1989, UT1-UTC=-0.1267512, Accumulated Leap Seconds=24
  { 2447538.50, 56.312497 }, //12 January 1989, UT1-UTC=-0.1284973, Accumulated Leap Seconds=24
  { 2447539.50, 56.314342 }, //13 January 1989, UT1-UTC=-0.1303419, Accumulated Leap Seconds=24
  { 2447540.50, 56.316146 }, //14 January 1989, UT1-UTC=-0.1321460, Accumulated Leap Seconds=24
  { 2447541.50, 56.317774 }, //15 January 1989, UT1-UTC=-0.1337743, Accumulated Leap Seconds=24
  { 2447542.50, 56.319147 }, //16 January 1989, UT1-UTC=-0.1351474, Accumulated Leap Seconds=24
  { 2447543.50, 56.320254 }, //17 January 1989, UT1-UTC=-0.1362545, Accumulated Leap Seconds=24
  { 2447544.50, 56.321142 }, //18 January 1989, UT1-UTC=-0.1371424, Accumulated Leap Seconds=24
  { 2447545.50, 56.321898 }, //19 January 1989, UT1-UTC=-0.1378982, Accumulated Leap Seconds=24
  { 2447546.50, 56.322634 }, //20 January 1989, UT1-UTC=-0.1386342, Accumulated Leap Seconds=24
  { 2447547.50, 56.323459 }, //21 January 1989, UT1-UTC=-0.1394585, Accumulated Leap Seconds=24
  { 2447548.50, 56.324454 }, //22 January 1989, UT1-UTC=-0.1404543, Accumulated Leap Seconds=24
  { 2447549.50, 56.325670 }, //23 January 1989, UT1-UTC=-0.1416705, Accumulated Leap Seconds=24
  { 2447550.50, 56.327117 }, //24 January 1989, UT1-UTC=-0.1431174, Accumulated Leap Seconds=24
  { 2447551.50, 56.328758 }, //25 January 1989, UT1-UTC=-0.1447585, Accumulated Leap Seconds=24
  { 2447552.50, 56.330527 }, //26 January 1989, UT1-UTC=-0.1465270, Accumulated Leap Seconds=24
  { 2447553.50, 56.332344 }, //27 January 1989, UT1-UTC=-0.1483442, Accumulated Leap Seconds=24
  { 2447554.50, 56.334131 }, //28 January 1989, UT1-UTC=-0.1501307, Accumulated Leap Seconds=24
  { 2447555.50, 56.335816 }, //29 January 1989, UT1-UTC=-0.1518158, Accumulated Leap Seconds=24
  { 2447556.50, 56.337352 }, //30 January 1989, UT1-UTC=-0.1533522, Accumulated Leap Seconds=24
  { 2447557.50, 56.338714 }, //31 January 1989, UT1-UTC=-0.1547139, Accumulated Leap Seconds=24
  { 2447558.50, 56.339902 }, //01 February 1989, UT1-UTC=-0.1559020, Accumulated Leap Seconds=24
  { 2447559.50, 56.340951 }, //02 February 1989, UT1-UTC=-0.1569507, Accumulated Leap Seconds=24
  { 2447560.50, 56.341932 }, //03 February 1989, UT1-UTC=-0.1579319, Accumulated Leap Seconds=24
  { 2447561.50, 56.342955 }, //04 February 1989, UT1-UTC=-0.1589547, Accumulated Leap Seconds=24
  { 2447562.50, 56.344150 }, //05 February 1989, UT1-UTC=-0.1601504, Accumulated Leap Seconds=24
  { 2447563.50, 56.345629 }, //06 February 1989, UT1-UTC=-0.1616292, Accumulated Leap Seconds=24
  { 2447564.50, 56.347440 }, //07 February 1989, UT1-UTC=-0.1634401, Accumulated Leap Seconds=24
  { 2447565.50, 56.349544 }, //08 February 1989, UT1-UTC=-0.1655440, Accumulated Leap Seconds=24
  { 2447566.50, 56.351814 }, //09 February 1989, UT1-UTC=-0.1678139, Accumulated Leap Seconds=24
  { 2447567.50, 56.354063 }, //10 February 1989, UT1-UTC=-0.1700631, Accumulated Leap Seconds=24
  { 2447568.50, 56.356116 }, //11 February 1989, UT1-UTC=-0.1721158, Accumulated Leap Seconds=24
  { 2447569.50, 56.357857 }, //12 February 1989, UT1-UTC=-0.1738568, Accumulated Leap Seconds=24
  { 2447570.50, 56.359252 }, //13 February 1989, UT1-UTC=-0.1752522, Accumulated Leap Seconds=24
  { 2447571.50, 56.360352 }, //14 February 1989, UT1-UTC=-0.1763520, Accumulated Leap Seconds=24
  { 2447572.50, 56.361254 }, //15 February 1989, UT1-UTC=-0.1772543, Accumulated Leap Seconds=24
  { 2447573.50, 56.362070 }, //16 February 1989, UT1-UTC=-0.1780702, Accumulated Leap Seconds=24
  { 2447574.50, 56.362900 }, //17 February 1989, UT1-UTC=-0.1788996, Accumulated Leap Seconds=24
  { 2447575.50, 56.363822 }, //18 February 1989, UT1-UTC=-0.1798216, Accumulated Leap Seconds=24
  { 2447576.50, 56.364901 }, //19 February 1989, UT1-UTC=-0.1809010, Accumulated Leap Seconds=24
  { 2447577.50, 56.366164 }, //20 February 1989, UT1-UTC=-0.1821639, Accumulated Leap Seconds=24
  { 2447578.50, 56.367603 }, //21 February 1989, UT1-UTC=-0.1836029, Accumulated Leap Seconds=24
  { 2447579.50, 56.369178 }, //22 February 1989, UT1-UTC=-0.1851780, Accumulated Leap Seconds=24
  { 2447580.50, 56.370818 }, //23 February 1989, UT1-UTC=-0.1868181, Accumulated Leap Seconds=24
  { 2447581.50, 56.372454 }, //24 February 1989, UT1-UTC=-0.1884540, Accumulated Leap Seconds=24
  { 2447582.50, 56.374024 }, //25 February 1989, UT1-UTC=-0.1900241, Accumulated Leap Seconds=24
  { 2447583.50, 56.375481 }, //26 February 1989, UT1-UTC=-0.1914809, Accumulated Leap Seconds=24
  { 2447584.50, 56.376796 }, //27 February 1989, UT1-UTC=-0.1927958, Accumulated Leap Seconds=24
  { 2447585.50, 56.377962 }, //28 February 1989, UT1-UTC=-0.1939620, Accumulated Leap Seconds=24
  { 2447586.50, 56.378999 }, //01 March 1989, UT1-UTC=-0.1949995, Accumulated Leap Seconds=24
  { 2447587.50, 56.379963 }, //02 March 1989, UT1-UTC=-0.1959628, Accumulated Leap Seconds=24
  { 2447588.50, 56.380944 }, //03 March 1989, UT1-UTC=-0.1969440, Accumulated Leap Seconds=24
  { 2447589.50, 56.382063 }, //04 March 1989, UT1-UTC=-0.1980628, Accumulated Leap Seconds=24
  { 2447590.50, 56.383442 }, //05 March 1989, UT1-UTC=-0.1994421, Accumulated Leap Seconds=24
  { 2447591.50, 56.385175 }, //06 March 1989, UT1-UTC=-0.2011754, Accumulated Leap Seconds=24
  { 2447592.50, 56.387276 }, //07 March 1989, UT1-UTC=-0.2032762, Accumulated Leap Seconds=24
  { 2447593.50, 56.389662 }, //08 March 1989, UT1-UTC=-0.2056620, Accumulated Leap Seconds=24
  { 2447594.50, 56.392177 }, //09 March 1989, UT1-UTC=-0.2081766, Accumulated Leap Seconds=24
  { 2447595.50, 56.394637 }, //10 March 1989, UT1-UTC=-0.2106368, Accumulated Leap Seconds=24
  { 2447596.50, 56.396872 }, //11 March 1989, UT1-UTC=-0.2128720, Accumulated Leap Seconds=24
  { 2447597.50, 56.398786 }, //12 March 1989, UT1-UTC=-0.2147857, Accumulated Leap Seconds=24
  { 2447598.50, 56.400376 }, //13 March 1989, UT1-UTC=-0.2163757, Accumulated Leap Seconds=24
  { 2447599.50, 56.401708 }, //14 March 1989, UT1-UTC=-0.2177083, Accumulated Leap Seconds=24
  { 2447600.50, 56.402885 }, //15 March 1989, UT1-UTC=-0.2188855, Accumulated Leap Seconds=24
  { 2447601.50, 56.404036 }, //16 March 1989, UT1-UTC=-0.2200363, Accumulated Leap Seconds=24
  { 2447602.50, 56.405267 }, //17 March 1989, UT1-UTC=-0.2212672, Accumulated Leap Seconds=24
  { 2447603.50, 56.406647 }, //18 March 1989, UT1-UTC=-0.2226472, Accumulated Leap Seconds=24
  { 2447604.50, 56.408212 }, //19 March 1989, UT1-UTC=-0.2242124, Accumulated Leap Seconds=24
  { 2447605.50, 56.409969 }, //20 March 1989, UT1-UTC=-0.2259687, Accumulated Leap Seconds=24
  { 2447606.50, 56.411875 }, //21 March 1989, UT1-UTC=-0.2278751, Accumulated Leap Seconds=24
  { 2447607.50, 56.413863 }, //22 March 1989, UT1-UTC=-0.2298630, Accumulated Leap Seconds=24
  { 2447608.50, 56.415855 }, //23 March 1989, UT1-UTC=-0.2318548, Accumulated Leap Seconds=24
  { 2447609.50, 56.417773 }, //24 March 1989, UT1-UTC=-0.2337735, Accumulated Leap Seconds=24
  { 2447610.50, 56.419557 }, //25 March 1989, UT1-UTC=-0.2355568, Accumulated Leap Seconds=24
  { 2447611.50, 56.421169 }, //26 March 1989, UT1-UTC=-0.2371691, Accumulated Leap Seconds=24
  { 2447612.50, 56.422600 }, //27 March 1989, UT1-UTC=-0.2385998, Accumulated Leap Seconds=24
  { 2447613.50, 56.423860 }, //28 March 1989, UT1-UTC=-0.2398599, Accumulated Leap Seconds=24
  { 2447614.50, 56.424981 }, //29 March 1989, UT1-UTC=-0.2409806, Accumulated Leap Seconds=24
  { 2447615.50, 56.426021 }, //30 March 1989, UT1-UTC=-0.2420207, Accumulated Leap Seconds=24
  { 2447616.50, 56.427080 }, //31 March 1989, UT1-UTC=-0.2430805, Accumulated Leap Seconds=24
  { 2447617.50, 56.428284 }, //01 April 1989, UT1-UTC=-0.2442839, Accumulated Leap Seconds=24
  { 2447618.50, 56.429740 }, //02 April 1989, UT1-UTC=-0.2457404, Accumulated Leap Seconds=24
  { 2447619.50, 56.431511 }, //03 April 1989, UT1-UTC=-0.2475107, Accumulated Leap Seconds=24
  { 2447620.50, 56.433570 }, //04 April 1989, UT1-UTC=-0.2495702, Accumulated Leap Seconds=24
  { 2447621.50, 56.435799 }, //05 April 1989, UT1-UTC=-0.2517986, Accumulated Leap Seconds=24
  { 2447622.50, 56.438016 }, //06 April 1989, UT1-UTC=-0.2540165, Accumulated Leap Seconds=24
  { 2447623.50, 56.440048 }, //07 April 1989, UT1-UTC=-0.2560478, Accumulated Leap Seconds=24
  { 2447624.50, 56.441780 }, //08 April 1989, UT1-UTC=-0.2577798, Accumulated Leap Seconds=24
  { 2447625.50, 56.443196 }, //09 April 1989, UT1-UTC=-0.2591964, Accumulated Leap Seconds=24
  { 2447626.50, 56.444369 }, //10 April 1989, UT1-UTC=-0.2603694, Accumulated Leap Seconds=24
  { 2447627.50, 56.445423 }, //11 April 1989, UT1-UTC=-0.2614231, Accumulated Leap Seconds=24
  { 2447628.50, 56.446488 }, //12 April 1989, UT1-UTC=-0.2624876, Accumulated Leap Seconds=24
  { 2447629.50, 56.447664 }, //13 April 1989, UT1-UTC=-0.2636642, Accumulated Leap Seconds=24
  { 2447630.50, 56.449013 }, //14 April 1989, UT1-UTC=-0.2650128, Accumulated Leap Seconds=24
  { 2447631.50, 56.450555 }, //15 April 1989, UT1-UTC=-0.2665550, Accumulated Leap Seconds=24
  { 2447632.50, 56.452283 }, //16 April 1989, UT1-UTC=-0.2682829, Accumulated Leap Seconds=24
  { 2447633.50, 56.454166 }, //17 April 1989, UT1-UTC=-0.2701665, Accumulated Leap Seconds=24
  { 2447634.50, 56.456162 }, //18 April 1989, UT1-UTC=-0.2721616, Accumulated Leap Seconds=24
  { 2447635.50, 56.458215 }, //19 April 1989, UT1-UTC=-0.2742155, Accumulated Leap Seconds=24
  { 2447636.50, 56.460272 }, //20 April 1989, UT1-UTC=-0.2762725, Accumulated Leap Seconds=24
  { 2447637.50, 56.462285 }, //21 April 1989, UT1-UTC=-0.2782853, Accumulated Leap Seconds=24
  { 2447638.50, 56.464221 }, //22 April 1989, UT1-UTC=-0.2802210, Accumulated Leap Seconds=24
  { 2447639.50, 56.466055 }, //23 April 1989, UT1-UTC=-0.2820548, Accumulated Leap Seconds=24
  { 2447640.50, 56.467769 }, //24 April 1989, UT1-UTC=-0.2837691, Accumulated Leap Seconds=24
  { 2447641.50, 56.469374 }, //25 April 1989, UT1-UTC=-0.2853743, Accumulated Leap Seconds=24
  { 2447642.50, 56.470913 }, //26 April 1989, UT1-UTC=-0.2869129, Accumulated Leap Seconds=24
  { 2447643.50, 56.472464 }, //27 April 1989, UT1-UTC=-0.2884636, Accumulated Leap Seconds=24
  { 2447644.50, 56.474118 }, //28 April 1989, UT1-UTC=-0.2901180, Accumulated Leap Seconds=24
  { 2447645.50, 56.475960 }, //29 April 1989, UT1-UTC=-0.2919599, Accumulated Leap Seconds=24
  { 2447646.50, 56.478051 }, //30 April 1989, UT1-UTC=-0.2940507, Accumulated Leap Seconds=24
  { 2447647.50, 56.480395 }, //01 May 1989, UT1-UTC=-0.2963947, Accumulated Leap Seconds=24
  { 2447648.50, 56.482919 }, //02 May 1989, UT1-UTC=-0.2989186, Accumulated Leap Seconds=24
  { 2447649.50, 56.485482 }, //03 May 1989, UT1-UTC=-0.3014823, Accumulated Leap Seconds=24
  { 2447650.50, 56.487918 }, //04 May 1989, UT1-UTC=-0.3039179, Accumulated Leap Seconds=24
  { 2447651.50, 56.490081 }, //05 May 1989, UT1-UTC=-0.3060813, Accumulated Leap Seconds=24
  { 2447652.50, 56.491903 }, //06 May 1989, UT1-UTC=-0.3079029, Accumulated Leap Seconds=24
  { 2447653.50, 56.493408 }, //07 May 1989, UT1-UTC=-0.3094084, Accumulated Leap Seconds=24
  { 2447654.50, 56.494700 }, //08 May 1989, UT1-UTC=-0.3106996, Accumulated Leap Seconds=24
  { 2447655.50, 56.495912 }, //09 May 1989, UT1-UTC=-0.3119116, Accumulated Leap Seconds=24
  { 2447656.50, 56.497167 }, //10 May 1989, UT1-UTC=-0.3131666, Accumulated Leap Seconds=24
  { 2447657.50, 56.498544 }, //11 May 1989, UT1-UTC=-0.3145445, Accumulated Leap Seconds=24
  { 2447658.50, 56.500078 }, //12 May 1989, UT1-UTC=-0.3160782, Accumulated Leap Seconds=24
  { 2447659.50, 56.501762 }, //13 May 1989, UT1-UTC=-0.3177621, Accumulated Leap Seconds=24
  { 2447660.50, 56.503565 }, //14 May 1989, UT1-UTC=-0.3195647, Accumulated Leap Seconds=24
  { 2447661.50, 56.505436 }, //15 May 1989, UT1-UTC=-0.3214363, Accumulated Leap Seconds=24
  { 2447662.50, 56.507320 }, //16 May 1989, UT1-UTC=-0.3233196, Accumulated Leap Seconds=24
  { 2447663.50, 56.509157 }, //17 May 1989, UT1-UTC=-0.3251571, Accumulated Leap Seconds=24
  { 2447664.50, 56.510901 }, //18 May 1989, UT1-UTC=-0.3269014, Accumulated Leap Seconds=24
  { 2447665.50, 56.512519 }, //19 May 1989, UT1-UTC=-0.3285192, Accumulated Leap Seconds=24
  { 2447666.50, 56.513994 }, //20 May 1989, UT1-UTC=-0.3299942, Accumulated Leap Seconds=24
  { 2447667.50, 56.515337 }, //21 May 1989, UT1-UTC=-0.3313368, Accumulated Leap Seconds=24
  { 2447668.50, 56.516583 }, //22 May 1989, UT1-UTC=-0.3325832, Accumulated Leap Seconds=24
  { 2447669.50, 56.517795 }, //23 May 1989, UT1-UTC=-0.3337945, Accumulated Leap Seconds=24
  { 2447670.50, 56.519051 }, //24 May 1989, UT1-UTC=-0.3350515, Accumulated Leap Seconds=24
  { 2447671.50, 56.520438 }, //25 May 1989, UT1-UTC=-0.3364376, Accumulated Leap Seconds=24
  { 2447672.50, 56.522028 }, //26 May 1989, UT1-UTC=-0.3380276, Accumulated Leap Seconds=24
  { 2447673.50, 56.523876 }, //27 May 1989, UT1-UTC=-0.3398757, Accumulated Leap Seconds=24
  { 2447674.50, 56.525988 }, //28 May 1989, UT1-UTC=-0.3419885, Accumulated Leap Seconds=24
  { 2447675.50, 56.528311 }, //29 May 1989, UT1-UTC=-0.3443113, Accumulated Leap Seconds=24
  { 2447676.50, 56.530727 }, //30 May 1989, UT1-UTC=-0.3467271, Accumulated Leap Seconds=24
  { 2447677.50, 56.533080 }, //31 May 1989, UT1-UTC=-0.3490799, Accumulated Leap Seconds=24
  { 2447678.50, 56.535216 }, //01 June 1989, UT1-UTC=-0.3512164, Accumulated Leap Seconds=24
  { 2447679.50, 56.537029 }, //02 June 1989, UT1-UTC=-0.3530292, Accumulated Leap Seconds=24
  { 2447680.50, 56.538488 }, //03 June 1989, UT1-UTC=-0.3544885, Accumulated Leap Seconds=24
  { 2447681.50, 56.539648 }, //04 June 1989, UT1-UTC=-0.3556479, Accumulated Leap Seconds=24
  { 2447682.50, 56.540620 }, //05 June 1989, UT1-UTC=-0.3566198, Accumulated Leap Seconds=24
  { 2447683.50, 56.541541 }, //06 June 1989, UT1-UTC=-0.3575406, Accumulated Leap Seconds=24
  { 2447684.50, 56.542532 }, //07 June 1989, UT1-UTC=-0.3585321, Accumulated Leap Seconds=24
  { 2447685.50, 56.543668 }, //08 June 1989, UT1-UTC=-0.3596685, Accumulated Leap Seconds=24
  { 2447686.50, 56.544966 }, //09 June 1989, UT1-UTC=-0.3609657, Accumulated Leap Seconds=24
  { 2447687.50, 56.546401 }, //10 June 1989, UT1-UTC=-0.3624006, Accumulated Leap Seconds=24
  { 2447688.50, 56.547925 }, //11 June 1989, UT1-UTC=-0.3639254, Accumulated Leap Seconds=24
  { 2447689.50, 56.549480 }, //12 June 1989, UT1-UTC=-0.3654801, Accumulated Leap Seconds=24
  { 2447690.50, 56.550991 }, //13 June 1989, UT1-UTC=-0.3669911, Accumulated Leap Seconds=24
  { 2447691.50, 56.552395 }, //14 June 1989, UT1-UTC=-0.3683952, Accumulated Leap Seconds=24
  { 2447692.50, 56.553643 }, //15 June 1989, UT1-UTC=-0.3696435, Accumulated Leap Seconds=24
  { 2447693.50, 56.554705 }, //16 June 1989, UT1-UTC=-0.3707053, Accumulated Leap Seconds=24
  { 2447694.50, 56.555574 }, //17 June 1989, UT1-UTC=-0.3715743, Accumulated Leap Seconds=24
  { 2447695.50, 56.556274 }, //18 June 1989, UT1-UTC=-0.3722738, Accumulated Leap Seconds=24
  { 2447696.50, 56.556866 }, //19 June 1989, UT1-UTC=-0.3728655, Accumulated Leap Seconds=24
  { 2447697.50, 56.557441 }, //20 June 1989, UT1-UTC=-0.3734413, Accumulated Leap Seconds=24
  { 2447698.50, 56.558107 }, //21 June 1989, UT1-UTC=-0.3741075, Accumulated Leap Seconds=24
  { 2447699.50, 56.558963 }, //22 June 1989, UT1-UTC=-0.3749631, Accumulated Leap Seconds=24
  { 2447700.50, 56.560071 }, //23 June 1989, UT1-UTC=-0.3760711, Accumulated Leap Seconds=24
  { 2447701.50, 56.561429 }, //24 June 1989, UT1-UTC=-0.3774292, Accumulated Leap Seconds=24
  { 2447702.50, 56.562974 }, //25 June 1989, UT1-UTC=-0.3789744, Accumulated Leap Seconds=24
  { 2447703.50, 56.564594 }, //26 June 1989, UT1-UTC=-0.3805941, Accumulated Leap Seconds=24
  { 2447704.50, 56.566146 }, //27 June 1989, UT1-UTC=-0.3821463, Accumulated Leap Seconds=24
  { 2447705.50, 56.567493 }, //28 June 1989, UT1-UTC=-0.3834928, Accumulated Leap Seconds=24
  { 2447706.50, 56.568546 }, //29 June 1989, UT1-UTC=-0.3845465, Accumulated Leap Seconds=24
  { 2447707.50, 56.569283 }, //30 June 1989, UT1-UTC=-0.3852834, Accumulated Leap Seconds=24
  { 2447708.50, 56.569749 }, //01 July 1989, UT1-UTC=-0.3857487, Accumulated Leap Seconds=24
  { 2447709.50, 56.570045 }, //02 July 1989, UT1-UTC=-0.3860450, Accumulated Leap Seconds=24
  { 2447710.50, 56.570301 }, //03 July 1989, UT1-UTC=-0.3863009, Accumulated Leap Seconds=24
  { 2447711.50, 56.570642 }, //04 July 1989, UT1-UTC=-0.3866416, Accumulated Leap Seconds=24
  { 2447712.50, 56.571155 }, //05 July 1989, UT1-UTC=-0.3871546, Accumulated Leap Seconds=24
  { 2447713.50, 56.571875 }, //06 July 1989, UT1-UTC=-0.3878747, Accumulated Leap Seconds=24
  { 2447714.50, 56.572791 }, //07 July 1989, UT1-UTC=-0.3887914, Accumulated Leap Seconds=24
  { 2447715.50, 56.573860 }, //08 July 1989, UT1-UTC=-0.3898604, Accumulated Leap Seconds=24
  { 2447716.50, 56.575015 }, //09 July 1989, UT1-UTC=-0.3910151, Accumulated Leap Seconds=24
  { 2447717.50, 56.576176 }, //10 July 1989, UT1-UTC=-0.3921757, Accumulated Leap Seconds=24
  { 2447718.50, 56.577272 }, //11 July 1989, UT1-UTC=-0.3932722, Accumulated Leap Seconds=24
  { 2447719.50, 56.578267 }, //12 July 1989, UT1-UTC=-0.3942669, Accumulated Leap Seconds=24
  { 2447720.50, 56.579142 }, //13 July 1989, UT1-UTC=-0.3951416, Accumulated Leap Seconds=24
  { 2447721.50, 56.579901 }, //14 July 1989, UT1-UTC=-0.3959008, Accumulated Leap Seconds=24
  { 2447722.50, 56.580576 }, //15 July 1989, UT1-UTC=-0.3965763, Accumulated Leap Seconds=24
  { 2447723.50, 56.581213 }, //16 July 1989, UT1-UTC=-0.3972135, Accumulated Leap Seconds=24
  { 2447724.50, 56.581863 }, //17 July 1989, UT1-UTC=-0.3978628, Accumulated Leap Seconds=24
  { 2447725.50, 56.582590 }, //18 July 1989, UT1-UTC=-0.3985897, Accumulated Leap Seconds=24
  { 2447726.50, 56.583486 }, //19 July 1989, UT1-UTC=-0.3994865, Accumulated Leap Seconds=24
  { 2447727.50, 56.584631 }, //20 July 1989, UT1-UTC=-0.4006310, Accumulated Leap Seconds=24
  { 2447728.50, 56.586054 }, //21 July 1989, UT1-UTC=-0.4020545, Accumulated Leap Seconds=24
  { 2447729.50, 56.587720 }, //22 July 1989, UT1-UTC=-0.4037197, Accumulated Leap Seconds=24
  { 2447730.50, 56.589521 }, //23 July 1989, UT1-UTC=-0.4055208, Accumulated Leap Seconds=24
  { 2447731.50, 56.591299 }, //24 July 1989, UT1-UTC=-0.4072988, Accumulated Leap Seconds=24
  { 2447732.50, 56.592902 }, //25 July 1989, UT1-UTC=-0.4089021, Accumulated Leap Seconds=24
  { 2447733.50, 56.594240 }, //26 July 1989, UT1-UTC=-0.4102402, Accumulated Leap Seconds=24
  { 2447734.50, 56.595281 }, //27 July 1989, UT1-UTC=-0.4112808, Accumulated Leap Seconds=24
  { 2447735.50, 56.596047 }, //28 July 1989, UT1-UTC=-0.4120472, Accumulated Leap Seconds=24
  { 2447736.50, 56.596626 }, //29 July 1989, UT1-UTC=-0.4126264, Accumulated Leap Seconds=24
  { 2447737.50, 56.597127 }, //30 July 1989, UT1-UTC=-0.4131275, Accumulated Leap Seconds=24
  { 2447738.50, 56.597658 }, //31 July 1989, UT1-UTC=-0.4136584, Accumulated Leap Seconds=24
  { 2447739.50, 56.598310 }, //01 August 1989, UT1-UTC=-0.4143102, Accumulated Leap Seconds=24
  { 2447740.50, 56.599129 }, //02 August 1989, UT1-UTC=-0.4151292, Accumulated Leap Seconds=24
  { 2447741.50, 56.600115 }, //03 August 1989, UT1-UTC=-0.4161154, Accumulated Leap Seconds=24
  { 2447742.50, 56.601233 }, //04 August 1989, UT1-UTC=-0.4172334, Accumulated Leap Seconds=24
  { 2447743.50, 56.602430 }, //05 August 1989, UT1-UTC=-0.4184299, Accumulated Leap Seconds=24
  { 2447744.50, 56.603643 }, //06 August 1989, UT1-UTC=-0.4196426, Accumulated Leap Seconds=24
  { 2447745.50, 56.604817 }, //07 August 1989, UT1-UTC=-0.4208167, Accumulated Leap Seconds=24
  { 2447746.50, 56.605902 }, //08 August 1989, UT1-UTC=-0.4219023, Accumulated Leap Seconds=24
  { 2447747.50, 56.606863 }, //09 August 1989, UT1-UTC=-0.4228632, Accumulated Leap Seconds=24
  { 2447748.50, 56.607693 }, //10 August 1989, UT1-UTC=-0.4236929, Accumulated Leap Seconds=24
  { 2447749.50, 56.608402 }, //11 August 1989, UT1-UTC=-0.4244021, Accumulated Leap Seconds=24
  { 2447750.50, 56.609014 }, //12 August 1989, UT1-UTC=-0.4250144, Accumulated Leap Seconds=24
  { 2447751.50, 56.609595 }, //13 August 1989, UT1-UTC=-0.4255955, Accumulated Leap Seconds=24
  { 2447752.50, 56.610242 }, //14 August 1989, UT1-UTC=-0.4262419, Accumulated Leap Seconds=24
  { 2447753.50, 56.611064 }, //15 August 1989, UT1-UTC=-0.4270639, Accumulated Leap Seconds=24
  { 2447754.50, 56.612160 }, //16 August 1989, UT1-UTC=-0.4281600, Accumulated Leap Seconds=24
  { 2447755.50, 56.613584 }, //17 August 1989, UT1-UTC=-0.4295845, Accumulated Leap Seconds=24
  { 2447756.50, 56.615326 }, //18 August 1989, UT1-UTC=-0.4313258, Accumulated Leap Seconds=24
  { 2447757.50, 56.617293 }, //19 August 1989, UT1-UTC=-0.4332933, Accumulated Leap Seconds=24
  { 2447758.50, 56.619322 }, //20 August 1989, UT1-UTC=-0.4353218, Accumulated Leap Seconds=24
  { 2447759.50, 56.621229 }, //21 August 1989, UT1-UTC=-0.4372287, Accumulated Leap Seconds=24
  { 2447760.50, 56.622871 }, //22 August 1989, UT1-UTC=-0.4388712, Accumulated Leap Seconds=24
  { 2447761.50, 56.624185 }, //23 August 1989, UT1-UTC=-0.4401855, Accumulated Leap Seconds=24
  { 2447762.50, 56.625193 }, //24 August 1989, UT1-UTC=-0.4411928, Accumulated Leap Seconds=24
  { 2447763.50, 56.625976 }, //25 August 1989, UT1-UTC=-0.4419764, Accumulated Leap Seconds=24
  { 2447764.50, 56.626647 }, //26 August 1989, UT1-UTC=-0.4426466, Accumulated Leap Seconds=24
  { 2447765.50, 56.627312 }, //27 August 1989, UT1-UTC=-0.4433117, Accumulated Leap Seconds=24
  { 2447766.50, 56.628072 }, //28 August 1989, UT1-UTC=-0.4440718, Accumulated Leap Seconds=24
  { 2447767.50, 56.628997 }, //29 August 1989, UT1-UTC=-0.4449967, Accumulated Leap Seconds=24
  { 2447768.50, 56.630114 }, //30 August 1989, UT1-UTC=-0.4461135, Accumulated Leap Seconds=24
  { 2447769.50, 56.631409 }, //31 August 1989, UT1-UTC=-0.4474090, Accumulated Leap Seconds=24
  { 2447770.50, 56.632833 }, //01 September 1989, UT1-UTC=-0.4488326, Accumulated Leap Seconds=24
  { 2447771.50, 56.634310 }, //02 September 1989, UT1-UTC=-0.4503104, Accumulated Leap Seconds=24
  { 2447772.50, 56.635759 }, //03 September 1989, UT1-UTC=-0.4517592, Accumulated Leap Seconds=24
  { 2447773.50, 56.637103 }, //04 September 1989, UT1-UTC=-0.4531034, Accumulated Leap Seconds=24
  { 2447774.50, 56.638286 }, //05 September 1989, UT1-UTC=-0.4542861, Accumulated Leap Seconds=24
  { 2447775.50, 56.639279 }, //06 September 1989, UT1-UTC=-0.4552793, Accumulated Leap Seconds=24
  { 2447776.50, 56.640099 }, //07 September 1989, UT1-UTC=-0.4560988, Accumulated Leap Seconds=24
  { 2447777.50, 56.640791 }, //08 September 1989, UT1-UTC=-0.4567912, Accumulated Leap Seconds=24
  { 2447778.50, 56.641424 }, //09 September 1989, UT1-UTC=-0.4574237, Accumulated Leap Seconds=24
  { 2447779.50, 56.642076 }, //10 September 1989, UT1-UTC=-0.4580759, Accumulated Leap Seconds=24
  { 2447780.50, 56.642841 }, //11 September 1989, UT1-UTC=-0.4588413, Accumulated Leap Seconds=24
  { 2447781.50, 56.643814 }, //12 September 1989, UT1-UTC=-0.4598142, Accumulated Leap Seconds=24
  { 2447782.50, 56.645086 }, //13 September 1989, UT1-UTC=-0.4610861, Accumulated Leap Seconds=24
  { 2447783.50, 56.646701 }, //14 September 1989, UT1-UTC=-0.4627008, Accumulated Leap Seconds=24
  { 2447784.50, 56.648626 }, //15 September 1989, UT1-UTC=-0.4646257, Accumulated Leap Seconds=24
  { 2447785.50, 56.650749 }, //16 September 1989, UT1-UTC=-0.4667494, Accumulated Leap Seconds=24
  { 2447786.50, 56.652894 }, //17 September 1989, UT1-UTC=-0.4688936, Accumulated Leap Seconds=24
  { 2447787.50, 56.654880 }, //18 September 1989, UT1-UTC=-0.4708799, Accumulated Leap Seconds=24
  { 2447788.50, 56.656588 }, //19 September 1989, UT1-UTC=-0.4725877, Accumulated Leap Seconds=24
  { 2447789.50, 56.657986 }, //20 September 1989, UT1-UTC=-0.4739864, Accumulated Leap Seconds=24
  { 2447790.50, 56.659135 }, //21 September 1989, UT1-UTC=-0.4751349, Accumulated Leap Seconds=24
  { 2447791.50, 56.660149 }, //22 September 1989, UT1-UTC=-0.4761491, Accumulated Leap Seconds=24
  { 2447792.50, 56.661158 }, //23 September 1989, UT1-UTC=-0.4771580, Accumulated Leap Seconds=24
  { 2447793.50, 56.662266 }, //24 September 1989, UT1-UTC=-0.4782659, Accumulated Leap Seconds=24
  { 2447794.50, 56.663536 }, //25 September 1989, UT1-UTC=-0.4795362, Accumulated Leap Seconds=24
  { 2447795.50, 56.664998 }, //26 September 1989, UT1-UTC=-0.4809976, Accumulated Leap Seconds=24
  { 2447796.50, 56.666641 }, //27 September 1989, UT1-UTC=-0.4826408, Accumulated Leap Seconds=24
  { 2447797.50, 56.668423 }, //28 September 1989, UT1-UTC=-0.4844230, Accumulated Leap Seconds=24
  { 2447798.50, 56.670277 }, //29 September 1989, UT1-UTC=-0.4862767, Accumulated Leap Seconds=24
  { 2447799.50, 56.672122 }, //30 September 1989, UT1-UTC=-0.4881223, Accumulated Leap Seconds=24
  { 2447800.50, 56.673881 }, //01 October 1989, UT1-UTC=-0.4898814, Accumulated Leap Seconds=24
  { 2447801.50, 56.675503 }, //02 October 1989, UT1-UTC=-0.4915025, Accumulated Leap Seconds=24
  { 2447802.50, 56.676966 }, //03 October 1989, UT1-UTC=-0.4929657, Accumulated Leap Seconds=24
  { 2447803.50, 56.678278 }, //04 October 1989, UT1-UTC=-0.4942776, Accumulated Leap Seconds=24
  { 2447804.50, 56.679471 }, //05 October 1989, UT1-UTC=-0.4954708, Accumulated Leap Seconds=24
  { 2447805.50, 56.680601 }, //06 October 1989, UT1-UTC=-0.4966012, Accumulated Leap Seconds=24
  { 2447806.50, 56.681739 }, //07 October 1989, UT1-UTC=-0.4977390, Accumulated Leap Seconds=24
  { 2447807.50, 56.682968 }, //08 October 1989, UT1-UTC=-0.4989679, Accumulated Leap Seconds=24
  { 2447808.50, 56.684376 }, //09 October 1989, UT1-UTC=-0.5003764, Accumulated Leap Seconds=24
  { 2447809.50, 56.686043 }, //10 October 1989, UT1-UTC=-0.5020430, Accumulated Leap Seconds=24
  { 2447810.50, 56.688019 }, //11 October 1989, UT1-UTC=-0.5040190, Accumulated Leap Seconds=24
  { 2447811.50, 56.690309 }, //12 October 1989, UT1-UTC=-0.5063092, Accumulated Leap Seconds=24
  { 2447812.50, 56.692836 }, //13 October 1989, UT1-UTC=-0.5088360, Accumulated Leap Seconds=24
  { 2447813.50, 56.695453 }, //14 October 1989, UT1-UTC=-0.5114535, Accumulated Leap Seconds=24
  { 2447814.50, 56.697988 }, //15 October 1989, UT1-UTC=-0.5139882, Accumulated Leap Seconds=24
  { 2447815.50, 56.700302 }, //16 October 1989, UT1-UTC=-0.5163023, Accumulated Leap Seconds=24
  { 2447816.50, 56.702355 }, //17 October 1989, UT1-UTC=-0.5183552, Accumulated Leap Seconds=24
  { 2447817.50, 56.704181 }, //18 October 1989, UT1-UTC=-0.5201807, Accumulated Leap Seconds=24
  { 2447818.50, 56.705882 }, //19 October 1989, UT1-UTC=-0.5218821, Accumulated Leap Seconds=24
  { 2447819.50, 56.707587 }, //20 October 1989, UT1-UTC=-0.5235875, Accumulated Leap Seconds=24
  { 2447820.50, 56.709397 }, //21 October 1989, UT1-UTC=-0.5253966, Accumulated Leap Seconds=24
  { 2447821.50, 56.711369 }, //22 October 1989, UT1-UTC=-0.5273690, Accumulated Leap Seconds=24
  { 2447822.50, 56.713520 }, //23 October 1989, UT1-UTC=-0.5295201, Accumulated Leap Seconds=24
  { 2447823.50, 56.715824 }, //24 October 1989, UT1-UTC=-0.5318237, Accumulated Leap Seconds=24
  { 2447824.50, 56.718218 }, //25 October 1989, UT1-UTC=-0.5342181, Accumulated Leap Seconds=24
  { 2447825.50, 56.720617 }, //26 October 1989, UT1-UTC=-0.5366175, Accumulated Leap Seconds=24
  { 2447826.50, 56.722973 }, //27 October 1989, UT1-UTC=-0.5389732, Accumulated Leap Seconds=24
  { 2447827.50, 56.725271 }, //28 October 1989, UT1-UTC=-0.5412707, Accumulated Leap Seconds=24
  { 2447828.50, 56.727472 }, //29 October 1989, UT1-UTC=-0.5434715, Accumulated Leap Seconds=24
  { 2447829.50, 56.729535 }, //30 October 1989, UT1-UTC=-0.5455347, Accumulated Leap Seconds=24
  { 2447830.50, 56.731445 }, //31 October 1989, UT1-UTC=-0.5474450, Accumulated Leap Seconds=24
  { 2447831.50, 56.733212 }, //01 November 1989, UT1-UTC=-0.5492116, Accumulated Leap Seconds=24
  { 2447832.50, 56.734862 }, //02 November 1989, UT1-UTC=-0.5508622, Accumulated Leap Seconds=24
  { 2447833.50, 56.736443 }, //03 November 1989, UT1-UTC=-0.5524428, Accumulated Leap Seconds=24
  { 2447834.50, 56.738022 }, //04 November 1989, UT1-UTC=-0.5540217, Accumulated Leap Seconds=24
  { 2447835.50, 56.739688 }, //05 November 1989, UT1-UTC=-0.5556885, Accumulated Leap Seconds=24
  { 2447836.50, 56.741541 }, //06 November 1989, UT1-UTC=-0.5575408, Accumulated Leap Seconds=24
  { 2447837.50, 56.743652 }, //07 November 1989, UT1-UTC=-0.5596524, Accumulated Leap Seconds=24
  { 2447838.50, 56.746048 }, //08 November 1989, UT1-UTC=-0.5620482, Accumulated Leap Seconds=24
  { 2447839.50, 56.748695 }, //09 November 1989, UT1-UTC=-0.5646954, Accumulated Leap Seconds=24
  { 2447840.50, 56.751491 }, //10 November 1989, UT1-UTC=-0.5674907, Accumulated Leap Seconds=24
  { 2447841.50, 56.754270 }, //11 November 1989, UT1-UTC=-0.5702703, Accumulated Leap Seconds=24
  { 2447842.50, 56.756873 }, //12 November 1989, UT1-UTC=-0.5728730, Accumulated Leap Seconds=24
  { 2447843.50, 56.759195 }, //13 November 1989, UT1-UTC=-0.5751947, Accumulated Leap Seconds=24
  { 2447844.50, 56.761222 }, //14 November 1989, UT1-UTC=-0.5772224, Accumulated Leap Seconds=24
  { 2447845.50, 56.763028 }, //15 November 1989, UT1-UTC=-0.5790283, Accumulated Leap Seconds=24
  { 2447846.50, 56.764744 }, //16 November 1989, UT1-UTC=-0.5807437, Accumulated Leap Seconds=24
  { 2447847.50, 56.766505 }, //17 November 1989, UT1-UTC=-0.5825050, Accumulated Leap Seconds=24
  { 2447848.50, 56.768413 }, //18 November 1989, UT1-UTC=-0.5844129, Accumulated Leap Seconds=24
  { 2447849.50, 56.770516 }, //19 November 1989, UT1-UTC=-0.5865157, Accumulated Leap Seconds=24
  { 2447850.50, 56.772808 }, //20 November 1989, UT1-UTC=-0.5888078, Accumulated Leap Seconds=24
  { 2447851.50, 56.775250 }, //21 November 1989, UT1-UTC=-0.5912498, Accumulated Leap Seconds=24
  { 2447852.50, 56.777783 }, //22 November 1989, UT1-UTC=-0.5937826, Accumulated Leap Seconds=24
  { 2447853.50, 56.780343 }, //23 November 1989, UT1-UTC=-0.5963429, Accumulated Leap Seconds=24
  { 2447854.50, 56.782872 }, //24 November 1989, UT1-UTC=-0.5988717, Accumulated Leap Seconds=24
  { 2447855.50, 56.785319 }, //25 November 1989, UT1-UTC=-0.6013194, Accumulated Leap Seconds=24
  { 2447856.50, 56.787633 }, //26 November 1989, UT1-UTC=-0.6036329, Accumulated Leap Seconds=24
  { 2447857.50, 56.789773 }, //27 November 1989, UT1-UTC=-0.6057728, Accumulated Leap Seconds=24
  { 2447858.50, 56.791748 }, //28 November 1989, UT1-UTC=-0.6077485, Accumulated Leap Seconds=24
  { 2447859.50, 56.793600 }, //29 November 1989, UT1-UTC=-0.6095996, Accumulated Leap Seconds=24
  { 2447860.50, 56.795383 }, //30 November 1989, UT1-UTC=-0.6113828, Accumulated Leap Seconds=24
  { 2447861.50, 56.797160 }, //01 December 1989, UT1-UTC=-0.6131596, Accumulated Leap Seconds=24
  { 2447862.50, 56.798996 }, //02 December 1989, UT1-UTC=-0.6149960, Accumulated Leap Seconds=24
  { 2447863.50, 56.800952 }, //03 December 1989, UT1-UTC=-0.6169522, Accumulated Leap Seconds=24
  { 2447864.50, 56.803070 }, //04 December 1989, UT1-UTC=-0.6190704, Accumulated Leap Seconds=24
  { 2447865.50, 56.805363 }, //05 December 1989, UT1-UTC=-0.6213625, Accumulated Leap Seconds=24
  { 2447866.50, 56.807815 }, //06 December 1989, UT1-UTC=-0.6238147, Accumulated Leap Seconds=24
  { 2447867.50, 56.810369 }, //07 December 1989, UT1-UTC=-0.6263694, Accumulated Leap Seconds=24
  { 2447868.50, 56.812923 }, //08 December 1989, UT1-UTC=-0.6289230, Accumulated Leap Seconds=24
  { 2447869.50, 56.815351 }, //09 December 1989, UT1-UTC=-0.6313513, Accumulated Leap Seconds=24
  { 2447870.50, 56.817546 }, //10 December 1989, UT1-UTC=-0.6335460, Accumulated Leap Seconds=24
  { 2447871.50, 56.819457 }, //11 December 1989, UT1-UTC=-0.6354566, Accumulated Leap Seconds=24
  { 2447872.50, 56.821114 }, //12 December 1989, UT1-UTC=-0.6371141, Accumulated Leap Seconds=24
  { 2447873.50, 56.822617 }, //13 December 1989, UT1-UTC=-0.6386168, Accumulated Leap Seconds=24
  { 2447874.50, 56.824093 }, //14 December 1989, UT1-UTC=-0.6400927, Accumulated Leap Seconds=24
  { 2447875.50, 56.825650 }, //15 December 1989, UT1-UTC=-0.6416503, Accumulated Leap Seconds=24
  { 2447876.50, 56.827365 }, //16 December 1989, UT1-UTC=-0.6433649, Accumulated Leap Seconds=24
  { 2447877.50, 56.829258 }, //17 December 1989, UT1-UTC=-0.6452577, Accumulated Leap Seconds=24
  { 2447878.50, 56.831304 }, //18 December 1989, UT1-UTC=-0.6473040, Accumulated Leap Seconds=24
  { 2447879.50, 56.833453 }, //19 December 1989, UT1-UTC=-0.6494535, Accumulated Leap Seconds=24
  { 2447880.50, 56.835645 }, //20 December 1989, UT1-UTC=-0.6516454, Accumulated Leap Seconds=24
  { 2447881.50, 56.837810 }, //21 December 1989, UT1-UTC=-0.6538100, Accumulated Leap Seconds=24
  { 2447882.50, 56.839885 }, //22 December 1989, UT1-UTC=-0.6558848, Accumulated Leap Seconds=24
  { 2447883.50, 56.841820 }, //23 December 1989, UT1-UTC=-0.6578201, Accumulated Leap Seconds=24
  { 2447884.50, 56.843578 }, //24 December 1989, UT1-UTC=-0.6595778, Accumulated Leap Seconds=24
  { 2447885.50, 56.845147 }, //25 December 1989, UT1-UTC=-0.6611469, Accumulated Leap Seconds=24
  { 2447886.50, 56.846554 }, //26 December 1989, UT1-UTC=-0.6625538, Accumulated Leap Seconds=24
  { 2447887.50, 56.847851 }, //27 December 1989, UT1-UTC=-0.6638512, Accumulated Leap Seconds=24
  { 2447888.50, 56.849110 }, //28 December 1989, UT1-UTC=-0.6651098, Accumulated Leap Seconds=24
  { 2447889.50, 56.850409 }, //29 December 1989, UT1-UTC=-0.6664093, Accumulated Leap Seconds=24
  { 2447890.50, 56.851828 }, //30 December 1989, UT1-UTC=-0.6678277, Accumulated Leap Seconds=24
  { 2447891.50, 56.853435 }, //31 December 1989, UT1-UTC=-0.6694346, Accumulated Leap Seconds=24
  { 2447892.50, 56.855270 }, //01 January 1990, UT1-UTC= 0.3287299, Accumulated Leap Seconds=25
  { 2447893.50, 56.857324 }, //02 January 1990, UT1-UTC= 0.3266755, Accumulated Leap Seconds=25
  { 2447894.50, 56.859539 }, //03 January 1990, UT1-UTC= 0.3244612, Accumulated Leap Seconds=25
  { 2447895.50, 56.861817 }, //04 January 1990, UT1-UTC= 0.3221827, Accumulated Leap Seconds=25
  { 2447896.50, 56.864033 }, //05 January 1990, UT1-UTC= 0.3199666, Accumulated Leap Seconds=25
  { 2447897.50, 56.866067 }, //06 January 1990, UT1-UTC= 0.3179327, Accumulated Leap Seconds=25
  { 2447898.50, 56.867845 }, //07 January 1990, UT1-UTC= 0.3161554, Accumulated Leap Seconds=25
  { 2447899.50, 56.869352 }, //08 January 1990, UT1-UTC= 0.3146479, Accumulated Leap Seconds=25
  { 2447900.50, 56.870641 }, //09 January 1990, UT1-UTC= 0.3133590, Accumulated Leap Seconds=25
  { 2447901.50, 56.871826 }, //10 January 1990, UT1-UTC= 0.3121743, Accumulated Leap Seconds=25
  { 2447902.50, 56.873040 }, //11 January 1990, UT1-UTC= 0.3109595, Accumulated Leap Seconds=25
  { 2447903.50, 56.874398 }, //12 January 1990, UT1-UTC= 0.3096018, Accumulated Leap Seconds=25
  { 2447904.50, 56.875961 }, //13 January 1990, UT1-UTC= 0.3080388, Accumulated Leap Seconds=25
  { 2447905.50, 56.877729 }, //14 January 1990, UT1-UTC= 0.3062707, Accumulated Leap Seconds=25
  { 2447906.50, 56.879637 }, //15 January 1990, UT1-UTC= 0.3043627, Accumulated Leap Seconds=25
  { 2447907.50, 56.881597 }, //16 January 1990, UT1-UTC= 0.3024028, Accumulated Leap Seconds=25
  { 2447908.50, 56.883527 }, //17 January 1990, UT1-UTC= 0.3004731, Accumulated Leap Seconds=25
  { 2447909.50, 56.885362 }, //18 January 1990, UT1-UTC= 0.2986380, Accumulated Leap Seconds=25
  { 2447910.50, 56.887061 }, //19 January 1990, UT1-UTC= 0.2969392, Accumulated Leap Seconds=25
  { 2447911.50, 56.888615 }, //20 January 1990, UT1-UTC= 0.2953849, Accumulated Leap Seconds=25
  { 2447912.50, 56.890030 }, //21 January 1990, UT1-UTC= 0.2939698, Accumulated Leap Seconds=25
  { 2447913.50, 56.891332 }, //22 January 1990, UT1-UTC= 0.2926678, Accumulated Leap Seconds=25
  { 2447914.50, 56.892572 }, //23 January 1990, UT1-UTC= 0.2914283, Accumulated Leap Seconds=25
  { 2447915.50, 56.893818 }, //24 January 1990, UT1-UTC= 0.2901819, Accumulated Leap Seconds=25
  { 2447916.50, 56.895158 }, //25 January 1990, UT1-UTC= 0.2888420, Accumulated Leap Seconds=25
  { 2447917.50, 56.896685 }, //26 January 1990, UT1-UTC= 0.2873146, Accumulated Leap Seconds=25
  { 2447918.50, 56.898477 }, //27 January 1990, UT1-UTC= 0.2855226, Accumulated Leap Seconds=25
  { 2447919.50, 56.900573 }, //28 January 1990, UT1-UTC= 0.2834268, Accumulated Leap Seconds=25
  { 2447920.50, 56.902963 }, //29 January 1990, UT1-UTC= 0.2810372, Accumulated Leap Seconds=25
  { 2447921.50, 56.905593 }, //30 January 1990, UT1-UTC= 0.2784068, Accumulated Leap Seconds=25
  { 2447922.50, 56.908367 }, //31 January 1990, UT1-UTC= 0.2756331, Accumulated Leap Seconds=25
  { 2447923.50, 56.911138 }, //01 February 1990, UT1-UTC= 0.2728622, Accumulated Leap Seconds=25
  { 2447924.50, 56.913772 }, //02 February 1990, UT1-UTC= 0.2702280, Accumulated Leap Seconds=25
  { 2447925.50, 56.916183 }, //03 February 1990, UT1-UTC= 0.2678168, Accumulated Leap Seconds=25
  { 2447926.50, 56.918362 }, //04 February 1990, UT1-UTC= 0.2656382, Accumulated Leap Seconds=25
  { 2447927.50, 56.920372 }, //05 February 1990, UT1-UTC= 0.2636275, Accumulated Leap Seconds=25
  { 2447928.50, 56.922322 }, //06 February 1990, UT1-UTC= 0.2616782, Accumulated Leap Seconds=25
  { 2447929.50, 56.924327 }, //07 February 1990, UT1-UTC= 0.2596732, Accumulated Leap Seconds=25
  { 2447930.50, 56.926479 }, //08 February 1990, UT1-UTC= 0.2575205, Accumulated Leap Seconds=25
  { 2447931.50, 56.928835 }, //09 February 1990, UT1-UTC= 0.2551654, Accumulated Leap Seconds=25
  { 2447932.50, 56.931389 }, //10 February 1990, UT1-UTC= 0.2526112, Accumulated Leap Seconds=25
  { 2447933.50, 56.934092 }, //11 February 1990, UT1-UTC= 0.2499077, Accumulated Leap Seconds=25
  { 2447934.50, 56.936873 }, //12 February 1990, UT1-UTC= 0.2471273, Accumulated Leap Seconds=25
  { 2447935.50, 56.939650 }, //13 February 1990, UT1-UTC= 0.2443503, Accumulated Leap Seconds=25
  { 2447936.50, 56.942349 }, //14 February 1990, UT1-UTC= 0.2416508, Accumulated Leap Seconds=25
  { 2447937.50, 56.944914 }, //15 February 1990, UT1-UTC= 0.2390857, Accumulated Leap Seconds=25
  { 2447938.50, 56.947311 }, //16 February 1990, UT1-UTC= 0.2366892, Accumulated Leap Seconds=25
  { 2447939.50, 56.949533 }, //17 February 1990, UT1-UTC= 0.2344671, Accumulated Leap Seconds=25
  { 2447940.50, 56.951581 }, //18 February 1990, UT1-UTC= 0.2324190, Accumulated Leap Seconds=25
  { 2447941.50, 56.953470 }, //19 February 1990, UT1-UTC= 0.2305295, Accumulated Leap Seconds=25
  { 2447942.50, 56.955223 }, //20 February 1990, UT1-UTC= 0.2287768, Accumulated Leap Seconds=25
  { 2447943.50, 56.956893 }, //21 February 1990, UT1-UTC= 0.2271069, Accumulated Leap Seconds=25
  { 2447944.50, 56.958594 }, //22 February 1990, UT1-UTC= 0.2254056, Accumulated Leap Seconds=25
  { 2447945.50, 56.960446 }, //23 February 1990, UT1-UTC= 0.2235538, Accumulated Leap Seconds=25
  { 2447946.50, 56.962532 }, //24 February 1990, UT1-UTC= 0.2214677, Accumulated Leap Seconds=25
  { 2447947.50, 56.964888 }, //25 February 1990, UT1-UTC= 0.2191119, Accumulated Leap Seconds=25
  { 2447948.50, 56.967480 }, //26 February 1990, UT1-UTC= 0.2165203, Accumulated Leap Seconds=25
  { 2447949.50, 56.970205 }, //27 February 1990, UT1-UTC= 0.2137950, Accumulated Leap Seconds=25
  { 2447950.50, 56.972921 }, //28 February 1990, UT1-UTC= 0.2110790, Accumulated Leap Seconds=25
  { 2447951.50, 56.975472 }, //01 March 1990, UT1-UTC= 0.2085275, Accumulated Leap Seconds=25
  { 2447952.50, 56.977755 }, //02 March 1990, UT1-UTC= 0.2062450, Accumulated Leap Seconds=25
  { 2447953.50, 56.979753 }, //03 March 1990, UT1-UTC= 0.2042473, Accumulated Leap Seconds=25
  { 2447954.50, 56.981526 }, //04 March 1990, UT1-UTC= 0.2024744, Accumulated Leap Seconds=25
  { 2447955.50, 56.983180 }, //05 March 1990, UT1-UTC= 0.2008204, Accumulated Leap Seconds=25
  { 2447956.50, 56.984837 }, //06 March 1990, UT1-UTC= 0.1991629, Accumulated Leap Seconds=25
  { 2447957.50, 56.986611 }, //07 March 1990, UT1-UTC= 0.1973892, Accumulated Leap Seconds=25
  { 2447958.50, 56.988578 }, //08 March 1990, UT1-UTC= 0.1954224, Accumulated Leap Seconds=25
  { 2447959.50, 56.990770 }, //09 March 1990, UT1-UTC= 0.1932302, Accumulated Leap Seconds=25
  { 2447960.50, 56.993175 }, //10 March 1990, UT1-UTC= 0.1908252, Accumulated Leap Seconds=25
  { 2447961.50, 56.995720 }, //11 March 1990, UT1-UTC= 0.1882804, Accumulated Leap Seconds=25
  { 2447962.50, 56.998309 }, //12 March 1990, UT1-UTC= 0.1856908, Accumulated Leap Seconds=25
  { 2447963.50, 57.000851 }, //13 March 1990, UT1-UTC= 0.1831488, Accumulated Leap Seconds=25
  { 2447964.50, 57.003269 }, //14 March 1990, UT1-UTC= 0.1807307, Accumulated Leap Seconds=25
  { 2447965.50, 57.005518 }, //15 March 1990, UT1-UTC= 0.1784817, Accumulated Leap Seconds=25
  { 2447966.50, 57.007612 }, //16 March 1990, UT1-UTC= 0.1763881, Accumulated Leap Seconds=25
  { 2447967.50, 57.009586 }, //17 March 1990, UT1-UTC= 0.1744136, Accumulated Leap Seconds=25
  { 2447968.50, 57.011486 }, //18 March 1990, UT1-UTC= 0.1725138, Accumulated Leap Seconds=25
  { 2447969.50, 57.013372 }, //19 March 1990, UT1-UTC= 0.1706279, Accumulated Leap Seconds=25
  { 2447970.50, 57.015316 }, //20 March 1990, UT1-UTC= 0.1686844, Accumulated Leap Seconds=25
  { 2447971.50, 57.017386 }, //21 March 1990, UT1-UTC= 0.1666143, Accumulated Leap Seconds=25
  { 2447972.50, 57.019649 }, //22 March 1990, UT1-UTC= 0.1643506, Accumulated Leap Seconds=25
  { 2447973.50, 57.022162 }, //23 March 1990, UT1-UTC= 0.1618382, Accumulated Leap Seconds=25
  { 2447974.50, 57.024954 }, //24 March 1990, UT1-UTC= 0.1590458, Accumulated Leap Seconds=25
  { 2447975.50, 57.028017 }, //25 March 1990, UT1-UTC= 0.1559830, Accumulated Leap Seconds=25
  { 2447976.50, 57.031258 }, //26 March 1990, UT1-UTC= 0.1527418, Accumulated Leap Seconds=25
  { 2447977.50, 57.034525 }, //27 March 1990, UT1-UTC= 0.1494752, Accumulated Leap Seconds=25
  { 2447978.50, 57.037644 }, //28 March 1990, UT1-UTC= 0.1463557, Accumulated Leap Seconds=25
  { 2447979.50, 57.040474 }, //29 March 1990, UT1-UTC= 0.1435257, Accumulated Leap Seconds=25
  { 2447980.50, 57.042944 }, //30 March 1990, UT1-UTC= 0.1410556, Accumulated Leap Seconds=25
  { 2447981.50, 57.045106 }, //31 March 1990, UT1-UTC= 0.1388943, Accumulated Leap Seconds=25
  { 2447982.50, 57.047077 }, //01 April 1990, UT1-UTC= 0.1369228, Accumulated Leap Seconds=25
  { 2447983.50, 57.049000 }, //02 April 1990, UT1-UTC= 0.1350002, Accumulated Leap Seconds=25
  { 2447984.50, 57.051002 }, //03 April 1990, UT1-UTC= 0.1329978, Accumulated Leap Seconds=25
  { 2447985.50, 57.053171 }, //04 April 1990, UT1-UTC= 0.1308293, Accumulated Leap Seconds=25
  { 2447986.50, 57.055518 }, //05 April 1990, UT1-UTC= 0.1284821, Accumulated Leap Seconds=25
  { 2447987.50, 57.058008 }, //06 April 1990, UT1-UTC= 0.1259920, Accumulated Leap Seconds=25
  { 2447988.50, 57.060596 }, //07 April 1990, UT1-UTC= 0.1234035, Accumulated Leap Seconds=25
  { 2447989.50, 57.063227 }, //08 April 1990, UT1-UTC= 0.1207725, Accumulated Leap Seconds=25
  { 2447990.50, 57.065837 }, //09 April 1990, UT1-UTC= 0.1181634, Accumulated Leap Seconds=25
  { 2447991.50, 57.068359 }, //10 April 1990, UT1-UTC= 0.1156414, Accumulated Leap Seconds=25
  { 2447992.50, 57.070743 }, //11 April 1990, UT1-UTC= 0.1132566, Accumulated Leap Seconds=25
  { 2447993.50, 57.072965 }, //12 April 1990, UT1-UTC= 0.1110349, Accumulated Leap Seconds=25
  { 2447994.50, 57.075024 }, //13 April 1990, UT1-UTC= 0.1089757, Accumulated Leap Seconds=25
  { 2447995.50, 57.076937 }, //14 April 1990, UT1-UTC= 0.1070634, Accumulated Leap Seconds=25
  { 2447996.50, 57.078740 }, //15 April 1990, UT1-UTC= 0.1052598, Accumulated Leap Seconds=25
  { 2447997.50, 57.080501 }, //16 April 1990, UT1-UTC= 0.1034993, Accumulated Leap Seconds=25
  { 2447998.50, 57.082294 }, //17 April 1990, UT1-UTC= 0.1017055, Accumulated Leap Seconds=25
  { 2447999.50, 57.084203 }, //18 April 1990, UT1-UTC= 0.0997972, Accumulated Leap Seconds=25
  { 2448000.50, 57.086314 }, //19 April 1990, UT1-UTC= 0.0976858, Accumulated Leap Seconds=25
  { 2448001.50, 57.088684 }, //20 April 1990, UT1-UTC= 0.0953157, Accumulated Leap Seconds=25
  { 2448002.50, 57.091324 }, //21 April 1990, UT1-UTC= 0.0926757, Accumulated Leap Seconds=25
  { 2448003.50, 57.094184 }, //22 April 1990, UT1-UTC= 0.0898164, Accumulated Leap Seconds=25
  { 2448004.50, 57.097147 }, //23 April 1990, UT1-UTC= 0.0868525, Accumulated Leap Seconds=25
  { 2448005.50, 57.100056 }, //24 April 1990, UT1-UTC= 0.0839438, Accumulated Leap Seconds=25
  { 2448006.50, 57.102756 }, //25 April 1990, UT1-UTC= 0.0812443, Accumulated Leap Seconds=25
  { 2448007.50, 57.105132 }, //26 April 1990, UT1-UTC= 0.0788675, Accumulated Leap Seconds=25
  { 2448008.50, 57.107159 }, //27 April 1990, UT1-UTC= 0.0768409, Accumulated Leap Seconds=25
  { 2448009.50, 57.108900 }, //28 April 1990, UT1-UTC= 0.0751001, Accumulated Leap Seconds=25
  { 2448010.50, 57.110462 }, //29 April 1990, UT1-UTC= 0.0735384, Accumulated Leap Seconds=25
  { 2448011.50, 57.111990 }, //30 April 1990, UT1-UTC= 0.0720096, Accumulated Leap Seconds=25
  { 2448012.50, 57.113613 }, //01 May 1990, UT1-UTC= 0.0703872, Accumulated Leap Seconds=25
  { 2448013.50, 57.115397 }, //02 May 1990, UT1-UTC= 0.0686030, Accumulated Leap Seconds=25
  { 2448014.50, 57.117352 }, //03 May 1990, UT1-UTC= 0.0666480, Accumulated Leap Seconds=25
  { 2448015.50, 57.119441 }, //04 May 1990, UT1-UTC= 0.0645590, Accumulated Leap Seconds=25
  { 2448016.50, 57.121601 }, //05 May 1990, UT1-UTC= 0.0623985, Accumulated Leap Seconds=25
  { 2448017.50, 57.123758 }, //06 May 1990, UT1-UTC= 0.0602422, Accumulated Leap Seconds=25
  { 2448018.50, 57.125839 }, //07 May 1990, UT1-UTC= 0.0581609, Accumulated Leap Seconds=25
  { 2448019.50, 57.127789 }, //08 May 1990, UT1-UTC= 0.0562105, Accumulated Leap Seconds=25
  { 2448020.50, 57.129578 }, //09 May 1990, UT1-UTC= 0.0544221, Accumulated Leap Seconds=25
  { 2448021.50, 57.131217 }, //10 May 1990, UT1-UTC= 0.0527826, Accumulated Leap Seconds=25
  { 2448022.50, 57.132733 }, //11 May 1990, UT1-UTC= 0.0512668, Accumulated Leap Seconds=25
  { 2448023.50, 57.134166 }, //12 May 1990, UT1-UTC= 0.0498338, Accumulated Leap Seconds=25
  { 2448024.50, 57.135576 }, //13 May 1990, UT1-UTC= 0.0484244, Accumulated Leap Seconds=25
  { 2448025.50, 57.137031 }, //14 May 1990, UT1-UTC= 0.0469688, Accumulated Leap Seconds=25
  { 2448026.50, 57.138614 }, //15 May 1990, UT1-UTC= 0.0453861, Accumulated Leap Seconds=25
  { 2448027.50, 57.140395 }, //16 May 1990, UT1-UTC= 0.0436049, Accumulated Leap Seconds=25
  { 2448028.50, 57.142404 }, //17 May 1990, UT1-UTC= 0.0415965, Accumulated Leap Seconds=25
  { 2448029.50, 57.144650 }, //18 May 1990, UT1-UTC= 0.0393498, Accumulated Leap Seconds=25
  { 2448030.50, 57.147110 }, //19 May 1990, UT1-UTC= 0.0368903, Accumulated Leap Seconds=25
  { 2448031.50, 57.149716 }, //20 May 1990, UT1-UTC= 0.0342836, Accumulated Leap Seconds=25
  { 2448032.50, 57.152367 }, //21 May 1990, UT1-UTC= 0.0316328, Accumulated Leap Seconds=25
  { 2448033.50, 57.154935 }, //22 May 1990, UT1-UTC= 0.0290653, Accumulated Leap Seconds=25
  { 2448034.50, 57.157303 }, //23 May 1990, UT1-UTC= 0.0266965, Accumulated Leap Seconds=25
  { 2448035.50, 57.159393 }, //24 May 1990, UT1-UTC= 0.0246069, Accumulated Leap Seconds=25
  { 2448036.50, 57.161192 }, //25 May 1990, UT1-UTC= 0.0228081, Accumulated Leap Seconds=25
  { 2448037.50, 57.162783 }, //26 May 1990, UT1-UTC= 0.0212166, Accumulated Leap Seconds=25
  { 2448038.50, 57.164320 }, //27 May 1990, UT1-UTC= 0.0196804, Accumulated Leap Seconds=25
  { 2448039.50, 57.165939 }, //28 May 1990, UT1-UTC= 0.0180611, Accumulated Leap Seconds=25
  { 2448040.50, 57.167719 }, //29 May 1990, UT1-UTC= 0.0162813, Accumulated Leap Seconds=25
  { 2448041.50, 57.169663 }, //30 May 1990, UT1-UTC= 0.0143374, Accumulated Leap Seconds=25
  { 2448042.50, 57.171723 }, //31 May 1990, UT1-UTC= 0.0122766, Accumulated Leap Seconds=25
  { 2448043.50, 57.173831 }, //01 June 1990, UT1-UTC= 0.0101690, Accumulated Leap Seconds=25
  { 2448044.50, 57.175911 }, //02 June 1990, UT1-UTC= 0.0080890, Accumulated Leap Seconds=25
  { 2448045.50, 57.177899 }, //03 June 1990, UT1-UTC= 0.0061014, Accumulated Leap Seconds=25
  { 2448046.50, 57.179744 }, //04 June 1990, UT1-UTC= 0.0042559, Accumulated Leap Seconds=25
  { 2448047.50, 57.181411 }, //05 June 1990, UT1-UTC= 0.0025887, Accumulated Leap Seconds=25
  { 2448048.50, 57.182887 }, //06 June 1990, UT1-UTC= 0.0011131, Accumulated Leap Seconds=25
  { 2448049.50, 57.184193 }, //07 June 1990, UT1-UTC=-0.0001929, Accumulated Leap Seconds=25
  { 2448050.50, 57.185378 }, //08 June 1990, UT1-UTC=-0.0013779, Accumulated Leap Seconds=25
  { 2448051.50, 57.186510 }, //09 June 1990, UT1-UTC=-0.0025105, Accumulated Leap Seconds=25
  { 2448052.50, 57.187670 }, //10 June 1990, UT1-UTC=-0.0036700, Accumulated Leap Seconds=25
  { 2448053.50, 57.188930 }, //11 June 1990, UT1-UTC=-0.0049304, Accumulated Leap Seconds=25
  { 2448054.50, 57.190345 }, //12 June 1990, UT1-UTC=-0.0063454, Accumulated Leap Seconds=25
  { 2448055.50, 57.191940 }, //13 June 1990, UT1-UTC=-0.0079399, Accumulated Leap Seconds=25
  { 2448056.50, 57.193724 }, //14 June 1990, UT1-UTC=-0.0097238, Accumulated Leap Seconds=25
  { 2448057.50, 57.195687 }, //15 June 1990, UT1-UTC=-0.0116874, Accumulated Leap Seconds=25
  { 2448058.50, 57.197784 }, //16 June 1990, UT1-UTC=-0.0137844, Accumulated Leap Seconds=25
  { 2448059.50, 57.199933 }, //17 June 1990, UT1-UTC=-0.0159328, Accumulated Leap Seconds=25
  { 2448060.50, 57.202025 }, //18 June 1990, UT1-UTC=-0.0180254, Accumulated Leap Seconds=25
  { 2448061.50, 57.203960 }, //19 June 1990, UT1-UTC=-0.0199602, Accumulated Leap Seconds=25
  { 2448062.50, 57.205676 }, //20 June 1990, UT1-UTC=-0.0216762, Accumulated Leap Seconds=25
  { 2448063.50, 57.207164 }, //21 June 1990, UT1-UTC=-0.0231643, Accumulated Leap Seconds=25
  { 2448064.50, 57.208473 }, //22 June 1990, UT1-UTC=-0.0244731, Accumulated Leap Seconds=25
  { 2448065.50, 57.209694 }, //23 June 1990, UT1-UTC=-0.0256936, Accumulated Leap Seconds=25
  { 2448066.50, 57.210959 }, //24 June 1990, UT1-UTC=-0.0269587, Accumulated Leap Seconds=25
  { 2448067.50, 57.212370 }, //25 June 1990, UT1-UTC=-0.0283702, Accumulated Leap Seconds=25
  { 2448068.50, 57.213960 }, //26 June 1990, UT1-UTC=-0.0299604, Accumulated Leap Seconds=25
  { 2448069.50, 57.215692 }, //27 June 1990, UT1-UTC=-0.0316925, Accumulated Leap Seconds=25
  { 2448070.50, 57.217487 }, //28 June 1990, UT1-UTC=-0.0334874, Accumulated Leap Seconds=25
  { 2448071.50, 57.219273 }, //29 June 1990, UT1-UTC=-0.0352726, Accumulated Leap Seconds=25
  { 2448072.50, 57.220991 }, //30 June 1990, UT1-UTC=-0.0369906, Accumulated Leap Seconds=25
  { 2448073.50, 57.222607 }, //01 July 1990, UT1-UTC=-0.0386068, Accumulated Leap Seconds=25
  { 2448074.50, 57.224088 }, //02 July 1990, UT1-UTC=-0.0400881, Accumulated Leap Seconds=25
  { 2448075.50, 57.225402 }, //03 July 1990, UT1-UTC=-0.0414022, Accumulated Leap Seconds=25
  { 2448076.50, 57.226562 }, //04 July 1990, UT1-UTC=-0.0425616, Accumulated Leap Seconds=25
  { 2448077.50, 57.227590 }, //05 July 1990, UT1-UTC=-0.0435896, Accumulated Leap Seconds=25
  { 2448078.50, 57.228523 }, //06 July 1990, UT1-UTC=-0.0445233, Accumulated Leap Seconds=25
  { 2448079.50, 57.229411 }, //07 July 1990, UT1-UTC=-0.0454115, Accumulated Leap Seconds=25
  { 2448080.50, 57.230301 }, //08 July 1990, UT1-UTC=-0.0463011, Accumulated Leap Seconds=25
  { 2448081.50, 57.231268 }, //09 July 1990, UT1-UTC=-0.0472682, Accumulated Leap Seconds=25
  { 2448082.50, 57.232388 }, //10 July 1990, UT1-UTC=-0.0483881, Accumulated Leap Seconds=25
  { 2448083.50, 57.233705 }, //11 July 1990, UT1-UTC=-0.0497046, Accumulated Leap Seconds=25
  { 2448084.50, 57.235212 }, //12 July 1990, UT1-UTC=-0.0512119, Accumulated Leap Seconds=25
  { 2448085.50, 57.236849 }, //13 July 1990, UT1-UTC=-0.0528489, Accumulated Leap Seconds=25
  { 2448086.50, 57.238534 }, //14 July 1990, UT1-UTC=-0.0545337, Accumulated Leap Seconds=25
  { 2448087.50, 57.240159 }, //15 July 1990, UT1-UTC=-0.0561593, Accumulated Leap Seconds=25
  { 2448088.50, 57.241620 }, //16 July 1990, UT1-UTC=-0.0576203, Accumulated Leap Seconds=25
  { 2448089.50, 57.242841 }, //17 July 1990, UT1-UTC=-0.0588413, Accumulated Leap Seconds=25
  { 2448090.50, 57.243800 }, //18 July 1990, UT1-UTC=-0.0597996, Accumulated Leap Seconds=25
  { 2448091.50, 57.244540 }, //19 July 1990, UT1-UTC=-0.0605401, Accumulated Leap Seconds=25
  { 2448092.50, 57.245176 }, //20 July 1990, UT1-UTC=-0.0611762, Accumulated Leap Seconds=25
  { 2448093.50, 57.245851 }, //21 July 1990, UT1-UTC=-0.0618514, Accumulated Leap Seconds=25
  { 2448094.50, 57.246693 }, //22 July 1990, UT1-UTC=-0.0626927, Accumulated Leap Seconds=25
  { 2448095.50, 57.247770 }, //23 July 1990, UT1-UTC=-0.0637696, Accumulated Leap Seconds=25
  { 2448096.50, 57.249073 }, //24 July 1990, UT1-UTC=-0.0650727, Accumulated Leap Seconds=25
  { 2448097.50, 57.250532 }, //25 July 1990, UT1-UTC=-0.0665325, Accumulated Leap Seconds=25
  { 2448098.50, 57.252060 }, //26 July 1990, UT1-UTC=-0.0680598, Accumulated Leap Seconds=25
  { 2448099.50, 57.253576 }, //27 July 1990, UT1-UTC=-0.0695760, Accumulated Leap Seconds=25
  { 2448100.50, 57.255036 }, //28 July 1990, UT1-UTC=-0.0710357, Accumulated Leap Seconds=25
  { 2448101.50, 57.256400 }, //29 July 1990, UT1-UTC=-0.0723997, Accumulated Leap Seconds=25
  { 2448102.50, 57.257641 }, //30 July 1990, UT1-UTC=-0.0736414, Accumulated Leap Seconds=25
  { 2448103.50, 57.258751 }, //31 July 1990, UT1-UTC=-0.0747510, Accumulated Leap Seconds=25
  { 2448104.50, 57.259731 }, //01 August 1990, UT1-UTC=-0.0757310, Accumulated Leap Seconds=25
  { 2448105.50, 57.260595 }, //02 August 1990, UT1-UTC=-0.0765954, Accumulated Leap Seconds=25
  { 2448106.50, 57.261409 }, //03 August 1990, UT1-UTC=-0.0774086, Accumulated Leap Seconds=25
  { 2448107.50, 57.262260 }, //04 August 1990, UT1-UTC=-0.0782605, Accumulated Leap Seconds=25
  { 2448108.50, 57.263238 }, //05 August 1990, UT1-UTC=-0.0792382, Accumulated Leap Seconds=25
  { 2448109.50, 57.264417 }, //06 August 1990, UT1-UTC=-0.0804167, Accumulated Leap Seconds=25
  { 2448110.50, 57.265842 }, //07 August 1990, UT1-UTC=-0.0818421, Accumulated Leap Seconds=25
  { 2448111.50, 57.267519 }, //08 August 1990, UT1-UTC=-0.0835193, Accumulated Leap Seconds=25
  { 2448112.50, 57.269407 }, //09 August 1990, UT1-UTC=-0.0854070, Accumulated Leap Seconds=25
  { 2448113.50, 57.271419 }, //10 August 1990, UT1-UTC=-0.0874187, Accumulated Leap Seconds=25
  { 2448114.50, 57.273444 }, //11 August 1990, UT1-UTC=-0.0894445, Accumulated Leap Seconds=25
  { 2448115.50, 57.275372 }, //12 August 1990, UT1-UTC=-0.0913725, Accumulated Leap Seconds=25
  { 2448116.50, 57.277119 }, //13 August 1990, UT1-UTC=-0.0931194, Accumulated Leap Seconds=25
  { 2448117.50, 57.278645 }, //14 August 1990, UT1-UTC=-0.0946455, Accumulated Leap Seconds=25
  { 2448118.50, 57.279973 }, //15 August 1990, UT1-UTC=-0.0959732, Accumulated Leap Seconds=25
  { 2448119.50, 57.281191 }, //16 August 1990, UT1-UTC=-0.0971907, Accumulated Leap Seconds=25
  { 2448120.50, 57.282418 }, //17 August 1990, UT1-UTC=-0.0984181, Accumulated Leap Seconds=25
  { 2448121.50, 57.283778 }, //18 August 1990, UT1-UTC=-0.0997780, Accumulated Leap Seconds=25
  { 2448122.50, 57.285356 }, //19 August 1990, UT1-UTC=-0.1013560, Accumulated Leap Seconds=25
  { 2448123.50, 57.287178 }, //20 August 1990, UT1-UTC=-0.1031780, Accumulated Leap Seconds=25
  { 2448124.50, 57.289207 }, //21 August 1990, UT1-UTC=-0.1052071, Accumulated Leap Seconds=25
  { 2448125.50, 57.291361 }, //22 August 1990, UT1-UTC=-0.1073609, Accumulated Leap Seconds=25
  { 2448126.50, 57.293538 }, //23 August 1990, UT1-UTC=-0.1095377, Accumulated Leap Seconds=25
  { 2448127.50, 57.295649 }, //24 August 1990, UT1-UTC=-0.1116487, Accumulated Leap Seconds=25
  { 2448128.50, 57.297621 }, //25 August 1990, UT1-UTC=-0.1136207, Accumulated Leap Seconds=25
  { 2448129.50, 57.299410 }, //26 August 1990, UT1-UTC=-0.1154098, Accumulated Leap Seconds=25
  { 2448130.50, 57.301003 }, //27 August 1990, UT1-UTC=-0.1170033, Accumulated Leap Seconds=25
  { 2448131.50, 57.302418 }, //28 August 1990, UT1-UTC=-0.1184179, Accumulated Leap Seconds=25
  { 2448132.50, 57.303691 }, //29 August 1990, UT1-UTC=-0.1196906, Accumulated Leap Seconds=25
  { 2448133.50, 57.304875 }, //30 August 1990, UT1-UTC=-0.1208751, Accumulated Leap Seconds=25
  { 2448134.50, 57.306043 }, //31 August 1990, UT1-UTC=-0.1220434, Accumulated Leap Seconds=25
  { 2448135.50, 57.307274 }, //01 September 1990, UT1-UTC=-0.1232742, Accumulated Leap Seconds=25
  { 2448136.50, 57.308654 }, //02 September 1990, UT1-UTC=-0.1246539, Accumulated Leap Seconds=25
  { 2448137.50, 57.310257 }, //03 September 1990, UT1-UTC=-0.1262568, Accumulated Leap Seconds=25
  { 2448138.50, 57.312125 }, //04 September 1990, UT1-UTC=-0.1281251, Accumulated Leap Seconds=25
  { 2448139.50, 57.314254 }, //05 September 1990, UT1-UTC=-0.1302544, Accumulated Leap Seconds=25
  { 2448140.50, 57.316580 }, //06 September 1990, UT1-UTC=-0.1325801, Accumulated Leap Seconds=25
  { 2448141.50, 57.318967 }, //07 September 1990, UT1-UTC=-0.1349674, Accumulated Leap Seconds=25
  { 2448142.50, 57.321264 }, //08 September 1990, UT1-UTC=-0.1372637, Accumulated Leap Seconds=25
  { 2448143.50, 57.323347 }, //09 September 1990, UT1-UTC=-0.1393470, Accumulated Leap Seconds=25
  { 2448144.50, 57.325155 }, //10 September 1990, UT1-UTC=-0.1411555, Accumulated Leap Seconds=25
  { 2448145.50, 57.326703 }, //11 September 1990, UT1-UTC=-0.1427028, Accumulated Leap Seconds=25
  { 2448146.50, 57.328089 }, //12 September 1990, UT1-UTC=-0.1440895, Accumulated Leap Seconds=25
  { 2448147.50, 57.329460 }, //13 September 1990, UT1-UTC=-0.1454603, Accumulated Leap Seconds=25
  { 2448148.50, 57.330948 }, //14 September 1990, UT1-UTC=-0.1469478, Accumulated Leap Seconds=25
  { 2448149.50, 57.332646 }, //15 September 1990, UT1-UTC=-0.1486457, Accumulated Leap Seconds=25
  { 2448150.50, 57.334591 }, //16 September 1990, UT1-UTC=-0.1505908, Accumulated Leap Seconds=25
  { 2448151.50, 57.336757 }, //17 September 1990, UT1-UTC=-0.1527567, Accumulated Leap Seconds=25
  { 2448152.50, 57.339074 }, //18 September 1990, UT1-UTC=-0.1550745, Accumulated Leap Seconds=25
  { 2448153.50, 57.341451 }, //19 September 1990, UT1-UTC=-0.1574511, Accumulated Leap Seconds=25
  { 2448154.50, 57.343792 }, //20 September 1990, UT1-UTC=-0.1597918, Accumulated Leap Seconds=25
  { 2448155.50, 57.346019 }, //21 September 1990, UT1-UTC=-0.1620186, Accumulated Leap Seconds=25
  { 2448156.50, 57.348081 }, //22 September 1990, UT1-UTC=-0.1640808, Accumulated Leap Seconds=25
  { 2448157.50, 57.349968 }, //23 September 1990, UT1-UTC=-0.1659684, Accumulated Leap Seconds=25
  { 2448158.50, 57.351703 }, //24 September 1990, UT1-UTC=-0.1677029, Accumulated Leap Seconds=25
  { 2448159.50, 57.353327 }, //25 September 1990, UT1-UTC=-0.1693274, Accumulated Leap Seconds=25
  { 2448160.50, 57.354900 }, //26 September 1990, UT1-UTC=-0.1708998, Accumulated Leap Seconds=25
  { 2448161.50, 57.356486 }, //27 September 1990, UT1-UTC=-0.1724864, Accumulated Leap Seconds=25
  { 2448162.50, 57.358160 }, //28 September 1990, UT1-UTC=-0.1741597, Accumulated Leap Seconds=25
  { 2448163.50, 57.359990 }, //29 September 1990, UT1-UTC=-0.1759905, Accumulated Leap Seconds=25
  { 2448164.50, 57.362038 }, //30 September 1990, UT1-UTC=-0.1780385, Accumulated Leap Seconds=25
  { 2448165.50, 57.364337 }, //01 October 1990, UT1-UTC=-0.1803368, Accumulated Leap Seconds=25
  { 2448166.50, 57.366873 }, //02 October 1990, UT1-UTC=-0.1828734, Accumulated Leap Seconds=25
  { 2448167.50, 57.369589 }, //03 October 1990, UT1-UTC=-0.1855895, Accumulated Leap Seconds=25
  { 2448168.50, 57.372378 }, //04 October 1990, UT1-UTC=-0.1883780, Accumulated Leap Seconds=25
  { 2448169.50, 57.375108 }, //05 October 1990, UT1-UTC=-0.1911078, Accumulated Leap Seconds=25
  { 2448170.50, 57.377656 }, //06 October 1990, UT1-UTC=-0.1936561, Accumulated Leap Seconds=25
  { 2448171.50, 57.379943 }, //07 October 1990, UT1-UTC=-0.1959435, Accumulated Leap Seconds=25
  { 2448172.50, 57.381966 }, //08 October 1990, UT1-UTC=-0.1979658, Accumulated Leap Seconds=25
  { 2448173.50, 57.383793 }, //09 October 1990, UT1-UTC=-0.1997935, Accumulated Leap Seconds=25
  { 2448174.50, 57.385540 }, //10 October 1990, UT1-UTC=-0.2015404, Accumulated Leap Seconds=25
  { 2448175.50, 57.387326 }, //11 October 1990, UT1-UTC=-0.2033257, Accumulated Leap Seconds=25
  { 2448176.50, 57.389244 }, //12 October 1990, UT1-UTC=-0.2052441, Accumulated Leap Seconds=25
  { 2448177.50, 57.391337 }, //13 October 1990, UT1-UTC=-0.2073366, Accumulated Leap Seconds=25
  { 2448178.50, 57.393590 }, //14 October 1990, UT1-UTC=-0.2095896, Accumulated Leap Seconds=25
  { 2448179.50, 57.395950 }, //15 October 1990, UT1-UTC=-0.2119499, Accumulated Leap Seconds=25
  { 2448180.50, 57.398346 }, //16 October 1990, UT1-UTC=-0.2143460, Accumulated Leap Seconds=25
  { 2448181.50, 57.400707 }, //17 October 1990, UT1-UTC=-0.2167072, Accumulated Leap Seconds=25
  { 2448182.50, 57.402973 }, //18 October 1990, UT1-UTC=-0.2189730, Accumulated Leap Seconds=25
  { 2448183.50, 57.405102 }, //19 October 1990, UT1-UTC=-0.2211017, Accumulated Leap Seconds=25
  { 2448184.50, 57.407077 }, //20 October 1990, UT1-UTC=-0.2230769, Accumulated Leap Seconds=25
  { 2448185.50, 57.408898 }, //21 October 1990, UT1-UTC=-0.2248977, Accumulated Leap Seconds=25
  { 2448186.50, 57.410603 }, //22 October 1990, UT1-UTC=-0.2266028, Accumulated Leap Seconds=25
  { 2448187.50, 57.412259 }, //23 October 1990, UT1-UTC=-0.2282592, Accumulated Leap Seconds=25
  { 2448188.50, 57.413941 }, //24 October 1990, UT1-UTC=-0.2299413, Accumulated Leap Seconds=25
  { 2448189.50, 57.415710 }, //25 October 1990, UT1-UTC=-0.2317097, Accumulated Leap Seconds=25
  { 2448190.50, 57.417616 }, //26 October 1990, UT1-UTC=-0.2336164, Accumulated Leap Seconds=25
  { 2448191.50, 57.419709 }, //27 October 1990, UT1-UTC=-0.2357095, Accumulated Leap Seconds=25
  { 2448192.50, 57.422038 }, //28 October 1990, UT1-UTC=-0.2380376, Accumulated Leap Seconds=25
  { 2448193.50, 57.424625 }, //29 October 1990, UT1-UTC=-0.2406245, Accumulated Leap Seconds=25
  { 2448194.50, 57.427447 }, //30 October 1990, UT1-UTC=-0.2434467, Accumulated Leap Seconds=25
  { 2448195.50, 57.430424 }, //31 October 1990, UT1-UTC=-0.2464239, Accumulated Leap Seconds=25
  { 2448196.50, 57.433428 }, //01 November 1990, UT1-UTC=-0.2494281, Accumulated Leap Seconds=25
  { 2448197.50, 57.436313 }, //02 November 1990, UT1-UTC=-0.2523127, Accumulated Leap Seconds=25
  { 2448198.50, 57.438948 }, //03 November 1990, UT1-UTC=-0.2549484, Accumulated Leap Seconds=25
  { 2448199.50, 57.441266 }, //04 November 1990, UT1-UTC=-0.2572657, Accumulated Leap Seconds=25
  { 2448200.50, 57.443284 }, //05 November 1990, UT1-UTC=-0.2592841, Accumulated Leap Seconds=25
  { 2448201.50, 57.445131 }, //06 November 1990, UT1-UTC=-0.2611312, Accumulated Leap Seconds=25
  { 2448202.50, 57.446978 }, //07 November 1990, UT1-UTC=-0.2629783, Accumulated Leap Seconds=25
  { 2448203.50, 57.448971 }, //08 November 1990, UT1-UTC=-0.2649711, Accumulated Leap Seconds=25
  { 2448204.50, 57.451185 }, //09 November 1990, UT1-UTC=-0.2671853, Accumulated Leap Seconds=25
  { 2448205.50, 57.453619 }, //10 November 1990, UT1-UTC=-0.2696192, Accumulated Leap Seconds=25
  { 2448206.50, 57.456204 }, //11 November 1990, UT1-UTC=-0.2722039, Accumulated Leap Seconds=25
  { 2448207.50, 57.458848 }, //12 November 1990, UT1-UTC=-0.2748479, Accumulated Leap Seconds=25
  { 2448208.50, 57.461459 }, //13 November 1990, UT1-UTC=-0.2774593, Accumulated Leap Seconds=25
  { 2448209.50, 57.463959 }, //14 November 1990, UT1-UTC=-0.2799595, Accumulated Leap Seconds=25
  { 2448210.50, 57.466292 }, //15 November 1990, UT1-UTC=-0.2822923, Accumulated Leap Seconds=25
  { 2448211.50, 57.468432 }, //16 November 1990, UT1-UTC=-0.2844325, Accumulated Leap Seconds=25
  { 2448212.50, 57.470382 }, //17 November 1990, UT1-UTC=-0.2863825, Accumulated Leap Seconds=25
  { 2448213.50, 57.472170 }, //18 November 1990, UT1-UTC=-0.2881696, Accumulated Leap Seconds=25
  { 2448214.50, 57.473846 }, //19 November 1990, UT1-UTC=-0.2898456, Accumulated Leap Seconds=25
  { 2448215.50, 57.475495 }, //20 November 1990, UT1-UTC=-0.2914947, Accumulated Leap Seconds=25
  { 2448216.50, 57.477195 }, //21 November 1990, UT1-UTC=-0.2931951, Accumulated Leap Seconds=25
  { 2448217.50, 57.479008 }, //22 November 1990, UT1-UTC=-0.2950083, Accumulated Leap Seconds=25
  { 2448218.50, 57.480984 }, //23 November 1990, UT1-UTC=-0.2969842, Accumulated Leap Seconds=25
  { 2448219.50, 57.483157 }, //24 November 1990, UT1-UTC=-0.2991571, Accumulated Leap Seconds=25
  { 2448220.50, 57.485542 }, //25 November 1990, UT1-UTC=-0.3015423, Accumulated Leap Seconds=25
  { 2448221.50, 57.488128 }, //26 November 1990, UT1-UTC=-0.3041284, Accumulated Leap Seconds=25
  { 2448222.50, 57.490869 }, //27 November 1990, UT1-UTC=-0.3068688, Accumulated Leap Seconds=25
  { 2448223.50, 57.493681 }, //28 November 1990, UT1-UTC=-0.3096809, Accumulated Leap Seconds=25
  { 2448224.50, 57.496465 }, //29 November 1990, UT1-UTC=-0.3124648, Accumulated Leap Seconds=25
  { 2448225.50, 57.499120 }, //30 November 1990, UT1-UTC=-0.3151201, Accumulated Leap Seconds=25
  { 2448226.50, 57.501575 }, //01 December 1990, UT1-UTC=-0.3175747, Accumulated Leap Seconds=25
  { 2448227.50, 57.503812 }, //02 December 1990, UT1-UTC=-0.3198120, Accumulated Leap Seconds=25
  { 2448228.50, 57.505891 }, //03 December 1990, UT1-UTC=-0.3218909, Accumulated Leap Seconds=25
  { 2448229.50, 57.507925 }, //04 December 1990, UT1-UTC=-0.3239249, Accumulated Leap Seconds=25
  { 2448230.50, 57.510032 }, //05 December 1990, UT1-UTC=-0.3260323, Accumulated Leap Seconds=25
  { 2448231.50, 57.512292 }, //06 December 1990, UT1-UTC=-0.3282918, Accumulated Leap Seconds=25
  { 2448232.50, 57.514716 }, //07 December 1990, UT1-UTC=-0.3307156, Accumulated Leap Seconds=25
  { 2448233.50, 57.517257 }, //08 December 1990, UT1-UTC=-0.3332567, Accumulated Leap Seconds=25
  { 2448234.50, 57.519832 }, //09 December 1990, UT1-UTC=-0.3358323, Accumulated Leap Seconds=25
  { 2448235.50, 57.522351 }, //10 December 1990, UT1-UTC=-0.3383512, Accumulated Leap Seconds=25
  { 2448236.50, 57.524721 }, //11 December 1990, UT1-UTC=-0.3407208, Accumulated Leap Seconds=25
  { 2448237.50, 57.526874 }, //12 December 1990, UT1-UTC=-0.3428735, Accumulated Leap Seconds=25
  { 2448238.50, 57.528792 }, //13 December 1990, UT1-UTC=-0.3447925, Accumulated Leap Seconds=25
  { 2448239.50, 57.530500 }, //14 December 1990, UT1-UTC=-0.3465003, Accumulated Leap Seconds=25
  { 2448240.50, 57.532045 }, //15 December 1990, UT1-UTC=-0.3480446, Accumulated Leap Seconds=25
  { 2448241.50, 57.533482 }, //16 December 1990, UT1-UTC=-0.3494824, Accumulated Leap Seconds=25
  { 2448242.50, 57.534874 }, //17 December 1990, UT1-UTC=-0.3508740, Accumulated Leap Seconds=25
  { 2448243.50, 57.536286 }, //18 December 1990, UT1-UTC=-0.3522861, Accumulated Leap Seconds=25
  { 2448244.50, 57.537788 }, //19 December 1990, UT1-UTC=-0.3537879, Accumulated Leap Seconds=25
  { 2448245.50, 57.539454 }, //20 December 1990, UT1-UTC=-0.3554536, Accumulated Leap Seconds=25
  { 2448246.50, 57.541321 }, //21 December 1990, UT1-UTC=-0.3573212, Accumulated Leap Seconds=25
  { 2448247.50, 57.543398 }, //22 December 1990, UT1-UTC=-0.3593976, Accumulated Leap Seconds=25
  { 2448248.50, 57.545664 }, //23 December 1990, UT1-UTC=-0.3616644, Accumulated Leap Seconds=25
  { 2448249.50, 57.548077 }, //24 December 1990, UT1-UTC=-0.3640766, Accumulated Leap Seconds=25
  { 2448250.50, 57.550569 }, //25 December 1990, UT1-UTC=-0.3665691, Accumulated Leap Seconds=25
  { 2448251.50, 57.553061 }, //26 December 1990, UT1-UTC=-0.3690611, Accumulated Leap Seconds=25
  { 2448252.50, 57.555466 }, //27 December 1990, UT1-UTC=-0.3714658, Accumulated Leap Seconds=25
  { 2448253.50, 57.557711 }, //28 December 1990, UT1-UTC=-0.3737111, Accumulated Leap Seconds=25
  { 2448254.50, 57.559759 }, //29 December 1990, UT1-UTC=-0.3757593, Accumulated Leap Seconds=25
  { 2448255.50, 57.561637 }, //30 December 1990, UT1-UTC=-0.3776373, Accumulated Leap Seconds=25
  { 2448256.50, 57.563443 }, //31 December 1990, UT1-UTC=-0.3794432, Accumulated Leap Seconds=25
  { 2448257.50, 57.565313 }, //01 January 1991, UT1-UTC= 0.6186873, Accumulated Leap Seconds=26
  { 2448258.50, 57.567365 }, //02 January 1991, UT1-UTC= 0.6166345, Accumulated Leap Seconds=26
  { 2448259.50, 57.569661 }, //03 January 1991, UT1-UTC= 0.6143393, Accumulated Leap Seconds=26
  { 2448260.50, 57.572180 }, //04 January 1991, UT1-UTC= 0.6118200, Accumulated Leap Seconds=26
  { 2448261.50, 57.574840 }, //05 January 1991, UT1-UTC= 0.6091597, Accumulated Leap Seconds=26
  { 2448262.50, 57.577532 }, //06 January 1991, UT1-UTC= 0.6064680, Accumulated Leap Seconds=26
  { 2448263.50, 57.580159 }, //07 January 1991, UT1-UTC= 0.6038412, Accumulated Leap Seconds=26
  { 2448264.50, 57.582651 }, //08 January 1991, UT1-UTC= 0.6013486, Accumulated Leap Seconds=26
  { 2448265.50, 57.584965 }, //09 January 1991, UT1-UTC= 0.5990349, Accumulated Leap Seconds=26
  { 2448266.50, 57.587098 }, //10 January 1991, UT1-UTC= 0.5969015, Accumulated Leap Seconds=26
  { 2448267.50, 57.589077 }, //11 January 1991, UT1-UTC= 0.5949227, Accumulated Leap Seconds=26
  { 2448268.50, 57.590941 }, //12 January 1991, UT1-UTC= 0.5930593, Accumulated Leap Seconds=26
  { 2448269.50, 57.592736 }, //13 January 1991, UT1-UTC= 0.5912641, Accumulated Leap Seconds=26
  { 2448270.50, 57.594515 }, //14 January 1991, UT1-UTC= 0.5894847, Accumulated Leap Seconds=26
  { 2448271.50, 57.596340 }, //15 January 1991, UT1-UTC= 0.5876603, Accumulated Leap Seconds=26
  { 2448272.50, 57.598265 }, //16 January 1991, UT1-UTC= 0.5857348, Accumulated Leap Seconds=26
  { 2448273.50, 57.600330 }, //17 January 1991, UT1-UTC= 0.5836705, Accumulated Leap Seconds=26
  { 2448274.50, 57.602541 }, //18 January 1991, UT1-UTC= 0.5814590, Accumulated Leap Seconds=26
  { 2448275.50, 57.604880 }, //19 January 1991, UT1-UTC= 0.5791200, Accumulated Leap Seconds=26
  { 2448276.50, 57.607313 }, //20 January 1991, UT1-UTC= 0.5766870, Accumulated Leap Seconds=26
  { 2448277.50, 57.609789 }, //21 January 1991, UT1-UTC= 0.5742108, Accumulated Leap Seconds=26
  { 2448278.50, 57.612242 }, //22 January 1991, UT1-UTC= 0.5717582, Accumulated Leap Seconds=26
  { 2448279.50, 57.614606 }, //23 January 1991, UT1-UTC= 0.5693941, Accumulated Leap Seconds=26
  { 2448280.50, 57.616829 }, //24 January 1991, UT1-UTC= 0.5671710, Accumulated Leap Seconds=26
  { 2448281.50, 57.618870 }, //25 January 1991, UT1-UTC= 0.5651295, Accumulated Leap Seconds=26
  { 2448282.50, 57.620742 }, //26 January 1991, UT1-UTC= 0.5632577, Accumulated Leap Seconds=26
  { 2448283.50, 57.622517 }, //27 January 1991, UT1-UTC= 0.5614832, Accumulated Leap Seconds=26
  { 2448284.50, 57.624309 }, //28 January 1991, UT1-UTC= 0.5596912, Accumulated Leap Seconds=26
  { 2448285.50, 57.626237 }, //29 January 1991, UT1-UTC= 0.5577626, Accumulated Leap Seconds=26
  { 2448286.50, 57.628384 }, //30 January 1991, UT1-UTC= 0.5556160, Accumulated Leap Seconds=26
  { 2448287.50, 57.630768 }, //31 January 1991, UT1-UTC= 0.5532320, Accumulated Leap Seconds=26
  { 2448288.50, 57.633340 }, //01 February 1991, UT1-UTC= 0.5506604, Accumulated Leap Seconds=26
  { 2448289.50, 57.636006 }, //02 February 1991, UT1-UTC= 0.5479937, Accumulated Leap Seconds=26
  { 2448290.50, 57.638662 }, //03 February 1991, UT1-UTC= 0.5453384, Accumulated Leap Seconds=26
  { 2448291.50, 57.641201 }, //04 February 1991, UT1-UTC= 0.5427988, Accumulated Leap Seconds=26
  { 2448292.50, 57.643559 }, //05 February 1991, UT1-UTC= 0.5404409, Accumulated Leap Seconds=26
  { 2448293.50, 57.645713 }, //06 February 1991, UT1-UTC= 0.5382868, Accumulated Leap Seconds=26
  { 2448294.50, 57.647687 }, //07 February 1991, UT1-UTC= 0.5363129, Accumulated Leap Seconds=26
  { 2448295.50, 57.649525 }, //08 February 1991, UT1-UTC= 0.5344753, Accumulated Leap Seconds=26
  { 2448296.50, 57.651281 }, //09 February 1991, UT1-UTC= 0.5327192, Accumulated Leap Seconds=26
  { 2448297.50, 57.653025 }, //10 February 1991, UT1-UTC= 0.5309754, Accumulated Leap Seconds=26
  { 2448298.50, 57.654828 }, //11 February 1991, UT1-UTC= 0.5291720, Accumulated Leap Seconds=26
  { 2448299.50, 57.656755 }, //12 February 1991, UT1-UTC= 0.5272449, Accumulated Leap Seconds=26
  { 2448300.50, 57.658852 }, //13 February 1991, UT1-UTC= 0.5251484, Accumulated Leap Seconds=26
  { 2448301.50, 57.661139 }, //14 February 1991, UT1-UTC= 0.5228605, Accumulated Leap Seconds=26
  { 2448302.50, 57.663620 }, //15 February 1991, UT1-UTC= 0.5203802, Accumulated Leap Seconds=26
  { 2448303.50, 57.666246 }, //16 February 1991, UT1-UTC= 0.5177536, Accumulated Leap Seconds=26
  { 2448304.50, 57.668937 }, //17 February 1991, UT1-UTC= 0.5150635, Accumulated Leap Seconds=26
  { 2448305.50, 57.671585 }, //18 February 1991, UT1-UTC= 0.5124151, Accumulated Leap Seconds=26
  { 2448306.50, 57.674105 }, //19 February 1991, UT1-UTC= 0.5098949, Accumulated Leap Seconds=26
  { 2448307.50, 57.676444 }, //20 February 1991, UT1-UTC= 0.5075563, Accumulated Leap Seconds=26
  { 2448308.50, 57.678589 }, //21 February 1991, UT1-UTC= 0.5054113, Accumulated Leap Seconds=26
  { 2448309.50, 57.680585 }, //22 February 1991, UT1-UTC= 0.5034147, Accumulated Leap Seconds=26
  { 2448310.50, 57.682524 }, //23 February 1991, UT1-UTC= 0.5014756, Accumulated Leap Seconds=26
  { 2448311.50, 57.684499 }, //24 February 1991, UT1-UTC= 0.4995014, Accumulated Leap Seconds=26
  { 2448312.50, 57.686605 }, //25 February 1991, UT1-UTC= 0.4973952, Accumulated Leap Seconds=26
  { 2448313.50, 57.688926 }, //26 February 1991, UT1-UTC= 0.4950742, Accumulated Leap Seconds=26
  { 2448314.50, 57.691503 }, //27 February 1991, UT1-UTC= 0.4924967, Accumulated Leap Seconds=26
  { 2448315.50, 57.694319 }, //28 February 1991, UT1-UTC= 0.4896806, Accumulated Leap Seconds=26
  { 2448316.50, 57.697284 }, //01 March 1991, UT1-UTC= 0.4867156, Accumulated Leap Seconds=26
  { 2448317.50, 57.700266 }, //02 March 1991, UT1-UTC= 0.4837336, Accumulated Leap Seconds=26
  { 2448318.50, 57.703131 }, //03 March 1991, UT1-UTC= 0.4808686, Accumulated Leap Seconds=26
  { 2448319.50, 57.705776 }, //04 March 1991, UT1-UTC= 0.4782244, Accumulated Leap Seconds=26
  { 2448320.50, 57.708145 }, //05 March 1991, UT1-UTC= 0.4758554, Accumulated Leap Seconds=26
  { 2448321.50, 57.710245 }, //06 March 1991, UT1-UTC= 0.4737546, Accumulated Leap Seconds=26
  { 2448322.50, 57.712120 }, //07 March 1991, UT1-UTC= 0.4718801, Accumulated Leap Seconds=26
  { 2448323.50, 57.713834 }, //08 March 1991, UT1-UTC= 0.4701663, Accumulated Leap Seconds=26
  { 2448324.50, 57.715478 }, //09 March 1991, UT1-UTC= 0.4685217, Accumulated Leap Seconds=26
  { 2448325.50, 57.717148 }, //10 March 1991, UT1-UTC= 0.4668521, Accumulated Leap Seconds=26
  { 2448326.50, 57.718919 }, //11 March 1991, UT1-UTC= 0.4650811, Accumulated Leap Seconds=26
  { 2448327.50, 57.720857 }, //12 March 1991, UT1-UTC= 0.4631426, Accumulated Leap Seconds=26
  { 2448328.50, 57.723010 }, //13 March 1991, UT1-UTC= 0.4609901, Accumulated Leap Seconds=26
  { 2448329.50, 57.725388 }, //14 March 1991, UT1-UTC= 0.4586115, Accumulated Leap Seconds=26
  { 2448330.50, 57.727977 }, //15 March 1991, UT1-UTC= 0.4560228, Accumulated Leap Seconds=26
  { 2448331.50, 57.730770 }, //16 March 1991, UT1-UTC= 0.4532295, Accumulated Leap Seconds=26
  { 2448332.50, 57.733680 }, //17 March 1991, UT1-UTC= 0.4503202, Accumulated Leap Seconds=26
  { 2448333.50, 57.736571 }, //18 March 1991, UT1-UTC= 0.4474291, Accumulated Leap Seconds=26
  { 2448334.50, 57.739316 }, //19 March 1991, UT1-UTC= 0.4446838, Accumulated Leap Seconds=26
  { 2448335.50, 57.741830 }, //20 March 1991, UT1-UTC= 0.4421696, Accumulated Leap Seconds=26
  { 2448336.50, 57.744107 }, //21 March 1991, UT1-UTC= 0.4398935, Accumulated Leap Seconds=26
  { 2448337.50, 57.746203 }, //22 March 1991, UT1-UTC= 0.4377969, Accumulated Leap Seconds=26
  { 2448338.50, 57.748213 }, //23 March 1991, UT1-UTC= 0.4357868, Accumulated Leap Seconds=26
  { 2448339.50, 57.750241 }, //24 March 1991, UT1-UTC= 0.4337585, Accumulated Leap Seconds=26
  { 2448340.50, 57.752381 }, //25 March 1991, UT1-UTC= 0.4316193, Accumulated Leap Seconds=26
  { 2448341.50, 57.754730 }, //26 March 1991, UT1-UTC= 0.4292702, Accumulated Leap Seconds=26
  { 2448342.50, 57.757338 }, //27 March 1991, UT1-UTC= 0.4266623, Accumulated Leap Seconds=26
  { 2448343.50, 57.760141 }, //28 March 1991, UT1-UTC= 0.4238586, Accumulated Leap Seconds=26
  { 2448344.50, 57.763026 }, //29 March 1991, UT1-UTC= 0.4209736, Accumulated Leap Seconds=26
  { 2448345.50, 57.765866 }, //30 March 1991, UT1-UTC= 0.4181342, Accumulated Leap Seconds=26
  { 2448346.50, 57.768564 }, //31 March 1991, UT1-UTC= 0.4154357, Accumulated Leap Seconds=26
  { 2448347.50, 57.771077 }, //01 April 1991, UT1-UTC= 0.4129226, Accumulated Leap Seconds=26
  { 2448348.50, 57.773400 }, //02 April 1991, UT1-UTC= 0.4106000, Accumulated Leap Seconds=26
  { 2448349.50, 57.775559 }, //03 April 1991, UT1-UTC= 0.4084408, Accumulated Leap Seconds=26
  { 2448350.50, 57.777594 }, //04 April 1991, UT1-UTC= 0.4064062, Accumulated Leap Seconds=26
  { 2448351.50, 57.779551 }, //05 April 1991, UT1-UTC= 0.4044489, Accumulated Leap Seconds=26
  { 2448352.50, 57.781484 }, //06 April 1991, UT1-UTC= 0.4025161, Accumulated Leap Seconds=26
  { 2448353.50, 57.783444 }, //07 April 1991, UT1-UTC= 0.4005560, Accumulated Leap Seconds=26
  { 2448354.50, 57.785477 }, //08 April 1991, UT1-UTC= 0.3985228, Accumulated Leap Seconds=26
  { 2448355.50, 57.787622 }, //09 April 1991, UT1-UTC= 0.3963778, Accumulated Leap Seconds=26
  { 2448356.50, 57.789914 }, //10 April 1991, UT1-UTC= 0.3940858, Accumulated Leap Seconds=26
  { 2448357.50, 57.792368 }, //11 April 1991, UT1-UTC= 0.3916317, Accumulated Leap Seconds=26
  { 2448358.50, 57.794974 }, //12 April 1991, UT1-UTC= 0.3890260, Accumulated Leap Seconds=26
  { 2448359.50, 57.797687 }, //13 April 1991, UT1-UTC= 0.3863127, Accumulated Leap Seconds=26
  { 2448360.50, 57.800429 }, //14 April 1991, UT1-UTC= 0.3835710, Accumulated Leap Seconds=26
  { 2448361.50, 57.803084 }, //15 April 1991, UT1-UTC= 0.3809164, Accumulated Leap Seconds=26
  { 2448362.50, 57.805549 }, //16 April 1991, UT1-UTC= 0.3784514, Accumulated Leap Seconds=26
  { 2448363.50, 57.807775 }, //17 April 1991, UT1-UTC= 0.3762250, Accumulated Leap Seconds=26
  { 2448364.50, 57.809785 }, //18 April 1991, UT1-UTC= 0.3742153, Accumulated Leap Seconds=26
  { 2448365.50, 57.811668 }, //19 April 1991, UT1-UTC= 0.3723315, Accumulated Leap Seconds=26
  { 2448366.50, 57.813598 }, //20 April 1991, UT1-UTC= 0.3704025, Accumulated Leap Seconds=26
  { 2448367.50, 57.815691 }, //21 April 1991, UT1-UTC= 0.3683085, Accumulated Leap Seconds=26
  { 2448368.50, 57.818009 }, //22 April 1991, UT1-UTC= 0.3659907, Accumulated Leap Seconds=26
  { 2448369.50, 57.820553 }, //23 April 1991, UT1-UTC= 0.3634465, Accumulated Leap Seconds=26
  { 2448370.50, 57.823279 }, //24 April 1991, UT1-UTC= 0.3607214, Accumulated Leap Seconds=26
  { 2448371.50, 57.826101 }, //25 April 1991, UT1-UTC= 0.3578994, Accumulated Leap Seconds=26
  { 2448372.50, 57.828919 }, //26 April 1991, UT1-UTC= 0.3550805, Accumulated Leap Seconds=26
  { 2448373.50, 57.831643 }, //27 April 1991, UT1-UTC= 0.3523574, Accumulated Leap Seconds=26
  { 2448374.50, 57.834203 }, //28 April 1991, UT1-UTC= 0.3497972, Accumulated Leap Seconds=26
  { 2448375.50, 57.836567 }, //29 April 1991, UT1-UTC= 0.3474332, Accumulated Leap Seconds=26
  { 2448376.50, 57.838736 }, //30 April 1991, UT1-UTC= 0.3452638, Accumulated Leap Seconds=26
  { 2448377.50, 57.840743 }, //01 May 1991, UT1-UTC= 0.3432573, Accumulated Leap Seconds=26
  { 2448378.50, 57.842642 }, //02 May 1991, UT1-UTC= 0.3413582, Accumulated Leap Seconds=26
  { 2448379.50, 57.844499 }, //03 May 1991, UT1-UTC= 0.3395013, Accumulated Leap Seconds=26
  { 2448380.50, 57.846380 }, //04 May 1991, UT1-UTC= 0.3376197, Accumulated Leap Seconds=26
  { 2448381.50, 57.848353 }, //05 May 1991, UT1-UTC= 0.3356474, Accumulated Leap Seconds=26
  { 2448382.50, 57.850471 }, //06 May 1991, UT1-UTC= 0.3335288, Accumulated Leap Seconds=26
  { 2448383.50, 57.852776 }, //07 May 1991, UT1-UTC= 0.3312242, Accumulated Leap Seconds=26
  { 2448384.50, 57.855280 }, //08 May 1991, UT1-UTC= 0.3287201, Accumulated Leap Seconds=26
  { 2448385.50, 57.857961 }, //09 May 1991, UT1-UTC= 0.3260393, Accumulated Leap Seconds=26
  { 2448386.50, 57.860768 }, //10 May 1991, UT1-UTC= 0.3232324, Accumulated Leap Seconds=26
  { 2448387.50, 57.863621 }, //11 May 1991, UT1-UTC= 0.3203791, Accumulated Leap Seconds=26
  { 2448388.50, 57.866423 }, //12 May 1991, UT1-UTC= 0.3175768, Accumulated Leap Seconds=26
  { 2448389.50, 57.869081 }, //13 May 1991, UT1-UTC= 0.3149187, Accumulated Leap Seconds=26
  { 2448390.50, 57.871535 }, //14 May 1991, UT1-UTC= 0.3124646, Accumulated Leap Seconds=26
  { 2448391.50, 57.873784 }, //15 May 1991, UT1-UTC= 0.3102155, Accumulated Leap Seconds=26
  { 2448392.50, 57.875878 }, //16 May 1991, UT1-UTC= 0.3081224, Accumulated Leap Seconds=26
  { 2448393.50, 57.877907 }, //17 May 1991, UT1-UTC= 0.3060929, Accumulated Leap Seconds=26
  { 2448394.50, 57.879971 }, //18 May 1991, UT1-UTC= 0.3040287, Accumulated Leap Seconds=26
  { 2448395.50, 57.882145 }, //19 May 1991, UT1-UTC= 0.3018546, Accumulated Leap Seconds=26
  { 2448396.50, 57.884447 }, //20 May 1991, UT1-UTC= 0.2995529, Accumulated Leap Seconds=26
  { 2448397.50, 57.886835 }, //21 May 1991, UT1-UTC= 0.2971647, Accumulated Leap Seconds=26
  { 2448398.50, 57.889228 }, //22 May 1991, UT1-UTC= 0.2947720, Accumulated Leap Seconds=26
  { 2448399.50, 57.891524 }, //23 May 1991, UT1-UTC= 0.2924757, Accumulated Leap Seconds=26
  { 2448400.50, 57.893656 }, //24 May 1991, UT1-UTC= 0.2903437, Accumulated Leap Seconds=26
  { 2448401.50, 57.895587 }, //25 May 1991, UT1-UTC= 0.2884128, Accumulated Leap Seconds=26
  { 2448402.50, 57.897311 }, //26 May 1991, UT1-UTC= 0.2866890, Accumulated Leap Seconds=26
  { 2448403.50, 57.898852 }, //27 May 1991, UT1-UTC= 0.2851481, Accumulated Leap Seconds=26
  { 2448404.50, 57.900262 }, //28 May 1991, UT1-UTC= 0.2837384, Accumulated Leap Seconds=26
  { 2448405.50, 57.901604 }, //29 May 1991, UT1-UTC= 0.2823960, Accumulated Leap Seconds=26
  { 2448406.50, 57.902934 }, //30 May 1991, UT1-UTC= 0.2810657, Accumulated Leap Seconds=26
  { 2448407.50, 57.904309 }, //31 May 1991, UT1-UTC= 0.2796910, Accumulated Leap Seconds=26
  { 2448408.50, 57.905780 }, //01 June 1991, UT1-UTC= 0.2782199, Accumulated Leap Seconds=26
  { 2448409.50, 57.907393 }, //02 June 1991, UT1-UTC= 0.2766068, Accumulated Leap Seconds=26
  { 2448410.50, 57.909176 }, //03 June 1991, UT1-UTC= 0.2748243, Accumulated Leap Seconds=26
  { 2448411.50, 57.911134 }, //04 June 1991, UT1-UTC= 0.2728657, Accumulated Leap Seconds=26
  { 2448412.50, 57.913264 }, //05 June 1991, UT1-UTC= 0.2707358, Accumulated Leap Seconds=26
  { 2448413.50, 57.915545 }, //06 June 1991, UT1-UTC= 0.2684552, Accumulated Leap Seconds=26
  { 2448414.50, 57.917929 }, //07 June 1991, UT1-UTC= 0.2660706, Accumulated Leap Seconds=26
  { 2448415.50, 57.920347 }, //08 June 1991, UT1-UTC= 0.2636532, Accumulated Leap Seconds=26
  { 2448416.50, 57.922706 }, //09 June 1991, UT1-UTC= 0.2612938, Accumulated Leap Seconds=26
  { 2448417.50, 57.924931 }, //10 June 1991, UT1-UTC= 0.2590690, Accumulated Leap Seconds=26
  { 2448418.50, 57.926991 }, //11 June 1991, UT1-UTC= 0.2570088, Accumulated Leap Seconds=26
  { 2448419.50, 57.928909 }, //12 June 1991, UT1-UTC= 0.2550909, Accumulated Leap Seconds=26
  { 2448420.50, 57.930736 }, //13 June 1991, UT1-UTC= 0.2532644, Accumulated Leap Seconds=26
  { 2448421.50, 57.932577 }, //14 June 1991, UT1-UTC= 0.2514228, Accumulated Leap Seconds=26
  { 2448422.50, 57.934529 }, //15 June 1991, UT1-UTC= 0.2494713, Accumulated Leap Seconds=26
  { 2448423.50, 57.936629 }, //16 June 1991, UT1-UTC= 0.2473708, Accumulated Leap Seconds=26
  { 2448424.50, 57.938845 }, //17 June 1991, UT1-UTC= 0.2451548, Accumulated Leap Seconds=26
  { 2448425.50, 57.941088 }, //18 June 1991, UT1-UTC= 0.2429118, Accumulated Leap Seconds=26
  { 2448426.50, 57.943255 }, //19 June 1991, UT1-UTC= 0.2407451, Accumulated Leap Seconds=26
  { 2448427.50, 57.945259 }, //20 June 1991, UT1-UTC= 0.2387406, Accumulated Leap Seconds=26
  { 2448428.50, 57.947048 }, //21 June 1991, UT1-UTC= 0.2369518, Accumulated Leap Seconds=26
  { 2448429.50, 57.948601 }, //22 June 1991, UT1-UTC= 0.2353985, Accumulated Leap Seconds=26
  { 2448430.50, 57.949921 }, //23 June 1991, UT1-UTC= 0.2340790, Accumulated Leap Seconds=26
  { 2448431.50, 57.951029 }, //24 June 1991, UT1-UTC= 0.2329710, Accumulated Leap Seconds=26
  { 2448432.50, 57.951967 }, //25 June 1991, UT1-UTC= 0.2320333, Accumulated Leap Seconds=26
  { 2448433.50, 57.952792 }, //26 June 1991, UT1-UTC= 0.2312077, Accumulated Leap Seconds=26
  { 2448434.50, 57.953577 }, //27 June 1991, UT1-UTC= 0.2304227, Accumulated Leap Seconds=26
  { 2448435.50, 57.954397 }, //28 June 1991, UT1-UTC= 0.2296034, Accumulated Leap Seconds=26
  { 2448436.50, 57.955315 }, //29 June 1991, UT1-UTC= 0.2286853, Accumulated Leap Seconds=26
  { 2448437.50, 57.956369 }, //30 June 1991, UT1-UTC= 0.2276310, Accumulated Leap Seconds=26
  { 2448438.50, 57.957566 }, //01 July 1991, UT1-UTC= 0.2264337, Accumulated Leap Seconds=26
  { 2448439.50, 57.958894 }, //02 July 1991, UT1-UTC= 0.2251061, Accumulated Leap Seconds=26
  { 2448440.50, 57.960341 }, //03 July 1991, UT1-UTC= 0.2236593, Accumulated Leap Seconds=26
  { 2448441.50, 57.961876 }, //04 July 1991, UT1-UTC= 0.2221241, Accumulated Leap Seconds=26
  { 2448442.50, 57.963428 }, //05 July 1991, UT1-UTC= 0.2205722, Accumulated Leap Seconds=26
  { 2448443.50, 57.964911 }, //06 July 1991, UT1-UTC= 0.2190892, Accumulated Leap Seconds=26
  { 2448444.50, 57.966247 }, //07 July 1991, UT1-UTC= 0.2177529, Accumulated Leap Seconds=26
  { 2448445.50, 57.967388 }, //08 July 1991, UT1-UTC= 0.2166115, Accumulated Leap Seconds=26
  { 2448446.50, 57.968340 }, //09 July 1991, UT1-UTC= 0.2156604, Accumulated Leap Seconds=26
  { 2448447.50, 57.969173 }, //10 July 1991, UT1-UTC= 0.2148270, Accumulated Leap Seconds=26
  { 2448448.50, 57.970031 }, //11 July 1991, UT1-UTC= 0.2139694, Accumulated Leap Seconds=26
  { 2448449.50, 57.971066 }, //12 July 1991, UT1-UTC= 0.2129336, Accumulated Leap Seconds=26
  { 2448450.50, 57.972376 }, //13 July 1991, UT1-UTC= 0.2116242, Accumulated Leap Seconds=26
  { 2448451.50, 57.973970 }, //14 July 1991, UT1-UTC= 0.2100297, Accumulated Leap Seconds=26
  { 2448452.50, 57.975774 }, //15 July 1991, UT1-UTC= 0.2082262, Accumulated Leap Seconds=26
  { 2448453.50, 57.977647 }, //16 July 1991, UT1-UTC= 0.2063530, Accumulated Leap Seconds=26
  { 2448454.50, 57.979454 }, //17 July 1991, UT1-UTC= 0.2045463, Accumulated Leap Seconds=26
  { 2448455.50, 57.981110 }, //18 July 1991, UT1-UTC= 0.2028900, Accumulated Leap Seconds=26
  { 2448456.50, 57.982568 }, //19 July 1991, UT1-UTC= 0.2014318, Accumulated Leap Seconds=26
  { 2448457.50, 57.983804 }, //20 July 1991, UT1-UTC= 0.2001958, Accumulated Leap Seconds=26
  { 2448458.50, 57.984840 }, //21 July 1991, UT1-UTC= 0.1991599, Accumulated Leap Seconds=26
  { 2448459.50, 57.985732 }, //22 July 1991, UT1-UTC= 0.1982680, Accumulated Leap Seconds=26
  { 2448460.50, 57.986547 }, //23 July 1991, UT1-UTC= 0.1974532, Accumulated Leap Seconds=26
  { 2448461.50, 57.987349 }, //24 July 1991, UT1-UTC= 0.1966510, Accumulated Leap Seconds=26
  { 2448462.50, 57.988200 }, //25 July 1991, UT1-UTC= 0.1958000, Accumulated Leap Seconds=26
  { 2448463.50, 57.989160 }, //26 July 1991, UT1-UTC= 0.1948400, Accumulated Leap Seconds=26
  { 2448464.50, 57.990273 }, //27 July 1991, UT1-UTC= 0.1937267, Accumulated Leap Seconds=26
  { 2448465.50, 57.991555 }, //28 July 1991, UT1-UTC= 0.1924452, Accumulated Leap Seconds=26
  { 2448466.50, 57.992986 }, //29 July 1991, UT1-UTC= 0.1910139, Accumulated Leap Seconds=26
  { 2448467.50, 57.994507 }, //30 July 1991, UT1-UTC= 0.1894931, Accumulated Leap Seconds=26
  { 2448468.50, 57.996029 }, //31 July 1991, UT1-UTC= 0.1879714, Accumulated Leap Seconds=26
  { 2448469.50, 57.997493 }, //01 August 1991, UT1-UTC= 0.1865071, Accumulated Leap Seconds=26
  { 2448470.50, 57.998867 }, //02 August 1991, UT1-UTC= 0.1851333, Accumulated Leap Seconds=26
  { 2448471.50, 58.000122 }, //03 August 1991, UT1-UTC= 0.1838784, Accumulated Leap Seconds=26
  { 2448472.50, 58.001231 }, //04 August 1991, UT1-UTC= 0.1827692, Accumulated Leap Seconds=26
  { 2448473.50, 58.002196 }, //05 August 1991, UT1-UTC= 0.1818041, Accumulated Leap Seconds=26
  { 2448474.50, 58.003057 }, //06 August 1991, UT1-UTC= 0.1809429, Accumulated Leap Seconds=26
  { 2448475.50, 58.003900 }, //07 August 1991, UT1-UTC= 0.1800998, Accumulated Leap Seconds=26
  { 2448476.50, 58.004841 }, //08 August 1991, UT1-UTC= 0.1791588, Accumulated Leap Seconds=26
  { 2448477.50, 58.006000 }, //09 August 1991, UT1-UTC= 0.1779999, Accumulated Leap Seconds=26
  { 2448478.50, 58.007442 }, //10 August 1991, UT1-UTC= 0.1765578, Accumulated Leap Seconds=26
  { 2448479.50, 58.009148 }, //11 August 1991, UT1-UTC= 0.1748518, Accumulated Leap Seconds=26
  { 2448480.50, 58.011028 }, //12 August 1991, UT1-UTC= 0.1729715, Accumulated Leap Seconds=26
  { 2448481.50, 58.012965 }, //13 August 1991, UT1-UTC= 0.1710353, Accumulated Leap Seconds=26
  { 2448482.50, 58.014846 }, //14 August 1991, UT1-UTC= 0.1691544, Accumulated Leap Seconds=26
  { 2448483.50, 58.016579 }, //15 August 1991, UT1-UTC= 0.1674209, Accumulated Leap Seconds=26
  { 2448484.50, 58.018121 }, //16 August 1991, UT1-UTC= 0.1658790, Accumulated Leap Seconds=26
  { 2448485.50, 58.019472 }, //17 August 1991, UT1-UTC= 0.1645283, Accumulated Leap Seconds=26
  { 2448486.50, 58.020665 }, //18 August 1991, UT1-UTC= 0.1633349, Accumulated Leap Seconds=26
  { 2448487.50, 58.021756 }, //19 August 1991, UT1-UTC= 0.1622436, Accumulated Leap Seconds=26
  { 2448488.50, 58.022804 }, //20 August 1991, UT1-UTC= 0.1611961, Accumulated Leap Seconds=26
  { 2448489.50, 58.023865 }, //21 August 1991, UT1-UTC= 0.1601349, Accumulated Leap Seconds=26
  { 2448490.50, 58.024995 }, //22 August 1991, UT1-UTC= 0.1590054, Accumulated Leap Seconds=26
  { 2448491.50, 58.026246 }, //23 August 1991, UT1-UTC= 0.1577537, Accumulated Leap Seconds=26
  { 2448492.50, 58.027680 }, //24 August 1991, UT1-UTC= 0.1563201, Accumulated Leap Seconds=26
  { 2448493.50, 58.029312 }, //25 August 1991, UT1-UTC= 0.1546877, Accumulated Leap Seconds=26
  { 2448494.50, 58.031126 }, //26 August 1991, UT1-UTC= 0.1528737, Accumulated Leap Seconds=26
  { 2448495.50, 58.033078 }, //27 August 1991, UT1-UTC= 0.1509224, Accumulated Leap Seconds=26
  { 2448496.50, 58.035108 }, //28 August 1991, UT1-UTC= 0.1488923, Accumulated Leap Seconds=26
  { 2448497.50, 58.037131 }, //29 August 1991, UT1-UTC= 0.1468689, Accumulated Leap Seconds=26
  { 2448498.50, 58.039069 }, //30 August 1991, UT1-UTC= 0.1449313, Accumulated Leap Seconds=26
  { 2448499.50, 58.040876 }, //31 August 1991, UT1-UTC= 0.1431235, Accumulated Leap Seconds=26
  { 2448500.50, 58.042552 }, //01 September 1991, UT1-UTC= 0.1414483, Accumulated Leap Seconds=26
  { 2448501.50, 58.044136 }, //02 September 1991, UT1-UTC= 0.1398642, Accumulated Leap Seconds=26
  { 2448502.50, 58.045709 }, //03 September 1991, UT1-UTC= 0.1382909, Accumulated Leap Seconds=26
  { 2448503.50, 58.047376 }, //04 September 1991, UT1-UTC= 0.1366244, Accumulated Leap Seconds=26
  { 2448504.50, 58.049240 }, //05 September 1991, UT1-UTC= 0.1347595, Accumulated Leap Seconds=26
  { 2448505.50, 58.051369 }, //06 September 1991, UT1-UTC= 0.1326306, Accumulated Leap Seconds=26
  { 2448506.50, 58.053761 }, //07 September 1991, UT1-UTC= 0.1302394, Accumulated Leap Seconds=26
  { 2448507.50, 58.056341 }, //08 September 1991, UT1-UTC= 0.1276590, Accumulated Leap Seconds=26
  { 2448508.50, 58.059002 }, //09 September 1991, UT1-UTC= 0.1249975, Accumulated Leap Seconds=26
  { 2448509.50, 58.061628 }, //10 September 1991, UT1-UTC= 0.1223721, Accumulated Leap Seconds=26
  { 2448510.50, 58.064120 }, //11 September 1991, UT1-UTC= 0.1198796, Accumulated Leap Seconds=26
  { 2448511.50, 58.066411 }, //12 September 1991, UT1-UTC= 0.1175890, Accumulated Leap Seconds=26
  { 2448512.50, 58.068478 }, //13 September 1991, UT1-UTC= 0.1155225, Accumulated Leap Seconds=26
  { 2448513.50, 58.070338 }, //14 September 1991, UT1-UTC= 0.1136621, Accumulated Leap Seconds=26
  { 2448514.50, 58.072038 }, //15 September 1991, UT1-UTC= 0.1119620, Accumulated Leap Seconds=26
  { 2448515.50, 58.073635 }, //16 September 1991, UT1-UTC= 0.1103647, Accumulated Leap Seconds=26
  { 2448516.50, 58.075184 }, //17 September 1991, UT1-UTC= 0.1088160, Accumulated Leap Seconds=26
  { 2448517.50, 58.076741 }, //18 September 1991, UT1-UTC= 0.1072593, Accumulated Leap Seconds=26
  { 2448518.50, 58.078378 }, //19 September 1991, UT1-UTC= 0.1056220, Accumulated Leap Seconds=26
  { 2448519.50, 58.080157 }, //20 September 1991, UT1-UTC= 0.1038429, Accumulated Leap Seconds=26
  { 2448520.50, 58.082116 }, //21 September 1991, UT1-UTC= 0.1018845, Accumulated Leap Seconds=26
  { 2448521.50, 58.084263 }, //22 September 1991, UT1-UTC= 0.0997366, Accumulated Leap Seconds=26
  { 2448522.50, 58.086584 }, //23 September 1991, UT1-UTC= 0.0974157, Accumulated Leap Seconds=26
  { 2448523.50, 58.089037 }, //24 September 1991, UT1-UTC= 0.0949625, Accumulated Leap Seconds=26
  { 2448524.50, 58.091551 }, //25 September 1991, UT1-UTC= 0.0924487, Accumulated Leap Seconds=26
  { 2448525.50, 58.094022 }, //26 September 1991, UT1-UTC= 0.0899784, Accumulated Leap Seconds=26
  { 2448526.50, 58.096366 }, //27 September 1991, UT1-UTC= 0.0876344, Accumulated Leap Seconds=26
  { 2448527.50, 58.098535 }, //28 September 1991, UT1-UTC= 0.0854645, Accumulated Leap Seconds=26
  { 2448528.50, 58.100544 }, //29 September 1991, UT1-UTC= 0.0834557, Accumulated Leap Seconds=26
  { 2448529.50, 58.102450 }, //30 September 1991, UT1-UTC= 0.0815502, Accumulated Leap Seconds=26
  { 2448530.50, 58.104332 }, //01 October 1991, UT1-UTC= 0.0796681, Accumulated Leap Seconds=26
  { 2448531.50, 58.106279 }, //02 October 1991, UT1-UTC= 0.0777207, Accumulated Leap Seconds=26
  { 2448532.50, 58.108381 }, //03 October 1991, UT1-UTC= 0.0756190, Accumulated Leap Seconds=26
  { 2448533.50, 58.110677 }, //04 October 1991, UT1-UTC= 0.0733229, Accumulated Leap Seconds=26
  { 2448534.50, 58.113145 }, //05 October 1991, UT1-UTC= 0.0708552, Accumulated Leap Seconds=26
  { 2448535.50, 58.115709 }, //06 October 1991, UT1-UTC= 0.0682907, Accumulated Leap Seconds=26
  { 2448536.50, 58.118270 }, //07 October 1991, UT1-UTC= 0.0657297, Accumulated Leap Seconds=26
  { 2448537.50, 58.120733 }, //08 October 1991, UT1-UTC= 0.0632667, Accumulated Leap Seconds=26
  { 2448538.50, 58.123033 }, //09 October 1991, UT1-UTC= 0.0609668, Accumulated Leap Seconds=26
  { 2448539.50, 58.125128 }, //10 October 1991, UT1-UTC= 0.0588719, Accumulated Leap Seconds=26
  { 2448540.50, 58.127012 }, //11 October 1991, UT1-UTC= 0.0569879, Accumulated Leap Seconds=26
  { 2448541.50, 58.128712 }, //12 October 1991, UT1-UTC= 0.0552882, Accumulated Leap Seconds=26
  { 2448542.50, 58.130277 }, //13 October 1991, UT1-UTC= 0.0537235, Accumulated Leap Seconds=26
  { 2448543.50, 58.131760 }, //14 October 1991, UT1-UTC= 0.0522401, Accumulated Leap Seconds=26
  { 2448544.50, 58.133210 }, //15 October 1991, UT1-UTC= 0.0507904, Accumulated Leap Seconds=26
  { 2448545.50, 58.134678 }, //16 October 1991, UT1-UTC= 0.0493224, Accumulated Leap Seconds=26
  { 2448546.50, 58.136241 }, //17 October 1991, UT1-UTC= 0.0477593, Accumulated Leap Seconds=26
  { 2448547.50, 58.137961 }, //18 October 1991, UT1-UTC= 0.0460390, Accumulated Leap Seconds=26
  { 2448548.50, 58.139861 }, //19 October 1991, UT1-UTC= 0.0441385, Accumulated Leap Seconds=26
  { 2448549.50, 58.141936 }, //20 October 1991, UT1-UTC= 0.0420641, Accumulated Leap Seconds=26
  { 2448550.50, 58.144146 }, //21 October 1991, UT1-UTC= 0.0398538, Accumulated Leap Seconds=26
  { 2448551.50, 58.146428 }, //22 October 1991, UT1-UTC= 0.0375715, Accumulated Leap Seconds=26
  { 2448552.50, 58.148702 }, //23 October 1991, UT1-UTC= 0.0352978, Accumulated Leap Seconds=26
  { 2448553.50, 58.150898 }, //24 October 1991, UT1-UTC= 0.0331024, Accumulated Leap Seconds=26
  { 2448554.50, 58.152955 }, //25 October 1991, UT1-UTC= 0.0310446, Accumulated Leap Seconds=26
  { 2448555.50, 58.154882 }, //26 October 1991, UT1-UTC= 0.0291177, Accumulated Leap Seconds=26
  { 2448556.50, 58.156743 }, //27 October 1991, UT1-UTC= 0.0272569, Accumulated Leap Seconds=26
  { 2448557.50, 58.158636 }, //28 October 1991, UT1-UTC= 0.0253641, Accumulated Leap Seconds=26
  { 2448558.50, 58.160657 }, //29 October 1991, UT1-UTC= 0.0233431, Accumulated Leap Seconds=26
  { 2448559.50, 58.162871 }, //30 October 1991, UT1-UTC= 0.0211288, Accumulated Leap Seconds=26
  { 2448560.50, 58.165294 }, //31 October 1991, UT1-UTC= 0.0187057, Accumulated Leap Seconds=26
  { 2448561.50, 58.167913 }, //01 November 1991, UT1-UTC= 0.0160872, Accumulated Leap Seconds=26
  { 2448562.50, 58.170670 }, //02 November 1991, UT1-UTC= 0.0133297, Accumulated Leap Seconds=26
  { 2448563.50, 58.173467 }, //03 November 1991, UT1-UTC= 0.0105333, Accumulated Leap Seconds=26
  { 2448564.50, 58.176187 }, //04 November 1991, UT1-UTC= 0.0078135, Accumulated Leap Seconds=26
  { 2448565.50, 58.178732 }, //05 November 1991, UT1-UTC= 0.0052679, Accumulated Leap Seconds=26
  { 2448566.50, 58.181048 }, //06 November 1991, UT1-UTC= 0.0029516, Accumulated Leap Seconds=26
  { 2448567.50, 58.183146 }, //07 November 1991, UT1-UTC= 0.0008540, Accumulated Leap Seconds=26
  { 2448568.50, 58.185075 }, //08 November 1991, UT1-UTC=-0.0010755, Accumulated Leap Seconds=26
  { 2448569.50, 58.186908 }, //09 November 1991, UT1-UTC=-0.0029076, Accumulated Leap Seconds=26
  { 2448570.50, 58.188710 }, //10 November 1991, UT1-UTC=-0.0047103, Accumulated Leap Seconds=26
  { 2448571.50, 58.190544 }, //11 November 1991, UT1-UTC=-0.0065439, Accumulated Leap Seconds=26
  { 2448572.50, 58.192458 }, //12 November 1991, UT1-UTC=-0.0084584, Accumulated Leap Seconds=26
  { 2448573.50, 58.194491 }, //13 November 1991, UT1-UTC=-0.0104907, Accumulated Leap Seconds=26
  { 2448574.50, 58.196662 }, //14 November 1991, UT1-UTC=-0.0126622, Accumulated Leap Seconds=26
  { 2448575.50, 58.198985 }, //15 November 1991, UT1-UTC=-0.0149851, Accumulated Leap Seconds=26
  { 2448576.50, 58.201453 }, //16 November 1991, UT1-UTC=-0.0174530, Accumulated Leap Seconds=26
  { 2448577.50, 58.204037 }, //17 November 1991, UT1-UTC=-0.0200371, Accumulated Leap Seconds=26
  { 2448578.50, 58.206675 }, //18 November 1991, UT1-UTC=-0.0226747, Accumulated Leap Seconds=26
  { 2448579.50, 58.209285 }, //19 November 1991, UT1-UTC=-0.0252847, Accumulated Leap Seconds=26
  { 2448580.50, 58.211788 }, //20 November 1991, UT1-UTC=-0.0277885, Accumulated Leap Seconds=26
  { 2448581.50, 58.214135 }, //21 November 1991, UT1-UTC=-0.0301355, Accumulated Leap Seconds=26
  { 2448582.50, 58.216300 }, //22 November 1991, UT1-UTC=-0.0323002, Accumulated Leap Seconds=26
  { 2448583.50, 58.218330 }, //23 November 1991, UT1-UTC=-0.0343297, Accumulated Leap Seconds=26
  { 2448584.50, 58.220335 }, //24 November 1991, UT1-UTC=-0.0363352, Accumulated Leap Seconds=26
  { 2448585.50, 58.222449 }, //25 November 1991, UT1-UTC=-0.0384490, Accumulated Leap Seconds=26
  { 2448586.50, 58.224779 }, //26 November 1991, UT1-UTC=-0.0407792, Accumulated Leap Seconds=26
  { 2448587.50, 58.227372 }, //27 November 1991, UT1-UTC=-0.0433718, Accumulated Leap Seconds=26
  { 2448588.50, 58.230181 }, //28 November 1991, UT1-UTC=-0.0461809, Accumulated Leap Seconds=26
  { 2448589.50, 58.233110 }, //29 November 1991, UT1-UTC=-0.0491102, Accumulated Leap Seconds=26
  { 2448590.50, 58.236053 }, //30 November 1991, UT1-UTC=-0.0520528, Accumulated Leap Seconds=26
  { 2448591.50, 58.238909 }, //01 December 1991, UT1-UTC=-0.0549092, Accumulated Leap Seconds=26
  { 2448592.50, 58.241604 }, //02 December 1991, UT1-UTC=-0.0576043, Accumulated Leap Seconds=26
  { 2448593.50, 58.244096 }, //03 December 1991, UT1-UTC=-0.0600957, Accumulated Leap Seconds=26
  { 2448594.50, 58.246382 }, //04 December 1991, UT1-UTC=-0.0623817, Accumulated Leap Seconds=26
  { 2448595.50, 58.248489 }, //05 December 1991, UT1-UTC=-0.0644891, Accumulated Leap Seconds=26
  { 2448596.50, 58.250470 }, //06 December 1991, UT1-UTC=-0.0664703, Accumulated Leap Seconds=26
  { 2448597.50, 58.252383 }, //07 December 1991, UT1-UTC=-0.0683827, Accumulated Leap Seconds=26
  { 2448598.50, 58.254290 }, //08 December 1991, UT1-UTC=-0.0702903, Accumulated Leap Seconds=26
  { 2448599.50, 58.256252 }, //09 December 1991, UT1-UTC=-0.0722516, Accumulated Leap Seconds=26
  { 2448600.50, 58.258313 }, //10 December 1991, UT1-UTC=-0.0743130, Accumulated Leap Seconds=26
  { 2448601.50, 58.260505 }, //11 December 1991, UT1-UTC=-0.0765048, Accumulated Leap Seconds=26
  { 2448602.50, 58.262835 }, //12 December 1991, UT1-UTC=-0.0788346, Accumulated Leap Seconds=26
  { 2448603.50, 58.265287 }, //13 December 1991, UT1-UTC=-0.0812872, Accumulated Leap Seconds=26
  { 2448604.50, 58.267833 }, //14 December 1991, UT1-UTC=-0.0838333, Accumulated Leap Seconds=26
  { 2448605.50, 58.270432 }, //15 December 1991, UT1-UTC=-0.0864317, Accumulated Leap Seconds=26
  { 2448606.50, 58.273032 }, //16 December 1991, UT1-UTC=-0.0890322, Accumulated Leap Seconds=26
  { 2448607.50, 58.275570 }, //17 December 1991, UT1-UTC=-0.0915699, Accumulated Leap Seconds=26
  { 2448608.50, 58.277984 }, //18 December 1991, UT1-UTC=-0.0939843, Accumulated Leap Seconds=26
  { 2448609.50, 58.280224 }, //19 December 1991, UT1-UTC=-0.0962239, Accumulated Leap Seconds=26
  { 2448610.50, 58.282266 }, //20 December 1991, UT1-UTC=-0.0982662, Accumulated Leap Seconds=26
  { 2448611.50, 58.284179 }, //21 December 1991, UT1-UTC=-0.1001794, Accumulated Leap Seconds=26
  { 2448612.50, 58.286094 }, //22 December 1991, UT1-UTC=-0.1020941, Accumulated Leap Seconds=26
  { 2448613.50, 58.288129 }, //23 December 1991, UT1-UTC=-0.1041286, Accumulated Leap Seconds=26
  { 2448614.50, 58.290354 }, //24 December 1991, UT1-UTC=-0.1063544, Accumulated Leap Seconds=26
  { 2448615.50, 58.292780 }, //25 December 1991, UT1-UTC=-0.1087801, Accumulated Leap Seconds=26
  { 2448616.50, 58.295350 }, //26 December 1991, UT1-UTC=-0.1113499, Accumulated Leap Seconds=26
  { 2448617.50, 58.297968 }, //27 December 1991, UT1-UTC=-0.1139684, Accumulated Leap Seconds=26
  { 2448618.50, 58.300535 }, //28 December 1991, UT1-UTC=-0.1165351, Accumulated Leap Seconds=26
  { 2448619.50, 58.302964 }, //29 December 1991, UT1-UTC=-0.1189640, Accumulated Leap Seconds=26
  { 2448620.50, 58.305210 }, //30 December 1991, UT1-UTC=-0.1212098, Accumulated Leap Seconds=26
  { 2448621.50, 58.307268 }, //31 December 1991, UT1-UTC=-0.1232681, Accumulated Leap Seconds=26
  { 2448622.50, 58.309166 }, //01 January 1992, UT1-UTC=-0.1251659, Accumulated Leap Seconds=26
  { 2448623.50, 58.310955 }, //02 January 1992, UT1-UTC=-0.1269551, Accumulated Leap Seconds=26
  { 2448624.50, 58.312700 }, //03 January 1992, UT1-UTC=-0.1286996, Accumulated Leap Seconds=26
  { 2448625.50, 58.314467 }, //04 January 1992, UT1-UTC=-0.1304668, Accumulated Leap Seconds=26
  { 2448626.50, 58.316317 }, //05 January 1992, UT1-UTC=-0.1323166, Accumulated Leap Seconds=26
  { 2448627.50, 58.318296 }, //06 January 1992, UT1-UTC=-0.1342957, Accumulated Leap Seconds=26
  { 2448628.50, 58.320432 }, //07 January 1992, UT1-UTC=-0.1364324, Accumulated Leap Seconds=26
  { 2448629.50, 58.322734 }, //08 January 1992, UT1-UTC=-0.1387344, Accumulated Leap Seconds=26
  { 2448630.50, 58.325186 }, //09 January 1992, UT1-UTC=-0.1411861, Accumulated Leap Seconds=26
  { 2448631.50, 58.327759 }, //10 January 1992, UT1-UTC=-0.1437590, Accumulated Leap Seconds=26
  { 2448632.50, 58.330411 }, //11 January 1992, UT1-UTC=-0.1464111, Accumulated Leap Seconds=26
  { 2448633.50, 58.333084 }, //12 January 1992, UT1-UTC=-0.1490836, Accumulated Leap Seconds=26
  { 2448634.50, 58.335712 }, //13 January 1992, UT1-UTC=-0.1517124, Accumulated Leap Seconds=26
  { 2448635.50, 58.338237 }, //14 January 1992, UT1-UTC=-0.1542371, Accumulated Leap Seconds=26
  { 2448636.50, 58.340618 }, //15 January 1992, UT1-UTC=-0.1566178, Accumulated Leap Seconds=26
  { 2448637.50, 58.342864 }, //16 January 1992, UT1-UTC=-0.1588645, Accumulated Leap Seconds=26
  { 2448638.50, 58.345030 }, //17 January 1992, UT1-UTC=-0.1610304, Accumulated Leap Seconds=26
  { 2448639.50, 58.347213 }, //18 January 1992, UT1-UTC=-0.1632132, Accumulated Leap Seconds=26
  { 2448640.50, 58.349538 }, //19 January 1992, UT1-UTC=-0.1655376, Accumulated Leap Seconds=26
  { 2448641.50, 58.352107 }, //20 January 1992, UT1-UTC=-0.1681073, Accumulated Leap Seconds=26
  { 2448642.50, 58.354965 }, //21 January 1992, UT1-UTC=-0.1709648, Accumulated Leap Seconds=26
  { 2448643.50, 58.358059 }, //22 January 1992, UT1-UTC=-0.1740592, Accumulated Leap Seconds=26
  { 2448644.50, 58.361249 }, //23 January 1992, UT1-UTC=-0.1772490, Accumulated Leap Seconds=26
  { 2448645.50, 58.364386 }, //24 January 1992, UT1-UTC=-0.1803857, Accumulated Leap Seconds=26
  { 2448646.50, 58.367349 }, //25 January 1992, UT1-UTC=-0.1833489, Accumulated Leap Seconds=26
  { 2448647.50, 58.370078 }, //26 January 1992, UT1-UTC=-0.1860775, Accumulated Leap Seconds=26
  { 2448648.50, 58.372566 }, //27 January 1992, UT1-UTC=-0.1885657, Accumulated Leap Seconds=26
  { 2448649.50, 58.374847 }, //28 January 1992, UT1-UTC=-0.1908472, Accumulated Leap Seconds=26
  { 2448650.50, 58.376986 }, //29 January 1992, UT1-UTC=-0.1929856, Accumulated Leap Seconds=26
  { 2448651.50, 58.379068 }, //30 January 1992, UT1-UTC=-0.1950680, Accumulated Leap Seconds=26
  { 2448652.50, 58.381156 }, //31 January 1992, UT1-UTC=-0.1971564, Accumulated Leap Seconds=26
  { 2448653.50, 58.383302 }, //01 February 1992, UT1-UTC=-0.1993021, Accumulated Leap Seconds=26
  { 2448654.50, 58.385542 }, //02 February 1992, UT1-UTC=-0.2015417, Accumulated Leap Seconds=26
  { 2448655.50, 58.387897 }, //03 February 1992, UT1-UTC=-0.2038971, Accumulated Leap Seconds=26
  { 2448656.50, 58.390371 }, //04 February 1992, UT1-UTC=-0.2063706, Accumulated Leap Seconds=26
  { 2448657.50, 58.392953 }, //05 February 1992, UT1-UTC=-0.2089529, Accumulated Leap Seconds=26
  { 2448658.50, 58.395645 }, //06 February 1992, UT1-UTC=-0.2116447, Accumulated Leap Seconds=26
  { 2448659.50, 58.398410 }, //07 February 1992, UT1-UTC=-0.2144096, Accumulated Leap Seconds=26
  { 2448660.50, 58.401189 }, //08 February 1992, UT1-UTC=-0.2171890, Accumulated Leap Seconds=26
  { 2448661.50, 58.403904 }, //09 February 1992, UT1-UTC=-0.2199043, Accumulated Leap Seconds=26
  { 2448662.50, 58.406493 }, //10 February 1992, UT1-UTC=-0.2224926, Accumulated Leap Seconds=26
  { 2448663.50, 58.408912 }, //11 February 1992, UT1-UTC=-0.2249119, Accumulated Leap Seconds=26
  { 2448664.50, 58.411153 }, //12 February 1992, UT1-UTC=-0.2271526, Accumulated Leap Seconds=26
  { 2448665.50, 58.413251 }, //13 February 1992, UT1-UTC=-0.2292510, Accumulated Leap Seconds=26
  { 2448666.50, 58.415285 }, //14 February 1992, UT1-UTC=-0.2312855, Accumulated Leap Seconds=26
  { 2448667.50, 58.417365 }, //15 February 1992, UT1-UTC=-0.2333649, Accumulated Leap Seconds=26
  { 2448668.50, 58.419606 }, //16 February 1992, UT1-UTC=-0.2356063, Accumulated Leap Seconds=26
  { 2448669.50, 58.422091 }, //17 February 1992, UT1-UTC=-0.2380910, Accumulated Leap Seconds=26
  { 2448670.50, 58.424832 }, //18 February 1992, UT1-UTC=-0.2408320, Accumulated Leap Seconds=26
  { 2448671.50, 58.427764 }, //19 February 1992, UT1-UTC=-0.2437644, Accumulated Leap Seconds=26
  { 2448672.50, 58.430765 }, //20 February 1992, UT1-UTC=-0.2467653, Accumulated Leap Seconds=26
  { 2448673.50, 58.433694 }, //21 February 1992, UT1-UTC=-0.2496941, Accumulated Leap Seconds=26
  { 2448674.50, 58.436446 }, //22 February 1992, UT1-UTC=-0.2524464, Accumulated Leap Seconds=26
  { 2448675.50, 58.438985 }, //23 February 1992, UT1-UTC=-0.2549854, Accumulated Leap Seconds=26
  { 2448676.50, 58.441321 }, //24 February 1992, UT1-UTC=-0.2573213, Accumulated Leap Seconds=26
  { 2448677.50, 58.443500 }, //25 February 1992, UT1-UTC=-0.2594998, Accumulated Leap Seconds=26
  { 2448678.50, 58.445576 }, //26 February 1992, UT1-UTC=-0.2615761, Accumulated Leap Seconds=26
  { 2448679.50, 58.447575 }, //27 February 1992, UT1-UTC=-0.2635749, Accumulated Leap Seconds=26
  { 2448680.50, 58.449550 }, //28 February 1992, UT1-UTC=-0.2655501, Accumulated Leap Seconds=26
  { 2448681.50, 58.451561 }, //29 February 1992, UT1-UTC=-0.2675607, Accumulated Leap Seconds=26
  { 2448682.50, 58.453675 }, //01 March 1992, UT1-UTC=-0.2696748, Accumulated Leap Seconds=26
  { 2448683.50, 58.455944 }, //02 March 1992, UT1-UTC=-0.2719443, Accumulated Leap Seconds=26
  { 2448684.50, 58.458393 }, //03 March 1992, UT1-UTC=-0.2743931, Accumulated Leap Seconds=26
  { 2448685.50, 58.461016 }, //04 March 1992, UT1-UTC=-0.2770161, Accumulated Leap Seconds=26
  { 2448686.50, 58.463775 }, //05 March 1992, UT1-UTC=-0.2797752, Accumulated Leap Seconds=26
  { 2448687.50, 58.466608 }, //06 March 1992, UT1-UTC=-0.2826080, Accumulated Leap Seconds=26
  { 2448688.50, 58.469447 }, //07 March 1992, UT1-UTC=-0.2854469, Accumulated Leap Seconds=26
  { 2448689.50, 58.472230 }, //08 March 1992, UT1-UTC=-0.2882303, Accumulated Leap Seconds=26
  { 2448690.50, 58.474900 }, //09 March 1992, UT1-UTC=-0.2909004, Accumulated Leap Seconds=26
  { 2448691.50, 58.477422 }, //10 March 1992, UT1-UTC=-0.2934216, Accumulated Leap Seconds=26
  { 2448692.50, 58.479821 }, //11 March 1992, UT1-UTC=-0.2958209, Accumulated Leap Seconds=26
  { 2448693.50, 58.482213 }, //12 March 1992, UT1-UTC=-0.2982133, Accumulated Leap Seconds=26
  { 2448694.50, 58.484698 }, //13 March 1992, UT1-UTC=-0.3006978, Accumulated Leap Seconds=26
  { 2448695.50, 58.487367 }, //14 March 1992, UT1-UTC=-0.3033668, Accumulated Leap Seconds=26
  { 2448696.50, 58.490286 }, //15 March 1992, UT1-UTC=-0.3062863, Accumulated Leap Seconds=26
  { 2448697.50, 58.493485 }, //16 March 1992, UT1-UTC=-0.3094850, Accumulated Leap Seconds=26
  { 2448698.50, 58.496922 }, //17 March 1992, UT1-UTC=-0.3129225, Accumulated Leap Seconds=26
  { 2448699.50, 58.500492 }, //18 March 1992, UT1-UTC=-0.3164923, Accumulated Leap Seconds=26
  { 2448700.50, 58.504053 }, //19 March 1992, UT1-UTC=-0.3200528, Accumulated Leap Seconds=26
  { 2448701.50, 58.507473 }, //20 March 1992, UT1-UTC=-0.3234726, Accumulated Leap Seconds=26
  { 2448702.50, 58.510669 }, //21 March 1992, UT1-UTC=-0.3266688, Accumulated Leap Seconds=26
  { 2448703.50, 58.513610 }, //22 March 1992, UT1-UTC=-0.3296101, Accumulated Leap Seconds=26
  { 2448704.50, 58.516329 }, //23 March 1992, UT1-UTC=-0.3323293, Accumulated Leap Seconds=26
  { 2448705.50, 58.518894 }, //24 March 1992, UT1-UTC=-0.3348943, Accumulated Leap Seconds=26
  { 2448706.50, 58.521378 }, //25 March 1992, UT1-UTC=-0.3373785, Accumulated Leap Seconds=26
  { 2448707.50, 58.523839 }, //26 March 1992, UT1-UTC=-0.3398387, Accumulated Leap Seconds=26
  { 2448708.50, 58.526320 }, //27 March 1992, UT1-UTC=-0.3423203, Accumulated Leap Seconds=26
  { 2448709.50, 58.528861 }, //28 March 1992, UT1-UTC=-0.3448609, Accumulated Leap Seconds=26
  { 2448710.50, 58.531502 }, //29 March 1992, UT1-UTC=-0.3475017, Accumulated Leap Seconds=26
  { 2448711.50, 58.534267 }, //30 March 1992, UT1-UTC=-0.3502671, Accumulated Leap Seconds=26
  { 2448712.50, 58.537156 }, //31 March 1992, UT1-UTC=-0.3531561, Accumulated Leap Seconds=26
  { 2448713.50, 58.540144 }, //01 April 1992, UT1-UTC=-0.3561438, Accumulated Leap Seconds=26
  { 2448714.50, 58.543178 }, //02 April 1992, UT1-UTC=-0.3591784, Accumulated Leap Seconds=26
  { 2448715.50, 58.546194 }, //03 April 1992, UT1-UTC=-0.3621936, Accumulated Leap Seconds=26
  { 2448716.50, 58.549123 }, //04 April 1992, UT1-UTC=-0.3651228, Accumulated Leap Seconds=26
  { 2448717.50, 58.551915 }, //05 April 1992, UT1-UTC=-0.3679150, Accumulated Leap Seconds=26
  { 2448718.50, 58.554545 }, //06 April 1992, UT1-UTC=-0.3705455, Accumulated Leap Seconds=26
  { 2448719.50, 58.557025 }, //07 April 1992, UT1-UTC=-0.3730252, Accumulated Leap Seconds=26
  { 2448720.50, 58.559407 }, //08 April 1992, UT1-UTC=-0.3754075, Accumulated Leap Seconds=26
  { 2448721.50, 58.561786 }, //09 April 1992, UT1-UTC=-0.3777861, Accumulated Leap Seconds=26
  { 2448722.50, 58.564266 }, //10 April 1992, UT1-UTC=-0.3802656, Accumulated Leap Seconds=26
  { 2448723.50, 58.566937 }, //11 April 1992, UT1-UTC=-0.3829367, Accumulated Leap Seconds=26
  { 2448724.50, 58.569861 }, //12 April 1992, UT1-UTC=-0.3858606, Accumulated Leap Seconds=26
  { 2448725.50, 58.573045 }, //13 April 1992, UT1-UTC=-0.3890452, Accumulated Leap Seconds=26
  { 2448726.50, 58.576442 }, //14 April 1992, UT1-UTC=-0.3924419, Accumulated Leap Seconds=26
  { 2448727.50, 58.579952 }, //15 April 1992, UT1-UTC=-0.3959518, Accumulated Leap Seconds=26
  { 2448728.50, 58.583432 }, //16 April 1992, UT1-UTC=-0.3994317, Accumulated Leap Seconds=26
  { 2448729.50, 58.586740 }, //17 April 1992, UT1-UTC=-0.4027396, Accumulated Leap Seconds=26
  { 2448730.50, 58.589793 }, //18 April 1992, UT1-UTC=-0.4057934, Accumulated Leap Seconds=26
  { 2448731.50, 58.592573 }, //19 April 1992, UT1-UTC=-0.4085728, Accumulated Leap Seconds=26
  { 2448732.50, 58.595109 }, //20 April 1992, UT1-UTC=-0.4111091, Accumulated Leap Seconds=26
  { 2448733.50, 58.597464 }, //21 April 1992, UT1-UTC=-0.4134637, Accumulated Leap Seconds=26
  { 2448734.50, 58.599714 }, //22 April 1992, UT1-UTC=-0.4157139, Accumulated Leap Seconds=26
  { 2448735.50, 58.601955 }, //23 April 1992, UT1-UTC=-0.4179547, Accumulated Leap Seconds=26
  { 2448736.50, 58.604252 }, //24 April 1992, UT1-UTC=-0.4202520, Accumulated Leap Seconds=26
  { 2448737.50, 58.606648 }, //25 April 1992, UT1-UTC=-0.4226478, Accumulated Leap Seconds=26
  { 2448738.50, 58.609167 }, //26 April 1992, UT1-UTC=-0.4251665, Accumulated Leap Seconds=26
  { 2448739.50, 58.611807 }, //27 April 1992, UT1-UTC=-0.4278073, Accumulated Leap Seconds=26
  { 2448740.50, 58.614546 }, //28 April 1992, UT1-UTC=-0.4305458, Accumulated Leap Seconds=26
  { 2448741.50, 58.617334 }, //29 April 1992, UT1-UTC=-0.4333345, Accumulated Leap Seconds=26
  { 2448742.50, 58.620099 }, //30 April 1992, UT1-UTC=-0.4360990, Accumulated Leap Seconds=26
  { 2448743.50, 58.622771 }, //01 May 1992, UT1-UTC=-0.4387714, Accumulated Leap Seconds=26
  { 2448744.50, 58.625294 }, //02 May 1992, UT1-UTC=-0.4412945, Accumulated Leap Seconds=26
  { 2448745.50, 58.627637 }, //03 May 1992, UT1-UTC=-0.4436372, Accumulated Leap Seconds=26
  { 2448746.50, 58.629805 }, //04 May 1992, UT1-UTC=-0.4458046, Accumulated Leap Seconds=26
  { 2448747.50, 58.631842 }, //05 May 1992, UT1-UTC=-0.4478419, Accumulated Leap Seconds=26
  { 2448748.50, 58.633829 }, //06 May 1992, UT1-UTC=-0.4498295, Accumulated Leap Seconds=26
  { 2448749.50, 58.635867 }, //07 May 1992, UT1-UTC=-0.4518668, Accumulated Leap Seconds=26
  { 2448750.50, 58.638054 }, //08 May 1992, UT1-UTC=-0.4540540, Accumulated Leap Seconds=26
  { 2448751.50, 58.640452 }, //09 May 1992, UT1-UTC=-0.4564523, Accumulated Leap Seconds=26
  { 2448752.50, 58.643058 }, //10 May 1992, UT1-UTC=-0.4590577, Accumulated Leap Seconds=26
  { 2448753.50, 58.645812 }, //11 May 1992, UT1-UTC=-0.4618123, Accumulated Leap Seconds=26
  { 2448754.50, 58.648618 }, //12 May 1992, UT1-UTC=-0.4646183, Accumulated Leap Seconds=26
  { 2448755.50, 58.651373 }, //13 May 1992, UT1-UTC=-0.4673732, Accumulated Leap Seconds=26
  { 2448756.50, 58.654011 }, //14 May 1992, UT1-UTC=-0.4700106, Accumulated Leap Seconds=26
  { 2448757.50, 58.656464 }, //15 May 1992, UT1-UTC=-0.4724640, Accumulated Leap Seconds=26
  { 2448758.50, 58.658703 }, //16 May 1992, UT1-UTC=-0.4747028, Accumulated Leap Seconds=26
  { 2448759.50, 58.660752 }, //17 May 1992, UT1-UTC=-0.4767517, Accumulated Leap Seconds=26
  { 2448760.50, 58.662672 }, //18 May 1992, UT1-UTC=-0.4786725, Accumulated Leap Seconds=26
  { 2448761.50, 58.664543 }, //19 May 1992, UT1-UTC=-0.4805432, Accumulated Leap Seconds=26
  { 2448762.50, 58.666439 }, //20 May 1992, UT1-UTC=-0.4824386, Accumulated Leap Seconds=26
  { 2448763.50, 58.668407 }, //21 May 1992, UT1-UTC=-0.4844067, Accumulated Leap Seconds=26
  { 2448764.50, 58.670467 }, //22 May 1992, UT1-UTC=-0.4864668, Accumulated Leap Seconds=26
  { 2448765.50, 58.672613 }, //23 May 1992, UT1-UTC=-0.4886132, Accumulated Leap Seconds=26
  { 2448766.50, 58.674852 }, //24 May 1992, UT1-UTC=-0.4908520, Accumulated Leap Seconds=26
  { 2448767.50, 58.677169 }, //25 May 1992, UT1-UTC=-0.4931688, Accumulated Leap Seconds=26
  { 2448768.50, 58.679531 }, //26 May 1992, UT1-UTC=-0.4955306, Accumulated Leap Seconds=26
  { 2448769.50, 58.681887 }, //27 May 1992, UT1-UTC=-0.4978871, Accumulated Leap Seconds=26
  { 2448770.50, 58.684165 }, //28 May 1992, UT1-UTC=-0.5001655, Accumulated Leap Seconds=26
  { 2448771.50, 58.686299 }, //29 May 1992, UT1-UTC=-0.5022989, Accumulated Leap Seconds=26
  { 2448772.50, 58.688262 }, //30 May 1992, UT1-UTC=-0.5042618, Accumulated Leap Seconds=26
  { 2448773.50, 58.690045 }, //31 May 1992, UT1-UTC=-0.5060453, Accumulated Leap Seconds=26
  { 2448774.50, 58.691666 }, //01 June 1992, UT1-UTC=-0.5076662, Accumulated Leap Seconds=26
  { 2448775.50, 58.693196 }, //02 June 1992, UT1-UTC=-0.5091957, Accumulated Leap Seconds=26
  { 2448776.50, 58.694734 }, //03 June 1992, UT1-UTC=-0.5107342, Accumulated Leap Seconds=26
  { 2448777.50, 58.696388 }, //04 June 1992, UT1-UTC=-0.5123885, Accumulated Leap Seconds=26
  { 2448778.50, 58.698229 }, //05 June 1992, UT1-UTC=-0.5142293, Accumulated Leap Seconds=26
  { 2448779.50, 58.700260 }, //06 June 1992, UT1-UTC=-0.5162599, Accumulated Leap Seconds=26
  { 2448780.50, 58.702426 }, //07 June 1992, UT1-UTC=-0.5184265, Accumulated Leap Seconds=26
  { 2448781.50, 58.704636 }, //08 June 1992, UT1-UTC=-0.5206362, Accumulated Leap Seconds=26
  { 2448782.50, 58.706786 }, //09 June 1992, UT1-UTC=-0.5227857, Accumulated Leap Seconds=26
  { 2448783.50, 58.708794 }, //10 June 1992, UT1-UTC=-0.5247940, Accumulated Leap Seconds=26
  { 2448784.50, 58.710619 }, //11 June 1992, UT1-UTC=-0.5266195, Accumulated Leap Seconds=26
  { 2448785.50, 58.712255 }, //12 June 1992, UT1-UTC=-0.5282551, Accumulated Leap Seconds=26
  { 2448786.50, 58.713729 }, //13 June 1992, UT1-UTC=-0.5297292, Accumulated Leap Seconds=26
  { 2448787.50, 58.715095 }, //14 June 1992, UT1-UTC=-0.5310955, Accumulated Leap Seconds=26
  { 2448788.50, 58.716414 }, //15 June 1992, UT1-UTC=-0.5324140, Accumulated Leap Seconds=26
  { 2448789.50, 58.717739 }, //16 June 1992, UT1-UTC=-0.5337391, Accumulated Leap Seconds=26
  { 2448790.50, 58.719116 }, //17 June 1992, UT1-UTC=-0.5351158, Accumulated Leap Seconds=26
  { 2448791.50, 58.720576 }, //18 June 1992, UT1-UTC=-0.5365763, Accumulated Leap Seconds=26
  { 2448792.50, 58.722141 }, //19 June 1992, UT1-UTC=-0.5381414, Accumulated Leap Seconds=26
  { 2448793.50, 58.723818 }, //20 June 1992, UT1-UTC=-0.5398179, Accumulated Leap Seconds=26
  { 2448794.50, 58.725602 }, //21 June 1992, UT1-UTC=-0.5416018, Accumulated Leap Seconds=26
  { 2448795.50, 58.727459 }, //22 June 1992, UT1-UTC=-0.5434588, Accumulated Leap Seconds=26
  { 2448796.50, 58.729339 }, //23 June 1992, UT1-UTC=-0.5453387, Accumulated Leap Seconds=26
  { 2448797.50, 58.731186 }, //24 June 1992, UT1-UTC=-0.5471858, Accumulated Leap Seconds=26
  { 2448798.50, 58.732944 }, //25 June 1992, UT1-UTC=-0.5489437, Accumulated Leap Seconds=26
  { 2448799.50, 58.734569 }, //26 June 1992, UT1-UTC=-0.5505692, Accumulated Leap Seconds=26
  { 2448800.50, 58.736037 }, //27 June 1992, UT1-UTC=-0.5520372, Accumulated Leap Seconds=26
  { 2448801.50, 58.737350 }, //28 June 1992, UT1-UTC=-0.5533501, Accumulated Leap Seconds=26
  { 2448802.50, 58.738551 }, //29 June 1992, UT1-UTC=-0.5545515, Accumulated Leap Seconds=26
  { 2448803.50, 58.739722 }, //30 June 1992, UT1-UTC=-0.5557222, Accumulated Leap Seconds=26
  { 2448804.50, 58.740963 }, //01 July 1992, UT1-UTC= 0.4430372, Accumulated Leap Seconds=27
  { 2448805.50, 58.742335 }, //02 July 1992, UT1-UTC= 0.4416653, Accumulated Leap Seconds=27
  { 2448806.50, 58.743884 }, //03 July 1992, UT1-UTC= 0.4401162, Accumulated Leap Seconds=27
  { 2448807.50, 58.745596 }, //04 July 1992, UT1-UTC= 0.4384035, Accumulated Leap Seconds=27
  { 2448808.50, 58.747389 }, //05 July 1992, UT1-UTC= 0.4366113, Accumulated Leap Seconds=27
  { 2448809.50, 58.749147 }, //06 July 1992, UT1-UTC= 0.4348529, Accumulated Leap Seconds=27
  { 2448810.50, 58.750774 }, //07 July 1992, UT1-UTC= 0.4332264, Accumulated Leap Seconds=27
  { 2448811.50, 58.752214 }, //08 July 1992, UT1-UTC= 0.4317857, Accumulated Leap Seconds=27
  { 2448812.50, 58.753454 }, //09 July 1992, UT1-UTC= 0.4305456, Accumulated Leap Seconds=27
  { 2448813.50, 58.754511 }, //10 July 1992, UT1-UTC= 0.4294894, Accumulated Leap Seconds=27
  { 2448814.50, 58.755440 }, //11 July 1992, UT1-UTC= 0.4285595, Accumulated Leap Seconds=27
  { 2448815.50, 58.756301 }, //12 July 1992, UT1-UTC= 0.4276994, Accumulated Leap Seconds=27
  { 2448816.50, 58.757143 }, //13 July 1992, UT1-UTC= 0.4268572, Accumulated Leap Seconds=27
  { 2448817.50, 58.758013 }, //14 July 1992, UT1-UTC= 0.4259871, Accumulated Leap Seconds=27
  { 2448818.50, 58.758962 }, //15 July 1992, UT1-UTC= 0.4250377, Accumulated Leap Seconds=27
  { 2448819.50, 58.760036 }, //16 July 1992, UT1-UTC= 0.4239636, Accumulated Leap Seconds=27
  { 2448820.50, 58.761250 }, //17 July 1992, UT1-UTC= 0.4227496, Accumulated Leap Seconds=27
  { 2448821.50, 58.762591 }, //18 July 1992, UT1-UTC= 0.4214089, Accumulated Leap Seconds=27
  { 2448822.50, 58.764036 }, //19 July 1992, UT1-UTC= 0.4199636, Accumulated Leap Seconds=27
  { 2448823.50, 58.765554 }, //20 July 1992, UT1-UTC= 0.4184461, Accumulated Leap Seconds=27
  { 2448824.50, 58.767103 }, //21 July 1992, UT1-UTC= 0.4168969, Accumulated Leap Seconds=27
  { 2448825.50, 58.768643 }, //22 July 1992, UT1-UTC= 0.4153569, Accumulated Leap Seconds=27
  { 2448826.50, 58.770133 }, //23 July 1992, UT1-UTC= 0.4138666, Accumulated Leap Seconds=27
  { 2448827.50, 58.771544 }, //24 July 1992, UT1-UTC= 0.4124558, Accumulated Leap Seconds=27
  { 2448828.50, 58.772865 }, //25 July 1992, UT1-UTC= 0.4111349, Accumulated Leap Seconds=27
  { 2448829.50, 58.774117 }, //26 July 1992, UT1-UTC= 0.4098831, Accumulated Leap Seconds=27
  { 2448830.50, 58.775354 }, //27 July 1992, UT1-UTC= 0.4086463, Accumulated Leap Seconds=27
  { 2448831.50, 58.776641 }, //28 July 1992, UT1-UTC= 0.4073591, Accumulated Leap Seconds=27
  { 2448832.50, 58.778059 }, //29 July 1992, UT1-UTC= 0.4059414, Accumulated Leap Seconds=27
  { 2448833.50, 58.779690 }, //30 July 1992, UT1-UTC= 0.4043098, Accumulated Leap Seconds=27
  { 2448834.50, 58.781560 }, //31 July 1992, UT1-UTC= 0.4024395, Accumulated Leap Seconds=27
  { 2448835.50, 58.783619 }, //01 August 1992, UT1-UTC= 0.4003811, Accumulated Leap Seconds=27
  { 2448836.50, 58.785760 }, //02 August 1992, UT1-UTC= 0.3982402, Accumulated Leap Seconds=27
  { 2448837.50, 58.787856 }, //03 August 1992, UT1-UTC= 0.3961444, Accumulated Leap Seconds=27
  { 2448838.50, 58.789802 }, //04 August 1992, UT1-UTC= 0.3941976, Accumulated Leap Seconds=27
  { 2448839.50, 58.791542 }, //05 August 1992, UT1-UTC= 0.3924582, Accumulated Leap Seconds=27
  { 2448840.50, 58.793100 }, //06 August 1992, UT1-UTC= 0.3909000, Accumulated Leap Seconds=27
  { 2448841.50, 58.794493 }, //07 August 1992, UT1-UTC= 0.3895066, Accumulated Leap Seconds=27
  { 2448842.50, 58.795766 }, //08 August 1992, UT1-UTC= 0.3882335, Accumulated Leap Seconds=27
  { 2448843.50, 58.796969 }, //09 August 1992, UT1-UTC= 0.3870313, Accumulated Leap Seconds=27
  { 2448844.50, 58.798157 }, //10 August 1992, UT1-UTC= 0.3858431, Accumulated Leap Seconds=27
  { 2448845.50, 58.799395 }, //11 August 1992, UT1-UTC= 0.3846046, Accumulated Leap Seconds=27
  { 2448846.50, 58.800792 }, //12 August 1992, UT1-UTC= 0.3832084, Accumulated Leap Seconds=27
  { 2448847.50, 58.802384 }, //13 August 1992, UT1-UTC= 0.3816163, Accumulated Leap Seconds=27
  { 2448848.50, 58.804139 }, //14 August 1992, UT1-UTC= 0.3798610, Accumulated Leap Seconds=27
  { 2448849.50, 58.805999 }, //15 August 1992, UT1-UTC= 0.3780014, Accumulated Leap Seconds=27
  { 2448850.50, 58.807909 }, //16 August 1992, UT1-UTC= 0.3760911, Accumulated Leap Seconds=27
  { 2448851.50, 58.809820 }, //17 August 1992, UT1-UTC= 0.3741799, Accumulated Leap Seconds=27
  { 2448852.50, 58.811688 }, //18 August 1992, UT1-UTC= 0.3723119, Accumulated Leap Seconds=27
  { 2448853.50, 58.813486 }, //19 August 1992, UT1-UTC= 0.3705135, Accumulated Leap Seconds=27
  { 2448854.50, 58.815210 }, //20 August 1992, UT1-UTC= 0.3687900, Accumulated Leap Seconds=27
  { 2448855.50, 58.816856 }, //21 August 1992, UT1-UTC= 0.3671439, Accumulated Leap Seconds=27
  { 2448856.50, 58.818444 }, //22 August 1992, UT1-UTC= 0.3655557, Accumulated Leap Seconds=27
  { 2448857.50, 58.820023 }, //23 August 1992, UT1-UTC= 0.3639766, Accumulated Leap Seconds=27
  { 2448858.50, 58.821665 }, //24 August 1992, UT1-UTC= 0.3623345, Accumulated Leap Seconds=27
  { 2448859.50, 58.823450 }, //25 August 1992, UT1-UTC= 0.3605495, Accumulated Leap Seconds=27
  { 2448860.50, 58.825458 }, //26 August 1992, UT1-UTC= 0.3585419, Accumulated Leap Seconds=27
  { 2448861.50, 58.827750 }, //27 August 1992, UT1-UTC= 0.3562497, Accumulated Leap Seconds=27
  { 2448862.50, 58.830314 }, //28 August 1992, UT1-UTC= 0.3536861, Accumulated Leap Seconds=27
  { 2448863.50, 58.833040 }, //29 August 1992, UT1-UTC= 0.3509602, Accumulated Leap Seconds=27
  { 2448864.50, 58.835761 }, //30 August 1992, UT1-UTC= 0.3482388, Accumulated Leap Seconds=27
  { 2448865.50, 58.838306 }, //31 August 1992, UT1-UTC= 0.3456935, Accumulated Leap Seconds=27
  { 2448866.50, 58.840554 }, //01 September 1992, UT1-UTC= 0.3434457, Accumulated Leap Seconds=27
  { 2448867.50, 58.842477 }, //02 September 1992, UT1-UTC= 0.3415226, Accumulated Leap Seconds=27
  { 2448868.50, 58.844138 }, //03 September 1992, UT1-UTC= 0.3398621, Accumulated Leap Seconds=27
  { 2448869.50, 58.845614 }, //04 September 1992, UT1-UTC= 0.3383856, Accumulated Leap Seconds=27
  { 2448870.50, 58.846991 }, //05 September 1992, UT1-UTC= 0.3370085, Accumulated Leap Seconds=27
  { 2448871.50, 58.848362 }, //06 September 1992, UT1-UTC= 0.3356378, Accumulated Leap Seconds=27
  { 2448872.50, 58.849808 }, //07 September 1992, UT1-UTC= 0.3341918, Accumulated Leap Seconds=27
  { 2448873.50, 58.851392 }, //08 September 1992, UT1-UTC= 0.3326079, Accumulated Leap Seconds=27
  { 2448874.50, 58.853147 }, //09 September 1992, UT1-UTC= 0.3308526, Accumulated Leap Seconds=27
  { 2448875.50, 58.855061 }, //10 September 1992, UT1-UTC= 0.3289393, Accumulated Leap Seconds=27
  { 2448876.50, 58.857098 }, //11 September 1992, UT1-UTC= 0.3269017, Accumulated Leap Seconds=27
  { 2448877.50, 58.859213 }, //12 September 1992, UT1-UTC= 0.3247867, Accumulated Leap Seconds=27
  { 2448878.50, 58.861351 }, //13 September 1992, UT1-UTC= 0.3226486, Accumulated Leap Seconds=27
  { 2448879.50, 58.863453 }, //14 September 1992, UT1-UTC= 0.3205468, Accumulated Leap Seconds=27
  { 2448880.50, 58.865464 }, //15 September 1992, UT1-UTC= 0.3185359, Accumulated Leap Seconds=27
  { 2448881.50, 58.867350 }, //16 September 1992, UT1-UTC= 0.3166498, Accumulated Leap Seconds=27
  { 2448882.50, 58.869116 }, //17 September 1992, UT1-UTC= 0.3148841, Accumulated Leap Seconds=27
  { 2448883.50, 58.870791 }, //18 September 1992, UT1-UTC= 0.3132085, Accumulated Leap Seconds=27
  { 2448884.50, 58.872418 }, //19 September 1992, UT1-UTC= 0.3115818, Accumulated Leap Seconds=27
  { 2448885.50, 58.874053 }, //20 September 1992, UT1-UTC= 0.3099468, Accumulated Leap Seconds=27
  { 2448886.50, 58.875766 }, //21 September 1992, UT1-UTC= 0.3082337, Accumulated Leap Seconds=27
  { 2448887.50, 58.877623 }, //22 September 1992, UT1-UTC= 0.3063772, Accumulated Leap Seconds=27
  { 2448888.50, 58.879685 }, //23 September 1992, UT1-UTC= 0.3043152, Accumulated Leap Seconds=27
  { 2448889.50, 58.882005 }, //24 September 1992, UT1-UTC= 0.3019954, Accumulated Leap Seconds=27
  { 2448890.50, 58.884542 }, //25 September 1992, UT1-UTC= 0.2994579, Accumulated Leap Seconds=27
  { 2448891.50, 58.887192 }, //26 September 1992, UT1-UTC= 0.2968080, Accumulated Leap Seconds=27
  { 2448892.50, 58.889817 }, //27 September 1992, UT1-UTC= 0.2941827, Accumulated Leap Seconds=27
  { 2448893.50, 58.892301 }, //28 September 1992, UT1-UTC= 0.2916985, Accumulated Leap Seconds=27
  { 2448894.50, 58.894585 }, //29 September 1992, UT1-UTC= 0.2894155, Accumulated Leap Seconds=27
  { 2448895.50, 58.896664 }, //30 September 1992, UT1-UTC= 0.2873360, Accumulated Leap Seconds=27
  { 2448896.50, 58.898579 }, //01 October 1992, UT1-UTC= 0.2854210, Accumulated Leap Seconds=27
  { 2448897.50, 58.900391 }, //02 October 1992, UT1-UTC= 0.2836088, Accumulated Leap Seconds=27
  { 2448898.50, 58.902165 }, //03 October 1992, UT1-UTC= 0.2818348, Accumulated Leap Seconds=27
  { 2448899.50, 58.903973 }, //04 October 1992, UT1-UTC= 0.2800272, Accumulated Leap Seconds=27
  { 2448900.50, 58.905872 }, //05 October 1992, UT1-UTC= 0.2781277, Accumulated Leap Seconds=27
  { 2448901.50, 58.907904 }, //06 October 1992, UT1-UTC= 0.2760962, Accumulated Leap Seconds=27
  { 2448902.50, 58.910087 }, //07 October 1992, UT1-UTC= 0.2739126, Accumulated Leap Seconds=27
  { 2448903.50, 58.912412 }, //08 October 1992, UT1-UTC= 0.2715878, Accumulated Leap Seconds=27
  { 2448904.50, 58.914845 }, //09 October 1992, UT1-UTC= 0.2691545, Accumulated Leap Seconds=27
  { 2448905.50, 58.917338 }, //10 October 1992, UT1-UTC= 0.2666616, Accumulated Leap Seconds=27
  { 2448906.50, 58.919829 }, //11 October 1992, UT1-UTC= 0.2641714, Accumulated Leap Seconds=27
  { 2448907.50, 58.922256 }, //12 October 1992, UT1-UTC= 0.2617437, Accumulated Leap Seconds=27
  { 2448908.50, 58.924574 }, //13 October 1992, UT1-UTC= 0.2594263, Accumulated Leap Seconds=27
  { 2448909.50, 58.926764 }, //14 October 1992, UT1-UTC= 0.2572357, Accumulated Leap Seconds=27
  { 2448910.50, 58.928864 }, //15 October 1992, UT1-UTC= 0.2551356, Accumulated Leap Seconds=27
  { 2448911.50, 58.930908 }, //16 October 1992, UT1-UTC= 0.2530918, Accumulated Leap Seconds=27
  { 2448912.50, 58.932950 }, //17 October 1992, UT1-UTC= 0.2510500, Accumulated Leap Seconds=27
  { 2448913.50, 58.935079 }, //18 October 1992, UT1-UTC= 0.2489210, Accumulated Leap Seconds=27
  { 2448914.50, 58.937383 }, //19 October 1992, UT1-UTC= 0.2466168, Accumulated Leap Seconds=27
  { 2448915.50, 58.939926 }, //20 October 1992, UT1-UTC= 0.2440735, Accumulated Leap Seconds=27
  { 2448916.50, 58.942744 }, //21 October 1992, UT1-UTC= 0.2412558, Accumulated Leap Seconds=27
  { 2448917.50, 58.945810 }, //22 October 1992, UT1-UTC= 0.2381899, Accumulated Leap Seconds=27
  { 2448918.50, 58.949005 }, //23 October 1992, UT1-UTC= 0.2349949, Accumulated Leap Seconds=27
  { 2448919.50, 58.952176 }, //24 October 1992, UT1-UTC= 0.2318239, Accumulated Leap Seconds=27
  { 2448920.50, 58.955199 }, //25 October 1992, UT1-UTC= 0.2288010, Accumulated Leap Seconds=27
  { 2448921.50, 58.957990 }, //26 October 1992, UT1-UTC= 0.2260100, Accumulated Leap Seconds=27
  { 2448922.50, 58.960518 }, //27 October 1992, UT1-UTC= 0.2234818, Accumulated Leap Seconds=27
  { 2448923.50, 58.962819 }, //28 October 1992, UT1-UTC= 0.2211813, Accumulated Leap Seconds=27
  { 2448924.50, 58.964974 }, //29 October 1992, UT1-UTC= 0.2190256, Accumulated Leap Seconds=27
  { 2448925.50, 58.967072 }, //30 October 1992, UT1-UTC= 0.2169279, Accumulated Leap Seconds=27
  { 2448926.50, 58.969186 }, //31 October 1992, UT1-UTC= 0.2148141, Accumulated Leap Seconds=27
  { 2448927.50, 58.971368 }, //01 November 1992, UT1-UTC= 0.2126322, Accumulated Leap Seconds=27
  { 2448928.50, 58.973626 }, //02 November 1992, UT1-UTC= 0.2103739, Accumulated Leap Seconds=27
  { 2448929.50, 58.975941 }, //03 November 1992, UT1-UTC= 0.2080589, Accumulated Leap Seconds=27
  { 2448930.50, 58.978297 }, //04 November 1992, UT1-UTC= 0.2057030, Accumulated Leap Seconds=27
  { 2448931.50, 58.980717 }, //05 November 1992, UT1-UTC= 0.2032827, Accumulated Leap Seconds=27
  { 2448932.50, 58.983211 }, //06 November 1992, UT1-UTC= 0.2007890, Accumulated Leap Seconds=27
  { 2448933.50, 58.985723 }, //07 November 1992, UT1-UTC= 0.1982766, Accumulated Leap Seconds=27
  { 2448934.50, 58.988200 }, //08 November 1992, UT1-UTC= 0.1957996, Accumulated Leap Seconds=27
  { 2448935.50, 58.990599 }, //09 November 1992, UT1-UTC= 0.1934013, Accumulated Leap Seconds=27
  { 2448936.50, 58.992888 }, //10 November 1992, UT1-UTC= 0.1911117, Accumulated Leap Seconds=27
  { 2448937.50, 58.995071 }, //11 November 1992, UT1-UTC= 0.1889294, Accumulated Leap Seconds=27
  { 2448938.50, 58.997178 }, //12 November 1992, UT1-UTC= 0.1868219, Accumulated Leap Seconds=27
  { 2448939.50, 58.999256 }, //13 November 1992, UT1-UTC= 0.1847444, Accumulated Leap Seconds=27
  { 2448940.50, 59.001375 }, //14 November 1992, UT1-UTC= 0.1826248, Accumulated Leap Seconds=27
  { 2448941.50, 59.003619 }, //15 November 1992, UT1-UTC= 0.1803805, Accumulated Leap Seconds=27
  { 2448942.50, 59.006065 }, //16 November 1992, UT1-UTC= 0.1779354, Accumulated Leap Seconds=27
  { 2448943.50, 59.008745 }, //17 November 1992, UT1-UTC= 0.1752549, Accumulated Leap Seconds=27
  { 2448944.50, 59.011642 }, //18 November 1992, UT1-UTC= 0.1723580, Accumulated Leap Seconds=27
  { 2448945.50, 59.014673 }, //19 November 1992, UT1-UTC= 0.1693270, Accumulated Leap Seconds=27
  { 2448946.50, 59.017721 }, //20 November 1992, UT1-UTC= 0.1662786, Accumulated Leap Seconds=27
  { 2448947.50, 59.020669 }, //21 November 1992, UT1-UTC= 0.1633308, Accumulated Leap Seconds=27
  { 2448948.50, 59.023437 }, //22 November 1992, UT1-UTC= 0.1605629, Accumulated Leap Seconds=27
  { 2448949.50, 59.025987 }, //23 November 1992, UT1-UTC= 0.1580130, Accumulated Leap Seconds=27
  { 2448950.50, 59.028327 }, //24 November 1992, UT1-UTC= 0.1556732, Accumulated Leap Seconds=27
  { 2448951.50, 59.030508 }, //25 November 1992, UT1-UTC= 0.1534920, Accumulated Leap Seconds=27
  { 2448952.50, 59.032608 }, //26 November 1992, UT1-UTC= 0.1513920, Accumulated Leap Seconds=27
  { 2448953.50, 59.034703 }, //27 November 1992, UT1-UTC= 0.1492972, Accumulated Leap Seconds=27
  { 2448954.50, 59.036849 }, //28 November 1992, UT1-UTC= 0.1471512, Accumulated Leap Seconds=27
  { 2448955.50, 59.039080 }, //29 November 1992, UT1-UTC= 0.1449200, Accumulated Leap Seconds=27
  { 2448956.50, 59.041412 }, //30 November 1992, UT1-UTC= 0.1425879, Accumulated Leap Seconds=27
  { 2448957.50, 59.043837 }, //01 December 1992, UT1-UTC= 0.1401630, Accumulated Leap Seconds=27
  { 2448958.50, 59.046366 }, //02 December 1992, UT1-UTC= 0.1376336, Accumulated Leap Seconds=27
  { 2448959.50, 59.048999 }, //03 December 1992, UT1-UTC= 0.1350007, Accumulated Leap Seconds=27
  { 2448960.50, 59.051701 }, //04 December 1992, UT1-UTC= 0.1322991, Accumulated Leap Seconds=27
  { 2448961.50, 59.054402 }, //05 December 1992, UT1-UTC= 0.1295981, Accumulated Leap Seconds=27
  { 2448962.50, 59.057058 }, //06 December 1992, UT1-UTC= 0.1269417, Accumulated Leap Seconds=27
  { 2448963.50, 59.059640 }, //07 December 1992, UT1-UTC= 0.1243596, Accumulated Leap Seconds=27
  { 2448964.50, 59.062116 }, //08 December 1992, UT1-UTC= 0.1218842, Accumulated Leap Seconds=27
  { 2448965.50, 59.064488 }, //09 December 1992, UT1-UTC= 0.1195120, Accumulated Leap Seconds=27
  { 2448966.50, 59.066806 }, //10 December 1992, UT1-UTC= 0.1171939, Accumulated Leap Seconds=27
  { 2448967.50, 59.069139 }, //11 December 1992, UT1-UTC= 0.1148608, Accumulated Leap Seconds=27
  { 2448968.50, 59.071565 }, //12 December 1992, UT1-UTC= 0.1124347, Accumulated Leap Seconds=27
  { 2448969.50, 59.074151 }, //13 December 1992, UT1-UTC= 0.1098494, Accumulated Leap Seconds=27
  { 2448970.50, 59.076922 }, //14 December 1992, UT1-UTC= 0.1070776, Accumulated Leap Seconds=27
  { 2448971.50, 59.079859 }, //15 December 1992, UT1-UTC= 0.1041410, Accumulated Leap Seconds=27
  { 2448972.50, 59.082896 }, //16 December 1992, UT1-UTC= 0.1011043, Accumulated Leap Seconds=27
  { 2448973.50, 59.085925 }, //17 December 1992, UT1-UTC= 0.0980745, Accumulated Leap Seconds=27
  { 2448974.50, 59.088842 }, //18 December 1992, UT1-UTC= 0.0951582, Accumulated Leap Seconds=27
  { 2448975.50, 59.091563 }, //19 December 1992, UT1-UTC= 0.0924366, Accumulated Leap Seconds=27
  { 2448976.50, 59.094049 }, //20 December 1992, UT1-UTC= 0.0899514, Accumulated Leap Seconds=27
  { 2448977.50, 59.096302 }, //21 December 1992, UT1-UTC= 0.0876983, Accumulated Leap Seconds=27
  { 2448978.50, 59.098371 }, //22 December 1992, UT1-UTC= 0.0856286, Accumulated Leap Seconds=27
  { 2448979.50, 59.100339 }, //23 December 1992, UT1-UTC= 0.0836607, Accumulated Leap Seconds=27
  { 2448980.50, 59.102298 }, //24 December 1992, UT1-UTC= 0.0817015, Accumulated Leap Seconds=27
  { 2448981.50, 59.104324 }, //25 December 1992, UT1-UTC= 0.0796758, Accumulated Leap Seconds=27
  { 2448982.50, 59.106467 }, //26 December 1992, UT1-UTC= 0.0775328, Accumulated Leap Seconds=27
  { 2448983.50, 59.108754 }, //27 December 1992, UT1-UTC= 0.0752456, Accumulated Leap Seconds=27
  { 2448984.50, 59.111190 }, //28 December 1992, UT1-UTC= 0.0728104, Accumulated Leap Seconds=27
  { 2448985.50, 59.113756 }, //29 December 1992, UT1-UTC= 0.0702439, Accumulated Leap Seconds=27
  { 2448986.50, 59.116419 }, //30 December 1992, UT1-UTC= 0.0675806, Accumulated Leap Seconds=27
  { 2448987.50, 59.119128 }, //31 December 1992, UT1-UTC= 0.0648720, Accumulated Leap Seconds=27
  { 2448988.50, 59.121841 }, //01 January 1993, UT1-UTC= 0.0621586, Accumulated Leap Seconds=27
  { 2448989.50, 59.124525 }, //02 January 1993, UT1-UTC= 0.0594754, Accumulated Leap Seconds=27
  { 2448990.50, 59.127148 }, //03 January 1993, UT1-UTC= 0.0568516, Accumulated Leap Seconds=27
  { 2448991.50, 59.129696 }, //04 January 1993, UT1-UTC= 0.0543039, Accumulated Leap Seconds=27
  { 2448992.50, 59.132170 }, //05 January 1993, UT1-UTC= 0.0518298, Accumulated Leap Seconds=27
  { 2448993.50, 59.134595 }, //06 January 1993, UT1-UTC= 0.0494045, Accumulated Leap Seconds=27
  { 2448994.50, 59.137003 }, //07 January 1993, UT1-UTC= 0.0469970, Accumulated Leap Seconds=27
  { 2448995.50, 59.139444 }, //08 January 1993, UT1-UTC= 0.0445561, Accumulated Leap Seconds=27
  { 2448996.50, 59.141978 }, //09 January 1993, UT1-UTC= 0.0420224, Accumulated Leap Seconds=27
  { 2448997.50, 59.144665 }, //10 January 1993, UT1-UTC= 0.0393348, Accumulated Leap Seconds=27
  { 2448998.50, 59.147519 }, //11 January 1993, UT1-UTC= 0.0364810, Accumulated Leap Seconds=27
  { 2448999.50, 59.150491 }, //12 January 1993, UT1-UTC= 0.0335089, Accumulated Leap Seconds=27
  { 2449000.50, 59.153492 }, //13 January 1993, UT1-UTC= 0.0305078, Accumulated Leap Seconds=27
  { 2449001.50, 59.156421 }, //14 January 1993, UT1-UTC= 0.0275793, Accumulated Leap Seconds=27
  { 2449002.50, 59.159194 }, //15 January 1993, UT1-UTC= 0.0248062, Accumulated Leap Seconds=27
  { 2449003.50, 59.161771 }, //16 January 1993, UT1-UTC= 0.0222291, Accumulated Leap Seconds=27
  { 2449004.50, 59.164149 }, //17 January 1993, UT1-UTC= 0.0198512, Accumulated Leap Seconds=27
  { 2449005.50, 59.166359 }, //18 January 1993, UT1-UTC= 0.0176414, Accumulated Leap Seconds=27
  { 2449006.50, 59.168450 }, //19 January 1993, UT1-UTC= 0.0155495, Accumulated Leap Seconds=27
  { 2449007.50, 59.170490 }, //20 January 1993, UT1-UTC= 0.0135102, Accumulated Leap Seconds=27
  { 2449008.50, 59.172556 }, //21 January 1993, UT1-UTC= 0.0114436, Accumulated Leap Seconds=27
  { 2449009.50, 59.174701 }, //22 January 1993, UT1-UTC= 0.0092994, Accumulated Leap Seconds=27
  { 2449010.50, 59.176950 }, //23 January 1993, UT1-UTC= 0.0070502, Accumulated Leap Seconds=27
  { 2449011.50, 59.179320 }, //24 January 1993, UT1-UTC= 0.0046799, Accumulated Leap Seconds=27
  { 2449012.50, 59.181813 }, //25 January 1993, UT1-UTC= 0.0021865, Accumulated Leap Seconds=27
  { 2449013.50, 59.184419 }, //26 January 1993, UT1-UTC=-0.0004188, Accumulated Leap Seconds=27
  { 2449014.50, 59.187117 }, //27 January 1993, UT1-UTC=-0.0031169, Accumulated Leap Seconds=27
  { 2449015.50, 59.189859 }, //28 January 1993, UT1-UTC=-0.0058587, Accumulated Leap Seconds=27
  { 2449016.50, 59.192580 }, //29 January 1993, UT1-UTC=-0.0085798, Accumulated Leap Seconds=27
  { 2449017.50, 59.195237 }, //30 January 1993, UT1-UTC=-0.0112369, Accumulated Leap Seconds=27
  { 2449018.50, 59.197802 }, //31 January 1993, UT1-UTC=-0.0138020, Accumulated Leap Seconds=27
  { 2449019.50, 59.200269 }, //01 February 1993, UT1-UTC=-0.0162687, Accumulated Leap Seconds=27
  { 2449020.50, 59.202653 }, //02 February 1993, UT1-UTC=-0.0186532, Accumulated Leap Seconds=27
  { 2449021.50, 59.205006 }, //03 February 1993, UT1-UTC=-0.0210060, Accumulated Leap Seconds=27
  { 2449022.50, 59.207411 }, //04 February 1993, UT1-UTC=-0.0234108, Accumulated Leap Seconds=27
  { 2449023.50, 59.209960 }, //05 February 1993, UT1-UTC=-0.0259598, Accumulated Leap Seconds=27
  { 2449024.50, 59.212733 }, //06 February 1993, UT1-UTC=-0.0287326, Accumulated Leap Seconds=27
  { 2449025.50, 59.215762 }, //07 February 1993, UT1-UTC=-0.0317622, Accumulated Leap Seconds=27
  { 2449026.50, 59.219012 }, //08 February 1993, UT1-UTC=-0.0350123, Accumulated Leap Seconds=27
  { 2449027.50, 59.222380 }, //09 February 1993, UT1-UTC=-0.0383801, Accumulated Leap Seconds=27
  { 2449028.50, 59.225731 }, //10 February 1993, UT1-UTC=-0.0417313, Accumulated Leap Seconds=27
  { 2449029.50, 59.228927 }, //11 February 1993, UT1-UTC=-0.0449274, Accumulated Leap Seconds=27
  { 2449030.50, 59.231868 }, //12 February 1993, UT1-UTC=-0.0478682, Accumulated Leap Seconds=27
  { 2449031.50, 59.234524 }, //13 February 1993, UT1-UTC=-0.0505241, Accumulated Leap Seconds=27
  { 2449032.50, 59.236942 }, //14 February 1993, UT1-UTC=-0.0529420, Accumulated Leap Seconds=27
  { 2449033.50, 59.239206 }, //15 February 1993, UT1-UTC=-0.0552062, Accumulated Leap Seconds=27
  { 2449034.50, 59.241409 }, //16 February 1993, UT1-UTC=-0.0574086, Accumulated Leap Seconds=27
  { 2449035.50, 59.243637 }, //17 February 1993, UT1-UTC=-0.0596371, Accumulated Leap Seconds=27
  { 2449036.50, 59.245948 }, //18 February 1993, UT1-UTC=-0.0619484, Accumulated Leap Seconds=27
  { 2449037.50, 59.248384 }, //19 February 1993, UT1-UTC=-0.0643843, Accumulated Leap Seconds=27
  { 2449038.50, 59.250965 }, //20 February 1993, UT1-UTC=-0.0669654, Accumulated Leap Seconds=27
  { 2449039.50, 59.253677 }, //21 February 1993, UT1-UTC=-0.0696767, Accumulated Leap Seconds=27
  { 2449040.50, 59.256479 }, //22 February 1993, UT1-UTC=-0.0724794, Accumulated Leap Seconds=27
  { 2449041.50, 59.259317 }, //23 February 1993, UT1-UTC=-0.0753175, Accumulated Leap Seconds=27
  { 2449042.50, 59.262132 }, //24 February 1993, UT1-UTC=-0.0781319, Accumulated Leap Seconds=27
  { 2449043.50, 59.264879 }, //25 February 1993, UT1-UTC=-0.0808788, Accumulated Leap Seconds=27
  { 2449044.50, 59.267525 }, //26 February 1993, UT1-UTC=-0.0835249, Accumulated Leap Seconds=27
  { 2449045.50, 59.270050 }, //27 February 1993, UT1-UTC=-0.0860497, Accumulated Leap Seconds=27
  { 2449046.50, 59.272449 }, //28 February 1993, UT1-UTC=-0.0884491, Accumulated Leap Seconds=27
  { 2449047.50, 59.274737 }, //01 March 1993, UT1-UTC=-0.0907370, Accumulated Leap Seconds=27
  { 2449048.50, 59.276949 }, //02 March 1993, UT1-UTC=-0.0929494, Accumulated Leap Seconds=27
  { 2449049.50, 59.279152 }, //03 March 1993, UT1-UTC=-0.0951520, Accumulated Leap Seconds=27
  { 2449050.50, 59.281437 }, //04 March 1993, UT1-UTC=-0.0974371, Accumulated Leap Seconds=27
  { 2449051.50, 59.283891 }, //05 March 1993, UT1-UTC=-0.0998912, Accumulated Leap Seconds=27
  { 2449052.50, 59.286574 }, //06 March 1993, UT1-UTC=-0.1025743, Accumulated Leap Seconds=27
  { 2449053.50, 59.289510 }, //07 March 1993, UT1-UTC=-0.1055100, Accumulated Leap Seconds=27
  { 2449054.50, 59.292658 }, //08 March 1993, UT1-UTC=-0.1086584, Accumulated Leap Seconds=27
  { 2449055.50, 59.295923 }, //09 March 1993, UT1-UTC=-0.1119232, Accumulated Leap Seconds=27
  { 2449056.50, 59.299184 }, //10 March 1993, UT1-UTC=-0.1151838, Accumulated Leap Seconds=27
  { 2449057.50, 59.302321 }, //11 March 1993, UT1-UTC=-0.1183210, Accumulated Leap Seconds=27
  { 2449058.50, 59.305265 }, //12 March 1993, UT1-UTC=-0.1212646, Accumulated Leap Seconds=27
  { 2449059.50, 59.308010 }, //13 March 1993, UT1-UTC=-0.1240096, Accumulated Leap Seconds=27
  { 2449060.50, 59.310606 }, //14 March 1993, UT1-UTC=-0.1266065, Accumulated Leap Seconds=27
  { 2449061.50, 59.313123 }, //15 March 1993, UT1-UTC=-0.1291234, Accumulated Leap Seconds=27
  { 2449062.50, 59.315619 }, //16 March 1993, UT1-UTC=-0.1316186, Accumulated Leap Seconds=27
  { 2449063.50, 59.318142 }, //17 March 1993, UT1-UTC=-0.1341417, Accumulated Leap Seconds=27
  { 2449064.50, 59.320728 }, //18 March 1993, UT1-UTC=-0.1367277, Accumulated Leap Seconds=27
  { 2449065.50, 59.323397 }, //19 March 1993, UT1-UTC=-0.1393974, Accumulated Leap Seconds=27
  { 2449066.50, 59.326156 }, //20 March 1993, UT1-UTC=-0.1421558, Accumulated Leap Seconds=27
  { 2449067.50, 59.328991 }, //21 March 1993, UT1-UTC=-0.1449906, Accumulated Leap Seconds=27
  { 2449068.50, 59.331873 }, //22 March 1993, UT1-UTC=-0.1478732, Accumulated Leap Seconds=27
  { 2449069.50, 59.334764 }, //23 March 1993, UT1-UTC=-0.1507642, Accumulated Leap Seconds=27
  { 2449070.50, 59.337625 }, //24 March 1993, UT1-UTC=-0.1536251, Accumulated Leap Seconds=27
  { 2449071.50, 59.340411 }, //25 March 1993, UT1-UTC=-0.1564110, Accumulated Leap Seconds=27
  { 2449072.50, 59.343076 }, //26 March 1993, UT1-UTC=-0.1590763, Accumulated Leap Seconds=27
  { 2449073.50, 59.345616 }, //27 March 1993, UT1-UTC=-0.1616165, Accumulated Leap Seconds=27
  { 2449074.50, 59.348042 }, //28 March 1993, UT1-UTC=-0.1640415, Accumulated Leap Seconds=27
  { 2449075.50, 59.350377 }, //29 March 1993, UT1-UTC=-0.1663773, Accumulated Leap Seconds=27
  { 2449076.50, 59.352668 }, //30 March 1993, UT1-UTC=-0.1686679, Accumulated Leap Seconds=27
  { 2449077.50, 59.354982 }, //31 March 1993, UT1-UTC=-0.1709824, Accumulated Leap Seconds=27
  { 2449078.50, 59.357413 }, //01 April 1993, UT1-UTC=-0.1734134, Accumulated Leap Seconds=27
  { 2449079.50, 59.360042 }, //02 April 1993, UT1-UTC=-0.1760424, Accumulated Leap Seconds=27
  { 2449080.50, 59.362915 }, //03 April 1993, UT1-UTC=-0.1789153, Accumulated Leap Seconds=27
  { 2449081.50, 59.366024 }, //04 April 1993, UT1-UTC=-0.1820242, Accumulated Leap Seconds=27
  { 2449082.50, 59.369302 }, //05 April 1993, UT1-UTC=-0.1853021, Accumulated Leap Seconds=27
  { 2449083.50, 59.372642 }, //06 April 1993, UT1-UTC=-0.1886421, Accumulated Leap Seconds=27
  { 2449084.50, 59.375934 }, //07 April 1993, UT1-UTC=-0.1919338, Accumulated Leap Seconds=27
  { 2449085.50, 59.379079 }, //08 April 1993, UT1-UTC=-0.1950789, Accumulated Leap Seconds=27
  { 2449086.50, 59.382033 }, //09 April 1993, UT1-UTC=-0.1980334, Accumulated Leap Seconds=27
  { 2449087.50, 59.384818 }, //10 April 1993, UT1-UTC=-0.2008176, Accumulated Leap Seconds=27
  { 2449088.50, 59.387492 }, //11 April 1993, UT1-UTC=-0.2034918, Accumulated Leap Seconds=27
  { 2449089.50, 59.390124 }, //12 April 1993, UT1-UTC=-0.2061237, Accumulated Leap Seconds=27
  { 2449090.50, 59.392763 }, //13 April 1993, UT1-UTC=-0.2087629, Accumulated Leap Seconds=27
  { 2449091.50, 59.395456 }, //14 April 1993, UT1-UTC=-0.2114559, Accumulated Leap Seconds=27
  { 2449092.50, 59.398249 }, //15 April 1993, UT1-UTC=-0.2142492, Accumulated Leap Seconds=27
  { 2449093.50, 59.401134 }, //16 April 1993, UT1-UTC=-0.2171343, Accumulated Leap Seconds=27
  { 2449094.50, 59.404079 }, //17 April 1993, UT1-UTC=-0.2200786, Accumulated Leap Seconds=27
  { 2449095.50, 59.407048 }, //18 April 1993, UT1-UTC=-0.2230481, Accumulated Leap Seconds=27
  { 2449096.50, 59.410005 }, //19 April 1993, UT1-UTC=-0.2260048, Accumulated Leap Seconds=27
  { 2449097.50, 59.412909 }, //20 April 1993, UT1-UTC=-0.2289091, Accumulated Leap Seconds=27
  { 2449098.50, 59.415742 }, //21 April 1993, UT1-UTC=-0.2317422, Accumulated Leap Seconds=27
  { 2449099.50, 59.418521 }, //22 April 1993, UT1-UTC=-0.2345212, Accumulated Leap Seconds=27
  { 2449100.50, 59.421254 }, //23 April 1993, UT1-UTC=-0.2372538, Accumulated Leap Seconds=27
  { 2449101.50, 59.423913 }, //24 April 1993, UT1-UTC=-0.2399125, Accumulated Leap Seconds=27
  { 2449102.50, 59.426499 }, //25 April 1993, UT1-UTC=-0.2424995, Accumulated Leap Seconds=27
  { 2449103.50, 59.429044 }, //26 April 1993, UT1-UTC=-0.2450436, Accumulated Leap Seconds=27
  { 2449104.50, 59.431587 }, //27 April 1993, UT1-UTC=-0.2475871, Accumulated Leap Seconds=27
  { 2449105.50, 59.434213 }, //28 April 1993, UT1-UTC=-0.2502132, Accumulated Leap Seconds=27
  { 2449106.50, 59.437030 }, //29 April 1993, UT1-UTC=-0.2530302, Accumulated Leap Seconds=27
  { 2449107.50, 59.440094 }, //30 April 1993, UT1-UTC=-0.2560944, Accumulated Leap Seconds=27
  { 2449108.50, 59.443383 }, //01 May 1993, UT1-UTC=-0.2593827, Accumulated Leap Seconds=27
  { 2449109.50, 59.446825 }, //02 May 1993, UT1-UTC=-0.2628253, Accumulated Leap Seconds=27
  { 2449110.50, 59.450310 }, //03 May 1993, UT1-UTC=-0.2663097, Accumulated Leap Seconds=27
  { 2449111.50, 59.453701 }, //04 May 1993, UT1-UTC=-0.2697007, Accumulated Leap Seconds=27
  { 2449112.50, 59.456882 }, //05 May 1993, UT1-UTC=-0.2728818, Accumulated Leap Seconds=27
  { 2449113.50, 59.459803 }, //06 May 1993, UT1-UTC=-0.2758031, Accumulated Leap Seconds=27
  { 2449114.50, 59.462473 }, //07 May 1993, UT1-UTC=-0.2784728, Accumulated Leap Seconds=27
  { 2449115.50, 59.464959 }, //08 May 1993, UT1-UTC=-0.2809586, Accumulated Leap Seconds=27
  { 2449116.50, 59.467345 }, //09 May 1993, UT1-UTC=-0.2833450, Accumulated Leap Seconds=27
  { 2449117.50, 59.469706 }, //10 May 1993, UT1-UTC=-0.2857062, Accumulated Leap Seconds=27
  { 2449118.50, 59.472091 }, //11 May 1993, UT1-UTC=-0.2880913, Accumulated Leap Seconds=27
  { 2449119.50, 59.474515 }, //12 May 1993, UT1-UTC=-0.2905148, Accumulated Leap Seconds=27
  { 2449120.50, 59.476971 }, //13 May 1993, UT1-UTC=-0.2929705, Accumulated Leap Seconds=27
  { 2449121.50, 59.479471 }, //14 May 1993, UT1-UTC=-0.2954714, Accumulated Leap Seconds=27
  { 2449122.50, 59.482027 }, //15 May 1993, UT1-UTC=-0.2980267, Accumulated Leap Seconds=27
  { 2449123.50, 59.484616 }, //16 May 1993, UT1-UTC=-0.3006161, Accumulated Leap Seconds=27
  { 2449124.50, 59.487207 }, //17 May 1993, UT1-UTC=-0.3032068, Accumulated Leap Seconds=27
  { 2449125.50, 59.489762 }, //18 May 1993, UT1-UTC=-0.3057621, Accumulated Leap Seconds=27
  { 2449126.50, 59.492249 }, //19 May 1993, UT1-UTC=-0.3082495, Accumulated Leap Seconds=27
  { 2449127.50, 59.494638 }, //20 May 1993, UT1-UTC=-0.3106377, Accumulated Leap Seconds=27
  { 2449128.50, 59.496913 }, //21 May 1993, UT1-UTC=-0.3129125, Accumulated Leap Seconds=27
  { 2449129.50, 59.499091 }, //22 May 1993, UT1-UTC=-0.3150909, Accumulated Leap Seconds=27
  { 2449130.50, 59.501211 }, //23 May 1993, UT1-UTC=-0.3172109, Accumulated Leap Seconds=27
  { 2449131.50, 59.503335 }, //24 May 1993, UT1-UTC=-0.3193351, Accumulated Leap Seconds=27
  { 2449132.50, 59.505536 }, //25 May 1993, UT1-UTC=-0.3215356, Accumulated Leap Seconds=27
  { 2449133.50, 59.507875 }, //26 May 1993, UT1-UTC=-0.3238752, Accumulated Leap Seconds=27
  { 2449134.50, 59.510383 }, //27 May 1993, UT1-UTC=-0.3263831, Accumulated Leap Seconds=27
  { 2449135.50, 59.513059 }, //28 May 1993, UT1-UTC=-0.3290589, Accumulated Leap Seconds=27
  { 2449136.50, 59.515865 }, //29 May 1993, UT1-UTC=-0.3318646, Accumulated Leap Seconds=27
  { 2449137.50, 59.518726 }, //30 May 1993, UT1-UTC=-0.3347260, Accumulated Leap Seconds=27
  { 2449138.50, 59.521549 }, //31 May 1993, UT1-UTC=-0.3375493, Accumulated Leap Seconds=27
  { 2449139.50, 59.524242 }, //01 June 1993, UT1-UTC=-0.3402416, Accumulated Leap Seconds=27
  { 2449140.50, 59.526731 }, //02 June 1993, UT1-UTC=-0.3427310, Accumulated Leap Seconds=27
  { 2449141.50, 59.528985 }, //03 June 1993, UT1-UTC=-0.3449852, Accumulated Leap Seconds=27
  { 2449142.50, 59.531031 }, //04 June 1993, UT1-UTC=-0.3470306, Accumulated Leap Seconds=27
  { 2449143.50, 59.532937 }, //05 June 1993, UT1-UTC=-0.3489374, Accumulated Leap Seconds=27
  { 2449144.50, 59.534793 }, //06 June 1993, UT1-UTC=-0.3507926, Accumulated Leap Seconds=27
  { 2449145.50, 59.536675 }, //07 June 1993, UT1-UTC=-0.3526748, Accumulated Leap Seconds=27
  { 2449146.50, 59.538636 }, //08 June 1993, UT1-UTC=-0.3546356, Accumulated Leap Seconds=27
  { 2449147.50, 59.540693 }, //09 June 1993, UT1-UTC=-0.3566934, Accumulated Leap Seconds=27
  { 2449148.50, 59.542832 }, //10 June 1993, UT1-UTC=-0.3588322, Accumulated Leap Seconds=27
  { 2449149.50, 59.545040 }, //11 June 1993, UT1-UTC=-0.3610400, Accumulated Leap Seconds=27
  { 2449150.50, 59.547286 }, //12 June 1993, UT1-UTC=-0.3632859, Accumulated Leap Seconds=27
  { 2449151.50, 59.549527 }, //13 June 1993, UT1-UTC=-0.3655270, Accumulated Leap Seconds=27
  { 2449152.50, 59.551718 }, //14 June 1993, UT1-UTC=-0.3677184, Accumulated Leap Seconds=27
  { 2449153.50, 59.553820 }, //15 June 1993, UT1-UTC=-0.3698199, Accumulated Leap Seconds=27
  { 2449154.50, 59.555803 }, //16 June 1993, UT1-UTC=-0.3718032, Accumulated Leap Seconds=27
  { 2449155.50, 59.557667 }, //17 June 1993, UT1-UTC=-0.3736668, Accumulated Leap Seconds=27
  { 2449156.50, 59.559432 }, //18 June 1993, UT1-UTC=-0.3754318, Accumulated Leap Seconds=27
  { 2449157.50, 59.561128 }, //19 June 1993, UT1-UTC=-0.3771281, Accumulated Leap Seconds=27
  { 2449158.50, 59.562809 }, //20 June 1993, UT1-UTC=-0.3788093, Accumulated Leap Seconds=27
  { 2449159.50, 59.564545 }, //21 June 1993, UT1-UTC=-0.3805452, Accumulated Leap Seconds=27
  { 2449160.50, 59.566401 }, //22 June 1993, UT1-UTC=-0.3824011, Accumulated Leap Seconds=27
  { 2449161.50, 59.568417 }, //23 June 1993, UT1-UTC=-0.3844170, Accumulated Leap Seconds=27
  { 2449162.50, 59.570583 }, //24 June 1993, UT1-UTC=-0.3865828, Accumulated Leap Seconds=27
  { 2449163.50, 59.572856 }, //25 June 1993, UT1-UTC=-0.3888564, Accumulated Leap Seconds=27
  { 2449164.50, 59.575178 }, //26 June 1993, UT1-UTC=-0.3911783, Accumulated Leap Seconds=27
  { 2449165.50, 59.577470 }, //27 June 1993, UT1-UTC=-0.3934703, Accumulated Leap Seconds=27
  { 2449166.50, 59.579643 }, //28 June 1993, UT1-UTC=-0.3956430, Accumulated Leap Seconds=27
  { 2449167.50, 59.581627 }, //29 June 1993, UT1-UTC=-0.3976266, Accumulated Leap Seconds=27
  { 2449168.50, 59.583395 }, //30 June 1993, UT1-UTC=-0.3993948, Accumulated Leap Seconds=27
  { 2449169.50, 59.584979 }, //01 July 1993, UT1-UTC= 0.5990213, Accumulated Leap Seconds=28
  { 2449170.50, 59.586433 }, //02 July 1993, UT1-UTC= 0.5975665, Accumulated Leap Seconds=28
  { 2449171.50, 59.587826 }, //03 July 1993, UT1-UTC= 0.5961744, Accumulated Leap Seconds=28
  { 2449172.50, 59.589219 }, //04 July 1993, UT1-UTC= 0.5947809, Accumulated Leap Seconds=28
  { 2449173.50, 59.590660 }, //05 July 1993, UT1-UTC= 0.5933398, Accumulated Leap Seconds=28
  { 2449174.50, 59.592171 }, //06 July 1993, UT1-UTC= 0.5918288, Accumulated Leap Seconds=28
  { 2449175.50, 59.593749 }, //07 July 1993, UT1-UTC= 0.5902506, Accumulated Leap Seconds=28
  { 2449176.50, 59.595372 }, //08 July 1993, UT1-UTC= 0.5886275, Accumulated Leap Seconds=28
  { 2449177.50, 59.597007 }, //09 July 1993, UT1-UTC= 0.5869933, Accumulated Leap Seconds=28
  { 2449178.50, 59.598621 }, //10 July 1993, UT1-UTC= 0.5853792, Accumulated Leap Seconds=28
  { 2449179.50, 59.600191 }, //11 July 1993, UT1-UTC= 0.5838086, Accumulated Leap Seconds=28
  { 2449180.50, 59.601703 }, //12 July 1993, UT1-UTC= 0.5822970, Accumulated Leap Seconds=28
  { 2449181.50, 59.603150 }, //13 July 1993, UT1-UTC= 0.5808503, Accumulated Leap Seconds=28
  { 2449182.50, 59.604539 }, //14 July 1993, UT1-UTC= 0.5794606, Accumulated Leap Seconds=28
  { 2449183.50, 59.605880 }, //15 July 1993, UT1-UTC= 0.5781205, Accumulated Leap Seconds=28
  { 2449184.50, 59.607164 }, //16 July 1993, UT1-UTC= 0.5768360, Accumulated Leap Seconds=28
  { 2449185.50, 59.608417 }, //17 July 1993, UT1-UTC= 0.5755834, Accumulated Leap Seconds=28
  { 2449186.50, 59.609694 }, //18 July 1993, UT1-UTC= 0.5743061, Accumulated Leap Seconds=28
  { 2449187.50, 59.611081 }, //19 July 1993, UT1-UTC= 0.5729185, Accumulated Leap Seconds=28
  { 2449188.50, 59.612656 }, //20 July 1993, UT1-UTC= 0.5713440, Accumulated Leap Seconds=28
  { 2449189.50, 59.614461 }, //21 July 1993, UT1-UTC= 0.5695386, Accumulated Leap Seconds=28
  { 2449190.50, 59.616505 }, //22 July 1993, UT1-UTC= 0.5674946, Accumulated Leap Seconds=28
  { 2449191.50, 59.618713 }, //23 July 1993, UT1-UTC= 0.5652874, Accumulated Leap Seconds=28
  { 2449192.50, 59.620957 }, //24 July 1993, UT1-UTC= 0.5630427, Accumulated Leap Seconds=28
  { 2449193.50, 59.623117 }, //25 July 1993, UT1-UTC= 0.5608826, Accumulated Leap Seconds=28
  { 2449194.50, 59.625106 }, //26 July 1993, UT1-UTC= 0.5588936, Accumulated Leap Seconds=28
  { 2449195.50, 59.626888 }, //27 July 1993, UT1-UTC= 0.5571115, Accumulated Leap Seconds=28
  { 2449196.50, 59.628479 }, //28 July 1993, UT1-UTC= 0.5555207, Accumulated Leap Seconds=28
  { 2449197.50, 59.629943 }, //29 July 1993, UT1-UTC= 0.5540573, Accumulated Leap Seconds=28
  { 2449198.50, 59.631371 }, //30 July 1993, UT1-UTC= 0.5526292, Accumulated Leap Seconds=28
  { 2449199.50, 59.632822 }, //31 July 1993, UT1-UTC= 0.5511775, Accumulated Leap Seconds=28
  { 2449200.50, 59.634350 }, //01 August 1993, UT1-UTC= 0.5496503, Accumulated Leap Seconds=28
  { 2449201.50, 59.635985 }, //02 August 1993, UT1-UTC= 0.5480148, Accumulated Leap Seconds=28
  { 2449202.50, 59.637739 }, //03 August 1993, UT1-UTC= 0.5462606, Accumulated Leap Seconds=28
  { 2449203.50, 59.639604 }, //04 August 1993, UT1-UTC= 0.5443964, Accumulated Leap Seconds=28
  { 2449204.50, 59.641560 }, //05 August 1993, UT1-UTC= 0.5424400, Accumulated Leap Seconds=28
  { 2449205.50, 59.643567 }, //06 August 1993, UT1-UTC= 0.5404326, Accumulated Leap Seconds=28
  { 2449206.50, 59.645557 }, //07 August 1993, UT1-UTC= 0.5384428, Accumulated Leap Seconds=28
  { 2449207.50, 59.647478 }, //08 August 1993, UT1-UTC= 0.5365217, Accumulated Leap Seconds=28
  { 2449208.50, 59.649296 }, //09 August 1993, UT1-UTC= 0.5347039, Accumulated Leap Seconds=28
  { 2449209.50, 59.650991 }, //10 August 1993, UT1-UTC= 0.5330091, Accumulated Leap Seconds=28
  { 2449210.50, 59.652561 }, //11 August 1993, UT1-UTC= 0.5314394, Accumulated Leap Seconds=28
  { 2449211.50, 59.654042 }, //12 August 1993, UT1-UTC= 0.5299581, Accumulated Leap Seconds=28
  { 2449212.50, 59.655495 }, //13 August 1993, UT1-UTC= 0.5285050, Accumulated Leap Seconds=28
  { 2449213.50, 59.656991 }, //14 August 1993, UT1-UTC= 0.5270087, Accumulated Leap Seconds=28
  { 2449214.50, 59.658612 }, //15 August 1993, UT1-UTC= 0.5253878, Accumulated Leap Seconds=28
  { 2449215.50, 59.660427 }, //16 August 1993, UT1-UTC= 0.5235733, Accumulated Leap Seconds=28
  { 2449216.50, 59.662473 }, //17 August 1993, UT1-UTC= 0.5215268, Accumulated Leap Seconds=28
  { 2449217.50, 59.664748 }, //18 August 1993, UT1-UTC= 0.5192517, Accumulated Leap Seconds=28
  { 2449218.50, 59.667214 }, //19 August 1993, UT1-UTC= 0.5167860, Accumulated Leap Seconds=28
  { 2449219.50, 59.669782 }, //20 August 1993, UT1-UTC= 0.5142177, Accumulated Leap Seconds=28
  { 2449220.50, 59.672327 }, //21 August 1993, UT1-UTC= 0.5116731, Accumulated Leap Seconds=28
  { 2449221.50, 59.674731 }, //22 August 1993, UT1-UTC= 0.5092691, Accumulated Leap Seconds=28
  { 2449222.50, 59.676925 }, //23 August 1993, UT1-UTC= 0.5070750, Accumulated Leap Seconds=28
  { 2449223.50, 59.678894 }, //24 August 1993, UT1-UTC= 0.5051055, Accumulated Leap Seconds=28
  { 2449224.50, 59.680669 }, //25 August 1993, UT1-UTC= 0.5033306, Accumulated Leap Seconds=28
  { 2449225.50, 59.682312 }, //26 August 1993, UT1-UTC= 0.5016878, Accumulated Leap Seconds=28
  { 2449226.50, 59.683893 }, //27 August 1993, UT1-UTC= 0.5001067, Accumulated Leap Seconds=28
  { 2449227.50, 59.685486 }, //28 August 1993, UT1-UTC= 0.4985141, Accumulated Leap Seconds=28
  { 2449228.50, 59.687147 }, //29 August 1993, UT1-UTC= 0.4968532, Accumulated Leap Seconds=28
  { 2449229.50, 59.688911 }, //30 August 1993, UT1-UTC= 0.4950890, Accumulated Leap Seconds=28
  { 2449230.50, 59.690792 }, //31 August 1993, UT1-UTC= 0.4932083, Accumulated Leap Seconds=28
  { 2449231.50, 59.692783 }, //01 September 1993, UT1-UTC= 0.4912173, Accumulated Leap Seconds=28
  { 2449232.50, 59.694857 }, //02 September 1993, UT1-UTC= 0.4891434, Accumulated Leap Seconds=28
  { 2449233.50, 59.696960 }, //03 September 1993, UT1-UTC= 0.4870396, Accumulated Leap Seconds=28
  { 2449234.50, 59.699041 }, //04 September 1993, UT1-UTC= 0.4849585, Accumulated Leap Seconds=28
  { 2449235.50, 59.701060 }, //05 September 1993, UT1-UTC= 0.4829399, Accumulated Leap Seconds=28
  { 2449236.50, 59.702994 }, //06 September 1993, UT1-UTC= 0.4810059, Accumulated Leap Seconds=28
  { 2449237.50, 59.704842 }, //07 September 1993, UT1-UTC= 0.4791580, Accumulated Leap Seconds=28
  { 2449238.50, 59.706618 }, //08 September 1993, UT1-UTC= 0.4773818, Accumulated Leap Seconds=28
  { 2449239.50, 59.708344 }, //09 September 1993, UT1-UTC= 0.4756563, Accumulated Leap Seconds=28
  { 2449240.50, 59.710073 }, //10 September 1993, UT1-UTC= 0.4739272, Accumulated Leap Seconds=28
  { 2449241.50, 59.711881 }, //11 September 1993, UT1-UTC= 0.4721191, Accumulated Leap Seconds=28
  { 2449242.50, 59.713848 }, //12 September 1993, UT1-UTC= 0.4701520, Accumulated Leap Seconds=28
  { 2449243.50, 59.716042 }, //13 September 1993, UT1-UTC= 0.4679584, Accumulated Leap Seconds=28
  { 2449244.50, 59.718490 }, //14 September 1993, UT1-UTC= 0.4655095, Accumulated Leap Seconds=28
  { 2449245.50, 59.721168 }, //15 September 1993, UT1-UTC= 0.4628317, Accumulated Leap Seconds=28
  { 2449246.50, 59.723987 }, //16 September 1993, UT1-UTC= 0.4600133, Accumulated Leap Seconds=28
  { 2449247.50, 59.726815 }, //17 September 1993, UT1-UTC= 0.4571853, Accumulated Leap Seconds=28
  { 2449248.50, 59.729523 }, //18 September 1993, UT1-UTC= 0.4544769, Accumulated Leap Seconds=28
  { 2449249.50, 59.732029 }, //19 September 1993, UT1-UTC= 0.4519709, Accumulated Leap Seconds=28
  { 2449250.50, 59.734317 }, //20 September 1993, UT1-UTC= 0.4496832, Accumulated Leap Seconds=28
  { 2449251.50, 59.736434 }, //21 September 1993, UT1-UTC= 0.4475661, Accumulated Leap Seconds=28
  { 2449252.50, 59.738467 }, //22 September 1993, UT1-UTC= 0.4455329, Accumulated Leap Seconds=28
  { 2449253.50, 59.740492 }, //23 September 1993, UT1-UTC= 0.4435076, Accumulated Leap Seconds=28
  { 2449254.50, 59.742556 }, //24 September 1993, UT1-UTC= 0.4414444, Accumulated Leap Seconds=28
  { 2449255.50, 59.744698 }, //25 September 1993, UT1-UTC= 0.4393015, Accumulated Leap Seconds=28
  { 2449256.50, 59.746936 }, //26 September 1993, UT1-UTC= 0.4370639, Accumulated Leap Seconds=28
  { 2449257.50, 59.749262 }, //27 September 1993, UT1-UTC= 0.4347381, Accumulated Leap Seconds=28
  { 2449258.50, 59.751652 }, //28 September 1993, UT1-UTC= 0.4323477, Accumulated Leap Seconds=28
  { 2449259.50, 59.754069 }, //29 September 1993, UT1-UTC= 0.4299313, Accumulated Leap Seconds=28
  { 2449260.50, 59.756461 }, //30 September 1993, UT1-UTC= 0.4275388, Accumulated Leap Seconds=28
  { 2449261.50, 59.758805 }, //01 October 1993, UT1-UTC= 0.4251950, Accumulated Leap Seconds=28
  { 2449262.50, 59.761087 }, //02 October 1993, UT1-UTC= 0.4229131, Accumulated Leap Seconds=28
  { 2449263.50, 59.763298 }, //03 October 1993, UT1-UTC= 0.4207018, Accumulated Leap Seconds=28
  { 2449264.50, 59.765446 }, //04 October 1993, UT1-UTC= 0.4185543, Accumulated Leap Seconds=28
  { 2449265.50, 59.767553 }, //05 October 1993, UT1-UTC= 0.4164466, Accumulated Leap Seconds=28
  { 2449266.50, 59.769660 }, //06 October 1993, UT1-UTC= 0.4143404, Accumulated Leap Seconds=28
  { 2449267.50, 59.771806 }, //07 October 1993, UT1-UTC= 0.4121937, Accumulated Leap Seconds=28
  { 2449268.50, 59.774016 }, //08 October 1993, UT1-UTC= 0.4099838, Accumulated Leap Seconds=28
  { 2449269.50, 59.776337 }, //09 October 1993, UT1-UTC= 0.4076625, Accumulated Leap Seconds=28
  { 2449270.50, 59.778832 }, //10 October 1993, UT1-UTC= 0.4051683, Accumulated Leap Seconds=28
  { 2449271.50, 59.781548 }, //11 October 1993, UT1-UTC= 0.4024520, Accumulated Leap Seconds=28
  { 2449272.50, 59.784500 }, //12 October 1993, UT1-UTC= 0.3995004, Accumulated Leap Seconds=28
  { 2449273.50, 59.787653 }, //13 October 1993, UT1-UTC= 0.3963465, Accumulated Leap Seconds=28
  { 2449274.50, 59.790935 }, //14 October 1993, UT1-UTC= 0.3930652, Accumulated Leap Seconds=28
  { 2449275.50, 59.794199 }, //15 October 1993, UT1-UTC= 0.3898006, Accumulated Leap Seconds=28
  { 2449276.50, 59.797312 }, //16 October 1993, UT1-UTC= 0.3866883, Accumulated Leap Seconds=28
  { 2449277.50, 59.800194 }, //17 October 1993, UT1-UTC= 0.3838059, Accumulated Leap Seconds=28
  { 2449278.50, 59.802840 }, //18 October 1993, UT1-UTC= 0.3811603, Accumulated Leap Seconds=28
  { 2449279.50, 59.805301 }, //19 October 1993, UT1-UTC= 0.3786989, Accumulated Leap Seconds=28
  { 2449280.50, 59.807662 }, //20 October 1993, UT1-UTC= 0.3763377, Accumulated Leap Seconds=28
  { 2449281.50, 59.810016 }, //21 October 1993, UT1-UTC= 0.3739835, Accumulated Leap Seconds=28
  { 2449282.50, 59.812438 }, //22 October 1993, UT1-UTC= 0.3715615, Accumulated Leap Seconds=28
  { 2449283.50, 59.814956 }, //23 October 1993, UT1-UTC= 0.3690436, Accumulated Leap Seconds=28
  { 2449284.50, 59.817569 }, //24 October 1993, UT1-UTC= 0.3664313, Accumulated Leap Seconds=28
  { 2449285.50, 59.820259 }, //25 October 1993, UT1-UTC= 0.3637412, Accumulated Leap Seconds=28
  { 2449286.50, 59.823002 }, //26 October 1993, UT1-UTC= 0.3609983, Accumulated Leap Seconds=28
  { 2449287.50, 59.825765 }, //27 October 1993, UT1-UTC= 0.3582351, Accumulated Leap Seconds=28
  { 2449288.50, 59.828508 }, //28 October 1993, UT1-UTC= 0.3554925, Accumulated Leap Seconds=28
  { 2449289.50, 59.831189 }, //29 October 1993, UT1-UTC= 0.3528112, Accumulated Leap Seconds=28
  { 2449290.50, 59.833781 }, //30 October 1993, UT1-UTC= 0.3502189, Accumulated Leap Seconds=28
  { 2449291.50, 59.836268 }, //31 October 1993, UT1-UTC= 0.3477323, Accumulated Leap Seconds=28
  { 2449292.50, 59.838645 }, //01 November 1993, UT1-UTC= 0.3453552, Accumulated Leap Seconds=28
  { 2449293.50, 59.840925 }, //02 November 1993, UT1-UTC= 0.3430746, Accumulated Leap Seconds=28
  { 2449294.50, 59.843139 }, //03 November 1993, UT1-UTC= 0.3408610, Accumulated Leap Seconds=28
  { 2449295.50, 59.845319 }, //04 November 1993, UT1-UTC= 0.3386812, Accumulated Leap Seconds=28
  { 2449296.50, 59.847523 }, //05 November 1993, UT1-UTC= 0.3364773, Accumulated Leap Seconds=28
  { 2449297.50, 59.849834 }, //06 November 1993, UT1-UTC= 0.3341663, Accumulated Leap Seconds=28
  { 2449298.50, 59.852320 }, //07 November 1993, UT1-UTC= 0.3316799, Accumulated Leap Seconds=28
  { 2449299.50, 59.855020 }, //08 November 1993, UT1-UTC= 0.3289803, Accumulated Leap Seconds=28
  { 2449300.50, 59.857926 }, //09 November 1993, UT1-UTC= 0.3260735, Accumulated Leap Seconds=28
  { 2449301.50, 59.860985 }, //10 November 1993, UT1-UTC= 0.3230149, Accumulated Leap Seconds=28
  { 2449302.50, 59.864090 }, //11 November 1993, UT1-UTC= 0.3199099, Accumulated Leap Seconds=28
  { 2449303.50, 59.867118 }, //12 November 1993, UT1-UTC= 0.3168824, Accumulated Leap Seconds=28
  { 2449304.50, 59.869974 }, //13 November 1993, UT1-UTC= 0.3140265, Accumulated Leap Seconds=28
  { 2449305.50, 59.872614 }, //14 November 1993, UT1-UTC= 0.3113862, Accumulated Leap Seconds=28
  { 2449306.50, 59.875054 }, //15 November 1993, UT1-UTC= 0.3089460, Accumulated Leap Seconds=28
  { 2449307.50, 59.877356 }, //16 November 1993, UT1-UTC= 0.3066442, Accumulated Leap Seconds=28
  { 2449308.50, 59.879596 }, //17 November 1993, UT1-UTC= 0.3044037, Accumulated Leap Seconds=28
  { 2449309.50, 59.881841 }, //18 November 1993, UT1-UTC= 0.3021586, Accumulated Leap Seconds=28
  { 2449310.50, 59.884137 }, //19 November 1993, UT1-UTC= 0.2998625, Accumulated Leap Seconds=28
  { 2449311.50, 59.886501 }, //20 November 1993, UT1-UTC= 0.2974989, Accumulated Leap Seconds=28
  { 2449312.50, 59.888926 }, //21 November 1993, UT1-UTC= 0.2950742, Accumulated Leap Seconds=28
  { 2449313.50, 59.891392 }, //22 November 1993, UT1-UTC= 0.2926080, Accumulated Leap Seconds=28
  { 2449314.50, 59.893874 }, //23 November 1993, UT1-UTC= 0.2901262, Accumulated Leap Seconds=28
  { 2449315.50, 59.896341 }, //24 November 1993, UT1-UTC= 0.2876592, Accumulated Leap Seconds=28
  { 2449316.50, 59.898757 }, //25 November 1993, UT1-UTC= 0.2852429, Accumulated Leap Seconds=28
  { 2449317.50, 59.901084 }, //26 November 1993, UT1-UTC= 0.2829157, Accumulated Leap Seconds=28
  { 2449318.50, 59.903303 }, //27 November 1993, UT1-UTC= 0.2806969, Accumulated Leap Seconds=28
  { 2449319.50, 59.905405 }, //28 November 1993, UT1-UTC= 0.2785951, Accumulated Leap Seconds=28
  { 2449320.50, 59.907390 }, //29 November 1993, UT1-UTC= 0.2766102, Accumulated Leap Seconds=28
  { 2449321.50, 59.909263 }, //30 November 1993, UT1-UTC= 0.2747371, Accumulated Leap Seconds=28
  { 2449322.50, 59.911057 }, //01 December 1993, UT1-UTC= 0.2729433, Accumulated Leap Seconds=28
  { 2449323.50, 59.912869 }, //02 December 1993, UT1-UTC= 0.2711311, Accumulated Leap Seconds=28
  { 2449324.50, 59.914754 }, //03 December 1993, UT1-UTC= 0.2692456, Accumulated Leap Seconds=28
  { 2449325.50, 59.916776 }, //04 December 1993, UT1-UTC= 0.2672240, Accumulated Leap Seconds=28
  { 2449326.50, 59.918980 }, //05 December 1993, UT1-UTC= 0.2650202, Accumulated Leap Seconds=28
  { 2449327.50, 59.921380 }, //06 December 1993, UT1-UTC= 0.2626204, Accumulated Leap Seconds=28
  { 2449328.50, 59.923955 }, //07 December 1993, UT1-UTC= 0.2600449, Accumulated Leap Seconds=28
  { 2449329.50, 59.926652 }, //08 December 1993, UT1-UTC= 0.2573477, Accumulated Leap Seconds=28
  { 2449330.50, 59.929369 }, //09 December 1993, UT1-UTC= 0.2546311, Accumulated Leap Seconds=28
  { 2449331.50, 59.932012 }, //10 December 1993, UT1-UTC= 0.2519876, Accumulated Leap Seconds=28
  { 2449332.50, 59.934521 }, //11 December 1993, UT1-UTC= 0.2494785, Accumulated Leap Seconds=28
  { 2449333.50, 59.936882 }, //12 December 1993, UT1-UTC= 0.2471184, Accumulated Leap Seconds=28
  { 2449334.50, 59.939125 }, //13 December 1993, UT1-UTC= 0.2448753, Accumulated Leap Seconds=28
  { 2449335.50, 59.941322 }, //14 December 1993, UT1-UTC= 0.2426777, Accumulated Leap Seconds=28
  { 2449336.50, 59.943556 }, //15 December 1993, UT1-UTC= 0.2404437, Accumulated Leap Seconds=28
  { 2449337.50, 59.945899 }, //16 December 1993, UT1-UTC= 0.2381006, Accumulated Leap Seconds=28
  { 2449338.50, 59.948367 }, //17 December 1993, UT1-UTC= 0.2356329, Accumulated Leap Seconds=28
  { 2449339.50, 59.950925 }, //18 December 1993, UT1-UTC= 0.2330752, Accumulated Leap Seconds=28
  { 2449340.50, 59.953537 }, //19 December 1993, UT1-UTC= 0.2304630, Accumulated Leap Seconds=28
  { 2449341.50, 59.956174 }, //20 December 1993, UT1-UTC= 0.2278256, Accumulated Leap Seconds=28
  { 2449342.50, 59.958817 }, //21 December 1993, UT1-UTC= 0.2251828, Accumulated Leap Seconds=28
  { 2449343.50, 59.961440 }, //22 December 1993, UT1-UTC= 0.2225602, Accumulated Leap Seconds=28
  { 2449344.50, 59.963975 }, //23 December 1993, UT1-UTC= 0.2200253, Accumulated Leap Seconds=28
  { 2449345.50, 59.966403 }, //24 December 1993, UT1-UTC= 0.2175975, Accumulated Leap Seconds=28
  { 2449346.50, 59.968730 }, //25 December 1993, UT1-UTC= 0.2152701, Accumulated Leap Seconds=28
  { 2449347.50, 59.970961 }, //26 December 1993, UT1-UTC= 0.2130385, Accumulated Leap Seconds=28
  { 2449348.50, 59.973116 }, //27 December 1993, UT1-UTC= 0.2108839, Accumulated Leap Seconds=28
  { 2449349.50, 59.975229 }, //28 December 1993, UT1-UTC= 0.2087707, Accumulated Leap Seconds=28
  { 2449350.50, 59.977352 }, //29 December 1993, UT1-UTC= 0.2066483, Accumulated Leap Seconds=28
  { 2449351.50, 59.979548 }, //30 December 1993, UT1-UTC= 0.2044522, Accumulated Leap Seconds=28
  { 2449352.50, 59.981908 }, //31 December 1993, UT1-UTC= 0.2020922, Accumulated Leap Seconds=28
  { 2449353.50, 59.984454 }, //01 January 1994, UT1-UTC= 0.1995463, Accumulated Leap Seconds=28
  { 2449354.50, 59.987158 }, //02 January 1994, UT1-UTC= 0.1968423, Accumulated Leap Seconds=28
  { 2449355.50, 59.989957 }, //03 January 1994, UT1-UTC= 0.1940431, Accumulated Leap Seconds=28
  { 2449356.50, 59.992764 }, //04 January 1994, UT1-UTC= 0.1912364, Accumulated Leap Seconds=28
  { 2449357.50, 59.995486 }, //05 January 1994, UT1-UTC= 0.1885145, Accumulated Leap Seconds=28
  { 2449358.50, 59.998057 }, //06 January 1994, UT1-UTC= 0.1859434, Accumulated Leap Seconds=28
  { 2449359.50, 60.000451 }, //07 January 1994, UT1-UTC= 0.1835491, Accumulated Leap Seconds=28
  { 2449360.50, 60.002664 }, //08 January 1994, UT1-UTC= 0.1813362, Accumulated Leap Seconds=28
  { 2449361.50, 60.004731 }, //09 January 1994, UT1-UTC= 0.1792693, Accumulated Leap Seconds=28
  { 2449362.50, 60.006724 }, //10 January 1994, UT1-UTC= 0.1772761, Accumulated Leap Seconds=28
  { 2449363.50, 60.008717 }, //11 January 1994, UT1-UTC= 0.1752829, Accumulated Leap Seconds=28
  { 2449364.50, 60.010768 }, //12 January 1994, UT1-UTC= 0.1732318, Accumulated Leap Seconds=28
  { 2449365.50, 60.012945 }, //13 January 1994, UT1-UTC= 0.1710547, Accumulated Leap Seconds=28
  { 2449366.50, 60.015256 }, //14 January 1994, UT1-UTC= 0.1687436, Accumulated Leap Seconds=28
  { 2449367.50, 60.017631 }, //15 January 1994, UT1-UTC= 0.1663686, Accumulated Leap Seconds=28
  { 2449368.50, 60.019976 }, //16 January 1994, UT1-UTC= 0.1640235, Accumulated Leap Seconds=28
  { 2449369.50, 60.022268 }, //17 January 1994, UT1-UTC= 0.1617322, Accumulated Leap Seconds=28
  { 2449370.50, 60.024519 }, //18 January 1994, UT1-UTC= 0.1594805, Accumulated Leap Seconds=28
  { 2449371.50, 60.026707 }, //19 January 1994, UT1-UTC= 0.1572930, Accumulated Leap Seconds=28
  { 2449372.50, 60.028782 }, //20 January 1994, UT1-UTC= 0.1552178, Accumulated Leap Seconds=28
  { 2449373.50, 60.030772 }, //21 January 1994, UT1-UTC= 0.1532284, Accumulated Leap Seconds=28
  { 2449374.50, 60.032727 }, //22 January 1994, UT1-UTC= 0.1512732, Accumulated Leap Seconds=28
  { 2449375.50, 60.034655 }, //23 January 1994, UT1-UTC= 0.1493454, Accumulated Leap Seconds=28
  { 2449376.50, 60.036585 }, //24 January 1994, UT1-UTC= 0.1474146, Accumulated Leap Seconds=28
  { 2449377.50, 60.038554 }, //25 January 1994, UT1-UTC= 0.1454455, Accumulated Leap Seconds=28
  { 2449378.50, 60.040633 }, //26 January 1994, UT1-UTC= 0.1433675, Accumulated Leap Seconds=28
  { 2449379.50, 60.042906 }, //27 January 1994, UT1-UTC= 0.1410944, Accumulated Leap Seconds=28
  { 2449380.50, 60.045391 }, //28 January 1994, UT1-UTC= 0.1386086, Accumulated Leap Seconds=28
  { 2449381.50, 60.048059 }, //29 January 1994, UT1-UTC= 0.1359406, Accumulated Leap Seconds=28
  { 2449382.50, 60.050851 }, //30 January 1994, UT1-UTC= 0.1331486, Accumulated Leap Seconds=28
  { 2449383.50, 60.053676 }, //31 January 1994, UT1-UTC= 0.1303237, Accumulated Leap Seconds=28
  { 2449384.50, 60.056435 }, //01 February 1994, UT1-UTC= 0.1275650, Accumulated Leap Seconds=28
  { 2449385.50, 60.059035 }, //02 February 1994, UT1-UTC= 0.1249650, Accumulated Leap Seconds=28
  { 2449386.50, 60.061430 }, //03 February 1994, UT1-UTC= 0.1225700, Accumulated Leap Seconds=28
  { 2449387.50, 60.063630 }, //04 February 1994, UT1-UTC= 0.1203695, Accumulated Leap Seconds=28
  { 2449388.50, 60.065670 }, //05 February 1994, UT1-UTC= 0.1183296, Accumulated Leap Seconds=28
  { 2449389.50, 60.067608 }, //06 February 1994, UT1-UTC= 0.1163916, Accumulated Leap Seconds=28
  { 2449390.50, 60.069524 }, //07 February 1994, UT1-UTC= 0.1144762, Accumulated Leap Seconds=28
  { 2449391.50, 60.071500 }, //08 February 1994, UT1-UTC= 0.1125004, Accumulated Leap Seconds=28
  { 2449392.50, 60.073581 }, //09 February 1994, UT1-UTC= 0.1104190, Accumulated Leap Seconds=28
  { 2449393.50, 60.075772 }, //10 February 1994, UT1-UTC= 0.1082275, Accumulated Leap Seconds=28
  { 2449394.50, 60.078056 }, //11 February 1994, UT1-UTC= 0.1059438, Accumulated Leap Seconds=28
  { 2449395.50, 60.080392 }, //12 February 1994, UT1-UTC= 0.1036081, Accumulated Leap Seconds=28
  { 2449396.50, 60.082745 }, //13 February 1994, UT1-UTC= 0.1012548, Accumulated Leap Seconds=28
  { 2449397.50, 60.085089 }, //14 February 1994, UT1-UTC= 0.0989113, Accumulated Leap Seconds=28
  { 2449398.50, 60.087391 }, //15 February 1994, UT1-UTC= 0.0966092, Accumulated Leap Seconds=28
  { 2449399.50, 60.089620 }, //16 February 1994, UT1-UTC= 0.0943799, Accumulated Leap Seconds=28
  { 2449400.50, 60.091761 }, //17 February 1994, UT1-UTC= 0.0922394, Accumulated Leap Seconds=28
  { 2449401.50, 60.093832 }, //18 February 1994, UT1-UTC= 0.0901679, Accumulated Leap Seconds=28
  { 2449402.50, 60.095856 }, //19 February 1994, UT1-UTC= 0.0881438, Accumulated Leap Seconds=28
  { 2449403.50, 60.097868 }, //20 February 1994, UT1-UTC= 0.0861319, Accumulated Leap Seconds=28
  { 2449404.50, 60.099928 }, //21 February 1994, UT1-UTC= 0.0840724, Accumulated Leap Seconds=28
  { 2449405.50, 60.102100 }, //22 February 1994, UT1-UTC= 0.0818996, Accumulated Leap Seconds=28
  { 2449406.50, 60.104453 }, //23 February 1994, UT1-UTC= 0.0795473, Accumulated Leap Seconds=28
  { 2449407.50, 60.107055 }, //24 February 1994, UT1-UTC= 0.0769445, Accumulated Leap Seconds=28
  { 2449408.50, 60.109938 }, //25 February 1994, UT1-UTC= 0.0740622, Accumulated Leap Seconds=28
  { 2449409.50, 60.113072 }, //26 February 1994, UT1-UTC= 0.0709277, Accumulated Leap Seconds=28
  { 2449410.50, 60.116382 }, //27 February 1994, UT1-UTC= 0.0676180, Accumulated Leap Seconds=28
  { 2449411.50, 60.119753 }, //28 February 1994, UT1-UTC= 0.0642465, Accumulated Leap Seconds=28
  { 2449412.50, 60.123065 }, //01 March 1994, UT1-UTC= 0.0609350, Accumulated Leap Seconds=28
  { 2449413.50, 60.126225 }, //02 March 1994, UT1-UTC= 0.0577749, Accumulated Leap Seconds=28
  { 2449414.50, 60.129196 }, //03 March 1994, UT1-UTC= 0.0548037, Accumulated Leap Seconds=28
  { 2449415.50, 60.131976 }, //04 March 1994, UT1-UTC= 0.0520238, Accumulated Leap Seconds=28
  { 2449416.50, 60.134591 }, //05 March 1994, UT1-UTC= 0.0494091, Accumulated Leap Seconds=28
  { 2449417.50, 60.137097 }, //06 March 1994, UT1-UTC= 0.0469026, Accumulated Leap Seconds=28
  { 2449418.50, 60.139573 }, //07 March 1994, UT1-UTC= 0.0444265, Accumulated Leap Seconds=28
  { 2449419.50, 60.142077 }, //08 March 1994, UT1-UTC= 0.0419231, Accumulated Leap Seconds=28
  { 2449420.50, 60.144641 }, //09 March 1994, UT1-UTC= 0.0393590, Accumulated Leap Seconds=28
  { 2449421.50, 60.147280 }, //10 March 1994, UT1-UTC= 0.0367196, Accumulated Leap Seconds=28
  { 2449422.50, 60.149965 }, //11 March 1994, UT1-UTC= 0.0340346, Accumulated Leap Seconds=28
  { 2449423.50, 60.152629 }, //12 March 1994, UT1-UTC= 0.0313708, Accumulated Leap Seconds=28
  { 2449424.50, 60.155218 }, //13 March 1994, UT1-UTC= 0.0287815, Accumulated Leap Seconds=28
  { 2449425.50, 60.157715 }, //14 March 1994, UT1-UTC= 0.0262847, Accumulated Leap Seconds=28
  { 2449426.50, 60.160114 }, //15 March 1994, UT1-UTC= 0.0238858, Accumulated Leap Seconds=28
  { 2449427.50, 60.162423 }, //16 March 1994, UT1-UTC= 0.0215771, Accumulated Leap Seconds=28
  { 2449428.50, 60.164659 }, //17 March 1994, UT1-UTC= 0.0193410, Accumulated Leap Seconds=28
  { 2449429.50, 60.166817 }, //18 March 1994, UT1-UTC= 0.0171831, Accumulated Leap Seconds=28
  { 2449430.50, 60.168928 }, //19 March 1994, UT1-UTC= 0.0150715, Accumulated Leap Seconds=28
  { 2449431.50, 60.171039 }, //20 March 1994, UT1-UTC= 0.0129606, Accumulated Leap Seconds=28
  { 2449432.50, 60.173207 }, //21 March 1994, UT1-UTC= 0.0107926, Accumulated Leap Seconds=28
  { 2449433.50, 60.175495 }, //22 March 1994, UT1-UTC= 0.0085045, Accumulated Leap Seconds=28
  { 2449434.50, 60.177965 }, //23 March 1994, UT1-UTC= 0.0060354, Accumulated Leap Seconds=28
  { 2449435.50, 60.180665 }, //24 March 1994, UT1-UTC= 0.0033354, Accumulated Leap Seconds=28
  { 2449436.50, 60.183602 }, //25 March 1994, UT1-UTC= 0.0003978, Accumulated Leap Seconds=28
  { 2449437.50, 60.186750 }, //26 March 1994, UT1-UTC=-0.0027501, Accumulated Leap Seconds=28
  { 2449438.50, 60.190006 }, //27 March 1994, UT1-UTC=-0.0060058, Accumulated Leap Seconds=28
  { 2449439.50, 60.193231 }, //28 March 1994, UT1-UTC=-0.0092311, Accumulated Leap Seconds=28
  { 2449440.50, 60.196298 }, //29 March 1994, UT1-UTC=-0.0122978, Accumulated Leap Seconds=28
  { 2449441.50, 60.199132 }, //30 March 1994, UT1-UTC=-0.0151325, Accumulated Leap Seconds=28
  { 2449442.50, 60.201753 }, //31 March 1994, UT1-UTC=-0.0177530, Accumulated Leap Seconds=28
  { 2449443.50, 60.204218 }, //01 April 1994, UT1-UTC=-0.0202185, Accumulated Leap Seconds=28
  { 2449444.50, 60.206601 }, //02 April 1994, UT1-UTC=-0.0226006, Accumulated Leap Seconds=28
  { 2449445.50, 60.208976 }, //03 April 1994, UT1-UTC=-0.0249765, Accumulated Leap Seconds=28
  { 2449446.50, 60.211398 }, //04 April 1994, UT1-UTC=-0.0273982, Accumulated Leap Seconds=28
  { 2449447.50, 60.213875 }, //05 April 1994, UT1-UTC=-0.0298750, Accumulated Leap Seconds=28
  { 2449448.50, 60.216400 }, //06 April 1994, UT1-UTC=-0.0323998, Accumulated Leap Seconds=28
  { 2449449.50, 60.218976 }, //07 April 1994, UT1-UTC=-0.0349757, Accumulated Leap Seconds=28
  { 2449450.50, 60.221591 }, //08 April 1994, UT1-UTC=-0.0375914, Accumulated Leap Seconds=28
  { 2449451.50, 60.224220 }, //09 April 1994, UT1-UTC=-0.0402202, Accumulated Leap Seconds=28
  { 2449452.50, 60.226830 }, //10 April 1994, UT1-UTC=-0.0428300, Accumulated Leap Seconds=28
  { 2449453.50, 60.229386 }, //11 April 1994, UT1-UTC=-0.0453863, Accumulated Leap Seconds=28
  { 2449454.50, 60.231855 }, //12 April 1994, UT1-UTC=-0.0478555, Accumulated Leap Seconds=28
  { 2449455.50, 60.234219 }, //13 April 1994, UT1-UTC=-0.0502190, Accumulated Leap Seconds=28
  { 2449456.50, 60.236478 }, //14 April 1994, UT1-UTC=-0.0524782, Accumulated Leap Seconds=28
  { 2449457.50, 60.238647 }, //15 April 1994, UT1-UTC=-0.0546467, Accumulated Leap Seconds=28
  { 2449458.50, 60.240755 }, //16 April 1994, UT1-UTC=-0.0567550, Accumulated Leap Seconds=28
  { 2449459.50, 60.242866 }, //17 April 1994, UT1-UTC=-0.0588659, Accumulated Leap Seconds=28
  { 2449460.50, 60.245045 }, //18 April 1994, UT1-UTC=-0.0610451, Accumulated Leap Seconds=28
  { 2449461.50, 60.247353 }, //19 April 1994, UT1-UTC=-0.0633530, Accumulated Leap Seconds=28
  { 2449462.50, 60.249832 }, //20 April 1994, UT1-UTC=-0.0658316, Accumulated Leap Seconds=28
  { 2449463.50, 60.252488 }, //21 April 1994, UT1-UTC=-0.0684879, Accumulated Leap Seconds=28
  { 2449464.50, 60.255337 }, //22 April 1994, UT1-UTC=-0.0713369, Accumulated Leap Seconds=28
  { 2449465.50, 60.258367 }, //23 April 1994, UT1-UTC=-0.0743670, Accumulated Leap Seconds=28
  { 2449466.50, 60.261506 }, //24 April 1994, UT1-UTC=-0.0775057, Accumulated Leap Seconds=28
  { 2449467.50, 60.264633 }, //25 April 1994, UT1-UTC=-0.0806326, Accumulated Leap Seconds=28
  { 2449468.50, 60.267624 }, //26 April 1994, UT1-UTC=-0.0836236, Accumulated Leap Seconds=28
  { 2449469.50, 60.270404 }, //27 April 1994, UT1-UTC=-0.0864045, Accumulated Leap Seconds=28
  { 2449470.50, 60.272987 }, //28 April 1994, UT1-UTC=-0.0889870, Accumulated Leap Seconds=28
  { 2449471.50, 60.275454 }, //29 April 1994, UT1-UTC=-0.0914543, Accumulated Leap Seconds=28
  { 2449472.50, 60.277897 }, //30 April 1994, UT1-UTC=-0.0938970, Accumulated Leap Seconds=28
  { 2449473.50, 60.280375 }, //01 May 1994, UT1-UTC=-0.0963745, Accumulated Leap Seconds=28
  { 2449474.50, 60.282912 }, //02 May 1994, UT1-UTC=-0.0989117, Accumulated Leap Seconds=28
  { 2449475.50, 60.285501 }, //03 May 1994, UT1-UTC=-0.1015009, Accumulated Leap Seconds=28
  { 2449476.50, 60.288120 }, //04 May 1994, UT1-UTC=-0.1041202, Accumulated Leap Seconds=28
  { 2449477.50, 60.290754 }, //05 May 1994, UT1-UTC=-0.1067542, Accumulated Leap Seconds=28
  { 2449478.50, 60.293379 }, //06 May 1994, UT1-UTC=-0.1093791, Accumulated Leap Seconds=28
  { 2449479.50, 60.295956 }, //07 May 1994, UT1-UTC=-0.1119560, Accumulated Leap Seconds=28
  { 2449480.50, 60.298459 }, //08 May 1994, UT1-UTC=-0.1144588, Accumulated Leap Seconds=28
  { 2449481.50, 60.300874 }, //09 May 1994, UT1-UTC=-0.1168745, Accumulated Leap Seconds=28
  { 2449482.50, 60.303192 }, //10 May 1994, UT1-UTC=-0.1191924, Accumulated Leap Seconds=28
  { 2449483.50, 60.305421 }, //11 May 1994, UT1-UTC=-0.1214215, Accumulated Leap Seconds=28
  { 2449484.50, 60.307604 }, //12 May 1994, UT1-UTC=-0.1236039, Accumulated Leap Seconds=28
  { 2449485.50, 60.309764 }, //13 May 1994, UT1-UTC=-0.1257643, Accumulated Leap Seconds=28
  { 2449486.50, 60.311913 }, //14 May 1994, UT1-UTC=-0.1279126, Accumulated Leap Seconds=28
  { 2449487.50, 60.314100 }, //15 May 1994, UT1-UTC=-0.1300998, Accumulated Leap Seconds=28
  { 2449488.50, 60.316367 }, //16 May 1994, UT1-UTC=-0.1323669, Accumulated Leap Seconds=28
  { 2449489.50, 60.318745 }, //17 May 1994, UT1-UTC=-0.1347455, Accumulated Leap Seconds=28
  { 2449490.50, 60.321252 }, //18 May 1994, UT1-UTC=-0.1372521, Accumulated Leap Seconds=28
  { 2449491.50, 60.323888 }, //19 May 1994, UT1-UTC=-0.1398877, Accumulated Leap Seconds=28
  { 2449492.50, 60.326628 }, //20 May 1994, UT1-UTC=-0.1426278, Accumulated Leap Seconds=28
  { 2449493.50, 60.329406 }, //21 May 1994, UT1-UTC=-0.1454058, Accumulated Leap Seconds=28
  { 2449494.50, 60.332135 }, //22 May 1994, UT1-UTC=-0.1481354, Accumulated Leap Seconds=28
  { 2449495.50, 60.334731 }, //23 May 1994, UT1-UTC=-0.1507309, Accumulated Leap Seconds=28
  { 2449496.50, 60.337135 }, //24 May 1994, UT1-UTC=-0.1531355, Accumulated Leap Seconds=28
  { 2449497.50, 60.339333 }, //25 May 1994, UT1-UTC=-0.1553330, Accumulated Leap Seconds=28
  { 2449498.50, 60.341352 }, //26 May 1994, UT1-UTC=-0.1573521, Accumulated Leap Seconds=28
  { 2449499.50, 60.343273 }, //27 May 1994, UT1-UTC=-0.1592733, Accumulated Leap Seconds=28
  { 2449500.50, 60.345164 }, //28 May 1994, UT1-UTC=-0.1611645, Accumulated Leap Seconds=28
  { 2449501.50, 60.347075 }, //29 May 1994, UT1-UTC=-0.1630746, Accumulated Leap Seconds=28
  { 2449502.50, 60.349025 }, //30 May 1994, UT1-UTC=-0.1650252, Accumulated Leap Seconds=28
  { 2449503.50, 60.351017 }, //31 May 1994, UT1-UTC=-0.1670173, Accumulated Leap Seconds=28
  { 2449504.50, 60.353035 }, //01 June 1994, UT1-UTC=-0.1690352, Accumulated Leap Seconds=28
  { 2449505.50, 60.355041 }, //02 June 1994, UT1-UTC=-0.1710415, Accumulated Leap Seconds=28
  { 2449506.50, 60.356995 }, //03 June 1994, UT1-UTC=-0.1729949, Accumulated Leap Seconds=28
  { 2449507.50, 60.358865 }, //04 June 1994, UT1-UTC=-0.1748653, Accumulated Leap Seconds=28
  { 2449508.50, 60.360634 }, //05 June 1994, UT1-UTC=-0.1766344, Accumulated Leap Seconds=28
  { 2449509.50, 60.362293 }, //06 June 1994, UT1-UTC=-0.1782929, Accumulated Leap Seconds=28
  { 2449510.50, 60.363840 }, //07 June 1994, UT1-UTC=-0.1798400, Accumulated Leap Seconds=28
  { 2449511.50, 60.365294 }, //08 June 1994, UT1-UTC=-0.1812939, Accumulated Leap Seconds=28
  { 2449512.50, 60.366688 }, //09 June 1994, UT1-UTC=-0.1826882, Accumulated Leap Seconds=28
  { 2449513.50, 60.368075 }, //10 June 1994, UT1-UTC=-0.1840746, Accumulated Leap Seconds=28
  { 2449514.50, 60.369521 }, //11 June 1994, UT1-UTC=-0.1855206, Accumulated Leap Seconds=28
  { 2449515.50, 60.371076 }, //12 June 1994, UT1-UTC=-0.1870763, Accumulated Leap Seconds=28
  { 2449516.50, 60.372773 }, //13 June 1994, UT1-UTC=-0.1887730, Accumulated Leap Seconds=28
  { 2449517.50, 60.374626 }, //14 June 1994, UT1-UTC=-0.1906264, Accumulated Leap Seconds=28
  { 2449518.50, 60.376626 }, //15 June 1994, UT1-UTC=-0.1926260, Accumulated Leap Seconds=28
  { 2449519.50, 60.378723 }, //16 June 1994, UT1-UTC=-0.1947228, Accumulated Leap Seconds=28
  { 2449520.50, 60.380859 }, //17 June 1994, UT1-UTC=-0.1968590, Accumulated Leap Seconds=28
  { 2449521.50, 60.382966 }, //18 June 1994, UT1-UTC=-0.1989662, Accumulated Leap Seconds=28
  { 2449522.50, 60.384968 }, //19 June 1994, UT1-UTC=-0.2009678, Accumulated Leap Seconds=28
  { 2449523.50, 60.386795 }, //20 June 1994, UT1-UTC=-0.2027948, Accumulated Leap Seconds=28
  { 2449524.50, 60.388409 }, //21 June 1994, UT1-UTC=-0.2044087, Accumulated Leap Seconds=28
  { 2449525.50, 60.389822 }, //22 June 1994, UT1-UTC=-0.2058218, Accumulated Leap Seconds=28
  { 2449526.50, 60.391102 }, //23 June 1994, UT1-UTC=-0.2071025, Accumulated Leap Seconds=28
  { 2449527.50, 60.392322 }, //24 June 1994, UT1-UTC=-0.2083225, Accumulated Leap Seconds=28
  { 2449528.50, 60.393544 }, //25 June 1994, UT1-UTC=-0.2095437, Accumulated Leap Seconds=28
  { 2449529.50, 60.394808 }, //26 June 1994, UT1-UTC=-0.2108083, Accumulated Leap Seconds=28
  { 2449530.50, 60.396121 }, //27 June 1994, UT1-UTC=-0.2121210, Accumulated Leap Seconds=28
  { 2449531.50, 60.397461 }, //28 June 1994, UT1-UTC=-0.2134615, Accumulated Leap Seconds=28
  { 2449532.50, 60.398783 }, //29 June 1994, UT1-UTC=-0.2147828, Accumulated Leap Seconds=28
  { 2449533.50, 60.400023 }, //30 June 1994, UT1-UTC=-0.2160228, Accumulated Leap Seconds=28
  { 2449534.50, 60.401189 }, //01 July 1994, UT1-UTC= 0.7828109, Accumulated Leap Seconds=29
  { 2449535.50, 60.402302 }, //02 July 1994, UT1-UTC= 0.7816975, Accumulated Leap Seconds=29
  { 2449536.50, 60.403348 }, //03 July 1994, UT1-UTC= 0.7806515, Accumulated Leap Seconds=29
  { 2449537.50, 60.404321 }, //04 July 1994, UT1-UTC= 0.7796789, Accumulated Leap Seconds=29
  { 2449538.50, 60.405229 }, //05 July 1994, UT1-UTC= 0.7787705, Accumulated Leap Seconds=29
  { 2449539.50, 60.406096 }, //06 July 1994, UT1-UTC= 0.7779035, Accumulated Leap Seconds=29
  { 2449540.50, 60.406955 }, //07 July 1994, UT1-UTC= 0.7770447, Accumulated Leap Seconds=29
  { 2449541.50, 60.407870 }, //08 July 1994, UT1-UTC= 0.7761297, Accumulated Leap Seconds=29
  { 2449542.50, 60.408926 }, //09 July 1994, UT1-UTC= 0.7750741, Accumulated Leap Seconds=29
  { 2449543.50, 60.410167 }, //10 July 1994, UT1-UTC= 0.7738328, Accumulated Leap Seconds=29
  { 2449544.50, 60.411610 }, //11 July 1994, UT1-UTC= 0.7723900, Accumulated Leap Seconds=29
  { 2449545.50, 60.413245 }, //12 July 1994, UT1-UTC= 0.7707553, Accumulated Leap Seconds=29
  { 2449546.50, 60.415033 }, //13 July 1994, UT1-UTC= 0.7689669, Accumulated Leap Seconds=29
  { 2449547.50, 60.416921 }, //14 July 1994, UT1-UTC= 0.7670791, Accumulated Leap Seconds=29
  { 2449548.50, 60.418839 }, //15 July 1994, UT1-UTC= 0.7651605, Accumulated Leap Seconds=29
  { 2449549.50, 60.420744 }, //16 July 1994, UT1-UTC= 0.7632557, Accumulated Leap Seconds=29
  { 2449550.50, 60.422571 }, //17 July 1994, UT1-UTC= 0.7614293, Accumulated Leap Seconds=29
  { 2449551.50, 60.424268 }, //18 July 1994, UT1-UTC= 0.7597324, Accumulated Leap Seconds=29
  { 2449552.50, 60.425812 }, //19 July 1994, UT1-UTC= 0.7581883, Accumulated Leap Seconds=29
  { 2449553.50, 60.427212 }, //20 July 1994, UT1-UTC= 0.7567882, Accumulated Leap Seconds=29
  { 2449554.50, 60.428511 }, //21 July 1994, UT1-UTC= 0.7554893, Accumulated Leap Seconds=29
  { 2449555.50, 60.429791 }, //22 July 1994, UT1-UTC= 0.7542091, Accumulated Leap Seconds=29
  { 2449556.50, 60.431135 }, //23 July 1994, UT1-UTC= 0.7528648, Accumulated Leap Seconds=29
  { 2449557.50, 60.432580 }, //24 July 1994, UT1-UTC= 0.7514198, Accumulated Leap Seconds=29
  { 2449558.50, 60.434102 }, //25 July 1994, UT1-UTC= 0.7498981, Accumulated Leap Seconds=29
  { 2449559.50, 60.435655 }, //26 July 1994, UT1-UTC= 0.7483448, Accumulated Leap Seconds=29
  { 2449560.50, 60.437188 }, //27 July 1994, UT1-UTC= 0.7468115, Accumulated Leap Seconds=29
  { 2449561.50, 60.438660 }, //28 July 1994, UT1-UTC= 0.7453400, Accumulated Leap Seconds=29
  { 2449562.50, 60.440073 }, //29 July 1994, UT1-UTC= 0.7439266, Accumulated Leap Seconds=29
  { 2449563.50, 60.441446 }, //30 July 1994, UT1-UTC= 0.7425542, Accumulated Leap Seconds=29
  { 2449564.50, 60.442759 }, //31 July 1994, UT1-UTC= 0.7412412, Accumulated Leap Seconds=29
  { 2449565.50, 60.443996 }, //01 August 1994, UT1-UTC= 0.7400041, Accumulated Leap Seconds=29
  { 2449566.50, 60.445155 }, //02 August 1994, UT1-UTC= 0.7388451, Accumulated Leap Seconds=29
  { 2449567.50, 60.446263 }, //03 August 1994, UT1-UTC= 0.7377366, Accumulated Leap Seconds=29
  { 2449568.50, 60.447402 }, //04 August 1994, UT1-UTC= 0.7365978, Accumulated Leap Seconds=29
  { 2449569.50, 60.448625 }, //05 August 1994, UT1-UTC= 0.7353751, Accumulated Leap Seconds=29
  { 2449570.50, 60.449965 }, //06 August 1994, UT1-UTC= 0.7340345, Accumulated Leap Seconds=29
  { 2449571.50, 60.451447 }, //07 August 1994, UT1-UTC= 0.7325534, Accumulated Leap Seconds=29
  { 2449572.50, 60.453063 }, //08 August 1994, UT1-UTC= 0.7309369, Accumulated Leap Seconds=29
  { 2449573.50, 60.454782 }, //09 August 1994, UT1-UTC= 0.7292178, Accumulated Leap Seconds=29
  { 2449574.50, 60.456543 }, //10 August 1994, UT1-UTC= 0.7274568, Accumulated Leap Seconds=29
  { 2449575.50, 60.458310 }, //11 August 1994, UT1-UTC= 0.7256896, Accumulated Leap Seconds=29
  { 2449576.50, 60.460021 }, //12 August 1994, UT1-UTC= 0.7239789, Accumulated Leap Seconds=29
  { 2449577.50, 60.461607 }, //13 August 1994, UT1-UTC= 0.7223926, Accumulated Leap Seconds=29
  { 2449578.50, 60.463036 }, //14 August 1994, UT1-UTC= 0.7209637, Accumulated Leap Seconds=29
  { 2449579.50, 60.464327 }, //15 August 1994, UT1-UTC= 0.7196732, Accumulated Leap Seconds=29
  { 2449580.50, 60.465542 }, //16 August 1994, UT1-UTC= 0.7184578, Accumulated Leap Seconds=29
  { 2449581.50, 60.466756 }, //17 August 1994, UT1-UTC= 0.7172443, Accumulated Leap Seconds=29
  { 2449582.50, 60.468012 }, //18 August 1994, UT1-UTC= 0.7159879, Accumulated Leap Seconds=29
  { 2449583.50, 60.469359 }, //19 August 1994, UT1-UTC= 0.7146405, Accumulated Leap Seconds=29
  { 2449584.50, 60.470828 }, //20 August 1994, UT1-UTC= 0.7131720, Accumulated Leap Seconds=29
  { 2449585.50, 60.472423 }, //21 August 1994, UT1-UTC= 0.7115773, Accumulated Leap Seconds=29
  { 2449586.50, 60.474123 }, //22 August 1994, UT1-UTC= 0.7098773, Accumulated Leap Seconds=29
  { 2449587.50, 60.475887 }, //23 August 1994, UT1-UTC= 0.7081134, Accumulated Leap Seconds=29
  { 2449588.50, 60.477666 }, //24 August 1994, UT1-UTC= 0.7063344, Accumulated Leap Seconds=29
  { 2449589.50, 60.479406 }, //25 August 1994, UT1-UTC= 0.7045941, Accumulated Leap Seconds=29
  { 2449590.50, 60.481065 }, //26 August 1994, UT1-UTC= 0.7029353, Accumulated Leap Seconds=29
  { 2449591.50, 60.482661 }, //27 August 1994, UT1-UTC= 0.7013388, Accumulated Leap Seconds=29
  { 2449592.50, 60.484199 }, //28 August 1994, UT1-UTC= 0.6998011, Accumulated Leap Seconds=29
  { 2449593.50, 60.485681 }, //29 August 1994, UT1-UTC= 0.6983189, Accumulated Leap Seconds=29
  { 2449594.50, 60.487120 }, //30 August 1994, UT1-UTC= 0.6968799, Accumulated Leap Seconds=29
  { 2449595.50, 60.488545 }, //31 August 1994, UT1-UTC= 0.6954548, Accumulated Leap Seconds=29
  { 2449596.50, 60.490026 }, //01 September 1994, UT1-UTC= 0.6939743, Accumulated Leap Seconds=29
  { 2449597.50, 60.491634 }, //02 September 1994, UT1-UTC= 0.6923658, Accumulated Leap Seconds=29
  { 2449598.50, 60.493424 }, //03 September 1994, UT1-UTC= 0.6905762, Accumulated Leap Seconds=29
  { 2449599.50, 60.495433 }, //04 September 1994, UT1-UTC= 0.6885671, Accumulated Leap Seconds=29
  { 2449600.50, 60.497665 }, //05 September 1994, UT1-UTC= 0.6863347, Accumulated Leap Seconds=29
  { 2449601.50, 60.500079 }, //06 September 1994, UT1-UTC= 0.6839213, Accumulated Leap Seconds=29
  { 2449602.50, 60.502593 }, //07 September 1994, UT1-UTC= 0.6814073, Accumulated Leap Seconds=29
  { 2449603.50, 60.505124 }, //08 September 1994, UT1-UTC= 0.6788759, Accumulated Leap Seconds=29
  { 2449604.50, 60.507655 }, //09 September 1994, UT1-UTC= 0.6763445, Accumulated Leap Seconds=29
  { 2449605.50, 60.510109 }, //10 September 1994, UT1-UTC= 0.6738905, Accumulated Leap Seconds=29
  { 2449606.50, 60.512456 }, //11 September 1994, UT1-UTC= 0.6715435, Accumulated Leap Seconds=29
  { 2449607.50, 60.514712 }, //12 September 1994, UT1-UTC= 0.6692876, Accumulated Leap Seconds=29
  { 2449608.50, 60.516924 }, //13 September 1994, UT1-UTC= 0.6670763, Accumulated Leap Seconds=29
  { 2449609.50, 60.519146 }, //14 September 1994, UT1-UTC= 0.6648538, Accumulated Leap Seconds=29
  { 2449610.50, 60.521425 }, //15 September 1994, UT1-UTC= 0.6625752, Accumulated Leap Seconds=29
  { 2449611.50, 60.523803 }, //16 September 1994, UT1-UTC= 0.6601972, Accumulated Leap Seconds=29
  { 2449612.50, 60.526306 }, //17 September 1994, UT1-UTC= 0.6576935, Accumulated Leap Seconds=29
  { 2449613.50, 60.528901 }, //18 September 1994, UT1-UTC= 0.6550986, Accumulated Leap Seconds=29
  { 2449614.50, 60.531529 }, //19 September 1994, UT1-UTC= 0.6524714, Accumulated Leap Seconds=29
  { 2449615.50, 60.534127 }, //20 September 1994, UT1-UTC= 0.6498729, Accumulated Leap Seconds=29
  { 2449616.50, 60.536649 }, //21 September 1994, UT1-UTC= 0.6473512, Accumulated Leap Seconds=29
  { 2449617.50, 60.539082 }, //22 September 1994, UT1-UTC= 0.6449183, Accumulated Leap Seconds=29
  { 2449618.50, 60.541401 }, //23 September 1994, UT1-UTC= 0.6425989, Accumulated Leap Seconds=29
  { 2449619.50, 60.543591 }, //24 September 1994, UT1-UTC= 0.6404093, Accumulated Leap Seconds=29
  { 2449620.50, 60.545664 }, //25 September 1994, UT1-UTC= 0.6383363, Accumulated Leap Seconds=29
  { 2449621.50, 60.547652 }, //26 September 1994, UT1-UTC= 0.6363481, Accumulated Leap Seconds=29
  { 2449622.50, 60.549596 }, //27 September 1994, UT1-UTC= 0.6344044, Accumulated Leap Seconds=29
  { 2449623.50, 60.551538 }, //28 September 1994, UT1-UTC= 0.6324622, Accumulated Leap Seconds=29
  { 2449624.50, 60.553517 }, //29 September 1994, UT1-UTC= 0.6304832, Accumulated Leap Seconds=29
  { 2449625.50, 60.555587 }, //30 September 1994, UT1-UTC= 0.6284133, Accumulated Leap Seconds=29
  { 2449626.50, 60.557805 }, //01 October 1994, UT1-UTC= 0.6261946, Accumulated Leap Seconds=29
  { 2449627.50, 60.560208 }, //02 October 1994, UT1-UTC= 0.6237917, Accumulated Leap Seconds=29
  { 2449628.50, 60.562797 }, //03 October 1994, UT1-UTC= 0.6212025, Accumulated Leap Seconds=29
  { 2449629.50, 60.565549 }, //04 October 1994, UT1-UTC= 0.6184507, Accumulated Leap Seconds=29
  { 2449630.50, 60.568403 }, //05 October 1994, UT1-UTC= 0.6155973, Accumulated Leap Seconds=29
  { 2449631.50, 60.571256 }, //06 October 1994, UT1-UTC= 0.6127436, Accumulated Leap Seconds=29
  { 2449632.50, 60.573997 }, //07 October 1994, UT1-UTC= 0.6100033, Accumulated Leap Seconds=29
  { 2449633.50, 60.576564 }, //08 October 1994, UT1-UTC= 0.6074360, Accumulated Leap Seconds=29
  { 2449634.50, 60.578974 }, //09 October 1994, UT1-UTC= 0.6050261, Accumulated Leap Seconds=29
  { 2449635.50, 60.581286 }, //10 October 1994, UT1-UTC= 0.6027139, Accumulated Leap Seconds=29
  { 2449636.50, 60.583574 }, //11 October 1994, UT1-UTC= 0.6004259, Accumulated Leap Seconds=29
  { 2449637.50, 60.585900 }, //12 October 1994, UT1-UTC= 0.5981002, Accumulated Leap Seconds=29
  { 2449638.50, 60.588296 }, //13 October 1994, UT1-UTC= 0.5957037, Accumulated Leap Seconds=29
  { 2449639.50, 60.590765 }, //14 October 1994, UT1-UTC= 0.5932348, Accumulated Leap Seconds=29
  { 2449640.50, 60.593287 }, //15 October 1994, UT1-UTC= 0.5907126, Accumulated Leap Seconds=29
  { 2449641.50, 60.595825 }, //16 October 1994, UT1-UTC= 0.5881747, Accumulated Leap Seconds=29
  { 2449642.50, 60.598327 }, //17 October 1994, UT1-UTC= 0.5856728, Accumulated Leap Seconds=29
  { 2449643.50, 60.600737 }, //18 October 1994, UT1-UTC= 0.5832630, Accumulated Leap Seconds=29
  { 2449644.50, 60.603010 }, //19 October 1994, UT1-UTC= 0.5809904, Accumulated Leap Seconds=29
  { 2449645.50, 60.605143 }, //20 October 1994, UT1-UTC= 0.5788572, Accumulated Leap Seconds=29
  { 2449646.50, 60.607176 }, //21 October 1994, UT1-UTC= 0.5768235, Accumulated Leap Seconds=29
  { 2449647.50, 60.609147 }, //22 October 1994, UT1-UTC= 0.5748526, Accumulated Leap Seconds=29
  { 2449648.50, 60.611083 }, //23 October 1994, UT1-UTC= 0.5729171, Accumulated Leap Seconds=29
  { 2449649.50, 60.613017 }, //24 October 1994, UT1-UTC= 0.5709825, Accumulated Leap Seconds=29
  { 2449650.50, 60.614991 }, //25 October 1994, UT1-UTC= 0.5690086, Accumulated Leap Seconds=29
  { 2449651.50, 60.617048 }, //26 October 1994, UT1-UTC= 0.5669519, Accumulated Leap Seconds=29
  { 2449652.50, 60.619223 }, //27 October 1994, UT1-UTC= 0.5647765, Accumulated Leap Seconds=29
  { 2449653.50, 60.621540 }, //28 October 1994, UT1-UTC= 0.5624601, Accumulated Leap Seconds=29
  { 2449654.50, 60.624024 }, //29 October 1994, UT1-UTC= 0.5599758, Accumulated Leap Seconds=29
  { 2449655.50, 60.626689 }, //30 October 1994, UT1-UTC= 0.5573114, Accumulated Leap Seconds=29
  { 2449656.50, 60.629513 }, //31 October 1994, UT1-UTC= 0.5544868, Accumulated Leap Seconds=29
  { 2449657.50, 60.632445 }, //01 November 1994, UT1-UTC= 0.5515554, Accumulated Leap Seconds=29
  { 2449658.50, 60.635404 }, //02 November 1994, UT1-UTC= 0.5485957, Accumulated Leap Seconds=29
  { 2449659.50, 60.638302 }, //03 November 1994, UT1-UTC= 0.5456979, Accumulated Leap Seconds=29
  { 2449660.50, 60.641052 }, //04 November 1994, UT1-UTC= 0.5429476, Accumulated Leap Seconds=29
  { 2449661.50, 60.643625 }, //05 November 1994, UT1-UTC= 0.5403745, Accumulated Leap Seconds=29
  { 2449662.50, 60.646049 }, //06 November 1994, UT1-UTC= 0.5379513, Accumulated Leap Seconds=29
  { 2449663.50, 60.648385 }, //07 November 1994, UT1-UTC= 0.5356145, Accumulated Leap Seconds=29
  { 2449664.50, 60.650696 }, //08 November 1994, UT1-UTC= 0.5333038, Accumulated Leap Seconds=29
  { 2449665.50, 60.653025 }, //09 November 1994, UT1-UTC= 0.5309748, Accumulated Leap Seconds=29
  { 2449666.50, 60.655416 }, //10 November 1994, UT1-UTC= 0.5285843, Accumulated Leap Seconds=29
  { 2449667.50, 60.657871 }, //11 November 1994, UT1-UTC= 0.5261290, Accumulated Leap Seconds=29
  { 2449668.50, 60.660370 }, //12 November 1994, UT1-UTC= 0.5236296, Accumulated Leap Seconds=29
  { 2449669.50, 60.662880 }, //13 November 1994, UT1-UTC= 0.5211201, Accumulated Leap Seconds=29
  { 2449670.50, 60.665365 }, //14 November 1994, UT1-UTC= 0.5186349, Accumulated Leap Seconds=29
  { 2449671.50, 60.667799 }, //15 November 1994, UT1-UTC= 0.5162007, Accumulated Leap Seconds=29
  { 2449672.50, 60.670163 }, //16 November 1994, UT1-UTC= 0.5138370, Accumulated Leap Seconds=29
  { 2449673.50, 60.672440 }, //17 November 1994, UT1-UTC= 0.5115599, Accumulated Leap Seconds=29
  { 2449674.50, 60.674621 }, //18 November 1994, UT1-UTC= 0.5093791, Accumulated Leap Seconds=29
  { 2449675.50, 60.676713 }, //19 November 1994, UT1-UTC= 0.5072870, Accumulated Leap Seconds=29
  { 2449676.50, 60.678743 }, //20 November 1994, UT1-UTC= 0.5052570, Accumulated Leap Seconds=29
  { 2449677.50, 60.680757 }, //21 November 1994, UT1-UTC= 0.5032430, Accumulated Leap Seconds=29
  { 2449678.50, 60.682809 }, //22 November 1994, UT1-UTC= 0.5011910, Accumulated Leap Seconds=29
  { 2449679.50, 60.684953 }, //23 November 1994, UT1-UTC= 0.4990469, Accumulated Leap Seconds=29
  { 2449680.50, 60.687237 }, //24 November 1994, UT1-UTC= 0.4967631, Accumulated Leap Seconds=29
  { 2449681.50, 60.689678 }, //25 November 1994, UT1-UTC= 0.4943220, Accumulated Leap Seconds=29
  { 2449682.50, 60.692267 }, //26 November 1994, UT1-UTC= 0.4917334, Accumulated Leap Seconds=29
  { 2449683.50, 60.694974 }, //27 November 1994, UT1-UTC= 0.4890261, Accumulated Leap Seconds=29
  { 2449684.50, 60.697751 }, //28 November 1994, UT1-UTC= 0.4862486, Accumulated Leap Seconds=29
  { 2449685.50, 60.700527 }, //29 November 1994, UT1-UTC= 0.4834729, Accumulated Leap Seconds=29
  { 2449686.50, 60.703232 }, //30 November 1994, UT1-UTC= 0.4807677, Accumulated Leap Seconds=29
  { 2449687.50, 60.705857 }, //01 December 1994, UT1-UTC= 0.4781431, Accumulated Leap Seconds=29
  { 2449688.50, 60.708378 }, //02 December 1994, UT1-UTC= 0.4756221, Accumulated Leap Seconds=29
  { 2449689.50, 60.710784 }, //03 December 1994, UT1-UTC= 0.4732160, Accumulated Leap Seconds=29
  { 2449690.50, 60.713107 }, //04 December 1994, UT1-UTC= 0.4708927, Accumulated Leap Seconds=29
  { 2449691.50, 60.715407 }, //05 December 1994, UT1-UTC= 0.4685931, Accumulated Leap Seconds=29
  { 2449692.50, 60.717747 }, //06 December 1994, UT1-UTC= 0.4662530, Accumulated Leap Seconds=29
  { 2449693.50, 60.720161 }, //07 December 1994, UT1-UTC= 0.4638394, Accumulated Leap Seconds=29
  { 2449694.50, 60.722637 }, //08 December 1994, UT1-UTC= 0.4613633, Accumulated Leap Seconds=29
  { 2449695.50, 60.725143 }, //09 December 1994, UT1-UTC= 0.4588573, Accumulated Leap Seconds=29
  { 2449696.50, 60.727649 }, //10 December 1994, UT1-UTC= 0.4563510, Accumulated Leap Seconds=29
  { 2449697.50, 60.730134 }, //11 December 1994, UT1-UTC= 0.4538660, Accumulated Leap Seconds=29
  { 2449698.50, 60.732583 }, //12 December 1994, UT1-UTC= 0.4514165, Accumulated Leap Seconds=29
  { 2449699.50, 60.734991 }, //13 December 1994, UT1-UTC= 0.4490087, Accumulated Leap Seconds=29
  { 2449700.50, 60.737348 }, //14 December 1994, UT1-UTC= 0.4466520, Accumulated Leap Seconds=29
  { 2449701.50, 60.739617 }, //15 December 1994, UT1-UTC= 0.4443833, Accumulated Leap Seconds=29
  { 2449702.50, 60.741830 }, //16 December 1994, UT1-UTC= 0.4421698, Accumulated Leap Seconds=29
  { 2449703.50, 60.744035 }, //17 December 1994, UT1-UTC= 0.4399654, Accumulated Leap Seconds=29
  { 2449704.50, 60.746284 }, //18 December 1994, UT1-UTC= 0.4377156, Accumulated Leap Seconds=29
  { 2449705.50, 60.748628 }, //19 December 1994, UT1-UTC= 0.4353719, Accumulated Leap Seconds=29
  { 2449706.50, 60.751110 }, //20 December 1994, UT1-UTC= 0.4328901, Accumulated Leap Seconds=29
  { 2449707.50, 60.753753 }, //21 December 1994, UT1-UTC= 0.4302465, Accumulated Leap Seconds=29
  { 2449708.50, 60.756518 }, //22 December 1994, UT1-UTC= 0.4274816, Accumulated Leap Seconds=29
  { 2449709.50, 60.759389 }, //23 December 1994, UT1-UTC= 0.4246107, Accumulated Leap Seconds=29
  { 2449710.50, 60.762363 }, //24 December 1994, UT1-UTC= 0.4216372, Accumulated Leap Seconds=29
  { 2449711.50, 60.765412 }, //25 December 1994, UT1-UTC= 0.4185881, Accumulated Leap Seconds=29
  { 2449712.50, 60.768489 }, //26 December 1994, UT1-UTC= 0.4155109, Accumulated Leap Seconds=29
  { 2449713.50, 60.771536 }, //27 December 1994, UT1-UTC= 0.4124641, Accumulated Leap Seconds=29
  { 2449714.50, 60.774495 }, //28 December 1994, UT1-UTC= 0.4095048, Accumulated Leap Seconds=29
  { 2449715.50, 60.777334 }, //29 December 1994, UT1-UTC= 0.4066659, Accumulated Leap Seconds=29
  { 2449716.50, 60.780044 }, //30 December 1994, UT1-UTC= 0.4039556, Accumulated Leap Seconds=29
  { 2449717.50, 60.782693 }, //31 December 1994, UT1-UTC= 0.4013070, Accumulated Leap Seconds=29
  { 2449718.50, 60.785348 }, //01 January 1995, UT1-UTC= 0.3986518, Accumulated Leap Seconds=29
  { 2449719.50, 60.788070 }, //02 January 1995, UT1-UTC= 0.3959303, Accumulated Leap Seconds=29
  { 2449720.50, 60.790897 }, //03 January 1995, UT1-UTC= 0.3931028, Accumulated Leap Seconds=29
  { 2449721.50, 60.793831 }, //04 January 1995, UT1-UTC= 0.3901693, Accumulated Leap Seconds=29
  { 2449722.50, 60.796826 }, //05 January 1995, UT1-UTC= 0.3871735, Accumulated Leap Seconds=29
  { 2449723.50, 60.799809 }, //06 January 1995, UT1-UTC= 0.3841909, Accumulated Leap Seconds=29
  { 2449724.50, 60.802717 }, //07 January 1995, UT1-UTC= 0.3812834, Accumulated Leap Seconds=29
  { 2449725.50, 60.805502 }, //08 January 1995, UT1-UTC= 0.3784981, Accumulated Leap Seconds=29
  { 2449726.50, 60.808135 }, //09 January 1995, UT1-UTC= 0.3758648, Accumulated Leap Seconds=29
  { 2449727.50, 60.810602 }, //10 January 1995, UT1-UTC= 0.3733976, Accumulated Leap Seconds=29
  { 2449728.50, 60.812908 }, //11 January 1995, UT1-UTC= 0.3710918, Accumulated Leap Seconds=29
  { 2449729.50, 60.815095 }, //12 January 1995, UT1-UTC= 0.3689052, Accumulated Leap Seconds=29
  { 2449730.50, 60.817208 }, //13 January 1995, UT1-UTC= 0.3667916, Accumulated Leap Seconds=29
  { 2449731.50, 60.819288 }, //14 January 1995, UT1-UTC= 0.3647120, Accumulated Leap Seconds=29
  { 2449732.50, 60.821380 }, //15 January 1995, UT1-UTC= 0.3626203, Accumulated Leap Seconds=29
  { 2449733.50, 60.823521 }, //16 January 1995, UT1-UTC= 0.3604790, Accumulated Leap Seconds=29
  { 2449734.50, 60.825741 }, //17 January 1995, UT1-UTC= 0.3582594, Accumulated Leap Seconds=29
  { 2449735.50, 60.828070 }, //18 January 1995, UT1-UTC= 0.3559299, Accumulated Leap Seconds=29
  { 2449736.50, 60.830554 }, //19 January 1995, UT1-UTC= 0.3534465, Accumulated Leap Seconds=29
  { 2449737.50, 60.833176 }, //20 January 1995, UT1-UTC= 0.3508239, Accumulated Leap Seconds=29
  { 2449738.50, 60.835911 }, //21 January 1995, UT1-UTC= 0.3480889, Accumulated Leap Seconds=29
  { 2449739.50, 60.838722 }, //22 January 1995, UT1-UTC= 0.3452784, Accumulated Leap Seconds=29
  { 2449740.50, 60.841566 }, //23 January 1995, UT1-UTC= 0.3424344, Accumulated Leap Seconds=29
  { 2449741.50, 60.844394 }, //24 January 1995, UT1-UTC= 0.3396055, Accumulated Leap Seconds=29
  { 2449742.50, 60.847167 }, //25 January 1995, UT1-UTC= 0.3368327, Accumulated Leap Seconds=29
  { 2449743.50, 60.849880 }, //26 January 1995, UT1-UTC= 0.3341203, Accumulated Leap Seconds=29
  { 2449744.50, 60.852551 }, //27 January 1995, UT1-UTC= 0.3314486, Accumulated Leap Seconds=29
  { 2449745.50, 60.855203 }, //28 January 1995, UT1-UTC= 0.3287973, Accumulated Leap Seconds=29
  { 2449746.50, 60.857885 }, //29 January 1995, UT1-UTC= 0.3261154, Accumulated Leap Seconds=29
  { 2449747.50, 60.860637 }, //30 January 1995, UT1-UTC= 0.3233627, Accumulated Leap Seconds=29
  { 2449748.50, 60.863469 }, //31 January 1995, UT1-UTC= 0.3205310, Accumulated Leap Seconds=29
  { 2449749.50, 60.866350 }, //01 February 1995, UT1-UTC= 0.3176496, Accumulated Leap Seconds=29
  { 2449750.50, 60.869234 }, //02 February 1995, UT1-UTC= 0.3147658, Accumulated Leap Seconds=29
  { 2449751.50, 60.872096 }, //03 February 1995, UT1-UTC= 0.3119044, Accumulated Leap Seconds=29
  { 2449752.50, 60.874918 }, //04 February 1995, UT1-UTC= 0.3090820, Accumulated Leap Seconds=29
  { 2449753.50, 60.877666 }, //05 February 1995, UT1-UTC= 0.3063335, Accumulated Leap Seconds=29
  { 2449754.50, 60.880317 }, //06 February 1995, UT1-UTC= 0.3036832, Accumulated Leap Seconds=29
  { 2449755.50, 60.882846 }, //07 February 1995, UT1-UTC= 0.3011537, Accumulated Leap Seconds=29
  { 2449756.50, 60.885237 }, //08 February 1995, UT1-UTC= 0.2987626, Accumulated Leap Seconds=29
  { 2449757.50, 60.887499 }, //09 February 1995, UT1-UTC= 0.2965010, Accumulated Leap Seconds=29
  { 2449758.50, 60.889692 }, //10 February 1995, UT1-UTC= 0.2943080, Accumulated Leap Seconds=29
  { 2449759.50, 60.891857 }, //11 February 1995, UT1-UTC= 0.2921430, Accumulated Leap Seconds=29
  { 2449760.50, 60.894031 }, //12 February 1995, UT1-UTC= 0.2899689, Accumulated Leap Seconds=29
  { 2449761.50, 60.896265 }, //13 February 1995, UT1-UTC= 0.2877352, Accumulated Leap Seconds=29
  { 2449762.50, 60.898604 }, //14 February 1995, UT1-UTC= 0.2853965, Accumulated Leap Seconds=29
  { 2449763.50, 60.901084 }, //15 February 1995, UT1-UTC= 0.2829157, Accumulated Leap Seconds=29
  { 2449764.50, 60.903734 }, //16 February 1995, UT1-UTC= 0.2802655, Accumulated Leap Seconds=29
  { 2449765.50, 60.906517 }, //17 February 1995, UT1-UTC= 0.2774826, Accumulated Leap Seconds=29
  { 2449766.50, 60.909361 }, //18 February 1995, UT1-UTC= 0.2746389, Accumulated Leap Seconds=29
  { 2449767.50, 60.912200 }, //19 February 1995, UT1-UTC= 0.2718002, Accumulated Leap Seconds=29
  { 2449768.50, 60.914974 }, //20 February 1995, UT1-UTC= 0.2690260, Accumulated Leap Seconds=29
  { 2449769.50, 60.917645 }, //21 February 1995, UT1-UTC= 0.2663551, Accumulated Leap Seconds=29
  { 2449770.50, 60.920205 }, //22 February 1995, UT1-UTC= 0.2637950, Accumulated Leap Seconds=29
  { 2449771.50, 60.922678 }, //23 February 1995, UT1-UTC= 0.2613224, Accumulated Leap Seconds=29
  { 2449772.50, 60.925118 }, //24 February 1995, UT1-UTC= 0.2588823, Accumulated Leap Seconds=29
  { 2449773.50, 60.927601 }, //25 February 1995, UT1-UTC= 0.2563985, Accumulated Leap Seconds=29
  { 2449774.50, 60.930184 }, //26 February 1995, UT1-UTC= 0.2538155, Accumulated Leap Seconds=29
  { 2449775.50, 60.932896 }, //27 February 1995, UT1-UTC= 0.2511036, Accumulated Leap Seconds=29
  { 2449776.50, 60.935733 }, //28 February 1995, UT1-UTC= 0.2482672, Accumulated Leap Seconds=29
  { 2449777.50, 60.938667 }, //01 March 1995, UT1-UTC= 0.2453328, Accumulated Leap Seconds=29
  { 2449778.50, 60.941672 }, //02 March 1995, UT1-UTC= 0.2423278, Accumulated Leap Seconds=29
  { 2449779.50, 60.944717 }, //03 March 1995, UT1-UTC= 0.2392832, Accumulated Leap Seconds=29
  { 2449780.50, 60.947771 }, //04 March 1995, UT1-UTC= 0.2362293, Accumulated Leap Seconds=29
  { 2449781.50, 60.950780 }, //05 March 1995, UT1-UTC= 0.2332196, Accumulated Leap Seconds=29
  { 2449782.50, 60.953697 }, //06 March 1995, UT1-UTC= 0.2303025, Accumulated Leap Seconds=29
  { 2449783.50, 60.956492 }, //07 March 1995, UT1-UTC= 0.2275077, Accumulated Leap Seconds=29
  { 2449784.50, 60.959163 }, //08 March 1995, UT1-UTC= 0.2248365, Accumulated Leap Seconds=29
  { 2449785.50, 60.961786 }, //09 March 1995, UT1-UTC= 0.2222144, Accumulated Leap Seconds=29
  { 2449786.50, 60.964429 }, //10 March 1995, UT1-UTC= 0.2195706, Accumulated Leap Seconds=29
  { 2449787.50, 60.967129 }, //11 March 1995, UT1-UTC= 0.2168712, Accumulated Leap Seconds=29
  { 2449788.50, 60.969901 }, //12 March 1995, UT1-UTC= 0.2140985, Accumulated Leap Seconds=29
  { 2449789.50, 60.972763 }, //13 March 1995, UT1-UTC= 0.2112374, Accumulated Leap Seconds=29
  { 2449790.50, 60.975712 }, //14 March 1995, UT1-UTC= 0.2082877, Accumulated Leap Seconds=29
  { 2449791.50, 60.978753 }, //15 March 1995, UT1-UTC= 0.2052470, Accumulated Leap Seconds=29
  { 2449792.50, 60.981913 }, //16 March 1995, UT1-UTC= 0.2020871, Accumulated Leap Seconds=29
  { 2449793.50, 60.985170 }, //17 March 1995, UT1-UTC= 0.1988304, Accumulated Leap Seconds=29
  { 2449794.50, 60.988465 }, //18 March 1995, UT1-UTC= 0.1955348, Accumulated Leap Seconds=29
  { 2449795.50, 60.991720 }, //19 March 1995, UT1-UTC= 0.1922801, Accumulated Leap Seconds=29
  { 2449796.50, 60.994859 }, //20 March 1995, UT1-UTC= 0.1891405, Accumulated Leap Seconds=29
  { 2449797.50, 60.997811 }, //21 March 1995, UT1-UTC= 0.1861891, Accumulated Leap Seconds=29
  { 2449798.50, 61.000537 }, //22 March 1995, UT1-UTC= 0.1834628, Accumulated Leap Seconds=29
  { 2449799.50, 61.003077 }, //23 March 1995, UT1-UTC= 0.1809230, Accumulated Leap Seconds=29
  { 2449800.50, 61.005551 }, //24 March 1995, UT1-UTC= 0.1784494, Accumulated Leap Seconds=29
  { 2449801.50, 61.008063 }, //25 March 1995, UT1-UTC= 0.1759369, Accumulated Leap Seconds=29
  { 2449802.50, 61.010668 }, //26 March 1995, UT1-UTC= 0.1733321, Accumulated Leap Seconds=29
  { 2449803.50, 61.013382 }, //27 March 1995, UT1-UTC= 0.1706177, Accumulated Leap Seconds=29
  { 2449804.50, 61.016195 }, //28 March 1995, UT1-UTC= 0.1678049, Accumulated Leap Seconds=29
  { 2449805.50, 61.019072 }, //29 March 1995, UT1-UTC= 0.1649277, Accumulated Leap Seconds=29
  { 2449806.50, 61.021975 }, //30 March 1995, UT1-UTC= 0.1620246, Accumulated Leap Seconds=29
  { 2449807.50, 61.024857 }, //31 March 1995, UT1-UTC= 0.1591432, Accumulated Leap Seconds=29
  { 2449808.50, 61.027676 }, //01 April 1995, UT1-UTC= 0.1563243, Accumulated Leap Seconds=29
  { 2449809.50, 61.030410 }, //02 April 1995, UT1-UTC= 0.1535896, Accumulated Leap Seconds=29
  { 2449810.50, 61.033052 }, //03 April 1995, UT1-UTC= 0.1509483, Accumulated Leap Seconds=29
  { 2449811.50, 61.035602 }, //04 April 1995, UT1-UTC= 0.1483982, Accumulated Leap Seconds=29
  { 2449812.50, 61.038068 }, //05 April 1995, UT1-UTC= 0.1459323, Accumulated Leap Seconds=29
  { 2449813.50, 61.040488 }, //06 April 1995, UT1-UTC= 0.1435124, Accumulated Leap Seconds=29
  { 2449814.50, 61.042930 }, //07 April 1995, UT1-UTC= 0.1410697, Accumulated Leap Seconds=29
  { 2449815.50, 61.045437 }, //08 April 1995, UT1-UTC= 0.1385627, Accumulated Leap Seconds=29
  { 2449816.50, 61.048021 }, //09 April 1995, UT1-UTC= 0.1359785, Accumulated Leap Seconds=29
  { 2449817.50, 61.050693 }, //10 April 1995, UT1-UTC= 0.1333072, Accumulated Leap Seconds=29
  { 2449818.50, 61.053490 }, //11 April 1995, UT1-UTC= 0.1305101, Accumulated Leap Seconds=29
  { 2449819.50, 61.056437 }, //12 April 1995, UT1-UTC= 0.1275625, Accumulated Leap Seconds=29
  { 2449820.50, 61.059529 }, //13 April 1995, UT1-UTC= 0.1244709, Accumulated Leap Seconds=29
  { 2449821.50, 61.062682 }, //14 April 1995, UT1-UTC= 0.1213176, Accumulated Leap Seconds=29
  { 2449822.50, 61.065828 }, //15 April 1995, UT1-UTC= 0.1181716, Accumulated Leap Seconds=29
  { 2449823.50, 61.068916 }, //16 April 1995, UT1-UTC= 0.1150839, Accumulated Leap Seconds=29
  { 2449824.50, 61.071908 }, //17 April 1995, UT1-UTC= 0.1120916, Accumulated Leap Seconds=29
  { 2449825.50, 61.074782 }, //18 April 1995, UT1-UTC= 0.1092185, Accumulated Leap Seconds=29
  { 2449826.50, 61.077545 }, //19 April 1995, UT1-UTC= 0.1064552, Accumulated Leap Seconds=29
  { 2449827.50, 61.080219 }, //20 April 1995, UT1-UTC= 0.1037807, Accumulated Leap Seconds=29
  { 2449828.50, 61.082869 }, //21 April 1995, UT1-UTC= 0.1011314, Accumulated Leap Seconds=29
  { 2449829.50, 61.085563 }, //22 April 1995, UT1-UTC= 0.0984366, Accumulated Leap Seconds=29
  { 2449830.50, 61.088327 }, //23 April 1995, UT1-UTC= 0.0956726, Accumulated Leap Seconds=29
  { 2449831.50, 61.091166 }, //24 April 1995, UT1-UTC= 0.0928337, Accumulated Leap Seconds=29
  { 2449832.50, 61.094066 }, //25 April 1995, UT1-UTC= 0.0899338, Accumulated Leap Seconds=29
  { 2449833.50, 61.096996 }, //26 April 1995, UT1-UTC= 0.0870036, Accumulated Leap Seconds=29
  { 2449834.50, 61.099898 }, //27 April 1995, UT1-UTC= 0.0841019, Accumulated Leap Seconds=29
  { 2449835.50, 61.102709 }, //28 April 1995, UT1-UTC= 0.0812909, Accumulated Leap Seconds=29
  { 2449836.50, 61.105392 }, //29 April 1995, UT1-UTC= 0.0786080, Accumulated Leap Seconds=29
  { 2449837.50, 61.107934 }, //30 April 1995, UT1-UTC= 0.0760659, Accumulated Leap Seconds=29
  { 2449838.50, 61.110345 }, //01 May 1995, UT1-UTC= 0.0736552, Accumulated Leap Seconds=29
  { 2449839.50, 61.112680 }, //02 May 1995, UT1-UTC= 0.0713203, Accumulated Leap Seconds=29
  { 2449840.50, 61.115005 }, //03 May 1995, UT1-UTC= 0.0689948, Accumulated Leap Seconds=29
  { 2449841.50, 61.117368 }, //04 May 1995, UT1-UTC= 0.0666317, Accumulated Leap Seconds=29
  { 2449842.50, 61.119790 }, //05 May 1995, UT1-UTC= 0.0642096, Accumulated Leap Seconds=29
  { 2449843.50, 61.122278 }, //06 May 1995, UT1-UTC= 0.0617218, Accumulated Leap Seconds=29
  { 2449844.50, 61.124832 }, //07 May 1995, UT1-UTC= 0.0591679, Accumulated Leap Seconds=29
  { 2449845.50, 61.127460 }, //08 May 1995, UT1-UTC= 0.0565402, Accumulated Leap Seconds=29
  { 2449846.50, 61.130200 }, //09 May 1995, UT1-UTC= 0.0537998, Accumulated Leap Seconds=29
  { 2449847.50, 61.133073 }, //10 May 1995, UT1-UTC= 0.0509274, Accumulated Leap Seconds=29
  { 2449848.50, 61.136058 }, //11 May 1995, UT1-UTC= 0.0479419, Accumulated Leap Seconds=29
  { 2449849.50, 61.139096 }, //12 May 1995, UT1-UTC= 0.0449044, Accumulated Leap Seconds=29
  { 2449850.50, 61.142097 }, //13 May 1995, UT1-UTC= 0.0419025, Accumulated Leap Seconds=29
  { 2449851.50, 61.145018 }, //14 May 1995, UT1-UTC= 0.0389823, Accumulated Leap Seconds=29
  { 2449852.50, 61.147841 }, //15 May 1995, UT1-UTC= 0.0361589, Accumulated Leap Seconds=29
  { 2449853.50, 61.150554 }, //16 May 1995, UT1-UTC= 0.0334457, Accumulated Leap Seconds=29
  { 2449854.50, 61.153174 }, //17 May 1995, UT1-UTC= 0.0308256, Accumulated Leap Seconds=29
  { 2449855.50, 61.155738 }, //18 May 1995, UT1-UTC= 0.0282623, Accumulated Leap Seconds=29
  { 2449856.50, 61.158287 }, //19 May 1995, UT1-UTC= 0.0257127, Accumulated Leap Seconds=29
  { 2449857.50, 61.160849 }, //20 May 1995, UT1-UTC= 0.0231514, Accumulated Leap Seconds=29
  { 2449858.50, 61.163419 }, //21 May 1995, UT1-UTC= 0.0205808, Accumulated Leap Seconds=29
  { 2449859.50, 61.165970 }, //22 May 1995, UT1-UTC= 0.0180304, Accumulated Leap Seconds=29
  { 2449860.50, 61.168461 }, //23 May 1995, UT1-UTC= 0.0155386, Accumulated Leap Seconds=29
  { 2449861.50, 61.170856 }, //24 May 1995, UT1-UTC= 0.0131440, Accumulated Leap Seconds=29
  { 2449862.50, 61.173132 }, //25 May 1995, UT1-UTC= 0.0108681, Accumulated Leap Seconds=29
  { 2449863.50, 61.175316 }, //26 May 1995, UT1-UTC= 0.0086836, Accumulated Leap Seconds=29
  { 2449864.50, 61.177419 }, //27 May 1995, UT1-UTC= 0.0065814, Accumulated Leap Seconds=29
  { 2449865.50, 61.179430 }, //28 May 1995, UT1-UTC= 0.0045701, Accumulated Leap Seconds=29
  { 2449866.50, 61.181370 }, //29 May 1995, UT1-UTC= 0.0026303, Accumulated Leap Seconds=29
  { 2449867.50, 61.183265 }, //30 May 1995, UT1-UTC= 0.0007350, Accumulated Leap Seconds=29
  { 2449868.50, 61.185146 }, //31 May 1995, UT1-UTC=-0.0011460, Accumulated Leap Seconds=29
  { 2449869.50, 61.187046 }, //01 June 1995, UT1-UTC=-0.0030458, Accumulated Leap Seconds=29
  { 2449870.50, 61.189009 }, //02 June 1995, UT1-UTC=-0.0050094, Accumulated Leap Seconds=29
  { 2449871.50, 61.191072 }, //03 June 1995, UT1-UTC=-0.0070721, Accumulated Leap Seconds=29
  { 2449872.50, 61.193248 }, //04 June 1995, UT1-UTC=-0.0092482, Accumulated Leap Seconds=29
  { 2449873.50, 61.195523 }, //05 June 1995, UT1-UTC=-0.0115228, Accumulated Leap Seconds=29
  { 2449874.50, 61.197880 }, //06 June 1995, UT1-UTC=-0.0138798, Accumulated Leap Seconds=29
  { 2449875.50, 61.200297 }, //07 June 1995, UT1-UTC=-0.0162974, Accumulated Leap Seconds=29
  { 2449876.50, 61.202757 }, //08 June 1995, UT1-UTC=-0.0187573, Accumulated Leap Seconds=29
  { 2449877.50, 61.205203 }, //09 June 1995, UT1-UTC=-0.0212035, Accumulated Leap Seconds=29
  { 2449878.50, 61.207574 }, //10 June 1995, UT1-UTC=-0.0235742, Accumulated Leap Seconds=29
  { 2449879.50, 61.209833 }, //11 June 1995, UT1-UTC=-0.0258332, Accumulated Leap Seconds=29
  { 2449880.50, 61.211976 }, //12 June 1995, UT1-UTC=-0.0279756, Accumulated Leap Seconds=29
  { 2449881.50, 61.214031 }, //13 June 1995, UT1-UTC=-0.0300312, Accumulated Leap Seconds=29
  { 2449882.50, 61.216068 }, //14 June 1995, UT1-UTC=-0.0320683, Accumulated Leap Seconds=29
  { 2449883.50, 61.218149 }, //15 June 1995, UT1-UTC=-0.0341489, Accumulated Leap Seconds=29
  { 2449884.50, 61.220274 }, //16 June 1995, UT1-UTC=-0.0362738, Accumulated Leap Seconds=29
  { 2449885.50, 61.222466 }, //17 June 1995, UT1-UTC=-0.0384656, Accumulated Leap Seconds=29
  { 2449886.50, 61.224709 }, //18 June 1995, UT1-UTC=-0.0407087, Accumulated Leap Seconds=29
  { 2449887.50, 61.226956 }, //19 June 1995, UT1-UTC=-0.0429558, Accumulated Leap Seconds=29
  { 2449888.50, 61.229157 }, //20 June 1995, UT1-UTC=-0.0451571, Accumulated Leap Seconds=29
  { 2449889.50, 61.231269 }, //21 June 1995, UT1-UTC=-0.0472689, Accumulated Leap Seconds=29
  { 2449890.50, 61.233251 }, //22 June 1995, UT1-UTC=-0.0492511, Accumulated Leap Seconds=29
  { 2449891.50, 61.235065 }, //23 June 1995, UT1-UTC=-0.0510650, Accumulated Leap Seconds=29
  { 2449892.50, 61.236679 }, //24 June 1995, UT1-UTC=-0.0526787, Accumulated Leap Seconds=29
  { 2449893.50, 61.238096 }, //25 June 1995, UT1-UTC=-0.0540955, Accumulated Leap Seconds=29
  { 2449894.50, 61.239364 }, //26 June 1995, UT1-UTC=-0.0553638, Accumulated Leap Seconds=29
  { 2449895.50, 61.240533 }, //27 June 1995, UT1-UTC=-0.0565328, Accumulated Leap Seconds=29
  { 2449896.50, 61.241658 }, //28 June 1995, UT1-UTC=-0.0576580, Accumulated Leap Seconds=29
  { 2449897.50, 61.242807 }, //29 June 1995, UT1-UTC=-0.0588065, Accumulated Leap Seconds=29
  { 2449898.50, 61.244037 }, //30 June 1995, UT1-UTC=-0.0600368, Accumulated Leap Seconds=29
  { 2449899.50, 61.245389 }, //01 July 1995, UT1-UTC=-0.0613891, Accumulated Leap Seconds=29
  { 2449900.50, 61.246885 }, //02 July 1995, UT1-UTC=-0.0628846, Accumulated Leap Seconds=29
  { 2449901.50, 61.248522 }, //03 July 1995, UT1-UTC=-0.0645225, Accumulated Leap Seconds=29
  { 2449902.50, 61.250256 }, //04 July 1995, UT1-UTC=-0.0662561, Accumulated Leap Seconds=29
  { 2449903.50, 61.252021 }, //05 July 1995, UT1-UTC=-0.0680209, Accumulated Leap Seconds=29
  { 2449904.50, 61.253761 }, //06 July 1995, UT1-UTC=-0.0697609, Accumulated Leap Seconds=29
  { 2449905.50, 61.255447 }, //07 July 1995, UT1-UTC=-0.0714469, Accumulated Leap Seconds=29
  { 2449906.50, 61.257054 }, //08 July 1995, UT1-UTC=-0.0730542, Accumulated Leap Seconds=29
  { 2449907.50, 61.258572 }, //09 July 1995, UT1-UTC=-0.0745724, Accumulated Leap Seconds=29
  { 2449908.50, 61.260018 }, //10 July 1995, UT1-UTC=-0.0760182, Accumulated Leap Seconds=29
  { 2449909.50, 61.261434 }, //11 July 1995, UT1-UTC=-0.0774336, Accumulated Leap Seconds=29
  { 2449910.50, 61.262877 }, //12 July 1995, UT1-UTC=-0.0788772, Accumulated Leap Seconds=29
  { 2449911.50, 61.264394 }, //13 July 1995, UT1-UTC=-0.0803942, Accumulated Leap Seconds=29
  { 2449912.50, 61.265999 }, //14 July 1995, UT1-UTC=-0.0819987, Accumulated Leap Seconds=29
  { 2449913.50, 61.267699 }, //15 July 1995, UT1-UTC=-0.0836994, Accumulated Leap Seconds=29
  { 2449914.50, 61.269456 }, //16 July 1995, UT1-UTC=-0.0854564, Accumulated Leap Seconds=29
  { 2449915.50, 61.271203 }, //17 July 1995, UT1-UTC=-0.0872029, Accumulated Leap Seconds=29
  { 2449916.50, 61.272869 }, //18 July 1995, UT1-UTC=-0.0888687, Accumulated Leap Seconds=29
  { 2449917.50, 61.274403 }, //19 July 1995, UT1-UTC=-0.0904031, Accumulated Leap Seconds=29
  { 2449918.50, 61.275781 }, //20 July 1995, UT1-UTC=-0.0917809, Accumulated Leap Seconds=29
  { 2449919.50, 61.277000 }, //21 July 1995, UT1-UTC=-0.0929999, Accumulated Leap Seconds=29
  { 2449920.50, 61.278066 }, //22 July 1995, UT1-UTC=-0.0940663, Accumulated Leap Seconds=29
  { 2449921.50, 61.279008 }, //23 July 1995, UT1-UTC=-0.0950078, Accumulated Leap Seconds=29
  { 2449922.50, 61.279867 }, //24 July 1995, UT1-UTC=-0.0958674, Accumulated Leap Seconds=29
  { 2449923.50, 61.280687 }, //25 July 1995, UT1-UTC=-0.0966875, Accumulated Leap Seconds=29
  { 2449924.50, 61.281509 }, //26 July 1995, UT1-UTC=-0.0975086, Accumulated Leap Seconds=29
  { 2449925.50, 61.282368 }, //27 July 1995, UT1-UTC=-0.0983680, Accumulated Leap Seconds=29
  { 2449926.50, 61.283294 }, //28 July 1995, UT1-UTC=-0.0992940, Accumulated Leap Seconds=29
  { 2449927.50, 61.284314 }, //29 July 1995, UT1-UTC=-0.1003143, Accumulated Leap Seconds=29
  { 2449928.50, 61.285453 }, //30 July 1995, UT1-UTC=-0.1014530, Accumulated Leap Seconds=29
  { 2449929.50, 61.286723 }, //31 July 1995, UT1-UTC=-0.1027230, Accumulated Leap Seconds=29
  { 2449930.50, 61.288102 }, //01 August 1995, UT1-UTC=-0.1041024, Accumulated Leap Seconds=29
  { 2449931.50, 61.289542 }, //02 August 1995, UT1-UTC=-0.1055418, Accumulated Leap Seconds=29
  { 2449932.50, 61.290974 }, //03 August 1995, UT1-UTC=-0.1069739, Accumulated Leap Seconds=29
  { 2449933.50, 61.292359 }, //04 August 1995, UT1-UTC=-0.1083592, Accumulated Leap Seconds=29
  { 2449934.50, 61.293690 }, //05 August 1995, UT1-UTC=-0.1096903, Accumulated Leap Seconds=29
  { 2449935.50, 61.294990 }, //06 August 1995, UT1-UTC=-0.1109898, Accumulated Leap Seconds=29
  { 2449936.50, 61.296295 }, //07 August 1995, UT1-UTC=-0.1122951, Accumulated Leap Seconds=29
  { 2449937.50, 61.297648 }, //08 August 1995, UT1-UTC=-0.1136483, Accumulated Leap Seconds=29
  { 2449938.50, 61.299092 }, //09 August 1995, UT1-UTC=-0.1150922, Accumulated Leap Seconds=29
  { 2449939.50, 61.300654 }, //10 August 1995, UT1-UTC=-0.1166544, Accumulated Leap Seconds=29
  { 2449940.50, 61.302335 }, //11 August 1995, UT1-UTC=-0.1183355, Accumulated Leap Seconds=29
  { 2449941.50, 61.304096 }, //12 August 1995, UT1-UTC=-0.1200959, Accumulated Leap Seconds=29
  { 2449942.50, 61.305879 }, //13 August 1995, UT1-UTC=-0.1218791, Accumulated Leap Seconds=29
  { 2449943.50, 61.307626 }, //14 August 1995, UT1-UTC=-0.1236258, Accumulated Leap Seconds=29
  { 2449944.50, 61.309292 }, //15 August 1995, UT1-UTC=-0.1252919, Accumulated Leap Seconds=29
  { 2449945.50, 61.310859 }, //16 August 1995, UT1-UTC=-0.1268589, Accumulated Leap Seconds=29
  { 2449946.50, 61.312326 }, //17 August 1995, UT1-UTC=-0.1283263, Accumulated Leap Seconds=29
  { 2449947.50, 61.313701 }, //18 August 1995, UT1-UTC=-0.1297007, Accumulated Leap Seconds=29
  { 2449948.50, 61.315015 }, //19 August 1995, UT1-UTC=-0.1310153, Accumulated Leap Seconds=29
  { 2449949.50, 61.316314 }, //20 August 1995, UT1-UTC=-0.1323139, Accumulated Leap Seconds=29
  { 2449950.50, 61.317634 }, //21 August 1995, UT1-UTC=-0.1336336, Accumulated Leap Seconds=29
  { 2449951.50, 61.319007 }, //22 August 1995, UT1-UTC=-0.1350073, Accumulated Leap Seconds=29
  { 2449952.50, 61.320459 }, //23 August 1995, UT1-UTC=-0.1364590, Accumulated Leap Seconds=29
  { 2449953.50, 61.321997 }, //24 August 1995, UT1-UTC=-0.1379975, Accumulated Leap Seconds=29
  { 2449954.50, 61.323645 }, //25 August 1995, UT1-UTC=-0.1396448, Accumulated Leap Seconds=29
  { 2449955.50, 61.325440 }, //26 August 1995, UT1-UTC=-0.1414401, Accumulated Leap Seconds=29
  { 2449956.50, 61.327390 }, //27 August 1995, UT1-UTC=-0.1433903, Accumulated Leap Seconds=29
  { 2449957.50, 61.329454 }, //28 August 1995, UT1-UTC=-0.1454536, Accumulated Leap Seconds=29
  { 2449958.50, 61.331588 }, //29 August 1995, UT1-UTC=-0.1475881, Accumulated Leap Seconds=29
  { 2449959.50, 61.333700 }, //30 August 1995, UT1-UTC=-0.1496999, Accumulated Leap Seconds=29
  { 2449960.50, 61.335764 }, //31 August 1995, UT1-UTC=-0.1517644, Accumulated Leap Seconds=29
  { 2449961.50, 61.337780 }, //01 September 1995, UT1-UTC=-0.1537799, Accumulated Leap Seconds=29
  { 2449962.50, 61.339758 }, //02 September 1995, UT1-UTC=-0.1557584, Accumulated Leap Seconds=29
  { 2449963.50, 61.341719 }, //03 September 1995, UT1-UTC=-0.1577190, Accumulated Leap Seconds=29
  { 2449964.50, 61.343694 }, //04 September 1995, UT1-UTC=-0.1596935, Accumulated Leap Seconds=29
  { 2449965.50, 61.345719 }, //05 September 1995, UT1-UTC=-0.1617194, Accumulated Leap Seconds=29
  { 2449966.50, 61.347826 }, //06 September 1995, UT1-UTC=-0.1638257, Accumulated Leap Seconds=29
  { 2449967.50, 61.350019 }, //07 September 1995, UT1-UTC=-0.1660187, Accumulated Leap Seconds=29
  { 2449968.50, 61.352300 }, //08 September 1995, UT1-UTC=-0.1683000, Accumulated Leap Seconds=29
  { 2449969.50, 61.354645 }, //09 September 1995, UT1-UTC=-0.1706454, Accumulated Leap Seconds=29
  { 2449970.50, 61.356987 }, //10 September 1995, UT1-UTC=-0.1729869, Accumulated Leap Seconds=29
  { 2449971.50, 61.359256 }, //11 September 1995, UT1-UTC=-0.1752559, Accumulated Leap Seconds=29
  { 2449972.50, 61.361411 }, //12 September 1995, UT1-UTC=-0.1774114, Accumulated Leap Seconds=29
  { 2449973.50, 61.363442 }, //13 September 1995, UT1-UTC=-0.1794417, Accumulated Leap Seconds=29
  { 2449974.50, 61.365363 }, //14 September 1995, UT1-UTC=-0.1813634, Accumulated Leap Seconds=29
  { 2449975.50, 61.367215 }, //15 September 1995, UT1-UTC=-0.1832150, Accumulated Leap Seconds=29
  { 2449976.50, 61.369039 }, //16 September 1995, UT1-UTC=-0.1850386, Accumulated Leap Seconds=29
  { 2449977.50, 61.370866 }, //17 September 1995, UT1-UTC=-0.1868661, Accumulated Leap Seconds=29
  { 2449978.50, 61.372719 }, //18 September 1995, UT1-UTC=-0.1887192, Accumulated Leap Seconds=29
  { 2449979.50, 61.374617 }, //19 September 1995, UT1-UTC=-0.1906167, Accumulated Leap Seconds=29
  { 2449980.50, 61.376586 }, //20 September 1995, UT1-UTC=-0.1925865, Accumulated Leap Seconds=29
  { 2449981.50, 61.378700 }, //21 September 1995, UT1-UTC=-0.1946996, Accumulated Leap Seconds=29
  { 2449982.50, 61.380985 }, //22 September 1995, UT1-UTC=-0.1969854, Accumulated Leap Seconds=29
  { 2449983.50, 61.383434 }, //23 September 1995, UT1-UTC=-0.1994339, Accumulated Leap Seconds=29
  { 2449984.50, 61.386016 }, //24 September 1995, UT1-UTC=-0.2020161, Accumulated Leap Seconds=29
  { 2449985.50, 61.388684 }, //25 September 1995, UT1-UTC=-0.2046839, Accumulated Leap Seconds=29
  { 2449986.50, 61.391375 }, //26 September 1995, UT1-UTC=-0.2073748, Accumulated Leap Seconds=29
  { 2449987.50, 61.394020 }, //27 September 1995, UT1-UTC=-0.2100204, Accumulated Leap Seconds=29
  { 2449988.50, 61.396556 }, //28 September 1995, UT1-UTC=-0.2125563, Accumulated Leap Seconds=29
  { 2449989.50, 61.398976 }, //29 September 1995, UT1-UTC=-0.2149760, Accumulated Leap Seconds=29
  { 2449990.50, 61.401311 }, //30 September 1995, UT1-UTC=-0.2173114, Accumulated Leap Seconds=29
  { 2449991.50, 61.403616 }, //01 October 1995, UT1-UTC=-0.2196165, Accumulated Leap Seconds=29
  { 2449992.50, 61.405952 }, //02 October 1995, UT1-UTC=-0.2219523, Accumulated Leap Seconds=29
  { 2449993.50, 61.408366 }, //03 October 1995, UT1-UTC=-0.2243660, Accumulated Leap Seconds=29
  { 2449994.50, 61.410875 }, //04 October 1995, UT1-UTC=-0.2268749, Accumulated Leap Seconds=29
  { 2449995.50, 61.413455 }, //05 October 1995, UT1-UTC=-0.2294547, Accumulated Leap Seconds=29
  { 2449996.50, 61.416093 }, //06 October 1995, UT1-UTC=-0.2320927, Accumulated Leap Seconds=29
  { 2449997.50, 61.418768 }, //07 October 1995, UT1-UTC=-0.2347681, Accumulated Leap Seconds=29
  { 2449998.50, 61.421438 }, //08 October 1995, UT1-UTC=-0.2374380, Accumulated Leap Seconds=29
  { 2449999.50, 61.424040 }, //09 October 1995, UT1-UTC=-0.2400403, Accumulated Leap Seconds=29
  { 2450000.50, 61.426511 }, //10 October 1995, UT1-UTC=-0.2425108, Accumulated Leap Seconds=29
  { 2450001.50, 61.428819 }, //11 October 1995, UT1-UTC=-0.2448193, Accumulated Leap Seconds=29
  { 2450002.50, 61.430991 }, //12 October 1995, UT1-UTC=-0.2469912, Accumulated Leap Seconds=29
  { 2450003.50, 61.433039 }, //13 October 1995, UT1-UTC=-0.2490387, Accumulated Leap Seconds=29
  { 2450004.50, 61.434989 }, //14 October 1995, UT1-UTC=-0.2509891, Accumulated Leap Seconds=29
  { 2450005.50, 61.436900 }, //15 October 1995, UT1-UTC=-0.2529003, Accumulated Leap Seconds=29
  { 2450006.50, 61.438819 }, //16 October 1995, UT1-UTC=-0.2548192, Accumulated Leap Seconds=29
  { 2450007.50, 61.440793 }, //17 October 1995, UT1-UTC=-0.2567932, Accumulated Leap Seconds=29
  { 2450008.50, 61.442861 }, //18 October 1995, UT1-UTC=-0.2588614, Accumulated Leap Seconds=29
  { 2450009.50, 61.445021 }, //19 October 1995, UT1-UTC=-0.2610211, Accumulated Leap Seconds=29
  { 2450010.50, 61.447262 }, //20 October 1995, UT1-UTC=-0.2632617, Accumulated Leap Seconds=29
  { 2450011.50, 61.449605 }, //21 October 1995, UT1-UTC=-0.2656050, Accumulated Leap Seconds=29
  { 2450012.50, 61.452036 }, //22 October 1995, UT1-UTC=-0.2680359, Accumulated Leap Seconds=29
  { 2450013.50, 61.454503 }, //23 October 1995, UT1-UTC=-0.2705028, Accumulated Leap Seconds=29
  { 2450014.50, 61.456965 }, //24 October 1995, UT1-UTC=-0.2729652, Accumulated Leap Seconds=29
  { 2450015.50, 61.459391 }, //25 October 1995, UT1-UTC=-0.2753909, Accumulated Leap Seconds=29
  { 2450016.50, 61.461756 }, //26 October 1995, UT1-UTC=-0.2777560, Accumulated Leap Seconds=29
  { 2450017.50, 61.464050 }, //27 October 1995, UT1-UTC=-0.2800497, Accumulated Leap Seconds=29
  { 2450018.50, 61.466302 }, //28 October 1995, UT1-UTC=-0.2823023, Accumulated Leap Seconds=29
  { 2450019.50, 61.468577 }, //29 October 1995, UT1-UTC=-0.2845771, Accumulated Leap Seconds=29
  { 2450020.50, 61.470934 }, //30 October 1995, UT1-UTC=-0.2869342, Accumulated Leap Seconds=29
  { 2450021.50, 61.473416 }, //31 October 1995, UT1-UTC=-0.2894157, Accumulated Leap Seconds=29
  { 2450022.50, 61.476037 }, //01 November 1995, UT1-UTC=-0.2920366, Accumulated Leap Seconds=29
  { 2450023.50, 61.478778 }, //02 November 1995, UT1-UTC=-0.2947781, Accumulated Leap Seconds=29
  { 2450024.50, 61.481552 }, //03 November 1995, UT1-UTC=-0.2975519, Accumulated Leap Seconds=29
  { 2450025.50, 61.484275 }, //04 November 1995, UT1-UTC=-0.3002748, Accumulated Leap Seconds=29
  { 2450026.50, 61.486900 }, //05 November 1995, UT1-UTC=-0.3028996, Accumulated Leap Seconds=29
  { 2450027.50, 61.489398 }, //06 November 1995, UT1-UTC=-0.3053979, Accumulated Leap Seconds=29
  { 2450028.50, 61.491771 }, //07 November 1995, UT1-UTC=-0.3077707, Accumulated Leap Seconds=29
  { 2450029.50, 61.494033 }, //08 November 1995, UT1-UTC=-0.3100334, Accumulated Leap Seconds=29
  { 2450030.50, 61.496191 }, //09 November 1995, UT1-UTC=-0.3121911, Accumulated Leap Seconds=29
  { 2450031.50, 61.498240 }, //10 November 1995, UT1-UTC=-0.3142402, Accumulated Leap Seconds=29
  { 2450032.50, 61.500219 }, //11 November 1995, UT1-UTC=-0.3162194, Accumulated Leap Seconds=29
  { 2450033.50, 61.502192 }, //12 November 1995, UT1-UTC=-0.3181918, Accumulated Leap Seconds=29
  { 2450034.50, 61.504225 }, //13 November 1995, UT1-UTC=-0.3202248, Accumulated Leap Seconds=29
  { 2450035.50, 61.506368 }, //14 November 1995, UT1-UTC=-0.3223682, Accumulated Leap Seconds=29
  { 2450036.50, 61.508665 }, //15 November 1995, UT1-UTC=-0.3246648, Accumulated Leap Seconds=29
  { 2450037.50, 61.511156 }, //16 November 1995, UT1-UTC=-0.3271557, Accumulated Leap Seconds=29
  { 2450038.50, 61.513805 }, //17 November 1995, UT1-UTC=-0.3298054, Accumulated Leap Seconds=29
  { 2450039.50, 61.516557 }, //18 November 1995, UT1-UTC=-0.3325569, Accumulated Leap Seconds=29
  { 2450040.50, 61.519391 }, //19 November 1995, UT1-UTC=-0.3353905, Accumulated Leap Seconds=29
  { 2450041.50, 61.522265 }, //20 November 1995, UT1-UTC=-0.3382653, Accumulated Leap Seconds=29
  { 2450042.50, 61.525132 }, //21 November 1995, UT1-UTC=-0.3411317, Accumulated Leap Seconds=29
  { 2450043.50, 61.527947 }, //22 November 1995, UT1-UTC=-0.3439468, Accumulated Leap Seconds=29
  { 2450044.50, 61.530679 }, //23 November 1995, UT1-UTC=-0.3466791, Accumulated Leap Seconds=29
  { 2450045.50, 61.533339 }, //24 November 1995, UT1-UTC=-0.3493391, Accumulated Leap Seconds=29
  { 2450046.50, 61.535967 }, //25 November 1995, UT1-UTC=-0.3519674, Accumulated Leap Seconds=29
  { 2450047.50, 61.538612 }, //26 November 1995, UT1-UTC=-0.3546121, Accumulated Leap Seconds=29
  { 2450048.50, 61.541311 }, //27 November 1995, UT1-UTC=-0.3573108, Accumulated Leap Seconds=29
  { 2450049.50, 61.544071 }, //28 November 1995, UT1-UTC=-0.3600711, Accumulated Leap Seconds=29
  { 2450050.50, 61.546878 }, //29 November 1995, UT1-UTC=-0.3628777, Accumulated Leap Seconds=29
  { 2450051.50, 61.549696 }, //30 November 1995, UT1-UTC=-0.3656959, Accumulated Leap Seconds=29
  { 2450052.50, 61.552460 }, //01 December 1995, UT1-UTC=-0.3684599, Accumulated Leap Seconds=29
  { 2450053.50, 61.555122 }, //02 December 1995, UT1-UTC=-0.3711221, Accumulated Leap Seconds=29
  { 2450054.50, 61.557674 }, //03 December 1995, UT1-UTC=-0.3736736, Accumulated Leap Seconds=29
  { 2450055.50, 61.560121 }, //04 December 1995, UT1-UTC=-0.3761212, Accumulated Leap Seconds=29
  { 2450056.50, 61.562456 }, //05 December 1995, UT1-UTC=-0.3784556, Accumulated Leap Seconds=29
  { 2450057.50, 61.564681 }, //06 December 1995, UT1-UTC=-0.3806815, Accumulated Leap Seconds=29
  { 2450058.50, 61.566816 }, //07 December 1995, UT1-UTC=-0.3828161, Accumulated Leap Seconds=29
  { 2450059.50, 61.568907 }, //08 December 1995, UT1-UTC=-0.3849067, Accumulated Leap Seconds=29
  { 2450060.50, 61.571006 }, //09 December 1995, UT1-UTC=-0.3870061, Accumulated Leap Seconds=29
  { 2450061.50, 61.573161 }, //10 December 1995, UT1-UTC=-0.3891605, Accumulated Leap Seconds=29
  { 2450062.50, 61.575392 }, //11 December 1995, UT1-UTC=-0.3913921, Accumulated Leap Seconds=29
  { 2450063.50, 61.577714 }, //12 December 1995, UT1-UTC=-0.3937138, Accumulated Leap Seconds=29
  { 2450064.50, 61.580141 }, //13 December 1995, UT1-UTC=-0.3961415, Accumulated Leap Seconds=29
  { 2450065.50, 61.582689 }, //14 December 1995, UT1-UTC=-0.3986893, Accumulated Leap Seconds=29
  { 2450066.50, 61.585310 }, //15 December 1995, UT1-UTC=-0.4013100, Accumulated Leap Seconds=29
  { 2450067.50, 61.587956 }, //16 December 1995, UT1-UTC=-0.4039556, Accumulated Leap Seconds=29
  { 2450068.50, 61.590607 }, //17 December 1995, UT1-UTC=-0.4066069, Accumulated Leap Seconds=29
  { 2450069.50, 61.593260 }, //18 December 1995, UT1-UTC=-0.4092604, Accumulated Leap Seconds=29
  { 2450070.50, 61.595886 }, //19 December 1995, UT1-UTC=-0.4118858, Accumulated Leap Seconds=29
  { 2450071.50, 61.598458 }, //20 December 1995, UT1-UTC=-0.4144579, Accumulated Leap Seconds=29
  { 2450072.50, 61.600957 }, //21 December 1995, UT1-UTC=-0.4169567, Accumulated Leap Seconds=29
  { 2450073.50, 61.603401 }, //22 December 1995, UT1-UTC=-0.4194010, Accumulated Leap Seconds=29
  { 2450074.50, 61.605853 }, //23 December 1995, UT1-UTC=-0.4218530, Accumulated Leap Seconds=29
  { 2450075.50, 61.608392 }, //24 December 1995, UT1-UTC=-0.4243922, Accumulated Leap Seconds=29
  { 2450076.50, 61.611057 }, //25 December 1995, UT1-UTC=-0.4270568, Accumulated Leap Seconds=29
  { 2450077.50, 61.613814 }, //26 December 1995, UT1-UTC=-0.4298138, Accumulated Leap Seconds=29
  { 2450078.50, 61.616597 }, //27 December 1995, UT1-UTC=-0.4325970, Accumulated Leap Seconds=29
  { 2450079.50, 61.619331 }, //28 December 1995, UT1-UTC=-0.4353306, Accumulated Leap Seconds=29
  { 2450080.50, 61.621945 }, //29 December 1995, UT1-UTC=-0.4379451, Accumulated Leap Seconds=29
  { 2450081.50, 61.624387 }, //30 December 1995, UT1-UTC=-0.4403875, Accumulated Leap Seconds=29
  { 2450082.50, 61.626627 }, //31 December 1995, UT1-UTC=-0.4426274, Accumulated Leap Seconds=29
  { 2450083.50, 61.628659 }, //01 January 1996, UT1-UTC= 0.5553407, Accumulated Leap Seconds=30
  { 2450084.50, 61.630503 }, //02 January 1996, UT1-UTC= 0.5534970, Accumulated Leap Seconds=30
  { 2450085.50, 61.632198 }, //03 January 1996, UT1-UTC= 0.5518018, Accumulated Leap Seconds=30
  { 2450086.50, 61.633791 }, //04 January 1996, UT1-UTC= 0.5502086, Accumulated Leap Seconds=30
  { 2450087.50, 61.635334 }, //05 January 1996, UT1-UTC= 0.5486657, Accumulated Leap Seconds=30
  { 2450088.50, 61.636888 }, //06 January 1996, UT1-UTC= 0.5471115, Accumulated Leap Seconds=30
  { 2450089.50, 61.638520 }, //07 January 1996, UT1-UTC= 0.5454805, Accumulated Leap Seconds=30
  { 2450090.50, 61.640266 }, //08 January 1996, UT1-UTC= 0.5437342, Accumulated Leap Seconds=30
  { 2450091.50, 61.642122 }, //09 January 1996, UT1-UTC= 0.5418781, Accumulated Leap Seconds=30
  { 2450092.50, 61.644054 }, //10 January 1996, UT1-UTC= 0.5399456, Accumulated Leap Seconds=30
  { 2450093.50, 61.646030 }, //11 January 1996, UT1-UTC= 0.5379699, Accumulated Leap Seconds=30
  { 2450094.50, 61.648025 }, //12 January 1996, UT1-UTC= 0.5359751, Accumulated Leap Seconds=30
  { 2450095.50, 61.650025 }, //13 January 1996, UT1-UTC= 0.5339752, Accumulated Leap Seconds=30
  { 2450096.50, 61.652017 }, //14 January 1996, UT1-UTC= 0.5319830, Accumulated Leap Seconds=30
  { 2450097.50, 61.653984 }, //15 January 1996, UT1-UTC= 0.5300156, Accumulated Leap Seconds=30
  { 2450098.50, 61.655909 }, //16 January 1996, UT1-UTC= 0.5280906, Accumulated Leap Seconds=30
  { 2450099.50, 61.657768 }, //17 January 1996, UT1-UTC= 0.5262316, Accumulated Leap Seconds=30
  { 2450100.50, 61.659546 }, //18 January 1996, UT1-UTC= 0.5244541, Accumulated Leap Seconds=30
  { 2450101.50, 61.661299 }, //19 January 1996, UT1-UTC= 0.5227015, Accumulated Leap Seconds=30
  { 2450102.50, 61.663118 }, //20 January 1996, UT1-UTC= 0.5208818, Accumulated Leap Seconds=30
  { 2450103.50, 61.665063 }, //21 January 1996, UT1-UTC= 0.5189365, Accumulated Leap Seconds=30
  { 2450104.50, 61.667138 }, //22 January 1996, UT1-UTC= 0.5168623, Accumulated Leap Seconds=30
  { 2450105.50, 61.669284 }, //23 January 1996, UT1-UTC= 0.5147163, Accumulated Leap Seconds=30
  { 2450106.50, 61.671413 }, //24 January 1996, UT1-UTC= 0.5125874, Accumulated Leap Seconds=30
  { 2450107.50, 61.673468 }, //25 January 1996, UT1-UTC= 0.5105325, Accumulated Leap Seconds=30
  { 2450108.50, 61.675405 }, //26 January 1996, UT1-UTC= 0.5085954, Accumulated Leap Seconds=30
  { 2450109.50, 61.677205 }, //27 January 1996, UT1-UTC= 0.5067951, Accumulated Leap Seconds=30
  { 2450110.50, 61.678861 }, //28 January 1996, UT1-UTC= 0.5051385, Accumulated Leap Seconds=30
  { 2450111.50, 61.680389 }, //29 January 1996, UT1-UTC= 0.5036110, Accumulated Leap Seconds=30
  { 2450112.50, 61.681819 }, //30 January 1996, UT1-UTC= 0.5021810, Accumulated Leap Seconds=30
  { 2450113.50, 61.683196 }, //31 January 1996, UT1-UTC= 0.5008035, Accumulated Leap Seconds=30
  { 2450114.50, 61.684582 }, //01 February 1996, UT1-UTC= 0.4994181, Accumulated Leap Seconds=30
  { 2450115.50, 61.686029 }, //02 February 1996, UT1-UTC= 0.4979707, Accumulated Leap Seconds=30
  { 2450116.50, 61.687577 }, //03 February 1996, UT1-UTC= 0.4964234, Accumulated Leap Seconds=30
  { 2450117.50, 61.689235 }, //04 February 1996, UT1-UTC= 0.4947646, Accumulated Leap Seconds=30
  { 2450118.50, 61.691002 }, //05 February 1996, UT1-UTC= 0.4929982, Accumulated Leap Seconds=30
  { 2450119.50, 61.692876 }, //06 February 1996, UT1-UTC= 0.4911237, Accumulated Leap Seconds=30
  { 2450120.50, 61.694865 }, //07 February 1996, UT1-UTC= 0.4891346, Accumulated Leap Seconds=30
  { 2450121.50, 61.696955 }, //08 February 1996, UT1-UTC= 0.4870454, Accumulated Leap Seconds=30
  { 2450122.50, 61.699099 }, //09 February 1996, UT1-UTC= 0.4849013, Accumulated Leap Seconds=30
  { 2450123.50, 61.701243 }, //10 February 1996, UT1-UTC= 0.4827566, Accumulated Leap Seconds=30
  { 2450124.50, 61.703345 }, //11 February 1996, UT1-UTC= 0.4806545, Accumulated Leap Seconds=30
  { 2450125.50, 61.705382 }, //12 February 1996, UT1-UTC= 0.4786184, Accumulated Leap Seconds=30
  { 2450126.50, 61.707352 }, //13 February 1996, UT1-UTC= 0.4766483, Accumulated Leap Seconds=30
  { 2450127.50, 61.709280 }, //14 February 1996, UT1-UTC= 0.4747202, Accumulated Leap Seconds=30
  { 2450128.50, 61.711209 }, //15 February 1996, UT1-UTC= 0.4727912, Accumulated Leap Seconds=30
  { 2450129.50, 61.713205 }, //16 February 1996, UT1-UTC= 0.4707947, Accumulated Leap Seconds=30
  { 2450130.50, 61.715339 }, //17 February 1996, UT1-UTC= 0.4686611, Accumulated Leap Seconds=30
  { 2450131.50, 61.717645 }, //18 February 1996, UT1-UTC= 0.4663545, Accumulated Leap Seconds=30
  { 2450132.50, 61.720111 }, //19 February 1996, UT1-UTC= 0.4638889, Accumulated Leap Seconds=30
  { 2450133.50, 61.722678 }, //20 February 1996, UT1-UTC= 0.4613222, Accumulated Leap Seconds=30
  { 2450134.50, 61.725262 }, //21 February 1996, UT1-UTC= 0.4587376, Accumulated Leap Seconds=30
  { 2450135.50, 61.727783 }, //22 February 1996, UT1-UTC= 0.4562170, Accumulated Leap Seconds=30
  { 2450136.50, 61.730148 }, //23 February 1996, UT1-UTC= 0.4538521, Accumulated Leap Seconds=30
  { 2450137.50, 61.732289 }, //24 February 1996, UT1-UTC= 0.4517114, Accumulated Leap Seconds=30
  { 2450138.50, 61.734247 }, //25 February 1996, UT1-UTC= 0.4497529, Accumulated Leap Seconds=30
  { 2450139.50, 61.736094 }, //26 February 1996, UT1-UTC= 0.4479063, Accumulated Leap Seconds=30
  { 2450140.50, 61.737886 }, //27 February 1996, UT1-UTC= 0.4461141, Accumulated Leap Seconds=30
  { 2450141.50, 61.739659 }, //28 February 1996, UT1-UTC= 0.4443409, Accumulated Leap Seconds=30
  { 2450142.50, 61.741445 }, //29 February 1996, UT1-UTC= 0.4425555, Accumulated Leap Seconds=30
  { 2450143.50, 61.743306 }, //01 March 1996, UT1-UTC= 0.4406940, Accumulated Leap Seconds=30
  { 2450144.50, 61.745292 }, //02 March 1996, UT1-UTC= 0.4387084, Accumulated Leap Seconds=30
  { 2450145.50, 61.747420 }, //03 March 1996, UT1-UTC= 0.4365804, Accumulated Leap Seconds=30
  { 2450146.50, 61.749699 }, //04 March 1996, UT1-UTC= 0.4343005, Accumulated Leap Seconds=30
  { 2450147.50, 61.752116 }, //05 March 1996, UT1-UTC= 0.4318840, Accumulated Leap Seconds=30
  { 2450148.50, 61.754636 }, //06 March 1996, UT1-UTC= 0.4293637, Accumulated Leap Seconds=30
  { 2450149.50, 61.757221 }, //07 March 1996, UT1-UTC= 0.4267790, Accumulated Leap Seconds=30
  { 2450150.50, 61.759834 }, //08 March 1996, UT1-UTC= 0.4241658, Accumulated Leap Seconds=30
  { 2450151.50, 61.762440 }, //09 March 1996, UT1-UTC= 0.4215597, Accumulated Leap Seconds=30
  { 2450152.50, 61.765012 }, //10 March 1996, UT1-UTC= 0.4189877, Accumulated Leap Seconds=30
  { 2450153.50, 61.767527 }, //11 March 1996, UT1-UTC= 0.4164726, Accumulated Leap Seconds=30
  { 2450154.50, 61.769951 }, //12 March 1996, UT1-UTC= 0.4140486, Accumulated Leap Seconds=30
  { 2450155.50, 61.772258 }, //13 March 1996, UT1-UTC= 0.4117416, Accumulated Leap Seconds=30
  { 2450156.50, 61.774466 }, //14 March 1996, UT1-UTC= 0.4095339, Accumulated Leap Seconds=30
  { 2450157.50, 61.776650 }, //15 March 1996, UT1-UTC= 0.4073504, Accumulated Leap Seconds=30
  { 2450158.50, 61.778879 }, //16 March 1996, UT1-UTC= 0.4051208, Accumulated Leap Seconds=30
  { 2450159.50, 61.781189 }, //17 March 1996, UT1-UTC= 0.4028109, Accumulated Leap Seconds=30
  { 2450160.50, 61.783576 }, //18 March 1996, UT1-UTC= 0.4004241, Accumulated Leap Seconds=30
  { 2450161.50, 61.786010 }, //19 March 1996, UT1-UTC= 0.3979900, Accumulated Leap Seconds=30
  { 2450162.50, 61.788450 }, //20 March 1996, UT1-UTC= 0.3955500, Accumulated Leap Seconds=30
  { 2450163.50, 61.790847 }, //21 March 1996, UT1-UTC= 0.3931531, Accumulated Leap Seconds=30
  { 2450164.50, 61.793142 }, //22 March 1996, UT1-UTC= 0.3908577, Accumulated Leap Seconds=30
  { 2450165.50, 61.795303 }, //23 March 1996, UT1-UTC= 0.3886966, Accumulated Leap Seconds=30
  { 2450166.50, 61.797339 }, //24 March 1996, UT1-UTC= 0.3866611, Accumulated Leap Seconds=30
  { 2450167.50, 61.799285 }, //25 March 1996, UT1-UTC= 0.3847153, Accumulated Leap Seconds=30
  { 2450168.50, 61.801173 }, //26 March 1996, UT1-UTC= 0.3828266, Accumulated Leap Seconds=30
  { 2450169.50, 61.803038 }, //27 March 1996, UT1-UTC= 0.3809615, Accumulated Leap Seconds=30
  { 2450170.50, 61.804917 }, //28 March 1996, UT1-UTC= 0.3790829, Accumulated Leap Seconds=30
  { 2450171.50, 61.806852 }, //29 March 1996, UT1-UTC= 0.3771484, Accumulated Leap Seconds=30
  { 2450172.50, 61.808872 }, //30 March 1996, UT1-UTC= 0.3751277, Accumulated Leap Seconds=30
  { 2450173.50, 61.810998 }, //31 March 1996, UT1-UTC= 0.3730019, Accumulated Leap Seconds=30
  { 2450174.50, 61.813242 }, //01 April 1996, UT1-UTC= 0.3707575, Accumulated Leap Seconds=30
  { 2450175.50, 61.815613 }, //02 April 1996, UT1-UTC= 0.3683867, Accumulated Leap Seconds=30
  { 2450176.50, 61.818085 }, //03 April 1996, UT1-UTC= 0.3659152, Accumulated Leap Seconds=30
  { 2450177.50, 61.820589 }, //04 April 1996, UT1-UTC= 0.3634106, Accumulated Leap Seconds=30
  { 2450178.50, 61.823091 }, //05 April 1996, UT1-UTC= 0.3609086, Accumulated Leap Seconds=30
  { 2450179.50, 61.825565 }, //06 April 1996, UT1-UTC= 0.3584349, Accumulated Leap Seconds=30
  { 2450180.50, 61.827981 }, //07 April 1996, UT1-UTC= 0.3560192, Accumulated Leap Seconds=30
  { 2450181.50, 61.830326 }, //08 April 1996, UT1-UTC= 0.3536740, Accumulated Leap Seconds=30
  { 2450182.50, 61.832610 }, //09 April 1996, UT1-UTC= 0.3513897, Accumulated Leap Seconds=30
  { 2450183.50, 61.834874 }, //10 April 1996, UT1-UTC= 0.3491265, Accumulated Leap Seconds=30
  { 2450184.50, 61.837176 }, //11 April 1996, UT1-UTC= 0.3468236, Accumulated Leap Seconds=30
  { 2450185.50, 61.839565 }, //12 April 1996, UT1-UTC= 0.3444346, Accumulated Leap Seconds=30
  { 2450186.50, 61.842054 }, //13 April 1996, UT1-UTC= 0.3419458, Accumulated Leap Seconds=30
  { 2450187.50, 61.844619 }, //14 April 1996, UT1-UTC= 0.3393814, Accumulated Leap Seconds=30
  { 2450188.50, 61.847207 }, //15 April 1996, UT1-UTC= 0.3367926, Accumulated Leap Seconds=30
  { 2450189.50, 61.849766 }, //16 April 1996, UT1-UTC= 0.3342341, Accumulated Leap Seconds=30
  { 2450190.50, 61.852246 }, //17 April 1996, UT1-UTC= 0.3317535, Accumulated Leap Seconds=30
  { 2450191.50, 61.854620 }, //18 April 1996, UT1-UTC= 0.3293800, Accumulated Leap Seconds=30
  { 2450192.50, 61.856840 }, //19 April 1996, UT1-UTC= 0.3271601, Accumulated Leap Seconds=30
  { 2450193.50, 61.858907 }, //20 April 1996, UT1-UTC= 0.3250932, Accumulated Leap Seconds=30
  { 2450194.50, 61.860868 }, //21 April 1996, UT1-UTC= 0.3231316, Accumulated Leap Seconds=30
  { 2450195.50, 61.862780 }, //22 April 1996, UT1-UTC= 0.3212204, Accumulated Leap Seconds=30
  { 2450196.50, 61.864670 }, //23 April 1996, UT1-UTC= 0.3193302, Accumulated Leap Seconds=30
  { 2450197.50, 61.866571 }, //24 April 1996, UT1-UTC= 0.3174294, Accumulated Leap Seconds=30
  { 2450198.50, 61.868520 }, //25 April 1996, UT1-UTC= 0.3154795, Accumulated Leap Seconds=30
  { 2450199.50, 61.870543 }, //26 April 1996, UT1-UTC= 0.3134574, Accumulated Leap Seconds=30
  { 2450200.50, 61.872671 }, //27 April 1996, UT1-UTC= 0.3113287, Accumulated Leap Seconds=30
  { 2450201.50, 61.874923 }, //28 April 1996, UT1-UTC= 0.3090773, Accumulated Leap Seconds=30
  { 2450202.50, 61.877291 }, //29 April 1996, UT1-UTC= 0.3067091, Accumulated Leap Seconds=30
  { 2450203.50, 61.879763 }, //30 April 1996, UT1-UTC= 0.3042372, Accumulated Leap Seconds=30
  { 2450204.50, 61.882320 }, //01 May 1996, UT1-UTC= 0.3016797, Accumulated Leap Seconds=30
  { 2450205.50, 61.884939 }, //02 May 1996, UT1-UTC= 0.2990606, Accumulated Leap Seconds=30
  { 2450206.50, 61.887564 }, //03 May 1996, UT1-UTC= 0.2964355, Accumulated Leap Seconds=30
  { 2450207.50, 61.890128 }, //04 May 1996, UT1-UTC= 0.2938716, Accumulated Leap Seconds=30
  { 2450208.50, 61.892597 }, //05 May 1996, UT1-UTC= 0.2914033, Accumulated Leap Seconds=30
  { 2450209.50, 61.894974 }, //06 May 1996, UT1-UTC= 0.2890261, Accumulated Leap Seconds=30
  { 2450210.50, 61.897302 }, //07 May 1996, UT1-UTC= 0.2866977, Accumulated Leap Seconds=30
  { 2450211.50, 61.899643 }, //08 May 1996, UT1-UTC= 0.2843570, Accumulated Leap Seconds=30
  { 2450212.50, 61.902052 }, //09 May 1996, UT1-UTC= 0.2819478, Accumulated Leap Seconds=30
  { 2450213.50, 61.904554 }, //10 May 1996, UT1-UTC= 0.2794457, Accumulated Leap Seconds=30
  { 2450214.50, 61.907154 }, //11 May 1996, UT1-UTC= 0.2768461, Accumulated Leap Seconds=30
  { 2450215.50, 61.909821 }, //12 May 1996, UT1-UTC= 0.2741790, Accumulated Leap Seconds=30
  { 2450216.50, 61.912493 }, //13 May 1996, UT1-UTC= 0.2715067, Accumulated Leap Seconds=30
  { 2450217.50, 61.915105 }, //14 May 1996, UT1-UTC= 0.2688952, Accumulated Leap Seconds=30
  { 2450218.50, 61.917603 }, //15 May 1996, UT1-UTC= 0.2663971, Accumulated Leap Seconds=30
  { 2450219.50, 61.919961 }, //16 May 1996, UT1-UTC= 0.2640392, Accumulated Leap Seconds=30
  { 2450220.50, 61.922152 }, //17 May 1996, UT1-UTC= 0.2618483, Accumulated Leap Seconds=30
  { 2450221.50, 61.924175 }, //18 May 1996, UT1-UTC= 0.2598254, Accumulated Leap Seconds=30
  { 2450222.50, 61.926061 }, //19 May 1996, UT1-UTC= 0.2579389, Accumulated Leap Seconds=30
  { 2450223.50, 61.927846 }, //20 May 1996, UT1-UTC= 0.2561541, Accumulated Leap Seconds=30
  { 2450224.50, 61.929555 }, //21 May 1996, UT1-UTC= 0.2544452, Accumulated Leap Seconds=30
  { 2450225.50, 61.931219 }, //22 May 1996, UT1-UTC= 0.2527805, Accumulated Leap Seconds=30
  { 2450226.50, 61.932893 }, //23 May 1996, UT1-UTC= 0.2511069, Accumulated Leap Seconds=30
  { 2450227.50, 61.934611 }, //24 May 1996, UT1-UTC= 0.2493890, Accumulated Leap Seconds=30
  { 2450228.50, 61.936389 }, //25 May 1996, UT1-UTC= 0.2476111, Accumulated Leap Seconds=30
  { 2450229.50, 61.938239 }, //26 May 1996, UT1-UTC= 0.2457611, Accumulated Leap Seconds=30
  { 2450230.50, 61.940154 }, //27 May 1996, UT1-UTC= 0.2438458, Accumulated Leap Seconds=30
  { 2450231.50, 61.942110 }, //28 May 1996, UT1-UTC= 0.2418896, Accumulated Leap Seconds=30
  { 2450232.50, 61.944075 }, //29 May 1996, UT1-UTC= 0.2399248, Accumulated Leap Seconds=30
  { 2450233.50, 61.946019 }, //30 May 1996, UT1-UTC= 0.2379809, Accumulated Leap Seconds=30
  { 2450234.50, 61.947895 }, //31 May 1996, UT1-UTC= 0.2361053, Accumulated Leap Seconds=30
  { 2450235.50, 61.949676 }, //01 June 1996, UT1-UTC= 0.2343238, Accumulated Leap Seconds=30
  { 2450236.50, 61.951369 }, //02 June 1996, UT1-UTC= 0.2326310, Accumulated Leap Seconds=30
  { 2450237.50, 61.953016 }, //03 June 1996, UT1-UTC= 0.2309839, Accumulated Leap Seconds=30
  { 2450238.50, 61.954681 }, //04 June 1996, UT1-UTC= 0.2293191, Accumulated Leap Seconds=30
  { 2450239.50, 61.956423 }, //05 June 1996, UT1-UTC= 0.2275770, Accumulated Leap Seconds=30
  { 2450240.50, 61.958256 }, //06 June 1996, UT1-UTC= 0.2257442, Accumulated Leap Seconds=30
  { 2450241.50, 61.960178 }, //07 June 1996, UT1-UTC= 0.2238221, Accumulated Leap Seconds=30
  { 2450242.50, 61.962176 }, //08 June 1996, UT1-UTC= 0.2218240, Accumulated Leap Seconds=30
  { 2450243.50, 61.964226 }, //09 June 1996, UT1-UTC= 0.2197735, Accumulated Leap Seconds=30
  { 2450244.50, 61.966285 }, //10 June 1996, UT1-UTC= 0.2177147, Accumulated Leap Seconds=30
  { 2450245.50, 61.968287 }, //11 June 1996, UT1-UTC= 0.2157127, Accumulated Leap Seconds=30
  { 2450246.50, 61.970178 }, //12 June 1996, UT1-UTC= 0.2138219, Accumulated Leap Seconds=30
  { 2450247.50, 61.971926 }, //13 June 1996, UT1-UTC= 0.2120739, Accumulated Leap Seconds=30
  { 2450248.50, 61.973520 }, //14 June 1996, UT1-UTC= 0.2104798, Accumulated Leap Seconds=30
  { 2450249.50, 61.974963 }, //15 June 1996, UT1-UTC= 0.2090372, Accumulated Leap Seconds=30
  { 2450250.50, 61.976268 }, //16 June 1996, UT1-UTC= 0.2077315, Accumulated Leap Seconds=30
  { 2450251.50, 61.977498 }, //17 June 1996, UT1-UTC= 0.2065017, Accumulated Leap Seconds=30
  { 2450252.50, 61.978697 }, //18 June 1996, UT1-UTC= 0.2053034, Accumulated Leap Seconds=30
  { 2450253.50, 61.979907 }, //19 June 1996, UT1-UTC= 0.2040934, Accumulated Leap Seconds=30
  { 2450254.50, 61.981174 }, //20 June 1996, UT1-UTC= 0.2028256, Accumulated Leap Seconds=30
  { 2450255.50, 61.982526 }, //21 June 1996, UT1-UTC= 0.2014737, Accumulated Leap Seconds=30
  { 2450256.50, 61.983966 }, //22 June 1996, UT1-UTC= 0.2000335, Accumulated Leap Seconds=30
  { 2450257.50, 61.985488 }, //23 June 1996, UT1-UTC= 0.1985121, Accumulated Leap Seconds=30
  { 2450258.50, 61.987060 }, //24 June 1996, UT1-UTC= 0.1969397, Accumulated Leap Seconds=30
  { 2450259.50, 61.988642 }, //25 June 1996, UT1-UTC= 0.1953581, Accumulated Leap Seconds=30
  { 2450260.50, 61.990193 }, //26 June 1996, UT1-UTC= 0.1938074, Accumulated Leap Seconds=30
  { 2450261.50, 61.991700 }, //27 June 1996, UT1-UTC= 0.1923004, Accumulated Leap Seconds=30
  { 2450262.50, 61.993131 }, //28 June 1996, UT1-UTC= 0.1908688, Accumulated Leap Seconds=30
  { 2450263.50, 61.994460 }, //29 June 1996, UT1-UTC= 0.1895402, Accumulated Leap Seconds=30
  { 2450264.50, 61.995695 }, //30 June 1996, UT1-UTC= 0.1883047, Accumulated Leap Seconds=30
  { 2450265.50, 61.996874 }, //01 July 1996, UT1-UTC= 0.1871257, Accumulated Leap Seconds=30
  { 2450266.50, 61.998055 }, //02 July 1996, UT1-UTC= 0.1859454, Accumulated Leap Seconds=30
  { 2450267.50, 61.999304 }, //03 July 1996, UT1-UTC= 0.1846963, Accumulated Leap Seconds=30
  { 2450268.50, 62.000672 }, //04 July 1996, UT1-UTC= 0.1833284, Accumulated Leap Seconds=30
  { 2450269.50, 62.002122 }, //05 July 1996, UT1-UTC= 0.1818781, Accumulated Leap Seconds=30
  { 2450270.50, 62.003583 }, //06 July 1996, UT1-UTC= 0.1804172, Accumulated Leap Seconds=30
  { 2450271.50, 62.004994 }, //07 July 1996, UT1-UTC= 0.1790058, Accumulated Leap Seconds=30
  { 2450272.50, 62.006317 }, //08 July 1996, UT1-UTC= 0.1776827, Accumulated Leap Seconds=30
  { 2450273.50, 62.007531 }, //09 July 1996, UT1-UTC= 0.1764691, Accumulated Leap Seconds=30
  { 2450274.50, 62.008631 }, //10 July 1996, UT1-UTC= 0.1753690, Accumulated Leap Seconds=30
  { 2450275.50, 62.009633 }, //11 July 1996, UT1-UTC= 0.1743671, Accumulated Leap Seconds=30
  { 2450276.50, 62.010545 }, //12 July 1996, UT1-UTC= 0.1734554, Accumulated Leap Seconds=30
  { 2450277.50, 62.011382 }, //13 July 1996, UT1-UTC= 0.1726175, Accumulated Leap Seconds=30
  { 2450278.50, 62.012193 }, //14 July 1996, UT1-UTC= 0.1718071, Accumulated Leap Seconds=30
  { 2450279.50, 62.013020 }, //15 July 1996, UT1-UTC= 0.1709798, Accumulated Leap Seconds=30
  { 2450280.50, 62.013893 }, //16 July 1996, UT1-UTC= 0.1701068, Accumulated Leap Seconds=30
  { 2450281.50, 62.014829 }, //17 July 1996, UT1-UTC= 0.1691706, Accumulated Leap Seconds=30
  { 2450282.50, 62.015844 }, //18 July 1996, UT1-UTC= 0.1681560, Accumulated Leap Seconds=30
  { 2450283.50, 62.016950 }, //19 July 1996, UT1-UTC= 0.1670503, Accumulated Leap Seconds=30
  { 2450284.50, 62.018133 }, //20 July 1996, UT1-UTC= 0.1658675, Accumulated Leap Seconds=30
  { 2450285.50, 62.019377 }, //21 July 1996, UT1-UTC= 0.1646226, Accumulated Leap Seconds=30
  { 2450286.50, 62.020668 }, //22 July 1996, UT1-UTC= 0.1633317, Accumulated Leap Seconds=30
  { 2450287.50, 62.021989 }, //23 July 1996, UT1-UTC= 0.1620112, Accumulated Leap Seconds=30
  { 2450288.50, 62.023324 }, //24 July 1996, UT1-UTC= 0.1606759, Accumulated Leap Seconds=30
  { 2450289.50, 62.024662 }, //25 July 1996, UT1-UTC= 0.1593380, Accumulated Leap Seconds=30
  { 2450290.50, 62.025966 }, //26 July 1996, UT1-UTC= 0.1580340, Accumulated Leap Seconds=30
  { 2450291.50, 62.027210 }, //27 July 1996, UT1-UTC= 0.1567900, Accumulated Leap Seconds=30
  { 2450292.50, 62.028430 }, //28 July 1996, UT1-UTC= 0.1555698, Accumulated Leap Seconds=30
  { 2450293.50, 62.029695 }, //29 July 1996, UT1-UTC= 0.1543052, Accumulated Leap Seconds=30
  { 2450294.50, 62.031071 }, //30 July 1996, UT1-UTC= 0.1529285, Accumulated Leap Seconds=30
  { 2450295.50, 62.032606 }, //31 July 1996, UT1-UTC= 0.1513944, Accumulated Leap Seconds=30
  { 2450296.50, 62.034294 }, //01 August 1996, UT1-UTC= 0.1497062, Accumulated Leap Seconds=30
  { 2450297.50, 62.036058 }, //02 August 1996, UT1-UTC= 0.1479419, Accumulated Leap Seconds=30
  { 2450298.50, 62.037788 }, //03 August 1996, UT1-UTC= 0.1462116, Accumulated Leap Seconds=30
  { 2450299.50, 62.039382 }, //04 August 1996, UT1-UTC= 0.1446181, Accumulated Leap Seconds=30
  { 2450300.50, 62.040784 }, //05 August 1996, UT1-UTC= 0.1432162, Accumulated Leap Seconds=30
  { 2450301.50, 62.041990 }, //06 August 1996, UT1-UTC= 0.1420101, Accumulated Leap Seconds=30
  { 2450302.50, 62.043014 }, //07 August 1996, UT1-UTC= 0.1409863, Accumulated Leap Seconds=30
  { 2450303.50, 62.043875 }, //08 August 1996, UT1-UTC= 0.1401254, Accumulated Leap Seconds=30
  { 2450304.50, 62.044620 }, //09 August 1996, UT1-UTC= 0.1393797, Accumulated Leap Seconds=30
  { 2450305.50, 62.045308 }, //10 August 1996, UT1-UTC= 0.1386916, Accumulated Leap Seconds=30
  { 2450306.50, 62.045973 }, //11 August 1996, UT1-UTC= 0.1380267, Accumulated Leap Seconds=30
  { 2450307.50, 62.046665 }, //12 August 1996, UT1-UTC= 0.1373348, Accumulated Leap Seconds=30
  { 2450308.50, 62.047416 }, //13 August 1996, UT1-UTC= 0.1365844, Accumulated Leap Seconds=30
  { 2450309.50, 62.048239 }, //14 August 1996, UT1-UTC= 0.1357608, Accumulated Leap Seconds=30
  { 2450310.50, 62.049147 }, //15 August 1996, UT1-UTC= 0.1348527, Accumulated Leap Seconds=30
  { 2450311.50, 62.050147 }, //16 August 1996, UT1-UTC= 0.1338526, Accumulated Leap Seconds=30
  { 2450312.50, 62.051236 }, //17 August 1996, UT1-UTC= 0.1327641, Accumulated Leap Seconds=30
  { 2450313.50, 62.052387 }, //18 August 1996, UT1-UTC= 0.1316126, Accumulated Leap Seconds=30
  { 2450314.50, 62.053561 }, //19 August 1996, UT1-UTC= 0.1304387, Accumulated Leap Seconds=30
  { 2450315.50, 62.054730 }, //20 August 1996, UT1-UTC= 0.1292702, Accumulated Leap Seconds=30
  { 2450316.50, 62.055876 }, //21 August 1996, UT1-UTC= 0.1281239, Accumulated Leap Seconds=30
  { 2450317.50, 62.057001 }, //22 August 1996, UT1-UTC= 0.1269986, Accumulated Leap Seconds=30
  { 2450318.50, 62.058117 }, //23 August 1996, UT1-UTC= 0.1258829, Accumulated Leap Seconds=30
  { 2450319.50, 62.059227 }, //24 August 1996, UT1-UTC= 0.1247726, Accumulated Leap Seconds=30
  { 2450320.50, 62.060361 }, //25 August 1996, UT1-UTC= 0.1236394, Accumulated Leap Seconds=30
  { 2450321.50, 62.061574 }, //26 August 1996, UT1-UTC= 0.1224258, Accumulated Leap Seconds=30
  { 2450322.50, 62.062918 }, //27 August 1996, UT1-UTC= 0.1210817, Accumulated Leap Seconds=30
  { 2450323.50, 62.064421 }, //28 August 1996, UT1-UTC= 0.1195793, Accumulated Leap Seconds=30
  { 2450324.50, 62.066106 }, //29 August 1996, UT1-UTC= 0.1178937, Accumulated Leap Seconds=30
  { 2450325.50, 62.067902 }, //30 August 1996, UT1-UTC= 0.1160979, Accumulated Leap Seconds=30
  { 2450326.50, 62.069703 }, //31 August 1996, UT1-UTC= 0.1142972, Accumulated Leap Seconds=30
  { 2450327.50, 62.071411 }, //01 September 1996, UT1-UTC= 0.1125892, Accumulated Leap Seconds=30
  { 2450328.50, 62.072956 }, //02 September 1996, UT1-UTC= 0.1110439, Accumulated Leap Seconds=30
  { 2450329.50, 62.074312 }, //03 September 1996, UT1-UTC= 0.1096882, Accumulated Leap Seconds=30
  { 2450330.50, 62.075486 }, //04 September 1996, UT1-UTC= 0.1085140, Accumulated Leap Seconds=30
  { 2450331.50, 62.076530 }, //05 September 1996, UT1-UTC= 0.1074697, Accumulated Leap Seconds=30
  { 2450332.50, 62.077490 }, //06 September 1996, UT1-UTC= 0.1065102, Accumulated Leap Seconds=30
  { 2450333.50, 62.078419 }, //07 September 1996, UT1-UTC= 0.1055810, Accumulated Leap Seconds=30
  { 2450334.50, 62.079356 }, //08 September 1996, UT1-UTC= 0.1046443, Accumulated Leap Seconds=30
  { 2450335.50, 62.080354 }, //09 September 1996, UT1-UTC= 0.1036456, Accumulated Leap Seconds=30
  { 2450336.50, 62.081472 }, //10 September 1996, UT1-UTC= 0.1025275, Accumulated Leap Seconds=30
  { 2450337.50, 62.082766 }, //11 September 1996, UT1-UTC= 0.1012340, Accumulated Leap Seconds=30
  { 2450338.50, 62.084273 }, //12 September 1996, UT1-UTC= 0.0997267, Accumulated Leap Seconds=30
  { 2450339.50, 62.085951 }, //13 September 1996, UT1-UTC= 0.0980494, Accumulated Leap Seconds=30
  { 2450340.50, 62.087725 }, //14 September 1996, UT1-UTC= 0.0962748, Accumulated Leap Seconds=30
  { 2450341.50, 62.089539 }, //15 September 1996, UT1-UTC= 0.0944614, Accumulated Leap Seconds=30
  { 2450342.50, 62.091345 }, //16 September 1996, UT1-UTC= 0.0926553, Accumulated Leap Seconds=30
  { 2450343.50, 62.093096 }, //17 September 1996, UT1-UTC= 0.0909044, Accumulated Leap Seconds=30
  { 2450344.50, 62.094759 }, //18 September 1996, UT1-UTC= 0.0892408, Accumulated Leap Seconds=30
  { 2450345.50, 62.096354 }, //19 September 1996, UT1-UTC= 0.0876456, Accumulated Leap Seconds=30
  { 2450346.50, 62.097923 }, //20 September 1996, UT1-UTC= 0.0860766, Accumulated Leap Seconds=30
  { 2450347.50, 62.099522 }, //21 September 1996, UT1-UTC= 0.0844781, Accumulated Leap Seconds=30
  { 2450348.50, 62.101194 }, //22 September 1996, UT1-UTC= 0.0828060, Accumulated Leap Seconds=30
  { 2450349.50, 62.103010 }, //23 September 1996, UT1-UTC= 0.0809897, Accumulated Leap Seconds=30
  { 2450350.50, 62.105018 }, //24 September 1996, UT1-UTC= 0.0789821, Accumulated Leap Seconds=30
  { 2450351.50, 62.107218 }, //25 September 1996, UT1-UTC= 0.0767816, Accumulated Leap Seconds=30
  { 2450352.50, 62.109553 }, //26 September 1996, UT1-UTC= 0.0744470, Accumulated Leap Seconds=30
  { 2450353.50, 62.111914 }, //27 September 1996, UT1-UTC= 0.0720855, Accumulated Leap Seconds=30
  { 2450354.50, 62.114202 }, //28 September 1996, UT1-UTC= 0.0697982, Accumulated Leap Seconds=30
  { 2450355.50, 62.116352 }, //29 September 1996, UT1-UTC= 0.0676484, Accumulated Leap Seconds=30
  { 2450356.50, 62.118353 }, //30 September 1996, UT1-UTC= 0.0656471, Accumulated Leap Seconds=30
  { 2450357.50, 62.120231 }, //01 October 1996, UT1-UTC= 0.0637685, Accumulated Leap Seconds=30
  { 2450358.50, 62.122015 }, //02 October 1996, UT1-UTC= 0.0619853, Accumulated Leap Seconds=30
  { 2450359.50, 62.123672 }, //03 October 1996, UT1-UTC= 0.0603279, Accumulated Leap Seconds=30
  { 2450360.50, 62.125247 }, //04 October 1996, UT1-UTC= 0.0587533, Accumulated Leap Seconds=30
  { 2450361.50, 62.126800 }, //05 October 1996, UT1-UTC= 0.0571995, Accumulated Leap Seconds=30
  { 2450362.50, 62.128391 }, //06 October 1996, UT1-UTC= 0.0556093, Accumulated Leap Seconds=30
  { 2450363.50, 62.130043 }, //07 October 1996, UT1-UTC= 0.0539571, Accumulated Leap Seconds=30
  { 2450364.50, 62.131779 }, //08 October 1996, UT1-UTC= 0.0522207, Accumulated Leap Seconds=30
  { 2450365.50, 62.133615 }, //09 October 1996, UT1-UTC= 0.0503846, Accumulated Leap Seconds=30
  { 2450366.50, 62.135559 }, //10 October 1996, UT1-UTC= 0.0484412, Accumulated Leap Seconds=30
  { 2450367.50, 62.137591 }, //11 October 1996, UT1-UTC= 0.0464085, Accumulated Leap Seconds=30
  { 2450368.50, 62.139681 }, //12 October 1996, UT1-UTC= 0.0443186, Accumulated Leap Seconds=30
  { 2450369.50, 62.141797 }, //13 October 1996, UT1-UTC= 0.0422025, Accumulated Leap Seconds=30
  { 2450370.50, 62.143910 }, //14 October 1996, UT1-UTC= 0.0400901, Accumulated Leap Seconds=30
  { 2450371.50, 62.145984 }, //15 October 1996, UT1-UTC= 0.0380158, Accumulated Leap Seconds=30
  { 2450372.50, 62.147994 }, //16 October 1996, UT1-UTC= 0.0360063, Accumulated Leap Seconds=30
  { 2450373.50, 62.149935 }, //17 October 1996, UT1-UTC= 0.0340650, Accumulated Leap Seconds=30
  { 2450374.50, 62.151825 }, //18 October 1996, UT1-UTC= 0.0321752, Accumulated Leap Seconds=30
  { 2450375.50, 62.153731 }, //19 October 1996, UT1-UTC= 0.0302686, Accumulated Leap Seconds=30
  { 2450376.50, 62.155745 }, //20 October 1996, UT1-UTC= 0.0282546, Accumulated Leap Seconds=30
  { 2450377.50, 62.157889 }, //21 October 1996, UT1-UTC= 0.0261109, Accumulated Leap Seconds=30
  { 2450378.50, 62.160147 }, //22 October 1996, UT1-UTC= 0.0238534, Accumulated Leap Seconds=30
  { 2450379.50, 62.162501 }, //23 October 1996, UT1-UTC= 0.0214994, Accumulated Leap Seconds=30
  { 2450380.50, 62.164943 }, //24 October 1996, UT1-UTC= 0.0190565, Accumulated Leap Seconds=30
  { 2450381.50, 62.167396 }, //25 October 1996, UT1-UTC= 0.0166042, Accumulated Leap Seconds=30
  { 2450382.50, 62.169762 }, //26 October 1996, UT1-UTC= 0.0142378, Accumulated Leap Seconds=30
  { 2450383.50, 62.171983 }, //27 October 1996, UT1-UTC= 0.0120173, Accumulated Leap Seconds=30
  { 2450384.50, 62.174033 }, //28 October 1996, UT1-UTC= 0.0099672, Accumulated Leap Seconds=30
  { 2450385.50, 62.175923 }, //29 October 1996, UT1-UTC= 0.0080767, Accumulated Leap Seconds=30
  { 2450386.50, 62.177677 }, //30 October 1996, UT1-UTC= 0.0063228, Accumulated Leap Seconds=30
  { 2450387.50, 62.179332 }, //31 October 1996, UT1-UTC= 0.0046680, Accumulated Leap Seconds=30
  { 2450388.50, 62.180951 }, //01 November 1996, UT1-UTC= 0.0030492, Accumulated Leap Seconds=30
  { 2450389.50, 62.182581 }, //02 November 1996, UT1-UTC= 0.0014189, Accumulated Leap Seconds=30
  { 2450390.50, 62.184275 }, //03 November 1996, UT1-UTC=-0.0002751, Accumulated Leap Seconds=30
  { 2450391.50, 62.186076 }, //04 November 1996, UT1-UTC=-0.0020760, Accumulated Leap Seconds=30
  { 2450392.50, 62.188006 }, //05 November 1996, UT1-UTC=-0.0040059, Accumulated Leap Seconds=30
  { 2450393.50, 62.190066 }, //06 November 1996, UT1-UTC=-0.0060663, Accumulated Leap Seconds=30
  { 2450394.50, 62.192203 }, //07 November 1996, UT1-UTC=-0.0082034, Accumulated Leap Seconds=30
  { 2450395.50, 62.194382 }, //08 November 1996, UT1-UTC=-0.0103823, Accumulated Leap Seconds=30
  { 2450396.50, 62.196566 }, //09 November 1996, UT1-UTC=-0.0125659, Accumulated Leap Seconds=30
  { 2450397.50, 62.198701 }, //10 November 1996, UT1-UTC=-0.0147006, Accumulated Leap Seconds=30
  { 2450398.50, 62.200780 }, //11 November 1996, UT1-UTC=-0.0167801, Accumulated Leap Seconds=30
  { 2450399.50, 62.202821 }, //12 November 1996, UT1-UTC=-0.0188207, Accumulated Leap Seconds=30
  { 2450400.50, 62.204845 }, //13 November 1996, UT1-UTC=-0.0208447, Accumulated Leap Seconds=30
  { 2450401.50, 62.206817 }, //14 November 1996, UT1-UTC=-0.0228168, Accumulated Leap Seconds=30
  { 2450402.50, 62.208698 }, //15 November 1996, UT1-UTC=-0.0246977, Accumulated Leap Seconds=30
  { 2450403.50, 62.210539 }, //16 November 1996, UT1-UTC=-0.0265391, Accumulated Leap Seconds=30
  { 2450404.50, 62.212455 }, //17 November 1996, UT1-UTC=-0.0284548, Accumulated Leap Seconds=30
  { 2450405.50, 62.214476 }, //18 November 1996, UT1-UTC=-0.0304757, Accumulated Leap Seconds=30
  { 2450406.50, 62.216584 }, //19 November 1996, UT1-UTC=-0.0325842, Accumulated Leap Seconds=30
  { 2450407.50, 62.218742 }, //20 November 1996, UT1-UTC=-0.0347420, Accumulated Leap Seconds=30
  { 2450408.50, 62.220903 }, //21 November 1996, UT1-UTC=-0.0369032, Accumulated Leap Seconds=30
  { 2450409.50, 62.223002 }, //22 November 1996, UT1-UTC=-0.0390021, Accumulated Leap Seconds=30
  { 2450410.50, 62.224982 }, //23 November 1996, UT1-UTC=-0.0409816, Accumulated Leap Seconds=30
  { 2450411.50, 62.226827 }, //24 November 1996, UT1-UTC=-0.0428270, Accumulated Leap Seconds=30
  { 2450412.50, 62.228559 }, //25 November 1996, UT1-UTC=-0.0445593, Accumulated Leap Seconds=30
  { 2450413.50, 62.230208 }, //26 November 1996, UT1-UTC=-0.0462081, Accumulated Leap Seconds=30
  { 2450414.50, 62.231811 }, //27 November 1996, UT1-UTC=-0.0478106, Accumulated Leap Seconds=30
  { 2450415.50, 62.233386 }, //28 November 1996, UT1-UTC=-0.0493859, Accumulated Leap Seconds=30
  { 2450416.50, 62.234960 }, //29 November 1996, UT1-UTC=-0.0509597, Accumulated Leap Seconds=30
  { 2450417.50, 62.236559 }, //30 November 1996, UT1-UTC=-0.0525595, Accumulated Leap Seconds=30
  { 2450418.50, 62.238205 }, //01 December 1996, UT1-UTC=-0.0542046, Accumulated Leap Seconds=30
  { 2450419.50, 62.239894 }, //02 December 1996, UT1-UTC=-0.0558944, Accumulated Leap Seconds=30
  { 2450420.50, 62.241613 }, //03 December 1996, UT1-UTC=-0.0576134, Accumulated Leap Seconds=30
  { 2450421.50, 62.243355 }, //04 December 1996, UT1-UTC=-0.0593555, Accumulated Leap Seconds=30
  { 2450422.50, 62.245138 }, //05 December 1996, UT1-UTC=-0.0611377, Accumulated Leap Seconds=30
  { 2450423.50, 62.246946 }, //06 December 1996, UT1-UTC=-0.0629464, Accumulated Leap Seconds=30
  { 2450424.50, 62.248739 }, //07 December 1996, UT1-UTC=-0.0647395, Accumulated Leap Seconds=30
  { 2450425.50, 62.250487 }, //08 December 1996, UT1-UTC=-0.0664871, Accumulated Leap Seconds=30
  { 2450426.50, 62.252178 }, //09 December 1996, UT1-UTC=-0.0681783, Accumulated Leap Seconds=30
  { 2450427.50, 62.253825 }, //10 December 1996, UT1-UTC=-0.0698246, Accumulated Leap Seconds=30
  { 2450428.50, 62.255457 }, //11 December 1996, UT1-UTC=-0.0714568, Accumulated Leap Seconds=30
  { 2450429.50, 62.257118 }, //12 December 1996, UT1-UTC=-0.0731184, Accumulated Leap Seconds=30
  { 2450430.50, 62.258848 }, //13 December 1996, UT1-UTC=-0.0748482, Accumulated Leap Seconds=30
  { 2450431.50, 62.260688 }, //14 December 1996, UT1-UTC=-0.0766884, Accumulated Leap Seconds=30
  { 2450432.50, 62.262657 }, //15 December 1996, UT1-UTC=-0.0786567, Accumulated Leap Seconds=30
  { 2450433.50, 62.264731 }, //16 December 1996, UT1-UTC=-0.0807312, Accumulated Leap Seconds=30
  { 2450434.50, 62.266872 }, //17 December 1996, UT1-UTC=-0.0828724, Accumulated Leap Seconds=30
  { 2450435.50, 62.269040 }, //18 December 1996, UT1-UTC=-0.0850402, Accumulated Leap Seconds=30
  { 2450436.50, 62.271198 }, //19 December 1996, UT1-UTC=-0.0871980, Accumulated Leap Seconds=30
  { 2450437.50, 62.273283 }, //20 December 1996, UT1-UTC=-0.0892835, Accumulated Leap Seconds=30
  { 2450438.50, 62.275251 }, //21 December 1996, UT1-UTC=-0.0912506, Accumulated Leap Seconds=30
  { 2450439.50, 62.277105 }, //22 December 1996, UT1-UTC=-0.0931049, Accumulated Leap Seconds=30
  { 2450440.50, 62.278882 }, //23 December 1996, UT1-UTC=-0.0948824, Accumulated Leap Seconds=30
  { 2450441.50, 62.280605 }, //24 December 1996, UT1-UTC=-0.0966050, Accumulated Leap Seconds=30
  { 2450442.50, 62.282288 }, //25 December 1996, UT1-UTC=-0.0982878, Accumulated Leap Seconds=30
  { 2450443.50, 62.283962 }, //26 December 1996, UT1-UTC=-0.0999625, Accumulated Leap Seconds=30
  { 2450444.50, 62.285663 }, //27 December 1996, UT1-UTC=-0.1016632, Accumulated Leap Seconds=30
  { 2450445.50, 62.287416 }, //28 December 1996, UT1-UTC=-0.1034160, Accumulated Leap Seconds=30
  { 2450446.50, 62.289234 }, //29 December 1996, UT1-UTC=-0.1052344, Accumulated Leap Seconds=30
  { 2450447.50, 62.291120 }, //30 December 1996, UT1-UTC=-0.1071200, Accumulated Leap Seconds=30
  { 2450448.50, 62.293063 }, //31 December 1996, UT1-UTC=-0.1090635, Accumulated Leap Seconds=30
  { 2450449.50, 62.295049 }, //01 January 1997, UT1-UTC=-0.1110486, Accumulated Leap Seconds=30
  { 2450450.50, 62.297053 }, //02 January 1997, UT1-UTC=-0.1130531, Accumulated Leap Seconds=30
  { 2450451.50, 62.299055 }, //03 January 1997, UT1-UTC=-0.1150555, Accumulated Leap Seconds=30
  { 2450452.50, 62.301029 }, //04 January 1997, UT1-UTC=-0.1170293, Accumulated Leap Seconds=30
  { 2450453.50, 62.302955 }, //05 January 1997, UT1-UTC=-0.1189555, Accumulated Leap Seconds=30
  { 2450454.50, 62.304832 }, //06 January 1997, UT1-UTC=-0.1208321, Accumulated Leap Seconds=30
  { 2450455.50, 62.306672 }, //07 January 1997, UT1-UTC=-0.1226723, Accumulated Leap Seconds=30
  { 2450456.50, 62.308512 }, //08 January 1997, UT1-UTC=-0.1245120, Accumulated Leap Seconds=30
  { 2450457.50, 62.310397 }, //09 January 1997, UT1-UTC=-0.1263974, Accumulated Leap Seconds=30
  { 2450458.50, 62.312363 }, //10 January 1997, UT1-UTC=-0.1283633, Accumulated Leap Seconds=30
  { 2450459.50, 62.314434 }, //11 January 1997, UT1-UTC=-0.1304342, Accumulated Leap Seconds=30
  { 2450460.50, 62.316595 }, //12 January 1997, UT1-UTC=-0.1325949, Accumulated Leap Seconds=30
  { 2450461.50, 62.318811 }, //13 January 1997, UT1-UTC=-0.1348112, Accumulated Leap Seconds=30
  { 2450462.50, 62.321047 }, //14 January 1997, UT1-UTC=-0.1370468, Accumulated Leap Seconds=30
  { 2450463.50, 62.323249 }, //15 January 1997, UT1-UTC=-0.1392488, Accumulated Leap Seconds=30
  { 2450464.50, 62.325363 }, //16 January 1997, UT1-UTC=-0.1413634, Accumulated Leap Seconds=30
  { 2450465.50, 62.327344 }, //17 January 1997, UT1-UTC=-0.1433445, Accumulated Leap Seconds=30
  { 2450466.50, 62.329164 }, //18 January 1997, UT1-UTC=-0.1451642, Accumulated Leap Seconds=30
  { 2450467.50, 62.330822 }, //19 January 1997, UT1-UTC=-0.1468222, Accumulated Leap Seconds=30
  { 2450468.50, 62.332351 }, //20 January 1997, UT1-UTC=-0.1483507, Accumulated Leap Seconds=30
  { 2450469.50, 62.333794 }, //21 January 1997, UT1-UTC=-0.1497942, Accumulated Leap Seconds=30
  { 2450470.50, 62.335207 }, //22 January 1997, UT1-UTC=-0.1512068, Accumulated Leap Seconds=30
  { 2450471.50, 62.336650 }, //23 January 1997, UT1-UTC=-0.1526505, Accumulated Leap Seconds=30
  { 2450472.50, 62.338129 }, //24 January 1997, UT1-UTC=-0.1541289, Accumulated Leap Seconds=30
  { 2450473.50, 62.339619 }, //25 January 1997, UT1-UTC=-0.1556195, Accumulated Leap Seconds=30
  { 2450474.50, 62.341145 }, //26 January 1997, UT1-UTC=-0.1571450, Accumulated Leap Seconds=30
  { 2450475.50, 62.342726 }, //27 January 1997, UT1-UTC=-0.1587262, Accumulated Leap Seconds=30
  { 2450476.50, 62.344365 }, //28 January 1997, UT1-UTC=-0.1603647, Accumulated Leap Seconds=30
  { 2450477.50, 62.346024 }, //29 January 1997, UT1-UTC=-0.1620237, Accumulated Leap Seconds=30
  { 2450478.50, 62.347627 }, //30 January 1997, UT1-UTC=-0.1636273, Accumulated Leap Seconds=30
  { 2450479.50, 62.349165 }, //31 January 1997, UT1-UTC=-0.1651655, Accumulated Leap Seconds=30
  { 2450480.50, 62.350648 }, //01 February 1997, UT1-UTC=-0.1666479, Accumulated Leap Seconds=30
  { 2450481.50, 62.352034 }, //02 February 1997, UT1-UTC=-0.1680344, Accumulated Leap Seconds=30
  { 2450482.50, 62.353331 }, //03 February 1997, UT1-UTC=-0.1693311, Accumulated Leap Seconds=30
  { 2450483.50, 62.354593 }, //04 February 1997, UT1-UTC=-0.1705927, Accumulated Leap Seconds=30
  { 2450484.50, 62.355897 }, //05 February 1997, UT1-UTC=-0.1718975, Accumulated Leap Seconds=30
  { 2450485.50, 62.357340 }, //06 February 1997, UT1-UTC=-0.1733399, Accumulated Leap Seconds=30
  { 2450486.50, 62.358965 }, //07 February 1997, UT1-UTC=-0.1749655, Accumulated Leap Seconds=30
  { 2450487.50, 62.360784 }, //08 February 1997, UT1-UTC=-0.1767845, Accumulated Leap Seconds=30
  { 2450488.50, 62.362750 }, //09 February 1997, UT1-UTC=-0.1787497, Accumulated Leap Seconds=30
  { 2450489.50, 62.364788 }, //10 February 1997, UT1-UTC=-0.1807885, Accumulated Leap Seconds=30
  { 2450490.50, 62.366861 }, //11 February 1997, UT1-UTC=-0.1828615, Accumulated Leap Seconds=30
  { 2450491.50, 62.368925 }, //12 February 1997, UT1-UTC=-0.1849246, Accumulated Leap Seconds=30
  { 2450492.50, 62.370870 }, //13 February 1997, UT1-UTC=-0.1868698, Accumulated Leap Seconds=30
  { 2450493.50, 62.372667 }, //14 February 1997, UT1-UTC=-0.1886672, Accumulated Leap Seconds=30
  { 2450494.50, 62.374350 }, //15 February 1997, UT1-UTC=-0.1903496, Accumulated Leap Seconds=30
  { 2450495.50, 62.375942 }, //16 February 1997, UT1-UTC=-0.1919421, Accumulated Leap Seconds=30
  { 2450496.50, 62.377472 }, //17 February 1997, UT1-UTC=-0.1934717, Accumulated Leap Seconds=30
  { 2450497.50, 62.378994 }, //18 February 1997, UT1-UTC=-0.1949944, Accumulated Leap Seconds=30
  { 2450498.50, 62.380550 }, //19 February 1997, UT1-UTC=-0.1965501, Accumulated Leap Seconds=30
  { 2450499.50, 62.382159 }, //20 February 1997, UT1-UTC=-0.1981592, Accumulated Leap Seconds=30
  { 2450500.50, 62.383847 }, //21 February 1997, UT1-UTC=-0.1998471, Accumulated Leap Seconds=30
  { 2450501.50, 62.385624 }, //22 February 1997, UT1-UTC=-0.2016242, Accumulated Leap Seconds=30
  { 2450502.50, 62.387499 }, //23 February 1997, UT1-UTC=-0.2034987, Accumulated Leap Seconds=30
  { 2450503.50, 62.389456 }, //24 February 1997, UT1-UTC=-0.2054556, Accumulated Leap Seconds=30
  { 2450504.50, 62.391475 }, //25 February 1997, UT1-UTC=-0.2074754, Accumulated Leap Seconds=30
  { 2450505.50, 62.393507 }, //26 February 1997, UT1-UTC=-0.2095069, Accumulated Leap Seconds=30
  { 2450506.50, 62.395520 }, //27 February 1997, UT1-UTC=-0.2115197, Accumulated Leap Seconds=30
  { 2450507.50, 62.397527 }, //28 February 1997, UT1-UTC=-0.2135268, Accumulated Leap Seconds=30
  { 2450508.50, 62.399538 }, //01 March 1997, UT1-UTC=-0.2155381, Accumulated Leap Seconds=30
  { 2450509.50, 62.401555 }, //02 March 1997, UT1-UTC=-0.2175554, Accumulated Leap Seconds=30
  { 2450510.50, 62.403581 }, //03 March 1997, UT1-UTC=-0.2195808, Accumulated Leap Seconds=30
  { 2450511.50, 62.405631 }, //04 March 1997, UT1-UTC=-0.2216314, Accumulated Leap Seconds=30
  { 2450512.50, 62.407748 }, //05 March 1997, UT1-UTC=-0.2237477, Accumulated Leap Seconds=30
  { 2450513.50, 62.409986 }, //06 March 1997, UT1-UTC=-0.2259856, Accumulated Leap Seconds=30
  { 2450514.50, 62.412409 }, //07 March 1997, UT1-UTC=-0.2284086, Accumulated Leap Seconds=30
  { 2450515.50, 62.415037 }, //08 March 1997, UT1-UTC=-0.2310370, Accumulated Leap Seconds=30
  { 2450516.50, 62.417831 }, //09 March 1997, UT1-UTC=-0.2338305, Accumulated Leap Seconds=30
  { 2450517.50, 62.420696 }, //10 March 1997, UT1-UTC=-0.2366965, Accumulated Leap Seconds=30
  { 2450518.50, 62.423540 }, //11 March 1997, UT1-UTC=-0.2395397, Accumulated Leap Seconds=30
  { 2450519.50, 62.426283 }, //12 March 1997, UT1-UTC=-0.2422826, Accumulated Leap Seconds=30
  { 2450520.50, 62.428880 }, //13 March 1997, UT1-UTC=-0.2448799, Accumulated Leap Seconds=30
  { 2450521.50, 62.431343 }, //14 March 1997, UT1-UTC=-0.2473432, Accumulated Leap Seconds=30
  { 2450522.50, 62.433717 }, //15 March 1997, UT1-UTC=-0.2497171, Accumulated Leap Seconds=30
  { 2450523.50, 62.436052 }, //16 March 1997, UT1-UTC=-0.2520524, Accumulated Leap Seconds=30
  { 2450524.50, 62.438387 }, //17 March 1997, UT1-UTC=-0.2543874, Accumulated Leap Seconds=30
  { 2450525.50, 62.440743 }, //18 March 1997, UT1-UTC=-0.2567432, Accumulated Leap Seconds=30
  { 2450526.50, 62.443141 }, //19 March 1997, UT1-UTC=-0.2591406, Accumulated Leap Seconds=30
  { 2450527.50, 62.445592 }, //20 March 1997, UT1-UTC=-0.2615925, Accumulated Leap Seconds=30
  { 2450528.50, 62.448062 }, //21 March 1997, UT1-UTC=-0.2640623, Accumulated Leap Seconds=30
  { 2450529.50, 62.450560 }, //22 March 1997, UT1-UTC=-0.2665600, Accumulated Leap Seconds=30
  { 2450530.50, 62.453100 }, //23 March 1997, UT1-UTC=-0.2691000, Accumulated Leap Seconds=30
  { 2450531.50, 62.455667 }, //24 March 1997, UT1-UTC=-0.2716668, Accumulated Leap Seconds=30
  { 2450532.50, 62.458240 }, //25 March 1997, UT1-UTC=-0.2742399, Accumulated Leap Seconds=30
  { 2450533.50, 62.460794 }, //26 March 1997, UT1-UTC=-0.2767937, Accumulated Leap Seconds=30
  { 2450534.50, 62.463303 }, //27 March 1997, UT1-UTC=-0.2793032, Accumulated Leap Seconds=30
  { 2450535.50, 62.465759 }, //28 March 1997, UT1-UTC=-0.2817595, Accumulated Leap Seconds=30
  { 2450536.50, 62.468176 }, //29 March 1997, UT1-UTC=-0.2841759, Accumulated Leap Seconds=30
  { 2450537.50, 62.470568 }, //30 March 1997, UT1-UTC=-0.2865684, Accumulated Leap Seconds=30
  { 2450538.50, 62.472962 }, //31 March 1997, UT1-UTC=-0.2889622, Accumulated Leap Seconds=30
  { 2450539.50, 62.475395 }, //01 April 1997, UT1-UTC=-0.2913950, Accumulated Leap Seconds=30
  { 2450540.50, 62.477903 }, //02 April 1997, UT1-UTC=-0.2939026, Accumulated Leap Seconds=30
  { 2450541.50, 62.480517 }, //03 April 1997, UT1-UTC=-0.2965173, Accumulated Leap Seconds=30
  { 2450542.50, 62.483267 }, //04 April 1997, UT1-UTC=-0.2992667, Accumulated Leap Seconds=30
  { 2450543.50, 62.486150 }, //05 April 1997, UT1-UTC=-0.3021503, Accumulated Leap Seconds=30
  { 2450544.50, 62.489115 }, //06 April 1997, UT1-UTC=-0.3051146, Accumulated Leap Seconds=30
  { 2450545.50, 62.492072 }, //07 April 1997, UT1-UTC=-0.3080719, Accumulated Leap Seconds=30
  { 2450546.50, 62.494934 }, //08 April 1997, UT1-UTC=-0.3109337, Accumulated Leap Seconds=30
  { 2450547.50, 62.497640 }, //09 April 1997, UT1-UTC=-0.3136397, Accumulated Leap Seconds=30
  { 2450548.50, 62.500189 }, //10 April 1997, UT1-UTC=-0.3161888, Accumulated Leap Seconds=30
  { 2450549.50, 62.502595 }, //11 April 1997, UT1-UTC=-0.3185955, Accumulated Leap Seconds=30
  { 2450550.50, 62.504892 }, //12 April 1997, UT1-UTC=-0.3208919, Accumulated Leap Seconds=30
  { 2450551.50, 62.507114 }, //13 April 1997, UT1-UTC=-0.3231144, Accumulated Leap Seconds=30
  { 2450552.50, 62.509302 }, //14 April 1997, UT1-UTC=-0.3253019, Accumulated Leap Seconds=30
  { 2450553.50, 62.511485 }, //15 April 1997, UT1-UTC=-0.3274855, Accumulated Leap Seconds=30
  { 2450554.50, 62.513693 }, //16 April 1997, UT1-UTC=-0.3296933, Accumulated Leap Seconds=30
  { 2450555.50, 62.515967 }, //17 April 1997, UT1-UTC=-0.3319670, Accumulated Leap Seconds=30
  { 2450556.50, 62.518286 }, //18 April 1997, UT1-UTC=-0.3342865, Accumulated Leap Seconds=30
  { 2450557.50, 62.520625 }, //19 April 1997, UT1-UTC=-0.3366253, Accumulated Leap Seconds=30
  { 2450558.50, 62.522985 }, //20 April 1997, UT1-UTC=-0.3389850, Accumulated Leap Seconds=30
  { 2450559.50, 62.525348 }, //21 April 1997, UT1-UTC=-0.3413476, Accumulated Leap Seconds=30
  { 2450560.50, 62.527669 }, //22 April 1997, UT1-UTC=-0.3436692, Accumulated Leap Seconds=30
  { 2450561.50, 62.529916 }, //23 April 1997, UT1-UTC=-0.3459163, Accumulated Leap Seconds=30
  { 2450562.50, 62.532082 }, //24 April 1997, UT1-UTC=-0.3480818, Accumulated Leap Seconds=30
  { 2450563.50, 62.534170 }, //25 April 1997, UT1-UTC=-0.3501698, Accumulated Leap Seconds=30
  { 2450564.50, 62.536190 }, //26 April 1997, UT1-UTC=-0.3521904, Accumulated Leap Seconds=30
  { 2450565.50, 62.538157 }, //27 April 1997, UT1-UTC=-0.3541570, Accumulated Leap Seconds=30
  { 2450566.50, 62.540102 }, //28 April 1997, UT1-UTC=-0.3561023, Accumulated Leap Seconds=30
  { 2450567.50, 62.542078 }, //29 April 1997, UT1-UTC=-0.3580781, Accumulated Leap Seconds=30
  { 2450568.50, 62.544135 }, //30 April 1997, UT1-UTC=-0.3601355, Accumulated Leap Seconds=30
  { 2450569.50, 62.546309 }, //01 May 1997, UT1-UTC=-0.3623091, Accumulated Leap Seconds=30
  { 2450570.50, 62.548590 }, //02 May 1997, UT1-UTC=-0.3645905, Accumulated Leap Seconds=30
  { 2450571.50, 62.550979 }, //03 May 1997, UT1-UTC=-0.3669788, Accumulated Leap Seconds=30
  { 2450572.50, 62.553456 }, //04 May 1997, UT1-UTC=-0.3694563, Accumulated Leap Seconds=30
  { 2450573.50, 62.555966 }, //05 May 1997, UT1-UTC=-0.3719660, Accumulated Leap Seconds=30
  { 2450574.50, 62.558440 }, //06 May 1997, UT1-UTC=-0.3744398, Accumulated Leap Seconds=30
  { 2450575.50, 62.560836 }, //07 May 1997, UT1-UTC=-0.3768360, Accumulated Leap Seconds=30
  { 2450576.50, 62.563092 }, //08 May 1997, UT1-UTC=-0.3790923, Accumulated Leap Seconds=30
  { 2450577.50, 62.565216 }, //09 May 1997, UT1-UTC=-0.3812157, Accumulated Leap Seconds=30
  { 2450578.50, 62.567260 }, //10 May 1997, UT1-UTC=-0.3832600, Accumulated Leap Seconds=30
  { 2450579.50, 62.569249 }, //11 May 1997, UT1-UTC=-0.3852491, Accumulated Leap Seconds=30
  { 2450580.50, 62.571207 }, //12 May 1997, UT1-UTC=-0.3872073, Accumulated Leap Seconds=30
  { 2450581.50, 62.573157 }, //13 May 1997, UT1-UTC=-0.3891574, Accumulated Leap Seconds=30
  { 2450582.50, 62.575111 }, //14 May 1997, UT1-UTC=-0.3911112, Accumulated Leap Seconds=30
  { 2450583.50, 62.577072 }, //15 May 1997, UT1-UTC=-0.3930716, Accumulated Leap Seconds=30
  { 2450584.50, 62.579048 }, //16 May 1997, UT1-UTC=-0.3950484, Accumulated Leap Seconds=30
  { 2450585.50, 62.581060 }, //17 May 1997, UT1-UTC=-0.3970597, Accumulated Leap Seconds=30
  { 2450586.50, 62.583110 }, //18 May 1997, UT1-UTC=-0.3991102, Accumulated Leap Seconds=30
  { 2450587.50, 62.585187 }, //19 May 1997, UT1-UTC=-0.4011869, Accumulated Leap Seconds=30
  { 2450588.50, 62.587278 }, //20 May 1997, UT1-UTC=-0.4032776, Accumulated Leap Seconds=30
  { 2450589.50, 62.589367 }, //21 May 1997, UT1-UTC=-0.4053668, Accumulated Leap Seconds=30
  { 2450590.50, 62.591421 }, //22 May 1997, UT1-UTC=-0.4074215, Accumulated Leap Seconds=30
  { 2450591.50, 62.593427 }, //23 May 1997, UT1-UTC=-0.4094272, Accumulated Leap Seconds=30
  { 2450592.50, 62.595400 }, //24 May 1997, UT1-UTC=-0.4114004, Accumulated Leap Seconds=30
  { 2450593.50, 62.597370 }, //25 May 1997, UT1-UTC=-0.4133700, Accumulated Leap Seconds=30
  { 2450594.50, 62.599376 }, //26 May 1997, UT1-UTC=-0.4153756, Accumulated Leap Seconds=30
  { 2450595.50, 62.601468 }, //27 May 1997, UT1-UTC=-0.4174680, Accumulated Leap Seconds=30
  { 2450596.50, 62.603691 }, //28 May 1997, UT1-UTC=-0.4196912, Accumulated Leap Seconds=30
  { 2450597.50, 62.606063 }, //29 May 1997, UT1-UTC=-0.4220631, Accumulated Leap Seconds=30
  { 2450598.50, 62.608554 }, //30 May 1997, UT1-UTC=-0.4245540, Accumulated Leap Seconds=30
  { 2450599.50, 62.611092 }, //31 May 1997, UT1-UTC=-0.4270924, Accumulated Leap Seconds=30
  { 2450600.50, 62.613603 }, //01 June 1997, UT1-UTC=-0.4296031, Accumulated Leap Seconds=30
  { 2450601.50, 62.616024 }, //02 June 1997, UT1-UTC=-0.4320241, Accumulated Leap Seconds=30
  { 2450602.50, 62.618291 }, //03 June 1997, UT1-UTC=-0.4342910, Accumulated Leap Seconds=30
  { 2450603.50, 62.620362 }, //04 June 1997, UT1-UTC=-0.4363620, Accumulated Leap Seconds=30
  { 2450604.50, 62.622218 }, //05 June 1997, UT1-UTC=-0.4382183, Accumulated Leap Seconds=30
  { 2450605.50, 62.623888 }, //06 June 1997, UT1-UTC=-0.4398880, Accumulated Leap Seconds=30
  { 2450606.50, 62.625440 }, //07 June 1997, UT1-UTC=-0.4414402, Accumulated Leap Seconds=30
  { 2450607.50, 62.626925 }, //08 June 1997, UT1-UTC=-0.4429250, Accumulated Leap Seconds=30
  { 2450608.50, 62.628385 }, //09 June 1997, UT1-UTC=-0.4443850, Accumulated Leap Seconds=30
  { 2450609.50, 62.629859 }, //10 June 1997, UT1-UTC=-0.4458595, Accumulated Leap Seconds=30
  { 2450610.50, 62.631378 }, //11 June 1997, UT1-UTC=-0.4473781, Accumulated Leap Seconds=30
  { 2450611.50, 62.632949 }, //12 June 1997, UT1-UTC=-0.4489494, Accumulated Leap Seconds=30
  { 2450612.50, 62.634554 }, //13 June 1997, UT1-UTC=-0.4505541, Accumulated Leap Seconds=30
  { 2450613.50, 62.636163 }, //14 June 1997, UT1-UTC=-0.4521635, Accumulated Leap Seconds=30
  { 2450614.50, 62.637749 }, //15 June 1997, UT1-UTC=-0.4537489, Accumulated Leap Seconds=30
  { 2450615.50, 62.639279 }, //16 June 1997, UT1-UTC=-0.4552787, Accumulated Leap Seconds=30
  { 2450616.50, 62.640727 }, //17 June 1997, UT1-UTC=-0.4567268, Accumulated Leap Seconds=30
  { 2450617.50, 62.642072 }, //18 June 1997, UT1-UTC=-0.4580718, Accumulated Leap Seconds=30
  { 2450618.50, 62.643271 }, //19 June 1997, UT1-UTC=-0.4592707, Accumulated Leap Seconds=30
  { 2450619.50, 62.644335 }, //20 June 1997, UT1-UTC=-0.4603355, Accumulated Leap Seconds=30
  { 2450620.50, 62.645324 }, //21 June 1997, UT1-UTC=-0.4613241, Accumulated Leap Seconds=30
  { 2450621.50, 62.646302 }, //22 June 1997, UT1-UTC=-0.4623024, Accumulated Leap Seconds=30
  { 2450622.50, 62.647340 }, //23 June 1997, UT1-UTC=-0.4633401, Accumulated Leap Seconds=30
  { 2450623.50, 62.648481 }, //24 June 1997, UT1-UTC=-0.4644808, Accumulated Leap Seconds=30
  { 2450624.50, 62.649732 }, //25 June 1997, UT1-UTC=-0.4657320, Accumulated Leap Seconds=30
  { 2450625.50, 62.651069 }, //26 June 1997, UT1-UTC=-0.4670695, Accumulated Leap Seconds=30
  { 2450626.50, 62.652444 }, //27 June 1997, UT1-UTC=-0.4684438, Accumulated Leap Seconds=30
  { 2450627.50, 62.653783 }, //28 June 1997, UT1-UTC=-0.4697828, Accumulated Leap Seconds=30
  { 2450628.50, 62.655024 }, //29 June 1997, UT1-UTC=-0.4710244, Accumulated Leap Seconds=30
  { 2450629.50, 62.656129 }, //30 June 1997, UT1-UTC=-0.4721291, Accumulated Leap Seconds=30
  { 2450630.50, 62.657074 }, //01 July 1997, UT1-UTC= 0.5269261, Accumulated Leap Seconds=31
  { 2450631.50, 62.657858 }, //02 July 1997, UT1-UTC= 0.5261415, Accumulated Leap Seconds=31
  { 2450632.50, 62.658499 }, //03 July 1997, UT1-UTC= 0.5255008, Accumulated Leap Seconds=31
  { 2450633.50, 62.659044 }, //04 July 1997, UT1-UTC= 0.5249556, Accumulated Leap Seconds=31
  { 2450634.50, 62.659562 }, //05 July 1997, UT1-UTC= 0.5244381, Accumulated Leap Seconds=31
  { 2450635.50, 62.660114 }, //06 July 1997, UT1-UTC= 0.5238855, Accumulated Leap Seconds=31
  { 2450636.50, 62.660748 }, //07 July 1997, UT1-UTC= 0.5232517, Accumulated Leap Seconds=31
  { 2450637.50, 62.661480 }, //08 July 1997, UT1-UTC= 0.5225198, Accumulated Leap Seconds=31
  { 2450638.50, 62.662305 }, //09 July 1997, UT1-UTC= 0.5216947, Accumulated Leap Seconds=31
  { 2450639.50, 62.663210 }, //10 July 1997, UT1-UTC= 0.5207898, Accumulated Leap Seconds=31
  { 2450640.50, 62.664188 }, //11 July 1997, UT1-UTC= 0.5198116, Accumulated Leap Seconds=31
  { 2450641.50, 62.665227 }, //12 July 1997, UT1-UTC= 0.5187733, Accumulated Leap Seconds=31
  { 2450642.50, 62.666316 }, //13 July 1997, UT1-UTC= 0.5176843, Accumulated Leap Seconds=31
  { 2450643.50, 62.667432 }, //14 July 1997, UT1-UTC= 0.5165684, Accumulated Leap Seconds=31
  { 2450644.50, 62.668541 }, //15 July 1997, UT1-UTC= 0.5154588, Accumulated Leap Seconds=31
  { 2450645.50, 62.669606 }, //16 July 1997, UT1-UTC= 0.5143944, Accumulated Leap Seconds=31
  { 2450646.50, 62.670637 }, //17 July 1997, UT1-UTC= 0.5133634, Accumulated Leap Seconds=31
  { 2450647.50, 62.671701 }, //18 July 1997, UT1-UTC= 0.5122995, Accumulated Leap Seconds=31
  { 2450648.50, 62.672862 }, //19 July 1997, UT1-UTC= 0.5111376, Accumulated Leap Seconds=31
  { 2450649.50, 62.674170 }, //20 July 1997, UT1-UTC= 0.5098301, Accumulated Leap Seconds=31
  { 2450650.50, 62.675666 }, //21 July 1997, UT1-UTC= 0.5083338, Accumulated Leap Seconds=31
  { 2450651.50, 62.677359 }, //22 July 1997, UT1-UTC= 0.5066406, Accumulated Leap Seconds=31
  { 2450652.50, 62.679217 }, //23 July 1997, UT1-UTC= 0.5047831, Accumulated Leap Seconds=31
  { 2450653.50, 62.681183 }, //24 July 1997, UT1-UTC= 0.5028170, Accumulated Leap Seconds=31
  { 2450654.50, 62.683207 }, //25 July 1997, UT1-UTC= 0.5007926, Accumulated Leap Seconds=31
  { 2450655.50, 62.685181 }, //26 July 1997, UT1-UTC= 0.4988191, Accumulated Leap Seconds=31
  { 2450656.50, 62.687032 }, //27 July 1997, UT1-UTC= 0.4969678, Accumulated Leap Seconds=31
  { 2450657.50, 62.688725 }, //28 July 1997, UT1-UTC= 0.4952746, Accumulated Leap Seconds=31
  { 2450658.50, 62.690262 }, //29 July 1997, UT1-UTC= 0.4937384, Accumulated Leap Seconds=31
  { 2450659.50, 62.691658 }, //30 July 1997, UT1-UTC= 0.4923415, Accumulated Leap Seconds=31
  { 2450660.50, 62.692943 }, //31 July 1997, UT1-UTC= 0.4910565, Accumulated Leap Seconds=31
  { 2450661.50, 62.694174 }, //01 August 1997, UT1-UTC= 0.4898258, Accumulated Leap Seconds=31
  { 2450662.50, 62.695406 }, //02 August 1997, UT1-UTC= 0.4885938, Accumulated Leap Seconds=31
  { 2450663.50, 62.696662 }, //03 August 1997, UT1-UTC= 0.4873380, Accumulated Leap Seconds=31
  { 2450664.50, 62.697950 }, //04 August 1997, UT1-UTC= 0.4860495, Accumulated Leap Seconds=31
  { 2450665.50, 62.699288 }, //05 August 1997, UT1-UTC= 0.4847120, Accumulated Leap Seconds=31
  { 2450666.50, 62.700674 }, //06 August 1997, UT1-UTC= 0.4833255, Accumulated Leap Seconds=31
  { 2450667.50, 62.702102 }, //07 August 1997, UT1-UTC= 0.4818982, Accumulated Leap Seconds=31
  { 2450668.50, 62.703564 }, //08 August 1997, UT1-UTC= 0.4804359, Accumulated Leap Seconds=31
  { 2450669.50, 62.705042 }, //09 August 1997, UT1-UTC= 0.4789580, Accumulated Leap Seconds=31
  { 2450670.50, 62.706497 }, //10 August 1997, UT1-UTC= 0.4775031, Accumulated Leap Seconds=31
  { 2450671.50, 62.707893 }, //11 August 1997, UT1-UTC= 0.4761068, Accumulated Leap Seconds=31
  { 2450672.50, 62.709213 }, //12 August 1997, UT1-UTC= 0.4747865, Accumulated Leap Seconds=31
  { 2450673.50, 62.710460 }, //13 August 1997, UT1-UTC= 0.4735395, Accumulated Leap Seconds=31
  { 2450674.50, 62.711647 }, //14 August 1997, UT1-UTC= 0.4723531, Accumulated Leap Seconds=31
  { 2450675.50, 62.712808 }, //15 August 1997, UT1-UTC= 0.4711922, Accumulated Leap Seconds=31
  { 2450676.50, 62.714012 }, //16 August 1997, UT1-UTC= 0.4699884, Accumulated Leap Seconds=31
  { 2450677.50, 62.715335 }, //17 August 1997, UT1-UTC= 0.4686650, Accumulated Leap Seconds=31
  { 2450678.50, 62.716822 }, //18 August 1997, UT1-UTC= 0.4671775, Accumulated Leap Seconds=31
  { 2450679.50, 62.718476 }, //19 August 1997, UT1-UTC= 0.4655244, Accumulated Leap Seconds=31
  { 2450680.50, 62.720275 }, //20 August 1997, UT1-UTC= 0.4637249, Accumulated Leap Seconds=31
  { 2450681.50, 62.722173 }, //21 August 1997, UT1-UTC= 0.4618272, Accumulated Leap Seconds=31
  { 2450682.50, 62.724058 }, //22 August 1997, UT1-UTC= 0.4599417, Accumulated Leap Seconds=31
  { 2450683.50, 62.725836 }, //23 August 1997, UT1-UTC= 0.4581639, Accumulated Leap Seconds=31
  { 2450684.50, 62.727459 }, //24 August 1997, UT1-UTC= 0.4565411, Accumulated Leap Seconds=31
  { 2450685.50, 62.728920 }, //25 August 1997, UT1-UTC= 0.4550797, Accumulated Leap Seconds=31
  { 2450686.50, 62.730266 }, //26 August 1997, UT1-UTC= 0.4537342, Accumulated Leap Seconds=31
  { 2450687.50, 62.731565 }, //27 August 1997, UT1-UTC= 0.4524346, Accumulated Leap Seconds=31
  { 2450688.50, 62.732867 }, //28 August 1997, UT1-UTC= 0.4511327, Accumulated Leap Seconds=31
  { 2450689.50, 62.734178 }, //29 August 1997, UT1-UTC= 0.4498222, Accumulated Leap Seconds=31
  { 2450690.50, 62.735510 }, //30 August 1997, UT1-UTC= 0.4484896, Accumulated Leap Seconds=31
  { 2450691.50, 62.736885 }, //31 August 1997, UT1-UTC= 0.4471150, Accumulated Leap Seconds=31
  { 2450692.50, 62.738327 }, //01 September 1997, UT1-UTC= 0.4456729, Accumulated Leap Seconds=31
  { 2450693.50, 62.739847 }, //02 September 1997, UT1-UTC= 0.4441532, Accumulated Leap Seconds=31
  { 2450694.50, 62.741428 }, //03 September 1997, UT1-UTC= 0.4425718, Accumulated Leap Seconds=31
  { 2450695.50, 62.743043 }, //04 September 1997, UT1-UTC= 0.4409570, Accumulated Leap Seconds=31
  { 2450696.50, 62.744652 }, //05 September 1997, UT1-UTC= 0.4393482, Accumulated Leap Seconds=31
  { 2450697.50, 62.746232 }, //06 September 1997, UT1-UTC= 0.4377683, Accumulated Leap Seconds=31
  { 2450698.50, 62.747778 }, //07 September 1997, UT1-UTC= 0.4362221, Accumulated Leap Seconds=31
  { 2450699.50, 62.749287 }, //08 September 1997, UT1-UTC= 0.4347132, Accumulated Leap Seconds=31
  { 2450700.50, 62.750724 }, //09 September 1997, UT1-UTC= 0.4332759, Accumulated Leap Seconds=31
  { 2450701.50, 62.752084 }, //10 September 1997, UT1-UTC= 0.4319162, Accumulated Leap Seconds=31
  { 2450702.50, 62.753460 }, //11 September 1997, UT1-UTC= 0.4305396, Accumulated Leap Seconds=31
  { 2450703.50, 62.754920 }, //12 September 1997, UT1-UTC= 0.4290799, Accumulated Leap Seconds=31
  { 2450704.50, 62.756498 }, //13 September 1997, UT1-UTC= 0.4275021, Accumulated Leap Seconds=31
  { 2450705.50, 62.758230 }, //14 September 1997, UT1-UTC= 0.4257695, Accumulated Leap Seconds=31
  { 2450706.50, 62.760171 }, //15 September 1997, UT1-UTC= 0.4238285, Accumulated Leap Seconds=31
  { 2450707.50, 62.762338 }, //16 September 1997, UT1-UTC= 0.4216620, Accumulated Leap Seconds=31
  { 2450708.50, 62.764689 }, //17 September 1997, UT1-UTC= 0.4193113, Accumulated Leap Seconds=31
  { 2450709.50, 62.767130 }, //18 September 1997, UT1-UTC= 0.4168703, Accumulated Leap Seconds=31
  { 2450710.50, 62.769539 }, //19 September 1997, UT1-UTC= 0.4144613, Accumulated Leap Seconds=31
  { 2450711.50, 62.771810 }, //20 September 1997, UT1-UTC= 0.4121901, Accumulated Leap Seconds=31
  { 2450712.50, 62.773897 }, //21 September 1997, UT1-UTC= 0.4101030, Accumulated Leap Seconds=31
  { 2450713.50, 62.775817 }, //22 September 1997, UT1-UTC= 0.4081827, Accumulated Leap Seconds=31
  { 2450714.50, 62.777617 }, //23 September 1997, UT1-UTC= 0.4063825, Accumulated Leap Seconds=31
  { 2450715.50, 62.779355 }, //24 September 1997, UT1-UTC= 0.4046454, Accumulated Leap Seconds=31
  { 2450716.50, 62.781081 }, //25 September 1997, UT1-UTC= 0.4029194, Accumulated Leap Seconds=31
  { 2450717.50, 62.782851 }, //26 September 1997, UT1-UTC= 0.4011487, Accumulated Leap Seconds=31
  { 2450718.50, 62.784709 }, //27 September 1997, UT1-UTC= 0.3992914, Accumulated Leap Seconds=31
  { 2450719.50, 62.786638 }, //28 September 1997, UT1-UTC= 0.3973618, Accumulated Leap Seconds=31
  { 2450720.50, 62.788618 }, //29 September 1997, UT1-UTC= 0.3953822, Accumulated Leap Seconds=31
  { 2450721.50, 62.790619 }, //30 September 1997, UT1-UTC= 0.3933814, Accumulated Leap Seconds=31
  { 2450722.50, 62.792630 }, //01 October 1997, UT1-UTC= 0.3913695, Accumulated Leap Seconds=31
  { 2450723.50, 62.794689 }, //02 October 1997, UT1-UTC= 0.3893106, Accumulated Leap Seconds=31
  { 2450724.50, 62.796761 }, //03 October 1997, UT1-UTC= 0.3872390, Accumulated Leap Seconds=31
  { 2450725.50, 62.798789 }, //04 October 1997, UT1-UTC= 0.3852107, Accumulated Leap Seconds=31
  { 2450726.50, 62.800747 }, //05 October 1997, UT1-UTC= 0.3832534, Accumulated Leap Seconds=31
  { 2450727.50, 62.802633 }, //06 October 1997, UT1-UTC= 0.3813666, Accumulated Leap Seconds=31
  { 2450728.50, 62.804474 }, //07 October 1997, UT1-UTC= 0.3795260, Accumulated Leap Seconds=31
  { 2450729.50, 62.806308 }, //08 October 1997, UT1-UTC= 0.3776920, Accumulated Leap Seconds=31
  { 2450730.50, 62.808188 }, //09 October 1997, UT1-UTC= 0.3758123, Accumulated Leap Seconds=31
  { 2450731.50, 62.810154 }, //10 October 1997, UT1-UTC= 0.3738457, Accumulated Leap Seconds=31
  { 2450732.50, 62.812238 }, //11 October 1997, UT1-UTC= 0.3717615, Accumulated Leap Seconds=31
  { 2450733.50, 62.814471 }, //12 October 1997, UT1-UTC= 0.3695290, Accumulated Leap Seconds=31
  { 2450734.50, 62.816864 }, //13 October 1997, UT1-UTC= 0.3671361, Accumulated Leap Seconds=31
  { 2450735.50, 62.819386 }, //14 October 1997, UT1-UTC= 0.3646143, Accumulated Leap Seconds=31
  { 2450736.50, 62.821976 }, //15 October 1997, UT1-UTC= 0.3620244, Accumulated Leap Seconds=31
  { 2450737.50, 62.824565 }, //16 October 1997, UT1-UTC= 0.3594354, Accumulated Leap Seconds=31
  { 2450738.50, 62.827072 }, //17 October 1997, UT1-UTC= 0.3569280, Accumulated Leap Seconds=31
  { 2450739.50, 62.829425 }, //18 October 1997, UT1-UTC= 0.3545749, Accumulated Leap Seconds=31
  { 2450740.50, 62.831595 }, //19 October 1997, UT1-UTC= 0.3524048, Accumulated Leap Seconds=31
  { 2450741.50, 62.833603 }, //20 October 1997, UT1-UTC= 0.3503967, Accumulated Leap Seconds=31
  { 2450742.50, 62.835489 }, //21 October 1997, UT1-UTC= 0.3485114, Accumulated Leap Seconds=31
  { 2450743.50, 62.837293 }, //22 October 1997, UT1-UTC= 0.3467065, Accumulated Leap Seconds=31
  { 2450744.50, 62.839051 }, //23 October 1997, UT1-UTC= 0.3449494, Accumulated Leap Seconds=31
  { 2450745.50, 62.840806 }, //24 October 1997, UT1-UTC= 0.3431936, Accumulated Leap Seconds=31
  { 2450746.50, 62.842603 }, //25 October 1997, UT1-UTC= 0.3413970, Accumulated Leap Seconds=31
  { 2450747.50, 62.844459 }, //26 October 1997, UT1-UTC= 0.3395408, Accumulated Leap Seconds=31
  { 2450748.50, 62.846379 }, //27 October 1997, UT1-UTC= 0.3376205, Accumulated Leap Seconds=31
  { 2450749.50, 62.848357 }, //28 October 1997, UT1-UTC= 0.3356433, Accumulated Leap Seconds=31
  { 2450750.50, 62.850387 }, //29 October 1997, UT1-UTC= 0.3336132, Accumulated Leap Seconds=31
  { 2450751.50, 62.852460 }, //30 October 1997, UT1-UTC= 0.3315395, Accumulated Leap Seconds=31
  { 2450752.50, 62.854568 }, //31 October 1997, UT1-UTC= 0.3294321, Accumulated Leap Seconds=31
  { 2450753.50, 62.856699 }, //01 November 1997, UT1-UTC= 0.3273014, Accumulated Leap Seconds=31
  { 2450754.50, 62.858816 }, //02 November 1997, UT1-UTC= 0.3251839, Accumulated Leap Seconds=31
  { 2450755.50, 62.860893 }, //03 November 1997, UT1-UTC= 0.3231068, Accumulated Leap Seconds=31
  { 2450756.50, 62.862932 }, //04 November 1997, UT1-UTC= 0.3210681, Accumulated Leap Seconds=31
  { 2450757.50, 62.864973 }, //05 November 1997, UT1-UTC= 0.3190272, Accumulated Leap Seconds=31
  { 2450758.50, 62.867102 }, //06 November 1997, UT1-UTC= 0.3168982, Accumulated Leap Seconds=31
  { 2450759.50, 62.869344 }, //07 November 1997, UT1-UTC= 0.3146564, Accumulated Leap Seconds=31
  { 2450760.50, 62.871695 }, //08 November 1997, UT1-UTC= 0.3123050, Accumulated Leap Seconds=31
  { 2450761.50, 62.874168 }, //09 November 1997, UT1-UTC= 0.3098320, Accumulated Leap Seconds=31
  { 2450762.50, 62.876774 }, //10 November 1997, UT1-UTC= 0.3072260, Accumulated Leap Seconds=31
  { 2450763.50, 62.879454 }, //11 November 1997, UT1-UTC= 0.3045463, Accumulated Leap Seconds=31
  { 2450764.50, 62.882115 }, //12 November 1997, UT1-UTC= 0.3018847, Accumulated Leap Seconds=31
  { 2450765.50, 62.884663 }, //13 November 1997, UT1-UTC= 0.2993371, Accumulated Leap Seconds=31
  { 2450766.50, 62.887000 }, //14 November 1997, UT1-UTC= 0.2970003, Accumulated Leap Seconds=31
  { 2450767.50, 62.889077 }, //15 November 1997, UT1-UTC= 0.2949232, Accumulated Leap Seconds=31
  { 2450768.50, 62.890927 }, //16 November 1997, UT1-UTC= 0.2930731, Accumulated Leap Seconds=31
  { 2450769.50, 62.892618 }, //17 November 1997, UT1-UTC= 0.2913821, Accumulated Leap Seconds=31
  { 2450770.50, 62.894214 }, //18 November 1997, UT1-UTC= 0.2897863, Accumulated Leap Seconds=31
  { 2450771.50, 62.895779 }, //19 November 1997, UT1-UTC= 0.2882205, Accumulated Leap Seconds=31
  { 2450772.50, 62.897368 }, //20 November 1997, UT1-UTC= 0.2866322, Accumulated Leap Seconds=31
  { 2450773.50, 62.898973 }, //21 November 1997, UT1-UTC= 0.2850269, Accumulated Leap Seconds=31
  { 2450774.50, 62.900580 }, //22 November 1997, UT1-UTC= 0.2834204, Accumulated Leap Seconds=31
  { 2450775.50, 62.902200 }, //23 November 1997, UT1-UTC= 0.2818001, Accumulated Leap Seconds=31
  { 2450776.50, 62.903839 }, //24 November 1997, UT1-UTC= 0.2801608, Accumulated Leap Seconds=31
  { 2450777.50, 62.905488 }, //25 November 1997, UT1-UTC= 0.2785119, Accumulated Leap Seconds=31
  { 2450778.50, 62.907126 }, //26 November 1997, UT1-UTC= 0.2768741, Accumulated Leap Seconds=31
  { 2450779.50, 62.908724 }, //27 November 1997, UT1-UTC= 0.2752764, Accumulated Leap Seconds=31
  { 2450780.50, 62.910265 }, //28 November 1997, UT1-UTC= 0.2737345, Accumulated Leap Seconds=31
  { 2450781.50, 62.911745 }, //29 November 1997, UT1-UTC= 0.2722546, Accumulated Leap Seconds=31
  { 2450782.50, 62.913170 }, //30 November 1997, UT1-UTC= 0.2708301, Accumulated Leap Seconds=31
  { 2450783.50, 62.914561 }, //01 December 1997, UT1-UTC= 0.2694393, Accumulated Leap Seconds=31
  { 2450784.50, 62.915945 }, //02 December 1997, UT1-UTC= 0.2680547, Accumulated Leap Seconds=31
  { 2450785.50, 62.917354 }, //03 December 1997, UT1-UTC= 0.2666459, Accumulated Leap Seconds=31
  { 2450786.50, 62.918813 }, //04 December 1997, UT1-UTC= 0.2651873, Accumulated Leap Seconds=31
  { 2450787.50, 62.920377 }, //05 December 1997, UT1-UTC= 0.2636229, Accumulated Leap Seconds=31
  { 2450788.50, 62.922105 }, //06 December 1997, UT1-UTC= 0.2618948, Accumulated Leap Seconds=31
  { 2450789.50, 62.924046 }, //07 December 1997, UT1-UTC= 0.2599537, Accumulated Leap Seconds=31
  { 2450790.50, 62.926196 }, //08 December 1997, UT1-UTC= 0.2578036, Accumulated Leap Seconds=31
  { 2450791.50, 62.928463 }, //09 December 1997, UT1-UTC= 0.2555369, Accumulated Leap Seconds=31
  { 2450792.50, 62.930735 }, //10 December 1997, UT1-UTC= 0.2532653, Accumulated Leap Seconds=31
  { 2450793.50, 62.932916 }, //11 December 1997, UT1-UTC= 0.2510836, Accumulated Leap Seconds=31
  { 2450794.50, 62.934964 }, //12 December 1997, UT1-UTC= 0.2490356, Accumulated Leap Seconds=31
  { 2450795.50, 62.936847 }, //13 December 1997, UT1-UTC= 0.2471526, Accumulated Leap Seconds=31
  { 2450796.50, 62.938558 }, //14 December 1997, UT1-UTC= 0.2454420, Accumulated Leap Seconds=31
  { 2450797.50, 62.940135 }, //15 December 1997, UT1-UTC= 0.2438651, Accumulated Leap Seconds=31
  { 2450798.50, 62.941641 }, //16 December 1997, UT1-UTC= 0.2423586, Accumulated Leap Seconds=31
  { 2450799.50, 62.943120 }, //17 December 1997, UT1-UTC= 0.2408796, Accumulated Leap Seconds=31
  { 2450800.50, 62.944609 }, //18 December 1997, UT1-UTC= 0.2393906, Accumulated Leap Seconds=31
  { 2450801.50, 62.946130 }, //19 December 1997, UT1-UTC= 0.2378700, Accumulated Leap Seconds=31
  { 2450802.50, 62.947708 }, //20 December 1997, UT1-UTC= 0.2362915, Accumulated Leap Seconds=31
  { 2450803.50, 62.949349 }, //21 December 1997, UT1-UTC= 0.2346512, Accumulated Leap Seconds=31
  { 2450804.50, 62.951032 }, //22 December 1997, UT1-UTC= 0.2329681, Accumulated Leap Seconds=31
  { 2450805.50, 62.952746 }, //23 December 1997, UT1-UTC= 0.2312540, Accumulated Leap Seconds=31
  { 2450806.50, 62.954451 }, //24 December 1997, UT1-UTC= 0.2295493, Accumulated Leap Seconds=31
  { 2450807.50, 62.956101 }, //25 December 1997, UT1-UTC= 0.2278985, Accumulated Leap Seconds=31
  { 2450808.50, 62.957672 }, //26 December 1997, UT1-UTC= 0.2263280, Accumulated Leap Seconds=31
  { 2450809.50, 62.959144 }, //27 December 1997, UT1-UTC= 0.2248562, Accumulated Leap Seconds=31
  { 2450810.50, 62.960516 }, //28 December 1997, UT1-UTC= 0.2234837, Accumulated Leap Seconds=31
  { 2450811.50, 62.961816 }, //29 December 1997, UT1-UTC= 0.2221838, Accumulated Leap Seconds=31
  { 2450812.50, 62.963104 }, //30 December 1997, UT1-UTC= 0.2208956, Accumulated Leap Seconds=31
  { 2450813.50, 62.964439 }, //31 December 1997, UT1-UTC= 0.2195610, Accumulated Leap Seconds=31
  { 2450814.50, 62.965869 }, //01 January 1998, UT1-UTC= 0.2181311, Accumulated Leap Seconds=31
  { 2450815.50, 62.967448 }, //02 January 1998, UT1-UTC= 0.2165517, Accumulated Leap Seconds=31
  { 2450816.50, 62.969208 }, //03 January 1998, UT1-UTC= 0.2147923, Accumulated Leap Seconds=31
  { 2450817.50, 62.971143 }, //04 January 1998, UT1-UTC= 0.2128567, Accumulated Leap Seconds=31
  { 2450818.50, 62.973206 }, //05 January 1998, UT1-UTC= 0.2107935, Accumulated Leap Seconds=31
  { 2450819.50, 62.975280 }, //06 January 1998, UT1-UTC= 0.2087199, Accumulated Leap Seconds=31
  { 2450820.50, 62.977261 }, //07 January 1998, UT1-UTC= 0.2067387, Accumulated Leap Seconds=31
  { 2450821.50, 62.979107 }, //08 January 1998, UT1-UTC= 0.2048928, Accumulated Leap Seconds=31
  { 2450822.50, 62.980788 }, //09 January 1998, UT1-UTC= 0.2032115, Accumulated Leap Seconds=31
  { 2450823.50, 62.982306 }, //10 January 1998, UT1-UTC= 0.2016942, Accumulated Leap Seconds=31
  { 2450824.50, 62.983721 }, //11 January 1998, UT1-UTC= 0.2002789, Accumulated Leap Seconds=31
  { 2450825.50, 62.985114 }, //12 January 1998, UT1-UTC= 0.1988863, Accumulated Leap Seconds=31
  { 2450826.50, 62.986538 }, //13 January 1998, UT1-UTC= 0.1974619, Accumulated Leap Seconds=31
  { 2450827.50, 62.988045 }, //14 January 1998, UT1-UTC= 0.1959551, Accumulated Leap Seconds=31
  { 2450828.50, 62.989641 }, //15 January 1998, UT1-UTC= 0.1943587, Accumulated Leap Seconds=31
  { 2450829.50, 62.991296 }, //16 January 1998, UT1-UTC= 0.1927044, Accumulated Leap Seconds=31
  { 2450830.50, 62.992983 }, //17 January 1998, UT1-UTC= 0.1910170, Accumulated Leap Seconds=31
  { 2450831.50, 62.994701 }, //18 January 1998, UT1-UTC= 0.1892995, Accumulated Leap Seconds=31
  { 2450832.50, 62.996448 }, //19 January 1998, UT1-UTC= 0.1875517, Accumulated Leap Seconds=31
  { 2450833.50, 62.998206 }, //20 January 1998, UT1-UTC= 0.1857944, Accumulated Leap Seconds=31
  { 2450834.50, 62.999947 }, //21 January 1998, UT1-UTC= 0.1840535, Accumulated Leap Seconds=31
  { 2450835.50, 63.001667 }, //22 January 1998, UT1-UTC= 0.1823334, Accumulated Leap Seconds=31
  { 2450836.50, 63.003368 }, //23 January 1998, UT1-UTC= 0.1806316, Accumulated Leap Seconds=31
  { 2450837.50, 63.005064 }, //24 January 1998, UT1-UTC= 0.1789359, Accumulated Leap Seconds=31
  { 2450838.50, 63.006763 }, //25 January 1998, UT1-UTC= 0.1772365, Accumulated Leap Seconds=31
  { 2450839.50, 63.008469 }, //26 January 1998, UT1-UTC= 0.1755314, Accumulated Leap Seconds=31
  { 2450840.50, 63.010233 }, //27 January 1998, UT1-UTC= 0.1737675, Accumulated Leap Seconds=31
  { 2450841.50, 63.012143 }, //28 January 1998, UT1-UTC= 0.1718573, Accumulated Leap Seconds=31
  { 2450842.50, 63.014274 }, //29 January 1998, UT1-UTC= 0.1697264, Accumulated Leap Seconds=31
  { 2450843.50, 63.016619 }, //30 January 1998, UT1-UTC= 0.1673805, Accumulated Leap Seconds=31
  { 2450844.50, 63.019115 }, //31 January 1998, UT1-UTC= 0.1648847, Accumulated Leap Seconds=31
  { 2450845.50, 63.021663 }, //01 February 1998, UT1-UTC= 0.1623368, Accumulated Leap Seconds=31
  { 2450846.50, 63.024214 }, //02 February 1998, UT1-UTC= 0.1597864, Accumulated Leap Seconds=31
  { 2450847.50, 63.026723 }, //03 February 1998, UT1-UTC= 0.1572768, Accumulated Leap Seconds=31
  { 2450848.50, 63.029162 }, //04 February 1998, UT1-UTC= 0.1548382, Accumulated Leap Seconds=31
  { 2450849.50, 63.031528 }, //05 February 1998, UT1-UTC= 0.1524722, Accumulated Leap Seconds=31
  { 2450850.50, 63.033789 }, //06 February 1998, UT1-UTC= 0.1502108, Accumulated Leap Seconds=31
  { 2450851.50, 63.035929 }, //07 February 1998, UT1-UTC= 0.1480705, Accumulated Leap Seconds=31
  { 2450852.50, 63.037999 }, //08 February 1998, UT1-UTC= 0.1460010, Accumulated Leap Seconds=31
  { 2450853.50, 63.040036 }, //09 February 1998, UT1-UTC= 0.1439637, Accumulated Leap Seconds=31
  { 2450854.50, 63.042018 }, //10 February 1998, UT1-UTC= 0.1419822, Accumulated Leap Seconds=31
  { 2450855.50, 63.043944 }, //11 February 1998, UT1-UTC= 0.1400562, Accumulated Leap Seconds=31
  { 2450856.50, 63.045880 }, //12 February 1998, UT1-UTC= 0.1381201, Accumulated Leap Seconds=31
  { 2450857.50, 63.047847 }, //13 February 1998, UT1-UTC= 0.1361528, Accumulated Leap Seconds=31
  { 2450858.50, 63.049838 }, //14 February 1998, UT1-UTC= 0.1341616, Accumulated Leap Seconds=31
  { 2450859.50, 63.051835 }, //15 February 1998, UT1-UTC= 0.1321653, Accumulated Leap Seconds=31
  { 2450860.50, 63.053821 }, //16 February 1998, UT1-UTC= 0.1301790, Accumulated Leap Seconds=31
  { 2450861.50, 63.055783 }, //17 February 1998, UT1-UTC= 0.1282172, Accumulated Leap Seconds=31
  { 2450862.50, 63.057688 }, //18 February 1998, UT1-UTC= 0.1263122, Accumulated Leap Seconds=31
  { 2450863.50, 63.059499 }, //19 February 1998, UT1-UTC= 0.1245008, Accumulated Leap Seconds=31
  { 2450864.50, 63.061271 }, //20 February 1998, UT1-UTC= 0.1227290, Accumulated Leap Seconds=31
  { 2450865.50, 63.063061 }, //21 February 1998, UT1-UTC= 0.1209391, Accumulated Leap Seconds=31
  { 2450866.50, 63.064872 }, //22 February 1998, UT1-UTC= 0.1191277, Accumulated Leap Seconds=31
  { 2450867.50, 63.066735 }, //23 February 1998, UT1-UTC= 0.1172648, Accumulated Leap Seconds=31
  { 2450868.50, 63.068702 }, //24 February 1998, UT1-UTC= 0.1152984, Accumulated Leap Seconds=31
  { 2450869.50, 63.070819 }, //25 February 1998, UT1-UTC= 0.1131814, Accumulated Leap Seconds=31
  { 2450870.50, 63.073098 }, //26 February 1998, UT1-UTC= 0.1109021, Accumulated Leap Seconds=31
  { 2450871.50, 63.075519 }, //27 February 1998, UT1-UTC= 0.1084813, Accumulated Leap Seconds=31
  { 2450872.50, 63.078065 }, //28 February 1998, UT1-UTC= 0.1059353, Accumulated Leap Seconds=31
  { 2450873.50, 63.080705 }, //01 March 1998, UT1-UTC= 0.1032948, Accumulated Leap Seconds=31
  { 2450874.50, 63.083337 }, //02 March 1998, UT1-UTC= 0.1006629, Accumulated Leap Seconds=31
  { 2450875.50, 63.085857 }, //03 March 1998, UT1-UTC= 0.0981429, Accumulated Leap Seconds=31
  { 2450876.50, 63.088201 }, //04 March 1998, UT1-UTC= 0.0957989, Accumulated Leap Seconds=31
  { 2450877.50, 63.090375 }, //05 March 1998, UT1-UTC= 0.0936253, Accumulated Leap Seconds=31
  { 2450878.50, 63.092412 }, //06 March 1998, UT1-UTC= 0.0915884, Accumulated Leap Seconds=31
  { 2450879.50, 63.094347 }, //07 March 1998, UT1-UTC= 0.0896525, Accumulated Leap Seconds=31
  { 2450880.50, 63.096217 }, //08 March 1998, UT1-UTC= 0.0877829, Accumulated Leap Seconds=31
  { 2450881.50, 63.098058 }, //09 March 1998, UT1-UTC= 0.0859420, Accumulated Leap Seconds=31
  { 2450882.50, 63.099917 }, //10 March 1998, UT1-UTC= 0.0840831, Accumulated Leap Seconds=31
  { 2450883.50, 63.101850 }, //11 March 1998, UT1-UTC= 0.0821495, Accumulated Leap Seconds=31
  { 2450884.50, 63.103905 }, //12 March 1998, UT1-UTC= 0.0800953, Accumulated Leap Seconds=31
  { 2450885.50, 63.106013 }, //13 March 1998, UT1-UTC= 0.0779874, Accumulated Leap Seconds=31
  { 2450886.50, 63.108137 }, //14 March 1998, UT1-UTC= 0.0758629, Accumulated Leap Seconds=31
  { 2450887.50, 63.110254 }, //15 March 1998, UT1-UTC= 0.0737455, Accumulated Leap Seconds=31
  { 2450888.50, 63.112335 }, //16 March 1998, UT1-UTC= 0.0716654, Accumulated Leap Seconds=31
  { 2450889.50, 63.114357 }, //17 March 1998, UT1-UTC= 0.0696432, Accumulated Leap Seconds=31
  { 2450890.50, 63.116301 }, //18 March 1998, UT1-UTC= 0.0676988, Accumulated Leap Seconds=31
  { 2450891.50, 63.118159 }, //19 March 1998, UT1-UTC= 0.0658410, Accumulated Leap Seconds=31
  { 2450892.50, 63.119946 }, //20 March 1998, UT1-UTC= 0.0640541, Accumulated Leap Seconds=31
  { 2450893.50, 63.121694 }, //21 March 1998, UT1-UTC= 0.0623061, Accumulated Leap Seconds=31
  { 2450894.50, 63.123447 }, //22 March 1998, UT1-UTC= 0.0605529, Accumulated Leap Seconds=31
  { 2450895.50, 63.125254 }, //23 March 1998, UT1-UTC= 0.0587462, Accumulated Leap Seconds=31
  { 2450896.50, 63.127164 }, //24 March 1998, UT1-UTC= 0.0568356, Accumulated Leap Seconds=31
  { 2450897.50, 63.129236 }, //25 March 1998, UT1-UTC= 0.0547640, Accumulated Leap Seconds=31
  { 2450898.50, 63.131544 }, //26 March 1998, UT1-UTC= 0.0524558, Accumulated Leap Seconds=31
  { 2450899.50, 63.134054 }, //27 March 1998, UT1-UTC= 0.0499457, Accumulated Leap Seconds=31
  { 2450900.50, 63.136646 }, //28 March 1998, UT1-UTC= 0.0473542, Accumulated Leap Seconds=31
  { 2450901.50, 63.139217 }, //29 March 1998, UT1-UTC= 0.0447827, Accumulated Leap Seconds=31
  { 2450902.50, 63.141704 }, //30 March 1998, UT1-UTC= 0.0422962, Accumulated Leap Seconds=31
  { 2450903.50, 63.144031 }, //31 March 1998, UT1-UTC= 0.0399693, Accumulated Leap Seconds=31
  { 2450904.50, 63.146172 }, //01 April 1998, UT1-UTC= 0.0378282, Accumulated Leap Seconds=31
  { 2450905.50, 63.148160 }, //02 April 1998, UT1-UTC= 0.0358396, Accumulated Leap Seconds=31
  { 2450906.50, 63.150038 }, //03 April 1998, UT1-UTC= 0.0339623, Accumulated Leap Seconds=31
  { 2450907.50, 63.151856 }, //04 April 1998, UT1-UTC= 0.0321444, Accumulated Leap Seconds=31
  { 2450908.50, 63.153666 }, //05 April 1998, UT1-UTC= 0.0303344, Accumulated Leap Seconds=31
  { 2450909.50, 63.155501 }, //06 April 1998, UT1-UTC= 0.0284990, Accumulated Leap Seconds=31
  { 2450910.50, 63.157378 }, //07 April 1998, UT1-UTC= 0.0266224, Accumulated Leap Seconds=31
  { 2450911.50, 63.159302 }, //08 April 1998, UT1-UTC= 0.0246978, Accumulated Leap Seconds=31
  { 2450912.50, 63.161266 }, //09 April 1998, UT1-UTC= 0.0227341, Accumulated Leap Seconds=31
  { 2450913.50, 63.163215 }, //10 April 1998, UT1-UTC= 0.0207847, Accumulated Leap Seconds=31
  { 2450914.50, 63.165153 }, //11 April 1998, UT1-UTC= 0.0188465, Accumulated Leap Seconds=31
  { 2450915.50, 63.167105 }, //12 April 1998, UT1-UTC= 0.0168945, Accumulated Leap Seconds=31
  { 2450916.50, 63.169052 }, //13 April 1998, UT1-UTC= 0.0149480, Accumulated Leap Seconds=31
  { 2450917.50, 63.170940 }, //14 April 1998, UT1-UTC= 0.0130601, Accumulated Leap Seconds=31
  { 2450918.50, 63.172742 }, //15 April 1998, UT1-UTC= 0.0112583, Accumulated Leap Seconds=31
  { 2450919.50, 63.174478 }, //16 April 1998, UT1-UTC= 0.0095222, Accumulated Leap Seconds=31
  { 2450920.50, 63.176146 }, //17 April 1998, UT1-UTC= 0.0078539, Accumulated Leap Seconds=31
  { 2450921.50, 63.177803 }, //18 April 1998, UT1-UTC= 0.0061968, Accumulated Leap Seconds=31
  { 2450922.50, 63.179500 }, //19 April 1998, UT1-UTC= 0.0044997, Accumulated Leap Seconds=31
  { 2450923.50, 63.181284 }, //20 April 1998, UT1-UTC= 0.0027157, Accumulated Leap Seconds=31
  { 2450924.50, 63.183204 }, //21 April 1998, UT1-UTC= 0.0007964, Accumulated Leap Seconds=31
  { 2450925.50, 63.185299 }, //22 April 1998, UT1-UTC=-0.0012989, Accumulated Leap Seconds=31
  { 2450926.50, 63.187579 }, //23 April 1998, UT1-UTC=-0.0035787, Accumulated Leap Seconds=31
  { 2450927.50, 63.190035 }, //24 April 1998, UT1-UTC=-0.0060350, Accumulated Leap Seconds=31
  { 2450928.50, 63.192599 }, //25 April 1998, UT1-UTC=-0.0085988, Accumulated Leap Seconds=31
  { 2450929.50, 63.195134 }, //26 April 1998, UT1-UTC=-0.0111338, Accumulated Leap Seconds=31
  { 2450930.50, 63.197517 }, //27 April 1998, UT1-UTC=-0.0135172, Accumulated Leap Seconds=31
  { 2450931.50, 63.199692 }, //28 April 1998, UT1-UTC=-0.0156916, Accumulated Leap Seconds=31
  { 2450932.50, 63.201666 }, //29 April 1998, UT1-UTC=-0.0176657, Accumulated Leap Seconds=31
  { 2450933.50, 63.203506 }, //30 April 1998, UT1-UTC=-0.0195062, Accumulated Leap Seconds=31
  { 2450934.50, 63.205295 }, //01 May 1998, UT1-UTC=-0.0212946, Accumulated Leap Seconds=31
  { 2450935.50, 63.207076 }, //02 May 1998, UT1-UTC=-0.0230758, Accumulated Leap Seconds=31
  { 2450936.50, 63.208858 }, //03 May 1998, UT1-UTC=-0.0248584, Accumulated Leap Seconds=31
  { 2450937.50, 63.210639 }, //04 May 1998, UT1-UTC=-0.0266392, Accumulated Leap Seconds=31
  { 2450938.50, 63.212418 }, //05 May 1998, UT1-UTC=-0.0284175, Accumulated Leap Seconds=31
  { 2450939.50, 63.214209 }, //06 May 1998, UT1-UTC=-0.0302087, Accumulated Leap Seconds=31
  { 2450940.50, 63.216052 }, //07 May 1998, UT1-UTC=-0.0320523, Accumulated Leap Seconds=31
  { 2450941.50, 63.217900 }, //08 May 1998, UT1-UTC=-0.0339004, Accumulated Leap Seconds=31
  { 2450942.50, 63.219714 }, //09 May 1998, UT1-UTC=-0.0357143, Accumulated Leap Seconds=31
  { 2450943.50, 63.221491 }, //10 May 1998, UT1-UTC=-0.0374909, Accumulated Leap Seconds=31
  { 2450944.50, 63.223206 }, //11 May 1998, UT1-UTC=-0.0392064, Accumulated Leap Seconds=31
  { 2450945.50, 63.224836 }, //12 May 1998, UT1-UTC=-0.0408359, Accumulated Leap Seconds=31
  { 2450946.50, 63.226369 }, //13 May 1998, UT1-UTC=-0.0423686, Accumulated Leap Seconds=31
  { 2450947.50, 63.227816 }, //14 May 1998, UT1-UTC=-0.0438156, Accumulated Leap Seconds=31
  { 2450948.50, 63.229222 }, //15 May 1998, UT1-UTC=-0.0452223, Accumulated Leap Seconds=31
  { 2450949.50, 63.230662 }, //16 May 1998, UT1-UTC=-0.0466623, Accumulated Leap Seconds=31
  { 2450950.50, 63.232212 }, //17 May 1998, UT1-UTC=-0.0482125, Accumulated Leap Seconds=31
  { 2450951.50, 63.233927 }, //18 May 1998, UT1-UTC=-0.0499273, Accumulated Leap Seconds=31
  { 2450952.50, 63.235838 }, //19 May 1998, UT1-UTC=-0.0518377, Accumulated Leap Seconds=31
  { 2450953.50, 63.237944 }, //20 May 1998, UT1-UTC=-0.0539439, Accumulated Leap Seconds=31
  { 2450954.50, 63.240206 }, //21 May 1998, UT1-UTC=-0.0562065, Accumulated Leap Seconds=31
  { 2450955.50, 63.242553 }, //22 May 1998, UT1-UTC=-0.0585527, Accumulated Leap Seconds=31
  { 2450956.50, 63.244945 }, //23 May 1998, UT1-UTC=-0.0609448, Accumulated Leap Seconds=31
  { 2450957.50, 63.247271 }, //24 May 1998, UT1-UTC=-0.0632712, Accumulated Leap Seconds=31
  { 2450958.50, 63.249424 }, //25 May 1998, UT1-UTC=-0.0654240, Accumulated Leap Seconds=31
  { 2450959.50, 63.251357 }, //26 May 1998, UT1-UTC=-0.0673566, Accumulated Leap Seconds=31
  { 2450960.50, 63.253071 }, //27 May 1998, UT1-UTC=-0.0690710, Accumulated Leap Seconds=31
  { 2450961.50, 63.254592 }, //28 May 1998, UT1-UTC=-0.0705917, Accumulated Leap Seconds=31
  { 2450962.50, 63.255976 }, //29 May 1998, UT1-UTC=-0.0719757, Accumulated Leap Seconds=31
  { 2450963.50, 63.257297 }, //30 May 1998, UT1-UTC=-0.0732972, Accumulated Leap Seconds=31
  { 2450964.50, 63.258613 }, //31 May 1998, UT1-UTC=-0.0746130, Accumulated Leap Seconds=31
  { 2450965.50, 63.259944 }, //01 June 1998, UT1-UTC=-0.0759441, Accumulated Leap Seconds=31
  { 2450966.50, 63.261271 }, //02 June 1998, UT1-UTC=-0.0772710, Accumulated Leap Seconds=31
  { 2450967.50, 63.262582 }, //03 June 1998, UT1-UTC=-0.0785817, Accumulated Leap Seconds=31
  { 2450968.50, 63.263878 }, //04 June 1998, UT1-UTC=-0.0798776, Accumulated Leap Seconds=31
  { 2450969.50, 63.265145 }, //05 June 1998, UT1-UTC=-0.0811453, Accumulated Leap Seconds=31
  { 2450970.50, 63.266356 }, //06 June 1998, UT1-UTC=-0.0823558, Accumulated Leap Seconds=31
  { 2450971.50, 63.267458 }, //07 June 1998, UT1-UTC=-0.0834578, Accumulated Leap Seconds=31
  { 2450972.50, 63.268432 }, //08 June 1998, UT1-UTC=-0.0844325, Accumulated Leap Seconds=31
  { 2450973.50, 63.269284 }, //09 June 1998, UT1-UTC=-0.0852839, Accumulated Leap Seconds=31
  { 2450974.50, 63.270032 }, //10 June 1998, UT1-UTC=-0.0860324, Accumulated Leap Seconds=31
  { 2450975.50, 63.270726 }, //11 June 1998, UT1-UTC=-0.0867264, Accumulated Leap Seconds=31
  { 2450976.50, 63.271406 }, //12 June 1998, UT1-UTC=-0.0874057, Accumulated Leap Seconds=31
  { 2450977.50, 63.272102 }, //13 June 1998, UT1-UTC=-0.0881018, Accumulated Leap Seconds=31
  { 2450978.50, 63.272861 }, //14 June 1998, UT1-UTC=-0.0888606, Accumulated Leap Seconds=31
  { 2450979.50, 63.273719 }, //15 June 1998, UT1-UTC=-0.0897194, Accumulated Leap Seconds=31
  { 2450980.50, 63.274653 }, //16 June 1998, UT1-UTC=-0.0906527, Accumulated Leap Seconds=31
  { 2450981.50, 63.275635 }, //17 June 1998, UT1-UTC=-0.0916352, Accumulated Leap Seconds=31
  { 2450982.50, 63.276642 }, //18 June 1998, UT1-UTC=-0.0926416, Accumulated Leap Seconds=31
  { 2450983.50, 63.277617 }, //19 June 1998, UT1-UTC=-0.0936165, Accumulated Leap Seconds=31
  { 2450984.50, 63.278541 }, //20 June 1998, UT1-UTC=-0.0945414, Accumulated Leap Seconds=31
  { 2450985.50, 63.279373 }, //21 June 1998, UT1-UTC=-0.0953728, Accumulated Leap Seconds=31
  { 2450986.50, 63.280063 }, //22 June 1998, UT1-UTC=-0.0960629, Accumulated Leap Seconds=31
  { 2450987.50, 63.280596 }, //23 June 1998, UT1-UTC=-0.0965964, Accumulated Leap Seconds=31
  { 2450988.50, 63.281014 }, //24 June 1998, UT1-UTC=-0.0970144, Accumulated Leap Seconds=31
  { 2450989.50, 63.281399 }, //25 June 1998, UT1-UTC=-0.0973995, Accumulated Leap Seconds=31
  { 2450990.50, 63.281792 }, //26 June 1998, UT1-UTC=-0.0977920, Accumulated Leap Seconds=31
  { 2450991.50, 63.282219 }, //27 June 1998, UT1-UTC=-0.0982186, Accumulated Leap Seconds=31
  { 2450992.50, 63.282686 }, //28 June 1998, UT1-UTC=-0.0986858, Accumulated Leap Seconds=31
  { 2450993.50, 63.283206 }, //29 June 1998, UT1-UTC=-0.0992064, Accumulated Leap Seconds=31
  { 2450994.50, 63.283787 }, //30 June 1998, UT1-UTC=-0.0997867, Accumulated Leap Seconds=31
  { 2450995.50, 63.284409 }, //01 July 1998, UT1-UTC=-0.1004088, Accumulated Leap Seconds=31
  { 2450996.50, 63.285033 }, //02 July 1998, UT1-UTC=-0.1010330, Accumulated Leap Seconds=31
  { 2450997.50, 63.285598 }, //03 July 1998, UT1-UTC=-0.1015981, Accumulated Leap Seconds=31
  { 2450998.50, 63.286067 }, //04 July 1998, UT1-UTC=-0.1020668, Accumulated Leap Seconds=31
  { 2450999.50, 63.286440 }, //05 July 1998, UT1-UTC=-0.1024404, Accumulated Leap Seconds=31
  { 2451000.50, 63.286719 }, //06 July 1998, UT1-UTC=-0.1027186, Accumulated Leap Seconds=31
  { 2451001.50, 63.286897 }, //07 July 1998, UT1-UTC=-0.1028967, Accumulated Leap Seconds=31
  { 2451002.50, 63.286982 }, //08 July 1998, UT1-UTC=-0.1029824, Accumulated Leap Seconds=31
  { 2451003.50, 63.287003 }, //09 July 1998, UT1-UTC=-0.1030035, Accumulated Leap Seconds=31
  { 2451004.50, 63.287042 }, //10 July 1998, UT1-UTC=-0.1030415, Accumulated Leap Seconds=31
  { 2451005.50, 63.287195 }, //11 July 1998, UT1-UTC=-0.1031947, Accumulated Leap Seconds=31
  { 2451006.50, 63.287523 }, //12 July 1998, UT1-UTC=-0.1035226, Accumulated Leap Seconds=31
  { 2451007.50, 63.288030 }, //13 July 1998, UT1-UTC=-0.1040298, Accumulated Leap Seconds=31
  { 2451008.50, 63.288681 }, //14 July 1998, UT1-UTC=-0.1046812, Accumulated Leap Seconds=31
  { 2451009.50, 63.289419 }, //15 July 1998, UT1-UTC=-0.1054190, Accumulated Leap Seconds=31
  { 2451010.50, 63.290174 }, //16 July 1998, UT1-UTC=-0.1061736, Accumulated Leap Seconds=31
  { 2451011.50, 63.290866 }, //17 July 1998, UT1-UTC=-0.1068659, Accumulated Leap Seconds=31
  { 2451012.50, 63.291439 }, //18 July 1998, UT1-UTC=-0.1074393, Accumulated Leap Seconds=31
  { 2451013.50, 63.291872 }, //19 July 1998, UT1-UTC=-0.1078724, Accumulated Leap Seconds=31
  { 2451014.50, 63.292164 }, //20 July 1998, UT1-UTC=-0.1081643, Accumulated Leap Seconds=31
  { 2451015.50, 63.292331 }, //21 July 1998, UT1-UTC=-0.1083311, Accumulated Leap Seconds=31
  { 2451016.50, 63.292423 }, //22 July 1998, UT1-UTC=-0.1084226, Accumulated Leap Seconds=31
  { 2451017.50, 63.292522 }, //23 July 1998, UT1-UTC=-0.1085224, Accumulated Leap Seconds=31
  { 2451018.50, 63.292684 }, //24 July 1998, UT1-UTC=-0.1086844, Accumulated Leap Seconds=31
  { 2451019.50, 63.292945 }, //25 July 1998, UT1-UTC=-0.1089446, Accumulated Leap Seconds=31
  { 2451020.50, 63.293314 }, //26 July 1998, UT1-UTC=-0.1093143, Accumulated Leap Seconds=31
  { 2451021.50, 63.293762 }, //27 July 1998, UT1-UTC=-0.1097623, Accumulated Leap Seconds=31
  { 2451022.50, 63.294247 }, //28 July 1998, UT1-UTC=-0.1102469, Accumulated Leap Seconds=31
  { 2451023.50, 63.294730 }, //29 July 1998, UT1-UTC=-0.1107304, Accumulated Leap Seconds=31
  { 2451024.50, 63.295193 }, //30 July 1998, UT1-UTC=-0.1111935, Accumulated Leap Seconds=31
  { 2451025.50, 63.295670 }, //31 July 1998, UT1-UTC=-0.1116704, Accumulated Leap Seconds=31
  { 2451026.50, 63.296137 }, //01 August 1998, UT1-UTC=-0.1121369, Accumulated Leap Seconds=31
  { 2451027.50, 63.296552 }, //02 August 1998, UT1-UTC=-0.1125522, Accumulated Leap Seconds=31
  { 2451028.50, 63.296918 }, //03 August 1998, UT1-UTC=-0.1129176, Accumulated Leap Seconds=31
  { 2451029.50, 63.297247 }, //04 August 1998, UT1-UTC=-0.1132469, Accumulated Leap Seconds=31
  { 2451030.50, 63.297560 }, //05 August 1998, UT1-UTC=-0.1135605, Accumulated Leap Seconds=31
  { 2451031.50, 63.297884 }, //06 August 1998, UT1-UTC=-0.1138838, Accumulated Leap Seconds=31
  { 2451032.50, 63.298274 }, //07 August 1998, UT1-UTC=-0.1142743, Accumulated Leap Seconds=31
  { 2451033.50, 63.298795 }, //08 August 1998, UT1-UTC=-0.1147954, Accumulated Leap Seconds=31
  { 2451034.50, 63.299484 }, //09 August 1998, UT1-UTC=-0.1154840, Accumulated Leap Seconds=31
  { 2451035.50, 63.300339 }, //10 August 1998, UT1-UTC=-0.1163391, Accumulated Leap Seconds=31
  { 2451036.50, 63.301309 }, //11 August 1998, UT1-UTC=-0.1173091, Accumulated Leap Seconds=31
  { 2451037.50, 63.302317 }, //12 August 1998, UT1-UTC=-0.1183170, Accumulated Leap Seconds=31
  { 2451038.50, 63.303287 }, //13 August 1998, UT1-UTC=-0.1192872, Accumulated Leap Seconds=31
  { 2451039.50, 63.304133 }, //14 August 1998, UT1-UTC=-0.1201329, Accumulated Leap Seconds=31
  { 2451040.50, 63.304800 }, //15 August 1998, UT1-UTC=-0.1207999, Accumulated Leap Seconds=31
  { 2451041.50, 63.305276 }, //16 August 1998, UT1-UTC=-0.1212755, Accumulated Leap Seconds=31
  { 2451042.50, 63.305593 }, //17 August 1998, UT1-UTC=-0.1215926, Accumulated Leap Seconds=31
  { 2451043.50, 63.305827 }, //18 August 1998, UT1-UTC=-0.1218266, Accumulated Leap Seconds=31
  { 2451044.50, 63.306036 }, //19 August 1998, UT1-UTC=-0.1220355, Accumulated Leap Seconds=31
  { 2451045.50, 63.306237 }, //20 August 1998, UT1-UTC=-0.1222367, Accumulated Leap Seconds=31
  { 2451046.50, 63.306495 }, //21 August 1998, UT1-UTC=-0.1224955, Accumulated Leap Seconds=31
  { 2451047.50, 63.306859 }, //22 August 1998, UT1-UTC=-0.1228588, Accumulated Leap Seconds=31
  { 2451048.50, 63.307315 }, //23 August 1998, UT1-UTC=-0.1233150, Accumulated Leap Seconds=31
  { 2451049.50, 63.307841 }, //24 August 1998, UT1-UTC=-0.1238411, Accumulated Leap Seconds=31
  { 2451050.50, 63.308424 }, //25 August 1998, UT1-UTC=-0.1244240, Accumulated Leap Seconds=31
  { 2451051.50, 63.309062 }, //26 August 1998, UT1-UTC=-0.1250624, Accumulated Leap Seconds=31
  { 2451052.50, 63.309742 }, //27 August 1998, UT1-UTC=-0.1257425, Accumulated Leap Seconds=31
  { 2451053.50, 63.310364 }, //28 August 1998, UT1-UTC=-0.1263636, Accumulated Leap Seconds=31
  { 2451054.50, 63.310921 }, //29 August 1998, UT1-UTC=-0.1269212, Accumulated Leap Seconds=31
  { 2451055.50, 63.311468 }, //30 August 1998, UT1-UTC=-0.1274676, Accumulated Leap Seconds=31
  { 2451056.50, 63.312027 }, //31 August 1998, UT1-UTC=-0.1280275, Accumulated Leap Seconds=31
  { 2451057.50, 63.312609 }, //01 September 1998, UT1-UTC=-0.1286092, Accumulated Leap Seconds=31
  { 2451058.50, 63.313234 }, //02 September 1998, UT1-UTC=-0.1292343, Accumulated Leap Seconds=31
  { 2451059.50, 63.313924 }, //03 September 1998, UT1-UTC=-0.1299243, Accumulated Leap Seconds=31
  { 2451060.50, 63.314722 }, //04 September 1998, UT1-UTC=-0.1307220, Accumulated Leap Seconds=31
  { 2451061.50, 63.315683 }, //05 September 1998, UT1-UTC=-0.1316830, Accumulated Leap Seconds=31
  { 2451062.50, 63.316837 }, //06 September 1998, UT1-UTC=-0.1328370, Accumulated Leap Seconds=31
  { 2451063.50, 63.318150 }, //07 September 1998, UT1-UTC=-0.1341501, Accumulated Leap Seconds=31
  { 2451064.50, 63.319573 }, //08 September 1998, UT1-UTC=-0.1355729, Accumulated Leap Seconds=31
  { 2451065.50, 63.321032 }, //09 September 1998, UT1-UTC=-0.1370316, Accumulated Leap Seconds=31
  { 2451066.50, 63.322443 }, //10 September 1998, UT1-UTC=-0.1384432, Accumulated Leap Seconds=31
  { 2451067.50, 63.323702 }, //11 September 1998, UT1-UTC=-0.1397025, Accumulated Leap Seconds=31
  { 2451068.50, 63.324756 }, //12 September 1998, UT1-UTC=-0.1407564, Accumulated Leap Seconds=31
  { 2451069.50, 63.325625 }, //13 September 1998, UT1-UTC=-0.1416251, Accumulated Leap Seconds=31
  { 2451070.50, 63.326353 }, //14 September 1998, UT1-UTC=-0.1423535, Accumulated Leap Seconds=31
  { 2451071.50, 63.327036 }, //15 September 1998, UT1-UTC=-0.1430362, Accumulated Leap Seconds=31
  { 2451072.50, 63.327748 }, //16 September 1998, UT1-UTC=-0.1437484, Accumulated Leap Seconds=31
  { 2451073.50, 63.328525 }, //17 September 1998, UT1-UTC=-0.1445251, Accumulated Leap Seconds=31
  { 2451074.50, 63.329383 }, //18 September 1998, UT1-UTC=-0.1453833, Accumulated Leap Seconds=31
  { 2451075.50, 63.330344 }, //19 September 1998, UT1-UTC=-0.1463441, Accumulated Leap Seconds=31
  { 2451076.50, 63.331409 }, //20 September 1998, UT1-UTC=-0.1474095, Accumulated Leap Seconds=31
  { 2451077.50, 63.332534 }, //21 September 1998, UT1-UTC=-0.1485342, Accumulated Leap Seconds=31
  { 2451078.50, 63.333673 }, //22 September 1998, UT1-UTC=-0.1496726, Accumulated Leap Seconds=31
  { 2451079.50, 63.334780 }, //23 September 1998, UT1-UTC=-0.1507802, Accumulated Leap Seconds=31
  { 2451080.50, 63.335829 }, //24 September 1998, UT1-UTC=-0.1518292, Accumulated Leap Seconds=31
  { 2451081.50, 63.336826 }, //25 September 1998, UT1-UTC=-0.1528258, Accumulated Leap Seconds=31
  { 2451082.50, 63.337774 }, //26 September 1998, UT1-UTC=-0.1537743, Accumulated Leap Seconds=31
  { 2451083.50, 63.338670 }, //27 September 1998, UT1-UTC=-0.1546698, Accumulated Leap Seconds=31
  { 2451084.50, 63.339505 }, //28 September 1998, UT1-UTC=-0.1555055, Accumulated Leap Seconds=31
  { 2451085.50, 63.340318 }, //29 September 1998, UT1-UTC=-0.1563181, Accumulated Leap Seconds=31
  { 2451086.50, 63.341172 }, //30 September 1998, UT1-UTC=-0.1571720, Accumulated Leap Seconds=31
  { 2451087.50, 63.342162 }, //01 October 1998, UT1-UTC=-0.1581622, Accumulated Leap Seconds=31
  { 2451088.50, 63.343346 }, //02 October 1998, UT1-UTC=-0.1593458, Accumulated Leap Seconds=31
  { 2451089.50, 63.344750 }, //03 October 1998, UT1-UTC=-0.1607503, Accumulated Leap Seconds=31
  { 2451090.50, 63.346382 }, //04 October 1998, UT1-UTC=-0.1623819, Accumulated Leap Seconds=31
  { 2451091.50, 63.348199 }, //05 October 1998, UT1-UTC=-0.1641995, Accumulated Leap Seconds=31
  { 2451092.50, 63.350100 }, //06 October 1998, UT1-UTC=-0.1661005, Accumulated Leap Seconds=31
  { 2451093.50, 63.351977 }, //07 October 1998, UT1-UTC=-0.1679768, Accumulated Leap Seconds=31
  { 2451094.50, 63.353772 }, //08 October 1998, UT1-UTC=-0.1697717, Accumulated Leap Seconds=31
  { 2451095.50, 63.355430 }, //09 October 1998, UT1-UTC=-0.1714298, Accumulated Leap Seconds=31
  { 2451096.50, 63.356952 }, //10 October 1998, UT1-UTC=-0.1729521, Accumulated Leap Seconds=31
  { 2451097.50, 63.358344 }, //11 October 1998, UT1-UTC=-0.1743443, Accumulated Leap Seconds=31
  { 2451098.50, 63.359627 }, //12 October 1998, UT1-UTC=-0.1756272, Accumulated Leap Seconds=31
  { 2451099.50, 63.360859 }, //13 October 1998, UT1-UTC=-0.1768590, Accumulated Leap Seconds=31
  { 2451100.50, 63.362108 }, //14 October 1998, UT1-UTC=-0.1781084, Accumulated Leap Seconds=31
  { 2451101.50, 63.363440 }, //15 October 1998, UT1-UTC=-0.1794396, Accumulated Leap Seconds=31
  { 2451102.50, 63.364827 }, //16 October 1998, UT1-UTC=-0.1808271, Accumulated Leap Seconds=31
  { 2451103.50, 63.366244 }, //17 October 1998, UT1-UTC=-0.1822437, Accumulated Leap Seconds=31
  { 2451104.50, 63.367689 }, //18 October 1998, UT1-UTC=-0.1836888, Accumulated Leap Seconds=31
  { 2451105.50, 63.369154 }, //19 October 1998, UT1-UTC=-0.1851536, Accumulated Leap Seconds=31
  { 2451106.50, 63.370616 }, //20 October 1998, UT1-UTC=-0.1866161, Accumulated Leap Seconds=31
  { 2451107.50, 63.372065 }, //21 October 1998, UT1-UTC=-0.1880647, Accumulated Leap Seconds=31
  { 2451108.50, 63.373479 }, //22 October 1998, UT1-UTC=-0.1894793, Accumulated Leap Seconds=31
  { 2451109.50, 63.374814 }, //23 October 1998, UT1-UTC=-0.1908137, Accumulated Leap Seconds=31
  { 2451110.50, 63.376073 }, //24 October 1998, UT1-UTC=-0.1920735, Accumulated Leap Seconds=31
  { 2451111.50, 63.377303 }, //25 October 1998, UT1-UTC=-0.1933035, Accumulated Leap Seconds=31
  { 2451112.50, 63.378523 }, //26 October 1998, UT1-UTC=-0.1945232, Accumulated Leap Seconds=31
  { 2451113.50, 63.379716 }, //27 October 1998, UT1-UTC=-0.1957162, Accumulated Leap Seconds=31
  { 2451114.50, 63.380901 }, //28 October 1998, UT1-UTC=-0.1969014, Accumulated Leap Seconds=31
  { 2451115.50, 63.382160 }, //29 October 1998, UT1-UTC=-0.1981601, Accumulated Leap Seconds=31
  { 2451116.50, 63.383590 }, //30 October 1998, UT1-UTC=-0.1995900, Accumulated Leap Seconds=31
  { 2451117.50, 63.385245 }, //31 October 1998, UT1-UTC=-0.2012454, Accumulated Leap Seconds=31
  { 2451118.50, 63.387130 }, //01 November 1998, UT1-UTC=-0.2031303, Accumulated Leap Seconds=31
  { 2451119.50, 63.389193 }, //02 November 1998, UT1-UTC=-0.2051926, Accumulated Leap Seconds=31
  { 2451120.50, 63.391325 }, //03 November 1998, UT1-UTC=-0.2073251, Accumulated Leap Seconds=31
  { 2451121.50, 63.393409 }, //04 November 1998, UT1-UTC=-0.2094093, Accumulated Leap Seconds=31
  { 2451122.50, 63.395365 }, //05 November 1998, UT1-UTC=-0.2113650, Accumulated Leap Seconds=31
  { 2451123.50, 63.397141 }, //06 November 1998, UT1-UTC=-0.2131407, Accumulated Leap Seconds=31
  { 2451124.50, 63.398755 }, //07 November 1998, UT1-UTC=-0.2147551, Accumulated Leap Seconds=31
  { 2451125.50, 63.400249 }, //08 November 1998, UT1-UTC=-0.2162487, Accumulated Leap Seconds=31
  { 2451126.50, 63.401682 }, //09 November 1998, UT1-UTC=-0.2176819, Accumulated Leap Seconds=31
  { 2451127.50, 63.403120 }, //10 November 1998, UT1-UTC=-0.2191202, Accumulated Leap Seconds=31
  { 2451128.50, 63.404591 }, //11 November 1998, UT1-UTC=-0.2205909, Accumulated Leap Seconds=31
  { 2451129.50, 63.406090 }, //12 November 1998, UT1-UTC=-0.2220896, Accumulated Leap Seconds=31
  { 2451130.50, 63.407608 }, //13 November 1998, UT1-UTC=-0.2236081, Accumulated Leap Seconds=31
  { 2451131.50, 63.409132 }, //14 November 1998, UT1-UTC=-0.2251319, Accumulated Leap Seconds=31
  { 2451132.50, 63.410641 }, //15 November 1998, UT1-UTC=-0.2266414, Accumulated Leap Seconds=31
  { 2451133.50, 63.412114 }, //16 November 1998, UT1-UTC=-0.2281143, Accumulated Leap Seconds=31
  { 2451134.50, 63.413545 }, //17 November 1998, UT1-UTC=-0.2295452, Accumulated Leap Seconds=31
  { 2451135.50, 63.414934 }, //18 November 1998, UT1-UTC=-0.2309338, Accumulated Leap Seconds=31
  { 2451136.50, 63.416280 }, //19 November 1998, UT1-UTC=-0.2322805, Accumulated Leap Seconds=31
  { 2451137.50, 63.417578 }, //20 November 1998, UT1-UTC=-0.2335782, Accumulated Leap Seconds=31
  { 2451138.50, 63.418834 }, //21 November 1998, UT1-UTC=-0.2348337, Accumulated Leap Seconds=31
  { 2451139.50, 63.420062 }, //22 November 1998, UT1-UTC=-0.2360622, Accumulated Leap Seconds=31
  { 2451140.50, 63.421288 }, //23 November 1998, UT1-UTC=-0.2372881, Accumulated Leap Seconds=31
  { 2451141.50, 63.422533 }, //24 November 1998, UT1-UTC=-0.2385329, Accumulated Leap Seconds=31
  { 2451142.50, 63.423819 }, //25 November 1998, UT1-UTC=-0.2398188, Accumulated Leap Seconds=31
  { 2451143.50, 63.425204 }, //26 November 1998, UT1-UTC=-0.2412040, Accumulated Leap Seconds=31
  { 2451144.50, 63.426740 }, //27 November 1998, UT1-UTC=-0.2427405, Accumulated Leap Seconds=31
  { 2451145.50, 63.428440 }, //28 November 1998, UT1-UTC=-0.2444404, Accumulated Leap Seconds=31
  { 2451146.50, 63.430269 }, //29 November 1998, UT1-UTC=-0.2462693, Accumulated Leap Seconds=31
  { 2451147.50, 63.432132 }, //30 November 1998, UT1-UTC=-0.2481323, Accumulated Leap Seconds=31
  { 2451148.50, 63.433930 }, //01 December 1998, UT1-UTC=-0.2499302, Accumulated Leap Seconds=31
  { 2451149.50, 63.435594 }, //02 December 1998, UT1-UTC=-0.2515939, Accumulated Leap Seconds=31
  { 2451150.50, 63.437116 }, //03 December 1998, UT1-UTC=-0.2531158, Accumulated Leap Seconds=31
  { 2451151.50, 63.438480 }, //04 December 1998, UT1-UTC=-0.2544798, Accumulated Leap Seconds=31
  { 2451152.50, 63.439707 }, //05 December 1998, UT1-UTC=-0.2557069, Accumulated Leap Seconds=31
  { 2451153.50, 63.440849 }, //06 December 1998, UT1-UTC=-0.2568494, Accumulated Leap Seconds=31
  { 2451154.50, 63.441950 }, //07 December 1998, UT1-UTC=-0.2579503, Accumulated Leap Seconds=31
  { 2451155.50, 63.443032 }, //08 December 1998, UT1-UTC=-0.2590320, Accumulated Leap Seconds=31
  { 2451156.50, 63.444105 }, //09 December 1998, UT1-UTC=-0.2601053, Accumulated Leap Seconds=31
  { 2451157.50, 63.445218 }, //10 December 1998, UT1-UTC=-0.2612184, Accumulated Leap Seconds=31
  { 2451158.50, 63.446391 }, //11 December 1998, UT1-UTC=-0.2623909, Accumulated Leap Seconds=31
  { 2451159.50, 63.447583 }, //12 December 1998, UT1-UTC=-0.2635827, Accumulated Leap Seconds=31
  { 2451160.50, 63.448736 }, //13 December 1998, UT1-UTC=-0.2647357, Accumulated Leap Seconds=31
  { 2451161.50, 63.449827 }, //14 December 1998, UT1-UTC=-0.2658275, Accumulated Leap Seconds=31
  { 2451162.50, 63.450846 }, //15 December 1998, UT1-UTC=-0.2668465, Accumulated Leap Seconds=31
  { 2451163.50, 63.451779 }, //16 December 1998, UT1-UTC=-0.2677791, Accumulated Leap Seconds=31
  { 2451164.50, 63.452616 }, //17 December 1998, UT1-UTC=-0.2686158, Accumulated Leap Seconds=31
  { 2451165.50, 63.453372 }, //18 December 1998, UT1-UTC=-0.2693720, Accumulated Leap Seconds=31
  { 2451166.50, 63.454059 }, //19 December 1998, UT1-UTC=-0.2700595, Accumulated Leap Seconds=31
  { 2451167.50, 63.454701 }, //20 December 1998, UT1-UTC=-0.2707009, Accumulated Leap Seconds=31
  { 2451168.50, 63.455336 }, //21 December 1998, UT1-UTC=-0.2713359, Accumulated Leap Seconds=31
  { 2451169.50, 63.456001 }, //22 December 1998, UT1-UTC=-0.2720006, Accumulated Leap Seconds=31
  { 2451170.50, 63.456744 }, //23 December 1998, UT1-UTC=-0.2727442, Accumulated Leap Seconds=31
  { 2451171.50, 63.457650 }, //24 December 1998, UT1-UTC=-0.2736499, Accumulated Leap Seconds=31
  { 2451172.50, 63.458750 }, //25 December 1998, UT1-UTC=-0.2747502, Accumulated Leap Seconds=31
  { 2451173.50, 63.460001 }, //26 December 1998, UT1-UTC=-0.2760006, Accumulated Leap Seconds=31
  { 2451174.50, 63.461331 }, //27 December 1998, UT1-UTC=-0.2773312, Accumulated Leap Seconds=31
  { 2451175.50, 63.462675 }, //28 December 1998, UT1-UTC=-0.2786753, Accumulated Leap Seconds=31
  { 2451176.50, 63.463977 }, //29 December 1998, UT1-UTC=-0.2799775, Accumulated Leap Seconds=31
  { 2451177.50, 63.465210 }, //30 December 1998, UT1-UTC=-0.2812099, Accumulated Leap Seconds=31
  { 2451178.50, 63.466334 }, //31 December 1998, UT1-UTC=-0.2823341, Accumulated Leap Seconds=31
  { 2451179.50, 63.467337 }, //01 January 1999, UT1-UTC= 0.7166631, Accumulated Leap Seconds=32
  { 2451180.50, 63.468246 }, //02 January 1999, UT1-UTC= 0.7157538, Accumulated Leap Seconds=32
  { 2451181.50, 63.469107 }, //03 January 1999, UT1-UTC= 0.7148929, Accumulated Leap Seconds=32
  { 2451182.50, 63.469958 }, //04 January 1999, UT1-UTC= 0.7140423, Accumulated Leap Seconds=32
  { 2451183.50, 63.470818 }, //05 January 1999, UT1-UTC= 0.7131819, Accumulated Leap Seconds=32
  { 2451184.50, 63.471720 }, //06 January 1999, UT1-UTC= 0.7122795, Accumulated Leap Seconds=32
  { 2451185.50, 63.472735 }, //07 January 1999, UT1-UTC= 0.7112652, Accumulated Leap Seconds=32
  { 2451186.50, 63.473816 }, //08 January 1999, UT1-UTC= 0.7101837, Accumulated Leap Seconds=32
  { 2451187.50, 63.474895 }, //09 January 1999, UT1-UTC= 0.7091047, Accumulated Leap Seconds=32
  { 2451188.50, 63.475965 }, //10 January 1999, UT1-UTC= 0.7080354, Accumulated Leap Seconds=32
  { 2451189.50, 63.477006 }, //11 January 1999, UT1-UTC= 0.7069937, Accumulated Leap Seconds=32
  { 2451190.50, 63.477952 }, //12 January 1999, UT1-UTC= 0.7060484, Accumulated Leap Seconds=32
  { 2451191.50, 63.478765 }, //13 January 1999, UT1-UTC= 0.7052346, Accumulated Leap Seconds=32
  { 2451192.50, 63.479500 }, //14 January 1999, UT1-UTC= 0.7045002, Accumulated Leap Seconds=32
  { 2451193.50, 63.480182 }, //15 January 1999, UT1-UTC= 0.7038183, Accumulated Leap Seconds=32
  { 2451194.50, 63.480792 }, //16 January 1999, UT1-UTC= 0.7032077, Accumulated Leap Seconds=32
  { 2451195.50, 63.481348 }, //17 January 1999, UT1-UTC= 0.7026524, Accumulated Leap Seconds=32
  { 2451196.50, 63.481893 }, //18 January 1999, UT1-UTC= 0.7021070, Accumulated Leap Seconds=32
  { 2451197.50, 63.482509 }, //19 January 1999, UT1-UTC= 0.7014909, Accumulated Leap Seconds=32
  { 2451198.50, 63.483263 }, //20 January 1999, UT1-UTC= 0.7007374, Accumulated Leap Seconds=32
  { 2451199.50, 63.484203 }, //21 January 1999, UT1-UTC= 0.6997973, Accumulated Leap Seconds=32
  { 2451200.50, 63.485328 }, //22 January 1999, UT1-UTC= 0.6986720, Accumulated Leap Seconds=32
  { 2451201.50, 63.486597 }, //23 January 1999, UT1-UTC= 0.6974028, Accumulated Leap Seconds=32
  { 2451202.50, 63.487965 }, //24 January 1999, UT1-UTC= 0.6960345, Accumulated Leap Seconds=32
  { 2451203.50, 63.489375 }, //25 January 1999, UT1-UTC= 0.6946253, Accumulated Leap Seconds=32
  { 2451204.50, 63.490726 }, //26 January 1999, UT1-UTC= 0.6932741, Accumulated Leap Seconds=32
  { 2451205.50, 63.491975 }, //27 January 1999, UT1-UTC= 0.6920250, Accumulated Leap Seconds=32
  { 2451206.50, 63.493158 }, //28 January 1999, UT1-UTC= 0.6908424, Accumulated Leap Seconds=32
  { 2451207.50, 63.494301 }, //29 January 1999, UT1-UTC= 0.6896986, Accumulated Leap Seconds=32
  { 2451208.50, 63.495455 }, //30 January 1999, UT1-UTC= 0.6885450, Accumulated Leap Seconds=32
  { 2451209.50, 63.496638 }, //31 January 1999, UT1-UTC= 0.6873616, Accumulated Leap Seconds=32
  { 2451210.50, 63.497864 }, //01 February 1999, UT1-UTC= 0.6861358, Accumulated Leap Seconds=32
  { 2451211.50, 63.499136 }, //02 February 1999, UT1-UTC= 0.6848638, Accumulated Leap Seconds=32
  { 2451212.50, 63.500445 }, //03 February 1999, UT1-UTC= 0.6835548, Accumulated Leap Seconds=32
  { 2451213.50, 63.501774 }, //04 February 1999, UT1-UTC= 0.6822259, Accumulated Leap Seconds=32
  { 2451214.50, 63.503107 }, //05 February 1999, UT1-UTC= 0.6808925, Accumulated Leap Seconds=32
  { 2451215.50, 63.504429 }, //06 February 1999, UT1-UTC= 0.6795712, Accumulated Leap Seconds=32
  { 2451216.50, 63.505712 }, //07 February 1999, UT1-UTC= 0.6782882, Accumulated Leap Seconds=32
  { 2451217.50, 63.506924 }, //08 February 1999, UT1-UTC= 0.6770761, Accumulated Leap Seconds=32
  { 2451218.50, 63.508045 }, //09 February 1999, UT1-UTC= 0.6759552, Accumulated Leap Seconds=32
  { 2451219.50, 63.509088 }, //10 February 1999, UT1-UTC= 0.6749119, Accumulated Leap Seconds=32
  { 2451220.50, 63.510057 }, //11 February 1999, UT1-UTC= 0.6739428, Accumulated Leap Seconds=32
  { 2451221.50, 63.510964 }, //12 February 1999, UT1-UTC= 0.6730364, Accumulated Leap Seconds=32
  { 2451222.50, 63.511861 }, //13 February 1999, UT1-UTC= 0.6721393, Accumulated Leap Seconds=32
  { 2451223.50, 63.512820 }, //14 February 1999, UT1-UTC= 0.6711804, Accumulated Leap Seconds=32
  { 2451224.50, 63.513886 }, //15 February 1999, UT1-UTC= 0.6701140, Accumulated Leap Seconds=32
  { 2451225.50, 63.515082 }, //16 February 1999, UT1-UTC= 0.6689183, Accumulated Leap Seconds=32
  { 2451226.50, 63.516426 }, //17 February 1999, UT1-UTC= 0.6675736, Accumulated Leap Seconds=32
  { 2451227.50, 63.517948 }, //18 February 1999, UT1-UTC= 0.6660522, Accumulated Leap Seconds=32
  { 2451228.50, 63.519606 }, //19 February 1999, UT1-UTC= 0.6643942, Accumulated Leap Seconds=32
  { 2451229.50, 63.521303 }, //20 February 1999, UT1-UTC= 0.6626968, Accumulated Leap Seconds=32
  { 2451230.50, 63.522959 }, //21 February 1999, UT1-UTC= 0.6610413, Accumulated Leap Seconds=32
  { 2451231.50, 63.524527 }, //22 February 1999, UT1-UTC= 0.6594725, Accumulated Leap Seconds=32
  { 2451232.50, 63.525949 }, //23 February 1999, UT1-UTC= 0.6580508, Accumulated Leap Seconds=32
  { 2451233.50, 63.527189 }, //24 February 1999, UT1-UTC= 0.6568111, Accumulated Leap Seconds=32
  { 2451234.50, 63.528269 }, //25 February 1999, UT1-UTC= 0.6557312, Accumulated Leap Seconds=32
  { 2451235.50, 63.529232 }, //26 February 1999, UT1-UTC= 0.6547679, Accumulated Leap Seconds=32
  { 2451236.50, 63.530125 }, //27 February 1999, UT1-UTC= 0.6538752, Accumulated Leap Seconds=32
  { 2451237.50, 63.531012 }, //28 February 1999, UT1-UTC= 0.6529877, Accumulated Leap Seconds=32
  { 2451238.50, 63.531933 }, //01 March 1999, UT1-UTC= 0.6520673, Accumulated Leap Seconds=32
  { 2451239.50, 63.532875 }, //02 March 1999, UT1-UTC= 0.6511253, Accumulated Leap Seconds=32
  { 2451240.50, 63.533825 }, //03 March 1999, UT1-UTC= 0.6501754, Accumulated Leap Seconds=32
  { 2451241.50, 63.534802 }, //04 March 1999, UT1-UTC= 0.6491977, Accumulated Leap Seconds=32
  { 2451242.50, 63.535816 }, //05 March 1999, UT1-UTC= 0.6481838, Accumulated Leap Seconds=32
  { 2451243.50, 63.536826 }, //06 March 1999, UT1-UTC= 0.6471735, Accumulated Leap Seconds=32
  { 2451244.50, 63.537742 }, //07 March 1999, UT1-UTC= 0.6462583, Accumulated Leap Seconds=32
  { 2451245.50, 63.538571 }, //08 March 1999, UT1-UTC= 0.6454292, Accumulated Leap Seconds=32
  { 2451246.50, 63.539332 }, //09 March 1999, UT1-UTC= 0.6446676, Accumulated Leap Seconds=32
  { 2451247.50, 63.540043 }, //10 March 1999, UT1-UTC= 0.6439574, Accumulated Leap Seconds=32
  { 2451248.50, 63.540735 }, //11 March 1999, UT1-UTC= 0.6432645, Accumulated Leap Seconds=32
  { 2451249.50, 63.541451 }, //12 March 1999, UT1-UTC= 0.6425490, Accumulated Leap Seconds=32
  { 2451250.50, 63.542235 }, //13 March 1999, UT1-UTC= 0.6417652, Accumulated Leap Seconds=32
  { 2451251.50, 63.543136 }, //14 March 1999, UT1-UTC= 0.6408641, Accumulated Leap Seconds=32
  { 2451252.50, 63.544199 }, //15 March 1999, UT1-UTC= 0.6398005, Accumulated Leap Seconds=32
  { 2451253.50, 63.545439 }, //16 March 1999, UT1-UTC= 0.6385608, Accumulated Leap Seconds=32
  { 2451254.50, 63.546840 }, //17 March 1999, UT1-UTC= 0.6371604, Accumulated Leap Seconds=32
  { 2451255.50, 63.548372 }, //18 March 1999, UT1-UTC= 0.6356283, Accumulated Leap Seconds=32
  { 2451256.50, 63.549993 }, //19 March 1999, UT1-UTC= 0.6340067, Accumulated Leap Seconds=32
  { 2451257.50, 63.551655 }, //20 March 1999, UT1-UTC= 0.6323448, Accumulated Leap Seconds=32
  { 2451258.50, 63.553317 }, //21 March 1999, UT1-UTC= 0.6306827, Accumulated Leap Seconds=32
  { 2451259.50, 63.554918 }, //22 March 1999, UT1-UTC= 0.6290823, Accumulated Leap Seconds=32
  { 2451260.50, 63.556379 }, //23 March 1999, UT1-UTC= 0.6276210, Accumulated Leap Seconds=32
  { 2451261.50, 63.557682 }, //24 March 1999, UT1-UTC= 0.6263176, Accumulated Leap Seconds=32
  { 2451262.50, 63.558885 }, //25 March 1999, UT1-UTC= 0.6251154, Accumulated Leap Seconds=32
  { 2451263.50, 63.560034 }, //26 March 1999, UT1-UTC= 0.6239661, Accumulated Leap Seconds=32
  { 2451264.50, 63.561210 }, //27 March 1999, UT1-UTC= 0.6227899, Accumulated Leap Seconds=32
  { 2451265.50, 63.562452 }, //28 March 1999, UT1-UTC= 0.6215475, Accumulated Leap Seconds=32
  { 2451266.50, 63.563773 }, //29 March 1999, UT1-UTC= 0.6202271, Accumulated Leap Seconds=32
  { 2451267.50, 63.565145 }, //30 March 1999, UT1-UTC= 0.6188549, Accumulated Leap Seconds=32
  { 2451268.50, 63.566534 }, //31 March 1999, UT1-UTC= 0.6174657, Accumulated Leap Seconds=32
  { 2451269.50, 63.567944 }, //01 April 1999, UT1-UTC= 0.6160559, Accumulated Leap Seconds=32
  { 2451270.50, 63.569364 }, //02 April 1999, UT1-UTC= 0.6146358, Accumulated Leap Seconds=32
  { 2451271.50, 63.570751 }, //03 April 1999, UT1-UTC= 0.6132488, Accumulated Leap Seconds=32
  { 2451272.50, 63.572056 }, //04 April 1999, UT1-UTC= 0.6119436, Accumulated Leap Seconds=32
  { 2451273.50, 63.573280 }, //05 April 1999, UT1-UTC= 0.6107195, Accumulated Leap Seconds=32
  { 2451274.50, 63.574450 }, //06 April 1999, UT1-UTC= 0.6095502, Accumulated Leap Seconds=32
  { 2451275.50, 63.575594 }, //07 April 1999, UT1-UTC= 0.6084063, Accumulated Leap Seconds=32
  { 2451276.50, 63.576698 }, //08 April 1999, UT1-UTC= 0.6073019, Accumulated Leap Seconds=32
  { 2451277.50, 63.577784 }, //09 April 1999, UT1-UTC= 0.6062160, Accumulated Leap Seconds=32
  { 2451278.50, 63.578921 }, //10 April 1999, UT1-UTC= 0.6050794, Accumulated Leap Seconds=32
  { 2451279.50, 63.580162 }, //11 April 1999, UT1-UTC= 0.6038380, Accumulated Leap Seconds=32
  { 2451280.50, 63.581563 }, //12 April 1999, UT1-UTC= 0.6024367, Accumulated Leap Seconds=32
  { 2451281.50, 63.583169 }, //13 April 1999, UT1-UTC= 0.6008313, Accumulated Leap Seconds=32
  { 2451282.50, 63.584967 }, //14 April 1999, UT1-UTC= 0.5990325, Accumulated Leap Seconds=32
  { 2451283.50, 63.586873 }, //15 April 1999, UT1-UTC= 0.5971267, Accumulated Leap Seconds=32
  { 2451284.50, 63.588811 }, //16 April 1999, UT1-UTC= 0.5951894, Accumulated Leap Seconds=32
  { 2451285.50, 63.590699 }, //17 April 1999, UT1-UTC= 0.5933008, Accumulated Leap Seconds=32
  { 2451286.50, 63.592457 }, //18 April 1999, UT1-UTC= 0.5915425, Accumulated Leap Seconds=32
  { 2451287.50, 63.594063 }, //19 April 1999, UT1-UTC= 0.5899368, Accumulated Leap Seconds=32
  { 2451288.50, 63.595520 }, //20 April 1999, UT1-UTC= 0.5884796, Accumulated Leap Seconds=32
  { 2451289.50, 63.596864 }, //21 April 1999, UT1-UTC= 0.5871357, Accumulated Leap Seconds=32
  { 2451290.50, 63.598130 }, //22 April 1999, UT1-UTC= 0.5858700, Accumulated Leap Seconds=32
  { 2451291.50, 63.599348 }, //23 April 1999, UT1-UTC= 0.5846524, Accumulated Leap Seconds=32
  { 2451292.50, 63.600590 }, //24 April 1999, UT1-UTC= 0.5834096, Accumulated Leap Seconds=32
  { 2451293.50, 63.601896 }, //25 April 1999, UT1-UTC= 0.5821038, Accumulated Leap Seconds=32
  { 2451294.50, 63.603273 }, //26 April 1999, UT1-UTC= 0.5807272, Accumulated Leap Seconds=32
  { 2451295.50, 63.604712 }, //27 April 1999, UT1-UTC= 0.5792883, Accumulated Leap Seconds=32
  { 2451296.50, 63.606185 }, //28 April 1999, UT1-UTC= 0.5778153, Accumulated Leap Seconds=32
  { 2451297.50, 63.607657 }, //29 April 1999, UT1-UTC= 0.5763425, Accumulated Leap Seconds=32
  { 2451298.50, 63.609089 }, //30 April 1999, UT1-UTC= 0.5749110, Accumulated Leap Seconds=32
  { 2451299.50, 63.610443 }, //01 May 1999, UT1-UTC= 0.5735568, Accumulated Leap Seconds=32
  { 2451300.50, 63.611699 }, //02 May 1999, UT1-UTC= 0.5723009, Accumulated Leap Seconds=32
  { 2451301.50, 63.612839 }, //03 May 1999, UT1-UTC= 0.5711608, Accumulated Leap Seconds=32
  { 2451302.50, 63.613847 }, //04 May 1999, UT1-UTC= 0.5701531, Accumulated Leap Seconds=32
  { 2451303.50, 63.614715 }, //05 May 1999, UT1-UTC= 0.5692849, Accumulated Leap Seconds=32
  { 2451304.50, 63.615475 }, //06 May 1999, UT1-UTC= 0.5685253, Accumulated Leap Seconds=32
  { 2451305.50, 63.616198 }, //07 May 1999, UT1-UTC= 0.5678023, Accumulated Leap Seconds=32
  { 2451306.50, 63.616976 }, //08 May 1999, UT1-UTC= 0.5670242, Accumulated Leap Seconds=32
  { 2451307.50, 63.617874 }, //09 May 1999, UT1-UTC= 0.5661260, Accumulated Leap Seconds=32
  { 2451308.50, 63.618934 }, //10 May 1999, UT1-UTC= 0.5650662, Accumulated Leap Seconds=32
  { 2451309.50, 63.620162 }, //11 May 1999, UT1-UTC= 0.5638383, Accumulated Leap Seconds=32
  { 2451310.50, 63.621558 }, //12 May 1999, UT1-UTC= 0.5624417, Accumulated Leap Seconds=32
  { 2451311.50, 63.623098 }, //13 May 1999, UT1-UTC= 0.5609021, Accumulated Leap Seconds=32
  { 2451312.50, 63.624692 }, //14 May 1999, UT1-UTC= 0.5593077, Accumulated Leap Seconds=32
  { 2451313.50, 63.626229 }, //15 May 1999, UT1-UTC= 0.5577711, Accumulated Leap Seconds=32
  { 2451314.50, 63.627616 }, //16 May 1999, UT1-UTC= 0.5563838, Accumulated Leap Seconds=32
  { 2451315.50, 63.628822 }, //17 May 1999, UT1-UTC= 0.5551776, Accumulated Leap Seconds=32
  { 2451316.50, 63.629895 }, //18 May 1999, UT1-UTC= 0.5541047, Accumulated Leap Seconds=32
  { 2451317.50, 63.630923 }, //19 May 1999, UT1-UTC= 0.5530771, Accumulated Leap Seconds=32
  { 2451318.50, 63.631969 }, //20 May 1999, UT1-UTC= 0.5520312, Accumulated Leap Seconds=32
  { 2451319.50, 63.633067 }, //21 May 1999, UT1-UTC= 0.5509326, Accumulated Leap Seconds=32
  { 2451320.50, 63.634258 }, //22 May 1999, UT1-UTC= 0.5497423, Accumulated Leap Seconds=32
  { 2451321.50, 63.635525 }, //23 May 1999, UT1-UTC= 0.5484752, Accumulated Leap Seconds=32
  { 2451322.50, 63.636816 }, //24 May 1999, UT1-UTC= 0.5471842, Accumulated Leap Seconds=32
  { 2451323.50, 63.638117 }, //25 May 1999, UT1-UTC= 0.5458829, Accumulated Leap Seconds=32
  { 2451324.50, 63.639383 }, //26 May 1999, UT1-UTC= 0.5446169, Accumulated Leap Seconds=32
  { 2451325.50, 63.640521 }, //27 May 1999, UT1-UTC= 0.5434790, Accumulated Leap Seconds=32
  { 2451326.50, 63.641512 }, //28 May 1999, UT1-UTC= 0.5424877, Accumulated Leap Seconds=32
  { 2451327.50, 63.642379 }, //29 May 1999, UT1-UTC= 0.5416207, Accumulated Leap Seconds=32
  { 2451328.50, 63.643143 }, //30 May 1999, UT1-UTC= 0.5408573, Accumulated Leap Seconds=32
  { 2451329.50, 63.643821 }, //31 May 1999, UT1-UTC= 0.5401792, Accumulated Leap Seconds=32
  { 2451330.50, 63.644429 }, //01 June 1999, UT1-UTC= 0.5395709, Accumulated Leap Seconds=32
  { 2451331.50, 63.645001 }, //02 June 1999, UT1-UTC= 0.5389991, Accumulated Leap Seconds=32
  { 2451332.50, 63.645579 }, //03 June 1999, UT1-UTC= 0.5384210, Accumulated Leap Seconds=32
  { 2451333.50, 63.646203 }, //04 June 1999, UT1-UTC= 0.5377971, Accumulated Leap Seconds=32
  { 2451334.50, 63.646922 }, //05 June 1999, UT1-UTC= 0.5370776, Accumulated Leap Seconds=32
  { 2451335.50, 63.647776 }, //06 June 1999, UT1-UTC= 0.5362240, Accumulated Leap Seconds=32
  { 2451336.50, 63.648765 }, //07 June 1999, UT1-UTC= 0.5352345, Accumulated Leap Seconds=32
  { 2451337.50, 63.649912 }, //08 June 1999, UT1-UTC= 0.5340876, Accumulated Leap Seconds=32
  { 2451338.50, 63.651206 }, //09 June 1999, UT1-UTC= 0.5327944, Accumulated Leap Seconds=32
  { 2451339.50, 63.652579 }, //10 June 1999, UT1-UTC= 0.5314212, Accumulated Leap Seconds=32
  { 2451340.50, 63.653942 }, //11 June 1999, UT1-UTC= 0.5300582, Accumulated Leap Seconds=32
  { 2451341.50, 63.655196 }, //12 June 1999, UT1-UTC= 0.5288036, Accumulated Leap Seconds=32
  { 2451342.50, 63.656255 }, //13 June 1999, UT1-UTC= 0.5277451, Accumulated Leap Seconds=32
  { 2451343.50, 63.657116 }, //14 June 1999, UT1-UTC= 0.5268842, Accumulated Leap Seconds=32
  { 2451344.50, 63.657835 }, //15 June 1999, UT1-UTC= 0.5261647, Accumulated Leap Seconds=32
  { 2451345.50, 63.658474 }, //16 June 1999, UT1-UTC= 0.5255264, Accumulated Leap Seconds=32
  { 2451346.50, 63.659088 }, //17 June 1999, UT1-UTC= 0.5249123, Accumulated Leap Seconds=32
  { 2451347.50, 63.659744 }, //18 June 1999, UT1-UTC= 0.5242564, Accumulated Leap Seconds=32
  { 2451348.50, 63.660457 }, //19 June 1999, UT1-UTC= 0.5235431, Accumulated Leap Seconds=32
  { 2451349.50, 63.661201 }, //20 June 1999, UT1-UTC= 0.5227985, Accumulated Leap Seconds=32
  { 2451350.50, 63.661946 }, //21 June 1999, UT1-UTC= 0.5220543, Accumulated Leap Seconds=32
  { 2451351.50, 63.662666 }, //22 June 1999, UT1-UTC= 0.5213344, Accumulated Leap Seconds=32
  { 2451352.50, 63.663310 }, //23 June 1999, UT1-UTC= 0.5206902, Accumulated Leap Seconds=32
  { 2451353.50, 63.663813 }, //24 June 1999, UT1-UTC= 0.5201869, Accumulated Leap Seconds=32
  { 2451354.50, 63.664175 }, //25 June 1999, UT1-UTC= 0.5198245, Accumulated Leap Seconds=32
  { 2451355.50, 63.664408 }, //26 June 1999, UT1-UTC= 0.5195920, Accumulated Leap Seconds=32
  { 2451356.50, 63.664511 }, //27 June 1999, UT1-UTC= 0.5194892, Accumulated Leap Seconds=32
  { 2451357.50, 63.664507 }, //28 June 1999, UT1-UTC= 0.5194934, Accumulated Leap Seconds=32
  { 2451358.50, 63.664419 }, //29 June 1999, UT1-UTC= 0.5195805, Accumulated Leap Seconds=32
  { 2451359.50, 63.664296 }, //30 June 1999, UT1-UTC= 0.5197040, Accumulated Leap Seconds=32
  { 2451360.50, 63.664181 }, //01 July 1999, UT1-UTC= 0.5198185, Accumulated Leap Seconds=32
  { 2451361.50, 63.664113 }, //02 July 1999, UT1-UTC= 0.5198866, Accumulated Leap Seconds=32
  { 2451362.50, 63.664141 }, //03 July 1999, UT1-UTC= 0.5198591, Accumulated Leap Seconds=32
  { 2451363.50, 63.664300 }, //04 July 1999, UT1-UTC= 0.5196997, Accumulated Leap Seconds=32
  { 2451364.50, 63.664647 }, //05 July 1999, UT1-UTC= 0.5193533, Accumulated Leap Seconds=32
  { 2451365.50, 63.665166 }, //06 July 1999, UT1-UTC= 0.5188338, Accumulated Leap Seconds=32
  { 2451366.50, 63.665804 }, //07 July 1999, UT1-UTC= 0.5181964, Accumulated Leap Seconds=32
  { 2451367.50, 63.666468 }, //08 July 1999, UT1-UTC= 0.5175319, Accumulated Leap Seconds=32
  { 2451368.50, 63.667071 }, //09 July 1999, UT1-UTC= 0.5169290, Accumulated Leap Seconds=32
  { 2451369.50, 63.667561 }, //10 July 1999, UT1-UTC= 0.5164390, Accumulated Leap Seconds=32
  { 2451370.50, 63.667936 }, //11 July 1999, UT1-UTC= 0.5160635, Accumulated Leap Seconds=32
  { 2451371.50, 63.668234 }, //12 July 1999, UT1-UTC= 0.5157658, Accumulated Leap Seconds=32
  { 2451372.50, 63.668501 }, //13 July 1999, UT1-UTC= 0.5154992, Accumulated Leap Seconds=32
  { 2451373.50, 63.668789 }, //14 July 1999, UT1-UTC= 0.5152114, Accumulated Leap Seconds=32
  { 2451374.50, 63.669150 }, //15 July 1999, UT1-UTC= 0.5148502, Accumulated Leap Seconds=32
  { 2451375.50, 63.669590 }, //16 July 1999, UT1-UTC= 0.5144098, Accumulated Leap Seconds=32
  { 2451376.50, 63.670104 }, //17 July 1999, UT1-UTC= 0.5138963, Accumulated Leap Seconds=32
  { 2451377.50, 63.670651 }, //18 July 1999, UT1-UTC= 0.5133493, Accumulated Leap Seconds=32
  { 2451378.50, 63.671178 }, //19 July 1999, UT1-UTC= 0.5128217, Accumulated Leap Seconds=32
  { 2451379.50, 63.671659 }, //20 July 1999, UT1-UTC= 0.5123413, Accumulated Leap Seconds=32
  { 2451380.50, 63.672074 }, //21 July 1999, UT1-UTC= 0.5119259, Accumulated Leap Seconds=32
  { 2451381.50, 63.672407 }, //22 July 1999, UT1-UTC= 0.5115925, Accumulated Leap Seconds=32
  { 2451382.50, 63.672640 }, //23 July 1999, UT1-UTC= 0.5113601, Accumulated Leap Seconds=32
  { 2451383.50, 63.672773 }, //24 July 1999, UT1-UTC= 0.5112268, Accumulated Leap Seconds=32
  { 2451384.50, 63.672821 }, //25 July 1999, UT1-UTC= 0.5111795, Accumulated Leap Seconds=32
  { 2451385.50, 63.672797 }, //26 July 1999, UT1-UTC= 0.5112034, Accumulated Leap Seconds=32
  { 2451386.50, 63.672755 }, //27 July 1999, UT1-UTC= 0.5112447, Accumulated Leap Seconds=32
  { 2451387.50, 63.672753 }, //28 July 1999, UT1-UTC= 0.5112465, Accumulated Leap Seconds=32
  { 2451388.50, 63.672836 }, //29 July 1999, UT1-UTC= 0.5111641, Accumulated Leap Seconds=32
  { 2451389.50, 63.673033 }, //30 July 1999, UT1-UTC= 0.5109674, Accumulated Leap Seconds=32
  { 2451390.50, 63.673398 }, //31 July 1999, UT1-UTC= 0.5106019, Accumulated Leap Seconds=32
  { 2451391.50, 63.673940 }, //01 August 1999, UT1-UTC= 0.5100597, Accumulated Leap Seconds=32
  { 2451392.50, 63.674616 }, //02 August 1999, UT1-UTC= 0.5093835, Accumulated Leap Seconds=32
  { 2451393.50, 63.675417 }, //03 August 1999, UT1-UTC= 0.5085834, Accumulated Leap Seconds=32
  { 2451394.50, 63.676275 }, //04 August 1999, UT1-UTC= 0.5077248, Accumulated Leap Seconds=32
  { 2451395.50, 63.677048 }, //05 August 1999, UT1-UTC= 0.5069517, Accumulated Leap Seconds=32
  { 2451396.50, 63.677664 }, //06 August 1999, UT1-UTC= 0.5063363, Accumulated Leap Seconds=32
  { 2451397.50, 63.678114 }, //07 August 1999, UT1-UTC= 0.5058855, Accumulated Leap Seconds=32
  { 2451398.50, 63.678427 }, //08 August 1999, UT1-UTC= 0.5055726, Accumulated Leap Seconds=32
  { 2451399.50, 63.678643 }, //09 August 1999, UT1-UTC= 0.5053566, Accumulated Leap Seconds=32
  { 2451400.50, 63.678845 }, //10 August 1999, UT1-UTC= 0.5051554, Accumulated Leap Seconds=32
  { 2451401.50, 63.679118 }, //11 August 1999, UT1-UTC= 0.5048818, Accumulated Leap Seconds=32
  { 2451402.50, 63.679506 }, //12 August 1999, UT1-UTC= 0.5044943, Accumulated Leap Seconds=32
  { 2451403.50, 63.679969 }, //13 August 1999, UT1-UTC= 0.5040306, Accumulated Leap Seconds=32
  { 2451404.50, 63.680524 }, //14 August 1999, UT1-UTC= 0.5034755, Accumulated Leap Seconds=32
  { 2451405.50, 63.681166 }, //15 August 1999, UT1-UTC= 0.5028335, Accumulated Leap Seconds=32
  { 2451406.50, 63.681849 }, //16 August 1999, UT1-UTC= 0.5021506, Accumulated Leap Seconds=32
  { 2451407.50, 63.682524 }, //17 August 1999, UT1-UTC= 0.5014756, Accumulated Leap Seconds=32
  { 2451408.50, 63.683133 }, //18 August 1999, UT1-UTC= 0.5008665, Accumulated Leap Seconds=32
  { 2451409.50, 63.683646 }, //19 August 1999, UT1-UTC= 0.5003538, Accumulated Leap Seconds=32
  { 2451410.50, 63.684050 }, //20 August 1999, UT1-UTC= 0.4999497, Accumulated Leap Seconds=32
  { 2451411.50, 63.684355 }, //21 August 1999, UT1-UTC= 0.4996446, Accumulated Leap Seconds=32
  { 2451412.50, 63.684624 }, //22 August 1999, UT1-UTC= 0.4993757, Accumulated Leap Seconds=32
  { 2451413.50, 63.684892 }, //23 August 1999, UT1-UTC= 0.4991079, Accumulated Leap Seconds=32
  { 2451414.50, 63.685187 }, //24 August 1999, UT1-UTC= 0.4988128, Accumulated Leap Seconds=32
  { 2451415.50, 63.685561 }, //25 August 1999, UT1-UTC= 0.4984385, Accumulated Leap Seconds=32
  { 2451416.50, 63.686111 }, //26 August 1999, UT1-UTC= 0.4978885, Accumulated Leap Seconds=32
  { 2451417.50, 63.686886 }, //27 August 1999, UT1-UTC= 0.4971141, Accumulated Leap Seconds=32
  { 2451418.50, 63.687863 }, //28 August 1999, UT1-UTC= 0.4961367, Accumulated Leap Seconds=32
  { 2451419.50, 63.688987 }, //29 August 1999, UT1-UTC= 0.4950132, Accumulated Leap Seconds=32
  { 2451420.50, 63.690205 }, //30 August 1999, UT1-UTC= 0.4937950, Accumulated Leap Seconds=32
  { 2451421.50, 63.691440 }, //31 August 1999, UT1-UTC= 0.4925597, Accumulated Leap Seconds=32
  { 2451422.50, 63.692603 }, //01 September 1999, UT1-UTC= 0.4913970, Accumulated Leap Seconds=32
  { 2451423.50, 63.693626 }, //02 September 1999, UT1-UTC= 0.4903736, Accumulated Leap Seconds=32
  { 2451424.50, 63.694504 }, //03 September 1999, UT1-UTC= 0.4894959, Accumulated Leap Seconds=32
  { 2451425.50, 63.695267 }, //04 September 1999, UT1-UTC= 0.4887330, Accumulated Leap Seconds=32
  { 2451426.50, 63.695958 }, //05 September 1999, UT1-UTC= 0.4880416, Accumulated Leap Seconds=32
  { 2451427.50, 63.696623 }, //06 September 1999, UT1-UTC= 0.4873767, Accumulated Leap Seconds=32
  { 2451428.50, 63.697298 }, //07 September 1999, UT1-UTC= 0.4867018, Accumulated Leap Seconds=32
  { 2451429.50, 63.698003 }, //08 September 1999, UT1-UTC= 0.4859969, Accumulated Leap Seconds=32
  { 2451430.50, 63.698741 }, //09 September 1999, UT1-UTC= 0.4852594, Accumulated Leap Seconds=32
  { 2451431.50, 63.699511 }, //10 September 1999, UT1-UTC= 0.4844892, Accumulated Leap Seconds=32
  { 2451432.50, 63.700296 }, //11 September 1999, UT1-UTC= 0.4837042, Accumulated Leap Seconds=32
  { 2451433.50, 63.701045 }, //12 September 1999, UT1-UTC= 0.4829547, Accumulated Leap Seconds=32
  { 2451434.50, 63.701728 }, //13 September 1999, UT1-UTC= 0.4822724, Accumulated Leap Seconds=32
  { 2451435.50, 63.702324 }, //14 September 1999, UT1-UTC= 0.4816757, Accumulated Leap Seconds=32
  { 2451436.50, 63.702820 }, //15 September 1999, UT1-UTC= 0.4811804, Accumulated Leap Seconds=32
  { 2451437.50, 63.703199 }, //16 September 1999, UT1-UTC= 0.4808012, Accumulated Leap Seconds=32
  { 2451438.50, 63.703493 }, //17 September 1999, UT1-UTC= 0.4805073, Accumulated Leap Seconds=32
  { 2451439.50, 63.703760 }, //18 September 1999, UT1-UTC= 0.4802400, Accumulated Leap Seconds=32
  { 2451440.50, 63.704063 }, //19 September 1999, UT1-UTC= 0.4799374, Accumulated Leap Seconds=32
  { 2451441.50, 63.704445 }, //20 September 1999, UT1-UTC= 0.4795551, Accumulated Leap Seconds=32
  { 2451442.50, 63.704963 }, //21 September 1999, UT1-UTC= 0.4790366, Accumulated Leap Seconds=32
  { 2451443.50, 63.705666 }, //22 September 1999, UT1-UTC= 0.4783341, Accumulated Leap Seconds=32
  { 2451444.50, 63.706553 }, //23 September 1999, UT1-UTC= 0.4774473, Accumulated Leap Seconds=32
  { 2451445.50, 63.707596 }, //24 September 1999, UT1-UTC= 0.4764040, Accumulated Leap Seconds=32
  { 2451446.50, 63.708732 }, //25 September 1999, UT1-UTC= 0.4752684, Accumulated Leap Seconds=32
  { 2451447.50, 63.709898 }, //26 September 1999, UT1-UTC= 0.4741016, Accumulated Leap Seconds=32
  { 2451448.50, 63.711061 }, //27 September 1999, UT1-UTC= 0.4729385, Accumulated Leap Seconds=32
  { 2451449.50, 63.712148 }, //28 September 1999, UT1-UTC= 0.4718520, Accumulated Leap Seconds=32
  { 2451450.50, 63.713108 }, //29 September 1999, UT1-UTC= 0.4708922, Accumulated Leap Seconds=32
  { 2451451.50, 63.713961 }, //30 September 1999, UT1-UTC= 0.4700389, Accumulated Leap Seconds=32
  { 2451452.50, 63.714707 }, //01 October 1999, UT1-UTC= 0.4692934, Accumulated Leap Seconds=32
  { 2451453.50, 63.715355 }, //02 October 1999, UT1-UTC= 0.4686452, Accumulated Leap Seconds=32
  { 2451454.50, 63.716011 }, //03 October 1999, UT1-UTC= 0.4679887, Accumulated Leap Seconds=32
  { 2451455.50, 63.716799 }, //04 October 1999, UT1-UTC= 0.4672008, Accumulated Leap Seconds=32
  { 2451456.50, 63.717751 }, //05 October 1999, UT1-UTC= 0.4662491, Accumulated Leap Seconds=32
  { 2451457.50, 63.718861 }, //06 October 1999, UT1-UTC= 0.4651385, Accumulated Leap Seconds=32
  { 2451458.50, 63.720103 }, //07 October 1999, UT1-UTC= 0.4638969, Accumulated Leap Seconds=32
  { 2451459.50, 63.721419 }, //08 October 1999, UT1-UTC= 0.4625806, Accumulated Leap Seconds=32
  { 2451460.50, 63.722768 }, //09 October 1999, UT1-UTC= 0.4612317, Accumulated Leap Seconds=32
  { 2451461.50, 63.724102 }, //10 October 1999, UT1-UTC= 0.4598978, Accumulated Leap Seconds=32
  { 2451462.50, 63.725364 }, //11 October 1999, UT1-UTC= 0.4586355, Accumulated Leap Seconds=32
  { 2451463.50, 63.726505 }, //12 October 1999, UT1-UTC= 0.4574952, Accumulated Leap Seconds=32
  { 2451464.50, 63.727525 }, //13 October 1999, UT1-UTC= 0.4564746, Accumulated Leap Seconds=32
  { 2451465.50, 63.728471 }, //14 October 1999, UT1-UTC= 0.4555292, Accumulated Leap Seconds=32
  { 2451466.50, 63.729327 }, //15 October 1999, UT1-UTC= 0.4546726, Accumulated Leap Seconds=32
  { 2451467.50, 63.730093 }, //16 October 1999, UT1-UTC= 0.4539074, Accumulated Leap Seconds=32
  { 2451468.50, 63.730818 }, //17 October 1999, UT1-UTC= 0.4531824, Accumulated Leap Seconds=32
  { 2451469.50, 63.731570 }, //18 October 1999, UT1-UTC= 0.4524298, Accumulated Leap Seconds=32
  { 2451470.50, 63.732408 }, //19 October 1999, UT1-UTC= 0.4515918, Accumulated Leap Seconds=32
  { 2451471.50, 63.733400 }, //20 October 1999, UT1-UTC= 0.4505998, Accumulated Leap Seconds=32
  { 2451472.50, 63.734631 }, //21 October 1999, UT1-UTC= 0.4493691, Accumulated Leap Seconds=32
  { 2451473.50, 63.736121 }, //22 October 1999, UT1-UTC= 0.4478794, Accumulated Leap Seconds=32
  { 2451474.50, 63.737835 }, //23 October 1999, UT1-UTC= 0.4461654, Accumulated Leap Seconds=32
  { 2451475.50, 63.739702 }, //24 October 1999, UT1-UTC= 0.4442978, Accumulated Leap Seconds=32
  { 2451476.50, 63.741620 }, //25 October 1999, UT1-UTC= 0.4423803, Accumulated Leap Seconds=32
  { 2451477.50, 63.743478 }, //26 October 1999, UT1-UTC= 0.4405215, Accumulated Leap Seconds=32
  { 2451478.50, 63.745194 }, //27 October 1999, UT1-UTC= 0.4388057, Accumulated Leap Seconds=32
  { 2451479.50, 63.746699 }, //28 October 1999, UT1-UTC= 0.4373007, Accumulated Leap Seconds=32
  { 2451480.50, 63.748032 }, //29 October 1999, UT1-UTC= 0.4359680, Accumulated Leap Seconds=32
  { 2451481.50, 63.749278 }, //30 October 1999, UT1-UTC= 0.4347220, Accumulated Leap Seconds=32
  { 2451482.50, 63.750518 }, //31 October 1999, UT1-UTC= 0.4334822, Accumulated Leap Seconds=32
  { 2451483.50, 63.751805 }, //01 November 1999, UT1-UTC= 0.4321945, Accumulated Leap Seconds=32
  { 2451484.50, 63.753160 }, //02 November 1999, UT1-UTC= 0.4308404, Accumulated Leap Seconds=32
  { 2451485.50, 63.754593 }, //03 November 1999, UT1-UTC= 0.4294067, Accumulated Leap Seconds=32
  { 2451486.50, 63.756098 }, //04 November 1999, UT1-UTC= 0.4279022, Accumulated Leap Seconds=32
  { 2451487.50, 63.757614 }, //05 November 1999, UT1-UTC= 0.4263864, Accumulated Leap Seconds=32
  { 2451488.50, 63.759106 }, //06 November 1999, UT1-UTC= 0.4248938, Accumulated Leap Seconds=32
  { 2451489.50, 63.760551 }, //07 November 1999, UT1-UTC= 0.4234493, Accumulated Leap Seconds=32
  { 2451490.50, 63.761918 }, //08 November 1999, UT1-UTC= 0.4220819, Accumulated Leap Seconds=32
  { 2451491.50, 63.763224 }, //09 November 1999, UT1-UTC= 0.4207760, Accumulated Leap Seconds=32
  { 2451492.50, 63.764483 }, //10 November 1999, UT1-UTC= 0.4195172, Accumulated Leap Seconds=32
  { 2451493.50, 63.765676 }, //11 November 1999, UT1-UTC= 0.4183245, Accumulated Leap Seconds=32
  { 2451494.50, 63.766803 }, //12 November 1999, UT1-UTC= 0.4171970, Accumulated Leap Seconds=32
  { 2451495.50, 63.767914 }, //13 November 1999, UT1-UTC= 0.4160858, Accumulated Leap Seconds=32
  { 2451496.50, 63.769086 }, //14 November 1999, UT1-UTC= 0.4149142, Accumulated Leap Seconds=32
  { 2451497.50, 63.770344 }, //15 November 1999, UT1-UTC= 0.4136561, Accumulated Leap Seconds=32
  { 2451498.50, 63.771684 }, //16 November 1999, UT1-UTC= 0.4123155, Accumulated Leap Seconds=32
  { 2451499.50, 63.773136 }, //17 November 1999, UT1-UTC= 0.4108637, Accumulated Leap Seconds=32
  { 2451500.50, 63.774751 }, //18 November 1999, UT1-UTC= 0.4092486, Accumulated Leap Seconds=32
  { 2451501.50, 63.776489 }, //19 November 1999, UT1-UTC= 0.4075114, Accumulated Leap Seconds=32
  { 2451502.50, 63.778295 }, //20 November 1999, UT1-UTC= 0.4057054, Accumulated Leap Seconds=32
  { 2451503.50, 63.780119 }, //21 November 1999, UT1-UTC= 0.4038815, Accumulated Leap Seconds=32
  { 2451504.50, 63.781863 }, //22 November 1999, UT1-UTC= 0.4021374, Accumulated Leap Seconds=32
  { 2451505.50, 63.783437 }, //23 November 1999, UT1-UTC= 0.4005630, Accumulated Leap Seconds=32
  { 2451506.50, 63.784801 }, //24 November 1999, UT1-UTC= 0.3991994, Accumulated Leap Seconds=32
  { 2451507.50, 63.785970 }, //25 November 1999, UT1-UTC= 0.3980295, Accumulated Leap Seconds=32
  { 2451508.50, 63.787000 }, //26 November 1999, UT1-UTC= 0.3970002, Accumulated Leap Seconds=32
  { 2451509.50, 63.787973 }, //27 November 1999, UT1-UTC= 0.3960273, Accumulated Leap Seconds=32
  { 2451510.50, 63.788979 }, //28 November 1999, UT1-UTC= 0.3950205, Accumulated Leap Seconds=32
  { 2451511.50, 63.790093 }, //29 November 1999, UT1-UTC= 0.3939067, Accumulated Leap Seconds=32
  { 2451512.50, 63.791350 }, //30 November 1999, UT1-UTC= 0.3926501, Accumulated Leap Seconds=32
  { 2451513.50, 63.792717 }, //01 December 1999, UT1-UTC= 0.3912830, Accumulated Leap Seconds=32
  { 2451514.50, 63.794098 }, //02 December 1999, UT1-UTC= 0.3899017, Accumulated Leap Seconds=32
  { 2451515.50, 63.795430 }, //03 December 1999, UT1-UTC= 0.3885700, Accumulated Leap Seconds=32
  { 2451516.50, 63.796681 }, //04 December 1999, UT1-UTC= 0.3873187, Accumulated Leap Seconds=32
  { 2451517.50, 63.797838 }, //05 December 1999, UT1-UTC= 0.3861617, Accumulated Leap Seconds=32
  { 2451518.50, 63.798892 }, //06 December 1999, UT1-UTC= 0.3851078, Accumulated Leap Seconds=32
  { 2451519.50, 63.799849 }, //07 December 1999, UT1-UTC= 0.3841514, Accumulated Leap Seconds=32
  { 2451520.50, 63.800713 }, //08 December 1999, UT1-UTC= 0.3832869, Accumulated Leap Seconds=32
  { 2451521.50, 63.801513 }, //09 December 1999, UT1-UTC= 0.3824869, Accumulated Leap Seconds=32
  { 2451522.50, 63.802308 }, //10 December 1999, UT1-UTC= 0.3816924, Accumulated Leap Seconds=32
  { 2451523.50, 63.803142 }, //11 December 1999, UT1-UTC= 0.3808583, Accumulated Leap Seconds=32
  { 2451524.50, 63.804039 }, //12 December 1999, UT1-UTC= 0.3799612, Accumulated Leap Seconds=32
  { 2451525.50, 63.804999 }, //13 December 1999, UT1-UTC= 0.3790014, Accumulated Leap Seconds=32
  { 2451526.50, 63.806071 }, //14 December 1999, UT1-UTC= 0.3779287, Accumulated Leap Seconds=32
  { 2451527.50, 63.807281 }, //15 December 1999, UT1-UTC= 0.3767190, Accumulated Leap Seconds=32
  { 2451528.50, 63.808609 }, //16 December 1999, UT1-UTC= 0.3753910, Accumulated Leap Seconds=32
  { 2451529.50, 63.810051 }, //17 December 1999, UT1-UTC= 0.3739490, Accumulated Leap Seconds=32
  { 2451530.50, 63.811572 }, //18 December 1999, UT1-UTC= 0.3724280, Accumulated Leap Seconds=32
  { 2451531.50, 63.813109 }, //19 December 1999, UT1-UTC= 0.3708914, Accumulated Leap Seconds=32
  { 2451532.50, 63.814582 }, //20 December 1999, UT1-UTC= 0.3694183, Accumulated Leap Seconds=32
  { 2451533.50, 63.815922 }, //21 December 1999, UT1-UTC= 0.3680781, Accumulated Leap Seconds=32
  { 2451534.50, 63.817101 }, //22 December 1999, UT1-UTC= 0.3668986, Accumulated Leap Seconds=32
  { 2451535.50, 63.818166 }, //23 December 1999, UT1-UTC= 0.3658344, Accumulated Leap Seconds=32
  { 2451536.50, 63.819187 }, //24 December 1999, UT1-UTC= 0.3648134, Accumulated Leap Seconds=32
  { 2451537.50, 63.820232 }, //25 December 1999, UT1-UTC= 0.3637676, Accumulated Leap Seconds=32
  { 2451538.50, 63.821358 }, //26 December 1999, UT1-UTC= 0.3626416, Accumulated Leap Seconds=32
  { 2451539.50, 63.822582 }, //27 December 1999, UT1-UTC= 0.3614176, Accumulated Leap Seconds=32
  { 2451540.50, 63.823880 }, //28 December 1999, UT1-UTC= 0.3601199, Accumulated Leap Seconds=32
  { 2451541.50, 63.825185 }, //29 December 1999, UT1-UTC= 0.3588150, Accumulated Leap Seconds=32
  { 2451542.50, 63.826418 }, //30 December 1999, UT1-UTC= 0.3575820, Accumulated Leap Seconds=32
  { 2451543.50, 63.827532 }, //31 December 1999, UT1-UTC= 0.3564680, Accumulated Leap Seconds=32
  { 2451544.50, 63.828522 }, //01 January 2000, UT1-UTC= 0.3554779, Accumulated Leap Seconds=32
  { 2451545.50, 63.829399 }, //02 January 2000, UT1-UTC= 0.3546013, Accumulated Leap Seconds=32
  { 2451546.50, 63.830149 }, //03 January 2000, UT1-UTC= 0.3538511, Accumulated Leap Seconds=32
  { 2451547.50, 63.830759 }, //04 January 2000, UT1-UTC= 0.3532407, Accumulated Leap Seconds=32
  { 2451548.50, 63.831242 }, //05 January 2000, UT1-UTC= 0.3527584, Accumulated Leap Seconds=32
  { 2451549.50, 63.831652 }, //06 January 2000, UT1-UTC= 0.3523478, Accumulated Leap Seconds=32
  { 2451550.50, 63.832047 }, //07 January 2000, UT1-UTC= 0.3519531, Accumulated Leap Seconds=32
  { 2451551.50, 63.832455 }, //08 January 2000, UT1-UTC= 0.3515451, Accumulated Leap Seconds=32
  { 2451552.50, 63.832899 }, //09 January 2000, UT1-UTC= 0.3511013, Accumulated Leap Seconds=32
  { 2451553.50, 63.833442 }, //10 January 2000, UT1-UTC= 0.3505581, Accumulated Leap Seconds=32
  { 2451554.50, 63.834116 }, //11 January 2000, UT1-UTC= 0.3498837, Accumulated Leap Seconds=32
  { 2451555.50, 63.834942 }, //12 January 2000, UT1-UTC= 0.3490579, Accumulated Leap Seconds=32
  { 2451556.50, 63.835921 }, //13 January 2000, UT1-UTC= 0.3480789, Accumulated Leap Seconds=32
  { 2451557.50, 63.837047 }, //14 January 2000, UT1-UTC= 0.3469530, Accumulated Leap Seconds=32
  { 2451558.50, 63.838247 }, //15 January 2000, UT1-UTC= 0.3457527, Accumulated Leap Seconds=32
  { 2451559.50, 63.839429 }, //16 January 2000, UT1-UTC= 0.3445709, Accumulated Leap Seconds=32
  { 2451560.50, 63.840521 }, //17 January 2000, UT1-UTC= 0.3434786, Accumulated Leap Seconds=32
  { 2451561.50, 63.841506 }, //18 January 2000, UT1-UTC= 0.3424942, Accumulated Leap Seconds=32
  { 2451562.50, 63.842399 }, //19 January 2000, UT1-UTC= 0.3416014, Accumulated Leap Seconds=32
  { 2451563.50, 63.843232 }, //20 January 2000, UT1-UTC= 0.3407681, Accumulated Leap Seconds=32
  { 2451564.50, 63.844076 }, //21 January 2000, UT1-UTC= 0.3399239, Accumulated Leap Seconds=32
  { 2451565.50, 63.844991 }, //22 January 2000, UT1-UTC= 0.3390087, Accumulated Leap Seconds=32
  { 2451566.50, 63.846021 }, //23 January 2000, UT1-UTC= 0.3379788, Accumulated Leap Seconds=32
  { 2451567.50, 63.847177 }, //24 January 2000, UT1-UTC= 0.3368230, Accumulated Leap Seconds=32
  { 2451568.50, 63.848435 }, //25 January 2000, UT1-UTC= 0.3355650, Accumulated Leap Seconds=32
  { 2451569.50, 63.849737 }, //26 January 2000, UT1-UTC= 0.3342630, Accumulated Leap Seconds=32
  { 2451570.50, 63.851029 }, //27 January 2000, UT1-UTC= 0.3329705, Accumulated Leap Seconds=32
  { 2451571.50, 63.852241 }, //28 January 2000, UT1-UTC= 0.3317595, Accumulated Leap Seconds=32
  { 2451572.50, 63.853306 }, //29 January 2000, UT1-UTC= 0.3306943, Accumulated Leap Seconds=32
  { 2451573.50, 63.854219 }, //30 January 2000, UT1-UTC= 0.3297810, Accumulated Leap Seconds=32
  { 2451574.50, 63.855003 }, //31 January 2000, UT1-UTC= 0.3289974, Accumulated Leap Seconds=32
  { 2451575.50, 63.855687 }, //01 February 2000, UT1-UTC= 0.3283129, Accumulated Leap Seconds=32
  { 2451576.50, 63.856297 }, //02 February 2000, UT1-UTC= 0.3277029, Accumulated Leap Seconds=32
  { 2451577.50, 63.856835 }, //03 February 2000, UT1-UTC= 0.3271651, Accumulated Leap Seconds=32
  { 2451578.50, 63.857331 }, //04 February 2000, UT1-UTC= 0.3266687, Accumulated Leap Seconds=32
  { 2451579.50, 63.857864 }, //05 February 2000, UT1-UTC= 0.3261358, Accumulated Leap Seconds=32
  { 2451580.50, 63.858512 }, //06 February 2000, UT1-UTC= 0.3254875, Accumulated Leap Seconds=32
  { 2451581.50, 63.859309 }, //07 February 2000, UT1-UTC= 0.3246905, Accumulated Leap Seconds=32
  { 2451582.50, 63.860277 }, //08 February 2000, UT1-UTC= 0.3237227, Accumulated Leap Seconds=32
  { 2451583.50, 63.861405 }, //09 February 2000, UT1-UTC= 0.3225948, Accumulated Leap Seconds=32
  { 2451584.50, 63.862630 }, //10 February 2000, UT1-UTC= 0.3213703, Accumulated Leap Seconds=32
  { 2451585.50, 63.863848 }, //11 February 2000, UT1-UTC= 0.3201523, Accumulated Leap Seconds=32
  { 2451586.50, 63.864950 }, //12 February 2000, UT1-UTC= 0.3190495, Accumulated Leap Seconds=32
  { 2451587.50, 63.865884 }, //13 February 2000, UT1-UTC= 0.3181164, Accumulated Leap Seconds=32
  { 2451588.50, 63.866677 }, //14 February 2000, UT1-UTC= 0.3173225, Accumulated Leap Seconds=32
  { 2451589.50, 63.867365 }, //15 February 2000, UT1-UTC= 0.3166345, Accumulated Leap Seconds=32
  { 2451590.50, 63.868007 }, //16 February 2000, UT1-UTC= 0.3159933, Accumulated Leap Seconds=32
  { 2451591.50, 63.868657 }, //17 February 2000, UT1-UTC= 0.3153425, Accumulated Leap Seconds=32
  { 2451592.50, 63.869368 }, //18 February 2000, UT1-UTC= 0.3146323, Accumulated Leap Seconds=32
  { 2451593.50, 63.870200 }, //19 February 2000, UT1-UTC= 0.3137999, Accumulated Leap Seconds=32
  { 2451594.50, 63.871212 }, //20 February 2000, UT1-UTC= 0.3127883, Accumulated Leap Seconds=32
  { 2451595.50, 63.872391 }, //21 February 2000, UT1-UTC= 0.3116093, Accumulated Leap Seconds=32
  { 2451596.50, 63.873647 }, //22 February 2000, UT1-UTC= 0.3103534, Accumulated Leap Seconds=32
  { 2451597.50, 63.874880 }, //23 February 2000, UT1-UTC= 0.3091197, Accumulated Leap Seconds=32
  { 2451598.50, 63.876024 }, //24 February 2000, UT1-UTC= 0.3079762, Accumulated Leap Seconds=32
  { 2451599.50, 63.877039 }, //25 February 2000, UT1-UTC= 0.3069611, Accumulated Leap Seconds=32
  { 2451600.50, 63.877909 }, //26 February 2000, UT1-UTC= 0.3060912, Accumulated Leap Seconds=32
  { 2451601.50, 63.878649 }, //27 February 2000, UT1-UTC= 0.3053514, Accumulated Leap Seconds=32
  { 2451602.50, 63.879289 }, //28 February 2000, UT1-UTC= 0.3047112, Accumulated Leap Seconds=32
  { 2451603.50, 63.879852 }, //29 February 2000, UT1-UTC= 0.3041475, Accumulated Leap Seconds=32
  { 2451604.50, 63.880385 }, //01 March 2000, UT1-UTC= 0.3036146, Accumulated Leap Seconds=32
  { 2451605.50, 63.880952 }, //02 March 2000, UT1-UTC= 0.3030484, Accumulated Leap Seconds=32
  { 2451606.50, 63.881587 }, //03 March 2000, UT1-UTC= 0.3024127, Accumulated Leap Seconds=32
  { 2451607.50, 63.882331 }, //04 March 2000, UT1-UTC= 0.3016694, Accumulated Leap Seconds=32
  { 2451608.50, 63.883226 }, //05 March 2000, UT1-UTC= 0.3007739, Accumulated Leap Seconds=32
  { 2451609.50, 63.884301 }, //06 March 2000, UT1-UTC= 0.2996988, Accumulated Leap Seconds=32
  { 2451610.50, 63.885543 }, //07 March 2000, UT1-UTC= 0.2984566, Accumulated Leap Seconds=32
  { 2451611.50, 63.886903 }, //08 March 2000, UT1-UTC= 0.2970966, Accumulated Leap Seconds=32
  { 2451612.50, 63.888303 }, //09 March 2000, UT1-UTC= 0.2956967, Accumulated Leap Seconds=32
  { 2451613.50, 63.889669 }, //10 March 2000, UT1-UTC= 0.2943311, Accumulated Leap Seconds=32
  { 2451614.50, 63.890941 }, //11 March 2000, UT1-UTC= 0.2930592, Accumulated Leap Seconds=32
  { 2451615.50, 63.892082 }, //12 March 2000, UT1-UTC= 0.2919176, Accumulated Leap Seconds=32
  { 2451616.50, 63.893078 }, //13 March 2000, UT1-UTC= 0.2909222, Accumulated Leap Seconds=32
  { 2451617.50, 63.893941 }, //14 March 2000, UT1-UTC= 0.2900590, Accumulated Leap Seconds=32
  { 2451618.50, 63.894719 }, //15 March 2000, UT1-UTC= 0.2892809, Accumulated Leap Seconds=32
  { 2451619.50, 63.895510 }, //16 March 2000, UT1-UTC= 0.2884896, Accumulated Leap Seconds=32
  { 2451620.50, 63.896367 }, //17 March 2000, UT1-UTC= 0.2876329, Accumulated Leap Seconds=32
  { 2451621.50, 63.897291 }, //18 March 2000, UT1-UTC= 0.2867091, Accumulated Leap Seconds=32
  { 2451622.50, 63.898265 }, //19 March 2000, UT1-UTC= 0.2857348, Accumulated Leap Seconds=32
  { 2451623.50, 63.899270 }, //20 March 2000, UT1-UTC= 0.2847302, Accumulated Leap Seconds=32
  { 2451624.50, 63.900308 }, //21 March 2000, UT1-UTC= 0.2836923, Accumulated Leap Seconds=32
  { 2451625.50, 63.901365 }, //22 March 2000, UT1-UTC= 0.2826350, Accumulated Leap Seconds=32
  { 2451626.50, 63.902381 }, //23 March 2000, UT1-UTC= 0.2816194, Accumulated Leap Seconds=32
  { 2451627.50, 63.903281 }, //24 March 2000, UT1-UTC= 0.2807193, Accumulated Leap Seconds=32
  { 2451628.50, 63.904037 }, //25 March 2000, UT1-UTC= 0.2799627, Accumulated Leap Seconds=32
  { 2451629.50, 63.904646 }, //26 March 2000, UT1-UTC= 0.2793535, Accumulated Leap Seconds=32
  { 2451630.50, 63.905110 }, //27 March 2000, UT1-UTC= 0.2788899, Accumulated Leap Seconds=32
  { 2451631.50, 63.905507 }, //28 March 2000, UT1-UTC= 0.2784925, Accumulated Leap Seconds=32
  { 2451632.50, 63.905906 }, //29 March 2000, UT1-UTC= 0.2780937, Accumulated Leap Seconds=32
  { 2451633.50, 63.906327 }, //30 March 2000, UT1-UTC= 0.2776725, Accumulated Leap Seconds=32
  { 2451634.50, 63.906834 }, //31 March 2000, UT1-UTC= 0.2771664, Accumulated Leap Seconds=32
  { 2451635.50, 63.907502 }, //01 April 2000, UT1-UTC= 0.2764975, Accumulated Leap Seconds=32
  { 2451636.50, 63.908382 }, //02 April 2000, UT1-UTC= 0.2756181, Accumulated Leap Seconds=32
  { 2451637.50, 63.909468 }, //03 April 2000, UT1-UTC= 0.2745324, Accumulated Leap Seconds=32
  { 2451638.50, 63.910718 }, //04 April 2000, UT1-UTC= 0.2732821, Accumulated Leap Seconds=32
  { 2451639.50, 63.912055 }, //05 April 2000, UT1-UTC= 0.2719447, Accumulated Leap Seconds=32
  { 2451640.50, 63.913370 }, //06 April 2000, UT1-UTC= 0.2706298, Accumulated Leap Seconds=32
  { 2451641.50, 63.914581 }, //07 April 2000, UT1-UTC= 0.2694194, Accumulated Leap Seconds=32
  { 2451642.50, 63.915654 }, //08 April 2000, UT1-UTC= 0.2683463, Accumulated Leap Seconds=32
  { 2451643.50, 63.916588 }, //09 April 2000, UT1-UTC= 0.2674115, Accumulated Leap Seconds=32
  { 2451644.50, 63.917419 }, //10 April 2000, UT1-UTC= 0.2665807, Accumulated Leap Seconds=32
  { 2451645.50, 63.918212 }, //11 April 2000, UT1-UTC= 0.2657881, Accumulated Leap Seconds=32
  { 2451646.50, 63.919041 }, //12 April 2000, UT1-UTC= 0.2649589, Accumulated Leap Seconds=32
  { 2451647.50, 63.919970 }, //13 April 2000, UT1-UTC= 0.2640299, Accumulated Leap Seconds=32
  { 2451648.50, 63.921044 }, //14 April 2000, UT1-UTC= 0.2629560, Accumulated Leap Seconds=32
  { 2451649.50, 63.922281 }, //15 April 2000, UT1-UTC= 0.2617192, Accumulated Leap Seconds=32
  { 2451650.50, 63.923657 }, //16 April 2000, UT1-UTC= 0.2603431, Accumulated Leap Seconds=32
  { 2451651.50, 63.925116 }, //17 April 2000, UT1-UTC= 0.2588842, Accumulated Leap Seconds=32
  { 2451652.50, 63.926575 }, //18 April 2000, UT1-UTC= 0.2574245, Accumulated Leap Seconds=32
  { 2451653.50, 63.927964 }, //19 April 2000, UT1-UTC= 0.2560356, Accumulated Leap Seconds=32
  { 2451654.50, 63.929236 }, //20 April 2000, UT1-UTC= 0.2547639, Accumulated Leap Seconds=32
  { 2451655.50, 63.930391 }, //21 April 2000, UT1-UTC= 0.2536094, Accumulated Leap Seconds=32
  { 2451656.50, 63.931425 }, //22 April 2000, UT1-UTC= 0.2525748, Accumulated Leap Seconds=32
  { 2451657.50, 63.932324 }, //23 April 2000, UT1-UTC= 0.2516761, Accumulated Leap Seconds=32
  { 2451658.50, 63.933093 }, //24 April 2000, UT1-UTC= 0.2509068, Accumulated Leap Seconds=32
  { 2451659.50, 63.933764 }, //25 April 2000, UT1-UTC= 0.2502359, Accumulated Leap Seconds=32
  { 2451660.50, 63.934387 }, //26 April 2000, UT1-UTC= 0.2496127, Accumulated Leap Seconds=32
  { 2451661.50, 63.935040 }, //27 April 2000, UT1-UTC= 0.2489595, Accumulated Leap Seconds=32
  { 2451662.50, 63.935818 }, //28 April 2000, UT1-UTC= 0.2481815, Accumulated Leap Seconds=32
  { 2451663.50, 63.936785 }, //29 April 2000, UT1-UTC= 0.2472151, Accumulated Leap Seconds=32
  { 2451664.50, 63.937948 }, //30 April 2000, UT1-UTC= 0.2460516, Accumulated Leap Seconds=32
  { 2451665.50, 63.939279 }, //01 May 2000, UT1-UTC= 0.2447213, Accumulated Leap Seconds=32
  { 2451666.50, 63.940740 }, //02 May 2000, UT1-UTC= 0.2432598, Accumulated Leap Seconds=32
  { 2451667.50, 63.942255 }, //03 May 2000, UT1-UTC= 0.2417452, Accumulated Leap Seconds=32
  { 2451668.50, 63.943745 }, //04 May 2000, UT1-UTC= 0.2402546, Accumulated Leap Seconds=32
  { 2451669.50, 63.945115 }, //05 May 2000, UT1-UTC= 0.2388845, Accumulated Leap Seconds=32
  { 2451670.50, 63.946300 }, //06 May 2000, UT1-UTC= 0.2376998, Accumulated Leap Seconds=32
  { 2451671.50, 63.947311 }, //07 May 2000, UT1-UTC= 0.2366889, Accumulated Leap Seconds=32
  { 2451672.50, 63.948204 }, //08 May 2000, UT1-UTC= 0.2357958, Accumulated Leap Seconds=32
  { 2451673.50, 63.949057 }, //09 May 2000, UT1-UTC= 0.2349430, Accumulated Leap Seconds=32
  { 2451674.50, 63.949951 }, //10 May 2000, UT1-UTC= 0.2340491, Accumulated Leap Seconds=32
  { 2451675.50, 63.950963 }, //11 May 2000, UT1-UTC= 0.2330370, Accumulated Leap Seconds=32
  { 2451676.50, 63.952084 }, //12 May 2000, UT1-UTC= 0.2319156, Accumulated Leap Seconds=32
  { 2451677.50, 63.953294 }, //13 May 2000, UT1-UTC= 0.2307060, Accumulated Leap Seconds=32
  { 2451678.50, 63.954559 }, //14 May 2000, UT1-UTC= 0.2294409, Accumulated Leap Seconds=32
  { 2451679.50, 63.955827 }, //15 May 2000, UT1-UTC= 0.2281735, Accumulated Leap Seconds=32
  { 2451680.50, 63.957025 }, //16 May 2000, UT1-UTC= 0.2269746, Accumulated Leap Seconds=32
  { 2451681.50, 63.958113 }, //17 May 2000, UT1-UTC= 0.2258874, Accumulated Leap Seconds=32
  { 2451682.50, 63.959102 }, //18 May 2000, UT1-UTC= 0.2248976, Accumulated Leap Seconds=32
  { 2451683.50, 63.959976 }, //19 May 2000, UT1-UTC= 0.2240242, Accumulated Leap Seconds=32
  { 2451684.50, 63.960715 }, //20 May 2000, UT1-UTC= 0.2232846, Accumulated Leap Seconds=32
  { 2451685.50, 63.961332 }, //21 May 2000, UT1-UTC= 0.2226679, Accumulated Leap Seconds=32
  { 2451686.50, 63.961864 }, //22 May 2000, UT1-UTC= 0.2221362, Accumulated Leap Seconds=32
  { 2451687.50, 63.962365 }, //23 May 2000, UT1-UTC= 0.2216349, Accumulated Leap Seconds=32
  { 2451688.50, 63.962883 }, //24 May 2000, UT1-UTC= 0.2211173, Accumulated Leap Seconds=32
  { 2451689.50, 63.963430 }, //25 May 2000, UT1-UTC= 0.2205696, Accumulated Leap Seconds=32
  { 2451690.50, 63.964028 }, //26 May 2000, UT1-UTC= 0.2199716, Accumulated Leap Seconds=32
  { 2451691.50, 63.964701 }, //27 May 2000, UT1-UTC= 0.2192988, Accumulated Leap Seconds=32
  { 2451692.50, 63.965472 }, //28 May 2000, UT1-UTC= 0.2185279, Accumulated Leap Seconds=32
  { 2451693.50, 63.966340 }, //29 May 2000, UT1-UTC= 0.2176599, Accumulated Leap Seconds=32
  { 2451694.50, 63.967271 }, //30 May 2000, UT1-UTC= 0.2167292, Accumulated Leap Seconds=32
  { 2451695.50, 63.968207 }, //31 May 2000, UT1-UTC= 0.2157935, Accumulated Leap Seconds=32
  { 2451696.50, 63.969074 }, //01 June 2000, UT1-UTC= 0.2149256, Accumulated Leap Seconds=32
  { 2451697.50, 63.969819 }, //02 June 2000, UT1-UTC= 0.2141808, Accumulated Leap Seconds=32
  { 2451698.50, 63.970424 }, //03 June 2000, UT1-UTC= 0.2135763, Accumulated Leap Seconds=32
  { 2451699.50, 63.970911 }, //04 June 2000, UT1-UTC= 0.2130893, Accumulated Leap Seconds=32
  { 2451700.50, 63.971342 }, //05 June 2000, UT1-UTC= 0.2126575, Accumulated Leap Seconds=32
  { 2451701.50, 63.971821 }, //06 June 2000, UT1-UTC= 0.2121793, Accumulated Leap Seconds=32
  { 2451702.50, 63.972415 }, //07 June 2000, UT1-UTC= 0.2115854, Accumulated Leap Seconds=32
  { 2451703.50, 63.973164 }, //08 June 2000, UT1-UTC= 0.2108364, Accumulated Leap Seconds=32
  { 2451704.50, 63.974038 }, //09 June 2000, UT1-UTC= 0.2099624, Accumulated Leap Seconds=32
  { 2451705.50, 63.974960 }, //10 June 2000, UT1-UTC= 0.2090401, Accumulated Leap Seconds=32
  { 2451706.50, 63.975848 }, //11 June 2000, UT1-UTC= 0.2081524, Accumulated Leap Seconds=32
  { 2451707.50, 63.976639 }, //12 June 2000, UT1-UTC= 0.2073608, Accumulated Leap Seconds=32
  { 2451708.50, 63.977285 }, //13 June 2000, UT1-UTC= 0.2067150, Accumulated Leap Seconds=32
  { 2451709.50, 63.977742 }, //14 June 2000, UT1-UTC= 0.2062577, Accumulated Leap Seconds=32
  { 2451710.50, 63.977989 }, //15 June 2000, UT1-UTC= 0.2060111, Accumulated Leap Seconds=32
  { 2451711.50, 63.978038 }, //16 June 2000, UT1-UTC= 0.2059622, Accumulated Leap Seconds=32
  { 2451712.50, 63.977945 }, //17 June 2000, UT1-UTC= 0.2060551, Accumulated Leap Seconds=32
  { 2451713.50, 63.977778 }, //18 June 2000, UT1-UTC= 0.2062217, Accumulated Leap Seconds=32
  { 2451714.50, 63.977584 }, //19 June 2000, UT1-UTC= 0.2064164, Accumulated Leap Seconds=32
  { 2451715.50, 63.977404 }, //20 June 2000, UT1-UTC= 0.2065961, Accumulated Leap Seconds=32
  { 2451716.50, 63.977291 }, //21 June 2000, UT1-UTC= 0.2067086, Accumulated Leap Seconds=32
  { 2451717.50, 63.977290 }, //22 June 2000, UT1-UTC= 0.2067104, Accumulated Leap Seconds=32
  { 2451718.50, 63.977395 }, //23 June 2000, UT1-UTC= 0.2066051, Accumulated Leap Seconds=32
  { 2451719.50, 63.977595 }, //24 June 2000, UT1-UTC= 0.2064052, Accumulated Leap Seconds=32
  { 2451720.50, 63.977890 }, //25 June 2000, UT1-UTC= 0.2061103, Accumulated Leap Seconds=32
  { 2451721.50, 63.978276 }, //26 June 2000, UT1-UTC= 0.2057241, Accumulated Leap Seconds=32
  { 2451722.50, 63.978711 }, //27 June 2000, UT1-UTC= 0.2052886, Accumulated Leap Seconds=32
  { 2451723.50, 63.979148 }, //28 June 2000, UT1-UTC= 0.2048518, Accumulated Leap Seconds=32
  { 2451724.50, 63.979518 }, //29 June 2000, UT1-UTC= 0.2044819, Accumulated Leap Seconds=32
  { 2451725.50, 63.979759 }, //30 June 2000, UT1-UTC= 0.2042407, Accumulated Leap Seconds=32
  { 2451726.50, 63.979860 }, //01 July 2000, UT1-UTC= 0.2041396, Accumulated Leap Seconds=32
  { 2451727.50, 63.979873 }, //02 July 2000, UT1-UTC= 0.2041269, Accumulated Leap Seconds=32
  { 2451728.50, 63.979898 }, //03 July 2000, UT1-UTC= 0.2041020, Accumulated Leap Seconds=32
  { 2451729.50, 63.980033 }, //04 July 2000, UT1-UTC= 0.2039671, Accumulated Leap Seconds=32
  { 2451730.50, 63.980336 }, //05 July 2000, UT1-UTC= 0.2036636, Accumulated Leap Seconds=32
  { 2451731.50, 63.980814 }, //06 July 2000, UT1-UTC= 0.2031858, Accumulated Leap Seconds=32
  { 2451732.50, 63.981428 }, //07 July 2000, UT1-UTC= 0.2025722, Accumulated Leap Seconds=32
  { 2451733.50, 63.982093 }, //08 July 2000, UT1-UTC= 0.2019067, Accumulated Leap Seconds=32
  { 2451734.50, 63.982732 }, //09 July 2000, UT1-UTC= 0.2012677, Accumulated Leap Seconds=32
  { 2451735.50, 63.983293 }, //10 July 2000, UT1-UTC= 0.2007068, Accumulated Leap Seconds=32
  { 2451736.50, 63.983745 }, //11 July 2000, UT1-UTC= 0.2002548, Accumulated Leap Seconds=32
  { 2451737.50, 63.984085 }, //12 July 2000, UT1-UTC= 0.1999154, Accumulated Leap Seconds=32
  { 2451738.50, 63.984294 }, //13 July 2000, UT1-UTC= 0.1997056, Accumulated Leap Seconds=32
  { 2451739.50, 63.984386 }, //14 July 2000, UT1-UTC= 0.1996141, Accumulated Leap Seconds=32
  { 2451740.50, 63.984360 }, //15 July 2000, UT1-UTC= 0.1996400, Accumulated Leap Seconds=32
  { 2451741.50, 63.984229 }, //16 July 2000, UT1-UTC= 0.1997705, Accumulated Leap Seconds=32
  { 2451742.50, 63.984046 }, //17 July 2000, UT1-UTC= 0.1999541, Accumulated Leap Seconds=32
  { 2451743.50, 63.983851 }, //18 July 2000, UT1-UTC= 0.2001486, Accumulated Leap Seconds=32
  { 2451744.50, 63.983677 }, //19 July 2000, UT1-UTC= 0.2003232, Accumulated Leap Seconds=32
  { 2451745.50, 63.983548 }, //20 July 2000, UT1-UTC= 0.2004517, Accumulated Leap Seconds=32
  { 2451746.50, 63.983468 }, //21 July 2000, UT1-UTC= 0.2005316, Accumulated Leap Seconds=32
  { 2451747.50, 63.983434 }, //22 July 2000, UT1-UTC= 0.2005659, Accumulated Leap Seconds=32
  { 2451748.50, 63.983460 }, //23 July 2000, UT1-UTC= 0.2005401, Accumulated Leap Seconds=32
  { 2451749.50, 63.983537 }, //24 July 2000, UT1-UTC= 0.2004625, Accumulated Leap Seconds=32
  { 2451750.50, 63.983633 }, //25 July 2000, UT1-UTC= 0.2003671, Accumulated Leap Seconds=32
  { 2451751.50, 63.983709 }, //26 July 2000, UT1-UTC= 0.2002905, Accumulated Leap Seconds=32
  { 2451752.50, 63.983724 }, //27 July 2000, UT1-UTC= 0.2002763, Accumulated Leap Seconds=32
  { 2451753.50, 63.983650 }, //28 July 2000, UT1-UTC= 0.2003495, Accumulated Leap Seconds=32
  { 2451754.50, 63.983503 }, //29 July 2000, UT1-UTC= 0.2004970, Accumulated Leap Seconds=32
  { 2451755.50, 63.983341 }, //30 July 2000, UT1-UTC= 0.2006586, Accumulated Leap Seconds=32
  { 2451756.50, 63.983253 }, //31 July 2000, UT1-UTC= 0.2007471, Accumulated Leap Seconds=32
  { 2451757.50, 63.983308 }, //01 August 2000, UT1-UTC= 0.2006923, Accumulated Leap Seconds=32
  { 2451758.50, 63.983527 }, //02 August 2000, UT1-UTC= 0.2004733, Accumulated Leap Seconds=32
  { 2451759.50, 63.983876 }, //03 August 2000, UT1-UTC= 0.2001244, Accumulated Leap Seconds=32
  { 2451760.50, 63.984281 }, //04 August 2000, UT1-UTC= 0.1997185, Accumulated Leap Seconds=32
  { 2451761.50, 63.984670 }, //05 August 2000, UT1-UTC= 0.1993296, Accumulated Leap Seconds=32
  { 2451762.50, 63.984979 }, //06 August 2000, UT1-UTC= 0.1990213, Accumulated Leap Seconds=32
  { 2451763.50, 63.985162 }, //07 August 2000, UT1-UTC= 0.1988380, Accumulated Leap Seconds=32
  { 2451764.50, 63.985201 }, //08 August 2000, UT1-UTC= 0.1987994, Accumulated Leap Seconds=32
  { 2451765.50, 63.985102 }, //09 August 2000, UT1-UTC= 0.1988978, Accumulated Leap Seconds=32
  { 2451766.50, 63.984911 }, //10 August 2000, UT1-UTC= 0.1990886, Accumulated Leap Seconds=32
  { 2451767.50, 63.984671 }, //11 August 2000, UT1-UTC= 0.1993289, Accumulated Leap Seconds=32
  { 2451768.50, 63.984421 }, //12 August 2000, UT1-UTC= 0.1995786, Accumulated Leap Seconds=32
  { 2451769.50, 63.984203 }, //13 August 2000, UT1-UTC= 0.1997968, Accumulated Leap Seconds=32
  { 2451770.50, 63.984054 }, //14 August 2000, UT1-UTC= 0.1999461, Accumulated Leap Seconds=32
  { 2451771.50, 63.984008 }, //15 August 2000, UT1-UTC= 0.1999916, Accumulated Leap Seconds=32
  { 2451772.50, 63.984097 }, //16 August 2000, UT1-UTC= 0.1999027, Accumulated Leap Seconds=32
  { 2451773.50, 63.984354 }, //17 August 2000, UT1-UTC= 0.1996462, Accumulated Leap Seconds=32
  { 2451774.50, 63.984776 }, //18 August 2000, UT1-UTC= 0.1992241, Accumulated Leap Seconds=32
  { 2451775.50, 63.985330 }, //19 August 2000, UT1-UTC= 0.1986702, Accumulated Leap Seconds=32
  { 2451776.50, 63.985972 }, //20 August 2000, UT1-UTC= 0.1980283, Accumulated Leap Seconds=32
  { 2451777.50, 63.986652 }, //21 August 2000, UT1-UTC= 0.1973479, Accumulated Leap Seconds=32
  { 2451778.50, 63.987321 }, //22 August 2000, UT1-UTC= 0.1966790, Accumulated Leap Seconds=32
  { 2451779.50, 63.987937 }, //23 August 2000, UT1-UTC= 0.1960633, Accumulated Leap Seconds=32
  { 2451780.50, 63.988480 }, //24 August 2000, UT1-UTC= 0.1955199, Accumulated Leap Seconds=32
  { 2451781.50, 63.988926 }, //25 August 2000, UT1-UTC= 0.1950741, Accumulated Leap Seconds=32
  { 2451782.50, 63.989322 }, //26 August 2000, UT1-UTC= 0.1946783, Accumulated Leap Seconds=32
  { 2451783.50, 63.989759 }, //27 August 2000, UT1-UTC= 0.1942415, Accumulated Leap Seconds=32
  { 2451784.50, 63.990311 }, //28 August 2000, UT1-UTC= 0.1936888, Accumulated Leap Seconds=32
  { 2451785.50, 63.991025 }, //29 August 2000, UT1-UTC= 0.1929748, Accumulated Leap Seconds=32
  { 2451786.50, 63.991893 }, //30 August 2000, UT1-UTC= 0.1921072, Accumulated Leap Seconds=32
  { 2451787.50, 63.992847 }, //31 August 2000, UT1-UTC= 0.1911533, Accumulated Leap Seconds=32
  { 2451788.50, 63.993801 }, //01 September 2000, UT1-UTC= 0.1901989, Accumulated Leap Seconds=32
  { 2451789.50, 63.994674 }, //02 September 2000, UT1-UTC= 0.1893256, Accumulated Leap Seconds=32
  { 2451790.50, 63.995408 }, //03 September 2000, UT1-UTC= 0.1885919, Accumulated Leap Seconds=32
  { 2451791.50, 63.995973 }, //04 September 2000, UT1-UTC= 0.1880273, Accumulated Leap Seconds=32
  { 2451792.50, 63.996365 }, //05 September 2000, UT1-UTC= 0.1876349, Accumulated Leap Seconds=32
  { 2451793.50, 63.996597 }, //06 September 2000, UT1-UTC= 0.1874028, Accumulated Leap Seconds=32
  { 2451794.50, 63.996690 }, //07 September 2000, UT1-UTC= 0.1873098, Accumulated Leap Seconds=32
  { 2451795.50, 63.996678 }, //08 September 2000, UT1-UTC= 0.1873223, Accumulated Leap Seconds=32
  { 2451796.50, 63.996614 }, //09 September 2000, UT1-UTC= 0.1873861, Accumulated Leap Seconds=32
  { 2451797.50, 63.996573 }, //10 September 2000, UT1-UTC= 0.1874266, Accumulated Leap Seconds=32
  { 2451798.50, 63.996599 }, //11 September 2000, UT1-UTC= 0.1874013, Accumulated Leap Seconds=32
  { 2451799.50, 63.996713 }, //12 September 2000, UT1-UTC= 0.1872869, Accumulated Leap Seconds=32
  { 2451800.50, 63.996950 }, //13 September 2000, UT1-UTC= 0.1870496, Accumulated Leap Seconds=32
  { 2451801.50, 63.997363 }, //14 September 2000, UT1-UTC= 0.1866365, Accumulated Leap Seconds=32
  { 2451802.50, 63.997945 }, //15 September 2000, UT1-UTC= 0.1860550, Accumulated Leap Seconds=32
  { 2451803.50, 63.998644 }, //16 September 2000, UT1-UTC= 0.1853555, Accumulated Leap Seconds=32
  { 2451804.50, 63.999376 }, //17 September 2000, UT1-UTC= 0.1846236, Accumulated Leap Seconds=32
  { 2451805.50, 64.000062 }, //18 September 2000, UT1-UTC= 0.1839384, Accumulated Leap Seconds=32
  { 2451806.50, 64.000642 }, //19 September 2000, UT1-UTC= 0.1833584, Accumulated Leap Seconds=32
  { 2451807.50, 64.001104 }, //20 September 2000, UT1-UTC= 0.1828959, Accumulated Leap Seconds=32
  { 2451808.50, 64.001499 }, //21 September 2000, UT1-UTC= 0.1825005, Accumulated Leap Seconds=32
  { 2451809.50, 64.001857 }, //22 September 2000, UT1-UTC= 0.1821426, Accumulated Leap Seconds=32
  { 2451810.50, 64.002237 }, //23 September 2000, UT1-UTC= 0.1817630, Accumulated Leap Seconds=32
  { 2451811.50, 64.002705 }, //24 September 2000, UT1-UTC= 0.1812951, Accumulated Leap Seconds=32
  { 2451812.50, 64.003325 }, //25 September 2000, UT1-UTC= 0.1806746, Accumulated Leap Seconds=32
  { 2451813.50, 64.004137 }, //26 September 2000, UT1-UTC= 0.1798634, Accumulated Leap Seconds=32
  { 2451814.50, 64.005124 }, //27 September 2000, UT1-UTC= 0.1788756, Accumulated Leap Seconds=32
  { 2451815.50, 64.006234 }, //28 September 2000, UT1-UTC= 0.1777662, Accumulated Leap Seconds=32
  { 2451816.50, 64.007366 }, //29 September 2000, UT1-UTC= 0.1766344, Accumulated Leap Seconds=32
  { 2451817.50, 64.008423 }, //30 September 2000, UT1-UTC= 0.1755771, Accumulated Leap Seconds=32
  { 2451818.50, 64.009338 }, //01 October 2000, UT1-UTC= 0.1746616, Accumulated Leap Seconds=32
  { 2451819.50, 64.010068 }, //02 October 2000, UT1-UTC= 0.1739322, Accumulated Leap Seconds=32
  { 2451820.50, 64.010595 }, //03 October 2000, UT1-UTC= 0.1734054, Accumulated Leap Seconds=32
  { 2451821.50, 64.010940 }, //04 October 2000, UT1-UTC= 0.1730604, Accumulated Leap Seconds=32
  { 2451822.50, 64.011181 }, //05 October 2000, UT1-UTC= 0.1728192, Accumulated Leap Seconds=32
  { 2451823.50, 64.011408 }, //06 October 2000, UT1-UTC= 0.1725918, Accumulated Leap Seconds=32
  { 2451824.50, 64.011699 }, //07 October 2000, UT1-UTC= 0.1723008, Accumulated Leap Seconds=32
  { 2451825.50, 64.012114 }, //08 October 2000, UT1-UTC= 0.1718859, Accumulated Leap Seconds=32
  { 2451826.50, 64.012692 }, //09 October 2000, UT1-UTC= 0.1713077, Accumulated Leap Seconds=32
  { 2451827.50, 64.013455 }, //10 October 2000, UT1-UTC= 0.1705445, Accumulated Leap Seconds=32
  { 2451828.50, 64.014409 }, //11 October 2000, UT1-UTC= 0.1695914, Accumulated Leap Seconds=32
  { 2451829.50, 64.015527 }, //12 October 2000, UT1-UTC= 0.1684726, Accumulated Leap Seconds=32
  { 2451830.50, 64.016771 }, //13 October 2000, UT1-UTC= 0.1672293, Accumulated Leap Seconds=32
  { 2451831.50, 64.018084 }, //14 October 2000, UT1-UTC= 0.1659165, Accumulated Leap Seconds=32
  { 2451832.50, 64.019400 }, //15 October 2000, UT1-UTC= 0.1646002, Accumulated Leap Seconds=32
  { 2451833.50, 64.020654 }, //16 October 2000, UT1-UTC= 0.1633460, Accumulated Leap Seconds=32
  { 2451834.50, 64.021799 }, //17 October 2000, UT1-UTC= 0.1622008, Accumulated Leap Seconds=32
  { 2451835.50, 64.022826 }, //18 October 2000, UT1-UTC= 0.1611743, Accumulated Leap Seconds=32
  { 2451836.50, 64.023778 }, //19 October 2000, UT1-UTC= 0.1602224, Accumulated Leap Seconds=32
  { 2451837.50, 64.024698 }, //20 October 2000, UT1-UTC= 0.1593019, Accumulated Leap Seconds=32
  { 2451838.50, 64.025663 }, //21 October 2000, UT1-UTC= 0.1583373, Accumulated Leap Seconds=32
  { 2451839.50, 64.026751 }, //22 October 2000, UT1-UTC= 0.1572489, Accumulated Leap Seconds=32
  { 2451840.50, 64.028003 }, //23 October 2000, UT1-UTC= 0.1559969, Accumulated Leap Seconds=32
  { 2451841.50, 64.029416 }, //24 October 2000, UT1-UTC= 0.1545842, Accumulated Leap Seconds=32
  { 2451842.50, 64.030949 }, //25 October 2000, UT1-UTC= 0.1530514, Accumulated Leap Seconds=32
  { 2451843.50, 64.032531 }, //26 October 2000, UT1-UTC= 0.1514687, Accumulated Leap Seconds=32
  { 2451844.50, 64.034086 }, //27 October 2000, UT1-UTC= 0.1499145, Accumulated Leap Seconds=32
  { 2451845.50, 64.035542 }, //28 October 2000, UT1-UTC= 0.1484579, Accumulated Leap Seconds=32
  { 2451846.50, 64.036858 }, //29 October 2000, UT1-UTC= 0.1471418, Accumulated Leap Seconds=32
  { 2451847.50, 64.038025 }, //30 October 2000, UT1-UTC= 0.1459752, Accumulated Leap Seconds=32
  { 2451848.50, 64.039052 }, //31 October 2000, UT1-UTC= 0.1449482, Accumulated Leap Seconds=32
  { 2451849.50, 64.039962 }, //01 November 2000, UT1-UTC= 0.1440379, Accumulated Leap Seconds=32
  { 2451850.50, 64.040775 }, //02 November 2000, UT1-UTC= 0.1432245, Accumulated Leap Seconds=32
  { 2451851.50, 64.041579 }, //03 November 2000, UT1-UTC= 0.1424206, Accumulated Leap Seconds=32
  { 2451852.50, 64.042422 }, //04 November 2000, UT1-UTC= 0.1415779, Accumulated Leap Seconds=32
  { 2451853.50, 64.043321 }, //05 November 2000, UT1-UTC= 0.1406790, Accumulated Leap Seconds=32
  { 2451854.50, 64.044298 }, //06 November 2000, UT1-UTC= 0.1397020, Accumulated Leap Seconds=32
  { 2451855.50, 64.045369 }, //07 November 2000, UT1-UTC= 0.1386307, Accumulated Leap Seconds=32
  { 2451856.50, 64.046544 }, //08 November 2000, UT1-UTC= 0.1374558, Accumulated Leap Seconds=32
  { 2451857.50, 64.047821 }, //09 November 2000, UT1-UTC= 0.1361794, Accumulated Leap Seconds=32
  { 2451858.50, 64.049159 }, //10 November 2000, UT1-UTC= 0.1348407, Accumulated Leap Seconds=32
  { 2451859.50, 64.050489 }, //11 November 2000, UT1-UTC= 0.1335115, Accumulated Leap Seconds=32
  { 2451860.50, 64.051718 }, //12 November 2000, UT1-UTC= 0.1322815, Accumulated Leap Seconds=32
  { 2451861.50, 64.052771 }, //13 November 2000, UT1-UTC= 0.1312288, Accumulated Leap Seconds=32
  { 2451862.50, 64.053620 }, //14 November 2000, UT1-UTC= 0.1303795, Accumulated Leap Seconds=32
  { 2451863.50, 64.054312 }, //15 November 2000, UT1-UTC= 0.1296880, Accumulated Leap Seconds=32
  { 2451864.50, 64.054925 }, //16 November 2000, UT1-UTC= 0.1290752, Accumulated Leap Seconds=32
  { 2451865.50, 64.055533 }, //17 November 2000, UT1-UTC= 0.1284673, Accumulated Leap Seconds=32
  { 2451866.50, 64.056220 }, //18 November 2000, UT1-UTC= 0.1277803, Accumulated Leap Seconds=32
  { 2451867.50, 64.057045 }, //19 November 2000, UT1-UTC= 0.1269550, Accumulated Leap Seconds=32
  { 2451868.50, 64.058020 }, //20 November 2000, UT1-UTC= 0.1259796, Accumulated Leap Seconds=32
  { 2451869.50, 64.059119 }, //21 November 2000, UT1-UTC= 0.1248810, Accumulated Leap Seconds=32
  { 2451870.50, 64.060291 }, //22 November 2000, UT1-UTC= 0.1237094, Accumulated Leap Seconds=32
  { 2451871.50, 64.061474 }, //23 November 2000, UT1-UTC= 0.1225261, Accumulated Leap Seconds=32
  { 2451872.50, 64.062573 }, //24 November 2000, UT1-UTC= 0.1214273, Accumulated Leap Seconds=32
  { 2451873.50, 64.063526 }, //25 November 2000, UT1-UTC= 0.1204737, Accumulated Leap Seconds=32
  { 2451874.50, 64.064311 }, //26 November 2000, UT1-UTC= 0.1196892, Accumulated Leap Seconds=32
  { 2451875.50, 64.064942 }, //27 November 2000, UT1-UTC= 0.1190582, Accumulated Leap Seconds=32
  { 2451876.50, 64.065463 }, //28 November 2000, UT1-UTC= 0.1185366, Accumulated Leap Seconds=32
  { 2451877.50, 64.065949 }, //29 November 2000, UT1-UTC= 0.1180511, Accumulated Leap Seconds=32
  { 2451878.50, 64.066464 }, //30 November 2000, UT1-UTC= 0.1175362, Accumulated Leap Seconds=32
  { 2451879.50, 64.067043 }, //01 December 2000, UT1-UTC= 0.1169571, Accumulated Leap Seconds=32
  { 2451880.50, 64.067716 }, //02 December 2000, UT1-UTC= 0.1162837, Accumulated Leap Seconds=32
  { 2451881.50, 64.068523 }, //03 December 2000, UT1-UTC= 0.1154766, Accumulated Leap Seconds=32
  { 2451882.50, 64.069474 }, //04 December 2000, UT1-UTC= 0.1145258, Accumulated Leap Seconds=32
  { 2451883.50, 64.070560 }, //05 December 2000, UT1-UTC= 0.1134396, Accumulated Leap Seconds=32
  { 2451884.50, 64.071763 }, //06 December 2000, UT1-UTC= 0.1122369, Accumulated Leap Seconds=32
  { 2451885.50, 64.073058 }, //07 December 2000, UT1-UTC= 0.1109424, Accumulated Leap Seconds=32
  { 2451886.50, 64.074390 }, //08 December 2000, UT1-UTC= 0.1096102, Accumulated Leap Seconds=32
  { 2451887.50, 64.075676 }, //09 December 2000, UT1-UTC= 0.1083239, Accumulated Leap Seconds=32
  { 2451888.50, 64.076849 }, //10 December 2000, UT1-UTC= 0.1071513, Accumulated Leap Seconds=32
  { 2451889.50, 64.077860 }, //11 December 2000, UT1-UTC= 0.1061402, Accumulated Leap Seconds=32
  { 2451890.50, 64.078698 }, //12 December 2000, UT1-UTC= 0.1053021, Accumulated Leap Seconds=32
  { 2451891.50, 64.079404 }, //13 December 2000, UT1-UTC= 0.1045962, Accumulated Leap Seconds=32
  { 2451892.50, 64.080057 }, //14 December 2000, UT1-UTC= 0.1039425, Accumulated Leap Seconds=32
  { 2451893.50, 64.080748 }, //15 December 2000, UT1-UTC= 0.1032518, Accumulated Leap Seconds=32
  { 2451894.50, 64.081521 }, //16 December 2000, UT1-UTC= 0.1024791, Accumulated Leap Seconds=32
  { 2451895.50, 64.082379 }, //17 December 2000, UT1-UTC= 0.1016208, Accumulated Leap Seconds=32
  { 2451896.50, 64.083308 }, //18 December 2000, UT1-UTC= 0.1006920, Accumulated Leap Seconds=32
  { 2451897.50, 64.084266 }, //19 December 2000, UT1-UTC= 0.0997342, Accumulated Leap Seconds=32
  { 2451898.50, 64.085205 }, //20 December 2000, UT1-UTC= 0.0987954, Accumulated Leap Seconds=32
  { 2451899.50, 64.086070 }, //21 December 2000, UT1-UTC= 0.0979297, Accumulated Leap Seconds=32
  { 2451900.50, 64.086815 }, //22 December 2000, UT1-UTC= 0.0971851, Accumulated Leap Seconds=32
  { 2451901.50, 64.087430 }, //23 December 2000, UT1-UTC= 0.0965696, Accumulated Leap Seconds=32
  { 2451902.50, 64.087923 }, //24 December 2000, UT1-UTC= 0.0960771, Accumulated Leap Seconds=32
  { 2451903.50, 64.088304 }, //25 December 2000, UT1-UTC= 0.0956962, Accumulated Leap Seconds=32
  { 2451904.50, 64.088600 }, //26 December 2000, UT1-UTC= 0.0954001, Accumulated Leap Seconds=32
  { 2451905.50, 64.088849 }, //27 December 2000, UT1-UTC= 0.0951508, Accumulated Leap Seconds=32
  { 2451906.50, 64.089090 }, //28 December 2000, UT1-UTC= 0.0949096, Accumulated Leap Seconds=32
  { 2451907.50, 64.089364 }, //29 December 2000, UT1-UTC= 0.0946359, Accumulated Leap Seconds=32
  { 2451908.50, 64.089717 }, //30 December 2000, UT1-UTC= 0.0942828, Accumulated Leap Seconds=32
  { 2451909.50, 64.090188 }, //31 December 2000, UT1-UTC= 0.0938120, Accumulated Leap Seconds=32
  { 2451910.50, 64.090788 }, //01 January 2001, UT1-UTC= 0.0932119, Accumulated Leap Seconds=32
  { 2451911.50, 64.091509 }, //02 January 2001, UT1-UTC= 0.0924914, Accumulated Leap Seconds=32
  { 2451912.50, 64.092331 }, //03 January 2001, UT1-UTC= 0.0916692, Accumulated Leap Seconds=32
  { 2451913.50, 64.093225 }, //04 January 2001, UT1-UTC= 0.0907750, Accumulated Leap Seconds=32
  { 2451914.50, 64.094132 }, //05 January 2001, UT1-UTC= 0.0898678, Accumulated Leap Seconds=32
  { 2451915.50, 64.094981 }, //06 January 2001, UT1-UTC= 0.0890192, Accumulated Leap Seconds=32
  { 2451916.50, 64.095709 }, //07 January 2001, UT1-UTC= 0.0882913, Accumulated Leap Seconds=32
  { 2451917.50, 64.096290 }, //08 January 2001, UT1-UTC= 0.0877100, Accumulated Leap Seconds=32
  { 2451918.50, 64.096759 }, //09 January 2001, UT1-UTC= 0.0872412, Accumulated Leap Seconds=32
  { 2451919.50, 64.097179 }, //10 January 2001, UT1-UTC= 0.0868214, Accumulated Leap Seconds=32
  { 2451920.50, 64.097602 }, //11 January 2001, UT1-UTC= 0.0863978, Accumulated Leap Seconds=32
  { 2451921.50, 64.098108 }, //12 January 2001, UT1-UTC= 0.0858918, Accumulated Leap Seconds=32
  { 2451922.50, 64.098772 }, //13 January 2001, UT1-UTC= 0.0852278, Accumulated Leap Seconds=32
  { 2451923.50, 64.099603 }, //14 January 2001, UT1-UTC= 0.0843969, Accumulated Leap Seconds=32
  { 2451924.50, 64.100535 }, //15 January 2001, UT1-UTC= 0.0834651, Accumulated Leap Seconds=32
  { 2451925.50, 64.101480 }, //16 January 2001, UT1-UTC= 0.0825200, Accumulated Leap Seconds=32
  { 2451926.50, 64.102338 }, //17 January 2001, UT1-UTC= 0.0816624, Accumulated Leap Seconds=32
  { 2451927.50, 64.103020 }, //18 January 2001, UT1-UTC= 0.0809795, Accumulated Leap Seconds=32
  { 2451928.50, 64.103508 }, //19 January 2001, UT1-UTC= 0.0804921, Accumulated Leap Seconds=32
  { 2451929.50, 64.103821 }, //20 January 2001, UT1-UTC= 0.0801787, Accumulated Leap Seconds=32
  { 2451930.50, 64.103980 }, //21 January 2001, UT1-UTC= 0.0800201, Accumulated Leap Seconds=32
  { 2451931.50, 64.104019 }, //22 January 2001, UT1-UTC= 0.0799806, Accumulated Leap Seconds=32
  { 2451932.50, 64.103976 }, //23 January 2001, UT1-UTC= 0.0800240, Accumulated Leap Seconds=32
  { 2451933.50, 64.103893 }, //24 January 2001, UT1-UTC= 0.0801074, Accumulated Leap Seconds=32
  { 2451934.50, 64.103883 }, //25 January 2001, UT1-UTC= 0.0801174, Accumulated Leap Seconds=32
  { 2451935.50, 64.104000 }, //26 January 2001, UT1-UTC= 0.0800005, Accumulated Leap Seconds=32
  { 2451936.50, 64.104252 }, //27 January 2001, UT1-UTC= 0.0797485, Accumulated Leap Seconds=32
  { 2451937.50, 64.104640 }, //28 January 2001, UT1-UTC= 0.0793597, Accumulated Leap Seconds=32
  { 2451938.50, 64.105143 }, //29 January 2001, UT1-UTC= 0.0788571, Accumulated Leap Seconds=32
  { 2451939.50, 64.105714 }, //30 January 2001, UT1-UTC= 0.0782861, Accumulated Leap Seconds=32
  { 2451940.50, 64.106293 }, //31 January 2001, UT1-UTC= 0.0777066, Accumulated Leap Seconds=32
  { 2451941.50, 64.106808 }, //01 February 2001, UT1-UTC= 0.0771923, Accumulated Leap Seconds=32
  { 2451942.50, 64.107223 }, //02 February 2001, UT1-UTC= 0.0767767, Accumulated Leap Seconds=32
  { 2451943.50, 64.107509 }, //03 February 2001, UT1-UTC= 0.0764905, Accumulated Leap Seconds=32
  { 2451944.50, 64.107654 }, //04 February 2001, UT1-UTC= 0.0763463, Accumulated Leap Seconds=32
  { 2451945.50, 64.107693 }, //05 February 2001, UT1-UTC= 0.0763070, Accumulated Leap Seconds=32
  { 2451946.50, 64.107710 }, //06 February 2001, UT1-UTC= 0.0762904, Accumulated Leap Seconds=32
  { 2451947.50, 64.107821 }, //07 February 2001, UT1-UTC= 0.0761786, Accumulated Leap Seconds=32
  { 2451948.50, 64.108163 }, //08 February 2001, UT1-UTC= 0.0758370, Accumulated Leap Seconds=32
  { 2451949.50, 64.108779 }, //09 February 2001, UT1-UTC= 0.0752207, Accumulated Leap Seconds=32
  { 2451950.50, 64.109645 }, //10 February 2001, UT1-UTC= 0.0743550, Accumulated Leap Seconds=32
  { 2451951.50, 64.110687 }, //11 February 2001, UT1-UTC= 0.0733133, Accumulated Leap Seconds=32
  { 2451952.50, 64.111815 }, //12 February 2001, UT1-UTC= 0.0721847, Accumulated Leap Seconds=32
  { 2451953.50, 64.112962 }, //13 February 2001, UT1-UTC= 0.0710379, Accumulated Leap Seconds=32
  { 2451954.50, 64.114071 }, //14 February 2001, UT1-UTC= 0.0699293, Accumulated Leap Seconds=32
  { 2451955.50, 64.115051 }, //15 February 2001, UT1-UTC= 0.0689493, Accumulated Leap Seconds=32
  { 2451956.50, 64.115880 }, //16 February 2001, UT1-UTC= 0.0681196, Accumulated Leap Seconds=32
  { 2451957.50, 64.116584 }, //17 February 2001, UT1-UTC= 0.0674165, Accumulated Leap Seconds=32
  { 2451958.50, 64.117195 }, //18 February 2001, UT1-UTC= 0.0668053, Accumulated Leap Seconds=32
  { 2451959.50, 64.117778 }, //19 February 2001, UT1-UTC= 0.0662222, Accumulated Leap Seconds=32
  { 2451960.50, 64.118382 }, //20 February 2001, UT1-UTC= 0.0656180, Accumulated Leap Seconds=32
  { 2451961.50, 64.119031 }, //21 February 2001, UT1-UTC= 0.0649693, Accumulated Leap Seconds=32
  { 2451962.50, 64.119728 }, //22 February 2001, UT1-UTC= 0.0642715, Accumulated Leap Seconds=32
  { 2451963.50, 64.120526 }, //23 February 2001, UT1-UTC= 0.0634742, Accumulated Leap Seconds=32
  { 2451964.50, 64.121481 }, //24 February 2001, UT1-UTC= 0.0625189, Accumulated Leap Seconds=32
  { 2451965.50, 64.122613 }, //25 February 2001, UT1-UTC= 0.0613865, Accumulated Leap Seconds=32
  { 2451966.50, 64.123909 }, //26 February 2001, UT1-UTC= 0.0600910, Accumulated Leap Seconds=32
  { 2451967.50, 64.125321 }, //27 February 2001, UT1-UTC= 0.0586786, Accumulated Leap Seconds=32
  { 2451968.50, 64.126780 }, //28 February 2001, UT1-UTC= 0.0572195, Accumulated Leap Seconds=32
  { 2451969.50, 64.128213 }, //01 March 2001, UT1-UTC= 0.0557875, Accumulated Leap Seconds=32
  { 2451970.50, 64.129540 }, //02 March 2001, UT1-UTC= 0.0544602, Accumulated Leap Seconds=32
  { 2451971.50, 64.130718 }, //03 March 2001, UT1-UTC= 0.0532819, Accumulated Leap Seconds=32
  { 2451972.50, 64.131762 }, //04 March 2001, UT1-UTC= 0.0522376, Accumulated Leap Seconds=32
  { 2451973.50, 64.132717 }, //05 March 2001, UT1-UTC= 0.0512832, Accumulated Leap Seconds=32
  { 2451974.50, 64.133651 }, //06 March 2001, UT1-UTC= 0.0503495, Accumulated Leap Seconds=32
  { 2451975.50, 64.134656 }, //07 March 2001, UT1-UTC= 0.0493442, Accumulated Leap Seconds=32
  { 2451976.50, 64.135866 }, //08 March 2001, UT1-UTC= 0.0481341, Accumulated Leap Seconds=32
  { 2451977.50, 64.137299 }, //09 March 2001, UT1-UTC= 0.0467010, Accumulated Leap Seconds=32
  { 2451978.50, 64.138877 }, //10 March 2001, UT1-UTC= 0.0451229, Accumulated Leap Seconds=32
  { 2451979.50, 64.140523 }, //11 March 2001, UT1-UTC= 0.0434773, Accumulated Leap Seconds=32
  { 2451980.50, 64.142120 }, //12 March 2001, UT1-UTC= 0.0418802, Accumulated Leap Seconds=32
  { 2451981.50, 64.143563 }, //13 March 2001, UT1-UTC= 0.0404374, Accumulated Leap Seconds=32
  { 2451982.50, 64.144770 }, //14 March 2001, UT1-UTC= 0.0392295, Accumulated Leap Seconds=32
  { 2451983.50, 64.145705 }, //15 March 2001, UT1-UTC= 0.0382952, Accumulated Leap Seconds=32
  { 2451984.50, 64.146427 }, //16 March 2001, UT1-UTC= 0.0375726, Accumulated Leap Seconds=32
  { 2451985.50, 64.147011 }, //17 March 2001, UT1-UTC= 0.0369887, Accumulated Leap Seconds=32
  { 2451986.50, 64.147495 }, //18 March 2001, UT1-UTC= 0.0365050, Accumulated Leap Seconds=32
  { 2451987.50, 64.147921 }, //19 March 2001, UT1-UTC= 0.0360785, Accumulated Leap Seconds=32
  { 2451988.50, 64.148346 }, //20 March 2001, UT1-UTC= 0.0356539, Accumulated Leap Seconds=32
  { 2451989.50, 64.148832 }, //21 March 2001, UT1-UTC= 0.0351677, Accumulated Leap Seconds=32
  { 2451990.50, 64.149426 }, //22 March 2001, UT1-UTC= 0.0345744, Accumulated Leap Seconds=32
  { 2451991.50, 64.150119 }, //23 March 2001, UT1-UTC= 0.0338807, Accumulated Leap Seconds=32
  { 2451992.50, 64.150905 }, //24 March 2001, UT1-UTC= 0.0330948, Accumulated Leap Seconds=32
  { 2451993.50, 64.151827 }, //25 March 2001, UT1-UTC= 0.0321732, Accumulated Leap Seconds=32
  { 2451994.50, 64.152866 }, //26 March 2001, UT1-UTC= 0.0311344, Accumulated Leap Seconds=32
  { 2451995.50, 64.153964 }, //27 March 2001, UT1-UTC= 0.0300364, Accumulated Leap Seconds=32
  { 2451996.50, 64.155047 }, //28 March 2001, UT1-UTC= 0.0289533, Accumulated Leap Seconds=32
  { 2451997.50, 64.156057 }, //29 March 2001, UT1-UTC= 0.0279431, Accumulated Leap Seconds=32
  { 2451998.50, 64.156942 }, //30 March 2001, UT1-UTC= 0.0270584, Accumulated Leap Seconds=32
  { 2451999.50, 64.157721 }, //31 March 2001, UT1-UTC= 0.0262790, Accumulated Leap Seconds=32
  { 2452000.50, 64.158421 }, //01 April 2001, UT1-UTC= 0.0255789, Accumulated Leap Seconds=32
  { 2452001.50, 64.159091 }, //02 April 2001, UT1-UTC= 0.0249086, Accumulated Leap Seconds=32
  { 2452002.50, 64.159803 }, //03 April 2001, UT1-UTC= 0.0241974, Accumulated Leap Seconds=32
  { 2452003.50, 64.160647 }, //04 April 2001, UT1-UTC= 0.0233525, Accumulated Leap Seconds=32
  { 2452004.50, 64.161706 }, //05 April 2001, UT1-UTC= 0.0222941, Accumulated Leap Seconds=32
  { 2452005.50, 64.162969 }, //06 April 2001, UT1-UTC= 0.0210310, Accumulated Leap Seconds=32
  { 2452006.50, 64.164387 }, //07 April 2001, UT1-UTC= 0.0196128, Accumulated Leap Seconds=32
  { 2452007.50, 64.165882 }, //08 April 2001, UT1-UTC= 0.0181175, Accumulated Leap Seconds=32
  { 2452008.50, 64.167287 }, //09 April 2001, UT1-UTC= 0.0167131, Accumulated Leap Seconds=32
  { 2452009.50, 64.168485 }, //10 April 2001, UT1-UTC= 0.0155146, Accumulated Leap Seconds=32
  { 2452010.50, 64.169441 }, //11 April 2001, UT1-UTC= 0.0145593, Accumulated Leap Seconds=32
  { 2452011.50, 64.170161 }, //12 April 2001, UT1-UTC= 0.0138392, Accumulated Leap Seconds=32
  { 2452012.50, 64.170683 }, //13 April 2001, UT1-UTC= 0.0133170, Accumulated Leap Seconds=32
  { 2452013.50, 64.171052 }, //14 April 2001, UT1-UTC= 0.0129482, Accumulated Leap Seconds=32
  { 2452014.50, 64.171355 }, //15 April 2001, UT1-UTC= 0.0126452, Accumulated Leap Seconds=32
  { 2452015.50, 64.171691 }, //16 April 2001, UT1-UTC= 0.0123092, Accumulated Leap Seconds=32
  { 2452016.50, 64.172152 }, //17 April 2001, UT1-UTC= 0.0118484, Accumulated Leap Seconds=32
  { 2452017.50, 64.172773 }, //18 April 2001, UT1-UTC= 0.0112269, Accumulated Leap Seconds=32
  { 2452018.50, 64.173565 }, //19 April 2001, UT1-UTC= 0.0104350, Accumulated Leap Seconds=32
  { 2452019.50, 64.174471 }, //20 April 2001, UT1-UTC= 0.0095289, Accumulated Leap Seconds=32
  { 2452020.50, 64.175444 }, //21 April 2001, UT1-UTC= 0.0085563, Accumulated Leap Seconds=32
  { 2452021.50, 64.176448 }, //22 April 2001, UT1-UTC= 0.0075520, Accumulated Leap Seconds=32
  { 2452022.50, 64.177423 }, //23 April 2001, UT1-UTC= 0.0065774, Accumulated Leap Seconds=32
  { 2452023.50, 64.178329 }, //24 April 2001, UT1-UTC= 0.0056709, Accumulated Leap Seconds=32
  { 2452024.50, 64.179156 }, //25 April 2001, UT1-UTC= 0.0048440, Accumulated Leap Seconds=32
  { 2452025.50, 64.179921 }, //26 April 2001, UT1-UTC= 0.0040790, Accumulated Leap Seconds=32
  { 2452026.50, 64.180598 }, //27 April 2001, UT1-UTC= 0.0034022, Accumulated Leap Seconds=32
  { 2452027.50, 64.181181 }, //28 April 2001, UT1-UTC= 0.0028191, Accumulated Leap Seconds=32
  { 2452028.50, 64.181757 }, //29 April 2001, UT1-UTC= 0.0022428, Accumulated Leap Seconds=32
  { 2452029.50, 64.182427 }, //30 April 2001, UT1-UTC= 0.0015733, Accumulated Leap Seconds=32
  { 2452030.50, 64.183272 }, //01 May 2001, UT1-UTC= 0.0007278, Accumulated Leap Seconds=32
  { 2452031.50, 64.184326 }, //02 May 2001, UT1-UTC=-0.0003259, Accumulated Leap Seconds=32
  { 2452032.50, 64.185593 }, //03 May 2001, UT1-UTC=-0.0015934, Accumulated Leap Seconds=32
  { 2452033.50, 64.187035 }, //04 May 2001, UT1-UTC=-0.0030352, Accumulated Leap Seconds=32
  { 2452034.50, 64.188593 }, //05 May 2001, UT1-UTC=-0.0045933, Accumulated Leap Seconds=32
  { 2452035.50, 64.190133 }, //06 May 2001, UT1-UTC=-0.0061331, Accumulated Leap Seconds=32
  { 2452036.50, 64.191537 }, //07 May 2001, UT1-UTC=-0.0075369, Accumulated Leap Seconds=32
  { 2452037.50, 64.192740 }, //08 May 2001, UT1-UTC=-0.0087404, Accumulated Leap Seconds=32
  { 2452038.50, 64.193737 }, //09 May 2001, UT1-UTC=-0.0097369, Accumulated Leap Seconds=32
  { 2452039.50, 64.194532 }, //10 May 2001, UT1-UTC=-0.0105319, Accumulated Leap Seconds=32
  { 2452040.50, 64.195176 }, //11 May 2001, UT1-UTC=-0.0111764, Accumulated Leap Seconds=32
  { 2452041.50, 64.195750 }, //12 May 2001, UT1-UTC=-0.0117496, Accumulated Leap Seconds=32
  { 2452042.50, 64.196325 }, //13 May 2001, UT1-UTC=-0.0123247, Accumulated Leap Seconds=32
  { 2452043.50, 64.196961 }, //14 May 2001, UT1-UTC=-0.0129605, Accumulated Leap Seconds=32
  { 2452044.50, 64.197704 }, //15 May 2001, UT1-UTC=-0.0137043, Accumulated Leap Seconds=32
  { 2452045.50, 64.198569 }, //16 May 2001, UT1-UTC=-0.0145695, Accumulated Leap Seconds=32
  { 2452046.50, 64.199537 }, //17 May 2001, UT1-UTC=-0.0155375, Accumulated Leap Seconds=32
  { 2452047.50, 64.200601 }, //18 May 2001, UT1-UTC=-0.0166007, Accumulated Leap Seconds=32
  { 2452048.50, 64.201727 }, //19 May 2001, UT1-UTC=-0.0177270, Accumulated Leap Seconds=32
  { 2452049.50, 64.202863 }, //20 May 2001, UT1-UTC=-0.0188626, Accumulated Leap Seconds=32
  { 2452050.50, 64.203942 }, //21 May 2001, UT1-UTC=-0.0199416, Accumulated Leap Seconds=32
  { 2452051.50, 64.204903 }, //22 May 2001, UT1-UTC=-0.0209025, Accumulated Leap Seconds=32
  { 2452052.50, 64.205678 }, //23 May 2001, UT1-UTC=-0.0216782, Accumulated Leap Seconds=32
  { 2452053.50, 64.206237 }, //24 May 2001, UT1-UTC=-0.0222369, Accumulated Leap Seconds=32
  { 2452054.50, 64.206603 }, //25 May 2001, UT1-UTC=-0.0226032, Accumulated Leap Seconds=32
  { 2452055.50, 64.206824 }, //26 May 2001, UT1-UTC=-0.0228237, Accumulated Leap Seconds=32
  { 2452056.50, 64.207003 }, //27 May 2001, UT1-UTC=-0.0230030, Accumulated Leap Seconds=32
  { 2452057.50, 64.207246 }, //28 May 2001, UT1-UTC=-0.0232464, Accumulated Leap Seconds=32
  { 2452058.50, 64.207618 }, //29 May 2001, UT1-UTC=-0.0236177, Accumulated Leap Seconds=32
  { 2452059.50, 64.208132 }, //30 May 2001, UT1-UTC=-0.0241318, Accumulated Leap Seconds=32
  { 2452060.50, 64.208748 }, //31 May 2001, UT1-UTC=-0.0247483, Accumulated Leap Seconds=32
  { 2452061.50, 64.209397 }, //01 June 2001, UT1-UTC=-0.0253975, Accumulated Leap Seconds=32
  { 2452062.50, 64.210017 }, //02 June 2001, UT1-UTC=-0.0260175, Accumulated Leap Seconds=32
  { 2452063.50, 64.210518 }, //03 June 2001, UT1-UTC=-0.0265184, Accumulated Leap Seconds=32
  { 2452064.50, 64.210828 }, //04 June 2001, UT1-UTC=-0.0268285, Accumulated Leap Seconds=32
  { 2452065.50, 64.210912 }, //05 June 2001, UT1-UTC=-0.0269122, Accumulated Leap Seconds=32
  { 2452066.50, 64.210759 }, //06 June 2001, UT1-UTC=-0.0267590, Accumulated Leap Seconds=32
  { 2452067.50, 64.210402 }, //07 June 2001, UT1-UTC=-0.0264018, Accumulated Leap Seconds=32
  { 2452068.50, 64.209929 }, //08 June 2001, UT1-UTC=-0.0259287, Accumulated Leap Seconds=32
  { 2452069.50, 64.209433 }, //09 June 2001, UT1-UTC=-0.0254333, Accumulated Leap Seconds=32
  { 2452070.50, 64.209001 }, //10 June 2001, UT1-UTC=-0.0250007, Accumulated Leap Seconds=32
  { 2452071.50, 64.208689 }, //11 June 2001, UT1-UTC=-0.0246893, Accumulated Leap Seconds=32
  { 2452072.50, 64.208524 }, //12 June 2001, UT1-UTC=-0.0245243, Accumulated Leap Seconds=32
  { 2452073.50, 64.208500 }, //13 June 2001, UT1-UTC=-0.0244996, Accumulated Leap Seconds=32
  { 2452074.50, 64.208585 }, //14 June 2001, UT1-UTC=-0.0245855, Accumulated Leap Seconds=32
  { 2452075.50, 64.208715 }, //15 June 2001, UT1-UTC=-0.0247149, Accumulated Leap Seconds=32
  { 2452076.50, 64.208884 }, //16 June 2001, UT1-UTC=-0.0248844, Accumulated Leap Seconds=32
  { 2452077.50, 64.209086 }, //17 June 2001, UT1-UTC=-0.0250857, Accumulated Leap Seconds=32
  { 2452078.50, 64.209277 }, //18 June 2001, UT1-UTC=-0.0252770, Accumulated Leap Seconds=32
  { 2452079.50, 64.209409 }, //19 June 2001, UT1-UTC=-0.0254095, Accumulated Leap Seconds=32
  { 2452080.50, 64.209432 }, //20 June 2001, UT1-UTC=-0.0254321, Accumulated Leap Seconds=32
  { 2452081.50, 64.209332 }, //21 June 2001, UT1-UTC=-0.0253320, Accumulated Leap Seconds=32
  { 2452082.50, 64.209178 }, //22 June 2001, UT1-UTC=-0.0251783, Accumulated Leap Seconds=32
  { 2452083.50, 64.209010 }, //23 June 2001, UT1-UTC=-0.0250102, Accumulated Leap Seconds=32
  { 2452084.50, 64.208897 }, //24 June 2001, UT1-UTC=-0.0248971, Accumulated Leap Seconds=32
  { 2452085.50, 64.208920 }, //25 June 2001, UT1-UTC=-0.0249203, Accumulated Leap Seconds=32
  { 2452086.50, 64.209157 }, //26 June 2001, UT1-UTC=-0.0251571, Accumulated Leap Seconds=32
  { 2452087.50, 64.209622 }, //27 June 2001, UT1-UTC=-0.0256219, Accumulated Leap Seconds=32
  { 2452088.50, 64.210234 }, //28 June 2001, UT1-UTC=-0.0262345, Accumulated Leap Seconds=32
  { 2452089.50, 64.210837 }, //29 June 2001, UT1-UTC=-0.0268374, Accumulated Leap Seconds=32
  { 2452090.50, 64.211331 }, //30 June 2001, UT1-UTC=-0.0273305, Accumulated Leap Seconds=32
  { 2452091.50, 64.211663 }, //01 July 2001, UT1-UTC=-0.0276628, Accumulated Leap Seconds=32
  { 2452092.50, 64.211807 }, //02 July 2001, UT1-UTC=-0.0278068, Accumulated Leap Seconds=32
  { 2452093.50, 64.211772 }, //03 July 2001, UT1-UTC=-0.0277724, Accumulated Leap Seconds=32
  { 2452094.50, 64.211592 }, //04 July 2001, UT1-UTC=-0.0275917, Accumulated Leap Seconds=32
  { 2452095.50, 64.211321 }, //05 July 2001, UT1-UTC=-0.0273214, Accumulated Leap Seconds=32
  { 2452096.50, 64.211017 }, //06 July 2001, UT1-UTC=-0.0270168, Accumulated Leap Seconds=32
  { 2452097.50, 64.210717 }, //07 July 2001, UT1-UTC=-0.0267172, Accumulated Leap Seconds=32
  { 2452098.50, 64.210488 }, //08 July 2001, UT1-UTC=-0.0264876, Accumulated Leap Seconds=32
  { 2452099.50, 64.210356 }, //09 July 2001, UT1-UTC=-0.0263559, Accumulated Leap Seconds=32
  { 2452100.50, 64.210296 }, //10 July 2001, UT1-UTC=-0.0262957, Accumulated Leap Seconds=32
  { 2452101.50, 64.210294 }, //11 July 2001, UT1-UTC=-0.0262937, Accumulated Leap Seconds=32
  { 2452102.50, 64.210342 }, //12 July 2001, UT1-UTC=-0.0263420, Accumulated Leap Seconds=32
  { 2452103.50, 64.210442 }, //13 July 2001, UT1-UTC=-0.0264423, Accumulated Leap Seconds=32
  { 2452104.50, 64.210574 }, //14 July 2001, UT1-UTC=-0.0265745, Accumulated Leap Seconds=32
  { 2452105.50, 64.210678 }, //15 July 2001, UT1-UTC=-0.0266776, Accumulated Leap Seconds=32
  { 2452106.50, 64.210700 }, //16 July 2001, UT1-UTC=-0.0266996, Accumulated Leap Seconds=32
  { 2452107.50, 64.210605 }, //17 July 2001, UT1-UTC=-0.0266054, Accumulated Leap Seconds=32
  { 2452108.50, 64.210377 }, //18 July 2001, UT1-UTC=-0.0263767, Accumulated Leap Seconds=32
  { 2452109.50, 64.209994 }, //19 July 2001, UT1-UTC=-0.0259942, Accumulated Leap Seconds=32
  { 2452110.50, 64.209508 }, //20 July 2001, UT1-UTC=-0.0255082, Accumulated Leap Seconds=32
  { 2452111.50, 64.209006 }, //21 July 2001, UT1-UTC=-0.0250060, Accumulated Leap Seconds=32
  { 2452112.50, 64.208613 }, //22 July 2001, UT1-UTC=-0.0246132, Accumulated Leap Seconds=32
  { 2452113.50, 64.208417 }, //23 July 2001, UT1-UTC=-0.0244171, Accumulated Leap Seconds=32
  { 2452114.50, 64.208438 }, //24 July 2001, UT1-UTC=-0.0244378, Accumulated Leap Seconds=32
  { 2452115.50, 64.208625 }, //25 July 2001, UT1-UTC=-0.0246254, Accumulated Leap Seconds=32
  { 2452116.50, 64.208855 }, //26 July 2001, UT1-UTC=-0.0248554, Accumulated Leap Seconds=32
  { 2452117.50, 64.209036 }, //27 July 2001, UT1-UTC=-0.0250362, Accumulated Leap Seconds=32
  { 2452118.50, 64.209071 }, //28 July 2001, UT1-UTC=-0.0250710, Accumulated Leap Seconds=32
  { 2452119.50, 64.208897 }, //29 July 2001, UT1-UTC=-0.0248973, Accumulated Leap Seconds=32
  { 2452120.50, 64.208520 }, //30 July 2001, UT1-UTC=-0.0245203, Accumulated Leap Seconds=32
  { 2452121.50, 64.207975 }, //31 July 2001, UT1-UTC=-0.0239750, Accumulated Leap Seconds=32
  { 2452122.50, 64.207317 }, //01 August 2001, UT1-UTC=-0.0233173, Accumulated Leap Seconds=32
  { 2452123.50, 64.206603 }, //02 August 2001, UT1-UTC=-0.0226031, Accumulated Leap Seconds=32
  { 2452124.50, 64.205931 }, //03 August 2001, UT1-UTC=-0.0219307, Accumulated Leap Seconds=32
  { 2452125.50, 64.205374 }, //04 August 2001, UT1-UTC=-0.0213741, Accumulated Leap Seconds=32
  { 2452126.50, 64.204965 }, //05 August 2001, UT1-UTC=-0.0209647, Accumulated Leap Seconds=32
  { 2452127.50, 64.204713 }, //06 August 2001, UT1-UTC=-0.0207134, Accumulated Leap Seconds=32
  { 2452128.50, 64.204609 }, //07 August 2001, UT1-UTC=-0.0206093, Accumulated Leap Seconds=32
  { 2452129.50, 64.204645 }, //08 August 2001, UT1-UTC=-0.0206453, Accumulated Leap Seconds=32
  { 2452130.50, 64.204828 }, //09 August 2001, UT1-UTC=-0.0208280, Accumulated Leap Seconds=32
  { 2452131.50, 64.205118 }, //10 August 2001, UT1-UTC=-0.0211180, Accumulated Leap Seconds=32
  { 2452132.50, 64.205488 }, //11 August 2001, UT1-UTC=-0.0214883, Accumulated Leap Seconds=32
  { 2452133.50, 64.205858 }, //12 August 2001, UT1-UTC=-0.0218582, Accumulated Leap Seconds=32
  { 2452134.50, 64.206168 }, //13 August 2001, UT1-UTC=-0.0221683, Accumulated Leap Seconds=32
  { 2452135.50, 64.206376 }, //14 August 2001, UT1-UTC=-0.0223759, Accumulated Leap Seconds=32
  { 2452136.50, 64.206472 }, //15 August 2001, UT1-UTC=-0.0224719, Accumulated Leap Seconds=32
  { 2452137.50, 64.206495 }, //16 August 2001, UT1-UTC=-0.0224951, Accumulated Leap Seconds=32
  { 2452138.50, 64.206542 }, //17 August 2001, UT1-UTC=-0.0225424, Accumulated Leap Seconds=32
  { 2452139.50, 64.206735 }, //18 August 2001, UT1-UTC=-0.0227349, Accumulated Leap Seconds=32
  { 2452140.50, 64.207147 }, //19 August 2001, UT1-UTC=-0.0231468, Accumulated Leap Seconds=32
  { 2452141.50, 64.207811 }, //20 August 2001, UT1-UTC=-0.0238110, Accumulated Leap Seconds=32
  { 2452142.50, 64.208672 }, //21 August 2001, UT1-UTC=-0.0246716, Accumulated Leap Seconds=32
  { 2452143.50, 64.209634 }, //22 August 2001, UT1-UTC=-0.0256336, Accumulated Leap Seconds=32
  { 2452144.50, 64.210572 }, //23 August 2001, UT1-UTC=-0.0265722, Accumulated Leap Seconds=32
  { 2452145.50, 64.211331 }, //24 August 2001, UT1-UTC=-0.0273305, Accumulated Leap Seconds=32
  { 2452146.50, 64.211886 }, //25 August 2001, UT1-UTC=-0.0278859, Accumulated Leap Seconds=32
  { 2452147.50, 64.212191 }, //26 August 2001, UT1-UTC=-0.0281912, Accumulated Leap Seconds=32
  { 2452148.50, 64.212270 }, //27 August 2001, UT1-UTC=-0.0282702, Accumulated Leap Seconds=32
  { 2452149.50, 64.212181 }, //28 August 2001, UT1-UTC=-0.0281810, Accumulated Leap Seconds=32
  { 2452150.50, 64.211987 }, //29 August 2001, UT1-UTC=-0.0279875, Accumulated Leap Seconds=32
  { 2452151.50, 64.211770 }, //30 August 2001, UT1-UTC=-0.0277699, Accumulated Leap Seconds=32
  { 2452152.50, 64.211625 }, //31 August 2001, UT1-UTC=-0.0276246, Accumulated Leap Seconds=32
  { 2452153.50, 64.211557 }, //01 September 2001, UT1-UTC=-0.0275565, Accumulated Leap Seconds=32
  { 2452154.50, 64.211624 }, //02 September 2001, UT1-UTC=-0.0276237, Accumulated Leap Seconds=32
  { 2452155.50, 64.211846 }, //03 September 2001, UT1-UTC=-0.0278457, Accumulated Leap Seconds=32
  { 2452156.50, 64.212210 }, //04 September 2001, UT1-UTC=-0.0282101, Accumulated Leap Seconds=32
  { 2452157.50, 64.212679 }, //05 September 2001, UT1-UTC=-0.0286795, Accumulated Leap Seconds=32
  { 2452158.50, 64.213222 }, //06 September 2001, UT1-UTC=-0.0292220, Accumulated Leap Seconds=32
  { 2452159.50, 64.213723 }, //07 September 2001, UT1-UTC=-0.0297228, Accumulated Leap Seconds=32
  { 2452160.50, 64.214125 }, //08 September 2001, UT1-UTC=-0.0301251, Accumulated Leap Seconds=32
  { 2452161.50, 64.214399 }, //09 September 2001, UT1-UTC=-0.0303991, Accumulated Leap Seconds=32
  { 2452162.50, 64.214541 }, //10 September 2001, UT1-UTC=-0.0305407, Accumulated Leap Seconds=32
  { 2452163.50, 64.214558 }, //11 September 2001, UT1-UTC=-0.0305585, Accumulated Leap Seconds=32
  { 2452164.50, 64.214489 }, //12 September 2001, UT1-UTC=-0.0304893, Accumulated Leap Seconds=32
  { 2452165.50, 64.214390 }, //13 September 2001, UT1-UTC=-0.0303902, Accumulated Leap Seconds=32
  { 2452166.50, 64.214377 }, //14 September 2001, UT1-UTC=-0.0303768, Accumulated Leap Seconds=32
  { 2452167.50, 64.214559 }, //15 September 2001, UT1-UTC=-0.0305592, Accumulated Leap Seconds=32
  { 2452168.50, 64.215023 }, //16 September 2001, UT1-UTC=-0.0310227, Accumulated Leap Seconds=32
  { 2452169.50, 64.215772 }, //17 September 2001, UT1-UTC=-0.0317716, Accumulated Leap Seconds=32
  { 2452170.50, 64.216708 }, //18 September 2001, UT1-UTC=-0.0327081, Accumulated Leap Seconds=32
  { 2452171.50, 64.217683 }, //19 September 2001, UT1-UTC=-0.0336833, Accumulated Leap Seconds=32
  { 2452172.50, 64.218578 }, //20 September 2001, UT1-UTC=-0.0345782, Accumulated Leap Seconds=32
  { 2452173.50, 64.219319 }, //21 September 2001, UT1-UTC=-0.0353188, Accumulated Leap Seconds=32
  { 2452174.50, 64.219865 }, //22 September 2001, UT1-UTC=-0.0358650, Accumulated Leap Seconds=32
  { 2452175.50, 64.220229 }, //23 September 2001, UT1-UTC=-0.0362289, Accumulated Leap Seconds=32
  { 2452176.50, 64.220432 }, //24 September 2001, UT1-UTC=-0.0364315, Accumulated Leap Seconds=32
  { 2452177.50, 64.220515 }, //25 September 2001, UT1-UTC=-0.0365148, Accumulated Leap Seconds=32
  { 2452178.50, 64.220572 }, //26 September 2001, UT1-UTC=-0.0365719, Accumulated Leap Seconds=32
  { 2452179.50, 64.220665 }, //27 September 2001, UT1-UTC=-0.0366649, Accumulated Leap Seconds=32
  { 2452180.50, 64.220863 }, //28 September 2001, UT1-UTC=-0.0368626, Accumulated Leap Seconds=32
  { 2452181.50, 64.221207 }, //29 September 2001, UT1-UTC=-0.0372070, Accumulated Leap Seconds=32
  { 2452182.50, 64.221682 }, //30 September 2001, UT1-UTC=-0.0376819, Accumulated Leap Seconds=32
  { 2452183.50, 64.222286 }, //01 October 2001, UT1-UTC=-0.0382858, Accumulated Leap Seconds=32
  { 2452184.50, 64.223027 }, //02 October 2001, UT1-UTC=-0.0390267, Accumulated Leap Seconds=32
  { 2452185.50, 64.223884 }, //03 October 2001, UT1-UTC=-0.0398842, Accumulated Leap Seconds=32
  { 2452186.50, 64.224766 }, //04 October 2001, UT1-UTC=-0.0407664, Accumulated Leap Seconds=32
  { 2452187.50, 64.225660 }, //05 October 2001, UT1-UTC=-0.0416597, Accumulated Leap Seconds=32
  { 2452188.50, 64.226507 }, //06 October 2001, UT1-UTC=-0.0425070, Accumulated Leap Seconds=32
  { 2452189.50, 64.227221 }, //07 October 2001, UT1-UTC=-0.0432212, Accumulated Leap Seconds=32
  { 2452190.50, 64.227809 }, //08 October 2001, UT1-UTC=-0.0438093, Accumulated Leap Seconds=32
  { 2452191.50, 64.228312 }, //09 October 2001, UT1-UTC=-0.0443119, Accumulated Leap Seconds=32
  { 2452192.50, 64.228795 }, //10 October 2001, UT1-UTC=-0.0447946, Accumulated Leap Seconds=32
  { 2452193.50, 64.229347 }, //11 October 2001, UT1-UTC=-0.0453471, Accumulated Leap Seconds=32
  { 2452194.50, 64.230039 }, //12 October 2001, UT1-UTC=-0.0460388, Accumulated Leap Seconds=32
  { 2452195.50, 64.230935 }, //13 October 2001, UT1-UTC=-0.0469352, Accumulated Leap Seconds=32
  { 2452196.50, 64.232105 }, //14 October 2001, UT1-UTC=-0.0481049, Accumulated Leap Seconds=32
  { 2452197.50, 64.233525 }, //15 October 2001, UT1-UTC=-0.0495247, Accumulated Leap Seconds=32
  { 2452198.50, 64.235092 }, //16 October 2001, UT1-UTC=-0.0510923, Accumulated Leap Seconds=32
  { 2452199.50, 64.236676 }, //17 October 2001, UT1-UTC=-0.0526756, Accumulated Leap Seconds=32
  { 2452200.50, 64.238138 }, //18 October 2001, UT1-UTC=-0.0541385, Accumulated Leap Seconds=32
  { 2452201.50, 64.239364 }, //19 October 2001, UT1-UTC=-0.0553641, Accumulated Leap Seconds=32
  { 2452202.50, 64.240310 }, //20 October 2001, UT1-UTC=-0.0563100, Accumulated Leap Seconds=32
  { 2452203.50, 64.241017 }, //21 October 2001, UT1-UTC=-0.0570173, Accumulated Leap Seconds=32
  { 2452204.50, 64.241586 }, //22 October 2001, UT1-UTC=-0.0575865, Accumulated Leap Seconds=32
  { 2452205.50, 64.242094 }, //23 October 2001, UT1-UTC=-0.0580938, Accumulated Leap Seconds=32
  { 2452206.50, 64.242603 }, //24 October 2001, UT1-UTC=-0.0586034, Accumulated Leap Seconds=32
  { 2452207.50, 64.243148 }, //25 October 2001, UT1-UTC=-0.0591482, Accumulated Leap Seconds=32
  { 2452208.50, 64.243806 }, //26 October 2001, UT1-UTC=-0.0598058, Accumulated Leap Seconds=32
  { 2452209.50, 64.244627 }, //27 October 2001, UT1-UTC=-0.0606271, Accumulated Leap Seconds=32
  { 2452210.50, 64.245600 }, //28 October 2001, UT1-UTC=-0.0616005, Accumulated Leap Seconds=32
  { 2452211.50, 64.246673 }, //29 October 2001, UT1-UTC=-0.0626731, Accumulated Leap Seconds=32
  { 2452212.50, 64.247794 }, //30 October 2001, UT1-UTC=-0.0637941, Accumulated Leap Seconds=32
  { 2452213.50, 64.248900 }, //31 October 2001, UT1-UTC=-0.0649004, Accumulated Leap Seconds=32
  { 2452214.50, 64.249962 }, //01 November 2001, UT1-UTC=-0.0659625, Accumulated Leap Seconds=32
  { 2452215.50, 64.250911 }, //02 November 2001, UT1-UTC=-0.0669114, Accumulated Leap Seconds=32
  { 2452216.50, 64.251679 }, //03 November 2001, UT1-UTC=-0.0676788, Accumulated Leap Seconds=32
  { 2452217.50, 64.252309 }, //04 November 2001, UT1-UTC=-0.0683087, Accumulated Leap Seconds=32
  { 2452218.50, 64.252848 }, //05 November 2001, UT1-UTC=-0.0688485, Accumulated Leap Seconds=32
  { 2452219.50, 64.253338 }, //06 November 2001, UT1-UTC=-0.0693383, Accumulated Leap Seconds=32
  { 2452220.50, 64.253818 }, //07 November 2001, UT1-UTC=-0.0698185, Accumulated Leap Seconds=32
  { 2452221.50, 64.254392 }, //08 November 2001, UT1-UTC=-0.0703925, Accumulated Leap Seconds=32
  { 2452222.50, 64.255211 }, //09 November 2001, UT1-UTC=-0.0712107, Accumulated Leap Seconds=32
  { 2452223.50, 64.256345 }, //10 November 2001, UT1-UTC=-0.0723447, Accumulated Leap Seconds=32
  { 2452224.50, 64.257715 }, //11 November 2001, UT1-UTC=-0.0737149, Accumulated Leap Seconds=32
  { 2452225.50, 64.259240 }, //12 November 2001, UT1-UTC=-0.0752402, Accumulated Leap Seconds=32
  { 2452226.50, 64.260824 }, //13 November 2001, UT1-UTC=-0.0768242, Accumulated Leap Seconds=32
  { 2452227.50, 64.262360 }, //14 November 2001, UT1-UTC=-0.0783598, Accumulated Leap Seconds=32
  { 2452228.50, 64.263680 }, //15 November 2001, UT1-UTC=-0.0796801, Accumulated Leap Seconds=32
  { 2452229.50, 64.264715 }, //16 November 2001, UT1-UTC=-0.0807152, Accumulated Leap Seconds=32
  { 2452230.50, 64.265530 }, //17 November 2001, UT1-UTC=-0.0815299, Accumulated Leap Seconds=32
  { 2452231.50, 64.266187 }, //18 November 2001, UT1-UTC=-0.0821868, Accumulated Leap Seconds=32
  { 2452232.50, 64.266757 }, //19 November 2001, UT1-UTC=-0.0827571, Accumulated Leap Seconds=32
  { 2452233.50, 64.267304 }, //20 November 2001, UT1-UTC=-0.0833038, Accumulated Leap Seconds=32
  { 2452234.50, 64.267864 }, //21 November 2001, UT1-UTC=-0.0838642, Accumulated Leap Seconds=32
  { 2452235.50, 64.268476 }, //22 November 2001, UT1-UTC=-0.0844760, Accumulated Leap Seconds=32
  { 2452236.50, 64.269189 }, //23 November 2001, UT1-UTC=-0.0851887, Accumulated Leap Seconds=32
  { 2452237.50, 64.269991 }, //24 November 2001, UT1-UTC=-0.0859913, Accumulated Leap Seconds=32
  { 2452238.50, 64.270876 }, //25 November 2001, UT1-UTC=-0.0868759, Accumulated Leap Seconds=32
  { 2452239.50, 64.271832 }, //26 November 2001, UT1-UTC=-0.0878323, Accumulated Leap Seconds=32
  { 2452240.50, 64.272823 }, //27 November 2001, UT1-UTC=-0.0888226, Accumulated Leap Seconds=32
  { 2452241.50, 64.273793 }, //28 November 2001, UT1-UTC=-0.0897932, Accumulated Leap Seconds=32
  { 2452242.50, 64.274684 }, //29 November 2001, UT1-UTC=-0.0906844, Accumulated Leap Seconds=32
  { 2452243.50, 64.275463 }, //30 November 2001, UT1-UTC=-0.0914634, Accumulated Leap Seconds=32
  { 2452244.50, 64.276097 }, //01 December 2001, UT1-UTC=-0.0920973, Accumulated Leap Seconds=32
  { 2452245.50, 64.276539 }, //02 December 2001, UT1-UTC=-0.0925390, Accumulated Leap Seconds=32
  { 2452246.50, 64.276848 }, //03 December 2001, UT1-UTC=-0.0928478, Accumulated Leap Seconds=32
  { 2452247.50, 64.277111 }, //04 December 2001, UT1-UTC=-0.0931113, Accumulated Leap Seconds=32
  { 2452248.50, 64.277407 }, //05 December 2001, UT1-UTC=-0.0934066, Accumulated Leap Seconds=32
  { 2452249.50, 64.277841 }, //06 December 2001, UT1-UTC=-0.0938409, Accumulated Leap Seconds=32
  { 2452250.50, 64.278481 }, //07 December 2001, UT1-UTC=-0.0944807, Accumulated Leap Seconds=32
  { 2452251.50, 64.279334 }, //08 December 2001, UT1-UTC=-0.0953339, Accumulated Leap Seconds=32
  { 2452252.50, 64.280386 }, //09 December 2001, UT1-UTC=-0.0963856, Accumulated Leap Seconds=32
  { 2452253.50, 64.281564 }, //10 December 2001, UT1-UTC=-0.0975643, Accumulated Leap Seconds=32
  { 2452254.50, 64.282748 }, //11 December 2001, UT1-UTC=-0.0987479, Accumulated Leap Seconds=32
  { 2452255.50, 64.283813 }, //12 December 2001, UT1-UTC=-0.0998134, Accumulated Leap Seconds=32
  { 2452256.50, 64.284708 }, //13 December 2001, UT1-UTC=-0.1007082, Accumulated Leap Seconds=32
  { 2452257.50, 64.285406 }, //14 December 2001, UT1-UTC=-0.1014059, Accumulated Leap Seconds=32
  { 2452258.50, 64.285886 }, //15 December 2001, UT1-UTC=-0.1018858, Accumulated Leap Seconds=32
  { 2452259.50, 64.286220 }, //16 December 2001, UT1-UTC=-0.1022203, Accumulated Leap Seconds=32
  { 2452260.50, 64.286528 }, //17 December 2001, UT1-UTC=-0.1025281, Accumulated Leap Seconds=32
  { 2452261.50, 64.286898 }, //18 December 2001, UT1-UTC=-0.1028980, Accumulated Leap Seconds=32
  { 2452262.50, 64.287369 }, //19 December 2001, UT1-UTC=-0.1033688, Accumulated Leap Seconds=32
  { 2452263.50, 64.287952 }, //20 December 2001, UT1-UTC=-0.1039517, Accumulated Leap Seconds=32
  { 2452264.50, 64.288741 }, //21 December 2001, UT1-UTC=-0.1047412, Accumulated Leap Seconds=32
  { 2452265.50, 64.289734 }, //22 December 2001, UT1-UTC=-0.1057340, Accumulated Leap Seconds=32
  { 2452266.50, 64.290832 }, //23 December 2001, UT1-UTC=-0.1068325, Accumulated Leap Seconds=32
  { 2452267.50, 64.291993 }, //24 December 2001, UT1-UTC=-0.1079933, Accumulated Leap Seconds=32
  { 2452268.50, 64.293180 }, //25 December 2001, UT1-UTC=-0.1091801, Accumulated Leap Seconds=32
  { 2452269.50, 64.294355 }, //26 December 2001, UT1-UTC=-0.1103550, Accumulated Leap Seconds=32
  { 2452270.50, 64.295483 }, //27 December 2001, UT1-UTC=-0.1114831, Accumulated Leap Seconds=32
  { 2452271.50, 64.296525 }, //28 December 2001, UT1-UTC=-0.1125251, Accumulated Leap Seconds=32
  { 2452272.50, 64.297444 }, //29 December 2001, UT1-UTC=-0.1134437, Accumulated Leap Seconds=32
  { 2452273.50, 64.298256 }, //30 December 2001, UT1-UTC=-0.1142562, Accumulated Leap Seconds=32
  { 2452274.50, 64.299021 }, //31 December 2001, UT1-UTC=-0.1150209, Accumulated Leap Seconds=32
  { 2452275.50, 64.299804 }, //01 January 2002, UT1-UTC=-0.1158037, Accumulated Leap Seconds=32
  { 2452276.50, 64.300669 }, //02 January 2002, UT1-UTC=-0.1166693, Accumulated Leap Seconds=32
  { 2452277.50, 64.301699 }, //03 January 2002, UT1-UTC=-0.1176985, Accumulated Leap Seconds=32
  { 2452278.50, 64.302899 }, //04 January 2002, UT1-UTC=-0.1188986, Accumulated Leap Seconds=32
  { 2452279.50, 64.304211 }, //05 January 2002, UT1-UTC=-0.1202113, Accumulated Leap Seconds=32
  { 2452280.50, 64.305547 }, //06 January 2002, UT1-UTC=-0.1215473, Accumulated Leap Seconds=32
  { 2452281.50, 64.306837 }, //07 January 2002, UT1-UTC=-0.1228365, Accumulated Leap Seconds=32
  { 2452282.50, 64.307961 }, //08 January 2002, UT1-UTC=-0.1239610, Accumulated Leap Seconds=32
  { 2452283.50, 64.308807 }, //09 January 2002, UT1-UTC=-0.1248067, Accumulated Leap Seconds=32
  { 2452284.50, 64.309401 }, //10 January 2002, UT1-UTC=-0.1254012, Accumulated Leap Seconds=32
  { 2452285.50, 64.309789 }, //11 January 2002, UT1-UTC=-0.1257888, Accumulated Leap Seconds=32
  { 2452286.50, 64.309999 }, //12 January 2002, UT1-UTC=-0.1259989, Accumulated Leap Seconds=32
  { 2452287.50, 64.310059 }, //13 January 2002, UT1-UTC=-0.1260590, Accumulated Leap Seconds=32
  { 2452288.50, 64.310049 }, //14 January 2002, UT1-UTC=-0.1260491, Accumulated Leap Seconds=32
  { 2452289.50, 64.310053 }, //15 January 2002, UT1-UTC=-0.1260530, Accumulated Leap Seconds=32
  { 2452290.50, 64.310143 }, //16 January 2002, UT1-UTC=-0.1261427, Accumulated Leap Seconds=32
  { 2452291.50, 64.310307 }, //17 January 2002, UT1-UTC=-0.1263067, Accumulated Leap Seconds=32
  { 2452292.50, 64.310579 }, //18 January 2002, UT1-UTC=-0.1265787, Accumulated Leap Seconds=32
  { 2452293.50, 64.311012 }, //19 January 2002, UT1-UTC=-0.1270116, Accumulated Leap Seconds=32
  { 2452294.50, 64.311535 }, //20 January 2002, UT1-UTC=-0.1275351, Accumulated Leap Seconds=32
  { 2452295.50, 64.312105 }, //21 January 2002, UT1-UTC=-0.1281046, Accumulated Leap Seconds=32
  { 2452296.50, 64.312701 }, //22 January 2002, UT1-UTC=-0.1287014, Accumulated Leap Seconds=32
  { 2452297.50, 64.313301 }, //23 January 2002, UT1-UTC=-0.1293014, Accumulated Leap Seconds=32
  { 2452298.50, 64.313855 }, //24 January 2002, UT1-UTC=-0.1298552, Accumulated Leap Seconds=32
  { 2452299.50, 64.314299 }, //25 January 2002, UT1-UTC=-0.1302989, Accumulated Leap Seconds=32
  { 2452300.50, 64.314633 }, //26 January 2002, UT1-UTC=-0.1306333, Accumulated Leap Seconds=32
  { 2452301.50, 64.314940 }, //27 January 2002, UT1-UTC=-0.1309397, Accumulated Leap Seconds=32
  { 2452302.50, 64.315327 }, //28 January 2002, UT1-UTC=-0.1313269, Accumulated Leap Seconds=32
  { 2452303.50, 64.315902 }, //29 January 2002, UT1-UTC=-0.1319022, Accumulated Leap Seconds=32
  { 2452304.50, 64.316738 }, //30 January 2002, UT1-UTC=-0.1327378, Accumulated Leap Seconds=32
  { 2452305.50, 64.317868 }, //31 January 2002, UT1-UTC=-0.1338676, Accumulated Leap Seconds=32
  { 2452306.50, 64.319186 }, //01 February 2002, UT1-UTC=-0.1351858, Accumulated Leap Seconds=32
  { 2452307.50, 64.320644 }, //02 February 2002, UT1-UTC=-0.1366441, Accumulated Leap Seconds=32
  { 2452308.50, 64.322144 }, //03 February 2002, UT1-UTC=-0.1381439, Accumulated Leap Seconds=32
  { 2452309.50, 64.323521 }, //04 February 2002, UT1-UTC=-0.1395213, Accumulated Leap Seconds=32
  { 2452310.50, 64.324717 }, //05 February 2002, UT1-UTC=-0.1407169, Accumulated Leap Seconds=32
  { 2452311.50, 64.325709 }, //06 February 2002, UT1-UTC=-0.1417091, Accumulated Leap Seconds=32
  { 2452312.50, 64.326448 }, //07 February 2002, UT1-UTC=-0.1424476, Accumulated Leap Seconds=32
  { 2452313.50, 64.327016 }, //08 February 2002, UT1-UTC=-0.1430156, Accumulated Leap Seconds=32
  { 2452314.50, 64.327440 }, //09 February 2002, UT1-UTC=-0.1434398, Accumulated Leap Seconds=32
  { 2452315.50, 64.327815 }, //10 February 2002, UT1-UTC=-0.1438149, Accumulated Leap Seconds=32
  { 2452316.50, 64.328249 }, //11 February 2002, UT1-UTC=-0.1442491, Accumulated Leap Seconds=32
  { 2452317.50, 64.328793 }, //12 February 2002, UT1-UTC=-0.1447929, Accumulated Leap Seconds=32
  { 2452318.50, 64.329416 }, //13 February 2002, UT1-UTC=-0.1454163, Accumulated Leap Seconds=32
  { 2452319.50, 64.330148 }, //14 February 2002, UT1-UTC=-0.1461477, Accumulated Leap Seconds=32
  { 2452320.50, 64.330980 }, //15 February 2002, UT1-UTC=-0.1469803, Accumulated Leap Seconds=32
  { 2452321.50, 64.331911 }, //16 February 2002, UT1-UTC=-0.1479109, Accumulated Leap Seconds=32
  { 2452322.50, 64.332942 }, //17 February 2002, UT1-UTC=-0.1489420, Accumulated Leap Seconds=32
  { 2452323.50, 64.334009 }, //18 February 2002, UT1-UTC=-0.1500088, Accumulated Leap Seconds=32
  { 2452324.50, 64.335041 }, //19 February 2002, UT1-UTC=-0.1510412, Accumulated Leap Seconds=32
  { 2452325.50, 64.336000 }, //20 February 2002, UT1-UTC=-0.1519999, Accumulated Leap Seconds=32
  { 2452326.50, 64.336876 }, //21 February 2002, UT1-UTC=-0.1528758, Accumulated Leap Seconds=32
  { 2452327.50, 64.337647 }, //22 February 2002, UT1-UTC=-0.1536466, Accumulated Leap Seconds=32
  { 2452328.50, 64.338341 }, //23 February 2002, UT1-UTC=-0.1543413, Accumulated Leap Seconds=32
  { 2452329.50, 64.339000 }, //24 February 2002, UT1-UTC=-0.1549999, Accumulated Leap Seconds=32
  { 2452330.50, 64.339754 }, //25 February 2002, UT1-UTC=-0.1557544, Accumulated Leap Seconds=32
  { 2452331.50, 64.340715 }, //26 February 2002, UT1-UTC=-0.1567153, Accumulated Leap Seconds=32
  { 2452332.50, 64.341943 }, //27 February 2002, UT1-UTC=-0.1579427, Accumulated Leap Seconds=32
  { 2452333.50, 64.343414 }, //28 February 2002, UT1-UTC=-0.1594139, Accumulated Leap Seconds=32
  { 2452334.50, 64.345013 }, //01 March 2002, UT1-UTC=-0.1610130, Accumulated Leap Seconds=32
  { 2452335.50, 64.346676 }, //02 March 2002, UT1-UTC=-0.1626763, Accumulated Leap Seconds=32
  { 2452336.50, 64.348281 }, //03 March 2002, UT1-UTC=-0.1642811, Accumulated Leap Seconds=32
  { 2452337.50, 64.349671 }, //04 March 2002, UT1-UTC=-0.1656714, Accumulated Leap Seconds=32
  { 2452338.50, 64.350792 }, //05 March 2002, UT1-UTC=-0.1667923, Accumulated Leap Seconds=32
  { 2452339.50, 64.351692 }, //06 March 2002, UT1-UTC=-0.1676924, Accumulated Leap Seconds=32
  { 2452340.50, 64.352420 }, //07 March 2002, UT1-UTC=-0.1684204, Accumulated Leap Seconds=32
  { 2452341.50, 64.353027 }, //08 March 2002, UT1-UTC=-0.1690271, Accumulated Leap Seconds=32
  { 2452342.50, 64.353578 }, //09 March 2002, UT1-UTC=-0.1695783, Accumulated Leap Seconds=32
  { 2452343.50, 64.354115 }, //10 March 2002, UT1-UTC=-0.1701147, Accumulated Leap Seconds=32
  { 2452344.50, 64.354726 }, //11 March 2002, UT1-UTC=-0.1707265, Accumulated Leap Seconds=32
  { 2452345.50, 64.355474 }, //12 March 2002, UT1-UTC=-0.1714742, Accumulated Leap Seconds=32
  { 2452346.50, 64.356370 }, //13 March 2002, UT1-UTC=-0.1723703, Accumulated Leap Seconds=32
  { 2452347.50, 64.357359 }, //14 March 2002, UT1-UTC=-0.1733587, Accumulated Leap Seconds=32
  { 2452348.50, 64.358353 }, //15 March 2002, UT1-UTC=-0.1743534, Accumulated Leap Seconds=32
  { 2452349.50, 64.359290 }, //16 March 2002, UT1-UTC=-0.1752898, Accumulated Leap Seconds=32
  { 2452350.50, 64.360179 }, //17 March 2002, UT1-UTC=-0.1761789, Accumulated Leap Seconds=32
  { 2452351.50, 64.361019 }, //18 March 2002, UT1-UTC=-0.1770191, Accumulated Leap Seconds=32
  { 2452352.50, 64.361769 }, //19 March 2002, UT1-UTC=-0.1777692, Accumulated Leap Seconds=32
  { 2452353.50, 64.362368 }, //20 March 2002, UT1-UTC=-0.1783677, Accumulated Leap Seconds=32
  { 2452354.50, 64.362833 }, //21 March 2002, UT1-UTC=-0.1788335, Accumulated Leap Seconds=32
  { 2452355.50, 64.363222 }, //22 March 2002, UT1-UTC=-0.1792219, Accumulated Leap Seconds=32
  { 2452356.50, 64.363603 }, //23 March 2002, UT1-UTC=-0.1796032, Accumulated Leap Seconds=32
  { 2452357.50, 64.364049 }, //24 March 2002, UT1-UTC=-0.1800487, Accumulated Leap Seconds=32
  { 2452358.50, 64.364658 }, //25 March 2002, UT1-UTC=-0.1806578, Accumulated Leap Seconds=32
  { 2452359.50, 64.365514 }, //26 March 2002, UT1-UTC=-0.1815139, Accumulated Leap Seconds=32
  { 2452360.50, 64.366663 }, //27 March 2002, UT1-UTC=-0.1826625, Accumulated Leap Seconds=32
  { 2452361.50, 64.368094 }, //28 March 2002, UT1-UTC=-0.1840935, Accumulated Leap Seconds=32
  { 2452362.50, 64.369654 }, //29 March 2002, UT1-UTC=-0.1856545, Accumulated Leap Seconds=32
  { 2452363.50, 64.371129 }, //30 March 2002, UT1-UTC=-0.1871290, Accumulated Leap Seconds=32
  { 2452364.50, 64.372410 }, //31 March 2002, UT1-UTC=-0.1884097, Accumulated Leap Seconds=32
  { 2452365.50, 64.373458 }, //01 April 2002, UT1-UTC=-0.1894584, Accumulated Leap Seconds=32
  { 2452366.50, 64.374263 }, //02 April 2002, UT1-UTC=-0.1902632, Accumulated Leap Seconds=32
  { 2452367.50, 64.374858 }, //03 April 2002, UT1-UTC=-0.1908576, Accumulated Leap Seconds=32
  { 2452368.50, 64.375294 }, //04 April 2002, UT1-UTC=-0.1912938, Accumulated Leap Seconds=32
  { 2452369.50, 64.375673 }, //05 April 2002, UT1-UTC=-0.1916733, Accumulated Leap Seconds=32
  { 2452370.50, 64.376073 }, //06 April 2002, UT1-UTC=-0.1920732, Accumulated Leap Seconds=32
  { 2452371.50, 64.376564 }, //07 April 2002, UT1-UTC=-0.1925641, Accumulated Leap Seconds=32
  { 2452372.50, 64.377149 }, //08 April 2002, UT1-UTC=-0.1931491, Accumulated Leap Seconds=32
  { 2452373.50, 64.377838 }, //09 April 2002, UT1-UTC=-0.1938376, Accumulated Leap Seconds=32
  { 2452374.50, 64.378639 }, //10 April 2002, UT1-UTC=-0.1946386, Accumulated Leap Seconds=32
  { 2452375.50, 64.379543 }, //11 April 2002, UT1-UTC=-0.1955434, Accumulated Leap Seconds=32
  { 2452376.50, 64.380507 }, //12 April 2002, UT1-UTC=-0.1965074, Accumulated Leap Seconds=32
  { 2452377.50, 64.381444 }, //13 April 2002, UT1-UTC=-0.1974443, Accumulated Leap Seconds=32
  { 2452378.50, 64.382290 }, //14 April 2002, UT1-UTC=-0.1982901, Accumulated Leap Seconds=32
  { 2452379.50, 64.383008 }, //15 April 2002, UT1-UTC=-0.1990076, Accumulated Leap Seconds=32
  { 2452380.50, 64.383563 }, //16 April 2002, UT1-UTC=-0.1995629, Accumulated Leap Seconds=32
  { 2452381.50, 64.383916 }, //17 April 2002, UT1-UTC=-0.1999155, Accumulated Leap Seconds=32
  { 2452382.50, 64.384194 }, //18 April 2002, UT1-UTC=-0.2001942, Accumulated Leap Seconds=32
  { 2452383.50, 64.384454 }, //19 April 2002, UT1-UTC=-0.2004544, Accumulated Leap Seconds=32
  { 2452384.50, 64.384674 }, //20 April 2002, UT1-UTC=-0.2006743, Accumulated Leap Seconds=32
  { 2452385.50, 64.384966 }, //21 April 2002, UT1-UTC=-0.2009659, Accumulated Leap Seconds=32
  { 2452386.50, 64.385462 }, //22 April 2002, UT1-UTC=-0.2014624, Accumulated Leap Seconds=32
  { 2452387.50, 64.386254 }, //23 April 2002, UT1-UTC=-0.2022539, Accumulated Leap Seconds=32
  { 2452388.50, 64.387313 }, //24 April 2002, UT1-UTC=-0.2033128, Accumulated Leap Seconds=32
  { 2452389.50, 64.388587 }, //25 April 2002, UT1-UTC=-0.2045866, Accumulated Leap Seconds=32
  { 2452390.50, 64.389994 }, //26 April 2002, UT1-UTC=-0.2059938, Accumulated Leap Seconds=32
  { 2452391.50, 64.391365 }, //27 April 2002, UT1-UTC=-0.2073654, Accumulated Leap Seconds=32
  { 2452392.50, 64.392533 }, //28 April 2002, UT1-UTC=-0.2085331, Accumulated Leap Seconds=32
  { 2452393.50, 64.393407 }, //29 April 2002, UT1-UTC=-0.2094074, Accumulated Leap Seconds=32
  { 2452394.50, 64.393996 }, //30 April 2002, UT1-UTC=-0.2099957, Accumulated Leap Seconds=32
  { 2452395.50, 64.394329 }, //01 May 2002, UT1-UTC=-0.2103291, Accumulated Leap Seconds=32
  { 2452396.50, 64.394553 }, //02 May 2002, UT1-UTC=-0.2105531, Accumulated Leap Seconds=32
  { 2452397.50, 64.394754 }, //03 May 2002, UT1-UTC=-0.2107538, Accumulated Leap Seconds=32
  { 2452398.50, 64.394992 }, //04 May 2002, UT1-UTC=-0.2109919, Accumulated Leap Seconds=32
  { 2452399.50, 64.395363 }, //05 May 2002, UT1-UTC=-0.2113629, Accumulated Leap Seconds=32
  { 2452400.50, 64.395911 }, //06 May 2002, UT1-UTC=-0.2119108, Accumulated Leap Seconds=32
  { 2452401.50, 64.396616 }, //07 May 2002, UT1-UTC=-0.2126161, Accumulated Leap Seconds=32
  { 2452402.50, 64.397446 }, //08 May 2002, UT1-UTC=-0.2134461, Accumulated Leap Seconds=32
  { 2452403.50, 64.398360 }, //09 May 2002, UT1-UTC=-0.2143605, Accumulated Leap Seconds=32
  { 2452404.50, 64.399327 }, //10 May 2002, UT1-UTC=-0.2153272, Accumulated Leap Seconds=32
  { 2452405.50, 64.400294 }, //11 May 2002, UT1-UTC=-0.2162937, Accumulated Leap Seconds=32
  { 2452406.50, 64.401196 }, //12 May 2002, UT1-UTC=-0.2171961, Accumulated Leap Seconds=32
  { 2452407.50, 64.402012 }, //13 May 2002, UT1-UTC=-0.2180117, Accumulated Leap Seconds=32
  { 2452408.50, 64.402729 }, //14 May 2002, UT1-UTC=-0.2187289, Accumulated Leap Seconds=32
  { 2452409.50, 64.403317 }, //15 May 2002, UT1-UTC=-0.2193166, Accumulated Leap Seconds=32
  { 2452410.50, 64.403856 }, //16 May 2002, UT1-UTC=-0.2198559, Accumulated Leap Seconds=32
  { 2452411.50, 64.404351 }, //17 May 2002, UT1-UTC=-0.2203514, Accumulated Leap Seconds=32
  { 2452412.50, 64.404863 }, //18 May 2002, UT1-UTC=-0.2208634, Accumulated Leap Seconds=32
  { 2452413.50, 64.405540 }, //19 May 2002, UT1-UTC=-0.2215396, Accumulated Leap Seconds=32
  { 2452414.50, 64.406449 }, //20 May 2002, UT1-UTC=-0.2224490, Accumulated Leap Seconds=32
  { 2452415.50, 64.407619 }, //21 May 2002, UT1-UTC=-0.2236193, Accumulated Leap Seconds=32
  { 2452416.50, 64.408988 }, //22 May 2002, UT1-UTC=-0.2249880, Accumulated Leap Seconds=32
  { 2452417.50, 64.410417 }, //23 May 2002, UT1-UTC=-0.2264168, Accumulated Leap Seconds=32
  { 2452418.50, 64.411798 }, //24 May 2002, UT1-UTC=-0.2277978, Accumulated Leap Seconds=32
  { 2452419.50, 64.413004 }, //25 May 2002, UT1-UTC=-0.2290038, Accumulated Leap Seconds=32
  { 2452420.50, 64.413922 }, //26 May 2002, UT1-UTC=-0.2299216, Accumulated Leap Seconds=32
  { 2452421.50, 64.414524 }, //27 May 2002, UT1-UTC=-0.2305238, Accumulated Leap Seconds=32
  { 2452422.50, 64.414850 }, //28 May 2002, UT1-UTC=-0.2308498, Accumulated Leap Seconds=32
  { 2452423.50, 64.414982 }, //29 May 2002, UT1-UTC=-0.2309822, Accumulated Leap Seconds=32
  { 2452424.50, 64.415023 }, //30 May 2002, UT1-UTC=-0.2310228, Accumulated Leap Seconds=32
  { 2452425.50, 64.415061 }, //31 May 2002, UT1-UTC=-0.2310614, Accumulated Leap Seconds=32
  { 2452426.50, 64.415116 }, //01 June 2002, UT1-UTC=-0.2311156, Accumulated Leap Seconds=32
  { 2452427.50, 64.415221 }, //02 June 2002, UT1-UTC=-0.2312210, Accumulated Leap Seconds=32
  { 2452428.50, 64.415385 }, //03 June 2002, UT1-UTC=-0.2313850, Accumulated Leap Seconds=32
  { 2452429.50, 64.415601 }, //04 June 2002, UT1-UTC=-0.2316011, Accumulated Leap Seconds=32
  { 2452430.50, 64.415817 }, //05 June 2002, UT1-UTC=-0.2318168, Accumulated Leap Seconds=32
  { 2452431.50, 64.415936 }, //06 June 2002, UT1-UTC=-0.2319358, Accumulated Leap Seconds=32
  { 2452432.50, 64.415953 }, //07 June 2002, UT1-UTC=-0.2319527, Accumulated Leap Seconds=32
  { 2452433.50, 64.415866 }, //08 June 2002, UT1-UTC=-0.2318656, Accumulated Leap Seconds=32
  { 2452434.50, 64.415637 }, //09 June 2002, UT1-UTC=-0.2316373, Accumulated Leap Seconds=32
  { 2452435.50, 64.415256 }, //10 June 2002, UT1-UTC=-0.2312558, Accumulated Leap Seconds=32
  { 2452436.50, 64.414726 }, //11 June 2002, UT1-UTC=-0.2307256, Accumulated Leap Seconds=32
  { 2452437.50, 64.414069 }, //12 June 2002, UT1-UTC=-0.2300693, Accumulated Leap Seconds=32
  { 2452438.50, 64.413390 }, //13 June 2002, UT1-UTC=-0.2293896, Accumulated Leap Seconds=32
  { 2452439.50, 64.412701 }, //14 June 2002, UT1-UTC=-0.2287007, Accumulated Leap Seconds=32
  { 2452440.50, 64.412185 }, //15 June 2002, UT1-UTC=-0.2281849, Accumulated Leap Seconds=32
  { 2452441.50, 64.411908 }, //16 June 2002, UT1-UTC=-0.2279079, Accumulated Leap Seconds=32
  { 2452442.50, 64.411893 }, //17 June 2002, UT1-UTC=-0.2278930, Accumulated Leap Seconds=32
  { 2452443.50, 64.412080 }, //18 June 2002, UT1-UTC=-0.2280801, Accumulated Leap Seconds=32
  { 2452444.50, 64.412359 }, //19 June 2002, UT1-UTC=-0.2283589, Accumulated Leap Seconds=32
  { 2452445.50, 64.412683 }, //20 June 2002, UT1-UTC=-0.2286830, Accumulated Leap Seconds=32
  { 2452446.50, 64.412987 }, //21 June 2002, UT1-UTC=-0.2289872, Accumulated Leap Seconds=32
  { 2452447.50, 64.413218 }, //22 June 2002, UT1-UTC=-0.2292182, Accumulated Leap Seconds=32
  { 2452448.50, 64.413302 }, //23 June 2002, UT1-UTC=-0.2293020, Accumulated Leap Seconds=32
  { 2452449.50, 64.413245 }, //24 June 2002, UT1-UTC=-0.2292446, Accumulated Leap Seconds=32
  { 2452450.50, 64.413110 }, //25 June 2002, UT1-UTC=-0.2291098, Accumulated Leap Seconds=32
  { 2452451.50, 64.412962 }, //26 June 2002, UT1-UTC=-0.2289625, Accumulated Leap Seconds=32
  { 2452452.50, 64.412828 }, //27 June 2002, UT1-UTC=-0.2288278, Accumulated Leap Seconds=32
  { 2452453.50, 64.412777 }, //28 June 2002, UT1-UTC=-0.2287767, Accumulated Leap Seconds=32
  { 2452454.50, 64.412835 }, //29 June 2002, UT1-UTC=-0.2288352, Accumulated Leap Seconds=32
  { 2452455.50, 64.412969 }, //30 June 2002, UT1-UTC=-0.2289689, Accumulated Leap Seconds=32
  { 2452456.50, 64.413206 }, //01 July 2002, UT1-UTC=-0.2292064, Accumulated Leap Seconds=32
  { 2452457.50, 64.413564 }, //02 July 2002, UT1-UTC=-0.2295636, Accumulated Leap Seconds=32
  { 2452458.50, 64.414011 }, //03 July 2002, UT1-UTC=-0.2300111, Accumulated Leap Seconds=32
  { 2452459.50, 64.414429 }, //04 July 2002, UT1-UTC=-0.2304294, Accumulated Leap Seconds=32
  { 2452460.50, 64.414786 }, //05 July 2002, UT1-UTC=-0.2307860, Accumulated Leap Seconds=32
  { 2452461.50, 64.415027 }, //06 July 2002, UT1-UTC=-0.2310266, Accumulated Leap Seconds=32
  { 2452462.50, 64.415123 }, //07 July 2002, UT1-UTC=-0.2311227, Accumulated Leap Seconds=32
  { 2452463.50, 64.415079 }, //08 July 2002, UT1-UTC=-0.2310794, Accumulated Leap Seconds=32
  { 2452464.50, 64.414914 }, //09 July 2002, UT1-UTC=-0.2309140, Accumulated Leap Seconds=32
  { 2452465.50, 64.414668 }, //10 July 2002, UT1-UTC=-0.2306679, Accumulated Leap Seconds=32
  { 2452466.50, 64.414390 }, //11 July 2002, UT1-UTC=-0.2303905, Accumulated Leap Seconds=32
  { 2452467.50, 64.414169 }, //12 July 2002, UT1-UTC=-0.2301687, Accumulated Leap Seconds=32
  { 2452468.50, 64.414184 }, //13 July 2002, UT1-UTC=-0.2301842, Accumulated Leap Seconds=32
  { 2452469.50, 64.414434 }, //14 July 2002, UT1-UTC=-0.2304341, Accumulated Leap Seconds=32
  { 2452470.50, 64.414894 }, //15 July 2002, UT1-UTC=-0.2308944, Accumulated Leap Seconds=32
  { 2452471.50, 64.415473 }, //16 July 2002, UT1-UTC=-0.2314728, Accumulated Leap Seconds=32
  { 2452472.50, 64.416047 }, //17 July 2002, UT1-UTC=-0.2320466, Accumulated Leap Seconds=32
  { 2452473.50, 64.416515 }, //18 July 2002, UT1-UTC=-0.2325153, Accumulated Leap Seconds=32
  { 2452474.50, 64.416806 }, //19 July 2002, UT1-UTC=-0.2328064, Accumulated Leap Seconds=32
  { 2452475.50, 64.416826 }, //20 July 2002, UT1-UTC=-0.2328256, Accumulated Leap Seconds=32
  { 2452476.50, 64.416549 }, //21 July 2002, UT1-UTC=-0.2325494, Accumulated Leap Seconds=32
  { 2452477.50, 64.416043 }, //22 July 2002, UT1-UTC=-0.2320432, Accumulated Leap Seconds=32
  { 2452478.50, 64.415400 }, //23 July 2002, UT1-UTC=-0.2313996, Accumulated Leap Seconds=32
  { 2452479.50, 64.414692 }, //24 July 2002, UT1-UTC=-0.2306922, Accumulated Leap Seconds=32
  { 2452480.50, 64.413988 }, //25 July 2002, UT1-UTC=-0.2299879, Accumulated Leap Seconds=32
  { 2452481.50, 64.413406 }, //26 July 2002, UT1-UTC=-0.2294060, Accumulated Leap Seconds=32
  { 2452482.50, 64.412964 }, //27 July 2002, UT1-UTC=-0.2289642, Accumulated Leap Seconds=32
  { 2452483.50, 64.412655 }, //28 July 2002, UT1-UTC=-0.2286554, Accumulated Leap Seconds=32
  { 2452484.50, 64.412457 }, //29 July 2002, UT1-UTC=-0.2284570, Accumulated Leap Seconds=32
  { 2452485.50, 64.412300 }, //30 July 2002, UT1-UTC=-0.2282995, Accumulated Leap Seconds=32
  { 2452486.50, 64.412106 }, //31 July 2002, UT1-UTC=-0.2281059, Accumulated Leap Seconds=32
  { 2452487.50, 64.411846 }, //01 August 2002, UT1-UTC=-0.2278464, Accumulated Leap Seconds=32
  { 2452488.50, 64.411512 }, //02 August 2002, UT1-UTC=-0.2275119, Accumulated Leap Seconds=32
  { 2452489.50, 64.411097 }, //03 August 2002, UT1-UTC=-0.2270973, Accumulated Leap Seconds=32
  { 2452490.50, 64.410556 }, //04 August 2002, UT1-UTC=-0.2265564, Accumulated Leap Seconds=32
  { 2452491.50, 64.409892 }, //05 August 2002, UT1-UTC=-0.2258918, Accumulated Leap Seconds=32
  { 2452492.50, 64.409160 }, //06 August 2002, UT1-UTC=-0.2251603, Accumulated Leap Seconds=32
  { 2452493.50, 64.408445 }, //07 August 2002, UT1-UTC=-0.2244446, Accumulated Leap Seconds=32
  { 2452494.50, 64.407856 }, //08 August 2002, UT1-UTC=-0.2238560, Accumulated Leap Seconds=32
  { 2452495.50, 64.407511 }, //09 August 2002, UT1-UTC=-0.2235111, Accumulated Leap Seconds=32
  { 2452496.50, 64.407477 }, //10 August 2002, UT1-UTC=-0.2234774, Accumulated Leap Seconds=32
  { 2452497.50, 64.407737 }, //11 August 2002, UT1-UTC=-0.2237373, Accumulated Leap Seconds=32
  { 2452498.50, 64.408218 }, //12 August 2002, UT1-UTC=-0.2242181, Accumulated Leap Seconds=32
  { 2452499.50, 64.408797 }, //13 August 2002, UT1-UTC=-0.2247973, Accumulated Leap Seconds=32
  { 2452500.50, 64.409321 }, //14 August 2002, UT1-UTC=-0.2253212, Accumulated Leap Seconds=32
  { 2452501.50, 64.409702 }, //15 August 2002, UT1-UTC=-0.2257019, Accumulated Leap Seconds=32
  { 2452502.50, 64.409859 }, //16 August 2002, UT1-UTC=-0.2258593, Accumulated Leap Seconds=32
  { 2452503.50, 64.409780 }, //17 August 2002, UT1-UTC=-0.2257801, Accumulated Leap Seconds=32
  { 2452504.50, 64.409524 }, //18 August 2002, UT1-UTC=-0.2255239, Accumulated Leap Seconds=32
  { 2452505.50, 64.409164 }, //19 August 2002, UT1-UTC=-0.2251639, Accumulated Leap Seconds=32
  { 2452506.50, 64.408774 }, //20 August 2002, UT1-UTC=-0.2247744, Accumulated Leap Seconds=32
  { 2452507.50, 64.408440 }, //21 August 2002, UT1-UTC=-0.2244398, Accumulated Leap Seconds=32
  { 2452508.50, 64.408268 }, //22 August 2002, UT1-UTC=-0.2242676, Accumulated Leap Seconds=32
  { 2452509.50, 64.408267 }, //23 August 2002, UT1-UTC=-0.2242667, Accumulated Leap Seconds=32
  { 2452510.50, 64.408354 }, //24 August 2002, UT1-UTC=-0.2243537, Accumulated Leap Seconds=32
  { 2452511.50, 64.408557 }, //25 August 2002, UT1-UTC=-0.2245567, Accumulated Leap Seconds=32
  { 2452512.50, 64.408842 }, //26 August 2002, UT1-UTC=-0.2248424, Accumulated Leap Seconds=32
  { 2452513.50, 64.409162 }, //27 August 2002, UT1-UTC=-0.2251621, Accumulated Leap Seconds=32
  { 2452514.50, 64.409469 }, //28 August 2002, UT1-UTC=-0.2254686, Accumulated Leap Seconds=32
  { 2452515.50, 64.409704 }, //29 August 2002, UT1-UTC=-0.2257036, Accumulated Leap Seconds=32
  { 2452516.50, 64.409814 }, //30 August 2002, UT1-UTC=-0.2258141, Accumulated Leap Seconds=32
  { 2452517.50, 64.409791 }, //31 August 2002, UT1-UTC=-0.2257909, Accumulated Leap Seconds=32
  { 2452518.50, 64.409654 }, //01 September 2002, UT1-UTC=-0.2256536, Accumulated Leap Seconds=32
  { 2452519.50, 64.409428 }, //02 September 2002, UT1-UTC=-0.2254281, Accumulated Leap Seconds=32
  { 2452520.50, 64.409168 }, //03 September 2002, UT1-UTC=-0.2251676, Accumulated Leap Seconds=32
  { 2452521.50, 64.408974 }, //04 September 2002, UT1-UTC=-0.2249739, Accumulated Leap Seconds=32
  { 2452522.50, 64.408953 }, //05 September 2002, UT1-UTC=-0.2249528, Accumulated Leap Seconds=32
  { 2452523.50, 64.409159 }, //06 September 2002, UT1-UTC=-0.2251587, Accumulated Leap Seconds=32
  { 2452524.50, 64.409690 }, //07 September 2002, UT1-UTC=-0.2256897, Accumulated Leap Seconds=32
  { 2452525.50, 64.410539 }, //08 September 2002, UT1-UTC=-0.2265390, Accumulated Leap Seconds=32
  { 2452526.50, 64.411565 }, //09 September 2002, UT1-UTC=-0.2275649, Accumulated Leap Seconds=32
  { 2452527.50, 64.412580 }, //10 September 2002, UT1-UTC=-0.2285800, Accumulated Leap Seconds=32
  { 2452528.50, 64.413409 }, //11 September 2002, UT1-UTC=-0.2294085, Accumulated Leap Seconds=32
  { 2452529.50, 64.413958 }, //12 September 2002, UT1-UTC=-0.2299584, Accumulated Leap Seconds=32
  { 2452530.50, 64.414288 }, //13 September 2002, UT1-UTC=-0.2302885, Accumulated Leap Seconds=32
  { 2452531.50, 64.414426 }, //14 September 2002, UT1-UTC=-0.2304264, Accumulated Leap Seconds=32
  { 2452532.50, 64.414433 }, //15 September 2002, UT1-UTC=-0.2304327, Accumulated Leap Seconds=32
  { 2452533.50, 64.414393 }, //16 September 2002, UT1-UTC=-0.2303933, Accumulated Leap Seconds=32
  { 2452534.50, 64.414370 }, //17 September 2002, UT1-UTC=-0.2303701, Accumulated Leap Seconds=32
  { 2452535.50, 64.414397 }, //18 September 2002, UT1-UTC=-0.2303971, Accumulated Leap Seconds=32
  { 2452536.50, 64.414518 }, //19 September 2002, UT1-UTC=-0.2305177, Accumulated Leap Seconds=32
  { 2452537.50, 64.414748 }, //20 September 2002, UT1-UTC=-0.2307477, Accumulated Leap Seconds=32
  { 2452538.50, 64.415080 }, //21 September 2002, UT1-UTC=-0.2310795, Accumulated Leap Seconds=32
  { 2452539.50, 64.415490 }, //22 September 2002, UT1-UTC=-0.2314898, Accumulated Leap Seconds=32
  { 2452540.50, 64.415941 }, //23 September 2002, UT1-UTC=-0.2319408, Accumulated Leap Seconds=32
  { 2452541.50, 64.416385 }, //24 September 2002, UT1-UTC=-0.2323847, Accumulated Leap Seconds=32
  { 2452542.50, 64.416755 }, //25 September 2002, UT1-UTC=-0.2327552, Accumulated Leap Seconds=32
  { 2452543.50, 64.417004 }, //26 September 2002, UT1-UTC=-0.2330039, Accumulated Leap Seconds=32
  { 2452544.50, 64.417103 }, //27 September 2002, UT1-UTC=-0.2331031, Accumulated Leap Seconds=32
  { 2452545.50, 64.417145 }, //28 September 2002, UT1-UTC=-0.2331446, Accumulated Leap Seconds=32
  { 2452546.50, 64.417066 }, //29 September 2002, UT1-UTC=-0.2330660, Accumulated Leap Seconds=32
  { 2452547.50, 64.416913 }, //30 September 2002, UT1-UTC=-0.2329126, Accumulated Leap Seconds=32
  { 2452548.50, 64.416783 }, //01 October 2002, UT1-UTC=-0.2327832, Accumulated Leap Seconds=32
  { 2452549.50, 64.416802 }, //02 October 2002, UT1-UTC=-0.2328021, Accumulated Leap Seconds=32
  { 2452550.50, 64.417001 }, //03 October 2002, UT1-UTC=-0.2330013, Accumulated Leap Seconds=32
  { 2452551.50, 64.417454 }, //04 October 2002, UT1-UTC=-0.2334542, Accumulated Leap Seconds=32
  { 2452552.50, 64.418256 }, //05 October 2002, UT1-UTC=-0.2342555, Accumulated Leap Seconds=32
  { 2452553.50, 64.419307 }, //06 October 2002, UT1-UTC=-0.2353067, Accumulated Leap Seconds=32
  { 2452554.50, 64.420457 }, //07 October 2002, UT1-UTC=-0.2364571, Accumulated Leap Seconds=32
  { 2452555.50, 64.421533 }, //08 October 2002, UT1-UTC=-0.2375330, Accumulated Leap Seconds=32
  { 2452556.50, 64.422403 }, //09 October 2002, UT1-UTC=-0.2384033, Accumulated Leap Seconds=32
  { 2452557.50, 64.423008 }, //10 October 2002, UT1-UTC=-0.2390083, Accumulated Leap Seconds=32
  { 2452558.50, 64.423363 }, //11 October 2002, UT1-UTC=-0.2393631, Accumulated Leap Seconds=32
  { 2452559.50, 64.423540 }, //12 October 2002, UT1-UTC=-0.2395402, Accumulated Leap Seconds=32
  { 2452560.50, 64.423657 }, //13 October 2002, UT1-UTC=-0.2396567, Accumulated Leap Seconds=32
  { 2452561.50, 64.423822 }, //14 October 2002, UT1-UTC=-0.2398217, Accumulated Leap Seconds=32
  { 2452562.50, 64.424091 }, //15 October 2002, UT1-UTC=-0.2400912, Accumulated Leap Seconds=32
  { 2452563.50, 64.424483 }, //16 October 2002, UT1-UTC=-0.2404829, Accumulated Leap Seconds=32
  { 2452564.50, 64.425016 }, //17 October 2002, UT1-UTC=-0.2410163, Accumulated Leap Seconds=32
  { 2452565.50, 64.425696 }, //18 October 2002, UT1-UTC=-0.2416962, Accumulated Leap Seconds=32
  { 2452566.50, 64.426430 }, //19 October 2002, UT1-UTC=-0.2424298, Accumulated Leap Seconds=32
  { 2452567.50, 64.427189 }, //20 October 2002, UT1-UTC=-0.2431891, Accumulated Leap Seconds=32
  { 2452568.50, 64.427946 }, //21 October 2002, UT1-UTC=-0.2439456, Accumulated Leap Seconds=32
  { 2452569.50, 64.428677 }, //22 October 2002, UT1-UTC=-0.2446767, Accumulated Leap Seconds=32
  { 2452570.50, 64.429347 }, //23 October 2002, UT1-UTC=-0.2453475, Accumulated Leap Seconds=32
  { 2452571.50, 64.429909 }, //24 October 2002, UT1-UTC=-0.2459090, Accumulated Leap Seconds=32
  { 2452572.50, 64.430336 }, //25 October 2002, UT1-UTC=-0.2463360, Accumulated Leap Seconds=32
  { 2452573.50, 64.430629 }, //26 October 2002, UT1-UTC=-0.2466288, Accumulated Leap Seconds=32
  { 2452574.50, 64.430846 }, //27 October 2002, UT1-UTC=-0.2468462, Accumulated Leap Seconds=32
  { 2452575.50, 64.431029 }, //28 October 2002, UT1-UTC=-0.2470293, Accumulated Leap Seconds=32
  { 2452576.50, 64.431203 }, //29 October 2002, UT1-UTC=-0.2472034, Accumulated Leap Seconds=32
  { 2452577.50, 64.431516 }, //30 October 2002, UT1-UTC=-0.2475164, Accumulated Leap Seconds=32
  { 2452578.50, 64.432081 }, //31 October 2002, UT1-UTC=-0.2480814, Accumulated Leap Seconds=32
  { 2452579.50, 64.432920 }, //01 November 2002, UT1-UTC=-0.2489200, Accumulated Leap Seconds=32
  { 2452580.50, 64.434022 }, //02 November 2002, UT1-UTC=-0.2500217, Accumulated Leap Seconds=32
  { 2452581.50, 64.435263 }, //03 November 2002, UT1-UTC=-0.2512630, Accumulated Leap Seconds=32
  { 2452582.50, 64.436491 }, //04 November 2002, UT1-UTC=-0.2524910, Accumulated Leap Seconds=32
  { 2452583.50, 64.437587 }, //05 November 2002, UT1-UTC=-0.2535866, Accumulated Leap Seconds=32
  { 2452584.50, 64.438400 }, //06 November 2002, UT1-UTC=-0.2544001, Accumulated Leap Seconds=32
  { 2452585.50, 64.438977 }, //07 November 2002, UT1-UTC=-0.2549773, Accumulated Leap Seconds=32
  { 2452586.50, 64.439326 }, //08 November 2002, UT1-UTC=-0.2553262, Accumulated Leap Seconds=32
  { 2452587.50, 64.439526 }, //09 November 2002, UT1-UTC=-0.2555263, Accumulated Leap Seconds=32
  { 2452588.50, 64.439625 }, //10 November 2002, UT1-UTC=-0.2556253, Accumulated Leap Seconds=32
  { 2452589.50, 64.439733 }, //11 November 2002, UT1-UTC=-0.2557333, Accumulated Leap Seconds=32
  { 2452590.50, 64.439923 }, //12 November 2002, UT1-UTC=-0.2559230, Accumulated Leap Seconds=32
  { 2452591.50, 64.440253 }, //13 November 2002, UT1-UTC=-0.2562531, Accumulated Leap Seconds=32
  { 2452592.50, 64.440745 }, //14 November 2002, UT1-UTC=-0.2567446, Accumulated Leap Seconds=32
  { 2452593.50, 64.441338 }, //15 November 2002, UT1-UTC=-0.2573380, Accumulated Leap Seconds=32
  { 2452594.50, 64.441978 }, //16 November 2002, UT1-UTC=-0.2579784, Accumulated Leap Seconds=32
  { 2452595.50, 64.442628 }, //17 November 2002, UT1-UTC=-0.2586282, Accumulated Leap Seconds=32
  { 2452596.50, 64.443240 }, //18 November 2002, UT1-UTC=-0.2592403, Accumulated Leap Seconds=32
  { 2452597.50, 64.443767 }, //19 November 2002, UT1-UTC=-0.2597669, Accumulated Leap Seconds=32
  { 2452598.50, 64.444139 }, //20 November 2002, UT1-UTC=-0.2601395, Accumulated Leap Seconds=32
  { 2452599.50, 64.444381 }, //21 November 2002, UT1-UTC=-0.2603806, Accumulated Leap Seconds=32
  { 2452600.50, 64.444572 }, //22 November 2002, UT1-UTC=-0.2605723, Accumulated Leap Seconds=32
  { 2452601.50, 64.444733 }, //23 November 2002, UT1-UTC=-0.2607330, Accumulated Leap Seconds=32
  { 2452602.50, 64.444893 }, //24 November 2002, UT1-UTC=-0.2608929, Accumulated Leap Seconds=32
  { 2452603.50, 64.445144 }, //25 November 2002, UT1-UTC=-0.2611440, Accumulated Leap Seconds=32
  { 2452604.50, 64.445566 }, //26 November 2002, UT1-UTC=-0.2615656, Accumulated Leap Seconds=32
  { 2452605.50, 64.446212 }, //27 November 2002, UT1-UTC=-0.2622115, Accumulated Leap Seconds=32
  { 2452606.50, 64.447130 }, //28 November 2002, UT1-UTC=-0.2631298, Accumulated Leap Seconds=32
  { 2452607.50, 64.448284 }, //29 November 2002, UT1-UTC=-0.2642837, Accumulated Leap Seconds=32
  { 2452608.50, 64.449629 }, //30 November 2002, UT1-UTC=-0.2656293, Accumulated Leap Seconds=32
  { 2452609.50, 64.451053 }, //01 December 2002, UT1-UTC=-0.2670529, Accumulated Leap Seconds=32
  { 2452610.50, 64.452408 }, //02 December 2002, UT1-UTC=-0.2684078, Accumulated Leap Seconds=32
  { 2452611.50, 64.453566 }, //03 December 2002, UT1-UTC=-0.2695660, Accumulated Leap Seconds=32
  { 2452612.50, 64.454448 }, //04 December 2002, UT1-UTC=-0.2704478, Accumulated Leap Seconds=32
  { 2452613.50, 64.455057 }, //05 December 2002, UT1-UTC=-0.2710566, Accumulated Leap Seconds=32
  { 2452614.50, 64.455424 }, //06 December 2002, UT1-UTC=-0.2714236, Accumulated Leap Seconds=32
  { 2452615.50, 64.455740 }, //07 December 2002, UT1-UTC=-0.2717396, Accumulated Leap Seconds=32
  { 2452616.50, 64.456121 }, //08 December 2002, UT1-UTC=-0.2721209, Accumulated Leap Seconds=32
  { 2452617.50, 64.456641 }, //09 December 2002, UT1-UTC=-0.2726412, Accumulated Leap Seconds=32
  { 2452618.50, 64.457336 }, //10 December 2002, UT1-UTC=-0.2733357, Accumulated Leap Seconds=32
  { 2452619.50, 64.458191 }, //11 December 2002, UT1-UTC=-0.2741907, Accumulated Leap Seconds=32
  { 2452620.50, 64.459186 }, //12 December 2002, UT1-UTC=-0.2751856, Accumulated Leap Seconds=32
  { 2452621.50, 64.460197 }, //13 December 2002, UT1-UTC=-0.2761973, Accumulated Leap Seconds=32
  { 2452622.50, 64.461257 }, //14 December 2002, UT1-UTC=-0.2772572, Accumulated Leap Seconds=32
  { 2452623.50, 64.462260 }, //15 December 2002, UT1-UTC=-0.2782595, Accumulated Leap Seconds=32
  { 2452624.50, 64.463143 }, //16 December 2002, UT1-UTC=-0.2791429, Accumulated Leap Seconds=32
  { 2452625.50, 64.463866 }, //17 December 2002, UT1-UTC=-0.2798662, Accumulated Leap Seconds=32
  { 2452626.50, 64.464432 }, //18 December 2002, UT1-UTC=-0.2804324, Accumulated Leap Seconds=32
  { 2452627.50, 64.464807 }, //19 December 2002, UT1-UTC=-0.2808068, Accumulated Leap Seconds=32
  { 2452628.50, 64.465029 }, //20 December 2002, UT1-UTC=-0.2810295, Accumulated Leap Seconds=32
  { 2452629.50, 64.465210 }, //21 December 2002, UT1-UTC=-0.2812100, Accumulated Leap Seconds=32
  { 2452630.50, 64.465387 }, //22 December 2002, UT1-UTC=-0.2813870, Accumulated Leap Seconds=32
  { 2452631.50, 64.465638 }, //23 December 2002, UT1-UTC=-0.2816384, Accumulated Leap Seconds=32
  { 2452632.50, 64.466081 }, //24 December 2002, UT1-UTC=-0.2820811, Accumulated Leap Seconds=32
  { 2452633.50, 64.466800 }, //25 December 2002, UT1-UTC=-0.2828001, Accumulated Leap Seconds=32
  { 2452634.50, 64.467765 }, //26 December 2002, UT1-UTC=-0.2837648, Accumulated Leap Seconds=32
  { 2452635.50, 64.468896 }, //27 December 2002, UT1-UTC=-0.2848960, Accumulated Leap Seconds=32
  { 2452636.50, 64.470082 }, //28 December 2002, UT1-UTC=-0.2860823, Accumulated Leap Seconds=32
  { 2452637.50, 64.471207 }, //29 December 2002, UT1-UTC=-0.2872075, Accumulated Leap Seconds=32
  { 2452638.50, 64.472165 }, //30 December 2002, UT1-UTC=-0.2881649, Accumulated Leap Seconds=32
  { 2452639.50, 64.472901 }, //31 December 2002, UT1-UTC=-0.2889012, Accumulated Leap Seconds=32
  { 2452640.50, 64.473428 }, //01 January 2003, UT1-UTC=-0.2894276, Accumulated Leap Seconds=32
  { 2452641.50, 64.473785 }, //02 January 2003, UT1-UTC=-0.2897847, Accumulated Leap Seconds=32
  { 2452642.50, 64.474091 }, //03 January 2003, UT1-UTC=-0.2900905, Accumulated Leap Seconds=32
  { 2452643.50, 64.474430 }, //04 January 2003, UT1-UTC=-0.2904298, Accumulated Leap Seconds=32
  { 2452644.50, 64.474811 }, //05 January 2003, UT1-UTC=-0.2908112, Accumulated Leap Seconds=32
  { 2452645.50, 64.475324 }, //06 January 2003, UT1-UTC=-0.2913240, Accumulated Leap Seconds=32
  { 2452646.50, 64.476019 }, //07 January 2003, UT1-UTC=-0.2920188, Accumulated Leap Seconds=32
  { 2452647.50, 64.476840 }, //08 January 2003, UT1-UTC=-0.2928401, Accumulated Leap Seconds=32
  { 2452648.50, 64.477714 }, //09 January 2003, UT1-UTC=-0.2937142, Accumulated Leap Seconds=32
  { 2452649.50, 64.478625 }, //10 January 2003, UT1-UTC=-0.2946250, Accumulated Leap Seconds=32
  { 2452650.50, 64.479555 }, //11 January 2003, UT1-UTC=-0.2955554, Accumulated Leap Seconds=32
  { 2452651.50, 64.480459 }, //12 January 2003, UT1-UTC=-0.2964593, Accumulated Leap Seconds=32
  { 2452652.50, 64.481254 }, //13 January 2003, UT1-UTC=-0.2972536, Accumulated Leap Seconds=32
  { 2452653.50, 64.481903 }, //14 January 2003, UT1-UTC=-0.2979028, Accumulated Leap Seconds=32
  { 2452654.50, 64.482410 }, //15 January 2003, UT1-UTC=-0.2984095, Accumulated Leap Seconds=32
  { 2452655.50, 64.482761 }, //16 January 2003, UT1-UTC=-0.2987611, Accumulated Leap Seconds=32
  { 2452656.50, 64.482994 }, //17 January 2003, UT1-UTC=-0.2989943, Accumulated Leap Seconds=32
  { 2452657.50, 64.483203 }, //18 January 2003, UT1-UTC=-0.2992026, Accumulated Leap Seconds=32
  { 2452658.50, 64.483485 }, //19 January 2003, UT1-UTC=-0.2994845, Accumulated Leap Seconds=32
  { 2452659.50, 64.483895 }, //20 January 2003, UT1-UTC=-0.2998952, Accumulated Leap Seconds=32
  { 2452660.50, 64.484452 }, //21 January 2003, UT1-UTC=-0.3004516, Accumulated Leap Seconds=32
  { 2452661.50, 64.485158 }, //22 January 2003, UT1-UTC=-0.3011576, Accumulated Leap Seconds=32
  { 2452662.50, 64.485980 }, //23 January 2003, UT1-UTC=-0.3019797, Accumulated Leap Seconds=32
  { 2452663.50, 64.486861 }, //24 January 2003, UT1-UTC=-0.3028613, Accumulated Leap Seconds=32
  { 2452664.50, 64.487754 }, //25 January 2003, UT1-UTC=-0.3037536, Accumulated Leap Seconds=32
  { 2452665.50, 64.488515 }, //26 January 2003, UT1-UTC=-0.3045155, Accumulated Leap Seconds=32
  { 2452666.50, 64.489042 }, //27 January 2003, UT1-UTC=-0.3050423, Accumulated Leap Seconds=32
  { 2452667.50, 64.489322 }, //28 January 2003, UT1-UTC=-0.3053217, Accumulated Leap Seconds=32
  { 2452668.50, 64.489400 }, //29 January 2003, UT1-UTC=-0.3054000, Accumulated Leap Seconds=32
  { 2452669.50, 64.489389 }, //30 January 2003, UT1-UTC=-0.3053888, Accumulated Leap Seconds=32
  { 2452670.50, 64.489353 }, //31 January 2003, UT1-UTC=-0.3053530, Accumulated Leap Seconds=32
  { 2452671.50, 64.489338 }, //01 February 2003, UT1-UTC=-0.3053377, Accumulated Leap Seconds=32
  { 2452672.50, 64.489421 }, //02 February 2003, UT1-UTC=-0.3054213, Accumulated Leap Seconds=32
  { 2452673.50, 64.489683 }, //03 February 2003, UT1-UTC=-0.3056832, Accumulated Leap Seconds=32
  { 2452674.50, 64.490173 }, //04 February 2003, UT1-UTC=-0.3061727, Accumulated Leap Seconds=32
  { 2452675.50, 64.490869 }, //05 February 2003, UT1-UTC=-0.3068687, Accumulated Leap Seconds=32
  { 2452676.50, 64.491646 }, //06 February 2003, UT1-UTC=-0.3076458, Accumulated Leap Seconds=32
  { 2452677.50, 64.492425 }, //07 February 2003, UT1-UTC=-0.3084246, Accumulated Leap Seconds=32
  { 2452678.50, 64.493132 }, //08 February 2003, UT1-UTC=-0.3091320, Accumulated Leap Seconds=32
  { 2452679.50, 64.493734 }, //09 February 2003, UT1-UTC=-0.3097340, Accumulated Leap Seconds=32
  { 2452680.50, 64.494225 }, //10 February 2003, UT1-UTC=-0.3102252, Accumulated Leap Seconds=32
  { 2452681.50, 64.494588 }, //11 February 2003, UT1-UTC=-0.3105885, Accumulated Leap Seconds=32
  { 2452682.50, 64.494842 }, //12 February 2003, UT1-UTC=-0.3108423, Accumulated Leap Seconds=32
  { 2452683.50, 64.495079 }, //13 February 2003, UT1-UTC=-0.3110792, Accumulated Leap Seconds=32
  { 2452684.50, 64.495373 }, //14 February 2003, UT1-UTC=-0.3113728, Accumulated Leap Seconds=32
  { 2452685.50, 64.495730 }, //15 February 2003, UT1-UTC=-0.3117305, Accumulated Leap Seconds=32
  { 2452686.50, 64.496182 }, //16 February 2003, UT1-UTC=-0.3121823, Accumulated Leap Seconds=32
  { 2452687.50, 64.496853 }, //17 February 2003, UT1-UTC=-0.3128526, Accumulated Leap Seconds=32
  { 2452688.50, 64.497786 }, //18 February 2003, UT1-UTC=-0.3137861, Accumulated Leap Seconds=32
  { 2452689.50, 64.498935 }, //19 February 2003, UT1-UTC=-0.3149354, Accumulated Leap Seconds=32
  { 2452690.50, 64.500186 }, //20 February 2003, UT1-UTC=-0.3161861, Accumulated Leap Seconds=32
  { 2452691.50, 64.501426 }, //21 February 2003, UT1-UTC=-0.3174260, Accumulated Leap Seconds=32
  { 2452692.50, 64.502518 }, //22 February 2003, UT1-UTC=-0.3185178, Accumulated Leap Seconds=32
  { 2452693.50, 64.503363 }, //23 February 2003, UT1-UTC=-0.3193628, Accumulated Leap Seconds=32
  { 2452694.50, 64.503957 }, //24 February 2003, UT1-UTC=-0.3199575, Accumulated Leap Seconds=32
  { 2452695.50, 64.504348 }, //25 February 2003, UT1-UTC=-0.3203477, Accumulated Leap Seconds=32
  { 2452696.50, 64.504606 }, //26 February 2003, UT1-UTC=-0.3206057, Accumulated Leap Seconds=32
  { 2452697.50, 64.504827 }, //27 February 2003, UT1-UTC=-0.3208274, Accumulated Leap Seconds=32
  { 2452698.50, 64.505062 }, //28 February 2003, UT1-UTC=-0.3210616, Accumulated Leap Seconds=32
  { 2452699.50, 64.505334 }, //01 March 2003, UT1-UTC=-0.3213342, Accumulated Leap Seconds=32
  { 2452700.50, 64.505729 }, //02 March 2003, UT1-UTC=-0.3217285, Accumulated Leap Seconds=32
  { 2452701.50, 64.506277 }, //03 March 2003, UT1-UTC=-0.3222769, Accumulated Leap Seconds=32
  { 2452702.50, 64.506934 }, //04 March 2003, UT1-UTC=-0.3229339, Accumulated Leap Seconds=32
  { 2452703.50, 64.507598 }, //05 March 2003, UT1-UTC=-0.3235981, Accumulated Leap Seconds=32
  { 2452704.50, 64.508284 }, //06 March 2003, UT1-UTC=-0.3242836, Accumulated Leap Seconds=32
  { 2452705.50, 64.508988 }, //07 March 2003, UT1-UTC=-0.3249880, Accumulated Leap Seconds=32
  { 2452706.50, 64.509634 }, //08 March 2003, UT1-UTC=-0.3256338, Accumulated Leap Seconds=32
  { 2452707.50, 64.510205 }, //09 March 2003, UT1-UTC=-0.3262047, Accumulated Leap Seconds=32
  { 2452708.50, 64.510699 }, //10 March 2003, UT1-UTC=-0.3266992, Accumulated Leap Seconds=32
  { 2452709.50, 64.511113 }, //11 March 2003, UT1-UTC=-0.3271130, Accumulated Leap Seconds=32
  { 2452710.50, 64.511441 }, //12 March 2003, UT1-UTC=-0.3274410, Accumulated Leap Seconds=32
  { 2452711.50, 64.511767 }, //13 March 2003, UT1-UTC=-0.3277667, Accumulated Leap Seconds=32
  { 2452712.50, 64.512122 }, //14 March 2003, UT1-UTC=-0.3281217, Accumulated Leap Seconds=32
  { 2452713.50, 64.512623 }, //15 March 2003, UT1-UTC=-0.3286228, Accumulated Leap Seconds=32
  { 2452714.50, 64.513353 }, //16 March 2003, UT1-UTC=-0.3293535, Accumulated Leap Seconds=32
  { 2452715.50, 64.514382 }, //17 March 2003, UT1-UTC=-0.3303825, Accumulated Leap Seconds=32
  { 2452716.50, 64.515706 }, //18 March 2003, UT1-UTC=-0.3317064, Accumulated Leap Seconds=32
  { 2452717.50, 64.517215 }, //19 March 2003, UT1-UTC=-0.3332146, Accumulated Leap Seconds=32
  { 2452718.50, 64.518755 }, //20 March 2003, UT1-UTC=-0.3347550, Accumulated Leap Seconds=32
  { 2452719.50, 64.520155 }, //21 March 2003, UT1-UTC=-0.3361548, Accumulated Leap Seconds=32
  { 2452720.50, 64.521291 }, //22 March 2003, UT1-UTC=-0.3372912, Accumulated Leap Seconds=32
  { 2452721.50, 64.522120 }, //23 March 2003, UT1-UTC=-0.3381200, Accumulated Leap Seconds=32
  { 2452722.50, 64.522663 }, //24 March 2003, UT1-UTC=-0.3386631, Accumulated Leap Seconds=32
  { 2452723.50, 64.523002 }, //25 March 2003, UT1-UTC=-0.3390021, Accumulated Leap Seconds=32
  { 2452724.50, 64.523243 }, //26 March 2003, UT1-UTC=-0.3392432, Accumulated Leap Seconds=32
  { 2452725.50, 64.523530 }, //27 March 2003, UT1-UTC=-0.3395301, Accumulated Leap Seconds=32
  { 2452726.50, 64.523936 }, //28 March 2003, UT1-UTC=-0.3399356, Accumulated Leap Seconds=32
  { 2452727.50, 64.524478 }, //29 March 2003, UT1-UTC=-0.3404782, Accumulated Leap Seconds=32
  { 2452728.50, 64.525174 }, //30 March 2003, UT1-UTC=-0.3411741, Accumulated Leap Seconds=32
  { 2452729.50, 64.526002 }, //31 March 2003, UT1-UTC=-0.3420023, Accumulated Leap Seconds=32
  { 2452730.50, 64.526919 }, //01 April 2003, UT1-UTC=-0.3429189, Accumulated Leap Seconds=32
  { 2452731.50, 64.527875 }, //02 April 2003, UT1-UTC=-0.3438752, Accumulated Leap Seconds=32
  { 2452732.50, 64.528859 }, //03 April 2003, UT1-UTC=-0.3448588, Accumulated Leap Seconds=32
  { 2452733.50, 64.529789 }, //04 April 2003, UT1-UTC=-0.3457889, Accumulated Leap Seconds=32
  { 2452734.50, 64.530603 }, //05 April 2003, UT1-UTC=-0.3466028, Accumulated Leap Seconds=32
  { 2452735.50, 64.531323 }, //06 April 2003, UT1-UTC=-0.3473228, Accumulated Leap Seconds=32
  { 2452736.50, 64.531916 }, //07 April 2003, UT1-UTC=-0.3479159, Accumulated Leap Seconds=32
  { 2452737.50, 64.532364 }, //08 April 2003, UT1-UTC=-0.3483635, Accumulated Leap Seconds=32
  { 2452738.50, 64.532696 }, //09 April 2003, UT1-UTC=-0.3486963, Accumulated Leap Seconds=32
  { 2452739.50, 64.533016 }, //10 April 2003, UT1-UTC=-0.3490161, Accumulated Leap Seconds=32
  { 2452740.50, 64.533410 }, //11 April 2003, UT1-UTC=-0.3494104, Accumulated Leap Seconds=32
  { 2452741.50, 64.533957 }, //12 April 2003, UT1-UTC=-0.3499565, Accumulated Leap Seconds=32
  { 2452742.50, 64.534757 }, //13 April 2003, UT1-UTC=-0.3507569, Accumulated Leap Seconds=32
  { 2452743.50, 64.535858 }, //14 April 2003, UT1-UTC=-0.3518583, Accumulated Leap Seconds=32
  { 2452744.50, 64.537221 }, //15 April 2003, UT1-UTC=-0.3532211, Accumulated Leap Seconds=32
  { 2452745.50, 64.538705 }, //16 April 2003, UT1-UTC=-0.3547053, Accumulated Leap Seconds=32
  { 2452746.50, 64.540127 }, //17 April 2003, UT1-UTC=-0.3561266, Accumulated Leap Seconds=32
  { 2452747.50, 64.541296 }, //18 April 2003, UT1-UTC=-0.3572959, Accumulated Leap Seconds=32
  { 2452748.50, 64.542153 }, //19 April 2003, UT1-UTC=-0.3581530, Accumulated Leap Seconds=32
  { 2452749.50, 64.542637 }, //20 April 2003, UT1-UTC=-0.3586367, Accumulated Leap Seconds=32
  { 2452750.50, 64.542822 }, //21 April 2003, UT1-UTC=-0.3588223, Accumulated Leap Seconds=32
  { 2452751.50, 64.542880 }, //22 April 2003, UT1-UTC=-0.3588803, Accumulated Leap Seconds=32
  { 2452752.50, 64.542912 }, //23 April 2003, UT1-UTC=-0.3589122, Accumulated Leap Seconds=32
  { 2452753.50, 64.543029 }, //24 April 2003, UT1-UTC=-0.3590290, Accumulated Leap Seconds=32
  { 2452754.50, 64.543289 }, //25 April 2003, UT1-UTC=-0.3592890, Accumulated Leap Seconds=32
  { 2452755.50, 64.543682 }, //26 April 2003, UT1-UTC=-0.3596822, Accumulated Leap Seconds=32
  { 2452756.50, 64.544228 }, //27 April 2003, UT1-UTC=-0.3602284, Accumulated Leap Seconds=32
  { 2452757.50, 64.544899 }, //28 April 2003, UT1-UTC=-0.3608987, Accumulated Leap Seconds=32
  { 2452758.50, 64.545645 }, //29 April 2003, UT1-UTC=-0.3616448, Accumulated Leap Seconds=32
  { 2452759.50, 64.546403 }, //30 April 2003, UT1-UTC=-0.3624026, Accumulated Leap Seconds=32
  { 2452760.50, 64.547094 }, //01 May 2003, UT1-UTC=-0.3630942, Accumulated Leap Seconds=32
  { 2452761.50, 64.547628 }, //02 May 2003, UT1-UTC=-0.3636280, Accumulated Leap Seconds=32
  { 2452762.50, 64.547973 }, //03 May 2003, UT1-UTC=-0.3639732, Accumulated Leap Seconds=32
  { 2452763.50, 64.548128 }, //04 May 2003, UT1-UTC=-0.3641276, Accumulated Leap Seconds=32
  { 2452764.50, 64.548114 }, //05 May 2003, UT1-UTC=-0.3641144, Accumulated Leap Seconds=32
  { 2452765.50, 64.548018 }, //06 May 2003, UT1-UTC=-0.3640185, Accumulated Leap Seconds=32
  { 2452766.50, 64.547916 }, //07 May 2003, UT1-UTC=-0.3639165, Accumulated Leap Seconds=32
  { 2452767.50, 64.547832 }, //08 May 2003, UT1-UTC=-0.3638322, Accumulated Leap Seconds=32
  { 2452768.50, 64.547906 }, //09 May 2003, UT1-UTC=-0.3639056, Accumulated Leap Seconds=32
  { 2452769.50, 64.548200 }, //10 May 2003, UT1-UTC=-0.3642002, Accumulated Leap Seconds=32
  { 2452770.50, 64.548731 }, //11 May 2003, UT1-UTC=-0.3647308, Accumulated Leap Seconds=32
  { 2452771.50, 64.549519 }, //12 May 2003, UT1-UTC=-0.3655190, Accumulated Leap Seconds=32
  { 2452772.50, 64.550554 }, //13 May 2003, UT1-UTC=-0.3665539, Accumulated Leap Seconds=32
  { 2452773.50, 64.551757 }, //14 May 2003, UT1-UTC=-0.3677575, Accumulated Leap Seconds=32
  { 2452774.50, 64.552942 }, //15 May 2003, UT1-UTC=-0.3689420, Accumulated Leap Seconds=32
  { 2452775.50, 64.553910 }, //16 May 2003, UT1-UTC=-0.3699096, Accumulated Leap Seconds=32
  { 2452776.50, 64.554540 }, //17 May 2003, UT1-UTC=-0.3705404, Accumulated Leap Seconds=32
  { 2452777.50, 64.554887 }, //18 May 2003, UT1-UTC=-0.3708873, Accumulated Leap Seconds=32
  { 2452778.50, 64.555057 }, //19 May 2003, UT1-UTC=-0.3710574, Accumulated Leap Seconds=32
  { 2452779.50, 64.555209 }, //20 May 2003, UT1-UTC=-0.3712090, Accumulated Leap Seconds=32
  { 2452780.50, 64.555425 }, //21 May 2003, UT1-UTC=-0.3714253, Accumulated Leap Seconds=32
  { 2452781.50, 64.555742 }, //22 May 2003, UT1-UTC=-0.3717424, Accumulated Leap Seconds=32
  { 2452782.50, 64.556198 }, //23 May 2003, UT1-UTC=-0.3721978, Accumulated Leap Seconds=32
  { 2452783.50, 64.556768 }, //24 May 2003, UT1-UTC=-0.3727682, Accumulated Leap Seconds=32
  { 2452784.50, 64.557413 }, //25 May 2003, UT1-UTC=-0.3734133, Accumulated Leap Seconds=32
  { 2452785.50, 64.558060 }, //26 May 2003, UT1-UTC=-0.3740601, Accumulated Leap Seconds=32
  { 2452786.50, 64.558653 }, //27 May 2003, UT1-UTC=-0.3746528, Accumulated Leap Seconds=32
  { 2452787.50, 64.559158 }, //28 May 2003, UT1-UTC=-0.3751579, Accumulated Leap Seconds=32
  { 2452788.50, 64.559540 }, //29 May 2003, UT1-UTC=-0.3755405, Accumulated Leap Seconds=32
  { 2452789.50, 64.559772 }, //30 May 2003, UT1-UTC=-0.3757718, Accumulated Leap Seconds=32
  { 2452790.50, 64.559823 }, //31 May 2003, UT1-UTC=-0.3758233, Accumulated Leap Seconds=32
  { 2452791.50, 64.559673 }, //01 June 2003, UT1-UTC=-0.3756729, Accumulated Leap Seconds=32
  { 2452792.50, 64.559367 }, //02 June 2003, UT1-UTC=-0.3753667, Accumulated Leap Seconds=32
  { 2452793.50, 64.558970 }, //03 June 2003, UT1-UTC=-0.3749699, Accumulated Leap Seconds=32
  { 2452794.50, 64.558515 }, //04 June 2003, UT1-UTC=-0.3745145, Accumulated Leap Seconds=32
  { 2452795.50, 64.558060 }, //05 June 2003, UT1-UTC=-0.3740604, Accumulated Leap Seconds=32
  { 2452796.50, 64.557745 }, //06 June 2003, UT1-UTC=-0.3737450, Accumulated Leap Seconds=32
  { 2452797.50, 64.557661 }, //07 June 2003, UT1-UTC=-0.3736611, Accumulated Leap Seconds=32
  { 2452798.50, 64.557832 }, //08 June 2003, UT1-UTC=-0.3738321, Accumulated Leap Seconds=32
  { 2452799.50, 64.558238 }, //09 June 2003, UT1-UTC=-0.3742377, Accumulated Leap Seconds=32
  { 2452800.50, 64.558771 }, //10 June 2003, UT1-UTC=-0.3747707, Accumulated Leap Seconds=32
  { 2452801.50, 64.559271 }, //11 June 2003, UT1-UTC=-0.3752706, Accumulated Leap Seconds=32
  { 2452802.50, 64.559596 }, //12 June 2003, UT1-UTC=-0.3755955, Accumulated Leap Seconds=32
  { 2452803.50, 64.559629 }, //13 June 2003, UT1-UTC=-0.3756285, Accumulated Leap Seconds=32
  { 2452804.50, 64.559332 }, //14 June 2003, UT1-UTC=-0.3753325, Accumulated Leap Seconds=32
  { 2452805.50, 64.558783 }, //15 June 2003, UT1-UTC=-0.3747830, Accumulated Leap Seconds=32
  { 2452806.50, 64.558104 }, //16 June 2003, UT1-UTC=-0.3741045, Accumulated Leap Seconds=32
  { 2452807.50, 64.557434 }, //17 June 2003, UT1-UTC=-0.3734337, Accumulated Leap Seconds=32
  { 2452808.50, 64.556872 }, //18 June 2003, UT1-UTC=-0.3728718, Accumulated Leap Seconds=32
  { 2452809.50, 64.556472 }, //19 June 2003, UT1-UTC=-0.3724717, Accumulated Leap Seconds=32
  { 2452810.50, 64.556240 }, //20 June 2003, UT1-UTC=-0.3722403, Accumulated Leap Seconds=32
  { 2452811.50, 64.556122 }, //21 June 2003, UT1-UTC=-0.3721221, Accumulated Leap Seconds=32
  { 2452812.50, 64.556050 }, //22 June 2003, UT1-UTC=-0.3720500, Accumulated Leap Seconds=32
  { 2452813.50, 64.555965 }, //23 June 2003, UT1-UTC=-0.3719645, Accumulated Leap Seconds=32
  { 2452814.50, 64.555811 }, //24 June 2003, UT1-UTC=-0.3718111, Accumulated Leap Seconds=32
  { 2452815.50, 64.555541 }, //25 June 2003, UT1-UTC=-0.3715413, Accumulated Leap Seconds=32
  { 2452816.50, 64.555126 }, //26 June 2003, UT1-UTC=-0.3711262, Accumulated Leap Seconds=32
  { 2452817.50, 64.554566 }, //27 June 2003, UT1-UTC=-0.3705659, Accumulated Leap Seconds=32
  { 2452818.50, 64.553862 }, //28 June 2003, UT1-UTC=-0.3698619, Accumulated Leap Seconds=32
  { 2452819.50, 64.553037 }, //29 June 2003, UT1-UTC=-0.3690369, Accumulated Leap Seconds=32
  { 2452820.50, 64.552135 }, //30 June 2003, UT1-UTC=-0.3681352, Accumulated Leap Seconds=32
  { 2452821.50, 64.551229 }, //01 July 2003, UT1-UTC=-0.3672293, Accumulated Leap Seconds=32
  { 2452822.50, 64.550398 }, //02 July 2003, UT1-UTC=-0.3663979, Accumulated Leap Seconds=32
  { 2452823.50, 64.549688 }, //03 July 2003, UT1-UTC=-0.3656883, Accumulated Leap Seconds=32
  { 2452824.50, 64.549135 }, //04 July 2003, UT1-UTC=-0.3651348, Accumulated Leap Seconds=32
  { 2452825.50, 64.548773 }, //05 July 2003, UT1-UTC=-0.3647734, Accumulated Leap Seconds=32
  { 2452826.50, 64.548561 }, //06 July 2003, UT1-UTC=-0.3645612, Accumulated Leap Seconds=32
  { 2452827.50, 64.548420 }, //07 July 2003, UT1-UTC=-0.3644204, Accumulated Leap Seconds=32
  { 2452828.50, 64.548263 }, //08 July 2003, UT1-UTC=-0.3642627, Accumulated Leap Seconds=32
  { 2452829.50, 64.548005 }, //09 July 2003, UT1-UTC=-0.3640050, Accumulated Leap Seconds=32
  { 2452830.50, 64.547533 }, //10 July 2003, UT1-UTC=-0.3635334, Accumulated Leap Seconds=32
  { 2452831.50, 64.546816 }, //11 July 2003, UT1-UTC=-0.3628156, Accumulated Leap Seconds=32
  { 2452832.50, 64.545917 }, //12 July 2003, UT1-UTC=-0.3619172, Accumulated Leap Seconds=32
  { 2452833.50, 64.544953 }, //13 July 2003, UT1-UTC=-0.3609529, Accumulated Leap Seconds=32
  { 2452834.50, 64.544025 }, //14 July 2003, UT1-UTC=-0.3600250, Accumulated Leap Seconds=32
  { 2452835.50, 64.543223 }, //15 July 2003, UT1-UTC=-0.3592228, Accumulated Leap Seconds=32
  { 2452836.50, 64.542619 }, //16 July 2003, UT1-UTC=-0.3586193, Accumulated Leap Seconds=32
  { 2452837.50, 64.542238 }, //17 July 2003, UT1-UTC=-0.3582378, Accumulated Leap Seconds=32
  { 2452838.50, 64.542022 }, //18 July 2003, UT1-UTC=-0.3580219, Accumulated Leap Seconds=32
  { 2452839.50, 64.541894 }, //19 July 2003, UT1-UTC=-0.3578936, Accumulated Leap Seconds=32
  { 2452840.50, 64.541786 }, //20 July 2003, UT1-UTC=-0.3577857, Accumulated Leap Seconds=32
  { 2452841.50, 64.541634 }, //21 July 2003, UT1-UTC=-0.3576341, Accumulated Leap Seconds=32
  { 2452842.50, 64.541400 }, //22 July 2003, UT1-UTC=-0.3573996, Accumulated Leap Seconds=32
  { 2452843.50, 64.541060 }, //23 July 2003, UT1-UTC=-0.3570603, Accumulated Leap Seconds=32
  { 2452844.50, 64.540607 }, //24 July 2003, UT1-UTC=-0.3566067, Accumulated Leap Seconds=32
  { 2452845.50, 64.540044 }, //25 July 2003, UT1-UTC=-0.3560443, Accumulated Leap Seconds=32
  { 2452846.50, 64.539390 }, //26 July 2003, UT1-UTC=-0.3553902, Accumulated Leap Seconds=32
  { 2452847.50, 64.538730 }, //27 July 2003, UT1-UTC=-0.3547300, Accumulated Leap Seconds=32
  { 2452848.50, 64.538123 }, //28 July 2003, UT1-UTC=-0.3541233, Accumulated Leap Seconds=32
  { 2452849.50, 64.537607 }, //29 July 2003, UT1-UTC=-0.3536072, Accumulated Leap Seconds=32
  { 2452850.50, 64.537214 }, //30 July 2003, UT1-UTC=-0.3532143, Accumulated Leap Seconds=32
  { 2452851.50, 64.537033 }, //31 July 2003, UT1-UTC=-0.3530332, Accumulated Leap Seconds=32
  { 2452852.50, 64.537091 }, //01 August 2003, UT1-UTC=-0.3530906, Accumulated Leap Seconds=32
  { 2452853.50, 64.537318 }, //02 August 2003, UT1-UTC=-0.3533179, Accumulated Leap Seconds=32
  { 2452854.50, 64.537656 }, //03 August 2003, UT1-UTC=-0.3536562, Accumulated Leap Seconds=32
  { 2452855.50, 64.538037 }, //04 August 2003, UT1-UTC=-0.3540369, Accumulated Leap Seconds=32
  { 2452856.50, 64.538321 }, //05 August 2003, UT1-UTC=-0.3543207, Accumulated Leap Seconds=32
  { 2452857.50, 64.538360 }, //06 August 2003, UT1-UTC=-0.3543599, Accumulated Leap Seconds=32
  { 2452858.50, 64.538162 }, //07 August 2003, UT1-UTC=-0.3541618, Accumulated Leap Seconds=32
  { 2452859.50, 64.537744 }, //08 August 2003, UT1-UTC=-0.3537440, Accumulated Leap Seconds=32
  { 2452860.50, 64.537178 }, //09 August 2003, UT1-UTC=-0.3531782, Accumulated Leap Seconds=32
  { 2452861.50, 64.536588 }, //10 August 2003, UT1-UTC=-0.3525882, Accumulated Leap Seconds=32
  { 2452862.50, 64.536087 }, //11 August 2003, UT1-UTC=-0.3520874, Accumulated Leap Seconds=32
  { 2452863.50, 64.535768 }, //12 August 2003, UT1-UTC=-0.3517684, Accumulated Leap Seconds=32
  { 2452864.50, 64.535677 }, //13 August 2003, UT1-UTC=-0.3516766, Accumulated Leap Seconds=32
  { 2452865.50, 64.535766 }, //14 August 2003, UT1-UTC=-0.3517656, Accumulated Leap Seconds=32
  { 2452866.50, 64.535988 }, //15 August 2003, UT1-UTC=-0.3519884, Accumulated Leap Seconds=32
  { 2452867.50, 64.536294 }, //16 August 2003, UT1-UTC=-0.3522936, Accumulated Leap Seconds=32
  { 2452868.50, 64.536603 }, //17 August 2003, UT1-UTC=-0.3526033, Accumulated Leap Seconds=32
  { 2452869.50, 64.536838 }, //18 August 2003, UT1-UTC=-0.3528384, Accumulated Leap Seconds=32
  { 2452870.50, 64.536904 }, //19 August 2003, UT1-UTC=-0.3529045, Accumulated Leap Seconds=32
  { 2452871.50, 64.536729 }, //20 August 2003, UT1-UTC=-0.3527289, Accumulated Leap Seconds=32
  { 2452872.50, 64.536360 }, //21 August 2003, UT1-UTC=-0.3523604, Accumulated Leap Seconds=32
  { 2452873.50, 64.535842 }, //22 August 2003, UT1-UTC=-0.3518420, Accumulated Leap Seconds=32
  { 2452874.50, 64.535221 }, //23 August 2003, UT1-UTC=-0.3512211, Accumulated Leap Seconds=32
  { 2452875.50, 64.534581 }, //24 August 2003, UT1-UTC=-0.3505807, Accumulated Leap Seconds=32
  { 2452876.50, 64.533987 }, //25 August 2003, UT1-UTC=-0.3499870, Accumulated Leap Seconds=32
  { 2452877.50, 64.533531 }, //26 August 2003, UT1-UTC=-0.3495314, Accumulated Leap Seconds=32
  { 2452878.50, 64.533293 }, //27 August 2003, UT1-UTC=-0.3492927, Accumulated Leap Seconds=32
  { 2452879.50, 64.533415 }, //28 August 2003, UT1-UTC=-0.3494149, Accumulated Leap Seconds=32
  { 2452880.50, 64.533862 }, //29 August 2003, UT1-UTC=-0.3498616, Accumulated Leap Seconds=32
  { 2452881.50, 64.534525 }, //30 August 2003, UT1-UTC=-0.3505251, Accumulated Leap Seconds=32
  { 2452882.50, 64.535276 }, //31 August 2003, UT1-UTC=-0.3512761, Accumulated Leap Seconds=32
  { 2452883.50, 64.535947 }, //01 September 2003, UT1-UTC=-0.3519472, Accumulated Leap Seconds=32
  { 2452884.50, 64.536416 }, //02 September 2003, UT1-UTC=-0.3524163, Accumulated Leap Seconds=32
  { 2452885.50, 64.536634 }, //03 September 2003, UT1-UTC=-0.3526344, Accumulated Leap Seconds=32
  { 2452886.50, 64.536586 }, //04 September 2003, UT1-UTC=-0.3525856, Accumulated Leap Seconds=32
  { 2452887.50, 64.536358 }, //05 September 2003, UT1-UTC=-0.3523584, Accumulated Leap Seconds=32
  { 2452888.50, 64.536069 }, //06 September 2003, UT1-UTC=-0.3520693, Accumulated Leap Seconds=32
  { 2452889.50, 64.535775 }, //07 September 2003, UT1-UTC=-0.3517748, Accumulated Leap Seconds=32
  { 2452890.50, 64.535576 }, //08 September 2003, UT1-UTC=-0.3515758, Accumulated Leap Seconds=32
  { 2452891.50, 64.535556 }, //09 September 2003, UT1-UTC=-0.3515564, Accumulated Leap Seconds=32
  { 2452892.50, 64.535737 }, //10 September 2003, UT1-UTC=-0.3517370, Accumulated Leap Seconds=32
  { 2452893.50, 64.536060 }, //11 September 2003, UT1-UTC=-0.3520603, Accumulated Leap Seconds=32
  { 2452894.50, 64.536460 }, //12 September 2003, UT1-UTC=-0.3524604, Accumulated Leap Seconds=32
  { 2452895.50, 64.536833 }, //13 September 2003, UT1-UTC=-0.3528330, Accumulated Leap Seconds=32
  { 2452896.50, 64.537132 }, //14 September 2003, UT1-UTC=-0.3531319, Accumulated Leap Seconds=32
  { 2452897.50, 64.537343 }, //15 September 2003, UT1-UTC=-0.3533428, Accumulated Leap Seconds=32
  { 2452898.50, 64.537453 }, //16 September 2003, UT1-UTC=-0.3534535, Accumulated Leap Seconds=32
  { 2452899.50, 64.537458 }, //17 September 2003, UT1-UTC=-0.3534576, Accumulated Leap Seconds=32
  { 2452900.50, 64.537346 }, //18 September 2003, UT1-UTC=-0.3533465, Accumulated Leap Seconds=32
  { 2452901.50, 64.537089 }, //19 September 2003, UT1-UTC=-0.3530887, Accumulated Leap Seconds=32
  { 2452902.50, 64.536797 }, //20 September 2003, UT1-UTC=-0.3527968, Accumulated Leap Seconds=32
  { 2452903.50, 64.536534 }, //21 September 2003, UT1-UTC=-0.3525335, Accumulated Leap Seconds=32
  { 2452904.50, 64.536389 }, //22 September 2003, UT1-UTC=-0.3523885, Accumulated Leap Seconds=32
  { 2452905.50, 64.536469 }, //23 September 2003, UT1-UTC=-0.3524686, Accumulated Leap Seconds=32
  { 2452906.50, 64.536814 }, //24 September 2003, UT1-UTC=-0.3528143, Accumulated Leap Seconds=32
  { 2452907.50, 64.537419 }, //25 September 2003, UT1-UTC=-0.3534192, Accumulated Leap Seconds=32
  { 2452908.50, 64.538241 }, //26 September 2003, UT1-UTC=-0.3542415, Accumulated Leap Seconds=32
  { 2452909.50, 64.539172 }, //27 September 2003, UT1-UTC=-0.3551718, Accumulated Leap Seconds=32
  { 2452910.50, 64.540060 }, //28 September 2003, UT1-UTC=-0.3560603, Accumulated Leap Seconds=32
  { 2452911.50, 64.540776 }, //29 September 2003, UT1-UTC=-0.3567764, Accumulated Leap Seconds=32
  { 2452912.50, 64.541263 }, //30 September 2003, UT1-UTC=-0.3572628, Accumulated Leap Seconds=32
  { 2452913.50, 64.541495 }, //01 October 2003, UT1-UTC=-0.3574947, Accumulated Leap Seconds=32
  { 2452914.50, 64.541543 }, //02 October 2003, UT1-UTC=-0.3575434, Accumulated Leap Seconds=32
  { 2452915.50, 64.541525 }, //03 October 2003, UT1-UTC=-0.3575253, Accumulated Leap Seconds=32
  { 2452916.50, 64.541556 }, //04 October 2003, UT1-UTC=-0.3575563, Accumulated Leap Seconds=32
  { 2452917.50, 64.541712 }, //05 October 2003, UT1-UTC=-0.3577116, Accumulated Leap Seconds=32
  { 2452918.50, 64.542050 }, //06 October 2003, UT1-UTC=-0.3580504, Accumulated Leap Seconds=32
  { 2452919.50, 64.542600 }, //07 October 2003, UT1-UTC=-0.3586005, Accumulated Leap Seconds=32
  { 2452920.50, 64.543308 }, //08 October 2003, UT1-UTC=-0.3593077, Accumulated Leap Seconds=32
  { 2452921.50, 64.544054 }, //09 October 2003, UT1-UTC=-0.3600542, Accumulated Leap Seconds=32
  { 2452922.50, 64.544757 }, //10 October 2003, UT1-UTC=-0.3607569, Accumulated Leap Seconds=32
  { 2452923.50, 64.545405 }, //11 October 2003, UT1-UTC=-0.3614052, Accumulated Leap Seconds=32
  { 2452924.50, 64.545945 }, //12 October 2003, UT1-UTC=-0.3619451, Accumulated Leap Seconds=32
  { 2452925.50, 64.546344 }, //13 October 2003, UT1-UTC=-0.3623444, Accumulated Leap Seconds=32
  { 2452926.50, 64.546594 }, //14 October 2003, UT1-UTC=-0.3625940, Accumulated Leap Seconds=32
  { 2452927.50, 64.546701 }, //15 October 2003, UT1-UTC=-0.3627009, Accumulated Leap Seconds=32
  { 2452928.50, 64.546685 }, //16 October 2003, UT1-UTC=-0.3626855, Accumulated Leap Seconds=32
  { 2452929.50, 64.546600 }, //17 October 2003, UT1-UTC=-0.3625996, Accumulated Leap Seconds=32
  { 2452930.50, 64.546527 }, //18 October 2003, UT1-UTC=-0.3625268, Accumulated Leap Seconds=32
  { 2452931.50, 64.546573 }, //19 October 2003, UT1-UTC=-0.3625735, Accumulated Leap Seconds=32
  { 2452932.50, 64.546835 }, //20 October 2003, UT1-UTC=-0.3628352, Accumulated Leap Seconds=32
  { 2452933.50, 64.547382 }, //21 October 2003, UT1-UTC=-0.3633816, Accumulated Leap Seconds=32
  { 2452934.50, 64.548221 }, //22 October 2003, UT1-UTC=-0.3642207, Accumulated Leap Seconds=32
  { 2452935.50, 64.549318 }, //23 October 2003, UT1-UTC=-0.3653182, Accumulated Leap Seconds=32
  { 2452936.50, 64.550603 }, //24 October 2003, UT1-UTC=-0.3666035, Accumulated Leap Seconds=32
  { 2452937.50, 64.551915 }, //25 October 2003, UT1-UTC=-0.3679152, Accumulated Leap Seconds=32
  { 2452938.50, 64.553075 }, //26 October 2003, UT1-UTC=-0.3690749, Accumulated Leap Seconds=32
  { 2452939.50, 64.553934 }, //27 October 2003, UT1-UTC=-0.3699340, Accumulated Leap Seconds=32
  { 2452940.50, 64.554426 }, //28 October 2003, UT1-UTC=-0.3704257, Accumulated Leap Seconds=32
  { 2452941.50, 64.554584 }, //29 October 2003, UT1-UTC=-0.3705842, Accumulated Leap Seconds=32
  { 2452942.50, 64.554542 }, //30 October 2003, UT1-UTC=-0.3705420, Accumulated Leap Seconds=32
  { 2452943.50, 64.554434 }, //31 October 2003, UT1-UTC=-0.3704336, Accumulated Leap Seconds=32
  { 2452944.50, 64.554363 }, //01 November 2003, UT1-UTC=-0.3703634, Accumulated Leap Seconds=32
  { 2452945.50, 64.554425 }, //02 November 2003, UT1-UTC=-0.3704255, Accumulated Leap Seconds=32
  { 2452946.50, 64.554670 }, //03 November 2003, UT1-UTC=-0.3706700, Accumulated Leap Seconds=32
  { 2452947.50, 64.555095 }, //04 November 2003, UT1-UTC=-0.3710949, Accumulated Leap Seconds=32
  { 2452948.50, 64.555645 }, //05 November 2003, UT1-UTC=-0.3716455, Accumulated Leap Seconds=32
  { 2452949.50, 64.556216 }, //06 November 2003, UT1-UTC=-0.3722165, Accumulated Leap Seconds=32
  { 2452950.50, 64.556738 }, //07 November 2003, UT1-UTC=-0.3727379, Accumulated Leap Seconds=32
  { 2452951.50, 64.557167 }, //08 November 2003, UT1-UTC=-0.3731670, Accumulated Leap Seconds=32
  { 2452952.50, 64.557499 }, //09 November 2003, UT1-UTC=-0.3734991, Accumulated Leap Seconds=32
  { 2452953.50, 64.557710 }, //10 November 2003, UT1-UTC=-0.3737101, Accumulated Leap Seconds=32
  { 2452954.50, 64.557806 }, //11 November 2003, UT1-UTC=-0.3738058, Accumulated Leap Seconds=32
  { 2452955.50, 64.557742 }, //12 November 2003, UT1-UTC=-0.3737423, Accumulated Leap Seconds=32
  { 2452956.50, 64.557685 }, //13 November 2003, UT1-UTC=-0.3736849, Accumulated Leap Seconds=32
  { 2452957.50, 64.557687 }, //14 November 2003, UT1-UTC=-0.3736874, Accumulated Leap Seconds=32
  { 2452958.50, 64.557744 }, //15 November 2003, UT1-UTC=-0.3737445, Accumulated Leap Seconds=32
  { 2452959.50, 64.557904 }, //16 November 2003, UT1-UTC=-0.3739045, Accumulated Leap Seconds=32
  { 2452960.50, 64.558257 }, //17 November 2003, UT1-UTC=-0.3742567, Accumulated Leap Seconds=32
  { 2452961.50, 64.558858 }, //18 November 2003, UT1-UTC=-0.3748576, Accumulated Leap Seconds=32
  { 2452962.50, 64.559673 }, //19 November 2003, UT1-UTC=-0.3756732, Accumulated Leap Seconds=32
  { 2452963.50, 64.560614 }, //20 November 2003, UT1-UTC=-0.3766143, Accumulated Leap Seconds=32
  { 2452964.50, 64.561657 }, //21 November 2003, UT1-UTC=-0.3776569, Accumulated Leap Seconds=32
  { 2452965.50, 64.562759 }, //22 November 2003, UT1-UTC=-0.3787591, Accumulated Leap Seconds=32
  { 2452966.50, 64.563692 }, //23 November 2003, UT1-UTC=-0.3796923, Accumulated Leap Seconds=32
  { 2452967.50, 64.564340 }, //24 November 2003, UT1-UTC=-0.3803395, Accumulated Leap Seconds=32
  { 2452968.50, 64.564691 }, //25 November 2003, UT1-UTC=-0.3806915, Accumulated Leap Seconds=32
  { 2452969.50, 64.564766 }, //26 November 2003, UT1-UTC=-0.3807660, Accumulated Leap Seconds=32
  { 2452970.50, 64.564708 }, //27 November 2003, UT1-UTC=-0.3807079, Accumulated Leap Seconds=32
  { 2452971.50, 64.564674 }, //28 November 2003, UT1-UTC=-0.3806736, Accumulated Leap Seconds=32
  { 2452972.50, 64.564777 }, //29 November 2003, UT1-UTC=-0.3807771, Accumulated Leap Seconds=32
  { 2452973.50, 64.565038 }, //30 November 2003, UT1-UTC=-0.3810378, Accumulated Leap Seconds=32
  { 2452974.50, 64.565430 }, //01 December 2003, UT1-UTC=-0.3814298, Accumulated Leap Seconds=32
  { 2452975.50, 64.565926 }, //02 December 2003, UT1-UTC=-0.3819260, Accumulated Leap Seconds=32
  { 2452976.50, 64.566461 }, //03 December 2003, UT1-UTC=-0.3824611, Accumulated Leap Seconds=32
  { 2452977.50, 64.566960 }, //04 December 2003, UT1-UTC=-0.3829600, Accumulated Leap Seconds=32
  { 2452978.50, 64.567344 }, //05 December 2003, UT1-UTC=-0.3833443, Accumulated Leap Seconds=32
  { 2452979.50, 64.567626 }, //06 December 2003, UT1-UTC=-0.3836264, Accumulated Leap Seconds=32
  { 2452980.50, 64.567789 }, //07 December 2003, UT1-UTC=-0.3837894, Accumulated Leap Seconds=32
  { 2452981.50, 64.567784 }, //08 December 2003, UT1-UTC=-0.3837836, Accumulated Leap Seconds=32
  { 2452982.50, 64.567567 }, //09 December 2003, UT1-UTC=-0.3835670, Accumulated Leap Seconds=32
  { 2452983.50, 64.567196 }, //10 December 2003, UT1-UTC=-0.3831965, Accumulated Leap Seconds=32
  { 2452984.50, 64.566808 }, //11 December 2003, UT1-UTC=-0.3828080, Accumulated Leap Seconds=32
  { 2452985.50, 64.566431 }, //12 December 2003, UT1-UTC=-0.3824312, Accumulated Leap Seconds=32
  { 2452986.50, 64.566144 }, //13 December 2003, UT1-UTC=-0.3821437, Accumulated Leap Seconds=32
  { 2452987.50, 64.565973 }, //14 December 2003, UT1-UTC=-0.3819729, Accumulated Leap Seconds=32
  { 2452988.50, 64.565996 }, //15 December 2003, UT1-UTC=-0.3819964, Accumulated Leap Seconds=32
  { 2452989.50, 64.566251 }, //16 December 2003, UT1-UTC=-0.3822506, Accumulated Leap Seconds=32
  { 2452990.50, 64.566753 }, //17 December 2003, UT1-UTC=-0.3827534, Accumulated Leap Seconds=32
  { 2452991.50, 64.567453 }, //18 December 2003, UT1-UTC=-0.3834535, Accumulated Leap Seconds=32
  { 2452992.50, 64.568227 }, //19 December 2003, UT1-UTC=-0.3842275, Accumulated Leap Seconds=32
  { 2452993.50, 64.569000 }, //20 December 2003, UT1-UTC=-0.3849995, Accumulated Leap Seconds=32
  { 2452994.50, 64.569678 }, //21 December 2003, UT1-UTC=-0.3856785, Accumulated Leap Seconds=32
  { 2452995.50, 64.570129 }, //22 December 2003, UT1-UTC=-0.3861289, Accumulated Leap Seconds=32
  { 2452996.50, 64.570305 }, //23 December 2003, UT1-UTC=-0.3863051, Accumulated Leap Seconds=32
  { 2452997.50, 64.570312 }, //24 December 2003, UT1-UTC=-0.3863118, Accumulated Leap Seconds=32
  { 2452998.50, 64.570380 }, //25 December 2003, UT1-UTC=-0.3863800, Accumulated Leap Seconds=32
  { 2452999.50, 64.570559 }, //26 December 2003, UT1-UTC=-0.3865590, Accumulated Leap Seconds=32
  { 2453000.50, 64.570881 }, //27 December 2003, UT1-UTC=-0.3868809, Accumulated Leap Seconds=32
  { 2453001.50, 64.571344 }, //28 December 2003, UT1-UTC=-0.3873439, Accumulated Leap Seconds=32
  { 2453002.50, 64.571905 }, //29 December 2003, UT1-UTC=-0.3879047, Accumulated Leap Seconds=32
  { 2453003.50, 64.572512 }, //30 December 2003, UT1-UTC=-0.3885122, Accumulated Leap Seconds=32
  { 2453004.50, 64.573094 }, //31 December 2003, UT1-UTC=-0.3890938, Accumulated Leap Seconds=32
  { 2453005.50, 64.573611 }, //01 January 2004, UT1-UTC=-0.3896111, Accumulated Leap Seconds=32
  { 2453006.50, 64.574077 }, //02 January 2004, UT1-UTC=-0.3900767, Accumulated Leap Seconds=32
  { 2453007.50, 64.574426 }, //03 January 2004, UT1-UTC=-0.3904257, Accumulated Leap Seconds=32
  { 2453008.50, 64.574593 }, //04 January 2004, UT1-UTC=-0.3905932, Accumulated Leap Seconds=32
  { 2453009.50, 64.574604 }, //05 January 2004, UT1-UTC=-0.3906040, Accumulated Leap Seconds=32
  { 2453010.50, 64.574521 }, //06 January 2004, UT1-UTC=-0.3905213, Accumulated Leap Seconds=32
  { 2453011.50, 64.574442 }, //07 January 2004, UT1-UTC=-0.3904423, Accumulated Leap Seconds=32
  { 2453012.50, 64.574462 }, //08 January 2004, UT1-UTC=-0.3904618, Accumulated Leap Seconds=32
  { 2453013.50, 64.574610 }, //09 January 2004, UT1-UTC=-0.3906104, Accumulated Leap Seconds=32
  { 2453014.50, 64.574973 }, //10 January 2004, UT1-UTC=-0.3909731, Accumulated Leap Seconds=32
  { 2453015.50, 64.575587 }, //11 January 2004, UT1-UTC=-0.3915875, Accumulated Leap Seconds=32
  { 2453016.50, 64.576462 }, //12 January 2004, UT1-UTC=-0.3924623, Accumulated Leap Seconds=32
  { 2453017.50, 64.577570 }, //13 January 2004, UT1-UTC=-0.3935699, Accumulated Leap Seconds=32
  { 2453018.50, 64.578842 }, //14 January 2004, UT1-UTC=-0.3948418, Accumulated Leap Seconds=32
  { 2453019.50, 64.580110 }, //15 January 2004, UT1-UTC=-0.3961097, Accumulated Leap Seconds=32
  { 2453020.50, 64.581263 }, //16 January 2004, UT1-UTC=-0.3972627, Accumulated Leap Seconds=32
  { 2453021.50, 64.582224 }, //17 January 2004, UT1-UTC=-0.3982241, Accumulated Leap Seconds=32
  { 2453022.50, 64.582920 }, //18 January 2004, UT1-UTC=-0.3989195, Accumulated Leap Seconds=32
  { 2453023.50, 64.583378 }, //19 January 2004, UT1-UTC=-0.3993777, Accumulated Leap Seconds=32
  { 2453024.50, 64.583642 }, //20 January 2004, UT1-UTC=-0.3996423, Accumulated Leap Seconds=32
  { 2453025.50, 64.583840 }, //21 January 2004, UT1-UTC=-0.3998404, Accumulated Leap Seconds=32
  { 2453026.50, 64.584069 }, //22 January 2004, UT1-UTC=-0.4000692, Accumulated Leap Seconds=32
  { 2453027.50, 64.584433 }, //23 January 2004, UT1-UTC=-0.4004329, Accumulated Leap Seconds=32
  { 2453028.50, 64.584949 }, //24 January 2004, UT1-UTC=-0.4009494, Accumulated Leap Seconds=32
  { 2453029.50, 64.585619 }, //25 January 2004, UT1-UTC=-0.4016188, Accumulated Leap Seconds=32
  { 2453030.50, 64.586424 }, //26 January 2004, UT1-UTC=-0.4024243, Accumulated Leap Seconds=32
  { 2453031.50, 64.587300 }, //27 January 2004, UT1-UTC=-0.4033002, Accumulated Leap Seconds=32
  { 2453032.50, 64.588071 }, //28 January 2004, UT1-UTC=-0.4040711, Accumulated Leap Seconds=32
  { 2453033.50, 64.588664 }, //29 January 2004, UT1-UTC=-0.4046641, Accumulated Leap Seconds=32
  { 2453034.50, 64.589043 }, //30 January 2004, UT1-UTC=-0.4050432, Accumulated Leap Seconds=32
  { 2453035.50, 64.589164 }, //31 January 2004, UT1-UTC=-0.4051645, Accumulated Leap Seconds=32
  { 2453036.50, 64.589114 }, //01 February 2004, UT1-UTC=-0.4051142, Accumulated Leap Seconds=32
  { 2453037.50, 64.588944 }, //02 February 2004, UT1-UTC=-0.4049439, Accumulated Leap Seconds=32
  { 2453038.50, 64.588702 }, //03 February 2004, UT1-UTC=-0.4047021, Accumulated Leap Seconds=32
  { 2453039.50, 64.588425 }, //04 February 2004, UT1-UTC=-0.4044246, Accumulated Leap Seconds=32
  { 2453040.50, 64.588180 }, //05 February 2004, UT1-UTC=-0.4041796, Accumulated Leap Seconds=32
  { 2453041.50, 64.588114 }, //06 February 2004, UT1-UTC=-0.4041138, Accumulated Leap Seconds=32
  { 2453042.50, 64.588221 }, //07 February 2004, UT1-UTC=-0.4042213, Accumulated Leap Seconds=32
  { 2453043.50, 64.588633 }, //08 February 2004, UT1-UTC=-0.4046330, Accumulated Leap Seconds=32
  { 2453044.50, 64.589292 }, //09 February 2004, UT1-UTC=-0.4052923, Accumulated Leap Seconds=32
  { 2453045.50, 64.590121 }, //10 February 2004, UT1-UTC=-0.4061206, Accumulated Leap Seconds=32
  { 2453046.50, 64.591039 }, //11 February 2004, UT1-UTC=-0.4070395, Accumulated Leap Seconds=32
  { 2453047.50, 64.591956 }, //12 February 2004, UT1-UTC=-0.4079564, Accumulated Leap Seconds=32
  { 2453048.50, 64.592755 }, //13 February 2004, UT1-UTC=-0.4087555, Accumulated Leap Seconds=32
  { 2453049.50, 64.593345 }, //14 February 2004, UT1-UTC=-0.4093455, Accumulated Leap Seconds=32
  { 2453050.50, 64.593695 }, //15 February 2004, UT1-UTC=-0.4096955, Accumulated Leap Seconds=32
  { 2453051.50, 64.593864 }, //16 February 2004, UT1-UTC=-0.4098643, Accumulated Leap Seconds=32
  { 2453052.50, 64.593954 }, //17 February 2004, UT1-UTC=-0.4099541, Accumulated Leap Seconds=32
  { 2453053.50, 64.594058 }, //18 February 2004, UT1-UTC=-0.4100584, Accumulated Leap Seconds=32
  { 2453054.50, 64.594318 }, //19 February 2004, UT1-UTC=-0.4103177, Accumulated Leap Seconds=32
  { 2453055.50, 64.594811 }, //20 February 2004, UT1-UTC=-0.4108109, Accumulated Leap Seconds=32
  { 2453056.50, 64.595554 }, //21 February 2004, UT1-UTC=-0.4115541, Accumulated Leap Seconds=32
  { 2453057.50, 64.596469 }, //22 February 2004, UT1-UTC=-0.4124685, Accumulated Leap Seconds=32
  { 2453058.50, 64.597518 }, //23 February 2004, UT1-UTC=-0.4135182, Accumulated Leap Seconds=32
  { 2453059.50, 64.598597 }, //24 February 2004, UT1-UTC=-0.4145967, Accumulated Leap Seconds=32
  { 2453060.50, 64.599570 }, //25 February 2004, UT1-UTC=-0.4155697, Accumulated Leap Seconds=32
  { 2453061.50, 64.600364 }, //26 February 2004, UT1-UTC=-0.4163641, Accumulated Leap Seconds=32
  { 2453062.50, 64.600966 }, //27 February 2004, UT1-UTC=-0.4169658, Accumulated Leap Seconds=32
  { 2453063.50, 64.601377 }, //28 February 2004, UT1-UTC=-0.4173773, Accumulated Leap Seconds=32
  { 2453064.50, 64.601528 }, //29 February 2004, UT1-UTC=-0.4175276, Accumulated Leap Seconds=32
  { 2453065.50, 64.601476 }, //01 March 2004, UT1-UTC=-0.4174759, Accumulated Leap Seconds=32
  { 2453066.50, 64.601334 }, //02 March 2004, UT1-UTC=-0.4173344, Accumulated Leap Seconds=32
  { 2453067.50, 64.601212 }, //03 March 2004, UT1-UTC=-0.4172123, Accumulated Leap Seconds=32
  { 2453068.50, 64.601185 }, //04 March 2004, UT1-UTC=-0.4171845, Accumulated Leap Seconds=32
  { 2453069.50, 64.601341 }, //05 March 2004, UT1-UTC=-0.4173411, Accumulated Leap Seconds=32
  { 2453070.50, 64.601716 }, //06 March 2004, UT1-UTC=-0.4177157, Accumulated Leap Seconds=32
  { 2453071.50, 64.602335 }, //07 March 2004, UT1-UTC=-0.4183349, Accumulated Leap Seconds=32
  { 2453072.50, 64.603174 }, //08 March 2004, UT1-UTC=-0.4191739, Accumulated Leap Seconds=32
  { 2453073.50, 64.604115 }, //09 March 2004, UT1-UTC=-0.4201152, Accumulated Leap Seconds=32
  { 2453074.50, 64.605010 }, //10 March 2004, UT1-UTC=-0.4210096, Accumulated Leap Seconds=32
  { 2453075.50, 64.605733 }, //11 March 2004, UT1-UTC=-0.4217335, Accumulated Leap Seconds=32
  { 2453076.50, 64.606202 }, //12 March 2004, UT1-UTC=-0.4222020, Accumulated Leap Seconds=32
  { 2453077.50, 64.606444 }, //13 March 2004, UT1-UTC=-0.4224436, Accumulated Leap Seconds=32
  { 2453078.50, 64.606502 }, //14 March 2004, UT1-UTC=-0.4225018, Accumulated Leap Seconds=32
  { 2453079.50, 64.606453 }, //15 March 2004, UT1-UTC=-0.4224526, Accumulated Leap Seconds=32
  { 2453080.50, 64.606455 }, //16 March 2004, UT1-UTC=-0.4224550, Accumulated Leap Seconds=32
  { 2453081.50, 64.606566 }, //17 March 2004, UT1-UTC=-0.4225659, Accumulated Leap Seconds=32
  { 2453082.50, 64.606905 }, //18 March 2004, UT1-UTC=-0.4229047, Accumulated Leap Seconds=32
  { 2453083.50, 64.607497 }, //19 March 2004, UT1-UTC=-0.4234972, Accumulated Leap Seconds=32
  { 2453084.50, 64.608361 }, //20 March 2004, UT1-UTC=-0.4243612, Accumulated Leap Seconds=32
  { 2453085.50, 64.609449 }, //21 March 2004, UT1-UTC=-0.4254491, Accumulated Leap Seconds=32
  { 2453086.50, 64.610591 }, //22 March 2004, UT1-UTC=-0.4265908, Accumulated Leap Seconds=32
  { 2453087.50, 64.611720 }, //23 March 2004, UT1-UTC=-0.4277204, Accumulated Leap Seconds=32
  { 2453088.50, 64.612809 }, //24 March 2004, UT1-UTC=-0.4288092, Accumulated Leap Seconds=32
  { 2453089.50, 64.613777 }, //25 March 2004, UT1-UTC=-0.4297774, Accumulated Leap Seconds=32
  { 2453090.50, 64.614597 }, //26 March 2004, UT1-UTC=-0.4305967, Accumulated Leap Seconds=32
  { 2453091.50, 64.615243 }, //27 March 2004, UT1-UTC=-0.4312433, Accumulated Leap Seconds=32
  { 2453092.50, 64.615725 }, //28 March 2004, UT1-UTC=-0.4317248, Accumulated Leap Seconds=32
  { 2453093.50, 64.616107 }, //29 March 2004, UT1-UTC=-0.4321075, Accumulated Leap Seconds=32
  { 2453094.50, 64.616497 }, //30 March 2004, UT1-UTC=-0.4324971, Accumulated Leap Seconds=32
  { 2453095.50, 64.616972 }, //31 March 2004, UT1-UTC=-0.4329721, Accumulated Leap Seconds=32
  { 2453096.50, 64.617615 }, //01 April 2004, UT1-UTC=-0.4336147, Accumulated Leap Seconds=32
  { 2453097.50, 64.618473 }, //02 April 2004, UT1-UTC=-0.4344729, Accumulated Leap Seconds=32
  { 2453098.50, 64.619545 }, //03 April 2004, UT1-UTC=-0.4355449, Accumulated Leap Seconds=32
  { 2453099.50, 64.620876 }, //04 April 2004, UT1-UTC=-0.4368762, Accumulated Leap Seconds=32
  { 2453100.50, 64.622401 }, //05 April 2004, UT1-UTC=-0.4384012, Accumulated Leap Seconds=32
  { 2453101.50, 64.623950 }, //06 April 2004, UT1-UTC=-0.4399498, Accumulated Leap Seconds=32
  { 2453102.50, 64.625407 }, //07 April 2004, UT1-UTC=-0.4414071, Accumulated Leap Seconds=32
  { 2453103.50, 64.626619 }, //08 April 2004, UT1-UTC=-0.4426187, Accumulated Leap Seconds=32
  { 2453104.50, 64.627511 }, //09 April 2004, UT1-UTC=-0.4435111, Accumulated Leap Seconds=32
  { 2453105.50, 64.628179 }, //10 April 2004, UT1-UTC=-0.4441791, Accumulated Leap Seconds=32
  { 2453106.50, 64.628632 }, //11 April 2004, UT1-UTC=-0.4446324, Accumulated Leap Seconds=32
  { 2453107.50, 64.628954 }, //12 April 2004, UT1-UTC=-0.4449541, Accumulated Leap Seconds=32
  { 2453108.50, 64.629344 }, //13 April 2004, UT1-UTC=-0.4453438, Accumulated Leap Seconds=32
  { 2453109.50, 64.629929 }, //14 April 2004, UT1-UTC=-0.4459293, Accumulated Leap Seconds=32
  { 2453110.50, 64.630731 }, //15 April 2004, UT1-UTC=-0.4467310, Accumulated Leap Seconds=32
  { 2453111.50, 64.631683 }, //16 April 2004, UT1-UTC=-0.4476832, Accumulated Leap Seconds=32
  { 2453112.50, 64.632730 }, //17 April 2004, UT1-UTC=-0.4487301, Accumulated Leap Seconds=32
  { 2453113.50, 64.633770 }, //18 April 2004, UT1-UTC=-0.4497703, Accumulated Leap Seconds=32
  { 2453114.50, 64.634733 }, //19 April 2004, UT1-UTC=-0.4507330, Accumulated Leap Seconds=32
  { 2453115.50, 64.635550 }, //20 April 2004, UT1-UTC=-0.4515498, Accumulated Leap Seconds=32
  { 2453116.50, 64.636139 }, //21 April 2004, UT1-UTC=-0.4521395, Accumulated Leap Seconds=32
  { 2453117.50, 64.636502 }, //22 April 2004, UT1-UTC=-0.4525024, Accumulated Leap Seconds=32
  { 2453118.50, 64.636630 }, //23 April 2004, UT1-UTC=-0.4526300, Accumulated Leap Seconds=32
  { 2453119.50, 64.636541 }, //24 April 2004, UT1-UTC=-0.4525408, Accumulated Leap Seconds=32
  { 2453120.50, 64.636348 }, //25 April 2004, UT1-UTC=-0.4523480, Accumulated Leap Seconds=32
  { 2453121.50, 64.636118 }, //26 April 2004, UT1-UTC=-0.4521176, Accumulated Leap Seconds=32
  { 2453122.50, 64.635968 }, //27 April 2004, UT1-UTC=-0.4519678, Accumulated Leap Seconds=32
  { 2453123.50, 64.635976 }, //28 April 2004, UT1-UTC=-0.4519755, Accumulated Leap Seconds=32
  { 2453124.50, 64.636210 }, //29 April 2004, UT1-UTC=-0.4522099, Accumulated Leap Seconds=32
  { 2453125.50, 64.636676 }, //30 April 2004, UT1-UTC=-0.4526761, Accumulated Leap Seconds=32
  { 2453126.50, 64.637440 }, //01 May 2004, UT1-UTC=-0.4534397, Accumulated Leap Seconds=32
  { 2453127.50, 64.638446 }, //02 May 2004, UT1-UTC=-0.4544456, Accumulated Leap Seconds=32
  { 2453128.50, 64.639569 }, //03 May 2004, UT1-UTC=-0.4555694, Accumulated Leap Seconds=32
  { 2453129.50, 64.640704 }, //04 May 2004, UT1-UTC=-0.4567044, Accumulated Leap Seconds=32
  { 2453130.50, 64.641756 }, //05 May 2004, UT1-UTC=-0.4577565, Accumulated Leap Seconds=32
  { 2453131.50, 64.642608 }, //06 May 2004, UT1-UTC=-0.4586077, Accumulated Leap Seconds=32
  { 2453132.50, 64.643242 }, //07 May 2004, UT1-UTC=-0.4592423, Accumulated Leap Seconds=32
  { 2453133.50, 64.643682 }, //08 May 2004, UT1-UTC=-0.4596825, Accumulated Leap Seconds=32
  { 2453134.50, 64.644009 }, //09 May 2004, UT1-UTC=-0.4600094, Accumulated Leap Seconds=32
  { 2453135.50, 64.644378 }, //10 May 2004, UT1-UTC=-0.4603782, Accumulated Leap Seconds=32
  { 2453136.50, 64.644881 }, //11 May 2004, UT1-UTC=-0.4608811, Accumulated Leap Seconds=32
  { 2453137.50, 64.645540 }, //12 May 2004, UT1-UTC=-0.4615400, Accumulated Leap Seconds=32
  { 2453138.50, 64.646362 }, //13 May 2004, UT1-UTC=-0.4623622, Accumulated Leap Seconds=32
  { 2453139.50, 64.647326 }, //14 May 2004, UT1-UTC=-0.4633256, Accumulated Leap Seconds=32
  { 2453140.50, 64.648366 }, //15 May 2004, UT1-UTC=-0.4643657, Accumulated Leap Seconds=32
  { 2453141.50, 64.649406 }, //16 May 2004, UT1-UTC=-0.4654063, Accumulated Leap Seconds=32
  { 2453142.50, 64.650345 }, //17 May 2004, UT1-UTC=-0.4663453, Accumulated Leap Seconds=32
  { 2453143.50, 64.651149 }, //18 May 2004, UT1-UTC=-0.4671494, Accumulated Leap Seconds=32
  { 2453144.50, 64.651751 }, //19 May 2004, UT1-UTC=-0.4677510, Accumulated Leap Seconds=32
  { 2453145.50, 64.652165 }, //20 May 2004, UT1-UTC=-0.4681655, Accumulated Leap Seconds=32
  { 2453146.50, 64.652355 }, //21 May 2004, UT1-UTC=-0.4683547, Accumulated Leap Seconds=32
  { 2453147.50, 64.652424 }, //22 May 2004, UT1-UTC=-0.4684237, Accumulated Leap Seconds=32
  { 2453148.50, 64.652387 }, //23 May 2004, UT1-UTC=-0.4683869, Accumulated Leap Seconds=32
  { 2453149.50, 64.652279 }, //24 May 2004, UT1-UTC=-0.4682793, Accumulated Leap Seconds=32
  { 2453150.50, 64.652262 }, //25 May 2004, UT1-UTC=-0.4682624, Accumulated Leap Seconds=32
  { 2453151.50, 64.652304 }, //26 May 2004, UT1-UTC=-0.4683043, Accumulated Leap Seconds=32
  { 2453152.50, 64.652480 }, //27 May 2004, UT1-UTC=-0.4684797, Accumulated Leap Seconds=32
  { 2453153.50, 64.652850 }, //28 May 2004, UT1-UTC=-0.4688498, Accumulated Leap Seconds=32
  { 2453154.50, 64.653328 }, //29 May 2004, UT1-UTC=-0.4693282, Accumulated Leap Seconds=32
  { 2453155.50, 64.653875 }, //30 May 2004, UT1-UTC=-0.4698745, Accumulated Leap Seconds=32
  { 2453156.50, 64.654411 }, //31 May 2004, UT1-UTC=-0.4704110, Accumulated Leap Seconds=32
  { 2453157.50, 64.654867 }, //01 June 2004, UT1-UTC=-0.4708674, Accumulated Leap Seconds=32
  { 2453158.50, 64.655151 }, //02 June 2004, UT1-UTC=-0.4711511, Accumulated Leap Seconds=32
  { 2453159.50, 64.655172 }, //03 June 2004, UT1-UTC=-0.4711724, Accumulated Leap Seconds=32
  { 2453160.50, 64.655015 }, //04 June 2004, UT1-UTC=-0.4710146, Accumulated Leap Seconds=32
  { 2453161.50, 64.654717 }, //05 June 2004, UT1-UTC=-0.4707173, Accumulated Leap Seconds=32
  { 2453162.50, 64.654440 }, //06 June 2004, UT1-UTC=-0.4704398, Accumulated Leap Seconds=32
  { 2453163.50, 64.654302 }, //07 June 2004, UT1-UTC=-0.4703015, Accumulated Leap Seconds=32
  { 2453164.50, 64.654372 }, //08 June 2004, UT1-UTC=-0.4703722, Accumulated Leap Seconds=32
  { 2453165.50, 64.654595 }, //09 June 2004, UT1-UTC=-0.4705952, Accumulated Leap Seconds=32
  { 2453166.50, 64.654982 }, //10 June 2004, UT1-UTC=-0.4709819, Accumulated Leap Seconds=32
  { 2453167.50, 64.655425 }, //11 June 2004, UT1-UTC=-0.4714248, Accumulated Leap Seconds=32
  { 2453168.50, 64.655897 }, //12 June 2004, UT1-UTC=-0.4718974, Accumulated Leap Seconds=32
  { 2453169.50, 64.656309 }, //13 June 2004, UT1-UTC=-0.4723091, Accumulated Leap Seconds=32
  { 2453170.50, 64.656571 }, //14 June 2004, UT1-UTC=-0.4725712, Accumulated Leap Seconds=32
  { 2453171.50, 64.656623 }, //15 June 2004, UT1-UTC=-0.4726232, Accumulated Leap Seconds=32
  { 2453172.50, 64.656415 }, //16 June 2004, UT1-UTC=-0.4724155, Accumulated Leap Seconds=32
  { 2453173.50, 64.656003 }, //17 June 2004, UT1-UTC=-0.4720032, Accumulated Leap Seconds=32
  { 2453174.50, 64.655465 }, //18 June 2004, UT1-UTC=-0.4714649, Accumulated Leap Seconds=32
  { 2453175.50, 64.654858 }, //19 June 2004, UT1-UTC=-0.4708583, Accumulated Leap Seconds=32
  { 2453176.50, 64.654293 }, //20 June 2004, UT1-UTC=-0.4702929, Accumulated Leap Seconds=32
  { 2453177.50, 64.653771 }, //21 June 2004, UT1-UTC=-0.4697710, Accumulated Leap Seconds=32
  { 2453178.50, 64.653367 }, //22 June 2004, UT1-UTC=-0.4693670, Accumulated Leap Seconds=32
  { 2453179.50, 64.653125 }, //23 June 2004, UT1-UTC=-0.4691251, Accumulated Leap Seconds=32
  { 2453180.50, 64.653029 }, //24 June 2004, UT1-UTC=-0.4690291, Accumulated Leap Seconds=32
  { 2453181.50, 64.653118 }, //25 June 2004, UT1-UTC=-0.4691181, Accumulated Leap Seconds=32
  { 2453182.50, 64.653343 }, //26 June 2004, UT1-UTC=-0.4693427, Accumulated Leap Seconds=32
  { 2453183.50, 64.653547 }, //27 June 2004, UT1-UTC=-0.4695472, Accumulated Leap Seconds=32
  { 2453184.50, 64.653683 }, //28 June 2004, UT1-UTC=-0.4696828, Accumulated Leap Seconds=32
  { 2453185.50, 64.653669 }, //29 June 2004, UT1-UTC=-0.4696693, Accumulated Leap Seconds=32
  { 2453186.50, 64.653469 }, //30 June 2004, UT1-UTC=-0.4694692, Accumulated Leap Seconds=32
  { 2453187.50, 64.653002 }, //01 July 2004, UT1-UTC=-0.4690021, Accumulated Leap Seconds=32
  { 2453188.50, 64.652374 }, //02 July 2004, UT1-UTC=-0.4683745, Accumulated Leap Seconds=32
  { 2453189.50, 64.651605 }, //03 July 2004, UT1-UTC=-0.4676054, Accumulated Leap Seconds=32
  { 2453190.50, 64.650881 }, //04 July 2004, UT1-UTC=-0.4668810, Accumulated Leap Seconds=32
  { 2453191.50, 64.650358 }, //05 July 2004, UT1-UTC=-0.4663585, Accumulated Leap Seconds=32
  { 2453192.50, 64.650073 }, //06 July 2004, UT1-UTC=-0.4660735, Accumulated Leap Seconds=32
  { 2453193.50, 64.649948 }, //07 July 2004, UT1-UTC=-0.4659482, Accumulated Leap Seconds=32
  { 2453194.50, 64.649866 }, //08 July 2004, UT1-UTC=-0.4658656, Accumulated Leap Seconds=32
  { 2453195.50, 64.649747 }, //09 July 2004, UT1-UTC=-0.4657474, Accumulated Leap Seconds=32
  { 2453196.50, 64.649525 }, //10 July 2004, UT1-UTC=-0.4655254, Accumulated Leap Seconds=32
  { 2453197.50, 64.649147 }, //11 July 2004, UT1-UTC=-0.4651475, Accumulated Leap Seconds=32
  { 2453198.50, 64.648604 }, //12 July 2004, UT1-UTC=-0.4646036, Accumulated Leap Seconds=32
  { 2453199.50, 64.647873 }, //13 July 2004, UT1-UTC=-0.4638727, Accumulated Leap Seconds=32
  { 2453200.50, 64.646948 }, //14 July 2004, UT1-UTC=-0.4629481, Accumulated Leap Seconds=32
  { 2453201.50, 64.645923 }, //15 July 2004, UT1-UTC=-0.4619229, Accumulated Leap Seconds=32
  { 2453202.50, 64.644871 }, //16 July 2004, UT1-UTC=-0.4608714, Accumulated Leap Seconds=32
  { 2453203.50, 64.643860 }, //17 July 2004, UT1-UTC=-0.4598601, Accumulated Leap Seconds=32
  { 2453204.50, 64.642968 }, //18 July 2004, UT1-UTC=-0.4589683, Accumulated Leap Seconds=32
  { 2453205.50, 64.642231 }, //19 July 2004, UT1-UTC=-0.4582311, Accumulated Leap Seconds=32
  { 2453206.50, 64.641669 }, //20 July 2004, UT1-UTC=-0.4576692, Accumulated Leap Seconds=32
  { 2453207.50, 64.641318 }, //21 July 2004, UT1-UTC=-0.4573183, Accumulated Leap Seconds=32
  { 2453208.50, 64.641191 }, //22 July 2004, UT1-UTC=-0.4571907, Accumulated Leap Seconds=32
  { 2453209.50, 64.641210 }, //23 July 2004, UT1-UTC=-0.4572103, Accumulated Leap Seconds=32
  { 2453210.50, 64.641298 }, //24 July 2004, UT1-UTC=-0.4572983, Accumulated Leap Seconds=32
  { 2453211.50, 64.641357 }, //25 July 2004, UT1-UTC=-0.4573568, Accumulated Leap Seconds=32
  { 2453212.50, 64.641283 }, //26 July 2004, UT1-UTC=-0.4572825, Accumulated Leap Seconds=32
  { 2453213.50, 64.640994 }, //27 July 2004, UT1-UTC=-0.4569944, Accumulated Leap Seconds=32
  { 2453214.50, 64.640476 }, //28 July 2004, UT1-UTC=-0.4564759, Accumulated Leap Seconds=32
  { 2453215.50, 64.639774 }, //29 July 2004, UT1-UTC=-0.4557740, Accumulated Leap Seconds=32
  { 2453216.50, 64.639027 }, //30 July 2004, UT1-UTC=-0.4550266, Accumulated Leap Seconds=32
  { 2453217.50, 64.638375 }, //31 July 2004, UT1-UTC=-0.4543754, Accumulated Leap Seconds=32
  { 2453218.50, 64.637927 }, //01 August 2004, UT1-UTC=-0.4539271, Accumulated Leap Seconds=32
  { 2453219.50, 64.637780 }, //02 August 2004, UT1-UTC=-0.4537798, Accumulated Leap Seconds=32
  { 2453220.50, 64.637901 }, //03 August 2004, UT1-UTC=-0.4539005, Accumulated Leap Seconds=32
  { 2453221.50, 64.638152 }, //04 August 2004, UT1-UTC=-0.4541519, Accumulated Leap Seconds=32
  { 2453222.50, 64.638403 }, //05 August 2004, UT1-UTC=-0.4544035, Accumulated Leap Seconds=32
  { 2453223.50, 64.638557 }, //06 August 2004, UT1-UTC=-0.4545573, Accumulated Leap Seconds=32
  { 2453224.50, 64.638614 }, //07 August 2004, UT1-UTC=-0.4546140, Accumulated Leap Seconds=32
  { 2453225.50, 64.638519 }, //08 August 2004, UT1-UTC=-0.4545193, Accumulated Leap Seconds=32
  { 2453226.50, 64.638252 }, //09 August 2004, UT1-UTC=-0.4542521, Accumulated Leap Seconds=32
  { 2453227.50, 64.637819 }, //10 August 2004, UT1-UTC=-0.4538186, Accumulated Leap Seconds=32
  { 2453228.50, 64.637231 }, //11 August 2004, UT1-UTC=-0.4532315, Accumulated Leap Seconds=32
  { 2453229.50, 64.636618 }, //12 August 2004, UT1-UTC=-0.4526178, Accumulated Leap Seconds=32
  { 2453230.50, 64.635964 }, //13 August 2004, UT1-UTC=-0.4519642, Accumulated Leap Seconds=32
  { 2453231.50, 64.635396 }, //14 August 2004, UT1-UTC=-0.4513955, Accumulated Leap Seconds=32
  { 2453232.50, 64.634946 }, //15 August 2004, UT1-UTC=-0.4509461, Accumulated Leap Seconds=32
  { 2453233.50, 64.634690 }, //16 August 2004, UT1-UTC=-0.4506899, Accumulated Leap Seconds=32
  { 2453234.50, 64.634677 }, //17 August 2004, UT1-UTC=-0.4506768, Accumulated Leap Seconds=32
  { 2453235.50, 64.634907 }, //18 August 2004, UT1-UTC=-0.4509071, Accumulated Leap Seconds=32
  { 2453236.50, 64.635323 }, //19 August 2004, UT1-UTC=-0.4513232, Accumulated Leap Seconds=32
  { 2453237.50, 64.635857 }, //20 August 2004, UT1-UTC=-0.4518570, Accumulated Leap Seconds=32
  { 2453238.50, 64.636402 }, //21 August 2004, UT1-UTC=-0.4524024, Accumulated Leap Seconds=32
  { 2453239.50, 64.636818 }, //22 August 2004, UT1-UTC=-0.4528184, Accumulated Leap Seconds=32
  { 2453240.50, 64.637012 }, //23 August 2004, UT1-UTC=-0.4530124, Accumulated Leap Seconds=32
  { 2453241.50, 64.636983 }, //24 August 2004, UT1-UTC=-0.4529826, Accumulated Leap Seconds=32
  { 2453242.50, 64.636811 }, //25 August 2004, UT1-UTC=-0.4528113, Accumulated Leap Seconds=32
  { 2453243.50, 64.636466 }, //26 August 2004, UT1-UTC=-0.4524656, Accumulated Leap Seconds=32
  { 2453244.50, 64.636110 }, //27 August 2004, UT1-UTC=-0.4521097, Accumulated Leap Seconds=32
  { 2453245.50, 64.635840 }, //28 August 2004, UT1-UTC=-0.4518399, Accumulated Leap Seconds=32
  { 2453246.50, 64.635820 }, //29 August 2004, UT1-UTC=-0.4518199, Accumulated Leap Seconds=32
  { 2453247.50, 64.636083 }, //30 August 2004, UT1-UTC=-0.4520827, Accumulated Leap Seconds=32
  { 2453248.50, 64.636575 }, //31 August 2004, UT1-UTC=-0.4525752, Accumulated Leap Seconds=32
  { 2453249.50, 64.637161 }, //01 September 2004, UT1-UTC=-0.4531610, Accumulated Leap Seconds=32
  { 2453250.50, 64.637711 }, //02 September 2004, UT1-UTC=-0.4537106, Accumulated Leap Seconds=32
  { 2453251.50, 64.638086 }, //03 September 2004, UT1-UTC=-0.4540860, Accumulated Leap Seconds=32
  { 2453252.50, 64.638275 }, //04 September 2004, UT1-UTC=-0.4542748, Accumulated Leap Seconds=32
  { 2453253.50, 64.638195 }, //05 September 2004, UT1-UTC=-0.4541946, Accumulated Leap Seconds=32
  { 2453254.50, 64.637811 }, //06 September 2004, UT1-UTC=-0.4538114, Accumulated Leap Seconds=32
  { 2453255.50, 64.637192 }, //07 September 2004, UT1-UTC=-0.4531921, Accumulated Leap Seconds=32
  { 2453256.50, 64.636424 }, //08 September 2004, UT1-UTC=-0.4524242, Accumulated Leap Seconds=32
  { 2453257.50, 64.635621 }, //09 September 2004, UT1-UTC=-0.4516210, Accumulated Leap Seconds=32
  { 2453258.50, 64.634887 }, //10 September 2004, UT1-UTC=-0.4508872, Accumulated Leap Seconds=32
  { 2453259.50, 64.634351 }, //11 September 2004, UT1-UTC=-0.4503511, Accumulated Leap Seconds=32
  { 2453260.50, 64.634080 }, //12 September 2004, UT1-UTC=-0.4500804, Accumulated Leap Seconds=32
  { 2453261.50, 64.634102 }, //13 September 2004, UT1-UTC=-0.4501020, Accumulated Leap Seconds=32
  { 2453262.50, 64.634422 }, //14 September 2004, UT1-UTC=-0.4504218, Accumulated Leap Seconds=32
  { 2453263.50, 64.634972 }, //15 September 2004, UT1-UTC=-0.4509723, Accumulated Leap Seconds=32
  { 2453264.50, 64.635640 }, //16 September 2004, UT1-UTC=-0.4516403, Accumulated Leap Seconds=32
  { 2453265.50, 64.636298 }, //17 September 2004, UT1-UTC=-0.4522982, Accumulated Leap Seconds=32
  { 2453266.50, 64.636857 }, //18 September 2004, UT1-UTC=-0.4528574, Accumulated Leap Seconds=32
  { 2453267.50, 64.637157 }, //19 September 2004, UT1-UTC=-0.4531570, Accumulated Leap Seconds=32
  { 2453268.50, 64.637197 }, //20 September 2004, UT1-UTC=-0.4531973, Accumulated Leap Seconds=32
  { 2453269.50, 64.637032 }, //21 September 2004, UT1-UTC=-0.4530320, Accumulated Leap Seconds=32
  { 2453270.50, 64.636743 }, //22 September 2004, UT1-UTC=-0.4527430, Accumulated Leap Seconds=32
  { 2453271.50, 64.636446 }, //23 September 2004, UT1-UTC=-0.4524462, Accumulated Leap Seconds=32
  { 2453272.50, 64.636268 }, //24 September 2004, UT1-UTC=-0.4522676, Accumulated Leap Seconds=32
  { 2453273.50, 64.636329 }, //25 September 2004, UT1-UTC=-0.4523286, Accumulated Leap Seconds=32
  { 2453274.50, 64.636646 }, //26 September 2004, UT1-UTC=-0.4526464, Accumulated Leap Seconds=32
  { 2453275.50, 64.637184 }, //27 September 2004, UT1-UTC=-0.4531841, Accumulated Leap Seconds=32
  { 2453276.50, 64.637930 }, //28 September 2004, UT1-UTC=-0.4539305, Accumulated Leap Seconds=32
  { 2453277.50, 64.638742 }, //29 September 2004, UT1-UTC=-0.4547416, Accumulated Leap Seconds=32
  { 2453278.50, 64.639438 }, //30 September 2004, UT1-UTC=-0.4554377, Accumulated Leap Seconds=32
  { 2453279.50, 64.639961 }, //01 October 2004, UT1-UTC=-0.4559614, Accumulated Leap Seconds=32
  { 2453280.50, 64.640348 }, //02 October 2004, UT1-UTC=-0.4563478, Accumulated Leap Seconds=32
  { 2453281.50, 64.640560 }, //03 October 2004, UT1-UTC=-0.4565597, Accumulated Leap Seconds=32
  { 2453282.50, 64.640634 }, //04 October 2004, UT1-UTC=-0.4566339, Accumulated Leap Seconds=32
  { 2453283.50, 64.640647 }, //05 October 2004, UT1-UTC=-0.4566472, Accumulated Leap Seconds=32
  { 2453284.50, 64.640655 }, //06 October 2004, UT1-UTC=-0.4566548, Accumulated Leap Seconds=32
  { 2453285.50, 64.640741 }, //07 October 2004, UT1-UTC=-0.4567407, Accumulated Leap Seconds=32
  { 2453286.50, 64.640929 }, //08 October 2004, UT1-UTC=-0.4569290, Accumulated Leap Seconds=32
  { 2453287.50, 64.641137 }, //09 October 2004, UT1-UTC=-0.4571373, Accumulated Leap Seconds=32
  { 2453288.50, 64.641579 }, //10 October 2004, UT1-UTC=-0.4575789, Accumulated Leap Seconds=32
  { 2453289.50, 64.642231 }, //11 October 2004, UT1-UTC=-0.4582313, Accumulated Leap Seconds=32
  { 2453290.50, 64.643095 }, //12 October 2004, UT1-UTC=-0.4590946, Accumulated Leap Seconds=32
  { 2453291.50, 64.644153 }, //13 October 2004, UT1-UTC=-0.4601526, Accumulated Leap Seconds=32
  { 2453292.50, 64.645287 }, //14 October 2004, UT1-UTC=-0.4612869, Accumulated Leap Seconds=32
  { 2453293.50, 64.646352 }, //15 October 2004, UT1-UTC=-0.4623517, Accumulated Leap Seconds=32
  { 2453294.50, 64.647298 }, //16 October 2004, UT1-UTC=-0.4632979, Accumulated Leap Seconds=32
  { 2453295.50, 64.647902 }, //17 October 2004, UT1-UTC=-0.4639024, Accumulated Leap Seconds=32
  { 2453296.50, 64.648175 }, //18 October 2004, UT1-UTC=-0.4641753, Accumulated Leap Seconds=32
  { 2453297.50, 64.648233 }, //19 October 2004, UT1-UTC=-0.4642333, Accumulated Leap Seconds=32
  { 2453298.50, 64.648212 }, //20 October 2004, UT1-UTC=-0.4642125, Accumulated Leap Seconds=32
  { 2453299.50, 64.648265 }, //21 October 2004, UT1-UTC=-0.4642655, Accumulated Leap Seconds=32
  { 2453300.50, 64.648504 }, //22 October 2004, UT1-UTC=-0.4645035, Accumulated Leap Seconds=32
  { 2453301.50, 64.648966 }, //23 October 2004, UT1-UTC=-0.4649655, Accumulated Leap Seconds=32
  { 2453302.50, 64.649620 }, //24 October 2004, UT1-UTC=-0.4656198, Accumulated Leap Seconds=32
  { 2453303.50, 64.650428 }, //25 October 2004, UT1-UTC=-0.4664284, Accumulated Leap Seconds=32
  { 2453304.50, 64.651317 }, //26 October 2004, UT1-UTC=-0.4673169, Accumulated Leap Seconds=32
  { 2453305.50, 64.652220 }, //27 October 2004, UT1-UTC=-0.4682203, Accumulated Leap Seconds=32
  { 2453306.50, 64.653041 }, //28 October 2004, UT1-UTC=-0.4690408, Accumulated Leap Seconds=32
  { 2453307.50, 64.653693 }, //29 October 2004, UT1-UTC=-0.4696932, Accumulated Leap Seconds=32
  { 2453308.50, 64.654108 }, //30 October 2004, UT1-UTC=-0.4701077, Accumulated Leap Seconds=32
  { 2453309.50, 64.654286 }, //31 October 2004, UT1-UTC=-0.4702857, Accumulated Leap Seconds=32
  { 2453310.50, 64.654315 }, //01 November 2004, UT1-UTC=-0.4703152, Accumulated Leap Seconds=32
  { 2453311.50, 64.654287 }, //02 November 2004, UT1-UTC=-0.4702875, Accumulated Leap Seconds=32
  { 2453312.50, 64.654266 }, //03 November 2004, UT1-UTC=-0.4702657, Accumulated Leap Seconds=32
  { 2453313.50, 64.654298 }, //04 November 2004, UT1-UTC=-0.4702980, Accumulated Leap Seconds=32
  { 2453314.50, 64.654460 }, //05 November 2004, UT1-UTC=-0.4704602, Accumulated Leap Seconds=32
  { 2453315.50, 64.654849 }, //06 November 2004, UT1-UTC=-0.4708486, Accumulated Leap Seconds=32
  { 2453316.50, 64.655503 }, //07 November 2004, UT1-UTC=-0.4715030, Accumulated Leap Seconds=32
  { 2453317.50, 64.656419 }, //08 November 2004, UT1-UTC=-0.4724189, Accumulated Leap Seconds=32
  { 2453318.50, 64.657556 }, //09 November 2004, UT1-UTC=-0.4735563, Accumulated Leap Seconds=32
  { 2453319.50, 64.658853 }, //10 November 2004, UT1-UTC=-0.4748534, Accumulated Leap Seconds=32
  { 2453320.50, 64.660163 }, //11 November 2004, UT1-UTC=-0.4761631, Accumulated Leap Seconds=32
  { 2453321.50, 64.661341 }, //12 November 2004, UT1-UTC=-0.4773411, Accumulated Leap Seconds=32
  { 2453322.50, 64.662352 }, //13 November 2004, UT1-UTC=-0.4783524, Accumulated Leap Seconds=32
  { 2453323.50, 64.663059 }, //14 November 2004, UT1-UTC=-0.4790588, Accumulated Leap Seconds=32
  { 2453324.50, 64.663503 }, //15 November 2004, UT1-UTC=-0.4795027, Accumulated Leap Seconds=32
  { 2453325.50, 64.663833 }, //16 November 2004, UT1-UTC=-0.4798327, Accumulated Leap Seconds=32
  { 2453326.50, 64.664174 }, //17 November 2004, UT1-UTC=-0.4801739, Accumulated Leap Seconds=32
  { 2453327.50, 64.664665 }, //18 November 2004, UT1-UTC=-0.4806647, Accumulated Leap Seconds=32
  { 2453328.50, 64.665382 }, //19 November 2004, UT1-UTC=-0.4813818, Accumulated Leap Seconds=32
  { 2453329.50, 64.666305 }, //20 November 2004, UT1-UTC=-0.4823048, Accumulated Leap Seconds=32
  { 2453330.50, 64.667391 }, //21 November 2004, UT1-UTC=-0.4833914, Accumulated Leap Seconds=32
  { 2453331.50, 64.668522 }, //22 November 2004, UT1-UTC=-0.4845220, Accumulated Leap Seconds=32
  { 2453332.50, 64.669622 }, //23 November 2004, UT1-UTC=-0.4856220, Accumulated Leap Seconds=32
  { 2453333.50, 64.670604 }, //24 November 2004, UT1-UTC=-0.4866044, Accumulated Leap Seconds=32
  { 2453334.50, 64.671377 }, //25 November 2004, UT1-UTC=-0.4873770, Accumulated Leap Seconds=32
  { 2453335.50, 64.671945 }, //26 November 2004, UT1-UTC=-0.4879449, Accumulated Leap Seconds=32
  { 2453336.50, 64.672264 }, //27 November 2004, UT1-UTC=-0.4882643, Accumulated Leap Seconds=32
  { 2453337.50, 64.672381 }, //28 November 2004, UT1-UTC=-0.4883812, Accumulated Leap Seconds=32
  { 2453338.50, 64.672390 }, //29 November 2004, UT1-UTC=-0.4883904, Accumulated Leap Seconds=32
  { 2453339.50, 64.672361 }, //30 November 2004, UT1-UTC=-0.4883606, Accumulated Leap Seconds=32
  { 2453340.50, 64.672316 }, //01 December 2004, UT1-UTC=-0.4883164, Accumulated Leap Seconds=32
  { 2453341.50, 64.672446 }, //02 December 2004, UT1-UTC=-0.4884463, Accumulated Leap Seconds=32
  { 2453342.50, 64.672732 }, //03 December 2004, UT1-UTC=-0.4887324, Accumulated Leap Seconds=32
  { 2453343.50, 64.673180 }, //04 December 2004, UT1-UTC=-0.4891801, Accumulated Leap Seconds=32
  { 2453344.50, 64.673842 }, //05 December 2004, UT1-UTC=-0.4898425, Accumulated Leap Seconds=32
  { 2453345.50, 64.674709 }, //06 December 2004, UT1-UTC=-0.4907088, Accumulated Leap Seconds=32
  { 2453346.50, 64.675723 }, //07 December 2004, UT1-UTC=-0.4917230, Accumulated Leap Seconds=32
  { 2453347.50, 64.676816 }, //08 December 2004, UT1-UTC=-0.4928163, Accumulated Leap Seconds=32
  { 2453348.50, 64.677858 }, //09 December 2004, UT1-UTC=-0.4938577, Accumulated Leap Seconds=32
  { 2453349.50, 64.678681 }, //10 December 2004, UT1-UTC=-0.4946810, Accumulated Leap Seconds=32
  { 2453350.50, 64.679313 }, //11 December 2004, UT1-UTC=-0.4953129, Accumulated Leap Seconds=32
  { 2453351.50, 64.679699 }, //12 December 2004, UT1-UTC=-0.4956992, Accumulated Leap Seconds=32
  { 2453352.50, 64.679909 }, //13 December 2004, UT1-UTC=-0.4959089, Accumulated Leap Seconds=32
  { 2453353.50, 64.680104 }, //14 December 2004, UT1-UTC=-0.4961039, Accumulated Leap Seconds=32
  { 2453354.50, 64.680483 }, //15 December 2004, UT1-UTC=-0.4964835, Accumulated Leap Seconds=32
  { 2453355.50, 64.681090 }, //16 December 2004, UT1-UTC=-0.4970905, Accumulated Leap Seconds=32
  { 2453356.50, 64.681906 }, //17 December 2004, UT1-UTC=-0.4979061, Accumulated Leap Seconds=32
  { 2453357.50, 64.682904 }, //18 December 2004, UT1-UTC=-0.4989039, Accumulated Leap Seconds=32
  { 2453358.50, 64.683963 }, //19 December 2004, UT1-UTC=-0.4999633, Accumulated Leap Seconds=32
  { 2453359.50, 64.685023 }, //20 December 2004, UT1-UTC=-0.5010228, Accumulated Leap Seconds=32
  { 2453360.50, 64.685941 }, //21 December 2004, UT1-UTC=-0.5019410, Accumulated Leap Seconds=32
  { 2453361.50, 64.686589 }, //22 December 2004, UT1-UTC=-0.5025887, Accumulated Leap Seconds=32
  { 2453362.50, 64.687075 }, //23 December 2004, UT1-UTC=-0.5030749, Accumulated Leap Seconds=32
  { 2453363.50, 64.687352 }, //24 December 2004, UT1-UTC=-0.5033518, Accumulated Leap Seconds=32
  { 2453364.50, 64.687395 }, //25 December 2004, UT1-UTC=-0.5033951, Accumulated Leap Seconds=32
  { 2453365.50, 64.687288 }, //26 December 2004, UT1-UTC=-0.5032880, Accumulated Leap Seconds=32
  { 2453366.50, 64.687131 }, //27 December 2004, UT1-UTC=-0.5031306, Accumulated Leap Seconds=32
  { 2453367.50, 64.687004 }, //28 December 2004, UT1-UTC=-0.5030036, Accumulated Leap Seconds=32
  { 2453368.50, 64.686949 }, //29 December 2004, UT1-UTC=-0.5029488, Accumulated Leap Seconds=32
  { 2453369.50, 64.686973 }, //30 December 2004, UT1-UTC=-0.5029735, Accumulated Leap Seconds=32
  { 2453370.50, 64.687188 }, //31 December 2004, UT1-UTC=-0.5031882, Accumulated Leap Seconds=32
  { 2453371.50, 64.687631 }, //01 January 2005, UT1-UTC=-0.5036311, Accumulated Leap Seconds=32
  { 2453372.50, 64.688263 }, //02 January 2005, UT1-UTC=-0.5042627, Accumulated Leap Seconds=32
  { 2453373.50, 64.689009 }, //03 January 2005, UT1-UTC=-0.5050088, Accumulated Leap Seconds=32
  { 2453374.50, 64.689848 }, //04 January 2005, UT1-UTC=-0.5058481, Accumulated Leap Seconds=32
  { 2453375.50, 64.690691 }, //05 January 2005, UT1-UTC=-0.5066907, Accumulated Leap Seconds=32
  { 2453376.50, 64.691395 }, //06 January 2005, UT1-UTC=-0.5073947, Accumulated Leap Seconds=32
  { 2453377.50, 64.691878 }, //07 January 2005, UT1-UTC=-0.5078776, Accumulated Leap Seconds=32
  { 2453378.50, 64.692158 }, //08 January 2005, UT1-UTC=-0.5081577, Accumulated Leap Seconds=32
  { 2453379.50, 64.692328 }, //09 January 2005, UT1-UTC=-0.5083283, Accumulated Leap Seconds=32
  { 2453380.50, 64.692507 }, //10 January 2005, UT1-UTC=-0.5085068, Accumulated Leap Seconds=32
  { 2453381.50, 64.692784 }, //11 January 2005, UT1-UTC=-0.5087841, Accumulated Leap Seconds=32
  { 2453382.50, 64.693335 }, //12 January 2005, UT1-UTC=-0.5093349, Accumulated Leap Seconds=32
  { 2453383.50, 64.694181 }, //13 January 2005, UT1-UTC=-0.5101808, Accumulated Leap Seconds=32
  { 2453384.50, 64.695269 }, //14 January 2005, UT1-UTC=-0.5112689, Accumulated Leap Seconds=32
  { 2453385.50, 64.696456 }, //15 January 2005, UT1-UTC=-0.5124561, Accumulated Leap Seconds=32
  { 2453386.50, 64.697655 }, //16 January 2005, UT1-UTC=-0.5136546, Accumulated Leap Seconds=32
  { 2453387.50, 64.698761 }, //17 January 2005, UT1-UTC=-0.5147612, Accumulated Leap Seconds=32
  { 2453388.50, 64.699702 }, //18 January 2005, UT1-UTC=-0.5157019, Accumulated Leap Seconds=32
  { 2453389.50, 64.700415 }, //19 January 2005, UT1-UTC=-0.5164151, Accumulated Leap Seconds=32
  { 2453390.50, 64.700891 }, //20 January 2005, UT1-UTC=-0.5168915, Accumulated Leap Seconds=32
  { 2453391.50, 64.701218 }, //21 January 2005, UT1-UTC=-0.5172180, Accumulated Leap Seconds=32
  { 2453392.50, 64.701403 }, //22 January 2005, UT1-UTC=-0.5174033, Accumulated Leap Seconds=32
  { 2453393.50, 64.701485 }, //23 January 2005, UT1-UTC=-0.5174851, Accumulated Leap Seconds=32
  { 2453394.50, 64.701525 }, //24 January 2005, UT1-UTC=-0.5175253, Accumulated Leap Seconds=32
  { 2453395.50, 64.701582 }, //25 January 2005, UT1-UTC=-0.5175818, Accumulated Leap Seconds=32
  { 2453396.50, 64.701706 }, //26 January 2005, UT1-UTC=-0.5177056, Accumulated Leap Seconds=32
  { 2453397.50, 64.701973 }, //27 January 2005, UT1-UTC=-0.5179730, Accumulated Leap Seconds=32
  { 2453398.50, 64.702436 }, //28 January 2005, UT1-UTC=-0.5184360, Accumulated Leap Seconds=32
  { 2453399.50, 64.703053 }, //29 January 2005, UT1-UTC=-0.5190534, Accumulated Leap Seconds=32
  { 2453400.50, 64.703740 }, //30 January 2005, UT1-UTC=-0.5197397, Accumulated Leap Seconds=32
  { 2453401.50, 64.704463 }, //31 January 2005, UT1-UTC=-0.5204631, Accumulated Leap Seconds=32
  { 2453402.50, 64.705191 }, //01 February 2005, UT1-UTC=-0.5211905, Accumulated Leap Seconds=32
  { 2453403.50, 64.705874 }, //02 February 2005, UT1-UTC=-0.5218738, Accumulated Leap Seconds=32
  { 2453404.50, 64.706405 }, //03 February 2005, UT1-UTC=-0.5224047, Accumulated Leap Seconds=32
  { 2453405.50, 64.706754 }, //04 February 2005, UT1-UTC=-0.5227538, Accumulated Leap Seconds=32
  { 2453406.50, 64.706942 }, //05 February 2005, UT1-UTC=-0.5229425, Accumulated Leap Seconds=32
  { 2453407.50, 64.707176 }, //06 February 2005, UT1-UTC=-0.5231756, Accumulated Leap Seconds=32
  { 2453408.50, 64.707545 }, //07 February 2005, UT1-UTC=-0.5235446, Accumulated Leap Seconds=32
  { 2453409.50, 64.708069 }, //08 February 2005, UT1-UTC=-0.5240695, Accumulated Leap Seconds=32
  { 2453410.50, 64.708876 }, //09 February 2005, UT1-UTC=-0.5248763, Accumulated Leap Seconds=32
  { 2453411.50, 64.710032 }, //10 February 2005, UT1-UTC=-0.5260316, Accumulated Leap Seconds=32
  { 2453412.50, 64.711455 }, //11 February 2005, UT1-UTC=-0.5274547, Accumulated Leap Seconds=32
  { 2453413.50, 64.713023 }, //12 February 2005, UT1-UTC=-0.5290234, Accumulated Leap Seconds=32
  { 2453414.50, 64.714555 }, //13 February 2005, UT1-UTC=-0.5305546, Accumulated Leap Seconds=32
  { 2453415.50, 64.715899 }, //14 February 2005, UT1-UTC=-0.5318992, Accumulated Leap Seconds=32
  { 2453416.50, 64.716979 }, //15 February 2005, UT1-UTC=-0.5329786, Accumulated Leap Seconds=32
  { 2453417.50, 64.717759 }, //16 February 2005, UT1-UTC=-0.5337595, Accumulated Leap Seconds=32
  { 2453418.50, 64.718304 }, //17 February 2005, UT1-UTC=-0.5343039, Accumulated Leap Seconds=32
  { 2453419.50, 64.718695 }, //18 February 2005, UT1-UTC=-0.5346954, Accumulated Leap Seconds=32
  { 2453420.50, 64.719121 }, //19 February 2005, UT1-UTC=-0.5351206, Accumulated Leap Seconds=32
  { 2453421.50, 64.719640 }, //20 February 2005, UT1-UTC=-0.5356405, Accumulated Leap Seconds=32
  { 2453422.50, 64.720288 }, //21 February 2005, UT1-UTC=-0.5362883, Accumulated Leap Seconds=32
  { 2453423.50, 64.721105 }, //22 February 2005, UT1-UTC=-0.5371051, Accumulated Leap Seconds=32
  { 2453424.50, 64.722099 }, //23 February 2005, UT1-UTC=-0.5380987, Accumulated Leap Seconds=32
  { 2453425.50, 64.723295 }, //24 February 2005, UT1-UTC=-0.5392947, Accumulated Leap Seconds=32
  { 2453426.50, 64.724706 }, //25 February 2005, UT1-UTC=-0.5407065, Accumulated Leap Seconds=32
  { 2453427.50, 64.726305 }, //26 February 2005, UT1-UTC=-0.5423051, Accumulated Leap Seconds=32
  { 2453428.50, 64.728027 }, //27 February 2005, UT1-UTC=-0.5440274, Accumulated Leap Seconds=32
  { 2453429.50, 64.729726 }, //28 February 2005, UT1-UTC=-0.5457256, Accumulated Leap Seconds=32
  { 2453430.50, 64.731343 }, //01 March 2005, UT1-UTC=-0.5473433, Accumulated Leap Seconds=32
  { 2453431.50, 64.732826 }, //02 March 2005, UT1-UTC=-0.5488256, Accumulated Leap Seconds=32
  { 2453432.50, 64.734107 }, //03 March 2005, UT1-UTC=-0.5501067, Accumulated Leap Seconds=32
  { 2453433.50, 64.735212 }, //04 March 2005, UT1-UTC=-0.5512121, Accumulated Leap Seconds=32
  { 2453434.50, 64.736216 }, //05 March 2005, UT1-UTC=-0.5522160, Accumulated Leap Seconds=32
  { 2453435.50, 64.737158 }, //06 March 2005, UT1-UTC=-0.5531581, Accumulated Leap Seconds=32
  { 2453436.50, 64.738124 }, //07 March 2005, UT1-UTC=-0.5541239, Accumulated Leap Seconds=32
  { 2453437.50, 64.739230 }, //08 March 2005, UT1-UTC=-0.5552297, Accumulated Leap Seconds=32
  { 2453438.50, 64.740500 }, //09 March 2005, UT1-UTC=-0.5565004, Accumulated Leap Seconds=32
  { 2453439.50, 64.741891 }, //10 March 2005, UT1-UTC=-0.5578906, Accumulated Leap Seconds=32
  { 2453440.50, 64.743377 }, //11 March 2005, UT1-UTC=-0.5593770, Accumulated Leap Seconds=32
  { 2453441.50, 64.744905 }, //12 March 2005, UT1-UTC=-0.5609054, Accumulated Leap Seconds=32
  { 2453442.50, 64.746288 }, //13 March 2005, UT1-UTC=-0.5622876, Accumulated Leap Seconds=32
  { 2453443.50, 64.747396 }, //14 March 2005, UT1-UTC=-0.5633959, Accumulated Leap Seconds=32
  { 2453444.50, 64.748160 }, //15 March 2005, UT1-UTC=-0.5641599, Accumulated Leap Seconds=32
  { 2453445.50, 64.748583 }, //16 March 2005, UT1-UTC=-0.5645828, Accumulated Leap Seconds=32
  { 2453446.50, 64.748760 }, //17 March 2005, UT1-UTC=-0.5647604, Accumulated Leap Seconds=32
  { 2453447.50, 64.748807 }, //18 March 2005, UT1-UTC=-0.5648075, Accumulated Leap Seconds=32
  { 2453448.50, 64.748849 }, //19 March 2005, UT1-UTC=-0.5648490, Accumulated Leap Seconds=32
  { 2453449.50, 64.748937 }, //20 March 2005, UT1-UTC=-0.5649368, Accumulated Leap Seconds=32
  { 2453450.50, 64.749114 }, //21 March 2005, UT1-UTC=-0.5651139, Accumulated Leap Seconds=32
  { 2453451.50, 64.749442 }, //22 March 2005, UT1-UTC=-0.5654418, Accumulated Leap Seconds=32
  { 2453452.50, 64.750024 }, //23 March 2005, UT1-UTC=-0.5660242, Accumulated Leap Seconds=32
  { 2453453.50, 64.750873 }, //24 March 2005, UT1-UTC=-0.5668735, Accumulated Leap Seconds=32
  { 2453454.50, 64.751938 }, //25 March 2005, UT1-UTC=-0.5679385, Accumulated Leap Seconds=32
  { 2453455.50, 64.753128 }, //26 March 2005, UT1-UTC=-0.5691276, Accumulated Leap Seconds=32
  { 2453456.50, 64.754319 }, //27 March 2005, UT1-UTC=-0.5703189, Accumulated Leap Seconds=32
  { 2453457.50, 64.755410 }, //28 March 2005, UT1-UTC=-0.5714102, Accumulated Leap Seconds=32
  { 2453458.50, 64.756319 }, //29 March 2005, UT1-UTC=-0.5723188, Accumulated Leap Seconds=32
  { 2453459.50, 64.756978 }, //30 March 2005, UT1-UTC=-0.5729782, Accumulated Leap Seconds=32
  { 2453460.50, 64.757354 }, //31 March 2005, UT1-UTC=-0.5733537, Accumulated Leap Seconds=32
  { 2453461.50, 64.757531 }, //01 April 2005, UT1-UTC=-0.5735312, Accumulated Leap Seconds=32
  { 2453462.50, 64.757623 }, //02 April 2005, UT1-UTC=-0.5736228, Accumulated Leap Seconds=32
  { 2453463.50, 64.757847 }, //03 April 2005, UT1-UTC=-0.5738473, Accumulated Leap Seconds=32
  { 2453464.50, 64.758313 }, //04 April 2005, UT1-UTC=-0.5743133, Accumulated Leap Seconds=32
  { 2453465.50, 64.759076 }, //05 April 2005, UT1-UTC=-0.5750763, Accumulated Leap Seconds=32
  { 2453466.50, 64.760062 }, //06 April 2005, UT1-UTC=-0.5760620, Accumulated Leap Seconds=32
  { 2453467.50, 64.761231 }, //07 April 2005, UT1-UTC=-0.5772309, Accumulated Leap Seconds=32
  { 2453468.50, 64.762517 }, //08 April 2005, UT1-UTC=-0.5785169, Accumulated Leap Seconds=32
  { 2453469.50, 64.763796 }, //09 April 2005, UT1-UTC=-0.5797958, Accumulated Leap Seconds=32
  { 2453470.50, 64.764854 }, //10 April 2005, UT1-UTC=-0.5808544, Accumulated Leap Seconds=32
  { 2453471.50, 64.765585 }, //11 April 2005, UT1-UTC=-0.5815846, Accumulated Leap Seconds=32
  { 2453472.50, 64.766065 }, //12 April 2005, UT1-UTC=-0.5820651, Accumulated Leap Seconds=32
  { 2453473.50, 64.766345 }, //13 April 2005, UT1-UTC=-0.5823454, Accumulated Leap Seconds=32
  { 2453474.50, 64.766586 }, //14 April 2005, UT1-UTC=-0.5825858, Accumulated Leap Seconds=32
  { 2453475.50, 64.766843 }, //15 April 2005, UT1-UTC=-0.5828435, Accumulated Leap Seconds=32
  { 2453476.50, 64.767175 }, //16 April 2005, UT1-UTC=-0.5831746, Accumulated Leap Seconds=32
  { 2453477.50, 64.767653 }, //17 April 2005, UT1-UTC=-0.5836531, Accumulated Leap Seconds=32
  { 2453478.50, 64.768300 }, //18 April 2005, UT1-UTC=-0.5843004, Accumulated Leap Seconds=32
  { 2453479.50, 64.769145 }, //19 April 2005, UT1-UTC=-0.5851451, Accumulated Leap Seconds=32
  { 2453480.50, 64.770181 }, //20 April 2005, UT1-UTC=-0.5861815, Accumulated Leap Seconds=32
  { 2453481.50, 64.771395 }, //21 April 2005, UT1-UTC=-0.5873950, Accumulated Leap Seconds=32
  { 2453482.50, 64.772758 }, //22 April 2005, UT1-UTC=-0.5887581, Accumulated Leap Seconds=32
  { 2453483.50, 64.774175 }, //23 April 2005, UT1-UTC=-0.5901745, Accumulated Leap Seconds=32
  { 2453484.50, 64.775602 }, //24 April 2005, UT1-UTC=-0.5916018, Accumulated Leap Seconds=32
  { 2453485.50, 64.776923 }, //25 April 2005, UT1-UTC=-0.5929231, Accumulated Leap Seconds=32
  { 2453486.50, 64.778046 }, //26 April 2005, UT1-UTC=-0.5940459, Accumulated Leap Seconds=32
  { 2453487.50, 64.778902 }, //27 April 2005, UT1-UTC=-0.5949024, Accumulated Leap Seconds=32
  { 2453488.50, 64.779502 }, //28 April 2005, UT1-UTC=-0.5955016, Accumulated Leap Seconds=32
  { 2453489.50, 64.779957 }, //29 April 2005, UT1-UTC=-0.5959571, Accumulated Leap Seconds=32
  { 2453490.50, 64.780468 }, //30 April 2005, UT1-UTC=-0.5964681, Accumulated Leap Seconds=32
  { 2453491.50, 64.781114 }, //01 May 2005, UT1-UTC=-0.5971143, Accumulated Leap Seconds=32
  { 2453492.50, 64.781990 }, //02 May 2005, UT1-UTC=-0.5979901, Accumulated Leap Seconds=32
  { 2453493.50, 64.783113 }, //03 May 2005, UT1-UTC=-0.5991133, Accumulated Leap Seconds=32
  { 2453494.50, 64.784440 }, //04 May 2005, UT1-UTC=-0.6004401, Accumulated Leap Seconds=32
  { 2453495.50, 64.785849 }, //05 May 2005, UT1-UTC=-0.6018485, Accumulated Leap Seconds=32
  { 2453496.50, 64.787197 }, //06 May 2005, UT1-UTC=-0.6031968, Accumulated Leap Seconds=32
  { 2453497.50, 64.788406 }, //07 May 2005, UT1-UTC=-0.6044061, Accumulated Leap Seconds=32
  { 2453498.50, 64.789440 }, //08 May 2005, UT1-UTC=-0.6054396, Accumulated Leap Seconds=32
  { 2453499.50, 64.790232 }, //09 May 2005, UT1-UTC=-0.6062322, Accumulated Leap Seconds=32
  { 2453500.50, 64.790792 }, //10 May 2005, UT1-UTC=-0.6067919, Accumulated Leap Seconds=32
  { 2453501.50, 64.791171 }, //11 May 2005, UT1-UTC=-0.6071706, Accumulated Leap Seconds=32
  { 2453502.50, 64.791429 }, //12 May 2005, UT1-UTC=-0.6074286, Accumulated Leap Seconds=32
  { 2453503.50, 64.791676 }, //13 May 2005, UT1-UTC=-0.6076761, Accumulated Leap Seconds=32
  { 2453504.50, 64.791927 }, //14 May 2005, UT1-UTC=-0.6079271, Accumulated Leap Seconds=32
  { 2453505.50, 64.792308 }, //15 May 2005, UT1-UTC=-0.6083077, Accumulated Leap Seconds=32
  { 2453506.50, 64.792865 }, //16 May 2005, UT1-UTC=-0.6088646, Accumulated Leap Seconds=32
  { 2453507.50, 64.793586 }, //17 May 2005, UT1-UTC=-0.6095860, Accumulated Leap Seconds=32
  { 2453508.50, 64.794413 }, //18 May 2005, UT1-UTC=-0.6104126, Accumulated Leap Seconds=32
  { 2453509.50, 64.795367 }, //19 May 2005, UT1-UTC=-0.6113669, Accumulated Leap Seconds=32
  { 2453510.50, 64.796394 }, //20 May 2005, UT1-UTC=-0.6123944, Accumulated Leap Seconds=32
  { 2453511.50, 64.797354 }, //21 May 2005, UT1-UTC=-0.6133540, Accumulated Leap Seconds=32
  { 2453512.50, 64.798219 }, //22 May 2005, UT1-UTC=-0.6142188, Accumulated Leap Seconds=32
  { 2453513.50, 64.798838 }, //23 May 2005, UT1-UTC=-0.6148382, Accumulated Leap Seconds=32
  { 2453514.50, 64.799132 }, //24 May 2005, UT1-UTC=-0.6151316, Accumulated Leap Seconds=32
  { 2453515.50, 64.799113 }, //25 May 2005, UT1-UTC=-0.6151126, Accumulated Leap Seconds=32
  { 2453516.50, 64.798859 }, //26 May 2005, UT1-UTC=-0.6148593, Accumulated Leap Seconds=32
  { 2453517.50, 64.798560 }, //27 May 2005, UT1-UTC=-0.6145596, Accumulated Leap Seconds=32
  { 2453518.50, 64.798407 }, //28 May 2005, UT1-UTC=-0.6144072, Accumulated Leap Seconds=32
  { 2453519.50, 64.798490 }, //29 May 2005, UT1-UTC=-0.6144904, Accumulated Leap Seconds=32
  { 2453520.50, 64.798824 }, //30 May 2005, UT1-UTC=-0.6148240, Accumulated Leap Seconds=32
  { 2453521.50, 64.799383 }, //31 May 2005, UT1-UTC=-0.6153832, Accumulated Leap Seconds=32
  { 2453522.50, 64.800093 }, //01 June 2005, UT1-UTC=-0.6160929, Accumulated Leap Seconds=32
  { 2453523.50, 64.800820 }, //02 June 2005, UT1-UTC=-0.6168203, Accumulated Leap Seconds=32
  { 2453524.50, 64.801375 }, //03 June 2005, UT1-UTC=-0.6173750, Accumulated Leap Seconds=32
  { 2453525.50, 64.801671 }, //04 June 2005, UT1-UTC=-0.6176712, Accumulated Leap Seconds=32
  { 2453526.50, 64.801769 }, //05 June 2005, UT1-UTC=-0.6177695, Accumulated Leap Seconds=32
  { 2453527.50, 64.801632 }, //06 June 2005, UT1-UTC=-0.6176318, Accumulated Leap Seconds=32
  { 2453528.50, 64.801303 }, //07 June 2005, UT1-UTC=-0.6173033, Accumulated Leap Seconds=32
  { 2453529.50, 64.800834 }, //08 June 2005, UT1-UTC=-0.6168337, Accumulated Leap Seconds=32
  { 2453530.50, 64.800260 }, //09 June 2005, UT1-UTC=-0.6162601, Accumulated Leap Seconds=32
  { 2453531.50, 64.799740 }, //10 June 2005, UT1-UTC=-0.6157397, Accumulated Leap Seconds=32
  { 2453532.50, 64.799344 }, //11 June 2005, UT1-UTC=-0.6153436, Accumulated Leap Seconds=32
  { 2453533.50, 64.799058 }, //12 June 2005, UT1-UTC=-0.6150583, Accumulated Leap Seconds=32
  { 2453534.50, 64.798900 }, //13 June 2005, UT1-UTC=-0.6148997, Accumulated Leap Seconds=32
  { 2453535.50, 64.798906 }, //14 June 2005, UT1-UTC=-0.6149058, Accumulated Leap Seconds=32
  { 2453536.50, 64.799070 }, //15 June 2005, UT1-UTC=-0.6150703, Accumulated Leap Seconds=32
  { 2453537.50, 64.799378 }, //16 June 2005, UT1-UTC=-0.6153776, Accumulated Leap Seconds=32
  { 2453538.50, 64.799773 }, //17 June 2005, UT1-UTC=-0.6157731, Accumulated Leap Seconds=32
  { 2453539.50, 64.800124 }, //18 June 2005, UT1-UTC=-0.6161240, Accumulated Leap Seconds=32
  { 2453540.50, 64.800300 }, //19 June 2005, UT1-UTC=-0.6162996, Accumulated Leap Seconds=32
  { 2453541.50, 64.800233 }, //20 June 2005, UT1-UTC=-0.6162332, Accumulated Leap Seconds=32
  { 2453542.50, 64.799941 }, //21 June 2005, UT1-UTC=-0.6159410, Accumulated Leap Seconds=32
  { 2453543.50, 64.799464 }, //22 June 2005, UT1-UTC=-0.6154641, Accumulated Leap Seconds=32
  { 2453544.50, 64.798956 }, //23 June 2005, UT1-UTC=-0.6149561, Accumulated Leap Seconds=32
  { 2453545.50, 64.798575 }, //24 June 2005, UT1-UTC=-0.6145745, Accumulated Leap Seconds=32
  { 2453546.50, 64.798415 }, //25 June 2005, UT1-UTC=-0.6144151, Accumulated Leap Seconds=32
  { 2453547.50, 64.798536 }, //26 June 2005, UT1-UTC=-0.6145364, Accumulated Leap Seconds=32
  { 2453548.50, 64.798801 }, //27 June 2005, UT1-UTC=-0.6148010, Accumulated Leap Seconds=32
  { 2453549.50, 64.799127 }, //28 June 2005, UT1-UTC=-0.6151267, Accumulated Leap Seconds=32
  { 2453550.50, 64.799392 }, //29 June 2005, UT1-UTC=-0.6153920, Accumulated Leap Seconds=32
  { 2453551.50, 64.799544 }, //30 June 2005, UT1-UTC=-0.6155443, Accumulated Leap Seconds=32
  { 2453552.50, 64.799456 }, //01 July 2005, UT1-UTC=-0.6154561, Accumulated Leap Seconds=32
  { 2453553.50, 64.799103 }, //02 July 2005, UT1-UTC=-0.6151027, Accumulated Leap Seconds=32
  { 2453554.50, 64.798447 }, //03 July 2005, UT1-UTC=-0.6144473, Accumulated Leap Seconds=32
  { 2453555.50, 64.797536 }, //04 July 2005, UT1-UTC=-0.6135364, Accumulated Leap Seconds=32
  { 2453556.50, 64.796509 }, //05 July 2005, UT1-UTC=-0.6125092, Accumulated Leap Seconds=32
  { 2453557.50, 64.795458 }, //06 July 2005, UT1-UTC=-0.6114583, Accumulated Leap Seconds=32
  { 2453558.50, 64.794465 }, //07 July 2005, UT1-UTC=-0.6104654, Accumulated Leap Seconds=32
  { 2453559.50, 64.793613 }, //08 July 2005, UT1-UTC=-0.6096134, Accumulated Leap Seconds=32
  { 2453560.50, 64.792896 }, //09 July 2005, UT1-UTC=-0.6088955, Accumulated Leap Seconds=32
  { 2453561.50, 64.792333 }, //10 July 2005, UT1-UTC=-0.6083329, Accumulated Leap Seconds=32
  { 2453562.50, 64.791994 }, //11 July 2005, UT1-UTC=-0.6079942, Accumulated Leap Seconds=32
  { 2453563.50, 64.791833 }, //12 July 2005, UT1-UTC=-0.6078326, Accumulated Leap Seconds=32
  { 2453564.50, 64.791765 }, //13 July 2005, UT1-UTC=-0.6077653, Accumulated Leap Seconds=32
  { 2453565.50, 64.791752 }, //14 July 2005, UT1-UTC=-0.6077515, Accumulated Leap Seconds=32
  { 2453566.50, 64.791678 }, //15 July 2005, UT1-UTC=-0.6076784, Accumulated Leap Seconds=32
  { 2453567.50, 64.791473 }, //16 July 2005, UT1-UTC=-0.6074731, Accumulated Leap Seconds=32
  { 2453568.50, 64.791107 }, //17 July 2005, UT1-UTC=-0.6071073, Accumulated Leap Seconds=32
  { 2453569.50, 64.790492 }, //18 July 2005, UT1-UTC=-0.6064925, Accumulated Leap Seconds=32
  { 2453570.50, 64.789696 }, //19 July 2005, UT1-UTC=-0.6056959, Accumulated Leap Seconds=32
  { 2453571.50, 64.788837 }, //20 July 2005, UT1-UTC=-0.6048368, Accumulated Leap Seconds=32
  { 2453572.50, 64.788040 }, //21 July 2005, UT1-UTC=-0.6040402, Accumulated Leap Seconds=32
  { 2453573.50, 64.787401 }, //22 July 2005, UT1-UTC=-0.6034009, Accumulated Leap Seconds=32
  { 2453574.50, 64.787150 }, //23 July 2005, UT1-UTC=-0.6031503, Accumulated Leap Seconds=32
  { 2453575.50, 64.787252 }, //24 July 2005, UT1-UTC=-0.6032519, Accumulated Leap Seconds=32
  { 2453576.50, 64.787647 }, //25 July 2005, UT1-UTC=-0.6036473, Accumulated Leap Seconds=32
  { 2453577.50, 64.788140 }, //26 July 2005, UT1-UTC=-0.6041398, Accumulated Leap Seconds=32
  { 2453578.50, 64.788576 }, //27 July 2005, UT1-UTC=-0.6045763, Accumulated Leap Seconds=32
  { 2453579.50, 64.788837 }, //28 July 2005, UT1-UTC=-0.6048366, Accumulated Leap Seconds=32
  { 2453580.50, 64.788849 }, //29 July 2005, UT1-UTC=-0.6048490, Accumulated Leap Seconds=32
  { 2453581.50, 64.788656 }, //30 July 2005, UT1-UTC=-0.6046557, Accumulated Leap Seconds=32
  { 2453582.50, 64.788209 }, //31 July 2005, UT1-UTC=-0.6042090, Accumulated Leap Seconds=32
  { 2453583.50, 64.787642 }, //01 August 2005, UT1-UTC=-0.6036424, Accumulated Leap Seconds=32
  { 2453584.50, 64.787022 }, //02 August 2005, UT1-UTC=-0.6030216, Accumulated Leap Seconds=32
  { 2453585.50, 64.786371 }, //03 August 2005, UT1-UTC=-0.6023707, Accumulated Leap Seconds=32
  { 2453586.50, 64.785762 }, //04 August 2005, UT1-UTC=-0.6017619, Accumulated Leap Seconds=32
  { 2453587.50, 64.785252 }, //05 August 2005, UT1-UTC=-0.6012517, Accumulated Leap Seconds=32
  { 2453588.50, 64.784955 }, //06 August 2005, UT1-UTC=-0.6009546, Accumulated Leap Seconds=32
  { 2453589.50, 64.784889 }, //07 August 2005, UT1-UTC=-0.6008886, Accumulated Leap Seconds=32
  { 2453590.50, 64.785060 }, //08 August 2005, UT1-UTC=-0.6010604, Accumulated Leap Seconds=32
  { 2453591.50, 64.785390 }, //09 August 2005, UT1-UTC=-0.6013898, Accumulated Leap Seconds=32
  { 2453592.50, 64.785837 }, //10 August 2005, UT1-UTC=-0.6018369, Accumulated Leap Seconds=32
  { 2453593.50, 64.786254 }, //11 August 2005, UT1-UTC=-0.6022535, Accumulated Leap Seconds=32
  { 2453594.50, 64.786552 }, //12 August 2005, UT1-UTC=-0.6025523, Accumulated Leap Seconds=32
  { 2453595.50, 64.786666 }, //13 August 2005, UT1-UTC=-0.6026659, Accumulated Leap Seconds=32
  { 2453596.50, 64.786577 }, //14 August 2005, UT1-UTC=-0.6025770, Accumulated Leap Seconds=32
  { 2453597.50, 64.786222 }, //15 August 2005, UT1-UTC=-0.6022223, Accumulated Leap Seconds=32
  { 2453598.50, 64.785756 }, //16 August 2005, UT1-UTC=-0.6017561, Accumulated Leap Seconds=32
  { 2453599.50, 64.785347 }, //17 August 2005, UT1-UTC=-0.6013473, Accumulated Leap Seconds=32
  { 2453600.50, 64.784991 }, //18 August 2005, UT1-UTC=-0.6009906, Accumulated Leap Seconds=32
  { 2453601.50, 64.784890 }, //19 August 2005, UT1-UTC=-0.6008900, Accumulated Leap Seconds=32
  { 2453602.50, 64.785130 }, //20 August 2005, UT1-UTC=-0.6011302, Accumulated Leap Seconds=32
  { 2453603.50, 64.785703 }, //21 August 2005, UT1-UTC=-0.6017028, Accumulated Leap Seconds=32
  { 2453604.50, 64.786445 }, //22 August 2005, UT1-UTC=-0.6024450, Accumulated Leap Seconds=32
  { 2453605.50, 64.787083 }, //23 August 2005, UT1-UTC=-0.6030829, Accumulated Leap Seconds=32
  { 2453606.50, 64.787499 }, //24 August 2005, UT1-UTC=-0.6034994, Accumulated Leap Seconds=32
  { 2453607.50, 64.787652 }, //25 August 2005, UT1-UTC=-0.6036519, Accumulated Leap Seconds=32
  { 2453608.50, 64.787529 }, //26 August 2005, UT1-UTC=-0.6035295, Accumulated Leap Seconds=32
  { 2453609.50, 64.787067 }, //27 August 2005, UT1-UTC=-0.6030669, Accumulated Leap Seconds=32
  { 2453610.50, 64.786405 }, //28 August 2005, UT1-UTC=-0.6024048, Accumulated Leap Seconds=32
  { 2453611.50, 64.785562 }, //29 August 2005, UT1-UTC=-0.6015615, Accumulated Leap Seconds=32
  { 2453612.50, 64.784649 }, //30 August 2005, UT1-UTC=-0.6006488, Accumulated Leap Seconds=32
  { 2453613.50, 64.783794 }, //31 August 2005, UT1-UTC=-0.5997937, Accumulated Leap Seconds=32
  { 2453614.50, 64.783095 }, //01 September 2005, UT1-UTC=-0.5990950, Accumulated Leap Seconds=32
  { 2453615.50, 64.782644 }, //02 September 2005, UT1-UTC=-0.5986444, Accumulated Leap Seconds=32
  { 2453616.50, 64.782494 }, //03 September 2005, UT1-UTC=-0.5984938, Accumulated Leap Seconds=32
  { 2453617.50, 64.782609 }, //04 September 2005, UT1-UTC=-0.5986091, Accumulated Leap Seconds=32
  { 2453618.50, 64.782849 }, //05 September 2005, UT1-UTC=-0.5988493, Accumulated Leap Seconds=32
  { 2453619.50, 64.783217 }, //06 September 2005, UT1-UTC=-0.5992174, Accumulated Leap Seconds=32
  { 2453620.50, 64.783661 }, //07 September 2005, UT1-UTC=-0.5996615, Accumulated Leap Seconds=32
  { 2453621.50, 64.784131 }, //08 September 2005, UT1-UTC=-0.6001311, Accumulated Leap Seconds=32
  { 2453622.50, 64.784514 }, //09 September 2005, UT1-UTC=-0.6005137, Accumulated Leap Seconds=32
  { 2453623.50, 64.784724 }, //10 September 2005, UT1-UTC=-0.6007243, Accumulated Leap Seconds=32
  { 2453624.50, 64.784739 }, //11 September 2005, UT1-UTC=-0.6007394, Accumulated Leap Seconds=32
  { 2453625.50, 64.784569 }, //12 September 2005, UT1-UTC=-0.6005686, Accumulated Leap Seconds=32
  { 2453626.50, 64.784392 }, //13 September 2005, UT1-UTC=-0.6003922, Accumulated Leap Seconds=32
  { 2453627.50, 64.784374 }, //14 September 2005, UT1-UTC=-0.6003735, Accumulated Leap Seconds=32
  { 2453628.50, 64.784592 }, //15 September 2005, UT1-UTC=-0.6005921, Accumulated Leap Seconds=32
  { 2453629.50, 64.785141 }, //16 September 2005, UT1-UTC=-0.6011414, Accumulated Leap Seconds=32
  { 2453630.50, 64.786067 }, //17 September 2005, UT1-UTC=-0.6020670, Accumulated Leap Seconds=32
  { 2453631.50, 64.787275 }, //18 September 2005, UT1-UTC=-0.6032750, Accumulated Leap Seconds=32
  { 2453632.50, 64.788569 }, //19 September 2005, UT1-UTC=-0.6045695, Accumulated Leap Seconds=32
  { 2453633.50, 64.789783 }, //20 September 2005, UT1-UTC=-0.6057829, Accumulated Leap Seconds=32
  { 2453634.50, 64.790745 }, //21 September 2005, UT1-UTC=-0.6067447, Accumulated Leap Seconds=32
  { 2453635.50, 64.791391 }, //22 September 2005, UT1-UTC=-0.6073910, Accumulated Leap Seconds=32
  { 2453636.50, 64.791720 }, //23 September 2005, UT1-UTC=-0.6077203, Accumulated Leap Seconds=32
  { 2453637.50, 64.791771 }, //24 September 2005, UT1-UTC=-0.6077713, Accumulated Leap Seconds=32
  { 2453638.50, 64.791655 }, //25 September 2005, UT1-UTC=-0.6076554, Accumulated Leap Seconds=32
  { 2453639.50, 64.791462 }, //26 September 2005, UT1-UTC=-0.6074621, Accumulated Leap Seconds=32
  { 2453640.50, 64.791313 }, //27 September 2005, UT1-UTC=-0.6073125, Accumulated Leap Seconds=32
  { 2453641.50, 64.791221 }, //28 September 2005, UT1-UTC=-0.6072208, Accumulated Leap Seconds=32
  { 2453642.50, 64.791303 }, //29 September 2005, UT1-UTC=-0.6073035, Accumulated Leap Seconds=32
  { 2453643.50, 64.791604 }, //30 September 2005, UT1-UTC=-0.6076038, Accumulated Leap Seconds=32
  { 2453644.50, 64.792060 }, //01 October 2005, UT1-UTC=-0.6080604, Accumulated Leap Seconds=32
  { 2453645.50, 64.792723 }, //02 October 2005, UT1-UTC=-0.6087229, Accumulated Leap Seconds=32
  { 2453646.50, 64.793474 }, //03 October 2005, UT1-UTC=-0.6094741, Accumulated Leap Seconds=32
  { 2453647.50, 64.794217 }, //04 October 2005, UT1-UTC=-0.6102174, Accumulated Leap Seconds=32
  { 2453648.50, 64.794865 }, //05 October 2005, UT1-UTC=-0.6108649, Accumulated Leap Seconds=32
  { 2453649.50, 64.795353 }, //06 October 2005, UT1-UTC=-0.6113533, Accumulated Leap Seconds=32
  { 2453650.50, 64.795696 }, //07 October 2005, UT1-UTC=-0.6116964, Accumulated Leap Seconds=32
  { 2453651.50, 64.795937 }, //08 October 2005, UT1-UTC=-0.6119373, Accumulated Leap Seconds=32
  { 2453652.50, 64.796022 }, //09 October 2005, UT1-UTC=-0.6120218, Accumulated Leap Seconds=32
  { 2453653.50, 64.796021 }, //10 October 2005, UT1-UTC=-0.6120214, Accumulated Leap Seconds=32
  { 2453654.50, 64.796055 }, //11 October 2005, UT1-UTC=-0.6120554, Accumulated Leap Seconds=32
  { 2453655.50, 64.796224 }, //12 October 2005, UT1-UTC=-0.6122245, Accumulated Leap Seconds=32
  { 2453656.50, 64.796641 }, //13 October 2005, UT1-UTC=-0.6126408, Accumulated Leap Seconds=32
  { 2453657.50, 64.797355 }, //14 October 2005, UT1-UTC=-0.6133548, Accumulated Leap Seconds=32
  { 2453658.50, 64.798384 }, //15 October 2005, UT1-UTC=-0.6143842, Accumulated Leap Seconds=32
  { 2453659.50, 64.799552 }, //16 October 2005, UT1-UTC=-0.6155524, Accumulated Leap Seconds=32
  { 2453660.50, 64.800706 }, //17 October 2005, UT1-UTC=-0.6167060, Accumulated Leap Seconds=32
  { 2453661.50, 64.801741 }, //18 October 2005, UT1-UTC=-0.6177406, Accumulated Leap Seconds=32
  { 2453662.50, 64.802575 }, //19 October 2005, UT1-UTC=-0.6185748, Accumulated Leap Seconds=32
  { 2453663.50, 64.803158 }, //20 October 2005, UT1-UTC=-0.6191583, Accumulated Leap Seconds=32
  { 2453664.50, 64.803512 }, //21 October 2005, UT1-UTC=-0.6195121, Accumulated Leap Seconds=32
  { 2453665.50, 64.803722 }, //22 October 2005, UT1-UTC=-0.6197224, Accumulated Leap Seconds=32
  { 2453666.50, 64.803867 }, //23 October 2005, UT1-UTC=-0.6198675, Accumulated Leap Seconds=32
  { 2453667.50, 64.804004 }, //24 October 2005, UT1-UTC=-0.6200037, Accumulated Leap Seconds=32
  { 2453668.50, 64.804236 }, //25 October 2005, UT1-UTC=-0.6202363, Accumulated Leap Seconds=32
  { 2453669.50, 64.804540 }, //26 October 2005, UT1-UTC=-0.6205397, Accumulated Leap Seconds=32
  { 2453670.50, 64.804978 }, //27 October 2005, UT1-UTC=-0.6209777, Accumulated Leap Seconds=32
  { 2453671.50, 64.805643 }, //28 October 2005, UT1-UTC=-0.6216430, Accumulated Leap Seconds=32
  { 2453672.50, 64.806469 }, //29 October 2005, UT1-UTC=-0.6224689, Accumulated Leap Seconds=32
  { 2453673.50, 64.807486 }, //30 October 2005, UT1-UTC=-0.6234856, Accumulated Leap Seconds=32
  { 2453674.50, 64.808580 }, //31 October 2005, UT1-UTC=-0.6245799, Accumulated Leap Seconds=32
  { 2453675.50, 64.809642 }, //01 November 2005, UT1-UTC=-0.6256421, Accumulated Leap Seconds=32
  { 2453676.50, 64.810568 }, //02 November 2005, UT1-UTC=-0.6265680, Accumulated Leap Seconds=32
  { 2453677.50, 64.811302 }, //03 November 2005, UT1-UTC=-0.6273016, Accumulated Leap Seconds=32
  { 2453678.50, 64.811833 }, //04 November 2005, UT1-UTC=-0.6278328, Accumulated Leap Seconds=32
  { 2453679.50, 64.812082 }, //05 November 2005, UT1-UTC=-0.6280822, Accumulated Leap Seconds=32
  { 2453680.50, 64.812252 }, //06 November 2005, UT1-UTC=-0.6282523, Accumulated Leap Seconds=32
  { 2453681.50, 64.812408 }, //07 November 2005, UT1-UTC=-0.6284076, Accumulated Leap Seconds=32
  { 2453682.50, 64.812730 }, //08 November 2005, UT1-UTC=-0.6287301, Accumulated Leap Seconds=32
  { 2453683.50, 64.813315 }, //09 November 2005, UT1-UTC=-0.6293147, Accumulated Leap Seconds=32
  { 2453684.50, 64.814293 }, //10 November 2005, UT1-UTC=-0.6302931, Accumulated Leap Seconds=32
  { 2453685.50, 64.815549 }, //11 November 2005, UT1-UTC=-0.6315489, Accumulated Leap Seconds=32
  { 2453686.50, 64.816994 }, //12 November 2005, UT1-UTC=-0.6329940, Accumulated Leap Seconds=32
  { 2453687.50, 64.818452 }, //13 November 2005, UT1-UTC=-0.6344525, Accumulated Leap Seconds=32
  { 2453688.50, 64.819795 }, //14 November 2005, UT1-UTC=-0.6357953, Accumulated Leap Seconds=32
  { 2453689.50, 64.820902 }, //15 November 2005, UT1-UTC=-0.6369022, Accumulated Leap Seconds=32
  { 2453690.50, 64.821707 }, //16 November 2005, UT1-UTC=-0.6377074, Accumulated Leap Seconds=32
  { 2453691.50, 64.822262 }, //17 November 2005, UT1-UTC=-0.6382619, Accumulated Leap Seconds=32
  { 2453692.50, 64.822601 }, //18 November 2005, UT1-UTC=-0.6386010, Accumulated Leap Seconds=32
  { 2453693.50, 64.822765 }, //19 November 2005, UT1-UTC=-0.6387653, Accumulated Leap Seconds=32
  { 2453694.50, 64.823000 }, //20 November 2005, UT1-UTC=-0.6390005, Accumulated Leap Seconds=32
  { 2453695.50, 64.823332 }, //21 November 2005, UT1-UTC=-0.6393319, Accumulated Leap Seconds=32
  { 2453696.50, 64.823805 }, //22 November 2005, UT1-UTC=-0.6398046, Accumulated Leap Seconds=32
  { 2453697.50, 64.824405 }, //23 November 2005, UT1-UTC=-0.6404055, Accumulated Leap Seconds=32
  { 2453698.50, 64.825146 }, //24 November 2005, UT1-UTC=-0.6411462, Accumulated Leap Seconds=32
  { 2453699.50, 64.826022 }, //25 November 2005, UT1-UTC=-0.6420224, Accumulated Leap Seconds=32
  { 2453700.50, 64.827012 }, //26 November 2005, UT1-UTC=-0.6430118, Accumulated Leap Seconds=32
  { 2453701.50, 64.828062 }, //27 November 2005, UT1-UTC=-0.6440616, Accumulated Leap Seconds=32
  { 2453702.50, 64.829074 }, //28 November 2005, UT1-UTC=-0.6450739, Accumulated Leap Seconds=32
  { 2453703.50, 64.829954 }, //29 November 2005, UT1-UTC=-0.6459536, Accumulated Leap Seconds=32
  { 2453704.50, 64.830628 }, //30 November 2005, UT1-UTC=-0.6466279, Accumulated Leap Seconds=32
  { 2453705.50, 64.831089 }, //01 December 2005, UT1-UTC=-0.6470888, Accumulated Leap Seconds=32
  { 2453706.50, 64.831353 }, //02 December 2005, UT1-UTC=-0.6473535, Accumulated Leap Seconds=32
  { 2453707.50, 64.831493 }, //03 December 2005, UT1-UTC=-0.6474934, Accumulated Leap Seconds=32
  { 2453708.50, 64.831623 }, //04 December 2005, UT1-UTC=-0.6476229, Accumulated Leap Seconds=32
  { 2453709.50, 64.831739 }, //05 December 2005, UT1-UTC=-0.6477390, Accumulated Leap Seconds=32
  { 2453710.50, 64.832003 }, //06 December 2005, UT1-UTC=-0.6480033, Accumulated Leap Seconds=32
  { 2453711.50, 64.832565 }, //07 December 2005, UT1-UTC=-0.6485652, Accumulated Leap Seconds=32
  { 2453712.50, 64.833323 }, //08 December 2005, UT1-UTC=-0.6493234, Accumulated Leap Seconds=32
  { 2453713.50, 64.834265 }, //09 December 2005, UT1-UTC=-0.6502652, Accumulated Leap Seconds=32
  { 2453714.50, 64.835242 }, //10 December 2005, UT1-UTC=-0.6512420, Accumulated Leap Seconds=32
  { 2453715.50, 64.836128 }, //11 December 2005, UT1-UTC=-0.6521276, Accumulated Leap Seconds=32
  { 2453716.50, 64.836883 }, //12 December 2005, UT1-UTC=-0.6528826, Accumulated Leap Seconds=32
  { 2453717.50, 64.837456 }, //13 December 2005, UT1-UTC=-0.6534555, Accumulated Leap Seconds=32
  { 2453718.50, 64.837806 }, //14 December 2005, UT1-UTC=-0.6538055, Accumulated Leap Seconds=32
  { 2453719.50, 64.837936 }, //15 December 2005, UT1-UTC=-0.6539359, Accumulated Leap Seconds=32
  { 2453720.50, 64.838000 }, //16 December 2005, UT1-UTC=-0.6540002, Accumulated Leap Seconds=32
  { 2453721.50, 64.838038 }, //17 December 2005, UT1-UTC=-0.6540381, Accumulated Leap Seconds=32
  { 2453722.50, 64.838123 }, //18 December 2005, UT1-UTC=-0.6541231, Accumulated Leap Seconds=32
  { 2453723.50, 64.838315 }, //19 December 2005, UT1-UTC=-0.6543147, Accumulated Leap Seconds=32
  { 2453724.50, 64.838658 }, //20 December 2005, UT1-UTC=-0.6546584, Accumulated Leap Seconds=32
  { 2453725.50, 64.839228 }, //21 December 2005, UT1-UTC=-0.6552283, Accumulated Leap Seconds=32
  { 2453726.50, 64.839963 }, //22 December 2005, UT1-UTC=-0.6559628, Accumulated Leap Seconds=32
  { 2453727.50, 64.840824 }, //23 December 2005, UT1-UTC=-0.6568244, Accumulated Leap Seconds=32
  { 2453728.50, 64.841759 }, //24 December 2005, UT1-UTC=-0.6577590, Accumulated Leap Seconds=32
  { 2453729.50, 64.842683 }, //25 December 2005, UT1-UTC=-0.6586830, Accumulated Leap Seconds=32
  { 2453730.50, 64.843547 }, //26 December 2005, UT1-UTC=-0.6595466, Accumulated Leap Seconds=32
  { 2453731.50, 64.844277 }, //27 December 2005, UT1-UTC=-0.6602768, Accumulated Leap Seconds=32
  { 2453732.50, 64.844783 }, //28 December 2005, UT1-UTC=-0.6607828, Accumulated Leap Seconds=32
  { 2453733.50, 64.845094 }, //29 December 2005, UT1-UTC=-0.6610944, Accumulated Leap Seconds=32
  { 2453734.50, 64.845140 }, //30 December 2005, UT1-UTC=-0.6611399, Accumulated Leap Seconds=32
  { 2453735.50, 64.845124 }, //31 December 2005, UT1-UTC=-0.6611236, Accumulated Leap Seconds=32
  { 2453736.50, 64.845183 }, //01 January 2006, UT1-UTC= 0.3388174, Accumulated Leap Seconds=33
  { 2453737.50, 64.845415 }, //02 January 2006, UT1-UTC= 0.3385849, Accumulated Leap Seconds=33
  { 2453738.50, 64.845892 }, //03 January 2006, UT1-UTC= 0.3381081, Accumulated Leap Seconds=33
  { 2453739.50, 64.846590 }, //04 January 2006, UT1-UTC= 0.3374102, Accumulated Leap Seconds=33
  { 2453740.50, 64.847457 }, //05 January 2006, UT1-UTC= 0.3365429, Accumulated Leap Seconds=33
  { 2453741.50, 64.848413 }, //06 January 2006, UT1-UTC= 0.3355865, Accumulated Leap Seconds=33
  { 2453742.50, 64.849321 }, //07 January 2006, UT1-UTC= 0.3346786, Accumulated Leap Seconds=33
  { 2453743.50, 64.850043 }, //08 January 2006, UT1-UTC= 0.3339572, Accumulated Leap Seconds=33
  { 2453744.50, 64.850506 }, //09 January 2006, UT1-UTC= 0.3334938, Accumulated Leap Seconds=33
  { 2453745.50, 64.850709 }, //10 January 2006, UT1-UTC= 0.3332906, Accumulated Leap Seconds=33
  { 2453746.50, 64.850692 }, //11 January 2006, UT1-UTC= 0.3333082, Accumulated Leap Seconds=33
  { 2453747.50, 64.850500 }, //12 January 2006, UT1-UTC= 0.3334999, Accumulated Leap Seconds=33
  { 2453748.50, 64.850227 }, //13 January 2006, UT1-UTC= 0.3337734, Accumulated Leap Seconds=33
  { 2453749.50, 64.850000 }, //14 January 2006, UT1-UTC= 0.3339995, Accumulated Leap Seconds=33
  { 2453750.50, 64.849911 }, //15 January 2006, UT1-UTC= 0.3340894, Accumulated Leap Seconds=33
  { 2453751.50, 64.849924 }, //16 January 2006, UT1-UTC= 0.3340758, Accumulated Leap Seconds=33
  { 2453752.50, 64.850056 }, //17 January 2006, UT1-UTC= 0.3339443, Accumulated Leap Seconds=33
  { 2453753.50, 64.850354 }, //18 January 2006, UT1-UTC= 0.3336460, Accumulated Leap Seconds=33
  { 2453754.50, 64.850867 }, //19 January 2006, UT1-UTC= 0.3331328, Accumulated Leap Seconds=33
  { 2453755.50, 64.851455 }, //20 January 2006, UT1-UTC= 0.3325448, Accumulated Leap Seconds=33
  { 2453756.50, 64.852118 }, //21 January 2006, UT1-UTC= 0.3318820, Accumulated Leap Seconds=33
  { 2453757.50, 64.852809 }, //22 January 2006, UT1-UTC= 0.3311906, Accumulated Leap Seconds=33
  { 2453758.50, 64.853459 }, //23 January 2006, UT1-UTC= 0.3305408, Accumulated Leap Seconds=33
  { 2453759.50, 64.854010 }, //24 January 2006, UT1-UTC= 0.3299901, Accumulated Leap Seconds=33
  { 2453760.50, 64.854483 }, //25 January 2006, UT1-UTC= 0.3295166, Accumulated Leap Seconds=33
  { 2453761.50, 64.854880 }, //26 January 2006, UT1-UTC= 0.3291204, Accumulated Leap Seconds=33
  { 2453762.50, 64.855275 }, //27 January 2006, UT1-UTC= 0.3287245, Accumulated Leap Seconds=33
  { 2453763.50, 64.855726 }, //28 January 2006, UT1-UTC= 0.3282743, Accumulated Leap Seconds=33
  { 2453764.50, 64.856304 }, //29 January 2006, UT1-UTC= 0.3276956, Accumulated Leap Seconds=33
  { 2453765.50, 64.857135 }, //30 January 2006, UT1-UTC= 0.3268650, Accumulated Leap Seconds=33
  { 2453766.50, 64.858271 }, //31 January 2006, UT1-UTC= 0.3257288, Accumulated Leap Seconds=33
  { 2453767.50, 64.859701 }, //01 February 2006, UT1-UTC= 0.3242987, Accumulated Leap Seconds=33
  { 2453768.50, 64.861274 }, //02 February 2006, UT1-UTC= 0.3227261, Accumulated Leap Seconds=33
  { 2453769.50, 64.862818 }, //03 February 2006, UT1-UTC= 0.3211822, Accumulated Leap Seconds=33
  { 2453770.50, 64.864243 }, //04 February 2006, UT1-UTC= 0.3197569, Accumulated Leap Seconds=33
  { 2453771.50, 64.865372 }, //05 February 2006, UT1-UTC= 0.3186277, Accumulated Leap Seconds=33
  { 2453772.50, 64.866212 }, //06 February 2006, UT1-UTC= 0.3177880, Accumulated Leap Seconds=33
  { 2453773.50, 64.866783 }, //07 February 2006, UT1-UTC= 0.3172169, Accumulated Leap Seconds=33
  { 2453774.50, 64.867181 }, //08 February 2006, UT1-UTC= 0.3168191, Accumulated Leap Seconds=33
  { 2453775.50, 64.867532 }, //09 February 2006, UT1-UTC= 0.3164682, Accumulated Leap Seconds=33
  { 2453776.50, 64.867922 }, //10 February 2006, UT1-UTC= 0.3160777, Accumulated Leap Seconds=33
  { 2453777.50, 64.868398 }, //11 February 2006, UT1-UTC= 0.3156025, Accumulated Leap Seconds=33
  { 2453778.50, 64.869005 }, //12 February 2006, UT1-UTC= 0.3149950, Accumulated Leap Seconds=33
  { 2453779.50, 64.869778 }, //13 February 2006, UT1-UTC= 0.3142223, Accumulated Leap Seconds=33
  { 2453780.50, 64.870660 }, //14 February 2006, UT1-UTC= 0.3133398, Accumulated Leap Seconds=33
  { 2453781.50, 64.871624 }, //15 February 2006, UT1-UTC= 0.3123764, Accumulated Leap Seconds=33
  { 2453782.50, 64.872648 }, //16 February 2006, UT1-UTC= 0.3113515, Accumulated Leap Seconds=33
  { 2453783.50, 64.873738 }, //17 February 2006, UT1-UTC= 0.3102624, Accumulated Leap Seconds=33
  { 2453784.50, 64.874854 }, //18 February 2006, UT1-UTC= 0.3091458, Accumulated Leap Seconds=33
  { 2453785.50, 64.875911 }, //19 February 2006, UT1-UTC= 0.3080886, Accumulated Leap Seconds=33
  { 2453786.50, 64.876866 }, //20 February 2006, UT1-UTC= 0.3071335, Accumulated Leap Seconds=33
  { 2453787.50, 64.877627 }, //21 February 2006, UT1-UTC= 0.3063731, Accumulated Leap Seconds=33
  { 2453788.50, 64.878173 }, //22 February 2006, UT1-UTC= 0.3058274, Accumulated Leap Seconds=33
  { 2453789.50, 64.878576 }, //23 February 2006, UT1-UTC= 0.3054235, Accumulated Leap Seconds=33
  { 2453790.50, 64.878950 }, //24 February 2006, UT1-UTC= 0.3050502, Accumulated Leap Seconds=33
  { 2453791.50, 64.879486 }, //25 February 2006, UT1-UTC= 0.3045144, Accumulated Leap Seconds=33
  { 2453792.50, 64.880295 }, //26 February 2006, UT1-UTC= 0.3037047, Accumulated Leap Seconds=33
  { 2453793.50, 64.881507 }, //27 February 2006, UT1-UTC= 0.3024929, Accumulated Leap Seconds=33
  { 2453794.50, 64.883106 }, //28 February 2006, UT1-UTC= 0.3008944, Accumulated Leap Seconds=33
  { 2453795.50, 64.884993 }, //01 March 2006, UT1-UTC= 0.2990071, Accumulated Leap Seconds=33
  { 2453796.50, 64.886990 }, //02 March 2006, UT1-UTC= 0.2970097, Accumulated Leap Seconds=33
  { 2453797.50, 64.888814 }, //03 March 2006, UT1-UTC= 0.2951858, Accumulated Leap Seconds=33
  { 2453798.50, 64.890398 }, //04 March 2006, UT1-UTC= 0.2936023, Accumulated Leap Seconds=33
  { 2453799.50, 64.891669 }, //05 March 2006, UT1-UTC= 0.2923314, Accumulated Leap Seconds=33
  { 2453800.50, 64.892591 }, //06 March 2006, UT1-UTC= 0.2914088, Accumulated Leap Seconds=33
  { 2453801.50, 64.893236 }, //07 March 2006, UT1-UTC= 0.2907642, Accumulated Leap Seconds=33
  { 2453802.50, 64.893697 }, //08 March 2006, UT1-UTC= 0.2903027, Accumulated Leap Seconds=33
  { 2453803.50, 64.894076 }, //09 March 2006, UT1-UTC= 0.2899238, Accumulated Leap Seconds=33
  { 2453804.50, 64.894431 }, //10 March 2006, UT1-UTC= 0.2895687, Accumulated Leap Seconds=33
  { 2453805.50, 64.894825 }, //11 March 2006, UT1-UTC= 0.2891746, Accumulated Leap Seconds=33
  { 2453806.50, 64.895327 }, //12 March 2006, UT1-UTC= 0.2886732, Accumulated Leap Seconds=33
  { 2453807.50, 64.896020 }, //13 March 2006, UT1-UTC= 0.2879796, Accumulated Leap Seconds=33
  { 2453808.50, 64.896881 }, //14 March 2006, UT1-UTC= 0.2871185, Accumulated Leap Seconds=33
  { 2453809.50, 64.897886 }, //15 March 2006, UT1-UTC= 0.2861136, Accumulated Leap Seconds=33
  { 2453810.50, 64.898939 }, //16 March 2006, UT1-UTC= 0.2850614, Accumulated Leap Seconds=33
  { 2453811.50, 64.899963 }, //17 March 2006, UT1-UTC= 0.2840371, Accumulated Leap Seconds=33
  { 2453812.50, 64.900898 }, //18 March 2006, UT1-UTC= 0.2831017, Accumulated Leap Seconds=33
  { 2453813.50, 64.901781 }, //19 March 2006, UT1-UTC= 0.2822194, Accumulated Leap Seconds=33
  { 2453814.50, 64.902530 }, //20 March 2006, UT1-UTC= 0.2814702, Accumulated Leap Seconds=33
  { 2453815.50, 64.903115 }, //21 March 2006, UT1-UTC= 0.2808846, Accumulated Leap Seconds=33
  { 2453816.50, 64.903606 }, //22 March 2006, UT1-UTC= 0.2803937, Accumulated Leap Seconds=33
  { 2453817.50, 64.904103 }, //23 March 2006, UT1-UTC= 0.2798968, Accumulated Leap Seconds=33
  { 2453818.50, 64.904675 }, //24 March 2006, UT1-UTC= 0.2793249, Accumulated Leap Seconds=33
  { 2453819.50, 64.905480 }, //25 March 2006, UT1-UTC= 0.2785200, Accumulated Leap Seconds=33
  { 2453820.50, 64.906663 }, //26 March 2006, UT1-UTC= 0.2773367, Accumulated Leap Seconds=33
  { 2453821.50, 64.908206 }, //27 March 2006, UT1-UTC= 0.2757941, Accumulated Leap Seconds=33
  { 2453822.50, 64.910035 }, //28 March 2006, UT1-UTC= 0.2739651, Accumulated Leap Seconds=33
  { 2453823.50, 64.912083 }, //29 March 2006, UT1-UTC= 0.2719172, Accumulated Leap Seconds=33
  { 2453824.50, 64.914138 }, //30 March 2006, UT1-UTC= 0.2698617, Accumulated Leap Seconds=33
  { 2453825.50, 64.915982 }, //31 March 2006, UT1-UTC= 0.2680182, Accumulated Leap Seconds=33
  { 2453826.50, 64.917499 }, //01 April 2006, UT1-UTC= 0.2665009, Accumulated Leap Seconds=33
  { 2453827.50, 64.918628 }, //02 April 2006, UT1-UTC= 0.2653722, Accumulated Leap Seconds=33
  { 2453828.50, 64.919446 }, //03 April 2006, UT1-UTC= 0.2645537, Accumulated Leap Seconds=33
  { 2453829.50, 64.920041 }, //04 April 2006, UT1-UTC= 0.2639591, Accumulated Leap Seconds=33
  { 2453830.50, 64.920591 }, //05 April 2006, UT1-UTC= 0.2634089, Accumulated Leap Seconds=33
  { 2453831.50, 64.921158 }, //06 April 2006, UT1-UTC= 0.2628423, Accumulated Leap Seconds=33
  { 2453832.50, 64.921806 }, //07 April 2006, UT1-UTC= 0.2621936, Accumulated Leap Seconds=33
  { 2453833.50, 64.922579 }, //08 April 2006, UT1-UTC= 0.2614213, Accumulated Leap Seconds=33
  { 2453834.50, 64.923564 }, //09 April 2006, UT1-UTC= 0.2604365, Accumulated Leap Seconds=33
  { 2453835.50, 64.924742 }, //10 April 2006, UT1-UTC= 0.2592584, Accumulated Leap Seconds=33
  { 2453836.50, 64.925970 }, //11 April 2006, UT1-UTC= 0.2580300, Accumulated Leap Seconds=33
  { 2453837.50, 64.927223 }, //12 April 2006, UT1-UTC= 0.2567767, Accumulated Leap Seconds=33
  { 2453838.50, 64.928464 }, //13 April 2006, UT1-UTC= 0.2555359, Accumulated Leap Seconds=33
  { 2453839.50, 64.929647 }, //14 April 2006, UT1-UTC= 0.2543535, Accumulated Leap Seconds=33
  { 2453840.50, 64.930668 }, //15 April 2006, UT1-UTC= 0.2533323, Accumulated Leap Seconds=33
  { 2453841.50, 64.931490 }, //16 April 2006, UT1-UTC= 0.2525104, Accumulated Leap Seconds=33
  { 2453842.50, 64.932142 }, //17 April 2006, UT1-UTC= 0.2518580, Accumulated Leap Seconds=33
  { 2453843.50, 64.932676 }, //18 April 2006, UT1-UTC= 0.2513237, Accumulated Leap Seconds=33
  { 2453844.50, 64.933169 }, //19 April 2006, UT1-UTC= 0.2508310, Accumulated Leap Seconds=33
  { 2453845.50, 64.933694 }, //20 April 2006, UT1-UTC= 0.2503061, Accumulated Leap Seconds=33
  { 2453846.50, 64.934337 }, //21 April 2006, UT1-UTC= 0.2496625, Accumulated Leap Seconds=33
  { 2453847.50, 64.935220 }, //22 April 2006, UT1-UTC= 0.2487799, Accumulated Leap Seconds=33
  { 2453848.50, 64.936436 }, //23 April 2006, UT1-UTC= 0.2475639, Accumulated Leap Seconds=33
  { 2453849.50, 64.937982 }, //24 April 2006, UT1-UTC= 0.2460181, Accumulated Leap Seconds=33
  { 2453850.50, 64.939741 }, //25 April 2006, UT1-UTC= 0.2442587, Accumulated Leap Seconds=33
  { 2453851.50, 64.941574 }, //26 April 2006, UT1-UTC= 0.2424260, Accumulated Leap Seconds=33
  { 2453852.50, 64.943352 }, //27 April 2006, UT1-UTC= 0.2406482, Accumulated Leap Seconds=33
  { 2453853.50, 64.944957 }, //28 April 2006, UT1-UTC= 0.2390430, Accumulated Leap Seconds=33
  { 2453854.50, 64.946290 }, //29 April 2006, UT1-UTC= 0.2377098, Accumulated Leap Seconds=33
  { 2453855.50, 64.947285 }, //30 April 2006, UT1-UTC= 0.2367145, Accumulated Leap Seconds=33
  { 2453856.50, 64.948030 }, //01 May 2006, UT1-UTC= 0.2359702, Accumulated Leap Seconds=33
  { 2453857.50, 64.948728 }, //02 May 2006, UT1-UTC= 0.2352722, Accumulated Leap Seconds=33
  { 2453858.50, 64.949514 }, //03 May 2006, UT1-UTC= 0.2344862, Accumulated Leap Seconds=33
  { 2453859.50, 64.950380 }, //04 May 2006, UT1-UTC= 0.2336195, Accumulated Leap Seconds=33
  { 2453860.50, 64.951379 }, //05 May 2006, UT1-UTC= 0.2326215, Accumulated Leap Seconds=33
  { 2453861.50, 64.952498 }, //06 May 2006, UT1-UTC= 0.2315021, Accumulated Leap Seconds=33
  { 2453862.50, 64.953780 }, //07 May 2006, UT1-UTC= 0.2302197, Accumulated Leap Seconds=33
  { 2453863.50, 64.955144 }, //08 May 2006, UT1-UTC= 0.2288564, Accumulated Leap Seconds=33
  { 2453864.50, 64.956603 }, //09 May 2006, UT1-UTC= 0.2273970, Accumulated Leap Seconds=33
  { 2453865.50, 64.958113 }, //10 May 2006, UT1-UTC= 0.2258866, Accumulated Leap Seconds=33
  { 2453866.50, 64.959610 }, //11 May 2006, UT1-UTC= 0.2243900, Accumulated Leap Seconds=33
  { 2453867.50, 64.961048 }, //12 May 2006, UT1-UTC= 0.2229523, Accumulated Leap Seconds=33
  { 2453868.50, 64.962368 }, //13 May 2006, UT1-UTC= 0.2216316, Accumulated Leap Seconds=33
  { 2453869.50, 64.963507 }, //14 May 2006, UT1-UTC= 0.2204928, Accumulated Leap Seconds=33
  { 2453870.50, 64.964443 }, //15 May 2006, UT1-UTC= 0.2195575, Accumulated Leap Seconds=33
  { 2453871.50, 64.965276 }, //16 May 2006, UT1-UTC= 0.2187243, Accumulated Leap Seconds=33
  { 2453872.50, 64.966033 }, //17 May 2006, UT1-UTC= 0.2179668, Accumulated Leap Seconds=33
  { 2453873.50, 64.966791 }, //18 May 2006, UT1-UTC= 0.2172085, Accumulated Leap Seconds=33
  { 2453874.50, 64.967708 }, //19 May 2006, UT1-UTC= 0.2162920, Accumulated Leap Seconds=33
  { 2453875.50, 64.968854 }, //20 May 2006, UT1-UTC= 0.2151456, Accumulated Leap Seconds=33
  { 2453876.50, 64.970205 }, //21 May 2006, UT1-UTC= 0.2137946, Accumulated Leap Seconds=33
  { 2453877.50, 64.971709 }, //22 May 2006, UT1-UTC= 0.2122906, Accumulated Leap Seconds=33
  { 2453878.50, 64.973291 }, //23 May 2006, UT1-UTC= 0.2107089, Accumulated Leap Seconds=33
  { 2453879.50, 64.974828 }, //24 May 2006, UT1-UTC= 0.2091717, Accumulated Leap Seconds=33
  { 2453880.50, 64.976191 }, //25 May 2006, UT1-UTC= 0.2078088, Accumulated Leap Seconds=33
  { 2453881.50, 64.977311 }, //26 May 2006, UT1-UTC= 0.2066891, Accumulated Leap Seconds=33
  { 2453882.50, 64.978072 }, //27 May 2006, UT1-UTC= 0.2059276, Accumulated Leap Seconds=33
  { 2453883.50, 64.978504 }, //28 May 2006, UT1-UTC= 0.2054963, Accumulated Leap Seconds=33
  { 2453884.50, 64.978712 }, //29 May 2006, UT1-UTC= 0.2052877, Accumulated Leap Seconds=33
  { 2453885.50, 64.978862 }, //30 May 2006, UT1-UTC= 0.2051379, Accumulated Leap Seconds=33
  { 2453886.50, 64.979083 }, //31 May 2006, UT1-UTC= 0.2049170, Accumulated Leap Seconds=33
  { 2453887.50, 64.979388 }, //01 June 2006, UT1-UTC= 0.2046119, Accumulated Leap Seconds=33
  { 2453888.50, 64.979868 }, //02 June 2006, UT1-UTC= 0.2041320, Accumulated Leap Seconds=33
  { 2453889.50, 64.980503 }, //03 June 2006, UT1-UTC= 0.2034972, Accumulated Leap Seconds=33
  { 2453890.50, 64.981252 }, //04 June 2006, UT1-UTC= 0.2027478, Accumulated Leap Seconds=33
  { 2453891.50, 64.982042 }, //05 June 2006, UT1-UTC= 0.2019577, Accumulated Leap Seconds=33
  { 2453892.50, 64.982801 }, //06 June 2006, UT1-UTC= 0.2011993, Accumulated Leap Seconds=33
  { 2453893.50, 64.983513 }, //07 June 2006, UT1-UTC= 0.2004874, Accumulated Leap Seconds=33
  { 2453894.50, 64.984082 }, //08 June 2006, UT1-UTC= 0.1999179, Accumulated Leap Seconds=33
  { 2453895.50, 64.984434 }, //09 June 2006, UT1-UTC= 0.1995655, Accumulated Leap Seconds=33
  { 2453896.50, 64.984522 }, //10 June 2006, UT1-UTC= 0.1994777, Accumulated Leap Seconds=33
  { 2453897.50, 64.984365 }, //11 June 2006, UT1-UTC= 0.1996352, Accumulated Leap Seconds=33
  { 2453898.50, 64.984016 }, //12 June 2006, UT1-UTC= 0.1999840, Accumulated Leap Seconds=33
  { 2453899.50, 64.983625 }, //13 June 2006, UT1-UTC= 0.2003747, Accumulated Leap Seconds=33
  { 2453900.50, 64.983334 }, //14 June 2006, UT1-UTC= 0.2006663, Accumulated Leap Seconds=33
  { 2453901.50, 64.983226 }, //15 June 2006, UT1-UTC= 0.2007738, Accumulated Leap Seconds=33
  { 2453902.50, 64.983376 }, //16 June 2006, UT1-UTC= 0.2006238, Accumulated Leap Seconds=33
  { 2453903.50, 64.983854 }, //17 June 2006, UT1-UTC= 0.2001456, Accumulated Leap Seconds=33
  { 2453904.50, 64.984599 }, //18 June 2006, UT1-UTC= 0.1994009, Accumulated Leap Seconds=33
  { 2453905.50, 64.985456 }, //19 June 2006, UT1-UTC= 0.1985436, Accumulated Leap Seconds=33
  { 2453906.50, 64.986300 }, //20 June 2006, UT1-UTC= 0.1977000, Accumulated Leap Seconds=33
  { 2453907.50, 64.986986 }, //21 June 2006, UT1-UTC= 0.1970141, Accumulated Leap Seconds=33
  { 2453908.50, 64.987502 }, //22 June 2006, UT1-UTC= 0.1964981, Accumulated Leap Seconds=33
  { 2453909.50, 64.987783 }, //23 June 2006, UT1-UTC= 0.1962165, Accumulated Leap Seconds=33
  { 2453910.50, 64.987886 }, //24 June 2006, UT1-UTC= 0.1961139, Accumulated Leap Seconds=33
  { 2453911.50, 64.987804 }, //25 June 2006, UT1-UTC= 0.1961956, Accumulated Leap Seconds=33
  { 2453912.50, 64.987690 }, //26 June 2006, UT1-UTC= 0.1963098, Accumulated Leap Seconds=33
  { 2453913.50, 64.987682 }, //27 June 2006, UT1-UTC= 0.1963182, Accumulated Leap Seconds=33
  { 2453914.50, 64.987824 }, //28 June 2006, UT1-UTC= 0.1961762, Accumulated Leap Seconds=33
  { 2453915.50, 64.988160 }, //29 June 2006, UT1-UTC= 0.1958401, Accumulated Leap Seconds=33
  { 2453916.50, 64.988728 }, //30 June 2006, UT1-UTC= 0.1952719, Accumulated Leap Seconds=33
  { 2453917.50, 64.989477 }, //01 July 2006, UT1-UTC= 0.1945228, Accumulated Leap Seconds=33
  { 2453918.50, 64.990394 }, //02 July 2006, UT1-UTC= 0.1936064, Accumulated Leap Seconds=33
  { 2453919.50, 64.991360 }, //03 July 2006, UT1-UTC= 0.1926397, Accumulated Leap Seconds=33
  { 2453920.50, 64.992326 }, //04 July 2006, UT1-UTC= 0.1916739, Accumulated Leap Seconds=33
  { 2453921.50, 64.993274 }, //05 July 2006, UT1-UTC= 0.1907265, Accumulated Leap Seconds=33
  { 2453922.50, 64.994091 }, //06 July 2006, UT1-UTC= 0.1899089, Accumulated Leap Seconds=33
  { 2453923.50, 64.994698 }, //07 July 2006, UT1-UTC= 0.1893020, Accumulated Leap Seconds=33
  { 2453924.50, 64.995103 }, //08 July 2006, UT1-UTC= 0.1888968, Accumulated Leap Seconds=33
  { 2453925.50, 64.995236 }, //09 July 2006, UT1-UTC= 0.1887639, Accumulated Leap Seconds=33
  { 2453926.50, 64.995257 }, //10 July 2006, UT1-UTC= 0.1887432, Accumulated Leap Seconds=33
  { 2453927.50, 64.995242 }, //11 July 2006, UT1-UTC= 0.1887579, Accumulated Leap Seconds=33
  { 2453928.50, 64.995324 }, //12 July 2006, UT1-UTC= 0.1886764, Accumulated Leap Seconds=33
  { 2453929.50, 64.995593 }, //13 July 2006, UT1-UTC= 0.1884070, Accumulated Leap Seconds=33
  { 2453930.50, 64.996191 }, //14 July 2006, UT1-UTC= 0.1878092, Accumulated Leap Seconds=33
  { 2453931.50, 64.997100 }, //15 July 2006, UT1-UTC= 0.1869002, Accumulated Leap Seconds=33
  { 2453932.50, 64.998206 }, //16 July 2006, UT1-UTC= 0.1857936, Accumulated Leap Seconds=33
  { 2453933.50, 64.999287 }, //17 July 2006, UT1-UTC= 0.1847131, Accumulated Leap Seconds=33
  { 2453934.50, 65.000209 }, //18 July 2006, UT1-UTC= 0.1837911, Accumulated Leap Seconds=33
  { 2453935.50, 65.000907 }, //19 July 2006, UT1-UTC= 0.1830931, Accumulated Leap Seconds=33
  { 2453936.50, 65.001332 }, //20 July 2006, UT1-UTC= 0.1826682, Accumulated Leap Seconds=33
  { 2453937.50, 65.001486 }, //21 July 2006, UT1-UTC= 0.1825142, Accumulated Leap Seconds=33
  { 2453938.50, 65.001411 }, //22 July 2006, UT1-UTC= 0.1825888, Accumulated Leap Seconds=33
  { 2453939.50, 65.001181 }, //23 July 2006, UT1-UTC= 0.1828187, Accumulated Leap Seconds=33
  { 2453940.50, 65.000918 }, //24 July 2006, UT1-UTC= 0.1830816, Accumulated Leap Seconds=33
  { 2453941.50, 65.000762 }, //25 July 2006, UT1-UTC= 0.1832384, Accumulated Leap Seconds=33
  { 2453942.50, 65.000731 }, //26 July 2006, UT1-UTC= 0.1832688, Accumulated Leap Seconds=33
  { 2453943.50, 65.000865 }, //27 July 2006, UT1-UTC= 0.1831348, Accumulated Leap Seconds=33
  { 2453944.50, 65.001106 }, //28 July 2006, UT1-UTC= 0.1828941, Accumulated Leap Seconds=33
  { 2453945.50, 65.001454 }, //29 July 2006, UT1-UTC= 0.1825462, Accumulated Leap Seconds=33
  { 2453946.50, 65.001886 }, //30 July 2006, UT1-UTC= 0.1821139, Accumulated Leap Seconds=33
  { 2453947.50, 65.002353 }, //31 July 2006, UT1-UTC= 0.1816465, Accumulated Leap Seconds=33
  { 2453948.50, 65.002815 }, //01 August 2006, UT1-UTC= 0.1811845, Accumulated Leap Seconds=33
  { 2453949.50, 65.003206 }, //02 August 2006, UT1-UTC= 0.1807943, Accumulated Leap Seconds=33
  { 2453950.50, 65.003427 }, //03 August 2006, UT1-UTC= 0.1805732, Accumulated Leap Seconds=33
  { 2453951.50, 65.003501 }, //04 August 2006, UT1-UTC= 0.1804988, Accumulated Leap Seconds=33
  { 2453952.50, 65.003455 }, //05 August 2006, UT1-UTC= 0.1805451, Accumulated Leap Seconds=33
  { 2453953.50, 65.003294 }, //06 August 2006, UT1-UTC= 0.1807064, Accumulated Leap Seconds=33
  { 2453954.50, 65.003090 }, //07 August 2006, UT1-UTC= 0.1809102, Accumulated Leap Seconds=33
  { 2453955.50, 65.003006 }, //08 August 2006, UT1-UTC= 0.1809936, Accumulated Leap Seconds=33
  { 2453956.50, 65.003165 }, //09 August 2006, UT1-UTC= 0.1808353, Accumulated Leap Seconds=33
  { 2453957.50, 65.003668 }, //10 August 2006, UT1-UTC= 0.1803323, Accumulated Leap Seconds=33
  { 2453958.50, 65.004505 }, //11 August 2006, UT1-UTC= 0.1794948, Accumulated Leap Seconds=33
  { 2453959.50, 65.005564 }, //12 August 2006, UT1-UTC= 0.1784364, Accumulated Leap Seconds=33
  { 2453960.50, 65.006679 }, //13 August 2006, UT1-UTC= 0.1773212, Accumulated Leap Seconds=33
  { 2453961.50, 65.007642 }, //14 August 2006, UT1-UTC= 0.1763583, Accumulated Leap Seconds=33
  { 2453962.50, 65.008348 }, //15 August 2006, UT1-UTC= 0.1756524, Accumulated Leap Seconds=33
  { 2453963.50, 65.008773 }, //16 August 2006, UT1-UTC= 0.1752268, Accumulated Leap Seconds=33
  { 2453964.50, 65.008947 }, //17 August 2006, UT1-UTC= 0.1750531, Accumulated Leap Seconds=33
  { 2453965.50, 65.008907 }, //18 August 2006, UT1-UTC= 0.1750933, Accumulated Leap Seconds=33
  { 2453966.50, 65.008745 }, //19 August 2006, UT1-UTC= 0.1752551, Accumulated Leap Seconds=33
  { 2453967.50, 65.008568 }, //20 August 2006, UT1-UTC= 0.1754316, Accumulated Leap Seconds=33
  { 2453968.50, 65.008522 }, //21 August 2006, UT1-UTC= 0.1754776, Accumulated Leap Seconds=33
  { 2453969.50, 65.008661 }, //22 August 2006, UT1-UTC= 0.1753393, Accumulated Leap Seconds=33
  { 2453970.50, 65.008969 }, //23 August 2006, UT1-UTC= 0.1750308, Accumulated Leap Seconds=33
  { 2453971.50, 65.009477 }, //24 August 2006, UT1-UTC= 0.1745229, Accumulated Leap Seconds=33
  { 2453972.50, 65.010120 }, //25 August 2006, UT1-UTC= 0.1738797, Accumulated Leap Seconds=33
  { 2453973.50, 65.010832 }, //26 August 2006, UT1-UTC= 0.1731683, Accumulated Leap Seconds=33
  { 2453974.50, 65.011556 }, //27 August 2006, UT1-UTC= 0.1724444, Accumulated Leap Seconds=33
  { 2453975.50, 65.012257 }, //28 August 2006, UT1-UTC= 0.1717434, Accumulated Leap Seconds=33
  { 2453976.50, 65.012878 }, //29 August 2006, UT1-UTC= 0.1711223, Accumulated Leap Seconds=33
  { 2453977.50, 65.013380 }, //30 August 2006, UT1-UTC= 0.1706200, Accumulated Leap Seconds=33
  { 2453978.50, 65.013697 }, //31 August 2006, UT1-UTC= 0.1703026, Accumulated Leap Seconds=33
  { 2453979.50, 65.013819 }, //01 September 2006, UT1-UTC= 0.1701807, Accumulated Leap Seconds=33
  { 2453980.50, 65.013852 }, //02 September 2006, UT1-UTC= 0.1701477, Accumulated Leap Seconds=33
  { 2453981.50, 65.013851 }, //03 September 2006, UT1-UTC= 0.1701488, Accumulated Leap Seconds=33
  { 2453982.50, 65.013958 }, //04 September 2006, UT1-UTC= 0.1700416, Accumulated Leap Seconds=33
  { 2453983.50, 65.014338 }, //05 September 2006, UT1-UTC= 0.1696617, Accumulated Leap Seconds=33
  { 2453984.50, 65.015049 }, //06 September 2006, UT1-UTC= 0.1689509, Accumulated Leap Seconds=33
  { 2453985.50, 65.016127 }, //07 September 2006, UT1-UTC= 0.1678727, Accumulated Leap Seconds=33
  { 2453986.50, 65.017529 }, //08 September 2006, UT1-UTC= 0.1664709, Accumulated Leap Seconds=33
  { 2453987.50, 65.019096 }, //09 September 2006, UT1-UTC= 0.1649044, Accumulated Leap Seconds=33
  { 2453988.50, 65.020623 }, //10 September 2006, UT1-UTC= 0.1633765, Accumulated Leap Seconds=33
  { 2453989.50, 65.021938 }, //11 September 2006, UT1-UTC= 0.1620623, Accumulated Leap Seconds=33
  { 2453990.50, 65.022927 }, //12 September 2006, UT1-UTC= 0.1610725, Accumulated Leap Seconds=33
  { 2453991.50, 65.023631 }, //13 September 2006, UT1-UTC= 0.1603686, Accumulated Leap Seconds=33
  { 2453992.50, 65.024170 }, //14 September 2006, UT1-UTC= 0.1598297, Accumulated Leap Seconds=33
  { 2453993.50, 65.024519 }, //15 September 2006, UT1-UTC= 0.1594810, Accumulated Leap Seconds=33
  { 2453994.50, 65.024821 }, //16 September 2006, UT1-UTC= 0.1591795, Accumulated Leap Seconds=33
  { 2453995.50, 65.025156 }, //17 September 2006, UT1-UTC= 0.1588435, Accumulated Leap Seconds=33
  { 2453996.50, 65.025616 }, //18 September 2006, UT1-UTC= 0.1583836, Accumulated Leap Seconds=33
  { 2453997.50, 65.026271 }, //19 September 2006, UT1-UTC= 0.1577286, Accumulated Leap Seconds=33
  { 2453998.50, 65.027118 }, //20 September 2006, UT1-UTC= 0.1568821, Accumulated Leap Seconds=33
  { 2453999.50, 65.028104 }, //21 September 2006, UT1-UTC= 0.1558963, Accumulated Leap Seconds=33
  { 2454000.50, 65.029229 }, //22 September 2006, UT1-UTC= 0.1547711, Accumulated Leap Seconds=33
  { 2454001.50, 65.030418 }, //23 September 2006, UT1-UTC= 0.1535816, Accumulated Leap Seconds=33
  { 2454002.50, 65.031560 }, //24 September 2006, UT1-UTC= 0.1524397, Accumulated Leap Seconds=33
  { 2454003.50, 65.032634 }, //25 September 2006, UT1-UTC= 0.1513663, Accumulated Leap Seconds=33
  { 2454004.50, 65.033601 }, //26 September 2006, UT1-UTC= 0.1503988, Accumulated Leap Seconds=33
  { 2454005.50, 65.034447 }, //27 September 2006, UT1-UTC= 0.1495535, Accumulated Leap Seconds=33
  { 2454006.50, 65.035167 }, //28 September 2006, UT1-UTC= 0.1488330, Accumulated Leap Seconds=33
  { 2454007.50, 65.035777 }, //29 September 2006, UT1-UTC= 0.1482233, Accumulated Leap Seconds=33
  { 2454008.50, 65.036410 }, //30 September 2006, UT1-UTC= 0.1475897, Accumulated Leap Seconds=33
  { 2454009.50, 65.037143 }, //01 October 2006, UT1-UTC= 0.1468568, Accumulated Leap Seconds=33
  { 2454010.50, 65.037999 }, //02 October 2006, UT1-UTC= 0.1460013, Accumulated Leap Seconds=33
  { 2454011.50, 65.039116 }, //03 October 2006, UT1-UTC= 0.1448836, Accumulated Leap Seconds=33
  { 2454012.50, 65.040591 }, //04 October 2006, UT1-UTC= 0.1434088, Accumulated Leap Seconds=33
  { 2454013.50, 65.042413 }, //05 October 2006, UT1-UTC= 0.1415869, Accumulated Leap Seconds=33
  { 2454014.50, 65.044551 }, //06 October 2006, UT1-UTC= 0.1394488, Accumulated Leap Seconds=33
  { 2454015.50, 65.046818 }, //07 October 2006, UT1-UTC= 0.1371819, Accumulated Leap Seconds=33
  { 2454016.50, 65.048993 }, //08 October 2006, UT1-UTC= 0.1350067, Accumulated Leap Seconds=33
  { 2454017.50, 65.050876 }, //09 October 2006, UT1-UTC= 0.1331237, Accumulated Leap Seconds=33
  { 2454018.50, 65.052414 }, //10 October 2006, UT1-UTC= 0.1315863, Accumulated Leap Seconds=33
  { 2454019.50, 65.053640 }, //11 October 2006, UT1-UTC= 0.1303598, Accumulated Leap Seconds=33
  { 2454020.50, 65.054680 }, //12 October 2006, UT1-UTC= 0.1293198, Accumulated Leap Seconds=33
  { 2454021.50, 65.055605 }, //13 October 2006, UT1-UTC= 0.1283950, Accumulated Leap Seconds=33
  { 2454022.50, 65.056431 }, //14 October 2006, UT1-UTC= 0.1275689, Accumulated Leap Seconds=33
  { 2454023.50, 65.057320 }, //15 October 2006, UT1-UTC= 0.1266799, Accumulated Leap Seconds=33
  { 2454024.50, 65.058328 }, //16 October 2006, UT1-UTC= 0.1256723, Accumulated Leap Seconds=33
  { 2454025.50, 65.059492 }, //17 October 2006, UT1-UTC= 0.1245078, Accumulated Leap Seconds=33
  { 2454026.50, 65.060796 }, //18 October 2006, UT1-UTC= 0.1232041, Accumulated Leap Seconds=33
  { 2454027.50, 65.062209 }, //19 October 2006, UT1-UTC= 0.1217907, Accumulated Leap Seconds=33
  { 2454028.50, 65.063694 }, //20 October 2006, UT1-UTC= 0.1203059, Accumulated Leap Seconds=33
  { 2454029.50, 65.065179 }, //21 October 2006, UT1-UTC= 0.1188213, Accumulated Leap Seconds=33
  { 2454030.50, 65.066625 }, //22 October 2006, UT1-UTC= 0.1173748, Accumulated Leap Seconds=33
  { 2454031.50, 65.067978 }, //23 October 2006, UT1-UTC= 0.1160217, Accumulated Leap Seconds=33
  { 2454032.50, 65.069206 }, //24 October 2006, UT1-UTC= 0.1147936, Accumulated Leap Seconds=33
  { 2454033.50, 65.070258 }, //25 October 2006, UT1-UTC= 0.1137423, Accumulated Leap Seconds=33
  { 2454034.50, 65.071144 }, //26 October 2006, UT1-UTC= 0.1128561, Accumulated Leap Seconds=33
  { 2454035.50, 65.071950 }, //27 October 2006, UT1-UTC= 0.1120498, Accumulated Leap Seconds=33
  { 2454036.50, 65.072687 }, //28 October 2006, UT1-UTC= 0.1113128, Accumulated Leap Seconds=33
  { 2454037.50, 65.073488 }, //29 October 2006, UT1-UTC= 0.1105119, Accumulated Leap Seconds=33
  { 2454038.50, 65.074471 }, //30 October 2006, UT1-UTC= 0.1095287, Accumulated Leap Seconds=33
  { 2454039.50, 65.075724 }, //31 October 2006, UT1-UTC= 0.1082756, Accumulated Leap Seconds=33
  { 2454040.50, 65.077260 }, //01 November 2006, UT1-UTC= 0.1067403, Accumulated Leap Seconds=33
  { 2454041.50, 65.079099 }, //02 November 2006, UT1-UTC= 0.1049005, Accumulated Leap Seconds=33
  { 2454042.50, 65.081105 }, //03 November 2006, UT1-UTC= 0.1028952, Accumulated Leap Seconds=33
  { 2454043.50, 65.083084 }, //04 November 2006, UT1-UTC= 0.1009160, Accumulated Leap Seconds=33
  { 2454044.50, 65.084868 }, //05 November 2006, UT1-UTC= 0.0991322, Accumulated Leap Seconds=33
  { 2454045.50, 65.086276 }, //06 November 2006, UT1-UTC= 0.0977237, Accumulated Leap Seconds=33
  { 2454046.50, 65.087356 }, //07 November 2006, UT1-UTC= 0.0966440, Accumulated Leap Seconds=33
  { 2454047.50, 65.088165 }, //08 November 2006, UT1-UTC= 0.0958346, Accumulated Leap Seconds=33
  { 2454048.50, 65.088855 }, //09 November 2006, UT1-UTC= 0.0951452, Accumulated Leap Seconds=33
  { 2454049.50, 65.089531 }, //10 November 2006, UT1-UTC= 0.0944694, Accumulated Leap Seconds=33
  { 2454050.50, 65.090249 }, //11 November 2006, UT1-UTC= 0.0937511, Accumulated Leap Seconds=33
  { 2454051.50, 65.091195 }, //12 November 2006, UT1-UTC= 0.0928053, Accumulated Leap Seconds=33
  { 2454052.50, 65.092256 }, //13 November 2006, UT1-UTC= 0.0917441, Accumulated Leap Seconds=33
  { 2454053.50, 65.093440 }, //14 November 2006, UT1-UTC= 0.0905599, Accumulated Leap Seconds=33
  { 2454054.50, 65.094688 }, //15 November 2006, UT1-UTC= 0.0893120, Accumulated Leap Seconds=33
  { 2454055.50, 65.095911 }, //16 November 2006, UT1-UTC= 0.0880885, Accumulated Leap Seconds=33
  { 2454056.50, 65.097155 }, //17 November 2006, UT1-UTC= 0.0868451, Accumulated Leap Seconds=33
  { 2454057.50, 65.098344 }, //18 November 2006, UT1-UTC= 0.0856560, Accumulated Leap Seconds=33
  { 2454058.50, 65.099422 }, //19 November 2006, UT1-UTC= 0.0845782, Accumulated Leap Seconds=33
  { 2454059.50, 65.100385 }, //20 November 2006, UT1-UTC= 0.0836147, Accumulated Leap Seconds=33
  { 2454060.50, 65.101222 }, //21 November 2006, UT1-UTC= 0.0827777, Accumulated Leap Seconds=33
  { 2454061.50, 65.101928 }, //22 November 2006, UT1-UTC= 0.0820722, Accumulated Leap Seconds=33
  { 2454062.50, 65.102537 }, //23 November 2006, UT1-UTC= 0.0814628, Accumulated Leap Seconds=33
  { 2454063.50, 65.103075 }, //24 November 2006, UT1-UTC= 0.0809251, Accumulated Leap Seconds=33
  { 2454064.50, 65.103681 }, //25 November 2006, UT1-UTC= 0.0803188, Accumulated Leap Seconds=33
  { 2454065.50, 65.104491 }, //26 November 2006, UT1-UTC= 0.0795089, Accumulated Leap Seconds=33
  { 2454066.50, 65.105587 }, //27 November 2006, UT1-UTC= 0.0784135, Accumulated Leap Seconds=33
  { 2454067.50, 65.106979 }, //28 November 2006, UT1-UTC= 0.0770213, Accumulated Leap Seconds=33
  { 2454068.50, 65.108616 }, //29 November 2006, UT1-UTC= 0.0753836, Accumulated Leap Seconds=33
  { 2454069.50, 65.110395 }, //30 November 2006, UT1-UTC= 0.0736049, Accumulated Leap Seconds=33
  { 2454070.50, 65.112197 }, //01 December 2006, UT1-UTC= 0.0718030, Accumulated Leap Seconds=33
  { 2454071.50, 65.113872 }, //02 December 2006, UT1-UTC= 0.0701278, Accumulated Leap Seconds=33
  { 2454072.50, 65.115255 }, //03 December 2006, UT1-UTC= 0.0687446, Accumulated Leap Seconds=33
  { 2454073.50, 65.116359 }, //04 December 2006, UT1-UTC= 0.0676410, Accumulated Leap Seconds=33
  { 2454074.50, 65.117247 }, //05 December 2006, UT1-UTC= 0.0667525, Accumulated Leap Seconds=33
  { 2454075.50, 65.118000 }, //06 December 2006, UT1-UTC= 0.0660001, Accumulated Leap Seconds=33
  { 2454076.50, 65.118688 }, //07 December 2006, UT1-UTC= 0.0653118, Accumulated Leap Seconds=33
  { 2454077.50, 65.119493 }, //08 December 2006, UT1-UTC= 0.0645070, Accumulated Leap Seconds=33
  { 2454078.50, 65.120440 }, //09 December 2006, UT1-UTC= 0.0635599, Accumulated Leap Seconds=33
  { 2454079.50, 65.121571 }, //10 December 2006, UT1-UTC= 0.0624287, Accumulated Leap Seconds=33
  { 2454080.50, 65.122862 }, //11 December 2006, UT1-UTC= 0.0611376, Accumulated Leap Seconds=33
  { 2454081.50, 65.124246 }, //12 December 2006, UT1-UTC= 0.0597538, Accumulated Leap Seconds=33
  { 2454082.50, 65.125693 }, //13 December 2006, UT1-UTC= 0.0583074, Accumulated Leap Seconds=33
  { 2454083.50, 65.127090 }, //14 December 2006, UT1-UTC= 0.0569097, Accumulated Leap Seconds=33
  { 2454084.50, 65.128351 }, //15 December 2006, UT1-UTC= 0.0556485, Accumulated Leap Seconds=33
  { 2454085.50, 65.129568 }, //16 December 2006, UT1-UTC= 0.0544320, Accumulated Leap Seconds=33
  { 2454086.50, 65.130609 }, //17 December 2006, UT1-UTC= 0.0533912, Accumulated Leap Seconds=33
  { 2454087.50, 65.131432 }, //18 December 2006, UT1-UTC= 0.0525678, Accumulated Leap Seconds=33
  { 2454088.50, 65.132133 }, //19 December 2006, UT1-UTC= 0.0518672, Accumulated Leap Seconds=33
  { 2454089.50, 65.132746 }, //20 December 2006, UT1-UTC= 0.0512543, Accumulated Leap Seconds=33
  { 2454090.50, 65.133273 }, //21 December 2006, UT1-UTC= 0.0507274, Accumulated Leap Seconds=33
  { 2454091.50, 65.133809 }, //22 December 2006, UT1-UTC= 0.0501906, Accumulated Leap Seconds=33
  { 2454092.50, 65.134521 }, //23 December 2006, UT1-UTC= 0.0494794, Accumulated Leap Seconds=33
  { 2454093.50, 65.135466 }, //24 December 2006, UT1-UTC= 0.0485339, Accumulated Leap Seconds=33
  { 2454094.50, 65.136687 }, //25 December 2006, UT1-UTC= 0.0473128, Accumulated Leap Seconds=33
  { 2454095.50, 65.138153 }, //26 December 2006, UT1-UTC= 0.0458470, Accumulated Leap Seconds=33
  { 2454096.50, 65.139781 }, //27 December 2006, UT1-UTC= 0.0442193, Accumulated Leap Seconds=33
  { 2454097.50, 65.141428 }, //28 December 2006, UT1-UTC= 0.0425721, Accumulated Leap Seconds=33
  { 2454098.50, 65.142975 }, //29 December 2006, UT1-UTC= 0.0410254, Accumulated Leap Seconds=33
  { 2454099.50, 65.144392 }, //30 December 2006, UT1-UTC= 0.0396077, Accumulated Leap Seconds=33
  { 2454100.50, 65.145536 }, //31 December 2006, UT1-UTC= 0.0384637, Accumulated Leap Seconds=33
  { 2454101.50, 65.146403 }, //01 January 2007, UT1-UTC= 0.0375966, Accumulated Leap Seconds=33
  { 2454102.50, 65.147059 }, //02 January 2007, UT1-UTC= 0.0369414, Accumulated Leap Seconds=33
  { 2454103.50, 65.147597 }, //03 January 2007, UT1-UTC= 0.0364027, Accumulated Leap Seconds=33
  { 2454104.50, 65.148142 }, //04 January 2007, UT1-UTC= 0.0358582, Accumulated Leap Seconds=33
  { 2454105.50, 65.148738 }, //05 January 2007, UT1-UTC= 0.0352623, Accumulated Leap Seconds=33
  { 2454106.50, 65.149536 }, //06 January 2007, UT1-UTC= 0.0344638, Accumulated Leap Seconds=33
  { 2454107.50, 65.150526 }, //07 January 2007, UT1-UTC= 0.0334739, Accumulated Leap Seconds=33
  { 2454108.50, 65.151668 }, //08 January 2007, UT1-UTC= 0.0323324, Accumulated Leap Seconds=33
  { 2454109.50, 65.152934 }, //09 January 2007, UT1-UTC= 0.0310659, Accumulated Leap Seconds=33
  { 2454110.50, 65.154285 }, //10 January 2007, UT1-UTC= 0.0297155, Accumulated Leap Seconds=33
  { 2454111.50, 65.155652 }, //11 January 2007, UT1-UTC= 0.0283476, Accumulated Leap Seconds=33
  { 2454112.50, 65.156921 }, //12 January 2007, UT1-UTC= 0.0270789, Accumulated Leap Seconds=33
  { 2454113.50, 65.158141 }, //13 January 2007, UT1-UTC= 0.0258594, Accumulated Leap Seconds=33
  { 2454114.50, 65.159229 }, //14 January 2007, UT1-UTC= 0.0247709, Accumulated Leap Seconds=33
  { 2454115.50, 65.160159 }, //15 January 2007, UT1-UTC= 0.0238413, Accumulated Leap Seconds=33
  { 2454116.50, 65.160973 }, //16 January 2007, UT1-UTC= 0.0230268, Accumulated Leap Seconds=33
  { 2454117.50, 65.161800 }, //17 January 2007, UT1-UTC= 0.0221999, Accumulated Leap Seconds=33
  { 2454118.50, 65.162601 }, //18 January 2007, UT1-UTC= 0.0213993, Accumulated Leap Seconds=33
  { 2454119.50, 65.163537 }, //19 January 2007, UT1-UTC= 0.0204626, Accumulated Leap Seconds=33
  { 2454120.50, 65.164777 }, //20 January 2007, UT1-UTC= 0.0192233, Accumulated Leap Seconds=33
  { 2454121.50, 65.166321 }, //21 January 2007, UT1-UTC= 0.0176790, Accumulated Leap Seconds=33
  { 2454122.50, 65.168140 }, //22 January 2007, UT1-UTC= 0.0158602, Accumulated Leap Seconds=33
  { 2454123.50, 65.170181 }, //23 January 2007, UT1-UTC= 0.0138192, Accumulated Leap Seconds=33
  { 2454124.50, 65.172324 }, //24 January 2007, UT1-UTC= 0.0116764, Accumulated Leap Seconds=33
  { 2454125.50, 65.174381 }, //25 January 2007, UT1-UTC= 0.0096191, Accumulated Leap Seconds=33
  { 2454126.50, 65.176209 }, //26 January 2007, UT1-UTC= 0.0077913, Accumulated Leap Seconds=33
  { 2454127.50, 65.177792 }, //27 January 2007, UT1-UTC= 0.0062075, Accumulated Leap Seconds=33
  { 2454128.50, 65.179101 }, //28 January 2007, UT1-UTC= 0.0048994, Accumulated Leap Seconds=33
  { 2454129.50, 65.180182 }, //29 January 2007, UT1-UTC= 0.0038184, Accumulated Leap Seconds=33
  { 2454130.50, 65.181170 }, //30 January 2007, UT1-UTC= 0.0028299, Accumulated Leap Seconds=33
  { 2454131.50, 65.182167 }, //31 January 2007, UT1-UTC= 0.0018325, Accumulated Leap Seconds=33
  { 2454132.50, 65.183264 }, //01 February 2007, UT1-UTC= 0.0007362, Accumulated Leap Seconds=33
  { 2454133.50, 65.184520 }, //02 February 2007, UT1-UTC=-0.0005201, Accumulated Leap Seconds=33
  { 2454134.50, 65.185959 }, //03 February 2007, UT1-UTC=-0.0019589, Accumulated Leap Seconds=33
  { 2454135.50, 65.187539 }, //04 February 2007, UT1-UTC=-0.0035391, Accumulated Leap Seconds=33
  { 2454136.50, 65.189208 }, //05 February 2007, UT1-UTC=-0.0052080, Accumulated Leap Seconds=33
  { 2454137.50, 65.190922 }, //06 February 2007, UT1-UTC=-0.0069220, Accumulated Leap Seconds=33
  { 2454138.50, 65.192607 }, //07 February 2007, UT1-UTC=-0.0086066, Accumulated Leap Seconds=33
  { 2454139.50, 65.194216 }, //08 February 2007, UT1-UTC=-0.0102163, Accumulated Leap Seconds=33
  { 2454140.50, 65.195661 }, //09 February 2007, UT1-UTC=-0.0116614, Accumulated Leap Seconds=33
  { 2454141.50, 65.196933 }, //10 February 2007, UT1-UTC=-0.0129331, Accumulated Leap Seconds=33
  { 2454142.50, 65.198050 }, //11 February 2007, UT1-UTC=-0.0140499, Accumulated Leap Seconds=33
  { 2454143.50, 65.199004 }, //12 February 2007, UT1-UTC=-0.0150040, Accumulated Leap Seconds=33
  { 2454144.50, 65.199806 }, //13 February 2007, UT1-UTC=-0.0158061, Accumulated Leap Seconds=33
  { 2454145.50, 65.200547 }, //14 February 2007, UT1-UTC=-0.0165468, Accumulated Leap Seconds=33
  { 2454146.50, 65.201289 }, //15 February 2007, UT1-UTC=-0.0172886, Accumulated Leap Seconds=33
  { 2454147.50, 65.202124 }, //16 February 2007, UT1-UTC=-0.0181243, Accumulated Leap Seconds=33
  { 2454148.50, 65.203134 }, //17 February 2007, UT1-UTC=-0.0191337, Accumulated Leap Seconds=33
  { 2454149.50, 65.204366 }, //18 February 2007, UT1-UTC=-0.0203665, Accumulated Leap Seconds=33
  { 2454150.50, 65.205876 }, //19 February 2007, UT1-UTC=-0.0218756, Accumulated Leap Seconds=33
  { 2454151.50, 65.207472 }, //20 February 2007, UT1-UTC=-0.0234719, Accumulated Leap Seconds=33
  { 2454152.50, 65.209002 }, //21 February 2007, UT1-UTC=-0.0250016, Accumulated Leap Seconds=33
  { 2454153.50, 65.210291 }, //22 February 2007, UT1-UTC=-0.0262915, Accumulated Leap Seconds=33
  { 2454154.50, 65.211275 }, //23 February 2007, UT1-UTC=-0.0272747, Accumulated Leap Seconds=33
  { 2454155.50, 65.212032 }, //24 February 2007, UT1-UTC=-0.0280325, Accumulated Leap Seconds=33
  { 2454156.50, 65.212576 }, //25 February 2007, UT1-UTC=-0.0285756, Accumulated Leap Seconds=33
  { 2454157.50, 65.213017 }, //26 February 2007, UT1-UTC=-0.0290170, Accumulated Leap Seconds=33
  { 2454158.50, 65.213438 }, //27 February 2007, UT1-UTC=-0.0294377, Accumulated Leap Seconds=33
  { 2454159.50, 65.213925 }, //28 February 2007, UT1-UTC=-0.0299252, Accumulated Leap Seconds=33
  { 2454160.50, 65.214536 }, //01 March 2007, UT1-UTC=-0.0305358, Accumulated Leap Seconds=33
  { 2454161.50, 65.215302 }, //02 March 2007, UT1-UTC=-0.0313021, Accumulated Leap Seconds=33
  { 2454162.50, 65.216255 }, //03 March 2007, UT1-UTC=-0.0322546, Accumulated Leap Seconds=33
  { 2454163.50, 65.217343 }, //04 March 2007, UT1-UTC=-0.0333426, Accumulated Leap Seconds=33
  { 2454164.50, 65.218540 }, //05 March 2007, UT1-UTC=-0.0345401, Accumulated Leap Seconds=33
  { 2454165.50, 65.219777 }, //06 March 2007, UT1-UTC=-0.0357775, Accumulated Leap Seconds=33
  { 2454166.50, 65.220981 }, //07 March 2007, UT1-UTC=-0.0369806, Accumulated Leap Seconds=33
  { 2454167.50, 65.222160 }, //08 March 2007, UT1-UTC=-0.0381604, Accumulated Leap Seconds=33
  { 2454168.50, 65.223229 }, //09 March 2007, UT1-UTC=-0.0392287, Accumulated Leap Seconds=33
  { 2454169.50, 65.224123 }, //10 March 2007, UT1-UTC=-0.0401227, Accumulated Leap Seconds=33
  { 2454170.50, 65.224860 }, //11 March 2007, UT1-UTC=-0.0408601, Accumulated Leap Seconds=33
  { 2454171.50, 65.225471 }, //12 March 2007, UT1-UTC=-0.0414710, Accumulated Leap Seconds=33
  { 2454172.50, 65.226023 }, //13 March 2007, UT1-UTC=-0.0420227, Accumulated Leap Seconds=33
  { 2454173.50, 65.226599 }, //14 March 2007, UT1-UTC=-0.0425989, Accumulated Leap Seconds=33
  { 2454174.50, 65.227280 }, //15 March 2007, UT1-UTC=-0.0432801, Accumulated Leap Seconds=33
  { 2454175.50, 65.228212 }, //16 March 2007, UT1-UTC=-0.0442119, Accumulated Leap Seconds=33
  { 2454176.50, 65.229534 }, //17 March 2007, UT1-UTC=-0.0455336, Accumulated Leap Seconds=33
  { 2454177.50, 65.231270 }, //18 March 2007, UT1-UTC=-0.0472700, Accumulated Leap Seconds=33
  { 2454178.50, 65.233265 }, //19 March 2007, UT1-UTC=-0.0492653, Accumulated Leap Seconds=33
  { 2454179.50, 65.235366 }, //20 March 2007, UT1-UTC=-0.0513664, Accumulated Leap Seconds=33
  { 2454180.50, 65.237400 }, //21 March 2007, UT1-UTC=-0.0533998, Accumulated Leap Seconds=33
  { 2454181.50, 65.239120 }, //22 March 2007, UT1-UTC=-0.0551204, Accumulated Leap Seconds=33
  { 2454182.50, 65.240415 }, //23 March 2007, UT1-UTC=-0.0564152, Accumulated Leap Seconds=33
  { 2454183.50, 65.241355 }, //24 March 2007, UT1-UTC=-0.0573549, Accumulated Leap Seconds=33
  { 2454184.50, 65.242114 }, //25 March 2007, UT1-UTC=-0.0581136, Accumulated Leap Seconds=33
  { 2454185.50, 65.242816 }, //26 March 2007, UT1-UTC=-0.0588164, Accumulated Leap Seconds=33
  { 2454186.50, 65.243543 }, //27 March 2007, UT1-UTC=-0.0595433, Accumulated Leap Seconds=33
  { 2454187.50, 65.244398 }, //28 March 2007, UT1-UTC=-0.0603980, Accumulated Leap Seconds=33
  { 2454188.50, 65.245450 }, //29 March 2007, UT1-UTC=-0.0614503, Accumulated Leap Seconds=33
  { 2454189.50, 65.246653 }, //30 March 2007, UT1-UTC=-0.0626531, Accumulated Leap Seconds=33
  { 2454190.50, 65.247969 }, //31 March 2007, UT1-UTC=-0.0639692, Accumulated Leap Seconds=33
  { 2454191.50, 65.249371 }, //01 April 2007, UT1-UTC=-0.0653713, Accumulated Leap Seconds=33
  { 2454192.50, 65.250889 }, //02 April 2007, UT1-UTC=-0.0668895, Accumulated Leap Seconds=33
  { 2454193.50, 65.252447 }, //03 April 2007, UT1-UTC=-0.0684467, Accumulated Leap Seconds=33
  { 2454194.50, 65.253974 }, //04 April 2007, UT1-UTC=-0.0699735, Accumulated Leap Seconds=33
  { 2454195.50, 65.255423 }, //05 April 2007, UT1-UTC=-0.0714227, Accumulated Leap Seconds=33
  { 2454196.50, 65.256772 }, //06 April 2007, UT1-UTC=-0.0727718, Accumulated Leap Seconds=33
  { 2454197.50, 65.258025 }, //07 April 2007, UT1-UTC=-0.0740254, Accumulated Leap Seconds=33
  { 2454198.50, 65.259253 }, //08 April 2007, UT1-UTC=-0.0752528, Accumulated Leap Seconds=33
  { 2454199.50, 65.260454 }, //09 April 2007, UT1-UTC=-0.0764542, Accumulated Leap Seconds=33
  { 2454200.50, 65.261642 }, //10 April 2007, UT1-UTC=-0.0776424, Accumulated Leap Seconds=33
  { 2454201.50, 65.262894 }, //11 April 2007, UT1-UTC=-0.0788935, Accumulated Leap Seconds=33
  { 2454202.50, 65.264270 }, //12 April 2007, UT1-UTC=-0.0802701, Accumulated Leap Seconds=33
  { 2454203.50, 65.265849 }, //13 April 2007, UT1-UTC=-0.0818486, Accumulated Leap Seconds=33
  { 2454204.50, 65.267750 }, //14 April 2007, UT1-UTC=-0.0837498, Accumulated Leap Seconds=33
  { 2454205.50, 65.269885 }, //15 April 2007, UT1-UTC=-0.0858846, Accumulated Leap Seconds=33
  { 2454206.50, 65.272167 }, //16 April 2007, UT1-UTC=-0.0881668, Accumulated Leap Seconds=33
  { 2454207.50, 65.274420 }, //17 April 2007, UT1-UTC=-0.0904205, Accumulated Leap Seconds=33
  { 2454208.50, 65.276426 }, //18 April 2007, UT1-UTC=-0.0924261, Accumulated Leap Seconds=33
  { 2454209.50, 65.278065 }, //19 April 2007, UT1-UTC=-0.0940647, Accumulated Leap Seconds=33
  { 2454210.50, 65.279312 }, //20 April 2007, UT1-UTC=-0.0953119, Accumulated Leap Seconds=33
  { 2454211.50, 65.280252 }, //21 April 2007, UT1-UTC=-0.0962523, Accumulated Leap Seconds=33
  { 2454212.50, 65.281059 }, //22 April 2007, UT1-UTC=-0.0970589, Accumulated Leap Seconds=33
  { 2454213.50, 65.281887 }, //23 April 2007, UT1-UTC=-0.0978874, Accumulated Leap Seconds=33
  { 2454214.50, 65.282805 }, //24 April 2007, UT1-UTC=-0.0988054, Accumulated Leap Seconds=33
  { 2454215.50, 65.283825 }, //25 April 2007, UT1-UTC=-0.0998247, Accumulated Leap Seconds=33
  { 2454216.50, 65.285005 }, //26 April 2007, UT1-UTC=-0.1010047, Accumulated Leap Seconds=33
  { 2454217.50, 65.286325 }, //27 April 2007, UT1-UTC=-0.1023246, Accumulated Leap Seconds=33
  { 2454218.50, 65.287770 }, //28 April 2007, UT1-UTC=-0.1037702, Accumulated Leap Seconds=33
  { 2454219.50, 65.289253 }, //29 April 2007, UT1-UTC=-0.1052528, Accumulated Leap Seconds=33
  { 2454220.50, 65.290703 }, //30 April 2007, UT1-UTC=-0.1067029, Accumulated Leap Seconds=33
  { 2454221.50, 65.292064 }, //01 May 2007, UT1-UTC=-0.1080645, Accumulated Leap Seconds=33
  { 2454222.50, 65.293293 }, //02 May 2007, UT1-UTC=-0.1092932, Accumulated Leap Seconds=33
  { 2454223.50, 65.294392 }, //03 May 2007, UT1-UTC=-0.1103915, Accumulated Leap Seconds=33
  { 2454224.50, 65.295338 }, //04 May 2007, UT1-UTC=-0.1113384, Accumulated Leap Seconds=33
  { 2454225.50, 65.296162 }, //05 May 2007, UT1-UTC=-0.1121618, Accumulated Leap Seconds=33
  { 2454226.50, 65.296877 }, //06 May 2007, UT1-UTC=-0.1128767, Accumulated Leap Seconds=33
  { 2454227.50, 65.297519 }, //07 May 2007, UT1-UTC=-0.1135195, Accumulated Leap Seconds=33
  { 2454228.50, 65.298218 }, //08 May 2007, UT1-UTC=-0.1142181, Accumulated Leap Seconds=33
  { 2454229.50, 65.299133 }, //09 May 2007, UT1-UTC=-0.1151328, Accumulated Leap Seconds=33
  { 2454230.50, 65.300318 }, //10 May 2007, UT1-UTC=-0.1163178, Accumulated Leap Seconds=33
  { 2454231.50, 65.301820 }, //11 May 2007, UT1-UTC=-0.1178205, Accumulated Leap Seconds=33
  { 2454232.50, 65.303668 }, //12 May 2007, UT1-UTC=-0.1196679, Accumulated Leap Seconds=33
  { 2454233.50, 65.305815 }, //13 May 2007, UT1-UTC=-0.1218147, Accumulated Leap Seconds=33
  { 2454234.50, 65.308026 }, //14 May 2007, UT1-UTC=-0.1240264, Accumulated Leap Seconds=33
  { 2454235.50, 65.310100 }, //15 May 2007, UT1-UTC=-0.1260998, Accumulated Leap Seconds=33
  { 2454236.50, 65.311850 }, //16 May 2007, UT1-UTC=-0.1278501, Accumulated Leap Seconds=33
  { 2454237.50, 65.313221 }, //17 May 2007, UT1-UTC=-0.1292208, Accumulated Leap Seconds=33
  { 2454238.50, 65.314320 }, //18 May 2007, UT1-UTC=-0.1303198, Accumulated Leap Seconds=33
  { 2454239.50, 65.315229 }, //19 May 2007, UT1-UTC=-0.1312294, Accumulated Leap Seconds=33
  { 2454240.50, 65.316102 }, //20 May 2007, UT1-UTC=-0.1321018, Accumulated Leap Seconds=33
  { 2454241.50, 65.317031 }, //21 May 2007, UT1-UTC=-0.1330305, Accumulated Leap Seconds=33
  { 2454242.50, 65.317947 }, //22 May 2007, UT1-UTC=-0.1339470, Accumulated Leap Seconds=33
  { 2454243.50, 65.318925 }, //23 May 2007, UT1-UTC=-0.1349255, Accumulated Leap Seconds=33
  { 2454244.50, 65.320048 }, //24 May 2007, UT1-UTC=-0.1360483, Accumulated Leap Seconds=33
  { 2454245.50, 65.321331 }, //25 May 2007, UT1-UTC=-0.1373306, Accumulated Leap Seconds=33
  { 2454246.50, 65.322634 }, //26 May 2007, UT1-UTC=-0.1386344, Accumulated Leap Seconds=33
  { 2454247.50, 65.323876 }, //27 May 2007, UT1-UTC=-0.1398765, Accumulated Leap Seconds=33
  { 2454248.50, 65.325012 }, //28 May 2007, UT1-UTC=-0.1410123, Accumulated Leap Seconds=33
  { 2454249.50, 65.326000 }, //29 May 2007, UT1-UTC=-0.1420000, Accumulated Leap Seconds=33
  { 2454250.50, 65.326822 }, //30 May 2007, UT1-UTC=-0.1428218, Accumulated Leap Seconds=33
  { 2454251.50, 65.327473 }, //31 May 2007, UT1-UTC=-0.1434728, Accumulated Leap Seconds=33
  { 2454252.50, 65.327940 }, //01 June 2007, UT1-UTC=-0.1439403, Accumulated Leap Seconds=33
  { 2454253.50, 65.328216 }, //02 June 2007, UT1-UTC=-0.1442162, Accumulated Leap Seconds=33
  { 2454254.50, 65.328377 }, //03 June 2007, UT1-UTC=-0.1443772, Accumulated Leap Seconds=33
  { 2454255.50, 65.328564 }, //04 June 2007, UT1-UTC=-0.1445635, Accumulated Leap Seconds=33
  { 2454256.50, 65.328929 }, //05 June 2007, UT1-UTC=-0.1449290, Accumulated Leap Seconds=33
  { 2454257.50, 65.329516 }, //06 June 2007, UT1-UTC=-0.1455156, Accumulated Leap Seconds=33
  { 2454258.50, 65.330287 }, //07 June 2007, UT1-UTC=-0.1462873, Accumulated Leap Seconds=33
  { 2454259.50, 65.331245 }, //08 June 2007, UT1-UTC=-0.1472454, Accumulated Leap Seconds=33
  { 2454260.50, 65.332321 }, //09 June 2007, UT1-UTC=-0.1483210, Accumulated Leap Seconds=33
  { 2454261.50, 65.333451 }, //10 June 2007, UT1-UTC=-0.1494508, Accumulated Leap Seconds=33
  { 2454262.50, 65.334549 }, //11 June 2007, UT1-UTC=-0.1505489, Accumulated Leap Seconds=33
  { 2454263.50, 65.335459 }, //12 June 2007, UT1-UTC=-0.1514586, Accumulated Leap Seconds=33
  { 2454264.50, 65.336053 }, //13 June 2007, UT1-UTC=-0.1520530, Accumulated Leap Seconds=33
  { 2454265.50, 65.336328 }, //14 June 2007, UT1-UTC=-0.1523285, Accumulated Leap Seconds=33
  { 2454266.50, 65.336367 }, //15 June 2007, UT1-UTC=-0.1523672, Accumulated Leap Seconds=33
  { 2454267.50, 65.336298 }, //16 June 2007, UT1-UTC=-0.1522976, Accumulated Leap Seconds=33
  { 2454268.50, 65.336304 }, //17 June 2007, UT1-UTC=-0.1523038, Accumulated Leap Seconds=33
  { 2454269.50, 65.336493 }, //18 June 2007, UT1-UTC=-0.1524927, Accumulated Leap Seconds=33
  { 2454270.50, 65.336866 }, //19 June 2007, UT1-UTC=-0.1528665, Accumulated Leap Seconds=33
  { 2454271.50, 65.337426 }, //20 June 2007, UT1-UTC=-0.1534259, Accumulated Leap Seconds=33
  { 2454272.50, 65.338131 }, //21 June 2007, UT1-UTC=-0.1541307, Accumulated Leap Seconds=33
  { 2454273.50, 65.338883 }, //22 June 2007, UT1-UTC=-0.1548830, Accumulated Leap Seconds=33
  { 2454274.50, 65.339591 }, //23 June 2007, UT1-UTC=-0.1555914, Accumulated Leap Seconds=33
  { 2454275.50, 65.340210 }, //24 June 2007, UT1-UTC=-0.1562105, Accumulated Leap Seconds=33
  { 2454276.50, 65.340728 }, //25 June 2007, UT1-UTC=-0.1567277, Accumulated Leap Seconds=33
  { 2454277.50, 65.341098 }, //26 June 2007, UT1-UTC=-0.1570984, Accumulated Leap Seconds=33
  { 2454278.50, 65.341330 }, //27 June 2007, UT1-UTC=-0.1573298, Accumulated Leap Seconds=33
  { 2454279.50, 65.341425 }, //28 June 2007, UT1-UTC=-0.1574245, Accumulated Leap Seconds=33
  { 2454280.50, 65.341416 }, //29 June 2007, UT1-UTC=-0.1574159, Accumulated Leap Seconds=33
  { 2454281.50, 65.341377 }, //30 June 2007, UT1-UTC=-0.1573773, Accumulated Leap Seconds=33
  { 2454282.50, 65.341337 }, //01 July 2007, UT1-UTC=-0.1573366, Accumulated Leap Seconds=33
  { 2454283.50, 65.341421 }, //02 July 2007, UT1-UTC=-0.1574206, Accumulated Leap Seconds=33
  { 2454284.50, 65.341734 }, //03 July 2007, UT1-UTC=-0.1577341, Accumulated Leap Seconds=33
  { 2454285.50, 65.342300 }, //04 July 2007, UT1-UTC=-0.1583003, Accumulated Leap Seconds=33
  { 2454286.50, 65.343068 }, //05 July 2007, UT1-UTC=-0.1590684, Accumulated Leap Seconds=33
  { 2454287.50, 65.343981 }, //06 July 2007, UT1-UTC=-0.1599805, Accumulated Leap Seconds=33
  { 2454288.50, 65.344938 }, //07 July 2007, UT1-UTC=-0.1609384, Accumulated Leap Seconds=33
  { 2454289.50, 65.345819 }, //08 July 2007, UT1-UTC=-0.1618187, Accumulated Leap Seconds=33
  { 2454290.50, 65.346453 }, //09 July 2007, UT1-UTC=-0.1624532, Accumulated Leap Seconds=33
  { 2454291.50, 65.346776 }, //10 July 2007, UT1-UTC=-0.1627757, Accumulated Leap Seconds=33
  { 2454292.50, 65.346857 }, //11 July 2007, UT1-UTC=-0.1628569, Accumulated Leap Seconds=33
  { 2454293.50, 65.346719 }, //12 July 2007, UT1-UTC=-0.1627189, Accumulated Leap Seconds=33
  { 2454294.50, 65.346471 }, //13 July 2007, UT1-UTC=-0.1624711, Accumulated Leap Seconds=33
  { 2454295.50, 65.346174 }, //14 July 2007, UT1-UTC=-0.1621739, Accumulated Leap Seconds=33
  { 2454296.50, 65.345996 }, //15 July 2007, UT1-UTC=-0.1619959, Accumulated Leap Seconds=33
  { 2454297.50, 65.345990 }, //16 July 2007, UT1-UTC=-0.1619895, Accumulated Leap Seconds=33
  { 2454298.50, 65.346156 }, //17 July 2007, UT1-UTC=-0.1621563, Accumulated Leap Seconds=33
  { 2454299.50, 65.346480 }, //18 July 2007, UT1-UTC=-0.1624805, Accumulated Leap Seconds=33
  { 2454300.50, 65.346882 }, //19 July 2007, UT1-UTC=-0.1628822, Accumulated Leap Seconds=33
  { 2454301.50, 65.347262 }, //20 July 2007, UT1-UTC=-0.1632619, Accumulated Leap Seconds=33
  { 2454302.50, 65.347555 }, //21 July 2007, UT1-UTC=-0.1635550, Accumulated Leap Seconds=33
  { 2454303.50, 65.347723 }, //22 July 2007, UT1-UTC=-0.1637231, Accumulated Leap Seconds=33
  { 2454304.50, 65.347713 }, //23 July 2007, UT1-UTC=-0.1637131, Accumulated Leap Seconds=33
  { 2454305.50, 65.347512 }, //24 July 2007, UT1-UTC=-0.1635122, Accumulated Leap Seconds=33
  { 2454306.50, 65.347126 }, //25 July 2007, UT1-UTC=-0.1631264, Accumulated Leap Seconds=33
  { 2454307.50, 65.346572 }, //26 July 2007, UT1-UTC=-0.1625716, Accumulated Leap Seconds=33
  { 2454308.50, 65.345951 }, //27 July 2007, UT1-UTC=-0.1619507, Accumulated Leap Seconds=33
  { 2454309.50, 65.345355 }, //28 July 2007, UT1-UTC=-0.1613552, Accumulated Leap Seconds=33
  { 2454310.50, 65.344926 }, //29 July 2007, UT1-UTC=-0.1609255, Accumulated Leap Seconds=33
  { 2454311.50, 65.344739 }, //30 July 2007, UT1-UTC=-0.1607387, Accumulated Leap Seconds=33
  { 2454312.50, 65.344824 }, //31 July 2007, UT1-UTC=-0.1608236, Accumulated Leap Seconds=33
  { 2454313.50, 65.345188 }, //01 August 2007, UT1-UTC=-0.1611881, Accumulated Leap Seconds=33
  { 2454314.50, 65.345814 }, //02 August 2007, UT1-UTC=-0.1618138, Accumulated Leap Seconds=33
  { 2454315.50, 65.346607 }, //03 August 2007, UT1-UTC=-0.1626073, Accumulated Leap Seconds=33
  { 2454316.50, 65.347363 }, //04 August 2007, UT1-UTC=-0.1633628, Accumulated Leap Seconds=33
  { 2454317.50, 65.347960 }, //05 August 2007, UT1-UTC=-0.1639595, Accumulated Leap Seconds=33
  { 2454318.50, 65.348304 }, //06 August 2007, UT1-UTC=-0.1643038, Accumulated Leap Seconds=33
  { 2454319.50, 65.348351 }, //07 August 2007, UT1-UTC=-0.1643512, Accumulated Leap Seconds=33
  { 2454320.50, 65.348138 }, //08 August 2007, UT1-UTC=-0.1641383, Accumulated Leap Seconds=33
  { 2454321.50, 65.347751 }, //09 August 2007, UT1-UTC=-0.1637508, Accumulated Leap Seconds=33
  { 2454322.50, 65.347288 }, //10 August 2007, UT1-UTC=-0.1632878, Accumulated Leap Seconds=33
  { 2454323.50, 65.346905 }, //11 August 2007, UT1-UTC=-0.1629055, Accumulated Leap Seconds=33
  { 2454324.50, 65.346727 }, //12 August 2007, UT1-UTC=-0.1627268, Accumulated Leap Seconds=33
  { 2454325.50, 65.346746 }, //13 August 2007, UT1-UTC=-0.1627462, Accumulated Leap Seconds=33
  { 2454326.50, 65.346892 }, //14 August 2007, UT1-UTC=-0.1628922, Accumulated Leap Seconds=33
  { 2454327.50, 65.347132 }, //15 August 2007, UT1-UTC=-0.1631324, Accumulated Leap Seconds=33
  { 2454328.50, 65.347408 }, //16 August 2007, UT1-UTC=-0.1634081, Accumulated Leap Seconds=33
  { 2454329.50, 65.347639 }, //17 August 2007, UT1-UTC=-0.1636393, Accumulated Leap Seconds=33
  { 2454330.50, 65.347856 }, //18 August 2007, UT1-UTC=-0.1638565, Accumulated Leap Seconds=33
  { 2454331.50, 65.347993 }, //19 August 2007, UT1-UTC=-0.1639926, Accumulated Leap Seconds=33
  { 2454332.50, 65.348001 }, //20 August 2007, UT1-UTC=-0.1640014, Accumulated Leap Seconds=33
  { 2454333.50, 65.347853 }, //21 August 2007, UT1-UTC=-0.1638526, Accumulated Leap Seconds=33
  { 2454334.50, 65.347545 }, //22 August 2007, UT1-UTC=-0.1635452, Accumulated Leap Seconds=33
  { 2454335.50, 65.347115 }, //23 August 2007, UT1-UTC=-0.1631149, Accumulated Leap Seconds=33
  { 2454336.50, 65.346611 }, //24 August 2007, UT1-UTC=-0.1626115, Accumulated Leap Seconds=33
  { 2454337.50, 65.346173 }, //25 August 2007, UT1-UTC=-0.1621729, Accumulated Leap Seconds=33
  { 2454338.50, 65.345902 }, //26 August 2007, UT1-UTC=-0.1619021, Accumulated Leap Seconds=33
  { 2454339.50, 65.345899 }, //27 August 2007, UT1-UTC=-0.1618994, Accumulated Leap Seconds=33
  { 2454340.50, 65.346205 }, //28 August 2007, UT1-UTC=-0.1622055, Accumulated Leap Seconds=33
  { 2454341.50, 65.346843 }, //29 August 2007, UT1-UTC=-0.1628426, Accumulated Leap Seconds=33
  { 2454342.50, 65.347711 }, //30 August 2007, UT1-UTC=-0.1637113, Accumulated Leap Seconds=33
  { 2454343.50, 65.348689 }, //31 August 2007, UT1-UTC=-0.1646890, Accumulated Leap Seconds=33
  { 2454344.50, 65.349640 }, //01 September 2007, UT1-UTC=-0.1656400, Accumulated Leap Seconds=33
  { 2454345.50, 65.350398 }, //02 September 2007, UT1-UTC=-0.1663977, Accumulated Leap Seconds=33
  { 2454346.50, 65.350887 }, //03 September 2007, UT1-UTC=-0.1668870, Accumulated Leap Seconds=33
  { 2454347.50, 65.351120 }, //04 September 2007, UT1-UTC=-0.1671203, Accumulated Leap Seconds=33
  { 2454348.50, 65.351184 }, //05 September 2007, UT1-UTC=-0.1671837, Accumulated Leap Seconds=33
  { 2454349.50, 65.351226 }, //06 September 2007, UT1-UTC=-0.1672262, Accumulated Leap Seconds=33
  { 2454350.50, 65.351353 }, //07 September 2007, UT1-UTC=-0.1673525, Accumulated Leap Seconds=33
  { 2454351.50, 65.351646 }, //08 September 2007, UT1-UTC=-0.1676457, Accumulated Leap Seconds=33
  { 2454352.50, 65.352150 }, //09 September 2007, UT1-UTC=-0.1681500, Accumulated Leap Seconds=33
  { 2454353.50, 65.352869 }, //10 September 2007, UT1-UTC=-0.1688687, Accumulated Leap Seconds=33
  { 2454354.50, 65.353698 }, //11 September 2007, UT1-UTC=-0.1696978, Accumulated Leap Seconds=33
  { 2454355.50, 65.354593 }, //12 September 2007, UT1-UTC=-0.1705927, Accumulated Leap Seconds=33
  { 2454356.50, 65.355555 }, //13 September 2007, UT1-UTC=-0.1715547, Accumulated Leap Seconds=33
  { 2454357.50, 65.356452 }, //14 September 2007, UT1-UTC=-0.1724516, Accumulated Leap Seconds=33
  { 2454358.50, 65.357230 }, //15 September 2007, UT1-UTC=-0.1732301, Accumulated Leap Seconds=33
  { 2454359.50, 65.357864 }, //16 September 2007, UT1-UTC=-0.1738645, Accumulated Leap Seconds=33
  { 2454360.50, 65.358348 }, //17 September 2007, UT1-UTC=-0.1743479, Accumulated Leap Seconds=33
  { 2454361.50, 65.358711 }, //18 September 2007, UT1-UTC=-0.1747108, Accumulated Leap Seconds=33
  { 2454362.50, 65.358961 }, //19 September 2007, UT1-UTC=-0.1749608, Accumulated Leap Seconds=33
  { 2454363.50, 65.359178 }, //20 September 2007, UT1-UTC=-0.1751775, Accumulated Leap Seconds=33
  { 2454364.50, 65.359478 }, //21 September 2007, UT1-UTC=-0.1754780, Accumulated Leap Seconds=33
  { 2454365.50, 65.359935 }, //22 September 2007, UT1-UTC=-0.1759346, Accumulated Leap Seconds=33
  { 2454366.50, 65.360599 }, //23 September 2007, UT1-UTC=-0.1765990, Accumulated Leap Seconds=33
  { 2454367.50, 65.361547 }, //24 September 2007, UT1-UTC=-0.1775465, Accumulated Leap Seconds=33
  { 2454368.50, 65.362791 }, //25 September 2007, UT1-UTC=-0.1787912, Accumulated Leap Seconds=33
  { 2454369.50, 65.364395 }, //26 September 2007, UT1-UTC=-0.1803951, Accumulated Leap Seconds=33
  { 2454370.50, 65.366119 }, //27 September 2007, UT1-UTC=-0.1821194, Accumulated Leap Seconds=33
  { 2454371.50, 65.367751 }, //28 September 2007, UT1-UTC=-0.1837513, Accumulated Leap Seconds=33
  { 2454372.50, 65.369171 }, //29 September 2007, UT1-UTC=-0.1851709, Accumulated Leap Seconds=33
  { 2454373.50, 65.370300 }, //30 September 2007, UT1-UTC=-0.1863000, Accumulated Leap Seconds=33
  { 2454374.50, 65.371131 }, //01 October 2007, UT1-UTC=-0.1871307, Accumulated Leap Seconds=33
  { 2454375.50, 65.371715 }, //02 October 2007, UT1-UTC=-0.1877154, Accumulated Leap Seconds=33
  { 2454376.50, 65.372223 }, //03 October 2007, UT1-UTC=-0.1882226, Accumulated Leap Seconds=33
  { 2454377.50, 65.372756 }, //04 October 2007, UT1-UTC=-0.1887562, Accumulated Leap Seconds=33
  { 2454378.50, 65.373371 }, //05 October 2007, UT1-UTC=-0.1893708, Accumulated Leap Seconds=33
  { 2454379.50, 65.374086 }, //06 October 2007, UT1-UTC=-0.1900862, Accumulated Leap Seconds=33
  { 2454380.50, 65.374949 }, //07 October 2007, UT1-UTC=-0.1909486, Accumulated Leap Seconds=33
  { 2454381.50, 65.375933 }, //08 October 2007, UT1-UTC=-0.1919326, Accumulated Leap Seconds=33
  { 2454382.50, 65.376978 }, //09 October 2007, UT1-UTC=-0.1929784, Accumulated Leap Seconds=33
  { 2454383.50, 65.378058 }, //10 October 2007, UT1-UTC=-0.1940585, Accumulated Leap Seconds=33
  { 2454384.50, 65.379158 }, //11 October 2007, UT1-UTC=-0.1951579, Accumulated Leap Seconds=33
  { 2454385.50, 65.380161 }, //12 October 2007, UT1-UTC=-0.1961614, Accumulated Leap Seconds=33
  { 2454386.50, 65.381024 }, //13 October 2007, UT1-UTC=-0.1970240, Accumulated Leap Seconds=33
  { 2454387.50, 65.381703 }, //14 October 2007, UT1-UTC=-0.1977025, Accumulated Leap Seconds=33
  { 2454388.50, 65.382223 }, //15 October 2007, UT1-UTC=-0.1982226, Accumulated Leap Seconds=33
  { 2454389.50, 65.382649 }, //16 October 2007, UT1-UTC=-0.1986487, Accumulated Leap Seconds=33
  { 2454390.50, 65.383032 }, //17 October 2007, UT1-UTC=-0.1990321, Accumulated Leap Seconds=33
  { 2454391.50, 65.383481 }, //18 October 2007, UT1-UTC=-0.1994807, Accumulated Leap Seconds=33
  { 2454392.50, 65.384036 }, //19 October 2007, UT1-UTC=-0.2000363, Accumulated Leap Seconds=33
  { 2454393.50, 65.384752 }, //20 October 2007, UT1-UTC=-0.2007521, Accumulated Leap Seconds=33
  { 2454394.50, 65.385681 }, //21 October 2007, UT1-UTC=-0.2016807, Accumulated Leap Seconds=33
  { 2454395.50, 65.386828 }, //22 October 2007, UT1-UTC=-0.2028284, Accumulated Leap Seconds=33
  { 2454396.50, 65.388218 }, //23 October 2007, UT1-UTC=-0.2042182, Accumulated Leap Seconds=33
  { 2454397.50, 65.389838 }, //24 October 2007, UT1-UTC=-0.2058379, Accumulated Leap Seconds=33
  { 2454398.50, 65.391530 }, //25 October 2007, UT1-UTC=-0.2075304, Accumulated Leap Seconds=33
  { 2454399.50, 65.393062 }, //26 October 2007, UT1-UTC=-0.2090621, Accumulated Leap Seconds=33
  { 2454400.50, 65.394252 }, //27 October 2007, UT1-UTC=-0.2102517, Accumulated Leap Seconds=33
  { 2454401.50, 65.395080 }, //28 October 2007, UT1-UTC=-0.2110800, Accumulated Leap Seconds=33
  { 2454402.50, 65.395662 }, //29 October 2007, UT1-UTC=-0.2116624, Accumulated Leap Seconds=33
  { 2454403.50, 65.396155 }, //30 October 2007, UT1-UTC=-0.2121553, Accumulated Leap Seconds=33
  { 2454404.50, 65.396643 }, //31 October 2007, UT1-UTC=-0.2126427, Accumulated Leap Seconds=33
  { 2454405.50, 65.397200 }, //01 November 2007, UT1-UTC=-0.2131998, Accumulated Leap Seconds=33
  { 2454406.50, 65.397886 }, //02 November 2007, UT1-UTC=-0.2138860, Accumulated Leap Seconds=33
  { 2454407.50, 65.398736 }, //03 November 2007, UT1-UTC=-0.2147359, Accumulated Leap Seconds=33
  { 2454408.50, 65.399722 }, //04 November 2007, UT1-UTC=-0.2157221, Accumulated Leap Seconds=33
  { 2454409.50, 65.400744 }, //05 November 2007, UT1-UTC=-0.2167438, Accumulated Leap Seconds=33
  { 2454410.50, 65.401789 }, //06 November 2007, UT1-UTC=-0.2177889, Accumulated Leap Seconds=33
  { 2454411.50, 65.402842 }, //07 November 2007, UT1-UTC=-0.2188419, Accumulated Leap Seconds=33
  { 2454412.50, 65.403891 }, //08 November 2007, UT1-UTC=-0.2198909, Accumulated Leap Seconds=33
  { 2454413.50, 65.404838 }, //09 November 2007, UT1-UTC=-0.2208385, Accumulated Leap Seconds=33
  { 2454414.50, 65.405599 }, //10 November 2007, UT1-UTC=-0.2215991, Accumulated Leap Seconds=33
  { 2454415.50, 65.406172 }, //11 November 2007, UT1-UTC=-0.2221723, Accumulated Leap Seconds=33
  { 2454416.50, 65.406586 }, //12 November 2007, UT1-UTC=-0.2225862, Accumulated Leap Seconds=33
  { 2454417.50, 65.406909 }, //13 November 2007, UT1-UTC=-0.2229088, Accumulated Leap Seconds=33
  { 2454418.50, 65.407255 }, //14 November 2007, UT1-UTC=-0.2232552, Accumulated Leap Seconds=33
  { 2454419.50, 65.407718 }, //15 November 2007, UT1-UTC=-0.2237181, Accumulated Leap Seconds=33
  { 2454420.50, 65.408438 }, //16 November 2007, UT1-UTC=-0.2244383, Accumulated Leap Seconds=33
  { 2454421.50, 65.409429 }, //17 November 2007, UT1-UTC=-0.2254288, Accumulated Leap Seconds=33
  { 2454422.50, 65.410698 }, //18 November 2007, UT1-UTC=-0.2266985, Accumulated Leap Seconds=33
  { 2454423.50, 65.412277 }, //19 November 2007, UT1-UTC=-0.2282772, Accumulated Leap Seconds=33
  { 2454424.50, 65.414114 }, //20 November 2007, UT1-UTC=-0.2301145, Accumulated Leap Seconds=33
  { 2454425.50, 65.416095 }, //21 November 2007, UT1-UTC=-0.2320951, Accumulated Leap Seconds=33
  { 2454426.50, 65.418095 }, //22 November 2007, UT1-UTC=-0.2340954, Accumulated Leap Seconds=33
  { 2454427.50, 65.419938 }, //23 November 2007, UT1-UTC=-0.2359377, Accumulated Leap Seconds=33
  { 2454428.50, 65.421505 }, //24 November 2007, UT1-UTC=-0.2375048, Accumulated Leap Seconds=33
  { 2454429.50, 65.422764 }, //25 November 2007, UT1-UTC=-0.2387638, Accumulated Leap Seconds=33
  { 2454430.50, 65.423777 }, //26 November 2007, UT1-UTC=-0.2397770, Accumulated Leap Seconds=33
  { 2454431.50, 65.424684 }, //27 November 2007, UT1-UTC=-0.2406843, Accumulated Leap Seconds=33
  { 2454432.50, 65.425644 }, //28 November 2007, UT1-UTC=-0.2416439, Accumulated Leap Seconds=33
  { 2454433.50, 65.426752 }, //29 November 2007, UT1-UTC=-0.2427516, Accumulated Leap Seconds=33
  { 2454434.50, 65.428060 }, //30 November 2007, UT1-UTC=-0.2440595, Accumulated Leap Seconds=33
  { 2454435.50, 65.429555 }, //01 December 2007, UT1-UTC=-0.2455547, Accumulated Leap Seconds=33
  { 2454436.50, 65.431134 }, //02 December 2007, UT1-UTC=-0.2471345, Accumulated Leap Seconds=33
  { 2454437.50, 65.432739 }, //03 December 2007, UT1-UTC=-0.2487391, Accumulated Leap Seconds=33
  { 2454438.50, 65.434252 }, //04 December 2007, UT1-UTC=-0.2502523, Accumulated Leap Seconds=33
  { 2454439.50, 65.435588 }, //05 December 2007, UT1-UTC=-0.2515877, Accumulated Leap Seconds=33
  { 2454440.50, 65.436672 }, //06 December 2007, UT1-UTC=-0.2526718, Accumulated Leap Seconds=33
  { 2454441.50, 65.437506 }, //07 December 2007, UT1-UTC=-0.2535059, Accumulated Leap Seconds=33
  { 2454442.50, 65.438088 }, //08 December 2007, UT1-UTC=-0.2540880, Accumulated Leap Seconds=33
  { 2454443.50, 65.438446 }, //09 December 2007, UT1-UTC=-0.2544462, Accumulated Leap Seconds=33
  { 2454444.50, 65.438655 }, //10 December 2007, UT1-UTC=-0.2546547, Accumulated Leap Seconds=33
  { 2454445.50, 65.438809 }, //11 December 2007, UT1-UTC=-0.2548093, Accumulated Leap Seconds=33
  { 2454446.50, 65.439034 }, //12 December 2007, UT1-UTC=-0.2550337, Accumulated Leap Seconds=33
  { 2454447.50, 65.439391 }, //13 December 2007, UT1-UTC=-0.2553910, Accumulated Leap Seconds=33
  { 2454448.50, 65.439908 }, //14 December 2007, UT1-UTC=-0.2559077, Accumulated Leap Seconds=33
  { 2454449.50, 65.440638 }, //15 December 2007, UT1-UTC=-0.2566382, Accumulated Leap Seconds=33
  { 2454450.50, 65.441625 }, //16 December 2007, UT1-UTC=-0.2576246, Accumulated Leap Seconds=33
  { 2454451.50, 65.442833 }, //17 December 2007, UT1-UTC=-0.2588329, Accumulated Leap Seconds=33
  { 2454452.50, 65.444155 }, //18 December 2007, UT1-UTC=-0.2601553, Accumulated Leap Seconds=33
  { 2454453.50, 65.445465 }, //19 December 2007, UT1-UTC=-0.2614654, Accumulated Leap Seconds=33
  { 2454454.50, 65.446731 }, //20 December 2007, UT1-UTC=-0.2627308, Accumulated Leap Seconds=33
  { 2454455.50, 65.447833 }, //21 December 2007, UT1-UTC=-0.2638328, Accumulated Leap Seconds=33
  { 2454456.50, 65.448614 }, //22 December 2007, UT1-UTC=-0.2646138, Accumulated Leap Seconds=33
  { 2454457.50, 65.449167 }, //23 December 2007, UT1-UTC=-0.2651672, Accumulated Leap Seconds=33
  { 2454458.50, 65.449641 }, //24 December 2007, UT1-UTC=-0.2656413, Accumulated Leap Seconds=33
  { 2454459.50, 65.450158 }, //25 December 2007, UT1-UTC=-0.2661582, Accumulated Leap Seconds=33
  { 2454460.50, 65.450806 }, //26 December 2007, UT1-UTC=-0.2668056, Accumulated Leap Seconds=33
  { 2454461.50, 65.451625 }, //27 December 2007, UT1-UTC=-0.2676248, Accumulated Leap Seconds=33
  { 2454462.50, 65.452620 }, //28 December 2007, UT1-UTC=-0.2686198, Accumulated Leap Seconds=33
  { 2454463.50, 65.453758 }, //29 December 2007, UT1-UTC=-0.2697583, Accumulated Leap Seconds=33
  { 2454464.50, 65.455023 }, //30 December 2007, UT1-UTC=-0.2710235, Accumulated Leap Seconds=33
  { 2454465.50, 65.456245 }, //31 December 2007, UT1-UTC=-0.2722454, Accumulated Leap Seconds=33
  { 2454466.50, 65.457349 }, //01 January 2008, UT1-UTC=-0.2733487, Accumulated Leap Seconds=33
  { 2454467.50, 65.458325 }, //02 January 2008, UT1-UTC=-0.2743246, Accumulated Leap Seconds=33
  { 2454468.50, 65.459192 }, //03 January 2008, UT1-UTC=-0.2751923, Accumulated Leap Seconds=33
  { 2454469.50, 65.459954 }, //04 January 2008, UT1-UTC=-0.2759541, Accumulated Leap Seconds=33
  { 2454470.50, 65.460611 }, //05 January 2008, UT1-UTC=-0.2766112, Accumulated Leap Seconds=33
  { 2454471.50, 65.461205 }, //06 January 2008, UT1-UTC=-0.2772046, Accumulated Leap Seconds=33
  { 2454472.50, 65.461744 }, //07 January 2008, UT1-UTC=-0.2777443, Accumulated Leap Seconds=33
  { 2454473.50, 65.462260 }, //08 January 2008, UT1-UTC=-0.2782602, Accumulated Leap Seconds=33
  { 2454474.50, 65.462847 }, //09 January 2008, UT1-UTC=-0.2788466, Accumulated Leap Seconds=33
  { 2454475.50, 65.463572 }, //10 January 2008, UT1-UTC=-0.2795717, Accumulated Leap Seconds=33
  { 2454476.50, 65.464459 }, //11 January 2008, UT1-UTC=-0.2804593, Accumulated Leap Seconds=33
  { 2454477.50, 65.465550 }, //12 January 2008, UT1-UTC=-0.2815502, Accumulated Leap Seconds=33
  { 2454478.50, 65.466827 }, //13 January 2008, UT1-UTC=-0.2828270, Accumulated Leap Seconds=33
  { 2454479.50, 65.468206 }, //14 January 2008, UT1-UTC=-0.2842063, Accumulated Leap Seconds=33
  { 2454480.50, 65.469592 }, //15 January 2008, UT1-UTC=-0.2855919, Accumulated Leap Seconds=33
  { 2454481.50, 65.470910 }, //16 January 2008, UT1-UTC=-0.2869096, Accumulated Leap Seconds=33
  { 2454482.50, 65.472001 }, //17 January 2008, UT1-UTC=-0.2880006, Accumulated Leap Seconds=33
  { 2454483.50, 65.472795 }, //18 January 2008, UT1-UTC=-0.2887946, Accumulated Leap Seconds=33
  { 2454484.50, 65.473356 }, //19 January 2008, UT1-UTC=-0.2893557, Accumulated Leap Seconds=33
  { 2454485.50, 65.473784 }, //20 January 2008, UT1-UTC=-0.2897838, Accumulated Leap Seconds=33
  { 2454486.50, 65.474214 }, //21 January 2008, UT1-UTC=-0.2902138, Accumulated Leap Seconds=33
  { 2454487.50, 65.474806 }, //22 January 2008, UT1-UTC=-0.2908058, Accumulated Leap Seconds=33
  { 2454488.50, 65.475645 }, //23 January 2008, UT1-UTC=-0.2916448, Accumulated Leap Seconds=33
  { 2454489.50, 65.476758 }, //24 January 2008, UT1-UTC=-0.2927578, Accumulated Leap Seconds=33
  { 2454490.50, 65.478091 }, //25 January 2008, UT1-UTC=-0.2940911, Accumulated Leap Seconds=33
  { 2454491.50, 65.479590 }, //26 January 2008, UT1-UTC=-0.2955900, Accumulated Leap Seconds=33
  { 2454492.50, 65.481124 }, //27 January 2008, UT1-UTC=-0.2971245, Accumulated Leap Seconds=33
  { 2454493.50, 65.482604 }, //28 January 2008, UT1-UTC=-0.2986041, Accumulated Leap Seconds=33
  { 2454494.50, 65.483962 }, //29 January 2008, UT1-UTC=-0.2999616, Accumulated Leap Seconds=33
  { 2454495.50, 65.485109 }, //30 January 2008, UT1-UTC=-0.3011086, Accumulated Leap Seconds=33
  { 2454496.50, 65.486020 }, //31 January 2008, UT1-UTC=-0.3020201, Accumulated Leap Seconds=33
  { 2454497.50, 65.486752 }, //01 February 2008, UT1-UTC=-0.3027520, Accumulated Leap Seconds=33
  { 2454498.50, 65.487392 }, //02 February 2008, UT1-UTC=-0.3033916, Accumulated Leap Seconds=33
  { 2454499.50, 65.487970 }, //03 February 2008, UT1-UTC=-0.3039696, Accumulated Leap Seconds=33
  { 2454500.50, 65.488530 }, //04 February 2008, UT1-UTC=-0.3045304, Accumulated Leap Seconds=33
  { 2454501.50, 65.489205 }, //05 February 2008, UT1-UTC=-0.3052051, Accumulated Leap Seconds=33
  { 2454502.50, 65.490069 }, //06 February 2008, UT1-UTC=-0.3060687, Accumulated Leap Seconds=33
  { 2454503.50, 65.491147 }, //07 February 2008, UT1-UTC=-0.3071474, Accumulated Leap Seconds=33
  { 2454504.50, 65.492466 }, //08 February 2008, UT1-UTC=-0.3084659, Accumulated Leap Seconds=33
  { 2454505.50, 65.493973 }, //09 February 2008, UT1-UTC=-0.3099733, Accumulated Leap Seconds=33
  { 2454506.50, 65.495569 }, //10 February 2008, UT1-UTC=-0.3115686, Accumulated Leap Seconds=33
  { 2454507.50, 65.497229 }, //11 February 2008, UT1-UTC=-0.3132294, Accumulated Leap Seconds=33
  { 2454508.50, 65.498821 }, //12 February 2008, UT1-UTC=-0.3148210, Accumulated Leap Seconds=33
  { 2454509.50, 65.500175 }, //13 February 2008, UT1-UTC=-0.3161747, Accumulated Leap Seconds=33
  { 2454510.50, 65.501323 }, //14 February 2008, UT1-UTC=-0.3173231, Accumulated Leap Seconds=33
  { 2454511.50, 65.502233 }, //15 February 2008, UT1-UTC=-0.3182327, Accumulated Leap Seconds=33
  { 2454512.50, 65.502919 }, //16 February 2008, UT1-UTC=-0.3189186, Accumulated Leap Seconds=33
  { 2454513.50, 65.503478 }, //17 February 2008, UT1-UTC=-0.3194775, Accumulated Leap Seconds=33
  { 2454514.50, 65.504078 }, //18 February 2008, UT1-UTC=-0.3200785, Accumulated Leap Seconds=33
  { 2454515.50, 65.504877 }, //19 February 2008, UT1-UTC=-0.3208770, Accumulated Leap Seconds=33
  { 2454516.50, 65.505925 }, //20 February 2008, UT1-UTC=-0.3219252, Accumulated Leap Seconds=33
  { 2454517.50, 65.507157 }, //21 February 2008, UT1-UTC=-0.3231568, Accumulated Leap Seconds=33
  { 2454518.50, 65.508496 }, //22 February 2008, UT1-UTC=-0.3244960, Accumulated Leap Seconds=33
  { 2454519.50, 65.509838 }, //23 February 2008, UT1-UTC=-0.3258381, Accumulated Leap Seconds=33
  { 2454520.50, 65.511112 }, //24 February 2008, UT1-UTC=-0.3271121, Accumulated Leap Seconds=33
  { 2454521.50, 65.512287 }, //25 February 2008, UT1-UTC=-0.3282865, Accumulated Leap Seconds=33
  { 2454522.50, 65.513282 }, //26 February 2008, UT1-UTC=-0.3292819, Accumulated Leap Seconds=33
  { 2454523.50, 65.514057 }, //27 February 2008, UT1-UTC=-0.3300575, Accumulated Leap Seconds=33
  { 2454524.50, 65.514609 }, //28 February 2008, UT1-UTC=-0.3306086, Accumulated Leap Seconds=33
  { 2454525.50, 65.514968 }, //29 February 2008, UT1-UTC=-0.3309683, Accumulated Leap Seconds=33
  { 2454526.50, 65.515201 }, //01 March 2008, UT1-UTC=-0.3312012, Accumulated Leap Seconds=33
  { 2454527.50, 65.515436 }, //02 March 2008, UT1-UTC=-0.3314358, Accumulated Leap Seconds=33
  { 2454528.50, 65.515782 }, //03 March 2008, UT1-UTC=-0.3317821, Accumulated Leap Seconds=33
  { 2454529.50, 65.516240 }, //04 March 2008, UT1-UTC=-0.3322401, Accumulated Leap Seconds=33
  { 2454530.50, 65.516968 }, //05 March 2008, UT1-UTC=-0.3329681, Accumulated Leap Seconds=33
  { 2454531.50, 65.517990 }, //06 March 2008, UT1-UTC=-0.3339902, Accumulated Leap Seconds=33
  { 2454532.50, 65.519268 }, //07 March 2008, UT1-UTC=-0.3352683, Accumulated Leap Seconds=33
  { 2454533.50, 65.520755 }, //08 March 2008, UT1-UTC=-0.3367549, Accumulated Leap Seconds=33
  { 2454534.50, 65.522335 }, //09 March 2008, UT1-UTC=-0.3383351, Accumulated Leap Seconds=33
  { 2454535.50, 65.523810 }, //10 March 2008, UT1-UTC=-0.3398101, Accumulated Leap Seconds=33
  { 2454536.50, 65.525083 }, //11 March 2008, UT1-UTC=-0.3410830, Accumulated Leap Seconds=33
  { 2454537.50, 65.526085 }, //12 March 2008, UT1-UTC=-0.3420846, Accumulated Leap Seconds=33
  { 2454538.50, 65.526812 }, //13 March 2008, UT1-UTC=-0.3428122, Accumulated Leap Seconds=33
  { 2454539.50, 65.527333 }, //14 March 2008, UT1-UTC=-0.3433331, Accumulated Leap Seconds=33
  { 2454540.50, 65.527770 }, //15 March 2008, UT1-UTC=-0.3437704, Accumulated Leap Seconds=33
  { 2454541.50, 65.528277 }, //16 March 2008, UT1-UTC=-0.3442772, Accumulated Leap Seconds=33
  { 2454542.50, 65.528937 }, //17 March 2008, UT1-UTC=-0.3449369, Accumulated Leap Seconds=33
  { 2454543.50, 65.529811 }, //18 March 2008, UT1-UTC=-0.3458107, Accumulated Leap Seconds=33
  { 2454544.50, 65.530964 }, //19 March 2008, UT1-UTC=-0.3469636, Accumulated Leap Seconds=33
  { 2454545.50, 65.532300 }, //20 March 2008, UT1-UTC=-0.3483000, Accumulated Leap Seconds=33
  { 2454546.50, 65.533718 }, //21 March 2008, UT1-UTC=-0.3497181, Accumulated Leap Seconds=33
  { 2454547.50, 65.535215 }, //22 March 2008, UT1-UTC=-0.3512150, Accumulated Leap Seconds=33
  { 2454548.50, 65.536666 }, //23 March 2008, UT1-UTC=-0.3526664, Accumulated Leap Seconds=33
  { 2454549.50, 65.538036 }, //24 March 2008, UT1-UTC=-0.3540362, Accumulated Leap Seconds=33
  { 2454550.50, 65.539276 }, //25 March 2008, UT1-UTC=-0.3552761, Accumulated Leap Seconds=33
  { 2454551.50, 65.540342 }, //26 March 2008, UT1-UTC=-0.3563422, Accumulated Leap Seconds=33
  { 2454552.50, 65.541226 }, //27 March 2008, UT1-UTC=-0.3572263, Accumulated Leap Seconds=33
  { 2454553.50, 65.541962 }, //28 March 2008, UT1-UTC=-0.3579623, Accumulated Leap Seconds=33
  { 2454554.50, 65.542599 }, //29 March 2008, UT1-UTC=-0.3585988, Accumulated Leap Seconds=33
  { 2454555.50, 65.543254 }, //30 March 2008, UT1-UTC=-0.3592544, Accumulated Leap Seconds=33
  { 2454556.50, 65.544024 }, //31 March 2008, UT1-UTC=-0.3600236, Accumulated Leap Seconds=33
  { 2454557.50, 65.544992 }, //01 April 2008, UT1-UTC=-0.3609916, Accumulated Leap Seconds=33
  { 2454558.50, 65.546178 }, //02 April 2008, UT1-UTC=-0.3621785, Accumulated Leap Seconds=33
  { 2454559.50, 65.547636 }, //03 April 2008, UT1-UTC=-0.3636364, Accumulated Leap Seconds=33
  { 2454560.50, 65.549282 }, //04 April 2008, UT1-UTC=-0.3652822, Accumulated Leap Seconds=33
  { 2454561.50, 65.551120 }, //05 April 2008, UT1-UTC=-0.3671200, Accumulated Leap Seconds=33
  { 2454562.50, 65.553045 }, //06 April 2008, UT1-UTC=-0.3690451, Accumulated Leap Seconds=33
  { 2454563.50, 65.554804 }, //07 April 2008, UT1-UTC=-0.3708038, Accumulated Leap Seconds=33
  { 2454564.50, 65.556212 }, //08 April 2008, UT1-UTC=-0.3722121, Accumulated Leap Seconds=33
  { 2454565.50, 65.557278 }, //09 April 2008, UT1-UTC=-0.3732783, Accumulated Leap Seconds=33
  { 2454566.50, 65.558045 }, //10 April 2008, UT1-UTC=-0.3740450, Accumulated Leap Seconds=33
  { 2454567.50, 65.558628 }, //11 April 2008, UT1-UTC=-0.3746282, Accumulated Leap Seconds=33
  { 2454568.50, 65.559168 }, //12 April 2008, UT1-UTC=-0.3751680, Accumulated Leap Seconds=33
  { 2454569.50, 65.559775 }, //13 April 2008, UT1-UTC=-0.3757750, Accumulated Leap Seconds=33
  { 2454570.50, 65.560580 }, //14 April 2008, UT1-UTC=-0.3765804, Accumulated Leap Seconds=33
  { 2454571.50, 65.561649 }, //15 April 2008, UT1-UTC=-0.3776493, Accumulated Leap Seconds=33
  { 2454572.50, 65.562911 }, //16 April 2008, UT1-UTC=-0.3789112, Accumulated Leap Seconds=33
  { 2454573.50, 65.564253 }, //17 April 2008, UT1-UTC=-0.3802534, Accumulated Leap Seconds=33
  { 2454574.50, 65.565609 }, //18 April 2008, UT1-UTC=-0.3816095, Accumulated Leap Seconds=33
  { 2454575.50, 65.566934 }, //19 April 2008, UT1-UTC=-0.3829337, Accumulated Leap Seconds=33
  { 2454576.50, 65.568167 }, //20 April 2008, UT1-UTC=-0.3841666, Accumulated Leap Seconds=33
  { 2454577.50, 65.569235 }, //21 April 2008, UT1-UTC=-0.3852347, Accumulated Leap Seconds=33
  { 2454578.50, 65.570156 }, //22 April 2008, UT1-UTC=-0.3861557, Accumulated Leap Seconds=33
  { 2454579.50, 65.570927 }, //23 April 2008, UT1-UTC=-0.3869274, Accumulated Leap Seconds=33
  { 2454580.50, 65.571595 }, //24 April 2008, UT1-UTC=-0.3875950, Accumulated Leap Seconds=33
  { 2454581.50, 65.572199 }, //25 April 2008, UT1-UTC=-0.3881989, Accumulated Leap Seconds=33
  { 2454582.50, 65.572769 }, //26 April 2008, UT1-UTC=-0.3887692, Accumulated Leap Seconds=33
  { 2454583.50, 65.573394 }, //27 April 2008, UT1-UTC=-0.3893942, Accumulated Leap Seconds=33
  { 2454584.50, 65.574199 }, //28 April 2008, UT1-UTC=-0.3901987, Accumulated Leap Seconds=33
  { 2454585.50, 65.575215 }, //29 April 2008, UT1-UTC=-0.3912154, Accumulated Leap Seconds=33
  { 2454586.50, 65.576511 }, //30 April 2008, UT1-UTC=-0.3925115, Accumulated Leap Seconds=33
  { 2454587.50, 65.578077 }, //01 May 2008, UT1-UTC=-0.3940768, Accumulated Leap Seconds=33
  { 2454588.50, 65.579780 }, //02 May 2008, UT1-UTC=-0.3957801, Accumulated Leap Seconds=33
  { 2454589.50, 65.581614 }, //03 May 2008, UT1-UTC=-0.3976139, Accumulated Leap Seconds=33
  { 2454590.50, 65.583463 }, //04 May 2008, UT1-UTC=-0.3994632, Accumulated Leap Seconds=33
  { 2454591.50, 65.585147 }, //05 May 2008, UT1-UTC=-0.4011471, Accumulated Leap Seconds=33
  { 2454592.50, 65.586571 }, //06 May 2008, UT1-UTC=-0.4025710, Accumulated Leap Seconds=33
  { 2454593.50, 65.587662 }, //07 May 2008, UT1-UTC=-0.4036620, Accumulated Leap Seconds=33
  { 2454594.50, 65.588556 }, //08 May 2008, UT1-UTC=-0.4045560, Accumulated Leap Seconds=33
  { 2454595.50, 65.589414 }, //09 May 2008, UT1-UTC=-0.4054143, Accumulated Leap Seconds=33
  { 2454596.50, 65.590353 }, //10 May 2008, UT1-UTC=-0.4063530, Accumulated Leap Seconds=33
  { 2454597.50, 65.591479 }, //11 May 2008, UT1-UTC=-0.4074787, Accumulated Leap Seconds=33
  { 2454598.50, 65.592878 }, //12 May 2008, UT1-UTC=-0.4088775, Accumulated Leap Seconds=33
  { 2454599.50, 65.594467 }, //13 May 2008, UT1-UTC=-0.4104669, Accumulated Leap Seconds=33
  { 2454600.50, 65.596139 }, //14 May 2008, UT1-UTC=-0.4121390, Accumulated Leap Seconds=33
  { 2454601.50, 65.597811 }, //15 May 2008, UT1-UTC=-0.4138108, Accumulated Leap Seconds=33
  { 2454602.50, 65.599407 }, //16 May 2008, UT1-UTC=-0.4154074, Accumulated Leap Seconds=33
  { 2454603.50, 65.600828 }, //17 May 2008, UT1-UTC=-0.4168283, Accumulated Leap Seconds=33
  { 2454604.50, 65.602013 }, //18 May 2008, UT1-UTC=-0.4180126, Accumulated Leap Seconds=33
  { 2454605.50, 65.603016 }, //19 May 2008, UT1-UTC=-0.4190159, Accumulated Leap Seconds=33
  { 2454606.50, 65.603810 }, //20 May 2008, UT1-UTC=-0.4198097, Accumulated Leap Seconds=33
  { 2454607.50, 65.604363 }, //21 May 2008, UT1-UTC=-0.4203632, Accumulated Leap Seconds=33
  { 2454608.50, 65.604738 }, //22 May 2008, UT1-UTC=-0.4207380, Accumulated Leap Seconds=33
  { 2454609.50, 65.605030 }, //23 May 2008, UT1-UTC=-0.4210295, Accumulated Leap Seconds=33
  { 2454610.50, 65.605325 }, //24 May 2008, UT1-UTC=-0.4213254, Accumulated Leap Seconds=33
  { 2454611.50, 65.605721 }, //25 May 2008, UT1-UTC=-0.4217205, Accumulated Leap Seconds=33
  { 2454612.50, 65.606294 }, //26 May 2008, UT1-UTC=-0.4222936, Accumulated Leap Seconds=33
  { 2454613.50, 65.607071 }, //27 May 2008, UT1-UTC=-0.4230707, Accumulated Leap Seconds=33
  { 2454614.50, 65.608022 }, //28 May 2008, UT1-UTC=-0.4240223, Accumulated Leap Seconds=33
  { 2454615.50, 65.609129 }, //29 May 2008, UT1-UTC=-0.4251295, Accumulated Leap Seconds=33
  { 2454616.50, 65.610350 }, //30 May 2008, UT1-UTC=-0.4263505, Accumulated Leap Seconds=33
  { 2454617.50, 65.611594 }, //31 May 2008, UT1-UTC=-0.4275942, Accumulated Leap Seconds=33
  { 2454618.50, 65.612728 }, //01 June 2008, UT1-UTC=-0.4287280, Accumulated Leap Seconds=33
  { 2454619.50, 65.613673 }, //02 June 2008, UT1-UTC=-0.4296730, Accumulated Leap Seconds=33
  { 2454620.50, 65.614332 }, //03 June 2008, UT1-UTC=-0.4303320, Accumulated Leap Seconds=33
  { 2454621.50, 65.614767 }, //04 June 2008, UT1-UTC=-0.4307670, Accumulated Leap Seconds=33
  { 2454622.50, 65.615090 }, //05 June 2008, UT1-UTC=-0.4310902, Accumulated Leap Seconds=33
  { 2454623.50, 65.615440 }, //06 June 2008, UT1-UTC=-0.4314399, Accumulated Leap Seconds=33
  { 2454624.50, 65.615902 }, //07 June 2008, UT1-UTC=-0.4319020, Accumulated Leap Seconds=33
  { 2454625.50, 65.616529 }, //08 June 2008, UT1-UTC=-0.4325288, Accumulated Leap Seconds=33
  { 2454626.50, 65.617315 }, //09 June 2008, UT1-UTC=-0.4333150, Accumulated Leap Seconds=33
  { 2454627.50, 65.618258 }, //10 June 2008, UT1-UTC=-0.4342579, Accumulated Leap Seconds=33
  { 2454628.50, 65.619266 }, //11 June 2008, UT1-UTC=-0.4352663, Accumulated Leap Seconds=33
  { 2454629.50, 65.620183 }, //12 June 2008, UT1-UTC=-0.4361835, Accumulated Leap Seconds=33
  { 2454630.50, 65.620931 }, //13 June 2008, UT1-UTC=-0.4369308, Accumulated Leap Seconds=33
  { 2454631.50, 65.621492 }, //14 June 2008, UT1-UTC=-0.4374919, Accumulated Leap Seconds=33
  { 2454632.50, 65.621821 }, //15 June 2008, UT1-UTC=-0.4378213, Accumulated Leap Seconds=33
  { 2454633.50, 65.621989 }, //16 June 2008, UT1-UTC=-0.4379886, Accumulated Leap Seconds=33
  { 2454634.50, 65.622060 }, //17 June 2008, UT1-UTC=-0.4380601, Accumulated Leap Seconds=33
  { 2454635.50, 65.622026 }, //18 June 2008, UT1-UTC=-0.4380263, Accumulated Leap Seconds=33
  { 2454636.50, 65.621951 }, //19 June 2008, UT1-UTC=-0.4379508, Accumulated Leap Seconds=33
  { 2454637.50, 65.621902 }, //20 June 2008, UT1-UTC=-0.4379019, Accumulated Leap Seconds=33
  { 2454638.50, 65.621909 }, //21 June 2008, UT1-UTC=-0.4379087, Accumulated Leap Seconds=33
  { 2454639.50, 65.622096 }, //22 June 2008, UT1-UTC=-0.4380958, Accumulated Leap Seconds=33
  { 2454640.50, 65.622520 }, //23 June 2008, UT1-UTC=-0.4385196, Accumulated Leap Seconds=33
  { 2454641.50, 65.623194 }, //24 June 2008, UT1-UTC=-0.4391943, Accumulated Leap Seconds=33
  { 2454642.50, 65.624063 }, //25 June 2008, UT1-UTC=-0.4400628, Accumulated Leap Seconds=33
  { 2454643.50, 65.625033 }, //26 June 2008, UT1-UTC=-0.4410332, Accumulated Leap Seconds=33
  { 2454644.50, 65.626062 }, //27 June 2008, UT1-UTC=-0.4420623, Accumulated Leap Seconds=33
  { 2454645.50, 65.626981 }, //28 June 2008, UT1-UTC=-0.4429814, Accumulated Leap Seconds=33
  { 2454646.50, 65.627770 }, //29 June 2008, UT1-UTC=-0.4437703, Accumulated Leap Seconds=33
  { 2454647.50, 65.628380 }, //30 June 2008, UT1-UTC=-0.4443796, Accumulated Leap Seconds=33
  { 2454648.50, 65.628750 }, //01 July 2008, UT1-UTC=-0.4447505, Accumulated Leap Seconds=33
  { 2454649.50, 65.628936 }, //02 July 2008, UT1-UTC=-0.4449365, Accumulated Leap Seconds=33
  { 2454650.50, 65.629080 }, //03 July 2008, UT1-UTC=-0.4450798, Accumulated Leap Seconds=33
  { 2454651.50, 65.629356 }, //04 July 2008, UT1-UTC=-0.4453557, Accumulated Leap Seconds=33
  { 2454652.50, 65.629826 }, //05 July 2008, UT1-UTC=-0.4458260, Accumulated Leap Seconds=33
  { 2454653.50, 65.630466 }, //06 July 2008, UT1-UTC=-0.4464665, Accumulated Leap Seconds=33
  { 2454654.50, 65.631246 }, //07 July 2008, UT1-UTC=-0.4472457, Accumulated Leap Seconds=33
  { 2454655.50, 65.632063 }, //08 July 2008, UT1-UTC=-0.4480625, Accumulated Leap Seconds=33
  { 2454656.50, 65.632747 }, //09 July 2008, UT1-UTC=-0.4487466, Accumulated Leap Seconds=33
  { 2454657.50, 65.633324 }, //10 July 2008, UT1-UTC=-0.4493235, Accumulated Leap Seconds=33
  { 2454658.50, 65.633727 }, //11 July 2008, UT1-UTC=-0.4497275, Accumulated Leap Seconds=33
  { 2454659.50, 65.633868 }, //12 July 2008, UT1-UTC=-0.4498679, Accumulated Leap Seconds=33
  { 2454660.50, 65.633781 }, //13 July 2008, UT1-UTC=-0.4497806, Accumulated Leap Seconds=33
  { 2454661.50, 65.633538 }, //14 July 2008, UT1-UTC=-0.4495377, Accumulated Leap Seconds=33
  { 2454662.50, 65.633185 }, //15 July 2008, UT1-UTC=-0.4491848, Accumulated Leap Seconds=33
  { 2454663.50, 65.632772 }, //16 July 2008, UT1-UTC=-0.4487718, Accumulated Leap Seconds=33
  { 2454664.50, 65.632403 }, //17 July 2008, UT1-UTC=-0.4484031, Accumulated Leap Seconds=33
  { 2454665.50, 65.632161 }, //18 July 2008, UT1-UTC=-0.4481609, Accumulated Leap Seconds=33
  { 2454666.50, 65.632067 }, //19 July 2008, UT1-UTC=-0.4480668, Accumulated Leap Seconds=33
  { 2454667.50, 65.632194 }, //20 July 2008, UT1-UTC=-0.4481945, Accumulated Leap Seconds=33
  { 2454668.50, 65.632612 }, //21 July 2008, UT1-UTC=-0.4486120, Accumulated Leap Seconds=33
  { 2454669.50, 65.633227 }, //22 July 2008, UT1-UTC=-0.4492268, Accumulated Leap Seconds=33
  { 2454670.50, 65.633946 }, //23 July 2008, UT1-UTC=-0.4499465, Accumulated Leap Seconds=33
  { 2454671.50, 65.634717 }, //24 July 2008, UT1-UTC=-0.4507169, Accumulated Leap Seconds=33
  { 2454672.50, 65.635469 }, //25 July 2008, UT1-UTC=-0.4514691, Accumulated Leap Seconds=33
  { 2454673.50, 65.636101 }, //26 July 2008, UT1-UTC=-0.4521011, Accumulated Leap Seconds=33
  { 2454674.50, 65.636516 }, //27 July 2008, UT1-UTC=-0.4525161, Accumulated Leap Seconds=33
  { 2454675.50, 65.636714 }, //28 July 2008, UT1-UTC=-0.4527138, Accumulated Leap Seconds=33
  { 2454676.50, 65.636769 }, //29 July 2008, UT1-UTC=-0.4527693, Accumulated Leap Seconds=33
  { 2454677.50, 65.636734 }, //30 July 2008, UT1-UTC=-0.4527343, Accumulated Leap Seconds=33
  { 2454678.50, 65.636792 }, //31 July 2008, UT1-UTC=-0.4527922, Accumulated Leap Seconds=33
  { 2454679.50, 65.637009 }, //01 August 2008, UT1-UTC=-0.4530091, Accumulated Leap Seconds=33
  { 2454680.50, 65.637439 }, //02 August 2008, UT1-UTC=-0.4534390, Accumulated Leap Seconds=33
  { 2454681.50, 65.638051 }, //03 August 2008, UT1-UTC=-0.4540511, Accumulated Leap Seconds=33
  { 2454682.50, 65.638839 }, //04 August 2008, UT1-UTC=-0.4548394, Accumulated Leap Seconds=33
  { 2454683.50, 65.639626 }, //05 August 2008, UT1-UTC=-0.4556265, Accumulated Leap Seconds=33
  { 2454684.50, 65.640292 }, //06 August 2008, UT1-UTC=-0.4562919, Accumulated Leap Seconds=33
  { 2454685.50, 65.640804 }, //07 August 2008, UT1-UTC=-0.4568038, Accumulated Leap Seconds=33
  { 2454686.50, 65.641136 }, //08 August 2008, UT1-UTC=-0.4571358, Accumulated Leap Seconds=33
  { 2454687.50, 65.641279 }, //09 August 2008, UT1-UTC=-0.4572793, Accumulated Leap Seconds=33
  { 2454688.50, 65.641273 }, //10 August 2008, UT1-UTC=-0.4572728, Accumulated Leap Seconds=33
  { 2454689.50, 65.641143 }, //11 August 2008, UT1-UTC=-0.4571428, Accumulated Leap Seconds=33
  { 2454690.50, 65.640931 }, //12 August 2008, UT1-UTC=-0.4569308, Accumulated Leap Seconds=33
  { 2454691.50, 65.640718 }, //13 August 2008, UT1-UTC=-0.4567183, Accumulated Leap Seconds=33
  { 2454692.50, 65.640597 }, //14 August 2008, UT1-UTC=-0.4565975, Accumulated Leap Seconds=33
  { 2454693.50, 65.640612 }, //15 August 2008, UT1-UTC=-0.4566119, Accumulated Leap Seconds=33
  { 2454694.50, 65.640800 }, //16 August 2008, UT1-UTC=-0.4568004, Accumulated Leap Seconds=33
  { 2454695.50, 65.641147 }, //17 August 2008, UT1-UTC=-0.4571473, Accumulated Leap Seconds=33
  { 2454696.50, 65.641712 }, //18 August 2008, UT1-UTC=-0.4577121, Accumulated Leap Seconds=33
  { 2454697.50, 65.642462 }, //19 August 2008, UT1-UTC=-0.4584618, Accumulated Leap Seconds=33
  { 2454698.50, 65.643322 }, //20 August 2008, UT1-UTC=-0.4593217, Accumulated Leap Seconds=33
  { 2454699.50, 65.644135 }, //21 August 2008, UT1-UTC=-0.4601349, Accumulated Leap Seconds=33
  { 2454700.50, 65.644780 }, //22 August 2008, UT1-UTC=-0.4607799, Accumulated Leap Seconds=33
  { 2454701.50, 65.645245 }, //23 August 2008, UT1-UTC=-0.4612450, Accumulated Leap Seconds=33
  { 2454702.50, 65.645491 }, //24 August 2008, UT1-UTC=-0.4614906, Accumulated Leap Seconds=33
  { 2454703.50, 65.645579 }, //25 August 2008, UT1-UTC=-0.4615794, Accumulated Leap Seconds=33
  { 2454704.50, 65.645635 }, //26 August 2008, UT1-UTC=-0.4616352, Accumulated Leap Seconds=33
  { 2454705.50, 65.645766 }, //27 August 2008, UT1-UTC=-0.4617661, Accumulated Leap Seconds=33
  { 2454706.50, 65.646055 }, //28 August 2008, UT1-UTC=-0.4620553, Accumulated Leap Seconds=33
  { 2454707.50, 65.646540 }, //29 August 2008, UT1-UTC=-0.4625399, Accumulated Leap Seconds=33
  { 2454708.50, 65.647315 }, //30 August 2008, UT1-UTC=-0.4633152, Accumulated Leap Seconds=33
  { 2454709.50, 65.648280 }, //31 August 2008, UT1-UTC=-0.4642801, Accumulated Leap Seconds=33
  { 2454710.50, 65.649338 }, //01 September 2008, UT1-UTC=-0.4653375, Accumulated Leap Seconds=33
  { 2454711.50, 65.650393 }, //02 September 2008, UT1-UTC=-0.4663928, Accumulated Leap Seconds=33
  { 2454712.50, 65.651280 }, //03 September 2008, UT1-UTC=-0.4672797, Accumulated Leap Seconds=33
  { 2454713.50, 65.651959 }, //04 September 2008, UT1-UTC=-0.4679589, Accumulated Leap Seconds=33
  { 2454714.50, 65.652480 }, //05 September 2008, UT1-UTC=-0.4684798, Accumulated Leap Seconds=33
  { 2454715.50, 65.652781 }, //06 September 2008, UT1-UTC=-0.4687813, Accumulated Leap Seconds=33
  { 2454716.50, 65.652961 }, //07 September 2008, UT1-UTC=-0.4689610, Accumulated Leap Seconds=33
  { 2454717.50, 65.653082 }, //08 September 2008, UT1-UTC=-0.4690819, Accumulated Leap Seconds=33
  { 2454718.50, 65.653201 }, //09 September 2008, UT1-UTC=-0.4692015, Accumulated Leap Seconds=33
  { 2454719.50, 65.653446 }, //10 September 2008, UT1-UTC=-0.4694463, Accumulated Leap Seconds=33
  { 2454720.50, 65.653848 }, //11 September 2008, UT1-UTC=-0.4698483, Accumulated Leap Seconds=33
  { 2454721.50, 65.654478 }, //12 September 2008, UT1-UTC=-0.4704784, Accumulated Leap Seconds=33
  { 2454722.50, 65.655409 }, //13 September 2008, UT1-UTC=-0.4714086, Accumulated Leap Seconds=33
  { 2454723.50, 65.656632 }, //14 September 2008, UT1-UTC=-0.4726317, Accumulated Leap Seconds=33
  { 2454724.50, 65.658095 }, //15 September 2008, UT1-UTC=-0.4740946, Accumulated Leap Seconds=33
  { 2454725.50, 65.659709 }, //16 September 2008, UT1-UTC=-0.4757093, Accumulated Leap Seconds=33
  { 2454726.50, 65.661329 }, //17 September 2008, UT1-UTC=-0.4773286, Accumulated Leap Seconds=33
  { 2454727.50, 65.662799 }, //18 September 2008, UT1-UTC=-0.4787992, Accumulated Leap Seconds=33
  { 2454728.50, 65.664021 }, //19 September 2008, UT1-UTC=-0.4800210, Accumulated Leap Seconds=33
  { 2454729.50, 65.664986 }, //20 September 2008, UT1-UTC=-0.4809865, Accumulated Leap Seconds=33
  { 2454730.50, 65.665685 }, //21 September 2008, UT1-UTC=-0.4816851, Accumulated Leap Seconds=33
  { 2454731.50, 65.666262 }, //22 September 2008, UT1-UTC=-0.4822619, Accumulated Leap Seconds=33
  { 2454732.50, 65.666856 }, //23 September 2008, UT1-UTC=-0.4828561, Accumulated Leap Seconds=33
  { 2454733.50, 65.667503 }, //24 September 2008, UT1-UTC=-0.4835035, Accumulated Leap Seconds=33
  { 2454734.50, 65.668302 }, //25 September 2008, UT1-UTC=-0.4843018, Accumulated Leap Seconds=33
  { 2454735.50, 65.669365 }, //26 September 2008, UT1-UTC=-0.4853648, Accumulated Leap Seconds=33
  { 2454736.50, 65.670614 }, //27 September 2008, UT1-UTC=-0.4866135, Accumulated Leap Seconds=33
  { 2454737.50, 65.671984 }, //28 September 2008, UT1-UTC=-0.4879836, Accumulated Leap Seconds=33
  { 2454738.50, 65.673419 }, //29 September 2008, UT1-UTC=-0.4894192, Accumulated Leap Seconds=33
  { 2454739.50, 65.674784 }, //30 September 2008, UT1-UTC=-0.4907835, Accumulated Leap Seconds=33
  { 2454740.50, 65.675993 }, //01 October 2008, UT1-UTC=-0.4919928, Accumulated Leap Seconds=33
  { 2454741.50, 65.676992 }, //02 October 2008, UT1-UTC=-0.4929921, Accumulated Leap Seconds=33
  { 2454742.50, 65.677745 }, //03 October 2008, UT1-UTC=-0.4937449, Accumulated Leap Seconds=33
  { 2454743.50, 65.678285 }, //04 October 2008, UT1-UTC=-0.4942848, Accumulated Leap Seconds=33
  { 2454744.50, 65.678678 }, //05 October 2008, UT1-UTC=-0.4946776, Accumulated Leap Seconds=33
  { 2454745.50, 65.679025 }, //06 October 2008, UT1-UTC=-0.4950251, Accumulated Leap Seconds=33
  { 2454746.50, 65.679358 }, //07 October 2008, UT1-UTC=-0.4953584, Accumulated Leap Seconds=33
  { 2454747.50, 65.679836 }, //08 October 2008, UT1-UTC=-0.4958358, Accumulated Leap Seconds=33
  { 2454748.50, 65.680510 }, //09 October 2008, UT1-UTC=-0.4965102, Accumulated Leap Seconds=33
  { 2454749.50, 65.681413 }, //10 October 2008, UT1-UTC=-0.4974134, Accumulated Leap Seconds=33
  { 2454750.50, 65.682549 }, //11 October 2008, UT1-UTC=-0.4985488, Accumulated Leap Seconds=33
  { 2454751.50, 65.683912 }, //12 October 2008, UT1-UTC=-0.4999116, Accumulated Leap Seconds=33
  { 2454752.50, 65.685435 }, //13 October 2008, UT1-UTC=-0.5014348, Accumulated Leap Seconds=33
  { 2454753.50, 65.687033 }, //14 October 2008, UT1-UTC=-0.5030326, Accumulated Leap Seconds=33
  { 2454754.50, 65.688535 }, //15 October 2008, UT1-UTC=-0.5045348, Accumulated Leap Seconds=33
  { 2454755.50, 65.689923 }, //16 October 2008, UT1-UTC=-0.5059233, Accumulated Leap Seconds=33
  { 2454756.50, 65.691136 }, //17 October 2008, UT1-UTC=-0.5071357, Accumulated Leap Seconds=33
  { 2454757.50, 65.692155 }, //18 October 2008, UT1-UTC=-0.5081548, Accumulated Leap Seconds=33
  { 2454758.50, 65.693054 }, //19 October 2008, UT1-UTC=-0.5090541, Accumulated Leap Seconds=33
  { 2454759.50, 65.694026 }, //20 October 2008, UT1-UTC=-0.5100264, Accumulated Leap Seconds=33
  { 2454760.50, 65.695112 }, //21 October 2008, UT1-UTC=-0.5111122, Accumulated Leap Seconds=33
  { 2454761.50, 65.696328 }, //22 October 2008, UT1-UTC=-0.5123276, Accumulated Leap Seconds=33
  { 2454762.50, 65.697729 }, //23 October 2008, UT1-UTC=-0.5137295, Accumulated Leap Seconds=33
  { 2454763.50, 65.699322 }, //24 October 2008, UT1-UTC=-0.5153218, Accumulated Leap Seconds=33
  { 2454764.50, 65.700978 }, //25 October 2008, UT1-UTC=-0.5169776, Accumulated Leap Seconds=33
  { 2454765.50, 65.702620 }, //26 October 2008, UT1-UTC=-0.5186202, Accumulated Leap Seconds=33
  { 2454766.50, 65.704212 }, //27 October 2008, UT1-UTC=-0.5202121, Accumulated Leap Seconds=33
  { 2454767.50, 65.705696 }, //28 October 2008, UT1-UTC=-0.5216963, Accumulated Leap Seconds=33
  { 2454768.50, 65.706999 }, //29 October 2008, UT1-UTC=-0.5229987, Accumulated Leap Seconds=33
  { 2454769.50, 65.708071 }, //30 October 2008, UT1-UTC=-0.5240711, Accumulated Leap Seconds=33
  { 2454770.50, 65.708951 }, //31 October 2008, UT1-UTC=-0.5249506, Accumulated Leap Seconds=33
  { 2454771.50, 65.709697 }, //01 November 2008, UT1-UTC=-0.5256966, Accumulated Leap Seconds=33
  { 2454772.50, 65.710309 }, //02 November 2008, UT1-UTC=-0.5263089, Accumulated Leap Seconds=33
  { 2454773.50, 65.710927 }, //03 November 2008, UT1-UTC=-0.5269267, Accumulated Leap Seconds=33
  { 2454774.50, 65.711591 }, //04 November 2008, UT1-UTC=-0.5275908, Accumulated Leap Seconds=33
  { 2454775.50, 65.712302 }, //05 November 2008, UT1-UTC=-0.5283016, Accumulated Leap Seconds=33
  { 2454776.50, 65.713143 }, //06 November 2008, UT1-UTC=-0.5291431, Accumulated Leap Seconds=33
  { 2454777.50, 65.714182 }, //07 November 2008, UT1-UTC=-0.5301816, Accumulated Leap Seconds=33
  { 2454778.50, 65.715430 }, //08 November 2008, UT1-UTC=-0.5314301, Accumulated Leap Seconds=33
  { 2454779.50, 65.716882 }, //09 November 2008, UT1-UTC=-0.5328820, Accumulated Leap Seconds=33
  { 2454780.50, 65.718505 }, //10 November 2008, UT1-UTC=-0.5345047, Accumulated Leap Seconds=33
  { 2454781.50, 65.720180 }, //11 November 2008, UT1-UTC=-0.5361795, Accumulated Leap Seconds=33
  { 2454782.50, 65.721797 }, //12 November 2008, UT1-UTC=-0.5377971, Accumulated Leap Seconds=33
  { 2454783.50, 65.723194 }, //13 November 2008, UT1-UTC=-0.5391942, Accumulated Leap Seconds=33
  { 2454784.50, 65.724310 }, //14 November 2008, UT1-UTC=-0.5403097, Accumulated Leap Seconds=33
  { 2454785.50, 65.725169 }, //15 November 2008, UT1-UTC=-0.5411691, Accumulated Leap Seconds=33
  { 2454786.50, 65.725924 }, //16 November 2008, UT1-UTC=-0.5419239, Accumulated Leap Seconds=33
  { 2454787.50, 65.726770 }, //17 November 2008, UT1-UTC=-0.5427698, Accumulated Leap Seconds=33
  { 2454788.50, 65.727856 }, //18 November 2008, UT1-UTC=-0.5438561, Accumulated Leap Seconds=33
  { 2454789.50, 65.729269 }, //19 November 2008, UT1-UTC=-0.5452686, Accumulated Leap Seconds=33
  { 2454790.50, 65.730932 }, //20 November 2008, UT1-UTC=-0.5469323, Accumulated Leap Seconds=33
  { 2454791.50, 65.732762 }, //21 November 2008, UT1-UTC=-0.5487621, Accumulated Leap Seconds=33
  { 2454792.50, 65.734643 }, //22 November 2008, UT1-UTC=-0.5506431, Accumulated Leap Seconds=33
  { 2454793.50, 65.736470 }, //23 November 2008, UT1-UTC=-0.5524701, Accumulated Leap Seconds=33
  { 2454794.50, 65.738167 }, //24 November 2008, UT1-UTC=-0.5541669, Accumulated Leap Seconds=33
  { 2454795.50, 65.739704 }, //25 November 2008, UT1-UTC=-0.5557042, Accumulated Leap Seconds=33
  { 2454796.50, 65.741070 }, //26 November 2008, UT1-UTC=-0.5570702, Accumulated Leap Seconds=33
  { 2454797.50, 65.742314 }, //27 November 2008, UT1-UTC=-0.5583143, Accumulated Leap Seconds=33
  { 2454798.50, 65.743388 }, //28 November 2008, UT1-UTC=-0.5593880, Accumulated Leap Seconds=33
  { 2454799.50, 65.744333 }, //29 November 2008, UT1-UTC=-0.5603332, Accumulated Leap Seconds=33
  { 2454800.50, 65.745201 }, //30 November 2008, UT1-UTC=-0.5612014, Accumulated Leap Seconds=33
  { 2454801.50, 65.746092 }, //01 December 2008, UT1-UTC=-0.5620920, Accumulated Leap Seconds=33
  { 2454802.50, 65.747064 }, //02 December 2008, UT1-UTC=-0.5630643, Accumulated Leap Seconds=33
  { 2454803.50, 65.748118 }, //03 December 2008, UT1-UTC=-0.5641177, Accumulated Leap Seconds=33
  { 2454804.50, 65.749290 }, //04 December 2008, UT1-UTC=-0.5652902, Accumulated Leap Seconds=33
  { 2454805.50, 65.750647 }, //05 December 2008, UT1-UTC=-0.5666473, Accumulated Leap Seconds=33
  { 2454806.50, 65.752187 }, //06 December 2008, UT1-UTC=-0.5681873, Accumulated Leap Seconds=33
  { 2454807.50, 65.753830 }, //07 December 2008, UT1-UTC=-0.5698301, Accumulated Leap Seconds=33
  { 2454808.50, 65.755519 }, //08 December 2008, UT1-UTC=-0.5715190, Accumulated Leap Seconds=33
  { 2454809.50, 65.757121 }, //09 December 2008, UT1-UTC=-0.5731210, Accumulated Leap Seconds=33
  { 2454810.50, 65.758494 }, //10 December 2008, UT1-UTC=-0.5744942, Accumulated Leap Seconds=33
  { 2454811.50, 65.759614 }, //11 December 2008, UT1-UTC=-0.5756138, Accumulated Leap Seconds=33
  { 2454812.50, 65.760487 }, //12 December 2008, UT1-UTC=-0.5764868, Accumulated Leap Seconds=33
  { 2454813.50, 65.761155 }, //13 December 2008, UT1-UTC=-0.5771552, Accumulated Leap Seconds=33
  { 2454814.50, 65.761862 }, //14 December 2008, UT1-UTC=-0.5778624, Accumulated Leap Seconds=33
  { 2454815.50, 65.762761 }, //15 December 2008, UT1-UTC=-0.5787611, Accumulated Leap Seconds=33
  { 2454816.50, 65.763919 }, //16 December 2008, UT1-UTC=-0.5799185, Accumulated Leap Seconds=33
  { 2454817.50, 65.765251 }, //17 December 2008, UT1-UTC=-0.5812507, Accumulated Leap Seconds=33
  { 2454818.50, 65.766668 }, //18 December 2008, UT1-UTC=-0.5826679, Accumulated Leap Seconds=33
  { 2454819.50, 65.768124 }, //19 December 2008, UT1-UTC=-0.5841241, Accumulated Leap Seconds=33
  { 2454820.50, 65.769514 }, //20 December 2008, UT1-UTC=-0.5855142, Accumulated Leap Seconds=33
  { 2454821.50, 65.770736 }, //21 December 2008, UT1-UTC=-0.5867361, Accumulated Leap Seconds=33
  { 2454822.50, 65.771753 }, //22 December 2008, UT1-UTC=-0.5877529, Accumulated Leap Seconds=33
  { 2454823.50, 65.772583 }, //23 December 2008, UT1-UTC=-0.5885829, Accumulated Leap Seconds=33
  { 2454824.50, 65.773212 }, //24 December 2008, UT1-UTC=-0.5892121, Accumulated Leap Seconds=33
  { 2454825.50, 65.773620 }, //25 December 2008, UT1-UTC=-0.5896202, Accumulated Leap Seconds=33
  { 2454826.50, 65.773870 }, //26 December 2008, UT1-UTC=-0.5898703, Accumulated Leap Seconds=33
  { 2454827.50, 65.774062 }, //27 December 2008, UT1-UTC=-0.5900625, Accumulated Leap Seconds=33
  { 2454828.50, 65.774317 }, //28 December 2008, UT1-UTC=-0.5903168, Accumulated Leap Seconds=33
  { 2454829.50, 65.774666 }, //29 December 2008, UT1-UTC=-0.5906664, Accumulated Leap Seconds=33
  { 2454830.50, 65.775150 }, //30 December 2008, UT1-UTC=-0.5911498, Accumulated Leap Seconds=33
  { 2454831.50, 65.775869 }, //31 December 2008, UT1-UTC=-0.5918692, Accumulated Leap Seconds=33
  { 2454832.50, 65.776836 }, //01 January 2009, UT1-UTC= 0.4071638, Accumulated Leap Seconds=34
  { 2454833.50, 65.778016 }, //02 January 2009, UT1-UTC= 0.4059839, Accumulated Leap Seconds=34
  { 2454834.50, 65.779322 }, //03 January 2009, UT1-UTC= 0.4046780, Accumulated Leap Seconds=34
  { 2454835.50, 65.780648 }, //04 January 2009, UT1-UTC= 0.4033522, Accumulated Leap Seconds=34
  { 2454836.50, 65.781926 }, //05 January 2009, UT1-UTC= 0.4020741, Accumulated Leap Seconds=34
  { 2454837.50, 65.783090 }, //06 January 2009, UT1-UTC= 0.4009105, Accumulated Leap Seconds=34
  { 2454838.50, 65.784054 }, //07 January 2009, UT1-UTC= 0.3999456, Accumulated Leap Seconds=34
  { 2454839.50, 65.784773 }, //08 January 2009, UT1-UTC= 0.3992267, Accumulated Leap Seconds=34
  { 2454840.50, 65.785311 }, //09 January 2009, UT1-UTC= 0.3986887, Accumulated Leap Seconds=34
  { 2454841.50, 65.785869 }, //10 January 2009, UT1-UTC= 0.3981312, Accumulated Leap Seconds=34
  { 2454842.50, 65.786543 }, //11 January 2009, UT1-UTC= 0.3974573, Accumulated Leap Seconds=34
  { 2454843.50, 65.787490 }, //12 January 2009, UT1-UTC= 0.3965104, Accumulated Leap Seconds=34
  { 2454844.50, 65.788717 }, //13 January 2009, UT1-UTC= 0.3952828, Accumulated Leap Seconds=34
  { 2454845.50, 65.790148 }, //14 January 2009, UT1-UTC= 0.3938521, Accumulated Leap Seconds=34
  { 2454846.50, 65.791657 }, //15 January 2009, UT1-UTC= 0.3923430, Accumulated Leap Seconds=34
  { 2454847.50, 65.793110 }, //16 January 2009, UT1-UTC= 0.3908897, Accumulated Leap Seconds=34
  { 2454848.50, 65.794383 }, //17 January 2009, UT1-UTC= 0.3896172, Accumulated Leap Seconds=34
  { 2454849.50, 65.795446 }, //18 January 2009, UT1-UTC= 0.3885536, Accumulated Leap Seconds=34
  { 2454850.50, 65.796236 }, //19 January 2009, UT1-UTC= 0.3877636, Accumulated Leap Seconds=34
  { 2454851.50, 65.796748 }, //20 January 2009, UT1-UTC= 0.3872518, Accumulated Leap Seconds=34
  { 2454852.50, 65.797026 }, //21 January 2009, UT1-UTC= 0.3869740, Accumulated Leap Seconds=34
  { 2454853.50, 65.797158 }, //22 January 2009, UT1-UTC= 0.3868424, Accumulated Leap Seconds=34
  { 2454854.50, 65.797213 }, //23 January 2009, UT1-UTC= 0.3867870, Accumulated Leap Seconds=34
  { 2454855.50, 65.797259 }, //24 January 2009, UT1-UTC= 0.3867405, Accumulated Leap Seconds=34
  { 2454856.50, 65.797382 }, //25 January 2009, UT1-UTC= 0.3866180, Accumulated Leap Seconds=34
  { 2454857.50, 65.797671 }, //26 January 2009, UT1-UTC= 0.3863294, Accumulated Leap Seconds=34
  { 2454858.50, 65.798130 }, //27 January 2009, UT1-UTC= 0.3858698, Accumulated Leap Seconds=34
  { 2454859.50, 65.798801 }, //28 January 2009, UT1-UTC= 0.3851985, Accumulated Leap Seconds=34
  { 2454860.50, 65.799632 }, //29 January 2009, UT1-UTC= 0.3843680, Accumulated Leap Seconds=34
  { 2454861.50, 65.800566 }, //30 January 2009, UT1-UTC= 0.3834340, Accumulated Leap Seconds=34
  { 2454862.50, 65.801549 }, //31 January 2009, UT1-UTC= 0.3824508, Accumulated Leap Seconds=34
  { 2454863.50, 65.802461 }, //01 February 2009, UT1-UTC= 0.3815386, Accumulated Leap Seconds=34
  { 2454864.50, 65.803289 }, //02 February 2009, UT1-UTC= 0.3807110, Accumulated Leap Seconds=34
  { 2454865.50, 65.803957 }, //03 February 2009, UT1-UTC= 0.3800428, Accumulated Leap Seconds=34
  { 2454866.50, 65.804490 }, //04 February 2009, UT1-UTC= 0.3795104, Accumulated Leap Seconds=34
  { 2454867.50, 65.804925 }, //05 February 2009, UT1-UTC= 0.3790747, Accumulated Leap Seconds=34
  { 2454868.50, 65.805370 }, //06 February 2009, UT1-UTC= 0.3786300, Accumulated Leap Seconds=34
  { 2454869.50, 65.805909 }, //07 February 2009, UT1-UTC= 0.3780910, Accumulated Leap Seconds=34
  { 2454870.50, 65.806670 }, //08 February 2009, UT1-UTC= 0.3773302, Accumulated Leap Seconds=34
  { 2454871.50, 65.807694 }, //09 February 2009, UT1-UTC= 0.3763064, Accumulated Leap Seconds=34
  { 2454872.50, 65.808978 }, //10 February 2009, UT1-UTC= 0.3750224, Accumulated Leap Seconds=34
  { 2454873.50, 65.810410 }, //11 February 2009, UT1-UTC= 0.3735904, Accumulated Leap Seconds=34
  { 2454874.50, 65.811844 }, //12 February 2009, UT1-UTC= 0.3721556, Accumulated Leap Seconds=34
  { 2454875.50, 65.813116 }, //13 February 2009, UT1-UTC= 0.3708836, Accumulated Leap Seconds=34
  { 2454876.50, 65.814077 }, //14 February 2009, UT1-UTC= 0.3699233, Accumulated Leap Seconds=34
  { 2454877.50, 65.814789 }, //15 February 2009, UT1-UTC= 0.3692112, Accumulated Leap Seconds=34
  { 2454878.50, 65.815322 }, //16 February 2009, UT1-UTC= 0.3686782, Accumulated Leap Seconds=34
  { 2454879.50, 65.815696 }, //17 February 2009, UT1-UTC= 0.3683044, Accumulated Leap Seconds=34
  { 2454880.50, 65.815935 }, //18 February 2009, UT1-UTC= 0.3680654, Accumulated Leap Seconds=34
  { 2454881.50, 65.816082 }, //19 February 2009, UT1-UTC= 0.3679182, Accumulated Leap Seconds=34
  { 2454882.50, 65.816225 }, //20 February 2009, UT1-UTC= 0.3677752, Accumulated Leap Seconds=34
  { 2454883.50, 65.816401 }, //21 February 2009, UT1-UTC= 0.3675990, Accumulated Leap Seconds=34
  { 2454884.50, 65.816665 }, //22 February 2009, UT1-UTC= 0.3673346, Accumulated Leap Seconds=34
  { 2454885.50, 65.817138 }, //23 February 2009, UT1-UTC= 0.3668617, Accumulated Leap Seconds=34
  { 2454886.50, 65.817855 }, //24 February 2009, UT1-UTC= 0.3661446, Accumulated Leap Seconds=34
  { 2454887.50, 65.818787 }, //25 February 2009, UT1-UTC= 0.3652133, Accumulated Leap Seconds=34
  { 2454888.50, 65.819909 }, //26 February 2009, UT1-UTC= 0.3640914, Accumulated Leap Seconds=34
  { 2454889.50, 65.821164 }, //27 February 2009, UT1-UTC= 0.3628361, Accumulated Leap Seconds=34
  { 2454890.50, 65.822441 }, //28 February 2009, UT1-UTC= 0.3615594, Accumulated Leap Seconds=34
  { 2454891.50, 65.823669 }, //01 March 2009, UT1-UTC= 0.3603305, Accumulated Leap Seconds=34
  { 2454892.50, 65.824804 }, //02 March 2009, UT1-UTC= 0.3591963, Accumulated Leap Seconds=34
  { 2454893.50, 65.825806 }, //03 March 2009, UT1-UTC= 0.3581938, Accumulated Leap Seconds=34
  { 2454894.50, 65.826694 }, //04 March 2009, UT1-UTC= 0.3573063, Accumulated Leap Seconds=34
  { 2454895.50, 65.827460 }, //05 March 2009, UT1-UTC= 0.3565404, Accumulated Leap Seconds=34
  { 2454896.50, 65.828220 }, //06 March 2009, UT1-UTC= 0.3557802, Accumulated Leap Seconds=34
  { 2454897.50, 65.829122 }, //07 March 2009, UT1-UTC= 0.3548776, Accumulated Leap Seconds=34
  { 2454898.50, 65.830230 }, //08 March 2009, UT1-UTC= 0.3537698, Accumulated Leap Seconds=34
  { 2454899.50, 65.831611 }, //09 March 2009, UT1-UTC= 0.3523894, Accumulated Leap Seconds=34
  { 2454900.50, 65.833241 }, //10 March 2009, UT1-UTC= 0.3507589, Accumulated Leap Seconds=34
  { 2454901.50, 65.834984 }, //11 March 2009, UT1-UTC= 0.3490158, Accumulated Leap Seconds=34
  { 2454902.50, 65.836713 }, //12 March 2009, UT1-UTC= 0.3472866, Accumulated Leap Seconds=34
  { 2454903.50, 65.838299 }, //13 March 2009, UT1-UTC= 0.3457012, Accumulated Leap Seconds=34
  { 2454904.50, 65.839667 }, //14 March 2009, UT1-UTC= 0.3443332, Accumulated Leap Seconds=34
  { 2454905.50, 65.840803 }, //15 March 2009, UT1-UTC= 0.3431967, Accumulated Leap Seconds=34
  { 2454906.50, 65.841723 }, //16 March 2009, UT1-UTC= 0.3422766, Accumulated Leap Seconds=34
  { 2454907.50, 65.842458 }, //17 March 2009, UT1-UTC= 0.3415421, Accumulated Leap Seconds=34
  { 2454908.50, 65.843073 }, //18 March 2009, UT1-UTC= 0.3409269, Accumulated Leap Seconds=34
  { 2454909.50, 65.843652 }, //19 March 2009, UT1-UTC= 0.3403477, Accumulated Leap Seconds=34
  { 2454910.50, 65.844291 }, //20 March 2009, UT1-UTC= 0.3397094, Accumulated Leap Seconds=34
  { 2454911.50, 65.845067 }, //21 March 2009, UT1-UTC= 0.3389330, Accumulated Leap Seconds=34
  { 2454912.50, 65.846078 }, //22 March 2009, UT1-UTC= 0.3379222, Accumulated Leap Seconds=34
  { 2454913.50, 65.847323 }, //23 March 2009, UT1-UTC= 0.3366767, Accumulated Leap Seconds=34
  { 2454914.50, 65.848777 }, //24 March 2009, UT1-UTC= 0.3352230, Accumulated Leap Seconds=34
  { 2454915.50, 65.850380 }, //25 March 2009, UT1-UTC= 0.3336195, Accumulated Leap Seconds=34
  { 2454916.50, 65.852041 }, //26 March 2009, UT1-UTC= 0.3319592, Accumulated Leap Seconds=34
  { 2454917.50, 65.853701 }, //27 March 2009, UT1-UTC= 0.3302986, Accumulated Leap Seconds=34
  { 2454918.50, 65.855278 }, //28 March 2009, UT1-UTC= 0.3287224, Accumulated Leap Seconds=34
  { 2454919.50, 65.856655 }, //29 March 2009, UT1-UTC= 0.3273451, Accumulated Leap Seconds=34
  { 2454920.50, 65.857787 }, //30 March 2009, UT1-UTC= 0.3262132, Accumulated Leap Seconds=34
  { 2454921.50, 65.858708 }, //31 March 2009, UT1-UTC= 0.3252924, Accumulated Leap Seconds=34
  { 2454922.50, 65.859504 }, //01 April 2009, UT1-UTC= 0.3244964, Accumulated Leap Seconds=34
  { 2454923.50, 65.860291 }, //02 April 2009, UT1-UTC= 0.3237088, Accumulated Leap Seconds=34
  { 2454924.50, 65.861166 }, //03 April 2009, UT1-UTC= 0.3228337, Accumulated Leap Seconds=34
  { 2454925.50, 65.862188 }, //04 April 2009, UT1-UTC= 0.3218122, Accumulated Leap Seconds=34
  { 2454926.50, 65.863386 }, //05 April 2009, UT1-UTC= 0.3206140, Accumulated Leap Seconds=34
  { 2454927.50, 65.864831 }, //06 April 2009, UT1-UTC= 0.3191685, Accumulated Leap Seconds=34
  { 2454928.50, 65.866504 }, //07 April 2009, UT1-UTC= 0.3174956, Accumulated Leap Seconds=34
  { 2454929.50, 65.868283 }, //08 April 2009, UT1-UTC= 0.3157168, Accumulated Leap Seconds=34
  { 2454930.50, 65.870019 }, //09 April 2009, UT1-UTC= 0.3139813, Accumulated Leap Seconds=34
  { 2454931.50, 65.871593 }, //10 April 2009, UT1-UTC= 0.3124073, Accumulated Leap Seconds=34
  { 2454932.50, 65.872977 }, //11 April 2009, UT1-UTC= 0.3110229, Accumulated Leap Seconds=34
  { 2454933.50, 65.874139 }, //12 April 2009, UT1-UTC= 0.3098608, Accumulated Leap Seconds=34
  { 2454934.50, 65.875067 }, //13 April 2009, UT1-UTC= 0.3089327, Accumulated Leap Seconds=34
  { 2454935.50, 65.875834 }, //14 April 2009, UT1-UTC= 0.3081661, Accumulated Leap Seconds=34
  { 2454936.50, 65.876562 }, //15 April 2009, UT1-UTC= 0.3074379, Accumulated Leap Seconds=34
  { 2454937.50, 65.877347 }, //16 April 2009, UT1-UTC= 0.3066535, Accumulated Leap Seconds=34
  { 2454938.50, 65.878248 }, //17 April 2009, UT1-UTC= 0.3057521, Accumulated Leap Seconds=34
  { 2454939.50, 65.879272 }, //18 April 2009, UT1-UTC= 0.3047280, Accumulated Leap Seconds=34
  { 2454940.50, 65.880453 }, //19 April 2009, UT1-UTC= 0.3035474, Accumulated Leap Seconds=34
  { 2454941.50, 65.881821 }, //20 April 2009, UT1-UTC= 0.3021794, Accumulated Leap Seconds=34
  { 2454942.50, 65.883409 }, //21 April 2009, UT1-UTC= 0.3005907, Accumulated Leap Seconds=34
  { 2454943.50, 65.885169 }, //22 April 2009, UT1-UTC= 0.2988310, Accumulated Leap Seconds=34
  { 2454944.50, 65.886987 }, //23 April 2009, UT1-UTC= 0.2970129, Accumulated Leap Seconds=34
  { 2454945.50, 65.888743 }, //24 April 2009, UT1-UTC= 0.2952567, Accumulated Leap Seconds=34
  { 2454946.50, 65.890351 }, //25 April 2009, UT1-UTC= 0.2936485, Accumulated Leap Seconds=34
  { 2454947.50, 65.891770 }, //26 April 2009, UT1-UTC= 0.2922298, Accumulated Leap Seconds=34
  { 2454948.50, 65.892977 }, //27 April 2009, UT1-UTC= 0.2910235, Accumulated Leap Seconds=34
  { 2454949.50, 65.894028 }, //28 April 2009, UT1-UTC= 0.2899716, Accumulated Leap Seconds=34
  { 2454950.50, 65.895030 }, //29 April 2009, UT1-UTC= 0.2889698, Accumulated Leap Seconds=34
  { 2454951.50, 65.896097 }, //30 April 2009, UT1-UTC= 0.2879032, Accumulated Leap Seconds=34
  { 2454952.50, 65.897301 }, //01 May 2009, UT1-UTC= 0.2866992, Accumulated Leap Seconds=34
  { 2454953.50, 65.898704 }, //02 May 2009, UT1-UTC= 0.2852964, Accumulated Leap Seconds=34
  { 2454954.50, 65.900315 }, //03 May 2009, UT1-UTC= 0.2836850, Accumulated Leap Seconds=34
  { 2454955.50, 65.902074 }, //04 May 2009, UT1-UTC= 0.2819259, Accumulated Leap Seconds=34
  { 2454956.50, 65.903865 }, //05 May 2009, UT1-UTC= 0.2801345, Accumulated Leap Seconds=34
  { 2454957.50, 65.905568 }, //06 May 2009, UT1-UTC= 0.2784317, Accumulated Leap Seconds=34
  { 2454958.50, 65.907099 }, //07 May 2009, UT1-UTC= 0.2769010, Accumulated Leap Seconds=34
  { 2454959.50, 65.908354 }, //08 May 2009, UT1-UTC= 0.2756458, Accumulated Leap Seconds=34
  { 2454960.50, 65.909324 }, //09 May 2009, UT1-UTC= 0.2746759, Accumulated Leap Seconds=34
  { 2454961.50, 65.910078 }, //10 May 2009, UT1-UTC= 0.2739222, Accumulated Leap Seconds=34
  { 2454962.50, 65.910687 }, //11 May 2009, UT1-UTC= 0.2733129, Accumulated Leap Seconds=34
  { 2454963.50, 65.911227 }, //12 May 2009, UT1-UTC= 0.2727733, Accumulated Leap Seconds=34
  { 2454964.50, 65.911760 }, //13 May 2009, UT1-UTC= 0.2722396, Accumulated Leap Seconds=34
  { 2454965.50, 65.912348 }, //14 May 2009, UT1-UTC= 0.2716517, Accumulated Leap Seconds=34
  { 2454966.50, 65.913058 }, //15 May 2009, UT1-UTC= 0.2709424, Accumulated Leap Seconds=34
  { 2454967.50, 65.913910 }, //16 May 2009, UT1-UTC= 0.2700901, Accumulated Leap Seconds=34
  { 2454968.50, 65.914893 }, //17 May 2009, UT1-UTC= 0.2691068, Accumulated Leap Seconds=34
  { 2454969.50, 65.916010 }, //18 May 2009, UT1-UTC= 0.2679896, Accumulated Leap Seconds=34
  { 2454970.50, 65.917300 }, //19 May 2009, UT1-UTC= 0.2666999, Accumulated Leap Seconds=34
  { 2454971.50, 65.918738 }, //20 May 2009, UT1-UTC= 0.2652618, Accumulated Leap Seconds=34
  { 2454972.50, 65.920222 }, //21 May 2009, UT1-UTC= 0.2637775, Accumulated Leap Seconds=34
  { 2454973.50, 65.921624 }, //22 May 2009, UT1-UTC= 0.2623763, Accumulated Leap Seconds=34
  { 2454974.50, 65.922828 }, //23 May 2009, UT1-UTC= 0.2611715, Accumulated Leap Seconds=34
  { 2454975.50, 65.923856 }, //24 May 2009, UT1-UTC= 0.2601438, Accumulated Leap Seconds=34
  { 2454976.50, 65.924723 }, //25 May 2009, UT1-UTC= 0.2592765, Accumulated Leap Seconds=34
  { 2454977.50, 65.925475 }, //26 May 2009, UT1-UTC= 0.2585250, Accumulated Leap Seconds=34
  { 2454978.50, 65.926215 }, //27 May 2009, UT1-UTC= 0.2577849, Accumulated Leap Seconds=34
  { 2454979.50, 65.927065 }, //28 May 2009, UT1-UTC= 0.2569352, Accumulated Leap Seconds=34
  { 2454980.50, 65.928114 }, //29 May 2009, UT1-UTC= 0.2558863, Accumulated Leap Seconds=34
  { 2454981.50, 65.929393 }, //30 May 2009, UT1-UTC= 0.2546074, Accumulated Leap Seconds=34
  { 2454982.50, 65.930834 }, //31 May 2009, UT1-UTC= 0.2531661, Accumulated Leap Seconds=34
  { 2454983.50, 65.932291 }, //01 June 2009, UT1-UTC= 0.2517090, Accumulated Leap Seconds=34
  { 2454984.50, 65.933645 }, //02 June 2009, UT1-UTC= 0.2503554, Accumulated Leap Seconds=34
  { 2454985.50, 65.934838 }, //03 June 2009, UT1-UTC= 0.2491618, Accumulated Leap Seconds=34
  { 2454986.50, 65.935818 }, //04 June 2009, UT1-UTC= 0.2481821, Accumulated Leap Seconds=34
  { 2454987.50, 65.936577 }, //05 June 2009, UT1-UTC= 0.2474231, Accumulated Leap Seconds=34
  { 2454988.50, 65.937105 }, //06 June 2009, UT1-UTC= 0.2468948, Accumulated Leap Seconds=34
  { 2454989.50, 65.937490 }, //07 June 2009, UT1-UTC= 0.2465104, Accumulated Leap Seconds=34
  { 2454990.50, 65.937769 }, //08 June 2009, UT1-UTC= 0.2462312, Accumulated Leap Seconds=34
  { 2454991.50, 65.938026 }, //09 June 2009, UT1-UTC= 0.2459743, Accumulated Leap Seconds=34
  { 2454992.50, 65.938335 }, //10 June 2009, UT1-UTC= 0.2456650, Accumulated Leap Seconds=34
  { 2454993.50, 65.938750 }, //11 June 2009, UT1-UTC= 0.2452496, Accumulated Leap Seconds=34
  { 2454994.50, 65.939333 }, //12 June 2009, UT1-UTC= 0.2446667, Accumulated Leap Seconds=34
  { 2454995.50, 65.940072 }, //13 June 2009, UT1-UTC= 0.2439277, Accumulated Leap Seconds=34
  { 2454996.50, 65.940920 }, //14 June 2009, UT1-UTC= 0.2430796, Accumulated Leap Seconds=34
  { 2454997.50, 65.941845 }, //15 June 2009, UT1-UTC= 0.2421545, Accumulated Leap Seconds=34
  { 2454998.50, 65.942839 }, //16 June 2009, UT1-UTC= 0.2411614, Accumulated Leap Seconds=34
  { 2454999.50, 65.943875 }, //17 June 2009, UT1-UTC= 0.2401249, Accumulated Leap Seconds=34
  { 2455000.50, 65.944838 }, //18 June 2009, UT1-UTC= 0.2391618, Accumulated Leap Seconds=34
  { 2455001.50, 65.945646 }, //19 June 2009, UT1-UTC= 0.2383535, Accumulated Leap Seconds=34
  { 2455002.50, 65.946258 }, //20 June 2009, UT1-UTC= 0.2377416, Accumulated Leap Seconds=34
  { 2455003.50, 65.946651 }, //21 June 2009, UT1-UTC= 0.2373492, Accumulated Leap Seconds=34
  { 2455004.50, 65.946869 }, //22 June 2009, UT1-UTC= 0.2371310, Accumulated Leap Seconds=34
  { 2455005.50, 65.946998 }, //23 June 2009, UT1-UTC= 0.2370015, Accumulated Leap Seconds=34
  { 2455006.50, 65.947175 }, //24 June 2009, UT1-UTC= 0.2368248, Accumulated Leap Seconds=34
  { 2455007.50, 65.947464 }, //25 June 2009, UT1-UTC= 0.2365362, Accumulated Leap Seconds=34
  { 2455008.50, 65.947959 }, //26 June 2009, UT1-UTC= 0.2360414, Accumulated Leap Seconds=34
  { 2455009.50, 65.948630 }, //27 June 2009, UT1-UTC= 0.2353703, Accumulated Leap Seconds=34
  { 2455010.50, 65.949372 }, //28 June 2009, UT1-UTC= 0.2346282, Accumulated Leap Seconds=34
  { 2455011.50, 65.950070 }, //29 June 2009, UT1-UTC= 0.2339304, Accumulated Leap Seconds=34
  { 2455012.50, 65.950607 }, //30 June 2009, UT1-UTC= 0.2333927, Accumulated Leap Seconds=34
  { 2455013.50, 65.950911 }, //01 July 2009, UT1-UTC= 0.2330890, Accumulated Leap Seconds=34
  { 2455014.50, 65.950974 }, //02 July 2009, UT1-UTC= 0.2330259, Accumulated Leap Seconds=34
  { 2455015.50, 65.950838 }, //03 July 2009, UT1-UTC= 0.2331616, Accumulated Leap Seconds=34
  { 2455016.50, 65.950574 }, //04 July 2009, UT1-UTC= 0.2334258, Accumulated Leap Seconds=34
  { 2455017.50, 65.950208 }, //05 July 2009, UT1-UTC= 0.2337924, Accumulated Leap Seconds=34
  { 2455018.50, 65.949783 }, //06 July 2009, UT1-UTC= 0.2342165, Accumulated Leap Seconds=34
  { 2455019.50, 65.949376 }, //07 July 2009, UT1-UTC= 0.2346239, Accumulated Leap Seconds=34
  { 2455020.50, 65.949039 }, //08 July 2009, UT1-UTC= 0.2349612, Accumulated Leap Seconds=34
  { 2455021.50, 65.948851 }, //09 July 2009, UT1-UTC= 0.2351490, Accumulated Leap Seconds=34
  { 2455022.50, 65.948803 }, //10 July 2009, UT1-UTC= 0.2351971, Accumulated Leap Seconds=34
  { 2455023.50, 65.948880 }, //11 July 2009, UT1-UTC= 0.2351196, Accumulated Leap Seconds=34
  { 2455024.50, 65.949109 }, //12 July 2009, UT1-UTC= 0.2348913, Accumulated Leap Seconds=34
  { 2455025.50, 65.949409 }, //13 July 2009, UT1-UTC= 0.2345908, Accumulated Leap Seconds=34
  { 2455026.50, 65.949712 }, //14 July 2009, UT1-UTC= 0.2342881, Accumulated Leap Seconds=34
  { 2455027.50, 65.950000 }, //15 July 2009, UT1-UTC= 0.2340001, Accumulated Leap Seconds=34
  { 2455028.50, 65.950197 }, //16 July 2009, UT1-UTC= 0.2338029, Accumulated Leap Seconds=34
  { 2455029.50, 65.950254 }, //17 July 2009, UT1-UTC= 0.2337458, Accumulated Leap Seconds=34
  { 2455030.50, 65.950194 }, //18 July 2009, UT1-UTC= 0.2338058, Accumulated Leap Seconds=34
  { 2455031.50, 65.950052 }, //19 July 2009, UT1-UTC= 0.2339479, Accumulated Leap Seconds=34
  { 2455032.50, 65.949900 }, //20 July 2009, UT1-UTC= 0.2341004, Accumulated Leap Seconds=34
  { 2455033.50, 65.949820 }, //21 July 2009, UT1-UTC= 0.2341796, Accumulated Leap Seconds=34
  { 2455034.50, 65.949900 }, //22 July 2009, UT1-UTC= 0.2341003, Accumulated Leap Seconds=34
  { 2455035.50, 65.950181 }, //23 July 2009, UT1-UTC= 0.2338189, Accumulated Leap Seconds=34
  { 2455036.50, 65.950711 }, //24 July 2009, UT1-UTC= 0.2332889, Accumulated Leap Seconds=34
  { 2455037.50, 65.951420 }, //25 July 2009, UT1-UTC= 0.2325804, Accumulated Leap Seconds=34
  { 2455038.50, 65.952186 }, //26 July 2009, UT1-UTC= 0.2318139, Accumulated Leap Seconds=34
  { 2455039.50, 65.952889 }, //27 July 2009, UT1-UTC= 0.2311110, Accumulated Leap Seconds=34
  { 2455040.50, 65.953444 }, //28 July 2009, UT1-UTC= 0.2305564, Accumulated Leap Seconds=34
  { 2455041.50, 65.953749 }, //29 July 2009, UT1-UTC= 0.2302513, Accumulated Leap Seconds=34
  { 2455042.50, 65.953822 }, //30 July 2009, UT1-UTC= 0.2301777, Accumulated Leap Seconds=34
  { 2455043.50, 65.953689 }, //31 July 2009, UT1-UTC= 0.2303110, Accumulated Leap Seconds=34
  { 2455044.50, 65.953411 }, //01 August 2009, UT1-UTC= 0.2305895, Accumulated Leap Seconds=34
  { 2455045.50, 65.953065 }, //02 August 2009, UT1-UTC= 0.2309348, Accumulated Leap Seconds=34
  { 2455046.50, 65.952781 }, //03 August 2009, UT1-UTC= 0.2312188, Accumulated Leap Seconds=34
  { 2455047.50, 65.952604 }, //04 August 2009, UT1-UTC= 0.2313959, Accumulated Leap Seconds=34
  { 2455048.50, 65.952521 }, //05 August 2009, UT1-UTC= 0.2314794, Accumulated Leap Seconds=34
  { 2455049.50, 65.952616 }, //06 August 2009, UT1-UTC= 0.2313842, Accumulated Leap Seconds=34
  { 2455050.50, 65.952940 }, //07 August 2009, UT1-UTC= 0.2310598, Accumulated Leap Seconds=34
  { 2455051.50, 65.953396 }, //08 August 2009, UT1-UTC= 0.2306036, Accumulated Leap Seconds=34
  { 2455052.50, 65.953977 }, //09 August 2009, UT1-UTC= 0.2300233, Accumulated Leap Seconds=34
  { 2455053.50, 65.954644 }, //10 August 2009, UT1-UTC= 0.2293560, Accumulated Leap Seconds=34
  { 2455054.50, 65.955294 }, //11 August 2009, UT1-UTC= 0.2287061, Accumulated Leap Seconds=34
  { 2455055.50, 65.955837 }, //12 August 2009, UT1-UTC= 0.2281630, Accumulated Leap Seconds=34
  { 2455056.50, 65.956215 }, //13 August 2009, UT1-UTC= 0.2277853, Accumulated Leap Seconds=34
  { 2455057.50, 65.956399 }, //14 August 2009, UT1-UTC= 0.2276013, Accumulated Leap Seconds=34
  { 2455058.50, 65.956397 }, //15 August 2009, UT1-UTC= 0.2276029, Accumulated Leap Seconds=34
  { 2455059.50, 65.956324 }, //16 August 2009, UT1-UTC= 0.2276759, Accumulated Leap Seconds=34
  { 2455060.50, 65.956267 }, //17 August 2009, UT1-UTC= 0.2277326, Accumulated Leap Seconds=34
  { 2455061.50, 65.956369 }, //18 August 2009, UT1-UTC= 0.2276306, Accumulated Leap Seconds=34
  { 2455062.50, 65.956756 }, //19 August 2009, UT1-UTC= 0.2272438, Accumulated Leap Seconds=34
  { 2455063.50, 65.957419 }, //20 August 2009, UT1-UTC= 0.2265811, Accumulated Leap Seconds=34
  { 2455064.50, 65.958318 }, //21 August 2009, UT1-UTC= 0.2256824, Accumulated Leap Seconds=34
  { 2455065.50, 65.959325 }, //22 August 2009, UT1-UTC= 0.2246746, Accumulated Leap Seconds=34
  { 2455066.50, 65.960273 }, //23 August 2009, UT1-UTC= 0.2237272, Accumulated Leap Seconds=34
  { 2455067.50, 65.961088 }, //24 August 2009, UT1-UTC= 0.2229120, Accumulated Leap Seconds=34
  { 2455068.50, 65.961691 }, //25 August 2009, UT1-UTC= 0.2223091, Accumulated Leap Seconds=34
  { 2455069.50, 65.962057 }, //26 August 2009, UT1-UTC= 0.2219434, Accumulated Leap Seconds=34
  { 2455070.50, 65.962253 }, //27 August 2009, UT1-UTC= 0.2217470, Accumulated Leap Seconds=34
  { 2455071.50, 65.962357 }, //28 August 2009, UT1-UTC= 0.2216431, Accumulated Leap Seconds=34
  { 2455072.50, 65.962429 }, //29 August 2009, UT1-UTC= 0.2215710, Accumulated Leap Seconds=34
  { 2455073.50, 65.962496 }, //30 August 2009, UT1-UTC= 0.2215045, Accumulated Leap Seconds=34
  { 2455074.50, 65.962613 }, //31 August 2009, UT1-UTC= 0.2213869, Accumulated Leap Seconds=34
  { 2455075.50, 65.962833 }, //01 September 2009, UT1-UTC= 0.2211670, Accumulated Leap Seconds=34
  { 2455076.50, 65.963210 }, //02 September 2009, UT1-UTC= 0.2207904, Accumulated Leap Seconds=34
  { 2455077.50, 65.963768 }, //03 September 2009, UT1-UTC= 0.2202321, Accumulated Leap Seconds=34
  { 2455078.50, 65.964499 }, //04 September 2009, UT1-UTC= 0.2195012, Accumulated Leap Seconds=34
  { 2455079.50, 65.965371 }, //05 September 2009, UT1-UTC= 0.2186289, Accumulated Leap Seconds=34
  { 2455080.50, 65.966328 }, //06 September 2009, UT1-UTC= 0.2176721, Accumulated Leap Seconds=34
  { 2455081.50, 65.967271 }, //07 September 2009, UT1-UTC= 0.2167292, Accumulated Leap Seconds=34
  { 2455082.50, 65.968132 }, //08 September 2009, UT1-UTC= 0.2158682, Accumulated Leap Seconds=34
  { 2455083.50, 65.968837 }, //09 September 2009, UT1-UTC= 0.2151630, Accumulated Leap Seconds=34
  { 2455084.50, 65.969346 }, //10 September 2009, UT1-UTC= 0.2146539, Accumulated Leap Seconds=34
  { 2455085.50, 65.969703 }, //11 September 2009, UT1-UTC= 0.2142969, Accumulated Leap Seconds=34
  { 2455086.50, 65.969965 }, //12 September 2009, UT1-UTC= 0.2140350, Accumulated Leap Seconds=34
  { 2455087.50, 65.970260 }, //13 September 2009, UT1-UTC= 0.2137395, Accumulated Leap Seconds=34
  { 2455088.50, 65.970685 }, //14 September 2009, UT1-UTC= 0.2133151, Accumulated Leap Seconds=34
  { 2455089.50, 65.971311 }, //15 September 2009, UT1-UTC= 0.2126885, Accumulated Leap Seconds=34
  { 2455090.50, 65.972181 }, //16 September 2009, UT1-UTC= 0.2118194, Accumulated Leap Seconds=34
  { 2455091.50, 65.973326 }, //17 September 2009, UT1-UTC= 0.2106744, Accumulated Leap Seconds=34
  { 2455092.50, 65.974669 }, //18 September 2009, UT1-UTC= 0.2093311, Accumulated Leap Seconds=34
  { 2455093.50, 65.976014 }, //19 September 2009, UT1-UTC= 0.2079858, Accumulated Leap Seconds=34
  { 2455094.50, 65.977291 }, //20 September 2009, UT1-UTC= 0.2067089, Accumulated Leap Seconds=34
  { 2455095.50, 65.978378 }, //21 September 2009, UT1-UTC= 0.2056216, Accumulated Leap Seconds=34
  { 2455096.50, 65.979173 }, //22 September 2009, UT1-UTC= 0.2048271, Accumulated Leap Seconds=34
  { 2455097.50, 65.979698 }, //23 September 2009, UT1-UTC= 0.2043018, Accumulated Leap Seconds=34
  { 2455098.50, 65.980080 }, //24 September 2009, UT1-UTC= 0.2039199, Accumulated Leap Seconds=34
  { 2455099.50, 65.980383 }, //25 September 2009, UT1-UTC= 0.2036174, Accumulated Leap Seconds=34
  { 2455100.50, 65.980664 }, //26 September 2009, UT1-UTC= 0.2033359, Accumulated Leap Seconds=34
  { 2455101.50, 65.980944 }, //27 September 2009, UT1-UTC= 0.2030560, Accumulated Leap Seconds=34
  { 2455102.50, 65.981307 }, //28 September 2009, UT1-UTC= 0.2026932, Accumulated Leap Seconds=34
  { 2455103.50, 65.981896 }, //29 September 2009, UT1-UTC= 0.2021040, Accumulated Leap Seconds=34
  { 2455104.50, 65.982758 }, //30 September 2009, UT1-UTC= 0.2012420, Accumulated Leap Seconds=34
  { 2455105.50, 65.983865 }, //01 October 2009, UT1-UTC= 0.2001353, Accumulated Leap Seconds=34
  { 2455106.50, 65.985106 }, //02 October 2009, UT1-UTC= 0.1988940, Accumulated Leap Seconds=34
  { 2455107.50, 65.986399 }, //03 October 2009, UT1-UTC= 0.1976013, Accumulated Leap Seconds=34
  { 2455108.50, 65.987731 }, //04 October 2009, UT1-UTC= 0.1962694, Accumulated Leap Seconds=34
  { 2455109.50, 65.989002 }, //05 October 2009, UT1-UTC= 0.1949983, Accumulated Leap Seconds=34
  { 2455110.50, 65.990095 }, //06 October 2009, UT1-UTC= 0.1939051, Accumulated Leap Seconds=34
  { 2455111.50, 65.991024 }, //07 October 2009, UT1-UTC= 0.1929757, Accumulated Leap Seconds=34
  { 2455112.50, 65.991778 }, //08 October 2009, UT1-UTC= 0.1922218, Accumulated Leap Seconds=34
  { 2455113.50, 65.992407 }, //09 October 2009, UT1-UTC= 0.1915935, Accumulated Leap Seconds=34
  { 2455114.50, 65.992994 }, //10 October 2009, UT1-UTC= 0.1910059, Accumulated Leap Seconds=34
  { 2455115.50, 65.993615 }, //11 October 2009, UT1-UTC= 0.1903848, Accumulated Leap Seconds=34
  { 2455116.50, 65.994328 }, //12 October 2009, UT1-UTC= 0.1896720, Accumulated Leap Seconds=34
  { 2455117.50, 65.995268 }, //13 October 2009, UT1-UTC= 0.1887316, Accumulated Leap Seconds=34
  { 2455118.50, 65.996453 }, //14 October 2009, UT1-UTC= 0.1875469, Accumulated Leap Seconds=34
  { 2455119.50, 65.997827 }, //15 October 2009, UT1-UTC= 0.1861733, Accumulated Leap Seconds=34
  { 2455120.50, 65.999315 }, //16 October 2009, UT1-UTC= 0.1846849, Accumulated Leap Seconds=34
  { 2455121.50, 66.000797 }, //17 October 2009, UT1-UTC= 0.1832026, Accumulated Leap Seconds=34
  { 2455122.50, 66.002122 }, //18 October 2009, UT1-UTC= 0.1818784, Accumulated Leap Seconds=34
  { 2455123.50, 66.003224 }, //19 October 2009, UT1-UTC= 0.1807756, Accumulated Leap Seconds=34
  { 2455124.50, 66.004103 }, //20 October 2009, UT1-UTC= 0.1798975, Accumulated Leap Seconds=34
  { 2455125.50, 66.004802 }, //21 October 2009, UT1-UTC= 0.1791977, Accumulated Leap Seconds=34
  { 2455126.50, 66.005401 }, //22 October 2009, UT1-UTC= 0.1785993, Accumulated Leap Seconds=34
  { 2455127.50, 66.005973 }, //23 October 2009, UT1-UTC= 0.1780274, Accumulated Leap Seconds=34
  { 2455128.50, 66.006494 }, //24 October 2009, UT1-UTC= 0.1775061, Accumulated Leap Seconds=34
  { 2455129.50, 66.007055 }, //25 October 2009, UT1-UTC= 0.1769450, Accumulated Leap Seconds=34
  { 2455130.50, 66.007776 }, //26 October 2009, UT1-UTC= 0.1762240, Accumulated Leap Seconds=34
  { 2455131.50, 66.008664 }, //27 October 2009, UT1-UTC= 0.1753358, Accumulated Leap Seconds=34
  { 2455132.50, 66.009721 }, //28 October 2009, UT1-UTC= 0.1742790, Accumulated Leap Seconds=34
  { 2455133.50, 66.010899 }, //29 October 2009, UT1-UTC= 0.1731010, Accumulated Leap Seconds=34
  { 2455134.50, 66.012147 }, //30 October 2009, UT1-UTC= 0.1718527, Accumulated Leap Seconds=34
  { 2455135.50, 66.013417 }, //31 October 2009, UT1-UTC= 0.1705827, Accumulated Leap Seconds=34
  { 2455136.50, 66.014673 }, //01 November 2009, UT1-UTC= 0.1693267, Accumulated Leap Seconds=34
  { 2455137.50, 66.015796 }, //02 November 2009, UT1-UTC= 0.1682037, Accumulated Leap Seconds=34
  { 2455138.50, 66.016779 }, //03 November 2009, UT1-UTC= 0.1672210, Accumulated Leap Seconds=34
  { 2455139.50, 66.017600 }, //04 November 2009, UT1-UTC= 0.1663999, Accumulated Leap Seconds=34
  { 2455140.50, 66.018302 }, //05 November 2009, UT1-UTC= 0.1656982, Accumulated Leap Seconds=34
  { 2455141.50, 66.018970 }, //06 November 2009, UT1-UTC= 0.1650303, Accumulated Leap Seconds=34
  { 2455142.50, 66.019635 }, //07 November 2009, UT1-UTC= 0.1643645, Accumulated Leap Seconds=34
  { 2455143.50, 66.020392 }, //08 November 2009, UT1-UTC= 0.1636083, Accumulated Leap Seconds=34
  { 2455144.50, 66.021282 }, //09 November 2009, UT1-UTC= 0.1627181, Accumulated Leap Seconds=34
  { 2455145.50, 66.022315 }, //10 November 2009, UT1-UTC= 0.1616846, Accumulated Leap Seconds=34
  { 2455146.50, 66.023539 }, //11 November 2009, UT1-UTC= 0.1604613, Accumulated Leap Seconds=34
  { 2455147.50, 66.024866 }, //12 November 2009, UT1-UTC= 0.1591338, Accumulated Leap Seconds=34
  { 2455148.50, 66.026187 }, //13 November 2009, UT1-UTC= 0.1578132, Accumulated Leap Seconds=34
  { 2455149.50, 66.027339 }, //14 November 2009, UT1-UTC= 0.1566610, Accumulated Leap Seconds=34
  { 2455150.50, 66.028284 }, //15 November 2009, UT1-UTC= 0.1557155, Accumulated Leap Seconds=34
  { 2455151.50, 66.029036 }, //16 November 2009, UT1-UTC= 0.1549638, Accumulated Leap Seconds=34
  { 2455152.50, 66.029597 }, //17 November 2009, UT1-UTC= 0.1544033, Accumulated Leap Seconds=34
  { 2455153.50, 66.030054 }, //18 November 2009, UT1-UTC= 0.1539462, Accumulated Leap Seconds=34
  { 2455154.50, 66.030442 }, //19 November 2009, UT1-UTC= 0.1535580, Accumulated Leap Seconds=34
  { 2455155.50, 66.030780 }, //20 November 2009, UT1-UTC= 0.1532202, Accumulated Leap Seconds=34
  { 2455156.50, 66.031212 }, //21 November 2009, UT1-UTC= 0.1527877, Accumulated Leap Seconds=34
  { 2455157.50, 66.031836 }, //22 November 2009, UT1-UTC= 0.1521636, Accumulated Leap Seconds=34
  { 2455158.50, 66.032686 }, //23 November 2009, UT1-UTC= 0.1513143, Accumulated Leap Seconds=34
  { 2455159.50, 66.033705 }, //24 November 2009, UT1-UTC= 0.1502950, Accumulated Leap Seconds=34
  { 2455160.50, 66.034839 }, //25 November 2009, UT1-UTC= 0.1491607, Accumulated Leap Seconds=34
  { 2455161.50, 66.036087 }, //26 November 2009, UT1-UTC= 0.1479131, Accumulated Leap Seconds=34
  { 2455162.50, 66.037428 }, //27 November 2009, UT1-UTC= 0.1465720, Accumulated Leap Seconds=34
  { 2455163.50, 66.038782 }, //28 November 2009, UT1-UTC= 0.1452184, Accumulated Leap Seconds=34
  { 2455164.50, 66.040044 }, //29 November 2009, UT1-UTC= 0.1439564, Accumulated Leap Seconds=34
  { 2455165.50, 66.041154 }, //30 November 2009, UT1-UTC= 0.1428458, Accumulated Leap Seconds=34
  { 2455166.50, 66.042049 }, //01 December 2009, UT1-UTC= 0.1419510, Accumulated Leap Seconds=34
  { 2455167.50, 66.042756 }, //02 December 2009, UT1-UTC= 0.1412438, Accumulated Leap Seconds=34
  { 2455168.50, 66.043389 }, //03 December 2009, UT1-UTC= 0.1406106, Accumulated Leap Seconds=34
  { 2455169.50, 66.044038 }, //04 December 2009, UT1-UTC= 0.1399617, Accumulated Leap Seconds=34
  { 2455170.50, 66.044820 }, //05 December 2009, UT1-UTC= 0.1391796, Accumulated Leap Seconds=34
  { 2455171.50, 66.045751 }, //06 December 2009, UT1-UTC= 0.1382485, Accumulated Leap Seconds=34
  { 2455172.50, 66.046881 }, //07 December 2009, UT1-UTC= 0.1371190, Accumulated Leap Seconds=34
  { 2455173.50, 66.048163 }, //08 December 2009, UT1-UTC= 0.1358371, Accumulated Leap Seconds=34
  { 2455174.50, 66.049591 }, //09 December 2009, UT1-UTC= 0.1344085, Accumulated Leap Seconds=34
  { 2455175.50, 66.050987 }, //10 December 2009, UT1-UTC= 0.1330130, Accumulated Leap Seconds=34
  { 2455176.50, 66.052233 }, //11 December 2009, UT1-UTC= 0.1317667, Accumulated Leap Seconds=34
  { 2455177.50, 66.053263 }, //12 December 2009, UT1-UTC= 0.1307371, Accumulated Leap Seconds=34
  { 2455178.50, 66.054092 }, //13 December 2009, UT1-UTC= 0.1299079, Accumulated Leap Seconds=34
  { 2455179.50, 66.054794 }, //14 December 2009, UT1-UTC= 0.1292060, Accumulated Leap Seconds=34
  { 2455180.50, 66.055395 }, //15 December 2009, UT1-UTC= 0.1286051, Accumulated Leap Seconds=34
  { 2455181.50, 66.055968 }, //16 December 2009, UT1-UTC= 0.1280323, Accumulated Leap Seconds=34
  { 2455182.50, 66.056570 }, //17 December 2009, UT1-UTC= 0.1274299, Accumulated Leap Seconds=34
  { 2455183.50, 66.057231 }, //18 December 2009, UT1-UTC= 0.1267693, Accumulated Leap Seconds=34
  { 2455184.50, 66.057984 }, //19 December 2009, UT1-UTC= 0.1260159, Accumulated Leap Seconds=34
  { 2455185.50, 66.058877 }, //20 December 2009, UT1-UTC= 0.1251234, Accumulated Leap Seconds=34
  { 2455186.50, 66.059929 }, //21 December 2009, UT1-UTC= 0.1240707, Accumulated Leap Seconds=34
  { 2455187.50, 66.061150 }, //22 December 2009, UT1-UTC= 0.1228495, Accumulated Leap Seconds=34
  { 2455188.50, 66.062460 }, //23 December 2009, UT1-UTC= 0.1215400, Accumulated Leap Seconds=34
  { 2455189.50, 66.063801 }, //24 December 2009, UT1-UTC= 0.1201994, Accumulated Leap Seconds=34
  { 2455190.50, 66.065044 }, //25 December 2009, UT1-UTC= 0.1189558, Accumulated Leap Seconds=34
  { 2455191.50, 66.066193 }, //26 December 2009, UT1-UTC= 0.1178071, Accumulated Leap Seconds=34
  { 2455192.50, 66.067219 }, //27 December 2009, UT1-UTC= 0.1167809, Accumulated Leap Seconds=34
  { 2455193.50, 66.068070 }, //28 December 2009, UT1-UTC= 0.1159300, Accumulated Leap Seconds=34
  { 2455194.50, 66.068706 }, //29 December 2009, UT1-UTC= 0.1152943, Accumulated Leap Seconds=34
  { 2455195.50, 66.069130 }, //30 December 2009, UT1-UTC= 0.1148704, Accumulated Leap Seconds=34
  { 2455196.50, 66.069486 }, //31 December 2009, UT1-UTC= 0.1145137, Accumulated Leap Seconds=34
  { 2455197.50, 66.069922 }, //01 January 2010, UT1-UTC= 0.1140783, Accumulated Leap Seconds=34
  { 2455198.50, 66.070555 }, //02 January 2010, UT1-UTC= 0.1134454, Accumulated Leap Seconds=34
  { 2455199.50, 66.071474 }, //03 January 2010, UT1-UTC= 0.1125259, Accumulated Leap Seconds=34
  { 2455200.50, 66.072629 }, //04 January 2010, UT1-UTC= 0.1113713, Accumulated Leap Seconds=34
  { 2455201.50, 66.073982 }, //05 January 2010, UT1-UTC= 0.1100183, Accumulated Leap Seconds=34
  { 2455202.50, 66.075375 }, //06 January 2010, UT1-UTC= 0.1086245, Accumulated Leap Seconds=34
  { 2455203.50, 66.076653 }, //07 January 2010, UT1-UTC= 0.1073468, Accumulated Leap Seconds=34
  { 2455204.50, 66.077740 }, //08 January 2010, UT1-UTC= 0.1062598, Accumulated Leap Seconds=34
  { 2455205.50, 66.078602 }, //09 January 2010, UT1-UTC= 0.1053984, Accumulated Leap Seconds=34
  { 2455206.50, 66.079185 }, //10 January 2010, UT1-UTC= 0.1048153, Accumulated Leap Seconds=34
  { 2455207.50, 66.079581 }, //11 January 2010, UT1-UTC= 0.1044188, Accumulated Leap Seconds=34
  { 2455208.50, 66.079862 }, //12 January 2010, UT1-UTC= 0.1041377, Accumulated Leap Seconds=34
  { 2455209.50, 66.080120 }, //13 January 2010, UT1-UTC= 0.1038803, Accumulated Leap Seconds=34
  { 2455210.50, 66.080354 }, //14 January 2010, UT1-UTC= 0.1036455, Accumulated Leap Seconds=34
  { 2455211.50, 66.080673 }, //15 January 2010, UT1-UTC= 0.1033269, Accumulated Leap Seconds=34
  { 2455212.50, 66.081143 }, //16 January 2010, UT1-UTC= 0.1028570, Accumulated Leap Seconds=34
  { 2455213.50, 66.081785 }, //17 January 2010, UT1-UTC= 0.1022150, Accumulated Leap Seconds=34
  { 2455214.50, 66.082632 }, //18 January 2010, UT1-UTC= 0.1013681, Accumulated Leap Seconds=34
  { 2455215.50, 66.083588 }, //19 January 2010, UT1-UTC= 0.1004121, Accumulated Leap Seconds=34
  { 2455216.50, 66.084578 }, //20 January 2010, UT1-UTC= 0.0994218, Accumulated Leap Seconds=34
  { 2455217.50, 66.085578 }, //21 January 2010, UT1-UTC= 0.0984222, Accumulated Leap Seconds=34
  { 2455218.50, 66.086602 }, //22 January 2010, UT1-UTC= 0.0973980, Accumulated Leap Seconds=34
  { 2455219.50, 66.087588 }, //23 January 2010, UT1-UTC= 0.0964122, Accumulated Leap Seconds=34
  { 2455220.50, 66.088487 }, //24 January 2010, UT1-UTC= 0.0955129, Accumulated Leap Seconds=34
  { 2455221.50, 66.089274 }, //25 January 2010, UT1-UTC= 0.0947260, Accumulated Leap Seconds=34
  { 2455222.50, 66.089911 }, //26 January 2010, UT1-UTC= 0.0940895, Accumulated Leap Seconds=34
  { 2455223.50, 66.090462 }, //27 January 2010, UT1-UTC= 0.0935376, Accumulated Leap Seconds=34
  { 2455224.50, 66.091070 }, //28 January 2010, UT1-UTC= 0.0929297, Accumulated Leap Seconds=34
  { 2455225.50, 66.091897 }, //29 January 2010, UT1-UTC= 0.0921034, Accumulated Leap Seconds=34
  { 2455226.50, 66.092994 }, //30 January 2010, UT1-UTC= 0.0910058, Accumulated Leap Seconds=34
  { 2455227.50, 66.094403 }, //31 January 2010, UT1-UTC= 0.0895969, Accumulated Leap Seconds=34
  { 2455228.50, 66.096134 }, //01 February 2010, UT1-UTC= 0.0878657, Accumulated Leap Seconds=34
  { 2455229.50, 66.098019 }, //02 February 2010, UT1-UTC= 0.0859807, Accumulated Leap Seconds=34
  { 2455230.50, 66.099876 }, //03 February 2010, UT1-UTC= 0.0841237, Accumulated Leap Seconds=34
  { 2455231.50, 66.101576 }, //04 February 2010, UT1-UTC= 0.0824237, Accumulated Leap Seconds=34
  { 2455232.50, 66.102996 }, //05 February 2010, UT1-UTC= 0.0810036, Accumulated Leap Seconds=34
  { 2455233.50, 66.104149 }, //06 February 2010, UT1-UTC= 0.0798511, Accumulated Leap Seconds=34
  { 2455234.50, 66.105086 }, //07 February 2010, UT1-UTC= 0.0789140, Accumulated Leap Seconds=34
  { 2455235.50, 66.105838 }, //08 February 2010, UT1-UTC= 0.0781619, Accumulated Leap Seconds=34
  { 2455236.50, 66.106460 }, //09 February 2010, UT1-UTC= 0.0775395, Accumulated Leap Seconds=34
  { 2455237.50, 66.107012 }, //10 February 2010, UT1-UTC= 0.0769876, Accumulated Leap Seconds=34
  { 2455238.50, 66.107613 }, //11 February 2010, UT1-UTC= 0.0763870, Accumulated Leap Seconds=34
  { 2455239.50, 66.108347 }, //12 February 2010, UT1-UTC= 0.0756531, Accumulated Leap Seconds=34
  { 2455240.50, 66.109243 }, //13 February 2010, UT1-UTC= 0.0747570, Accumulated Leap Seconds=34
  { 2455241.50, 66.110325 }, //14 February 2010, UT1-UTC= 0.0736753, Accumulated Leap Seconds=34
  { 2455242.50, 66.111571 }, //15 February 2010, UT1-UTC= 0.0724293, Accumulated Leap Seconds=34
  { 2455243.50, 66.112956 }, //16 February 2010, UT1-UTC= 0.0710437, Accumulated Leap Seconds=34
  { 2455244.50, 66.114418 }, //17 February 2010, UT1-UTC= 0.0695816, Accumulated Leap Seconds=34
  { 2455245.50, 66.115941 }, //18 February 2010, UT1-UTC= 0.0680586, Accumulated Leap Seconds=34
  { 2455246.50, 66.117421 }, //19 February 2010, UT1-UTC= 0.0665789, Accumulated Leap Seconds=34
  { 2455247.50, 66.118803 }, //20 February 2010, UT1-UTC= 0.0651974, Accumulated Leap Seconds=34
  { 2455248.50, 66.120051 }, //21 February 2010, UT1-UTC= 0.0639494, Accumulated Leap Seconds=34
  { 2455249.50, 66.121173 }, //22 February 2010, UT1-UTC= 0.0628269, Accumulated Leap Seconds=34
  { 2455250.50, 66.122182 }, //23 February 2010, UT1-UTC= 0.0618181, Accumulated Leap Seconds=34
  { 2455251.50, 66.123136 }, //24 February 2010, UT1-UTC= 0.0608641, Accumulated Leap Seconds=34
  { 2455252.50, 66.124215 }, //25 February 2010, UT1-UTC= 0.0597845, Accumulated Leap Seconds=34
  { 2455253.50, 66.125534 }, //26 February 2010, UT1-UTC= 0.0584662, Accumulated Leap Seconds=34
  { 2455254.50, 66.127102 }, //27 February 2010, UT1-UTC= 0.0568980, Accumulated Leap Seconds=34
  { 2455255.50, 66.128962 }, //28 February 2010, UT1-UTC= 0.0550378, Accumulated Leap Seconds=34
  { 2455256.50, 66.131012 }, //01 March 2010, UT1-UTC= 0.0529884, Accumulated Leap Seconds=34
  { 2455257.50, 66.133076 }, //02 March 2010, UT1-UTC= 0.0509242, Accumulated Leap Seconds=34
  { 2455258.50, 66.134954 }, //03 March 2010, UT1-UTC= 0.0490461, Accumulated Leap Seconds=34
  { 2455259.50, 66.136605 }, //04 March 2010, UT1-UTC= 0.0473946, Accumulated Leap Seconds=34
  { 2455260.50, 66.137980 }, //05 March 2010, UT1-UTC= 0.0460197, Accumulated Leap Seconds=34
  { 2455261.50, 66.139068 }, //06 March 2010, UT1-UTC= 0.0449323, Accumulated Leap Seconds=34
  { 2455262.50, 66.139975 }, //07 March 2010, UT1-UTC= 0.0440250, Accumulated Leap Seconds=34
  { 2455263.50, 66.140735 }, //08 March 2010, UT1-UTC= 0.0432650, Accumulated Leap Seconds=34
  { 2455264.50, 66.141537 }, //09 March 2010, UT1-UTC= 0.0424633, Accumulated Leap Seconds=34
  { 2455265.50, 66.142456 }, //10 March 2010, UT1-UTC= 0.0415440, Accumulated Leap Seconds=34
  { 2455266.50, 66.143394 }, //11 March 2010, UT1-UTC= 0.0406058, Accumulated Leap Seconds=34
  { 2455267.50, 66.144404 }, //12 March 2010, UT1-UTC= 0.0395956, Accumulated Leap Seconds=34
  { 2455268.50, 66.145521 }, //13 March 2010, UT1-UTC= 0.0384795, Accumulated Leap Seconds=34
  { 2455269.50, 66.146618 }, //14 March 2010, UT1-UTC= 0.0373822, Accumulated Leap Seconds=34
  { 2455270.50, 66.147744 }, //15 March 2010, UT1-UTC= 0.0362561, Accumulated Leap Seconds=34
  { 2455271.50, 66.148947 }, //16 March 2010, UT1-UTC= 0.0350533, Accumulated Leap Seconds=34
  { 2455272.50, 66.150178 }, //17 March 2010, UT1-UTC= 0.0338216, Accumulated Leap Seconds=34
  { 2455273.50, 66.151407 }, //18 March 2010, UT1-UTC= 0.0325925, Accumulated Leap Seconds=34
  { 2455274.50, 66.152510 }, //19 March 2010, UT1-UTC= 0.0314904, Accumulated Leap Seconds=34
  { 2455275.50, 66.153429 }, //20 March 2010, UT1-UTC= 0.0305709, Accumulated Leap Seconds=34
  { 2455276.50, 66.154189 }, //21 March 2010, UT1-UTC= 0.0298113, Accumulated Leap Seconds=34
  { 2455277.50, 66.154798 }, //22 March 2010, UT1-UTC= 0.0292025, Accumulated Leap Seconds=34
  { 2455278.50, 66.155341 }, //23 March 2010, UT1-UTC= 0.0286592, Accumulated Leap Seconds=34
  { 2455279.50, 66.155935 }, //24 March 2010, UT1-UTC= 0.0280651, Accumulated Leap Seconds=34
  { 2455280.50, 66.156721 }, //25 March 2010, UT1-UTC= 0.0272792, Accumulated Leap Seconds=34
  { 2455281.50, 66.157858 }, //26 March 2010, UT1-UTC= 0.0261417, Accumulated Leap Seconds=34
  { 2455282.50, 66.159333 }, //27 March 2010, UT1-UTC= 0.0246671, Accumulated Leap Seconds=34
  { 2455283.50, 66.161052 }, //28 March 2010, UT1-UTC= 0.0229478, Accumulated Leap Seconds=34
  { 2455284.50, 66.162965 }, //29 March 2010, UT1-UTC= 0.0210345, Accumulated Leap Seconds=34
  { 2455285.50, 66.164946 }, //30 March 2010, UT1-UTC= 0.0190545, Accumulated Leap Seconds=34
  { 2455286.50, 66.166777 }, //31 March 2010, UT1-UTC= 0.0172228, Accumulated Leap Seconds=34
  { 2455287.50, 66.168271 }, //01 April 2010, UT1-UTC= 0.0157287, Accumulated Leap Seconds=34
  { 2455288.50, 66.169435 }, //02 April 2010, UT1-UTC= 0.0145646, Accumulated Leap Seconds=34
  { 2455289.50, 66.170351 }, //03 April 2010, UT1-UTC= 0.0136485, Accumulated Leap Seconds=34
  { 2455290.50, 66.171142 }, //04 April 2010, UT1-UTC= 0.0128581, Accumulated Leap Seconds=34
  { 2455291.50, 66.171930 }, //05 April 2010, UT1-UTC= 0.0120699, Accumulated Leap Seconds=34
  { 2455292.50, 66.172797 }, //06 April 2010, UT1-UTC= 0.0112035, Accumulated Leap Seconds=34
  { 2455293.50, 66.173764 }, //07 April 2010, UT1-UTC= 0.0102362, Accumulated Leap Seconds=34
  { 2455294.50, 66.174863 }, //08 April 2010, UT1-UTC= 0.0091373, Accumulated Leap Seconds=34
  { 2455295.50, 66.176125 }, //09 April 2010, UT1-UTC= 0.0078750, Accumulated Leap Seconds=34
  { 2455296.50, 66.177490 }, //10 April 2010, UT1-UTC= 0.0065104, Accumulated Leap Seconds=34
  { 2455297.50, 66.178941 }, //11 April 2010, UT1-UTC= 0.0050587, Accumulated Leap Seconds=34
  { 2455298.50, 66.180484 }, //12 April 2010, UT1-UTC= 0.0035160, Accumulated Leap Seconds=34
  { 2455299.50, 66.182041 }, //13 April 2010, UT1-UTC= 0.0019593, Accumulated Leap Seconds=34
  { 2455300.50, 66.183602 }, //14 April 2010, UT1-UTC= 0.0003977, Accumulated Leap Seconds=34
  { 2455301.50, 66.185095 }, //15 April 2010, UT1-UTC=-0.0010948, Accumulated Leap Seconds=34
  { 2455302.50, 66.186465 }, //16 April 2010, UT1-UTC=-0.0024650, Accumulated Leap Seconds=34
  { 2455303.50, 66.187674 }, //17 April 2010, UT1-UTC=-0.0036737, Accumulated Leap Seconds=34
  { 2455304.50, 66.188773 }, //18 April 2010, UT1-UTC=-0.0047732, Accumulated Leap Seconds=34
  { 2455305.50, 66.189804 }, //19 April 2010, UT1-UTC=-0.0058036, Accumulated Leap Seconds=34
  { 2455306.50, 66.190847 }, //20 April 2010, UT1-UTC=-0.0068473, Accumulated Leap Seconds=34
  { 2455307.50, 66.192001 }, //21 April 2010, UT1-UTC=-0.0080007, Accumulated Leap Seconds=34
  { 2455308.50, 66.193375 }, //22 April 2010, UT1-UTC=-0.0093746, Accumulated Leap Seconds=34
  { 2455309.50, 66.194972 }, //23 April 2010, UT1-UTC=-0.0109721, Accumulated Leap Seconds=34
  { 2455310.50, 66.196737 }, //24 April 2010, UT1-UTC=-0.0127370, Accumulated Leap Seconds=34
  { 2455311.50, 66.198615 }, //25 April 2010, UT1-UTC=-0.0146146, Accumulated Leap Seconds=34
  { 2455312.50, 66.200483 }, //26 April 2010, UT1-UTC=-0.0164828, Accumulated Leap Seconds=34
  { 2455313.50, 66.202294 }, //27 April 2010, UT1-UTC=-0.0182937, Accumulated Leap Seconds=34
  { 2455314.50, 66.203929 }, //28 April 2010, UT1-UTC=-0.0199293, Accumulated Leap Seconds=34
  { 2455315.50, 66.205277 }, //29 April 2010, UT1-UTC=-0.0212773, Accumulated Leap Seconds=34
  { 2455316.50, 66.206342 }, //30 April 2010, UT1-UTC=-0.0223424, Accumulated Leap Seconds=34
  { 2455317.50, 66.207163 }, //01 May 2010, UT1-UTC=-0.0231627, Accumulated Leap Seconds=34
  { 2455318.50, 66.207814 }, //02 May 2010, UT1-UTC=-0.0238145, Accumulated Leap Seconds=34
  { 2455319.50, 66.208384 }, //03 May 2010, UT1-UTC=-0.0243844, Accumulated Leap Seconds=34
  { 2455320.50, 66.208912 }, //04 May 2010, UT1-UTC=-0.0249123, Accumulated Leap Seconds=34
  { 2455321.50, 66.209471 }, //05 May 2010, UT1-UTC=-0.0254707, Accumulated Leap Seconds=34
  { 2455322.50, 66.210100 }, //06 May 2010, UT1-UTC=-0.0261002, Accumulated Leap Seconds=34
  { 2455323.50, 66.210831 }, //07 May 2010, UT1-UTC=-0.0268309, Accumulated Leap Seconds=34
  { 2455324.50, 66.211709 }, //08 May 2010, UT1-UTC=-0.0277086, Accumulated Leap Seconds=34
  { 2455325.50, 66.212753 }, //09 May 2010, UT1-UTC=-0.0287534, Accumulated Leap Seconds=34
  { 2455326.50, 66.213922 }, //10 May 2010, UT1-UTC=-0.0299219, Accumulated Leap Seconds=34
  { 2455327.50, 66.215144 }, //11 May 2010, UT1-UTC=-0.0311440, Accumulated Leap Seconds=34
  { 2455328.50, 66.216334 }, //12 May 2010, UT1-UTC=-0.0323336, Accumulated Leap Seconds=34
  { 2455329.50, 66.217396 }, //13 May 2010, UT1-UTC=-0.0333959, Accumulated Leap Seconds=34
  { 2455330.50, 66.218288 }, //14 May 2010, UT1-UTC=-0.0342881, Accumulated Leap Seconds=34
  { 2455331.50, 66.219000 }, //15 May 2010, UT1-UTC=-0.0349997, Accumulated Leap Seconds=34
  { 2455332.50, 66.219650 }, //16 May 2010, UT1-UTC=-0.0356503, Accumulated Leap Seconds=34
  { 2455333.50, 66.220347 }, //17 May 2010, UT1-UTC=-0.0363467, Accumulated Leap Seconds=34
  { 2455334.50, 66.221158 }, //18 May 2010, UT1-UTC=-0.0371575, Accumulated Leap Seconds=34
  { 2455335.50, 66.222217 }, //19 May 2010, UT1-UTC=-0.0382165, Accumulated Leap Seconds=34
  { 2455336.50, 66.223515 }, //20 May 2010, UT1-UTC=-0.0395151, Accumulated Leap Seconds=34
  { 2455337.50, 66.225017 }, //21 May 2010, UT1-UTC=-0.0410167, Accumulated Leap Seconds=34
  { 2455338.50, 66.226672 }, //22 May 2010, UT1-UTC=-0.0426724, Accumulated Leap Seconds=34
  { 2455339.50, 66.228362 }, //23 May 2010, UT1-UTC=-0.0443621, Accumulated Leap Seconds=34
  { 2455340.50, 66.229913 }, //24 May 2010, UT1-UTC=-0.0459129, Accumulated Leap Seconds=34
  { 2455341.50, 66.231226 }, //25 May 2010, UT1-UTC=-0.0472264, Accumulated Leap Seconds=34
  { 2455342.50, 66.232267 }, //26 May 2010, UT1-UTC=-0.0482667, Accumulated Leap Seconds=34
  { 2455343.50, 66.233091 }, //27 May 2010, UT1-UTC=-0.0490912, Accumulated Leap Seconds=34
  { 2455344.50, 66.233798 }, //28 May 2010, UT1-UTC=-0.0497983, Accumulated Leap Seconds=34
  { 2455345.50, 66.234343 }, //29 May 2010, UT1-UTC=-0.0503434, Accumulated Leap Seconds=34
  { 2455346.50, 66.234744 }, //30 May 2010, UT1-UTC=-0.0507442, Accumulated Leap Seconds=34
  { 2455347.50, 66.235122 }, //31 May 2010, UT1-UTC=-0.0511222, Accumulated Leap Seconds=34
  { 2455348.50, 66.235585 }, //01 June 2010, UT1-UTC=-0.0515846, Accumulated Leap Seconds=34
  { 2455349.50, 66.236161 }, //02 June 2010, UT1-UTC=-0.0521615, Accumulated Leap Seconds=34
  { 2455350.50, 66.236830 }, //03 June 2010, UT1-UTC=-0.0528302, Accumulated Leap Seconds=34
  { 2455351.50, 66.237573 }, //04 June 2010, UT1-UTC=-0.0535728, Accumulated Leap Seconds=34
  { 2455352.50, 66.238314 }, //05 June 2010, UT1-UTC=-0.0543140, Accumulated Leap Seconds=34
  { 2455353.50, 66.239023 }, //06 June 2010, UT1-UTC=-0.0550229, Accumulated Leap Seconds=34
  { 2455354.50, 66.239683 }, //07 June 2010, UT1-UTC=-0.0556829, Accumulated Leap Seconds=34
  { 2455355.50, 66.240219 }, //08 June 2010, UT1-UTC=-0.0562189, Accumulated Leap Seconds=34
  { 2455356.50, 66.240662 }, //09 June 2010, UT1-UTC=-0.0566620, Accumulated Leap Seconds=34
  { 2455357.50, 66.240971 }, //10 June 2010, UT1-UTC=-0.0569711, Accumulated Leap Seconds=34
  { 2455358.50, 66.241164 }, //11 June 2010, UT1-UTC=-0.0571637, Accumulated Leap Seconds=34
  { 2455359.50, 66.241245 }, //12 June 2010, UT1-UTC=-0.0572449, Accumulated Leap Seconds=34
  { 2455360.50, 66.241295 }, //13 June 2010, UT1-UTC=-0.0572951, Accumulated Leap Seconds=34
  { 2455361.50, 66.241395 }, //14 June 2010, UT1-UTC=-0.0573947, Accumulated Leap Seconds=34
  { 2455362.50, 66.241599 }, //15 June 2010, UT1-UTC=-0.0575990, Accumulated Leap Seconds=34
  { 2455363.50, 66.241950 }, //16 June 2010, UT1-UTC=-0.0579501, Accumulated Leap Seconds=34
  { 2455364.50, 66.242454 }, //17 June 2010, UT1-UTC=-0.0584543, Accumulated Leap Seconds=34
  { 2455365.50, 66.243089 }, //18 June 2010, UT1-UTC=-0.0590887, Accumulated Leap Seconds=34
  { 2455366.50, 66.243740 }, //19 June 2010, UT1-UTC=-0.0597400, Accumulated Leap Seconds=34
  { 2455367.50, 66.244198 }, //20 June 2010, UT1-UTC=-0.0601981, Accumulated Leap Seconds=34
  { 2455368.50, 66.244410 }, //21 June 2010, UT1-UTC=-0.0604096, Accumulated Leap Seconds=34
  { 2455369.50, 66.244405 }, //22 June 2010, UT1-UTC=-0.0604055, Accumulated Leap Seconds=34
  { 2455370.50, 66.244151 }, //23 June 2010, UT1-UTC=-0.0601505, Accumulated Leap Seconds=34
  { 2455371.50, 66.243702 }, //24 June 2010, UT1-UTC=-0.0597020, Accumulated Leap Seconds=34
  { 2455372.50, 66.243130 }, //25 June 2010, UT1-UTC=-0.0591297, Accumulated Leap Seconds=34
  { 2455373.50, 66.242516 }, //26 June 2010, UT1-UTC=-0.0585156, Accumulated Leap Seconds=34
  { 2455374.50, 66.241934 }, //27 June 2010, UT1-UTC=-0.0579343, Accumulated Leap Seconds=34
  { 2455375.50, 66.241444 }, //28 June 2010, UT1-UTC=-0.0574436, Accumulated Leap Seconds=34
  { 2455376.50, 66.241104 }, //29 June 2010, UT1-UTC=-0.0571041, Accumulated Leap Seconds=34
  { 2455377.50, 66.240915 }, //30 June 2010, UT1-UTC=-0.0569155, Accumulated Leap Seconds=34
  { 2455378.50, 66.240855 }, //01 July 2010, UT1-UTC=-0.0568549, Accumulated Leap Seconds=34
  { 2455379.50, 66.240831 }, //02 July 2010, UT1-UTC=-0.0568310, Accumulated Leap Seconds=34
  { 2455380.50, 66.240870 }, //03 July 2010, UT1-UTC=-0.0568704, Accumulated Leap Seconds=34
  { 2455381.50, 66.240906 }, //04 July 2010, UT1-UTC=-0.0569063, Accumulated Leap Seconds=34
  { 2455382.50, 66.240851 }, //05 July 2010, UT1-UTC=-0.0568509, Accumulated Leap Seconds=34
  { 2455383.50, 66.240669 }, //06 July 2010, UT1-UTC=-0.0566688, Accumulated Leap Seconds=34
  { 2455384.50, 66.240331 }, //07 July 2010, UT1-UTC=-0.0563313, Accumulated Leap Seconds=34
  { 2455385.50, 66.239821 }, //08 July 2010, UT1-UTC=-0.0558206, Accumulated Leap Seconds=34
  { 2455386.50, 66.239222 }, //09 July 2010, UT1-UTC=-0.0552222, Accumulated Leap Seconds=34
  { 2455387.50, 66.238605 }, //10 July 2010, UT1-UTC=-0.0546055, Accumulated Leap Seconds=34
  { 2455388.50, 66.238066 }, //11 July 2010, UT1-UTC=-0.0540657, Accumulated Leap Seconds=34
  { 2455389.50, 66.237671 }, //12 July 2010, UT1-UTC=-0.0536709, Accumulated Leap Seconds=34
  { 2455390.50, 66.237482 }, //13 July 2010, UT1-UTC=-0.0534821, Accumulated Leap Seconds=34
  { 2455391.50, 66.237544 }, //14 July 2010, UT1-UTC=-0.0535439, Accumulated Leap Seconds=34
  { 2455392.50, 66.237765 }, //15 July 2010, UT1-UTC=-0.0537651, Accumulated Leap Seconds=34
  { 2455393.50, 66.238068 }, //16 July 2010, UT1-UTC=-0.0540685, Accumulated Leap Seconds=34
  { 2455394.50, 66.238389 }, //17 July 2010, UT1-UTC=-0.0543890, Accumulated Leap Seconds=34
  { 2455395.50, 66.238576 }, //18 July 2010, UT1-UTC=-0.0545765, Accumulated Leap Seconds=34
  { 2455396.50, 66.238628 }, //19 July 2010, UT1-UTC=-0.0546280, Accumulated Leap Seconds=34
  { 2455397.50, 66.238429 }, //20 July 2010, UT1-UTC=-0.0544292, Accumulated Leap Seconds=34
  { 2455398.50, 66.238012 }, //21 July 2010, UT1-UTC=-0.0540123, Accumulated Leap Seconds=34
  { 2455399.50, 66.237377 }, //22 July 2010, UT1-UTC=-0.0533767, Accumulated Leap Seconds=34
  { 2455400.50, 66.236603 }, //23 July 2010, UT1-UTC=-0.0526032, Accumulated Leap Seconds=34
  { 2455401.50, 66.235877 }, //24 July 2010, UT1-UTC=-0.0518769, Accumulated Leap Seconds=34
  { 2455402.50, 66.235168 }, //25 July 2010, UT1-UTC=-0.0511681, Accumulated Leap Seconds=34
  { 2455403.50, 66.234588 }, //26 July 2010, UT1-UTC=-0.0505881, Accumulated Leap Seconds=34
  { 2455404.50, 66.234194 }, //27 July 2010, UT1-UTC=-0.0501943, Accumulated Leap Seconds=34
  { 2455405.50, 66.233977 }, //28 July 2010, UT1-UTC=-0.0499766, Accumulated Leap Seconds=34
  { 2455406.50, 66.233863 }, //29 July 2010, UT1-UTC=-0.0498633, Accumulated Leap Seconds=34
  { 2455407.50, 66.233779 }, //30 July 2010, UT1-UTC=-0.0497795, Accumulated Leap Seconds=34
  { 2455408.50, 66.233695 }, //31 July 2010, UT1-UTC=-0.0496946, Accumulated Leap Seconds=34
  { 2455409.50, 66.233542 }, //01 August 2010, UT1-UTC=-0.0495423, Accumulated Leap Seconds=34
  { 2455410.50, 66.233293 }, //02 August 2010, UT1-UTC=-0.0492932, Accumulated Leap Seconds=34
  { 2455411.50, 66.232951 }, //03 August 2010, UT1-UTC=-0.0489515, Accumulated Leap Seconds=34
  { 2455412.50, 66.232559 }, //04 August 2010, UT1-UTC=-0.0485588, Accumulated Leap Seconds=34
  { 2455413.50, 66.232122 }, //05 August 2010, UT1-UTC=-0.0481220, Accumulated Leap Seconds=34
  { 2455414.50, 66.231680 }, //06 August 2010, UT1-UTC=-0.0476799, Accumulated Leap Seconds=34
  { 2455415.50, 66.231278 }, //07 August 2010, UT1-UTC=-0.0472782, Accumulated Leap Seconds=34
  { 2455416.50, 66.231034 }, //08 August 2010, UT1-UTC=-0.0470343, Accumulated Leap Seconds=34
  { 2455417.50, 66.231043 }, //09 August 2010, UT1-UTC=-0.0470428, Accumulated Leap Seconds=34
  { 2455418.50, 66.231350 }, //10 August 2010, UT1-UTC=-0.0473497, Accumulated Leap Seconds=34
  { 2455419.50, 66.231949 }, //11 August 2010, UT1-UTC=-0.0479491, Accumulated Leap Seconds=34
  { 2455420.50, 66.232700 }, //12 August 2010, UT1-UTC=-0.0487004, Accumulated Leap Seconds=34
  { 2455421.50, 66.233431 }, //13 August 2010, UT1-UTC=-0.0494313, Accumulated Leap Seconds=34
  { 2455422.50, 66.234052 }, //14 August 2010, UT1-UTC=-0.0500523, Accumulated Leap Seconds=34
  { 2455423.50, 66.234458 }, //15 August 2010, UT1-UTC=-0.0504581, Accumulated Leap Seconds=34
  { 2455424.50, 66.234607 }, //16 August 2010, UT1-UTC=-0.0506069, Accumulated Leap Seconds=34
  { 2455425.50, 66.234533 }, //17 August 2010, UT1-UTC=-0.0505327, Accumulated Leap Seconds=34
  { 2455426.50, 66.234245 }, //18 August 2010, UT1-UTC=-0.0502453, Accumulated Leap Seconds=34
  { 2455427.50, 66.233850 }, //19 August 2010, UT1-UTC=-0.0498500, Accumulated Leap Seconds=34
  { 2455428.50, 66.233468 }, //20 August 2010, UT1-UTC=-0.0494675, Accumulated Leap Seconds=34
  { 2455429.50, 66.233140 }, //21 August 2010, UT1-UTC=-0.0491402, Accumulated Leap Seconds=34
  { 2455430.50, 66.232922 }, //22 August 2010, UT1-UTC=-0.0489221, Accumulated Leap Seconds=34
  { 2455431.50, 66.232900 }, //23 August 2010, UT1-UTC=-0.0489005, Accumulated Leap Seconds=34
  { 2455432.50, 66.233039 }, //24 August 2010, UT1-UTC=-0.0490389, Accumulated Leap Seconds=34
  { 2455433.50, 66.233263 }, //25 August 2010, UT1-UTC=-0.0492630, Accumulated Leap Seconds=34
  { 2455434.50, 66.233602 }, //26 August 2010, UT1-UTC=-0.0496017, Accumulated Leap Seconds=34
  { 2455435.50, 66.234004 }, //27 August 2010, UT1-UTC=-0.0500041, Accumulated Leap Seconds=34
  { 2455436.50, 66.234340 }, //28 August 2010, UT1-UTC=-0.0503404, Accumulated Leap Seconds=34
  { 2455437.50, 66.234617 }, //29 August 2010, UT1-UTC=-0.0506168, Accumulated Leap Seconds=34
  { 2455438.50, 66.234808 }, //30 August 2010, UT1-UTC=-0.0508076, Accumulated Leap Seconds=34
  { 2455439.50, 66.234902 }, //31 August 2010, UT1-UTC=-0.0509022, Accumulated Leap Seconds=34
  { 2455440.50, 66.234911 }, //01 September 2010, UT1-UTC=-0.0509107, Accumulated Leap Seconds=34
  { 2455441.50, 66.234825 }, //02 September 2010, UT1-UTC=-0.0508249, Accumulated Leap Seconds=34
  { 2455442.50, 66.234738 }, //03 September 2010, UT1-UTC=-0.0507385, Accumulated Leap Seconds=34
  { 2455443.50, 66.234710 }, //04 September 2010, UT1-UTC=-0.0507099, Accumulated Leap Seconds=34
  { 2455444.50, 66.234841 }, //05 September 2010, UT1-UTC=-0.0508409, Accumulated Leap Seconds=34
  { 2455445.50, 66.235231 }, //06 September 2010, UT1-UTC=-0.0512306, Accumulated Leap Seconds=34
  { 2455446.50, 66.235918 }, //07 September 2010, UT1-UTC=-0.0519176, Accumulated Leap Seconds=34
  { 2455447.50, 66.236850 }, //08 September 2010, UT1-UTC=-0.0528496, Accumulated Leap Seconds=34
  { 2455448.50, 66.237928 }, //09 September 2010, UT1-UTC=-0.0539279, Accumulated Leap Seconds=34
  { 2455449.50, 66.238966 }, //10 September 2010, UT1-UTC=-0.0549659, Accumulated Leap Seconds=34
  { 2455450.50, 66.239764 }, //11 September 2010, UT1-UTC=-0.0557636, Accumulated Leap Seconds=34
  { 2455451.50, 66.240296 }, //12 September 2010, UT1-UTC=-0.0562956, Accumulated Leap Seconds=34
  { 2455452.50, 66.240550 }, //13 September 2010, UT1-UTC=-0.0565504, Accumulated Leap Seconds=34
  { 2455453.50, 66.240555 }, //14 September 2010, UT1-UTC=-0.0565547, Accumulated Leap Seconds=34
  { 2455454.50, 66.240429 }, //15 September 2010, UT1-UTC=-0.0564287, Accumulated Leap Seconds=34
  { 2455455.50, 66.240295 }, //16 September 2010, UT1-UTC=-0.0562948, Accumulated Leap Seconds=34
  { 2455456.50, 66.240185 }, //17 September 2010, UT1-UTC=-0.0561851, Accumulated Leap Seconds=34
  { 2455457.50, 66.240175 }, //18 September 2010, UT1-UTC=-0.0561754, Accumulated Leap Seconds=34
  { 2455458.50, 66.240306 }, //19 September 2010, UT1-UTC=-0.0563058, Accumulated Leap Seconds=34
  { 2455459.50, 66.240562 }, //20 September 2010, UT1-UTC=-0.0565618, Accumulated Leap Seconds=34
  { 2455460.50, 66.240912 }, //21 September 2010, UT1-UTC=-0.0569117, Accumulated Leap Seconds=34
  { 2455461.50, 66.241303 }, //22 September 2010, UT1-UTC=-0.0573028, Accumulated Leap Seconds=34
  { 2455462.50, 66.241760 }, //23 September 2010, UT1-UTC=-0.0577600, Accumulated Leap Seconds=34
  { 2455463.50, 66.242270 }, //24 September 2010, UT1-UTC=-0.0582699, Accumulated Leap Seconds=34
  { 2455464.50, 66.242724 }, //25 September 2010, UT1-UTC=-0.0587245, Accumulated Leap Seconds=34
  { 2455465.50, 66.243157 }, //26 September 2010, UT1-UTC=-0.0591569, Accumulated Leap Seconds=34
  { 2455466.50, 66.243488 }, //27 September 2010, UT1-UTC=-0.0594877, Accumulated Leap Seconds=34
  { 2455467.50, 66.243681 }, //28 September 2010, UT1-UTC=-0.0596806, Accumulated Leap Seconds=34
  { 2455468.50, 66.243792 }, //29 September 2010, UT1-UTC=-0.0597925, Accumulated Leap Seconds=34
  { 2455469.50, 66.243891 }, //30 September 2010, UT1-UTC=-0.0598906, Accumulated Leap Seconds=34
  { 2455470.50, 66.244109 }, //01 October 2010, UT1-UTC=-0.0601095, Accumulated Leap Seconds=34
  { 2455471.50, 66.244452 }, //02 October 2010, UT1-UTC=-0.0604520, Accumulated Leap Seconds=34
  { 2455472.50, 66.245093 }, //03 October 2010, UT1-UTC=-0.0610928, Accumulated Leap Seconds=34
  { 2455473.50, 66.246128 }, //04 October 2010, UT1-UTC=-0.0621279, Accumulated Leap Seconds=34
  { 2455474.50, 66.247487 }, //05 October 2010, UT1-UTC=-0.0634867, Accumulated Leap Seconds=34
  { 2455475.50, 66.249099 }, //06 October 2010, UT1-UTC=-0.0650995, Accumulated Leap Seconds=34
  { 2455476.50, 66.250808 }, //07 October 2010, UT1-UTC=-0.0668083, Accumulated Leap Seconds=34
  { 2455477.50, 66.252434 }, //08 October 2010, UT1-UTC=-0.0684337, Accumulated Leap Seconds=34
  { 2455478.50, 66.253847 }, //09 October 2010, UT1-UTC=-0.0698473, Accumulated Leap Seconds=34
  { 2455479.50, 66.255041 }, //10 October 2010, UT1-UTC=-0.0710410, Accumulated Leap Seconds=34
  { 2455480.50, 66.256012 }, //11 October 2010, UT1-UTC=-0.0720120, Accumulated Leap Seconds=34
  { 2455481.50, 66.256833 }, //12 October 2010, UT1-UTC=-0.0728332, Accumulated Leap Seconds=34
  { 2455482.50, 66.257611 }, //13 October 2010, UT1-UTC=-0.0736113, Accumulated Leap Seconds=34
  { 2455483.50, 66.258373 }, //14 October 2010, UT1-UTC=-0.0743731, Accumulated Leap Seconds=34
  { 2455484.50, 66.259218 }, //15 October 2010, UT1-UTC=-0.0752176, Accumulated Leap Seconds=34
  { 2455485.50, 66.260108 }, //16 October 2010, UT1-UTC=-0.0761076, Accumulated Leap Seconds=34
  { 2455486.50, 66.261084 }, //17 October 2010, UT1-UTC=-0.0770841, Accumulated Leap Seconds=34
  { 2455487.50, 66.262177 }, //18 October 2010, UT1-UTC=-0.0781773, Accumulated Leap Seconds=34
  { 2455488.50, 66.263346 }, //19 October 2010, UT1-UTC=-0.0793457, Accumulated Leap Seconds=34
  { 2455489.50, 66.264591 }, //20 October 2010, UT1-UTC=-0.0805907, Accumulated Leap Seconds=34
  { 2455490.50, 66.265813 }, //21 October 2010, UT1-UTC=-0.0818131, Accumulated Leap Seconds=34
  { 2455491.50, 66.266943 }, //22 October 2010, UT1-UTC=-0.0829431, Accumulated Leap Seconds=34
  { 2455492.50, 66.267955 }, //23 October 2010, UT1-UTC=-0.0839553, Accumulated Leap Seconds=34
  { 2455493.50, 66.268807 }, //24 October 2010, UT1-UTC=-0.0848073, Accumulated Leap Seconds=34
  { 2455494.50, 66.269519 }, //25 October 2010, UT1-UTC=-0.0855185, Accumulated Leap Seconds=34
  { 2455495.50, 66.270120 }, //26 October 2010, UT1-UTC=-0.0861200, Accumulated Leap Seconds=34
  { 2455496.50, 66.270661 }, //27 October 2010, UT1-UTC=-0.0866612, Accumulated Leap Seconds=34
  { 2455497.50, 66.271193 }, //28 October 2010, UT1-UTC=-0.0871935, Accumulated Leap Seconds=34
  { 2455498.50, 66.271830 }, //29 October 2010, UT1-UTC=-0.0878296, Accumulated Leap Seconds=34
  { 2455499.50, 66.272654 }, //30 October 2010, UT1-UTC=-0.0886538, Accumulated Leap Seconds=34
  { 2455500.50, 66.273736 }, //31 October 2010, UT1-UTC=-0.0897362, Accumulated Leap Seconds=34
  { 2455501.50, 66.275112 }, //01 November 2010, UT1-UTC=-0.0911123, Accumulated Leap Seconds=34
  { 2455502.50, 66.276715 }, //02 November 2010, UT1-UTC=-0.0927155, Accumulated Leap Seconds=34
  { 2455503.50, 66.278454 }, //03 November 2010, UT1-UTC=-0.0944538, Accumulated Leap Seconds=34
  { 2455504.50, 66.280154 }, //04 November 2010, UT1-UTC=-0.0961543, Accumulated Leap Seconds=34
  { 2455505.50, 66.281713 }, //05 November 2010, UT1-UTC=-0.0977133, Accumulated Leap Seconds=34
  { 2455506.50, 66.283068 }, //06 November 2010, UT1-UTC=-0.0990679, Accumulated Leap Seconds=34
  { 2455507.50, 66.284154 }, //07 November 2010, UT1-UTC=-0.1001542, Accumulated Leap Seconds=34
  { 2455508.50, 66.284979 }, //08 November 2010, UT1-UTC=-0.1009786, Accumulated Leap Seconds=34
  { 2455509.50, 66.285671 }, //09 November 2010, UT1-UTC=-0.1016712, Accumulated Leap Seconds=34
  { 2455510.50, 66.286256 }, //10 November 2010, UT1-UTC=-0.1022564, Accumulated Leap Seconds=34
  { 2455511.50, 66.286845 }, //11 November 2010, UT1-UTC=-0.1028446, Accumulated Leap Seconds=34
  { 2455512.50, 66.287528 }, //12 November 2010, UT1-UTC=-0.1035280, Accumulated Leap Seconds=34
  { 2455513.50, 66.288345 }, //13 November 2010, UT1-UTC=-0.1043448, Accumulated Leap Seconds=34
  { 2455514.50, 66.289325 }, //14 November 2010, UT1-UTC=-0.1053251, Accumulated Leap Seconds=34
  { 2455515.50, 66.290422 }, //15 November 2010, UT1-UTC=-0.1064222, Accumulated Leap Seconds=34
  { 2455516.50, 66.291587 }, //16 November 2010, UT1-UTC=-0.1075873, Accumulated Leap Seconds=34
  { 2455517.50, 66.292771 }, //17 November 2010, UT1-UTC=-0.1087714, Accumulated Leap Seconds=34
  { 2455518.50, 66.293877 }, //18 November 2010, UT1-UTC=-0.1098769, Accumulated Leap Seconds=34
  { 2455519.50, 66.294896 }, //19 November 2010, UT1-UTC=-0.1108959, Accumulated Leap Seconds=34
  { 2455520.50, 66.295736 }, //20 November 2010, UT1-UTC=-0.1117357, Accumulated Leap Seconds=34
  { 2455521.50, 66.296411 }, //21 November 2010, UT1-UTC=-0.1124108, Accumulated Leap Seconds=34
  { 2455522.50, 66.296971 }, //22 November 2010, UT1-UTC=-0.1129715, Accumulated Leap Seconds=34
  { 2455523.50, 66.297459 }, //23 November 2010, UT1-UTC=-0.1134586, Accumulated Leap Seconds=34
  { 2455524.50, 66.297951 }, //24 November 2010, UT1-UTC=-0.1139513, Accumulated Leap Seconds=34
  { 2455525.50, 66.298507 }, //25 November 2010, UT1-UTC=-0.1145072, Accumulated Leap Seconds=34
  { 2455526.50, 66.299206 }, //26 November 2010, UT1-UTC=-0.1152063, Accumulated Leap Seconds=34
  { 2455527.50, 66.300132 }, //27 November 2010, UT1-UTC=-0.1161318, Accumulated Leap Seconds=34
  { 2455528.50, 66.301321 }, //28 November 2010, UT1-UTC=-0.1173210, Accumulated Leap Seconds=34
  { 2455529.50, 66.302681 }, //29 November 2010, UT1-UTC=-0.1186808, Accumulated Leap Seconds=34
  { 2455530.50, 66.304064 }, //30 November 2010, UT1-UTC=-0.1200639, Accumulated Leap Seconds=34
  { 2455531.50, 66.305433 }, //01 December 2010, UT1-UTC=-0.1214334, Accumulated Leap Seconds=34
  { 2455532.50, 66.306668 }, //02 December 2010, UT1-UTC=-0.1226680, Accumulated Leap Seconds=34
  { 2455533.50, 66.307665 }, //03 December 2010, UT1-UTC=-0.1236645, Accumulated Leap Seconds=34
  { 2455534.50, 66.308405 }, //04 December 2010, UT1-UTC=-0.1244053, Accumulated Leap Seconds=34
  { 2455535.50, 66.308931 }, //05 December 2010, UT1-UTC=-0.1249308, Accumulated Leap Seconds=34
  { 2455536.50, 66.309362 }, //06 December 2010, UT1-UTC=-0.1253618, Accumulated Leap Seconds=34
  { 2455537.50, 66.309777 }, //07 December 2010, UT1-UTC=-0.1257772, Accumulated Leap Seconds=34
  { 2455538.50, 66.310247 }, //08 December 2010, UT1-UTC=-0.1262471, Accumulated Leap Seconds=34
  { 2455539.50, 66.310798 }, //09 December 2010, UT1-UTC=-0.1267983, Accumulated Leap Seconds=34
  { 2455540.50, 66.311468 }, //10 December 2010, UT1-UTC=-0.1274676, Accumulated Leap Seconds=34
  { 2455541.50, 66.312185 }, //11 December 2010, UT1-UTC=-0.1281845, Accumulated Leap Seconds=34
  { 2455542.50, 66.312905 }, //12 December 2010, UT1-UTC=-0.1289045, Accumulated Leap Seconds=34
  { 2455543.50, 66.313618 }, //13 December 2010, UT1-UTC=-0.1296181, Accumulated Leap Seconds=34
  { 2455544.50, 66.314344 }, //14 December 2010, UT1-UTC=-0.1303443, Accumulated Leap Seconds=34
  { 2455545.50, 66.314985 }, //15 December 2010, UT1-UTC=-0.1309848, Accumulated Leap Seconds=34
  { 2455546.50, 66.315537 }, //16 December 2010, UT1-UTC=-0.1315368, Accumulated Leap Seconds=34
  { 2455547.50, 66.316037 }, //17 December 2010, UT1-UTC=-0.1320367, Accumulated Leap Seconds=34
  { 2455548.50, 66.316416 }, //18 December 2010, UT1-UTC=-0.1324163, Accumulated Leap Seconds=34
  { 2455549.50, 66.316690 }, //19 December 2010, UT1-UTC=-0.1326901, Accumulated Leap Seconds=34
  { 2455550.50, 66.316824 }, //20 December 2010, UT1-UTC=-0.1328237, Accumulated Leap Seconds=34
  { 2455551.50, 66.316898 }, //21 December 2010, UT1-UTC=-0.1328980, Accumulated Leap Seconds=34
  { 2455552.50, 66.317098 }, //22 December 2010, UT1-UTC=-0.1330978, Accumulated Leap Seconds=34
  { 2455553.50, 66.317441 }, //23 December 2010, UT1-UTC=-0.1334410, Accumulated Leap Seconds=34
  { 2455554.50, 66.317993 }, //24 December 2010, UT1-UTC=-0.1339930, Accumulated Leap Seconds=34
  { 2455555.50, 66.318751 }, //25 December 2010, UT1-UTC=-0.1347515, Accumulated Leap Seconds=34
  { 2455556.50, 66.319706 }, //26 December 2010, UT1-UTC=-0.1357064, Accumulated Leap Seconds=34
  { 2455557.50, 66.320822 }, //27 December 2010, UT1-UTC=-0.1368218, Accumulated Leap Seconds=34
  { 2455558.50, 66.321940 }, //28 December 2010, UT1-UTC=-0.1379396, Accumulated Leap Seconds=34
  { 2455559.50, 66.322930 }, //29 December 2010, UT1-UTC=-0.1389297, Accumulated Leap Seconds=34
  { 2455560.50, 66.323653 }, //30 December 2010, UT1-UTC=-0.1396535, Accumulated Leap Seconds=34
  { 2455561.50, 66.324224 }, //31 December 2010, UT1-UTC=-0.1402235, Accumulated Leap Seconds=34
  { 2455562.50, 66.324557 }, //01 January 2011, UT1-UTC=-0.1405568, Accumulated Leap Seconds=34
  { 2455563.50, 66.324785 }, //02 January 2011, UT1-UTC=-0.1407850, Accumulated Leap Seconds=34
  { 2455564.50, 66.324970 }, //03 January 2011, UT1-UTC=-0.1409701, Accumulated Leap Seconds=34
  { 2455565.50, 66.325132 }, //04 January 2011, UT1-UTC=-0.1411325, Accumulated Leap Seconds=34
  { 2455566.50, 66.325341 }, //05 January 2011, UT1-UTC=-0.1413415, Accumulated Leap Seconds=34
  { 2455567.50, 66.325639 }, //06 January 2011, UT1-UTC=-0.1416387, Accumulated Leap Seconds=34
  { 2455568.50, 66.326066 }, //07 January 2011, UT1-UTC=-0.1420664, Accumulated Leap Seconds=34
  { 2455569.50, 66.326575 }, //08 January 2011, UT1-UTC=-0.1425749, Accumulated Leap Seconds=34
  { 2455570.50, 66.327127 }, //09 January 2011, UT1-UTC=-0.1431271, Accumulated Leap Seconds=34
  { 2455571.50, 66.327732 }, //10 January 2011, UT1-UTC=-0.1437316, Accumulated Leap Seconds=34
  { 2455572.50, 66.328327 }, //11 January 2011, UT1-UTC=-0.1443269, Accumulated Leap Seconds=34
  { 2455573.50, 66.328907 }, //12 January 2011, UT1-UTC=-0.1449070, Accumulated Leap Seconds=34
  { 2455574.50, 66.329462 }, //13 January 2011, UT1-UTC=-0.1454623, Accumulated Leap Seconds=34
  { 2455575.50, 66.329950 }, //14 January 2011, UT1-UTC=-0.1459496, Accumulated Leap Seconds=34
  { 2455576.50, 66.330323 }, //15 January 2011, UT1-UTC=-0.1463227, Accumulated Leap Seconds=34
  { 2455577.50, 66.330633 }, //16 January 2011, UT1-UTC=-0.1466326, Accumulated Leap Seconds=34
  { 2455578.50, 66.330896 }, //17 January 2011, UT1-UTC=-0.1468964, Accumulated Leap Seconds=34
  { 2455579.50, 66.331177 }, //18 January 2011, UT1-UTC=-0.1471767, Accumulated Leap Seconds=34
  { 2455580.50, 66.331560 }, //19 January 2011, UT1-UTC=-0.1475601, Accumulated Leap Seconds=34
  { 2455581.50, 66.332076 }, //20 January 2011, UT1-UTC=-0.1480758, Accumulated Leap Seconds=34
  { 2455582.50, 66.332800 }, //21 January 2011, UT1-UTC=-0.1488003, Accumulated Leap Seconds=34
  { 2455583.50, 66.333707 }, //22 January 2011, UT1-UTC=-0.1497068, Accumulated Leap Seconds=34
  { 2455584.50, 66.334817 }, //23 January 2011, UT1-UTC=-0.1508170, Accumulated Leap Seconds=34
  { 2455585.50, 66.335998 }, //24 January 2011, UT1-UTC=-0.1519977, Accumulated Leap Seconds=34
  { 2455586.50, 66.337088 }, //25 January 2011, UT1-UTC=-0.1530882, Accumulated Leap Seconds=34
  { 2455587.50, 66.338013 }, //26 January 2011, UT1-UTC=-0.1540134, Accumulated Leap Seconds=34
  { 2455588.50, 66.338725 }, //27 January 2011, UT1-UTC=-0.1547247, Accumulated Leap Seconds=34
  { 2455589.50, 66.339224 }, //28 January 2011, UT1-UTC=-0.1552242, Accumulated Leap Seconds=34
  { 2455590.50, 66.339582 }, //29 January 2011, UT1-UTC=-0.1555816, Accumulated Leap Seconds=34
  { 2455591.50, 66.339891 }, //30 January 2011, UT1-UTC=-0.1558913, Accumulated Leap Seconds=34
  { 2455592.50, 66.340204 }, //31 January 2011, UT1-UTC=-0.1562042, Accumulated Leap Seconds=34
  { 2455593.50, 66.340571 }, //01 February 2011, UT1-UTC=-0.1565713, Accumulated Leap Seconds=34
  { 2455594.50, 66.341039 }, //02 February 2011, UT1-UTC=-0.1570388, Accumulated Leap Seconds=34
  { 2455595.50, 66.341684 }, //03 February 2011, UT1-UTC=-0.1576845, Accumulated Leap Seconds=34
  { 2455596.50, 66.342418 }, //04 February 2011, UT1-UTC=-0.1584176, Accumulated Leap Seconds=34
  { 2455597.50, 66.343240 }, //05 February 2011, UT1-UTC=-0.1592403, Accumulated Leap Seconds=34
  { 2455598.50, 66.343998 }, //06 February 2011, UT1-UTC=-0.1599975, Accumulated Leap Seconds=34
  { 2455599.50, 66.344757 }, //07 February 2011, UT1-UTC=-0.1607568, Accumulated Leap Seconds=34
  { 2455600.50, 66.345460 }, //08 February 2011, UT1-UTC=-0.1614604, Accumulated Leap Seconds=34
  { 2455601.50, 66.346080 }, //09 February 2011, UT1-UTC=-0.1620804, Accumulated Leap Seconds=34
  { 2455602.50, 66.346624 }, //10 February 2011, UT1-UTC=-0.1626243, Accumulated Leap Seconds=34
  { 2455603.50, 66.347087 }, //11 February 2011, UT1-UTC=-0.1630870, Accumulated Leap Seconds=34
  { 2455604.50, 66.347470 }, //12 February 2011, UT1-UTC=-0.1634699, Accumulated Leap Seconds=34
  { 2455605.50, 66.347815 }, //13 February 2011, UT1-UTC=-0.1638153, Accumulated Leap Seconds=34
  { 2455606.50, 66.348176 }, //14 February 2011, UT1-UTC=-0.1641760, Accumulated Leap Seconds=34
  { 2455607.50, 66.348659 }, //15 February 2011, UT1-UTC=-0.1646592, Accumulated Leap Seconds=34
  { 2455608.50, 66.349365 }, //16 February 2011, UT1-UTC=-0.1653654, Accumulated Leap Seconds=34
  { 2455609.50, 66.350349 }, //17 February 2011, UT1-UTC=-0.1663492, Accumulated Leap Seconds=34
  { 2455610.50, 66.351653 }, //18 February 2011, UT1-UTC=-0.1676529, Accumulated Leap Seconds=34
  { 2455611.50, 66.353142 }, //19 February 2011, UT1-UTC=-0.1691421, Accumulated Leap Seconds=34
  { 2455612.50, 66.354763 }, //20 February 2011, UT1-UTC=-0.1707627, Accumulated Leap Seconds=34
  { 2455613.50, 66.356391 }, //21 February 2011, UT1-UTC=-0.1723914, Accumulated Leap Seconds=34
  { 2455614.50, 66.357827 }, //22 February 2011, UT1-UTC=-0.1738274, Accumulated Leap Seconds=34
  { 2455615.50, 66.358998 }, //23 February 2011, UT1-UTC=-0.1749979, Accumulated Leap Seconds=34
  { 2455616.50, 66.359847 }, //24 February 2011, UT1-UTC=-0.1758466, Accumulated Leap Seconds=34
  { 2455617.50, 66.360475 }, //25 February 2011, UT1-UTC=-0.1764754, Accumulated Leap Seconds=34
  { 2455618.50, 66.360951 }, //26 February 2011, UT1-UTC=-0.1769511, Accumulated Leap Seconds=34
  { 2455619.50, 66.361393 }, //27 February 2011, UT1-UTC=-0.1773927, Accumulated Leap Seconds=34
  { 2455620.50, 66.361836 }, //28 February 2011, UT1-UTC=-0.1778362, Accumulated Leap Seconds=34
  { 2455621.50, 66.362443 }, //01 March 2011, UT1-UTC=-0.1784433, Accumulated Leap Seconds=34
  { 2455622.50, 66.363192 }, //02 March 2011, UT1-UTC=-0.1791925, Accumulated Leap Seconds=34
  { 2455623.50, 66.364080 }, //03 March 2011, UT1-UTC=-0.1800797, Accumulated Leap Seconds=34
  { 2455624.50, 66.365156 }, //04 March 2011, UT1-UTC=-0.1811560, Accumulated Leap Seconds=34
  { 2455625.50, 66.366322 }, //05 March 2011, UT1-UTC=-0.1823223, Accumulated Leap Seconds=34
  { 2455626.50, 66.367533 }, //06 March 2011, UT1-UTC=-0.1835329, Accumulated Leap Seconds=34
  { 2455627.50, 66.368703 }, //07 March 2011, UT1-UTC=-0.1847033, Accumulated Leap Seconds=34
  { 2455628.50, 66.369749 }, //08 March 2011, UT1-UTC=-0.1857495, Accumulated Leap Seconds=34
  { 2455629.50, 66.370738 }, //09 March 2011, UT1-UTC=-0.1867381, Accumulated Leap Seconds=34
  { 2455630.50, 66.371623 }, //10 March 2011, UT1-UTC=-0.1876230, Accumulated Leap Seconds=34
  { 2455631.50, 66.372399 }, //11 March 2011, UT1-UTC=-0.1883986, Accumulated Leap Seconds=34
  { 2455632.50, 66.373121 }, //12 March 2011, UT1-UTC=-0.1891211, Accumulated Leap Seconds=34
  { 2455633.50, 66.373684 }, //13 March 2011, UT1-UTC=-0.1896843, Accumulated Leap Seconds=34
  { 2455634.50, 66.374292 }, //14 March 2011, UT1-UTC=-0.1902920, Accumulated Leap Seconds=34
  { 2455635.50, 66.375023 }, //15 March 2011, UT1-UTC=-0.1910232, Accumulated Leap Seconds=34
  { 2455636.50, 66.375995 }, //16 March 2011, UT1-UTC=-0.1919951, Accumulated Leap Seconds=34
  { 2455637.50, 66.377222 }, //17 March 2011, UT1-UTC=-0.1932222, Accumulated Leap Seconds=34
  { 2455638.50, 66.378652 }, //18 March 2011, UT1-UTC=-0.1946523, Accumulated Leap Seconds=34
  { 2455639.50, 66.380212 }, //19 March 2011, UT1-UTC=-0.1962118, Accumulated Leap Seconds=34
  { 2455640.50, 66.381805 }, //20 March 2011, UT1-UTC=-0.1978045, Accumulated Leap Seconds=34
  { 2455641.50, 66.383341 }, //21 March 2011, UT1-UTC=-0.1993411, Accumulated Leap Seconds=34
  { 2455642.50, 66.384722 }, //22 March 2011, UT1-UTC=-0.2007220, Accumulated Leap Seconds=34
  { 2455643.50, 66.385833 }, //23 March 2011, UT1-UTC=-0.2018334, Accumulated Leap Seconds=34
  { 2455644.50, 66.386757 }, //24 March 2011, UT1-UTC=-0.2027574, Accumulated Leap Seconds=34
  { 2455645.50, 66.387553 }, //25 March 2011, UT1-UTC=-0.2035533, Accumulated Leap Seconds=34
  { 2455646.50, 66.388432 }, //26 March 2011, UT1-UTC=-0.2044324, Accumulated Leap Seconds=34
  { 2455647.50, 66.389311 }, //27 March 2011, UT1-UTC=-0.2053112, Accumulated Leap Seconds=34
  { 2455648.50, 66.390333 }, //28 March 2011, UT1-UTC=-0.2063335, Accumulated Leap Seconds=34
  { 2455649.50, 66.391520 }, //29 March 2011, UT1-UTC=-0.2075201, Accumulated Leap Seconds=34
  { 2455650.50, 66.392848 }, //30 March 2011, UT1-UTC=-0.2088485, Accumulated Leap Seconds=34
  { 2455651.50, 66.394234 }, //31 March 2011, UT1-UTC=-0.2102342, Accumulated Leap Seconds=34
  { 2455652.50, 66.395710 }, //01 April 2011, UT1-UTC=-0.2117101, Accumulated Leap Seconds=34
  { 2455653.50, 66.397193 }, //02 April 2011, UT1-UTC=-0.2131933, Accumulated Leap Seconds=34
  { 2455654.50, 66.398686 }, //03 April 2011, UT1-UTC=-0.2146861, Accumulated Leap Seconds=34
  { 2455655.50, 66.400058 }, //04 April 2011, UT1-UTC=-0.2160584, Accumulated Leap Seconds=34
  { 2455656.50, 66.401272 }, //05 April 2011, UT1-UTC=-0.2172718, Accumulated Leap Seconds=34
  { 2455657.50, 66.402318 }, //06 April 2011, UT1-UTC=-0.2183180, Accumulated Leap Seconds=34
  { 2455658.50, 66.403231 }, //07 April 2011, UT1-UTC=-0.2192312, Accumulated Leap Seconds=34
  { 2455659.50, 66.403987 }, //08 April 2011, UT1-UTC=-0.2199870, Accumulated Leap Seconds=34
  { 2455660.50, 66.404727 }, //09 April 2011, UT1-UTC=-0.2207270, Accumulated Leap Seconds=34
  { 2455661.50, 66.405490 }, //10 April 2011, UT1-UTC=-0.2214895, Accumulated Leap Seconds=34
  { 2455662.50, 66.406337 }, //11 April 2011, UT1-UTC=-0.2223371, Accumulated Leap Seconds=34
  { 2455663.50, 66.407308 }, //12 April 2011, UT1-UTC=-0.2233083, Accumulated Leap Seconds=34
  { 2455664.50, 66.408479 }, //13 April 2011, UT1-UTC=-0.2244786, Accumulated Leap Seconds=34
  { 2455665.50, 66.409795 }, //14 April 2011, UT1-UTC=-0.2257947, Accumulated Leap Seconds=34
  { 2455666.50, 66.411301 }, //15 April 2011, UT1-UTC=-0.2273013, Accumulated Leap Seconds=34
  { 2455667.50, 66.412922 }, //16 April 2011, UT1-UTC=-0.2289220, Accumulated Leap Seconds=34
  { 2455668.50, 66.414580 }, //17 April 2011, UT1-UTC=-0.2305802, Accumulated Leap Seconds=34
  { 2455669.50, 66.416111 }, //18 April 2011, UT1-UTC=-0.2321108, Accumulated Leap Seconds=34
  { 2455670.50, 66.417454 }, //19 April 2011, UT1-UTC=-0.2334543, Accumulated Leap Seconds=34
  { 2455671.50, 66.418561 }, //20 April 2011, UT1-UTC=-0.2345609, Accumulated Leap Seconds=34
  { 2455672.50, 66.419449 }, //21 April 2011, UT1-UTC=-0.2354495, Accumulated Leap Seconds=34
  { 2455673.50, 66.420176 }, //22 April 2011, UT1-UTC=-0.2361763, Accumulated Leap Seconds=34
  { 2455674.50, 66.420889 }, //23 April 2011, UT1-UTC=-0.2368889, Accumulated Leap Seconds=34
  { 2455675.50, 66.421684 }, //24 April 2011, UT1-UTC=-0.2376842, Accumulated Leap Seconds=34
  { 2455676.50, 66.422557 }, //25 April 2011, UT1-UTC=-0.2385571, Accumulated Leap Seconds=34
  { 2455677.50, 66.423524 }, //26 April 2011, UT1-UTC=-0.2395237, Accumulated Leap Seconds=34
  { 2455678.50, 66.424562 }, //27 April 2011, UT1-UTC=-0.2405616, Accumulated Leap Seconds=34
  { 2455679.50, 66.425625 }, //28 April 2011, UT1-UTC=-0.2416246, Accumulated Leap Seconds=34
  { 2455680.50, 66.426769 }, //29 April 2011, UT1-UTC=-0.2427694, Accumulated Leap Seconds=34
  { 2455681.50, 66.427867 }, //30 April 2011, UT1-UTC=-0.2438674, Accumulated Leap Seconds=34
  { 2455682.50, 66.428903 }, //01 May 2011, UT1-UTC=-0.2449030, Accumulated Leap Seconds=34
  { 2455683.50, 66.429874 }, //02 May 2011, UT1-UTC=-0.2458739, Accumulated Leap Seconds=34
  { 2455684.50, 66.430843 }, //03 May 2011, UT1-UTC=-0.2468430, Accumulated Leap Seconds=34
  { 2455685.50, 66.431690 }, //04 May 2011, UT1-UTC=-0.2476898, Accumulated Leap Seconds=34
  { 2455686.50, 66.432572 }, //05 May 2011, UT1-UTC=-0.2485719, Accumulated Leap Seconds=34
  { 2455687.50, 66.433394 }, //06 May 2011, UT1-UTC=-0.2493942, Accumulated Leap Seconds=34
  { 2455688.50, 66.434214 }, //07 May 2011, UT1-UTC=-0.2502137, Accumulated Leap Seconds=34
  { 2455689.50, 66.435105 }, //08 May 2011, UT1-UTC=-0.2511055, Accumulated Leap Seconds=34
  { 2455690.50, 66.436117 }, //09 May 2011, UT1-UTC=-0.2521172, Accumulated Leap Seconds=34
  { 2455691.50, 66.437326 }, //10 May 2011, UT1-UTC=-0.2533262, Accumulated Leap Seconds=34
  { 2455692.50, 66.438814 }, //11 May 2011, UT1-UTC=-0.2548141, Accumulated Leap Seconds=34
  { 2455693.50, 66.440427 }, //12 May 2011, UT1-UTC=-0.2564269, Accumulated Leap Seconds=34
  { 2455694.50, 66.442167 }, //13 May 2011, UT1-UTC=-0.2581670, Accumulated Leap Seconds=34
  { 2455695.50, 66.443984 }, //14 May 2011, UT1-UTC=-0.2599844, Accumulated Leap Seconds=34
  { 2455696.50, 66.445808 }, //15 May 2011, UT1-UTC=-0.2618083, Accumulated Leap Seconds=34
  { 2455697.50, 66.447430 }, //16 May 2011, UT1-UTC=-0.2634301, Accumulated Leap Seconds=34
  { 2455698.50, 66.448839 }, //17 May 2011, UT1-UTC=-0.2648395, Accumulated Leap Seconds=34
  { 2455699.50, 66.449996 }, //18 May 2011, UT1-UTC=-0.2659958, Accumulated Leap Seconds=34
  { 2455700.50, 66.450991 }, //19 May 2011, UT1-UTC=-0.2669906, Accumulated Leap Seconds=34
  { 2455701.50, 66.451790 }, //20 May 2011, UT1-UTC=-0.2677904, Accumulated Leap Seconds=34
  { 2455702.50, 66.452564 }, //21 May 2011, UT1-UTC=-0.2685639, Accumulated Leap Seconds=34
  { 2455703.50, 66.453390 }, //22 May 2011, UT1-UTC=-0.2693903, Accumulated Leap Seconds=34
  { 2455704.50, 66.454204 }, //23 May 2011, UT1-UTC=-0.2702037, Accumulated Leap Seconds=34
  { 2455705.50, 66.455091 }, //24 May 2011, UT1-UTC=-0.2710909, Accumulated Leap Seconds=34
  { 2455706.50, 66.455973 }, //25 May 2011, UT1-UTC=-0.2719731, Accumulated Leap Seconds=34
  { 2455707.50, 66.456888 }, //26 May 2011, UT1-UTC=-0.2728881, Accumulated Leap Seconds=34
  { 2455708.50, 66.457902 }, //27 May 2011, UT1-UTC=-0.2739016, Accumulated Leap Seconds=34
  { 2455709.50, 66.458877 }, //28 May 2011, UT1-UTC=-0.2748769, Accumulated Leap Seconds=34
  { 2455710.50, 66.459777 }, //29 May 2011, UT1-UTC=-0.2757765, Accumulated Leap Seconds=34
  { 2455711.50, 66.460592 }, //30 May 2011, UT1-UTC=-0.2765919, Accumulated Leap Seconds=34
  { 2455712.50, 66.461265 }, //31 May 2011, UT1-UTC=-0.2772650, Accumulated Leap Seconds=34
  { 2455713.50, 66.461867 }, //01 June 2011, UT1-UTC=-0.2778675, Accumulated Leap Seconds=34
  { 2455714.50, 66.462411 }, //02 June 2011, UT1-UTC=-0.2784108, Accumulated Leap Seconds=34
  { 2455715.50, 66.462902 }, //03 June 2011, UT1-UTC=-0.2789021, Accumulated Leap Seconds=34
  { 2455716.50, 66.463405 }, //04 June 2011, UT1-UTC=-0.2794053, Accumulated Leap Seconds=34
  { 2455717.50, 66.464073 }, //05 June 2011, UT1-UTC=-0.2800725, Accumulated Leap Seconds=34
  { 2455718.50, 66.464891 }, //06 June 2011, UT1-UTC=-0.2808911, Accumulated Leap Seconds=34
  { 2455719.50, 66.465861 }, //07 June 2011, UT1-UTC=-0.2818610, Accumulated Leap Seconds=34
  { 2455720.50, 66.467000 }, //08 June 2011, UT1-UTC=-0.2830003, Accumulated Leap Seconds=34
  { 2455721.50, 66.468183 }, //09 June 2011, UT1-UTC=-0.2841834, Accumulated Leap Seconds=34
  { 2455722.50, 66.469307 }, //10 June 2011, UT1-UTC=-0.2853072, Accumulated Leap Seconds=34
  { 2455723.50, 66.470366 }, //11 June 2011, UT1-UTC=-0.2863664, Accumulated Leap Seconds=34
  { 2455724.50, 66.471278 }, //12 June 2011, UT1-UTC=-0.2872785, Accumulated Leap Seconds=34
  { 2455725.50, 66.471931 }, //13 June 2011, UT1-UTC=-0.2879310, Accumulated Leap Seconds=34
  { 2455726.50, 66.472434 }, //14 June 2011, UT1-UTC=-0.2884341, Accumulated Leap Seconds=34
  { 2455727.50, 66.472785 }, //15 June 2011, UT1-UTC=-0.2887848, Accumulated Leap Seconds=34
  { 2455728.50, 66.472977 }, //16 June 2011, UT1-UTC=-0.2889773, Accumulated Leap Seconds=34
  { 2455729.50, 66.473185 }, //17 June 2011, UT1-UTC=-0.2891847, Accumulated Leap Seconds=34
  { 2455730.50, 66.473421 }, //18 June 2011, UT1-UTC=-0.2894214, Accumulated Leap Seconds=34
  { 2455731.50, 66.473714 }, //19 June 2011, UT1-UTC=-0.2897142, Accumulated Leap Seconds=34
  { 2455732.50, 66.474069 }, //20 June 2011, UT1-UTC=-0.2900687, Accumulated Leap Seconds=34
  { 2455733.50, 66.474470 }, //21 June 2011, UT1-UTC=-0.2904697, Accumulated Leap Seconds=34
  { 2455734.50, 66.474877 }, //22 June 2011, UT1-UTC=-0.2908771, Accumulated Leap Seconds=34
  { 2455735.50, 66.475241 }, //23 June 2011, UT1-UTC=-0.2912407, Accumulated Leap Seconds=34
  { 2455736.50, 66.475519 }, //24 June 2011, UT1-UTC=-0.2915191, Accumulated Leap Seconds=34
  { 2455737.50, 66.475682 }, //25 June 2011, UT1-UTC=-0.2916821, Accumulated Leap Seconds=34
  { 2455738.50, 66.475739 }, //26 June 2011, UT1-UTC=-0.2917394, Accumulated Leap Seconds=34
  { 2455739.50, 66.475714 }, //27 June 2011, UT1-UTC=-0.2917139, Accumulated Leap Seconds=34
  { 2455740.50, 66.475610 }, //28 June 2011, UT1-UTC=-0.2916100, Accumulated Leap Seconds=34
  { 2455741.50, 66.475433 }, //29 June 2011, UT1-UTC=-0.2914334, Accumulated Leap Seconds=34
  { 2455742.50, 66.475168 }, //30 June 2011, UT1-UTC=-0.2911683, Accumulated Leap Seconds=34
  { 2455743.50, 66.474884 }, //01 July 2011, UT1-UTC=-0.2908837, Accumulated Leap Seconds=34
  { 2455744.50, 66.474658 }, //02 July 2011, UT1-UTC=-0.2906578, Accumulated Leap Seconds=34
  { 2455745.50, 66.474534 }, //03 July 2011, UT1-UTC=-0.2905343, Accumulated Leap Seconds=34
  { 2455746.50, 66.474687 }, //04 July 2011, UT1-UTC=-0.2906868, Accumulated Leap Seconds=34
  { 2455747.50, 66.475071 }, //05 July 2011, UT1-UTC=-0.2910715, Accumulated Leap Seconds=34
  { 2455748.50, 66.475566 }, //06 July 2011, UT1-UTC=-0.2915655, Accumulated Leap Seconds=34
  { 2455749.50, 66.476076 }, //07 July 2011, UT1-UTC=-0.2920759, Accumulated Leap Seconds=34
  { 2455750.50, 66.476486 }, //08 July 2011, UT1-UTC=-0.2924862, Accumulated Leap Seconds=34
  { 2455751.50, 66.476784 }, //09 July 2011, UT1-UTC=-0.2927840, Accumulated Leap Seconds=34
  { 2455752.50, 66.476853 }, //10 July 2011, UT1-UTC=-0.2928534, Accumulated Leap Seconds=34
  { 2455753.50, 66.476734 }, //11 July 2011, UT1-UTC=-0.2927340, Accumulated Leap Seconds=34
  { 2455754.50, 66.476487 }, //12 July 2011, UT1-UTC=-0.2924866, Accumulated Leap Seconds=34
  { 2455755.50, 66.476156 }, //13 July 2011, UT1-UTC=-0.2921557, Accumulated Leap Seconds=34
  { 2455756.50, 66.475834 }, //14 July 2011, UT1-UTC=-0.2918336, Accumulated Leap Seconds=34
  { 2455757.50, 66.475572 }, //15 July 2011, UT1-UTC=-0.2915718, Accumulated Leap Seconds=34
  { 2455758.50, 66.475385 }, //16 July 2011, UT1-UTC=-0.2913846, Accumulated Leap Seconds=34
  { 2455759.50, 66.475332 }, //17 July 2011, UT1-UTC=-0.2913317, Accumulated Leap Seconds=34
  { 2455760.50, 66.475421 }, //18 July 2011, UT1-UTC=-0.2914214, Accumulated Leap Seconds=34
  { 2455761.50, 66.475591 }, //19 July 2011, UT1-UTC=-0.2915911, Accumulated Leap Seconds=34
  { 2455762.50, 66.475861 }, //20 July 2011, UT1-UTC=-0.2918613, Accumulated Leap Seconds=34
  { 2455763.50, 66.476027 }, //21 July 2011, UT1-UTC=-0.2920271, Accumulated Leap Seconds=34
  { 2455764.50, 66.476153 }, //22 July 2011, UT1-UTC=-0.2921533, Accumulated Leap Seconds=34
  { 2455765.50, 66.476207 }, //23 July 2011, UT1-UTC=-0.2922072, Accumulated Leap Seconds=34
  { 2455766.50, 66.476192 }, //24 July 2011, UT1-UTC=-0.2921918, Accumulated Leap Seconds=34
  { 2455767.50, 66.476085 }, //25 July 2011, UT1-UTC=-0.2920855, Accumulated Leap Seconds=34
  { 2455768.50, 66.475866 }, //26 July 2011, UT1-UTC=-0.2918658, Accumulated Leap Seconds=34
  { 2455769.50, 66.475560 }, //27 July 2011, UT1-UTC=-0.2915599, Accumulated Leap Seconds=34
  { 2455770.50, 66.475213 }, //28 July 2011, UT1-UTC=-0.2912130, Accumulated Leap Seconds=34
  { 2455771.50, 66.474928 }, //29 July 2011, UT1-UTC=-0.2909276, Accumulated Leap Seconds=34
  { 2455772.50, 66.474735 }, //30 July 2011, UT1-UTC=-0.2907354, Accumulated Leap Seconds=34
  { 2455773.50, 66.474795 }, //31 July 2011, UT1-UTC=-0.2907947, Accumulated Leap Seconds=34
  { 2455774.50, 66.475128 }, //01 August 2011, UT1-UTC=-0.2911281, Accumulated Leap Seconds=34
  { 2455775.50, 66.475651 }, //02 August 2011, UT1-UTC=-0.2916507, Accumulated Leap Seconds=34
  { 2455776.50, 66.476314 }, //03 August 2011, UT1-UTC=-0.2923144, Accumulated Leap Seconds=34
  { 2455777.50, 66.477000 }, //04 August 2011, UT1-UTC=-0.2929997, Accumulated Leap Seconds=34
  { 2455778.50, 66.477595 }, //05 August 2011, UT1-UTC=-0.2935954, Accumulated Leap Seconds=34
  { 2455779.50, 66.478003 }, //06 August 2011, UT1-UTC=-0.2940030, Accumulated Leap Seconds=34
  { 2455780.50, 66.478150 }, //07 August 2011, UT1-UTC=-0.2941502, Accumulated Leap Seconds=34
  { 2455781.50, 66.478103 }, //08 August 2011, UT1-UTC=-0.2941028, Accumulated Leap Seconds=34
  { 2455782.50, 66.477973 }, //09 August 2011, UT1-UTC=-0.2939728, Accumulated Leap Seconds=34
  { 2455783.50, 66.477789 }, //10 August 2011, UT1-UTC=-0.2937887, Accumulated Leap Seconds=34
  { 2455784.50, 66.477621 }, //11 August 2011, UT1-UTC=-0.2936214, Accumulated Leap Seconds=34
  { 2455785.50, 66.477486 }, //12 August 2011, UT1-UTC=-0.2934862, Accumulated Leap Seconds=34
  { 2455786.50, 66.477477 }, //13 August 2011, UT1-UTC=-0.2934767, Accumulated Leap Seconds=34
  { 2455787.50, 66.477579 }, //14 August 2011, UT1-UTC=-0.2935789, Accumulated Leap Seconds=34
  { 2455788.50, 66.477704 }, //15 August 2011, UT1-UTC=-0.2937044, Accumulated Leap Seconds=34
  { 2455789.50, 66.477933 }, //16 August 2011, UT1-UTC=-0.2939326, Accumulated Leap Seconds=34
  { 2455790.50, 66.478171 }, //17 August 2011, UT1-UTC=-0.2941710, Accumulated Leap Seconds=34
  { 2455791.50, 66.478435 }, //18 August 2011, UT1-UTC=-0.2944346, Accumulated Leap Seconds=34
  { 2455792.50, 66.478619 }, //19 August 2011, UT1-UTC=-0.2946193, Accumulated Leap Seconds=34
  { 2455793.50, 66.478688 }, //20 August 2011, UT1-UTC=-0.2946877, Accumulated Leap Seconds=34
  { 2455794.50, 66.478668 }, //21 August 2011, UT1-UTC=-0.2946678, Accumulated Leap Seconds=34
  { 2455795.50, 66.478563 }, //22 August 2011, UT1-UTC=-0.2945629, Accumulated Leap Seconds=34
  { 2455796.50, 66.478405 }, //23 August 2011, UT1-UTC=-0.2944054, Accumulated Leap Seconds=34
  { 2455797.50, 66.478212 }, //24 August 2011, UT1-UTC=-0.2942118, Accumulated Leap Seconds=34
  { 2455798.50, 66.478130 }, //25 August 2011, UT1-UTC=-0.2941299, Accumulated Leap Seconds=34
  { 2455799.50, 66.478206 }, //26 August 2011, UT1-UTC=-0.2942063, Accumulated Leap Seconds=34
  { 2455800.50, 66.478536 }, //27 August 2011, UT1-UTC=-0.2945359, Accumulated Leap Seconds=34
  { 2455801.50, 66.479107 }, //28 August 2011, UT1-UTC=-0.2951073, Accumulated Leap Seconds=34
  { 2455802.50, 66.479906 }, //29 August 2011, UT1-UTC=-0.2959055, Accumulated Leap Seconds=34
  { 2455803.50, 66.480896 }, //30 August 2011, UT1-UTC=-0.2968963, Accumulated Leap Seconds=34
  { 2455804.50, 66.481899 }, //31 August 2011, UT1-UTC=-0.2978993, Accumulated Leap Seconds=34
  { 2455805.50, 66.482868 }, //01 September 2011, UT1-UTC=-0.2988678, Accumulated Leap Seconds=34
  { 2455806.50, 66.483708 }, //02 September 2011, UT1-UTC=-0.2997080, Accumulated Leap Seconds=34
  { 2455807.50, 66.484298 }, //03 September 2011, UT1-UTC=-0.3002984, Accumulated Leap Seconds=34
  { 2455808.50, 66.484662 }, //04 September 2011, UT1-UTC=-0.3006615, Accumulated Leap Seconds=34
  { 2455809.50, 66.484888 }, //05 September 2011, UT1-UTC=-0.3008882, Accumulated Leap Seconds=34
  { 2455810.50, 66.485053 }, //06 September 2011, UT1-UTC=-0.3010531, Accumulated Leap Seconds=34
  { 2455811.50, 66.485219 }, //07 September 2011, UT1-UTC=-0.3012189, Accumulated Leap Seconds=34
  { 2455812.50, 66.485496 }, //08 September 2011, UT1-UTC=-0.3014957, Accumulated Leap Seconds=34
  { 2455813.50, 66.485869 }, //09 September 2011, UT1-UTC=-0.3018690, Accumulated Leap Seconds=34
  { 2455814.50, 66.486319 }, //10 September 2011, UT1-UTC=-0.3023195, Accumulated Leap Seconds=34
  { 2455815.50, 66.486836 }, //11 September 2011, UT1-UTC=-0.3028359, Accumulated Leap Seconds=34
  { 2455816.50, 66.487431 }, //12 September 2011, UT1-UTC=-0.3034312, Accumulated Leap Seconds=34
  { 2455817.50, 66.488087 }, //13 September 2011, UT1-UTC=-0.3040873, Accumulated Leap Seconds=34
  { 2455818.50, 66.488800 }, //14 September 2011, UT1-UTC=-0.3048003, Accumulated Leap Seconds=34
  { 2455819.50, 66.489520 }, //15 September 2011, UT1-UTC=-0.3055205, Accumulated Leap Seconds=34
  { 2455820.50, 66.490141 }, //16 September 2011, UT1-UTC=-0.3061407, Accumulated Leap Seconds=34
  { 2455821.50, 66.490653 }, //17 September 2011, UT1-UTC=-0.3066527, Accumulated Leap Seconds=34
  { 2455822.50, 66.491078 }, //18 September 2011, UT1-UTC=-0.3070784, Accumulated Leap Seconds=34
  { 2455823.50, 66.491491 }, //19 September 2011, UT1-UTC=-0.3074913, Accumulated Leap Seconds=34
  { 2455824.50, 66.491901 }, //20 September 2011, UT1-UTC=-0.3079013, Accumulated Leap Seconds=34
  { 2455825.50, 66.492398 }, //21 September 2011, UT1-UTC=-0.3083976, Accumulated Leap Seconds=34
  { 2455826.50, 66.493041 }, //22 September 2011, UT1-UTC=-0.3090412, Accumulated Leap Seconds=34
  { 2455827.50, 66.493905 }, //23 September 2011, UT1-UTC=-0.3099054, Accumulated Leap Seconds=34
  { 2455828.50, 66.495011 }, //24 September 2011, UT1-UTC=-0.3110111, Accumulated Leap Seconds=34
  { 2455829.50, 66.496403 }, //25 September 2011, UT1-UTC=-0.3124025, Accumulated Leap Seconds=34
  { 2455830.50, 66.498000 }, //26 September 2011, UT1-UTC=-0.3139997, Accumulated Leap Seconds=34
  { 2455831.50, 66.499758 }, //27 September 2011, UT1-UTC=-0.3157575, Accumulated Leap Seconds=34
  { 2455832.50, 66.501524 }, //28 September 2011, UT1-UTC=-0.3175236, Accumulated Leap Seconds=34
  { 2455833.50, 66.503136 }, //29 September 2011, UT1-UTC=-0.3191363, Accumulated Leap Seconds=34
  { 2455834.50, 66.504480 }, //30 September 2011, UT1-UTC=-0.3204805, Accumulated Leap Seconds=34
  { 2455835.50, 66.505617 }, //01 October 2011, UT1-UTC=-0.3216165, Accumulated Leap Seconds=34
  { 2455836.50, 66.506615 }, //02 October 2011, UT1-UTC=-0.3226155, Accumulated Leap Seconds=34
  { 2455837.50, 66.507531 }, //03 October 2011, UT1-UTC=-0.3235315, Accumulated Leap Seconds=34
  { 2455838.50, 66.508453 }, //04 October 2011, UT1-UTC=-0.3244529, Accumulated Leap Seconds=34
  { 2455839.50, 66.509463 }, //05 October 2011, UT1-UTC=-0.3254629, Accumulated Leap Seconds=34
  { 2455840.50, 66.510608 }, //06 October 2011, UT1-UTC=-0.3266080, Accumulated Leap Seconds=34
  { 2455841.50, 66.511888 }, //07 October 2011, UT1-UTC=-0.3278883, Accumulated Leap Seconds=34
  { 2455842.50, 66.513235 }, //08 October 2011, UT1-UTC=-0.3292353, Accumulated Leap Seconds=34
  { 2455843.50, 66.514627 }, //09 October 2011, UT1-UTC=-0.3306270, Accumulated Leap Seconds=34
  { 2455844.50, 66.516013 }, //10 October 2011, UT1-UTC=-0.3320125, Accumulated Leap Seconds=34
  { 2455845.50, 66.517360 }, //11 October 2011, UT1-UTC=-0.3333600, Accumulated Leap Seconds=34
  { 2455846.50, 66.518645 }, //12 October 2011, UT1-UTC=-0.3346448, Accumulated Leap Seconds=34
  { 2455847.50, 66.519813 }, //13 October 2011, UT1-UTC=-0.3358130, Accumulated Leap Seconds=34
  { 2455848.50, 66.520766 }, //14 October 2011, UT1-UTC=-0.3367665, Accumulated Leap Seconds=34
  { 2455849.50, 66.521538 }, //15 October 2011, UT1-UTC=-0.3375380, Accumulated Leap Seconds=34
  { 2455850.50, 66.522143 }, //16 October 2011, UT1-UTC=-0.3381432, Accumulated Leap Seconds=34
  { 2455851.50, 66.522653 }, //17 October 2011, UT1-UTC=-0.3386527, Accumulated Leap Seconds=34
  { 2455852.50, 66.523161 }, //18 October 2011, UT1-UTC=-0.3391610, Accumulated Leap Seconds=34
  { 2455853.50, 66.523726 }, //19 October 2011, UT1-UTC=-0.3397260, Accumulated Leap Seconds=34
  { 2455854.50, 66.524407 }, //20 October 2011, UT1-UTC=-0.3404070, Accumulated Leap Seconds=34
  { 2455855.50, 66.525273 }, //21 October 2011, UT1-UTC=-0.3412734, Accumulated Leap Seconds=34
  { 2455856.50, 66.526363 }, //22 October 2011, UT1-UTC=-0.3423628, Accumulated Leap Seconds=34
  { 2455857.50, 66.527677 }, //23 October 2011, UT1-UTC=-0.3436772, Accumulated Leap Seconds=34
  { 2455858.50, 66.529158 }, //24 October 2011, UT1-UTC=-0.3451585, Accumulated Leap Seconds=34
  { 2455859.50, 66.530746 }, //25 October 2011, UT1-UTC=-0.3467464, Accumulated Leap Seconds=34
  { 2455860.50, 66.532273 }, //26 October 2011, UT1-UTC=-0.3482733, Accumulated Leap Seconds=34
  { 2455861.50, 66.533641 }, //27 October 2011, UT1-UTC=-0.3496405, Accumulated Leap Seconds=34
  { 2455862.50, 66.534809 }, //28 October 2011, UT1-UTC=-0.3508091, Accumulated Leap Seconds=34
  { 2455863.50, 66.535767 }, //29 October 2011, UT1-UTC=-0.3517675, Accumulated Leap Seconds=34
  { 2455864.50, 66.536602 }, //30 October 2011, UT1-UTC=-0.3526018, Accumulated Leap Seconds=34
  { 2455865.50, 66.537411 }, //31 October 2011, UT1-UTC=-0.3534112, Accumulated Leap Seconds=34
  { 2455866.50, 66.538291 }, //01 November 2011, UT1-UTC=-0.3542912, Accumulated Leap Seconds=34
  { 2455867.50, 66.539295 }, //02 November 2011, UT1-UTC=-0.3552952, Accumulated Leap Seconds=34
  { 2455868.50, 66.540416 }, //03 November 2011, UT1-UTC=-0.3564158, Accumulated Leap Seconds=34
  { 2455869.50, 66.541667 }, //04 November 2011, UT1-UTC=-0.3576671, Accumulated Leap Seconds=34
  { 2455870.50, 66.543065 }, //05 November 2011, UT1-UTC=-0.3590652, Accumulated Leap Seconds=34
  { 2455871.50, 66.544481 }, //06 November 2011, UT1-UTC=-0.3604810, Accumulated Leap Seconds=34
  { 2455872.50, 66.545872 }, //07 November 2011, UT1-UTC=-0.3618717, Accumulated Leap Seconds=34
  { 2455873.50, 66.547204 }, //08 November 2011, UT1-UTC=-0.3632037, Accumulated Leap Seconds=34
  { 2455874.50, 66.548368 }, //09 November 2011, UT1-UTC=-0.3643678, Accumulated Leap Seconds=34
  { 2455875.50, 66.549356 }, //10 November 2011, UT1-UTC=-0.3653556, Accumulated Leap Seconds=34
  { 2455876.50, 66.550231 }, //11 November 2011, UT1-UTC=-0.3662311, Accumulated Leap Seconds=34
  { 2455877.50, 66.550958 }, //12 November 2011, UT1-UTC=-0.3669581, Accumulated Leap Seconds=34
  { 2455878.50, 66.551638 }, //13 November 2011, UT1-UTC=-0.3676382, Accumulated Leap Seconds=34
  { 2455879.50, 66.552295 }, //14 November 2011, UT1-UTC=-0.3682952, Accumulated Leap Seconds=34
  { 2455880.50, 66.553001 }, //15 November 2011, UT1-UTC=-0.3690015, Accumulated Leap Seconds=34
  { 2455881.50, 66.553801 }, //16 November 2011, UT1-UTC=-0.3698014, Accumulated Leap Seconds=34
  { 2455882.50, 66.554739 }, //17 November 2011, UT1-UTC=-0.3707393, Accumulated Leap Seconds=34
  { 2455883.50, 66.555872 }, //18 November 2011, UT1-UTC=-0.3718721, Accumulated Leap Seconds=34
  { 2455884.50, 66.557199 }, //19 November 2011, UT1-UTC=-0.3731986, Accumulated Leap Seconds=34
  { 2455885.50, 66.558704 }, //20 November 2011, UT1-UTC=-0.3747042, Accumulated Leap Seconds=34
  { 2455886.50, 66.560247 }, //21 November 2011, UT1-UTC=-0.3762473, Accumulated Leap Seconds=34
  { 2455887.50, 66.561735 }, //22 November 2011, UT1-UTC=-0.3777355, Accumulated Leap Seconds=34
  { 2455888.50, 66.563158 }, //23 November 2011, UT1-UTC=-0.3791582, Accumulated Leap Seconds=34
  { 2455889.50, 66.564401 }, //24 November 2011, UT1-UTC=-0.3804010, Accumulated Leap Seconds=34
  { 2455890.50, 66.565425 }, //25 November 2011, UT1-UTC=-0.3814251, Accumulated Leap Seconds=34
  { 2455891.50, 66.566276 }, //26 November 2011, UT1-UTC=-0.3822758, Accumulated Leap Seconds=34
  { 2455892.50, 66.567042 }, //27 November 2011, UT1-UTC=-0.3830416, Accumulated Leap Seconds=34
  { 2455893.50, 66.567803 }, //28 November 2011, UT1-UTC=-0.3838034, Accumulated Leap Seconds=34
  { 2455894.50, 66.568618 }, //29 November 2011, UT1-UTC=-0.3846179, Accumulated Leap Seconds=34
  { 2455895.50, 66.569529 }, //30 November 2011, UT1-UTC=-0.3855292, Accumulated Leap Seconds=34
  { 2455896.50, 66.570563 }, //01 December 2011, UT1-UTC=-0.3865628, Accumulated Leap Seconds=34
  { 2455897.50, 66.571725 }, //02 December 2011, UT1-UTC=-0.3877251, Accumulated Leap Seconds=34
  { 2455898.50, 66.572924 }, //03 December 2011, UT1-UTC=-0.3889243, Accumulated Leap Seconds=34
  { 2455899.50, 66.574014 }, //04 December 2011, UT1-UTC=-0.3900143, Accumulated Leap Seconds=34
  { 2455900.50, 66.575029 }, //05 December 2011, UT1-UTC=-0.3910293, Accumulated Leap Seconds=34
  { 2455901.50, 66.575934 }, //06 December 2011, UT1-UTC=-0.3919339, Accumulated Leap Seconds=34
  { 2455902.50, 66.576712 }, //07 December 2011, UT1-UTC=-0.3927123, Accumulated Leap Seconds=34
  { 2455903.50, 66.577350 }, //08 December 2011, UT1-UTC=-0.3933500, Accumulated Leap Seconds=34
  { 2455904.50, 66.577907 }, //09 December 2011, UT1-UTC=-0.3939074, Accumulated Leap Seconds=34
  { 2455905.50, 66.578457 }, //10 December 2011, UT1-UTC=-0.3944575, Accumulated Leap Seconds=34
  { 2455906.50, 66.579014 }, //11 December 2011, UT1-UTC=-0.3950136, Accumulated Leap Seconds=34
  { 2455907.50, 66.579595 }, //12 December 2011, UT1-UTC=-0.3955950, Accumulated Leap Seconds=34
  { 2455908.50, 66.580276 }, //13 December 2011, UT1-UTC=-0.3962761, Accumulated Leap Seconds=34
  { 2455909.50, 66.581123 }, //14 December 2011, UT1-UTC=-0.3971231, Accumulated Leap Seconds=34
  { 2455910.50, 66.582161 }, //15 December 2011, UT1-UTC=-0.3981611, Accumulated Leap Seconds=34
  { 2455911.50, 66.583388 }, //16 December 2011, UT1-UTC=-0.3993875, Accumulated Leap Seconds=34
  { 2455912.50, 66.584762 }, //17 December 2011, UT1-UTC=-0.4007618, Accumulated Leap Seconds=34
  { 2455913.50, 66.586232 }, //18 December 2011, UT1-UTC=-0.4022323, Accumulated Leap Seconds=34
  { 2455914.50, 66.587725 }, //19 December 2011, UT1-UTC=-0.4037255, Accumulated Leap Seconds=34
  { 2455915.50, 66.589117 }, //20 December 2011, UT1-UTC=-0.4051174, Accumulated Leap Seconds=34
  { 2455916.50, 66.590358 }, //21 December 2011, UT1-UTC=-0.4063582, Accumulated Leap Seconds=34
  { 2455917.50, 66.591392 }, //22 December 2011, UT1-UTC=-0.4073919, Accumulated Leap Seconds=34
  { 2455918.50, 66.592214 }, //23 December 2011, UT1-UTC=-0.4082143, Accumulated Leap Seconds=34
  { 2455919.50, 66.592944 }, //24 December 2011, UT1-UTC=-0.4089439, Accumulated Leap Seconds=34
  { 2455920.50, 66.593738 }, //25 December 2011, UT1-UTC=-0.4097384, Accumulated Leap Seconds=34
  { 2455921.50, 66.594696 }, //26 December 2011, UT1-UTC=-0.4106962, Accumulated Leap Seconds=34
  { 2455922.50, 66.595857 }, //27 December 2011, UT1-UTC=-0.4118570, Accumulated Leap Seconds=34
  { 2455923.50, 66.597201 }, //28 December 2011, UT1-UTC=-0.4132011, Accumulated Leap Seconds=34
  { 2455924.50, 66.598683 }, //29 December 2011, UT1-UTC=-0.4146832, Accumulated Leap Seconds=34
  { 2455925.50, 66.600201 }, //30 December 2011, UT1-UTC=-0.4162013, Accumulated Leap Seconds=34
  { 2455926.50, 66.601660 }, //31 December 2011, UT1-UTC=-0.4176597, Accumulated Leap Seconds=34
  { 2455927.50, 66.603020 }, //01 January 2012, UT1-UTC=-0.4190198, Accumulated Leap Seconds=34
  { 2455928.50, 66.604263 }, //02 January 2012, UT1-UTC=-0.4202632, Accumulated Leap Seconds=34
  { 2455929.50, 66.605396 }, //03 January 2012, UT1-UTC=-0.4213964, Accumulated Leap Seconds=34
  { 2455930.50, 66.606393 }, //04 January 2012, UT1-UTC=-0.4223926, Accumulated Leap Seconds=34
  { 2455931.50, 66.607270 }, //05 January 2012, UT1-UTC=-0.4232705, Accumulated Leap Seconds=34
  { 2455932.50, 66.608054 }, //06 January 2012, UT1-UTC=-0.4240541, Accumulated Leap Seconds=34
  { 2455933.50, 66.608777 }, //07 January 2012, UT1-UTC=-0.4247770, Accumulated Leap Seconds=34
  { 2455934.50, 66.609533 }, //08 January 2012, UT1-UTC=-0.4255331, Accumulated Leap Seconds=34
  { 2455935.50, 66.610363 }, //09 January 2012, UT1-UTC=-0.4263634, Accumulated Leap Seconds=34
  { 2455936.50, 66.611284 }, //10 January 2012, UT1-UTC=-0.4272844, Accumulated Leap Seconds=34
  { 2455937.50, 66.612360 }, //11 January 2012, UT1-UTC=-0.4283596, Accumulated Leap Seconds=34
  { 2455938.50, 66.613596 }, //12 January 2012, UT1-UTC=-0.4295961, Accumulated Leap Seconds=34
  { 2455939.50, 66.615004 }, //13 January 2012, UT1-UTC=-0.4310044, Accumulated Leap Seconds=34
  { 2455940.50, 66.616521 }, //14 January 2012, UT1-UTC=-0.4325210, Accumulated Leap Seconds=34
  { 2455941.50, 66.618008 }, //15 January 2012, UT1-UTC=-0.4340075, Accumulated Leap Seconds=34
  { 2455942.50, 66.619413 }, //16 January 2012, UT1-UTC=-0.4354132, Accumulated Leap Seconds=34
  { 2455943.50, 66.620670 }, //17 January 2012, UT1-UTC=-0.4366697, Accumulated Leap Seconds=34
  { 2455944.50, 66.621724 }, //18 January 2012, UT1-UTC=-0.4377243, Accumulated Leap Seconds=34
  { 2455945.50, 66.622578 }, //19 January 2012, UT1-UTC=-0.4385780, Accumulated Leap Seconds=34
  { 2455946.50, 66.623316 }, //20 January 2012, UT1-UTC=-0.4393155, Accumulated Leap Seconds=34
  { 2455947.50, 66.624001 }, //21 January 2012, UT1-UTC=-0.4400007, Accumulated Leap Seconds=34
  { 2455948.50, 66.624786 }, //22 January 2012, UT1-UTC=-0.4407855, Accumulated Leap Seconds=34
  { 2455949.50, 66.625699 }, //23 January 2012, UT1-UTC=-0.4416988, Accumulated Leap Seconds=34
  { 2455950.50, 66.626704 }, //24 January 2012, UT1-UTC=-0.4427038, Accumulated Leap Seconds=34
  { 2455951.50, 66.627795 }, //25 January 2012, UT1-UTC=-0.4437950, Accumulated Leap Seconds=34
  { 2455952.50, 66.628910 }, //26 January 2012, UT1-UTC=-0.4449104, Accumulated Leap Seconds=34
  { 2455953.50, 66.630024 }, //27 January 2012, UT1-UTC=-0.4460237, Accumulated Leap Seconds=34
  { 2455954.50, 66.631072 }, //28 January 2012, UT1-UTC=-0.4470718, Accumulated Leap Seconds=34
  { 2455955.50, 66.631982 }, //29 January 2012, UT1-UTC=-0.4479821, Accumulated Leap Seconds=34
  { 2455956.50, 66.632785 }, //30 January 2012, UT1-UTC=-0.4487852, Accumulated Leap Seconds=34
  { 2455957.50, 66.633465 }, //31 January 2012, UT1-UTC=-0.4494654, Accumulated Leap Seconds=34
  { 2455958.50, 66.633969 }, //01 February 2012, UT1-UTC=-0.4499689, Accumulated Leap Seconds=34
  { 2455959.50, 66.634384 }, //02 February 2012, UT1-UTC=-0.4503843, Accumulated Leap Seconds=34
  { 2455960.50, 66.634812 }, //03 February 2012, UT1-UTC=-0.4508117, Accumulated Leap Seconds=34
  { 2455961.50, 66.635259 }, //04 February 2012, UT1-UTC=-0.4512588, Accumulated Leap Seconds=34
  { 2455962.50, 66.635751 }, //05 February 2012, UT1-UTC=-0.4517515, Accumulated Leap Seconds=34
  { 2455963.50, 66.636395 }, //06 February 2012, UT1-UTC=-0.4523955, Accumulated Leap Seconds=34
  { 2455964.50, 66.637298 }, //07 February 2012, UT1-UTC=-0.4532978, Accumulated Leap Seconds=34
  { 2455965.50, 66.638405 }, //08 February 2012, UT1-UTC=-0.4544046, Accumulated Leap Seconds=34
  { 2455966.50, 66.639723 }, //09 February 2012, UT1-UTC=-0.4557230, Accumulated Leap Seconds=34
  { 2455967.50, 66.641160 }, //10 February 2012, UT1-UTC=-0.4571597, Accumulated Leap Seconds=34
  { 2455968.50, 66.642634 }, //11 February 2012, UT1-UTC=-0.4586340, Accumulated Leap Seconds=34
  { 2455969.50, 66.644038 }, //12 February 2012, UT1-UTC=-0.4600383, Accumulated Leap Seconds=34
  { 2455970.50, 66.645318 }, //13 February 2012, UT1-UTC=-0.4613177, Accumulated Leap Seconds=34
  { 2455971.50, 66.646375 }, //14 February 2012, UT1-UTC=-0.4623751, Accumulated Leap Seconds=34
  { 2455972.50, 66.647176 }, //15 February 2012, UT1-UTC=-0.4631761, Accumulated Leap Seconds=34
  { 2455973.50, 66.647787 }, //16 February 2012, UT1-UTC=-0.4637866, Accumulated Leap Seconds=34
  { 2455974.50, 66.648340 }, //17 February 2012, UT1-UTC=-0.4643397, Accumulated Leap Seconds=34
  { 2455975.50, 66.648901 }, //18 February 2012, UT1-UTC=-0.4649014, Accumulated Leap Seconds=34
  { 2455976.50, 66.649564 }, //19 February 2012, UT1-UTC=-0.4655638, Accumulated Leap Seconds=34
  { 2455977.50, 66.650350 }, //20 February 2012, UT1-UTC=-0.4663497, Accumulated Leap Seconds=34
  { 2455978.50, 66.651232 }, //21 February 2012, UT1-UTC=-0.4672324, Accumulated Leap Seconds=34
  { 2455979.50, 66.652152 }, //22 February 2012, UT1-UTC=-0.4681517, Accumulated Leap Seconds=34
  { 2455980.50, 66.653045 }, //23 February 2012, UT1-UTC=-0.4690449, Accumulated Leap Seconds=34
  { 2455981.50, 66.653884 }, //24 February 2012, UT1-UTC=-0.4698839, Accumulated Leap Seconds=34
  { 2455982.50, 66.654611 }, //25 February 2012, UT1-UTC=-0.4706106, Accumulated Leap Seconds=34
  { 2455983.50, 66.655269 }, //26 February 2012, UT1-UTC=-0.4712692, Accumulated Leap Seconds=34
  { 2455984.50, 66.655809 }, //27 February 2012, UT1-UTC=-0.4718090, Accumulated Leap Seconds=34
  { 2455985.50, 66.656269 }, //28 February 2012, UT1-UTC=-0.4722693, Accumulated Leap Seconds=34
  { 2455986.50, 66.656631 }, //29 February 2012, UT1-UTC=-0.4726308, Accumulated Leap Seconds=34
  { 2455987.50, 66.656912 }, //01 March 2012, UT1-UTC=-0.4729117, Accumulated Leap Seconds=34
  { 2455988.50, 66.657189 }, //02 March 2012, UT1-UTC=-0.4731888, Accumulated Leap Seconds=34
  { 2455989.50, 66.657543 }, //03 March 2012, UT1-UTC=-0.4735434, Accumulated Leap Seconds=34
  { 2455990.50, 66.658013 }, //04 March 2012, UT1-UTC=-0.4740126, Accumulated Leap Seconds=34
  { 2455991.50, 66.658632 }, //05 March 2012, UT1-UTC=-0.4746319, Accumulated Leap Seconds=34
  { 2455992.50, 66.659448 }, //06 March 2012, UT1-UTC=-0.4754483, Accumulated Leap Seconds=34
  { 2455993.50, 66.660478 }, //07 March 2012, UT1-UTC=-0.4764782, Accumulated Leap Seconds=34
  { 2455994.50, 66.661697 }, //08 March 2012, UT1-UTC=-0.4776966, Accumulated Leap Seconds=34
  { 2455995.50, 66.663072 }, //09 March 2012, UT1-UTC=-0.4790716, Accumulated Leap Seconds=34
  { 2455996.50, 66.664520 }, //10 March 2012, UT1-UTC=-0.4805201, Accumulated Leap Seconds=34
  { 2455997.50, 66.666009 }, //11 March 2012, UT1-UTC=-0.4820089, Accumulated Leap Seconds=34
  { 2455998.50, 66.667383 }, //12 March 2012, UT1-UTC=-0.4833830, Accumulated Leap Seconds=34
  { 2455999.50, 66.668612 }, //13 March 2012, UT1-UTC=-0.4846116, Accumulated Leap Seconds=34
  { 2456000.50, 66.669760 }, //14 March 2012, UT1-UTC=-0.4857596, Accumulated Leap Seconds=34
  { 2456001.50, 66.670874 }, //15 March 2012, UT1-UTC=-0.4868744, Accumulated Leap Seconds=34
  { 2456002.50, 66.672046 }, //16 March 2012, UT1-UTC=-0.4880460, Accumulated Leap Seconds=34
  { 2456003.50, 66.673362 }, //17 March 2012, UT1-UTC=-0.4893618, Accumulated Leap Seconds=34
  { 2456004.50, 66.674819 }, //18 March 2012, UT1-UTC=-0.4908192, Accumulated Leap Seconds=34
  { 2456005.50, 66.676341 }, //19 March 2012, UT1-UTC=-0.4923412, Accumulated Leap Seconds=34
  { 2456006.50, 66.677913 }, //20 March 2012, UT1-UTC=-0.4939128, Accumulated Leap Seconds=34
  { 2456007.50, 66.679503 }, //21 March 2012, UT1-UTC=-0.4955028, Accumulated Leap Seconds=34
  { 2456008.50, 66.681000 }, //22 March 2012, UT1-UTC=-0.4970001, Accumulated Leap Seconds=34
  { 2456009.50, 66.682400 }, //23 March 2012, UT1-UTC=-0.4984003, Accumulated Leap Seconds=34
  { 2456010.50, 66.683741 }, //24 March 2012, UT1-UTC=-0.4997414, Accumulated Leap Seconds=34
  { 2456011.50, 66.685002 }, //25 March 2012, UT1-UTC=-0.5010021, Accumulated Leap Seconds=34
  { 2456012.50, 66.686141 }, //26 March 2012, UT1-UTC=-0.5021407, Accumulated Leap Seconds=34
  { 2456013.50, 66.687198 }, //27 March 2012, UT1-UTC=-0.5031977, Accumulated Leap Seconds=34
  { 2456014.50, 66.688204 }, //28 March 2012, UT1-UTC=-0.5042038, Accumulated Leap Seconds=34
  { 2456015.50, 66.689184 }, //29 March 2012, UT1-UTC=-0.5051845, Accumulated Leap Seconds=34
  { 2456016.50, 66.690196 }, //30 March 2012, UT1-UTC=-0.5061958, Accumulated Leap Seconds=34
  { 2456017.50, 66.691281 }, //31 March 2012, UT1-UTC=-0.5072813, Accumulated Leap Seconds=34
  { 2456018.50, 66.692501 }, //01 April 2012, UT1-UTC=-0.5085011, Accumulated Leap Seconds=34
  { 2456019.50, 66.693882 }, //02 April 2012, UT1-UTC=-0.5098823, Accumulated Leap Seconds=34
  { 2456020.50, 66.695433 }, //03 April 2012, UT1-UTC=-0.5114332, Accumulated Leap Seconds=34
  { 2456021.50, 66.697165 }, //04 April 2012, UT1-UTC=-0.5131647, Accumulated Leap Seconds=34
  { 2456022.50, 66.698984 }, //05 April 2012, UT1-UTC=-0.5149838, Accumulated Leap Seconds=34
  { 2456023.50, 66.700876 }, //06 April 2012, UT1-UTC=-0.5168757, Accumulated Leap Seconds=34
  { 2456024.50, 66.702652 }, //07 April 2012, UT1-UTC=-0.5186518, Accumulated Leap Seconds=34
  { 2456025.50, 66.704195 }, //08 April 2012, UT1-UTC=-0.5201950, Accumulated Leap Seconds=34
  { 2456026.50, 66.705524 }, //09 April 2012, UT1-UTC=-0.5215237, Accumulated Leap Seconds=34
  { 2456027.50, 66.706664 }, //10 April 2012, UT1-UTC=-0.5226642, Accumulated Leap Seconds=34
  { 2456028.50, 66.707696 }, //11 April 2012, UT1-UTC=-0.5236960, Accumulated Leap Seconds=34
  { 2456029.50, 66.708731 }, //12 April 2012, UT1-UTC=-0.5247306, Accumulated Leap Seconds=34
  { 2456030.50, 66.709839 }, //13 April 2012, UT1-UTC=-0.5258385, Accumulated Leap Seconds=34
  { 2456031.50, 66.711044 }, //14 April 2012, UT1-UTC=-0.5270442, Accumulated Leap Seconds=34
  { 2456032.50, 66.712287 }, //15 April 2012, UT1-UTC=-0.5282869, Accumulated Leap Seconds=34
  { 2456033.50, 66.713563 }, //16 April 2012, UT1-UTC=-0.5295627, Accumulated Leap Seconds=34
  { 2456034.50, 66.714875 }, //17 April 2012, UT1-UTC=-0.5308746, Accumulated Leap Seconds=34
  { 2456035.50, 66.716206 }, //18 April 2012, UT1-UTC=-0.5322055, Accumulated Leap Seconds=34
  { 2456036.50, 66.717485 }, //19 April 2012, UT1-UTC=-0.5334851, Accumulated Leap Seconds=34
  { 2456037.50, 66.718722 }, //20 April 2012, UT1-UTC=-0.5347225, Accumulated Leap Seconds=34
  { 2456038.50, 66.719851 }, //21 April 2012, UT1-UTC=-0.5358514, Accumulated Leap Seconds=34
  { 2456039.50, 66.720877 }, //22 April 2012, UT1-UTC=-0.5368770, Accumulated Leap Seconds=34
  { 2456040.50, 66.721801 }, //23 April 2012, UT1-UTC=-0.5378007, Accumulated Leap Seconds=34
  { 2456041.50, 66.722631 }, //24 April 2012, UT1-UTC=-0.5386314, Accumulated Leap Seconds=34
  { 2456042.50, 66.723393 }, //25 April 2012, UT1-UTC=-0.5393933, Accumulated Leap Seconds=34
  { 2456043.50, 66.724148 }, //26 April 2012, UT1-UTC=-0.5401479, Accumulated Leap Seconds=34
  { 2456044.50, 66.724941 }, //27 April 2012, UT1-UTC=-0.5409412, Accumulated Leap Seconds=34
  { 2456045.50, 66.725791 }, //28 April 2012, UT1-UTC=-0.5417915, Accumulated Leap Seconds=34
  { 2456046.50, 66.726687 }, //29 April 2012, UT1-UTC=-0.5426867, Accumulated Leap Seconds=34
  { 2456047.50, 66.727688 }, //30 April 2012, UT1-UTC=-0.5436883, Accumulated Leap Seconds=34
  { 2456048.50, 66.728873 }, //01 May 2012, UT1-UTC=-0.5448729, Accumulated Leap Seconds=34
  { 2456049.50, 66.730231 }, //02 May 2012, UT1-UTC=-0.5462308, Accumulated Leap Seconds=34
  { 2456050.50, 66.731702 }, //03 May 2012, UT1-UTC=-0.5477020, Accumulated Leap Seconds=34
  { 2456051.50, 66.733184 }, //04 May 2012, UT1-UTC=-0.5491839, Accumulated Leap Seconds=34
  { 2456052.50, 66.734584 }, //05 May 2012, UT1-UTC=-0.5505838, Accumulated Leap Seconds=34
  { 2456053.50, 66.735863 }, //06 May 2012, UT1-UTC=-0.5518634, Accumulated Leap Seconds=34
  { 2456054.50, 66.736955 }, //07 May 2012, UT1-UTC=-0.5529546, Accumulated Leap Seconds=34
  { 2456055.50, 66.737874 }, //08 May 2012, UT1-UTC=-0.5538741, Accumulated Leap Seconds=34
  { 2456056.50, 66.738696 }, //09 May 2012, UT1-UTC=-0.5546964, Accumulated Leap Seconds=34
  { 2456057.50, 66.739565 }, //10 May 2012, UT1-UTC=-0.5555650, Accumulated Leap Seconds=34
  { 2456058.50, 66.740491 }, //11 May 2012, UT1-UTC=-0.5564914, Accumulated Leap Seconds=34
  { 2456059.50, 66.741460 }, //12 May 2012, UT1-UTC=-0.5574596, Accumulated Leap Seconds=34
  { 2456060.50, 66.742483 }, //13 May 2012, UT1-UTC=-0.5584826, Accumulated Leap Seconds=34
  { 2456061.50, 66.743543 }, //14 May 2012, UT1-UTC=-0.5595429, Accumulated Leap Seconds=34
  { 2456062.50, 66.744611 }, //15 May 2012, UT1-UTC=-0.5606109, Accumulated Leap Seconds=34
  { 2456063.50, 66.745696 }, //16 May 2012, UT1-UTC=-0.5616964, Accumulated Leap Seconds=34
  { 2456064.50, 66.746693 }, //17 May 2012, UT1-UTC=-0.5626929, Accumulated Leap Seconds=34
  { 2456065.50, 66.747599 }, //18 May 2012, UT1-UTC=-0.5635986, Accumulated Leap Seconds=34
  { 2456066.50, 66.748342 }, //19 May 2012, UT1-UTC=-0.5643424, Accumulated Leap Seconds=34
  { 2456067.50, 66.748931 }, //20 May 2012, UT1-UTC=-0.5649315, Accumulated Leap Seconds=34
  { 2456068.50, 66.749396 }, //21 May 2012, UT1-UTC=-0.5653962, Accumulated Leap Seconds=34
  { 2456069.50, 66.749794 }, //22 May 2012, UT1-UTC=-0.5657943, Accumulated Leap Seconds=34
  { 2456070.50, 66.750175 }, //23 May 2012, UT1-UTC=-0.5661755, Accumulated Leap Seconds=34
  { 2456071.50, 66.750597 }, //24 May 2012, UT1-UTC=-0.5665975, Accumulated Leap Seconds=34
  { 2456072.50, 66.751092 }, //25 May 2012, UT1-UTC=-0.5670919, Accumulated Leap Seconds=34
  { 2456073.50, 66.751704 }, //26 May 2012, UT1-UTC=-0.5677043, Accumulated Leap Seconds=34
  { 2456074.50, 66.752445 }, //27 May 2012, UT1-UTC=-0.5684455, Accumulated Leap Seconds=34
  { 2456075.50, 66.753319 }, //28 May 2012, UT1-UTC=-0.5693193, Accumulated Leap Seconds=34
  { 2456076.50, 66.754332 }, //29 May 2012, UT1-UTC=-0.5703324, Accumulated Leap Seconds=34
  { 2456077.50, 66.755458 }, //30 May 2012, UT1-UTC=-0.5714580, Accumulated Leap Seconds=34
  { 2456078.50, 66.756657 }, //31 May 2012, UT1-UTC=-0.5726569, Accumulated Leap Seconds=34
  { 2456079.50, 66.757872 }, //01 June 2012, UT1-UTC=-0.5738719, Accumulated Leap Seconds=34
  { 2456080.50, 66.759013 }, //02 June 2012, UT1-UTC=-0.5750134, Accumulated Leap Seconds=34
  { 2456081.50, 66.759996 }, //03 June 2012, UT1-UTC=-0.5759962, Accumulated Leap Seconds=34
  { 2456082.50, 66.760819 }, //04 June 2012, UT1-UTC=-0.5768192, Accumulated Leap Seconds=34
  { 2456083.50, 66.761452 }, //05 June 2012, UT1-UTC=-0.5774521, Accumulated Leap Seconds=34
  { 2456084.50, 66.762023 }, //06 June 2012, UT1-UTC=-0.5780231, Accumulated Leap Seconds=34
  { 2456085.50, 66.762674 }, //07 June 2012, UT1-UTC=-0.5786736, Accumulated Leap Seconds=34
  { 2456086.50, 66.763443 }, //08 June 2012, UT1-UTC=-0.5794428, Accumulated Leap Seconds=34
  { 2456087.50, 66.764344 }, //09 June 2012, UT1-UTC=-0.5803438, Accumulated Leap Seconds=34
  { 2456088.50, 66.765241 }, //10 June 2012, UT1-UTC=-0.5812409, Accumulated Leap Seconds=34
  { 2456089.50, 66.766064 }, //11 June 2012, UT1-UTC=-0.5820641, Accumulated Leap Seconds=34
  { 2456090.50, 66.766842 }, //12 June 2012, UT1-UTC=-0.5828421, Accumulated Leap Seconds=34
  { 2456091.50, 66.767533 }, //13 June 2012, UT1-UTC=-0.5835328, Accumulated Leap Seconds=34
  { 2456092.50, 66.768088 }, //14 June 2012, UT1-UTC=-0.5840883, Accumulated Leap Seconds=34
  { 2456093.50, 66.768496 }, //15 June 2012, UT1-UTC=-0.5844962, Accumulated Leap Seconds=34
  { 2456094.50, 66.768758 }, //16 June 2012, UT1-UTC=-0.5847581, Accumulated Leap Seconds=34
  { 2456095.50, 66.768935 }, //17 June 2012, UT1-UTC=-0.5849351, Accumulated Leap Seconds=34
  { 2456096.50, 66.768985 }, //18 June 2012, UT1-UTC=-0.5849851, Accumulated Leap Seconds=34
  { 2456097.50, 66.768952 }, //19 June 2012, UT1-UTC=-0.5849518, Accumulated Leap Seconds=34
  { 2456098.50, 66.768882 }, //20 June 2012, UT1-UTC=-0.5848824, Accumulated Leap Seconds=34
  { 2456099.50, 66.768800 }, //21 June 2012, UT1-UTC=-0.5847996, Accumulated Leap Seconds=34
  { 2456100.50, 66.768750 }, //22 June 2012, UT1-UTC=-0.5847501, Accumulated Leap Seconds=34
  { 2456101.50, 66.768799 }, //23 June 2012, UT1-UTC=-0.5847991, Accumulated Leap Seconds=34
  { 2456102.50, 66.769014 }, //24 June 2012, UT1-UTC=-0.5850136, Accumulated Leap Seconds=34
  { 2456103.50, 66.769325 }, //25 June 2012, UT1-UTC=-0.5853254, Accumulated Leap Seconds=34
  { 2456104.50, 66.769723 }, //26 June 2012, UT1-UTC=-0.5857231, Accumulated Leap Seconds=34
  { 2456105.50, 66.770164 }, //27 June 2012, UT1-UTC=-0.5861642, Accumulated Leap Seconds=34
  { 2456106.50, 66.770535 }, //28 June 2012, UT1-UTC=-0.5865347, Accumulated Leap Seconds=34
  { 2456107.50, 66.770762 }, //29 June 2012, UT1-UTC=-0.5867620, Accumulated Leap Seconds=34
  { 2456108.50, 66.770837 }, //30 June 2012, UT1-UTC=-0.5868367, Accumulated Leap Seconds=34
  { 2456109.50, 66.770763 }, //01 July 2012, UT1-UTC= 0.4132375, Accumulated Leap Seconds=35
  { 2456110.50, 66.770566 }, //02 July 2012, UT1-UTC= 0.4134344, Accumulated Leap Seconds=35
  { 2456111.50, 66.770334 }, //03 July 2012, UT1-UTC= 0.4136664, Accumulated Leap Seconds=35
  { 2456112.50, 66.770163 }, //04 July 2012, UT1-UTC= 0.4138372, Accumulated Leap Seconds=35
  { 2456113.50, 66.770060 }, //05 July 2012, UT1-UTC= 0.4139396, Accumulated Leap Seconds=35
  { 2456114.50, 66.770089 }, //06 July 2012, UT1-UTC= 0.4139105, Accumulated Leap Seconds=35
  { 2456115.50, 66.770278 }, //07 July 2012, UT1-UTC= 0.4137221, Accumulated Leap Seconds=35
  { 2456116.50, 66.770533 }, //08 July 2012, UT1-UTC= 0.4134669, Accumulated Leap Seconds=35
  { 2456117.50, 66.770839 }, //09 July 2012, UT1-UTC= 0.4131610, Accumulated Leap Seconds=35
  { 2456118.50, 66.771097 }, //10 July 2012, UT1-UTC= 0.4129032, Accumulated Leap Seconds=35
  { 2456119.50, 66.771272 }, //11 July 2012, UT1-UTC= 0.4127281, Accumulated Leap Seconds=35
  { 2456120.50, 66.771311 }, //12 July 2012, UT1-UTC= 0.4126893, Accumulated Leap Seconds=35
  { 2456121.50, 66.771128 }, //13 July 2012, UT1-UTC= 0.4128718, Accumulated Leap Seconds=35
  { 2456122.50, 66.770879 }, //14 July 2012, UT1-UTC= 0.4131207, Accumulated Leap Seconds=35
  { 2456123.50, 66.770581 }, //15 July 2012, UT1-UTC= 0.4134191, Accumulated Leap Seconds=35
  { 2456124.50, 66.770211 }, //16 July 2012, UT1-UTC= 0.4137890, Accumulated Leap Seconds=35
  { 2456125.50, 66.769932 }, //17 July 2012, UT1-UTC= 0.4140682, Accumulated Leap Seconds=35
  { 2456126.50, 66.769786 }, //18 July 2012, UT1-UTC= 0.4142135, Accumulated Leap Seconds=35
  { 2456127.50, 66.769750 }, //19 July 2012, UT1-UTC= 0.4142499, Accumulated Leap Seconds=35
  { 2456128.50, 66.769928 }, //20 July 2012, UT1-UTC= 0.4140718, Accumulated Leap Seconds=35
  { 2456129.50, 66.770267 }, //21 July 2012, UT1-UTC= 0.4137326, Accumulated Leap Seconds=35
  { 2456130.50, 66.770741 }, //22 July 2012, UT1-UTC= 0.4132590, Accumulated Leap Seconds=35
  { 2456131.50, 66.771296 }, //23 July 2012, UT1-UTC= 0.4127044, Accumulated Leap Seconds=35
  { 2456132.50, 66.771871 }, //24 July 2012, UT1-UTC= 0.4121292, Accumulated Leap Seconds=35
  { 2456133.50, 66.772418 }, //25 July 2012, UT1-UTC= 0.4115821, Accumulated Leap Seconds=35
  { 2456134.50, 66.772884 }, //26 July 2012, UT1-UTC= 0.4111164, Accumulated Leap Seconds=35
  { 2456135.50, 66.773214 }, //27 July 2012, UT1-UTC= 0.4107859, Accumulated Leap Seconds=35
  { 2456136.50, 66.773452 }, //28 July 2012, UT1-UTC= 0.4105478, Accumulated Leap Seconds=35
  { 2456137.50, 66.773580 }, //29 July 2012, UT1-UTC= 0.4104204, Accumulated Leap Seconds=35
  { 2456138.50, 66.773657 }, //30 July 2012, UT1-UTC= 0.4103435, Accumulated Leap Seconds=35
  { 2456139.50, 66.773788 }, //31 July 2012, UT1-UTC= 0.4102122, Accumulated Leap Seconds=35
  { 2456140.50, 66.774043 }, //01 August 2012, UT1-UTC= 0.4099573, Accumulated Leap Seconds=35
  { 2456141.50, 66.774430 }, //02 August 2012, UT1-UTC= 0.4095700, Accumulated Leap Seconds=35
  { 2456142.50, 66.774942 }, //03 August 2012, UT1-UTC= 0.4090577, Accumulated Leap Seconds=35
  { 2456143.50, 66.775484 }, //04 August 2012, UT1-UTC= 0.4085156, Accumulated Leap Seconds=35
  { 2456144.50, 66.776078 }, //05 August 2012, UT1-UTC= 0.4079222, Accumulated Leap Seconds=35
  { 2456145.50, 66.776633 }, //06 August 2012, UT1-UTC= 0.4073667, Accumulated Leap Seconds=35
  { 2456146.50, 66.777057 }, //07 August 2012, UT1-UTC= 0.4069433, Accumulated Leap Seconds=35
  { 2456147.50, 66.777309 }, //08 August 2012, UT1-UTC= 0.4066915, Accumulated Leap Seconds=35
  { 2456148.50, 66.777414 }, //09 August 2012, UT1-UTC= 0.4065860, Accumulated Leap Seconds=35
  { 2456149.50, 66.777405 }, //10 August 2012, UT1-UTC= 0.4065948, Accumulated Leap Seconds=35
  { 2456150.50, 66.777332 }, //11 August 2012, UT1-UTC= 0.4066675, Accumulated Leap Seconds=35
  { 2456151.50, 66.777181 }, //12 August 2012, UT1-UTC= 0.4068189, Accumulated Leap Seconds=35
  { 2456152.50, 66.776980 }, //13 August 2012, UT1-UTC= 0.4070203, Accumulated Leap Seconds=35
  { 2456153.50, 66.776820 }, //14 August 2012, UT1-UTC= 0.4071800, Accumulated Leap Seconds=35
  { 2456154.50, 66.776772 }, //15 August 2012, UT1-UTC= 0.4072279, Accumulated Leap Seconds=35
  { 2456155.50, 66.776877 }, //16 August 2012, UT1-UTC= 0.4071232, Accumulated Leap Seconds=35
  { 2456156.50, 66.777148 }, //17 August 2012, UT1-UTC= 0.4068515, Accumulated Leap Seconds=35
  { 2456157.50, 66.777563 }, //18 August 2012, UT1-UTC= 0.4064372, Accumulated Leap Seconds=35
  { 2456158.50, 66.778136 }, //19 August 2012, UT1-UTC= 0.4058643, Accumulated Leap Seconds=35
  { 2456159.50, 66.778814 }, //20 August 2012, UT1-UTC= 0.4051864, Accumulated Leap Seconds=35
  { 2456160.50, 66.779498 }, //21 August 2012, UT1-UTC= 0.4045017, Accumulated Leap Seconds=35
  { 2456161.50, 66.780103 }, //22 August 2012, UT1-UTC= 0.4038969, Accumulated Leap Seconds=35
  { 2456162.50, 66.780585 }, //23 August 2012, UT1-UTC= 0.4034152, Accumulated Leap Seconds=35
  { 2456163.50, 66.780897 }, //24 August 2012, UT1-UTC= 0.4031028, Accumulated Leap Seconds=35
  { 2456164.50, 66.781077 }, //25 August 2012, UT1-UTC= 0.4029231, Accumulated Leap Seconds=35
  { 2456165.50, 66.781202 }, //26 August 2012, UT1-UTC= 0.4027980, Accumulated Leap Seconds=35
  { 2456166.50, 66.781393 }, //27 August 2012, UT1-UTC= 0.4026068, Accumulated Leap Seconds=35
  { 2456167.50, 66.781741 }, //28 August 2012, UT1-UTC= 0.4022594, Accumulated Leap Seconds=35
  { 2456168.50, 66.782280 }, //29 August 2012, UT1-UTC= 0.4017200, Accumulated Leap Seconds=35
  { 2456169.50, 66.782973 }, //30 August 2012, UT1-UTC= 0.4010269, Accumulated Leap Seconds=35
  { 2456170.50, 66.783764 }, //31 August 2012, UT1-UTC= 0.4002356, Accumulated Leap Seconds=35
  { 2456171.50, 66.784629 }, //01 September 2012, UT1-UTC= 0.3993712, Accumulated Leap Seconds=35
  { 2456172.50, 66.785458 }, //02 September 2012, UT1-UTC= 0.3985418, Accumulated Leap Seconds=35
  { 2456173.50, 66.786263 }, //03 September 2012, UT1-UTC= 0.3977367, Accumulated Leap Seconds=35
  { 2456174.50, 66.786997 }, //04 September 2012, UT1-UTC= 0.3970028, Accumulated Leap Seconds=35
  { 2456175.50, 66.787571 }, //05 September 2012, UT1-UTC= 0.3964291, Accumulated Leap Seconds=35
  { 2456176.50, 66.787955 }, //06 September 2012, UT1-UTC= 0.3960454, Accumulated Leap Seconds=35
  { 2456177.50, 66.788262 }, //07 September 2012, UT1-UTC= 0.3957383, Accumulated Leap Seconds=35
  { 2456178.50, 66.788510 }, //08 September 2012, UT1-UTC= 0.3954904, Accumulated Leap Seconds=35
  { 2456179.50, 66.788699 }, //09 September 2012, UT1-UTC= 0.3953013, Accumulated Leap Seconds=35
  { 2456180.50, 66.788956 }, //10 September 2012, UT1-UTC= 0.3950435, Accumulated Leap Seconds=35
  { 2456181.50, 66.789341 }, //11 September 2012, UT1-UTC= 0.3946592, Accumulated Leap Seconds=35
  { 2456182.50, 66.789900 }, //12 September 2012, UT1-UTC= 0.3941003, Accumulated Leap Seconds=35
  { 2456183.50, 66.790643 }, //13 September 2012, UT1-UTC= 0.3933570, Accumulated Leap Seconds=35
  { 2456184.50, 66.791609 }, //14 September 2012, UT1-UTC= 0.3923912, Accumulated Leap Seconds=35
  { 2456185.50, 66.792843 }, //15 September 2012, UT1-UTC= 0.3911574, Accumulated Leap Seconds=35
  { 2456186.50, 66.794245 }, //16 September 2012, UT1-UTC= 0.3897550, Accumulated Leap Seconds=35
  { 2456187.50, 66.795660 }, //17 September 2012, UT1-UTC= 0.3883395, Accumulated Leap Seconds=35
  { 2456188.50, 66.797024 }, //18 September 2012, UT1-UTC= 0.3869764, Accumulated Leap Seconds=35
  { 2456189.50, 66.798233 }, //19 September 2012, UT1-UTC= 0.3857667, Accumulated Leap Seconds=35
  { 2456190.50, 66.799276 }, //20 September 2012, UT1-UTC= 0.3847240, Accumulated Leap Seconds=35
  { 2456191.50, 66.800109 }, //21 September 2012, UT1-UTC= 0.3838911, Accumulated Leap Seconds=35
  { 2456192.50, 66.800767 }, //22 September 2012, UT1-UTC= 0.3832327, Accumulated Leap Seconds=35
  { 2456193.50, 66.801418 }, //23 September 2012, UT1-UTC= 0.3825820, Accumulated Leap Seconds=35
  { 2456194.50, 66.802132 }, //24 September 2012, UT1-UTC= 0.3818681, Accumulated Leap Seconds=35
  { 2456195.50, 66.802975 }, //25 September 2012, UT1-UTC= 0.3810252, Accumulated Leap Seconds=35
  { 2456196.50, 66.803990 }, //26 September 2012, UT1-UTC= 0.3800104, Accumulated Leap Seconds=35
  { 2456197.50, 66.805197 }, //27 September 2012, UT1-UTC= 0.3788030, Accumulated Leap Seconds=35
  { 2456198.50, 66.806534 }, //28 September 2012, UT1-UTC= 0.3774658, Accumulated Leap Seconds=35
  { 2456199.50, 66.807873 }, //29 September 2012, UT1-UTC= 0.3761275, Accumulated Leap Seconds=35
  { 2456200.50, 66.809165 }, //30 September 2012, UT1-UTC= 0.3748350, Accumulated Leap Seconds=35
  { 2456201.50, 66.810324 }, //01 October 2012, UT1-UTC= 0.3736760, Accumulated Leap Seconds=35
  { 2456202.50, 66.811308 }, //02 October 2012, UT1-UTC= 0.3726921, Accumulated Leap Seconds=35
  { 2456203.50, 66.812090 }, //03 October 2012, UT1-UTC= 0.3719102, Accumulated Leap Seconds=35
  { 2456204.50, 66.812754 }, //04 October 2012, UT1-UTC= 0.3712461, Accumulated Leap Seconds=35
  { 2456205.50, 66.813328 }, //05 October 2012, UT1-UTC= 0.3706723, Accumulated Leap Seconds=35
  { 2456206.50, 66.813837 }, //06 October 2012, UT1-UTC= 0.3701625, Accumulated Leap Seconds=35
  { 2456207.50, 66.814331 }, //07 October 2012, UT1-UTC= 0.3696688, Accumulated Leap Seconds=35
  { 2456208.50, 66.814886 }, //08 October 2012, UT1-UTC= 0.3691137, Accumulated Leap Seconds=35
  { 2456209.50, 66.815582 }, //09 October 2012, UT1-UTC= 0.3684180, Accumulated Leap Seconds=35
  { 2456210.50, 66.816447 }, //10 October 2012, UT1-UTC= 0.3675530, Accumulated Leap Seconds=35
  { 2456211.50, 66.817510 }, //11 October 2012, UT1-UTC= 0.3664901, Accumulated Leap Seconds=35
  { 2456212.50, 66.818739 }, //12 October 2012, UT1-UTC= 0.3652605, Accumulated Leap Seconds=35
  { 2456213.50, 66.820114 }, //13 October 2012, UT1-UTC= 0.3638862, Accumulated Leap Seconds=35
  { 2456214.50, 66.821601 }, //14 October 2012, UT1-UTC= 0.3623986, Accumulated Leap Seconds=35
  { 2456215.50, 66.823099 }, //15 October 2012, UT1-UTC= 0.3609005, Accumulated Leap Seconds=35
  { 2456216.50, 66.824514 }, //16 October 2012, UT1-UTC= 0.3594861, Accumulated Leap Seconds=35
  { 2456217.50, 66.825772 }, //17 October 2012, UT1-UTC= 0.3582281, Accumulated Leap Seconds=35
  { 2456218.50, 66.826836 }, //18 October 2012, UT1-UTC= 0.3571637, Accumulated Leap Seconds=35
  { 2456219.50, 66.827729 }, //19 October 2012, UT1-UTC= 0.3562709, Accumulated Leap Seconds=35
  { 2456220.50, 66.828509 }, //20 October 2012, UT1-UTC= 0.3554906, Accumulated Leap Seconds=35
  { 2456221.50, 66.829268 }, //21 October 2012, UT1-UTC= 0.3547319, Accumulated Leap Seconds=35
  { 2456222.50, 66.830088 }, //22 October 2012, UT1-UTC= 0.3539122, Accumulated Leap Seconds=35
  { 2456223.50, 66.830983 }, //23 October 2012, UT1-UTC= 0.3530166, Accumulated Leap Seconds=35
  { 2456224.50, 66.831985 }, //24 October 2012, UT1-UTC= 0.3520151, Accumulated Leap Seconds=35
  { 2456225.50, 66.833077 }, //25 October 2012, UT1-UTC= 0.3509232, Accumulated Leap Seconds=35
  { 2456226.50, 66.834228 }, //26 October 2012, UT1-UTC= 0.3497723, Accumulated Leap Seconds=35
  { 2456227.50, 66.835395 }, //27 October 2012, UT1-UTC= 0.3486052, Accumulated Leap Seconds=35
  { 2456228.50, 66.836506 }, //28 October 2012, UT1-UTC= 0.3474935, Accumulated Leap Seconds=35
  { 2456229.50, 66.837523 }, //29 October 2012, UT1-UTC= 0.3464766, Accumulated Leap Seconds=35
  { 2456230.50, 66.838444 }, //30 October 2012, UT1-UTC= 0.3455560, Accumulated Leap Seconds=35
  { 2456231.50, 66.839290 }, //31 October 2012, UT1-UTC= 0.3447103, Accumulated Leap Seconds=35
  { 2456232.50, 66.840048 }, //01 November 2012, UT1-UTC= 0.3439520, Accumulated Leap Seconds=35
  { 2456233.50, 66.840771 }, //02 November 2012, UT1-UTC= 0.3432288, Accumulated Leap Seconds=35
  { 2456234.50, 66.841458 }, //03 November 2012, UT1-UTC= 0.3425421, Accumulated Leap Seconds=35
  { 2456235.50, 66.842193 }, //04 November 2012, UT1-UTC= 0.3418066, Accumulated Leap Seconds=35
  { 2456236.50, 66.843040 }, //05 November 2012, UT1-UTC= 0.3409600, Accumulated Leap Seconds=35
  { 2456237.50, 66.843997 }, //06 November 2012, UT1-UTC= 0.3400028, Accumulated Leap Seconds=35
  { 2456238.50, 66.845052 }, //07 November 2012, UT1-UTC= 0.3389475, Accumulated Leap Seconds=35
  { 2456239.50, 66.846268 }, //08 November 2012, UT1-UTC= 0.3377324, Accumulated Leap Seconds=35
  { 2456240.50, 66.847619 }, //09 November 2012, UT1-UTC= 0.3363812, Accumulated Leap Seconds=35
  { 2456241.50, 66.849017 }, //10 November 2012, UT1-UTC= 0.3349834, Accumulated Leap Seconds=35
  { 2456242.50, 66.850466 }, //11 November 2012, UT1-UTC= 0.3335335, Accumulated Leap Seconds=35
  { 2456243.50, 66.851972 }, //12 November 2012, UT1-UTC= 0.3320284, Accumulated Leap Seconds=35
  { 2456244.50, 66.853467 }, //13 November 2012, UT1-UTC= 0.3305328, Accumulated Leap Seconds=35
  { 2456245.50, 66.854865 }, //14 November 2012, UT1-UTC= 0.3291351, Accumulated Leap Seconds=35
  { 2456246.50, 66.856181 }, //15 November 2012, UT1-UTC= 0.3278191, Accumulated Leap Seconds=35
  { 2456247.50, 66.857477 }, //16 November 2012, UT1-UTC= 0.3265233, Accumulated Leap Seconds=35
  { 2456248.50, 66.858780 }, //17 November 2012, UT1-UTC= 0.3252203, Accumulated Leap Seconds=35
  { 2456249.50, 66.860205 }, //18 November 2012, UT1-UTC= 0.3237945, Accumulated Leap Seconds=35
  { 2456250.50, 66.861812 }, //19 November 2012, UT1-UTC= 0.3221881, Accumulated Leap Seconds=35
  { 2456251.50, 66.863527 }, //20 November 2012, UT1-UTC= 0.3204726, Accumulated Leap Seconds=35
  { 2456252.50, 66.865267 }, //21 November 2012, UT1-UTC= 0.3187333, Accumulated Leap Seconds=35
  { 2456253.50, 66.866985 }, //22 November 2012, UT1-UTC= 0.3170146, Accumulated Leap Seconds=35
  { 2456254.50, 66.868653 }, //23 November 2012, UT1-UTC= 0.3153467, Accumulated Leap Seconds=35
  { 2456255.50, 66.870234 }, //24 November 2012, UT1-UTC= 0.3137657, Accumulated Leap Seconds=35
  { 2456256.50, 66.871670 }, //25 November 2012, UT1-UTC= 0.3123299, Accumulated Leap Seconds=35
  { 2456257.50, 66.872949 }, //26 November 2012, UT1-UTC= 0.3110510, Accumulated Leap Seconds=35
  { 2456258.50, 66.874111 }, //27 November 2012, UT1-UTC= 0.3098885, Accumulated Leap Seconds=35
  { 2456259.50, 66.875147 }, //28 November 2012, UT1-UTC= 0.3088531, Accumulated Leap Seconds=35
  { 2456260.50, 66.876087 }, //29 November 2012, UT1-UTC= 0.3079131, Accumulated Leap Seconds=35
  { 2456261.50, 66.876988 }, //30 November 2012, UT1-UTC= 0.3070124, Accumulated Leap Seconds=35
  { 2456262.50, 66.877860 }, //01 December 2012, UT1-UTC= 0.3061399, Accumulated Leap Seconds=35
  { 2456263.50, 66.878814 }, //02 December 2012, UT1-UTC= 0.3051859, Accumulated Leap Seconds=35
  { 2456264.50, 66.879848 }, //03 December 2012, UT1-UTC= 0.3041521, Accumulated Leap Seconds=35
  { 2456265.50, 66.880957 }, //04 December 2012, UT1-UTC= 0.3030432, Accumulated Leap Seconds=35
  { 2456266.50, 66.882160 }, //05 December 2012, UT1-UTC= 0.3018403, Accumulated Leap Seconds=35
  { 2456267.50, 66.883461 }, //06 December 2012, UT1-UTC= 0.3005391, Accumulated Leap Seconds=35
  { 2456268.50, 66.884822 }, //07 December 2012, UT1-UTC= 0.2991782, Accumulated Leap Seconds=35
  { 2456269.50, 66.886191 }, //08 December 2012, UT1-UTC= 0.2978093, Accumulated Leap Seconds=35
  { 2456270.50, 66.887546 }, //09 December 2012, UT1-UTC= 0.2964536, Accumulated Leap Seconds=35
  { 2456271.50, 66.888877 }, //10 December 2012, UT1-UTC= 0.2951226, Accumulated Leap Seconds=35
  { 2456272.50, 66.890101 }, //11 December 2012, UT1-UTC= 0.2938993, Accumulated Leap Seconds=35
  { 2456273.50, 66.891187 }, //12 December 2012, UT1-UTC= 0.2928127, Accumulated Leap Seconds=35
  { 2456274.50, 66.892142 }, //13 December 2012, UT1-UTC= 0.2918578, Accumulated Leap Seconds=35
  { 2456275.50, 66.893050 }, //14 December 2012, UT1-UTC= 0.2909503, Accumulated Leap Seconds=35
  { 2456276.50, 66.893982 }, //15 December 2012, UT1-UTC= 0.2900183, Accumulated Leap Seconds=35
  { 2456277.50, 66.894980 }, //16 December 2012, UT1-UTC= 0.2890199, Accumulated Leap Seconds=35
  { 2456278.50, 66.896052 }, //17 December 2012, UT1-UTC= 0.2879482, Accumulated Leap Seconds=35
  { 2456279.50, 66.897179 }, //18 December 2012, UT1-UTC= 0.2868211, Accumulated Leap Seconds=35
  { 2456280.50, 66.898298 }, //19 December 2012, UT1-UTC= 0.2857017, Accumulated Leap Seconds=35
  { 2456281.50, 66.899321 }, //20 December 2012, UT1-UTC= 0.2846791, Accumulated Leap Seconds=35
  { 2456282.50, 66.900197 }, //21 December 2012, UT1-UTC= 0.2838031, Accumulated Leap Seconds=35
  { 2456283.50, 66.900948 }, //22 December 2012, UT1-UTC= 0.2830518, Accumulated Leap Seconds=35
  { 2456284.50, 66.901664 }, //23 December 2012, UT1-UTC= 0.2823359, Accumulated Leap Seconds=35
  { 2456285.50, 66.902330 }, //24 December 2012, UT1-UTC= 0.2816704, Accumulated Leap Seconds=35
  { 2456286.50, 66.902895 }, //25 December 2012, UT1-UTC= 0.2811052, Accumulated Leap Seconds=35
  { 2456287.50, 66.903377 }, //26 December 2012, UT1-UTC= 0.2806228, Accumulated Leap Seconds=35
  { 2456288.50, 66.903836 }, //27 December 2012, UT1-UTC= 0.2801644, Accumulated Leap Seconds=35
  { 2456289.50, 66.904290 }, //28 December 2012, UT1-UTC= 0.2797098, Accumulated Leap Seconds=35
  { 2456290.50, 66.904760 }, //29 December 2012, UT1-UTC= 0.2792396, Accumulated Leap Seconds=35
  { 2456291.50, 66.905352 }, //30 December 2012, UT1-UTC= 0.2786482, Accumulated Leap Seconds=35
  { 2456292.50, 66.906073 }, //31 December 2012, UT1-UTC= 0.2779265, Accumulated Leap Seconds=35
  { 2456293.50, 66.906909 }, //01 January 2013, UT1-UTC= 0.2770909, Accumulated Leap Seconds=35
  { 2456294.50, 66.907870 }, //02 January 2013, UT1-UTC= 0.2761301, Accumulated Leap Seconds=35
  { 2456295.50, 66.908961 }, //03 January 2013, UT1-UTC= 0.2750390, Accumulated Leap Seconds=35
  { 2456296.50, 66.910141 }, //04 January 2013, UT1-UTC= 0.2738592, Accumulated Leap Seconds=35
  { 2456297.50, 66.911341 }, //05 January 2013, UT1-UTC= 0.2726594, Accumulated Leap Seconds=35
  { 2456298.50, 66.912528 }, //06 January 2013, UT1-UTC= 0.2714718, Accumulated Leap Seconds=35
  { 2456299.50, 66.913604 }, //07 January 2013, UT1-UTC= 0.2703958, Accumulated Leap Seconds=35
  { 2456300.50, 66.914537 }, //08 January 2013, UT1-UTC= 0.2694627, Accumulated Leap Seconds=35
  { 2456301.50, 66.915387 }, //09 January 2013, UT1-UTC= 0.2686133, Accumulated Leap Seconds=35
  { 2456302.50, 66.916238 }, //10 January 2013, UT1-UTC= 0.2677618, Accumulated Leap Seconds=35
  { 2456303.50, 66.917193 }, //11 January 2013, UT1-UTC= 0.2668071, Accumulated Leap Seconds=35
  { 2456304.50, 66.918274 }, //12 January 2013, UT1-UTC= 0.2657257, Accumulated Leap Seconds=35
  { 2456305.50, 66.919512 }, //13 January 2013, UT1-UTC= 0.2644880, Accumulated Leap Seconds=35
  { 2456306.50, 66.920917 }, //14 January 2013, UT1-UTC= 0.2630828, Accumulated Leap Seconds=35
  { 2456307.50, 66.922434 }, //15 January 2013, UT1-UTC= 0.2615655, Accumulated Leap Seconds=35
  { 2456308.50, 66.923926 }, //16 January 2013, UT1-UTC= 0.2600741, Accumulated Leap Seconds=35
  { 2456309.50, 66.925320 }, //17 January 2013, UT1-UTC= 0.2586797, Accumulated Leap Seconds=35
  { 2456310.50, 66.926624 }, //18 January 2013, UT1-UTC= 0.2573759, Accumulated Leap Seconds=35
  { 2456311.50, 66.927876 }, //19 January 2013, UT1-UTC= 0.2561239, Accumulated Leap Seconds=35
  { 2456312.50, 66.929037 }, //20 January 2013, UT1-UTC= 0.2549625, Accumulated Leap Seconds=35
  { 2456313.50, 66.930137 }, //21 January 2013, UT1-UTC= 0.2538633, Accumulated Leap Seconds=35
  { 2456314.50, 66.931198 }, //22 January 2013, UT1-UTC= 0.2528018, Accumulated Leap Seconds=35
  { 2456315.50, 66.932263 }, //23 January 2013, UT1-UTC= 0.2517366, Accumulated Leap Seconds=35
  { 2456316.50, 66.933356 }, //24 January 2013, UT1-UTC= 0.2506442, Accumulated Leap Seconds=35
  { 2456317.50, 66.934484 }, //25 January 2013, UT1-UTC= 0.2495164, Accumulated Leap Seconds=35
  { 2456318.50, 66.935610 }, //26 January 2013, UT1-UTC= 0.2483898, Accumulated Leap Seconds=35
  { 2456319.50, 66.936858 }, //27 January 2013, UT1-UTC= 0.2471418, Accumulated Leap Seconds=35
  { 2456320.50, 66.938236 }, //28 January 2013, UT1-UTC= 0.2457638, Accumulated Leap Seconds=35
  { 2456321.50, 66.939698 }, //29 January 2013, UT1-UTC= 0.2443016, Accumulated Leap Seconds=35
  { 2456322.50, 66.941192 }, //30 January 2013, UT1-UTC= 0.2428084, Accumulated Leap Seconds=35
  { 2456323.50, 66.942763 }, //31 January 2013, UT1-UTC= 0.2412368, Accumulated Leap Seconds=35
  { 2456324.50, 66.944259 }, //01 February 2013, UT1-UTC= 0.2397410, Accumulated Leap Seconds=35
  { 2456325.50, 66.945640 }, //02 February 2013, UT1-UTC= 0.2383595, Accumulated Leap Seconds=35
  { 2456326.50, 66.946864 }, //03 February 2013, UT1-UTC= 0.2371356, Accumulated Leap Seconds=35
  { 2456327.50, 66.947963 }, //04 February 2013, UT1-UTC= 0.2360374, Accumulated Leap Seconds=35
  { 2456328.50, 66.948970 }, //05 February 2013, UT1-UTC= 0.2350300, Accumulated Leap Seconds=35
  { 2456329.50, 66.949892 }, //06 February 2013, UT1-UTC= 0.2341084, Accumulated Leap Seconds=35
  { 2456330.50, 66.950826 }, //07 February 2013, UT1-UTC= 0.2331743, Accumulated Leap Seconds=35
  { 2456331.50, 66.951863 }, //08 February 2013, UT1-UTC= 0.2321369, Accumulated Leap Seconds=35
  { 2456332.50, 66.953060 }, //09 February 2013, UT1-UTC= 0.2309402, Accumulated Leap Seconds=35
  { 2456333.50, 66.954395 }, //10 February 2013, UT1-UTC= 0.2296051, Accumulated Leap Seconds=35
  { 2456334.50, 66.955804 }, //11 February 2013, UT1-UTC= 0.2281962, Accumulated Leap Seconds=35
  { 2456335.50, 66.957225 }, //12 February 2013, UT1-UTC= 0.2267749, Accumulated Leap Seconds=35
  { 2456336.50, 66.958553 }, //13 February 2013, UT1-UTC= 0.2254471, Accumulated Leap Seconds=35
  { 2456337.50, 66.959713 }, //14 February 2013, UT1-UTC= 0.2242867, Accumulated Leap Seconds=35
  { 2456338.50, 66.960712 }, //15 February 2013, UT1-UTC= 0.2232879, Accumulated Leap Seconds=35
  { 2456339.50, 66.961616 }, //16 February 2013, UT1-UTC= 0.2223839, Accumulated Leap Seconds=35
  { 2456340.50, 66.962413 }, //17 February 2013, UT1-UTC= 0.2215870, Accumulated Leap Seconds=35
  { 2456341.50, 66.963102 }, //18 February 2013, UT1-UTC= 0.2208981, Accumulated Leap Seconds=35
  { 2456342.50, 66.963723 }, //19 February 2013, UT1-UTC= 0.2202771, Accumulated Leap Seconds=35
  { 2456343.50, 66.964359 }, //20 February 2013, UT1-UTC= 0.2196415, Accumulated Leap Seconds=35
  { 2456344.50, 66.965068 }, //21 February 2013, UT1-UTC= 0.2189318, Accumulated Leap Seconds=35
  { 2456345.50, 66.965919 }, //22 February 2013, UT1-UTC= 0.2180807, Accumulated Leap Seconds=35
  { 2456346.50, 66.966949 }, //23 February 2013, UT1-UTC= 0.2170512, Accumulated Leap Seconds=35
  { 2456347.50, 66.968155 }, //24 February 2013, UT1-UTC= 0.2158446, Accumulated Leap Seconds=35
  { 2456348.50, 66.969517 }, //25 February 2013, UT1-UTC= 0.2144827, Accumulated Leap Seconds=35
  { 2456349.50, 66.971072 }, //26 February 2013, UT1-UTC= 0.2129276, Accumulated Leap Seconds=35
  { 2456350.50, 66.972753 }, //27 February 2013, UT1-UTC= 0.2112466, Accumulated Leap Seconds=35
  { 2456351.50, 66.974508 }, //28 February 2013, UT1-UTC= 0.2094922, Accumulated Leap Seconds=35
  { 2456352.50, 66.976251 }, //01 March 2013, UT1-UTC= 0.2077492, Accumulated Leap Seconds=35
  { 2456353.50, 66.977938 }, //02 March 2013, UT1-UTC= 0.2060622, Accumulated Leap Seconds=35
  { 2456354.50, 66.979469 }, //03 March 2013, UT1-UTC= 0.2045307, Accumulated Leap Seconds=35
  { 2456355.50, 66.980928 }, //04 March 2013, UT1-UTC= 0.2030719, Accumulated Leap Seconds=35
  { 2456356.50, 66.982393 }, //05 March 2013, UT1-UTC= 0.2016072, Accumulated Leap Seconds=35
  { 2456357.50, 66.983886 }, //06 March 2013, UT1-UTC= 0.2001142, Accumulated Leap Seconds=35
  { 2456358.50, 66.985436 }, //07 March 2013, UT1-UTC= 0.1985640, Accumulated Leap Seconds=35
  { 2456359.50, 66.987092 }, //08 March 2013, UT1-UTC= 0.1969081, Accumulated Leap Seconds=35
  { 2456360.50, 66.988847 }, //09 March 2013, UT1-UTC= 0.1951532, Accumulated Leap Seconds=35
  { 2456361.50, 66.990718 }, //10 March 2013, UT1-UTC= 0.1932825, Accumulated Leap Seconds=35
  { 2456362.50, 66.992619 }, //11 March 2013, UT1-UTC= 0.1913807, Accumulated Leap Seconds=35
  { 2456363.50, 66.994495 }, //12 March 2013, UT1-UTC= 0.1895051, Accumulated Leap Seconds=35
  { 2456364.50, 66.996338 }, //13 March 2013, UT1-UTC= 0.1876624, Accumulated Leap Seconds=35
  { 2456365.50, 66.998089 }, //14 March 2013, UT1-UTC= 0.1859111, Accumulated Leap Seconds=35
  { 2456366.50, 66.999695 }, //15 March 2013, UT1-UTC= 0.1843053, Accumulated Leap Seconds=35
  { 2456367.50, 67.001166 }, //16 March 2013, UT1-UTC= 0.1828344, Accumulated Leap Seconds=35
  { 2456368.50, 67.002538 }, //17 March 2013, UT1-UTC= 0.1814622, Accumulated Leap Seconds=35
  { 2456369.50, 67.003799 }, //18 March 2013, UT1-UTC= 0.1802006, Accumulated Leap Seconds=35
  { 2456370.50, 67.004995 }, //19 March 2013, UT1-UTC= 0.1790053, Accumulated Leap Seconds=35
  { 2456371.50, 67.006135 }, //20 March 2013, UT1-UTC= 0.1778653, Accumulated Leap Seconds=35
  { 2456372.50, 67.007314 }, //21 March 2013, UT1-UTC= 0.1766857, Accumulated Leap Seconds=35
  { 2456373.50, 67.008525 }, //22 March 2013, UT1-UTC= 0.1754752, Accumulated Leap Seconds=35
  { 2456374.50, 67.009844 }, //23 March 2013, UT1-UTC= 0.1741560, Accumulated Leap Seconds=35
  { 2456375.50, 67.011339 }, //24 March 2013, UT1-UTC= 0.1726614, Accumulated Leap Seconds=35
  { 2456376.50, 67.012985 }, //25 March 2013, UT1-UTC= 0.1710147, Accumulated Leap Seconds=35
  { 2456377.50, 67.014769 }, //26 March 2013, UT1-UTC= 0.1692309, Accumulated Leap Seconds=35
  { 2456378.50, 67.016674 }, //27 March 2013, UT1-UTC= 0.1673260, Accumulated Leap Seconds=35
  { 2456379.50, 67.018650 }, //28 March 2013, UT1-UTC= 0.1653499, Accumulated Leap Seconds=35
  { 2456380.50, 67.020584 }, //29 March 2013, UT1-UTC= 0.1634160, Accumulated Leap Seconds=35
  { 2456381.50, 67.022441 }, //30 March 2013, UT1-UTC= 0.1615592, Accumulated Leap Seconds=35
  { 2456382.50, 67.024175 }, //31 March 2013, UT1-UTC= 0.1598249, Accumulated Leap Seconds=35
  { 2456383.50, 67.025813 }, //01 April 2013, UT1-UTC= 0.1581874, Accumulated Leap Seconds=35
  { 2456384.50, 67.027387 }, //02 April 2013, UT1-UTC= 0.1566131, Accumulated Leap Seconds=35
  { 2456385.50, 67.028971 }, //03 April 2013, UT1-UTC= 0.1550285, Accumulated Leap Seconds=35
  { 2456386.50, 67.030575 }, //04 April 2013, UT1-UTC= 0.1534245, Accumulated Leap Seconds=35
  { 2456387.50, 67.032259 }, //05 April 2013, UT1-UTC= 0.1517410, Accumulated Leap Seconds=35
  { 2456388.50, 67.034089 }, //06 April 2013, UT1-UTC= 0.1499114, Accumulated Leap Seconds=35
  { 2456389.50, 67.035978 }, //07 April 2013, UT1-UTC= 0.1480220, Accumulated Leap Seconds=35
  { 2456390.50, 67.037854 }, //08 April 2013, UT1-UTC= 0.1461464, Accumulated Leap Seconds=35
  { 2456391.50, 67.039699 }, //09 April 2013, UT1-UTC= 0.1443015, Accumulated Leap Seconds=35
  { 2456392.50, 67.041451 }, //10 April 2013, UT1-UTC= 0.1425486, Accumulated Leap Seconds=35
  { 2456393.50, 67.043056 }, //11 April 2013, UT1-UTC= 0.1409438, Accumulated Leap Seconds=35
  { 2456394.50, 67.044492 }, //12 April 2013, UT1-UTC= 0.1395077, Accumulated Leap Seconds=35
  { 2456395.50, 67.045786 }, //13 April 2013, UT1-UTC= 0.1382141, Accumulated Leap Seconds=35
  { 2456396.50, 67.046968 }, //14 April 2013, UT1-UTC= 0.1370315, Accumulated Leap Seconds=35
  { 2456397.50, 67.048065 }, //15 April 2013, UT1-UTC= 0.1359352, Accumulated Leap Seconds=35
  { 2456398.50, 67.049166 }, //16 April 2013, UT1-UTC= 0.1348339, Accumulated Leap Seconds=35
  { 2456399.50, 67.050315 }, //17 April 2013, UT1-UTC= 0.1336846, Accumulated Leap Seconds=35
  { 2456400.50, 67.051476 }, //18 April 2013, UT1-UTC= 0.1325236, Accumulated Leap Seconds=35
  { 2456401.50, 67.052714 }, //19 April 2013, UT1-UTC= 0.1312859, Accumulated Leap Seconds=35
  { 2456402.50, 67.054114 }, //20 April 2013, UT1-UTC= 0.1298863, Accumulated Leap Seconds=35
  { 2456403.50, 67.055656 }, //21 April 2013, UT1-UTC= 0.1283439, Accumulated Leap Seconds=35
  { 2456404.50, 67.057289 }, //22 April 2013, UT1-UTC= 0.1267108, Accumulated Leap Seconds=35
  { 2456405.50, 67.059038 }, //23 April 2013, UT1-UTC= 0.1249617, Accumulated Leap Seconds=35
  { 2456406.50, 67.060863 }, //24 April 2013, UT1-UTC= 0.1231367, Accumulated Leap Seconds=35
  { 2456407.50, 67.062678 }, //25 April 2013, UT1-UTC= 0.1213216, Accumulated Leap Seconds=35
  { 2456408.50, 67.064429 }, //26 April 2013, UT1-UTC= 0.1195707, Accumulated Leap Seconds=35
  { 2456409.50, 67.066046 }, //27 April 2013, UT1-UTC= 0.1179535, Accumulated Leap Seconds=35
  { 2456410.50, 67.067504 }, //28 April 2013, UT1-UTC= 0.1164958, Accumulated Leap Seconds=35
  { 2456411.50, 67.068873 }, //29 April 2013, UT1-UTC= 0.1151272, Accumulated Leap Seconds=35
  { 2456412.50, 67.070230 }, //30 April 2013, UT1-UTC= 0.1137697, Accumulated Leap Seconds=35
  { 2456413.50, 67.071629 }, //01 May 2013, UT1-UTC= 0.1123714, Accumulated Leap Seconds=35
  { 2456414.50, 67.073105 }, //02 May 2013, UT1-UTC= 0.1108950, Accumulated Leap Seconds=35
  { 2456415.50, 67.074704 }, //03 May 2013, UT1-UTC= 0.1092964, Accumulated Leap Seconds=35
  { 2456416.50, 67.076355 }, //04 May 2013, UT1-UTC= 0.1076446, Accumulated Leap Seconds=35
  { 2456417.50, 67.078060 }, //05 May 2013, UT1-UTC= 0.1059401, Accumulated Leap Seconds=35
  { 2456418.50, 67.079723 }, //06 May 2013, UT1-UTC= 0.1042767, Accumulated Leap Seconds=35
  { 2456419.50, 67.081326 }, //07 May 2013, UT1-UTC= 0.1026739, Accumulated Leap Seconds=35
  { 2456420.50, 67.082869 }, //08 May 2013, UT1-UTC= 0.1011310, Accumulated Leap Seconds=35
  { 2456421.50, 67.084248 }, //09 May 2013, UT1-UTC= 0.0997519, Accumulated Leap Seconds=35
  { 2456422.50, 67.085441 }, //10 May 2013, UT1-UTC= 0.0985591, Accumulated Leap Seconds=35
  { 2456423.50, 67.086537 }, //11 May 2013, UT1-UTC= 0.0974628, Accumulated Leap Seconds=35
  { 2456424.50, 67.087533 }, //12 May 2013, UT1-UTC= 0.0964673, Accumulated Leap Seconds=35
  { 2456425.50, 67.088444 }, //13 May 2013, UT1-UTC= 0.0955559, Accumulated Leap Seconds=35
  { 2456426.50, 67.089332 }, //14 May 2013, UT1-UTC= 0.0946677, Accumulated Leap Seconds=35
  { 2456427.50, 67.090268 }, //15 May 2013, UT1-UTC= 0.0937318, Accumulated Leap Seconds=35
  { 2456428.50, 67.091273 }, //16 May 2013, UT1-UTC= 0.0927271, Accumulated Leap Seconds=35
  { 2456429.50, 67.092397 }, //17 May 2013, UT1-UTC= 0.0916030, Accumulated Leap Seconds=35
  { 2456430.50, 67.093648 }, //18 May 2013, UT1-UTC= 0.0903520, Accumulated Leap Seconds=35
  { 2456431.50, 67.095043 }, //19 May 2013, UT1-UTC= 0.0889571, Accumulated Leap Seconds=35
  { 2456432.50, 67.096528 }, //20 May 2013, UT1-UTC= 0.0874723, Accumulated Leap Seconds=35
  { 2456433.50, 67.098051 }, //21 May 2013, UT1-UTC= 0.0859485, Accumulated Leap Seconds=35
  { 2456434.50, 67.099546 }, //22 May 2013, UT1-UTC= 0.0844544, Accumulated Leap Seconds=35
  { 2456435.50, 67.100954 }, //23 May 2013, UT1-UTC= 0.0830456, Accumulated Leap Seconds=35
  { 2456436.50, 67.102261 }, //24 May 2013, UT1-UTC= 0.0817386, Accumulated Leap Seconds=35
  { 2456437.50, 67.103472 }, //25 May 2013, UT1-UTC= 0.0805279, Accumulated Leap Seconds=35
  { 2456438.50, 67.104547 }, //26 May 2013, UT1-UTC= 0.0794526, Accumulated Leap Seconds=35
  { 2456439.50, 67.105516 }, //27 May 2013, UT1-UTC= 0.0784835, Accumulated Leap Seconds=35
  { 2456440.50, 67.106432 }, //28 May 2013, UT1-UTC= 0.0775684, Accumulated Leap Seconds=35
  { 2456441.50, 67.107325 }, //29 May 2013, UT1-UTC= 0.0766749, Accumulated Leap Seconds=35
  { 2456442.50, 67.108246 }, //30 May 2013, UT1-UTC= 0.0757541, Accumulated Leap Seconds=35
  { 2456443.50, 67.109189 }, //31 May 2013, UT1-UTC= 0.0748112, Accumulated Leap Seconds=35
  { 2456444.50, 67.110018 }, //01 June 2013, UT1-UTC= 0.0739816, Accumulated Leap Seconds=35
  { 2456445.50, 67.110786 }, //02 June 2013, UT1-UTC= 0.0732142, Accumulated Leap Seconds=35
  { 2456446.50, 67.111474 }, //03 June 2013, UT1-UTC= 0.0725261, Accumulated Leap Seconds=35
  { 2456447.50, 67.112081 }, //04 June 2013, UT1-UTC= 0.0719188, Accumulated Leap Seconds=35
  { 2456448.50, 67.112558 }, //05 June 2013, UT1-UTC= 0.0714419, Accumulated Leap Seconds=35
  { 2456449.50, 67.112928 }, //06 June 2013, UT1-UTC= 0.0710720, Accumulated Leap Seconds=35
  { 2456450.50, 67.113206 }, //07 June 2013, UT1-UTC= 0.0707938, Accumulated Leap Seconds=35
  { 2456451.50, 67.113393 }, //08 June 2013, UT1-UTC= 0.0706070, Accumulated Leap Seconds=35
  { 2456452.50, 67.113485 }, //09 June 2013, UT1-UTC= 0.0705147, Accumulated Leap Seconds=35
  { 2456453.50, 67.113577 }, //10 June 2013, UT1-UTC= 0.0704234, Accumulated Leap Seconds=35
  { 2456454.50, 67.113758 }, //11 June 2013, UT1-UTC= 0.0702420, Accumulated Leap Seconds=35
  { 2456455.50, 67.114002 }, //12 June 2013, UT1-UTC= 0.0699977, Accumulated Leap Seconds=35
  { 2456456.50, 67.114344 }, //13 June 2013, UT1-UTC= 0.0696561, Accumulated Leap Seconds=35
  { 2456457.50, 67.114843 }, //14 June 2013, UT1-UTC= 0.0691571, Accumulated Leap Seconds=35
  { 2456458.50, 67.115465 }, //15 June 2013, UT1-UTC= 0.0685349, Accumulated Leap Seconds=35
  { 2456459.50, 67.116194 }, //16 June 2013, UT1-UTC= 0.0678064, Accumulated Leap Seconds=35
  { 2456460.50, 67.116994 }, //17 June 2013, UT1-UTC= 0.0670063, Accumulated Leap Seconds=35
  { 2456461.50, 67.117869 }, //18 June 2013, UT1-UTC= 0.0661313, Accumulated Leap Seconds=35
  { 2456462.50, 67.118705 }, //19 June 2013, UT1-UTC= 0.0652946, Accumulated Leap Seconds=35
  { 2456463.50, 67.119506 }, //20 June 2013, UT1-UTC= 0.0644936, Accumulated Leap Seconds=35
  { 2456464.50, 67.120229 }, //21 June 2013, UT1-UTC= 0.0637705, Accumulated Leap Seconds=35
  { 2456465.50, 67.120822 }, //22 June 2013, UT1-UTC= 0.0631776, Accumulated Leap Seconds=35
  { 2456466.50, 67.121376 }, //23 June 2013, UT1-UTC= 0.0626242, Accumulated Leap Seconds=35
  { 2456467.50, 67.121951 }, //24 June 2013, UT1-UTC= 0.0620486, Accumulated Leap Seconds=35
  { 2456468.50, 67.122621 }, //25 June 2013, UT1-UTC= 0.0613785, Accumulated Leap Seconds=35
  { 2456469.50, 67.123359 }, //26 June 2013, UT1-UTC= 0.0606405, Accumulated Leap Seconds=35
  { 2456470.50, 67.124157 }, //27 June 2013, UT1-UTC= 0.0598425, Accumulated Leap Seconds=35
  { 2456471.50, 67.124945 }, //28 June 2013, UT1-UTC= 0.0590551, Accumulated Leap Seconds=35
  { 2456472.50, 67.125662 }, //29 June 2013, UT1-UTC= 0.0583377, Accumulated Leap Seconds=35
  { 2456473.50, 67.126242 }, //30 June 2013, UT1-UTC= 0.0577583, Accumulated Leap Seconds=35
  { 2456474.50, 67.126640 }, //01 July 2013, UT1-UTC= 0.0573599, Accumulated Leap Seconds=35
  { 2456475.50, 67.126831 }, //02 July 2013, UT1-UTC= 0.0571685, Accumulated Leap Seconds=35
  { 2456476.50, 67.126848 }, //03 July 2013, UT1-UTC= 0.0571524, Accumulated Leap Seconds=35
  { 2456477.50, 67.126694 }, //04 July 2013, UT1-UTC= 0.0573061, Accumulated Leap Seconds=35
  { 2456478.50, 67.126414 }, //05 July 2013, UT1-UTC= 0.0575859, Accumulated Leap Seconds=35
  { 2456479.50, 67.126080 }, //06 July 2013, UT1-UTC= 0.0579201, Accumulated Leap Seconds=35
  { 2456480.50, 67.125729 }, //07 July 2013, UT1-UTC= 0.0582709, Accumulated Leap Seconds=35
  { 2456481.50, 67.125425 }, //08 July 2013, UT1-UTC= 0.0585745, Accumulated Leap Seconds=35
  { 2456482.50, 67.125231 }, //09 July 2013, UT1-UTC= 0.0587690, Accumulated Leap Seconds=35
  { 2456483.50, 67.125165 }, //10 July 2013, UT1-UTC= 0.0588351, Accumulated Leap Seconds=35
  { 2456484.50, 67.125244 }, //11 July 2013, UT1-UTC= 0.0587556, Accumulated Leap Seconds=35
  { 2456485.50, 67.125471 }, //12 July 2013, UT1-UTC= 0.0585291, Accumulated Leap Seconds=35
  { 2456486.50, 67.125784 }, //13 July 2013, UT1-UTC= 0.0582157, Accumulated Leap Seconds=35
  { 2456487.50, 67.126165 }, //14 July 2013, UT1-UTC= 0.0578346, Accumulated Leap Seconds=35
  { 2456488.50, 67.126584 }, //15 July 2013, UT1-UTC= 0.0574160, Accumulated Leap Seconds=35
  { 2456489.50, 67.127030 }, //16 July 2013, UT1-UTC= 0.0569702, Accumulated Leap Seconds=35
  { 2456490.50, 67.127463 }, //17 July 2013, UT1-UTC= 0.0565365, Accumulated Leap Seconds=35
  { 2456491.50, 67.127819 }, //18 July 2013, UT1-UTC= 0.0561813, Accumulated Leap Seconds=35
  { 2456492.50, 67.128128 }, //19 July 2013, UT1-UTC= 0.0558717, Accumulated Leap Seconds=35
  { 2456493.50, 67.128363 }, //20 July 2013, UT1-UTC= 0.0556366, Accumulated Leap Seconds=35
  { 2456494.50, 67.128583 }, //21 July 2013, UT1-UTC= 0.0554172, Accumulated Leap Seconds=35
  { 2456495.50, 67.128843 }, //22 July 2013, UT1-UTC= 0.0551567, Accumulated Leap Seconds=35
  { 2456496.50, 67.129218 }, //23 July 2013, UT1-UTC= 0.0547824, Accumulated Leap Seconds=35
  { 2456497.50, 67.129722 }, //24 July 2013, UT1-UTC= 0.0542778, Accumulated Leap Seconds=35
  { 2456498.50, 67.130341 }, //25 July 2013, UT1-UTC= 0.0536586, Accumulated Leap Seconds=35
  { 2456499.50, 67.130996 }, //26 July 2013, UT1-UTC= 0.0530038, Accumulated Leap Seconds=35
  { 2456500.50, 67.131642 }, //27 July 2013, UT1-UTC= 0.0523582, Accumulated Leap Seconds=35
  { 2456501.50, 67.132179 }, //28 July 2013, UT1-UTC= 0.0518211, Accumulated Leap Seconds=35
  { 2456502.50, 67.132604 }, //29 July 2013, UT1-UTC= 0.0513956, Accumulated Leap Seconds=35
  { 2456503.50, 67.132902 }, //30 July 2013, UT1-UTC= 0.0510983, Accumulated Leap Seconds=35
  { 2456504.50, 67.133060 }, //31 July 2013, UT1-UTC= 0.0509405, Accumulated Leap Seconds=35
  { 2456505.50, 67.133139 }, //01 August 2013, UT1-UTC= 0.0508609, Accumulated Leap Seconds=35
  { 2456506.50, 67.133136 }, //02 August 2013, UT1-UTC= 0.0508637, Accumulated Leap Seconds=35
  { 2456507.50, 67.133078 }, //03 August 2013, UT1-UTC= 0.0509223, Accumulated Leap Seconds=35
  { 2456508.50, 67.133044 }, //04 August 2013, UT1-UTC= 0.0509561, Accumulated Leap Seconds=35
  { 2456509.50, 67.133073 }, //05 August 2013, UT1-UTC= 0.0509268, Accumulated Leap Seconds=35
  { 2456510.50, 67.133218 }, //06 August 2013, UT1-UTC= 0.0507820, Accumulated Leap Seconds=35
  { 2456511.50, 67.133440 }, //07 August 2013, UT1-UTC= 0.0505595, Accumulated Leap Seconds=35
  { 2456512.50, 67.133759 }, //08 August 2013, UT1-UTC= 0.0502410, Accumulated Leap Seconds=35
  { 2456513.50, 67.134167 }, //09 August 2013, UT1-UTC= 0.0498328, Accumulated Leap Seconds=35
  { 2456514.50, 67.134668 }, //10 August 2013, UT1-UTC= 0.0493321, Accumulated Leap Seconds=35
  { 2456515.50, 67.135228 }, //11 August 2013, UT1-UTC= 0.0487723, Accumulated Leap Seconds=35
  { 2456516.50, 67.135755 }, //12 August 2013, UT1-UTC= 0.0482448, Accumulated Leap Seconds=35
  { 2456517.50, 67.136208 }, //13 August 2013, UT1-UTC= 0.0477920, Accumulated Leap Seconds=35
  { 2456518.50, 67.136602 }, //14 August 2013, UT1-UTC= 0.0473982, Accumulated Leap Seconds=35
  { 2456519.50, 67.136942 }, //15 August 2013, UT1-UTC= 0.0470580, Accumulated Leap Seconds=35
  { 2456520.50, 67.137247 }, //16 August 2013, UT1-UTC= 0.0467527, Accumulated Leap Seconds=35
  { 2456521.50, 67.137583 }, //17 August 2013, UT1-UTC= 0.0464170, Accumulated Leap Seconds=35
  { 2456522.50, 67.138028 }, //18 August 2013, UT1-UTC= 0.0459724, Accumulated Leap Seconds=35
  { 2456523.50, 67.138611 }, //19 August 2013, UT1-UTC= 0.0453891, Accumulated Leap Seconds=35
  { 2456524.50, 67.139327 }, //20 August 2013, UT1-UTC= 0.0446732, Accumulated Leap Seconds=35
  { 2456525.50, 67.140172 }, //21 August 2013, UT1-UTC= 0.0438280, Accumulated Leap Seconds=35
  { 2456526.50, 67.141052 }, //22 August 2013, UT1-UTC= 0.0429478, Accumulated Leap Seconds=35
  { 2456527.50, 67.141915 }, //23 August 2013, UT1-UTC= 0.0420853, Accumulated Leap Seconds=35
  { 2456528.50, 67.142717 }, //24 August 2013, UT1-UTC= 0.0412834, Accumulated Leap Seconds=35
  { 2456529.50, 67.143408 }, //25 August 2013, UT1-UTC= 0.0405919, Accumulated Leap Seconds=35
  { 2456530.50, 67.144003 }, //26 August 2013, UT1-UTC= 0.0399971, Accumulated Leap Seconds=35
  { 2456531.50, 67.144470 }, //27 August 2013, UT1-UTC= 0.0395295, Accumulated Leap Seconds=35
  { 2456532.50, 67.144807 }, //28 August 2013, UT1-UTC= 0.0391929, Accumulated Leap Seconds=35
  { 2456533.50, 67.145087 }, //29 August 2013, UT1-UTC= 0.0389134, Accumulated Leap Seconds=35
  { 2456534.50, 67.145317 }, //30 August 2013, UT1-UTC= 0.0386829, Accumulated Leap Seconds=35
  { 2456535.50, 67.145544 }, //31 August 2013, UT1-UTC= 0.0384563, Accumulated Leap Seconds=35
  { 2456536.50, 67.145797 }, //01 September 2013, UT1-UTC= 0.0382030, Accumulated Leap Seconds=35
  { 2456537.50, 67.146150 }, //02 September 2013, UT1-UTC= 0.0378499, Accumulated Leap Seconds=35
  { 2456538.50, 67.146614 }, //03 September 2013, UT1-UTC= 0.0373860, Accumulated Leap Seconds=35
  { 2456539.50, 67.147206 }, //04 September 2013, UT1-UTC= 0.0367941, Accumulated Leap Seconds=35
  { 2456540.50, 67.147939 }, //05 September 2013, UT1-UTC= 0.0360611, Accumulated Leap Seconds=35
  { 2456541.50, 67.148817 }, //06 September 2013, UT1-UTC= 0.0351828, Accumulated Leap Seconds=35
  { 2456542.50, 67.149772 }, //07 September 2013, UT1-UTC= 0.0342281, Accumulated Leap Seconds=35
  { 2456543.50, 67.150740 }, //08 September 2013, UT1-UTC= 0.0332595, Accumulated Leap Seconds=35
  { 2456544.50, 67.151644 }, //09 September 2013, UT1-UTC= 0.0323559, Accumulated Leap Seconds=35
  { 2456545.50, 67.152462 }, //10 September 2013, UT1-UTC= 0.0315380, Accumulated Leap Seconds=35
  { 2456546.50, 67.153217 }, //11 September 2013, UT1-UTC= 0.0307827, Accumulated Leap Seconds=35
  { 2456547.50, 67.153893 }, //12 September 2013, UT1-UTC= 0.0301072, Accumulated Leap Seconds=35
  { 2456548.50, 67.154523 }, //13 September 2013, UT1-UTC= 0.0294774, Accumulated Leap Seconds=35
  { 2456549.50, 67.155196 }, //14 September 2013, UT1-UTC= 0.0288044, Accumulated Leap Seconds=35
  { 2456550.50, 67.155982 }, //15 September 2013, UT1-UTC= 0.0280182, Accumulated Leap Seconds=35
  { 2456551.50, 67.156923 }, //16 September 2013, UT1-UTC= 0.0270773, Accumulated Leap Seconds=35
  { 2456552.50, 67.158038 }, //17 September 2013, UT1-UTC= 0.0259619, Accumulated Leap Seconds=35
  { 2456553.50, 67.159326 }, //18 September 2013, UT1-UTC= 0.0246739, Accumulated Leap Seconds=35
  { 2456554.50, 67.160695 }, //19 September 2013, UT1-UTC= 0.0233047, Accumulated Leap Seconds=35
  { 2456555.50, 67.162053 }, //20 September 2013, UT1-UTC= 0.0219471, Accumulated Leap Seconds=35
  { 2456556.50, 67.163265 }, //21 September 2013, UT1-UTC= 0.0207346, Accumulated Leap Seconds=35
  { 2456557.50, 67.164348 }, //22 September 2013, UT1-UTC= 0.0196520, Accumulated Leap Seconds=35
  { 2456558.50, 67.165296 }, //23 September 2013, UT1-UTC= 0.0187040, Accumulated Leap Seconds=35
  { 2456559.50, 67.166174 }, //24 September 2013, UT1-UTC= 0.0178257, Accumulated Leap Seconds=35
  { 2456560.50, 67.167026 }, //25 September 2013, UT1-UTC= 0.0169739, Accumulated Leap Seconds=35
  { 2456561.50, 67.167815 }, //26 September 2013, UT1-UTC= 0.0161847, Accumulated Leap Seconds=35
  { 2456562.50, 67.168564 }, //27 September 2013, UT1-UTC= 0.0154357, Accumulated Leap Seconds=35
  { 2456563.50, 67.169267 }, //28 September 2013, UT1-UTC= 0.0147334, Accumulated Leap Seconds=35
  { 2456564.50, 67.169992 }, //29 September 2013, UT1-UTC= 0.0140076, Accumulated Leap Seconds=35
  { 2456565.50, 67.170816 }, //30 September 2013, UT1-UTC= 0.0131841, Accumulated Leap Seconds=35
  { 2456566.50, 67.171740 }, //01 October 2013, UT1-UTC= 0.0122600, Accumulated Leap Seconds=35
  { 2456567.50, 67.172791 }, //02 October 2013, UT1-UTC= 0.0112094, Accumulated Leap Seconds=35
  { 2456568.50, 67.173976 }, //03 October 2013, UT1-UTC= 0.0100236, Accumulated Leap Seconds=35
  { 2456569.50, 67.175198 }, //04 October 2013, UT1-UTC= 0.0088021, Accumulated Leap Seconds=35
  { 2456570.50, 67.176489 }, //05 October 2013, UT1-UTC= 0.0075108, Accumulated Leap Seconds=35
  { 2456571.50, 67.177787 }, //06 October 2013, UT1-UTC= 0.0062132, Accumulated Leap Seconds=35
  { 2456572.50, 67.179051 }, //07 October 2013, UT1-UTC= 0.0049493, Accumulated Leap Seconds=35
  { 2456573.50, 67.180251 }, //08 October 2013, UT1-UTC= 0.0037486, Accumulated Leap Seconds=35
  { 2456574.50, 67.181416 }, //09 October 2013, UT1-UTC= 0.0025843, Accumulated Leap Seconds=35
  { 2456575.50, 67.182532 }, //10 October 2013, UT1-UTC= 0.0014680, Accumulated Leap Seconds=35
  { 2456576.50, 67.183617 }, //11 October 2013, UT1-UTC= 0.0003834, Accumulated Leap Seconds=35
  { 2456577.50, 67.184770 }, //12 October 2013, UT1-UTC=-0.0007695, Accumulated Leap Seconds=35
  { 2456578.50, 67.185957 }, //13 October 2013, UT1-UTC=-0.0019572, Accumulated Leap Seconds=35
  { 2456579.50, 67.187277 }, //14 October 2013, UT1-UTC=-0.0032770, Accumulated Leap Seconds=35
  { 2456580.50, 67.188719 }, //15 October 2013, UT1-UTC=-0.0047187, Accumulated Leap Seconds=35
  { 2456581.50, 67.190234 }, //16 October 2013, UT1-UTC=-0.0062343, Accumulated Leap Seconds=35
  { 2456582.50, 67.191748 }, //17 October 2013, UT1-UTC=-0.0077481, Accumulated Leap Seconds=35
  { 2456583.50, 67.193240 }, //18 October 2013, UT1-UTC=-0.0092401, Accumulated Leap Seconds=35
  { 2456584.50, 67.194596 }, //19 October 2013, UT1-UTC=-0.0105956, Accumulated Leap Seconds=35
  { 2456585.50, 67.195769 }, //20 October 2013, UT1-UTC=-0.0117687, Accumulated Leap Seconds=35
  { 2456586.50, 67.196766 }, //21 October 2013, UT1-UTC=-0.0127656, Accumulated Leap Seconds=35
  { 2456587.50, 67.197631 }, //22 October 2013, UT1-UTC=-0.0136313, Accumulated Leap Seconds=35
  { 2456588.50, 67.198420 }, //23 October 2013, UT1-UTC=-0.0144201, Accumulated Leap Seconds=35
  { 2456589.50, 67.199178 }, //24 October 2013, UT1-UTC=-0.0151783, Accumulated Leap Seconds=35
  { 2456590.50, 67.200025 }, //25 October 2013, UT1-UTC=-0.0160253, Accumulated Leap Seconds=35
  { 2456591.50, 67.201009 }, //26 October 2013, UT1-UTC=-0.0170091, Accumulated Leap Seconds=35
  { 2456592.50, 67.202064 }, //27 October 2013, UT1-UTC=-0.0180639, Accumulated Leap Seconds=35
  { 2456593.50, 67.203253 }, //28 October 2013, UT1-UTC=-0.0192528, Accumulated Leap Seconds=35
  { 2456594.50, 67.204606 }, //29 October 2013, UT1-UTC=-0.0206063, Accumulated Leap Seconds=35
  { 2456595.50, 67.206057 }, //30 October 2013, UT1-UTC=-0.0220570, Accumulated Leap Seconds=35
  { 2456596.50, 67.207586 }, //31 October 2013, UT1-UTC=-0.0235861, Accumulated Leap Seconds=35
  { 2456597.50, 67.209107 }, //01 November 2013, UT1-UTC=-0.0251069, Accumulated Leap Seconds=35
  { 2456598.50, 67.210525 }, //02 November 2013, UT1-UTC=-0.0265246, Accumulated Leap Seconds=35
  { 2456599.50, 67.211884 }, //03 November 2013, UT1-UTC=-0.0278841, Accumulated Leap Seconds=35
  { 2456600.50, 67.213209 }, //04 November 2013, UT1-UTC=-0.0292093, Accumulated Leap Seconds=35
  { 2456601.50, 67.214482 }, //05 November 2013, UT1-UTC=-0.0304821, Accumulated Leap Seconds=35
  { 2456602.50, 67.215726 }, //06 November 2013, UT1-UTC=-0.0317261, Accumulated Leap Seconds=35
  { 2456603.50, 67.216891 }, //07 November 2013, UT1-UTC=-0.0328912, Accumulated Leap Seconds=35
  { 2456604.50, 67.218033 }, //08 November 2013, UT1-UTC=-0.0340328, Accumulated Leap Seconds=35
  { 2456605.50, 67.219197 }, //09 November 2013, UT1-UTC=-0.0351967, Accumulated Leap Seconds=35
  { 2456606.50, 67.220445 }, //10 November 2013, UT1-UTC=-0.0364450, Accumulated Leap Seconds=35
  { 2456607.50, 67.221792 }, //11 November 2013, UT1-UTC=-0.0377924, Accumulated Leap Seconds=35
  { 2456608.50, 67.223229 }, //12 November 2013, UT1-UTC=-0.0392291, Accumulated Leap Seconds=35
  { 2456609.50, 67.224723 }, //13 November 2013, UT1-UTC=-0.0407233, Accumulated Leap Seconds=35
  { 2456610.50, 67.226188 }, //14 November 2013, UT1-UTC=-0.0421877, Accumulated Leap Seconds=35
  { 2456611.50, 67.227563 }, //15 November 2013, UT1-UTC=-0.0435631, Accumulated Leap Seconds=35
  { 2456612.50, 67.228772 }, //16 November 2013, UT1-UTC=-0.0447718, Accumulated Leap Seconds=35
  { 2456613.50, 67.229869 }, //17 November 2013, UT1-UTC=-0.0458689, Accumulated Leap Seconds=35
  { 2456614.50, 67.230872 }, //18 November 2013, UT1-UTC=-0.0468719, Accumulated Leap Seconds=35
  { 2456615.50, 67.231804 }, //19 November 2013, UT1-UTC=-0.0478037, Accumulated Leap Seconds=35
  { 2456616.50, 67.232691 }, //20 November 2013, UT1-UTC=-0.0486906, Accumulated Leap Seconds=35
  { 2456617.50, 67.233572 }, //21 November 2013, UT1-UTC=-0.0495717, Accumulated Leap Seconds=35
  { 2456618.50, 67.234453 }, //22 November 2013, UT1-UTC=-0.0504531, Accumulated Leap Seconds=35
  { 2456619.50, 67.235415 }, //23 November 2013, UT1-UTC=-0.0514154, Accumulated Leap Seconds=35
  { 2456620.50, 67.236543 }, //24 November 2013, UT1-UTC=-0.0525427, Accumulated Leap Seconds=35
  { 2456621.50, 67.237742 }, //25 November 2013, UT1-UTC=-0.0537417, Accumulated Leap Seconds=35
  { 2456622.50, 67.239000 }, //26 November 2013, UT1-UTC=-0.0549998, Accumulated Leap Seconds=35
  { 2456623.50, 67.240353 }, //27 November 2013, UT1-UTC=-0.0563533, Accumulated Leap Seconds=35
  { 2456624.50, 67.241785 }, //28 November 2013, UT1-UTC=-0.0577847, Accumulated Leap Seconds=35
  { 2456625.50, 67.243230 }, //29 November 2013, UT1-UTC=-0.0592297, Accumulated Leap Seconds=35
  { 2456626.50, 67.244613 }, //30 November 2013, UT1-UTC=-0.0606129, Accumulated Leap Seconds=35
  { 2456627.50, 67.245981 }, //01 December 2013, UT1-UTC=-0.0619812, Accumulated Leap Seconds=35
  { 2456628.50, 67.247276 }, //02 December 2013, UT1-UTC=-0.0632761, Accumulated Leap Seconds=35
  { 2456629.50, 67.248483 }, //03 December 2013, UT1-UTC=-0.0644831, Accumulated Leap Seconds=35
  { 2456630.50, 67.249597 }, //04 December 2013, UT1-UTC=-0.0655974, Accumulated Leap Seconds=35
  { 2456631.50, 67.250748 }, //05 December 2013, UT1-UTC=-0.0667483, Accumulated Leap Seconds=35
  { 2456632.50, 67.251949 }, //06 December 2013, UT1-UTC=-0.0679488, Accumulated Leap Seconds=35
  { 2456633.50, 67.253255 }, //07 December 2013, UT1-UTC=-0.0692549, Accumulated Leap Seconds=35
  { 2456634.50, 67.254643 }, //08 December 2013, UT1-UTC=-0.0706430, Accumulated Leap Seconds=35
  { 2456635.50, 67.256071 }, //09 December 2013, UT1-UTC=-0.0720710, Accumulated Leap Seconds=35
  { 2456636.50, 67.257465 }, //10 December 2013, UT1-UTC=-0.0734653, Accumulated Leap Seconds=35
  { 2456637.50, 67.258760 }, //11 December 2013, UT1-UTC=-0.0747597, Accumulated Leap Seconds=35
  { 2456638.50, 67.259962 }, //12 December 2013, UT1-UTC=-0.0759619, Accumulated Leap Seconds=35
  { 2456639.50, 67.261056 }, //13 December 2013, UT1-UTC=-0.0770561, Accumulated Leap Seconds=35
  { 2456640.50, 67.262041 }, //14 December 2013, UT1-UTC=-0.0780415, Accumulated Leap Seconds=35
  { 2456641.50, 67.262878 }, //15 December 2013, UT1-UTC=-0.0788776, Accumulated Leap Seconds=35
  { 2456642.50, 67.263601 }, //16 December 2013, UT1-UTC=-0.0796010, Accumulated Leap Seconds=35
  { 2456643.50, 67.264240 }, //17 December 2013, UT1-UTC=-0.0802402, Accumulated Leap Seconds=35
  { 2456644.50, 67.264834 }, //18 December 2013, UT1-UTC=-0.0808341, Accumulated Leap Seconds=35
  { 2456645.50, 67.265477 }, //19 December 2013, UT1-UTC=-0.0814769, Accumulated Leap Seconds=35
  { 2456646.50, 67.266195 }, //20 December 2013, UT1-UTC=-0.0821954, Accumulated Leap Seconds=35
  { 2456647.50, 67.267042 }, //21 December 2013, UT1-UTC=-0.0830417, Accumulated Leap Seconds=35
  { 2456648.50, 67.268036 }, //22 December 2013, UT1-UTC=-0.0840356, Accumulated Leap Seconds=35
  { 2456649.50, 67.269159 }, //23 December 2013, UT1-UTC=-0.0851591, Accumulated Leap Seconds=35
  { 2456650.50, 67.270404 }, //24 December 2013, UT1-UTC=-0.0864035, Accumulated Leap Seconds=35
  { 2456651.50, 67.271727 }, //25 December 2013, UT1-UTC=-0.0877268, Accumulated Leap Seconds=35
  { 2456652.50, 67.273100 }, //26 December 2013, UT1-UTC=-0.0890999, Accumulated Leap Seconds=35
  { 2456653.50, 67.274528 }, //27 December 2013, UT1-UTC=-0.0905278, Accumulated Leap Seconds=35
  { 2456654.50, 67.275954 }, //28 December 2013, UT1-UTC=-0.0919538, Accumulated Leap Seconds=35
  { 2456655.50, 67.277349 }, //29 December 2013, UT1-UTC=-0.0933492, Accumulated Leap Seconds=35
  { 2456656.50, 67.278658 }, //30 December 2013, UT1-UTC=-0.0946579, Accumulated Leap Seconds=35
  { 2456657.50, 67.279880 }, //31 December 2013, UT1-UTC=-0.0958804, Accumulated Leap Seconds=35
  { 2456658.50, 67.281038 }, //01 January 2014, UT1-UTC=-0.0970383, Accumulated Leap Seconds=35
  { 2456659.50, 67.282234 }, //02 January 2014, UT1-UTC=-0.0982338, Accumulated Leap Seconds=35
  { 2456660.50, 67.283509 }, //03 January 2014, UT1-UTC=-0.0995090, Accumulated Leap Seconds=35
  { 2456661.50, 67.284833 }, //04 January 2014, UT1-UTC=-0.1008335, Accumulated Leap Seconds=35
  { 2456662.50, 67.286236 }, //05 January 2014, UT1-UTC=-0.1022355, Accumulated Leap Seconds=35
  { 2456663.50, 67.287653 }, //06 January 2014, UT1-UTC=-0.1036535, Accumulated Leap Seconds=35
  { 2456664.50, 67.288989 }, //07 January 2014, UT1-UTC=-0.1049890, Accumulated Leap Seconds=35
  { 2456665.50, 67.290171 }, //08 January 2014, UT1-UTC=-0.1061710, Accumulated Leap Seconds=35
  { 2456666.50, 67.291190 }, //09 January 2014, UT1-UTC=-0.1071896, Accumulated Leap Seconds=35
  { 2456667.50, 67.292090 }, //10 January 2014, UT1-UTC=-0.1080903, Accumulated Leap Seconds=35
  { 2456668.50, 67.292849 }, //11 January 2014, UT1-UTC=-0.1088486, Accumulated Leap Seconds=35
  { 2456669.50, 67.293455 }, //12 January 2014, UT1-UTC=-0.1094555, Accumulated Leap Seconds=35
  { 2456670.50, 67.294027 }, //13 January 2014, UT1-UTC=-0.1100272, Accumulated Leap Seconds=35
  { 2456671.50, 67.294576 }, //14 January 2014, UT1-UTC=-0.1105759, Accumulated Leap Seconds=35
  { 2456672.50, 67.295158 }, //15 January 2014, UT1-UTC=-0.1111581, Accumulated Leap Seconds=35
  { 2456673.50, 67.295814 }, //16 January 2014, UT1-UTC=-0.1118144, Accumulated Leap Seconds=35
  { 2456674.50, 67.296595 }, //17 January 2014, UT1-UTC=-0.1125953, Accumulated Leap Seconds=35
  { 2456675.50, 67.297498 }, //18 January 2014, UT1-UTC=-0.1134977, Accumulated Leap Seconds=35
  { 2456676.50, 67.298549 }, //19 January 2014, UT1-UTC=-0.1145490, Accumulated Leap Seconds=35
  { 2456677.50, 67.299717 }, //20 January 2014, UT1-UTC=-0.1157167, Accumulated Leap Seconds=35
  { 2456678.50, 67.300949 }, //21 January 2014, UT1-UTC=-0.1169494, Accumulated Leap Seconds=35
  { 2456679.50, 67.302229 }, //22 January 2014, UT1-UTC=-0.1182290, Accumulated Leap Seconds=35
  { 2456680.50, 67.303515 }, //23 January 2014, UT1-UTC=-0.1195153, Accumulated Leap Seconds=35
  { 2456681.50, 67.304723 }, //24 January 2014, UT1-UTC=-0.1207227, Accumulated Leap Seconds=35
  { 2456682.50, 67.305832 }, //25 January 2014, UT1-UTC=-0.1218321, Accumulated Leap Seconds=35
  { 2456683.50, 67.306891 }, //26 January 2014, UT1-UTC=-0.1228909, Accumulated Leap Seconds=35
  { 2456684.50, 67.307850 }, //27 January 2014, UT1-UTC=-0.1238500, Accumulated Leap Seconds=35
  { 2456685.50, 67.308794 }, //28 January 2014, UT1-UTC=-0.1247940, Accumulated Leap Seconds=35
  { 2456686.50, 67.309816 }, //29 January 2014, UT1-UTC=-0.1258157, Accumulated Leap Seconds=35
  { 2456687.50, 67.310961 }, //30 January 2014, UT1-UTC=-0.1269615, Accumulated Leap Seconds=35
  { 2456688.50, 67.312264 }, //31 January 2014, UT1-UTC=-0.1282637, Accumulated Leap Seconds=35
  { 2456689.50, 67.313645 }, //01 February 2014, UT1-UTC=-0.1296452, Accumulated Leap Seconds=35
  { 2456690.50, 67.315052 }, //02 February 2014, UT1-UTC=-0.1310519, Accumulated Leap Seconds=35
  { 2456691.50, 67.316450 }, //03 February 2014, UT1-UTC=-0.1324499, Accumulated Leap Seconds=35
  { 2456692.50, 67.317737 }, //04 February 2014, UT1-UTC=-0.1337375, Accumulated Leap Seconds=35
  { 2456693.50, 67.318870 }, //05 February 2014, UT1-UTC=-0.1348701, Accumulated Leap Seconds=35
  { 2456694.50, 67.319775 }, //06 February 2014, UT1-UTC=-0.1357755, Accumulated Leap Seconds=35
  { 2456695.50, 67.320546 }, //07 February 2014, UT1-UTC=-0.1365457, Accumulated Leap Seconds=35
  { 2456696.50, 67.321227 }, //08 February 2014, UT1-UTC=-0.1372273, Accumulated Leap Seconds=35
  { 2456697.50, 67.321817 }, //09 February 2014, UT1-UTC=-0.1378173, Accumulated Leap Seconds=35
  { 2456698.50, 67.322382 }, //10 February 2014, UT1-UTC=-0.1383820, Accumulated Leap Seconds=35
  { 2456699.50, 67.323000 }, //11 February 2014, UT1-UTC=-0.1390005, Accumulated Leap Seconds=35
  { 2456700.50, 67.323669 }, //12 February 2014, UT1-UTC=-0.1396690, Accumulated Leap Seconds=35
  { 2456701.50, 67.324469 }, //13 February 2014, UT1-UTC=-0.1404686, Accumulated Leap Seconds=35
  { 2456702.50, 67.325399 }, //14 February 2014, UT1-UTC=-0.1413991, Accumulated Leap Seconds=35
  { 2456703.50, 67.326449 }, //15 February 2014, UT1-UTC=-0.1424488, Accumulated Leap Seconds=35
  { 2456704.50, 67.327646 }, //16 February 2014, UT1-UTC=-0.1436459, Accumulated Leap Seconds=35
  { 2456705.50, 67.328945 }, //17 February 2014, UT1-UTC=-0.1449452, Accumulated Leap Seconds=35
  { 2456706.50, 67.330308 }, //18 February 2014, UT1-UTC=-0.1463082, Accumulated Leap Seconds=35
  { 2456707.50, 67.331706 }, //19 February 2014, UT1-UTC=-0.1477058, Accumulated Leap Seconds=35
  { 2456708.50, 67.333058 }, //20 February 2014, UT1-UTC=-0.1490579, Accumulated Leap Seconds=35
  { 2456709.50, 67.334344 }, //21 February 2014, UT1-UTC=-0.1503436, Accumulated Leap Seconds=35
  { 2456710.50, 67.335542 }, //22 February 2014, UT1-UTC=-0.1515416, Accumulated Leap Seconds=35
  { 2456711.50, 67.336676 }, //23 February 2014, UT1-UTC=-0.1526764, Accumulated Leap Seconds=35
  { 2456712.50, 67.337818 }, //24 February 2014, UT1-UTC=-0.1538181, Accumulated Leap Seconds=35
  { 2456713.50, 67.339036 }, //25 February 2014, UT1-UTC=-0.1550359, Accumulated Leap Seconds=35
  { 2456714.50, 67.340392 }, //26 February 2014, UT1-UTC=-0.1563919, Accumulated Leap Seconds=35
  { 2456715.50, 67.341929 }, //27 February 2014, UT1-UTC=-0.1579295, Accumulated Leap Seconds=35
  { 2456716.50, 67.343734 }, //28 February 2014, UT1-UTC=-0.1597342, Accumulated Leap Seconds=35
  { 2456717.50, 67.345697 }, //01 March 2014, UT1-UTC=-0.1616968, Accumulated Leap Seconds=35
  { 2456718.50, 67.347690 }, //02 March 2014, UT1-UTC=-0.1636895, Accumulated Leap Seconds=35
  { 2456719.50, 67.349660 }, //03 March 2014, UT1-UTC=-0.1656599, Accumulated Leap Seconds=35
  { 2456720.50, 67.351522 }, //04 March 2014, UT1-UTC=-0.1675222, Accumulated Leap Seconds=35
  { 2456721.50, 67.353220 }, //05 March 2014, UT1-UTC=-0.1692196, Accumulated Leap Seconds=35
  { 2456722.50, 67.354739 }, //06 March 2014, UT1-UTC=-0.1707389, Accumulated Leap Seconds=35
  { 2456723.50, 67.356109 }, //07 March 2014, UT1-UTC=-0.1721094, Accumulated Leap Seconds=35
  { 2456724.50, 67.357344 }, //08 March 2014, UT1-UTC=-0.1733443, Accumulated Leap Seconds=35
  { 2456725.50, 67.358469 }, //09 March 2014, UT1-UTC=-0.1744687, Accumulated Leap Seconds=35
  { 2456726.50, 67.359548 }, //10 March 2014, UT1-UTC=-0.1755480, Accumulated Leap Seconds=35
  { 2456727.50, 67.360624 }, //11 March 2014, UT1-UTC=-0.1766243, Accumulated Leap Seconds=35
  { 2456728.50, 67.361674 }, //12 March 2014, UT1-UTC=-0.1776736, Accumulated Leap Seconds=35
  { 2456729.50, 67.362779 }, //13 March 2014, UT1-UTC=-0.1787788, Accumulated Leap Seconds=35
  { 2456730.50, 67.363977 }, //14 March 2014, UT1-UTC=-0.1799770, Accumulated Leap Seconds=35
  { 2456731.50, 67.365236 }, //15 March 2014, UT1-UTC=-0.1812363, Accumulated Leap Seconds=35
  { 2456732.50, 67.366513 }, //16 March 2014, UT1-UTC=-0.1825130, Accumulated Leap Seconds=35
  { 2456733.50, 67.367786 }, //17 March 2014, UT1-UTC=-0.1837860, Accumulated Leap Seconds=35
  { 2456734.50, 67.369030 }, //18 March 2014, UT1-UTC=-0.1850303, Accumulated Leap Seconds=35
  { 2456735.50, 67.370253 }, //19 March 2014, UT1-UTC=-0.1862528, Accumulated Leap Seconds=35
  { 2456736.50, 67.371441 }, //20 March 2014, UT1-UTC=-0.1874414, Accumulated Leap Seconds=35
  { 2456737.50, 67.372584 }, //21 March 2014, UT1-UTC=-0.1885837, Accumulated Leap Seconds=35
  { 2456738.50, 67.373681 }, //22 March 2014, UT1-UTC=-0.1896806, Accumulated Leap Seconds=35
  { 2456739.50, 67.374820 }, //23 March 2014, UT1-UTC=-0.1908195, Accumulated Leap Seconds=35
  { 2456740.50, 67.376046 }, //24 March 2014, UT1-UTC=-0.1920456, Accumulated Leap Seconds=35
  { 2456741.50, 67.377349 }, //25 March 2014, UT1-UTC=-0.1933490, Accumulated Leap Seconds=35
  { 2456742.50, 67.378788 }, //26 March 2014, UT1-UTC=-0.1947880, Accumulated Leap Seconds=35
  { 2456743.50, 67.380400 }, //27 March 2014, UT1-UTC=-0.1964000, Accumulated Leap Seconds=35
  { 2456744.50, 67.382100 }, //28 March 2014, UT1-UTC=-0.1981001, Accumulated Leap Seconds=35
  { 2456745.50, 67.383876 }, //29 March 2014, UT1-UTC=-0.1998763, Accumulated Leap Seconds=35
  { 2456746.50, 67.385679 }, //30 March 2014, UT1-UTC=-0.2016788, Accumulated Leap Seconds=35
  { 2456747.50, 67.387414 }, //31 March 2014, UT1-UTC=-0.2034139, Accumulated Leap Seconds=35
  { 2456748.50, 67.389003 }, //01 April 2014, UT1-UTC=-0.2050030, Accumulated Leap Seconds=35
  { 2456749.50, 67.390461 }, //02 April 2014, UT1-UTC=-0.2064612, Accumulated Leap Seconds=35
  { 2456750.50, 67.391827 }, //03 April 2014, UT1-UTC=-0.2078270, Accumulated Leap Seconds=35
  { 2456751.50, 67.393078 }, //04 April 2014, UT1-UTC=-0.2090778, Accumulated Leap Seconds=35
  { 2456752.50, 67.394267 }, //05 April 2014, UT1-UTC=-0.2102667, Accumulated Leap Seconds=35
  { 2456753.50, 67.395479 }, //06 April 2014, UT1-UTC=-0.2114790, Accumulated Leap Seconds=35
  { 2456754.50, 67.396679 }, //07 April 2014, UT1-UTC=-0.2126790, Accumulated Leap Seconds=35
  { 2456755.50, 67.397868 }, //08 April 2014, UT1-UTC=-0.2138681, Accumulated Leap Seconds=35
  { 2456756.50, 67.399068 }, //09 April 2014, UT1-UTC=-0.2150684, Accumulated Leap Seconds=35
  { 2456757.50, 67.400344 }, //10 April 2014, UT1-UTC=-0.2163440, Accumulated Leap Seconds=35
  { 2456758.50, 67.401669 }, //11 April 2014, UT1-UTC=-0.2176690, Accumulated Leap Seconds=35
  { 2456759.50, 67.403072 }, //12 April 2014, UT1-UTC=-0.2190724, Accumulated Leap Seconds=35
  { 2456760.50, 67.404490 }, //13 April 2014, UT1-UTC=-0.2204899, Accumulated Leap Seconds=35
  { 2456761.50, 67.405974 }, //14 April 2014, UT1-UTC=-0.2219743, Accumulated Leap Seconds=35
  { 2456762.50, 67.407426 }, //15 April 2014, UT1-UTC=-0.2234264, Accumulated Leap Seconds=35
  { 2456763.50, 67.408722 }, //16 April 2014, UT1-UTC=-0.2247221, Accumulated Leap Seconds=35
  { 2456764.50, 67.409929 }, //17 April 2014, UT1-UTC=-0.2259292, Accumulated Leap Seconds=35
  { 2456765.50, 67.411131 }, //18 April 2014, UT1-UTC=-0.2271312, Accumulated Leap Seconds=35
  { 2456766.50, 67.412324 }, //19 April 2014, UT1-UTC=-0.2283242, Accumulated Leap Seconds=35
  { 2456767.50, 67.413567 }, //20 April 2014, UT1-UTC=-0.2295665, Accumulated Leap Seconds=35
  { 2456768.50, 67.414867 }, //21 April 2014, UT1-UTC=-0.2308673, Accumulated Leap Seconds=35
  { 2456769.50, 67.416296 }, //22 April 2014, UT1-UTC=-0.2322960, Accumulated Leap Seconds=35
  { 2456770.50, 67.417898 }, //23 April 2014, UT1-UTC=-0.2338981, Accumulated Leap Seconds=35
  { 2456771.50, 67.419623 }, //24 April 2014, UT1-UTC=-0.2356233, Accumulated Leap Seconds=35
  { 2456772.50, 67.421478 }, //25 April 2014, UT1-UTC=-0.2374777, Accumulated Leap Seconds=35
  { 2456773.50, 67.423475 }, //26 April 2014, UT1-UTC=-0.2394754, Accumulated Leap Seconds=35
  { 2456774.50, 67.425458 }, //27 April 2014, UT1-UTC=-0.2414582, Accumulated Leap Seconds=35
  { 2456775.50, 67.427320 }, //28 April 2014, UT1-UTC=-0.2433201, Accumulated Leap Seconds=35
  { 2456776.50, 67.429039 }, //29 April 2014, UT1-UTC=-0.2450388, Accumulated Leap Seconds=35
  { 2456777.50, 67.430537 }, //30 April 2014, UT1-UTC=-0.2465370, Accumulated Leap Seconds=35
  { 2456778.50, 67.431843 }, //01 May 2014, UT1-UTC=-0.2478433, Accumulated Leap Seconds=35
  { 2456779.50, 67.433011 }, //02 May 2014, UT1-UTC=-0.2490107, Accumulated Leap Seconds=35
  { 2456780.50, 67.434133 }, //03 May 2014, UT1-UTC=-0.2501332, Accumulated Leap Seconds=35
  { 2456781.50, 67.435233 }, //04 May 2014, UT1-UTC=-0.2512326, Accumulated Leap Seconds=35
  { 2456782.50, 67.436357 }, //05 May 2014, UT1-UTC=-0.2523575, Accumulated Leap Seconds=35
  { 2456783.50, 67.437537 }, //06 May 2014, UT1-UTC=-0.2535367, Accumulated Leap Seconds=35
  { 2456784.50, 67.438785 }, //07 May 2014, UT1-UTC=-0.2547854, Accumulated Leap Seconds=35
  { 2456785.50, 67.440106 }, //08 May 2014, UT1-UTC=-0.2561064, Accumulated Leap Seconds=35
  { 2456786.50, 67.441521 }, //09 May 2014, UT1-UTC=-0.2575214, Accumulated Leap Seconds=35
  { 2456787.50, 67.442953 }, //10 May 2014, UT1-UTC=-0.2589526, Accumulated Leap Seconds=35
  { 2456788.50, 67.444405 }, //11 May 2014, UT1-UTC=-0.2604045, Accumulated Leap Seconds=35
  { 2456789.50, 67.445797 }, //12 May 2014, UT1-UTC=-0.2617971, Accumulated Leap Seconds=35
  { 2456790.50, 67.447121 }, //13 May 2014, UT1-UTC=-0.2631211, Accumulated Leap Seconds=35
  { 2456791.50, 67.448295 }, //14 May 2014, UT1-UTC=-0.2642948, Accumulated Leap Seconds=35
  { 2456792.50, 67.449405 }, //15 May 2014, UT1-UTC=-0.2654046, Accumulated Leap Seconds=35
  { 2456793.50, 67.450433 }, //16 May 2014, UT1-UTC=-0.2664327, Accumulated Leap Seconds=35
  { 2456794.50, 67.451429 }, //17 May 2014, UT1-UTC=-0.2674294, Accumulated Leap Seconds=35
  { 2456795.50, 67.452474 }, //18 May 2014, UT1-UTC=-0.2684738, Accumulated Leap Seconds=35
  { 2456796.50, 67.453612 }, //19 May 2014, UT1-UTC=-0.2696121, Accumulated Leap Seconds=35
  { 2456797.50, 67.454892 }, //20 May 2014, UT1-UTC=-0.2708920, Accumulated Leap Seconds=35
  { 2456798.50, 67.456293 }, //21 May 2014, UT1-UTC=-0.2722925, Accumulated Leap Seconds=35
  { 2456799.50, 67.457770 }, //22 May 2014, UT1-UTC=-0.2737699, Accumulated Leap Seconds=35
  { 2456800.50, 67.459218 }, //23 May 2014, UT1-UTC=-0.2752180, Accumulated Leap Seconds=35
  { 2456801.50, 67.460642 }, //24 May 2014, UT1-UTC=-0.2766420, Accumulated Leap Seconds=35
  { 2456802.50, 67.461920 }, //25 May 2014, UT1-UTC=-0.2779204, Accumulated Leap Seconds=35
  { 2456803.50, 67.463023 }, //26 May 2014, UT1-UTC=-0.2790232, Accumulated Leap Seconds=35
  { 2456804.50, 67.463955 }, //27 May 2014, UT1-UTC=-0.2799550, Accumulated Leap Seconds=35
  { 2456805.50, 67.464711 }, //28 May 2014, UT1-UTC=-0.2807109, Accumulated Leap Seconds=35
  { 2456806.50, 67.465341 }, //29 May 2014, UT1-UTC=-0.2813415, Accumulated Leap Seconds=35
  { 2456807.50, 67.465827 }, //30 May 2014, UT1-UTC=-0.2818273, Accumulated Leap Seconds=35
  { 2456808.50, 67.466232 }, //31 May 2014, UT1-UTC=-0.2822320, Accumulated Leap Seconds=35
  { 2456809.50, 67.466621 }, //01 June 2014, UT1-UTC=-0.2826209, Accumulated Leap Seconds=35
  { 2456810.50, 67.467039 }, //02 June 2014, UT1-UTC=-0.2830388, Accumulated Leap Seconds=35
  { 2456811.50, 67.467486 }, //03 June 2014, UT1-UTC=-0.2834856, Accumulated Leap Seconds=35
  { 2456812.50, 67.467976 }, //04 June 2014, UT1-UTC=-0.2839761, Accumulated Leap Seconds=35
  { 2456813.50, 67.468584 }, //05 June 2014, UT1-UTC=-0.2845840, Accumulated Leap Seconds=35
  { 2456814.50, 67.469218 }, //06 June 2014, UT1-UTC=-0.2852180, Accumulated Leap Seconds=35
  { 2456815.50, 67.469951 }, //07 June 2014, UT1-UTC=-0.2859513, Accumulated Leap Seconds=35
  { 2456816.50, 67.470692 }, //08 June 2014, UT1-UTC=-0.2866921, Accumulated Leap Seconds=35
  { 2456817.50, 67.471393 }, //09 June 2014, UT1-UTC=-0.2873932, Accumulated Leap Seconds=35
  { 2456818.50, 67.472067 }, //10 June 2014, UT1-UTC=-0.2880675, Accumulated Leap Seconds=35
  { 2456819.50, 67.472762 }, //11 June 2014, UT1-UTC=-0.2887623, Accumulated Leap Seconds=35
  { 2456820.50, 67.473396 }, //12 June 2014, UT1-UTC=-0.2893958, Accumulated Leap Seconds=35
  { 2456821.50, 67.474041 }, //13 June 2014, UT1-UTC=-0.2900413, Accumulated Leap Seconds=35
  { 2456822.50, 67.474694 }, //14 June 2014, UT1-UTC=-0.2906944, Accumulated Leap Seconds=35
  { 2456823.50, 67.475433 }, //15 June 2014, UT1-UTC=-0.2914330, Accumulated Leap Seconds=35
  { 2456824.50, 67.476261 }, //16 June 2014, UT1-UTC=-0.2922606, Accumulated Leap Seconds=35
  { 2456825.50, 67.477230 }, //17 June 2014, UT1-UTC=-0.2932302, Accumulated Leap Seconds=35
  { 2456826.50, 67.478370 }, //18 June 2014, UT1-UTC=-0.2943704, Accumulated Leap Seconds=35
  { 2456827.50, 67.479504 }, //19 June 2014, UT1-UTC=-0.2955044, Accumulated Leap Seconds=35
  { 2456828.50, 67.480679 }, //20 June 2014, UT1-UTC=-0.2966786, Accumulated Leap Seconds=35
  { 2456829.50, 67.481746 }, //21 June 2014, UT1-UTC=-0.2977463, Accumulated Leap Seconds=35
  { 2456830.50, 67.482677 }, //22 June 2014, UT1-UTC=-0.2986769, Accumulated Leap Seconds=35
  { 2456831.50, 67.483400 }, //23 June 2014, UT1-UTC=-0.2994001, Accumulated Leap Seconds=35
  { 2456832.50, 67.483990 }, //24 June 2014, UT1-UTC=-0.2999901, Accumulated Leap Seconds=35
  { 2456833.50, 67.484390 }, //25 June 2014, UT1-UTC=-0.3003898, Accumulated Leap Seconds=35
  { 2456834.50, 67.484713 }, //26 June 2014, UT1-UTC=-0.3007127, Accumulated Leap Seconds=35
  { 2456835.50, 67.484870 }, //27 June 2014, UT1-UTC=-0.3008696, Accumulated Leap Seconds=35
  { 2456836.50, 67.485012 }, //28 June 2014, UT1-UTC=-0.3010117, Accumulated Leap Seconds=35
  { 2456837.50, 67.485232 }, //29 June 2014, UT1-UTC=-0.3012320, Accumulated Leap Seconds=35
  { 2456838.50, 67.485519 }, //30 June 2014, UT1-UTC=-0.3015187, Accumulated Leap Seconds=35
  { 2456839.50, 67.485846 }, //01 July 2014, UT1-UTC=-0.3018459, Accumulated Leap Seconds=35
  { 2456840.50, 67.486218 }, //02 July 2014, UT1-UTC=-0.3022180, Accumulated Leap Seconds=35
  { 2456841.50, 67.486686 }, //03 July 2014, UT1-UTC=-0.3026865, Accumulated Leap Seconds=35
  { 2456842.50, 67.487236 }, //04 July 2014, UT1-UTC=-0.3032358, Accumulated Leap Seconds=35
  { 2456843.50, 67.487772 }, //05 July 2014, UT1-UTC=-0.3037720, Accumulated Leap Seconds=35
  { 2456844.50, 67.488274 }, //06 July 2014, UT1-UTC=-0.3042743, Accumulated Leap Seconds=35
  { 2456845.50, 67.488703 }, //07 July 2014, UT1-UTC=-0.3047034, Accumulated Leap Seconds=35
  { 2456846.50, 67.489050 }, //08 July 2014, UT1-UTC=-0.3050498, Accumulated Leap Seconds=35
  { 2456847.50, 67.489353 }, //09 July 2014, UT1-UTC=-0.3053529, Accumulated Leap Seconds=35
  { 2456848.50, 67.489617 }, //10 July 2014, UT1-UTC=-0.3056174, Accumulated Leap Seconds=35
  { 2456849.50, 67.489971 }, //11 July 2014, UT1-UTC=-0.3059711, Accumulated Leap Seconds=35
  { 2456850.50, 67.490370 }, //12 July 2014, UT1-UTC=-0.3063702, Accumulated Leap Seconds=35
  { 2456851.50, 67.490912 }, //13 July 2014, UT1-UTC=-0.3069123, Accumulated Leap Seconds=35
  { 2456852.50, 67.491626 }, //14 July 2014, UT1-UTC=-0.3076262, Accumulated Leap Seconds=35
  { 2456853.50, 67.492508 }, //15 July 2014, UT1-UTC=-0.3085082, Accumulated Leap Seconds=35
  { 2456854.50, 67.493416 }, //16 July 2014, UT1-UTC=-0.3094163, Accumulated Leap Seconds=35
  { 2456855.50, 67.494345 }, //17 July 2014, UT1-UTC=-0.3103450, Accumulated Leap Seconds=35
  { 2456856.50, 67.495170 }, //18 July 2014, UT1-UTC=-0.3111697, Accumulated Leap Seconds=35
  { 2456857.50, 67.495876 }, //19 July 2014, UT1-UTC=-0.3118756, Accumulated Leap Seconds=35
  { 2456858.50, 67.496422 }, //20 July 2014, UT1-UTC=-0.3124224, Accumulated Leap Seconds=35
  { 2456859.50, 67.496771 }, //21 July 2014, UT1-UTC=-0.3127713, Accumulated Leap Seconds=35
  { 2456860.50, 67.496985 }, //22 July 2014, UT1-UTC=-0.3129852, Accumulated Leap Seconds=35
  { 2456861.50, 67.497090 }, //23 July 2014, UT1-UTC=-0.3130897, Accumulated Leap Seconds=35
  { 2456862.50, 67.497107 }, //24 July 2014, UT1-UTC=-0.3131074, Accumulated Leap Seconds=35
  { 2456863.50, 67.497146 }, //25 July 2014, UT1-UTC=-0.3131465, Accumulated Leap Seconds=35
  { 2456864.50, 67.497177 }, //26 July 2014, UT1-UTC=-0.3131767, Accumulated Leap Seconds=35
  { 2456865.50, 67.497268 }, //27 July 2014, UT1-UTC=-0.3132680, Accumulated Leap Seconds=35
  { 2456866.50, 67.497426 }, //28 July 2014, UT1-UTC=-0.3134264, Accumulated Leap Seconds=35
  { 2456867.50, 67.497654 }, //29 July 2014, UT1-UTC=-0.3136543, Accumulated Leap Seconds=35
  { 2456868.50, 67.498028 }, //30 July 2014, UT1-UTC=-0.3140280, Accumulated Leap Seconds=35
  { 2456869.50, 67.498460 }, //31 July 2014, UT1-UTC=-0.3144596, Accumulated Leap Seconds=35
  { 2456870.50, 67.498915 }, //01 August 2014, UT1-UTC=-0.3149147, Accumulated Leap Seconds=35
  { 2456871.50, 67.499317 }, //02 August 2014, UT1-UTC=-0.3153168, Accumulated Leap Seconds=35
  { 2456872.50, 67.499648 }, //03 August 2014, UT1-UTC=-0.3156480, Accumulated Leap Seconds=35
  { 2456873.50, 67.499924 }, //04 August 2014, UT1-UTC=-0.3159244, Accumulated Leap Seconds=35
  { 2456874.50, 67.500111 }, //05 August 2014, UT1-UTC=-0.3161113, Accumulated Leap Seconds=35
  { 2456875.50, 67.500239 }, //06 August 2014, UT1-UTC=-0.3162391, Accumulated Leap Seconds=35
  { 2456876.50, 67.500321 }, //07 August 2014, UT1-UTC=-0.3163210, Accumulated Leap Seconds=35
  { 2456877.50, 67.500406 }, //08 August 2014, UT1-UTC=-0.3164063, Accumulated Leap Seconds=35
  { 2456878.50, 67.500658 }, //09 August 2014, UT1-UTC=-0.3166580, Accumulated Leap Seconds=35
  { 2456879.50, 67.501055 }, //10 August 2014, UT1-UTC=-0.3170554, Accumulated Leap Seconds=35
  { 2456880.50, 67.501657 }, //11 August 2014, UT1-UTC=-0.3176567, Accumulated Leap Seconds=35
  { 2456881.50, 67.502368 }, //12 August 2014, UT1-UTC=-0.3183680, Accumulated Leap Seconds=35
  { 2456882.50, 67.503187 }, //13 August 2014, UT1-UTC=-0.3191870, Accumulated Leap Seconds=35
  { 2456883.50, 67.504041 }, //14 August 2014, UT1-UTC=-0.3200406, Accumulated Leap Seconds=35
  { 2456884.50, 67.504799 }, //15 August 2014, UT1-UTC=-0.3207990, Accumulated Leap Seconds=35
  { 2456885.50, 67.505429 }, //16 August 2014, UT1-UTC=-0.3214286, Accumulated Leap Seconds=35
  { 2456886.50, 67.505928 }, //17 August 2014, UT1-UTC=-0.3219281, Accumulated Leap Seconds=35
  { 2456887.50, 67.506346 }, //18 August 2014, UT1-UTC=-0.3223465, Accumulated Leap Seconds=35
  { 2456888.50, 67.506664 }, //19 August 2014, UT1-UTC=-0.3226636, Accumulated Leap Seconds=35
  { 2456889.50, 67.506854 }, //20 August 2014, UT1-UTC=-0.3228541, Accumulated Leap Seconds=35
  { 2456890.50, 67.507049 }, //21 August 2014, UT1-UTC=-0.3230495, Accumulated Leap Seconds=35
  { 2456891.50, 67.507215 }, //22 August 2014, UT1-UTC=-0.3232151, Accumulated Leap Seconds=35
  { 2456892.50, 67.507438 }, //23 August 2014, UT1-UTC=-0.3234376, Accumulated Leap Seconds=35
  { 2456893.50, 67.507669 }, //24 August 2014, UT1-UTC=-0.3236686, Accumulated Leap Seconds=35
  { 2456894.50, 67.507899 }, //25 August 2014, UT1-UTC=-0.3238992, Accumulated Leap Seconds=35
  { 2456895.50, 67.508194 }, //26 August 2014, UT1-UTC=-0.3241938, Accumulated Leap Seconds=35
  { 2456896.50, 67.508590 }, //27 August 2014, UT1-UTC=-0.3245904, Accumulated Leap Seconds=35
  { 2456897.50, 67.509072 }, //28 August 2014, UT1-UTC=-0.3250723, Accumulated Leap Seconds=35
  { 2456898.50, 67.509570 }, //29 August 2014, UT1-UTC=-0.3255697, Accumulated Leap Seconds=35
  { 2456899.50, 67.510057 }, //30 August 2014, UT1-UTC=-0.3260573, Accumulated Leap Seconds=35
  { 2456900.50, 67.510587 }, //31 August 2014, UT1-UTC=-0.3265866, Accumulated Leap Seconds=35
  { 2456901.50, 67.511094 }, //01 September 2014, UT1-UTC=-0.3270936, Accumulated Leap Seconds=35
  { 2456902.50, 67.511619 }, //02 September 2014, UT1-UTC=-0.3276187, Accumulated Leap Seconds=35
  { 2456903.50, 67.512138 }, //03 September 2014, UT1-UTC=-0.3281385, Accumulated Leap Seconds=35
  { 2456904.50, 67.512692 }, //04 September 2014, UT1-UTC=-0.3286925, Accumulated Leap Seconds=35
  { 2456905.50, 67.513271 }, //05 September 2014, UT1-UTC=-0.3292707, Accumulated Leap Seconds=35
  { 2456906.50, 67.514045 }, //06 September 2014, UT1-UTC=-0.3300446, Accumulated Leap Seconds=35
  { 2456907.50, 67.514979 }, //07 September 2014, UT1-UTC=-0.3309788, Accumulated Leap Seconds=35
  { 2456908.50, 67.516107 }, //08 September 2014, UT1-UTC=-0.3321068, Accumulated Leap Seconds=35
  { 2456909.50, 67.517386 }, //09 September 2014, UT1-UTC=-0.3333858, Accumulated Leap Seconds=35
  { 2456910.50, 67.518691 }, //10 September 2014, UT1-UTC=-0.3346911, Accumulated Leap Seconds=35
  { 2456911.50, 67.519894 }, //11 September 2014, UT1-UTC=-0.3358939, Accumulated Leap Seconds=35
  { 2456912.50, 67.520934 }, //12 September 2014, UT1-UTC=-0.3369338, Accumulated Leap Seconds=35
  { 2456913.50, 67.521770 }, //13 September 2014, UT1-UTC=-0.3377698, Accumulated Leap Seconds=35
  { 2456914.50, 67.522477 }, //14 September 2014, UT1-UTC=-0.3384770, Accumulated Leap Seconds=35
  { 2456915.50, 67.523096 }, //15 September 2014, UT1-UTC=-0.3390958, Accumulated Leap Seconds=35
  { 2456916.50, 67.523670 }, //16 September 2014, UT1-UTC=-0.3396704, Accumulated Leap Seconds=35
  { 2456917.50, 67.524259 }, //17 September 2014, UT1-UTC=-0.3402588, Accumulated Leap Seconds=35
  { 2456918.50, 67.524930 }, //18 September 2014, UT1-UTC=-0.3409302, Accumulated Leap Seconds=35
  { 2456919.50, 67.525610 }, //19 September 2014, UT1-UTC=-0.3416097, Accumulated Leap Seconds=35
  { 2456920.50, 67.526311 }, //20 September 2014, UT1-UTC=-0.3423108, Accumulated Leap Seconds=35
  { 2456921.50, 67.527060 }, //21 September 2014, UT1-UTC=-0.3430601, Accumulated Leap Seconds=35
  { 2456922.50, 67.527905 }, //22 September 2014, UT1-UTC=-0.3439051, Accumulated Leap Seconds=35
  { 2456923.50, 67.528834 }, //23 September 2014, UT1-UTC=-0.3448340, Accumulated Leap Seconds=35
  { 2456924.50, 67.529702 }, //24 September 2014, UT1-UTC=-0.3457021, Accumulated Leap Seconds=35
  { 2456925.50, 67.530665 }, //25 September 2014, UT1-UTC=-0.3466653, Accumulated Leap Seconds=35
  { 2456926.50, 67.531574 }, //26 September 2014, UT1-UTC=-0.3475740, Accumulated Leap Seconds=35
  { 2456927.50, 67.532366 }, //27 September 2014, UT1-UTC=-0.3483662, Accumulated Leap Seconds=35
  { 2456928.50, 67.533132 }, //28 September 2014, UT1-UTC=-0.3491320, Accumulated Leap Seconds=35
  { 2456929.50, 67.533840 }, //29 September 2014, UT1-UTC=-0.3498404, Accumulated Leap Seconds=35
  { 2456930.50, 67.534560 }, //30 September 2014, UT1-UTC=-0.3505599, Accumulated Leap Seconds=35
  { 2456931.50, 67.535266 }, //01 October 2014, UT1-UTC=-0.3512658, Accumulated Leap Seconds=35
  { 2456932.50, 67.536054 }, //02 October 2014, UT1-UTC=-0.3520544, Accumulated Leap Seconds=35
  { 2456933.50, 67.536914 }, //03 October 2014, UT1-UTC=-0.3529137, Accumulated Leap Seconds=35
  { 2456934.50, 67.538004 }, //04 October 2014, UT1-UTC=-0.3540035, Accumulated Leap Seconds=35
  { 2456935.50, 67.539299 }, //05 October 2014, UT1-UTC=-0.3552989, Accumulated Leap Seconds=35
  { 2456936.50, 67.540741 }, //06 October 2014, UT1-UTC=-0.3567413, Accumulated Leap Seconds=35
  { 2456937.50, 67.542296 }, //07 October 2014, UT1-UTC=-0.3582963, Accumulated Leap Seconds=35
  { 2456938.50, 67.543948 }, //08 October 2014, UT1-UTC=-0.3599479, Accumulated Leap Seconds=35
  { 2456939.50, 67.545468 }, //09 October 2014, UT1-UTC=-0.3614677, Accumulated Leap Seconds=35
  { 2456940.50, 67.546825 }, //10 October 2014, UT1-UTC=-0.3628252, Accumulated Leap Seconds=35
  { 2456941.50, 67.548000 }, //11 October 2014, UT1-UTC=-0.3640000, Accumulated Leap Seconds=35
  { 2456942.50, 67.549093 }, //12 October 2014, UT1-UTC=-0.3650925, Accumulated Leap Seconds=35
  { 2456943.50, 67.550112 }, //13 October 2014, UT1-UTC=-0.3661121, Accumulated Leap Seconds=35
  { 2456944.50, 67.551029 }, //14 October 2014, UT1-UTC=-0.3670286, Accumulated Leap Seconds=35
  { 2456945.50, 67.551912 }, //15 October 2014, UT1-UTC=-0.3679116, Accumulated Leap Seconds=35
  { 2456946.50, 67.552825 }, //16 October 2014, UT1-UTC=-0.3688248, Accumulated Leap Seconds=35
  { 2456947.50, 67.553788 }, //17 October 2014, UT1-UTC=-0.3697884, Accumulated Leap Seconds=35
  { 2456948.50, 67.554758 }, //18 October 2014, UT1-UTC=-0.3707585, Accumulated Leap Seconds=35
  { 2456949.50, 67.555860 }, //19 October 2014, UT1-UTC=-0.3718601, Accumulated Leap Seconds=35
  { 2456950.50, 67.557023 }, //20 October 2014, UT1-UTC=-0.3730225, Accumulated Leap Seconds=35
  { 2456951.50, 67.558284 }, //21 October 2014, UT1-UTC=-0.3742835, Accumulated Leap Seconds=35
  { 2456952.50, 67.559622 }, //22 October 2014, UT1-UTC=-0.3756224, Accumulated Leap Seconds=35
  { 2456953.50, 67.561027 }, //23 October 2014, UT1-UTC=-0.3770269, Accumulated Leap Seconds=35
  { 2456954.50, 67.562297 }, //24 October 2014, UT1-UTC=-0.3782970, Accumulated Leap Seconds=35
  { 2456955.50, 67.563481 }, //25 October 2014, UT1-UTC=-0.3794815, Accumulated Leap Seconds=35
  { 2456956.50, 67.564589 }, //26 October 2014, UT1-UTC=-0.3805888, Accumulated Leap Seconds=35
  { 2456957.50, 67.565604 }, //27 October 2014, UT1-UTC=-0.3816040, Accumulated Leap Seconds=35
  { 2456958.50, 67.566589 }, //28 October 2014, UT1-UTC=-0.3825892, Accumulated Leap Seconds=35
  { 2456959.50, 67.567593 }, //29 October 2014, UT1-UTC=-0.3835930, Accumulated Leap Seconds=35
  { 2456960.50, 67.568672 }, //30 October 2014, UT1-UTC=-0.3846721, Accumulated Leap Seconds=35
  { 2456961.50, 67.569820 }, //31 October 2014, UT1-UTC=-0.3858203, Accumulated Leap Seconds=35
  { 2456962.50, 67.571103 }, //01 November 2014, UT1-UTC=-0.3871030, Accumulated Leap Seconds=35
  { 2456963.50, 67.572635 }, //02 November 2014, UT1-UTC=-0.3886351, Accumulated Leap Seconds=35
  { 2456964.50, 67.574328 }, //03 November 2014, UT1-UTC=-0.3903277, Accumulated Leap Seconds=35
  { 2456965.50, 67.576018 }, //04 November 2014, UT1-UTC=-0.3920178, Accumulated Leap Seconds=35
  { 2456966.50, 67.577656 }, //05 November 2014, UT1-UTC=-0.3936560, Accumulated Leap Seconds=35
  { 2456967.50, 67.579175 }, //06 November 2014, UT1-UTC=-0.3951751, Accumulated Leap Seconds=35
  { 2456968.50, 67.580518 }, //07 November 2014, UT1-UTC=-0.3965180, Accumulated Leap Seconds=35
  { 2456969.50, 67.581751 }, //08 November 2014, UT1-UTC=-0.3977507, Accumulated Leap Seconds=35
  { 2456970.50, 67.582760 }, //09 November 2014, UT1-UTC=-0.3987604, Accumulated Leap Seconds=35
  { 2456971.50, 67.583661 }, //10 November 2014, UT1-UTC=-0.3996613, Accumulated Leap Seconds=35
  { 2456972.50, 67.584526 }, //11 November 2014, UT1-UTC=-0.4005263, Accumulated Leap Seconds=35
  { 2456973.50, 67.585359 }, //12 November 2014, UT1-UTC=-0.4013595, Accumulated Leap Seconds=35
  { 2456974.50, 67.586274 }, //13 November 2014, UT1-UTC=-0.4022739, Accumulated Leap Seconds=35
  { 2456975.50, 67.587227 }, //14 November 2014, UT1-UTC=-0.4032271, Accumulated Leap Seconds=35
  { 2456976.50, 67.588234 }, //15 November 2014, UT1-UTC=-0.4042344, Accumulated Leap Seconds=35
  { 2456977.50, 67.589284 }, //16 November 2014, UT1-UTC=-0.4052845, Accumulated Leap Seconds=35
  { 2456978.50, 67.590443 }, //17 November 2014, UT1-UTC=-0.4064431, Accumulated Leap Seconds=35
  { 2456979.50, 67.591688 }, //18 November 2014, UT1-UTC=-0.4076876, Accumulated Leap Seconds=35
  { 2456980.50, 67.592951 }, //19 November 2014, UT1-UTC=-0.4089511, Accumulated Leap Seconds=35
  { 2456981.50, 67.594198 }, //20 November 2014, UT1-UTC=-0.4101976, Accumulated Leap Seconds=35
  { 2456982.50, 67.595347 }, //21 November 2014, UT1-UTC=-0.4113468, Accumulated Leap Seconds=35
  { 2456983.50, 67.596448 }, //22 November 2014, UT1-UTC=-0.4124480, Accumulated Leap Seconds=35
  { 2456984.50, 67.597509 }, //23 November 2014, UT1-UTC=-0.4135088, Accumulated Leap Seconds=35
  { 2456985.50, 67.598526 }, //24 November 2014, UT1-UTC=-0.4145265, Accumulated Leap Seconds=35
  { 2456986.50, 67.599529 }, //25 November 2014, UT1-UTC=-0.4155291, Accumulated Leap Seconds=35
  { 2456987.50, 67.600517 }, //26 November 2014, UT1-UTC=-0.4165166, Accumulated Leap Seconds=35
  { 2456988.50, 67.601551 }, //27 November 2014, UT1-UTC=-0.4175513, Accumulated Leap Seconds=35
  { 2456989.50, 67.602716 }, //28 November 2014, UT1-UTC=-0.4187161, Accumulated Leap Seconds=35
  { 2456990.50, 67.604060 }, //29 November 2014, UT1-UTC=-0.4200603, Accumulated Leap Seconds=35
  { 2456991.50, 67.605512 }, //30 November 2014, UT1-UTC=-0.4215121, Accumulated Leap Seconds=35
  { 2456992.50, 67.607025 }, //01 December 2014, UT1-UTC=-0.4230253, Accumulated Leap Seconds=35
  { 2456993.50, 67.608487 }, //02 December 2014, UT1-UTC=-0.4244870, Accumulated Leap Seconds=35
  { 2456994.50, 67.609879 }, //03 December 2014, UT1-UTC=-0.4258790, Accumulated Leap Seconds=35
  { 2456995.50, 67.611213 }, //04 December 2014, UT1-UTC=-0.4272133, Accumulated Leap Seconds=35
  { 2456996.50, 67.612434 }, //05 December 2014, UT1-UTC=-0.4284339, Accumulated Leap Seconds=35
  { 2456997.50, 67.613539 }, //06 December 2014, UT1-UTC=-0.4295389, Accumulated Leap Seconds=35
  { 2456998.50, 67.614507 }, //07 December 2014, UT1-UTC=-0.4305074, Accumulated Leap Seconds=35
  { 2456999.50, 67.615454 }, //08 December 2014, UT1-UTC=-0.4314540, Accumulated Leap Seconds=35
  { 2457000.50, 67.616441 }, //09 December 2014, UT1-UTC=-0.4324409, Accumulated Leap Seconds=35
  { 2457001.50, 67.617477 }, //10 December 2014, UT1-UTC=-0.4334773, Accumulated Leap Seconds=35
  { 2457002.50, 67.618536 }, //11 December 2014, UT1-UTC=-0.4345364, Accumulated Leap Seconds=35
  { 2457003.50, 67.619705 }, //12 December 2014, UT1-UTC=-0.4357052, Accumulated Leap Seconds=35
  { 2457004.50, 67.620938 }, //13 December 2014, UT1-UTC=-0.4369381, Accumulated Leap Seconds=35
  { 2457005.50, 67.622162 }, //14 December 2014, UT1-UTC=-0.4381615, Accumulated Leap Seconds=35
  { 2457006.50, 67.623386 }, //15 December 2014, UT1-UTC=-0.4393863, Accumulated Leap Seconds=35
  { 2457007.50, 67.624656 }, //16 December 2014, UT1-UTC=-0.4406563, Accumulated Leap Seconds=35
  { 2457008.50, 67.625969 }, //17 December 2014, UT1-UTC=-0.4419688, Accumulated Leap Seconds=35
  { 2457009.50, 67.627209 }, //18 December 2014, UT1-UTC=-0.4432094, Accumulated Leap Seconds=35
  { 2457010.50, 67.628393 }, //19 December 2014, UT1-UTC=-0.4443930, Accumulated Leap Seconds=35
  { 2457011.50, 67.629474 }, //20 December 2014, UT1-UTC=-0.4454738, Accumulated Leap Seconds=35
  { 2457012.50, 67.630456 }, //21 December 2014, UT1-UTC=-0.4464559, Accumulated Leap Seconds=35
  { 2457013.50, 67.631371 }, //22 December 2014, UT1-UTC=-0.4473714, Accumulated Leap Seconds=35
  { 2457014.50, 67.632345 }, //23 December 2014, UT1-UTC=-0.4483446, Accumulated Leap Seconds=35
  { 2457015.50, 67.633404 }, //24 December 2014, UT1-UTC=-0.4494041, Accumulated Leap Seconds=35
  { 2457016.50, 67.634579 }, //25 December 2014, UT1-UTC=-0.4505794, Accumulated Leap Seconds=35
  { 2457017.50, 67.635943 }, //26 December 2014, UT1-UTC=-0.4519435, Accumulated Leap Seconds=35
  { 2457018.50, 67.637371 }, //27 December 2014, UT1-UTC=-0.4533706, Accumulated Leap Seconds=35
  { 2457019.50, 67.638883 }, //28 December 2014, UT1-UTC=-0.4548828, Accumulated Leap Seconds=35
  { 2457020.50, 67.640386 }, //29 December 2014, UT1-UTC=-0.4563856, Accumulated Leap Seconds=35
  { 2457021.50, 67.641717 }, //30 December 2014, UT1-UTC=-0.4577165, Accumulated Leap Seconds=35
  { 2457022.50, 67.642892 }, //31 December 2014, UT1-UTC=-0.4588917, Accumulated Leap Seconds=35
  { 2457023.50, 67.643917 }, //01 January 2015, UT1-UTC=-0.4599167, Accumulated Leap Seconds=35
  { 2457024.50, 67.644830 }, //02 January 2015, UT1-UTC=-0.4608302, Accumulated Leap Seconds=35
  { 2457025.50, 67.645654 }, //03 January 2015, UT1-UTC=-0.4616545, Accumulated Leap Seconds=35
  { 2457026.50, 67.646376 }, //04 January 2015, UT1-UTC=-0.4623764, Accumulated Leap Seconds=35
  { 2457027.50, 67.646997 }, //05 January 2015, UT1-UTC=-0.4629975, Accumulated Leap Seconds=35
  { 2457028.50, 67.647658 }, //06 January 2015, UT1-UTC=-0.4636584, Accumulated Leap Seconds=35
  { 2457029.50, 67.648314 }, //07 January 2015, UT1-UTC=-0.4643139, Accumulated Leap Seconds=35
  { 2457030.50, 67.649089 }, //08 January 2015, UT1-UTC=-0.4650890, Accumulated Leap Seconds=35
  { 2457031.50, 67.650012 }, //09 January 2015, UT1-UTC=-0.4660117, Accumulated Leap Seconds=35
  { 2457032.50, 67.651061 }, //10 January 2015, UT1-UTC=-0.4670609, Accumulated Leap Seconds=35
  { 2457033.50, 67.652187 }, //11 January 2015, UT1-UTC=-0.4681869, Accumulated Leap Seconds=35
  { 2457034.50, 67.653291 }, //12 January 2015, UT1-UTC=-0.4692912, Accumulated Leap Seconds=35
  { 2457035.50, 67.654390 }, //13 January 2015, UT1-UTC=-0.4703905, Accumulated Leap Seconds=35
  { 2457036.50, 67.655531 }, //14 January 2015, UT1-UTC=-0.4715311, Accumulated Leap Seconds=35
  { 2457037.50, 67.656678 }, //15 January 2015, UT1-UTC=-0.4726777, Accumulated Leap Seconds=35
  { 2457038.50, 67.657783 }, //16 January 2015, UT1-UTC=-0.4737830, Accumulated Leap Seconds=35
  { 2457039.50, 67.658894 }, //17 January 2015, UT1-UTC=-0.4748937, Accumulated Leap Seconds=35
  { 2457040.50, 67.659944 }, //18 January 2015, UT1-UTC=-0.4759444, Accumulated Leap Seconds=35
  { 2457041.50, 67.661020 }, //19 January 2015, UT1-UTC=-0.4770205, Accumulated Leap Seconds=35
  { 2457042.50, 67.662158 }, //20 January 2015, UT1-UTC=-0.4781581, Accumulated Leap Seconds=35
  { 2457043.50, 67.663385 }, //21 January 2015, UT1-UTC=-0.4793853, Accumulated Leap Seconds=35
  { 2457044.50, 67.664814 }, //22 January 2015, UT1-UTC=-0.4808142, Accumulated Leap Seconds=35
  { 2457045.50, 67.666341 }, //23 January 2015, UT1-UTC=-0.4823410, Accumulated Leap Seconds=35
  { 2457046.50, 67.667859 }, //24 January 2015, UT1-UTC=-0.4838588, Accumulated Leap Seconds=35
  { 2457047.50, 67.669455 }, //25 January 2015, UT1-UTC=-0.4854546, Accumulated Leap Seconds=35
  { 2457048.50, 67.670784 }, //26 January 2015, UT1-UTC=-0.4867840, Accumulated Leap Seconds=35
  { 2457049.50, 67.672007 }, //27 January 2015, UT1-UTC=-0.4880067, Accumulated Leap Seconds=35
  { 2457050.50, 67.673063 }, //28 January 2015, UT1-UTC=-0.4890626, Accumulated Leap Seconds=35
  { 2457051.50, 67.673980 }, //29 January 2015, UT1-UTC=-0.4899801, Accumulated Leap Seconds=35
  { 2457052.50, 67.674823 }, //30 January 2015, UT1-UTC=-0.4908230, Accumulated Leap Seconds=35
  { 2457053.50, 67.675658 }, //31 January 2015, UT1-UTC=-0.4916583, Accumulated Leap Seconds=35
  { 2457054.50, 67.676527 }, //01 February 2015, UT1-UTC=-0.4925272, Accumulated Leap Seconds=35
  { 2457055.50, 67.677426 }, //02 February 2015, UT1-UTC=-0.4934260, Accumulated Leap Seconds=35
  { 2457056.50, 67.678431 }, //03 February 2015, UT1-UTC=-0.4944307, Accumulated Leap Seconds=35
  { 2457057.50, 67.679514 }, //04 February 2015, UT1-UTC=-0.4955137, Accumulated Leap Seconds=35
  { 2457058.50, 67.680649 }, //05 February 2015, UT1-UTC=-0.4966490, Accumulated Leap Seconds=35
  { 2457059.50, 67.681863 }, //06 February 2015, UT1-UTC=-0.4978631, Accumulated Leap Seconds=35
  { 2457060.50, 67.683144 }, //07 February 2015, UT1-UTC=-0.4991436, Accumulated Leap Seconds=35
  { 2457061.50, 67.684517 }, //08 February 2015, UT1-UTC=-0.5005175, Accumulated Leap Seconds=35
  { 2457062.50, 67.685881 }, //09 February 2015, UT1-UTC=-0.5018809, Accumulated Leap Seconds=35
  { 2457063.50, 67.687163 }, //10 February 2015, UT1-UTC=-0.5031630, Accumulated Leap Seconds=35
  { 2457064.50, 67.688359 }, //11 February 2015, UT1-UTC=-0.5043586, Accumulated Leap Seconds=35
  { 2457065.50, 67.689528 }, //12 February 2015, UT1-UTC=-0.5055278, Accumulated Leap Seconds=35
  { 2457066.50, 67.690702 }, //13 February 2015, UT1-UTC=-0.5067017, Accumulated Leap Seconds=35
  { 2457067.50, 67.691896 }, //14 February 2015, UT1-UTC=-0.5078957, Accumulated Leap Seconds=35
  { 2457068.50, 67.693055 }, //15 February 2015, UT1-UTC=-0.5090549, Accumulated Leap Seconds=35
  { 2457069.50, 67.694276 }, //16 February 2015, UT1-UTC=-0.5102758, Accumulated Leap Seconds=35
  { 2457070.50, 67.695588 }, //17 February 2015, UT1-UTC=-0.5115883, Accumulated Leap Seconds=35
  { 2457071.50, 67.697078 }, //18 February 2015, UT1-UTC=-0.5130784, Accumulated Leap Seconds=35
  { 2457072.50, 67.698756 }, //19 February 2015, UT1-UTC=-0.5147561, Accumulated Leap Seconds=35
  { 2457073.50, 67.700508 }, //20 February 2015, UT1-UTC=-0.5165075, Accumulated Leap Seconds=35
  { 2457074.50, 67.702267 }, //21 February 2015, UT1-UTC=-0.5182666, Accumulated Leap Seconds=35
  { 2457075.50, 67.703952 }, //22 February 2015, UT1-UTC=-0.5199520, Accumulated Leap Seconds=35
  { 2457076.50, 67.705510 }, //23 February 2015, UT1-UTC=-0.5215099, Accumulated Leap Seconds=35
  { 2457077.50, 67.706888 }, //24 February 2015, UT1-UTC=-0.5228885, Accumulated Leap Seconds=35
  { 2457078.50, 67.708022 }, //25 February 2015, UT1-UTC=-0.5240220, Accumulated Leap Seconds=35
  { 2457079.50, 67.709021 }, //26 February 2015, UT1-UTC=-0.5250211, Accumulated Leap Seconds=35
  { 2457080.50, 67.709928 }, //27 February 2015, UT1-UTC=-0.5259277, Accumulated Leap Seconds=35
  { 2457081.50, 67.710800 }, //28 February 2015, UT1-UTC=-0.5268002, Accumulated Leap Seconds=35
  { 2457082.50, 67.711669 }, //01 March 2015, UT1-UTC=-0.5276693, Accumulated Leap Seconds=35
  { 2457083.50, 67.712586 }, //02 March 2015, UT1-UTC=-0.5285859, Accumulated Leap Seconds=35
  { 2457084.50, 67.713570 }, //03 March 2015, UT1-UTC=-0.5295700, Accumulated Leap Seconds=35
  { 2457085.50, 67.714606 }, //04 March 2015, UT1-UTC=-0.5306064, Accumulated Leap Seconds=35
  { 2457086.50, 67.715740 }, //05 March 2015, UT1-UTC=-0.5317397, Accumulated Leap Seconds=35
  { 2457087.50, 67.716877 }, //06 March 2015, UT1-UTC=-0.5328766, Accumulated Leap Seconds=35
  { 2457088.50, 67.718098 }, //07 March 2015, UT1-UTC=-0.5340977, Accumulated Leap Seconds=35
  { 2457089.50, 67.719343 }, //08 March 2015, UT1-UTC=-0.5353429, Accumulated Leap Seconds=35
  { 2457090.50, 67.720656 }, //09 March 2015, UT1-UTC=-0.5366558, Accumulated Leap Seconds=35
  { 2457091.50, 67.722045 }, //10 March 2015, UT1-UTC=-0.5380450, Accumulated Leap Seconds=35
  { 2457092.50, 67.723440 }, //11 March 2015, UT1-UTC=-0.5394402, Accumulated Leap Seconds=35
  { 2457093.50, 67.724850 }, //12 March 2015, UT1-UTC=-0.5408497, Accumulated Leap Seconds=35
  { 2457094.50, 67.726268 }, //13 March 2015, UT1-UTC=-0.5422684, Accumulated Leap Seconds=35
  { 2457095.50, 67.727678 }, //14 March 2015, UT1-UTC=-0.5436778, Accumulated Leap Seconds=35
  { 2457096.50, 67.729207 }, //15 March 2015, UT1-UTC=-0.5452073, Accumulated Leap Seconds=35
  { 2457097.50, 67.730811 }, //16 March 2015, UT1-UTC=-0.5468113, Accumulated Leap Seconds=35
  { 2457098.50, 67.732545 }, //17 March 2015, UT1-UTC=-0.5485449, Accumulated Leap Seconds=35
  { 2457099.50, 67.734460 }, //18 March 2015, UT1-UTC=-0.5504600, Accumulated Leap Seconds=35
  { 2457100.50, 67.736582 }, //19 March 2015, UT1-UTC=-0.5525818, Accumulated Leap Seconds=35
  { 2457101.50, 67.738767 }, //20 March 2015, UT1-UTC=-0.5547671, Accumulated Leap Seconds=35
  { 2457102.50, 67.741051 }, //21 March 2015, UT1-UTC=-0.5570514, Accumulated Leap Seconds=35
  { 2457103.50, 67.743263 }, //22 March 2015, UT1-UTC=-0.5592635, Accumulated Leap Seconds=35
  { 2457104.50, 67.745303 }, //23 March 2015, UT1-UTC=-0.5613029, Accumulated Leap Seconds=35
  { 2457105.50, 67.747221 }, //24 March 2015, UT1-UTC=-0.5632208, Accumulated Leap Seconds=35
  { 2457106.50, 67.748983 }, //25 March 2015, UT1-UTC=-0.5649826, Accumulated Leap Seconds=35
  { 2457107.50, 67.750639 }, //26 March 2015, UT1-UTC=-0.5666387, Accumulated Leap Seconds=35
  { 2457108.50, 67.752112 }, //27 March 2015, UT1-UTC=-0.5681125, Accumulated Leap Seconds=35
  { 2457109.50, 67.753529 }, //28 March 2015, UT1-UTC=-0.5695294, Accumulated Leap Seconds=35
  { 2457110.50, 67.754924 }, //29 March 2015, UT1-UTC=-0.5709237, Accumulated Leap Seconds=35
  { 2457111.50, 67.756292 }, //30 March 2015, UT1-UTC=-0.5722921, Accumulated Leap Seconds=35
  { 2457112.50, 67.757665 }, //31 March 2015, UT1-UTC=-0.5736649, Accumulated Leap Seconds=35
  { 2457113.50, 67.759063 }, //01 April 2015, UT1-UTC=-0.5750634, Accumulated Leap Seconds=35
  { 2457114.50, 67.760417 }, //02 April 2015, UT1-UTC=-0.5764171, Accumulated Leap Seconds=35
  { 2457115.50, 67.761735 }, //03 April 2015, UT1-UTC=-0.5777352, Accumulated Leap Seconds=35
  { 2457116.50, 67.762993 }, //04 April 2015, UT1-UTC=-0.5789928, Accumulated Leap Seconds=35
  { 2457117.50, 67.764202 }, //05 April 2015, UT1-UTC=-0.5802022, Accumulated Leap Seconds=35
  { 2457118.50, 67.765426 }, //06 April 2015, UT1-UTC=-0.5814261, Accumulated Leap Seconds=35
  { 2457119.50, 67.766655 }, //07 April 2015, UT1-UTC=-0.5826549, Accumulated Leap Seconds=35
  { 2457120.50, 67.767871 }, //08 April 2015, UT1-UTC=-0.5838712, Accumulated Leap Seconds=35
  { 2457121.50, 67.769132 }, //09 April 2015, UT1-UTC=-0.5851323, Accumulated Leap Seconds=35
  { 2457122.50, 67.770404 }, //10 April 2015, UT1-UTC=-0.5864044, Accumulated Leap Seconds=35
  { 2457123.50, 67.771751 }, //11 April 2015, UT1-UTC=-0.5877510, Accumulated Leap Seconds=35
  { 2457124.50, 67.773263 }, //12 April 2015, UT1-UTC=-0.5892630, Accumulated Leap Seconds=35
  { 2457125.50, 67.774911 }, //13 April 2015, UT1-UTC=-0.5909108, Accumulated Leap Seconds=35
  { 2457126.50, 67.776675 }, //14 April 2015, UT1-UTC=-0.5926746, Accumulated Leap Seconds=35
  { 2457127.50, 67.778548 }, //15 April 2015, UT1-UTC=-0.5945485, Accumulated Leap Seconds=35
  { 2457128.50, 67.780540 }, //16 April 2015, UT1-UTC=-0.5965396, Accumulated Leap Seconds=35
  { 2457129.50, 67.782559 }, //17 April 2015, UT1-UTC=-0.5985591, Accumulated Leap Seconds=35
  { 2457130.50, 67.784431 }, //18 April 2015, UT1-UTC=-0.6004308, Accumulated Leap Seconds=35
  { 2457131.50, 67.786209 }, //19 April 2015, UT1-UTC=-0.6022092, Accumulated Leap Seconds=35
  { 2457132.50, 67.787721 }, //20 April 2015, UT1-UTC=-0.6037213, Accumulated Leap Seconds=35
  { 2457133.50, 67.789038 }, //21 April 2015, UT1-UTC=-0.6050379, Accumulated Leap Seconds=35
  { 2457134.50, 67.790146 }, //22 April 2015, UT1-UTC=-0.6061459, Accumulated Leap Seconds=35
  { 2457135.50, 67.791200 }, //23 April 2015, UT1-UTC=-0.6071996, Accumulated Leap Seconds=35
  { 2457136.50, 67.792214 }, //24 April 2015, UT1-UTC=-0.6082139, Accumulated Leap Seconds=35
  { 2457137.50, 67.793267 }, //25 April 2015, UT1-UTC=-0.6092667, Accumulated Leap Seconds=35
  { 2457138.50, 67.794445 }, //26 April 2015, UT1-UTC=-0.6104453, Accumulated Leap Seconds=35
  { 2457139.50, 67.795685 }, //27 April 2015, UT1-UTC=-0.6116849, Accumulated Leap Seconds=35
  { 2457140.50, 67.796990 }, //28 April 2015, UT1-UTC=-0.6129902, Accumulated Leap Seconds=35
  { 2457141.50, 67.798361 }, //29 April 2015, UT1-UTC=-0.6143607, Accumulated Leap Seconds=35
  { 2457142.50, 67.799741 }, //30 April 2015, UT1-UTC=-0.6157415, Accumulated Leap Seconds=35
  { 2457143.50, 67.801154 }, //01 May 2015, UT1-UTC=-0.6171542, Accumulated Leap Seconds=35
  { 2457144.50, 67.802511 }, //02 May 2015, UT1-UTC=-0.6185112, Accumulated Leap Seconds=35
  { 2457145.50, 67.803777 }, //03 May 2015, UT1-UTC=-0.6197775, Accumulated Leap Seconds=35
  { 2457146.50, 67.804970 }, //04 May 2015, UT1-UTC=-0.6209697, Accumulated Leap Seconds=35
  { 2457147.50, 67.806073 }, //05 May 2015, UT1-UTC=-0.6220729, Accumulated Leap Seconds=35
  { 2457148.50, 67.807092 }, //06 May 2015, UT1-UTC=-0.6230920, Accumulated Leap Seconds=35
  { 2457149.50, 67.808098 }, //07 May 2015, UT1-UTC=-0.6240983, Accumulated Leap Seconds=35
  { 2457150.50, 67.809169 }, //08 May 2015, UT1-UTC=-0.6251695, Accumulated Leap Seconds=35
  { 2457151.50, 67.810284 }, //09 May 2015, UT1-UTC=-0.6262838, Accumulated Leap Seconds=35
  { 2457152.50, 67.811541 }, //10 May 2015, UT1-UTC=-0.6275410, Accumulated Leap Seconds=35
  { 2457153.50, 67.812955 }, //11 May 2015, UT1-UTC=-0.6289545, Accumulated Leap Seconds=35
  { 2457154.50, 67.814580 }, //12 May 2015, UT1-UTC=-0.6305802, Accumulated Leap Seconds=35
  { 2457155.50, 67.816375 }, //13 May 2015, UT1-UTC=-0.6323748, Accumulated Leap Seconds=35
  { 2457156.50, 67.818264 }, //14 May 2015, UT1-UTC=-0.6342644, Accumulated Leap Seconds=35
  { 2457157.50, 67.820154 }, //15 May 2015, UT1-UTC=-0.6361535, Accumulated Leap Seconds=35
  { 2457158.50, 67.822000 }, //16 May 2015, UT1-UTC=-0.6379997, Accumulated Leap Seconds=35
  { 2457159.50, 67.823746 }, //17 May 2015, UT1-UTC=-0.6397461, Accumulated Leap Seconds=35
  { 2457160.50, 67.825318 }, //18 May 2015, UT1-UTC=-0.6413176, Accumulated Leap Seconds=35
  { 2457161.50, 67.826741 }, //19 May 2015, UT1-UTC=-0.6427407, Accumulated Leap Seconds=35
  { 2457162.50, 67.828051 }, //20 May 2015, UT1-UTC=-0.6440508, Accumulated Leap Seconds=35
  { 2457163.50, 67.829286 }, //21 May 2015, UT1-UTC=-0.6452856, Accumulated Leap Seconds=35
  { 2457164.50, 67.830409 }, //22 May 2015, UT1-UTC=-0.6464092, Accumulated Leap Seconds=35
  { 2457165.50, 67.831516 }, //23 May 2015, UT1-UTC=-0.6475158, Accumulated Leap Seconds=35
  { 2457166.50, 67.832537 }, //24 May 2015, UT1-UTC=-0.6485374, Accumulated Leap Seconds=35
  { 2457167.50, 67.833557 }, //25 May 2015, UT1-UTC=-0.6495572, Accumulated Leap Seconds=35
  { 2457168.50, 67.834568 }, //26 May 2015, UT1-UTC=-0.6505676, Accumulated Leap Seconds=35
  { 2457169.50, 67.835617 }, //27 May 2015, UT1-UTC=-0.6516169, Accumulated Leap Seconds=35
  { 2457170.50, 67.836609 }, //28 May 2015, UT1-UTC=-0.6526091, Accumulated Leap Seconds=35
  { 2457171.50, 67.837638 }, //29 May 2015, UT1-UTC=-0.6536380, Accumulated Leap Seconds=35
  { 2457172.50, 67.838568 }, //30 May 2015, UT1-UTC=-0.6545684, Accumulated Leap Seconds=35
  { 2457173.50, 67.839442 }, //31 May 2015, UT1-UTC=-0.6554421, Accumulated Leap Seconds=35
  { 2457174.50, 67.840213 }, //01 June 2015, UT1-UTC=-0.6562130, Accumulated Leap Seconds=35
  { 2457175.50, 67.840913 }, //02 June 2015, UT1-UTC=-0.6569133, Accumulated Leap Seconds=35
  { 2457176.50, 67.841549 }, //03 June 2015, UT1-UTC=-0.6575488, Accumulated Leap Seconds=35
  { 2457177.50, 67.842192 }, //04 June 2015, UT1-UTC=-0.6581924, Accumulated Leap Seconds=35
  { 2457178.50, 67.842824 }, //05 June 2015, UT1-UTC=-0.6588244, Accumulated Leap Seconds=35
  { 2457179.50, 67.843558 }, //06 June 2015, UT1-UTC=-0.6595579, Accumulated Leap Seconds=35
  { 2457180.50, 67.844412 }, //07 June 2015, UT1-UTC=-0.6604117, Accumulated Leap Seconds=35
  { 2457181.50, 67.845409 }, //08 June 2015, UT1-UTC=-0.6614093, Accumulated Leap Seconds=35
  { 2457182.50, 67.846492 }, //09 June 2015, UT1-UTC=-0.6624918, Accumulated Leap Seconds=35
  { 2457183.50, 67.847531 }, //10 June 2015, UT1-UTC=-0.6635311, Accumulated Leap Seconds=35
  { 2457184.50, 67.848547 }, //11 June 2015, UT1-UTC=-0.6645469, Accumulated Leap Seconds=35
  { 2457185.50, 67.849500 }, //12 June 2015, UT1-UTC=-0.6654997, Accumulated Leap Seconds=35
  { 2457186.50, 67.850325 }, //13 June 2015, UT1-UTC=-0.6663246, Accumulated Leap Seconds=35
  { 2457187.50, 67.850933 }, //14 June 2015, UT1-UTC=-0.6669333, Accumulated Leap Seconds=35
  { 2457188.50, 67.851353 }, //15 June 2015, UT1-UTC=-0.6673534, Accumulated Leap Seconds=35
  { 2457189.50, 67.851748 }, //16 June 2015, UT1-UTC=-0.6677485, Accumulated Leap Seconds=35
  { 2457190.50, 67.851968 }, //17 June 2015, UT1-UTC=-0.6679675, Accumulated Leap Seconds=35
  { 2457191.50, 67.852184 }, //18 June 2015, UT1-UTC=-0.6681839, Accumulated Leap Seconds=35
  { 2457192.50, 67.852427 }, //19 June 2015, UT1-UTC=-0.6684274, Accumulated Leap Seconds=35
  { 2457193.50, 67.852822 }, //20 June 2015, UT1-UTC=-0.6688221, Accumulated Leap Seconds=35
  { 2457194.50, 67.853341 }, //21 June 2015, UT1-UTC=-0.6693410, Accumulated Leap Seconds=35
  { 2457195.50, 67.853919 }, //22 June 2015, UT1-UTC=-0.6699194, Accumulated Leap Seconds=35
  { 2457196.50, 67.854565 }, //23 June 2015, UT1-UTC=-0.6705649, Accumulated Leap Seconds=35
  { 2457197.50, 67.855326 }, //24 June 2015, UT1-UTC=-0.6713261, Accumulated Leap Seconds=35
  { 2457198.50, 67.856136 }, //25 June 2015, UT1-UTC=-0.6721359, Accumulated Leap Seconds=35
  { 2457199.50, 67.856958 }, //26 June 2015, UT1-UTC=-0.6729582, Accumulated Leap Seconds=35
  { 2457200.50, 67.857809 }, //27 June 2015, UT1-UTC=-0.6738090, Accumulated Leap Seconds=35
  { 2457201.50, 67.858637 }, //28 June 2015, UT1-UTC=-0.6746368, Accumulated Leap Seconds=35
  { 2457202.50, 67.859384 }, //29 June 2015, UT1-UTC=-0.6753836, Accumulated Leap Seconds=35
  { 2457203.50, 67.860032 }, //30 June 2015, UT1-UTC=-0.6760316, Accumulated Leap Seconds=35
  { 2457204.50, 67.860632 }, //01 July 2015, UT1-UTC= 0.3233682, Accumulated Leap Seconds=36
  { 2457205.50, 67.861237 }, //02 July 2015, UT1-UTC= 0.3227635, Accumulated Leap Seconds=36
  { 2457206.50, 67.861898 }, //03 July 2015, UT1-UTC= 0.3221020, Accumulated Leap Seconds=36
  { 2457207.50, 67.862730 }, //04 July 2015, UT1-UTC= 0.3212702, Accumulated Leap Seconds=36
  { 2457208.50, 67.863746 }, //05 July 2015, UT1-UTC= 0.3202538, Accumulated Leap Seconds=36
  { 2457209.50, 67.864973 }, //06 July 2015, UT1-UTC= 0.3190270, Accumulated Leap Seconds=36
  { 2457210.50, 67.866231 }, //07 July 2015, UT1-UTC= 0.3177691, Accumulated Leap Seconds=36
  { 2457211.50, 67.867625 }, //08 July 2015, UT1-UTC= 0.3163753, Accumulated Leap Seconds=36
  { 2457212.50, 67.868908 }, //09 July 2015, UT1-UTC= 0.3150922, Accumulated Leap Seconds=36
  { 2457213.50, 67.870151 }, //10 July 2015, UT1-UTC= 0.3138490, Accumulated Leap Seconds=36
  { 2457214.50, 67.871274 }, //11 July 2015, UT1-UTC= 0.3127263, Accumulated Leap Seconds=36
  { 2457215.50, 67.872203 }, //12 July 2015, UT1-UTC= 0.3117972, Accumulated Leap Seconds=36
  { 2457216.50, 67.872921 }, //13 July 2015, UT1-UTC= 0.3110788, Accumulated Leap Seconds=36
  { 2457217.50, 67.873444 }, //14 July 2015, UT1-UTC= 0.3105558, Accumulated Leap Seconds=36
  { 2457218.50, 67.873889 }, //15 July 2015, UT1-UTC= 0.3101112, Accumulated Leap Seconds=36
  { 2457219.50, 67.874276 }, //16 July 2015, UT1-UTC= 0.3097239, Accumulated Leap Seconds=36
  { 2457220.50, 67.874724 }, //17 July 2015, UT1-UTC= 0.3092758, Accumulated Leap Seconds=36
  { 2457221.50, 67.875175 }, //18 July 2015, UT1-UTC= 0.3088245, Accumulated Leap Seconds=36
  { 2457222.50, 67.875646 }, //19 July 2015, UT1-UTC= 0.3083544, Accumulated Leap Seconds=36
  { 2457223.50, 67.876113 }, //20 July 2015, UT1-UTC= 0.3078874, Accumulated Leap Seconds=36
  { 2457224.50, 67.876659 }, //21 July 2015, UT1-UTC= 0.3073406, Accumulated Leap Seconds=36
  { 2457225.50, 67.877209 }, //22 July 2015, UT1-UTC= 0.3067906, Accumulated Leap Seconds=36
  { 2457226.50, 67.877750 }, //23 July 2015, UT1-UTC= 0.3062503, Accumulated Leap Seconds=36
  { 2457227.50, 67.878205 }, //24 July 2015, UT1-UTC= 0.3057953, Accumulated Leap Seconds=36
  { 2457228.50, 67.878704 }, //25 July 2015, UT1-UTC= 0.3052963, Accumulated Leap Seconds=36
  { 2457229.50, 67.879106 }, //26 July 2015, UT1-UTC= 0.3048942, Accumulated Leap Seconds=36
  { 2457230.50, 67.879508 }, //27 July 2015, UT1-UTC= 0.3044921, Accumulated Leap Seconds=36
  { 2457231.50, 67.879901 }, //28 July 2015, UT1-UTC= 0.3040989, Accumulated Leap Seconds=36
  { 2457232.50, 67.880256 }, //29 July 2015, UT1-UTC= 0.3037437, Accumulated Leap Seconds=36
  { 2457233.50, 67.880729 }, //30 July 2015, UT1-UTC= 0.3032715, Accumulated Leap Seconds=36
  { 2457234.50, 67.881343 }, //31 July 2015, UT1-UTC= 0.3026575, Accumulated Leap Seconds=36
  { 2457235.50, 67.882158 }, //01 August 2015, UT1-UTC= 0.3018424, Accumulated Leap Seconds=36
  { 2457236.50, 67.883154 }, //02 August 2015, UT1-UTC= 0.3008463, Accumulated Leap Seconds=36
  { 2457237.50, 67.884314 }, //03 August 2015, UT1-UTC= 0.2996859, Accumulated Leap Seconds=36
  { 2457238.50, 67.885549 }, //04 August 2015, UT1-UTC= 0.2984513, Accumulated Leap Seconds=36
  { 2457239.50, 67.886778 }, //05 August 2015, UT1-UTC= 0.2972219, Accumulated Leap Seconds=36
  { 2457240.50, 67.887987 }, //06 August 2015, UT1-UTC= 0.2960130, Accumulated Leap Seconds=36
  { 2457241.50, 67.888999 }, //07 August 2015, UT1-UTC= 0.2950006, Accumulated Leap Seconds=36
  { 2457242.50, 67.889879 }, //08 August 2015, UT1-UTC= 0.2941210, Accumulated Leap Seconds=36
  { 2457243.50, 67.890644 }, //09 August 2015, UT1-UTC= 0.2933557, Accumulated Leap Seconds=36
  { 2457244.50, 67.891280 }, //10 August 2015, UT1-UTC= 0.2927201, Accumulated Leap Seconds=36
  { 2457245.50, 67.891859 }, //11 August 2015, UT1-UTC= 0.2921410, Accumulated Leap Seconds=36
  { 2457246.50, 67.892463 }, //12 August 2015, UT1-UTC= 0.2915366, Accumulated Leap Seconds=36
  { 2457247.50, 67.893078 }, //13 August 2015, UT1-UTC= 0.2909220, Accumulated Leap Seconds=36
  { 2457248.50, 67.893812 }, //14 August 2015, UT1-UTC= 0.2901880, Accumulated Leap Seconds=36
  { 2457249.50, 67.894581 }, //15 August 2015, UT1-UTC= 0.2894194, Accumulated Leap Seconds=36
  { 2457250.50, 67.895396 }, //16 August 2015, UT1-UTC= 0.2886043, Accumulated Leap Seconds=36
  { 2457251.50, 67.896237 }, //17 August 2015, UT1-UTC= 0.2877626, Accumulated Leap Seconds=36
  { 2457252.50, 67.897073 }, //18 August 2015, UT1-UTC= 0.2869267, Accumulated Leap Seconds=36
  { 2457253.50, 67.897869 }, //19 August 2015, UT1-UTC= 0.2861307, Accumulated Leap Seconds=36
  { 2457254.50, 67.898692 }, //20 August 2015, UT1-UTC= 0.2853077, Accumulated Leap Seconds=36
  { 2457255.50, 67.899532 }, //21 August 2015, UT1-UTC= 0.2844675, Accumulated Leap Seconds=36
  { 2457256.50, 67.900361 }, //22 August 2015, UT1-UTC= 0.2836391, Accumulated Leap Seconds=36
  { 2457257.50, 67.901136 }, //23 August 2015, UT1-UTC= 0.2828636, Accumulated Leap Seconds=36
  { 2457258.50, 67.901886 }, //24 August 2015, UT1-UTC= 0.2821140, Accumulated Leap Seconds=36
  { 2457259.50, 67.902689 }, //25 August 2015, UT1-UTC= 0.2813110, Accumulated Leap Seconds=36
  { 2457260.50, 67.903572 }, //26 August 2015, UT1-UTC= 0.2804278, Accumulated Leap Seconds=36
  { 2457261.50, 67.904534 }, //27 August 2015, UT1-UTC= 0.2794655, Accumulated Leap Seconds=36
  { 2457262.50, 67.905652 }, //28 August 2015, UT1-UTC= 0.2783477, Accumulated Leap Seconds=36
  { 2457263.50, 67.906956 }, //29 August 2015, UT1-UTC= 0.2770440, Accumulated Leap Seconds=36
  { 2457264.50, 67.908463 }, //30 August 2015, UT1-UTC= 0.2755372, Accumulated Leap Seconds=36
  { 2457265.50, 67.910198 }, //31 August 2015, UT1-UTC= 0.2738018, Accumulated Leap Seconds=36
  { 2457266.50, 67.912010 }, //01 September 2015, UT1-UTC= 0.2719899, Accumulated Leap Seconds=36
  { 2457267.50, 67.913805 }, //02 September 2015, UT1-UTC= 0.2701950, Accumulated Leap Seconds=36
  { 2457268.50, 67.915502 }, //03 September 2015, UT1-UTC= 0.2684981, Accumulated Leap Seconds=36
  { 2457269.50, 67.916981 }, //04 September 2015, UT1-UTC= 0.2670190, Accumulated Leap Seconds=36
  { 2457270.50, 67.918323 }, //05 September 2015, UT1-UTC= 0.2656770, Accumulated Leap Seconds=36
  { 2457271.50, 67.919522 }, //06 September 2015, UT1-UTC= 0.2644780, Accumulated Leap Seconds=36
  { 2457272.50, 67.920638 }, //07 September 2015, UT1-UTC= 0.2633617, Accumulated Leap Seconds=36
  { 2457273.50, 67.921686 }, //08 September 2015, UT1-UTC= 0.2623142, Accumulated Leap Seconds=36
  { 2457274.50, 67.922743 }, //09 September 2015, UT1-UTC= 0.2612568, Accumulated Leap Seconds=36
  { 2457275.50, 67.923803 }, //10 September 2015, UT1-UTC= 0.2601972, Accumulated Leap Seconds=36
  { 2457276.50, 67.924848 }, //11 September 2015, UT1-UTC= 0.2591523, Accumulated Leap Seconds=36
  { 2457277.50, 67.926026 }, //12 September 2015, UT1-UTC= 0.2579742, Accumulated Leap Seconds=36
  { 2457278.50, 67.927323 }, //13 September 2015, UT1-UTC= 0.2566772, Accumulated Leap Seconds=36
  { 2457279.50, 67.928636 }, //14 September 2015, UT1-UTC= 0.2553638, Accumulated Leap Seconds=36
  { 2457280.50, 67.929948 }, //15 September 2015, UT1-UTC= 0.2540516, Accumulated Leap Seconds=36
  { 2457281.50, 67.931255 }, //16 September 2015, UT1-UTC= 0.2527454, Accumulated Leap Seconds=36
  { 2457282.50, 67.932495 }, //17 September 2015, UT1-UTC= 0.2515046, Accumulated Leap Seconds=36
  { 2457283.50, 67.933679 }, //18 September 2015, UT1-UTC= 0.2503208, Accumulated Leap Seconds=36
  { 2457284.50, 67.934866 }, //19 September 2015, UT1-UTC= 0.2491338, Accumulated Leap Seconds=36
  { 2457285.50, 67.936024 }, //20 September 2015, UT1-UTC= 0.2479762, Accumulated Leap Seconds=36
  { 2457286.50, 67.937250 }, //21 September 2015, UT1-UTC= 0.2467495, Accumulated Leap Seconds=36
  { 2457287.50, 67.938492 }, //22 September 2015, UT1-UTC= 0.2455077, Accumulated Leap Seconds=36
  { 2457288.50, 67.939824 }, //23 September 2015, UT1-UTC= 0.2441763, Accumulated Leap Seconds=36
  { 2457289.50, 67.941239 }, //24 September 2015, UT1-UTC= 0.2427613, Accumulated Leap Seconds=36
  { 2457290.50, 67.942820 }, //25 September 2015, UT1-UTC= 0.2411804, Accumulated Leap Seconds=36
  { 2457291.50, 67.944623 }, //26 September 2015, UT1-UTC= 0.2393773, Accumulated Leap Seconds=36
  { 2457292.50, 67.946652 }, //27 September 2015, UT1-UTC= 0.2373484, Accumulated Leap Seconds=36
  { 2457293.50, 67.948779 }, //28 September 2015, UT1-UTC= 0.2352207, Accumulated Leap Seconds=36
  { 2457294.50, 67.950875 }, //29 September 2015, UT1-UTC= 0.2331245, Accumulated Leap Seconds=36
  { 2457295.50, 67.952852 }, //30 September 2015, UT1-UTC= 0.2311479, Accumulated Leap Seconds=36
  { 2457296.50, 67.954646 }, //01 October 2015, UT1-UTC= 0.2293538, Accumulated Leap Seconds=36
  { 2457297.50, 67.956199 }, //02 October 2015, UT1-UTC= 0.2278013, Accumulated Leap Seconds=36
  { 2457298.50, 67.957538 }, //03 October 2015, UT1-UTC= 0.2264625, Accumulated Leap Seconds=36
  { 2457299.50, 67.958771 }, //04 October 2015, UT1-UTC= 0.2252291, Accumulated Leap Seconds=36
  { 2457300.50, 67.959974 }, //05 October 2015, UT1-UTC= 0.2240258, Accumulated Leap Seconds=36
  { 2457301.50, 67.961252 }, //06 October 2015, UT1-UTC= 0.2227476, Accumulated Leap Seconds=36
  { 2457302.50, 67.962587 }, //07 October 2015, UT1-UTC= 0.2214125, Accumulated Leap Seconds=36
  { 2457303.50, 67.963977 }, //08 October 2015, UT1-UTC= 0.2200226, Accumulated Leap Seconds=36
  { 2457304.50, 67.965397 }, //09 October 2015, UT1-UTC= 0.2186028, Accumulated Leap Seconds=36
  { 2457305.50, 67.966943 }, //10 October 2015, UT1-UTC= 0.2170570, Accumulated Leap Seconds=36
  { 2457306.50, 67.968506 }, //11 October 2015, UT1-UTC= 0.2154943, Accumulated Leap Seconds=36
  { 2457307.50, 67.970134 }, //12 October 2015, UT1-UTC= 0.2138664, Accumulated Leap Seconds=36
  { 2457308.50, 67.971775 }, //13 October 2015, UT1-UTC= 0.2122251, Accumulated Leap Seconds=36
  { 2457309.50, 67.973448 }, //14 October 2015, UT1-UTC= 0.2105521, Accumulated Leap Seconds=36
  { 2457310.50, 67.974998 }, //15 October 2015, UT1-UTC= 0.2090018, Accumulated Leap Seconds=36
  { 2457311.50, 67.976399 }, //16 October 2015, UT1-UTC= 0.2076005, Accumulated Leap Seconds=36
  { 2457312.50, 67.977804 }, //17 October 2015, UT1-UTC= 0.2061956, Accumulated Leap Seconds=36
  { 2457313.50, 67.979277 }, //18 October 2015, UT1-UTC= 0.2047225, Accumulated Leap Seconds=36
  { 2457314.50, 67.980820 }, //19 October 2015, UT1-UTC= 0.2031797, Accumulated Leap Seconds=36
  { 2457315.50, 67.982376 }, //20 October 2015, UT1-UTC= 0.2016244, Accumulated Leap Seconds=36
  { 2457316.50, 67.983940 }, //21 October 2015, UT1-UTC= 0.2000598, Accumulated Leap Seconds=36
  { 2457317.50, 67.985608 }, //22 October 2015, UT1-UTC= 0.1983916, Accumulated Leap Seconds=36
  { 2457318.50, 67.987465 }, //23 October 2015, UT1-UTC= 0.1965353, Accumulated Leap Seconds=36
  { 2457319.50, 67.989535 }, //24 October 2015, UT1-UTC= 0.1944652, Accumulated Leap Seconds=36
  { 2457320.50, 67.991755 }, //25 October 2015, UT1-UTC= 0.1922446, Accumulated Leap Seconds=36
  { 2457321.50, 67.994053 }, //26 October 2015, UT1-UTC= 0.1899469, Accumulated Leap Seconds=36
  { 2457322.50, 67.996344 }, //27 October 2015, UT1-UTC= 0.1876556, Accumulated Leap Seconds=36
  { 2457323.50, 67.998528 }, //28 October 2015, UT1-UTC= 0.1854718, Accumulated Leap Seconds=36
  { 2457324.50, 68.000562 }, //29 October 2015, UT1-UTC= 0.1834377, Accumulated Leap Seconds=36
  { 2457325.50, 68.002338 }, //30 October 2015, UT1-UTC= 0.1816623, Accumulated Leap Seconds=36
  { 2457326.50, 68.003976 }, //31 October 2015, UT1-UTC= 0.1800239, Accumulated Leap Seconds=36
  { 2457327.50, 68.005484 }, //01 November 2015, UT1-UTC= 0.1785161, Accumulated Leap Seconds=36
  { 2457328.50, 68.006968 }, //02 November 2015, UT1-UTC= 0.1770322, Accumulated Leap Seconds=36
  { 2457329.50, 68.008405 }, //03 November 2015, UT1-UTC= 0.1755949, Accumulated Leap Seconds=36
  { 2457330.50, 68.009818 }, //04 November 2015, UT1-UTC= 0.1741824, Accumulated Leap Seconds=36
  { 2457331.50, 68.011250 }, //05 November 2015, UT1-UTC= 0.1727503, Accumulated Leap Seconds=36
  { 2457332.50, 68.012748 }, //06 November 2015, UT1-UTC= 0.1712525, Accumulated Leap Seconds=36
  { 2457333.50, 68.014348 }, //07 November 2015, UT1-UTC= 0.1696520, Accumulated Leap Seconds=36
  { 2457334.50, 68.015963 }, //08 November 2015, UT1-UTC= 0.1680366, Accumulated Leap Seconds=36
  { 2457335.50, 68.017570 }, //09 November 2015, UT1-UTC= 0.1664302, Accumulated Leap Seconds=36
  { 2457336.50, 68.019126 }, //10 November 2015, UT1-UTC= 0.1648740, Accumulated Leap Seconds=36
  { 2457337.50, 68.020644 }, //11 November 2015, UT1-UTC= 0.1633562, Accumulated Leap Seconds=36
  { 2457338.50, 68.022110 }, //12 November 2015, UT1-UTC= 0.1618895, Accumulated Leap Seconds=36
  { 2457339.50, 68.023525 }, //13 November 2015, UT1-UTC= 0.1604752, Accumulated Leap Seconds=36
  { 2457340.50, 68.024882 }, //14 November 2015, UT1-UTC= 0.1591181, Accumulated Leap Seconds=36
  { 2457341.50, 68.026281 }, //15 November 2015, UT1-UTC= 0.1577185, Accumulated Leap Seconds=36
  { 2457342.50, 68.027648 }, //16 November 2015, UT1-UTC= 0.1563516, Accumulated Leap Seconds=36
  { 2457343.50, 68.029073 }, //17 November 2015, UT1-UTC= 0.1549268, Accumulated Leap Seconds=36
  { 2457344.50, 68.030557 }, //18 November 2015, UT1-UTC= 0.1534431, Accumulated Leap Seconds=36
  { 2457345.50, 68.032117 }, //19 November 2015, UT1-UTC= 0.1518834, Accumulated Leap Seconds=36
  { 2457346.50, 68.033750 }, //20 November 2015, UT1-UTC= 0.1502498, Accumulated Leap Seconds=36
  { 2457347.50, 68.035483 }, //21 November 2015, UT1-UTC= 0.1485169, Accumulated Leap Seconds=36
  { 2457348.50, 68.037313 }, //22 November 2015, UT1-UTC= 0.1466873, Accumulated Leap Seconds=36
  { 2457349.50, 68.039109 }, //23 November 2015, UT1-UTC= 0.1448906, Accumulated Leap Seconds=36
  { 2457350.50, 68.040925 }, //24 November 2015, UT1-UTC= 0.1430747, Accumulated Leap Seconds=36
  { 2457351.50, 68.042616 }, //25 November 2015, UT1-UTC= 0.1413842, Accumulated Leap Seconds=36
  { 2457352.50, 68.044216 }, //26 November 2015, UT1-UTC= 0.1397837, Accumulated Leap Seconds=36
  { 2457353.50, 68.045653 }, //27 November 2015, UT1-UTC= 0.1383474, Accumulated Leap Seconds=36
  { 2457354.50, 68.047014 }, //28 November 2015, UT1-UTC= 0.1369856, Accumulated Leap Seconds=36
  { 2457355.50, 68.048401 }, //29 November 2015, UT1-UTC= 0.1355993, Accumulated Leap Seconds=36
  { 2457356.50, 68.049871 }, //30 November 2015, UT1-UTC= 0.1341290, Accumulated Leap Seconds=36
  { 2457357.50, 68.051412 }, //01 December 2015, UT1-UTC= 0.1325880, Accumulated Leap Seconds=36
  { 2457358.50, 68.053004 }, //02 December 2015, UT1-UTC= 0.1309957, Accumulated Leap Seconds=36
  { 2457359.50, 68.054690 }, //03 December 2015, UT1-UTC= 0.1293098, Accumulated Leap Seconds=36
  { 2457360.50, 68.056378 }, //04 December 2015, UT1-UTC= 0.1276216, Accumulated Leap Seconds=36
  { 2457361.50, 68.058076 }, //05 December 2015, UT1-UTC= 0.1259237, Accumulated Leap Seconds=36
  { 2457362.50, 68.059724 }, //06 December 2015, UT1-UTC= 0.1242763, Accumulated Leap Seconds=36
  { 2457363.50, 68.061403 }, //07 December 2015, UT1-UTC= 0.1225967, Accumulated Leap Seconds=36
  { 2457364.50, 68.063055 }, //08 December 2015, UT1-UTC= 0.1209451, Accumulated Leap Seconds=36
  { 2457365.50, 68.064685 }, //09 December 2015, UT1-UTC= 0.1193154, Accumulated Leap Seconds=36
  { 2457366.50, 68.066303 }, //10 December 2015, UT1-UTC= 0.1176968, Accumulated Leap Seconds=36
  { 2457367.50, 68.067796 }, //11 December 2015, UT1-UTC= 0.1162041, Accumulated Leap Seconds=36
  { 2457368.50, 68.069219 }, //12 December 2015, UT1-UTC= 0.1147810, Accumulated Leap Seconds=36
  { 2457369.50, 68.070639 }, //13 December 2015, UT1-UTC= 0.1133612, Accumulated Leap Seconds=36
  { 2457370.50, 68.072019 }, //14 December 2015, UT1-UTC= 0.1119809, Accumulated Leap Seconds=36
  { 2457371.50, 68.073489 }, //15 December 2015, UT1-UTC= 0.1105114, Accumulated Leap Seconds=36
  { 2457372.50, 68.075079 }, //16 December 2015, UT1-UTC= 0.1089211, Accumulated Leap Seconds=36
  { 2457373.50, 68.076782 }, //17 December 2015, UT1-UTC= 0.1072180, Accumulated Leap Seconds=36
  { 2457374.50, 68.078602 }, //18 December 2015, UT1-UTC= 0.1053984, Accumulated Leap Seconds=36
  { 2457375.50, 68.080495 }, //19 December 2015, UT1-UTC= 0.1035047, Accumulated Leap Seconds=36
  { 2457376.50, 68.082497 }, //20 December 2015, UT1-UTC= 0.1015032, Accumulated Leap Seconds=36
  { 2457377.50, 68.084479 }, //21 December 2015, UT1-UTC= 0.0995206, Accumulated Leap Seconds=36
  { 2457378.50, 68.086364 }, //22 December 2015, UT1-UTC= 0.0976362, Accumulated Leap Seconds=36
  { 2457379.50, 68.088138 }, //23 December 2015, UT1-UTC= 0.0958616, Accumulated Leap Seconds=36
  { 2457380.50, 68.089696 }, //24 December 2015, UT1-UTC= 0.0943042, Accumulated Leap Seconds=36
  { 2457381.50, 68.091096 }, //25 December 2015, UT1-UTC= 0.0929038, Accumulated Leap Seconds=36
  { 2457382.50, 68.092555 }, //26 December 2015, UT1-UTC= 0.0914454, Accumulated Leap Seconds=36
  { 2457383.50, 68.093921 }, //27 December 2015, UT1-UTC= 0.0900786, Accumulated Leap Seconds=36
  { 2457384.50, 68.095406 }, //28 December 2015, UT1-UTC= 0.0885937, Accumulated Leap Seconds=36
  { 2457385.50, 68.097022 }, //29 December 2015, UT1-UTC= 0.0869781, Accumulated Leap Seconds=36
  { 2457386.50, 68.098729 }, //30 December 2015, UT1-UTC= 0.0852709, Accumulated Leap Seconds=36
  { 2457387.50, 68.100524 }, //31 December 2015, UT1-UTC= 0.0834758, Accumulated Leap Seconds=36
  { 2457388.50, 68.102420 }, //01 January 2016, UT1-UTC= 0.0815795, Accumulated Leap Seconds=36
  { 2457389.50, 68.104363 }, //02 January 2016, UT1-UTC= 0.0796373, Accumulated Leap Seconds=36
  { 2457390.50, 68.106285 }, //03 January 2016, UT1-UTC= 0.0777154, Accumulated Leap Seconds=36
  { 2457391.50, 68.108226 }, //04 January 2016, UT1-UTC= 0.0757743, Accumulated Leap Seconds=36
  { 2457392.50, 68.110154 }, //05 January 2016, UT1-UTC= 0.0738461, Accumulated Leap Seconds=36
  { 2457393.50, 68.112037 }, //06 January 2016, UT1-UTC= 0.0719625, Accumulated Leap Seconds=36
  { 2457394.50, 68.113879 }, //07 January 2016, UT1-UTC= 0.0701214, Accumulated Leap Seconds=36
  { 2457395.50, 68.115684 }, //08 January 2016, UT1-UTC= 0.0683163, Accumulated Leap Seconds=36
  { 2457396.50, 68.117416 }, //09 January 2016, UT1-UTC= 0.0665840, Accumulated Leap Seconds=36
  { 2457397.50, 68.119177 }, //10 January 2016, UT1-UTC= 0.0648234, Accumulated Leap Seconds=36
  { 2457398.50, 68.120974 }, //11 January 2016, UT1-UTC= 0.0630258, Accumulated Leap Seconds=36
  { 2457399.50, 68.122936 }, //12 January 2016, UT1-UTC= 0.0610640, Accumulated Leap Seconds=36
  { 2457400.50, 68.124992 }, //13 January 2016, UT1-UTC= 0.0590080, Accumulated Leap Seconds=36
  { 2457401.50, 68.127121 }, //14 January 2016, UT1-UTC= 0.0568793, Accumulated Leap Seconds=36
  { 2457402.50, 68.129367 }, //15 January 2016, UT1-UTC= 0.0546329, Accumulated Leap Seconds=36
  { 2457403.50, 68.131517 }, //16 January 2016, UT1-UTC= 0.0524834, Accumulated Leap Seconds=36
  { 2457404.50, 68.133647 }, //17 January 2016, UT1-UTC= 0.0503529, Accumulated Leap Seconds=36
  { 2457405.50, 68.135583 }, //18 January 2016, UT1-UTC= 0.0484173, Accumulated Leap Seconds=36
  { 2457406.50, 68.137489 }, //19 January 2016, UT1-UTC= 0.0465106, Accumulated Leap Seconds=36
  { 2457407.50, 68.139260 }, //20 January 2016, UT1-UTC= 0.0447403, Accumulated Leap Seconds=36
  { 2457408.50, 68.140893 }, //21 January 2016, UT1-UTC= 0.0431073, Accumulated Leap Seconds=36
  { 2457409.50, 68.142460 }, //22 January 2016, UT1-UTC= 0.0415405, Accumulated Leap Seconds=36
  { 2457410.50, 68.143983 }, //23 January 2016, UT1-UTC= 0.0400168, Accumulated Leap Seconds=36
  { 2457411.50, 68.145577 }, //24 January 2016, UT1-UTC= 0.0384231, Accumulated Leap Seconds=36
  { 2457412.50, 68.147185 }, //25 January 2016, UT1-UTC= 0.0368154, Accumulated Leap Seconds=36
  { 2457413.50, 68.148817 }, //26 January 2016, UT1-UTC= 0.0351828, Accumulated Leap Seconds=36
  { 2457414.50, 68.150425 }, //27 January 2016, UT1-UTC= 0.0335746, Accumulated Leap Seconds=36
  { 2457415.50, 68.152010 }, //28 January 2016, UT1-UTC= 0.0319901, Accumulated Leap Seconds=36
  { 2457416.50, 68.153547 }, //29 January 2016, UT1-UTC= 0.0304532, Accumulated Leap Seconds=36
  { 2457417.50, 68.155063 }, //30 January 2016, UT1-UTC= 0.0289370, Accumulated Leap Seconds=36
  { 2457418.50, 68.156408 }, //31 January 2016, UT1-UTC= 0.0275923, Accumulated Leap Seconds=36
  { 2457419.50, 68.157713 }, //01 February 2016, UT1-UTC= 0.0262873, Accumulated Leap Seconds=36
  { 2457420.50, 68.159009 }, //02 February 2016, UT1-UTC= 0.0249913, Accumulated Leap Seconds=36
  { 2457421.50, 68.160201 }, //03 February 2016, UT1-UTC= 0.0237994, Accumulated Leap Seconds=36
  { 2457422.50, 68.161383 }, //04 February 2016, UT1-UTC= 0.0226169, Accumulated Leap Seconds=36
  { 2457423.50, 68.162573 }, //05 February 2016, UT1-UTC= 0.0214271, Accumulated Leap Seconds=36
  { 2457424.50, 68.163811 }, //06 February 2016, UT1-UTC= 0.0201890, Accumulated Leap Seconds=36
  { 2457425.50, 68.165349 }, //07 February 2016, UT1-UTC= 0.0186505, Accumulated Leap Seconds=36
  { 2457426.50, 68.166893 }, //08 February 2016, UT1-UTC= 0.0171067, Accumulated Leap Seconds=36
  { 2457427.50, 68.168691 }, //09 February 2016, UT1-UTC= 0.0153086, Accumulated Leap Seconds=36
  { 2457428.50, 68.170687 }, //10 February 2016, UT1-UTC= 0.0133135, Accumulated Leap Seconds=36
  { 2457429.50, 68.172788 }, //11 February 2016, UT1-UTC= 0.0112116, Accumulated Leap Seconds=36
  { 2457430.50, 68.174834 }, //12 February 2016, UT1-UTC= 0.0091659, Accumulated Leap Seconds=36
  { 2457431.50, 68.176871 }, //13 February 2016, UT1-UTC= 0.0071291, Accumulated Leap Seconds=36
  { 2457432.50, 68.178759 }, //14 February 2016, UT1-UTC= 0.0052412, Accumulated Leap Seconds=36
  { 2457433.50, 68.180496 }, //15 February 2016, UT1-UTC= 0.0035036, Accumulated Leap Seconds=36
  { 2457434.50, 68.182088 }, //16 February 2016, UT1-UTC= 0.0019117, Accumulated Leap Seconds=36
  { 2457435.50, 68.183531 }, //17 February 2016, UT1-UTC= 0.0004688, Accumulated Leap Seconds=36
  { 2457436.50, 68.184856 }, //18 February 2016, UT1-UTC=-0.0008558, Accumulated Leap Seconds=36
  { 2457437.50, 68.186144 }, //19 February 2016, UT1-UTC=-0.0021443, Accumulated Leap Seconds=36
  { 2457438.50, 68.187467 }, //20 February 2016, UT1-UTC=-0.0034666, Accumulated Leap Seconds=36
  { 2457439.50, 68.188887 }, //21 February 2016, UT1-UTC=-0.0048874, Accumulated Leap Seconds=36
  { 2457440.50, 68.190431 }, //22 February 2016, UT1-UTC=-0.0064308, Accumulated Leap Seconds=36
  { 2457441.50, 68.192075 }, //23 February 2016, UT1-UTC=-0.0080748, Accumulated Leap Seconds=36
  { 2457442.50, 68.193792 }, //24 February 2016, UT1-UTC=-0.0097922, Accumulated Leap Seconds=36
  { 2457443.50, 68.195576 }, //25 February 2016, UT1-UTC=-0.0115757, Accumulated Leap Seconds=36
  { 2457444.50, 68.197419 }, //26 February 2016, UT1-UTC=-0.0134193, Accumulated Leap Seconds=36
  { 2457445.50, 68.199216 }, //27 February 2016, UT1-UTC=-0.0152164, Accumulated Leap Seconds=36
  { 2457446.50, 68.200966 }, //28 February 2016, UT1-UTC=-0.0169658, Accumulated Leap Seconds=36
  { 2457447.50, 68.202691 }, //29 February 2016, UT1-UTC=-0.0186907, Accumulated Leap Seconds=36
  { 2457448.50, 68.204365 }, //01 March 2016, UT1-UTC=-0.0203653, Accumulated Leap Seconds=36
  { 2457449.50, 68.206012 }, //02 March 2016, UT1-UTC=-0.0220122, Accumulated Leap Seconds=36
  { 2457450.50, 68.207702 }, //03 March 2016, UT1-UTC=-0.0237020, Accumulated Leap Seconds=36
  { 2457451.50, 68.209442 }, //04 March 2016, UT1-UTC=-0.0254416, Accumulated Leap Seconds=36
  { 2457452.50, 68.211236 }, //05 March 2016, UT1-UTC=-0.0272359, Accumulated Leap Seconds=36
  { 2457453.50, 68.213184 }, //06 March 2016, UT1-UTC=-0.0291843, Accumulated Leap Seconds=36
  { 2457454.50, 68.215309 }, //07 March 2016, UT1-UTC=-0.0313091, Accumulated Leap Seconds=36
  { 2457455.50, 68.217560 }, //08 March 2016, UT1-UTC=-0.0335602, Accumulated Leap Seconds=36
  { 2457456.50, 68.219917 }, //09 March 2016, UT1-UTC=-0.0359172, Accumulated Leap Seconds=36
  { 2457457.50, 68.222365 }, //10 March 2016, UT1-UTC=-0.0383652, Accumulated Leap Seconds=36
  { 2457458.50, 68.224862 }, //11 March 2016, UT1-UTC=-0.0408624, Accumulated Leap Seconds=36
  { 2457459.50, 68.227282 }, //12 March 2016, UT1-UTC=-0.0432816, Accumulated Leap Seconds=36
  { 2457460.50, 68.229580 }, //13 March 2016, UT1-UTC=-0.0455797, Accumulated Leap Seconds=36
  { 2457461.50, 68.231747 }, //14 March 2016, UT1-UTC=-0.0477466, Accumulated Leap Seconds=36
  { 2457462.50, 68.233769 }, //15 March 2016, UT1-UTC=-0.0497693, Accumulated Leap Seconds=36
  { 2457463.50, 68.235648 }, //16 March 2016, UT1-UTC=-0.0516483, Accumulated Leap Seconds=36
  { 2457464.50, 68.237466 }, //17 March 2016, UT1-UTC=-0.0534664, Accumulated Leap Seconds=36
  { 2457465.50, 68.239284 }, //18 March 2016, UT1-UTC=-0.0552843, Accumulated Leap Seconds=36
  { 2457466.50, 68.241143 }, //19 March 2016, UT1-UTC=-0.0571428, Accumulated Leap Seconds=36
  { 2457467.50, 68.243084 }, //20 March 2016, UT1-UTC=-0.0590845, Accumulated Leap Seconds=36
  { 2457468.50, 68.245099 }, //21 March 2016, UT1-UTC=-0.0610988, Accumulated Leap Seconds=36
  { 2457469.50, 68.247186 }, //22 March 2016, UT1-UTC=-0.0631860, Accumulated Leap Seconds=36
  { 2457470.50, 68.249293 }, //23 March 2016, UT1-UTC=-0.0652933, Accumulated Leap Seconds=36
  { 2457471.50, 68.251340 }, //24 March 2016, UT1-UTC=-0.0673398, Accumulated Leap Seconds=36
  { 2457472.50, 68.253354 }, //25 March 2016, UT1-UTC=-0.0693540, Accumulated Leap Seconds=36
  { 2457473.50, 68.255390 }, //26 March 2016, UT1-UTC=-0.0713899, Accumulated Leap Seconds=36
  { 2457474.50, 68.257399 }, //27 March 2016, UT1-UTC=-0.0733993, Accumulated Leap Seconds=36
  { 2457475.50, 68.259336 }, //28 March 2016, UT1-UTC=-0.0753359, Accumulated Leap Seconds=36
  { 2457476.50, 68.261184 }, //29 March 2016, UT1-UTC=-0.0771843, Accumulated Leap Seconds=36
  { 2457477.50, 68.262961 }, //30 March 2016, UT1-UTC=-0.0789609, Accumulated Leap Seconds=36
  { 2457478.50, 68.264700 }, //31 March 2016, UT1-UTC=-0.0807003, Accumulated Leap Seconds=36
  { 2457479.50, 68.266451 }, //01 April 2016, UT1-UTC=-0.0824507, Accumulated Leap Seconds=36
  { 2457480.50, 68.268246 }, //02 April 2016, UT1-UTC=-0.0842461, Accumulated Leap Seconds=36
  { 2457481.50, 68.270150 }, //03 April 2016, UT1-UTC=-0.0861498, Accumulated Leap Seconds=36
  { 2457482.50, 68.272153 }, //04 April 2016, UT1-UTC=-0.0881531, Accumulated Leap Seconds=36
  { 2457483.50, 68.274299 }, //05 April 2016, UT1-UTC=-0.0902987, Accumulated Leap Seconds=36
  { 2457484.50, 68.276577 }, //06 April 2016, UT1-UTC=-0.0925765, Accumulated Leap Seconds=36
  { 2457485.50, 68.278934 }, //07 April 2016, UT1-UTC=-0.0949338, Accumulated Leap Seconds=36
  { 2457486.50, 68.281274 }, //08 April 2016, UT1-UTC=-0.0972744, Accumulated Leap Seconds=36
  { 2457487.50, 68.283575 }, //09 April 2016, UT1-UTC=-0.0995752, Accumulated Leap Seconds=36
  { 2457488.50, 68.285722 }, //10 April 2016, UT1-UTC=-0.1017219, Accumulated Leap Seconds=36
  { 2457489.50, 68.287665 }, //11 April 2016, UT1-UTC=-0.1036651, Accumulated Leap Seconds=36
  { 2457490.50, 68.289474 }, //12 April 2016, UT1-UTC=-0.1054736, Accumulated Leap Seconds=36
  { 2457491.50, 68.291168 }, //13 April 2016, UT1-UTC=-0.1071675, Accumulated Leap Seconds=36
  { 2457492.50, 68.292754 }, //14 April 2016, UT1-UTC=-0.1087543, Accumulated Leap Seconds=36
  { 2457493.50, 68.294296 }, //15 April 2016, UT1-UTC=-0.1102959, Accumulated Leap Seconds=36
  { 2457494.50, 68.295895 }, //16 April 2016, UT1-UTC=-0.1118952, Accumulated Leap Seconds=36
  { 2457495.50, 68.297549 }, //17 April 2016, UT1-UTC=-0.1135492, Accumulated Leap Seconds=36
  { 2457496.50, 68.299216 }, //18 April 2016, UT1-UTC=-0.1152155, Accumulated Leap Seconds=36
  { 2457497.50, 68.300908 }, //19 April 2016, UT1-UTC=-0.1169076, Accumulated Leap Seconds=36
  { 2457498.50, 68.302631 }, //20 April 2016, UT1-UTC=-0.1186310, Accumulated Leap Seconds=36
  { 2457499.50, 68.304326 }, //21 April 2016, UT1-UTC=-0.1203262, Accumulated Leap Seconds=36
  { 2457500.50, 68.305971 }, //22 April 2016, UT1-UTC=-0.1219708, Accumulated Leap Seconds=36
  { 2457501.50, 68.307581 }, //23 April 2016, UT1-UTC=-0.1235814, Accumulated Leap Seconds=36
  { 2457502.50, 68.309116 }, //24 April 2016, UT1-UTC=-0.1251158, Accumulated Leap Seconds=36
  { 2457503.50, 68.310519 }, //25 April 2016, UT1-UTC=-0.1265186, Accumulated Leap Seconds=36
  { 2457504.50, 68.311798 }, //26 April 2016, UT1-UTC=-0.1277979, Accumulated Leap Seconds=36
  { 2457505.50, 68.313029 }, //27 April 2016, UT1-UTC=-0.1290288, Accumulated Leap Seconds=36
  { 2457506.50, 68.314317 }, //28 April 2016, UT1-UTC=-0.1303167, Accumulated Leap Seconds=36
  { 2457507.50, 68.315701 }, //29 April 2016, UT1-UTC=-0.1317012, Accumulated Leap Seconds=36
  { 2457508.50, 68.317193 }, //30 April 2016, UT1-UTC=-0.1331931, Accumulated Leap Seconds=36
  { 2457509.50, 68.318817 }, //01 May 2016, UT1-UTC=-0.1348171, Accumulated Leap Seconds=36
  { 2457510.50, 68.320586 }, //02 May 2016, UT1-UTC=-0.1365862, Accumulated Leap Seconds=36
  { 2457511.50, 68.322540 }, //03 May 2016, UT1-UTC=-0.1385399, Accumulated Leap Seconds=36
  { 2457512.50, 68.324670 }, //04 May 2016, UT1-UTC=-0.1406699, Accumulated Leap Seconds=36
  { 2457513.50, 68.326863 }, //05 May 2016, UT1-UTC=-0.1428627, Accumulated Leap Seconds=36
  { 2457514.50, 68.329010 }, //06 May 2016, UT1-UTC=-0.1450098, Accumulated Leap Seconds=36
  { 2457515.50, 68.331030 }, //07 May 2016, UT1-UTC=-0.1470298, Accumulated Leap Seconds=36
  { 2457516.50, 68.332878 }, //08 May 2016, UT1-UTC=-0.1488783, Accumulated Leap Seconds=36
  { 2457517.50, 68.334549 }, //09 May 2016, UT1-UTC=-0.1505493, Accumulated Leap Seconds=36
  { 2457518.50, 68.336073 }, //10 May 2016, UT1-UTC=-0.1520726, Accumulated Leap Seconds=36
  { 2457519.50, 68.337475 }, //11 May 2016, UT1-UTC=-0.1534752, Accumulated Leap Seconds=36
  { 2457520.50, 68.338831 }, //12 May 2016, UT1-UTC=-0.1548314, Accumulated Leap Seconds=36
  { 2457521.50, 68.340259 }, //13 May 2016, UT1-UTC=-0.1562592, Accumulated Leap Seconds=36
  { 2457522.50, 68.341809 }, //14 May 2016, UT1-UTC=-0.1578095, Accumulated Leap Seconds=36
  { 2457523.50, 68.343459 }, //15 May 2016, UT1-UTC=-0.1594593, Accumulated Leap Seconds=36
  { 2457524.50, 68.345178 }, //16 May 2016, UT1-UTC=-0.1611781, Accumulated Leap Seconds=36
  { 2457525.50, 68.346928 }, //17 May 2016, UT1-UTC=-0.1629276, Accumulated Leap Seconds=36
  { 2457526.50, 68.348654 }, //18 May 2016, UT1-UTC=-0.1646540, Accumulated Leap Seconds=36
  { 2457527.50, 68.350296 }, //19 May 2016, UT1-UTC=-0.1662957, Accumulated Leap Seconds=36
  { 2457528.50, 68.351864 }, //20 May 2016, UT1-UTC=-0.1678640, Accumulated Leap Seconds=36
  { 2457529.50, 68.353378 }, //21 May 2016, UT1-UTC=-0.1693776, Accumulated Leap Seconds=36
  { 2457530.50, 68.354842 }, //22 May 2016, UT1-UTC=-0.1708417, Accumulated Leap Seconds=36
  { 2457531.50, 68.356305 }, //23 May 2016, UT1-UTC=-0.1723050, Accumulated Leap Seconds=36
  { 2457532.50, 68.357760 }, //24 May 2016, UT1-UTC=-0.1737600, Accumulated Leap Seconds=36
  { 2457533.50, 68.359170 }, //25 May 2016, UT1-UTC=-0.1751700, Accumulated Leap Seconds=36
  { 2457534.50, 68.360536 }, //26 May 2016, UT1-UTC=-0.1765358, Accumulated Leap Seconds=36
  { 2457535.50, 68.361918 }, //27 May 2016, UT1-UTC=-0.1779183, Accumulated Leap Seconds=36
  { 2457536.50, 68.363389 }, //28 May 2016, UT1-UTC=-0.1793894, Accumulated Leap Seconds=36
  { 2457537.50, 68.364976 }, //29 May 2016, UT1-UTC=-0.1809761, Accumulated Leap Seconds=36
  { 2457538.50, 68.366702 }, //30 May 2016, UT1-UTC=-0.1827016, Accumulated Leap Seconds=36
  { 2457539.50, 68.368513 }, //31 May 2016, UT1-UTC=-0.1845129, Accumulated Leap Seconds=36
  { 2457540.50, 68.370356 }, //01 June 2016, UT1-UTC=-0.1863564, Accumulated Leap Seconds=36
  { 2457541.50, 68.372195 }, //02 June 2016, UT1-UTC=-0.1881947, Accumulated Leap Seconds=36
  { 2457542.50, 68.373937 }, //03 June 2016, UT1-UTC=-0.1899367, Accumulated Leap Seconds=36
  { 2457543.50, 68.375490 }, //04 June 2016, UT1-UTC=-0.1914902, Accumulated Leap Seconds=36
  { 2457544.50, 68.376844 }, //05 June 2016, UT1-UTC=-0.1928440, Accumulated Leap Seconds=36
  { 2457545.50, 68.378049 }, //06 June 2016, UT1-UTC=-0.1940489, Accumulated Leap Seconds=36
  { 2457546.50, 68.379150 }, //07 June 2016, UT1-UTC=-0.1951500, Accumulated Leap Seconds=36
  { 2457547.50, 68.380185 }, //08 June 2016, UT1-UTC=-0.1961851, Accumulated Leap Seconds=36
  { 2457548.50, 68.381171 }, //09 June 2016, UT1-UTC=-0.1971713, Accumulated Leap Seconds=36
  { 2457549.50, 68.382113 }, //10 June 2016, UT1-UTC=-0.1981130, Accumulated Leap Seconds=36
  { 2457550.50, 68.383042 }, //11 June 2016, UT1-UTC=-0.1990420, Accumulated Leap Seconds=36
  { 2457551.50, 68.383971 }, //12 June 2016, UT1-UTC=-0.1999712, Accumulated Leap Seconds=36
  { 2457552.50, 68.384860 }, //13 June 2016, UT1-UTC=-0.2008601, Accumulated Leap Seconds=36
  { 2457553.50, 68.385680 }, //14 June 2016, UT1-UTC=-0.2016797, Accumulated Leap Seconds=36
  { 2457554.50, 68.386426 }, //15 June 2016, UT1-UTC=-0.2024261, Accumulated Leap Seconds=36
  { 2457555.50, 68.387061 }, //16 June 2016, UT1-UTC=-0.2030606, Accumulated Leap Seconds=36
  { 2457556.50, 68.387581 }, //17 June 2016, UT1-UTC=-0.2035810, Accumulated Leap Seconds=36
  { 2457557.50, 68.388030 }, //18 June 2016, UT1-UTC=-0.2040299, Accumulated Leap Seconds=36
  { 2457558.50, 68.388359 }, //19 June 2016, UT1-UTC=-0.2043590, Accumulated Leap Seconds=36
  { 2457559.50, 68.388617 }, //20 June 2016, UT1-UTC=-0.2046174, Accumulated Leap Seconds=36
  { 2457560.50, 68.388834 }, //21 June 2016, UT1-UTC=-0.2048343, Accumulated Leap Seconds=36
  { 2457561.50, 68.389081 }, //22 June 2016, UT1-UTC=-0.2050810, Accumulated Leap Seconds=36
  { 2457562.50, 68.389432 }, //23 June 2016, UT1-UTC=-0.2054319, Accumulated Leap Seconds=36
  { 2457563.50, 68.389890 }, //24 June 2016, UT1-UTC=-0.2058903, Accumulated Leap Seconds=36
  { 2457564.50, 68.390513 }, //25 June 2016, UT1-UTC=-0.2065126, Accumulated Leap Seconds=36
  { 2457565.50, 68.391317 }, //26 June 2016, UT1-UTC=-0.2073165, Accumulated Leap Seconds=36
  { 2457566.50, 68.392262 }, //27 June 2016, UT1-UTC=-0.2082620, Accumulated Leap Seconds=36
  { 2457567.50, 68.393307 }, //28 June 2016, UT1-UTC=-0.2093075, Accumulated Leap Seconds=36
  { 2457568.50, 68.394414 }, //29 June 2016, UT1-UTC=-0.2104145, Accumulated Leap Seconds=36
  { 2457569.50, 68.395482 }, //30 June 2016, UT1-UTC=-0.2114824, Accumulated Leap Seconds=36
  { 2457570.50, 68.396436 }, //01 July 2016, UT1-UTC=-0.2124356, Accumulated Leap Seconds=36
  { 2457571.50, 68.397314 }, //02 July 2016, UT1-UTC=-0.2133138, Accumulated Leap Seconds=36
  { 2457572.50, 68.398103 }, //03 July 2016, UT1-UTC=-0.2141035, Accumulated Leap Seconds=36
  { 2457573.50, 68.398752 }, //04 July 2016, UT1-UTC=-0.2147524, Accumulated Leap Seconds=36
  { 2457574.50, 68.399314 }, //05 July 2016, UT1-UTC=-0.2153143, Accumulated Leap Seconds=36
  { 2457575.50, 68.399836 }, //06 July 2016, UT1-UTC=-0.2158356, Accumulated Leap Seconds=36
  { 2457576.50, 68.400348 }, //07 July 2016, UT1-UTC=-0.2163479, Accumulated Leap Seconds=36
  { 2457577.50, 68.400902 }, //08 July 2016, UT1-UTC=-0.2169018, Accumulated Leap Seconds=36
  { 2457578.50, 68.401510 }, //09 July 2016, UT1-UTC=-0.2175095, Accumulated Leap Seconds=36
  { 2457579.50, 68.402145 }, //10 July 2016, UT1-UTC=-0.2181454, Accumulated Leap Seconds=36
  { 2457580.50, 68.402691 }, //11 July 2016, UT1-UTC=-0.2186913, Accumulated Leap Seconds=36
  { 2457581.50, 68.403142 }, //12 July 2016, UT1-UTC=-0.2191422, Accumulated Leap Seconds=36
  { 2457582.50, 68.403541 }, //13 July 2016, UT1-UTC=-0.2195406, Accumulated Leap Seconds=36
  { 2457583.50, 68.403847 }, //14 July 2016, UT1-UTC=-0.2198472, Accumulated Leap Seconds=36
  { 2457584.50, 68.404006 }, //15 July 2016, UT1-UTC=-0.2200059, Accumulated Leap Seconds=36
  { 2457585.50, 68.404047 }, //16 July 2016, UT1-UTC=-0.2200471, Accumulated Leap Seconds=36
  { 2457586.50, 68.404032 }, //17 July 2016, UT1-UTC=-0.2200325, Accumulated Leap Seconds=36
  { 2457587.50, 68.404006 }, //18 July 2016, UT1-UTC=-0.2200059, Accumulated Leap Seconds=36
  { 2457588.50, 68.403998 }, //19 July 2016, UT1-UTC=-0.2199975, Accumulated Leap Seconds=36
  { 2457589.50, 68.404069 }, //20 July 2016, UT1-UTC=-0.2200691, Accumulated Leap Seconds=36
  { 2457590.50, 68.404250 }, //21 July 2016, UT1-UTC=-0.2202501, Accumulated Leap Seconds=36
  { 2457591.50, 68.404554 }, //22 July 2016, UT1-UTC=-0.2205537, Accumulated Leap Seconds=36
  { 2457592.50, 68.404971 }, //23 July 2016, UT1-UTC=-0.2209708, Accumulated Leap Seconds=36
  { 2457593.50, 68.405450 }, //24 July 2016, UT1-UTC=-0.2214499, Accumulated Leap Seconds=36
  { 2457594.50, 68.405995 }, //25 July 2016, UT1-UTC=-0.2219951, Accumulated Leap Seconds=36
  { 2457595.50, 68.406600 }, //26 July 2016, UT1-UTC=-0.2226003, Accumulated Leap Seconds=36
  { 2457596.50, 68.407201 }, //27 July 2016, UT1-UTC=-0.2232010, Accumulated Leap Seconds=36
  { 2457597.50, 68.407760 }, //28 July 2016, UT1-UTC=-0.2237600, Accumulated Leap Seconds=36
  { 2457598.50, 68.408266 }, //29 July 2016, UT1-UTC=-0.2242661, Accumulated Leap Seconds=36
  { 2457599.50, 68.408711 }, //30 July 2016, UT1-UTC=-0.2247111, Accumulated Leap Seconds=36
  { 2457600.50, 68.409098 }, //31 July 2016, UT1-UTC=-0.2250985, Accumulated Leap Seconds=36
  { 2457601.50, 68.409447 }, //01 August 2016, UT1-UTC=-0.2254472, Accumulated Leap Seconds=36
  { 2457602.50, 68.409841 }, //02 August 2016, UT1-UTC=-0.2258411, Accumulated Leap Seconds=36
  { 2457603.50, 68.410339 }, //03 August 2016, UT1-UTC=-0.2263390, Accumulated Leap Seconds=36
  { 2457604.50, 68.410919 }, //04 August 2016, UT1-UTC=-0.2269191, Accumulated Leap Seconds=36
  { 2457605.50, 68.411568 }, //05 August 2016, UT1-UTC=-0.2275676, Accumulated Leap Seconds=36
  { 2457606.50, 68.412255 }, //06 August 2016, UT1-UTC=-0.2282547, Accumulated Leap Seconds=36
  { 2457607.50, 68.412963 }, //07 August 2016, UT1-UTC=-0.2289632, Accumulated Leap Seconds=36
  { 2457608.50, 68.413673 }, //08 August 2016, UT1-UTC=-0.2296730, Accumulated Leap Seconds=36
  { 2457609.50, 68.414366 }, //09 August 2016, UT1-UTC=-0.2303661, Accumulated Leap Seconds=36
  { 2457610.50, 68.415028 }, //10 August 2016, UT1-UTC=-0.2310283, Accumulated Leap Seconds=36
  { 2457611.50, 68.415606 }, //11 August 2016, UT1-UTC=-0.2316058, Accumulated Leap Seconds=36
  { 2457612.50, 68.416082 }, //12 August 2016, UT1-UTC=-0.2320822, Accumulated Leap Seconds=36
  { 2457613.50, 68.416444 }, //13 August 2016, UT1-UTC=-0.2324437, Accumulated Leap Seconds=36
  { 2457614.50, 68.416714 }, //14 August 2016, UT1-UTC=-0.2327144, Accumulated Leap Seconds=36
  { 2457615.50, 68.416975 }, //15 August 2016, UT1-UTC=-0.2329751, Accumulated Leap Seconds=36
  { 2457616.50, 68.417323 }, //16 August 2016, UT1-UTC=-0.2333231, Accumulated Leap Seconds=36
  { 2457617.50, 68.417796 }, //17 August 2016, UT1-UTC=-0.2337959, Accumulated Leap Seconds=36
  { 2457618.50, 68.418410 }, //18 August 2016, UT1-UTC=-0.2344103, Accumulated Leap Seconds=36
  { 2457619.50, 68.419233 }, //19 August 2016, UT1-UTC=-0.2352334, Accumulated Leap Seconds=36
  { 2457620.50, 68.420229 }, //20 August 2016, UT1-UTC=-0.2362291, Accumulated Leap Seconds=36
  { 2457621.50, 68.421322 }, //21 August 2016, UT1-UTC=-0.2373224, Accumulated Leap Seconds=36
  { 2457622.50, 68.422444 }, //22 August 2016, UT1-UTC=-0.2384439, Accumulated Leap Seconds=36
  { 2457623.50, 68.423546 }, //23 August 2016, UT1-UTC=-0.2395462, Accumulated Leap Seconds=36
  { 2457624.50, 68.424574 }, //24 August 2016, UT1-UTC=-0.2405738, Accumulated Leap Seconds=36
  { 2457625.50, 68.425496 }, //25 August 2016, UT1-UTC=-0.2414956, Accumulated Leap Seconds=36
  { 2457626.50, 68.426326 }, //26 August 2016, UT1-UTC=-0.2423264, Accumulated Leap Seconds=36
  { 2457627.50, 68.427064 }, //27 August 2016, UT1-UTC=-0.2430636, Accumulated Leap Seconds=36
  { 2457628.50, 68.427729 }, //28 August 2016, UT1-UTC=-0.2437289, Accumulated Leap Seconds=36
  { 2457629.50, 68.428333 }, //29 August 2016, UT1-UTC=-0.2443330, Accumulated Leap Seconds=36
  { 2457630.50, 68.428960 }, //30 August 2016, UT1-UTC=-0.2449597, Accumulated Leap Seconds=36
  { 2457631.50, 68.429680 }, //31 August 2016, UT1-UTC=-0.2456797, Accumulated Leap Seconds=36
  { 2457632.50, 68.430461 }, //01 September 2016, UT1-UTC=-0.2464611, Accumulated Leap Seconds=36
  { 2457633.50, 68.431292 }, //02 September 2016, UT1-UTC=-0.2472920, Accumulated Leap Seconds=36
  { 2457634.50, 68.432177 }, //03 September 2016, UT1-UTC=-0.2481766, Accumulated Leap Seconds=36
  { 2457635.50, 68.433119 }, //04 September 2016, UT1-UTC=-0.2491189, Accumulated Leap Seconds=36
  { 2457636.50, 68.434086 }, //05 September 2016, UT1-UTC=-0.2500864, Accumulated Leap Seconds=36
  { 2457637.50, 68.435033 }, //06 September 2016, UT1-UTC=-0.2510326, Accumulated Leap Seconds=36
  { 2457638.50, 68.435919 }, //07 September 2016, UT1-UTC=-0.2519190, Accumulated Leap Seconds=36
  { 2457639.50, 68.436746 }, //08 September 2016, UT1-UTC=-0.2527462, Accumulated Leap Seconds=36
  { 2457640.50, 68.437511 }, //09 September 2016, UT1-UTC=-0.2535109, Accumulated Leap Seconds=36
  { 2457641.50, 68.438194 }, //10 September 2016, UT1-UTC=-0.2541938, Accumulated Leap Seconds=36
  { 2457642.50, 68.438892 }, //11 September 2016, UT1-UTC=-0.2548922, Accumulated Leap Seconds=36
  { 2457643.50, 68.439633 }, //12 September 2016, UT1-UTC=-0.2556327, Accumulated Leap Seconds=36
  { 2457644.50, 68.440407 }, //13 September 2016, UT1-UTC=-0.2564073, Accumulated Leap Seconds=36
  { 2457645.50, 68.441238 }, //14 September 2016, UT1-UTC=-0.2572379, Accumulated Leap Seconds=36
  { 2457646.50, 68.442296 }, //15 September 2016, UT1-UTC=-0.2582958, Accumulated Leap Seconds=36
  { 2457647.50, 68.443545 }, //16 September 2016, UT1-UTC=-0.2595455, Accumulated Leap Seconds=36
  { 2457648.50, 68.444937 }, //17 September 2016, UT1-UTC=-0.2609369, Accumulated Leap Seconds=36
  { 2457649.50, 68.446405 }, //18 September 2016, UT1-UTC=-0.2624053, Accumulated Leap Seconds=36
  { 2457650.50, 68.447902 }, //19 September 2016, UT1-UTC=-0.2639019, Accumulated Leap Seconds=36
  { 2457651.50, 68.449366 }, //20 September 2016, UT1-UTC=-0.2653663, Accumulated Leap Seconds=36
  { 2457652.50, 68.450741 }, //21 September 2016, UT1-UTC=-0.2667412, Accumulated Leap Seconds=36
  { 2457653.50, 68.451975 }, //22 September 2016, UT1-UTC=-0.2679749, Accumulated Leap Seconds=36
  { 2457654.50, 68.453108 }, //23 September 2016, UT1-UTC=-0.2691085, Accumulated Leap Seconds=36
  { 2457655.50, 68.454190 }, //24 September 2016, UT1-UTC=-0.2701901, Accumulated Leap Seconds=36
  { 2457656.50, 68.455257 }, //25 September 2016, UT1-UTC=-0.2712571, Accumulated Leap Seconds=36
  { 2457657.50, 68.456385 }, //26 September 2016, UT1-UTC=-0.2723850, Accumulated Leap Seconds=36
  { 2457658.50, 68.457577 }, //27 September 2016, UT1-UTC=-0.2735775, Accumulated Leap Seconds=36
  { 2457659.50, 68.458815 }, //28 September 2016, UT1-UTC=-0.2748146, Accumulated Leap Seconds=36
  { 2457660.50, 68.460150 }, //29 September 2016, UT1-UTC=-0.2761502, Accumulated Leap Seconds=36
  { 2457661.50, 68.461575 }, //30 September 2016, UT1-UTC=-0.2775748, Accumulated Leap Seconds=36
  { 2457662.50, 68.462979 }, //01 October 2016, UT1-UTC=-0.2789791, Accumulated Leap Seconds=36
  { 2457663.50, 68.464340 }, //02 October 2016, UT1-UTC=-0.2803398, Accumulated Leap Seconds=36
  { 2457664.50, 68.465604 }, //03 October 2016, UT1-UTC=-0.2816038, Accumulated Leap Seconds=36
  { 2457665.50, 68.466773 }, //04 October 2016, UT1-UTC=-0.2827729, Accumulated Leap Seconds=36
  { 2457666.50, 68.467889 }, //05 October 2016, UT1-UTC=-0.2838894, Accumulated Leap Seconds=36
  { 2457667.50, 68.468945 }, //06 October 2016, UT1-UTC=-0.2849447, Accumulated Leap Seconds=36
  { 2457668.50, 68.469934 }, //07 October 2016, UT1-UTC=-0.2859339, Accumulated Leap Seconds=36
  { 2457669.50, 68.470900 }, //08 October 2016, UT1-UTC=-0.2869004, Accumulated Leap Seconds=36
  { 2457670.50, 68.471873 }, //09 October 2016, UT1-UTC=-0.2878725, Accumulated Leap Seconds=36
  { 2457671.50, 68.472932 }, //10 October 2016, UT1-UTC=-0.2889315, Accumulated Leap Seconds=36
  { 2457672.50, 68.474151 }, //11 October 2016, UT1-UTC=-0.2901510, Accumulated Leap Seconds=36
  { 2457673.50, 68.475565 }, //12 October 2016, UT1-UTC=-0.2915646, Accumulated Leap Seconds=36
  { 2457674.50, 68.477181 }, //13 October 2016, UT1-UTC=-0.2931808, Accumulated Leap Seconds=36
  { 2457675.50, 68.479011 }, //14 October 2016, UT1-UTC=-0.2950111, Accumulated Leap Seconds=36
  { 2457676.50, 68.481021 }, //15 October 2016, UT1-UTC=-0.2970207, Accumulated Leap Seconds=36
  { 2457677.50, 68.483114 }, //16 October 2016, UT1-UTC=-0.2991143, Accumulated Leap Seconds=36
  { 2457678.50, 68.485174 }, //17 October 2016, UT1-UTC=-0.3011743, Accumulated Leap Seconds=36
  { 2457679.50, 68.487093 }, //18 October 2016, UT1-UTC=-0.3030930, Accumulated Leap Seconds=36
  { 2457680.50, 68.488857 }, //19 October 2016, UT1-UTC=-0.3048574, Accumulated Leap Seconds=36
  { 2457681.50, 68.490461 }, //20 October 2016, UT1-UTC=-0.3064606, Accumulated Leap Seconds=36
  { 2457682.50, 68.491907 }, //21 October 2016, UT1-UTC=-0.3079075, Accumulated Leap Seconds=36
  { 2457683.50, 68.493316 }, //22 October 2016, UT1-UTC=-0.3093158, Accumulated Leap Seconds=36
  { 2457684.50, 68.494745 }, //23 October 2016, UT1-UTC=-0.3107455, Accumulated Leap Seconds=36
  { 2457685.50, 68.496168 }, //24 October 2016, UT1-UTC=-0.3121681, Accumulated Leap Seconds=36
  { 2457686.50, 68.497592 }, //25 October 2016, UT1-UTC=-0.3135924, Accumulated Leap Seconds=36
  { 2457687.50, 68.499065 }, //26 October 2016, UT1-UTC=-0.3150654, Accumulated Leap Seconds=36
  { 2457688.50, 68.500535 }, //27 October 2016, UT1-UTC=-0.3165353, Accumulated Leap Seconds=36
  { 2457689.50, 68.502004 }, //28 October 2016, UT1-UTC=-0.3180035, Accumulated Leap Seconds=36
  { 2457690.50, 68.503498 }, //29 October 2016, UT1-UTC=-0.3194982, Accumulated Leap Seconds=36
  { 2457691.50, 68.504998 }, //30 October 2016, UT1-UTC=-0.3209980, Accumulated Leap Seconds=36
  { 2457692.50, 68.506448 }, //31 October 2016, UT1-UTC=-0.3224484, Accumulated Leap Seconds=36
  { 2457693.50, 68.507818 }, //01 November 2016, UT1-UTC=-0.3238180, Accumulated Leap Seconds=36
  { 2457694.50, 68.509132 }, //02 November 2016, UT1-UTC=-0.3251318, Accumulated Leap Seconds=36
  { 2457695.50, 68.510393 }, //03 November 2016, UT1-UTC=-0.3263927, Accumulated Leap Seconds=36
  { 2457696.50, 68.511625 }, //04 November 2016, UT1-UTC=-0.3276250, Accumulated Leap Seconds=36
  { 2457697.50, 68.512870 }, //05 November 2016, UT1-UTC=-0.3288702, Accumulated Leap Seconds=36
  { 2457698.50, 68.514196 }, //06 November 2016, UT1-UTC=-0.3301963, Accumulated Leap Seconds=36
  { 2457699.50, 68.515672 }, //07 November 2016, UT1-UTC=-0.3316717, Accumulated Leap Seconds=36
  { 2457700.50, 68.517270 }, //08 November 2016, UT1-UTC=-0.3332698, Accumulated Leap Seconds=36
  { 2457701.50, 68.519005 }, //09 November 2016, UT1-UTC=-0.3350051, Accumulated Leap Seconds=36
  { 2457702.50, 68.520901 }, //10 November 2016, UT1-UTC=-0.3369012, Accumulated Leap Seconds=36
  { 2457703.50, 68.522945 }, //11 November 2016, UT1-UTC=-0.3389450, Accumulated Leap Seconds=36
  { 2457704.50, 68.525097 }, //12 November 2016, UT1-UTC=-0.3410974, Accumulated Leap Seconds=36
  { 2457705.50, 68.527236 }, //13 November 2016, UT1-UTC=-0.3432364, Accumulated Leap Seconds=36
  { 2457706.50, 68.529259 }, //14 November 2016, UT1-UTC=-0.3452590, Accumulated Leap Seconds=36
  { 2457707.50, 68.531100 }, //15 November 2016, UT1-UTC=-0.3470999, Accumulated Leap Seconds=36
  { 2457708.50, 68.532693 }, //16 November 2016, UT1-UTC=-0.3486935, Accumulated Leap Seconds=36
  { 2457709.50, 68.534072 }, //17 November 2016, UT1-UTC=-0.3500719, Accumulated Leap Seconds=36
  { 2457710.50, 68.535358 }, //18 November 2016, UT1-UTC=-0.3513579, Accumulated Leap Seconds=36
  { 2457711.50, 68.536624 }, //19 November 2016, UT1-UTC=-0.3526242, Accumulated Leap Seconds=36
  { 2457712.50, 68.537944 }, //20 November 2016, UT1-UTC=-0.3539436, Accumulated Leap Seconds=36
  { 2457713.50, 68.539327 }, //21 November 2016, UT1-UTC=-0.3553270, Accumulated Leap Seconds=36
  { 2457714.50, 68.540807 }, //22 November 2016, UT1-UTC=-0.3568065, Accumulated Leap Seconds=36
  { 2457715.50, 68.542332 }, //23 November 2016, UT1-UTC=-0.3583322, Accumulated Leap Seconds=36
  { 2457716.50, 68.543898 }, //24 November 2016, UT1-UTC=-0.3598979, Accumulated Leap Seconds=36
  { 2457717.50, 68.545480 }, //25 November 2016, UT1-UTC=-0.3614801, Accumulated Leap Seconds=36
  { 2457718.50, 68.547035 }, //26 November 2016, UT1-UTC=-0.3630352, Accumulated Leap Seconds=36
  { 2457719.50, 68.548514 }, //27 November 2016, UT1-UTC=-0.3645141, Accumulated Leap Seconds=36
  { 2457720.50, 68.549904 }, //28 November 2016, UT1-UTC=-0.3659039, Accumulated Leap Seconds=36
  { 2457721.50, 68.551215 }, //29 November 2016, UT1-UTC=-0.3672153, Accumulated Leap Seconds=36
  { 2457722.50, 68.552473 }, //30 November 2016, UT1-UTC=-0.3684730, Accumulated Leap Seconds=36
  { 2457723.50, 68.553702 }, //01 December 2016, UT1-UTC=-0.3697018, Accumulated Leap Seconds=36
  { 2457724.50, 68.554917 }, //02 December 2016, UT1-UTC=-0.3709165, Accumulated Leap Seconds=36
  { 2457725.50, 68.556125 }, //03 December 2016, UT1-UTC=-0.3721246, Accumulated Leap Seconds=36
  { 2457726.50, 68.557334 }, //04 December 2016, UT1-UTC=-0.3733343, Accumulated Leap Seconds=36
  { 2457727.50, 68.558646 }, //05 December 2016, UT1-UTC=-0.3746458, Accumulated Leap Seconds=36
  { 2457728.50, 68.560048 }, //06 December 2016, UT1-UTC=-0.3760479, Accumulated Leap Seconds=36
  { 2457729.50, 68.561547 }, //07 December 2016, UT1-UTC=-0.3775474, Accumulated Leap Seconds=36
  { 2457730.50, 68.563172 }, //08 December 2016, UT1-UTC=-0.3791717, Accumulated Leap Seconds=36
  { 2457731.50, 68.564912 }, //09 December 2016, UT1-UTC=-0.3809123, Accumulated Leap Seconds=36
  { 2457732.50, 68.566729 }, //10 December 2016, UT1-UTC=-0.3827288, Accumulated Leap Seconds=36
  { 2457733.50, 68.568517 }, //11 December 2016, UT1-UTC=-0.3845167, Accumulated Leap Seconds=36
  { 2457734.50, 68.570187 }, //12 December 2016, UT1-UTC=-0.3861868, Accumulated Leap Seconds=36
  { 2457735.50, 68.571701 }, //13 December 2016, UT1-UTC=-0.3877010, Accumulated Leap Seconds=36
  { 2457736.50, 68.573022 }, //14 December 2016, UT1-UTC=-0.3890224, Accumulated Leap Seconds=36
  { 2457737.50, 68.574251 }, //15 December 2016, UT1-UTC=-0.3902508, Accumulated Leap Seconds=36
  { 2457738.50, 68.575491 }, //16 December 2016, UT1-UTC=-0.3914912, Accumulated Leap Seconds=36
  { 2457739.50, 68.576760 }, //17 December 2016, UT1-UTC=-0.3927602, Accumulated Leap Seconds=36
  { 2457740.50, 68.578024 }, //18 December 2016, UT1-UTC=-0.3940241, Accumulated Leap Seconds=36
  { 2457741.50, 68.579282 }, //19 December 2016, UT1-UTC=-0.3952822, Accumulated Leap Seconds=36
  { 2457742.50, 68.580562 }, //20 December 2016, UT1-UTC=-0.3965624, Accumulated Leap Seconds=36
  { 2457743.50, 68.581837 }, //21 December 2016, UT1-UTC=-0.3978370, Accumulated Leap Seconds=36
  { 2457744.50, 68.583102 }, //22 December 2016, UT1-UTC=-0.3991024, Accumulated Leap Seconds=36
  { 2457745.50, 68.584339 }, //23 December 2016, UT1-UTC=-0.4003390, Accumulated Leap Seconds=36
  { 2457746.50, 68.585510 }, //24 December 2016, UT1-UTC=-0.4015098, Accumulated Leap Seconds=36
  { 2457747.50, 68.586582 }, //25 December 2016, UT1-UTC=-0.4025817, Accumulated Leap Seconds=36
  { 2457748.50, 68.587553 }, //26 December 2016, UT1-UTC=-0.4035526, Accumulated Leap Seconds=36
  { 2457749.50, 68.588454 }, //27 December 2016, UT1-UTC=-0.4044543, Accumulated Leap Seconds=36
  { 2457750.50, 68.589285 }, //28 December 2016, UT1-UTC=-0.4052850, Accumulated Leap Seconds=36
  { 2457751.50, 68.590088 }, //29 December 2016, UT1-UTC=-0.4060885, Accumulated Leap Seconds=36
  { 2457752.50, 68.590918 }, //30 December 2016, UT1-UTC=-0.4069180, Accumulated Leap Seconds=36
  { 2457753.50, 68.591760 }, //31 December 2016, UT1-UTC=-0.4077601, Accumulated Leap Seconds=36
  { 2457754.50, 68.592718 }, //01 January 2017, UT1-UTC= 0.5912821, Accumulated Leap Seconds=37
  { 2457755.50, 68.593825 }, //02 January 2017, UT1-UTC= 0.5901752, Accumulated Leap Seconds=37
  { 2457756.50, 68.595059 }, //03 January 2017, UT1-UTC= 0.5889406, Accumulated Leap Seconds=37
  { 2457757.50, 68.596437 }, //04 January 2017, UT1-UTC= 0.5875626, Accumulated Leap Seconds=37
  { 2457758.50, 68.597983 }, //05 January 2017, UT1-UTC= 0.5860173, Accumulated Leap Seconds=37
  { 2457759.50, 68.599574 }, //06 January 2017, UT1-UTC= 0.5844262, Accumulated Leap Seconds=37
  { 2457760.50, 68.601156 }, //07 January 2017, UT1-UTC= 0.5828444, Accumulated Leap Seconds=37
  { 2457761.50, 68.602695 }, //08 January 2017, UT1-UTC= 0.5813054, Accumulated Leap Seconds=37
  { 2457762.50, 68.604148 }, //09 January 2017, UT1-UTC= 0.5798520, Accumulated Leap Seconds=37
  { 2457763.50, 68.605481 }, //10 January 2017, UT1-UTC= 0.5785191, Accumulated Leap Seconds=37
  { 2457764.50, 68.606680 }, //11 January 2017, UT1-UTC= 0.5773203, Accumulated Leap Seconds=37
  { 2457765.50, 68.607842 }, //12 January 2017, UT1-UTC= 0.5761577, Accumulated Leap Seconds=37
  { 2457766.50, 68.609024 }, //13 January 2017, UT1-UTC= 0.5749761, Accumulated Leap Seconds=37
  { 2457767.50, 68.610253 }, //14 January 2017, UT1-UTC= 0.5737466, Accumulated Leap Seconds=37
  { 2457768.50, 68.611561 }, //15 January 2017, UT1-UTC= 0.5724390, Accumulated Leap Seconds=37
  { 2457769.50, 68.612918 }, //16 January 2017, UT1-UTC= 0.5710822, Accumulated Leap Seconds=37
  { 2457770.50, 68.614261 }, //17 January 2017, UT1-UTC= 0.5697394, Accumulated Leap Seconds=37
  { 2457771.50, 68.615526 }, //18 January 2017, UT1-UTC= 0.5684741, Accumulated Leap Seconds=37
  { 2457772.50, 68.616734 }, //19 January 2017, UT1-UTC= 0.5672660, Accumulated Leap Seconds=37
  { 2457773.50, 68.617853 }, //20 January 2017, UT1-UTC= 0.5661469, Accumulated Leap Seconds=37
  { 2457774.50, 68.618924 }, //21 January 2017, UT1-UTC= 0.5650763, Accumulated Leap Seconds=37
  { 2457775.50, 68.619946 }, //22 January 2017, UT1-UTC= 0.5640535, Accumulated Leap Seconds=37
  { 2457776.50, 68.620892 }, //23 January 2017, UT1-UTC= 0.5631082, Accumulated Leap Seconds=37
  { 2457777.50, 68.621742 }, //24 January 2017, UT1-UTC= 0.5622575, Accumulated Leap Seconds=37
  { 2457778.50, 68.622546 }, //25 January 2017, UT1-UTC= 0.5614538, Accumulated Leap Seconds=37
  { 2457779.50, 68.623335 }, //26 January 2017, UT1-UTC= 0.5606647, Accumulated Leap Seconds=37
  { 2457780.50, 68.624133 }, //27 January 2017, UT1-UTC= 0.5598669, Accumulated Leap Seconds=37
  { 2457781.50, 68.625018 }, //28 January 2017, UT1-UTC= 0.5589818, Accumulated Leap Seconds=37
  { 2457782.50, 68.626012 }, //29 January 2017, UT1-UTC= 0.5579876, Accumulated Leap Seconds=37
  { 2457783.50, 68.627160 }, //30 January 2017, UT1-UTC= 0.5568403, Accumulated Leap Seconds=37
  { 2457784.50, 68.628428 }, //31 January 2017, UT1-UTC= 0.5555718, Accumulated Leap Seconds=37
  { 2457785.50, 68.629811 }, //01 February 2017, UT1-UTC= 0.5541893, Accumulated Leap Seconds=37
  { 2457786.50, 68.631265 }, //02 February 2017, UT1-UTC= 0.5527345, Accumulated Leap Seconds=37
  { 2457787.50, 68.632735 }, //03 February 2017, UT1-UTC= 0.5512649, Accumulated Leap Seconds=37
  { 2457788.50, 68.634167 }, //04 February 2017, UT1-UTC= 0.5498326, Accumulated Leap Seconds=37
  { 2457789.50, 68.635528 }, //05 February 2017, UT1-UTC= 0.5484722, Accumulated Leap Seconds=37
  { 2457790.50, 68.636806 }, //06 February 2017, UT1-UTC= 0.5471941, Accumulated Leap Seconds=37
  { 2457791.50, 68.638038 }, //07 February 2017, UT1-UTC= 0.5459615, Accumulated Leap Seconds=37
  { 2457792.50, 68.639224 }, //08 February 2017, UT1-UTC= 0.5447764, Accumulated Leap Seconds=37
  { 2457793.50, 68.640432 }, //09 February 2017, UT1-UTC= 0.5435676, Accumulated Leap Seconds=37
  { 2457794.50, 68.641730 }, //10 February 2017, UT1-UTC= 0.5422695, Accumulated Leap Seconds=37
  { 2457795.50, 68.643141 }, //11 February 2017, UT1-UTC= 0.5408590, Accumulated Leap Seconds=37
  { 2457796.50, 68.644687 }, //12 February 2017, UT1-UTC= 0.5393132, Accumulated Leap Seconds=37
  { 2457797.50, 68.646321 }, //13 February 2017, UT1-UTC= 0.5376793, Accumulated Leap Seconds=37
  { 2457798.50, 68.647998 }, //14 February 2017, UT1-UTC= 0.5360017, Accumulated Leap Seconds=37
  { 2457799.50, 68.649661 }, //15 February 2017, UT1-UTC= 0.5343385, Accumulated Leap Seconds=37
  { 2457800.50, 68.651288 }, //16 February 2017, UT1-UTC= 0.5327118, Accumulated Leap Seconds=37
  { 2457801.50, 68.652810 }, //17 February 2017, UT1-UTC= 0.5311897, Accumulated Leap Seconds=37
  { 2457802.50, 68.654194 }, //18 February 2017, UT1-UTC= 0.5298063, Accumulated Leap Seconds=37
  { 2457803.50, 68.655370 }, //19 February 2017, UT1-UTC= 0.5286304, Accumulated Leap Seconds=37
  { 2457804.50, 68.656415 }, //20 February 2017, UT1-UTC= 0.5275852, Accumulated Leap Seconds=37
  { 2457805.50, 68.657384 }, //21 February 2017, UT1-UTC= 0.5266157, Accumulated Leap Seconds=37
  { 2457806.50, 68.658307 }, //22 February 2017, UT1-UTC= 0.5256930, Accumulated Leap Seconds=37
  { 2457807.50, 68.659248 }, //23 February 2017, UT1-UTC= 0.5247522, Accumulated Leap Seconds=37
  { 2457808.50, 68.660291 }, //24 February 2017, UT1-UTC= 0.5237091, Accumulated Leap Seconds=37
  { 2457809.50, 68.661386 }, //25 February 2017, UT1-UTC= 0.5226139, Accumulated Leap Seconds=37
  { 2457810.50, 68.662600 }, //26 February 2017, UT1-UTC= 0.5214002, Accumulated Leap Seconds=37
  { 2457811.50, 68.663937 }, //27 February 2017, UT1-UTC= 0.5200629, Accumulated Leap Seconds=37
  { 2457812.50, 68.665437 }, //28 February 2017, UT1-UTC= 0.5185628, Accumulated Leap Seconds=37
  { 2457813.50, 68.667063 }, //01 March 2017, UT1-UTC= 0.5169373, Accumulated Leap Seconds=37
  { 2457814.50, 68.668704 }, //02 March 2017, UT1-UTC= 0.5152957, Accumulated Leap Seconds=37
  { 2457815.50, 68.670258 }, //03 March 2017, UT1-UTC= 0.5137419, Accumulated Leap Seconds=37
  { 2457816.50, 68.671729 }, //04 March 2017, UT1-UTC= 0.5122705, Accumulated Leap Seconds=37
  { 2457817.50, 68.673126 }, //05 March 2017, UT1-UTC= 0.5108741, Accumulated Leap Seconds=37
  { 2457818.50, 68.674446 }, //06 March 2017, UT1-UTC= 0.5095544, Accumulated Leap Seconds=37
  { 2457819.50, 68.675676 }, //07 March 2017, UT1-UTC= 0.5083238, Accumulated Leap Seconds=37
  { 2457820.50, 68.676948 }, //08 March 2017, UT1-UTC= 0.5070521, Accumulated Leap Seconds=37
  { 2457821.50, 68.678277 }, //09 March 2017, UT1-UTC= 0.5057233, Accumulated Leap Seconds=37
  { 2457822.50, 68.679677 }, //10 March 2017, UT1-UTC= 0.5043233, Accumulated Leap Seconds=37
  { 2457823.50, 68.681142 }, //11 March 2017, UT1-UTC= 0.5028581, Accumulated Leap Seconds=37
  { 2457824.50, 68.682698 }, //12 March 2017, UT1-UTC= 0.5013018, Accumulated Leap Seconds=37
  { 2457825.50, 68.684332 }, //13 March 2017, UT1-UTC= 0.4996678, Accumulated Leap Seconds=37
  { 2457826.50, 68.686034 }, //14 March 2017, UT1-UTC= 0.4979656, Accumulated Leap Seconds=37
  { 2457827.50, 68.687687 }, //15 March 2017, UT1-UTC= 0.4963128, Accumulated Leap Seconds=37
  { 2457828.50, 68.689212 }, //16 March 2017, UT1-UTC= 0.4947876, Accumulated Leap Seconds=37
  { 2457829.50, 68.690600 }, //17 March 2017, UT1-UTC= 0.4933998, Accumulated Leap Seconds=37
  { 2457830.50, 68.691880 }, //18 March 2017, UT1-UTC= 0.4921200, Accumulated Leap Seconds=37
  { 2457831.50, 68.693068 }, //19 March 2017, UT1-UTC= 0.4909323, Accumulated Leap Seconds=37
  { 2457832.50, 68.694196 }, //20 March 2017, UT1-UTC= 0.4898042, Accumulated Leap Seconds=37
  { 2457833.50, 68.695335 }, //21 March 2017, UT1-UTC= 0.4886648, Accumulated Leap Seconds=37
  { 2457834.50, 68.696501 }, //22 March 2017, UT1-UTC= 0.4874986, Accumulated Leap Seconds=37
  { 2457835.50, 68.697761 }, //23 March 2017, UT1-UTC= 0.4862389, Accumulated Leap Seconds=37
  { 2457836.50, 68.699156 }, //24 March 2017, UT1-UTC= 0.4848439, Accumulated Leap Seconds=37
  { 2457837.50, 68.700628 }, //25 March 2017, UT1-UTC= 0.4833715, Accumulated Leap Seconds=37
  { 2457838.50, 68.702286 }, //26 March 2017, UT1-UTC= 0.4817144, Accumulated Leap Seconds=37
  { 2457839.50, 68.704104 }, //27 March 2017, UT1-UTC= 0.4798964, Accumulated Leap Seconds=37
  { 2457840.50, 68.706011 }, //28 March 2017, UT1-UTC= 0.4779892, Accumulated Leap Seconds=37
  { 2457841.50, 68.707993 }, //29 March 2017, UT1-UTC= 0.4760074, Accumulated Leap Seconds=37
  { 2457842.50, 68.709973 }, //30 March 2017, UT1-UTC= 0.4740274, Accumulated Leap Seconds=37
  { 2457843.50, 68.711823 }, //31 March 2017, UT1-UTC= 0.4721774, Accumulated Leap Seconds=37
  { 2457844.50, 68.713521 }, //01 April 2017, UT1-UTC= 0.4704792, Accumulated Leap Seconds=37
  { 2457845.50, 68.715091 }, //02 April 2017, UT1-UTC= 0.4689092, Accumulated Leap Seconds=37
  { 2457846.50, 68.716517 }, //03 April 2017, UT1-UTC= 0.4674833, Accumulated Leap Seconds=37
  { 2457847.50, 68.717849 }, //04 April 2017, UT1-UTC= 0.4661513, Accumulated Leap Seconds=37
  { 2457848.50, 68.719194 }, //05 April 2017, UT1-UTC= 0.4648063, Accumulated Leap Seconds=37
  { 2457849.50, 68.720565 }, //06 April 2017, UT1-UTC= 0.4634347, Accumulated Leap Seconds=37
  { 2457850.50, 68.722085 }, //07 April 2017, UT1-UTC= 0.4619147, Accumulated Leap Seconds=37
  { 2457851.50, 68.723697 }, //08 April 2017, UT1-UTC= 0.4603031, Accumulated Leap Seconds=37
  { 2457852.50, 68.725370 }, //09 April 2017, UT1-UTC= 0.4586299, Accumulated Leap Seconds=37
  { 2457853.50, 68.727018 }, //10 April 2017, UT1-UTC= 0.4569819, Accumulated Leap Seconds=37
  { 2457854.50, 68.728594 }, //11 April 2017, UT1-UTC= 0.4554060, Accumulated Leap Seconds=37
  { 2457855.50, 68.730153 }, //12 April 2017, UT1-UTC= 0.4538469, Accumulated Leap Seconds=37
  { 2457856.50, 68.731741 }, //13 April 2017, UT1-UTC= 0.4522586, Accumulated Leap Seconds=37
  { 2457857.50, 68.733300 }, //14 April 2017, UT1-UTC= 0.4506998, Accumulated Leap Seconds=37
  { 2457858.50, 68.734784 }, //15 April 2017, UT1-UTC= 0.4492158, Accumulated Leap Seconds=37
  { 2457859.50, 68.736201 }, //16 April 2017, UT1-UTC= 0.4477989, Accumulated Leap Seconds=37
  { 2457860.50, 68.737639 }, //17 April 2017, UT1-UTC= 0.4463615, Accumulated Leap Seconds=37
  { 2457861.50, 68.739108 }, //18 April 2017, UT1-UTC= 0.4448916, Accumulated Leap Seconds=37
  { 2457862.50, 68.740589 }, //19 April 2017, UT1-UTC= 0.4434115, Accumulated Leap Seconds=37
  { 2457863.50, 68.742097 }, //20 April 2017, UT1-UTC= 0.4419032, Accumulated Leap Seconds=37
  { 2457864.50, 68.743643 }, //21 April 2017, UT1-UTC= 0.4403567, Accumulated Leap Seconds=37
  { 2457865.50, 68.745365 }, //22 April 2017, UT1-UTC= 0.4386346, Accumulated Leap Seconds=37
  { 2457866.50, 68.747271 }, //23 April 2017, UT1-UTC= 0.4367286, Accumulated Leap Seconds=37
  { 2457867.50, 68.749353 }, //24 April 2017, UT1-UTC= 0.4346468, Accumulated Leap Seconds=37
  { 2457868.50, 68.751541 }, //25 April 2017, UT1-UTC= 0.4324586, Accumulated Leap Seconds=37
  { 2457869.50, 68.753736 }, //26 April 2017, UT1-UTC= 0.4302640, Accumulated Leap Seconds=37
  { 2457870.50, 68.755863 }, //27 April 2017, UT1-UTC= 0.4281367, Accumulated Leap Seconds=37
  { 2457871.50, 68.757748 }, //28 April 2017, UT1-UTC= 0.4262524, Accumulated Leap Seconds=37
  { 2457872.50, 68.759411 }, //29 April 2017, UT1-UTC= 0.4245890, Accumulated Leap Seconds=37
  { 2457873.50, 68.760885 }, //30 April 2017, UT1-UTC= 0.4231149, Accumulated Leap Seconds=37
  { 2457874.50, 68.762276 }, //01 May 2017, UT1-UTC= 0.4217245, Accumulated Leap Seconds=37
  { 2457875.50, 68.763656 }, //02 May 2017, UT1-UTC= 0.4203435, Accumulated Leap Seconds=37
  { 2457876.50, 68.765036 }, //03 May 2017, UT1-UTC= 0.4189635, Accumulated Leap Seconds=37
  { 2457877.50, 68.766421 }, //04 May 2017, UT1-UTC= 0.4175785, Accumulated Leap Seconds=37
  { 2457878.50, 68.767888 }, //05 May 2017, UT1-UTC= 0.4161125, Accumulated Leap Seconds=37
  { 2457879.50, 68.769459 }, //06 May 2017, UT1-UTC= 0.4145406, Accumulated Leap Seconds=37
  { 2457880.50, 68.771091 }, //07 May 2017, UT1-UTC= 0.4129087, Accumulated Leap Seconds=37
  { 2457881.50, 68.772664 }, //08 May 2017, UT1-UTC= 0.4113358, Accumulated Leap Seconds=37
  { 2457882.50, 68.774072 }, //09 May 2017, UT1-UTC= 0.4099284, Accumulated Leap Seconds=37
  { 2457883.50, 68.775419 }, //10 May 2017, UT1-UTC= 0.4085807, Accumulated Leap Seconds=37
  { 2457884.50, 68.776704 }, //11 May 2017, UT1-UTC= 0.4072957, Accumulated Leap Seconds=37
  { 2457885.50, 68.777903 }, //12 May 2017, UT1-UTC= 0.4060969, Accumulated Leap Seconds=37
  { 2457886.50, 68.779012 }, //13 May 2017, UT1-UTC= 0.4049881, Accumulated Leap Seconds=37
  { 2457887.50, 68.780062 }, //14 May 2017, UT1-UTC= 0.4039375, Accumulated Leap Seconds=37
  { 2457888.50, 68.781040 }, //15 May 2017, UT1-UTC= 0.4029604, Accumulated Leap Seconds=37
  { 2457889.50, 68.782005 }, //16 May 2017, UT1-UTC= 0.4019948, Accumulated Leap Seconds=37
  { 2457890.50, 68.783079 }, //17 May 2017, UT1-UTC= 0.4009211, Accumulated Leap Seconds=37
  { 2457891.50, 68.784223 }, //18 May 2017, UT1-UTC= 0.3997768, Accumulated Leap Seconds=37
  { 2457892.50, 68.785427 }, //19 May 2017, UT1-UTC= 0.3985733, Accumulated Leap Seconds=37
  { 2457893.50, 68.786718 }, //20 May 2017, UT1-UTC= 0.3972820, Accumulated Leap Seconds=37
  { 2457894.50, 68.788141 }, //21 May 2017, UT1-UTC= 0.3958587, Accumulated Leap Seconds=37
  { 2457895.50, 68.789706 }, //22 May 2017, UT1-UTC= 0.3942937, Accumulated Leap Seconds=37
  { 2457896.50, 68.791368 }, //23 May 2017, UT1-UTC= 0.3926322, Accumulated Leap Seconds=37
  { 2457897.50, 68.793067 }, //24 May 2017, UT1-UTC= 0.3909326, Accumulated Leap Seconds=37
  { 2457898.50, 68.794689 }, //25 May 2017, UT1-UTC= 0.3893107, Accumulated Leap Seconds=37
  { 2457899.50, 68.796166 }, //26 May 2017, UT1-UTC= 0.3878337, Accumulated Leap Seconds=37
  { 2457900.50, 68.797462 }, //27 May 2017, UT1-UTC= 0.3865379, Accumulated Leap Seconds=37
  { 2457901.50, 68.798559 }, //28 May 2017, UT1-UTC= 0.3854409, Accumulated Leap Seconds=37
  { 2457902.50, 68.799612 }, //29 May 2017, UT1-UTC= 0.3843881, Accumulated Leap Seconds=37
  { 2457903.50, 68.800745 }, //30 May 2017, UT1-UTC= 0.3832551, Accumulated Leap Seconds=37
  { 2457904.50, 68.801966 }, //31 May 2017, UT1-UTC= 0.3820342, Accumulated Leap Seconds=37
  { 2457905.50, 68.803284 }, //01 June 2017, UT1-UTC= 0.3807157, Accumulated Leap Seconds=37
  { 2457906.50, 68.804654 }, //02 June 2017, UT1-UTC= 0.3793461, Accumulated Leap Seconds=37
  { 2457907.50, 68.806024 }, //03 June 2017, UT1-UTC= 0.3779760, Accumulated Leap Seconds=37
  { 2457908.50, 68.807318 }, //04 June 2017, UT1-UTC= 0.3766818, Accumulated Leap Seconds=37
  { 2457909.50, 68.808520 }, //05 June 2017, UT1-UTC= 0.3754796, Accumulated Leap Seconds=37
  { 2457910.50, 68.809633 }, //06 June 2017, UT1-UTC= 0.3743671, Accumulated Leap Seconds=37
  { 2457911.50, 68.810615 }, //07 June 2017, UT1-UTC= 0.3733854, Accumulated Leap Seconds=37
  { 2457912.50, 68.811409 }, //08 June 2017, UT1-UTC= 0.3725906, Accumulated Leap Seconds=37
  { 2457913.50, 68.812050 }, //09 June 2017, UT1-UTC= 0.3719500, Accumulated Leap Seconds=37
  { 2457914.50, 68.812542 }, //10 June 2017, UT1-UTC= 0.3714582, Accumulated Leap Seconds=37
  { 2457915.50, 68.812959 }, //11 June 2017, UT1-UTC= 0.3710412, Accumulated Leap Seconds=37
  { 2457916.50, 68.813367 }, //12 June 2017, UT1-UTC= 0.3706325, Accumulated Leap Seconds=37
  { 2457917.50, 68.813740 }, //13 June 2017, UT1-UTC= 0.3702599, Accumulated Leap Seconds=37
  { 2457918.50, 68.814135 }, //14 June 2017, UT1-UTC= 0.3698652, Accumulated Leap Seconds=37
  { 2457919.50, 68.814639 }, //15 June 2017, UT1-UTC= 0.3693614, Accumulated Leap Seconds=37
  { 2457920.50, 68.815255 }, //16 June 2017, UT1-UTC= 0.3687448, Accumulated Leap Seconds=37
  { 2457921.50, 68.815924 }, //17 June 2017, UT1-UTC= 0.3680758, Accumulated Leap Seconds=37
  { 2457922.50, 68.816716 }, //18 June 2017, UT1-UTC= 0.3672843, Accumulated Leap Seconds=37
  { 2457923.50, 68.817533 }, //19 June 2017, UT1-UTC= 0.3664668, Accumulated Leap Seconds=37
  { 2457924.50, 68.818327 }, //20 June 2017, UT1-UTC= 0.3656730, Accumulated Leap Seconds=37
  { 2457925.50, 68.819183 }, //21 June 2017, UT1-UTC= 0.3648167, Accumulated Leap Seconds=37
  { 2457926.50, 68.819896 }, //22 June 2017, UT1-UTC= 0.3641044, Accumulated Leap Seconds=37
  { 2457927.50, 68.820462 }, //23 June 2017, UT1-UTC= 0.3635383, Accumulated Leap Seconds=37
  { 2457928.50, 68.820892 }, //24 June 2017, UT1-UTC= 0.3631081, Accumulated Leap Seconds=37
  { 2457929.50, 68.821293 }, //25 June 2017, UT1-UTC= 0.3627069, Accumulated Leap Seconds=37
  { 2457930.50, 68.821713 }, //26 June 2017, UT1-UTC= 0.3622868, Accumulated Leap Seconds=37
  { 2457931.50, 68.822198 }, //27 June 2017, UT1-UTC= 0.3618024, Accumulated Leap Seconds=37
  { 2457932.50, 68.822764 }, //28 June 2017, UT1-UTC= 0.3612356, Accumulated Leap Seconds=37
  { 2457933.50, 68.823357 }, //29 June 2017, UT1-UTC= 0.3606428, Accumulated Leap Seconds=37
  { 2457934.50, 68.823927 }, //30 June 2017, UT1-UTC= 0.3600727, Accumulated Leap Seconds=37
  { 2457935.50, 68.824484 }, //01 July 2017, UT1-UTC= 0.3595162, Accumulated Leap Seconds=37
  { 2457936.50, 68.825069 }, //02 July 2017, UT1-UTC= 0.3589314, Accumulated Leap Seconds=37
  { 2457937.50, 68.825574 }, //03 July 2017, UT1-UTC= 0.3584260, Accumulated Leap Seconds=37
  { 2457938.50, 68.825907 }, //04 July 2017, UT1-UTC= 0.3580926, Accumulated Leap Seconds=37
  { 2457939.50, 68.826149 }, //05 July 2017, UT1-UTC= 0.3578507, Accumulated Leap Seconds=37
  { 2457940.50, 68.826320 }, //06 July 2017, UT1-UTC= 0.3576795, Accumulated Leap Seconds=37
  { 2457941.50, 68.826411 }, //07 July 2017, UT1-UTC= 0.3575894, Accumulated Leap Seconds=37
  { 2457942.50, 68.826493 }, //08 July 2017, UT1-UTC= 0.3575071, Accumulated Leap Seconds=37
  { 2457943.50, 68.826575 }, //09 July 2017, UT1-UTC= 0.3574252, Accumulated Leap Seconds=37
  { 2457944.50, 68.826667 }, //10 July 2017, UT1-UTC= 0.3573335, Accumulated Leap Seconds=37
  { 2457945.50, 68.826847 }, //11 July 2017, UT1-UTC= 0.3571525, Accumulated Leap Seconds=37
  { 2457946.50, 68.827202 }, //12 July 2017, UT1-UTC= 0.3567985, Accumulated Leap Seconds=37
  { 2457947.50, 68.827660 }, //13 July 2017, UT1-UTC= 0.3563397, Accumulated Leap Seconds=37
  { 2457948.50, 68.828225 }, //14 July 2017, UT1-UTC= 0.3557750, Accumulated Leap Seconds=37
  { 2457949.50, 68.828867 }, //15 July 2017, UT1-UTC= 0.3551328, Accumulated Leap Seconds=37
  { 2457950.50, 68.829489 }, //16 July 2017, UT1-UTC= 0.3545113, Accumulated Leap Seconds=37
  { 2457951.50, 68.830066 }, //17 July 2017, UT1-UTC= 0.3539339, Accumulated Leap Seconds=37
  { 2457952.50, 68.830613 }, //18 July 2017, UT1-UTC= 0.3533869, Accumulated Leap Seconds=37
  { 2457953.50, 68.831130 }, //19 July 2017, UT1-UTC= 0.3528699, Accumulated Leap Seconds=37
  { 2457954.50, 68.831524 }, //20 July 2017, UT1-UTC= 0.3524756, Accumulated Leap Seconds=37
  { 2457955.50, 68.831902 }, //21 July 2017, UT1-UTC= 0.3520979, Accumulated Leap Seconds=37
  { 2457956.50, 68.832215 }, //22 July 2017, UT1-UTC= 0.3517853, Accumulated Leap Seconds=37
  { 2457957.50, 68.832494 }, //23 July 2017, UT1-UTC= 0.3515060, Accumulated Leap Seconds=37
  { 2457958.50, 68.832810 }, //24 July 2017, UT1-UTC= 0.3511897, Accumulated Leap Seconds=37
  { 2457959.50, 68.833223 }, //25 July 2017, UT1-UTC= 0.3507771, Accumulated Leap Seconds=37
  { 2457960.50, 68.833773 }, //26 July 2017, UT1-UTC= 0.3502269, Accumulated Leap Seconds=37
  { 2457961.50, 68.834445 }, //27 July 2017, UT1-UTC= 0.3495550, Accumulated Leap Seconds=37
  { 2457962.50, 68.835168 }, //28 July 2017, UT1-UTC= 0.3488315, Accumulated Leap Seconds=37
  { 2457963.50, 68.835839 }, //29 July 2017, UT1-UTC= 0.3481606, Accumulated Leap Seconds=37
  { 2457964.50, 68.836470 }, //30 July 2017, UT1-UTC= 0.3475298, Accumulated Leap Seconds=37
  { 2457965.50, 68.836981 }, //31 July 2017, UT1-UTC= 0.3470191, Accumulated Leap Seconds=37
  { 2457966.50, 68.837343 }, //01 August 2017, UT1-UTC= 0.3466573, Accumulated Leap Seconds=37
  { 2457967.50, 68.837582 }, //02 August 2017, UT1-UTC= 0.3464176, Accumulated Leap Seconds=37
  { 2457968.50, 68.837721 }, //03 August 2017, UT1-UTC= 0.3462795, Accumulated Leap Seconds=37
  { 2457969.50, 68.837793 }, //04 August 2017, UT1-UTC= 0.3462068, Accumulated Leap Seconds=37
  { 2457970.50, 68.837869 }, //05 August 2017, UT1-UTC= 0.3461313, Accumulated Leap Seconds=37
  { 2457971.50, 68.838012 }, //06 August 2017, UT1-UTC= 0.3459884, Accumulated Leap Seconds=37
  { 2457972.50, 68.838146 }, //07 August 2017, UT1-UTC= 0.3458541, Accumulated Leap Seconds=37
  { 2457973.50, 68.838366 }, //08 August 2017, UT1-UTC= 0.3456339, Accumulated Leap Seconds=37
  { 2457974.50, 68.838709 }, //09 August 2017, UT1-UTC= 0.3452908, Accumulated Leap Seconds=37
  { 2457975.50, 68.839188 }, //10 August 2017, UT1-UTC= 0.3448115, Accumulated Leap Seconds=37
  { 2457976.50, 68.839819 }, //11 August 2017, UT1-UTC= 0.3441808, Accumulated Leap Seconds=37
  { 2457977.50, 68.840514 }, //12 August 2017, UT1-UTC= 0.3434856, Accumulated Leap Seconds=37
  { 2457978.50, 68.841189 }, //13 August 2017, UT1-UTC= 0.3428105, Accumulated Leap Seconds=37
  { 2457979.50, 68.841817 }, //14 August 2017, UT1-UTC= 0.3421832, Accumulated Leap Seconds=37
  { 2457980.50, 68.842343 }, //15 August 2017, UT1-UTC= 0.3416571, Accumulated Leap Seconds=37
  { 2457981.50, 68.842763 }, //16 August 2017, UT1-UTC= 0.3412370, Accumulated Leap Seconds=37
  { 2457982.50, 68.843064 }, //17 August 2017, UT1-UTC= 0.3409361, Accumulated Leap Seconds=37
  { 2457983.50, 68.843259 }, //18 August 2017, UT1-UTC= 0.3407410, Accumulated Leap Seconds=37
  { 2457984.50, 68.843481 }, //19 August 2017, UT1-UTC= 0.3405194, Accumulated Leap Seconds=37
  { 2457985.50, 68.843696 }, //20 August 2017, UT1-UTC= 0.3403042, Accumulated Leap Seconds=37
  { 2457986.50, 68.844032 }, //21 August 2017, UT1-UTC= 0.3399683, Accumulated Leap Seconds=37
  { 2457987.50, 68.844509 }, //22 August 2017, UT1-UTC= 0.3394911, Accumulated Leap Seconds=37
  { 2457988.50, 68.845011 }, //23 August 2017, UT1-UTC= 0.3389885, Accumulated Leap Seconds=37
  { 2457989.50, 68.845517 }, //24 August 2017, UT1-UTC= 0.3384825, Accumulated Leap Seconds=37
  { 2457990.50, 68.846092 }, //25 August 2017, UT1-UTC= 0.3379082, Accumulated Leap Seconds=37
  { 2457991.50, 68.846569 }, //26 August 2017, UT1-UTC= 0.3374314, Accumulated Leap Seconds=37
  { 2457992.50, 68.846934 }, //27 August 2017, UT1-UTC= 0.3370656, Accumulated Leap Seconds=37
  { 2457993.50, 68.847199 }, //28 August 2017, UT1-UTC= 0.3368008, Accumulated Leap Seconds=37
  { 2457994.50, 68.847403 }, //29 August 2017, UT1-UTC= 0.3365967, Accumulated Leap Seconds=37
  { 2457995.50, 68.847535 }, //30 August 2017, UT1-UTC= 0.3364652, Accumulated Leap Seconds=37
  { 2457996.50, 68.847593 }, //31 August 2017, UT1-UTC= 0.3364071, Accumulated Leap Seconds=37
  { 2457997.50, 68.847693 }, //01 September 2017, UT1-UTC= 0.3363070, Accumulated Leap Seconds=37
  { 2457998.50, 68.847817 }, //02 September 2017, UT1-UTC= 0.3361832, Accumulated Leap Seconds=37
  { 2457999.50, 68.847980 }, //03 September 2017, UT1-UTC= 0.3360197, Accumulated Leap Seconds=37
  { 2458000.50, 68.848208 }, //04 September 2017, UT1-UTC= 0.3357915, Accumulated Leap Seconds=37
  { 2458001.50, 68.848596 }, //05 September 2017, UT1-UTC= 0.3354041, Accumulated Leap Seconds=37
  { 2458002.50, 68.849182 }, //06 September 2017, UT1-UTC= 0.3348184, Accumulated Leap Seconds=37
  { 2458003.50, 68.849945 }, //07 September 2017, UT1-UTC= 0.3340548, Accumulated Leap Seconds=37
  { 2458004.50, 68.850826 }, //08 September 2017, UT1-UTC= 0.3331740, Accumulated Leap Seconds=37
  { 2458005.50, 68.851774 }, //09 September 2017, UT1-UTC= 0.3322264, Accumulated Leap Seconds=37
  { 2458006.50, 68.852732 }, //10 September 2017, UT1-UTC= 0.3312681, Accumulated Leap Seconds=37
  { 2458007.50, 68.853624 }, //11 September 2017, UT1-UTC= 0.3303762, Accumulated Leap Seconds=37
  { 2458008.50, 68.854430 }, //12 September 2017, UT1-UTC= 0.3295699, Accumulated Leap Seconds=37
  { 2458009.50, 68.855107 }, //13 September 2017, UT1-UTC= 0.3288928, Accumulated Leap Seconds=37
  { 2458010.50, 68.855706 }, //14 September 2017, UT1-UTC= 0.3282938, Accumulated Leap Seconds=37
  { 2458011.50, 68.856287 }, //15 September 2017, UT1-UTC= 0.3277127, Accumulated Leap Seconds=37
  { 2458012.50, 68.856875 }, //16 September 2017, UT1-UTC= 0.3271251, Accumulated Leap Seconds=37
  { 2458013.50, 68.857630 }, //17 September 2017, UT1-UTC= 0.3263704, Accumulated Leap Seconds=37
  { 2458014.50, 68.858548 }, //18 September 2017, UT1-UTC= 0.3254522, Accumulated Leap Seconds=37
  { 2458015.50, 68.859618 }, //19 September 2017, UT1-UTC= 0.3243822, Accumulated Leap Seconds=37
  { 2458016.50, 68.860730 }, //20 September 2017, UT1-UTC= 0.3232700, Accumulated Leap Seconds=37
  { 2458017.50, 68.861822 }, //21 September 2017, UT1-UTC= 0.3221784, Accumulated Leap Seconds=37
  { 2458018.50, 68.862893 }, //22 September 2017, UT1-UTC= 0.3211066, Accumulated Leap Seconds=37
  { 2458019.50, 68.863852 }, //23 September 2017, UT1-UTC= 0.3201480, Accumulated Leap Seconds=37
  { 2458020.50, 68.864705 }, //24 September 2017, UT1-UTC= 0.3192947, Accumulated Leap Seconds=37
  { 2458021.50, 68.865421 }, //25 September 2017, UT1-UTC= 0.3185789, Accumulated Leap Seconds=37
  { 2458022.50, 68.865994 }, //26 September 2017, UT1-UTC= 0.3180063, Accumulated Leap Seconds=37
  { 2458023.50, 68.866520 }, //27 September 2017, UT1-UTC= 0.3174801, Accumulated Leap Seconds=37
  { 2458024.50, 68.867028 }, //28 September 2017, UT1-UTC= 0.3169719, Accumulated Leap Seconds=37
  { 2458025.50, 68.867612 }, //29 September 2017, UT1-UTC= 0.3163884, Accumulated Leap Seconds=37
  { 2458026.50, 68.868212 }, //30 September 2017, UT1-UTC= 0.3157875, Accumulated Leap Seconds=37
  { 2458027.50, 68.868857 }, //01 October 2017, UT1-UTC= 0.3151433, Accumulated Leap Seconds=37
  { 2458028.50, 68.869496 }, //02 October 2017, UT1-UTC= 0.3145042, Accumulated Leap Seconds=37
  { 2458029.50, 68.870245 }, //03 October 2017, UT1-UTC= 0.3137553, Accumulated Leap Seconds=37
  { 2458030.50, 68.871120 }, //04 October 2017, UT1-UTC= 0.3128801, Accumulated Leap Seconds=37
  { 2458031.50, 68.872153 }, //05 October 2017, UT1-UTC= 0.3118472, Accumulated Leap Seconds=37
  { 2458032.50, 68.873297 }, //06 October 2017, UT1-UTC= 0.3107030, Accumulated Leap Seconds=37
  { 2458033.50, 68.874454 }, //07 October 2017, UT1-UTC= 0.3095458, Accumulated Leap Seconds=37
  { 2458034.50, 68.875597 }, //08 October 2017, UT1-UTC= 0.3084027, Accumulated Leap Seconds=37
  { 2458035.50, 68.876633 }, //09 October 2017, UT1-UTC= 0.3073670, Accumulated Leap Seconds=37
  { 2458036.50, 68.877532 }, //10 October 2017, UT1-UTC= 0.3064679, Accumulated Leap Seconds=37
  { 2458037.50, 68.878313 }, //11 October 2017, UT1-UTC= 0.3056873, Accumulated Leap Seconds=37
  { 2458038.50, 68.879059 }, //12 October 2017, UT1-UTC= 0.3049414, Accumulated Leap Seconds=37
  { 2458039.50, 68.879822 }, //13 October 2017, UT1-UTC= 0.3041776, Accumulated Leap Seconds=37
  { 2458040.50, 68.880696 }, //14 October 2017, UT1-UTC= 0.3033042, Accumulated Leap Seconds=37
  { 2458041.50, 68.881669 }, //15 October 2017, UT1-UTC= 0.3023310, Accumulated Leap Seconds=37
  { 2458042.50, 68.882781 }, //16 October 2017, UT1-UTC= 0.3012194, Accumulated Leap Seconds=37
  { 2458043.50, 68.884059 }, //17 October 2017, UT1-UTC= 0.2999410, Accumulated Leap Seconds=37
  { 2458044.50, 68.885421 }, //18 October 2017, UT1-UTC= 0.2985785, Accumulated Leap Seconds=37
  { 2458045.50, 68.886810 }, //19 October 2017, UT1-UTC= 0.2971898, Accumulated Leap Seconds=37
  { 2458046.50, 68.888184 }, //20 October 2017, UT1-UTC= 0.2958162, Accumulated Leap Seconds=37
  { 2458047.50, 68.889454 }, //21 October 2017, UT1-UTC= 0.2945455, Accumulated Leap Seconds=37
  { 2458048.50, 68.890622 }, //22 October 2017, UT1-UTC= 0.2933783, Accumulated Leap Seconds=37
  { 2458049.50, 68.891731 }, //23 October 2017, UT1-UTC= 0.2922689, Accumulated Leap Seconds=37
  { 2458050.50, 68.892701 }, //24 October 2017, UT1-UTC= 0.2912988, Accumulated Leap Seconds=37
  { 2458051.50, 68.893554 }, //25 October 2017, UT1-UTC= 0.2904457, Accumulated Leap Seconds=37
  { 2458052.50, 68.894378 }, //26 October 2017, UT1-UTC= 0.2896225, Accumulated Leap Seconds=37
  { 2458053.50, 68.895205 }, //27 October 2017, UT1-UTC= 0.2887951, Accumulated Leap Seconds=37
  { 2458054.50, 68.896058 }, //28 October 2017, UT1-UTC= 0.2879423, Accumulated Leap Seconds=37
  { 2458055.50, 68.897006 }, //29 October 2017, UT1-UTC= 0.2869943, Accumulated Leap Seconds=37
  { 2458056.50, 68.898078 }, //30 October 2017, UT1-UTC= 0.2859215, Accumulated Leap Seconds=37
  { 2458057.50, 68.899244 }, //31 October 2017, UT1-UTC= 0.2847563, Accumulated Leap Seconds=37
  { 2458058.50, 68.900581 }, //01 November 2017, UT1-UTC= 0.2834186, Accumulated Leap Seconds=37
  { 2458059.50, 68.902026 }, //02 November 2017, UT1-UTC= 0.2819741, Accumulated Leap Seconds=37
  { 2458060.50, 68.903557 }, //03 November 2017, UT1-UTC= 0.2804434, Accumulated Leap Seconds=37
  { 2458061.50, 68.905108 }, //04 November 2017, UT1-UTC= 0.2788916, Accumulated Leap Seconds=37
  { 2458062.50, 68.906572 }, //05 November 2017, UT1-UTC= 0.2774282, Accumulated Leap Seconds=37
  { 2458063.50, 68.907886 }, //06 November 2017, UT1-UTC= 0.2761141, Accumulated Leap Seconds=37
  { 2458064.50, 68.909041 }, //07 November 2017, UT1-UTC= 0.2749591, Accumulated Leap Seconds=37
  { 2458065.50, 68.910083 }, //08 November 2017, UT1-UTC= 0.2739175, Accumulated Leap Seconds=37
  { 2458066.50, 68.911145 }, //09 November 2017, UT1-UTC= 0.2728553, Accumulated Leap Seconds=37
  { 2458067.50, 68.912245 }, //10 November 2017, UT1-UTC= 0.2717553, Accumulated Leap Seconds=37
  { 2458068.50, 68.913435 }, //11 November 2017, UT1-UTC= 0.2705650, Accumulated Leap Seconds=37
  { 2458069.50, 68.914751 }, //12 November 2017, UT1-UTC= 0.2692491, Accumulated Leap Seconds=37
  { 2458070.50, 68.916183 }, //13 November 2017, UT1-UTC= 0.2678169, Accumulated Leap Seconds=37
  { 2458071.50, 68.917687 }, //14 November 2017, UT1-UTC= 0.2663129, Accumulated Leap Seconds=37
  { 2458072.50, 68.919178 }, //15 November 2017, UT1-UTC= 0.2648225, Accumulated Leap Seconds=37
  { 2458073.50, 68.920585 }, //16 November 2017, UT1-UTC= 0.2634146, Accumulated Leap Seconds=37
  { 2458074.50, 68.921854 }, //17 November 2017, UT1-UTC= 0.2621456, Accumulated Leap Seconds=37
  { 2458075.50, 68.923007 }, //18 November 2017, UT1-UTC= 0.2609930, Accumulated Leap Seconds=37
  { 2458076.50, 68.924045 }, //19 November 2017, UT1-UTC= 0.2599548, Accumulated Leap Seconds=37
  { 2458077.50, 68.924905 }, //20 November 2017, UT1-UTC= 0.2590952, Accumulated Leap Seconds=37
  { 2458078.50, 68.925597 }, //21 November 2017, UT1-UTC= 0.2584030, Accumulated Leap Seconds=37
  { 2458079.50, 68.926228 }, //22 November 2017, UT1-UTC= 0.2577722, Accumulated Leap Seconds=37
  { 2458080.50, 68.926872 }, //23 November 2017, UT1-UTC= 0.2571279, Accumulated Leap Seconds=37
  { 2458081.50, 68.927557 }, //24 November 2017, UT1-UTC= 0.2564427, Accumulated Leap Seconds=37
  { 2458082.50, 68.928282 }, //25 November 2017, UT1-UTC= 0.2557184, Accumulated Leap Seconds=37
  { 2458083.50, 68.929059 }, //26 November 2017, UT1-UTC= 0.2549406, Accumulated Leap Seconds=37
  { 2458084.50, 68.930007 }, //27 November 2017, UT1-UTC= 0.2539932, Accumulated Leap Seconds=37
  { 2458085.50, 68.931165 }, //28 November 2017, UT1-UTC= 0.2528346, Accumulated Leap Seconds=37
  { 2458086.50, 68.932520 }, //29 November 2017, UT1-UTC= 0.2514797, Accumulated Leap Seconds=37
  { 2458087.50, 68.933985 }, //30 November 2017, UT1-UTC= 0.2500151, Accumulated Leap Seconds=37
  { 2458088.50, 68.935500 }, //01 December 2017, UT1-UTC= 0.2485001, Accumulated Leap Seconds=37
  { 2458089.50, 68.937030 }, //02 December 2017, UT1-UTC= 0.2469699, Accumulated Leap Seconds=37
  { 2458090.50, 68.938438 }, //03 December 2017, UT1-UTC= 0.2455622, Accumulated Leap Seconds=37
  { 2458091.50, 68.939722 }, //04 December 2017, UT1-UTC= 0.2442778, Accumulated Leap Seconds=37
  { 2458092.50, 68.940928 }, //05 December 2017, UT1-UTC= 0.2430722, Accumulated Leap Seconds=37
  { 2458093.50, 68.942121 }, //06 December 2017, UT1-UTC= 0.2418790, Accumulated Leap Seconds=37
  { 2458094.50, 68.943344 }, //07 December 2017, UT1-UTC= 0.2406564, Accumulated Leap Seconds=37
  { 2458095.50, 68.944612 }, //08 December 2017, UT1-UTC= 0.2393884, Accumulated Leap Seconds=37
  { 2458096.50, 68.945951 }, //09 December 2017, UT1-UTC= 0.2380494, Accumulated Leap Seconds=37
  { 2458097.50, 68.947386 }, //10 December 2017, UT1-UTC= 0.2366137, Accumulated Leap Seconds=37
  { 2458098.50, 68.948849 }, //11 December 2017, UT1-UTC= 0.2351508, Accumulated Leap Seconds=37
  { 2458099.50, 68.950273 }, //12 December 2017, UT1-UTC= 0.2337272, Accumulated Leap Seconds=37
  { 2458100.50, 68.951614 }, //13 December 2017, UT1-UTC= 0.2323861, Accumulated Leap Seconds=37
  { 2458101.50, 68.952816 }, //14 December 2017, UT1-UTC= 0.2311842, Accumulated Leap Seconds=37
  { 2458102.50, 68.953864 }, //15 December 2017, UT1-UTC= 0.2301356, Accumulated Leap Seconds=37
  { 2458103.50, 68.954709 }, //16 December 2017, UT1-UTC= 0.2292912, Accumulated Leap Seconds=37
  { 2458104.50, 68.955432 }, //17 December 2017, UT1-UTC= 0.2285676, Accumulated Leap Seconds=37
  { 2458105.50, 68.956059 }, //18 December 2017, UT1-UTC= 0.2279407, Accumulated Leap Seconds=37
  { 2458106.50, 68.956584 }, //19 December 2017, UT1-UTC= 0.2274163, Accumulated Leap Seconds=37
  { 2458107.50, 68.957087 }, //20 December 2017, UT1-UTC= 0.2269129, Accumulated Leap Seconds=37
  { 2458108.50, 68.957637 }, //21 December 2017, UT1-UTC= 0.2263633, Accumulated Leap Seconds=37
  { 2458109.50, 68.958177 }, //22 December 2017, UT1-UTC= 0.2258234, Accumulated Leap Seconds=37
  { 2458110.50, 68.958770 }, //23 December 2017, UT1-UTC= 0.2252297, Accumulated Leap Seconds=37
  { 2458111.50, 68.959513 }, //24 December 2017, UT1-UTC= 0.2244866, Accumulated Leap Seconds=37
  { 2458112.50, 68.960378 }, //25 December 2017, UT1-UTC= 0.2236216, Accumulated Leap Seconds=37
  { 2458113.50, 68.961335 }, //26 December 2017, UT1-UTC= 0.2226648, Accumulated Leap Seconds=37
  { 2458114.50, 68.962398 }, //27 December 2017, UT1-UTC= 0.2216024, Accumulated Leap Seconds=37
  { 2458115.50, 68.963543 }, //28 December 2017, UT1-UTC= 0.2204568, Accumulated Leap Seconds=37
  { 2458116.50, 68.964692 }, //29 December 2017, UT1-UTC= 0.2193082, Accumulated Leap Seconds=37
  { 2458117.50, 68.965756 }, //30 December 2017, UT1-UTC= 0.2182443, Accumulated Leap Seconds=37
  { 2458118.50, 68.966760 }, //31 December 2017, UT1-UTC= 0.2172403, Accumulated Leap Seconds=37
  { 2458119.50, 68.967642 }, //01 January 2018, UT1-UTC= 0.2163577, Accumulated Leap Seconds=37
  { 2458120.50, 68.968413 }, //02 January 2018, UT1-UTC= 0.2155874, Accumulated Leap Seconds=37
  { 2458121.50, 68.969139 }, //03 January 2018, UT1-UTC= 0.2148609, Accumulated Leap Seconds=37
  { 2458122.50, 68.969947 }, //04 January 2018, UT1-UTC= 0.2140527, Accumulated Leap Seconds=37
  { 2458123.50, 68.970812 }, //05 January 2018, UT1-UTC= 0.2131882, Accumulated Leap Seconds=37
  { 2458124.50, 68.971710 }, //06 January 2018, UT1-UTC= 0.2122898, Accumulated Leap Seconds=37
  { 2458125.50, 68.972574 }, //07 January 2018, UT1-UTC= 0.2114257, Accumulated Leap Seconds=37
  { 2458126.50, 68.973405 }, //08 January 2018, UT1-UTC= 0.2105948, Accumulated Leap Seconds=37
  { 2458127.50, 68.974090 }, //09 January 2018, UT1-UTC= 0.2099097, Accumulated Leap Seconds=37
  { 2458128.50, 68.974653 }, //10 January 2018, UT1-UTC= 0.2093470, Accumulated Leap Seconds=37
  { 2458129.50, 68.975094 }, //11 January 2018, UT1-UTC= 0.2089056, Accumulated Leap Seconds=37
  { 2458130.50, 68.975485 }, //12 January 2018, UT1-UTC= 0.2085151, Accumulated Leap Seconds=37
  { 2458131.50, 68.975720 }, //13 January 2018, UT1-UTC= 0.2082798, Accumulated Leap Seconds=37
  { 2458132.50, 68.975825 }, //14 January 2018, UT1-UTC= 0.2081755, Accumulated Leap Seconds=37
  { 2458133.50, 68.975902 }, //15 January 2018, UT1-UTC= 0.2080984, Accumulated Leap Seconds=37
  { 2458134.50, 68.975923 }, //16 January 2018, UT1-UTC= 0.2080772, Accumulated Leap Seconds=37
  { 2458135.50, 68.976025 }, //17 January 2018, UT1-UTC= 0.2079746, Accumulated Leap Seconds=37
  { 2458136.50, 68.976168 }, //18 January 2018, UT1-UTC= 0.2078316, Accumulated Leap Seconds=37
  { 2458137.50, 68.976353 }, //19 January 2018, UT1-UTC= 0.2076469, Accumulated Leap Seconds=37
  { 2458138.50, 68.976688 }, //20 January 2018, UT1-UTC= 0.2073123, Accumulated Leap Seconds=37
  { 2458139.50, 68.977206 }, //21 January 2018, UT1-UTC= 0.2067940, Accumulated Leap Seconds=37
  { 2458140.50, 68.977884 }, //22 January 2018, UT1-UTC= 0.2061160, Accumulated Leap Seconds=37
  { 2458141.50, 68.978686 }, //23 January 2018, UT1-UTC= 0.2053143, Accumulated Leap Seconds=37
  { 2458142.50, 68.979603 }, //24 January 2018, UT1-UTC= 0.2043973, Accumulated Leap Seconds=37
  { 2458143.50, 68.980560 }, //25 January 2018, UT1-UTC= 0.2034402, Accumulated Leap Seconds=37
  { 2458144.50, 68.981568 }, //26 January 2018, UT1-UTC= 0.2024318, Accumulated Leap Seconds=37
  { 2458145.50, 68.982530 }, //27 January 2018, UT1-UTC= 0.2014701, Accumulated Leap Seconds=37
  { 2458146.50, 68.983422 }, //28 January 2018, UT1-UTC= 0.2005783, Accumulated Leap Seconds=37
  { 2458147.50, 68.984313 }, //29 January 2018, UT1-UTC= 0.1996865, Accumulated Leap Seconds=37
  { 2458148.50, 68.985243 }, //30 January 2018, UT1-UTC= 0.1987569, Accumulated Leap Seconds=37
  { 2458149.50, 68.986326 }, //31 January 2018, UT1-UTC= 0.1976740, Accumulated Leap Seconds=37
  { 2458150.50, 68.987535 }, //01 February 2018, UT1-UTC= 0.1964646, Accumulated Leap Seconds=37
  { 2458151.50, 68.988918 }, //02 February 2018, UT1-UTC= 0.1950819, Accumulated Leap Seconds=37
  { 2458152.50, 68.990502 }, //03 February 2018, UT1-UTC= 0.1934980, Accumulated Leap Seconds=37
  { 2458153.50, 68.992183 }, //04 February 2018, UT1-UTC= 0.1918172, Accumulated Leap Seconds=37
  { 2458154.50, 68.993858 }, //05 February 2018, UT1-UTC= 0.1901421, Accumulated Leap Seconds=37
  { 2458155.50, 68.995434 }, //06 February 2018, UT1-UTC= 0.1885659, Accumulated Leap Seconds=37
  { 2458156.50, 68.996846 }, //07 February 2018, UT1-UTC= 0.1871541, Accumulated Leap Seconds=37
  { 2458157.50, 68.998106 }, //08 February 2018, UT1-UTC= 0.1858940, Accumulated Leap Seconds=37
  { 2458158.50, 68.999198 }, //09 February 2018, UT1-UTC= 0.1848016, Accumulated Leap Seconds=37
  { 2458159.50, 69.000087 }, //10 February 2018, UT1-UTC= 0.1839131, Accumulated Leap Seconds=37
  { 2458160.50, 69.000860 }, //11 February 2018, UT1-UTC= 0.1831397, Accumulated Leap Seconds=37
  { 2458161.50, 69.001614 }, //12 February 2018, UT1-UTC= 0.1823862, Accumulated Leap Seconds=37
  { 2458162.50, 69.002368 }, //13 February 2018, UT1-UTC= 0.1816316, Accumulated Leap Seconds=37
  { 2458163.50, 69.003141 }, //14 February 2018, UT1-UTC= 0.1808591, Accumulated Leap Seconds=37
  { 2458164.50, 69.003964 }, //15 February 2018, UT1-UTC= 0.1800361, Accumulated Leap Seconds=37
  { 2458165.50, 69.004848 }, //16 February 2018, UT1-UTC= 0.1791523, Accumulated Leap Seconds=37
  { 2458166.50, 69.005821 }, //17 February 2018, UT1-UTC= 0.1781794, Accumulated Leap Seconds=37
  { 2458167.50, 69.006882 }, //18 February 2018, UT1-UTC= 0.1771176, Accumulated Leap Seconds=37
  { 2458168.50, 69.008023 }, //19 February 2018, UT1-UTC= 0.1759766, Accumulated Leap Seconds=37
  { 2458169.50, 69.009194 }, //20 February 2018, UT1-UTC= 0.1748059, Accumulated Leap Seconds=37
  { 2458170.50, 69.010340 }, //21 February 2018, UT1-UTC= 0.1736605, Accumulated Leap Seconds=37
  { 2458171.50, 69.011436 }, //22 February 2018, UT1-UTC= 0.1725641, Accumulated Leap Seconds=37
  { 2458172.50, 69.012460 }, //23 February 2018, UT1-UTC= 0.1715396, Accumulated Leap Seconds=37
  { 2458173.50, 69.013364 }, //24 February 2018, UT1-UTC= 0.1706360, Accumulated Leap Seconds=37
  { 2458174.50, 69.014173 }, //25 February 2018, UT1-UTC= 0.1698271, Accumulated Leap Seconds=37
  { 2458175.50, 69.014951 }, //26 February 2018, UT1-UTC= 0.1690487, Accumulated Leap Seconds=37
  { 2458176.50, 69.015752 }, //27 February 2018, UT1-UTC= 0.1682480, Accumulated Leap Seconds=37
  { 2458177.50, 69.016597 }, //28 February 2018, UT1-UTC= 0.1674030, Accumulated Leap Seconds=37
  { 2458178.50, 69.017553 }, //01 March 2018, UT1-UTC= 0.1664473, Accumulated Leap Seconds=37
  { 2458179.50, 69.018708 }, //02 March 2018, UT1-UTC= 0.1652916, Accumulated Leap Seconds=37
  { 2458180.50, 69.019973 }, //03 March 2018, UT1-UTC= 0.1640270, Accumulated Leap Seconds=37
  { 2458181.50, 69.021213 }, //04 March 2018, UT1-UTC= 0.1627874, Accumulated Leap Seconds=37
  { 2458182.50, 69.022421 }, //05 March 2018, UT1-UTC= 0.1615785, Accumulated Leap Seconds=37
  { 2458183.50, 69.023516 }, //06 March 2018, UT1-UTC= 0.1604839, Accumulated Leap Seconds=37
  { 2458184.50, 69.024459 }, //07 March 2018, UT1-UTC= 0.1595412, Accumulated Leap Seconds=37
  { 2458185.50, 69.025363 }, //08 March 2018, UT1-UTC= 0.1586365, Accumulated Leap Seconds=37
  { 2458186.50, 69.026175 }, //09 March 2018, UT1-UTC= 0.1578251, Accumulated Leap Seconds=37
  { 2458187.50, 69.026889 }, //10 March 2018, UT1-UTC= 0.1571114, Accumulated Leap Seconds=37
  { 2458188.50, 69.027573 }, //11 March 2018, UT1-UTC= 0.1564273, Accumulated Leap Seconds=37
  { 2458189.50, 69.028270 }, //12 March 2018, UT1-UTC= 0.1557302, Accumulated Leap Seconds=37
  { 2458190.50, 69.029028 }, //13 March 2018, UT1-UTC= 0.1549723, Accumulated Leap Seconds=37
  { 2458191.50, 69.029855 }, //14 March 2018, UT1-UTC= 0.1541452, Accumulated Leap Seconds=37
  { 2458192.50, 69.030758 }, //15 March 2018, UT1-UTC= 0.1532421, Accumulated Leap Seconds=37
  { 2458193.50, 69.031755 }, //16 March 2018, UT1-UTC= 0.1522448, Accumulated Leap Seconds=37
  { 2458194.50, 69.032834 }, //17 March 2018, UT1-UTC= 0.1511656, Accumulated Leap Seconds=37
  { 2458195.50, 69.033973 }, //18 March 2018, UT1-UTC= 0.1500269, Accumulated Leap Seconds=37
  { 2458196.50, 69.035171 }, //19 March 2018, UT1-UTC= 0.1488290, Accumulated Leap Seconds=37
  { 2458197.50, 69.036413 }, //20 March 2018, UT1-UTC= 0.1475874, Accumulated Leap Seconds=37
  { 2458198.50, 69.037652 }, //21 March 2018, UT1-UTC= 0.1463478, Accumulated Leap Seconds=37
  { 2458199.50, 69.038843 }, //22 March 2018, UT1-UTC= 0.1451572, Accumulated Leap Seconds=37
  { 2458200.50, 69.039935 }, //23 March 2018, UT1-UTC= 0.1440653, Accumulated Leap Seconds=37
  { 2458201.50, 69.040898 }, //24 March 2018, UT1-UTC= 0.1431017, Accumulated Leap Seconds=37
  { 2458202.50, 69.041767 }, //25 March 2018, UT1-UTC= 0.1422333, Accumulated Leap Seconds=37
  { 2458203.50, 69.042619 }, //26 March 2018, UT1-UTC= 0.1413808, Accumulated Leap Seconds=37
  { 2458204.50, 69.043582 }, //27 March 2018, UT1-UTC= 0.1404181, Accumulated Leap Seconds=37
  { 2458205.50, 69.044663 }, //28 March 2018, UT1-UTC= 0.1393369, Accumulated Leap Seconds=37
  { 2458206.50, 69.045850 }, //29 March 2018, UT1-UTC= 0.1381503, Accumulated Leap Seconds=37
  { 2458207.50, 69.047130 }, //30 March 2018, UT1-UTC= 0.1368700, Accumulated Leap Seconds=37
  { 2458208.50, 69.048520 }, //31 March 2018, UT1-UTC= 0.1354801, Accumulated Leap Seconds=37
  { 2458209.50, 69.049908 }, //01 April 2018, UT1-UTC= 0.1340919, Accumulated Leap Seconds=37
  { 2458210.50, 69.051236 }, //02 April 2018, UT1-UTC= 0.1327638, Accumulated Leap Seconds=37
  { 2458211.50, 69.052452 }, //03 April 2018, UT1-UTC= 0.1315479, Accumulated Leap Seconds=37
  { 2458212.50, 69.053507 }, //04 April 2018, UT1-UTC= 0.1304933, Accumulated Leap Seconds=37
  { 2458213.50, 69.054383 }, //05 April 2018, UT1-UTC= 0.1296169, Accumulated Leap Seconds=37
  { 2458214.50, 69.055133 }, //06 April 2018, UT1-UTC= 0.1288673, Accumulated Leap Seconds=37
  { 2458215.50, 69.055818 }, //07 April 2018, UT1-UTC= 0.1281820, Accumulated Leap Seconds=37
  { 2458216.50, 69.056474 }, //08 April 2018, UT1-UTC= 0.1275261, Accumulated Leap Seconds=37
  { 2458217.50, 69.057135 }, //09 April 2018, UT1-UTC= 0.1268650, Accumulated Leap Seconds=37
  { 2458218.50, 69.057816 }, //10 April 2018, UT1-UTC= 0.1261843, Accumulated Leap Seconds=37
  { 2458219.50, 69.058516 }, //11 April 2018, UT1-UTC= 0.1254843, Accumulated Leap Seconds=37
  { 2458220.50, 69.059311 }, //12 April 2018, UT1-UTC= 0.1246885, Accumulated Leap Seconds=37
  { 2458221.50, 69.060235 }, //13 April 2018, UT1-UTC= 0.1237648, Accumulated Leap Seconds=37
  { 2458222.50, 69.061272 }, //14 April 2018, UT1-UTC= 0.1227279, Accumulated Leap Seconds=37
  { 2458223.50, 69.062448 }, //15 April 2018, UT1-UTC= 0.1215516, Accumulated Leap Seconds=37
  { 2458224.50, 69.063778 }, //16 April 2018, UT1-UTC= 0.1202219, Accumulated Leap Seconds=37
  { 2458225.50, 69.065141 }, //17 April 2018, UT1-UTC= 0.1188591, Accumulated Leap Seconds=37
  { 2458226.50, 69.066398 }, //18 April 2018, UT1-UTC= 0.1176018, Accumulated Leap Seconds=37
  { 2458227.50, 69.067509 }, //19 April 2018, UT1-UTC= 0.1164911, Accumulated Leap Seconds=37
  { 2458228.50, 69.068540 }, //20 April 2018, UT1-UTC= 0.1154595, Accumulated Leap Seconds=37
  { 2458229.50, 69.069499 }, //21 April 2018, UT1-UTC= 0.1145009, Accumulated Leap Seconds=37
  { 2458230.50, 69.070458 }, //22 April 2018, UT1-UTC= 0.1135425, Accumulated Leap Seconds=37
  { 2458231.50, 69.071466 }, //23 April 2018, UT1-UTC= 0.1125337, Accumulated Leap Seconds=37
  { 2458232.50, 69.072602 }, //24 April 2018, UT1-UTC= 0.1113982, Accumulated Leap Seconds=37
  { 2458233.50, 69.073921 }, //25 April 2018, UT1-UTC= 0.1100794, Accumulated Leap Seconds=37
  { 2458234.50, 69.075347 }, //26 April 2018, UT1-UTC= 0.1086530, Accumulated Leap Seconds=37
  { 2458235.50, 69.076855 }, //27 April 2018, UT1-UTC= 0.1071446, Accumulated Leap Seconds=37
  { 2458236.50, 69.078382 }, //28 April 2018, UT1-UTC= 0.1056176, Accumulated Leap Seconds=37
  { 2458237.50, 69.079872 }, //29 April 2018, UT1-UTC= 0.1041281, Accumulated Leap Seconds=37
  { 2458238.50, 69.081182 }, //30 April 2018, UT1-UTC= 0.1028179, Accumulated Leap Seconds=37
  { 2458239.50, 69.082343 }, //01 May 2018, UT1-UTC= 0.1016567, Accumulated Leap Seconds=37
  { 2458240.50, 69.083377 }, //02 May 2018, UT1-UTC= 0.1006233, Accumulated Leap Seconds=37
  { 2458241.50, 69.084265 }, //03 May 2018, UT1-UTC= 0.0997347, Accumulated Leap Seconds=37
  { 2458242.50, 69.085023 }, //04 May 2018, UT1-UTC= 0.0989771, Accumulated Leap Seconds=37
  { 2458243.50, 69.085713 }, //05 May 2018, UT1-UTC= 0.0982874, Accumulated Leap Seconds=37
  { 2458244.50, 69.086382 }, //06 May 2018, UT1-UTC= 0.0976177, Accumulated Leap Seconds=37
  { 2458245.50, 69.087030 }, //07 May 2018, UT1-UTC= 0.0969698, Accumulated Leap Seconds=37
  { 2458246.50, 69.087705 }, //08 May 2018, UT1-UTC= 0.0962955, Accumulated Leap Seconds=37
  { 2458247.50, 69.088479 }, //09 May 2018, UT1-UTC= 0.0955205, Accumulated Leap Seconds=37
  { 2458248.50, 69.089367 }, //10 May 2018, UT1-UTC= 0.0946331, Accumulated Leap Seconds=37
  { 2458249.50, 69.090314 }, //11 May 2018, UT1-UTC= 0.0936859, Accumulated Leap Seconds=37
  { 2458250.50, 69.091343 }, //12 May 2018, UT1-UTC= 0.0926566, Accumulated Leap Seconds=37
  { 2458251.50, 69.092474 }, //13 May 2018, UT1-UTC= 0.0915264, Accumulated Leap Seconds=37
  { 2458252.50, 69.093573 }, //14 May 2018, UT1-UTC= 0.0904266, Accumulated Leap Seconds=37
  { 2458253.50, 69.094593 }, //15 May 2018, UT1-UTC= 0.0894075, Accumulated Leap Seconds=37
  { 2458254.50, 69.095463 }, //16 May 2018, UT1-UTC= 0.0885366, Accumulated Leap Seconds=37
  { 2458255.50, 69.096171 }, //17 May 2018, UT1-UTC= 0.0878291, Accumulated Leap Seconds=37
  { 2458256.50, 69.096802 }, //18 May 2018, UT1-UTC= 0.0871981, Accumulated Leap Seconds=37
  { 2458257.50, 69.097345 }, //19 May 2018, UT1-UTC= 0.0866552, Accumulated Leap Seconds=37
  { 2458258.50, 69.097936 }, //20 May 2018, UT1-UTC= 0.0860637, Accumulated Leap Seconds=37
  { 2458259.50, 69.098575 }, //21 May 2018, UT1-UTC= 0.0854250, Accumulated Leap Seconds=37
  { 2458260.50, 69.099340 }, //22 May 2018, UT1-UTC= 0.0846600, Accumulated Leap Seconds=37
  { 2458261.50, 69.100272 }, //23 May 2018, UT1-UTC= 0.0837281, Accumulated Leap Seconds=37
  { 2458262.50, 69.101310 }, //24 May 2018, UT1-UTC= 0.0826900, Accumulated Leap Seconds=37
  { 2458263.50, 69.102354 }, //25 May 2018, UT1-UTC= 0.0816457, Accumulated Leap Seconds=37
  { 2458264.50, 69.103350 }, //26 May 2018, UT1-UTC= 0.0806496, Accumulated Leap Seconds=37
  { 2458265.50, 69.104300 }, //27 May 2018, UT1-UTC= 0.0797000, Accumulated Leap Seconds=37
  { 2458266.50, 69.105162 }, //28 May 2018, UT1-UTC= 0.0788383, Accumulated Leap Seconds=37
  { 2458267.50, 69.105830 }, //29 May 2018, UT1-UTC= 0.0781702, Accumulated Leap Seconds=37
  { 2458268.50, 69.106303 }, //30 May 2018, UT1-UTC= 0.0776965, Accumulated Leap Seconds=37
  { 2458269.50, 69.106670 }, //31 May 2018, UT1-UTC= 0.0773296, Accumulated Leap Seconds=37
  { 2458270.50, 69.107003 }, //01 June 2018, UT1-UTC= 0.0769966, Accumulated Leap Seconds=37
  { 2458271.50, 69.107292 }, //02 June 2018, UT1-UTC= 0.0767081, Accumulated Leap Seconds=37
  { 2458272.50, 69.107605 }, //03 June 2018, UT1-UTC= 0.0763948, Accumulated Leap Seconds=37
  { 2458273.50, 69.107954 }, //04 June 2018, UT1-UTC= 0.0760462, Accumulated Leap Seconds=37
  { 2458274.50, 69.108390 }, //05 June 2018, UT1-UTC= 0.0756097, Accumulated Leap Seconds=37
  { 2458275.50, 69.108891 }, //06 June 2018, UT1-UTC= 0.0751095, Accumulated Leap Seconds=37
  { 2458276.50, 69.109508 }, //07 June 2018, UT1-UTC= 0.0744923, Accumulated Leap Seconds=37
  { 2458277.50, 69.110218 }, //08 June 2018, UT1-UTC= 0.0737824, Accumulated Leap Seconds=37
  { 2458278.50, 69.111012 }, //09 June 2018, UT1-UTC= 0.0729877, Accumulated Leap Seconds=37
  { 2458279.50, 69.111894 }, //10 June 2018, UT1-UTC= 0.0721058, Accumulated Leap Seconds=37
  { 2458280.50, 69.112734 }, //11 June 2018, UT1-UTC= 0.0712665, Accumulated Leap Seconds=37
  { 2458281.50, 69.113445 }, //12 June 2018, UT1-UTC= 0.0705547, Accumulated Leap Seconds=37
  { 2458282.50, 69.114002 }, //13 June 2018, UT1-UTC= 0.0699977, Accumulated Leap Seconds=37
  { 2458283.50, 69.114408 }, //14 June 2018, UT1-UTC= 0.0695920, Accumulated Leap Seconds=37
  { 2458284.50, 69.114683 }, //15 June 2018, UT1-UTC= 0.0693172, Accumulated Leap Seconds=37
  { 2458285.50, 69.114886 }, //16 June 2018, UT1-UTC= 0.0691140, Accumulated Leap Seconds=37
  { 2458286.50, 69.115113 }, //17 June 2018, UT1-UTC= 0.0688867, Accumulated Leap Seconds=37
  { 2458287.50, 69.115394 }, //18 June 2018, UT1-UTC= 0.0686057, Accumulated Leap Seconds=37
  { 2458288.50, 69.115781 }, //19 June 2018, UT1-UTC= 0.0682192, Accumulated Leap Seconds=37
  { 2458289.50, 69.116242 }, //20 June 2018, UT1-UTC= 0.0677580, Accumulated Leap Seconds=37
  { 2458290.50, 69.116675 }, //21 June 2018, UT1-UTC= 0.0673250, Accumulated Leap Seconds=37
  { 2458291.50, 69.116998 }, //22 June 2018, UT1-UTC= 0.0670016, Accumulated Leap Seconds=37
  { 2458292.50, 69.117156 }, //23 June 2018, UT1-UTC= 0.0668439, Accumulated Leap Seconds=37
  { 2458293.50, 69.117124 }, //24 June 2018, UT1-UTC= 0.0668756, Accumulated Leap Seconds=37
  { 2458294.50, 69.116874 }, //25 June 2018, UT1-UTC= 0.0671262, Accumulated Leap Seconds=37
  { 2458295.50, 69.116473 }, //26 June 2018, UT1-UTC= 0.0675266, Accumulated Leap Seconds=37
  { 2458296.50, 69.115908 }, //27 June 2018, UT1-UTC= 0.0680923, Accumulated Leap Seconds=37
  { 2458297.50, 69.115273 }, //28 June 2018, UT1-UTC= 0.0687270, Accumulated Leap Seconds=37
  { 2458298.50, 69.114672 }, //29 June 2018, UT1-UTC= 0.0693284, Accumulated Leap Seconds=37
  { 2458299.50, 69.114035 }, //30 June 2018, UT1-UTC= 0.0699649, Accumulated Leap Seconds=37
  { 2458300.50, 69.113403 }, //01 July 2018, UT1-UTC= 0.0705973, Accumulated Leap Seconds=37
  { 2458301.50, 69.112818 }, //02 July 2018, UT1-UTC= 0.0711817, Accumulated Leap Seconds=37
  { 2458302.50, 69.112321 }, //03 July 2018, UT1-UTC= 0.0716793, Accumulated Leap Seconds=37
  { 2458303.50, 69.111947 }, //04 July 2018, UT1-UTC= 0.0720528, Accumulated Leap Seconds=37
  { 2458304.50, 69.111700 }, //05 July 2018, UT1-UTC= 0.0723000, Accumulated Leap Seconds=37
  { 2458305.50, 69.111632 }, //06 July 2018, UT1-UTC= 0.0723675, Accumulated Leap Seconds=37
  { 2458306.50, 69.111681 }, //07 July 2018, UT1-UTC= 0.0723193, Accumulated Leap Seconds=37
  { 2458307.50, 69.111802 }, //08 July 2018, UT1-UTC= 0.0721982, Accumulated Leap Seconds=37
  { 2458308.50, 69.111966 }, //09 July 2018, UT1-UTC= 0.0720344, Accumulated Leap Seconds=37
  { 2458309.50, 69.112111 }, //10 July 2018, UT1-UTC= 0.0718895, Accumulated Leap Seconds=37
  { 2458310.50, 69.112226 }, //11 July 2018, UT1-UTC= 0.0717743, Accumulated Leap Seconds=37
  { 2458311.50, 69.112260 }, //12 July 2018, UT1-UTC= 0.0717395, Accumulated Leap Seconds=37
  { 2458312.50, 69.112221 }, //13 July 2018, UT1-UTC= 0.0717787, Accumulated Leap Seconds=37
  { 2458313.50, 69.112212 }, //14 July 2018, UT1-UTC= 0.0717883, Accumulated Leap Seconds=37
  { 2458314.50, 69.112313 }, //15 July 2018, UT1-UTC= 0.0716867, Accumulated Leap Seconds=37
  { 2458315.50, 69.112588 }, //16 July 2018, UT1-UTC= 0.0714123, Accumulated Leap Seconds=37
  { 2458316.50, 69.113022 }, //17 July 2018, UT1-UTC= 0.0709782, Accumulated Leap Seconds=37
  { 2458317.50, 69.113554 }, //18 July 2018, UT1-UTC= 0.0704463, Accumulated Leap Seconds=37
  { 2458318.50, 69.114098 }, //19 July 2018, UT1-UTC= 0.0699020, Accumulated Leap Seconds=37
  { 2458319.50, 69.114549 }, //20 July 2018, UT1-UTC= 0.0694513, Accumulated Leap Seconds=37
  { 2458320.50, 69.114838 }, //21 July 2018, UT1-UTC= 0.0691622, Accumulated Leap Seconds=37
  { 2458321.50, 69.114954 }, //22 July 2018, UT1-UTC= 0.0690456, Accumulated Leap Seconds=37
  { 2458322.50, 69.114931 }, //23 July 2018, UT1-UTC= 0.0690692, Accumulated Leap Seconds=37
  { 2458323.50, 69.114822 }, //24 July 2018, UT1-UTC= 0.0691781, Accumulated Leap Seconds=37
  { 2458324.50, 69.114670 }, //25 July 2018, UT1-UTC= 0.0693304, Accumulated Leap Seconds=37
  { 2458325.50, 69.114490 }, //26 July 2018, UT1-UTC= 0.0695103, Accumulated Leap Seconds=37
  { 2458326.50, 69.114339 }, //27 July 2018, UT1-UTC= 0.0696607, Accumulated Leap Seconds=37
  { 2458327.50, 69.114217 }, //28 July 2018, UT1-UTC= 0.0697827, Accumulated Leap Seconds=37
  { 2458328.50, 69.114111 }, //29 July 2018, UT1-UTC= 0.0698892, Accumulated Leap Seconds=37
  { 2458329.50, 69.114055 }, //30 July 2018, UT1-UTC= 0.0699453, Accumulated Leap Seconds=37
  { 2458330.50, 69.114068 }, //31 July 2018, UT1-UTC= 0.0699325, Accumulated Leap Seconds=37
  { 2458331.50, 69.114190 }, //01 August 2018, UT1-UTC= 0.0698102, Accumulated Leap Seconds=37
  { 2458332.50, 69.114424 }, //02 August 2018, UT1-UTC= 0.0695763, Accumulated Leap Seconds=37
  { 2458333.50, 69.114737 }, //03 August 2018, UT1-UTC= 0.0692629, Accumulated Leap Seconds=37
  { 2458334.50, 69.115061 }, //04 August 2018, UT1-UTC= 0.0689387, Accumulated Leap Seconds=37
  { 2458335.50, 69.115333 }, //05 August 2018, UT1-UTC= 0.0686671, Accumulated Leap Seconds=37
  { 2458336.50, 69.115539 }, //06 August 2018, UT1-UTC= 0.0684611, Accumulated Leap Seconds=37
  { 2458337.50, 69.115667 }, //07 August 2018, UT1-UTC= 0.0683332, Accumulated Leap Seconds=37
  { 2458338.50, 69.115707 }, //08 August 2018, UT1-UTC= 0.0682926, Accumulated Leap Seconds=37
  { 2458339.50, 69.115721 }, //09 August 2018, UT1-UTC= 0.0682790, Accumulated Leap Seconds=37
  { 2458340.50, 69.115803 }, //10 August 2018, UT1-UTC= 0.0681971, Accumulated Leap Seconds=37
  { 2458341.50, 69.116033 }, //11 August 2018, UT1-UTC= 0.0679668, Accumulated Leap Seconds=37
  { 2458342.50, 69.116463 }, //12 August 2018, UT1-UTC= 0.0675370, Accumulated Leap Seconds=37
  { 2458343.50, 69.117042 }, //13 August 2018, UT1-UTC= 0.0669575, Accumulated Leap Seconds=37
  { 2458344.50, 69.117662 }, //14 August 2018, UT1-UTC= 0.0663382, Accumulated Leap Seconds=37
  { 2458345.50, 69.118266 }, //15 August 2018, UT1-UTC= 0.0657339, Accumulated Leap Seconds=37
  { 2458346.50, 69.118819 }, //16 August 2018, UT1-UTC= 0.0651808, Accumulated Leap Seconds=37
  { 2458347.50, 69.119235 }, //17 August 2018, UT1-UTC= 0.0647654, Accumulated Leap Seconds=37
  { 2458348.50, 69.119511 }, //18 August 2018, UT1-UTC= 0.0644887, Accumulated Leap Seconds=37
  { 2458349.50, 69.119600 }, //19 August 2018, UT1-UTC= 0.0643999, Accumulated Leap Seconds=37
  { 2458350.50, 69.119573 }, //20 August 2018, UT1-UTC= 0.0644274, Accumulated Leap Seconds=37
  { 2458351.50, 69.119472 }, //21 August 2018, UT1-UTC= 0.0645280, Accumulated Leap Seconds=37
  { 2458352.50, 69.119312 }, //22 August 2018, UT1-UTC= 0.0646884, Accumulated Leap Seconds=37
  { 2458353.50, 69.119148 }, //23 August 2018, UT1-UTC= 0.0648517, Accumulated Leap Seconds=37
  { 2458354.50, 69.118998 }, //24 August 2018, UT1-UTC= 0.0650018, Accumulated Leap Seconds=37
  { 2458355.50, 69.118838 }, //25 August 2018, UT1-UTC= 0.0651619, Accumulated Leap Seconds=37
  { 2458356.50, 69.118769 }, //26 August 2018, UT1-UTC= 0.0652315, Accumulated Leap Seconds=37
  { 2458357.50, 69.118891 }, //27 August 2018, UT1-UTC= 0.0651086, Accumulated Leap Seconds=37
  { 2458358.50, 69.119189 }, //28 August 2018, UT1-UTC= 0.0648107, Accumulated Leap Seconds=37
  { 2458359.50, 69.119548 }, //29 August 2018, UT1-UTC= 0.0644521, Accumulated Leap Seconds=37
  { 2458360.50, 69.119905 }, //30 August 2018, UT1-UTC= 0.0640949, Accumulated Leap Seconds=37
  { 2458361.50, 69.120296 }, //31 August 2018, UT1-UTC= 0.0637042, Accumulated Leap Seconds=37
  { 2458362.50, 69.120720 }, //01 September 2018, UT1-UTC= 0.0632797, Accumulated Leap Seconds=37
  { 2458363.50, 69.121048 }, //02 September 2018, UT1-UTC= 0.0629525, Accumulated Leap Seconds=37
  { 2458364.50, 69.121253 }, //03 September 2018, UT1-UTC= 0.0627471, Accumulated Leap Seconds=37
  { 2458365.50, 69.121382 }, //04 September 2018, UT1-UTC= 0.0626180, Accumulated Leap Seconds=37
  { 2458366.50, 69.121477 }, //05 September 2018, UT1-UTC= 0.0625225, Accumulated Leap Seconds=37
  { 2458367.50, 69.121603 }, //06 September 2018, UT1-UTC= 0.0623972, Accumulated Leap Seconds=37
  { 2458368.50, 69.121830 }, //07 September 2018, UT1-UTC= 0.0621696, Accumulated Leap Seconds=37
  { 2458369.50, 69.122249 }, //08 September 2018, UT1-UTC= 0.0617506, Accumulated Leap Seconds=37
  { 2458370.50, 69.122857 }, //09 September 2018, UT1-UTC= 0.0611433, Accumulated Leap Seconds=37
  { 2458371.50, 69.123680 }, //10 September 2018, UT1-UTC= 0.0603195, Accumulated Leap Seconds=37
  { 2458372.50, 69.124651 }, //11 September 2018, UT1-UTC= 0.0593494, Accumulated Leap Seconds=37
  { 2458373.50, 69.125631 }, //12 September 2018, UT1-UTC= 0.0583688, Accumulated Leap Seconds=37
  { 2458374.50, 69.126486 }, //13 September 2018, UT1-UTC= 0.0575139, Accumulated Leap Seconds=37
  { 2458375.50, 69.127142 }, //14 September 2018, UT1-UTC= 0.0568579, Accumulated Leap Seconds=37
  { 2458376.50, 69.127597 }, //15 September 2018, UT1-UTC= 0.0564025, Accumulated Leap Seconds=37
  { 2458377.50, 69.127876 }, //16 September 2018, UT1-UTC= 0.0561236, Accumulated Leap Seconds=37
  { 2458378.50, 69.128005 }, //17 September 2018, UT1-UTC= 0.0559947, Accumulated Leap Seconds=37
  { 2458379.50, 69.128088 }, //18 September 2018, UT1-UTC= 0.0559116, Accumulated Leap Seconds=37
  { 2458380.50, 69.128138 }, //19 September 2018, UT1-UTC= 0.0558621, Accumulated Leap Seconds=37
  { 2458381.50, 69.128187 }, //20 September 2018, UT1-UTC= 0.0558132, Accumulated Leap Seconds=37
  { 2458382.50, 69.128336 }, //21 September 2018, UT1-UTC= 0.0556637, Accumulated Leap Seconds=37
  { 2458383.50, 69.128611 }, //22 September 2018, UT1-UTC= 0.0553887, Accumulated Leap Seconds=37
  { 2458384.50, 69.129026 }, //23 September 2018, UT1-UTC= 0.0549741, Accumulated Leap Seconds=37
  { 2458385.50, 69.129622 }, //24 September 2018, UT1-UTC= 0.0543782, Accumulated Leap Seconds=37
  { 2458386.50, 69.130387 }, //25 September 2018, UT1-UTC= 0.0536128, Accumulated Leap Seconds=37
  { 2458387.50, 69.131261 }, //26 September 2018, UT1-UTC= 0.0527386, Accumulated Leap Seconds=37
  { 2458388.50, 69.132158 }, //27 September 2018, UT1-UTC= 0.0518424, Accumulated Leap Seconds=37
  { 2458389.50, 69.133088 }, //28 September 2018, UT1-UTC= 0.0509123, Accumulated Leap Seconds=37
  { 2458390.50, 69.133991 }, //29 September 2018, UT1-UTC= 0.0500090, Accumulated Leap Seconds=37
  { 2458391.50, 69.134815 }, //30 September 2018, UT1-UTC= 0.0491855, Accumulated Leap Seconds=37
  { 2458392.50, 69.135558 }, //01 October 2018, UT1-UTC= 0.0484422, Accumulated Leap Seconds=37
  { 2458393.50, 69.136244 }, //02 October 2018, UT1-UTC= 0.0477556, Accumulated Leap Seconds=37
  { 2458394.50, 69.136905 }, //03 October 2018, UT1-UTC= 0.0470950, Accumulated Leap Seconds=37
  { 2458395.50, 69.137617 }, //04 October 2018, UT1-UTC= 0.0463829, Accumulated Leap Seconds=37
  { 2458396.50, 69.138467 }, //05 October 2018, UT1-UTC= 0.0455331, Accumulated Leap Seconds=37
  { 2458397.50, 69.139572 }, //06 October 2018, UT1-UTC= 0.0444284, Accumulated Leap Seconds=37
  { 2458398.50, 69.140833 }, //07 October 2018, UT1-UTC= 0.0431670, Accumulated Leap Seconds=37
  { 2458399.50, 69.142207 }, //08 October 2018, UT1-UTC= 0.0417926, Accumulated Leap Seconds=37
  { 2458400.50, 69.143653 }, //09 October 2018, UT1-UTC= 0.0403473, Accumulated Leap Seconds=37
  { 2458401.50, 69.145027 }, //10 October 2018, UT1-UTC= 0.0389725, Accumulated Leap Seconds=37
  { 2458402.50, 69.146263 }, //11 October 2018, UT1-UTC= 0.0377371, Accumulated Leap Seconds=37
  { 2458403.50, 69.147323 }, //12 October 2018, UT1-UTC= 0.0366770, Accumulated Leap Seconds=37
  { 2458404.50, 69.148196 }, //13 October 2018, UT1-UTC= 0.0358042, Accumulated Leap Seconds=37
  { 2458405.50, 69.148899 }, //14 October 2018, UT1-UTC= 0.0351009, Accumulated Leap Seconds=37
  { 2458406.50, 69.149491 }, //15 October 2018, UT1-UTC= 0.0345090, Accumulated Leap Seconds=37
  { 2458407.50, 69.149951 }, //16 October 2018, UT1-UTC= 0.0340490, Accumulated Leap Seconds=37
  { 2458408.50, 69.150364 }, //17 October 2018, UT1-UTC= 0.0336358, Accumulated Leap Seconds=37
  { 2458409.50, 69.150816 }, //18 October 2018, UT1-UTC= 0.0331836, Accumulated Leap Seconds=37
  { 2458410.50, 69.151366 }, //19 October 2018, UT1-UTC= 0.0326338, Accumulated Leap Seconds=37
  { 2458411.50, 69.152019 }, //20 October 2018, UT1-UTC= 0.0319812, Accumulated Leap Seconds=37
  { 2458412.50, 69.152809 }, //21 October 2018, UT1-UTC= 0.0311914, Accumulated Leap Seconds=37
  { 2458413.50, 69.153788 }, //22 October 2018, UT1-UTC= 0.0302123, Accumulated Leap Seconds=37
  { 2458414.50, 69.154911 }, //23 October 2018, UT1-UTC= 0.0290892, Accumulated Leap Seconds=37
  { 2458415.50, 69.156150 }, //24 October 2018, UT1-UTC= 0.0278497, Accumulated Leap Seconds=37
  { 2458416.50, 69.157448 }, //25 October 2018, UT1-UTC= 0.0265518, Accumulated Leap Seconds=37
  { 2458417.50, 69.158705 }, //26 October 2018, UT1-UTC= 0.0252947, Accumulated Leap Seconds=37
  { 2458418.50, 69.159871 }, //27 October 2018, UT1-UTC= 0.0241285, Accumulated Leap Seconds=37
  { 2458419.50, 69.160941 }, //28 October 2018, UT1-UTC= 0.0230588, Accumulated Leap Seconds=37
  { 2458420.50, 69.161905 }, //29 October 2018, UT1-UTC= 0.0220948, Accumulated Leap Seconds=37
  { 2458421.50, 69.162796 }, //30 October 2018, UT1-UTC= 0.0212045, Accumulated Leap Seconds=37
  { 2458422.50, 69.163660 }, //31 October 2018, UT1-UTC= 0.0203398, Accumulated Leap Seconds=37
  { 2458423.50, 69.164590 }, //01 November 2018, UT1-UTC= 0.0194100, Accumulated Leap Seconds=37
  { 2458424.50, 69.165661 }, //02 November 2018, UT1-UTC= 0.0183392, Accumulated Leap Seconds=37
  { 2458425.50, 69.166846 }, //03 November 2018, UT1-UTC= 0.0171535, Accumulated Leap Seconds=37
  { 2458426.50, 69.168094 }, //04 November 2018, UT1-UTC= 0.0159064, Accumulated Leap Seconds=37
  { 2458427.50, 69.169348 }, //05 November 2018, UT1-UTC= 0.0146517, Accumulated Leap Seconds=37
  { 2458428.50, 69.170587 }, //06 November 2018, UT1-UTC= 0.0134133, Accumulated Leap Seconds=37
  { 2458429.50, 69.171770 }, //07 November 2018, UT1-UTC= 0.0122295, Accumulated Leap Seconds=37
  { 2458430.50, 69.172807 }, //08 November 2018, UT1-UTC= 0.0111928, Accumulated Leap Seconds=37
  { 2458431.50, 69.173742 }, //09 November 2018, UT1-UTC= 0.0102582, Accumulated Leap Seconds=37
  { 2458432.50, 69.174530 }, //10 November 2018, UT1-UTC= 0.0094704, Accumulated Leap Seconds=37
  { 2458433.50, 69.175233 }, //11 November 2018, UT1-UTC= 0.0087666, Accumulated Leap Seconds=37
  { 2458434.50, 69.175909 }, //12 November 2018, UT1-UTC= 0.0080908, Accumulated Leap Seconds=37
  { 2458435.50, 69.176558 }, //13 November 2018, UT1-UTC= 0.0074419, Accumulated Leap Seconds=37
  { 2458436.50, 69.177269 }, //14 November 2018, UT1-UTC= 0.0067305, Accumulated Leap Seconds=37
  { 2458437.50, 69.178042 }, //15 November 2018, UT1-UTC= 0.0059584, Accumulated Leap Seconds=37
  { 2458438.50, 69.178905 }, //16 November 2018, UT1-UTC= 0.0050952, Accumulated Leap Seconds=37
  { 2458439.50, 69.179859 }, //17 November 2018, UT1-UTC= 0.0041410, Accumulated Leap Seconds=37
  { 2458440.50, 69.180921 }, //18 November 2018, UT1-UTC= 0.0030793, Accumulated Leap Seconds=37
  { 2458441.50, 69.182130 }, //19 November 2018, UT1-UTC= 0.0018698, Accumulated Leap Seconds=37
  { 2458442.50, 69.183475 }, //20 November 2018, UT1-UTC= 0.0005255, Accumulated Leap Seconds=37
  { 2458443.50, 69.184874 }, //21 November 2018, UT1-UTC=-0.0008739, Accumulated Leap Seconds=37
  { 2458444.50, 69.186264 }, //22 November 2018, UT1-UTC=-0.0022644, Accumulated Leap Seconds=37
  { 2458445.50, 69.187586 }, //23 November 2018, UT1-UTC=-0.0035859, Accumulated Leap Seconds=37
  { 2458446.50, 69.188808 }, //24 November 2018, UT1-UTC=-0.0048081, Accumulated Leap Seconds=37
  { 2458447.50, 69.189892 }, //25 November 2018, UT1-UTC=-0.0058923, Accumulated Leap Seconds=37
  { 2458448.50, 69.190861 }, //26 November 2018, UT1-UTC=-0.0068614, Accumulated Leap Seconds=37
  { 2458449.50, 69.191819 }, //27 November 2018, UT1-UTC=-0.0078195, Accumulated Leap Seconds=37
  { 2458450.50, 69.192816 }, //28 November 2018, UT1-UTC=-0.0088161, Accumulated Leap Seconds=37
  { 2458451.50, 69.193925 }, //29 November 2018, UT1-UTC=-0.0099250, Accumulated Leap Seconds=37
  { 2458452.50, 69.195138 }, //30 November 2018, UT1-UTC=-0.0111375, Accumulated Leap Seconds=37
  { 2458453.50, 69.196423 }, //01 December 2018, UT1-UTC=-0.0124228, Accumulated Leap Seconds=37
  { 2458454.50, 69.197734 }, //02 December 2018, UT1-UTC=-0.0137338, Accumulated Leap Seconds=37
  { 2458455.50, 69.198992 }, //03 December 2018, UT1-UTC=-0.0149921, Accumulated Leap Seconds=37
  { 2458456.50, 69.200115 }, //04 December 2018, UT1-UTC=-0.0161149, Accumulated Leap Seconds=37
  { 2458457.50, 69.200995 }, //05 December 2018, UT1-UTC=-0.0169955, Accumulated Leap Seconds=37
  { 2458458.50, 69.201681 }, //06 December 2018, UT1-UTC=-0.0176813, Accumulated Leap Seconds=37
  { 2458459.50, 69.202237 }, //07 December 2018, UT1-UTC=-0.0182366, Accumulated Leap Seconds=37
  { 2458460.50, 69.202731 }, //08 December 2018, UT1-UTC=-0.0187315, Accumulated Leap Seconds=37
  { 2458461.50, 69.203134 }, //09 December 2018, UT1-UTC=-0.0191338, Accumulated Leap Seconds=37
  { 2458462.50, 69.203491 }, //10 December 2018, UT1-UTC=-0.0194914, Accumulated Leap Seconds=37
  { 2458463.50, 69.203882 }, //11 December 2018, UT1-UTC=-0.0198816, Accumulated Leap Seconds=37
  { 2458464.50, 69.204350 }, //12 December 2018, UT1-UTC=-0.0203504, Accumulated Leap Seconds=37
  { 2458465.50, 69.204888 }, //13 December 2018, UT1-UTC=-0.0208876, Accumulated Leap Seconds=37
  { 2458466.50, 69.205493 }, //14 December 2018, UT1-UTC=-0.0214930, Accumulated Leap Seconds=37
  { 2458467.50, 69.206195 }, //15 December 2018, UT1-UTC=-0.0221946, Accumulated Leap Seconds=37
  { 2458468.50, 69.207008 }, //16 December 2018, UT1-UTC=-0.0230078, Accumulated Leap Seconds=37
  { 2458469.50, 69.207941 }, //17 December 2018, UT1-UTC=-0.0239410, Accumulated Leap Seconds=37
  { 2458470.50, 69.208967 }, //18 December 2018, UT1-UTC=-0.0249665, Accumulated Leap Seconds=37
  { 2458471.50, 69.210005 }, //19 December 2018, UT1-UTC=-0.0260051, Accumulated Leap Seconds=37
  { 2458472.50, 69.210907 }, //20 December 2018, UT1-UTC=-0.0269070, Accumulated Leap Seconds=37
  { 2458473.50, 69.211635 }, //21 December 2018, UT1-UTC=-0.0276346, Accumulated Leap Seconds=37
  { 2458474.50, 69.212213 }, //22 December 2018, UT1-UTC=-0.0282128, Accumulated Leap Seconds=37
  { 2458475.50, 69.212679 }, //23 December 2018, UT1-UTC=-0.0286787, Accumulated Leap Seconds=37
  { 2458476.50, 69.213140 }, //24 December 2018, UT1-UTC=-0.0291401, Accumulated Leap Seconds=37
  { 2458477.50, 69.213669 }, //25 December 2018, UT1-UTC=-0.0296689, Accumulated Leap Seconds=37
  { 2458478.50, 69.214294 }, //26 December 2018, UT1-UTC=-0.0302944, Accumulated Leap Seconds=37
  { 2458479.50, 69.215076 }, //27 December 2018, UT1-UTC=-0.0310765, Accumulated Leap Seconds=37
  { 2458480.50, 69.216011 }, //28 December 2018, UT1-UTC=-0.0320114, Accumulated Leap Seconds=37
  { 2458481.50, 69.217085 }, //29 December 2018, UT1-UTC=-0.0330854, Accumulated Leap Seconds=37
  { 2458482.50, 69.218166 }, //30 December 2018, UT1-UTC=-0.0341662, Accumulated Leap Seconds=37
  { 2458483.50, 69.219195 }, //31 December 2018, UT1-UTC=-0.0351948, Accumulated Leap Seconds=37
  { 2458484.50, 69.220163 }, //01 January 2019, UT1-UTC=-0.0361632, Accumulated Leap Seconds=37
  { 2458485.50, 69.221045 }, //02 January 2019, UT1-UTC=-0.0370452, Accumulated Leap Seconds=37
  { 2458486.50, 69.221758 }, //03 January 2019, UT1-UTC=-0.0377584, Accumulated Leap Seconds=37
  { 2458487.50, 69.222271 }, //04 January 2019, UT1-UTC=-0.0382710, Accumulated Leap Seconds=37
  { 2458488.50, 69.222723 }, //05 January 2019, UT1-UTC=-0.0387231, Accumulated Leap Seconds=37
  { 2458489.50, 69.223111 }, //06 January 2019, UT1-UTC=-0.0391107, Accumulated Leap Seconds=37
  { 2458490.50, 69.223432 }, //07 January 2019, UT1-UTC=-0.0394318, Accumulated Leap Seconds=37
  { 2458491.50, 69.223761 }, //08 January 2019, UT1-UTC=-0.0397606, Accumulated Leap Seconds=37
  { 2458492.50, 69.224123 }, //09 January 2019, UT1-UTC=-0.0401230, Accumulated Leap Seconds=37
  { 2458493.50, 69.224582 }, //10 January 2019, UT1-UTC=-0.0405821, Accumulated Leap Seconds=37
  { 2458494.50, 69.225091 }, //11 January 2019, UT1-UTC=-0.0410915, Accumulated Leap Seconds=37
  { 2458495.50, 69.225734 }, //12 January 2019, UT1-UTC=-0.0417339, Accumulated Leap Seconds=37
  { 2458496.50, 69.226465 }, //13 January 2019, UT1-UTC=-0.0424652, Accumulated Leap Seconds=37
  { 2458497.50, 69.227266 }, //14 January 2019, UT1-UTC=-0.0432663, Accumulated Leap Seconds=37
  { 2458498.50, 69.228143 }, //15 January 2019, UT1-UTC=-0.0441428, Accumulated Leap Seconds=37
  { 2458499.50, 69.229047 }, //16 January 2019, UT1-UTC=-0.0450467, Accumulated Leap Seconds=37
  { 2458500.50, 69.229913 }, //17 January 2019, UT1-UTC=-0.0459130, Accumulated Leap Seconds=37
  { 2458501.50, 69.230725 }, //18 January 2019, UT1-UTC=-0.0467252, Accumulated Leap Seconds=37
  { 2458502.50, 69.231481 }, //19 January 2019, UT1-UTC=-0.0474805, Accumulated Leap Seconds=37
  { 2458503.50, 69.232218 }, //20 January 2019, UT1-UTC=-0.0482177, Accumulated Leap Seconds=37
  { 2458504.50, 69.233007 }, //21 January 2019, UT1-UTC=-0.0490075, Accumulated Leap Seconds=37
  { 2458505.50, 69.233919 }, //22 January 2019, UT1-UTC=-0.0499191, Accumulated Leap Seconds=37
  { 2458506.50, 69.234977 }, //23 January 2019, UT1-UTC=-0.0509772, Accumulated Leap Seconds=37
  { 2458507.50, 69.236230 }, //24 January 2019, UT1-UTC=-0.0522303, Accumulated Leap Seconds=37
  { 2458508.50, 69.237634 }, //25 January 2019, UT1-UTC=-0.0536338, Accumulated Leap Seconds=37
  { 2458509.50, 69.239069 }, //26 January 2019, UT1-UTC=-0.0550685, Accumulated Leap Seconds=37
  { 2458510.50, 69.240486 }, //27 January 2019, UT1-UTC=-0.0564856, Accumulated Leap Seconds=37
  { 2458511.50, 69.241808 }, //28 January 2019, UT1-UTC=-0.0578081, Accumulated Leap Seconds=37
  { 2458512.50, 69.242911 }, //29 January 2019, UT1-UTC=-0.0589114, Accumulated Leap Seconds=37
  { 2458513.50, 69.243832 }, //30 January 2019, UT1-UTC=-0.0598320, Accumulated Leap Seconds=37
  { 2458514.50, 69.244576 }, //31 January 2019, UT1-UTC=-0.0605762, Accumulated Leap Seconds=37
  { 2458515.50, 69.245156 }, //01 February 2019, UT1-UTC=-0.0611564, Accumulated Leap Seconds=37
  { 2458516.50, 69.245662 }, //02 February 2019, UT1-UTC=-0.0616624, Accumulated Leap Seconds=37
  { 2458517.50, 69.246128 }, //03 February 2019, UT1-UTC=-0.0621277, Accumulated Leap Seconds=37
  { 2458518.50, 69.246653 }, //04 February 2019, UT1-UTC=-0.0626533, Accumulated Leap Seconds=37
  { 2458519.50, 69.247318 }, //05 February 2019, UT1-UTC=-0.0633185, Accumulated Leap Seconds=37
  { 2458520.50, 69.248089 }, //06 February 2019, UT1-UTC=-0.0640895, Accumulated Leap Seconds=37
  { 2458521.50, 69.248984 }, //07 February 2019, UT1-UTC=-0.0649842, Accumulated Leap Seconds=37
  { 2458522.50, 69.249979 }, //08 February 2019, UT1-UTC=-0.0659790, Accumulated Leap Seconds=37
  { 2458523.50, 69.251067 }, //09 February 2019, UT1-UTC=-0.0670674, Accumulated Leap Seconds=37
  { 2458524.50, 69.252229 }, //10 February 2019, UT1-UTC=-0.0682290, Accumulated Leap Seconds=37
  { 2458525.50, 69.253433 }, //11 February 2019, UT1-UTC=-0.0694334, Accumulated Leap Seconds=37
  { 2458526.50, 69.254616 }, //12 February 2019, UT1-UTC=-0.0706159, Accumulated Leap Seconds=37
  { 2458527.50, 69.255794 }, //13 February 2019, UT1-UTC=-0.0717940, Accumulated Leap Seconds=37
  { 2458528.50, 69.256926 }, //14 February 2019, UT1-UTC=-0.0729259, Accumulated Leap Seconds=37
  { 2458529.50, 69.257966 }, //15 February 2019, UT1-UTC=-0.0739665, Accumulated Leap Seconds=37
  { 2458530.50, 69.258954 }, //16 February 2019, UT1-UTC=-0.0749540, Accumulated Leap Seconds=37
  { 2458531.50, 69.259918 }, //17 February 2019, UT1-UTC=-0.0759185, Accumulated Leap Seconds=37
  { 2458532.50, 69.260930 }, //18 February 2019, UT1-UTC=-0.0769301, Accumulated Leap Seconds=37
  { 2458533.50, 69.262101 }, //19 February 2019, UT1-UTC=-0.0781015, Accumulated Leap Seconds=37
  { 2458534.50, 69.263456 }, //20 February 2019, UT1-UTC=-0.0794558, Accumulated Leap Seconds=37
  { 2458535.50, 69.264945 }, //21 February 2019, UT1-UTC=-0.0809446, Accumulated Leap Seconds=37
  { 2458536.50, 69.266487 }, //22 February 2019, UT1-UTC=-0.0824874, Accumulated Leap Seconds=37
  { 2458537.50, 69.267992 }, //23 February 2019, UT1-UTC=-0.0839918, Accumulated Leap Seconds=37
  { 2458538.50, 69.269379 }, //24 February 2019, UT1-UTC=-0.0853787, Accumulated Leap Seconds=37
  { 2458539.50, 69.270533 }, //25 February 2019, UT1-UTC=-0.0865327, Accumulated Leap Seconds=37
  { 2458540.50, 69.271483 }, //26 February 2019, UT1-UTC=-0.0874829, Accumulated Leap Seconds=37
  { 2458541.50, 69.272212 }, //27 February 2019, UT1-UTC=-0.0882119, Accumulated Leap Seconds=37
  { 2458542.50, 69.272796 }, //28 February 2019, UT1-UTC=-0.0887955, Accumulated Leap Seconds=37
  { 2458543.50, 69.273276 }, //01 March 2019, UT1-UTC=-0.0892758, Accumulated Leap Seconds=37
  { 2458544.50, 69.273715 }, //02 March 2019, UT1-UTC=-0.0897148, Accumulated Leap Seconds=37
  { 2458545.50, 69.274210 }, //03 March 2019, UT1-UTC=-0.0902098, Accumulated Leap Seconds=37
  { 2458546.50, 69.274725 }, //04 March 2019, UT1-UTC=-0.0907251, Accumulated Leap Seconds=37
  { 2458547.50, 69.275334 }, //05 March 2019, UT1-UTC=-0.0913344, Accumulated Leap Seconds=37
  { 2458548.50, 69.276043 }, //06 March 2019, UT1-UTC=-0.0920430, Accumulated Leap Seconds=37
  { 2458549.50, 69.276878 }, //07 March 2019, UT1-UTC=-0.0928782, Accumulated Leap Seconds=37
  { 2458550.50, 69.277828 }, //08 March 2019, UT1-UTC=-0.0938285, Accumulated Leap Seconds=37
  { 2458551.50, 69.278822 }, //09 March 2019, UT1-UTC=-0.0948222, Accumulated Leap Seconds=37
  { 2458552.50, 69.279867 }, //10 March 2019, UT1-UTC=-0.0958673, Accumulated Leap Seconds=37
  { 2458553.50, 69.280942 }, //11 March 2019, UT1-UTC=-0.0969416, Accumulated Leap Seconds=37
  { 2458554.50, 69.281963 }, //12 March 2019, UT1-UTC=-0.0979631, Accumulated Leap Seconds=37
  { 2458555.50, 69.282882 }, //13 March 2019, UT1-UTC=-0.0988823, Accumulated Leap Seconds=37
  { 2458556.50, 69.283702 }, //14 March 2019, UT1-UTC=-0.0997021, Accumulated Leap Seconds=37
  { 2458557.50, 69.284469 }, //15 March 2019, UT1-UTC=-0.1004690, Accumulated Leap Seconds=37
  { 2458558.50, 69.285269 }, //16 March 2019, UT1-UTC=-0.1012687, Accumulated Leap Seconds=37
  { 2458559.50, 69.286138 }, //17 March 2019, UT1-UTC=-0.1021385, Accumulated Leap Seconds=37
  { 2458560.50, 69.287134 }, //18 March 2019, UT1-UTC=-0.1031338, Accumulated Leap Seconds=37
  { 2458561.50, 69.288342 }, //19 March 2019, UT1-UTC=-0.1043416, Accumulated Leap Seconds=37
  { 2458562.50, 69.289750 }, //20 March 2019, UT1-UTC=-0.1057496, Accumulated Leap Seconds=37
  { 2458563.50, 69.291324 }, //21 March 2019, UT1-UTC=-0.1073241, Accumulated Leap Seconds=37
  { 2458564.50, 69.293003 }, //22 March 2019, UT1-UTC=-0.1090030, Accumulated Leap Seconds=37
  { 2458565.50, 69.294633 }, //23 March 2019, UT1-UTC=-0.1106334, Accumulated Leap Seconds=37
  { 2458566.50, 69.296124 }, //24 March 2019, UT1-UTC=-0.1121243, Accumulated Leap Seconds=37
  { 2458567.50, 69.297399 }, //25 March 2019, UT1-UTC=-0.1133992, Accumulated Leap Seconds=37
  { 2458568.50, 69.298443 }, //26 March 2019, UT1-UTC=-0.1144431, Accumulated Leap Seconds=37
  { 2458569.50, 69.299318 }, //27 March 2019, UT1-UTC=-0.1153175, Accumulated Leap Seconds=37
  { 2458570.50, 69.300097 }, //28 March 2019, UT1-UTC=-0.1160974, Accumulated Leap Seconds=37
  { 2458571.50, 69.300810 }, //29 March 2019, UT1-UTC=-0.1168102, Accumulated Leap Seconds=37
  { 2458572.50, 69.301540 }, //30 March 2019, UT1-UTC=-0.1175397, Accumulated Leap Seconds=37
  { 2458573.50, 69.302324 }, //31 March 2019, UT1-UTC=-0.1183242, Accumulated Leap Seconds=37
  { 2458574.50, 69.303198 }, //01 April 2019, UT1-UTC=-0.1191979, Accumulated Leap Seconds=37
  { 2458575.50, 69.304165 }, //02 April 2019, UT1-UTC=-0.1201648, Accumulated Leap Seconds=37
  { 2458576.50, 69.305246 }, //03 April 2019, UT1-UTC=-0.1212465, Accumulated Leap Seconds=37
  { 2458577.50, 69.306473 }, //04 April 2019, UT1-UTC=-0.1224726, Accumulated Leap Seconds=37
  { 2458578.50, 69.307757 }, //05 April 2019, UT1-UTC=-0.1237568, Accumulated Leap Seconds=37
  { 2458579.50, 69.309040 }, //06 April 2019, UT1-UTC=-0.1250404, Accumulated Leap Seconds=37
  { 2458580.50, 69.310335 }, //07 April 2019, UT1-UTC=-0.1263347, Accumulated Leap Seconds=37
  { 2458581.50, 69.311579 }, //08 April 2019, UT1-UTC=-0.1275788, Accumulated Leap Seconds=37
  { 2458582.50, 69.312742 }, //09 April 2019, UT1-UTC=-0.1287422, Accumulated Leap Seconds=37
  { 2458583.50, 69.313774 }, //10 April 2019, UT1-UTC=-0.1297738, Accumulated Leap Seconds=37
  { 2458584.50, 69.314658 }, //11 April 2019, UT1-UTC=-0.1306581, Accumulated Leap Seconds=37
  { 2458585.50, 69.315416 }, //12 April 2019, UT1-UTC=-0.1314155, Accumulated Leap Seconds=37
  { 2458586.50, 69.316211 }, //13 April 2019, UT1-UTC=-0.1322115, Accumulated Leap Seconds=37
  { 2458587.50, 69.317062 }, //14 April 2019, UT1-UTC=-0.1330617, Accumulated Leap Seconds=37
  { 2458588.50, 69.318044 }, //15 April 2019, UT1-UTC=-0.1340441, Accumulated Leap Seconds=37
  { 2458589.50, 69.319233 }, //16 April 2019, UT1-UTC=-0.1352327, Accumulated Leap Seconds=37
  { 2458590.50, 69.320600 }, //17 April 2019, UT1-UTC=-0.1365999, Accumulated Leap Seconds=37
  { 2458591.50, 69.322075 }, //18 April 2019, UT1-UTC=-0.1380752, Accumulated Leap Seconds=37
  { 2458592.50, 69.323490 }, //19 April 2019, UT1-UTC=-0.1394900, Accumulated Leap Seconds=37
  { 2458593.50, 69.324803 }, //20 April 2019, UT1-UTC=-0.1408027, Accumulated Leap Seconds=37
  { 2458594.50, 69.325930 }, //21 April 2019, UT1-UTC=-0.1419298, Accumulated Leap Seconds=37
  { 2458595.50, 69.326810 }, //22 April 2019, UT1-UTC=-0.1428101, Accumulated Leap Seconds=37
  { 2458596.50, 69.327476 }, //23 April 2019, UT1-UTC=-0.1434763, Accumulated Leap Seconds=37
  { 2458597.50, 69.328017 }, //24 April 2019, UT1-UTC=-0.1440169, Accumulated Leap Seconds=37
  { 2458598.50, 69.328513 }, //25 April 2019, UT1-UTC=-0.1445128, Accumulated Leap Seconds=37
  { 2458599.50, 69.329004 }, //26 April 2019, UT1-UTC=-0.1450041, Accumulated Leap Seconds=37
  { 2458600.50, 69.329563 }, //27 April 2019, UT1-UTC=-0.1455633, Accumulated Leap Seconds=37
  { 2458601.50, 69.330175 }, //28 April 2019, UT1-UTC=-0.1461748, Accumulated Leap Seconds=37
  { 2458602.50, 69.330839 }, //29 April 2019, UT1-UTC=-0.1468390, Accumulated Leap Seconds=37
  { 2458603.50, 69.331632 }, //30 April 2019, UT1-UTC=-0.1476321, Accumulated Leap Seconds=37
  { 2458604.50, 69.332567 }, //01 May 2019, UT1-UTC=-0.1485675, Accumulated Leap Seconds=37
  { 2458605.50, 69.333612 }, //02 May 2019, UT1-UTC=-0.1496123, Accumulated Leap Seconds=37
  { 2458606.50, 69.334656 }, //03 May 2019, UT1-UTC=-0.1506564, Accumulated Leap Seconds=37
  { 2458607.50, 69.335631 }, //04 May 2019, UT1-UTC=-0.1516315, Accumulated Leap Seconds=37
  { 2458608.50, 69.336533 }, //05 May 2019, UT1-UTC=-0.1525334, Accumulated Leap Seconds=37
  { 2458609.50, 69.337366 }, //06 May 2019, UT1-UTC=-0.1533663, Accumulated Leap Seconds=37
  { 2458610.50, 69.338093 }, //07 May 2019, UT1-UTC=-0.1540934, Accumulated Leap Seconds=37
  { 2458611.50, 69.338711 }, //08 May 2019, UT1-UTC=-0.1547113, Accumulated Leap Seconds=37
  { 2458612.50, 69.339280 }, //09 May 2019, UT1-UTC=-0.1552802, Accumulated Leap Seconds=37
  { 2458613.50, 69.339827 }, //10 May 2019, UT1-UTC=-0.1558271, Accumulated Leap Seconds=37
  { 2458614.50, 69.340399 }, //11 May 2019, UT1-UTC=-0.1563994, Accumulated Leap Seconds=37
  { 2458615.50, 69.341094 }, //12 May 2019, UT1-UTC=-0.1570941, Accumulated Leap Seconds=37
  { 2458616.50, 69.341942 }, //13 May 2019, UT1-UTC=-0.1579420, Accumulated Leap Seconds=37
  { 2458617.50, 69.342971 }, //14 May 2019, UT1-UTC=-0.1589712, Accumulated Leap Seconds=37
  { 2458618.50, 69.344180 }, //15 May 2019, UT1-UTC=-0.1601803, Accumulated Leap Seconds=37
  { 2458619.50, 69.345457 }, //16 May 2019, UT1-UTC=-0.1614572, Accumulated Leap Seconds=37
  { 2458620.50, 69.346712 }, //17 May 2019, UT1-UTC=-0.1627123, Accumulated Leap Seconds=37
  { 2458621.50, 69.347795 }, //18 May 2019, UT1-UTC=-0.1637952, Accumulated Leap Seconds=37
  { 2458622.50, 69.348605 }, //19 May 2019, UT1-UTC=-0.1646049, Accumulated Leap Seconds=37
  { 2458623.50, 69.349206 }, //20 May 2019, UT1-UTC=-0.1652061, Accumulated Leap Seconds=37
  { 2458624.50, 69.349595 }, //21 May 2019, UT1-UTC=-0.1655949, Accumulated Leap Seconds=37
  { 2458625.50, 69.349853 }, //22 May 2019, UT1-UTC=-0.1658535, Accumulated Leap Seconds=37
  { 2458626.50, 69.350076 }, //23 May 2019, UT1-UTC=-0.1660761, Accumulated Leap Seconds=37
  { 2458627.50, 69.350313 }, //24 May 2019, UT1-UTC=-0.1663131, Accumulated Leap Seconds=37
  { 2458628.50, 69.350590 }, //25 May 2019, UT1-UTC=-0.1665897, Accumulated Leap Seconds=37
  { 2458629.50, 69.350908 }, //26 May 2019, UT1-UTC=-0.1669079, Accumulated Leap Seconds=37
  { 2458630.50, 69.351309 }, //27 May 2019, UT1-UTC=-0.1673088, Accumulated Leap Seconds=37
  { 2458631.50, 69.351776 }, //28 May 2019, UT1-UTC=-0.1677762, Accumulated Leap Seconds=37
  { 2458632.50, 69.352288 }, //29 May 2019, UT1-UTC=-0.1682876, Accumulated Leap Seconds=37
  { 2458633.50, 69.352872 }, //30 May 2019, UT1-UTC=-0.1688717, Accumulated Leap Seconds=37
  { 2458634.50, 69.353467 }, //31 May 2019, UT1-UTC=-0.1694668, Accumulated Leap Seconds=37
  { 2458635.50, 69.354051 }, //01 June 2019, UT1-UTC=-0.1700507, Accumulated Leap Seconds=37
  { 2458636.50, 69.354589 }, //02 June 2019, UT1-UTC=-0.1705893, Accumulated Leap Seconds=37
  { 2458637.50, 69.354973 }, //03 June 2019, UT1-UTC=-0.1709735, Accumulated Leap Seconds=37
  { 2458638.50, 69.355220 }, //04 June 2019, UT1-UTC=-0.1712196, Accumulated Leap Seconds=37
  { 2458639.50, 69.355393 }, //05 June 2019, UT1-UTC=-0.1713926, Accumulated Leap Seconds=37
  { 2458640.50, 69.355519 }, //06 June 2019, UT1-UTC=-0.1715186, Accumulated Leap Seconds=37
  { 2458641.50, 69.355640 }, //07 June 2019, UT1-UTC=-0.1716405, Accumulated Leap Seconds=37
  { 2458642.50, 69.355829 }, //08 June 2019, UT1-UTC=-0.1718291, Accumulated Leap Seconds=37
  { 2458643.50, 69.356139 }, //09 June 2019, UT1-UTC=-0.1721388, Accumulated Leap Seconds=37
  { 2458644.50, 69.356610 }, //10 June 2019, UT1-UTC=-0.1726102, Accumulated Leap Seconds=37
  { 2458645.50, 69.357214 }, //11 June 2019, UT1-UTC=-0.1732135, Accumulated Leap Seconds=37
  { 2458646.50, 69.357870 }, //12 June 2019, UT1-UTC=-0.1738701, Accumulated Leap Seconds=37
  { 2458647.50, 69.358495 }, //13 June 2019, UT1-UTC=-0.1744949, Accumulated Leap Seconds=37
  { 2458648.50, 69.359015 }, //14 June 2019, UT1-UTC=-0.1750150, Accumulated Leap Seconds=37
  { 2458649.50, 69.359362 }, //15 June 2019, UT1-UTC=-0.1753616, Accumulated Leap Seconds=37
  { 2458650.50, 69.359471 }, //16 June 2019, UT1-UTC=-0.1754711, Accumulated Leap Seconds=37
  { 2458651.50, 69.359408 }, //17 June 2019, UT1-UTC=-0.1754081, Accumulated Leap Seconds=37
  { 2458652.50, 69.359244 }, //18 June 2019, UT1-UTC=-0.1752440, Accumulated Leap Seconds=37
  { 2458653.50, 69.359012 }, //19 June 2019, UT1-UTC=-0.1750116, Accumulated Leap Seconds=37
  { 2458654.50, 69.358768 }, //20 June 2019, UT1-UTC=-0.1747685, Accumulated Leap Seconds=37
  { 2458655.50, 69.358574 }, //21 June 2019, UT1-UTC=-0.1745736, Accumulated Leap Seconds=37
  { 2458656.50, 69.358438 }, //22 June 2019, UT1-UTC=-0.1744381, Accumulated Leap Seconds=37
  { 2458657.50, 69.358352 }, //23 June 2019, UT1-UTC=-0.1743521, Accumulated Leap Seconds=37
  { 2458658.50, 69.358371 }, //24 June 2019, UT1-UTC=-0.1743707, Accumulated Leap Seconds=37
  { 2458659.50, 69.358416 }, //25 June 2019, UT1-UTC=-0.1744160, Accumulated Leap Seconds=37
  { 2458660.50, 69.358479 }, //26 June 2019, UT1-UTC=-0.1744791, Accumulated Leap Seconds=37
  { 2458661.50, 69.358542 }, //27 June 2019, UT1-UTC=-0.1745418, Accumulated Leap Seconds=37
  { 2458662.50, 69.358553 }, //28 June 2019, UT1-UTC=-0.1745529, Accumulated Leap Seconds=37
  { 2458663.50, 69.358537 }, //29 June 2019, UT1-UTC=-0.1745373, Accumulated Leap Seconds=37
  { 2458664.50, 69.358421 }, //30 June 2019, UT1-UTC=-0.1744209, Accumulated Leap Seconds=37
  { 2458665.50, 69.358172 }, //01 July 2019, UT1-UTC=-0.1741722, Accumulated Leap Seconds=37
  { 2458666.50, 69.357807 }, //02 July 2019, UT1-UTC=-0.1738071, Accumulated Leap Seconds=37
  { 2458667.50, 69.357323 }, //03 July 2019, UT1-UTC=-0.1733232, Accumulated Leap Seconds=37
  { 2458668.50, 69.356772 }, //04 July 2019, UT1-UTC=-0.1727722, Accumulated Leap Seconds=37
  { 2458669.50, 69.356296 }, //05 July 2019, UT1-UTC=-0.1722958, Accumulated Leap Seconds=37
  { 2458670.50, 69.356008 }, //06 July 2019, UT1-UTC=-0.1720076, Accumulated Leap Seconds=37
  { 2458671.50, 69.355934 }, //07 July 2019, UT1-UTC=-0.1719339, Accumulated Leap Seconds=37
  { 2458672.50, 69.355947 }, //08 July 2019, UT1-UTC=-0.1719471, Accumulated Leap Seconds=37
  { 2458673.50, 69.356011 }, //09 July 2019, UT1-UTC=-0.1720110, Accumulated Leap Seconds=37
  { 2458674.50, 69.356063 }, //10 July 2019, UT1-UTC=-0.1720628, Accumulated Leap Seconds=37
  { 2458675.50, 69.356024 }, //11 July 2019, UT1-UTC=-0.1720240, Accumulated Leap Seconds=37
  { 2458676.50, 69.355793 }, //12 July 2019, UT1-UTC=-0.1717933, Accumulated Leap Seconds=37
  { 2458677.50, 69.355341 }, //13 July 2019, UT1-UTC=-0.1713410, Accumulated Leap Seconds=37
  { 2458678.50, 69.354672 }, //14 July 2019, UT1-UTC=-0.1706722, Accumulated Leap Seconds=37
  { 2458679.50, 69.353813 }, //15 July 2019, UT1-UTC=-0.1698133, Accumulated Leap Seconds=37
  { 2458680.50, 69.352858 }, //16 July 2019, UT1-UTC=-0.1688584, Accumulated Leap Seconds=37
  { 2458681.50, 69.351937 }, //17 July 2019, UT1-UTC=-0.1679368, Accumulated Leap Seconds=37
  { 2458682.50, 69.351065 }, //18 July 2019, UT1-UTC=-0.1670647, Accumulated Leap Seconds=37
  { 2458683.50, 69.350290 }, //19 July 2019, UT1-UTC=-0.1662905, Accumulated Leap Seconds=37
  { 2458684.50, 69.349639 }, //20 July 2019, UT1-UTC=-0.1656386, Accumulated Leap Seconds=37
  { 2458685.50, 69.349134 }, //21 July 2019, UT1-UTC=-0.1651336, Accumulated Leap Seconds=37
  { 2458686.50, 69.348655 }, //22 July 2019, UT1-UTC=-0.1646551, Accumulated Leap Seconds=37
  { 2458687.50, 69.348206 }, //23 July 2019, UT1-UTC=-0.1642060, Accumulated Leap Seconds=37
  { 2458688.50, 69.347871 }, //24 July 2019, UT1-UTC=-0.1638710, Accumulated Leap Seconds=37
  { 2458689.50, 69.347555 }, //25 July 2019, UT1-UTC=-0.1635554, Accumulated Leap Seconds=37
  { 2458690.50, 69.347194 }, //26 July 2019, UT1-UTC=-0.1631943, Accumulated Leap Seconds=37
  { 2458691.50, 69.346789 }, //27 July 2019, UT1-UTC=-0.1627890, Accumulated Leap Seconds=37
  { 2458692.50, 69.346366 }, //28 July 2019, UT1-UTC=-0.1623658, Accumulated Leap Seconds=37
  { 2458693.50, 69.345847 }, //29 July 2019, UT1-UTC=-0.1618466, Accumulated Leap Seconds=37
  { 2458694.50, 69.345241 }, //30 July 2019, UT1-UTC=-0.1612414, Accumulated Leap Seconds=37
  { 2458695.50, 69.344654 }, //31 July 2019, UT1-UTC=-0.1606543, Accumulated Leap Seconds=37
  { 2458696.50, 69.344159 }, //01 August 2019, UT1-UTC=-0.1601594, Accumulated Leap Seconds=37
  { 2458697.50, 69.343849 }, //02 August 2019, UT1-UTC=-0.1598491, Accumulated Leap Seconds=37
  { 2458698.50, 69.343766 }, //03 August 2019, UT1-UTC=-0.1597661, Accumulated Leap Seconds=37
  { 2458699.50, 69.343891 }, //04 August 2019, UT1-UTC=-0.1598905, Accumulated Leap Seconds=37
  { 2458700.50, 69.344156 }, //05 August 2019, UT1-UTC=-0.1601557, Accumulated Leap Seconds=37
  { 2458701.50, 69.344447 }, //06 August 2019, UT1-UTC=-0.1604474, Accumulated Leap Seconds=37
  { 2458702.50, 69.344659 }, //07 August 2019, UT1-UTC=-0.1606591, Accumulated Leap Seconds=37
  { 2458703.50, 69.344699 }, //08 August 2019, UT1-UTC=-0.1606992, Accumulated Leap Seconds=37
  { 2458704.50, 69.344514 }, //09 August 2019, UT1-UTC=-0.1605139, Accumulated Leap Seconds=37
  { 2458705.50, 69.344086 }, //10 August 2019, UT1-UTC=-0.1600862, Accumulated Leap Seconds=37
  { 2458706.50, 69.343495 }, //11 August 2019, UT1-UTC=-0.1594952, Accumulated Leap Seconds=37
  { 2458707.50, 69.342817 }, //12 August 2019, UT1-UTC=-0.1588168, Accumulated Leap Seconds=37
  { 2458708.50, 69.342045 }, //13 August 2019, UT1-UTC=-0.1580450, Accumulated Leap Seconds=37
  { 2458709.50, 69.341327 }, //14 August 2019, UT1-UTC=-0.1573268, Accumulated Leap Seconds=37
  { 2458710.50, 69.340743 }, //15 August 2019, UT1-UTC=-0.1567430, Accumulated Leap Seconds=37
  { 2458711.50, 69.340276 }, //16 August 2019, UT1-UTC=-0.1562759, Accumulated Leap Seconds=37
  { 2458712.50, 69.339842 }, //17 August 2019, UT1-UTC=-0.1558423, Accumulated Leap Seconds=37
  { 2458713.50, 69.339503 }, //18 August 2019, UT1-UTC=-0.1555028, Accumulated Leap Seconds=37
  { 2458714.50, 69.339215 }, //19 August 2019, UT1-UTC=-0.1552151, Accumulated Leap Seconds=37
  { 2458715.50, 69.338987 }, //20 August 2019, UT1-UTC=-0.1549871, Accumulated Leap Seconds=37
  { 2458716.50, 69.338821 }, //21 August 2019, UT1-UTC=-0.1548215, Accumulated Leap Seconds=37
  { 2458717.50, 69.338679 }, //22 August 2019, UT1-UTC=-0.1546787, Accumulated Leap Seconds=37
  { 2458718.50, 69.338523 }, //23 August 2019, UT1-UTC=-0.1545227, Accumulated Leap Seconds=37
  { 2458719.50, 69.338351 }, //24 August 2019, UT1-UTC=-0.1543507, Accumulated Leap Seconds=37
  { 2458720.50, 69.338081 }, //25 August 2019, UT1-UTC=-0.1540811, Accumulated Leap Seconds=37
  { 2458721.50, 69.337706 }, //26 August 2019, UT1-UTC=-0.1537064, Accumulated Leap Seconds=37
  { 2458722.50, 69.337305 }, //27 August 2019, UT1-UTC=-0.1533047, Accumulated Leap Seconds=37
  { 2458723.50, 69.336991 }, //28 August 2019, UT1-UTC=-0.1529913, Accumulated Leap Seconds=37
  { 2458724.50, 69.336817 }, //29 August 2019, UT1-UTC=-0.1528174, Accumulated Leap Seconds=37
  { 2458725.50, 69.336861 }, //30 August 2019, UT1-UTC=-0.1528609, Accumulated Leap Seconds=37
  { 2458726.50, 69.337126 }, //31 August 2019, UT1-UTC=-0.1531259, Accumulated Leap Seconds=37
  { 2458727.50, 69.337633 }, //01 September 2019, UT1-UTC=-0.1536329, Accumulated Leap Seconds=37
  { 2458728.50, 69.338265 }, //02 September 2019, UT1-UTC=-0.1542651, Accumulated Leap Seconds=37
  { 2458729.50, 69.338814 }, //03 September 2019, UT1-UTC=-0.1548143, Accumulated Leap Seconds=37
  { 2458730.50, 69.339264 }, //04 September 2019, UT1-UTC=-0.1552635, Accumulated Leap Seconds=37
  { 2458731.50, 69.339524 }, //05 September 2019, UT1-UTC=-0.1555240, Accumulated Leap Seconds=37
  { 2458732.50, 69.339542 }, //06 September 2019, UT1-UTC=-0.1555421, Accumulated Leap Seconds=37
  { 2458733.50, 69.339368 }, //07 September 2019, UT1-UTC=-0.1553677, Accumulated Leap Seconds=37
  { 2458734.50, 69.339106 }, //08 September 2019, UT1-UTC=-0.1551056, Accumulated Leap Seconds=37
  { 2458735.50, 69.338717 }, //09 September 2019, UT1-UTC=-0.1547169, Accumulated Leap Seconds=37
  { 2458736.50, 69.338301 }, //10 September 2019, UT1-UTC=-0.1543014, Accumulated Leap Seconds=37
  { 2458737.50, 69.337939 }, //11 September 2019, UT1-UTC=-0.1539390, Accumulated Leap Seconds=37
  { 2458738.50, 69.337596 }, //12 September 2019, UT1-UTC=-0.1535956, Accumulated Leap Seconds=37
  { 2458739.50, 69.337369 }, //13 September 2019, UT1-UTC=-0.1533692, Accumulated Leap Seconds=37
  { 2458740.50, 69.337299 }, //14 September 2019, UT1-UTC=-0.1532988, Accumulated Leap Seconds=37
  { 2458741.50, 69.337325 }, //15 September 2019, UT1-UTC=-0.1533251, Accumulated Leap Seconds=37
  { 2458742.50, 69.337337 }, //16 September 2019, UT1-UTC=-0.1533367, Accumulated Leap Seconds=37
  { 2458743.50, 69.337324 }, //17 September 2019, UT1-UTC=-0.1533243, Accumulated Leap Seconds=37
  { 2458744.50, 69.337282 }, //18 September 2019, UT1-UTC=-0.1532823, Accumulated Leap Seconds=37
  { 2458745.50, 69.337151 }, //19 September 2019, UT1-UTC=-0.1531511, Accumulated Leap Seconds=37
  { 2458746.50, 69.336911 }, //20 September 2019, UT1-UTC=-0.1529114, Accumulated Leap Seconds=37
  { 2458747.50, 69.336552 }, //21 September 2019, UT1-UTC=-0.1525524, Accumulated Leap Seconds=37
  { 2458748.50, 69.336108 }, //22 September 2019, UT1-UTC=-0.1521081, Accumulated Leap Seconds=37
  { 2458749.50, 69.335646 }, //23 September 2019, UT1-UTC=-0.1516463, Accumulated Leap Seconds=37
  { 2458750.50, 69.335209 }, //24 September 2019, UT1-UTC=-0.1512088, Accumulated Leap Seconds=37
  { 2458751.50, 69.334915 }, //25 September 2019, UT1-UTC=-0.1509148, Accumulated Leap Seconds=37
  { 2458752.50, 69.334903 }, //26 September 2019, UT1-UTC=-0.1509029, Accumulated Leap Seconds=37
  { 2458753.50, 69.335215 }, //27 September 2019, UT1-UTC=-0.1512152, Accumulated Leap Seconds=37
  { 2458754.50, 69.335802 }, //28 September 2019, UT1-UTC=-0.1518024, Accumulated Leap Seconds=37
  { 2458755.50, 69.336506 }, //29 September 2019, UT1-UTC=-0.1525061, Accumulated Leap Seconds=37
  { 2458756.50, 69.337197 }, //30 September 2019, UT1-UTC=-0.1531969, Accumulated Leap Seconds=37
  { 2458757.50, 69.337742 }, //01 October 2019, UT1-UTC=-0.1537424, Accumulated Leap Seconds=37
  { 2458758.50, 69.337982 }, //02 October 2019, UT1-UTC=-0.1539819, Accumulated Leap Seconds=37
  { 2458759.50, 69.337930 }, //03 October 2019, UT1-UTC=-0.1539302, Accumulated Leap Seconds=37
  { 2458760.50, 69.337687 }, //04 October 2019, UT1-UTC=-0.1536867, Accumulated Leap Seconds=37
  { 2458761.50, 69.337299 }, //05 October 2019, UT1-UTC=-0.1532989, Accumulated Leap Seconds=37
  { 2458762.50, 69.336834 }, //06 October 2019, UT1-UTC=-0.1528341, Accumulated Leap Seconds=37
  { 2458763.50, 69.336374 }, //07 October 2019, UT1-UTC=-0.1523745, Accumulated Leap Seconds=37
  { 2458764.50, 69.335972 }, //08 October 2019, UT1-UTC=-0.1519718, Accumulated Leap Seconds=37
  { 2458765.50, 69.335646 }, //09 October 2019, UT1-UTC=-0.1516461, Accumulated Leap Seconds=37
  { 2458766.50, 69.335433 }, //10 October 2019, UT1-UTC=-0.1514326, Accumulated Leap Seconds=37
  { 2458767.50, 69.335374 }, //11 October 2019, UT1-UTC=-0.1513740, Accumulated Leap Seconds=37
  { 2458768.50, 69.335478 }, //12 October 2019, UT1-UTC=-0.1514780, Accumulated Leap Seconds=37
  { 2458769.50, 69.335673 }, //13 October 2019, UT1-UTC=-0.1516730, Accumulated Leap Seconds=37
  { 2458770.50, 69.335935 }, //14 October 2019, UT1-UTC=-0.1519347, Accumulated Leap Seconds=37
  { 2458771.50, 69.336240 }, //15 October 2019, UT1-UTC=-0.1522400, Accumulated Leap Seconds=37
  { 2458772.50, 69.336573 }, //16 October 2019, UT1-UTC=-0.1525734, Accumulated Leap Seconds=37
  { 2458773.50, 69.336884 }, //17 October 2019, UT1-UTC=-0.1528842, Accumulated Leap Seconds=37
  { 2458774.50, 69.337135 }, //18 October 2019, UT1-UTC=-0.1531350, Accumulated Leap Seconds=37
  { 2458775.50, 69.337321 }, //19 October 2019, UT1-UTC=-0.1533206, Accumulated Leap Seconds=37
  { 2458776.50, 69.337419 }, //20 October 2019, UT1-UTC=-0.1534187, Accumulated Leap Seconds=37
  { 2458777.50, 69.337490 }, //21 October 2019, UT1-UTC=-0.1534904, Accumulated Leap Seconds=37
  { 2458778.50, 69.337599 }, //22 October 2019, UT1-UTC=-0.1535986, Accumulated Leap Seconds=37
  { 2458779.50, 69.337818 }, //23 October 2019, UT1-UTC=-0.1538178, Accumulated Leap Seconds=37
  { 2458780.50, 69.338240 }, //24 October 2019, UT1-UTC=-0.1542404, Accumulated Leap Seconds=37
  { 2458781.50, 69.338915 }, //25 October 2019, UT1-UTC=-0.1549155, Accumulated Leap Seconds=37
  { 2458782.50, 69.339816 }, //26 October 2019, UT1-UTC=-0.1558157, Accumulated Leap Seconds=37
  { 2458783.50, 69.340812 }, //27 October 2019, UT1-UTC=-0.1568124, Accumulated Leap Seconds=37
  { 2458784.50, 69.341742 }, //28 October 2019, UT1-UTC=-0.1577418, Accumulated Leap Seconds=37
  { 2458785.50, 69.342452 }, //29 October 2019, UT1-UTC=-0.1584521, Accumulated Leap Seconds=37
  { 2458786.50, 69.342919 }, //30 October 2019, UT1-UTC=-0.1589190, Accumulated Leap Seconds=37
  { 2458787.50, 69.343150 }, //31 October 2019, UT1-UTC=-0.1591500, Accumulated Leap Seconds=37
  { 2458788.50, 69.343219 }, //01 November 2019, UT1-UTC=-0.1592191, Accumulated Leap Seconds=37
  { 2458789.50, 69.343203 }, //02 November 2019, UT1-UTC=-0.1592033, Accumulated Leap Seconds=37
  { 2458790.50, 69.343177 }, //03 November 2019, UT1-UTC=-0.1591771, Accumulated Leap Seconds=37
  { 2458791.50, 69.343217 }, //04 November 2019, UT1-UTC=-0.1592174, Accumulated Leap Seconds=37
  { 2458792.50, 69.343284 }, //05 November 2019, UT1-UTC=-0.1592844, Accumulated Leap Seconds=37
  { 2458793.50, 69.343451 }, //06 November 2019, UT1-UTC=-0.1594509, Accumulated Leap Seconds=37
  { 2458794.50, 69.343776 }, //07 November 2019, UT1-UTC=-0.1597762, Accumulated Leap Seconds=37
  { 2458795.50, 69.344251 }, //08 November 2019, UT1-UTC=-0.1602505, Accumulated Leap Seconds=37
  { 2458796.50, 69.344864 }, //09 November 2019, UT1-UTC=-0.1608641, Accumulated Leap Seconds=37
  { 2458797.50, 69.345559 }, //10 November 2019, UT1-UTC=-0.1615591, Accumulated Leap Seconds=37
  { 2458798.50, 69.346267 }, //11 November 2019, UT1-UTC=-0.1622667, Accumulated Leap Seconds=37
  { 2458799.50, 69.346901 }, //12 November 2019, UT1-UTC=-0.1629007, Accumulated Leap Seconds=37
  { 2458800.50, 69.347435 }, //13 November 2019, UT1-UTC=-0.1634352, Accumulated Leap Seconds=37
  { 2458801.50, 69.347865 }, //14 November 2019, UT1-UTC=-0.1638651, Accumulated Leap Seconds=37
  { 2458802.50, 69.348175 }, //15 November 2019, UT1-UTC=-0.1641749, Accumulated Leap Seconds=37
  { 2458803.50, 69.348388 }, //16 November 2019, UT1-UTC=-0.1643885, Accumulated Leap Seconds=37
  { 2458804.50, 69.348470 }, //17 November 2019, UT1-UTC=-0.1644699, Accumulated Leap Seconds=37
  { 2458805.50, 69.348518 }, //18 November 2019, UT1-UTC=-0.1645181, Accumulated Leap Seconds=37
  { 2458806.50, 69.348658 }, //19 November 2019, UT1-UTC=-0.1646584, Accumulated Leap Seconds=37
  { 2458807.50, 69.348964 }, //20 November 2019, UT1-UTC=-0.1649642, Accumulated Leap Seconds=37
  { 2458808.50, 69.349553 }, //21 November 2019, UT1-UTC=-0.1655534, Accumulated Leap Seconds=37
  { 2458809.50, 69.350340 }, //22 November 2019, UT1-UTC=-0.1663395, Accumulated Leap Seconds=37
  { 2458810.50, 69.351211 }, //23 November 2019, UT1-UTC=-0.1672110, Accumulated Leap Seconds=37
  { 2458811.50, 69.352096 }, //24 November 2019, UT1-UTC=-0.1680955, Accumulated Leap Seconds=37
  { 2458812.50, 69.352901 }, //25 November 2019, UT1-UTC=-0.1689013, Accumulated Leap Seconds=37
  { 2458813.50, 69.353475 }, //26 November 2019, UT1-UTC=-0.1694753, Accumulated Leap Seconds=37
  { 2458814.50, 69.353787 }, //27 November 2019, UT1-UTC=-0.1697867, Accumulated Leap Seconds=37
  { 2458815.50, 69.353946 }, //28 November 2019, UT1-UTC=-0.1699463, Accumulated Leap Seconds=37
  { 2458816.50, 69.354020 }, //29 November 2019, UT1-UTC=-0.1700196, Accumulated Leap Seconds=37
  { 2458817.50, 69.354032 }, //30 November 2019, UT1-UTC=-0.1700320, Accumulated Leap Seconds=37
  { 2458818.50, 69.354014 }, //01 December 2019, UT1-UTC=-0.1700144, Accumulated Leap Seconds=37
  { 2458819.50, 69.354067 }, //02 December 2019, UT1-UTC=-0.1700672, Accumulated Leap Seconds=37
  { 2458820.50, 69.354205 }, //03 December 2019, UT1-UTC=-0.1702051, Accumulated Leap Seconds=37
  { 2458821.50, 69.354424 }, //04 December 2019, UT1-UTC=-0.1704239, Accumulated Leap Seconds=37
  { 2458822.50, 69.354692 }, //05 December 2019, UT1-UTC=-0.1706918, Accumulated Leap Seconds=37
  { 2458823.50, 69.355027 }, //06 December 2019, UT1-UTC=-0.1710269, Accumulated Leap Seconds=37
  { 2458824.50, 69.355423 }, //07 December 2019, UT1-UTC=-0.1714231, Accumulated Leap Seconds=37
  { 2458825.50, 69.355864 }, //08 December 2019, UT1-UTC=-0.1718637, Accumulated Leap Seconds=37
  { 2458826.50, 69.356268 }, //09 December 2019, UT1-UTC=-0.1722681, Accumulated Leap Seconds=37
  { 2458827.50, 69.356535 }, //10 December 2019, UT1-UTC=-0.1725355, Accumulated Leap Seconds=37
  { 2458828.50, 69.356624 }, //11 December 2019, UT1-UTC=-0.1726237, Accumulated Leap Seconds=37
  { 2458829.50, 69.356599 }, //12 December 2019, UT1-UTC=-0.1725987, Accumulated Leap Seconds=37
  { 2458830.50, 69.356460 }, //13 December 2019, UT1-UTC=-0.1724603, Accumulated Leap Seconds=37
  { 2458831.50, 69.356275 }, //14 December 2019, UT1-UTC=-0.1722749, Accumulated Leap Seconds=37
  { 2458832.50, 69.356102 }, //15 December 2019, UT1-UTC=-0.1721025, Accumulated Leap Seconds=37
  { 2458833.50, 69.355998 }, //16 December 2019, UT1-UTC=-0.1719983, Accumulated Leap Seconds=37
  { 2458834.50, 69.356034 }, //17 December 2019, UT1-UTC=-0.1720340, Accumulated Leap Seconds=37
  { 2458835.50, 69.356248 }, //18 December 2019, UT1-UTC=-0.1722479, Accumulated Leap Seconds=37
  { 2458836.50, 69.356718 }, //19 December 2019, UT1-UTC=-0.1727185, Accumulated Leap Seconds=37
  { 2458837.50, 69.357355 }, //20 December 2019, UT1-UTC=-0.1733550, Accumulated Leap Seconds=37
  { 2458838.50, 69.358038 }, //21 December 2019, UT1-UTC=-0.1740384, Accumulated Leap Seconds=37
  { 2458839.50, 69.358719 }, //22 December 2019, UT1-UTC=-0.1747194, Accumulated Leap Seconds=37
  { 2458840.50, 69.359312 }, //23 December 2019, UT1-UTC=-0.1753123, Accumulated Leap Seconds=37
  { 2458841.50, 69.359757 }, //24 December 2019, UT1-UTC=-0.1757574, Accumulated Leap Seconds=37
  { 2458842.50, 69.360024 }, //25 December 2019, UT1-UTC=-0.1760245, Accumulated Leap Seconds=37
  { 2458843.50, 69.360146 }, //26 December 2019, UT1-UTC=-0.1761465, Accumulated Leap Seconds=37
  { 2458844.50, 69.360209 }, //27 December 2019, UT1-UTC=-0.1762087, Accumulated Leap Seconds=37
  { 2458845.50, 69.360251 }, //28 December 2019, UT1-UTC=-0.1762508, Accumulated Leap Seconds=37
  { 2458846.50, 69.360326 }, //29 December 2019, UT1-UTC=-0.1763259, Accumulated Leap Seconds=37
  { 2458847.50, 69.360492 }, //30 December 2019, UT1-UTC=-0.1764922, Accumulated Leap Seconds=37
  { 2458848.50, 69.360769 }, //31 December 2019, UT1-UTC=-0.1767688, Accumulated Leap Seconds=37
  { 2458849.50, 69.361155 }, //01 January 2020, UT1-UTC=-0.1771554, Accumulated Leap Seconds=37
  { 2458850.50, 69.361627 }, //02 January 2020, UT1-UTC=-0.1776274, Accumulated Leap Seconds=37
  { 2458851.50, 69.362115 }, //03 January 2020, UT1-UTC=-0.1781148, Accumulated Leap Seconds=37
  { 2458852.50, 69.362589 }, //04 January 2020, UT1-UTC=-0.1785885, Accumulated Leap Seconds=37
  { 2458853.50, 69.363082 }, //05 January 2020, UT1-UTC=-0.1790818, Accumulated Leap Seconds=37
  { 2458854.50, 69.363428 }, //06 January 2020, UT1-UTC=-0.1794276, Accumulated Leap Seconds=37
  { 2458855.50, 69.363551 }, //07 January 2020, UT1-UTC=-0.1795514, Accumulated Leap Seconds=37
  { 2458856.50, 69.363530 }, //08 January 2020, UT1-UTC=-0.1795302, Accumulated Leap Seconds=37
  { 2458857.50, 69.363375 }, //09 January 2020, UT1-UTC=-0.1793750, Accumulated Leap Seconds=37
  { 2458858.50, 69.363117 }, //10 January 2020, UT1-UTC=-0.1791174, Accumulated Leap Seconds=37
  { 2458859.50, 69.362879 }, //11 January 2020, UT1-UTC=-0.1788786, Accumulated Leap Seconds=37
  { 2458860.50, 69.362790 }, //12 January 2020, UT1-UTC=-0.1787902, Accumulated Leap Seconds=37
  { 2458861.50, 69.362938 }, //13 January 2020, UT1-UTC=-0.1789375, Accumulated Leap Seconds=37
  { 2458862.50, 69.363328 }, //14 January 2020, UT1-UTC=-0.1793277, Accumulated Leap Seconds=37
  { 2458863.50, 69.363964 }, //15 January 2020, UT1-UTC=-0.1799645, Accumulated Leap Seconds=37
  { 2458864.50, 69.364793 }, //16 January 2020, UT1-UTC=-0.1807929, Accumulated Leap Seconds=37
  { 2458865.50, 69.365743 }, //17 January 2020, UT1-UTC=-0.1817427, Accumulated Leap Seconds=37
  { 2458866.50, 69.366785 }, //18 January 2020, UT1-UTC=-0.1827849, Accumulated Leap Seconds=37
  { 2458867.50, 69.367757 }, //19 January 2020, UT1-UTC=-0.1837574, Accumulated Leap Seconds=37
  { 2458868.50, 69.368602 }, //20 January 2020, UT1-UTC=-0.1846024, Accumulated Leap Seconds=37
  { 2458869.50, 69.369284 }, //21 January 2020, UT1-UTC=-0.1852838, Accumulated Leap Seconds=37
  { 2458870.50, 69.369801 }, //22 January 2020, UT1-UTC=-0.1858012, Accumulated Leap Seconds=37
  { 2458871.50, 69.370164 }, //23 January 2020, UT1-UTC=-0.1861639, Accumulated Leap Seconds=37
  { 2458872.50, 69.370469 }, //24 January 2020, UT1-UTC=-0.1864690, Accumulated Leap Seconds=37
  { 2458873.50, 69.370784 }, //25 January 2020, UT1-UTC=-0.1867835, Accumulated Leap Seconds=37
  { 2458874.50, 69.371192 }, //26 January 2020, UT1-UTC=-0.1871917, Accumulated Leap Seconds=37
  { 2458875.50, 69.371708 }, //27 January 2020, UT1-UTC=-0.1877080, Accumulated Leap Seconds=37
  { 2458876.50, 69.372327 }, //28 January 2020, UT1-UTC=-0.1883268, Accumulated Leap Seconds=37
  { 2458877.50, 69.373052 }, //29 January 2020, UT1-UTC=-0.1890522, Accumulated Leap Seconds=37
  { 2458878.50, 69.373813 }, //30 January 2020, UT1-UTC=-0.1898134, Accumulated Leap Seconds=37
  { 2458879.50, 69.374520 }, //31 January 2020, UT1-UTC=-0.1905200, Accumulated Leap Seconds=37
  { 2458880.50, 69.375170 }, //01 February 2020, UT1-UTC=-0.1911703, Accumulated Leap Seconds=37
  { 2458881.50, 69.375720 }, //02 February 2020, UT1-UTC=-0.1917202, Accumulated Leap Seconds=37
  { 2458882.50, 69.376187 }, //03 February 2020, UT1-UTC=-0.1921865, Accumulated Leap Seconds=37
  { 2458883.50, 69.376514 }, //04 February 2020, UT1-UTC=-0.1925140, Accumulated Leap Seconds=37
  { 2458884.50, 69.376715 }, //05 February 2020, UT1-UTC=-0.1927151, Accumulated Leap Seconds=37
  { 2458885.50, 69.376845 }, //06 February 2020, UT1-UTC=-0.1928451, Accumulated Leap Seconds=37
  { 2458886.50, 69.376930 }, //07 February 2020, UT1-UTC=-0.1929304, Accumulated Leap Seconds=37
  { 2458887.50, 69.377019 }, //08 February 2020, UT1-UTC=-0.1930187, Accumulated Leap Seconds=37
  { 2458888.50, 69.377268 }, //09 February 2020, UT1-UTC=-0.1932681, Accumulated Leap Seconds=37
  { 2458889.50, 69.377793 }, //10 February 2020, UT1-UTC=-0.1937929, Accumulated Leap Seconds=37
  { 2458890.50, 69.378581 }, //11 February 2020, UT1-UTC=-0.1945811, Accumulated Leap Seconds=37
  { 2458891.50, 69.379552 }, //12 February 2020, UT1-UTC=-0.1955524, Accumulated Leap Seconds=37
  { 2458892.50, 69.380540 }, //13 February 2020, UT1-UTC=-0.1965396, Accumulated Leap Seconds=37
  { 2458893.50, 69.381428 }, //14 February 2020, UT1-UTC=-0.1974284, Accumulated Leap Seconds=37
  { 2458894.50, 69.382113 }, //15 February 2020, UT1-UTC=-0.1981127, Accumulated Leap Seconds=37
  { 2458895.50, 69.382629 }, //16 February 2020, UT1-UTC=-0.1986289, Accumulated Leap Seconds=37
  { 2458896.50, 69.382940 }, //17 February 2020, UT1-UTC=-0.1989401, Accumulated Leap Seconds=37
  { 2458897.50, 69.383076 }, //18 February 2020, UT1-UTC=-0.1990761, Accumulated Leap Seconds=37
  { 2458898.50, 69.383126 }, //19 February 2020, UT1-UTC=-0.1991257, Accumulated Leap Seconds=37
  { 2458899.50, 69.383152 }, //20 February 2020, UT1-UTC=-0.1991524, Accumulated Leap Seconds=37
  { 2458900.50, 69.383237 }, //21 February 2020, UT1-UTC=-0.1992371, Accumulated Leap Seconds=37
  { 2458901.50, 69.383508 }, //22 February 2020, UT1-UTC=-0.1995078, Accumulated Leap Seconds=37
  { 2458902.50, 69.383974 }, //23 February 2020, UT1-UTC=-0.1999742, Accumulated Leap Seconds=37
  { 2458903.50, 69.384575 }, //24 February 2020, UT1-UTC=-0.2005754, Accumulated Leap Seconds=37
  { 2458904.50, 69.385249 }, //25 February 2020, UT1-UTC=-0.2012489, Accumulated Leap Seconds=37
  { 2458905.50, 69.385971 }, //26 February 2020, UT1-UTC=-0.2019712, Accumulated Leap Seconds=37
  { 2458906.50, 69.386736 }, //27 February 2020, UT1-UTC=-0.2027359, Accumulated Leap Seconds=37
  { 2458907.50, 69.387521 }, //28 February 2020, UT1-UTC=-0.2035207, Accumulated Leap Seconds=37
  { 2458908.50, 69.388280 }, //29 February 2020, UT1-UTC=-0.2042796, Accumulated Leap Seconds=37
  { 2458909.50, 69.388990 }, //01 March 2020, UT1-UTC=-0.2049904, Accumulated Leap Seconds=37
  { 2458910.50, 69.389625 }, //02 March 2020, UT1-UTC=-0.2056255, Accumulated Leap Seconds=37
  { 2458911.50, 69.390169 }, //03 March 2020, UT1-UTC=-0.2061689, Accumulated Leap Seconds=37
  { 2458912.50, 69.390666 }, //04 March 2020, UT1-UTC=-0.2066659, Accumulated Leap Seconds=37
  { 2458913.50, 69.391141 }, //05 March 2020, UT1-UTC=-0.2071407, Accumulated Leap Seconds=37
  { 2458914.50, 69.391683 }, //06 March 2020, UT1-UTC=-0.2076825, Accumulated Leap Seconds=37
  { 2458915.50, 69.392381 }, //07 March 2020, UT1-UTC=-0.2083814, Accumulated Leap Seconds=37
  { 2458916.50, 69.393258 }, //08 March 2020, UT1-UTC=-0.2092576, Accumulated Leap Seconds=37
  { 2458917.50, 69.394358 }, //09 March 2020, UT1-UTC=-0.2103578, Accumulated Leap Seconds=37
  { 2458918.50, 69.395715 }, //10 March 2020, UT1-UTC=-0.2117147, Accumulated Leap Seconds=37
  { 2458919.50, 69.397267 }, //11 March 2020, UT1-UTC=-0.2132675, Accumulated Leap Seconds=37
  { 2458920.50, 69.398795 }, //12 March 2020, UT1-UTC=-0.2147950, Accumulated Leap Seconds=37
  { 2458921.50, 69.400140 }, //13 March 2020, UT1-UTC=-0.2161404, Accumulated Leap Seconds=37
  { 2458922.50, 69.401282 }, //14 March 2020, UT1-UTC=-0.2172820, Accumulated Leap Seconds=37
  { 2458923.50, 69.402166 }, //15 March 2020, UT1-UTC=-0.2181664, Accumulated Leap Seconds=37
  { 2458924.50, 69.402794 }, //16 March 2020, UT1-UTC=-0.2187942, Accumulated Leap Seconds=37
  { 2458925.50, 69.403272 }, //17 March 2020, UT1-UTC=-0.2192723, Accumulated Leap Seconds=37
  { 2458926.50, 69.403647 }, //18 March 2020, UT1-UTC=-0.2196470, Accumulated Leap Seconds=37
  { 2458927.50, 69.403942 }, //19 March 2020, UT1-UTC=-0.2199424, Accumulated Leap Seconds=37
  { 2458928.50, 69.404242 }, //20 March 2020, UT1-UTC=-0.2202417, Accumulated Leap Seconds=37
  { 2458929.50, 69.404585 }, //21 March 2020, UT1-UTC=-0.2205851, Accumulated Leap Seconds=37
  { 2458930.50, 69.404949 }, //22 March 2020, UT1-UTC=-0.2209490, Accumulated Leap Seconds=37
  { 2458931.50, 69.405378 }, //23 March 2020, UT1-UTC=-0.2213784, Accumulated Leap Seconds=37
  { 2458932.50, 69.405931 }, //24 March 2020, UT1-UTC=-0.2219310, Accumulated Leap Seconds=37
  { 2458933.50, 69.406494 }, //25 March 2020, UT1-UTC=-0.2224940, Accumulated Leap Seconds=37
  { 2458934.50, 69.407053 }, //26 March 2020, UT1-UTC=-0.2230530, Accumulated Leap Seconds=37
  { 2458935.50, 69.407573 }, //27 March 2020, UT1-UTC=-0.2235729, Accumulated Leap Seconds=37
  { 2458936.50, 69.408033 }, //28 March 2020, UT1-UTC=-0.2240333, Accumulated Leap Seconds=37
  { 2458937.50, 69.408433 }, //29 March 2020, UT1-UTC=-0.2244327, Accumulated Leap Seconds=37
  { 2458938.50, 69.408710 }, //30 March 2020, UT1-UTC=-0.2247096, Accumulated Leap Seconds=37
  { 2458939.50, 69.408943 }, //31 March 2020, UT1-UTC=-0.2249429, Accumulated Leap Seconds=37
  { 2458940.50, 69.409164 }, //01 April 2020, UT1-UTC=-0.2251639, Accumulated Leap Seconds=37
  { 2458941.50, 69.409438 }, //02 April 2020, UT1-UTC=-0.2254376, Accumulated Leap Seconds=37
  { 2458942.50, 69.409815 }, //03 April 2020, UT1-UTC=-0.2258149, Accumulated Leap Seconds=37
  { 2458943.50, 69.410426 }, //04 April 2020, UT1-UTC=-0.2264256, Accumulated Leap Seconds=37
  { 2458944.50, 69.411327 }, //05 April 2020, UT1-UTC=-0.2273265, Accumulated Leap Seconds=37
  { 2458945.50, 69.412545 }, //06 April 2020, UT1-UTC=-0.2285447, Accumulated Leap Seconds=37
  { 2458946.50, 69.414009 }, //07 April 2020, UT1-UTC=-0.2300093, Accumulated Leap Seconds=37
  { 2458947.50, 69.415621 }, //08 April 2020, UT1-UTC=-0.2316212, Accumulated Leap Seconds=37
  { 2458948.50, 69.417204 }, //09 April 2020, UT1-UTC=-0.2332043, Accumulated Leap Seconds=37
  { 2458949.50, 69.418632 }, //10 April 2020, UT1-UTC=-0.2346315, Accumulated Leap Seconds=37
  { 2458950.50, 69.419776 }, //11 April 2020, UT1-UTC=-0.2357756, Accumulated Leap Seconds=37
  { 2458951.50, 69.420627 }, //12 April 2020, UT1-UTC=-0.2366269, Accumulated Leap Seconds=37
  { 2458952.50, 69.421231 }, //13 April 2020, UT1-UTC=-0.2372315, Accumulated Leap Seconds=37
  { 2458953.50, 69.421621 }, //14 April 2020, UT1-UTC=-0.2376207, Accumulated Leap Seconds=37
  { 2458954.50, 69.421856 }, //15 April 2020, UT1-UTC=-0.2378557, Accumulated Leap Seconds=37
  { 2458955.50, 69.422017 }, //16 April 2020, UT1-UTC=-0.2380168, Accumulated Leap Seconds=37
  { 2458956.50, 69.422233 }, //17 April 2020, UT1-UTC=-0.2382326, Accumulated Leap Seconds=37
  { 2458957.50, 69.422573 }, //18 April 2020, UT1-UTC=-0.2385729, Accumulated Leap Seconds=37
  { 2458958.50, 69.423014 }, //19 April 2020, UT1-UTC=-0.2390137, Accumulated Leap Seconds=37
  { 2458959.50, 69.423515 }, //20 April 2020, UT1-UTC=-0.2395153, Accumulated Leap Seconds=37
  { 2458960.50, 69.423995 }, //21 April 2020, UT1-UTC=-0.2399946, Accumulated Leap Seconds=37
  { 2458961.50, 69.424481 }, //22 April 2020, UT1-UTC=-0.2404811, Accumulated Leap Seconds=37
  { 2458962.50, 69.424961 }, //23 April 2020, UT1-UTC=-0.2409606, Accumulated Leap Seconds=37
  { 2458963.50, 69.425370 }, //24 April 2020, UT1-UTC=-0.2413705, Accumulated Leap Seconds=37
  { 2458964.50, 69.425680 }, //25 April 2020, UT1-UTC=-0.2416796, Accumulated Leap Seconds=37
  { 2458965.50, 69.425868 }, //26 April 2020, UT1-UTC=-0.2418681, Accumulated Leap Seconds=37
  { 2458966.50, 69.425977 }, //27 April 2020, UT1-UTC=-0.2419768, Accumulated Leap Seconds=37
  { 2458967.50, 69.426069 }, //28 April 2020, UT1-UTC=-0.2420687, Accumulated Leap Seconds=37
  { 2458968.50, 69.426135 }, //29 April 2020, UT1-UTC=-0.2421354, Accumulated Leap Seconds=37
  { 2458969.50, 69.426237 }, //30 April 2020, UT1-UTC=-0.2422373, Accumulated Leap Seconds=37
  { 2458970.50, 69.426466 }, //01 May 2020, UT1-UTC=-0.2424662, Accumulated Leap Seconds=37
  { 2458971.50, 69.426919 }, //02 May 2020, UT1-UTC=-0.2429192, Accumulated Leap Seconds=37
  { 2458972.50, 69.427644 }, //03 May 2020, UT1-UTC=-0.2436439, Accumulated Leap Seconds=37
  { 2458973.50, 69.428598 }, //04 May 2020, UT1-UTC=-0.2445982, Accumulated Leap Seconds=37
  { 2458974.50, 69.429674 }, //05 May 2020, UT1-UTC=-0.2456736, Accumulated Leap Seconds=37
  { 2458975.50, 69.430782 }, //06 May 2020, UT1-UTC=-0.2467817, Accumulated Leap Seconds=37
  { 2458976.50, 69.431845 }, //07 May 2020, UT1-UTC=-0.2478452, Accumulated Leap Seconds=37
  { 2458977.50, 69.432699 }, //08 May 2020, UT1-UTC=-0.2486991, Accumulated Leap Seconds=37
  { 2458978.50, 69.433290 }, //09 May 2020, UT1-UTC=-0.2492900, Accumulated Leap Seconds=37
  { 2458979.50, 69.433739 }, //10 May 2020, UT1-UTC=-0.2497392, Accumulated Leap Seconds=37
  { 2458980.50, 69.433995 }, //11 May 2020, UT1-UTC=-0.2499950, Accumulated Leap Seconds=37
  { 2458981.50, 69.434153 }, //12 May 2020, UT1-UTC=-0.2501527, Accumulated Leap Seconds=37
  { 2458982.50, 69.434330 }, //13 May 2020, UT1-UTC=-0.2503303, Accumulated Leap Seconds=37
  { 2458983.50, 69.434535 }, //14 May 2020, UT1-UTC=-0.2505353, Accumulated Leap Seconds=37
  { 2458984.50, 69.434752 }, //15 May 2020, UT1-UTC=-0.2507515, Accumulated Leap Seconds=37
  { 2458985.50, 69.435076 }, //16 May 2020, UT1-UTC=-0.2510759, Accumulated Leap Seconds=37
  { 2458986.50, 69.435486 }, //17 May 2020, UT1-UTC=-0.2514862, Accumulated Leap Seconds=37
  { 2458987.50, 69.436004 }, //18 May 2020, UT1-UTC=-0.2520039, Accumulated Leap Seconds=37
  { 2458988.50, 69.436609 }, //19 May 2020, UT1-UTC=-0.2526091, Accumulated Leap Seconds=37
  { 2458989.50, 69.437142 }, //20 May 2020, UT1-UTC=-0.2531420, Accumulated Leap Seconds=37
  { 2458990.50, 69.437620 }, //21 May 2020, UT1-UTC=-0.2536200, Accumulated Leap Seconds=37
  { 2458991.50, 69.437967 }, //22 May 2020, UT1-UTC=-0.2539670, Accumulated Leap Seconds=37
  { 2458992.50, 69.438144 }, //23 May 2020, UT1-UTC=-0.2541436, Accumulated Leap Seconds=37
  { 2458993.50, 69.438153 }, //24 May 2020, UT1-UTC=-0.2541533, Accumulated Leap Seconds=37
  { 2458994.50, 69.438012 }, //25 May 2020, UT1-UTC=-0.2540118, Accumulated Leap Seconds=37
  { 2458995.50, 69.437832 }, //26 May 2020, UT1-UTC=-0.2538321, Accumulated Leap Seconds=37
  { 2458996.50, 69.437679 }, //27 May 2020, UT1-UTC=-0.2536790, Accumulated Leap Seconds=37
  { 2458997.50, 69.437550 }, //28 May 2020, UT1-UTC=-0.2535495, Accumulated Leap Seconds=37
  { 2458998.50, 69.437550 }, //29 May 2020, UT1-UTC=-0.2535504, Accumulated Leap Seconds=37
  { 2458999.50, 69.437714 }, //30 May 2020, UT1-UTC=-0.2537140, Accumulated Leap Seconds=37
  { 2459000.50, 69.438087 }, //31 May 2020, UT1-UTC=-0.2540873, Accumulated Leap Seconds=37
  { 2459001.50, 69.438633 }, //01 June 2020, UT1-UTC=-0.2546335, Accumulated Leap Seconds=37
  { 2459002.50, 69.439252 }, //02 June 2020, UT1-UTC=-0.2552518, Accumulated Leap Seconds=37
  { 2459003.50, 69.439800 }, //03 June 2020, UT1-UTC=-0.2558002, Accumulated Leap Seconds=37
  { 2459004.50, 69.440136 }, //04 June 2020, UT1-UTC=-0.2561358, Accumulated Leap Seconds=37
  { 2459005.50, 69.440208 }, //05 June 2020, UT1-UTC=-0.2562081, Accumulated Leap Seconds=37
  { 2459006.50, 69.440053 }, //06 June 2020, UT1-UTC=-0.2560530, Accumulated Leap Seconds=37
  { 2459007.50, 69.439689 }, //07 June 2020, UT1-UTC=-0.2556889, Accumulated Leap Seconds=37
  { 2459008.50, 69.439185 }, //08 June 2020, UT1-UTC=-0.2551855, Accumulated Leap Seconds=37
  { 2459009.50, 69.438626 }, //09 June 2020, UT1-UTC=-0.2546259, Accumulated Leap Seconds=37
  { 2459010.50, 69.437981 }, //10 June 2020, UT1-UTC=-0.2539811, Accumulated Leap Seconds=37
  { 2459011.50, 69.437376 }, //11 June 2020, UT1-UTC=-0.2533761, Accumulated Leap Seconds=37
  { 2459012.50, 69.436731 }, //12 June 2020, UT1-UTC=-0.2527310, Accumulated Leap Seconds=37
  { 2459013.50, 69.436098 }, //13 June 2020, UT1-UTC=-0.2520982, Accumulated Leap Seconds=37
  { 2459014.50, 69.435564 }, //14 June 2020, UT1-UTC=-0.2515642, Accumulated Leap Seconds=37
  { 2459015.50, 69.435133 }, //15 June 2020, UT1-UTC=-0.2511335, Accumulated Leap Seconds=37
  { 2459016.50, 69.434749 }, //16 June 2020, UT1-UTC=-0.2507493, Accumulated Leap Seconds=37
  { 2459017.50, 69.434280 }, //17 June 2020, UT1-UTC=-0.2502801, Accumulated Leap Seconds=37
  { 2459018.50, 69.433709 }, //18 June 2020, UT1-UTC=-0.2497090, Accumulated Leap Seconds=37
  { 2459019.50, 69.433032 }, //19 June 2020, UT1-UTC=-0.2490317, Accumulated Leap Seconds=37
  { 2459020.50, 69.432102 }, //20 June 2020, UT1-UTC=-0.2481022, Accumulated Leap Seconds=37
  { 2459021.50, 69.431045 }, //21 June 2020, UT1-UTC=-0.2470448, Accumulated Leap Seconds=37
  { 2459022.50, 69.429886 }, //22 June 2020, UT1-UTC=-0.2458865, Accumulated Leap Seconds=37
  { 2459023.50, 69.428702 }, //23 June 2020, UT1-UTC=-0.2447016, Accumulated Leap Seconds=37
  { 2459024.50, 69.427573 }, //24 June 2020, UT1-UTC=-0.2435726, Accumulated Leap Seconds=37
  { 2459025.50, 69.426600 }, //25 June 2020, UT1-UTC=-0.2426000, Accumulated Leap Seconds=37
  { 2459026.50, 69.425866 }, //26 June 2020, UT1-UTC=-0.2418664, Accumulated Leap Seconds=37
  { 2459027.50, 69.425362 }, //27 June 2020, UT1-UTC=-0.2413618, Accumulated Leap Seconds=37
  { 2459028.50, 69.425025 }, //28 June 2020, UT1-UTC=-0.2410246, Accumulated Leap Seconds=37
  { 2459029.50, 69.424810 }, //29 June 2020, UT1-UTC=-0.2408104, Accumulated Leap Seconds=37
  { 2459030.50, 69.424568 }, //30 June 2020, UT1-UTC=-0.2405680, Accumulated Leap Seconds=37
  { 2459031.50, 69.424133 }, //01 July 2020, UT1-UTC=-0.2401335, Accumulated Leap Seconds=37
  { 2459032.50, 69.423485 }, //02 July 2020, UT1-UTC=-0.2394847, Accumulated Leap Seconds=37
  { 2459033.50, 69.422600 }, //03 July 2020, UT1-UTC=-0.2385999, Accumulated Leap Seconds=37
  { 2459034.50, 69.421499 }, //04 July 2020, UT1-UTC=-0.2374993, Accumulated Leap Seconds=37
  { 2459035.50, 69.420218 }, //05 July 2020, UT1-UTC=-0.2362178, Accumulated Leap Seconds=37
  { 2459036.50, 69.418878 }, //06 July 2020, UT1-UTC=-0.2348785, Accumulated Leap Seconds=37
  { 2459037.50, 69.417601 }, //07 July 2020, UT1-UTC=-0.2336011, Accumulated Leap Seconds=37
  { 2459038.50, 69.416395 }, //08 July 2020, UT1-UTC=-0.2323948, Accumulated Leap Seconds=37
  { 2459039.50, 69.415266 }, //09 July 2020, UT1-UTC=-0.2312661, Accumulated Leap Seconds=37
  { 2459040.50, 69.414208 }, //10 July 2020, UT1-UTC=-0.2302077, Accumulated Leap Seconds=37
  { 2459041.50, 69.413242 }, //11 July 2020, UT1-UTC=-0.2292418, Accumulated Leap Seconds=37
  { 2459042.50, 69.412237 }, //12 July 2020, UT1-UTC=-0.2282373, Accumulated Leap Seconds=37
  { 2459043.50, 69.411192 }, //13 July 2020, UT1-UTC=-0.2271922, Accumulated Leap Seconds=37
  { 2459044.50, 69.410120 }, //14 July 2020, UT1-UTC=-0.2261196, Accumulated Leap Seconds=37
  { 2459045.50, 69.408965 }, //15 July 2020, UT1-UTC=-0.2249647, Accumulated Leap Seconds=37
  { 2459046.50, 69.407756 }, //16 July 2020, UT1-UTC=-0.2237563, Accumulated Leap Seconds=37
  { 2459047.50, 69.406474 }, //17 July 2020, UT1-UTC=-0.2224735, Accumulated Leap Seconds=37
  { 2459048.50, 69.405122 }, //18 July 2020, UT1-UTC=-0.2211221, Accumulated Leap Seconds=37
  { 2459049.50, 69.403697 }, //19 July 2020, UT1-UTC=-0.2196966, Accumulated Leap Seconds=37
  { 2459050.50, 69.402223 }, //20 July 2020, UT1-UTC=-0.2182229, Accumulated Leap Seconds=37
  { 2459051.50, 69.400857 }, //21 July 2020, UT1-UTC=-0.2168569, Accumulated Leap Seconds=37
  { 2459052.50, 69.399693 }, //22 July 2020, UT1-UTC=-0.2156934, Accumulated Leap Seconds=37
  { 2459053.50, 69.398740 }, //23 July 2020, UT1-UTC=-0.2147397, Accumulated Leap Seconds=37
  { 2459054.50, 69.397989 }, //24 July 2020, UT1-UTC=-0.2139895, Accumulated Leap Seconds=37
  { 2459055.50, 69.397428 }, //25 July 2020, UT1-UTC=-0.2134277, Accumulated Leap Seconds=37
  { 2459056.50, 69.396964 }, //26 July 2020, UT1-UTC=-0.2129642, Accumulated Leap Seconds=37
  { 2459057.50, 69.396451 }, //27 July 2020, UT1-UTC=-0.2124511, Accumulated Leap Seconds=37
  { 2459058.50, 69.395825 }, //28 July 2020, UT1-UTC=-0.2118250, Accumulated Leap Seconds=37
  { 2459059.50, 69.395086 }, //29 July 2020, UT1-UTC=-0.2110863, Accumulated Leap Seconds=37
  { 2459060.50, 69.394185 }, //30 July 2020, UT1-UTC=-0.2101852, Accumulated Leap Seconds=37
  { 2459061.50, 69.393173 }, //31 July 2020, UT1-UTC=-0.2091731, Accumulated Leap Seconds=37
  { 2459062.50, 69.392124 }, //01 August 2020, UT1-UTC=-0.2081241, Accumulated Leap Seconds=37
  { 2459063.50, 69.391052 }, //02 August 2020, UT1-UTC=-0.2070517, Accumulated Leap Seconds=37
  { 2459064.50, 69.390068 }, //03 August 2020, UT1-UTC=-0.2060679, Accumulated Leap Seconds=37
  { 2459065.50, 69.389219 }, //04 August 2020, UT1-UTC=-0.2052187, Accumulated Leap Seconds=37
  { 2459066.50, 69.388463 }, //05 August 2020, UT1-UTC=-0.2044629, Accumulated Leap Seconds=37
  { 2459067.50, 69.387881 }, //06 August 2020, UT1-UTC=-0.2038807, Accumulated Leap Seconds=37
  { 2459068.50, 69.387420 }, //07 August 2020, UT1-UTC=-0.2034204, Accumulated Leap Seconds=37
  { 2459069.50, 69.387041 }, //08 August 2020, UT1-UTC=-0.2030407, Accumulated Leap Seconds=37
  { 2459070.50, 69.386727 }, //09 August 2020, UT1-UTC=-0.2027274, Accumulated Leap Seconds=37
  { 2459071.50, 69.386337 }, //10 August 2020, UT1-UTC=-0.2023368, Accumulated Leap Seconds=37
  { 2459072.50, 69.385832 }, //11 August 2020, UT1-UTC=-0.2018318, Accumulated Leap Seconds=37
  { 2459073.50, 69.385197 }, //12 August 2020, UT1-UTC=-0.2011971, Accumulated Leap Seconds=37
  { 2459074.50, 69.384391 }, //13 August 2020, UT1-UTC=-0.2003908, Accumulated Leap Seconds=37
  { 2459075.50, 69.383493 }, //14 August 2020, UT1-UTC=-0.1994927, Accumulated Leap Seconds=37
  { 2459076.50, 69.382461 }, //15 August 2020, UT1-UTC=-0.1984615, Accumulated Leap Seconds=37
  { 2459077.50, 69.381279 }, //16 August 2020, UT1-UTC=-0.1972792, Accumulated Leap Seconds=37
  { 2459078.50, 69.380107 }, //17 August 2020, UT1-UTC=-0.1961068, Accumulated Leap Seconds=37
  { 2459079.50, 69.379096 }, //18 August 2020, UT1-UTC=-0.1950958, Accumulated Leap Seconds=37
  { 2459080.50, 69.378295 }, //19 August 2020, UT1-UTC=-0.1942946, Accumulated Leap Seconds=37
  { 2459081.50, 69.377665 }, //20 August 2020, UT1-UTC=-0.1936647, Accumulated Leap Seconds=37
  { 2459082.50, 69.377349 }, //21 August 2020, UT1-UTC=-0.1933494, Accumulated Leap Seconds=37
  { 2459083.50, 69.377238 }, //22 August 2020, UT1-UTC=-0.1932377, Accumulated Leap Seconds=37
  { 2459084.50, 69.377178 }, //23 August 2020, UT1-UTC=-0.1931778, Accumulated Leap Seconds=37
  { 2459085.50, 69.377031 }, //24 August 2020, UT1-UTC=-0.1930310, Accumulated Leap Seconds=37
  { 2459086.50, 69.376704 }, //25 August 2020, UT1-UTC=-0.1927039, Accumulated Leap Seconds=37
  { 2459087.50, 69.376095 }, //26 August 2020, UT1-UTC=-0.1920949, Accumulated Leap Seconds=37
  { 2459088.50, 69.375202 }, //27 August 2020, UT1-UTC=-0.1912021, Accumulated Leap Seconds=37
  { 2459089.50, 69.374122 }, //28 August 2020, UT1-UTC=-0.1901225, Accumulated Leap Seconds=37
  { 2459090.50, 69.372891 }, //29 August 2020, UT1-UTC=-0.1888915, Accumulated Leap Seconds=37
  { 2459091.50, 69.371601 }, //30 August 2020, UT1-UTC=-0.1876013, Accumulated Leap Seconds=37
  { 2459092.50, 69.370382 }, //31 August 2020, UT1-UTC=-0.1863822, Accumulated Leap Seconds=37
  { 2459093.50, 69.369342 }, //01 September 2020, UT1-UTC=-0.1853422, Accumulated Leap Seconds=37
  { 2459094.50, 69.368421 }, //02 September 2020, UT1-UTC=-0.1844207, Accumulated Leap Seconds=37
  { 2459095.50, 69.367644 }, //03 September 2020, UT1-UTC=-0.1836441, Accumulated Leap Seconds=37
  { 2459096.50, 69.367008 }, //04 September 2020, UT1-UTC=-0.1830079, Accumulated Leap Seconds=37
  { 2459097.50, 69.366438 }, //05 September 2020, UT1-UTC=-0.1824377, Accumulated Leap Seconds=37
  { 2459098.50, 69.365864 }, //06 September 2020, UT1-UTC=-0.1818637, Accumulated Leap Seconds=37
  { 2459099.50, 69.365304 }, //07 September 2020, UT1-UTC=-0.1813038, Accumulated Leap Seconds=37
  { 2459100.50, 69.364675 }, //08 September 2020, UT1-UTC=-0.1806747, Accumulated Leap Seconds=37
  { 2459101.50, 69.363960 }, //09 September 2020, UT1-UTC=-0.1799600, Accumulated Leap Seconds=37
  { 2459102.50, 69.363144 }, //10 September 2020, UT1-UTC=-0.1791442, Accumulated Leap Seconds=37
  { 2459103.50, 69.362349 }, //11 September 2020, UT1-UTC=-0.1783491, Accumulated Leap Seconds=37
  { 2459104.50, 69.361533 }, //12 September 2020, UT1-UTC=-0.1775331, Accumulated Leap Seconds=37
  { 2459105.50, 69.360742 }, //13 September 2020, UT1-UTC=-0.1767424, Accumulated Leap Seconds=37
  { 2459106.50, 69.360111 }, //14 September 2020, UT1-UTC=-0.1761111, Accumulated Leap Seconds=37
  { 2459107.50, 69.359711 }, //15 September 2020, UT1-UTC=-0.1757107, Accumulated Leap Seconds=37
  { 2459108.50, 69.359555 }, //16 September 2020, UT1-UTC=-0.1755555, Accumulated Leap Seconds=37
  { 2459109.50, 69.359676 }, //17 September 2020, UT1-UTC=-0.1756758, Accumulated Leap Seconds=37
  { 2459110.50, 69.359978 }, //18 September 2020, UT1-UTC=-0.1759784, Accumulated Leap Seconds=37
  { 2459111.50, 69.360403 }, //19 September 2020, UT1-UTC=-0.1764030, Accumulated Leap Seconds=37
  { 2459112.50, 69.360789 }, //20 September 2020, UT1-UTC=-0.1767895, Accumulated Leap Seconds=37
  { 2459113.50, 69.361006 }, //21 September 2020, UT1-UTC=-0.1770063, Accumulated Leap Seconds=37
  { 2459114.50, 69.361019 }, //22 September 2020, UT1-UTC=-0.1770188, Accumulated Leap Seconds=37
  { 2459115.50, 69.360811 }, //23 September 2020, UT1-UTC=-0.1768106, Accumulated Leap Seconds=37
  { 2459116.50, 69.360457 }, //24 September 2020, UT1-UTC=-0.1764569, Accumulated Leap Seconds=37
  { 2459117.50, 69.360008 }, //25 September 2020, UT1-UTC=-0.1760082, Accumulated Leap Seconds=37
  { 2459118.50, 69.359564 }, //26 September 2020, UT1-UTC=-0.1755640, Accumulated Leap Seconds=37
  { 2459119.50, 69.359037 }, //27 September 2020, UT1-UTC=-0.1750367, Accumulated Leap Seconds=37
  { 2459120.50, 69.358483 }, //28 September 2020, UT1-UTC=-0.1744831, Accumulated Leap Seconds=37
  { 2459121.50, 69.358055 }, //29 September 2020, UT1-UTC=-0.1740546, Accumulated Leap Seconds=37
  { 2459122.50, 69.357718 }, //30 September 2020, UT1-UTC=-0.1737183, Accumulated Leap Seconds=37
  { 2459123.50, 69.357478 }, //01 October 2020, UT1-UTC=-0.1734782, Accumulated Leap Seconds=37
  { 2459124.50, 69.357252 }, //02 October 2020, UT1-UTC=-0.1732518, Accumulated Leap Seconds=37
  { 2459125.50, 69.357089 }, //03 October 2020, UT1-UTC=-0.1730890, Accumulated Leap Seconds=37
  { 2459126.50, 69.356952 }, //04 October 2020, UT1-UTC=-0.1729516, Accumulated Leap Seconds=37
  { 2459127.50, 69.356781 }, //05 October 2020, UT1-UTC=-0.1727809, Accumulated Leap Seconds=37
  { 2459128.50, 69.356508 }, //06 October 2020, UT1-UTC=-0.1725084, Accumulated Leap Seconds=37
  { 2459129.50, 69.356119 }, //07 October 2020, UT1-UTC=-0.1721191, Accumulated Leap Seconds=37
  { 2459130.50, 69.355606 }, //08 October 2020, UT1-UTC=-0.1716062, Accumulated Leap Seconds=37
  { 2459131.50, 69.354953 }, //09 October 2020, UT1-UTC=-0.1709530, Accumulated Leap Seconds=37
  { 2459132.50, 69.354345 }, //10 October 2020, UT1-UTC=-0.1703455, Accumulated Leap Seconds=37
  { 2459133.50, 69.353833 }, //11 October 2020, UT1-UTC=-0.1698329, Accumulated Leap Seconds=37
  { 2459134.50, 69.353429 }, //12 October 2020, UT1-UTC=-0.1694286, Accumulated Leap Seconds=37
  { 2459135.50, 69.353290 }, //13 October 2020, UT1-UTC=-0.1692903, Accumulated Leap Seconds=37
  { 2459136.50, 69.353485 }, //14 October 2020, UT1-UTC=-0.1694855, Accumulated Leap Seconds=37
  { 2459137.50, 69.353928 }, //15 October 2020, UT1-UTC=-0.1699282, Accumulated Leap Seconds=37
  { 2459138.50, 69.354637 }, //16 October 2020, UT1-UTC=-0.1706366, Accumulated Leap Seconds=37
  { 2459139.50, 69.355433 }, //17 October 2020, UT1-UTC=-0.1714326, Accumulated Leap Seconds=37
  { 2459140.50, 69.356209 }, //18 October 2020, UT1-UTC=-0.1722094, Accumulated Leap Seconds=37
  { 2459141.50, 69.356737 }, //19 October 2020, UT1-UTC=-0.1727370, Accumulated Leap Seconds=37
  { 2459142.50, 69.357014 }, //20 October 2020, UT1-UTC=-0.1730140, Accumulated Leap Seconds=37
  { 2459143.50, 69.357091 }, //21 October 2020, UT1-UTC=-0.1730915, Accumulated Leap Seconds=37
  { 2459144.50, 69.357076 }, //22 October 2020, UT1-UTC=-0.1730757, Accumulated Leap Seconds=37
  { 2459145.50, 69.357119 }, //23 October 2020, UT1-UTC=-0.1731186, Accumulated Leap Seconds=37
  { 2459146.50, 69.357193 }, //24 October 2020, UT1-UTC=-0.1731930, Accumulated Leap Seconds=37
  { 2459147.50, 69.357321 }, //25 October 2020, UT1-UTC=-0.1733207, Accumulated Leap Seconds=37
  { 2459148.50, 69.357498 }, //26 October 2020, UT1-UTC=-0.1734985, Accumulated Leap Seconds=37
  { 2459149.50, 69.357762 }, //27 October 2020, UT1-UTC=-0.1737620, Accumulated Leap Seconds=37
  { 2459150.50, 69.358107 }, //28 October 2020, UT1-UTC=-0.1741066, Accumulated Leap Seconds=37
  { 2459151.50, 69.358464 }, //29 October 2020, UT1-UTC=-0.1744639, Accumulated Leap Seconds=37
  { 2459152.50, 69.358771 }, //30 October 2020, UT1-UTC=-0.1747709, Accumulated Leap Seconds=37
  { 2459153.50, 69.359062 }, //31 October 2020, UT1-UTC=-0.1750625, Accumulated Leap Seconds=37
  { 2459154.50, 69.359324 }, //01 November 2020, UT1-UTC=-0.1753242, Accumulated Leap Seconds=37
  { 2459155.50, 69.359492 }, //02 November 2020, UT1-UTC=-0.1754919, Accumulated Leap Seconds=37
  { 2459156.50, 69.359538 }, //03 November 2020, UT1-UTC=-0.1755382, Accumulated Leap Seconds=37
  { 2459157.50, 69.359484 }, //04 November 2020, UT1-UTC=-0.1754841, Accumulated Leap Seconds=37
  { 2459158.50, 69.359309 }, //05 November 2020, UT1-UTC=-0.1753093, Accumulated Leap Seconds=37
  { 2459159.50, 69.359030 }, //06 November 2020, UT1-UTC=-0.1750298, Accumulated Leap Seconds=37
  { 2459160.50, 69.358723 }, //07 November 2020, UT1-UTC=-0.1747231, Accumulated Leap Seconds=37
  { 2459161.50, 69.358517 }, //08 November 2020, UT1-UTC=-0.1745166, Accumulated Leap Seconds=37
  { 2459162.50, 69.358566 }, //09 November 2020, UT1-UTC=-0.1745658, Accumulated Leap Seconds=37
  { 2459163.50, 69.358925 }, //10 November 2020, UT1-UTC=-0.1749252, Accumulated Leap Seconds=37
  { 2459164.50, 69.359491 }, //11 November 2020, UT1-UTC=-0.1754913, Accumulated Leap Seconds=37
  { 2459165.50, 69.360261 }, //12 November 2020, UT1-UTC=-0.1762615, Accumulated Leap Seconds=37
  { 2459166.50, 69.361147 }, //13 November 2020, UT1-UTC=-0.1771474, Accumulated Leap Seconds=37
  { 2459167.50, 69.361973 }, //14 November 2020, UT1-UTC=-0.1779726, Accumulated Leap Seconds=37
  { 2459168.50, 69.362636 }, //15 November 2020, UT1-UTC=-0.1786358, Accumulated Leap Seconds=37
  { 2459169.50, 69.363014 }, //16 November 2020, UT1-UTC=-0.1790143, Accumulated Leap Seconds=37
  { 2459170.50, 69.363119 }, //17 November 2020, UT1-UTC=-0.1791188, Accumulated Leap Seconds=37
  { 2459171.50, 69.363020 }, //18 November 2020, UT1-UTC=-0.1790201, Accumulated Leap Seconds=37
  { 2459172.50, 69.362745 }, //19 November 2020, UT1-UTC=-0.1787447, Accumulated Leap Seconds=37
  { 2459173.50, 69.362450 }, //20 November 2020, UT1-UTC=-0.1784505, Accumulated Leap Seconds=37
  { 2459174.50, 69.362194 }, //21 November 2020, UT1-UTC=-0.1781942, Accumulated Leap Seconds=37
  { 2459175.50, 69.362074 }, //22 November 2020, UT1-UTC=-0.1780743, Accumulated Leap Seconds=37
  { 2459176.50, 69.362147 }, //23 November 2020, UT1-UTC=-0.1781469, Accumulated Leap Seconds=37
  { 2459177.50, 69.362355 }, //24 November 2020, UT1-UTC=-0.1783547, Accumulated Leap Seconds=37
  { 2459178.50, 69.362641 }, //25 November 2020, UT1-UTC=-0.1786407, Accumulated Leap Seconds=37
  { 2459179.50, 69.362963 }, //26 November 2020, UT1-UTC=-0.1789629, Accumulated Leap Seconds=37
  { 2459180.50, 69.363135 }, //27 November 2020, UT1-UTC=-0.1791349, Accumulated Leap Seconds=37
  { 2459181.50, 69.363286 }, //28 November 2020, UT1-UTC=-0.1792864, Accumulated Leap Seconds=37
  { 2459182.50, 69.363305 }, //29 November 2020, UT1-UTC=-0.1793053, Accumulated Leap Seconds=37
  { 2459183.50, 69.363212 }, //30 November 2020, UT1-UTC=-0.1792119, Accumulated Leap Seconds=37
  { 2459184.50, 69.363024 }, //01 December 2020, UT1-UTC=-0.1790244, Accumulated Leap Seconds=37
  { 2459185.50, 69.362717 }, //02 December 2020, UT1-UTC=-0.1787170, Accumulated Leap Seconds=37
  { 2459186.50, 69.362321 }, //03 December 2020, UT1-UTC=-0.1783210, Accumulated Leap Seconds=37
  { 2459187.50, 69.361923 }, //04 December 2020, UT1-UTC=-0.1779234, Accumulated Leap Seconds=37
  { 2459188.50, 69.361628 }, //05 December 2020, UT1-UTC=-0.1776276, Accumulated Leap Seconds=37
  { 2459189.50, 69.361540 }, //06 December 2020, UT1-UTC=-0.1775402, Accumulated Leap Seconds=37
  { 2459190.50, 69.361646 }, //07 December 2020, UT1-UTC=-0.1776460, Accumulated Leap Seconds=37
  { 2459191.50, 69.361974 }, //08 December 2020, UT1-UTC=-0.1779740, Accumulated Leap Seconds=37
  { 2459192.50, 69.362494 }, //09 December 2020, UT1-UTC=-0.1784944, Accumulated Leap Seconds=37
  { 2459193.50, 69.363104 }, //10 December 2020, UT1-UTC=-0.1791038, Accumulated Leap Seconds=37
  { 2459194.50, 69.363735 }, //11 December 2020, UT1-UTC=-0.1797353, Accumulated Leap Seconds=37
  { 2459195.50, 69.364180 }, //12 December 2020, UT1-UTC=-0.1801801, Accumulated Leap Seconds=37
  { 2459196.50, 69.364399 }, //13 December 2020, UT1-UTC=-0.1803991, Accumulated Leap Seconds=37
  { 2459197.50, 69.364377 }, //14 December 2020, UT1-UTC=-0.1803768, Accumulated Leap Seconds=37
  { 2459198.50, 69.364117 }, //15 December 2020, UT1-UTC=-0.1801173, Accumulated Leap Seconds=37
  { 2459199.50, 69.363684 }, //16 December 2020, UT1-UTC=-0.1796837, Accumulated Leap Seconds=37
  { 2459200.50, 69.363165 }, //17 December 2020, UT1-UTC=-0.1791649, Accumulated Leap Seconds=37
  { 2459201.50, 69.362769 }, //18 December 2020, UT1-UTC=-0.1787687, Accumulated Leap Seconds=37
  { 2459202.50, 69.362462 }, //19 December 2020, UT1-UTC=-0.1784625, Accumulated Leap Seconds=37
  { 2459203.50, 69.362312 }, //20 December 2020, UT1-UTC=-0.1783116, Accumulated Leap Seconds=37
  { 2459204.50, 69.362298 }, //21 December 2020, UT1-UTC=-0.1782977, Accumulated Leap Seconds=37
  { 2459205.50, 69.362394 }, //22 December 2020, UT1-UTC=-0.1783938, Accumulated Leap Seconds=37
  { 2459206.50, 69.362563 }, //23 December 2020, UT1-UTC=-0.1785629, Accumulated Leap Seconds=37
  { 2459207.50, 69.362636 }, //24 December 2020, UT1-UTC=-0.1786356, Accumulated Leap Seconds=37
  { 2459208.50, 69.362635 }, //25 December 2020, UT1-UTC=-0.1786349, Accumulated Leap Seconds=37
  { 2459209.50, 69.362503 }, //26 December 2020, UT1-UTC=-0.1785028, Accumulated Leap Seconds=37
  { 2459210.50, 69.362208 }, //27 December 2020, UT1-UTC=-0.1782078, Accumulated Leap Seconds=37
  { 2459211.50, 69.361772 }, //28 December 2020, UT1-UTC=-0.1777715, Accumulated Leap Seconds=37
  { 2459212.50, 69.361239 }, //29 December 2020, UT1-UTC=-0.1772387, Accumulated Leap Seconds=37
  { 2459213.50, 69.360642 }, //30 December 2020, UT1-UTC=-0.1766424, Accumulated Leap Seconds=37
  { 2459214.50, 69.359981 }, //31 December 2020, UT1-UTC=-0.1759809, Accumulated Leap Seconds=37
  { 2459215.50, 69.359334 }, //01 January 2021, UT1-UTC=-0.1753340, Accumulated Leap Seconds=37
  { 2459216.50, 69.358812 }, //02 January 2021, UT1-UTC=-0.1748118, Accumulated Leap Seconds=37
  { 2459217.50, 69.358473 }, //03 January 2021, UT1-UTC=-0.1744728, Accumulated Leap Seconds=37
  { 2459218.50, 69.358326 }, //04 January 2021, UT1-UTC=-0.1743263, Accumulated Leap Seconds=37
  { 2459219.50, 69.358380 }, //05 January 2021, UT1-UTC=-0.1743799, Accumulated Leap Seconds=37
  { 2459220.50, 69.358574 }, //06 January 2021, UT1-UTC=-0.1745738, Accumulated Leap Seconds=37
  { 2459221.50, 69.358835 }, //07 January 2021, UT1-UTC=-0.1748355, Accumulated Leap Seconds=37
  { 2459222.50, 69.359060 }, //08 January 2021, UT1-UTC=-0.1750599, Accumulated Leap Seconds=37
  { 2459223.50, 69.359087 }, //09 January 2021, UT1-UTC=-0.1750871, Accumulated Leap Seconds=37
  { 2459224.50, 69.358867 }, //10 January 2021, UT1-UTC=-0.1748670, Accumulated Leap Seconds=37
  { 2459225.50, 69.358463 }, //11 January 2021, UT1-UTC=-0.1744634, Accumulated Leap Seconds=37
  { 2459226.50, 69.357966 }, //12 January 2021, UT1-UTC=-0.1739663, Accumulated Leap Seconds=37
  { 2459227.50, 69.357417 }, //13 January 2021, UT1-UTC=-0.1734174, Accumulated Leap Seconds=37
  { 2459228.50, 69.356884 }, //14 January 2021, UT1-UTC=-0.1728842, Accumulated Leap Seconds=37
  { 2459229.50, 69.356469 }, //15 January 2021, UT1-UTC=-0.1724694, Accumulated Leap Seconds=37
  { 2459230.50, 69.356040 }, //16 January 2021, UT1-UTC=-0.1720402, Accumulated Leap Seconds=37
  { 2459231.50, 69.355766 }, //17 January 2021, UT1-UTC=-0.1717660, Accumulated Leap Seconds=37
  { 2459232.50, 69.355586 }, //18 January 2021, UT1-UTC=-0.1715857, Accumulated Leap Seconds=37
  { 2459233.50, 69.355416 }, //19 January 2021, UT1-UTC=-0.1714164, Accumulated Leap Seconds=37
  { 2459234.50, 69.355273 }, //20 January 2021, UT1-UTC=-0.1712735, Accumulated Leap Seconds=37
  { 2459235.50, 69.355090 }, //21 January 2021, UT1-UTC=-0.1710897, Accumulated Leap Seconds=37
  { 2459236.50, 69.354796 }, //22 January 2021, UT1-UTC=-0.1707958, Accumulated Leap Seconds=37
  { 2459237.50, 69.354479 }, //23 January 2021, UT1-UTC=-0.1704792, Accumulated Leap Seconds=37
  { 2459238.50, 69.354010 }, //24 January 2021, UT1-UTC=-0.1700095, Accumulated Leap Seconds=37
  { 2459239.50, 69.353415 }, //25 January 2021, UT1-UTC=-0.1694151, Accumulated Leap Seconds=37
  { 2459240.50, 69.352720 }, //26 January 2021, UT1-UTC=-0.1687202, Accumulated Leap Seconds=37
  { 2459241.50, 69.351955 }, //27 January 2021, UT1-UTC=-0.1679551, Accumulated Leap Seconds=37
  { 2459242.50, 69.351309 }, //28 January 2021, UT1-UTC=-0.1673088, Accumulated Leap Seconds=37
  { 2459243.50, 69.350911 }, //29 January 2021, UT1-UTC=-0.1669112, Accumulated Leap Seconds=37
  { 2459244.50, 69.350682 }, //30 January 2021, UT1-UTC=-0.1666818, Accumulated Leap Seconds=37
  { 2459245.50, 69.350688 }, //31 January 2021, UT1-UTC=-0.1666882, Accumulated Leap Seconds=37
  { 2459246.50, 69.351013 }, //01 February 2021, UT1-UTC=-0.1670132, Accumulated Leap Seconds=37
  { 2459247.50, 69.351560 }, //02 February 2021, UT1-UTC=-0.1675605, Accumulated Leap Seconds=37
  { 2459248.50, 69.352217 }, //03 February 2021, UT1-UTC=-0.1682167, Accumulated Leap Seconds=37
  { 2459249.50, 69.352834 }, //04 February 2021, UT1-UTC=-0.1688337, Accumulated Leap Seconds=37
  { 2459250.50, 69.353285 }, //05 February 2021, UT1-UTC=-0.1692846, Accumulated Leap Seconds=37
  { 2459251.50, 69.353532 }, //06 February 2021, UT1-UTC=-0.1695318, Accumulated Leap Seconds=37
  { 2459252.50, 69.353577 }, //07 February 2021, UT1-UTC=-0.1695769, Accumulated Leap Seconds=37
  { 2459253.50, 69.353477 }, //08 February 2021, UT1-UTC=-0.1694775, Accumulated Leap Seconds=37
  { 2459254.50, 69.353330 }, //09 February 2021, UT1-UTC=-0.1693299, Accumulated Leap Seconds=37
  { 2459255.50, 69.353251 }, //10 February 2021, UT1-UTC=-0.1692512, Accumulated Leap Seconds=37
  { 2459256.50, 69.353258 }, //11 February 2021, UT1-UTC=-0.1692582, Accumulated Leap Seconds=37
  { 2459257.50, 69.353409 }, //12 February 2021, UT1-UTC=-0.1694093, Accumulated Leap Seconds=37
  { 2459258.50, 69.353718 }, //13 February 2021, UT1-UTC=-0.1697183, Accumulated Leap Seconds=37
  { 2459259.50, 69.354184 }, //14 February 2021, UT1-UTC=-0.1701837, Accumulated Leap Seconds=37
  { 2459260.50, 69.354732 }, //15 February 2021, UT1-UTC=-0.1707316, Accumulated Leap Seconds=37
  { 2459261.50, 69.355226 }, //16 February 2021, UT1-UTC=-0.1712262, Accumulated Leap Seconds=37
  { 2459262.50, 69.355659 }, //17 February 2021, UT1-UTC=-0.1716587, Accumulated Leap Seconds=37
  { 2459263.50, 69.355934 }, //18 February 2021, UT1-UTC=-0.1719343, Accumulated Leap Seconds=37
  { 2459264.50, 69.355997 }, //19 February 2021, UT1-UTC=-0.1719965, Accumulated Leap Seconds=37
  { 2459265.50, 69.355895 }, //20 February 2021, UT1-UTC=-0.1718954, Accumulated Leap Seconds=37
  { 2459266.50, 69.355559 }, //21 February 2021, UT1-UTC=-0.1715594, Accumulated Leap Seconds=37
  { 2459267.50, 69.355043 }, //22 February 2021, UT1-UTC=-0.1710427, Accumulated Leap Seconds=37
  { 2459268.50, 69.354455 }, //23 February 2021, UT1-UTC=-0.1704546, Accumulated Leap Seconds=37
  { 2459269.50, 69.353816 }, //24 February 2021, UT1-UTC=-0.1698162, Accumulated Leap Seconds=37
  { 2459270.50, 69.353332 }, //25 February 2021, UT1-UTC=-0.1693323, Accumulated Leap Seconds=37
  { 2459271.50, 69.353052 }, //26 February 2021, UT1-UTC=-0.1690518, Accumulated Leap Seconds=37
  { 2459272.50, 69.353042 }, //27 February 2021, UT1-UTC=-0.1690418, Accumulated Leap Seconds=37
  { 2459273.50, 69.353289 }, //28 February 2021, UT1-UTC=-0.1692889, Accumulated Leap Seconds=37
  { 2459274.50, 69.353791 }, //01 March 2021, UT1-UTC=-0.1697915, Accumulated Leap Seconds=37
  { 2459275.50, 69.354408 }, //02 March 2021, UT1-UTC=-0.1704082, Accumulated Leap Seconds=37
  { 2459276.50, 69.355006 }, //03 March 2021, UT1-UTC=-0.1710064, Accumulated Leap Seconds=37
  { 2459277.50, 69.355574 }, //04 March 2021, UT1-UTC=-0.1715740, Accumulated Leap Seconds=37
  { 2459278.50, 69.355908 }, //05 March 2021, UT1-UTC=-0.1719079, Accumulated Leap Seconds=37
  { 2459279.50, 69.355960 }, //06 March 2021, UT1-UTC=-0.1719597, Accumulated Leap Seconds=37
  { 2459280.50, 69.355830 }, //07 March 2021, UT1-UTC=-0.1718304, Accumulated Leap Seconds=37
  { 2459281.50, 69.355596 }, //08 March 2021, UT1-UTC=-0.1715958, Accumulated Leap Seconds=37
  { 2459282.50, 69.355314 }, //09 March 2021, UT1-UTC=-0.1713142, Accumulated Leap Seconds=37
  { 2459283.50, 69.355105 }, //10 March 2021, UT1-UTC=-0.1711055, Accumulated Leap Seconds=37
  { 2459284.50, 69.355021 }, //11 March 2021, UT1-UTC=-0.1710208, Accumulated Leap Seconds=37
  { 2459285.50, 69.355074 }, //12 March 2021, UT1-UTC=-0.1710743, Accumulated Leap Seconds=37
  { 2459286.50, 69.355281 }, //13 March 2021, UT1-UTC=-0.1712814, Accumulated Leap Seconds=37
  { 2459287.50, 69.355605 }, //14 March 2021, UT1-UTC=-0.1716047, Accumulated Leap Seconds=37
  { 2459288.50, 69.355944 }, //15 March 2021, UT1-UTC=-0.1719443, Accumulated Leap Seconds=37
  { 2459289.50, 69.356229 }, //16 March 2021, UT1-UTC=-0.1722295, Accumulated Leap Seconds=37
  { 2459290.50, 69.356371 }, //17 March 2021, UT1-UTC=-0.1723715, Accumulated Leap Seconds=37
  { 2459291.50, 69.356377 }, //18 March 2021, UT1-UTC=-0.1723775, Accumulated Leap Seconds=37
  { 2459292.50, 69.356285 }, //19 March 2021, UT1-UTC=-0.1722853, Accumulated Leap Seconds=37
  { 2459293.50, 69.356014 }, //20 March 2021, UT1-UTC=-0.1720141, Accumulated Leap Seconds=37
  { 2459294.50, 69.355668 }, //21 March 2021, UT1-UTC=-0.1716685, Accumulated Leap Seconds=37
  { 2459295.50, 69.355311 }, //22 March 2021, UT1-UTC=-0.1713111, Accumulated Leap Seconds=37
  { 2459296.50, 69.354963 }, //23 March 2021, UT1-UTC=-0.1709631, Accumulated Leap Seconds=37
  { 2459297.50, 69.354673 }, //24 March 2021, UT1-UTC=-0.1706729, Accumulated Leap Seconds=37
  { 2459298.50, 69.354524 }, //25 March 2021, UT1-UTC=-0.1705245, Accumulated Leap Seconds=37
  { 2459299.50, 69.354585 }, //26 March 2021, UT1-UTC=-0.1705846, Accumulated Leap Seconds=37
  { 2459300.50, 69.354926 }, //27 March 2021, UT1-UTC=-0.1709259, Accumulated Leap Seconds=37
  { 2459301.50, 69.355480 }, //28 March 2021, UT1-UTC=-0.1714805, Accumulated Leap Seconds=37
  { 2459302.50, 69.356252 }, //29 March 2021, UT1-UTC=-0.1722520, Accumulated Leap Seconds=37
  { 2459303.50, 69.357046 }, //30 March 2021, UT1-UTC=-0.1730457, Accumulated Leap Seconds=37
  { 2459304.50, 69.357725 }, //31 March 2021, UT1-UTC=-0.1737246, Accumulated Leap Seconds=37
  { 2459305.50, 69.358222 }, //01 April 2021, UT1-UTC=-0.1742218, Accumulated Leap Seconds=37
  { 2459306.50, 69.358484 }, //02 April 2021, UT1-UTC=-0.1744839, Accumulated Leap Seconds=37
  { 2459307.50, 69.358553 }, //03 April 2021, UT1-UTC=-0.1745528, Accumulated Leap Seconds=37
  { 2459308.50, 69.358477 }, //04 April 2021, UT1-UTC=-0.1744768, Accumulated Leap Seconds=37
  { 2459309.50, 69.358372 }, //05 April 2021, UT1-UTC=-0.1743720, Accumulated Leap Seconds=37
  { 2459310.50, 69.358408 }, //06 April 2021, UT1-UTC=-0.1744081, Accumulated Leap Seconds=37
  { 2459311.50, 69.358601 }, //07 April 2021, UT1-UTC=-0.1746012, Accumulated Leap Seconds=37
  { 2459312.50, 69.358863 }, //08 April 2021, UT1-UTC=-0.1748628, Accumulated Leap Seconds=37
  { 2459313.50, 69.359281 }, //09 April 2021, UT1-UTC=-0.1752813, Accumulated Leap Seconds=37
  { 2459314.50, 69.359755 }, //10 April 2021, UT1-UTC=-0.1757552, Accumulated Leap Seconds=37
  { 2459315.50, 69.360284 }, //11 April 2021, UT1-UTC=-0.1762845, Accumulated Leap Seconds=37
  { 2459316.50, 69.360833 }, //12 April 2021, UT1-UTC=-0.1768331, Accumulated Leap Seconds=37
  { 2459317.50, 69.361251 }, //13 April 2021, UT1-UTC=-0.1772515, Accumulated Leap Seconds=37
  { 2459318.50, 69.361616 }, //14 April 2021, UT1-UTC=-0.1776162, Accumulated Leap Seconds=37
  { 2459319.50, 69.361857 }, //15 April 2021, UT1-UTC=-0.1778569, Accumulated Leap Seconds=37
  { 2459320.50, 69.361991 }, //16 April 2021, UT1-UTC=-0.1779908, Accumulated Leap Seconds=37
  { 2459321.50, 69.362018 }, //17 April 2021, UT1-UTC=-0.1780185, Accumulated Leap Seconds=37
  { 2459322.50, 69.361974 }, //18 April 2021, UT1-UTC=-0.1779744, Accumulated Leap Seconds=37
  { 2459323.50, 69.361869 }, //19 April 2021, UT1-UTC=-0.1778687, Accumulated Leap Seconds=37
  { 2459324.50, 69.361843 }, //20 April 2021, UT1-UTC=-0.1778427, Accumulated Leap Seconds=37
  { 2459325.50, 69.361862 }, //21 April 2021, UT1-UTC=-0.1778615, Accumulated Leap Seconds=37
  { 2459326.50, 69.362046 }, //22 April 2021, UT1-UTC=-0.1780459, Accumulated Leap Seconds=37
  { 2459327.50, 69.362525 }, //23 April 2021, UT1-UTC=-0.1785253, Accumulated Leap Seconds=37
  { 2459328.50, 69.363155 }, //24 April 2021, UT1-UTC=-0.1791552, Accumulated Leap Seconds=37
  { 2459329.50, 69.363962 }, //25 April 2021, UT1-UTC=-0.1799625, Accumulated Leap Seconds=37
  { 2459330.50, 69.364887 }, //26 April 2021, UT1-UTC=-0.1808869, Accumulated Leap Seconds=37
  { 2459331.50, 69.365898 }, //27 April 2021, UT1-UTC=-0.1818983, Accumulated Leap Seconds=37
  { 2459332.50, 69.366703 }, //28 April 2021, UT1-UTC=-0.1827032, Accumulated Leap Seconds=37
  { 2459333.50, 69.367201 }, //29 April 2021, UT1-UTC=-0.1832012, Accumulated Leap Seconds=37
  { 2459334.50, 69.367369 }, //30 April 2021, UT1-UTC=-0.1833694, Accumulated Leap Seconds=37
  { 2459335.50, 69.367301 }, //01 May 2021, UT1-UTC=-0.1833006, Accumulated Leap Seconds=37
  { 2459336.50, 69.367085 }, //02 May 2021, UT1-UTC=-0.1830849, Accumulated Leap Seconds=37
  { 2459337.50, 69.366789 }, //03 May 2021, UT1-UTC=-0.1827894, Accumulated Leap Seconds=37
  { 2459338.50, 69.366519 }, //04 May 2021, UT1-UTC=-0.1825186, Accumulated Leap Seconds=37
  { 2459339.50, 69.366387 }, //05 May 2021, UT1-UTC=-0.1823873, Accumulated Leap Seconds=37
  { 2459340.50, 69.366396 }, //06 May 2021, UT1-UTC=-0.1823958, Accumulated Leap Seconds=37
  { 2459341.50, 69.366557 }, //07 May 2021, UT1-UTC=-0.1825566, Accumulated Leap Seconds=37
  { 2459342.50, 69.366822 }, //08 May 2021, UT1-UTC=-0.1828216, Accumulated Leap Seconds=37
  { 2459343.50, 69.367047 }, //09 May 2021, UT1-UTC=-0.1830469, Accumulated Leap Seconds=37
  { 2459344.50, 69.367223 }, //10 May 2021, UT1-UTC=-0.1832230, Accumulated Leap Seconds=37
  { 2459345.50, 69.367354 }, //11 May 2021, UT1-UTC=-0.1833539, Accumulated Leap Seconds=37
  { 2459346.50, 69.367441 }, //12 May 2021, UT1-UTC=-0.1834411, Accumulated Leap Seconds=37
  { 2459347.50, 69.367301 }, //13 May 2021, UT1-UTC=-0.1833015, Accumulated Leap Seconds=37
  { 2459348.50, 69.367098 }, //14 May 2021, UT1-UTC=-0.1830977, Accumulated Leap Seconds=37
  { 2459349.50, 69.366823 }, //15 May 2021, UT1-UTC=-0.1828230, Accumulated Leap Seconds=37
  { 2459350.50, 69.366535 }, //16 May 2021, UT1-UTC=-0.1825350, Accumulated Leap Seconds=37
  { 2459351.50, 69.366260 }, //17 May 2021, UT1-UTC=-0.1822596, Accumulated Leap Seconds=37
  { 2459352.50, 69.366023 }, //18 May 2021, UT1-UTC=-0.1820234, Accumulated Leap Seconds=37
  { 2459353.50, 69.365971 }, //19 May 2021, UT1-UTC=-0.1819708, Accumulated Leap Seconds=37
  { 2459354.50, 69.366073 }, //20 May 2021, UT1-UTC=-0.1820729, Accumulated Leap Seconds=37
  { 2459355.50, 69.366507 }, //21 May 2021, UT1-UTC=-0.1825069, Accumulated Leap Seconds=37
  { 2459356.50, 69.367027 }, //22 May 2021, UT1-UTC=-0.1830266, Accumulated Leap Seconds=37
  { 2459357.50, 69.367716 }, //23 May 2021, UT1-UTC=-0.1837161, Accumulated Leap Seconds=37
  { 2459358.50, 69.368435 }, //24 May 2021, UT1-UTC=-0.1844354, Accumulated Leap Seconds=37
  { 2459359.50, 69.369041 }, //25 May 2021, UT1-UTC=-0.1850412, Accumulated Leap Seconds=37
  { 2459360.50, 69.369417 }, //26 May 2021, UT1-UTC=-0.1854166, Accumulated Leap Seconds=37
  { 2459361.50, 69.369542 }, //27 May 2021, UT1-UTC=-0.1855424, Accumulated Leap Seconds=37
  { 2459362.50, 69.369404 }, //28 May 2021, UT1-UTC=-0.1854036, Accumulated Leap Seconds=37
  { 2459363.50, 69.369050 }, //29 May 2021, UT1-UTC=-0.1850501, Accumulated Leap Seconds=37
  { 2459364.50, 69.368589 }, //30 May 2021, UT1-UTC=-0.1845894, Accumulated Leap Seconds=37
  { 2459365.50, 69.368161 }, //31 May 2021, UT1-UTC=-0.1841611, Accumulated Leap Seconds=37
  { 2459366.50, 69.367863 }, //01 June 2021, UT1-UTC=-0.1838630, Accumulated Leap Seconds=37
  { 2459367.50, 69.367704 }, //02 June 2021, UT1-UTC=-0.1837036, Accumulated Leap Seconds=37
  { 2459368.50, 69.367660 }, //03 June 2021, UT1-UTC=-0.1836598, Accumulated Leap Seconds=37
  { 2459369.50, 69.367710 }, //04 June 2021, UT1-UTC=-0.1837104, Accumulated Leap Seconds=37
  { 2459370.50, 69.367740 }, //05 June 2021, UT1-UTC=-0.1837396, Accumulated Leap Seconds=37
  { 2459371.50, 69.367666 }, //06 June 2021, UT1-UTC=-0.1836660, Accumulated Leap Seconds=37
  { 2459372.50, 69.367471 }, //07 June 2021, UT1-UTC=-0.1834706, Accumulated Leap Seconds=37
  { 2459373.50, 69.367129 }, //08 June 2021, UT1-UTC=-0.1831290, Accumulated Leap Seconds=37
  { 2459374.50, 69.366575 }, //09 June 2021, UT1-UTC=-0.1825746, Accumulated Leap Seconds=37
  { 2459375.50, 69.365902 }, //10 June 2021, UT1-UTC=-0.1819017, Accumulated Leap Seconds=37
  { 2459376.50, 69.365101 }, //11 June 2021, UT1-UTC=-0.1811009, Accumulated Leap Seconds=37
  { 2459377.50, 69.364171 }, //12 June 2021, UT1-UTC=-0.1801705, Accumulated Leap Seconds=37
  { 2459378.50, 69.363266 }, //13 June 2021, UT1-UTC=-0.1792665, Accumulated Leap Seconds=37
  { 2459379.50, 69.362346 }, //14 June 2021, UT1-UTC=-0.1783462, Accumulated Leap Seconds=37
  { 2459380.50, 69.361572 }, //15 June 2021, UT1-UTC=-0.1775718, Accumulated Leap Seconds=37
  { 2459381.50, 69.360975 }, //16 June 2021, UT1-UTC=-0.1769755, Accumulated Leap Seconds=37
  { 2459382.50, 69.360507 }, //17 June 2021, UT1-UTC=-0.1765067, Accumulated Leap Seconds=37
  { 2459383.50, 69.360283 }, //18 June 2021, UT1-UTC=-0.1762826, Accumulated Leap Seconds=37
  { 2459384.50, 69.360208 }, //19 June 2021, UT1-UTC=-0.1762076, Accumulated Leap Seconds=37
  { 2459385.50, 69.360239 }, //20 June 2021, UT1-UTC=-0.1762386, Accumulated Leap Seconds=37
  { 2459386.50, 69.360206 }, //21 June 2021, UT1-UTC=-0.1762061, Accumulated Leap Seconds=37
  { 2459387.50, 69.359952 }, //22 June 2021, UT1-UTC=-0.1759518, Accumulated Leap Seconds=37
  { 2459388.50, 69.359471 }, //23 June 2021, UT1-UTC=-0.1754706, Accumulated Leap Seconds=37
  { 2459389.50, 69.358663 }, //24 June 2021, UT1-UTC=-0.1746633, Accumulated Leap Seconds=37
  { 2459390.50, 69.357579 }, //25 June 2021, UT1-UTC=-0.1735795, Accumulated Leap Seconds=37
  { 2459391.50, 69.356387 }, //26 June 2021, UT1-UTC=-0.1723867, Accumulated Leap Seconds=37
  { 2459392.50, 69.355140 }, //27 June 2021, UT1-UTC=-0.1711396, Accumulated Leap Seconds=37
  { 2459393.50, 69.353970 }, //28 June 2021, UT1-UTC=-0.1699698, Accumulated Leap Seconds=37
  { 2459394.50, 69.352983 }, //29 June 2021, UT1-UTC=-0.1689827, Accumulated Leap Seconds=37
  { 2459395.50, 69.352192 }, //30 June 2021, UT1-UTC=-0.1681924, Accumulated Leap Seconds=37
  { 2459396.50, 69.351419 }, //01 July 2021, UT1-UTC=-0.1674190, Accumulated Leap Seconds=37
  { 2459397.50, 69.350724 }, //02 July 2021, UT1-UTC=-0.1667236, Accumulated Leap Seconds=37
  { 2459398.50, 69.349989 }, //03 July 2021, UT1-UTC=-0.1659893, Accumulated Leap Seconds=37
  { 2459399.50, 69.349171 }, //04 July 2021, UT1-UTC=-0.1651715, Accumulated Leap Seconds=37
  { 2459400.50, 69.348163 }, //05 July 2021, UT1-UTC=-0.1641626, Accumulated Leap Seconds=37
  { 2459401.50, 69.346980 }, //06 July 2021, UT1-UTC=-0.1629798, Accumulated Leap Seconds=37
  { 2459402.50, 69.345807 }, //07 July 2021, UT1-UTC=-0.1618074, Accumulated Leap Seconds=37
  { 2459403.50, 69.344451 }, //08 July 2021, UT1-UTC=-0.1604511, Accumulated Leap Seconds=37
  { 2459404.50, 69.343025 }, //09 July 2021, UT1-UTC=-0.1590251, Accumulated Leap Seconds=37
  { 2459405.50, 69.341565 }, //10 July 2021, UT1-UTC=-0.1575647, Accumulated Leap Seconds=37
  { 2459406.50, 69.340219 }, //11 July 2021, UT1-UTC=-0.1562192, Accumulated Leap Seconds=37
  { 2459407.50, 69.338968 }, //12 July 2021, UT1-UTC=-0.1549683, Accumulated Leap Seconds=37
  { 2459408.50, 69.337897 }, //13 July 2021, UT1-UTC=-0.1538972, Accumulated Leap Seconds=37
  { 2459409.50, 69.337008 }, //14 July 2021, UT1-UTC=-0.1530083, Accumulated Leap Seconds=37
  { 2459410.50, 69.336397 }, //15 July 2021, UT1-UTC=-0.1523970, Accumulated Leap Seconds=37
  { 2459411.50, 69.335991 }, //16 July 2021, UT1-UTC=-0.1519912, Accumulated Leap Seconds=37
  { 2459412.50, 69.335746 }, //17 July 2021, UT1-UTC=-0.1517456, Accumulated Leap Seconds=37
  { 2459413.50, 69.335531 }, //18 July 2021, UT1-UTC=-0.1515307, Accumulated Leap Seconds=37
  { 2459414.50, 69.335190 }, //19 July 2021, UT1-UTC=-0.1511896, Accumulated Leap Seconds=37
  { 2459415.50, 69.334704 }, //20 July 2021, UT1-UTC=-0.1507035, Accumulated Leap Seconds=37
  { 2459416.50, 69.334077 }, //21 July 2021, UT1-UTC=-0.1500775, Accumulated Leap Seconds=37
  { 2459417.50, 69.333301 }, //22 July 2021, UT1-UTC=-0.1493008, Accumulated Leap Seconds=37
  { 2459418.50, 69.332386 }, //23 July 2021, UT1-UTC=-0.1483865, Accumulated Leap Seconds=37
  { 2459419.50, 69.331474 }, //24 July 2021, UT1-UTC=-0.1474742, Accumulated Leap Seconds=37
  { 2459420.50, 69.330703 }, //25 July 2021, UT1-UTC=-0.1467032, Accumulated Leap Seconds=37
  { 2459421.50, 69.330074 }, //26 July 2021, UT1-UTC=-0.1460741, Accumulated Leap Seconds=37
  { 2459422.50, 69.329535 }, //27 July 2021, UT1-UTC=-0.1455355, Accumulated Leap Seconds=37
  { 2459423.50, 69.329121 }, //28 July 2021, UT1-UTC=-0.1451213, Accumulated Leap Seconds=37
  { 2459424.50, 69.328760 }, //29 July 2021, UT1-UTC=-0.1447597, Accumulated Leap Seconds=37
  { 2459425.50, 69.328366 }, //30 July 2021, UT1-UTC=-0.1443658, Accumulated Leap Seconds=37
  { 2459426.50, 69.327926 }, //31 July 2021, UT1-UTC=-0.1439257, Accumulated Leap Seconds=37
  { 2459427.50, 69.327343 }, //01 August 2021, UT1-UTC=-0.1433427, Accumulated Leap Seconds=37
  { 2459428.50, 69.326647 }, //02 August 2021, UT1-UTC=-0.1426475, Accumulated Leap Seconds=37
  { 2459429.50, 69.325831 }, //03 August 2021, UT1-UTC=-0.1418307, Accumulated Leap Seconds=37
  { 2459430.50, 69.324888 }, //04 August 2021, UT1-UTC=-0.1408875, Accumulated Leap Seconds=37
  { 2459431.50, 69.323857 }, //05 August 2021, UT1-UTC=-0.1398571, Accumulated Leap Seconds=37
  { 2459432.50, 69.322769 }, //06 August 2021, UT1-UTC=-0.1387687, Accumulated Leap Seconds=37
  { 2459433.50, 69.321752 }, //07 August 2021, UT1-UTC=-0.1377521, Accumulated Leap Seconds=37
  { 2459434.50, 69.320744 }, //08 August 2021, UT1-UTC=-0.1367436, Accumulated Leap Seconds=37
  { 2459435.50, 69.319826 }, //09 August 2021, UT1-UTC=-0.1358262, Accumulated Leap Seconds=37
  { 2459436.50, 69.319082 }, //10 August 2021, UT1-UTC=-0.1350822, Accumulated Leap Seconds=37
  { 2459437.50, 69.318512 }, //11 August 2021, UT1-UTC=-0.1345123, Accumulated Leap Seconds=37
  { 2459438.50, 69.318100 }, //12 August 2021, UT1-UTC=-0.1340997, Accumulated Leap Seconds=37
  { 2459439.50, 69.317837 }, //13 August 2021, UT1-UTC=-0.1338369, Accumulated Leap Seconds=37
  { 2459440.50, 69.317510 }, //14 August 2021, UT1-UTC=-0.1335099, Accumulated Leap Seconds=37
  { 2459441.50, 69.317125 }, //15 August 2021, UT1-UTC=-0.1331252, Accumulated Leap Seconds=37
  { 2459442.50, 69.316560 }, //16 August 2021, UT1-UTC=-0.1325601, Accumulated Leap Seconds=37
  { 2459443.50, 69.315755 }, //17 August 2021, UT1-UTC=-0.1317555, Accumulated Leap Seconds=37
  { 2459444.50, 69.314714 }, //18 August 2021, UT1-UTC=-0.1307144, Accumulated Leap Seconds=37
  { 2459445.50, 69.313564 }, //19 August 2021, UT1-UTC=-0.1295642, Accumulated Leap Seconds=37
  { 2459446.50, 69.312361 }, //20 August 2021, UT1-UTC=-0.1283612, Accumulated Leap Seconds=37
  { 2459447.50, 69.311232 }, //21 August 2021, UT1-UTC=-0.1272321, Accumulated Leap Seconds=37
  { 2459448.50, 69.310319 }, //22 August 2021, UT1-UTC=-0.1263191, Accumulated Leap Seconds=37
  { 2459449.50, 69.309574 }, //23 August 2021, UT1-UTC=-0.1255745, Accumulated Leap Seconds=37
  { 2459450.50, 69.308966 }, //24 August 2021, UT1-UTC=-0.1249661, Accumulated Leap Seconds=37
  { 2459451.50, 69.308454 }, //25 August 2021, UT1-UTC=-0.1244538, Accumulated Leap Seconds=37
  { 2459452.50, 69.308009 }, //26 August 2021, UT1-UTC=-0.1240091, Accumulated Leap Seconds=37
  { 2459453.50, 69.307491 }, //27 August 2021, UT1-UTC=-0.1234907, Accumulated Leap Seconds=37
  { 2459454.50, 69.306891 }, //28 August 2021, UT1-UTC=-0.1228913, Accumulated Leap Seconds=37
  { 2459455.50, 69.306107 }, //29 August 2021, UT1-UTC=-0.1221074, Accumulated Leap Seconds=37
  { 2459456.50, 69.305265 }, //30 August 2021, UT1-UTC=-0.1212649, Accumulated Leap Seconds=37
  { 2459457.50, 69.304326 }, //31 August 2021, UT1-UTC=-0.1203260, Accumulated Leap Seconds=37
  { 2459458.50, 69.303329 }, //01 September 2021, UT1-UTC=-0.1193293, Accumulated Leap Seconds=37
  { 2459459.50, 69.302273 }, //02 September 2021, UT1-UTC=-0.1182735, Accumulated Leap Seconds=37
  { 2459460.50, 69.301218 }, //03 September 2021, UT1-UTC=-0.1172182, Accumulated Leap Seconds=37
  { 2459461.50, 69.300338 }, //04 September 2021, UT1-UTC=-0.1163385, Accumulated Leap Seconds=37
  { 2459462.50, 69.299528 }, //05 September 2021, UT1-UTC=-0.1155283, Accumulated Leap Seconds=37
  { 2459463.50, 69.298925 }, //06 September 2021, UT1-UTC=-0.1149253, Accumulated Leap Seconds=37
  { 2459464.50, 69.298567 }, //07 September 2021, UT1-UTC=-0.1145666, Accumulated Leap Seconds=37
  { 2459465.50, 69.298429 }, //08 September 2021, UT1-UTC=-0.1144291, Accumulated Leap Seconds=37
  { 2459466.50, 69.298456 }, //09 September 2021, UT1-UTC=-0.1144558, Accumulated Leap Seconds=37
  { 2459467.50, 69.298497 }, //10 September 2021, UT1-UTC=-0.1144973, Accumulated Leap Seconds=37
  { 2459468.50, 69.298514 }, //11 September 2021, UT1-UTC=-0.1145144, Accumulated Leap Seconds=37
  { 2459469.50, 69.298322 }, //12 September 2021, UT1-UTC=-0.1143218, Accumulated Leap Seconds=37
  { 2459470.50, 69.297907 }, //13 September 2021, UT1-UTC=-0.1139072, Accumulated Leap Seconds=37
  { 2459471.50, 69.297240 }, //14 September 2021, UT1-UTC=-0.1132401, Accumulated Leap Seconds=37
  { 2459472.50, 69.296449 }, //15 September 2021, UT1-UTC=-0.1124488, Accumulated Leap Seconds=37
  { 2459473.50, 69.295575 }, //16 September 2021, UT1-UTC=-0.1115747, Accumulated Leap Seconds=37
  { 2459474.50, 69.294783 }, //17 September 2021, UT1-UTC=-0.1107835, Accumulated Leap Seconds=37
  { 2459475.50, 69.294152 }, //18 September 2021, UT1-UTC=-0.1101522, Accumulated Leap Seconds=37
  { 2459476.50, 69.293794 }, //19 September 2021, UT1-UTC=-0.1097936, Accumulated Leap Seconds=37
  { 2459477.50, 69.293504 }, //20 September 2021, UT1-UTC=-0.1095036, Accumulated Leap Seconds=37
  { 2459478.50, 69.293377 }, //21 September 2021, UT1-UTC=-0.1093766, Accumulated Leap Seconds=37
  { 2459479.50, 69.293244 }, //22 September 2021, UT1-UTC=-0.1092443, Accumulated Leap Seconds=37
  { 2459480.50, 69.293181 }, //23 September 2021, UT1-UTC=-0.1091809, Accumulated Leap Seconds=37
  { 2459481.50, 69.293067 }, //24 September 2021, UT1-UTC=-0.1090666, Accumulated Leap Seconds=37
  { 2459482.50, 69.292850 }, //25 September 2021, UT1-UTC=-0.1088502, Accumulated Leap Seconds=37
  { 2459483.50, 69.292505 }, //26 September 2021, UT1-UTC=-0.1085049, Accumulated Leap Seconds=37
  { 2459484.50, 69.292036 }, //27 September 2021, UT1-UTC=-0.1080357, Accumulated Leap Seconds=37
  { 2459485.50, 69.291398 }, //28 September 2021, UT1-UTC=-0.1073977, Accumulated Leap Seconds=37
  { 2459486.50, 69.290719 }, //29 September 2021, UT1-UTC=-0.1067194, Accumulated Leap Seconds=37
  { 2459487.50, 69.289991 }, //30 September 2021, UT1-UTC=-0.1059905, Accumulated Leap Seconds=37
  { 2459488.50, 69.289249 }, //01 October 2021, UT1-UTC=-0.1052490, Accumulated Leap Seconds=37
  { 2459489.50, 69.288697 }, //02 October 2021, UT1-UTC=-0.1046972, Accumulated Leap Seconds=37
  { 2459490.50, 69.288322 }, //03 October 2021, UT1-UTC=-0.1043221, Accumulated Leap Seconds=37
  { 2459491.50, 69.288115 }, //04 October 2021, UT1-UTC=-0.1041148, Accumulated Leap Seconds=37
  { 2459492.50, 69.288161 }, //05 October 2021, UT1-UTC=-0.1041607, Accumulated Leap Seconds=37
  { 2459493.50, 69.288493 }, //06 October 2021, UT1-UTC=-0.1044930, Accumulated Leap Seconds=37
  { 2459494.50, 69.289040 }, //07 October 2021, UT1-UTC=-0.1050400, Accumulated Leap Seconds=37
  { 2459495.50, 69.289552 }, //08 October 2021, UT1-UTC=-0.1055523, Accumulated Leap Seconds=37
  { 2459496.50, 69.289930 }, //09 October 2021, UT1-UTC=-0.1059297, Accumulated Leap Seconds=37
  { 2459497.50, 69.290053 }, //10 October 2021, UT1-UTC=-0.1060531, Accumulated Leap Seconds=37
  { 2459498.50, 69.289956 }, //11 October 2021, UT1-UTC=-0.1059562, Accumulated Leap Seconds=37
  { 2459499.50, 69.289707 }, //12 October 2021, UT1-UTC=-0.1057066, Accumulated Leap Seconds=37
  { 2459500.50, 69.289372 }, //13 October 2021, UT1-UTC=-0.1053724, Accumulated Leap Seconds=37
  { 2459501.50, 69.289050 }, //14 October 2021, UT1-UTC=-0.1050497, Accumulated Leap Seconds=37
  { 2459502.50, 69.288816 }, //15 October 2021, UT1-UTC=-0.1048161, Accumulated Leap Seconds=37
  { 2459503.50, 69.288714 }, //16 October 2021, UT1-UTC=-0.1047139, Accumulated Leap Seconds=37
  { 2459504.50, 69.288781 }, //17 October 2021, UT1-UTC=-0.1047806, Accumulated Leap Seconds=37
  { 2459505.50, 69.288975 }, //18 October 2021, UT1-UTC=-0.1049752, Accumulated Leap Seconds=37
  { 2459506.50, 69.289321 }, //19 October 2021, UT1-UTC=-0.1053208, Accumulated Leap Seconds=37
  { 2459507.50, 69.289645 }, //20 October 2021, UT1-UTC=-0.1056449, Accumulated Leap Seconds=37
  { 2459508.50, 69.289935 }, //21 October 2021, UT1-UTC=-0.1059347, Accumulated Leap Seconds=37
  { 2459509.50, 69.290147 }, //22 October 2021, UT1-UTC=-0.1061467, Accumulated Leap Seconds=37
  { 2459510.50, 69.290165 }, //23 October 2021, UT1-UTC=-0.1061646, Accumulated Leap Seconds=37
  { 2459511.50, 69.289935 }, //24 October 2021, UT1-UTC=-0.1059346, Accumulated Leap Seconds=37
  { 2459512.50, 69.289563 }, //25 October 2021, UT1-UTC=-0.1055629, Accumulated Leap Seconds=37
  { 2459513.50, 69.289070 }, //26 October 2021, UT1-UTC=-0.1050701, Accumulated Leap Seconds=37
  { 2459514.50, 69.288559 }, //27 October 2021, UT1-UTC=-0.1045590, Accumulated Leap Seconds=37
  { 2459515.50, 69.288128 }, //28 October 2021, UT1-UTC=-0.1041280, Accumulated Leap Seconds=37
  { 2459516.50, 69.287764 }, //29 October 2021, UT1-UTC=-0.1037640, Accumulated Leap Seconds=37
  { 2459517.50, 69.287563 }, //30 October 2021, UT1-UTC=-0.1035626, Accumulated Leap Seconds=37
  { 2459518.50, 69.287638 }, //31 October 2021, UT1-UTC=-0.1036382, Accumulated Leap Seconds=37
  { 2459519.50, 69.288039 }, //01 November 2021, UT1-UTC=-0.1040393, Accumulated Leap Seconds=37
  { 2459520.50, 69.288680 }, //02 November 2021, UT1-UTC=-0.1046797, Accumulated Leap Seconds=37
  { 2459521.50, 69.289525 }, //03 November 2021, UT1-UTC=-0.1055246, Accumulated Leap Seconds=37
  { 2459522.50, 69.290341 }, //04 November 2021, UT1-UTC=-0.1063411, Accumulated Leap Seconds=37
  { 2459523.50, 69.291034 }, //05 November 2021, UT1-UTC=-0.1070341, Accumulated Leap Seconds=37
  { 2459524.50, 69.291420 }, //06 November 2021, UT1-UTC=-0.1074204, Accumulated Leap Seconds=37
  { 2459525.50, 69.291484 }, //07 November 2021, UT1-UTC=-0.1074842, Accumulated Leap Seconds=37
  { 2459526.50, 69.291368 }, //08 November 2021, UT1-UTC=-0.1073682, Accumulated Leap Seconds=37
  { 2459527.50, 69.291173 }, //09 November 2021, UT1-UTC=-0.1071734, Accumulated Leap Seconds=37
  { 2459528.50, 69.290969 }, //10 November 2021, UT1-UTC=-0.1069692, Accumulated Leap Seconds=37
  { 2459529.50, 69.290869 }, //11 November 2021, UT1-UTC=-0.1068686, Accumulated Leap Seconds=37
  { 2459530.50, 69.290989 }, //12 November 2021, UT1-UTC=-0.1069889, Accumulated Leap Seconds=37
  { 2459531.50, 69.291249 }, //13 November 2021, UT1-UTC=-0.1072489, Accumulated Leap Seconds=37
  { 2459532.50, 69.291641 }, //14 November 2021, UT1-UTC=-0.1076406, Accumulated Leap Seconds=37
  { 2459533.50, 69.292005 }, //15 November 2021, UT1-UTC=-0.1080050, Accumulated Leap Seconds=37
  { 2459534.50, 69.292269 }, //16 November 2021, UT1-UTC=-0.1082693, Accumulated Leap Seconds=37
  { 2459535.50, 69.292462 }, //17 November 2021, UT1-UTC=-0.1084617, Accumulated Leap Seconds=37
  { 2459536.50, 69.292531 }, //18 November 2021, UT1-UTC=-0.1085313, Accumulated Leap Seconds=37
  { 2459537.50, 69.292462 }, //19 November 2021, UT1-UTC=-0.1084622, Accumulated Leap Seconds=37
  { 2459538.50, 69.292189 }, //20 November 2021, UT1-UTC=-0.1081890, Accumulated Leap Seconds=37
  { 2459539.50, 69.291770 }, //21 November 2021, UT1-UTC=-0.1077701, Accumulated Leap Seconds=37
  { 2459540.50, 69.291312 }, //22 November 2021, UT1-UTC=-0.1073121, Accumulated Leap Seconds=37
  { 2459541.50, 69.290739 }, //23 November 2021, UT1-UTC=-0.1067392, Accumulated Leap Seconds=37
  { 2459542.50, 69.290211 }, //24 November 2021, UT1-UTC=-0.1062114, Accumulated Leap Seconds=37
  { 2459543.50, 69.289727 }, //25 November 2021, UT1-UTC=-0.1057271, Accumulated Leap Seconds=37
  { 2459544.50, 69.289380 }, //26 November 2021, UT1-UTC=-0.1053797, Accumulated Leap Seconds=37
  { 2459545.50, 69.289317 }, //27 November 2021, UT1-UTC=-0.1053166, Accumulated Leap Seconds=37
  { 2459546.50, 69.289510 }, //28 November 2021, UT1-UTC=-0.1055097, Accumulated Leap Seconds=37
  { 2459547.50, 69.289845 }, //29 November 2021, UT1-UTC=-0.1058449, Accumulated Leap Seconds=37
  { 2459548.50, 69.290250 }, //30 November 2021, UT1-UTC=-0.1062498, Accumulated Leap Seconds=37
  { 2459549.50, 69.290806 }, //01 December 2021, UT1-UTC=-0.1068058, Accumulated Leap Seconds=37
  { 2459550.50, 69.291335 }, //02 December 2021, UT1-UTC=-0.1073352, Accumulated Leap Seconds=37
  { 2459551.50, 69.291760 }, //03 December 2021, UT1-UTC=-0.1077605, Accumulated Leap Seconds=37
  { 2459552.50, 69.291887 }, //04 December 2021, UT1-UTC=-0.1078868, Accumulated Leap Seconds=37
  { 2459553.50, 69.291759 }, //05 December 2021, UT1-UTC=-0.1077592, Accumulated Leap Seconds=37
  { 2459554.50, 69.291523 }, //06 December 2021, UT1-UTC=-0.1075231, Accumulated Leap Seconds=37
  { 2459555.50, 69.291278 }, //07 December 2021, UT1-UTC=-0.1072782, Accumulated Leap Seconds=37
  { 2459556.50, 69.291059 }, //08 December 2021, UT1-UTC=-0.1070594, Accumulated Leap Seconds=37
  { 2459557.50, 69.291089 }, //09 December 2021, UT1-UTC=-0.1070893, Accumulated Leap Seconds=37
  { 2459558.50, 69.291391 }, //10 December 2021, UT1-UTC=-0.1073912, Accumulated Leap Seconds=37
  { 2459559.50, 69.291747 }, //11 December 2021, UT1-UTC=-0.1077471, Accumulated Leap Seconds=37
  { 2459560.50, 69.292188 }, //12 December 2021, UT1-UTC=-0.1081883, Accumulated Leap Seconds=37
  { 2459561.50, 69.292639 }, //13 December 2021, UT1-UTC=-0.1086391, Accumulated Leap Seconds=37
  { 2459562.50, 69.292959 }, //14 December 2021, UT1-UTC=-0.1089586, Accumulated Leap Seconds=37
  { 2459563.50, 69.293194 }, //15 December 2021, UT1-UTC=-0.1091937, Accumulated Leap Seconds=37
  { 2459564.50, 69.293294 }, //16 December 2021, UT1-UTC=-0.1092943, Accumulated Leap Seconds=37
  { 2459565.50, 69.293265 }, //17 December 2021, UT1-UTC=-0.1092654, Accumulated Leap Seconds=37
  { 2459566.50, 69.293035 }, //18 December 2021, UT1-UTC=-0.1090354, Accumulated Leap Seconds=37
  { 2459567.50, 69.292734 }, //19 December 2021, UT1-UTC=-0.1087344, Accumulated Leap Seconds=37
  { 2459568.50, 69.292393 }, //20 December 2021, UT1-UTC=-0.1083929, Accumulated Leap Seconds=37
  { 2459569.50, 69.291998 }, //21 December 2021, UT1-UTC=-0.1079977, Accumulated Leap Seconds=37
  { 2459570.50, 69.291593 }, //22 December 2021, UT1-UTC=-0.1075927, Accumulated Leap Seconds=37
  { 2459571.50, 69.291320 }, //23 December 2021, UT1-UTC=-0.1073202, Accumulated Leap Seconds=37
  { 2459572.50, 69.291257 }, //24 December 2021, UT1-UTC=-0.1072569, Accumulated Leap Seconds=37
  { 2459573.50, 69.291367 }, //25 December 2021, UT1-UTC=-0.1073673, Accumulated Leap Seconds=37
  { 2459574.50, 69.291722 }, //26 December 2021, UT1-UTC=-0.1077217, Accumulated Leap Seconds=37
  { 2459575.50, 69.292293 }, //27 December 2021, UT1-UTC=-0.1082929, Accumulated Leap Seconds=37
  { 2459576.50, 69.292922 }, //28 December 2021, UT1-UTC=-0.1089216, Accumulated Leap Seconds=37
  { 2459577.50, 69.293553 }, //29 December 2021, UT1-UTC=-0.1095530, Accumulated Leap Seconds=37
  { 2459578.50, 69.294095 }, //30 December 2021, UT1-UTC=-0.1100946, Accumulated Leap Seconds=37
  { 2459579.50, 69.294421 }, //31 December 2021, UT1-UTC=-0.1104206, Accumulated Leap Seconds=37
  { 2459580.50, 69.294493 }, //01 January 2022, UT1-UTC=-0.1104934, Accumulated Leap Seconds=37
  { 2459581.50, 69.294360 }, //02 January 2022, UT1-UTC=-0.1103601, Accumulated Leap Seconds=37
  { 2459582.50, 69.294124 }, //03 January 2022, UT1-UTC=-0.1101240, Accumulated Leap Seconds=37
  { 2459583.50, 69.293934 }, //04 January 2022, UT1-UTC=-0.1099342, Accumulated Leap Seconds=37
  { 2459584.50, 69.293851 }, //05 January 2022, UT1-UTC=-0.1098513, Accumulated Leap Seconds=37
  { 2459585.50, 69.294062 }, //06 January 2022, UT1-UTC=-0.1100624, Accumulated Leap Seconds=37
  { 2459586.50, 69.294370 }, //07 January 2022, UT1-UTC=-0.1103699, Accumulated Leap Seconds=37
  { 2459587.50, 69.294774 }, //08 January 2022, UT1-UTC=-0.1107744, Accumulated Leap Seconds=37
  { 2459588.50, 69.295219 }, //09 January 2022, UT1-UTC=-0.1112190, Accumulated Leap Seconds=37
  { 2459589.50, 69.295608 }, //10 January 2022, UT1-UTC=-0.1116079, Accumulated Leap Seconds=37
  { 2459590.50, 69.295845 }, //11 January 2022, UT1-UTC=-0.1118450, Accumulated Leap Seconds=37
  { 2459591.50, 69.295926 }, //12 January 2022, UT1-UTC=-0.1119256, Accumulated Leap Seconds=37
  { 2459592.50, 69.295866 }, //13 January 2022, UT1-UTC=-0.1118664, Accumulated Leap Seconds=37
  { 2459593.50, 69.295600 }, //14 January 2022, UT1-UTC=-0.1116003, Accumulated Leap Seconds=37
  { 2459594.50, 69.295109 }, //15 January 2022, UT1-UTC=-0.1111090, Accumulated Leap Seconds=37
  { 2459595.50, 69.294570 }, //16 January 2022, UT1-UTC=-0.1105700, Accumulated Leap Seconds=37
  { 2459596.50, 69.293999 }, //17 January 2022, UT1-UTC=-0.1099985, Accumulated Leap Seconds=37
  { 2459597.50, 69.293337 }, //18 January 2022, UT1-UTC=-0.1093370, Accumulated Leap Seconds=37
  { 2459598.50, 69.292800 }, //19 January 2022, UT1-UTC=-0.1087995, Accumulated Leap Seconds=37
  { 2459599.50, 69.292345 }, //20 January 2022, UT1-UTC=-0.1083454, Accumulated Leap Seconds=37
  { 2459600.50, 69.292219 }, //21 January 2022, UT1-UTC=-0.1082186, Accumulated Leap Seconds=37, Predicted value
  { 2459601.50, 69.292288 }, //22 January 2022, UT1-UTC=-0.1082882, Accumulated Leap Seconds=37, Predicted value
  { 2459602.50, 69.292510 }, //23 January 2022, UT1-UTC=-0.1085105, Accumulated Leap Seconds=37, Predicted value
  { 2459603.50, 69.292831 }, //24 January 2022, UT1-UTC=-0.1088312, Accumulated Leap Seconds=37, Predicted value
  { 2459604.50, 69.293167 }, //25 January 2022, UT1-UTC=-0.1091668, Accumulated Leap Seconds=37, Predicted value
  { 2459605.50, 69.293425 }, //26 January 2022, UT1-UTC=-0.1094246, Accumulated Leap Seconds=37, Predicted value
  { 2459606.50, 69.293511 }, //27 January 2022, UT1-UTC=-0.1095115, Accumulated Leap Seconds=37, Predicted value
  { 2459607.50, 69.293381 }, //28 January 2022, UT1-UTC=-0.1093807, Accumulated Leap Seconds=37, Predicted value
  { 2459608.50, 69.293053 }, //29 January 2022, UT1-UTC=-0.1090531, Accumulated Leap Seconds=37, Predicted value
  { 2459609.50, 69.292617 }, //30 January 2022, UT1-UTC=-0.1086167, Accumulated Leap Seconds=37, Predicted value
  { 2459610.50, 69.292201 }, //31 January 2022, UT1-UTC=-0.1082011, Accumulated Leap Seconds=37, Predicted value
  { 2459611.50, 69.291941 }, //01 February 2022, UT1-UTC=-0.1079409, Accumulated Leap Seconds=37, Predicted value
  { 2459612.50, 69.291884 }, //02 February 2022, UT1-UTC=-0.1078838, Accumulated Leap Seconds=37, Predicted value
  { 2459613.50, 69.292061 }, //03 February 2022, UT1-UTC=-0.1080607, Accumulated Leap Seconds=37, Predicted value
  { 2459614.50, 69.292434 }, //04 February 2022, UT1-UTC=-0.1084341, Accumulated Leap Seconds=37, Predicted value
  { 2459615.50, 69.292911 }, //05 February 2022, UT1-UTC=-0.1089114, Accumulated Leap Seconds=37, Predicted value
  { 2459616.50, 69.293394 }, //06 February 2022, UT1-UTC=-0.1093939, Accumulated Leap Seconds=37, Predicted value
  { 2459617.50, 69.293800 }, //07 February 2022, UT1-UTC=-0.1098004, Accumulated Leap Seconds=37, Predicted value
  { 2459618.50, 69.294074 }, //08 February 2022, UT1-UTC=-0.1100739, Accumulated Leap Seconds=37, Predicted value
  { 2459619.50, 69.294192 }, //09 February 2022, UT1-UTC=-0.1101923, Accumulated Leap Seconds=37, Predicted value
  { 2459620.50, 69.294148 }, //10 February 2022, UT1-UTC=-0.1101484, Accumulated Leap Seconds=37, Predicted value
  { 2459621.50, 69.293965 }, //11 February 2022, UT1-UTC=-0.1099646, Accumulated Leap Seconds=37, Predicted value
  { 2459622.50, 69.293685 }, //12 February 2022, UT1-UTC=-0.1096847, Accumulated Leap Seconds=37, Predicted value
  { 2459623.50, 69.293367 }, //13 February 2022, UT1-UTC=-0.1093674, Accumulated Leap Seconds=37, Predicted value
  { 2459624.50, 69.293087 }, //14 February 2022, UT1-UTC=-0.1090872, Accumulated Leap Seconds=37, Predicted value
  { 2459625.50, 69.292914 }, //15 February 2022, UT1-UTC=-0.1089139, Accumulated Leap Seconds=37, Predicted value
  { 2459626.50, 69.292912 }, //16 February 2022, UT1-UTC=-0.1089119, Accumulated Leap Seconds=37, Predicted value
  { 2459627.50, 69.293131 }, //17 February 2022, UT1-UTC=-0.1091307, Accumulated Leap Seconds=37, Predicted value
  { 2459628.50, 69.293583 }, //18 February 2022, UT1-UTC=-0.1095826, Accumulated Leap Seconds=37, Predicted value
  { 2459629.50, 69.294236 }, //19 February 2022, UT1-UTC=-0.1102361, Accumulated Leap Seconds=37, Predicted value
  { 2459630.50, 69.295020 }, //20 February 2022, UT1-UTC=-0.1110199, Accumulated Leap Seconds=37, Predicted value
  { 2459631.50, 69.295839 }, //21 February 2022, UT1-UTC=-0.1118394, Accumulated Leap Seconds=37, Predicted value
  { 2459632.50, 69.296591 }, //22 February 2022, UT1-UTC=-0.1125912, Accumulated Leap Seconds=37, Predicted value
  { 2459633.50, 69.297182 }, //23 February 2022, UT1-UTC=-0.1131821, Accumulated Leap Seconds=37, Predicted value
  { 2459634.50, 69.297556 }, //24 February 2022, UT1-UTC=-0.1135555, Accumulated Leap Seconds=37, Predicted value
  { 2459635.50, 69.297710 }, //25 February 2022, UT1-UTC=-0.1137105, Accumulated Leap Seconds=37, Predicted value
  { 2459636.50, 69.297711 }, //26 February 2022, UT1-UTC=-0.1137107, Accumulated Leap Seconds=37, Predicted value
  { 2459637.50, 69.297661 }, //27 February 2022, UT1-UTC=-0.1136613, Accumulated Leap Seconds=37, Predicted value
  { 2459638.50, 69.297691 }, //28 February 2022, UT1-UTC=-0.1136913, Accumulated Leap Seconds=37, Predicted value
  { 2459639.50, 69.297905 }, //01 March 2022, UT1-UTC=-0.1139045, Accumulated Leap Seconds=37, Predicted value
  { 2459640.50, 69.298341 }, //02 March 2022, UT1-UTC=-0.1143407, Accumulated Leap Seconds=37, Predicted value
  { 2459641.50, 69.298977 }, //03 March 2022, UT1-UTC=-0.1149767, Accumulated Leap Seconds=37, Predicted value
  { 2459642.50, 69.299733 }, //04 March 2022, UT1-UTC=-0.1157326, Accumulated Leap Seconds=37, Predicted value
  { 2459643.50, 69.300505 }, //05 March 2022, UT1-UTC=-0.1165054, Accumulated Leap Seconds=37, Predicted value
  { 2459644.50, 69.301198 }, //06 March 2022, UT1-UTC=-0.1171978, Accumulated Leap Seconds=37, Predicted value
  { 2459645.50, 69.301743 }, //07 March 2022, UT1-UTC=-0.1177434, Accumulated Leap Seconds=37, Predicted value
  { 2459646.50, 69.302107 }, //08 March 2022, UT1-UTC=-0.1181072, Accumulated Leap Seconds=37, Predicted value
  { 2459647.50, 69.302278 }, //09 March 2022, UT1-UTC=-0.1182783, Accumulated Leap Seconds=37, Predicted value
  { 2459648.50, 69.302273 }, //10 March 2022, UT1-UTC=-0.1182733, Accumulated Leap Seconds=37, Predicted value
  { 2459649.50, 69.302130 }, //11 March 2022, UT1-UTC=-0.1181301, Accumulated Leap Seconds=37, Predicted value
  { 2459650.50, 69.301911 }, //12 March 2022, UT1-UTC=-0.1179111, Accumulated Leap Seconds=37, Predicted value
  { 2459651.50, 69.301689 }, //13 March 2022, UT1-UTC=-0.1176895, Accumulated Leap Seconds=37, Predicted value
  { 2459652.50, 69.301539 }, //14 March 2022, UT1-UTC=-0.1175391, Accumulated Leap Seconds=37, Predicted value
  { 2459653.50, 69.301529 }, //15 March 2022, UT1-UTC=-0.1175289, Accumulated Leap Seconds=37, Predicted value
  { 2459654.50, 69.301719 }, //16 March 2022, UT1-UTC=-0.1177188, Accumulated Leap Seconds=37, Predicted value
  { 2459655.50, 69.302140 }, //17 March 2022, UT1-UTC=-0.1181400, Accumulated Leap Seconds=37, Predicted value
  { 2459656.50, 69.302785 }, //18 March 2022, UT1-UTC=-0.1187855, Accumulated Leap Seconds=37, Predicted value
  { 2459657.50, 69.303601 }, //19 March 2022, UT1-UTC=-0.1196007, Accumulated Leap Seconds=37, Predicted value
  { 2459658.50, 69.304482 }, //20 March 2022, UT1-UTC=-0.1204824, Accumulated Leap Seconds=37, Predicted value
  { 2459659.50, 69.305307 }, //21 March 2022, UT1-UTC=-0.1213070, Accumulated Leap Seconds=37, Predicted value
  { 2459660.50, 69.305965 }, //22 March 2022, UT1-UTC=-0.1219649, Accumulated Leap Seconds=37, Predicted value
  { 2459661.50, 69.306393 }, //23 March 2022, UT1-UTC=-0.1223926, Accumulated Leap Seconds=37, Predicted value
  { 2459662.50, 69.306586 }, //24 March 2022, UT1-UTC=-0.1225862, Accumulated Leap Seconds=37, Predicted value
  { 2459663.50, 69.306601 }, //25 March 2022, UT1-UTC=-0.1226007, Accumulated Leap Seconds=37, Predicted value
  { 2459664.50, 69.306536 }, //26 March 2022, UT1-UTC=-0.1225364, Accumulated Leap Seconds=37, Predicted value
  { 2459665.50, 69.306508 }, //27 March 2022, UT1-UTC=-0.1225082, Accumulated Leap Seconds=37, Predicted value
  { 2459666.50, 69.306619 }, //28 March 2022, UT1-UTC=-0.1226193, Accumulated Leap Seconds=37, Predicted value
  { 2459667.50, 69.306935 }, //29 March 2022, UT1-UTC=-0.1229351, Accumulated Leap Seconds=37, Predicted value
  { 2459668.50, 69.307461 }, //30 March 2022, UT1-UTC=-0.1234614, Accumulated Leap Seconds=37, Predicted value
  { 2459669.50, 69.308148 }, //31 March 2022, UT1-UTC=-0.1241476, Accumulated Leap Seconds=37, Predicted value
  { 2459670.50, 69.308904 }, //01 April 2022, UT1-UTC=-0.1249036, Accumulated Leap Seconds=37, Predicted value
  { 2459671.50, 69.309626 }, //02 April 2022, UT1-UTC=-0.1256262, Accumulated Leap Seconds=37, Predicted value
  { 2459672.50, 69.310225 }, //03 April 2022, UT1-UTC=-0.1262246, Accumulated Leap Seconds=37, Predicted value
  { 2459673.50, 69.310641 }, //04 April 2022, UT1-UTC=-0.1266406, Accumulated Leap Seconds=37, Predicted value
  { 2459674.50, 69.310849 }, //05 April 2022, UT1-UTC=-0.1268491, Accumulated Leap Seconds=37, Predicted value
  { 2459675.50, 69.310856 }, //06 April 2022, UT1-UTC=-0.1268565, Accumulated Leap Seconds=37, Predicted value
  { 2459676.50, 69.310698 }, //07 April 2022, UT1-UTC=-0.1266980, Accumulated Leap Seconds=37, Predicted value
  { 2459677.50, 69.310430 }, //08 April 2022, UT1-UTC=-0.1264299, Accumulated Leap Seconds=37, Predicted value
  { 2459678.50, 69.310120 }, //09 April 2022, UT1-UTC=-0.1261201, Accumulated Leap Seconds=37, Predicted value
  { 2459679.50, 69.309843 }, //10 April 2022, UT1-UTC=-0.1258435, Accumulated Leap Seconds=37, Predicted value
  { 2459680.50, 69.309674 }, //11 April 2022, UT1-UTC=-0.1256741, Accumulated Leap Seconds=37, Predicted value
  { 2459681.50, 69.309679 }, //12 April 2022, UT1-UTC=-0.1256788, Accumulated Leap Seconds=37, Predicted value
  { 2459682.50, 69.309904 }, //13 April 2022, UT1-UTC=-0.1259039, Accumulated Leap Seconds=37, Predicted value
  { 2459683.50, 69.310364 }, //14 April 2022, UT1-UTC=-0.1263639, Accumulated Leap Seconds=37, Predicted value
  { 2459684.50, 69.311026 }, //15 April 2022, UT1-UTC=-0.1270257, Accumulated Leap Seconds=37, Predicted value
  { 2459685.50, 69.311806 }, //16 April 2022, UT1-UTC=-0.1278061, Accumulated Leap Seconds=37, Predicted value
  { 2459686.50, 69.312581 }, //17 April 2022, UT1-UTC=-0.1285808, Accumulated Leap Seconds=37, Predicted value
  { 2459687.50, 69.313218 }, //18 April 2022, UT1-UTC=-0.1292181, Accumulated Leap Seconds=37, Predicted value
  { 2459688.50, 69.313619 }, //19 April 2022, UT1-UTC=-0.1296191, Accumulated Leap Seconds=37, Predicted value
  { 2459689.50, 69.313752 }, //20 April 2022, UT1-UTC=-0.1297518, Accumulated Leap Seconds=37, Predicted value
  { 2459690.50, 69.313660 }, //21 April 2022, UT1-UTC=-0.1296599, Accumulated Leap Seconds=37, Predicted value
  { 2459691.50, 69.313443 }, //22 April 2022, UT1-UTC=-0.1294435, Accumulated Leap Seconds=37, Predicted value
  { 2459692.50, 69.313226 }, //23 April 2022, UT1-UTC=-0.1292258, Accumulated Leap Seconds=37, Predicted value
  { 2459693.50, 69.313117 }, //24 April 2022, UT1-UTC=-0.1291171, Accumulated Leap Seconds=37, Predicted value
  { 2459694.50, 69.313189 }, //25 April 2022, UT1-UTC=-0.1291892, Accumulated Leap Seconds=37, Predicted value
  { 2459695.50, 69.313462 }, //26 April 2022, UT1-UTC=-0.1294615, Accumulated Leap Seconds=37, Predicted value
  { 2459696.50, 69.313899 }, //27 April 2022, UT1-UTC=-0.1298989, Accumulated Leap Seconds=37, Predicted value
  { 2459697.50, 69.314428 }, //28 April 2022, UT1-UTC=-0.1304280, Accumulated Leap Seconds=37, Predicted value
  { 2459698.50, 69.314957 }, //29 April 2022, UT1-UTC=-0.1309568, Accumulated Leap Seconds=37, Predicted value
  { 2459699.50, 69.315397 }, //30 April 2022, UT1-UTC=-0.1313966, Accumulated Leap Seconds=37, Predicted value
  { 2459700.50, 69.315678 }, //01 May 2022, UT1-UTC=-0.1316784, Accumulated Leap Seconds=37, Predicted value
  { 2459701.50, 69.315763 }, //02 May 2022, UT1-UTC=-0.1317633, Accumulated Leap Seconds=37, Predicted value
  { 2459702.50, 69.315644 }, //03 May 2022, UT1-UTC=-0.1316442, Accumulated Leap Seconds=37, Predicted value
  { 2459703.50, 69.315344 }, //04 May 2022, UT1-UTC=-0.1313440, Accumulated Leap Seconds=37, Predicted value
  { 2459704.50, 69.314908 }, //05 May 2022, UT1-UTC=-0.1309081, Accumulated Leap Seconds=37, Predicted value
  { 2459705.50, 69.314400 }, //06 May 2022, UT1-UTC=-0.1304001, Accumulated Leap Seconds=37, Predicted value
  { 2459706.50, 69.313892 }, //07 May 2022, UT1-UTC=-0.1298918, Accumulated Leap Seconds=37, Predicted value
  { 2459707.50, 69.313454 }, //08 May 2022, UT1-UTC=-0.1294538, Accumulated Leap Seconds=37, Predicted value
  { 2459708.50, 69.313150 }, //09 May 2022, UT1-UTC=-0.1291500, Accumulated Leap Seconds=37, Predicted value
  { 2459709.50, 69.313031 }, //10 May 2022, UT1-UTC=-0.1290306, Accumulated Leap Seconds=37, Predicted value
  { 2459710.50, 69.313124 }, //11 May 2022, UT1-UTC=-0.1291239, Accumulated Leap Seconds=37, Predicted value
  { 2459711.50, 69.313425 }, //12 May 2022, UT1-UTC=-0.1294251, Accumulated Leap Seconds=37, Predicted value
  { 2459712.50, 69.313883 }, //13 May 2022, UT1-UTC=-0.1298827, Accumulated Leap Seconds=37, Predicted value
  { 2459713.50, 69.314397 }, //14 May 2022, UT1-UTC=-0.1303966, Accumulated Leap Seconds=37, Predicted value
  { 2459714.50, 69.314834 }, //15 May 2022, UT1-UTC=-0.1308339, Accumulated Leap Seconds=37, Predicted value
  { 2459715.50, 69.315070 }, //16 May 2022, UT1-UTC=-0.1310700, Accumulated Leap Seconds=37, Predicted value
  { 2459716.50, 69.315032 }, //17 May 2022, UT1-UTC=-0.1310318, Accumulated Leap Seconds=37, Predicted value
  { 2459717.50, 69.314727 }, //18 May 2022, UT1-UTC=-0.1307273, Accumulated Leap Seconds=37, Predicted value
  { 2459718.50, 69.314241 }, //19 May 2022, UT1-UTC=-0.1302410, Accumulated Leap Seconds=37, Predicted value
  { 2459719.50, 69.313702 }, //20 May 2022, UT1-UTC=-0.1297018, Accumulated Leap Seconds=37, Predicted value
  { 2459720.50, 69.313235 }, //21 May 2022, UT1-UTC=-0.1292353, Accumulated Leap Seconds=37, Predicted value
  { 2459721.50, 69.312924 }, //22 May 2022, UT1-UTC=-0.1289244, Accumulated Leap Seconds=37, Predicted value
  { 2459722.50, 69.312792 }, //23 May 2022, UT1-UTC=-0.1287924, Accumulated Leap Seconds=37, Predicted value
  { 2459723.50, 69.312807 }, //24 May 2022, UT1-UTC=-0.1288070, Accumulated Leap Seconds=37, Predicted value
  { 2459724.50, 69.312900 }, //25 May 2022, UT1-UTC=-0.1288996, Accumulated Leap Seconds=37, Predicted value
  { 2459725.50, 69.312986 }, //26 May 2022, UT1-UTC=-0.1289862, Accumulated Leap Seconds=37, Predicted value
  { 2459726.50, 69.312985 }, //27 May 2022, UT1-UTC=-0.1289853, Accumulated Leap Seconds=37, Predicted value
  { 2459727.50, 69.312831 }, //28 May 2022, UT1-UTC=-0.1288306, Accumulated Leap Seconds=37, Predicted value
  { 2459728.50, 69.312479 }, //29 May 2022, UT1-UTC=-0.1284787, Accumulated Leap Seconds=37, Predicted value
  { 2459729.50, 69.311914 }, //30 May 2022, UT1-UTC=-0.1279141, Accumulated Leap Seconds=37, Predicted value
  { 2459730.50, 69.311150 }, //31 May 2022, UT1-UTC=-0.1271496, Accumulated Leap Seconds=37, Predicted value
  { 2459731.50, 69.310223 }, //01 June 2022, UT1-UTC=-0.1262234, Accumulated Leap Seconds=37, Predicted value
  { 2459732.50, 69.309196 }, //02 June 2022, UT1-UTC=-0.1251957, Accumulated Leap Seconds=37, Predicted value
  { 2459733.50, 69.308138 }, //03 June 2022, UT1-UTC=-0.1241379, Accumulated Leap Seconds=37, Predicted value
  { 2459734.50, 69.307124 }, //04 June 2022, UT1-UTC=-0.1231235, Accumulated Leap Seconds=37, Predicted value
  { 2459735.50, 69.306217 }, //05 June 2022, UT1-UTC=-0.1222174, Accumulated Leap Seconds=37, Predicted value
  { 2459736.50, 69.305469 }, //06 June 2022, UT1-UTC=-0.1214692, Accumulated Leap Seconds=37, Predicted value
  { 2459737.50, 69.304910 }, //07 June 2022, UT1-UTC=-0.1209096, Accumulated Leap Seconds=37, Predicted value
  { 2459738.50, 69.304544 }, //08 June 2022, UT1-UTC=-0.1205441, Accumulated Leap Seconds=37, Predicted value
  { 2459739.50, 69.304345 }, //09 June 2022, UT1-UTC=-0.1203449, Accumulated Leap Seconds=37, Predicted value
  { 2459740.50, 69.304243 }, //10 June 2022, UT1-UTC=-0.1202432, Accumulated Leap Seconds=37, Predicted value
  { 2459741.50, 69.304131 }, //11 June 2022, UT1-UTC=-0.1201309, Accumulated Leap Seconds=37, Predicted value
  { 2459742.50, 69.303885 }, //12 June 2022, UT1-UTC=-0.1198853, Accumulated Leap Seconds=37, Predicted value
  { 2459743.50, 69.303407 }, //13 June 2022, UT1-UTC=-0.1194070, Accumulated Leap Seconds=37, Predicted value
  { 2459744.50, 69.302661 }, //14 June 2022, UT1-UTC=-0.1186614, Accumulated Leap Seconds=37, Predicted value
  { 2459745.50, 69.301699 }, //15 June 2022, UT1-UTC=-0.1176991, Accumulated Leap Seconds=37, Predicted value
  { 2459746.50, 69.300639 }, //16 June 2022, UT1-UTC=-0.1166389, Accumulated Leap Seconds=37, Predicted value
  { 2459747.50, 69.299622 }, //17 June 2022, UT1-UTC=-0.1156221, Accumulated Leap Seconds=37, Predicted value
  { 2459748.50, 69.298760 }, //18 June 2022, UT1-UTC=-0.1147597, Accumulated Leap Seconds=37, Predicted value
  { 2459749.50, 69.298097 }, //19 June 2022, UT1-UTC=-0.1140971, Accumulated Leap Seconds=37, Predicted value
  { 2459750.50, 69.297609 }, //20 June 2022, UT1-UTC=-0.1136089, Accumulated Leap Seconds=37, Predicted value
  { 2459751.50, 69.297220 }, //21 June 2022, UT1-UTC=-0.1132199, Accumulated Leap Seconds=37, Predicted value
  { 2459752.50, 69.296837 }, //22 June 2022, UT1-UTC=-0.1128366, Accumulated Leap Seconds=37, Predicted value
  { 2459753.50, 69.296372 }, //23 June 2022, UT1-UTC=-0.1123723, Accumulated Leap Seconds=37, Predicted value
  { 2459754.50, 69.295761 }, //24 June 2022, UT1-UTC=-0.1117612, Accumulated Leap Seconds=37, Predicted value
  { 2459755.50, 69.294962 }, //25 June 2022, UT1-UTC=-0.1109618, Accumulated Leap Seconds=37, Predicted value
  { 2459756.50, 69.293959 }, //26 June 2022, UT1-UTC=-0.1099588, Accumulated Leap Seconds=37, Predicted value
  { 2459757.50, 69.292761 }, //27 June 2022, UT1-UTC=-0.1087614, Accumulated Leap Seconds=37, Predicted value
  { 2459758.50, 69.291402 }, //28 June 2022, UT1-UTC=-0.1074021, Accumulated Leap Seconds=37, Predicted value
  { 2459759.50, 69.289935 }, //29 June 2022, UT1-UTC=-0.1059351, Accumulated Leap Seconds=37, Predicted value
  { 2459760.50, 69.288430 }, //30 June 2022, UT1-UTC=-0.1044297, Accumulated Leap Seconds=37, Predicted value
  { 2459761.50, 69.286959 }, //01 July 2022, UT1-UTC=-0.1029586, Accumulated Leap Seconds=37, Predicted value
  { 2459762.50, 69.285590 }, //02 July 2022, UT1-UTC=-0.1015899, Accumulated Leap Seconds=37, Predicted value
  { 2459763.50, 69.284376 }, //03 July 2022, UT1-UTC=-0.1003757, Accumulated Leap Seconds=37, Predicted value
  { 2459764.50, 69.283346 }, //04 July 2022, UT1-UTC=-0.0993462, Accumulated Leap Seconds=37, Predicted value
  { 2459765.50, 69.282506 }, //05 July 2022, UT1-UTC=-0.0985061, Accumulated Leap Seconds=37, Predicted value
  { 2459766.50, 69.281833 }, //06 July 2022, UT1-UTC=-0.0978329, Accumulated Leap Seconds=37, Predicted value
  { 2459767.50, 69.281274 }, //07 July 2022, UT1-UTC=-0.0972737, Accumulated Leap Seconds=37, Predicted value
  { 2459768.50, 69.280746 }, //08 July 2022, UT1-UTC=-0.0967456, Accumulated Leap Seconds=37, Predicted value
  { 2459769.50, 69.280145 }, //09 July 2022, UT1-UTC=-0.0961446, Accumulated Leap Seconds=37, Predicted value
  { 2459770.50, 69.279370 }, //10 July 2022, UT1-UTC=-0.0953703, Accumulated Leap Seconds=37, Predicted value
  { 2459771.50, 69.278363 }, //11 July 2022, UT1-UTC=-0.0943626, Accumulated Leap Seconds=37, Predicted value
  { 2459772.50, 69.277133 }, //12 July 2022, UT1-UTC=-0.0931326, Accumulated Leap Seconds=37, Predicted value
  { 2459773.50, 69.275768 }, //13 July 2022, UT1-UTC=-0.0917683, Accumulated Leap Seconds=37, Predicted value
  { 2459774.50, 69.274407 }, //14 July 2022, UT1-UTC=-0.0904066, Accumulated Leap Seconds=37, Predicted value
  { 2459775.50, 69.273183 }, //15 July 2022, UT1-UTC=-0.0891826, Accumulated Leap Seconds=37, Predicted value
  { 2459776.50, 69.272177 }, //16 July 2022, UT1-UTC=-0.0881766, Accumulated Leap Seconds=37, Predicted value
  { 2459777.50, 69.271389 }, //17 July 2022, UT1-UTC=-0.0873891, Accumulated Leap Seconds=37, Predicted value
  { 2459778.50, 69.270748 }, //18 July 2022, UT1-UTC=-0.0867480, Accumulated Leap Seconds=37, Predicted value
  { 2459779.50, 69.270143 }, //19 July 2022, UT1-UTC=-0.0861428, Accumulated Leap Seconds=37, Predicted value
  { 2459780.50, 69.269462 }, //20 July 2022, UT1-UTC=-0.0854623, Accumulated Leap Seconds=37, Predicted value
  { 2459781.50, 69.268621 }, //21 July 2022, UT1-UTC=-0.0846213, Accumulated Leap Seconds=37, Predicted value
  { 2459782.50, 69.267569 }, //22 July 2022, UT1-UTC=-0.0835692, Accumulated Leap Seconds=37, Predicted value
  { 2459783.50, 69.266288 }, //23 July 2022, UT1-UTC=-0.0822875, Accumulated Leap Seconds=37, Predicted value
  { 2459784.50, 69.264785 }, //24 July 2022, UT1-UTC=-0.0807847, Accumulated Leap Seconds=37, Predicted value
  { 2459785.50, 69.263092 }, //25 July 2022, UT1-UTC=-0.0790924, Accumulated Leap Seconds=37, Predicted value
  { 2459786.50, 69.261264 }, //26 July 2022, UT1-UTC=-0.0772642, Accumulated Leap Seconds=37, Predicted value
  { 2459787.50, 69.259372 }, //27 July 2022, UT1-UTC=-0.0753721, Accumulated Leap Seconds=37, Predicted value
  { 2459788.50, 69.257501 }, //28 July 2022, UT1-UTC=-0.0735012, Accumulated Leap Seconds=37, Predicted value
  { 2459789.50, 69.255740 }, //29 July 2022, UT1-UTC=-0.0717401, Accumulated Leap Seconds=37, Predicted value
  { 2459790.50, 69.254169 }, //30 July 2022, UT1-UTC=-0.0701688, Accumulated Leap Seconds=37, Predicted value
  { 2459791.50, 69.252845 }, //31 July 2022, UT1-UTC=-0.0688451, Accumulated Leap Seconds=37, Predicted value
  { 2459792.50, 69.251793 }, //01 August 2022, UT1-UTC=-0.0677932, Accumulated Leap Seconds=37, Predicted value
  { 2459793.50, 69.250998 }, //02 August 2022, UT1-UTC=-0.0669985, Accumulated Leap Seconds=37, Predicted value
  { 2459794.50, 69.250410 }, //03 August 2022, UT1-UTC=-0.0664103, Accumulated Leap Seconds=37, Predicted value
  { 2459795.50, 69.249949 }, //04 August 2022, UT1-UTC=-0.0659493, Accumulated Leap Seconds=37, Predicted value
  { 2459796.50, 69.249519 }, //05 August 2022, UT1-UTC=-0.0655194, Accumulated Leap Seconds=37, Predicted value
  { 2459797.50, 69.249025 }, //06 August 2022, UT1-UTC=-0.0650253, Accumulated Leap Seconds=37, Predicted value
  { 2459798.50, 69.248395 }, //07 August 2022, UT1-UTC=-0.0643953, Accumulated Leap Seconds=37, Predicted value
  { 2459799.50, 69.247607 }, //08 August 2022, UT1-UTC=-0.0636066, Accumulated Leap Seconds=37, Predicted value
  { 2459800.50, 69.246702 }, //09 August 2022, UT1-UTC=-0.0627016, Accumulated Leap Seconds=37, Predicted value
  { 2459801.50, 69.245781 }, //10 August 2022, UT1-UTC=-0.0617811, Accumulated Leap Seconds=37, Predicted value
  { 2459802.50, 69.244971 }, //11 August 2022, UT1-UTC=-0.0609713, Accumulated Leap Seconds=37, Predicted value
  { 2459803.50, 69.244374 }, //12 August 2022, UT1-UTC=-0.0603745, Accumulated Leap Seconds=37, Predicted value
  { 2459804.50, 69.244025 }, //13 August 2022, UT1-UTC=-0.0600255, Accumulated Leap Seconds=37, Predicted value
  { 2459805.50, 69.243876 }, //14 August 2022, UT1-UTC=-0.0598760, Accumulated Leap Seconds=37, Predicted value
  { 2459806.50, 69.243814 }, //15 August 2022, UT1-UTC=-0.0598145, Accumulated Leap Seconds=37, Predicted value
  { 2459807.50, 69.243710 }, //16 August 2022, UT1-UTC=-0.0597099, Accumulated Leap Seconds=37, Predicted value
  { 2459808.50, 69.243453 }, //17 August 2022, UT1-UTC=-0.0594532, Accumulated Leap Seconds=37, Predicted value
  { 2459809.50, 69.242981 }, //18 August 2022, UT1-UTC=-0.0589808, Accumulated Leap Seconds=37, Predicted value
  { 2459810.50, 69.242274 }, //19 August 2022, UT1-UTC=-0.0582741, Accumulated Leap Seconds=37, Predicted value
  { 2459811.50, 69.241347 }, //20 August 2022, UT1-UTC=-0.0573470, Accumulated Leap Seconds=37, Predicted value
  { 2459812.50, 69.240233 }, //21 August 2022, UT1-UTC=-0.0562334, Accumulated Leap Seconds=37, Predicted value
  { 2459813.50, 69.238982 }, //22 August 2022, UT1-UTC=-0.0549823, Accumulated Leap Seconds=37, Predicted value
  { 2459814.50, 69.237656 }, //23 August 2022, UT1-UTC=-0.0536558, Accumulated Leap Seconds=37, Predicted value
  { 2459815.50, 69.236329 }, //24 August 2022, UT1-UTC=-0.0523294, Accumulated Leap Seconds=37, Predicted value
  { 2459816.50, 69.235093 }, //25 August 2022, UT1-UTC=-0.0510933, Accumulated Leap Seconds=37, Predicted value
  { 2459817.50, 69.234023 }, //26 August 2022, UT1-UTC=-0.0500229, Accumulated Leap Seconds=37, Predicted value
  { 2459818.50, 69.233147 }, //27 August 2022, UT1-UTC=-0.0491471, Accumulated Leap Seconds=37, Predicted value
  { 2459819.50, 69.232470 }, //28 August 2022, UT1-UTC=-0.0484703, Accumulated Leap Seconds=37, Predicted value
  { 2459820.50, 69.231990 }, //29 August 2022, UT1-UTC=-0.0479903, Accumulated Leap Seconds=37, Predicted value
  { 2459821.50, 69.231624 }, //30 August 2022, UT1-UTC=-0.0476239, Accumulated Leap Seconds=37, Predicted value
  { 2459822.50, 69.231289 }, //31 August 2022, UT1-UTC=-0.0472887, Accumulated Leap Seconds=37, Predicted value
  { 2459823.50, 69.230958 }, //01 September 2022, UT1-UTC=-0.0469576, Accumulated Leap Seconds=37, Predicted value
  { 2459824.50, 69.230597 }, //02 September 2022, UT1-UTC=-0.0465971, Accumulated Leap Seconds=37, Predicted value
  { 2459825.50, 69.230098 }, //03 September 2022, UT1-UTC=-0.0460982, Accumulated Leap Seconds=37, Predicted value
  { 2459826.50, 69.229493 }, //04 September 2022, UT1-UTC=-0.0454929, Accumulated Leap Seconds=37, Predicted value
  { 2459827.50, 69.228807 }, //05 September 2022, UT1-UTC=-0.0448065, Accumulated Leap Seconds=37, Predicted value
  { 2459828.50, 69.228043 }, //06 September 2022, UT1-UTC=-0.0440430, Accumulated Leap Seconds=37, Predicted value
  { 2459829.50, 69.227406 }, //07 September 2022, UT1-UTC=-0.0434064, Accumulated Leap Seconds=37, Predicted value
  { 2459830.50, 69.226985 }, //08 September 2022, UT1-UTC=-0.0429851, Accumulated Leap Seconds=37, Predicted value
  { 2459831.50, 69.226810 }, //09 September 2022, UT1-UTC=-0.0428101, Accumulated Leap Seconds=37, Predicted value
  { 2459832.50, 69.226931 }, //10 September 2022, UT1-UTC=-0.0429310, Accumulated Leap Seconds=37, Predicted value
  { 2459833.50, 69.227222 }, //11 September 2022, UT1-UTC=-0.0432219, Accumulated Leap Seconds=37, Predicted value
  { 2459834.50, 69.227675 }, //12 September 2022, UT1-UTC=-0.0436752, Accumulated Leap Seconds=37, Predicted value
  { 2459835.50, 69.228074 }, //13 September 2022, UT1-UTC=-0.0440736, Accumulated Leap Seconds=37, Predicted value
  { 2459836.50, 69.228288 }, //14 September 2022, UT1-UTC=-0.0442882, Accumulated Leap Seconds=37, Predicted value
  { 2459837.50, 69.228355 }, //15 September 2022, UT1-UTC=-0.0443549, Accumulated Leap Seconds=37, Predicted value
  { 2459838.50, 69.228150 }, //16 September 2022, UT1-UTC=-0.0441495, Accumulated Leap Seconds=37, Predicted value
  { 2459839.50, 69.227791 }, //17 September 2022, UT1-UTC=-0.0437909, Accumulated Leap Seconds=37, Predicted value
  { 2459840.50, 69.227282 }, //18 September 2022, UT1-UTC=-0.0432820, Accumulated Leap Seconds=37, Predicted value
  { 2459841.50, 69.226655 }, //19 September 2022, UT1-UTC=-0.0426552, Accumulated Leap Seconds=37, Predicted value
  { 2459842.50, 69.226078 }, //20 September 2022, UT1-UTC=-0.0420785, Accumulated Leap Seconds=37, Predicted value
  { 2459843.50, 69.225581 }, //21 September 2022, UT1-UTC=-0.0415806, Accumulated Leap Seconds=37, Predicted value
  { 2459844.50, 69.225241 }, //22 September 2022, UT1-UTC=-0.0412414, Accumulated Leap Seconds=37, Predicted value
  { 2459845.50, 69.225142 }, //23 September 2022, UT1-UTC=-0.0411419, Accumulated Leap Seconds=37, Predicted value
  { 2459846.50, 69.225282 }, //24 September 2022, UT1-UTC=-0.0412822, Accumulated Leap Seconds=37, Predicted value
  { 2459847.50, 69.225596 }, //25 September 2022, UT1-UTC=-0.0415957, Accumulated Leap Seconds=37, Predicted value
  { 2459848.50, 69.226065 }, //26 September 2022, UT1-UTC=-0.0420650, Accumulated Leap Seconds=37, Predicted value
  { 2459849.50, 69.226614 }, //27 September 2022, UT1-UTC=-0.0426140, Accumulated Leap Seconds=37, Predicted value
  { 2459850.50, 69.227139 }, //28 September 2022, UT1-UTC=-0.0431392, Accumulated Leap Seconds=37, Predicted value
  { 2459851.50, 69.227579 }, //29 September 2022, UT1-UTC=-0.0435786, Accumulated Leap Seconds=37, Predicted value
  { 2459852.50, 69.227895 }, //30 September 2022, UT1-UTC=-0.0438952, Accumulated Leap Seconds=37, Predicted value
  { 2459853.50, 69.228050 }, //01 October 2022, UT1-UTC=-0.0440503, Accumulated Leap Seconds=37, Predicted value
  { 2459854.50, 69.228031 }, //02 October 2022, UT1-UTC=-0.0440306, Accumulated Leap Seconds=37, Predicted value
  { 2459855.50, 69.227877 }, //03 October 2022, UT1-UTC=-0.0438771, Accumulated Leap Seconds=37, Predicted value
  { 2459856.50, 69.227743 }, //04 October 2022, UT1-UTC=-0.0437433, Accumulated Leap Seconds=37, Predicted value
  { 2459857.50, 69.227756 }, //05 October 2022, UT1-UTC=-0.0437561, Accumulated Leap Seconds=37, Predicted value
  { 2459858.50, 69.228010 }, //06 October 2022, UT1-UTC=-0.0440100, Accumulated Leap Seconds=37, Predicted value
  { 2459859.50, 69.228521 }, //07 October 2022, UT1-UTC=-0.0445205, Accumulated Leap Seconds=37, Predicted value
  { 2459860.50, 69.229403 }, //08 October 2022, UT1-UTC=-0.0454030, Accumulated Leap Seconds=37, Predicted value
  { 2459861.50, 69.230323 }, //09 October 2022, UT1-UTC=-0.0463231, Accumulated Leap Seconds=37, Predicted value
  { 2459862.50, 69.231295 }, //10 October 2022, UT1-UTC=-0.0472946, Accumulated Leap Seconds=37, Predicted value
  { 2459863.50, 69.232107 }, //11 October 2022, UT1-UTC=-0.0481068, Accumulated Leap Seconds=37, Predicted value
  { 2459864.50, 69.232748 }, //12 October 2022, UT1-UTC=-0.0487482, Accumulated Leap Seconds=37, Predicted value
  { 2459865.50, 69.233137 }, //13 October 2022, UT1-UTC=-0.0491373, Accumulated Leap Seconds=37, Predicted value
  { 2459866.50, 69.233244 }, //14 October 2022, UT1-UTC=-0.0492439, Accumulated Leap Seconds=37, Predicted value
  { 2459867.50, 69.233117 }, //15 October 2022, UT1-UTC=-0.0491167, Accumulated Leap Seconds=37, Predicted value
  { 2459868.50, 69.232782 }, //16 October 2022, UT1-UTC=-0.0487819, Accumulated Leap Seconds=37, Predicted value
  { 2459869.50, 69.232470 }, //17 October 2022, UT1-UTC=-0.0484705, Accumulated Leap Seconds=37, Predicted value
  { 2459870.50, 69.232094 }, //18 October 2022, UT1-UTC=-0.0480936, Accumulated Leap Seconds=37, Predicted value
  { 2459871.50, 69.231760 }, //19 October 2022, UT1-UTC=-0.0477603, Accumulated Leap Seconds=37, Predicted value
  { 2459872.50, 69.231582 }, //20 October 2022, UT1-UTC=-0.0475815, Accumulated Leap Seconds=37, Predicted value
  { 2459873.50, 69.231640 }, //21 October 2022, UT1-UTC=-0.0476402, Accumulated Leap Seconds=37, Predicted value
  { 2459874.50, 69.231858 }, //22 October 2022, UT1-UTC=-0.0478575, Accumulated Leap Seconds=37, Predicted value
  { 2459875.50, 69.232207 }, //23 October 2022, UT1-UTC=-0.0482072, Accumulated Leap Seconds=37, Predicted value
  { 2459876.50, 69.232667 }, //24 October 2022, UT1-UTC=-0.0486667, Accumulated Leap Seconds=37, Predicted value
  { 2459877.50, 69.233137 }, //25 October 2022, UT1-UTC=-0.0491374, Accumulated Leap Seconds=37, Predicted value
  { 2459878.50, 69.233579 }, //26 October 2022, UT1-UTC=-0.0495795, Accumulated Leap Seconds=37, Predicted value
  { 2459879.50, 69.233916 }, //27 October 2022, UT1-UTC=-0.0499162, Accumulated Leap Seconds=37, Predicted value
  { 2459880.50, 69.234051 }, //28 October 2022, UT1-UTC=-0.0500513, Accumulated Leap Seconds=37, Predicted value
  { 2459881.50, 69.233980 }, //29 October 2022, UT1-UTC=-0.0499799, Accumulated Leap Seconds=37, Predicted value
  { 2459882.50, 69.233806 }, //30 October 2022, UT1-UTC=-0.0498063, Accumulated Leap Seconds=37, Predicted value
  { 2459883.50, 69.233635 }, //31 October 2022, UT1-UTC=-0.0496349, Accumulated Leap Seconds=37, Predicted value
  { 2459884.50, 69.233580 }, //01 November 2022, UT1-UTC=-0.0495798, Accumulated Leap Seconds=37, Predicted value
  { 2459885.50, 69.233589 }, //02 November 2022, UT1-UTC=-0.0495886, Accumulated Leap Seconds=37, Predicted value
  { 2459886.50, 69.233826 }, //03 November 2022, UT1-UTC=-0.0498264, Accumulated Leap Seconds=37, Predicted value
  { 2459887.50, 69.234325 }, //04 November 2022, UT1-UTC=-0.0503253, Accumulated Leap Seconds=37, Predicted value
  { 2459888.50, 69.235083 }, //05 November 2022, UT1-UTC=-0.0510832, Accumulated Leap Seconds=37, Predicted value
  { 2459889.50, 69.235828 }, //06 November 2022, UT1-UTC=-0.0518284, Accumulated Leap Seconds=37, Predicted value
  { 2459890.50, 69.236515 }, //07 November 2022, UT1-UTC=-0.0525146, Accumulated Leap Seconds=37, Predicted value
  { 2459891.50, 69.237091 }, //08 November 2022, UT1-UTC=-0.0530906, Accumulated Leap Seconds=37, Predicted value
  { 2459892.50, 69.237389 }, //09 November 2022, UT1-UTC=-0.0533887, Accumulated Leap Seconds=37, Predicted value
  { 2459893.50, 69.237470 }, //10 November 2022, UT1-UTC=-0.0534698, Accumulated Leap Seconds=37, Predicted value
  { 2459894.50, 69.237316 }, //11 November 2022, UT1-UTC=-0.0533159, Accumulated Leap Seconds=37, Predicted value
  { 2459895.50, 69.236979 }, //12 November 2022, UT1-UTC=-0.0529790, Accumulated Leap Seconds=37, Predicted value
  { 2459896.50, 69.236516 }, //13 November 2022, UT1-UTC=-0.0525163, Accumulated Leap Seconds=37, Predicted value
  { 2459897.50, 69.236032 }, //14 November 2022, UT1-UTC=-0.0520320, Accumulated Leap Seconds=37, Predicted value
  { 2459898.50, 69.235561 }, //15 November 2022, UT1-UTC=-0.0515613, Accumulated Leap Seconds=37, Predicted value
  { 2459899.50, 69.235290 }, //16 November 2022, UT1-UTC=-0.0512903, Accumulated Leap Seconds=37, Predicted value
  { 2459900.50, 69.235170 }, //17 November 2022, UT1-UTC=-0.0511696, Accumulated Leap Seconds=37, Predicted value
  { 2459901.50, 69.235242 }, //18 November 2022, UT1-UTC=-0.0512415, Accumulated Leap Seconds=37, Predicted value
  { 2459902.50, 69.235549 }, //19 November 2022, UT1-UTC=-0.0515491, Accumulated Leap Seconds=37, Predicted value
  { 2459903.50, 69.235984 }, //20 November 2022, UT1-UTC=-0.0519835, Accumulated Leap Seconds=37, Predicted value
  { 2459904.50, 69.236561 }, //21 November 2022, UT1-UTC=-0.0525614, Accumulated Leap Seconds=37, Predicted value
  { 2459905.50, 69.237058 }, //22 November 2022, UT1-UTC=-0.0530581, Accumulated Leap Seconds=37, Predicted value
  { 2459906.50, 69.237310 }, //23 November 2022, UT1-UTC=-0.0533102, Accumulated Leap Seconds=37, Predicted value
  { 2459907.50, 69.237308 }, //24 November 2022, UT1-UTC=-0.0533084, Accumulated Leap Seconds=37, Predicted value
  { 2459908.50, 69.237043 }, //25 November 2022, UT1-UTC=-0.0530435, Accumulated Leap Seconds=37, Predicted value
  { 2459909.50, 69.236536 }, //26 November 2022, UT1-UTC=-0.0525365, Accumulated Leap Seconds=37, Predicted value
  { 2459910.50, 69.235916 }, //27 November 2022, UT1-UTC=-0.0519155, Accumulated Leap Seconds=37, Predicted value
  { 2459911.50, 69.235307 }, //28 November 2022, UT1-UTC=-0.0513070, Accumulated Leap Seconds=37, Predicted value
  { 2459912.50, 69.234867 }, //29 November 2022, UT1-UTC=-0.0508672, Accumulated Leap Seconds=37, Predicted value
  { 2459913.50, 69.234661 }, //30 November 2022, UT1-UTC=-0.0506609, Accumulated Leap Seconds=37, Predicted value
  { 2459914.50, 69.234618 }, //01 December 2022, UT1-UTC=-0.0506182, Accumulated Leap Seconds=37, Predicted value
  { 2459915.50, 69.234761 }, //02 December 2022, UT1-UTC=-0.0507608, Accumulated Leap Seconds=37, Predicted value
  { 2459916.50, 69.234939 }, //03 December 2022, UT1-UTC=-0.0509390, Accumulated Leap Seconds=37, Predicted value
  { 2459917.50, 69.235078 }, //04 December 2022, UT1-UTC=-0.0510784, Accumulated Leap Seconds=37, Predicted value
  { 2459918.50, 69.235081 }, //05 December 2022, UT1-UTC=-0.0510807, Accumulated Leap Seconds=37, Predicted value
  { 2459919.50, 69.234887 }, //06 December 2022, UT1-UTC=-0.0508874, Accumulated Leap Seconds=37, Predicted value
  { 2459920.50, 69.234450 }, //07 December 2022, UT1-UTC=-0.0504500, Accumulated Leap Seconds=37, Predicted value
  { 2459921.50, 69.233744 }, //08 December 2022, UT1-UTC=-0.0497437, Accumulated Leap Seconds=37, Predicted value
  { 2459922.50, 69.232833 }, //09 December 2022, UT1-UTC=-0.0488326, Accumulated Leap Seconds=37, Predicted value
  { 2459923.50, 69.231863 }, //10 December 2022, UT1-UTC=-0.0478633, Accumulated Leap Seconds=37, Predicted value
  { 2459924.50, 69.230830 }, //11 December 2022, UT1-UTC=-0.0468298, Accumulated Leap Seconds=37, Predicted value
  { 2459925.50, 69.229867 }, //12 December 2022, UT1-UTC=-0.0458673, Accumulated Leap Seconds=37, Predicted value
  { 2459926.50, 69.229089 }, //13 December 2022, UT1-UTC=-0.0450891, Accumulated Leap Seconds=37, Predicted value
  { 2459927.50, 69.228507 }, //14 December 2022, UT1-UTC=-0.0445070, Accumulated Leap Seconds=37, Predicted value
  { 2459928.50, 69.228134 }, //15 December 2022, UT1-UTC=-0.0441339, Accumulated Leap Seconds=37, Predicted value
  { 2459929.50, 69.227925 }, //16 December 2022, UT1-UTC=-0.0439250, Accumulated Leap Seconds=37, Predicted value
  { 2459930.50, 69.227858 }, //17 December 2022, UT1-UTC=-0.0438580, Accumulated Leap Seconds=37, Predicted value
  { 2459931.50, 69.227909 }, //18 December 2022, UT1-UTC=-0.0439086, Accumulated Leap Seconds=37, Predicted value
  { 2459932.50, 69.227987 }, //19 December 2022, UT1-UTC=-0.0439867, Accumulated Leap Seconds=37, Predicted value
  { 2459933.50, 69.227983 }, //20 December 2022, UT1-UTC=-0.0439830, Accumulated Leap Seconds=37, Predicted value
  { 2459934.50, 69.227835 }, //21 December 2022, UT1-UTC=-0.0438355, Accumulated Leap Seconds=37, Predicted value
  { 2459935.50, 69.227452 }, //22 December 2022, UT1-UTC=-0.0434525, Accumulated Leap Seconds=37, Predicted value
  { 2459936.50, 69.226846 }, //23 December 2022, UT1-UTC=-0.0428462, Accumulated Leap Seconds=37, Predicted value
  { 2459937.50, 69.226168 }, //24 December 2022, UT1-UTC=-0.0421684, Accumulated Leap Seconds=37, Predicted value
  { 2459938.50, 69.225524 }, //25 December 2022, UT1-UTC=-0.0415244, Accumulated Leap Seconds=37, Predicted value
  { 2459939.50, 69.224920 }, //26 December 2022, UT1-UTC=-0.0409202, Accumulated Leap Seconds=37, Predicted value
  { 2459940.50, 69.224543 }, //27 December 2022, UT1-UTC=-0.0405434, Accumulated Leap Seconds=37, Predicted value
  { 2459941.50, 69.224395 }, //28 December 2022, UT1-UTC=-0.0403950, Accumulated Leap Seconds=37, Predicted value
  { 2459942.50, 69.224370 }, //29 December 2022, UT1-UTC=-0.0403705, Accumulated Leap Seconds=37, Predicted value
  { 2459943.50, 69.224410 }, //30 December 2022, UT1-UTC=-0.0404101, Accumulated Leap Seconds=37, Predicted value
  { 2459944.50, 69.224376 }, //31 December 2022, UT1-UTC=-0.0403757, Accumulated Leap Seconds=37, Predicted value
  { 2459945.50, 69.224183 }, //01 January 2023, UT1-UTC=-0.0401835, Accumulated Leap Seconds=37, Predicted value
  { 2459946.50, 69.223743 }, //02 January 2023, UT1-UTC=-0.0397432, Accumulated Leap Seconds=37, Predicted value
  { 2459947.50, 69.223158 }, //03 January 2023, UT1-UTC=-0.0391579, Accumulated Leap Seconds=37, Predicted value
  { 2459948.50, 69.222394 }, //04 January 2023, UT1-UTC=-0.0383940, Accumulated Leap Seconds=37, Predicted value
  { 2459949.50, 69.221552 }, //05 January 2023, UT1-UTC=-0.0375522, Accumulated Leap Seconds=37, Predicted value
  { 2459950.50, 69.220718 }, //06 January 2023, UT1-UTC=-0.0367182, Accumulated Leap Seconds=37, Predicted value
  { 2459951.50, 69.219884 }, //07 January 2023, UT1-UTC=-0.0358842, Accumulated Leap Seconds=37, Predicted value
  { 2459952.50, 69.219188 }, //08 January 2023, UT1-UTC=-0.0351884, Accumulated Leap Seconds=37, Predicted value
  { 2459953.50, 69.218674 }, //09 January 2023, UT1-UTC=-0.0346738, Accumulated Leap Seconds=37, Predicted value
  { 2459954.50, 69.218343 }, //10 January 2023, UT1-UTC=-0.0343429, Accumulated Leap Seconds=37, Predicted value
  { 2459955.50, 69.218269 }, //11 January 2023, UT1-UTC=-0.0342689, Accumulated Leap Seconds=37, Predicted value
  { 2459956.50, 69.218372 }, //12 January 2023, UT1-UTC=-0.0343723, Accumulated Leap Seconds=37, Predicted value
  { 2459957.50, 69.218603 }, //13 January 2023, UT1-UTC=-0.0346025, Accumulated Leap Seconds=37, Predicted value
  { 2459958.50, 69.218937 }, //14 January 2023, UT1-UTC=-0.0349375, Accumulated Leap Seconds=37, Predicted value
  { 2459959.50, 69.219316 }, //15 January 2023, UT1-UTC=-0.0353163, Accumulated Leap Seconds=37, Predicted value
  { 2459960.50, 69.219666 }, //16 January 2023, UT1-UTC=-0.0356662, Accumulated Leap Seconds=37, Predicted value
  { 2459961.50, 69.219835 }, //17 January 2023, UT1-UTC=-0.0358354, Accumulated Leap Seconds=37, Predicted value
  { 2459962.50, 69.219780 }, //18 January 2023, UT1-UTC=-0.0357804, Accumulated Leap Seconds=37, Predicted value
  { 2459963.50, 69.219502 }, //19 January 2023, UT1-UTC=-0.0355019, Accumulated Leap Seconds=37, Predicted value
  { 2459964.50, 69.219072 }, //20 January 2023, UT1-UTC=-0.0350721, Accumulated Leap Seconds=37, Predicted value
  { 2459965.50, 69.218623 }, //21 January 2023, UT1-UTC=-0.0346234, Accumulated Leap Seconds=37, Predicted value
  { 2459966.50, 69.218325 }, //22 January 2023, UT1-UTC=-0.0343252, Accumulated Leap Seconds=37, Predicted value
  { 2459967.50, 69.218264 }, //23 January 2023, UT1-UTC=-0.0342642, Accumulated Leap Seconds=37, Predicted value
  { 2459968.50, 69.218444 }, //24 January 2023, UT1-UTC=-0.0344437, Accumulated Leap Seconds=37, Predicted value
  { 2459969.50, 69.218800 }, //25 January 2023, UT1-UTC=-0.0347995, Accumulated Leap Seconds=37, Predicted value
  { 2459970.50, 69.219181 }, //26 January 2023, UT1-UTC=-0.0351808, Accumulated Leap Seconds=37, Predicted value
  { 2459971.50, 69.219544 }, //27 January 2023, UT1-UTC=-0.0355438, Accumulated Leap Seconds=37, Predicted value
  { 2459972.50, 69.219801 }, //28 January 2023, UT1-UTC=-0.0358013, Accumulated Leap Seconds=37, Predicted value

//Values from https://cddis.nasa.gov/archive/products/iers/deltat.preds
  { 2460036.50, 71.540    }, //2023.25, Predicted value
  { 2460127.50, 71.670    }, //2023.50, Predicted value
  { 2460219.50, 71.740    }, //2023.75, Predicted value
  { 2460310.50, 71.880    }, //2024.00, Predicted value
  { 2460401.50, 72.030    }, //2024.25, Predicted value
  { 2460493.50, 72.150    }, //2024.50, Predicted value
  { 2460584.50, 72.220    }, //2024.75, Predicted value
  { 2460675.50, 72.360    }, //2025.00, Predicted value
  { 2460767.50, 72.500    }, //2025.25, Predicted value
  { 2460858.50, 72.620    }, //2025.50, Predicted value
  { 2460949.50, 72.690    }, //2025.75, Predicted value
  { 2461041.50, 72.830    }, //2026.00, Predicted value
  { 2461132.50, 72.980    }, //2026.25, Predicted value
  { 2461223.50, 73.100    }, //2026.50, Predicted value
  { 2461314.50, 73.170    }, //2026.75, Predicted value
  { 2461406.50, 73.320    }, //2027.00, Predicted value
  { 2461497.50, 73.460    }, //2027.25, Predicted value
  { 2461588.50, 73.580    }, //2027.50, Predicted value
  { 2461680.50, 73.660    }  //2027.75, Predicted value
} };
#endif //#ifndef AAPLUS_DELTAT_NO_HIGH_PRECISION

struct LeapSecondCoefficient
{
  double JD;
  double LeapSeconds;
  double BaseMJD;
  double Coefficient;
};

constexpr array<LeapSecondCoefficient, 41> g_LeapSecondCoefficients //Cumulative leap second values as taken from https://cddis.nasa.gov/archive/products/iers/tai-utc.dat
{ {
  { 2437300.5, 1.4228180, 37300, 0.001296  }, //1 Jan 1961
  { 2437512.5, 1.3728180, 37300, 0.001296  },
  { 2437665.5, 1.8458580, 37665, 0.0011232 },
  { 2438334.5, 1.9458580, 37665, 0.0011232 },
  { 2438395.5, 3.2401300, 38761, 0.001296  },
  { 2438486.5, 3.3401300, 38761, 0.001296  },
  { 2438639.5, 3.4401300, 38761, 0.001296  },
  { 2438761.5, 3.5401300, 38761, 0.001296  },
  { 2438820.5, 3.6401300, 38761, 0.001296  },
  { 2438942.5, 3.7401300, 38761, 0.001296  },
  { 2439004.5, 3.8401300, 38761, 0.001296  },
  { 2439126.5, 4.3131700, 39126, 0.002592  },
  { 2439887.5, 4.2131700, 39126, 0.002592  },
  { 2441317.5, 10.0,      41317, 0.0       },
  { 2441499.5, 11.0,      41317, 0.0       },
  { 2441683.5, 12.0,      41317, 0.0       },
  { 2442048.5, 13.0,      41317, 0.0       },
  { 2442413.5, 14.0,      41317, 0.0       },
  { 2442778.5, 15.0,      41317, 0.0       },
  { 2443144.5, 16.0,      41317, 0.0       },
  { 2443509.5, 17.0,      41317, 0.0       },
  { 2443874.5, 18.0,      41317, 0.0       },
  { 2444239.5, 19.0,      41317, 0.0       },
  { 2444786.5, 20.0,      41317, 0.0       },
  { 2445151.5, 21.0,      41317, 0.0       },
  { 2445516.5, 22.0,      41317, 0.0       },
  { 2446247.5, 23.0,      41317, 0.0       },
  { 2447161.5, 24.0,      41317, 0.0       },
  { 2447892.5, 25.0,      41317, 0.0       },
  { 2448257.5, 26.0,      41317, 0.0       },
  { 2448804.5, 27.0,      41317, 0.0       },
  { 2449169.5, 28.0,      41317, 0.0       },
  { 2449534.5, 29.0,      41317, 0.0       },
  { 2450083.5, 30.0,      41317, 0.0       },
  { 2450630.5, 31.0,      41317, 0.0       },
  { 2451179.5, 32.0,      41317, 0.0       },
  { 2453736.5, 33.0,      41317, 0.0       },
  { 2454832.5, 34.0,      41317, 0.0       },
  { 2456109.5, 35.0,      41317, 0.0       },
  { 2457204.5, 36.0,      41317, 0.0       },
  { 2457754.5, 37.0,      41317, 0.0       } //1 January 2017
} };


////////////////////////////////// Implementation /////////////////////////////

CAADynamicalTime::DELTAT_PROC CAADynamicalTime::sm_pDeltaTProc = nullptr;

CAADynamicalTime::DELTAT_PROC CAADynamicalTime::SetUserDefinedDeltaT(DELTAT_PROC pProc) noexcept
{
  DELTAT_PROC pOld = sm_pDeltaTProc;
  sm_pDeltaTProc = pProc;
  return pOld;
}

double CAADynamicalTime::DeltaT(double JD)
{
  //Allow end-user customization of the DeltaT function
  if (sm_pDeltaTProc != nullptr)
    return sm_pDeltaTProc(JD);

  //What will be the return value from the method
  double Delta = 0;

  //Determine if we can use the lookup table
#ifndef AAPLUS_DELTAT_NO_HIGH_PRECISION
  constexpr size_t nLookupElements = g_DeltaTValues.size();
  assert(nLookupElements >= 2);
  if ((JD >= g_DeltaTValues[0].JD) && (JD < g_DeltaTValues[nLookupElements - 1].JD))
  {
    //Find the index in the lookup table which contains the JD value closest to the JD input parameter
    DeltaTValue toFind{JD, 0};
    const auto iter = std::upper_bound(g_DeltaTValues.cbegin(), g_DeltaTValues.cend(), toFind, [](const DeltaTValue& element1, const DeltaTValue& element2) noexcept -> bool
      {
        return element1.JD < element2.JD;
      });
    assert(iter != g_DeltaTValues.cend());
    const size_t nFoundIndex = std::distance(g_DeltaTValues.cbegin(), iter);

    //Now do a simple linear interpolation of the DeltaT values from the lookup table
  #ifdef _MSC_VER
    #pragma warning(suppress : 28020)
  #endif //#ifdef _MSC_VER
    Delta = (JD - g_DeltaTValues[nFoundIndex - 1].JD) / (g_DeltaTValues[nFoundIndex].JD - g_DeltaTValues[nFoundIndex - 1].JD) * (g_DeltaTValues[nFoundIndex].DeltaT - g_DeltaTValues[nFoundIndex - 1].DeltaT) + g_DeltaTValues[nFoundIndex - 1].DeltaT;
  }
  else
#endif //#ifndef AAPLUS_DELTAT_NO_HIGH_PRECISION
  {
    const CAADate date(JD, CAADate::AfterPapalReform(JD));
    const double y = date.FractionalYear();

    //Use the polynomial expressions from Espenak & Meeus 2006. References: http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html and
    //http://www.staff.science.uu.nl/~gent0113/deltat/deltat_old.htm (Espenak & Meeus 2006 section)
    if (y < -500)
    {
      const double u = (y - 1820)/100.0;
      const double u2 = u*u;
      Delta = -20 + (32*u2);
    }
    else if (y < 500)
    {
      const double u = y/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      const double u5 = u4*u;
      const double u6 = u5*u;
      Delta = 10583.6 + (-1014.41*u) + (33.78311*u2) + (-5.952053*u3) + (-0.1798452*u4) + (0.022174192*u5) + (0.0090316521*u6);
    }
    else if (y < 1600)
    {
      const double u = (y - 1000)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      const double u5 = u4*u;
      const double u6 = u5*u;
      Delta = 1574.2 + (-556.01*u) + (71.23472*u2) + (0.319781*u3) + (-0.8503463*u4) + (-0.005050998*u5) + (0.0083572073*u6);
    }
    else if (y < 1700)
    {
      const double u = (y - 1600)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      Delta = 120 + (-98.08*u) + (-153.2*u2) + (u3/0.007129);
    }
    else if (y < 1800)
    {
      const double u = (y - 1700)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      Delta = 8.83 + (16.03*u) + (-59.285*u2) + (133.36*u3) + (-u4/0.01174);
    }
    else if (y < 1860)
    {
      const double u = (y - 1800)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      const double u5 = u4*u;
      const double u6 = u5*u;
      const double u7 = u6*u;
      Delta = 13.72 + (-33.2447*u) + (68.612*u2) + (4111.6*u3) + (-37436*u4) + (121272*u5) + (-169900*u6) + (87500*u7);
    }
    else if (y < 1900)
    {
      const double u = (y - 1860)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      const double u5 = u4*u;
      Delta = 7.62 + (57.37*u) + (-2517.54*u2) + (16806.68*u3) + (-44736.24*u4) + (u5/0.0000233174);
    }
    else if (y < 1920)
    {
      const double u = (y - 1900)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      Delta = -2.79 + (149.4119*u) + (-598.939*u2) + (6196.6*u3) + (-19700*u4);
    }
    else if (y < 1941)
    {
      const double u = (y - 1920)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      Delta = 21.20 + (84.493*u) + (-761.00*u2) + (2093.6*u3);
    }
    else if (y < 1961)
    {
      const double u = (y - 1950)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      Delta = 29.07 + (40.7*u) + (-u2/0.0233) + (u3/0.002547);
    }
    else if (y < 1986)
    {
      const double u = (y - 1975)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      Delta = 45.45 + 106.7*u - u2/0.026 - u3/0.000718;
    }
    else if (y < 2005)
    {
      const double u = (y - 2000)/100.0;
      const double u2 = u*u;
      const double u3 = u2*u;
      const double u4 = u3*u;
      const double u5 = u4*u;
      Delta = 63.86 + (33.45*u) + (-603.74*u2) + (1727.5*u3) + (65181.4*u4) + (237359.9*u5);
    }
    else if (y < 2050)
    {
      const double u = (y - 2000)/100.0;
      const double u2 = u*u;
      Delta = 62.92 + (32.217*u) + (55.89*u2);
    }
    else if (y < 2150)
    {
      const double u = (y - 1820)/100.0;
      const double u2 = u*u;
      Delta = -205.72 + (56.28*u) + (32*u2);
    }
    else
    {
      const double u = (y - 1820)/100.0;
      const double u2 = u*u;
      Delta = -20 + (32*u2);
    }
  }

  return Delta;
}

double CAADynamicalTime::CumulativeLeapSeconds(double JD)
{
  //What will be the return value from the method
  double LeapSeconds = 0;

  if (JD >= g_LeapSecondCoefficients[0].JD)
  {
    constexpr size_t nLookupElements = g_LeapSecondCoefficients.size();
    assert(nLookupElements >= 2);
    if (JD < g_LeapSecondCoefficients[nLookupElements - 1].JD)
    {
      //Find the index in the lookup table which contains the JD value closest to the JD input parameter
      LeapSecondCoefficient toFind{JD, 0, 0, 0};
      const auto iter = std::upper_bound(g_LeapSecondCoefficients.cbegin(), g_LeapSecondCoefficients.cend(), toFind, [](const LeapSecondCoefficient& element1, const LeapSecondCoefficient& element2) noexcept -> bool
        {
          return element1.JD < element2.JD;
        });

      assert(iter != g_LeapSecondCoefficients.cend());
      const size_t nFoundIndex = std::distance(g_LeapSecondCoefficients.cbegin(), iter);
      LeapSeconds = g_LeapSecondCoefficients[nFoundIndex - 1].LeapSeconds + (JD - 2400000.5 - g_LeapSecondCoefficients[nFoundIndex - 1].BaseMJD) * g_LeapSecondCoefficients[nFoundIndex - 1].Coefficient;
    }
    else
      LeapSeconds = g_LeapSecondCoefficients[nLookupElements - 1].LeapSeconds + (JD - 2400000.5 - g_LeapSecondCoefficients[nLookupElements - 1].BaseMJD) * g_LeapSecondCoefficients[nLookupElements - 1].Coefficient;
  }

  return LeapSeconds;
}

double CAADynamicalTime::TT2UTC(double JD)
{
  //Outside of the range 1 January 1961 (first entry in g_LeapSecondCoefficients) to 500 days after the last leap second (last entry in g_LeapSecondCoefficients),
  //we implement TT2UTC as TT2UT1
  constexpr size_t nLookupElements = g_LeapSecondCoefficients.size();
  assert(nLookupElements >= 2);
  if ((JD < g_LeapSecondCoefficients[0].JD) || (JD > (g_LeapSecondCoefficients[nLookupElements - 1].JD + 500)))
    return TT2UT1(JD);

  const double DT = DeltaT(JD);
  const double UT1 = JD - (DT / 86400.0);
  const double LeapSeconds = CumulativeLeapSeconds(JD);
  return ((DT - LeapSeconds - 32.184) / 86400.0) + UT1;
}

double CAADynamicalTime::UTC2TT(double JD)
{
  //Outside of the range 1 January 1961 (first entry in g_LeapSecondCoefficients) to 500 days after the last leap second (last entry in g_LeapSecondCoefficients),
  //we implement TT2UTC as TT2UT1
  constexpr size_t nLookupElements = g_LeapSecondCoefficients.size();
  assert(nLookupElements >= 2);
  if ((JD < g_LeapSecondCoefficients[0].JD) || (JD >(g_LeapSecondCoefficients[nLookupElements - 1].JD + 500)))
    return UT12TT(JD);

  const double DT = DeltaT(JD);
  const double LeapSeconds = CumulativeLeapSeconds(JD);
  const double UT1 = JD - ((DT - LeapSeconds - 32.184) / 86400.0);
  return UT1 + (DT / 86400.0);
}

double CAADynamicalTime::TT2UT1(double JD)
{
  return JD - (DeltaT(JD) / 86400.0);
}

double CAADynamicalTime::UT12TT(double JD)
{
  return JD + (DeltaT(JD) / 86400.0);
}

double CAADynamicalTime::UT1MinusUTC(double JD)
{
  const double JDUTC = JD + ((CAADynamicalTime::DeltaT(JD) - CumulativeLeapSeconds(JD) - 32.184) / 86400);
  return (JD - JDUTC) * 86400;
}
