/*
Module : AAELP2000.cpp
Purpose: Implementation for the algorithms for ELP2000-82B
Created: PJN / 31-12-2015
History: PJN / 31-12-2015 1. Initial public release.
         PJN / 04-01-2016 1. Major cleanup and refactoring of the code. The new reimplementation should prove
                          faster and easier to maintain going forward.
                          2. Updated copyright details
         PJN / 30-07-2017 1. Removed unnecessary SECOND_2_RAD define. 
                          2. Updated various CAAELP2000 methods to use "const" parameters.
         PJN / 18-08-2019 1. Fixed some further compiler warnings when using VC 2019 Preview v16.3.0 Preview 2.0
         PJN / 18-04-2020 1. Reworked C arrays to use std::array
         PJN / 29-04-2020 1. Fixed a compilation issue on GCC where size_t was undefined in various modules.
                          Thanks to Bert Devlieghe for reporting this bug.
         PJN / 01-06-2020 1. Fixed a bug in CAAELP2000::SunMeanAnomaly where some local variables were not 
                          initialized. Thanks to "Carlos" for reporting this bug.
                          2. Renamed CAAELP2000::MoonMeanLongitude to CAAELP2000::MoonMeanMeanLongitude.
                          3. Renamed CAAELP2000::MoonMeanLongitudeLunarPerigee to 
                          CAAELP2000::MeanLongitudeLunarPerigee.
                          4. Renamed CAAELP2000::MoonMeanLongitudeLunarAscendingNode to 
                          CAAELP2000::MeanLongitudeLunarAscendingNode.
                          5. Renamed CAAELP2000::EarthMoonBarycentreMeanLongitude to 
                          CAAELP2000::MeanHeliocentricMeanLongitudeEarthMoonBarycentre.
                          6. Renamed CAAELP2000::EarthMoonBarycentreMeanLongitudeOfPerihelion to 
                          CAAELP2000::MeanLongitudeOfPerilhelionOfEarthMoonBarycentre

Copyright (c) 2015 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "AAELP2000.h"
#include "AACoordinateTransformation.h"
#include "AA3DCoordinate.h"
#include "AADefines.h"
#include <cmath>
#include <cassert>
#include <array>
using namespace std;


////////////////////////////// Macros / Defines ///////////////////////////////

#ifdef _MSC_VER
#pragma warning(disable : 26446 26481 26485)
#endif //#ifdef _MSC_VER

//Lunar constants
constexpr double AM = .074801329518;
constexpr double DTASM = 0.022921886117733679; //ALFA * 2.0 / (AM * 3.0) where ALFA .002571881335

//Lunar arguments
constexpr array<double, 15> g_W
{ {
  3.8103444305883079,      //CAACoordinateTransformation::DegreesToRadians(218 + 18/60.0 + 59.95571/3600.0),
  1.4547885323225087,      //CAACoordinateTransformation::DegreesToRadians(83 + 21/60.0 + 11.67475/3600.0),
  2.1824391972168398,      //CAACoordinateTransformation::DegreesToRadians(125 + 2/60.0 + 40.39816/3600.0),
  8399.6847317739139,      //1732559343.73604"
  70.993304818359618,      //14643420.2632"
  -33.781426356625921,     //-6967919.3622"
  -2.8547283984772807e-5 , //-5.8883"
  -0.00018557504160038375, //-38.2776"
  3.0844816019550893e-5,   //6.3622"
  3.2017095500473753e-8,   //0.006604"
  -2.1839401892941265e-7,  //-0.045047"
  3.6967043184602116e-8,   //0.007625"
  -1.5363745554361197e-10, //-0.00003169"
  1.0327016221314225e-9,   //0.00021301"
  -1.7385418604587960e-10  //-0.00003586"
} };

constexpr array<double, 5> g_EARTH
{ {
  1.7534703431506580,     //CAACoordinateTransformation::DegreesToRadians(100 + 27/60.0 + 59.22059/3600.0)
  628.30758496215537,     //129597742.2758"
  -9.7932363584126268e-8, //-0.0202"
  4.3633231299858238e-11, //0.000009"
  7.2722052166430391e-13  //0.00000015"
} };

constexpr array<double, 5> g_PERI
{ {
  1.7965955233587829,      //CAACoordinateTransformation::DegreesToRadians(102 + 56/60.0 + 14.42753/3600.0),
  0.0056297936673156855,   //1161.2283"
  2.5826024792704977e-6,   //0.5327"
  -6.6904287993115966e-10, //-0.000138"
  0.0
} };

//Planetary arguments
constexpr array<array<double, 2>, 8> g_P
{ {
  { 4.4026088424029615,  2608.7903141574106 }, //CAACoordinateTransformation::DegreesToRadians(252 + 15/60.0 + 3.25986/3600.0),   538101628.68898"
  { 3.1761466969075944,  1021.3285546211089 }, //CAACoordinateTransformation::DegreesToRadians(181 + 58/60.0 + 47.28305/3600.0),  210664136.43355"
  { 1.7534703431506580,  628.30758496215537 }, //g_EARTH[0],                                                                      g_EARTH[1]
  { 6.2034809133999449,  334.06124314922965	}, //CAACoordinateTransformation::DegreesToRadians(355 + 25/60.0 + 59.78866/3600.0),  68905077.59284"
  { 0.59954649738867349, 52.969096509472053	}, //CAACoordinateTransformation::DegreesToRadians(34  + 21/60.0 + 5.34212/3600.0),   10925660.42861"
  { 0.87401675651848076, 21.329909543800007	}, //CAACoordinateTransformation::DegreesToRadians(50  + 4/60.0  + 38.89694/3600.0),  4399609.65932"
  { 5.4812938716049908,  7.4781598567143535	}, //CAACoordinateTransformation::DegreesToRadians(314 + 3/60.0  + 18.01841/3600.0),  1542481.19393"
  { 5.3118862867834666,  3.8133035637584558	}  //CAACoordinateTransformation::DegreesToRadians(304 + 20/60.0 + 55.19575/3600.0),  786550.32074"
} };

//Corrections of the constants (fit to DE200/LE200)
constexpr double g_DELNU = 3.2093561586235289e-10;  //0.55604" / g_W[3]
constexpr double g_DELE  = 8.6733167550495987e-8;   //0.01789"
constexpr double g_DELG  = -3.9105071518295170e-7;  //-0.08066"
constexpr double g_DELNP = -3.7078095034525493e-11; //-0.06424" / g_W[3]
constexpr double g_DELEP = -6.2439153990097128e-7;  //-0.12879"

constexpr array<double, 2> g_ZETA
{ {
  3.8103444305883079, //g_W[0]
  8399.7091135222672  //g_W[3] + 5029.0966"
} };

//The main ELP2000-82b tables
constexpr array<ELP2000MainProblemCoefficient, 1023> g_ELP1
{ {
 { {  0,  0,  0,  2 },   -411.60287, {    168.48, -18433.81,   -121.62,      0.40,     -0.18,      0.00 } },
 { {  0,  0,  0,  4 },      0.42034, {     -0.39,     37.65,      0.57,      0.00,      0.00,      0.00 } },
 { {  0,  0,  0,  6 },     -0.00059, {      0.00,     -0.08,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -6 },      0.00016, {      0.00,      0.02,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -4 },     -0.08019, {      0.77,     -7.22,     -1.46,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -2 },     39.53393, {   -395.24,   1788.33,    720.91,     -2.60,      0.90,      0.00 } },
 { {  0,  0,  1,  0 },  22639.55000, {      0.00,      0.00, 412529.61,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1,  2 },    -45.10032, {     17.41,  -2019.78,   -830.20,      0.09,     -0.02,      0.00 } },
 { {  0,  0,  1,  4 },      0.09092, {     -0.07,      8.14,      1.73,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1,  6 },     -0.00019, {      0.00,     -0.02,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -4 },      0.00068, {      0.07,      0.06,      0.03,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -2 },     -1.37259, {     -7.58,    -62.49,    -50.12,     -0.07,      0.00,      0.00 } },
 { {  0,  0,  2,  0 },    769.02326, {   -257.51,    -47.39,  28008.99,     -6.83,     -0.56,      0.00 } },
 { {  0,  0,  2,  2 },     -3.99767, {      2.40,   -178.94,   -146.18,      0.05,      0.00,      0.00 } },
 { {  0,  0,  2,  4 },      0.01297, {     -0.01,      1.16,      0.48,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2,  6 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -4 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -2 },     -0.05684, {      0.38,     -2.65,     -3.13,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3,  0 },     36.12364, {    -28.42,     -5.10,   1972.74,     -0.93,     -0.08,      0.00 } },
 { {  0,  0,  3,  2 },     -0.32866, {      0.32,    -14.70,    -17.99,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3,  4 },      0.00153, {      0.00,      0.14,      0.08,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4, -2 },     -0.00354, {      0.04,     -0.17,     -0.26,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4,  0 },      1.93367, {     -2.65,     -0.44,    140.75,     -0.10,      0.00,      0.00 } },
 { {  0,  0,  4,  2 },     -0.02602, {      0.04,     -1.16,     -1.90,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4,  4 },      0.00016, {      0.00,      0.01,      0.01,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5, -2 },     -0.00023, {      0.00,     -0.01,     -0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  0 },      0.11100, {     -0.23,     -0.04,     10.10,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  2 },     -0.00201, {      0.00,     -0.09,     -0.18,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6,  0 },      0.00665, {     -0.02,      0.00,      0.73,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6,  2 },     -0.00015, {      0.00,      0.00,     -0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  7,  0 },      0.00041, {      0.00,      0.00,      0.05,      0.00,      0.00,      0.00 } },
 { {  0,  0,  7,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  8,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -7,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -6,  0 },     -0.00024, {      0.00,      0.00,     -0.03,     -0.01,      0.00,      0.00 } },
 { {  0,  1, -5, -2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -5,  0 },     -0.00332, {     -0.07,      0.00,     -0.30,     -0.20,      0.00,      0.00 } },
 { {  0,  1, -4, -2 },      0.00064, {      0.01,      0.03,      0.05,      0.04,      0.00,      0.00 } },
 { {  0,  1, -4,  0 },     -0.04702, {     -1.04,      0.03,     -3.42,     -2.82,      0.00,      0.00 } },
 { {  0,  1, -4,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -3, -4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -3, -2 },      0.00623, {      0.14,      0.28,      0.34,      0.37,      0.00,      0.00 } },
 { {  0,  1, -3,  0 },     -0.66940, {    -15.55,      0.38,    -36.53,    -40.10,     -0.02,      0.00 } },
 { {  0,  1, -3,  2 },      0.00106, {      0.02,      0.05,      0.06,      0.07,      0.00,      0.00 } },
 { {  0,  1, -2, -4 },     -0.00017, {      0.00,     -0.02,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  1, -2, -2 },      0.05165, {      1.17,      2.29,      1.89,      3.09,      0.00,      0.00 } },
 { {  0,  1, -2,  0 },     -9.67938, {   -235.29,      4.90,   -352.52,   -579.94,     -0.15,      0.00 } },
 { {  0,  1, -2,  2 },      0.02429, {      0.74,      1.09,      0.89,      1.46,      0.00,      0.00 } },
 { {  0,  1, -2,  4 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -1, -4 },     -0.00063, {     -0.01,     -0.06,     -0.01,     -0.04,      0.00,      0.00 } },
 { {  0,  1, -1, -2 },      0.30157, {      6.60,     13.34,      5.61,     18.07,      0.00,      0.00 } },
 { {  0,  1, -1,  0 },   -147.32654, {  -3778.62,     68.68,  -2688.53,  -8829.17,     -0.76,      0.00 } },
 { {  0,  1, -1,  2 },     -0.08049, {      0.25,     -3.63,     -1.39,     -4.82,     -0.01,      0.00 } },
 { {  0,  1, -1,  4 },      0.00013, {      0.00,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  0, -4 },     -0.00023, {      0.01,     -0.02,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  1,  0, -2 },      0.07501, {     -6.72,      3.04,      0.85,      4.46,      0.00,      0.00 } },
 { {  0,  1,  0,  0 },   -666.44186, {  -5206.84,    258.79,   -555.98, -39887.79,      1.73,      0.00 } },
 { {  0,  1,  0,  2 },      0.41342, {      1.50,     18.15,      1.19,     24.72,      0.01,      0.00 } },
 { {  0,  1,  0,  4 },     -0.00083, {      0.00,     -0.07,      0.00,     -0.05,      0.00,      0.00 } },
 { {  0,  1,  1, -4 },      0.00011, {      0.00,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  1, -2 },     -0.08282, {      1.73,     -3.73,     -1.50,     -4.95,      0.00,      0.00 } },
 { {  0,  1,  1,  0 },   -109.38419, {  -2193.78,     51.64,  -2018.13,  -6556.10,      0.54,      0.00 } },
 { {  0,  1,  1,  2 },      0.26338, {      4.53,     11.65,      5.03,     15.78,      0.00,      0.00 } },
 { {  0,  1,  1,  4 },     -0.00062, {      0.00,     -0.06,     -0.01,     -0.04,      0.00,      0.00 } },
 { {  0,  1,  2, -4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  2, -2 },      0.02677, {      1.10,      1.20,      0.97,      1.60,      0.00,      0.00 } },
 { {  0,  1,  2,  0 },     -7.63041, {   -156.07,      4.09,   -279.89,   -457.41,      0.12,      0.00 } },
 { {  0,  1,  2,  2 },      0.04338, {      0.81,      1.92,      1.61,      2.60,      0.00,      0.00 } },
 { {  0,  1,  2,  4 },     -0.00015, {      0.00,     -0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  3, -2 },      0.00166, {      0.06,      0.08,      0.09,      0.10,      0.00,      0.00 } },
 { {  0,  1,  3,  0 },     -0.54594, {    -11.18,      0.34,    -29.97,    -32.73,      0.01,      0.00 } },
 { {  0,  1,  3,  2 },      0.00526, {      0.10,      0.23,      0.29,      0.32,      0.00,      0.00 } },
 { {  0,  1,  3,  4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  4, -2 },      0.00014, {      0.00,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { {  0,  1,  4,  0 },     -0.03947, {     -0.80,      0.03,     -2.89,     -2.37,      0.00,      0.00 } },
 { {  0,  1,  4,  2 },      0.00055, {      0.01,      0.02,      0.04,      0.03,      0.00,      0.00 } },
 { {  0,  1,  5, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  5,  0 },     -0.00287, {     -0.06,      0.00,     -0.26,     -0.17,      0.00,      0.00 } },
 { {  0,  1,  5,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  6,  0 },     -0.00021, {      0.00,      0.00,     -0.02,     -0.01,      0.00,      0.00 } },
 { {  0,  1,  7,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -5,  0 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  2, -4, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -4,  0 },     -0.00119, {     -0.04,      0.00,     -0.09,     -0.14,      0.00,      0.00 } },
 { {  0,  2, -3, -2 },      0.00014, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  0,  2, -3,  0 },     -0.01514, {     -0.50,      0.01,     -0.83,     -1.81,      0.00,      0.00 } },
 { {  0,  2, -3,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -2, -2 },      0.00100, {      0.03,      0.04,      0.04,      0.12,      0.00,      0.00 } },
 { {  0,  2, -2,  0 },     -0.19311, {     -6.23,      0.13,     -7.04,    -23.13,      0.00,      0.00 } },
 { {  0,  2, -2,  2 },      0.00060, {      0.02,      0.03,      0.02,      0.07,      0.00,      0.00 } },
 { {  0,  2, -1, -2 },      0.00473, {      0.12,      0.21,      0.09,      0.57,      0.00,      0.00 } },
 { {  0,  2, -1,  0 },     -2.56813, {    -81.06,      1.44,    -46.94,   -307.62,      0.00,      0.00 } },
 { {  0,  2, -1,  2 },      0.00013, {      0.05,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  0,  2,  0, -2 },     -0.00355, {     -0.26,     -0.16,      0.02,     -0.42,      0.00,      0.00 } },
 { {  0,  2,  0,  0 },     -7.44804, {    -12.64,      2.98,     -9.32,   -891.27,      0.03,      0.00 } },
 { {  0,  2,  0,  2 },      0.00519, {     -0.03,      0.23,      0.02,      0.62,      0.00,      0.00 } },
 { {  0,  2,  0,  4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  1, -2 },     -0.00040, {      0.08,     -0.02,      0.00,     -0.05,      0.00,      0.00 } },
 { {  0,  2,  1,  0 },     -1.16177, {    -19.75,      0.43,    -21.52,   -139.13,      0.00,      0.00 } },
 { {  0,  2,  1,  2 },      0.00270, {      0.03,      0.12,      0.05,      0.32,      0.00,      0.00 } },
 { {  0,  2,  2, -2 },      0.00040, {      0.02,      0.02,      0.02,      0.05,      0.00,      0.00 } },
 { {  0,  2,  2,  0 },     -0.06513, {     -0.79,      0.02,     -2.39,     -7.80,      0.00,      0.00 } },
 { {  0,  2,  2,  2 },      0.00035, {      0.00,      0.02,      0.01,      0.04,      0.00,      0.00 } },
 { {  0,  2,  3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  3,  0 },     -0.00341, {      0.00,      0.00,     -0.19,     -0.41,      0.00,      0.00 } },
 { {  0,  2,  3,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  4,  0 },     -0.00015, {      0.00,      0.00,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  0,  3, -4,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -3,  0 },     -0.00035, {     -0.01,      0.00,     -0.02,     -0.06,      0.00,      0.00 } },
 { {  0,  3, -2, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -2,  0 },     -0.00416, {     -0.15,      0.00,     -0.15,     -0.75,      0.00,      0.00 } },
 { {  0,  3, -2,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -1, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  0,  3, -1,  0 },     -0.05142, {     -1.82,      0.03,     -0.94,     -9.24,      0.00,      0.00 } },
 { {  0,  3, -1,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  0, -2 },     -0.00016, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  0,  3,  0,  0 },     -0.10136, {      1.16,      0.04,     -0.20,    -18.19,      0.00,      0.00 } },
 { {  0,  3,  0,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  0,  3,  1,  0 },     -0.01796, {     -0.30,      0.00,     -0.33,     -3.23,      0.00,      0.00 } },
 { {  0,  3,  1,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  2,  0 },     -0.00095, {     -0.01,      0.00,     -0.03,     -0.17,      0.00,      0.00 } },
 { {  0,  3,  3,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4, -2,  0 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  0,  4, -1,  0 },     -0.00108, {     -0.04,      0.00,     -0.02,     -0.26,      0.00,      0.00 } },
 { {  0,  4,  0,  0 },     -0.00128, {      0.06,      0.00,      0.00,     -0.30,      0.00,      0.00 } },
 { {  0,  4,  1,  0 },     -0.00032, {      0.00,      0.00,      0.00,     -0.08,      0.00,      0.00 } },
 { {  0,  4,  2,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  5, -1,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -4,  0,  0 },      0.00011, {      0.00,      0.00,      0.00,      0.03,      0.04,      0.00 } },
 { {  1, -3, -2,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -3, -1,  0 },      0.00020, {      0.00,      0.00,      0.00,      0.04,      0.08,      0.00 } },
 { {  1, -3,  0,  0 },      0.00254, {      0.03,      0.00,      0.00,      0.45,      0.99,      0.00 } },
 { {  1, -3,  1,  0 },      0.00007, {      0.00,      0.00,      0.00,      0.01,      0.03,      0.00 } },
 { {  1, -3,  2,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2, -3,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1, -2, -2,  0 },      0.00025, {      0.00,      0.00,      0.01,      0.03,      0.10,      0.00 } },
 { {  1, -2, -1, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1, -2, -1,  0 },      0.00229, {      0.06,      0.00,      0.05,      0.27,      0.89,      0.00 } },
 { {  1, -2,  0, -2 },     -0.00032, {      0.00,     -0.01,      0.00,     -0.04,     -0.12,      0.00 } },
 { {  1, -2,  0,  0 },      0.04304, {     -0.01,      0.00,      0.06,      5.14,     16.74,     -0.09 } },
 { {  1, -2,  0,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -2,  1, -2 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -2,  1,  0 },     -0.00033, {     -0.16,      0.00,     -0.01,     -0.04,     -0.13,      0.00 } },
 { {  1, -2,  2,  0 },     -0.00037, {     -0.03,      0.00,     -0.02,     -0.04,     -0.14,      0.00 } },
 { {  1, -2,  3,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1, -1, -5,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1, -4,  0 },      0.00014, {      0.00,      0.00,      0.01,      0.00,      0.05,      0.00 } },
 { {  1, -1, -3, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1, -3,  0 },      0.00069, {      0.02,      0.00,      0.04,      0.04,      0.27,      0.00 } },
 { {  1, -1, -3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1, -2, -2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  1, -1, -2,  0 },     -0.00694, {     -0.25,      0.00,     -0.23,     -0.42,     -2.70,      0.01 } },
 { {  1, -1, -2,  2 },      0.00013, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  1, -1, -1, -2 },     -0.00100, {      0.00,     -0.04,     -0.02,     -0.06,     -0.39,      0.00 } },
 { {  1, -1, -1,  0 },     -0.13636, {     -3.24,      0.19,     -2.41,     -8.17,    -53.02,      0.28 } },
 { {  1, -1, -1,  2 },      0.00013, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  1, -1,  0, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  0, -2 },     -0.01281, {     -0.13,     -0.57,      0.00,     -0.77,     -4.98,      0.03 } },
 { {  1, -1,  0,  0 },     -0.55694, {    -40.23,      1.70,     -0.01,    -33.62,   -216.55,      1.14 } },
 { {  1, -1,  0,  2 },      0.00155, {      0.09,      0.07,      0.00,      0.09,      0.60,      0.00 } },
 { {  1, -1,  1, -2 },     -0.00212, {     -0.02,     -0.09,     -0.04,     -0.13,     -0.83,      0.00 } },
 { {  1, -1,  1,  0 },     -0.12073, {     -6.32,      0.25,     -2.38,     -7.23,    -46.94,      0.25 } },
 { {  1, -1,  1,  2 },      0.00060, {      0.03,      0.03,      0.01,      0.04,      0.23,      0.00 } },
 { {  1, -1,  2, -2 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  1, -1,  2,  0 },     -0.01483, {     -0.72,      0.03,     -0.56,     -0.89,     -5.77,      0.03 } },
 { {  1, -1,  2,  2 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  1, -1,  3, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  3,  0 },     -0.00149, {     -0.07,      0.00,     -0.08,     -0.09,     -0.58,      0.00 } },
 { {  1, -1,  3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  4,  0 },     -0.00014, {      0.00,      0.00,     -0.01,      0.00,     -0.05,      0.00 } },
 { {  1, -1,  5,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -6,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1,  0, -5, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -5,  0 },     -0.00072, {     -0.02,      0.00,     -0.06,      0.00,     -0.28,      0.00 } },
 { {  1,  0, -4, -2 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  1,  0, -4,  0 },     -0.00955, {     -0.28,      0.01,     -0.70,      0.00,     -3.71,      0.02 } },
 { {  1,  0, -4,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -3, -2 },      0.00069, {      0.02,      0.03,      0.04,      0.00,      0.27,      0.00 } },
 { {  1,  0, -3,  0 },     -0.12812, {     -3.66,      0.13,     -7.03,     -0.05,    -49.82,      0.26 } },
 { {  1,  0, -3,  2 },      0.00038, {      0.06,      0.02,      0.03,      0.00,      0.14,      0.00 } },
 { {  1,  0, -2, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -2, -2 },      0.00289, {      0.14,      0.12,      0.11,      0.00,      1.12,      0.00 } },
 { {  1,  0, -2,  0 },     -1.75296, {    -48.87,      1.67,    -64.28,     -0.51,   -681.60,      3.59 } },
 { {  1,  0, -2,  2 },     -0.00142, {      0.05,     -0.07,     -0.05,      0.00,     -0.55,      0.00 } },
 { {  1,  0, -1, -4 },      0.00019, {      0.00,      0.02,      0.00,      0.00,      0.07,      0.00 } },
 { {  1,  0, -1, -2 },     -0.02379, {      0.12,     -1.09,     -0.42,      0.00,     -9.25,      0.05 } },
 { {  1,  0, -1,  0 },    -18.58467, {   -437.66,     18.91,   -346.51,     -1.00,  -7226.22,     38.10 } },
 { {  1,  0, -1,  2 },      0.01598, {      0.32,      0.69,      0.30,      0.00,      6.21,     -0.03 } },
 { {  1,  0, -1,  4 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  1,  0,  0, -4 },      0.00123, {      0.01,      0.11,      0.00,      0.00,      0.48,      0.00 } },
 { {  1,  0,  0, -2 },     -0.58332, {     -5.89,    -25.98,     -0.12,      0.00,   -226.81,      1.20 } },
 { {  1,  0,  0,  0 },   -124.98806, {  -2831.88,    136.18,    -86.60,     -2.68, -48598.15,    256.21 } },
 { {  1,  0,  0,  2 },      0.25429, {      5.73,     11.11,      0.26,      0.00,     98.87,     -0.52 } },
 { {  1,  0,  0,  4 },     -0.00054, {     -0.01,     -0.05,      0.00,      0.00,     -0.21,      0.00 } },
 { {  1,  0,  1, -4 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  1,  0,  1, -2 },     -0.04012, {     -0.22,     -1.80,     -0.72,      0.00,    -15.60,      0.08 } },
 { {  1,  0,  1,  0 },     -8.45308, {   -187.78,     10.10,   -158.62,      0.77,  -3286.75,     17.33 } },
 { {  1,  0,  1,  2 },      0.04450, {      0.98,      1.94,      0.84,      0.00,     17.30,     -0.09 } },
 { {  1,  0,  1,  4 },     -0.00015, {      0.00,     -0.01,      0.00,      0.00,     -0.06,      0.00 } },
 { {  1,  0,  2, -2 },     -0.00077, {      0.03,     -0.03,     -0.03,      0.00,     -0.30,      0.00 } },
 { {  1,  0,  2,  0 },     -0.58399, {    -12.36,      0.75,    -21.52,      0.13,   -227.07,      1.20 } },
 { {  1,  0,  2,  2 },      0.00540, {      0.12,      0.24,      0.20,      0.00,      2.10,     -0.01 } },
 { {  1,  0,  2,  4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  0,  3, -2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1,  0,  3,  0 },     -0.04074, {     -0.81,      0.06,     -2.24,      0.02,    -15.84,      0.08 } },
 { {  1,  0,  3,  2 },      0.00056, {      0.01,      0.02,      0.03,      0.00,      0.22,      0.00 } },
 { {  1,  0,  4,  0 },     -0.00285, {     -0.05,      0.00,     -0.21,      0.00,     -1.11,      0.00 } },
 { {  1,  0,  4,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1,  0,  5,  0 },     -0.00020, {      0.00,      0.00,     -0.02,      0.00,     -0.08,      0.00 } },
 { {  1,  0,  6,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1, -5,  0 },      0.00011, {      0.00,      0.00,      0.01,      0.00,      0.04,      0.00 } },
 { {  1,  1, -4, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1, -4,  0 },      0.00155, {      0.02,      0.00,      0.11,      0.09,      0.60,      0.00 } },
 { {  1,  1, -3, -2 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  1,  1, -3,  0 },      0.02296, {      0.30,      0.00,      1.26,      1.38,      8.93,     -0.05 } },
 { {  1,  1, -3,  2 },     -0.00006, {      0.02,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1,  1, -2, -2 },     -0.00071, {      0.00,     -0.03,     -0.03,     -0.04,     -0.27,      0.00 } },
 { {  1,  1, -2,  0 },      0.34966, {      5.36,     -0.11,     12.83,     20.95,    135.95,     -0.72 } },
 { {  1,  1, -2,  2 },      0.00063, {      0.00,      0.03,      0.02,      0.04,      0.25,      0.00 } },
 { {  1,  1, -1, -2 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1,  1, -1,  0 },      1.07773, {      9.38,     -1.15,     21.49,     64.24,    418.88,     -2.21 } },
 { {  1,  1, -1,  2 },      0.00000, {      0.03,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  0, -2 },      0.00017, {      0.03,      0.00,      0.07,      0.00,      0.07,      0.00 } },
 { {  1,  1,  0,  0 },     17.95512, {      6.59,    -11.64,     21.84,   1074.24,   6981.34,    -36.80 } },
 { {  1,  1,  0,  2 },     -0.03587, {      0.00,     -1.58,     -0.05,     -2.15,    -13.95,      0.07 } },
 { {  1,  1,  0,  4 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  1,  1, -2 },     -0.00157, {      0.36,     -0.04,     -0.04,     -0.07,     -0.56,      0.00 } },
 { {  1,  1,  1,  0 },      1.26186, {      1.83,     -0.90,     24.50,     75.50,    490.64,     -2.59 } },
 { {  1,  1,  1,  2 },     -0.00659, {      0.00,     -0.29,     -0.13,     -0.39,     -2.56,      0.01 } },
 { {  1,  1,  1,  4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  2, -2 },      0.00028, {      0.03,      0.01,      0.00,      0.02,      0.11,      0.00 } },
 { {  1,  1,  2,  0 },      0.09205, {      0.22,     -0.07,      3.46,      5.51,     35.79,     -0.19 } },
 { {  1,  1,  2,  2 },     -0.00084, {      0.00,     -0.04,     -0.03,     -0.05,     -0.33,      0.00 } },
 { {  1,  1,  3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  3,  0 },      0.00679, {      0.02,      0.00,      0.38,      0.41,      2.64,     -0.01 } },
 { {  1,  1,  3,  2 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  1,  1,  4,  0 },      0.00050, {      0.00,      0.00,      0.04,      0.03,      0.20,      0.00 } },
 { {  1,  1,  5,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1,  2, -4,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1,  2, -3,  0 },      0.00038, {      0.02,      0.00,      0.02,      0.05,      0.15,      0.00 } },
 { {  1,  2, -2,  0 },      0.00361, {      0.16,      0.00,      0.13,      0.43,      1.40,      0.00 } },
 { {  1,  2, -1,  0 },     -0.00300, {     -0.18,      0.00,     -0.05,     -0.36,     -1.16,      0.00 } },
 { {  1,  2, -1,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2,  0, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.01,      0.01,      0.00 } },
 { {  1,  2,  0,  0 },     -0.03968, {     -0.61,      0.04,     -0.07,     -4.75,    -15.43,      0.08 } },
 { {  1,  2,  0,  2 },      0.00011, {      0.00,      0.00,      0.00,      0.01,      0.04,      0.00 } },
 { {  1,  2,  1, -2 },      0.00002, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2,  1,  0 },     -0.00849, {     -0.16,      0.00,     -0.17,     -1.02,     -3.30,      0.02 } },
 { {  1,  2,  1,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1,  2,  2,  0 },     -0.00102, {     -0.02,      0.00,     -0.04,     -0.12,     -0.40,      0.00 } },
 { {  1,  2,  2,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2,  3,  0 },     -0.00011, {      0.00,      0.00,      0.00,     -0.01,     -0.04,      0.00 } },
 { {  1,  2,  4,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  3, -2,  0 },      0.00007, {      0.00,      0.00,      0.00,      0.01,      0.03,      0.00 } },
 { {  1,  3, -1,  0 },     -0.00011, {      0.00,      0.00,      0.00,     -0.02,     -0.04,      0.00 } },
 { {  1,  3,  0, -2 },      0.00000, {      0.02,     -0.01,      0.01,      0.02,      0.00,      0.00 } },
 { {  1,  3,  0,  0 },     -0.00056, {     -0.01,      0.00,      0.00,     -0.10,     -0.22,      0.00 } },
 { {  1,  3,  1,  0 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,     -0.03,      0.00 } },
 { {  1,  4,  0,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -6,  0,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -5, -2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -5, -1,  0 },      0.00027, {      0.00,      0.00,      0.00,      0.08,      0.00,      0.00 } },
 { {  2, -5,  0,  0 },      0.00050, {      0.02,      0.00,      0.00,      0.15,      0.00,      0.00 } },
 { {  2, -5,  1,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2, -4, -3,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4, -2,  0 },      0.00032, {      0.01,      0.00,      0.01,      0.08,      0.00,      0.00 } },
 { {  2, -4, -1,  0 },      0.00812, {      0.24,      0.00,      0.15,      1.94,      0.00,      0.00 } },
 { {  2, -4, -1,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4,  0, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4,  0,  0 },      0.01331, {      0.58,      0.00,      0.02,      3.18,      0.00,      0.00 } },
 { {  2, -4,  0,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4,  1, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4,  1,  0 },      0.00133, {      0.06,      0.00,      0.03,      0.32,      0.00,      0.00 } },
 { {  2, -4,  2,  0 },      0.00012, {      0.00,      0.00,      0.00,      0.03,      0.00,      0.00 } },
 { {  2, -4,  3,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -3,  0 },      0.00049, {      0.01,      0.00,      0.03,      0.09,      0.00,      0.00 } },
 { {  2, -3, -2, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -2,  0 },      0.00965, {      0.26,      0.00,      0.35,      1.73,      0.00,      0.00 } },
 { {  2, -3, -2,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -1, -2 },     -0.00021, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  2, -3, -1,  0 },      0.24697, {      6.33,     -0.03,      4.52,     44.32,      0.00,      0.00 } },
 { {  2, -3, -1,  2 },     -0.00052, {     -0.01,     -0.02,      0.00,     -0.09,      0.00,      0.00 } },
 { {  2, -3,  0, -2 },      0.00163, {      0.00,      0.07,      0.00,      0.29,      0.00,      0.00 } },
 { {  2, -3,  0,  0 },      0.33983, {     13.41,     -0.12,      0.74,     60.98,      0.00,      0.00 } },
 { {  2, -3,  0,  2 },     -0.00075, {     -0.03,     -0.03,      0.00,     -0.13,      0.00,      0.00 } },
 { {  2, -3,  1, -2 },     -0.00065, {     -0.02,     -0.03,     -0.01,     -0.12,      0.00,      0.00 } },
 { {  2, -3,  1,  0 },      0.03300, {      1.42,     -0.02,      0.67,      5.92,      0.00,      0.00 } },
 { {  2, -3,  1,  2 },     -0.00016, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  2, -3,  2, -2 },     -0.00005, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2, -3,  2,  0 },      0.00295, {      0.14,      0.00,      0.11,      0.53,      0.00,      0.00 } },
 { {  2, -3,  2,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  3,  0 },      0.00026, {      0.01,      0.00,      0.01,      0.05,      0.00,      0.00 } },
 { {  2, -3,  4,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -5,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -4, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -4,  0 },      0.00088, {      0.01,      0.00,      0.06,      0.10,      0.00,      0.00 } },
 { {  2, -2, -3, -2 },     -0.00013, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2, -2, -3,  0 },      0.01544, {      0.29,      0.00,      0.84,      1.85,      0.00,      0.00 } },
 { {  2, -2, -2, -2 },     -0.00111, {     -0.03,     -0.05,     -0.04,     -0.13,      0.00,      0.00 } },
 { {  2, -2, -2,  0 },      0.29422, {      6.88,     -0.03,     10.73,     35.21,      0.00,      0.00 } },
 { {  2, -2, -2,  2 },     -0.00072, {      0.00,     -0.03,     -0.03,     -0.09,      0.00,      0.00 } },
 { {  2, -2, -1, -2 },     -0.00434, {     -0.20,     -0.19,     -0.08,     -0.52,      0.00,      0.00 } },
 { {  2, -2, -1,  0 },      7.37173, {    165.38,     -1.06,    134.79,    881.89,     -0.04,      0.00 } },
 { {  2, -2, -1,  2 },     -0.01541, {     -0.33,     -0.69,     -0.29,     -1.84,      0.00,      0.00 } },
 { {  2, -2, -1,  4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  0, -4 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  0, -2 },      0.06314, {      0.43,      2.82,     -0.04,      7.55,      0.00,      0.00 } },
 { {  2, -2,  0,  0 },      8.05076, {    289.14,     -3.14,     21.09,    962.91,     -0.06,      0.00 } },
 { {  2, -2,  0,  2 },     -0.01816, {     -0.62,     -0.81,     -0.12,     -2.17,      0.00,      0.00 } },
 { {  2, -2,  0,  4 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  1, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  1, -2 },     -0.01644, {     -0.38,     -0.74,     -0.30,     -1.97,      0.00,      0.00 } },
 { {  2, -2,  1,  0 },      0.75173, {     29.04,     -0.43,     15.47,     89.90,      0.00,      0.00 } },
 { {  2, -2,  1,  2 },     -0.00378, {     -0.14,     -0.17,     -0.09,     -0.45,      0.00,      0.00 } },
 { {  2, -2,  1,  4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  2, -2 },     -0.00133, {     -0.03,     -0.06,     -0.05,     -0.16,      0.00,      0.00 } },
 { {  2, -2,  2,  0 },      0.06439, {      2.63,     -0.05,      2.50,      7.70,      0.00,      0.00 } },
 { {  2, -2,  2,  2 },     -0.00055, {     -0.02,     -0.02,     -0.02,     -0.07,      0.00,      0.00 } },
 { {  2, -2,  3, -2 },     -0.00011, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2, -2,  3,  0 },      0.00538, {      0.23,      0.00,      0.31,      0.64,      0.00,      0.00 } },
 { {  2, -2,  3,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  4,  0 },      0.00044, {      0.02,      0.00,      0.03,      0.05,      0.00,      0.00 } },
 { {  2, -2,  5,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -6,  0 },      0.00010, {      0.00,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { {  2, -1, -5, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -5,  0 },      0.00178, {      0.01,      0.00,      0.16,      0.11,      0.00,      0.00 } },
 { {  2, -1, -4, -2 },     -0.00036, {      0.00,     -0.02,     -0.03,     -0.02,      0.00,      0.00 } },
 { {  2, -1, -4,  0 },      0.02949, {      0.41,      0.00,      2.15,      1.77,      0.00,      0.00 } },
 { {  2, -1, -4,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -3, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -3, -2 },     -0.00365, {     -0.08,     -0.16,     -0.20,     -0.22,      0.00,      0.00 } },
 { {  2, -1, -3,  0 },      0.47842, {      8.72,     -0.05,     26.15,     28.63,      0.00,      0.00 } },
 { {  2, -1, -3,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -2, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -2, -2 },     -0.02804, {     -0.74,     -1.25,     -1.03,     -1.68,      0.00,      0.00 } },
 { {  2, -1, -2,  0 },      8.60582, {    187.04,     -1.14,    313.85,    515.03,      0.02,      0.00 } },
 { {  2, -1, -2,  2 },     -0.02157, {     -0.21,     -0.97,     -0.79,     -1.29,      0.00,      0.00 } },
 { {  2, -1, -2,  4 },      0.00012, {      0.00,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -1, -4 },     -0.00015, {      0.00,     -0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -1, -2 },     -0.06103, {     -3.57,     -2.72,     -1.22,     -3.65,      0.00,      0.00 } },
 { {  2, -1, -1,  0 },    205.44315, {   4157.78,    -32.99,   3751.43,  12284.72,     -0.13,      0.00 } },
 { {  2, -1, -1,  2 },     -0.42624, {     -8.30,    -19.01,     -7.93,    -25.48,      0.00,      0.00 } },
 { {  2, -1, -1,  4 },      0.00093, {      0.02,      0.08,      0.02,      0.06,      0.00,      0.00 } },
 { {  2, -1,  0, -4 },     -0.00286, {     -0.02,     -0.26,      0.00,     -0.17,      0.00,      0.00 } },
 { {  2, -1,  0, -2 },      2.14619, {     18.77,     95.81,     -0.84,    128.22,      0.05,      0.00 } },
 { {  2, -1,  0,  0 },    164.73458, {   5378.28,    -77.43,    545.71,   9844.15,     -0.18,      0.00 } },
 { {  2, -1,  0,  2 },     -0.38215, {    -11.75,    -16.94,     -3.10,    -22.84,      0.00,      0.00 } },
 { {  2, -1,  0,  4 },      0.00089, {      0.03,      0.08,      0.01,      0.05,      0.00,      0.00 } },
 { {  2, -1,  1, -4 },      0.00018, {      0.00,      0.02,      0.00,      0.01,      0.00,      0.00 } },
 { {  2, -1,  1, -2 },     -0.37453, {     -6.50,    -16.89,     -6.91,    -22.36,      0.00,      0.00 } },
 { {  2, -1,  1,  0 },     14.53078, {    497.46,     -9.38,    307.25,    867.99,      0.00,      0.00 } },
 { {  2, -1,  1,  2 },     -0.07373, {     -2.45,     -3.26,     -1.77,     -4.40,      0.00,      0.00 } },
 { {  2, -1,  1,  4 },      0.00025, {      0.00,      0.02,      0.00,      0.02,      0.00,      0.00 } },
 { {  2, -1,  2, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  2, -2 },     -0.02887, {     -0.54,     -1.30,     -1.06,     -1.72,      0.00,      0.00 } },
 { {  2, -1,  2,  0 },      1.17704, {     41.72,     -0.96,     46.23,     70.28,      0.00,      0.00 } },
 { {  2, -1,  2,  2 },     -0.01007, {     -0.35,     -0.44,     -0.42,     -0.60,      0.00,      0.00 } },
 { {  2, -1,  2,  4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  3, -2 },     -0.00232, {     -0.05,     -0.10,     -0.13,     -0.14,      0.00,      0.00 } },
 { {  2, -1,  3,  0 },      0.09320, {      3.39,     -0.09,      5.36,      5.56,      0.00,      0.00 } },
 { {  2, -1,  3,  2 },     -0.00117, {     -0.04,     -0.05,     -0.07,     -0.07,      0.00,      0.00 } },
 { {  2, -1,  4, -2 },     -0.00019, {      0.00,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  2, -1,  4,  0 },      0.00730, {      0.27,      0.00,      0.55,      0.44,      0.00,      0.00 } },
 { {  2, -1,  4,  2 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  5, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  5,  0 },      0.00057, {      0.02,      0.00,      0.05,      0.03,      0.00,      0.00 } },
 { {  2, -1,  5,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  6,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -8,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -7,  0 },      0.00036, {      0.00,      0.00,      0.05,      0.00,      0.00,      0.00 } },
 { {  2,  0, -6, -2 },     -0.00011, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0, -6,  0 },      0.00497, {      0.10,      0.00,      0.54,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5, -2 },     -0.00111, {     -0.02,     -0.05,     -0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5,  0 },      0.06850, {      1.40,     -0.03,      6.24,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4, -2 },     -0.01077, {     -0.24,     -0.48,     -0.79,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4,  0 },      0.94827, {     19.71,     -0.30,     69.11,      0.06,      0.00,      0.00 } },
 { {  2,  0, -4,  2 },     -0.00054, {      0.00,     -0.03,     -0.04,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3, -4 },      0.00024, {      0.00,      0.02,      0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3, -2 },     -0.09154, {     -2.11,     -4.08,     -5.03,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3,  0 },     13.19400, {    279.87,     -3.19,    721.29,      0.65,      0.14,      0.00 } },
 { {  2,  0, -3,  2 },      0.00467, {      0.38,      0.20,      0.24,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3,  4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -4 },      0.00037, {      0.03,      0.03,      0.02,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -2 },     -0.56065, {    -14.54,    -24.99,    -20.69,     -0.01,      0.00,      0.00 } },
 { {  2,  0, -2,  0 },    211.65487, {   4685.54,    -42.06,   7715.64,      7.86,      2.12,     -0.01 } },
 { {  2,  0, -2,  2 },     -0.53572, {     -4.50,    -24.12,    -19.50,      0.04,     -0.01,      0.00 } },
 { {  2,  0, -2,  4 },      0.00282, {      0.02,      0.25,      0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -6 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -4 },     -0.00850, {      0.00,     -0.76,     -0.14,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -2 },      0.17903, {    -34.06,      7.98,      1.01,      0.03,      0.18,      0.00 } },
 { {  2,  0, -1,  0 },   4586.43061, {  87132.46,   -842.12,  83586.18,   -191.17,     20.31,     -0.17 } },
 { {  2,  0, -1,  2 },     -9.36601, {   -173.58,   -417.66,   -173.70,      0.48,     -0.06,      0.00 } },
 { {  2,  0, -1,  4 },      0.01993, {      0.36,      1.78,      0.39,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1,  6 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -6 },      0.00015, {      0.00,      0.02,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -4 },     -0.07518, {     -0.70,     -6.72,      0.06,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -2 },     55.17801, {    530.97,   2463.55,    -15.35,     -6.96,      1.20,      0.00 } },
 { {  2,  0,  0,  0 },   2369.91227, {  69551.14,  -1472.50,  10817.07,   -255.36,     22.07,     -0.15 } },
 { {  2,  0,  0,  2 },     -5.74170, {   -155.85,   -254.00,    -59.77,      0.64,     -0.10,      0.00 } },
 { {  2,  0,  0,  4 },      0.01372, {      0.35,      1.22,      0.21,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0,  6 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -6 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -4 },      0.00056, {      0.10,      0.05,      0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -2 },     -6.38325, {    -72.70,   -288.51,   -117.47,      1.16,      0.00,      0.00 } },
 { {  2,  0,  1,  0 },    191.95575, {   5619.27,   -153.09,   4250.55,    -30.10,      2.78,     -0.02 } },
 { {  2,  0,  1,  2 },     -0.99023, {    -27.73,    -43.70,    -25.33,      0.16,     -0.02,      0.00 } },
 { {  2,  0,  1,  4 },      0.00344, {      0.09,      0.31,      0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1,  6 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -4 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -2 },     -0.45380, {     -4.94,    -20.52,    -16.63,      0.10,      0.00,      0.00 } },
 { {  2,  0,  2,  0 },     14.37964, {    417.80,    -13.71,    577.47,     -3.08,      0.28,      0.00 } },
 { {  2,  0,  2,  2 },     -0.12386, {     -3.49,     -5.45,     -5.28,      0.03,      0.00,      0.00 } },
 { {  2,  0,  2,  4 },      0.00059, {      0.02,      0.05,      0.03,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -2 },     -0.03336, {     -0.35,     -1.51,     -1.83,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3,  0 },      1.05949, {     30.46,     -1.16,     61.74,     -0.30,      0.03,      0.00 } },
 { {  2,  0,  3,  2 },     -0.01332, {     -0.37,     -0.59,     -0.80,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3,  4 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4, -2 },     -0.00247, {     -0.03,     -0.11,     -0.18,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4,  0 },      0.07765, {      2.20,     -0.10,      5.93,     -0.03,      0.00,      0.00 } },
 { {  2,  0,  4,  2 },     -0.00131, {     -0.04,     -0.06,     -0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4,  4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5, -2 },     -0.00018, {      0.00,      0.00,     -0.02,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5,  0 },      0.00568, {      0.16,      0.00,      0.54,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5,  2 },     -0.00012, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6,  0 },      0.00041, {      0.01,      0.00,      0.05,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  7,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -7,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -6,  0 },      0.00015, {      0.01,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  2,  1, -5, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -5,  0 },      0.00173, {      0.12,      0.00,      0.16,      0.10,      0.00,      0.00 } },
 { {  2,  1, -4, -2 },     -0.00020, {     -0.02,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  2,  1, -4,  0 },      0.01819, {      1.37,     -0.02,      1.32,      1.10,      0.00,      0.00 } },
 { {  2,  1, -4,  2 },     -0.00003, {      0.04,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2,  1, -3, -2 },     -0.00116, {     -0.10,     -0.05,     -0.06,     -0.07,      0.00,      0.00 } },
 { {  2,  1, -3,  0 },      0.17624, {     15.80,     -0.23,      9.59,     10.63,      0.05,      0.00 } },
 { {  2,  1, -3,  2 },     -0.00024, {      0.00,     -0.01,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  2,  1, -2, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -2, -2 },     -0.00346, {     -0.43,     -0.15,     -0.13,     -0.21,      0.00,      0.00 } },
 { {  2,  1, -2,  0 },      2.48897, {    257.30,     -3.72,     89.71,    150.19,      0.26,      0.00 } },
 { {  2,  1, -2,  2 },     -0.00001, {     -0.34,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -1, -4 },      0.00017, {      0.00,      0.02,      0.00,      0.01,      0.00,      0.00 } },
 { {  2,  1, -1, -2 },     -0.00228, {      0.76,     -0.11,     -0.05,     -0.13,      0.00,      0.00 } },
 { {  2,  1, -1,  0 },    -28.39810, {    915.46,    -16.70,   -528.73,  -1693.30,     -0.38,      0.00 } },
 { {  2,  1, -1,  2 },      0.06742, {     -1.63,      3.04,      1.29,      4.02,      0.00,      0.00 } },
 { {  2,  1, -1,  4 },     -0.00016, {      0.00,     -0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  0, -4 },      0.00198, {      0.03,      0.18,      0.00,      0.12,      0.00,      0.00 } },
 { {  2,  1,  0, -2 },     -1.43724, {    -23.32,    -64.11,     -0.16,    -86.03,     -0.04,      0.00 } },
 { {  2,  1,  0,  0 },    -24.35910, {   -661.00,     29.13,   -128.06,  -1457.61,     -5.21,      0.03 } },
 { {  2,  1,  0,  2 },      0.06541, {      1.46,      2.86,      0.75,      3.91,      0.03,      0.00 } },
 { {  2,  1,  0,  4 },     -0.00017, {      0.00,     -0.02,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2,  1,  1, -4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  1, -2 },      0.08472, {     -0.38,      3.85,      1.59,      5.05,      0.00,      0.00 } },
 { {  2,  1,  1,  0 },     -2.91464, {   -100.96,      4.26,    -66.13,   -174.49,     -0.75,      0.00 } },
 { {  2,  1,  1,  2 },      0.01605, {      0.51,      0.70,      0.42,      0.96,      0.00,      0.00 } },
 { {  2,  1,  1,  4 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  2, -2 },      0.00889, {      0.07,      0.40,      0.33,      0.53,      0.00,      0.00 } },
 { {  2,  1,  2,  0 },     -0.29023, {    -11.13,      0.47,    -11.80,    -17.38,     -0.08,      0.00 } },
 { {  2,  1,  2,  2 },      0.00262, {      0.10,      0.11,      0.11,      0.16,      0.00,      0.00 } },
 { {  2,  1,  2,  4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  3, -2 },      0.00085, {      0.01,      0.04,      0.05,      0.05,      0.00,      0.00 } },
 { {  2,  1,  3,  0 },     -0.02676, {     -1.08,      0.05,     -1.57,     -1.60,      0.00,      0.00 } },
 { {  2,  1,  3,  2 },      0.00035, {      0.01,      0.02,      0.02,      0.02,      0.00,      0.00 } },
 { {  2,  1,  4, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  4,  0 },     -0.00236, {     -0.10,      0.00,     -0.18,     -0.14,      0.00,      0.00 } },
 { {  2,  1,  4,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  5,  0 },     -0.00020, {      0.00,      0.00,     -0.02,     -0.01,      0.00,      0.00 } },
 { {  2,  1,  6,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -5,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -4,  0 },     -0.00049, {      0.00,      0.00,     -0.04,     -0.06,      0.00,      0.00 } },
 { {  2,  2, -3, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -3,  0 },     -0.00944, {     -0.26,      0.00,     -0.53,     -1.14,      0.00,      0.00 } },
 { {  2,  2, -3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -2, -2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -2,  0 },     -0.24990, {    -10.76,      0.11,     -9.20,    -29.88,     -0.02,      0.00 } },
 { {  2,  2, -2,  2 },      0.00060, {      0.02,      0.03,      0.02,      0.07,      0.00,      0.00 } },
 { {  2,  2, -1, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -1, -2 },     -0.00654, {     -0.24,     -0.29,     -0.09,     -0.78,      0.00,      0.00 } },
 { {  2,  2, -1,  0 },     -2.52138, {    -84.66,      0.74,    -46.07,   -302.13,      0.07,      0.00 } },
 { {  2,  2, -1,  2 },      0.00515, {      0.17,      0.23,      0.10,      0.62,      0.00,      0.00 } },
 { {  2,  2, -1,  4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  0, -4 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  0, -2 },     -0.02349, {     -0.29,     -0.90,      0.49,     -2.75,      0.04,      0.00 } },
 { {  2,  2,  0,  0 },     -0.18576, {     -6.48,      0.33,     -6.56,    -22.26,      0.39,      0.00 } },
 { {  2,  2,  0,  2 },      0.00096, {      0.03,      0.04,      0.03,      0.11,      0.00,      0.00 } },
 { {  2,  2,  1, -2 },     -0.00200, {     -0.12,     -0.07,      0.00,     -0.23,      0.00,      0.00 } },
 { {  2,  2,  1,  0 },     -0.01343, {     -0.46,      0.03,     -0.71,     -1.61,      0.06,      0.00 } },
 { {  2,  2,  1,  2 },      0.00012, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2,  2,  2, -2 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2,  2,  2,  0 },     -0.00066, {     -0.01,      0.00,     -0.06,     -0.08,      0.00,      0.00 } },
 { {  2,  2,  3, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3, -3,  0 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2,  3, -2,  0 },     -0.00093, {     -0.06,      0.00,     -0.03,     -0.17,      0.00,      0.00 } },
 { {  2,  3, -1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3, -1,  0 },     -0.00188, {     -0.12,      0.00,     -0.03,     -0.34,      0.00,      0.00 } },
 { {  2,  3,  0, -2 },      0.00007, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2,  3,  0,  0 },      0.00038, {      0.03,      0.00,      0.02,      0.07,      0.00,      0.00 } },
 { {  2,  3,  1, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3,  1,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  4, -2,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  4, -1,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -4, -1,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -4,  0,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -3, -2,  0 },     -0.00017, {      0.00,      0.00,      0.00,     -0.03,     -0.07,      0.00 } },
 { {  3, -3, -1,  0 },     -0.00053, {     -0.02,      0.00,      0.00,     -0.09,     -0.21,      0.00 } },
 { {  3, -3,  0, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -3,  0,  0 },      0.00030, {      0.00,      0.00,      0.00,      0.05,      0.12,      0.00 } },
 { {  3, -3,  1,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  3, -2, -3,  0 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3, -2, -2,  0 },     -0.00399, {     -0.10,      0.00,     -0.15,     -0.48,     -1.55,      0.00 } },
 { {  3, -2, -1, -2 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3, -2, -1,  0 },     -0.01182, {     -0.50,      0.00,     -0.23,     -1.41,     -4.60,      0.02 } },
 { {  3, -2, -1,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  3, -2,  0, -2 },     -0.00102, {     -0.02,     -0.05,      0.00,     -0.12,     -0.40,      0.00 } },
 { {  3, -2,  0,  0 },      0.00518, {      0.05,      0.00,     -0.05,      0.62,      2.01,     -0.01 } },
 { {  3, -2,  1, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -2,  1,  0 },      0.00037, {      0.00,      0.00,      0.00,      0.04,      0.15,      0.00 } },
 { {  3, -1, -4,  0 },     -0.00013, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  3, -1, -3,  0 },     -0.00297, {     -0.06,      0.00,     -0.16,     -0.18,     -1.16,      0.00 } },
 { {  3, -1, -2,  0 },     -0.08311, {     -1.91,      0.02,     -3.04,     -4.98,    -32.32,      0.17 } },
 { {  3, -1, -2,  2 },      0.00020, {      0.00,      0.00,      0.00,      0.01,      0.08,      0.00 } },
 { {  3, -1, -1, -2 },     -0.00189, {     -0.02,     -0.08,     -0.03,     -0.11,     -0.73,      0.00 } },
 { {  3, -1, -1,  0 },     -0.23141, {     -9.24,      0.19,     -4.48,    -13.78,    -89.98,      0.47 } },
 { {  3, -1, -1,  2 },      0.00075, {      0.03,      0.03,      0.02,      0.05,      0.29,      0.00 } },
 { {  3, -1,  0, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  0, -2 },     -0.01948, {     -0.34,     -0.87,      0.00,     -1.16,     -7.57,      0.04 } },
 { {  3, -1,  0,  0 },      0.06507, {     -0.18,      0.16,     -1.04,      3.93,     25.30,     -0.13 } },
 { {  3, -1,  0,  2 },      0.00020, {      0.02,      0.00,      0.00,      0.01,      0.08,      0.00 } },
 { {  3, -1,  1, -2 },     -0.00084, {      0.00,     -0.04,     -0.01,     -0.05,     -0.33,      0.00 } },
 { {  3, -1,  1,  0 },      0.00330, {     -0.25,      0.03,     -0.05,      0.21,      1.28,      0.00 } },
 { {  3, -1,  1,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3, -1,  2, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -1,  2,  0 },     -0.00011, {     -0.05,      0.00,     -0.01,      0.00,     -0.04,      0.00 } },
 { {  3, -1,  2,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  3,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  0, -5,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  0, -4, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -4,  0 },     -0.00174, {      0.00,      0.00,     -0.13,      0.00,     -0.68,      0.00 } },
 { {  3,  0, -3, -2 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  3,  0, -3,  0 },     -0.04442, {     -0.91,      0.00,     -2.43,     -0.03,    -17.27,      0.09 } },
 { {  3,  0, -3,  2 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  3,  0, -2, -2 },     -0.00028, {      0.02,     -0.01,      0.00,      0.00,     -0.11,      0.00 } },
 { {  3,  0, -2,  0 },     -1.22412, {    -28.10,      0.42,    -44.73,     -0.15,   -475.97,      2.51 } },
 { {  3,  0, -2,  2 },      0.00298, {      0.07,      0.13,      0.11,      0.00,      1.16,      0.00 } },
 { {  3,  0, -2,  4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -1, -4 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  0, -1, -2 },     -0.02979, {     -0.39,     -1.33,     -0.54,      0.00,    -11.59,      0.06 } },
 { {  3,  0, -1,  0 },     -3.20968, {   -121.62,      3.09,    -62.39,      1.61,  -1248.01,      6.58 } },
 { {  3,  0, -1,  2 },      0.01215, {      0.48,      0.53,      0.25,      0.00,      4.72,     -0.02 } },
 { {  3,  0, -1,  4 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  3,  0,  0, -4 },      0.00025, {      0.00,      0.02,      0.00,      0.00,      0.10,      0.00 } },
 { {  3,  0,  0, -2 },     -0.25304, {     -3.60,    -11.27,      0.10,      0.08,    -98.39,      0.52 } },
 { {  3,  0,  0,  0 },      0.40423, {    -12.92,      2.28,    -14.17,      1.05,    157.17,     -0.83 } },
 { {  3,  0,  0,  2 },      0.00375, {      0.24,      0.16,      0.11,      0.00,      1.46,      0.00 } },
 { {  3,  0,  0,  4 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  1, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  0,  1, -2 },     -0.01062, {     -0.05,     -0.47,     -0.18,      0.00,     -4.13,      0.02 } },
 { {  3,  0,  1,  0 },      0.00037, {     -3.59,      0.36,     -1.36,      0.18,      0.14,      0.00 } },
 { {  3,  0,  1,  2 },      0.00083, {      0.06,      0.03,      0.03,      0.00,      0.32,      0.00 } },
 { {  3,  0,  2, -2 },     -0.00051, {      0.00,     -0.02,     -0.02,      0.00,     -0.20,      0.00 } },
 { {  3,  0,  2,  0 },     -0.00389, {     -0.50,      0.04,     -0.26,      0.02,     -1.51,      0.00 } },
 { {  3,  0,  2,  2 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  3,  0,  3, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  3,  0 },     -0.00062, {     -0.05,      0.00,     -0.04,      0.00,     -0.24,      0.00 } },
 { {  3,  0,  3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  4,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3,  1, -5,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  1, -4,  0 },     -0.00059, {     -0.02,      0.00,     -0.04,     -0.04,     -0.23,      0.00 } },
 { {  3,  1, -3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -3,  0 },     -0.00623, {     -0.30,      0.00,     -0.34,     -0.37,     -2.42,      0.01 } },
 { {  3,  1, -3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -2, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -2,  0 },     -0.04189, {     -2.51,      0.02,     -1.50,     -2.52,    -16.29,      0.09 } },
 { {  3,  1, -2,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  3,  1, -1, -2 },      0.00116, {      0.02,      0.05,      0.02,      0.07,      0.45,      0.00 } },
 { {  3,  1, -1,  0 },      0.27377, {      5.06,     -0.24,      5.22,     16.36,    106.45,     -0.56 } },
 { {  3,  1, -1,  2 },     -0.00138, {     -0.02,     -0.06,     -0.03,     -0.08,     -0.53,      0.00 } },
 { {  3,  1,  0, -2 },      0.00891, {      0.14,      0.40,     -0.01,      0.53,      3.46,     -0.02 } },
 { {  3,  1,  0,  0 },      0.14990, {      4.45,     -0.27,      1.61,      8.95,     58.29,     -0.31 } },
 { {  3,  1,  0,  2 },     -0.00096, {     -0.03,     -0.04,     -0.01,     -0.06,     -0.37,      0.00 } },
 { {  3,  1,  1, -2 },      0.00020, {      0.00,      0.00,      0.00,      0.01,      0.08,      0.00 } },
 { {  3,  1,  1,  0 },      0.02244, {      0.70,     -0.04,      0.59,      1.34,      8.73,     -0.05 } },
 { {  3,  1,  1,  2 },     -0.00022, {      0.00,      0.00,      0.00,     -0.01,     -0.09,      0.00 } },
 { {  3,  1,  2, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1,  2,  0 },      0.00256, {      0.08,      0.00,      0.11,      0.15,      1.00,      0.00 } },
 { {  3,  1,  2,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  3,  1,  3,  0 },      0.00026, {      0.00,      0.00,      0.02,      0.02,      0.10,      0.00 } },
 { {  3,  1,  4,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2, -4,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2, -3,  0 },     -0.00020, {      0.00,      0.00,     -0.01,     -0.02,     -0.08,      0.00 } },
 { {  3,  2, -2,  0 },     -0.00036, {     -0.06,      0.00,     -0.01,     -0.04,     -0.14,      0.00 } },
 { {  3,  2, -1, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2, -1,  0 },     -0.00127, {      0.07,      0.00,     -0.03,     -0.15,     -0.49,      0.00 } },
 { {  3,  2,  0, -2 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,     -0.04,      0.00 } },
 { {  3,  2,  0,  0 },     -0.00191, {     -0.05,      0.00,     -0.02,     -0.23,     -0.74,      0.00 } },
 { {  3,  2,  0,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2,  1,  0 },     -0.00039, {     -0.01,      0.00,      0.00,     -0.05,     -0.15,      0.00 } },
 { {  3,  2,  2,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  3, -3,  0 },      0.00000, {      0.00,      0.00,      0.00,     -0.02,     -0.03,      0.00 } },
 { {  3,  3, -2, -2 },      0.00000, {     -0.03,      0.01,     -0.03,      0.00,     -0.08,      0.00 } },
 { {  3,  3, -2,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  3,  3, -1,  0 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,     -0.05,      0.00 } },
 { {  3,  3,  0, -2 },     -0.00002, {     -0.09,      0.07,     -0.08,     -0.08,      0.05,      0.00 } },
 { {  3,  3,  0,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -5, -2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -5, -1,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  4, -5,  0,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -4, -3,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -4, -2,  0 },      0.00033, {      0.02,      0.00,      0.01,      0.08,      0.00,      0.00 } },
 { {  4, -4, -1,  0 },      0.00086, {      0.05,      0.00,      0.02,      0.21,      0.00,      0.00 } },
 { {  4, -4,  0,  0 },      0.00054, {      0.04,      0.00,      0.00,      0.13,      0.00,      0.00 } },
 { {  4, -4,  1,  0 },      0.00009, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4, -4,  2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -3,  0 },      0.00027, {      0.01,      0.00,      0.01,      0.05,      0.00,      0.00 } },
 { {  4, -3, -2,  0 },      0.00755, {      0.33,      0.00,      0.28,      1.35,      0.00,      0.00 } },
 { {  4, -3, -2,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -1, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -1,  0 },      0.01741, {      0.97,     -0.01,      0.36,      3.12,      0.00,      0.00 } },
 { {  4, -3, -1,  2 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -3,  0, -2 },      0.00007, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  4, -3,  0,  0 },      0.00971, {      0.65,      0.00,      0.11,      1.74,      0.00,      0.00 } },
 { {  4, -3,  0,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3,  1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3,  1,  0 },      0.00160, {      0.11,      0.00,      0.04,      0.29,      0.00,      0.00 } },
 { {  4, -3,  1,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3,  2,  0 },      0.00020, {      0.01,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  4, -3,  3,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -5,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -4,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -3,  0 },      0.00560, {      0.27,      0.00,      0.31,      0.67,      0.00,      0.00 } },
 { {  4, -2, -3,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -2, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -2,  0 },      0.15781, {      6.46,     -0.06,      5.77,     18.87,      0.00,      0.00 } },
 { {  4, -2, -2,  2 },     -0.00090, {     -0.04,     -0.04,     -0.03,     -0.11,      0.00,      0.00 } },
 { {  4, -2, -1, -2 },      0.00108, {      0.00,      0.05,      0.02,      0.13,      0.00,      0.00 } },
 { {  4, -2, -1,  0 },      0.30874, {     16.37,     -0.23,      6.49,     36.90,      0.00,      0.00 } },
 { {  4, -2, -1,  2 },     -0.00160, {     -0.08,     -0.07,     -0.04,     -0.19,      0.00,      0.00 } },
 { {  4, -2,  0, -2 },      0.00091, {      0.07,      0.04,     -0.02,      0.11,      0.00,      0.00 } },
 { {  4, -2,  0,  0 },      0.15227, {      9.62,     -0.17,      1.88,     18.19,      0.01,      0.00 } },
 { {  4, -2,  0,  2 },     -0.00089, {     -0.05,     -0.04,     -0.02,     -0.11,      0.00,      0.00 } },
 { {  4, -2,  1, -2 },     -0.00052, {     -0.02,     -0.02,     -0.01,     -0.06,      0.00,      0.00 } },
 { {  4, -2,  1,  0 },      0.02411, {      1.58,     -0.03,      0.66,      2.88,      0.00,      0.00 } },
 { {  4, -2,  1,  2 },     -0.00021, {     -0.01,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -2,  2, -2 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  4, -2,  2,  0 },      0.00290, {      0.20,      0.00,      0.13,      0.35,      0.00,      0.00 } },
 { {  4, -2,  2,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  3, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  3,  0 },      0.00031, {      0.02,      0.00,      0.02,      0.04,      0.00,      0.00 } },
 { {  4, -2,  4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -6,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -5,  0 },     -0.00034, {     -0.01,      0.00,     -0.03,     -0.02,      0.00,      0.00 } },
 { {  4, -1, -4, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -4,  0 },     -0.00020, {      0.00,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  4, -1, -3, -2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -3,  0 },      0.09982, {      4.58,     -0.04,      5.46,      5.98,      0.02,      0.00 } },
 { {  4, -1, -3,  2 },     -0.00078, {     -0.03,     -0.04,     -0.04,     -0.05,      0.00,      0.00 } },
 { {  4, -1, -2, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -2, -2 },      0.00043, {      0.00,      0.02,      0.02,      0.03,      0.00,      0.00 } },
 { {  4, -1, -2,  0 },      2.73198, {    106.73,     -1.04,     99.83,    163.20,      0.11,      0.00 } },
 { {  4, -1, -2,  2 },     -0.01533, {     -0.59,     -0.68,     -0.57,     -0.92,      0.00,      0.00 } },
 { {  4, -1, -2,  4 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -1, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -1, -2 },      0.02289, {      0.27,      1.02,      0.41,      1.36,      0.02,      0.00 } },
 { {  4, -1, -1,  0 },      4.37416, {    221.00,     -3.71,     93.87,    260.99,     -0.05,      0.00 } },
 { {  4, -1, -1,  2 },     -0.02273, {     -1.12,     -1.00,     -0.56,     -1.36,      0.00,      0.00 } },
 { {  4, -1, -1,  4 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  0, -4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  0, -2 },      0.00666, {      0.75,      0.26,     -0.27,      0.39,      0.02,      0.00 } },
 { {  4, -1,  0,  0 },      1.87083, {    111.80,     -2.35,     25.65,    111.49,      0.16,      0.00 } },
 { {  4, -1,  0,  2 },     -0.01114, {     -0.64,     -0.49,     -0.23,     -0.66,      0.00,      0.00 } },
 { {  4, -1,  0,  4 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  1, -4 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  1, -2 },     -0.00732, {     -0.28,     -0.33,     -0.18,     -0.44,      0.00,      0.00 } },
 { {  4, -1,  1,  0 },      0.28251, {     17.39,     -0.42,      8.05,     16.82,      0.02,      0.00 } },
 { {  4, -1,  1,  2 },     -0.00251, {     -0.15,     -0.11,     -0.08,     -0.15,      0.00,      0.00 } },
 { {  4, -1,  1,  4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  2, -2 },     -0.00121, {     -0.05,     -0.05,     -0.05,     -0.07,      0.00,      0.00 } },
 { {  4, -1,  2,  0 },      0.03274, {      2.05,     -0.06,      1.49,      1.95,      0.00,      0.00 } },
 { {  4, -1,  2,  2 },     -0.00041, {     -0.03,     -0.02,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  4, -1,  3, -2 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  3,  0 },      0.00336, {      0.21,      0.00,      0.21,      0.20,      0.00,      0.00 } },
 { {  4, -1,  3,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  4, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  4,  0 },      0.00032, {      0.02,      0.00,      0.03,      0.02,      0.00,      0.00 } },
 { {  4, -1,  5,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -7,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -6, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -6,  0 },     -0.00062, {     -0.03,      0.00,     -0.07,      0.00,      0.00,      0.00 } },
 { {  4,  0, -5, -2 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -5,  0 },     -0.00431, {     -0.18,      0.00,     -0.39,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4, -2 },      0.00042, {      0.02,      0.02,      0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  0 },     -0.00070, {      0.03,      0.00,     -0.05,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3, -2 },      0.00023, {      0.02,      0.01,      0.01,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3,  0 },      1.18682, {     52.89,     -0.50,     64.88,      0.22,      0.25,      0.00 } },
 { {  4,  0, -3,  2 },     -0.00934, {     -0.29,     -0.42,     -0.51,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3,  4 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -4 },      0.00013, {      0.00,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -2 },      0.00662, {      0.09,      0.29,      0.24,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2,  0 },     30.77247, {   1164.74,    -12.42,   1123.35,     -4.25,      1.18,      0.00 } },
 { {  4,  0, -2,  2 },     -0.16977, {     -6.36,     -7.54,     -6.26,      0.03,      0.00,      0.00 } },
 { {  4,  0, -2,  4 },      0.00061, {      0.02,      0.05,      0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -4 },      0.00017, {      0.01,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -2 },      0.30086, {      5.06,     13.37,      5.39,     -0.19,      0.15,      0.00 } },
 { {  4,  0, -1,  0 },     38.42974, {   1853.28,    -37.86,    849.77,     -9.84,      0.08,      0.00 } },
 { {  4,  0, -1,  2 },     -0.20134, {     -9.45,     -8.85,     -5.16,      0.05,      0.00,      0.00 } },
 { {  4,  0, -1,  4 },      0.00072, {      0.03,      0.06,      0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -2 },     -0.02391, {      4.23,     -1.47,     -2.51,     -0.04,      0.16,      0.00 } },
 { {  4,  0,  0,  0 },     13.89903, {    782.37,    -19.88,    215.74,     -5.22,      1.37,      0.00 } },
 { {  4,  0,  0,  2 },     -0.08500, {     -4.56,     -3.71,     -1.97,      0.03,      0.00,      0.00 } },
 { {  4,  0,  0,  4 },      0.00035, {      0.02,      0.03,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -4 },     -0.00061, {      0.00,     -0.05,     -0.01,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -2 },     -0.06601, {     -2.12,     -2.98,     -1.62,      0.03,      0.02,      0.00 } },
 { {  4,  0,  1,  0 },      1.97772, {    113.34,     -3.31,     59.21,     -0.94,      0.18,      0.00 } },
 { {  4,  0,  1,  2 },     -0.01781, {     -0.99,     -0.77,     -0.63,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1,  4 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2, -4 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2, -2 },     -0.00980, {     -0.35,     -0.44,     -0.41,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2,  0 },      0.21853, {     12.57,     -0.41,     10.19,     -0.13,      0.02,      0.00 } },
 { {  4,  0,  2,  2 },     -0.00274, {     -0.15,     -0.12,     -0.14,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2,  4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3, -2 },     -0.00109, {     -0.04,     -0.05,     -0.07,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3,  0 },      0.02149, {      1.24,     -0.04,      1.38,     -0.02,      0.00,      0.00 } },
 { {  4,  0,  3,  2 },     -0.00036, {     -0.02,     -0.02,     -0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4, -2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4,  0 },      0.00198, {      0.11,      0.00,      0.16,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  5,  0 },      0.00017, {      0.01,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0,  6,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -6,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -5,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -4,  0 },      0.00064, {      0.06,      0.00,      0.05,      0.04,      0.00,      0.00 } },
 { {  4,  1, -3, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -3,  0 },      0.02682, {      3.50,     -0.05,      1.45,      1.60,      0.02,      0.00 } },
 { {  4,  1, -3,  2 },     -0.00003, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -2, -2 },     -0.00039, {     -0.02,     -0.02,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  4,  1, -2,  0 },     -0.35759, {      7.75,     -0.18,    -13.13,    -21.49,     -0.04,      0.00 } },
 { {  4,  1, -2,  2 },      0.00220, {     -0.03,      0.10,      0.08,      0.13,      0.00,      0.00 } },
 { {  4,  1, -1, -2 },     -0.01875, {     -0.68,     -0.83,     -0.35,     -1.12,      0.00,      0.00 } },
 { {  4,  1, -1,  0 },     -0.63523, {    -17.62,      0.74,    -14.32,    -38.11,     -0.21,      0.00 } },
 { {  4,  1, -1,  2 },      0.00353, {      0.09,      0.15,      0.09,      0.21,      0.00,      0.00 } },
 { {  4,  1, -1,  4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  0, -2 },      0.00051, {     -0.28,      0.04,      0.02,      0.03,     -0.01,      0.00 } },
 { {  4,  1,  0,  0 },     -0.28911, {    -15.10,      0.59,     -4.73,    -17.30,     -0.06,      0.00 } },
 { {  4,  1,  0,  2 },      0.00186, {      0.09,      0.08,      0.05,      0.11,      0.00,      0.00 } },
 { {  4,  1,  1, -4 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1, -2 },      0.00183, {      0.03,      0.08,      0.04,      0.11,      0.00,      0.00 } },
 { {  4,  1,  1,  0 },     -0.05070, {     -2.99,      0.12,     -1.55,     -3.03,     -0.01,      0.00 } },
 { {  4,  1,  1,  2 },      0.00048, {      0.03,      0.02,      0.02,      0.03,      0.00,      0.00 } },
 { {  4,  1,  2, -2 },      0.00032, {      0.01,      0.01,      0.01,      0.02,      0.00,      0.00 } },
 { {  4,  1,  2,  0 },     -0.00668, {     -0.42,      0.02,     -0.32,     -0.40,      0.00,      0.00 } },
 { {  4,  1,  2,  2 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  3, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  3,  0 },     -0.00077, {     -0.05,      0.00,     -0.05,     -0.05,      0.00,      0.00 } },
 { {  4,  1,  3,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  4,  0 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -4,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -3,  0 },     -0.00369, {     -0.24,      0.00,     -0.20,     -0.44,      0.00,      0.00 } },
 { {  4,  2, -3,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -2, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -2,  0 },     -0.03514, {     -2.04,      0.02,     -1.29,     -4.21,      0.00,      0.00 } },
 { {  4,  2, -2,  2 },      0.00019, {      0.01,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4,  2, -1, -2 },      0.00090, {      0.06,      0.04,      0.02,      0.11,      0.00,      0.00 } },
 { {  4,  2, -1,  0 },     -0.02171, {     -1.51,      0.03,     -0.57,     -2.60,      0.01,      0.00 } },
 { {  4,  2, -1,  2 },      0.00012, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  4,  2,  0, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2,  0,  0 },     -0.00208, {     -0.16,      0.00,     -0.12,     -0.25,      0.00,      0.00 } },
 { {  4,  2,  0,  2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2,  1, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2,  1,  0 },     -0.00008, {      0.00,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  4,  2,  2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  3, -2,  0 },      0.00020, {      0.00,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  4,  3, -1,  0 },      0.00023, {      0.01,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  4,  3,  0,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  4, -2, -2 },      0.00000, {     -0.02,      0.00,      0.00,      0.04,     -0.06,      0.00 } },
 { {  5, -3, -2,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5, -3,  0,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2, -3,  0 },     -0.00013, {      0.00,      0.00,      0.00,     -0.02,     -0.05,      0.00 } },
 { {  5, -2, -2,  0 },     -0.00052, {     -0.03,      0.00,     -0.02,     -0.06,     -0.20,      0.00 } },
 { {  5, -2, -1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5, -2, -1,  0 },     -0.00011, {     -0.02,      0.00,      0.00,     -0.01,     -0.04,      0.00 } },
 { {  5, -2,  0, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2,  0,  0 },      0.00012, {      0.00,      0.00,      0.00,      0.01,      0.05,      0.00 } },
 { {  5, -2,  1,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -4,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5, -1, -3,  0 },     -0.00189, {     -0.08,      0.00,     -0.10,     -0.11,     -0.73,      0.00 } },
 { {  5, -1, -3,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -2, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5, -1, -2,  0 },     -0.00693, {     -0.40,      0.00,     -0.26,     -0.41,     -2.70,      0.01 } },
 { {  5, -1, -2,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  5, -1, -1, -2 },     -0.00040, {     -0.02,     -0.02,      0.00,     -0.02,     -0.16,      0.00 } },
 { {  5, -1, -1,  0 },     -0.00154, {     -0.21,      0.01,     -0.08,     -0.09,     -0.60,      0.00 } },
 { {  5, -1, -1,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  5, -1,  0, -2 },     -0.00021, {      0.00,      0.00,      0.00,     -0.01,     -0.08,      0.00 } },
 { {  5, -1,  0,  0 },      0.00102, {      0.01,      0.00,     -0.02,      0.06,      0.40,      0.00 } },
 { {  5, -1,  1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5, -1,  1,  0 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  5,  0, -4,  0 },     -0.00059, {     -0.03,      0.00,     -0.04,      0.00,     -0.23,      0.00 } },
 { {  5,  0, -3,  0 },     -0.01686, {     -0.72,      0.00,     -0.92,      0.00,     -6.56,      0.03 } },
 { {  5,  0, -3,  2 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  5,  0, -2, -2 },     -0.00031, {      0.00,     -0.01,     -0.01,      0.00,     -0.12,      0.00 } },
 { {  5,  0, -2,  0 },     -0.05725, {     -3.18,      0.07,     -2.18,      0.05,    -22.26,      0.12 } },
 { {  5,  0, -2,  2 },      0.00039, {      0.02,      0.02,      0.02,      0.00,      0.15,      0.00 } },
 { {  5,  0, -1, -2 },     -0.00335, {     -0.11,     -0.15,     -0.06,      0.00,     -1.30,      0.00 } },
 { {  5,  0, -1,  0 },     -0.01350, {     -1.57,      0.09,     -0.64,      0.06,     -5.25,      0.03 } },
 { {  5,  0, -1,  2 },      0.00024, {      0.02,      0.01,      0.00,      0.00,      0.09,      0.00 } },
 { {  5,  0,  0, -4 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0,  0, -2 },     -0.00137, {     -0.06,     -0.06,      0.00,      0.00,     -0.53,      0.00 } },
 { {  5,  0,  0,  0 },      0.00439, {     -0.05,      0.03,     -0.18,      0.03,      1.71,      0.00 } },
 { {  5,  0,  0,  2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  5,  0,  1, -2 },     -0.00019, {      0.00,      0.00,      0.00,      0.00,     -0.07,      0.00 } },
 { {  5,  0,  1,  0 },      0.00050, {     -0.03,      0.00,     -0.02,      0.00,      0.20,      0.00 } },
 { {  5,  0,  1,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0,  2, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0,  2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1, -4,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  1, -3,  0 },     -0.00046, {     -0.05,      0.00,     -0.02,     -0.03,     -0.18,      0.00 } },
 { {  5,  1, -2, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1, -2,  0 },      0.00392, {      0.12,      0.00,      0.15,      0.23,      1.52,      0.00 } },
 { {  5,  1, -2,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5,  1, -1, -2 },      0.00016, {      0.00,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  5,  1, -1,  0 },      0.00475, {      0.24,     -0.01,      0.12,      0.28,      1.85,      0.00 } },
 { {  5,  1, -1,  2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  1,  0, -2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  5,  1,  0,  0 },      0.00181, {      0.11,      0.00,      0.05,      0.11,      0.70,      0.00 } },
 { {  5,  1,  0,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1,  1,  0 },      0.00035, {      0.02,      0.00,      0.01,      0.02,      0.14,      0.00 } },
 { {  5,  1,  2,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  5,  2, -3,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  2, -2,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  2, -1,  0 },     -0.00008, {      0.00,      0.00,      0.00,     -0.01,     -0.03,      0.00 } },
 { {  5,  2,  0,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  2,  1,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  3, -2, -2 },      0.00000, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  6, -4, -3,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -4, -2,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -4, -1,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -4,  0,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3, -3,  0 },      0.00020, {      0.01,      0.00,      0.01,      0.03,      0.00,      0.00 } },
 { {  6, -3, -2,  0 },      0.00059, {      0.04,      0.00,      0.02,      0.11,      0.00,      0.00 } },
 { {  6, -3, -1,  0 },      0.00059, {      0.05,      0.00,      0.02,      0.11,      0.00,      0.00 } },
 { {  6, -3,  0,  0 },      0.00024, {      0.02,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  6, -3,  1,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -4,  0 },      0.00014, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  6, -2, -3,  0 },      0.00312, {      0.19,      0.00,      0.17,      0.37,      0.00,      0.00 } },
 { {  6, -2, -3,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -2, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -2,  0 },      0.00836, {      0.59,      0.00,      0.33,      1.00,      0.00,      0.00 } },
 { {  6, -2, -2,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -1, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -1,  0 },      0.00744, {      0.61,     -0.01,      0.21,      0.89,      0.00,      0.00 } },
 { {  6, -2, -1,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  0, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  0,  0 },      0.00285, {      0.26,      0.00,      0.07,      0.34,      0.00,      0.00 } },
 { {  6, -2,  0,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  1, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  1,  0 },      0.00056, {      0.05,      0.00,      0.02,      0.07,      0.00,      0.00 } },
 { {  6, -2,  2,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  3,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -5,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -4,  0 },      0.00176, {      0.11,      0.00,      0.13,      0.11,      0.00,      0.00 } },
 { {  6, -1, -4,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -3,  0 },      0.03900, {      2.28,     -0.02,      2.14,      2.33,      0.00,      0.00 } },
 { {  6, -1, -3,  2 },     -0.00039, {     -0.02,     -0.02,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  6, -1, -2, -2 },      0.00044, {      0.01,      0.02,      0.02,      0.03,      0.00,      0.00 } },
 { {  6, -1, -2,  0 },      0.09033, {      6.23,     -0.11,      3.59,      5.38,      0.00,      0.00 } },
 { {  6, -1, -2,  2 },     -0.00080, {     -0.05,     -0.04,     -0.03,     -0.05,      0.00,      0.00 } },
 { {  6, -1, -1, -2 },      0.00018, {      0.02,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  6, -1, -1,  0 },      0.07142, {      5.64,     -0.12,      2.05,      4.25,      0.00,      0.00 } },
 { {  6, -1, -1,  2 },     -0.00064, {     -0.05,     -0.03,     -0.02,     -0.04,      0.00,      0.00 } },
 { {  6, -1,  0, -2 },     -0.00025, {      0.00,     -0.01,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  6, -1,  0,  0 },      0.02510, {      2.16,     -0.05,      0.66,      1.49,      0.00,      0.00 } },
 { {  6, -1,  0,  2 },     -0.00026, {     -0.02,     -0.01,      0.00,     -0.02,      0.00,      0.00 } },
 { {  6, -1,  1, -2 },     -0.00016, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  1,  0 },      0.00477, {      0.42,     -0.01,      0.18,      0.28,      0.00,      0.00 } },
 { {  6, -1,  1,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  2, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  2,  0 },      0.00069, {      0.06,      0.00,      0.04,      0.04,      0.00,      0.00 } },
 { {  6, -1,  2,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  3,  0 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -5,  0 },      0.00020, {      0.01,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  0 },      0.01376, {      0.88,      0.00,      1.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  2 },     -0.00017, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3, -2 },      0.00018, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3,  0 },      0.29255, {     16.72,     -0.19,     16.04,     -0.08,      0.03,      0.00 } },
 { {  6,  0, -3,  2 },     -0.00287, {     -0.16,     -0.13,     -0.16,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3,  4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2, -2 },      0.00364, {      0.13,      0.16,      0.13,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2,  0 },      0.57156, {     38.34,     -0.75,     22.97,     -0.26,      0.02,      0.00 } },
 { {  6,  0, -2,  2 },     -0.00509, {     -0.34,     -0.22,     -0.22,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2,  4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1, -2 },      0.00007, {      0.08,      0.00,     -0.03,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1,  0 },      0.39451, {     30.10,     -0.74,     11.79,     -0.25,      0.03,      0.00 } },
 { {  6,  0, -1,  2 },     -0.00359, {     -0.27,     -0.16,     -0.13,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1,  4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -4 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -2 },     -0.00214, {     -0.07,     -0.10,     -0.07,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0,  0 },      0.12616, {     10.43,     -0.29,      3.57,     -0.10,      0.02,      0.00 } },
 { {  6,  0,  0,  2 },     -0.00132, {     -0.11,     -0.06,     -0.05,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1, -4 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1, -2 },     -0.00094, {     -0.05,     -0.04,     -0.03,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1,  0 },      0.02285, {      1.93,     -0.06,      0.91,     -0.02,      0.00,      0.00 } },
 { {  6,  0,  1,  2 },     -0.00031, {     -0.02,     -0.01,     -0.01,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2, -2 },     -0.00018, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2,  0 },      0.00318, {      0.27,      0.00,      0.18,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2,  2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  3, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  3,  0 },      0.00038, {      0.03,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  6,  0,  4,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -5,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -4,  0 },      0.00028, {      0.05,      0.00,      0.02,      0.02,      0.00,      0.00 } },
 { {  6,  1, -3,  0 },     -0.00478, {      0.05,      0.00,     -0.26,     -0.29,      0.00,      0.00 } },
 { {  6,  1, -3,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -2, -2 },     -0.00026, {     -0.01,     -0.01,      0.00,     -0.02,      0.00,      0.00 } },
 { {  6,  1, -2,  0 },     -0.01282, {     -0.47,      0.02,     -0.52,     -0.77,      0.00,      0.00 } },
 { {  6,  1, -2,  2 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -1, -2 },     -0.00007, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -1,  0 },     -0.01064, {     -0.66,      0.02,     -0.32,     -0.64,      0.00,      0.00 } },
 { {  6,  1, -1,  2 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  0, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  0,  0 },     -0.00396, {     -0.31,      0.01,     -0.12,     -0.24,      0.00,      0.00 } },
 { {  6,  1,  0,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  1, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  1,  0 },     -0.00083, {     -0.07,      0.00,     -0.03,     -0.05,      0.00,      0.00 } },
 { {  6,  1,  1,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  2,  0 },     -0.00013, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  3,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -4,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -3,  0 },     -0.00051, {     -0.04,      0.00,     -0.03,     -0.06,      0.00,      0.00 } },
 { {  6,  2, -2, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -2,  0 },     -0.00061, {     -0.06,      0.00,     -0.03,     -0.07,      0.00,      0.00 } },
 { {  6,  2, -1, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -1,  0 },     -0.00021, {     -0.02,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  6,  2,  0,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  3, -2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -2, -3,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -2, -2,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -4,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  7, -1, -3,  0 },     -0.00016, {     -0.01,      0.00,      0.00,     -0.01,     -0.06,      0.00 } },
 { {  7, -1, -2,  0 },     -0.00012, {     -0.01,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  7, -1,  0,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -5,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -4,  0 },     -0.00025, {     -0.02,      0.00,     -0.02,      0.00,     -0.10,      0.00 } },
 { {  7,  0, -3,  0 },     -0.00098, {     -0.07,      0.00,     -0.06,      0.00,     -0.38,      0.00 } },
 { {  7,  0, -3,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -2, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  7,  0, -2,  0 },     -0.00068, {     -0.07,      0.00,     -0.03,      0.00,     -0.26,      0.00 } },
 { {  7,  0, -2,  2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -1, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  7,  0, -1,  0 },     -0.00005, {     -0.02,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  7,  0,  0, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0,  0,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  7,  1, -3,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  7,  1, -2,  0 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  7,  1, -1,  0 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  7,  1,  0,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  8, -3, -3,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -3, -2,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -3, -1,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -4,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -3,  0 },      0.00019, {      0.02,      0.00,      0.01,      0.02,      0.00,      0.00 } },
 { {  8, -2, -2,  0 },      0.00025, {      0.03,      0.00,      0.01,      0.03,      0.00,      0.00 } },
 { {  8, -2, -1,  0 },      0.00015, {      0.02,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  8, -2,  0,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2,  1,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -5,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -4,  0 },      0.00058, {      0.04,      0.00,      0.04,      0.03,      0.00,      0.00 } },
 { {  8, -1, -3,  0 },      0.00170, {      0.15,      0.00,      0.10,      0.10,      0.00,      0.00 } },
 { {  8, -1, -3,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -2,  0 },      0.00196, {      0.19,      0.00,      0.09,      0.12,      0.00,      0.00 } },
 { {  8, -1, -2,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -1,  0 },      0.00112, {      0.12,      0.00,      0.04,      0.07,      0.00,      0.00 } },
 { {  8, -1, -1,  2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  0,  0 },      0.00037, {      0.04,      0.00,      0.02,      0.02,      0.00,      0.00 } },
 { {  8, -1,  1,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  2,  0 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -5,  0 },      0.00018, {      0.01,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  8,  0, -4,  0 },      0.00324, {      0.25,      0.00,      0.24,      0.00,      0.00,      0.00 } },
 { {  8,  0, -4,  2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3, -2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3,  0 },      0.00840, {      0.72,     -0.01,      0.49,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3,  2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2, -2 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2,  0 },      0.00866, {      0.83,     -0.02,      0.40,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2,  2 },     -0.00011, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1,  0 },      0.00455, {      0.47,     -0.01,      0.18,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0,  0 },      0.00139, {      0.15,      0.00,      0.06,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  1, -2 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  1,  0 },      0.00029, {      0.03,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  8,  0,  2,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -4,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -3,  0 },     -0.00024, {     -0.01,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  8,  1, -2,  0 },     -0.00029, {     -0.02,      0.00,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  8,  1, -1,  0 },     -0.00017, {     -0.02,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  8,  1,  0,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1,  1,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  2, -3,  0 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  9,  0, -4,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  9,  0, -3,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -3,  0 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -2,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -1,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -5,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -4,  0 },      0.00012, {      0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -3,  0 },      0.00017, {      0.02,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { { 10,  0, -2,  0 },      0.00013, {      0.02,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -1,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0,  0,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } }
} };

constexpr array<ELP2000MainProblemCoefficient, 918> g_ELP2
{ {
 { {  0,  0,  0,  1 },  18461.40000, {      0.00, 412529.61,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  0,  3 },     -6.29664, {      7.68,   -422.65,    -13.21,      0.02,     -0.02,      0.00 } },
 { {  0,  0,  0,  5 },      0.00592, {      0.00,      0.66,      0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -5 },     -0.00236, {      0.02,     -0.26,     -0.05,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -3 },      2.79871, {    -17.13,    188.68,     50.78,     -0.11,      0.04,      0.00 } },
 { {  0,  0,  1, -1 },    999.70079, {   -563.82,  22508.54,  18298.82,     -0.92,     -0.21,      0.00 } },
 { {  0,  0,  1,  1 },   1010.17430, {    -93.16,  22571.83,  18386.36,     -0.76,     -0.17,      0.00 } },
 { {  0,  0,  1,  3 },     -1.01941, {      0.95,    -68.42,    -19.88,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1,  5 },      0.00156, {      0.00,      0.17,      0.03,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -5 },     -0.00044, {      0.00,     -0.05,     -0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -3 },     -0.13035, {     -0.40,     -8.78,     -4.77,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -1 },     31.75985, {   -167.64,    723.95,   1159.97,     -1.09,      0.22,      0.00 } },
 { {  0,  0,  2,  1 },     61.91229, {    -23.73,   1382.40,   2253.18,     -0.62,     -0.06,      0.00 } },
 { {  0,  0,  2,  3 },     -0.11787, {      0.11,     -7.91,     -4.41,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2,  5 },      0.00026, {      0.00,      0.03,      0.01,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -5 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -3 },     -0.00146, {      0.04,     -0.10,     -0.08,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -1 },      1.58131, {    -10.25,     36.06,     86.53,     -0.08,      0.00,      0.00 } },
 { {  0,  0,  3,  1 },      3.98407, {     -3.30,     88.86,    217.43,     -0.11,     -0.01,      0.00 } },
 { {  0,  0,  3,  3 },     -0.01181, {      0.01,     -0.79,     -0.65,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3,  5 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4, -3 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4, -1 },      0.09157, {     -0.70,      2.09,      6.68,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4,  1 },      0.26325, {     -0.37,      5.86,     19.15,     -0.01,      0.00,      0.00 } },
 { {  0,  0,  4,  3 },     -0.00109, {      0.00,     -0.07,     -0.08,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5, -1 },      0.00566, {     -0.05,      0.13,      0.51,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  1 },      0.01768, {     -0.04,      0.39,      1.61,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  3 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6, -1 },      0.00036, {      0.00,      0.00,      0.04,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6,  1 },      0.00120, {      0.00,      0.03,      0.13,      0.00,      0.00,      0.00 } },
 { {  0,  0,  7, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  7,  1 },      0.00008, {      0.00,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { {  0,  1, -6, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -6,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -5, -1 },     -0.00053, {     -0.01,     -0.01,     -0.05,     -0.03,      0.00,      0.00 } },
 { {  0,  1, -5,  1 },     -0.00014, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  0,  1, -4, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -4, -1 },     -0.00646, {     -0.14,     -0.14,     -0.47,     -0.39,      0.00,      0.00 } },
 { {  0,  1, -4,  1 },     -0.00186, {     -0.02,     -0.04,     -0.14,     -0.11,      0.00,      0.00 } },
 { {  0,  1, -3, -3 },      0.00023, {      0.00,      0.01,      0.01,      0.01,      0.00,      0.00 } },
 { {  0,  1, -3, -1 },     -0.07479, {     -1.69,     -1.63,     -4.08,     -4.48,      0.00,      0.00 } },
 { {  0,  1, -3,  1 },     -0.02365, {     -0.35,     -0.53,     -1.29,     -1.42,      0.00,      0.00 } },
 { {  0,  1, -3,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -2, -3 },      0.00154, {      0.03,      0.10,      0.06,      0.09,      0.00,      0.00 } },
 { {  0,  1, -2, -1 },     -0.79322, {    -18.59,    -17.32,    -28.86,    -47.52,     -0.01,      0.00 } },
 { {  0,  1, -2,  1 },     -0.30129, {     -4.79,     -6.71,    -11.01,    -18.04,     -0.01,      0.00 } },
 { {  0,  1, -2,  3 },      0.00191, {      0.06,      0.13,      0.07,      0.12,      0.00,      0.00 } },
 { {  0,  1, -1, -5 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -1, -3 },      0.00679, {      0.14,      0.45,      0.14,      0.41,      0.00,      0.00 } },
 { {  0,  1, -1, -1 },     -6.73173, {   -159.36,   -146.94,   -122.89,   -403.36,     -0.03,      0.00 } },
 { {  0,  1, -1,  1 },     -5.63260, {   -148.81,   -123.96,   -102.80,   -337.62,     -0.04,      0.00 } },
 { {  0,  1, -1,  3 },     -0.00911, {     -0.20,     -0.61,     -0.15,     -0.55,      0.00,      0.00 } },
 { {  0,  1,  0, -3 },      0.00010, {     -0.17,      0.00,      0.06,      0.00,      0.00,      0.00 } },
 { {  0,  1,  0, -1 },     -4.83983, {     70.86,   -100.97,     -8.74,   -289.05,      0.02,      0.00 } },
 { {  0,  1,  0,  1 },     -6.46036, {     28.14,   -136.25,    -20.21,   -386.09,      0.29,      0.00 } },
 { {  0,  1,  0,  3 },      0.01157, {      0.06,      0.77,      0.04,      0.69,      0.00,      0.00 } },
 { {  0,  1,  0,  5 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  1, -3 },     -0.00703, {     -0.04,     -0.47,     -0.13,     -0.42,      0.00,      0.00 } },
 { {  0,  1,  1, -1 },     -5.07614, {   -113.16,   -112.23,    -93.12,   -304.28,      0.00,      0.00 } },
 { {  0,  1,  1,  1 },     -5.31151, {    -99.03,   -115.96,    -99.19,   -318.30,      0.06,      0.00 } },
 { {  0,  1,  1,  3 },      0.00668, {      0.10,      0.45,      0.13,      0.40,      0.00,      0.00 } },
 { {  0,  1,  1,  5 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  2, -3 },      0.00141, {      0.04,      0.09,      0.05,      0.08,      0.00,      0.00 } },
 { {  0,  1,  2, -1 },     -0.31292, {     -4.92,     -6.96,    -11.47,    -18.75,      0.00,      0.00 } },
 { {  0,  1,  2,  1 },     -0.63884, {    -12.56,    -13.95,    -23.51,    -38.29,      0.01,      0.00 } },
 { {  0,  1,  2,  3 },      0.00135, {      0.02,      0.09,      0.05,      0.08,      0.00,      0.00 } },
 { {  0,  1,  3, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  3, -1 },     -0.02419, {     -0.36,     -0.54,     -1.33,     -1.45,      0.00,      0.00 } },
 { {  0,  1,  3,  1 },     -0.06176, {     -1.23,     -1.35,     -3.40,     -3.70,      0.00,      0.00 } },
 { {  0,  1,  3,  3 },      0.00020, {      0.00,      0.01,      0.01,      0.01,      0.00,      0.00 } },
 { {  0,  1,  4, -1 },     -0.00191, {     -0.03,     -0.04,     -0.14,     -0.11,      0.00,      0.00 } },
 { {  0,  1,  4,  1 },     -0.00548, {     -0.11,     -0.12,     -0.40,     -0.33,      0.00,      0.00 } },
 { {  0,  1,  4,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  5, -1 },     -0.00015, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  0,  1,  5,  1 },     -0.00046, {      0.00,     -0.01,     -0.04,     -0.03,      0.00,      0.00 } },
 { {  0,  1,  6, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  6,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -5, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -4, -1 },     -0.00016, {      0.00,      0.00,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  0,  2, -4,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -3, -1 },     -0.00169, {     -0.05,     -0.04,     -0.09,     -0.20,      0.00,      0.00 } },
 { {  0,  2, -3,  1 },     -0.00049, {     -0.01,     -0.01,     -0.03,     -0.06,      0.00,      0.00 } },
 { {  0,  2, -2, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -2, -1 },     -0.01571, {     -0.49,     -0.34,     -0.57,     -1.88,      0.00,      0.00 } },
 { {  0,  2, -2,  1 },     -0.00536, {     -0.12,     -0.12,     -0.20,     -0.64,      0.00,      0.00 } },
 { {  0,  2, -2,  3 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -1, -3 },      0.00010, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  0,  2, -1, -1 },     -0.11335, {     -3.25,     -2.46,     -2.07,    -13.58,      0.00,      0.00 } },
 { {  0,  2, -1,  1 },     -0.09511, {     -3.17,     -2.08,     -1.73,    -11.39,      0.00,      0.00 } },
 { {  0,  2, -1,  3 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  2,  0, -3 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  0,  2,  0, -1 },     -0.01801, {      3.13,     -0.33,     -0.14,     -2.14,      0.00,      0.00 } },
 { {  0,  2,  0,  1 },     -0.05729, {      2.09,     -1.19,     -0.39,     -6.84,      0.00,      0.00 } },
 { {  0,  2,  0,  3 },      0.00016, {      0.00,      0.01,      0.00,      0.02,      0.00,      0.00 } },
 { {  0,  2,  1, -3 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  1, -1 },     -0.06187, {     -1.29,     -1.37,     -1.13,     -7.41,      0.00,      0.00 } },
 { {  0,  2,  1,  1 },     -0.05504, {     -0.81,     -1.21,     -1.04,     -6.59,      0.00,      0.00 } },
 { {  0,  2,  1,  3 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  2, -3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  2, -1 },     -0.00284, {     -0.02,     -0.06,     -0.10,     -0.34,      0.00,      0.00 } },
 { {  0,  2,  2,  1 },     -0.00528, {     -0.06,     -0.12,     -0.19,     -0.63,      0.00,      0.00 } },
 { {  0,  2,  2,  3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  3, -1 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  0,  2,  3,  1 },     -0.00037, {      0.00,      0.00,     -0.02,     -0.04,      0.00,      0.00 } },
 { {  0,  2,  4,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -3, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -3,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -2, -1 },     -0.00033, {     -0.01,      0.00,     -0.01,     -0.06,      0.00,      0.00 } },
 { {  0,  3, -2,  1 },     -0.00011, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  0,  3, -1, -1 },     -0.00216, {     -0.07,     -0.05,     -0.04,     -0.39,      0.00,      0.00 } },
 { {  0,  3, -1,  1 },     -0.00190, {     -0.07,     -0.04,     -0.03,     -0.34,      0.00,      0.00 } },
 { {  0,  3,  0, -1 },      0.00066, {      0.11,      0.02,      0.00,      0.12,      0.00,      0.00 } },
 { {  0,  3,  0,  1 },     -0.00031, {      0.08,      0.00,      0.00,     -0.06,      0.00,      0.00 } },
 { {  0,  3,  1, -1 },     -0.00109, {     -0.02,     -0.02,     -0.02,     -0.19,      0.00,      0.00 } },
 { {  0,  3,  1,  1 },     -0.00083, {     -0.01,     -0.02,     -0.02,     -0.15,      0.00,      0.00 } },
 { {  0,  3,  2, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  2,  1 },     -0.00008, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  4, -1, -1 },     -0.00004, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  0,  4, -1,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4,  0, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4,  0,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4,  1, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4,  1,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -3, -1, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1, -3,  0, -1 },      0.00011, {      0.00,      0.00,      0.00,      0.02,      0.04,      0.00 } },
 { {  1, -3,  0,  1 },      0.00012, {      0.00,      0.00,      0.00,      0.02,      0.05,      0.00 } },
 { {  1, -3,  1, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2, -2, -1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1, -2, -2,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1, -2, -1, -1 },      0.00064, {      0.00,      0.01,      0.01,      0.08,      0.25,      0.00 } },
 { {  1, -2, -1,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2,  0, -1 },      0.00204, {      0.02,      0.05,      0.00,      0.24,      0.79,      0.00 } },
 { {  1, -2,  0,  1 },      0.00214, {      0.00,      0.05,      0.00,      0.26,      0.83,      0.00 } },
 { {  1, -2,  1, -1 },     -0.00060, {     -0.01,     -0.01,     -0.01,     -0.07,     -0.23,      0.00 } },
 { {  1, -2,  1,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1, -2,  2, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -2,  2,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1, -1, -4, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1, -1, -3, -1 },      0.00018, {      0.00,      0.00,      0.01,      0.01,      0.07,      0.00 } },
 { {  1, -1, -3,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1, -1, -2, -1 },      0.00096, {      0.00,      0.02,      0.04,      0.06,      0.37,      0.00 } },
 { {  1, -1, -2,  1 },     -0.00137, {     -0.03,     -0.03,     -0.05,     -0.08,     -0.53,      0.00 } },
 { {  1, -1, -1, -3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1, -1, -1, -1 },      0.01031, {      0.07,      0.23,      0.20,      0.62,      4.01,     -0.02 } },
 { {  1, -1, -1,  1 },      0.00080, {      0.04,      0.03,      0.00,      0.05,      0.31,      0.00 } },
 { {  1, -1,  0, -3 },     -0.00052, {      0.00,     -0.03,      0.00,     -0.03,     -0.20,      0.00 } },
 { {  1, -1,  0, -1 },     -0.01346, {     -1.07,     -0.27,      0.00,     -0.82,     -5.23,      0.03 } },
 { {  1, -1,  0,  1 },     -0.01829, {     -1.52,     -0.33,      0.00,     -1.11,     -7.11,      0.04 } },
 { {  1, -1,  0,  3 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1, -1,  1, -3 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -1,  1, -1 },     -0.02012, {     -0.42,     -0.45,     -0.38,     -1.20,     -7.83,      0.04 } },
 { {  1, -1,  1,  1 },     -0.00845, {     -0.45,     -0.17,     -0.17,     -0.51,     -3.29,      0.02 } },
 { {  1, -1,  1,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  2, -1 },     -0.00175, {     -0.04,     -0.04,     -0.06,     -0.10,     -0.68,      0.00 } },
 { {  1, -1,  2,  1 },     -0.00146, {     -0.07,     -0.03,     -0.06,     -0.09,     -0.57,      0.00 } },
 { {  1, -1,  3, -1 },     -0.00015, {      0.00,      0.00,      0.00,      0.00,     -0.06,      0.00 } },
 { {  1, -1,  3,  1 },     -0.00018, {      0.00,      0.00,     -0.01,     -0.01,     -0.07,      0.00 } },
 { {  1, -1,  4, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  4,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -5, -1 },     -0.00011, {      0.00,      0.00,     -0.01,      0.00,     -0.04,      0.00 } },
 { {  1,  0, -5,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  0, -4, -1 },     -0.00122, {     -0.04,     -0.03,     -0.09,      0.00,     -0.47,      0.00 } },
 { {  1,  0, -4,  1 },     -0.00043, {      0.00,      0.00,     -0.03,      0.00,     -0.17,      0.00 } },
 { {  1,  0, -3, -1 },     -0.01255, {     -0.37,     -0.27,     -0.69,      0.00,     -4.88,      0.03 } },
 { {  1,  0, -3,  1 },     -0.00537, {     -0.12,     -0.12,     -0.29,      0.00,     -2.09,      0.01 } },
 { {  1,  0, -3,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -2, -3 },     -0.00025, {      0.00,     -0.02,      0.00,      0.00,     -0.10,      0.00 } },
 { {  1,  0, -2, -1 },     -0.10964, {     -3.19,     -2.33,     -4.03,     -0.02,    -42.63,      0.22 } },
 { {  1,  0, -2,  1 },     -0.07846, {     -2.00,     -1.72,     -2.87,     -0.03,    -30.51,      0.16 } },
 { {  1,  0, -2,  3 },     -0.00020, {      0.00,     -0.01,      0.00,      0.00,     -0.08,      0.00 } },
 { {  1,  0, -1, -3 },     -0.00369, {     -0.04,     -0.25,     -0.07,      0.00,     -1.44,      0.00 } },
 { {  1,  0, -1, -1 },     -0.42989, {    -13.13,     -8.83,     -8.15,     -0.01,   -167.15,      0.88 } },
 { {  1,  0, -1,  1 },      0.13928, {      1.86,      3.37,      2.46,      0.04,     54.16,     -0.29 } },
 { {  1,  0, -1,  3 },      0.00234, {      0.05,      0.16,      0.04,      0.00,      0.91,      0.00 } },
 { {  1,  0,  0, -5 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1,  0,  0, -3 },     -0.03226, {     -0.41,     -2.15,      0.00,      0.00,    -12.54,      0.07 } },
 { {  1,  0,  0, -1 },     -4.80578, {   -102.50,   -103.39,     -3.12,     -0.09,  -1868.59,      9.85 } },
 { {  1,  0,  0,  1 },     -5.36844, {   -119.82,   -113.86,     -3.16,     -0.08,  -2087.37,     11.00 } },
 { {  1,  0,  0,  3 },      0.00645, {      0.15,      0.43,      0.02,      0.00,      2.51,     -0.01 } },
 { {  1,  0,  0,  5 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0,  1, -5 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0,  1, -3 },     -0.00019, {      0.01,     -0.01,      0.00,      0.00,     -0.07,      0.00 } },
 { {  1,  0,  1, -1 },     -0.58893, {     -7.87,    -13.10,    -10.90,      0.03,   -228.99,      1.21 } },
 { {  1,  0,  1,  1 },     -0.66741, {    -14.68,    -14.15,    -12.47,      0.07,   -259.50,      1.37 } },
 { {  1,  0,  1,  3 },      0.00142, {      0.03,      0.09,      0.03,      0.00,      0.55,      0.00 } },
 { {  1,  0,  2, -1 },     -0.03636, {     -0.40,     -0.81,     -1.34,      0.00,    -14.14,      0.07 } },
 { {  1,  0,  2,  1 },     -0.06383, {     -1.35,     -1.35,     -2.35,      0.02,    -24.82,      0.13 } },
 { {  1,  0,  2,  3 },      0.00021, {      0.00,      0.01,      0.00,      0.00,      0.08,      0.00 } },
 { {  1,  0,  3, -1 },     -0.00247, {     -0.02,     -0.05,     -0.14,      0.00,     -0.96,      0.00 } },
 { {  1,  0,  3,  1 },     -0.00553, {     -0.11,     -0.12,     -0.30,      0.00,     -2.15,      0.01 } },
 { {  1,  0,  3,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0,  4, -1 },     -0.00017, {      0.00,      0.00,     -0.01,      0.00,     -0.07,      0.00 } },
 { {  1,  0,  4,  1 },     -0.00045, {      0.00,      0.00,     -0.03,      0.00,     -0.18,      0.00 } },
 { {  1,  0,  5, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0,  5,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  1, -5, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1, -4, -1 },      0.00017, {      0.00,      0.00,      0.01,      0.01,      0.07,      0.00 } },
 { {  1,  1, -4,  1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  1, -3, -1 },      0.00186, {      0.02,      0.04,      0.10,      0.11,      0.72,      0.00 } },
 { {  1,  1, -3,  1 },      0.00093, {      0.00,      0.02,      0.05,      0.06,      0.36,      0.00 } },
 { {  1,  1, -2, -3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1, -2, -1 },      0.01597, {      0.20,      0.35,      0.59,      0.96,      6.21,     -0.03 } },
 { {  1,  1, -2,  1 },      0.01680, {      0.23,      0.38,      0.61,      1.01,      6.53,     -0.03 } },
 { {  1,  1, -2,  3 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  1, -1, -3 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  1, -1, -1 },      0.00635, {     -0.38,      0.07,      0.20,      0.36,      2.46,     -0.01 } },
 { {  1,  1, -1,  1 },     -0.05590, {     -0.18,     -1.29,     -0.97,     -3.36,    -21.74,      0.11 } },
 { {  1,  1, -1,  3 },     -0.00021, {      0.00,     -0.01,      0.00,     -0.01,     -0.08,      0.00 } },
 { {  1,  1,  0, -3 },      0.00080, {      0.00,      0.05,      0.00,      0.05,      0.31,      0.00 } },
 { {  1,  1,  0, -1 },      0.80426, {      0.22,     17.53,      1.11,     48.12,    312.71,     -1.65 } },
 { {  1,  1,  0,  1 },      0.80263, {      0.34,     17.41,      0.85,     48.02,    312.08,     -1.65 } },
 { {  1,  1,  0,  3 },     -0.00083, {      0.00,     -0.06,      0.00,     -0.05,     -0.32,      0.00 } },
 { {  1,  1,  1, -3 },     -0.00021, {      0.01,     -0.01,      0.00,     -0.01,     -0.08,      0.00 } },
 { {  1,  1,  1, -1 },      0.03465, {      0.09,      0.75,      0.71,      2.07,     13.48,     -0.07 } },
 { {  1,  1,  1,  1 },      0.10176, {      0.15,      2.20,      1.97,      6.09,     39.57,     -0.21 } },
 { {  1,  1,  1,  3 },     -0.00020, {      0.00,     -0.01,      0.00,     -0.01,     -0.08,      0.00 } },
 { {  1,  1,  2, -3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  2, -1 },      0.00317, {      0.00,      0.07,      0.12,      0.19,      1.23,      0.00 } },
 { {  1,  1,  2,  1 },      0.01016, {      0.03,      0.22,      0.38,      0.61,      3.95,     -0.02 } },
 { {  1,  1,  2,  3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  1,  3, -1 },      0.00027, {      0.00,      0.00,      0.02,      0.02,      0.11,      0.00 } },
 { {  1,  1,  3,  1 },      0.00093, {      0.00,      0.02,      0.05,      0.06,      0.36,      0.00 } },
 { {  1,  1,  4, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  4,  1 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  2, -3, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  1,  2, -3,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2, -2, -1 },      0.00017, {      0.00,      0.00,      0.00,      0.02,      0.06,      0.00 } },
 { {  1,  2, -2,  1 },      0.00016, {      0.00,      0.00,      0.00,      0.02,      0.06,      0.00 } },
 { {  1,  2, -1, -1 },     -0.00008, {      0.00,      0.00,      0.00,     -0.01,     -0.03,      0.00 } },
 { {  1,  2, -1,  1 },     -0.00039, {     -0.02,      0.00,      0.00,     -0.05,     -0.15,      0.00 } },
 { {  1,  2,  0, -1 },     -0.00170, {     -0.03,     -0.04,      0.00,     -0.20,     -0.66,      0.00 } },
 { {  1,  2,  0,  1 },     -0.00204, {     -0.03,     -0.04,      0.00,     -0.24,     -0.79,      0.00 } },
 { {  1,  2,  1, -1 },     -0.00022, {      0.00,      0.00,      0.00,     -0.03,     -0.09,      0.00 } },
 { {  1,  2,  1,  1 },     -0.00072, {     -0.01,     -0.02,     -0.01,     -0.09,     -0.28,      0.00 } },
 { {  1,  2,  2, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  2,  2,  1 },     -0.00012, {      0.00,      0.00,      0.00,     -0.01,     -0.05,      0.00 } },
 { {  1,  2,  3,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  3,  0, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  1,  3,  0,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  2, -5, -1, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -5, -1,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -5,  0, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2, -5,  0,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4, -2, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4, -1, -1 },      0.00031, {      0.00,      0.00,      0.00,      0.07,      0.00,      0.00 } },
 { {  2, -4, -1,  1 },      0.00034, {      0.00,      0.00,      0.00,      0.08,      0.00,      0.00 } },
 { {  2, -4,  0, -1 },      0.00118, {      0.03,      0.03,      0.00,      0.28,      0.00,      0.00 } },
 { {  2, -4,  0,  1 },      0.00062, {      0.03,      0.01,      0.00,      0.15,      0.00,      0.00 } },
 { {  2, -4,  1, -1 },      0.00008, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2, -4,  1,  1 },      0.00010, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2, -4,  2,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -3, -1 },      0.00006, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2, -3, -2, -1 },      0.00077, {      0.02,      0.02,      0.03,      0.14,      0.00,      0.00 } },
 { {  2, -3, -2,  1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -1, -1 },      0.00914, {      0.25,      0.20,      0.17,      1.64,      0.00,      0.00 } },
 { {  2, -3, -1,  1 },      0.01042, {      0.25,      0.23,      0.19,      1.87,      0.00,      0.00 } },
 { {  2, -3, -1,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  0, -3 },      0.00008, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2, -3,  0, -1 },      0.03647, {      0.91,      0.81,      0.02,      6.54,      0.00,      0.00 } },
 { {  2, -3,  0,  1 },      0.01603, {      0.62,      0.35,      0.06,      2.88,      0.00,      0.00 } },
 { {  2, -3,  0,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  1, -3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  1, -1 },      0.00240, {      0.07,      0.05,      0.05,      0.43,      0.00,      0.00 } },
 { {  2, -3,  1,  1 },      0.00256, {      0.11,      0.06,      0.05,      0.46,      0.00,      0.00 } },
 { {  2, -3,  2, -1 },      0.00021, {      0.00,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  2, -3,  2,  1 },      0.00031, {      0.01,      0.00,      0.01,      0.06,      0.00,      0.00 } },
 { {  2, -3,  3, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  3,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -4, -1 },      0.00012, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2, -2, -4,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -3, -1 },      0.00178, {      0.03,      0.04,      0.10,      0.21,      0.00,      0.00 } },
 { {  2, -2, -3,  1 },      0.00030, {      0.00,      0.00,      0.02,      0.04,      0.00,      0.00 } },
 { {  2, -2, -2, -1 },      0.02285, {      0.53,      0.51,      0.83,      2.73,      0.00,      0.00 } },
 { {  2, -2, -2,  1 },     -0.00192, {      0.00,     -0.04,     -0.07,     -0.23,      0.00,      0.00 } },
 { {  2, -2, -2,  3 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2, -2, -1, -3 },      0.00024, {      0.00,      0.02,      0.00,      0.03,      0.00,      0.00 } },
 { {  2, -2, -1, -1 },      0.26865, {      6.32,      5.97,      4.91,     32.14,      0.00,      0.00 } },
 { {  2, -2, -1,  1 },      0.31474, {      6.78,      6.97,      5.75,     37.65,      0.00,      0.00 } },
 { {  2, -2, -1,  3 },     -0.00045, {      0.00,     -0.03,      0.00,     -0.05,      0.00,      0.00 } },
 { {  2, -2,  0, -3 },      0.00273, {      0.03,      0.18,      0.00,      0.33,      0.00,      0.00 } },
 { {  2, -2,  0, -1 },      1.08587, {     23.51,     24.25,      0.47,    129.87,      0.00,      0.00 } },
 { {  2, -2,  0,  1 },      0.38353, {     13.48,      8.45,      1.72,     45.87,      0.00,      0.00 } },
 { {  2, -2,  0,  3 },     -0.00044, {     -0.01,     -0.03,      0.00,     -0.05,      0.00,      0.00 } },
 { {  2, -2,  1, -3 },     -0.00072, {     -0.02,     -0.05,     -0.01,     -0.09,      0.00,      0.00 } },
 { {  2, -2,  1, -1 },      0.06915, {      1.69,      1.51,      1.31,      8.27,      0.00,      0.00 } },
 { {  2, -2,  1,  1 },      0.05848, {      2.24,      1.28,      1.26,      6.99,      0.00,      0.00 } },
 { {  2, -2,  1,  3 },     -0.00011, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2, -2,  2, -3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  2, -1 },      0.00555, {      0.16,      0.12,      0.21,      0.66,      0.00,      0.00 } },
 { {  2, -2,  2,  1 },      0.00677, {      0.27,      0.15,      0.27,      0.81,      0.00,      0.00 } },
 { {  2, -2,  2,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  3, -1 },      0.00046, {      0.01,      0.00,      0.03,      0.06,      0.00,      0.00 } },
 { {  2, -2,  3,  1 },      0.00070, {      0.03,      0.02,      0.04,      0.08,      0.00,      0.00 } },
 { {  2, -2,  4, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  4,  1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -6, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -5, -1 },      0.00029, {      0.00,      0.00,      0.03,      0.02,      0.00,      0.00 } },
 { {  2, -1, -5,  1 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -4, -1 },      0.00416, {      0.06,      0.09,      0.30,      0.25,      0.00,      0.00 } },
 { {  2, -1, -4,  1 },      0.00080, {      0.00,      0.02,      0.06,      0.05,      0.00,      0.00 } },
 { {  2, -1, -3, -3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -3, -1 },      0.05502, {      0.99,      1.23,      3.01,      3.29,      0.00,      0.00 } },
 { {  2, -1, -3,  1 },      0.00935, {      0.12,      0.21,      0.51,      0.56,      0.00,      0.00 } },
 { {  2, -1, -3,  3 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -2, -3 },      0.00033, {     -0.01,      0.02,      0.01,      0.02,      0.00,      0.00 } },
 { {  2, -1, -2, -1 },      0.65025, {     13.75,     14.46,     23.70,     38.91,      0.00,      0.00 } },
 { {  2, -1, -2,  1 },     -0.06208, {     -0.48,     -1.40,     -2.27,     -3.71,      0.00,      0.00 } },
 { {  2, -1, -2,  3 },     -0.00267, {     -0.03,     -0.18,     -0.10,     -0.16,      0.00,      0.00 } },
 { {  2, -1, -1, -5 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -1, -3 },      0.01034, {      0.06,      0.69,      0.17,      0.62,      0.00,      0.00 } },
 { {  2, -1, -1, -1 },      7.43488, {    154.99,    165.14,    135.90,    444.61,     -0.02,      0.00 } },
 { {  2, -1, -1,  1 },      8.86853, {    174.25,    196.33,    161.68,    530.29,     -0.01,      0.00 } },
 { {  2, -1, -1,  3 },     -0.01177, {     -0.23,     -0.79,     -0.24,     -0.70,      0.00,      0.00 } },
 { {  2, -1, -1,  5 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  0, -5 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  0, -3 },      0.08815, {      0.98,      5.90,     -0.13,      5.27,      0.00,      0.00 } },
 { {  2, -1,  0, -1 },     29.57794, {    553.39,    661.64,     12.43,   1767.42,      0.16,      0.00 } },
 { {  2, -1,  0,  1 },      7.95891, {    253.11,    175.21,     44.35,    475.60,      0.00,      0.00 } },
 { {  2, -1,  0,  3 },     -0.00944, {     -0.28,     -0.63,     -0.11,     -0.56,      0.00,      0.00 } },
 { {  2, -1,  0,  5 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  1, -5 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  1, -3 },     -0.01669, {     -0.30,     -1.12,     -0.31,     -1.00,      0.00,      0.00 } },
 { {  2, -1,  1, -1 },      1.76606, {     35.98,     38.73,     33.43,    105.50,      0.02,      0.00 } },
 { {  2, -1,  1,  1 },      1.13466, {     38.36,     24.88,     25.43,     67.78,      0.00,      0.00 } },
 { {  2, -1,  1,  3 },     -0.00224, {     -0.07,     -0.15,     -0.06,     -0.13,      0.00,      0.00 } },
 { {  2, -1,  2, -3 },     -0.00061, {     -0.01,     -0.04,     -0.02,     -0.04,      0.00,      0.00 } },
 { {  2, -1,  2, -1 },      0.12897, {      2.89,      2.78,      4.82,      7.70,      0.00,      0.00 } },
 { {  2, -1,  2,  1 },      0.12387, {      4.35,      2.71,      4.97,      7.40,      0.00,      0.00 } },
 { {  2, -1,  2,  3 },     -0.00036, {     -0.01,     -0.02,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  2, -1,  3, -3 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  3, -1 },      0.00985, {      0.24,      0.21,      0.55,      0.59,      0.00,      0.00 } },
 { {  2, -1,  3,  1 },      0.01211, {      0.44,      0.26,      0.70,      0.72,      0.00,      0.00 } },
 { {  2, -1,  3,  3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  4, -1 },      0.00076, {      0.02,      0.02,      0.06,      0.05,      0.00,      0.00 } },
 { {  2, -1,  4,  1 },      0.00111, {      0.04,      0.02,      0.08,      0.07,      0.00,      0.00 } },
 { {  2, -1,  5, -1 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  5,  1 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -7, -1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -7,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -6, -1 },      0.00092, {      0.02,      0.02,      0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0, -6,  1 },      0.00020, {      0.00,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5, -3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5, -1 },      0.01127, {      0.23,      0.25,      1.03,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5,  1 },      0.00230, {      0.03,      0.05,      0.21,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4, -3 },     -0.00014, {      0.00,      0.00,     -0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4, -1 },      0.13381, {      2.77,      2.96,      9.75,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4,  1 },      0.02496, {      0.39,      0.56,      1.82,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3, -3 },      0.00023, {     -0.02,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3, -1 },      1.51564, {     31.76,     33.59,     82.84,      0.07,      0.00,      0.00 } },
 { {  2,  0, -3,  1 },      0.25408, {      3.88,      5.75,     13.92,      0.00,      0.01,      0.00 } },
 { {  2,  0, -3,  3 },      0.00232, {      0.05,      0.16,      0.13,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3,  5 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -5 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -3 },      0.02045, {      0.10,      1.37,      0.70,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -1 },     15.56635, {    330.35,    344.99,    567.47,      0.23,      0.03,      0.00 } },
 { {  2,  0, -2,  1 },     -1.62443, {    -20.88,    -36.38,    -59.20,      0.18,     -0.02,      0.00 } },
 { {  2,  0, -2,  3 },     -0.06561, {     -0.77,     -4.41,     -2.38,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2,  5 },      0.00012, {      0.00,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -5 },     -0.00046, {      0.00,     -0.05,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -3 },      0.32907, {      3.62,     22.03,      5.64,     -0.02,      0.01,      0.00 } },
 { {  2,  0, -1, -1 },    166.57528, {   3219.29,   3695.90,   3042.80,     -5.99,      0.39,      0.00 } },
 { {  2,  0, -1,  1 },    199.48515, {   3726.28,   4413.25,   3627.39,     -8.68,      0.73,      0.00 } },
 { {  2,  0, -1,  3 },     -0.24484, {     -4.55,    -16.39,     -4.94,      0.02,      0.00,      0.00 } },
 { {  2,  0, -1,  5 },      0.00038, {      0.00,      0.04,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -5 },     -0.00221, {     -0.02,     -0.25,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -3 },      2.18637, {     23.85,    146.45,     -2.38,     -0.27,      0.03,      0.00 } },
 { {  2,  0,  0, -1 },    623.65783, {   9963.62,  13978.19,    245.16,    -70.26,      5.00,     -0.03 } },
 { {  2,  0,  0,  1 },    117.26161, {   3313.25,   2575.66,    874.52,    -12.65,      1.79,     -0.01 } },
 { {  2,  0,  0,  3 },     -0.14453, {     -3.71,     -9.64,     -2.05,      0.02,      0.00,      0.00 } },
 { {  2,  0,  0,  5 },      0.00026, {      0.00,      0.03,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -5 },     -0.00075, {      0.00,     -0.08,     -0.01,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -3 },     -0.29116, {     -3.49,    -19.63,     -5.36,      0.05,      0.00,      0.00 } },
 { {  2,  0,  1, -1 },     33.35743, {    512.77,    735.76,    630.31,     -5.21,      0.43,      0.00 } },
 { {  2,  0,  1,  1 },     15.12165, {    433.44,    331.25,    359.20,     -2.40,      0.31,      0.00 } },
 { {  2,  0,  1,  3 },     -0.03038, {     -0.82,     -2.02,     -0.85,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1,  5 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -5 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -3 },     -0.00995, {     -0.12,     -0.68,     -0.36,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -1 },      2.14618, {     33.22,     46.62,     80.12,     -0.45,      0.04,      0.00 } },
 { {  2,  0,  2,  1 },      1.51976, {     43.48,     33.21,     62.71,     -0.33,      0.04,      0.00 } },
 { {  2,  0,  2,  3 },     -0.00451, {     -0.12,     -0.30,     -0.20,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2,  5 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -3 },     -0.00093, {      0.00,     -0.06,     -0.05,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -1 },      0.14642, {      2.28,      3.13,      8.16,     -0.04,      0.00,      0.00 } },
 { {  2,  0,  3,  1 },      0.13795, {      3.92,      3.01,      8.15,     -0.04,      0.00,      0.00 } },
 { {  2,  0,  3,  3 },     -0.00056, {     -0.02,     -0.04,     -0.03,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4, -3 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4, -1 },      0.01027, {      0.16,      0.22,      0.76,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4,  1 },      0.01186, {      0.33,      0.26,      0.91,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4,  3 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5, -1 },      0.00073, {      0.01,      0.02,      0.07,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5,  1 },      0.00099, {      0.03,      0.02,      0.09,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6, -1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6,  1 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -6, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -5, -1 },      0.00028, {      0.02,      0.00,      0.03,      0.02,      0.00,      0.00 } },
 { {  2,  1, -5,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -4, -1 },      0.00247, {      0.19,      0.05,      0.18,      0.15,      0.00,      0.00 } },
 { {  2,  1, -4,  1 },      0.00034, {      0.03,      0.00,      0.03,      0.02,      0.00,      0.00 } },
 { {  2,  1, -3, -3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -3, -1 },      0.01818, {      1.67,      0.38,      0.99,      1.10,      0.00,      0.00 } },
 { {  2,  1, -3,  1 },      0.00162, {      0.17,      0.03,      0.09,      0.10,      0.00,      0.00 } },
 { {  2,  1, -2, -3 },     -0.00015, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -2, -1 },      0.07913, {     12.08,      1.57,      2.84,      4.80,      0.00,      0.00 } },
 { {  2,  1, -2,  1 },      0.05429, {      3.18,      1.16,      1.97,      3.26,      0.00,      0.00 } },
 { {  2,  1, -2,  3 },      0.00017, {     -0.03,      0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -1, -3 },     -0.00342, {      0.05,     -0.23,     -0.06,     -0.20,      0.00,      0.00 } },
 { {  2,  1, -1, -1 },     -0.79105, {     41.53,    -18.40,    -14.91,    -47.09,     -0.01,      0.00 } },
 { {  2,  1, -1,  1 },     -1.31788, {     39.71,    -29.98,    -24.58,    -78.59,      0.00,      0.00 } },
 { {  2,  1, -1,  3 },      0.00192, {     -0.03,      0.13,      0.04,      0.11,      0.00,      0.00 } },
 { {  2,  1,  0, -5 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  0, -3 },     -0.05457, {     -0.95,     -3.65,      0.01,     -3.27,      0.00,      0.00 } },
 { {  2,  1,  0, -1 },    -12.09470, {   -205.76,   -271.06,     -3.14,   -723.84,     -0.31,      0.00 } },
 { {  2,  1,  0,  1 },     -1.26433, {    -30.65,    -27.38,    -10.59,    -75.65,     -0.41,      0.00 } },
 { {  2,  1,  0,  3 },      0.00177, {      0.04,      0.12,      0.03,      0.11,      0.00,      0.00 } },
 { {  2,  1,  1, -5 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  1, -3 },      0.00465, {      0.02,      0.31,      0.09,      0.28,      0.00,      0.00 } },
 { {  2,  1,  1, -1 },     -0.82275, {    -17.02,    -18.20,    -15.41,    -49.25,     -0.04,      0.00 } },
 { {  2,  1,  1,  1 },     -0.23702, {     -7.85,     -5.10,     -5.79,    -14.19,     -0.08,      0.00 } },
 { {  2,  1,  1,  3 },      0.00052, {      0.02,      0.03,      0.02,      0.03,      0.00,      0.00 } },
 { {  2,  1,  2, -3 },      0.00021, {      0.00,      0.01,      0.00,      0.01,      0.00,      0.00 } },
 { {  2,  1,  2, -1 },     -0.06283, {     -1.47,     -1.37,     -2.34,     -3.76,      0.00,      0.00 } },
 { {  2,  1,  2,  1 },     -0.03142, {     -1.17,     -0.67,     -1.31,     -1.88,     -0.01,      0.00 } },
 { {  2,  1,  2,  3 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  3, -3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  3, -1 },     -0.00496, {     -0.13,     -0.11,     -0.28,     -0.30,      0.00,      0.00 } },
 { {  2,  1,  3,  1 },     -0.00355, {     -0.14,     -0.08,     -0.21,     -0.21,      0.00,      0.00 } },
 { {  2,  1,  3,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  4, -1 },     -0.00040, {     -0.01,      0.00,     -0.03,     -0.02,      0.00,      0.00 } },
 { {  2,  1,  4,  1 },     -0.00037, {     -0.02,      0.00,     -0.03,     -0.02,      0.00,      0.00 } },
 { {  2,  1,  5, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  5,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -4, -1 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -4,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -3, -1 },     -0.00102, {     -0.03,     -0.02,     -0.06,     -0.12,      0.00,      0.00 } },
 { {  2,  2, -3,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -2, -3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -2, -1 },     -0.01262, {     -0.47,     -0.28,     -0.46,     -1.51,      0.00,      0.00 } },
 { {  2,  2, -2,  1 },     -0.00504, {     -0.27,     -0.11,     -0.19,     -0.60,      0.00,      0.00 } },
 { {  2,  2, -2,  3 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -1, -3 },     -0.00046, {     -0.02,     -0.03,      0.00,     -0.06,      0.00,      0.00 } },
 { {  2,  2, -1, -1 },     -0.10535, {     -3.66,     -2.34,     -1.91,    -12.62,      0.00,      0.00 } },
 { {  2,  2, -1,  1 },     -0.11330, {     -3.82,     -2.50,     -2.08,    -13.58,      0.00,      0.00 } },
 { {  2,  2, -1,  3 },      0.00012, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  2,  2,  0, -3 },     -0.00103, {     -0.02,     -0.06,      0.02,     -0.12,      0.00,      0.00 } },
 { {  2,  2,  0, -1 },     -0.13415, {     -2.41,     -3.04,     -0.26,    -16.07,      0.00,      0.00 } },
 { {  2,  2,  0,  1 },     -0.01482, {     -0.52,     -0.32,     -0.53,     -1.78,      0.03,      0.00 } },
 { {  2,  2,  0,  3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  1, -3 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  1, -1 },     -0.00617, {     -0.08,     -0.14,     -0.13,     -0.74,      0.00,      0.00 } },
 { {  2,  2,  1,  1 },     -0.00145, {     -0.05,     -0.03,     -0.08,     -0.17,      0.00,      0.00 } },
 { {  2,  2,  2, -1 },     -0.00025, {      0.00,      0.00,     -0.01,     -0.03,      0.00,      0.00 } },
 { {  2,  2,  2,  1 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2,  3, -2, -1 },     -0.00006, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  2,  3, -2,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3, -1, -1 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2,  3, -1,  1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3,  0, -1 },      0.00037, {      0.01,      0.00,      0.00,      0.07,      0.00,      0.00 } },
 { {  2,  3,  0,  1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3,  1, -1 },      0.00006, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  3, -3, -1, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -3, -1,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  3, -3,  0, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -2, -3, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -2, -2, -1 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3, -2, -2,  1 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3, -2, -1, -1 },     -0.00101, {     -0.03,     -0.02,     -0.02,     -0.12,     -0.39,      0.00 } },
 { {  3, -2, -1,  1 },     -0.00063, {     -0.03,     -0.01,     -0.01,     -0.08,     -0.25,      0.00 } },
 { {  3, -2,  0, -3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -2,  0, -1 },     -0.00131, {     -0.05,     -0.03,      0.00,     -0.16,     -0.51,      0.00 } },
 { {  3, -2,  0,  1 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  3, -2,  1, -1 },     -0.00014, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3, -2,  2, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1, -4, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1, -3, -1 },     -0.00022, {      0.00,      0.00,     -0.01,     -0.01,     -0.08,      0.00 } },
 { {  3, -1, -2, -1 },     -0.00282, {     -0.07,     -0.06,     -0.10,     -0.17,     -1.10,      0.00 } },
 { {  3, -1, -2,  1 },     -0.00329, {     -0.07,     -0.07,     -0.12,     -0.20,     -1.28,      0.00 } },
 { {  3, -1, -1, -3 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  3, -1, -1, -1 },     -0.02104, {     -0.53,     -0.47,     -0.39,     -1.26,     -8.18,      0.04 } },
 { {  3, -1, -1,  1 },     -0.01356, {     -0.55,     -0.29,     -0.27,     -0.81,     -5.27,      0.03 } },
 { {  3, -1, -1,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  0, -3 },     -0.00076, {     -0.01,     -0.05,      0.00,     -0.05,     -0.30,      0.00 } },
 { {  3, -1,  0, -1 },     -0.02572, {     -0.94,     -0.60,     -0.07,     -1.53,    -10.00,      0.05 } },
 { {  3, -1,  0,  1 },     -0.00042, {     -0.19,      0.00,     -0.10,     -0.02,     -0.16,      0.00 } },
 { {  3, -1,  0,  3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  1, -1 },     -0.00275, {     -0.11,     -0.06,     -0.06,     -0.16,     -1.07,      0.00 } },
 { {  3, -1,  1,  1 },     -0.00030, {     -0.05,      0.00,     -0.02,     -0.02,     -0.12,      0.00 } },
 { {  3, -1,  2, -1 },     -0.00027, {     -0.01,      0.00,     -0.01,     -0.02,     -0.10,      0.00 } },
 { {  3, -1,  2,  1 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3, -1,  3, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  3,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -4, -1 },     -0.00016, {      0.00,      0.00,     -0.01,      0.00,     -0.06,      0.00 } },
 { {  3,  0, -4,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  3,  0, -3, -1 },     -0.00287, {     -0.04,     -0.06,     -0.16,      0.00,     -1.12,      0.00 } },
 { {  3,  0, -3,  1 },      0.00015, {      0.00,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  3,  0, -3,  3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -2, -3 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3,  0, -2, -1 },     -0.03941, {     -0.88,     -0.87,     -1.44,      0.00,    -15.32,      0.08 } },
 { {  3,  0, -2,  1 },     -0.04852, {     -1.04,     -1.06,     -1.77,      0.00,    -18.87,      0.10 } },
 { {  3,  0, -2,  3 },      0.00016, {      0.00,      0.01,      0.00,      0.00,      0.06,      0.00 } },
 { {  3,  0, -1, -3 },     -0.00128, {     -0.01,     -0.09,     -0.02,      0.00,     -0.50,      0.00 } },
 { {  3,  0, -1, -1 },     -0.30517, {     -7.43,     -6.74,     -5.66,      0.08,   -118.66,      0.63 } },
 { {  3,  0, -1,  1 },     -0.20593, {     -7.88,     -4.39,     -4.05,      0.12,    -80.07,      0.42 } },
 { {  3,  0, -1,  3 },      0.00038, {      0.02,      0.02,      0.00,      0.00,      0.15,      0.00 } },
 { {  3,  0,  0, -3 },     -0.01009, {     -0.13,     -0.68,      0.00,      0.00,     -3.92,      0.02 } },
 { {  3,  0,  0, -1 },     -0.35183, {    -11.87,     -8.15,     -0.99,      0.22,   -136.80,      0.72 } },
 { {  3,  0,  0,  1 },     -0.02840, {     -2.77,     -0.49,     -1.37,      0.10,    -11.04,      0.06 } },
 { {  3,  0,  0,  3 },      0.00016, {      0.00,      0.01,      0.00,      0.00,      0.06,      0.00 } },
 { {  3,  0,  1, -3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  1, -1 },     -0.03611, {     -1.26,     -0.83,     -0.75,      0.03,    -14.04,      0.07 } },
 { {  3,  0,  1,  1 },     -0.00658, {     -0.56,     -0.12,     -0.29,      0.02,     -2.56,      0.01 } },
 { {  3,  0,  1,  3 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  0,  2, -1 },     -0.00322, {     -0.12,     -0.07,     -0.13,      0.00,     -1.25,      0.00 } },
 { {  3,  0,  2,  1 },     -0.00108, {     -0.08,     -0.02,     -0.06,      0.00,     -0.42,      0.00 } },
 { {  3,  0,  3, -1 },     -0.00027, {     -0.01,      0.00,     -0.02,      0.00,     -0.11,      0.00 } },
 { {  3,  0,  3,  1 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  3,  0,  4, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  4,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -4, -1 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3,  1, -4,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -3, -1 },     -0.00044, {     -0.02,      0.00,     -0.02,     -0.03,     -0.17,      0.00 } },
 { {  3,  1, -3,  1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  3,  1, -2, -1 },     -0.00152, {     -0.09,     -0.03,     -0.06,     -0.09,     -0.59,      0.00 } },
 { {  3,  1, -2,  1 },     -0.00261, {     -0.12,     -0.06,     -0.09,     -0.16,     -1.02,      0.00 } },
 { {  3,  1, -1, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.01,      0.00 } },
 { {  3,  1, -1, -1 },      0.01321, {      0.23,      0.29,      0.25,      0.79,      5.14,     -0.03 } },
 { {  3,  1, -1,  1 },      0.02083, {      0.37,      0.45,      0.39,      1.24,      8.10,     -0.04 } },
 { {  3,  1, -1,  3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  1,  0, -3 },      0.00031, {      0.00,      0.02,      0.00,      0.02,      0.12,      0.00 } },
 { {  3,  1,  0, -1 },      0.03436, {      0.68,      0.77,      0.09,      2.05,     13.36,     -0.07 } },
 { {  3,  1,  0,  1 },      0.01351, {      0.39,      0.29,      0.17,      0.81,      5.25,     -0.03 } },
 { {  3,  1,  0,  3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  3,  1,  1, -1 },      0.00426, {      0.09,      0.09,      0.09,      0.25,      1.66,      0.00 } },
 { {  3,  1,  1,  1 },      0.00258, {      0.08,      0.05,      0.07,      0.15,      1.00,      0.00 } },
 { {  3,  1,  2, -1 },      0.00044, {      0.00,      0.00,      0.02,      0.03,      0.17,      0.00 } },
 { {  3,  1,  2,  1 },      0.00035, {      0.01,      0.00,      0.02,      0.02,      0.14,      0.00 } },
 { {  3,  1,  3, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  1,  3,  1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  2, -2, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2, -2,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  2, -1, -1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  2, -1,  1 },     -0.00010, {      0.00,      0.00,      0.00,     -0.01,     -0.04,      0.00 } },
 { {  3,  2,  0, -1 },     -0.00056, {      0.00,     -0.01,      0.00,     -0.07,     -0.22,      0.00 } },
 { {  3,  2,  0,  1 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3,  2,  1, -1 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,     -0.03,      0.00 } },
 { {  3,  2,  1,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  2,  2, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  3, -1, -3 },      0.00001, {      0.00,      0.00,     -0.02,     -0.01,      0.03,      0.00 } },
 { {  3,  3, -1, -1 },      0.00000, {     -0.01,      0.00,     -0.01,     -0.02,      0.01,      0.00 } },
 { {  3,  3, -1,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -4, -2,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -4, -1, -1 },      0.00007, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4, -4, -1,  1 },      0.00007, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4, -4,  0, -1 },      0.00008, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4, -4,  0,  1 },      0.00004, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  4, -4,  1, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -2, -1 },      0.00013, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4, -3, -2,  1 },      0.00059, {      0.03,      0.01,      0.02,      0.11,      0.00,      0.00 } },
 { {  4, -3, -1, -1 },      0.00165, {      0.07,      0.04,      0.03,      0.30,      0.00,      0.00 } },
 { {  4, -3, -1,  1 },      0.00134, {      0.07,      0.03,      0.03,      0.24,      0.00,      0.00 } },
 { {  4, -3,  0, -3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3,  0, -1 },      0.00161, {      0.08,      0.04,      0.00,      0.29,      0.00,      0.00 } },
 { {  4, -3,  0,  1 },      0.00080, {      0.05,      0.02,      0.01,      0.14,      0.00,      0.00 } },
 { {  4, -3,  1, -1 },      0.00024, {      0.01,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  4, -3,  1,  1 },      0.00017, {      0.01,      0.00,      0.00,      0.03,      0.00,      0.00 } },
 { {  4, -3,  2, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3,  2,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -4, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -3, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -3,  1 },      0.00011, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  4, -2, -2, -3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -2, -1 },      0.00283, {      0.11,      0.06,      0.10,      0.34,      0.00,      0.00 } },
 { {  4, -2, -2,  1 },      0.01230, {      0.50,      0.27,      0.45,      1.47,      0.00,      0.00 } },
 { {  4, -2, -2,  3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -1, -3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -1, -1 },      0.03462, {      1.29,      0.77,      0.65,      4.14,      0.00,      0.00 } },
 { {  4, -2, -1,  1 },      0.02380, {      1.25,      0.52,      0.53,      2.85,      0.00,      0.00 } },
 { {  4, -2, -1,  3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  0, -3 },      0.00029, {      0.00,      0.02,      0.00,      0.04,      0.00,      0.00 } },
 { {  4, -2,  0, -1 },      0.02899, {      1.41,      0.65,      0.18,      3.46,      0.00,      0.00 } },
 { {  4, -2,  0,  1 },      0.01270, {      0.79,      0.27,      0.20,      1.52,      0.00,      0.00 } },
 { {  4, -2,  0,  3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  1, -3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  1, -1 },      0.00417, {      0.21,      0.09,      0.10,      0.50,      0.00,      0.00 } },
 { {  4, -2,  1,  1 },      0.00255, {      0.17,      0.06,      0.08,      0.30,      0.00,      0.00 } },
 { {  4, -2,  2, -1 },      0.00047, {      0.03,      0.01,      0.02,      0.06,      0.00,      0.00 } },
 { {  4, -2,  2,  1 },      0.00037, {      0.03,      0.00,      0.02,      0.04,      0.00,      0.00 } },
 { {  4, -2,  3, -1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  3,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -6, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -5, -1 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -4, -1 },     -0.00058, {     -0.02,     -0.01,     -0.04,     -0.03,      0.00,      0.00 } },
 { {  4, -1, -4,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -3, -3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -3, -1 },     -0.00090, {     -0.03,     -0.02,     -0.05,     -0.05,      0.00,      0.00 } },
 { {  4, -1, -3,  1 },      0.00191, {      0.12,      0.04,      0.10,      0.11,      0.00,      0.00 } },
 { {  4, -1, -3,  3 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -2, -3 },     -0.00033, {     -0.01,     -0.02,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  4, -1, -2, -1 },      0.05251, {      1.99,      1.16,      1.92,      3.14,      0.00,      0.00 } },
 { {  4, -1, -2,  1 },      0.21376, {      8.24,      4.68,      7.80,     12.77,      0.00,      0.00 } },
 { {  4, -1, -2,  3 },     -0.00060, {     -0.02,     -0.04,     -0.02,     -0.04,      0.00,      0.00 } },
 { {  4, -1, -1, -3 },     -0.00065, {     -0.04,     -0.04,     -0.01,     -0.04,      0.00,      0.00 } },
 { {  4, -1, -1, -1 },      0.59580, {     21.19,     13.27,     11.16,     35.53,      0.02,      0.00 } },
 { {  4, -1, -1,  1 },      0.33882, {     16.92,      7.36,      7.74,     20.21,      0.00,      0.00 } },
 { {  4, -1, -1,  3 },     -0.00075, {     -0.04,     -0.05,     -0.02,     -0.04,      0.00,      0.00 } },
 { {  4, -1,  0, -5 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  0, -3 },      0.00534, {      0.13,      0.36,      0.00,      0.32,      0.00,      0.00 } },
 { {  4, -1,  0, -1 },      0.41496, {     19.00,      9.23,      2.87,     24.71,      0.00,      0.00 } },
 { {  4, -1,  0,  1 },      0.15791, {      9.24,      3.41,      2.73,      9.41,      0.00,      0.00 } },
 { {  4, -1,  0,  3 },     -0.00037, {     -0.02,     -0.02,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -1,  1, -3 },     -0.00062, {     -0.02,     -0.04,     -0.01,     -0.04,      0.00,      0.00 } },
 { {  4, -1,  1, -1 },      0.05686, {      2.69,      1.24,      1.35,      3.38,      0.00,      0.00 } },
 { {  4, -1,  1,  1 },      0.03009, {      1.83,      0.65,      0.93,      1.79,      0.00,      0.00 } },
 { {  4, -1,  1,  3 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  2, -3 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  2, -1 },      0.00614, {      0.30,      0.13,      0.25,      0.37,      0.00,      0.00 } },
 { {  4, -1,  2,  1 },      0.00421, {      0.26,      0.09,      0.20,      0.25,      0.00,      0.00 } },
 { {  4, -1,  2,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  3, -3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  3, -1 },      0.00060, {      0.03,      0.01,      0.04,      0.04,      0.00,      0.00 } },
 { {  4, -1,  3,  1 },      0.00050, {      0.03,      0.01,      0.03,      0.03,      0.00,      0.00 } },
 { {  4, -1,  4, -1 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  4,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -7, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -6, -1 },     -0.00015, {      0.00,      0.00,     -0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0, -6,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -5, -1 },     -0.00116, {     -0.05,     -0.03,     -0.11,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4, -3 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4, -1 },     -0.00680, {     -0.28,     -0.15,     -0.50,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  1 },      0.00029, {      0.00,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3, -3 },     -0.00066, {     -0.02,     -0.04,     -0.04,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3, -1 },     -0.00771, {     -0.29,     -0.17,     -0.42,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3,  1 },      0.02174, {      1.30,      0.47,      1.19,      0.02,      0.00,      0.00 } },
 { {  4,  0, -3,  3 },     -0.00127, {     -0.04,     -0.09,     -0.07,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -3 },     -0.00390, {     -0.14,     -0.26,     -0.14,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -1 },      0.63371, {     23.67,     13.99,     23.16,     -0.03,      0.08,      0.00 } },
 { {  4,  0, -2,  1 },      2.41389, {     90.53,     52.84,     87.95,     -0.35,      0.07,      0.00 } },
 { {  4,  0, -2,  3 },     -0.00632, {     -0.24,     -0.42,     -0.24,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2,  5 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -5 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -3 },     -0.00554, {     -0.34,     -0.37,     -0.10,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -1 },      6.57962, {    223.57,    146.75,    123.07,     -1.94,      0.26,      0.00 } },
 { {  4,  0, -1,  1 },      2.99850, {    142.57,     65.02,     71.12,     -0.78,      0.06,      0.00 } },
 { {  4,  0, -1,  3 },     -0.00664, {     -0.31,     -0.44,     -0.19,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1,  5 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -5 },     -0.00013, {      0.00,     -0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -3 },      0.06257, {      1.51,      4.19,     -0.04,     -0.03,      0.02,      0.00 } },
 { {  4,  0,  0, -1 },      3.67449, {    157.55,     81.63,     29.57,     -1.61,      0.06,      0.00 } },
 { {  4,  0,  0,  1 },      1.19188, {     65.39,     25.66,     23.11,     -0.45,      0.08,      0.00 } },
 { {  4,  0,  0,  3 },     -0.00288, {     -0.15,     -0.19,     -0.08,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -3 },     -0.00552, {     -0.15,     -0.38,     -0.11,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -1 },      0.47338, {     20.49,     10.32,     11.59,     -0.25,      0.02,      0.00 } },
 { {  4,  0,  1,  1 },      0.21259, {     11.97,      4.56,      6.92,     -0.10,      0.01,      0.00 } },
 { {  4,  0,  1,  3 },     -0.00068, {     -0.04,     -0.04,     -0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2, -3 },     -0.00090, {     -0.02,     -0.06,     -0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2, -1 },      0.04834, {      2.10,      1.03,      2.04,     -0.03,      0.00,      0.00 } },
 { {  4,  0,  2,  1 },      0.02828, {      1.60,      0.60,      1.37,     -0.02,      0.00,      0.00 } },
 { {  4,  0,  2,  3 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3, -3 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3, -1 },      0.00449, {      0.20,      0.09,      0.27,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3,  1 },      0.00323, {      0.18,      0.07,      0.21,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4, -1 },      0.00040, {      0.02,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4,  1 },      0.00034, {      0.02,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0,  5, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  5,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -5, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -4, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -4,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -3, -1 },      0.00063, {      0.05,      0.01,      0.03,      0.04,      0.00,      0.00 } },
 { {  4,  1, -3,  1 },      0.00192, {      0.19,      0.04,      0.10,      0.11,      0.00,      0.00 } },
 { {  4,  1, -2, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -2, -1 },     -0.00203, {      0.57,     -0.05,     -0.08,     -0.12,      0.00,      0.00 } },
 { {  4,  1, -2,  1 },     -0.02957, {      0.54,     -0.67,     -1.08,     -1.78,      0.00,      0.00 } },
 { {  4,  1, -2,  3 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -1, -3 },     -0.00015, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -1, -1 },     -0.17191, {     -3.93,     -3.89,     -3.20,    -10.30,     -0.02,      0.00 } },
 { {  4,  1, -1,  1 },     -0.05097, {     -1.36,     -1.11,     -1.24,     -3.06,     -0.02,      0.00 } },
 { {  4,  1, -1,  3 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  0, -3 },     -0.00304, {     -0.09,     -0.20,      0.00,     -0.18,      0.00,      0.00 } },
 { {  4,  1,  0, -1 },     -0.11308, {     -4.83,     -2.51,     -0.93,     -6.76,     -0.02,      0.00 } },
 { {  4,  1,  0,  1 },     -0.02549, {     -1.26,     -0.54,     -0.52,     -1.53,      0.00,      0.00 } },
 { {  4,  1,  0,  3 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1, -3 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1, -1 },     -0.01692, {     -0.79,     -0.37,     -0.42,     -1.01,      0.00,      0.00 } },
 { {  4,  1,  1,  1 },     -0.00558, {     -0.32,     -0.12,     -0.19,     -0.33,      0.00,      0.00 } },
 { {  4,  1,  1,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  2, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  2, -1 },     -0.00196, {     -0.10,     -0.04,     -0.08,     -0.12,      0.00,      0.00 } },
 { {  4,  1,  2,  1 },     -0.00088, {     -0.05,     -0.02,     -0.04,     -0.05,      0.00,      0.00 } },
 { {  4,  1,  3, -1 },     -0.00020, {     -0.01,      0.00,     -0.01,     -0.01,      0.00,      0.00 } },
 { {  4,  1,  3,  1 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  4, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  4,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -3, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -3,  1 },     -0.00020, {     -0.01,      0.00,     -0.01,     -0.02,      0.00,      0.00 } },
 { {  4,  2, -2, -3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -2, -1 },     -0.00114, {     -0.07,     -0.02,     -0.04,     -0.14,      0.00,      0.00 } },
 { {  4,  2, -2,  1 },     -0.00277, {     -0.16,     -0.06,     -0.10,     -0.33,      0.00,      0.00 } },
 { {  4,  2, -1, -3 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -1, -1 },     -0.00438, {     -0.25,     -0.09,     -0.08,     -0.52,      0.00,      0.00 } },
 { {  4,  2, -1,  1 },     -0.00179, {     -0.12,     -0.04,     -0.05,     -0.21,      0.00,      0.00 } },
 { {  4,  2,  0, -1 },     -0.00056, {     -0.03,     -0.01,     -0.02,     -0.07,      0.00,      0.00 } },
 { {  4,  2,  0,  1 },     -0.00025, {     -0.02,      0.00,     -0.01,     -0.03,      0.00,      0.00 } },
 { {  4,  2,  1, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2,  1,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  3, -2,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  3, -1, -1 },      0.00008, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4,  3, -1,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  3,  0, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2, -2, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5, -2, -2,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5, -2, -1, -1 },     -0.00010, {      0.00,      0.00,      0.00,     -0.01,     -0.04,      0.00 } },
 { {  5, -2, -1,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2,  0, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -3, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5, -1, -3,  1 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  5, -1, -2, -1 },     -0.00057, {     -0.02,     -0.01,     -0.02,     -0.03,     -0.22,      0.00 } },
 { {  5, -1, -2,  1 },     -0.00061, {     -0.04,     -0.01,     -0.02,     -0.04,     -0.24,      0.00 } },
 { {  5, -1, -1, -3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -1, -1 },     -0.00127, {     -0.07,     -0.03,     -0.03,     -0.08,     -0.50,      0.00 } },
 { {  5, -1, -1,  1 },     -0.00029, {     -0.03,      0.00,     -0.01,     -0.02,     -0.11,      0.00 } },
 { {  5, -1,  0, -3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5, -1,  0, -1 },     -0.00032, {     -0.03,      0.00,      0.00,     -0.02,     -0.12,      0.00 } },
 { {  5, -1,  1, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  0, -4,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0, -3, -1 },     -0.00033, {     -0.01,      0.00,     -0.02,      0.00,     -0.13,      0.00 } },
 { {  5,  0, -3,  1 },     -0.00124, {     -0.05,     -0.03,     -0.07,      0.00,     -0.48,      0.00 } },
 { {  5,  0, -2, -1 },     -0.00505, {     -0.21,     -0.11,     -0.19,      0.00,     -1.96,      0.01 } },
 { {  5,  0, -2,  1 },     -0.00519, {     -0.29,     -0.11,     -0.20,      0.00,     -2.02,      0.01 } },
 { {  5,  0, -2,  3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0, -1, -3 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  5,  0, -1, -1 },     -0.01049, {     -0.54,     -0.24,     -0.22,      0.01,     -4.08,      0.02 } },
 { {  5,  0, -1,  1 },     -0.00245, {     -0.21,     -0.05,     -0.09,      0.00,     -0.95,      0.00 } },
 { {  5,  0, -1,  3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0,  0, -3 },     -0.00041, {     -0.01,     -0.03,      0.00,      0.00,     -0.16,      0.00 } },
 { {  5,  0,  0, -1 },     -0.00254, {     -0.19,     -0.06,     -0.07,      0.00,     -0.99,      0.00 } },
 { {  5,  0,  0,  1 },     -0.00027, {     -0.05,      0.00,     -0.03,      0.00,     -0.10,      0.00 } },
 { {  5,  0,  1, -3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  5,  0,  1, -1 },     -0.00045, {     -0.03,     -0.01,     -0.02,      0.00,     -0.17,      0.00 } },
 { {  5,  0,  1,  1 },     -0.00006, {     -0.01,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  0,  2, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  0,  2,  1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1, -3, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1, -3,  1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  1, -2, -1 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  5,  1, -2,  1 },      0.00041, {      0.01,      0.00,      0.02,      0.02,      0.16,      0.00 } },
 { {  5,  1, -1, -1 },      0.00097, {      0.04,      0.02,      0.02,      0.06,      0.38,      0.00 } },
 { {  5,  1, -1,  1 },      0.00053, {      0.03,      0.01,      0.01,      0.03,      0.21,      0.00 } },
 { {  5,  1,  0, -3 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1,  0, -1 },      0.00051, {      0.02,      0.01,      0.00,      0.03,      0.20,      0.00 } },
 { {  5,  1,  0,  1 },      0.00023, {      0.01,      0.00,      0.00,      0.01,      0.09,      0.00 } },
 { {  5,  1,  1, -1 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  5,  1,  1,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  5,  1,  2, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  2, -1, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  2,  0, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3, -3,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3, -2, -1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3, -2,  1 },      0.00006, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  6, -3, -1, -1 },      0.00010, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  6, -3, -1,  1 },      0.00006, {      0.00,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  6, -3,  0, -1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3,  0,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3,  1, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -3, -1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -3,  1 },      0.00033, {      0.02,      0.00,      0.02,      0.04,      0.00,      0.00 } },
 { {  6, -2, -2, -1 },      0.00088, {      0.05,      0.02,      0.03,      0.10,      0.00,      0.00 } },
 { {  6, -2, -2,  1 },      0.00087, {      0.06,      0.02,      0.04,      0.10,      0.00,      0.00 } },
 { {  6, -2, -1, -3 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -1, -1 },      0.00137, {      0.09,      0.03,      0.03,      0.16,      0.00,      0.00 } },
 { {  6, -2, -1,  1 },      0.00078, {      0.06,      0.02,      0.02,      0.09,      0.00,      0.00 } },
 { {  6, -2,  0, -1 },      0.00067, {      0.05,      0.01,      0.01,      0.08,      0.00,      0.00 } },
 { {  6, -2,  0,  1 },      0.00032, {      0.03,      0.00,      0.00,      0.04,      0.00,      0.00 } },
 { {  6, -2,  1, -1 },      0.00013, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  6, -2,  1,  1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  2, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  2,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -4, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -4,  1 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -3, -1 },      0.00089, {      0.05,      0.02,      0.05,      0.05,      0.00,      0.00 } },
 { {  6, -1, -3,  1 },      0.00415, {      0.24,      0.09,      0.23,      0.25,      0.00,      0.00 } },
 { {  6, -1, -3,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -2, -1 },      0.01091, {      0.59,      0.24,      0.40,      0.65,      0.00,      0.00 } },
 { {  6, -1, -2,  1 },      0.00939, {      0.64,      0.20,      0.38,      0.56,      0.00,      0.00 } },
 { {  6, -1, -2,  3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -1, -3 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -1, -1 },      0.01486, {      0.96,      0.33,      0.35,      0.88,      0.00,      0.00 } },
 { {  6, -1, -1,  1 },      0.00757, {      0.59,      0.16,      0.24,      0.45,      0.00,      0.00 } },
 { {  6, -1, -1,  3 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  0, -3 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  0, -1 },      0.00652, {      0.47,      0.14,      0.12,      0.39,      0.00,      0.00 } },
 { {  6, -1,  0,  1 },      0.00288, {      0.24,      0.06,      0.09,      0.17,      0.00,      0.00 } },
 { {  6, -1,  0,  3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  1, -1 },      0.00122, {      0.09,      0.03,      0.04,      0.07,      0.00,      0.00 } },
 { {  6, -1,  1,  1 },      0.00063, {      0.06,      0.01,      0.03,      0.04,      0.00,      0.00 } },
 { {  6, -1,  2, -1 },      0.00017, {      0.01,      0.00,      0.00,      0.01,      0.00,      0.00 } },
 { {  6, -1,  2,  1 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  3, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1,  3,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4, -1 },      0.00014, {      0.00,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  1 },      0.00072, {      0.05,      0.02,      0.05,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3, -1 },      0.00702, {      0.40,      0.15,      0.38,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3,  1 },      0.03118, {      1.77,      0.68,      1.71,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3,  3 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2, -3 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2, -1 },      0.08096, {      4.28,      1.79,      3.00,     -0.04,      0.00,      0.00 } },
 { {  6,  0, -2,  1 },      0.05963, {      3.97,      1.28,      2.46,     -0.03,      0.00,      0.00 } },
 { {  6,  0, -2,  3 },     -0.00021, {     -0.01,     -0.01,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1, -3 },      0.00107, {      0.04,      0.07,      0.02,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1, -1 },      0.09403, {      5.84,      2.08,      2.28,     -0.07,      0.00,      0.00 } },
 { {  6,  0, -1,  1 },      0.04217, {      3.17,      0.89,      1.37,     -0.03,      0.00,      0.00 } },
 { {  6,  0, -1,  3 },     -0.00014, {     -0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -3 },      0.00060, {      0.04,      0.04,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -1 },      0.03674, {      2.54,      0.80,      0.73,     -0.03,      0.00,      0.00 } },
 { {  6,  0,  0,  1 },      0.01465, {      1.19,      0.31,      0.47,     -0.01,      0.00,      0.00 } },
 { {  6,  0,  0,  3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1, -3 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1, -1 },      0.00654, {      0.46,      0.14,      0.22,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1,  1 },      0.00305, {      0.25,      0.06,      0.13,      0.00,      0.00,      0.00 } },
 { {  6,  0,  1,  3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2, -3 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2, -1 },      0.00088, {      0.06,      0.02,      0.04,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2,  1 },      0.00048, {      0.04,      0.01,      0.03,      0.00,      0.00,      0.00 } },
 { {  6,  0,  3, -1 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  3,  1 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  4, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -4,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -3, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -3,  1 },     -0.00054, {      0.00,     -0.01,     -0.03,     -0.03,      0.00,      0.00 } },
 { {  6,  1, -2, -1 },     -0.00256, {     -0.08,     -0.06,     -0.09,     -0.15,      0.00,      0.00 } },
 { {  6,  1, -2,  1 },     -0.00137, {     -0.05,     -0.03,     -0.06,     -0.08,      0.00,      0.00 } },
 { {  6,  1, -1, -3 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -1, -1 },     -0.00350, {     -0.19,     -0.08,     -0.08,     -0.21,      0.00,      0.00 } },
 { {  6,  1, -1,  1 },     -0.00116, {     -0.07,     -0.02,     -0.04,     -0.07,      0.00,      0.00 } },
 { {  6,  1,  0, -3 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  0, -1 },     -0.00153, {     -0.10,     -0.03,     -0.03,     -0.09,      0.00,      0.00 } },
 { {  6,  1,  0,  1 },     -0.00047, {     -0.04,      0.00,     -0.02,     -0.03,      0.00,      0.00 } },
 { {  6,  1,  1, -1 },     -0.00031, {     -0.02,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  6,  1,  1,  1 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  2, -1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  2,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -3, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -3,  1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -2, -1 },     -0.00009, {      0.00,      0.00,      0.00,     -0.01,      0.00,      0.00 } },
 { {  6,  2, -2,  1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -1, -1 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -1,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -3, -1 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -3,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -2, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  7, -1, -2,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -1, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -4,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -3, -1 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  7,  0, -3,  1 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  7,  0, -2, -1 },     -0.00024, {     -0.02,      0.00,      0.00,      0.00,     -0.09,      0.00 } },
 { {  7,  0, -2,  1 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  7,  0, -1, -3 },     -0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -1, -1 },     -0.00013, {     -0.01,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  7,  0, -1,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  7,  0,  0, -1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,     -0.01,      0.00 } },
 { {  7,  1, -2, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  1, -2,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  1, -1, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  1, -1,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -3, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -3,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -2, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -2,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -1, -1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -1,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2,  0, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -4, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -4,  1 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -3, -1 },      0.00019, {      0.01,      0.00,      0.01,      0.01,      0.00,      0.00 } },
 { {  8, -1, -3,  1 },      0.00022, {      0.02,      0.00,      0.01,      0.01,      0.00,      0.00 } },
 { {  8, -1, -2, -1 },      0.00038, {      0.03,      0.00,      0.02,      0.02,      0.00,      0.00 } },
 { {  8, -1, -2,  1 },      0.00025, {      0.02,      0.00,      0.01,      0.01,      0.00,      0.00 } },
 { {  8, -1, -1, -1 },      0.00029, {      0.03,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  8, -1, -1,  1 },      0.00015, {      0.02,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  0, -1 },      0.00011, {      0.01,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  0,  1 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  1, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  1,  1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -5,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -4, -1 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -4,  1 },      0.00042, {      0.03,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3, -1 },      0.00107, {      0.08,      0.02,      0.06,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3,  1 },      0.00108, {      0.09,      0.02,      0.06,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2, -3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2, -1 },      0.00189, {      0.15,      0.04,      0.08,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2,  1 },      0.00111, {      0.11,      0.02,      0.05,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1, -3 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1, -1 },      0.00130, {      0.12,      0.03,      0.04,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1,  1 },      0.00060, {      0.06,      0.01,      0.03,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0, -1 },      0.00045, {      0.04,      0.00,      0.01,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0,  1 },      0.00020, {      0.02,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  1, -1 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  1,  1 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  2, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -3, -1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -3,  1 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -2, -1 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -2,  1 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -1, -1 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -1,  1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1,  0, -1 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -4, -1 },      0.00001, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -4,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -3, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -3,  1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -2, -1 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -2,  1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -1, -1 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } }
}  };

constexpr array<ELP2000MainProblemCoefficient, 704> g_ELP3
{ {
 { {  0,  0,  0,  0 }, 385000.52719, {  -7992.63,    -11.06,  21578.08,     -4.53,     11.39,     -0.06 } },
 { {  0,  0,  0,  2 },     -3.14837, {   -204.48,   -138.94,    159.64,     -0.39,      0.12,      0.00 } },
 { {  0,  0,  0,  4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -4 },      0.00038, {      0.03,      0.03,     -0.03,      0.00,      0.00,      0.00 } },
 { {  0,  0,  1, -2 },     79.66183, {   -359.45,   3583.79,   1454.02,     -2.37,      0.85,      0.00 } },
 { {  0,  0,  1,  0 }, -20905.32206, {   6888.23,    -35.83, -380331.74,     22.31,      1.77,      0.00 } },
 { {  0,  0,  1,  2 },     -0.10326, {     -9.14,     -4.53,      8.08,      0.00,      0.02,      0.00 } },
 { {  0,  0,  2, -4 },     -0.00837, {      0.07,     -0.75,     -0.31,      0.00,      0.00,      0.00 } },
 { {  0,  0,  2, -2 },     -4.42124, {     18.29,   -198.91,   -161.90,      0.14,     -0.04,      0.00 } },
 { {  0,  0,  2,  0 },   -569.92332, {    374.44,     -1.99, -20737.33,      5.79,      0.44,      0.00 } },
 { {  0,  0,  2,  2 },     -0.00102, {     -0.37,     -0.04,      0.62,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -4 },      0.00033, {      0.00,      0.03,      0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3, -2 },     -0.00111, {      0.00,      0.00,     -0.05,      0.00,      0.00,      0.00 } },
 { {  0,  0,  3,  0 },    -23.21032, {     26.49,     -0.14,  -1266.64,      0.63,      0.05,      0.00 } },
 { {  0,  0,  3,  2 },      0.00026, {      0.00,      0.00,      0.06,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4, -2 },      0.00027, {      0.00,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  0,  0,  4,  0 },     -1.11693, {      1.98,      0.00,    -81.26,      0.06,      0.00,      0.00 } },
 { {  0,  0,  4,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  0,  5,  0 },     -0.05888, {      0.15,      0.00,     -5.35,      0.00,      0.00,      0.00 } },
 { {  0,  0,  6,  0 },     -0.00329, {      0.00,      0.00,     -0.36,      0.00,      0.00,      0.00 } },
 { {  0,  0,  7,  0 },     -0.00019, {      0.00,      0.00,     -0.02,      0.00,      0.00,      0.00 } },
 { {  0,  1, -6,  0 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -5,  0 },     -0.00174, {     -0.03,      0.00,     -0.16,     -0.10,      0.00,      0.00 } },
 { {  0,  1, -4,  0 },     -0.02674, {     -0.56,      0.00,     -1.94,     -1.60,      0.00,      0.00 } },
 { {  0,  1, -3,  0 },     -0.42242, {     -9.35,      0.20,    -23.04,    -25.30,      0.00,      0.00 } },
 { {  0,  1, -3,  2 },      0.00015, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -2, -2 },     -0.00019, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -2,  0 },     -7.00293, {   -161.63,      3.29,   -254.78,   -419.55,     -0.12,      0.00 } },
 { {  0,  1, -2,  2 },     -0.05020, {     -1.17,     -2.24,     -1.83,     -3.01,      0.00,      0.00 } },
 { {  0,  1, -2,  4 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1, -1, -2 },     -0.00769, {     -0.15,     -0.34,     -0.04,     -0.46,      0.00,      0.00 } },
 { {  0,  1, -1,  0 },   -129.62476, {  -3122.77,     61.80,  -2362.46,  -7767.84,     -0.74,      0.00 } },
 { {  0,  1, -1,  2 },      0.33465, {      7.26,     14.91,      5.99,     20.06,      0.00,      0.00 } },
 { {  0,  1, -1,  4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  0, -2 },     -0.18568, {     -4.48,     -8.29,      0.54,    -11.12,      0.00,      0.00 } },
 { {  0,  1,  0,  0 },     48.89010, {   1082.40,    -25.99,    156.78,   2926.50,     -3.76,      0.02 } },
 { {  0,  1,  0,  2 },     -0.15803, {     -3.35,     -7.06,     -0.08,     -9.46,     -0.03,      0.00 } },
 { {  0,  1,  1, -4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  1, -2 },     -0.24811, {     -2.72,    -11.08,     -4.50,    -14.87,      0.00,      0.00 } },
 { {  0,  1,  1,  0 },    104.75896, {   2165.07,    -47.62,   1929.10,   6279.07,     -0.60,      0.00 } },
 { {  0,  1,  1,  2 },     -0.00682, {     -0.05,     -0.31,     -0.17,     -0.41,      0.00,      0.00 } },
 { {  0,  1,  2, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  2, -2 },      0.03265, {      0.36,      1.45,      1.20,      1.96,      0.00,      0.00 } },
 { {  0,  1,  2,  0 },      5.75105, {    119.47,     -2.62,    210.76,    344.76,     -0.09,      0.00 } },
 { {  0,  1,  2,  2 },     -0.00030, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  3, -2 },     -0.00080, {     -0.04,     -0.04,     -0.04,     -0.05,      0.00,      0.00 } },
 { {  0,  1,  3,  0 },      0.35509, {      7.33,     -0.16,     19.48,     21.29,      0.00,      0.00 } },
 { {  0,  1,  4, -2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  1,  4,  0 },      0.02304, {      0.47,      0.00,      1.68,      1.38,      0.00,      0.00 } },
 { {  0,  1,  5,  0 },      0.00153, {      0.03,      0.00,      0.14,      0.09,      0.00,      0.00 } },
 { {  0,  1,  6,  0 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -5,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2, -4,  0 },     -0.00066, {     -0.02,      0.00,     -0.05,     -0.08,      0.00,      0.00 } },
 { {  0,  2, -3,  0 },     -0.00938, {     -0.29,      0.00,     -0.51,     -1.12,      0.00,      0.00 } },
 { {  0,  2, -2,  0 },     -0.13618, {     -4.10,      0.09,     -4.96,    -16.31,      0.00,      0.00 } },
 { {  0,  2, -2,  2 },     -0.00099, {     -0.03,     -0.04,     -0.04,     -0.12,      0.00,      0.00 } },
 { {  0,  2, -1, -2 },     -0.00031, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  0,  2, -1,  0 },     -2.11728, {    -59.78,      1.22,    -38.70,   -253.61,      0.00,      0.00 } },
 { {  0,  2, -1,  2 },      0.00450, {      0.12,      0.20,      0.08,      0.54,      0.00,      0.00 } },
 { {  0,  2,  0, -2 },     -0.00558, {     -0.14,     -0.25,      0.00,     -0.67,      0.00,      0.00 } },
 { {  0,  2,  0,  0 },      1.06575, {     16.47,     -0.60,      3.28,    127.54,      0.00,      0.00 } },
 { {  0,  2,  0,  2 },     -0.00342, {     -0.07,     -0.15,      0.00,     -0.41,      0.00,      0.00 } },
 { {  0,  2,  1, -2 },     -0.00237, {      0.00,     -0.11,     -0.04,     -0.28,      0.00,      0.00 } },
 { {  0,  2,  1,  0 },      1.16562, {     22.20,     -0.39,     21.49,    139.60,      0.00,      0.00 } },
 { {  0,  2,  1,  2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  2,  2, -2 },      0.00018, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  0,  2,  2,  0 },      0.05072, {      0.70,      0.00,      1.86,      6.07,      0.00,      0.00 } },
 { {  0,  2,  3,  0 },      0.00228, {      0.00,      0.00,      0.13,      0.27,      0.00,      0.00 } },
 { {  0,  2,  4,  0 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -3,  0 },     -0.00021, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  0,  3, -2,  0 },     -0.00287, {     -0.10,      0.00,     -0.10,     -0.51,      0.00,      0.00 } },
 { {  0,  3, -2,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3, -1,  0 },     -0.03894, {     -1.14,      0.03,     -0.71,     -6.99,      0.00,      0.00 } },
 { {  0,  3, -1,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  0, -2 },     -0.00017, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  0,  3,  0,  0 },      0.02127, {      0.03,      0.00,      0.08,      3.82,      0.00,      0.00 } },
 { {  0,  3,  0,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  3,  1,  0 },      0.01912, {      0.39,      0.00,      0.35,      3.43,      0.00,      0.00 } },
 { {  0,  3,  2,  0 },      0.00078, {      0.00,      0.00,      0.03,      0.14,      0.00,      0.00 } },
 { {  0,  3,  3,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4, -2,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  0,  4, -1,  0 },     -0.00073, {     -0.02,      0.00,      0.00,     -0.17,      0.00,      0.00 } },
 { {  0,  4,  0,  0 },      0.00034, {      0.00,      0.00,      0.00,      0.08,      0.00,      0.00 } },
 { {  0,  4,  1,  0 },      0.00037, {      0.00,      0.00,      0.00,      0.09,      0.00,      0.00 } },
 { {  1, -4,  0,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1, -3, -1,  0 },      0.00035, {      0.00,      0.00,      0.00,      0.06,      0.13,      0.00 } },
 { {  1, -3,  0,  0 },     -0.00161, {      0.00,      0.00,      0.00,     -0.29,     -0.63,      0.00 } },
 { {  1, -2, -3,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2, -2,  0 },      0.00029, {      0.00,      0.00,      0.00,      0.04,      0.11,      0.00 } },
 { {  1, -2, -1,  0 },      0.00844, {      0.10,      0.00,      0.17,      1.01,      3.28,      0.00 } },
 { {  1, -2,  0, -2 },     -0.00024, {      0.00,      0.00,      0.00,     -0.03,     -0.09,      0.00 } },
 { {  1, -2,  0,  0 },     -0.02948, {      0.18,      0.00,     -0.03,     -3.52,    -11.46,      0.06 } },
 { {  1, -2,  0,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2,  1, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -2,  1,  0 },      0.00159, {      0.16,      0.00,      0.03,      0.19,      0.62,      0.00 } },
 { {  1, -2,  2,  0 },      0.00036, {      0.03,      0.00,      0.00,      0.04,      0.14,      0.00 } },
 { {  1, -2,  3,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1, -4,  0 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  1, -1, -3,  0 },      0.00065, {      0.00,      0.00,      0.04,      0.04,      0.25,      0.00 } },
 { {  1, -1, -2, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1, -2,  0 },     -0.00422, {     -0.24,      0.00,     -0.13,     -0.26,     -1.64,      0.00 } },
 { {  1, -1, -2,  2 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -1, -1, -2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  1, -1, -1,  0 },      0.11411, {      0.00,     -0.02,      2.31,      6.80,     44.37,     -0.23 } },
 { {  1, -1, -1,  2 },      0.00021, {      0.00,      0.00,      0.00,      0.00,      0.08,      0.00 } },
 { {  1, -1,  0, -2 },     -0.01366, {     -0.21,     -0.61,      0.00,     -0.82,     -5.31,      0.03 } },
 { {  1, -1,  0,  0 },      0.49757, {     32.61,     -1.42,      0.14,     30.00,    193.46,     -1.02 } },
 { {  1, -1,  0,  2 },     -0.00074, {     -0.03,     -0.03,      0.00,     -0.04,     -0.29,      0.00 } },
 { {  1, -1,  1, -2 },      0.00044, {      0.00,      0.02,      0.00,      0.03,      0.17,      0.00 } },
 { {  1, -1,  1,  0 },      0.10998, {      5.32,     -0.17,      2.14,      6.59,     42.76,     -0.23 } },
 { {  1, -1,  1,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  1, -1,  2, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1, -1,  2,  0 },      0.01121, {      0.52,      0.00,      0.42,      0.67,      4.36,     -0.02 } },
 { {  1, -1,  3,  0 },      0.00097, {      0.04,      0.00,      0.05,      0.06,      0.38,      0.00 } },
 { {  1, -1,  4,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  0, -6,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  0, -5,  0 },     -0.00043, {      0.00,      0.00,     -0.04,      0.00,     -0.17,      0.00 } },
 { {  1,  0, -4,  0 },     -0.00636, {     -0.18,      0.00,     -0.46,      0.00,     -2.47,      0.00 } },
 { {  1,  0, -3, -2 },     -0.00019, {      0.00,      0.00,      0.00,      0.00,     -0.07,      0.00 } },
 { {  1,  0, -3,  0 },     -0.09938, {     -2.83,      0.09,     -5.45,     -0.04,    -38.64,      0.20 } },
 { {  1,  0, -3,  2 },     -0.00048, {      0.00,     -0.02,     -0.03,      0.00,     -0.19,      0.00 } },
 { {  1,  0, -2, -2 },     -0.00278, {     -0.04,     -0.12,     -0.10,      0.00,     -1.08,      0.00 } },
 { {  1,  0, -2,  0 },     -1.73852, {    -48.13,      1.59,    -63.71,     -0.50,   -675.99,      3.56 } },
 { {  1,  0, -2,  2 },      0.00457, {      0.03,      0.20,      0.16,      0.00,      1.78,      0.00 } },
 { {  1,  0, -1, -2 },     -0.04232, {     -0.60,     -1.87,     -0.77,      0.00,    -16.45,      0.09 } },
 { {  1,  0, -1,  0 },     -8.37909, {   -219.27,      9.53,   -156.07,     -0.84,  -3257.99,     17.18 } },
 { {  1,  0, -1,  2 },     -0.02471, {     -0.50,     -1.08,     -0.46,      0.00,     -9.61,      0.05 } },
 { {  1,  0,  0, -2 },     -0.79564, {    -11.23,    -35.24,     -0.22,      0.00,   -309.37,      1.63 } },
 { {  1,  0,  0,  0 },    108.74265, {   2351.06,   -120.48,     69.82,      2.01,  42281.56,   -222.90 } },
 { {  1,  0,  0,  2 },     -0.01874, {     -0.51,     -0.83,     -0.13,      0.00,     -7.29,      0.04 } },
 { {  1,  0,  1, -4 },      0.00017, {      0.00,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  1,  0,  1, -2 },      0.06055, {      0.87,      2.67,      1.14,      0.00,     23.54,     -0.12 } },
 { {  1,  0,  1,  0 },      6.32199, {    138.54,     -6.74,    118.10,     -0.56,   2458.13,    -12.96 } },
 { {  1,  0,  1,  2 },     -0.00164, {     -0.05,     -0.07,     -0.04,      0.00,     -0.64,      0.00 } },
 { {  1,  0,  2, -2 },      0.00203, {      0.00,      0.09,      0.07,      0.00,      0.79,      0.00 } },
 { {  1,  0,  2,  0 },      0.37852, {      7.87,     -0.40,     13.92,     -0.09,    147.18,     -0.78 } },
 { {  1,  0,  2,  2 },     -0.00013, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  1,  0,  3, -2 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  1,  0,  3,  0 },      0.02368, {      0.46,     -0.02,      1.30,      0.00,      9.21,     -0.05 } },
 { {  1,  0,  4,  0 },      0.00152, {      0.03,      0.00,      0.11,      0.00,      0.59,      0.00 } },
 { {  1,  0,  5,  0 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  1,  1, -5,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1,  1, -4,  0 },      0.00101, {      0.00,      0.00,      0.07,      0.06,      0.39,      0.00 } },
 { {  1,  1, -3,  0 },      0.01729, {      0.21,      0.00,      0.95,      1.04,      6.72,     -0.04 } },
 { {  1,  1, -3,  2 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.05,      0.00 } },
 { {  1,  1, -2, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  1, -2,  0 },      0.33226, {      4.67,     -0.10,     12.19,     19.90,    129.18,     -0.68 } },
 { {  1,  1, -2,  2 },     -0.00148, {      0.00,     -0.07,     -0.05,     -0.09,     -0.58,      0.00 } },
 { {  1,  1, -1, -2 },      0.00156, {      0.00,      0.07,      0.03,      0.09,      0.61,      0.00 } },
 { {  1,  1, -1,  0 },      0.85127, {     -2.92,     -0.50,     15.94,     50.91,    330.99,     -1.74 } },
 { {  1,  1, -1,  2 },      0.00331, {     -0.02,      0.15,      0.06,      0.20,      1.29,      0.00 } },
 { {  1,  1,  0, -2 },      0.03379, {      0.04,      1.48,      0.11,      2.02,     13.14,     -0.07 } },
 { {  1,  1,  0,  0 },    -16.67533, {     -4.78,     10.81,    -19.36,   -997.67,  -6483.73,     34.18 } },
 { {  1,  1,  0,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  1,  1,  1, -2 },     -0.00562, {      0.00,     -0.25,     -0.11,     -0.34,     -2.18,      0.00 } },
 { {  1,  1,  1,  0 },     -0.93335, {     -1.07,      0.64,    -18.08,    -55.84,   -362.91,      1.91 } },
 { {  1,  1,  2, -2 },     -0.00037, {     -0.02,      0.00,      0.00,     -0.02,     -0.14,      0.00 } },
 { {  1,  1,  2,  0 },     -0.05904, {     -0.12,      0.04,     -2.22,     -3.53,    -22.96,      0.12 } },
 { {  1,  1,  3, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  1,  3,  0 },     -0.00392, {      0.00,      0.00,     -0.22,     -0.23,     -1.52,      0.00 } },
 { {  1,  1,  4,  0 },     -0.00027, {      0.00,      0.00,     -0.02,      0.00,     -0.10,      0.00 } },
 { {  1,  2, -4,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2, -3,  0 },      0.00028, {      0.00,      0.00,      0.00,      0.03,      0.11,      0.00 } },
 { {  1,  2, -2,  0 },      0.00324, {      0.14,      0.00,      0.12,      0.39,      1.26,      0.00 } },
 { {  1,  2, -1,  0 },      0.00438, {      0.14,      0.00,      0.08,      0.52,      1.70,      0.00 } },
 { {  1,  2,  0, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2,  0,  0 },      0.03729, {      0.60,     -0.03,      0.07,      4.47,     14.50,     -0.08 } },
 { {  1,  2,  1, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  1,  2,  1,  0 },      0.00646, {      0.12,      0.00,      0.13,      0.77,      2.51,      0.00 } },
 { {  1,  2,  2,  0 },      0.00067, {      0.00,      0.00,      0.03,      0.08,      0.26,      0.00 } },
 { {  1,  2,  3,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1,  3, -2,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  1,  3, -1,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  1,  3,  0,  0 },      0.00054, {      0.00,      0.00,      0.00,      0.10,      0.21,      0.00 } },
 { {  1,  3,  1,  0 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  2, -6,  0,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -5, -1,  0 },     -0.00012, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  2, -5,  0,  0 },     -0.00058, {     -0.03,      0.00,      0.00,     -0.17,      0.00,      0.00 } },
 { {  2, -5,  1,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4, -2,  0 },      0.00036, {      0.00,      0.00,      0.00,      0.09,      0.00,      0.00 } },
 { {  2, -4, -1,  0 },     -0.00426, {     -0.08,      0.00,     -0.08,     -1.02,      0.00,      0.00 } },
 { {  2, -4,  0, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -4,  0,  0 },     -0.01578, {     -0.66,      0.00,      0.00,     -3.78,      0.00,      0.00 } },
 { {  2, -4,  1,  0 },     -0.00113, {     -0.05,      0.00,     -0.02,     -0.27,      0.00,      0.00 } },
 { {  2, -4,  2,  0 },     -0.00009, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2, -3, -4,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -3,  0 },      0.00046, {      0.00,      0.00,      0.02,      0.08,      0.00,      0.00 } },
 { {  2, -3, -2,  0 },      0.01116, {      0.26,      0.00,      0.41,      2.00,      0.00,      0.00 } },
 { {  2, -3, -2,  2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3, -1, -2 },      0.00026, {      0.00,      0.00,      0.00,      0.05,      0.00,      0.00 } },
 { {  2, -3, -1,  0 },     -0.14808, {     -2.64,      0.03,     -2.71,    -26.58,      0.00,      0.00 } },
 { {  2, -3, -1,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  0, -2 },      0.00104, {      0.03,      0.05,      0.00,      0.19,      0.00,      0.00 } },
 { {  2, -3,  0,  0 },     -0.41076, {    -15.67,      0.09,     -0.51,    -73.71,      0.00,      0.00 } },
 { {  2, -3,  1, -2 },      0.00027, {      0.00,      0.00,      0.00,      0.05,      0.00,      0.00 } },
 { {  2, -3,  1,  0 },     -0.02855, {     -1.20,      0.00,     -0.56,     -5.12,      0.00,      0.00 } },
 { {  2, -3,  2, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -3,  2,  0 },     -0.00209, {     -0.10,      0.00,     -0.08,     -0.38,      0.00,      0.00 } },
 { {  2, -3,  3,  0 },     -0.00016, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  2, -2, -5,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -4,  0 },      0.00067, {      0.00,      0.00,      0.05,      0.08,      0.00,      0.00 } },
 { {  2, -2, -3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2, -3,  0 },      0.01517, {      0.23,      0.00,      0.83,      1.81,      0.00,      0.00 } },
 { {  2, -2, -2, -2 },      0.00041, {      0.00,      0.00,      0.00,      0.05,      0.00,      0.00 } },
 { {  2, -2, -2,  0 },      0.34304, {      6.87,     -0.03,     12.50,     41.04,      0.00,      0.00 } },
 { {  2, -2, -2,  2 },      0.00099, {      0.00,      0.04,      0.04,      0.12,      0.00,      0.00 } },
 { {  2, -2, -1, -2 },      0.00957, {      0.13,      0.43,      0.17,      1.15,      0.00,      0.00 } },
 { {  2, -2, -1,  0 },     -4.95049, {    -84.90,      0.91,    -90.35,   -592.24,      0.03,      0.00 } },
 { {  2, -2, -1,  2 },      0.00168, {      0.04,      0.08,      0.04,      0.20,      0.00,      0.00 } },
 { {  2, -2,  0, -2 },      0.02839, {      0.78,      1.27,     -0.11,      3.40,      0.00,      0.00 } },
 { {  2, -2,  0,  0 },     -9.88519, {   -346.35,      1.99,    -14.48,  -1182.31,      0.07,      0.00 } },
 { {  2, -2,  1, -4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  1, -2 },      0.00792, {      0.09,      0.36,      0.14,      0.95,      0.00,      0.00 } },
 { {  2, -2,  1,  0 },     -0.65758, {    -25.02,      0.18,    -13.07,    -78.64,      0.00,      0.00 } },
 { {  2, -2,  2, -2 },      0.00084, {      0.02,      0.04,      0.03,      0.10,      0.00,      0.00 } },
 { {  2, -2,  2,  0 },     -0.04608, {     -1.86,      0.00,     -1.76,     -5.51,      0.00,      0.00 } },
 { {  2, -2,  3, -2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -2,  3,  0 },     -0.00335, {     -0.14,      0.00,     -0.19,     -0.40,      0.00,      0.00 } },
 { {  2, -2,  4,  0 },     -0.00025, {      0.00,      0.00,     -0.02,     -0.03,      0.00,      0.00 } },
 { {  2, -2,  5,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -6,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -5,  0 },      0.00119, {      0.00,      0.00,      0.11,      0.07,      0.00,      0.00 } },
 { {  2, -1, -4, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1, -4,  0 },      0.02334, {      0.28,      0.00,      1.70,      1.40,      0.00,      0.00 } },
 { {  2, -1, -3, -2 },      0.00079, {      0.00,      0.04,      0.04,      0.05,      0.00,      0.00 } },
 { {  2, -1, -3,  0 },      0.49506, {      7.68,      0.00,     27.04,     29.62,      0.00,      0.00 } },
 { {  2, -1, -3,  2 },     -0.00046, {      0.00,     -0.02,     -0.02,     -0.03,      0.00,      0.00 } },
 { {  2, -1, -2, -2 },      0.01518, {      0.17,      0.68,      0.54,      0.91,      0.00,      0.00 } },
 { {  2, -1, -2,  0 },     10.05654, {    188.87,     -1.29,    366.36,    601.57,     -0.03,      0.00 } },
 { {  2, -1, -2,  2 },      0.03024, {      0.36,      1.36,      1.10,      1.81,      0.00,      0.00 } },
 { {  2, -1, -1, -2 },      0.32336, {      4.56,     14.42,      5.73,     19.34,      0.00,      0.00 } },
 { {  2, -1, -1,  0 },   -152.14314, {  -2564.20,     30.40,  -2771.74,  -9097.84,      0.06,      0.00 } },
 { {  2, -1, -1,  2 },      0.03734, {      0.84,      1.68,      0.85,      2.23,      0.00,      0.00 } },
 { {  2, -1,  0, -4 },     -0.00035, {      0.00,     -0.03,      0.00,     -0.02,      0.00,      0.00 } },
 { {  2, -1,  0, -2 },      0.65710, {     17.83,     29.30,     -2.92,     39.26,      0.00,      0.00 } },
 { {  2, -1,  0,  0 },   -204.59357, {  -6583.80,     38.79,   -377.95, -12225.82,      0.27,      0.00 } },
 { {  2, -1,  0,  2 },      0.00108, {     -0.09,      0.05,      0.21,      0.06,      0.00,      0.00 } },
 { {  2, -1,  1, -4 },     -0.00085, {      0.00,     -0.08,      0.00,     -0.05,      0.00,      0.00 } },
 { {  2, -1,  1, -2 },      0.20942, {      1.81,      9.46,      3.82,     12.50,      0.00,      0.00 } },
 { {  2, -1,  1,  0 },    -12.83185, {   -434.14,      2.98,   -260.02,   -766.51,      0.02,      0.00 } },
 { {  2, -1,  1,  2 },      0.00000, {      0.00,      0.00,      0.02,      0.00,      0.00,      0.00 } },
 { {  2, -1,  2, -4 },      0.00013, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  2, -2 },      0.01804, {      0.34,      0.81,      0.66,      1.08,      0.00,      0.00 } },
 { {  2, -1,  2,  0 },     -0.84883, {    -29.75,      0.22,    -32.80,    -50.68,      0.00,      0.00 } },
 { {  2, -1,  3, -2 },      0.00127, {      0.03,      0.06,      0.07,      0.08,      0.00,      0.00 } },
 { {  2, -1,  3,  0 },     -0.05850, {     -2.10,      0.00,     -3.33,     -3.49,      0.00,      0.00 } },
 { {  2, -1,  4, -2 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2, -1,  4,  0 },     -0.00413, {     -0.15,      0.00,     -0.31,     -0.25,      0.00,      0.00 } },
 { {  2, -1,  5,  0 },     -0.00030, {      0.00,      0.00,     -0.03,      0.00,      0.00,      0.00 } },
 { {  2, -1,  6,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -7,  0 },      0.00020, {      0.00,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  2,  0, -6,  0 },      0.00304, {      0.06,      0.00,      0.33,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5, -2 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5,  0 },      0.04740, {      0.93,      0.00,      4.32,      0.00,      0.00,      0.00 } },
 { {  2,  0, -5,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4, -2 },      0.00188, {      0.03,      0.08,      0.14,      0.00,      0.00,      0.00 } },
 { {  2,  0, -4,  0 },      0.77854, {     15.43,     -0.13,     56.71,      0.05,      0.00,      0.00 } },
 { {  2,  0, -4,  2 },      0.00048, {      0.02,      0.02,      0.04,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3, -2 },      0.02910, {      0.44,      1.30,      1.57,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3,  0 },     14.40262, {    282.85,     -2.35,    786.99,      0.58,      0.14,      0.00 } },
 { {  2,  0, -3,  2 },     -0.02008, {     -0.43,     -0.89,     -1.09,      0.00,      0.00,      0.00 } },
 { {  2,  0, -3,  4 },     -0.00018, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -4 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2, -2 },      0.47263, {      7.20,     21.06,     16.92,      0.00,      0.00,      0.00 } },
 { {  2,  0, -2,  0 },    246.15768, {   4806.36,    -51.78,   8962.91,     -0.43,      1.11,      0.00 } },
 { {  2,  0, -2,  2 },      0.77405, {      9.68,     34.65,     28.17,     -0.04,      0.00,      0.00 } },
 { {  2,  0, -2,  4 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -4 },     -0.00039, {      0.00,     -0.04,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0, -1, -2 },      8.75170, {    133.45,    390.01,    155.98,     -0.28,      0.20,      0.00 } },
 { {  2,  0, -1,  0 },  -3699.10468, { -63127.05,    818.00, -67236.74,    147.86,    -15.95,      0.14 } },
 { {  2,  0, -1,  2 },      0.59633, {     12.13,     26.76,     14.46,     -0.08,      0.03,      0.00 } },
 { {  2,  0,  0, -4 },     -0.00745, {     -0.11,     -0.67,     -0.10,      0.00,      0.00,      0.00 } },
 { {  2,  0,  0, -2 },     10.32129, {    282.08,    460.16,    -60.76,     -1.29,      0.06,      0.00 } },
 { {  2,  0,  0,  0 },  -2955.96651, { -86674.51,    507.99,  -7597.69,    323.54,    -19.09,      0.14 } },
 { {  2,  0,  0,  2 },      0.03143, {     -0.69,      1.43,      2.99,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -4 },     -0.01866, {     -0.11,     -1.67,     -0.33,      0.00,      0.00,      0.00 } },
 { {  2,  0,  1, -2 },      4.13118, {     20.86,    186.87,     75.09,     -0.76,      0.00,      0.00 } },
 { {  2,  0,  1,  0 },   -170.73274, {  -4963.30,     29.72,  -3579.16,     26.82,     -1.82,      0.00 } },
 { {  2,  0,  1,  2 },      0.00181, {     -0.12,      0.08,      0.32,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -4 },      0.00219, {      0.00,      0.20,      0.08,      0.00,      0.00,      0.00 } },
 { {  2,  0,  2, -2 },      0.28399, {      2.90,     12.83,     10.35,     -0.06,      0.00,      0.00 } },
 { {  2,  0,  2,  0 },    -10.44472, {   -300.48,      1.81,   -410.75,      2.24,     -0.17,      0.00 } },
 { {  2,  0,  2,  2 },      0.00014, {      0.00,      0.00,      0.03,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -4 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3, -2 },      0.01811, {      0.19,      0.82,      0.99,      0.00,      0.00,      0.00 } },
 { {  2,  0,  3,  0 },     -0.66968, {    -19.02,      0.12,    -38.59,      0.19,      0.00,      0.00 } },
 { {  2,  0,  4, -2 },      0.00119, {      0.00,      0.05,      0.09,      0.00,      0.00,      0.00 } },
 { {  2,  0,  4,  0 },     -0.04422, {     -1.24,      0.00,     -3.36,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5, -2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  0,  5,  0 },     -0.00298, {     -0.08,      0.00,     -0.28,      0.00,      0.00,      0.00 } },
 { {  2,  0,  6,  0 },     -0.00020, {      0.00,      0.00,     -0.02,      0.00,      0.00,      0.00 } },
 { {  2,  1, -6,  0 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -5,  0 },      0.00116, {      0.08,      0.00,      0.11,      0.07,      0.00,      0.00 } },
 { {  2,  1, -4, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -4,  0 },      0.01417, {      1.08,      0.00,      1.03,      0.85,      0.00,      0.00 } },
 { {  2,  1, -3, -2 },      0.00021, {      0.02,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1, -3,  0 },      0.16858, {     15.79,     -0.24,      9.16,     10.17,      0.05,      0.00 } },
 { {  2,  1, -3,  2 },      0.00030, {      0.00,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  2,  1, -2, -2 },      0.00052, {      0.27,      0.02,      0.00,      0.03,      0.00,      0.00 } },
 { {  2,  1, -2,  0 },      0.14368, {    122.13,     -1.91,      4.58,      9.23,      0.13,      0.00 } },
 { {  2,  1, -2,  2 },     -0.00203, {      0.31,     -0.10,     -0.08,     -0.12,      0.00,      0.00 } },
 { {  2,  1, -1, -2 },     -0.04736, {      2.58,     -2.16,     -0.89,     -2.82,      0.00,      0.00 } },
 { {  2,  1, -1,  0 },     24.20935, {   -837.26,     13.30,    449.34,   1443.45,      0.26,      0.00 } },
 { {  2,  1, -1,  2 },     -0.00588, {     -0.10,     -0.26,     -0.14,     -0.35,      0.00,      0.00 } },
 { {  2,  1,  0, -4 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  0, -2 },     -0.13572, {     -4.94,     -6.04,      0.54,     -8.13,      0.03,      0.00 } },
 { {  2,  1,  0,  0 },     30.82498, {    873.36,     -8.40,     94.40,   1844.57,      3.86,     -0.02 } },
 { {  2,  1,  0,  2 },     -0.00143, {     -0.04,     -0.06,     -0.04,     -0.09,      0.00,      0.00 } },
 { {  2,  1,  1, -4 },      0.00040, {      0.00,      0.04,      0.00,      0.02,      0.00,      0.00 } },
 { {  2,  1,  1, -2 },     -0.05527, {      1.02,     -2.52,     -1.02,     -3.29,      0.00,      0.00 } },
 { {  2,  1,  1,  0 },      2.61650, {     90.88,     -1.01,     56.02,    156.64,      0.49,      0.00 } },
 { {  2,  1,  1,  2 },     -0.00015, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  2, -4 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  2, -2 },     -0.00537, {     -0.03,     -0.24,     -0.20,     -0.32,      0.00,      0.00 } },
 { {  2,  1,  2,  0 },      0.21252, {      8.11,     -0.09,      8.45,     12.72,      0.05,      0.00 } },
 { {  2,  1,  3, -2 },     -0.00044, {      0.00,     -0.02,     -0.02,     -0.03,      0.00,      0.00 } },
 { {  2,  1,  3,  0 },      0.01704, {      0.68,      0.00,      0.99,      1.02,      0.00,      0.00 } },
 { {  2,  1,  4, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  1,  4,  0 },      0.00135, {      0.06,      0.00,      0.10,      0.08,      0.00,      0.00 } },
 { {  2,  1,  5,  0 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -4,  0 },     -0.00041, {      0.00,      0.00,     -0.03,     -0.05,      0.00,      0.00 } },
 { {  2,  2, -3, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2, -3,  0 },     -0.01036, {     -0.29,      0.00,     -0.58,     -1.25,      0.00,      0.00 } },
 { {  2,  2, -2, -2 },     -0.00055, {      0.00,     -0.02,     -0.02,     -0.07,      0.00,      0.00 } },
 { {  2,  2, -2,  0 },     -0.10888, {     -2.72,      0.00,     -3.96,    -13.05,      0.00,      0.00 } },
 { {  2,  2, -2,  2 },     -0.00046, {      0.00,     -0.02,      0.00,     -0.06,      0.00,      0.00 } },
 { {  2,  2, -1, -2 },     -0.01072, {     -0.37,     -0.47,     -0.17,     -1.28,      0.00,      0.00 } },
 { {  2,  2, -1,  0 },      2.35380, {     79.52,     -0.72,     42.86,    282.05,     -0.06,      0.00 } },
 { {  2,  2, -1,  2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  0, -2 },      0.00080, {      0.02,      0.04,      0.04,      0.10,      0.00,      0.00 } },
 { {  2,  2,  0,  0 },      0.14764, {      4.88,     -0.05,      4.90,     17.69,     -0.29,      0.00 } },
 { {  2,  2,  1, -2 },      0.00187, {      0.11,      0.07,      0.00,      0.22,      0.00,      0.00 } },
 { {  2,  2,  1,  0 },      0.00912, {      0.29,      0.00,      0.46,      1.09,     -0.04,      0.00 } },
 { {  2,  2,  2, -2 },      0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  2,  2,  0 },      0.00036, {      0.00,      0.00,      0.03,      0.04,      0.00,      0.00 } },
 { {  2,  3, -3,  0 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3, -2,  0 },     -0.00052, {     -0.02,      0.00,     -0.02,     -0.09,      0.00,      0.00 } },
 { {  2,  3, -1, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  2,  3, -1,  0 },      0.00208, {      0.14,      0.00,      0.03,      0.37,      0.00,      0.00 } },
 { {  2,  3,  0,  0 },     -0.00024, {     -0.02,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  2,  4, -1,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -4, -1,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -4,  0,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -3, -2,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  3, -3, -1,  0 },      0.00065, {      0.03,      0.00,      0.00,      0.12,      0.25,      0.00 } },
 { {  3, -3,  0, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -3,  0,  0 },     -0.00055, {      0.00,      0.00,      0.00,     -0.10,     -0.21,      0.00 } },
 { {  3, -3,  1,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -2, -3,  0 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  3, -2, -2,  0 },      0.00224, {      0.03,      0.00,      0.08,      0.27,      0.87,      0.00 } },
 { {  3, -2, -1, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -2, -1,  0 },      0.01389, {      0.55,      0.00,      0.26,      1.66,      5.40,     -0.03 } },
 { {  3, -2,  0, -2 },      0.00070, {      0.00,      0.03,      0.00,      0.08,      0.27,      0.00 } },
 { {  3, -2,  0,  0 },     -0.01048, {     -0.23,      0.00,      0.04,     -1.25,     -4.07,      0.02 } },
 { {  3, -2,  1, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -2,  1,  0 },     -0.00067, {      0.00,      0.00,      0.00,     -0.08,     -0.26,      0.00 } },
 { {  3, -2,  2,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1, -4,  0 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  3, -1, -3,  0 },     -0.00334, {     -0.05,      0.00,     -0.18,     -0.20,     -1.30,      0.00 } },
 { {  3, -1, -2, -2 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  3, -1, -2,  0 },      0.05293, {      0.92,      0.00,      1.93,      3.17,     20.58,     -0.11 } },
 { {  3, -1, -2,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3, -1, -1, -2 },     -0.00089, {      0.00,     -0.04,      0.00,     -0.05,     -0.34,      0.00 } },
 { {  3, -1, -1,  0 },      0.25561, {      9.61,     -0.15,      4.83,     15.24,     99.38,     -0.52 } },
 { {  3, -1,  0, -2 },      0.01441, {      0.21,      0.64,      0.00,      0.86,      5.61,     -0.03 } },
 { {  3, -1,  0,  0 },     -0.15708, {     -2.88,     -0.04,      0.83,     -9.42,    -61.08,      0.32 } },
 { {  3, -1,  1, -2 },      0.00049, {      0.00,      0.02,      0.00,      0.03,      0.19,      0.00 } },
 { {  3, -1,  1,  0 },     -0.00887, {     -0.03,      0.00,     -0.09,     -0.54,     -3.45,      0.00 } },
 { {  3, -1,  2, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3, -1,  2,  0 },     -0.00030, {      0.00,      0.00,      0.00,      0.00,     -0.12,      0.00 } },
 { {  3,  0, -5,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -4,  0 },     -0.00170, {      0.00,      0.00,     -0.12,      0.00,     -0.66,      0.00 } },
 { {  3,  0, -3, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0, -3,  0 },     -0.04843, {     -0.72,      0.00,     -2.65,     -0.03,    -18.83,      0.10 } },
 { {  3,  0, -3,  2 },     -0.00013, {      0.00,      0.00,      0.00,      0.00,     -0.05,      0.00 } },
 { {  3,  0, -2, -2 },     -0.00206, {      0.00,     -0.09,     -0.07,      0.00,     -0.80,      0.00 } },
 { {  3,  0, -2,  0 },      0.86243, {     16.64,     -0.34,     31.48,      0.13,    335.34,     -1.77 } },
 { {  3,  0, -2,  2 },     -0.00101, {     -0.03,     -0.04,     -0.04,      0.00,     -0.39,      0.00 } },
 { {  3,  0, -1, -2 },     -0.00868, {      0.00,     -0.39,     -0.14,      0.00,     -3.38,      0.00 } },
 { {  3,  0, -1,  0 },      3.25823, {    117.77,     -2.27,     61.91,     -1.34,   1266.88,     -6.68 } },
 { {  3,  0, -1,  2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  3,  0,  0, -4 },     -0.00018, {      0.00,      0.00,      0.00,      0.00,     -0.07,      0.00 } },
 { {  3,  0,  0, -2 },      0.20099, {      2.46,      8.95,     -0.04,     -0.07,     78.15,     -0.41 } },
 { {  3,  0,  0,  0 },     -1.41893, {    -19.43,     -0.69,     10.76,     -0.63,   -551.71,      2.91 } },
 { {  3,  0,  0,  2 },     -0.00015, {      0.00,      0.00,      0.00,      0.00,     -0.06,      0.00 } },
 { {  3,  0,  1, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  0,  1, -2 },      0.00671, {      0.02,      0.30,      0.12,      0.00,      2.61,      0.00 } },
 { {  3,  0,  1,  0 },     -0.06545, {      0.72,     -0.10,     -0.32,     -0.11,    -25.45,      0.13 } },
 { {  3,  0,  1,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  0,  2, -2 },      0.00035, {      0.00,      0.00,      0.00,      0.00,      0.14,      0.00 } },
 { {  3,  0,  2,  0 },     -0.00111, {      0.21,      0.00,      0.03,      0.00,     -0.43,      0.00 } },
 { {  3,  0,  3,  0 },      0.00016, {      0.03,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  3,  0,  4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -5,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -4,  0 },     -0.00065, {     -0.03,      0.00,     -0.05,     -0.04,     -0.25,      0.00 } },
 { {  3,  1, -3,  0 },     -0.00680, {     -0.29,      0.00,     -0.37,     -0.41,     -2.64,      0.00 } },
 { {  3,  1, -3,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  1, -2, -2 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,     -0.04,      0.00 } },
 { {  3,  1, -2,  0 },      0.03810, {      2.03,      0.00,      1.37,      2.29,     14.82,     -0.08 } },
 { {  3,  1, -2,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  3,  1, -1, -2 },     -0.00023, {      0.00,      0.00,      0.00,      0.00,     -0.09,      0.00 } },
 { {  3,  1, -1,  0 },     -0.21259, {     -4.00,      0.17,     -4.03,    -12.70,    -82.66,      0.44 } },
 { {  3,  1, -1,  2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  3,  1,  0, -2 },     -0.00808, {     -0.12,     -0.36,      0.00,     -0.48,     -3.14,      0.00 } },
 { {  3,  1,  0,  0 },     -0.10766, {     -3.22,      0.12,     -1.09,     -6.42,    -41.86,      0.22 } },
 { {  3,  1,  1, -2 },     -0.00018, {      0.00,      0.00,      0.00,      0.00,     -0.07,      0.00 } },
 { {  3,  1,  1,  0 },     -0.01448, {     -0.44,      0.00,     -0.37,     -0.86,     -5.63,      0.03 } },
 { {  3,  1,  2,  0 },     -0.00150, {     -0.05,      0.00,     -0.07,     -0.09,     -0.58,      0.00 } },
 { {  3,  1,  3,  0 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,     -0.06,      0.00 } },
 { {  3,  2, -3,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  3,  2, -2,  0 },      0.00032, {      0.05,      0.00,      0.00,      0.04,      0.12,      0.00 } },
 { {  3,  2, -1,  0 },      0.00099, {     -0.05,      0.00,      0.02,      0.12,      0.39,      0.00 } },
 { {  3,  2,  0, -2 },      0.00009, {      0.00,      0.00,      0.00,      0.02,      0.04,      0.00 } },
 { {  3,  2,  0,  0 },      0.00162, {      0.04,      0.00,      0.00,      0.19,      0.63,      0.00 } },
 { {  3,  2,  1,  0 },      0.00028, {      0.00,      0.00,      0.00,      0.03,      0.11,      0.00 } },
 { {  3,  2,  2,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  3, -2, -2 },      0.00000, {     -0.03,      0.00,     -0.03,      0.00,     -0.07,      0.00 } },
 { {  3,  3, -2,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  3,  3, -1,  0 },      0.00010, {      0.00,      0.00,      0.00,      0.02,      0.04,      0.00 } },
 { {  3,  3,  0, -2 },      0.00002, {      0.08,     -0.06,      0.07,      0.08,     -0.05,      0.00 } },
 { {  4, -5, -1,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -5,  0,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -4, -2,  0 },     -0.00022, {      0.00,      0.00,      0.00,     -0.05,      0.00,      0.00 } },
 { {  4, -4, -1,  0 },     -0.00078, {     -0.05,      0.00,      0.00,     -0.19,      0.00,      0.00 } },
 { {  4, -4,  0,  0 },     -0.00045, {     -0.03,      0.00,      0.00,     -0.11,      0.00,      0.00 } },
 { {  4, -4,  1,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -3,  0 },     -0.00011, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -3, -2,  0 },     -0.00513, {     -0.22,      0.00,     -0.19,     -0.92,      0.00,      0.00 } },
 { {  4, -3, -1, -2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -3, -1,  0 },     -0.01571, {     -0.87,      0.00,     -0.31,     -2.82,      0.00,      0.00 } },
 { {  4, -3,  0, -2 },     -0.00022, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  4, -3,  0,  0 },     -0.00817, {     -0.54,      0.00,     -0.08,     -1.47,      0.00,      0.00 } },
 { {  4, -3,  1,  0 },     -0.00114, {     -0.08,      0.00,     -0.03,     -0.21,      0.00,      0.00 } },
 { {  4, -3,  2,  0 },     -0.00013, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -2, -5,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -4,  0 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -3, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -3,  0 },     -0.00239, {     -0.12,      0.00,     -0.13,     -0.29,      0.00,      0.00 } },
 { {  4, -2, -3,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -2, -2 },     -0.00036, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  4, -2, -2,  0 },     -0.10834, {     -4.38,      0.04,     -3.95,    -12.96,      0.00,      0.00 } },
 { {  4, -2, -2,  2 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2, -1, -2 },     -0.00145, {     -0.04,     -0.06,     -0.03,     -0.17,      0.00,      0.00 } },
 { {  4, -2, -1,  0 },     -0.27906, {    -14.76,      0.11,     -5.61,    -33.36,      0.00,      0.00 } },
 { {  4, -2, -1,  2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  0, -2 },     -0.00403, {     -0.18,     -0.18,      0.00,     -0.48,      0.00,      0.00 } },
 { {  4, -2,  0,  0 },     -0.12806, {     -8.09,      0.05,     -1.35,    -15.30,     -0.02,      0.00 } },
 { {  4, -2,  1, -2 },      0.00023, {      0.00,      0.00,      0.00,      0.03,      0.00,      0.00 } },
 { {  4, -2,  1,  0 },     -0.01728, {     -1.13,      0.00,     -0.45,     -2.06,      0.00,      0.00 } },
 { {  4, -2,  2, -2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -2,  2,  0 },     -0.00184, {     -0.12,      0.00,     -0.08,     -0.22,      0.00,      0.00 } },
 { {  4, -2,  3,  0 },     -0.00018, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  4, -1, -6,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -5,  0 },     -0.00041, {      0.00,      0.00,     -0.04,     -0.02,      0.00,      0.00 } },
 { {  4, -1, -4, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -4,  0 },     -0.00230, {     -0.06,      0.00,     -0.17,     -0.14,      0.00,      0.00 } },
 { {  4, -1, -3, -2 },     -0.00059, {     -0.02,     -0.03,     -0.03,     -0.04,      0.00,      0.00 } },
 { {  4, -1, -3,  0 },     -0.04316, {     -2.19,      0.02,     -2.36,     -2.59,      0.00,      0.00 } },
 { {  4, -1, -3,  2 },      0.00077, {      0.02,      0.03,      0.04,      0.05,      0.00,      0.00 } },
 { {  4, -1, -2, -2 },     -0.00661, {     -0.25,     -0.29,     -0.24,     -0.40,      0.00,      0.00 } },
 { {  4, -1, -2,  0 },     -1.89710, {    -73.13,      0.79,    -69.17,   -113.34,     -0.12,      0.00 } },
 { {  4, -1, -2,  2 },      0.00160, {      0.07,      0.07,      0.06,      0.10,      0.00,      0.00 } },
 { {  4, -1, -1, -4 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1, -1, -2 },     -0.02698, {     -0.67,     -1.20,     -0.50,     -1.61,      0.00,      0.00 } },
 { {  4, -1, -1,  0 },     -3.95812, {   -199.78,      1.54,    -80.75,   -236.16,      0.15,      0.00 } },
 { {  4, -1, -1,  2 },      0.00069, {      0.03,      0.03,      0.03,      0.04,      0.00,      0.00 } },
 { {  4, -1,  0, -4 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  0, -2 },     -0.05789, {     -2.54,     -2.57,      0.14,     -3.44,     -0.02,      0.00 } },
 { {  4, -1,  0,  0 },     -1.57145, {    -94.02,      0.59,    -18.48,    -93.65,     -0.26,      0.00 } },
 { {  4, -1,  0,  2 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  1, -4 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  1, -2 },      0.00272, {      0.12,      0.12,      0.07,      0.16,      0.00,      0.00 } },
 { {  4, -1,  1,  0 },     -0.20286, {    -12.45,      0.08,     -5.51,    -12.08,     -0.03,      0.00 } },
 { {  4, -1,  2, -2 },      0.00055, {      0.02,      0.02,      0.02,      0.03,      0.00,      0.00 } },
 { {  4, -1,  2,  0 },     -0.02079, {     -1.29,      0.00,     -0.92,     -1.24,      0.00,      0.00 } },
 { {  4, -1,  3, -2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4, -1,  3,  0 },     -0.00194, {     -0.12,      0.00,     -0.12,     -0.12,      0.00,      0.00 } },
 { {  4, -1,  4,  0 },     -0.00017, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -7,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -6,  0 },     -0.00055, {     -0.02,      0.00,     -0.06,      0.00,      0.00,      0.00 } },
 { {  4,  0, -5, -2 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -5,  0 },     -0.00542, {     -0.21,      0.00,     -0.49,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4, -2 },     -0.00072, {     -0.03,     -0.03,     -0.05,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  0 },     -0.02939, {     -0.87,      0.00,     -2.14,      0.00,      0.00,      0.00 } },
 { {  4,  0, -4,  2 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3, -2 },     -0.00744, {     -0.27,     -0.33,     -0.41,      0.00,      0.00,      0.00 } },
 { {  4,  0, -3,  0 },     -0.51423, {    -25.81,      0.28,    -28.10,     -0.19,     -0.15,      0.00 } },
 { {  4,  0, -3,  2 },      0.00925, {      0.29,      0.41,      0.50,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -4 },     -0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2, -2 },     -0.07982, {     -2.91,     -3.54,     -2.91,      0.00,      0.00,      0.00 } },
 { {  4,  0, -2,  0 },    -21.63627, {   -808.81,      9.47,   -788.01,      2.77,     -1.21,      0.00 } },
 { {  4,  0, -2,  2 },      0.01483, {      0.60,      0.66,      0.61,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -4 },     -0.00080, {     -0.02,     -0.07,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0, -1, -2 },     -0.32176, {     -7.98,    -14.29,     -5.95,      0.15,     -0.11,      0.00 } },
 { {  4,  0, -1,  0 },    -34.78245, {  -1678.73,     13.50,   -725.23,      8.97,      0.88,      0.00 } },
 { {  4,  0, -1,  2 },      0.00550, {      0.25,      0.25,      0.21,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -4 },      0.00032, {      0.00,      0.03,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  0, -2 },     -0.50793, {    -21.44,    -22.55,      1.37,      0.28,     -0.17,      0.00 } },
 { {  4,  0,  0,  0 },    -11.64993, {   -657.42,      4.13,   -155.60,      4.42,     -2.17,      0.00 } },
 { {  4,  0,  0,  2 },      0.00067, {      0.02,      0.03,      0.05,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -4 },      0.00050, {      0.00,      0.04,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1, -2 },      0.01938, {      0.76,      0.89,      0.57,      0.00,      0.00,      0.00 } },
 { {  4,  0,  1,  0 },     -1.42255, {    -81.30,      0.50,    -40.47,      0.68,     -0.21,      0.00 } },
 { {  4,  0,  1,  2 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2, -2 },      0.00400, {      0.16,      0.18,      0.17,      0.00,      0.00,      0.00 } },
 { {  4,  0,  2,  0 },     -0.13922, {     -7.97,      0.05,     -6.33,      0.08,      0.00,      0.00 } },
 { {  4,  0,  3, -2 },      0.00044, {      0.00,      0.02,      0.03,      0.00,      0.00,      0.00 } },
 { {  4,  0,  3,  0 },     -0.01243, {     -0.71,      0.00,     -0.78,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  0,  4,  0 },     -0.00106, {     -0.06,      0.00,     -0.09,      0.00,      0.00,      0.00 } },
 { {  4,  0,  5,  0 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -5,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -4,  0 },      0.00093, {      0.05,      0.00,      0.07,      0.06,      0.00,      0.00 } },
 { {  4,  1, -3,  0 },     -0.02399, {     -2.48,      0.03,     -1.30,     -1.43,      0.00,      0.00 } },
 { {  4,  1, -3,  2 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -2, -2 },      0.00036, {     -0.05,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  4,  1, -2,  0 },      0.23696, {     -7.25,      0.12,      8.69,     14.25,      0.02,      0.00 } },
 { {  4,  1, -2,  2 },     -0.00022, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1, -1, -2 },      0.01551, {      0.50,      0.69,      0.29,      0.92,      0.00,      0.00 } },
 { {  4,  1, -1,  0 },      0.57881, {     16.49,     -0.10,     12.27,     34.72,      0.13,      0.00 } },
 { {  4,  1, -1,  2 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  0, -2 },      0.01969, {      0.95,      0.87,      0.00,      1.18,      0.00,      0.00 } },
 { {  4,  1,  0,  0 },      0.24453, {     12.94,     -0.10,      3.44,     14.63,      0.04,      0.00 } },
 { {  4,  1,  0,  2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1, -4 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1, -2 },      0.00013, {      0.03,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  1,  0 },      0.03682, {      2.17,     -0.02,      1.07,      2.20,      0.00,      0.00 } },
 { {  4,  1,  2, -2 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  1,  2,  0 },      0.00430, {      0.27,      0.00,      0.20,      0.26,      0.00,      0.00 } },
 { {  4,  1,  3,  0 },      0.00045, {      0.03,      0.00,      0.03,      0.03,      0.00,      0.00 } },
 { {  4,  1,  4,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -4,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -3, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2, -3,  0 },      0.00252, {      0.17,      0.00,      0.14,      0.30,      0.00,      0.00 } },
 { {  4,  2, -2, -2 },      0.00021, {      0.00,      0.00,      0.00,      0.03,      0.00,      0.00 } },
 { {  4,  2, -2,  0 },      0.02658, {      1.57,      0.00,      0.97,      3.18,      0.00,      0.00 } },
 { {  4,  2, -1, -2 },     -0.00064, {     -0.04,     -0.03,      0.00,     -0.08,      0.00,      0.00 } },
 { {  4,  2, -1,  0 },      0.01868, {      1.31,      0.00,      0.45,      2.24,      0.00,      0.00 } },
 { {  4,  2,  0, -2 },      0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  2,  0,  0 },      0.00136, {      0.11,      0.00,      0.09,      0.16,      0.00,      0.00 } },
 { {  4,  2,  1,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  4,  3, -2,  0 },     -0.00015, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  4,  3, -1,  0 },     -0.00020, {      0.00,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  4,  3,  0,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -3, -2,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2, -3,  0 },      0.00009, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  5, -2, -2,  0 },      0.00044, {      0.03,      0.00,      0.00,      0.05,      0.17,      0.00 } },
 { {  5, -2, -1, -2 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -2, -1,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5, -2,  0,  0 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,     -0.06,      0.00 } },
 { {  5, -2,  1,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -3,  0 },      0.00128, {      0.06,      0.00,      0.07,      0.08,      0.50,      0.00 } },
 { {  5, -1, -2, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1, -2,  0 },      0.00573, {      0.32,      0.00,      0.22,      0.34,      2.23,      0.00 } },
 { {  5, -1, -1, -2 },      0.00034, {      0.00,      0.00,      0.00,      0.02,      0.13,      0.00 } },
 { {  5, -1, -1,  0 },     -0.00056, {      0.07,      0.00,      0.02,     -0.04,     -0.22,      0.00 } },
 { {  5, -1,  0, -2 },      0.00021, {      0.00,      0.00,      0.00,      0.00,      0.08,      0.00 } },
 { {  5, -1,  0,  0 },     -0.00153, {     -0.06,      0.00,      0.00,     -0.09,     -0.60,      0.00 } },
 { {  5, -1,  1, -2 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5, -1,  1,  0 },     -0.00019, {      0.00,      0.00,      0.00,      0.00,     -0.07,      0.00 } },
 { {  5,  0, -4,  0 },      0.00026, {      0.00,      0.00,      0.02,      0.00,      0.10,      0.00 } },
 { {  5,  0, -3, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0, -3,  0 },      0.01149, {      0.49,      0.00,      0.63,      0.00,      4.47,     -0.02 } },
 { {  5,  0, -3,  2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  0, -2, -2 },      0.00030, {      0.00,      0.00,      0.00,      0.00,      0.12,      0.00 } },
 { {  5,  0, -2,  0 },      0.04591, {      2.50,     -0.04,      1.73,     -0.03,     17.85,     -0.09 } },
 { {  5,  0, -1, -2 },      0.00281, {      0.10,      0.12,      0.05,      0.00,      1.09,      0.00 } },
 { {  5,  0, -1,  0 },     -0.00248, {      0.57,     -0.02,      0.22,     -0.04,     -0.96,      0.00 } },
 { {  5,  0,  0, -2 },      0.00142, {      0.07,      0.06,      0.00,      0.00,      0.55,      0.00 } },
 { {  5,  0,  0,  0 },     -0.00814, {     -0.25,      0.00,      0.07,      0.00,     -3.17,      0.00 } },
 { {  5,  0,  1, -2 },      0.00020, {      0.00,      0.00,      0.00,      0.00,      0.08,      0.00 } },
 { {  5,  0,  1,  0 },     -0.00089, {      0.00,      0.00,      0.00,      0.00,     -0.35,      0.00 } },
 { {  5,  0,  2,  0 },     -0.00006, {      0.00,      0.00,      0.00,      0.00,     -0.02,      0.00 } },
 { {  5,  1, -4,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  1, -3,  0 },      0.00038, {      0.04,      0.00,      0.02,      0.02,      0.15,      0.00 } },
 { {  5,  1, -2,  0 },     -0.00256, {     -0.07,      0.00,     -0.10,     -0.15,     -1.00,      0.00 } },
 { {  5,  1, -1, -2 },     -0.00015, {      0.00,      0.00,      0.00,      0.00,     -0.06,      0.00 } },
 { {  5,  1, -1,  0 },     -0.00311, {     -0.16,      0.00,     -0.08,     -0.19,     -1.21,      0.00 } },
 { {  5,  1,  0, -2 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  5,  1,  0,  0 },     -0.00111, {     -0.07,      0.00,     -0.03,     -0.07,     -0.43,      0.00 } },
 { {  5,  1,  1,  0 },     -0.00020, {      0.00,      0.00,      0.00,      0.00,     -0.08,      0.00 } },
 { {  5,  1,  2,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  5,  2, -1,  0 },      0.00007, {      0.00,      0.00,      0.00,      0.00,      0.03,      0.00 } },
 { {  5,  2,  0,  0 },      0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -4, -2,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -4, -1,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -3, -3,  0 },     -0.00012, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  6, -3, -2,  0 },     -0.00044, {     -0.03,      0.00,      0.00,     -0.08,      0.00,      0.00 } },
 { {  6, -3, -1,  0 },     -0.00043, {     -0.04,      0.00,      0.00,     -0.08,      0.00,      0.00 } },
 { {  6, -3,  0,  0 },     -0.00017, {      0.00,      0.00,      0.00,     -0.03,      0.00,      0.00 } },
 { {  6, -3,  1,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -4,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -3,  0 },     -0.00195, {     -0.12,      0.00,     -0.11,     -0.23,      0.00,      0.00 } },
 { {  6, -2, -2, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -2,  0 },     -0.00618, {     -0.44,      0.00,     -0.24,     -0.74,      0.00,      0.00 } },
 { {  6, -2, -1, -2 },     -0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2, -1,  0 },     -0.00541, {     -0.44,      0.00,     -0.14,     -0.65,      0.00,      0.00 } },
 { {  6, -2,  0, -2 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -2,  0,  0 },     -0.00194, {     -0.17,      0.00,     -0.04,     -0.23,      0.00,      0.00 } },
 { {  6, -2,  1,  0 },     -0.00035, {     -0.03,      0.00,      0.00,     -0.04,      0.00,      0.00 } },
 { {  6, -2,  2,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -4,  0 },     -0.00085, {     -0.06,      0.00,     -0.06,     -0.05,      0.00,      0.00 } },
 { {  6, -1, -3, -2 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -3,  0 },     -0.02436, {     -1.43,      0.00,     -1.33,     -1.45,      0.00,      0.00 } },
 { {  6, -1, -3,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -2, -2 },     -0.00041, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  6, -1, -2,  0 },     -0.06678, {     -4.61,      0.04,     -2.60,     -3.98,      0.00,      0.00 } },
 { {  6, -1, -2,  2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6, -1, -1, -2 },     -0.00121, {     -0.07,     -0.05,     -0.02,     -0.07,      0.00,      0.00 } },
 { {  6, -1, -1,  0 },     -0.05191, {     -4.10,      0.03,     -1.42,     -3.09,      0.00,      0.00 } },
 { {  6, -1,  0, -2 },     -0.00049, {     -0.04,     -0.02,      0.00,     -0.03,      0.00,      0.00 } },
 { {  6, -1,  0,  0 },     -0.01709, {     -1.47,      0.00,     -0.42,     -1.01,      0.00,      0.00 } },
 { {  6, -1,  1,  0 },     -0.00296, {     -0.26,      0.00,     -0.11,     -0.18,      0.00,      0.00 } },
 { {  6, -1,  2,  0 },     -0.00039, {     -0.04,      0.00,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  6, -1,  3,  0 },     -0.00005, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -5,  0 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  0 },     -0.00659, {     -0.45,      0.00,     -0.48,      0.00,      0.00,      0.00 } },
 { {  6,  0, -4,  2 },      0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3, -2 },     -0.00027, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -3,  0 },     -0.18316, {    -10.48,      0.12,    -10.02,      0.05,     -0.02,      0.00 } },
 { {  6,  0, -3,  2 },      0.00028, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2, -2 },     -0.00326, {     -0.13,     -0.14,     -0.12,      0.00,      0.00,      0.00 } },
 { {  6,  0, -2,  0 },     -0.42250, {    -28.36,      0.26,    -16.58,      0.19,      0.00,      0.00 } },
 { {  6,  0, -2,  2 },      0.00020, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1, -2 },     -0.00785, {     -0.44,     -0.35,     -0.13,      0.00,      0.00,      0.00 } },
 { {  6,  0, -1,  0 },     -0.28663, {    -21.86,      0.16,     -8.12,      0.18,     -0.03,      0.00 } },
 { {  6,  0, -1,  2 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -4 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0, -2 },     -0.00270, {     -0.19,     -0.12,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  0,  0 },     -0.08587, {     -7.10,      0.05,     -2.25,      0.07,     -0.02,      0.00 } },
 { {  6,  0,  1,  0 },     -0.01420, {     -1.20,      0.00,     -0.55,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  2,  0 },     -0.00182, {     -0.16,      0.00,     -0.10,      0.00,      0.00,      0.00 } },
 { {  6,  0,  3,  0 },     -0.00020, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  0,  4,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -4,  0 },     -0.00024, {     -0.04,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -3,  0 },      0.00278, {     -0.06,      0.00,      0.15,      0.17,      0.00,      0.00 } },
 { {  6,  1, -2, -2 },      0.00019, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1, -2,  0 },      0.00948, {      0.35,      0.00,      0.37,      0.57,      0.00,      0.00 } },
 { {  6,  1, -1, -2 },      0.00042, {      0.03,      0.00,      0.00,      0.02,      0.00,      0.00 } },
 { {  6,  1, -1,  0 },      0.00780, {      0.49,      0.00,      0.22,      0.47,      0.00,      0.00 } },
 { {  6,  1,  0, -2 },      0.00016, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  1,  0,  0 },      0.00272, {      0.21,      0.00,      0.07,      0.16,      0.00,      0.00 } },
 { {  6,  1,  1,  0 },      0.00052, {      0.04,      0.00,      0.02,      0.03,      0.00,      0.00 } },
 { {  6,  1,  2,  0 },      0.00008, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -4,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  6,  2, -3,  0 },      0.00034, {      0.03,      0.00,      0.02,      0.04,      0.00,      0.00 } },
 { {  6,  2, -2,  0 },      0.00044, {      0.04,      0.00,      0.00,      0.05,      0.00,      0.00 } },
 { {  6,  2, -1,  0 },      0.00014, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -4,  0 },      0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7, -1, -3,  0 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.04,      0.00 } },
 { {  7, -1, -2,  0 },      0.00006, {      0.00,      0.00,      0.00,      0.00,      0.02,      0.00 } },
 { {  7, -1, -1,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -4,  0 },      0.00015, {      0.00,      0.00,      0.00,      0.00,      0.06,      0.00 } },
 { {  7,  0, -3,  0 },      0.00066, {      0.05,      0.00,      0.04,      0.00,      0.26,      0.00 } },
 { {  7,  0, -2, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -2,  0 },      0.00032, {      0.04,      0.00,      0.00,      0.00,      0.13,      0.00 } },
 { {  7,  0, -1, -2 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  0, -1,  0 },     -0.00008, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  7,  0,  0,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  7,  1, -3,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  7,  1, -2,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,     -0.03,      0.00 } },
 { {  7,  1, -1,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -4,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -3,  0 },     -0.00013, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2, -2,  0 },     -0.00016, {      0.00,      0.00,      0.00,     -0.02,      0.00,      0.00 } },
 { {  8, -2, -1,  0 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -2,  0,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -4,  0 },     -0.00033, {     -0.03,      0.00,     -0.02,     -0.02,      0.00,      0.00 } },
 { {  8, -1, -3,  0 },     -0.00109, {     -0.10,      0.00,     -0.06,     -0.07,      0.00,      0.00 } },
 { {  8, -1, -2, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -2,  0 },     -0.00126, {     -0.12,      0.00,     -0.06,     -0.08,      0.00,      0.00 } },
 { {  8, -1, -1, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1, -1,  0 },     -0.00070, {     -0.08,      0.00,     -0.03,     -0.04,      0.00,      0.00 } },
 { {  8, -1,  0,  0 },     -0.00022, {     -0.02,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8, -1,  1,  0 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -5,  0 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -4,  0 },     -0.00184, {     -0.14,      0.00,     -0.13,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3, -2 },     -0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -3,  0 },     -0.00539, {     -0.46,      0.00,     -0.31,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2, -2 },     -0.00012, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -2,  0 },     -0.00559, {     -0.53,      0.00,     -0.25,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1, -2 },     -0.00009, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0, -1,  0 },     -0.00285, {     -0.29,      0.00,     -0.11,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0, -2 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  0,  0 },     -0.00083, {     -0.09,      0.00,     -0.03,      0.00,      0.00,      0.00 } },
 { {  8,  0,  1,  0 },     -0.00016, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  0,  2,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -4,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -3,  0 },      0.00015, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -2,  0 },      0.00019, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1, -1,  0 },      0.00011, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { {  8,  1,  0,  0 },      0.00004, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -3,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10, -1, -2,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -5,  0 },     -0.00002, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -4,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -3,  0 },     -0.00010, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -2,  0 },     -0.00007, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } },
 { { 10,  0, -1,  0 },     -0.00003, {      0.00,      0.00,      0.00,      0.00,      0.00,      0.00 } }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 347> g_ELP4
{ {
 {  0, {  0,  0,  0,  1 },    270.00000,   0.00003,     0.075 },
 {  0, {  0,  0,  0,  2 },      0.00000,   0.00037,     0.037 },
 {  0, {  0,  0,  1, -2 },    180.00000,   0.00480,     0.074 },
 {  0, {  0,  0,  1, -1 },     90.00000,   0.00001,     5.997 },
 {  0, {  0,  0,  2, -2 },      0.00000,   0.00085,     2.998 },
 {  0, {  0,  0,  2,  0 },    180.00000,   0.00025,     0.038 },
 {  0, {  0,  0,  3, -2 },      0.00000,   0.00003,     0.077 },
 {  0, {  0,  0,  3,  0 },    180.00000,   0.00002,     0.025 },
 {  0, {  0,  0,  4,  0 },    180.00000,   0.00001,     0.019 },
 {  0, {  0,  1, -2,  0 },    180.00000,   0.00003,     0.039 },
 {  0, {  0,  1, -1,  0 },    180.00000,   0.00048,     0.082 },
 {  0, {  0,  1, -1,  2 },    180.00000,   0.00001,     0.069 },
 {  0, {  0,  1,  0,  0 },    180.00000,   0.00047,     1.000 },
 {  0, {  0,  1,  1, -2 },    180.00000,   0.00005,     0.079 },
 {  0, {  0,  1,  1,  0 },    180.00000,   0.00019,     0.070 },
 {  0, {  0,  1,  2,  0 },    180.00000,   0.00001,     0.036 },
 {  0, {  0,  2, -1,  0 },    180.00000,   0.00003,     0.089 },
 {  0, {  1, -1, -1,  0 },    180.00000,   0.00001,     0.530 },
 {  0, {  1, -1,  0,  0 },    180.00000,   0.00015,     0.088 },
 {  0, {  1, -1,  1,  0 },    180.00000,   0.00001,     0.041 },
 {  0, {  1,  0, -2,  0 },    180.00000,   0.00001,     0.071 },
 {  0, {  1,  0, -1,  0 },    180.00000,   0.00004,     1.127 },
 {  0, {  1,  0,  0, -2 },      0.00000,   0.00001,     0.069 },
 {  0, {  1,  0,  0,  0 },    180.00000,   0.00100,     0.081 },
 {  0, {  1,  0,  1,  0 },    180.00000,   0.00006,     0.039 },
 {  0, {  1,  1, -1,  0 },      0.00000,   0.00019,     8.850 },
 {  0, {  1,  1,  0,  0 },    180.00000,   0.00070,     0.075 },
 {  0, {  1,  1,  1,  0 },    180.00000,   0.00005,     0.038 },
 {  0, {  2, -1, -2,  0 },    180.00000,   0.00002,     0.360 },
 {  0, {  2, -1, -1,  0 },      0.00000,   0.00061,     0.095 },
 {  0, {  2, -1,  0, -2 },      0.00000,   0.00002,     0.322 },
 {  0, {  2, -1,  0,  0 },      0.00000,   0.00001,     0.042 },
 {  0, {  2, -1,  1, -2 },      0.00000,   0.00004,     0.099 },
 {  0, {  2,  0, -3,  2 },    180.00000,   0.00001,     0.085 },
 {  0, {  2,  0, -2,  0 },    180.00000,   0.00077,     0.564 },
 {  0, {  2,  0, -2,  2 },      0.00000,   0.00006,     0.040 },
 {  0, {  2,  0, -1, -2 },    180.00000,   0.00005,     0.065 },
 {  0, {  2,  0, -1,  0 },      0.00000,   0.02206,     0.087 },
 {  0, {  2,  0,  0, -2 },      0.00000,   0.00064,     0.474 },
 {  0, {  2,  0,  0,  0 },      0.00000,   0.00102,     0.040 },
 {  0, {  2,  0,  1, -2 },      0.00000,   0.00089,     0.090 },
 {  0, {  2,  0,  1,  0 },      0.00000,   0.00004,     0.026 },
 {  0, {  2,  0,  2, -2 },      0.00000,   0.00007,     0.041 },
 {  0, {  2,  1, -2,  0 },      0.00000,   0.00006,     1.292 },
 {  0, {  2,  1, -1,  0 },    180.00000,   0.00006,     0.080 },
 {  0, {  2,  1,  0, -2 },    180.00000,   0.00002,     0.903 },
 {  0, {  2,  1,  0,  0 },    180.00000,   0.00001,     0.039 },
 {  0, {  2,  2, -1,  0 },      0.00000,   0.00009,     0.074 },
 {  0, {  3,  0, -1,  0 },    180.00000,   0.00001,     0.042 },
 {  0, {  4, -1, -2,  0 },      0.00000,   0.00002,     0.046 },
 {  0, {  4, -1, -1,  0 },      0.00000,   0.00002,     0.028 },
 {  0, {  4,  0, -2,  0 },      0.00000,   0.00026,     0.044 },
 {  0, {  4,  0, -1, -2 },      0.00000,   0.00002,     0.107 },
 {  0, {  4,  0, -1,  0 },      0.00000,   0.00019,     0.028 },
 {  0, {  4,  0,  0, -2 },      0.00000,   0.00002,     0.044 },
 {  0, {  4,  0,  0,  0 },      0.00000,   0.00002,     0.020 },
 {  0, {  4,  0,  1, -2 },      0.00000,   0.00001,     0.028 },
 {  1, { -6,  0,  0, -1 },      0.00100,   0.00001,     0.013 },
 {  1, { -6,  0,  1, -1 },      0.00100,   0.00003,     0.016 },
 {  1, { -6,  0,  2, -1 },      0.00068,   0.00003,     0.021 },
 {  1, { -6,  0,  3, -1 },      0.00036,   0.00001,     0.029 },
 {  1, { -4, -1,  0, -1 },    180.00103,   0.00001,     0.020 },
 {  1, { -4, -1,  1, -1 },    180.00081,   0.00002,     0.027 },
 {  1, { -4, -1,  1,  1 },    180.00072,   0.00002,     0.096 },
 {  1, { -4,  0, -2, -1 },      0.00099,   0.00002,     0.013 },
 {  1, { -4,  0, -2,  1 },    180.00050,   0.00001,     0.020 },
 {  1, { -4,  0, -1, -1 },      0.00109,   0.00014,     0.016 },
 {  1, { -4,  0, -1,  1 },    180.00050,   0.00009,     0.028 },
 {  1, { -4,  0,  0, -1 },      0.00100,   0.00075,     0.020 },
 {  1, { -4,  0,  0,  1 },    180.00017,   0.00011,     0.044 },
 {  1, { -4,  0,  1, -3 },    180.00093,   0.00002,     0.016 },
 {  1, { -4,  0,  1, -1 },      0.00080,   0.00141,     0.028 },
 {  1, { -4,  0,  1,  1 },      0.00091,   0.00022,     0.106 },
 {  1, { -4,  0,  2, -3 },    180.00091,   0.00001,     0.020 },
 {  1, { -4,  0,  2, -1 },      0.00029,   0.00065,     0.043 },
 {  1, { -4,  1, -1, -1 },      0.00098,   0.00002,     0.016 },
 {  1, { -4,  1,  0, -1 },      0.00099,   0.00010,     0.021 },
 {  1, { -4,  1,  1, -1 },      0.00080,   0.00017,     0.028 },
 {  1, { -4,  1,  1,  1 },      0.00082,   0.00002,     0.119 },
 {  1, { -4,  1,  2, -1 },      0.00030,   0.00006,     0.045 },
 {  1, { -4,  2,  0, -1 },      0.00094,   0.00001,     0.021 },
 {  1, { -4,  2,  1, -1 },      0.00098,   0.00001,     0.029 },
 {  1, { -3,  0,  0, -1 },      0.00082,   0.00004,     0.027 },
 {  1, { -3,  0,  0,  1 },    180.00073,   0.00023,     0.097 },
 {  1, { -3,  0,  1, -1 },    180.00068,   0.00006,     0.042 },
 {  1, { -3,  0,  1,  1 },    180.00072,   0.00003,     0.340 },
 {  1, { -3,  1,  0,  1 },    180.00072,   0.00002,     0.107 },
 {  1, { -2, -2,  1, -1 },    180.00031,   0.00002,     0.074 },
 {  1, { -2, -2,  1,  1 },    180.00029,   0.00001,     0.075 },
 {  1, { -2, -1, -2, -1 },    180.00123,   0.00002,     0.019 },
 {  1, { -2, -1, -2,  1 },      0.00056,   0.00001,     0.039 },
 {  1, { -2, -1, -1, -1 },    180.00119,   0.00012,     0.026 },
 {  1, { -2, -1, -1,  1 },      0.00065,   0.00009,     0.082 },
 {  1, { -2, -1,  0, -1 },    180.00109,   0.00053,     0.039 },
 {  1, { -2, -1,  0,  1 },    180.00076,   0.00123,     0.949 },
 {  1, { -2, -1,  1, -1 },    180.00047,   0.00053,     0.080 },
 {  1, { -2, -1,  2, -1 },    180.00148,   0.00006,     1.388 },
 {  1, { -2,  0, -3, -1 },      0.00122,   0.00009,     0.015 },
 {  1, { -2,  0, -3,  1 },    180.00044,   0.00005,     0.027 },
 {  1, { -2,  0, -2, -3 },    180.00093,   0.00001,     0.013 },
 {  1, { -2,  0, -2, -1 },      0.00119,   0.00097,     0.019 },
 {  1, { -2,  0, -2,  1 },    180.00050,   0.00065,     0.041 },
 {  1, { -2,  0, -1, -3 },    180.00101,   0.00008,     0.015 },
 {  1, { -2,  0, -1, -1 },      0.00113,   0.00913,     0.026 },
 {  1, { -2,  0, -1,  1 },    180.00049,   0.00887,     0.089 },
 {  1, { -2,  0,  0, -3 },    180.00092,   0.00037,     0.019 },
 {  1, { -2,  0,  0, -1 },      0.00100,   0.06569,     0.040 },
 {  1, { -2,  0,  0,  0 },     90.00000,   0.00006,     0.088 },
 {  1, { -2,  0,  0,  1 },      0.00076,   0.04746,     0.487 },
 {  1, { -2,  0,  0,  3 },    180.00077,   0.00013,     0.035 },
 {  1, { -2,  0,  1, -3 },    180.00073,   0.00044,     0.026 },
 {  1, { -2,  0,  1, -1 },      0.00029,   0.04962,     0.087 },
 {  1, { -2,  0,  1,  1 },      0.00045,   0.00026,     0.065 },
 {  1, { -2,  0,  1,  3 },    180.00076,   0.00001,     0.024 },
 {  1, { -2,  0,  2, -3 },    180.00043,   0.00001,     0.040 },
 {  1, { -2,  0,  2, -1 },    180.00208,   0.00249,     0.581 },
 {  1, { -2,  0,  2,  1 },    180.00070,   0.00048,     0.035 },
 {  1, { -2,  0,  3, -1 },    180.00155,   0.00045,     0.067 },
 {  1, { -2,  0,  3,  1 },    180.00073,   0.00008,     0.024 },
 {  1, { -2,  0,  4, -1 },    180.00145,   0.00005,     0.035 },
 {  1, { -2,  1, -2, -1 },      0.00118,   0.00008,     0.020 },
 {  1, { -2,  1, -2,  1 },    180.00049,   0.00004,     0.043 },
 {  1, { -2,  1, -1, -1 },      0.00111,   0.00072,     0.027 },
 {  1, { -2,  1, -1,  1 },    180.00052,   0.00049,     0.098 },
 {  1, { -2,  1,  0, -3 },    180.00091,   0.00003,     0.020 },
 {  1, { -2,  1,  0, -1 },      0.00098,   0.00482,     0.042 },
 {  1, { -2,  1,  0,  1 },      0.00075,   0.00187,     0.327 },
 {  1, { -2,  1,  1, -3 },    180.00089,   0.00002,     0.027 },
 {  1, { -2,  1,  1, -1 },      0.00031,   0.00225,     0.095 },
 {  1, { -2,  1,  1,  1 },    180.00078,   0.00005,     0.061 },
 {  1, { -2,  1,  2, -1 },    180.00225,   0.00009,     0.368 },
 {  1, { -2,  1,  2,  1 },    180.00071,   0.00003,     0.034 },
 {  1, { -2,  1,  3, -1 },    180.00164,   0.00001,     0.063 },
 {  1, { -2,  2, -1, -1 },      0.00111,   0.00004,     0.028 },
 {  1, { -2,  2, -1,  1 },    180.00054,   0.00002,     0.109 },
 {  1, { -2,  2,  0, -1 },      0.00097,   0.00024,     0.044 },
 {  1, { -2,  2,  0,  1 },      0.00072,   0.00006,     0.247 },
 {  1, { -2,  2,  1, -1 },      0.00031,   0.00008,     0.105 },
 {  1, { -2,  3,  0, -1 },      0.00094,   0.00001,     0.046 },
 {  1, { -1, -1, -1, -1 },      0.00122,   0.00004,     0.037 },
 {  1, { -1, -1,  0, -1 },      0.00107,   0.00019,     0.075 },
 {  1, { -1, -1,  0,  1 },      0.00000,   0.00001,     0.075 },
 {  1, { -1, -1,  1, -1 },    180.00149,   0.00001,     5.998 },
 {  1, { -1,  0, -2, -1 },    180.00131,   0.00003,     0.026 },
 {  1, { -1,  0, -1, -1 },    180.00131,   0.00023,     0.039 },
 {  1, { -1,  0, -1,  1 },    180.00122,   0.00002,     0.857 },
 {  1, { -1,  0,  0, -3 },      0.00093,   0.00001,     0.025 },
 {  1, { -1,  0,  0, -1 },    180.00134,   0.00075,     0.081 },
 {  1, { -1,  0,  0,  1 },    180.00067,   0.00056,     0.069 },
 {  1, { -1,  0,  1, -1 },      0.00127,   0.00028,     1.200 },
 {  1, { -1,  0,  1,  1 },    180.00072,   0.00002,     0.036 },
 {  1, { -1,  0,  2, -1 },      0.00129,   0.00006,     0.071 },
 {  1, { -1,  1,  0, -1 },    179.99990,   0.00004,     0.088 },
 {  1, { -1,  1,  0,  1 },    180.00072,   0.00001,     0.065 },
 {  1, {  0, -2, -1, -1 },    180.00138,   0.00002,     0.065 },
 {  1, {  0, -2,  0, -1 },      0.00072,   0.00003,     0.487 },
 {  1, {  0, -2,  1, -1 },      0.00116,   0.00007,     0.089 },
 {  1, {  0, -1, -3, -1 },    180.00133,   0.00003,     0.024 },
 {  1, {  0, -1, -2, -1 },    180.00133,   0.00029,     0.036 },
 {  1, {  0, -1, -2,  1 },      0.00025,   0.00003,     1.389 },
 {  1, {  0, -1, -1, -3 },      0.00090,   0.00001,     0.024 },
 {  1, {  0, -1, -1, -1 },    180.00141,   0.00187,     0.070 },
 {  1, {  0, -1, -1,  1 },    359.99928,   0.00008,     0.080 },
 {  1, {  0, -1,  0, -3 },      0.00092,   0.00001,     0.036 },
 {  1, {  0, -1,  0, -1 },      0.00067,   0.00282,     0.949 },
 {  1, {  0, -1,  0,  1 },      0.00045,   0.00010,     0.039 },
 {  1, {  0, -1,  1, -1 },      0.00118,   0.00387,     0.082 },
 {  1, {  0, -1,  1,  1 },      0.00070,   0.00026,     0.026 },
 {  1, {  0, -1,  2, -1 },      0.00122,   0.00047,     0.039 },
 {  1, {  0, -1,  2,  1 },      0.00070,   0.00004,     0.019 },
 {  1, {  0, -1,  3, -1 },      0.00122,   0.00005,     0.026 },
 {  1, {  0,  0, -5, -1 },      0.00127,   0.00001,     0.015 },
 {  1, {  0,  0, -4, -1 },      0.00129,   0.00016,     0.019 },
 {  1, {  0,  0, -3, -3 },    180.00113,   0.00003,     0.015 },
 {  1, {  0,  0, -3, -1 },      0.00129,   0.00228,     0.025 },
 {  1, {  0,  0, -3,  1 },    180.00029,   0.00013,     0.077 },
 {  1, {  0,  0, -2, -3 },    180.00109,   0.00032,     0.019 },
 {  1, {  0,  0, -2, -1 },      0.00128,   0.03279,     0.038 },
 {  1, {  0,  0, -2,  1 },    180.00024,   0.00387,     3.574 },
 {  1, {  0,  0, -1, -3 },    180.00102,   0.00261,     0.025 },
 {  1, {  0,  0, -1, -1 },      0.00127,   0.49141,     0.075 },
 {  1, {  0,  0, -1,  0 },     90.00000,   0.00010,     8.847 },
 {  1, {  0,  0, -1,  1 },      0.00051,   0.05036,     0.074 },
 {  1, {  0,  0, -1,  3 },    180.00060,   0.00017,     0.025 },
 {  1, {  0,  0,  0, -5 },      0.00086,   0.00003,     0.019 },
 {  1, {  0,  0,  0, -3 },    180.00086,   0.01492,     0.037 },
 {  1, {  0,  0,  0, -1 },      0.00094,   7.06304,    18.613 },
 {  1, {  0,  0,  0,  0 },    270.00000,   0.00027,     0.075 },
 {  1, {  0,  0,  0,  1 },      0.00071,   0.36061,     0.037 },
 {  1, {  0,  0,  0,  3 },    180.00071,   0.00074,     0.019 },
 {  1, {  0,  0,  1, -3 },    180.00054,   0.00050,     0.073 },
 {  1, {  0,  0,  1, -1 },      0.00127,   0.49331,     0.076 },
 {  1, {  0,  0,  1,  0 },    270.00000,   0.00002,     0.038 },
 {  1, {  0,  0,  1,  1 },      0.00070,   0.03838,     0.025 },
 {  1, {  0,  0,  1,  3 },    180.00074,   0.00015,     0.015 },
 {  1, {  0,  0,  2, -3 },    180.00061,   0.00004,     2.582 },
 {  1, {  0,  0,  2, -1 },      0.00126,   0.03402,     0.038 },
 {  1, {  0,  0,  2,  1 },      0.00069,   0.00331,     0.019 },
 {  1, {  0,  0,  2,  3 },    180.00073,   0.00002,     0.012 },
 {  1, {  0,  0,  3, -1 },      0.00125,   0.00242,     0.025 },
 {  1, {  0,  0,  3,  1 },      0.00067,   0.00027,     0.015 },
 {  1, {  0,  0,  4, -1 },      0.00125,   0.00017,     0.019 },
 {  1, {  0,  0,  4,  1 },      0.00067,   0.00002,     0.013 },
 {  1, {  0,  0,  5, -1 },      0.00127,   0.00001,     0.015 },
 {  1, {  0,  1, -3, -1 },      0.00133,   0.00004,     0.026 },
 {  1, {  0,  1, -2, -1 },      0.00135,   0.00037,     0.039 },
 {  1, {  0,  1, -2,  1 },    180.00050,   0.00003,     0.781 },
 {  1, {  0,  1, -1, -3 },    180.00090,   0.00001,     0.026 },
 {  1, {  0,  1, -1, -1 },      0.00143,   0.00249,     0.081 },
 {  1, {  0,  1, -1,  1 },      0.00272,   0.00002,     0.069 },
 {  1, {  0,  1,  0, -1 },    180.00079,   0.00240,     1.057 },
 {  1, {  0,  1,  0,  1 },    180.00076,   0.00034,     0.036 },
 {  1, {  0,  1,  1, -1 },    180.00116,   0.00292,     0.070 },
 {  1, {  0,  1,  1,  1 },    180.00073,   0.00022,     0.024 },
 {  1, {  0,  1,  2, -1 },    180.00121,   0.00037,     0.036 },
 {  1, {  0,  1,  2,  1 },    180.00071,   0.00004,     0.018 },
 {  1, {  0,  1,  3, -1 },    180.00122,   0.00004,     0.025 },
 {  1, {  0,  2, -1, -1 },      0.00146,   0.00004,     0.088 },
 {  1, {  0,  2,  0, -1 },    180.00072,   0.00003,     0.514 },
 {  1, {  0,  2,  1, -1 },    180.00119,   0.00003,     0.066 },
 {  1, {  1, -1,  0, -1 },    180.00080,   0.00003,     0.088 },
 {  1, {  1,  0, -2, -1 },      0.00182,   0.00003,     0.070 },
 {  1, {  1,  0, -1, -1 },    179.99929,   0.00011,     1.063 },
 {  1, {  1,  0, -1,  1 },    180.00052,   0.00002,     0.039 },
 {  1, {  1,  0,  0, -1 },    180.00088,   0.00342,     0.081 },
 {  1, {  1,  0,  0,  1 },    180.00072,   0.00021,     0.026 },
 {  1, {  1,  0,  1, -1 },    180.00104,   0.00043,     0.039 },
 {  1, {  1,  0,  1,  1 },    180.00071,   0.00004,     0.019 },
 {  1, {  1,  0,  2, -1 },    180.00110,   0.00004,     0.026 },
 {  1, {  1,  1, -1, -1 },    180.00007,   0.00006,    16.874 },
 {  1, {  1,  1,  0, -1 },      0.00059,   0.00064,     0.075 },
 {  1, {  1,  1,  0,  1 },      0.00072,   0.00003,     0.025 },
 {  1, {  1,  1,  1, -1 },      0.00085,   0.00007,     0.038 },
 {  1, {  2, -3,  0, -1 },      0.00094,   0.00001,     0.046 },
 {  1, {  2, -2, -1, -1 },      0.00040,   0.00010,     0.106 },
 {  1, {  2, -2, -1,  1 },      0.00072,   0.00001,     0.028 },
 {  1, {  2, -2,  0, -1 },      0.00090,   0.00031,     0.044 },
 {  1, {  2, -2,  0,  1 },      0.00072,   0.00002,     0.020 },
 {  1, {  2, -2,  1, -1 },      0.00104,   0.00005,     0.028 },
 {  1, {  2, -1, -3, -1 },    180.00162,   0.00001,     0.062 },
 {  1, {  2, -1, -2, -1 },    180.00296,   0.00006,     0.354 },
 {  1, {  2, -1, -2,  1 },      0.00051,   0.00003,     0.042 },
 {  1, {  2, -1, -1, -1 },      0.00041,   0.00288,     0.096 },
 {  1, {  2, -1, -1,  1 },      0.00069,   0.00037,     0.027 },
 {  1, {  2, -1,  0, -1 },      0.00091,   0.00646,     0.042 },
 {  1, {  2, -1,  0,  1 },      0.00072,   0.00031,     0.020 },
 {  1, {  2, -1,  1, -1 },      0.00103,   0.00091,     0.027 },
 {  1, {  2, -1,  1,  1 },      0.00069,   0.00006,     0.016 },
 {  1, {  2, -1,  2, -1 },      0.00106,   0.00010,     0.020 },
 {  1, {  2,  0, -4, -1 },    180.00153,   0.00005,     0.035 },
 {  1, {  2,  0, -3, -1 },    180.00173,   0.00037,     0.066 },
 {  1, {  2,  0, -3,  1 },    180.00003,   0.00005,     0.085 },
 {  1, {  2,  0, -2, -3 },      0.00121,   0.00002,     0.035 },
 {  1, {  2,  0, -2, -1 },    180.00322,   0.00142,     0.547 },
 {  1, {  2,  0, -2,  1 },      0.00052,   0.00067,     0.040 },
 {  1, {  2,  0, -1, -1 },      0.00042,   0.06456,     0.088 },
 {  1, {  2,  0, -1,  1 },      0.00071,   0.00815,     0.026 },
 {  1, {  2,  0, -1,  3 },    180.00036,   0.00004,     0.015 },
 {  1, {  2,  0,  0, -3 },    179.99827,   0.00006,     0.463 },
 {  1, {  2,  0,  0, -1 },      0.00090,   0.09642,     0.041 },
 {  1, {  2,  0,  0,  1 },      0.00070,   0.00480,     0.019 },
 {  1, {  2,  0,  0,  3 },    180.00038,   0.00002,     0.013 },
 {  1, {  2,  0,  1, -3 },    180.00114,   0.00014,     0.090 },
 {  1, {  2,  0,  1, -1 },      0.00102,   0.01234,     0.026 },
 {  1, {  2,  0,  1,  1 },      0.00070,   0.00080,     0.015 },
 {  1, {  2,  0,  2, -3 },    180.00118,   0.00002,     0.041 },
 {  1, {  2,  0,  2, -1 },      0.00107,   0.00126,     0.020 },
 {  1, {  2,  0,  2,  1 },      0.00073,   0.00009,     0.013 },
 {  1, {  2,  0,  3, -1 },      0.00111,   0.00012,     0.016 },
 {  1, {  2,  0,  3,  1 },      0.00062,   0.00001,     0.011 },
 {  1, {  2,  1, -1, -1 },    180.00004,   0.00017,     0.080 },
 {  1, {  2,  1, -1,  1 },    180.00073,   0.00006,     0.025 },
 {  1, {  2,  1,  0, -1 },    180.00088,   0.00113,     0.039 },
 {  1, {  2,  1,  0,  1 },    180.00073,   0.00005,     0.019 },
 {  1, {  2,  1,  1, -1 },    180.00098,   0.00021,     0.026 },
 {  1, {  2,  1,  1,  1 },    180.00099,   0.00001,     0.015 },
 {  1, {  2,  1,  2, -1 },    180.00105,   0.00003,     0.019 },
 {  1, {  2,  2, -1, -1 },    180.00033,   0.00005,     0.074 },
 {  1, {  3,  0, -2, -1 },    179.99961,   0.00001,     0.095 },
 {  1, {  3,  0, -1, -1 },    180.00070,   0.00013,     0.042 },
 {  1, {  3,  0, -1,  1 },    180.00072,   0.00001,     0.020 },
 {  1, {  3,  1, -1, -1 },      0.00073,   0.00001,     0.040 },
 {  1, {  3,  1,  0, -1 },      0.00088,   0.00001,     0.026 },
 {  1, {  4, -2, -1, -1 },      0.00091,   0.00002,     0.029 },
 {  1, {  4, -2,  0, -1 },      0.00094,   0.00001,     0.021 },
 {  1, {  4, -1, -2, -1 },      0.00041,   0.00008,     0.046 },
 {  1, {  4, -1, -2,  1 },      0.00092,   0.00001,     0.021 },
 {  1, {  4, -1, -1, -1 },      0.00077,   0.00024,     0.028 },
 {  1, {  4, -1, -1,  1 },      0.00071,   0.00002,     0.016 },
 {  1, {  4, -1,  0, -1 },      0.00090,   0.00015,     0.021 },
 {  1, {  4, -1,  1, -1 },      0.00091,   0.00003,     0.016 },
 {  1, {  4,  0, -3, -1 },    359.99818,   0.00001,     0.104 },
 {  1, {  4,  0, -3,  1 },      0.00056,   0.00001,     0.027 },
 {  1, {  4,  0, -2, -1 },      0.00042,   0.00088,     0.044 },
 {  1, {  4,  0, -2,  1 },      0.00070,   0.00014,     0.020 },
 {  1, {  4,  0, -1, -1 },      0.00078,   0.00217,     0.028 },
 {  1, {  4,  0, -1,  1 },      0.00071,   0.00016,     0.016 },
 {  1, {  4,  0,  0, -1 },      0.00090,   0.00115,     0.020 },
 {  1, {  4,  0,  0,  1 },      0.00072,   0.00006,     0.013 },
 {  1, {  4,  0,  1, -1 },      0.00098,   0.00021,     0.016 },
 {  1, {  4,  0,  1,  1 },      0.00038,   0.00001,     0.011 },
 {  1, {  4,  0,  2, -1 },      0.00093,   0.00003,     0.013 },
 {  1, {  4,  1, -1, -1 },    180.00077,   0.00004,     0.027 },
 {  1, {  4,  1,  0, -1 },    180.00089,   0.00003,     0.020 },
 {  1, {  6,  0, -3, -1 },      0.00037,   0.00001,     0.029 },
 {  1, {  6,  0, -2, -1 },      0.00067,   0.00004,     0.021 },
 {  1, {  6,  0, -1, -1 },      0.00090,   0.00004,     0.016 },
 {  1, {  6,  0,  0, -1 },      0.00090,   0.00002,     0.013 },
 {  2, { -2, -1,  0,  0 },      0.00000,   0.00001,     1.000 },
 {  2, { -2,  0, -1, -2 },    180.00000,   0.00004,     0.026 },
 {  2, { -2,  0, -1,  0 },      0.00000,   0.00013,     0.089 },
 {  2, { -2,  0,  0, -2 },    180.00000,   0.00027,     0.040 },
 {  2, { -2,  0,  0,  0 },    180.00000,   0.00043,     0.500 },
 {  2, { -2,  0,  1, -2 },    180.00000,   0.00025,     0.086 },
 {  2, { -2,  0,  1,  0 },    180.00000,   0.00002,     0.066 },
 {  2, { -2,  0,  2, -2 },      0.00000,   0.00001,     0.600 },
 {  2, { -2,  1,  0, -2 },    180.00000,   0.00002,     0.042 },
 {  2, { -2,  1,  0,  0 },    180.00000,   0.00003,     0.333 },
 {  2, { -2,  1,  1, -2 },    180.00000,   0.00001,     0.094 },
 {  2, {  0, -1,  0, -2 },    180.00000,   0.00003,     0.903 },
 {  2, {  0, -1,  1, -2 },    180.00000,   0.00002,     0.082 },
 {  2, {  0,  0, -2, -2 },    180.00000,   0.00016,     0.038 },
 {  2, {  0,  0, -2,  0 },      0.00000,   0.00009,     4.424 },
 {  2, {  0,  0, -1, -4 },      0.00000,   0.00001,     0.025 },
 {  2, {  0,  0, -1, -2 },    180.00000,   0.00239,     0.075 },
 {  2, {  0,  0, -1, -1 },    270.00000,   0.00001,    16.863 },
 {  2, {  0,  0, -1,  0 },    180.00000,   0.00080,     0.074 },
 {  2, {  0,  0,  0, -4 },      0.00000,   0.00007,     0.037 },
 {  2, {  0,  0,  0, -2 },    180.00000,   0.03638,     9.307 },
 {  2, {  0,  0,  0, -1 },     90.00000,   0.00003,     0.075 },
 {  2, {  0,  0,  0,  0 },    180.00000,   0.00239,     0.037 },
 {  2, {  0,  0,  1, -2 },    180.00000,   0.00241,     0.076 },
 {  2, {  0,  0,  1,  0 },    180.00000,   0.00025,     0.025 },
 {  2, {  0,  0,  2, -2 },    180.00000,   0.00017,     0.038 },
 {  2, {  0,  0,  2,  0 },    180.00000,   0.00002,     0.019 },
 {  2, {  0,  1,  0, -2 },      0.00000,   0.00002,     1.120 },
 {  2, {  0,  1,  1, -2 },      0.00000,   0.00001,     0.071 },
 {  2, {  1,  0,  0, -2 },      0.00000,   0.00002,     0.082 },
 {  2, {  1,  1,  0, -2 },    180.00000,   0.00004,     0.075 },
 {  2, {  2, -1, -1, -2 },    180.00000,   0.00002,     0.096 },
 {  2, {  2, -1,  0, -2 },    180.00000,   0.00004,     0.042 },
 {  2, {  2,  0, -1, -2 },    180.00000,   0.00039,     0.088 },
 {  2, {  2,  0, -1,  0 },    180.00000,   0.00006,     0.026 },
 {  2, {  2,  0,  0, -2 },    180.00000,   0.00057,     0.041 },
 {  2, {  2,  0,  0,  0 },    180.00000,   0.00003,     0.019 },
 {  2, {  2,  0,  1, -2 },    180.00000,   0.00007,     0.026 },
 {  2, {  4,  0, -1, -2 },    180.00000,   0.00001,     0.028 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 316> g_ELP5
{ {
 {  0, {  0,  0,  0,  3 },      0.00000,   0.00003,     0.025 },
 {  0, {  0,  0,  1, -3 },    180.00000,   0.00021,     0.037 },
 {  0, {  0,  0,  1, -1 },      0.00000,   0.00056,     5.997 },
 {  0, {  0,  0,  1,  1 },      0.00000,   0.00047,     0.037 },
 {  0, {  0,  0,  2, -3 },      0.00000,   0.00004,     0.073 },
 {  0, {  0,  0,  2, -1 },    180.00000,   0.00144,     0.076 },
 {  0, {  0,  0,  2,  1 },      0.00000,   0.00003,     0.025 },
 {  0, {  0,  0,  3, -1 },    180.00000,   0.00007,     0.038 },
 {  0, {  0,  1, -1, -1 },    180.00000,   0.00002,     0.039 },
 {  0, {  0,  1, -1,  1 },    180.00000,   0.00002,     0.857 },
 {  0, {  0,  1,  0, -1 },    180.00000,   0.00004,     0.081 },
 {  0, {  0,  1,  0,  1 },    180.00000,   0.00004,     0.069 },
 {  0, {  0,  1,  1, -1 },    180.00000,   0.00001,     1.200 },
 {  0, {  0,  1,  1,  1 },    180.00000,   0.00002,     0.036 },
 {  0, {  1,  0,  0, -1 },    180.00000,   0.00004,     0.949 },
 {  0, {  1,  0,  0,  1 },    180.00000,   0.00005,     0.039 },
 {  0, {  1,  1, -1, -1 },      0.00000,   0.00001,     0.075 },
 {  0, {  1,  1, -1,  1 },      0.00000,   0.00001,     0.074 },
 {  0, {  1,  1,  0, -1 },    180.00000,   0.00003,    18.600 },
 {  0, {  1,  1,  0,  1 },    180.00000,   0.00003,     0.037 },
 {  0, {  2, -1, -1, -1 },      0.00000,   0.00003,     0.340 },
 {  0, {  2, -1, -1,  1 },      0.00000,   0.00004,     0.042 },
 {  0, {  2, -1,  0, -1 },    180.00000,   0.00006,     0.097 },
 {  0, {  2,  0, -2, -1 },    180.00000,   0.00005,     0.066 },
 {  0, {  2,  0, -2,  1 },    180.00000,   0.00021,     0.086 },
 {  0, {  2,  0, -1, -1 },      0.00000,   0.00126,     0.515 },
 {  0, {  2,  0, -1,  1 },      0.00000,   0.00108,     0.040 },
 {  0, {  2,  0,  0, -3 },      0.00000,   0.00004,     0.064 },
 {  0, {  2,  0,  0, -1 },    180.00000,   0.00217,     0.088 },
 {  0, {  2,  0,  0,  1 },      0.00000,   0.00013,     0.026 },
 {  0, {  2,  0,  1, -3 },      0.00000,   0.00003,     0.440 },
 {  0, {  2,  0,  1, -1 },    180.00000,   0.00008,     0.041 },
 {  0, {  2,  0,  2, -1 },    180.00000,   0.00002,     0.026 },
 {  0, {  2,  1, -1, -1 },    180.00000,   0.00001,     1.063 },
 {  0, {  2,  1,  0, -1 },      0.00000,   0.00009,     0.081 },
 {  0, {  2,  2,  0, -1 },      0.00000,   0.00001,     0.075 },
 {  0, {  4,  0, -2,  1 },      0.00000,   0.00003,     0.027 },
 {  0, {  4,  0, -1, -1 },      0.00000,   0.00001,     0.044 },
 {  0, {  4,  0, -1,  1 },      0.00000,   0.00002,     0.020 },
 {  0, {  4,  0,  0, -1 },    180.00000,   0.00001,     0.028 },
 {  1, { -6,  0,  0,  0 },      0.00076,   0.00002,     0.016 },
 {  1, { -6,  0,  1,  0 },      0.00075,   0.00005,     0.021 },
 {  1, { -6,  0,  2,  0 },      0.00075,   0.00004,     0.029 },
 {  1, { -4, -1,  0,  0 },    180.00075,   0.00006,     0.027 },
 {  1, { -4, -1,  1,  0 },    180.00074,   0.00008,     0.042 },
 {  1, { -4,  0, -2,  0 },      0.00076,   0.00003,     0.016 },
 {  1, { -4,  0, -1, -2 },      0.00095,   0.00002,     0.013 },
 {  1, { -4,  0, -1,  0 },      0.00075,   0.00025,     0.020 },
 {  1, { -4,  0, -1,  2 },    180.00046,   0.00001,     0.044 },
 {  1, { -4,  0,  0, -2 },      0.00094,   0.00009,     0.016 },
 {  1, { -4,  0,  0,  0 },      0.00073,   0.00188,     0.028 },
 {  1, { -4,  0,  0,  2 },      0.00072,   0.00009,     0.108 },
 {  1, { -4,  0,  1, -2 },      0.00083,   0.00017,     0.020 },
 {  1, { -4,  0,  1,  0 },      0.00071,   0.00322,     0.044 },
 {  1, { -4,  0,  2, -2 },      0.00055,   0.00009,     0.027 },
 {  1, { -4,  0,  2,  0 },      0.00070,   0.00029,     0.104 },
 {  1, { -4,  1, -1,  0 },      0.00075,   0.00003,     0.021 },
 {  1, { -4,  1,  0, -2 },      0.00094,   0.00001,     0.016 },
 {  1, { -4,  1,  0,  0 },      0.00075,   0.00021,     0.028 },
 {  1, { -4,  1,  1, -2 },      0.00094,   0.00002,     0.021 },
 {  1, { -4,  1,  1,  0 },      0.00072,   0.00029,     0.046 },
 {  1, { -4,  1,  2, -2 },      0.00093,   0.00001,     0.028 },
 {  1, { -4,  1,  2,  0 },      0.00072,   0.00002,     0.116 },
 {  1, { -4,  2,  0,  0 },      0.00072,   0.00001,     0.029 },
 {  1, { -4,  2,  1,  0 },      0.00072,   0.00002,     0.048 },
 {  1, { -3, -1,  0,  0 },      0.00072,   0.00002,     0.040 },
 {  1, { -3,  0, -1,  0 },    180.00072,   0.00002,     0.027 },
 {  1, { -3,  0,  0,  0 },    180.00074,   0.00017,     0.042 },
 {  1, { -3,  0,  0,  2 },    180.00072,   0.00001,     0.322 },
 {  1, { -3,  0,  1,  0 },    180.00073,   0.00014,     0.095 },
 {  1, { -3,  0,  2,  0 },    180.00072,   0.00002,     0.360 },
 {  1, { -3,  1,  0,  0 },    180.00072,   0.00001,     0.044 },
 {  1, { -2, -2,  0,  0 },    180.00037,   0.00013,     0.075 },
 {  1, { -2, -2,  1,  0 },    180.00072,   0.00005,     8.853 },
 {  1, { -2, -1, -2,  0 },    180.00075,   0.00003,     0.026 },
 {  1, { -2, -1, -1, -2 },    180.00095,   0.00001,     0.019 },
 {  1, { -2, -1, -1,  0 },    180.00073,   0.00041,     0.039 },
 {  1, { -2, -1,  0, -2 },    180.00094,   0.00006,     0.026 },
 {  1, { -2, -1,  0,  0 },    180.00071,   0.00589,     0.081 },
 {  1, { -2, -1,  0,  2 },    180.00079,   0.00007,     0.069 },
 {  1, { -2, -1,  1, -2 },    180.00066,   0.00005,     0.039 },
 {  1, { -2, -1,  1,  0 },    180.00074,   0.00034,     1.127 },
 {  1, { -2, -1,  2,  0 },      0.00072,   0.00003,     0.071 },
 {  1, { -2,  0, -3, -2 },      0.00094,   0.00001,     0.013 },
 {  1, { -2,  0, -3,  0 },      0.00077,   0.00007,     0.020 },
 {  1, { -2,  0, -2, -2 },      0.00111,   0.00014,     0.015 },
 {  1, { -2,  0, -2,  0 },      0.00077,   0.00110,     0.026 },
 {  1, { -2,  0, -2,  2 },    180.00047,   0.00002,     0.091 },
 {  1, { -2,  0, -1, -2 },      0.00104,   0.00104,     0.019 },
 {  1, { -2,  0, -1,  0 },      0.00075,   0.01647,     0.041 },
 {  1, { -2,  0, -1,  2 },    180.00059,   0.00048,     0.450 },
 {  1, { -2,  0,  0, -4 },    180.00090,   0.00001,     0.015 },
 {  1, { -2,  0,  0, -2 },      0.00093,   0.00562,     0.026 },
 {  1, { -2,  0,  0,  0 },      0.00072,   0.29854,     0.088 },
 {  1, { -2,  0,  0,  2 },      0.00077,   0.00273,     0.065 },
 {  1, { -2,  0,  1, -4 },    180.00088,   0.00002,     0.019 },
 {  1, { -2,  0,  1, -2 },      0.00065,   0.00562,     0.040 },
 {  1, { -2,  0,  1,  0 },      0.00071,   0.07332,     0.530 },
 {  1, { -2,  0,  1,  2 },      0.00072,   0.00043,     0.035 },
 {  1, { -2,  0,  2, -2 },    179.99951,   0.00002,     0.085 },
 {  1, { -2,  0,  2,  0 },      0.00066,   0.00672,     0.066 },
 {  1, { -2,  0,  2,  2 },      0.00072,   0.00003,     0.024 },
 {  1, { -2,  0,  3,  0 },      0.00063,   0.00063,     0.035 },
 {  1, { -2,  0,  4,  0 },      0.00061,   0.00005,     0.024 },
 {  1, { -2,  1, -2, -2 },      0.00094,   0.00001,     0.016 },
 {  1, { -2,  1, -2,  0 },      0.00075,   0.00006,     0.027 },
 {  1, { -2,  1, -1, -2 },      0.00104,   0.00008,     0.020 },
 {  1, { -2,  1, -1,  0 },      0.00075,   0.00086,     0.042 },
 {  1, { -2,  1, -1,  2 },    180.00060,   0.00003,     0.311 },
 {  1, { -2,  1,  0, -2 },      0.00094,   0.00038,     0.027 },
 {  1, { -2,  1,  0,  0 },      0.00072,   0.01404,     0.096 },
 {  1, { -2,  1,  0,  2 },      0.00076,   0.00011,     0.061 },
 {  1, { -2,  1,  1, -2 },      0.00065,   0.00025,     0.042 },
 {  1, { -2,  1,  1,  0 },      0.00071,   0.00328,     0.346 },
 {  1, { -2,  1,  1,  2 },      0.00073,   0.00001,     0.034 },
 {  1, { -2,  1,  2,  0 },      0.00068,   0.00028,     0.062 },
 {  1, { -2,  1,  3,  0 },      0.00072,   0.00002,     0.034 },
 {  1, { -2,  2, -1,  0 },      0.00074,   0.00003,     0.044 },
 {  1, { -2,  2,  0, -2 },      0.00093,   0.00002,     0.028 },
 {  1, { -2,  2,  0,  0 },      0.00073,   0.00051,     0.107 },
 {  1, { -2,  2,  1, -2 },      0.00091,   0.00001,     0.044 },
 {  1, { -2,  2,  1,  0 },      0.00072,   0.00012,     0.257 },
 {  1, { -2,  2,  2,  0 },      0.00072,   0.00001,     0.058 },
 {  1, { -2,  3,  0,  0 },      0.00072,   0.00002,     0.119 },
 {  1, { -1, -1, -1,  0 },      0.00072,   0.00002,     0.075 },
 {  1, { -1, -1,  0, -2 },      0.00094,   0.00002,     0.037 },
 {  1, { -1, -1,  0,  0 },      0.00072,   0.00036, 25770.018 },
 {  1, { -1,  0, -2,  0 },    180.00072,   0.00002,     0.039 },
 {  1, { -1,  0, -1, -2 },    180.00112,   0.00003,     0.026 },
 {  1, { -1,  0, -1,  0 },    180.00072,   0.00029,     0.082 },
 {  1, { -1,  0,  0, -2 },    180.00099,   0.00013,     0.039 },
 {  1, { -1,  0,  0,  0 },    180.00073,   0.00212,     1.000 },
 {  1, { -1,  0,  0,  2 },    180.00069,   0.00005,     0.036 },
 {  1, { -1,  0,  1,  0 },    180.00074,   0.00017,     0.070 },
 {  1, { -1,  0,  2,  0 },    180.00071,   0.00004,     0.036 },
 {  1, {  0, -2, -1,  0 },    180.00072,   0.00003,     0.530 },
 {  1, {  0, -2,  1,  0 },    180.00071,   0.00005,     0.041 },
 {  1, {  0, -1, -3,  0 },    180.00072,   0.00001,     0.036 },
 {  1, {  0, -1, -2, -2 },    180.00119,   0.00004,     0.024 },
 {  1, {  0, -1, -2,  0 },    180.00080,   0.00014,     0.071 },
 {  1, {  0, -1, -1, -2 },    180.00114,   0.00018,     0.036 },
 {  1, {  0, -1, -1,  0 },    180.00076,   0.00227,     1.127 },
 {  1, {  0, -1,  0, -2 },    180.00192,   0.00002,     0.069 },
 {  1, {  0, -1,  0,  0 },    180.00058,   0.00252,     0.081 },
 {  1, {  0, -1,  1, -2 },      0.00202,   0.00004,     0.819 },
 {  1, {  0, -1,  1,  0 },    180.00068,   0.00276,     0.039 },
 {  1, {  0, -1,  2,  0 },    180.00065,   0.00031,     0.026 },
 {  1, {  0, -1,  3,  0 },    180.00069,   0.00003,     0.019 },
 {  1, {  0,  0, -4, -2 },      0.00121,   0.00003,     0.015 },
 {  1, {  0,  0, -4,  0 },      0.00064,   0.00005,     0.025 },
 {  1, {  0,  0, -3, -2 },      0.00119,   0.00033,     0.019 },
 {  1, {  0,  0, -3,  0 },      0.00069,   0.00091,     0.038 },
 {  1, {  0,  0, -2, -4 },    180.00089,   0.00001,     0.015 },
 {  1, {  0,  0, -2, -2 },      0.00116,   0.00377,     0.025 },
 {  1, {  0,  0, -2,  0 },      0.00067,   0.01745,     0.076 },
 {  1, {  0,  0, -2,  2 },    180.00038,   0.00033,     0.073 },
 {  1, {  0,  0, -1, -4 },    180.00098,   0.00008,     0.019 },
 {  1, {  0,  0, -1, -2 },      0.00109,   0.03977,     0.037 },
 {  1, {  0,  0, -1,  0 },      0.00075,   0.45586,     8.847 },
 {  1, {  0,  0, -1,  2 },      0.00062,   0.00437,     0.037 },
 {  1, {  0,  0,  0, -4 },    180.00086,   0.00034,     0.025 },
 {  1, {  0,  0,  0, -2 },      0.00086,   0.32622,     0.074 },
 {  1, {  0,  0,  0, -1 },    270.00000,   0.00001,    18.613 },
 {  1, {  0,  0,  0,  0 },    180.00071,   8.04508,     0.075 },
 {  1, {  0,  0,  0,  1 },    270.00000,   0.00001,     0.037 },
 {  1, {  0,  0,  0,  2 },      0.00071,   0.00834,     0.025 },
 {  1, {  0,  0,  0,  4 },    180.00032,   0.00002,     0.015 },
 {  1, {  0,  0,  1, -4 },    180.00067,   0.00009,     0.037 },
 {  1, {  0,  0,  1, -2 },      0.00464,   0.00239,     4.536 },
 {  1, {  0,  0,  1,  0 },    180.00069,   0.41571,     0.038 },
 {  1, {  0,  0,  1,  2 },      0.00070,   0.00133,     0.019 },
 {  1, {  0,  0,  2, -2 },      0.00147,   0.00079,     0.077 },
 {  1, {  0,  0,  2,  0 },    180.00066,   0.02403,     0.025 },
 {  1, {  0,  0,  2,  2 },      0.00071,   0.00015,     0.015 },
 {  1, {  0,  0,  3, -2 },      0.00137,   0.00007,     0.038 },
 {  1, {  0,  0,  3,  0 },    180.00062,   0.00146,     0.019 },
 {  1, {  0,  0,  3,  2 },      0.00075,   0.00002,     0.013 },
 {  1, {  0,  0,  4,  0 },    180.00058,   0.00009,     0.015 },
 {  1, {  0,  1, -3,  0 },      0.00072,   0.00001,     0.039 },
 {  1, {  0,  1, -2, -2 },      0.00119,   0.00004,     0.026 },
 {  1, {  0,  1, -2,  0 },      0.00077,   0.00014,     0.082 },
 {  1, {  0,  1, -1, -2 },      0.00114,   0.00023,     0.039 },
 {  1, {  0,  1, -1,  0 },      0.00074,   0.00257,     0.898 },
 {  1, {  0,  1, -1,  2 },      0.00087,   0.00001,     0.036 },
 {  1, {  0,  1,  0, -2 },      0.00071,   0.00003,     0.080 },
 {  1, {  0,  1,  0,  0 },      0.00095,   0.00206,     0.070 },
 {  1, {  0,  1,  0,  2 },    180.00072,   0.00002,     0.024 },
 {  1, {  0,  1,  1, -2 },    180.00217,   0.00003,     1.283 },
 {  1, {  0,  1,  1,  0 },      0.00070,   0.00212,     0.036 },
 {  1, {  0,  1,  2,  0 },      0.00065,   0.00025,     0.024 },
 {  1, {  0,  1,  3,  0 },      0.00069,   0.00002,     0.018 },
 {  1, {  0,  2, -1,  0 },      0.00072,   0.00004,     0.473 },
 {  1, {  0,  2,  0,  0 },      0.00088,   0.00002,     0.065 },
 {  1, {  0,  2,  1,  0 },      0.00072,   0.00002,     0.035 },
 {  1, {  1,  0, -2,  0 },      0.00072,   0.00004,     1.292 },
 {  1, {  1,  0, -1,  0 },    180.00137,   0.00003,     0.080 },
 {  1, {  1,  0,  0, -2 },    180.00102,   0.00003,     0.903 },
 {  1, {  1,  0,  0,  0 },      0.00071,   0.00218,     0.039 },
 {  1, {  1,  0,  1, -2 },    180.00131,   0.00001,     0.082 },
 {  1, {  1,  0,  1,  0 },      0.00067,   0.00026,     0.026 },
 {  1, {  1,  0,  2,  0 },      0.00068,   0.00002,     0.019 },
 {  1, {  1,  1, -1,  0 },      0.00022,   0.00008,     0.074 },
 {  1, {  1,  1,  0, -2 },      0.00028,   0.00001,     9.303 },
 {  1, {  1,  1,  0,  0 },    180.00073,   0.00032,     0.037 },
 {  1, {  1,  1,  1,  0 },    180.00070,   0.00004,     0.025 },
 {  1, {  2, -2, -1,  0 },    180.00071,   0.00013,     0.044 },
 {  1, {  2, -2,  0, -2 },      0.00102,   0.00002,     0.108 },
 {  1, {  2, -2,  0,  0 },    180.00070,   0.00015,     0.028 },
 {  1, {  2, -2,  1,  0 },    180.00069,   0.00002,     0.020 },
 {  1, {  2, -1, -2, -2 },    180.00146,   0.00002,     0.062 },
 {  1, {  2, -1, -2,  0 },      0.00091,   0.00002,     0.094 },
 {  1, {  2, -1, -1, -2 },    180.00253,   0.00003,     0.334 },
 {  1, {  2, -1, -1,  0 },    180.00073,   0.00371,     0.042 },
 {  1, {  2, -1, -1,  2 },      0.00072,   0.00002,     0.020 },
 {  1, {  2, -1,  0, -2 },      0.00101,   0.00049,     0.097 },
 {  1, {  2, -1,  0,  0 },    180.00070,   0.00315,     0.027 },
 {  1, {  2, -1,  0,  2 },      0.00073,   0.00001,     0.016 },
 {  1, {  2, -1,  1, -2 },      0.00112,   0.00008,     0.043 },
 {  1, {  2, -1,  1,  0 },    180.00066,   0.00043,     0.020 },
 {  1, {  2, -1,  2,  0 },    180.00067,   0.00004,     0.016 },
 {  1, {  2,  0, -4,  0 },    180.00072,   0.00001,     0.067 },
 {  1, {  2,  0, -3, -2 },    180.00138,   0.00007,     0.035 },
 {  1, {  2,  0, -3,  0 },    180.00074,   0.00013,     0.602 },
 {  1, {  2,  0, -2, -2 },    180.00150,   0.00038,     0.066 },
 {  1, {  2,  0, -2,  0 },      0.00164,   0.00033,     0.086 },
 {  1, {  2,  0, -2,  2 },      0.00062,   0.00010,     0.026 },
 {  1, {  2,  0, -1, -4 },    180.00083,   0.00001,     0.035 },
 {  1, {  2,  0, -1, -2 },    180.00240,   0.00066,     0.501 },
 {  1, {  2,  0, -1,  0 },    180.00073,   0.08350,     0.040 },
 {  1, {  2,  0, -1,  2 },      0.00069,   0.00033,     0.019 },
 {  1, {  2,  0,  0, -4 },    180.00072,   0.00004,     0.064 },
 {  1, {  2,  0,  0, -2 },      0.00102,   0.01005,     0.089 },
 {  1, {  2,  0,  0,  0 },    180.00070,   0.04638,     0.026 },
 {  1, {  2,  0,  0,  2 },      0.00073,   0.00018,     0.015 },
 {  1, {  2,  0,  1, -2 },      0.00112,   0.00139,     0.041 },
 {  1, {  2,  0,  1,  0 },    180.00067,   0.00563,     0.019 },
 {  1, {  2,  0,  1,  2 },      0.00072,   0.00004,     0.013 },
 {  1, {  2,  0,  2, -2 },      0.00115,   0.00014,     0.026 },
 {  1, {  2,  0,  2,  0 },    180.00062,   0.00054,     0.015 },
 {  1, {  2,  0,  3, -2 },      0.00093,   0.00001,     0.020 },
 {  1, {  2,  0,  3,  0 },    180.00066,   0.00005,     0.013 },
 {  1, {  2,  1, -2,  0 },    180.00072,   0.00002,     0.079 },
 {  1, {  2,  1, -1, -2 },      0.00121,   0.00001,     1.006 },
 {  1, {  2,  1, -1,  0 },      0.00074,   0.00055,     0.039 },
 {  1, {  2,  1,  0, -2 },    180.00103,   0.00019,     0.082 },
 {  1, {  2,  1,  0,  0 },      0.00071,   0.00049,     0.026 },
 {  1, {  2,  1,  1, -2 },    180.00114,   0.00003,     0.039 },
 {  1, {  2,  1,  1,  0 },      0.00069,   0.00009,     0.019 },
 {  1, {  2,  1,  2,  0 },      0.00067,   0.00001,     0.015 },
 {  1, {  2,  2, -1,  0 },      0.00071,   0.00005,     0.037 },
 {  1, {  3,  0, -2,  0 },      0.00072,   0.00002,     0.042 },
 {  1, {  3,  0, -1,  0 },      0.00070,   0.00008,     0.027 },
 {  1, {  3,  0,  0,  0 },      0.00072,   0.00001,     0.020 },
 {  1, {  4, -2, -1,  0 },    180.00072,   0.00001,     0.021 },
 {  1, {  4, -1, -2,  0 },    180.00070,   0.00008,     0.028 },
 {  1, {  4, -1, -1, -2 },      0.00074,   0.00002,     0.046 },
 {  1, {  4, -1, -1,  0 },    180.00069,   0.00013,     0.021 },
 {  1, {  4, -1,  0, -2 },      0.00096,   0.00002,     0.029 },
 {  1, {  4, -1,  0,  0 },    180.00067,   0.00006,     0.016 },
 {  1, {  4, -1,  1,  0 },    180.00066,   0.00001,     0.013 },
 {  1, {  4,  0, -3,  0 },    180.00034,   0.00001,     0.043 },
 {  1, {  4,  0, -2,  0 },    180.00074,   0.00097,     0.028 },
 {  1, {  4,  0, -1, -2 },      0.00074,   0.00019,     0.044 },
 {  1, {  4,  0, -1,  0 },    180.00071,   0.00113,     0.020 },
 {  1, {  4,  0, -1,  2 },      0.00056,   0.00001,     0.013 },
 {  1, {  4,  0,  0, -2 },      0.00096,   0.00019,     0.028 },
 {  1, {  4,  0,  0,  0 },    180.00067,   0.00043,     0.016 },
 {  1, {  4,  0,  1, -2 },      0.00103,   0.00004,     0.020 },
 {  1, {  4,  0,  1,  0 },    180.00066,   0.00007,     0.013 },
 {  1, {  4,  1, -2,  0 },      0.00072,   0.00001,     0.027 },
 {  1, {  4,  1, -1,  0 },      0.00069,   0.00002,     0.020 },
 {  1, {  6,  0, -3,  0 },    180.00069,   0.00001,     0.021 },
 {  1, {  6,  0, -2,  0 },    180.00067,   0.00002,     0.016 },
 {  1, {  6,  0, -1,  0 },    180.00066,   0.00001,     0.013 },
 {  2, { -4,  0,  0, -1 },    180.00000,   0.00002,     0.028 },
 {  2, { -4,  0,  1, -1 },    180.00000,   0.00003,     0.044 },
 {  2, { -2, -1,  0, -1 },      0.00000,   0.00006,     0.081 },
 {  2, { -2,  0, -2, -1 },    180.00000,   0.00001,     0.026 },
 {  2, { -2,  0, -1, -1 },    180.00000,   0.00016,     0.041 },
 {  2, { -2,  0,  0, -3 },    180.00000,   0.00003,     0.026 },
 {  2, { -2,  0,  0, -1 },    180.00000,   0.00293,     0.088 },
 {  2, { -2,  0,  0,  1 },    180.00000,   0.00014,     0.065 },
 {  2, { -2,  0,  1, -3 },    180.00000,   0.00003,     0.040 },
 {  2, { -2,  0,  1, -1 },    180.00000,   0.00072,     0.545 },
 {  2, { -2,  0,  1,  1 },    180.00000,   0.00001,     0.035 },
 {  2, { -2,  0,  2, -1 },    180.00000,   0.00007,     0.066 },
 {  2, { -2,  1,  0, -1 },    180.00000,   0.00014,     0.096 },
 {  2, { -2,  1,  1, -1 },    180.00000,   0.00003,     0.353 },
 {  2, { -1,  0,  0, -1 },      0.00000,   0.00002,     1.057 },
 {  2, {  0, -1, -1, -1 },      0.00000,   0.00002,     1.063 },
 {  2, {  0, -1,  0, -1 },      0.00000,   0.00003,     0.081 },
 {  2, {  0, -1,  1, -1 },      0.00000,   0.00003,     0.039 },
 {  2, {  0,  0, -3, -1 },    180.00000,   0.00001,     0.038 },
 {  2, {  0,  0, -2, -3 },    180.00000,   0.00002,     0.025 },
 {  2, {  0,  0, -2, -1 },    180.00000,   0.00020,     0.076 },
 {  2, {  0,  0, -2,  1 },    180.00000,   0.00005,     0.073 },
 {  2, {  0,  0, -1, -3 },    180.00000,   0.00019,     0.037 },
 {  2, {  0,  0, -1, -1 },    180.00000,   0.00441,    16.863 },
 {  2, {  0,  0, -1,  1 },    180.00000,   0.00006,     0.037 },
 {  2, {  0,  0,  0, -3 },    180.00000,   0.00156,     0.074 },
 {  2, {  0,  0,  0, -1 },      0.00000,   0.08042,     0.075 },
 {  2, {  0,  0,  0,  1 },    180.00000,   0.00003,     0.025 },
 {  2, {  0,  0,  1, -1 },      0.00000,   0.00425,     0.038 },
 {  2, {  0,  0,  2, -1 },      0.00000,   0.00026,     0.025 },
 {  2, {  0,  0,  3, -1 },      0.00000,   0.00002,     0.019 },
 {  2, {  0,  1, -1, -1 },    180.00000,   0.00002,     0.944 },
 {  2, {  0,  1,  1, -1 },    180.00000,   0.00002,     0.036 },
 {  2, {  1,  0,  0, -1 },    180.00000,   0.00002,     0.039 },
 {  2, {  1,  1, -1, -1 },    180.00000,   0.00002,     0.074 },
 {  2, {  2, -1, -1, -1 },      0.00000,   0.00004,     0.042 },
 {  2, {  2, -1,  0, -1 },      0.00000,   0.00003,     0.027 },
 {  2, {  2,  0, -1, -1 },      0.00000,   0.00085,     0.040 },
 {  2, {  2,  0,  0, -3 },    180.00000,   0.00005,     0.089 },
 {  2, {  2,  0,  0, -1 },      0.00000,   0.00049,     0.026 },
 {  2, {  2,  0,  1, -1 },      0.00000,   0.00006,     0.019 },
 {  2, {  4,  0, -1, -1 },      0.00000,   0.00001,     0.020 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 237> g_ELP6
{ {
 {  0, {  0,  0,  0,  0 },     90.00000,   0.04301, 99999.999 },
 {  0, {  0,  0,  0,  1 },    180.00000,   0.00003,     0.075 },
 {  0, {  0,  0,  0,  2 },    270.00000,   0.00004,     0.037 },
 {  0, {  0,  0,  1, -2 },    270.00000,   0.00444,     0.074 },
 {  0, {  0,  0,  1,  0 },     90.00000,   0.00531,     0.075 },
 {  0, {  0,  0,  2, -2 },     90.00000,   0.00026,     2.998 },
 {  0, {  0,  0,  2,  0 },     90.00000,   0.00042,     0.038 },
 {  0, {  0,  0,  2,  2 },    270.00000,   0.00004,     0.019 },
 {  0, {  0,  0,  3, -2 },    270.00000,   0.00002,     0.077 },
 {  0, {  0,  0,  3,  0 },     90.00000,   0.00003,     0.025 },
 {  0, {  0,  1, -2,  0 },    270.00000,   0.00004,     0.039 },
 {  0, {  0,  1, -1,  0 },    270.00000,   0.00037,     0.082 },
 {  0, {  0,  1,  0,  0 },     90.00000,   0.00009,     1.000 },
 {  0, {  0,  1,  1, -2 },    270.00000,   0.00005,     0.079 },
 {  0, {  0,  1,  1,  0 },     90.00000,   0.00017,     0.070 },
 {  0, {  0,  1,  2,  0 },     90.00000,   0.00002,     0.036 },
 {  0, {  0,  2,  0,  0 },    270.00000,   0.00002,     0.500 },
 {  0, {  1, -1,  0,  0 },     90.00000,   0.00012,     0.088 },
 {  0, {  1,  0, -1,  0 },    270.00000,   0.00005,     1.127 },
 {  0, {  1,  0,  0,  0 },     90.00000,   0.00088,     0.081 },
 {  0, {  1,  0,  1,  0 },     90.00000,   0.00006,     0.039 },
 {  0, {  1,  1, -1,  0 },    270.00000,   0.00004,     8.850 },
 {  0, {  1,  1,  0,  0 },     90.00000,   0.00064,     0.075 },
 {  0, {  1,  1,  1,  0 },     90.00000,   0.00004,     0.038 },
 {  0, {  2, -1, -1,  0 },    270.00000,   0.00053,     0.095 },
 {  0, {  2, -1,  1, -2 },    270.00000,   0.00003,     0.099 },
 {  0, {  2,  0, -3,  0 },     90.00000,   0.00003,     0.067 },
 {  0, {  2,  0, -2,  0 },     90.00000,   0.00082,     0.564 },
 {  0, {  2,  0, -2,  2 },    270.00000,   0.00003,     0.040 },
 {  0, {  2,  0, -1,  0 },    270.00000,   0.01961,     0.087 },
 {  0, {  2,  0, -1,  2 },    270.00000,   0.00003,     0.026 },
 {  0, {  2,  0,  0, -2 },     90.00000,   0.00021,     0.474 },
 {  0, {  2,  0,  0,  0 },    270.00000,   0.00047,     0.040 },
 {  0, {  2,  0,  1, -2 },    270.00000,   0.00071,     0.090 },
 {  0, {  2,  0,  2, -2 },    270.00000,   0.00005,     0.041 },
 {  0, {  2,  1, -1,  0 },     90.00000,   0.00008,     0.080 },
 {  0, {  2,  1,  0,  0 },     90.00000,   0.00003,     0.039 },
 {  0, {  2,  2, -1,  0 },    270.00000,   0.00009,     0.074 },
 {  0, {  4, -1, -1,  0 },    270.00000,   0.00002,     0.028 },
 {  0, {  4,  0, -2,  0 },    270.00000,   0.00017,     0.044 },
 {  0, {  4,  0, -1,  0 },    270.00000,   0.00017,     0.028 },
 {  0, {  4,  0,  0,  0 },    270.00000,   0.00002,     0.020 },
 {  1, { -6,  0,  1, -1 },     90.00094,   0.00002,     0.016 },
 {  1, { -6,  0,  2, -1 },     90.00070,   0.00003,     0.021 },
 {  1, { -4, -1,  1, -1 },    270.00094,   0.00003,     0.027 },
 {  1, { -4,  0, -1, -1 },     90.00103,   0.00012,     0.016 },
 {  1, { -4,  0, -1,  1 },    270.00027,   0.00005,     0.028 },
 {  1, { -4,  0,  0, -1 },     90.00096,   0.00075,     0.020 },
 {  1, { -4,  0,  0,  1 },     90.00086,   0.00040,     0.044 },
 {  1, { -4,  0,  1, -1 },     90.00079,   0.00149,     0.028 },
 {  1, { -4,  0,  1,  1 },     90.00079,   0.00027,     0.106 },
 {  1, { -4,  0,  2, -1 },     90.00028,   0.00046,     0.043 },
 {  1, { -4,  0,  2,  1 },     90.00067,   0.00008,     0.261 },
 {  1, { -4,  1,  0, -1 },     90.00095,   0.00010,     0.021 },
 {  1, { -4,  1,  0,  1 },     90.00072,   0.00005,     0.046 },
 {  1, { -4,  1,  1, -1 },     90.00079,   0.00017,     0.028 },
 {  1, { -4,  1,  1,  1 },     90.00072,   0.00002,     0.119 },
 {  1, { -4,  1,  2, -1 },     90.00029,   0.00004,     0.045 },
 {  1, { -3,  0,  0, -1 },     90.00092,   0.00008,     0.027 },
 {  1, { -3,  0,  0,  1 },    270.00072,   0.00018,     0.097 },
 {  1, { -3,  0,  1, -1 },    270.00069,   0.00007,     0.042 },
 {  1, { -2, -2,  1, -1 },    270.00024,   0.00002,     0.074 },
 {  1, { -2, -1, -1, -1 },    270.00109,   0.00014,     0.026 },
 {  1, { -2, -1, -1,  1 },     90.00059,   0.00006,     0.082 },
 {  1, { -2, -1,  0, -1 },    270.00097,   0.00096,     0.039 },
 {  1, { -2, -1,  0,  1 },     90.00072,   0.00012,     0.949 },
 {  1, { -2, -1,  1, -1 },    270.00047,   0.00044,     0.080 },
 {  1, { -2, -1,  1,  1 },     90.00065,   0.00005,     0.070 },
 {  1, { -2, -1,  2, -1 },     90.00072,   0.00002,     1.388 },
 {  1, { -2,  0, -3, -1 },     90.00117,   0.00007,     0.015 },
 {  1, { -2,  0, -3,  1 },    270.00047,   0.00003,     0.027 },
 {  1, { -2,  0, -2, -1 },     90.00113,   0.00080,     0.019 },
 {  1, { -2,  0, -2,  1 },    270.00044,   0.00044,     0.041 },
 {  1, { -2,  0, -1, -1 },     90.00108,   0.00925,     0.026 },
 {  1, { -2,  0, -1,  1 },    270.00046,   0.00617,     0.089 },
 {  1, { -2,  0, -1,  3 },     90.00074,   0.00003,     0.064 },
 {  1, { -2,  0,  0, -1 },     90.00096,   0.09554,     0.040 },
 {  1, { -2,  0,  0,  0 },    180.00000,   0.00005,     0.088 },
 {  1, { -2,  0,  0,  1 },    270.00079,   0.00854,     0.487 },
 {  1, { -2,  0,  1, -3 },    270.00088,   0.00003,     0.026 },
 {  1, { -2,  0,  1, -1 },     90.00028,   0.03855,     0.087 },
 {  1, { -2,  0,  1,  1 },    270.00070,   0.00785,     0.065 },
 {  1, { -2,  0,  2, -3 },    270.00030,   0.00002,     0.040 },
 {  1, { -2,  0,  2, -1 },     90.00216,   0.00280,     0.581 },
 {  1, { -2,  0,  2,  1 },    270.00071,   0.00040,     0.035 },
 {  1, { -2,  0,  3, -1 },     90.00159,   0.00047,     0.067 },
 {  1, { -2,  0,  3,  1 },    270.00091,   0.00002,     0.024 },
 {  1, { -2,  0,  4, -1 },     90.00149,   0.00004,     0.035 },
 {  1, { -2,  1, -2, -1 },     90.00112,   0.00007,     0.020 },
 {  1, { -2,  1, -2,  1 },    270.00050,   0.00002,     0.043 },
 {  1, { -2,  1, -1, -1 },     90.00109,   0.00069,     0.027 },
 {  1, { -2,  1, -1,  1 },    270.00048,   0.00028,     0.098 },
 {  1, { -2,  1,  0, -1 },     90.00096,   0.00657,     0.042 },
 {  1, { -2,  1,  0,  1 },    270.00077,   0.00056,     0.327 },
 {  1, { -2,  1,  1, -1 },     90.00029,   0.00160,     0.095 },
 {  1, { -2,  1,  1,  1 },    270.00071,   0.00029,     0.061 },
 {  1, { -2,  1,  2, -1 },     90.00227,   0.00011,     0.368 },
 {  1, { -2,  2, -1, -1 },     90.00094,   0.00003,     0.028 },
 {  1, { -2,  2,  0, -1 },     90.00095,   0.00032,     0.044 },
 {  1, { -2,  2,  0,  1 },    270.00072,   0.00003,     0.247 },
 {  1, { -2,  2,  1, -1 },     90.00030,   0.00005,     0.105 },
 {  1, { -1, -1, -1, -1 },     90.00121,   0.00003,     0.037 },
 {  1, { -1, -1,  0, -1 },     90.00107,   0.00018,     0.075 },
 {  1, { -1, -1,  0,  1 },    270.00052,   0.00004,     0.075 },
 {  1, { -1,  0, -1, -1 },    270.00128,   0.00018,     0.039 },
 {  1, { -1,  0, -1,  1 },    270.00072,   0.00005,     0.857 },
 {  1, { -1,  0,  0, -1 },    270.00137,   0.00063,     0.081 },
 {  1, { -1,  0,  0,  1 },     90.00068,   0.00075,     0.069 },
 {  1, { -1,  0,  1, -1 },    270.00123,   0.00014,     1.200 },
 {  1, { -1,  0,  1,  1 },     90.00072,   0.00004,     0.036 },
 {  1, { -1,  0,  2, -1 },    270.00129,   0.00006,     0.071 },
 {  1, { -1,  1,  0, -1 },    269.99992,   0.00003,     0.088 },
 {  1, {  0, -2, -1, -1 },    270.00137,   0.00002,     0.065 },
 {  1, {  0, -2,  1, -1 },    270.00115,   0.00006,     0.089 },
 {  1, {  0, -1, -3, -1 },    270.00127,   0.00002,     0.024 },
 {  1, {  0, -1, -2, -1 },    270.00134,   0.00022,     0.036 },
 {  1, {  0, -1, -2,  1 },    270.00073,   0.00003,     1.389 },
 {  1, {  0, -1, -1, -1 },    270.00140,   0.00181,     0.070 },
 {  1, {  0, -1, -1,  1 },     90.00162,   0.00010,     0.080 },
 {  1, {  0, -1,  0, -1 },     90.00067,   0.00028,     0.949 },
 {  1, {  0, -1,  0,  1 },     90.00087,   0.00013,     0.039 },
 {  1, {  0, -1,  1, -1 },    270.00118,   0.00341,     0.082 },
 {  1, {  0, -1,  1,  1 },    270.00000,   0.00003,     0.026 },
 {  1, {  0, -1,  2, -1 },    270.00121,   0.00034,     0.039 },
 {  1, {  0, -1,  3, -1 },    270.00127,   0.00003,     0.026 },
 {  1, {  0,  0, -4, -1 },     90.00128,   0.00010,     0.019 },
 {  1, {  0,  0, -3, -1 },     90.00128,   0.00150,     0.025 },
 {  1, {  0,  0, -3,  1 },    270.00000,   0.00006,     0.077 },
 {  1, {  0,  0, -2, -1 },     90.00128,   0.02469,     0.038 },
 {  1, {  0,  0, -2,  1 },     90.00062,   0.00485,     3.574 },
 {  1, {  0,  0, -1, -1 },     90.00128,   0.45276,     0.075 },
 {  1, {  0,  0, -1,  1 },    270.00061,   0.08508,     0.074 },
 {  1, {  0,  0,  0, -3 },     90.00067,   0.00012,     0.037 },
 {  1, {  0,  0,  0, -1 },     90.00001,   0.02809,    18.613 },
 {  1, {  0,  0,  0,  0 },    180.00000,   0.00025,     0.075 },
 {  1, {  0,  0,  0,  1 },     90.00050,   0.00390,     0.037 },
 {  1, {  0,  0,  1, -3 },    270.00037,   0.00108,     0.073 },
 {  1, {  0,  0,  1, -1 },    270.00126,   0.45648,     0.076 },
 {  1, {  0,  0,  1,  1 },     90.00024,   0.00025,     0.025 },
 {  1, {  0,  0,  2, -3 },    270.00280,   0.00004,     2.582 },
 {  1, {  0,  0,  2, -1 },    270.00126,   0.02488,     0.038 },
 {  1, {  0,  0,  2,  1 },     90.00000,   0.00002,     0.019 },
 {  1, {  0,  0,  3, -1 },    270.00126,   0.00153,     0.025 },
 {  1, {  0,  0,  3,  1 },    270.00000,   0.00002,     0.015 },
 {  1, {  0,  0,  4, -1 },    270.00126,   0.00010,     0.019 },
 {  1, {  0,  1, -3, -1 },     90.00127,   0.00003,     0.026 },
 {  1, {  0,  1, -2, -1 },     90.00134,   0.00027,     0.039 },
 {  1, {  0,  1, -2,  1 },     90.00072,   0.00005,     0.781 },
 {  1, {  0,  1, -1, -1 },     90.00143,   0.00218,     0.081 },
 {  1, {  0,  1, -1,  1 },    270.00084,   0.00025,     0.069 },
 {  1, {  0,  1,  0, -1 },     90.00080,   0.00024,     1.057 },
 {  1, {  0,  1,  0,  1 },     90.00064,   0.00015,     0.036 },
 {  1, {  0,  1,  1, -1 },     90.00116,   0.00279,     0.070 },
 {  1, {  0,  1,  2, -1 },     90.00121,   0.00028,     0.036 },
 {  1, {  0,  1,  3, -1 },     90.00127,   0.00002,     0.025 },
 {  1, {  0,  2, -1, -1 },     90.00147,   0.00003,     0.088 },
 {  1, {  0,  2,  1, -1 },     90.00120,   0.00003,     0.066 },
 {  1, {  1, -1,  0, -1 },     90.00080,   0.00002,     0.088 },
 {  1, {  1,  0, -2, -1 },     90.00180,   0.00003,     0.070 },
 {  1, {  1,  0, -1, -1 },    269.99953,   0.00006,     1.063 },
 {  1, {  1,  0, -1,  1 },     90.00072,   0.00002,     0.039 },
 {  1, {  1,  0,  0, -1 },     90.00088,   0.00299,     0.081 },
 {  1, {  1,  0,  0,  1 },     90.00051,   0.00002,     0.026 },
 {  1, {  1,  0,  1, -1 },     90.00104,   0.00031,     0.039 },
 {  1, {  1,  0,  2, -1 },     90.00090,   0.00002,     0.026 },
 {  1, {  1,  1,  0, -1 },    270.00058,   0.00059,     0.075 },
 {  1, {  1,  1,  1, -1 },    270.00085,   0.00005,     0.038 },
 {  1, {  2, -2, -1, -1 },    270.00041,   0.00007,     0.106 },
 {  1, {  2, -2,  0, -1 },    270.00093,   0.00036,     0.044 },
 {  1, {  2, -2,  1, -1 },    270.00094,   0.00003,     0.028 },
 {  1, {  2, -1, -2, -1 },    270.00281,   0.00008,     0.354 },
 {  1, {  2, -1, -2,  1 },    270.00059,   0.00003,     0.042 },
 {  1, {  2, -1, -1, -1 },    270.00042,   0.00218,     0.096 },
 {  1, {  2, -1, -1,  1 },    270.00039,   0.00006,     0.027 },
 {  1, {  2, -1,  0, -1 },    270.00092,   0.00745,     0.042 },
 {  1, {  2, -1,  0,  1 },    270.00000,   0.00002,     0.020 },
 {  1, {  2, -1,  1, -1 },    270.00105,   0.00075,     0.027 },
 {  1, {  2, -1,  2, -1 },    270.00112,   0.00007,     0.020 },
 {  1, {  2,  0, -4, -1 },    270.00148,   0.00004,     0.035 },
 {  1, {  2,  0, -3, -1 },    270.00168,   0.00042,     0.066 },
 {  1, {  2,  0, -3,  1 },     90.00024,   0.00006,     0.085 },
 {  1, {  2,  0, -2, -3 },    270.00122,   0.00002,     0.035 },
 {  1, {  2,  0, -2, -1 },    270.00300,   0.00177,     0.547 },
 {  1, {  2,  0, -2,  1 },    270.00064,   0.00078,     0.040 },
 {  1, {  2,  0, -1, -3 },    270.00120,   0.00018,     0.065 },
 {  1, {  2,  0, -1, -1 },    270.00043,   0.05343,     0.088 },
 {  1, {  2,  0, -1,  1 },    270.00062,   0.00059,     0.026 },
 {  1, {  2,  0,  0, -1 },    270.00092,   0.10689,     0.041 },
 {  1, {  2,  0,  0,  1 },    270.00095,   0.00002,     0.019 },
 {  1, {  2,  0,  1, -3 },     90.00107,   0.00010,     0.090 },
 {  1, {  2,  0,  1, -1 },    270.00105,   0.00992,     0.026 },
 {  1, {  2,  0,  2, -1 },    270.00111,   0.00084,     0.020 },
 {  1, {  2,  0,  3, -1 },    270.00115,   0.00007,     0.016 },
 {  1, {  2,  1, -2, -1 },     90.00072,   0.00002,     1.208 },
 {  1, {  2,  1, -1, -1 },     90.00008,   0.00016,     0.080 },
 {  1, {  2,  1,  0, -1 },     90.00092,   0.00115,     0.039 },
 {  1, {  2,  1,  1, -1 },     90.00104,   0.00016,     0.026 },
 {  1, {  2,  2, -1, -1 },     90.00033,   0.00005,     0.074 },
 {  1, {  3,  0, -1, -1 },     90.00070,   0.00012,     0.042 },
 {  1, {  3,  0,  0, -1 },    270.00096,   0.00007,     0.027 },
 {  1, {  4, -1, -2, -1 },    270.00042,   0.00006,     0.046 },
 {  1, {  4, -1, -1, -1 },    270.00078,   0.00020,     0.028 },
 {  1, {  4, -1,  0, -1 },    270.00093,   0.00011,     0.021 },
 {  1, {  4,  0, -2, -1 },    270.00042,   0.00063,     0.044 },
 {  1, {  4,  0, -1, -1 },    270.00078,   0.00176,     0.028 },
 {  1, {  4,  0,  0, -1 },    270.00093,   0.00084,     0.020 },
 {  1, {  4,  0,  1, -1 },    270.00105,   0.00013,     0.016 },
 {  1, {  4,  1, -1, -1 },     90.00094,   0.00003,     0.027 },
 {  1, {  6,  0, -2, -1 },    270.00070,   0.00003,     0.021 },
 {  1, {  6,  0, -1, -1 },    270.00094,   0.00002,     0.016 },
 {  2, { -2,  0, -1, -2 },    270.00000,   0.00005,     0.026 },
 {  2, { -2,  0, -1,  0 },     90.00000,   0.00009,     0.089 },
 {  2, { -2,  0,  0, -2 },    270.00000,   0.00047,     0.040 },
 {  2, { -2,  0,  0,  0 },     90.00000,   0.00004,     0.500 },
 {  2, { -2,  0,  1, -2 },    270.00000,   0.00019,     0.086 },
 {  2, { -2,  1,  0, -2 },    270.00000,   0.00004,     0.042 },
 {  2, {  0, -1,  0,  0 },     90.00000,   0.00002,     0.039 },
 {  2, {  0,  0, -2, -2 },    270.00000,   0.00012,     0.038 },
 {  2, {  0,  0, -2,  0 },    270.00000,   0.00005,     4.424 },
 {  2, {  0,  0, -1, -2 },    270.00000,   0.00219,     0.075 },
 {  2, {  0,  0, -1,  0 },     90.00000,   0.00077,     0.074 },
 {  2, {  0,  0,  0, -2 },    270.00000,   0.00028,     9.307 },
 {  2, {  0,  0,  0, -1 },      0.00000,   0.00003,     0.075 },
 {  2, {  0,  0,  0,  0 },     90.00000,   0.00448,     0.037 },
 {  2, {  0,  0,  1, -2 },     90.00000,   0.00223,     0.076 },
 {  2, {  0,  0,  1,  0 },     90.00000,   0.00050,     0.025 },
 {  2, {  0,  0,  2, -2 },     90.00000,   0.00012,     0.038 },
 {  2, {  0,  0,  2,  0 },     90.00000,   0.00006,     0.019 },
 {  2, {  1,  0,  0, -2 },    270.00000,   0.00002,     0.082 },
 {  2, {  1,  0,  0,  0 },     90.00000,   0.00003,     0.026 },
 {  2, {  1,  1,  0, -2 },     90.00000,   0.00004,     0.075 },
 {  2, {  2, -1,  0, -2 },     90.00000,   0.00004,     0.042 },
 {  2, {  2,  0, -1, -2 },     90.00000,   0.00032,     0.088 },
 {  2, {  2,  0, -1,  0 },     90.00000,   0.00009,     0.026 },
 {  2, {  2,  0,  0, -2 },     90.00000,   0.00056,     0.041 },
 {  2, {  2,  0,  0,  0 },     90.00000,   0.00005,     0.019 },
 {  2, {  2,  0,  1, -2 },     90.00000,   0.00005,     0.026 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 14> g_ELP7
{ {
 {  1, { -2,  0,  0, -1 },    180.00000,   0.00003,     0.040 },
 {  1, { -2,  0,  0,  1 },    180.00000,   0.00002,     0.487 },
 {  1, { -2,  0,  1, -1 },    180.00000,   0.00002,     0.087 },
 {  1, {  0,  0, -2, -1 },    180.00000,   0.00001,     0.038 },
 {  1, {  0,  0, -1, -1 },    180.00000,   0.00021,     0.075 },
 {  1, {  0,  0, -1,  1 },    180.00000,   0.00002,     0.074 },
 {  1, {  0,  0,  0, -1 },    180.00000,   0.00300,    18.613 },
 {  1, {  0,  0,  0,  1 },    180.00000,   0.00015,     0.037 },
 {  1, {  0,  0,  1, -1 },    180.00000,   0.00021,     0.076 },
 {  1, {  0,  0,  1,  1 },    180.00000,   0.00002,     0.025 },
 {  1, {  0,  0,  2, -1 },    180.00000,   0.00002,     0.038 },
 {  1, {  2,  0, -1, -1 },    180.00000,   0.00003,     0.088 },
 {  1, {  2,  0,  0, -1 },    180.00000,   0.00004,     0.041 },
 {  2, {  0,  0,  0, -2 },      0.00000,   0.00004,     9.307 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 11> g_ELP8
{ {
 {  1, { -2,  0,  0,  0 },    180.00000,   0.00012,     0.088 },
 {  1, { -2,  0,  1,  0 },    180.00000,   0.00003,     0.530 },
 {  1, {  0,  0, -1, -2 },    180.00000,   0.00001,     0.037 },
 {  1, {  0,  0, -1,  0 },    180.00000,   0.00019,     8.847 },
 {  1, {  0,  0,  0, -2 },    180.00000,   0.00014,     0.074 },
 {  1, {  0,  0,  0,  0 },      0.00000,   0.00342,     0.075 },
 {  1, {  0,  0,  1,  0 },      0.00000,   0.00018,     0.038 },
 {  1, {  0,  0,  2,  0 },      0.00000,   0.00001,     0.025 },
 {  1, {  2,  0, -1,  0 },      0.00000,   0.00004,     0.040 },
 {  1, {  2,  0,  0,  0 },      0.00000,   0.00002,     0.026 },
 {  2, {  0,  0,  0, -1 },    180.00000,   0.00009,     0.075 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 8> g_ELP9
{ {
 {  0, {  0,  0,  0,  0 },    270.00000,   0.00004, 99999.999 },
 {  1, { -2,  0,  0, -1 },    270.00000,   0.00004,     0.040 },
 {  1, { -2,  0,  1, -1 },    270.00000,   0.00002,     0.087 },
 {  1, {  0,  0, -1, -1 },    270.00000,   0.00019,     0.075 },
 {  1, {  0,  0, -1,  1 },     90.00000,   0.00003,     0.074 },
 {  1, {  0,  0,  1, -1 },     90.00000,   0.00019,     0.076 },
 {  1, {  2,  0, -1, -1 },     90.00000,   0.00002,     0.088 },
 {  1, {  2,  0,  0, -1 },     90.00000,   0.00004,     0.041 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 14328> g_ELP10
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 },    359.99831,   0.00020,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },    359.98254,   0.00007,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    359.93674,   0.00962,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   2 },    180.03508,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    359.95864,   0.00048,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0 },    359.94375,   0.00005,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    180.00175,   0.00004,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    180.01810,   0.00071,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },     97.01810,   0.00011,     0.949 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    180.01529,   0.00414,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1 },    248.64591,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1 },    259.46800,   0.00001,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },    180.01744,   0.00029,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0 },    180.94522,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    185.23351,   0.00003,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -2 },      0.04208,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },      0.07845,   0.00212,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },      0.12204,   0.03726,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2 },    180.10246,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2 },      0.23669,   0.00024,     0.474 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },      0.09994,   0.01585,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   2 },    180.11193,   0.00005,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2 },    180.01593,   0.00007,     0.090 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   0 },      0.11688,   0.00135,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    359.93335,   0.00009,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.12456,   0.00006,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,  -1 },    214.90087,   0.00001,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -3,   0 },      0.56402,   0.00006,     0.103 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0 },      0.12267,   0.00049,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.11476,   0.00057,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2 },      0.02034,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0 },      0.17048,   0.00020,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   1,   0 },      0.11019,   0.00005,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.12372,   0.00001,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0 },    319.83798,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   1,   0 },     49.17104,   0.00005,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0 },    261.18864,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },    261.05178,   0.00031,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },    260.13667,   0.00074,   164.772 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0 },    261.05178,   0.00031,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   2,   0 },    261.18864,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0 },     47.39808,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0 },    314.65028,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   0 },    153.55642,   0.00003,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },    333.58419,   0.00015,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },    333.76745,   0.00035,    82.386 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },    333.58419,   0.00015,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },    153.55642,   0.00003,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   0,   0 },    146.21801,   0.00007,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   1,   0 },    255.18835,   0.00018,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   2,   0 },    282.49407,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -2,   0 },    101.05952,   0.00008,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -1,   0 },    101.14709,   0.00125,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 },     95.09731,   0.00297,  4234.399 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   1,   0 },    101.14709,   0.00125,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0 },     94.53733,   0.00008,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,  -2,   0 },    282.49406,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,  -1,   0 },    255.29088,   0.00018,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,   0,   0 },    146.45132,   0.00007,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,  -1,   0 },    184.37205,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0 },    187.93148,   0.00017,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0 },    173.34889,   0.00030,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0 },    237.22913,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0 },    240.14508,   0.00047,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 },    242.32873,   0.00110,    84.021 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0 },    240.30167,   0.00047,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   2,   0 },    237.20081,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -1,   0 },    175.30397,   0.00030,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0 },    189.59104,   0.00017,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,   1,   0 },    186.06611,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 },    348.77302,   0.00001,  2117.199 },
 { {   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 },    359.69428,   0.00005,    85.722 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   1,   0 },    161.93169,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0 },    341.87569,   0.00013,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 },    341.49509,   0.00030,    42.010 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   0,   1,   0 },    341.87569,   0.00013,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -1,   0 },    161.93169,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -1,   0 },    152.66611,   0.00001,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 },    169.57246,   0.00002,    28.007 },
 { {   0,   0,   0,   0,   0,   0,   3,   0,   0,   1,   0 },    152.66611,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,  -2,   0,   0 },    290.10491,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,  -2,   1,   0 },    125.75292,   0.00001,     0.087 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   0,  -1,   0 },    306.44677,   0.00008,     0.075 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0 },    312.56240,   0.00015,   568.851 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   1,   0 },    306.44677,   0.00008,     0.075 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   2,  -1,   0 },    125.75292,   0.00001,     0.087 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   2,   0,   0 },    289.80972,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0 },    271.99675,   0.00001,    98.581 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0 },    267.90587,   0.00005,     0.020 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   1,   0 },    268.66358,   0.00011,     0.028 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   2,   0 },    269.91976,   0.00006,     0.044 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0 },    268.03939,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -2,   0 },    266.75880,   0.00006,     0.020 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -1,   0 },    266.27143,   0.00039,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -1,   2 },     87.84840,   0.00001,     0.090 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,  -2 },    267.94260,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0 },    267.68227,   0.00390,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   2 },    267.86053,   0.00008,     0.467 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   0 },    269.19797,   0.00584,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   2 },     87.84840,   0.00001,     0.065 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   2,   0 },    272.36342,   0.00020,     0.553 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0 },     87.45204,   0.00012,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -1,   1,   0 },     87.84840,   0.00002,     1.086 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0 },    260.45677,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0 },    260.87968,   0.00058,     0.038 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -2 },     81.50986,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },    260.78714,   0.00854,     0.076 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   2 },    261.50986,   0.00001,     0.073 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },    260.21409,   0.01901,    29.457 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -2 },    261.50986,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0 },    260.75637,   0.00850,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   2 },     81.50986,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   2,   0 },    260.82740,   0.00058,     0.038 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   3,   0 },    260.37448,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   1,  -1,   0 },     87.24961,   0.00002,     1.172 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,  -1 },     66.17808,   0.00002,     0.981 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0 },     87.69139,   0.00010,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -2,   0 },    272.85986,   0.00019,     0.575 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,  -2 },     87.24969,   0.00001,     0.065 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,   0 },    268.86171,   0.00583,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,  -2 },    267.23846,   0.00008,     0.482 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0 },    267.10688,   0.00392,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   2 },    267.15910,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   1,  -2 },     87.24969,   0.00001,     0.089 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   1,   0 },    266.09869,   0.00038,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   2,   0 },    266.30173,   0.00005,     0.019 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   3,  -2,   0 },    267.09052,   0.00001,     0.094 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   3,   0,   0 },    267.08441,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,  -2,   0 },    269.49802,   0.00006,     0.043 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,  -1,   0 },    268.41974,   0.00011,     0.028 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,   0,   0 },    266.95444,   0.00005,     0.020 },
 { {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0 },    211.80941,   0.00003,    31.066 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,  -1,   0 },    166.00397,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0 },    168.35859,   0.00022,     0.041 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   1,   0 },    162.80939,   0.00049,     0.088 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   2,   0 },    164.04160,   0.00001,     0.543 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   1 },    324.97162,   0.00001,     0.892 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0 },    325.00445,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0 },    326.15530,   0.00050,     0.076 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 },    326.83340,   0.00111,    14.729 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,   0 },    326.42167,   0.00051,     0.075 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   2,   0 },    324.95852,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,  -2,   0 },    165.60627,   0.00001,     0.586 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,  -1,   0 },    163.99685,   0.00043,     0.087 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,   0,   0 },    177.84361,   0.00014,     0.040 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,   1,   0 },    169.74560,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,  -2,   0,   0 },     48.74729,   0.00003,     0.041 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,  -2,   1,   0 },     64.88200,   0.00003,     0.088 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   0,  -1,   0 },    259.33632,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 },    272.75725,   0.00010,     9.819 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   0,   1,   0 },    259.33632,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   2,   0,   0 },     48.78047,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   0,   5,   0,   0,   0,  -1,   0 },    309.55348,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0 },    307.74906,   0.00004,     5.891 },
 { {   0,   0,   0,   0,   0,   5,   0,   0,   0,   1,   0 },    309.55348,   0.00002,     0.074 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,  -2,   0,   0 },    200.35369,   0.00008,     0.040 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,  -2,   1,   0 },    193.55743,   0.00016,     0.086 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,  -1,   0 },    258.01073,   0.00012,     0.075 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 },    269.35285,   0.00022,    11.705 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,   1,   0 },    252.22892,   0.00012,     0.076 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   2,  -1,   0 },    187.23229,   0.00017,     0.088 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   2,   0,   0 },    193.55450,   0.00009,     0.041 },
 { {   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 },    190.61995,   0.00002,    19.422 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,  -2,   0,   0 },    134.88846,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,  -2,   1,   0 },    123.44026,   0.00009,     0.087 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   0,  -1,   0 },    235.51258,   0.00010,     0.075 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 },    274.55463,   0.00040,    57.013 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   0,   1,   0 },    235.51258,   0.00010,     0.076 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   2,  -1,   0 },    120.51745,   0.00009,     0.087 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   2,   0,   0 },    131.88420,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,  -1,   0 },    355.56927,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   0,   0 },    352.34787,   0.00025,     0.040 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   1,   0 },      6.18326,   0.00046,     0.087 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   2,   0 },     34.77802,   0.00001,     0.559 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -2,   0 },    298.25858,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -1,   0 },    292.79512,   0.00062,     0.076 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 },    273.87630,   0.00129,    60.947 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   1,   0 },    292.62392,   0.00063,     0.075 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   2,   0 },    297.73414,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,  -2,   0 },     31.06782,   0.00001,     0.569 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,  -1,   0 },      3.95795,   0.00046,     0.087 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,   0,   0 },    350.33585,   0.00025,     0.040 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,   1,   0 },    353.65395,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,  -1,   0 },    177.14581,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   0,   0 },    177.14581,   0.00017,     0.041 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   1,   0 },    177.26391,   0.00032,     0.087 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   2,   0 },    177.14581,   0.00001,     0.548 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,  -2,   0 },    175.97413,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,  -1,   0 },    176.59350,   0.00012,     0.076 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },    348.75412,   0.00022,    19.859 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,   1,   0 },    176.41129,   0.00012,     0.075 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,   2,   0 },    175.99131,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,  -2,   0 },    176.87332,   0.00001,     0.580 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,  -1,   0 },    176.95520,   0.00032,     0.087 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,   0,   0 },    176.87333,   0.00016,     0.040 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,   1,   0 },    176.87333,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -6,   0,   0 },    177.97731,   0.00001,     0.013 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -6,   1,   0 },    354.88320,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -6,   2,   0 },    354.68177,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,  -2,   0 },    176.55938,   0.00001,     0.013 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,  -1,   0 },    348.73438,   0.00006,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   0,   0 },      0.26903,   0.00040,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   1,   0 },      0.19699,   0.00102,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   1,   2 },    355.97611,   0.00002,     0.108 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   2,   0 },      1.47626,   0.00078,     0.044 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   3,   0 },    351.56075,   0.00003,     0.104 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 },    279.65630,   0.00005,     0.027 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -3,   1,   0 },    204.05158,   0.00005,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -3,   2,   0 },     71.27124,   0.00003,     0.095 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -3,   0 },    355.56911,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -2,   0 },    344.56521,   0.00021,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,  -2 },    175.11988,   0.00002,     0.015 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   0 },    350.45816,   0.00286,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   2 },    169.14222,   0.00008,     0.090 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,  -2 },    174.54380,   0.00005,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 },    355.13454,   0.03309,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   2 },    354.99965,   0.00072,     0.456 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,  -2 },    180.41095,   0.00013,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   0 },    359.80946,   0.05756,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   2 },    353.42167,   0.00003,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   2,   0 },      5.19536,   0.00253,     0.538 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   3,   0 },     21.27165,   0.00012,     0.066 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   0 },    196.41179,   0.00010,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   1 },     78.50308,   0.00007,     0.083 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,  -1 },    258.87805,   0.00008,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 },    188.32808,   0.00100,     0.081 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   1 },     78.90433,   0.00034,     0.879 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,  -1 },    258.53595,   0.00008,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   0 },    192.24655,   0.00016,     1.030 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   1 },     78.48339,   0.00005,     0.069 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   2,   0 },    155.94596,   0.00002,     0.070 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -4,   0 },    289.44877,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -3,   0 },    292.78316,   0.00019,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -2 },    122.50459,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 },    295.89119,   0.00273,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,  -2 },    113.49087,   0.00008,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },    294.98414,   0.04046,     0.076 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   2 },    292.25162,   0.00006,     0.073 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -2 },    257.49480,   0.00003,     0.037 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1 },    256.85503,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },    286.69628,   0.08634,    11.862 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   2 },    211.84037,   0.00003,     0.037 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,  -2 },    293.46032,   0.00010,     0.074 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 },    293.67110,   0.04040,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   2 },    112.53679,   0.00008,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,   0 },    293.08941,   0.00279,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,   2 },    122.47574,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   3,   0 },    291.65112,   0.00019,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   4,   0 },    288.78628,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -2,   0 },    207.98979,   0.00001,     0.071 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,  -1 },     66.05885,   0.00006,     0.070 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0 },    186.72334,   0.00013,     1.246 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   1 },    258.55283,   0.00010,     0.079 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,  -1 },     77.02895,   0.00041,     1.032 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },    177.95329,   0.00074,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   1 },    258.55474,   0.00009,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   1,  -1 },    105.42894,   0.00006,     0.081 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   1,   0 },    174.33869,   0.00007,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   2,  -1 },     77.20036,   0.00001,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -3,   0 },     26.87354,   0.00012,     0.067 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },      7.90949,   0.00237,     0.592 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,  -2 },    352.29972,   0.00003,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    356.13718,   0.05274,     0.086 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   2 },    177.36829,   0.00012,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,  -2 },    345.34930,   0.00069,     0.494 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    346.26872,   0.02848,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   2 },    173.97248,   0.00004,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,  -2 },    163.49062,   0.00007,     0.089 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,   0 },    337.94966,   0.00255,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,   2 },    172.03927,   0.00001,     0.015 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   2,   0 },    331.01462,   0.00020,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   3,   0 },    356.16922,   0.00002,     0.015 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,  -1,  -1 },    358.70983,   0.00001,     0.095 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,  -1,   0 },    155.03044,   0.00004,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,   0,  -1 },    178.41128,   0.00002,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,   0,   0 },     22.35132,   0.00002,     0.027 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -3,   0 },    350.39718,   0.00003,     0.102 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -2,   0 },    355.55117,   0.00070,     0.043 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -1,  -2 },    348.73653,   0.00002,     0.106 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -1,   0 },    347.83821,   0.00089,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   0,  -2 },    347.03093,   0.00001,     0.044 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   0,   0 },    342.84700,   0.00034,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   1,   0 },    356.22590,   0.00006,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   2,   0 },    166.65213,   0.00001,     0.013 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   6,  -2,   0 },    351.09840,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   6,  -1,   0 },    343.77250,   0.00001,     0.016 },
 { {   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 },    272.58905,   0.00002,     8.457 },
 { {   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 },     94.56449,   0.00004,     6.570 },
 { {   0,   0,   0,   0,   1,   3,   0,   0,   0,   0,   0 },     37.10841,   0.00002,     5.372 },
 { {   0,   0,   0,   0,   2,  -7,   0,   0,   0,  -1,   0 },      1.58325,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 },    355.25511,   0.00008,    14.487 },
 { {   0,   0,   0,   0,   2,  -7,   0,   0,   0,   1,   0 },      1.58325,   0.00001,     0.076 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,  -2,   0,   0 },    101.18294,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,  -2,   1,   0 },    101.18292,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   0,  -1,   0 },    108.51356,   0.00004,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 },    110.07308,   0.00004,    28.507 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   0,   1,   0 },    108.51356,   0.00004,     0.076 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   2,  -1,   0 },    100.40235,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   2,   0,   0 },    100.40235,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,  -2,   0,   0 },     15.09223,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,  -2,   1,   0 },    171.28442,   0.00008,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,  -2,   2,   0 },    180.21790,   0.00001,     0.564 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -2,   0 },      0.04391,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -1,   0 },    359.84776,   0.00064,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 },    358.47067,   0.00146,  1597.835 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   1,   0 },    359.84776,   0.00064,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   2,   0 },      0.04391,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   2,  -2,   0 },    180.21794,   0.00001,     0.564 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   2,  -1,   0 },    171.19345,   0.00008,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   2,   0,   0 },     14.87532,   0.00004,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   4,  -2,   0,   0,   0 },      0.41099,   0.00001,  2566.169 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   0,   0 },    256.07455,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   1,   0 },    190.13554,   0.00012,     0.028 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   2,   0 },    177.02568,   0.00018,     0.044 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -2,   0 },    331.24839,   0.00008,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -1,   0 },    334.64637,   0.00060,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -1,   2 },    165.96095,   0.00002,     0.090 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   0 },    259.59890,   0.00155,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   2 },    280.67919,   0.00003,     0.475 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,  -2 },    344.54200,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   0 },    178.71268,   0.01400,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   2,   0 },    173.17456,   0.00131,     0.564 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   3,   0 },    170.01126,   0.00011,     0.067 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   4,   0 },    166.78471,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -1,  -1,   0 },    165.96095,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -1,   0,   0 },     62.25128,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -1,   1,   0 },    353.43385,   0.00006,     1.129 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -4,   0 },    345.96095,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -3,   0 },    346.65157,   0.00033,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,  -2 },    165.96095,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   0 },    346.45996,   0.00463,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,  -2 },    166.30689,   0.00013,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   0 },    346.48704,   0.06826,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   2 },    346.06876,   0.00012,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,  -2 },    342.38792,   0.00004,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 },    347.76653,   0.16376,   883.328 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   2 },    342.38792,   0.00004,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,  -2 },    346.06876,   0.00012,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   0 },    346.48972,   0.06826,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   2 },    166.30689,   0.00013,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   0 },    346.46192,   0.00463,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   2 },    165.96095,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   3,   0 },    346.65267,   0.00033,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   4,   0 },    345.96095,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,  -1,   0 },    353.37022,   0.00006,     1.126 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,   0,   0 },     72.26397,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,   1,   0 },    165.96095,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -4,   0 },    166.81401,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -3,   0 },    169.97581,   0.00012,     0.067 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -2,   0 },    173.13536,   0.00132,     0.563 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   0 },    178.60854,   0.01408,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   2 },    344.54200,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,  -2 },    278.18936,   0.00003,     0.474 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,   0 },    255.53360,   0.00155,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   1,  -2 },    165.96095,   0.00002,     0.090 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   1,   0 },    334.57965,   0.00059,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   2,   0 },    330.62667,   0.00008,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -2,   0 },    177.21943,   0.00018,     0.044 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -1,   0 },    189.85757,   0.00013,     0.028 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,   0,   0 },    253.41593,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,  -1,   0 },    250.78551,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,   0,   0 },    250.78550,   0.00011,     0.040 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,   1,   0 },    250.14160,   0.00020,     0.087 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,  -1,   0 },    262.69957,   0.00005,     0.076 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 },    315.35814,   0.00006,    30.473 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   1,   0 },    263.03964,   0.00005,     0.075 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,  -1,   0 },    250.40581,   0.00020,     0.087 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,   0,   0 },    251.03215,   0.00010,     0.040 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,   1,   0 },    251.03215,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,  -1,   0 },    300.40427,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,   0,   0 },    300.40427,   0.00013,     0.041 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,   1,   0 },    300.34516,   0.00022,     0.088 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,  -2,   0 },    300.71945,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,  -1,   0 },    300.84519,   0.00021,     0.076 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 },    301.99465,   0.00028,    14.978 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,   1,   0 },    300.84249,   0.00021,     0.075 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,   2,   0 },    300.71937,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,  -1,   0 },    300.29574,   0.00020,     0.087 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,   0,   0 },    300.37005,   0.00012,     0.040 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,   1,   0 },    300.36991,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -4,   1,   0 },    359.18166,   0.00001,     0.028 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,  -1,   0 },    359.17699,   0.00005,     0.026 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   0,   0 },    359.17518,   0.00029,     0.041 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   1,   0 },    359.39149,   0.00054,     0.088 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   2,   0 },    359.18166,   0.00002,     0.533 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,  -2,   0 },    356.78732,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,  -1,   0 },    356.71654,   0.00022,     0.076 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 },    356.94334,   0.00034,     9.929 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   1,   0 },    356.11336,   0.00020,     0.075 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   2,   0 },    356.83512,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,  -2,   0 },    358.65389,   0.00002,     0.598 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,  -1,   0 },    358.80579,   0.00049,     0.086 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   0,   0 },    358.64690,   0.00026,     0.040 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   1,   0 },    358.65385,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   2,  -1,   0,   0,  -2,   0,   0 },    165.33870,   0.00002,     0.041 },
 { {   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 },    355.11826,   0.00007,     7.426 },
 { {   0,   0,   0,   0,   2,  -1,   0,   0,   2,   0,   0 },    160.28913,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -4,   1,   0 },    134.38632,   0.00004,     0.028 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -4,   2,   0 },    119.98922,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,  -1,   0 },     11.33048,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 },     18.03421,   0.00065,     0.041 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   2 },     22.01561,   0.00002,     0.439 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,  -1 },    132.88991,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,   0 },     34.78287,   0.00099,     0.088 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   2,  -1 },    131.86228,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   2,   0 },     57.79356,   0.00005,     0.515 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0 },    140.96350,   0.00008,     0.082 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   1 },     46.10753,   0.00005,     0.818 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,   1,  -1 },    336.53896,   0.00003,     0.081 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -3,   0 },    319.08135,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 },    317.51869,   0.00028,     0.038 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1 },    133.31470,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0 },    318.59829,   0.00426,     0.076 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1 },    133.17116,   0.00035,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 },    316.17703,   0.00817,     5.931 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1 },    132.88991,   0.00228,   540.002 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   0 },    317.84238,   0.00425,     0.074 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1 },    133.67290,   0.00014,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,   0 },    316.09272,   0.00030,     0.037 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1 },    133.31470,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   3,   0 },    320.20899,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,  -1 },    109.05424,   0.00009,     1.130 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0 },    154.02387,   0.00003,     0.080 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -3,   0 },    247.15310,   0.00001,     0.067 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -2,   0 },     80.73462,   0.00005,     0.623 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -1,   0 },    353.56529,   0.00124,     0.086 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,  -2 },    334.62829,   0.00004,     0.516 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,  -1 },    313.84607,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   0 },    333.89330,   0.00127,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   1,  -1 },    132.88991,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   1,   0 },    330.39236,   0.00014,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   4,  -2,   0 },      8.91455,   0.00001,     0.043 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   4,  -1,   0 },    321.13691,   0.00002,     0.027 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   4,   0,   0 },    350.59390,   0.00001,     0.020 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,  -3,   0 },    119.78860,   0.00002,     0.068 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,  -2,   0 },    119.78860,   0.00002,     0.697 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,  -1,   0 },    119.78860,   0.00002,     0.085 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,   0,   0 },    119.78860,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 },     56.26588,   0.00005,    65.464 },
 { {   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 },     24.63163,   0.00002,    20.316 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,  -1,   0 },     84.56325,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   0,   0 },     83.71082,   0.00021,     0.041 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   1,   0 },     84.49231,   0.00038,     0.088 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,  -2,   0 },     88.64495,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,  -1,   0 },     88.12667,   0.00028,     0.076 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 },     86.22903,   0.00051,    12.023 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   1,   0 },     88.11011,   0.00028,     0.075 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   2,   0 },     87.56673,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,  -2,   0 },     84.80206,   0.00002,     0.591 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,  -1,   0 },     84.31461,   0.00038,     0.086 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   0 },     84.80157,   0.00021,     0.040 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,   1,   0 },     84.80210,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,  -2,   0,   0 },    120.93600,   0.00003,     0.041 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,  -2,   1,   0 },    120.93448,   0.00005,     0.088 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   0,  -1,   0 },    122.74231,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 },    161.95432,   0.00002,     8.538 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   0,   1,   0 },    122.74231,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   2,  -1,   0 },    120.50365,   0.00004,     0.086 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   2,   0,   0 },    120.50365,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,  -2,   0,   0 },    342.07391,   0.00004,     0.041 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,  -2,   1,   0 },    342.07391,   0.00006,     0.088 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   0,  -1,   0 },    350.34054,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 },     69.78928,   0.00004,     6.620 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   0,   1,   0 },    350.34054,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   2,  -1,   0 },    337.34517,   0.00004,     0.086 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   2,   0,   0 },    337.34519,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,  -2,   0,   0 },    356.81054,   0.00003,     0.041 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,  -2,   1,   0 },    356.81054,   0.00004,     0.089 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   0,  -1,   0 },    335.80159,   0.00004,     0.077 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 },    332.18900,   0.00009,     5.405 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   0,   1,   0 },    335.80159,   0.00004,     0.074 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   2,  -1,   0 },    346.51731,   0.00004,     0.086 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   2,   0,   0 },    347.94688,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -4,   2,   0 },    324.19590,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -2,   0,   0 },    110.59353,   0.00010,     0.041 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -2,   1,   0 },     87.86833,   0.00008,     0.089 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -1,   1,  -1 },    326.39334,   0.00002,     0.081 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,  -2,   0 },    314.35081,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,  -1,   0 },    313.53502,   0.00032,     0.077 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 },    310.21919,   0.00052,     3.954 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   1,   0 },    315.89488,   0.00036,     0.074 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   2,   0 },    314.12374,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   1,   0,  -1 },    115.20065,   0.00001,     1.249 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   2,  -1,   0 },      1.89253,   0.00004,     0.085 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   2,   0,   0 },    325.41769,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   4, -11,   3,   0,   0,   0,   0 },     47.87058,   0.00005,  1975.368 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,  -2,   0,   0 },     39.19362,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,  -2,   1,   0 },     13.36738,   0.00007,     0.087 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,  -2,   0 },    177.39404,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,  -1,   0 },    178.52240,   0.00027,     0.075 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,   0 },    184.68914,   0.00075,   441.664 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   1,   0 },    178.52240,   0.00027,     0.075 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   2,   0 },    177.39404,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   2,  -1,   0 },     13.11872,   0.00007,     0.087 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   2,   0,   0 },     37.90135,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 },    263.49004,   0.00002,    31.562 },
 { {   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 },    311.03589,   0.00003,    15.237 },
 { {   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 },    188.32594,   0.00003,    10.042 },
 { {   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 },     54.71950,   0.00002,     7.489 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,  -2,   0,   0 },     74.66964,   0.00002,     0.041 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,  -2,   1,   0 },     74.66964,   0.00003,     0.088 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   0,  -1,   0 },     95.68604,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 },    248.08734,   0.00005,     5.971 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   0,   1,  -1 },    345.24785,   0.00007,  1389.349 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   0,   1,   0 },     95.68604,   0.00002,     0.074 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   2,  -1,   0 },     81.34084,   0.00003,     0.086 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   2,   0,   0 },     81.35018,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 },      2.65862,   0.00003,     4.965 },
 { {   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 },     14.00981,   0.00001,     4.249 },
 { {   0,   0,   0,   0,   4,   0,   0,   0,  -2,   0,   0 },    114.00320,   0.00001,     0.041 },
 { {   0,   0,   0,   0,   4,   0,   0,   0,   0,  -1,   0 },    306.23084,   0.00001,     0.077 },
 { {   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 },    300.53586,   0.00002,     2.965 },
 { {   0,   0,   0,   0,   4,   0,   0,   0,   0,   1,   0 },    306.23084,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   4,   0,   0,   0,   0,   2,  -2 },    117.63494,   0.00002,   270.001 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,  -3,   0 },    320.02187,   0.00002,     0.068 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,  -2,   0 },    320.02187,   0.00002,     0.695 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,  -1,   0 },    320.02187,   0.00002,     0.085 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,   0,   0 },    320.02187,   0.00002,     0.040 },
 { {   0,   0,   0,   1,  -2,   0,   0,   0,  -2,   0,   0 },    150.15402,   0.00002,     0.041 },
 { {   0,   0,   0,   1,  -2,   0,   0,   0,  -2,   1,   0 },    330.22232,   0.00003,     0.090 },
 { {   0,   0,   0,   1,  -1,   0,   0,   0,  -2,   0,   0 },    357.16334,   0.00001,     0.041 },
 { {   0,   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0 },    175.45328,   0.00001,     0.078 },
 { {   0,   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0 },    175.45328,   0.00001,     0.073 },
 { {   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },    171.79341,   0.00001,     2.009 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -4,   0,   0 },     39.41179,   0.00001,     0.020 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -4,   2,   0 },     38.91086,   0.00001,     0.045 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -1,   0 },     39.08111,   0.00008,     0.027 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0 },     40.90204,   0.00088,     0.041 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   2 },     38.57425,   0.00002,     0.379 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0 },     39.20557,   0.00130,     0.091 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0 },     38.86732,   0.00008,     0.434 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   1 },    128.70847,   0.00002,     0.086 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1 },    308.37441,   0.00001,     0.040 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1 },    128.33297,   0.00008,     0.631 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   2,  -1 },    308.37441,   0.00001,     0.799 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0 },     20.12128,   0.00009,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },    206.82943,   0.00165,     1.881 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0 },    346.93250,   0.00005,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   2,   0 },      9.97369,   0.00001,     0.037 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -2,   1 },    308.75876,   0.00001,     5.303 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1 },    128.68576,   0.00003,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1 },    308.70847,   0.00012,     0.077 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1 },    128.80532,   0.00023,     1.915 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0 },     29.72397,   0.00004,     0.078 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   1 },    308.77291,   0.00005,     0.038 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1 },    128.68576,   0.00003,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -2,   0 },     34.25400,   0.00006,     0.805 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1,   0 },     30.33359,   0.00087,     0.083 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,  -2 },     26.41865,   0.00001,     0.635 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,   0 },     26.55092,   0.00033,     0.040 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   1,   0 },     30.52286,   0.00004,     0.026 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   4,  -1,   0 },     28.77501,   0.00001,     0.027 },
 { {   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 },     13.20679,   0.00001,     1.234 },
 { {   0,   0,   0,   2,  -2,   0,   0,   0,  -2,   0,   0 },      4.74462,   0.00001,     0.042 },
 { {   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },      0.16638,   0.00005,     1.118 },
 { {   0,   0,   0,   2,  -1,   0,   0,   0,  -2,   0,   0 },     26.03638,   0.00001,     0.042 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,  -2,   0,   0 },     27.14722,   0.00003,     0.042 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,  -2,   1,   0 },     27.14722,   0.00004,     0.095 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0 },     23.58979,   0.00001,     0.082 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0 },     15.82321,   0.00007,     1.005 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0 },     23.58979,   0.00001,     0.070 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,  -1,   0 },    309.50266,   0.00001,     0.016 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,   0,   0 },    124.90725,   0.00005,     0.021 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,   1,   0 },    125.50039,   0.00009,     0.028 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,   2,   0 },    129.58322,   0.00004,     0.046 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -3,   2,  -1 },    343.79532,   0.00001,     0.044 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -2,   0 },    135.39105,   0.00002,     0.020 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -1,   0 },    129.44100,   0.00024,     0.027 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0 },    132.93660,   0.00224,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   2 },    133.76976,   0.00003,     0.315 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   1,   0 },    134.97789,   0.00218,     0.096 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   2,   0 },    136.90760,   0.00007,     0.352 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -2,  -1 },    163.74195,   0.00002,     0.020 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1,  -1 },    163.73806,   0.00015,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1,   1 },    163.75176,   0.00003,     0.090 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,  -1 },    163.71185,   0.00081,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   1 },    344.27531,   0.00027,     0.472 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1,  -1 },    343.79779,   0.00076,     0.087 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   2,  -1 },    343.76057,   0.00014,     0.561 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   3,  -1 },    343.76077,   0.00001,     0.066 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0 },    147.90760,   0.00010,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0 },    157.82075,   0.00103,     0.082 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1 },    343.72199,   0.00002,     0.081 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },    183.35287,   0.00321,     0.940 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,   0 },    210.56812,   0.00043,     0.070 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   2,   0 },    262.25762,   0.00004,     0.036 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -3,  -1 },    163.74835,   0.00005,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -2,  -1 },    163.74835,   0.00074,     0.038 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -3 },    343.77101,   0.00003,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -1 },    163.74686,   0.01093,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,   1 },    343.54163,   0.00020,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,  -3 },    343.82568,   0.00008,     0.037 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,  -1 },    163.72199,   0.06369,   104.137 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,   0 },     46.57369,   0.00002,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,   1 },    343.75650,   0.00172,     0.037 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -3 },    163.70549,   0.00002,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -1 },    163.74984,   0.01090,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,   1 },    343.74948,   0.00021,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   2,  -1 },    163.74835,   0.00074,     0.038 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   3,  -1 },    163.74835,   0.00005,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -2,   0 },     18.89332,   0.00002,     1.407 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1,   0 },    320.29087,   0.00009,     0.080 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,  -2 },    281.41134,   0.00001,     0.958 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,  -1 },    343.72199,   0.00002,     0.081 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,   0 },    279.95494,   0.00019,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   1,   0 },    308.30928,   0.00003,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -3,  -1 },    343.76077,   0.00001,     0.067 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -2,  -1 },    343.76077,   0.00014,     0.567 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1,  -1 },    343.79308,   0.00083,     0.087 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1,   1 },    343.75180,   0.00004,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   0,  -1 },    163.71963,   0.00067,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   0,   1 },    343.75183,   0.00002,     0.019 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   1,  -1 },    163.73806,   0.00015,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   2,  -1 },    163.74195,   0.00002,     0.020 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   5,  -2,  -1 },    343.79532,   0.00001,     0.044 },
 { {   0,   0,   0,   3,  -2,   0,   0,   0,  -2,   0,   0 },    207.95868,   0.00001,     0.043 },
 { {   0,   0,   0,   3,  -2,   0,   0,   0,   2,  -1,   0 },     31.10931,   0.00003,     0.077 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   0,   0 },    326.00928,   0.00003,     0.043 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   1,   0 },    325.94090,   0.00003,     0.101 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0 },    279.57717,   0.00003,     0.086 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 },    268.26171,   0.00016,     0.627 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,   1,   0 },    293.39121,   0.00001,     0.067 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   1,   0,  -1 },      2.51896,   0.00002,     1.847 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -3,   0 },    146.58733,   0.00001,     0.074 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -2,   0 },     75.41993,   0.00001,     5.588 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -1,   0 },     90.42249,   0.00011,     0.076 },
 { {   0,   0,   0,   3,   1,   0,   0,   0,   2,  -3,   0 },    105.22410,   0.00001,     0.075 },
 { {   0,   0,   0,   3,   1,   0,   0,   0,   2,  -1,   0 },    285.22410,   0.00001,     0.076 },
 { {   0,   0,   0,   4,  -2,   0,   0,   0,   2,  -1,   0 },    236.75260,   0.00002,     0.074 },
 { {   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 },    256.66994,   0.00001,     0.490 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,  -2,   0,   0 },     37.01647,   0.00005,     0.044 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,  -2,   1,   0 },     37.43035,   0.00005,     0.107 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0 },    349.80173,   0.00002,     0.470 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   0,  -2 },    142.18734,   0.00001,    52.068 },
 { {   0,   0,   1,  -8,   3,   0,   0,   0,  -3,   1,   2 },    235.58943,   0.00001,   166.090 },
 { {   0,   0,   1,  -8,   3,   0,   0,   0,   2,  -1,   0 },    235.63732,   0.00001,     0.118 },
 { {   0,   0,   1,  -8,   3,   0,   0,   0,   2,   0,   0 },    235.63970,   0.00003,     0.046 },
 { {   0,   0,   1,  -7,   2,   0,   0,   0,  -2,   0,   0 },     51.76752,   0.00001,     0.037 },
 { {   0,   0,   1,  -7,   2,   0,   0,   0,  -2,   2,   0 },     51.76752,   0.00001,     1.283 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   2,   0,   0 },    119.90542,   0.00002,     0.044 },
 { {   0,   0,   1,  -6,   1,   0,   0,   0,  -2,   0,   2 },     77.24613,   0.00002,   578.012 },
 { {   0,   0,   1,  -5,  -2,   0,   0,   0,  -2,   1,   0 },    356.21498,   0.00001,     0.075 },
 { {   0,   0,   1,  -5,  -2,   0,   0,   0,  -2,   3,   0 },    176.21498,   0.00001,     0.076 },
 { {   0,   0,   1,  -5,  -1,   0,   0,   0,  -2,   1,   0 },    208.94037,   0.00002,     0.076 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   1,   0 },     81.69836,   0.00004,     0.076 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0 },    247.04513,   0.00004,     0.603 },
 { {   0,   0,   1,  -5,   2,   0,   0,   0,  -2,   1,   0 },    112.91539,   0.00002,     0.077 },
 { {   0,   0,   1,  -4,  -1,   0,   0,   0,  -2,   1,   0 },     52.50618,   0.00001,     0.079 },
 { {   0,   0,   1,  -4,  -1,   0,   0,   0,  -2,   3,   0 },     52.50618,   0.00001,     0.072 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   0,   0 },    225.00962,   0.00003,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   1,   0 },    174.06698,   0.00003,     0.079 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   2,   0 },     47.86823,   0.00001,     1.545 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,  -1,   1 },    345.16456,   0.00006,     0.075 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,  -1 },    165.71481,   0.00003,     0.037 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   1 },    345.86786,   0.00038,    13.704 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   1,   1 },    345.16456,   0.00006,     0.076 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0 },      9.97217,   0.00009,     0.070 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0 },     15.11214,   0.00083,     0.888 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0 },     31.49630,   0.00035,     0.082 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   2,   0 },     18.53482,   0.00002,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   1,   0,  -1 },    167.84474,   0.00001,     0.459 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -2,   0 },     41.77853,   0.00004,     0.345 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -1,   0 },     41.78518,   0.00085,     0.097 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,  -2 },     41.50084,   0.00001,     0.309 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,   0 },     39.88448,   0.00078,     0.042 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   1,   0 },     41.14943,   0.00011,     0.027 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   4,  -1,   0 },     43.65921,   0.00002,     0.028 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   4,   0,   0 },     43.14806,   0.00002,     0.021 },
 { {   0,   0,   1,  -4,   0,   2,   0,   0,  -1,   0,   1 },    241.06175,   0.00001,   197.046 },
 { {   0,   0,   1,  -4,   1,   0,   0,   0,   2,  -2,   0 },    277.35236,   0.00001,     0.355 },
 { {   0,   0,   1,  -4,   1,   0,   0,   0,   2,   0,   0 },    277.35236,   0.00001,     0.042 },
 { {   0,   0,   1,  -4,   2,   0,   0,   0,   0,   0,   0 },    122.72522,   0.00004,     1.044 },
 { {   0,   0,   1,  -4,   2,   0,   0,   0,   2,  -1,   0 },    115.72613,   0.00002,     0.095 },
 { {   0,   0,   1,  -4,   2,   0,   0,   0,   2,   0,   0 },    102.26017,   0.00002,     0.042 },
 { {   0,   0,   1,  -4,   3,   0,   0,   0,   0,   0,   0 },    137.23345,   0.00003,     1.144 },
 { {   0,   0,   1,  -4,   3,   0,   0,   0,   2,   0,   0 },    137.46624,   0.00001,     0.042 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,  -2,   0 },    228.86324,   0.00001,     0.019 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   0,   0 },    236.42011,   0.00005,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   1,   0 },    126.30991,   0.00015,     0.083 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   2,   0 },      8.58705,   0.00001,     0.848 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   3,   0 },    302.28516,   0.00002,     0.069 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,  -1 },    201.08161,   0.00002,     0.038 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,   0 },    122.42917,   0.00002,     0.077 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,   1 },     20.96441,   0.00014,     2.180 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   1,  -1 },    201.28465,   0.00009,     0.076 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0 },    280.01886,   0.00028,     0.072 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0 },    301.97868,   0.00159,     1.681 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0 },    293.64113,   0.00022,     0.079 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   1,   0,  -1 },     24.27550,   0.00006,     0.607 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   1,   1,  -1 },     21.28465,   0.00002,     0.086 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,  -2,   0 },    356.47454,   0.00001,     0.422 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,  -1,   0 },     81.33996,   0.00045,     0.092 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,   0 },     74.85425,   0.00032,     0.041 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   1,   0 },     80.57214,   0.00004,     0.027 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   4,  -1,   0 },     81.04778,   0.00001,     0.028 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   4,   0,   0 },    262.65349,   0.00002,     0.020 },
 { {   0,   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0 },    161.60315,   0.00002,     1.958 },
 { {   0,   0,   1,  -2,  -3,   0,   0,   0,   0,   0,   0 },    103.73270,   0.00002,     3.162 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -6,   2,   0 },    318.97517,   0.00001,     0.021 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,  -1,   0 },    318.63956,   0.00005,     0.016 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   0,   0 },    319.65512,   0.00026,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   0,   2 },    318.48112,   0.00003,     0.044 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   1,   0 },    318.28395,   0.00066,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   1,   2 },    318.43015,   0.00001,     0.106 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   2,   0 },    321.10029,   0.00043,     0.043 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   0,   0 },    317.24140,   0.00002,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   0,   1 },    273.34108,   0.00005,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   1,   0 },    137.98844,   0.00002,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   1,   1 },     93.26775,   0.00004,     0.095 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -3,   0 },    318.83036,   0.00002,     0.015 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -2,   0 },    317.09601,   0.00021,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   0 },    316.90522,   0.00241,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   2 },    137.02518,   0.00007,     0.089 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,  -2 },    139.62246,   0.00004,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   0 },    318.22943,   0.02301,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   2 },    318.00010,   0.00054,     0.489 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,  -2 },    141.08190,   0.00007,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   0 },    320.91988,   0.03194,     0.087 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   2 },    319.46143,   0.00002,     0.065 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   2,   0 },    327.67688,   0.00088,     0.585 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   3,   0 },    343.19358,   0.00003,     0.067 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,  -1,   0 },    137.20200,   0.00007,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,  -1,   1 },    273.31472,   0.00007,     0.082 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,  -1 },    232.98699,   0.00003,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   0 },    139.08029,   0.00057,     0.080 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   1 },     56.72062,   0.00012,     1.010 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   1,  -1 },    232.97824,   0.00001,     0.079 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   1,   0 },    144.71037,   0.00005,     1.214 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   1,   1 },     93.31472,   0.00005,     0.070 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   3,  -1 },    232.97824,   0.00001,     0.072 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   0 },    310.95751,   0.00001,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   0 },    313.25246,   0.00030,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -2 },    134.71038,   0.00003,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0 },    313.27156,   0.00431,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -2 },    134.82829,   0.00015,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0 },    313.30347,   0.06323,     0.075 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2 },    313.24076,   0.00012,     0.074 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2 },    143.30288,   0.00015,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },    318.13776,   0.32455,    15.781 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2 },    145.38488,   0.00013,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2 },    312.47618,   0.00009,     0.073 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0 },    313.47920,   0.06406,     0.076 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   2 },    134.82829,   0.00015,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0 },    313.52666,   0.00433,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   2 },    134.71140,   0.00003,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   3,   0 },    313.18788,   0.00030,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   4,   0 },    311.14008,   0.00001,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,  -1,   0 },    143.54052,   0.00007,     1.052 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,  -1,   1 },    244.81403,   0.00003,     0.080 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,  -1 },     52.65308,   0.00012,     0.895 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,   0 },    141.66373,   0.00067,     0.081 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   1,   0 },    137.13196,   0.00009,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -3,   0 },    341.19829,   0.00003,     0.066 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -2,   0 },    325.20289,   0.00111,     0.544 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,  -2 },    319.84005,   0.00002,     0.065 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   0 },    321.96125,   0.03549,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   2 },    141.94831,   0.00008,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,  -2 },    320.04579,   0.00054,     0.461 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   0 },    320.19758,   0.02513,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   2 },    139.80638,   0.00004,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,  -2 },    138.18975,   0.00007,     0.090 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,   0 },    319.04223,   0.00255,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,   2 },    140.12475,   0.00001,     0.015 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,   0 },    317.56724,   0.00022,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   3,  -1,   0 },    140.02816,   0.00003,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   3,   0,   0 },    321.38522,   0.00002,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -3,   0 },    323.77231,   0.00001,     0.104 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -2,   0 },    322.00414,   0.00049,     0.044 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -1,  -2 },    320.26210,   0.00001,     0.107 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -1,   0 },    321.23525,   0.00070,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   0,   0 },    319.08476,   0.00030,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   1,   0 },    319.30405,   0.00006,     0.016 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   6,  -2,   0 },    320.02325,   0.00002,     0.021 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   6,  -1,   0 },    319.61294,   0.00001,     0.016 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,  -1,   0 },     33.82493,   0.00003,     0.026 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   0,   0 },     33.60930,   0.00021,     0.040 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   1,   0 },     34.90305,   0.00035,     0.087 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   2,   0 },     33.82493,   0.00001,     0.562 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,  -2,   0 },     26.34471,   0.00002,     0.038 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,  -1,   0 },     27.52486,   0.00036,     0.075 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0,   0 },     31.83398,   0.00197,   220.858 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   1,   0 },     27.50267,   0.00036,     0.075 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   2,   0 },     26.34872,   0.00002,     0.038 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,  -2,   0 },     33.77499,   0.00001,     0.565 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,  -1,   0 },     34.85337,   0.00035,     0.087 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,   0,   0 },     33.56257,   0.00021,     0.040 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,   1,   0 },     33.77499,   0.00003,     0.026 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,  -2,   0,   0 },    195.15893,   0.00004,     0.040 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,  -2,   1,   0 },    195.75764,   0.00005,     0.087 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,  -1,   0 },    193.10597,   0.00006,     0.076 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0 },    196.88833,   0.00026,    47.766 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,   1,   0 },    193.10597,   0.00006,     0.075 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   2,  -1,   0 },    194.23383,   0.00005,     0.087 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   2,   0,   0 },    196.00816,   0.00003,     0.040 },
 { {   0,   0,   1,  -2,   2,   0,   0,   0,   0,   0,   0 },      7.22764,   0.00001,     9.502 },
 { {   0,   0,   1,  -2,   3,   0,   0,   0,   0,   0,   0 },    199.12221,   0.00001,     5.276 },
 { {   0,   0,   1,  -1,  -2,   0,   0,   0,  -2,   0,   0 },    121.10022,   0.00002,     0.041 },
 { {   0,   0,   1,  -1,  -2,   0,   0,   0,   0,  -1,   0 },    301.79181,   0.00001,     0.077 },
 { {   0,   0,   1,  -1,  -2,   0,   0,   0,   0,   1,   0 },    301.79181,   0.00001,     0.074 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -6,   2,   0 },      0.64900,   0.00001,     0.021 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   0,   0 },      0.73534,   0.00005,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   1,   0 },      0.65606,   0.00015,     0.028 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   2,   0 },      0.65537,   0.00009,     0.044 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -3,   2,   0 },      0.60489,   0.00001,     0.099 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -2,   0 },      0.64635,   0.00003,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -1,   0 },      0.64804,   0.00039,     0.027 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -1,   2 },    180.62741,   0.00001,     0.094 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,  -2 },    180.64806,   0.00001,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   0 },      0.65541,   0.00466,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   2 },      0.63749,   0.00009,     0.388 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   0 },      0.68035,   0.00692,     0.091 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   2 },    180.62741,   0.00001,     0.063 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   2,   0 },      0.68420,   0.00032,     0.446 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   3,   0 },      0.63493,   0.00002,     0.065 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   0 },    180.47095,   0.00017,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   1 },     92.76982,   0.00004,     0.657 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   1,   0 },    180.62741,   0.00001,     0.738 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0 },      0.52829,   0.00022,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0 },      0.53930,   0.00342,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0 },      0.88394,   0.01051,     2.135 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2 },    181.11135,   0.00002,     0.076 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0 },      0.43288,   0.00296,     0.073 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0 },      0.39412,   0.00019,     0.037 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -2 },    181.11135,   0.00002,     0.075 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   3,   0 },      0.28699,   0.00001,     0.025 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,  -1,   1 },    274.24036,   0.00004,     0.077 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,  -1 },     94.55317,   0.00013,     1.708 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0 },      0.28330,   0.00013,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   1 },    274.46515,   0.00002,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -2,   0 },    359.85577,   0.00015,     0.766 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -1,   0 },      0.34936,   0.00450,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,  -2 },      0.25086,   0.00007,     0.610 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0 },      0.32528,   0.00218,     0.040 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   1,   0 },      0.33794,   0.00022,     0.026 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   2,   0 },      0.28876,   0.00002,     0.019 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   3,  -2,   0 },    180.19743,   0.00001,     0.090 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,  -2,   0 },      0.41949,   0.00007,     0.043 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,  -1,   0 },      0.35246,   0.00007,     0.027 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,   0,   0 },      0.30629,   0.00005,     0.020 },
 { {   0,   0,   1,  -1,   1,   0,   0,   0,   2,   0,   0 },    212.06686,   0.00001,     0.040 },
 { {   0,   0,   1,  -1,   3,   0,   0,   0,   1,  -2,   1 },    127.67456,   0.00003,  1023.362 },
 { {   0,   0,   1,   0,  -6,   0,   0,   0,   0,   0,   0 },    334.15538,   0.00003,     2.024 },
 { {   0,   0,   1,   0,  -6,   5,   0,   0,   0,   0,   0 },    198.89487,   0.00004,     1.506 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,  -2,   0,   0 },    327.39228,   0.00005,     0.041 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,  -2,   1,   0 },    328.06938,   0.00006,     0.092 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   0,  -1,   0 },    323.41315,   0.00007,     0.079 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 },    322.87209,   0.00030,     1.729 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   0,   1,   0 },    323.41315,   0.00007,     0.072 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   2,  -1,   0 },    326.88174,   0.00002,     0.083 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   2,   0,   0 },    325.92462,   0.00002,     0.040 },
 { {   0,   0,   1,   0,  -5,   2,   0,   0,  -2,   0,   0 },    312.39644,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -5,   2,   0,   0,   0,   0,   0 },    302.19921,   0.00007,     1.547 },
 { {   0,   0,   1,   0,  -5,   3,   0,   0,  -2,   0,   0 },    310.90387,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -5,   3,   0,   0,   0,   0,   0 },    309.51439,   0.00005,     1.470 },
 { {   0,   0,   1,   0,  -5,   4,   0,   0,   0,   0,   0 },    186.07310,   0.00002,     1.400 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,  -2,   0,   0 },    183.85027,   0.00004,     0.042 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,  -2,   1,   0 },    184.61091,   0.00004,     0.093 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   0,  -1,   0 },    180.24308,   0.00004,     0.080 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   0,   0,   0 },    172.89602,   0.00018,     1.337 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   0,   1,   0 },    180.24308,   0.00004,     0.071 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   2,  -1,   0 },    337.15396,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -5,   7,   0,   0,   0,   0,   0 },    167.61398,   0.00002,     1.225 },
 { {   0,   0,   1,   0,  -5,  10,   0,   0,   0,   0,   0 },    343.19316,   0.00002,     1.089 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -4,   1,   0 },    317.60841,   0.00002,     0.028 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    317.22604,   0.00005,     0.027 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   0,   0 },    317.72609,   0.00045,     0.042 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   1,   0 },    317.45100,   0.00064,     0.092 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   2,   0 },    317.77153,   0.00002,     0.410 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -1,  -1,   1 },    121.20491,   0.00002,     0.087 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -1,   0,   1 },    122.26208,   0.00002,     0.583 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -2,   0 },    313.48645,   0.00004,     0.039 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -1,   0 },    313.56750,   0.00059,     0.079 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 },    313.05906,   0.00267,     1.509 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   1,   0 },    312.99376,   0.00050,     0.072 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   2,   0 },    313.51592,   0.00004,     0.037 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,  -2,   1 },    301.44525,   0.00004,    17.401 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,  -1,   1 },    301.20488,   0.00012,     0.076 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,   0,  -1 },    120.55930,   0.00010,     2.558 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,   0,   0 },    313.39291,   0.00002,     0.077 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,   0,   1 },    300.64203,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,  -2,   0 },    128.40108,   0.00002,     0.900 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,  -1,   0 },    332.01177,   0.00012,     0.082 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,   0,   0 },    316.31844,   0.00014,     0.039 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,   1,   0 },    316.99535,   0.00001,     0.026 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,  -2,   0,   0 },    313.23833,   0.00006,     0.042 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,  -2,   1,   0 },    313.70581,   0.00007,     0.093 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,  -1,   0 },    308.34030,   0.00006,     0.080 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,   0,   0 },    304.37716,   0.00027,     1.369 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,   1,   0 },    308.34030,   0.00006,     0.071 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,  -3,   1 },    281.30192,   0.00001,     0.075 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,  -2,   1 },    281.09191,   0.00012,    95.906 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,  -1,   1 },    101.05464,   0.00006,     0.075 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,   0,  -1 },    100.99207,   0.00001,     3.095 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   2,  -1,   0 },     15.12737,   0.00001,     0.082 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   2,   0,   0 },    314.73020,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,  -2,   0,   0 },      6.00875,   0.00004,     0.042 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,  -2,   1,   0 },      6.00874,   0.00003,     0.093 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,   0,  -1,   0 },      5.63245,   0.00002,     0.080 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,   0,   0,   0 },    356.24475,   0.00013,     1.308 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,   0,   1,   0 },      5.63245,   0.00002,     0.071 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,   2,  -1,   0 },    160.08416,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -4,   4,   0,   0,  -2,   0,   0 },    239.61591,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -4,   4,   0,   0,   0,   0,   0 },    238.96661,   0.00005,     1.252 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,  -2,   0,   0 },    117.63895,   0.00008,     0.042 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,  -2,   1,   0 },    118.32161,   0.00011,     0.094 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,  -1,   0 },    111.88052,   0.00011,     0.080 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,   0,   0 },    107.79183,   0.00055,     1.201 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,   1,   0 },    110.21753,   0.00009,     0.071 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   1,   0,  -1 },    271.91529,   0.00007,     4.520 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,  -3,   0 },    300.30406,   0.00002,     0.070 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,  -2,   0 },    274.52734,   0.00001,     1.062 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,  -1,   0 },     98.10967,   0.00004,     0.081 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,   0,   0 },    127.64824,   0.00003,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   0,   0 },    308.92791,   0.00005,     0.021 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   1,   0 },    309.20352,   0.00013,     0.028 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   2,   0 },    310.42636,   0.00006,     0.045 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -3,   0,   0 },    313.10603,   0.00001,     0.028 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -2,   0 },    308.98080,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    309.25262,   0.00039,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,  -2 },    128.81414,   0.00001,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },    309.45272,   0.00436,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   2 },    309.61675,   0.00007,     0.350 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,   0 },    309.40223,   0.00587,     0.093 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   2,   0 },    307.95789,   0.00027,     0.397 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   3,   0 },    307.85103,   0.00002,     0.063 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,  -2,   1 },    290.66152,   0.00005,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,  -1,   1 },    290.63435,   0.00051,     0.087 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,  -1 },    291.84522,   0.00003,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,   0 },    128.69461,   0.00020,     0.086 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,   1 },    112.17551,   0.00009,     0.555 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   1,  -1 },    291.45099,   0.00005,     0.085 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   1,   0 },    131.16083,   0.00002,     0.612 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   1,   1 },    111.36112,   0.00003,     0.066 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -2,   0 },    306.40303,   0.00031,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,  -2 },    125.23097,   0.00002,     0.025 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,   0 },    306.55823,   0.00458,     0.080 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 },    304.60699,   0.02314,     1.339 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   0 },    304.88943,   0.00415,     0.071 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   2 },    125.23097,   0.00002,     0.024 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   2,   0 },    304.62567,   0.00027,     0.037 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -3,   1 },    290.62181,   0.00007,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -2,  -1 },    290.92232,   0.00002,     0.037 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -2,   1 },    290.64489,   0.00158,    37.265 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,  -1 },    111.02527,   0.00016,     0.074 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,   0 },    304.95930,   0.00012,     7.147 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,   1 },    110.66854,   0.00247,     0.075 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,  -1 },    110.93672,   0.00119,     3.261 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,   0 },    308.03001,   0.00033,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,   1 },    291.62425,   0.00006,     0.038 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   1,  -1 },    111.02527,   0.00016,     0.077 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   1,   0 },    307.99315,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   1,   1 },    290.92405,   0.00002,     0.025 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -3,   0 },    294.03154,   0.00003,     0.070 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -2,   0 },    111.89138,   0.00031,     0.974 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,   0 },     88.25841,   0.00147,     0.082 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,  -2 },    308.87316,   0.00005,     0.735 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,   0 },    317.17854,   0.00083,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   1,   0 },    314.52759,   0.00010,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   3,  -3,   1 },    290.66156,   0.00002,     0.087 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   3,  -2,   1 },    110.69773,   0.00002,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   3,  -1,   1 },    110.66968,   0.00002,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -3,   0 },    112.95176,   0.00020,     0.096 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -2,   0 },    112.27720,   0.00050,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -1,   0 },    111.44731,   0.00030,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,   0,   0 },    105.44695,   0.00003,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   6,  -3,   0 },    112.98414,   0.00002,     0.028 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   6,  -2,   0 },    113.01972,   0.00001,     0.021 },
 { {   0,   0,   1,   0,  -3,   1,   0,   0,  -2,   0,   0 },    130.31720,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -3,   1,   0,   0,   0,   0,   0 },    115.94410,   0.00008,     1.280 },
 { {   0,   0,   1,   0,  -3,   1,   0,   0,   2,  -1,   0 },    104.37510,   0.00004,     0.082 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,  -2,  -1,   0 },    356.28551,   0.00001,     0.027 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,  -2,   0,   0 },    356.28552,   0.00014,     0.042 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,  -2,   1,   0 },    356.33506,   0.00018,     0.094 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,  -1,   0 },    355.20730,   0.00015,     0.080 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,   0,   0 },    353.30112,   0.00059,     1.227 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,   1,   0 },    355.02334,   0.00011,     0.071 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,  -3,   0 },      0.64187,   0.00002,     0.070 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,  -2,   0 },      1.16561,   0.00002,     1.043 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,  -1,   0 },    357.65883,   0.00006,     0.081 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,   0,   0 },    357.61155,   0.00005,     0.039 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   4,  -3,   0 },    177.90569,   0.00001,     0.095 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   4,  -2,   0 },    177.49266,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,  -2,   0,   0 },     60.48779,   0.00007,     0.042 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,  -2,   1,   0 },     60.50212,   0.00007,     0.094 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,  -1,   0 },     60.35673,   0.00005,     0.081 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,   0,   0 },     60.30729,   0.00021,     1.178 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,   1,   0 },     60.35673,   0.00005,     0.071 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   2,  -1,   0 },     57.64029,   0.00002,     0.081 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   2,   0,   0 },     58.49142,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -3,   4,   0,   0,  -2,   0,   0 },    296.75959,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -3,   4,   0,   0,   0,   0,   0 },    291.95568,   0.00003,     1.133 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -4,   1,   0 },    172.51650,   0.00001,     0.028 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,  -1,   0 },    172.55521,   0.00004,     0.027 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   0,   0 },    172.56976,   0.00041,     0.042 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   1,   0 },    172.46418,   0.00055,     0.095 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   2,   0 },    172.51632,   0.00002,     0.372 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -1,   0,   0 },    172.58186,   0.00003,     0.087 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,  -2,   0 },    173.03974,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,  -1,   0 },    172.86210,   0.00038,     0.081 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   0,   0 },    172.32635,   0.00171,     1.091 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   1,   0 },    172.94415,   0.00025,     0.071 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   2,   0 },    173.00815,   0.00002,     0.036 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   1,  -1,   0 },    172.58701,   0.00006,    33.536 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   1,   0,  -1 },     12.25880,   0.00001,     7.303 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   1,   0,   0 },    352.58186,   0.00017,     0.075 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,  -3,   0 },      1.54121,   0.00001,     0.071 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,  -1,   0 },    171.32092,   0.00012,     0.081 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,   0,   0 },    172.45015,   0.00008,     0.039 },
 { {   0,   0,   1,   0,  -3,   6,  -3,   0,   0,   0,   0 },    349.49838,   0.00002,     1.093 },
 { {   0,   0,   1,   0,  -2,  -5,   0,   0,   0,   0,   0 },    321.16517,   0.00003,     1.511 },
 { {   0,   0,   1,   0,  -2,  -2,   0,   0,   0,   0,   0 },     97.39312,   0.00003,     1.310 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -6,   1,   0 },    306.29079,   0.00002,     0.017 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -6,   2,   0 },    305.91426,   0.00003,     0.021 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,  -1,   0 },    305.66531,   0.00007,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   0,   0 },    306.66166,   0.00038,     0.021 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   1,   0 },    306.60629,   0.00090,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   1,   2 },    306.35447,   0.00001,     0.117 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   2,   0 },    305.82923,   0.00060,     0.045 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   3,   0 },    301.60541,   0.00003,     0.113 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   0,   0 },    308.20751,   0.00001,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   1,   0 },    126.16501,   0.00006,     0.043 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   1,   1 },    280.64001,   0.00001,     0.104 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -3,   0 },    305.76630,   0.00003,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -2,   0 },    304.55917,   0.00026,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,  -2 },    125.58929,   0.00002,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    305.55372,   0.00313,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,   2 },    125.69253,   0.00007,     0.097 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,  -2 },    124.64040,   0.00007,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    306.02804,   0.03474,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   2 },    306.27197,   0.00050,     0.340 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,  -2 },    125.60605,   0.00013,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },    305.99416,   0.04570,     0.094 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   2 },    305.37331,   0.00002,     0.062 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   2,   0 },    305.97085,   0.00196,     0.384 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   3,   0 },    308.15563,   0.00010,     0.063 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -2,   1 },    280.51623,   0.00005,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,  -1 },    280.50617,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   0 },    125.78375,   0.00015,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   1 },    280.51864,   0.00069,     0.088 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,  -1 },    280.50828,   0.00016,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0 },    125.85704,   0.00233,     0.087 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   1 },    100.50269,   0.00104,     0.530 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,  -1 },    280.53888,   0.00040,     0.085 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,   0 },    125.98196,   0.00025,     0.582 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,   1 },    100.55030,   0.00018,     0.066 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   2,  -1 },    280.51641,   0.00003,     0.644 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   2,   0 },    125.80993,   0.00002,     0.067 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -3,   0 },    300.04224,   0.00020,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   0 },    300.11507,   0.00261,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,  -2 },    122.26721,   0.00009,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   0 },    301.00167,   0.03780,     0.080 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   2 },    291.77386,   0.00002,     0.069 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -2 },    118.84971,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -1 },    100.51448,   0.00002,     0.079 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },    301.00599,   0.16425,     1.203 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   2 },    121.44426,   0.00011,     0.036 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   0 },    301.49768,   0.02889,     0.071 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   2 },    122.03429,   0.00008,     0.024 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   2,   0 },    301.26671,   0.00202,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   3,   0 },    301.52272,   0.00013,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -3,   0 },    308.25410,   0.00001,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -3,   1 },    280.54049,   0.00002,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,  -1 },    100.52254,   0.00008,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   0 },    310.57872,   0.00018,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   1 },    280.53725,   0.00055,     8.998 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,  -1 },    100.52330,   0.00155,     0.074 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   0 },    305.89040,   0.00327,    17.980 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   1 },    100.52447,   0.00445,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -3 },    283.50810,   0.00001,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -1 },    100.53010,   0.01183,     4.497 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },    306.19317,   0.00697,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   1 },    280.50610,   0.00203,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,  -1 },    100.52417,   0.00165,     0.077 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   0 },    306.13874,   0.00047,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   1 },    280.50699,   0.00024,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   2,  -1 },    100.52321,   0.00011,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   2,   0 },    306.09015,   0.00003,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   2,   1 },    280.50738,   0.00002,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -4,   0 },    292.64492,   0.00001,     0.036 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -3,   0 },    275.01305,   0.00004,     0.070 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -2,   0 },    288.46989,   0.00049,     1.061 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },    301.71105,   0.01819,     0.081 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   2 },    130.79514,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,  -2 },    304.98612,   0.00040,     0.784 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },    303.96800,   0.00872,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   2 },    125.80893,   0.00002,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   1,  -2 },    125.00996,   0.00002,     0.083 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   1,   0 },    303.24276,   0.00099,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   2,   0 },    305.07496,   0.00010,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,  -1 },    280.51840,   0.00002,     0.501 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,   0 },    305.54406,   0.00003,     0.088 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,   1 },    100.52410,   0.00006,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,  -1 },    280.47228,   0.00007,     0.089 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,   0 },    305.92318,   0.00012,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,  -1 },    100.66050,   0.00003,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,   0 },    305.89891,   0.00008,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,   1 },    280.50722,   0.00003,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -3,   0 },    103.08555,   0.00018,     0.095 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -2,   0 },    289.48098,   0.00069,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -1,   0 },    289.22982,   0.00098,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,   0,   0 },    297.36032,   0.00019,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,   1,   0 },    304.95958,   0.00001,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   6,  -3,   0 },    286.19275,   0.00002,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   6,  -2,   0 },    283.29088,   0.00004,     0.021 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   6,  -1,   0 },    283.05891,   0.00001,     0.016 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,  -2,   0,   0 },    182.97699,   0.00006,     0.042 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,  -2,   1,   0 },    182.97699,   0.00007,     0.094 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   0,  -1,   0 },    182.53313,   0.00005,     0.081 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   0,   0,   0 },    181.62176,   0.00019,     1.156 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   0,   1,   0 },    182.53313,   0.00005,     0.071 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   2,  -1,   0 },    181.10935,   0.00003,     0.081 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   2,   0,   0 },    181.10937,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,  -2,  -1,   0 },     28.72600,   0.00001,     0.027 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,  -2,   0,   0 },     28.72600,   0.00010,     0.042 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,  -2,   1,   0 },     29.40456,   0.00013,     0.094 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   0,  -1,   0 },     23.13930,   0.00010,     0.081 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   0,   0,   0 },     22.87782,   0.00039,     1.112 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   0,   1,   0 },     20.60965,   0.00007,     0.071 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   1,  -1,   0 },     12.80519,   0.00008,    81.442 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   1,   0,  -1 },    114.86911,   0.00002,     6.474 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   1,   0,   0 },    191.06371,   0.00007,     0.075 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   2,  -1,   0 },      3.51477,   0.00003,     0.081 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   2,   0,   0 },      9.77407,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -2,   3,   0,   0,  -2,   0,   0 },    177.03037,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -2,   3,   0,   0,   0,   0,   0 },    237.17738,   0.00006,     1.072 },
 { {   0,   0,   1,   0,  -2,   3,   0,   0,   1,   0,  -1 },    108.70734,   0.00004,     8.297 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,  -2,   0,   0 },    348.12064,   0.00006,     0.042 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,  -2,   1,   0 },    348.12065,   0.00007,     0.095 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,  -1,   0 },    349.13373,   0.00005,     0.081 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,   0,   0 },    350.55637,   0.00026,     1.034 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,   1,   0 },    349.13373,   0.00005,     0.070 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   1,   0,  -1 },    186.90352,   0.00001,    11.550 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   2,  -1,   0 },    351.98818,   0.00003,     0.080 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   2,   0,   0 },    346.48027,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -4,   1,   0 },    244.13912,   0.00002,     0.028 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -4,   2,   0 },    269.65440,   0.00001,     0.046 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,  -1,   0 },    128.30763,   0.00005,     0.027 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   0,   0 },    193.44975,   0.00028,     0.042 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   1,   0 },    244.37235,   0.00078,     0.095 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   2,   0 },    253.99687,   0.00006,     0.360 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -2,   0 },     97.68447,   0.00006,     0.039 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -1,   0 },    112.55607,   0.00042,     0.082 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   0,   0 },    200.60998,   0.00068,     0.999 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   1,   0 },    256.23495,   0.00037,     0.070 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   2,   0 },    259.54650,   0.00003,     0.036 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,   0,  -1 },    266.23098,   0.00004,    19.000 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,   0,   1 },     86.29926,   0.00005,     0.037 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,  -3,   0 },    343.74828,   0.00001,     0.071 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,  -2,   0 },    215.12622,   0.00001,     1.294 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,  -1,   0 },     86.05616,   0.00007,     0.080 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,   0,   0 },    181.73904,   0.00003,     0.039 },
 { {   0,   0,   1,   0,  -2,   6,   0,   0,   1,   0,  -1 },     90.45864,   0.00003,    53.519 },
 { {   0,   0,   1,   0,  -2,   7,   0,   0,   1,   0,  -1 },     48.45830,   0.00001,    65.520 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,  -2,   0,   0 },    318.20729,   0.00004,     0.042 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,  -2,   1,   0 },    318.20763,   0.00005,     0.093 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   0,  -1,   0 },    314.50609,   0.00004,     0.080 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   0,   0,   0 },    310.64360,   0.00019,     1.341 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   0,   1,   0 },    314.50609,   0.00004,     0.071 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   2,   0,   0 },    319.24170,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -1,  -3,   0,   0,   0,   0,   0 },    338.79903,   0.00001,     1.229 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,  -2,   0,   0 },    107.01942,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,  -2,   1,   0 },    107.01944,   0.00002,     0.094 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,   0,  -1,   0 },    107.01961,   0.00001,     0.081 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,   0,   0,   0 },    106.34498,   0.00012,     1.180 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,   0,   1,   0 },    107.01961,   0.00001,     0.071 },
 { {   0,   0,   1,   0,  -1,  -1,   0,   0,  -2,   0,   0 },    310.20686,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -1,  -1,   0,   0,   0,   0,   0 },    307.27779,   0.00005,     1.134 },
 { {   0,   0,   1,   0,  -1,  -1,   0,   0,   1,   0,  -1 },    293.88307,   0.00001,     5.813 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,  -1,   0 },      1.89589,   0.00001,     0.012 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   0,   0 },      0.49190,   0.00003,     0.014 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   1,   0 },      1.35464,   0.00008,     0.017 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   2,   0 },      6.79550,   0.00009,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   3,   0 },      1.72902,   0.00004,     0.030 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -5,   1,   0 },      2.10260,   0.00001,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -2,   0 },      1.06774,   0.00005,     0.013 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -1,   0 },    359.98801,   0.00028,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -1,   2 },      1.76835,   0.00001,     0.029 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,   0 },      2.84647,   0.00188,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,   2 },      1.78234,   0.00005,     0.046 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,  -2 },    181.69844,   0.00002,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,   0 },      2.42316,   0.00443,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,   2 },      2.13908,   0.00002,     0.118 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   2,  -2 },    181.66913,   0.00001,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   2,   0 },      2.18960,   0.00282,     0.045 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   3,   0 },      1.64297,   0.00007,     0.114 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,  -2,   0 },      1.83938,   0.00001,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,  -1,   0 },      1.67504,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   0,   0 },      1.48038,   0.00015,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   1,   0 },    183.90071,   0.00009,     0.044 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   2,   0 },    182.24984,   0.00008,     0.103 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -3,   0 },      1.63224,   0.00011,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -2,   0 },      1.59954,   0.00119,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -2,   2 },    181.81898,   0.00002,     0.043 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,  -2 },    181.65124,   0.00007,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,   0 },      1.96510,   0.01455,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,   2 },    181.60944,   0.00035,     0.098 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,  -2 },    181.66637,   0.00040,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },      1.80280,   0.16497,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   2 },      1.80214,   0.00228,     0.331 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,  -2 },    181.71558,   0.00045,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,   0 },      1.74333,   0.21118,     0.095 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,   2 },    181.63980,   0.00006,     0.061 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,  -2 },    181.54035,   0.00003,     0.041 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,   0 },      1.70905,   0.00889,     0.372 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,   2 },    181.73432,   0.00005,     0.034 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   3,   0 },      2.18550,   0.00053,     0.063 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   4,   0 },      2.21086,   0.00004,     0.034 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -2,   0 },      0.65203,   0.00008,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0 },      0.56029,   0.00080,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   1 },    277.89267,   0.00004,     0.089 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,  -2 },    180.71018,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,  -1 },    274.61602,   0.00001,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0 },      0.62111,   0.00978,     0.087 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   1 },     96.66415,   0.00008,     0.508 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,  -1 },    275.64169,   0.00003,     0.086 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   0 },      0.51838,   0.00074,     0.555 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   1 },     93.04832,   0.00001,     0.066 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   2,   0 },      0.73364,   0.00008,     0.066 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -4,   0 },      0.96831,   0.00004,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -3,   0 },      0.89273,   0.00067,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,  -2 },    181.16483,   0.00006,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,   0 },      1.25695,   0.00943,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,   2 },    181.75147,   0.00003,     0.801 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,  -2 },    181.35410,   0.00031,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   0 },      1.23582,   0.14347,     0.081 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   2 },    358.13664,   0.00006,     0.069 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -2 },    183.26644,   0.00008,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },      1.22890,   0.63880,     1.092 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2 },    181.20410,   0.00032,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,  -2 },    348.26208,   0.00004,     0.079 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   0 },      1.11168,   0.10993,     0.071 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   2 },    181.16154,   0.00027,     0.024 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,  -2 },    183.12005,   0.00002,     1.718 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   0 },      1.03637,   0.00766,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   2 },    180.93895,   0.00004,     0.018 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   3,   0 },      1.11285,   0.00055,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   4,   0 },      1.06401,   0.00004,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -3,   0 },    181.13764,   0.00005,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,  -1 },     97.00740,   0.00001,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,   0 },    180.86275,   0.00049,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,   1 },    277.38904,   0.00001,     5.117 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,  -2 },    181.06667,   0.00003,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,  -1 },     96.64287,   0.00021,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   0 },      0.69736,   0.02338,    34.860 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   1 },     96.67509,   0.00024,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   2 },    181.06667,   0.00003,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -2 },    180.73433,   0.00011,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -1 },     96.79066,   0.00160,     7.243 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   0 },    180.70913,   0.06711,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   1 },    276.84940,   0.00035,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   2 },      0.71019,   0.00014,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,  -1 },     96.64287,   0.00021,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   0 },    180.71141,   0.00460,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   1 },    276.81041,   0.00004,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   2 },      0.71019,   0.00002,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,  -2 },    180.46107,   0.00001,     0.077 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,   0 },    180.70831,   0.00033,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   3,   0 },    180.71019,   0.00002,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -4,   0 },    182.37915,   0.00001,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    182.16774,   0.00008,     0.071 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    179.69145,   0.00083,     1.165 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,  -2 },      1.62934,   0.00003,     0.069 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,   0 },      0.57920,   0.05108,     0.081 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,   2 },    180.81680,   0.00011,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,  -2 },      0.31118,   0.00162,     0.839 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },      0.43396,   0.03111,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,   2 },    181.51628,   0.00005,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,  -2 },    179.20290,   0.00008,     0.083 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,   0 },      0.42724,   0.00346,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   2,   0 },      1.42808,   0.00031,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   3,   0 },      0.65552,   0.00003,     0.015 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -3,   0 },      0.73734,   0.00003,     0.573 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -2,   0 },      0.74424,   0.00061,     0.087 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,   0 },    180.69106,   0.00066,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,   0 },    180.74540,   0.00060,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   1,   0 },    180.71015,   0.00009,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -3,   0 },    181.42944,   0.00002,     0.094 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -2,   0 },      1.31560,   0.00069,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,  -2 },      0.22172,   0.00003,     0.097 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,   0 },    359.77109,   0.00093,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,  -2 },      0.28365,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,   0 },      1.64409,   0.00037,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   1,   0 },      0.63262,   0.00008,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   5,  -1,   0 },    180.70929,   0.00002,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   6,  -2,   0 },      0.81201,   0.00002,     0.021 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,  -2,   0,   0 },    225.45822,   0.00004,     0.042 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,  -2,   1,   0 },    225.45822,   0.00004,     0.095 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,   0,  -1,   0 },    219.57781,   0.00001,     0.081 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,   0,   0,   0 },    206.65393,   0.00009,     1.053 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,   0,   1,   0 },    219.57781,   0.00001,     0.070 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,   1,   0,  -1 },    278.33701,   0.00001,     9.604 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,  -2,   0,   0 },     49.88717,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,  -2,   1,   0 },     48.16415,   0.00004,     0.095 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,   0,  -1,   0 },     36.61654,   0.00001,     0.081 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,   0,   0,   0 },     14.77425,   0.00007,     1.017 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,   0,   1,   0 },     36.61654,   0.00001,     0.070 },
 { {   0,   0,   1,   0,  -1,   3,   0,   0,   0,   0,   0 },    189.44046,   0.00002,     0.983 },
 { {   0,   0,   1,   0,  -1,   4,   0,   0,   1,   0,  -1 },     43.79417,   0.00003,   439.132 },
 { {   0,   0,   1,   0,  -1,   5,   0,   0,  -2,   0,   0 },    234.93438,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -1,   5,   0,   0,   0,   0,   0 },    222.79349,   0.00002,     0.921 },
 { {   0,   0,   1,   0,  -1,   5,   0,   0,   1,   0,  -1 },    105.76899,   0.00004,    31.575 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,  -1,   0 },    314.84897,   0.00003,     0.027 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   0,   0 },    314.55212,   0.00018,     0.042 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   1,   0 },    314.86532,   0.00025,     0.094 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,  -1,   0 },    311.70887,   0.00023,     0.080 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   0,   0 },    307.67562,   0.00093,     1.204 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   1,   0 },    310.48677,   0.00017,     0.071 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   2,   0 },    312.43034,   0.00001,     0.037 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   1,  -1,   1 },    109.26887,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   1,   0,  -1 },    109.14854,   0.00007,     4.474 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   1,   0,   0 },    309.96776,   0.00003,     0.076 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   2,  -3,   0 },    198.39726,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   2,  -1,   0 },    298.49700,   0.00011,     0.081 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   2,   0,   0 },    313.30902,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,  -2,   0,   0 },    108.78012,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,  -2,   1,   0 },    109.04454,   0.00005,     0.094 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   0,  -1,   0 },    110.45712,   0.00004,     0.081 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   0,   0,   0 },    113.27543,   0.00019,     1.157 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   0,   1,   0 },    110.45712,   0.00004,     0.071 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   1,  -1,   0 },    110.38551,   0.00001,    43.858 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   1,   0,  -1 },    299.23779,   0.00003,     5.276 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   2,  -1,   0 },    108.76629,   0.00002,     0.081 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -4,   1,   0 },     16.83284,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,  -1,   0 },     16.68556,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   0,   0 },     16.50939,   0.00038,     0.042 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   1,   0 },     17.03548,   0.00046,     0.094 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   2,   0 },     17.04571,   0.00001,     0.374 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -1,   0,   0 },     17.01466,   0.00008,     0.087 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -1,   0,   1 },    206.54776,   0.00004,     0.512 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,  -2,   0 },     16.92896,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,  -1,   0 },     16.84004,   0.00037,     0.081 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0 },     16.54744,   0.00172,     1.113 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   1,   0 },     16.76787,   0.00027,     0.071 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   2,   0 },     16.93190,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,  -1 },    206.53066,   0.00005,     0.075 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,   0 },     16.97403,   0.00045,    89.713 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,   1 },    206.53071,   0.00007,     0.074 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,  -1 },    206.52104,   0.00037,     6.426 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,   0 },    197.01738,   0.00037,     0.075 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,   1 },     26.52278,   0.00008,     0.037 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   1,  -1 },    206.53066,   0.00005,     0.076 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   1,   0 },    197.01467,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   2,  -1,   0 },     16.79434,   0.00013,     0.081 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   2,   0,   0 },     16.94454,   0.00007,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   0,   0 },    285.95026,   0.00003,     0.021 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   1,   0 },    285.82211,   0.00007,     0.028 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   2,   0 },    285.94679,   0.00005,     0.045 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,  -2,   0 },    286.00582,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,  -1,   0 },    285.68158,   0.00021,     0.027 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,   0 },    285.71665,   0.00245,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,   2 },    286.27724,   0.00004,     0.329 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   1,   0 },    285.98662,   0.00312,     0.095 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   2,   0 },    286.40825,   0.00014,     0.370 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,  -1,   1 },    293.92221,   0.00005,     0.089 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,  -1 },    293.92192,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,   0 },    286.30010,   0.00008,     0.087 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,   1 },    113.92209,   0.00018,     0.504 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   1,  -1 },    293.92209,   0.00006,     0.086 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   1,   1 },    113.92218,   0.00004,     0.066 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -2,   0 },    285.61471,   0.00017,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -1,   0 },    285.65697,   0.00245,     0.081 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0 },    285.55600,   0.01096,     1.073 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   1,   0 },    285.69271,   0.00186,     0.070 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   2,   0 },    285.63324,   0.00012,     0.036 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   3,   0 },    105.97684,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -2,  -1 },    113.92217,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -2,   0 },    110.51666,   0.00001,     0.076 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -2,   1 },    293.92231,   0.00002,     4.720 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,  -1 },    113.92217,   0.00039,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   0 },    285.92979,   0.00010,    22.176 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   1 },    113.92219,   0.00038,     0.074 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,  -1 },    113.92222,   0.00307,     8.220 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   0 },    106.27291,   0.00066,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   1 },    293.92222,   0.00073,     0.037 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,  -1 },    113.92217,   0.00039,     0.076 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,   0 },    106.27293,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,   1 },    293.92221,   0.00008,     0.025 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   2,  -1 },    113.92217,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -2,   0 },    107.46550,   0.00002,     1.188 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -1,   0 },    286.69128,   0.00062,     0.081 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   0,  -2 },    286.40676,   0.00002,     0.851 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   0,   0 },    286.25059,   0.00042,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   1,   0 },    286.07365,   0.00006,     0.026 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,  -1,  -1 },    293.92207,   0.00003,     0.088 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,  -1,   1 },    293.92221,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   4,  -2,   0 },    283.11632,   0.00001,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   4,  -1,   0 },    286.49969,   0.00001,     0.027 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,  -2,   0 },      0.72216,   0.00001,     0.013 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   0,   0 },      0.22081,   0.00012,     0.021 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   1,   0 },      0.66907,   0.00027,     0.028 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   2,   0 },      0.70082,   0.00013,     0.045 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -3,   0,   0 },      0.84442,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -2,   0 },      0.68489,   0.00009,     0.020 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -1,   0 },      0.64394,   0.00085,     0.027 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -1,   2 },    180.69733,   0.00002,     0.098 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,  -2 },    180.66154,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,   0 },      0.72506,   0.00959,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,   2 },      0.71252,   0.00014,     0.325 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   1,   0 },      0.68492,   0.01211,     0.095 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   1,   2 },    180.69733,   0.00002,     0.061 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.75517,   0.00053,     0.365 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   3,   0 },      0.81129,   0.00003,     0.063 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,  -1,   0 },      0.29648,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,   0,   0 },    358.50648,   0.00004,     0.088 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,   0,   1 },    169.36160,   0.00003,     0.495 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -3,   0 },      0.44799,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   0 },      0.48417,   0.00058,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,  -2 },    180.45616,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   0 },      0.45799,   0.00863,     0.081 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   2 },      0.46461,   0.00001,     0.069 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -2 },    180.52663,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0 },      0.43296,   0.03897,     1.035 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   2 },    180.59961,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,  -2 },      0.46461,   0.00001,     0.079 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,   0 },      0.40477,   0.00654,     0.070 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,   2 },    180.45616,   0.00003,     0.024 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   2,   0 },      0.41053,   0.00044,     0.036 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   3,   0 },      0.28688,   0.00003,     0.025 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -2,   0 },    181.88657,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,  -1 },    170.10657,   0.00008,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,   0 },    181.44488,   0.00001,    12.652 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,   1 },    170.16104,   0.00006,     0.074 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,  -1 },    170.12632,   0.00066,    11.401 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   0 },    180.31263,   0.00154,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   1 },    350.12312,   0.00020,     0.037 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,  -1 },    170.10657,   0.00008,     0.076 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,   0 },    180.31836,   0.00011,     0.038 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,   1 },    350.12312,   0.00002,     0.025 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -2,   0 },    179.99601,   0.00011,     1.238 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -1,   0 },      0.18215,   0.00225,     0.080 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   0,  -2 },      0.22788,   0.00009,     0.876 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   0,   0 },      0.14437,   0.00164,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   1,   0 },      0.41038,   0.00020,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   3,  -1,   0 },    180.29648,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   4,  -2,   0 },      0.27118,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   4,  -1,   0 },      0.44369,   0.00003,     0.027 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   4,   0,   0 },      0.14700,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,   0,  -3,   0,  -2,   0,   0 },     92.89554,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   0,   0 },     92.95149,   0.00007,     1.037 },
 { {   0,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0 },    100.70523,   0.00001,     1.024 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,  -1,   0 },    276.99875,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   0,   0 },    276.99852,   0.00013,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   1,   0 },    276.99629,   0.00018,     0.095 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,  -1,   0 },    277.01334,   0.00015,     0.081 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0 },    276.99993,   0.00068,     1.024 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0 },    277.01924,   0.00011,     0.070 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,  -1 },     74.07492,   0.00004,    12.892 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   1 },    254.07509,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -1,   0 },    276.18312,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0 },    276.82618,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,   0,  -1,  -2,   0,   0,   0 },    273.08670,   0.00001,     1.025 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,  -1,   0 },    353.64441,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0 },    353.72737,   0.00023,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   1,   0 },    353.57297,   0.00031,     0.095 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0 },    354.94461,   0.00025,     0.082 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0 },    356.07118,   0.00104,     1.012 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0 },    355.46605,   0.00017,     0.070 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0 },    354.70883,   0.00001,     0.036 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0 },    176.64363,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -1,   0 },      7.27170,   0.00004,     0.080 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0 },    358.44811,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   0,   0 },    283.59447,   0.00009,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   1,   0 },    283.59496,   0.00012,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,  -1,   0 },    283.58667,   0.00010,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0 },    283.57315,   0.00042,     1.012 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0 },    283.58574,   0.00009,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0,  -1 },    131.78514,   0.00006,    15.174 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0,   1 },    311.78513,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   2,  -1,   0 },    283.53188,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0,   0 },    283.56864,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0 },      1.69966,   0.00009,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -2,   1,   0 },      1.71613,   0.00011,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0 },      1.37850,   0.00009,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0 },      0.87907,   0.00041,     1.006 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0 },      1.37850,   0.00009,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   2,  -1,   0 },    358.20265,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   2,   0,   0 },      0.34816,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   0 },     72.28017,   0.00012,     0.021 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   1,   0 },     66.93395,   0.00012,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   2,   0 },     68.06377,   0.00007,     0.046 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -2,   0 },     60.29767,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,   0 },     67.94537,   0.00030,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.63066,   0.00309,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1 },    103.34668,   0.00002,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   2 },     61.97076,   0.00004,     0.322 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },     63.15579,   0.00390,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },     60.56540,   0.00019,     0.360 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0 },     89.51559,   0.00004,     0.062 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -2,   1 },    112.42127,   0.00001,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     95.13371,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   0 },     84.88227,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   1 },    278.39287,   0.00010,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -1 },     91.71870,   0.00022,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    108.92132,   0.00004,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    279.63289,   0.00112,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,  -1 },     91.64297,   0.00044,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   1 },    287.43049,   0.00021,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   2,  -1 },     88.90418,   0.00003,     0.581 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   1 },    312.27707,   0.00003,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0 },     91.64231,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.34874,   0.00025,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.29781,   0.00355,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },    281.28859,   0.00010,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     95.38394,   0.01699,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },    103.35181,   0.00008,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1 },    283.35135,   0.00001,     1.200 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     88.40924,   0.00267,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0 },     89.85306,   0.00019,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    276.12701,   0.00007,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    274.42730,   0.00091,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0 },    262.19415,   0.00008,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   1 },    288.20132,   0.00085,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    298.97935,   0.00456,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    264.99536,   0.00116,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    112.29596,   0.00625,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -1 },    274.14807,   0.00100,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0 },    264.80305,   0.00008,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   1 },    111.92507,   0.00070,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   2,  -1 },    271.43734,   0.00006,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   2,   1 },    109.28790,   0.00007,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    293.73261,   0.00002,     0.071 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    303.05433,   0.00020,     1.292 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },     16.06155,   0.00242,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -2 },     52.61615,   0.00004,     0.903 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },     99.83747,   0.00006,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     45.43483,   0.00093,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   1,   0 },     57.73479,   0.00012,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -3,  -1 },     77.06307,   0.00002,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -2,  -1 },     85.72930,   0.00001,     0.547 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -2,   1 },    289.29715,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,  -1 },     96.60247,   0.00022,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,   0 },    264.88227,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,   1 },    111.96123,   0.00014,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,  -1 },    111.43777,   0.00021,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,   1 },    111.96122,   0.00009,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   1,  -1 },    111.96395,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -2,   0 },     35.66816,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -1,   0 },      5.70828,   0.00003,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   4,   0,   0 },    354.15150,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   1,  -2,  -2,   0,   0 },    163.79136,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0 },    176.83445,   0.00006,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0 },     83.74899,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0 },     84.42506,   0.00001,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0 },     86.97644,   0.00004,     0.988 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,   0,   1,   0 },     84.42506,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   0,   1,  -3,   0,  -2,   0,   0 },     24.73129,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0 },     23.31355,   0.00003,     1.002 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -4,   1,   0 },    165.68770,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,  -1,   0 },    165.72063,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   0,   0 },    165.36176,   0.00044,     0.042 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   1,   0 },    166.62476,   0.00047,     0.096 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   2,   0 },    165.65309,   0.00003,     0.356 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   0,  -1 },    334.35754,   0.00004,     0.040 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   0,   1 },    334.90504,   0.00004,     0.479 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   1,  -1 },    154.07284,   0.00010,     0.087 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   2,  -1 },    154.14483,   0.00001,     0.570 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,  -1,   0 },    164.60535,   0.00022,     0.082 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0 },    167.08140,   0.00066,     0.967 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,   1,   0 },    171.44269,   0.00006,     0.070 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -2,  -1 },    334.18174,   0.00007,     0.038 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,  -1 },    334.18174,   0.00099,     0.075 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,   1 },    334.29832,   0.00002,     0.074 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,  -1 },    334.27632,   0.00507,    50.462 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,   1 },    334.18806,   0.00057,     0.037 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   1,  -1 },    334.18174,   0.00099,     0.076 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   1,   1 },    334.18959,   0.00006,     0.025 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   2,  -1 },    334.18174,   0.00007,     0.038 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   2,  -1,   0 },    348.99977,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   2,   0,   0 },    348.37132,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,  -2,  -1 },    154.14483,   0.00001,     0.557 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,  -1,  -1 },    154.04424,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,   0,  -1 },    334.24135,   0.00008,     0.040 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -2,   0,   0 },     63.32997,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -2,   1,   0 },     63.41285,   0.00002,     0.096 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -1,   0,  -1 },    236.29092,   0.00001,     0.040 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -1,   1,  -1 },     57.45282,   0.00001,     0.087 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   0,   0,   0 },     66.09502,   0.00004,     0.936 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,  -1,  -1 },    236.72933,   0.00017,     0.076 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,  -1 },    236.29092,   0.00095,    70.769 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,   1 },     55.59023,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   1,  -1 },    236.72933,   0.00017,     0.075 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   3,  -1,  -1 },     57.45282,   0.00001,     0.087 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   3,   0,  -1 },    236.29092,   0.00001,     0.040 },
 { {   0,   0,   1,   0,   0,   3,   0,   0,   1,   0,  -1 },    139.65231,   0.00001,    20.799 },
 { {   0,   0,   1,   0,   1,  -6,   3,   0,   0,   0,   0 },    192.39273,   0.00002,     1.091 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -4,   1,   0 },     11.66801,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,  -1,   0 },     11.69393,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,   0 },     11.70527,   0.00045,     0.042 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   1,   0 },     11.62128,   0.00061,     0.095 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   2,   0 },     11.67816,   0.00002,     0.372 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -1,   0,   0 },     11.20761,   0.00004,     0.087 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,  -2,   0 },     11.93144,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,  -1,   0 },     11.83028,   0.00041,     0.081 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   0,   0 },     11.17867,   0.00182,     1.093 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   1,   0 },     11.74529,   0.00029,     0.071 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   2,   0 },     11.96065,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   1,  -1,   0 },     10.91598,   0.00007,    36.292 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   1,   0,   0 },    191.20763,   0.00019,     0.075 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,  -3,   0 },    189.78991,   0.00001,     0.071 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,  -1,   0 },      8.78702,   0.00012,     0.081 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,   0,   0 },     10.38319,   0.00007,     0.039 },
 { {   0,   0,   1,   0,   1,  -4,   0,   0,  -2,   0,   0 },    239.60996,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   1,  -4,   0,   0,   0,   0,   0 },    239.60591,   0.00003,     1.054 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,  -2,  -1,   0 },    308.90079,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,  -2,   0,   0 },    308.90079,   0.00009,     0.042 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,  -2,   1,   0 },    308.90705,   0.00012,     0.095 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   0,  -1,   0 },    308.37261,   0.00009,     0.081 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   0,   0,   0 },    306.81822,   0.00040,     1.018 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   0,   1,   0 },    308.15254,   0.00007,     0.070 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   1,   0,  -1 },    136.65653,   0.00001,    14.024 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   2,  -1,   0 },    298.88018,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   2,   0,   0 },    305.22858,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -4,   1,   0 },    183.23221,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,  -1,   0 },    183.21251,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   0,   0 },    183.15774,   0.00045,     0.042 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   1,   0 },    183.22892,   0.00057,     0.096 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   2,   0 },    183.26614,   0.00003,     0.358 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,  -2,   0 },    182.25835,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,  -1,   0 },    182.07120,   0.00043,     0.082 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   0,   0 },    181.05494,   0.00186,     0.984 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   1,   0 },    181.59678,   0.00027,     0.070 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   2,   0 },    182.36747,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   1,   0,  -1 },     79.18208,   0.00002,    26.769 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   1,   0,   0 },      0.57133,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   2,  -1,   0 },    175.51858,   0.00007,     0.080 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   2,   0,   0 },    179.15315,   0.00006,     0.039 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -2,   0,   0 },     66.36315,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -2,   1,   0 },     66.36315,   0.00002,     0.096 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   0,  -1 },    273.73523,   0.00006,     0.040 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   1,  -1 },     94.19933,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   2,  -1 },     93.99044,   0.00001,     0.565 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   0,   0,   0 },     34.37238,   0.00004,     0.952 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -2,  -1 },    273.91123,   0.00006,     0.038 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -1,  -1 },    273.91123,   0.00087,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,  -1 },    273.73523,   0.00491,   293.315 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,   1 },    274.41768,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   1,  -1 },    273.91123,   0.00087,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   2,  -1 },    273.91123,   0.00006,     0.038 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   3,  -2,  -1 },     93.99044,   0.00001,     0.563 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   3,  -1,  -1 },     94.19933,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   3,   0,  -1 },    273.73523,   0.00006,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,  -1,   0 },     72.57794,   0.00002,     0.016 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   0,   0 },    246.27304,   0.00009,     0.021 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   1,   0 },    249.37683,   0.00017,     0.028 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   2,   0 },    251.65810,   0.00009,     0.046 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -3,   0,   0 },    247.55197,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -2,   0 },    244.67831,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -1,   0 },    241.02063,   0.00035,     0.027 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   0 },    244.62727,   0.00408,     0.042 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   2 },    245.48442,   0.00006,     0.313 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   1,   0 },    247.51034,   0.00474,     0.096 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   2,   0 },    251.71272,   0.00022,     0.350 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,  -1,  -1 },     75.53653,   0.00006,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,  -1,   1 },     76.89821,   0.00003,     0.090 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,  -1 },     75.54788,   0.00041,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,   0 },     59.88138,   0.00003,     0.089 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,   1 },    256.09304,   0.00036,     0.468 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   1,  -1 },    255.89411,   0.00022,     0.087 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   1,   1 },    254.50488,   0.00002,     0.065 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   2,  -1 },    255.37659,   0.00006,     0.554 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -2,   0 },    226.44437,   0.00014,     0.039 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -1,   0 },    227.42682,   0.00188,     0.082 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0 },    235.35696,   0.00720,     0.922 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   1,   0 },    241.15890,   0.00109,     0.070 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   2,   0 },    245.65800,   0.00005,     0.036 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -3,  -1 },     75.62742,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -2,  -1 },     75.35587,   0.00029,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,  -1 },     75.57587,   0.00426,     0.076 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,   0 },     57.70321,   0.00002,     5.069 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,   1 },    255.79887,   0.00029,     0.073 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,  -3 },    257.53717,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,  -1 },     75.47182,   0.02540,    32.746 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   0 },     38.92634,   0.00007,     0.074 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   1 },    255.51532,   0.00153,     0.037 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   1,  -1 },     75.57867,   0.00424,     0.075 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   1,   1 },    255.50852,   0.00016,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   2,  -1 },     75.57727,   0.00029,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   3,  -1 },     75.62742,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,  -2,   0 },     19.85100,   0.00004,     1.450 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },     60.53865,   0.00017,     0.080 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,   0,   0 },     96.92248,   0.00010,     0.039 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -2,  -1 },    255.62742,   0.00005,     0.574 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -1,  -1 },    255.74048,   0.00033,     0.087 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -1,   1 },    255.50861,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   0,  -1 },     75.47718,   0.00023,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   0,   1 },    255.50880,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   1,  -1 },     75.53653,   0.00006,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   4,  -1,   0 },     67.03728,   0.00001,     0.027 },
 { {   0,   0,   1,   0,   1,   5,   0,   0,   0,  -1,   0 },    299.54401,   0.00002,     0.083 },
 { {   0,   0,   1,   0,   1,   5,   0,   0,   0,   1,   0 },    299.54401,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   1,   5,   0,   0,   2,  -2,   0 },    299.68645,   0.00001,     1.923 },
 { {   0,   0,   1,   0,   1,   5,   0,   0,   2,   0,   0 },    299.68645,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   2,  -8,   0,   0,   0,   0,   0 },    205.29039,   0.00003,     1.115 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,  -2,   0,   0 },    113.76645,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,  -2,   1,   0 },    113.76645,   0.00005,     0.095 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   0,  -1,   0 },    113.66936,   0.00003,     0.081 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   0,   0,   0 },    113.55864,   0.00016,     1.074 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   0,   1,   0 },    113.66936,   0.00003,     0.070 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   1,   0,  -1 },    302.63460,   0.00004,     8.144 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,  -2,   0,   0 },    191.22542,   0.00008,     0.042 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,  -2,   1,   0 },    191.23019,   0.00008,     0.095 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,  -1,   0 },    191.09835,   0.00006,     0.081 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   0,   0 },    190.86072,   0.00028,     1.036 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   1,   0 },    191.09835,   0.00006,     0.070 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   2,  -1,   0 },    189.89382,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   2,   0,   0 },    190.32994,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,  -2,   0,   0 },    257.32753,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,  -2,   1,   0 },    257.32753,   0.00002,     0.095 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,   0,  -1,   0 },    257.32753,   0.00001,     0.082 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,   0,   0,   0 },    255.73960,   0.00009,     0.999 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,   0,   1,   0 },    257.32753,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,  -1,   0 },     67.01560,   0.00001,     0.016 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   0,   0 },     65.53301,   0.00005,     0.021 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   1,   0 },     66.52706,   0.00014,     0.028 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   2,   0 },     66.95799,   0.00008,     0.046 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -2,   0 },     66.93071,   0.00004,     0.020 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -1,   0 },     67.26564,   0.00032,     0.027 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,  -2 },    246.82026,   0.00001,     0.020 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   0 },     67.17110,   0.00423,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   2 },     67.08534,   0.00006,     0.322 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,   0 },     66.72935,   0.00586,     0.095 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   2,   0 },     66.57709,   0.00029,     0.361 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   3,   0 },     67.22225,   0.00002,     0.062 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,  -1,   1 },     58.99911,   0.00002,     0.089 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,   0,   0 },    255.30291,   0.00001,     0.088 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,   0,   1 },    239.04433,   0.00003,     0.487 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -3,   0 },     65.68512,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -2,   0 },     66.66060,   0.00019,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,   0 },     65.89732,   0.00331,     0.082 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   0,   0 },     65.64939,   0.01696,     1.001 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,   0 },     65.39064,   0.00313,     0.070 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   2,   0 },     64.96312,   0.00022,     0.036 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,  -1,  -1 },    239.05526,   0.00001,     0.075 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,  -1,   0 },    246.60583,   0.00004,     8.940 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,  -1,   1 },    239.05342,   0.00008,     0.074 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,  -1 },    239.05618,   0.00070,    18.216 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   0 },    245.30844,   0.00046,     0.075 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   1 },     59.05928,   0.00064,     0.037 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   1,  -1 },    239.05526,   0.00001,     0.076 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   1,   0 },    245.35194,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   1,   1 },     59.05936,   0.00007,     0.025 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,  -2,   0 },    259.23696,   0.00005,     1.290 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,  -1,   0 },     65.18715,   0.00063,     0.080 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,  -2 },     64.63706,   0.00004,     0.902 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,   0 },     62.22515,   0.00065,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   1,   0 },     64.59615,   0.00010,     0.026 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   3,   0,  -1 },     59.06088,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   4,  -2,   0 },     78.00011,   0.00001,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   4,  -1,   0 },     64.47137,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,  -2,   0,   0 },    283.38647,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   0,   0,   0 },    295.93949,   0.00010,     0.968 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,  -1,  -1 },    322.38408,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   0,  -1 },    323.17235,   0.00013,    47.735 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   0,   1 },    323.19213,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   1,  -1 },    322.38408,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,  -2,   0,   0 },    183.77514,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,  -2,   1,   0 },    183.77514,   0.00004,     0.096 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   0,  -1,   0 },    183.77514,   0.00001,     0.082 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   0,   0,   0 },    175.38763,   0.00009,     0.937 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   0,   1,   0 },    183.77514,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   1,  -1,  -1 },     30.02094,   0.00003,     0.076 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   1,   0,  -1 },     30.29032,   0.00018,    76.932 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   1,   1,  -1 },     30.02094,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,  -2,   0,   0 },    240.32391,   0.00005,     0.042 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,  -2,   1,   0 },    240.32391,   0.00006,     0.096 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   0,  -1,   0 },    235.66831,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   0,   0,   0 },    227.60587,   0.00009,     0.909 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   0,   1,   0 },    235.66831,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   1,  -1,  -1 },     93.61838,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   1,   0,  -1 },     95.38929,   0.00016,    21.301 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   1,   1,  -1 },     93.61838,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -2,  -1,   0 },    248.30068,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -2,   0,   0 },    247.04035,   0.00010,     0.042 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -2,   1,   0 },    257.94737,   0.00010,     0.097 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   0,  -1,   0 },    229.80346,   0.00010,     0.083 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   0,   0,   0 },    229.89870,   0.00022,     0.856 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   0,   1,   0 },    232.38744,   0.00003,     0.069 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,  -1,  -1 },     75.79516,   0.00001,     0.076 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,  -1 },     81.42625,   0.00009,     8.708 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,   1,  -1 },     75.79516,   0.00001,     0.075 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   2,  -3,   0 },    213.30060,   0.00001,     0.072 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   2,  -1,   0 },     42.22462,   0.00002,     0.079 },
 { {   0,   0,   1,   0,   2,   5,   0,   0,   2,  -2,   0 },    110.16034,   0.00001,     2.295 },
 { {   0,   0,   1,   0,   2,   5,   0,   0,   2,   0,   0 },    110.16034,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   3, -10,   0,   0,   0,   0,   0 },    200.93049,   0.00002,     1.095 },
 { {   0,   0,   1,   0,   3,  -7,   0,   0,  -2,   0,   0 },     12.46788,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   3,  -7,   0,   0,   0,   0,   0 },     10.65371,   0.00004,     0.985 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,  -2,   0,   0 },    328.93024,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,  -2,   1,   0 },    329.76928,   0.00003,     0.096 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   0,  -1,   0 },    324.33559,   0.00001,     0.082 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   0,   0,   0 },    315.40898,   0.00005,     0.923 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   0,   1,   0 },    324.33559,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,  -1,  -1 },    170.73723,   0.00003,     0.076 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,   0,  -1 },    172.01480,   0.00019,    34.006 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,   1,  -1 },    170.73723,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   3,  -4,   0,   0,   0,   0,   0 },    355.87705,   0.00002,     0.895 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,  -2,   0,   0 },    258.14034,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,  -2,   1,   0 },    258.14034,   0.00004,     0.095 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,   0,  -1,   0 },    258.14034,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,   0,   0,   0 },    256.50076,   0.00013,     1.002 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,   0,   1,   0 },    258.14034,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   4,  -1,   0,   0,   2,  -3,   0 },    306.94900,   0.00001,     0.073 },
 { {   0,   0,   1,   0,   4,  -1,   0,   0,   2,  -1,   0 },    306.94900,   0.00001,     0.078 },
 { {   0,   0,   1,   0,   5,  -5,   0,   0,   0,  -1,   0 },    139.77728,   0.00002,     0.083 },
 { {   0,   0,   1,   0,   5,  -5,   0,   0,   0,   1,   0 },    139.77728,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   5,  -5,   0,   0,   2,  -2,   0 },    138.95793,   0.00001,     1.915 },
 { {   0,   0,   1,   0,   5,  -5,   0,   0,   2,   0,   0 },    138.95793,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   5,  -3,   0,   0,   2,  -3,   0 },     68.55096,   0.00001,     0.073 },
 { {   0,   0,   1,   0,   5,  -3,   0,   0,   2,  -1,   0 },     68.55096,   0.00001,     0.078 },
 { {   0,   0,   1,   0,   5,  -2,   0,   0,   2,  -2,   0 },    127.14117,   0.00002,     2.379 },
 { {   0,   0,   1,   0,   5,  -2,   0,   0,   2,   0,   0 },    127.14117,   0.00002,     0.038 },
 { {   0,   0,   1,   0,   6,  -5,   0,   0,   2,  -2,   0 },    310.39361,   0.00001,     2.284 },
 { {   0,   0,   1,   0,   6,  -5,   0,   0,   2,   0,   0 },    310.39361,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   6,  -2,   0,   0,   2,  -3,   0 },    117.75750,   0.00001,     0.074 },
 { {   0,   0,   1,   0,   6,  -2,   0,   0,   2,  -1,   0 },    117.75750,   0.00001,     0.077 },
 { {   0,   0,   1,   0,  10,  -2,   0,   0,   2,  -2,   0 },    259.85363,   0.00002,   884.771 },
 { {   0,   0,   1,   0,  14, -12,   0,   0,   2,  -2,   0 },    262.86194,   0.00002,   881.891 },
 { {   0,   0,   1,   1,  -2,   0,   0,   0,   2,  -3,   0 },    358.15527,   0.00001,     0.073 },
 { {   0,   0,   1,   1,  -2,   0,   0,   0,   2,  -1,   0 },    358.15527,   0.00001,     0.078 },
 { {   0,   0,   1,   1,  -1,   0,   0,   0,   2,  -1,   0 },     26.16957,   0.00002,     0.077 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,  -1,   0 },    274.12870,   0.00002,     0.027 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0 },    271.85660,   0.00014,     0.043 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,   1,   0 },    278.92924,   0.00012,     0.101 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -1,   0,   1 },    156.99256,   0.00001,     0.387 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0 },    252.58294,   0.00015,     0.085 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0 },    246.60047,   0.00053,     0.653 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   1,   0 },    256.44133,   0.00009,     0.068 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   1,   0,  -1 },    335.07416,   0.00005,     2.092 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -2,   0 },     58.06653,   0.00002,     4.127 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -1,   0 },     61.76801,   0.00028,     0.077 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,   0,   0 },     59.59974,   0.00004,     0.038 },
 { {   0,   0,   1,   1,   1,   0,   0,   0,   2,  -1,   0 },    274.21665,   0.00002,     0.076 },
 { {   0,   0,   1,   1,   2,   0,   0,   0,   2,  -2,   0 },    302.50854,   0.00001,    13.564 },
 { {   0,   0,   1,   1,   3,   0,   0,   0,   2,  -2,   0 },    284.64977,   0.00001,    94.519 },
 { {   0,   0,   1,   2,  -3,   0,   0,   0,   2,  -2,   0 },     59.88152,   0.00002,    27.444 },
 { {   0,   0,   1,   2,  -3,   0,   0,   0,   2,  -1,   0 },    244.52168,   0.00004,     0.075 },
 { {   0,   0,   1,   2,  -2,   0,   0,   0,   0,   0,   0 },    209.28680,   0.00001,     0.528 },
 { {   0,   0,   1,   2,  -1,   0,   0,   0,   2,  -1,   0 },    240.79457,   0.00002,     0.074 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,  -1,   0 },     41.41904,   0.00001,     0.028 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   0,   0 },     38.00699,   0.00015,     0.044 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   1,   0 },     42.87670,   0.00010,     0.106 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   0,  -1 },     60.78506,   0.00005,     0.042 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   1,  -1 },    240.85840,   0.00004,     0.095 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   0,  -1,   0 },     51.49893,   0.00003,     0.089 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   0,   1,   0 },     68.96779,   0.00001,     0.065 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,  -1,  -1 },     60.81142,   0.00007,     0.082 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   0,  -1 },     34.02751,   0.00002,     0.991 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   1,  -1 },    240.81142,   0.00005,     0.070 },
 { {   0,   0,   1,   2,   1,   0,   0,   0,   2,   0,  -2 },    281.76775,   0.00001,    24.912 },
 { {   0,   0,   1,   3,  -2,   0,   0,   0,   2,  -1,   0 },    249.68674,   0.00002,     0.072 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -3,   0 },    310.74834,   0.00001,     0.080 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -1,   0 },    310.74834,   0.00001,     0.071 },
 { {   0,   0,   1,   4,  -2,   0,   0,   0,   2,  -1,   0 },    278.81867,   0.00002,     0.069 },
 { {   0,   0,   1,   4,  -2,   0,   0,   0,   2,   0,   0 },     99.01894,   0.00002,     0.036 },
 { {   0,   0,   1,   4,  -1,   0,   0,   0,   2,   0,   0 },    125.41708,   0.00001,     0.036 },
 { {   0,   0,   1,   5,  -2,   0,   0,   0,   2,  -1,   0 },    307.83083,   0.00002,     0.067 },
 { {   0,   0,   1,   5,  -2,   0,   0,   0,   2,   0,   0 },    308.07822,   0.00001,     0.035 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   1,   0 },    173.57239,   0.00001,     0.073 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   3,   0 },    173.57239,   0.00001,     0.078 },
 { {   0,   0,   2,  -8,   2,   0,   0,   0,  -2,   0,   0 },     51.63315,   0.00002,     0.037 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,  -2,   1,   0 },    150.66331,   0.00002,     0.074 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,  -2,   3,   0 },    149.44349,   0.00002,     0.077 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   0,   0,   0 },    132.70290,   0.00024,     0.500 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   0,   1,   0 },    132.50010,   0.00009,     0.089 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,  -2,   0 },    132.69629,   0.00002,     0.265 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,  -1,   0 },    132.85383,   0.00047,     0.105 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   0,   0 },    132.63605,   0.00056,     0.044 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   1,   0 },    132.69105,   0.00006,     0.028 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   4,  -1,   0 },    132.68818,   0.00004,     0.029 },
 { {   0,   0,   2,  -8,   5,   0,   0,   0,  -2,   1,   0 },    288.00126,   0.00003,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   1,   0 },    107.52921,   0.00001,     0.076 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   2,   0 },    202.41912,   0.00003,    19.147 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   3,   0 },    107.52921,   0.00001,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   0,   0,   0 },    222.07617,   0.00002,     0.581 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   1,   1,  -1 },    246.78022,   0.00002,     0.095 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   1,   2,  -1 },    247.37610,   0.00001,     0.042 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   1,   0 },     55.73276,   0.00003,     0.079 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   3,   0 },     63.48169,   0.00001,     0.072 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -1,   0,   1 },    310.26949,   0.00002,     7.335 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,  -1,   0 },    334.03154,   0.00001,     0.069 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0,   0 },    333.33255,   0.00016,     0.840 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,   1,   0 },      7.54164,   0.00008,     0.083 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,  -1,   0 },     16.71917,   0.00015,     0.097 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,   0,   0 },     16.21113,   0.00017,     0.042 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,   1,   0 },     16.18443,   0.00001,     0.027 },
 { {   0,   0,   2,  -6,   0,   4,   0,   0,  -1,   0,   1 },    310.17119,   0.00004,  1827.649 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   0,  -1,   0 },     67.95139,   0.00001,     0.070 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   0,   0,   0 },     73.47596,   0.00007,     0.979 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   0,   1,   0 },     67.95139,   0.00001,     0.082 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   2,  -1,   0 },     74.06337,   0.00005,     0.096 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   2,   0,   0 },     73.92457,   0.00005,     0.042 },
 { {   0,   0,   2,  -6,   3,   0,   0,   0,   0,   0,   0 },    107.44873,   0.00006,     1.067 },
 { {   0,   0,   2,  -6,   3,   0,   0,   0,   2,   0,   0 },    105.85378,   0.00002,     0.042 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   0,   0 },    245.90524,   0.00003,     0.039 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   1,   0 },    116.31294,   0.00002,     0.082 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   0,   1 },    351.49736,   0.00007,     2.529 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   1,  -1 },    171.96109,   0.00008,     0.076 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   2,  -1 },    171.76867,   0.00002,    16.174 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0 },    269.80967,   0.00018,     0.072 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0 },    279.16298,   0.00109,     1.519 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0 },    280.48589,   0.00019,     0.079 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   1,   0,  -1 },    353.27429,   0.00003,     0.584 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,  -1,   0 },      8.86017,   0.00017,     0.092 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,   0,   0 },      5.15994,   0.00015,     0.042 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,   1,   0 },      2.08273,   0.00001,     0.027 },
 { {   0,   0,   2,  -5,   3,   0,   0,   0,   0,   0,   0 },    137.74216,   0.00002,     2.466 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   0,   0 },    299.59485,   0.00008,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   1,   0 },    302.02284,   0.00020,     0.028 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   2,   0 },    302.99130,   0.00012,     0.043 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -3,   0,   0 },    300.69963,   0.00002,     0.027 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -2,   0 },    297.81188,   0.00009,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    299.89252,   0.00071,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,   2 },    121.21714,   0.00001,     0.089 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,  -2 },    122.02516,   0.00001,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   0 },    300.82014,   0.00691,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   2 },    300.95259,   0.00017,     0.505 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,  -2 },    121.86509,   0.00002,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,   0 },    302.55753,   0.01024,     0.086 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   2,   0 },    306.85022,   0.00032,     0.607 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   3,   0 },    309.21538,   0.00002,     0.067 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   0 },    120.13350,   0.00016,     0.080 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   1 },     23.36834,   0.00006,     1.079 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   1,   0 },    123.11933,   0.00002,     1.315 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -3,   0 },    297.64178,   0.00008,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0 },    297.93859,   0.00116,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -2 },    118.83348,   0.00004,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0 },    297.82534,   0.01712,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2 },    297.74734,   0.00003,     0.074 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -2 },    123.13915,   0.00004,     0.037 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0 },    300.13913,   0.09178,     7.890 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1 },     23.32594,   0.00002,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   2 },    123.13915,   0.00004,     0.037 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -2 },    297.74734,   0.00003,     0.073 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0 },    298.04083,   0.01770,     0.076 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   2 },    118.83348,   0.00004,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0 },    297.87865,   0.00118,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   3,   0 },    297.59736,   0.00008,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,  -1,   0 },    123.74970,   0.00003,     0.986 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,  -1 },     24.12689,   0.00007,     0.847 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,   0 },    112.47582,   0.00031,     0.082 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   1,   0 },    121.80711,   0.00002,     0.039 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -3,   0 },    308.37580,   0.00002,     0.066 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -2,   0 },    305.38874,   0.00045,     0.526 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   0 },    303.41117,   0.01284,     0.088 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   2 },    122.56962,   0.00003,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,  -2 },    302.51295,   0.00019,     0.448 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   0 },    302.71544,   0.00874,     0.041 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   2 },    122.65793,   0.00002,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,  -2 },    122.77107,   0.00001,     0.091 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,   0 },    301.63427,   0.00086,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   2,   0 },    301.40410,   0.00009,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   3,  -1,   0 },     95.30828,   0.00002,     0.042 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   3,   0,   0 },    303.30441,   0.00002,     0.027 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -2,   0 },    303.45346,   0.00015,     0.044 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -1,   0 },    302.80630,   0.00026,     0.028 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,   0,   0 },    302.34644,   0.00011,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,   1,   0 },    302.68559,   0.00002,     0.016 },
 { {   0,   0,   2,  -4,   0,   1,   0,   0,   0,   0,   0 },    199.20656,   0.00001,    10.777 },
 { {   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0,   0 },      0.62821,   0.00006,    16.995 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,  -2,   0,   0 },    155.28116,   0.00007,     0.040 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,  -2,   1,   0 },    155.69021,   0.00011,     0.087 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   0,  -1,   0 },    153.34101,   0.00012,     0.075 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   0,   0,   0 },    154.61292,   0.00077,    23.567 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   0,   1,   0 },    153.38690,   0.00012,     0.076 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   2,  -1,   0 },    155.99854,   0.00012,     0.087 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   2,   0,   0 },    155.47036,   0.00007,     0.040 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,  -1,   0 },      0.61461,   0.00003,     0.026 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   0,   0 },      0.55256,   0.00019,     0.040 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   1,   0 },      0.85198,   0.00031,     0.087 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   2,   0 },      0.61462,   0.00001,     0.551 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,  -2,   0 },    359.01236,   0.00002,     0.038 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,  -1,   0 },    359.24257,   0.00034,     0.076 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0,   0 },      0.03516,   0.00176,    23.883 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   1,   0 },    359.18071,   0.00033,     0.075 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   2,   0 },    359.01937,   0.00002,     0.038 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,  -2,   0 },      0.50893,   0.00001,     0.577 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,  -1,   0 },      0.72562,   0.00027,     0.087 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,   0,   0 },      0.50830,   0.00019,     0.040 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,   1,   0 },      0.53883,   0.00002,     0.026 },
 { {   0,   0,   2,  -4,   3,   0,   0,   0,   0,   0,   0 },    176.23417,   0.00005,     7.926 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,  -1,   0 },    330.91174,   0.00001,     0.016 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   0,   0 },    330.89879,   0.00008,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   1,   0 },    330.81922,   0.00022,     0.028 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   2,   0 },    330.92324,   0.00013,     0.044 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -3,   0,   0 },    331.11863,   0.00002,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -2,   0 },    331.70256,   0.00006,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -1,   0 },    330.68115,   0.00068,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,  -2 },    150.76639,   0.00002,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   0 },    330.97751,   0.00739,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   2 },    331.00142,   0.00014,     0.398 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,  -2 },    150.79928,   0.00002,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,   0 },    330.92395,   0.01085,     0.090 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   2,   0 },    331.07667,   0.00047,     0.459 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   3,   0 },    331.37341,   0.00003,     0.065 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   0 },    150.67536,   0.00025,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   1 },     52.85066,   0.00006,     0.686 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   1,   0 },    150.76250,   0.00004,     0.774 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   0 },    329.73544,   0.00004,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0 },    329.81147,   0.00057,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -2 },    150.04332,   0.00003,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0 },    329.85153,   0.00849,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   2 },    329.79971,   0.00001,     0.071 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2 },    150.74641,   0.00002,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0 },    330.28695,   0.04014,     2.470 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2 },    150.74641,   0.00002,     0.037 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2 },    154.79834,   0.00003,     0.076 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0 },    329.78144,   0.00753,     0.073 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   2 },    150.04332,   0.00003,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2 },    153.55605,   0.00001,    13.999 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0 },    329.74212,   0.00052,     0.037 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0 },    329.65605,   0.00003,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,  -1,  -1 },     53.53236,   0.00002,     0.072 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,  -1,   1 },    233.38522,   0.00007,     0.077 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,  -1 },     53.41627,   0.00022,     1.541 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,   0 },    331.10928,   0.00006,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,   1 },    233.39520,   0.00003,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   1,  -1 },     53.53236,   0.00002,     0.079 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -2,   0 },    331.04394,   0.00022,     0.730 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,  -2 },    150.52247,   0.00001,     0.067 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    331.23176,   0.00653,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,  -2 },    330.50098,   0.00010,     0.587 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,   0 },    330.41642,   0.00368,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   1,  -2 },    150.52247,   0.00001,     0.087 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   1,   0 },    330.45866,   0.00038,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   2,   0 },    330.43897,   0.00004,     0.019 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,  -2,   0 },    331.08323,   0.00007,     0.043 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,  -1,   0 },    330.73217,   0.00012,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,   0,   0 },    330.83801,   0.00005,     0.020 },
 { {   0,   0,   2,  -3,   1,   0,   0,   0,   0,   0,   0 },    353.36656,   0.00002,     2.044 },
 { {   0,   0,   2,  -3,   1,   0,   0,   0,   2,   0,   0 },    183.00956,   0.00001,     0.040 },
 { {   0,   0,   2,  -3,   2,   0,   0,   0,   2,  -3,   0 },    209.84056,   0.00001,     0.069 },
 { {   0,   0,   2,  -3,   2,   0,   0,   0,   2,  -1,   0 },    209.84056,   0.00001,     0.083 },
 { {   0,   0,   2,  -3,   2,   0,   0,   0,   2,   0,   0 },     29.77752,   0.00001,     0.040 },
 { {   0,   0,   2,  -3,   3,   0,   0,   0,   0,   0,   0 },    351.96947,   0.00001,     1.520 },
 { {   0,   0,   2,  -2,  -3,   0,   0,   0,   0,   0,   0 },    320.72418,   0.00001,     1.463 },
 { {   0,   0,   2,  -2,  -2,   0,   0,   0,  -2,   0,   0 },    120.93760,   0.00001,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -6,   0,   0 },    177.24420,   0.00001,     0.014 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -6,   1,   0 },    359.04198,   0.00002,     0.017 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -6,   2,   0 },    358.93111,   0.00004,     0.021 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,  -1,   0 },      0.97546,   0.00009,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,   0 },    357.17070,   0.00051,     0.021 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,   2 },    358.67143,   0.00001,     0.046 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   1,   0 },    357.61898,   0.00119,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   1,   2 },    358.69452,   0.00002,     0.119 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   2,   0 },    358.45138,   0.00077,     0.045 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   3,   0 },    359.18001,   0.00003,     0.114 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -3,  -1,   0 },    358.26192,   0.00001,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -3,   0,   0 },    358.92979,   0.00004,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -3,   1,   0 },    176.34655,   0.00002,     0.044 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -3,   0 },    358.81215,   0.00004,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -2,   0 },    359.08935,   0.00033,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,  -2 },    178.91350,   0.00002,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,   0 },    358.13118,   0.00404,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,   2 },    178.72998,   0.00007,     0.098 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,  -2 },    179.09959,   0.00009,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   0 },    358.57975,   0.04612,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   2 },    358.76710,   0.00064,     0.329 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,  -2 },    178.89931,   0.00017,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   0 },    358.86873,   0.05904,     0.095 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   2 },    359.28638,   0.00002,     0.061 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   2,   0 },    359.11621,   0.00254,     0.369 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   3,   0 },    358.22612,   0.00013,     0.063 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,  -1,   0 },      2.19520,   0.00004,     0.041 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,  -1,   1 },    279.51216,   0.00003,     0.089 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,  -1 },    100.08715,   0.00001,     0.040 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,   0 },      0.46997,   0.00119,     0.087 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,   1 },     98.75784,   0.00006,     0.502 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   1,   0 },      1.62333,   0.00007,     0.548 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0 },    359.84872,   0.00021,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0 },    359.28474,   0.00274,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -2 },    179.55695,   0.00011,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0 },    359.51637,   0.04213,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2 },    178.52060,   0.00002,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0 },    359.76487,   0.19443,     1.068 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2 },    179.69211,   0.00014,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0 },      0.07117,   0.03259,     0.070 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   2 },    179.66457,   0.00009,     0.024 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0 },      0.40226,   0.00229,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2 },    179.66760,   0.00001,     0.018 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0 },    359.97411,   0.00015,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -3,   0 },    180.43621,   0.00002,     0.038 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -2,   0 },    180.50246,   0.00017,     0.076 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,  -1 },     99.34031,   0.00014,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   0 },      0.27548,   0.00139,    20.152 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   1 },     99.53251,   0.00013,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -2 },    180.23488,   0.00002,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -1 },     99.52623,   0.00113,     8.537 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   0 },    180.27603,   0.01201,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   1 },    279.44407,   0.00028,     0.037 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   2 },      0.23486,   0.00002,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,  -1 },     99.34031,   0.00014,     0.076 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,   0 },    180.23089,   0.00083,     0.038 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,   1 },    279.40921,   0.00003,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   2,   0 },    180.23489,   0.00006,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -3,   0 },    183.61355,   0.00001,     0.071 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -2,   0 },    180.68775,   0.00035,     1.194 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,   0 },      0.45156,   0.01315,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,   2 },    179.96311,   0.00003,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -2 },      0.84425,   0.00049,     0.854 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   0 },      0.89077,   0.00875,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   2 },    179.52810,   0.00002,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   1,  -2 },    180.79372,   0.00002,     0.083 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   1,   0 },      1.04119,   0.00101,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   2,   0 },    359.61722,   0.00012,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -2,   0 },      0.40036,   0.00006,     0.087 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -1,   0 },    180.22454,   0.00014,     0.040 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,   0,   0 },    180.01815,   0.00009,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -2,   0 },    359.98857,   0.00016,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -1,   0 },      1.39166,   0.00027,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,   0,   0 },      0.78600,   0.00009,     0.020 },
 { {   0,   0,   2,  -2,   0,   2,   0,   0,  -2,   0,   0 },    290.92236,   0.00002,     0.042 },
 { {   0,   0,   2,  -2,   0,   2,   0,   0,  -2,   1,   0 },    290.92236,   0.00002,     0.095 },
 { {   0,   0,   2,  -2,   0,   2,   0,   0,   0,   0,   0 },    290.94554,   0.00003,     0.996 },
 { {   0,   0,   2,  -2,   1,   0,   0,   0,   0,   0,   0 },     97.77295,   0.00001,     0.980 },
 { {   0,   0,   2,  -2,   1,   0,   0,   0,   1,   0,  -1 },    120.93336,   0.00001,    30.461 },
 { {   0,   0,   2,  -2,   1,   0,   0,   0,   2,  -3,   0 },    204.71958,   0.00001,     0.071 },
 { {   0,   0,   2,  -2,   1,   0,   0,   0,   2,  -1,   0 },    204.71958,   0.00001,     0.080 },
 { {   0,   0,   2,  -2,   2,   0,   0,   0,   1,   0,  -1 },    165.35486,   0.00001,    19.427 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -4,   1,   0 },    255.72463,   0.00002,     0.029 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,  -1,   0 },    255.53513,   0.00005,     0.027 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,   0 },    255.00621,   0.00050,     0.043 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   1,   0 },    257.41131,   0.00053,     0.100 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   2,   0 },    260.26274,   0.00003,     0.308 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0 },    234.62398,   0.00039,     0.085 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0 },    228.74962,   0.00140,     0.681 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0 },    232.83971,   0.00017,     0.068 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,  -1,  -1 },    308.94239,   0.00002,     0.078 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,  -1,   1 },    129.69201,   0.00002,     0.071 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,  -1 },    309.05654,   0.00017,     2.412 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,   0 },    210.13122,   0.00002,     0.072 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,   1 },    128.95929,   0.00002,     0.037 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   1,  -1 },    308.94239,   0.00002,     0.073 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -2,   0 },     44.31594,   0.00007,     3.271 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,   0 },     39.71529,   0.00055,     0.077 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,   0,   0 },     43.77641,   0.00009,     0.038 },
 { {   0,   0,   2,  -1,   2,   0,   0,   0,   2,  -1,   0 },    281.91898,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -7,   0,   0,   0,   0,   0,   0 },    233.41129,   0.00001,     0.709 },
 { {   0,   0,   2,   0,  -7,   3,   0,   0,   0,   0,   0 },    197.99317,   0.00001,     0.662 },
 { {   0,   0,   2,   0,  -7,   5,   0,   0,  -2,   0,   0 },    102.11525,   0.00002,     0.043 },
 { {   0,   0,   2,   0,  -7,   5,   0,   0,   0,   0,   0 },    102.48371,   0.00003,     0.633 },
 { {   0,   0,   2,   0,  -7,  10,   0,   0,   2,  -3,   0 },    247.91736,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -7,  10,   0,   0,   2,  -2,   0 },     77.16868,   0.00003,    40.542 },
 { {   0,   0,   2,   0,  -7,  10,   0,   0,   2,  -1,   0 },     67.91736,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,  -2,   0,   0 },    218.52011,   0.00004,     0.043 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,  -2,   1,   0 },    218.75520,   0.00004,     0.100 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,   0,  -1,   0 },    219.60412,   0.00001,     0.085 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,   0,   0,   0 },    220.52707,   0.00010,     0.669 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,   0,   1,   0 },    219.60412,   0.00001,     0.068 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,   2,  -1,   0 },     41.06397,   0.00005,     0.077 },
 { {   0,   0,   2,   0,  -6,   2,   0,   0,  -2,   0,   0 },    199.25922,   0.00002,     0.043 },
 { {   0,   0,   2,   0,  -6,   2,   0,   0,   0,   0,   0 },    200.88252,   0.00003,     0.640 },
 { {   0,   0,   2,   0,  -6,   3,   0,   0,  -2,   0,   0 },    217.47330,   0.00001,     0.043 },
 { {   0,   0,   2,   0,  -6,   3,   0,   0,   0,   0,   0 },    219.21827,   0.00003,     0.627 },
 { {   0,   0,   2,   0,  -6,   3,   0,   0,   2,  -1,   0 },    210.60076,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -6,   4,   0,   0,  -2,   0,   0 },    129.70901,   0.00002,     0.043 },
 { {   0,   0,   2,   0,  -6,   4,   0,   0,   0,   0,   0 },    129.85882,   0.00002,     0.614 },
 { {   0,   0,   2,   0,  -6,   4,   0,   0,   2,  -1,   0 },    131.00069,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,  -2,   0,   0 },     80.01938,   0.00005,     0.043 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,  -2,   1,   0 },     80.01942,   0.00004,     0.102 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   0,  -1,   0 },     79.27518,   0.00005,     0.086 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   0,   0,   0 },     80.88266,   0.00010,     0.601 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   0,   1,   0 },     80.01998,   0.00001,     0.067 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   2,  -3,   0 },    139.15169,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   2,  -2,   0 },    255.40186,   0.00003,     9.083 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   2,  -1,   0 },    259.07576,   0.00019,     0.076 },
 { {   0,   0,   2,   0,  -6,   7,   0,   0,   0,   0,   0 },    353.83378,   0.00001,     0.577 },
 { {   0,   0,   2,   0,  -6,   8,   0,   0,   2,  -2,   0 },    231.83393,   0.00003,   121.093 },
 { {   0,   0,   2,   0,  -6,  10,   0,   0,   0,   0,   0 },    161.67157,   0.00001,     0.545 },
 { {   0,   0,   2,   0,  -6,  10,   0,   0,   2,  -3,   0 },    335.23345,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -6,  10,   0,   0,   2,  -1,   0 },    155.23345,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,  -1,   0 },    209.09753,   0.00003,     0.027 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   0,   0 },    209.17203,   0.00028,     0.043 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   1,   0 },    209.10470,   0.00032,     0.101 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,  -2,   0 },    209.96875,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,  -1,   0 },    209.89266,   0.00039,     0.086 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 },    210.15374,   0.00084,     0.634 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,   1,   0 },    210.75493,   0.00014,     0.067 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   1,   0,  -1 },    298.55109,   0.00001,     1.906 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -2,   0 },     29.34102,   0.00012,     5.114 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -1,   0 },     29.50803,   0.00078,     0.077 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,   0,   0 },     29.83593,   0.00008,     0.038 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,  -2,   0,   0 },    190.25918,   0.00007,     0.043 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,  -2,   1,   0 },    190.25918,   0.00006,     0.102 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   0,  -1,   0 },    191.18147,   0.00008,     0.086 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   0,   0,   0 },    190.84789,   0.00016,     0.607 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   0,   1,   0 },    190.65957,   0.00003,     0.067 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   2,  -3,   0 },    201.88455,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   2,  -2,   0 },     10.34054,   0.00003,     7.834 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   2,  -1,   0 },     11.48589,   0.00024,     0.076 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,  -2,   0,   0 },    214.00698,   0.00004,     0.043 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,  -2,   1,   0 },    214.00698,   0.00004,     0.102 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   0,  -2,   0 },     33.61296,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   0,  -1,   0 },    212.48717,   0.00004,     0.086 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   0,   0,   0 },    212.69708,   0.00008,     0.595 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   2,  -3,   0 },    102.51300,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   2,  -2,   0 },     34.69405,   0.00003,    10.673 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   2,  -1,   0 },     32.10257,   0.00016,     0.076 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,  -2,   0,   0 },     64.01451,   0.00002,     0.043 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   0,  -1,   0 },     58.28773,   0.00003,     0.087 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   0,   0,   0 },     62.33730,   0.00005,     0.583 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   2,  -3,   0 },     73.73707,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   2,  -2,   0 },    239.28470,   0.00005,    16.738 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   2,  -1,   0 },    238.28747,   0.00016,     0.076 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,  -2,  -1,   0 },     56.73236,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,  -2,   0,   0 },     56.73999,   0.00014,     0.043 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,  -2,   1,   0 },     56.74690,   0.00015,     0.103 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,  -2,   0 },     53.80032,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,  -1,   0 },     54.49299,   0.00087,     0.087 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,   0,   0 },     55.92918,   0.00048,     0.572 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,   1,   0 },     56.48239,   0.00006,     0.067 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -3,   0 },    247.87335,   0.00004,     0.075 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -2,   0 },    233.93304,   0.00271,    38.763 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -1,   0 },    234.22455,   0.00410,     0.076 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,   0,   0 },    231.36839,   0.00028,     0.038 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,   1,   0 },    234.21721,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -3,   0 },    232.03012,   0.00005,     0.087 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -2,   0 },    234.13389,   0.00008,     0.040 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -1,   0 },    234.20316,   0.00004,     0.026 },
 { {   0,   0,   2,   0,  -5,   6,   0,   0,   2,  -3,   0 },    168.60619,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -5,   6,   0,   0,   2,  -2,   0 },    178.18081,   0.00007,   122.705 },
 { {   0,   0,   2,   0,  -5,   6,   0,   0,   2,  -1,   0 },    348.60619,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -4,   0,   0 },    199.89215,   0.00003,     0.021 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -4,   1,   0 },    199.90065,   0.00006,     0.029 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -4,   2,   0 },    199.90246,   0.00004,     0.047 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,  -2,   0 },    199.12213,   0.00001,     0.020 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    199.85974,   0.00023,     0.028 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   0 },    199.37704,   0.00248,     0.043 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   2 },    199.64855,   0.00003,     0.265 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   1,   0 },    199.81212,   0.00258,     0.102 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   2,   0 },    199.55181,   0.00011,     0.291 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -1,   0,   0 },     20.20545,   0.00003,     0.093 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -1,   0,   1 },    295.24783,   0.00001,     0.368 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -3,   0 },    199.86928,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -2,   0 },    200.26118,   0.00025,     0.040 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -1,   0 },    200.25056,   0.00425,     0.086 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 },    200.47414,   0.00672,     0.601 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   1,   0 },    201.01344,   0.00112,     0.067 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   2,   0 },    201.04392,   0.00007,     0.035 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   1,   0,  -1 },    289.34098,   0.00008,     1.642 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   1,   0,   0 },     21.43224,   0.00004,     0.071 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -3,   0 },     19.92361,   0.00010,     0.075 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -2,   0 },     19.83495,   0.00273,     8.990 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,  -2 },     19.94000,   0.00002,     0.073 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   0 },     19.76044,   0.01312,     0.076 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   2 },    199.94000,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   0,  -2 },     20.01370,   0.00001,     2.248 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   0,   0 },     21.57931,   0.00105,     0.038 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   1,   0 },     20.04940,   0.00008,     0.025 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -3,   0 },     19.63919,   0.00004,     0.088 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -2,   0 },     19.84786,   0.00020,     0.041 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -1,   0 },     20.01036,   0.00013,     0.026 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,  -2,   0,   0 },     15.30410,   0.00002,     0.043 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,   0,   0,   0 },     15.65554,   0.00004,     0.589 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,   2,  -3,   0 },     27.32675,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,   2,  -2,   0 },    192.35467,   0.00002,    12.939 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,   2,  -1,   0 },    189.72303,   0.00006,     0.076 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,  -1,   0 },    182.70411,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   0,   0 },    182.71354,   0.00018,     0.043 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   1,   0 },    182.72216,   0.00019,     0.103 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -2,   0 },    182.19300,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -1,   0 },    182.41605,   0.00072,     0.087 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   0,   0 },    182.28698,   0.00057,     0.578 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   1,   0 },    182.46177,   0.00008,     0.067 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -2,   0 },      2.27787,   0.00129,    23.075 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -1,   0 },      2.38272,   0.00304,     0.076 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,   0,   0 },      2.25718,   0.00022,     0.038 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   4,  -3,   0 },    359.10763,   0.00003,     0.087 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   4,  -2,   0 },      2.24429,   0.00005,     0.040 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   4,  -1,   0 },      2.36893,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,  -2,   0,   0 },    239.16234,   0.00007,     0.044 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,  -2,   1,   0 },    239.16234,   0.00007,     0.103 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -2,   0 },     57.23415,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -1,   0 },    237.18925,   0.00085,     0.087 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,   0,   0 },    237.95305,   0.00026,     0.567 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,   1,   0 },    238.75155,   0.00003,     0.067 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -4,   0 },     52.64018,   0.00002,     0.038 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -3,   0 },     52.61078,   0.00023,     0.075 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -2,   0 },     57.32317,   0.00819,   106.494 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -1,   0 },     57.16351,   0.00495,     0.075 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,   0,   0 },     57.21868,   0.00034,     0.038 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,   1,   0 },     57.14692,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -3,   0 },     58.91499,   0.00014,     0.087 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -2,   0 },     57.34952,   0.00016,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -1,   0 },     57.20429,   0.00005,     0.026 },
 { {   0,   0,   2,   0,  -4,   4,  -2,   0,   2,  -2,   0 },    174.85996,   0.00040,  1326.783 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,  -2,   0,   0 },    117.58238,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   0,  -1,   0 },    296.93826,   0.00004,     0.087 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   0,   0,   0 },    117.28863,   0.00003,     0.556 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   2,  -2,   0 },    295.36360,   0.00012,    40.721 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   2,  -1,   0 },    116.93835,   0.00020,     0.075 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -4,   1,   0 },    351.22838,   0.00002,     0.029 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,  -1,   0 },    351.22739,   0.00005,     0.028 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   0,   0 },    351.23654,   0.00047,     0.044 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   1,   0 },    351.23550,   0.00050,     0.104 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   2,   0 },    351.23018,   0.00001,     0.277 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -2,   0 },    170.98422,   0.00009,     0.041 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -1,   0 },    171.19781,   0.00103,     0.088 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,   0,   0 },    351.46713,   0.00105,     0.546 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,   1,   0 },    351.29284,   0.00017,     0.066 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,   2,   0 },    349.14985,   0.00003,     0.035 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -3,   0 },    148.63306,   0.00005,     0.076 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -2,   0 },    170.86324,   0.00133,    17.093 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,   0 },    351.23859,   0.00632,     0.075 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   0,  -2 },    167.59622,   0.00001,     3.636 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   0,   0 },    351.49083,   0.00035,     0.038 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   1,   0 },    351.21474,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   2,   0 },    168.24317,   0.00001,     0.019 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   4,  -3,   0 },    183.23586,   0.00002,     0.087 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   4,  -2,   0 },    350.48814,   0.00007,     0.040 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   4,  -1,   0 },    351.19689,   0.00006,     0.026 },
 { {   0,   0,   2,   0,  -4,   6,  -3,   0,   0,   0,   0 },    168.76255,   0.00001,     0.546 },
 { {   0,   0,   2,   0,  -4,   6,  -3,   0,   2,  -1,   0 },    173.23991,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -3,  -5,   0,   0,   0,   0,   0 },    219.39782,   0.00001,     0.634 },
 { {   0,   0,   2,   0,  -3,  -1,   0,   0,   0,   0,   0 },    154.75835,   0.00001,     0.584 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -6,   1,   0 },    190.06698,   0.00001,     0.017 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,  -1,   0 },    189.99846,   0.00004,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   0,   0 },    189.12578,   0.00019,     0.021 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   0,   2 },    190.23547,   0.00002,     0.048 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   1,   0 },    190.62647,   0.00043,     0.029 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   2,   0 },    190.34099,   0.00023,     0.047 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -3,   0,   0 },    190.19286,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -3,   1,   0 },     10.20972,   0.00002,     0.045 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -3,   0 },    190.11682,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -2,   0 },    190.27943,   0.00020,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    190.29227,   0.00220,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,   2 },     10.25022,   0.00002,     0.106 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,  -2 },     10.40597,   0.00004,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   0 },    190.20751,   0.01841,     0.043 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   2 },    190.22067,   0.00016,     0.259 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,  -2 },     10.38090,   0.00004,     0.027 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   0 },    190.29786,   0.01814,     0.103 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   2,   0 },    190.40548,   0.00075,     0.284 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   3,   0 },    190.32909,   0.00005,     0.060 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,  -1,   0 },     10.08577,   0.00006,     0.042 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,  -1,   1 },    187.73280,   0.00002,     0.096 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,   0,   0 },     11.03822,   0.00032,     0.094 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,   0,   1 },    292.54951,   0.00004,     0.357 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,   1,   0 },     10.42894,   0.00004,     0.380 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -3,   0 },    190.43392,   0.00011,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,  -2 },     10.11312,   0.00002,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   0 },    190.26869,   0.00320,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   2 },     10.05397,   0.00004,     0.481 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,  -2 },     10.17332,   0.00019,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   0 },    190.17970,   0.09518,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   2 },    190.69419,   0.00001,     0.065 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,  -2 },      8.63024,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 },    190.43600,   0.05167,     0.572 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   2 },     10.39390,   0.00006,     0.035 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   0 },    190.51572,   0.00813,     0.067 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   2 },     10.61530,   0.00003,     0.024 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   2,   0 },    190.67758,   0.00058,     0.035 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   3,   0 },    190.76576,   0.00004,     0.024 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -2,   0 },    190.64834,   0.00004,     0.081 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -1,  -1 },    276.01323,   0.00003,     0.080 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -1,   0 },     10.20756,   0.00044,     1.163 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -1,   1 },     99.37819,   0.00003,     0.070 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   0,  -1 },    278.28609,   0.00031,     1.442 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   0,   0 },      8.01131,   0.00039,     0.071 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   0,   1 },    100.50400,   0.00002,     0.036 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   1,  -1 },    276.01323,   0.00003,     0.072 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   1,   0 },      8.09704,   0.00003,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -5,   0 },     10.15314,   0.00005,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -4,   0 },     10.15390,   0.00064,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,  -2 },    190.07777,   0.00004,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   0 },     10.15170,   0.00936,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,  -2 },    190.01079,   0.00024,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   0 },     10.03835,   0.28281,    37.133 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   2 },    190.04435,   0.00029,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,  -2 },     10.07701,   0.00070,     0.073 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   0 },     10.07001,   0.44532,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   2 },    190.08255,   0.00092,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,  -2 },     11.69679,   0.00014,     2.774 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   0 },     10.08580,   0.03174,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   2 },    190.08286,   0.00016,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,  -2 },    190.20619,   0.00005,     0.078 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,   0 },      9.99365,   0.00223,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,   2 },    190.08587,   0.00002,     0.015 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   2,   0 },     10.10132,   0.00017,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   3,  -2,   0 },    190.01305,   0.00013,     0.081 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   3,  -1,   0 },    190.08285,   0.00017,     0.039 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   3,   0,   0 },    190.08585,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -4,   0 },      9.87614,   0.00012,     0.555 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,   0 },      9.99181,   0.00440,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,   2 },    190.02539,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,  -2 },     10.05206,   0.00004,     0.469 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   0 },      9.96731,   0.00897,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   2 },    190.06895,   0.00005,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,  -2 },    190.04988,   0.00011,     0.090 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,   0 },     10.08150,   0.00447,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,   2 },    190.08535,   0.00002,     0.015 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   0,   0 },     10.15916,   0.00066,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   1,   0 },     10.37202,   0.00006,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -4,   0 },      9.99388,   0.00006,     0.044 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -3,   0 },     10.04952,   0.00017,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -2,   0 },     10.05748,   0.00015,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -1,   0 },     10.08511,   0.00005,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   1,   0,   2,  -3,   0 },      0.73083,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   1,   0,   2,  -2,   0 },    180.73749,   0.00002,    66.542 },
 { {   0,   0,   2,   0,  -3,   0,   1,   0,   2,  -1,   0 },    180.73083,   0.00001,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   2,   0,   2,  -2,   0 },    349.75714,   0.00030,   319.860 },
 { {   0,   0,   2,   0,  -3,   0,   2,   0,   2,  -1,   0 },    349.77796,   0.00007,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,  -2,   0,   0 },      3.60212,   0.00007,     0.044 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,  -2,   1,   0 },      3.60212,   0.00007,     0.103 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -3,   0 },    181.05821,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -2,   0 },    181.31415,   0.00020,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -1,   0 },    181.39771,   0.00121,     0.087 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,   0,   0 },      5.95027,   0.00005,     0.561 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,   1,   0 },      3.60212,   0.00001,     0.067 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -4,   0 },    179.81568,   0.00002,     0.038 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -3,   0 },    179.82209,   0.00037,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -2,   0 },    181.29764,   0.01126,   142.500 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -1,   0 },      1.45638,   0.00479,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,   0,   0 },      1.46226,   0.00031,     0.038 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,   1,   0 },      1.43829,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -3,   0 },    181.91874,   0.00018,     0.087 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -2,   0 },    181.25099,   0.00006,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -1,   0 },      1.48846,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,  -2,   0,   0 },    202.89251,   0.00008,     0.044 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,  -2,   1,   0 },    202.89251,   0.00008,     0.103 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,  -2,   0 },     11.29152,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,  -1,   0 },     10.64759,   0.00018,     0.087 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,   0,   0 },    197.63421,   0.00015,     0.551 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,   1,   0 },    195.91362,   0.00003,     0.066 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -3,   0 },      1.44742,   0.00004,     0.076 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -2,   0 },     10.51372,   0.00033,    24.411 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -1,   0 },    191.46556,   0.00106,     0.075 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,   0,   0 },    193.45760,   0.00006,     0.038 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   4,  -2,   0 },    191.14204,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -3,   3,   0,   0,   0,   0,   0 },     10.26092,   0.00001,     0.541 },
 { {   0,   0,   2,   0,  -3,   3,   0,   0,   2,  -3,   0 },    254.91214,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -3,   3,   0,   0,   2,  -1,   0 },    334.07521,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,  -2,   0,   0 },     76.38436,   0.00005,     0.044 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,  -2,   1,   0 },     76.02045,   0.00004,     0.105 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   0,  -1,   0 },     77.50562,   0.00001,     0.088 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   0,   0,   0 },     83.82384,   0.00004,     0.522 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   0,   1,   0 },     77.50562,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   2,  -2,   0 },    159.39784,   0.00002,     7.002 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   2,  -1,   0 },    130.21854,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   2,   0,   0 },    159.39784,   0.00002,     0.038 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,  -2,   0,   0 },    208.35504,   0.00003,     0.043 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,  -2,   1,   0 },    208.35147,   0.00003,     0.102 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   0,  -1,   0 },    209.41199,   0.00002,     0.086 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   0,   0,   0 },    209.45979,   0.00007,     0.602 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   2,  -3,   0 },    209.88474,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   2,  -2,   0 },     28.39752,   0.00001,     8.900 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   2,  -1,   0 },     29.41198,   0.00011,     0.076 },
 { {   0,   0,   2,   0,  -2,  -3,   0,   0,   0,   0,   0 },    181.09496,   0.00001,     0.578 },
 { {   0,   0,   2,   0,  -2,  -3,   0,   0,   2,  -3,   0 },    178.90967,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -2,  -3,   0,   0,   2,  -1,   0 },    358.90967,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,  -2,   0,   0 },    336.86475,   0.00001,     0.044 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   0,  -1,   0 },    345.69335,   0.00015,     0.087 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   0,   0,   0 },    339.00971,   0.00003,     0.567 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -3,   0 },    182.36408,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -2,   0 },    165.35711,   0.00141,    95.036 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -1,   0 },    165.63609,   0.00091,     0.076 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,   0,   0 },    165.66216,   0.00006,     0.038 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   4,  -3,   0 },    163.79148,   0.00003,     0.087 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   4,  -2,   0 },    165.37158,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,  -2,   0,   0 },    146.97517,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   0,  -1,   0 },    348.08262,   0.00004,     0.087 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   0,   0,   0 },    150.52119,   0.00002,     0.556 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   2,  -3,   0 },    160.40109,   0.00004,     0.076 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   2,  -2,   0 },    349.51166,   0.00010,    42.689 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   2,  -1,   0 },    168.08249,   0.00019,     0.075 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -6,   1,   0 },    180.98162,   0.00004,     0.017 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -6,   2,   0 },    181.35240,   0.00003,     0.022 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -6,   3,   0 },    180.86455,   0.00001,     0.031 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,  -1,   0 },    178.53433,   0.00012,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   0,   0 },    181.89761,   0.00093,     0.021 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,  -2 },      0.86279,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,   0 },    181.41909,   0.00211,     0.029 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,   2 },    180.91846,   0.00002,     0.133 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   2,   0 },    181.28067,   0.00121,     0.047 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   3,   0 },    180.89533,   0.00004,     0.127 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,  -1,   0 },    181.11348,   0.00004,     0.021 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,   0,   0 },    181.39627,   0.00005,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,   1,   0 },      0.95358,   0.00012,     0.045 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,   2,   0 },      1.14025,   0.00002,     0.114 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -3,   0 },    181.69019,   0.00003,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -2,   0 },    181.58709,   0.00022,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,  -2 },      0.81445,   0.00004,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    181.40070,   0.00527,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,   2 },      0.86642,   0.00016,     0.107 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,  -2 },      0.80053,   0.00019,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    180.95918,   0.08408,     0.044 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   2 },    180.90069,   0.00079,     0.254 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,  -2 },      0.82326,   0.00024,     0.027 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   0 },    180.86110,   0.08911,     0.104 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   2 },      0.96278,   0.00004,     0.058 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   2,   0 },    180.86009,   0.00380,     0.277 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   3,   0 },    180.98602,   0.00025,     0.059 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   4,   0 },    180.87677,   0.00002,     0.033 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,  -2,   0 },    180.32900,   0.00003,     0.027 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,  -1,   0 },    179.51267,   0.00006,     0.042 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,  -1,   1 },    177.58718,   0.00004,     0.096 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,   0 },      1.33764,   0.00129,     0.095 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,   1 },    353.59803,   0.00006,     0.347 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   1,   0 },      0.46288,   0.00015,     0.368 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   2,   0 },      0.18606,   0.00002,     0.063 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -4,   0 },      0.07057,   0.00009,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -3,   0 },      0.01426,   0.00107,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,  -2 },    180.06121,   0.00009,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.98542,   0.01434,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,  -2 },    179.95924,   0.00037,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   0 },    359.98450,   0.18256,     0.088 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   2 },      2.46009,   0.00002,     0.065 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,  -2 },    180.09868,   0.00007,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 },    180.43608,   0.18457,     0.546 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   2 },      0.42763,   0.00011,     0.035 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,  -2 },      0.36703,   0.00014,     0.085 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,   0 },    180.40927,   0.03337,     0.066 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,   2 },      0.31138,   0.00007,     0.024 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   2,   0 },    180.33774,   0.00230,     0.035 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   2,   2 },      0.51257,   0.00001,     0.018 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   3,   0 },    180.40999,   0.00015,     0.024 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -2,   0 },    180.09718,   0.00016,     0.081 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -1,  -1 },    172.49802,   0.00001,     0.080 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -1,   0 },    180.06030,   0.00068,     1.059 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -1,   1 },    177.58718,   0.00002,     0.070 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,  -1 },    203.82142,   0.00007,     1.286 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,   0 },      0.60156,   0.00128,     0.070 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   1,  -1 },    172.49802,   0.00001,     0.071 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   1,   0 },      0.18446,   0.00011,     0.036 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -5,   0 },      0.05492,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -4,   0 },      0.05611,   0.00035,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,  -2 },    180.05103,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,   0 },      0.06638,   0.00553,     0.076 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,  -2 },    180.07733,   0.00029,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   0 },      0.08642,   0.24515,    17.430 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   2 },    180.05452,   0.00011,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,  -2 },    180.11956,   0.00251,     0.074 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   0 },    180.11977,   1.14307,     0.075 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   2 },      0.12242,   0.00227,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,  -2 },      0.10286,   0.00158,     3.621 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,   0 },    180.13179,   0.06863,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,   2 },      0.14340,   0.00038,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,  -2 },    180.10327,   0.00045,     0.077 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,   0 },    180.17824,   0.00504,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,   2 },      0.12183,   0.00005,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,  -2 },    180.11372,   0.00004,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,   0 },    179.96460,   0.00034,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   3,   0 },    179.94793,   0.00003,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -2,   0 },      0.18229,   0.00004,     0.080 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -1,   0 },      0.12383,   0.00047,     0.039 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,   0,   0 },      0.14286,   0.00005,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -5,   0 },      0.10001,   0.00001,     0.067 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -4,   0 },      0.09968,   0.00015,     0.583 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,   0 },      0.09014,   0.00429,     0.087 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,   2 },    180.07773,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,  -2 },      0.11489,   0.00008,     0.488 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,   0 },    180.12877,   0.01229,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,   2 },      0.12918,   0.00007,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,  -2 },      0.12353,   0.00043,     0.089 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,   0 },    180.12629,   0.01038,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,   2 },      0.12185,   0.00006,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,  -2 },      0.26860,   0.00002,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,   0 },    180.11499,   0.00141,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   1,   0 },    180.12985,   0.00012,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   5,  -2,   0 },      0.12198,   0.00002,     0.027 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -4,   0 },      0.08889,   0.00006,     0.043 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -3,   0 },    180.14107,   0.00013,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -2,   0 },    180.12825,   0.00026,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -1,   0 },    180.12965,   0.00012,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,   0,   0 },    180.12185,   0.00003,     0.013 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   8,  -3,   0 },    180.12198,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,  -2,   0,   0 },     41.73264,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,   0,   0,   0 },     33.74277,   0.00003,     0.536 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,   2,  -3,   0 },    331.13216,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,   2,  -1,   0 },      6.00222,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,  -2,   0,   0 },    283.48898,   0.00005,     0.044 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,  -2,   1,   0 },    283.49158,   0.00005,     0.104 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   0,  -1,   0 },    100.10121,   0.00004,     0.088 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   0,   0,   0 },    282.63102,   0.00009,     0.527 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   0,   1,   0 },    283.14100,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   2,  -2,   0 },    104.88439,   0.00002,     7.983 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   2,  -1,   0 },    280.95432,   0.00024,     0.075 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,  -2,   0,   0 },    169.17608,   0.00006,     0.044 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,  -2,   1,   0 },    169.16969,   0.00005,     0.105 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   0,  -1,   0 },    351.46129,   0.00004,     0.088 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   0,   0,   0 },    170.30239,   0.00011,     0.517 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   0,   1,   0 },    169.54836,   0.00002,     0.066 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   2,  -2,   0 },    348.06403,   0.00002,     6.281 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   2,  -1,   0 },    170.93398,   0.00027,     0.075 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   2,   0,   0 },    170.93398,   0.00002,     0.037 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,  -2,   0,   0 },    106.95316,   0.00003,     0.044 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,  -2,   1,   0 },    168.16519,   0.00002,     0.105 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,   0,   0,   0 },     95.44981,   0.00002,     0.500 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,   2,  -3,   0 },    140.76610,   0.00003,     0.077 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,   2,  -1,   0 },    134.07606,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,  -2,  -1,   0 },    198.78129,   0.00001,     0.028 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,  -2,   0,   0 },    198.76425,   0.00013,     0.043 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,  -2,   1,   0 },    198.76374,   0.00014,     0.103 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,  -2,   0 },    198.85517,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,  -1,   0 },    198.98587,   0.00078,     0.087 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,   0,   0 },    199.27773,   0.00046,     0.573 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,   1,   0 },    199.10271,   0.00006,     0.067 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -3,   0 },     54.52565,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -2,   0 },     18.66198,   0.00224,    35.635 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -1,   0 },     18.96250,   0.00362,     0.076 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,   0,   0 },     18.95365,   0.00026,     0.038 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   4,  -3,   0 },     15.26933,   0.00005,     0.087 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   4,  -2,   0 },     18.71783,   0.00007,     0.040 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   4,  -1,   0 },     18.95085,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -1,  -4,   0,   0,   2,  -3,   0 },    196.55183,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -1,  -4,   0,   0,   2,  -2,   0 },    299.69332,   0.00011,   169.910 },
 { {   0,   0,   2,   0,  -1,  -4,   0,   0,   2,  -1,   0 },    156.41913,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -1,  -3,   0,   0,   0,   0,   0 },    137.58176,   0.00001,     0.551 },
 { {   0,   0,   2,   0,  -1,  -3,   0,   0,   2,  -3,   0 },    309.04588,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -1,  -3,   0,   0,   2,  -2,   0 },    297.15271,   0.00001,    25.105 },
 { {   0,   0,   2,   0,  -1,  -3,   0,   0,   2,  -1,   0 },    252.20237,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,  -2,   0,   0 },    351.91841,   0.00006,     0.044 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,  -2,   1,   0 },    351.91841,   0.00005,     0.104 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   0,  -1,   0 },    169.62663,   0.00008,     0.088 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   0,   0,   0 },    351.03939,   0.00011,     0.541 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   0,   1,   0 },    351.91842,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -3,   0 },    146.21699,   0.00001,     0.076 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -2,   0 },    171.27366,   0.00007,    13.554 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -1,   0 },    349.90833,   0.00046,     0.075 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,   0,   0 },    349.90849,   0.00003,     0.038 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,  -2,   0,   0 },    173.33467,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   0,   0,   0 },    173.45135,   0.00003,     0.531 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   2,  -3,   0 },    358.17782,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   2,  -2,   0 },      3.17709,   0.00001,     9.283 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   2,  -1,   0 },    169.60113,   0.00004,     0.075 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -6,   1,   0 },    261.78835,   0.00002,     0.017 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -6,   2,   0 },    261.98526,   0.00001,     0.022 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -6,   3,   0 },    260.34782,   0.00001,     0.031 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,  -1,   0 },    261.37442,   0.00009,     0.016 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   0,   0 },    261.48637,   0.00028,     0.021 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   1,   0 },    262.43306,   0.00063,     0.029 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   2,   0 },    262.77725,   0.00030,     0.048 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -3,   0,   0 },    262.90082,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -3,   1,   0 },     86.46774,   0.00002,     0.046 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -2,   0 },    263.93325,   0.00014,     0.020 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -1,   0 },    263.28669,   0.00154,     0.028 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -1,   2 },     84.73850,   0.00003,     0.108 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,  -2 },     85.32613,   0.00004,     0.020 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    264.37885,   0.01644,     0.044 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   2 },    264.37894,   0.00014,     0.249 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,  -2 },     84.99393,   0.00002,     0.027 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,   0 },    264.83212,   0.01537,     0.105 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,   2 },     84.85461,   0.00003,     0.058 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   2,   0 },    264.94744,   0.00064,     0.271 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   3,   0 },    264.28399,   0.00006,     0.059 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,  -2,   0 },    262.50277,   0.00001,     0.027 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,  -1,   0 },    257.76751,   0.00006,     0.042 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,   0 },    254.07292,   0.00053,     0.096 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,   1 },    273.55272,   0.00003,     0.337 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   1,   0 },    244.21901,   0.00003,     0.357 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -3,   0 },    227.57587,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -2,   0 },    247.83149,   0.00038,     0.041 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -1,   0 },    245.45749,   0.00456,     0.088 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -2 },     90.37415,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 },    270.16044,   0.01589,     0.522 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   1,   0 },    271.98372,   0.00251,     0.066 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   2,   0 },    276.51276,   0.00015,     0.035 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   3,   0 },    236.07786,   0.00002,     0.024 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -2,   0 },     84.36933,   0.00002,     0.082 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -1,  -1 },    257.66853,   0.00003,     0.081 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -1,   0 },    248.94579,   0.00004,     0.972 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -1,   1 },     79.55285,   0.00003,     0.069 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,  -1 },    257.93250,   0.00028,     1.160 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   0 },    253.04605,   0.00032,     0.070 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   1 },     78.79461,   0.00002,     0.036 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   1,  -1 },    257.66853,   0.00003,     0.071 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   1,   0 },    257.77326,   0.00005,     0.036 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -3,   0 },    260.51065,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    201.31508,   0.00062,     7.058 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,  -2 },    320.37553,   0.00004,     0.074 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,  -1 },     79.89304,   0.00003,    39.875 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    307.24896,   0.01299,     0.075 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,   2 },    125.66925,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,  -2 },    194.82026,   0.00004,     5.213 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,  -1 },     80.55549,   0.00004,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,   0 },    299.59680,   0.00072,     0.038 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   1,  -2 },    339.13026,   0.00002,     0.077 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   1,   0 },    282.54731,   0.00005,     0.025 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   3,  -1,   0 },    257.77236,   0.00001,     0.039 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -2,   0 },    304.78109,   0.00017,     0.040 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -1,   0 },    305.63416,   0.00011,     0.026 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,   0,   0 },    329.37532,   0.00002,     0.019 },
 { {   0,   0,   2,   0,  -1,   5,   0,   0,   2,  -2,   0 },    128.62117,   0.00002,     3.211 },
 { {   0,   0,   2,   0,  -1,   5,   0,   0,   2,   0,   0 },    128.62117,   0.00002,     0.037 },
 { {   0,   0,   2,   0,   0,  -7,   0,   0,   2,  -2,   0 },     15.19887,   0.00004,    85.805 },
 { {   0,   0,   2,   0,   0,  -7,   0,   0,   2,  -1,   0 },      8.46291,   0.00002,     0.076 },
 { {   0,   0,   2,   0,   0,  -6,   0,   0,   2,  -2,   0 },      8.23446,   0.00002,    44.857 },
 { {   0,   0,   2,   0,   0,  -6,   0,   0,   2,  -1,   0 },    191.28581,   0.00003,     0.075 },
 { {   0,   0,   2,   0,   0,  -6,   3,   0,   0,   0,   0 },     12.02526,   0.00001,     0.546 },
 { {   0,   0,   2,   0,   0,  -6,   3,   0,   2,  -1,   0 },      7.96907,   0.00003,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -4,   1,   0 },    187.84183,   0.00002,     0.029 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,  -1,   0 },    187.15708,   0.00005,     0.028 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,   0 },    185.89922,   0.00047,     0.044 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   1,   0 },    189.05948,   0.00052,     0.104 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   2,   0 },    189.10414,   0.00001,     0.277 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -2,   0 },      7.61759,   0.00010,     0.041 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -1,   0 },      7.37374,   0.00111,     0.088 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,   0,   0 },    187.72480,   0.00108,     0.546 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,   1,   0 },    188.49423,   0.00017,     0.066 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,   2,   0 },    188.79404,   0.00003,     0.035 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -3,   0 },     15.96455,   0.00004,     0.076 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -2,   0 },      7.95474,   0.00149,    17.781 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,   0 },    187.64895,   0.00675,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   0,  -2 },      8.63899,   0.00001,     3.607 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   0,   0 },    187.22867,   0.00038,     0.038 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   1,   0 },    187.66530,   0.00003,     0.025 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   2,   0 },      8.69629,   0.00001,     0.019 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   4,  -3,   0 },      5.14538,   0.00002,     0.087 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   4,  -2,   0 },    187.72134,   0.00007,     0.040 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   4,  -1,   0 },    187.66997,   0.00006,     0.026 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,  -2,   0,   0 },      5.37312,   0.00009,     0.044 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,  -2,   1,   0 },      5.37314,   0.00010,     0.104 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,  -2,   0 },    186.88581,   0.00001,     0.041 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,  -1,   0 },    186.23149,   0.00013,     0.088 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,   0,   0 },      7.60341,   0.00025,     0.536 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,   1,   0 },      6.05630,   0.00005,     0.066 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -3,   0 },    268.23134,   0.00001,     0.076 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -2,   0 },    185.53792,   0.00010,    11.088 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -1,   0 },      6.18464,   0.00085,     0.075 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,   0,   0 },      6.53457,   0.00004,     0.038 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   4,  -2,   0 },      6.91948,   0.00001,     0.040 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -4,   1,   0 },    272.93798,   0.00003,     0.029 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,  -1,   0 },    273.04698,   0.00008,     0.028 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   0,   0 },    273.18074,   0.00071,     0.044 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   1,   0 },    272.73608,   0.00071,     0.104 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   2,   0 },    272.72041,   0.00004,     0.272 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -2,   0 },     93.12669,   0.00004,     0.041 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -1,   0 },     92.94190,   0.00050,     0.088 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   0,   0 },    273.01802,   0.00150,     0.527 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   1,   0 },    272.83346,   0.00026,     0.066 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   2,   0 },    272.74790,   0.00002,     0.035 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   1,   0,  -1 },    294.42020,   0.00002,     1.184 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -3,   0 },    117.05041,   0.00003,     0.076 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -2,   0 },     92.75404,   0.00030,     8.056 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -1,   0 },    272.87559,   0.00457,     0.075 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,   0,  -2 },     92.71831,   0.00002,     4.776 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,   0,   0 },    272.95177,   0.00022,     0.038 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   4,  -2,   0 },    273.02068,   0.00006,     0.040 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   4,  -1,   0 },    272.88570,   0.00004,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,  -1,   0 },      0.21680,   0.00001,     0.016 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,   0,   0 },    180.20269,   0.00005,     0.021 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,   1,   0 },    180.20507,   0.00008,     0.029 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,   2,   0 },    180.21268,   0.00004,     0.048 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,  -2,   0 },    180.22454,   0.00003,     0.020 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,  -1,   0 },    180.23063,   0.00024,     0.028 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   0,   0 },    180.23405,   0.00358,     0.044 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   0,   2 },    180.23247,   0.00001,     0.248 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   1,   0 },    180.21168,   0.00358,     0.105 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   2,   0 },    180.23834,   0.00016,     0.270 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -1,   0,   0 },    359.92071,   0.00004,     0.096 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -1,   0,   1 },     10.52651,   0.00001,     0.335 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -3,   0 },      0.45413,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   0 },    359.85186,   0.00008,     0.041 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -1,   0 },    359.88318,   0.00168,     0.088 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0 },    180.10531,   0.00731,     0.518 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   1,   0 },    180.09598,   0.00131,     0.066 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   2,   0 },    180.01831,   0.00009,     0.035 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   3,   0 },      0.24797,   0.00001,     0.024 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   1,   0,  -1 },    189.78115,   0.00001,     1.139 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   1,   0,   0 },    359.92908,   0.00005,     0.070 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -3,   0 },    173.22245,   0.00003,     0.076 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -2,   0 },    359.98617,   0.00078,     6.326 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,  -2 },    180.01787,   0.00008,     0.074 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,  -1 },    189.71188,   0.00001,   115.334 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,   0 },    180.03353,   0.01873,     0.075 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,   2 },      0.03430,   0.00004,     0.025 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   0,  -2 },    359.89129,   0.00007,     5.700 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   0,   0 },    180.06486,   0.00088,     0.037 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   1,  -2 },    180.00555,   0.00004,     0.076 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   1,   0 },    180.00555,   0.00004,     0.025 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -3,   0 },    357.22957,   0.00001,     0.086 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -2,   0 },    180.04375,   0.00025,     0.040 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -1,   0 },    180.03992,   0.00015,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,   0,   0 },    180.04866,   0.00002,     0.019 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -4,   0,   0 },    258.56395,   0.00001,     0.021 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -4,   1,   0 },    258.09616,   0.00002,     0.029 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,  -1,   0 },    258.37827,   0.00012,     0.028 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   0,   0 },    258.42306,   0.00101,     0.044 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   0,   2 },    258.44774,   0.00001,     0.245 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   1,   0 },    258.45887,   0.00098,     0.105 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   2,   0 },    258.48153,   0.00005,     0.267 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,  -1,   0 },    256.50120,   0.00027,     0.089 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0 },    259.08132,   0.00107,     0.509 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   1,   0 },    259.62742,   0.00015,     0.066 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   1,   0,  -1 },    246.32225,   0.00002,     1.096 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -2,   0 },    254.48282,   0.00002,     5.207 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -1,   0 },    262.64338,   0.00077,     0.074 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,   0,   0 },    259.25116,   0.00004,     0.037 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   4,  -2,   0 },    262.41934,   0.00001,     0.040 },
 { {   0,   0,   2,   0,   0,   0,  -3,   0,  -2,   0,   0 },    353.60684,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   0,   0,  -3,   0,   0,   0,   0 },    353.69978,   0.00003,     0.509 },
 { {   0,   0,   2,   0,   0,   0,  -3,   0,   2,  -1,   0 },    354.08193,   0.00003,     0.074 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,  -2,   0,   0 },    178.79845,   0.00009,     0.044 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,  -2,   1,   0 },    178.79845,   0.00008,     0.105 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,  -1,   0 },      1.37115,   0.00004,     0.089 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0 },    179.39604,   0.00015,     0.506 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   1,   0 },    176.84794,   0.00002,     0.066 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   2,  -1,   0 },    179.76067,   0.00028,     0.074 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0 },    359.98692,   0.00002,     0.037 },
 { {   0,   0,   2,   0,   0,   0,  -1,   0,  -2,   0,   0 },    249.04721,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0 },    247.25364,   0.00002,     0.503 },
 { {   0,   0,   2,   0,   0,   0,   0,  -2,  -2,   0,   0 },    180.21905,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0 },    180.12520,   0.00005,     0.503 },
 { {   0,   0,   2,   0,   0,   0,   0,  -2,   2,  -1,   0 },    180.01897,   0.00007,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,  -1,  -2,   0,   0 },    258.84540,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0,   0 },    258.67998,   0.00001,     0.502 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -4,   0,   0 },    357.22781,   0.00002,     0.021 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -4,   1,   0 },      0.89775,   0.00002,     0.029 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    349.55878,   0.00005,     0.028 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.68949,   0.00062,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.94570,   0.00055,     0.105 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   2,   0 },     11.44747,   0.00006,     0.265 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,  -1 },     11.15341,   0.00005,     0.042 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0 },    203.56274,   0.00004,     0.096 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   1 },    194.37784,   0.00011,     0.327 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   1,  -1 },     10.39726,   0.00001,     0.095 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    356.89361,   0.00014,     0.089 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0 },     20.93860,   0.00316,     0.500 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0 },     20.40505,   0.00035,     0.066 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   2,   0 },     46.07493,   0.00002,     0.035 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -1,   0 },    204.53669,   0.00001,     0.898 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,  -1 },      8.09133,   0.00024,     1.057 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0 },    204.62437,   0.00007,     0.070 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   1 },    186.15503,   0.00005,     0.036 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    210.82671,   0.00003,     0.077 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },     23.14997,   0.00203,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -2 },     24.35768,   0.00002,     9.300 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },    192.71213,   0.00008,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0 },      9.80521,   0.00009,     0.037 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   1,   0 },    202.23961,   0.00001,     0.025 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   4,  -2,   0 },     20.87349,   0.00003,     0.040 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   4,  -1,   0 },     23.05208,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,   1,   0,   0,  -2,   0,   0 },     63.32040,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   0,   1,   0,   0,  -2,   1,   0 },     63.29034,   0.00001,     0.106 },
 { {   0,   0,   2,   0,   0,   1,   0,   0,   0,   0,   0 },     66.74270,   0.00001,     0.492 },
 { {   0,   0,   2,   0,   0,   2,   0,   0,   2,  -1,   0 },    307.95213,   0.00002,     0.074 },
 { {   0,   0,   2,   0,   1,  -9,   0,   0,   2,  -2,   0 },     86.50807,   0.00001,   210.377 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,  -2,   0,   0 },    287.02856,   0.00004,     0.044 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,  -2,   1,   0 },    289.35855,   0.00004,     0.105 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   0,  -1,   0 },    250.30444,   0.00001,     0.088 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   0,   0,   0 },    315.21386,   0.00004,     0.522 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   2,  -3,   0 },    281.60941,   0.00001,     0.076 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   2,  -2,   0 },      7.34085,   0.00002,     7.115 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   2,  -1,   0 },     18.63096,   0.00008,     0.075 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,  -2,   0,   0 },     79.22284,   0.00005,     0.044 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,  -2,   1,   0 },     79.22422,   0.00004,     0.106 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,   0,  -1,   0 },     78.83298,   0.00001,     0.089 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,   0,   0,   0 },     77.84341,   0.00004,     0.496 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,   0,   1,   0 },     78.83298,   0.00001,     0.065 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,   2,  -3,   0 },    258.43092,   0.00001,     0.077 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,   2,  -1,   0 },     78.43092,   0.00001,     0.074 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -4,   0,   0 },    323.19960,   0.00001,     0.021 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -4,   2,   0 },    143.03238,   0.00001,     0.048 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,  -1,   0 },    139.48499,   0.00004,     0.028 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   0,   0 },    139.78365,   0.00025,     0.044 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   1,   0 },    141.84827,   0.00024,     0.106 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -1,   0,  -1 },    332.53742,   0.00002,     0.042 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -1,   1,  -1 },    152.83598,   0.00001,     0.096 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   0,  -1,   0 },    136.65996,   0.00009,     0.090 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   0,   0,   0 },    133.56749,   0.00015,     0.480 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   1,  -1,  -1 },    332.64034,   0.00003,     0.082 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   1,   0,  -1 },    332.97325,   0.00002,     0.970 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   1,   1,  -1 },    152.69049,   0.00001,     0.070 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,  -1,   0 },     67.11161,   0.00007,     0.074 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -2 },    166.38704,   0.00002,    43.055 },
 { {   0,   0,   2,   0,   1,   2,   0,   0,   2,  -3,   0 },    135.69703,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   1,   2,   0,   0,   2,  -1,   0 },    135.69703,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   1,   5,   0,   0,   2,  -3,   0 },    110.05814,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   1,   5,   0,   0,   2,  -2,   0 },    110.44811,   0.00002,     2.083 },
 { {   0,   0,   2,   0,   1,   5,   0,   0,   2,  -1,   0 },    110.05814,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   1,   5,   0,   0,   2,   0,   0 },    110.44811,   0.00002,     0.037 },
 { {   0,   0,   2,   0,   2, -10,   0,   0,   2,  -3,   0 },    197.61599,   0.00002,     0.076 },
 { {   0,   0,   2,   0,   2, -10,   0,   0,   2,  -2,   0 },    197.67967,   0.00001,    18.146 },
 { {   0,   0,   2,   0,   2, -10,   0,   0,   2,  -1,   0 },     17.61599,   0.00002,     0.075 },
 { {   0,   0,   2,   0,   2,  -8,   0,   0,  -2,   0,   0 },    101.76991,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   2,  -8,   0,   0,   0,   0,   0 },    101.73640,   0.00004,     0.527 },
 { {   0,   0,   2,   0,   2,  -8,   0,   0,   2,  -1,   0 },    101.40928,   0.00007,     0.075 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,  -2,   0,   0 },      8.27321,   0.00006,     0.044 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,  -2,   1,   0 },      8.27321,   0.00006,     0.105 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   0,  -1,   0 },    188.85531,   0.00004,     0.088 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   0,   0,   0 },      8.29109,   0.00012,     0.518 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   0,   1,   0 },      8.27321,   0.00001,     0.066 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,  -2,   0 },    188.74877,   0.00002,     6.371 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,  -1,   0 },      8.73506,   0.00028,     0.075 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,   0,   0 },      8.73503,   0.00002,     0.037 },
 { {   0,   0,   2,   0,   2,  -6,   0,   0,   2,  -3,   0 },    268.74331,   0.00001,     0.077 },
 { {   0,   0,   2,   0,   2,  -6,   0,   0,   2,  -1,   0 },    268.74331,   0.00001,     0.074 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -4,   1,   0 },    323.36776,   0.00001,     0.029 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,  -1,   0 },    323.31743,   0.00004,     0.028 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   0,   0 },    323.36974,   0.00039,     0.044 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   1,   0 },    323.22404,   0.00040,     0.105 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   2,   0 },    323.36242,   0.00002,     0.265 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,  -1,   0 },    321.82601,   0.00013,     0.089 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,   0,   0 },    322.39471,   0.00039,     0.500 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,   1,   0 },    322.71539,   0.00005,     0.066 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   2,  -3,   0 },    348.16303,   0.00003,     0.077 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   2,  -2,   0 },    323.69884,   0.00001,     4.447 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   2,  -1,   0 },    328.81856,   0.00015,     0.074 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,  -2,   0,   0 },    138.26363,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,  -3,   0 },    128.32362,   0.00003,     0.078 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,  -1,   0 },    308.32362,   0.00003,     0.073 },
 { {   0,   0,   2,   0,   3,  -5,   0,   0,   2,  -2,   0 },    328.85444,   0.00002,     3.235 },
 { {   0,   0,   2,   0,   3,  -5,   0,   0,   2,   0,   0 },    328.52329,   0.00003,     0.037 },
 { {   0,   0,   2,   0,   3,  -2,   0,   0,   2,  -3,   0 },    316.21833,   0.00001,     0.078 },
 { {   0,   0,   2,   0,   3,  -2,   0,   0,   2,  -1,   0 },    316.21833,   0.00001,     0.073 },
 { {   0,   0,   2,   0,   3,  -1,   0,   0,   2,  -3,   0 },    126.93963,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   3,  -1,   0,   0,   2,  -1,   0 },    126.93963,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -3,   0,   0,   2,  -3,   0 },    248.54160,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   4,  -3,   0,   0,   2,  -1,   0 },    248.54160,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -3,   0,   0,   2,   0,   0 },    248.54160,   0.00001,     0.037 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -3,   0 },    127.33748,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -2,   0 },    127.13180,   0.00003,     2.019 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -1,   0 },    127.33748,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,   0,  -2 },    307.07676,   0.00001,     6.181 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,   0,   0 },    127.13180,   0.00003,     0.037 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,  -3,   0 },    264.91699,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,  -2,   0 },    310.68138,   0.00001,     2.093 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,  -1,   0 },    264.91699,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,   0,   0 },    310.68138,   0.00001,     0.037 },
 { {   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0 },    228.60012,   0.00001,     0.409 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0 },    253.11000,   0.00002,     0.395 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -3,   0 },    283.57909,   0.00001,     0.080 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -1,   0 },    283.57909,   0.00001,     0.071 },
 { {   0,   0,   2,   2,  -1,   0,   0,   0,   2,  -3,   0 },    102.54136,   0.00001,     0.083 },
 { {   0,   0,   2,   2,  -1,   0,   0,   0,   2,  -1,   0 },    102.54136,   0.00001,     0.069 },
 { {   0,   0,   2,   2,  -1,   0,   0,   0,   2,   0,   0 },    282.54136,   0.00002,     0.036 },
 { {   0,   0,   3, -10,   2,   0,   0,   0,  -2,   0,   0 },     22.51141,   0.00001,     0.037 },
 { {   0,   0,   3, -10,   2,   0,   0,   0,  -2,   2,   0 },    202.51141,   0.00001,     2.672 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   1,   0 },    207.61692,   0.00009,     0.075 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   2,   0 },     26.83788,   0.00010,    89.761 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,   1,   0 },     27.61658,   0.00002,     0.087 },
 { {   0,   0,   3,  -9,   0,   0,   0,   2,  -2,   2,   0 },     44.97077,   0.00001,  1002.658 },
 { {   0,   0,   3,  -9,   0,   0,   1,   0,  -2,   2,   0 },      0.07407,   0.00004,  1313.736 },
 { {   0,   0,   3,  -9,   2,   0,   0,   0,  -2,   0,   0 },     51.51960,   0.00003,     0.038 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,  -1,   0 },    348.73215,   0.00001,     0.069 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,   0,   0 },    304.59287,   0.00004,     0.798 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,   1,   0 },    348.73215,   0.00001,     0.083 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,  -1,   0 },    349.66384,   0.00004,     0.098 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,   0,   0 },    349.65687,   0.00004,     0.043 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   0,  -1,   0 },     21.86394,   0.00001,     0.070 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   0,   0,   0 },     21.82577,   0.00009,     1.001 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   0,   1,   0 },     21.86394,   0.00001,     0.082 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   2,  -1,   0 },     22.31594,   0.00007,     0.095 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   2,   0,   0 },     21.99916,   0.00007,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -4,   0,   0 },    209.72639,   0.00002,     0.020 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -4,   1,   0 },    209.71654,   0.00004,     0.027 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -4,   2,   0 },    209.71688,   0.00002,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,  -2,   0 },    209.71556,   0.00002,     0.019 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    208.76462,   0.00007,     0.026 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   0,   0 },    209.49356,   0.00053,     0.039 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   0,   2 },    209.71735,   0.00003,     0.903 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   1,   0 },    210.39021,   0.00037,     0.080 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   2,   0 },     32.09918,   0.00003,     1.293 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -1,   0,   0 },     28.93912,   0.00004,     0.075 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -2,   0 },     38.22120,   0.00003,     0.036 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -1,   0 },     31.31024,   0.00101,     0.070 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   0,   0 },     29.79747,   0.01121,     0.999 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,   0 },     29.21785,   0.00360,     0.082 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   2,   0 },     28.91381,   0.00031,     0.039 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   3,   0 },     27.91575,   0.00001,     0.026 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   1,  -2,   0 },     29.90611,   0.00001,     0.066 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -3,   0 },     30.36812,   0.00003,     0.062 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -2,   0 },     30.30248,   0.00038,     0.360 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,  -2 },    209.74692,   0.00002,     0.061 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,   0 },     29.96125,   0.00971,     0.095 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,  -2 },     29.74450,   0.00013,     0.322 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   0 },     29.68106,   0.00905,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   2 },    209.74901,   0.00002,     0.020 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   1,  -2 },    209.74644,   0.00001,     0.099 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   1,   0 },     29.46493,   0.00089,     0.027 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   2,   0 },     29.36321,   0.00010,     0.020 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   3,  -1,   0 },    209.74023,   0.00001,     0.044 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   3,   0,   0 },     29.74023,   0.00002,     0.028 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,  -2,   0 },     30.02831,   0.00019,     0.046 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,  -1,   0 },     29.80919,   0.00039,     0.028 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   0,   0 },     29.65492,   0.00016,     0.021 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   1,   0 },     29.73229,   0.00004,     0.016 },
 { {   0,   0,   3,  -8,   5,  -5,   0,   0,   2,   0,   0 },    165.95272,   0.00002,     0.042 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   2,  -1 },    144.54893,   0.00019,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   3,  -1 },    144.40986,   0.00023,     0.087 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   0,   0 },    203.69163,   0.00003,     0.039 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,  -1 },    322.16169,   0.00005,     0.038 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,   1 },    322.27495,   0.00004,     3.012 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,  -1 },    322.16254,   0.00070,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,  -3 },    328.43370,   0.00001,     0.037 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,  -1 },    144.43917,   0.01806,   650.233 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,   1 },    328.43370,   0.00001,     0.037 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   3,  -1 },    144.47758,   0.00086,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   4,  -1 },    144.47814,   0.00006,     0.038 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0 },    250.46817,   0.00008,     0.072 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0,   0 },    255.38725,   0.00040,     1.385 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   1,   0 },    251.46978,   0.00008,     0.080 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   0,  -1 },    143.17675,   0.00002,     0.563 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   1,  -1 },    143.84352,   0.00054,     0.087 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   2,  -1 },    144.34912,   0.00023,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   3,  -1 },    144.39416,   0.00003,     0.026 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -1,   0 },    351.93488,   0.00009,     0.093 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,   0,   0 },    341.78371,   0.00009,     0.042 },
 { {   0,   0,   3,  -7,   1,   0,   0,   0,   0,   0,   0 },     92.84746,   0.00001,     1.569 },
 { {   0,   0,   3,  -7,   2,   0,   0,   0,   0,   0,   0 },    300.59636,   0.00001,     1.808 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -4,   0,   0 },    271.97640,   0.00002,     0.020 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -4,   2,   0 },    273.16172,   0.00002,     0.043 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,  -1,   0 },    272.00825,   0.00013,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,   0 },    271.35434,   0.00113,     0.040 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,   2 },    271.74655,   0.00003,     0.522 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   1,   0 },    273.64510,   0.00165,     0.086 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   2,   0 },    280.93973,   0.00004,     0.631 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   0 },     93.23811,   0.00002,     0.080 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   1 },    352.31167,   0.00003,     1.158 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0 },    268.66588,   0.00020,     0.037 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0 },    268.91653,   0.00310,     0.074 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1 },    170.65860,   0.00002,    42.830 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0 },    271.34254,   0.01673,     5.260 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0 },    269.15751,   0.00325,     0.077 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0 },    269.24534,   0.00022,     0.038 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   3,   0 },    268.53487,   0.00001,     0.025 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,  -2,   0 },     94.11047,   0.00001,     0.070 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,   0,  -1 },    353.64689,   0.00003,     0.804 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,   0,   0 },     93.02865,   0.00003,     0.082 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -2,   0 },    278.56985,   0.00008,     0.509 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,   0 },    275.23688,   0.00234,     0.089 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,  -2 },    274.74556,   0.00003,     0.435 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,   0 },    274.96465,   0.00162,     0.041 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   1,   0 },    273.48571,   0.00017,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   4,  -2,   0 },    274.53363,   0.00003,     0.044 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   4,  -1,   0 },    274.52485,   0.00003,     0.028 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   4,   0,   0 },    274.68577,   0.00002,     0.020 },
 { {   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0,   0 },    331.07661,   0.00002,     8.183 },
 { {   0,   0,   3,  -6,   1,   0,   0,   0,  -2,   0,   0 },    124.88875,   0.00001,     0.040 },
 { {   0,   0,   3,  -6,   1,   0,   0,   0,   0,   0,   0 },    126.45061,   0.00009,     9.452 },
 { {   0,   0,   3,  -6,   1,   0,   0,   0,   2,   0,   0 },    125.45024,   0.00002,     0.041 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -4,   1,   0 },    331.15956,   0.00001,     0.028 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,  -1,   0 },    331.15955,   0.00005,     0.026 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   0,   0 },    331.07540,   0.00046,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   0,   2 },    331.15956,   0.00001,     0.479 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   1,   0 },    331.36313,   0.00072,     0.087 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   2,   0 },    331.15429,   0.00003,     0.571 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -2,   0 },    329.91149,   0.00006,     0.038 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -1,   0 },    329.95950,   0.00092,     0.075 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0,   0 },    330.47414,   0.00491,    46.520 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   1,   0 },    329.96819,   0.00092,     0.076 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   2,   0 },    329.92302,   0.00006,     0.038 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,  -2,   0 },    331.72637,   0.00002,     0.557 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,  -1,   0 },    331.44039,   0.00073,     0.087 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   0,  -2 },    331.23132,   0.00001,     0.470 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   0,   0 },    331.12252,   0.00047,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   1,   0 },    331.21866,   0.00005,     0.026 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   4,  -2,   0 },    331.40405,   0.00001,     0.044 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   4,  -1,   0 },    331.23132,   0.00002,     0.028 },
 { {   0,   0,   3,  -6,   3,   0,   0,   0,   0,   0,   0 },    235.66375,   0.00005,    15.922 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,  -2,   0,   0 },    127.66330,   0.00001,     0.040 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,  -2,   1,   0 },    127.66330,   0.00002,     0.087 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,   0,   0,   0 },    179.45246,   0.00001,   229.298 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,   2,  -1,   0 },    127.77034,   0.00002,     0.087 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,   2,   0,   0 },    127.77034,   0.00001,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   0,   0 },    301.53802,   0.00004,     0.020 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   1,   0 },    301.59268,   0.00011,     0.028 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   2,   0 },    301.79753,   0.00006,     0.044 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,  -2,   0 },    301.72929,   0.00003,     0.020 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,  -1,   0 },    301.39694,   0.00030,     0.027 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   0 },    301.84789,   0.00341,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   2 },    301.88879,   0.00006,     0.408 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   1,   0 },    301.79270,   0.00510,     0.090 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   2,   0 },    302.04841,   0.00022,     0.473 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   3,   0 },    301.73212,   0.00001,     0.065 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   0 },    121.50926,   0.00011,     0.083 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   1 },     22.82202,   0.00003,     0.717 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   1,   0 },    121.83608,   0.00001,     0.814 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -3,   0 },    299.69505,   0.00001,     0.025 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0 },    300.21053,   0.00029,     0.038 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0 },    300.17010,   0.00410,     0.077 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0 },    300.80387,   0.02009,     2.928 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -2 },    118.46630,   0.00018,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0 },    300.06852,   0.00374,     0.074 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2 },    118.12651,   0.00037,   123.995 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0 },    299.95449,   0.00024,     0.037 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -2 },    115.26960,   0.00001,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   3,   0 },    299.55187,   0.00001,     0.025 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,  -1,   0 },    121.18257,   0.00001,     1.833 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,  -1,   1 },    203.17617,   0.00003,     0.078 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,  -1 },     23.01226,   0.00013,     1.404 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,   0 },    303.32113,   0.00001,     0.079 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,   1 },    203.02177,   0.00002,     0.038 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -2,   0 },    301.95798,   0.00011,     0.698 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,   0 },    302.07160,   0.00341,     0.085 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,  -2 },    301.24254,   0.00005,     0.566 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,   0 },    301.13669,   0.00196,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   1,  -2 },    298.51055,   0.00003,     0.087 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   1,   0 },    301.18386,   0.00020,     0.026 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   2,   0 },    301.38256,   0.00002,     0.019 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,  -2,   0 },    301.69647,   0.00004,     0.043 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,  -1,   0 },    301.42743,   0.00007,     0.027 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,   0,   0 },    301.62577,   0.00004,     0.020 },
 { {   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0,   0 },      0.40637,   0.00002,     1.960 },
 { {   0,   0,   3,  -5,   2,   0,   0,   0,   2,   0,   0 },      0.98515,   0.00001,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,  -1,   0 },    330.16111,   0.00003,     0.016 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   0,   0 },    329.00970,   0.00013,     0.021 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   1,   0 },    327.46365,   0.00029,     0.028 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   2,   0 },    328.46994,   0.00018,     0.045 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -3,   1,   0 },    150.32139,   0.00004,     0.044 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -2,   0 },    328.80224,   0.00009,     0.020 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    327.55457,   0.00097,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -1,   2 },    148.61491,   0.00001,     0.097 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,  -2 },    148.74651,   0.00003,     0.020 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   0 },    328.29186,   0.01099,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   2 },    328.45312,   0.00015,     0.335 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,  -2 },    148.73909,   0.00003,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,   0 },    328.67574,   0.01431,     0.094 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   2,   0 },    329.42285,   0.00063,     0.378 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   3,   0 },    328.17078,   0.00004,     0.063 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,  -1,   1 },    234.81545,   0.00005,     0.088 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,  -1 },     54.52628,   0.00001,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   0 },    151.03220,   0.00227,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   1 },     54.51751,   0.00006,     0.519 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   1,   0 },    150.81731,   0.00024,     0.568 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   2,   0 },    151.11338,   0.00002,     0.067 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   0 },    330.06033,   0.00004,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0 },    329.66250,   0.00066,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -2 },    149.91072,   0.00002,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0 },    329.61383,   0.00994,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2 },    329.93597,   0.00001,     0.069 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2 },    149.84907,   0.00003,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0 },    330.08714,   0.04667,     1.145 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2 },    149.73435,   0.00003,     0.036 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0 },    330.63019,   0.00789,     0.071 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,   2 },    149.91072,   0.00002,     0.024 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0 },    330.98090,   0.00055,     0.037 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0 },    331.04282,   0.00004,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -3,   0 },    331.55475,   0.00004,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -2,   0 },    331.55649,   0.00061,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,  -2 },    151.65098,   0.00001,     0.037 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,  -1 },     54.79510,   0.00012,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   0 },    331.25359,   0.01425,    72.752 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   1 },     54.81358,   0.00022,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   2 },    151.65098,   0.00001,     0.037 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -2 },    331.23561,   0.00002,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -1 },     54.84877,   0.00095,     5.540 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },    331.23293,   0.01126,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   1 },    234.83983,   0.00018,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   2 },    151.23553,   0.00002,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,  -1 },     54.79510,   0.00012,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,   0 },    331.23687,   0.00076,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,   1 },    234.83811,   0.00002,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   2,   0 },    331.23562,   0.00005,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -2,   0 },    315.20699,   0.00004,     1.110 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -1,  -2 },    151.34018,   0.00001,     0.069 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -1,   0 },    330.11336,   0.00475,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,  -2 },    331.23377,   0.00013,     0.810 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,   0 },    331.65667,   0.00260,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   1,  -2 },    151.34018,   0.00001,     0.083 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   1,   0 },    332.14263,   0.00029,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   2,   0 },    330.32884,   0.00003,     0.019 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -2,   0 },    331.05931,   0.00019,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -1,   0 },    331.23745,   0.00030,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,   0,   0 },    331.23723,   0.00012,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,  -2,   0 },    322.41723,   0.00007,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,  -1,   0 },    330.58913,   0.00008,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,   0,   0 },    335.37803,   0.00003,     0.020 },
 { {   0,   0,   3,  -4,   1,   0,   0,   0,  -2,   0,   0 },    158.53103,   0.00001,     0.042 },
 { {   0,   0,   3,  -4,   1,   0,   0,   0,   0,   0,   0 },    176.70724,   0.00006,     1.044 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,  -2,   0,   0 },    236.72569,   0.00003,     0.042 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,  -2,   1,   0 },    236.72569,   0.00004,     0.096 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,   0,  -1,   0 },    232.89325,   0.00001,     0.082 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,   0,   0,   0 },    222.57221,   0.00008,     0.960 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,   0,   1,   0 },    232.89325,   0.00001,     0.070 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,   2,   0,   0 },     29.50165,   0.00001,     0.039 },
 { {   0,   0,   3,  -3,  -3,   0,   0,   0,   0,   0,   0 },    307.06808,   0.00001,     0.868 },
 { {   0,   0,   3,  -3,  -2,   0,   0,   0,  -2,   0,   0 },    120.84255,   0.00001,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   0,   0 },    191.31089,   0.00005,     0.021 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   1,   0 },    190.41720,   0.00010,     0.029 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   2,   0 },    186.91965,   0.00005,     0.046 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,  -2,   0 },    187.44462,   0.00004,     0.020 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,  -1,   0 },    188.22724,   0.00032,     0.027 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   0 },    187.75965,   0.00405,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   2 },    188.06188,   0.00003,     0.285 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,   0 },    187.34444,   0.00445,     0.099 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   2,   0 },    187.87178,   0.00020,     0.315 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   0,   0 },      5.85755,   0.00007,     0.091 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   2,   0 },      1.05370,   0.00001,     0.064 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -3,   0 },    182.26009,   0.00001,     0.026 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0 },    183.05121,   0.00032,     0.040 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0 },    183.14913,   0.00455,     0.084 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0 },    182.64615,   0.01454,     0.712 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0 },    182.41704,   0.00228,     0.068 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0 },    182.49060,   0.00015,     0.036 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,  -2,   0 },      2.65712,   0.00002,     0.079 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,  -1,  -1 },    279.36498,   0.00001,     0.077 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,  -1,   0 },      3.87204,   0.00001,     1.931 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   0,  -1 },    277.64218,   0.00014,     2.848 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   0,   0 },      8.37493,   0.00004,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   1,  -1 },    279.36498,   0.00001,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -3,   0 },    355.44561,   0.00003,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -2,   0 },      1.11029,   0.00072,     2.710 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,   0 },      0.86750,   0.00639,     0.078 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,  -2 },    359.11052,   0.00002,     1.423 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,   0 },      1.04873,   0.00062,     0.038 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   1,   0 },    356.66735,   0.00004,     0.025 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   4,  -2,   0 },      0.56060,   0.00008,     0.041 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   4,  -1,   0 },      0.57210,   0.00008,     0.027 },
 { {   0,   0,   3,  -3,   2,   0,   0,   0,   2,  -3,   0 },     60.84329,   0.00001,     0.074 },
 { {   0,   0,   3,  -3,   2,   0,   0,   0,   2,  -1,   0 },     60.84329,   0.00001,     0.077 },
 { {   0,   0,   3,  -2,  -2,   0,   0,   0,   2,  -2,   0 },    257.52570,   0.00003,   166.820 },
 { {   0,   0,   3,  -2,  -2,   0,   0,   0,   2,  -1,   0 },    200.05796,   0.00002,     0.075 },
 { {   0,   0,   3,  -2,  -1,   0,   0,   0,   2,  -2,   0 },     11.99109,   0.00002,    12.770 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,  -1,   0 },     75.93781,   0.00003,     0.016 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   0,   0 },    257.10852,   0.00011,     0.021 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   1,   0 },    256.42089,   0.00018,     0.029 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   2,   0 },    256.52895,   0.00009,     0.048 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -3,   0,   0 },    256.48334,   0.00001,     0.028 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,  -2,   0 },    256.86423,   0.00003,     0.020 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,  -1,   0 },    256.61174,   0.00039,     0.028 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   0 },    256.49156,   0.00459,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   2 },    256.84439,   0.00004,     0.247 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   1,   0 },    256.92620,   0.00423,     0.105 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   2,   0 },    256.90742,   0.00019,     0.269 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -1,   0,   0 },    257.31521,   0.00008,     0.096 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -1,   2,   0 },    257.26573,   0.00001,     0.062 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -3,   0 },     77.25593,   0.00001,     0.026 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0 },    256.07867,   0.00012,     0.041 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0 },    255.82151,   0.00138,     0.088 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1 },    310.84469,   0.00003,     0.087 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0 },    254.39593,   0.00459,     0.516 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0 },    256.02376,   0.00074,     0.066 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0 },    256.32633,   0.00004,     0.035 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,  -1 },    307.66275,   0.00009,     1.133 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,   0 },    257.23286,   0.00007,     0.070 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -3,   0 },    237.15108,   0.00001,     0.076 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },    255.58156,   0.00014,     6.150 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,  -1 },    130.33601,   0.00064,   241.391 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    259.02532,   0.00196,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,  -1 },    130.75842,   0.00021,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },    259.56942,   0.00017,     0.037 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   4,  -2,   0 },    259.11230,   0.00003,     0.040 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   4,  -1,   0 },    258.99152,   0.00002,     0.026 },
 { {   0,   0,   3,  -2,   2,   0,   0,   0,   2,   0,  -2 },    340.27142,   0.00001,   436.527 },
 { {   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   0 },    180.16039,   0.00001,     0.435 },
 { {   0,   0,   3,  -1,  -2,   0,   0,   0,   2,  -3,   0 },      1.51409,   0.00001,     0.079 },
 { {   0,   0,   3,  -1,  -2,   0,   0,   0,   2,  -1,   0 },    181.51409,   0.00001,     0.073 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,  -2,   0,   0 },    159.02383,   0.00004,     0.045 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,  -2,   1,   0 },    158.96542,   0.00002,     0.111 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0 },    221.08242,   0.00007,     0.405 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   1,   0,  -1 },    339.04593,   0.00002,     0.707 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -1,   0 },    233.53070,   0.00003,     0.072 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,  -2,   0,   0 },    204.45137,   0.00005,     0.045 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,  -2,   1,   0 },    204.75111,   0.00002,     0.111 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   0,  -1,   0 },     32.37048,   0.00002,     0.093 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   0,   0,   0 },    209.15255,   0.00007,     0.401 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   0,   1,   0 },    208.68407,   0.00001,     0.063 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   2,  -1,   0 },    210.63943,   0.00013,     0.072 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   2,   0,   0 },     32.04839,   0.00003,     0.037 },
 { {   0,   0,   3,   0,  -6,   2,   0,   0,   0,   0,   0 },    185.33787,   0.00001,     0.390 },
 { {   0,   0,   3,   0,  -6,   2,   0,   0,   2,  -3,   0 },     10.49486,   0.00001,     0.080 },
 { {   0,   0,   3,   0,  -6,   2,   0,   0,   2,  -1,   0 },    190.49486,   0.00001,     0.071 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,  -2,   0,   0 },     32.00637,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   0,   0,   0 },     37.13100,   0.00003,     0.375 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,  -3,   0 },    244.17585,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,  -2,   0 },     69.03290,   0.00002,     1.124 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,  -1,   0 },     46.73641,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,  -1,   0 },    194.01581,   0.00004,     0.028 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   0,   0 },    193.92881,   0.00026,     0.045 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   1,   0 },    193.83859,   0.00023,     0.112 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,  -1,   0 },     40.99694,   0.00002,     0.094 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 },    198.09669,   0.00044,     0.388 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,   1,   0 },    198.63818,   0.00009,     0.063 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,   2,   0 },    198.92986,   0.00001,     0.034 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   1,   0,  -1 },    290.76128,   0.00002,     0.656 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -2,   0 },    200.59454,   0.00003,     1.243 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -1,   0 },    199.52518,   0.00103,     0.071 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   0,  -2 },    200.60244,   0.00004,     2.123 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   0,   0 },     21.56655,   0.00012,     0.037 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   1,   0 },     20.64237,   0.00003,     0.025 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   4,  -2,   0 },    199.39722,   0.00001,     0.039 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   4,   0,   0 },    200.60130,   0.00002,     0.019 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,  -2,   0,   0 },    166.25768,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   0,   0,   0 },    174.14888,   0.00003,     0.378 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,  -3,   0 },      1.52660,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,  -2,   0 },    181.25175,   0.00001,     1.146 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,  -1,   0 },    181.19608,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -5,   3,   0,   0,   2,  -3,   0 },     59.69935,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -5,   3,   0,   0,   2,  -2,   0 },    239.69935,   0.00001,     1.103 },
 { {   0,   0,   3,   0,  -5,   3,   0,   0,   2,  -1,   0 },    239.69935,   0.00001,     0.071 },
 { {   0,   0,   3,   0,  -5,   3,   0,   0,   2,   0,   0 },     59.69935,   0.00001,     0.036 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,  -2,   0,   0 },    343.07928,   0.00008,     0.045 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,  -2,   1,   0 },    343.20785,   0.00006,     0.115 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   0,  -1,   0 },    261.11823,   0.00002,     0.095 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   0,   0,   0 },    347.11876,   0.00011,     0.364 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   0,   1,   0 },    342.15283,   0.00002,     0.062 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   0,   2,   0 },    348.35145,   0.00001,     0.034 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,  -3,   0 },    338.74679,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,  -1,   0 },    345.49412,   0.00020,     0.070 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,   0,   0 },    168.29184,   0.00006,     0.036 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,   2,   0 },    168.47486,   0.00001,     0.019 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -4,   0,   0 },    182.49177,   0.00003,     0.021 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -4,   1,   0 },    182.54691,   0.00003,     0.030 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -4,   2,   0 },    182.48803,   0.00001,     0.049 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    182.36502,   0.00014,     0.028 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,   0 },    181.65196,   0.00167,     0.045 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,   2 },    182.02730,   0.00001,     0.210 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   1,   0 },    182.40301,   0.00145,     0.113 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   2,   0 },    180.23132,   0.00006,     0.225 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,  -2,   0 },     39.14450,   0.00002,     0.042 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,  -1,   0 },     48.27073,   0.00030,     0.094 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 },    186.61024,   0.00241,     0.376 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   1,   0 },    187.12548,   0.00067,     0.063 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   2,   0 },    192.84973,   0.00003,     0.034 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,  -3,   0 },     14.46989,   0.00005,     0.040 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,  -2,   0 },     14.13625,   0.00018,     0.087 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,   0,  -1 },    280.19094,   0.00006,     0.622 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -3,   0 },    186.92375,   0.00003,     0.081 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -2,   0 },    184.62105,   0.00017,     1.125 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -1,   0 },    189.35457,   0.00593,     0.071 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   0,  -2 },    190.42992,   0.00023,     1.801 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   0,   0 },     12.52838,   0.00088,     0.036 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   1,   0 },     10.34107,   0.00009,     0.025 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   2,   0 },     10.22636,   0.00002,     0.019 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -4,   0 },     14.63823,   0.00014,     0.075 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -3,   0 },     14.46989,   0.00452,   569.346 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -2,   0 },    193.70045,   0.00038,     0.075 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -1,   0 },    193.70035,   0.00003,     0.038 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   4,  -2,   0 },    188.43030,   0.00009,     0.039 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   4,  -1,   0 },    189.80523,   0.00005,     0.026 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   4,   0,   0 },      7.15468,   0.00003,     0.019 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   5,  -4,   0 },     14.40380,   0.00007,     0.087 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   5,  -3,   0 },     14.46989,   0.00005,     0.040 },
 { {   0,   0,   3,   0,  -4,   1,   0,   0,   2,  -3,   0 },    181.30296,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -4,   1,   0,   0,   2,  -1,   0 },      1.30296,   0.00001,     0.071 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,  -2,   0,   0 },     98.24051,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,   0,  -1,   0 },     79.44695,   0.00003,     0.095 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,   0,   0,   0 },    113.47590,   0.00002,     0.366 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,   2,  -3,   0 },     17.86930,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,   2,  -1,   0 },    195.95218,   0.00001,     0.070 },
 { {   0,   0,   3,   0,  -4,   3,   0,   0,   0,  -1,   0 },    134.21038,   0.00004,     0.095 },
 { {   0,   0,   3,   0,  -4,   3,   0,   0,   2,  -1,   0 },    134.21165,   0.00002,     0.070 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,  -2,   0,   0 },    246.12093,   0.00006,     0.046 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,  -2,   1,   0 },    246.15473,   0.00004,     0.116 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,   0,  -1,   0 },     68.27753,   0.00006,     0.096 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,   0,   0,   0 },    244.33239,   0.00003,     0.353 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,   2,  -1,   0 },     68.27658,   0.00003,     0.070 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   0,   0 },    150.03212,   0.00009,     0.021 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   1,   0 },    162.44700,   0.00017,     0.030 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   2,   0 },    165.84814,   0.00006,     0.049 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -3,   0,   0 },    160.87005,   0.00001,     0.029 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,  -2,   0 },    153.40802,   0.00006,     0.021 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    153.28759,   0.00053,     0.028 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,  -2 },    346.34287,   0.00002,     0.020 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   0 },    163.62400,   0.00655,     0.045 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   2 },    166.31023,   0.00005,     0.206 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   1,   0 },    165.26267,   0.00556,     0.114 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   2,   0 },    165.27831,   0.00025,     0.221 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   3,   0 },    165.24948,   0.00002,     0.056 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -1,   0,   0 },    347.99896,   0.00009,     0.104 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -2,   0 },     72.30594,   0.00010,     0.042 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -1,   0 },     75.77000,   0.00442,     0.095 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,  -2 },    351.76535,   0.00001,     0.041 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 },    169.78399,   0.00851,     0.364 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   1,  -2 },    180.09146,   0.00001,     0.092 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   1,   0 },    175.09060,   0.00246,     0.062 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   2,   0 },    174.69319,   0.00017,     0.034 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   3,   0 },    177.41519,   0.00001,     0.024 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   1,   0,   0 },    355.03284,   0.00008,     0.066 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -3,   0 },    224.05574,   0.00008,     0.081 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -2,   0 },    151.43656,   0.00069,     1.028 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,  -2 },    179.68394,   0.00003,     0.079 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,   0 },    173.71802,   0.02166,     0.070 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,   2 },    359.40561,   0.00005,     0.024 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,  -2 },    180.50849,   0.00083,     1.564 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,   0 },      5.44054,   0.00368,     0.036 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,   2 },    179.63226,   0.00002,     0.018 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   1,  -2 },    358.94407,   0.00002,     0.072 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   1,   0 },      6.32409,   0.00032,     0.025 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   2,   0 },    359.91898,   0.00007,     0.019 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   3,  -4,   0 },    180.09471,   0.00001,     0.076 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   3,  -2,   0 },    211.41965,   0.00005,     0.075 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   3,  -1,   0 },    267.14884,   0.00002,     0.038 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -3,   0 },     78.86859,   0.00003,     0.080 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -2,   0 },    160.78539,   0.00032,     0.039 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -1,  -2 },    180.09146,   0.00001,     0.083 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -1,   0 },    146.22767,   0.00011,     0.026 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   0,   0 },    359.72684,   0.00004,     0.019 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   2,   0 },      0.03516,   0.00001,     0.013 },
 { {   0,   0,   3,   0,  -3,   1,   0,   0,   0,  -1,   0 },     78.50121,   0.00005,     0.095 },
 { {   0,   0,   3,   0,  -3,   1,   0,   0,   2,  -1,   0 },     78.50073,   0.00003,     0.070 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,  -2,   0,   0 },    190.40447,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   0,   0,   0 },    194.93615,   0.00003,     0.376 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,  -3,   0 },     21.12747,   0.00002,     0.081 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,  -2,   0 },    199.60506,   0.00001,     1.127 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,  -1,   0 },    199.10379,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   3,  -3,   0 },     18.09192,   0.00018,  1601.744 },
 { {   0,   0,   3,   0,  -2,  -2,   0,   0,   2,  -3,   0 },    172.54515,   0.00001,     0.082 },
 { {   0,   0,   3,   0,  -2,  -2,   0,   0,   2,  -1,   0 },    352.54515,   0.00001,     0.070 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   0,   0 },     75.88388,   0.00016,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   1,   0 },     76.30885,   0.00033,     0.030 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   2,   0 },     75.60761,   0.00010,     0.050 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -3,   0,   0 },     76.28782,   0.00002,     0.029 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -3,   3,   0 },    256.26387,   0.00001,     0.182 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -3,   0 },    257.33259,   0.00001,     0.016 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -2,   0 },     75.01697,   0.00005,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -1,   0 },     75.25630,   0.00053,     0.028 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,  -2 },    255.46846,   0.00002,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,   0 },     75.71608,   0.00860,     0.046 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,   2 },     75.71200,   0.00006,     0.202 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   1,   0 },     75.66754,   0.00702,     0.116 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   2,   0 },     76.74344,   0.00034,     0.217 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   3,   0 },     77.14481,   0.00002,     0.056 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -1,   0,   0 },    254.40682,   0.00006,     0.105 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -3,   0 },    256.44724,   0.00008,     0.027 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -2,   0 },    256.88794,   0.00076,     0.042 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -1,  -2 },     77.18490,   0.00002,     0.027 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -1,   0 },    256.83321,   0.00813,     0.096 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 },     73.98285,   0.00487,     0.353 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   1,   0 },     73.08045,   0.00095,     0.062 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   2,   0 },     70.45478,   0.00006,     0.034 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,  -2,  -1 },     75.99754,   0.00002,     0.040 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,  -1,  -1 },     77.09600,   0.00012,     0.087 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,   0,  -1 },    259.42867,   0.00006,     0.563 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -3,   0 },    262.37671,   0.00003,     0.082 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -2,   0 },    259.65668,   0.00060,     0.946 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -1,   0 },    268.93847,   0.00452,     0.070 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   0,  -2 },    316.39743,   0.00004,     1.382 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   0,   0 },    241.95199,   0.00063,     0.036 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   1,   0 },    228.28328,   0.00006,     0.024 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -3,  -1 },     59.85798,   0.00001,     0.075 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,  -1 },     75.99754,   0.00161,   277.143 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,   0 },     77.20480,   0.00002,     0.074 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,  -1 },    257.55559,   0.00042,     0.075 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,   0 },     77.18490,   0.00007,     0.037 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,   0,  -1 },    257.28611,   0.00003,     0.038 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -3,   0 },     77.97372,   0.00001,     0.080 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -2,   0 },    263.06044,   0.00013,     0.039 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -1,   0 },    256.84235,   0.00016,     0.026 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,   0,   0 },    263.46336,   0.00003,     0.019 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   5,  -3,  -1 },     77.27166,   0.00003,     0.087 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   5,  -2,  -1 },     75.99754,   0.00002,     0.040 },
 { {   0,   0,   3,   0,  -2,   5,   0,   0,   2,  -3,   0 },    308.90894,   0.00001,     0.083 },
 { {   0,   0,   3,   0,  -2,   5,   0,   0,   2,  -1,   0 },    308.90894,   0.00001,     0.069 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,  -2,   0,   0 },    171.21417,   0.00007,     0.045 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,  -2,   1,   0 },    173.43575,   0.00005,     0.114 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   0,  -1,   0 },     45.11452,   0.00005,     0.095 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   0,   0,   0 },    183.03229,   0.00011,     0.364 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   0,   2,   0 },    188.72979,   0.00001,     0.034 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,  -3,   0 },    188.23031,   0.00002,     0.081 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,  -1,   0 },    183.06585,   0.00018,     0.070 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,   0,   0 },      8.39022,   0.00006,     0.036 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,   2,   0 },      8.70826,   0.00001,     0.019 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -4,   0,   0 },    166.86567,   0.00003,     0.021 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -4,   1,   0 },    163.24095,   0.00004,     0.030 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,  -1,   0 },    164.96772,   0.00012,     0.029 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    165.05969,   0.00105,     0.046 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   1,   0 },    165.50009,   0.00078,     0.117 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    166.01848,   0.00004,     0.213 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -1,   0,   0 },    154.83633,   0.00002,     0.106 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,  -2,   0 },     14.75438,   0.00001,     0.042 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,  -1,   0 },    143.96842,   0.00015,     0.097 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 },    174.40671,   0.00035,     0.343 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,   1,   0 },    182.25516,   0.00005,     0.062 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },     53.19664,   0.00003,     0.069 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,  -2,   0,   0 },     74.80576,   0.00006,     0.046 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,  -2,   1,   0 },     83.04526,   0.00005,     0.116 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   0,  -1,   0 },    264.72844,   0.00006,     0.096 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   0,   0,   0 },     63.97861,   0.00004,     0.353 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   2,  -1,   0 },    304.68435,   0.00004,     0.070 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   2,   0,   0 },    185.46661,   0.00002,     0.036 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   3,  -2,  -1 },    171.66373,   0.00004,   403.851 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,  -2,   0,   0 },    270.37657,   0.00006,     0.046 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,  -2,   1,   0 },    270.77835,   0.00002,     0.116 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   0,   0,   0 },    271.39488,   0.00005,     0.349 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   0,   1,   0 },    269.97123,   0.00002,     0.062 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   2,  -1,   0 },    272.72768,   0.00012,     0.070 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   2,   0,   0 },     92.72768,   0.00005,     0.036 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,  -1,   0 },    176.80007,   0.00002,     0.028 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   0,   0 },    176.77092,   0.00020,     0.046 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   1,   0 },    176.62709,   0.00016,     0.117 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,  -1,   0 },    352.95809,   0.00003,     0.097 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,   0,   0 },    177.60960,   0.00021,     0.345 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,   1,   0 },    178.07279,   0.00007,     0.062 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,  -2,   0 },    180.00873,   0.00002,     0.890 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,  -1,   0 },    180.44439,   0.00044,     0.070 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   0,  -2 },    180.00873,   0.00001,     1.265 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   0,   0 },      0.00560,   0.00009,     0.036 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   1,   0 },    359.96741,   0.00003,     0.024 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -4,   1,   0 },     76.95217,   0.00001,     0.030 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,  -1,   0 },     76.94075,   0.00005,     0.029 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   0,   0 },     76.94705,   0.00033,     0.046 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   1,   0 },     76.94757,   0.00026,     0.117 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,  -2,   0 },    257.10551,   0.00001,     0.042 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,  -1,   0 },    257.54878,   0.00005,     0.097 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,   0,   0 },     76.65210,   0.00022,     0.341 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,   1,   0 },     76.65316,   0.00003,     0.062 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   2,  -2,   0 },    257.11960,   0.00001,     0.864 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   2,  -1,   0 },    257.08369,   0.00007,     0.069 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   2,   0,   0 },     81.30098,   0.00001,     0.036 },
 { {   0,   0,   3,   0,   0,  -1,   0,   0,  -2,   0,   0 },    155.89028,   0.00010,     0.046 },
 { {   0,   0,   3,   0,   0,  -1,   0,   0,  -2,   1,   0 },    155.89712,   0.00006,     0.117 },
 { {   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,   0 },    157.56747,   0.00002,     0.337 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },    268.74508,   0.00008,     0.046 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   1,   0 },    290.94861,   0.00004,     0.118 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   1 },    120.82067,   0.00002,     0.247 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    308.36335,   0.00002,     0.098 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    292.64900,   0.00021,     0.333 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    294.45273,   0.00003,     0.062 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,  -1 },    304.13372,   0.00003,     0.514 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -3,   0 },    216.15544,   0.00003,     0.083 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.09836,   0.00003,     0.816 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    303.94270,   0.00004,     0.069 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     78.12629,   0.00003,     0.036 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   4,   0,   0 },     51.16061,   0.00002,     0.019 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,  -2,   0,   0 },     38.97324,   0.00002,     0.046 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,   2,  -3,   0 },    306.86712,   0.00001,     0.084 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,   2,  -2,   0 },    126.86719,   0.00001,     0.763 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,   2,  -1,   0 },    126.86712,   0.00001,     0.069 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,   2,   0,   0 },    305.38578,   0.00002,     0.036 },
 { {   0,   0,   3,   0,   2,  -5,   0,   0,  -2,   0,   0 },    220.30897,   0.00003,     0.046 },
 { {   0,   0,   3,   0,   2,  -5,   0,   0,   0,   0,   0 },    219.03288,   0.00001,     0.333 },
 { {   0,   0,   3,   0,   2,  -5,   0,   0,   2,  -3,   0 },    149.14221,   0.00001,     0.083 },
 { {   0,   0,   3,   0,   2,  -5,   0,   0,   2,  -1,   0 },    149.14221,   0.00001,     0.069 },
 { {   0,   0,   3,   0,   3,  -2,   0,   0,   2,  -3,   0 },    307.12243,   0.00001,     0.084 },
 { {   0,   0,   3,   0,   3,  -2,   0,   0,   2,  -2,   0 },    127.12243,   0.00001,     0.709 },
 { {   0,   0,   3,   0,   3,  -2,   0,   0,   2,  -1,   0 },    307.12243,   0.00001,     0.068 },
 { {   0,   0,   3,   0,   3,  -2,   0,   0,   2,   0,   0 },    127.12243,   0.00001,     0.036 },
 { {   0,   0,   3,   0,   4,  -5,   0,   0,   2,   0,   0 },    192.61099,   0.00001,     0.036 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,  -2,   0 },    168.97108,   0.00002,     0.437 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,   0,   0 },    168.92470,   0.00003,     0.035 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,  -2,   0 },    184.68820,   0.00001,     0.354 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,   0,   0 },    184.68820,   0.00001,     0.034 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   1,   0 },    332.22778,   0.00005,     0.075 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   3,   0 },    323.17001,   0.00001,     0.076 },
 { {   0,   0,   4, -11,   0,   2,   0,   0,  -2,   2,   0 },    356.98663,   0.00001,   151.221 },
 { {   0,   0,   4, -11,   1,   0,   0,   0,  -2,   1,   0 },    350.56273,   0.00002,     0.075 },
 { {   0,   0,   4, -11,   1,   0,   0,   0,  -2,   2,   0 },    218.87448,   0.00002,   102.094 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,   0,   0 },    259.10251,   0.00001,     0.759 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   2,   0,   0 },    323.07357,   0.00002,     0.043 },
 { {   0,   0,   4, -10,   2,   0,   0,   0,  -2,   0,   0 },     51.32151,   0.00002,     0.039 },
 { {   0,   0,   4, -10,   2,   0,   0,   0,  -2,   1,   0 },     51.50867,   0.00001,     0.079 },
 { {   0,   0,   4, -10,   2,   0,   0,   0,  -2,   2,   0 },    231.24107,   0.00001,     1.598 },
 { {   0,   0,   4, -10,   2,   0,   0,   0,  -2,   3,   0 },     51.50867,   0.00001,     0.072 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -1,   0,   1 },    293.38240,   0.00002,     3.723 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   0,  -1,   0 },    255.11163,   0.00002,     0.071 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   0,   0,   0 },    231.55152,   0.00016,     1.274 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   0,   1,   0 },    271.08625,   0.00003,     0.080 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -1,   0 },    325.70712,   0.00006,     0.093 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,   0,   0 },    313.11309,   0.00005,     0.042 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,  -1,   0 },    242.61029,   0.00003,     0.026 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   0,   0 },    243.31159,   0.00023,     0.040 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   1,   0 },    245.01650,   0.00032,     0.085 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0 },    240.72582,   0.00005,     0.037 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0 },    240.02370,   0.00069,     0.074 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0 },    242.87629,   0.00373,     3.945 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,   0 },    240.76922,   0.00074,     0.077 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   2,   0 },    241.24033,   0.00006,     0.038 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   1,   0,  -1 },    324.55670,   0.00001,     0.765 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -2,   0 },    247.12224,   0.00002,     0.493 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -1,   0 },    248.26064,   0.00053,     0.089 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,   0 },    246.51899,   0.00038,     0.041 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   1,   0 },    246.92435,   0.00004,     0.026 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   4,  -1,   0 },    246.94665,   0.00001,     0.028 },
 { {   0,   0,   4,  -8,   1,   0,   0,   0,   0,   0,   0 },     94.08514,   0.00001,     5.911 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -4,   0,   0 },     99.17335,   0.00001,     0.020 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -4,   1,   0 },    279.17335,   0.00003,     0.028 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -4,   2,   0 },    279.17348,   0.00001,     0.044 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,  -1,   0 },    279.03666,   0.00007,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   0,   0 },    279.04653,   0.00064,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   0,   2 },    279.17335,   0.00002,     0.474 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   1,   0 },    279.60917,   0.00099,     0.087 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   2,   0 },    279.17356,   0.00004,     0.564 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -2,   0 },    276.71815,   0.00008,     0.038 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -1,   0 },    276.83449,   0.00125,     0.075 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   0,   0 },    278.08371,   0.00650,  1750.265 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   1,   0 },    276.83348,   0.00125,     0.075 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   2,   0 },    276.71763,   0.00008,     0.038 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -2,   0 },    279.17001,   0.00004,     0.564 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -1,   0 },    279.60572,   0.00099,     0.087 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   0,  -2 },    279.17021,   0.00002,     0.475 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   0,   0 },    279.04359,   0.00064,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   1,   0 },    279.03378,   0.00007,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   4,  -2,   0 },    279.17009,   0.00001,     0.044 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   4,  -1,   0 },    279.17021,   0.00003,     0.028 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   4,   0,   0 },     99.17021,   0.00001,     0.020 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,  -2,   0,   0 },    301.45402,   0.00002,     0.040 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,  -2,   1,   0 },    301.19091,   0.00001,     0.086 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   0,  -1,   0 },    301.51487,   0.00005,     0.075 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   0,   0,   0 },    300.21814,   0.00018,    11.784 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   0,   1,   0 },    301.51487,   0.00005,     0.076 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   2,  -1,   0 },    301.53249,   0.00001,     0.088 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   2,   0,   0 },    301.64415,   0.00002,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   0,   0 },    286.77996,   0.00002,     0.013 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   1,   0 },    286.78089,   0.00004,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   2,   0 },    286.78203,   0.00005,     0.021 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   3,   0 },    286.78214,   0.00002,     0.029 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,  -1,   0 },    286.54532,   0.00017,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   0 },    286.68832,   0.00106,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   2 },    286.78020,   0.00002,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,  -2 },    106.78140,   0.00002,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   0 },    286.87773,   0.00264,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   2 },    286.77901,   0.00002,     0.107 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   2,   0 },    287.11011,   0.00189,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   3,   0 },    287.69514,   0.00005,     0.103 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,  -1,   0 },    286.78051,   0.00003,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   0,   0 },    286.77618,   0.00003,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   1,   0 },    106.94286,   0.00017,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   2,   0 },    107.29832,   0.00006,     0.094 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -3,   0 },    286.21074,   0.00008,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -2,   0 },    286.20069,   0.00076,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -2,   2 },    106.78088,   0.00001,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,  -2 },    106.63383,   0.00006,     0.015 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    286.40935,   0.00876,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   2 },    106.37678,   0.00027,     0.090 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,  -2 },    106.70915,   0.00024,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   0 },    286.68957,   0.09002,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   2 },    286.68001,   0.00210,     0.475 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,  -2 },    107.07842,   0.00033,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   0 },    287.11216,   0.13960,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,  -2 },    106.78331,   0.00002,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   0 },    287.81515,   0.00484,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   2 },    106.78153,   0.00003,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   3,   0 },    288.98969,   0.00022,     0.067 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   4,   0 },    289.77464,   0.00001,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,  -1,   0 },    106.19302,   0.00022,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   0,   0 },    106.69076,   0.00237,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   1,   0 },    107.80252,   0.00021,     1.128 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   2,   0 },    109.47326,   0.00002,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -4,   0 },    284.82440,   0.00005,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -3,   0 },    284.93758,   0.00080,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,  -2 },    105.02794,   0.00006,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   0 },    284.98930,   0.01162,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   2 },    104.67763,   0.00002,     3.003 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,  -2 },    105.24010,   0.00039,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   0 },    284.98776,   0.17083,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   2 },    284.45177,   0.00024,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,  -2 },    107.22398,   0.00034,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   0 },    285.98707,   0.90114,  1783.359 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   2 },    107.22659,   0.00034,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,  -2 },    284.42756,   0.00023,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   0 },    284.98767,   0.17082,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   2 },    105.24074,   0.00039,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,  -2 },    104.67825,   0.00002,     2.993 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   0 },    284.98981,   0.01162,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   2 },    105.02829,   0.00006,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   3,   0 },    284.93803,   0.00080,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   4,   0 },    284.82479,   0.00005,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,  -2,   0 },    109.47486,   0.00002,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,  -1,   0 },    107.80120,   0.00022,     1.127 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   0,   0 },    106.69093,   0.00237,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   1,   0 },    106.19418,   0.00022,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -4,   0 },    289.77337,   0.00001,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -3,   0 },    288.98983,   0.00022,     0.067 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,  -2 },    106.78416,   0.00003,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   0 },    287.81538,   0.00485,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   2 },    106.78354,   0.00002,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   0 },    287.11388,   0.13971,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   2 },    107.08017,   0.00033,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -2 },    286.68261,   0.00210,     0.474 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   0 },    286.69269,   0.09009,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   2 },    106.71065,   0.00024,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,  -2 },    106.37929,   0.00027,     0.090 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   0 },    286.41041,   0.00875,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   2 },    106.63554,   0.00006,     0.015 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   2,  -2 },    106.78406,   0.00001,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   2,   0 },    286.20210,   0.00076,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   3,   0 },    286.21296,   0.00008,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,  -2,   0 },    107.30215,   0.00006,     0.094 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,  -1,   0 },    106.94518,   0.00017,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   0,   0 },    286.78777,   0.00003,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   1,   0 },    286.78446,   0.00003,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -3,   0 },    287.69301,   0.00005,     0.103 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -2,   0 },    287.11186,   0.00189,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,  -2 },    286.78482,   0.00002,     0.107 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,   0 },    286.87976,   0.00264,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,   2 },    106.78391,   0.00002,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,  -2 },    286.78391,   0.00002,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,   0 },    286.69301,   0.00107,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   1,   0 },    286.54818,   0.00017,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -3,   0 },    286.78362,   0.00002,     0.029 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -2,   0 },    286.78374,   0.00005,     0.021 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -1,   0 },    286.78382,   0.00004,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,   0,   0 },    286.78387,   0.00002,     0.013 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,  -2,   0,   0 },     61.30125,   0.00007,     0.040 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,  -2,   1,   0 },     61.30125,   0.00015,     0.087 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   0,  -1,   0 },    241.30037,   0.00002,     0.075 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   0,   0,   0 },    241.30081,   0.00001,   590.730 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   0,   1,   0 },    241.30125,   0.00002,     0.075 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   2,  -1,   0 },     61.30037,   0.00015,     0.087 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   2,   0,   0 },     61.30037,   0.00007,     0.040 },
 { {   0,   0,   4,  -7,  -2,   0,   0,   0,   0,   0,   0 },     33.69604,   0.00001,     9.124 },
 { {   0,   0,   4,  -7,  -1,   0,   0,   0,   0,   0,   0 },    220.18463,   0.00001,     5.157 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   0,   0 },    272.71125,   0.00003,     0.020 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   1,   0 },    272.52220,   0.00002,     0.028 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   2,   0 },    272.57606,   0.00003,     0.044 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -3,   2,  -1 },     41.50224,   0.00001,     0.042 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -3,   3,  -1 },     41.47425,   0.00001,     0.095 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,  -2,   0 },    272.77672,   0.00002,     0.020 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,  -1,   0 },    272.42072,   0.00015,     0.027 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   0 },    273.02089,   0.00170,     0.041 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   2 },    272.82899,   0.00004,     0.419 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   1,   0 },    272.81422,   0.00261,     0.089 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   2,   0 },    272.88077,   0.00010,     0.487 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   0 },     92.30986,   0.00003,     0.083 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   1 },    353.51657,   0.00001,     0.751 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   2,   0 },     92.62687,   0.00001,     0.069 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0 },    270.56356,   0.00014,     0.038 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0 },    270.60119,   0.00219,     0.077 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0 },    271.44643,   0.01099,     3.594 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0 },    270.40845,   0.00201,     0.074 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0 },    270.32088,   0.00013,     0.037 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   3,   0 },     92.53218,   0.00001,     0.025 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,  -2,   0 },    272.16968,   0.00001,     0.072 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,   0,  -1 },    353.41415,   0.00007,     1.289 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,   0,   0 },     92.16968,   0.00001,     0.079 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -2,   0 },    272.04631,   0.00007,     0.669 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,   0 },    272.77206,   0.00189,     0.085 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,  -2 },    272.16968,   0.00003,     0.547 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,   0 },    272.18908,   0.00109,     0.040 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   1,   0 },    271.98986,   0.00013,     0.026 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   4,  -2,   0 },    272.34000,   0.00002,     0.043 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   4,   0,   0 },    272.38272,   0.00002,     0.020 },
 { {   0,   0,   4,  -7,   1,   0,   0,   0,   0,   0,   0 },    318.22523,   0.00001,     2.759 },
 { {   0,   0,   4,  -7,   2,   0,   0,   0,   0,   0,   0 },    331.25358,   0.00002,     2.238 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   0,   0 },    300.53917,   0.00004,     0.021 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   1,   0 },    300.53006,   0.00008,     0.028 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   2,   0 },    300.41270,   0.00005,     0.045 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,  -2,   0 },    300.51729,   0.00003,     0.020 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,  -1,   0 },    300.59992,   0.00028,     0.027 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   0 },    300.38934,   0.00332,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   2 },    300.45823,   0.00004,     0.343 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   1,   0 },    300.61624,   0.00439,     0.094 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   2,   0 },    300.42864,   0.00020,     0.387 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,  -1,   1 },    204.00194,   0.00004,     0.088 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   0 },    121.58608,   0.00021,     0.087 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   1 },     23.52022,   0.00004,     0.537 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0 },    300.90066,   0.00021,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0 },    300.19798,   0.00296,     0.080 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0 },    301.21730,   0.01410,     1.235 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0 },    301.39358,   0.00237,     0.071 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0 },    301.28065,   0.00015,     0.037 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -2,   0 },    299.73138,   0.00002,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -2,   1 },    204.01170,   0.00003,    11.156 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,  -1 },     23.99072,   0.00005,     0.074 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   0 },    301.28450,   0.00023,    12.968 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   1 },     24.00194,   0.00022,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,  -1 },     24.02114,   0.00042,     4.101 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   0 },    302.22085,   0.00061,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   1 },    204.01481,   0.00008,     0.038 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   1,  -1 },     23.99072,   0.00005,     0.077 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   1,   0 },    302.24105,   0.00004,     0.038 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -2,   0 },    282.46053,   0.00006,     1.037 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -1,   0 },    296.09565,   0.00209,     0.081 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,  -2 },    302.10962,   0.00004,     0.771 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,   0 },    301.06135,   0.00099,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   1,   0 },    301.78517,   0.00013,     0.026 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,  -3,   0 },     74.81841,   0.00001,     0.095 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,  -2,   0 },    284.26226,   0.00004,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,  -1,   0 },    276.31860,   0.00005,     0.027 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,   0,   0 },    302.23472,   0.00002,     0.020 },
 { {   0,   0,   4,  -6,   1,   0,   0,   0,   0,   0,   0 },    150.72078,   0.00002,     1.118 },
 { {   0,   0,   4,  -6,   1,   0,   0,   0,   2,   0,   0 },    153.57136,   0.00001,     0.039 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,  -2,   0,   0 },    281.53588,   0.00004,     0.042 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,  -2,   1,   0 },    289.64352,   0.00004,     0.095 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,   0,  -1,   0 },    328.67411,   0.00001,     0.081 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,   0,   0,   0 },    330.48698,   0.00010,     1.022 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,   0,   1,   0 },    328.67411,   0.00001,     0.070 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,   2,   0,   0 },      0.57942,   0.00002,     0.039 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -4,   0,   0 },    150.54280,   0.00003,     0.021 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -4,   1,   0 },    155.87821,   0.00007,     0.029 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -4,   2,   0 },    156.15847,   0.00004,     0.046 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,  -2,   0 },    156.63247,   0.00002,     0.020 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,  -1,   0 },    155.63310,   0.00021,     0.027 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,   0 },    156.49504,   0.00253,     0.043 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,   2 },    156.60079,   0.00003,     0.290 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   1,   0 },    156.19895,   0.00282,     0.099 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   2,   0 },    156.94985,   0.00012,     0.321 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -1,   0,   0 },    336.16138,   0.00004,     0.091 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0 },    153.10385,   0.00019,     0.040 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0 },    153.16602,   0.00292,     0.084 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0 },    152.76092,   0.00948,     0.745 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0 },    152.58201,   0.00152,     0.069 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0 },    152.64466,   0.00010,     0.036 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,  -2,   0 },    336.94431,   0.00002,     0.078 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,  -1,  -1 },    236.07738,   0.00003,     0.077 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,  -1,   1 },     56.45125,   0.00002,     0.072 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,  -1 },    236.08784,   0.00026,     3.475 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,   0 },    336.94431,   0.00002,     0.073 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,   1 },     56.04727,   0.00003,     0.037 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   1,  -1 },    236.07738,   0.00003,     0.074 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -2,   0 },    331.81893,   0.00040,     2.313 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -1,   0 },    331.33500,   0.00382,     0.078 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,  -2 },    332.47709,   0.00002,     1.306 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,   0 },    333.25408,   0.00041,     0.038 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   1,   0 },    330.01258,   0.00003,     0.025 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   4,  -2,   0 },    331.20460,   0.00005,     0.041 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   4,  -1,   0 },    331.41834,   0.00005,     0.027 },
 { {   0,   0,   4,  -5,   2,   0,   0,   0,   2,   0,   0 },     29.28089,   0.00001,     0.038 },
 { {   0,   0,   4,  -4,  -2,   0,   0,   0,   2,  -2,   0 },    304.81126,   0.00001,    14.417 },
 { {   0,   0,   4,  -4,   0,  -3,   0,   0,   2,  -1,   0 },     43.45149,   0.00002,     0.075 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -4,   0,   0 },    213.10792,   0.00004,     0.021 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -4,   1,   0 },    211.51417,   0.00004,     0.029 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -4,   2,   0 },    205.21000,   0.00003,     0.047 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,  -2,   0 },    205.25096,   0.00001,     0.020 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    216.57460,   0.00012,     0.028 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   0 },    206.72557,   0.00175,     0.044 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   2 },    206.27167,   0.00002,     0.251 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   1,   0 },    205.67151,   0.00173,     0.104 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   2,   0 },    207.27878,   0.00007,     0.274 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -1,  -1,   0 },    228.31666,   0.00001,     0.042 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -1,   0,   0 },     48.29721,   0.00009,     0.095 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -3,   0 },      2.37878,   0.00002,     0.026 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0 },    357.27706,   0.00022,     0.041 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0 },    356.65726,   0.00303,     0.088 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0 },    191.18699,   0.00404,     0.534 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0 },    189.86500,   0.00068,     0.066 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0 },    188.01414,   0.00005,     0.035 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,  -1 },    278.91983,   0.00007,     1.220 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,   0 },     48.29858,   0.00005,     0.070 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -3,   0 },    350.40644,   0.00004,     0.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -2,   0 },      0.33668,   0.00214,    10.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,  -2 },    181.44751,   0.00006,     0.074 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   0 },    181.91276,   0.02008,     0.075 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   2 },      1.97449,   0.00004,     0.025 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,  -2 },    357.84690,   0.00007,     4.269 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,   0 },    182.43699,   0.00104,     0.038 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   1,  -2 },    180.64281,   0.00002,     0.077 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   1,   0 },    180.63716,   0.00003,     0.025 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -3,   0 },      2.41753,   0.00004,     0.086 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -2,   0 },    182.14808,   0.00024,     0.040 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -1,   0 },    182.22544,   0.00016,     0.026 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,   0,   0 },    181.82093,   0.00003,     0.019 },
 { {   0,   0,   4,  -4,   2,   0,   0,   0,   2,  -1,   0 },    241.92932,   0.00002,     0.074 },
 { {   0,   0,   4,  -3,  -2,   0,   0,   0,   0,   0,   0 },    151.59258,   0.00001,     0.447 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -4,   0,   0 },    258.78829,   0.00002,     0.021 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -4,   1,   0 },     76.83305,   0.00001,     0.030 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,  -1,   0 },     76.46745,   0.00004,     0.028 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   0,   0 },     76.97083,   0.00037,     0.045 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   1,   0 },     77.02255,   0.00034,     0.110 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0 },     86.80608,   0.00018,     0.092 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0 },     92.53030,   0.00025,     0.416 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0 },     70.81756,   0.00004,     0.064 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   1,   0,  -1 },    310.01902,   0.00006,     0.740 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -2,   0 },     73.16255,   0.00001,     1.585 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -1,   0 },     41.96687,   0.00026,     0.072 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,   0 },    195.29004,   0.00003,     0.037 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,  -1,   0 },    154.68854,   0.00003,     0.029 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   0,   0 },    154.62830,   0.00026,     0.046 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   1,   0 },    155.37539,   0.00022,     0.117 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0 },    154.77604,   0.00005,     0.097 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0 },    165.13910,   0.00010,     0.341 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0 },    159.50898,   0.00001,     0.062 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -3,   0 },    186.59820,   0.00001,     0.083 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -1,   0 },    186.59820,   0.00001,     0.069 },
 { {   0,   0,   4,   0,  -7,   0,   0,   0,   0,   0,   0 },    209.68924,   0.00001,     0.293 },
 { {   0,   0,   4,   0,  -7,   0,   0,   0,   2,  -1,   0 },    211.31862,   0.00004,     0.067 },
 { {   0,   0,   4,   0,  -7,   0,   0,   0,   2,   0,   0 },     30.78033,   0.00002,     0.036 },
 { {   0,   0,   4,   0,  -7,   5,   0,   0,   2,  -2,   0 },     17.57651,   0.00001,     0.554 },
 { {   0,   0,   4,   0,  -7,   5,   0,   0,   2,   0,   0 },    197.57651,   0.00001,     0.035 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,  -2,   0,   0 },    195.35908,   0.00006,     0.047 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,  -2,   1,   0 },    195.41881,   0.00003,     0.125 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   0,  -1,   0 },     22.90172,   0.00002,     0.102 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   0,   0,   0 },    198.49698,   0.00005,     0.286 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   0,   1,   0 },    198.45822,   0.00002,     0.060 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,  -2,   0 },    200.62488,   0.00001,     0.581 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,  -1,   0 },    200.62679,   0.00021,     0.067 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,   0,   0 },     20.51236,   0.00009,     0.035 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,   1,   0 },     20.59541,   0.00002,     0.024 },
 { {   0,   0,   4,   0,  -6,   2,   0,   0,   2,  -2,   0 },    180.86660,   0.00001,     0.559 },
 { {   0,   0,   4,   0,  -6,   2,   0,   0,   2,   0,   0 },      0.86660,   0.00001,     0.035 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,  -2,   0,   0 },    341.14143,   0.00002,     0.047 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   0,   0,   0 },    345.15246,   0.00001,     0.273 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,  -2,   0 },    348.38664,   0.00002,     0.529 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,  -1,   0 },    348.18925,   0.00004,     0.066 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,   0,   0 },    168.53215,   0.00001,     0.035 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,  -1,   0 },    183.44329,   0.00003,     0.029 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   0,   0 },    183.37433,   0.00024,     0.047 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   1,   0 },    182.93256,   0.00019,     0.127 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,  -1,   0 },     15.46280,   0.00006,     0.103 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 },    186.63934,   0.00019,     0.279 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   1,   0 },    190.63390,   0.00010,     0.059 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   2,   0 },    189.96826,   0.00001,     0.033 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   1,   0,  -1 },    279.54852,   0.00001,     0.396 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -2,   0 },    190.30882,   0.00003,     0.554 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -1,   0 },    190.28079,   0.00103,     0.066 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   0,  -2 },    190.33379,   0.00003,     0.680 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   0,   0 },     10.32275,   0.00038,     0.035 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   1,   0 },     10.30275,   0.00007,     0.024 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   4,  -2,   0 },    190.19487,   0.00002,     0.038 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   4,  -1,   0 },     10.30239,   0.00001,     0.025 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -4,   1,   0 },    166.59743,   0.00003,     0.031 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    166.82614,   0.00008,     0.029 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   0,   0 },    164.95994,   0.00076,     0.047 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   1,   0 },    166.03743,   0.00054,     0.128 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   2,   0 },    164.98070,   0.00002,     0.184 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,  -2,   0 },      6.57990,   0.00001,     0.044 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,  -1,   0 },    355.89655,   0.00020,     0.104 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 },    170.23424,   0.00044,     0.273 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   1,   0 },    175.74620,   0.00034,     0.059 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   2,   0 },    178.88334,   0.00002,     0.033 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -2,   0 },    180.12376,   0.00010,     0.529 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -1,   0 },    179.86950,   0.00308,     0.066 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   0,  -2 },    179.79519,   0.00010,     0.643 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   0,   0 },      0.66210,   0.00125,     0.035 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   1,   0 },    359.63586,   0.00013,     0.024 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   2,   0 },      0.69812,   0.00001,     0.018 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,  -2,   0 },    179.20680,   0.00005,     0.038 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,   0,   0 },    359.38266,   0.00003,     0.019 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,   1,   0 },    179.79519,   0.00001,     0.015 },
 { {   0,   0,   4,   0,  -3,  -5,   0,   0,   2,  -2,   0 },    198.20250,   0.00001,     0.555 },
 { {   0,   0,   4,   0,  -3,  -5,   0,   0,   2,   0,   0 },     18.20250,   0.00001,     0.035 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -4,   0,   0 },    242.83085,   0.00002,     0.022 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -4,   1,   0 },     63.53677,   0.00003,     0.031 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,  -1,   0 },     64.17913,   0.00007,     0.029 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   0,   0 },     63.97047,   0.00069,     0.048 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   1,   0 },     64.00815,   0.00050,     0.129 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   2,   0 },     64.09748,   0.00003,     0.181 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,  -3,   0 },    220.11468,   0.00001,     0.028 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,  -2,   0 },    265.04503,   0.00002,     0.044 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,  -1,   0 },    339.56530,   0.00015,     0.105 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 },     64.17514,   0.00025,     0.267 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,   1,   0 },     63.12464,   0.00006,     0.059 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   1,   0,  -1 },    259.10916,   0.00001,     0.371 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,  -2,   0 },      5.10737,   0.00001,     0.507 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,  -1,   0 },      2.19620,   0.00025,     0.066 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,   0,   0 },    175.37062,   0.00009,     0.035 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,   1,   0 },    204.59145,   0.00002,     0.024 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -3,   0 },      8.58659,   0.00001,     0.088 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -2,   0 },    188.58593,   0.00001,     0.530 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -1,   0 },    188.49814,   0.00004,     0.066 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,   0,   0 },      8.51434,   0.00001,     0.035 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -4,   1,   0 },    331.91163,   0.00003,     0.031 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -4,   3,   0 },    153.91645,   0.00002,     0.170 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    331.07351,   0.00006,     0.029 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   0,   0 },    331.49117,   0.00056,     0.048 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   1,   0 },    331.06287,   0.00038,     0.131 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   2,   0 },    331.61048,   0.00002,     0.178 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,  -2,   0 },    154.13843,   0.00004,     0.044 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,  -1,   0 },    153.64961,   0.00029,     0.106 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 },    329.33388,   0.00010,     0.261 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   2,  -3,   0 },    153.64088,   0.00001,     0.089 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   2,  -2,   0 },    154.30014,   0.00001,     0.486 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   2,  -1,   0 },    156.54069,   0.00004,     0.065 },
 { {   0,   0,   4,   0,  -2,   5,   0,   0,   2,  -3,   0 },    119.42302,   0.00002,     0.091 },
 { {   0,   0,   4,   0,  -2,   5,   0,   0,   2,  -2,   0 },    119.28068,   0.00002,     0.449 },
 { {   0,   0,   4,   0,  -2,   5,   0,   0,   2,  -1,   0 },    119.42302,   0.00002,     0.065 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,  -2,   0,   0 },     66.11883,   0.00006,     0.048 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,  -2,   1,   0 },     64.73120,   0.00003,     0.132 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   0,   0,   0 },     73.78519,   0.00001,     0.255 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    317.91238,   0.00003,     0.065 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   2,   0,   0 },    168.04955,   0.00007,     0.035 },
 { {   0,   0,   4,   0,   0,  -5,   0,   0,   2,   0,   0 },     92.67342,   0.00001,     0.035 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,  -2,   0,   0 },    176.17132,   0.00003,     0.048 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,   0,   0,   0 },    177.49605,   0.00001,     0.259 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,   2,  -1,   0 },    179.88721,   0.00005,     0.065 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,   2,   0,   0 },    359.88722,   0.00003,     0.035 },
 { {   0,   0,   4,   0,   0,  -3,   0,   0,  -2,   0,   0 },     74.94416,   0.00002,     0.048 },
 { {   0,   0,   4,   0,   0,  -2,   0,   0,  -2,   0,   0 },    333.83422,   0.00003,     0.048 },
 { {   0,   0,   4,   0,   0,  -2,   0,   0,   0,   0,   0 },    334.42674,   0.00001,     0.254 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0 },    200.82458,   0.00003,     0.250 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -3,   0 },    306.54187,   0.00003,     0.091 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    128.06472,   0.00004,     0.449 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,  -2 },    308.22368,   0.00003,     0.528 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    128.14308,   0.00008,     0.035 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21587,   0.00002,     0.018 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   4,  -2,   0 },    308.11929,   0.00001,     0.037 },
 { {   0,   0,   4,   0,   1,   0,   0,   0,   2,  -1,   0 },    119.69063,   0.00004,     0.064 },
 { {   0,   0,   4,   0,   1,   0,   0,   0,   2,   0,   0 },    297.34397,   0.00006,     0.035 },
 { {   0,   0,   4,   0,   2,  -5,   0,   0,   2,  -3,   0 },    318.69333,   0.00002,     0.091 },
 { {   0,   0,   4,   0,   2,  -5,   0,   0,   2,  -2,   0 },    319.51461,   0.00002,     0.449 },
 { {   0,   0,   4,   0,   2,  -5,   0,   0,   2,  -1,   0 },    318.69333,   0.00002,     0.065 },
 { {   0,   0,   4,   1,  -2,   0,   0,   0,   2,   0,   0 },     98.62210,   0.00002,     0.034 },
 { {   0,   0,   4,   1,  -1,   0,   0,   0,   2,   0,   0 },    305.30141,   0.00002,     0.034 },
 { {   0,   0,   4,   2,  -2,   0,   0,   0,   2,   0,   0 },    127.97736,   0.00002,     0.034 },
 { {   0,   0,   5, -14,   4,   0,   0,   0,  -2,   0,   2 },    258.79860,   0.00002,   855.192 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,  -2,   1,   0 },    324.39407,   0.00002,     0.078 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   0,   0,   0 },    230.88636,   0.00005,     1.179 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,  -1,   0 },    305.78646,   0.00005,     0.094 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,   0,   0 },    306.92615,   0.00005,     0.042 },
 { {   0,   0,   5, -11,   1,   0,   0,   0,   0,   0,   0 },    216.97294,   0.00001,     1.309 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   0,   0 },    212.75052,   0.00006,     0.040 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   1,   0 },    212.49394,   0.00008,     0.085 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0 },    212.13522,   0.00017,     0.074 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   0,   0 },    214.34109,   0.00093,     3.156 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   1,   0 },    213.37875,   0.00019,     0.077 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   2,  -1,   0 },    221.37705,   0.00012,     0.090 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   2,   0,   0 },    219.00556,   0.00010,     0.041 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   2,   1,   0 },    219.67365,   0.00002,     0.027 },
 { {   0,   0,   5, -10,   2,   0,   0,   0,   0,   0,   0 },    273.49303,   0.00002,     6.746 },
 { {   0,   0,   5, -10,   3,   0,   0,   0,   0,   0,   0 },     39.88925,   0.00003,    15.643 },
 { {   0,   0,   5,  -9,  -2,   0,   0,   0,   0,  -1,   0 },    178.69326,   0.00002,     0.076 },
 { {   0,   0,   5,  -9,  -2,   0,   0,   0,   0,   0,   0 },    358.15307,   0.00001,    21.630 },
 { {   0,   0,   5,  -9,  -2,   0,   0,   0,   0,   1,   0 },    178.69326,   0.00002,     0.075 },
 { {   0,   0,   5,  -9,  -1,   0,   0,   0,   0,  -1,   0 },     24.30080,   0.00001,     0.076 },
 { {   0,   0,   5,  -9,  -1,   0,   0,   0,   0,   1,   0 },     24.30080,   0.00001,     0.075 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -4,   0,   0 },    243.82592,   0.00002,     0.020 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -4,   2,   0 },    243.70326,   0.00002,     0.044 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,  -1,   0 },    243.41253,   0.00010,     0.026 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,   0 },    243.74940,   0.00095,     0.041 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,   2 },    243.84146,   0.00002,     0.431 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,   0 },    244.04576,   0.00146,     0.089 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   2,   0 },    243.87032,   0.00006,     0.503 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -1,   0,   0 },     63.28656,   0.00001,     0.082 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -1,   2,   0 },     63.28656,   0.00001,     0.070 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0 },    240.43723,   0.00007,     0.038 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0 },    241.18261,   0.00130,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0 },    242.29669,   0.00676,     4.655 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0 },    240.95514,   0.00125,     0.074 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0 },    240.77241,   0.00008,     0.037 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   1,   0,  -1 },    324.74671,   0.00002,     1.192 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -2,   0 },    243.01247,   0.00005,     0.641 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,   0 },    243.52582,   0.00115,     0.085 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,  -2 },    243.15886,   0.00002,     0.528 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,   0 },    243.13692,   0.00068,     0.040 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   1,   0 },    242.90020,   0.00008,     0.026 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   4,  -2,   0 },    243.17054,   0.00001,     0.043 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   4,  -1,   0 },    243.21348,   0.00002,     0.027 },
 { {   0,   0,   5,  -9,   1,   0,   0,   0,   0,   0,   0 },    274.96725,   0.00001,     3.343 },
 { {   0,   0,   5,  -9,   2,   0,   0,   0,   0,   0,   0 },    302.11384,   0.00001,     2.608 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -4,   0,   0 },    271.97857,   0.00002,     0.021 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -4,   1,   0 },    271.99112,   0.00001,     0.028 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -4,   2,   0 },    272.08664,   0.00002,     0.045 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,  -1,   0 },    271.97805,   0.00010,     0.027 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   0 },    271.98380,   0.00110,     0.042 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   2 },    271.97491,   0.00002,     0.350 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   1,   0 },    272.00141,   0.00151,     0.093 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   2,   0 },    271.97450,   0.00005,     0.397 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,  -1,   1 },    174.29636,   0.00007,     0.087 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,   0,   0 },     92.81867,   0.00002,     0.086 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,   0,   1 },    354.31648,   0.00001,     0.555 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0 },    271.88281,   0.00006,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0 },    273.52474,   0.00092,     0.080 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0 },    272.05502,   0.00474,     1.340 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0 },    271.88587,   0.00080,     0.071 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0 },    271.85455,   0.00005,     0.037 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -3,   1 },    174.47222,   0.00001,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -2,   1 },    174.30157,   0.00022,    38.061 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   0 },    273.15148,   0.00003,     7.118 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   1 },    354.29617,   0.00034,     0.075 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   0,  -1 },    354.30935,   0.00017,     3.255 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   0,   0 },    272.81866,   0.00012,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -2,   0 },      2.25083,   0.00002,     0.973 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -1,   0 },    287.37653,   0.00066,     0.082 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,   0 },    276.30214,   0.00035,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   1,   0 },    273.05120,   0.00004,     0.026 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   4,  -3,   0 },     44.70941,   0.00001,     0.096 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   4,  -2,   0 },     44.97180,   0.00004,     0.042 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   4,  -1,   0 },     17.57901,   0.00002,     0.027 },
 { {   0,   0,   5,  -8,   1,   0,   0,   0,   0,   0,   0 },    308.63033,   0.00001,     1.204 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,  -2,   0,   0 },    174.66766,   0.00007,     0.042 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,  -2,   1,   0 },    174.78519,   0.00007,     0.095 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,   0,  -1,   0 },    172.88920,   0.00001,     0.082 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,   0,   0,   0 },    172.23896,   0.00010,     0.999 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,   0,   1,   0 },    172.88920,   0.00001,     0.070 },
 { {   0,   0,   5,  -8,   2,   0,   0,   0,   0,   0,   0 },    327.56132,   0.00002,     1.093 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -6,   1,   0 },    184.53121,   0.00001,     0.017 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,  -2,   0 },    184.42887,   0.00002,     0.013 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,  -1,   0 },    184.15184,   0.00003,     0.016 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   0,   0 },    184.36099,   0.00017,     0.021 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   1,   0 },    184.59433,   0.00040,     0.028 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   2,   0 },    184.95020,   0.00017,     0.046 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -3,   0,   0 },    184.68618,   0.00002,     0.028 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -3,   1,   0 },      4.68618,   0.00001,     0.044 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -3,   3,   0 },      4.68618,   0.00001,     0.273 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -2,   0 },    184.28617,   0.00011,     0.020 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    184.49564,   0.00088,     0.027 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -1,   2 },      4.96957,   0.00001,     0.099 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,  -2 },      5.09862,   0.00002,     0.020 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   0 },    184.81755,   0.00897,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   2 },    184.77332,   0.00014,     0.322 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   0 },    185.26257,   0.00962,     0.095 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   2 },      4.96957,   0.00001,     0.061 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   2,   0 },    185.80785,   0.00037,     0.361 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   3,   0 },    185.99934,   0.00003,     0.062 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -1,   0,   0 },      9.17382,   0.00004,     0.088 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -1,   1,   0 },      4.56870,   0.00001,     0.530 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -3,   0 },    183.91012,   0.00001,     0.026 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -2,   0 },    184.89408,   0.00031,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,   0 },    185.63585,   0.00353,     0.082 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0,   0 },    187.28372,   0.01091,     1.001 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,   0 },    191.80940,   0.00097,     0.070 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   2,   0 },    267.96509,   0.00001,     0.036 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   1,   0,   0 },    158.37729,   0.00004,     0.075 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -2,   0 },    180.86025,   0.00003,     1.291 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -1,   0 },      0.53934,   0.00040,     0.080 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   0,  -2 },      1.73539,   0.00004,     0.902 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   0,   0 },      1.02559,   0.00052,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   1,   0 },      1.67263,   0.00004,     0.026 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   2,   0 },      3.72668,   0.00002,     0.019 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   4,  -2,   0 },      3.75594,   0.00002,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   4,  -1,   0 },      3.31546,   0.00004,     0.027 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   4,   0,   0 },      6.12257,   0.00001,     0.020 },
 { {   0,   0,   5,  -8,   5,  -5,   0,   0,  -2,   0,   0 },    319.67278,   0.00002,     0.042 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -4,   0,   0 },    126.48274,   0.00002,     0.021 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -4,   1,   0 },    126.37873,   0.00002,     0.029 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -4,   2,   0 },    126.74498,   0.00002,     0.046 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,  -2,   0 },    126.60846,   0.00002,     0.020 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,  -1,   0 },    126.35822,   0.00011,     0.027 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   0 },    127.36827,   0.00141,     0.043 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   2 },    126.69205,   0.00002,     0.295 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,   0 },    126.21393,   0.00158,     0.098 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   2,   0 },    127.21115,   0.00007,     0.328 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -1,   0,   1 },    206.01899,   0.00001,     0.429 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0 },    124.55419,   0.00011,     0.040 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0 },    124.12575,   0.00159,     0.083 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0 },    123.47048,   0.00534,     0.782 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0 },    123.14434,   0.00086,     0.069 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0 },    123.26854,   0.00006,     0.036 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,  -1,  -1 },    204.44314,   0.00003,     0.077 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,  -1,   1 },     24.82448,   0.00002,     0.072 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   0,  -1 },    204.45995,   0.00026,     4.456 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   0,   1 },     24.42966,   0.00003,     0.037 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   1,  -1 },    204.44314,   0.00003,     0.074 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -3,   0 },    298.58475,   0.00001,     0.073 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -2,   0 },    304.13691,   0.00018,     2.017 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,   0 },    301.81663,   0.00178,     0.078 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,  -2 },    302.45037,   0.00002,     1.206 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,   0 },    302.94665,   0.00019,     0.038 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   1,   0 },    302.79273,   0.00002,     0.025 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   4,  -2,   0 },    301.35623,   0.00002,     0.041 },
 { {   0,   0,   5,  -7,   2,   0,   0,   0,   2,   0,   0 },      0.52397,   0.00001,     0.038 },
 { {   0,   0,   5,  -7,   3,   0,   0,   0,   2,  -1,   0 },    137.11477,   0.00002,     0.077 },
 { {   0,   0,   5,  -6,  -2,   0,   0,   0,   2,  -1,   0 },    274.03927,   0.00003,     0.076 },
 { {   0,   0,   5,  -6,  -1,   0,   0,   0,   2,  -3,   0 },    268.86865,   0.00001,     0.075 },
 { {   0,   0,   5,  -6,  -1,   0,   0,   0,   2,  -2,   0 },    301.35082,   0.00001,    20.650 },
 { {   0,   0,   5,  -6,  -1,   0,   0,   0,   2,  -1,   0 },    295.76891,   0.00003,     0.076 },
 { {   0,   0,   5,  -6,   0,  -1,   0,   0,   2,  -2,   0 },    269.30943,   0.00006,   518.400 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -4,   1,   0 },    165.92111,   0.00002,     0.029 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,  -2,   0 },    330.10195,   0.00001,     0.020 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,  -1,   0 },    235.33022,   0.00002,     0.028 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,   0 },    169.25784,   0.00073,     0.044 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   1,   0 },    166.76850,   0.00085,     0.103 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   2,   0 },    169.60530,   0.00003,     0.279 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -3,   0 },    331.85017,   0.00006,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0 },    331.45395,   0.00072,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,  -2 },    151.74184,   0.00002,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0 },    331.17373,   0.00838,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0 },    157.54097,   0.00195,     0.553 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0 },    156.02960,   0.00038,     0.066 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0 },    154.39915,   0.00003,     0.035 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,  -1,   0 },    151.74187,   0.00004,     1.084 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   0,  -1 },    237.30839,   0.00010,     1.323 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   0,   0 },    149.01946,   0.00002,     0.071 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -4,   0 },    332.79334,   0.00003,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -3,   0 },    332.76317,   0.00044,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,  -2 },    152.05447,   0.00001,     0.037 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,   0 },    331.68291,   0.01682,    27.873 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,   2 },    152.05447,   0.00001,     0.037 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,  -2 },    151.70597,   0.00009,     0.074 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   0 },    151.73240,   0.04282,     0.075 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   2 },    331.74180,   0.00009,     0.025 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,  -2 },    331.55923,   0.00006,     3.360 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,   0 },    151.78604,   0.00263,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,   2 },    331.74176,   0.00002,     0.019 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   1,  -2 },    151.50608,   0.00001,     0.077 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   1,   0 },    151.70539,   0.00016,     0.025 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   3,  -1,   0 },    331.74191,   0.00002,     0.039 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -3,   0 },    331.33561,   0.00028,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -2,   0 },    151.79152,   0.00038,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -1,   0 },    151.75356,   0.00037,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,   0,   0 },    151.73124,   0.00006,     0.019 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   6,  -2,   0 },    151.75346,   0.00001,     0.020 },
 { {   0,   0,   5,  -6,   1,   0,   0,   0,   0,   0,   0 },      5.38713,   0.00001,     0.528 },
 { {   0,   0,   5,  -6,   2,   0,   0,   0,   2,  -2,   0 },    209.02969,   0.00002,     4.890 },
 { {   0,   0,   5,  -5,  -1,   0,   0,   0,  -2,   0,   0 },    327.43867,   0.00001,     0.044 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,  -1,   0 },    137.14940,   0.00004,     0.028 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,   0 },    133.48680,   0.00024,     0.045 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   1,   0 },    136.28595,   0.00021,     0.109 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0 },      4.13794,   0.00002,     0.041 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0 },     19.19138,   0.00028,     0.092 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0 },    173.74343,   0.00128,     0.427 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0 },    175.78756,   0.00022,     0.064 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0 },    171.96334,   0.00002,     0.035 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   1,   0,  -1 },    282.35470,   0.00004,     0.777 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -2,   0 },    179.24188,   0.00005,     1.762 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,   0 },    179.30738,   0.00261,     0.072 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,  -2 },    180.70395,   0.00018,     4.273 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,   0 },      2.36053,   0.00020,     0.037 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   1,   0 },      0.38197,   0.00005,     0.025 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   4,  -2,   0 },    179.52680,   0.00004,     0.040 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   4,  -1,   0 },    178.07452,   0.00001,     0.026 },
 { {   0,   0,   5,  -5,   2,   0,   0,   0,   0,   0,   0 },     57.36918,   0.00001,     0.398 },
 { {   0,   0,   5,  -4,  -2,   0,   0,   0,  -2,   0,   0 },    149.14299,   0.00001,     0.045 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,  -1,   0 },     86.58657,   0.00002,     0.028 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   0,   0 },     88.53538,   0.00017,     0.046 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   1,   0 },     87.04942,   0.00013,     0.116 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   0 },    255.58031,   0.00001,     0.042 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0 },    249.70883,   0.00016,     0.096 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0 },     86.60589,   0.00008,     0.348 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0 },     67.96159,   0.00004,     0.062 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   1,   0,  -1 },    309.95016,   0.00004,     0.550 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -1,   0 },    358.54316,   0.00013,     0.070 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,   0,   0 },    206.47643,   0.00005,     0.036 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   3,  -1,  -1 },    309.94012,   0.00003,     0.075 },
 { {   0,   0,   5,  -3,  -1,   0,   0,   0,   0,   0,   0 },    197.92809,   0.00001,     0.301 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,  -2,   0,   0 },    327.61837,   0.00003,     0.047 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0 },    315.90909,   0.00001,     0.294 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   0,   0 },     50.80135,   0.00002,     0.048 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   2,  -2,   0 },    266.38515,   0.00002,     0.462 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   2,   0,   0 },    266.38515,   0.00002,     0.035 },
 { {   0,   0,   5,   0,  -8,   0,   0,   0,   2,   0,   0 },     30.77839,   0.00001,     0.034 },
 { {   0,   0,   5,   0,  -7,   0,   0,   0,  -2,   0,   0 },    195.14696,   0.00002,     0.049 },
 { {   0,   0,   5,   0,  -7,   0,   0,   0,   2,  -1,   0 },    200.50710,   0.00005,     0.063 },
 { {   0,   0,   5,   0,  -7,   0,   0,   0,   2,   0,   0 },     20.50710,   0.00005,     0.034 },
 { {   0,   0,   5,   0,  -7,   5,   0,   0,   2,  -2,   0 },    347.68989,   0.00001,     0.356 },
 { {   0,   0,   5,   0,  -7,   5,   0,   0,   2,   0,   0 },    167.68989,   0.00001,     0.034 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,  -2,   0,   0 },    182.56487,   0.00005,     0.049 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,  -2,   1,   0 },    182.66173,   0.00002,     0.143 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   0,  -1,   0 },     10.05168,   0.00002,     0.114 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   0,   0,   0 },    180.19734,   0.00001,     0.223 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   0,   1,   0 },    190.05169,   0.00001,     0.056 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,  -2,   0 },    190.05168,   0.00001,     0.368 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,  -1,   0 },    190.05168,   0.00021,     0.063 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,   0,   0 },      9.87338,   0.00012,     0.034 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,   1,   0 },     10.05168,   0.00003,     0.024 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,  -2,   0,   0 },    165.09018,   0.00014,     0.050 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,  -2,   1,   0 },    164.99957,   0.00006,     0.145 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,  -1,   0 },     20.46082,   0.00002,     0.115 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,   0,   0 },    139.95599,   0.00002,     0.218 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,   1,   0 },    174.42576,   0.00005,     0.056 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -2,   0 },    181.76136,   0.00001,     0.357 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -1,   0 },    178.99764,   0.00050,     0.062 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   0,  -2 },    179.14282,   0.00002,     0.405 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   0,   0 },    359.31534,   0.00032,     0.034 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   1,   0 },    359.10517,   0.00006,     0.023 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   4,  -1,   0 },    359.14281,   0.00003,     0.025 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,  -2,   0,   0 },     62.82266,   0.00011,     0.050 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,  -2,   1,   0 },     62.95072,   0.00007,     0.147 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,  -1,   0 },    246.25800,   0.00002,     0.116 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,   0,   0 },     65.41392,   0.00002,     0.214 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,   1,   0 },     58.17690,   0.00001,     0.056 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   2,  -1,   0 },     18.49793,   0.00006,     0.062 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   2,   0,   0 },    195.99402,   0.00006,     0.034 },
 { {   0,   0,   5,   0,  -3,  -5,   0,   0,   2,  -2,   0 },    187.92329,   0.00001,     0.357 },
 { {   0,   0,   5,   0,  -3,  -5,   0,   0,   2,   0,   0 },      7.92329,   0.00001,     0.034 },
 { {   0,   0,   5,   0,  -3,   0,   0,   0,  -2,   0,   0 },    319.22020,   0.00006,     0.050 },
 { {   0,   0,   5,   0,  -3,   0,   0,   0,  -2,   1,   0 },    315.20942,   0.00003,     0.148 },
 { {   0,   0,   5,   0,  -3,   0,   0,   0,   0,   0,   0 },    316.15638,   0.00001,     0.211 },
 { {   0,   0,   5,   0,  -3,   5,   0,   0,   2,   0,   0 },    298.87575,   0.00001,     0.034 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,  -2,   0,   0 },    228.39841,   0.00004,     0.050 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,  -2,   1,   0 },    227.71796,   0.00001,     0.150 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,   2,  -1,   0 },    248.04316,   0.00002,     0.061 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,   2,   0,   0 },    250.55612,   0.00005,     0.034 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   0,  -1,   0 },    309.53050,   0.00001,     0.120 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   0,   1,   0 },    309.53050,   0.00001,     0.055 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   2,  -3,   0 },    130.41171,   0.00001,     0.099 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   2,  -2,   0 },    309.97545,   0.00004,     0.318 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   2,   0,   0 },    311.02969,   0.00004,     0.034 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.15328,   0.00002,     0.051 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.15328,   0.00002,     0.148 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,  -2,   0 },    158.01872,   0.00001,     0.046 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0 },     14.86713,   0.00004,     0.200 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   2,   0 },     37.36581,   0.00001,     0.032 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -2,   0 },    197.95044,   0.00003,     0.310 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   4,   0,   0 },    215.58729,   0.00003,     0.018 },
 { {   0,   0,   5,   0,   1,  -5,   0,   0,   2,   0,   0 },    138.27989,   0.00001,     0.034 },
 { {   0,   0,   6, -16,   4,   5,   0,   0,   0,   0,   0 },     98.55429,   0.00003,     0.500 },
 { {   0,   0,   6, -16,   4,   5,   0,   0,   2,  -1,   0 },     98.55429,   0.00006,     0.105 },
 { {   0,   0,   6, -16,   4,   5,   0,   0,   2,   0,   0 },     98.55429,   0.00007,     0.044 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,   2,   0,   0 },      0.78529,   0.00002,     0.044 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,  -1,   0 },    274.74257,   0.00001,     0.071 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,   0,   0 },    270.37797,   0.00006,     1.097 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,   1,   0 },    274.74257,   0.00001,     0.081 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   2,  -1,   0 },    283.39219,   0.00006,     0.095 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   2,   0,   0 },    283.39219,   0.00005,     0.042 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,  -2,   0,   0 },    182.30484,   0.00001,     0.040 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,  -2,   1,   0 },    182.30747,   0.00002,     0.084 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,  -1,   0 },    184.20191,   0.00005,     0.073 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   0,   0 },    186.11080,   0.00028,     2.630 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   1,   0 },    184.20191,   0.00005,     0.078 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   2,  -1,   0 },    193.94016,   0.00003,     0.090 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   2,   0,   0 },    193.94016,   0.00003,     0.041 },
 { {   0,   0,   6, -12,   2,   0,   0,   0,   0,  -1,   0 },    229.81270,   0.00001,     0.074 },
 { {   0,   0,   6, -12,   2,   0,   0,   0,   0,   1,   0 },    229.81270,   0.00001,     0.077 },
 { {   0,   0,   6, -12,   3,   0,   0,   0,   0,   0,   0 },     46.09829,   0.00005,     7.856 },
 { {   0,   0,   6, -11,  -2,   0,   0,   0,   0,  -1,   0 },    332.61770,   0.00001,     0.075 },
 { {   0,   0,   6, -11,  -2,   0,   0,   0,   0,   0,   0 },    150.58295,   0.00001,    58.360 },
 { {   0,   0,   6, -11,  -2,   0,   0,   0,   0,   1,   0 },    332.61770,   0.00001,     0.076 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -4,   1,   0 },    214.83415,   0.00001,     0.028 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -4,   2,   0 },    214.85246,   0.00001,     0.044 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,  -1,   0 },    214.83766,   0.00007,     0.026 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   0 },    214.72072,   0.00064,     0.041 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   2 },    214.85233,   0.00001,     0.443 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   1,   0 },    215.26000,   0.00099,     0.088 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   2,   0 },    214.88208,   0.00004,     0.519 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0 },    212.79378,   0.00006,     0.038 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,   0 },    211.86189,   0.00096,     0.076 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   0,   0 },    213.13032,   0.00506,     6.602 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,   0 },    211.63600,   0.00092,     0.075 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   2,   0 },    211.02357,   0.00005,     0.038 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   1,   0,  -1 },    295.64939,   0.00001,     1.108 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -2,   0 },    214.11202,   0.00003,     0.616 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -1,   0 },    214.68275,   0.00082,     0.086 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,  -2 },    214.22894,   0.00001,     0.511 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,   0 },    214.16791,   0.00050,     0.040 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   1,   0 },    214.29315,   0.00005,     0.026 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   4,  -2,   0 },    214.27082,   0.00001,     0.043 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   4,  -1,   0 },    214.26988,   0.00002,     0.027 },
 { {   0,   0,   6, -11,   3,   0,   0,   0,   0,   0,   0 },     78.42278,   0.00002,     2.473 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -4,   1,   0 },    243.25477,   0.00001,     0.028 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,  -1,   0 },    243.25397,   0.00004,     0.027 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,   0 },    243.24841,   0.00039,     0.042 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   1,   0 },    243.32955,   0.00054,     0.093 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   2,   0 },    243.25575,   0.00003,     0.407 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -1,  -1,   1 },    324.93216,   0.00002,     0.087 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0 },    240.91973,   0.00001,     0.039 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,   0 },    242.38046,   0.00033,     0.080 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   0,   0 },    242.97277,   0.00175,     1.464 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,   0 },    242.32738,   0.00027,     0.072 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   2,   0 },    242.37240,   0.00002,     0.037 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,  -2,   1 },    144.93679,   0.00006,    26.959 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,  -1,   1 },    144.93216,   0.00011,     0.076 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,   0,  -1 },    324.92941,   0.00007,     2.698 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,   0,   0 },    241.12026,   0.00002,     0.077 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,  -1,   0 },    248.55139,   0.00031,     0.082 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,   0,   0 },    243.96720,   0.00017,     0.039 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,   1,   0 },    244.03823,   0.00001,     0.026 },
 { {   0,   0,   6, -10,   2,   0,   0,   0,   0,   0,   0 },    122.96145,   0.00001,     1.174 },
 { {   0,   0,   6, -10,   3,   0,   0,   0,  -2,   0,   0 },    107.85647,   0.00001,     0.042 },
 { {   0,   0,   6, -10,   3,   0,   0,   0,   0,   0,   0 },    108.28442,   0.00006,     1.068 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -4,   1,   0 },     97.48760,   0.00003,     0.029 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,  -1,   0 },     97.42020,   0.00008,     0.027 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,   0 },     98.09205,   0.00078,     0.043 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,   2 },     97.55550,   0.00001,     0.301 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   1,   0 },     97.21796,   0.00091,     0.097 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   2,   0 },     96.95490,   0.00005,     0.335 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0 },     94.97117,   0.00004,     0.040 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0 },     95.27586,   0.00087,     0.083 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0 },     94.52828,   0.00297,     0.823 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0 },     94.54497,   0.00049,     0.069 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   2,   0 },     94.35962,   0.00003,     0.036 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,  -2,   0 },    271.77986,   0.00001,     0.077 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,  -1,  -1 },    174.34205,   0.00003,     0.076 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,  -1 },    174.36543,   0.00021,     6.208 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,   0 },     91.77986,   0.00001,     0.074 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,   1 },    354.33538,   0.00002,     0.037 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   1,  -2 },     94.31568,   0.00004,   176.039 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   1,  -1 },    174.34205,   0.00003,     0.075 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -2,   0 },    274.27676,   0.00006,     1.788 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,   0 },    272.60679,   0.00076,     0.079 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,   0,   0 },    275.10754,   0.00006,     0.039 },
 { {   0,   0,   6,  -8,  -2,   0,   0,   0,   0,   0,   0 },    244.02891,   0.00001,     0.634 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -4,   1,   0 },    132.65166,   0.00001,     0.029 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,  -1,   0 },    126.49048,   0.00010,     0.028 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,   0 },    131.57355,   0.00050,     0.043 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   1,   0 },    133.35712,   0.00039,     0.103 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   2,   0 },    137.24451,   0.00002,     0.284 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0 },    122.70987,   0.00021,     0.040 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0 },    122.62930,   0.00664,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0 },    124.51760,   0.00206,     0.573 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0 },    125.54360,   0.00031,     0.067 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0 },    127.04114,   0.00002,     0.035 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,  -1,   0 },    302.33645,   0.00003,     1.163 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,  -1 },    205.29451,   0.00009,     1.444 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,   0 },    122.24097,   0.00002,     0.071 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -4,   0 },    301.50396,   0.00005,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -3,   0 },    301.49163,   0.00067,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,  -2 },    122.33781,   0.00002,     0.037 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   0 },    302.35646,   0.02023,    36.376 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   2 },    122.33781,   0.00002,     0.037 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,  -2 },    302.33625,   0.00006,     0.073 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   0 },    302.34346,   0.03317,     0.076 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   2 },    122.33633,   0.00006,     0.025 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,  -2 },    302.50480,   0.00003,     2.770 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,   0 },    302.30181,   0.00243,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   1,   0 },    302.36876,   0.00018,     0.025 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -3,   0 },    302.72996,   0.00031,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -2,   0 },    302.35524,   0.00066,     0.040 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -1,   0 },    302.34702,   0.00034,     0.026 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,   0,   0 },    302.25641,   0.00003,     0.020 },
 { {   0,   0,   6,  -8,   0,   1,   0,   0,   2,  -2,   0 },    164.89208,   0.00002,   154.875 },
 { {   0,   0,   6,  -8,   1,   0,   0,   0,   0,   0,   0 },    356.85537,   0.00001,     0.546 },
 { {   0,   0,   6,  -8,   1,   0,   0,   0,   2,  -1,   0 },    288.17658,   0.00003,     0.075 },
 { {   0,   0,   6,  -8,   2,   0,   0,   0,   2,   0,   0 },    180.28671,   0.00001,     0.038 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -4,   1,   0 },     82.36849,   0.00003,     0.029 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,  -1,   0 },     82.67342,   0.00006,     0.028 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   0,   0 },     82.38089,   0.00054,     0.044 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   1,   0 },     82.82372,   0.00046,     0.105 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   2,   0 },     82.64769,   0.00002,     0.265 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,  -1,   0 },     84.18299,   0.00011,     0.089 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,   0,   0 },     84.49691,   0.00024,     0.500 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   2,  -3,   0 },    112.22020,   0.00001,     0.077 },
 { {   0,   0,   6,  -7,  -6,   0,   0,   0,   2,  -2,   0 },    194.08565,   0.00002,   612.237 },
 { {   0,   0,   6,  -7,  -2,   0,   0,   0,   2,  -1,  -2 },    272.41906,   0.00002,     0.075 },
 { {   0,   0,   6,  -7,  -2,   0,   0,   0,   2,   0,  -2 },    274.94253,   0.00010,   483.014 },
 { {   0,   0,   6,  -7,  -2,   0,   0,   0,   2,   1,  -2 },    272.41906,   0.00002,     0.075 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,  -1,   0 },    121.14559,   0.00003,     0.028 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,   0 },    115.83856,   0.00018,     0.045 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   1,   0 },    120.29311,   0.00017,     0.109 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0 },    333.39489,   0.00002,     0.041 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0 },    340.22583,   0.00030,     0.091 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0 },    146.13823,   0.00103,     0.439 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0 },    146.48406,   0.00021,     0.064 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0 },    145.72593,   0.00002,     0.035 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   1,   0,  -1 },    239.70707,   0.00006,     0.817 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -3,   0 },    154.81416,   0.00002,     0.078 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -2,   0 },    151.12720,   0.00003,     1.983 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,  -2 },    151.53437,   0.00001,     0.076 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,   0 },    150.75692,   0.00264,     0.073 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,  -2 },    151.53437,   0.00025,     5.859 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,   0 },    332.79125,   0.00017,     0.037 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   1,  -2 },    151.53437,   0.00001,     0.074 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   1,   0 },    331.35103,   0.00005,     0.025 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   4,  -2,   0 },    150.73103,   0.00004,     0.040 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,  -1,   0 },    149.65697,   0.00002,     0.028 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   0,   0 },    150.74381,   0.00012,     0.046 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   1,   0 },    136.40112,   0.00007,     0.115 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0 },    249.70853,   0.00003,     0.042 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0 },    231.97904,   0.00036,     0.096 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0 },    179.26855,   0.00063,     0.356 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0 },    173.09875,   0.00011,     0.062 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   1,   0,  -1 },    283.25603,   0.00003,     0.569 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -2,   0 },    202.11220,   0.00006,     0.965 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,   0 },    187.68271,   0.00119,     0.070 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,  -2 },    180.60929,   0.00004,     1.424 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,   0 },    353.50009,   0.00016,     0.036 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   1,   0 },      0.54597,   0.00004,     0.025 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   3,  -2,  -1 },    103.24638,   0.00004,    55.903 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   3,  -1,  -1 },    103.22061,   0.00005,     0.076 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   4,  -2,   0 },    179.88116,   0.00002,     0.039 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   4,  -1,   0 },      0.50711,   0.00001,     0.026 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,  -2,   0,   0 },     41.15904,   0.00005,     0.047 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,  -2,   1,   0 },     41.00893,   0.00002,     0.123 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0 },    236.59217,   0.00004,     0.101 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0 },     53.39052,   0.00002,     0.299 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0 },     34.11417,   0.00002,     0.060 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   1,   0,  -1 },    309.73176,   0.00002,     0.437 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,  -1,   0 },     30.13399,   0.00014,     0.067 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,   0,   0 },    208.65293,   0.00006,     0.036 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,  -2,   0,   0 },    327.93194,   0.00002,     0.048 },
 { {   0,   0,   6,   0,  -8,   0,   0,   0,   2,  -1,   0 },    200.24811,   0.00002,     0.059 },
 { {   0,   0,   6,   0,  -8,   0,   0,   0,   2,   0,   0 },     20.24811,   0.00002,     0.033 },
 { {   0,   0,   6,   0,  -7,   0,   0,   0,   0,   0,   0 },     10.05180,   0.00001,     0.185 },
 { {   0,   0,   6,   0,  -7,   0,   0,   0,   2,  -1,   0 },    189.58455,   0.00004,     0.059 },
 { {   0,   0,   6,   0,  -7,   0,   0,   0,   2,   0,   0 },      9.58455,   0.00004,     0.033 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,  -2,   0,   0 },    162.22589,   0.00002,     0.052 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   0,   0,   0 },      5.27556,   0.00002,     0.182 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,  -1,   0 },    178.11715,   0.00010,     0.059 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,   0,   0 },    358.11714,   0.00009,     0.033 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,   1,   0 },    358.11714,   0.00001,     0.023 },
 { {   0,   0,   6,   0,  -5,   0,   0,   0,  -2,   0,   0 },     59.89367,   0.00003,     0.052 },
 { {   0,   0,   6,   0,  -5,   0,   0,   0,   2,  -1,   0 },     23.28472,   0.00002,     0.059 },
 { {   0,   0,   6,   0,  -5,   0,   0,   0,   2,   0,   0 },    203.28534,   0.00002,     0.033 },
 { {   0,   0,   6,   0,  -2,   0,   0,   0,   0,   0,   0 },    308.35719,   0.00001,     0.171 },
 { {   0,   0,   6,   0,  -2,   0,   0,   0,   0,   2,   0 },    308.35719,   0.00001,     0.031 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,  -2,   0 },    322.32527,   0.00001,     0.237 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,  -1,   0 },    326.76203,   0.00002,     0.057 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,   0,   0 },    322.32527,   0.00001,     0.033 },
 { {   0,   0,   7, -17,   0,  -2,   0,   0,  -2,   0,   2 },    356.30915,   0.00004,   997.332 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,  -1,   0 },    175.61736,   0.00002,     0.026 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   0,   0 },    175.61736,   0.00002,     0.039 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   1,   0 },    175.61736,   0.00007,     0.080 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,  -1,   0 },    355.61736,   0.00020,     0.070 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   0,   0 },    355.61736,   0.00120,     1.000 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   1,   0 },    355.61736,   0.00027,     0.082 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -2,   0 },    355.61736,   0.00006,     0.360 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -1,   0 },    355.61736,   0.00115,     0.095 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   0,  -2 },    355.61736,   0.00001,     0.322 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   0,   0 },    355.61736,   0.00087,     0.042 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   1,   0 },    355.61736,   0.00011,     0.027 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   4,  -1,   0 },    355.61736,   0.00004,     0.028 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   4,   0,   0 },    355.61736,   0.00002,     0.021 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -2,   1,   0 },     78.45120,   0.00003,     0.080 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,  -1,   0 },    251.87076,   0.00001,     0.070 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   0,   0 },    256.02994,   0.00030,     1.025 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    254.79616,   0.00011,     0.081 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,  -1,   0 },    258.30474,   0.00028,     0.095 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   0,   0 },    257.80420,   0.00024,     0.042 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   1,   0 },    258.03064,   0.00003,     0.027 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,   0,   0,   0 },    158.05011,   0.00006,     2.254 },
 { {   0,   0,   7, -14,   1,   0,   0,   0,   0,   0,   0 },    186.56147,   0.00002,     2.783 },
 { {   0,   0,   7, -14,   3,   0,   0,   0,   0,   0,   0 },     17.05075,   0.00001,     5.245 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,  -2,   0,   0 },    116.21464,   0.00006,     0.040 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,  -2,   1,   0 },    116.67861,   0.00007,     0.087 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   0,  -1,   0 },    115.10657,   0.00013,     0.075 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   0,   0,   0 },    117.13709,   0.00055,   263.130 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   0,   1,   0 },    115.10657,   0.00013,     0.075 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   2,  -1,   0 },    116.67220,   0.00007,     0.087 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   2,   0,   0 },    116.20735,   0.00005,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -4,   1,   0 },    185.84609,   0.00003,     0.028 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -4,   2,   0 },    185.86143,   0.00001,     0.044 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,  -1,   0 },    185.84606,   0.00006,     0.026 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   0 },    185.69215,   0.00063,     0.041 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   2 },    185.84611,   0.00001,     0.455 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,   0 },    186.33582,   0.00098,     0.088 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   2,   0 },    185.87533,   0.00004,     0.537 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0 },    182.60034,   0.00007,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -1,   0 },    182.80550,   0.00107,     0.076 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,  -2 },      4.19360,   0.00001,     0.037 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   0 },    184.17328,   0.00558,    11.350 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   2 },      4.19360,   0.00001,     0.037 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   1,   0 },    182.66224,   0.00105,     0.075 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   2,   0 },    182.51179,   0.00007,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -2,   0 },    185.34571,   0.00004,     0.593 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -1,   0 },    185.80209,   0.00087,     0.086 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,  -2 },    185.38594,   0.00001,     0.495 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,   0 },    185.24565,   0.00055,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   1,   0 },    185.40037,   0.00005,     0.026 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   4,  -1,   0 },    185.38594,   0.00002,     0.028 },
 { {   0,   0,   7, -13,   1,   0,   0,   0,   0,   0,   0 },    220.09764,   0.00001,     5.800 },
 { {   0,   0,   7, -13,   3,   0,   0,   0,   0,   0,   0 },     49.52584,   0.00002,     2.932 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,  -1,   0 },    214.45568,   0.00002,     0.027 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   0,   0 },    214.41870,   0.00012,     0.041 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   1,   0 },    214.48787,   0.00020,     0.092 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,  -1,   0 },    213.56722,   0.00012,     0.079 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   0,   0 },    213.58112,   0.00066,     1.613 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   1,   0 },    213.54772,   0.00010,     0.072 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   1,   0,  -1 },    295.83579,   0.00002,     2.304 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   2,  -1,   0 },    215.06712,   0.00014,     0.083 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   2,   0,   0 },    214.81266,   0.00008,     0.039 },
 { {   0,   0,   7, -12,   3,   0,   0,   0,   0,   0,   0 },     78.43443,   0.00003,     1.146 },
 { {   0,   0,   7, -11,  -2,   0,   0,   0,   0,   0,   0 },    184.73943,   0.00001,     1.017 },
 { {   0,   0,   7, -11,  -1,   0,   0,   0,   1,   0,  -1 },    266.12183,   0.00001,    74.609 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -4,   1,   0 },     68.97653,   0.00002,     0.029 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,  -1,   0 },     68.94893,   0.00005,     0.027 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   0,   0 },     68.89240,   0.00047,     0.042 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   1,   0 },     68.61706,   0.00056,     0.097 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   2,   0 },     68.93320,   0.00003,     0.342 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0 },     66.10739,   0.00002,     0.039 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,   0 },     66.43955,   0.00049,     0.083 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   0,   0 },     66.05873,   0.00179,     0.868 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,   0 },     65.84482,   0.00030,     0.069 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   2,   0 },     66.74295,   0.00002,     0.036 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,  -1,  -1 },    144.33765,   0.00002,     0.076 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,  -1 },    144.60955,   0.00017,    10.235 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,   1 },    324.57725,   0.00002,     0.037 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   1,  -1 },    144.33765,   0.00002,     0.075 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,  -2,   0 },    247.54797,   0.00003,     1.606 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,  -1,   0 },    245.09640,   0.00029,     0.079 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,   0,   0 },    248.94981,   0.00002,     0.039 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,  -1,   0 },    101.11083,   0.00002,     0.028 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   0,   0 },    101.35731,   0.00016,     0.043 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   1,   0 },    102.11860,   0.00016,     0.102 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0 },     93.00024,   0.00005,     0.040 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -1,   0 },     93.77435,   0.00118,     0.086 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   0,   0 },     95.10740,   0.00080,     0.594 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   1,   0 },     96.15972,   0.00013,     0.067 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   1,   0,  -1 },    175.37810,   0.00007,     1.589 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -3,   0 },    273.49440,   0.00004,     0.075 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -2,   0 },    273.32950,   0.00111,    11.006 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,   0 },    273.15224,   0.00507,     0.076 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,  -2 },    273.47081,   0.00002,     2.356 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,   0 },    273.14583,   0.00043,     0.038 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   1,   0 },    273.24610,   0.00004,     0.025 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   4,  -3,   0 },    273.24207,   0.00002,     0.088 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   4,  -2,   0 },    273.15997,   0.00008,     0.041 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   4,  -1,   0 },    273.12664,   0.00005,     0.026 },
 { {   0,   0,   7, -10,   1,   0,   0,   0,   2,  -2,   0 },    124.26595,   0.00023,   152.550 },
 { {   0,   0,   7, -10,   1,   0,   0,   0,   2,  -1,   0 },    142.65434,   0.00008,     0.075 },
 { {   0,   0,   7, -10,   2,   0,   0,   0,   2,  -2,   0 },    331.20298,   0.00001,    12.862 },
 { {   0,   0,   7, -10,   2,   0,   0,   0,   2,   0,   0 },    151.20298,   0.00001,     0.038 },
 { {   0,   0,   7,  -9,  -2,   0,   0,   0,   0,   0,   0 },    221.75863,   0.00001,     0.489 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,  -2,   0 },     88.83785,   0.00001,     0.020 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   0,   0 },     92.45058,   0.00011,     0.044 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   1,   0 },     85.13168,   0.00009,     0.108 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   0 },    303.81596,   0.00002,     0.041 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0 },    306.95728,   0.00022,     0.091 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0 },    116.85078,   0.00059,     0.452 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0 },    114.65008,   0.00012,     0.065 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0 },    119.41381,   0.00002,     0.035 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   1,   0,  -1 },    206.60647,   0.00005,     0.861 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -2,   0 },    122.38397,   0.00002,     2.269 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,  -2 },    122.49052,   0.00002,     0.076 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,   0 },    121.76370,   0.00186,     0.073 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,  -2 },    122.49052,   0.00030,     9.319 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,   0 },    303.46146,   0.00011,     0.037 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   1,  -2 },    122.49052,   0.00002,     0.075 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   1,   0 },    302.35066,   0.00003,     0.025 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   4,  -2,   0 },    121.64116,   0.00003,     0.040 },
 { {   0,   0,   7,  -9,   2,   0,   0,   0,   2,  -2,   0 },    180.07526,   0.00001,     1.641 },
 { {   0,   0,   7,  -9,   2,   0,   0,   0,   2,   0,   0 },    180.07526,   0.00001,     0.037 },
 { {   0,   0,   7,  -8,  -2,   0,   0,   0,   0,   0,   0 },    273.10708,   0.00001,     0.388 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,  -1,   0 },    135.90354,   0.00002,     0.028 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   0,   0 },    136.55136,   0.00010,     0.045 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   1,   0 },    129.53296,   0.00007,     0.114 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   0 },    350.89078,   0.00001,     0.042 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0 },     10.71796,   0.00034,     0.095 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0 },    147.41646,   0.00055,     0.364 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0 },    148.58967,   0.00013,     0.062 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0 },    149.45525,   0.00001,     0.034 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   1,   0,  -1 },    241.33582,   0.00003,     0.591 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -2,   0 },    106.34533,   0.00002,     1.028 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -1,   0 },    144.77064,   0.00103,     0.070 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,  -2 },    151.32808,   0.00004,     1.565 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,   0 },    336.19107,   0.00020,     0.036 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   1,   0 },    331.41052,   0.00004,     0.025 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   4,  -2,   0 },    151.21284,   0.00002,     0.039 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   4,  -1,   0 },    331.43484,   0.00001,     0.026 },
 { {   0,   0,   7,  -8,   3,   0,   0,   0,  -2,   0,   0 },    339.86251,   0.00003,     0.046 },
 { {   0,   0,   7,  -8,   3,   0,   0,   0,  -2,   1,   0 },    339.84781,   0.00001,     0.118 },
 { {   0,   0,   7,  -8,   3,   0,   0,   0,   3,  -1,  -2 },    338.09533,   0.00002,   204.109 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,  -2,   0,   0 },    167.71851,   0.00006,     0.047 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0 },    292.33579,   0.00002,     0.100 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0 },    178.63939,   0.00032,     0.305 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0 },    176.36074,   0.00006,     0.060 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   1,   0,  -1 },    283.88445,   0.00002,     0.450 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -2,   0 },    180.40739,   0.00002,     0.665 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,   0 },    181.25212,   0.00050,     0.068 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,  -2 },    180.41764,   0.00002,     0.854 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,   0 },      0.85409,   0.00013,     0.036 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   1,   0 },      0.41049,   0.00003,     0.024 },
 { {   0,   0,   7,  -6,  -2,   0,   0,   0,   0,   0,   0 },    330.61425,   0.00002,     0.275 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   0,   0 },     43.28949,   0.00003,     0.048 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   1,   0 },     35.61937,   0.00001,     0.130 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0 },    181.78258,   0.00002,     0.106 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0 },     40.16200,   0.00002,     0.262 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0 },     29.83486,   0.00001,     0.059 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   1,   0,  -1 },    309.35216,   0.00001,     0.363 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -1,   0 },     29.84516,   0.00010,     0.065 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,   0,   0 },    209.85072,   0.00006,     0.035 },
 { {   0,   0,   7,   0,  -8,   0,   0,   0,   2,   0,   0 },      8.93110,   0.00002,     0.032 },
 { {   0,   0,   7,   0,  -7,   0,   0,   0,   2,  -1,   0 },    176.69842,   0.00003,     0.056 },
 { {   0,   0,   7,   0,  -7,   0,   0,   0,   2,   0,   0 },    356.69842,   0.00003,     0.032 },
 { {   0,   0,   8, -19,   0,   0,   0,   0,  -2,   0,   2 },    202.69618,   0.00003,   180.816 },
 { {   0,   0,   8, -19,   0,   3,   0,   0,   2,   0,   0 },    277.16115,   0.00003,     0.044 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,  -1,   0,   1 },    175.19440,   0.00005,    66.080 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   0,  -1,   0 },    234.30832,   0.00001,     0.070 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   0,   0,   0 },    234.34277,   0.00006,     0.963 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   0,   1,   0 },    234.30832,   0.00001,     0.082 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   2,  -1,   0 },    232.43473,   0.00005,     0.096 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   2,   0,   0 },    232.43473,   0.00004,     0.042 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,   0,   0,   0 },    121.39490,   0.00002,     1.973 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,  -1,   0 },    252.68044,   0.00002,     0.016 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   0,   0 },    252.68044,   0.00011,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   1,   0 },    252.68044,   0.00029,     0.028 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   2,   0 },    252.68044,   0.00020,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -3,   0,   0 },    252.68044,   0.00002,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -3,   3,   0 },     72.68044,   0.00001,     0.376 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -2,   0 },    252.68044,   0.00007,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,   0 },    252.68044,   0.00068,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,   2 },     72.68044,   0.00001,     0.090 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,  -2 },     72.68044,   0.00002,     0.019 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,   0 },    252.68044,   0.00865,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,   2 },    252.68044,   0.00021,     0.474 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,  -2 },     72.68044,   0.00003,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,   0 },    252.68044,   0.01709,     0.087 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   2,   0 },    252.68044,   0.00083,     0.564 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   3,   0 },    252.68044,   0.00006,     0.067 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -1,  -1,   0 },     72.68044,   0.00002,     0.039 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -1,   0,   0 },     72.68044,   0.00023,     0.081 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -1,   1,   0 },     72.68044,   0.00005,     1.127 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -3,   0 },     72.68044,   0.00002,     0.025 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -2,   0 },     72.68044,   0.00015,     0.038 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -1,   0 },     72.68044,   0.00199,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   0,   0 },     72.68044,   0.00188, 94316.280 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   1,   0 },     72.68044,   0.00199,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   2,   0 },     72.68044,   0.00015,     0.038 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   3,   0 },     72.68044,   0.00002,     0.025 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   1,  -1,   0 },     72.68044,   0.00005,     1.127 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   1,   0,   0 },     72.68044,   0.00023,     0.081 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   1,   1,   0 },     72.68044,   0.00002,     0.039 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -3,   0 },    252.68044,   0.00006,     0.067 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -2,   0 },    252.68044,   0.00083,     0.564 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,   0 },    252.68044,   0.01709,     0.087 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,   2 },     72.68044,   0.00003,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,  -2 },    252.68044,   0.00021,     0.474 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,   0 },    252.68044,   0.00865,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,   2 },     72.68044,   0.00002,     0.019 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,  -2 },     72.68044,   0.00001,     0.090 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,   0 },    252.68044,   0.00068,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   2,   0 },    252.68044,   0.00007,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   3,  -3,   0 },     72.68044,   0.00001,     0.376 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   3,   0,   0 },    252.68044,   0.00002,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -2,   0 },    252.68044,   0.00020,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -1,   0 },    252.68044,   0.00029,     0.028 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,   0,   0 },    252.68044,   0.00011,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,   1,   0 },    252.68044,   0.00002,     0.016 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,  -2,   0,   0 },    261.02476,   0.00007,     0.040 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,  -2,   1,   0 },    261.02476,   0.00012,     0.087 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,   0,   1,   0 },     81.02476,   0.00002,     0.075 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,   2,  -1,   0 },    261.02476,   0.00012,     0.087 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,   2,   0,   0 },    261.02476,   0.00007,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   0,   0 },    156.84007,   0.00003,     0.020 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   1,   0 },    156.84157,   0.00008,     0.028 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   2,   0 },    157.63371,   0.00004,     0.044 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -2,   0 },    156.83421,   0.00003,     0.020 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -1,   0 },    156.21153,   0.00026,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   0 },    156.70847,   0.00268,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   2 },    156.80777,   0.00005,     0.469 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,   0 },    157.36080,   0.00414,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   2,   0 },    158.41343,   0.00016,     0.556 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   3,   0 },    156.84541,   0.00001,     0.066 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -1,   0,   0 },    336.59225,   0.00010,     0.081 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -1,   1,   0 },    336.80777,   0.00001,     1.097 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -3,   0 },    152.80515,   0.00002,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0 },    153.80753,   0.00035,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,  -2 },    334.48396,   0.00002,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   0 },    153.77074,   0.00503,     0.076 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,  -2 },    336.81895,   0.00001,     0.037 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   0 },    155.30914,   0.02611,    40.427 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   2 },    336.81895,   0.00001,     0.037 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   0 },    153.72987,   0.00501,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   2 },    334.48396,   0.00002,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   2,   0 },    153.75276,   0.00034,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   3,   0 },    152.76463,   0.00002,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   1,  -1,   0 },    336.64584,   0.00001,     1.160 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   1,   0,   0 },    336.62054,   0.00007,     0.081 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -3,   0 },    156.63725,   0.00001,     0.067 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -2,   0 },    158.19709,   0.00015,     0.572 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,   0 },    157.22211,   0.00398,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,  -2 },    156.64584,   0.00006,     0.480 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,   0 },    156.53915,   0.00257,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   1,   0 },    156.06070,   0.00025,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   2,   0 },    156.67228,   0.00003,     0.020 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,  -2,   0 },    157.36888,   0.00005,     0.044 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,  -1,   0 },    156.69269,   0.00007,     0.028 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,   0,   0 },    156.67682,   0.00003,     0.020 },
 { {   0,   0,   8, -15,   2,   0,   0,   0,   0,  -1,   0 },     18.95348,   0.00001,     0.077 },
 { {   0,   0,   8, -15,   2,   0,   0,   0,   0,   1,   0 },     18.95348,   0.00001,     0.074 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   0,   0 },    185.61422,   0.00006,     0.041 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   1,   0 },    185.61429,   0.00009,     0.092 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0 },    182.67280,   0.00005,     0.079 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   0,   0 },    184.33493,   0.00029,     1.797 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   1,   0 },    182.67280,   0.00005,     0.072 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   2,  -1,   0 },    184.99505,   0.00007,     0.083 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   2,   0,   0 },    186.15995,   0.00004,     0.040 },
 { {   0,   0,   8, -14,   3,   0,   0,   0,   0,   0,   0 },     49.41348,   0.00001,     1.236 },
 { {   0,   0,   8, -13,   0,  -1,   0,   0,   1,  -1,  -1 },    140.83140,   0.00002,     0.075 },
 { {   0,   0,   8, -13,   0,  -1,   0,   0,   1,   0,  -1 },    173.13169,   0.00031,  2556.886 },
 { {   0,   0,   8, -13,   0,  -1,   0,   0,   1,   1,  -1 },    140.83140,   0.00002,     0.075 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -4,   1,   0 },     41.71831,   0.00001,     0.028 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,  -1,   0 },     41.68119,   0.00004,     0.027 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0 },     41.59314,   0.00035,     0.042 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   1,   0 },     41.86703,   0.00043,     0.096 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   2,   0 },     41.78255,   0.00002,     0.349 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0 },     38.30027,   0.00001,     0.039 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0 },     38.59660,   0.00032,     0.082 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   0,   0 },     37.64922,   0.00131,     0.919 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   1,   0 },     37.19069,   0.00020,     0.070 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,  -1,  -1 },    112.41175,   0.00002,     0.076 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   0,  -1 },    113.54244,   0.00018,    29.122 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   1,  -1 },    112.41175,   0.00002,     0.075 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,  -1,   0 },    216.29657,   0.00009,     0.080 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,   0,   0 },     37.63577,   0.00002,     0.039 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   0,   0 },     70.66549,   0.00008,     0.043 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   1,   0 },     71.22857,   0.00008,     0.101 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0 },     64.23246,   0.00002,     0.040 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,  -1,   0 },     64.80209,   0.00036,     0.086 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   0,   0 },     65.75956,   0.00036,     0.617 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   1,   0 },     66.93573,   0.00008,     0.067 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   1,   0,  -1 },    145.71911,   0.00004,     1.767 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -2,   0 },    244.82365,   0.00021,     6.484 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -1,   0 },    244.22044,   0.00142,     0.076 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,   0,   0 },    244.30365,   0.00015,     0.038 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   4,  -2,   0 },    244.13369,   0.00002,     0.041 },
 { {   0,   0,   8, -12,   2,   0,   0,   0,   2,  -2,   0 },    314.29913,   0.00004,    69.540 },
 { {   0,   0,   8, -12,   2,   0,   0,   0,   2,  -1,   0 },    129.28969,   0.00007,     0.075 },
 { {   0,   0,   8, -11,  -2,   0,   0,   0,   0,   0,   0 },    231.57264,   0.00001,     0.504 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   0,   0 },     52.26274,   0.00006,     0.044 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   1,   0 },     41.95414,   0.00004,     0.107 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,  -2,   0 },    272.43808,   0.00002,     0.041 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,  -1,   0 },    275.93330,   0.00017,     0.090 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   0,   0 },     86.94818,   0.00029,     0.465 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   1,   0 },     84.66270,   0.00006,     0.065 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   1,   0,  -1 },    175.95574,   0.00004,     0.911 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,  -2 },     93.55982,   0.00007,     0.076 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,   0 },     92.89823,   0.00116,     0.073 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,  -2 },     93.46844,   0.00065,    22.758 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,   0 },    274.11589,   0.00010,     0.037 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   1,  -2 },     93.55982,   0.00007,     0.075 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   1,   0 },    273.45393,   0.00001,     0.025 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   4,  -2,   0 },     92.73755,   0.00002,     0.040 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   0,   0 },    110.65126,   0.00007,     0.045 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   1,   0 },    103.42546,   0.00004,     0.114 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,  -2,   0 },    302.19589,   0.00001,     0.042 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0 },    323.65073,   0.00010,     0.095 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   0,   0 },    120.12038,   0.00034,     0.373 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   1,   0 },    119.81466,   0.00007,     0.063 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   1,   0,  -1 },    207.57264,   0.00003,     0.614 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -2,   0 },    118.51936,   0.00002,     1.100 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -1,   0 },    121.12253,   0.00078,     0.071 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,  -2 },    122.44212,   0.00003,     1.737 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,   0 },    302.48086,   0.00011,     0.036 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   1,   0 },    302.42019,   0.00003,     0.025 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   3,  -3,   0 },    302.78038,   0.00002,    45.185 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   3,  -2,   0 },    122.78039,   0.00003,     0.075 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   4,  -2,   0 },    122.33816,   0.00001,     0.039 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   4,   0,   0 },    122.44143,   0.00002,     0.019 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   0,   0 },    145.26843,   0.00006,     0.046 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   1,   0 },    144.66724,   0.00002,     0.121 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0 },    293.12405,   0.00002,     0.100 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0 },    149.55915,   0.00032,     0.311 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0 },    148.51052,   0.00006,     0.061 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   1,   0,  -1 },    242.90220,   0.00002,     0.463 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -2,   0 },    151.38245,   0.00002,     0.694 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,   0 },    151.40369,   0.00059,     0.068 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,  -2 },    151.38342,   0.00002,     0.903 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,   0 },    330.97177,   0.00016,     0.036 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   1,   0 },    331.20911,   0.00003,     0.024 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   4,  -2,   0 },    149.93422,   0.00001,     0.038 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   4,  -1,   0 },    331.37264,   0.00001,     0.025 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,  -2,   0,   0 },    159.66933,   0.00002,     0.048 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0 },    359.63061,   0.00003,     0.105 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0 },    177.89354,   0.00017,     0.267 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0 },    180.05446,   0.00001,     0.059 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   1,   0,  -1 },    284.29374,   0.00001,     0.371 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -2,   0 },    180.06132,   0.00001,     0.507 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -1,   0 },    179.80085,   0.00025,     0.066 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,   0 },      0.10057,   0.00011,     0.035 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   1,   0 },      0.06300,   0.00003,     0.024 },
 { {   0,   0,   8,  -7,  -2,   0,   0,   0,   0,   0,   0 },    330.15148,   0.00001,     0.243 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,  -2,   0,   0 },     45.81431,   0.00001,     0.049 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0 },     48.98582,   0.00002,     0.234 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0 },     30.56951,   0.00001,     0.057 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,  -1,   0 },     30.57776,   0.00007,     0.063 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,   0,   0 },    210.58259,   0.00005,     0.034 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,  -2,   0 },    174.22432,   0.00001,     0.036 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,  -1,   0 },    174.22409,   0.00003,     0.070 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   0,   0 },    174.22417,   0.00023,     1.000 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   1,   0 },    174.22412,   0.00005,     0.082 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,  -1,   0 },    174.22422,   0.00024,     0.095 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,   0,   0 },    174.22421,   0.00019,     0.042 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,   1,   0 },    174.22422,   0.00002,     0.027 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -4,   1,   0 },    127.72004,   0.00001,     0.028 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,  -1,   0 },    127.72003,   0.00004,     0.026 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,   0 },    127.55873,   0.00038,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   1,   0 },    128.20785,   0.00058,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   2,   0 },    127.69730,   0.00002,     0.576 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -2,   0 },    125.28546,   0.00006,     0.038 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,   0 },    125.08655,   0.00077,     0.075 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   0,   0 },    126.49116,   0.00407,    25.886 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,   0 },    125.17405,   0.00078,     0.076 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   2,   0 },    125.28842,   0.00006,     0.038 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -2,   0 },    128.05447,   0.00003,     0.552 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,   0 },    128.56203,   0.00061,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,   0 },    127.86486,   0.00040,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   1,   0 },    128.03477,   0.00004,     0.026 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   4,  -1,   0 },    128.03477,   0.00001,     0.028 },
 { {   0,   0,   9, -17,   1,   0,   0,   0,   0,   0,   0 },    173.43137,   0.00001,    21.895 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   0,   0 },    156.74971,   0.00003,     0.041 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   1,   0 },    156.74982,   0.00002,     0.091 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,  -1,   0 },    156.61556,   0.00002,     0.078 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   0,   0 },    154.68833,   0.00012,     2.028 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   1,   0 },    156.61556,   0.00002,     0.073 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   2,  -1,   0 },    155.26305,   0.00002,     0.084 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   2,   0,   0 },    156.37565,   0.00002,     0.040 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -4,   0,   0 },    149.74351,   0.00003,     0.021 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -4,   1,   0 },    149.74351,   0.00003,     0.028 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -4,   2,   0 },    149.74351,   0.00002,     0.046 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,  -1,   0 },    149.74351,   0.00010,     0.027 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   0,   0 },    149.74351,   0.00087,     0.042 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   0,   2 },    149.74351,   0.00001,     0.322 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   1,   0 },    149.74351,   0.00114,     0.095 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   2,   0 },    149.74351,   0.00006,     0.360 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,  -1,   0 },    149.74351,   0.00025,     0.082 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   0,   0 },    149.74351,   0.00121,     1.000 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   1,   0 },    149.74351,   0.00020,     0.070 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,  -2,   0 },    149.74351,   0.00001,     1.292 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,  -1,   0 },    329.74351,   0.00006,     0.080 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,   0,   0 },    329.74351,   0.00004,     0.039 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,   1,   0 },    329.74351,   0.00002,     0.026 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -4,   1,   0 },     21.75648,   0.00002,     0.028 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,  -1,   0 },     21.45906,   0.00007,     0.027 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,   0 },     23.47673,   0.00067,     0.042 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   1,   0 },     21.62820,   0.00081,     0.096 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   2,   0 },     22.44109,   0.00004,     0.357 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,  -2,   0 },     13.57675,   0.00002,     0.039 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,  -1,   0 },     15.10261,   0.00052,     0.082 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   0,   0 },     12.34844,   0.00209,     0.976 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   1,   0 },      9.71422,   0.00031,     0.070 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   2,   0 },      9.09427,   0.00001,     0.036 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,  -1,  -1 },    268.08074,   0.00002,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,  -1 },    267.80135,   0.00017,    34.449 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,   0 },    191.62814,   0.00004,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,   1 },    267.67828,   0.00004,     0.037 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   1,  -1 },    268.08074,   0.00002,     0.076 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,  -1,   0 },    334.76744,   0.00002,     0.080 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,   0,   0 },    355.08701,   0.00004,     0.039 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,  -2,   0,   0 },     40.52623,   0.00004,     0.043 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,  -2,   1,   0 },     40.52617,   0.00004,     0.101 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,  -1,   0 },     36.37737,   0.00012,     0.085 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,   0,   0 },     37.32408,   0.00014,     0.643 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,   1,   0 },     37.69541,   0.00003,     0.068 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   1,   0,  -1 },    116.33953,   0.00002,     1.989 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -2,   0 },    215.10341,   0.00005,     4.596 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -1,   0 },    215.87951,   0.00044,     0.077 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,   0,   0 },    215.71385,   0.00006,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,  -2,   0,   0 },    359.04974,   0.00003,     0.044 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,  -1,   0 },    245.57216,   0.00012,     0.090 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   0,  -2 },     64.77251,   0.00001,     0.040 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   0,   0 },     53.91719,   0.00012,     0.479 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   1,  -2 },    244.54592,   0.00002,     0.087 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   1,   0 },     57.26965,   0.00003,     0.065 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   1,   0,  -1 },    146.11804,   0.00002,     0.967 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -2,  -2 },     64.61893,   0.00001,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,  -2 },     64.65635,   0.00021,     0.075 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,   0 },     63.89922,   0.00068,     0.074 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,  -2 },     64.77251,   0.00095,    51.478 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,   0 },     64.20504,   0.00007,     0.037 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   1,  -2 },     64.65635,   0.00021,     0.076 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   2,  -2 },     64.61893,   0.00001,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   4,  -1,  -2 },    244.54592,   0.00002,     0.087 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   4,   0,  -2 },     64.77251,   0.00001,     0.040 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,  -2,   0,   0 },     81.84470,   0.00004,     0.045 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,  -2,   1,   0 },     81.74187,   0.00001,     0.113 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,  -1,   0 },    273.47142,   0.00007,     0.094 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,   0,   0 },     91.27758,   0.00019,     0.382 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,   1,   0 },     93.32587,   0.00003,     0.063 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   1,   0,  -1 },    176.57251,   0.00002,     0.639 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -2,   0 },     93.41766,   0.00001,     1.182 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,   0 },     93.42675,   0.00048,     0.071 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,  -2 },     93.43257,   0.00002,     1.952 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,   0 },    273.41799,   0.00007,     0.037 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   1,   0 },    273.42567,   0.00001,     0.025 },
 { {   0,   0,   9, -12,   2,   0,   0,   0,   2,  -3,   0 },    150.92973,   0.00001,     0.082 },
 { {   0,   0,   9, -12,   2,   0,   0,   0,   2,  -1,   0 },    330.92973,   0.00001,     0.070 },
 { {   0,   0,   9, -11,  -2,   0,   0,   0,   0,   0,   0 },    231.91089,   0.00001,     0.335 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,  -2,   0,   0 },    118.45055,   0.00005,     0.046 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0 },    303.69418,   0.00004,     0.099 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   0,   0 },    120.64349,   0.00022,     0.317 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   1,   0 },    121.21491,   0.00004,     0.061 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   1,   0,  -1 },    208.56607,   0.00002,     0.477 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -2,   0 },    122.37913,   0.00002,     0.726 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -1,   0 },    122.37679,   0.00044,     0.068 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,  -2 },    122.37970,   0.00001,     0.958 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,   0 },    301.61336,   0.00011,     0.036 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   1,   0 },    302.13487,   0.00002,     0.024 },
 { {   0,   0,   9, -10,  -2,   0,   0,   0,   0,   0,   0 },    272.36020,   0.00001,     0.285 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,  -2,   0,   0 },    144.46567,   0.00005,     0.047 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0 },    334.88969,   0.00001,     0.104 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   0,   0 },    149.73293,   0.00018,     0.272 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   1,   0 },    147.49130,   0.00003,     0.059 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   1,   0,  -1 },    244.37326,   0.00001,     0.380 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -2,   0 },    151.22352,   0.00002,     0.524 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,   0 },    151.20494,   0.00034,     0.066 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,  -2 },    151.22685,   0.00001,     0.635 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,   0 },    332.74377,   0.00013,     0.035 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   1,   0 },    331.68431,   0.00003,     0.024 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   4,   0,   0 },    151.22677,   0.00002,     0.019 },
 { {   0,   0,   9,  -9,  -2,   0,   0,   0,   0,   0,   0 },    301.21841,   0.00001,     0.247 },
 { {   0,   0,   9,  -9,  -1,   0,   0,   0,   0,   0,   0 },    146.96548,   0.00001,     0.242 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,  -2,   0,   0 },    154.12175,   0.00002,     0.049 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0 },    359.46552,   0.00001,     0.111 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0 },    175.65625,   0.00008,     0.237 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,  -1,   0 },    179.46994,   0.00012,     0.064 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,   0,   0 },    359.46827,   0.00008,     0.035 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,   1,  -4 },    186.67087,   0.00002,     0.087 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,  -1,  -4 },      4.41056,   0.00008,     0.075 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,   0,  -4 },    192.80477,   0.00053,  4535.679 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,   1,  -4 },      4.41056,   0.00008,     0.075 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   6,  -1,  -4 },    186.67087,   0.00002,     0.087 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,  -1,   0 },     31.02646,   0.00005,     0.062 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,   0,   0 },    211.03081,   0.00004,     0.034 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,   0,   0 },     59.71903,   0.00001,     0.033 },
 { {   0,   0,  10, -21,   2,   0,   0,   0,   0,   0,   0 },    230.14217,   0.00001,     1.003 },
 { {   0,   0,  10, -21,   2,   0,   0,   0,   2,   0,   0 },    230.96460,   0.00001,     0.042 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   0,   0 },     98.56202,   0.00002,     0.040 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   1,   0 },     98.56209,   0.00003,     0.086 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,  -1,   0 },     98.30797,   0.00004,     0.075 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   0,   0 },     98.02006,   0.00023,     9.804 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   1,   0 },     98.30797,   0.00004,     0.076 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   2,  -1,   0 },     99.59175,   0.00003,     0.088 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   2,   0,   0 },     99.59175,   0.00002,     0.041 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   0,   0 },     71.28735,   0.00003,     0.020 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   1,   0 },     71.28735,   0.00006,     0.028 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   2,   0 },     71.28735,   0.00005,     0.044 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,  -2,   0 },     71.28735,   0.00002,     0.020 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,  -1,   0 },     71.28735,   0.00013,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   0 },     71.28735,   0.00173,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   2 },     71.28735,   0.00005,     0.475 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,   0 },     71.28735,   0.00340,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   2,   0 },     71.28736,   0.00016,     0.564 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -1,   0,   0 },    251.28739,   0.00006,     0.081 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,  -2,   0 },    251.28725,   0.00003,     0.038 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,  -1,   0 },    251.28724,   0.00039,     0.075 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   0,   0 },    251.28731,   0.00038,  6408.190 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   1,   0 },    251.28738,   0.00039,     0.075 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   2,   0 },    251.28737,   0.00003,     0.038 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   1,   0,   0 },    251.28723,   0.00006,     0.081 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -2,   0 },     71.28727,   0.00016,     0.564 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,   0 },     71.28727,   0.00340,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,  -2 },     71.28727,   0.00005,     0.474 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,   0 },     71.28727,   0.00173,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   1,   0 },     71.28727,   0.00013,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   2,   0 },     71.28727,   0.00002,     0.020 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,  -2,   0 },     71.28727,   0.00005,     0.044 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,  -1,   0 },     71.28727,   0.00006,     0.028 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,   0,   0 },     71.28727,   0.00003,     0.020 },
 { {   0,   0,  10, -19,   1,   0,   0,   0,   0,   0,   0 },    288.81662,   0.00001,    56.510 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,  -2,   0,   0 },    127.80631,   0.00001,     0.041 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,   0,   0,   0 },    125.55748,   0.00006,     2.327 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,  -1,   0 },    142.20072,   0.00002,     0.027 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   0,   0 },    142.20068,   0.00010,     0.042 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   1,   0 },    141.24243,   0.00012,     0.095 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,  -1,   0 },    150.06601,   0.00011,     0.081 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   0,   0 },    152.28527,   0.00048,     1.040 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   1,   0 },    152.90587,   0.00008,     0.070 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   1,   0,  -1 },    238.09991,   0.00003,    10.823 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,  -3,   0 },    151.68212,   0.00001,     0.071 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,  -1,   0 },    163.22964,   0.00003,     0.080 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,   0,   0 },    160.54441,   0.00002,     0.039 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,  -2,   0,   0 },     10.77068,   0.00002,     0.043 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   0,  -1,   0 },      6.62790,   0.00003,     0.085 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   0,   0,   0 },      8.22504,   0.00007,     0.670 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   1,   0,  -1 },     78.89407,   0.00001,     2.276 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,  -2,   0 },    183.71144,   0.00001,     3.559 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,  -1,   0 },    186.62790,   0.00014,     0.077 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,   0,   0 },    186.62423,   0.00001,     0.038 },
 { {   0,   0,  10, -16,   4,   5,   0,   0,  -2,   0,   0 },     46.80658,   0.00008,     0.044 },
 { {   0,   0,  10, -16,   4,   5,   0,   0,  -2,   1,   0 },     46.80658,   0.00006,     0.105 },
 { {   0,   0,  10, -16,   4,   5,   0,   0,   0,   0,   0 },     46.80658,   0.00003,     0.500 },
 { {   0,   0,  10, -15,  -3,   0,   0,   0,   2,  -2,   0 },    166.86752,   0.00002,   455.769 },
 { {   0,   0,  10, -15,  -3,   0,   0,   0,   2,  -1,   0 },    193.06535,   0.00003,     0.075 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,  -2,   0,   0 },    289.18732,   0.00006,     0.044 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,  -2,   1,   0 },    289.14305,   0.00005,     0.106 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   0,  -1,   0 },    221.17729,   0.00007,     0.089 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   0,   0,   0 },    348.42267,   0.00005,     0.494 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   0,   1,   0 },    343.88518,   0.00001,     0.065 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   1,   0,  -1 },    116.63175,   0.00001,     1.030 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,   0 },     32.26885,   0.00038,     0.074 },
 { {   0,   0,  10, -15,   1,   0,   0,   0,   2,  -1,  -2 },    248.47226,   0.00004,     0.075 },
 { {   0,   0,  10, -15,   1,   0,   0,   0,   2,   0,  -2 },    245.97058,   0.00021,   662.430 },
 { {   0,   0,  10, -15,   1,   0,   0,   0,   2,   1,  -2 },    248.47226,   0.00004,     0.075 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,  -2,   0,   0 },     53.31792,   0.00002,     0.045 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,  -1,   0 },    244.42744,   0.00004,     0.093 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,   0,   0 },     62.75576,   0.00008,     0.391 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,   1,   0 },     64.47172,   0.00002,     0.063 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   1,   0,  -1 },    146.54604,   0.00001,     0.666 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -2,   0 },     64.45828,   0.00001,     1.278 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -1,   0 },     64.44371,   0.00026,     0.071 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,  -2 },     64.43925,   0.00001,     2.228 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,   0 },    244.43822,   0.00004,     0.037 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,  -2,   0,   0 },     91.53431,   0.00002,     0.046 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,  -1,   0 },    273.45032,   0.00003,     0.098 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,   0,   0 },     92.53091,   0.00011,     0.324 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,   1,   0 },     93.27498,   0.00002,     0.061 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   1,   0,  -1 },    177.21592,   0.00001,     0.492 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,  -2,   0 },     93.38324,   0.00001,     0.761 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,  -1,   0 },     93.37540,   0.00029,     0.069 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   0,   0 },    273.36840,   0.00008,     0.036 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   1,   0 },    273.37580,   0.00001,     0.024 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   0,   0 },    119.03409,   0.00003,     0.047 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,  -1,   0 },    303.35612,   0.00003,     0.104 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,   0,   0 },    120.37440,   0.00014,     0.276 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,   1,   0 },    121.96409,   0.00002,     0.059 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   1,   0,  -1 },    209.57110,   0.00001,     0.390 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -2,   0 },    122.27897,   0.00002,     0.542 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,   0 },    122.72965,   0.00027,     0.066 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,   0,   0 },    302.23105,   0.00010,     0.035 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,   1,   0 },    302.27929,   0.00003,     0.024 },
 { {   0,   0,  10, -11,  -2,   0,   0,   0,   0,   0,   0 },    272.09263,   0.00002,     0.251 },
 { {   0,   0,  10, -11,  -1,   0,   0,   0,   0,   0,   0 },    118.34455,   0.00001,     0.246 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,  -2,   0,   0 },    139.44620,   0.00002,     0.049 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0 },    333.31814,   0.00002,     0.110 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,   0,   0 },    148.86752,   0.00011,     0.241 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,   1,   0 },    150.45439,   0.00002,     0.057 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -2,   0 },    150.99562,   0.00001,     0.421 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -1,   0 },    151.83232,   0.00021,     0.064 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,   0,   0 },    330.92789,   0.00009,     0.035 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,   1,   0 },    330.99628,   0.00003,     0.024 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,  -1,   0 },    178.48945,   0.00001,     0.117 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,   0,   0 },    175.36727,   0.00003,     0.214 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,   1,   0 },    178.48945,   0.00001,     0.056 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   2,  -1,   0 },    178.48573,   0.00006,     0.062 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   2,   0,   0 },    358.48340,   0.00005,     0.034 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,  -1,   0 },     31.27144,   0.00004,     0.060 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,   0,   0 },    211.27546,   0.00003,     0.033 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,  -1,   1 },      8.02677,   0.00001,   720.555 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,   0,   0 },     66.34313,   0.00003,     6.047 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   0,   0 },    128.13641,   0.00011,     0.040 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   1,   0 },    129.48710,   0.00016,     0.087 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,  -2,   0 },    122.89095,   0.00001,     0.038 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,  -1,   0 },    123.52087,   0.00024,     0.075 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,   0,   0 },    126.34025,   0.00118,   308.674 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,   1,   0 },    123.51184,   0.00024,     0.075 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,   2,   0 },    122.89260,   0.00001,     0.038 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,  -1,   0 },    129.45563,   0.00016,     0.087 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,   0,   0 },    128.10754,   0.00011,     0.040 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,   0,   0 },     97.36771,   0.00003,     2.730 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   0,   0 },    123.83680,   0.00002,     0.042 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   1,   0 },    123.83684,   0.00002,     0.094 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,   0,   0,   0 },    126.92581,   0.00008,     1.114 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,   1,  -1,   0 },    129.03348,   0.00004,    90.987 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,  -1,   0 },    328.35041,   0.00003,     0.027 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,   0,   0 },    328.35041,   0.00017,     0.042 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,   1,   0 },    328.35041,   0.00023,     0.095 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,  -1,   0 },    328.35049,   0.00006,     0.082 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   0,   0 },    328.35045,   0.00024,     1.000 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   1,   0 },    328.35052,   0.00003,     0.070 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,  -2,   0,   0 },    341.28303,   0.00002,     0.043 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   0,   0,   0 },    339.51754,   0.00003,     0.699 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -1,   0 },    157.70399,   0.00005,     0.077 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   0,   0 },     30.04053,   0.00002,     0.044 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   0,  -1,   0 },    186.34927,   0.00005,     0.089 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   0,   0,   0 },     19.23934,   0.00004,     0.510 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,  -1,   0 },      6.34925,   0.00023,     0.074 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,   0,   0 },    186.56133,   0.00002,     0.037 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   0,  -1,   0 },    215.40756,   0.00002,     0.093 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   0,   0,   0 },     33.99889,   0.00004,     0.401 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   0,   1,   0 },     35.59992,   0.00001,     0.063 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,  -1,   0 },     35.47411,   0.00013,     0.072 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,   0 },    215.46833,   0.00004,     0.037 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,  -2,   0,   0 },     77.65965,   0.00002,     0.046 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   0,  -1,   0 },    244.39263,   0.00002,     0.098 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   0,   0,   0 },     66.62762,   0.00005,     0.331 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   0,   1,   0 },     64.39907,   0.00002,     0.061 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,  -1,   0 },     63.83504,   0.00015,     0.069 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,   0,   0 },    244.39198,   0.00005,     0.036 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,  -2,   0,   0 },     84.18739,   0.00003,     0.047 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   0,  -1,   0 },    273.41644,   0.00002,     0.103 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   0,   0,   0 },     91.67783,   0.00007,     0.281 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   0,   1,   0 },     86.37342,   0.00001,     0.059 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   1,   0,  -1 },    177.87898,   0.00001,     0.400 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -1,   0 },     93.29828,   0.00018,     0.066 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,   0,   0 },    273.29381,   0.00009,     0.035 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   4,  -4,   0 },    273.30915,   0.00001,   119.249 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,  -2,   0,   0 },    114.90080,   0.00003,     0.048 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   0,  -1,   0 },    302.13583,   0.00002,     0.109 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   0,   0,   0 },    120.06066,   0.00009,     0.245 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   0,   1,   0 },    122.13537,   0.00001,     0.058 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,  -2,   0 },    122.13616,   0.00001,     0.432 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,  -1,   0 },    122.13653,   0.00017,     0.064 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,   0,   0 },    302.17338,   0.00008,     0.035 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,   1,   0 },    302.13651,   0.00002,     0.024 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,  -2,   0,   0 },    139.86382,   0.00002,     0.050 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0 },    316.56513,   0.00001,     0.116 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   0,   0,   0 },    147.87437,   0.00006,     0.216 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,  -1,   0 },    150.67874,   0.00012,     0.062 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   0,   0 },    330.67810,   0.00009,     0.034 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   1,   0 },    330.67868,   0.00002,     0.023 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   0,   0,   0 },    168.79978,   0.00001,     0.194 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   2,  -1,   0 },    176.74877,   0.00003,     0.060 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   2,   0,   0 },    356.74689,   0.00004,     0.033 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,  -1,   0 },     29.14403,   0.00003,     0.058 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,   0,   0 },    211.35220,   0.00004,     0.033 },
 { {   0,   0,  12, -22,   0,   0,   0,   0,   0,   0,   0 },     68.37807,   0.00002,     3.301 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,   0,   0,   0 },     98.11281,   0.00002,     1.198 },
 { {   0,   0,  12, -21,   2,   0,   0,   0,  -2,   0,   0 },     12.62488,   0.00002,     0.042 },
 { {   0,   0,  12, -21,   2,   0,   0,   0,   0,   0,   0 },    357.91970,   0.00002,     0.997 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,   0,   0,   0 },    310.07391,   0.00001,     0.732 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,   2,  -1,   0 },    124.83364,   0.00002,     0.078 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,   0,  -1,   0 },    157.62406,   0.00003,     0.088 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,   2,  -2,   0 },    157.58496,   0.00001,     8.066 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,   2,  -1,   0 },    337.62406,   0.00015,     0.075 },
 { {   0,   0,  12, -19,   0,   3,   0,   0,  -2,   0,   0 },    225.41347,   0.00001,     0.044 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   0,   0,   0 },      7.76890,   0.00001,     0.412 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -1,   0 },      5.68777,   0.00008,     0.072 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,   0,   0 },    186.49731,   0.00002,     0.037 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   0,   0,   0 },     39.46090,   0.00002,     0.338 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   0,   1,   0 },     31.91638,   0.00001,     0.062 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,  -1,   0 },     35.42400,   0.00008,     0.069 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,   0,   0 },    215.41917,   0.00004,     0.036 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   0,  -1,   0 },    244.35172,   0.00001,     0.102 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   0,   0,   0 },     63.68756,   0.00003,     0.286 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   0,   1,   0 },     64.31469,   0.00001,     0.060 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -1,   0 },     64.33077,   0.00010,     0.067 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,   0,   0 },    244.32681,   0.00005,     0.035 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   0,  -1,   0 },    273.35912,   0.00001,     0.108 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   0,   0,   0 },     91.60928,   0.00003,     0.248 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   0,   1,   0 },     84.07559,   0.00001,     0.058 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -1,   0 },     93.19258,   0.00012,     0.064 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,   0,   0 },    274.50599,   0.00007,     0.035 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   0,  -1,   0 },    301.94564,   0.00001,     0.115 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   0,   0,   0 },    119.89293,   0.00005,     0.219 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,  -1,   0 },    121.94657,   0.00011,     0.062 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   0,   0 },    301.94624,   0.00009,     0.034 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   1,   0 },    301.94654,   0.00002,     0.024 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   0,   0,   0 },    145.86204,   0.00002,     0.197 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,  -1,   0 },    150.25383,   0.00008,     0.060 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,   0,   0 },    330.25303,   0.00007,     0.034 },
 { {   0,   0,  12, -12,   0,   0,   0,   0,   2,   0,   0 },    353.13841,   0.00002,     0.033 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,  -1,   0 },     31.32731,   0.00002,     0.057 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,   0,   0 },    211.27923,   0.00003,     0.032 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   0,  -1,   0 },    128.26363,   0.00003,     0.088 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -2,   0 },    128.67324,   0.00003,    16.498 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -1,   0 },    308.26367,   0.00013,     0.075 },
 { {   0,   0,  13, -20,   0,   0,   0,   0,   2,  -1,   0 },    336.06123,   0.00004,     0.072 },
 { {   0,   0,  13, -20,   0,   0,   0,   0,   2,   0,   0 },    157.52495,   0.00002,     0.037 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,   2,  -1,   0 },      6.45732,   0.00005,     0.070 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,   2,   0,   0 },    186.43785,   0.00002,     0.036 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   0,   0,   0 },     35.52489,   0.00001,     0.292 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,  -1,   0 },     35.36342,   0.00006,     0.067 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,   0,   0 },    215.35917,   0.00003,     0.036 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   0,   0,   0 },     63.31407,   0.00002,     0.252 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   0,   1,   0 },     64.21390,   0.00001,     0.058 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,  -1,   0 },     64.24512,   0.00007,     0.065 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,   0,   0 },    244.24135,   0.00004,     0.035 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   0,   0,   0 },     91.07986,   0.00002,     0.223 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   0,   1,   0 },    106.64265,   0.00002,     0.056 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,  -1,   0 },     93.06896,   0.00008,     0.063 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,   0,   0 },    273.06875,   0.00006,     0.034 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   0,   0,   0 },    118.46435,   0.00002,     0.199 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,  -1,   0 },    121.70374,   0.00008,     0.061 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,   0,   0 },    301.70335,   0.00006,     0.034 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   0,   0,   0 },    144.20260,   0.00001,     0.180 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,  -1,   0 },    149.69165,   0.00004,     0.059 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,   0,   0 },    329.69071,   0.00004,     0.033 },
 { {   0,   0,  13, -12,   0,   0,   0,   0,   2,  -1,   0 },     31.05255,   0.00002,     0.055 },
 { {   0,   0,  13, -12,   0,   0,   0,   0,   2,   0,   0 },    211.05376,   0.00002,     0.032 },
 { {   0,   0,  13, -11,   0,   0,   0,   0,   2,   0,   0 },     63.90512,   0.00002,     0.031 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -3,   0 },     98.66413,   0.00001,     0.026 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -2,   0 },     98.62704,   0.00007,     0.040 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -1,   0 },    278.67299,   0.00015,     0.087 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,   0,   0 },    278.66972,   0.00002,     0.565 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -4,   0 },     97.85932,   0.00002,     0.038 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -3,   0 },     97.85776,   0.00025,     0.075 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,  -2 },    278.66413,   0.00002,     0.037 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   0 },     98.66413,   0.00800,   362.998 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   2 },    278.66413,   0.00002,     0.037 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,   0 },     98.66819,   0.00162,     0.075 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,   0,   0 },     98.66815,   0.00011,     0.038 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -3,   0 },     98.98698,   0.00013,     0.087 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -2,   0 },     98.68906,   0.00011,     0.040 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -1,   0 },     98.66413,   0.00001,     0.026 },
 { {   0,   0,  14, -23,   4,   0,   0,   0,   2,   0,  -2 },     70.28412,   0.00007,  1053.903 },
 { {   0,   0,  14, -22,   0,   0,   0,   0,   2,  -1,   0 },    306.34888,   0.00002,     0.073 },
 { {   0,   0,  14, -21,   0,   0,   0,   0,   2,  -1,   0 },    337.49906,   0.00002,     0.070 },
 { {   0,   0,  14, -21,   0,   0,   0,   0,   2,   0,   0 },    157.47037,   0.00002,     0.036 },
 { {   0,   0,  14, -21,   0,   0,   0,   0,   3,  -2,  -1 },    238.32840,   0.00002,   146.030 },
 { {   0,   0,  14, -20,   0,   0,   0,   0,   2,  -1,   0 },      6.39697,   0.00003,     0.067 },
 { {   0,   0,  14, -20,   0,   0,   0,   0,   2,   0,   0 },    186.38192,   0.00002,     0.036 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   0,   0,   0 },     35.22775,   0.00002,     0.257 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,  -1,   0 },     35.28847,   0.00004,     0.065 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,   0,   0 },    215.28486,   0.00003,     0.035 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   0,   0,   0 },     62.86314,   0.00001,     0.226 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,  -1,   0 },     66.10269,   0.00005,     0.063 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,   0,   0 },    245.76130,   0.00005,     0.034 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   0,  -1,   0 },     92.89913,   0.00001,     0.121 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   0,   0,   0 },     90.42494,   0.00001,     0.202 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   0,   1,   0 },     92.89913,   0.00001,     0.055 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,  -1,   0 },     92.89878,   0.00005,     0.061 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,   0,   0 },    272.89855,   0.00005,     0.034 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   0,   0,   0 },    117.40803,   0.00001,     0.182 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   2,  -1,   0 },    121.40080,   0.00004,     0.059 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   2,   0,   0 },    301.40036,   0.00005,     0.033 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,   2,  -1,   0 },    149.66142,   0.00002,     0.057 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,   2,   0,   0 },    328.94480,   0.00004,     0.033 },
 { {   0,   0,  14, -13,   0,   0,   0,   0,   2,   0,   0 },    210.66032,   0.00001,     0.031 },
 { {   0,   0,  15, -25,   0,   0,   0,   0,   2,  -1,   0 },     67.30316,   0.00002,     0.076 },
 { {   0,   0,  15, -22,   0,   0,   0,   0,   2,   0,   0 },    157.41296,   0.00002,     0.036 },
 { {   0,   0,  15, -21,   0,   0,   0,   0,   2,  -1,   0 },      6.31568,   0.00002,     0.065 },
 { {   0,   0,  15, -21,   0,   0,   0,   0,   2,   0,   0 },    186.31565,   0.00002,     0.035 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,  -1,   0 },     35.19363,   0.00003,     0.063 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,   0,   0 },    215.19386,   0.00003,     0.034 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,  -1,   0 },     64.01040,   0.00003,     0.061 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,   0,   0 },    244.01032,   0.00004,     0.034 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   2,  -1,   0 },     92.69090,   0.00004,     0.059 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   2,   0,   0 },    272.69076,   0.00005,     0.033 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   2,  -1,   0 },    121.02830,   0.00003,     0.057 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   2,   0,   0 },    301.02779,   0.00003,     0.033 },
 { {   0,   0,  15, -16,   0,   0,   0,   0,   2,  -1,   0 },    147.93713,   0.00002,     0.056 },
 { {   0,   0,  15, -16,   0,   0,   0,   0,   2,   0,   0 },    327.93662,   0.00003,     0.032 },
 { {   0,   0,  15, -14,   0,   0,   0,   0,   2,   0,   0 },    206.56097,   0.00001,     0.031 },
 { {   0,   0,  16, -23,   0,   0,   0,   0,   2,   0,   0 },    157.34897,   0.00002,     0.035 },
 { {   0,   0,  16, -22,   0,   0,   0,   0,   2,  -1,   0 },      6.23663,   0.00002,     0.063 },
 { {   0,   0,  16, -22,   0,   0,   0,   0,   2,   0,   0 },    186.23660,   0.00001,     0.034 },
 { {   0,   0,  16, -21,   0,   0,   0,   0,   2,  -1,   0 },     35.08558,   0.00003,     0.061 },
 { {   0,   0,  16, -21,   0,   0,   0,   0,   2,   0,   0 },    215.08553,   0.00002,     0.034 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,  -1,   0 },     63.85112,   0.00003,     0.059 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,   0,   0 },    243.85170,   0.00004,     0.033 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   0,   0,   0 },     92.44158,   0.00002,     0.170 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   2,  -1,   0 },     92.44013,   0.00003,     0.058 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   2,   0,   0 },    272.44141,   0.00004,     0.033 },
 { {   0,   0,  16, -18,   0,   0,   0,   0,   2,  -1,   0 },    120.56953,   0.00002,     0.056 },
 { {   0,   0,  16, -18,   0,   0,   0,   0,   2,   0,   0 },    300.57319,   0.00004,     0.032 },
 { {   0,   0,  16, -17,   0,   0,   0,   0,   2,   0,   0 },    326.52740,   0.00002,     0.032 },
 { {   0,   0,  16, -15,   0,   0,   0,   0,   2,   0,   0 },    209.20669,   0.00002,     0.031 },
 { {   0,   0,  16,  -5,  -1,   0,   0,   0,   0,  -1,   0 },    272.09758,   0.00002,   658.208 },
 { {   0,   0,  17, -32,   0,   0,   0,   0,   0,   0,   0 },    282.42285,   0.00004,    71.966 },
 { {   0,   0,  17, -30,   0,   0,   0,   0,   1,   0,  -1 },    243.79378,   0.00003,   232.951 },
 { {   0,   0,  17, -28,   0,   0,   0,   0,   2,   0,  -2 },    221.72626,   0.00005,   188.324 },
 { {   0,   0,  17, -24,   0,   0,   0,   0,   2,   0,   0 },    157.27578,   0.00002,     0.035 },
 { {   0,   0,  17, -23,   0,   0,   0,   0,   2,   0,   0 },    186.14222,   0.00002,     0.034 },
 { {   0,   0,  17, -22,   0,   0,   0,   0,   2,  -1,   0 },     34.95659,   0.00002,     0.060 },
 { {   0,   0,  17, -22,   0,   0,   0,   0,   2,   0,   0 },    214.95654,   0.00002,     0.033 },
 { {   0,   0,  17, -21,   0,   0,   0,   0,   2,  -1,   0 },     63.66319,   0.00002,     0.058 },
 { {   0,   0,  17, -21,   0,   0,   0,   0,   2,   0,   0 },    243.66392,   0.00003,     0.033 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,   2,  -1,   0 },     92.14620,   0.00003,     0.056 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,   2,   0,   0 },    272.14608,   0.00003,     0.032 },
 { {   0,   0,  17, -19,   0,   0,   0,   0,   2,  -1,   0 },    120.01885,   0.00002,     0.054 },
 { {   0,   0,  17, -19,   0,   0,   0,   0,   2,   0,   0 },    300.01861,   0.00002,     0.032 },
 { {   0,   0,  17, -18,   0,   0,   0,   0,   2,   0,   0 },    327.21710,   0.00001,     0.031 },
 { {   0,   0,  18, -24,   0,   0,   0,   0,   2,   0,   0 },    186.03230,   0.00001,     0.033 },
 { {   0,   0,  18, -23,   0,   0,   0,   0,   2,  -1,   0 },     34.80544,   0.00002,     0.058 },
 { {   0,   0,  18, -23,   0,   0,   0,   0,   2,   0,   0 },    214.80539,   0.00002,     0.033 },
 { {   0,   0,  18, -22,   0,   0,   0,   0,   2,  -1,   0 },     63.44474,   0.00002,     0.056 },
 { {   0,   0,  18, -22,   0,   0,   0,   0,   2,   0,   0 },    243.44465,   0.00002,     0.032 },
 { {   0,   0,  18, -21,   0,   0,   0,   0,   2,   0,   0 },    271.79915,   0.00002,     0.032 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,   2,   0,   0 },    299.33957,   0.00002,     0.031 },
 { {   0,   0,  19, -25,   0,   0,   0,   0,   2,   0,   0 },    185.90488,   0.00001,     0.033 },
 { {   0,   0,  19, -24,   0,   0,   0,   0,   2,   0,   0 },    214.63055,   0.00001,     0.032 },
 { {   0,   0,  19, -23,   0,   0,   0,   0,   2,   0,   0 },    243.19122,   0.00002,     0.032 },
 { {   0,   0,  19, -22,   0,   0,   0,   0,   2,   0,   0 },    271.39354,   0.00002,     0.031 },
 { {   0,   0,  19, -21,   0,   0,   0,   0,   2,   0,   0 },    298.49992,   0.00001,     0.031 },
 { {   0,   0,  19, -17,  -3,   0,   0,   0,  -4,   3,   0 },    255.59386,   0.00003,  1361.910 },
 { {   0,   0,  20, -26,   0,   0,   0,   0,   2,   0,   0 },    185.75972,   0.00002,     0.032 },
 { {   0,   0,  20, -25,   0,   0,   0,   0,   2,   0,   0 },    214.43043,   0.00001,     0.032 },
 { {   0,   0,  20, -24,   0,   0,   0,   0,   2,   0,   0 },    242.90049,   0.00001,     0.031 },
 { {   0,   0,  20, -23,   0,   0,   0,   0,   2,   0,   0 },    270.91498,   0.00002,     0.031 },
 { {   0,   0,  20, -18,   0,   0,   0,   0,  -3,   1,   1 },    140.45460,   0.00016,   584.305 },
 { {   0,   0,  20, -16,   0,   0,  -1,   0,  -2,   1,   0 },    105.22322,   0.00002,  1701.651 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -4,   1,   0 },    327.06701,   0.00002,     0.040 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -4,   2,   0 },    327.06701,   0.00002,     0.087 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   0,   0 },    327.02107,   0.00007,     0.076 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   1,   0 },    327.06701,   0.00148,    88.385 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   2,   0 },    327.10461,   0.00009,     0.075 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,   0,   0,   0 },    327.06701,   0.00002,     0.087 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,   0,   1,   0 },    327.06701,   0.00002,     0.040 },
 { {   0,   0,  20, -13,   2,   0,   0,   0,   0,  -1,   0 },     59.40255,   0.00002,  1043.255 },
 { {   0,   0,  21, -26,   0,   0,   0,   0,   2,   0,   0 },    214.20177,   0.00001,     0.031 },
 { {   0,   0,  21, -25,   0,   0,   0,   0,   2,   0,   0 },    242.56605,   0.00001,     0.031 },
 { {   0,   0,  21, -24,   0,   0,   0,   0,   2,   0,   0 },    270.36093,   0.00001,     0.030 },
 { {   0,   0,  21, -18,   0,   0,   0,   0,  -2,   1,   0 },     15.76431,   0.00014,    19.211 },
 { {   0,   0,  22, -19,  -5,   0,   0,   0,  -2,   1,   0 },     87.78273,   0.00001,   190.106 },
 { {   0,   0,  23, -21,  -4,   0,   0,   0,  -2,   1,   0 },    326.30902,   0.00001,    63.795 },
 { {   0,   0,  23, -20,   0,   0,   0,   0,  -1,   0,   0 },    100.98487,   0.00009,   411.317 },
 { {   0,   0,  23, -20,   2,   0,   0,   0,  -1,   1,  -1 },    258.27890,   0.00008,  1726.003 },
 { {   0,   0,  26, -27,  -2,   0,   0,   0,  -2,   0,   0 },    268.21423,   0.00001,     0.075 },
 { {   0,   0,  26, -27,  -2,   0,   0,   0,  -2,   1,   0 },    356.23747,   0.00001,   171.793 },
 { {   0,   0,  26, -27,  -2,   0,   0,   0,  -2,   2,   0 },    268.21423,   0.00001,     0.075 },
 { {   0,   0,  26, -27,  -1,   0,   0,   0,  -2,   1,   0 },    103.45962,   0.00001,    12.742 },
 { {   0,   0,  26, -27,   0,   0,   0,   0,  -2,   1,   0 },    336.95700,   0.00001,     6.143 },
 { {   0,   0,  26, -24,   0,   0,   0,   0,   0,  -1,   0 },    278.29382,   0.00004,    61.818 },
 { {   0,   0,  26, -22,   0,   0,   0,   0,   1,  -1,  -1 },    229.82382,   0.00007,   152.117 },
 { {   0,   0,  27, -29,   0,   0,   0,   0,  -2,   1,   0 },     53.21564,   0.00002,    10.059 },
 { {   0,   0,  28, -31,   0,   0,  -3,   0,  -2,   1,   0 },     34.45428,   0.00013,  2901.861 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   0,   0 },     29.61085,   0.00002,     0.076 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   1,   0 },     28.79619,   0.00046,    27.739 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   2,   0 },     28.36335,   0.00003,     0.075 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   0,   0 },    305.75894,   0.00003,     0.075 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   1,   0 },    305.75894,   0.00050,    36.607 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   2,   0 },    305.75894,   0.00003,     0.076 },
 { {   0,   0,  29, -32,  -6,   0,   0,   0,  -2,   1,   0 },    125.45704,   0.00003,   685.011 },
 { {   0,   0,  29, -31,   0,   0,   0,   0,  -1,   1,  -1 },    211.61091,   0.00002,    56.451 },
 { {   0,   0,  29, -29,  -4,   0,   0,   0,   0,  -1,   0 },     52.72877,   0.00001,    84.635 },
 { {   0,   0,  29, -29,   0,   0,   0,   0,   0,   1,  -2 },    336.96819,   0.00001,   123.278 },
 { {   0,   0,  30, -35,   0,   0,   0,   0,  -2,   1,   0 },    250.32536,   0.00004,    11.027 },
 { {   0,   0,  30, -35,   2,   0,   0,   0,  -2,   1,   0 },     79.50916,   0.00002,    12.833 },
 { {   0,   0,  30, -33,  -3,   0,   0,   0,  -1,  -1,   1 },    125.60411,   0.00014,  2071.859 },
 { {   0,   0,  30, -33,  -1,   0,   0,   0,  -1,   0,   0 },    242.86547,   0.00003,   730.361 },
 { {   0,   0,  31, -37,   0,   0,   0,   0,  -2,   1,   0 },    179.39809,   0.00002,     6.491 },
 { {   0,   0,  31, -37,   2,   0,   0,   0,  -2,   1,   0 },     92.98304,   0.00001,    68.710 },
 { {   0,   0,  31, -34,   0,   0,   0,   0,   0,  -3,   2 },    244.98107,   0.00004,  2022.190 },
 { {   0,   0,  32, -39,   2,   0,   0,   0,  -2,   1,   0 },     22.69720,   0.00001,    20.486 },
 { {   0,   0,  32, -39,   3,   0,   0,   0,  -2,   1,   0 },    352.81873,   0.00001,    28.177 },
 { {   0,   0,  32, -35,   0,  -4,   0,   0,   0,  -1,   0 },    284.84382,   0.00005,  2015.126 },
 { {   0,   0,  33, -41,   3,   0,   0,   0,  -2,   1,   0 },    339.58641,   0.00001,    35.871 },
 { {   0,   0,  33, -40,  -3,   0,   0,   0,  -2,   1,   0 },     88.94043,   0.00004,   494.910 },
 { {   0,   0,  33, -37,  -1,   0,   0,   0,   0,  -1,   0 },      6.72121,   0.00001,    80.800 },
 { {   0,   0,  34, -43,   0,   0,   0,   0,  -2,  -1,   2 },    271.07256,   0.00001,    93.931 },
 { {   0,   0,  34, -43,   4,   0,   0,   0,  -2,   1,   0 },    261.07370,   0.00002,   144.041 },
 { {   0,   0,  34, -42,  -2,   0,   0,   0,  -2,   1,   0 },     69.86408,   0.00001,    52.868 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,  -1,   1 },     33.74756,   0.00017,   958.407 },
 { {   0,   0,  34, -41,   2,   0,   0,   0,  -1,  -1,   0 },    210.39978,   0.00001,     0.075 },
 { {   0,   0,  34, -41,   2,   0,   0,   0,  -1,   0,   0 },    210.39978,   0.00020,  1236.942 },
 { {   0,   0,  34, -41,   2,   0,   0,   0,  -1,   1,   0 },    210.39978,   0.00001,     0.075 },
 { {   0,   0,  34, -39,   0,   0,   0,   0,   0,  -1,   0 },     30.96216,   0.00018,   116.831 },
 { {   0,   0,  34, -37,   0,   0,   0,   0,   1,  -1,  -1 },    182.37116,   0.00002,    55.060 },
 { {   0,   0,  35, -44,  -2,   0,   0,   0,  -2,   1,   0 },     25.91301,   0.00002,    22.496 },
 { {   0,   0,  35, -44,  -1,   0,   0,   0,  -2,   1,   0 },    263.24490,   0.00002,    25.094 },
 { {   0,   0,  35, -41,   0,   0,   0,   0,   0,  -1,   0 },    352.96137,   0.00001,    18.245 },
 { {   0,   0,  36, -46,  -1,   0,   0,   0,  -2,   1,   0 },    221.49445,   0.00001,    42.523 },
 { {   0,   0,  36, -44,   0,  -2,   0,   0,  -1,   1,  -1 },    149.08247,   0.00001,   401.021 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   0,   0 },    222.48113,   0.00003,     0.075 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   1,   0 },    251.77849,   0.00016,   387.400 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   2,   0 },    233.58943,   0.00001,     0.075 },
 { {   0,   0,  37, -46,   0,   0,   0,   0,  -1,   1,  -1 },      2.11691,   0.00008,   142.421 },
 { {   0,   0,  37, -45,   2,   0,   0,   0,   0,  -1,   0 },    356.17714,   0.00003,    77.298 },
 { {   0,   0,  38, -47,   3,   0,   0,   0,   0,  -1,   0 },    297.28448,   0.00001,   125.022 },
 { {   0,   0,  40, -53,  -4,   0,   0,   0,  -2,   1,   0 },    270.34871,   0.00001,   561.851 },
 { {   0,   0,  40, -50,  -2,   0,   0,   0,   0,  -1,   0 },    352.60679,   0.00001,   116.607 },
 { {   0,   0,  40, -48,  -2,   0,   0,   0,   1,  -1,  -1 },    189.70638,   0.00003,   973.771 },
 { {   0,   0,  40, -48,   0,   0,   0,   0,   1,   0,  -2 },    301.00192,   0.00002,   347.370 },
 { {   0,   0,  41, -56,   3,   0,   0,   0,  -2,   1,   0 },    126.31894,   0.00001,   318.264 },
 { {   0,   0,  42, -54,   0,  -1,   0,   0,   0,  -2,   0 },    311.88046,   0.00001,     0.075 },
 { {   0,   0,  42, -54,   0,  -1,   0,   0,   0,  -1,   0 },    311.88046,   0.00021,  1533.043 },
 { {   0,   0,  42, -54,   0,  -1,   0,   0,   0,   0,   0 },    311.88046,   0.00001,     0.075 },
 { {   0,   0,  43, -55,  -2,   0,   0,   0,   0,   1,  -2 },    312.30343,   0.00001,  1957.040 },
 { {   0,   0,  43, -52,   0,   0,   0,   0,   2,  -1,  -2 },    273.60664,   0.00001,    92.025 },
 { {   0,   0,  44, -61,  -1,   0,   0,   0,  -2,   1,   0 },     31.13663,   0.00007,   820.282 },
 { {   0,   0,  44, -56,   1,   0,   0,   0,   1,  -1,  -1 },    297.36536,   0.00013,  2145.021 },
 { {   0,   0,  45, -63,   0,   0,   0,   0,  -2,   1,   0 },    178.89729,   0.00001,    45.137 },
 { {   0,   0,  48, -69,   2,   0,   0,   0,  -2,   1,   0 },     11.69227,   0.00005,  1518.942 },
 { {   0,   0,  48, -67,   0,   0,   0,   0,  -1,   0,   0 },    335.32219,   0.00008,   118.917 },
 { {   0,   1, -42,  51,   0,   0,   0,   0,   0,   1,   0 },    279.85878,   0.00001,   353.939 },
 { {   0,   1, -37,  45,   0,   0,   0,   0,   2,  -1,   0 },     92.72444,   0.00002,    30.256 },
 { {   0,   1, -36,  43,   0,   0,   0,   0,   2,  -1,   0 },     63.39727,   0.00001,    32.986 },
 { {   0,   1, -11,   0,   0,   0,   0,   0,   4,  -1,  -2 },    154.75786,   0.00005,  3701.205 },
 { {   0,   1, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.62915,   0.00001,     0.030 },
 { {   0,   1, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.62915,   0.00001,     0.152 },
 { {   0,   1,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    120.65475,   0.00002,     0.056 },
 { {   0,   1,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.50192,   0.00001,     0.032 },
 { {   0,   1,  -8,   0,   2,   0,   0,   0,  -2,   2,   0 },    142.50192,   0.00001,     0.226 },
 { {   0,   1,  -8,  10,   0,   0,   0,   0,  -1,   0,   1 },    214.60696,   0.00002,   258.482 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,  -2,   0,   0 },    145.52467,   0.00006,     0.040 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,  -2,   1,   0 },    145.64240,   0.00008,     0.087 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   0,  -1,   0 },    144.84768,   0.00011,     0.075 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   0,   0,   0 },    144.55296,   0.00051,   174.398 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   0,   1,   0 },    144.84768,   0.00011,     0.075 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   2,  -1,   0 },    145.64347,   0.00008,     0.087 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   2,   0,   0 },    145.52477,   0.00006,     0.040 },
 { {   0,   1,  -8,  13,   0,   0,   0,   0,   0,   0,   0 },    354.25830,   0.00001,     1.861 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    211.17739,   0.00001,     0.023 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    217.87908,   0.00003,     0.033 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    214.00551,   0.00002,     0.059 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    218.68307,   0.00001,     0.278 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },     36.87142,   0.00001,     0.104 },
 { {   0,   1,  -7,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.68665,   0.00005,     0.033 },
 { {   0,   1,  -7,   0,   2,   0,   0,   0,  -2,   1,   0 },    231.64116,   0.00002,     0.060 },
 { {   0,   1,  -7,   0,   2,   0,   0,   0,   0,  -2,   0 },    231.64262,   0.00001,     0.032 },
 { {   0,   1,  -7,   0,   2,   0,   0,   0,   0,   0,   0 },    231.64262,   0.00001,     0.192 },
 { {   0,   1,  -7,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.90846,   0.00002,     0.034 },
 { {   0,   1,  -7,   2,   0,   0,   0,   0,  -2,   1,   0 },     51.69452,   0.00002,     0.063 },
 { {   0,   1,  -7,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.90846,   0.00002,     0.394 },
 { {   0,   1,  -7,   2,   0,   0,   0,   0,  -2,   3,   0 },     51.69452,   0.00002,     0.093 },
 { {   0,   1,  -7,  11,   0,   0,   0,   0,   0,   0,   0 },    325.28939,   0.00001,     2.110 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.10009,   0.00002,     0.024 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    346.16470,   0.00005,     0.385 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    142.69345,   0.00001,     0.032 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.10009,   0.00003,     0.057 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    152.43253,   0.00005,     0.229 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    331.40107,   0.00003,     0.141 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },     52.69111,   0.00001,     0.034 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,  -2,   1,   0 },     50.46931,   0.00001,     0.063 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,   0,  -1,   0 },    230.46931,   0.00001,     0.057 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,   0,   1,   0 },    230.46931,   0.00001,     0.112 },
 { {   0,   1,  -6,   0,   2,   0,   0,   0,  -2,   0,   0 },    291.82436,   0.00005,     0.035 },
 { {   0,   1,  -6,   0,   2,   0,   0,   0,  -2,   1,   0 },    291.95665,   0.00001,     0.064 },
 { {   0,   1,  -6,   0,   2,   0,   0,   0,  -2,   3,   0 },    291.95665,   0.00001,     0.092 },
 { {   0,   1,  -6,   2,   0,   0,   0,   0,  -2,   0,   0 },    272.30579,   0.00003,     0.036 },
 { {   0,   1,  -6,   2,   0,   0,   0,   0,  -2,   2,   0 },    273.61358,   0.00002,     0.651 },
 { {   0,   1,  -6,   3,   0,   0,   0,   0,  -2,   0,   0 },    288.62810,   0.00001,     0.036 },
 { {   0,   1,  -6,   3,   0,   0,   0,   0,  -2,   1,   0 },    108.62810,   0.00002,     0.070 },
 { {   0,   1,  -6,   3,   0,   0,   0,   0,  -2,   2,   0 },    288.62810,   0.00001,     0.995 },
 { {   0,   1,  -6,   3,   0,   0,   0,   0,  -2,   3,   0 },    108.62810,   0.00002,     0.082 },
 { {   0,   1,  -6,   6,   0,   0,   0,   0,   0,   0,   0 },    209.49856,   0.00001,     0.844 },
 { {   0,   1,  -6,   8,   0,   0,   0,   0,   0,   0,   0 },    271.98085,   0.00001,     8.264 },
 { {   0,   1,  -5,   0,  -1,   0,   0,   0,  -2,   0,   0 },     58.49155,   0.00002,     0.035 },
 { {   0,   1,  -5,   0,  -1,   0,   0,   0,  -2,   2,   0 },     65.67359,   0.00001,     0.594 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.78036,   0.00002,     0.019 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -4,   3,   0 },     51.84184,   0.00003,     0.076 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,  -2,   0 },     51.77916,   0.00001,     0.018 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     53.78387,   0.00005,     0.036 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    231.14937,   0.00002,     0.067 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    232.02408,   0.00004,     0.086 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },     51.67657,   0.00001,     0.033 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    323.65236,   0.00004,     0.296 },
 { {   0,   1,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },    308.77167,   0.00003,     0.036 },
 { {   0,   1,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },    161.34807,   0.00003,     0.068 },
 { {   0,   1,  -5,   0,   1,   0,   0,   0,  -2,   2,   0 },    221.25700,   0.00001,     0.660 },
 { {   0,   1,  -5,   2,   0,   0,   0,   0,   0,   0,   0 },    172.33442,   0.00001,     0.433 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -4,   3,   0 },    181.63861,   0.00001,     0.087 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -2,   1,   0 },    354.31941,   0.00009,     0.075 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -2,   2,   0 },    133.11357,   0.00007,   184.957 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -2,   3,   0 },     13.99536,   0.00005,     0.075 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,   0,   1,   0 },    181.63861,   0.00001,     0.087 },
 { {   0,   1,  -5,   5,   0,   0,   0,   0,  -1,   2,  -1 },    330.72302,   0.00002,   238.317 },
 { {   0,   1,  -5,   5,   0,   0,   0,   0,   0,   0,   0 },     73.00507,   0.00001,     1.397 },
 { {   0,   1,  -5,   6,   0,   0,   0,   0,   0,   0,   0 },    260.62470,   0.00001,     5.424 },
 { {   0,   1,  -5,   8,  -3,   0,   0,   0,   0,   0,   0 },     71.94884,   0.00001,     1.597 },
 { {   0,   1,  -4,  -2,   0,   0,   0,   0,  -2,   0,   0 },     11.07071,   0.00003,     0.035 },
 { {   0,   1,  -4,  -2,   0,   0,   0,   0,  -2,   2,   0 },     11.02763,   0.00002,     0.601 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,  -2,   0,   0 },     54.54187,   0.00006,     0.037 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0 },     62.34610,   0.00010,     0.072 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,  -2,   3,   0 },    329.75732,   0.00001,     0.080 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,   0,   1,   0 },    242.34663,   0.00002,     0.093 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },    128.84343,   0.00002,     0.019 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -4,   2,   0 },    128.84343,   0.00002,     0.039 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    107.91554,   0.00006,     0.037 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    241.96897,   0.00007,     0.072 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    149.71842,   0.00003,     1.665 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    358.56416,   0.00003,     0.079 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    118.18029,   0.00002,     0.068 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    237.23392,   0.00004,     0.757 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    241.77089,   0.00003,     0.064 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    245.66895,   0.00028,     0.421 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    251.05414,   0.00004,     0.092 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },     60.72392,   0.00003,     0.292 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    243.90761,   0.00005,     0.110 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    242.99129,   0.00007,     0.045 },
 { {   0,   1,  -4,   0,   1,   0,   0,   0,  -2,   1,   0 },    308.66953,   0.00003,     0.073 },
 { {   0,   1,  -4,   0,   2,   0,   0,   0,   2,  -2,   0 },     90.68012,   0.00002,     0.251 },
 { {   0,   1,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },    270.68012,   0.00002,     0.044 },
 { {   0,   1,  -4,   0,   3,   0,   0,   0,  -2,   0,   2 },    357.63960,   0.00003,    71.265 },
 { {   0,   1,  -4,   0,   3,   0,   0,   0,   0,   0,   0 },    346.28792,   0.00001,     0.471 },
 { {   0,   1,  -4,   1,   0,   0,   0,   0,  -2,   1,   0 },    297.13022,   0.00003,     0.075 },
 { {   0,   1,  -4,   2,   0,   0,   0,   0,   0,   0,   0 },    272.79017,   0.00002,     0.763 },
 { {   0,   1,  -4,   3,   0,   0,   0,   0,   0,   0,   0 },    353.42390,   0.00002,     1.283 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   0,  -1,   0 },     29.73664,   0.00003,     0.074 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   0,   0,   0 },    216.78407,   0.00002,     4.037 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   0,   1,   0 },     29.73664,   0.00003,     0.077 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   2,  -2,   0 },     29.59891,   0.00001,     0.495 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   2,   0,   0 },     29.59891,   0.00001,     0.041 },
 { {   0,   1,  -3,  -1,   0,   0,   0,   0,  -2,   1,   0 },    270.41908,   0.00002,     0.075 },
 { {   0,   1,  -3,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.58262,   0.00003,     0.038 },
 { {   0,   1,  -3,   0,  -2,   0,   0,   0,  -2,   1,   0 },    230.08332,   0.00006,     0.077 },
 { {   0,   1,  -3,   0,  -2,   0,   0,   0,  -2,   3,   0 },     54.61614,   0.00001,     0.074 },
 { {   0,   1,  -3,   0,  -1,   0,   0,   0,  -2,   1,   0 },    304.09719,   0.00002,     0.077 },
 { {   0,   1,  -3,   0,  -1,   0,   0,   0,  -2,   3,   0 },     23.00902,   0.00001,     0.074 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -4,   2,   0 },    342.31271,   0.00002,     0.041 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -3,   0,   1 },    165.30624,   0.00001,     0.040 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -3,   1,   1 },    348.37569,   0.00001,     0.085 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    334.36294,   0.00002,     0.025 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    322.87761,   0.00008,     0.038 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    338.12280,   0.00109,     0.078 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    336.13776,   0.00010,     2.503 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    355.67555,   0.00005,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -2,   3 },    348.33066,   0.00002,    78.110 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    166.46533,   0.00018,     0.074 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    340.67074,   0.00013,     0.037 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    336.25151,   0.00016,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    165.27038,   0.00100,     3.118 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },    346.53273,   0.00006,     0.072 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },    336.22575,   0.00002,     2.051 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.46533,   0.00018,     0.077 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    137.50080,   0.00003,     0.036 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    157.33561,   0.00043,     0.068 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    157.20414,   0.00379,     0.728 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    163.80284,   0.00082,     0.084 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    161.21970,   0.00005,     0.040 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    346.68712,   0.00001,     0.064 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.59298,   0.00023,     0.412 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    337.68703,   0.00003,     0.091 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },    167.89477,   0.00009,     0.041 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    346.68712,   0.00001,     0.092 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    167.87036,   0.00004,     0.318 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    154.13172,   0.00021,     0.099 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    160.39055,   0.00023,     0.043 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    157.81932,   0.00002,     0.027 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    167.38208,   0.00002,     0.021 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,  -2,   1,   0 },    266.87316,   0.00001,     0.078 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,  -2,   3,   0 },     86.87316,   0.00001,     0.073 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,   0,   0,   0 },    271.77167,   0.00001,     0.775 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,   2,  -2,   0 },     89.35628,   0.00003,     0.326 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },     89.20831,   0.00002,     0.043 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.90700,   0.00003,     0.039 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,  -2,   1,   0 },    178.66527,   0.00005,     0.079 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,   0,  -2,   0 },    180.43500,   0.00001,     0.036 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,   0,   0,   0 },    180.87444,   0.00006,     0.829 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,   2,  -1,   0 },    179.86658,   0.00003,     0.097 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,   2,   0,   0 },    180.25544,   0.00001,     0.042 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,   2,   2,   0 },    179.89491,   0.00001,     0.020 },
 { {   0,   1,  -3,   0,   3,   0,   0,   0,   2,   0,   0 },    170.90680,   0.00004,     0.042 },
 { {   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0,   0 },    163.75181,   0.00002,     1.187 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.21342,   0.00001,     0.684 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   0,  -1,   0 },    357.73009,   0.00002,     0.074 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   0,   1,   0 },    357.73009,   0.00002,     0.077 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,  -3,   0 },      0.20649,   0.00003,     0.065 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,  -2,   0 },    179.45871,   0.00002,     0.480 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,  -1,   0 },      0.20649,   0.00003,     0.090 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,   0,   0 },    179.91393,   0.00004,     0.041 },
 { {   0,   1,  -3,   3,   0,   0,   0,   0,   0,   0,   0 },     36.95977,   0.00005,     4.534 },
 { {   0,   1,  -3,   6,   0,   0,   0,   0,   2,  -3,   0 },    103.21612,   0.00002,     0.076 },
 { {   0,   1,  -3,   6,   0,   0,   0,   0,   2,  -1,   0 },    103.21612,   0.00002,     0.075 },
 { {   0,   1,  -2,  -3,   0,   0,   0,   0,  -2,   1,   0 },     11.74799,   0.00001,     0.074 },
 { {   0,   1,  -2,  -3,   0,   0,   0,   0,  -2,   3,   0 },    110.50170,   0.00001,     0.077 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,  -2,   1,   0 },    204.02656,   0.00002,     0.084 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,  -2,   3,   0 },     24.02656,   0.00002,     0.069 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    279.59430,   0.00001,     0.448 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },     99.59430,   0.00001,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },    147.42174,   0.00002,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },    121.74622,   0.00005,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   2,   0 },    209.20339,   0.00001,     0.043 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   0,   1 },    256.66370,   0.00001,     0.042 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   1,   1 },     32.93228,   0.00002,     0.093 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   2,  -1 },     76.62258,   0.00001,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    101.82811,   0.00002,     0.019 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    114.28512,   0.00026,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1,   2 },    117.63752,   0.00002,     0.087 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.83823,   0.00176,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },     76.62069,   0.00010,     0.086 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   2 },    117.86322,   0.00006,     0.577 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    210.07755,   0.00123,     0.084 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.49947,   0.00021,     0.715 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },    276.27082,   0.00006,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -2,   1 },    256.65863,   0.00002,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     84.29029,   0.00004,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },     88.45118,   0.00002,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    256.64752,   0.00044,     0.080 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },     76.60526,   0.00038,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },     88.33229,   0.00032,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.71004,   0.00165,     1.472 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },     76.98180,   0.00087,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },     91.65932,   0.00007,     1.951 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },    256.63883,   0.00031,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,  -1 },     76.63207,   0.00008,     2.892 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,   1 },    256.66081,   0.00002,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },    107.20507,   0.00004,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    105.56738,   0.00067,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   2 },    295.42269,   0.00004,    24.413 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,  -2 },    286.22944,   0.00003,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    106.02572,   0.00913,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   1 },    256.58715,   0.00008,     4.814 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    299.02292,   0.00007,     0.076 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -2 },    282.16708,   0.00001,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },     76.33223,   0.00015,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    103.71567,   0.04478,     2.670 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    256.61446,   0.00034,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   2 },    282.16708,   0.00001,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -2 },    108.91975,   0.00001,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -1 },     76.62831,   0.00002,     1.848 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    106.63234,   0.00852,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   1 },    256.61446,   0.00002,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   2 },    286.22944,   0.00003,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },    106.00555,   0.00054,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   3,   0 },    106.83091,   0.00004,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,  -1 },    256.64954,   0.00001,     0.036 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,   1 },     76.67233,   0.00004,     0.913 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    256.68172,   0.00024,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },    272.59961,   0.00003,     0.793 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },     76.81626,   0.00057,     0.082 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    256.71743,   0.00218,     0.700 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    280.64284,   0.00042,     0.083 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },     78.67042,   0.00050,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    256.53462,   0.00046,     0.085 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,   1 },     78.19981,   0.00004,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   2,  -1 },    256.63635,   0.00002,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    285.99786,   0.00001,     0.065 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    225.60085,   0.00011,     0.465 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    228.43342,   0.00186,     0.090 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },    211.10122,   0.00001,     0.403 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    236.84642,   0.00137,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    217.53858,   0.00008,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },    259.63558,   0.00003,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    248.74426,   0.00002,     0.044 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    263.06973,   0.00004,     0.028 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    219.57130,   0.00001,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   1,   0 },     87.84108,   0.00002,     0.016 },
 { {   0,   1,  -2,   0,   0,   1,   0,   0,   0,  -2,   2 },    359.79477,   0.00001,   142.558 },
 { {   0,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0 },      0.03881,   0.00001,     2.937 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,  -2,   0,   0 },    173.28844,   0.00006,     0.040 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,  -2,   1,   0 },    177.97083,   0.00018,     0.085 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,  -1,   0 },    356.28430,   0.00003,     0.074 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0 },      1.63536,   0.00008,     3.446 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,   1,   0 },    357.50031,   0.00008,     0.077 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   2,  -2,   0 },      1.62593,   0.00001,     0.484 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   2,  -1,   0 },    358.69008,   0.00002,     0.089 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    177.52907,   0.00003,     0.041 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   2,   2,   0 },    178.72160,   0.00001,     0.020 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,  -2,   0,   0 },    235.69843,   0.00001,     0.040 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,  -2,   1,   0 },    240.84591,   0.00002,     0.086 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   0,  -1,   0 },     60.84591,   0.00008,     0.074 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   0,   0,   0 },    212.09890,   0.00002,     4.858 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   0,   1,   0 },     60.84591,   0.00008,     0.077 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   2,  -2,   0 },     83.01609,   0.00002,     0.505 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   2,  -1,   0 },    240.84591,   0.00002,     0.089 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   2,   0,   0 },     83.01609,   0.00002,     0.041 },
 { {   0,   1,  -2,   0,   3,   0,   0,   0,   0,   0,   0 },    221.82558,   0.00001,     8.226 },
 { {   0,   1,  -2,   0,   4,   0,   0,   0,   0,   0,   0 },    241.06307,   0.00002,    26.839 },
 { {   0,   1,  -2,   1,   0,   0,   0,   0,   0,  -1,   0 },    168.88743,   0.00002,     0.076 },
 { {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0 },    316.81706,   0.00001,     6.361 },
 { {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   1,   0 },    168.88743,   0.00002,     0.075 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   0,   0 },    220.79643,   0.00002,     0.042 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   1,   0 },    220.71872,   0.00001,     0.093 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   2,   0 },    220.94154,   0.00003,     0.406 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   3,   0 },    220.71872,   0.00001,     0.064 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   0,  -1,   0 },    222.50139,   0.00002,     0.080 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   0,   0,   0 },    280.74161,   0.00001,     1.452 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   0,   1,   0 },    222.50139,   0.00002,     0.072 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -3,   0 },     44.75552,   0.00004,     0.070 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -2,   0 },     44.75553,   0.00003,     0.922 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -1,   0 },     44.75552,   0.00004,     0.082 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,   0,   0 },     45.43762,   0.00002,     0.039 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.43573,   0.00002,     0.043 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.43573,   0.00002,     0.334 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,  -3,   0 },    236.45162,   0.00003,     0.072 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,  -2,   0 },    236.88038,   0.00002,     1.807 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,  -1,   0 },    236.45162,   0.00003,     0.079 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,   0,   0 },    236.88038,   0.00002,     0.039 },
 { {   0,   1,  -2,   4,   0,   0,   0,   0,   2,  -3,   0 },    268.58328,   0.00003,     0.075 },
 { {   0,   1,  -2,   4,   0,   0,   0,   0,   2,  -2,   0 },     97.55556,   0.00001,    45.963 },
 { {   0,   1,  -2,   4,   0,   0,   0,   0,   2,  -1,   0 },    217.11118,   0.00003,     0.076 },
 { {   0,   1,  -1,  -5,   0,   0,   0,   0,  -2,   1,   0 },     24.06686,   0.00002,     0.074 },
 { {   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0 },    120.66743,   0.00001,     2.284 },
 { {   0,   1,  -1,   0,  -2,   0,   0,   0,  -2,   0,   0 },     24.95662,   0.00002,     0.041 },
 { {   0,   1,  -1,   0,  -2,   0,   0,   0,  -2,   2,   0 },    204.95662,   0.00002,     0.448 },
 { {   0,   1,  -1,   0,  -2,   0,   0,   0,   0,   0,   0 },      6.13370,   0.00001,     2.189 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    211.70400,   0.00006,     0.041 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,  -2,   2,   0 },     91.85586,   0.00002,     0.432 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 },     72.78657,   0.00002,     1.848 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    193.92594,   0.00001,     0.069 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    193.92594,   0.00001,     0.811 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    193.92590,   0.00004,     0.083 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },    191.71094,   0.00007,     0.040 },
 { {   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0 },     88.27143,   0.00001,     1.690 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   0,   0 },    179.98711,   0.00002,     0.014 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   1,   0 },    179.98348,   0.00005,     0.017 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   2,   0 },    179.98327,   0.00006,     0.021 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   3,   0 },    179.77106,   0.00002,     0.030 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,  -2,   0 },    180.00084,   0.00001,     0.013 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,  -1,   0 },    179.99048,   0.00020,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    179.99291,   0.00113,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,   2 },    179.96845,   0.00001,     0.045 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,  -2 },    359.98272,   0.00002,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    179.99251,   0.00266,     0.028 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },    179.96731,   0.00175,     0.045 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   3,   0 },    180.35760,   0.00006,     0.110 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,  -1,   0 },    179.97191,   0.00003,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   0 },    179.81395,   0.00002,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   1,   0 },    359.99147,   0.00023,     0.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   2,   0 },      0.06095,   0.00004,     0.100 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   3,   0 },    359.98809,   0.00001,     0.304 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -3,   0 },    180.00386,   0.00009,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    180.01287,   0.00084,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    359.98429,   0.00004,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.98974,   0.00939,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   2 },    359.99045,   0.00023,     0.095 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,  -2 },    359.98662,   0.00025,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.98152,   0.09922,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   2 },    179.98419,   0.00164,     0.366 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,  -2 },    359.98269,   0.00031,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.98436,   0.13122,     0.092 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,  -2 },    359.97777,   0.00003,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.98818,   0.00497,     0.417 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },    179.94837,   0.00017,     0.064 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   4,   0 },    179.91990,   0.00001,     0.035 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   0 },      0.00588,   0.00047,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1 },     50.92030,   0.00003,     0.086 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    220.63296,   0.00001,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },      0.00756,   0.00666,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    149.12131,   0.00002,     0.596 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.00166,   0.00077,     0.661 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,  -1 },    220.63296,   0.00001,     0.743 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0 },    359.10505,   0.00002,     0.068 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -4,   0 },    180.01282,   0.00005,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -3,   0 },    180.01041,   0.00072,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,  -2 },      0.01037,   0.00005,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    180.00907,   0.01019,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   2 },      0.01331,   0.00001,     1.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,  -2 },      0.00796,   0.00037,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.00888,   0.15350,     0.079 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   2 },    180.04650,   0.00019,     0.070 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -2 },    359.99572,   0.00023,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    180.00988,   0.82155,     1.599 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   2 },      0.01039,   0.00050,     0.036 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -2 },    180.20508,   0.00012,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    180.01511,   0.12722,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   2 },      0.01333,   0.00031,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,  -2 },    359.88023,   0.00003,     3.425 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },    180.01602,   0.00885,     0.037 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   2 },      0.01182,   0.00006,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   3,   0 },    180.01777,   0.00062,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   4,   0 },    180.01934,   0.00004,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0 },    180.10248,   0.00025,     0.074 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   1 },    247.05434,   0.00003,    10.557 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },     31.60414,   0.00003,     0.073 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },    180.01289,   0.00292,     3.824 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    230.92029,   0.00013,     0.076 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -2 },    180.02054,   0.00003,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },     24.50142,   0.00025,     2.335 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    180.02010,   0.01218,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    189.74904,   0.00004,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   2 },      0.02150,   0.00002,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },     31.60414,   0.00003,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,   0 },    180.02159,   0.00075,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   2,   0 },    179.96511,   0.00005,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -4,   0 },    359.71514,   0.00001,     0.036 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    359.99555,   0.00176,     0.871 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -2 },      0.16506,   0.00002,     0.068 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.87973,   0.01292,     0.083 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   2 },    359.95374,   0.00003,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -2 },    180.02067,   0.00089,     0.675 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    180.01464,   0.02201,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   2 },    359.88784,   0.00002,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,  -2 },      0.08247,   0.00008,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,   0 },    179.99890,   0.00286,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   2,   0 },    180.08600,   0.00028,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   3,   0 },    180.06502,   0.00003,     0.015 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -2,   0 },    179.85485,   0.00003,     0.089 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.01830,   0.00017,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,   0,   0 },    180.01863,   0.00012,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,   2,   0 },    180.03914,   0.00002,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -3,   0 },      0.33775,   0.00006,     0.097 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.80687,   0.00019,     0.042 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -1,  -2 },    180.02983,   0.00002,     0.100 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -1,   0 },    179.97889,   0.00051,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   0,   0 },    179.99546,   0.00022,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   1,   0 },    179.99714,   0.00004,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   5,  -2,   0 },    180.01961,   0.00001,     0.028 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.02687,   0.00003,     0.016 },
 { {   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0 },     85.91332,   0.00002,     1.516 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,  -2,   0,   0 },    344.89843,   0.00004,     0.042 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,  -2,   1,   0 },    169.06361,   0.00001,     0.093 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   0,  -1,   0 },    349.06361,   0.00007,     0.080 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0 },    295.59737,   0.00001,     1.409 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   0,   1,   0 },    349.06361,   0.00007,     0.072 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   1,  -2,   1 },    105.14623,   0.00002,    95.933 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   1,  -1,   1 },     91.85473,   0.00009,     0.075 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,  -3,   0 },    314.47497,   0.00002,     0.070 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,  -2,   0 },    314.42020,   0.00002,     0.940 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,  -1,   0 },    315.09599,   0.00004,     0.082 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },    330.47710,   0.00007,     0.039 },
 { {   0,   1,  -1,   0,   2,   0,   0,   0,   0,   0,   0 },     74.45150,   0.00001,     1.259 },
 { {   0,   1,  -1,   0,   2,   0,   0,   0,   1,   0,  -1 },    354.96323,   0.00001,     3.852 },
 { {   0,   1,  -1,   2,   0,   0,   0,   0,   2,  -3,   0 },    211.59856,   0.00002,     0.075 },
 { {   0,   1,  -1,   2,   0,   0,   0,   0,   2,  -1,   0 },    231.87620,   0.00006,     0.076 },
 { {   0,   1,   0,  -7,   0,   0,   0,   0,  -2,   0,   2 },    176.64640,   0.00001,    88.775 },
 { {   0,   1,   0,  -6,   0,   0,   0,   0,  -2,   1,   0 },    206.52704,   0.00003,     0.077 },
 { {   0,   1,   0,  -3,   0,   0,   0,   0,   0,  -1,   0 },      0.65587,   0.00002,     0.076 },
 { {   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0 },    324.21707,   0.00003,    32.819 },
 { {   0,   1,   0,  -3,   0,   0,   0,   0,   0,   1,   0 },      0.65587,   0.00002,     0.075 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   2,   0 },    321.25531,   0.00002,     0.428 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -1,   0 },    321.56224,   0.00003,     0.079 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0 },    321.56224,   0.00003,     0.072 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -3,   0 },    139.61750,   0.00004,     0.069 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -2,   0 },    140.14083,   0.00002,     0.825 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -1,   0 },    139.61750,   0.00004,     0.083 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,   0,   0 },    140.14083,   0.00002,     0.040 },
 { {   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0 },    316.07824,   0.00001,     0.729 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },    303.46154,   0.00006,     0.043 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   2,  -3,   0 },    140.80377,   0.00002,     0.074 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.54950,   0.00001,     3.154 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   2,  -1,   0 },    320.80377,   0.00002,     0.077 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   2,   0,   0 },    128.54950,   0.00001,     0.038 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.15333,   0.00004,     0.043 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   1,   0 },      0.99083,   0.00003,     0.101 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.69208,   0.00001,     0.302 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -2,   0 },    180.99083,   0.00003,     0.040 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0 },      0.45420,   0.00004,     0.085 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },      4.89774,   0.00004,     0.649 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0 },      0.45420,   0.00004,     0.068 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -3,   0 },      3.60398,   0.00003,     0.074 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -2,   0 },      4.07373,   0.00002,     4.296 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -1,   0 },    183.60398,   0.00003,     0.077 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },      3.49427,   0.00007,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,  -2,   0 },     76.97330,   0.00001,     0.013 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,  -1,   0 },     77.86114,   0.00003,     0.016 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },     78.47725,   0.00024,     0.021 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },     77.30098,   0.00040,     0.029 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   2,   0 },     77.86318,   0.00016,     0.047 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0 },     76.36328,   0.00002,     0.028 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.36093,   0.00002,     0.045 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   1,  -1 },    103.33732,   0.00001,     0.028 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   1 },    103.33833,   0.00001,     0.114 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   3,   0 },    256.36328,   0.00001,     0.233 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -2,   0 },     76.87897,   0.00011,     0.020 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },     77.14490,   0.00093,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   2 },    256.12626,   0.00001,     0.105 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,  -2 },    257.03654,   0.00002,     0.020 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },     77.05540,   0.00942,     0.043 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   2 },     76.64842,   0.00010,     0.268 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },     76.96472,   0.00873,     0.101 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   2 },    256.12626,   0.00001,     0.059 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },     74.64992,   0.00033,     0.294 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   3,   0 },     85.01446,   0.00007,     0.060 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   4,   0 },     80.28938,   0.00001,     0.033 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.34053,   0.00007,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.79516,   0.00009,     0.095 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.30756,   0.00077,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    258.05341,   0.00022,     0.093 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.33926,   0.00198,     0.373 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,  -1 },    102.52540,   0.00045,     0.092 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   1 },    283.34399,   0.00022,     0.063 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,  -1 },     99.18276,   0.00001,     0.426 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   1 },    283.33860,   0.00001,     0.034 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0 },     84.44125,   0.00003,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0 },     80.77578,   0.00049,     0.040 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },     81.70343,   0.00654,     0.086 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },    282.86659,   0.00012,     0.085 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     82.23299,   0.01457,     0.615 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 },    103.33745,   0.00005,     0.066 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     81.22907,   0.00191,     0.067 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0 },     83.18798,   0.00011,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,  -1 },    103.33515,   0.00009,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    263.74264,   0.00001,     0.080 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   1 },    283.47606,   0.00003,     1.105 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.33358,   0.00141,     0.079 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    270.37704,   0.00003,     1.354 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.33733,   0.00088,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -3 },    283.33100,   0.00001,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.33371,   0.00894,     1.749 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    271.09250,   0.00017,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   1 },    283.33267,   0.00114,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.38366,   0.00126,     0.072 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   1,   1 },    283.33365,   0.00012,     0.025 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   2,  -1 },    103.33503,   0.00009,     0.037 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    269.57878,   0.00004,     0.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -3,   1 },    282.43438,   0.00003,    54.757 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    265.80280,   0.00195,     6.734 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   1 },    103.34166,   0.00003,     0.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,  -1 },    103.33116,   0.00004,     3.172 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.20661,   0.00821,     0.076 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },    103.33748,   0.00020,     0.077 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    275.31156,   0.00062,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    271.84206,   0.00005,     0.025 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -2,  -1 },    283.33586,   0.00002,     0.832 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -1,  -1 },    283.32713,   0.00018,     0.083 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -1,   1 },    283.33111,   0.00002,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.33675,   0.00008,     0.040 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   1,  -1 },    103.33381,   0.00001,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   4,  -3,   0 },    265.51795,   0.00003,     0.088 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   4,  -2,   0 },    273.65329,   0.00009,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.79278,   0.00007,     0.026 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,  -2,   0,   0 },     80.43315,   0.00001,     0.043 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,  -2,   2,   0 },    260.43315,   0.00001,     0.287 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 },     94.52045,   0.00001,     0.585 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,  -3,   0 },    285.80937,   0.00002,     0.075 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },    264.99470,   0.00002,    15.579 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    302.40922,   0.00006,     0.076 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    151.68345,   0.00001,     0.038 },
 { {   0,   1,   0,   0,   2,   0,   0,   0,   2,  -3,   0 },    149.99234,   0.00004,     0.076 },
 { {   0,   1,   0,   0,   2,   0,   0,   0,   2,  -2,   0 },    342.79825,   0.00002,    49.716 },
 { {   0,   1,   0,   0,   2,   0,   0,   0,   2,  -1,   0 },    111.60287,   0.00003,     0.075 },
 { {   0,   1,   1,  -5,   0,   0,   0,   0,   0,   0,   0 },    100.47026,   0.00002,    30.398 },
 { {   0,   1,   1,  -3,   0,   0,   0,   0,   1,   0,  -1 },    242.60296,   0.00001,    42.928 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,  -2,   0,   0 },    180.13117,   0.00003,     0.043 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,  -2,   1,   0 },      0.05172,   0.00003,     0.101 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,  -2,   3,   0 },      0.05172,   0.00001,     0.060 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,   0,  -1,   0 },      0.05172,   0.00001,     0.086 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },    232.00212,   0.00001,     0.640 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,   0,   1,   0 },      0.05172,   0.00001,     0.067 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,   2,  -3,   0 },    183.77456,   0.00001,     0.074 },
 { {   0,   1,   1,  -2,   0,   0,   0,   0,   2,  -1,   0 },    183.77456,   0.00001,     0.077 },
 { {   0,   1,   1,  -1,   0,   0,   0,   0,   2,   0,  -2 },     37.48181,   0.00002,    73.036 },
 { {   0,   1,   1,   0,  -4,   0,   0,   0,   0,   0,   0 },    199.68887,   0.00001,     0.437 },
 { {   0,   1,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },    191.53701,   0.00004,     0.045 },
 { {   0,   1,   1,   0,  -3,   0,   0,   0,  -2,   2,   0 },      9.14997,   0.00001,     0.241 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },    180.04313,   0.00009,     0.111 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,  -2,   3,   0 },      0.14769,   0.00001,     0.057 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   0,  -1,   0 },    180.06292,   0.00002,     0.093 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },    180.76211,   0.00005,     0.407 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   0,   1,   0 },    180.06292,   0.00002,     0.064 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   2,  -3,   0 },      0.34114,   0.00001,     0.080 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },    180.14229,   0.00005,     0.072 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },      0.09384,   0.00002,     0.037 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },     90.65294,   0.00003,     0.045 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,  -2,   2,   0 },     90.65294,   0.00003,     0.232 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,   0,  -2,   0 },    345.84756,   0.00003,     0.042 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },    189.16197,   0.00003,     0.394 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -4,   0,   0 },     96.29441,   0.00001,     0.021 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -4,   1,   0 },    348.79785,   0.00004,     0.030 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    351.83035,   0.00009,     0.028 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    355.71568,   0.00087,     0.045 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    357.26565,   0.00067,     0.113 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },      2.00013,   0.00006,     0.227 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,  -1 },      9.96778,   0.00004,     0.043 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    204.51726,   0.00005,     0.103 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    190.20605,   0.00011,     0.272 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   1,  -1 },      9.96777,   0.00001,     0.101 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -2,   0 },     17.24182,   0.00002,     0.042 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },      7.13155,   0.00041,     0.094 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     22.28319,   0.00273,     0.381 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     24.15493,   0.00027,     0.063 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   2,   0 },     18.78292,   0.00003,     0.034 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },     12.68482,   0.00016,     0.636 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    205.15282,   0.00005,     0.067 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    191.70508,   0.00003,     0.035 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    280.06138,   0.00002,     0.081 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },     13.60408,   0.00002,     1.174 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },     21.70667,   0.00066,     0.071 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,  -2 },     26.74414,   0.00001,     1.931 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    222.69193,   0.00004,     0.037 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   3,  -2,   0 },     23.89420,   0.00002,     0.076 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   4,  -2,   0 },     30.28179,   0.00001,     0.039 },
 { {   0,   1,   1,   0,   2,   0,   0,   0,   2,  -2,   0 },    308.15928,   0.00001,     0.980 },
 { {   0,   1,   1,   0,   2,   0,   0,   0,   2,  -1,   0 },    308.15928,   0.00003,     0.070 },
 { {   0,   1,   1,   0,   2,   0,   0,   0,   2,   0,   0 },    128.15928,   0.00001,     0.036 },
 { {   0,   1,   2,  -6,   0,   0,   0,   0,   0,   0,   0 },    273.78265,   0.00001,     2.297 },
 { {   0,   1,   2,  -4,   0,   0,   0,   0,  -2,   0,   0 },    330.64702,   0.00001,     0.043 },
 { {   0,   1,   2,  -4,   0,   0,   0,   0,  -2,   2,   0 },    330.64702,   0.00001,     0.306 },
 { {   0,   1,   2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    266.52214,   0.00002,     0.074 },
 { {   0,   1,   2,   0,  -3,   0,   0,   0,   0,   0,   0 },    241.27278,   0.00001,     0.297 },
 { {   0,   1,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    269.30771,   0.00002,     0.047 },
 { {   0,   1,   2,   0,  -2,   0,   0,   0,  -2,   2,   0 },     89.30771,   0.00002,     0.191 },
 { {   0,   1,   2,   0,  -2,   0,   0,   0,   0,   0,   0 },     90.17411,   0.00001,     0.289 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },    275.67863,   0.00009,     0.047 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,  -2,   1,   0 },    282.04749,   0.00004,     0.127 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   1 },    122.28630,   0.00002,     0.214 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    305.59708,   0.00003,     0.104 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    293.10873,   0.00020,     0.276 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    291.98587,   0.00002,     0.059 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   1,   0,  -1 },    305.43253,   0.00002,     0.389 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    219.45918,   0.00003,     0.088 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },     45.06026,   0.00004,     0.540 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    271.03308,   0.00003,     0.066 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,   0 },     72.87844,   0.00004,     0.035 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   1,   0 },     55.71089,   0.00001,     0.024 },
 { {   0,   1,   2,   0,   1,   0,   0,   0,   0,   0,   0 },    128.85934,   0.00002,     0.270 },
 { {   0,   1,   2,   0,   1,   0,   0,   0,   2,  -2,   0 },    126.89596,   0.00001,     0.517 },
 { {   0,   1,   2,   0,   1,   0,   0,   0,   2,  -1,   0 },    135.60273,   0.00001,     0.066 },
 { {   0,   1,   2,   0,   1,   0,   0,   0,   2,   0,   0 },    308.44416,   0.00012,     0.035 },
 { {   0,   1,   2,   3,   0,   0,   0,   0,   2,  -2,   0 },    184.68801,   0.00001,     0.290 },
 { {   0,   1,   2,   3,   0,   0,   0,   0,   2,   0,   0 },    184.68801,   0.00001,     0.033 },
 { {   0,   1,   3,  -8,   0,   0,   0,   0,   0,   0,   0 },    245.04722,   0.00001,     2.688 },
 { {   0,   1,   3,  -8,   3,   0,   0,   0,   0,   0,   0 },    108.11816,   0.00001,     1.600 },
 { {   0,   1,   3,  -5,   0,   0,   0,   0,   0,   0,   0 },    150.61476,   0.00001,     0.508 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -3,   0 },    320.59981,   0.00003,     0.095 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    309.07281,   0.00005,     0.361 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    320.59981,   0.00003,     0.062 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,   0,   0 },    300.07999,   0.00005,     0.034 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.46391,   0.00002,     0.050 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    131.90671,   0.00001,     0.156 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    128.21312,   0.00002,     0.116 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    208.25808,   0.00003,     0.216 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    128.21312,   0.00002,     0.056 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   1,   0,   0 },    128.21503,   0.00002,     0.059 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },    128.14134,   0.00003,     0.351 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    306.65229,   0.00002,     0.062 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   0,  -2 },    308.21503,   0.00003,     0.397 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    129.29893,   0.00004,     0.034 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21311,   0.00002,     0.018 },
 { {   0,   1,   3,   0,   1,   0,   0,   0,   2,  -2,   0 },    114.55933,   0.00001,     0.341 },
 { {   0,   1,   3,   0,   1,   0,   0,   0,   2,   0,   0 },    294.55933,   0.00001,     0.034 },
 { {   0,   1,   4,  -7,   0,   0,   0,   0,   2,  -1,   0 },    303.84243,   0.00002,     0.075 },
 { {   0,   1,   4,  -2,   0,   0,   0,   0,   2,  -2,   0 },    266.38569,   0.00002,     0.359 },
 { {   0,   1,   4,  -2,   0,   0,   0,   0,   2,   0,   0 },    266.38569,   0.00002,     0.034 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   0,   0,   0 },    309.24482,   0.00003,     0.180 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -3,   0 },    309.24732,   0.00003,     0.105 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    128.78330,   0.00002,     0.265 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    309.24732,   0.00003,     0.059 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,   0,   0 },    128.90837,   0.00003,     0.033 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.14815,   0.00002,     0.052 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.14815,   0.00002,     0.135 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   0,  -2,   0 },    122.08319,   0.00001,     0.048 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0 },    302.08319,   0.00001,     0.178 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,  -3,   0 },     37.36269,   0.00003,     0.106 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    207.60397,   0.00005,     0.260 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },     59.94299,   0.00001,     0.058 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    343.28691,   0.00002,     0.033 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   1,   0 },     34.49560,   0.00001,     0.023 },
 { {   0,   1,   5,  -9,   0,   0,   0,   0,   2,  -1,   0 },    268.47290,   0.00003,     0.075 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,  -3,   0 },    128.30401,   0.00001,     0.106 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,  -2,   0 },    128.18457,   0.00002,     0.264 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,  -1,   0 },    128.30401,   0.00001,     0.059 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,   0,   0 },    130.34336,   0.00004,     0.033 },
 { {   0,   1,   5,   0,  -3,   0,   0,   0,   2,   0,   0 },    138.99252,   0.00002,     0.032 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   0,   0,   0 },    128.31521,   0.00001,     0.155 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   0,   2,   0 },    308.31521,   0.00001,     0.030 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -3,   0 },    308.35760,   0.00002,     0.117 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.31457,   0.00002,     0.214 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -1,   0 },    128.35760,   0.00002,     0.056 },
 { {   0,   1,   5,   0,  -1,   0,   0,   0,   2,  -2,   0 },     38.69884,   0.00001,     0.210 },
 { {   0,   1,   5,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.69884,   0.00001,     0.032 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,  -1,   0 },    152.01252,   0.00001,     0.151 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,   1,   0 },    152.01252,   0.00001,     0.050 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   2,   2,   0 },    152.01252,   0.00002,     0.017 },
 { {   0,   1,   6, -11,   0,   0,   0,   0,   2,  -3,   0 },    142.27977,   0.00002,     0.075 },
 { {   0,   1,   6, -11,   0,   0,   0,   0,   2,  -2,   0 },     67.73113,   0.00010,   335.665 },
 { {   0,   1,   6, -11,   0,   0,   0,   0,   2,  -1,   0 },    142.27977,   0.00002,     0.075 },
 { {   0,   1,   6,   0,   0,   0,   0,   0,   2,   0,   0 },    239.44258,   0.00003,     0.031 },
 { {   0,   1,   9, -20,   0,   0,   0,   0,   0,   0,   0 },    248.07808,   0.00001,   122.527 },
 { {   0,   1,  29, -36,   0,   0,   0,   0,  -2,   1,   0 },     10.30751,   0.00001,   317.190 },
 { {   0,   1,  37, -51,   0,   0,   0,   0,  -2,   1,   0 },    155.60701,   0.00002,    35.857 },
 { {   0,   1,  43, -59,   0,   0,   0,   0,   0,  -2,   0 },     97.32856,   0.00001,     0.075 },
 { {   0,   1,  43, -59,   0,   0,   0,   0,   0,  -1,   0 },     97.32856,   0.00024,  2513.483 },
 { {   0,   1,  43, -59,   0,   0,   0,   0,   0,   0,   0 },     97.32856,   0.00001,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   1,   0 },    295.93028,   0.00001,     0.026 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   2,   0 },    295.93028,   0.00009,     0.040 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   3,   0 },    290.96245,   0.00015,     0.087 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   0,   0 },    301.49373,   0.00001,     0.038 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   1,   0 },    290.92112,   0.00028,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   2,  -2 },    115.93028,   0.00002,     0.037 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   2,   0 },    295.93028,   0.00808,  3054.914 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   2,   2 },    115.93028,   0.00002,     0.037 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   3,   0 },    321.43984,   0.00015,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   4,   0 },    301.49373,   0.00001,     0.038 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   1,   0 },    298.25198,   0.00012,     0.087 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   2,   0 },    295.93028,   0.00009,     0.040 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   3,   0 },    295.93028,   0.00001,     0.026 },
 { {   0,   2, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.62468,   0.00001,     0.031 },
 { {   0,   2, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.62468,   0.00001,     0.167 },
 { {   0,   2, -10,  11,   0,   0,   0,   0,   0,   0,   0 },     87.15411,   0.00001,     1.111 },
 { {   0,   2,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    121.15815,   0.00002,     0.058 },
 { {   0,   2,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.55182,   0.00002,     0.033 },
 { {   0,   2,  -9,   8,   0,   0,   0,   0,   0,   0,   0 },     46.49670,   0.00001,     0.669 },
 { {   0,   2,  -9,  11,   0,   0,   0,   0,   0,   0,   0 },    341.36343,   0.00001,    10.043 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    213.23919,   0.00001,     0.023 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    213.23919,   0.00001,     0.062 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    200.15588,   0.00001,     0.211 },
 { {   0,   2,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.68177,   0.00004,     0.034 },
 { {   0,   2,  -8,   0,   2,   0,   0,   0,  -2,   1,   0 },    231.63715,   0.00002,     0.062 },
 { {   0,   2,  -8,   0,   2,   0,   0,   0,   0,   0,   0 },    231.64133,   0.00002,     0.218 },
 { {   0,   2,  -8,   0,   3,   0,   0,   0,  -2,   0,   0 },    221.89006,   0.00001,     0.034 },
 { {   0,   2,  -8,   0,   3,   0,   0,   0,  -2,   2,   0 },    221.89006,   0.00001,     0.367 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.90827,   0.00002,     0.035 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   1,   0 },     51.77835,   0.00003,     0.066 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.90827,   0.00002,     0.523 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   3,   0 },     51.77835,   0.00003,     0.088 },
 { {   0,   2,  -8,   5,   0,   0,   0,   0,  -2,   0,   2 },    322.26441,   0.00001,    58.829 },
 { {   0,   2,  -8,   6,   0,   0,   0,   0,   0,   0,   0 },      0.26923,   0.00001,     0.642 },
 { {   0,   2,  -8,   7,   0,   0,   0,   0,   0,   0,   0 },     17.33684,   0.00002,     0.974 },
 { {   0,   2,  -8,   7,   0,   0,   0,   0,   2,   0,   0 },      2.16435,   0.00002,     0.042 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,  -2,   0,   0 },    263.80525,   0.00002,     0.040 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,  -2,   1,   0 },    263.80525,   0.00002,     0.087 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   0,  -1,   0 },    261.44265,   0.00004,     0.076 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   0,   0,   0 },    263.80525,   0.00017,    27.621 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   0,   1,   0 },    261.44265,   0.00004,     0.075 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   2,  -1,   0 },    263.80525,   0.00002,     0.087 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   2,   0,   0 },    263.80525,   0.00002,     0.040 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.01496,   0.00002,     0.024 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    138.46032,   0.00004,     0.035 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    132.05683,   0.00004,     0.066 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    344.65927,   0.00002,     0.506 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    152.59440,   0.00005,     0.089 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    142.75571,   0.00001,     0.033 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    157.11247,   0.00005,     0.267 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    114.12874,   0.00002,     0.044 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    329.12496,   0.00005,     0.129 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,  -2,   2,   0 },     51.20817,   0.00002,     0.529 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,  -2,   3,   0 },    231.20818,   0.00001,     0.088 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,   0,  -1,   0 },    230.46802,   0.00002,     0.059 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,   0,   1,   0 },    230.46802,   0.00002,     0.104 },
 { {   0,   2,  -7,   0,   2,   0,   0,   0,  -2,   0,   0 },    291.82275,   0.00005,     0.035 },
 { {   0,   2,  -7,   2,   0,   0,   0,   0,  -2,   0,   0 },    261.98103,   0.00001,     0.036 },
 { {   0,   2,  -7,   3,   0,   0,   0,   0,  -2,   0,   0 },    288.62681,   0.00001,     0.037 },
 { {   0,   2,  -7,   3,   0,   0,   0,   0,  -2,   0,   2 },    288.97260,   0.00004,    21.566 },
 { {   0,   2,  -7,   3,   0,   0,   0,   0,  -2,   2,   0 },    288.62681,   0.00001,     2.632 },
 { {   0,   2,  -7,   6,   0,   0,   0,   0,   0,   0,   0 },      8.61985,   0.00001,     1.789 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,  -2,   0,   0 },    234.68596,   0.00006,     0.040 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,  -2,   1,   0 },    234.09721,   0.00008,     0.087 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   0,  -1,   0 },    237.69409,   0.00009,     0.075 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   0,   0,   0 },    235.31761,   0.00053,    36.814 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   0,   1,   0 },    237.66683,   0.00009,     0.076 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   2,  -1,   0 },    234.01132,   0.00009,     0.087 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   2,   0,   0 },    234.59503,   0.00006,     0.040 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },     58.43599,   0.00002,     0.036 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   1,   0 },     65.74387,   0.00001,     0.070 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   2,   0 },     65.74387,   0.00001,     0.944 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   3,   0 },    245.74387,   0.00001,     0.082 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.78521,   0.00002,     0.019 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -4,   3,   0 },     64.64068,   0.00001,     0.080 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -3,   2,   0 },    231.78521,   0.00001,     0.075 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -3,   4,   0 },    231.78521,   0.00001,     0.076 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     52.50217,   0.00010,     0.036 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    276.42256,   0.00003,     0.070 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.90565,   0.00007,     1.026 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    237.35391,   0.00008,     0.081 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.90829,   0.00001,     0.062 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    324.57632,   0.00005,     0.364 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     31.25567,   0.00004,     0.095 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    214.43470,   0.00002,     0.045 },
 { {   0,   2,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },    324.96030,   0.00003,     0.036 },
 { {   0,   2,  -6,   0,   1,   0,   0,   0,  -2,   2,   0 },    220.95094,   0.00001,     1.123 },
 { {   0,   2,  -6,   2,   0,   0,   0,   0,  -2,   1,   0 },    213.89208,   0.00002,     0.076 },
 { {   0,   2,  -6,   2,   0,   0,   0,   0,  -2,   2,   0 },      4.22595,   0.00001,    11.308 },
 { {   0,   2,  -6,   2,   0,   0,   0,   0,  -2,   3,   0 },    213.89208,   0.00002,     0.075 },
 { {   0,   2,  -6,   3,   0,   0,   0,   0,   0,   0,   0 },    318.90582,   0.00002,     0.867 },
 { {   0,   2,  -6,   5,   0,   0,   0,   0,  -2,   0,   0 },    208.89421,   0.00002,     0.040 },
 { {   0,   2,  -6,   5,   0,   0,   0,   0,   0,   0,   0 },    222.98549,   0.00008,    11.046 },
 { {   0,   2,  -6,   5,   0,   0,   0,   0,   2,   0,   0 },    208.89421,   0.00001,     0.041 },
 { {   0,   2,  -6,   6,   0,   0,   0,   0,   0,   0,   0 },     56.33819,   0.00002,     2.267 },
 { {   0,   2,  -6,   8,  -3,   0,   0,   0,  -2,   0,   0 },    252.16881,   0.00002,     0.043 },
 { {   0,   2,  -6,   8,  -3,   0,   0,   0,   0,   0,   0 },    251.99727,   0.00003,     0.799 },
 { {   0,   2,  -5,  -3,   0,   0,   0,   0,  -2,   0,   0 },    355.31264,   0.00001,     0.036 },
 { {   0,   2,  -5,  -3,   0,   0,   0,   0,  -2,   1,   0 },    175.31264,   0.00001,     0.067 },
 { {   0,   2,  -5,  -3,   0,   0,   0,   0,  -2,   2,   0 },    355.31264,   0.00001,     0.637 },
 { {   0,   2,  -5,  -3,   0,   0,   0,   0,  -2,   3,   0 },    175.31264,   0.00001,     0.086 },
 { {   0,   2,  -5,  -2,   0,   0,   0,   0,  -2,   0,   0 },     11.10836,   0.00003,     0.036 },
 { {   0,   2,  -5,  -2,   0,   0,   0,   0,  -2,   2,   0 },     11.02487,   0.00002,     0.963 },
 { {   0,   2,  -5,   0,  -2,   0,   0,   0,  -2,   1,   0 },    341.49481,   0.00007,     0.075 },
 { {   0,   2,  -5,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.78912,   0.00001,     0.038 },
 { {   0,   2,  -5,   0,  -1,   0,   0,   0,  -2,   1,   0 },    151.72935,   0.00004,     0.075 },
 { {   0,   2,  -5,   0,   0,  -1,   0,   0,  -2,   1,   0 },    233.89110,   0.00003,     0.075 },
 { {   0,   2,  -5,   0,   0,  -1,   0,   0,  -2,   2,   0 },     63.36583,   0.00002,   112.558 },
 { {   0,   2,  -5,   0,   0,   0,  -2,   0,  -2,   1,   0 },    338.08050,   0.00004,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,  -2,   0,  -2,   2,   0 },    175.29949,   0.00011,   793.895 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    116.99722,   0.00002,     0.020 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },     67.66858,   0.00002,     0.026 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },     68.10832,   0.00004,     0.040 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,  -2,   0 },    133.86577,   0.00001,     0.019 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     79.93704,   0.00017,     0.038 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     67.61040,   0.00200,     0.076 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },     69.08897,   0.00133,    39.899 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },     54.54115,   0.00012,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    238.62227,   0.00007,     1.438 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    241.17739,   0.00004,     0.067 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    238.87386,   0.00037,     0.572 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    244.14903,   0.00042,     0.087 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },     62.24139,   0.00004,     0.357 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    131.45183,   0.00013,     0.103 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    146.96117,   0.00014,     0.044 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    129.72760,   0.00002,     0.028 },
 { {   0,   2,  -5,   0,   0,   1,   0,   0,  -2,   1,   0 },     48.70003,   0.00003,     0.076 },
 { {   0,   2,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },     60.00645,   0.00005,     0.076 },
 { {   0,   2,  -5,   0,   1,   0,   0,   0,  -2,   2,   0 },    230.59094,   0.00001,     9.144 },
 { {   0,   2,  -5,   0,   1,   0,   0,   0,  -2,   3,   0 },    121.48778,   0.00001,     0.075 },
 { {   0,   2,  -5,   0,   2,   0,   0,   0,  -2,   1,   0 },    270.71456,   0.00002,     0.077 },
 { {   0,   2,  -5,   2,   0,   0,   0,   0,   0,   0,   0 },    286.81265,   0.00001,     1.459 },
 { {   0,   2,  -5,   3,   0,   0,   0,   0,   0,   0,   0 },    187.79906,   0.00005,     6.498 },
 { {   0,   2,  -5,   4,   0,   0,   0,   0,   0,  -1,   0 },    209.59761,   0.00002,     0.078 },
 { {   0,   2,  -5,   4,   0,   0,   0,   0,   0,   1,   0 },    209.59761,   0.00002,     0.073 },
 { {   0,   2,  -5,   4,   0,   0,   0,   0,   2,  -2,   0 },     29.59761,   0.00001,     0.716 },
 { {   0,   2,  -5,   4,   0,   0,   0,   0,   2,   0,   0 },     29.59761,   0.00001,     0.040 },
 { {   0,   2,  -5,   5,   0,   0,   0,   0,   0,   0,   0 },     57.52426,   0.00002,     1.100 },
 { {   0,   2,  -5,   6,   0,   0,   0,   0,   0,   0,   0 },     86.33947,   0.00001,     0.694 },
 { {   0,   2,  -4,  -2,   0,   0,   0,   0,  -2,   1,   0 },    307.20658,   0.00002,     0.075 },
 { {   0,   2,  -4,  -2,   0,   0,   0,   0,  -2,   2,   0 },    109.86889,   0.00002,    26.107 },
 { {   0,   2,  -4,  -2,   0,   0,   0,   0,  -2,   3,   0 },    307.20658,   0.00002,     0.076 },
 { {   0,   2,  -4,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.84137,   0.00002,     0.039 },
 { {   0,   2,  -4,   0,  -2,   0,   0,   0,  -2,   1,   0 },     51.84137,   0.00003,     0.081 },
 { {   0,   2,  -4,   0,  -1,   0,   0,   0,   0,   0,   0 },    237.29818,   0.00001,     1.200 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },    355.59192,   0.00001,     0.020 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },    194.25785,   0.00002,     0.027 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    186.32604,   0.00003,     0.026 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    195.59847,   0.00027,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   2 },    193.07008,   0.00002,     0.736 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    325.83101,   0.00033,     0.082 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    349.18319,   0.00004,     0.976 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    187.50447,   0.00002,     0.070 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    158.33937,   0.00004,     0.038 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    350.49890,   0.00004,     0.077 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.48841,   0.00002,     0.038 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    157.37063,   0.00053,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    359.76709,   0.00007,     3.281 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    352.64615,   0.00056,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },    155.68934,   0.00015,     7.244 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    350.49890,   0.00004,     0.074 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,  -1 },    172.55335,   0.00031,    34.842 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },    182.47560,   0.00002,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   3,  -1 },    172.50249,   0.00001,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    166.44188,   0.00008,     0.037 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    152.77866,   0.00078,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    129.37164,   0.00413,     1.335 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    101.16167,   0.00060,     0.080 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },     88.76060,   0.00004,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },    337.82241,   0.00001,     0.611 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    169.60775,   0.00005,     0.085 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    349.87907,   0.00018,     0.555 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    334.76551,   0.00017,     0.086 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    169.76257,   0.00005,     0.040 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    173.00966,   0.00012,     0.087 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     28.31943,   0.00009,     0.396 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     19.21372,   0.00259,     0.093 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },     18.11640,   0.00004,     0.350 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     19.00872,   0.00226,     0.042 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },     16.50078,   0.00024,     0.027 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },     19.14295,   0.00002,     0.020 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },     16.61975,   0.00005,     0.045 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },     21.33165,   0.00007,     0.028 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },     65.87277,   0.00002,     0.021 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,   1,   0 },    191.38951,   0.00001,     0.016 },
 { {   0,   2,  -4,   0,   1,   0,   0,   0,   0,   0,   0 },    251.13038,   0.00003,     1.505 },
 { {   0,   2,  -4,   0,   1,   0,   0,   0,   2,   0,   0 },     89.35458,   0.00005,     0.042 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.90681,   0.00003,     0.039 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,  -2,   1,   0 },    179.90681,   0.00002,     0.083 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   0,  -2,   0 },    180.43371,   0.00001,     0.037 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   0,  -1,   0 },    179.82651,   0.00002,     0.072 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   0,   0,   0 },    181.31943,   0.00013,     1.723 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   0,   1,   0 },    179.82651,   0.00002,     0.079 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   0,   2,   0 },    180.27260,   0.00001,     0.039 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   2,  -1,   0 },    179.84936,   0.00003,     0.092 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },    179.84936,   0.00003,     0.041 },
 { {   0,   2,  -4,   0,   3,   0,   0,   0,   0,   0,   0 },    174.93635,   0.00002,     2.016 },
 { {   0,   2,  -4,   1,   0,   0,   0,   0,   0,   0,   0 },    276.84493,   0.00003,     4.603 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,  -2,   0,   0 },    358.68406,   0.00003,     0.041 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,  -2,   2,   0 },    356.71662,   0.00001,     0.479 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   0,  -1,   0 },    359.10399,   0.00004,     0.077 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   0,   0,   0 },      1.20360,   0.00009,     3.181 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   0,   1,   0 },    359.10399,   0.00004,     0.074 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   2,  -2,   0 },    179.45659,   0.00001,     0.685 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   2,   0,   0 },    179.91706,   0.00004,     0.040 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,  -2,   0,   0 },     57.97569,   0.00001,     0.043 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,  -2,   2,   0 },     57.97569,   0.00001,     0.317 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,   2,  -3,   0 },     61.47355,   0.00003,     0.073 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,   2,  -1,   0 },     61.47355,   0.00003,     0.078 },
 { {   0,   2,  -3,  -4,   0,   0,   0,   0,  -2,   1,   0 },    253.08880,   0.00002,     0.075 },
 { {   0,   2,  -3,  -4,   0,   0,   0,   0,  -2,   3,   0 },    253.08880,   0.00002,     0.076 },
 { {   0,   2,  -3,  -1,   0,   0,   0,   0,   0,   0,   0 },    286.82292,   0.00001,     3.563 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,  -2,   0,   0 },    300.94478,   0.00002,     0.040 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,  -2,   1,   0 },    318.41651,   0.00003,     0.087 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   0,  -1,   0 },    231.88655,   0.00004,     0.075 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   0,   0,   0 },      4.96385,   0.00003,   535.690 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   0,   1,   0 },    231.88655,   0.00004,     0.075 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   2,  -1,   0 },    318.42374,   0.00003,     0.087 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   2,   0,   0 },    300.91986,   0.00002,     0.040 },
 { {   0,   2,  -3,   0,  -2,   0,   0,   0,   0,   0,   0 },    101.06435,   0.00002,    12.131 },
 { {   0,   2,  -3,   0,  -1,   0,   0,   0,   0,  -1,   0 },    301.02922,   0.00003,     0.076 },
 { {   0,   2,  -3,   0,  -1,   0,   0,   0,   0,   0,   0 },    120.07986,   0.00003,     5.997 },
 { {   0,   2,  -3,   0,  -1,   0,   0,   0,   0,   1,   0 },    301.02922,   0.00003,     0.075 },
 { {   0,   2,  -3,   0,   0,  -1,   0,   0,   0,   0,   0 },    356.19840,   0.00001,     4.606 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -6,   1,   0 },    272.10054,   0.00002,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -6,   2,   0 },    271.16587,   0.00003,     0.021 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,  -1,   0 },    270.61327,   0.00007,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },    271.97385,   0.00053,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },    271.61743,   0.00121,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,   2 },    271.55365,   0.00002,     0.110 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   2,   0 },    271.60314,   0.00082,     0.044 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   3,   0 },    271.13561,   0.00002,     0.106 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,  -1,   0 },    271.76533,   0.00001,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   0,   1 },    256.63085,   0.00001,     0.043 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },     90.86246,   0.00005,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   1,   1 },     20.60820,   0.00002,     0.098 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -3,   0 },    271.20896,   0.00002,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -2,  -2 },     91.81624,   0.00001,     0.013 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },    271.14015,   0.00033,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -2 },     91.04291,   0.00002,     0.015 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    271.16102,   0.00395,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   2 },     90.98729,   0.00009,     0.092 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,  -2 },     90.57476,   0.00008,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.38094,   0.04214,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },    256.57767,   0.00003,     0.090 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   2 },    271.58433,   0.00082,     0.424 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,  -2 },     91.41525,   0.00014,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },    257.32478,   0.00001,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    271.56625,   0.06188,     0.089 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   2 },    270.93153,   0.00003,     0.064 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    271.64486,   0.00232,     0.494 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    271.82929,   0.00015,     0.065 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -2,   1 },    256.61086,   0.00002,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    112.47371,   0.00001,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },     90.64696,   0.00015,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    256.62461,   0.00038,     0.084 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },     76.78070,   0.00012,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },     90.62855,   0.00148,     0.083 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.86939,   0.00067,     0.766 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },     77.86533,   0.00029,     0.081 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },     91.72101,   0.00017,     0.879 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    256.87702,   0.00013,     0.069 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   2,  -1 },     76.64016,   0.00003,     1.030 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   2,   0 },     91.24277,   0.00003,     0.069 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -4,   0 },    267.79538,   0.00001,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },    268.02300,   0.00027,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,  -2 },     89.05704,   0.00003,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    268.68892,   0.00408,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,  -2 },     89.16174,   0.00016,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    257.33927,   0.00006,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    268.69648,   0.06030,     0.077 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },     76.36431,   0.00002,     2.394 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    269.82881,   0.00007,     0.072 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },     91.74862,   0.00015,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    257.32345,   0.00089,     0.076 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    269.95393,   0.34355,     3.983 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },     76.57767,   0.00016,     0.073 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   2 },     90.69380,   0.00021,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },    262.04651,   0.00012,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    256.56524,   0.00033,    11.864 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    268.56875,   0.05540,     0.074 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   2 },     88.79391,   0.00012,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,  -2 },     79.07792,   0.00001,    12.126 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,  -1 },    257.25906,   0.00002,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    268.57188,   0.00378,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   2 },     89.07711,   0.00003,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,   0 },    268.71498,   0.00025,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   4,   0 },    267.53098,   0.00001,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,  -1 },    256.57924,   0.00003,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   1 },     76.59594,   0.00008,     2.131 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    256.57648,   0.00045,     0.071 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    266.53007,   0.00006,     1.572 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },     76.65323,   0.00118,     0.078 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    256.56454,   0.00397,     1.246 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    119.92242,   0.00008,     0.079 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     77.12905,   0.00083,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    256.48123,   0.00056,     0.080 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },     87.54529,   0.00002,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,   1 },     76.99248,   0.00007,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   2,  -1 },    256.57875,   0.00003,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   2,   0 },    270.51720,   0.00001,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    271.30019,   0.00004,     0.068 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    272.13749,   0.00106,     0.657 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   2 },     90.66780,   0.00001,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    271.01820,   0.03451,     0.085 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   2 },     91.21551,   0.00007,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },    269.81514,   0.00058,     0.539 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },     77.46482,   0.00022,     0.086 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    269.98647,   0.02172,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   2 },     90.32606,   0.00005,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },     90.23846,   0.00007,     0.088 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,  -1 },     77.32478,   0.00001,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    269.38362,   0.00213,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    269.71175,   0.00021,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,  -3,  -1 },     76.63605,   0.00002,     0.063 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,  -2,   1 },     69.65221,   0.00001,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.14657,   0.00002,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0,  -1 },    256.64368,   0.00001,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },    270.04178,   0.00003,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -3,   0 },    273.67481,   0.00001,     0.100 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },    268.55872,   0.00037,     0.043 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,  -2 },    270.40941,   0.00001,     0.104 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    268.27515,   0.00057,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,  -2 },    270.51720,   0.00003,     0.044 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    267.53874,   0.00022,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   1,   0 },    271.36502,   0.00005,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   5,  -2,   0 },    270.14658,   0.00001,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   6,  -2,   0 },    275.20895,   0.00001,     0.021 },
 { {   0,   2,  -3,   0,   0,   1,   0,   0,   0,  -1,   0 },    179.78368,   0.00001,     0.077 },
 { {   0,   2,  -3,   0,   0,   1,   0,   0,   0,   0,   0 },    359.73986,   0.00001,     3.509 },
 { {   0,   2,  -3,   0,   0,   1,   0,   0,   0,   1,   0 },    179.78368,   0.00001,     0.074 },
 { {   0,   2,  -3,   0,   0,   2,   0,   0,   0,   0,   0 },     93.56181,   0.00001,     3.135 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,  -2,   0,   0 },    175.48205,   0.00009,     0.041 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,  -2,   1,   0 },      3.02982,   0.00005,     0.090 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,  -2,   2,   0 },    358.64663,   0.00001,     0.474 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   0,   0 },    359.25551,   0.00028,     2.982 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   1,  -2 },    181.35604,   0.00003,     0.075 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   2,  -2 },    134.65151,   0.00009,   544.385 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   2,   0 },    178.71738,   0.00002,     0.037 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   2,  -1,   0 },    358.56051,   0.00003,     0.085 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    177.27193,   0.00001,     0.040 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.45115,   0.00003,     0.041 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   0,   0,   0 },     55.78223,   0.00005,     2.383 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,  -2,   0 },     60.19491,   0.00003,     0.738 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,  -1,   0 },     60.30983,   0.00006,     0.084 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,   0,   0 },     60.19491,   0.00003,     0.040 },
 { {   0,   2,  -3,   0,   3,   0,   0,   0,   0,   0,   0 },     65.49288,   0.00002,     1.984 },
 { {   0,   2,  -3,   1,   0,   0,   0,   0,   0,   0,   0 },    359.26914,   0.00002,     1.278 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   0,   0 },     40.17471,   0.00001,     0.043 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   1,   0 },     40.51086,   0.00001,     0.098 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   2,   0 },     40.23663,   0.00003,     0.324 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   3,   0 },     40.51086,   0.00001,     0.061 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,   0,   0,   0 },    256.27172,   0.00002,     0.761 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,   2,  -2,   0 },     44.71905,   0.00003,     2.176 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,   2,   0,   0 },     45.36731,   0.00002,     0.038 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,  -3,   0 },    237.60405,   0.00003,     0.076 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,  -2,   0 },    236.87909,   0.00001,    13.873 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,  -1,   0 },    231.22138,   0.00005,     0.075 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,   0,   0 },    236.87909,   0.00001,     0.038 },
 { {   0,   2,  -2,  -2,   0,   0,   0,   0,   0,  -1,   0 },     47.36806,   0.00001,     0.077 },
 { {   0,   2,  -2,  -2,   0,   0,   0,   0,   0,   0,   0 },    151.68915,   0.00001,     5.328 },
 { {   0,   2,  -2,  -2,   0,   0,   0,   0,   0,   1,   0 },     47.36806,   0.00001,     0.074 },
 { {   0,   2,  -2,  -1,   0,   0,   0,   0,   1,  -3,   1 },    166.28235,   0.00001,     0.075 },
 { {   0,   2,  -2,  -1,   0,   0,   0,   0,   1,  -2,   1 },    166.28235,   0.00027,  1124.126 },
 { {   0,   2,  -2,  -1,   0,   0,   0,   0,   1,  -1,   1 },    166.28235,   0.00001,     0.075 },
 { {   0,   2,  -2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    205.43951,   0.00001,     0.042 },
 { {   0,   2,  -2,   0,  -2,   0,   0,   0,   0,   0,   0 },    341.34030,   0.00001,     0.924 },
 { {   0,   2,  -2,   0,  -2,   0,   0,   0,   1,   0,  -1 },    175.32019,   0.00001,    34.878 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    224.93197,   0.00009,     0.042 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   1,   0 },     12.46078,   0.00003,     0.097 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },      9.79677,   0.00004,     0.340 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   3,   0 },    192.04237,   0.00002,     0.062 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   0,  -2,   0 },    193.05632,   0.00001,     0.039 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   0,  -1,   0 },     11.56931,   0.00001,     0.083 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0 },     86.45017,   0.00002,     0.857 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   0,   1,   0 },     11.56931,   0.00001,     0.069 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   1,   0,  -1 },     76.96145,   0.00002,     8.852 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    193.99621,   0.00001,     1.647 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    192.20545,   0.00005,     0.079 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },    191.76544,   0.00006,     0.039 },
 { {   0,   2,  -2,   0,   0,  -1,   0,   0,   0,   0,   0 },    270.41206,   0.00001,     0.822 },
 { {   0,   2,  -2,   0,   0,  -1,   0,   0,   2,  -1,   0 },     92.48294,   0.00003,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   0,   0 },    359.92716,   0.00003,     0.014 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   1,   0 },      0.43930,   0.00003,     0.017 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   2,   0 },      0.54023,   0.00007,     0.022 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   3,   0 },      0.36763,   0.00002,     0.030 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -5,   2,   0 },    180.40373,   0.00001,     0.029 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -2,   0 },    359.90094,   0.00003,     0.013 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -1,   0 },    359.43053,   0.00020,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -1,   2 },      0.30838,   0.00001,     0.029 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,  -2 },    180.32343,   0.00001,     0.013 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },      0.70739,   0.00148,     0.021 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   2 },      0.28134,   0.00005,     0.047 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,  -2 },    180.30497,   0.00001,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },      0.58206,   0.00346,     0.029 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,   2 },      0.36811,   0.00002,     0.123 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   2,   0 },      0.49096,   0.00213,     0.046 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   3,   0 },      0.48042,   0.00005,     0.118 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,  -1,   0 },      0.43444,   0.00004,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0,   0 },      1.20965,   0.00004,     0.028 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   1,   0 },    180.30591,   0.00022,     0.044 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   2,   0 },    180.43321,   0.00008,     0.107 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -3,   0 },      0.29029,   0.00008,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.28459,   0.00084,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   2 },    180.32476,   0.00001,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    180.29238,   0.00007,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },      0.44680,   0.01120,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,   2 },    180.30006,   0.00024,     0.101 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,  -2 },    180.30278,   0.00032,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },      0.36659,   0.13432,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   2 },      0.35010,   0.00154,     0.298 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,  -2 },    180.30459,   0.00037,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.31177,   0.16090,     0.098 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   2 },    180.29156,   0.00008,     0.060 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,  -2 },    180.29900,   0.00003,     0.042 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.22764,   0.00704,     0.331 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   2 },    180.28326,   0.00003,     0.033 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },      0.23237,   0.00054,     0.061 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   4,   0 },      0.22117,   0.00005,     0.034 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -2,   0 },    180.60766,   0.00002,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },    180.11985,   0.00022,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    185.45464,   0.00003,     0.091 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    188.36537,   0.00011,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    180.06428,   0.00276,     0.090 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },      3.66832,   0.00020,     0.434 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },     11.97736,   0.00003,     0.089 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },    180.10336,   0.00029,     0.468 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   2,   0 },    179.93386,   0.00003,     0.065 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },      0.29678,   0.00025,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,  -2 },    180.23418,   0.00003,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.48600,   0.00376,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,  -2 },    180.35476,   0.00013,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },      0.39852,   0.05635,     0.083 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    180.33985,   0.00009,     0.067 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,  -2 },    359.05321,   0.00003,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },      0.20448,   0.30155,     0.799 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   2 },    180.18661,   0.00014,     0.036 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,  -2 },    358.82626,   0.00004,     0.081 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },      0.15681,   0.05071,     0.069 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   2 },    180.21408,   0.00014,     0.024 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },      0.07710,   0.00353,     0.036 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,   2 },    180.19259,   0.00002,     0.018 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   3,  -2 },    179.74152,   0.00002,     0.071 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   3,   0 },      0.22709,   0.00023,     0.024 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -2,  -1 },    187.48286,   0.00004,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,  -2 },    359.98947,   0.00002,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    187.48286,   0.00052,     0.077 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.06730,   0.00028,     2.747 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },      5.54364,   0.00019,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },    359.98872,   0.00036,     0.076 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    185.61233,   0.00334,     5.069 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },      0.05969,   0.00202,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },      5.87158,   0.00036,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -2 },    359.97178,   0.00027,    32.762 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    187.48286,   0.00052,     0.074 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,   0 },    359.85065,   0.00009,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,   1 },      5.89457,   0.00004,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   2,  -2 },      0.01147,   0.00004,     0.075 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   2,  -1 },    187.48286,   0.00004,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -4,   0 },      0.73561,   0.00003,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },      0.88037,   0.00036,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,  -2 },    179.87397,   0.00002,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    359.82862,   0.00559,     1.912 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   2 },    180.11298,   0.00002,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,  -2 },      0.07543,   0.00012,     0.071 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.99762,   0.13593,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   2 },    180.01441,   0.00029,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },    359.99290,   0.00223,     1.168 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    359.98667,   0.03553,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   2 },    180.08415,   0.00013,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,  -2 },    179.86882,   0.00011,     0.081 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },      0.00862,   0.00325,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,   2 },    180.08230,   0.00001,     0.015 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2,  -2 },    179.97657,   0.00002,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },      0.12832,   0.00031,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   3,   0 },      0.06198,   0.00002,     0.015 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -2,   0 },    180.03152,   0.00004,     0.084 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },     11.97736,   0.00003,     0.086 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.03273,   0.00006,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,  -2 },    179.98947,   0.00007,     0.087 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,  -1 },    185.39125,   0.00002,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,   0 },      0.00627,   0.00004,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -3,   0 },    358.60856,   0.00005,     0.091 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },      0.05894,   0.00181,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,  -2 },    359.89329,   0.00002,     0.094 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.01608,   0.00172,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    359.99954,   0.00049,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   1,   0 },      0.05559,   0.00007,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   5,  -3,   0 },    359.99437,   0.00001,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -3,   0 },      0.04685,   0.00002,     0.028 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -2,   0 },      0.04120,   0.00004,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.01076,   0.00005,     0.016 },
 { {   0,   2,  -2,   0,   0,   1,   0,   0,   2,  -1,   0 },     87.67754,   0.00003,     0.078 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   0,   0 },    346.48010,   0.00006,     0.043 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   1,   0 },    171.90457,   0.00004,     0.099 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   3,   0 },    352.22066,   0.00003,     0.061 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0 },    212.18312,   0.00002,     0.749 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,  -2,   0 },    349.58133,   0.00001,     2.280 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,  -1,   0 },    336.37806,   0.00007,     0.078 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    316.88751,   0.00005,     0.038 },
 { {   0,   2,  -2,   0,   2,   0,   0,   0,   0,   0,   0 },      7.31072,   0.00001,     0.704 },
 { {   0,   2,  -2,   1,   0,   0,   0,   0,   2,  -3,   0 },    226.01931,   0.00002,     0.075 },
 { {   0,   2,  -2,   1,   0,   0,   0,   0,   2,  -2,   0 },    358.89753,   0.00003,   114.769 },
 { {   0,   2,  -2,   1,   0,   0,   0,   0,   2,  -1,   0 },    226.01931,   0.00002,     0.075 },
 { {   0,   2,  -1,  -4,   0,   0,   0,   0,   0,   0,   0 },    161.41633,   0.00001,     8.045 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.12027,   0.00002,     0.042 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.12027,   0.00002,     0.412 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   0,  -1,   0 },    303.12027,   0.00001,     0.079 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   0,   1,   0 },    303.12027,   0.00001,     0.072 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   2,  -3,   0 },    303.56492,   0.00002,     0.070 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   2,  -2,   0 },    303.56492,   0.00002,     0.894 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   2,  -1,   0 },    303.56492,   0.00002,     0.082 },
 { {   0,   2,  -1,  -3,   0,   0,   0,   0,   2,   0,   0 },    303.56492,   0.00002,     0.039 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,  -2,   0,   0 },    127.73329,   0.00003,     0.042 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,  -2,   2,   0 },    135.09889,   0.00004,     0.338 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0 },    355.00834,   0.00001,     0.842 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,   2,  -2,   0 },    139.61280,   0.00003,     1.706 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,   2,   0,   0 },    139.69031,   0.00002,     0.039 },
 { {   0,   2,  -1,  -1,   0,   0,   0,   0,   2,  -3,   0 },     93.39166,   0.00001,     0.075 },
 { {   0,   2,  -1,  -1,   0,   0,   0,   0,   2,  -1,   0 },    273.39166,   0.00001,     0.076 },
 { {   0,   2,  -1,   0,  -3,   0,   0,   0,   0,   0,   0 },    129.63478,   0.00001,     0.500 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    116.19893,   0.00006,     0.044 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0 },    292.32948,   0.00004,     0.480 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,  -3,   0 },    321.18615,   0.00002,     0.077 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,  -1,   0 },    141.43973,   0.00005,     0.074 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,   0,  -2 },    170.19984,   0.00004,    39.852 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },      2.53658,   0.00003,     0.044 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,  -2,   1,   0 },    181.35098,   0.00010,     0.107 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.73204,   0.00001,     0.254 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,  -2,   0 },    185.13521,   0.00001,     0.041 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 },    176.70531,   0.00006,     0.462 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   1,   0 },    181.46864,   0.00004,     0.065 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    182.47868,   0.00002,     2.546 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    177.54832,   0.00016,     0.073 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,   0,  -2 },    183.04003,   0.00005,    16.889 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },      4.04220,   0.00009,     0.037 },
 { {   0,   2,  -1,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.21567,   0.00001,     0.044 },
 { {   0,   2,  -1,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.21567,   0.00001,     0.251 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -6,   1,   0 },    257.35926,   0.00001,     0.017 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,  -1,   0 },    257.07948,   0.00006,     0.017 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    256.11909,   0.00019,     0.021 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    257.06724,   0.00050,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },    256.51930,   0.00023,     0.048 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   3,   0 },    249.92783,   0.00001,     0.134 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   0,   0 },    257.53316,   0.00002,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.12676,   0.00002,     0.047 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,  -1 },    103.22816,   0.00001,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,   0 },     76.66489,   0.00002,     0.046 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,   1 },    103.29820,   0.00001,     0.122 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    257.14830,   0.00011,     0.020 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    257.05477,   0.00120,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   2 },     77.38636,   0.00002,     0.112 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,  -2 },     77.01752,   0.00003,     0.020 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    257.09235,   0.01345,     0.044 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   2 },    257.57055,   0.00010,     0.229 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,  -2 },     77.11750,   0.00001,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.14013,   0.01203,     0.108 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   2 },     77.25741,   0.00003,     0.057 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    258.44766,   0.00054,     0.248 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },    254.19294,   0.00003,     0.058 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.20645,   0.00004,     0.027 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.56645,   0.00010,     0.100 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.22377,   0.00055,     0.042 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.22003,   0.00132,     0.303 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    101.84842,   0.00033,     0.097 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   0 },     77.33003,   0.00001,     0.319 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   1 },    283.21235,   0.00013,     0.060 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    256.62916,   0.00019,     0.041 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.05510,   0.00262,     0.091 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -2 },     77.26835,   0.00001,     0.041 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.33020,   0.00001,     0.090 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    252.15770,   0.00873,     0.444 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },    103.32945,   0.00003,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    253.84566,   0.00147,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },    253.19634,   0.00010,     0.035 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    103.28896,   0.00004,     0.040 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,   1 },    283.31693,   0.00002,     0.653 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.29262,   0.00073,     0.083 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },     77.21289,   0.00001,     0.733 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.29445,   0.00049,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.30400,   0.00442,     0.835 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    276.74460,   0.00005,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    283.29610,   0.00049,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.45792,   0.00056,     0.069 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,   1 },    283.30465,   0.00005,     0.024 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   2,  -1 },    103.35838,   0.00003,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    261.49190,   0.00007,     0.078 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    256.29879,   0.00015,     2.096 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1 },    286.58674,   0.00003,     3.223 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    243.58589,   0.00200,     0.073 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -2 },    246.09659,   0.00020,     6.968 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.33796,   0.00024,     0.074 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     71.02165,   0.00013,     0.037 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,  -2,  -1 },    283.31134,   0.00003,     1.734 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1 },    283.32679,   0.00023,     0.079 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.30363,   0.00005,     0.039 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   4,  -2,   0 },    205.66746,   0.00001,     0.040 },
 { {   0,   2,  -1,   0,   1,   0,   0,   0,   2,  -3,   0 },     75.75072,   0.00001,     0.079 },
 { {   0,   2,  -1,   0,   1,   0,   0,   0,   2,  -1,   0 },    255.75072,   0.00001,     0.072 },
 { {   0,   2,   0,  -6,   0,   0,   0,   0,   0,   0,   0 },     68.54010,   0.00002,    16.410 },
 { {   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0,   0 },    110.67464,   0.00001,     1.687 },
 { {   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -3,   0 },    231.64130,   0.00002,     0.075 },
 { {   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -2,   0 },    151.77909,   0.00002,     8.473 },
 { {   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -1,   0 },    254.83756,   0.00004,     0.076 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   0,   0 },    359.78027,   0.00003,     0.044 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   1,   0 },    179.89419,   0.00003,     0.108 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   3,   0 },    179.89419,   0.00002,     0.058 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.89419,   0.00001,     0.041 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0 },    179.89419,   0.00001,     0.090 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },    179.89419,   0.00001,     0.457 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   0,   1,   0 },    179.89419,   0.00001,     0.065 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   2,  -3,   0 },    167.16048,   0.00001,     0.078 },
 { {   0,   2,   0,  -2,   0,   0,   0,   0,   2,  -1,   0 },    167.16048,   0.00001,     0.073 },
 { {   0,   2,   0,   0,  -3,   0,   0,   0,  -2,   0,   0 },      8.93760,   0.00004,     0.046 },
 { {   0,   2,   0,   0,  -3,   0,   0,   0,  -2,   2,   0 },      9.13343,   0.00001,     0.210 },
 { {   0,   2,   0,   0,  -3,   0,   0,   0,   0,   0,   0 },    190.88633,   0.00001,     0.334 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },    179.95400,   0.00002,     0.046 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   1,   0 },      0.11639,   0.00011,     0.119 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.30715,   0.00001,     0.206 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   3,   0 },      0.06345,   0.00002,     0.055 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,  -2,   0 },      0.06178,   0.00002,     0.043 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,  -1,   0 },      0.10574,   0.00001,     0.098 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0 },      0.00906,   0.00009,     0.324 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,   1,   0 },      0.10574,   0.00001,     0.061 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   2,  -3,   0 },      0.09365,   0.00001,     0.084 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   2,  -2,   0 },    180.05624,   0.00002,     0.764 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   2,  -1,   0 },    180.09365,   0.00001,     0.069 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   2,   0,   0 },      0.09365,   0.00001,     0.036 },
 { {   0,   2,   0,   0,  -1,   0,   0,   0,  -2,   0,   0 },     90.68973,   0.00002,     0.046 },
 { {   0,   2,   0,   0,  -1,   0,   0,   0,  -2,   2,   0 },     90.56956,   0.00003,     0.202 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },      3.40689,   0.00001,     0.022 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },    124.84287,   0.00002,     0.030 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },    140.70789,   0.00009,     0.029 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    126.54430,   0.00067,     0.047 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    126.23758,   0.00049,     0.121 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },     67.41730,   0.00002,     0.199 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },      9.19711,   0.00003,     0.044 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    204.97658,   0.00004,     0.110 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    190.30629,   0.00006,     0.232 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   1,  -1 },      9.19711,   0.00001,     0.108 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0 },     23.30462,   0.00002,     0.043 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },     43.60343,   0.00020,     0.100 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     31.13941,   0.00204,     0.308 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     30.17502,   0.00021,     0.061 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    110.15353,   0.00001,     0.034 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.52917,   0.00009,     0.455 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    205.52884,   0.00003,     0.064 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    118.52780,   0.00002,     0.085 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },      3.41174,   0.00001,     0.677 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     29.10268,   0.00039,     0.068 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    205.17307,   0.00008,     0.036 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    208.63998,   0.00001,     0.024 },
 { {   0,   2,   0,   0,   2,   0,   0,   0,   2,  -2,   0 },    308.15909,   0.00001,     0.608 },
 { {   0,   2,   0,   0,   2,   0,   0,   0,   2,   0,   0 },    128.55661,   0.00002,     0.036 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,  -2,   0,   0 },     19.93267,   0.00005,     0.040 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,  -2,   1,   0 },     19.93267,   0.00009,     0.087 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   0,  -1,   0 },     24.29953,   0.00002,     0.075 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   0,   0,   0 },      0.53309,   0.00012,   411.948 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   0,   1,   0 },     24.29953,   0.00002,     0.075 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   2,  -1,   0 },     20.00215,   0.00009,     0.087 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   2,   0,   0 },     20.00215,   0.00005,     0.040 },
 { {   0,   2,   1,  -5,   0,   0,   0,   0,   0,   0,   0 },    120.84494,   0.00002,     0.628 },
 { {   0,   2,   1,  -5,   0,   0,   0,   0,   2,  -3,   0 },    121.53824,   0.00001,     0.074 },
 { {   0,   2,   1,  -5,   0,   0,   0,   0,   2,  -1,   0 },    121.53824,   0.00001,     0.076 },
 { {   0,   2,   1,  -4,   0,   0,   0,   0,  -2,   0,   0 },    150.58257,   0.00001,     0.044 },
 { {   0,   2,   1,  -4,   0,   0,   0,   0,  -2,   2,   0 },    150.40174,   0.00002,     0.257 },
 { {   0,   2,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    269.20818,   0.00002,     0.048 },
 { {   0,   2,   1,   0,  -2,   0,   0,   0,  -2,   2,   0 },     89.20818,   0.00002,     0.171 },
 { {   0,   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },     18.29731,   0.00001,     0.240 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    334.01396,   0.00006,     0.049 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.30369,   0.00003,     0.138 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    123.54757,   0.00002,     0.189 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.81241,   0.00004,     0.111 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    295.54179,   0.00013,     0.235 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    288.18173,   0.00002,     0.057 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    306.20454,   0.00002,     0.313 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    274.63127,   0.00003,     0.064 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    113.34698,   0.00002,     0.034 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,  -3,   0 },    128.36404,   0.00004,     0.094 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,  -2,   0 },    126.86050,   0.00001,     0.390 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },    304.48471,   0.00002,     0.063 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,   0,   0 },    128.23677,   0.00007,     0.034 },
 { {   0,   2,   1,   2,   0,   0,   0,   0,   2,   0,   0 },    168.89538,   0.00001,     0.033 },
 { {   0,   2,   2, -10,   0,   0,   0,   0,   0,   0,   0 },     65.15393,   0.00001,    15.199 },
 { {   0,   2,   2,  -8,   0,   0,   0,   0,   0,   0,   0 },     60.62284,   0.00002,     1.002 },
 { {   0,   2,   2,  -8,   3,   0,   0,   0,  -2,   0,   0 },    288.33852,   0.00002,     0.043 },
 { {   0,   2,   2,  -8,   3,   0,   0,   0,   0,   0,   0 },    288.16652,   0.00003,     0.800 },
 { {   0,   2,   2,  -7,   0,   0,   0,   0,   0,   0,   0 },     92.67390,   0.00001,     0.654 },
 { {   0,   2,   2,   0,  -4,   0,   0,   0,   0,   0,   0 },    174.02942,   0.00001,     0.204 },
 { {   0,   2,   2,   0,  -2,   0,   0,   0,   2,  -2,   0 },    335.38947,   0.00001,     0.302 },
 { {   0,   2,   2,   0,  -2,   0,   0,   0,   2,   0,   0 },    335.38947,   0.00001,     0.034 },
 { {   0,   2,   2,   0,  -1,   0,   0,   0,   2,  -3,   0 },    141.30860,   0.00002,     0.101 },
 { {   0,   2,   2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    220.48647,   0.00001,     0.295 },
 { {   0,   2,   2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    141.30860,   0.00002,     0.060 },
 { {   0,   2,   2,   0,  -1,   0,   0,   0,   2,   0,   0 },    155.31746,   0.00006,     0.033 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },    131.03727,   0.00002,     0.051 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,  -2,   1,   0 },    308.24391,   0.00001,     0.160 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    128.24391,   0.00006,     0.125 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    221.30690,   0.00003,     0.190 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    128.24391,   0.00004,     0.054 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   2,   0 },    307.23895,   0.00002,     0.031 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    126.07098,   0.00006,     0.288 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    127.60822,   0.00004,     0.060 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    127.47903,   0.00011,     0.033 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   1,   0 },    128.29955,   0.00002,     0.023 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   2,   0 },    128.24391,   0.00006,     0.018 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   4,  -2,   0 },    306.07098,   0.00001,     0.035 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   4,  -1,   0 },    128.20013,   0.00001,     0.024 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   4,   0,   0 },    308.24391,   0.00001,     0.018 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   4,   1,   0 },    308.20013,   0.00001,     0.015 },
 { {   0,   2,   2,   0,   1,   0,   0,   0,   2,   0,   0 },    300.37662,   0.00004,     0.033 },
 { {   0,   2,   3,  -3,   0,   0,   0,   0,   2,  -2,   0 },    251.37255,   0.00001,     0.347 },
 { {   0,   2,   3,  -3,   0,   0,   0,   0,   2,   0,   0 },    251.37255,   0.00001,     0.034 },
 { {   0,   2,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },     84.58108,   0.00002,     0.293 },
 { {   0,   2,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },     84.58108,   0.00002,     0.033 },
 { {   0,   2,   3,   0,  -2,   0,   0,   0,   2,   0,   0 },     40.27182,   0.00001,     0.032 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   0,  -1,   0 },    309.48082,   0.00002,     0.141 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   0,   0,   0 },    129.53134,   0.00003,     0.162 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    128.78311,   0.00002,     0.228 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    129.48689,   0.00009,     0.057 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,   0,   0 },    129.48082,   0.00005,     0.032 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.09799,   0.00003,     0.054 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.09665,   0.00004,     0.125 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,  -2,   0 },    122.07128,   0.00001,     0.049 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    209.69225,   0.00003,     0.143 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     17.79855,   0.00002,     0.160 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    209.69225,   0.00003,     0.051 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   2,   0 },     25.10337,   0.00002,     0.031 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    193.38344,   0.00001,     0.032 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,  -3,   0 },    308.14647,   0.00001,     0.113 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,  -2,   0 },    308.22966,   0.00005,     0.227 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,  -1,   0 },    308.17809,   0.00002,     0.057 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,   0,   0 },    308.20545,   0.00006,     0.032 },
 { {   0,   2,   4,   0,  -3,   0,   0,   0,   2,   0,   0 },    318.10930,   0.00003,     0.032 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   0,   0,   0 },    308.37964,   0.00002,     0.141 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -3,   0 },    128.35803,   0.00002,     0.126 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.31438,   0.00002,     0.188 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -1,   0 },    128.35803,   0.00002,     0.054 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,   0,   0 },    308.33181,   0.00006,     0.031 },
 { {   0,   2,   4,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.64490,   0.00002,     0.031 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   0,  -1,   0 },    123.91039,   0.00002,     0.167 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    324.39307,   0.00001,     0.183 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    160.47021,   0.00001,     0.031 },
 { {   0,   2,   5,   0,   0,   0,   0,   0,   2,   0,   0 },    239.01520,   0.00002,     0.030 },
 { {   0,   2,  18, -19,   0,   0,   0,   0,  -2,  -1,   2 },    261.06286,   0.00001,  1216.760 },
 { {   0,   2,  21, -26,   0,   0,   0,   0,  -3,   1,   1 },    187.91231,   0.00006,  1396.538 },
 { {   0,   2,  21, -24,   0,   0,   0,   0,  -2,   1,   0 },    205.94538,   0.00001,   112.528 },
 { {   0,   2,  27, -28,   0,   0,   0,   0,   2,  -1,  -2 },    220.77143,   0.00007,  1662.840 },
 { {   0,   2,  38, -56,   0,   0,   0,   0,  -2,   1,   0 },     77.24243,   0.00002,   199.649 },
 { {   0,   3, -15,  23,   0,   0,   0,   0,   2,   0,  -2 },    135.34394,   0.00001,   449.647 },
 { {   0,   3, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.61400,   0.00001,     0.031 },
 { {   0,   3, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    121.85174,   0.00002,     0.060 },
 { {   0,   3, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    121.85174,   0.00002,     0.101 },
 { {   0,   3, -10,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.61107,   0.00002,     0.034 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    215.25255,   0.00001,     0.024 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     38.89178,   0.00002,     0.035 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    230.83874,   0.00002,     0.092 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    246.96305,   0.00002,     0.243 },
 { {   0,   3,  -9,   0,   1,   0,   0,   0,  -2,   1,   0 },    321.43776,   0.00001,     0.064 },
 { {   0,   3,  -9,   0,   1,   0,   0,   0,  -2,   3,   0 },    141.43776,   0.00001,     0.091 },
 { {   0,   3,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.67057,   0.00003,     0.035 },
 { {   0,   3,  -9,   0,   2,   0,   0,   0,  -2,   1,   0 },    231.59965,   0.00002,     0.065 },
 { {   0,   3,  -9,   0,   2,   0,   0,   0,  -2,   3,   0 },    231.59965,   0.00002,     0.090 },
 { {   0,   3,  -9,   0,   2,   0,   0,   0,   0,   0,   0 },    231.63857,   0.00002,     0.253 },
 { {   0,   3,  -9,   0,   3,   0,   0,   0,  -2,   0,   0 },    221.88730,   0.00001,     0.035 },
 { {   0,   3,  -9,   0,   3,   0,   0,   0,  -2,   2,   0 },    221.88730,   0.00001,     0.477 },
 { {   0,   3,  -9,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.90697,   0.00002,     0.036 },
 { {   0,   3,  -9,   2,   0,   0,   0,   0,  -2,   1,   0 },     51.90697,   0.00002,     0.069 },
 { {   0,   3,  -9,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.90697,   0.00002,     0.778 },
 { {   0,   3,  -9,   2,   0,   0,   0,   0,  -2,   3,   0 },     51.90697,   0.00002,     0.084 },
 { {   0,   3,  -9,   4,   0,   0,   0,   0,  -2,   1,   0 },    115.33170,   0.00001,     0.074 },
 { {   0,   3,  -9,   4,   0,   0,   0,   0,  -2,   3,   0 },    115.33170,   0.00001,     0.077 },
 { {   0,   3,  -9,   4,   0,   0,   0,   0,   0,   0,   0 },    351.35835,   0.00001,     0.501 },
 { {   0,   3,  -9,   4,   0,   0,   0,   0,   2,   0,   0 },    352.37395,   0.00002,     0.044 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    323.99749,   0.00004,     0.024 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    164.95916,   0.00002,     0.036 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     69.79296,   0.00001,     0.068 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    315.50587,   0.00003,     0.741 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    135.66770,   0.00003,     0.084 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    142.81721,   0.00001,     0.034 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    322.81721,   0.00001,     0.061 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    149.07685,   0.00003,     0.320 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    322.81721,   0.00001,     0.099 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    332.32187,   0.00003,     0.120 },
 { {   0,   3,  -8,   0,   1,   0,   0,   0,  -2,  -2,   0 },     50.46526,   0.00001,     0.018 },
 { {   0,   3,  -8,   0,   1,   0,   0,   0,  -2,   2,   0 },     51.20183,   0.00002,     0.790 },
 { {   0,   3,  -8,   0,   1,   0,   0,   0,  -2,   3,   0 },    231.20184,   0.00002,     0.083 },
 { {   0,   3,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },    291.81940,   0.00004,     0.036 },
 { {   0,   3,  -8,   0,   2,   0,   0,   0,  -2,   1,   0 },    111.89211,   0.00005,     0.069 },
 { {   0,   3,  -8,   0,   2,   0,   0,   0,  -2,   3,   0 },    292.29837,   0.00001,     0.083 },
 { {   0,   3,  -8,   2,   0,   0,   0,   0,  -2,   1,   0 },    244.90751,   0.00001,     0.074 },
 { {   0,   3,  -8,   2,   0,   0,   0,   0,  -2,   3,   0 },    244.90751,   0.00001,     0.077 },
 { {   0,   3,  -8,   3,   0,   0,   0,   0,  -2,   0,   0 },    288.62405,   0.00001,     0.038 },
 { {   0,   3,  -8,   3,   0,   0,   0,   0,  -2,   2,   0 },    288.62405,   0.00001,     4.071 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,  -2,   1,   0 },     69.34942,   0.00003,     0.080 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,  -2,   2,   0 },    250.76575,   0.00001,     1.286 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,  -1,   0 },    241.08423,   0.00002,     0.070 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,   0,   0 },    248.24262,   0.00040,     1.003 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,   1,   0 },    249.19898,   0.00015,     0.082 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,  -2,   0 },    249.84305,   0.00001,     0.361 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,  -1,   0 },    249.64800,   0.00036,     0.095 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,   0,   0 },    249.98770,   0.00030,     0.042 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,   1,   0 },    249.77644,   0.00003,     0.027 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   4,   0,   0 },     70.00990,   0.00002,     0.021 },
 { {   0,   3,  -8,   5,   0,   0,   0,   0,   0,   0,   0 },    312.58799,   0.00001,     2.151 },
 { {   0,   3,  -8,   6,   0,   0,   0,   0,   0,   0,   0 },    210.60693,   0.00005,    14.998 },
 { {   0,   3,  -8,   8,  -3,   0,   0,   0,   0,   0,   0 },    163.06266,   0.00001,     1.140 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },     58.37820,   0.00002,     0.037 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   0,   2 },     61.59035,   0.00003,     9.980 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   1,   0 },    267.24677,   0.00008,     0.073 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   2,   0 },     65.79319,   0.00001,     2.306 },
 { {   0,   3,  -7,   0,   0,  -1,   0,   0,  -2,   0,   2 },    324.22791,   0.00001,    20.062 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.78851,   0.00002,     0.019 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,  -1,   2 },     50.72777,   0.00003,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     52.81597,   0.00007,     0.037 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   2 },    295.91998,   0.00003,    62.897 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    297.14208,   0.00003,     0.074 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   2 },     50.72777,   0.00003,     0.076 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.91223,   0.00003,     2.862 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    267.97756,   0.00002,     0.077 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    208.10099,   0.00002,     0.065 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -1,   2 },     51.78851,   0.00002,     0.087 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    256.42079,   0.00006,     0.471 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    208.10099,   0.00002,     0.090 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    190.93430,   0.00009,     0.107 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    191.80962,   0.00010,     0.044 },
 { {   0,   3,  -7,   0,   0,   0,   1,   0,  -2,   0,   2 },    238.39143,   0.00001,   250.179 },
 { {   0,   3,  -7,   0,   0,   1,   0,   0,  -2,   0,   2 },    319.48502,   0.00001,    55.406 },
 { {   0,   3,  -7,   0,   0,   1,   0,   0,  -2,   1,   0 },    139.42666,   0.00003,     0.074 },
 { {   0,   3,  -7,   0,   0,   1,   0,   0,  -2,   3,   0 },    319.42666,   0.00003,     0.077 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   0,   0 },      0.68813,   0.00002,     0.037 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   0,   2 },    180.68813,   0.00002,    14.619 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   1,   0 },      2.61949,   0.00003,     0.074 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   2,   0 },     70.92436,   0.00002,     3.772 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   3,   0 },      2.61949,   0.00003,     0.077 },
 { {   0,   3,  -7,   0,   4,   0,   0,   0,  -2,   1,   0 },      4.21465,   0.00003,     0.075 },
 { {   0,   3,  -7,   0,   4,   0,   0,   0,  -2,   2,   0 },    163.49262,   0.00001,    81.999 },
 { {   0,   3,  -7,   2,   0,   0,   0,   0,  -2,   1,   0 },     24.87198,   0.00001,     0.080 },
 { {   0,   3,  -7,   2,   0,   0,   0,   0,  -2,   3,   0 },     24.87198,   0.00001,     0.072 },
 { {   0,   3,  -7,   2,   0,   0,   0,   0,  -1,   0,   1 },    357.91181,   0.00006,   158.825 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   0,   0 },    147.64087,   0.00004,     0.020 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   1,   0 },    147.53348,   0.00010,     0.028 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   2,   0 },    147.19788,   0.00006,     0.044 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,  -2,   0 },    147.64086,   0.00003,     0.020 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,  -1,   0 },    147.97531,   0.00032,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,   0 },    147.70586,   0.00338,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,   2 },    147.64085,   0.00007,     0.474 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   1,   0 },    147.32107,   0.00534,     0.087 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   2,   0 },    146.76182,   0.00020,     0.563 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -1,   0,   0 },    327.70570,   0.00010,     0.081 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -1,   1,   0 },    327.64085,   0.00001,     1.123 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -3,   0 },    150.31254,   0.00002,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -2,   0 },    149.41233,   0.00040,     0.038 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -1,  -2 },    328.88000,   0.00002,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -1,   0 },    149.42843,   0.00588,     0.075 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,  -2 },    327.00233,   0.00001,     0.037 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,   0 },    148.23103,   0.03245,   302.434 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,   2 },    327.00233,   0.00001,     0.037 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   1,   0 },    149.43192,   0.00588,     0.075 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   1,   2 },    328.88000,   0.00002,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   2,   0 },    149.41684,   0.00040,     0.038 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   3,   0 },    149.55441,   0.00003,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   1,  -1,   0 },    327.65037,   0.00001,     1.132 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   1,   0,   0 },    327.70200,   0.00010,     0.081 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -2,   0 },    146.77848,   0.00020,     0.565 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -1,   0 },    147.33055,   0.00531,     0.087 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   0,  -2 },    147.65036,   0.00007,     0.475 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   0,   0 },    147.71485,   0.00336,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   1,   0 },    147.98410,   0.00032,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   2,   0 },    147.65035,   0.00003,     0.020 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,  -2,   0 },    147.20513,   0.00006,     0.044 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,  -1,   0 },    147.53378,   0.00009,     0.028 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,   0,   0 },    147.65034,   0.00004,     0.020 },
 { {   0,   3,  -7,   5,   0,   0,   0,   0,   0,   0,   0 },    256.55257,   0.00001,     1.869 },
 { {   0,   3,  -7,   6,   0,   0,   0,   0,   1,   0,  -1 },    333.73017,   0.00002,    77.464 },
 { {   0,   3,  -7,   7,   0,   0,   0,   0,   0,   0,   0 },    267.32270,   0.00001,     0.626 },
 { {   0,   3,  -7,   7,   0,   0,   0,   0,   2,  -3,   0 },     86.41357,   0.00001,     0.074 },
 { {   0,   3,  -7,   7,   0,   0,   0,   0,   2,  -1,   0 },     86.41357,   0.00001,     0.076 },
 { {   0,   3,  -6,  -3,   0,   0,   0,   0,  -2,   0,   0 },    355.31245,   0.00001,     0.036 },
 { {   0,   3,  -6,  -3,   0,   0,   0,   0,  -2,   1,   0 },    175.31245,   0.00002,     0.070 },
 { {   0,   3,  -6,  -3,   0,   0,   0,   0,  -2,   2,   0 },    355.31245,   0.00001,     1.059 },
 { {   0,   3,  -6,  -3,   0,   0,   0,   0,  -2,   3,   0 },    175.31245,   0.00002,     0.081 },
 { {   0,   3,  -6,  -2,   0,   0,   0,   0,  -2,   0,   0 },     11.10559,   0.00003,     0.037 },
 { {   0,   3,  -6,  -2,   0,   0,   0,   0,  -2,   1,   0 },     10.66873,   0.00001,     0.073 },
 { {   0,   3,  -6,  -2,   0,   0,   0,   0,  -2,   2,   0 },     11.02041,   0.00002,     2.423 },
 { {   0,   3,  -6,  -2,   0,   0,   0,   0,  -2,   3,   0 },     10.66873,   0.00001,     0.078 },
 { {   0,   3,  -6,  -1,   0,   0,   0,   0,  -2,   1,   0 },     51.97211,   0.00002,     0.076 },
 { {   0,   3,  -6,  -1,   0,   0,   0,   0,  -2,   3,   0 },     51.97211,   0.00002,     0.075 },
 { {   0,   3,  -6,   0,  -2,   0,   0,   0,  -2,   1,   0 },      6.88882,   0.00004,     0.078 },
 { {   0,   3,  -6,   0,  -2,   0,   0,   0,  -2,   3,   0 },    180.28669,   0.00002,     0.073 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -4,   3,   0 },     55.93351,   0.00001,     0.092 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.75135,   0.00001,     0.039 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -2,   1,   0 },    236.77036,   0.00009,     0.079 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -2,   3,   0 },    232.45091,   0.00002,     0.072 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,   0,  -1,   0 },    235.75135,   0.00001,     0.069 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },    120.17613,   0.00002,     0.020 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   0,   1 },    238.81434,   0.00002,     0.040 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   1,   1 },     59.66610,   0.00003,     0.087 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    267.61811,   0.00002,     0.026 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    253.04635,   0.00005,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    241.55642,   0.00002,     0.079 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    163.81561,   0.00004,     0.072 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -2,   1 },    239.28017,   0.00002,     0.038 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     59.01215,   0.00002,     0.025 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },    239.13492,   0.00035,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },     59.01215,   0.00018,     0.037 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     23.16638,   0.00003,     0.074 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    238.81434,   0.00199,    14.355 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },     58.69281,   0.00004,     0.073 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    239.13492,   0.00035,     0.076 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   2,   0 },     84.68631,   0.00001,     0.077 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   2,   1 },    239.28017,   0.00002,     0.038 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    286.94914,   0.00001,     0.036 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    100.86536,   0.00015,     0.070 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     97.28147,   0.00129,     0.890 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     90.84710,   0.00043,     0.082 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },     87.29635,   0.00001,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },     59.66610,   0.00003,     0.088 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },     60.15302,   0.00009,     0.459 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    254.27681,   0.00002,     0.041 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.59936,   0.00008,     0.345 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.55601,   0.00153,     0.097 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },     88.76601,   0.00002,     0.310 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     90.34532,   0.00135,     0.042 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },     89.44139,   0.00016,     0.027 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.75389,   0.00003,     0.046 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },     90.21240,   0.00003,     0.028 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },     89.56666,   0.00004,     0.021 },
 { {   0,   3,  -6,   0,   0,   2,   0,   0,  -1,  -1,   1 },     74.16812,   0.00001,     0.075 },
 { {   0,   3,  -6,   0,   0,   2,   0,   0,  -1,   0,   1 },     95.03696,   0.00002,   565.470 },
 { {   0,   3,  -6,   0,   0,   2,   0,   0,  -1,   1,   1 },     74.16812,   0.00001,     0.075 },
 { {   0,   3,  -6,   0,   1,   0,   0,   0,  -1,   0,   1 },    154.83284,   0.00002,    68.309 },
 { {   0,   3,  -6,   0,   1,   0,   0,   0,   0,   0,   0 },    185.36692,   0.00002,     0.962 },
 { {   0,   3,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },    185.70081,   0.00003,     0.042 },
 { {   0,   3,  -6,   0,   2,   0,   0,   0,  -1,   0,   1 },     83.52106,   0.00001,    10.107 },
 { {   0,   3,  -6,   0,   2,   0,   0,   0,   0,   0,   0 },     69.29960,   0.00002,     1.047 },
 { {   0,   3,  -6,   0,   2,   0,   0,   0,   2,  -1,   0 },    288.72423,   0.00004,     0.095 },
 { {   0,   3,  -6,   0,   2,   0,   0,   0,   2,   0,   0 },    288.72423,   0.00003,     0.042 },
 { {   0,   3,  -6,   0,   3,   0,   0,   0,   0,   0,   0 },     59.30060,   0.00002,     1.149 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,  -2,   0,   0 },     94.49196,   0.00002,     0.040 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,  -2,   1,   0 },     94.49196,   0.00003,     0.087 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   0,  -1,   0 },     90.97182,   0.00001,     0.075 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   0,   0,   0 },     93.47569,   0.00017,    16.650 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   0,   1,   0 },     90.97182,   0.00001,     0.076 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   2,  -1,   0 },     94.68674,   0.00003,     0.088 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   2,   0,   0 },     94.68674,   0.00002,     0.041 },
 { {   0,   3,  -6,   3,   0,   0,   0,   0,   0,   0,   0 },    153.86559,   0.00002,     2.120 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,  -1,   0 },     50.45705,   0.00003,     0.027 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,   0,   0 },     49.06805,   0.00022,     0.042 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,   1,   0 },     50.36129,   0.00026,     0.095 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   0,  -1,   0 },     62.22752,   0.00005,     0.082 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   0,   0,   0 },    117.97570,   0.00009,     0.997 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   0,   1,   0 },    208.24052,   0.00003,     0.070 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   2,  -1,   0 },    224.71384,   0.00002,     0.080 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   2,   0,   0 },    219.82833,   0.00003,     0.039 },
 { {   0,   3,  -5,  -3,   0,   0,   0,   0,  -2,   1,   0 },     98.35678,   0.00004,     0.076 },
 { {   0,   3,  -5,  -3,   0,   0,   0,   0,  -2,   2,   0 },     46.82085,   0.00005,    18.007 },
 { {   0,   3,  -5,  -3,   0,   0,   0,   0,  -2,   3,   0 },     98.35678,   0.00004,     0.075 },
 { {   0,   3,  -5,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.84007,   0.00002,     0.040 },
 { {   0,   3,  -5,   0,  -2,   0,   0,   0,   0,  -1,   0 },    231.84007,   0.00002,     0.074 },
 { {   0,   3,  -5,   0,  -2,   0,   0,   0,   0,   1,   0 },    231.84007,   0.00002,     0.077 },
 { {   0,   3,  -5,   0,  -1,   0,   0,   0,   0,  -2,   0 },    216.78144,   0.00001,     0.037 },
 { {   0,   3,  -5,   0,  -1,   0,   0,   0,   0,   0,   0 },    101.04565,   0.00003,     4.814 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,  -1,   0 },    344.13331,   0.00002,     0.016 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    345.56658,   0.00015,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },    344.07999,   0.00034,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    346.99099,   0.00025,     0.043 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -3,   0,   0 },    342.27667,   0.00002,     0.027 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -3,   1,   0 },    166.64220,   0.00001,     0.042 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -2,   0 },    344.35960,   0.00011,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.99722,   0.00098,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,   2 },    178.38465,   0.00002,     0.089 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,  -2 },    165.30274,   0.00002,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    344.00427,   0.01070,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    348.40933,   0.00004,     0.087 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   2 },    344.16117,   0.00027,     0.504 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,  -2 },    165.19197,   0.00004,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    345.88283,   0.01917,     0.086 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    347.12077,   0.00079,     0.606 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    345.68671,   0.00006,     0.067 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,  -1,   0 },    165.93351,   0.00002,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },    169.84783,   0.00003,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    155.00823,   0.00028,     0.080 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    350.65074,   0.00013,     1.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.27593,   0.00003,     0.079 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   1,   0 },    162.65638,   0.00005,     1.310 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },    337.63123,   0.00006,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    338.45733,   0.00098,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,  -2 },    160.82585,   0.00004,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    338.06720,   0.01423,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   1 },    349.98796,   0.00007,    23.081 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   2 },    339.46665,   0.00003,     0.074 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,  -2 },    169.20530,   0.00007,     0.037 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    168.57698,   0.00004,     0.074 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    344.52995,   0.08277,     8.102 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    168.40932,   0.00019,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   2 },    170.50149,   0.00004,     0.037 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,  -2 },    339.46665,   0.00003,     0.073 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    338.66373,   0.01498,     0.076 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   2 },    160.82585,   0.00004,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    338.05798,   0.00100,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   3,   0 },    335.77310,   0.00006,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -2,   0 },    188.07619,   0.00002,     0.070 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    344.99950,   0.00003,     0.069 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },    101.28964,   0.00002,     0.990 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    167.36493,   0.00008,     0.081 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    347.19677,   0.00033,     0.849 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },     46.47518,   0.00022,     0.082 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    167.03510,   0.00008,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    344.99950,   0.00003,     0.083 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },     26.34752,   0.00004,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },    348.75147,   0.00006,     0.066 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    346.85523,   0.00095,     0.527 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    346.95020,   0.02299,     0.088 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   2 },    166.13531,   0.00005,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },    346.48512,   0.00030,     0.448 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    346.49874,   0.01424,     0.041 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   2 },    166.27070,   0.00003,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,  -2 },    166.12686,   0.00003,     0.091 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    346.15173,   0.00130,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    345.37025,   0.00011,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -2,   0 },     13.57567,   0.00007,     0.096 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -1,   0 },     17.44751,   0.00011,     0.042 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },      6.30678,   0.00007,     0.027 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },    349.96290,   0.00033,     0.044 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },    348.48291,   0.00043,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   0,  -2 },    346.31553,   0.00002,     0.044 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    343.79260,   0.00013,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   1,   0 },    345.73406,   0.00003,     0.016 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   6,  -1,   0 },    345.80908,   0.00001,     0.016 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,  -2,   0,   0 },    178.79426,   0.00002,     0.040 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   0,  -1,   0 },    179.67378,   0.00004,     0.075 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   0,   0,   0 },    178.71001,   0.00013,    18.007 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   0,   1,   0 },    179.67378,   0.00004,     0.076 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   2,   0,   0 },    178.81345,   0.00002,     0.041 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,  -2,   0,   0 },     87.11605,   0.00002,     0.040 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,  -2,   1,   0 },     87.11605,   0.00003,     0.087 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   0,  -1,   0 },     87.32583,   0.00004,     0.075 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   0,   0,   0 },     87.27908,   0.00017,    46.327 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   0,   1,   0 },     87.32583,   0.00004,     0.076 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   2,  -1,   0 },     87.12420,   0.00003,     0.087 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   2,   0,   0 },     87.12420,   0.00002,     0.040 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },     82.28955,   0.00010,     0.040 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },     82.27618,   0.00017,     0.087 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,  -2,   0 },     83.05196,   0.00001,     0.038 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,  -1,   0 },     82.29425,   0.00021,     0.075 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   0,   0 },     80.27411,   0.00098,    25.559 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   1,   0 },     82.46197,   0.00023,     0.076 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   2,   0 },     83.04501,   0.00001,     0.038 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,  -3,   0 },     89.65843,   0.00001,     0.066 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },     83.14199,   0.00020,     0.087 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },     84.60115,   0.00016,     0.040 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,   1,   0 },     83.99379,   0.00002,     0.026 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -4,   1,   0 },    178.98819,   0.00002,     0.028 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,  -1,   0 },    178.99840,   0.00006,     0.026 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.00260,   0.00060,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   0,   2 },    178.97104,   0.00001,     0.465 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   1,   0 },    178.64199,   0.00089,     0.087 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   2,   0 },    178.97133,   0.00004,     0.550 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -2,   0 },    180.22923,   0.00008,     0.038 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -1,   0 },    180.30021,   0.00106,     0.076 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   0,  -2 },    358.91000,   0.00001,     0.037 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   0,   0 },    178.90909,   0.00612,    22.135 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   0,   2 },    358.91000,   0.00001,     0.037 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   1,   0 },    180.31711,   0.00105,     0.075 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   2,   0 },    180.19349,   0.00008,     0.038 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,  -2,   0 },    175.36059,   0.00001,     0.578 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,  -1,   0 },    178.61060,   0.00083,     0.087 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   0,  -2 },    178.96647,   0.00001,     0.485 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   0,   0 },    178.92470,   0.00051,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   1,   0 },    178.94244,   0.00005,     0.026 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   4,  -1,   0 },    178.96647,   0.00002,     0.028 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,  -2,   0,   0 },    170.28352,   0.00002,     0.041 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,  -2,   1,   0 },    170.28352,   0.00002,     0.088 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   0,  -1,   0 },    169.26844,   0.00002,     0.076 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   0,   0,   0 },    167.78779,   0.00017,     7.723 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   0,   1,   0 },    169.26844,   0.00002,     0.075 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   2,  -1,   0 },    170.19941,   0.00004,     0.086 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   2,   0,   0 },    170.27290,   0.00002,     0.040 },
 { {   0,   3,  -5,   0,   4,   0,   0,   0,   0,   0,   0 },    158.73395,   0.00003,     4.678 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,  -2,   0,   0 },    359.16063,   0.00003,     0.042 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   0,  -1,   0 },    180.82289,   0.00002,     0.081 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   0,   0,   0 },    359.79892,   0.00010,     1.064 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   0,   1,   0 },    180.82289,   0.00002,     0.070 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   2,  -3,   0 },    179.45574,   0.00002,     0.071 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   2,  -1,   0 },    179.45574,   0.00002,     0.081 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,   2,   0,   0 },    181.82620,   0.00002,     0.039 },
 { {   0,   3,  -5,   3,   0,   0,   0,   0,   2,  -3,   0 },    210.03351,   0.00001,     0.074 },
 { {   0,   3,  -5,   3,   0,   0,   0,   0,   2,  -1,   0 },    210.03351,   0.00001,     0.077 },
 { {   0,   3,  -5,   4,   0,   0,   0,   0,  -2,   0,   0 },    312.52775,   0.00002,     0.044 },
 { {   0,   3,  -4,  -5,   0,   0,   0,   0,  -2,   2,   0 },    110.87966,   0.00002,   127.646 },
 { {   0,   3,  -4,  -3,   0,   0,   0,   0,  -1,   2,  -1 },    120.96053,   0.00001,   565.421 },
 { {   0,   3,  -4,  -2,   0,   0,   0,   0,   0,   0,   0 },    302.19030,   0.00001,     5.353 },
 { {   0,   3,  -4,  -1,   0,   0,   0,   0,   0,   0,   0 },    171.62440,   0.00002,     2.900 },
 { {   0,   3,  -4,   0,  -1,   0,   0,   0,   2,  -3,   0 },    279.70346,   0.00001,     0.070 },
 { {   0,   3,  -4,   0,  -1,   0,   0,   0,   2,  -1,   0 },     99.70346,   0.00001,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -6,   1,   0 },    271.18826,   0.00002,     0.017 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -6,   2,   0 },    270.65898,   0.00001,     0.021 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,  -1,   0 },    271.26639,   0.00008,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },    269.88491,   0.00044,     0.021 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,   2 },    270.62635,   0.00002,     0.046 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },    270.03838,   0.00100,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,   2 },    270.58473,   0.00001,     0.118 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   2,   0 },    270.03376,   0.00062,     0.045 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   3,   0 },    270.35452,   0.00002,     0.113 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   0,   0 },     90.31931,   0.00007,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   0,   1 },     76.62137,   0.00002,     0.044 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   1,   0 },     90.34284,   0.00019,     0.044 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   1,   1 },     76.58428,   0.00001,     0.105 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -3,   0 },    270.59041,   0.00002,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -2,   0 },    270.63288,   0.00026,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,  -2 },     90.60792,   0.00002,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.40307,   0.00334,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   2 },     90.59577,   0.00006,     0.097 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,  -2 },     90.60912,   0.00006,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.55741,   0.03668,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   2 },    270.59467,   0.00048,     0.335 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,  -2 },     90.62678,   0.00012,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.60642,   0.04662,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   2 },    270.71564,   0.00002,     0.062 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    270.79374,   0.00198,     0.377 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    270.72782,   0.00011,     0.063 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -2,   0 },    270.19254,   0.00008,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    270.26561,   0.00031,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },     76.45565,   0.00015,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   2 },    270.34111,   0.00002,     0.477 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -2 },    270.28216,   0.00002,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    256.58820,   0.00030,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     90.34969,   0.01197,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },     76.59968,   0.00102,     0.518 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    256.62226,   0.00028,     0.086 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },     90.33533,   0.00125,     0.567 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },     76.62822,   0.00010,     0.066 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2,  -1 },    256.58418,   0.00002,     0.626 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },     90.28287,   0.00011,     0.067 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -3,   0 },    271.42749,   0.00019,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.34748,   0.00265,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,  -2 },     90.86638,   0.00009,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.17391,   0.03945,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   2 },    272.81590,   0.00002,     0.069 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -2 },     91.09101,   0.00004,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    256.48346,   0.00007,     0.080 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    270.99316,   0.17511,     1.141 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },     76.31181,   0.00004,     0.070 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   2 },     90.87085,   0.00013,     0.036 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },    260.38588,   0.00002,     1.409 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    271.08666,   0.02911,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   2 },     90.92858,   0.00008,     0.024 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    271.11970,   0.00202,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   3,   0 },    271.05675,   0.00013,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -5,   0 },     90.31013,   0.00001,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -4,   0 },    269.95281,   0.00002,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -3,  -2 },    270.31013,   0.00001,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -3,   0 },    269.90285,   0.00028,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,  -2 },     90.16106,   0.00002,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    269.89523,   0.00405,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   1 },     92.56333,   0.00003,     6.397 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -2 },     90.35485,   0.00008,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },     77.79571,   0.00005,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },    270.34292,   0.10386,    95.796 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    255.84785,   0.00014,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   2 },     90.35485,   0.00008,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },    270.28794,   0.00012,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    256.47750,   0.00297,     5.644 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    270.28604,   0.06514,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     77.23388,   0.00047,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   2 },     90.29218,   0.00014,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     77.28880,   0.00008,     0.076 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    270.28959,   0.00443,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   1 },     76.22251,   0.00004,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   2 },     90.31013,   0.00001,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   2,   0 },    270.28758,   0.00031,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   3,   0 },    270.31013,   0.00001,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    166.94147,   0.00004,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     80.45605,   0.00042,     1.114 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,  -2 },     91.78270,   0.00002,     0.069 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },     81.61266,   0.00001,     0.940 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    282.08185,   0.00431,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },    271.82725,   0.00022,     0.812 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.52831,   0.00006,     0.082 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    272.39662,   0.00462,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,  -2 },     91.78270,   0.00002,     0.083 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    271.75114,   0.00056,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    270.76729,   0.00008,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -3,   0 },    271.33232,   0.00003,     0.560 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -2,   0 },    270.56825,   0.00149,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,  -2 },    270.34111,   0.00002,     0.472 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    256.69433,   0.00007,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,   0 },    270.33206,   0.00199,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,  -1 },    256.64516,   0.00002,     0.041 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,   0 },    270.29750,   0.00070,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   1,   0 },    270.29310,   0.00009,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -3,   0 },     93.38456,   0.00006,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },    282.78972,   0.00037,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },    281.15115,   0.00052,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    274.75562,   0.00008,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   1,   0 },     91.69537,   0.00003,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   5,  -3,   0 },    270.56524,   0.00002,     0.044 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   5,  -2,   0 },    270.40129,   0.00004,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   5,  -1,   0 },    270.32725,   0.00003,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   6,  -3,   0 },    282.98863,   0.00001,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   6,  -2,   0 },    283.10052,   0.00002,     0.021 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   0,   0 },    203.34915,   0.00002,     0.042 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   1,   0 },    354.52443,   0.00011,     0.095 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   2,   0 },     83.09588,   0.00001,     0.366 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   3,   0 },    355.98442,   0.00001,     0.063 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,  -1,   0 },    342.33430,   0.00001,     0.081 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   0,   0 },    356.76213,   0.00044,     1.041 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   1,   0 },    342.33430,   0.00001,     0.070 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   2,   0 },    178.68872,   0.00003,     0.036 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   2,  -2,   0 },    178.16590,   0.00001,     1.230 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   2,  -1,   0 },    327.78541,   0.00004,     0.080 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   2,   0,   0 },    178.16585,   0.00001,     0.039 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,  -2,   0,   0 },     83.49065,   0.00004,     0.042 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,  -2,   1,   0 },     83.49065,   0.00005,     0.096 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   0,  -1,   0 },    240.20488,   0.00001,     0.082 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   0,   0,   0 },    107.38748,   0.00004,     0.957 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   0,   1,   0 },    240.20488,   0.00001,     0.070 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,  -1,  -1 },    230.96169,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,   0,  -1 },    112.12242,   0.00006,   116.455 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,   1,  -1 },    230.96169,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   2,  -3,   0 },     60.69154,   0.00001,     0.072 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   2,  -2,   0 },     60.55203,   0.00001,     1.372 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   2,  -1,   0 },     88.41315,   0.00003,     0.080 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },     60.27259,   0.00004,     0.039 },
 { {   0,   3,  -4,   1,   0,   0,   0,   0,   2,  -3,   0 },    359.52787,   0.00001,     0.073 },
 { {   0,   3,  -4,   1,   0,   0,   0,   0,   2,  -1,   0 },    359.52787,   0.00001,     0.078 },
 { {   0,   3,  -4,   2,   0,   0,   0,   0,   2,  -3,   0 },    111.43075,   0.00002,     0.076 },
 { {   0,   3,  -4,   2,   0,   0,   0,   0,   2,  -1,   0 },    111.43075,   0.00002,     0.075 },
 { {   0,   3,  -4,   3,   0,   0,   0,   0,   2,  -3,   0 },     51.71642,   0.00001,     0.080 },
 { {   0,   3,  -4,   3,   0,   0,   0,   0,   2,  -2,   0 },    236.87633,   0.00001,     1.434 },
 { {   0,   3,  -4,   3,   0,   0,   0,   0,   2,  -1,   0 },     51.71642,   0.00001,     0.072 },
 { {   0,   3,  -4,   3,   0,   0,   0,   0,   2,   0,   0 },    236.87633,   0.00001,     0.037 },
 { {   0,   3,  -3,  -2,   0,   0,   0,   0,   0,   0,   0 },     33.60323,   0.00001,     1.230 },
 { {   0,   3,  -3,   0,  -2,   0,   0,   0,  -2,   0,   0 },    205.43932,   0.00001,     0.043 },
 { {   0,   3,  -3,   0,  -2,   0,   0,   0,   0,   0,   0 },    340.44798,   0.00001,     0.586 },
 { {   0,   3,  -3,   0,  -2,   0,   0,   0,   2,  -2,   0 },     14.76328,   0.00001,    15.139 },
 { {   0,   3,  -3,   0,  -2,   0,   0,   0,   2,  -1,   0 },     30.30952,   0.00003,     0.076 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -4,   2,   0 },     50.93853,   0.00001,     0.047 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    206.56985,   0.00008,     0.044 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   1,   0 },     13.05613,   0.00003,     0.103 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    272.29512,   0.00003,     0.280 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   3,   0 },     13.05613,   0.00002,     0.059 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   0,  -2,   0 },    193.05613,   0.00001,     0.040 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   0,  -1,   0 },     86.96348,   0.00001,     0.087 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   0,   0,   0 },     46.31070,   0.00003,     0.558 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -3,   0 },    283.22041,   0.00002,     0.076 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    214.12094,   0.00006,    54.802 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    262.56553,   0.00008,     0.075 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },    191.81054,   0.00005,     0.038 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   4,  -3,   0 },     86.96348,   0.00001,     0.087 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -3,   0 },    116.87152,   0.00005,     0.075 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -2,   0 },    122.06366,   0.00070,  1412.983 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -1,   0 },    116.87152,   0.00005,     0.075 },
 { {   0,   3,  -3,   0,   0,  -1,   0,   0,   2,  -3,   0 },     93.21748,   0.00001,     0.076 },
 { {   0,   3,  -3,   0,   0,  -1,   0,   0,   2,  -2,   0 },    272.89956,   0.00001,    14.577 },
 { {   0,   3,  -3,   0,   0,  -1,   0,   0,   2,  -1,   0 },     92.79085,   0.00007,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -6,   1,   0 },      4.00448,   0.00001,     0.017 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,  -1,   0 },    356.95714,   0.00003,     0.016 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },      7.10632,   0.00014,     0.021 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   0,   2 },      2.08724,   0.00002,     0.048 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },      4.92348,   0.00034,     0.029 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   2,   0 },      5.25703,   0.00017,     0.047 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   3,   0 },    179.59556,   0.00001,     0.128 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },    182.96974,   0.00002,     0.046 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.52395,   0.00024,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },      1.88109,   0.00215,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,   2 },    182.03987,   0.00003,     0.108 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,  -2 },    182.65012,   0.00003,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.82174,   0.01502,     0.044 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   2 },      3.19827,   0.00011,     0.251 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,  -2 },    182.69224,   0.00001,     0.027 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },      2.66594,   0.01487,     0.104 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   2 },    182.36440,   0.00003,     0.058 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },      1.64413,   0.00082,     0.274 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },      1.01269,   0.00013,     0.059 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,  -2,   0 },    177.13830,   0.00003,     0.027 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    176.84272,   0.00017,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    178.57843,   0.00003,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    340.39392,   0.00033,     0.095 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    354.24577,   0.00013,     0.341 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },      0.33464,   0.00001,     0.094 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },    359.43667,   0.00006,     0.362 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -4,   0 },      0.09727,   0.00004,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },      0.14369,   0.00063,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,  -2 },    179.98628,   0.00005,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.13999,   0.00884,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,  -2 },    180.11074,   0.00025,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },      0.12046,   0.12537,     0.088 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    179.63738,   0.00004,     0.065 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    359.76360,   0.00007,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },     11.31903,   0.00003,     0.087 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    179.09975,   0.04340,     0.533 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   2 },    358.64140,   0.00003,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },      0.12211,   0.00020,     0.085 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    179.18127,   0.00737,     0.066 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,   2 },    358.88014,   0.00003,     0.024 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,  -2 },    359.52647,   0.00002,     0.648 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    179.87451,   0.00054,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   3,   0 },    179.61127,   0.00005,     0.024 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },    174.99852,   0.00006,     0.082 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    182.09030,   0.00009,     0.080 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    183.04355,   0.00019,     1.011 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },      6.59203,   0.00006,     0.069 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,  -2 },      0.01178,   0.00003,     0.079 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    184.50964,   0.00071,     1.215 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    358.28258,   0.00226,     0.070 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },      4.69841,   0.00006,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    182.09030,   0.00009,     0.071 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    356.87102,   0.00016,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -4,   0 },      3.93814,   0.00007,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -3,  -2 },    180.58420,   0.00001,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },      3.13959,   0.00122,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,  -2 },    180.36827,   0.00013,     0.037 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },      0.04167,   0.07199,     9.751 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.99399,   0.00187,     0.074 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },    192.87703,   0.00004,    15.575 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.98144,   0.64371,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   2 },    359.97918,   0.00128,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },      0.03394,   0.00305,     4.330 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    191.31899,   0.00015,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    179.94277,   0.03030,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   2 },    359.96369,   0.00018,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },    180.01503,   0.00072,     0.077 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    179.84375,   0.00207,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,   2 },    359.98076,   0.00003,     0.015 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,  -2 },    180.01460,   0.00005,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    179.82011,   0.00008,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,  -2,   0 },    181.48858,   0.00002,     0.080 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },    359.02132,   0.00028,     0.039 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },      0.02506,   0.00008,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,   1,   0 },      0.06044,   0.00001,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -4,   0 },    358.63627,   0.00005,     0.598 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -3,   0 },    359.53643,   0.00133,     0.086 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,  -2 },      0.00847,   0.00004,     0.499 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.97859,   0.00765,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,   2 },    359.98458,   0.00005,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,  -2 },    359.99818,   0.00035,     0.089 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    179.96891,   0.00577,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,   2 },    359.98168,   0.00003,     0.015 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   0,  -2 },    359.86065,   0.00001,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    179.98428,   0.00071,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   1,   0 },    179.91368,   0.00004,     0.015 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   5,  -2,   0 },    359.98818,   0.00001,     0.027 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -4,   0 },    359.43351,   0.00002,     0.043 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -3,   0 },    180.22383,   0.00009,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -2,   0 },    179.98232,   0.00016,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -1,   0 },    179.96443,   0.00006,     0.016 },
 { {   0,   3,  -3,   0,   0,   1,   0,   0,   2,  -3,   0 },     87.68406,   0.00001,     0.076 },
 { {   0,   3,  -3,   0,   0,   1,   0,   0,   2,  -2,   0 },     87.67478,   0.00001,     7.326 },
 { {   0,   3,  -3,   0,   0,   1,   0,   0,   2,  -1,   0 },     87.68901,   0.00004,     0.075 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,  -2,   0,   0 },    328.01611,   0.00007,     0.044 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   0,  -1,   0 },    171.26545,   0.00001,     0.089 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   0,   0,   0 },    187.95884,   0.00005,     0.510 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   0,   1,   0 },    171.26545,   0.00001,     0.066 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,  -2,   0 },     37.40554,   0.00001,     5.352 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,  -1,   0 },    202.71394,   0.00004,     0.074 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    332.05920,   0.00004,     0.037 },
 { {   0,   3,  -3,   0,   2,   0,   0,   0,   2,  -3,   0 },    327.04126,   0.00001,     0.077 },
 { {   0,   3,  -3,   0,   2,   0,   0,   0,   2,  -1,   0 },    183.88643,   0.00003,     0.074 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,  -3,   0 },    303.51344,   0.00003,     0.073 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,  -2,   0 },    303.56473,   0.00001,     2.030 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    303.51344,   0.00003,     0.078 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,   0,   0 },    303.56473,   0.00001,     0.038 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   0,   0,   0 },    102.18607,   0.00001,     0.552 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   2,  -2,   0 },    125.99030,   0.00001,    25.521 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   2,  -1,   0 },    156.77626,   0.00002,     0.075 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   2,   0,   0 },    139.68199,   0.00002,     0.038 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,  -2,   0,   0 },     89.59678,   0.00002,     0.045 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,  -2,   1,   0 },    119.79153,   0.00001,     0.114 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,  -2,   2,   0 },    269.59678,   0.00002,     0.223 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,   2,  -3,   0 },    308.54545,   0.00001,     0.081 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,   2,  -1,   0 },    308.54545,   0.00001,     0.070 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },      2.31098,   0.00001,     0.046 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },      1.48892,   0.00003,     0.219 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   0,  -2,   0 },    184.99626,   0.00001,     0.042 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   0,  -1,   0 },      4.78524,   0.00002,     0.096 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0 },    179.61256,   0.00001,     0.358 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    184.78524,   0.00010,     0.070 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },      4.00442,   0.00008,     0.036 },
 { {   0,   3,  -2,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.21548,   0.00001,     0.046 },
 { {   0,   3,  -2,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.21548,   0.00001,     0.217 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },    256.56443,   0.00003,     0.021 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },    254.91802,   0.00002,     0.030 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.18224,   0.00002,     0.048 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -3,   1,   1 },    103.31586,   0.00002,     0.132 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    257.12618,   0.00002,     0.021 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    256.95967,   0.00029,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.61257,   0.00161,     0.046 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   2 },    258.14115,   0.00001,     0.201 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.16601,   0.00135,     0.116 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    255.30153,   0.00006,     0.215 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.14126,   0.00002,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.56347,   0.00009,     0.107 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.16444,   0.00042,     0.044 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.81215,   0.00003,     0.105 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.15656,   0.00087,     0.254 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.31588,   0.00022,     0.104 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },    283.23055,   0.00008,     0.058 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },    256.88451,   0.00003,     0.027 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    256.85865,   0.00047,     0.042 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    256.68904,   0.00594,     0.096 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    282.77428,   0.00004,     0.095 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    191.12153,   0.00041,     0.348 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    103.32282,   0.00003,     0.061 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    131.12366,   0.00005,     0.062 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },     78.94195,   0.00001,     0.034 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -2,   1 },    283.32371,   0.00001,     0.464 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    283.11592,   0.00005,     0.087 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.30635,   0.00026,     0.065 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.31010,   0.00274,     0.549 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    273.38652,   0.00003,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    283.29839,   0.00025,     0.035 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.30950,   0.00033,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,   1 },    283.29950,   0.00003,     0.024 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   2,  -1 },    103.30779,   0.00003,     0.035 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },     86.72780,   0.00004,     0.082 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    257.24598,   0.00046,     0.907 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    256.21778,   0.00365,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },    256.54111,   0.00006,     1.300 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.34109,   0.00008,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    257.54861,   0.00014,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,  -1 },    283.27393,   0.00043,    20.498 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,   0 },     77.04935,   0.00002,     0.074 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },    103.28746,   0.00239,     0.075 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },     77.04401,   0.00004,     0.037 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.29323,   0.00020,     0.038 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -3,   0 },    262.28163,   0.00002,     0.079 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    256.37580,   0.00010,     0.039 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    257.35289,   0.00008,     0.026 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   5,  -2,  -1 },    103.28731,   0.00003,     0.040 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   5,  -1,  -1 },    103.28742,   0.00002,     0.026 },
 { {   0,   3,  -1,  -6,   0,   0,   0,   0,   0,   0,   0 },     80.32198,   0.00001,     1.457 },
 { {   0,   3,  -1,  -4,   0,   0,   0,   0,   2,  -3,   0 },    146.94679,   0.00002,     0.075 },
 { {   0,   3,  -1,  -4,   0,   0,   0,   0,   2,  -2,   0 },    301.81664,   0.00003,    41.349 },
 { {   0,   3,  -1,  -3,   0,   0,   0,   0,   0,   0,   0 },    310.37725,   0.00001,     0.438 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.49502,   0.00003,     0.046 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   1,   0 },    179.89400,   0.00001,     0.115 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.49502,   0.00003,     0.218 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   3,   0 },    179.89400,   0.00001,     0.056 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.89400,   0.00001,     0.042 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   0,   0 },    179.89400,   0.00001,     0.355 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,  -2,  -2,   0 },    180.10555,   0.00001,     0.021 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,  -2,   1,   0 },    180.02449,   0.00001,     0.129 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.31088,   0.00001,     0.182 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.98955,   0.00003,     0.044 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,  -1,   0 },      0.10555,   0.00001,     0.105 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.98955,   0.00003,     0.270 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,   1,   0 },      0.10555,   0.00001,     0.059 },
 { {   0,   3,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },     90.49784,   0.00002,     0.048 },
 { {   0,   3,  -1,   0,  -1,   0,   0,   0,  -2,   2,   0 },     90.40749,   0.00003,     0.180 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },     39.25549,   0.00003,     0.029 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     43.92093,   0.00033,     0.048 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },     42.14167,   0.00022,     0.131 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },     13.04937,   0.00001,     0.177 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },      8.19771,   0.00002,     0.046 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    205.30930,   0.00004,     0.118 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    189.95591,   0.00004,     0.203 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },     87.78267,   0.00020,     0.107 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     26.19570,   0.00180,     0.258 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     23.69884,   0.00017,     0.058 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    180.35051,   0.00002,     0.096 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.27812,   0.00005,     0.354 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    208.71722,   0.00002,     0.062 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    182.61579,   0.00001,     0.034 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    165.63574,   0.00003,     0.090 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },     26.29583,   0.00002,     0.476 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },     35.47228,   0.00028,     0.065 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    200.50665,   0.00009,     0.035 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   1,   0 },    203.80726,   0.00001,     0.024 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   4,  -2,  -2 },    246.51321,   0.00007,   200.852 },
 { {   0,   3,   0,  -8,   3,   0,   0,   0,   0,   0,   0 },    199.22909,   0.00001,     1.142 },
 { {   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0,   0 },    120.46718,   0.00001,     0.451 },
 { {   0,   3,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },    269.02651,   0.00002,     0.050 },
 { {   0,   3,   0,   0,  -2,   0,   0,   0,  -2,   2,   0 },     89.02651,   0.00002,     0.154 },
 { {   0,   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },     19.87691,   0.00001,     0.209 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    293.59709,   0.00006,     0.050 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.94303,   0.00003,     0.151 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    124.54783,   0.00001,     0.169 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    287.50548,   0.00002,     0.119 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    291.78758,   0.00010,     0.205 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    292.04293,   0.00001,     0.055 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    307.11312,   0.00002,     0.262 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    218.63857,   0.00002,     0.098 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    218.63857,   0.00002,     0.061 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    192.36851,   0.00002,     0.034 },
 { {   0,   3,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    131.16529,   0.00006,     0.034 },
 { {   0,   3,   1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    147.83289,   0.00001,     0.249 },
 { {   0,   3,   1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    141.30841,   0.00003,     0.058 },
 { {   0,   3,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },    154.35232,   0.00003,     0.033 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    308.17522,   0.00002,     0.178 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    308.32687,   0.00002,     0.136 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    206.45084,   0.00002,     0.170 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    308.32687,   0.00002,     0.052 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    126.20660,   0.00005,     0.244 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    127.73779,   0.00006,     0.033 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   1,   0 },    308.23231,   0.00002,     0.023 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21162,   0.00002,     0.018 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   4,  -2,   0 },    306.20660,   0.00001,     0.035 },
 { {   0,   3,   1,   0,   1,   0,   0,   0,   2,  -2,   0 },    114.91505,   0.00001,     0.239 },
 { {   0,   3,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },    299.51773,   0.00003,     0.057 },
 { {   0,   3,   1,   0,   1,   0,   0,   0,   2,   0,   0 },    294.91505,   0.00001,     0.033 },
 { {   0,   3,   2,  -9,   0,   0,   0,   0,   2,   0,  -2 },    243.47824,   0.00001,    62.037 },
 { {   0,   3,   2,  -3,   0,   0,   0,   0,   2,  -2,   0 },    251.37236,   0.00001,     0.285 },
 { {   0,   3,   2,  -3,   0,   0,   0,   0,   2,   0,   0 },    251.37236,   0.00001,     0.033 },
 { {   0,   3,   2,   0,  -1,   0,   0,   0,   2,   0,   0 },    309.71109,   0.00002,     0.032 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.08239,   0.00004,     0.056 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.08098,   0.00005,     0.116 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0 },     41.79437,   0.00003,     0.145 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    211.83356,   0.00002,     0.032 },
 { {   0,   3,   3,  -2,   0,   0,   0,   0,   2,  -3,   0 },    308.14628,   0.00001,     0.122 },
 { {   0,   3,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },    308.09090,   0.00001,     0.198 },
 { {   0,   3,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    308.14628,   0.00001,     0.055 },
 { {   0,   3,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },    308.09090,   0.00001,     0.032 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    333.06063,   0.00003,     0.127 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },    325.40469,   0.00001,     0.164 },
 { {   0,   3,   5, -15,   0,   0,   0,   0,   2,  -3,   0 },    156.70371,   0.00003,     0.076 },
 { {   0,   3,   5, -15,   0,   0,   0,   0,   2,  -1,   0 },    156.70371,   0.00003,     0.075 },
 { {   0,   3,   5,   0,  -2,   0,   0,   0,  -4,   3,   0 },    184.66204,   0.00013,  4160.319 },
 { {   0,   3,   7, -19,   0,   0,   0,   0,   2,  -3,   0 },    304.87434,   0.00002,     0.075 },
 { {   0,   3,   7, -19,   0,   0,   0,   0,   2,  -2,   0 },    311.20488,   0.00025,  1812.670 },
 { {   0,   3,   7, -19,   0,   0,   0,   0,   2,  -1,   0 },    304.87434,   0.00002,     0.075 },
 { {   0,   3,  27, -37,   0,   0,   0,   0,  -1,  -1,   1 },     43.22752,   0.00001,   441.870 },
 { {   0,   3,  30, -44,   0,   0,   0,   0,  -2,   1,   0 },     38.91467,   0.00010,  1378.938 },
 { {   0,   4, -49,  55,   0,   0,   0,   0,   0,   1,   0 },    334.52157,   0.00007,  2078.312 },
 { {   0,   4, -16,  15,   0,   0,   0,   0,  -2,   1,   0 },    254.75046,   0.00001,     0.077 },
 { {   0,   4, -16,  15,   0,   0,   0,   0,  -2,   3,   0 },    254.75046,   0.00001,     0.074 },
 { {   0,   4, -15,  14,   0,   0,   0,   0,  -1,  -1,   1 },    286.27969,   0.00002,     0.075 },
 { {   0,   4, -15,  14,   0,   0,   0,   0,  -1,   0,   1 },    286.27969,   0.00030,  1778.605 },
 { {   0,   4, -15,  14,   0,   0,   0,   0,  -1,   1,   1 },    286.27969,   0.00002,     0.075 },
 { {   0,   4, -12,   7,   0,   0,   0,   0,  -2,   2,   0 },    279.68505,   0.00006,   476.157 },
 { {   0,   4, -11,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.67334,   0.00002,     0.034 },
 { {   0,   4, -11,   5,   0,   0,   0,   0,  -2,   2,   0 },    289.22781,   0.00002,    15.275 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,  -1,   0 },    217.10317,   0.00001,     0.024 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    217.47825,   0.00001,     0.035 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     98.18924,   0.00001,     0.067 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    217.47825,   0.00001,     0.580 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     66.96037,   0.00001,     0.060 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    246.96283,   0.00002,     0.286 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   1,   0 },     66.96037,   0.00001,     0.102 },
 { {   0,   4, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },    321.50003,   0.00001,     0.067 },
 { {   0,   4, -10,   0,   1,   0,   0,   0,  -2,   3,   0 },    321.50003,   0.00001,     0.086 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.66226,   0.00001,     0.036 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   1,   0 },     51.66226,   0.00001,     0.068 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   2,   0 },    231.66226,   0.00001,     0.643 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   3,   0 },     51.66226,   0.00001,     0.085 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,   0,  -2,   0 },     51.63411,   0.00001,     0.034 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,   0,   0,   0 },    231.63411,   0.00001,     0.300 },
 { {   0,   4, -10,   0,   3,   0,   0,   0,  -2,   0,   0 },    221.88283,   0.00003,     0.036 },
 { {   0,   4, -10,   0,   3,   0,   0,   0,  -2,   1,   0 },     41.88283,   0.00002,     0.068 },
 { {   0,   4, -10,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.90422,   0.00002,     0.037 },
 { {   0,   4, -10,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.90422,   0.00002,     1.514 },
 { {   0,   4, -10,   3,   0,   0,   0,   0,  -2,   1,   0 },    332.55143,   0.00002,     0.075 },
 { {   0,   4, -10,   3,   0,   0,   0,   0,  -2,   3,   0 },    332.55143,   0.00002,     0.076 },
 { {   0,   4,  -9,   0,  -1,   0,   0,   0,  -2,   1,   0 },    331.51345,   0.00003,     0.071 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,  -3,   0 },    324.00186,   0.00002,     0.015 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.00186,   0.00002,     0.025 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     20.06782,   0.00003,     0.072 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    312.08652,   0.00005,     1.381 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },     37.03815,   0.00003,     0.080 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    142.87579,   0.00001,     0.034 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    322.87579,   0.00001,     0.063 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    133.10769,   0.00004,     0.400 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    322.87579,   0.00001,     0.093 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    313.55881,   0.00003,     0.111 },
 { {   0,   4,  -9,   0,   1,   0,   0,   0,  -2,  -2,   0 },     50.46080,   0.00001,     0.019 },
 { {   0,   4,  -9,   0,   1,   0,   0,   0,  -2,   1,   0 },    115.74916,   0.00003,     0.072 },
 { {   0,   4,  -9,   0,   1,   0,   0,   0,  -2,   2,   0 },     51.19352,   0.00002,     1.564 },
 { {   0,   4,  -9,   0,   1,   0,   0,   0,   0,  -2,   0 },     50.46080,   0.00002,     0.035 },
 { {   0,   4,  -9,   0,   1,   0,   0,   0,   0,   0,   0 },     50.46080,   0.00002,     0.414 },
 { {   0,   4,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    291.84123,   0.00003,     0.037 },
 { {   0,   4,  -9,   0,   2,   0,   0,   0,  -2,   1,   0 },     32.18773,   0.00001,     0.072 },
 { {   0,   4,  -9,   0,   2,   0,   0,   0,  -2,   2,   0 },    111.76848,   0.00002,     1.801 },
 { {   0,   4,  -9,   0,   2,   0,   0,   0,  -2,   3,   0 },    112.35256,   0.00001,     0.079 },
 { {   0,   4,  -9,   1,   0,   0,   0,   0,  -2,   1,   0 },     51.33899,   0.00002,     0.074 },
 { {   0,   4,  -9,   1,   0,   0,   0,   0,  -2,   3,   0 },     51.33899,   0.00002,     0.077 },
 { {   0,   4,  -9,   2,   0,   0,   0,   0,  -2,   1,   0 },    320.07451,   0.00001,     0.077 },
 { {   0,   4,  -9,   2,   0,   0,   0,   0,  -2,   3,   0 },    320.07451,   0.00001,     0.074 },
 { {   0,   4,  -9,   3,   0,   0,   0,   0,  -2,   0,   0 },    288.61958,   0.00001,     0.039 },
 { {   0,   4,  -9,   3,   0,   0,   0,   0,  -2,   1,   0 },    288.61958,   0.00002,     0.081 },
 { {   0,   4,  -9,   3,   0,   0,   0,   0,  -2,   2,   0 },    288.61958,   0.00001,     1.148 },
 { {   0,   4,  -9,   3,   0,   0,   0,   0,  -2,   3,   0 },    288.61958,   0.00002,     0.071 },
 { {   0,   4,  -9,   5,   0,   0,   0,   0,   0,   0,   0 },     39.33194,   0.00001,     6.230 },
 { {   0,   4,  -9,   8,   0,   0,   0,   0,   2,  -3,   0 },    266.92423,   0.00002,     0.075 },
 { {   0,   4,  -9,   8,   0,   0,   0,   0,   2,  -1,   0 },    266.92423,   0.00002,     0.076 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,  -2,   0,   0 },    245.82703,   0.00001,     0.038 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,  -2,   1,   0 },    324.28442,   0.00001,     0.077 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,  -2,   3,   0 },    144.28442,   0.00001,     0.074 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,   0,  -1,   0 },    240.47421,   0.00001,     0.067 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,   0,   1,   0 },    240.47421,   0.00001,     0.086 },
 { {   0,   4,  -8,   0,   0,  -1,   0,   0,  -2,   1,   0 },    324.31545,   0.00006,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.78956,   0.00002,     0.020 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },     51.78956,   0.00002,     0.041 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     48.02229,   0.00003,     0.038 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    107.51763,   0.00004,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },     60.47962,   0.00001,     3.622 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    133.87210,   0.00001,     0.074 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    326.22191,   0.00001,     2.252 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,  -1,   2 },     51.78956,   0.00003,     0.083 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    294.86096,   0.00007,     0.668 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    177.76637,   0.00002,     0.043 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },     41.60262,   0.00003,     0.038 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,  -2,   1,   0 },    180.30031,   0.00010,     0.078 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,   0,  -1,   0 },     38.68353,   0.00001,     0.068 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,   0,   1,   0 },     38.68353,   0.00001,     0.084 },
 { {   0,   4,  -8,   2,   0,   0,   0,   0,   0,   0,   0 },    341.89298,   0.00001,     2.301 },
 { {   0,   4,  -8,   3,   0,   0,   0,   0,   0,   0,   0 },    226.96315,   0.00002,    10.294 },
 { {   0,   4,  -7,  -3,   0,   0,   0,   0,  -2,   0,   0 },    355.31115,   0.00001,     0.037 },
 { {   0,   4,  -7,  -3,   0,   0,   0,   0,  -2,   0,   2 },    103.20237,   0.00001,    68.630 },
 { {   0,   4,  -7,  -3,   0,   0,   0,   0,  -2,   2,   0 },    355.31115,   0.00001,     3.135 },
 { {   0,   4,  -7,  -2,   0,   0,   0,   0,  -2,   0,   0 },     11.22899,   0.00001,     0.038 },
 { {   0,   4,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.71193,   0.00001,     0.039 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,  -2,   0 },    132.72033,   0.00002,     0.019 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     66.78881,   0.00004,     0.040 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     57.67808,   0.00010,     0.083 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    319.84538,   0.00001,     0.784 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    244.81152,   0.00003,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },     63.13638,   0.00003,     1.799 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    241.73036,   0.00001,     0.077 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   3,  -1 },     61.73036,   0.00001,     0.074 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.52488,   0.00014,     0.073 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    241.52732,   0.00055,     2.008 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    247.71283,   0.00010,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     64.43377,   0.00006,     0.644 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },     80.06342,   0.00001,     0.440 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.19450,   0.00021,     0.091 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     88.95765,   0.00011,     0.041 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },     82.94358,   0.00002,     0.027 },
 { {   0,   4,  -7,   0,   1,   0,   0,   0,   0,   0,   0 },     12.05416,   0.00002,     2.418 },
 { {   0,   4,  -7,   0,   2,   0,   0,   0,   0,   0,   0 },     72.36345,   0.00001,     3.037 },
 { {   0,   4,  -7,   0,   3,   0,   0,   0,   0,   0,   0 },    103.38110,   0.00002,     4.082 },
 { {   0,   4,  -7,   0,   6,   0,   0,   0,  -2,   0,   0 },    157.57959,   0.00001,     0.040 },
 { {   0,   4,  -7,   0,   6,   0,   0,   0,   2,   0,   0 },    157.57959,   0.00001,     0.040 },
 { {   0,   4,  -7,   5,   0,   0,   0,   0,   0,   0,   0 },    238.19125,   0.00001,     0.463 },
 { {   0,   4,  -6,  -1,   0,   0,   0,   0,   0,   0,   0 },    238.42968,   0.00001,    33.794 },
 { {   0,   4,  -6,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.83732,   0.00002,     0.041 },
 { {   0,   4,  -6,   0,  -2,   0,   0,   0,   0,  -1,   0 },     51.83732,   0.00001,     0.077 },
 { {   0,   4,  -6,   0,  -2,   0,   0,   0,   0,   1,   0 },     51.83732,   0.00001,     0.074 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },    344.92454,   0.00004,     0.020 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   1,   0 },    344.23200,   0.00008,     0.028 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },    349.77722,   0.00004,     0.045 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -2,   0 },    345.27657,   0.00003,     0.020 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    346.34962,   0.00027,     0.027 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    345.36111,   0.00310,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   2 },    346.10899,   0.00005,     0.383 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    345.80733,   0.00411,     0.091 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    342.12819,   0.00017,     0.439 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },    354.99883,   0.00007,     0.085 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    162.35178,   0.00018,     0.084 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    354.17680,   0.00003,     0.643 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },    168.53828,   0.00002,     0.083 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },     24.04529,   0.00001,     0.068 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   3,  -1 },    348.13888,   0.00001,     0.069 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -3,   0 },    345.65490,   0.00002,     0.025 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    345.24027,   0.00039,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    342.65713,   0.00511,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,  -2 },    165.16519,   0.00001,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    348.54716,   0.00003,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    343.21324,   0.02587,     1.992 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   2 },    165.16519,   0.00001,     0.037 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    342.73542,   0.00439,     0.073 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    340.75118,   0.00028,     0.037 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   3,   0 },    339.48221,   0.00001,     0.025 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -2,   1 },    169.02593,   0.00005,     4.584 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },    349.21336,   0.00009,     0.072 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },    339.99733,   0.00005,     2.598 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    169.23207,   0.00031,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    349.41839,   0.00079,     1.813 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    340.60695,   0.00030,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    169.34230,   0.00017,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },    349.21336,   0.00009,     0.079 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },    340.44612,   0.00002,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    171.84725,   0.00008,     0.786 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    321.23219,   0.00013,     0.083 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },    344.72466,   0.00002,     0.623 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    341.92592,   0.00063,     0.040 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    344.23345,   0.00010,     0.026 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },    175.48330,   0.00002,     0.043 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    191.32337,   0.00002,     0.027 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },    165.32644,   0.00004,     0.020 },
 { {   0,   4,  -6,   0,   1,   0,   0,   0,   0,   0,   0 },     50.14859,   0.00002,     1.705 },
 { {   0,   4,  -6,   0,   1,   0,   0,   0,   2,  -2,   0 },    269.54365,   0.00003,     0.842 },
 { {   0,   4,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },     89.41278,   0.00002,     0.039 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.87138,   0.00005,     0.042 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   0,  -1,   0 },      0.31291,   0.00002,     0.079 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   0,   0,   0 },    177.04201,   0.00011,     1.491 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   0,   1,   0 },      0.31291,   0.00002,     0.072 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   2,  -3,   0 },    359.96757,   0.00001,     0.070 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   2,  -2,   0 },    359.96756,   0.00001,     0.906 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,   2,   0,   0 },    359.96756,   0.00001,     0.039 },
 { {   0,   4,  -6,   0,   3,   0,   0,   0,   0,   0,   0 },    169.45638,   0.00002,     1.325 },
 { {   0,   4,  -6,   1,   0,   0,   0,   0,   1,   0,  -1 },     75.18524,   0.00002,    50.030 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   0,  -1,   0 },    180.09795,   0.00001,     0.086 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   0,   1,   0 },    180.09795,   0.00001,     0.067 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   2,  -3,   0 },    128.67202,   0.00003,     0.074 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   2,  -1,   0 },    128.67202,   0.00003,     0.077 },
 { {   0,   4,  -6,   3,   0,   0,   0,   0,   2,   0,  -2 },    302.56492,   0.00001,    96.289 },
 { {   0,   4,  -5,  -3,   0,   0,   0,   0,   0,  -1,   0 },    327.97823,   0.00001,     0.075 },
 { {   0,   4,  -5,  -3,   0,   0,   0,   0,   0,   0,   0 },    162.57409,   0.00001,    10.758 },
 { {   0,   4,  -5,  -3,   0,   0,   0,   0,   0,   1,   0 },    327.97823,   0.00001,     0.076 },
 { {   0,   4,  -5,   0,  -1,   0,   0,   0,  -2,   0,   0 },    103.96532,   0.00001,     0.043 },
 { {   0,   4,  -5,   0,  -1,   0,   0,   0,  -2,   2,   0 },    283.96532,   0.00001,     0.313 },
 { {   0,   4,  -5,   0,  -1,   0,   0,   0,   2,  -1,   0 },     98.18988,   0.00004,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },     89.81349,   0.00003,     0.021 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },     93.55326,   0.00009,     0.029 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },     92.96538,   0.00006,     0.047 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -3,   0,   1 },     76.54321,   0.00002,     0.045 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -3,   1,   1 },     76.54286,   0.00002,     0.112 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,  -2,   0 },     94.10935,   0.00003,     0.020 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },     94.18336,   0.00020,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     94.01869,   0.00339,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   2 },     93.29249,   0.00002,     0.277 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     93.33289,   0.00410,     0.100 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },     94.90752,   0.00012,     0.305 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    270.74846,   0.00003,     0.060 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,  -2,   1 },     76.52144,   0.00001,     0.042 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     86.33834,   0.00003,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,  -1,   1 },     76.46689,   0.00004,     0.093 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },     76.52762,   0.00036,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    269.87125,   0.00006,     0.092 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.71154,   0.00083,     0.391 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   1,  -1 },     78.63936,   0.00016,     0.091 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   1,   1 },    256.53793,   0.00009,     0.063 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,  -1 },    256.36133,   0.00002,     0.450 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,   0 },    268.08870,   0.00001,     0.064 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },    272.52579,   0.00003,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.91037,   0.00034,     0.040 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.66153,   0.00541,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    256.45097,   0.00008,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    130.83992,   0.00044,     0.666 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },     76.36893,   0.00003,     0.067 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    262.23794,   0.00021,     0.068 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    279.76195,   0.00001,     0.036 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,  -1 },     76.51343,   0.00009,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,   0 },    278.41839,   0.00001,     0.079 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,   1 },    251.71225,   0.00001,     1.279 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },     76.51346,   0.00135,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },     89.42727,   0.00008,     1.626 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    256.44558,   0.00079,     0.071 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -3 },    257.15632,   0.00001,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.51091,   0.00987,     2.231 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },     88.05033,   0.00029,     0.072 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    255.84693,   0.00118,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },     76.51282,   0.00126,     0.073 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },     88.59621,   0.00002,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,   1 },    256.09624,   0.00013,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   2,  -1 },     76.51350,   0.00009,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -4,   0 },    275.49210,   0.00001,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },    299.97297,   0.00003,     0.074 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.85791,   0.00269,     3.678 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,  -2 },     91.51650,   0.00004,     0.072 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,  -1 },     76.46568,   0.00003,     2.280 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     91.39917,   0.02952,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   2 },    271.59219,   0.00006,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },     92.39326,   0.00025,     1.652 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.46313,   0.00017,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     92.15256,   0.00411,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },     92.83362,   0.00040,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },     92.22852,   0.00004,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,  -2,  -1 },    256.51527,   0.00001,     0.754 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,  -1,  -1 },    256.51602,   0.00013,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,  -1,   1 },    256.51317,   0.00003,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,  -1 },     76.51426,   0.00008,     0.040 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },     92.48061,   0.00002,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,   1 },    256.51319,   0.00002,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   1,  -1 },     76.51084,   0.00001,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -3,   0 },     93.05324,   0.00006,     0.089 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.66241,   0.00038,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },     89.11485,   0.00025,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },     96.13263,   0.00011,     0.020 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,   1,   0 },     92.64679,   0.00001,     0.016 },
 { {   0,   4,  -5,   0,   0,   1,   0,   0,   2,  -1,   0 },    359.71691,   0.00005,     0.077 },
 { {   0,   4,  -5,   0,   0,   2,   0,   0,   2,  -3,   0 },    261.38164,   0.00001,     0.074 },
 { {   0,   4,  -5,   0,   0,   2,   0,   0,   2,  -1,   0 },     81.38164,   0.00001,     0.077 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },    189.19521,   0.00010,     0.043 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },    188.10364,   0.00001,     0.101 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,  -1,   0 },    178.13818,   0.00004,     0.086 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,   0,   0 },    171.36038,   0.00003,     0.630 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,   1,   0 },    178.00947,   0.00003,     0.067 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,  -3,   0 },    359.76912,   0.00003,     0.074 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },     89.76551,   0.00005,     0.077 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },    178.71308,   0.00001,     0.038 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,   2,   0 },    358.71308,   0.00001,     0.019 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   0,   0,   0 },     62.74192,   0.00002,     0.599 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   2,  -3,   0 },    229.89415,   0.00002,     0.075 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   2,  -2,   0 },    240.31686,   0.00001,     9.679 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   2,  -1,   0 },     49.89415,   0.00002,     0.076 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   2,   0,   0 },     60.32331,   0.00004,     0.038 },
 { {   0,   4,  -5,   0,   3,   0,   0,   0,   2,  -3,   0 },     91.25535,   0.00001,     0.075 },
 { {   0,   4,  -5,   0,   3,   0,   0,   0,   2,  -2,   0 },     38.07320,   0.00002,    52.601 },
 { {   0,   4,  -5,   0,   3,   0,   0,   0,   2,  -1,   0 },    271.25535,   0.00001,     0.076 },
 { {   0,   4,  -5,   2,   0,   0,   0,   0,   2,  -3,   0 },    224.65763,   0.00004,     0.080 },
 { {   0,   4,  -5,   2,   0,   0,   0,   0,   2,  -1,   0 },    224.65763,   0.00004,     0.071 },
 { {   0,   4,  -5,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.42721,   0.00002,     0.046 },
 { {   0,   4,  -5,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.42721,   0.00002,     0.205 },
 { {   0,   4,  -4,  -8,   0,   0,   0,   0,  -2,   2,   0 },    295.86760,   0.00002,    44.178 },
 { {   0,   4,  -4,   0,  -2,   0,   0,   0,  -2,   0,   0 },    205.43802,   0.00001,     0.045 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,  -2,   0,   0 },    190.21632,   0.00005,     0.045 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0 },     13.05483,   0.00003,     0.110 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.51498,   0.00001,     0.239 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   0,  -2,   0 },    193.05483,   0.00001,     0.042 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   0,   0,   0 },     51.53623,   0.00001,     0.414 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,  -2,   0 },     14.07931,   0.00001,     1.553 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,  -1,   0 },     12.64858,   0.00006,     0.072 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,   0,   0 },    194.07931,   0.00001,     0.037 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },      0.08415,   0.00005,     0.021 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },    359.74681,   0.00009,     0.030 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   2,   0 },    359.85360,   0.00006,     0.049 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -3,   0,   0 },    357.74899,   0.00001,     0.029 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -2,   0 },    359.52272,   0.00008,     0.021 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    358.92468,   0.00037,     0.028 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.15210,   0.00469,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   2 },      0.73947,   0.00003,     0.217 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    359.20843,   0.00439,     0.111 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.08954,   0.00033,     0.234 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    358.92407,   0.00005,     0.057 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    180.03912,   0.00002,     0.043 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    255.41944,   0.00001,     0.043 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },      1.40456,   0.00014,     0.101 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },     15.51097,   0.00005,     0.281 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },    359.93932,   0.00004,     0.295 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -3,   0 },      0.79126,   0.00003,     0.027 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.57460,   0.00057,     0.042 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },      0.25543,   0.00882,     0.093 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   2 },    180.05005,   0.00002,     0.062 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,  -2 },    178.79230,   0.00010,     0.041 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    180.07081,   0.04626,     0.400 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   2 },    358.72617,   0.00007,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,  -2 },    180.09062,   0.00017,     0.090 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    180.07654,   0.00746,     0.063 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,   2 },      0.09367,   0.00003,     0.023 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    180.00993,   0.00056,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   3,   0 },    179.80978,   0.00004,     0.024 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },      0.94793,   0.00010,     0.086 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    189.95183,   0.00004,     0.085 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.67701,   0.00012,     0.619 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },      2.66834,   0.00002,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    190.90912,   0.00037,     0.690 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },      0.10228,   0.00111,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     10.94372,   0.00002,     0.035 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    189.95183,   0.00004,     0.068 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },      0.13912,   0.00006,     0.036 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    180.60757,   0.00014,     0.080 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,  -2 },    179.70368,   0.00001,     0.038 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.88331,   0.00190,     1.374 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,   2 },      0.02746,   0.00001,     0.036 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.69225,   0.00017,     0.078 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.00979,   0.08235,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   2 },      0.03058,   0.00016,     0.024 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },    180.01682,   0.00392,     2.535 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },      0.09583,   0.00909,     0.037 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   2 },    180.04491,   0.00001,     0.018 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },      0.08195,   0.00080,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    359.89201,   0.00009,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -3,  -1 },    176.99685,   0.00001,     0.074 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -3,   0 },    180.02281,   0.00005,     6.292 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -2,   0 },    180.10775,   0.00039,     0.076 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    356.99685,   0.00001,     0.077 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,   0,  -2 },    180.06452,   0.00004,     0.078 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,   0,   0 },      0.06882,   0.00004,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,   1,   0 },      0.06455,   0.00001,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -3,   0 },    178.03962,   0.00001,     0.082 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.04443,   0.00111,     0.039 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,  -2 },    180.29156,   0.00004,     0.084 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },    180.09393,   0.00053,     0.026 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,  -2 },      0.08177,   0.00002,     0.040 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    359.94088,   0.00004,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   1,   0 },      0.22960,   0.00005,     0.015 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   6,  -3,   0 },    180.01741,   0.00002,     0.027 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   6,  -2,   0 },    180.02385,   0.00002,     0.020 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.03861,   0.00002,     0.016 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,  -2,   0,   0 },    345.49724,   0.00005,     0.045 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,  -2,   1,   0 },    171.26416,   0.00003,     0.112 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   0,   0,   0 },    171.26416,   0.00002,     0.387 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   2,  -3,   0 },    168.80203,   0.00002,     0.080 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   2,  -1,   0 },    140.44758,   0.00006,     0.071 },
 { {   0,   4,  -3,  -4,   0,   0,   0,   0,   0,   0,   0 },     30.44665,   0.00001,     0.727 },
 { {   0,   4,  -3,  -3,   0,   0,   0,   0,   2,  -3,   0 },    304.59359,   0.00001,     0.076 },
 { {   0,   4,  -3,  -3,   0,   0,   0,   0,   2,  -2,   0 },    303.52601,   0.00004,     7.518 },
 { {   0,   4,  -3,  -3,   0,   0,   0,   0,   2,  -1,   0 },    304.59359,   0.00001,     0.075 },
 { {   0,   4,  -3,  -3,   0,   0,   0,   0,   2,   0,   0 },    303.56343,   0.00001,     0.038 },
 { {   0,   4,  -3,  -1,   0,   0,   0,   0,   0,   0,   0 },    180.43456,   0.00002,     0.337 },
 { {   0,   4,  -3,   0,  -2,   0,   0,   0,  -2,   0,   0 },     78.34422,   0.00001,     0.047 },
 { {   0,   4,  -3,   0,  -2,   0,   0,   0,  -2,   1,   0 },    119.79024,   0.00001,     0.123 },
 { {   0,   4,  -3,   0,  -2,   0,   0,   0,  -2,   2,   0 },     88.12706,   0.00003,     0.196 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.72242,   0.00001,     0.047 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.72242,   0.00001,     0.193 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   0,  -2,   0 },    181.70649,   0.00001,     0.043 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.09447,   0.00005,     0.293 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   0,   2,   0 },    181.13891,   0.00002,     0.033 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   2,  -3,   0 },      3.96421,   0.00002,     0.086 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    183.96421,   0.00002,     0.067 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },      3.96498,   0.00007,     0.036 },
 { {   0,   4,  -3,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.21419,   0.00001,     0.047 },
 { {   0,   4,  -3,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.21419,   0.00001,     0.191 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },     77.04944,   0.00002,     0.022 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },    252.25574,   0.00001,     0.031 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.29894,   0.00002,     0.050 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -3,   1,   1 },    103.31372,   0.00001,     0.144 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    253.41838,   0.00006,     0.029 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    251.39308,   0.00042,     0.047 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    253.26329,   0.00044,     0.125 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    273.91490,   0.00004,     0.190 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },     86.93375,   0.00001,     0.054 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    102.95531,   0.00003,     0.028 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.64678,   0.00007,     0.115 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    102.90743,   0.00033,     0.045 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.31101,   0.00059,     0.220 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.29964,   0.00017,     0.111 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    283.31055,   0.00005,     0.056 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    257.19022,   0.00006,     0.043 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    254.50173,   0.00060,     0.103 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.30511,   0.00003,     0.101 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    101.98035,   0.00048,     0.286 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    103.31637,   0.00002,     0.059 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     81.08860,   0.00010,     0.060 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.32050,   0.00029,     0.093 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.32776,   0.00017,     0.062 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.34236,   0.00196,     0.408 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },    283.32745,   0.00013,     0.034 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.32232,   0.00024,     0.064 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   1,   1 },    283.33009,   0.00002,     0.024 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   2,  -1 },    103.31921,   0.00001,     0.035 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },     80.97947,   0.00003,     0.087 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    258.84614,   0.00006,     0.579 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    259.39418,   0.00081,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },    259.84055,   0.00003,     0.717 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.34634,   0.00005,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     80.72265,   0.00025,     0.035 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },     76.58725,   0.00006,     0.024 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -2,  -1 },    103.31817,   0.00001,     1.483 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -1,  -1 },    103.32493,   0.00013,     0.072 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.32824,   0.00003,     0.037 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },     85.37853,   0.00001,     0.025 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },     95.41454,   0.00001,     0.019 },
 { {   0,   4,  -3,   0,   1,   0,   0,   0,   0,   0,   0 },     16.00514,   0.00001,     0.279 },
 { {   0,   4,  -2,  -4,   0,   0,   0,   0,   0,   0,   0 },    269.38322,   0.00001,     0.421 },
 { {   0,   4,  -2,  -4,   0,   0,   0,   0,   2,  -3,   0 },    302.01644,   0.00002,     0.079 },
 { {   0,   4,  -2,  -4,   0,   0,   0,   0,   2,  -1,   0 },    302.01644,   0.00002,     0.072 },
 { {   0,   4,  -2,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.65484,   0.00001,     0.047 },
 { {   0,   4,  -2,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.65484,   0.00001,     0.192 },
 { {   0,   4,  -2,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.89270,   0.00001,     0.043 },
 { {   0,   4,  -2,  -2,   0,   0,   0,   0,   0,   0,   0 },    179.89270,   0.00001,     0.291 },
 { {   0,   4,  -2,   0,  -3,   0,   0,   0,   0,   0,   0 },    189.69287,   0.00001,     0.235 },
 { {   0,   4,  -2,   0,  -2,   0,   0,   0,  -2,  -2,   0 },      0.10426,   0.00001,     0.021 },
 { {   0,   4,  -2,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.31092,   0.00001,     0.164 },
 { {   0,   4,  -2,   0,  -2,   0,   0,   0,   0,   0,   0 },    180.06591,   0.00002,     0.231 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },     90.16528,   0.00003,     0.049 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },     90.16528,   0.00003,     0.162 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0 },    347.76866,   0.00001,     0.226 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,   0,   2,   0 },    347.76866,   0.00001,     0.032 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },     30.76400,   0.00003,     0.030 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     31.88195,   0.00029,     0.049 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     29.68554,   0.00019,     0.143 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },     57.82761,   0.00001,     0.159 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    207.57942,   0.00003,     0.127 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    189.51236,   0.00002,     0.180 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    340.90676,   0.00002,     0.045 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     33.16800,   0.00019,     0.114 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     25.92218,   0.00147,     0.222 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     26.01936,   0.00015,     0.056 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },     11.20963,   0.00003,     0.290 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    206.08729,   0.00002,     0.059 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },     26.22489,   0.00002,     0.367 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     26.12690,   0.00024,     0.063 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    206.45540,   0.00010,     0.034 },
 { {   0,   4,  -1,  -7,   0,   0,   0,   0,   2,  -3,   0 },      2.92289,   0.00001,     0.075 },
 { {   0,   4,  -1,  -7,   0,   0,   0,   0,   2,  -2,   0 },    354.17540,   0.00003,   159.092 },
 { {   0,   4,  -1,  -7,   0,   0,   0,   0,   2,  -1,   0 },      2.92289,   0.00001,     0.075 },
 { {   0,   4,  -1,  -5,   0,   0,   0,   0,   0,   0,   0 },    121.79939,   0.00001,     0.352 },
 { {   0,   4,  -1,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.69105,   0.00001,     0.051 },
 { {   0,   4,  -1,   0,  -3,   0,   0,   0,  -2,   2,   0 },    176.69105,   0.00001,     0.142 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    288.33438,   0.00005,     0.052 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.27201,   0.00002,     0.167 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    288.80819,   0.00008,     0.182 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },    307.79836,   0.00002,     0.225 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },     37.35996,   0.00003,     0.268 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    293.32734,   0.00002,     0.059 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     72.70858,   0.00002,     0.033 },
 { {   0,   4,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },    132.61890,   0.00005,     0.033 },
 { {   0,   4,   0,   0,  -2,   0,   0,   0,   2,  -2,   0 },    153.69031,   0.00001,     0.219 },
 { {   0,   4,   0,   0,  -2,   0,   0,   0,   2,   0,   0 },    153.69031,   0.00001,     0.032 },
 { {   0,   4,   0,   0,  -1,   0,   0,   0,   2,  -2,   0 },    142.38052,   0.00002,     0.215 },
 { {   0,   4,   0,   0,  -1,   0,   0,   0,   2,  -1,   0 },    141.30711,   0.00002,     0.056 },
 { {   0,   4,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },    149.62697,   0.00004,     0.032 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    269.03362,   0.00004,     0.154 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    128.13758,   0.00001,     0.051 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    308.15282,   0.00004,     0.030 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   3,   0 },    128.13758,   0.00001,     0.022 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    308.14005,   0.00001,     0.117 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    127.98852,   0.00005,     0.212 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    129.66369,   0.00003,     0.032 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    308.13758,   0.00002,     0.022 },
 { {   0,   4,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },    299.16176,   0.00001,     0.208 },
 { {   0,   4,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    299.51644,   0.00003,     0.055 },
 { {   0,   4,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    296.64426,   0.00003,     0.032 },
 { {   0,   4,   1,   0,  -2,   0,   0,   0,   2,  -2,   0 },    225.02101,   0.00001,     0.180 },
 { {   0,   4,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },     45.02101,   0.00001,     0.031 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    129.24544,   0.00002,     0.131 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    129.24544,   0.00002,     0.053 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },    128.62194,   0.00004,     0.031 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,   2,   0 },    129.52985,   0.00001,     0.017 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.06274,   0.00004,     0.058 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.06668,   0.00005,     0.108 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     59.12941,   0.00002,     0.133 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    217.38922,   0.00001,     0.133 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    218.70738,   0.00002,     0.175 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    217.38922,   0.00001,     0.053 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    218.70738,   0.00002,     0.031 },
 { {   0,   4,   2,  -2,   0,   0,   0,   0,   2,  -2,   0 },    308.08815,   0.00001,     0.177 },
 { {   0,   4,   2,  -2,   0,   0,   0,   0,   2,   0,   0 },    308.08815,   0.00001,     0.031 },
 { {   0,   4,   2,   0,  -3,   0,   0,   0,   2,  -2,   0 },    318.10782,   0.00001,     0.154 },
 { {   0,   4,   2,   0,  -3,   0,   0,   0,   2,   0,   0 },    318.10782,   0.00001,     0.030 },
 { {   0,   4,   2,   0,  -2,   0,   0,   0,   0,   0,   0 },    308.35654,   0.00001,     0.120 },
 { {   0,   4,   2,   0,  -2,   0,   0,   0,   0,   2,   0 },    128.35654,   0.00001,     0.029 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    333.37104,   0.00002,     0.118 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    153.37104,   0.00003,     0.050 },
 { {   0,   4,   5,   0,   0,   0,   0,   0,  -2,   1,   0 },    169.96004,   0.00002,    49.129 },
 { {   0,   4,   7,   0,  -3,   0,   0,   0,   0,  -1,   0 },    163.13458,   0.00002,   152.083 },
 { {   0,   4,  22, -32,   0,   0,   0,   0,  -2,   1,   0 },    215.64594,   0.00001,   154.818 },
 { {   0,   5, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.62197,   0.00001,     0.033 },
 { {   0,   5, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.62197,   0.00001,     0.244 },
 { {   0,   5, -14,  15,   0,   0,   0,   0,   2,   0,  -2 },    337.84742,   0.00002,   214.986 },
 { {   0,   5, -13,   6,   0,   0,   0,   0,  -2,   1,   0 },    290.07513,   0.00004,     0.076 },
 { {   0,   5, -13,   6,   0,   0,   0,   0,  -2,   2,   0 },    110.09079,   0.00001,    10.901 },
 { {   0,   5, -13,   6,   0,   0,   0,   0,  -2,   3,   0 },    290.07513,   0.00004,     0.075 },
 { {   0,   5, -12,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.73484,   0.00002,     0.035 },
 { {   0,   5, -12,   4,   0,   0,   0,   0,  -2,   1,   0 },    343.45650,   0.00001,     0.076 },
 { {   0,   5, -12,   4,   0,   0,   0,   0,  -2,   3,   0 },    343.45650,   0.00001,     0.075 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    218.75432,   0.00002,     0.024 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    217.49970,   0.00001,     0.036 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    223.94577,   0.00002,     0.070 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    217.49970,   0.00001,     0.910 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },     66.96899,   0.00002,     0.062 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    266.75016,   0.00003,     0.348 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   1,   0 },     66.96899,   0.00002,     0.096 },
 { {   0,   5, -11,   0,   1,   0,   0,   0,  -2,   1,   0 },    141.90583,   0.00003,     0.070 },
 { {   0,   5, -11,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.65191,   0.00001,     0.036 },
 { {   0,   5, -11,   0,   2,   0,   0,   0,  -2,   1,   0 },     51.65192,   0.00004,     0.070 },
 { {   0,   5, -11,   0,   2,   0,   0,   0,  -2,   2,   0 },    231.65191,   0.00001,     1.076 },
 { {   0,   5, -11,   0,   3,   0,   0,   0,  -2,   0,   0 },    221.87649,   0.00002,     0.037 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.76932,   0.00004,     0.038 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   1,   0 },    184.69653,   0.00001,     0.075 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   2,   0 },    252.88815,   0.00003,    28.573 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   3,   0 },    184.69653,   0.00001,     0.076 },
 { {   0,   5, -10,   0,  -1,   0,   0,   0,  -2,   1,   0 },    156.69993,   0.00001,     0.074 },
 { {   0,   5, -10,   0,  -1,   0,   0,   0,  -2,   3,   0 },    336.69993,   0.00001,     0.076 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    308.00022,   0.00003,     0.038 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    301.75805,   0.00005,    10.166 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    144.02950,   0.00002,     0.070 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    121.17866,   0.00003,     0.534 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    237.79140,   0.00001,     0.274 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    294.83560,   0.00003,     0.104 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    274.34826,   0.00003,     0.044 },
 { {   0,   5, -10,   0,   0,   1,   0,   0,  -2,   1,   0 },    232.35572,   0.00002,     0.075 },
 { {   0,   5, -10,   0,   0,   3,   0,   0,  -2,   1,   0 },     43.35767,   0.00003,     0.075 },
 { {   0,   5, -10,   0,   0,   3,   0,   0,  -2,   3,   0 },     43.35767,   0.00003,     0.075 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },     51.18317,   0.00001,     0.038 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },    322.75829,   0.00005,     0.075 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },     88.45584,   0.00002,    71.113 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   3,   0 },    265.98929,   0.00001,     0.076 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,   0,  -1,   0 },     50.73187,   0.00002,     0.066 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,   0,   0,   0 },     50.45446,   0.00003,     0.559 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,   0,   1,   0 },     50.73187,   0.00002,     0.087 },
 { {   0,   5, -10,   0,   2,   0,   0,   0,  -2,   1,   0 },    199.28088,   0.00002,     0.076 },
 { {   0,   5, -10,   0,   2,   0,   0,   0,  -2,   2,   0 },    319.76186,   0.00002,    14.237 },
 { {   0,   5, -10,   0,   3,   0,   0,   0,  -2,   1,   0 },    287.22458,   0.00001,     0.076 },
 { {   0,   5, -10,   0,   3,   0,   0,   0,  -2,   3,   0 },    107.22458,   0.00001,     0.075 },
 { {   0,   5, -10,   3,   0,   0,   0,   0,  -2,   0,   0 },    288.61325,   0.00001,     0.040 },
 { {   0,   5, -10,   3,   0,   0,   0,   0,  -2,   2,   0 },    288.61325,   0.00001,     0.668 },
 { {   0,   5, -10,   3,   0,   0,   0,   0,   0,  -1,   0 },    108.61325,   0.00002,     0.074 },
 { {   0,   5, -10,   3,   0,   0,   0,   0,   0,   1,   0 },    108.61325,   0.00002,     0.077 },
 { {   0,   5, -10,   7,   0,   0,   0,   0,   2,  -3,   0 },     57.36439,   0.00001,     0.076 },
 { {   0,   5, -10,   7,   0,   0,   0,   0,   2,  -1,   0 },     57.36439,   0.00001,     0.075 },
 { {   0,   5,  -9,  -2,   0,   0,   0,   0,  -2,   1,   0 },    282.06778,   0.00001,     0.075 },
 { {   0,   5,  -9,  -2,   0,   0,   0,   0,  -2,   3,   0 },    282.06778,   0.00001,     0.076 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    242.12143,   0.00003,     0.039 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   1,   0 },    245.84929,   0.00001,     0.080 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   2,   0 },    234.93502,   0.00001,     1.224 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   3,   0 },     65.84929,   0.00001,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.78811,   0.00002,     0.020 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -4,   2,   0 },     54.26162,   0.00001,     0.042 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },     51.69148,   0.00001,     0.026 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     37.48569,   0.00005,     0.039 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.78811,   0.00001,     0.810 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     56.69065,   0.00006,     0.081 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },     44.85095,   0.00002,     1.109 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   2,   2 },    231.78811,   0.00001,     0.036 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },     83.21433,   0.00003,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    309.41372,   0.00013,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   1,   0 },    321.17371,   0.00015,    68.379 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   2,   0 },    232.23305,   0.00004,     0.075 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    197.52411,   0.00017,     1.146 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1,  -2 },     51.78811,   0.00002,     0.069 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    172.34970,   0.00007,     0.081 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    128.61003,   0.00002,     0.087 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    173.77229,   0.00022,     0.094 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    173.57201,   0.00020,     0.042 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },    173.89612,   0.00002,     0.027 },
 { {   0,   5,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },     41.51655,   0.00003,     0.039 },
 { {   0,   5,  -9,   0,   1,   0,   0,   0,  -2,   1,   0 },     41.89534,   0.00002,     0.082 },
 { {   0,   5,  -9,   0,   1,   0,   0,   0,  -2,   2,   0 },     43.66993,   0.00001,     1.014 },
 { {   0,   5,  -9,   0,   1,   0,   0,   0,   0,   0,   0 },    233.64020,   0.00001,     1.269 },
 { {   0,   5,  -9,   2,   0,   0,   0,   0,   0,   0,   0 },    254.95721,   0.00001,     5.236 },
 { {   0,   5,  -9,   3,   0,   0,   0,   0,   1,  -2,   1 },    188.34521,   0.00007,   413.749 },
 { {   0,   5,  -9,   5,   0,   0,   0,   0,   2,  -3,   0 },    199.05792,   0.00002,     0.076 },
 { {   0,   5,  -9,   5,   0,   0,   0,   0,   2,  -1,   0 },    199.05792,   0.00002,     0.075 },
 { {   0,   5,  -8,  -3,   0,   0,   0,   0,  -2,   0,   0 },    355.30839,   0.00001,     0.038 },
 { {   0,   5,  -8,  -3,   0,   0,   0,   0,  -2,   2,   0 },    355.30839,   0.00001,     3.262 },
 { {   0,   5,  -8,  -2,   0,   0,   0,   0,  -2,   1,   0 },     11.00576,   0.00002,     0.081 },
 { {   0,   5,  -8,  -2,   0,   0,   0,   0,  -2,   3,   0 },     11.00576,   0.00002,     0.071 },
 { {   0,   5,  -8,   0,  -2,   0,   0,   0,   0,  -1,   0 },     18.04124,   0.00003,     0.075 },
 { {   0,   5,  -8,   0,  -2,   0,   0,   0,   0,   1,   0 },     18.04124,   0.00003,     0.076 },
 { {   0,   5,  -8,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.67262,   0.00001,     0.040 },
 { {   0,   5,  -8,   0,  -1,   0,   0,   0,  -2,   1,   0 },     55.67262,   0.00003,     0.087 },
 { {   0,   5,  -8,   0,  -1,   0,   0,   0,   0,  -1,   0 },    147.01258,   0.00002,     0.076 },
 { {   0,   5,  -8,   0,  -1,   0,   0,   0,   0,   1,   0 },    147.01258,   0.00002,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },     80.54604,   0.00003,     0.020 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },     72.16531,   0.00006,     0.028 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },     86.04906,   0.00005,     0.044 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -2,   0 },     97.67962,   0.00003,     0.020 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },     71.27538,   0.00022,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.10378,   0.00231,     0.041 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   2 },     72.24117,   0.00005,     0.447 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     72.39651,   0.00337,     0.088 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },     59.62185,   0.00011,     0.526 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    251.68162,   0.00007,     0.082 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -3,   0 },     66.90891,   0.00001,     0.025 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },     67.59342,   0.00028,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.47888,   0.00418,     0.076 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,  -2 },    252.37603,   0.00001,     0.037 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    243.76431,   0.00006,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     70.23554,   0.02296,     7.836 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   2 },    252.37603,   0.00001,     0.037 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,  -1 },     63.75271,   0.00002,    25.549 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },     69.45228,   0.00403,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },     69.29890,   0.00027,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   3,   0 },     68.82349,   0.00001,     0.025 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },     67.68815,   0.00012,     1.080 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    255.04423,   0.00003,     0.080 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },    272.36718,   0.00001,     0.081 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },     73.52502,   0.00009,     0.607 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     72.29866,   0.00253,     0.086 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },     71.63609,   0.00004,     0.505 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     71.42136,   0.00166,     0.040 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },     71.03193,   0.00017,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   2,   0 },     71.63092,   0.00002,     0.019 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   3,   0,  -1 },    272.20436,   0.00001,     0.042 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },     72.85672,   0.00002,     0.043 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },     71.87079,   0.00003,     0.028 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },     71.63522,   0.00002,     0.020 },
 { {   0,   5,  -8,   0,   1,   0,   0,   0,   0,   0,   0 },    199.48777,   0.00001,     4.719 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   0,  -1,   0 },     91.99814,   0.00003,     0.077 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   0,   0,   0 },    267.14024,   0.00005,     3.376 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   0,   1,   0 },     91.99814,   0.00003,     0.074 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   2,  -1,   0 },    270.98712,   0.00002,     0.085 },
 { {   0,   5,  -8,   0,   3,   0,   0,   0,   0,   0,   0 },    193.73834,   0.00004,     2.628 },
 { {   0,   5,  -8,   2,   0,   0,   0,   0,   0,   0,   0 },     93.11365,   0.00001,     0.840 },
 { {   0,   5,  -8,   3,   0,   0,   0,   0,   2,  -3,   0 },    356.82216,   0.00003,     0.075 },
 { {   0,   5,  -8,   3,   0,   0,   0,   0,   2,  -2,   0 },    175.16167,   0.00002,    19.475 },
 { {   0,   5,  -8,   3,   0,   0,   0,   0,   2,  -1,   0 },    356.82216,   0.00003,     0.076 },
 { {   0,   5,  -8,   4,   0,   0,   0,   0,   0,   0,   0 },     30.59227,   0.00001,     0.444 },
 { {   0,   5,  -7,  -4,   0,   0,   0,   0,   0,   0,   0 },    112.33611,   0.00005,     1.001 },
 { {   0,   5,  -7,  -4,   0,   0,   0,   0,   2,  -1,   0 },    112.33521,   0.00004,     0.095 },
 { {   0,   5,  -7,  -4,   0,   0,   0,   0,   2,   0,   0 },    112.33521,   0.00004,     0.042 },
 { {   0,   5,  -7,  -3,   0,   0,   0,   0,   0,   0,   0 },     27.05346,   0.00001,     2.139 },
 { {   0,   5,  -7,  -2,   0,   0,   0,   0,   0,  -1,   0 },    166.34359,   0.00001,     0.076 },
 { {   0,   5,  -7,  -2,   0,   0,   0,   0,   0,   1,   0 },    166.34359,   0.00001,     0.075 },
 { {   0,   5,  -7,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.83285,   0.00001,     0.042 },
 { {   0,   5,  -7,   0,  -2,   0,   0,   0,  -2,   2,   0 },    231.83285,   0.00001,     0.366 },
 { {   0,   5,  -7,   0,  -1,   0,   0,   0,   1,   0,  -1 },    207.41341,   0.00002,    95.659 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },    167.94774,   0.00004,     0.021 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   1,   0 },    165.47318,   0.00007,     0.028 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },    165.11367,   0.00005,     0.046 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,  -2,   0 },    167.04578,   0.00003,     0.020 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    166.64575,   0.00023,     0.027 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    169.34099,   0.00290,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   2 },    168.29770,   0.00002,     0.309 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    168.34997,   0.00365,     0.097 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    173.54094,   0.00015,     0.345 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    169.32703,   0.00004,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1,   1 },    166.06523,   0.00006,     0.090 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    169.22225,   0.00034,     0.041 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    344.09104,   0.00008,     0.089 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    350.57298,   0.00045,     0.458 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    349.85223,   0.00009,     0.088 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,   1 },      0.61555,   0.00003,     0.065 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   2,  -1 },    349.42896,   0.00003,     0.541 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    158.72095,   0.00007,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    170.32184,   0.00142,     0.082 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    170.49227,   0.00818,     0.887 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    169.10744,   0.00129,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    175.28373,   0.00009,     0.036 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -2,  -1 },    169.35770,   0.00017,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -2,   0 },    350.19402,   0.00002,     0.077 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    169.37065,   0.00271,     0.076 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },    155.87557,   0.00002,     4.156 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    350.54413,   0.00046,     0.073 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,  -3 },    350.29441,   0.00002,     0.037 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },    169.26762,   0.01810,    13.541 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    154.98152,   0.00020,     0.074 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    349.30721,   0.00163,     0.037 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },    169.37065,   0.00271,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,   1 },    349.28233,   0.00018,     0.025 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   2,  -1 },    169.37065,   0.00018,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    137.60469,   0.00002,     0.072 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    160.98271,   0.00011,     1.547 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,  -2 },    195.24922,   0.00004,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    164.29973,   0.00281,     0.079 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,  -2 },    175.20334,   0.00007,     1.021 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    166.82596,   0.00093,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   1,  -2 },     15.24922,   0.00005,     0.081 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },    167.05949,   0.00010,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -2,  -1 },    349.43025,   0.00003,     0.588 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -1,  -1 },    349.57483,   0.00018,     0.087 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -1,   1 },    349.28236,   0.00004,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   0,  -1 },    169.26684,   0.00014,     0.040 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   0,   1 },    349.28241,   0.00002,     0.019 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   1,  -1 },    169.32703,   0.00004,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },    148.05110,   0.00002,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -1,  -2 },    195.30912,   0.00001,     0.095 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },    161.09427,   0.00005,     0.027 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,   0,  -2 },     15.23270,   0.00004,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },    166.81601,   0.00005,     0.020 },
 { {   0,   5,  -7,   0,   0,   1,   0,   0,   0,   0,   0 },    268.92224,   0.00001,     0.861 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.89829,   0.00002,     0.043 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,  -1,   0 },    359.85921,   0.00003,     0.084 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,   0,   0 },    175.80398,   0.00003,     0.771 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,   1,   0 },    359.85921,   0.00003,     0.069 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,   2,   0 },    359.88005,   0.00001,     0.036 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   2,  -1,   0 },    164.68671,   0.00004,     0.078 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   2,   0,   0 },    359.95362,   0.00001,     0.038 },
 { {   0,   5,  -7,   0,   3,   0,   0,   0,   0,   0,   0 },    169.63002,   0.00001,     0.724 },
 { {   0,   5,  -7,   0,   3,   0,   0,   0,   2,   0,   0 },    170.12878,   0.00003,     0.038 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -3,   0 },    337.23945,   0.00002,     0.078 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -2,   0 },      0.18132,   0.00003,     2.398 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -1,   0 },    337.23945,   0.00002,     0.073 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,   0,  -2 },    185.59480,   0.00002,    11.983 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,   0,   0 },      0.18132,   0.00003,     0.037 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -4,   1,   0 },      9.41215,   0.00001,     0.028 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,  -1,   0 },      9.41215,   0.00004,     0.026 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   0,   0 },      9.36573,   0.00032,     0.040 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   1,   0 },     10.18461,   0.00054,     0.087 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   2,   0 },      9.41215,   0.00002,     0.563 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -2,   0 },      5.59985,   0.00004,     0.038 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -1,   0 },      5.26116,   0.00053,     0.075 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   0,   0 },      9.00497,   0.00323,  1137.638 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   1,   0 },      5.25591,   0.00053,     0.075 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   2,   0 },      5.60034,   0.00004,     0.038 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,  -2,   0 },      9.40021,   0.00002,     0.564 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,  -1,   0 },     10.17222,   0.00054,     0.087 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,   0,   0 },      9.35498,   0.00032,     0.040 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,   1,   0 },      9.40021,   0.00004,     0.026 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   4,  -1,   0 },      9.40021,   0.00001,     0.028 },
 { {   0,   5,  -6,  -2,   0,   0,   0,   0,   1,   0,  -1 },     79.70022,   0.00003,    95.404 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.02458,   0.00001,     0.044 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,  -2,   2,   0 },    284.02458,   0.00001,     0.262 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   0,  -2,   0 },     27.69645,   0.00001,     0.041 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   0,   0,   0 },    207.69645,   0.00001,     0.489 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   2,  -3,   0 },    284.86700,   0.00002,     0.077 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   2,  -1,   0 },    104.86700,   0.00002,     0.074 },
 { {   0,   5,  -6,   0,   0,  -1,   0,   0,   2,   0,  -2 },      3.42159,   0.00002,    72.134 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,  -2,   0 },    273.77149,   0.00002,     0.020 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },     87.46616,   0.00014,     0.028 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.42560,   0.00113,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,   2 },     86.20463,   0.00001,     0.236 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,  -2 },    272.29577,   0.00001,     0.028 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },     85.33308,   0.00116,     0.107 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   2,  -2 },     92.32294,   0.00001,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },     82.61303,   0.00003,     0.256 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    282.77548,   0.00002,     0.058 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },    256.39642,   0.00001,     0.099 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },     75.47995,   0.00018,     0.042 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     93.81824,   0.00002,     0.098 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.59686,   0.00039,     0.314 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },     77.18686,   0.00008,     0.096 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    256.33222,   0.00004,     0.061 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -3,   0 },     92.13803,   0.00002,     0.027 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },     91.09474,   0.00023,     0.041 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,  -2 },    272.30456,   0.00012,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     91.39858,   0.00386,     0.090 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -2 },    272.29649,   0.00072,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },     76.38477,   0.00005,     0.089 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    272.66896,   0.00367,     0.470 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },     76.33156,   0.00002,     0.064 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,  -2 },     92.39080,   0.00026,     0.087 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    272.32204,   0.00084,     0.065 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,  -2 },     92.32294,   0.00008,     0.557 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    272.47157,   0.00005,     0.035 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -2,  -1 },     76.37594,   0.00003,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },     76.45309,   0.00046,     0.082 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },     92.55331,   0.00003,     0.806 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    256.38519,   0.00030,     0.068 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },     92.29756,   0.00003,     0.081 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.35549,   0.00310,     0.931 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },     90.73411,   0.00015,     0.069 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    253.95918,   0.00031,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },     76.59721,   0.00037,     0.070 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1,   1 },    254.70303,   0.00003,     0.024 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   2,  -1 },     76.37628,   0.00003,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -3,  -2 },    272.34704,   0.00004,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    133.38444,   0.00002,     0.078 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,  -2 },    272.31229,   0.00050,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    102.69627,   0.00011,     2.828 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -4 },     92.43617,   0.00001,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },    272.30952,   0.00742,     0.076 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -1 },    253.68024,   0.00003,     5.352 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    271.88133,   0.02267,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   2 },     91.81894,   0.00005,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -4 },     92.54150,   0.00007,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.29577,   0.05272,    49.790 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    256.36944,   0.00037,     0.074 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     92.87057,   0.00342,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },    272.31229,   0.00738,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -1 },    256.36944,   0.00002,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },     93.23879,   0.00045,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   2,  -2 },    272.31229,   0.00050,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },     92.36618,   0.00006,     0.019 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   3,  -2 },    272.31229,   0.00004,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,  -3,   0 },     91.84748,   0.00001,     1.875 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,  -1,  -1 },    262.23417,   0.00012,     0.080 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,   0,  -1 },     76.40167,   0.00003,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,   0,   0 },     92.45498,   0.00002,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -2,  -2 },     92.32294,   0.00008,     0.570 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },    273.21121,   0.00032,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,  -2 },     92.35993,   0.00039,     0.087 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    273.38024,   0.00015,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,  -2 },    272.29471,   0.00049,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },     98.61661,   0.00008,     0.019 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   1,  -2 },    272.30561,   0.00010,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   1,   0 },     92.57159,   0.00001,     0.015 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   6,  -2,  -2 },     92.32294,   0.00001,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   6,  -1,  -2 },    272.29577,   0.00001,     0.028 },
 { {   0,   5,  -6,   0,   0,   1,   0,   0,   2,  -3,   0 },    179.77012,   0.00002,     0.078 },
 { {   0,   5,  -6,   0,   0,   1,   0,   0,   2,  -1,   0 },    179.77012,   0.00002,     0.073 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },    191.56646,   0.00009,     0.044 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,  -2,   1,   0 },    189.25424,   0.00001,     0.108 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,   0,   0,   0 },    324.10636,   0.00002,     0.452 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,   2,  -3,   0 },    201.09942,   0.00002,     0.078 },
 { {   0,   5,  -6,   0,   2,   0,   0,   0,   0,   0,   0 },     72.94278,   0.00002,     0.436 },
 { {   0,   5,  -6,   0,   2,   0,   0,   0,   2,  -3,   0 },    257.29730,   0.00002,     0.079 },
 { {   0,   5,  -6,   0,   2,   0,   0,   0,   2,  -1,   0 },    257.29730,   0.00002,     0.073 },
 { {   0,   5,  -6,   1,   0,   0,   0,   0,   0,   0,   0 },    179.54975,   0.00001,     0.376 },
 { {   0,   5,  -6,   2,   0,   0,   0,   0,   2,  -3,   0 },    223.65699,   0.00002,     0.084 },
 { {   0,   5,  -6,   2,   0,   0,   0,   0,   2,  -2,   0 },     43.65699,   0.00002,     0.706 },
 { {   0,   5,  -6,   2,   0,   0,   0,   0,   2,  -1,   0 },    223.65699,   0.00002,     0.068 },
 { {   0,   5,  -6,   2,   0,   0,   0,   0,   2,   0,   0 },     43.65699,   0.00002,     0.036 },
 { {   0,   5,  -6,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.42088,   0.00002,     0.048 },
 { {   0,   5,  -6,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.42088,   0.00002,     0.182 },
 { {   0,   5,  -5,  -4,   0,   0,   0,   0,  -2,   0,   0 },    261.37065,   0.00003,     0.042 },
 { {   0,   5,  -5,  -4,   0,   0,   0,   0,  -2,   1,   0 },    261.37065,   0.00002,     0.095 },
 { {   0,   5,  -5,  -4,   0,   0,   0,   0,   0,   0,   0 },    244.15123,   0.00002,     0.999 },
 { {   0,   5,  -5,  -3,   0,   0,   0,   0,   2,  -3,   0 },     34.53318,   0.00001,     0.074 },
 { {   0,   5,  -5,  -3,   0,   0,   0,   0,   2,  -1,   0 },     34.53318,   0.00001,     0.077 },
 { {   0,   5,  -5,   0,  -2,   0,   0,   0,  -2,   0,   0 },    205.43527,   0.00001,     0.046 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,  -2,   0,   0 },    191.39795,   0.00006,     0.046 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,  -2,   1,   0 },     13.05207,   0.00001,     0.119 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.53189,   0.00001,     0.208 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,  -2,   3,   0 },    193.05207,   0.00001,     0.055 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,   0,  -2,   0 },    193.05207,   0.00001,     0.043 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,   0,   0,   0 },     54.38763,   0.00002,     0.329 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,  -1,   0 },    179.51709,   0.00001,     0.017 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    359.11543,   0.00003,     0.022 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },    358.79444,   0.00003,     0.030 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },      1.01011,   0.00002,     0.050 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    356.50464,   0.00004,     0.029 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    358.24987,   0.00120,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    358.30351,   0.00104,     0.120 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.89631,   0.00018,     0.204 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    356.39149,   0.00003,     0.055 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,  -1,   0 },    180.28601,   0.00001,     0.044 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },      0.06601,   0.00019,     0.108 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },     14.86339,   0.00004,     0.239 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.00057,   0.00005,     0.249 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },    356.53637,   0.00002,     0.027 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    359.83751,   0.00014,     0.043 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    358.67902,   0.00197,     0.099 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,  -2 },    175.22001,   0.00011,     0.042 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    180.02851,   0.03305,     0.320 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   2 },      0.11789,   0.00003,     0.033 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,  -2 },    184.32086,   0.00004,     0.096 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    180.05132,   0.00532,     0.061 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    180.02746,   0.00043,     0.034 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   3,   0 },    179.99427,   0.00004,     0.023 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   4,   0 },    180.02855,   0.00001,     0.018 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,  -2,   0 },    183.17846,   0.00002,     0.091 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    190.02421,   0.00002,     0.089 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.13708,   0.00009,     0.446 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,  -2 },    180.06353,   0.00004,     0.088 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    190.02421,   0.00027,     0.482 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },      0.29382,   0.00064,     0.065 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    190.02421,   0.00002,     0.065 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },      0.44244,   0.00003,     0.035 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.03008,   0.00124,     0.739 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,  -2 },    174.64815,   0.00009,     0.082 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.07299,   0.03759,     0.068 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   2 },      0.03914,   0.00008,     0.024 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.89773,   0.00126,     0.980 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },      0.04788,   0.00952,     0.036 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   2 },    180.02209,   0.00004,     0.018 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,  -2 },    355.44353,   0.00008,     0.070 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },      0.01666,   0.00080,     0.024 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    359.87478,   0.00012,     0.018 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -2,   0 },      0.15837,   0.00012,     0.073 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -1,  -2 },      0.06258,   0.00001,     7.515 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -1,  -1 },    179.36156,   0.00002,     0.074 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   0,  -2 },      0.06353,   0.00018,     0.075 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },      0.08032,   0.00003,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   1,   0 },      0.06634,   0.00001,     0.019 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -3,   0 },    182.09188,   0.00003,     0.078 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.06602,   0.00052,     0.038 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -1,  -2 },    180.04945,   0.00004,     0.080 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },    180.06925,   0.00016,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   0,  -2 },      0.01779,   0.00002,     0.039 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    359.97634,   0.00010,     0.019 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   1,   0 },      0.01350,   0.00004,     0.015 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.03501,   0.00001,     0.016 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },    345.67910,   0.00004,     0.046 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },    171.26140,   0.00001,     0.121 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,  -2,   3,   0 },    351.26140,   0.00001,     0.055 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   0,   0,   0 },    171.26140,   0.00002,     0.311 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,  -2,   0 },    166.28960,   0.00002,     0.695 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },    346.28964,   0.00004,     0.068 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },    158.61754,   0.00003,     0.036 },
 { {   0,   5,  -4,  -3,   0,   0,   0,   0,   2,  -3,   0 },    123.56068,   0.00002,     0.080 },
 { {   0,   5,  -4,  -3,   0,   0,   0,   0,   2,  -2,   0 },    303.56068,   0.00001,     1.318 },
 { {   0,   5,  -4,  -3,   0,   0,   0,   0,   2,  -1,   0 },    123.56068,   0.00002,     0.071 },
 { {   0,   5,  -4,  -3,   0,   0,   0,   0,   2,   0,   0 },    303.56068,   0.00001,     0.037 },
 { {   0,   5,  -4,  -1,   0,   0,   0,   0,   0,   0,   0 },    180.43180,   0.00001,     0.278 },
 { {   0,   5,  -4,   0,  -3,   0,   0,   0,   0,   0,   0 },    113.52150,   0.00002,     0.258 },
 { {   0,   5,  -4,   0,  -2,   0,   0,   0,   0,   0,   0 },    276.75251,   0.00001,     0.253 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.69827,   0.00001,     0.048 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0 },    181.48661,   0.00003,     0.134 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.69827,   0.00001,     0.172 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   3,   0 },      1.48661,   0.00003,     0.052 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   0,  -2,   0 },    181.70083,   0.00001,     0.045 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   0,  -1,   0 },    181.27481,   0.00001,     0.109 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.47093,   0.00008,     0.247 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   0,   1,   0 },    181.27481,   0.00001,     0.058 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    183.92490,   0.00002,     0.441 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    183.92490,   0.00002,     0.064 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   2,   0,   0 },      3.92490,   0.00002,     0.035 },
 { {   0,   5,  -4,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.21143,   0.00001,     0.048 },
 { {   0,   5,  -4,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.21143,   0.00001,     0.170 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.29032,   0.00002,     0.051 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    249.56000,   0.00002,     0.030 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    234.17859,   0.00008,     0.049 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    250.50591,   0.00018,     0.136 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    283.55545,   0.00002,     0.169 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    250.73214,   0.00001,     0.052 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.29720,   0.00002,     0.029 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.72475,   0.00005,     0.124 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.29470,   0.00029,     0.046 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.46054,   0.00042,     0.193 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.29323,   0.00014,     0.119 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    283.41504,   0.00003,     0.054 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    241.55316,   0.00001,     0.045 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    253.94078,   0.00019,     0.110 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.30956,   0.00001,     0.108 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     95.59954,   0.00029,     0.242 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     75.10590,   0.00005,     0.058 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.32065,   0.00022,     0.098 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.32669,   0.00012,     0.060 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.32721,   0.00145,     0.325 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    283.32388,   0.00008,     0.033 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.32371,   0.00016,     0.061 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    266.99682,   0.00004,     0.425 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.82534,   0.00052,     0.064 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },    262.83182,   0.00002,     0.495 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.34881,   0.00003,     0.065 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     81.92619,   0.00031,     0.035 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },     78.76471,   0.00007,     0.024 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    103.32345,   0.00006,     0.069 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.32679,   0.00003,     0.036 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },     83.78464,   0.00002,     0.025 },
 { {   0,   5,  -3, -13,   0,   0,   0,   0,  -2,   1,   0 },    328.44350,   0.00001,     0.075 },
 { {   0,   5,  -3, -13,   0,   0,   0,   0,  -2,   3,   0 },    328.44350,   0.00001,     0.075 },
 { {   0,   5,  -3,  -4,   0,   0,   0,   0,   0,   0,   0 },    303.53613,   0.00001,     0.333 },
 { {   0,   5,  -3,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.21945,   0.00001,     0.047 },
 { {   0,   5,  -3,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.21945,   0.00001,     0.188 },
 { {   0,   5,  -3,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.62159,   0.00001,     0.048 },
 { {   0,   5,  -3,  -2,   0,   0,   0,   0,  -2,   1,   0 },    179.68977,   0.00002,     0.135 },
 { {   0,   5,  -3,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.62159,   0.00001,     0.171 },
 { {   0,   5,  -3,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.88994,   0.00001,     0.045 },
 { {   0,   5,  -3,  -2,   0,   0,   0,   0,   0,   0,   0 },    179.88994,   0.00001,     0.246 },
 { {   0,   5,  -3,   0,  -2,   0,   0,   0,  -2,  -2,   0 },      0.10150,   0.00001,     0.022 },
 { {   0,   5,  -3,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.30858,   0.00002,     0.149 },
 { {   0,   5,  -3,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.76298,   0.00002,     0.046 },
 { {   0,   5,  -3,   0,  -1,   0,   0,   0,  -2,   0,   0 },     89.84334,   0.00002,     0.051 },
 { {   0,   5,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    269.84334,   0.00002,     0.147 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },     28.89356,   0.00003,     0.030 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     29.99973,   0.00028,     0.051 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     28.13268,   0.00017,     0.157 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },     25.99616,   0.00002,     0.145 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    185.74220,   0.00001,     0.048 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    205.64246,   0.00002,     0.138 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    188.99350,   0.00001,     0.162 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     25.76346,   0.00017,     0.123 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     26.39120,   0.00119,     0.195 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     25.99023,   0.00012,     0.054 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    149.87787,   0.00001,     0.032 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },      8.06302,   0.00002,     0.245 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    206.27331,   0.00002,     0.057 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    336.82029,   0.00002,     0.298 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     29.03489,   0.00019,     0.060 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    216.67313,   0.00010,     0.033 },
 { {   0,   5,  -2,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.68829,   0.00001,     0.053 },
 { {   0,   5,  -2,   0,  -3,   0,   0,   0,  -2,   2,   0 },    176.68829,   0.00001,     0.131 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    290.03497,   0.00002,     0.054 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    288.46315,   0.00002,     0.187 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    288.02033,   0.00008,     0.163 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    308.31413,   0.00001,     0.197 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    249.94198,   0.00002,     0.032 },
 { {   0,   5,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    133.50689,   0.00004,     0.032 },
 { {   0,   5,  -1,   0,  -2,   0,   0,   0,   2,  -2,   0 },    333.68755,   0.00001,     0.193 },
 { {   0,   5,  -1,   0,  -2,   0,   0,   0,   2,   0,   0 },    153.68755,   0.00001,     0.032 },
 { {   0,   5,  -1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    141.86789,   0.00001,     0.190 },
 { {   0,   5,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },    144.85198,   0.00008,     0.031 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    181.55070,   0.00001,     0.057 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    181.55070,   0.00001,     0.112 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    146.79769,   0.00001,     0.052 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    128.21247,   0.00002,     0.163 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    187.97454,   0.00002,     0.140 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    128.14199,   0.00003,     0.049 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    128.05908,   0.00001,     0.127 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    128.03385,   0.00002,     0.187 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    308.05908,   0.00001,     0.054 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    131.41007,   0.00002,     0.031 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21247,   0.00002,     0.017 },
 { {   0,   5,  -1,   0,   1,   0,   0,   0,   2,  -2,   0 },    299.51720,   0.00003,     0.184 },
 { {   0,   5,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },    298.03776,   0.00005,     0.031 },
 { {   0,   5,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },    129.52709,   0.00001,     0.124 },
 { {   0,   5,   0,   0,  -1,   0,   0,   0,   0,   2,   0 },    129.52709,   0.00001,     0.029 },
 { {   0,   5,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },    128.78376,   0.00004,     0.031 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.05032,   0.00005,     0.060 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.05032,   0.00005,     0.101 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    212.28191,   0.00003,     0.157 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    210.01950,   0.00002,     0.051 },
 { {   0,   5,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    153.60480,   0.00002,     0.049 },
 { {   0,   5,  31, -52,   0,   0,   0,   0,  -2,   1,   0 },    323.50777,   0.00002,   587.442 },
 { {   0,   6, -14,   4,   0,   0,   0,   0,  -2,   0,   2 },     79.80510,   0.00002,    79.413 },
 { {   0,   6, -14,   6,   0,   0,   0,   0,  -1,   0,   1 },    162.81738,   0.00001,   334.478 },
 { {   0,   6, -14,   8,   0,   0,   0,   0,   0,  -1,   0 },    174.65497,   0.00001,     0.075 },
 { {   0,   6, -14,   8,   0,   0,   0,   0,   0,   0,   0 },     53.99266,   0.00001,   151.217 },
 { {   0,   6, -14,   8,   0,   0,   0,   0,   0,   1,   0 },    174.65497,   0.00001,     0.075 },
 { {   0,   6, -13,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.79342,   0.00002,     0.036 },
 { {   0,   6, -13,   0,   2,   0,   0,   0,  -2,   1,   0 },    142.79342,   0.00002,     0.069 },
 { {   0,   6, -13,   4,   0,   0,   0,   0,  -2,   1,   0 },     81.32346,   0.00001,     0.079 },
 { {   0,   6, -13,   4,   0,   0,   0,   0,  -2,   3,   0 },     81.32346,   0.00001,     0.072 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -4,   0,   0 },     40.20754,   0.00002,     0.019 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    220.20754,   0.00002,     0.025 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,  -1,   2 },    220.20754,   0.00001,     0.075 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    219.60542,   0.00003,     0.037 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   2 },    218.34238,   0.00001,     7.177 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    220.20754,   0.00002,     0.073 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   1,   2 },    220.20754,   0.00001,     0.076 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    219.60542,   0.00003,     2.115 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    276.70150,   0.00002,     0.445 },
 { {   0,   6, -12,   0,   0,   4,   0,   0,  -2,   0,   2 },     46.32653,   0.00003,   282.735 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.63025,   0.00004,     0.037 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,  -2,   0,   2 },    202.98849,   0.00002,    34.155 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,   0,  -1,   0 },     51.61946,   0.00001,     0.065 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,   0,   1,   0 },     51.61946,   0.00001,     0.089 },
 { {   0,   6, -12,   0,   3,   0,   0,   0,  -2,   0,   0 },    221.86818,   0.00001,     0.037 },
 { {   0,   6, -12,   0,   3,   0,   0,   0,  -2,   1,   0 },     39.65357,   0.00001,     0.074 },
 { {   0,   6, -12,   0,   3,   0,   0,   0,  -2,   2,   0 },     41.86818,   0.00001,     4.547 },
 { {   0,   6, -12,   0,   3,   0,   0,   0,  -2,   3,   0 },     39.65357,   0.00001,     0.077 },
 { {   0,   6, -12,   0,   4,   0,   0,   0,  -2,   1,   0 },     32.27426,   0.00003,     0.075 },
 { {   0,   6, -12,   1,   0,   0,   0,   0,  -2,   1,   0 },    142.43043,   0.00001,     0.076 },
 { {   0,   6, -12,   1,   0,   0,   0,   0,  -2,   3,   0 },    142.43043,   0.00001,     0.075 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,  -2,   0,   0 },    233.64901,   0.00005,     0.039 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.76222,   0.00004,     1.693 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,   0,  -1,   0 },    231.83101,   0.00001,     0.069 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,   0,   1,   0 },    231.83101,   0.00001,     0.083 },
 { {   0,   6, -11,  -1,   0,   0,   0,   0,  -2,   1,   0 },     42.48004,   0.00002,     0.075 },
 { {   0,   6, -11,  -1,   0,   0,   0,   0,  -2,   2,   0 },    106.86328,   0.00002,   220.866 },
 { {   0,   6, -11,  -1,   0,   0,   0,   0,  -2,   3,   0 },     42.48004,   0.00002,     0.075 },
 { {   0,   6, -11,   0,  -1,   0,   0,   0,  -2,   1,   0 },    332.71339,   0.00003,     0.078 },
 { {   0,   6, -11,   0,  -1,   0,   0,   0,  -2,   3,   0 },    152.71339,   0.00003,     0.073 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -4,   3,   0 },    352.40018,   0.00001,     0.091 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,  -1,   2 },    324.02544,   0.00001,     0.081 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    358.53053,   0.00001,     0.038 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    208.82650,   0.00004,     0.079 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   1,   2 },    324.02544,   0.00001,     0.071 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    333.08213,   0.00006,     1.897 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    197.21763,   0.00004,     0.073 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    183.06757,   0.00002,     0.802 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    169.74328,   0.00002,     0.040 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    334.74031,   0.00002,     0.061 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },     57.85284,   0.00001,     0.331 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    334.74031,   0.00002,     0.098 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     57.85284,   0.00001,     0.043 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },     51.17075,   0.00001,     0.039 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,  -2,   1,   0 },    344.44769,   0.00003,     0.079 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },     51.17075,   0.00001,     1.635 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,   0,  -1,   0 },     50.43674,   0.00004,     0.069 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,   0,   0,   0 },     50.44615,   0.00003,     0.860 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,   0,   1,   0 },     50.43674,   0.00004,     0.083 },
 { {   0,   6, -11,   0,   2,   0,   0,   0,  -2,   1,   0 },    291.95740,   0.00003,     0.080 },
 { {   0,   6, -11,   0,   2,   0,   0,   0,  -2,   3,   0 },    291.95740,   0.00003,     0.072 },
 { {   0,   6, -11,   0,   3,   0,   0,   0,   0,   0,   0 },     35.64774,   0.00004,     1.006 },
 { {   0,   6, -11,   0,   3,   0,   0,   0,   2,  -1,   0 },     35.90102,   0.00005,     0.095 },
 { {   0,   6, -11,   0,   3,   0,   0,   0,   2,   0,   0 },     35.90102,   0.00004,     0.042 },
 { {   0,   6, -11,   2,   0,   0,   0,   0,  -2,   1,   0 },     95.39640,   0.00001,     0.086 },
 { {   0,   6, -11,   2,   0,   0,   0,   0,  -2,   3,   0 },     95.39640,   0.00001,     0.067 },
 { {   0,   6, -11,   2,   0,   0,   0,   0,   0,  -1,   0 },     98.60047,   0.00002,     0.074 },
 { {   0,   6, -11,   2,   0,   0,   0,   0,   0,   1,   0 },     98.60047,   0.00002,     0.076 },
 { {   0,   6, -11,   3,   0,   0,   0,   0,   0,  -1,   0 },    288.60493,   0.00001,     0.077 },
 { {   0,   6, -11,   3,   0,   0,   0,   0,   0,   1,   0 },    288.60493,   0.00001,     0.074 },
 { {   0,   6, -10,  -3,   0,   0,   0,   0,  -2,   1,   0 },     86.49969,   0.00002,     0.075 },
 { {   0,   6, -10,  -3,   0,   0,   0,   0,  -2,   3,   0 },     86.49969,   0.00002,     0.076 },
 { {   0,   6, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    254.10550,   0.00004,     0.040 },
 { {   0,   6, -10,   0,  -1,   0,   0,   0,  -2,   1,   0 },    241.62613,   0.00007,     0.085 },
 { {   0,   6, -10,   0,  -1,   0,   0,   0,   0,  -2,   2 },    323.89988,   0.00003,   425.789 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -4,   2,   0 },     17.61576,   0.00001,     0.043 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,  -2,   0 },     51.70494,   0.00002,     0.019 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     94.97298,   0.00006,     0.040 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.78407,   0.00003,     0.537 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    120.96343,   0.00007,     0.085 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },     60.20267,   0.00002,     0.655 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },     55.06804,   0.00003,     0.068 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    212.95089,   0.00004,     0.074 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    196.29676,   0.00008,     4.051 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    198.19644,   0.00005,     0.077 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    150.91811,   0.00001,     0.769 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    152.01518,   0.00010,     0.089 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    156.57538,   0.00007,     0.041 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },     41.43653,   0.00003,     0.040 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },     39.00000,   0.00003,     0.086 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,  -2,   3,   0 },    219.00000,   0.00003,     0.067 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,   0,  -1,   0 },     66.60308,   0.00001,     0.075 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,   0,   1,   0 },     66.60308,   0.00001,     0.076 },
 { {   0,   6, -10,   0,   2,   0,   0,   0,   0,  -1,   0 },    312.04460,   0.00001,     0.075 },
 { {   0,   6, -10,   0,   2,   0,   0,   0,   0,   0,   0 },    346.21080,   0.00002,    12.779 },
 { {   0,   6, -10,   0,   2,   0,   0,   0,   0,   1,   0 },    312.04460,   0.00001,     0.076 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -4,   1,   0 },    293.07924,   0.00001,     0.028 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,  -1,   0 },    293.07924,   0.00004,     0.026 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   0,   0 },    293.02768,   0.00039,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   1,   0 },    293.59106,   0.00061,     0.087 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   2,   0 },    293.07982,   0.00002,     0.562 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -2,   0 },    290.70547,   0.00005,     0.038 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -1,   0 },    290.54706,   0.00068,     0.075 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   0,   0 },    292.61428,   0.00373,   165.227 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   1,   0 },    290.54003,   0.00068,     0.075 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   2,   0 },    290.70524,   0.00005,     0.038 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,  -2,   0 },    293.07023,   0.00002,     0.566 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,  -1,   0 },    293.58645,   0.00060,     0.087 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   0,   0 },    293.01948,   0.00038,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   1,   0 },    293.07083,   0.00004,     0.026 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   4,  -1,   0 },    293.07083,   0.00001,     0.028 },
 { {   0,   6, -10,   0,   4,   0,   0,   0,   0,   0,   0 },    203.25175,   0.00003,    11.067 },
 { {   0,   6,  -9,  -3,   0,   0,   0,   0,  -2,   0,   0 },    355.30393,   0.00001,     0.039 },
 { {   0,   6,  -9,  -3,   0,   0,   0,   0,  -2,   1,   0 },    355.30393,   0.00001,     0.081 },
 { {   0,   6,  -9,  -3,   0,   0,   0,   0,  -2,   2,   0 },    355.30393,   0.00001,     1.073 },
 { {   0,   6,  -9,  -3,   0,   0,   0,   0,  -2,   3,   0 },    355.30393,   0.00001,     0.070 },
 { {   0,   6,  -9,  -2,   0,   0,   0,   0,  -2,   1,   0 },    191.21434,   0.00001,     0.085 },
 { {   0,   6,  -9,  -2,   0,   0,   0,   0,  -2,   3,   0 },    191.21434,   0.00001,     0.068 },
 { {   0,   6,  -9,  -2,   0,   0,   0,   0,   0,  -1,   0 },    190.99543,   0.00002,     0.074 },
 { {   0,   6,  -9,  -2,   0,   0,   0,   0,   0,   1,   0 },    190.99543,   0.00002,     0.077 },
 { {   0,   6,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.63411,   0.00001,     0.042 },
 { {   0,   6,  -9,   0,  -1,   0,   0,   0,   0,  -1,   0 },     55.63411,   0.00001,     0.079 },
 { {   0,   6,  -9,   0,  -1,   0,   0,   0,   0,   1,   0 },     55.63411,   0.00001,     0.072 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -4,   1,   0 },     71.47735,   0.00001,     0.028 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -4,   2,   0 },    122.82554,   0.00002,     0.045 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },     68.84068,   0.00004,     0.027 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.38850,   0.00038,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     71.74130,   0.00047,     0.093 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },     57.38348,   0.00002,     0.396 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,  -1,   1 },    246.55187,   0.00006,     0.087 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    250.05108,   0.00003,     0.086 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },     64.40922,   0.00002,     0.039 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     67.70389,   0.00048,     0.080 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     71.17581,   0.00254,     1.328 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },     70.98776,   0.00041,     0.071 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },     71.35558,   0.00003,     0.037 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -2,   1 },    245.66885,   0.00014,    30.407 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },     66.45329,   0.00002,     0.074 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },     69.38398,   0.00003,     7.470 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },     66.57558,   0.00029,     0.075 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },     66.75549,   0.00021,     3.326 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     70.05105,   0.00015,     0.076 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    247.82566,   0.00001,     0.038 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },     66.45329,   0.00002,     0.077 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    267.40540,   0.00006,     0.980 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    265.56828,   0.00063,     0.082 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    266.19141,   0.00009,     0.039 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -3,   0 },    269.33746,   0.00004,     0.096 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    270.10034,   0.00010,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.17960,   0.00005,     0.027 },
 { {   0,   6,  -9,   0,   2,   0,   0,   0,   0,   0,   0 },    271.00006,   0.00003,     1.085 },
 { {   0,   6,  -9,   0,   2,   0,   0,   0,   2,  -2,   0 },    271.04571,   0.00001,     1.173 },
 { {   0,   6,  -9,   0,   2,   0,   0,   0,   2,   0,   0 },    271.04571,   0.00001,     0.039 },
 { {   0,   6,  -9,   0,   3,   0,   0,   0,  -2,   0,   0 },    177.70577,   0.00002,     0.042 },
 { {   0,   6,  -9,   0,   3,   0,   0,   0,   0,   0,   0 },     21.77241,   0.00002,     0.994 },
 { {   0,   6,  -9,   2,   0,   0,   0,   0,   0,   0,   0 },     92.78783,   0.00001,     0.551 },
 { {   0,   6,  -9,   2,   0,   0,   0,   0,   2,  -3,   0 },      1.52469,   0.00003,     0.076 },
 { {   0,   6,  -9,   2,   0,   0,   0,   0,   2,  -1,   0 },      1.52469,   0.00003,     0.075 },
 { {   0,   6,  -8,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.82651,   0.00001,     0.043 },
 { {   0,   6,  -8,   0,  -2,   0,   0,   0,  -2,   2,   0 },    231.82651,   0.00001,     0.298 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,  -2,   0 },    347.13908,   0.00002,     0.020 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    350.84106,   0.00010,     0.028 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    308.48592,   0.00005,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    172.68091,   0.00025,     0.103 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    335.37027,   0.00001,     0.284 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },    168.61294,   0.00002,     0.042 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    357.49192,   0.00006,     0.356 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   1,  -1 },    211.02106,   0.00002,     0.093 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    348.50696,   0.00032,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,  -2 },    167.24585,   0.00003,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    347.12979,   0.01441,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    342.87289,   0.00262,     0.570 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    344.32861,   0.00035,     0.067 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    343.35192,   0.00001,     0.035 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -2,   0 },    347.20852,   0.00002,     0.081 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },    168.98609,   0.00010,     0.080 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    166.18968,   0.00008,     1.155 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    348.98829,   0.00006,     0.070 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    169.08977,   0.00082,     1.430 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    162.22273,   0.00009,     0.071 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    349.16770,   0.00008,     0.036 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },    168.98609,   0.00010,     0.072 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -4,   0 },    168.88015,   0.00013,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,  -2 },    347.54648,   0.00002,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    168.89483,   0.00192,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,  -2 },    347.10175,   0.00004,     0.037 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    167.12686,   0.05946,    47.898 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,   2 },    347.13417,   0.00007,     0.037 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,  -2 },    167.24627,   0.00013,     0.073 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },    173.17452,   0.00001,     5.330 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    167.22262,   0.07667,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   2 },    347.23207,   0.00017,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },    169.18906,   0.00006,     2.822 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    169.10173,   0.00007,     0.077 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    167.44083,   0.00558,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   2 },    347.24620,   0.00003,     0.019 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },    167.21770,   0.00041,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   2,   0 },    167.28968,   0.00003,     0.019 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   3,  -2,   0 },    347.11130,   0.00002,     0.081 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },    347.24557,   0.00003,     0.039 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -4,   0 },    164.60191,   0.00003,     0.557 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -3,   0 },    166.38562,   0.00093,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },    167.16516,   0.00166,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -1,  -2 },    347.24604,   0.00002,     0.090 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },    167.26694,   0.00080,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },    167.52563,   0.00013,     0.020 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   6,  -4,   0 },    166.39682,   0.00001,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   6,  -3,   0 },    166.97811,   0.00003,     0.028 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   6,  -2,   0 },    167.19840,   0.00002,     0.020 },
 { {   0,   6,  -8,   0,   0,   0,   1,   0,   2,  -2,   0 },     95.81277,   0.00001,   111.409 },
 { {   0,   6,  -8,   0,   0,   1,   0,   0,   2,  -3,   0 },    302.59967,   0.00001,     0.076 },
 { {   0,   6,  -8,   0,   0,   1,   0,   0,   2,  -2,   0 },     85.79865,   0.00007,    76.512 },
 { {   0,   6,  -8,   0,   0,   1,   0,   0,   2,  -1,   0 },    122.59967,   0.00001,     0.075 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   0,  -1,   0 },     64.22704,   0.00003,     0.088 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   0,   0,   0 },     91.25411,   0.00001,     0.544 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   2,  -2,   0 },     21.55475,   0.00003,    15.767 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   2,  -1,   0 },    244.22697,   0.00013,     0.075 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.89195,   0.00001,     0.044 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,  -2,   2,   0 },    359.89195,   0.00001,     0.271 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   0,  -1,   0 },    359.87174,   0.00001,     0.088 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   0,   1,   0 },    359.87174,   0.00001,     0.066 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   2,  -1,   0 },    357.28802,   0.00003,     0.075 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   2,   0,   0 },    359.86072,   0.00002,     0.038 },
 { {   0,   6,  -8,   0,   3,   0,   0,   0,   2,  -3,   0 },    169.17114,   0.00001,     0.077 },
 { {   0,   6,  -8,   0,   3,   0,   0,   0,   2,  -2,   0 },    350.12046,   0.00001,     4.310 },
 { {   0,   6,  -8,   0,   3,   0,   0,   0,   2,  -1,   0 },    169.17114,   0.00001,     0.074 },
 { {   0,   6,  -8,   0,   3,   0,   0,   0,   2,   0,   0 },    170.12046,   0.00001,     0.037 },
 { {   0,   6,  -8,   0,   4,   0,   0,   0,   2,   0,  -2 },    341.34112,   0.00002,    58.227 },
 { {   0,   6,  -8,   2,   0,   0,   0,   0,   2,  -2,   0 },      0.17123,   0.00002,     0.959 },
 { {   0,   6,  -8,   2,   0,   0,   0,   0,   2,   0,   0 },    353.93399,   0.00002,     0.036 },
 { {   0,   6,  -7,  -2,   0,   0,   0,   0,   2,  -3,   0 },    309.22247,   0.00001,     0.075 },
 { {   0,   6,  -7,  -2,   0,   0,   0,   0,   2,  -1,   0 },    309.22247,   0.00001,     0.076 },
 { {   0,   6,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.44998,   0.00002,     0.045 },
 { {   0,   6,  -7,   0,  -1,   0,   0,   0,   0,   0,   0 },    263.98409,   0.00001,     0.375 },
 { {   0,   6,  -7,   0,  -1,   0,   0,   0,   2,  -3,   0 },     99.88581,   0.00001,     0.081 },
 { {   0,   6,  -7,   0,  -1,   0,   0,   0,   2,  -1,   0 },    279.88581,   0.00001,     0.071 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },     99.19836,   0.00002,     0.028 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.94789,   0.00020,     0.045 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     88.46878,   0.00032,     0.115 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    283.82087,   0.00002,     0.221 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },     76.25386,   0.00011,     0.043 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.70230,   0.00001,     0.104 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.23740,   0.00023,     0.263 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },     76.27990,   0.00006,     0.102 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1,   1 },    256.26694,   0.00003,     0.059 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },     93.69652,   0.00004,     0.042 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.86916,   0.00037,     0.095 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    271.20703,   0.00395,     0.363 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    271.54781,   0.00077,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    270.76220,   0.00004,     0.034 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -2,   0 },    273.16474,   0.00003,     0.088 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },     76.25420,   0.00037,     0.087 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },     91.86102,   0.00002,     0.536 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    258.79442,   0.00016,     0.065 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.26226,   0.00160,     0.588 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },     91.45719,   0.00011,     0.066 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    255.61904,   0.00010,     0.035 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },     76.27434,   0.00018,     0.067 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   1,   1 },    255.99544,   0.00001,     0.024 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   2,   1 },     76.29613,   0.00002,     0.018 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    260.73963,   0.00020,     1.021 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,  -2 },    272.26596,   0.00002,     0.079 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,  -1 },    262.58658,   0.00001,     1.231 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.78630,   0.00811,     0.070 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.26596,   0.00031,     1.549 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    256.24924,   0.00008,     0.071 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     94.32226,   0.00140,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,  -2 },    272.26596,   0.00002,     0.072 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },     94.21570,   0.00019,     0.025 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },     90.14763,   0.00004,     0.019 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -3,  -1 },    256.28422,   0.00001,     0.075 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -3,   0 },    272.44311,   0.00002,    10.856 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -2,  -1 },    256.21780,   0.00021,    13.397 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -2,   0 },     92.34134,   0.00012,     0.075 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -1,  -1 },    256.21873,   0.00076,     0.076 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,   0,  -1 },    256.13173,   0.00002,     0.038 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },     92.44229,   0.00002,     0.025 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },    272.14228,   0.00013,     0.039 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },    274.54730,   0.00007,     0.026 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },     98.95262,   0.00007,     0.019 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,   1,   0 },     92.58547,   0.00001,     0.015 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,  -2,   0,   0 },    192.89631,   0.00009,     0.046 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,  -2,   0 },    358.83709,   0.00002,     0.042 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,  -1,   0 },    178.69843,   0.00001,     0.096 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,   0,   0 },    351.20367,   0.00003,     0.352 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,   1,   0 },    178.69843,   0.00001,     0.062 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   2,  -2,   0 },    178.61577,   0.00002,     0.940 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   2,   0,   0 },    178.61577,   0.00002,     0.036 },
 { {   0,   6,  -7,   0,   2,   0,   0,   0,   2,  -2,   0 },     60.18576,   0.00002,     0.871 },
 { {   0,   6,  -7,   0,   2,   0,   0,   0,   2,   0,   0 },     60.18576,   0.00002,     0.036 },
 { {   0,   6,  -7,   2,   0,   0,   0,   0,   2,  -2,   0 },     43.64868,   0.00001,     0.490 },
 { {   0,   6,  -7,   2,   0,   0,   0,   0,   2,   0,   0 },     43.64868,   0.00001,     0.035 },
 { {   0,   6,  -7,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.41256,   0.00002,     0.049 },
 { {   0,   6,  -7,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.41256,   0.00002,     0.163 },
 { {   0,   6,  -6,  -7,   0,   0,   0,   0,   0,  -1,   0 },    144.20847,   0.00001,     0.076 },
 { {   0,   6,  -6,  -7,   0,   0,   0,   0,   0,   1,   0 },    144.20847,   0.00001,     0.075 },
 { {   0,   6,  -6,   0,  -2,   0,   0,   0,  -2,   0,   0 },    205.43080,   0.00001,     0.047 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },    188.79198,   0.00008,     0.047 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.54745,   0.00001,     0.184 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,   0,  -2,   0 },    193.04761,   0.00001,     0.044 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,   0,   0,   0 },     13.04761,   0.00001,     0.273 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,   2,   0,   0 },      5.64958,   0.00003,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },    341.61926,   0.00001,     0.052 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    178.50458,   0.00001,     0.029 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    348.46414,   0.00004,     0.048 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   2 },    359.69322,   0.00001,     0.171 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    351.64109,   0.00005,     0.129 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    356.91432,   0.00010,     0.181 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },      0.20987,   0.00002,     0.053 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.98181,   0.00018,     0.116 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },     13.32498,   0.00003,     0.208 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.42026,   0.00003,     0.216 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    359.52172,   0.00008,     0.105 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,  -2 },    179.81608,   0.00001,     0.043 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    180.02165,   0.02263,     0.266 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,   2 },      0.08737,   0.00002,     0.033 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,  -2 },    180.01453,   0.00002,     0.102 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    180.03169,   0.00369,     0.059 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    179.99380,   0.00031,     0.033 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   3,   0 },    179.98052,   0.00002,     0.023 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },    190.38725,   0.00001,     0.095 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.01332,   0.00005,     0.349 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    190.38725,   0.00021,     0.370 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },      0.02095,   0.00041,     0.062 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },    190.38725,   0.00001,     0.063 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },    359.96052,   0.00001,     0.034 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },      1.63686,   0.00004,     0.089 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.03330,   0.00080,     0.505 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },    180.01469,   0.00003,     0.086 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.04574,   0.02116,     0.066 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   2 },      0.02249,   0.00004,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },    180.03541,   0.00067,     0.608 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },      0.06581,   0.00848,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   2 },    180.02422,   0.00003,     0.018 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },      0.02986,   0.00003,     0.067 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },      0.03027,   0.00070,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },      0.01420,   0.00009,     0.018 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,  -2,   0 },      0.06633,   0.00005,     0.070 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.06587,   0.00001,     0.036 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,   0,  -2 },      0.06823,   0.00003,     0.071 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,   0,   0 },      0.07936,   0.00003,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,   1,   0 },      0.06569,   0.00001,     0.018 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -3,   0 },    180.57654,   0.00003,     0.074 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.01801,   0.00030,     0.037 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -1,  -2 },    180.03309,   0.00001,     0.076 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    179.99508,   0.00004,     0.025 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },      0.02511,   0.00010,     0.019 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   1,   0 },      0.02831,   0.00004,     0.015 },
 { {   0,   6,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },    346.05912,   0.00005,     0.048 },
 { {   0,   6,  -6,   0,   1,   0,   0,   0,   0,   0,   0 },    165.72037,   0.00003,     0.261 },
 { {   0,   6,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },    166.05079,   0.00002,     0.035 },
 { {   0,   6,  -5,  -2,   0,   0,   0,   0,   2,  -2,   0 },    139.46662,   0.00001,     0.522 },
 { {   0,   6,  -5,  -2,   0,   0,   0,   0,   2,   0,   0 },    139.46662,   0.00001,     0.035 },
 { {   0,   6,  -5,   0,  -2,   0,   0,   0,   0,   0,   0 },    276.81950,   0.00002,     0.218 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,  -2,  -2,   0 },      1.27034,   0.00001,     0.021 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.66867,   0.00001,     0.155 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.50669,   0.00008,     0.214 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   2,  -3,   0 },    179.10711,   0.00001,     0.097 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   2,  -2,   0 },      3.88640,   0.00002,     0.345 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   2,  -1,   0 },    359.10711,   0.00001,     0.062 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   2,   0,   0 },      1.97336,   0.00003,     0.034 },
 { {   0,   6,  -5,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.20697,   0.00001,     0.050 },
 { {   0,   6,  -5,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.20697,   0.00001,     0.154 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.27935,   0.00002,     0.053 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    118.00521,   0.00003,     0.050 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    252.05267,   0.00008,     0.149 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    275.80168,   0.00003,     0.153 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.28383,   0.00003,     0.029 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.17499,   0.00001,     0.134 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.28108,   0.00023,     0.048 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.27580,   0.00029,     0.172 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.28519,   0.00009,     0.128 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    252.10680,   0.00002,     0.046 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    246.95665,   0.00008,     0.118 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.30082,   0.00001,     0.115 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     93.48317,   0.00013,     0.210 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },     73.35176,   0.00003,     0.056 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.31629,   0.00017,     0.105 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.32182,   0.00008,     0.058 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.32397,   0.00106,     0.270 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    283.31725,   0.00004,     0.033 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.31999,   0.00011,     0.059 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   1,   1 },    103.32149,   0.00001,     0.023 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    251.75208,   0.00003,     0.336 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    266.76575,   0.00039,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },    259.89924,   0.00001,     0.378 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.35227,   0.00002,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     87.28246,   0.00033,     0.034 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },     84.06530,   0.00006,     0.023 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },     77.37523,   0.00001,     0.018 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   3,  -1,  -1 },     99.93870,   0.00003,     0.066 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.32316,   0.00002,     0.035 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },     72.96510,   0.00001,     0.036 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },     86.12614,   0.00002,     0.024 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },     81.56586,   0.00002,     0.018 },
 { {   0,   6,  -5,   0,   1,   0,   0,   0,   2,  -2,   0 },    151.66028,   0.00001,     0.326 },
 { {   0,   6,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },    331.66028,   0.00001,     0.034 },
 { {   0,   6,  -4,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.21816,   0.00001,     0.049 },
 { {   0,   6,  -4,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.21816,   0.00001,     0.169 },
 { {   0,   6,  -4,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.59259,   0.00001,     0.050 },
 { {   0,   6,  -4,  -2,   0,   0,   0,   0,  -2,   1,   0 },    179.68476,   0.00003,     0.147 },
 { {   0,   6,  -4,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.59259,   0.00001,     0.155 },
 { {   0,   6,  -4,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.88548,   0.00001,     0.046 },
 { {   0,   6,  -4,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.77960,   0.00001,     0.052 },
 { {   0,   6,  -4,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.77960,   0.00001,     0.137 },
 { {   0,   6,  -4,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.30451,   0.00002,     0.052 },
 { {   0,   6,  -4,   0,  -2,   0,   0,   0,  -2,   1,   0 },      0.14797,   0.00004,     0.170 },
 { {   0,   6,  -4,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.30451,   0.00002,     0.136 },
 { {   0,   6,  -4,   0,  -2,   0,   0,   0,  -2,   3,   0 },      0.09704,   0.00001,     0.049 },
 { {   0,   6,  -4,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.74701,   0.00002,     0.048 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },     27.68259,   0.00003,     0.031 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     27.92781,   0.00026,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.50856,   0.00015,     0.175 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },     24.86978,   0.00002,     0.133 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    205.79741,   0.00002,     0.151 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    184.29595,   0.00001,     0.147 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     25.80379,   0.00017,     0.133 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     26.14700,   0.00097,     0.174 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     25.97302,   0.00009,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },      3.30471,   0.00001,     0.213 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    206.39026,   0.00002,     0.055 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    332.04744,   0.00003,     0.251 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     28.87562,   0.00015,     0.058 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    219.07327,   0.00011,     0.033 },
 { {   0,   6,  -3,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.68382,   0.00001,     0.055 },
 { {   0,   6,  -3,   0,  -3,   0,   0,   0,  -2,   2,   0 },    176.68382,   0.00001,     0.121 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    280.41200,   0.00002,     0.211 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    289.30330,   0.00007,     0.148 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    308.70608,   0.00001,     0.175 },
 { {   0,   6,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    125.38367,   0.00003,     0.032 },
 { {   0,   6,  -2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    220.73725,   0.00001,     0.170 },
 { {   0,   6,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },    151.80226,   0.00008,     0.031 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    128.23943,   0.00003,     0.182 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    159.37159,   0.00001,     0.129 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    128.23943,   0.00003,     0.048 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },    184.03956,   0.00001,     0.029 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    127.83510,   0.00001,     0.167 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    128.08597,   0.00003,     0.031 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21227,   0.00002,     0.017 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    308.00843,   0.00002,     0.017 },
 { {   0,   6,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    298.77412,   0.00008,     0.031 },
 { {   0,   6,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 },    129.52263,   0.00002,     0.115 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.03320,   0.00004,     0.063 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.03320,   0.00004,     0.095 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     79.42172,   0.00001,     0.114 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    216.98408,   0.00004,     0.143 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    216.98408,   0.00004,     0.030 },
 { {   0,   6,   0, -15,   0,   0,   0,   0,   2,  -3,   0 },    323.85091,   0.00001,     0.075 },
 { {   0,   6,   0, -15,   0,   0,   0,   0,   2,  -1,   0 },    323.85091,   0.00001,     0.075 },
 { {   0,   6,   0,  -2,   0,   0,   0,   0,   2,  -2,   0 },    308.30059,   0.00002,     0.145 },
 { {   0,   6,   0,  -2,   0,   0,   0,   0,   2,   0,   0 },    308.24178,   0.00003,     0.030 },
 { {   0,   6,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    327.13397,   0.00001,     0.029 },
 { {   0,   6,   4,   0,  -6,   0,   0,   0,   0,  -1,   0 },    124.17800,   0.00002,   118.454 },
 { {   0,   7, -42,  36,   0,   0,   0,   0,   2,  -1,   0 },    306.33764,   0.00001,   968.212 },
 { {   0,   7, -19,  11,   0,   0,   0,   0,  -2,   1,   0 },    158.55636,   0.00005,     0.075 },
 { {   0,   7, -19,  11,   0,   0,   0,   0,  -2,   2,   0 },    132.89984,   0.00007,   828.943 },
 { {   0,   7, -19,  11,   0,   0,   0,   0,  -2,   3,   0 },    158.55636,   0.00005,     0.075 },
 { {   0,   7, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.61472,   0.00002,     0.034 },
 { {   0,   7, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.62140,   0.00001,     0.351 },
 { {   0,   7, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    305.27208,   0.00002,     0.381 },
 { {   0,   7, -14,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.84799,   0.00002,     0.037 },
 { {   0,   7, -14,   0,   2,   0,   0,   0,  -2,   1,   0 },    322.94001,   0.00002,     0.072 },
 { {   0,   7, -14,   0,   2,   0,   0,   0,  -2,   3,   0 },    142.94001,   0.00002,     0.080 },
 { {   0,   7, -14,   2,   0,   0,   0,   0,  -2,   1,   0 },    323.05907,   0.00002,     0.077 },
 { {   0,   7, -14,   2,   0,   0,   0,   0,  -2,   3,   0 },    323.05907,   0.00002,     0.074 },
 { {   0,   7, -14,   5,   0,   0,   0,   0,   0,  -1,   0 },    289.17636,   0.00003,     0.076 },
 { {   0,   7, -14,   5,   0,   0,   0,   0,   0,   1,   0 },    289.17636,   0.00003,     0.075 },
 { {   0,   7, -13,   0,  -1,   0,   0,   0,  -2,   1,   0 },     51.98380,   0.00007,     0.076 },
 { {   0,   7, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    230.98389,   0.00001,    14.625 },
 { {   0,   7, -13,   0,  -1,   0,   0,   0,   0,   1,   0 },    231.98380,   0.00002,     0.087 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    221.48024,   0.00002,     0.025 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    219.96242,   0.00002,     0.038 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    244.05055,   0.00005,     0.076 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    219.96242,   0.00002,     6.550 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    100.89981,   0.00002,     0.075 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    268.19660,   0.00002,     0.617 },
 { {   0,   7, -13,   0,   1,   0,   0,   0,  -2,   1,   0 },    142.51747,   0.00001,     0.077 },
 { {   0,   7, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },    141.67468,   0.00001,     4.220 },
 { {   0,   7, -13,   0,   1,   0,   0,   0,  -2,   3,   0 },    142.51747,   0.00001,     0.074 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   1,   0 },    203.16216,   0.00001,     0.077 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   2,   0 },     51.57778,   0.00002,     3.113 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   3,   0 },    203.16216,   0.00001,     0.074 },
 { {   0,   7, -13,   0,   3,   0,   0,   0,  -2,   1,   0 },     41.58668,   0.00003,     0.078 },
 { {   0,   7, -13,   0,   3,   0,   0,   0,  -2,   3,   0 },     41.58668,   0.00003,     0.073 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   0,   0 },    233.72668,   0.00005,     0.040 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   1,   0 },    231.75298,   0.00003,     0.083 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.75298,   0.00003,     0.822 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   3,   0 },    231.75298,   0.00003,     0.069 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,   0,  -1,   0 },     51.75298,   0.00003,     0.072 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,   0,   1,   0 },     51.75298,   0.00003,     0.079 },
 { {   0,   7, -13,   3,   0,   0,   0,   0,   0,  -1,   0 },     20.15736,   0.00004,     0.075 },
 { {   0,   7, -13,   3,   0,   0,   0,   0,   0,   1,   0 },     20.15736,   0.00004,     0.076 },
 { {   0,   7, -12,  -2,   0,   0,   0,   0,  -2,   1,   0 },      1.73867,   0.00001,     0.076 },
 { {   0,   7, -12,  -2,   0,   0,   0,   0,  -2,   3,   0 },      1.73867,   0.00001,     0.075 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    346.66227,   0.00001,     0.039 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    357.99146,   0.00006,     0.083 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    334.22332,   0.00004,     0.868 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    346.88252,   0.00006,     0.069 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    173.17851,   0.00001,     0.037 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    205.74575,   0.00002,     0.072 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     77.65085,   0.00003,     1.609 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    192.77160,   0.00003,     0.079 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    142.71587,   0.00001,     0.039 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    237.87060,   0.00002,     0.417 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    237.87803,   0.00002,     0.092 },
 { {   0,   7, -12,   0,   1,   0,   0,   0,   0,   0,   0 },     50.43580,   0.00002,     1.862 },
 { {   0,   7, -12,   0,   2,   0,   0,   0,  -2,   1,   0 },    111.92313,   0.00003,     0.084 },
 { {   0,   7, -12,   0,   2,   0,   0,   0,   0,  -1,   0 },    291.92313,   0.00002,     0.073 },
 { {   0,   7, -12,   0,   2,   0,   0,   0,   0,   1,   0 },    291.92313,   0.00002,     0.078 },
 { {   0,   7, -12,   1,   0,   0,   0,   0,   0,  -1,   0 },    231.24612,   0.00001,     0.075 },
 { {   0,   7, -12,   1,   0,   0,   0,   0,   0,   1,   0 },    231.24612,   0.00001,     0.076 },
 { {   0,   7, -12,   2,   0,   0,   0,   0,  -2,   0,   0 },     95.38605,   0.00002,     0.041 },
 { {   0,   7, -12,   2,   0,   0,   0,   0,  -2,   2,   0 },     95.38605,   0.00002,     0.451 },
 { {   0,   7, -12,   2,   0,   0,   0,   0,   0,  -1,   0 },     95.38605,   0.00001,     0.078 },
 { {   0,   7, -12,   2,   0,   0,   0,   0,   0,   1,   0 },     95.38605,   0.00001,     0.073 },
 { {   0,   7, -11,  -4,   0,   0,   0,   0,  -2,   1,   0 },    344.27383,   0.00001,     0.076 },
 { {   0,   7, -11,  -4,   0,   0,   0,   0,  -2,   3,   0 },    344.27383,   0.00001,     0.075 },
 { {   0,   7, -11,   0,  -1,   0,   0,   0,  -2,   0,   0 },    242.17307,   0.00003,     0.041 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    145.90283,   0.00006,     0.041 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.77743,   0.00002,     0.402 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    126.37372,   0.00007,     0.090 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },     61.06912,   0.00001,     0.465 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },     51.94140,   0.00003,     0.065 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    130.28157,   0.00010,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    153.11345,   0.00058,     2.641 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    130.28157,   0.00010,     0.073 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    151.66055,   0.00002,     1.481 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   1,   0,   0 },     58.88694,   0.00002,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    172.45257,   0.00003,     0.084 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    153.03558,   0.00004,     0.040 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },     39.15162,   0.00005,     0.041 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,  -2,   1,   0 },    218.65854,   0.00003,     0.091 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },     38.65718,   0.00003,     0.447 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,   0,  -1,   0 },     38.92048,   0.00003,     0.078 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,   0,   1,   0 },     38.92048,   0.00003,     0.073 },
 { {   0,   7, -10,  -2,   0,   0,   0,   0,   0,  -1,   0 },     10.98302,   0.00002,     0.077 },
 { {   0,   7, -10,  -2,   0,   0,   0,   0,   0,   1,   0 },     10.98302,   0.00002,     0.074 },
 { {   0,   7, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    235.59653,   0.00001,     0.043 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -4,   0,   0 },    120.98243,   0.00003,     0.021 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    261.60273,   0.00008,     0.043 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    254.31865,   0.00011,     0.099 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     72.52235,   0.00011,     0.084 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     49.02518,   0.00010,     0.725 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    254.34708,   0.00001,     0.068 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   2,   0 },     72.63743,   0.00001,     0.036 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    248.42646,   0.00002,     0.077 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    248.42646,   0.00036,     3.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    250.08895,   0.00003,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,   1 },     68.37961,   0.00004,     0.037 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   1,  -1 },    248.42646,   0.00002,     0.074 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   2,  -3 },     67.85481,   0.00001,   116.052 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },     42.11926,   0.00001,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    254.51257,   0.00006,     2.529 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    253.77328,   0.00064,     0.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    253.81395,   0.00014,     0.038 },
 { {   0,   7, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },     24.47721,   0.00001,     0.043 },
 { {   0,   7, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },    204.47721,   0.00001,     0.309 },
 { {   0,   7, -10,   0,   1,   0,   0,   0,   2,  -3,   0 },    169.25274,   0.00002,     0.074 },
 { {   0,   7, -10,   0,   1,   0,   0,   0,   2,  -1,   0 },    169.25274,   0.00002,     0.077 },
 { {   0,   7, -10,   0,   2,   0,   0,   0,   2,  -3,   0 },     91.24294,   0.00001,     0.074 },
 { {   0,   7, -10,   0,   2,   0,   0,   0,   2,  -1,   0 },     91.24294,   0.00001,     0.077 },
 { {   0,   7, -10,   0,   2,   0,   0,   0,   2,   0,   0 },    271.10027,   0.00002,     0.038 },
 { {   0,   7, -10,   0,   3,   0,   0,   0,   2,  -3,   0 },     81.41201,   0.00002,     0.075 },
 { {   0,   7, -10,   0,   3,   0,   0,   0,   2,  -1,   0 },    214.09714,   0.00006,     0.076 },
 { {   0,   7, -10,   0,   4,   0,   0,   0,   2,  -1,   0 },    172.19332,   0.00003,     0.076 },
 { {   0,   7, -10,   2,   0,   0,   0,   0,   2,  -3,   0 },    271.31183,   0.00001,     0.079 },
 { {   0,   7, -10,   2,   0,   0,   0,   0,   2,  -1,   0 },    271.31183,   0.00001,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    171.55245,   0.00006,     0.045 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    165.24609,   0.00007,     0.110 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    225.48196,   0.00001,     0.241 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    356.80563,   0.00001,     0.057 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    168.27852,   0.00001,     0.043 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },      6.32091,   0.00004,     0.291 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    171.27810,   0.00029,     0.092 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    342.00375,   0.00101,     0.420 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    345.54539,   0.00018,     0.064 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,  -3,   1 },    168.93861,   0.00002,     0.086 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },    168.92485,   0.00004,     0.084 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },    348.93861,   0.00002,     0.067 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.98461,   0.00037,     0.755 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    163.21785,   0.00004,     0.068 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    348.93858,   0.00003,     0.035 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },    168.92485,   0.00004,     0.069 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    351.31644,   0.00006,     0.079 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    331.06974,   0.00001,     1.654 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    348.84970,   0.00238,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    350.08697,   0.00015,     3.688 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    349.06806,   0.00003,     0.073 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    174.39874,   0.00023,     0.037 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },    168.91441,   0.00004,     0.025 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   4,  -3,   0 },    171.79239,   0.00001,     0.083 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    347.89712,   0.00004,     0.039 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },    169.73018,   0.00003,     0.019 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.39475,   0.00001,     0.045 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.39475,   0.00001,     0.236 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   0,   0,   0 },     90.97002,   0.00001,     0.406 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,  -3,   0 },    269.89632,   0.00002,     0.080 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,  -2,   0 },     89.92696,   0.00002,     1.452 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,  -1,   0 },    269.89632,   0.00002,     0.072 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,   0,   0 },     89.92696,   0.00002,     0.037 },
 { {   0,   7,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    179.88364,   0.00001,     0.045 },
 { {   0,   7,  -9,   0,   2,   0,   0,   0,  -2,   2,   0 },    359.88364,   0.00001,     0.231 },
 { {   0,   7,  -9,   0,   2,   0,   0,   0,   2,  -3,   0 },    179.92269,   0.00001,     0.080 },
 { {   0,   7,  -9,   0,   2,   0,   0,   0,   2,  -1,   0 },    179.92269,   0.00001,     0.071 },
 { {   0,   7,  -9,   0,   2,   0,   0,   0,   2,   0,   0 },    359.84379,   0.00003,     0.037 },
 { {   0,   7,  -9,   0,   3,   0,   0,   0,   2,  -1,   0 },    350.11012,   0.00003,     0.071 },
 { {   0,   7,  -9,   2,   0,   0,   0,   0,   2,  -2,   0 },      0.15930,   0.00002,     0.600 },
 { {   0,   7,  -9,   4,   0,   0,   0,   0,   0,   0,   0 },     57.64351,   0.00001,     0.222 },
 { {   0,   7,  -8,  -3,   0,   0,   0,   0,   2,  -2,   0 },    350.56694,   0.00003,   104.252 },
 { {   0,   7,  -8,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.47209,   0.00002,     0.047 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     83.23138,   0.00002,     0.047 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     90.52699,   0.00016,     0.123 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    275.09681,   0.00001,     0.194 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,  -1,   1 },     76.15114,   0.00002,     0.113 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },     76.12400,   0.00007,     0.045 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },     93.67482,   0.00004,     0.111 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.10238,   0.00014,     0.226 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   1,  -1 },     76.13968,   0.00005,     0.109 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     89.65201,   0.00025,     0.101 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    271.72293,   0.00303,     0.296 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    271.81423,   0.00059,     0.060 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    271.63462,   0.00004,     0.033 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },     76.14435,   0.00014,     0.091 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },     91.92088,   0.00001,     0.401 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    256.14893,   0.00007,     0.063 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.14496,   0.00097,     0.430 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     90.99229,   0.00007,     0.064 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    256.12083,   0.00002,     0.034 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },     76.14224,   0.00010,     0.064 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    270.85810,   0.00003,     0.086 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.95782,   0.00009,     0.623 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.00995,   0.00422,     0.067 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.17587,   0.00014,     0.787 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    258.11372,   0.00004,     0.068 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     93.51051,   0.00147,     0.036 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },     93.92594,   0.00020,     0.024 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   2,   0 },     92.34454,   0.00003,     0.018 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,  -1,  -1 },     84.02925,   0.00009,     0.072 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,   0,  -1 },     76.14441,   0.00001,     0.037 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,   0,   0 },     92.43620,   0.00002,     0.025 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },    273.50069,   0.00007,     0.038 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },    261.87196,   0.00001,     0.025 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },     98.44385,   0.00007,     0.019 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,   1,   0 },     92.56636,   0.00001,     0.015 },
 { {   0,   7,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },    194.21063,   0.00008,     0.047 },
 { {   0,   7,  -8,   0,   1,   0,   0,   0,   0,   0,   0 },     11.58450,   0.00002,     0.289 },
 { {   0,   7,  -8,   0,   1,   0,   0,   0,   2,  -2,   0 },    178.56395,   0.00001,     0.592 },
 { {   0,   7,  -8,   0,   1,   0,   0,   0,   2,   0,   0 },    178.65268,   0.00004,     0.035 },
 { {   0,   7,  -8,   0,   2,   0,   0,   0,   2,   0,   0 },     60.17541,   0.00003,     0.035 },
 { {   0,   7,  -8,   0,   2,   0,   0,   0,   4,  -4,   0 },     60.17541,   0.00013,  1309.823 },
 { {   0,   7,  -8,   2,   0,   0,   0,   0,  -2,   1,   0 },     40.49581,   0.00001,     0.142 },
 { {   0,   7,  -8,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.40222,   0.00002,     0.051 },
 { {   0,   7,  -8,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.40222,   0.00002,     0.148 },
 { {   0,   7,  -7,   0,  -2,   0,   0,   0,   0,   0,   0 },    194.58968,   0.00002,     0.238 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },    187.37803,   0.00007,     0.049 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.56129,   0.00001,     0.165 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,   0,   0,   0 },    161.14844,   0.00002,     0.233 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,   2,  -1,   0 },    188.70149,   0.00003,     0.063 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,   2,   0,   0 },      5.57836,   0.00002,     0.034 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },    179.80136,   0.00002,     0.054 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.99434,   0.00004,     0.030 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    181.03795,   0.00031,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.85118,   0.00023,     0.141 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.73670,   0.00010,     0.163 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    359.75732,   0.00002,     0.052 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.95678,   0.00014,     0.125 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },     17.56019,   0.00002,     0.184 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.05143,   0.00002,     0.190 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    179.29296,   0.00002,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.43156,   0.00041,     0.113 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,  -2 },    179.92421,   0.00001,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    179.98587,   0.01547,     0.228 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,   2 },      0.02374,   0.00002,     0.032 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    180.07780,   0.00258,     0.057 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    180.18203,   0.00020,     0.032 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   3,   0 },    180.13016,   0.00002,     0.023 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.32124,   0.00005,     0.286 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },    190.86477,   0.00017,     0.301 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },      0.01887,   0.00023,     0.060 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   1,   0 },    359.95962,   0.00001,     0.033 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -4,   0 },      0.02143,   0.00001,     0.042 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },      0.07850,   0.00003,     0.094 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.20108,   0.00055,     0.384 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,  -2 },    180.02317,   0.00004,     0.091 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.05152,   0.01290,     0.063 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   2 },      0.03096,   0.00003,     0.023 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,  -2 },    180.02963,   0.00038,     0.440 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },      0.04712,   0.00708,     0.034 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00981,   0.00003,     0.018 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },      0.00567,   0.00059,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },    359.99526,   0.00008,     0.018 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,  -2,   0 },      0.01116,   0.00003,     0.067 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.06270,   0.00001,     0.035 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },      0.07645,   0.00003,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,   1,   0 },      0.06251,   0.00001,     0.018 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.90396,   0.00018,     0.037 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.92759,   0.00001,     0.025 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },      0.00404,   0.00009,     0.019 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,   1,   0 },      0.01732,   0.00003,     0.015 },
 { {   0,   7,  -7,   0,   1,   0,   0,   0,  -2,   0,   0 },    346.20478,   0.00005,     0.049 },
 { {   0,   7,  -7,   0,   1,   0,   0,   0,   0,  -2,   0 },    351.25059,   0.00001,     0.045 },
 { {   0,   7,  -7,   0,   1,   0,   0,   0,   0,   0,   0 },    171.25059,   0.00001,     0.224 },
 { {   0,   7,  -7,   0,   1,   0,   0,   0,   2,  -1,   0 },    346.91082,   0.00002,     0.063 },
 { {   0,   7,  -6,  -2,   0,   0,   0,   0,   2,  -2,   0 },    139.45627,   0.00001,     0.394 },
 { {   0,   7,  -6,  -2,   0,   0,   0,   0,   2,   0,   0 },    139.45627,   0.00001,     0.034 },
 { {   0,   7,  -6,   0,  -2,   0,   0,   0,  -2,   0,   0 },    108.07759,   0.00001,     0.051 },
 { {   0,   7,  -6,   0,  -2,   0,   0,   0,  -2,   2,   0 },    108.07759,   0.00001,     0.143 },
 { {   0,   7,  -6,   0,  -2,   0,   0,   0,   0,   0,   0 },    277.93501,   0.00001,     0.192 },
 { {   0,   7,  -6,   0,  -1,   0,   0,   0,  -2,  -2,   0 },    181.26401,   0.00001,     0.022 },
 { {   0,   7,  -6,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.63609,   0.00001,     0.141 },
 { {   0,   7,  -6,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.50885,   0.00008,     0.189 },
 { {   0,   7,  -6,   0,  -1,   0,   0,   0,   2,   0,   0 },      3.84881,   0.00003,     0.033 },
 { {   0,   7,  -6,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.20063,   0.00001,     0.052 },
 { {   0,   7,  -6,   0,   0,  -1,   0,   0,  -2,   2,   0 },      0.20063,   0.00001,     0.140 },
 { {   0,   7,  -6,   0,   0,  -1,   0,   0,   0,   0,   0 },    180.19092,   0.00001,     0.187 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -3,   0,   1 },    103.26602,   0.00002,     0.055 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.00920,   0.00004,     0.052 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.10478,   0.00003,     0.164 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    287.65431,   0.00003,     0.140 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.27212,   0.00002,     0.030 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.26506,   0.00018,     0.049 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.16314,   0.00022,     0.155 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.26897,   0.00007,     0.140 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    103.31514,   0.00001,     0.051 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    238.17758,   0.00003,     0.127 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.29041,   0.00001,     0.124 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    103.87008,   0.00003,     0.186 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     81.95330,   0.00003,     0.054 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    353.01259,   0.00001,     0.031 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.30932,   0.00014,     0.112 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.31570,   0.00006,     0.056 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.31793,   0.00081,     0.231 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.31359,   0.00009,     0.057 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   1,   1 },    103.31536,   0.00001,     0.023 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    258.70484,   0.00001,     0.277 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    269.10504,   0.00029,     0.059 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },    261.80123,   0.00001,     0.306 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    283.35077,   0.00002,     0.060 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     89.19962,   0.00033,     0.033 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },     85.48805,   0.00005,     0.023 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   3,   0,  -1 },    103.31603,   0.00002,     0.034 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },     87.86549,   0.00002,     0.024 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },     81.80114,   0.00003,     0.018 },
 { {   0,   7,  -5,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.21540,   0.00001,     0.050 },
 { {   0,   7,  -5,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.21540,   0.00001,     0.152 },
 { {   0,   7,  -5,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.56670,   0.00001,     0.051 },
 { {   0,   7,  -5,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.56670,   0.00001,     0.141 },
 { {   0,   7,  -5,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.87914,   0.00001,     0.047 },
 { {   0,   7,  -5,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.67581,   0.00001,     0.054 },
 { {   0,   7,  -5,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.67581,   0.00001,     0.127 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.29893,   0.00002,     0.125 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,   0,  -2,   0 },      0.11639,   0.00003,     0.049 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.91945,   0.00007,     0.161 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,   2,  -2,   0 },    179.98422,   0.00001,     0.225 },
 { {   0,   7,  -5,   0,  -1,   0,   0,   0,   0,   0,   0 },     89.24118,   0.00001,     0.159 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.83917,   0.00003,     0.032 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     27.82119,   0.00022,     0.054 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.61916,   0.00013,     0.196 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    205.91954,   0.00002,     0.167 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    342.07231,   0.00001,     0.050 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.39256,   0.00014,     0.145 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     26.20666,   0.00076,     0.157 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },     25.94863,   0.00007,     0.051 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    327.22099,   0.00003,     0.217 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.91958,   0.00011,     0.056 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    218.64306,   0.00011,     0.032 },
 { {   0,   7,  -4,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.67748,   0.00001,     0.057 },
 { {   0,   7,  -4,   0,  -3,   0,   0,   0,  -2,   2,   0 },    176.67748,   0.00001,     0.112 },
 { {   0,   7,  -4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    156.55444,   0.00001,     0.181 },
 { {   0,   7,  -4,   0,  -1,   0,   0,   0,   2,   0,   0 },    336.55444,   0.00001,     0.031 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    279.31321,   0.00002,     0.244 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    286.23738,   0.00005,     0.136 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    309.00800,   0.00001,     0.158 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     44.09399,   0.00003,     0.178 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     29.30085,   0.00001,     0.031 },
 { {   0,   7,  -4,   0,   1,   0,   0,   0,   2,  -2,   0 },    307.29456,   0.00003,     0.176 },
 { {   0,   7,  -4,   0,   1,   0,   0,   0,   2,   0,   0 },    311.49871,   0.00002,     0.031 },
 { {   0,   7,  -3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    294.47001,   0.00002,     0.153 },
 { {   0,   7,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },    157.40616,   0.00005,     0.030 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    128.23584,   0.00003,     0.205 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    184.33882,   0.00001,     0.119 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    128.23584,   0.00003,     0.046 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    130.59254,   0.00001,     0.151 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.19712,   0.00002,     0.050 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    127.78912,   0.00003,     0.030 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    128.25372,   0.00003,     0.022 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    128.21098,   0.00002,     0.017 },
 { {   0,   7,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    298.78214,   0.00007,     0.030 },
 { {   0,   7,  -2,  -2,   0,   0,   0,   0,   2,  -2,   0 },     84.56604,   0.00001,     0.153 },
 { {   0,   7,  -2,  -2,   0,   0,   0,   0,   2,   0,   0 },     84.56604,   0.00001,     0.030 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,  -4,   1,   2 },    307.56812,   0.00003,   224.432 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.01220,   0.00004,     0.065 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.01220,   0.00004,     0.090 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    106.15247,   0.00002,     0.107 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    221.97672,   0.00002,     0.132 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    221.97672,   0.00002,     0.029 },
 { {   0,   7,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    146.63392,   0.00005,     0.075 },
 { {   0,   7,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    333.57500,   0.00002,     9.702 },
 { {   0,   7,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.63392,   0.00005,     0.076 },
 { {   0,   7,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },     81.62314,   0.00007,   100.849 },
 { {   0,   7,   4,  -4,   0,   0,   0,   0,   0,  -1,   0 },    255.05749,   0.00004,   261.124 },
 { {   0,   8, -47,  39,   0,   0,   0,   0,   0,   1,   0 },     14.69784,   0.00001,   228.633 },
 { {   0,   8, -21,  19,   0,   0,   0,   0,   2,   0,  -2 },     29.96674,   0.00001,   743.522 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.62237,   0.00004,     0.035 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    109.62109,   0.00001,     0.065 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.62237,   0.00004,     0.450 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   3,   0 },    109.62109,   0.00001,     0.091 },
 { {   0,   8, -16,   0,   0,   0,   0,   0,  -3,   1,   2 },    189.09427,   0.00009,   784.579 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    135.62389,   0.00002,     0.074 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   3,   0 },    121.17650,   0.00004,     0.077 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    103.76468,   0.00002,     0.066 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     66.92964,   0.00004,     0.501 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     96.50738,   0.00004,     0.089 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },     84.10440,   0.00012,     0.105 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     83.49278,   0.00017,     0.044 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   2,   1,   0 },     83.70399,   0.00001,     0.028 },
 { {   0,   8, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    220.52410,   0.00001,     0.075 },
 { {   0,   8, -15,   0,   1,   0,   0,   0,  -2,   3,   0 },    220.52410,   0.00001,     0.076 },
 { {   0,   8, -15,   0,   2,   0,   0,   0,  -2,   0,   0 },    322.89803,   0.00001,     0.038 },
 { {   0,   8, -15,   0,   2,   0,   0,   0,  -2,   1,   0 },    144.67237,   0.00003,     0.075 },
 { {   0,   8, -15,   0,   3,   0,   0,   0,  -2,   1,   0 },    311.04069,   0.00002,     0.076 },
 { {   0,   8, -15,   0,   3,   0,   0,   0,  -2,   3,   0 },    311.04069,   0.00002,     0.075 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    196.53620,   0.00002,     0.026 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.91695,   0.00017,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   2 },    160.84741,   0.00001,     0.900 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    168.49561,   0.00011,     0.080 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    338.72447,   0.00001,     0.075 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    342.43482,   0.00030,     0.070 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    339.88725,   0.00297,     1.004 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    339.14058,   0.00092,     0.082 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   2,   0 },    338.52567,   0.00007,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    342.19956,   0.00010,     0.361 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    340.86790,   0.00265,     0.095 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,  -2 },    340.77699,   0.00003,     0.322 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    340.02778,   0.00242,     0.042 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    339.65027,   0.00024,     0.027 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   2,   0 },    340.37887,   0.00002,     0.020 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,  -2,   0 },    341.02069,   0.00005,     0.046 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,  -1,   0 },    340.12333,   0.00010,     0.028 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,   0,   0 },    339.86955,   0.00006,     0.021 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,   1,   0 },    160.78100,   0.00001,     0.016 },
 { {   0,   8, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.72472,   0.00002,     0.039 },
 { {   0,   8, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },    322.05580,   0.00001,     0.081 },
 { {   0,   8, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },    141.72472,   0.00002,     1.159 },
 { {   0,   8, -14,   0,   1,   0,   0,   0,  -2,   3,   0 },    142.05580,   0.00001,     0.071 },
 { {   0,   8, -14,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.54343,   0.00001,     0.039 },
 { {   0,   8, -14,   0,   2,   0,   0,   0,  -2,   2,   0 },     51.51862,   0.00001,     1.056 },
 { {   0,   8, -14,   0,   2,   0,   0,   0,   0,  -1,   0 },    231.62418,   0.00002,     0.071 },
 { {   0,   8, -14,   0,   2,   0,   0,   0,   0,   1,   0 },    231.62418,   0.00002,     0.080 },
 { {   0,   8, -14,   1,   0,   0,   0,   0,   0,   0,   0 },     57.72855,   0.00001,     2.155 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   0,   0 },    233.91101,   0.00004,     0.041 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   1,   0 },     51.80825,   0.00001,     0.088 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.74157,   0.00003,     0.543 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   3,   0 },     51.80825,   0.00001,     0.066 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,   0,  -1,   0 },     42.04044,   0.00002,     0.076 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,   0,   1,   0 },     42.04044,   0.00002,     0.075 },
 { {   0,   8, -13,   0,  -1,   0,   0,   0,   0,  -1,   0 },    151.61017,   0.00002,     0.075 },
 { {   0,   8, -13,   0,  -1,   0,   0,   0,   0,   1,   0 },    151.61017,   0.00002,     0.076 },
 { {   0,   8, -13,   0,   0,  -1,   0,   0,   0,  -1,   0 },     55.48736,   0.00001,     0.075 },
 { {   0,   8, -13,   0,   0,  -1,   0,   0,   0,   1,   0 },     55.48736,   0.00001,     0.076 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -6,   1,   0 },    236.61541,   0.00001,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -6,   2,   0 },    236.94322,   0.00002,     0.021 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,  -1,   0 },    233.07350,   0.00005,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   0,   0 },    234.82207,   0.00027,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   1,   0 },    237.51551,   0.00072,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   1,   2 },    237.88744,   0.00001,     0.107 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    236.54948,   0.00051,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   3,   0 },    234.33814,   0.00001,     0.103 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -3,   1,   0 },     57.30147,   0.00003,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -3,   0 },    280.67275,   0.00002,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -2,   0 },    236.10963,   0.00019,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,  -2 },     57.56075,   0.00001,     0.015 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    237.05628,   0.00227,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   2 },     57.18680,   0.00006,     0.090 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,  -2 },     56.92285,   0.00005,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    237.41676,   0.02418,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   2 },    237.88744,   0.00055,     0.474 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,  -2 },     57.95869,   0.00009,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    238.00817,   0.03878,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   2 },    236.32780,   0.00002,     0.065 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    241.53627,   0.00152,     0.562 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    240.35755,   0.00011,     0.067 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,  -1,   0 },     55.56603,   0.00007,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },     55.68143,   0.00065,     0.081 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   1,   0 },     59.35696,   0.00009,     1.122 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -3,   0 },    232.11210,   0.00018,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,  -2 },     54.21420,   0.00002,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    233.24949,   0.00263,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,  -2 },     54.36152,   0.00010,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    233.44625,   0.03878,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   2 },    232.42650,   0.00006,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,  -2 },     57.93744,   0.00011,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    235.75013,   0.22653,   238.919 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   2 },     57.96760,   0.00011,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,  -2 },    232.81680,   0.00006,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    233.44527,   0.03874,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   2 },     54.36152,   0.00010,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    233.46584,   0.00262,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   2 },     54.27133,   0.00002,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   3,   0 },    233.31325,   0.00018,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,  -1,   0 },     59.45463,   0.00009,     1.133 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   0,   0 },     58.12277,   0.00064,     0.081 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   1,   0 },     57.06079,   0.00007,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },    239.76901,   0.00011,     0.067 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    239.14661,   0.00149,     0.565 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,  -2 },    236.57729,   0.00002,     0.065 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    237.96935,   0.03850,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   2 },     57.91031,   0.00009,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,  -2 },    237.85679,   0.00055,     0.475 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    237.22280,   0.02393,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   2 },     56.71492,   0.00005,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,  -2 },     57.13173,   0.00006,     0.090 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,   0 },    236.29815,   0.00226,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,   2 },     57.38709,   0.00001,     0.015 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   2,   0 },    235.81563,   0.00018,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   3,   0 },    237.43235,   0.00002,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   3,  -1,   0 },     57.85682,   0.00003,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   3,   0,   0 },    238.67899,   0.00002,     0.027 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -3,   0 },    237.32225,   0.00001,     0.103 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -2,   0 },    237.95909,   0.00051,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -1,  -2 },    237.85449,   0.00001,     0.107 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -1,   0 },    237.32032,   0.00071,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   0,   0 },    234.65369,   0.00027,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   1,   0 },    236.48167,   0.00005,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   6,  -2,   0 },    237.29957,   0.00002,     0.021 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   6,  -1,   0 },    236.56764,   0.00001,     0.016 },
 { {   0,   8, -13,   0,   0,   1,   0,   0,   0,  -1,   0 },    231.33101,   0.00002,     0.076 },
 { {   0,   8, -13,   0,   0,   1,   0,   0,   0,   1,   0 },    231.33101,   0.00002,     0.075 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.44888,   0.00003,     0.041 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,  -2,   3,   0 },    229.98944,   0.00002,     0.066 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,  -2,   0 },    230.42338,   0.00001,     0.038 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,  -1,   0 },    160.86795,   0.00003,     0.076 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,   0,   0 },     50.42338,   0.00001,    11.301 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,   1,   0 },    160.86795,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   2,   0,   0,   0,  -2,   0,   0 },    111.91072,   0.00003,     0.041 },
 { {   0,   8, -13,   0,   2,   0,   0,   0,   0,  -1,   0 },    292.34976,   0.00002,     0.076 },
 { {   0,   8, -13,   0,   2,   0,   0,   0,   0,   0,   0 },     66.15411,   0.00001,     5.787 },
 { {   0,   8, -13,   0,   2,   0,   0,   0,   0,   1,   0 },    292.34976,   0.00002,     0.074 },
 { {   0,   8, -13,   2,   0,   0,   0,   0,  -2,   0,   0 },     95.37364,   0.00001,     0.042 },
 { {   0,   8, -13,   2,   0,   0,   0,   0,  -2,   2,   0 },     95.37364,   0.00001,     0.352 },
 { {   0,   8, -13,   2,   0,   0,   0,   0,   1,   0,  -1 },    248.76653,   0.00003,    72.525 },
 { {   0,   8, -12,  -2,   0,   0,   0,   0,   0,  -1,   0 },    282.32972,   0.00002,     0.075 },
 { {   0,   8, -12,  -2,   0,   0,   0,   0,   0,   1,   0 },    282.32972,   0.00002,     0.076 },
 { {   0,   8, -12,  -1,   0,   0,   0,   0,   0,   0,   0 },     81.34500,   0.00001,     2.116 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    242.39647,   0.00005,     0.042 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   1,   0 },     60.43680,   0.00001,     0.095 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },    240.43201,   0.00003,     0.371 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   3,   0 },    240.43680,   0.00001,     0.063 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,   0,  -1,   0 },     60.43680,   0.00001,     0.081 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,   0,   1,   0 },     60.43680,   0.00001,     0.071 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -4,   0,   0 },    127.99202,   0.00003,     0.021 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -4,   1,   0 },    122.52248,   0.00006,     0.028 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },    146.86481,   0.00001,     0.046 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    111.72135,   0.00014,     0.027 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,  -1,   2 },    231.76818,   0.00001,     0.099 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     99.95690,   0.00108,     0.042 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,   2 },    178.21671,   0.00001,     0.321 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     91.84276,   0.00102,     0.095 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,   2 },    231.76818,   0.00001,     0.061 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },     64.73554,   0.00006,     0.360 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    226.47710,   0.00002,     0.062 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,  -1,   1 },    333.73592,   0.00002,     0.089 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    153.72609,   0.00001,     0.040 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },    153.46828,   0.00002,     0.087 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },      4.92585,   0.00004,     0.039 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },      7.20604,   0.00070,     0.082 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    350.36883,   0.00388,     0.996 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   1,  -2 },    231.76818,   0.00003,     0.079 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    342.84084,   0.00083,     0.070 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    333.21241,   0.00008,     0.036 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -1,  -1 },    333.46325,   0.00005,     0.075 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },    135.38098,   0.00002,     8.534 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -1,   1 },    154.04091,   0.00004,     0.074 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    153.45595,   0.00028,    20.170 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    155.82241,   0.00016,     0.075 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,   1 },    333.48140,   0.00050,     0.037 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   1,  -1 },    333.46325,   0.00005,     0.076 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   1,   1 },    333.47607,   0.00005,     0.025 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    139.80500,   0.00003,     1.299 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    328.37378,   0.00040,     0.080 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   0,  -2 },    325.29317,   0.00002,     0.906 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    330.95691,   0.00038,     0.039 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   1,   0 },    324.37803,   0.00002,     0.026 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   3,   0,  -1 },    333.47607,   0.00002,     0.041 },
 { {   0,   8, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },     39.03936,   0.00007,     0.042 },
 { {   0,   8, -12,   0,   1,   0,   0,   0,  -2,   2,   0 },     38.05158,   0.00001,     0.349 },
 { {   0,   8, -11,  -4,   0,   0,   0,   0,   0,   0,   0 },     35.34457,   0.00001,     8.160 },
 { {   0,   8, -11,  -2,   0,   0,   0,   0,  -2,   0,   0 },    191.19158,   0.00002,     0.042 },
 { {   0,   8, -11,  -2,   0,   0,   0,   0,  -2,   2,   0 },    191.19158,   0.00002,     0.368 },
 { {   0,   8, -11,   0,  -1,   0,   0,   0,  -2,   2,   0 },     52.49172,   0.00001,     0.271 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -4,   0,   0 },    119.58915,   0.00002,     0.021 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,  -2,   0 },     64.68732,   0.00001,     0.020 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    307.99636,   0.00007,     0.044 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    281.91444,   0.00008,     0.106 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    322.02983,   0.00002,     0.265 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    256.01605,   0.00020,     0.089 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     79.71820,   0.00009,     0.499 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    247.87639,   0.00010,     1.052 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },     52.74876,   0.00002,     0.077 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    255.42287,   0.00003,     4.345 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },     76.07381,   0.00097,     0.074 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },     66.35732,   0.00005,     0.075 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    249.57390,   0.00002,     0.037 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   4,  -2,   0 },     76.43212,   0.00001,     0.040 },
 { {   0,   8, -11,   0,   0,   3,   0,   0,   2,   0,  -2 },    176.95398,   0.00006,   666.222 },
 { {   0,   8, -11,   0,   1,   0,   0,   0,   2,  -1,   0 },    349.97845,   0.00003,     0.074 },
 { {   0,   8, -11,   0,   2,   0,   0,   0,   2,  -2,   0 },     91.15031,   0.00001,     2.508 },
 { {   0,   8, -11,   0,   2,   0,   0,   0,   2,   0,  -2 },     91.15031,   0.00001,    15.323 },
 { {   0,   8, -11,   0,   2,   0,   0,   0,   2,   0,   0 },    271.15031,   0.00001,     0.037 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    162.84295,   0.00002,     0.046 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    168.44358,   0.00003,     0.118 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    335.29463,   0.00001,     0.209 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },      3.77679,   0.00002,     0.246 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    186.14487,   0.00005,     0.098 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    343.30806,   0.00075,     0.333 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    346.91804,   0.00012,     0.062 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    349.67325,   0.00002,     0.034 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    168.83439,   0.00001,     0.088 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,  -2 },    349.94455,   0.00002,     0.087 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.83439,   0.00021,     0.513 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    164.20843,   0.00003,     0.065 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   1,  -1 },    168.83439,   0.00001,     0.066 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    345.34885,   0.00002,     0.083 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    351.80220,   0.00003,     0.813 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    351.47067,   0.00104,     0.069 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,  -2 },    350.96508,   0.00004,     1.115 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    173.86039,   0.00024,     0.036 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   1,   0 },    170.26325,   0.00003,     0.024 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,  -2,  -2 },    168.65656,   0.00002,     0.075 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,  -1,  -2 },    349.78408,   0.00008,   103.674 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,  -1,  -1 },    168.88498,   0.00005,     0.074 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,   0,  -2 },    169.94454,   0.00012,     0.075 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   4,  -2,   0 },    349.38813,   0.00002,     0.039 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   4,  -1,   0 },    171.08730,   0.00001,     0.025 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   4,   0,   0 },    170.84468,   0.00003,     0.019 },
 { {   0,   8, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.45625,   0.00001,     0.046 },
 { {   0,   8, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.45625,   0.00001,     0.206 },
 { {   0,   8, -10,   0,   1,   0,   0,   0,   2,  -1,   0 },    269.97700,   0.00003,     0.069 },
 { {   0,   8, -10,   0,   1,   0,   0,   0,   2,   0,   0 },     89.97700,   0.00003,     0.036 },
 { {   0,   8, -10,   0,   3,   0,   0,   0,   0,   0,   0 },    168.80855,   0.00001,     0.307 },
 { {   0,   8, -10,   2,   0,   0,   0,   0,   2,  -3,   0 },    180.06053,   0.00001,     0.091 },
 { {   0,   8, -10,   2,   0,   0,   0,   0,   2,  -1,   0 },    180.06053,   0.00001,     0.064 },
 { {   0,   8, -10,   4,   0,   0,   0,   0,  -2,   0,   0 },     57.95233,   0.00001,     0.051 },
 { {   0,   8, -10,   4,   0,   0,   0,   0,  -2,   2,   0 },     57.95233,   0.00001,     0.145 },
 { {   0,   8, -10,   4,   0,   0,   0,   0,   0,   0,   0 },     57.63520,   0.00001,     0.195 },
 { {   0,   8,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.49748,   0.00002,     0.048 },
 { {   0,   8,  -9,   0,  -1,   0,   0,   0,   0,   0,   0 },    353.07234,   0.00001,     0.255 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.40191,   0.00004,     0.048 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     87.18922,   0.00008,     0.134 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    275.43402,   0.00001,     0.173 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },     75.95030,   0.00003,     0.046 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.40841,   0.00001,     0.120 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },    255.94317,   0.00013,     0.198 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   1,  -1 },     75.96678,   0.00002,     0.117 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     95.47770,   0.00007,     0.108 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    271.86537,   0.00226,     0.250 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    272.45045,   0.00047,     0.058 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    289.55645,   0.00002,     0.033 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   3,   0 },    273.19541,   0.00002,     0.023 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },     75.95853,   0.00010,     0.097 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },     91.97263,   0.00001,     0.321 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },    254.28392,   0.00004,     0.060 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },     75.97900,   0.00063,     0.339 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     90.79187,   0.00005,     0.061 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },     75.94837,   0.00006,     0.062 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    283.15383,   0.00006,     0.091 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    245.97426,   0.00008,     0.448 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.64014,   0.00267,     0.065 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31544,   0.00008,     0.527 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    255.93574,   0.00002,     0.065 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },     95.07514,   0.00134,     0.035 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },     94.37263,   0.00020,     0.024 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },     92.96444,   0.00002,     0.018 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   3,   0,   0 },     92.43333,   0.00002,     0.024 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,  -3,   0 },    102.53124,   0.00001,     0.073 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    283.41709,   0.00003,     0.037 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },     98.43039,   0.00005,     0.019 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },    192.74983,   0.00009,     0.048 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,   0,   0,   0 },      7.15018,   0.00003,     0.245 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,   2,  -1,   0 },    178.69890,   0.00005,     0.064 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,   2,   0,   0 },    178.54098,   0.00003,     0.035 },
 { {   0,   8,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.65672,   0.00001,     0.049 },
 { {   0,   8,  -9,   0,   2,   0,   0,   0,  -2,   2,   0 },    232.65672,   0.00001,     0.168 },
 { {   0,   8,  -9,   0,   2,   0,   0,   0,   2,  -2,   0 },    240.16300,   0.00001,     0.417 },
 { {   0,   8,  -9,   0,   2,   0,   0,   0,   2,   0,   0 },     60.16300,   0.00001,     0.035 },
 { {   0,   8,  -9,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.38980,   0.00001,     0.052 },
 { {   0,   8,  -9,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.38980,   0.00001,     0.136 },
 { {   0,   8,  -8,   0,  -2,   0,   0,   0,   0,   0,   0 },    201.54403,   0.00001,     0.207 },
 { {   0,   8,  -8,   0,  -1,   0,   0,   0,  -2,   0,   0 },    185.87168,   0.00007,     0.050 },
 { {   0,   8,  -8,   0,  -1,   0,   0,   0,  -2,   1,   0 },    299.10731,   0.00001,     0.152 },
 { {   0,   8,  -8,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.57318,   0.00001,     0.149 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    179.69950,   0.00002,     0.022 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.24969,   0.00002,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    179.81267,   0.00002,     0.056 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -4,   3,   0 },    180.44228,   0.00001,     0.213 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.37099,   0.00006,     0.030 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.37377,   0.00042,     0.051 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.86514,   0.00031,     0.154 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.49819,   0.00008,     0.148 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },      0.71826,   0.00001,     0.050 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.91455,   0.00011,     0.136 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },     17.33381,   0.00001,     0.165 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.06464,   0.00001,     0.170 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -3,   0 },      1.42510,   0.00002,     0.029 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    184.43161,   0.00001,     0.046 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.37418,   0.00052,     0.121 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,  -2 },    179.11327,   0.00001,     0.046 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    179.96606,   0.01070,     0.200 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    180.04475,   0.00180,     0.055 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    179.31585,   0.00017,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    359.99015,   0.00003,     0.243 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    190.47185,   0.00015,     0.253 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },      0.01680,   0.00016,     0.058 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },    180.05684,   0.00001,     0.033 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    186.37966,   0.00001,     0.100 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.54345,   0.00037,     0.310 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.99395,   0.00002,     0.097 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.03702,   0.00824,     0.061 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   2 },      0.01132,   0.00002,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.99081,   0.00025,     0.345 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },      0.04498,   0.00571,     0.034 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00359,   0.00002,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },      0.00802,   0.00048,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   2,   0 },      0.00349,   0.00007,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,  -4,   0 },      0.05685,   0.00001,     0.092 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.05254,   0.00001,     0.035 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,   0,   0 },      0.05686,   0.00004,     0.024 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,   1,   0 },      0.05682,   0.00001,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.62544,   0.00012,     0.036 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.04773,   0.00005,     0.024 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },    359.99998,   0.00010,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,   1,   0 },      0.00465,   0.00002,     0.015 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   6,  -1,   0 },      0.01731,   0.00001,     0.015 },
 { {   0,   8,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },    348.21364,   0.00007,     0.051 },
 { {   0,   8,  -8,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.70961,   0.00002,     0.156 },
 { {   0,   8,  -8,   0,   1,   0,   0,   0,   0,  -2,   0 },    351.24228,   0.00001,     0.047 },
 { {   0,   8,  -8,   0,   1,   0,   0,   0,   0,   0,   0 },    171.24228,   0.00001,     0.197 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.60188,   0.00001,     0.053 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,  -2,   1,   0 },    181.25569,   0.00001,     0.180 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.60188,   0.00001,     0.130 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,  -2,   3,   0 },    181.25569,   0.00001,     0.048 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.50798,   0.00007,     0.169 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,   2,   0,   0 },      3.81203,   0.00002,     0.033 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    104.47861,   0.00001,     0.053 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.07549,   0.00002,     0.183 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    256.70854,   0.00003,     0.129 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.25820,   0.00002,     0.030 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.25037,   0.00002,     0.161 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.21353,   0.00014,     0.051 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.24149,   0.00017,     0.142 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.26048,   0.00007,     0.153 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },     77.40470,   0.00003,     0.049 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    229.84838,   0.00001,     0.138 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.27818,   0.00001,     0.135 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     65.42057,   0.00003,     0.052 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.29977,   0.00011,     0.120 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.30693,   0.00003,     0.054 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.35529,   0.00060,     0.202 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    103.30005,   0.00002,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.30703,   0.00005,     0.055 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   1,   1 },    103.30689,   0.00002,     0.022 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   2,   1 },    103.30718,   0.00002,     0.017 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.78034,   0.00021,     0.057 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     90.31023,   0.00030,     0.033 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },     86.78562,   0.00005,     0.023 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },     89.80743,   0.00001,     0.024 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },     83.36842,   0.00003,     0.018 },
 { {   0,   8,  -6,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.21094,   0.00001,     0.052 },
 { {   0,   8,  -6,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.21094,   0.00001,     0.139 },
 { {   0,   8,  -6,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.54285,   0.00001,     0.053 },
 { {   0,   8,  -6,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.54285,   0.00001,     0.130 },
 { {   0,   8,  -6,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.57396,   0.00001,     0.056 },
 { {   0,   8,  -6,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.57396,   0.00001,     0.117 },
 { {   0,   8,  -6,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.29187,   0.00002,     0.116 },
 { {   0,   8,  -6,   0,  -2,   0,   0,   0,   0,  -2,   0 },      0.15984,   0.00001,     0.051 },
 { {   0,   8,  -6,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.81983,   0.00004,     0.146 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.48057,   0.00002,     0.032 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     26.52442,   0.00021,     0.056 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.06900,   0.00012,     0.223 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.01648,   0.00002,     0.186 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    342.15401,   0.00001,     0.051 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.88444,   0.00013,     0.160 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     26.02170,   0.00061,     0.143 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     25.87580,   0.00005,     0.049 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    282.56674,   0.00001,     0.191 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     22.16742,   0.00008,     0.054 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    214.13598,   0.00010,     0.032 },
 { {   0,   8,  -5,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.66917,   0.00001,     0.059 },
 { {   0,   8,  -5,   0,  -3,   0,   0,   0,  -2,   2,   0 },    176.66917,   0.00001,     0.105 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    277.30794,   0.00002,     0.288 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    283.91703,   0.00003,     0.125 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.47811,   0.00002,     0.161 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     38.76000,   0.00002,     0.051 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     39.47811,   0.00002,     0.031 },
 { {   0,   8,  -4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    323.43296,   0.00001,     0.140 },
 { {   0,   8,  -4,   0,  -1,   0,   0,   0,   2,   0,   0 },    220.07058,   0.00002,     0.030 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    308.19431,   0.00001,     0.235 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    289.76288,   0.00002,     0.111 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    308.20821,   0.00002,     0.028 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.06931,   0.00002,     0.049 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    129.00962,   0.00003,     0.030 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    128.11240,   0.00003,     0.021 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    128.20822,   0.00002,     0.017 },
 { {   0,   8,  -3,   0,  -2,   0,   0,   0,   2,  -2,   0 },    248.02064,   0.00001,     0.124 },
 { {   0,   8,  -3,   0,  -2,   0,   0,   0,   2,   0,   0 },     68.02064,   0.00001,     0.029 },
 { {   0,   8,  -3,   0,  -1,   0,   0,   0,   2,  -3,   0 },    129.50798,   0.00001,     0.196 },
 { {   0,   8,  -3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    309.50798,   0.00001,     0.047 },
 { {   0,   8,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },    129.23058,   0.00002,     0.029 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.10438,   0.00001,     0.085 },
 { {   0,   8,  -2,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.30651,   0.00001,     0.110 },
 { {   0,   8,  -2,   0,  -2,   0,   0,   0,   2,   0,   0 },    308.30651,   0.00001,     0.028 },
 { {   0,   8,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    334.03187,   0.00002,     0.108 },
 { {   0,   8,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    334.03187,   0.00002,     0.028 },
 { {   0,   8,  -1,  -1,   0,   0,   0,   0,  -2,   1,   0 },     63.28239,   0.00002,   108.269 },
 { {   0,   8,  -1,   0,  -6,   0,   0,   0,  -2,   1,   0 },    253.39248,   0.00003,    60.164 },
 { {   0,   9, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.61843,   0.00001,     0.036 },
 { {   0,   9, -18,   3,   0,   0,   0,   0,  -2,   2,   0 },    234.59025,   0.00020,   818.910 },
 { {   0,   9, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    208.11767,   0.00003,     0.072 },
 { {   0,   9, -17,   0,   3,   0,   0,   0,  -2,   0,   2 },    358.66248,   0.00005,   101.557 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    302.15500,   0.00003,     0.078 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   3,   0 },    268.30028,   0.00001,     0.073 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    306.94222,   0.00002,     0.730 },
 { {   0,   9, -16,   0,   2,   0,   0,   0,  -2,   1,   0 },    143.02614,   0.00002,     0.079 },
 { {   0,   9, -16,   0,   2,   0,   0,   0,  -2,   3,   0 },    323.02614,   0.00002,     0.072 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },    223.57250,   0.00001,     0.026 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    224.86748,   0.00001,     0.040 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    238.53156,   0.00004,     0.084 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    224.86748,   0.00001,     0.712 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },     59.19834,   0.00003,     0.073 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1,   2 },    223.57250,   0.00002,     0.076 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    238.24462,   0.00003,     2.701 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     59.19834,   0.00003,     0.078 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.77009,   0.00002,     0.040 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    141.77009,   0.00001,     0.085 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },    141.77009,   0.00002,     0.672 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,  -2,   3,   0 },    321.77009,   0.00001,     0.068 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,   0,  -1,   0 },    142.09657,   0.00002,     0.074 },
 { {   0,   9, -15,   0,   1,   0,   0,   0,   0,   1,   0 },    142.09657,   0.00002,     0.077 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.57187,   0.00003,     0.040 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,  -2,   1,   0 },     51.58703,   0.00006,     0.086 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,  -2,   2,   0 },     51.51245,   0.00001,     0.636 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,   0,  -1,   0 },     51.58980,   0.00002,     0.074 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,   0,   1,   0 },     51.58980,   0.00002,     0.077 },
 { {   0,   9, -15,   0,   3,   0,   0,   0,   0,  -1,   0 },    221.57463,   0.00003,     0.075 },
 { {   0,   9, -15,   0,   3,   0,   0,   0,   0,   1,   0 },    221.57463,   0.00003,     0.076 },
 { {   0,   9, -15,   0,   4,   0,   0,   0,   0,  -1,   0 },     32.23700,   0.00001,     0.075 },
 { {   0,   9, -15,   0,   4,   0,   0,   0,   0,   1,   0 },     32.23700,   0.00001,     0.076 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.74770,   0.00004,     0.042 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.72800,   0.00003,     0.405 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,   0,  -1,   0 },     51.79376,   0.00001,     0.080 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,   0,   1,   0 },     51.79376,   0.00001,     0.072 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -4,   2,   0 },    112.96262,   0.00001,     0.045 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.05464,   0.00004,     0.027 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    304.68850,   0.00007,     0.041 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    260.81165,   0.00003,     0.092 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    292.96262,   0.00007,     0.416 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    226.87554,   0.00002,     0.079 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    185.04520,   0.00005,     1.588 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    226.87554,   0.00002,     0.072 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   2,  -3,   0 },     57.87953,   0.00003,     0.069 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     57.87953,   0.00003,     0.083 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.38725,   0.00003,     0.042 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },    230.37061,   0.00001,     0.093 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,  -2,   3,   0 },     50.38725,   0.00003,     0.064 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,   0,  -1,   0 },    230.38725,   0.00004,     0.080 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,   0,   1,   0 },    230.38725,   0.00004,     0.072 },
 { {   0,   9, -14,   0,   2,   0,   0,   0,  -2,   0,   0 },    111.89623,   0.00003,     0.042 },
 { {   0,   9, -14,   2,   0,   0,   0,   0,  -2,   0,   0 },     95.35914,   0.00001,     0.043 },
 { {   0,   9, -14,   2,   0,   0,   0,   0,  -2,   2,   0 },     95.35914,   0.00001,     0.288 },
 { {   0,   9, -13,  -3,   0,   0,   0,   0,   0,  -1,   0 },    266.56794,   0.00003,     0.076 },
 { {   0,   9, -13,  -3,   0,   0,   0,   0,   0,   1,   0 },    266.56794,   0.00003,     0.075 },
 { {   0,   9, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    241.43970,   0.00008,     0.043 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     42.75384,   0.00003,     0.043 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    229.52965,   0.00002,     0.102 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },     39.89832,   0.00001,     0.294 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    231.77865,   0.00003,     0.060 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    184.14743,   0.00006,     0.086 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    168.80575,   0.00003,     0.614 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    231.77864,   0.00003,     0.067 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    333.07692,   0.00002,     1.736 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },     61.45742,   0.00004,     0.075 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    336.46775,   0.00004,     6.930 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    340.43404,   0.00024,     0.076 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,   0,   0 },     75.10642,   0.00002,     0.038 },
 { {   0,   9, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },     38.84072,   0.00007,     0.043 },
 { {   0,   9, -13,   0,   1,   0,   0,   0,   2,  -3,   0 },     36.65135,   0.00001,     0.075 },
 { {   0,   9, -13,   0,   1,   0,   0,   0,   2,  -1,   0 },    295.97874,   0.00002,     0.076 },
 { {   0,   9, -13,   0,   2,   0,   0,   0,   2,  -2,   0 },     15.67061,   0.00001,    41.153 },
 { {   0,   9, -12,  -2,   0,   0,   0,   0,  -2,   0,   0 },    191.17708,   0.00001,     0.043 },
 { {   0,   9, -12,  -2,   0,   0,   0,   0,  -2,   2,   0 },    191.17708,   0.00001,     0.299 },
 { {   0,   9, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },     53.65361,   0.00006,     0.045 },
 { {   0,   9, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },     50.80774,   0.00002,     0.232 },
 { {   0,   9, -12,   0,  -1,   0,   0,   0,   0,   0,   0 },    232.52583,   0.00002,     0.393 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    316.72830,   0.00002,     0.045 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    113.38005,   0.00002,     0.113 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    300.04424,   0.00003,     0.227 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    255.25519,   0.00006,     0.094 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     66.72645,   0.00016,     0.380 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    249.69559,   0.00005,     0.635 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },     82.16472,   0.00001,     1.169 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },     75.25844,   0.00024,     0.071 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    255.21482,   0.00006,     0.037 },
 { {   0,   9, -12,   0,   2,   0,   0,   0,   2,  -1,   0 },     91.19568,   0.00002,     0.070 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    197.79685,   0.00002,     0.047 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    341.84409,   0.00003,     0.185 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },     23.95418,   0.00001,     0.054 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },      8.27502,   0.00002,     0.214 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    316.65511,   0.00002,     0.104 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    344.55175,   0.00057,     0.276 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    348.22145,   0.00009,     0.059 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.78210,   0.00014,     0.388 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    165.11012,   0.00002,     0.063 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    351.68886,   0.00002,     0.539 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    352.73488,   0.00061,     0.066 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,  -2 },    351.93690,   0.00002,     0.657 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    173.67934,   0.00024,     0.035 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   1,   0 },    170.94579,   0.00003,     0.024 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   4,  -1,   0 },    171.73322,   0.00001,     0.025 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   4,   0,   0 },    171.82275,   0.00002,     0.019 },
 { {   0,   9, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.51484,   0.00001,     0.048 },
 { {   0,   9, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.51484,   0.00001,     0.182 },
 { {   0,   9, -11,   0,   1,   0,   0,   0,   2,   0,   0 },     90.02237,   0.00002,     0.035 },
 { {   0,   9, -11,   0,   2,   0,   0,   0,   2,  -2,   0 },    179.82719,   0.00002,     0.494 },
 { {   0,   9, -11,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.19091,   0.00001,     0.050 },
 { {   0,   9, -11,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.19091,   0.00001,     0.155 },
 { {   0,   9, -11,   2,   0,   0,   0,   0,   0,   0,   0 },      0.79419,   0.00001,     0.213 },
 { {   0,   9, -11,   2,   0,   0,   0,   0,   2,  -2,   0 },      0.04604,   0.00002,     0.343 },
 { {   0,   9, -11,   2,   0,   0,   0,   0,   2,   0,   0 },      0.04604,   0.00002,     0.034 },
 { {   0,   9, -11,   4,   0,   0,   0,   0,  -2,   0,   0 },     57.94199,   0.00001,     0.053 },
 { {   0,   9, -11,   4,   0,   0,   0,   0,  -2,   2,   0 },     57.94199,   0.00001,     0.133 },
 { {   0,   9, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.24495,   0.00002,     0.050 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    275.88960,   0.00005,     0.050 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     74.99589,   0.00002,     0.146 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },     94.05825,   0.00002,     0.156 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    271.08005,   0.00001,     0.051 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },     75.74334,   0.00002,     0.047 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },    254.98484,   0.00009,     0.176 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   1,  -1 },     75.72402,   0.00002,     0.126 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    267.46816,   0.00004,     0.116 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    271.75113,   0.00168,     0.216 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    272.20175,   0.00038,     0.056 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    272.37503,   0.00003,     0.032 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },     81.28961,   0.00007,     0.103 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,  -1,   1 },    255.78168,   0.00003,     0.058 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },     75.77984,   0.00045,     0.280 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   0 },     90.02300,   0.00003,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   1 },     75.80206,   0.00001,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   1,  -1 },     83.44531,   0.00005,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    278.31879,   0.00002,     0.096 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    244.34099,   0.00005,     0.350 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.60821,   0.00179,     0.062 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.30447,   0.00006,     0.396 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,   0 },     94.08280,   0.00117,     0.034 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   1,   0 },     93.63096,   0.00017,     0.023 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   2,   0 },     92.61005,   0.00002,     0.018 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   3,   0,   0 },     92.43184,   0.00002,     0.024 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,  -2,   0 },    279.10498,   0.00003,     0.036 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,  -1,   0 },    266.28676,   0.00001,     0.024 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,   0,   0 },     98.02000,   0.00004,     0.018 },
 { {   0,   9, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },    210.57620,   0.00004,     0.050 },
 { {   0,   9, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },    358.41690,   0.00002,     0.154 },
 { {   0,   9, -10,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.64840,   0.00001,     0.050 },
 { {   0,   9, -10,   0,   2,   0,   0,   0,  -2,   2,   0 },    232.64840,   0.00001,     0.152 },
 { {   0,   9, -10,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.37531,   0.00001,     0.054 },
 { {   0,   9, -10,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.37531,   0.00001,     0.125 },
 { {   0,   9, -10,   3,   0,   0,   0,   0,   0,   0,   0 },    236.82460,   0.00001,     0.161 },
 { {   0,   9,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    184.27806,   0.00007,     0.052 },
 { {   0,   9,  -9,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.58298,   0.00001,     0.137 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.13291,   0.00001,     0.033 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.59930,   0.00002,     0.058 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -4,   3,   0 },    180.18228,   0.00001,     0.245 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.16919,   0.00003,     0.022 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    181.59344,   0.00002,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.53109,   0.00039,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.23788,   0.00027,     0.171 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.38354,   0.00009,     0.135 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    359.55760,   0.00001,     0.048 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },     14.72923,   0.00001,     0.050 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.90568,   0.00009,     0.148 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },     17.20539,   0.00001,     0.150 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.00282,   0.00051,     0.131 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    179.95637,   0.00748,     0.178 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   2 },      0.00668,   0.00001,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    180.19934,   0.00128,     0.053 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    180.78636,   0.00013,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },      3.43332,   0.00001,     0.211 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.19234,   0.00011,     0.219 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },      0.29876,   0.00007,     0.056 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    191.19165,   0.00002,     0.032 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },    180.04863,   0.00001,     0.032 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    177.75565,   0.00002,     0.106 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.93227,   0.00027,     0.259 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.99702,   0.00536,     0.058 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    180.00763,   0.00017,     0.284 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },      0.06211,   0.00451,     0.033 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00327,   0.00002,     0.017 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },    359.99851,   0.00038,     0.023 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },    359.96693,   0.00005,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.04628,   0.00002,     0.034 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   3,   0,   0 },      0.04702,   0.00004,     0.023 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.99928,   0.00008,     0.035 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.00228,   0.00005,     0.024 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },      0.00026,   0.00008,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,   1,   0 },      0.00181,   0.00001,     0.015 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },    348.22708,   0.00007,     0.053 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.68412,   0.00002,     0.173 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,   0,   0,   0 },    175.71238,   0.00002,     0.175 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,   2,  -2,   0 },    166.88391,   0.00001,     0.254 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,   2,   0,   0 },    166.88391,   0.00001,     0.033 },
 { {   0,   9,  -9,  11,   0,   0,   0,   0,  -2,   1,   0 },    301.33698,   0.00001,   285.530 },
 { {   0,   9,  -8,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.08656,   0.00002,     0.051 },
 { {   0,   9,  -8,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.08656,   0.00002,     0.147 },
 { {   0,   9,  -8,  -2,   0,   0,   0,   0,  -2,   1,   0 },    136.29509,   0.00001,     0.169 },
 { {   0,   9,  -8,  -2,   0,   0,   0,   0,   0,   0,   0 },    311.83097,   0.00001,     0.180 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.56671,   0.00001,     0.120 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,   0,  -2,   0 },    180.96183,   0.00002,     0.050 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.36808,   0.00009,     0.153 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,   2,   0,   0 },    181.29379,   0.00002,     0.032 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    255.05675,   0.00002,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.16885,   0.00001,     0.206 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.23243,   0.00002,     0.179 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.23341,   0.00012,     0.052 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.23103,   0.00015,     0.130 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.24478,   0.00006,     0.169 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   2,   1 },    283.29525,   0.00002,     0.029 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    206.21134,   0.00001,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    346.14970,   0.00001,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.28764,   0.00009,     0.130 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.29563,   0.00002,     0.052 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.30228,   0.00045,     0.179 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    103.29284,   0.00003,     0.031 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.29580,   0.00004,     0.053 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    269.12671,   0.00020,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     89.78093,   0.00030,     0.032 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },     92.05941,   0.00003,     0.022 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },     90.36836,   0.00001,     0.023 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },     84.67979,   0.00002,     0.018 },
 { {   0,   9,  -7,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.20460,   0.00001,     0.053 },
 { {   0,   9,  -7,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.20460,   0.00001,     0.128 },
 { {   0,   9,  -7,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.52016,   0.00001,     0.055 },
 { {   0,   9,  -7,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.52016,   0.00001,     0.120 },
 { {   0,   9,  -7,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.47393,   0.00001,     0.058 },
 { {   0,   9,  -7,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.47393,   0.00001,     0.109 },
 { {   0,   9,  -7,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.28325,   0.00002,     0.108 },
 { {   0,   9,  -7,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.88546,   0.00005,     0.134 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.00739,   0.00002,     0.033 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     24.89620,   0.00021,     0.058 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.22459,   0.00012,     0.260 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.09360,   0.00002,     0.211 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },     19.98513,   0.00012,     0.178 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     26.55564,   0.00047,     0.131 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     15.93140,   0.00004,     0.048 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    315.87475,   0.00002,     0.171 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.86781,   0.00007,     0.052 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    210.59618,   0.00008,     0.031 },
 { {   0,   9,  -6,  -4,   0,   0,   0,   0,   0,   0,   0 },    150.73154,   0.00001,     0.154 },
 { {   0,   9,  -6,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.65882,   0.00002,     0.061 },
 { {   0,   9,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },     24.10460,   0.00001,     0.062 },
 { {   0,   9,  -6,   0,  -1,   0,   0,   0,  -2,   2,   0 },    204.10460,   0.00001,     0.097 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    275.91338,   0.00002,     0.351 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    262.41277,   0.00004,     0.116 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },     39.36768,   0.00001,     0.156 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.47368,   0.00001,     0.146 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     39.36768,   0.00001,     0.050 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     39.47368,   0.00001,     0.030 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    217.16163,   0.00001,     0.540 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    307.95784,   0.00002,     0.276 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    308.20375,   0.00002,     0.104 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    308.20375,   0.00002,     0.028 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    129.26479,   0.00002,     0.127 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.26061,   0.00001,     0.047 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    127.74175,   0.00006,     0.029 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    128.26061,   0.00001,     0.021 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    128.22370,   0.00004,     0.016 },
 { {   0,   9,  -4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    309.93161,   0.00001,     0.114 },
 { {   0,   9,  -4,   0,  -1,   0,   0,   0,   2,   0,   0 },    309.93161,   0.00001,     0.028 },
 { {   0,   9,  -4,   0,   0,  -1,   0,   0,  -3,   2,   0 },    128.09357,   0.00005,   969.469 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.08492,   0.00002,     0.071 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.08492,   0.00002,     0.081 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     36.07082,   0.00002,     0.113 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     36.07082,   0.00002,     0.028 },
 { {   0,   9,  -3,   0,  -2,   0,   0,   0,  -2,   0,   0 },     87.06853,   0.00001,     0.075 },
 { {   0,   9,  -3,   0,  -2,   0,   0,   0,  -2,   1,   0 },    349.73737,   0.00007,    48.456 },
 { {   0,   9,  -3,   0,  -2,   0,   0,   0,  -2,   2,   0 },    267.06853,   0.00001,     0.076 },
 { {   0,   9,  -3,   0,  -2,   0,   0,   0,   2,  -2,   0 },    128.38012,   0.00001,     0.103 },
 { {   0,   9,  -3,   0,  -2,   0,   0,   0,   2,   0,   0 },    308.38012,   0.00001,     0.028 },
 { {   0,   9,  -3,   0,  -1,   0,   0,   0,  -2,   1,   0 },    354.01718,   0.00002,    15.707 },
 { {   0,   9,  -3,   0,   0,  -4,   0,   0,  -2,   1,   0 },    201.98556,   0.00003,    82.115 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     21.57050,   0.00001,     0.076 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     62.47479,   0.00004,     6.758 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },     21.57050,   0.00001,     0.075 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    334.11426,   0.00002,     0.027 },
 { {   0,   9,  -2,   0,  -3,   0,   0,   0,  -1,   0,   0 },     80.89653,   0.00001,   124.238 },
 { {   0,   9,  -1,  -4,   0,   0,   0,   0,  -2,   1,   0 },     65.97184,   0.00004,    47.095 },
 { {   0,   9,  -1,   0,  -4,   0,   0,   0,   0,  -1,   0 },     29.00378,   0.00001,    27.222 },
 { {   0,   9,   8, -21,   0,   0,   0,   0,  -2,   1,   0 },    243.48136,   0.00003,    57.479 },
 { {   0,  10, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.61529,   0.00002,     0.036 },
 { {   0,  10, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.61592,   0.00001,     1.030 },
 { {   0,  10, -18,   0,   0,  -1,   0,   0,  -2,   2,   0 },    122.03949,   0.00001,   212.821 },
 { {   0,  10, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     66.91061,   0.00002,     0.076 },
 { {   0,  10, -18,   0,   0,   0,   0,   0,  -2,   3,   0 },    352.58192,   0.00001,     0.075 },
 { {   0,  10, -18,   0,   2,   0,   0,   0,  -2,   1,   0 },     44.20260,   0.00001,     0.077 },
 { {   0,  10, -18,   0,   2,   0,   0,   0,  -2,   3,   0 },     44.20260,   0.00001,     0.074 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    307.72584,   0.00002,     1.343 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    224.43386,   0.00001,     0.026 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    224.43386,   0.00001,     0.089 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    279.13835,   0.00002,     3.918 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },     44.43386,   0.00002,     0.085 },
 { {   0,  10, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.81085,   0.00003,     0.041 },
 { {   0,  10, -16,   0,   1,   0,   0,   0,   0,   2,  -2 },    321.81085,   0.00001,   166.045 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.54053,   0.00002,     0.041 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,  -2,   1,   0 },    231.56564,   0.00002,     0.090 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,  -2,   2,   0 },     51.54053,   0.00002,     0.455 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,  -2,   3,   0 },    231.56564,   0.00002,     0.065 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,   0,  -1,   0 },     51.54054,   0.00002,     0.078 },
 { {   0,  10, -16,   0,   2,   0,   0,   0,   0,   1,   0 },     51.54054,   0.00002,     0.073 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.06036,   0.00003,     0.027 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    357.63702,   0.00001,     0.043 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    328.78848,   0.00001,     0.098 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    339.27037,   0.00003,     0.330 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    173.31972,   0.00002,     0.797 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },     55.65212,   0.00001,     4.964 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },    237.88051,   0.00004,     0.073 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    237.88051,   0.00004,     0.079 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     57.91167,   0.00002,     0.038 },
 { {   0,  10, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.36742,   0.00003,     0.043 },
 { {   0,  10, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },     50.39233,   0.00004,     0.099 },
 { {   0,  10, -15,   0,   1,   0,   0,   0,   0,  -1,   0 },     50.39233,   0.00001,     0.084 },
 { {   0,  10, -15,   0,   1,   0,   0,   0,   0,   1,   0 },     50.39233,   0.00001,     0.069 },
 { {   0,  10, -15,   0,   2,   0,   0,   0,  -2,   0,   0 },    111.87966,   0.00002,     0.043 },
 { {   0,  10, -15,   1,   0,   0,   0,   0,   2,  -2,   0 },    231.20265,   0.00001,    77.527 },
 { {   0,  10, -15,   2,   0,   0,   0,   0,  -2,   0,   0 },     95.34258,   0.00001,     0.045 },
 { {   0,  10, -15,   2,   0,   0,   0,   0,  -2,   2,   0 },     95.34258,   0.00001,     0.244 },
 { {   0,  10, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    241.43379,   0.00007,     0.044 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.42795,   0.00003,     0.044 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    231.64926,   0.00001,     0.108 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     33.93299,   0.00004,     0.248 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   3,   0 },    231.64926,   0.00001,     0.058 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    231.74199,   0.00002,     0.099 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    163.72969,   0.00003,     0.443 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    333.74161,   0.00001,     0.832 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,  -1,  -2 },    229.68561,   0.00001,     0.076 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    155.56117,   0.00007,     0.073 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    126.20240,   0.00002,     0.037 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,   1,  -2 },    229.68561,   0.00001,     0.075 },
 { {   0,  10, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },     38.84074,   0.00003,     0.045 },
 { {   0,  10, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },    218.84074,   0.00003,     0.243 },
 { {   0,  10, -13,  -2,   0,   0,   0,   0,  -2,   0,   0 },    191.16052,   0.00001,     0.044 },
 { {   0,  10, -13,  -2,   0,   0,   0,   0,  -2,   2,   0 },    191.16052,   0.00001,     0.252 },
 { {   0,  10, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },     53.63801,   0.00006,     0.046 },
 { {   0,  10, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    230.81357,   0.00002,     0.202 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    109.17950,   0.00001,     0.122 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    240.27883,   0.00003,     0.100 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     67.56399,   0.00013,     0.307 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    250.34890,   0.00003,     0.454 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     76.42638,   0.00012,     0.068 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    255.97601,   0.00007,     0.036 },
 { {   0,  10, -13,   0,   1,   0,   0,   0,   0,   0,   0 },    353.34516,   0.00001,     0.299 },
 { {   0,  10, -13,   0,   2,   0,   0,   0,   0,   0,   0 },    269.92400,   0.00001,     0.292 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    353.91648,   0.00001,     0.049 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    334.71677,   0.00002,     0.166 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    353.80902,   0.00005,     0.111 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    344.85039,   0.00043,     0.235 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    348.93446,   0.00005,     0.057 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.72245,   0.00010,     0.312 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   1,   0,   1 },    348.75910,   0.00002,     0.033 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    354.09950,   0.00002,     0.403 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    353.91886,   0.00039,     0.064 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,  -2 },    352.78678,   0.00001,     0.466 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    174.42052,   0.00021,     0.034 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   1,   0 },    171.92863,   0.00002,     0.024 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   4,  -3,   0 },    196.00228,   0.00001,     0.072 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   4,  -1,   0 },    172.60333,   0.00001,     0.025 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   4,   0,   0 },    172.68422,   0.00002,     0.019 },
 { {   0,  10, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.56940,   0.00001,     0.049 },
 { {   0,  10, -12,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.56940,   0.00001,     0.164 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   0,   0,   0 },    179.68647,   0.00001,     0.226 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   2,  -2,   0 },    179.79272,   0.00001,     0.377 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   2,  -1,   0 },    359.81792,   0.00003,     0.063 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   2,   0,   0 },    179.80411,   0.00004,     0.034 },
 { {   0,  10, -12,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.18056,   0.00001,     0.051 },
 { {   0,  10, -12,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.18056,   0.00001,     0.141 },
 { {   0,  10, -12,   2,   0,   0,   0,   0,   0,   0,   0 },      0.77823,   0.00001,     0.188 },
 { {   0,  10, -12,   2,   0,   0,   0,   0,   2,  -2,   0 },      0.02948,   0.00001,     0.282 },
 { {   0,  10, -12,   2,   0,   0,   0,   0,   2,   0,   0 },      0.02948,   0.00001,     0.033 },
 { {   0,  10, -11,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.73647,   0.00002,     0.051 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -4,   2,   0 },    271.88964,   0.00001,     0.056 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    273.62919,   0.00005,     0.051 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },     90.92485,   0.00002,     0.142 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    271.16697,   0.00001,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },     63.55688,   0.00003,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    255.50825,   0.00007,     0.159 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   1,  -1 },     75.53061,   0.00001,     0.137 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.88218,   0.00006,     0.125 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    272.00785,   0.00122,     0.190 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    272.00653,   0.00028,     0.054 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    272.23966,   0.00001,     0.031 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,  -1,  -1 },     75.55271,   0.00003,     0.110 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,  -1,   1 },     75.55500,   0.00002,     0.056 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },     75.56544,   0.00032,     0.238 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   0 },     85.91167,   0.00002,     0.057 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   1 },     75.55812,   0.00001,     0.032 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   1,  -1 },     75.55336,   0.00001,     0.057 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    290.07502,   0.00002,     0.102 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    299.73446,   0.00001,     0.287 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.50818,   0.00123,     0.060 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31030,   0.00004,     0.318 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     93.37067,   0.00099,     0.033 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   1,   0 },     92.26654,   0.00013,     0.023 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   2,   0 },     92.37987,   0.00002,     0.018 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   3,   0,   0 },     92.43066,   0.00002,     0.024 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   4,  -2,   0 },    269.16626,   0.00003,     0.035 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   4,  -1,   0 },     99.61007,   0.00001,     0.024 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   4,   0,   0 },     97.77106,   0.00006,     0.018 },
 { {   0,  10, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },    215.61438,   0.00004,     0.052 },
 { {   0,  10, -11,   0,   1,   0,   0,   0,   2,  -1,   0 },    178.64461,   0.00003,     0.059 },
 { {   0,  10, -11,   0,   1,   0,   0,   0,   2,   0,   0 },    358.68846,   0.00004,     0.033 },
 { {   0,  10, -11,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.63806,   0.00002,     0.052 },
 { {   0,  10, -11,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.35874,   0.00001,     0.056 },
 { {   0,  10, -11,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.35874,   0.00001,     0.116 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    182.59762,   0.00006,     0.054 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,  -2,   1,   0 },      7.48510,   0.00001,     0.188 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.59062,   0.00001,     0.126 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,   2,   0,   0 },      8.65802,   0.00002,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.06199,   0.00001,     0.033 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.72516,   0.00002,     0.060 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,  -2,   0 },    359.79946,   0.00001,     0.022 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,  -1,   0 },    177.06724,   0.00002,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.30490,   0.00039,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    181.77674,   0.00028,     0.191 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.56750,   0.00009,     0.125 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },     14.59558,   0.00001,     0.051 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.93785,   0.00008,     0.164 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.21780,   0.00041,     0.143 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,  -2 },    179.94689,   0.00002,     0.049 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    179.93921,   0.00527,     0.160 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    179.93820,   0.00088,     0.051 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    180.89321,   0.00006,     0.031 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,  -1,   0 },    359.99568,   0.00002,     0.186 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.38026,   0.00009,     0.192 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    359.69306,   0.00004,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   1 },    191.37963,   0.00002,     0.031 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    181.13766,   0.00003,     0.114 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -2,  -2 },      0.00601,   0.00001,     0.045 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.44836,   0.00019,     0.223 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.99190,   0.00359,     0.056 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.99286,   0.00009,     0.241 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   0 },      0.07627,   0.00348,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00060,   0.00001,     0.017 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   1,   0 },      0.00500,   0.00029,     0.023 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   2,   0 },    359.98857,   0.00006,     0.017 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.03560,   0.00002,     0.033 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   3,   0,   0 },      0.03633,   0.00003,     0.023 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.71105,   0.00005,     0.034 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.04155,   0.00003,     0.024 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   0,   0 },      0.00317,   0.00006,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   1,   0 },    359.99256,   0.00002,     0.014 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   2,   0 },    359.99110,   0.00001,     0.012 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },    349.18014,   0.00004,     0.054 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.65835,   0.00002,     0.194 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },    346.54674,   0.00002,     0.123 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,   2,  -2,   0 },    346.86735,   0.00001,     0.219 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,   2,   0,   0 },    166.86735,   0.00001,     0.032 },
 { {   0,  10,  -9,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.07415,   0.00002,     0.052 },
 { {   0,  10,  -9,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.07415,   0.00002,     0.135 },
 { {   0,  10,  -9,  -2,   0,   0,   0,   0,   0,   0,   0 },    291.47238,   0.00002,     0.162 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.53094,   0.00001,     0.112 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,   0,  -2,   0 },    180.95163,   0.00002,     0.052 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.35758,   0.00008,     0.139 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,   2,  -2,   0 },      1.27391,   0.00001,     0.185 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.53792,   0.00003,     0.057 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    239.44230,   0.00001,     0.111 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.21229,   0.00002,     0.202 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.20163,   0.00009,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.21903,   0.00013,     0.120 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.22687,   0.00006,     0.189 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },    283.27658,   0.00002,     0.125 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    248.07819,   0.00003,     0.167 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    261.77161,   0.00004,     0.138 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,  -3,   1 },    103.28148,   0.00002,     0.149 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.39882,   0.00006,     0.142 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.28915,   0.00035,     0.161 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    103.27940,   0.00003,     0.030 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.28384,   0.00002,     0.051 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    269.53898,   0.00016,     0.053 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },     92.00775,   0.00023,     0.031 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },     98.30513,   0.00002,     0.022 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },     91.24120,   0.00001,     0.023 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },     85.78970,   0.00002,     0.018 },
 { {   0,  10,  -8,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.19629,   0.00001,     0.055 },
 { {   0,  10,  -8,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.19629,   0.00001,     0.119 },
 { {   0,  10,  -8,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.49794,   0.00001,     0.057 },
 { {   0,  10,  -8,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.49794,   0.00001,     0.112 },
 { {   0,  10,  -8,   0,  -3,   0,   0,   0,  -2,   0,   0 },      8.62943,   0.00002,     0.060 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.27301,   0.00002,     0.060 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.27301,   0.00002,     0.101 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.85724,   0.00002,     0.054 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.44492,   0.00001,     0.124 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },     25.99087,   0.00002,     0.034 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     25.44123,   0.00020,     0.061 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.64331,   0.00012,     0.310 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },     29.62051,   0.00001,     0.100 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    200.29906,   0.00002,     0.243 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.64449,   0.00008,     0.200 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     26.61226,   0.00038,     0.121 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     27.26989,   0.00005,     0.051 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    205.73275,   0.00006,     0.030 },
 { {   0,  10,  -7,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.64641,   0.00003,     0.064 },
 { {   0,  10,  -7,   0,  -3,   0,   0,   0,  -2,   1,   0 },    356.64641,   0.00001,     0.403 },
 { {   0,  10,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },     24.53125,   0.00001,     0.064 },
 { {   0,  10,  -7,   0,  -1,   0,   0,   0,  -2,   1,   0 },     24.52704,   0.00001,     0.433 },
 { {   0,  10,  -7,   0,  -1,   0,   0,   0,  -2,   2,   0 },     24.53125,   0.00001,     0.091 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    275.43550,   0.00002,     0.449 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    258.97351,   0.00003,     0.108 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },     38.59513,   0.00001,     0.134 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     38.59513,   0.00001,     0.029 },
 { {   0,  10,  -6,   0,  -1,   0,   0,   0,   2,   0,   0 },    321.26248,   0.00003,     0.029 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,  -3,   0,   2 },     48.44340,   0.00013,   167.135 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    220.95634,   0.00002,     0.815 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    308.20936,   0.00002,     0.333 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    308.21488,   0.00004,     0.098 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    308.25230,   0.00001,     0.043 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    308.21488,   0.00004,     0.027 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    128.19741,   0.00002,     0.209 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    128.08386,   0.00001,     0.118 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.19741,   0.00002,     0.046 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    129.12814,   0.00003,     0.029 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },    128.64926,   0.00002,     0.016 },
 { {   0,  10,  -6,   0,   1,   0,   0,   0,   2,  -2,   0 },    119.47180,   0.00001,     0.117 },
 { {   0,  10,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },    119.47180,   0.00001,     0.029 },
 { {   0,  10,  -5,   0,  -3,   0,   0,   0,  -2,   1,   0 },    199.94052,   0.00001,     2.086 },
 { {   0,  10,  -5,   0,  -2,   0,   0,   0,  -2,   1,   0 },    109.90642,   0.00001,     2.531 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    148.48031,   0.00002,    16.737 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    130.88979,   0.00008,     4.415 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.17937,   0.00002,     0.044 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     35.57982,   0.00001,     0.028 },
 { {   0,  10,  -4,  -2,   0,   0,   0,   0,  -2,   1,   0 },    128.10035,   0.00001,     3.450 },
 { {   0,  10,  -4,   0,  -9,   0,   0,   0,  -2,   1,   0 },     30.23056,   0.00001,    67.776 },
 { {   0,  10,  -4,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.31190,   0.00002,     0.097 },
 { {   0,  10,  -4,   0,  -2,   0,   0,   0,   2,   0,   0 },    308.31190,   0.00002,     0.027 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,  -2,   0 },    256.24244,   0.00001,     0.026 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    256.24244,   0.00007,     0.039 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.24244,   0.00004,     0.080 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },     76.24244,   0.00002,     0.075 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    234.20775,   0.00001,    18.782 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    244.74659,   0.00001,     8.810 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,  -3,   0 },    256.24244,   0.00002,     0.036 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    256.24244,   0.00015,     0.070 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     76.24244,   0.00020,     0.082 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     76.24244,   0.00003,     0.039 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     76.24244,   0.00028,     0.095 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     76.24244,   0.00045,     0.042 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     76.24244,   0.00006,     0.027 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },     76.24244,   0.00002,     0.028 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },     76.24244,   0.00001,     0.021 },
 { {   0,  10,  -3,   0,  -2,   0,   0,   0,   0,  -1,   0 },    209.51355,   0.00002,     5.913 },
 { {   0,  10,  -3,   0,  -1,   0,   0,   0,   0,  -1,   0 },    331.11891,   0.00002,    11.789 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,  -2,   0 },    333.30551,   0.00001,     0.016 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,  -1,   0 },    333.30551,   0.00008,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },    333.30551,   0.00016,     0.028 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },    333.30551,   0.00008,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -3,   0 },    333.30551,   0.00008,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },    333.30551,   0.00079,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    153.30551,   0.00003,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    333.30551,   0.00647,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,  -2 },    153.30551,   0.00004,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    333.58030,   0.00623,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    153.30551,   0.00002,     0.067 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -1,  -2,   0 },    153.30551,   0.00002,     0.039 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    153.30551,   0.00017,     0.081 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -5,   0 },    333.30551,   0.00001,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -4,   0 },    333.30551,   0.00015,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,  -2 },    153.30551,   0.00002,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },    333.07007,   0.00211,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,  -2 },    153.30551,   0.00018,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    333.10628,   0.03091,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   2 },    153.30551,   0.00005,     0.074 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,  -2 },    153.21025,   0.00111,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    333.30551,   0.56341,  1911.695 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    153.21025,   0.00111,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    153.30551,   0.00005,     0.074 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    333.03408,   0.03111,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   2 },    153.30551,   0.00018,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    333.07007,   0.00211,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   2 },    153.30551,   0.00002,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    333.30551,   0.00015,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   3,   0 },    333.30551,   0.00001,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    153.30551,   0.00017,     0.081 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    153.30551,   0.00002,     0.039 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -4,   0 },    153.30551,   0.00002,     0.067 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    333.50460,   0.00627,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   2 },    153.30551,   0.00004,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    333.30551,   0.00647,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   2 },    153.30551,   0.00003,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    333.30551,   0.00079,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    333.30551,   0.00008,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -3,   0 },    333.30551,   0.00008,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },    333.30551,   0.00016,     0.028 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    333.30551,   0.00008,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    333.30551,   0.00001,     0.016 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -4,  -1,   0 },    230.36858,   0.00001,     0.021 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },    230.36858,   0.00002,     0.028 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    230.36858,   0.00006,     0.027 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    230.36858,   0.00045,     0.042 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    230.36858,   0.00028,     0.095 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },    230.36858,   0.00003,     0.039 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    230.36858,   0.00020,     0.082 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     50.36858,   0.00015,     0.070 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     50.36858,   0.00002,     0.036 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },    230.36858,   0.00002,     0.075 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },     50.36858,   0.00004,     0.080 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     50.36858,   0.00007,     0.039 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     50.36858,   0.00001,     0.026 },
 { {   0,  10,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    127.43165,   0.00002,     0.044 },
 { {   0,  10,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    127.43165,   0.00001,     0.105 },
 { {   0,  10,   0,  -9,   0,   0,   0,   0,  -2,   1,   0 },    106.82924,   0.00001,    85.736 },
 { {   0,  10,   8, -24,   0,   0,   0,   0,  -2,   1,   0 },    220.31490,   0.00001,    76.498 },
 { {   0,  11, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.61086,   0.00002,     0.037 },
 { {   0,  11, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.62485,   0.00003,     0.074 },
 { {   0,  11, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.61228,   0.00001,     2.897 },
 { {   0,  11, -20,   0,   0,   0,   0,   0,  -2,   3,   0 },    289.62851,   0.00002,     0.077 },
 { {   0,  11, -20,   0,   4,   0,   0,   0,  -2,   2,   0 },    131.52075,   0.00003,   124.848 },
 { {   0,  11, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     30.14735,   0.00003,     0.079 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },     88.22700,   0.00001,     0.075 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    308.47135,   0.00002,     8.386 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,   1,   0 },     88.22700,   0.00001,     0.076 },
 { {   0,  11, -18,   0,   1,   0,   0,   0,   0,  -1,   0 },    222.58294,   0.00002,     0.075 },
 { {   0,  11, -18,   0,   1,   0,   0,   0,   0,   0,   0 },    312.53566,   0.00001,    28.621 },
 { {   0,  11, -18,   0,   1,   0,   0,   0,   0,   1,   0 },    222.58294,   0.00002,     0.076 },
 { {   0,  11, -18,   0,   2,   0,   0,   0,   0,  -1,   0 },    323.12340,   0.00002,     0.076 },
 { {   0,  11, -18,   0,   2,   0,   0,   0,   0,   1,   0 },    323.12340,   0.00002,     0.075 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    225.19576,   0.00001,     0.027 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    225.19576,   0.00001,     0.094 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    247.01744,   0.00002,     1.135 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,   1,  -1,   0 },    322.60146,   0.00003,   159.914 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },     45.19576,   0.00003,     0.081 },
 { {   0,  11, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.84715,   0.00003,     0.042 },
 { {   0,  11, -17,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.54960,   0.00005,     0.042 },
 { {   0,  11, -17,   0,   2,   0,   0,   0,  -2,   2,   0 },     51.55062,   0.00004,     0.354 },
 { {   0,  11, -17,   2,   0,   0,   0,   0,  -2,   0,   0 },     51.75857,   0.00002,     0.044 },
 { {   0,  11, -17,   2,   0,   0,   0,   0,  -2,   2,   0 },     51.75857,   0.00002,     0.269 },
 { {   0,  11, -16,   0,  -1,   0,   0,   0,   2,  -2,   0 },    331.86056,   0.00001,    44.577 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.06401,   0.00002,     0.028 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    315.78558,   0.00002,     0.044 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    231.02722,   0.00003,     0.104 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.11812,   0.00003,     0.274 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   3,   0 },    110.30296,   0.00002,     0.059 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    122.28390,   0.00002,     0.532 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },    237.86456,   0.00001,     9.369 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    235.88774,   0.00003,     0.075 },
 { {   0,  11, -16,   0,   1,   0,   0,   0,  -2,   1,   0 },     50.37370,   0.00001,     0.105 },
 { {   0,  11, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },    229.94597,   0.00002,     0.267 },
 { {   0,  11, -16,   0,   1,   0,   0,   0,  -2,   3,   0 },    230.37370,   0.00001,     0.059 },
 { {   0,  11, -16,   0,   2,   0,   0,   0,  -2,   0,   0 },    111.86104,   0.00001,     0.044 },
 { {   0,  11, -16,   0,   2,   0,   0,   0,  -2,   2,   0 },    291.86104,   0.00001,     0.262 },
 { {   0,  11, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    241.61175,   0.00003,     0.046 },
 { {   0,  11, -15,   0,  -1,   0,   0,   0,  -2,   2,   0 },     61.61175,   0.00003,     0.219 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.76952,   0.00003,     0.046 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     56.41650,   0.00001,     0.116 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     28.59571,   0.00002,     0.215 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    165.58676,   0.00004,     0.347 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    158.44545,   0.00001,     0.036 },
 { {   0,  11, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.45174,   0.00003,     0.048 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.13919,   0.00002,     0.048 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    255.72015,   0.00002,     0.107 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     67.06993,   0.00011,     0.258 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    250.92958,   0.00002,     0.354 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     75.39829,   0.00008,     0.065 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    256.70141,   0.00006,     0.035 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   4,  -2,  -2 },    259.25820,   0.00048,  1260.592 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   4,  -1,  -2 },    100.46620,   0.00002,     0.075 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    356.98137,   0.00003,     0.050 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    339.54095,   0.00003,     0.150 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -1,   0,   1 },     32.10246,   0.00002,     0.169 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },     23.66391,   0.00001,     0.046 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    342.28773,   0.00003,     0.119 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    347.65557,   0.00033,     0.205 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    348.37408,   0.00003,     0.055 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.69263,   0.00007,     0.261 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   1,   0,   1 },    348.68513,   0.00002,     0.033 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    356.76690,   0.00002,     0.322 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    353.37056,   0.00026,     0.061 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    175.57283,   0.00018,     0.034 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,   1,   0 },    173.01905,   0.00002,     0.023 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   4,  -1,   0 },    173.37300,   0.00001,     0.024 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   4,   0,   0 },    173.44622,   0.00002,     0.018 },
 { {   0,  11, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.61944,   0.00001,     0.051 },
 { {   0,  11, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.61944,   0.00001,     0.148 },
 { {   0,  11, -13,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.16814,   0.00001,     0.053 },
 { {   0,  11, -13,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.16814,   0.00001,     0.130 },
 { {   0,  11, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    124.50823,   0.00002,     0.053 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },    272.80414,   0.00001,     0.059 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    273.97431,   0.00004,     0.053 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     98.82512,   0.00003,     0.179 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },     93.49970,   0.00003,     0.131 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },     75.26809,   0.00003,     0.050 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,   1 },    255.24821,   0.00004,     0.144 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },     94.34013,   0.00002,     0.150 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },     88.29131,   0.00001,     0.048 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    282.10592,   0.00007,     0.136 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    272.03339,   0.00094,     0.170 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    275.63089,   0.00022,     0.052 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    277.27465,   0.00001,     0.031 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,  -1,  -1 },     75.29609,   0.00003,     0.119 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,  -1,   1 },     75.29861,   0.00001,     0.054 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },     75.29674,   0.00025,     0.207 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     83.89094,   0.00002,     0.055 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,   1 },     75.30087,   0.00001,     0.032 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   1,  -1 },     75.29695,   0.00001,     0.055 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    146.32001,   0.00002,     0.244 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.69940,   0.00087,     0.058 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31403,   0.00003,     0.265 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,   0 },     94.30713,   0.00083,     0.033 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   1,   0 },     92.70842,   0.00013,     0.023 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   3,   0,   0 },     92.42917,   0.00002,     0.023 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   4,  -2,   0 },    280.24971,   0.00003,     0.035 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   4,   0,   0 },     92.71122,   0.00002,     0.018 },
 { {   0,  11, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },    220.61314,   0.00003,     0.053 },
 { {   0,  11, -12,   0,   1,   0,   0,   0,   2,   0,   0 },    358.67481,   0.00003,     0.033 },
 { {   0,  11, -12,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.62564,   0.00001,     0.054 },
 { {   0,  11, -12,   0,   2,   0,   0,   0,  -2,   2,   0 },     52.62564,   0.00001,     0.128 },
 { {   0,  11, -12,   2,   0,   0,   0,   0,   0,   0,   0 },     39.90972,   0.00001,     0.144 },
 { {   0,  11, -12,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.34012,   0.00001,     0.058 },
 { {   0,  11, -12,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.34012,   0.00001,     0.108 },
 { {   0,  11, -11,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.05004,   0.00006,     0.056 },
 { {   0,  11, -11,   0,  -1,   0,   0,   0,  -2,   1,   0 },      7.70733,   0.00002,     0.213 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.02157,   0.00001,     0.034 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.55209,   0.00002,     0.062 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.28376,   0.00003,     0.032 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.87990,   0.00037,     0.056 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.07998,   0.00027,     0.217 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.34703,   0.00008,     0.116 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },     11.10314,   0.00002,     0.046 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },      0.37740,   0.00008,     0.182 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.97563,   0.00037,     0.157 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   0,  -2 },    179.93212,   0.00001,     0.050 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    179.92918,   0.00374,     0.145 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    179.97767,   0.00061,     0.050 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   2,  -2 },    359.93212,   0.00001,     0.153 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    179.97065,   0.00007,     0.030 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,  -2,   1 },     11.54399,   0.00002,     0.162 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,  -1,   0 },      2.16705,   0.00001,     0.167 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.54427,   0.00007,     0.172 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    358.50588,   0.00002,     0.052 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   2,   0 },      0.02478,   0.00001,     0.022 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -5,   0 },    179.97336,   0.00001,     0.029 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    181.16462,   0.00002,     0.123 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    181.06455,   0.00011,     0.196 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.97470,   0.00242,     0.054 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.97603,   0.00005,     0.210 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   0 },      0.10224,   0.00268,     0.032 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00538,   0.00001,     0.017 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   1,   0 },      0.01568,   0.00025,     0.022 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   2,   0 },    359.98932,   0.00005,     0.017 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.02246,   0.00001,     0.033 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   3,   0,   0 },      0.02479,   0.00004,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   3,   2,   0 },    180.02475,   0.00001,     0.014 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.38982,   0.00004,     0.034 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.09562,   0.00003,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,   0,   0 },      0.02542,   0.00005,     0.018 },
 { {   0,  11, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },    349.11875,   0.00004,     0.056 },
 { {   0,  11, -11,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.63197,   0.00002,     0.221 },
 { {   0,  11, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },    346.63312,   0.00002,     0.114 },
 { {   0,  11, -10,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.05965,   0.00001,     0.054 },
 { {   0,  11, -10,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.05965,   0.00001,     0.124 },
 { {   0,  11, -10,  -3,   0,   0,   0,   0,   0,   0,   0 },    303.91225,   0.00001,     0.159 },
 { {   0,  11, -10,   0,  -2,   0,   0,   0,  -2,   1,   0 },    119.06301,   0.00002,     0.265 },
 { {   0,  11, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.49464,   0.00001,     0.059 },
 { {   0,  11, -10,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.49464,   0.00001,     0.104 },
 { {   0,  11, -10,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.39473,   0.00006,     0.128 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    255.32161,   0.00003,     0.059 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    241.29776,   0.00001,     0.104 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.84186,   0.00001,     0.231 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.15773,   0.00006,     0.056 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.19783,   0.00010,     0.112 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.20678,   0.00006,     0.215 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    239.21673,   0.00002,     0.186 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    282.10196,   0.00003,     0.127 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.59759,   0.00005,     0.156 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,  -1,   1 },    103.26488,   0.00002,     0.049 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.27351,   0.00028,     0.146 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,   1 },    103.29518,   0.00002,     0.030 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.26736,   0.00001,     0.050 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    268.74023,   0.00015,     0.052 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   0,   0 },     92.72375,   0.00019,     0.031 },
 { {   0,  11,  -9,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.18594,   0.00001,     0.057 },
 { {   0,  11,  -9,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.18594,   0.00001,     0.110 },
 { {   0,  11,  -9,  -2,   0,   0,   0,   0,  -2,   0,   0 },    359.98607,   0.00004,     0.059 },
 { {   0,  11,  -9,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.47567,   0.00001,     0.104 },
 { {   0,  11,  -9,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.27871,   0.00001,     0.062 },
 { {   0,  11,  -9,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.27871,   0.00001,     0.096 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.26106,   0.00002,     0.062 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.26106,   0.00002,     0.095 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,   0,  -2,   0 },    180.18845,   0.00001,     0.056 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.72984,   0.00003,     0.115 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.15916,   0.00002,     0.034 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     24.73498,   0.00018,     0.063 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.53361,   0.00014,     0.384 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    191.39894,   0.00001,     0.094 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    196.96630,   0.00002,     0.287 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.69884,   0.00008,     0.229 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     28.71409,   0.00030,     0.113 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.78685,   0.00004,     0.049 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    205.70632,   0.00005,     0.030 },
 { {   0,  11,  -8,   0,  -3,   0,   0,   0,  -2,   0,   0 },    356.63192,   0.00001,     0.066 },
 { {   0,  11,  -8,   0,  -3,   0,   0,   0,  -2,   2,   0 },    356.63192,   0.00001,     0.088 },
 { {   0,  11,  -8,   0,  -2,   0,   0,   0,  -4,   3,   0 },    260.63491,   0.00004,   225.944 },
 { {   0,  11,  -8,   0,  -2,   0,   0,   0,  -2,   1,   0 },     86.71023,   0.00002,     0.565 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    278.55908,   0.00002,     0.625 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    223.30152,   0.00002,     0.029 },
 { {   0,  11,  -8,   0,   1,   0,   0,   0,   2,   0,   0 },    127.57039,   0.00002,     0.029 },
 { {   0,  11,  -8,   3,   0,   0,   0,   0,  -2,   0,   0 },    184.47740,   0.00001,     0.075 },
 { {   0,  11,  -8,   3,   0,   0,   0,   0,  -2,   1,   0 },     15.08901,   0.00002,   168.641 },
 { {   0,  11,  -8,   3,   0,   0,   0,   0,  -2,   2,   0 },    184.47740,   0.00001,     0.075 },
 { {   0,  11,  -7,   0,  -1,   0,   0,   0,   2,  -2,   0 },    141.24799,   0.00001,     0.111 },
 { {   0,  11,  -7,   0,  -1,   0,   0,   0,   2,   0,   0 },    321.24799,   0.00001,     0.028 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    228.41176,   0.00002,     0.072 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    227.65813,   0.00008,     1.664 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    309.55562,   0.00001,     0.092 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    308.19993,   0.00003,     0.027 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    128.62174,   0.00003,     0.241 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.62174,   0.00003,     0.045 },
 { {   0,  11,  -7,   0,   1,   0,   0,   0,   2,   0,   0 },    119.45731,   0.00002,     0.028 },
 { {   0,  11,  -6,   0,  -5,   0,   0,   0,  -2,   1,   0 },    260.35373,   0.00005,    44.436 },
 { {   0,  11,  -6,   0,  -4,   0,   0,   0,  -2,   1,   0 },    223.69449,   0.00002,    16.182 },
 { {   0,  11,  -6,   0,  -3,   0,   0,   0,  -2,   1,   0 },     39.27588,   0.00001,     6.845 },
 { {   0,  11,  -6,   0,  -2,   0,   0,   0,  -2,   1,   0 },     97.73317,   0.00002,     4.340 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.90834,   0.00003,     0.078 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    128.88696,   0.00002,     2.506 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    121.89966,   0.00002,     0.073 },
 { {   0,  11,  -5,   0,  -4,   0,   0,   0,  -1,   1,  -1 },    167.95366,   0.00005,   124.469 },
 { {   0,  11,  -5,   0,  -2,   0,   0,   0,   0,  -1,   0 },    308.35232,   0.00001,     1.840 },
 { {   0,  11,  -3,  -3,   0,   0,   0,   0,   0,  -1,   0 },     58.41366,   0.00003,    33.392 },
 { {   0,  11,   6, -20,   0,   0,   0,   0,   0,  -1,   0 },    321.98630,   0.00001,   115.147 },
 { {   0,  12, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.60503,   0.00002,     0.038 },
 { {   0,  12, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    109.62018,   0.00002,     0.077 },
 { {   0,  12, -21,   0,   0,   0,   0,   0,  -2,   3,   0 },    109.62018,   0.00002,     0.074 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    328.82095,   0.00001,     0.078 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,   0,   1,   0 },    328.82095,   0.00001,     0.073 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,  -1,   0 },    225.87272,   0.00001,     0.027 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    225.87272,   0.00001,     0.100 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    247.01742,   0.00002,     0.085 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    247.01906,   0.00002,     0.664 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    247.01742,   0.00002,     0.068 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   2,  -3,   0 },     28.06088,   0.00002,     0.074 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },     28.06088,   0.00002,     0.077 },
 { {   0,  12, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.87923,   0.00002,     0.043 },
 { {   0,  12, -18,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.52893,   0.00007,     0.043 },
 { {   0,  12, -18,   0,   3,   0,   0,   0,   2,  -2,   0 },    221.53620,   0.00003,    67.451 },
 { {   0,  12, -18,   0,   3,   0,   0,   0,   2,  -1,   0 },     41.77508,   0.00003,     0.076 },
 { {   0,  12, -18,   2,   0,   0,   0,   0,  -2,   0,   0 },     51.73791,   0.00002,     0.045 },
 { {   0,  12, -18,   2,   0,   0,   0,   0,  -2,   2,   0 },     51.73791,   0.00002,     0.230 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.06565,   0.00002,     0.028 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    207.34952,   0.00001,     0.111 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    314.64357,   0.00002,     0.234 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -3,   0 },     57.87334,   0.00003,     0.080 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    237.85973,   0.00001,     1.366 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },     57.87334,   0.00003,     0.071 },
 { {   0,  12, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },    229.92734,   0.00002,     0.045 },
 { {   0,  12, -17,   0,   1,   0,   0,   0,  -2,   2,   0 },    229.92734,   0.00002,     0.229 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -4,  -1,   0 },    231.70578,   0.00002,     0.017 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },    231.70578,   0.00002,     0.031 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,  -2,   2 },    231.70578,   0.00001,     0.048 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     61.83720,   0.00004,     0.047 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.70578,   0.00001,     0.178 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },     40.18465,   0.00004,     0.189 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    165.60834,   0.00004,     0.285 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    159.36817,   0.00001,     0.035 },
 { {   0,  12, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.41571,   0.00003,     0.049 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    110.01516,   0.00002,     0.049 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    173.96182,   0.00001,     0.143 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   3,   0 },    173.96182,   0.00001,     0.051 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     72.39297,   0.00008,     0.222 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    251.44617,   0.00002,     0.290 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },    255.98847,   0.00002,     0.095 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },     76.27763,   0.00003,     0.063 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    257.22277,   0.00006,     0.034 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -4,   0,   0 },    212.87298,   0.00002,     0.023 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    335.56589,   0.00002,     0.052 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    345.53377,   0.00002,     0.167 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    344.16886,   0.00001,     0.137 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    346.74969,   0.00025,     0.182 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.60920,   0.00005,     0.225 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   1,   0,   1 },    348.60290,   0.00001,     0.032 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    353.75822,   0.00001,     0.268 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    353.98414,   0.00020,     0.059 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    176.29242,   0.00016,     0.033 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    173.57165,   0.00001,     0.023 },
 { {   0,  12, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.66481,   0.00001,     0.052 },
 { {   0,  12, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.66481,   0.00001,     0.136 },
 { {   0,  12, -14,   0,   2,   0,   0,   0,   2,  -2,   0 },    179.77864,   0.00001,     0.256 },
 { {   0,  12, -14,   0,   2,   0,   0,   0,   2,   0,   0 },    179.77864,   0.00001,     0.033 },
 { {   0,  12, -14,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.15365,   0.00001,     0.055 },
 { {   0,  12, -14,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.15365,   0.00001,     0.120 },
 { {   0,  12, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.77736,   0.00002,     0.055 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    272.25167,   0.00001,     0.061 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    273.04276,   0.00003,     0.055 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },     92.07891,   0.00002,     0.121 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },     74.97625,   0.00003,     0.052 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },     74.97840,   0.00001,     0.166 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -3,   0 },     91.90220,   0.00001,     0.030 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.69216,   0.00002,     0.050 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    280.43281,   0.00007,     0.148 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    271.97336,   0.00077,     0.154 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    276.07429,   0.00016,     0.051 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,  -1,  -1 },     75.00785,   0.00001,     0.128 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },     75.00912,   0.00019,     0.183 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,   0 },     92.62393,   0.00001,     0.053 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,   1 },     75.01291,   0.00001,     0.031 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    279.14210,   0.00001,     0.211 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    273.02424,   0.00061,     0.056 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31573,   0.00002,     0.227 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,   0 },     93.48833,   0.00063,     0.032 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   1,   0 },     92.50519,   0.00011,     0.022 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   3,   0,   0 },     92.42513,   0.00002,     0.023 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   4,  -2,   0 },    271.71577,   0.00002,     0.034 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   4,   0,   0 },     92.43105,   0.00002,     0.018 },
 { {   0,  12, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    218.85521,   0.00004,     0.055 },
 { {   0,  12, -13,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.61115,   0.00001,     0.055 },
 { {   0,  12, -13,   0,   2,   0,   0,   0,  -2,   2,   0 },     52.61115,   0.00001,     0.118 },
 { {   0,  12, -13,   2,   0,   0,   0,   0,   0,   0,   0 },     39.89316,   0.00001,     0.132 },
 { {   0,  12, -13,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.31946,   0.00001,     0.060 },
 { {   0,  12, -13,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.31946,   0.00001,     0.101 },
 { {   0,  12, -12,   0,  -2,   0,   0,   0,  -2,   0,   0 },    207.22977,   0.00001,     0.057 },
 { {   0,  12, -12,   0,  -2,   0,   0,   0,  -2,   2,   0 },    207.22977,   0.00001,     0.110 },
 { {   0,  12, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    183.41748,   0.00004,     0.058 },
 { {   0,  12, -12,   0,  -1,   0,   0,   0,  -2,   1,   0 },      7.94129,   0.00002,     0.246 },
 { {   0,  12, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },    162.93244,   0.00001,     0.109 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.57791,   0.00002,     0.065 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,  -2,   0 },    359.93825,   0.00004,     0.023 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.19305,   0.00004,     0.033 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.30221,   0.00034,     0.058 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.48761,   0.00021,     0.252 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.20571,   0.00008,     0.108 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    176.45090,   0.00002,     0.044 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.93673,   0.00005,     0.206 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.92050,   0.00034,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    179.88323,   0.00270,     0.133 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    180.01314,   0.00048,     0.048 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    179.99298,   0.00003,     0.029 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.68767,   0.00005,     0.155 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   0,   1 },     11.68761,   0.00002,     0.030 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   2,   0 },      0.00921,   0.00001,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -3,   0 },    180.03438,   0.00002,     0.133 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.90222,   0.00009,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.95351,   0.00162,     0.053 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.95605,   0.00005,     0.185 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    359.96299,   0.00207,     0.031 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   2 },    359.94923,   0.00001,     0.017 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   1,   0 },    359.95828,   0.00020,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   2,   0 },    359.95906,   0.00003,     0.017 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.00688,   0.00001,     0.032 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   3,   0,   0 },      0.00921,   0.00004,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.12713,   0.00002,     0.033 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.95189,   0.00003,     0.023 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,   0,   0 },    359.95914,   0.00004,     0.018 },
 { {   0,  12, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },    347.64726,   0.00003,     0.058 },
 { {   0,  12, -12,   0,   1,   0,   0,   0,  -2,   2,   0 },    347.64726,   0.00003,     0.107 },
 { {   0,  12, -11,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.04309,   0.00001,     0.056 },
 { {   0,  12, -11,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.04309,   0.00001,     0.115 },
 { {   0,  12, -11,   0,  -2,   0,   0,   0,   0,   0,   0 },    114.91604,   0.00002,     0.120 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.45786,   0.00001,     0.061 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.45786,   0.00001,     0.098 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,   0,  -2,   0 },      1.20547,   0.00002,     0.055 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.25665,   0.00007,     0.119 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    267.45492,   0.00004,     0.062 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     84.54896,   0.00001,     0.336 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.82247,   0.00001,     0.271 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.13134,   0.00006,     0.058 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    284.04913,   0.00007,     0.105 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.18455,   0.00005,     0.248 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    346.20825,   0.00002,     0.118 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    114.14213,   0.00001,     0.046 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,   2,   0 },     81.90541,   0.00002,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.23076,   0.00004,     0.172 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,  -1,   1 },    103.24596,   0.00001,     0.048 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.21414,   0.00026,     0.134 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,   1 },    103.24543,   0.00002,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.30059,   0.00010,     0.050 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     91.71430,   0.00017,     0.030 },
 { {   0,  12, -10,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.17352,   0.00001,     0.059 },
 { {   0,  12, -10,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.17352,   0.00001,     0.103 },
 { {   0,  12, -10,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.45296,   0.00002,     0.061 },
 { {   0,  12, -10,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.45296,   0.00002,     0.098 },
 { {   0,  12, -10,   0,  -3,   0,   0,   0,  -2,   0,   0 },      7.18317,   0.00001,     0.065 },
 { {   0,  12, -10,   0,  -3,   0,   0,   0,  -2,   2,   0 },      7.18317,   0.00001,     0.091 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.24726,   0.00003,     0.065 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,  -2,   1,   0 },      0.24709,   0.00001,     0.466 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.24761,   0.00001,     0.090 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,   0,  -2,   0 },    180.17187,   0.00001,     0.058 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.70779,   0.00003,     0.107 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.20180,   0.00002,     0.035 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     25.53652,   0.00019,     0.066 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.98889,   0.00017,     0.506 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    199.19285,   0.00002,     0.349 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.62375,   0.00008,     0.267 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     32.69816,   0.00026,     0.105 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    198.77108,   0.00001,     0.044 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.76807,   0.00003,     0.048 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    214.19935,   0.00005,     0.029 },
 { {   0,  12,  -9,  -4,   0,   0,   0,   0,  -2,   1,   0 },    150.32456,   0.00001,     0.322 },
 { {   0,  12,  -9,   0,  -3,   0,   0,   0,  -3,   0,   2 },    355.64099,   0.00003,   127.390 },
 { {   0,  12,  -9,   0,  -2,   0,   0,   0,  -2,   0,   0 },     86.02425,   0.00002,     0.069 },
 { {   0,  12,  -9,   0,  -2,   0,   0,   0,  -2,   1,   0 },    252.44417,   0.00001,     0.874 },
 { {   0,  12,  -9,   0,  -2,   0,   0,   0,  -2,   2,   0 },    266.02425,   0.00002,     0.083 },
 { {   0,  12,  -9,   0,  -1,   0,   0,   0,  -3,   1,   1 },    193.65282,   0.00004,   166.720 },
 { {   0,  12,  -9,   0,   0,  -2,   0,   0,  -3,   1,   1 },    104.88515,   0.00002,    96.065 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,  -3,   1,   1 },    130.95117,   0.00006,    12.771 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    273.67422,   0.00003,     1.025 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    310.63301,   0.00002,     0.081 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    331.44532,   0.00002,     0.095 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    331.44532,   0.00003,     0.042 },
 { {   0,  12,  -9,   0,   1,   0,   0,   0,  -3,   2,   0 },    304.53833,   0.00002,   241.182 },
 { {   0,  12,  -8,   0,  -3,   0,   0,   0,  -2,   1,   0 },    235.41185,   0.00007,     4.379 },
 { {   0,  12,  -8,   0,  -2,   0,   0,   0,  -2,   0,   0 },    203.31811,   0.00001,     0.075 },
 { {   0,  12,  -8,   0,  -2,   0,   0,   0,  -2,   1,   0 },    157.24734,   0.00007,     6.941 },
 { {   0,  12,  -8,   0,  -2,   0,   0,   0,  -2,   2,   0 },    203.31811,   0.00001,     0.076 },
 { {   0,  12,  -8,   0,  -1,   0,   0,   0,  -2,   1,   0 },    283.55098,   0.00004,    16.733 },
 { {   0,  12,  -8,   0,   0,  -3,   0,   0,  -2,   1,   0 },    269.58359,   0.00001,    12.936 },
 { {   0,  12,  -8,   0,   0,  -2,   0,   0,  -2,   1,   0 },    190.06057,   0.00005,    23.065 },
 { {   0,  12,  -8,   0,   0,  -1,   0,   0,  -2,   1,   0 },    196.79564,   0.00005,   106.299 },
 { {   0,  12,  -8,   0,   0,   0,  -3,   0,  -2,   1,   0 },     18.39190,   0.00001,    89.563 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   0,   0 },    272.38258,   0.00003,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   1,   0 },    272.38258,   0.00048,  1357.743 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   2,   0 },    272.38258,   0.00003,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,  -1,   0,  -2,   1,   0 },    291.49674,   0.00004,    79.124 },
 { {   0,  12,  -8,   0,   0,   0,   0,  -2,  -2,   1,   0 },    205.16489,   0.00003,    80.631 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    228.50839,   0.00005,     0.026 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    228.48690,   0.00039,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    228.38865,   0.00037,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -3,   1,   0 },     48.50839,   0.00001,     0.081 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    228.15488,   0.00008,     0.038 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,  -2 },     48.50839,   0.00001,     0.025 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    228.15465,   0.00119,     0.076 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,  -2 },     48.50839,   0.00007,     0.037 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    228.50839,   0.03489,    40.749 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   2 },     48.50839,   0.00007,     0.037 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    228.61917,   0.00199,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   2 },     48.50839,   0.00001,     0.025 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    228.61908,   0.00014,     0.038 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -1,   1,   0 },     48.50839,   0.00001,     0.081 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    128.30583,   0.00001,     0.067 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    227.34513,   0.00053,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    228.52878,   0.00041,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    228.50839,   0.00005,     0.026 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    128.19132,   0.00003,     0.284 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.19132,   0.00003,     0.044 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    264.18591,   0.00001,     0.028 },
 { {   0,  12,  -8,   0,   0,   1,   0,   0,  -2,   1,   0 },    204.16174,   0.00002,    17.098 },
 { {   0,  12,  -8,   0,   1,   0,   0,   0,  -2,   1,   0 },     56.73550,   0.00003,     9.188 },
 { {   0,  12,  -7,  -2,   0,   0,   0,   0,  -2,   0,   0 },    288.13049,   0.00003,     0.075 },
 { {   0,  12,  -7,  -2,   0,   0,   0,   0,  -2,   1,   0 },    267.43157,   0.00005,    25.755 },
 { {   0,  12,  -7,  -2,   0,   0,   0,   0,  -2,   2,   0 },    288.13049,   0.00003,     0.076 },
 { {   0,  12,  -7,   0,  -2,   0,   0,   0,  -1,   0,   0 },     84.52563,   0.00002,    32.179 },
 { {   0,  12,  -7,   0,  -1,   0,   0,   0,   2,  -2,   0 },    129.18502,   0.00002,     0.094 },
 { {   0,  12,  -7,   0,  -1,   0,   0,   0,   2,   0,   0 },    309.18502,   0.00002,     0.027 },
 { {   0,  12,  -7,   0,   0,  -4,   0,   0,  -1,   0,   0 },    116.18114,   0.00002,   574.972 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -4,   1,   0 },    125.57146,   0.00003,     0.042 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },    125.57146,   0.00002,     0.096 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    125.57146,   0.00001,     0.082 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    149.22796,   0.00004,     0.075 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    316.29266,   0.00003,     7.271 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    149.22796,   0.00082,    34.218 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   2,  -1 },    149.22796,   0.00004,     0.076 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    144.72505,   0.00001,     0.071 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    223.97528,   0.00001,     1.334 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    324.72505,   0.00001,     0.080 },
 { {   0,  12,  -6,  -4,   0,   0,   0,   0,  -2,   1,   0 },    250.62107,   0.00001,     9.785 },
 { {   0,  12,  -6,  -2,   0,   0,   0,   0,  -1,   0,   0 },    308.06929,   0.00001,    13.484 },
 { {   0,  12,  -6,   0,  -3,   0,   0,   0,   0,  -2,   0 },     34.90652,   0.00002,     0.075 },
 { {   0,  12,  -6,   0,  -3,   0,   0,   0,   0,  -1,   0 },    212.06446,   0.00006,   418.437 },
 { {   0,  12,  -6,   0,  -3,   0,   0,   0,   0,   0,   0 },    167.32033,   0.00003,     0.075 },
 { {   0,  12,  -6,   0,  -2,   0,   0,   0,   0,  -1,   0 },     45.38939,   0.00002,    12.208 },
 { {   0,  12,  -6,   0,   1,   0,   0,   0,   0,   0,  -2 },    278.68671,   0.00001,     0.075 },
 { {   0,  12,  -6,   0,   1,   0,   0,   0,   0,   1,  -2 },    278.68671,   0.00027,   761.128 },
 { {   0,  12,  -6,   0,   1,   0,   0,   0,   0,   2,  -2 },    278.68671,   0.00001,     0.075 },
 { {   0,  12,  -3,  -6,   0,   0,   0,   0,   0,  -1,   0 },    282.81066,   0.00001,    16.552 },
 { {   0,  12,  -2,  -8,   0,   0,   0,   0,   0,  -1,   0 },    227.44503,   0.00008,   338.916 },
 { {   0,  12,   1, -17,   0,   0,   0,   0,  -2,   1,   0 },     23.94773,   0.00002,    70.966 },
 { {   0,  13, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },     24.22404,   0.00004,     0.075 },
 { {   0,  13, -23,   0,   0,   0,   0,   0,  -2,   3,   0 },     14.55687,   0.00002,     0.076 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.59071,   0.00002,     0.039 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.61015,   0.00004,     0.081 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.60121,   0.00001,     1.104 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   3,   0 },    289.61015,   0.00004,     0.071 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    309.84736,   0.00001,     0.097 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,  -2,   3,   0 },    309.84736,   0.00001,     0.062 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    309.84736,   0.00002,     0.082 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,   0,   1,   0 },    309.84736,   0.00002,     0.070 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,  -2,  -1,   0 },    226.47689,   0.00001,     0.028 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    226.47689,   0.00001,     0.107 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   2,  -3,   0 },    329.09123,   0.00001,     0.078 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },    277.23642,   0.00003,     0.073 },
 { {   0,  13, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.90730,   0.00001,     0.044 },
 { {   0,  13, -19,   0,   1,   0,   0,   0,  -2,   2,   0 },    321.90730,   0.00001,     0.251 },
 { {   0,  13, -19,   0,   2,   0,   0,   0,  -2,   0,   0 },     51.50556,   0.00004,     0.045 },
 { {   0,  13, -19,   0,   2,   0,   0,   0,  -2,   2,   0 },    231.50629,   0.00003,     0.245 },
 { {   0,  13, -19,   2,   0,   0,   0,   0,  -2,   0,   0 },     51.71525,   0.00001,     0.046 },
 { {   0,  13, -19,   2,   0,   0,   0,   0,  -2,   2,   0 },     51.71525,   0.00001,     0.201 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,  -1,   0 },    324.06536,   0.00002,     0.029 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    170.65998,   0.00002,     0.046 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    291.18251,   0.00001,     0.120 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.06536,   0.00003,     0.099 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    124.90756,   0.00002,     0.068 },
 { {   0,  13, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    230.30701,   0.00005,     0.046 },
 { {   0,  13, -18,   0,   1,   0,   0,   0,  -2,   2,   0 },    230.29745,   0.00004,     0.200 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -4,  -1,   0 },    231.68401,   0.00001,     0.017 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -4,   3,   0 },     51.68401,   0.00002,     0.179 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.55692,   0.00004,     0.049 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.68401,   0.00003,     0.160 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },     55.76071,   0.00001,     0.136 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },     41.86308,   0.00004,     0.169 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    158.57798,   0.00001,     0.242 },
 { {   0,  13, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.37932,   0.00003,     0.051 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    108.11079,   0.00002,     0.051 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     74.23946,   0.00006,     0.195 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    251.90606,   0.00001,     0.245 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    257.89750,   0.00005,     0.033 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    352.50918,   0.00005,     0.054 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    346.22672,   0.00002,     0.187 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    333.15816,   0.00003,     0.126 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    347.61143,   0.00017,     0.163 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.51723,   0.00004,     0.197 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   1,   0,   1 },    348.51193,   0.00001,     0.031 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    354.60515,   0.00014,     0.057 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    174.41677,   0.00016,     0.032 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   1,   0 },    174.46561,   0.00002,     0.023 },
 { {   0,  13, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.70557,   0.00001,     0.054 },
 { {   0,  13, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.70557,   0.00001,     0.125 },
 { {   0,  13, -15,   0,   2,   0,   0,   0,   2,   0,   0 },    179.75597,   0.00002,     0.032 },
 { {   0,  13, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.79436,   0.00002,     0.057 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -4,   2,   0 },    272.33858,   0.00001,     0.063 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    273.46920,   0.00003,     0.057 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     92.02382,   0.00002,     0.112 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },     74.65179,   0.00002,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    281.16137,   0.00006,     0.163 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    272.07699,   0.00061,     0.140 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    276.24303,   0.00012,     0.049 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   2,   0 },    274.83033,   0.00001,     0.030 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   3,   0 },    272.31784,   0.00001,     0.021 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,  -1,  -1 },     74.68747,   0.00001,     0.139 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },     74.68874,   0.00014,     0.165 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,   1 },     74.69079,   0.00002,     0.030 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.08180,   0.00041,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31549,   0.00002,     0.199 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,   0 },     94.75965,   0.00054,     0.031 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   1,   0 },     92.81847,   0.00010,     0.022 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   3,   0,   0 },     92.42414,   0.00001,     0.023 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   4,  -2,   0 },    271.51097,   0.00001,     0.033 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   4,   0,   0 },     92.46600,   0.00005,     0.018 },
 { {   0,  13, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    220.35571,   0.00004,     0.057 },
 { {   0,  13, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },    178.23363,   0.00002,     0.234 },
 { {   0,  13, -14,   0,   1,   0,   0,   0,   2,   0,   0 },    358.58266,   0.00003,     0.031 },
 { {   0,  13, -14,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.29680,   0.00001,     0.062 },
 { {   0,  13, -14,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.29680,   0.00001,     0.095 },
 { {   0,  13, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    182.24953,   0.00004,     0.060 },
 { {   0,  13, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    159.41786,   0.00001,     0.102 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.77296,   0.00002,     0.067 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.01877,   0.00002,     0.023 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.14179,   0.00004,     0.033 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.01576,   0.00032,     0.060 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.04128,   0.00023,     0.299 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.56271,   0.00009,     0.101 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.93238,   0.00005,     0.236 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -3,   0 },    358.45456,   0.00001,     0.032 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    338.76230,   0.00002,     0.054 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.88355,   0.00030,     0.195 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    180.01489,   0.00198,     0.123 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    180.07653,   0.00033,     0.047 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    178.46610,   0.00001,     0.029 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.81327,   0.00004,     0.141 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    359.99763,   0.00001,     0.049 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,   1 },     11.81322,   0.00001,     0.029 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },    179.84932,   0.00001,     0.145 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    178.73082,   0.00007,     0.157 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.07168,   0.00111,     0.051 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.93297,   0.00005,     0.166 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    359.91488,   0.00157,     0.030 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   1,   0 },    359.92418,   0.00015,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   2,   0 },    359.93047,   0.00004,     0.017 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   3,   0,   0 },    359.99126,   0.00003,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   4,   0,   0 },    359.92883,   0.00003,     0.017 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   4,   1,   0 },    359.93297,   0.00001,     0.014 },
 { {   0,  13, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    347.69253,   0.00003,     0.061 },
 { {   0,  13, -13,   0,   1,   0,   0,   0,  -2,   1,   0 },    347.69259,   0.00001,     0.306 },
 { {   0,  13, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },    347.69253,   0.00003,     0.100 },
 { {   0,  13, -12,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.02446,   0.00001,     0.058 },
 { {   0,  13, -12,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.02446,   0.00001,     0.107 },
 { {   0,  13, -12,   0,  -2,   0,   0,   0,   0,   0,   0 },    109.07302,   0.00002,     0.112 },
 { {   0,  13, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.42057,   0.00001,     0.064 },
 { {   0,  13, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.42057,   0.00001,     0.092 },
 { {   0,  13, -12,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.26145,   0.00005,     0.111 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    259.54112,   0.00004,     0.064 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     74.56401,   0.00002,     0.426 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    253.70534,   0.00001,     0.092 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.13939,   0.00001,     0.326 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.10430,   0.00005,     0.060 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,   1 },    287.10405,   0.00003,     0.098 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.16021,   0.00005,     0.294 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     38.92581,   0.00002,     0.110 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.22237,   0.00003,     0.193 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,  -1,   1 },    103.22457,   0.00001,     0.046 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.23302,   0.00021,     0.124 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,   1 },    103.22460,   0.00003,     0.029 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.71977,   0.00008,     0.049 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,   0,   0 },     92.09531,   0.00014,     0.030 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,   1,   0 },     93.49710,   0.00002,     0.021 },
 { {   0,  13, -11,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.15903,   0.00001,     0.062 },
 { {   0,  13, -11,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.15903,   0.00001,     0.097 },
 { {   0,  13, -11,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.42950,   0.00002,     0.064 },
 { {   0,  13, -11,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.42950,   0.00002,     0.092 },
 { {   0,  13, -11,   0,  -3,   0,   0,   0,  -2,   1,   0 },     10.43743,   0.00002,     0.624 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.23184,   0.00003,     0.068 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.23184,   0.00003,     0.085 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,   0,  -2,   0 },    179.79797,   0.00002,     0.060 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.79797,   0.00002,     0.100 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.70380,   0.00002,     0.036 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     26.51221,   0.00024,     0.068 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.96323,   0.00027,     0.741 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    199.53523,   0.00002,     0.084 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    203.60775,   0.00002,     0.447 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.66399,   0.00009,     0.320 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     29.28335,   0.00023,     0.099 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    200.01563,   0.00001,     0.043 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.75074,   0.00002,     0.046 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    205.52238,   0.00007,     0.029 },
 { {   0,  13, -10,   0,  -4,   0,   0,   0,  -2,   1,   0 },      7.12681,   0.00003,     1.455 },
 { {   0,  13, -10,   0,  -3,   0,   0,   0,  -2,   0,   0 },    350.98531,   0.00001,     0.072 },
 { {   0,  13, -10,   0,  -3,   0,   0,   0,  -2,   1,   0 },     68.97798,   0.00002,     1.659 },
 { {   0,  13, -10,   0,  -3,   0,   0,   0,  -2,   2,   0 },    350.98531,   0.00001,     0.079 },
 { {   0,  13, -10,   0,  -2,   0,   0,   0,  -2,   0,   0 },    256.53163,   0.00002,     0.073 },
 { {   0,  13, -10,   0,  -2,   0,   0,   0,  -2,   1,   0 },    248.22565,   0.00006,     1.929 },
 { {   0,  13, -10,   0,  -2,   0,   0,   0,  -2,   2,   0 },     76.53163,   0.00002,     0.079 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -4,   2,   0 },    105.89599,   0.00001,     0.085 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -2,   2 },    102.34519,   0.00002,     0.075 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -1,   2 },    102.34519,   0.00034,    60.894 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    281.05962,   0.00007,     0.073 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   1 },    307.79918,   0.00003,     5.459 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   2 },    102.34519,   0.00002,     0.076 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    280.36378,   0.00025,     2.858 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.77152,   0.00005,     0.077 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -1,   1,  -1 },    309.07531,   0.00001,     0.712 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    105.89599,   0.00001,     0.090 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.41831,   0.00002,     0.107 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   2,   0,   0 },     39.41831,   0.00002,     0.028 },
 { {   0,  13, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },    280.37537,   0.00002,     3.765 },
 { {   0,  13, -10,   2,   0,   0,   0,   0,  -1,  -1,   1 },    251.77003,   0.00003,   146.642 },
 { {   0,  13, -10,   4,   0,   0,   0,   0,   0,  -1,   0 },    351.01272,   0.00004,   359.272 },
 { {   0,  13,  -9,   0,  -7,   0,   0,   0,  -2,   1,   0 },     27.41798,   0.00001,    16.686 },
 { {   0,  13,  -9,   0,  -4,   0,   0,   0,  -2,   1,   0 },    352.59573,   0.00002,     3.197 },
 { {   0,  13,  -9,   0,  -2,   0,   0,   0,  -2,   0,   0 },     24.52528,   0.00001,     0.078 },
 { {   0,  13,  -9,   0,  -2,   0,   0,   0,  -2,   2,   0 },    204.52528,   0.00001,     0.073 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    189.53220,   0.00002,     0.079 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    229.03372,   0.00004,     1.538 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    251.81246,   0.00001,     0.072 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -1,  -1,   1 },     24.46001,   0.00009,    14.248 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     49.60070,   0.00007,     4.220 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    106.28062,   0.00002,     0.070 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    231.12356,   0.00001,     0.890 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    154.54748,   0.00002,     0.082 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },     67.14391,   0.00001,     0.097 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    348.12010,   0.00001,     0.027 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },    308.21504,   0.00002,     0.016 },
 { {   0,  13,  -9,   2,   0,   0,   0,   0,   0,  -1,   0 },    242.18415,   0.00002,    16.506 },
 { {   0,  13,  -8,  -3,   0,   0,   0,   0,  -2,   1,   0 },     31.66935,   0.00003,    18.178 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    134.57238,   0.00001,     0.086 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    314.57238,   0.00005,     0.075 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    122.65270,   0.00011,     8.068 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    314.57238,   0.00005,     0.076 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    134.57238,   0.00001,     0.088 },
 { {   0,  13,  -7,  -5,   0,   0,   0,   0,  -2,   1,   0 },    238.15252,   0.00003,   119.657 },
 { {   0,  13,  -7,  -3,   0,   0,   0,   0,  -1,   1,  -1 },    334.16114,   0.00017,   802.892 },
 { {   0,  13,  -6,  -7,   0,   0,   0,   0,  -2,   1,   0 },    328.34943,   0.00002,    13.942 },
 { {   0,  14, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    289.58224,   0.00002,     0.040 },
 { {   0,  14, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    289.59340,   0.00001,     0.653 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    310.47972,   0.00002,     0.103 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -3,   0 },     54.74072,   0.00004,     0.075 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },     17.25352,   0.00001,    59.908 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },     86.61140,   0.00004,     0.076 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     30.04707,   0.00001,     0.045 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    210.04707,   0.00001,     0.221 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },    227.01841,   0.00002,     0.070 },
 { {   0,  14, -20,   2,   0,   0,   0,   0,  -2,   0,   0 },     51.69061,   0.00001,     0.048 },
 { {   0,  14, -20,   2,   0,   0,   0,   0,  -2,   2,   0 },     51.69061,   0.00001,     0.179 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.99289,   0.00003,     0.048 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    147.88555,   0.00004,     0.105 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,   0,   1,   0 },    154.62472,   0.00001,     0.059 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     57.85934,   0.00001,     0.035 },
 { {   0,  14, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    230.27647,   0.00008,     0.048 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -4,  -1,   0 },    231.65986,   0.00001,     0.017 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -4,   1,   0 },     51.65986,   0.00002,     0.032 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     49.43860,   0.00003,     0.050 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.65986,   0.00002,     0.146 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     55.89697,   0.00001,     0.149 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.79625,   0.00005,     0.153 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     49.67245,   0.00002,     0.210 },
 { {   0,  14, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    218.52848,   0.00001,     0.050 },
 { {   0,  14, -18,   0,   1,   0,   0,   0,  -2,   2,   0 },    218.52848,   0.00001,     0.151 },
 { {   0,  14, -17,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.34239,   0.00002,     0.052 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    144.59176,   0.00001,     0.031 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    130.14358,   0.00002,     0.053 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    144.59176,   0.00001,     0.175 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    288.50911,   0.00001,     0.133 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     73.95543,   0.00005,     0.174 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    258.53593,   0.00004,     0.033 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -4,   2,   0 },    175.76018,   0.00001,     0.062 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },     34.94712,   0.00001,     0.032 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.85725,   0.00006,     0.056 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },      7.44721,   0.00003,     0.212 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    351.17808,   0.00004,     0.117 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },     16.97307,   0.00001,     0.154 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    346.78116,   0.00012,     0.148 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.41630,   0.00003,     0.175 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    355.16172,   0.00011,     0.055 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    174.96893,   0.00014,     0.032 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,   1,   0 },    175.02305,   0.00002,     0.022 },
 { {   0,  14, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.74187,   0.00001,     0.056 },
 { {   0,  14, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.74187,   0.00001,     0.116 },
 { {   0,  14, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.80895,   0.00002,     0.059 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -4,   2,   0 },    272.41652,   0.00001,     0.066 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    276.80442,   0.00001,     0.059 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     92.00518,   0.00003,     0.105 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },     74.29294,   0.00002,     0.055 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },    272.45282,   0.00001,     0.053 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    280.60248,   0.00004,     0.182 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    272.24255,   0.00048,     0.129 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    276.29800,   0.00010,     0.048 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },     74.33377,   0.00011,     0.149 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    189.27685,   0.00001,     0.167 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    271.87872,   0.00027,     0.052 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.31338,   0.00001,     0.177 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     95.68055,   0.00042,     0.031 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   1,   0 },     93.21093,   0.00008,     0.022 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   3,   0,   0 },     92.41992,   0.00001,     0.022 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   4,   0,   0 },     92.33189,   0.00004,     0.017 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    208.84951,   0.00003,     0.059 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    178.22955,   0.00003,     0.275 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },    227.29071,   0.00001,     0.104 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,   2,   0,   0 },    358.55802,   0.00002,     0.031 },
 { {   0,  14, -15,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.27216,   0.00001,     0.065 },
 { {   0,  14, -15,   3,   0,   0,   0,   0,  -2,   1,   0 },    236.27216,   0.00002,     0.470 },
 { {   0,  14, -15,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.27216,   0.00001,     0.090 },
 { {   0,  14, -15,   7,   0,   0,   0,   0,  -2,   1,   0 },     24.12651,   0.00002,   381.207 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    197.14197,   0.00004,     0.062 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,  -2,   1,   0 },    149.22386,   0.00001,     0.356 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,  -2,   2,   0 },    203.38769,   0.00001,     0.096 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,   0,   0,   0 },    272.54991,   0.00001,     0.115 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.77745,   0.00002,     0.070 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.91570,   0.00004,     0.034 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.74869,   0.00033,     0.063 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.12489,   0.00024,     0.367 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.56164,   0.00008,     0.095 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.92958,   0.00004,     0.277 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.86924,   0.00027,     0.222 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    180.27724,   0.00148,     0.114 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    180.11612,   0.00024,     0.045 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    191.92348,   0.00003,     0.130 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    359.25674,   0.00002,     0.047 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   1,   0,   1 },     11.92343,   0.00001,     0.029 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    175.96026,   0.00004,     0.143 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.91873,   0.00075,     0.049 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.90875,   0.00003,     0.150 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   0,   0 },      0.01970,   0.00120,     0.030 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    359.96080,   0.00016,     0.021 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   2,   0 },    359.92684,   0.00003,     0.017 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   3,  -3,   0 },    179.97100,   0.00001,     0.164 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   3,  -1,   0 },    179.97100,   0.00001,     0.031 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   3,   0,   0 },    359.96921,   0.00003,     0.022 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,  -2,   0 },    181.70141,   0.00002,     0.032 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.88101,   0.00001,     0.022 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,   0,   0 },    359.95643,   0.00003,     0.017 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,   1,   0 },    359.90876,   0.00001,     0.014 },
 { {   0,  14, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    346.82704,   0.00002,     0.063 },
 { {   0,  14, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },    347.72968,   0.00001,     0.379 },
 { {   0,  14, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },    346.82704,   0.00002,     0.094 },
 { {   0,  14, -13,  -3,   0,   0,   0,   0,  -2,   0,   0 },    303.00381,   0.00001,     0.060 },
 { {   0,  14, -13,  -3,   0,   0,   0,   0,  -2,   2,   0 },    303.00381,   0.00001,     0.101 },
 { {   0,  14, -13,   0,  -2,   0,   0,   0,  -2,   1,   0 },     95.17164,   0.00003,     0.528 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.38264,   0.00001,     0.066 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.38264,   0.00001,     0.087 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,   0,  -2,   0 },      1.24274,   0.00002,     0.059 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.24274,   0.00002,     0.103 },
 { {   0,  14, -13,   0,   0,  -1,   0,   0,  -4,   3,   0 },    179.99574,   0.00001,    63.874 },
 { {   0,  14, -13,   0,   0,  -1,   0,   0,  -2,   0,   0 },    180.09921,   0.00001,     0.067 },
 { {   0,  14, -13,   0,   0,  -1,   0,   0,  -2,   2,   0 },    180.09921,   0.00001,     0.087 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -4,   3,   0 },     78.48634,   0.00003,    20.160 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    266.37148,   0.00002,     0.067 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    266.37148,   0.00002,     0.087 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.08122,   0.00001,     0.409 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.07705,   0.00005,     0.062 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,   1 },    289.91059,   0.00002,     0.092 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.13381,   0.00006,     0.360 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     35.31336,   0.00001,     0.102 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.19881,   0.00003,     0.220 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.21077,   0.00017,     0.115 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,   1 },    103.20078,   0.00003,     0.028 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.57134,   0.00004,     0.047 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   0,   0 },     92.54742,   0.00011,     0.029 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   1,   0 },     93.90763,   0.00001,     0.021 },
 { {   0,  14, -13,   3,   0,   0,   0,   0,  -2,   1,   0 },    260.82777,   0.00002,     7.730 },
 { {   0,  14, -12,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.14247,   0.00001,     0.064 },
 { {   0,  14, -12,  -3,   0,   0,   0,   0,  -2,   1,   0 },    336.14247,   0.00001,     0.431 },
 { {   0,  14, -12,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.14247,   0.00001,     0.091 },
 { {   0,  14, -12,  -2,   0,   0,   0,   0,  -2,   0,   0 },      0.40510,   0.00002,     0.066 },
 { {   0,  14, -12,  -2,   0,   0,   0,   0,  -2,   2,   0 },      0.40510,   0.00002,     0.087 },
 { {   0,  14, -12,   0,  -3,   0,   0,   0,  -2,   0,   0 },     10.44918,   0.00005,     0.070 },
 { {   0,  14, -12,   0,  -3,   0,   0,   0,  -2,   1,   0 },     60.79464,   0.00001,     1.023 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,  -2,   0,   0 },      0.28475,   0.00002,     0.071 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,  -2,   1,   0 },    359.23321,   0.00003,     1.120 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,  -2,   2,   0 },      0.28475,   0.00002,     0.081 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,   0,  -2,   0 },    179.77415,   0.00002,     0.063 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.77415,   0.00002,     0.094 },
 { {   0,  14, -12,   0,  -1,   0,   0,   0,  -2,   1,   0 },    116.49780,   0.00005,     1.237 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },     19.62870,   0.00002,     0.082 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   2,  -1 },    352.33008,   0.00001,     0.075 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   3,  -1 },    352.33008,   0.00023,   240.318 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   4,  -1 },    352.33008,   0.00001,     0.075 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },     27.77479,   0.00002,     0.037 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     27.93050,   0.00034,     0.072 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.72280,   0.00065,     1.380 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    194.21334,   0.00005,     0.080 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    204.44145,   0.00003,     0.621 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.53577,   0.00010,     0.400 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     26.11804,   0.00022,     0.093 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,   2,   0 },     25.74959,   0.00002,     0.027 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.73302,   0.00001,     0.045 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    205.75124,   0.00006,     0.028 },
 { {   0,  14, -12,   3,   0,   0,   0,   0,  -1,   0,   0 },     99.71522,   0.00001,    61.092 },
 { {   0,  14, -11,  -4,   0,   0,   0,   0,  -2,   1,   0 },    150.05544,   0.00002,     0.540 },
 { {   0,  14, -11,  -3,   0,   0,   0,   0,  -2,   1,   0 },    359.80622,   0.00001,     0.758 },
 { {   0,  14, -11,  -2,   0,   0,   0,   0,  -2,   1,   0 },    344.26270,   0.00001,     1.269 },
 { {   0,  14, -11,   0,  -3,   0,   0,   0,  -2,   0,   0 },    108.64691,   0.00007,     0.076 },
 { {   0,  14, -11,   0,  -3,   0,   0,   0,  -2,   1,   0 },     81.02598,   0.00006,    44.104 },
 { {   0,  14, -11,   0,  -3,   0,   0,   0,  -2,   2,   0 },    328.24837,   0.00002,     0.075 },
 { {   0,  14, -11,   0,  -2,   0,   0,   0,  -2,   0,   0 },     71.86410,   0.00013,     0.076 },
 { {   0,  14, -11,   0,  -2,   0,   0,   0,  -2,   1,   0 },    338.23441,   0.00005,     9.348 },
 { {   0,  14, -11,   0,  -2,   0,   0,   0,  -2,   2,   0 },    145.22063,   0.00001,     0.075 },
 { {   0,  14, -11,   0,  -2,   0,   0,   0,   0,   0,   0 },    251.86371,   0.00003,     0.086 },
 { {   0,  14, -11,   0,  -1,   0,   0,   0,  -2,   1,   0 },     24.57093,   0.00006,     5.228 },
 { {   0,  14, -11,   0,   0,  -4,   0,   0,  -2,   1,   0 },     92.10541,   0.00001,     7.153 },
 { {   0,  14, -11,   0,   0,  -3,   0,   0,  -2,   1,   0 },    358.24802,   0.00003,     5.756 },
 { {   0,  14, -11,   0,   0,  -2,   0,   0,  -2,   1,   0 },    266.13598,   0.00001,     4.815 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    107.03863,   0.00003,     0.077 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    281.63014,   0.00026,     3.629 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    107.03863,   0.00003,     0.074 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -1,  -1,   1 },    128.80170,   0.00002,     2.249 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    101.66589,   0.00001,     1.636 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -1,   1,  -1 },    309.21827,   0.00003,     1.285 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.58528,   0.00003,     0.100 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     39.58528,   0.00003,     0.027 },
 { {   0,  14, -10,  -3,   0,   0,   0,   0,  -2,   1,   0 },    178.36788,   0.00003,     3.130 },
 { {   0,  14, -10,   0,  -2,   0,   0,   0,  -1,   1,  -1 },     12.46964,   0.00002,    18.793 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },     51.72905,   0.00002,     2.574 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,  -1,   1,  -1 },    202.05000,   0.00002,     4.508 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    308.12377,   0.00004,     0.073 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    224.07426,   0.00006,     2.006 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    264.21471,   0.00004,     0.078 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    128.19848,   0.00001,     0.025 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    308.19848,   0.00002,     0.440 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    308.19848,   0.00002,     0.041 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    308.19848,   0.00001,     0.027 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,   2,   0 },    308.19848,   0.00001,     0.016 },
 { {   0,  14,  -9,  -4,   0,   0,   0,   0,  -2,   1,   0 },    289.19325,   0.00001,     6.718 },
 { {   0,  14,  -9,   0,  -6,   0,   0,   0,   0,  -1,   0 },    287.40487,   0.00005,   234.929 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,  -2,   1,  -2 },    320.24046,   0.00002,     0.040 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,  -2,   2,  -2 },    320.24046,   0.00002,     0.087 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,  -1,   0 },     51.71689,   0.00001,     3.003 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   0,  -2 },    300.62284,   0.00011,     0.075 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   1,  -2 },    320.24046,   0.00205,  1808.606 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   2,  -2 },    318.81492,   0.00011,     0.075 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   2,   0,  -2 },    320.24046,   0.00002,     0.087 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   2,   1,  -2 },    320.24046,   0.00002,     0.040 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    315.55035,   0.00001,     0.078 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    135.55035,   0.00001,     0.073 },
 { {   0,  14,  -9,   3,   0,   0,   0,   0,   2,  -1,  -2 },     29.59174,   0.00003,    91.671 },
 { {   0,  14,  -8,  -3,   0,   0,   0,   0,   0,  -1,   0 },     96.28312,   0.00002,    10.697 },
 { {   0,  14,  -8,   0,  -2,   0,   0,   0,   2,  -2,   0 },    308.22025,   0.00001,     0.078 },
 { {   0,  14,  -8,   0,  -2,   0,   0,   0,   2,   0,   0 },    128.22025,   0.00001,     0.025 },
 { {   0,  14,  -7,  -8,   0,   0,   0,   0,  -2,   1,   0 },    302.03558,   0.00006,    45.223 },
 { {   0,  14,  -7,  -4,   0,   0,   0,   0,   0,   1,  -2 },    278.53756,   0.00001,    24.202 },
 { {   0,  14,  -6, -10,   0,   0,   0,   0,  -2,   1,   0 },     23.73039,   0.00002,    24.240 },
 { {   0,  14,  -1, -16,   0,   0,   0,   0,   0,  -1,   0 },     44.00534,   0.00002,   185.940 },
 { {   0,  15, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    289.58390,   0.00001,     0.078 },
 { {   0,  15, -24,   0,   0,   0,   0,   0,   0,   1,   0 },    289.58390,   0.00001,     0.073 },
 { {   0,  15, -23,  19,   0,   0,   0,   0,  -2,   1,   0 },     96.41331,   0.00001,   321.466 },
 { {   0,  15, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    311.07705,   0.00002,     0.110 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.27156,   0.00004,     0.049 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    169.00825,   0.00001,     0.162 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    147.30800,   0.00003,     0.113 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   2,  -2,   0 },     57.85362,   0.00002,     0.383 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },    144.05930,   0.00002,     0.063 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     57.84556,   0.00001,     0.034 },
 { {   0,  15, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    230.24809,   0.00007,     0.049 },
 { {   0,  15, -19,   0,  -1,   0,   0,   0,  -2,   0,   0 },     60.29258,   0.00003,     0.051 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -4,   1,   0 },     51.63337,   0.00002,     0.032 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,  -1,   0 },    231.63337,   0.00001,     0.031 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     48.83057,   0.00002,     0.052 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.63337,   0.00002,     0.134 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     52.87585,   0.00003,     0.164 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.76873,   0.00003,     0.140 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },     51.47605,   0.00001,     0.047 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,   0,   0,   0 },     50.74668,   0.00003,     0.186 },
 { {   0,  15, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    218.50190,   0.00002,     0.052 },
 { {   0,  15, -18,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.30479,   0.00001,     0.054 },
 { {   0,  15, -18,   0,  -1,   0,   0,   0,  -2,   2,   0 },    235.30479,   0.00001,     0.124 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    138.98630,   0.00001,     0.146 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     77.05503,   0.00002,     0.157 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    138.98630,   0.00001,     0.051 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    259.21647,   0.00001,     0.056 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    259.12024,   0.00003,     0.032 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.23392,   0.00005,     0.058 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },      5.46800,   0.00002,     0.244 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    350.46726,   0.00002,     0.109 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    349.86436,   0.00014,     0.135 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.30600,   0.00002,     0.158 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    355.64194,   0.00008,     0.053 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    175.56471,   0.00013,     0.031 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   1,   0 },    175.63841,   0.00001,     0.022 },
 { {   0,  15, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.77395,   0.00001,     0.058 },
 { {   0,  15, -17,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.77395,   0.00001,     0.108 },
 { {   0,  15, -17,   4,   0,   0,   0,   0,  -2,   0,   0 },     57.83657,   0.00001,     0.066 },
 { {   0,  15, -17,   4,   0,   0,   0,   0,  -2,   2,   0 },     57.83657,   0.00001,     0.089 },
 { {   0,  15, -17,  11,   0,   0,   0,   0,   0,  -1,   0 },     93.55997,   0.00001,    41.017 },
 { {   0,  15, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.82113,   0.00002,     0.061 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.12923,   0.00002,     0.061 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.30733,   0.00002,     0.323 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },     90.17508,   0.00003,     0.098 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },     73.88898,   0.00002,     0.057 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.85529,   0.00002,     0.205 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    272.30562,   0.00040,     0.119 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    272.67634,   0.00004,     0.046 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },     87.69337,   0.00006,     0.136 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   1,   0,   1 },    253.94436,   0.00002,     0.029 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },    221.24618,   0.00002,     0.151 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.06665,   0.00019,     0.050 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   0,   0 },     95.09055,   0.00033,     0.030 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   1,   0 },     92.97057,   0.00006,     0.022 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   3,  -2,   0 },     92.41536,   0.00001,     0.053 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   3,   0,   0 },     92.41536,   0.00001,     0.022 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },     92.24715,   0.00001,     0.022 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   4,   0,   0 },     92.26500,   0.00002,     0.017 },
 { {   0,  15, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    226.77197,   0.00002,     0.061 },
 { {   0,  15, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },    228.01062,   0.00001,     0.098 },
 { {   0,  15, -16,   3,   0,   0,   0,   0,  -2,   0,   0 },    236.24558,   0.00001,     0.068 },
 { {   0,  15, -16,   3,   0,   0,   0,   0,  -2,   2,   0 },    236.24558,   0.00001,     0.085 },
 { {   0,  15, -15,   0,  -2,   0,   0,   0,  -2,   0,   0 },    209.27884,   0.00002,     0.064 },
 { {   0,  15, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    189.55338,   0.00006,     0.065 },
 { {   0,  15, -15,   0,  -1,   0,   0,   0,  -2,   2,   0 },    185.37016,   0.00001,     0.090 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -6,   1,   2 },    359.60386,   0.00002,     0.040 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -6,   2,   2 },    359.60386,   0.00002,     0.087 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   0,   2 },    359.60386,   0.00012,     0.075 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   1,   2 },    359.60386,   0.00211,   115.724 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.77817,   0.00002,     0.074 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   2,   2 },    359.60386,   0.00012,     0.075 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.93779,   0.00002,     0.035 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.79088,   0.00033,     0.065 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   2 },    359.60386,   0.00002,     0.087 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.95063,   0.00028,     0.476 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   2 },    359.60386,   0.00002,     0.040 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.53061,   0.00009,     0.090 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.92880,   0.00005,     0.335 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.12186,   0.00002,     0.058 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.02806,   0.00025,     0.258 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    180.64600,   0.00113,     0.107 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    181.27990,   0.00019,     0.044 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.02030,   0.00003,     0.120 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   1,   0,   0 },      0.22285,   0.00001,     0.046 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    181.84093,   0.00004,     0.131 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.81196,   0.00050,     0.048 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.88219,   0.00002,     0.137 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    359.98553,   0.00088,     0.029 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   1,   0 },    359.92511,   0.00014,     0.021 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   2,   0 },    359.90857,   0.00002,     0.016 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   3,   0,   0 },    359.94664,   0.00003,     0.022 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   4,  -2,   0 },    178.73309,   0.00001,     0.031 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   4,   0,   0 },    359.93260,   0.00002,     0.017 },
 { {   0,  15, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    346.87338,   0.00005,     0.065 },
 { {   0,  15, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    346.87338,   0.00002,     0.496 },
 { {   0,  15, -15,   2,   0,   0,   0,   0,  -3,   1,   1 },    140.05630,   0.00002,    54.813 },
 { {   0,  15, -14,  -3,   0,   0,   0,   0,  -2,   0,   0 },    302.98114,   0.00001,     0.063 },
 { {   0,  15, -14,  -3,   0,   0,   0,   0,  -2,   2,   0 },    302.98114,   0.00001,     0.095 },
 { {   0,  15, -14,   0,  -3,   0,   0,   0,  -2,   0,   0 },    282.15080,   0.00001,     0.068 },
 { {   0,  15, -14,   0,  -3,   0,   0,   0,  -2,   2,   0 },    282.15080,   0.00001,     0.084 },
 { {   0,  15, -14,   0,  -2,   0,   0,   0,  -2,   0,   0 },     94.75609,   0.00006,     0.069 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.34410,   0.00002,     0.069 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.34410,   0.00002,     0.083 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,   0,  -2,   0 },      1.22032,   0.00002,     0.062 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.22032,   0.00002,     0.097 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   0,   1 },    283.19826,   0.00007,     0.075 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   1,   1 },    283.38829,   0.00073,    22.161 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   2,   1 },    283.37950,   0.00004,     0.076 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    277.53224,   0.00002,     0.070 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    110.34302,   0.00003,     0.910 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    233.43738,   0.00001,     0.082 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.04683,   0.00001,     0.550 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.04997,   0.00005,     0.065 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,   1 },    290.67189,   0.00002,     0.087 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.10539,   0.00007,     0.465 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    242.26553,   0.00001,     0.062 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     36.93639,   0.00002,     0.096 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    129.26008,   0.00002,     0.042 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.17285,   0.00003,     0.255 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.18681,   0.00014,     0.107 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,   1 },    103.17458,   0.00002,     0.028 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.82792,   0.00003,     0.046 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,   0,   0 },     91.32889,   0.00011,     0.028 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,   1,   0 },     94.10798,   0.00001,     0.021 },
 { {   0,  15, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },     48.63053,   0.00002,     0.986 },
 { {   0,  15, -14,   2,   0,   0,   0,   0,  -2,   1,   0 },    240.51233,   0.00005,    28.153 },
 { {   0,  15, -13,  -3,   0,   0,   0,   0,  -2,   0,   0 },    336.12384,   0.00001,     0.067 },
 { {   0,  15, -13,  -3,   0,   0,   0,   0,  -2,   2,   0 },    336.12384,   0.00001,     0.086 },
 { {   0,  15, -13,   0,  -4,   0,   0,   0,  -2,   1,   0 },     55.02833,   0.00002,     2.293 },
 { {   0,  15, -13,   0,  -3,   0,   0,   0,  -2,  -1,   2 },    181.77058,   0.00004,    54.678 },
 { {   0,  15, -13,   0,  -3,   0,   0,   0,  -2,   0,   0 },     53.02827,   0.00004,     0.073 },
 { {   0,  15, -13,   0,  -3,   0,   0,   0,  -2,   1,   0 },    359.05963,   0.00004,     2.843 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,  -2,   0,   0 },    358.81122,   0.00003,     0.074 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,  -2,   1,   0 },    348.75063,   0.00002,     3.738 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.17801,   0.00004,     0.077 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,   0,  -2,   0 },    359.74833,   0.00002,     0.065 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,   0,  -1,   0 },    359.75349,   0.00002,     0.490 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.74833,   0.00002,     0.089 },
 { {   0,  15, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    143.52499,   0.00004,     0.074 },
 { {   0,  15, -13,   0,  -1,   0,   0,   0,  -2,   1,   0 },     78.00496,   0.00010,     5.459 },
 { {   0,  15, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },     52.56796,   0.00004,     0.076 },
 { {   0,  15, -13,   0,   0,  -2,   0,   0,  -2,   0,   0 },      0.10078,   0.00002,     0.075 },
 { {   0,  15, -13,   0,   0,  -2,   0,   0,  -2,   1,   0 },    358.57376,   0.00004,     5.996 },
 { {   0,  15, -13,   0,   0,  -2,   0,   0,  -2,   2,   0 },      0.10078,   0.00002,     0.076 },
 { {   0,  15, -13,   0,   0,  -1,   0,   0,  -2,   1,   0 },     97.84768,   0.00005,     7.528 },
 { {   0,  15, -13,   0,   0,   0,  -2,   0,  -2,   1,   0 },    180.59082,   0.00002,     8.151 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   0,   0 },     26.27522,   0.00006,     0.026 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   1,   0 },     26.31508,   0.00032,     0.040 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },     26.39613,   0.00029,     0.086 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.24191,   0.00025,     0.038 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,  -1,   2 },    202.81352,   0.00001,     4.262 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     26.23847,   0.00364,     0.075 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },      3.94901,   0.00009,    14.735 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,  -2 },    206.32315,   0.00005,     0.037 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.32315,   0.02514,    10.112 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   2 },    206.32315,   0.00005,     0.037 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },     26.24572,   0.00134,     0.076 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },     26.24572,   0.00009,     0.038 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.28091,   0.00006,     1.014 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },     29.51352,   0.00008,     0.534 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     27.43744,   0.00010,     0.088 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   1,   0 },     26.33316,   0.00026,     0.041 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   2,   0 },     26.30169,   0.00003,     0.026 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    202.63794,   0.00001,     0.042 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    205.73031,   0.00002,     0.044 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    215.56343,   0.00004,     0.028 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   2,   2,   0 },    344.29833,   0.00001,     0.016 },
 { {   0,  15, -13,   0,   0,   2,   0,   0,  -2,   1,   0 },     16.03973,   0.00002,    32.265 },
 { {   0,  15, -13,   0,   0,   3,   0,   0,  -2,   1,   0 },    278.90911,   0.00001,   338.444 },
 { {   0,  15, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    287.82116,   0.00005,     0.075 },
 { {   0,  15, -13,   0,   1,   0,   0,   0,  -2,   1,   0 },     70.82335,   0.00004,    68.563 },
 { {   0,  15, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },    144.08411,   0.00001,     0.076 },
 { {   0,  15, -13,   0,   2,   0,   0,   0,  -2,   1,   0 },    254.40729,   0.00001,    14.344 },
 { {   0,  15, -12,  -4,   0,   0,   0,   0,  -2,   0,   0 },    150.02814,   0.00002,     0.069 },
 { {   0,  15, -12,  -4,   0,   0,   0,   0,  -2,   2,   0 },    150.02814,   0.00002,     0.083 },
 { {   0,  15, -12,  -3,   0,   0,   0,   0,  -2,   0,   0 },    359.78355,   0.00001,     0.072 },
 { {   0,  15, -12,  -3,   0,   0,   0,   0,  -2,   2,   0 },    359.78355,   0.00001,     0.080 },
 { {   0,  15, -12,  -2,   0,   0,   0,   0,  -2,   1,   0 },      2.18335,   0.00001,     6.163 },
 { {   0,  15, -12,   0,  -4,   0,   0,   0,  -2,   1,   0 },      7.06779,   0.00003,     1.773 },
 { {   0,  15, -12,   0,  -3,   0,   0,   0,  -2,   1,   0 },     74.16277,   0.00001,     1.543 },
 { {   0,  15, -12,   0,  -2,   0,   0,   0,  -2,   1,   0 },    253.50984,   0.00002,     1.365 },
 { {   0,  15, -12,   0,  -2,   0,   0,   0,  -1,  -2,   1 },    283.23830,   0.00001,     0.076 },
 { {   0,  15, -12,   0,  -2,   0,   0,   0,  -1,  -1,   1 },    281.98931,   0.00002,    81.636 },
 { {   0,  15, -12,   0,  -2,   0,   0,   0,  -1,   0,   1 },    283.23830,   0.00001,     0.075 },
 { {   0,  15, -12,   0,  -2,   0,   0,   0,  -1,   1,  -1 },    283.27862,   0.00003,     3.113 },
 { {   0,  15, -12,   0,   0,  -1,   0,   0,  -1,   0,   0 },    205.32375,   0.00002,    50.387 },
 { {   0,  15, -12,   0,   0,   0,  -1,   0,  -1,   0,   0 },     33.48860,   0.00001,   454.655 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -4,   1,   0 },    283.38622,   0.00002,     0.042 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },    283.45895,   0.00001,     0.095 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,  -1,   0 },     99.70101,   0.00002,     0.026 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   0,   0 },     99.70101,   0.00017,     0.040 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   1,   0 },     99.70101,   0.00016,     0.087 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    283.30498,   0.00002,     0.081 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    223.34392,   0.00001,     1.110 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -2,   0 },     99.70101,   0.00005,     0.038 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -1,   0 },     99.74920,   0.00082,     0.076 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -1,   1 },    128.93675,   0.00010,     5.529 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -2 },    279.70101,   0.00003,     0.037 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    309.31612,   0.00012,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },     99.70101,   0.01467,    70.915 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,   2 },    279.70101,   0.00003,     0.037 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },    309.25086,   0.00060,     6.551 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },     99.74920,   0.00082,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   2,  -1 },    309.25110,   0.00003,     0.076 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   2,   0 },     99.70101,   0.00005,     0.038 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    252.42539,   0.00002,     1.146 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },     99.70101,   0.00016,     0.087 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    129.34032,   0.00002,     0.088 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     99.70101,   0.00017,     0.040 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   1,   0 },     99.70101,   0.00002,     0.026 },
 { {   0,  15, -11,  -6,   0,   0,   0,   0,  -2,   1,   0 },    121.50360,   0.00001,     0.776 },
 { {   0,  15, -11,  -3,   0,   0,   0,   0,  -2,   1,   0 },    178.70092,   0.00003,     3.267 },
 { {   0,  15, -11,   0,  -3,   0,   0,   0,   0,  -1,   0 },    240.03332,   0.00002,     7.948 },
 { {   0,  15, -11,   0,  -2,   0,   0,   0,   0,  -1,   0 },    183.76439,   0.00008,    24.088 },
 { {   0,  15, -11,   0,  -2,   0,   0,   0,   0,   0,   0 },     25.72852,   0.00002,     0.076 },
 { {   0,  15, -11,   0,  -1,   0,   0,   0,   0,  -2,   0 },    144.73659,   0.00003,     0.076 },
 { {   0,  15, -11,   0,  -1,   0,   0,   0,   0,   0,   0 },    324.73659,   0.00003,     0.075 },
 { {   0,  15, -11,   0,   0,  -5,   0,   0,   0,  -1,   0 },     71.53021,   0.00001,    23.448 },
 { {   0,  15, -11,   0,   0,  -4,   0,   0,   0,  -1,   0 },     44.37392,   0.00001,   114.947 },
 { {   0,  15, -11,   0,   0,  -1,   0,   0,   0,  -1,   0 },    220.77948,   0.00002,    10.736 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,  -3,   0,   0 },    356.76409,   0.00001,     0.042 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    231.45322,   0.00002,     0.088 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    231.69791,   0.00004,     0.076 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    231.71999,   0.00082,     7.868 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,   0,  -1 },     22.88466,   0.00013,    25.212 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     59.22484,   0.00003,     0.075 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    128.17985,   0.00001,     0.025 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    308.17985,   0.00001,     0.607 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    277.04917,   0.00002,     0.040 },
 { {   0,  15, -10,  -2,   0,   0,   0,   0,   0,  -1,   0 },    262.65160,   0.00002,    15.693 },
 { {   0,  15, -10,   0,  -1,   0,   0,   0,   1,  -1,  -1 },     33.22582,   0.00002,    91.100 },
 { {   0,  15, -10,   0,   0,  -2,   0,   0,   1,  -1,  -1 },    127.83628,   0.00003,   184.133 },
 { {   0,  15, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    153.79544,   0.00006,    13.638 },
 { {   0,  15,  -9,  -8,   0,   0,   0,   0,  -2,  -1,   2 },    123.75341,   0.00001,    53.052 },
 { {   0,  15,  -9,  -8,   0,   0,   0,   0,  -2,   1,   0 },    303.75341,   0.00001,     2.838 },
 { {   0,  15,  -9,  -6,   0,   0,   0,   0,  -1,  -1,   1 },    118.58797,   0.00003,   108.145 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    207.17736,   0.00007,     0.040 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,  -2,   0,   0 },    216.91471,   0.00009,     0.087 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -3,   0 },    186.80597,   0.00002,     0.038 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -2,   0 },    184.30702,   0.00022,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -1,  -2 },     27.17736,   0.00001,     0.037 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -1,   0 },    207.17736,   0.00613,  2809.635 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -1,   2 },     27.17736,   0.00001,     0.037 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,   0,   0 },    189.14737,   0.00023,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,   1,   0 },    186.80597,   0.00002,     0.038 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   2,  -2,   0 },    216.91471,   0.00009,     0.087 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   2,  -1,   0 },    207.17736,   0.00007,     0.040 },
 { {   0,  15,  -9,   0,  -2,   0,   0,   0,   2,  -1,  -2 },    128.27240,   0.00001,     6.709 },
 { {   0,  15,  -9,   0,   0,  -2,   0,   0,   2,  -1,  -2 },    308.13634,   0.00001,    20.689 },
 { {   0,  15,  -9,   0,   0,  -1,   0,   0,   2,  -1,  -2 },     30.58134,   0.00001,    69.511 },
 { {   0,  15,  -6, -13,   0,   0,   0,   0,  -2,   1,   0 },    116.90781,   0.00003,    92.724 },
 { {   0,  15,  -6, -11,   0,   0,   0,   0,  -1,   1,  -1 },     27.31838,   0.00008,   846.029 },
 { {   0,  15,  -5, -15,   0,   0,   0,   0,  -2,   1,   0 },    122.90196,   0.00003,    13.486 },
 { {   0,  16, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },     32.54653,   0.00002,     0.075 },
 { {   0,  16, -26,   0,   0,   0,   0,   0,   0,   1,   0 },     32.54653,   0.00002,     0.075 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,  -2,  -2,   0 },     47.94606,   0.00002,     0.021 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     47.94606,   0.00002,     0.048 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    227.94606,   0.00002,     0.108 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,   0,   1,   0 },    227.94606,   0.00002,     0.058 },
 { {   0,  16, -22,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.42400,   0.00002,     0.049 },
 { {   0,  16, -22,   0,   2,   0,   0,   0,  -2,   2,   0 },    231.42400,   0.00002,     0.168 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -4,   0,   0 },    324.05367,   0.00002,     0.022 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -4,   2,   0 },    324.05367,   0.00002,     0.056 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    166.98984,   0.00005,     0.051 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    203.58647,   0.00002,     0.147 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },     56.41313,   0.00001,     0.047 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.05367,   0.00002,     0.121 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    236.41313,   0.00001,     0.200 },
 { {   0,  16, -21,   0,   1,   0,   0,   0,  -2,   0,   0 },    230.21169,   0.00005,     0.051 },
 { {   0,  16, -20,   0,  -1,   0,   0,   0,  -2,   0,   0 },     60.26411,   0.00002,     0.053 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -4,   1,   0 },     51.60459,   0.00002,     0.033 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,  -3,   0 },    231.60459,   0.00001,     0.017 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,  -1,   0 },    231.60459,   0.00003,     0.031 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     48.96770,   0.00002,     0.053 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.60459,   0.00002,     0.123 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },     53.19368,   0.00002,     0.183 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.74478,   0.00003,     0.128 },
 { {   0,  16, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    218.47343,   0.00001,     0.054 },
 { {   0,  16, -20,   0,   1,   0,   0,   0,  -2,   2,   0 },     38.47343,   0.00001,     0.127 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    134.49381,   0.00001,     0.224 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,  -2,   3,   0 },    134.49381,   0.00001,     0.045 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,   0,   0,   0 },     77.23391,   0.00002,     0.143 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,   2,  -2,   0 },     79.88991,   0.00002,     0.191 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    259.40383,   0.00001,     0.032 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,  -2,   0 },    217.17716,   0.00001,     0.023 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    342.58115,   0.00004,     0.060 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     17.35024,   0.00004,     0.288 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },      9.27677,   0.00002,     0.102 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    350.84317,   0.00012,     0.125 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    175.57518,   0.00001,     0.047 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.18284,   0.00001,     0.144 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    356.09281,   0.00006,     0.051 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    176.01791,   0.00011,     0.031 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,   1,   0 },    176.08779,   0.00001,     0.022 },
 { {   0,  16, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.80202,   0.00001,     0.060 },
 { {   0,  16, -18,   0,   1,   0,   0,   0,  -2,   2,   0 },     90.80202,   0.00001,     0.101 },
 { {   0,  16, -18,   6,   0,   0,   0,   0,  -2,   1,   0 },    280.12144,   0.00003,     3.530 },
 { {   0,  16, -17,   0,  -1,   0,   0,   0,  -2,   0,   0 },    104.83095,   0.00003,     0.063 },
 { {   0,  16, -17,   0,  -1,   0,   0,   0,  -2,   1,   0 },    104.51844,   0.00001,     0.391 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.84249,   0.00003,     0.064 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.17611,   0.00003,     0.404 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },     90.06826,   0.00002,     0.093 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },     73.45573,   0.00002,     0.060 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.91688,   0.00005,     0.235 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    271.76365,   0.00029,     0.111 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,   1,   0 },    269.61772,   0.00002,     0.045 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   1,   0,  -1 },     82.92838,   0.00004,     0.126 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   1,   0,   1 },    253.51392,   0.00001,     0.029 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.05656,   0.00016,     0.049 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   0,   0 },     91.40505,   0.00025,     0.030 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   1,   0 },     91.29801,   0.00003,     0.021 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   3,  -2,   0 },     92.40876,   0.00001,     0.051 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   3,   0,   0 },     92.40876,   0.00001,     0.022 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   4,  -2,   0 },     84.13604,   0.00001,     0.031 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   4,  -1,   0 },     91.92859,   0.00003,     0.022 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   4,   0,   0 },     92.30309,   0.00003,     0.017 },
 { {   0,  16, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },    198.13487,   0.00002,     0.064 },
 { {   0,  16, -17,   2,   0,   0,   0,   0,  -2,   0,   0 },     39.52792,   0.00001,     0.068 },
 { {   0,  16, -17,   2,   0,   0,   0,   0,  -2,   2,   0 },     39.52792,   0.00001,     0.084 },
 { {   0,  16, -17,   6,   0,   0,   0,   0,  -1,  -1,   1 },    335.82687,   0.00002,   284.671 },
 { {   0,  16, -17,   8,   0,   0,   0,   0,   0,  -1,   0 },    225.65979,   0.00006,   164.206 },
 { {   0,  16, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },    202.77330,   0.00006,     0.068 },
 { {   0,  16, -16,   0,  -1,   0,   0,   0,  -2,   1,   0 },    199.30399,   0.00003,     0.642 },
 { {   0,  16, -16,   0,   0,  -1,   0,   0,  -2,   1,   0 },     92.15805,   0.00001,     0.663 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -4,   2,   0 },    178.69276,   0.00002,     0.077 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    359.90691,   0.00002,     0.024 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.15815,   0.00003,     0.036 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.88742,   0.00037,     0.068 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.86179,   0.00042,     0.679 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.32120,   0.00007,     0.085 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -1,   0,   0 },      1.18803,   0.00005,     0.424 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    359.17352,   0.00005,     0.061 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.11892,   0.00022,     0.308 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    180.83726,   0.00086,     0.100 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    180.29125,   0.00013,     0.043 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.10539,   0.00002,     0.112 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.84969,   0.00034,     0.047 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.85333,   0.00002,     0.127 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    359.86029,   0.00064,     0.029 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   1,   0 },    359.85288,   0.00012,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   3,   0 },    359.85680,   0.00001,     0.013 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   3,   0,   0 },    359.92182,   0.00002,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.19023,   0.00001,     0.030 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   4,   0,   0 },    359.85649,   0.00002,     0.017 },
 { {   0,  16, -16,   0,   0,   1,   0,   0,  -2,   1,   0 },     87.35265,   0.00001,     0.695 },
 { {   0,  16, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    347.66656,   0.00007,     0.068 },
 { {   0,  16, -16,   2,   0,   0,   0,   0,  -2,   1,   0 },    306.64670,   0.00002,     2.439 },
 { {   0,  16, -16,   3,   0,   0,   0,   0,  -2,   1,   0 },    180.34653,   0.00001,     8.218 },
 { {   0,  16, -16,   6,   0,   0,   0,   0,   0,  -1,   0 },    185.91687,   0.00001,    17.459 },
 { {   0,  16, -15,  -3,   0,   0,   0,   0,  -2,   0,   0 },    302.95650,   0.00001,     0.065 },
 { {   0,  16, -15,  -3,   0,   0,   0,   0,  -2,   1,   0 },    302.95650,   0.00002,     0.483 },
 { {   0,  16, -15,  -3,   0,   0,   0,   0,  -2,   2,   0 },    302.95650,   0.00001,     0.089 },
 { {   0,  16, -15,  -2,   0,   0,   0,   0,  -2,   0,   0 },    272.70001,   0.00002,     0.068 },
 { {   0,  16, -15,  -2,   0,   0,   0,   0,  -2,   2,   0 },    272.70001,   0.00002,     0.085 },
 { {   0,  16, -15,   0,  -3,   0,   0,   0,  -2,   1,   0 },    102.45069,   0.00002,     1.377 },
 { {   0,  16, -15,   0,  -2,   0,   0,   0,  -2,   0,   0 },    105.24386,   0.00002,     0.072 },
 { {   0,  16, -15,   0,  -2,   0,   0,   0,  -2,   1,   0 },    110.89213,   0.00004,     1.558 },
 { {   0,  16, -15,   0,  -2,   0,   0,   0,  -2,   2,   0 },    105.24386,   0.00002,     0.079 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    178.30454,   0.00005,     0.072 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,  -2,   1,   0 },    150.49324,   0.00002,     1.793 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,  -2,   2,   0 },    178.30533,   0.00002,     0.079 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,   0,  -2,   0 },    359.65211,   0.00001,     0.064 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,   0,  -1,   0 },    181.19577,   0.00002,     0.429 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.68449,   0.00004,     0.092 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.40438,   0.00005,     0.073 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   1 },    103.01791,   0.00004,     3.262 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    100.32917,   0.00017,     2.113 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    245.44315,   0.00002,     0.078 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.55277,   0.00003,     0.838 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    103.04916,   0.00008,     0.068 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   0,   1 },    291.46619,   0.00002,     0.083 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.07499,   0.00010,     0.655 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    261.90888,   0.00002,     0.091 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,  -2,   1 },    283.14626,   0.00002,     0.097 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.14452,   0.00003,     0.303 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.16210,   0.00011,     0.100 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },     84.78625,   0.00001,     0.249 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     94.27352,   0.00009,     0.028 },
 { {   0,  16, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },     38.43103,   0.00003,     0.073 },
 { {   0,  16, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    342.92642,   0.00004,     2.570 },
 { {   0,  16, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },    250.70772,   0.00001,     0.078 },
 { {   0,  16, -15,   0,   2,   0,   0,   0,  -2,   1,   0 },     78.15364,   0.00003,     3.281 },
 { {   0,  16, -15,   1,   0,   0,   0,   0,  -2,   1,   0 },    202.31743,   0.00002,    17.148 },
 { {   0,  16, -14,  -3,   0,   0,   0,   0,  -3,   1,   1 },     28.49513,   0.00002,    68.243 },
 { {   0,  16, -14,  -3,   0,   0,   0,   0,  -2,   1,   0 },    328.65267,   0.00001,     0.936 },
 { {   0,  16, -14,  -2,   0,   0,   0,   0,  -2,   1,   0 },    182.89113,   0.00003,     1.863 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -4,   2,   0 },    265.33679,   0.00001,     0.087 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -2,   0,   0 },     85.33679,   0.00006,     0.075 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -2,   1,   0 },    187.27793,   0.00003,   197.991 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -2,   2,   0 },     85.33679,   0.00006,     0.075 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,   0,   0,   0 },    265.33679,   0.00001,     0.087 },
 { {   0,  16, -14,   0,  -5,   0,   0,   0,  -2,   1,   0 },     38.20982,   0.00002,     9.514 },
 { {   0,  16, -14,   0,  -3,   0,   0,   0,  -2,   0,   0 },    223.09637,   0.00003,     0.077 },
 { {   0,  16, -14,   0,  -3,   0,   0,   0,  -2,   1,   0 },    354.17589,   0.00004,     3.653 },
 { {   0,  16, -14,   0,  -3,   0,   0,   0,  -2,   2,   0 },    237.57516,   0.00001,     0.074 },
 { {   0,  16, -14,   0,  -2,   0,   0,   0,  -2,   0,   0 },    179.94719,   0.00002,     0.078 },
 { {   0,  16, -14,   0,  -2,   0,   0,   0,  -2,   1,   0 },    343.18353,   0.00002,     2.793 },
 { {   0,  16, -14,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.94719,   0.00002,     0.073 },
 { {   0,  16, -14,   0,  -2,   0,   0,   0,   0,   0,   0 },    359.69776,   0.00005,     0.084 },
 { {   0,  16, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },     61.84769,   0.00002,     0.078 },
 { {   0,  16, -14,   0,  -1,   0,   0,   0,  -2,   1,   0 },     79.05852,   0.00004,     2.261 },
 { {   0,  16, -14,   0,  -1,   0,   0,   0,  -2,   2,   0 },     94.66743,   0.00003,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -4,   2,   0 },    202.97231,   0.00001,     0.091 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    207.28357,   0.00001,     0.038 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    207.88469,   0.00030,     0.079 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.56171,   0.00064,     1.899 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     24.30970,   0.00009,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   3,   0 },     27.28357,   0.00001,     0.037 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,  -1,   0 },    206.35562,   0.00002,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,  -1,   1 },    166.85531,   0.00003,     5.165 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.35562,   0.00033,     2.775 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   1,  -1 },    330.25041,   0.00002,     1.897 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   1,   0 },    206.35562,   0.00002,     0.078 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },      8.82241,   0.00004,     0.069 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.61702,   0.00021,     0.802 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     28.06210,   0.00035,     0.083 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   2,  -3,   0 },    345.42944,   0.00001,     0.331 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    251.34359,   0.00001,     0.043 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    214.61470,   0.00004,     0.027 },
 { {   0,  16, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },     17.73914,   0.00002,     1.637 },
 { {   0,  16, -14,   1,   0,   0,   0,   0,  -1,   1,  -1 },    327.52623,   0.00002,   219.684 },
 { {   0,  16, -14,   2,   0,   0,   0,   0,   0,  -1,   0 },     92.58397,   0.00001,     5.434 },
 { {   0,  16, -13,  -3,   0,   0,   0,   0,  -2,   0,   0 },    191.20347,   0.00003,     0.075 },
 { {   0,  16, -13,  -3,   0,   0,   0,   0,  -2,   2,   0 },    191.20347,   0.00003,     0.076 },
 { {   0,  16, -13,  -2,   0,   0,   0,   0,  -2,   1,   0 },     25.24804,   0.00001,     2.159 },
 { {   0,  16, -13,   0,  -3,   0,   0,   0,  -2,   1,   0 },    356.52877,   0.00001,     0.785 },
 { {   0,  16, -13,   0,  -2,   0,   0,   0,  -1,   1,  -1 },    283.25596,   0.00002,     3.287 },
 { {   0,  16, -13,   0,  -1,   0,   0,   0,  -2,   1,   0 },     26.25312,   0.00001,     0.693 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },    309.44497,   0.00005,     2.115 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    221.41211,   0.00003,     0.074 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    247.53290,   0.00014,     4.042 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,   0,  -1 },    129.46189,   0.00002,     2.415 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    235.78290,   0.00006,     0.077 },
 { {   0,  16, -13,   0,   1,   0,   0,   0,   0,  -1,   0 },     30.26029,   0.00003,     6.132 },
 { {   0,  16, -13,   0,   2,   0,   0,   0,   0,  -1,   0 },    167.60586,   0.00002,    12.695 },
 { {   0,  16, -12,  -5,   0,   0,   0,   0,  -2,   1,   0 },    353.11110,   0.00003,     7.588 },
 { {   0,  16, -12,  -4,   0,   0,   0,   0,  -2,   1,   0 },     10.21447,   0.00003,     2.501 },
 { {   0,  16, -12,   0,  -5,   0,   0,   0,   0,   1,  -2 },    220.55911,   0.00003,   413.281 },
 { {   0,  16, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    308.13223,   0.00001,     0.080 },
 { {   0,  16, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    232.01001,   0.00002,     1.329 },
 { {   0,  16, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    308.13223,   0.00001,     0.071 },
 { {   0,  16, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    120.66421,   0.00001,    46.727 },
 { {   0,  16,  -8,  -9,   0,   0,   0,   0,   0,  -1,   0 },    124.59136,   0.00003,    30.730 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },     44.63381,   0.00004,     0.050 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    228.34527,   0.00001,     0.116 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,   0,   1,   0 },    228.34527,   0.00001,     0.056 },
 { {   0,  17, -23,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.39368,   0.00003,     0.050 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -4,   0,   0 },    324.04641,   0.00003,     0.023 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    218.65406,   0.00002,     0.052 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    297.11389,   0.00004,     0.135 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,   0,  -2,   0 },    117.61353,   0.00002,     0.048 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    187.10177,   0.00002,     0.178 },
 { {   0,  17, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },    230.47395,   0.00001,     0.053 },
 { {   0,  17, -22,   0,   1,   0,   0,   0,  -2,   2,   0 },     50.47395,   0.00001,     0.133 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   0,   0 },     39.34961,   0.00004,     0.075 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   1,   0 },     47.78167,   0.00081,  1463.573 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   2,   0 },     39.34961,   0.00004,     0.075 },
 { {   0,  17, -21,   0,  -1,   0,   0,   0,  -2,   0,   0 },     60.23378,   0.00001,     0.055 },
 { {   0,  17, -21,   0,  -1,   0,   0,   0,  -2,   2,   0 },    240.23378,   0.00001,     0.120 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,  -2,  -1,   0 },    231.57355,   0.00003,     0.032 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },     53.46299,   0.00002,     0.206 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },     51.57355,   0.00002,     0.055 },
 { {   0,  17, -20,   0,   0,   0,   0,   0,   0,   0,   0 },     73.54684,   0.00001,     0.131 },
 { {   0,  17, -20,   7,   0,   0,   0,   0,  -2,   1,   0 },    166.89097,   0.00001,     7.933 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    333.85507,   0.00004,     0.062 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     12.91628,   0.00003,     0.351 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },      8.50331,   0.00002,     0.096 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.12348,   0.00002,     0.216 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    349.89168,   0.00010,     0.116 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   0,   1,   0 },    176.02236,   0.00001,     0.046 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.05332,   0.00001,     0.132 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    176.44561,   0.00010,     0.030 },
 { {   0,  17, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.82635,   0.00002,     0.062 },
 { {   0,  17, -19,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.05050,   0.00001,     0.066 },
 { {   0,  17, -19,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.05050,   0.00001,     0.087 },
 { {   0,  17, -19,   6,   0,   0,   0,   0,  -2,   1,   0 },    280.26759,   0.00001,     2.922 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -4,   3,   0 },    276.83429,   0.00003,    13.546 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.60861,   0.00004,     0.066 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    272.20044,   0.00006,     0.541 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     90.21631,   0.00003,     0.088 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.89868,   0.00002,     0.059 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.76412,   0.00003,     0.276 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    272.16953,   0.00022,     0.104 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,   1,   0 },     93.55382,   0.00003,     0.044 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },     73.04001,   0.00002,     0.117 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,  -2,   0 },    272.29669,   0.00001,     0.127 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.03344,   0.00012,     0.047 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   0,   0 },     92.16562,   0.00019,     0.029 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   1,   0 },     92.23795,   0.00002,     0.021 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   4,  -1,   0 },     92.29666,   0.00001,     0.022 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   4,   0,   0 },     92.29664,   0.00003,     0.017 },
 { {   0,  17, -18,   0,   0,   2,   0,   0,  -4,   3,   0 },    251.91602,   0.00001,   168.646 },
 { {   0,  17, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    191.03605,   0.00003,     0.067 },
 { {   0,  17, -18,   0,   1,   0,   0,   0,  -2,   1,   0 },    208.76586,   0.00003,     0.567 },
 { {   0,  17, -18,   0,   2,   0,   0,   0,  -2,   0,   0 },    232.50800,   0.00002,     0.067 },
 { {   0,  17, -18,   3,   0,   0,   0,   0,  -2,  -1,   2 },     56.18679,   0.00001,    12.388 },
 { {   0,  17, -17,   0,  -1,   0,   0,   0,  -2,   0,   0 },    199.69564,   0.00006,     0.070 },
 { {   0,  17, -17,   0,  -1,   0,   0,   0,  -2,   1,   0 },    231.28288,   0.00002,     1.073 },
 { {   0,  17, -17,   0,   0,  -1,   0,   0,  -2,   0,   0 },     92.13148,   0.00001,     0.071 },
 { {   0,  17, -17,   0,   0,  -1,   0,   0,  -2,   1,   0 },     92.02303,   0.00002,     1.134 },
 { {   0,  17, -17,   0,   0,  -1,   0,   0,  -2,   2,   0 },    272.13148,   0.00001,     0.081 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.04474,   0.00001,     0.039 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -4,   2,   0 },    179.82054,   0.00002,     0.081 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -3,   2,   0 },      0.24149,   0.00033,    25.533 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -3,   3,   0 },      3.76342,   0.00002,     0.075 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,  -2,   0 },    359.88425,   0.00001,     0.025 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.93243,   0.00004,     0.037 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.71640,   0.00056,     0.071 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.04982,   0.00090,     1.179 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.43394,   0.00003,     0.081 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.63979,   0.00006,     0.576 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.58081,   0.00002,     0.526 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.62082,   0.00003,     0.063 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.94007,   0.00025,     0.381 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    180.89429,   0.00065,     0.094 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   1,   0 },    180.00123,   0.00011,     0.042 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.18018,   0.00002,     0.104 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.71579,   0.00024,     0.045 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   0,  -2 },    179.82221,   0.00001,     0.117 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    359.82754,   0.00045,     0.028 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   1,   0 },    359.82262,   0.00010,     0.021 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   3,   0,   0 },    359.89479,   0.00002,     0.021 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   4,  -2,   0 },    359.80181,   0.00002,     0.030 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   4,   0,   0 },    359.82059,   0.00004,     0.017 },
 { {   0,  17, -17,   0,   0,   1,   0,   0,  -2,   0,   0 },     87.32607,   0.00002,     0.071 },
 { {   0,  17, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },    347.22434,   0.00006,     0.071 },
 { {   0,  17, -17,   0,   1,   0,   0,   0,  -2,   1,   0 },     76.04659,   0.00001,     1.310 },
 { {   0,  17, -17,   0,   1,   0,   0,   0,  -2,   2,   0 },    347.80298,   0.00003,     0.080 },
 { {   0,  17, -17,   0,   2,   0,   0,   0,  -2,   1,   0 },    162.84261,   0.00001,     1.472 },
 { {   0,  17, -17,   2,   0,   0,   0,   0,  -2,   1,   0 },    350.82716,   0.00003,     4.640 },
 { {   0,  17, -17,   3,   0,   0,   0,   0,  -1,  -1,   1 },    337.90420,   0.00002,    37.096 },
 { {   0,  17, -16,  -3,   0,   0,   0,   0,  -2,   1,   0 },    304.06447,   0.00003,     0.693 },
 { {   0,  17, -16,  -2,   0,   0,   0,   0,  -2,   0,   0 },    122.13617,   0.00002,     0.071 },
 { {   0,  17, -16,  -2,   0,   0,   0,   0,  -2,   1,   0 },    109.90873,   0.00003,     1.097 },
 { {   0,  17, -16,  -2,   0,   0,   0,   0,  -2,   2,   0 },    122.13617,   0.00002,     0.081 },
 { {   0,  17, -16,   0,  -4,   0,   0,   0,  -2,   1,   0 },    139.89845,   0.00002,     5.404 },
 { {   0,  17, -16,   0,  -3,   0,   0,   0,  -2,   1,   0 },    130.19719,   0.00003,     9.925 },
 { {   0,  17, -16,   0,  -2,   0,   0,   0,  -2,   0,   0 },     37.68260,   0.00003,     0.075 },
 { {   0,  17, -16,   0,  -2,   0,   0,   0,  -2,   1,   0 },    351.85658,   0.00001,    60.783 },
 { {   0,  17, -16,   0,  -2,   0,   0,   0,  -2,   2,   0 },     91.57827,   0.00001,     0.076 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },     63.32530,   0.00003,     0.076 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   1,   0 },     16.89532,   0.00005,    14.738 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   2,   0 },    243.32530,   0.00003,     0.075 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,   0,  -1,   0 },    181.37809,   0.00003,     0.586 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.27327,   0.00005,     0.087 },
 { {   0,  17, -16,   0,   0,  -5,   0,   0,  -2,   1,   0 },     12.59878,   0.00001,    56.870 },
 { {   0,  17, -16,   0,   0,  -4,   0,   0,  -2,   1,   0 },    279.56829,   0.00002,    61.111 },
 { {   0,  17, -16,   0,   0,  -3,   0,   0,  -2,   1,   0 },    194.86660,   0.00004,    19.876 },
 { {   0,  17, -16,   0,   0,  -2,   0,   0,  -2,   0,   0 },    273.80064,   0.00001,     0.076 },
 { {   0,  17, -16,   0,   0,  -2,   0,   0,  -2,   1,   0 },    108.06230,   0.00008,    11.868 },
 { {   0,  17, -16,   0,   0,  -2,   0,   0,  -2,   2,   0 },     93.80064,   0.00001,     0.075 },
 { {   0,  17, -16,   0,   0,  -1,   0,   0,  -2,   1,   0 },    171.92737,   0.00003,     8.460 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },     96.97758,   0.00001,     0.041 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    282.37814,   0.00007,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   0,   1 },    102.98498,   0.00003,     3.136 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,  -1 },    283.02569,   0.00005,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },     96.97758,   0.00128,     6.572 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,  -1 },    283.04847,   0.00074,    68.493 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },     95.75301,   0.00012,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   3,  -1 },    283.02569,   0.00005,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -1,   1 },    282.97276,   0.00008,     1.760 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },    102.90380,   0.00016,     0.071 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,   0 },    275.00113,   0.00001,     1.361 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.01642,   0.00006,     0.079 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.04267,   0.00020,     1.109 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },    103.04344,   0.00001,     0.081 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.08676,   0.00001,     0.559 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    154.42321,   0.00001,     0.086 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   0,   1,   0 },     96.97758,   0.00001,     0.040 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.49999,   0.00004,     0.374 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.14346,   0.00008,     0.095 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,   0,   1 },    283.11556,   0.00002,     0.027 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   2,   0,   0 },     94.51898,   0.00007,     0.027 },
 { {   0,  17, -16,   0,   0,   1,   0,   0,  -2,   1,   0 },    352.43675,   0.00003,     5.373 },
 { {   0,  17, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    240.45729,   0.00001,     0.077 },
 { {   0,  17, -16,   0,   1,   0,   0,   0,  -2,   1,   0 },    306.85875,   0.00002,     4.229 },
 { {   0,  17, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },    240.45729,   0.00001,     0.074 },
 { {   0,  17, -16,   3,   0,   0,   0,   0,   0,  -1,   0 },    242.03602,   0.00005,    37.303 },
 { {   0,  17, -15,  -3,   0,   0,   0,   0,  -2,   0,   0 },    328.79301,   0.00002,     0.073 },
 { {   0,  17, -15,  -3,   0,   0,   0,   0,  -2,   1,   0 },     67.66396,   0.00002,     2.258 },
 { {   0,  17, -15,  -3,   0,   0,   0,   0,  -2,   2,   0 },    328.79301,   0.00002,     0.078 },
 { {   0,  17, -15,  -2,   0,   0,   0,   0,  -2,   0,   0 },      2.88709,   0.00004,     0.076 },
 { {   0,  17, -15,  -2,   0,   0,   0,   0,  -2,   1,   0 },    345.36011,   0.00001,    11.263 },
 { {   0,  17, -15,  -2,   0,   0,   0,   0,  -2,   2,   0 },      2.88709,   0.00004,     0.075 },
 { {   0,  17, -15,   0,  -3,   0,   0,   0,  -2,   0,   0 },    189.81997,   0.00001,     0.081 },
 { {   0,  17, -15,   0,  -3,   0,   0,   0,  -2,   2,   0 },      9.81997,   0.00001,     0.071 },
 { {   0,  17, -15,   0,  -3,   0,   0,   0,  -1,   0,   0 },     39.33563,   0.00001,    81.736 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,  -2,   0,   0 },    180.15128,   0.00002,     0.081 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,  -2,   1,   0 },      0.21729,   0.00002,     1.017 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,  -2,   2,   0 },    180.15128,   0.00002,     0.070 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,  -1,   0,   0 },    180.50977,   0.00002,    10.359 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,   0,  -2,   0 },      0.84694,   0.00001,     0.071 },
 { {   0,  17, -15,   0,  -2,   0,   0,   0,   0,   0,   0 },      0.30874,   0.00002,     0.080 },
 { {   0,  17, -15,   0,  -1,   0,   0,   0,  -1,   1,  -1 },    112.74040,   0.00002,    70.990 },
 { {   0,  17, -15,   0,   0,  -3,   0,   0,  -1,   1,  -1 },     96.92756,   0.00003,   289.567 },
 { {   0,  17, -15,   0,   0,  -2,   0,   0,  -2,   1,   0 },    179.83675,   0.00001,     0.922 },
 { {   0,  17, -15,   0,   0,  -2,   0,   0,  -1,   1,  -1 },     15.84109,   0.00005,    32.793 },
 { {   0,  17, -15,   0,   0,  -1,   0,   0,  -1,   1,  -1 },    102.90825,   0.00003,    15.518 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    204.96485,   0.00016,     0.083 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.15634,   0.00009,     0.868 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     18.02707,   0.00002,     0.069 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,  -1,   0 },    206.35980,   0.00002,     0.077 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    320.29031,   0.00002,     0.076 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.35980,   0.00042,     3.772 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,  -1 },    320.29031,   0.00030,    10.164 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,   0 },    206.35980,   0.00002,     0.074 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   2,  -1 },    320.29031,   0.00002,     0.075 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },     30.36622,   0.00006,     0.072 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.70587,   0.00067,     1.608 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     27.50277,   0.00047,     0.079 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     27.90664,   0.00003,     0.039 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    205.69113,   0.00001,     0.041 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    205.69264,   0.00003,     0.027 },
 { {   0,  17, -14,  -5,   0,   0,   0,   0,  -2,   1,   0 },     74.43934,   0.00001,     1.975 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -4,   2,   0 },    150.36572,   0.00001,     0.087 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -2,   0,   0 },    330.36572,   0.00005,     0.076 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -2,   1,   0 },    159.43954,   0.00003,    39.340 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -2,   2,   0 },    330.36572,   0.00005,     0.075 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,   0,   0,   0 },    150.36572,   0.00001,     0.087 },
 { {   0,  17, -14,  -2,   0,   0,   0,   0,  -1,   1,  -1 },    248.82296,   0.00002,    28.554 },
 { {   0,  17, -14,   0,  -5,   0,   0,   0,   0,  -1,   0 },    184.37019,   0.00003,    23.051 },
 { {   0,  17, -14,   0,  -4,   0,   0,   0,   0,  -2,   0 },      7.58347,   0.00001,     0.076 },
 { {   0,  17, -14,   0,  -4,   0,   0,   0,   0,  -1,   0 },    174.16510,   0.00003,    24.437 },
 { {   0,  17, -14,   0,  -4,   0,   0,   0,   0,   0,   0 },    187.58347,   0.00001,     0.075 },
 { {   0,  17, -14,   0,  -3,   0,   0,   0,   0,  -2,   0 },    178.06081,   0.00001,     0.076 },
 { {   0,  17, -14,   0,  -3,   0,   0,   0,   0,  -1,   0 },    245.41101,   0.00003,     7.986 },
 { {   0,  17, -14,   0,  -3,   0,   0,   0,   0,   0,   0 },    177.38720,   0.00004,     0.075 },
 { {   0,  17, -14,   0,  -2,   0,   0,   0,   0,  -1,   0 },    286.15525,   0.00006,     4.773 },
 { {   0,  17, -14,   0,  -1,   0,   0,   0,   0,  -1,   0 },    244.68342,   0.00001,     3.403 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    250.21232,   0.00004,     2.645 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,   0,  -1 },    130.46390,   0.00005,     4.731 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     63.14714,   0.00002,     0.073 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,   1,  -2 },    205.54534,   0.00003,    22.409 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   1,  -1,  -1 },    310.30248,   0.00002,     1.480 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    219.71345,   0.00002,     0.084 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    219.71345,   0.00002,     0.026 },
 { {   0,  17, -13,  -6,   0,   0,   0,   0,  -2,   1,   0 },    256.74918,   0.00005,    26.352 },
 { {   0,  17, -13,  -4,   0,   0,   0,   0,  -1,   1,  -1 },     64.79636,   0.00002,    35.279 },
 { {   0,  17, -13,   0,  -4,   0,   0,   0,   1,  -1,  -1 },    285.40276,   0.00003,    77.860 },
 { {   0,  17, -12,  -7,   0,   0,   0,   0,  -2,   1,   0 },    302.83974,   0.00002,     2.324 },
 { {   0,  17, -12,   0,  -3,   0,   0,   0,   2,  -1,  -2 },    253.64266,   0.00002,    56.514 },
 { {   0,  17,  -8, -12,   0,   0,   0,   0,   0,  -1,   0 },    335.14470,   0.00008,   482.725 },
 { {   0,  17,  -5, -19,   0,   0,   0,   0,  -1,   1,  -1 },    230.61990,   0.00003,   277.048 },
 { {   0,  18, -27,  17,   0,   0,   0,   0,  -2,   1,   0 },    231.88064,   0.00001,     5.444 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },     48.70841,   0.00002,     0.051 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    228.70841,   0.00002,     0.161 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,  -2,   3,   0 },    228.70841,   0.00002,     0.049 },
 { {   0,  18, -24,   0,   2,   0,   0,   0,  -2,   0,   0 },    231.36154,   0.00002,     0.052 },
 { {   0,  18, -24,  15,   0,   0,   0,   0,   0,  -1,   0 },     43.50231,   0.00001,    47.453 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -4,   0,   0 },    324.03758,   0.00002,     0.023 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    216.79848,   0.00003,     0.054 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    287.20323,   0.00001,     0.191 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    235.92588,   0.00004,     0.124 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,   0,  -2,   0 },     55.52144,   0.00001,     0.049 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    159.49298,   0.00003,     0.160 },
 { {   0,  18, -23,   9,   0,   0,   0,   0,  -2,   1,   0 },    296.25104,   0.00001,     2.288 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,  -1,   0 },    231.54030,   0.00002,     0.033 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    237.47322,   0.00001,     0.057 },
 { {   0,  18, -22,   8,   0,   0,   0,   0,  -2,   1,   0 },    127.21332,   0.00005,    32.099 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,  -2,  -2,   0 },    325.76918,   0.00001,     0.023 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    325.76918,   0.00001,     0.061 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    145.76918,   0.00002,     0.310 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,   0,   0,   0 },     74.88243,   0.00001,     0.121 },
 { {   0,  18, -21,   6,   0,   0,   0,   0,  -2,   1,   0 },     75.59869,   0.00006,    31.042 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    348.50527,   0.00004,     0.065 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },     13.95289,   0.00001,     0.450 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    352.31747,   0.00003,     0.091 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.64683,   0.00002,     0.250 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    352.37002,   0.00011,     0.108 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   0,   2,   0 },    356.96069,   0.00001,     0.028 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   2,   0,   0 },    176.83213,   0.00009,     0.029 },
 { {   0,  18, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    270.84716,   0.00001,     0.065 },
 { {   0,  18, -20,   0,   1,   0,   0,   0,  -2,   1,   0 },    270.84716,   0.00001,     0.468 },
 { {   0,  18, -20,   0,   1,   0,   0,   0,  -2,   2,   0 },    270.84716,   0.00001,     0.090 },
 { {   0,  18, -20,   2,   0,   0,   0,   0,  -2,   0,   0 },    177.50139,   0.00002,     0.069 },
 { {   0,  18, -20,   2,   0,   0,   0,   0,  -2,   1,   0 },    178.08269,   0.00003,     0.863 },
 { {   0,  18, -20,   2,   0,   0,   0,   0,  -2,   2,   0 },    177.50139,   0.00002,     0.083 },
 { {   0,  18, -20,   3,   0,   0,   0,   0,  -2,   1,   0 },    330.38235,   0.00003,     1.594 },
 { {   0,  18, -20,   4,   0,   0,   0,   0,  -2,   0,   0 },     57.03608,   0.00001,     0.075 },
 { {   0,  18, -20,   4,   0,   0,   0,   0,  -2,   1,   0 },    100.51436,   0.00003,    10.462 },
 { {   0,  18, -20,   4,   0,   0,   0,   0,  -2,   2,   0 },     57.03608,   0.00001,     0.076 },
 { {   0,  18, -19,   0,  -1,   0,   0,   0,  -2,   1,   0 },    344.34795,   0.00002,     0.765 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -5,   0,   2 },    270.03812,   0.00002,     0.040 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -5,   1,   2 },    270.03812,   0.00002,     0.087 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -4,   2,   0 },    271.78864,   0.00001,     0.079 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,  -1,   2 },    270.03812,   0.00009,     0.075 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   0,   2 },    270.03812,   0.00158,   556.277 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   1,   2 },    270.03812,   0.00009,     0.075 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.29777,   0.00006,     0.069 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    271.23290,   0.00011,     0.818 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.79007,   0.00007,     0.083 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,  -1,   2 },    270.03812,   0.00002,     0.087 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,   0,   2 },    270.03812,   0.00002,     0.040 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.92644,   0.00002,     0.062 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    273.17001,   0.00003,     0.334 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    271.66776,   0.00016,     0.097 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,   1,   0 },     93.69443,   0.00002,     0.043 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   1,   0,  -1 },     67.74931,   0.00002,     0.109 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,  -2,   0 },    314.32280,   0.00001,     0.118 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },    284.05536,   0.00010,     0.046 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     92.15793,   0.00015,     0.029 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   1,   0 },     92.23098,   0.00002,     0.021 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   4,  -1,   0 },     92.28782,   0.00001,     0.022 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   4,   0,   0 },     92.28781,   0.00002,     0.017 },
 { {   0,  18, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    209.72865,   0.00006,     0.069 },
 { {   0,  18, -19,   0,   1,   0,   0,   0,  -2,   1,   0 },    132.85497,   0.00004,     0.879 },
 { {   0,  18, -19,   2,   0,   0,   0,   0,  -2,   0,   0 },     38.55896,   0.00002,     0.075 },
 { {   0,  18, -19,   2,   0,   0,   0,   0,  -2,   1,   0 },    118.05057,   0.00001,     6.291 },
 { {   0,  18, -19,   2,   0,   0,   0,   0,  -2,   2,   0 },     38.55896,   0.00002,     0.076 },
 { {   0,  18, -19,   3,   0,   0,   0,   0,  -2,   0,   0 },     56.14540,   0.00003,     0.078 },
 { {   0,  18, -19,   3,   0,   0,   0,   0,  -2,   1,   0 },     56.22184,   0.00002,     2.683 },
 { {   0,  18, -19,   3,   0,   0,   0,   0,  -2,   2,   0 },     56.14540,   0.00003,     0.073 },
 { {   0,  18, -18,   0,  -2,   0,   0,   0,  -2,   1,   0 },    193.77666,   0.00004,     2.558 },
 { {   0,  18, -18,   0,  -1,   0,   0,   0,  -2,  -1,   2 },     92.64063,   0.00001,    37.190 },
 { {   0,  18, -18,   0,  -1,   0,   0,   0,  -2,   0,   0 },    232.47555,   0.00005,     0.074 },
 { {   0,  18, -18,   0,  -1,   0,   0,   0,  -2,   1,   0 },    265.15180,   0.00001,     3.261 },
 { {   0,  18, -18,   0,   0,  -1,   0,   0,  -2,   1,   0 },    271.32056,   0.00003,     3.902 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   0,   0 },    179.88810,   0.00001,     0.026 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   1,   0 },    179.95924,   0.00013,     0.040 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   2,   0 },    180.29310,   0.00010,     0.085 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,  -1,   0 },    176.49298,   0.00016,     0.037 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,  -1,   1 },    195.18854,   0.00001,     0.075 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,  -1,   2 },    359.90190,   0.00012,     8.993 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    178.35056,   0.00233,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   1 },    195.18854,   0.00026,    17.998 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,  -2 },    359.88810,   0.00002,     0.037 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.88826,   0.00933,     4.498 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   1 },    195.18854,   0.00001,     0.076 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   2 },    359.88810,   0.00002,     0.038 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.54313,   0.00055,     0.077 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   3,   0 },    179.54410,   0.00004,     0.038 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -1,   0,  -1 },    210.76857,   0.00002,     0.069 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -1,   0,   0 },    359.51226,   0.00010,     0.901 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.55291,   0.00003,     0.784 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },     41.00259,   0.00005,     0.066 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.77521,   0.00027,     0.501 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    189.71881,   0.00047,     0.089 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    183.27328,   0.00018,     0.041 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   2,   0 },    179.87006,   0.00002,     0.026 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.24582,   0.00001,     0.098 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -2,   0 },    230.11958,   0.00026,     0.105 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    221.64817,   0.00066,     0.044 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    348.19350,   0.00031,     0.028 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   1,   0 },    354.14783,   0.00007,     0.020 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   3,   0 },    359.78889,   0.00001,     0.013 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   3,   0,   0 },    359.86268,   0.00002,     0.021 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,  -3,   0 },    232.85604,   0.00001,     0.048 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,  -2,   0 },    233.87975,   0.00003,     0.029 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,  -1,   0 },    237.67122,   0.00002,     0.021 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,   0,   0 },    359.78572,   0.00003,     0.016 },
 { {   0,  18, -18,   0,   0,   1,   0,   0,  -2,   0,   0 },    266.90274,   0.00001,     0.074 },
 { {   0,  18, -18,   0,   0,   1,   0,   0,  -2,   2,   0 },     86.90274,   0.00001,     0.077 },
 { {   0,  18, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },     66.83402,   0.00001,     0.075 },
 { {   0,  18, -18,   0,   1,   0,   0,   0,  -2,   1,   0 },     32.63862,   0.00003,     7.246 },
 { {   0,  18, -18,   0,   1,   0,   0,   0,  -2,   2,   0 },     66.83402,   0.00001,     0.076 },
 { {   0,  18, -18,   0,   2,   0,   0,   0,  -2,   0,   0 },      9.68971,   0.00002,     0.075 },
 { {   0,  18, -18,   0,   2,   0,   0,   0,  -2,   1,   0 },     58.64426,   0.00005,    18.618 },
 { {   0,  18, -18,   0,   2,   0,   0,   0,  -2,   2,   0 },    189.68971,   0.00002,     0.076 },
 { {   0,  18, -18,   0,   3,   0,   0,   0,  -2,   1,   0 },     81.10356,   0.00004,    32.688 },
 { {   0,  18, -17,  -3,   0,   0,   0,   0,  -2,   0,   0 },    307.33765,   0.00002,     0.071 },
 { {   0,  18, -17,  -3,   0,   0,   0,   0,  -2,   2,   0 },    307.33765,   0.00002,     0.080 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,  -1,   2 },    300.97597,   0.00004,    20.910 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,   0,   0 },    109.59090,   0.00003,     0.074 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,   1,   0 },    247.34839,   0.00002,     3.500 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,   2,   0 },    109.59090,   0.00003,     0.077 },
 { {   0,  18, -17,   0,  -3,   0,   0,   0,  -2,   1,   0 },    120.81126,   0.00002,     1.906 },
 { {   0,  18, -17,   0,  -2,   0,   0,   0,  -2,   0,   0 },    278.46117,   0.00001,     0.079 },
 { {   0,  18, -17,   0,  -2,   0,   0,   0,  -2,   1,   0 },    109.59462,   0.00003,     1.642 },
 { {   0,  18, -17,   0,  -2,   0,   0,   0,  -2,   2,   0 },    278.46117,   0.00001,     0.072 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,  -2,   0,   0 },    358.22361,   0.00002,     0.080 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,  -2,   1,   0 },    179.19090,   0.00006,     1.442 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,  -2,   2,   0 },    358.22361,   0.00002,     0.072 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,  -1,  -1,   1 },    148.19089,   0.00002,    37.208 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,   0,   0,   0 },    181.28720,   0.00008,     0.082 },
 { {   0,  18, -17,   0,   0,  -2,   0,   0,  -1,  -1,   1 },    296.74053,   0.00005,    95.528 },
 { {   0,  18, -17,   0,   0,  -1,   0,   0,  -2,   1,   0 },    179.86798,   0.00001,     1.345 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,  -1,   0 },    309.47954,   0.00004,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,   0,   0 },    309.47954,   0.00007,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,   1,   0 },    309.47954,   0.00002,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,  -1,   0 },    129.47954,   0.00002,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,  -1,   1 },    283.01956,   0.00006,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   0,   0 },    129.47954,   0.00002,     0.040 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   0,   1 },    282.97678,   0.00012,     0.088 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   1,  -1 },    102.94981,   0.00002,     0.040 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   2,  -1 },    102.94987,   0.00002,     0.085 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -3,   0 },    309.47954,   0.00004,     0.019 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -2,   0 },    309.47954,   0.00030,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    309.47938,   0.00169,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    310.91251,   0.00101,     0.080 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    108.41834,   0.00004,     1.286 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -3,   1 },    283.01931,   0.00002,     0.038 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   0 },    129.47954,   0.00009,     0.038 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   1 },    283.01928,   0.00031,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    103.01126,   0.00004,     0.037 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   0 },    130.43148,   0.00060,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   1 },    283.01956,   0.00547,    17.413 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   3 },    103.01956,   0.00001,     0.037 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    102.99765,   0.00047,     0.074 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },    286.16330,   0.00046,     9.147 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   1 },    284.76810,   0.00001,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,  -1 },    102.94981,   0.00152,     3.622 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,   0 },    293.69413,   0.00004,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },    102.94978,   0.00008,     0.077 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -4,   0 },    309.47954,   0.00006,     0.025 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -3,   0 },    309.48016,   0.00053,     0.036 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,  -2 },    129.47954,   0.00003,     0.024 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    309.56196,   0.00380,     0.070 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -1,  -2 },    129.47954,   0.00003,     0.036 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    292.42425,   0.00004,     1.004 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.71312,   0.00002,     0.860 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    129.68695,   0.00478,     0.082 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   2 },    309.47954,   0.00003,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   1,   0 },    129.48252,   0.00063,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   2,   0 },    129.47954,   0.00007,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -2,   1 },    283.01986,   0.00006,     0.087 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.50577,   0.00005,     0.488 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,   0 },    309.47954,   0.00002,     0.088 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.01956,   0.00006,     0.040 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.16243,   0.00004,     0.089 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   0,   1 },    283.08247,   0.00001,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   1,  -1 },    102.94981,   0.00002,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -4,   0 },    309.47954,   0.00002,     0.062 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    129.48008,   0.00704,     0.095 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   2 },    309.47954,   0.00004,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    129.41497,   0.01137,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   2 },    309.47954,   0.00005,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,  -2 },    309.47954,   0.00001,     0.099 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    128.24801,   0.00152,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   2 },    309.47954,   0.00001,     0.016 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   1,   0 },    129.47954,   0.00016,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   2,   0 },    129.47954,   0.00002,     0.016 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   3,  -2,   0 },    309.47954,   0.00002,     0.044 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -3,   0 },    129.47954,   0.00019,     0.046 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -2,   0 },    129.47562,   0.00045,     0.028 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -1,   0 },    129.47954,   0.00026,     0.021 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,   0,   0 },    129.47954,   0.00005,     0.016 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   6,  -3,   0 },    129.47954,   0.00001,     0.021 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   6,  -2,   0 },    129.47954,   0.00001,     0.017 },
 { {   0,  18, -17,   0,   1,   0,   0,   0,  -1,   0,   0 },    222.20396,   0.00008,    39.962 },
 { {   0,  18, -17,   2,   0,   0,   0,   0,   0,  -1,   0 },    239.83112,   0.00004,    14.919 },
 { {   0,  18, -16,  -3,   0,   0,   0,   0,  -2,   0,   0 },    149.78381,   0.00002,     0.076 },
 { {   0,  18, -16,  -3,   0,   0,   0,   0,  -2,   1,   0 },    203.45474,   0.00003,     5.476 },
 { {   0,  18, -16,  -3,   0,   0,   0,   0,  -2,   2,   0 },    149.78381,   0.00002,     0.074 },
 { {   0,  18, -16,  -2,   0,   0,   0,   0,  -1,  -1,   1 },    168.39146,   0.00005,   168.341 },
 { {   0,  18, -16,  -2,   0,   0,   0,   0,   0,  -2,   0 },    179.67538,   0.00001,     0.070 },
 { {   0,  18, -16,  -2,   0,   0,   0,   0,   0,  -1,   0 },    179.62422,   0.00003,     0.944 },
 { {   0,  18, -16,  -2,   0,   0,   0,   0,   0,   0,   0 },    179.67538,   0.00001,     0.082 },
 { {   0,  18, -16,   0,  -3,   0,   0,   0,  -2,   1,   0 },    191.33583,   0.00002,     0.656 },
 { {   0,  18, -16,   0,  -3,   0,   0,   0,   0,  -2,   0 },    189.59894,   0.00002,     0.074 },
 { {   0,  18, -16,   0,  -3,   0,   0,   0,   0,  -1,   0 },     66.83316,   0.00002,     4.012 },
 { {   0,  18, -16,   0,  -3,   0,   0,   0,   0,   0,   0 },      9.59894,   0.00002,     0.077 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,  -2,   0,   0 },    179.96849,   0.00003,     0.086 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,  -2,   2,   0 },    179.96849,   0.00003,     0.067 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,   0,  -1,   0 },    357.07084,   0.00004,     6.063 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,   0,   0,   0 },    168.82504,   0.00003,     0.076 },
 { {   0,  18, -16,   0,  -1,   0,   0,   0,   0,  -1,   0 },     47.03790,   0.00006,    12.401 },
 { {   0,  18, -16,   0,   0,  -2,   0,   0,   0,  -1,   0 },    177.47527,   0.00002,    15.568 },
 { {   0,  18, -16,   0,   0,  -1,   0,   0,   0,  -1,   0 },    135.36282,   0.00004,    33.019 },
 { {   0,  18, -16,   0,   0,   0,  -1,   0,   0,  -1,   0 },    348.75512,   0.00003,   121.368 },
 { {   0,  18, -16,   0,   0,   0,   0,  -1,   0,  -1,   0 },    182.40839,   0.00001,   415.528 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,  -1,   0 },     26.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   0,   0 },     26.54261,   0.00007,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   1,   0 },     26.54261,   0.00008,     0.021 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   2,   0 },     26.54261,   0.00003,     0.029 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -3,   0 },     26.54261,   0.00005,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -2,   0 },     26.54261,   0.00034,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,  -2 },    206.54261,   0.00002,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   0 },     26.54238,   0.00193,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,  -2 },    206.54261,   0.00003,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   0 },     26.54161,   0.00406,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   2 },     26.54261,   0.00001,     0.107 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,  -2 },    206.54261,   0.00002,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },     26.53992,   0.00224,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   2,   0 },     26.54261,   0.00004,     0.103 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,  -1,   0 },     26.54261,   0.00004,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   0,   0 },    206.54261,   0.00022,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   1,   0 },    206.54261,   0.00004,     0.094 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -5,   0 },     26.54261,   0.00002,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -4,   0 },     26.54261,   0.00018,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,  -2 },    206.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   0 },     26.54287,   0.00200,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   2 },    206.54261,   0.00003,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,  -2 },    206.54261,   0.00017,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },     26.54274,   0.01999,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   2 },    206.54261,   0.00022,     0.090 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    206.54261,   0.00079,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },     26.54223,   0.16440,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   2 },     26.52638,   0.00004,     0.474 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -2 },    206.54224,   0.00098,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     26.53985,   0.16710,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,  -2 },    206.54261,   0.00004,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },     26.58338,   0.00081,     0.563 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   2 },     26.54261,   0.00004,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    206.46492,   0.00041,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   3,   0 },    206.54261,   0.00007,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -3,   0 },    206.54261,   0.00006,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -2,   0 },    206.54205,   0.00059,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,  -2 },     26.54261,   0.00003,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },    206.54261,   0.00432,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   2 },     26.54261,   0.00002,     0.069 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,   0,   0 },    206.61826,   0.00006,     1.123 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },     26.54261,   0.00006,     0.071 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -6,   0 },     26.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -5,   0 },     26.54155,   0.00028,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,  -2 },    206.54261,   0.00006,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,   0 },     26.54324,   0.00377,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,  -2 },    206.54207,   0.00056,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   0 },     26.54324,   0.05354,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -4 },     26.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -2 },    206.54294,   0.00465,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },     26.54324,   0.78811,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   2 },    206.54233,   0.00132,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -4 },     26.54261,   0.00006,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -2 },    206.54280,   0.02810,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.54261,  14.24883,   273.043 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   2 },    206.54280,   0.02810,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   4 },     26.54261,   0.00006,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -2 },    206.54023,   0.00141,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },    147.61395,   0.00001,     6.132 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     26.54560,   0.73930,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   2 },    206.54371,   0.00455,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   4 },     26.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,  -2 },     24.75914,   0.00003,     3.032 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   0 },     26.54560,   0.05022,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   2 },    206.54316,   0.00054,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   0 },     26.54559,   0.00354,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   2 },    206.54261,   0.00006,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   3,   0 },     26.54373,   0.00026,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   4,   0 },     26.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -3,   0 },     26.54261,   0.00006,     0.071 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,  -2 },     26.54261,   0.00002,     0.069 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   0 },    206.54261,   0.00432,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   2 },     26.54261,   0.00003,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    206.54318,   0.00057,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   1,   0 },    206.54261,   0.00006,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -5,   0 },    206.54261,   0.00007,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -4,   0 },    206.54448,   0.00047,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,  -2 },     26.54261,   0.00004,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,   0 },    206.62410,   0.00011,     0.565 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,   2 },    206.54261,   0.00004,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },     26.54227,   0.15711,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   2 },    206.54298,   0.00096,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,  -2 },     26.52638,   0.00004,     0.475 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },     26.54301,   0.16302,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   2 },    206.54261,   0.00079,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,  -2 },    206.54261,   0.00022,     0.090 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   0 },     26.54486,   0.01946,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   2 },    206.54261,   0.00017,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,  -2 },    206.54261,   0.00003,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   0 },     26.54412,   0.00193,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   2 },    206.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   2,   0 },     26.54261,   0.00018,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   3,   0 },     26.54261,   0.00002,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -3,   0 },    206.54261,   0.00004,     0.094 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -2,   0 },    206.54261,   0.00022,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -1,   0 },     26.54261,   0.00004,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -4,   0 },     26.54261,   0.00004,     0.103 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   0 },     26.54196,   0.00211,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   2 },    206.54261,   0.00002,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,  -2 },     26.54261,   0.00001,     0.107 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   0 },     26.54247,   0.00396,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   2 },    206.54261,   0.00003,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },     26.54284,   0.00191,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   2 },    206.54261,   0.00002,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   0,   0 },     26.54261,   0.00034,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   1,   0 },     26.54261,   0.00005,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -4,   0 },     26.54261,   0.00003,     0.029 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -3,   0 },     26.54261,   0.00008,     0.021 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -2,   0 },     26.54261,   0.00007,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -1,   0 },     26.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   1,   0,   0,   0,  -1,   0 },    251.56659,   0.00004,    26.589 },
 { {   0,  18, -16,   0,   1,   0,   0,   0,   0,  -1,   0 },    311.76977,   0.00002,    11.368 },
 { {   0,  18, -16,   2,   0,   0,   0,   0,   1,  -1,  -1 },    353.90966,   0.00003,    75.385 },
 { {   0,  18, -15,  -6,   0,   0,   0,   0,  -2,   1,   0 },    255.03250,   0.00001,     2.425 },
 { {   0,  18, -15,  -5,   0,   0,   0,   0,  -2,   1,   0 },     94.05393,   0.00002,     8.385 },
 { {   0,  18, -15,  -4,   0,   0,   0,   0,  -2,   0,   0 },    329.93569,   0.00002,     0.079 },
 { {   0,  18, -15,  -4,   0,   0,   0,   0,  -2,   2,   0 },    329.93569,   0.00002,     0.072 },
 { {   0,  18, -15,  -2,   0,   0,   0,   0,   0,  -1,   0 },    263.42466,   0.00001,    16.749 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -6,   0,   0 },    283.60568,   0.00001,     0.017 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -6,   1,   0 },    283.60568,   0.00001,     0.021 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -2,   0 },    283.60568,   0.00005,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -1,   0 },    283.60568,   0.00026,     0.021 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   0,   0 },    283.60568,   0.00045,     0.028 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   1,   0 },    283.60568,   0.00019,     0.046 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -3,   0,   0 },    103.60568,   0.00002,     0.044 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -4,   0 },    283.60568,   0.00002,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -3,   0 },    283.60568,   0.00016,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,  -2 },    103.60568,   0.00001,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   0 },    283.60583,   0.00151,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   2 },    103.60568,   0.00001,     0.099 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    103.60568,   0.00005,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },    283.60558,   0.01141,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,  -2 },    103.60568,   0.00004,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    283.60299,   0.00748,     0.095 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    283.55392,   0.00003,     0.360 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    103.60568,   0.00002,     0.062 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    310.09123,   0.00004,     0.040 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -1,  -1,   0 },    103.60568,   0.00002,     0.088 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -4,   0 },    283.60568,   0.00007,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -3,   0 },    283.60631,   0.00067,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,  -2 },    103.60568,   0.00003,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },    283.60631,   0.00513,     0.082 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    283.57068,   0.00003,     0.996 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -1,   2 },    283.60568,   0.00003,     0.036 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    103.60868,   0.00357,     0.070 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   2 },    283.60568,   0.00003,     0.024 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    103.60867,   0.00050,     0.036 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   2,   0 },    103.60568,   0.00006,     0.025 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -3,  -1 },    310.09123,   0.00001,     0.038 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -3,   0 },    103.60568,   0.00001,     0.076 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -2,  -1 },    310.09095,   0.00018,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,  -1 },    310.09123,   0.00332,    19.959 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.60568,   0.00062,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    310.11090,   0.00032,     0.076 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    283.60568,   0.00009,     0.038 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   1,  -1 },    310.09123,   0.00001,     0.038 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    104.31160,   0.00097,     0.080 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    103.60584,   0.00168,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    105.91066,   0.00030,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   1,   0 },    103.60568,   0.00004,     0.019 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -2,  -1 },    310.09150,   0.00004,     0.087 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -2,   0 },    283.60568,   0.00002,     0.040 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -1,  -1 },    310.09123,   0.00004,     0.041 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -1,   0 },    283.60568,   0.00002,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -3,   0 },    103.60568,   0.00002,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -2,   0 },    103.60568,   0.00007,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -1,   0 },    103.60568,   0.00004,     0.020 },
 { {   0,  18, -14,  -7,   0,   0,   0,   0,  -2,   0,   0 },    223.60251,   0.00001,     0.076 },
 { {   0,  18, -14,  -7,   0,   0,   0,   0,  -2,   1,   0 },     79.41596,   0.00002,    17.893 },
 { {   0,  18, -14,  -7,   0,   0,   0,   0,  -2,   2,   0 },    223.60251,   0.00001,     0.075 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -4,  -1,   0 },    180.66875,   0.00002,     0.021 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -4,   0,   0 },    180.66875,   0.00003,     0.029 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.66875,   0.00001,     0.048 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -2,   0 },    180.66875,   0.00008,     0.028 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    180.66875,   0.00056,     0.044 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.66606,   0.00027,     0.106 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   0,  -3,   0 },    180.66875,   0.00001,     0.041 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    180.66875,   0.00009,     0.089 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   0,   0,   0 },      1.35623,   0.00004,     0.066 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -2,  -2 },    308.11189,   0.00001,     0.075 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    344.83075,   0.00012,     0.074 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },      0.66875,   0.00001,     0.037 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,   0,  -2 },    308.11189,   0.00001,     0.076 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    308.07255,   0.00004,     0.025 },
 { {   0,  18, -14,   0,   0,   3,   0,   0,   2,  -1,  -2 },     17.16509,   0.00002,   494.047 },
 { {   0,  18, -13,  -9,   0,   0,   0,   0,  -2,   0,   0 },     45.37674,   0.00002,     0.075 },
 { {   0,  18, -13,  -9,   0,   0,   0,   0,  -2,   1,   0 },    101.32334,   0.00005,   133.722 },
 { {   0,  18, -13,  -9,   0,   0,   0,   0,  -2,   2,   0 },     45.37674,   0.00002,     0.075 },
 { {   0,  18, -13,  -7,   0,   0,   0,   0,  -1,   1,  -1 },     13.37177,   0.00011,   470.696 },
 { {   0,  18, -13,  -5,   0,   0,   0,   0,   0,   1,  -2 },      4.78415,   0.00001,    85.271 },
 { {   0,  18, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },     77.73182,   0.00002,     0.046 },
 { {   0,  18, -12, -11,   0,   0,   0,   0,  -2,   1,   0 },     29.01884,   0.00002,    14.115 },
 { {   0,  18, -10, -15,   0,   0,   0,   0,  -2,   1,   0 },    328.74187,   0.00001,     5.061 },
 { {   0,  18, -10, -13,   0,   0,   0,   0,  -1,   0,   0 },     35.98306,   0.00001,    47.117 },
 { {   0,  18,  -8, -15,   0,   0,   0,   0,   0,  -1,   0 },    160.31219,   0.00002,    35.213 },
 { {   0,  19, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    109.54692,   0.00001,     0.046 },
 { {   0,  19, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    109.54692,   0.00001,     0.215 },
 { {   0,  19, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },     49.03974,   0.00002,     0.053 },
 { {   0,  19, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    229.03974,   0.00002,     0.179 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -4,   0,   0 },    324.02721,   0.00002,     0.023 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -4,   2,   0 },     57.52747,   0.00001,     0.062 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    222.45300,   0.00004,     0.056 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.51153,   0.00005,     0.115 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    129.85920,   0.00002,     0.145 },
 { {   0,  19, -24,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.15432,   0.00002,     0.056 },
 { {   0,  19, -24,   0,   1,   0,   0,   0,  -2,   2,   0 },     50.15432,   0.00002,     0.114 },
 { {   0,  19, -24,   8,   0,   0,   0,   0,  -2,  -1,   2 },     87.68055,   0.00003,   101.947 },
 { {   0,  19, -24,   9,   0,   0,   0,   0,  -2,   1,   0 },    116.20816,   0.00001,     5.309 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -3,  -1,   1 },    313.46714,   0.00004,     0.040 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -3,   0,   1 },    313.46714,   0.00004,     0.087 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -3,   1 },    313.46714,   0.00001,     0.038 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -2,   1 },    313.46714,   0.00021,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -1,   1 },    313.46714,   0.00375,  4846.835 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,   0,   1 },    313.46714,   0.00021,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,   1,   1 },    313.46714,   0.00001,     0.038 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,   1,  -2,   1 },    313.46714,   0.00004,     0.087 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,   1,  -1,   1 },    313.46714,   0.00004,     0.040 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    232.74557,   0.00002,     0.059 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.50490,   0.00003,     0.186 },
 { {   0,  19, -22,   0,  -1,   0,   0,   0,  -2,   0,   0 },     53.11937,   0.00001,     0.063 },
 { {   0,  19, -22,   0,  -1,   0,   0,   0,  -2,   2,   0 },    233.11937,   0.00001,     0.095 },
 { {   0,  19, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    325.95129,   0.00002,     0.063 },
 { {   0,  19, -22,   5,   0,   0,   0,   0,  -2,   1,   0 },    242.77914,   0.00002,    16.228 },
 { {   0,  19, -21,   0,  -2,   0,   0,   0,  -4,   3,   0 },    224.03506,   0.00003,   116.125 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     12.42061,   0.00004,     0.067 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },      9.03660,   0.00003,     0.626 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    357.03422,   0.00002,     0.086 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    351.13897,   0.00002,     0.297 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    353.01293,   0.00010,     0.101 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   2,   0,   0 },    177.17767,   0.00007,     0.029 },
 { {   0,  19, -21,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.38441,   0.00003,     0.073 },
 { {   0,  19, -21,   2,   0,   0,   0,   0,  -2,   2,   0 },    179.38441,   0.00003,     0.079 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   0,   0 },     14.98714,   0.00001,     0.026 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   1,   0 },     14.98714,   0.00009,     0.040 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   2,   0 },     14.57317,   0.00007,     0.087 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,  -1,   0 },     14.98714,   0.00003,     0.038 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   0,   0 },     15.30223,   0.00048,     0.075 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,  -2 },    194.98714,   0.00002,     0.037 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,   0 },     14.98714,   0.00775,   573.318 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,   2 },    194.98714,   0.00002,     0.037 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   2,   0 },     15.30223,   0.00048,     0.075 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   3,   0 },     14.98714,   0.00003,     0.038 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   0,   0 },     14.57317,   0.00007,     0.087 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   1,   0 },     14.98714,   0.00009,     0.040 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   2,   0 },     14.98714,   0.00001,     0.026 },
 { {   0,  19, -21,   4,   0,   0,   0,   0,  -2,   1,   0 },     57.63100,   0.00003,     1.887 },
 { {   0,  19, -21,   5,   0,   0,   0,   0,  -1,   1,  -1 },    171.66387,   0.00002,   127.251 },
 { {   0,  19, -20,   0,  -1,   0,   0,   0,  -2,   0,   0 },    336.80923,   0.00003,     0.072 },
 { {   0,  19, -20,   0,  -1,   0,   0,   0,  -2,   1,   0 },    254.10989,   0.00003,     1.468 },
 { {   0,  19, -20,   0,  -1,   0,   0,   0,  -2,   2,   0 },     16.30361,   0.00002,     0.080 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.86974,   0.00020,     0.072 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.73573,   0.00032,     1.676 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.22679,   0.00001,     0.079 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.83555,   0.00001,     0.674 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },     71.91804,   0.00001,     0.606 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    275.13247,   0.00004,     0.422 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    267.83618,   0.00014,     0.092 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   1,   0 },     95.28488,   0.00002,     0.041 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   2,   0 },    272.24876,   0.00001,     0.027 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },     71.96169,   0.00001,     0.102 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.27745,   0.00008,     0.045 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     91.27001,   0.00011,     0.028 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   4,   0,   0 },     92.27745,   0.00001,     0.017 },
 { {   0,  19, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    131.06144,   0.00007,     0.073 },
 { {   0,  19, -20,   0,   1,   0,   0,   0,  -2,   1,   0 },    141.48762,   0.00003,     1.951 },
 { {   0,  19, -20,   0,   1,   0,   0,   0,   0,   0,   0 },    311.06144,   0.00001,     0.091 },
 { {   0,  19, -20,   0,   2,   0,   0,   0,  -2,   0,   0 },    237.78960,   0.00001,     0.073 },
 { {   0,  19, -20,   0,   2,   0,   0,   0,  -2,   1,   0 },    206.95491,   0.00004,     2.335 },
 { {   0,  19, -20,   0,   2,   0,   0,   0,  -2,   2,   0 },    237.78960,   0.00001,     0.078 },
 { {   0,  19, -20,   1,   0,   0,   0,   0,  -2,   1,   0 },    145.45247,   0.00002,    15.358 },
 { {   0,  19, -20,   3,   0,   0,   0,   0,  -1,   1,  -1 },    274.48467,   0.00005,    18.015 },
 { {   0,  19, -19,  -2,   0,   0,   0,   0,  -2,   0,   0 },     27.64647,   0.00001,     0.072 },
 { {   0,  19, -19,  -2,   0,   0,   0,   0,  -2,   2,   0 },     27.64647,   0.00001,     0.079 },
 { {   0,  19, -19,  -1,   0,   0,   0,   0,  -2,   1,   0 },    301.98610,   0.00002,     7.783 },
 { {   0,  19, -19,   0,  -5,   0,   0,   0,  -2,   1,   0 },    236.63442,   0.00003,    54.589 },
 { {   0,  19, -19,   0,  -4,   0,   0,   0,  -2,   1,   0 },    241.45076,   0.00005,    15.155 },
 { {   0,  19, -19,   0,  -3,   0,   0,   0,  -2,   0,   0 },     36.24752,   0.00001,     0.076 },
 { {   0,  19, -19,   0,  -3,   0,   0,   0,  -2,   1,   0 },    200.56318,   0.00002,     6.654 },
 { {   0,  19, -19,   0,  -3,   0,   0,   0,  -2,   2,   0 },    216.24752,   0.00001,     0.075 },
 { {   0,  19, -19,   0,  -2,   0,   0,   0,  -2,   1,   0 },    216.25336,   0.00004,     4.263 },
 { {   0,  19, -19,   0,  -1,   0,   0,   0,  -2,   0,   0 },     53.61634,   0.00002,     0.077 },
 { {   0,  19, -19,   0,  -1,   0,   0,   0,  -2,   1,   0 },    265.66106,   0.00001,     3.136 },
 { {   0,  19, -19,   0,  -1,   0,   0,   0,  -2,   2,   0 },     53.61634,   0.00002,     0.074 },
 { {   0,  19, -19,   0,  -1,   0,   0,   0,   0,  -1,   0 },    272.79391,   0.00001,     0.687 },
 { {   0,  19, -19,   0,   0,  -2,   0,   0,  -2,   1,   0 },    201.28363,   0.00002,     2.982 },
 { {   0,  19, -19,   0,   0,  -1,   0,   0,  -2,   1,   0 },    272.14754,   0.00004,     2.708 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -4,   1,   0 },    180.12040,   0.00002,     0.041 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -4,   2,   0 },    181.55054,   0.00002,     0.090 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,  -1,   0 },    359.53461,   0.00005,     0.038 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.70702,   0.00072,     0.078 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.12256,   0.00189,     2.480 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.18710,   0.00013,     0.073 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,  -1,   0 },      0.07775,   0.00002,     0.073 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,  -1,   1 },     16.01292,   0.00005,     3.154 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },    211.34903,   0.00003,     0.072 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,   0 },      0.07775,   0.00032,     2.067 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,   1 },     31.34903,   0.00002,     0.077 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.74482,   0.00008,     1.537 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.07775,   0.00002,     0.078 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.95150,   0.00047,     0.730 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    180.68062,   0.00084,     0.084 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   1,   0 },    180.42095,   0.00009,     0.040 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   1,  -1,  -1 },    192.30227,   0.00001,     0.412 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.75346,   0.00001,     0.099 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.75339,   0.00010,     0.043 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    359.75211,   0.00025,     0.027 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   1,   0 },    359.75343,   0.00006,     0.020 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   3,  -2,   0 },    359.83604,   0.00001,     0.044 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   3,   0,   0 },    359.83604,   0.00001,     0.020 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.75342,   0.00001,     0.021 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   4,   0,   0 },    359.74890,   0.00002,     0.016 },
 { {   0,  19, -19,   0,   0,   1,   0,   0,  -2,   0,   0 },    267.26727,   0.00001,     0.078 },
 { {   0,  19, -19,   0,   0,   1,   0,   0,  -2,   1,   0 },    267.04485,   0.00003,     2.288 },
 { {   0,  19, -19,   0,   0,   1,   0,   0,  -2,   2,   0 },    267.26727,   0.00001,     0.073 },
 { {   0,  19, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    163.93499,   0.00003,     0.078 },
 { {   0,  19, -19,   0,   2,   0,   0,   0,  -2,   1,   0 },    162.78382,   0.00001,     1.749 },
 { {   0,  19, -19,   2,   0,   0,   0,   0,   0,  -2,   0 },    330.08866,   0.00001,     0.074 },
 { {   0,  19, -19,   2,   0,   0,   0,   0,   0,   0,   0 },    330.08866,   0.00001,     0.077 },
 { {   0,  19, -18,  -3,   0,   0,   0,   0,  -2,   0,   0 },    103.51920,   0.00002,     0.074 },
 { {   0,  19, -18,  -3,   0,   0,   0,   0,  -2,   1,   0 },    144.74645,   0.00003,     5.213 },
 { {   0,  19, -18,  -3,   0,   0,   0,   0,  -2,   2,   0 },    103.51920,   0.00002,     0.077 },
 { {   0,  19, -18,  -2,   0,   0,   0,   0,  -2,   0,   0 },    329.18483,   0.00002,     0.077 },
 { {   0,  19, -18,  -2,   0,   0,   0,   0,  -2,   2,   0 },    329.18483,   0.00002,     0.074 },
 { {   0,  19, -18,  -1,   0,   0,   0,   0,  -1,   0,   0 },      0.59301,   0.00003,    64.559 },
 { {   0,  19, -18,   0,  -4,   0,   0,   0,  -1,   1,  -1 },    107.47491,   0.00003,    81.835 },
 { {   0,  19, -18,   0,  -2,   0,   0,   0,  -2,   0,   0 },    271.42506,   0.00001,     0.083 },
 { {   0,  19, -18,   0,  -2,   0,   0,   0,  -2,   1,   0 },     93.52359,   0.00003,     0.810 },
 { {   0,  19, -18,   0,  -2,   0,   0,   0,  -2,   2,   0 },     91.42506,   0.00001,     0.069 },
 { {   0,  19, -18,   0,   0,  -1,   0,   0,   0,  -1,   0 },    158.47561,   0.00001,     2.471 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,  -1,   1 },    282.89279,   0.00003,     1.464 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   0,  -1 },    282.97965,   0.00021,     0.077 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   0,   0 },    279.74905,   0.00002,     1.937 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   1,  -1 },    102.97241,   0.00066,     2.862 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   2,  -1 },    102.97192,   0.00004,     0.074 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },     74.56254,   0.00001,     0.073 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    211.21035,   0.00006,     2.697 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,   0,  -1 },    283.02710,   0.00008,     1.860 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    293.80340,   0.00001,     0.078 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.03882,   0.00008,     0.702 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.02454,   0.00016,     0.085 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },     94.34296,   0.00002,     0.041 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   2,   0,   0 },     94.37374,   0.00005,     0.027 },
 { {   0,  19, -18,   0,   1,   0,   0,   0,   0,  -1,   0 },     23.90290,   0.00003,     3.490 },
 { {   0,  19, -17,  -3,   0,   0,   0,   0,  -2,   0,   0 },    156.02931,   0.00001,     0.080 },
 { {   0,  19, -17,  -3,   0,   0,   0,   0,  -2,   2,   0 },    156.02931,   0.00001,     0.071 },
 { {   0,  19, -17,  -2,   0,   0,   0,   0,   0,  -2,   0 },    180.32189,   0.00001,     0.073 },
 { {   0,  19, -17,  -2,   0,   0,   0,   0,   0,   0,   0 },    180.32189,   0.00001,     0.078 },
 { {   0,  19, -17,  -1,   0,   0,   0,   0,   0,  -1,   0 },    303.86228,   0.00001,    10.256 },
 { {   0,  19, -17,   0,  -3,   0,   0,   0,   0,   1,  -2 },    190.20330,   0.00001,    23.387 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     28.93689,   0.00003,     0.092 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    205.20543,   0.00001,     0.064 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.46717,   0.00034,     1.589 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    207.50061,   0.00017,     0.072 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,   1,  -2 },     24.94287,   0.00003,     3.382 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   1,  -1,  -1 },    211.18219,   0.00003,     2.355 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    205.64927,   0.00002,     0.026 },
 { {   0,  19, -16,  -6,   0,   0,   0,   0,  -2,   1,   0 },     45.23537,   0.00003,     4.693 },
 { {   0,  19, -16,  -4,   0,   0,   0,   0,   0,  -1,   0 },    150.54000,   0.00003,     2.010 },
 { {   0,  19, -16,  -3,   0,   0,   0,   0,   0,  -2,   0 },    169.43258,   0.00001,     0.076 },
 { {   0,  19, -16,  -3,   0,   0,   0,   0,   0,  -1,   0 },    311.60767,   0.00006,    29.298 },
 { {   0,  19, -16,  -3,   0,   0,   0,   0,   0,   0,   0 },    349.43258,   0.00001,     0.075 },
 { {   0,  19, -16,   0,   0,   0,   0,   0,   1,  -1,  -1 },    310.14747,   0.00002,     1.738 },
 { {   0,  19, -15,  -5,   0,   0,   0,   0,   0,  -1,   0 },    354.40755,   0.00007,    34.206 },
 { {   0,  19, -15,  -3,   0,   0,   0,   0,   1,  -1,  -1 },    275.49538,   0.00002,    50.937 },
 { {   0,  19, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    308.04590,   0.00003,     0.071 },
 { {   0,  19, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    308.04590,   0.00003,     0.025 },
 { {   0,  19, -14, -10,   0,   0,   0,   0,  -2,   1,   0 },     33.80482,   0.00001,    11.580 },
 { {   0,  19, -14,  -7,   0,   0,   0,   0,   0,  -1,   0 },    123.71166,   0.00002,    10.799 },
 { {   0,  19, -11, -15,   0,   0,   0,   0,  -2,   1,   0 },    336.92477,   0.00001,     2.337 },
 { {   0,  19,  -7, -20,   0,   0,   0,   0,   0,  -1,   0 },    298.01758,   0.00004,   222.268 },
 { {   0,  20, -29,  15,   0,   0,   0,   0,  -2,   0,   0 },     15.07727,   0.00002,     0.075 },
 { {   0,  20, -29,  15,   0,   0,   0,   0,  -2,   1,   0 },    339.17294,   0.00006,   250.641 },
 { {   0,  20, -29,  15,   0,   0,   0,   0,  -2,   2,   0 },     15.07727,   0.00002,     0.075 },
 { {   0,  20, -28,  13,   0,   0,   0,   0,  -2,   1,   0 },    172.87021,   0.00001,    16.841 },
 { {   0,  20, -27,  11,   0,   0,   0,   0,  -2,   1,   0 },    116.70171,   0.00001,     8.147 },
 { {   0,  20, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     49.34291,   0.00002,     0.055 },
 { {   0,  20, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    229.34291,   0.00002,     0.201 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    226.90194,   0.00004,     0.058 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.64000,   0.00005,     0.108 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.01537,   0.00001,     0.174 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    126.55889,   0.00002,     0.133 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,   0,   1,   0 },    144.01537,   0.00001,     0.048 },
 { {   0,  20, -25,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.11869,   0.00003,     0.058 },
 { {   0,  20, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.46737,   0.00002,     0.211 },
 { {   0,  20, -23,   0,   0,   0,   0,   0,   2,   0,   0 },     81.87243,   0.00002,     0.029 },
 { {   0,  20, -23,   2,   0,   0,   0,   0,  -2,   1,   0 },    270.72912,   0.00001,     1.101 },
 { {   0,  20, -23,   3,   0,   0,   0,   0,  -2,   1,   0 },     94.79392,   0.00002,     2.658 },
 { {   0,  20, -23,   7,   0,   0,   0,   0,   0,  -1,   0 },    131.74389,   0.00003,    42.551 },
 { {   0,  20, -23,  11,   0,   0,   0,   0,   2,  -1,  -2 },    244.95399,   0.00001,   232.800 },
 { {   0,  20, -22,   0,  -2,   0,   0,   0,  -2,   1,   0 },     51.52638,   0.00001,     0.877 },
 { {   0,  20, -22,   0,  -1,   0,   0,   0,  -3,   1,   1 },    249.20154,   0.00009,   551.702 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -4,   1,   0 },    194.73920,   0.00001,     0.039 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    321.75317,   0.00003,     1.030 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },     13.43142,   0.00003,     0.081 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    351.66594,   0.00002,     0.364 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   0,   0 },      2.92699,   0.00016,     0.095 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   1,   0 },     14.73920,   0.00009,     0.042 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   2,   0 },     14.73920,   0.00001,     0.027 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   2,  -1,   0 },    157.08893,   0.00001,     0.045 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   2,   0,   0 },    177.48820,   0.00006,     0.028 },
 { {   0,  20, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },    269.80161,   0.00002,     0.071 },
 { {   0,  20, -22,   0,   1,   0,   0,   0,  -2,   1,   0 },    283.76275,   0.00001,     1.127 },
 { {   0,  20, -22,   0,   1,   0,   0,   0,  -2,   2,   0 },     89.80161,   0.00002,     0.081 },
 { {   0,  20, -22,   1,   0,   0,   0,   0,  -2,   1,   0 },    271.28256,   0.00001,     2.275 },
 { {   0,  20, -22,   2,   0,   0,   0,   0,  -2,   1,   0 },     25.57411,   0.00002,    10.859 },
 { {   0,  20, -22,   3,   0,   0,   0,   0,  -2,   1,   0 },      6.78803,   0.00001,     1.603 },
 { {   0,  20, -22,   5,   0,   0,   0,   0,   0,  -1,   0 },    152.74495,   0.00001,    11.512 },
 { {   0,  20, -21,   0,  -3,   0,   0,   0,  -2,   1,   0 },     54.01052,   0.00002,     4.461 },
 { {   0,  20, -21,   0,  -2,   0,   0,   0,  -2,   1,   0 },    283.70883,   0.00004,     7.149 },
 { {   0,  20, -21,   0,  -1,   0,   0,   0,  -2,   0,   0 },    288.28541,   0.00004,     0.075 },
 { {   0,  20, -21,   0,  -1,   0,   0,   0,  -2,   1,   0 },    331.62721,   0.00002,    17.993 },
 { {   0,  20, -21,   0,   0,  -2,   0,   0,  -2,   1,   0 },    229.48748,   0.00005,    25.530 },
 { {   0,  20, -21,   0,   0,  -1,   0,   0,  -2,   0,   0 },    347.79695,   0.00004,     0.075 },
 { {   0,  20, -21,   0,   0,  -1,   0,   0,  -2,   1,   0 },    333.79132,   0.00004,   191.502 },
 { {   0,  20, -21,   0,   0,  -1,   0,   0,  -2,   2,   0 },     78.09149,   0.00001,     0.075 },
 { {   0,  20, -21,   0,   0,   0,  -3,   0,  -2,   1,   0 },    261.38256,   0.00003,   143.271 },
 { {   0,  20, -21,   0,   0,   0,  -2,   0,  -2,   1,   0 },    334.48673,   0.00007,   203.168 },
 { {   0,  20, -21,   0,   0,   0,   0,  -2,  -2,   1,   0 },     98.36599,   0.00003,    60.286 },
 { {   0,  20, -21,   0,   0,   0,   0,  -1,  -2,   1,   0 },    330.07186,   0.00002,    44.137 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -6,   1,   0 },    271.80227,   0.00001,     0.020 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -6,   2,   0 },    271.80227,   0.00003,     0.028 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -6,   3,   0 },    271.80227,   0.00002,     0.044 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,  -1,   0 },    271.80227,   0.00001,     0.020 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   0,   0 },    271.92600,   0.00012,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   1,   0 },    271.80349,   0.00119,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   2,   0 },    271.50332,   0.00114,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -3,   1,   0 },     91.80227,   0.00003,     0.081 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,  -2,   0 },    271.80227,   0.00003,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,  -1,   0 },    272.36187,   0.00022,     0.038 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,  -2 },     91.80227,   0.00003,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.36059,   0.00318,     0.076 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   1 },     82.77305,   0.00001,     5.116 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   2 },     91.80227,   0.00001,     0.073 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,  -2 },     91.80227,   0.00021,     0.037 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    271.80227,   0.10651,    34.812 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   2 },     91.80227,   0.00021,     0.037 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,  -2 },     91.80227,   0.00001,     0.074 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    272.07968,   0.00606,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   2 },     91.80227,   0.00003,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   3,   0 },    272.07972,   0.00041,     0.038 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   4,   0 },    271.80227,   0.00003,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   0,  -1 },     88.26384,   0.00003,     0.070 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   0,   0 },     87.17070,   0.00003,     1.165 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   1,  -1 },     71.30361,   0.00003,     0.976 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   1,   0 },     91.80227,   0.00003,     0.081 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   2,  -1 },    296.89728,   0.00002,     0.082 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    272.12924,   0.00010,     0.573 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    271.38461,   0.00189,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   1,   0 },    271.75742,   0.00124,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   2,   0 },    271.89467,   0.00015,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   3,   0 },    271.80227,   0.00001,     0.020 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   1,   0,  -1 },    296.96521,   0.00002,     0.096 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   1,   1,  -1 },    296.96521,   0.00003,     0.042 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.16218,   0.00007,     0.043 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,   0,   0 },     91.47034,   0.00010,     0.028 },
 { {   0,  20, -21,   0,   0,   0,   0,   2,  -2,   1,   0 },    255.98243,   0.00003,    24.472 },
 { {   0,  20, -21,   0,   0,   0,   1,   0,  -2,   1,   0 },    219.05548,   0.00003,    24.614 },
 { {   0,  20, -21,   0,   0,   0,   2,   0,  -2,   1,   0 },    262.59358,   0.00003,    19.037 },
 { {   0,  20, -21,   0,   0,   2,   0,   0,  -2,   1,   0 },    231.09359,   0.00001,    10.350 },
 { {   0,  20, -21,   0,   0,   3,   0,   0,  -2,   1,   0 },    161.36050,   0.00001,     7.659 },
 { {   0,  20, -21,   0,   1,   0,   0,   0,  -2,   0,   0 },    297.24734,   0.00003,     0.076 },
 { {   0,  20, -21,   0,   1,   0,   0,   0,  -2,   1,   0 },    273.88184,   0.00006,     8.847 },
 { {   0,  20, -21,   0,   1,   0,   0,   0,  -2,   2,   0 },    178.38548,   0.00002,     0.075 },
 { {   0,  20, -21,   0,   2,   0,   0,   0,  -2,   0,   0 },     45.85071,   0.00002,     0.077 },
 { {   0,  20, -21,   0,   2,   0,   0,   0,  -2,   1,   0 },    240.50050,   0.00002,     5.068 },
 { {   0,  20, -21,   0,   2,   0,   0,   0,  -2,   2,   0 },     45.85071,   0.00002,     0.074 },
 { {   0,  20, -21,   3,   0,   0,   0,   0,   0,  -1,   0 },    298.69135,   0.00001,     6.656 },
 { {   0,  20, -20,  -2,   0,   0,   0,   0,  -2,   0,   0 },     30.08443,   0.00001,     0.075 },
 { {   0,  20, -20,  -2,   0,   0,   0,   0,  -2,   1,   0 },    226.16897,   0.00005,    28.867 },
 { {   0,  20, -20,  -2,   0,   0,   0,   0,  -2,   2,   0 },    210.08443,   0.00001,     0.076 },
 { {   0,  20, -20,   0,  -2,   0,   0,   0,  -2,   1,   0 },    190.35156,   0.00002,     1.163 },
 { {   0,  20, -20,   0,  -2,   0,   0,   0,  -2,   2,   0 },    190.66035,   0.00002,     0.071 },
 { {   0,  20, -20,   0,  -2,   0,   0,   0,  -1,   0,   0 },     19.60512,   0.00005,    37.187 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,  -2,   0,   0 },     15.31126,   0.00001,     0.081 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,  -2,   1,   0 },    200.88621,   0.00004,     1.059 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,  -2,   2,   0 },    195.31126,   0.00001,     0.070 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,  -1,   0,   0 },     59.53673,   0.00004,    17.418 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,   0,  -2,   0 },     92.80541,   0.00001,     0.071 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,   0,  -1,   0 },    300.97172,   0.00003,     1.205 },
 { {   0,  20, -20,   0,  -1,   0,   0,   0,   0,   0,   0 },    272.80541,   0.00001,     0.080 },
 { {   0,  20, -20,   0,   0,  -1,   0,   0,  -2,   1,   0 },     92.04054,   0.00001,     1.005 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -5,   2,  -1 },    194.02829,   0.00001,     0.028 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -4,   0,   0 },    168.86534,   0.00001,     0.027 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -4,   1,   0 },    168.86534,   0.00009,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -4,   2,   0 },    166.26619,   0.00004,     0.096 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   0,  -1 },    194.02829,   0.00006,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   0,   0 },      0.00384,   0.00003,     0.041 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,  -1 },    194.01066,   0.00051,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,   0 },      0.00384,   0.00003,     0.088 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   2,  -1 },    194.05968,   0.00047,     0.087 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.62350,   0.00030,     0.082 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   1,  -1 },     14.02829,   0.00001,     0.081 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    179.65063,   0.00021,     0.972 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    190.51079,   0.00003,     0.070 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -2,  -1 },    194.02829,   0.00001,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    193.80691,   0.00023,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,   0 },      0.00384,   0.00014,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,   1 },     15.99522,   0.00004,     3.242 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -3 },     14.02829,   0.00001,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.80687,   0.00340,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,   0 },      0.00384,   0.00258,     7.056 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -3 },     14.02829,   0.00009,     0.037 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    194.02829,   0.04280,    39.938 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.00384,   0.00014,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   1 },     14.02829,   0.00009,     0.037 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,  -1 },    193.99747,   0.00237,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,   1 },     14.02829,   0.00001,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   3,  -1 },    193.99739,   0.00016,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   4,  -1 },    194.02829,   0.00001,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    181.52494,   0.00003,     0.071 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.03835,   0.00117,     1.342 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    180.48907,   0.00094,     0.080 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   1,  -2 },    179.76262,   0.00001,     0.927 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   1,  -1 },     14.02829,   0.00001,     0.081 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   1,   0 },    184.34621,   0.00004,     0.039 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.00384,   0.00003,     0.086 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },    194.60739,   0.00026,     0.087 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   0,   0 },      0.00384,   0.00003,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   1,  -1 },    194.04691,   0.00048,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   2,  -1 },    194.02829,   0.00006,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.58700,   0.00003,     0.093 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.86244,   0.00012,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   0,   0 },    359.34623,   0.00019,     0.027 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   1,   0 },    359.71844,   0.00003,     0.020 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   3,  -2,   0 },    359.80266,   0.00001,     0.043 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   3,   0,  -1 },    194.02829,   0.00001,     0.028 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   3,   0,   0 },    359.80266,   0.00001,     0.020 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.71579,   0.00003,     0.021 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   4,   0,   0 },    359.71578,   0.00004,     0.016 },
 { {   0,  20, -20,   0,   0,   0,   2,   0,  -1,   1,  -1 },    152.74894,   0.00003,   809.552 },
 { {   0,  20, -20,   0,   0,   1,   0,   0,  -1,   1,  -1 },    135.29029,   0.00001,   112.250 },
 { {   0,  20, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    167.82006,   0.00001,     0.082 },
 { {   0,  20, -20,   0,   1,   0,   0,   0,  -2,   1,   0 },    347.00214,   0.00003,     0.898 },
 { {   0,  20, -20,   0,   1,   0,   0,   0,  -2,   2,   0 },    347.82006,   0.00001,     0.070 },
 { {   0,  20, -20,   0,   1,   0,   0,   0,   0,  -1,   0 },    347.82006,   0.00001,     1.513 },
 { {   0,  20, -20,   0,   1,   0,   0,   0,   0,   0,   0 },    170.98437,   0.00002,     0.079 },
 { {   0,  20, -20,   2,   0,   0,   0,   0,   0,   1,  -2 },     59.26186,   0.00001,    64.783 },
 { {   0,  20, -19,  -4,   0,   0,   0,   0,  -2,   1,   0 },      4.43767,   0.00003,    10.203 },
 { {   0,  20, -19,  -3,   0,   0,   0,   0,  -2,   0,   0 },    123.99606,   0.00003,     0.078 },
 { {   0,  20, -19,  -3,   0,   0,   0,   0,  -2,   1,   0 },     99.32701,   0.00004,     2.306 },
 { {   0,  20, -19,  -3,   0,   0,   0,   0,  -2,   2,   0 },    123.99606,   0.00003,     0.073 },
 { {   0,  20, -19,  -2,   0,   0,   0,   0,  -2,   1,   0 },    101.04586,   0.00001,     1.036 },
 { {   0,  20, -19,  -2,   0,   0,   0,   0,  -1,   0,   0 },    297.01328,   0.00002,    12.764 },
 { {   0,  20, -19,  -2,   0,   0,   0,   0,  -1,   1,  -1 },     77.24374,   0.00003,    11.311 },
 { {   0,  20, -19,  -1,   0,   0,   0,   0,   0,  -1,   0 },    180.23681,   0.00002,     3.610 },
 { {   0,  20, -19,   0,  -4,   0,   0,   0,   0,  -1,   0 },    117.49807,   0.00003,    12.120 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,  -2,   0 },    107.82783,   0.00006,     0.075 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,  -1,   0 },    116.74397,   0.00047,   556.895 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,   0,   0 },    161.26547,   0.00005,     0.075 },
 { {   0,  20, -19,   0,  -2,   0,   0,   0,   0,  -1,   0 },    121.12913,   0.00003,    11.615 },
 { {   0,  20, -19,   0,  -1,   0,   0,   0,   0,  -1,   0 },      0.85892,   0.00001,     5.868 },
 { {   0,  20, -19,   0,  -1,   0,   0,   0,   0,   0,  -1 },    325.93978,   0.00002,   274.159 },
 { {   0,  20, -19,   0,   0,  -2,   0,   0,   0,  -1,   0 },     96.11209,   0.00004,     5.353 },
 { {   0,  20, -19,   0,   0,  -1,   0,   0,   0,  -1,   0 },    339.63949,   0.00001,     4.530 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -3,   1,  -1 },     91.09136,   0.00003,     0.042 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -3,   2,  -1 },     91.09136,   0.00002,     0.095 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   0,  -1 },    282.98935,   0.00002,     0.041 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   1,  -1 },    282.98935,   0.00002,     0.088 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },    296.37771,   0.00004,     0.081 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   1,  -1 },    102.52409,   0.00006,     1.026 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   2,  -1 },    271.01797,   0.00001,     0.070 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },    252.08480,   0.00001,     0.077 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,  -1,  -1 },    282.98935,   0.00011,     0.076 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    208.72693,   0.00011,     3.926 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,  -1 },    282.98935,   0.00200,    11.370 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    261.44088,   0.00002,     0.074 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   1,  -2 },    279.79124,   0.00010,    12.690 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   1,  -1 },    282.98935,   0.00011,     0.075 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,  -2,  -1 },    103.00182,   0.00001,     0.071 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.00155,   0.00019,     1.251 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.00834,   0.00015,     0.080 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,  -1,  -1 },    282.98935,   0.00002,     0.086 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },     94.48577,   0.00001,     0.040 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,   0,  -1 },    282.98935,   0.00002,     0.040 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     89.58860,   0.00003,     0.026 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   4,  -2,   0 },     94.51541,   0.00001,     0.027 },
 { {   0,  20, -19,   0,   1,   0,   0,   0,   0,   1,  -2 },    242.34213,   0.00006,   181.839 },
 { {   0,  20, -18,  -6,   0,   0,   0,   0,  -2,   1,   0 },    139.29457,   0.00003,     6.197 },
 { {   0,  20, -18,  -5,   0,   0,   0,   0,  -2,   1,   0 },    214.43786,   0.00002,     2.701 },
 { {   0,  20, -18,  -4,   0,   0,   0,   0,  -1,   0,   0 },     51.31473,   0.00002,    66.754 },
 { {   0,  20, -18,  -2,   0,   0,   0,   0,  -2,   1,   0 },    179.44861,   0.00002,     0.509 },
 { {   0,  20, -18,   0,  -2,   0,   0,   0,   0,  -1,   0 },    359.54844,   0.00003,     0.921 },
 { {   0,  20, -18,   0,  -2,   0,   0,   0,   1,  -1,  -1 },     47.01136,   0.00003,    30.927 },
 { {   0,  20, -18,   0,  -1,   0,   0,   0,   1,  -1,  -1 },    256.07697,   0.00002,     8.574 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     20.96995,   0.00002,     0.098 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    153.69135,   0.00001,     0.083 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },     28.26734,   0.00005,     0.797 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    199.95526,   0.00008,     0.069 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,  -1,  -2 },    206.45318,   0.00003,     0.076 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,  -1,  -1 },    213.56812,   0.00011,     4.977 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   0,  -2 },    206.45318,   0.00063,    29.252 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },      5.89453,   0.00002,     0.074 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   1,  -2 },    206.45318,   0.00003,     0.075 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    205.62567,   0.00002,     0.025 },
 { {   0,  20, -17,   0,  -1,   0,   0,   0,   2,  -1,  -2 },    232.83495,   0.00001,    15.905 },
 { {   0,  20, -17,   0,   0,   0,   0,   0,   2,   0,  -3 },    129.73514,   0.00003,    51.076 },
 { {   0,  20, -16,  -6,   0,   0,   0,   0,   0,  -1,   0 },     41.50025,   0.00002,    15.479 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,  -2,   0 },    127.30124,   0.00003,     0.075 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,  -1,   0 },    100.59376,   0.00053,   809.761 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,   0,   0 },    127.30124,   0.00003,     0.075 },
 { {   0,  20, -15,  -6,   0,   0,   0,   0,   1,  -1,  -1 },     77.76865,   0.00002,    92.271 },
 { {   0,  20, -15,  -4,   0,   0,   0,   0,   2,  -1,  -2 },    302.15620,   0.00001,    48.923 },
 { {   0,  20, -14, -10,   0,   0,   0,   0,   0,  -1,   0 },     63.63286,   0.00002,    16.095 },
 { {   0,  20, -13, -15,   0,   0,   0,   0,  -2,   1,   0 },    222.74364,   0.00001,    18.705 },
 { {   0,  20, -12, -17,   0,   0,   0,   0,  -2,   1,   0 },    146.77366,   0.00002,   100.952 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -3,   1,  -1 },    354.13759,   0.00005,     0.040 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -3,   2,  -1 },    354.13759,   0.00005,     0.087 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,  -1,  -1 },    354.13759,   0.00002,     0.038 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   0,  -1 },    354.13759,   0.00022,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   1,  -1 },    354.13759,   0.00407,  3300.556 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   2,  -1 },    354.13759,   0.00022,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   3,  -1 },    354.13759,   0.00002,     0.038 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,   1,   0,  -1 },    354.13759,   0.00005,     0.087 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,   1,   1,  -1 },    354.13759,   0.00005,     0.040 },
 { {   0,  21, -29,  15,   0,   0,   0,   0,   0,  -1,   0 },    202.98938,   0.00004,     6.920 },
 { {   0,  21, -28,  10,   0,   0,   0,   0,  -2,   1,   0 },    210.97604,   0.00002,    34.593 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,  -1,   0 },    144.00210,   0.00001,     0.033 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    236.74721,   0.00005,     0.060 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    144.00210,   0.00001,     0.299 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.74721,   0.00005,     0.101 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.00210,   0.00001,     0.195 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    100.86375,   0.00001,     0.123 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,   0,   1,   0 },    144.00210,   0.00001,     0.047 },
 { {   0,  21, -26,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.08139,   0.00003,     0.061 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.42777,   0.00003,     0.025 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.42777,   0.00002,     0.243 },
 { {   0,  21, -24,   0,  -1,   0,   0,   0,  -2,   1,   0 },     55.06012,   0.00001,     0.699 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    299.37223,   0.00001,     0.068 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    299.37223,   0.00001,     0.084 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,   2,   0,   0 },     82.30157,   0.00001,     0.029 },
 { {   0,  21, -24,   2,   0,   0,   0,   0,  -2,   0,   0 },    270.74062,   0.00001,     0.074 },
 { {   0,  21, -24,   2,   0,   0,   0,   0,  -2,   1,   0 },    296.45138,   0.00001,     3.541 },
 { {   0,  21, -24,   2,   0,   0,   0,   0,  -2,   2,   0 },    270.74062,   0.00001,     0.077 },
 { {   0,  21, -24,   3,   0,   0,   0,   0,  -2,   1,   0 },     99.91011,   0.00002,     4.011 },
 { {   0,  21, -24,   6,   0,   0,   0,   0,   0,  -1,   0 },     26.69430,   0.00003,    14.217 },
 { {   0,  21, -23,   0,  -2,   0,   0,   0,  -2,   0,   0 },     51.49075,   0.00001,     0.073 },
 { {   0,  21, -23,   0,  -2,   0,   0,   0,  -2,   1,   0 },    230.36001,   0.00002,     1.944 },
 { {   0,  21, -23,   0,  -2,   0,   0,   0,  -2,   2,   0 },    231.49075,   0.00001,     0.078 },
 { {   0,  21, -23,   0,  -1,   0,   0,   0,  -2,   0,   0 },    189.50207,   0.00001,     0.073 },
 { {   0,  21, -23,   0,  -1,   0,   0,   0,  -2,   1,   0 },    339.08254,   0.00003,     2.325 },
 { {   0,  21, -23,   0,  -1,   0,   0,   0,  -2,   2,   0 },      9.50207,   0.00001,     0.078 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -2,   2 },    169.56912,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -1,   2 },    166.67445,   0.00054,    81.722 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    328.80337,   0.00004,     0.074 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   0,   2 },    169.56912,   0.00004,     0.076 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    344.94325,   0.00021,     2.892 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    166.37843,   0.00002,     0.077 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    351.45086,   0.00002,     0.472 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    352.97649,   0.00008,     0.090 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   2,  -1,   0 },    177.81792,   0.00001,     0.044 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   2,   0,   0 },    177.77300,   0.00005,     0.028 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,  -2,   2 },     15.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,  -1,   2 },     15.43642,   0.00081,  2987.643 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,   0,   2 },     15.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   0,   1,   0,   0,  -2,   1,   0 },    265.27388,   0.00002,     3.207 },
 { {   0,  21, -23,   0,   1,   0,   0,   0,  -2,   0,   0 },    273.66686,   0.00007,     0.074 },
 { {   0,  21, -23,   0,   1,   0,   0,   0,  -2,   1,   0 },    209.35700,   0.00003,     3.825 },
 { {   0,  21, -23,   0,   1,   0,   0,   0,  -2,   2,   0 },     89.94270,   0.00002,     0.077 },
 { {   0,  21, -23,   0,   2,   0,   0,   0,  -2,   1,   0 },    291.82028,   0.00003,     5.645 },
 { {   0,  21, -23,   0,   3,   0,   0,   0,  -2,   1,   0 },    307.47074,   0.00003,    10.772 },
 { {   0,  21, -23,   0,   4,   0,   0,   0,  -2,   0,   0 },    340.82882,   0.00002,     0.075 },
 { {   0,  21, -23,   0,   4,   0,   0,   0,  -2,   1,   0 },    211.96555,   0.00005,   117.194 },
 { {   0,  21, -23,   0,   4,   0,   0,   0,  -2,   2,   0 },    160.82882,   0.00002,     0.075 },
 { {   0,  21, -23,   1,   0,   0,   0,   0,  -2,   1,   0 },     92.50793,   0.00001,     5.378 },
 { {   0,  21, -23,   2,   0,   0,   0,   0,  -2,   0,   0 },    336.46116,   0.00002,     0.080 },
 { {   0,  21, -23,   2,   0,   0,   0,   0,  -2,   2,   0 },    336.46116,   0.00002,     0.072 },
 { {   0,  21, -23,   2,   0,   0,   0,   0,  -1,  -1,   1 },    178.68074,   0.00009,   379.678 },
 { {   0,  21, -23,   4,   0,   0,   0,   0,   0,  -1,   0 },     59.68772,   0.00024,   143.494 },
 { {   0,  21, -23,   6,   0,   0,   0,   0,   1,  -1,  -1 },    140.22023,   0.00002,    60.344 },
 { {   0,  21, -22,   0,  -2,   0,   0,   0,  -2,   0,   0 },    296.97463,   0.00001,     0.078 },
 { {   0,  21, -22,   0,  -2,   0,   0,   0,  -2,   2,   0 },    116.97463,   0.00001,     0.073 },
 { {   0,  21, -22,   0,  -1,   0,   0,   0,  -2,   0,   0 },    282.14875,   0.00003,     0.079 },
 { {   0,  21, -22,   0,  -1,   0,   0,   0,  -2,   1,   0 },    288.44185,   0.00004,     1.755 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -4,   2,   0 },     91.78587,   0.00001,     0.092 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.09438,   0.00011,     0.079 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    268.74734,   0.00014,     1.528 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    271.78587,   0.00007,     0.072 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -2,   1 },    250.55274,   0.00003,     0.075 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   0 },     91.99959,   0.00002,     0.074 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   1 },    250.55328,   0.00062,    15.151 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },     70.61512,   0.00003,     0.073 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.99959,   0.00030,     4.296 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,   1 },    250.57042,   0.00010,     0.076 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   1,  -1 },     70.63567,   0.00013,     2.503 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   1,   0 },     91.99959,   0.00002,     0.077 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    272.37283,   0.00015,     0.893 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    272.57359,   0.00026,     0.082 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,   1,   0 },     94.21672,   0.00002,     0.039 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -3,   0 },     92.25031,   0.00002,     0.346 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -2,   0 },    280.33224,   0.00002,     0.097 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -1,   0 },    278.26530,   0.00005,     0.042 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,   0,   0 },     92.25182,   0.00012,     0.027 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,   1,   0 },     92.25236,   0.00001,     0.020 },
 { {   0,  21, -22,   0,   0,   2,   0,   0,  -1,  -1,   1 },    229.05570,   0.00004,   527.988 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },     21.41385,   0.00004,     0.080 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,  -2,   1,   0 },    128.46767,   0.00003,     1.354 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,  -2,   2,   0 },    264.83912,   0.00001,     0.071 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,  -1,  -1,   1 },    171.85514,   0.00003,    54.640 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,   0,  -1,   0 },    177.89085,   0.00002,     0.966 },
 { {   0,  21, -22,   0,   2,   0,   0,   0,  -2,   1,   0 },    233.23315,   0.00003,     1.215 },
 { {   0,  21, -22,   0,   2,   0,   0,   0,  -1,  -1,   1 },    264.37005,   0.00002,     9.746 },
 { {   0,  21, -22,   0,   2,   0,   0,   0,  -1,   0,   0 },     53.23314,   0.00003,    15.589 },
 { {   0,  21, -22,   2,   0,   0,   0,   0,   0,  -1,   0 },     55.53629,   0.00001,    17.731 },
 { {   0,  21, -22,   3,   0,   0,   0,   0,   0,  -1,   0 },    236.04781,   0.00001,     2.104 },
 { {   0,  21, -21,  -3,   0,   0,   0,   0,  -2,   0,   0 },     13.84813,   0.00003,     0.076 },
 { {   0,  21, -21,  -3,   0,   0,   0,   0,  -2,   1,   0 },    329.41123,   0.00002,    16.893 },
 { {   0,  21, -21,  -3,   0,   0,   0,   0,  -2,   2,   0 },     13.84813,   0.00003,     0.075 },
 { {   0,  21, -21,   0,  -2,   0,   0,   0,  -2,   1,   0 },     25.12433,   0.00001,     0.673 },
 { {   0,  21, -21,   0,  -2,   0,   0,   0,   0,  -1,   0 },    353.88568,   0.00001,     3.468 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,  -2,   0,   0 },    102.26006,   0.00002,     0.086 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,  -2,   1,   0 },      5.48897,   0.00002,     0.637 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,  -2,   2,   0 },    102.26006,   0.00002,     0.067 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,   0,  -1,   0 },    282.35594,   0.00004,     4.900 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,   0,   0,   0 },    263.74555,   0.00004,     0.077 },
 { {   0,  21, -21,   0,   0,  -1,   0,   0,   0,  -1,   0 },    268.56549,   0.00003,     6.506 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -4,   2,   0 },      1.64813,   0.00001,     0.102 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -2,   0 },    180.02373,   0.00004,     0.026 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.97132,   0.00032,     0.040 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.99408,   0.00042,     0.086 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    359.91997,   0.00001,     0.605 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    181.65593,   0.00006,     0.067 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -1,   0,  -1 },     32.32431,   0.00002,     0.079 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -1,   0,   0 },    357.86943,   0.00006,     1.303 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -1,   1,  -1 },    192.93843,   0.00006,     1.665 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -3,   0 },    180.12862,   0.00012,     0.038 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -2,  -2 },    359.94393,   0.00001,     0.025 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },    180.12886,   0.00171,     0.075 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,  -2 },    359.94393,   0.00006,     0.037 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.94393,   0.03238,     8.350 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,   2 },    359.94393,   0.00006,     0.037 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,  -1 },     13.86949,   0.00007,     3.490 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    179.87972,   0.00542,     0.076 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   2 },    359.94393,   0.00001,     0.025 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   1,  -2 },    179.72878,   0.00007,     2.206 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   1,   0 },    179.88916,   0.00034,     0.038 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.50293,   0.00003,     0.852 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.77277,   0.00038,     0.088 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.88914,   0.00049,     0.041 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,   0,   0 },    359.40050,   0.00006,     0.026 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,   1,   0 },    359.67886,   0.00002,     0.020 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.67610,   0.00001,     0.020 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   4,   0,   0 },    359.67609,   0.00003,     0.016 },
 { {   0,  21, -21,   0,   0,   1,   0,   0,   0,  -1,   0 },    249.03019,   0.00002,    11.653 },
 { {   0,  21, -21,   0,   0,   3,   0,   0,   0,  -1,   0 },    245.38156,   0.00003,    55.794 },
 { {   0,  21, -21,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.30346,   0.00002,     0.575 },
 { {   0,  21, -21,   0,   2,   0,   0,   0,   0,  -1,   0 },     53.96289,   0.00005,    20.475 },
 { {   0,  21, -20,  -5,   0,   0,   0,   0,  -2,   1,   0 },    302.34336,   0.00010,   239.709 },
 { {   0,  21, -20,  -3,   0,   0,   0,   0,  -1,   1,  -1 },    353.61109,   0.00008,   184.128 },
 { {   0,  21, -20,  -2,   0,   0,   0,   0,   0,  -1,   0 },    113.00661,   0.00003,     5.461 },
 { {   0,  21, -20,  -1,   0,   0,   0,   0,   0,   1,  -2 },      0.08161,   0.00001,    66.517 },
 { {   0,  21, -20,   0,  -2,   0,   0,   0,   0,  -1,   0 },     91.26065,   0.00002,     1.405 },
 { {   0,  21, -20,   0,  -1,   0,   0,   0,   0,  -2,   0 },    181.04329,   0.00002,     0.080 },
 { {   0,  21, -20,   0,  -1,   0,   0,   0,   0,   0,   0 },      1.04329,   0.00002,     0.071 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -2,  -1,   0 },     77.00700,   0.00003,     0.042 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.80670,   0.00002,     0.094 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    102.93263,   0.00003,     0.040 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },    283.86502,   0.00006,     0.086 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    102.89871,   0.00002,     0.625 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   0,   0,  -1 },    282.96276,   0.00002,     1.402 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -3,  -1 },    102.91603,   0.00001,     0.037 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -3,   0 },      6.55100,   0.00002,     0.075 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,  -1 },    102.91551,   0.00016,     0.074 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,   0 },      6.55100,   0.00034,   147.570 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,  -1 },    102.93263,   0.00298,     5.763 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,   0 },      6.55100,   0.00002,     0.075 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },    103.37069,   0.00064,     0.076 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   1,  -1 },    103.37703,   0.00004,     0.038 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     81.25020,   0.00002,     0.026 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   3,  -2,  -1 },    102.95002,   0.00003,     0.088 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   3,  -1,  -1 },    102.93263,   0.00003,     0.041 },
 { {   0,  21, -20,   0,   2,   0,   0,   0,   1,  -1,  -1 },     41.14601,   0.00003,   203.061 },
 { {   0,  21, -19,  -7,   0,   0,   0,   0,  -2,   1,   0 },    285.40936,   0.00002,    14.806 },
 { {   0,  21, -19,  -4,   0,   0,   0,   0,   0,  -1,   0 },    219.29743,   0.00002,     4.057 },
 { {   0,  21, -19,  -3,   0,   0,   0,   0,   0,  -1,   0 },    329.17652,   0.00001,     3.507 },
 { {   0,  21, -19,  -3,   0,   0,   0,   0,   0,   1,  -2 },    154.16443,   0.00001,    20.690 },
 { {   0,  21, -19,  -2,   0,   0,   0,   0,  -2,   0,   0 },    359.57900,   0.00001,     0.094 },
 { {   0,  21, -19,  -2,   0,   0,   0,   0,  -2,   2,   0 },    359.57900,   0.00001,     0.063 },
 { {   0,  21, -19,  -2,   0,   0,   0,   0,   0,  -1,   0 },    180.25102,   0.00001,     1.224 },
 { {   0,  21, -19,   0,  -2,   0,   0,   0,  -2,   1,   0 },    179.78249,   0.00001,     0.287 },
 { {   0,  21, -19,   0,   0,  -3,   0,   0,   2,  -3,   0 },     92.43266,   0.00001,   228.825 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.46541,   0.00002,     0.532 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    206.46720,   0.00003,     0.066 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   2,  -3,   0 },     25.40088,   0.00003,     9.415 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   2,  -1,  -2 },     25.69786,   0.00003,     4.400 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    163.54967,   0.00002,     0.025 },
 { {   0,  21, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    128.02652,   0.00002,     0.065 },
 { {   0,  21, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    128.02652,   0.00002,     0.024 },
 { {   0,  21, -16,  -9,   0,   0,   0,   0,   0,  -1,   0 },     53.04904,   0.00005,    10.518 },
 { {   0,  21, -15, -11,   0,   0,   0,   0,   0,  -1,   0 },     11.24359,   0.00002,    31.541 },
 { {   0,  21, -14, -13,   0,   0,   0,   0,   0,  -1,   0 },    351.47519,   0.00002,    31.583 },
 { {   0,  22, -29,   9,   0,   0,   0,   0,  -2,   1,   0 },    179.31636,   0.00002,    15.401 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   0,   0 },    318.16930,   0.00003,     0.075 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   1,   0 },    206.04256,   0.00024,   640.093 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   2,   0 },    318.16930,   0.00003,     0.075 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,  -1,   0 },    143.98743,   0.00001,     0.034 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    236.54734,   0.00004,     0.063 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },    170.86652,   0.00002,     0.368 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.54734,   0.00004,     0.095 },
 { {   0,  22, -27,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.04246,   0.00003,     0.063 },
 { {   0,  22, -27,   0,   1,   0,   0,   0,  -2,   1,   0 },     50.04246,   0.00001,     0.379 },
 { {   0,  22, -27,   5,   0,   0,   0,   0,  -2,   1,   0 },    212.49624,   0.00003,    16.180 },
 { {   0,  22, -26,   0,  -1,   0,   0,   0,  -2,   1,   0 },     60.05587,   0.00001,     0.554 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.38614,   0.00002,     0.025 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.38614,   0.00002,     0.286 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,   0,   1,   0 },     51.38614,   0.00002,     0.043 },
 { {   0,  22, -26,   0,   1,   0,   0,   0,  -2,   1,   0 },    218.26520,   0.00001,     0.611 },
 { {   0,  22, -26,   3,   0,   0,   0,   0,  -2,   1,   0 },    165.76042,   0.00006,     7.989 },
 { {   0,  22, -26,   4,   0,   0,   0,   0,  -2,   1,   0 },    299.85471,   0.00001,     2.460 },
 { {   0,  22, -25,   0,  -2,   0,   0,   0,  -3,   2,   0 },    353.47430,   0.00003,   547.201 },
 { {   0,  22, -25,   0,  -1,   0,   0,   0,  -2,   0,   0 },     55.01568,   0.00003,     0.071 },
 { {   0,  22, -25,   0,  -1,   0,   0,   0,  -2,   1,   0 },    232.12775,   0.00002,     1.243 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    107.32915,   0.00001,     0.072 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    287.32915,   0.00001,     0.080 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,   2,   0,   0 },     82.68351,   0.00001,     0.028 },
 { {   0,  22, -25,   0,   1,   0,   0,   0,  -2,   1,   0 },    325.27517,   0.00003,     1.573 },
 { {   0,  22, -25,   0,   2,   0,   0,   0,  -2,   1,   0 },    265.71061,   0.00001,     1.813 },
 { {   0,  22, -25,   3,   0,   0,   0,   0,  -1,   0,   0 },    187.16862,   0.00001,    82.081 },
 { {   0,  22, -24,  -2,   0,   0,   0,   0,  -2,   1,   0 },    234.59689,   0.00001,     1.276 },
 { {   0,  22, -24,   0,  -4,   0,   0,   0,  -2,   1,   0 },     96.69689,   0.00002,    17.408 },
 { {   0,  22, -24,   0,  -3,   0,   0,   0,  -2,   0,   0 },     61.44067,   0.00001,     0.076 },
 { {   0,  22, -24,   0,  -3,   0,   0,   0,  -2,   1,   0 },     22.45366,   0.00004,    37.232 },
 { {   0,  22, -24,   0,  -3,   0,   0,   0,  -2,   2,   0 },     61.44067,   0.00001,     0.075 },
 { {   0,  22, -24,   0,  -2,   0,   0,   0,  -2,   0,   0 },     45.10611,   0.00005,     0.076 },
 { {   0,  22, -24,   0,  -2,   0,   0,   0,  -2,   1,   0 },    350.55313,   0.00003,     8.996 },
 { {   0,  22, -24,   0,  -2,   0,   0,   0,  -2,   2,   0 },     36.90369,   0.00002,     0.075 },
 { {   0,  22, -24,   0,  -1,   0,   0,   0,  -2,   1,   0 },     23.32894,   0.00003,     5.116 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    142.38789,   0.00002,     0.077 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.10567,   0.00024,     3.574 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    342.20044,   0.00003,     0.074 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,   0,   0 },    168.50168,   0.00001,     1.647 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,   1,  -1 },    167.04185,   0.00001,     1.292 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.88644,   0.00003,     0.669 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    352.56528,   0.00005,     0.085 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   2,  -1,   0 },    178.07714,   0.00001,     0.043 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   2,   0,   0 },    178.03482,   0.00004,     0.027 },
 { {   0,  22, -24,   0,   0,   1,   0,   0,  -2,   1,   0 },    265.52150,   0.00001,     3.188 },
 { {   0,  22, -24,   0,   1,   0,   0,   0,  -2,   0,   0 },    111.02878,   0.00003,     0.078 },
 { {   0,  22, -24,   0,   1,   0,   0,   0,  -2,   1,   0 },    206.35408,   0.00002,     2.747 },
 { {   0,  22, -24,   0,   1,   0,   0,   0,  -2,   2,   0 },     18.62382,   0.00001,     0.073 },
 { {   0,  22, -24,   3,   0,   0,   0,   0,   0,  -1,   0 },    110.30917,   0.00002,     9.920 },
 { {   0,  22, -23,  -3,   0,   0,   0,   0,  -2,   1,   0 },    132.71109,   0.00003,     3.172 },
 { {   0,  22, -23,  -2,   0,   0,   0,   0,  -2,   1,   0 },    243.32376,   0.00001,     4.621 },
 { {   0,  22, -23,   0,  -3,   0,   0,   0,  -1,   1,  -1 },     90.68128,   0.00001,    37.167 },
 { {   0,  22, -23,   0,  -2,   0,   0,   0,  -1,   1,  -1 },     67.05925,   0.00002,    17.422 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    106.24455,   0.00008,     0.084 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    264.93181,   0.00002,     0.781 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.55520,   0.00001,     0.069 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    249.82742,   0.00003,     0.077 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,   0 },     93.28807,   0.00008,     2.546 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   1,  -1 },     69.98837,   0.00029,     4.425 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   2,  -1 },     69.98834,   0.00002,     0.074 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    272.40772,   0.00059,     2.023 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    272.16585,   0.00031,     0.078 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,   1,   0 },     92.23769,   0.00001,     0.038 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   1,  -1,  -1 },     69.94416,   0.00001,     0.646 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   2,   0,   0 },     92.23714,   0.00009,     0.027 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   2,   1,   0 },     92.23769,   0.00001,     0.020 },
 { {   0,  22, -23,   0,   1,   0,   0,   0,   0,  -1,   0 },    135.06817,   0.00002,     2.439 },
 { {   0,  22, -23,   0,   1,   0,   0,   0,   0,   0,   0 },    177.70094,   0.00003,     0.078 },
 { {   0,  22, -23,   0,   2,   0,   0,   0,   0,  -1,   0 },    226.38663,   0.00003,     3.071 },
 { {   0,  22, -23,   1,   0,   0,   0,   0,   0,  -1,   0 },     96.45579,   0.00002,    26.710 },
 { {   0,  22, -22,  -4,   0,   0,   0,   0,  -2,   1,   0 },     40.23933,   0.00003,     6.535 },
 { {   0,  22, -22,  -3,   0,   0,   0,   0,  -1,  -1,   1 },     40.26711,   0.00002,    28.144 },
 { {   0,  22, -22,  -1,   0,   0,   0,   0,   0,  -1,   0 },    269.07398,   0.00003,    38.568 },
 { {   0,  22, -22,   0,  -2,   0,   0,   0,   0,  -1,   0 },    199.68106,   0.00004,     2.966 },
 { {   0,  22, -22,   0,  -2,   0,   0,   0,   0,   1,  -2 },    345.16992,   0.00006,   275.284 },
 { {   0,  22, -22,   0,  -1,   0,   0,   0,  -2,   0,   0 },    189.53799,   0.00001,     0.090 },
 { {   0,  22, -22,   0,  -1,   0,   0,   0,  -2,   2,   0 },      9.53799,   0.00001,     0.065 },
 { {   0,  22, -22,   0,  -1,   0,   0,   0,   0,  -2,   0 },      9.53799,   0.00001,     0.078 },
 { {   0,  22, -22,   0,  -1,   0,   0,   0,   0,  -1,   0 },    186.92089,   0.00002,     2.373 },
 { {   0,  22, -22,   0,  -1,   0,   0,   0,   0,   0,   0 },    189.53799,   0.00001,     0.073 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,  -1,   0 },    179.79678,   0.00001,     0.041 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    170.70902,   0.00014,     0.091 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    175.71949,   0.00002,     0.439 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    177.06422,   0.00004,     0.064 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -1,   0,   0 },    354.43227,   0.00001,     0.718 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.45315,   0.00001,     0.816 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,  -2,   0 },    179.71600,   0.00012,     0.078 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.79992,   0.00128,     1.977 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,  -2 },      0.11333,   0.00001,     0.076 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,  -1 },     13.56099,   0.00004,     2.950 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,   0 },      2.10529,   0.00042,     0.073 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   1,  -2 },    179.54555,   0.00036,     5.806 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   1,   0 },      0.76849,   0.00006,     0.037 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   2,  -2 },    179.54387,   0.00002,     0.074 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   2,   0 },    179.63436,   0.00001,     0.025 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   1,  -1,  -1 },    192.52457,   0.00012,     1.825 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.43394,   0.00005,     0.079 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.66843,   0.00007,     0.040 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   2,   0,   0 },    359.63444,   0.00010,     0.026 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   4,   0,   0 },    359.63436,   0.00001,     0.016 },
 { {   0,  22, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },    350.91483,   0.00001,     0.092 },
 { {   0,  22, -22,   0,   1,   0,   0,   0,  -2,   2,   0 },    170.91483,   0.00001,     0.064 },
 { {   0,  22, -22,   1,   0,   0,   0,   0,   1,  -1,  -1 },     83.66897,   0.00001,    61.253 },
 { {   0,  22, -22,   3,   0,   0,   0,   0,   2,  -1,  -2 },    266.02502,   0.00002,   148.742 },
 { {   0,  22, -21,  -6,   0,   0,   0,   0,  -2,   1,   0 },    192.11340,   0.00002,    11.153 },
 { {   0,  22, -21,  -2,   0,   0,   0,   0,   0,  -1,   0 },    257.93440,   0.00003,     2.261 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,  -1,   0,  -1 },    102.92773,   0.00003,     0.091 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,  -2,  -1 },    102.92058,   0.00002,     0.078 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,  -1,  -1 },    102.92084,   0.00032,     2.212 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,   0,  -1 },    282.92880,   0.00011,     0.073 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },    267.17887,   0.00001,     0.077 },
 { {   0,  22, -20,  -8,   0,   0,   0,   0,  -2,   1,   0 },    135.54913,   0.00005,    38.025 },
 { {   0,  22, -20,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.74858,   0.00001,     0.109 },
 { {   0,  22, -20,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.74858,   0.00001,     0.058 },
 { {   0,  22, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.43805,   0.00001,     0.399 },
 { {   0,  22, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     25.56607,   0.00002,     0.025 },
 { {   0,  22, -19,  -6,   0,   0,   0,   0,   0,   1,  -2 },    213.12571,   0.00001,    55.980 },
 { {   0,  22, -18,   0,   0,   0,   0,   0,   2,  -2,   0 },    127.99438,   0.00001,     0.063 },
 { {   0,  22, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    127.99438,   0.00001,     0.024 },
 { {   0,  22, -14, -16,   0,   0,   0,   0,   0,  -1,   0 },     64.72480,   0.00005,   838.533 },
 { {   0,  23, -36,  19,   0,   0,   0,   0,  -2,   1,   0 },    146.74605,   0.00001,   137.056 },
 { {   0,  23, -33,  15,   0,   0,   0,   0,  -1,   0,   0 },    252.22554,   0.00002,   155.060 },
 { {   0,  23, -31,  13,   0,   0,   0,   0,   0,  -1,   0 },    268.39259,   0.00002,    23.163 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    236.64860,   0.00004,     0.065 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },    143.97142,   0.00003,     0.477 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.64860,   0.00004,     0.090 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,   0,   0,   0 },     57.26002,   0.00002,     0.107 },
 { {   0,  23, -28,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.00194,   0.00003,     0.066 },
 { {   0,  23, -28,   0,   1,   0,   0,   0,  -2,   1,   0 },     50.00194,   0.00002,     0.497 },
 { {   0,  23, -28,   2,   0,   0,   0,   0,  -2,   1,   0 },     94.95220,   0.00001,     0.970 },
 { {   0,  23, -28,   3,   0,   0,   0,   0,  -2,   1,   0 },    108.22457,   0.00003,     2.001 },
 { {   0,  23, -28,   4,   0,   0,   0,   0,  -2,   1,   0 },     39.49929,   0.00001,    31.219 },
 { {   0,  23, -27,   0,  -1,   0,   0,   0,  -2,   0,   0 },     60.01536,   0.00001,     0.069 },
 { {   0,  23, -27,   0,  -1,   0,   0,   0,  -2,   2,   0 },    240.01536,   0.00001,     0.083 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -4,   0,   0 },    231.34252,   0.00002,     0.026 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },     56.81643,   0.00003,     0.070 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   0,   2 },    231.34252,   0.00002,     0.080 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },     74.65162,   0.00001,     0.914 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },     41.21608,   0.00001,     0.082 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },     51.34252,   0.00001,     0.349 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,   0,   1,   0 },     51.34252,   0.00001,     0.042 },
 { {   0,  23, -27,   0,   1,   0,   0,   0,  -2,   0,   0 },    218.22468,   0.00002,     0.070 },
 { {   0,  23, -27,   0,   1,   0,   0,   0,  -2,   1,   0 },     38.29573,   0.00003,     0.990 },
 { {   0,  23, -27,   0,   2,   0,   0,   0,  -2,   1,   0 },    205.18086,   0.00001,     1.080 },
 { {   0,  23, -27,   2,   0,   0,   0,   0,  -2,   0,   0 },     28.22856,   0.00005,     0.075 },
 { {   0,  23, -27,   2,   0,   0,   0,   0,  -2,   1,   0 },    302.89215,   0.00004,    31.913 },
 { {   0,  23, -27,   2,   0,   0,   0,   0,  -2,   2,   0 },     28.22856,   0.00005,     0.076 },
 { {   0,  23, -27,   4,   0,   0,   0,   0,  -1,   1,  -1 },     26.01303,   0.00002,    46.014 },
 { {   0,  23, -27,   6,   0,   0,   0,   0,   0,   1,  -2 },    175.11891,   0.00001,    82.442 },
 { {   0,  23, -26,  -2,   0,   0,   0,   0,  -2,   1,   0 },    190.77014,   0.00001,     0.864 },
 { {   0,  23, -26,   0,  -1,   0,   0,   0,  -2,   0,   0 },    243.61488,   0.00003,     0.074 },
 { {   0,  23, -26,   0,  -1,   0,   0,   0,  -2,   1,   0 },     91.42537,   0.00001,     5.586 },
 { {   0,  23, -26,   0,   0,  -2,   0,   0,  -2,   1,   0 },    161.45105,   0.00001,     6.150 },
 { {   0,  23, -26,   0,   0,  -1,   0,   0,  -2,   1,   0 },    244.69455,   0.00003,     7.773 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     65.70841,   0.00008,     0.075 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },     71.69479,   0.00055,    10.559 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },     85.31386,   0.00005,     0.076 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,   2,   0,   0 },     83.06564,   0.00001,     0.028 },
 { {   0,  23, -26,   0,   0,   2,   0,   0,  -2,   1,   0 },    347.15894,   0.00003,    37.303 },
 { {   0,  23, -26,   0,   0,   3,   0,   0,  -2,   1,   0 },    237.03584,   0.00003,   140.052 },
 { {   0,  23, -26,   0,   1,   0,   0,   0,  -2,   0,   0 },    272.16512,   0.00007,     0.075 },
 { {   0,  23, -26,   0,   1,   0,   0,   0,  -2,   1,   0 },    214.48033,   0.00006,    96.157 },
 { {   0,  23, -26,   0,   1,   0,   0,   0,  -2,   2,   0 },    165.16452,   0.00003,     0.075 },
 { {   0,  23, -26,   0,   2,   0,   0,   0,  -2,   1,   0 },     84.59294,   0.00003,    13.531 },
 { {   0,  23, -26,   0,   2,   0,   0,   0,  -2,   2,   0 },     79.50844,   0.00003,     0.075 },
 { {   0,  23, -26,   0,   3,   0,   0,   0,  -2,   1,   0 },    184.75191,   0.00002,     6.321 },
 { {   0,  23, -26,   2,   0,   0,   0,   0,  -1,   1,  -1 },    304.59995,   0.00001,    11.751 },
 { {   0,  23, -26,   4,   0,   0,   0,   0,   0,  -1,   0 },     30.44894,   0.00001,     3.876 },
 { {   0,  23, -25,  -4,   0,   0,   0,   0,  -3,  -1,   2 },    353.14913,   0.00001,     0.075 },
 { {   0,  23, -25,  -4,   0,   0,   0,   0,  -3,   0,   2 },    353.14913,   0.00019,  1088.552 },
 { {   0,  23, -25,  -4,   0,   0,   0,   0,  -3,   1,   2 },    353.14913,   0.00001,     0.075 },
 { {   0,  23, -25,   0,  -2,   0,   0,   0,  -2,   1,   0 },     51.41451,   0.00001,     1.357 },
 { {   0,  23, -25,   0,  -2,   0,   0,   0,  -1,  -1,   1 },    342.74520,   0.00003,    60.846 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,  -1,   0 },    168.18245,   0.00002,     0.026 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   0,   0 },    168.18245,   0.00014,     0.040 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   1,   0 },    168.18245,   0.00014,     0.087 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    171.20588,   0.00001,     0.081 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    171.20588,   0.00001,     0.071 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -2,   0 },    168.18245,   0.00005,     0.038 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -1,   0 },    168.18245,   0.00069,     0.076 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -1,   1 },    346.82564,   0.00002,     5.404 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,  -2 },    348.18245,   0.00002,     0.037 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,  -1 },    166.79698,   0.00004,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,   0 },    168.18245,   0.01249,    54.684 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,   2 },    348.18245,   0.00002,     0.037 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.87428,   0.00030,     6.735 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,   0 },    168.18245,   0.00069,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   2,  -1 },    166.87469,   0.00002,     0.076 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   2,   0 },    168.18245,   0.00005,     0.038 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.43526,   0.00006,     1.151 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    352.91650,   0.00006,     0.081 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,  -1,   0 },    168.18245,   0.00014,     0.087 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,   0,   0 },    168.18245,   0.00014,     0.040 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,   1,   0 },    168.18245,   0.00002,     0.026 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   2,   0,   0 },    358.39837,   0.00001,     0.027 },
 { {   0,  23, -25,   0,   1,   0,   0,   0,  -2,   0,   0 },     90.90554,   0.00001,     0.082 },
 { {   0,  23, -25,   0,   1,   0,   0,   0,  -2,   1,   0 },    269.86119,   0.00003,     1.011 },
 { {   0,  23, -25,   0,   1,   0,   0,   0,  -2,   2,   0 },    270.90554,   0.00001,     0.070 },
 { {   0,  23, -25,   0,   1,   0,   0,   0,  -1,   0,   0 },     90.90554,   0.00001,     9.748 },
 { {   0,  23, -25,   2,   0,   0,   0,   0,   0,  -1,   0 },    355.03871,   0.00003,     5.138 },
 { {   0,  23, -24,  -3,   0,   0,   0,   0,  -2,   1,   0 },    132.52370,   0.00002,     3.223 },
 { {   0,  23, -24,   0,  -2,   0,   0,   0,   0,  -2,   0 },    320.64855,   0.00001,     0.075 },
 { {   0,  23, -24,   0,  -2,   0,   0,   0,   0,  -1,   0 },     13.68162,   0.00005,    26.788 },
 { {   0,  23, -24,   0,  -2,   0,   0,   0,   0,   0,   0 },    140.64855,   0.00001,     0.076 },
 { {   0,  23, -24,   0,  -1,   0,   0,   0,   0,  -2,   0 },    205.73844,   0.00002,     0.076 },
 { {   0,  23, -24,   0,  -1,   0,   0,   0,   0,  -1,   0 },    253.37584,   0.00003,    21.289 },
 { {   0,  23, -24,   0,  -1,   0,   0,   0,   0,   0,   0 },    163.82986,   0.00006,     0.075 },
 { {   0,  23, -24,   0,   0,  -3,   0,   0,   0,  -1,   0 },     22.45596,   0.00002,    33.975 },
 { {   0,  23, -24,   0,   0,  -1,   0,   0,   0,  -1,   0 },    354.69049,   0.00006,    10.274 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,  -1,   0 },    272.32584,   0.00008,     0.041 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   0,  -1 },    249.19500,   0.00003,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.23487,   0.00020,     0.088 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   1,  -1 },    249.19500,   0.00003,     0.087 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    271.75335,   0.00001,     0.525 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.98073,   0.00001,     0.982 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -1,   1,  -1 },     69.11403,   0.00002,     1.174 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -3,   0 },    271.61478,   0.00002,     0.038 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.61544,   0.00036,     0.076 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,  -2 },     92.28110,   0.00001,     0.037 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,  -1 },    249.19500,   0.00014,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    272.28110,   0.00629,     7.618 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,   2 },     92.28110,   0.00001,     0.037 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,  -1 },    249.19500,   0.00251,    28.187 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,   0 },     93.66294,   0.00038,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,  -2 },    272.15908,   0.00009,     4.945 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,  -1 },    249.19500,   0.00014,     0.076 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,   0 },     93.53059,   0.00003,     0.038 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   1,  -1,  -1 },     69.14894,   0.00003,     1.084 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -2,   0 },    272.98755,   0.00007,     0.086 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -1,  -1 },    249.19500,   0.00003,     0.087 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.22276,   0.00005,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,   0,  -1 },    249.19500,   0.00003,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,   0,   0 },     92.22167,   0.00009,     0.026 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,  -2,   0,   0 },    359.01342,   0.00001,     0.089 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,  -2,   1,   0 },    179.01342,   0.00001,     0.503 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,  -2,   2,   0 },    179.01342,   0.00001,     0.066 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,   0,  -2,   0 },    357.28149,   0.00001,     0.077 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,   0,  -1,   0 },     29.91054,   0.00002,     4.639 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,   0,   0,   0 },    357.28149,   0.00001,     0.074 },
 { {   0,  23, -24,   0,   1,   0,   0,   0,   0,   1,  -2 },    178.39309,   0.00001,     8.480 },
 { {   0,  23, -24,   0,   2,   0,   0,   0,   0,  -1,   0 },    227.62696,   0.00004,     3.335 },
 { {   0,  23, -23,  -2,   0,   0,   0,   0,   0,  -1,   0 },    214.36848,   0.00002,    14.726 },
 { {   0,  23, -23,   0,  -1,   0,   0,   0,   1,  -1,  -1 },     33.50446,   0.00010,   147.246 },
 { {   0,  23, -23,   0,   0,  -2,   0,   0,   1,  -1,  -1 },    163.77054,   0.00001,   103.989 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    171.93451,   0.00008,     0.097 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.48128,   0.00004,     0.062 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    192.75345,   0.00005,     0.041 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.22752,   0.00011,     0.088 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,  -2,   0 },    190.86690,   0.00001,     0.082 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.40950,   0.00019,     0.884 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   0,   0 },      1.25032,   0.00011,     0.070 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   1,   0 },    359.59068,   0.00002,     0.036 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   2,   0 },    179.58667,   0.00002,     0.024 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -3,  -1 },    192.75653,   0.00002,     0.038 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -2,  -1 },    192.75686,   0.00024,     0.076 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -2,   0 },    359.86167,   0.00001,     4.094 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,  -2 },    359.67965,   0.00002,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,  -1 },    192.75345,   0.00443,    12.901 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -2 },    359.67965,   0.00030,    11.205 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -1 },     16.14215,   0.00008,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   1,  -2 },    359.67965,   0.00002,     0.076 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,  -3,   0 },      0.00509,   0.00001,     1.556 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.59066,   0.00002,     1.024 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.43648,   0.00003,     0.039 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,   0,   0 },    359.58949,   0.00013,     0.026 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,   1,   0 },    359.59068,   0.00002,     0.019 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   3,  -2,  -1 },    192.74996,   0.00005,     0.087 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   3,  -1,  -1 },    192.75345,   0.00005,     0.040 },
 { {   0,  23, -22,  -6,   0,   0,   0,   0,  -1,  -1,   1 },    287.01482,   0.00002,   197.578 },
 { {   0,  23, -22,  -4,   0,   0,   0,   0,   0,  -1,   0 },    217.19059,   0.00016,   220.195 },
 { {   0,  23, -22,  -2,   0,   0,   0,   0,   1,  -1,  -1 },    281.83276,   0.00003,    70.700 },
 { {   0,  23, -22,   0,  -1,   0,   0,   0,   2,  -1,  -2 },    180.75207,   0.00001,    16.514 },
 { {   0,  23, -22,   0,   0,  -1,   0,   0,   2,  -1,  -2 },    180.65569,   0.00001,    98.033 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   0,  -1,  -2 },    274.98030,   0.00004,     0.040 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   0,   0,  -2 },    274.98030,   0.00003,     0.087 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   1,  -1,  -1 },    103.43098,   0.00004,     0.928 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   1,   0,  -1 },    282.88414,   0.00004,     0.070 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -3,  -2 },    274.98030,   0.00001,     0.038 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -2,  -2 },    274.98126,   0.00017,     0.076 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -1,  -2 },    274.98030,   0.00305,    42.111 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,   0,  -2 },    276.69207,   0.00011,     0.075 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,   1,  -2 },    274.98030,   0.00001,     0.038 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   4,  -2,  -2 },    274.98030,   0.00003,     0.087 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   4,  -1,  -2 },    274.98030,   0.00004,     0.040 },
 { {   0,  23, -21,  -6,   0,   0,   0,   0,   0,  -1,   0 },     49.53566,   0.00003,    16.999 },
 { {   0,  23, -21,  -2,   0,   0,   0,   0,   2,  -1,  -2 },    179.78697,   0.00001,    25.239 },
 { {   0,  23, -21,   0,   0,   0,   0,   0,   2,   0,   0 },     20.00646,   0.00001,     0.024 },
 { {   0,  23, -19, -13,   0,   0,   0,   0,  -2,   1,   0 },     95.35390,   0.00005,   151.533 },
 { {   0,  23, -19, -11,   0,   0,   0,   0,  -1,   1,  -1 },    328.06014,   0.00002,   332.942 },
 { {   0,  24, -30,   4,   0,   0,   0,   0,  -2,   1,   0 },    291.62825,   0.00003,     2.920 },
 { {   0,  24, -30,   6,   0,   0,   0,   0,  -1,  -1,   1 },    347.03519,   0.00009,  1486.561 },
 { {   0,  24, -30,   8,   0,   0,   0,   0,   0,  -1,   0 },    107.93235,   0.00003,    97.320 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    318.18570,   0.00003,     0.068 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    143.95409,   0.00002,     0.681 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,  -2,   2,   0 },    233.47713,   0.00001,     0.085 },
 { {   0,  24, -29,   0,   1,   0,   0,   0,  -2,   0,   0 },     49.95990,   0.00003,     0.068 },
 { {   0,  24, -29,   0,   1,   0,   0,   0,  -2,   1,   0 },    229.91029,   0.00003,     0.722 },
 { {   0,  24, -29,   2,   0,   0,   0,   0,  -2,   0,   0 },     94.91015,   0.00001,     0.073 },
 { {   0,  24, -29,   2,   0,   0,   0,   0,  -2,   1,   0 },    115.17334,   0.00001,     2.464 },
 { {   0,  24, -29,   2,   0,   0,   0,   0,  -2,   2,   0 },     94.91015,   0.00001,     0.078 },
 { {   0,  24, -29,   6,   0,   0,   0,   0,   0,  -1,   0 },    115.40239,   0.00001,    18.835 },
 { {   0,  24, -28,   0,  -1,   0,   0,   0,  -2,   1,   0 },     21.70337,   0.00002,     1.807 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },     97.74809,   0.00001,     0.073 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },     42.48034,   0.00001,     2.131 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },     97.74809,   0.00001,     0.078 },
 { {   0,  24, -28,   0,   1,   0,   0,   0,  -2,   0,   0 },     38.24715,   0.00006,     0.073 },
 { {   0,  24, -28,   0,   2,   0,   0,   0,  -2,   1,   0 },    290.92554,   0.00003,     3.327 },
 { {   0,  24, -28,   0,   2,   0,   0,   0,  -2,   2,   0 },    205.72722,   0.00002,     0.077 },
 { {   0,  24, -27,  -1,   0,   0,   0,   0,  -2,   0,   0 },    218.20623,   0.00002,     0.075 },
 { {   0,  24, -27,  -1,   0,   0,   0,   0,  -2,   1,   0 },    218.20623,   0.00028,  1155.779 },
 { {   0,  24, -27,  -1,   0,   0,   0,   0,  -2,   2,   0 },    218.20623,   0.00002,     0.075 },
 { {   0,  24, -27,   0,  -2,   0,   0,   0,  -2,   1,   0 },    356.00863,   0.00001,     2.761 },
 { {   0,  24, -27,   0,  -1,   0,   0,   0,  -2,   0,   0 },    234.92349,   0.00003,     0.078 },
 { {   0,  24, -27,   0,  -1,   0,   0,   0,  -2,   1,   0 },    238.24589,   0.00003,     2.240 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },     75.28694,   0.00002,     1.884 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.23809,   0.00001,     0.804 },
 { {   0,  24, -27,   0,   2,   0,   0,   0,  -1,  -1,   1 },    128.32489,   0.00002,    48.022 },
 { {   0,  24, -27,   3,   0,   0,   0,   0,   0,   1,  -2 },    147.55131,   0.00002,    54.525 },
 { {   0,  24, -26,  -4,   0,   0,   0,   0,  -2,   1,   0 },    173.26326,   0.00001,     1.678 },
 { {   0,  24, -26,  -3,   0,   0,   0,   0,  -2,   0,   0 },    209.67391,   0.00002,     0.075 },
 { {   0,  24, -26,  -3,   0,   0,   0,   0,  -2,   1,   0 },     12.01906,   0.00007,    15.568 },
 { {   0,  24, -26,  -3,   0,   0,   0,   0,  -2,   2,   0 },    209.67391,   0.00002,     0.076 },
 { {   0,  24, -26,   0,  -1,   0,   0,   0,   0,  -1,   0 },    190.30690,   0.00001,     3.053 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    164.47616,   0.00004,     0.085 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    145.95644,   0.00001,     0.068 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.65002,   0.00002,     2.096 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },    352.76050,   0.00005,     0.074 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.13490,   0.00055,     4.112 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    351.57139,   0.00017,     0.077 },
 { {   0,  24, -26,   0,   1,   0,   0,   0,   0,  -2,   0 },     90.90931,   0.00001,     0.075 },
 { {   0,  24, -26,   0,   1,   0,   0,   0,   0,  -1,   0 },    214.34367,   0.00004,     6.294 },
 { {   0,  24, -26,   0,   1,   0,   0,   0,   0,   0,   0 },    270.90931,   0.00001,     0.076 },
 { {   0,  24, -26,   0,   2,   0,   0,   0,   0,  -1,   0 },    301.43826,   0.00003,    13.407 },
 { {   0,  24, -26,   0,   3,   0,   0,   0,   0,  -2,   0 },    155.46496,   0.00001,     0.075 },
 { {   0,  24, -26,   0,   3,   0,   0,   0,   0,  -1,   0 },    155.46496,   0.00019,   102.937 },
 { {   0,  24, -26,   0,   3,   0,   0,   0,   0,   0,   0 },    155.46496,   0.00001,     0.075 },
 { {   0,  24, -25,  -5,   0,   0,   0,   0,  -2,   1,   0 },    352.93979,   0.00001,     7.837 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.11520,   0.00003,     0.093 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.65702,   0.00015,     1.321 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,   0,   0 },     92.07652,   0.00009,     0.071 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,   1,  -2 },    272.14532,   0.00001,     2.362 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,  -2,  -1 },     68.28310,   0.00001,     0.074 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,  -2,   0 },     92.27462,   0.00001,     7.680 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,  -1,  -1 },     68.28288,   0.00020,     3.367 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,   0,  -1 },     68.28289,   0.00007,     0.077 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   2,  -1,   0 },     92.20468,   0.00002,     0.039 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   2,   0,   0 },     92.20433,   0.00007,     0.026 },
 { {   0,  24, -25,   0,   1,   0,   0,   0,   0,  -1,   0 },    178.97611,   0.00001,     1.189 },
 { {   0,  24, -24,  -7,   0,   0,   0,   0,  -2,   1,   0 },    143.72223,   0.00001,     5.236 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.49373,   0.00007,     0.103 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.38178,   0.00001,     0.060 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -4,   0 },    179.43126,   0.00001,     0.026 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -3,   0 },    179.43126,   0.00012,     0.040 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },      0.76840,   0.00001,     0.087 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.58933,   0.00006,     0.569 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    359.87784,   0.00007,     0.067 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,   1,   0 },    359.54503,   0.00001,     0.035 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.62043,   0.00004,     1.422 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -5,   0 },    179.42979,   0.00004,     0.038 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -4,   0 },    179.42976,   0.00056,     0.075 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,  -2 },    359.43126,   0.00002,     0.037 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,   0 },    179.43126,   0.01017,    58.088 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,   2 },    359.43126,   0.00002,     0.037 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.48665,   0.00106,     0.076 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.55185,   0.00011,     2.851 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.48244,   0.00007,     0.038 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,   0,   0 },    359.54378,   0.00010,     0.025 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,   1,   0 },    359.54502,   0.00001,     0.019 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -4,   0 },    179.43276,   0.00011,     0.087 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -3,   0 },    179.43126,   0.00012,     0.040 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.43032,   0.00001,     0.026 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,   1,  -1,  -1 },    102.83859,   0.00001,     0.587 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,   2,   0,   0 },    273.95941,   0.00002,     0.025 },
 { {   0,  24, -22,  -7,   0,   0,   0,   0,   0,  -1,   0 },    259.07143,   0.00003,    28.562 },
 { {   0,  24, -22,   0,   0,   0,   0,   0,   2,   0,   0 },     20.46127,   0.00001,     0.024 },
 { {   0,  24, -21,  -9,   0,   0,   0,   0,   0,  -1,   0 },    260.39049,   0.00002,    35.266 },
 { {   0,  25, -35,  11,   0,   0,   0,   0,  -2,   1,   0 },    125.39758,   0.00002,   205.390 },
 { {   0,  25, -33,   7,   0,   0,   0,   0,  -2,   1,   0 },    316.69450,   0.00002,     8.206 },
 { {   0,  25, -31,   2,   0,   0,   0,   0,  -2,   1,   0 },    231.24439,   0.00003,     1.282 },
 { {   0,  25, -31,   3,   0,   0,   0,   0,  -2,   1,   0 },    132.60731,   0.00001,     4.023 },
 { {   0,  25, -31,   4,   0,   0,   0,   0,  -2,   1,   0 },    291.59064,   0.00003,     3.533 },
 { {   0,  25, -30,   0,  -1,   0,   0,   0,  -2,   1,   0 },    152.02113,   0.00001,     1.078 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    293.15553,   0.00003,     0.071 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   2,   0 },    246.80950,   0.00002,     0.081 },
 { {   0,  25, -30,   0,   1,   0,   0,   0,  -2,   0,   0 },    229.86582,   0.00005,     0.071 },
 { {   0,  25, -30,   1,   0,   0,   0,   0,  -2,   1,   0 },     50.72670,   0.00001,     3.205 },
 { {   0,  25, -30,   2,   0,   0,   0,   0,  -2,   0,   0 },    274.86663,   0.00002,     0.077 },
 { {   0,  25, -30,   2,   0,   0,   0,   0,  -2,   1,   0 },     88.33759,   0.00002,     4.552 },
 { {   0,  25, -30,   2,   0,   0,   0,   0,  -2,   2,   0 },    274.86663,   0.00002,     0.074 },
 { {   0,  25, -30,   5,   0,   0,   0,   0,   0,  -1,   0 },     74.25677,   0.00003,    24.543 },
 { {   0,  25, -29,   0,  -3,   0,   0,   0,  -2,   1,   0 },    137.89015,   0.00001,    10.355 },
 { {   0,  25, -29,   0,  -2,   0,   0,   0,  -2,   0,   0 },    312.14482,   0.00006,     0.075 },
 { {   0,  25, -29,   0,  -2,   0,   0,   0,  -2,   1,   0 },    218.68261,   0.00003,    81.524 },
 { {   0,  25, -29,   0,  -2,   0,   0,   0,  -2,   2,   0 },     66.59932,   0.00001,     0.076 },
 { {   0,  25, -29,   0,  -1,   0,   0,   0,  -2,   0,   0 },    186.99279,   0.00003,     0.076 },
 { {   0,  25, -29,   0,  -1,   0,   0,   0,  -2,   1,   0 },    192.25738,   0.00002,    13.882 },
 { {   0,  25, -29,   0,  -1,   0,   0,   0,  -2,   2,   0 },    224.16544,   0.00002,     0.075 },
 { {   0,  25, -29,   0,   0,  -2,   0,   0,  -2,   1,   0 },    245.98700,   0.00001,    11.307 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,  -1,   0 },     51.24951,   0.00001,     0.038 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   0,   2 },     51.24951,   0.00002,     0.073 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    105.49336,   0.00003,     6.396 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   2,   0 },    228.23362,   0.00002,     0.075 },
 { {   0,  25, -29,   0,   0,   1,   0,   0,  -2,   1,   0 },    139.02138,   0.00001,     5.255 },
 { {   0,  25, -29,   0,   1,   0,   0,   0,  -2,   0,   0 },    220.29406,   0.00002,     0.077 },
 { {   0,  25, -29,   0,   1,   0,   0,   0,  -2,   1,   0 },     82.67868,   0.00002,     4.156 },
 { {   0,  25, -29,   3,   0,   0,   0,   0,   0,  -2,   0 },    358.07189,   0.00001,     0.075 },
 { {   0,  25, -29,   3,   0,   0,   0,   0,   0,  -1,   0 },    284.28691,   0.00001,    44.205 },
 { {   0,  25, -29,   3,   0,   0,   0,   0,   0,   0,   0 },    358.07189,   0.00001,     0.076 },
 { {   0,  25, -28,  -2,   0,   0,   0,   0,  -2,   0,   0 },    190.52767,   0.00002,     0.076 },
 { {   0,  25, -28,  -2,   0,   0,   0,   0,  -2,   2,   0 },    190.52767,   0.00002,     0.075 },
 { {   0,  25, -28,   0,  -1,   0,   0,   0,  -1,   1,  -1 },      9.23648,   0.00002,    54.729 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -1,   0,   0 },    253.20580,   0.00001,     3.713 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -1,   1,  -1 },    253.36858,   0.00013,     9.749 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,  -1,   0 },     43.01180,   0.00001,     1.619 },
 { {   0,  25, -28,   0,   1,   0,   0,   0,   0,  -1,   0 },    283.66296,   0.00001,     1.875 },
 { {   0,  25, -27,  -4,   0,   0,   0,   0,  -2,   0,   0 },    173.13905,   0.00002,     0.076 },
 { {   0,  25, -27,  -4,   0,   0,   0,   0,  -2,   2,   0 },    173.13905,   0.00002,     0.075 },
 { {   0,  25, -27,   0,  -4,   0,   0,   0,   0,  -1,   0 },     72.96086,   0.00001,    22.170 },
 { {   0,  25, -27,   0,  -3,   0,   0,   0,   0,  -1,   0 },     62.26536,   0.00001,     7.727 },
 { {   0,  25, -27,   0,  -1,   0,   0,   0,   0,  -1,   0 },    190.53926,   0.00001,     3.356 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    176.16032,   0.00002,     0.065 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,  -2,   0 },    179.51764,   0.00004,     0.078 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },    353.09459,   0.00015,     2.616 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   0,  -2 },    356.31599,   0.00002,     0.076 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   0,  -1 },    346.36652,   0.00002,     4.639 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   0,   0 },    177.21198,   0.00010,     0.073 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   1,  -2 },    356.31599,   0.00034,    20.488 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   2,  -2 },    356.31599,   0.00002,     0.075 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.59911,   0.00002,     1.489 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   2,  -2,   0 },    357.88212,   0.00001,     0.084 },
 { {   0,  25, -27,   0,   1,   0,   0,   0,   0,  -1,   0 },    270.63228,   0.00003,     2.143 },
 { {   0,  25, -26,  -2,   0,   0,   0,   0,   0,   1,  -2 },    124.38911,   0.00001,    68.695 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     91.62336,   0.00002,     0.099 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.62336,   0.00002,     0.061 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -1,   1,  -3 },     67.08515,   0.00001,     0.040 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -1,   2,  -3 },     67.08515,   0.00001,     0.087 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },     91.56907,   0.00001,     0.084 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },    276.85499,   0.00002,     0.723 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,   0,   0 },     91.94076,   0.00005,     0.068 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,  -1,  -1 },     67.33934,   0.00012,     3.044 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   0,  -3 },     67.08515,   0.00005,     0.075 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   0,  -2 },     92.29347,   0.00003,     6.180 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   0,  -1 },    247.18588,   0.00003,     0.074 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   1,  -3 },     67.08515,   0.00094,   201.839 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   2,  -3 },     67.08515,   0.00005,     0.075 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   2,  -1,  -2 },    272.18568,   0.00001,     1.379 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   2,  -1,   0 },     92.18605,   0.00002,     0.038 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   2,   0,   0 },     92.18570,   0.00005,     0.025 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   3,   0,  -3 },     67.08515,   0.00001,     0.087 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   3,   1,  -3 },     67.08515,   0.00001,     0.040 },
 { {   0,  25, -25,  -8,   0,   0,   0,   0,  -2,   1,   0 },    115.83891,   0.00001,    10.295 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.44774,   0.00006,     0.110 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   0,  -2,   0 },    359.55143,   0.00003,     0.092 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    359.88013,   0.00005,     0.064 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.67918,   0.00001,     0.072 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,  -1,  -2 },    179.50544,   0.00014,     3.639 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.48910,   0.00001,     0.037 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,   0,   0 },    359.49691,   0.00008,     0.025 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,   2,   0 },    359.50180,   0.00001,     0.015 },
 { {   0,  25, -24,  -3,   0,   0,   0,   0,   2,  -3,   0 },    123.09736,   0.00001,    75.446 },
 { {   0,  25, -24,   0,   0,   0,   0,   0,   2,   0,   0 },    274.79987,   0.00001,     0.024 },
 { {   0,  25, -21, -15,   0,   0,   0,   0,  -2,   1,   0 },    207.92314,   0.00003,     5.523 },
 { {   0,  25, -21, -12,   0,   0,   0,   0,   0,  -1,   0 },    302.10364,   0.00005,   473.049 },
 { {   0,  25, -18, -21,   0,   0,   0,   0,  -2,   1,   0 },    301.59579,   0.00001,   110.782 },
 { {   0,  26, -37,  15,   0,   0,   0,   0,   0,  -1,   0 },    308.09385,   0.00001,    59.214 },
 { {   0,  26, -36,  10,   0,   0,   0,   0,  -2,   1,   0 },    167.15205,   0.00001,    10.131 },
 { {   0,  26, -35,   8,   0,   0,   0,   0,  -2,   1,   0 },     65.16050,   0.00003,    28.298 },
 { {   0,  26, -34,   6,   0,   0,   0,   0,  -2,   0,   0 },    109.50776,   0.00001,     0.075 },
 { {   0,  26, -34,   6,   0,   0,   0,   0,  -2,   2,   0 },    109.50776,   0.00001,     0.076 },
 { {   0,  26, -33,   4,   0,   0,   0,   0,  -2,   0,   0 },     80.74875,   0.00002,     0.075 },
 { {   0,  26, -33,   4,   0,   0,   0,   0,  -2,   1,   0 },    291.65188,   0.00001,    10.941 },
 { {   0,  26, -33,   4,   0,   0,   0,   0,  -2,   2,   0 },     80.74875,   0.00002,     0.076 },
 { {   0,  26, -32,   0,   1,   0,   0,   0,  -2,   1,   0 },    142.00107,   0.00001,     0.882 },
 { {   0,  26, -32,   2,   0,   0,   0,   0,  -2,   0,   0 },    231.19979,   0.00004,     0.075 },
 { {   0,  26, -32,   2,   0,   0,   0,   0,  -2,   1,   0 },    208.74074,   0.00004,     6.461 },
 { {   0,  26, -32,   2,   0,   0,   0,   0,  -2,   2,   0 },    231.19979,   0.00004,     0.076 },
 { {   0,  26, -32,   3,   0,   0,   0,   0,  -2,   1,   0 },     20.15985,   0.00002,     2.653 },
 { {   0,  26, -32,   4,   0,   0,   0,   0,  -1,   0,   0 },    109.64501,   0.00001,    46.311 },
 { {   0,  26, -31,   0,  -1,   0,   0,   0,  -2,   0,   0 },    152.01448,   0.00002,     0.074 },
 { {   0,  26, -31,   0,  -1,   0,   0,   0,  -2,   1,   0 },    163.44569,   0.00002,     3.307 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,  -2,   2 },    143.91561,   0.00002,     0.076 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   0,   2 },    143.91561,   0.00002,     0.075 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   1,   0 },    140.10012,   0.00001,     4.584 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   2,   0 },    150.46288,   0.00002,     0.077 },
 { {   0,  26, -31,   0,   0,   2,   0,   0,  -2,   1,   0 },    312.59881,   0.00003,     6.656 },
 { {   0,  26, -31,   0,   1,   0,   0,   0,  -2,   0,   0 },    109.35886,   0.00007,     0.075 },
 { {   0,  26, -31,   0,   3,   0,   0,   0,  -2,   0,   0 },    286.32040,   0.00002,     0.076 },
 { {   0,  26, -31,   0,   3,   0,   0,   0,  -2,   1,   0 },    264.64932,   0.00005,    28.754 },
 { {   0,  26, -31,   1,   0,   0,   0,   0,  -2,   1,   0 },     50.68176,   0.00001,     3.189 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,  -1,   0 },     21.86259,   0.00002,     0.039 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },     33.93243,   0.00002,     0.080 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   1,   0 },     51.20022,   0.00003,     1.279 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.00643,   0.00001,     0.071 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -1,  -1,   1 },    158.66232,   0.00001,    16.230 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -1,   0,   0 },    336.42869,   0.00002,     9.511 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   0,  -2,   0 },    357.93767,   0.00003,     0.070 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   0,   0,   0 },    177.93767,   0.00004,     0.082 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   2,  -2,   0 },    177.93767,   0.00005,     0.095 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   2,  -1,   0 },    177.93767,   0.00008,     0.042 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   2,   0,   0 },    177.93767,   0.00001,     0.027 },
 { {   0,  26, -30,   2,   0,   0,   0,   0,   0,  -1,   0 },     71.12624,   0.00003,    14.042 },
 { {   0,  26, -29,  -4,   0,   0,   0,   0,  -2,  -1,   2 },    265.15296,   0.00003,    88.047 },
 { {   0,  26, -29,  -4,   0,   0,   0,   0,  -2,   1,   0 },    265.30443,   0.00001,     2.900 },
 { {   0,  26, -29,  -3,   0,   0,   0,   0,  -2,   0,   0 },    174.80303,   0.00001,     0.077 },
 { {   0,  26, -29,  -3,   0,   0,   0,   0,  -2,   1,   0 },    125.73411,   0.00004,     5.353 },
 { {   0,  26, -29,  -3,   0,   0,   0,   0,  -2,   2,   0 },    174.80303,   0.00001,     0.074 },
 { {   0,  26, -29,  -2,   0,   0,   0,   0,  -2,   1,   0 },    190.63963,   0.00001,     1.392 },
 { {   0,  26, -29,  -2,   0,   0,   0,   0,  -1,  -1,   1 },    138.15485,   0.00002,   569.866 },
 { {   0,  26, -29,   0,  -1,   0,   0,   0,   0,  -1,   0 },    154.67102,   0.00002,    13.080 },
 { {   0,  26, -29,   0,   0,  -1,   0,   0,   0,  -1,   0 },    141.06638,   0.00001,    38.315 },
 { {   0,  26, -29,   0,   0,   0,  -1,   0,   0,  -1,   0 },    270.16505,   0.00001,   246.676 },
 { {   0,  26, -29,   0,   0,   0,   0,  -2,   0,  -1,   0 },    173.11127,   0.00003,   233.097 },
 { {   0,  26, -29,   0,   0,   0,   0,  -1,   0,  -1,   0 },    308.27387,   0.00003,   562.132 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -4,  -1,   0 },     75.00074,   0.00001,     0.020 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -4,   0,   0 },     75.00074,   0.00003,     0.028 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -4,   1,   0 },     75.00074,   0.00002,     0.044 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -3,   0 },     75.00074,   0.00001,     0.020 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -2,   0 },     75.00074,   0.00015,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -1,   0 },     74.98814,   0.00121,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.77825,   0.00131,     0.087 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -1,  -1,   0 },    255.00074,   0.00003,     0.081 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -4,   0 },     75.00074,   0.00003,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -3,   0 },     75.04732,   0.00039,     0.038 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,  -2 },    255.00074,   0.00003,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   0 },     75.04726,   0.00579,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   2 },    255.00074,   0.00001,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,  -2 },    255.00074,   0.00021,     0.037 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   0 },     75.00074,   0.10489,   127.422 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   2 },    255.00074,   0.00021,     0.037 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,  -2 },    255.00074,   0.00001,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,  -1 },     72.64665,   0.00001,     6.293 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   0 },     75.28520,   0.00505,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   2 },    255.00074,   0.00003,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   1,   0 },     75.28512,   0.00034,     0.038 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   2,   0 },     75.00074,   0.00003,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   1,  -1,   0 },    255.00074,   0.00003,     0.081 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -2,   0 },     74.95368,   0.00116,     0.087 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -1,   0 },     75.01355,   0.00120,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,   0,   0 },     75.00074,   0.00015,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,   1,   0 },     75.00074,   0.00001,     0.020 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   4,  -3,   0 },     75.00074,   0.00002,     0.044 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   4,  -2,   0 },     75.00074,   0.00003,     0.028 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   4,  -1,   0 },     75.00074,   0.00001,     0.020 },
 { {   0,  26, -29,   0,   0,   1,   0,   0,   0,  -1,   0 },    336.28701,   0.00003,    23.926 },
 { {   0,  26, -29,   0,   1,   0,   0,   0,   0,  -1,   0 },     14.10252,   0.00003,    10.852 },
 { {   0,  26, -29,   0,   2,   0,   0,   0,   0,  -1,   0 },    271.64168,   0.00001,     5.667 },
 { {   0,  26, -29,   2,   0,   0,   0,   0,   1,  -1,  -1 },     49.89204,   0.00002,    57.304 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -2,  -2,   0 },    332.06381,   0.00001,     0.027 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -2,  -1,   0 },    332.06381,   0.00008,     0.042 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    332.06381,   0.00005,     0.095 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    166.27827,   0.00004,     0.040 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,  -2,   0 },    332.05047,   0.00004,     0.082 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,  -1,   0 },    351.42578,   0.00002,     0.992 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,   0,   0 },    159.94225,   0.00007,     0.070 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -3,  -1 },    166.27827,   0.00001,     0.038 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -2,  -1 },    166.26439,   0.00017,     0.075 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.27827,   0.00314,    21.779 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   0,  -1 },    166.30776,   0.00031,     0.076 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   1,  -1 },    166.27827,   0.00001,     0.038 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   2,  -1,   0 },    152.06381,   0.00001,     0.039 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   2,   0,   0 },    358.96800,   0.00002,     0.026 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   3,  -2,  -1 },    166.29218,   0.00003,     0.087 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   3,  -1,  -1 },    166.27827,   0.00004,     0.041 },
 { {   0,  26, -28,   0,   0,   1,   0,   0,   1,  -1,  -1 },    301.76116,   0.00001,    83.548 },
 { {   0,  26, -28,   0,   1,   0,   0,   0,   1,  -1,  -1 },     75.29450,   0.00002,    26.051 },
 { {   0,  26, -28,   2,   0,   0,   0,   0,   2,  -1,  -2 },    179.50859,   0.00001,    27.538 },
 { {   0,  26, -27,  -7,   0,   0,   0,   0,  -2,   1,   0 },     55.96385,   0.00002,    16.646 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },    272.12403,   0.00001,     0.498 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   0,   0,   0 },     92.08159,   0.00001,     0.066 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   1,  -1,  -1 },     66.31105,   0.00001,     1.048 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -3,   0 },    272.16564,   0.00002,     4.277 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -2,  -2 },    272.16399,   0.00002,     0.075 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -1,  -2 },    272.16399,   0.00044,    10.032 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -1,   0 },     92.16617,   0.00001,     0.037 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,   0,  -2 },    272.16399,   0.00002,     0.076 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,   0,   0 },     92.16582,   0.00004,     0.025 },
 { {   0,  26, -27,   0,   1,   0,   0,   0,   2,  -1,  -2 },    262.65658,   0.00001,    65.026 },
 { {   0,  26, -26,  -9,   0,   0,   0,   0,  -2,   0,   0 },    206.76176,   0.00002,     0.075 },
 { {   0,  26, -26,  -9,   0,   0,   0,   0,  -2,   1,   0 },    319.16551,   0.00007,   303.701 },
 { {   0,  26, -26,  -9,   0,   0,   0,   0,  -2,   2,   0 },    206.76176,   0.00002,     0.075 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.65360,   0.00002,     0.055 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.75270,   0.00001,     0.332 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.44880,   0.00002,     0.036 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   2,   0,   0 },    359.44808,   0.00007,     0.024 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   3,  -2,  -2 },    359.56137,   0.00012,    75.143 },
 { {   0,  26, -25,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.52581,   0.00001,     0.132 },
 { {   0,  26, -25,   0,   0,   0,   0,   0,   2,   0,   0 },    274.79975,   0.00001,     0.024 },
 { {   0,  26, -24,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.59620,   0.00001,     0.151 },
 { {   0,  26, -23, -15,   0,   0,   0,   0,  -2,   1,   0 },    119.99959,   0.00001,     5.171 },
 { {   0,  27, -37,   8,   0,   0,   0,   0,  -2,  -1,   2 },     50.45768,   0.00001,   177.824 },
 { {   0,  27, -36,   7,   0,   0,   0,   0,  -2,   1,   0 },    260.29398,   0.00002,     7.741 },
 { {   0,  27, -35,   5,   0,   0,   0,   0,  -2,   1,   0 },    200.76337,   0.00001,    15.196 },
 { {   0,  27, -34,   2,   0,   0,   0,   0,  -2,   1,   0 },    323.31839,   0.00002,     1.890 },
 { {   0,  27, -34,   3,   0,   0,   0,   0,  -2,   1,   0 },     29.44002,   0.00003,   409.620 },
 { {   0,  27, -33,   0,   0,   0,   0,   0,  -2,   1,   0 },    224.92449,   0.00001,     1.687 },
 { {   0,  27, -33,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.99257,   0.00003,     0.073 },
 { {   0,  27, -33,   0,   2,   0,   0,   0,  -2,   1,   0 },    223.37302,   0.00002,     2.359 },
 { {   0,  27, -33,   0,   3,   0,   0,   0,  -2,  -1,   2 },    219.41841,   0.00002,   161.700 },
 { {   0,  27, -33,   0,   3,   0,   0,   0,  -2,   1,   0 },     41.05198,   0.00003,     2.944 },
 { {   0,  27, -33,   0,   7,   0,   0,   0,  -2,   1,   0 },     26.59010,   0.00004,   403.129 },
 { {   0,  27, -33,   2,   0,   0,   0,   0,  -2,   1,   0 },    198.47589,   0.00002,     2.124 },
 { {   0,  27, -32,   0,  -1,   0,   0,   0,  -2,   1,   0 },    203.45161,   0.00002,     3.095 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.34022,   0.00003,     0.078 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   2,   0 },    117.22769,   0.00006,     0.073 },
 { {   0,  27, -32,   0,   1,   0,   0,   0,  -2,   0,   0 },     49.94357,   0.00002,     0.078 },
 { {   0,  27, -32,   0,   1,   0,   0,   0,  -2,   1,   0 },    109.98005,   0.00003,     2.034 },
 { {   0,  27, -32,   0,   1,   0,   0,   0,  -2,   2,   0 },    229.69058,   0.00003,     0.073 },
 { {   0,  27, -32,   0,   2,   0,   0,   0,  -2,   1,   0 },    111.31242,   0.00002,     1.736 },
 { {   0,  27, -32,   2,   0,   0,   0,   0,   0,  -1,   0 },     94.77534,   0.00001,     3.298 },
 { {   0,  27, -32,   3,   0,   0,   0,   0,   0,  -1,   0 },    287.98387,   0.00001,     4.378 },
 { {   0,  27, -31,  -2,   0,   0,   0,   0,  -2,   1,   0 },    102.55264,   0.00003,     2.907 },
 { {   0,  27, -31,  -1,   0,   0,   0,   0,  -1,   0,   0 },    143.08560,   0.00001,    88.463 },
 { {   0,  27, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    231.14911,   0.00003,     0.084 },
 { {   0,  27, -31,   0,   0,   0,   0,   0,   0,  -1,   0 },    165.11215,   0.00002,     2.728 },
 { {   0,  27, -31,   0,   1,   0,   0,   0,   0,  -1,   0 },     38.06726,   0.00003,     3.542 },
 { {   0,  27, -31,   0,   4,   0,   0,   0,   0,  -1,   0 },    250.53801,   0.00002,    34.001 },
 { {   0,  27, -30,  -4,   0,   0,   0,   0,  -2,   1,   0 },    265.29593,   0.00001,     3.563 },
 { {   0,  27, -30,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.22583,   0.00001,     1.579 },
 { {   0,  27, -30,   0,   0,   0,   0,   0,   1,  -1,  -1 },    254.63911,   0.00001,     2.379 },
 { {   0,  27, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    188.80521,   0.00001,     0.102 },
 { {   0,  27, -29,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.11362,   0.00002,     1.725 },
 { {   0,  27, -29,   0,   0,   0,   0,   0,   2,  -3,   0 },    359.38343,   0.00001,     7.110 },
 { {   0,  27, -29,   0,   0,   0,   0,   0,   2,   0,   0 },    359.14104,   0.00002,     0.025 },
 { {   0,  27, -28,  -5,   0,   0,   0,   0,   0,  -1,   0 },    173.99408,   0.00002,    39.921 },
 { {   0,  27, -28,  -3,   0,   0,   0,   0,   1,  -1,  -1 },    111.10989,   0.00002,    64.739 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },     91.56917,   0.00001,     0.113 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.56917,   0.00001,     0.057 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,   2,  -2,   0 },    272.14261,   0.00002,     0.071 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,   2,  -1,  -2 },    272.14470,   0.00001,     1.902 },
 { {   0,  27, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.31192,   0.00002,     0.127 },
 { {   0,  27, -27,   0,   0,   0,   0,   0,   0,   0,   0 },    134.80294,   0.00001,     0.059 },
 { {   0,  27, -27,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.39760,   0.00002,     0.035 },
 { {   0,  27, -27,   0,   0,   0,   0,   0,   2,   0,   0 },    359.39688,   0.00005,     0.024 },
 { {   0,  27, -25,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.55416,   0.00001,     0.166 },
 { {   0,  27, -20, -20,   0,   0,   0,   0,   0,  -2,   0 },    329.81539,   0.00002,     0.075 },
 { {   0,  27, -20, -20,   0,   0,   0,   0,   0,  -1,   0 },    329.81539,   0.00031,  3189.335 },
 { {   0,  27, -20, -20,   0,   0,   0,   0,   0,   0,   0 },    329.81539,   0.00002,     0.075 },
 { {   0,  27, -19,   0,  -2,   0,   0,   0,  -2,   0,   0 },     72.90633,   0.00003,    58.910 },
 { {   0,  28, -42,  15,   0,   0,   0,   0,  -2,   1,   0 },    210.60298,   0.00002,   122.320 },
 { {   0,  28, -39,  11,   0,   0,   0,   0,  -1,   0,   0 },    136.46429,   0.00002,   179.530 },
 { {   0,  28, -37,   9,   0,   0,   0,   0,   0,  -1,   0 },    166.44272,   0.00002,    22.701 },
 { {   0,  28, -36,   4,   0,   0,   0,   0,  -2,   1,   0 },    309.95515,   0.00001,     6.264 },
 { {   0,  28, -36,   7,   0,   0,   0,   0,   0,  -1,   0 },    167.61974,   0.00001,    51.772 },
 { {   0,  28, -35,   0,   2,   0,   0,   0,  -2,   1,   0 },    323.15573,   0.00001,     1.252 },
 { {   0,  28, -35,   2,   0,   0,   0,   0,  -2,   0,   0 },    143.30848,   0.00003,     0.076 },
 { {   0,  28, -35,   2,   0,   0,   0,   0,  -2,   1,   0 },    337.76064,   0.00002,    10.387 },
 { {   0,  28, -35,   2,   0,   0,   0,   0,  -2,   2,   0 },    143.30848,   0.00003,     0.075 },
 { {   0,  28, -35,   5,   0,   0,   0,   0,   0,  -1,   0 },    108.57078,   0.00002,    12.094 },
 { {   0,  28, -34,   0,  -1,   0,   0,   0,  -2,   1,   0 },     80.62719,   0.00002,    19.458 },
 { {   0,  28, -34,   0,   0,  -1,   0,   0,  -2,   1,   0 },    106.33250,   0.00016,   964.920 },
 { {   0,  28, -34,   0,   0,  -1,   0,   0,  -2,   2,   0 },    332.64653,   0.00003,     0.075 },
 { {   0,  28, -34,   0,   0,   0,  -1,   0,  -2,   1,   0 },    224.27100,   0.00003,    47.598 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -4,   2,   0 },    242.47435,   0.00002,     0.087 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.47435,   0.00008,     0.076 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   2,   0 },     62.47435,   0.00008,     0.075 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,   0,   0,   0 },    242.47435,   0.00002,     0.087 },
 { {   0,  28, -34,   0,   0,   1,   0,   0,  -2,   1,   0 },    139.01818,   0.00002,    14.957 },
 { {   0,  28, -34,   0,   0,   2,   0,   0,  -2,   1,   0 },     49.61081,   0.00001,     9.920 },
 { {   0,  28, -34,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.54621,   0.00002,     0.076 },
 { {   0,  28, -34,   0,   1,   0,   0,   0,  -2,   1,   0 },    304.19952,   0.00002,     8.531 },
 { {   0,  28, -34,   0,   1,   0,   0,   0,  -2,   2,   0 },    141.54621,   0.00002,     0.075 },
 { {   0,  28, -34,   0,   2,   0,   0,   0,  -2,   0,   0 },    230.94938,   0.00007,     0.077 },
 { {   0,  28, -34,   3,   0,   0,   0,   0,   0,  -1,   0 },     20.14121,   0.00001,     6.847 },
 { {   0,  28, -33,  -2,   0,   0,   0,   0,  -2,   0,   0 },      9.67191,   0.00002,     0.075 },
 { {   0,  28, -33,  -2,   0,   0,   0,   0,  -2,   1,   0 },    176.05232,   0.00002,    32.834 },
 { {   0,  28, -33,  -2,   0,   0,   0,   0,  -2,   2,   0 },      9.67191,   0.00002,     0.076 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,  -1,   1,  -1 },     64.92306,   0.00014,    47.954 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,  -2,   0 },     21.41270,   0.00001,     0.071 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,   0,   0 },    201.41270,   0.00001,     0.080 },
 { {   0,  28, -33,   0,   1,   0,   0,   0,  -2,   1,   0 },     49.77741,   0.00001,     0.895 },
 { {   0,  28, -32,  -3,   0,   0,   0,   0,  -2,   1,   0 },     86.84420,   0.00001,     2.284 },
 { {   0,  28, -32,   0,  -3,   0,   0,   0,   0,  -1,   0 },    181.44296,   0.00005,   167.191 },
 { {   0,  28, -32,   0,  -2,   0,   0,   0,   0,  -1,   0 },    115.51685,   0.00003,    11.076 },
 { {   0,  28, -32,   0,  -1,   0,   0,   0,   0,  -1,   0 },     15.76863,   0.00001,     5.728 },
 { {   0,  28, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    231.09626,   0.00002,     0.089 },
 { {   0,  28, -32,   0,   0,   0,   0,   0,   0,  -1,   0 },    119.61074,   0.00002,     3.863 },
 { {   0,  28, -32,   0,   1,   0,   0,   0,   0,  -1,   0 },    218.00015,   0.00001,     2.914 },
 { {   0,  28, -32,   0,   1,   0,   0,   0,   0,   1,  -2 },    200.24925,   0.00001,   103.254 },
 { {   0,  28, -31,  -4,   0,   0,   0,   0,  -1,   0,   0 },     85.28578,   0.00001,    52.176 },
 { {   0,  28, -31,  -2,   0,   0,   0,   0,   0,  -1,   0 },     10.28812,   0.00001,     5.114 },
 { {   0,  28, -31,   0,   0,   0,   0,   0,   1,  -1,  -1 },    254.75886,   0.00004,     4.875 },
 { {   0,  28, -31,   0,   0,   0,   0,   0,   1,   0,  -2 },    258.06405,   0.00003,    26.061 },
 { {   0,  28, -30,  -4,   0,   0,   0,   0,   0,  -1,   0 },    173.08077,   0.00002,     7.567 },
 { {   0,  28, -30,   0,   0,   0,   0,   0,   2,  -1,  -2 },    359.25304,   0.00004,     6.607 },
 { {   0,  28, -30,   0,   0,   0,   0,   0,   2,   0,  -3 },    345.93281,   0.00005,    64.963 },
 { {   0,  28, -30,   0,   0,   0,   0,   0,   2,   0,   0 },    359.30144,   0.00001,     0.025 },
 { {   0,  28, -29,  -6,   0,   0,   0,   0,   0,  -1,   0 },    144.34012,   0.00001,    14.537 },
 { {   0,  28, -29,   0,   0,   0,   0,   0,   0,   0,   0 },    114.15406,   0.00002,     0.061 },
 { {   0,  28, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    359.51377,   0.00001,     0.052 },
 { {   0,  28, -28,   0,   0,   0,   0,   0,   0,   0,   0 },    179.34388,   0.00002,     0.057 },
 { {   0,  28, -28,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.34392,   0.00001,     0.034 },
 { {   0,  28, -28,   0,   0,   0,   0,   0,   2,   0,   0 },    359.34391,   0.00004,     0.024 },
 { {   0,  28, -25, -17,   0,   0,   0,   0,  -2,   1,   0 },    337.72025,   0.00001,   174.819 },
 { {   0,  29, -41,  10,   0,   0,   0,   0,  -2,   1,   0 },     30.90737,   0.00002,    40.450 },
 { {   0,  29, -37,   6,   0,   0,   0,   0,   0,  -1,   0 },    109.36483,   0.00001,    13.419 },
 { {   0,  29, -36,   0,   0,   0,   0,   0,  -2,   0,   2 },    136.73346,   0.00002,     0.075 },
 { {   0,  29, -36,   0,   2,   0,   0,   0,  -2,   1,   0 },    110.60882,   0.00003,     5.782 },
 { {   0,  29, -36,   0,   3,   0,   0,   0,  -2,   1,   0 },    310.05465,   0.00001,    11.280 },
 { {   0,  29, -36,   0,   4,   0,   0,   0,  -2,   0,   0 },    226.90054,   0.00003,     0.075 },
 { {   0,  29, -36,   0,   4,   0,   0,   0,  -2,   1,   0 },    163.96849,   0.00004,   230.025 },
 { {   0,  29, -36,   0,   4,   0,   0,   0,  -2,   2,   0 },    303.59416,   0.00001,     0.075 },
 { {   0,  29, -36,   2,   0,   0,   0,   0,  -1,  -1,   1 },     14.18380,   0.00001,   644.451 },
 { {   0,  29, -36,   4,   0,   0,   0,   0,   0,  -2,   0 },    260.60581,   0.00001,     0.076 },
 { {   0,  29, -36,   4,   0,   0,   0,   0,   0,  -1,   0 },     87.16729,   0.00002,    89.650 },
 { {   0,  29, -36,   4,   0,   0,   0,   0,   0,   0,   0 },    260.60581,   0.00001,     0.075 },
 { {   0,  29, -35,   0,   0,   0,   0,   0,  -2,   0,   0 },    149.40427,   0.00001,     0.079 },
 { {   0,  29, -35,   0,   0,   0,   0,   0,  -2,   2,   0 },    149.40427,   0.00001,     0.072 },
 { {   0,  29, -35,   0,   1,   0,   0,   0,  -2,   1,   0 },    141.97057,   0.00002,     1.346 },
 { {   0,  29, -35,   0,   2,   0,   0,   0,  -2,   1,   0 },    230.90181,   0.00001,     1.209 },
 { {   0,  29, -35,   0,   2,   0,   0,   0,  -1,   0,   0 },     50.90181,   0.00001,    16.677 },
 { {   0,  29, -35,   2,   0,   0,   0,   0,   0,  -1,   0 },    245.23861,   0.00002,    19.152 },
 { {   0,  29, -34,  -3,   0,   0,   0,   0,  -2,   1,   0 },     86.67820,   0.00002,    15.778 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },     57.10944,   0.00001,     0.086 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,  -2,   0 },    237.10944,   0.00005,     0.075 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,   0,   0 },    237.10944,   0.00005,     0.076 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   2,  -2,   0 },     57.10944,   0.00001,     0.088 },
 { {   0,  29, -34,   0,   0,   1,   0,   0,   0,  -1,   0 },    230.82361,   0.00003,    12.250 },
 { {   0,  29, -34,   0,   0,   2,   0,   0,   0,  -1,   0 },    307.04575,   0.00002,    20.971 },
 { {   0,  29, -34,   0,   1,   0,   0,   0,   0,  -2,   0 },    198.30919,   0.00003,     0.075 },
 { {   0,  29, -34,   0,   1,   0,   0,   0,   0,  -1,   0 },    244.25843,   0.00002,    31.972 },
 { {   0,  29, -34,   0,   1,   0,   0,   0,   0,   0,   0 },    264.88645,   0.00002,     0.076 },
 { {   0,  29, -34,   0,   2,   0,   0,   0,   0,  -1,   0 },     11.17643,   0.00002,    18.859 },
 { {   0,  29, -33,  -2,   0,   0,   0,   0,   0,  -1,   0 },    282.73877,   0.00001,     5.588 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -3,   0 },    336.27180,   0.00003,     0.075 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -2,   0 },    335.62831,   0.00048,   385.960 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -1,  -1 },    340.96531,   0.00001,     5.905 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -1,   0 },    287.13522,   0.00003,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,  -1,  -1,  -1 },    162.70535,   0.00003,     0.040 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,  -1,   0,  -1 },    162.70535,   0.00003,     0.087 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -2,  -1 },    162.70535,   0.00012,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -1,  -1 },    162.70535,   0.00226,  1352.997 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,   0,  -1 },    162.70535,   0.00012,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   3,  -2,  -1 },    162.70535,   0.00003,     0.087 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   3,  -1,  -1 },    162.70535,   0.00003,     0.040 },
 { {   0,  29, -30,   0,   0,   0,   0,   0,   4,  -3,  -2 },     92.53337,   0.00007,   348.536 },
 { {   0,  29, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.22788,   0.00001,     0.152 },
 { {   0,  29, -29,   0,   0,   0,   0,   0,   2,   0,   0 },    179.27295,   0.00001,     0.023 },
 { {   0,  29, -27, -13,   0,   0,   0,   0,   0,  -1,   0 },    148.85254,   0.00001,    36.394 },
 { {   0,  29, -24,   3,   0,   0,   0,   0,  -2,   0,   0 },    128.04773,   0.00003,   441.048 },
 { {   0,  30, -41,   7,   0,   0,   0,   0,  -2,   1,   0 },    229.64315,   0.00001,   173.980 },
 { {   0,  30, -40,   5,   0,   0,   0,   0,  -2,   1,   0 },     78.53930,   0.00002,    17.355 },
 { {   0,  30, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },     41.55296,   0.00002,     0.072 },
 { {   0,  30, -38,   3,   0,   0,   0,   0,  -1,   0,   0 },    143.31355,   0.00002,   125.038 },
 { {   0,  30, -37,   0,   0,   0,   0,   0,  -2,   1,   0 },    335.68751,   0.00001,     3.522 },
 { {   0,  30, -37,   0,   1,   0,   0,   0,  -2,   1,   0 },     49.62683,   0.00001,     2.716 },
 { {   0,  30, -37,   0,   2,   0,   0,   0,  -2,   1,   0 },    143.13428,   0.00003,     2.210 },
 { {   0,  30, -36,   0,   2,   0,   0,   0,   0,  -1,   0 },    230.85150,   0.00002,     3.111 },
 { {   0,  30, -36,   1,   0,   0,   0,   0,   0,  -1,   0 },    322.77564,   0.00004,    24.024 },
 { {   0,  30, -35,  -1,   0,   0,   0,   0,   0,  -1,   0 },    262.19871,   0.00005,    45.992 },
 { {   0,  30, -35,   3,   0,   0,   0,   0,   2,  -1,  -2 },    149.73494,   0.00002,   394.087 },
 { {   0,  30, -31,   0,   0,   0,   0,   0,   2,   0,   0 },    272.07474,   0.00002,     0.024 },
 { {   0,  30, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.17946,   0.00002,     0.167 },
 { {   0,  30, -25,   0,  -1,   0,   0,   0,  -3,   0,   1 },     27.77363,   0.00003,   428.143 },
 { {   0,  30, -24,   0,   0,  -1,   0,   0,  -2,   0,   0 },    300.84357,   0.00006,   174.065 },
 { {   0,  30, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.67061,   0.00008,    35.458 },
 { {   0,  31, -43,  11,   0,   0,   0,   0,   0,  -1,   0 },     53.72279,   0.00002,    62.465 },
 { {   0,  31, -41,   4,   0,   0,   0,   0,  -2,   1,   0 },     77.39350,   0.00001,    27.611 },
 { {   0,  31, -40,   2,   0,   0,   0,   0,  -2,   1,   0 },    146.21829,   0.00001,    36.833 },
 { {   0,  31, -39,   0,   0,   1,   0,   0,  -2,   1,   0 },    316.04461,   0.00001,    17.677 },
 { {   0,  31, -39,   0,   0,   2,   0,   0,  -2,   1,   0 },    217.68354,   0.00003,    44.205 },
 { {   0,  31, -39,   0,   0,   3,   0,   0,  -2,   1,   0 },    304.05991,   0.00001,    88.294 },
 { {   0,  31, -39,   0,   1,   0,   0,   0,  -2,   1,   0 },     74.79371,   0.00007,   160.924 },
 { {   0,  31, -39,   0,   2,   0,   0,   0,  -2,   0,   0 },    230.51442,   0.00001,     0.076 },
 { {   0,  31, -39,   0,   2,   0,   0,   0,  -2,   1,   0 },    175.36625,   0.00003,    12.806 },
 { {   0,  31, -39,   0,   2,   0,   0,   0,  -2,   2,   0 },     50.51442,   0.00001,     0.075 },
 { {   0,  31, -39,   0,   3,   0,   0,   0,  -2,   1,   0 },     40.76315,   0.00001,     6.158 },
 { {   0,  31, -38,   2,   0,   0,   0,   0,   0,  -1,   0 },    143.32708,   0.00001,     5.029 },
 { {   0,  31, -37,   0,  -2,   0,   0,   0,   0,  -1,   0 },     84.80914,   0.00001,    30.171 },
 { {   0,  31, -37,   0,  -1,   0,   0,   0,   0,  -1,   0 },    299.46869,   0.00005,    19.547 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,  -1,   0 },      5.27181,   0.00002,     7.382 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,   0,   0 },     67.42806,   0.00003,     0.075 },
 { {   0,  31, -37,   0,   1,   0,   0,   0,   0,  -1,   0 },    321.89953,   0.00003,     4.550 },
 { {   0,  31, -37,   0,   2,   0,   0,   0,   0,  -1,   0 },    230.80010,   0.00001,     3.289 },
 { {   0,  31, -36,  -2,   0,   0,   0,   0,   0,  -1,   0 },    190.02869,   0.00001,    13.871 },
 { {   0,  31, -36,   0,   1,   0,   0,   0,   1,   0,  -2 },    229.62679,   0.00012,  1326.246 },
 { {   0,  31, -35,   0,   0,   0,   0,   0,   2,  -1,  -2 },    170.52019,   0.00005,    35.801 },
 { {   0,  31, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.20627,   0.00002,     0.120 },
 { {   0,  31, -32,   0,   0,   0,   0,   0,   2,   0,   0 },    272.04902,   0.00002,     0.023 },
 { {   0,  31, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.13561,   0.00001,     0.187 },
 { {   0,  31, -31,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.17991,   0.00001,     0.057 },
 { {   0,  31, -31,   0,   0,   0,   0,   0,   2,   0,   0 },    179.17177,   0.00002,     0.023 },
 { {   0,  31, -29,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.37167,   0.00001,     0.285 },
 { {   0,  31, -23,  -5,   0,   0,   0,   0,  -2,   0,   0 },    181.77744,   0.00001,   213.001 },
 { {   0,  32, -43,   8,   0,   0,   0,   0,   0,  -1,   0 },    355.17483,   0.00002,    69.152 },
 { {   0,  32, -40,  -1,   0,   0,   0,   0,  -2,   1,   0 },     29.51394,   0.00002,   301.177 },
 { {   0,  32, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },    222.27359,   0.00002,     0.079 },
 { {   0,  32, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.49853,   0.00003,     0.085 },
 { {   0,  32, -39,   0,   1,   0,   0,   0,   0,  -1,   0 },    229.46907,   0.00002,     6.464 },
 { {   0,  32, -39,   0,   3,   0,   0,   0,   0,  -1,   0 },    166.03763,   0.00004,    71.941 },
 { {   0,  32, -33,   0,   0,   0,   0,   0,   2,   0,   0 },    272.02221,   0.00001,     0.023 },
 { {   0,  32, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.08797,   0.00001,     0.212 },
 { {   0,  32, -32,   0,   0,   0,   0,   0,   2,   0,   0 },    179.11633,   0.00002,     0.022 },
 { {   0,  32, -31,   7,   0,   0,   0,   0,  -2,   0,   0 },    310.80428,   0.00004,   962.299 },
 { {   0,  32, -28,   0,  -2,   0,   0,   0,  -3,   1,   0 },    184.51566,   0.00002,   425.427 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   1,  -1 },    140.41562,   0.00002,     0.075 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   2,  -1 },    140.41562,   0.00029,  2005.078 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   3,  -1 },    140.41562,   0.00002,     0.075 },
 { {   0,  32, -27,   0,  -3,   0,   0,   0,  -2,   0,   0 },    260.75522,   0.00003,    37.971 },
 { {   0,  32, -25,   0,  -2,   0,   0,   0,   0,   0,  -2 },     25.23827,   0.00002,   321.594 },
 { {   0,  33, -42,   0,   0,   0,   0,   0,  -2,   0,   0 },    109.02329,   0.00002,     0.076 },
 { {   0,  33, -42,   0,   0,   0,   0,   0,  -2,   2,   0 },    109.02329,   0.00002,     0.075 },
 { {   0,  33, -42,   3,   0,   0,   0,   0,   0,  -1,   0 },    233.59181,   0.00003,    54.240 },
 { {   0,  33, -40,   0,   0,   0,   0,   0,   0,   0,   0 },    137.73018,   0.00002,     0.073 },
 { {   0,  33, -39,   0,   0,   0,   0,   0,   1,   1,  -3 },     42.03161,   0.00010,  1300.532 },
 { {   0,  33, -38,   0,   0,   0,   0,   0,   0,   0,   0 },    323.74386,   0.00002,     0.064 },
 { {   0,  33, -37,   0,   0,   0,   0,   0,   0,   0,   0 },    230.80731,   0.00002,     0.060 },
 { {   0,  33, -33,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.03818,   0.00001,     0.244 },
 { {   0,  33, -33,   0,   0,   0,   0,   0,   2,   0,   0 },    179.05531,   0.00001,     0.022 },
 { {   0,  33, -30,   0,   0,   0,   0,   0,  -3,   0,   1 },    129.70212,   0.00002,    24.119 },
 { {   0,  33, -30,   2,   0,   0,   0,   0,  -2,   0,   0 },    272.63357,   0.00003,    31.389 },
 { {   0,  33, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },     53.96838,   0.00007,    10.501 },
 { {   0,  33, -29,   0,   1,   0,   0,   0,  -2,   0,   0 },    298.53955,   0.00004,    91.552 },
 { {   0,  34, -44,   0,   3,   0,   0,   0,  -2,   1,   0 },    175.43677,   0.00002,    25.665 },
 { {   0,  34, -42,   0,   0,   0,   0,   0,   0,   0,   0 },    263.98247,   0.00002,     0.075 },
 { {   0,  34, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },     51.80425,   0.00002,     0.106 },
 { {   0,  34, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    178.98881,   0.00001,     0.288 },
 { {   0,  34, -32,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.19975,   0.00002,     0.610 },
 { {   0,  34, -30,  -1,   0,   0,   0,   0,  -2,   0,   0 },     97.60566,   0.00011,   720.298 },
 { {   0,  35, -47,   3,   0,   0,   0,   0,  -2,   1,   0 },    306.60794,   0.00002,   150.902 },
 { {   0,  35, -46,   0,   7,   0,   0,   0,  -2,   1,   0 },    316.29287,   0.00003,   586.536 },
 { {   0,  35, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },    323.68504,   0.00003,     0.127 },
 { {   0,  35, -33,   0,   0,   0,   0,   0,  -2,   0,   0 },     14.01754,   0.00003,     1.185 },
 { {   0,  35, -32,   0,  -2,   0,   0,   0,  -2,   0,   0 },     25.04919,   0.00007,    78.189 },
 { {   0,  35, -32,   0,   0,  -3,   0,   0,  -2,   0,   0 },    177.30859,   0.00002,    18.528 },
 { {   0,  35, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    243.55198,   0.00004,     6.418 },
 { {   0,  35, -32,   0,   0,   0,   0,   0,  -2,   1,  -1 },    308.95020,   0.00001,    91.428 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -3,   0,  -1 },    113.91542,   0.00002,     0.040 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -3,   1,  -1 },    113.91542,   0.00002,     0.087 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,  -1,  -1 },    113.91542,   0.00007,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,   0,  -1 },    113.91542,   0.00136,  4784.938 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,   1,  -1 },    113.91542,   0.00007,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,   1,  -1,  -1 },    113.91542,   0.00002,     0.087 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,   1,   0,  -1 },    113.91542,   0.00002,     0.040 },
 { {   0,  35, -30,  -4,   0,   0,   0,   0,  -2,   0,   0 },    173.45657,   0.00003,    34.386 },
 { {   0,  36, -41,   0,   0,   0,   0,   0,  -2,   0,   0 },    323.65412,   0.00002,     0.139 },
 { {   0,  36, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },    230.61602,   0.00002,     0.161 },
 { {   0,  36, -37,   6,   0,   0,   0,   0,  -2,   0,   0 },    112.50701,   0.00002,    35.024 },
 { {   0,  36, -36,   4,   0,   0,   0,   0,  -2,   0,   0 },     84.31809,   0.00001,    10.879 },
 { {   0,  36, -35,   0,   2,   0,   0,   0,  -3,   0,   1 },    111.01562,   0.00004,   269.063 },
 { {   0,  36, -35,   2,   0,   0,   0,   0,  -2,   0,   0 },     66.73544,   0.00001,     6.440 },
 { {   0,  36, -34,   0,  -3,   0,   0,   0,  -2,   0,   0 },    189.18887,   0.00001,     2.121 },
 { {   0,  36, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    205.64049,   0.00006,     4.573 },
 { {   0,  36, -34,   0,   1,   0,   0,   0,  -2,   0,   0 },     13.79102,   0.00003,     7.443 },
 { {   0,  36, -34,   0,   2,   0,   0,   0,  -2,   0,   0 },    194.21500,   0.00001,    19.981 },
 { {   0,  36, -32,   0,   0,   0,   0,   0,   0,  -2,   0 },    237.57534,   0.00004,   136.522 },
 { {   0,  36, -29,  -9,   0,   0,   0,   0,  -2,   0,   0 },    176.91497,   0.00003,   262.068 },
 { {   0,  37, -37,   3,   0,   0,   0,   0,  -2,   0,   0 },     16.50375,   0.00013,   521.324 },
 { {   0,  37, -36,   0,  -2,   0,   0,   0,  -2,   0,   0 },    269.94562,   0.00001,     1.313 },
 { {   0,  37, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    294.19720,   0.00001,     1.686 },
 { {   0,  37, -35,   0,  -5,   0,   0,   0,  -2,   0,   0 },     10.84497,   0.00002,    68.230 },
 { {   0,  37, -35,   0,  -3,   0,   0,   0,  -2,   0,   0 },    216.74335,   0.00003,     6.496 },
 { {   0,  37, -35,   0,  -1,   0,   0,   0,  -2,   0,   0 },     39.37725,   0.00001,     3.100 },
 { {   0,  38, -45,  15,   0,   0,   0,   0,  -2,   0,   0 },    180.04277,   0.00003,   130.682 },
 { {   0,  38, -42,   6,   0,   0,   0,   0,  -4,   2,   0 },    237.22078,   0.00001,   286.659 },
 { {   0,  38, -39,   3,   0,   0,   0,   0,  -2,   0,   0 },     55.20445,   0.00001,     2.684 },
 { {   0,  38, -38,   0,  -1,   0,   0,   0,  -3,   0,   1 },    242.38914,   0.00006,   540.584 },
 { {   0,  38, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },    170.48717,   0.00001,     1.033 },
 { {   0,  38, -37,   0,  -3,   0,   0,   0,  -2,   0,   0 },    281.50948,   0.00003,     4.535 },
 { {   0,  38, -37,   0,  -2,   0,   0,   0,  -2,   0,   0 },    295.97422,   0.00004,     7.341 },
 { {   0,  38, -37,   0,  -1,   0,   0,   0,  -2,   0,   0 },     59.34438,   0.00003,    19.262 },
 { {   0,  38, -37,   0,   0,  -2,   0,   0,  -2,   0,   0 },    289.73904,   0.00001,    28.163 },
 { {   0,  38, -37,   0,   0,  -1,   0,   0,  -2,   0,   0 },    316.81822,   0.00016,   641.251 },
 { {   0,  38, -37,   0,   0,   0,  -3,   0,  -2,   0,   0 },     93.54456,   0.00003,   301.443 },
 { {   0,  38, -37,   0,   0,   0,  -2,   0,  -2,   0,   0 },    279.03123,   0.00004,   116.490 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,  -1,   0 },    307.19129,   0.00003,     0.076 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   0,   0 },    307.19129,   0.00061,    30.876 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   1,   0 },    307.19129,   0.00003,     0.075 },
 { {   0,  38, -37,   0,   1,   0,   0,   0,  -2,   0,   0 },    195.08401,   0.00004,     8.570 },
 { {   0,  38, -36,  -2,   0,   0,   0,   0,  -2,   0,   0 },     77.38994,   0.00003,    32.280 },
 { {   0,  38, -36,   0,   0,   0,   0,   0,  -1,   0,  -1 },     39.93914,   0.00004,    46.780 },
 { {   0,  38, -35,  -4,   0,   0,   0,   0,  -2,   0,   0 },    322.55580,   0.00001,    10.599 },
 { {   0,  38, -35,   0,  -3,   0,   0,   0,   0,  -2,   0 },      0.31228,   0.00002,   183.214 },
 { {   0,  38, -28, -17,   0,   0,   0,   0,  -2,   0,   0 },    193.48464,   0.00005,   160.172 },
 { {   0,  39, -41,   4,   0,   0,   0,   0,  -2,   0,   0 },    236.79477,   0.00001,     4.644 },
 { {   0,  39, -40,   2,   0,   0,   0,   0,  -2,   0,   0 },    219.00396,   0.00002,     3.588 },
 { {   0,  39, -39,   0,  -2,   0,   0,   0,  -2,   0,   0 },    198.89773,   0.00001,     1.958 },
 { {   0,  39, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },    357.33923,   0.00011,     2.923 },
 { {   0,  39, -39,   0,   0,   1,   0,   0,  -2,   0,   0 },    266.34334,   0.00001,     3.245 },
 { {   0,  39, -39,   0,   4,   0,   0,   0,  -2,   0,   0 },    255.37927,   0.00004,   205.320 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -4,   0,   0 },     16.24881,   0.00002,     0.040 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -4,   1,   0 },     16.24881,   0.00002,     0.087 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,  -1,   0 },     16.24881,   0.00010,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,   0,   0 },     16.24881,   0.00180,  1963.481 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,   1,   0 },     16.24881,   0.00010,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,   0,  -1,   0 },     16.24881,   0.00002,     0.087 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,   0,   0,   0 },     16.24881,   0.00002,     0.040 },
 { {   0,  39, -36,  -3,   0,   0,   0,   0,  -1,   0,  -1 },    218.02684,   0.00005,   109.969 },
 { {   0,  40, -45,   9,   0,   0,   0,   0,  -2,   0,   0 },    115.25240,   0.00002,    14.579 },
 { {   0,  40, -44,   7,   0,   0,   0,   0,  -2,   0,   0 },    217.59888,   0.00002,   191.399 },
 { {   0,  40, -43,   5,   0,   0,   0,   0,  -2,   0,   0 },    237.08787,   0.00003,    17.199 },
 { {   0,  40, -41,   0,  -2,   0,   0,   0,  -3,   1,   0 },    101.61372,   0.00004,   544.976 },
 { {   0,  40, -41,   0,   0,   0,   0,   0,  -3,   2,  -1 },     68.52461,   0.00002,   271.239 },
 { {   0,  40, -41,   1,   0,   0,   0,   0,  -2,   0,   0 },    358.59464,   0.00001,     5.409 },
 { {   0,  40, -40,   0,  -4,   0,   0,   0,  -2,   0,   0 },     53.89583,   0.00003,    18.594 },
 { {   0,  40, -40,   0,  -3,   0,   0,   0,  -2,   0,   0 },     15.71844,   0.00002,    32.764 },
 { {   0,  40, -40,   0,  -2,   0,   0,   0,  -2,   0,   0 },     17.07395,   0.00003,     8.709 },
 { {   0,  40, -40,   0,  -1,   0,   0,   0,  -2,   0,   0 },     54.20251,   0.00003,     5.022 },
 { {   0,  40, -40,   0,   0,  -1,   0,   0,  -2,   0,   0 },    270.95273,   0.00001,     4.008 },
 { {   0,  40, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },    354.68350,   0.00007,     3.528 },
 { {   0,  40, -39,  -3,   0,   0,   0,   0,  -2,   0,   0 },    227.69039,   0.00001,     3.209 },
 { {   0,  40, -39,  -2,   0,   0,   0,   0,  -2,   0,   0 },    316.91957,   0.00002,     4.544 },
 { {   0,  40, -39,   0,  -1,   0,   0,   0,  -2,   0,   0 },    179.82030,   0.00001,     0.834 },
 { {   0,  40, -39,   0,   0,   0,   0,   0,  -1,   0,  -1 },    282.06234,   0.00002,     4.354 },
 { {   0,  40, -38,  -4,   0,   0,   0,   0,  -2,   0,   0 },    302.69767,   0.00001,     6.382 },
 { {   0,  40, -38,   0,  -2,   0,   0,   0,   0,   0,  -2 },    351.98310,   0.00004,   137.080 },
 { {   0,  40, -37,  -6,   0,   0,   0,   0,  -2,   0,   0 },     81.44068,   0.00001,    10.715 },
 { {   0,  41, -45,   6,   0,   0,   0,   0,  -2,   0,   0 },     57.43194,   0.00001,    10.095 },
 { {   0,  41, -43,   2,   0,   0,   0,   0,  -2,   0,   0 },    351.24548,   0.00004,    36.137 },
 { {   0,  41, -42,   0,  -1,   0,   0,   0,  -2,   0,   0 },    149.47275,   0.00002,     5.703 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,  -1,   0 },     91.32596,   0.00003,     0.075 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   0,   0 },     91.32596,   0.00048,    10.984 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   1,   0 },     91.32596,   0.00003,     0.076 },
 { {   0,  41, -42,   0,   0,   1,   0,   0,  -2,   0,   0 },    356.83254,   0.00001,    17.515 },
 { {   0,  41, -42,   0,   0,   2,   0,   0,  -2,   0,   0 },     62.29644,   0.00004,    43.206 },
 { {   0,  41, -42,   0,   2,   0,   0,   0,  -2,   0,   0 },     39.59155,   0.00004,    12.892 },
 { {   0,  41, -41,   0,   0,   0,   0,   0,  -2,   0,   0 },    178.60666,   0.00001,     1.100 },
 { {   0,  41, -41,   0,   0,   0,   0,   0,  -1,   0,  -1 },     13.35268,   0.00003,     6.906 },
 { {   0,  41, -35, -13,   0,   0,   0,   0,  -2,   0,   0 },     55.38966,   0.00001,   340.508 },
 { {   0,  42, -44,   0,   1,   0,   0,   0,  -2,   0,   0 },    109.41966,   0.00003,     2.623 },
 { {   0,  42, -43,  -1,   0,   0,   0,   0,  -2,   0,   0 },     17.71809,   0.00002,   357.500 },
 { {   0,  42, -43,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.96104,   0.00001,     1.871 },
 { {   0,  42, -43,   0,   1,   0,   0,   0,  -2,   0,   0 },    358.07360,   0.00001,     1.616 },
 { {   0,  42, -42,  -3,   0,   0,   0,   0,  -2,   0,   0 },     30.67588,   0.00004,    16.510 },
 { {   0,  43, -51,  11,   0,   0,   0,   0,  -2,   0,   0 },     50.50552,   0.00003,   117.217 },
 { {   0,  43, -46,   2,   0,   0,   0,   0,  -2,   0,   0 },     91.34680,   0.00001,     4.477 },
 { {   0,  43, -45,   0,  -2,   0,   0,   0,  -2,   0,   0 },    187.67692,   0.00006,   116.226 },
 { {   0,  43, -45,   0,  -1,   0,   0,   0,  -2,   0,   0 },    283.24310,   0.00004,    13.210 },
 { {   0,  43, -45,   0,   1,   0,   0,   0,  -2,   0,   0 },    123.22300,   0.00002,     4.093 },
 { {   0,  43, -44,  -2,   0,   0,   0,   0,  -2,   0,   0 },    157.92112,   0.00002,    10.348 },
 { {   0,  43, -44,   0,  -1,   0,   0,   0,  -1,   0,  -1 },    350.86562,   0.00001,    45.591 },
 { {   0,  43, -43,  -4,   0,   0,   0,   0,  -2,   0,   0 },     39.58259,   0.00002,    30.060 },
 { {   0,  43, -43,   0,  -1,   0,   0,   0,   0,   0,  -2 },     92.38516,   0.00002,    31.417 },
 { {   0,  43, -43,   0,   0,   0,   0,   0,   0,   0,  -2 },      0.77865,   0.00003,    19.058 },
 { {   0,  43, -42,  -6,   0,   0,   0,   0,  -2,   0,   0 },     26.21171,   0.00003,    33.222 },
 { {   0,  43, -42,  -2,   0,   0,   0,   0,   0,   0,  -2 },    101.66231,   0.00002,    91.787 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,  -1,   0,  -3 },    281.55562,   0.00008,     0.040 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,  -1,   1,  -3 },    281.55562,   0.00007,     0.087 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,  -2,  -3 },    281.55562,   0.00002,     0.038 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,  -1,  -3 },    281.55562,   0.00036,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -5 },    101.55562,   0.00001,     0.037 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -3 },    281.55562,   0.00660,   773.987 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -1 },    101.55562,   0.00001,     0.037 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   1,  -3 },    281.55562,   0.00036,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   2,  -3 },    281.55562,   0.00002,     0.038 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   3,  -1,  -3 },    281.55562,   0.00007,     0.087 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   3,   0,  -3 },    281.55562,   0.00008,     0.040 },
 { {   0,  44, -48,   2,   0,   0,   0,   0,  -2,   0,   0 },    179.07244,   0.00001,     6.618 },
 { {   0,  44, -47,   0,   0,   0,   0,   0,  -2,   0,   0 },    123.34583,   0.00001,     4.663 },
 { {   0,  44, -47,   0,   2,   0,   0,   0,  -2,   0,   0 },    274.19765,   0.00002,    21.804 },
 { {   0,  44, -47,   0,   3,   0,   0,   0,  -2,   0,   0 },    336.29113,   0.00001,    26.014 },
 { {   0,  44, -45,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.28317,   0.00007,    86.878 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -4,   0,   0 },    296.45270,   0.00002,     0.040 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -4,   1,   0 },    296.45270,   0.00002,     0.087 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,  -1,   0 },    296.45270,   0.00010,     0.075 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,   0,   0 },    296.45270,   0.00183,  2704.790 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,   1,   0 },    296.45270,   0.00010,     0.075 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,   0,  -1,   0 },    296.45270,   0.00002,     0.087 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,   0,   0,   0 },    296.45270,   0.00002,     0.040 },
 { {   0,  45, -50,   3,   0,   0,   0,   0,  -2,   0,   0 },    276.00610,   0.00005,   163.835 },
 { {   0,  45, -48,   0,   0,   0,   0,   0,  -2,   0,   0 },     94.86623,   0.00001,     2.433 },
 { {   0,  46, -58,  15,   0,   0,   0,   0,  -2,   0,   0 },    284.34780,   0.00001,    84.476 },
 { {   0,  46, -51,   2,   0,   0,   0,   0,  -2,   0,   0 },     94.41690,   0.00003,    10.006 },
 { {   0,  46, -50,   0,   0,  -1,   0,   0,  -2,   0,   0 },    108.84125,   0.00008,   380.797 },
 { {   0,  46, -50,   0,   0,   0,   0,   0,  -2,  -1,   0 },    229.90240,   0.00001,     0.076 },
 { {   0,  46, -50,   0,   0,   0,   0,   0,  -2,   1,   0 },    229.90240,   0.00001,     0.075 },
 { {   0,  46, -50,   0,   1,   0,   0,   0,  -2,   0,   0 },    225.49997,   0.00002,     8.273 },
 { {   0,  46, -49,   0,   0,   0,   0,   0,  -1,   0,  -1 },     67.05250,   0.00009,    58.170 },
 { {   0,  46, -49,   2,   0,   0,   0,   0,   0,   0,  -2 },     26.03227,   0.00001,   131.783 },
 { {   0,  46, -48,  -4,   0,   0,   0,   0,  -2,   0,   0 },    352.04148,   0.00002,    11.091 },
 { {   0,  46, -41, -17,   0,   0,   0,   0,  -2,   0,   0 },    129.74459,   0.00002,   485.961 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,  -1,   0 },    259.65428,   0.00004,     0.075 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,   0,   0 },    259.65428,   0.00066,  1460.021 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,   1,   0 },    259.65428,   0.00004,     0.075 },
 { {   0,  47, -50,  -3,   0,   0,   0,   0,  -2,   0,   0 },    105.95508,   0.00001,    14.916 },
 { {   0,  47, -49,   0,   0,   0,   0,   0,   1,  -3,   0 },    125.66930,   0.00001,   933.289 },
 { {   0,  48, -57,   7,   0,   0,   0,   0,  -2,   0,   0 },     78.51579,   0.00001,   106.267 },
 { {   0,  49, -57,   4,   0,   0,   0,   0,  -2,   0,   0 },    142.75870,   0.00001,    25.075 },
 { {   0,  49, -56,   2,   0,   0,   0,   0,  -2,   0,   0 },     90.92478,   0.00002,    42.576 },
 { {   0,  49, -55,   0,   0,   1,   0,   0,  -2,   0,   0 },    322.53267,   0.00001,    18.901 },
 { {   0,  49, -55,   0,   0,   2,   0,   0,  -2,   0,   0 },    228.15531,   0.00001,    52.744 },
 { {   0,  49, -55,   0,   1,   0,   0,   0,  -2,   0,   0 },      6.88496,   0.00012,   391.900 },
 { {   0,  49, -48, -13,   0,   0,   0,   0,  -2,   0,   0 },    237.79204,   0.00001,   800.815 },
 { {   0,  50, -56,  -1,   0,   0,   0,   0,  -2,   0,   0 },    161.89377,   0.00003,   143.211 },
 { {   0,  51, -58,   0,  -2,   0,   0,   0,  -2,   0,   0 },      2.21636,   0.00003,   226.325 },
 { {   0,  51, -55,   0,   0,   0,   0,   0,   1,   0,  -3 },    162.53003,   0.00002,   345.601 },
 { {   0,  52, -60,   0,   2,   0,   0,   0,  -2,   0,   0 },    218.72807,   0.00003,    23.994 },
 { {   0,  53, -63,   3,   0,   0,   0,   0,  -2,   0,   0 },    320.47392,   0.00002,    97.189 },
 { {   0,  55, -65,   0,   4,   0,   0,   0,  -2,   0,   0 },     92.48135,   0.00005,   285.666 },
 { {   0,  59, -58,   0,   1,   0,   0,   0,  -2,  -1,   0 },    204.33634,   0.00003,   325.228 },
 { {   0,  61, -64,   2,   0,   0,   0,   0,  -4,   1,   0 },    114.66924,   0.00002,   949.691 },
 { {   0,  61, -61,   0,  -2,   0,   0,   0,  -2,  -1,   0 },      9.57199,   0.00005,   202.367 },
 { {   1, -51,  54,   0,   0,   0,   0,   0,   2,   0,   0 },    131.05411,   0.00001,    82.651 },
 { {   1, -43,  41,   0,   0,   0,   0,   0,   2,   0,   0 },     32.95991,   0.00004,   126.367 },
 { {   1, -35,  28,   0,   0,   0,   0,   0,   2,   0,   0 },    288.16234,   0.00007,   268.243 },
 { {   1, -33,  38,   0,   0,   0,   0,   0,   2,  -1,   0 },    108.33477,   0.00001,   118.531 },
 { {   1, -31,  33,   0,   0,   0,   0,   0,   0,   1,   0 },    103.66695,   0.00001,    60.144 },
 { {   1, -30,  33,   0,   0,   0,   0,   0,   2,  -1,   0 },    113.56021,   0.00003,     7.584 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -4,   1,   2 },     77.18279,   0.00001,     0.040 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   0,   2 },     77.18279,   0.00005,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   1,   2 },     77.18279,   0.00088,  1374.129 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   2,   2 },     77.18279,   0.00005,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,   0,   1,   2 },     77.18279,   0.00001,     0.040 },
 { {   1, -28,  30,   0,   0,   0,   0,   0,   2,  -1,   0 },    205.27405,   0.00001,     8.391 },
 { {   1, -25,  25,   0,   0,   0,   0,   0,   2,  -2,   0 },     72.86625,   0.00003,     0.075 },
 { {   1, -25,  25,   0,   0,   0,   0,   0,   2,  -1,   0 },    135.06025,   0.00003,   235.235 },
 { {   1, -25,  25,   0,   0,   0,   0,   0,   2,   0,   0 },     72.86625,   0.00003,     0.075 },
 { {   1, -23,  20,   0,   0,   0,   0,   0,   0,   1,   0 },    150.68724,   0.00002,    48.049 },
 { {   1, -23,  22,   0,   0,   0,   0,   0,   2,  -1,   0 },    345.52179,   0.00002,     4.052 },
 { {   1, -21,  19,   0,   0,   0,   0,   0,   2,  -1,   0 },     77.34438,   0.00001,     2.009 },
 { {   1, -20,  13,   0,   0,   0,   0,   0,  -2,   1,   2 },    203.51144,   0.00001,   203.531 },
 { {   1, -20,  16,   0,   0,   0,   0,   0,   1,   0,   0 },     67.84951,   0.00004,    96.401 },
 { {   1, -20,  17,   0,   0,   0,   0,   0,   2,  -1,   0 },    149.79295,   0.00004,     8.106 },
 { {   1, -10,  10,   0,   0,   0,   0,   0,  -2,   0,   2 },     87.98710,   0.00001,   227.782 },
 { {   1,  -9,   7,   0,   0,   0,   0,   0,   0,   0,   0 },     76.87347,   0.00001,     0.288 },
 { {   1,  -8,   6,   0,   0,   0,   0,   0,   0,   0,   0 },     76.86713,   0.00001,     0.351 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   0,  -2,   0 },    114.25737,   0.00001,     0.040 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   0,  -1,   0 },    114.25737,   0.00001,     0.087 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -3,   0 },    111.33037,   0.00004,     0.075 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -2,   0 },    114.25737,   0.00092,  1698.830 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -1,   0 },    129.45385,   0.00006,     0.075 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   4,  -3,   0 },    114.25737,   0.00001,     0.087 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   4,  -2,   0 },    114.25737,   0.00001,     0.040 },
 { {   1,  -6,   4,   0,   0,   0,   0,   0,  -2,   1,   0 },     76.90106,   0.00002,     0.076 },
 { {   1,  -6,   4,   0,   0,   0,   0,   0,  -2,   3,   0 },     76.90106,   0.00002,     0.074 },
 { {   1,  -6,   5,   0,   0,   0,   0,   0,   0,   0,   0 },    348.42115,   0.00002,     1.664 },
 { {   1,  -5,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     80.05975,   0.00001,     1.025 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.70769,   0.00002,     0.040 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,  -2,   1,   0 },    351.70769,   0.00002,     0.087 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   0,  -1,   0 },    348.53644,   0.00003,     0.076 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   0,   0,   0 },    351.13666,   0.00016,    40.859 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   0,   1,   0 },    348.53644,   0.00003,     0.075 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },    351.75147,   0.00002,     0.087 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    351.75147,   0.00002,     0.040 },
 { {   1,  -4,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    174.20518,   0.00001,     1.538 },
 { {   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   2,  -1 },    339.27729,   0.00002,   236.591 },
 { {   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    153.84855,   0.00002,     1.380 },
 { {   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    336.11243,   0.00002,     1.899 },
 { {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    257.13574,   0.00001,     0.655 },
 { {   1,   0, -11,  11,   0,   0,   0,   0,   0,   0,   0 },    229.87126,   0.00001,     1.001 },
 { {   1,   0, -11,  11,   0,   0,   0,   0,   2,   0,   0 },    229.87126,   0.00001,     0.042 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,  -1,   0 },    126.93434,   0.00002,     0.026 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,   0,   0 },    126.93434,   0.00009,     0.040 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,   1,   0 },    126.93434,   0.00019,     0.087 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   0,  -1,   0 },    306.93434,   0.00002,     0.075 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   0,   0,   0 },    306.93434,   0.00001,  1618.870 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   0,   1,   0 },    306.93434,   0.00002,     0.075 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,  -1,   0 },    126.93434,   0.00019,     0.087 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,   0,   0 },    126.93434,   0.00009,     0.040 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,   1,   0 },    126.93434,   0.00002,     0.026 },
 { {   1,   0,  -9,  11,   0,   0,   0,   0,  -2,   0,   0 },     23.99741,   0.00001,     0.042 },
 { {   1,   0,  -9,  11,   0,   0,   0,   0,   0,   0,   0 },     23.99741,   0.00001,     0.999 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    128.87340,   0.00003,     0.075 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    128.81148,   0.00001,    13.534 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    128.87340,   0.00003,     0.076 },
 { {   1,   0,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    357.07582,   0.00002,     1.179 },
 { {   1,   0,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    320.45554,   0.00002,     0.094 },
 { {   1,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    320.55205,   0.00003,     0.042 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },    209.54941,   0.00001,     0.028 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    209.54959,   0.00004,     0.026 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    209.61102,   0.00037,     0.041 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    209.07557,   0.00054,     0.088 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    209.65789,   0.00002,     0.519 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    210.83710,   0.00005,     0.038 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    211.76798,   0.00069,     0.076 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    210.62744,   0.00343,     6.575 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    211.98110,   0.00068,     0.075 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    211.76522,   0.00005,     0.038 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    210.15770,   0.00043,     0.086 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    210.47236,   0.00027,     0.040 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    210.32785,   0.00003,     0.026 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },    136.66278,   0.00001,     0.029 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    136.63648,   0.00005,     0.027 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    136.63611,   0.00041,     0.042 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    136.68663,   0.00052,     0.097 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    136.71670,   0.00002,     0.342 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    136.85267,   0.00028,     0.083 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    137.55570,   0.00130,     0.868 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    137.14070,   0.00020,     0.069 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    289.95008,   0.00004,     0.076 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    289.92226,   0.00028,    10.170 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    318.13697,   0.00002,     0.074 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },    109.58839,   0.00003,     0.037 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    289.95008,   0.00004,     0.075 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    138.68164,   0.00019,     0.079 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    138.62924,   0.00007,     0.039 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    274.17776,   0.00009,     0.044 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    268.11549,   0.00011,     0.107 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    255.50152,   0.00002,     0.041 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.60911,   0.00021,     0.090 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     68.94317,   0.00019,     0.465 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     64.70620,   0.00003,     0.065 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },     33.45054,   0.00002,     0.910 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,  -2 },     77.59292,   0.00008,     0.076 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     75.74895,   0.00119,     0.073 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },     76.38184,   0.00070,    22.442 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    259.48578,   0.00010,     0.037 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   1,  -2 },     77.59292,   0.00008,     0.075 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    254.29338,   0.00002,     0.025 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },     75.67593,   0.00002,     0.040 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    175.84948,   0.00005,     0.046 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    176.24209,   0.00003,     0.120 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },    176.95710,   0.00001,     0.055 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    179.16778,   0.00003,     0.099 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    179.45421,   0.00034,     0.317 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    178.83487,   0.00004,     0.061 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },     53.92661,   0.00001,     0.477 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    177.49408,   0.00012,     0.068 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    358.03369,   0.00006,     0.036 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    290.91409,   0.00005,     0.049 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    293.64188,   0.00002,     0.136 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    313.93738,   0.00001,     0.192 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    289.46980,   0.00012,     0.241 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    279.21122,   0.00003,     0.064 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    106.53061,   0.00002,     0.035 },
 { {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     52.92676,   0.00002,     0.194 },
 { {   1,   0,  27, -37,   0,   0,   0,   0,  -2,   1,   0 },    190.38280,   0.00002,   509.270 },
 { {   1,   1,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    342.10263,   0.00003,     0.075 },
 { {   1,   1,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    144.39918,   0.00004,   278.019 },
 { {   1,   1,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    342.10263,   0.00003,     0.075 },
 { {   1,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    256.85055,   0.00001,     0.294 },
 { {   1,   3,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    308.59592,   0.00002,     0.076 },
 { {   1,   3,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    308.59592,   0.00002,     0.075 },
 { {   1,  10,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    188.56156,   0.00003,    13.438 },
 { {   1,  13, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     76.70400,   0.00001,     0.835 },
 { {   1,  13, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    157.47711,   0.00001,    35.540 },
 { {   1,  14, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     76.67742,   0.00001,     0.072 },
 { {   1,  14, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },     76.67742,   0.00001,     0.079 },
 { {   1,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     76.83770,   0.00003,     0.076 },
 { {   1,  15, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },     76.68984,   0.00002,    18.797 },
 { {   1,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },     76.83770,   0.00003,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -3,   1,  -1 },      4.34116,   0.00003,     0.040 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -3,   2,  -1 },      4.34116,   0.00003,     0.087 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,  -1,  -1 },      4.34116,   0.00001,     0.038 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },      2.38969,   0.00019,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   1,  -1 },      4.34116,   0.00280,  1772.175 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },      2.38969,   0.00019,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   3,  -1 },      4.34116,   0.00001,     0.038 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },      4.34116,   0.00003,     0.087 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,   1,   1,  -1 },      4.34116,   0.00003,     0.040 },
 { {   1,  16, -19,   0,   0,   0,   0,   0,  -2,  -1,   2 },    162.16957,   0.00001,    81.712 },
 { {   1,  16, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    291.95086,   0.00001,    10.494 },
 { {   1,  18, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },     95.60438,   0.00004,    14.239 },
 { {   1,  18, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.61573,   0.00002,     6.421 },
 { {   1,  21, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },     24.61435,   0.00007,    30.938 },
 { {   1,  23, -30,   0,   0,   0,   0,   0,  -2,   1,   0 },     34.31800,   0.00003,    17.426 },
 { {   1,  23, -29,   0,   0,   0,   0,   0,  -1,   1,  -1 },    296.91300,   0.00003,   276.148 },
 { {   1,  26, -35,   0,   0,   0,   0,   0,  -2,   1,   0 },    173.28188,   0.00001,    15.142 },
 { {   1,  36, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },     73.07787,   0.00003,    15.023 },
 { {   1,  41, -45,   0,   0,   0,   0,   0,  -1,   0,  -1 },    272.50388,   0.00001,   137.294 },
 { {   2, -56,  58,   0,   0,   0,   0,   0,   2,   0,   0 },    272.07051,   0.00005,    80.804 },
 { {   2, -48,  45,   0,   0,   0,   0,   0,   2,   0,   0 },    261.88899,   0.00001,    60.382 },
 { {   2, -45,  38,   0,   0,   0,   0,   0,   0,   0,   2 },     10.29663,   0.00002,  1510.143 },
 { {   2, -42,  37,   0,   0,   0,   0,   0,   4,  -2,   0 },    223.19313,   0.00003,   231.663 },
 { {   2, -36,  37,   0,   0,   0,   0,   0,   0,   1,   0 },    357.80269,   0.00002,    24.330 },
 { {   2, -33,  32,   0,   0,   0,   0,   0,   0,   1,   0 },    350.92054,   0.00003,    12.147 },
 { {   2, -33,  34,   0,   0,   0,   0,   0,   2,  -1,   0 },    352.26054,   0.00001,     6.961 },
 { {   2, -32,  32,   0,   0,   0,   0,   0,   2,  -1,   0 },     78.99747,   0.00001,     4.332 },
 { {   2, -30,  29,   0,   0,   0,   0,   0,   2,  -2,   0 },    351.36750,   0.00002,     0.076 },
 { {   2, -30,  29,   0,   0,   0,   0,   0,   2,  -1,   0 },    268.25318,   0.00001,    49.447 },
 { {   2, -30,  29,   0,   0,   0,   0,   0,   2,   0,   0 },    351.36750,   0.00002,     0.075 },
 { {   2, -28,  24,   0,   0,   0,   0,   0,   0,   1,   0 },    145.00481,   0.00003,    22.081 },
 { {   2, -28,  25,   0,   0,   0,   0,   0,   1,   1,  -1 },     61.19473,   0.00002,   117.956 },
 { {   2, -27,  22,   0,   0,   0,   0,   0,   0,  -1,   2 },    223.81676,   0.00003,   231.235 },
 { {   2, -27,  24,   0,   0,   0,   0,   0,   2,  -1,   0 },    141.20533,   0.00003,     9.689 },
 { {   2, -25,  20,   0,   0,   0,   0,   0,   1,   0,   0 },     42.34068,   0.00003,    28.696 },
 { {   2, -25,  21,   0,   0,   0,   0,   0,   2,  -1,   0 },    144.45551,   0.00001,     6.764 },
 { {   2, -24,  19,   0,   0,   0,   0,   0,   2,  -2,   0 },     42.29277,   0.00001,     0.074 },
 { {   2, -24,  19,   0,   0,   0,   0,   0,   2,   0,   0 },     42.29277,   0.00001,     0.077 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -4,   0 },     27.44429,   0.00003,     0.075 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -3,   0 },     27.44429,   0.00058,  1528.620 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -2,   0 },     27.44429,   0.00003,     0.075 },
 { {   2, -22,  16,   0,   0,   0,   0,   0,   2,  -1,   0 },    327.81499,   0.00004,    40.968 },
 { {   2, -20,  12,   0,   0,   0,   0,   0,   1,   1,  -1 },    271.12666,   0.00002,   232.978 },
 { {   2, -12,  13,   0,   0,   0,   0,   0,   2,  -3,   0 },    185.69581,   0.00001,     0.076 },
 { {   2, -12,  13,   0,   0,   0,   0,   0,   2,  -2,   0 },      3.11023,   0.00002,    41.866 },
 { {   2, -12,  13,   0,   0,   0,   0,   0,   2,  -1,   0 },    185.69581,   0.00001,     0.075 },
 { {   2, -11,   6,   0,   0,   0,   0,   0,   0,   0,   0 },    222.26879,   0.00001,     0.280 },
 { {   2, -10,   8,   0,   0,   0,   0,   0,   0,   0,   0 },    114.97925,   0.00001,    20.429 },
 { {   2,  -9,   8,   0,   0,   0,   0,   0,   2,  -2,   0 },    170.51661,   0.00001,    10.046 },
 { {   2,  -8,   3,   0,   0,   0,   0,   0,  -2,   1,   0 },     42.96954,   0.00001,     0.076 },
 { {   2,  -8,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    222.22620,   0.00004,    13.511 },
 { {   2,  -8,   3,   0,   0,   0,   0,   0,  -2,   3,   0 },     42.96954,   0.00001,     0.075 },
 { {   2,  -7,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    146.92333,   0.00005,    13.427 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    342.14626,   0.00001,     0.555 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    314.45685,   0.00004,     0.076 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    136.02991,   0.00003,    35.623 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    314.45685,   0.00004,     0.075 },
 { {   2,  -3,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    222.01215,   0.00001,     0.074 },
 { {   2,  -3,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    222.01215,   0.00001,     0.077 },
 { {   2,  -2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    273.69942,   0.00003,     0.074 },
 { {   2,  -2,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    273.69942,   0.00003,     0.076 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    296.79910,   0.00002,     0.087 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    113.79542,   0.00015,     0.075 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    105.29889,   0.00071,  1587.748 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    113.79542,   0.00015,     0.075 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    296.79910,   0.00002,     0.087 },
 { {   2,  -2,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },     42.01085,   0.00002,     0.077 },
 { {   2,  -2,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     42.01085,   0.00002,     0.074 },
 { {   2,   0,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    208.86962,   0.00001,     0.303 },
 { {   2,   0,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    209.41210,   0.00002,     0.068 },
 { {   2,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     29.43848,   0.00001,     0.036 },
 { {   2,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    321.33896,   0.00002,     0.232 },
 { {   2,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    320.50724,   0.00003,     0.063 },
 { {   2,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    140.51526,   0.00003,     0.034 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.79508,   0.00002,     0.051 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     71.90720,   0.00014,     0.189 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     70.90165,   0.00006,     0.060 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    250.90172,   0.00006,     0.033 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.16665,   0.00007,     0.054 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    180.29143,   0.00005,     0.193 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    180.40064,   0.00005,     0.144 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    180.23590,   0.00032,     0.159 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    180.17430,   0.00003,     0.051 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.89226,   0.00010,     0.056 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    359.88508,   0.00010,     0.032 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    359.90059,   0.00003,     0.023 },
 { {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    280.26363,   0.00002,     0.239 },
 { {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    290.55670,   0.00004,     0.137 },
 { {   2,   0,   3,   0,   0,   0,   0,   0,  -2,   0,   1 },    243.09409,   0.00002,    92.605 },
 { {   2,   0,   3,   0,   0,   0,   0,   0,  -2,   1,   0 },     64.56438,   0.00001,     5.632 },
 { {   2,   0,  17, -26,   0,   0,   0,   0,  -2,   1,   0 },    331.63943,   0.00004,   743.009 },
 { {   2,   1, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     93.74501,   0.00002,     0.075 },
 { {   2,   1, -10,   0,   0,   0,   0,   0,   0,   1,   0 },     93.74501,   0.00002,     0.076 },
 { {   2,   5,  -6,   0,   0,   0,   0,   0,  -3,   1,   1 },    118.16615,   0.00002,   261.378 },
 { {   2,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    291.46423,   0.00002,    19.007 },
 { {   2,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    324.36967,   0.00003,   867.364 },
 { {   2,  10, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    148.51042,   0.00001,     1.084 },
 { {   2,  10, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    216.99130,   0.00002,    12.874 },
 { {   2,  11, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    148.48578,   0.00001,     0.074 },
 { {   2,  11, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    332.49315,   0.00002,     3.369 },
 { {   2,  11, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    148.48578,   0.00001,     0.077 },
 { {   2,  11, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    240.99468,   0.00001,    14.121 },
 { {   2,  13, -19,   0,   0,   0,   0,   0,  -3,   1,   1 },    322.76352,   0.00002,   124.822 },
 { {   2,  13, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     39.86755,   0.00003,     0.075 },
 { {   2,  13, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     39.86755,   0.00003,     0.076 },
 { {   2,  14, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    313.64604,   0.00003,     2.380 },
 { {   2,  14, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    221.80121,   0.00003,     1.725 },
 { {   2,  14, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    140.84115,   0.00002,     5.148 },
 { {   2,  15, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    150.51108,   0.00002,     4.871 },
 { {   2,  16, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },     56.55430,   0.00002,     5.911 },
 { {   2,  16, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    216.89577,   0.00001,    17.606 },
 { {   2,  16, -20,   0,   0,   0,   0,   0,   1,  -2,  -1 },    145.01232,   0.00001,     0.075 },
 { {   2,  16, -20,   0,   0,   0,   0,   0,   1,  -1,  -1 },    145.01232,   0.00018,   329.750 },
 { {   2,  16, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },    145.01232,   0.00001,     0.075 },
 { {   2,  18, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },     79.29607,   0.00003,    12.216 },
 { {   2,  19, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },     93.21236,   0.00002,     3.418 },
 { {   2,  19, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },     41.53023,   0.00004,    15.008 },
 { {   2,  20, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    273.25045,   0.00001,     3.004 },
 { {   2,  24, -34,   0,   0,   0,   0,   0,   0,  -1,   0 },    185.16247,   0.00001,    16.398 },
 { {   2,  31, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },     84.86887,   0.00002,    23.758 },
 { {   2,  34, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },     40.81796,   0.00003,     6.042 },
 { {   3, -27,  20,   0,   0,   0,   0,   0,   2,  -1,   0 },    119.78825,   0.00002,    20.457 },
 { {   3,  -9,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    115.89615,   0.00001,     5.766 },
 { {   3,   0,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    321.64766,   0.00001,     0.134 },
 { {   3,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    141.06907,   0.00002,     0.031 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.90417,   0.00003,     0.061 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     73.06046,   0.00002,     0.331 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     72.12771,   0.00001,     0.208 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     72.26608,   0.00007,     0.118 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     71.19749,   0.00002,     0.050 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    251.19817,   0.00003,     0.030 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    181.15390,   0.00009,     0.065 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    181.32315,   0.00009,     0.494 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.11043,   0.00004,     0.263 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    181.62589,   0.00013,     0.106 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.44543,   0.00002,     0.048 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },      0.44573,   0.00005,     0.029 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   1,   1 },    312.75570,   0.00044,    35.335 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   2,   1 },    312.74905,   0.00002,     0.075 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     95.68887,   0.00012,     0.070 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     95.78247,   0.00013,     0.975 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    199.65617,   0.00001,     0.082 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     98.50362,   0.00002,     0.357 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    116.68690,   0.00005,     0.096 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    199.50229,   0.00002,     0.042 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    283.75096,   0.00002,     0.028 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },     96.56536,   0.00004,     0.026 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },     96.56536,   0.00034,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },     96.65389,   0.00032,     0.087 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },     96.29486,   0.00016,     0.038 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     96.29482,   0.00234,     0.075 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,  -2 },    276.56536,   0.00006,     0.037 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },     96.56536,   0.02923,    39.271 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   2 },    276.56536,   0.00006,     0.037 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },     96.48044,   0.00164,     0.076 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },     96.48036,   0.00011,     0.038 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     95.46738,   0.00018,     0.087 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     96.56536,   0.00034,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },     96.56536,   0.00004,     0.026 },
 { {   3,   0,   0,   0,  -3,   0,   0,   0,  -1,  -1,   1 },     98.15329,   0.00001,   729.668 },
 { {   3,   0,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },    353.62843,   0.00002,     0.042 },
 { {   3,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   0 },    353.62843,   0.00001,     0.095 },
 { {   3,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    353.62843,   0.00001,     0.081 },
 { {   3,   0,   0,   0,   0,   0,   0,   0,  -1,   1,  -1 },     49.24634,   0.00005,    12.622 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -4,   1,   0 },    353.96206,   0.00002,     0.040 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -4,   2,   0 },    353.96206,   0.00002,     0.087 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   0,   0 },    353.96206,   0.00009,     0.075 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   1,   0 },    353.96206,   0.00158,  1373.316 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   2,   0 },    353.96206,   0.00009,     0.075 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,   0,   0,   0 },    353.96206,   0.00002,     0.087 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    353.96206,   0.00002,     0.040 },
 { {   3,   2,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },    239.57624,   0.00002,     0.075 },
 { {   3,   2,  -6,   0,   0,   0,   0,   0,  -4,   3,   0 },    239.57624,   0.00031,  2494.868 },
 { {   3,   2,  -6,   0,   0,   0,   0,   0,  -4,   4,   0 },    239.57624,   0.00002,     0.075 },
 { {   3,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    345.57675,   0.00001,    21.579 },
 { {   3,  13, -20,   0,   0,   0,   0,   0,   0,   1,  -2 },    293.86041,   0.00008,   837.450 },
 { {   4, -29,  18,   0,   0,   0,   0,   0,   1,  -1,   1 },    342.68185,   0.00003,   270.651 },
 { {   4, -29,  19,   0,   0,   0,   0,   0,   2,  -1,   0 },    262.73450,   0.00002,    19.972 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -4,   3,   0 },    184.98192,   0.00001,     0.087 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   1,   0 },     10.24706,   0.00010,     0.075 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   2,   0 },     19.56685,   0.00069,  2144.232 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   3,   0 },     10.24706,   0.00010,     0.075 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,   0,   1,   0 },    184.98192,   0.00001,     0.087 },
 { {   4,  -7,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    314.11913,   0.00001,     0.075 },
 { {   4,  -7,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    314.11913,   0.00001,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -4,   1,   0 },     30.68945,   0.00002,     0.040 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -4,   2,   0 },     30.68945,   0.00002,     0.087 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },     30.68945,   0.00008,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   1,   0 },     30.68945,   0.00152,  1010.649 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },     30.68945,   0.00008,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     30.68945,   0.00002,     0.087 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,   0,   1,   0 },     30.68945,   0.00002,     0.040 },
 { {   4,   0,  -6,   0,   0,   0,   0,   0,  -3,   2,   0 },    143.41195,   0.00004,    73.341 },
 { {   4,   0,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    322.83154,   0.00002,     0.071 },
 { {   4,   0,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    322.87694,   0.00003,     1.145 },
 { {   4,   0,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    141.58981,   0.00001,     0.028 },
 { {   4,   0,  -6,   3,   0,   0,   0,   0,  -1,  -1,   1 },      4.44188,   0.00002,   700.544 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },     73.47726,   0.00003,     0.041 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },     73.47011,   0.00003,     0.088 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    253.74923,   0.00001,     0.038 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    253.74917,   0.00019,     0.076 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     73.47726,   0.00281,     7.897 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },     73.48408,   0.00016,     0.075 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },     73.48398,   0.00001,     0.038 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     71.05964,   0.00004,     0.607 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     72.79041,   0.00017,     0.086 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },     73.47726,   0.00003,     0.040 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    251.68541,   0.00002,     0.028 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -3,   1,  -1 },     24.75992,   0.00001,     0.041 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -3,   2,  -1 },     24.75942,   0.00001,     0.088 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.88890,   0.00012,     0.082 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    183.00504,   0.00007,     0.888 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    205.04924,   0.00001,     2.461 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    205.28899,   0.00002,     0.076 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },      2.43596,   0.00011,     4.173 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },     24.75992,   0.00114,    13.725 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   2,  -1 },     24.76040,   0.00006,     0.075 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    182.98484,   0.00002,     0.072 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    183.02988,   0.00035,     1.545 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    182.69471,   0.00024,     0.079 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },     24.82347,   0.00003,     0.087 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     24.75992,   0.00001,     0.040 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },      1.50649,   0.00003,     0.027 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     99.39311,   0.00002,     0.090 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,  -2 },    100.63105,   0.00005,     0.040 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     99.30534,   0.00002,     0.470 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   2,  -2 },    100.63105,   0.00005,     0.087 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },     99.28339,   0.00004,     0.078 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,  -2 },    100.63105,   0.00002,     0.038 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     99.36455,   0.00073,     2.836 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    100.77271,   0.00016,     0.076 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    313.25341,   0.00002,     5.381 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    279.29738,   0.00020,     0.073 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },    100.63105,   0.00427,    52.370 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   2,  -2 },    100.73599,   0.00024,     0.075 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   3,  -2 },    100.63105,   0.00002,     0.038 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    133.67317,   0.00002,     1.426 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },    100.63105,   0.00005,     0.087 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    283.54026,   0.00002,     0.026 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },    100.63105,   0.00005,     0.040 },
 { {   4,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     35.96047,   0.00002,     0.739 },
 { {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    215.97147,   0.00002,     0.068 },
 { {   4,   0,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },     52.05762,   0.00004,     3.346 },
 { {   4,   3, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     61.20913,   0.00002,     0.075 },
 { {   4,   3, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    332.74186,   0.00008,   312.886 },
 { {   4,   5,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    191.66472,   0.00003,   661.131 },
 { {   4,   6, -11,   0,   0,   0,   0,   0,   2,  -1,  -2 },    135.17538,   0.00001,   560.934 },
 { {   4,  11, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    214.85098,   0.00004,   135.472 },
 { {   4,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     22.16816,   0.00002,   374.117 },
 { {   4,  17, -31,   0,   0,   0,   0,   0,   0,  -1,   0 },    313.52255,   0.00001,    74.095 },
 { {   4,  21, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     21.64885,   0.00001,   145.805 },
 { {   4,  29, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.14764,   0.00002,    90.547 },
 { {   5, -32,  18,   0,   0,   0,   0,   0,   0,   1,   0 },    257.06814,   0.00009,  1880.011 },
 { {   5, -14,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    335.89670,   0.00001,   319.437 },
 { {   5, -12,  12,   0,   0,   0,   0,   0,   0,  -1,   0 },    250.86844,   0.00003,   633.672 },
 { {   5,  -4,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    111.03403,   0.00001,     0.075 },
 { {   5,  -4,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    111.03403,   0.00021,   383.522 },
 { {   5,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    111.03403,   0.00001,     0.075 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   0,   0 },    303.08124,   0.00003,     0.075 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   1,   0 },    303.08124,   0.00053,  1971.354 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   2,   0 },    303.08124,   0.00003,     0.075 },
 { {   5,   0,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    214.19978,   0.00002,     2.019 },
 { {   5,   0,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    322.55127,   0.00006,     1.981 },
 { {   5,   0,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    142.90780,   0.00003,     0.073 },
 { {   5,   0,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     73.66632,   0.00002,     0.665 },
 { {   5,   0,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    246.33393,   0.00003,     0.068 },
 { {   5,   0,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    252.40635,   0.00001,     0.025 },
 { {   5,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },      3.42225,   0.00001,     0.025 },
 { {   5,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    283.79840,   0.00001,     0.024 },
 { {   5,   4, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    188.67468,   0.00002,   147.212 },
 { {   6, -39,  27,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.43578,   0.00006,   892.417 },
 { {   6, -31,  14,   0,   0,   0,   0,   0,   2,  -1,   0 },     97.57590,   0.00001,   326.268 },
 { {   6,   0,  -3,   0,   0,   0,   0,   0,  -5,   2,   1 },    155.15702,   0.00005,   352.558 },
 { {   6,   0,   0,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    231.51527,   0.00001,     0.075 },
 { {   6,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },    231.51527,   0.00026,   155.167 },
 { {   6,   0,   0,   0,  -2,   0,   0,   0,  -2,   1,   0 },    231.51527,   0.00001,     0.075 },
 { {   6,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },    235.02094,   0.00001,     0.076 },
 { {   6,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    235.02094,   0.00024,     5.713 },
 { {   6,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    235.02094,   0.00001,     0.074 },
 { {   7,   0,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    280.64524,   0.00002,     0.598 },
 { {   7,   0,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    208.70560,   0.00010,     1.486 },
 { {   7,   0,  -4,   0,  -4,   0,   0,   0,  -2,   0,   0 },     41.48953,   0.00003,    99.278 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    138.00381,   0.00001,     0.077 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    138.00381,   0.00025,     3.057 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    138.00381,   0.00001,     0.074 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,   2,  -2 },    137.63269,   0.00002,   157.008 },
 { {   7,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },     83.20050,   0.00003,     3.658 },
 { {   8,  -4,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     31.09549,   0.00003,    43.751 },
 { {   8,   0,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    261.32095,   0.00002,     1.920 },
 { {   8,   0,  -9,   1,   0,   0,   0,   0,  -2,   0,   0 },      7.56929,   0.00002,    91.577 },
 { {   8,   4, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    233.87755,   0.00001,    53.558 },
 { {   9, -19,   5,   0,   0,   0,   0,   0,  -2,   1,   0 },    104.65047,   0.00002,   467.061 },
 { {   9,  -3, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    220.49519,   0.00001,    95.129 },
 { {   9,  -1, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.25700,   0.00006,   172.311 },
 { {  10,   0, -12,  -9,   0,   0,   0,   0,  -2,   0,   0 },    352.79254,   0.00001,   574.613 },
 { {  11,   0, -22,   2,   0,   0,   0,   0,  -2,   0,   0 },    113.75419,   0.00005,   890.800 },
 { {  12,   0, -12,   0,   2,   0,   0,   0,  -2,  -1,   0 },    207.32678,   0.00017,  2204.260 },
 { {  17,   0, -22,   0,   0,   0,   0,   0,  -5,   1,   0 },    332.53824,   0.00002,   411.722 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 5233> g_ELP11
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 },    179.93197,   0.00068,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1 },    359.92861,   0.00007,     5.997 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1 },    359.97739,   0.00040,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1 },    359.99484,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   1 },    359.97313,   0.00006,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   1 },    359.89152,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1 },    180.03960,   0.00001,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    356.53630,   0.00002,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },    180.03960,   0.00001,     0.080 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    180.01772,   0.00018,     0.949 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },     76.40971,   0.00052,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1 },    180.01674,   0.00019,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },     45.97098,   0.00007,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   1 },    180.66992,   0.00001,     0.622 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1 },      0.06750,   0.00015,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1 },      0.09788,   0.00143,     0.515 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1 },      0.10790,   0.00162,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },      0.14435,   0.00321,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   1 },      0.12799,   0.00081,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1 },      0.15148,   0.00021,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   1 },      0.14122,   0.00010,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },     34.89815,   0.00007,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 },     34.89815,   0.00006,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -2,   1 },      0.12097,   0.00003,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1 },      0.11370,   0.00008,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1 },      0.10938,   0.00004,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.11543,   0.00005,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   1 },      0.10670,   0.00001,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -1 },    261.18864,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   1 },    261.18864,   0.00001,     5.786 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,  -1 },    261.18864,   0.00001,     6.223 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   1 },    261.18864,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -1,  -1 },     99.71523,   0.00005,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -1,   1 },    102.00892,   0.00005,     6.005 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,  -1 },    330.59903,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   1 },    330.59903,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   1,  -1 },    102.00892,   0.00005,     5.988 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   1,   1 },     99.71523,   0.00005,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   1 },    184.37205,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,  -1 },    236.39187,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   1 },    236.39187,   0.00002,     5.597 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,  -1 },    236.39187,   0.00002,     6.458 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   1 },    236.39187,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,  -1 },    186.06606,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   1,  -1 },    267.93817,   0.00001,     0.020 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -1,  -1 },    267.93448,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -1,   1 },    266.91265,   0.00008,     0.041 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,  -1 },    267.32323,   0.00021,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   1 },    267.41538,   0.00101,     0.089 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,  -1 },    269.36401,   0.00024,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   1 },    269.57119,   0.00020,     0.506 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   2,   1 },    273.81956,   0.00001,     0.066 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0 },     63.31391,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1 },    261.43413,   0.00004,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1 },    261.29928,   0.00002,     0.077 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1 },    260.89638,   0.00038,     0.038 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1 },    261.08984,   0.00040,     4.983 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1 },     77.86313,   0.00007,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   1 },     78.07459,   0.00007,     0.074 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1 },    261.05509,   0.00039,     7.530 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   1 },    260.83894,   0.00038,     0.037 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1 },    261.31633,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   2,   1 },    261.44244,   0.00004,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0 },     66.17816,   0.00006,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -2,  -1 },    273.51283,   0.00001,     0.066 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,  -1 },    269.41265,   0.00020,     0.524 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,   1 },    269.09604,   0.00024,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,  -1 },    266.95423,   0.00103,     0.088 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   1 },    267.30520,   0.00020,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   1,  -1 },    266.45286,   0.00008,     0.041 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   1,   1 },    267.72236,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,  -1,   1 },    267.44696,   0.00001,     0.020 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,  -1 },    167.01997,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   1 },    164.85104,   0.00006,     0.089 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0 },    324.55694,   0.00004,     0.081 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,  -1 },    324.48914,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,   1 },    324.48914,   0.00002,     4.262 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1 },    151.65684,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   1 },    151.65684,   0.00002,     0.074 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,  -1 },    324.48914,   0.00002,    10.115 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,   1 },    324.48914,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   1,   0,   0 },    139.43202,   0.00002,     0.080 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,   0,  -1 },    168.87091,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,  -2,   0,   1 },    200.35349,   0.00002,     0.088 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   2,   0,  -1 },    193.55451,   0.00002,     0.089 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   0,  -1 },    355.56931,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   0,   1 },    355.56923,   0.00007,     0.089 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -1,  -1 },    298.68935,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -1,   1 },    298.68935,   0.00003,     5.460 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,  -1 },     47.02190,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   1 },     47.02190,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   1,  -1 },    298.68935,   0.00003,     6.651 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   1,   1 },    298.68935,   0.00003,     0.037 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,   0,  -1 },    353.65395,   0.00007,     0.088 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,   0,   1 },    353.65395,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   0,   1 },    177.14581,   0.00004,     0.089 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,   0,  -1 },    176.87334,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   0,  -1 },    351.25342,   0.00004,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   0,   1 },    354.03243,   0.00010,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   1,  -1 },    354.16463,   0.00008,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   1,   1 },    358.47110,   0.00015,     0.044 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   2,  -1 },      0.79676,   0.00006,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   2,   1 },    354.86504,   0.00002,     0.106 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -2,  -1 },    355.37667,   0.00004,     0.015 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -2,   1 },    344.45685,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,  -1 },    347.06646,   0.00024,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   1 },    349.19248,   0.00050,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,  -1 },    354.59152,   0.00164,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   1 },    354.87377,   0.00837,     0.089 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   3 },    357.18904,   0.00003,     0.064 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,  -1 },    359.57886,   0.00249,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   1 },    358.81236,   0.00208,     0.494 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   2,  -1 },    184.95790,   0.00004,     0.086 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   2,   1 },      6.05703,   0.00020,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   0 },     78.88306,   0.00006,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,  -1 },    178.57582,   0.00005,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 },     79.94322,   0.00119,     0.081 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   1 },    176.76452,   0.00005,     0.879 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   0 },    258.88306,   0.00006,     1.030 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -3,  -1 },    291.10766,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -1 },    295.12411,   0.00022,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1 },    293.67293,   0.00010,     0.077 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,  -1 },    295.95314,   0.00178,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   1 },    294.28372,   0.00181,     3.983 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1 },    102.39298,   0.00053,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   1 },    103.21455,   0.00049,     0.074 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1 },    294.08048,   0.00181,    12.128 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   1 },    294.52112,   0.00180,     0.037 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1 },    293.15461,   0.00010,     0.076 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,   1 },    293.90406,   0.00021,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   3,   1 },    290.19587,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0 },    258.38906,   0.00009,     1.246 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   2 },    257.20040,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,  -2 },     78.52681,   0.00001,     0.069 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,  -1 },    174.06643,   0.00003,     1.032 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },     78.54130,   0.00142,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   1 },    174.06773,   0.00003,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   1,   0 },     78.55035,   0.00008,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -2,  -1 },      6.80143,   0.00018,     0.066 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -2,   1 },    154.62865,   0.00003,     0.085 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,  -1 },    356.90677,   0.00194,     0.539 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   1 },    356.39811,   0.00230,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,  -3 },    346.45726,   0.00001,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,  -1 },    346.14598,   0.00779,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   1 },    346.81478,   0.00143,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,  -1 },    343.48259,   0.00047,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,   1 },    339.26577,   0.00021,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   2,  -1 },    325.27306,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   2,   1 },    351.91873,   0.00002,     0.015 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,  -1,   0 },    178.44571,   0.00002,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -2,  -1 },    352.08542,   0.00002,     0.104 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -2,   1 },    357.78140,   0.00006,     0.027 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -1,  -1 },    353.41059,   0.00014,     0.044 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -1,   1 },    351.70810,   0.00007,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   0,  -1 },    350.54032,   0.00010,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   0,   1 },    352.65785,   0.00003,     0.016 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -1,  -1 },      0.04391,   0.00003,     0.037 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -1,   1 },      0.04391,   0.00002,     5.974 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   1,  -1 },      0.04391,   0.00002,     6.019 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   1,   1 },      0.04391,   0.00003,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   1,   1 },    165.96095,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   2,  -1 },    165.63973,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -2,   1 },    345.15966,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -1,  -1 },    333.09842,   0.00005,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -1,   1 },    329.35036,   0.00010,     0.041 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,  -1 },    251.06582,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   1 },    263.99803,   0.00041,     0.088 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,  -1 },    178.13381,   0.00061,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   1 },    178.93841,   0.00050,     0.516 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   2,   1 },    169.58626,   0.00009,     0.066 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   3,   1 },    166.63073,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -3,  -1 },    345.96095,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -3,   1 },    345.96095,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,  -1 },    346.25039,   0.00037,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   1 },    346.25189,   0.00019,     0.076 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,  -1 },    346.45683,   0.00299,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   1 },    346.26978,   0.00305,     6.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,  -1 },    161.53001,   0.00086,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   1 },    161.53396,   0.00086,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,  -1 },    346.27178,   0.00305,     5.956 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   1 },    346.45910,   0.00299,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,  -1 },    346.25133,   0.00019,     0.076 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   1 },    346.25018,   0.00037,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   3,  -1 },    345.96095,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   3,   1 },    345.96095,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,   0,   0 },    342.20207,   0.00002,     0.081 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -3,  -1 },    166.64833,   0.00001,     0.035 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -2,  -1 },    169.56755,   0.00009,     0.066 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,  -1 },    178.84686,   0.00050,     0.515 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   1 },    178.04116,   0.00062,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,  -1 },    261.34897,   0.00040,     0.088 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,   1 },    249.99781,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   1,  -1 },    329.11836,   0.00010,     0.041 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   1,   1 },    333.02115,   0.00005,     0.019 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   2,  -1 },    345.12184,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -2,   1 },    165.63638,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -1,  -1 },    165.96095,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,   0,   1 },    250.78551,   0.00002,     0.089 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,  -1 },    240.49213,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   1 },    240.49213,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,   0,  -1 },    251.03214,   0.00002,     0.088 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,   0,   1 },    300.40427,   0.00003,     0.089 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,   0,  -1 },    300.36962,   0.00002,     0.088 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   0,  -1 },    359.17033,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   0,   1 },    359.17917,   0.00007,     0.089 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,  -1,  -1 },    356.92413,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,  -1,   1 },    356.92413,   0.00001,     3.739 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   1,  -1 },    356.92413,   0.00001,    15.142 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   1,   1 },    356.92413,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   0,  -1 },    358.65392,   0.00007,     0.088 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   0,   1 },    358.65380,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -3,   0,   0 },    336.55581,   0.00002,     0.027 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -3,   1,   0 },    336.55586,   0.00002,     0.042 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,  -1 },     19.82961,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   1 },     14.23182,   0.00016,     0.090 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,  -1 },     23.79540,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,   1 },     20.42049,   0.00005,     0.474 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,  -1,   0 },    336.55572,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0 },     30.81684,   0.00020,     0.082 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -1 },    320.30722,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   1 },    320.30722,   0.00001,     0.077 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1 },    319.76670,   0.00020,     0.038 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1 },    320.05216,   0.00019,     2.982 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,  -2 },    132.88991,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1 },    143.79604,   0.00008,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   1 },    142.16659,   0.00008,     0.074 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,  -2 },    136.63823,   0.00006,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1 },    319.87148,   0.00020,   540.002 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   0 },    132.72694,   0.00010,     0.074 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   1 },    320.12063,   0.00019,     0.037 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1 },    320.30722,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,   0 },    132.88991,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,   1 },    320.30722,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0 },    101.53678,   0.00028,     0.080 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -1,  -1 },      2.13348,   0.00005,     0.564 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -1,   1 },    356.14373,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,  -1 },    333.81865,   0.00035,     0.087 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   1 },    336.70338,   0.00006,     0.026 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   0,  -1 },     83.97737,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   0,   1 },     84.75757,   0.00006,     0.089 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,  -1,  -1 },     87.88220,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,  -1,   1 },     87.88220,   0.00001,     4.001 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   1,  -1 },     87.88220,   0.00001,    11.964 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   1,   1 },     87.88220,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,  -1 },     84.80211,   0.00005,     0.088 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   1 },     84.80209,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -1,   0,   0 },     24.02389,   0.00001,     0.083 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -1,   0,   2 },     24.02389,   0.00001,     0.068 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,  -1,  -1 },    314.35081,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,  -1,   1 },    314.35081,   0.00001,     2.383 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,  -1 },    134.79108,   0.00001,     0.076 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   1 },    134.79108,   0.00001,     0.073 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   1,  -1 },    314.35081,   0.00001,    11.607 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   1,   1 },    314.35081,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   1,   0,   0 },    107.97562,   0.00004,     0.079 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,  -1,  -1 },    177.39404,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,  -1 },    181.08871,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,   1 },    181.08871,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   1,   1 },    177.39404,   0.00001,     0.037 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,  -1 },     38.91751,   0.00005,     0.027 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1 },     38.65254,   0.00020,     0.093 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1,  -1 },     38.65549,   0.00006,     0.041 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1,   1 },     38.65675,   0.00006,     0.404 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0 },    128.06358,   0.00002,     0.040 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },    127.99528,   0.00026,     0.084 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0 },    308.06358,   0.00002,     0.705 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1 },     27.22252,   0.00001,     0.038 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1 },     27.22252,   0.00001,     1.432 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1 },    189.21294,   0.00002,     0.078 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1 },    193.86844,   0.00002,     0.072 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1 },     27.22252,   0.00001,     2.740 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   1 },     27.22252,   0.00001,     0.037 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0 },    308.77663,   0.00005,     2.814 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0 },    128.77680,   0.00087,     0.078 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   1,   0 },    128.77663,   0.00005,     0.038 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1,  -1 },     31.07464,   0.00004,     0.710 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1,   1 },     31.07573,   0.00004,     0.039 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,  -1 },     27.84341,   0.00011,     0.084 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,   1 },     32.69161,   0.00002,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -1,  -1 },    136.60870,   0.00001,     0.020 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,  -1 },    131.46616,   0.00011,     0.027 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   1 },    132.33948,   0.00038,     0.098 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   1,  -1 },    136.42319,   0.00010,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   1,   1 },    139.59362,   0.00008,     0.333 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1,   0 },    343.75667,   0.00004,     0.041 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,  -2 },    163.74761,   0.00005,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0 },    343.88402,   0.00131,     0.088 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1,  -2 },    343.75556,   0.00001,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1,   0 },    343.75311,   0.00038,     0.513 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   2,   0 },    343.75408,   0.00004,     0.066 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,  -1 },    148.51213,   0.00008,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,   1 },    145.97182,   0.00005,     0.813 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1 },    202.19193,   0.00003,     0.081 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   1 },    207.37132,   0.00003,     0.069 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,  -1 },    183.47784,   0.00002,     1.115 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,   1 },    195.85001,   0.00003,     0.036 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -2,  -2 },    163.75558,   0.00007,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -2,   0 },    343.74723,   0.00004,     0.076 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -2 },    163.76157,   0.00059,     0.037 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,   0 },    343.74996,   0.00163,     5.670 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,  -2 },    163.82402,   0.00188,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,   0 },    163.75018,   0.03884,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,   2 },    343.75118,   0.00004,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -2 },    163.72838,   0.00039,     6.363 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,   0 },    163.74937,   0.00261,     0.037 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   2,  -2 },    163.74326,   0.00002,     0.076 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   2,   0 },    163.74921,   0.00019,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,  -1 },    277.59253,   0.00007,     0.081 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,   1 },    103.58964,   0.00001,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1,  -2 },    343.80445,   0.00005,     0.518 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1,   0 },    163.74551,   0.00038,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   0,  -2 },    163.67271,   0.00013,     0.088 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   0,   0 },    163.74608,   0.00028,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   1,  -2 },    163.72493,   0.00002,     0.041 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   1,   0 },    163.74462,   0.00004,     0.019 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   1,   0,   0 },      1.87727,   0.00006,     0.072 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -1,   0,   0 },    145.97363,   0.00002,     0.071 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,  -1,   0 },    345.71481,   0.00004,     0.037 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   0 },    345.71813,   0.00078,     0.074 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   2 },    345.88377,   0.00001,     0.075 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   1,   0 },    165.71481,   0.00004,     4.171 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -1 },     25.35364,   0.00001,     0.069 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   1 },     33.71797,   0.00003,     0.081 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   1,   0,   0 },    164.69789,   0.00004,     0.089 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -1,  -1 },     41.03148,   0.00004,     0.326 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -1,   1 },     41.03083,   0.00004,     0.042 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,  -1 },     41.00287,   0.00012,     0.098 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,   1 },     39.42259,   0.00004,     0.027 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,  -1,   0 },     21.08161,   0.00003,     0.038 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,   0 },     21.08161,   0.00051,     0.077 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   1,   0 },    201.08161,   0.00003,     3.425 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1 },    325.41258,   0.00002,     0.071 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1 },    325.41258,   0.00002,     0.078 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   1,   0,   0 },     23.41995,   0.00016,     0.085 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,  -1 },     83.58432,   0.00007,     0.093 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,   1 },     79.19742,   0.00002,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   0,  -1 },    319.40267,   0.00002,     0.016 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   0,   1 },    318.78174,   0.00008,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   1,  -1 },    318.95058,   0.00006,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   1,   1 },    319.55382,   0.00010,     0.044 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   2,  -1 },    321.52419,   0.00003,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   2,   1 },    319.15823,   0.00001,     0.104 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   0,   0 },    273.31123,   0.00002,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   1,   0 },    273.31125,   0.00002,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -2,  -1 },    319.20274,   0.00002,     0.015 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -2,   1 },    313.43785,   0.00001,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,  -1 },    318.20184,   0.00019,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   1 },    317.41767,   0.00043,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,  -1 },    318.43293,   0.00116,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   1 },    318.17659,   0.00615,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   3 },    321.19337,   0.00002,     0.065 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,  -1 },    321.35166,   0.00141,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   1 },    321.43461,   0.00113,     0.533 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   2,  -1 },    134.41425,   0.00003,     0.085 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   2,   1 },    327.21693,   0.00006,     0.066 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,  -1,   0 },     52.99205,   0.00002,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,  -1 },    139.11613,   0.00003,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   0 },     52.99174,   0.00038,     0.080 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   1 },    139.46717,   0.00002,     1.010 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   1,   0 },    232.99205,   0.00002,     1.214 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -3,  -1 },    313.60374,   0.00004,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1 },    313.64198,   0.00035,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1 },    312.97875,   0.00017,     0.076 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1 },    314.04103,   0.00297,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1 },    312.90846,   0.00265,     9.672 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1 },    324.71916,   0.00258,     0.074 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1 },    324.90632,   0.00253,     0.075 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1 },    312.90551,   0.00267,     4.346 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1 },    314.21189,   0.00299,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1 },    313.06461,   0.00017,     0.077 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   1 },    313.80605,   0.00036,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   3,   1 },    313.71692,   0.00004,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,  -1,   0 },    232.66818,   0.00002,     1.052 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,  -1 },    137.69416,   0.00003,     0.895 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,   0 },     50.41465,   0.00039,     0.081 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,   1 },    137.80926,   0.00004,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   1,   0 },     52.66818,   0.00002,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -2,  -1 },    326.21148,   0.00008,     0.066 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -2,   1 },    142.87779,   0.00002,     0.086 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,  -1 },    321.59414,   0.00128,     0.499 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   1 },    321.81656,   0.00157,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,  -3 },    319.69152,   0.00002,     0.064 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,  -1 },    320.08151,   0.00634,     0.089 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   1 },    320.13750,   0.00127,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,  -1 },    318.60690,   0.00045,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,   1 },    318.61091,   0.00022,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,  -1 },    313.54800,   0.00001,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,   1 },    320.16164,   0.00002,     0.015 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -2,  -1 },    319.98162,   0.00001,     0.106 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -2,   1 },    322.39290,   0.00003,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -1,  -1 },    321.07161,   0.00010,     0.044 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -1,   1 },    319.71584,   0.00006,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   0,  -1 },    319.77961,   0.00009,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   0,   1 },    319.30288,   0.00003,     0.016 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   0,  -1 },     33.82493,   0.00001,     0.026 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   0,   1 },     33.82493,   0.00005,     0.088 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,  -1,  -1 },     28.87695,   0.00002,     0.037 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,  -1,   1 },     26.57329,   0.00001,     5.838 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0,  -1 },     44.03817,   0.00002,     0.075 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0,   1 },     44.03817,   0.00002,     0.074 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   1,  -1 },     26.57329,   0.00001,     6.164 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   1,   1 },     28.87695,   0.00002,     0.037 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,   0,  -1 },     33.77499,   0.00005,     0.088 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,   0,   1 },     33.77499,   0.00001,     0.026 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,  -1 },    196.88833,   0.00001,     0.075 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   1 },    196.88833,   0.00001,     0.074 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   1,  -1 },      0.64803,   0.00001,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -1,  -1 },      0.64287,   0.00005,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -1,   1 },      0.63762,   0.00008,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,  -1 },      0.64947,   0.00022,     0.027 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   1 },      0.63791,   0.00102,     0.092 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,  -1 },      0.68639,   0.00030,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   1 },      0.67872,   0.00026,     0.415 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   2,   1 },      0.63781,   0.00003,     0.064 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   0 },     92.49532,   0.00014,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1 },      0.40408,   0.00002,     0.025 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1 },      0.58625,   0.00017,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1 },      0.48341,   0.00017,     1.575 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1 },      2.47987,   0.00009,     0.077 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1 },      3.47280,   0.00006,     0.072 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1 },      0.37430,   0.00014,     3.316 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1 },      0.47112,   0.00013,     0.037 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1 },      0.40193,   0.00002,     0.025 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,  -1,   0 },    274.46515,   0.00002,     2.388 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0 },     94.46515,   0.00046,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   1,   0 },     94.46515,   0.00003,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -1,  -1 },      0.35803,   0.00016,     0.679 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -1,   1 },      0.36103,   0.00021,     0.039 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,  -1 },      0.28839,   0.00071,     0.085 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   1 },      0.33807,   0.00011,     0.026 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   1,  -1 },      0.25930,   0.00003,     0.040 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   1,   1 },      0.30457,   0.00001,     0.019 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   1,   0,   0 },    129.72692,   0.00002,     0.077 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   1,   0,   0 },    328.56107,   0.00003,     0.076 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   0,  -1 },    317.01304,   0.00003,     0.027 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   0,   1 },    317.56134,   0.00009,     0.094 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   1,  -1 },    317.17349,   0.00003,     0.041 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   1,   1 },    317.17408,   0.00003,     0.384 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -1,   0,   0 },    122.26209,   0.00006,     0.085 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -1,  -1 },    313.45519,   0.00003,     0.038 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -1,   1 },    313.72619,   0.00002,     1.206 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,  -1 },    314.67902,   0.00002,     0.078 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   1 },    314.67902,   0.00002,     0.071 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   1,  -1 },    313.72619,   0.00002,     2.016 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   1,   1 },    313.45519,   0.00003,     0.037 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,  -1,   0 },    300.64203,   0.00002,     4.460 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,   0,   0 },    120.64203,   0.00037,     0.077 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,   1,   0 },    120.64203,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,   0,  -1 },    316.85271,   0.00003,     0.083 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,   0,   0 },    101.05921,   0.00005,     0.076 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,  -1,   0,   0 },    271.81407,   0.00003,     0.087 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   1,   0,   0 },    271.89429,   0.00028,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   0,   1 },    309.20586,   0.00002,     0.028 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   1,  -1 },    309.30383,   0.00001,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -1,  -1 },    308.72530,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -1,   1 },    309.23156,   0.00005,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,  -1 },    308.48055,   0.00023,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   1 },    309.02448,   0.00085,     0.095 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,  -1 },    309.09498,   0.00026,     0.041 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,   1 },    309.04054,   0.00023,     0.372 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   2,   1 },    308.37714,   0.00003,     0.063 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,  -1,   2 },    290.64572,   0.00002,     0.508 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,   0 },    111.64125,   0.00048,     0.086 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   1,   0 },    291.16086,   0.00007,     0.612 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -2,  -1 },    305.36312,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -2,   1 },    299.78544,   0.00002,     0.081 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,  -1 },    306.30471,   0.00022,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,   1 },    306.10966,   0.00024,     1.094 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,  -1 },    303.24119,   0.00020,     0.079 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   1 },    304.59841,   0.00022,     0.071 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,  -1 },    304.88123,   0.00018,     1.723 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   1 },    304.68426,   0.00020,     0.036 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   2,  -1 },    305.57996,   0.00001,     0.072 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   2,   1 },    305.39692,   0.00003,     0.025 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -3,   0 },    110.91875,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -2,   0 },    110.98078,   0.00035,     0.075 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -2,   2 },    290.62235,   0.00004,     0.074 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,   0 },    291.10467,   0.00016,     7.147 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,   2 },    110.66970,   0.00012,     0.037 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,  -2 },    112.97505,   0.00003,     0.073 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,   0 },    110.92166,   0.00472,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   1,   0 },    110.92407,   0.00025,     0.038 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -2,  -1 },    112.98692,   0.00001,     0.069 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,  -1 },     88.63337,   0.00006,     0.838 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,   1 },     88.19020,   0.00007,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,  -1 },    308.41591,   0.00047,     0.083 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,   1 },    323.44147,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   1,  -1 },    308.67856,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   3,  -1,   0 },    110.92413,   0.00005,     0.041 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   3,   0,   0 },    110.92417,   0.00003,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -3,   1 },    112.98994,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -2,  -1 },    112.95625,   0.00003,     0.097 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -2,   1 },    112.99951,   0.00004,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -1,  -1 },    113.02274,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -1,   1 },    113.02248,   0.00002,     0.020 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   1,   0,   0 },    109.39031,   0.00002,     0.076 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   0,  -1 },    172.56769,   0.00002,     0.027 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   0,   1 },    172.53186,   0.00007,     0.096 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,  -1,  -1 },    173.00124,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,  -1,   1 },    173.00124,   0.00001,     0.923 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   0,  -1 },    171.13715,   0.00001,     0.080 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   0,   1 },    171.13715,   0.00001,     0.070 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   1,  -1 },    173.00124,   0.00001,     1.333 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   1,   1 },    173.00124,   0.00001,     0.036 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   1,   0,   0 },      4.12068,   0.00007,     0.075 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,   0,  -1 },    172.49600,   0.00002,     0.082 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,  -1,   1 },    305.97672,   0.00002,     0.021 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   0,  -1 },    305.25529,   0.00003,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   0,   1 },    305.62310,   0.00008,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   1,  -1 },    305.62745,   0.00007,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   1,   1 },    306.19034,   0.00011,     0.046 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   2,  -1 },    305.59588,   0.00005,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   2,   1 },    305.48706,   0.00001,     0.115 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   0,   0 },    100.50635,   0.00003,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   1,   0 },    100.50635,   0.00005,     0.043 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -2,  -1 },    305.79455,   0.00004,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -2,   1 },    306.02965,   0.00002,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,  -1 },    304.73794,   0.00026,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,   1 },    305.08745,   0.00043,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,  -1 },    305.95343,   0.00169,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   1 },    305.98316,   0.00663,     0.095 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   3 },    306.68180,   0.00002,     0.061 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,  -1 },    305.73983,   0.00201,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   1 },    305.70839,   0.00168,     0.361 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   2,  -1 },    131.18668,   0.00002,     0.092 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   2,   1 },    306.22014,   0.00016,     0.062 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   3,   1 },    305.79972,   0.00002,     0.034 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   0 },    100.50191,   0.00015,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   2 },    280.52528,   0.00003,     0.487 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,  -1 },    125.85143,   0.00011,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0 },    100.50673,   0.00349,     0.087 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   1 },    125.82852,   0.00010,     0.530 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   2 },    100.50444,   0.00004,     0.065 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,  -2 },    101.29685,   0.00001,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,   0 },    280.50859,   0.00086,     0.582 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   2,   0 },    280.50978,   0.00008,     0.067 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -3,  -1 },    302.36815,   0.00001,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,  -1 },    300.18484,   0.00022,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   1 },    302.14370,   0.00010,     0.082 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,  -1 },    300.77514,   0.00177,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   1 },    300.69538,   0.00147,     1.002 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -1 },    300.48941,   0.00115,     0.079 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   1 },    300.94576,   0.00148,     0.070 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   3 },    149.83671,   0.00001,     0.024 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,  -1 },    301.64591,   0.00134,     1.505 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   1 },    301.42728,   0.00139,     0.036 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -1 },    302.05472,   0.00006,     0.072 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   2,   1 },    301.02148,   0.00015,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   3,   1 },    301.47419,   0.00001,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -3,   0 },    100.52470,   0.00004,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,  -1 },    305.51949,   0.00002,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   0 },    100.52668,   0.00061,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   2 },    280.54444,   0.00002,     0.074 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,  -2 },    101.13342,   0.00009,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,  -1 },    306.25983,   0.00006,     0.074 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   0 },    280.50698,   0.00266,    17.980 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   1 },    308.30407,   0.00003,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   2 },    100.52633,   0.00019,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -2 },    103.09351,   0.00038,     0.073 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -1 },    306.09016,   0.00030,     4.497 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },    100.50752,   0.04964,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   1 },    306.06097,   0.00032,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   2 },    280.49972,   0.00003,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,  -2 },     99.69207,   0.00006,     2.570 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,  -1 },    306.09016,   0.00002,     0.077 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   0 },    100.50794,   0.00279,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   1 },    306.09015,   0.00004,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   2,   0 },    100.50713,   0.00016,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -2,  -1 },    289.41574,   0.00003,     0.070 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -2,   1 },    144.18597,   0.00002,     0.080 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,  -1 },    301.80675,   0.00063,     0.901 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   1 },    301.29000,   0.00083,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,  -3 },    304.87871,   0.00001,     0.068 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,  -1 },    305.89091,   0.00338,     0.082 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   1 },    302.43516,   0.00051,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   1,  -1 },    305.72842,   0.00023,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   1,   1 },    305.40154,   0.00008,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,   0 },    280.52470,   0.00002,     0.088 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,   0 },    100.50770,   0.00054,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,  -2 },     99.25223,   0.00003,     0.090 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,   0 },    100.50801,   0.00032,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   1,   0 },    100.50731,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -3,   1 },    103.44847,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -2,  -1 },    100.75656,   0.00001,     0.096 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -2,   1 },    289.32177,   0.00006,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -1,  -1 },    290.79664,   0.00013,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -1,   1 },    285.65733,   0.00007,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,   0,  -1 },    306.05664,   0.00002,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,   0,   1 },    130.71738,   0.00002,     0.016 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   1,   0,   0 },    111.74346,   0.00009,     0.075 },
 { {   0,   0,   1,   0,  -2,   3,   0,   0,   1,   0,   0 },    108.19019,   0.00024,     0.075 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,   0,  -1 },    350.55637,   0.00001,     0.080 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,   0,   1 },    350.55637,   0.00001,     0.069 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   1,   0,   0 },    188.55882,   0.00009,     0.075 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   0,  -1 },    179.22582,   0.00001,     0.027 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   0,   1 },    188.00181,   0.00004,     0.097 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   1,  -1 },    271.10212,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   1,   1 },    271.10212,   0.00002,     0.340 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -1,  -1 },     91.10212,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -1,   1 },     91.10212,   0.00001,     0.856 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   1,  -1 },    271.10212,   0.00001,     1.199 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   1,   1 },    271.10212,   0.00001,     0.036 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,  -1,   0 },     86.29898,   0.00006,     8.763 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,   0,   0 },    266.29936,   0.00105,     0.075 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,   1,   0 },    266.29936,   0.00006,     0.038 },
 { {   0,   0,   1,   0,  -2,   6,   0,   0,   1,   0,   0 },    270.45839,   0.00008,     0.075 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   1,   0,   0 },    110.43573,   0.00003,     0.076 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,   1,   0,   0 },    271.70895,   0.00003,     0.076 },
 { {   0,   0,   1,   0,  -1,  -1,   0,   0,   1,   0,   0 },    293.88307,   0.00003,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   1,  -1 },      1.76268,   0.00001,     0.014 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   1,   1 },      1.76268,   0.00001,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   3,  -1 },      1.66425,   0.00001,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -1,  -1 },      0.82999,   0.00003,     0.013 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -1,   1 },      1.38188,   0.00006,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,  -1 },      0.47473,   0.00017,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,   1 },      1.21567,   0.00041,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,  -1 },      1.54823,   0.00037,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,   1 },      1.67271,   0.00062,     0.046 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   2,  -1 },      1.72742,   0.00021,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   2,   1 },      1.49130,   0.00003,     0.116 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   0,  -1 },      2.01887,   0.00002,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -2,  -1 },      1.59843,   0.00015,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -2,   1 },      1.68689,   0.00016,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,  -1 },      1.57500,   0.00114,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,   1 },      1.66607,   0.00182,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,  -1 },      1.79585,   0.00804,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   1 },      1.77997,   0.03056,     0.096 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   3 },      2.89794,   0.00009,     0.061 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,  -1 },      1.64413,   0.00916,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,   1 },      1.61873,   0.00765,     0.350 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,  -1 },    183.33568,   0.00005,     0.093 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,   1 },      1.74668,   0.00068,     0.062 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   3,   1 },      1.84666,   0.00007,     0.034 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,  -1 },      0.70571,   0.00009,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0 },     96.04769,   0.00003,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   1 },      0.57956,   0.00007,     0.089 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,  -1 },      0.62586,   0.00043,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0 },     95.30881,   0.00026,     0.087 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   1 },      0.62970,   0.00033,     0.508 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   0 },    276.60259,   0.00010,     0.555 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   1 },      0.79443,   0.00009,     0.066 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -3,  -1 },      1.24128,   0.00008,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -3,   1 },      0.79997,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,  -1 },      1.31137,   0.00077,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,   1 },      1.08966,   0.00033,     0.082 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,  -3 },      1.10092,   0.00001,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,  -1 },      1.26914,   0.00657,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   1 },      1.39630,   0.00554,     0.924 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   3 },      0.98504,   0.00003,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -1 },      1.42005,   0.00450,     0.080 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1 },      1.30219,   0.00591,     0.070 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   3 },    181.40702,   0.00002,     0.024 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,  -3 },      2.03388,   0.00001,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,  -1 },      1.12110,   0.00511,     1.335 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   1 },      1.09775,   0.00530,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,  -1 },      0.73030,   0.00029,     0.071 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   1 },      1.13481,   0.00064,     0.024 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   3,  -1 },      1.03968,   0.00002,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   3,   1 },      1.08422,   0.00006,     0.018 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,  -1 },      1.53722,   0.00001,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,   0 },     97.38832,   0.00003,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,   1 },    181.02795,   0.00005,     5.117 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,  -1 },      1.08244,   0.00051,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   0 },    276.81058,   0.00043,    34.860 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   1 },      1.05607,   0.00076,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -2 },     83.85380,   0.00004,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -1 },    180.71271,   0.00301,     7.243 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   0 },     96.81286,   0.00802,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   1 },    180.70716,   0.00297,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,  -1 },    180.71019,   0.00019,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   0 },     96.81034,   0.00043,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   1 },    180.71069,   0.00037,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,   0 },     96.81031,   0.00003,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,   1 },    180.71019,   0.00004,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -3,  -1 },    181.83972,   0.00002,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -2,  -1 },      1.39315,   0.00003,     0.070 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -2,   1 },    170.52361,   0.00003,     0.080 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,  -1 },      1.06533,   0.00170,     0.976 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,   1 },      0.84663,   0.00234,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,  -3 },      0.78448,   0.00006,     0.068 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,  -1 },      0.45426,   0.01409,     0.082 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,   1 },      0.59378,   0.00163,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,  -1 },      0.69308,   0.00090,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,   1 },      1.45479,   0.00028,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   2,  -1 },      0.75788,   0.00007,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   2,   1 },      1.20167,   0.00004,     0.015 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -2,  -1 },      0.74583,   0.00001,     0.523 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -2,   1 },      0.71384,   0.00004,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,  -1 },      0.39757,   0.00003,     0.088 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,   0 },     96.81002,   0.00009,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,   1 },    180.71563,   0.00005,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,  -1 },    180.71018,   0.00010,     0.041 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,   0 },     96.80979,   0.00005,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,   1 },    180.71008,   0.00005,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -2,  -1 },      0.97576,   0.00002,     0.096 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -2,   1 },      0.85242,   0.00006,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,  -1 },      0.51788,   0.00021,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,   1 },      0.93003,   0.00006,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,  -1 },      0.73857,   0.00014,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,   1 },      1.34395,   0.00003,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   5,  -2,  -1 },      0.71104,   0.00001,     0.044 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,   1,   0,   0 },    269.90517,   0.00005,     0.075 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,   1,   0,   0 },     68.96682,   0.00003,     0.075 },
 { {   0,   0,   1,   0,  -1,   3,   0,   0,   1,   0,   0 },     87.55473,   0.00010,     0.075 },
 { {   0,   0,   1,   0,  -1,   5,   0,   0,   1,   0,   0 },    103.14401,   0.00006,     0.074 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   0,  -1 },    314.81920,   0.00001,     0.027 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   0,   1 },    314.86471,   0.00003,     0.095 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -1,   0,   0 },    109.26883,   0.00003,     0.087 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   1,   0,   0 },    108.15544,   0.00027,     0.076 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   1,   0,   0 },    299.23514,   0.00015,     0.076 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   0,  -1 },     16.61592,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   0,   1 },     16.90939,   0.00006,     0.096 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -1,   0,   0 },    206.56574,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,  -1,  -1 },     16.76453,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,  -1,   1 },     16.93339,   0.00001,     0.939 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,  -1 },     15.18055,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   1 },     15.18055,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   1,  -1 },     16.93339,   0.00001,     1.367 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   1,   1 },     16.76453,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,   0 },     26.52278,   0.00009,    89.713 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,  -2 },    206.52081,   0.00001,     0.074 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,  -1 },    197.01456,   0.00002,     6.426 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,   0 },    206.52277,   0.00172,     0.075 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,   1 },    197.01469,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   1,   0 },    206.52278,   0.00009,     0.038 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   2,   0,  -1 },     17.08334,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   3,  -1,   0 },    206.52278,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,  -1,  -1 },    285.81874,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,  -1,   1 },    285.83145,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,  -1 },    285.69547,   0.00012,     0.027 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,   1 },    286.10670,   0.00046,     0.096 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   1,  -1 },    285.95522,   0.00015,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   1,   1 },    286.00324,   0.00013,     0.348 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,  -1,   0 },    113.92204,   0.00002,     0.040 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,   0 },    113.92205,   0.00044,     0.087 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   1,   0 },    293.92219,   0.00022,     0.550 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -1,  -1 },    285.66875,   0.00013,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -1,   1 },    285.85935,   0.00011,     0.910 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,  -1 },    284.29961,   0.00013,     0.080 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   1 },    284.19621,   0.00012,     0.070 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   1,  -1 },    285.80422,   0.00010,     1.307 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   1,   1 },    285.53821,   0.00009,     0.036 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -2,   0 },    113.92233,   0.00004,     0.076 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,  -2 },    113.92218,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   0 },    293.92221,   0.00087,    22.176 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   2 },    113.92220,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,  -2 },    113.92222,   0.00009,     0.074 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,  -1 },    106.27276,   0.00003,     8.220 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   0 },    113.92221,   0.01638,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   1 },    106.27295,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   2 },    293.92221,   0.00002,     0.025 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,  -2 },    113.92216,   0.00002,     3.467 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,   0 },    113.92221,   0.00091,     0.038 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   2,   0 },    113.92221,   0.00005,     0.025 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -1,  -1 },    288.49915,   0.00001,     0.991 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -1,   1 },    287.31749,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   0,  -1 },    286.32677,   0.00019,     0.082 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   0,   1 },    286.11560,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,  -1,   0 },    113.92221,   0.00017,     0.040 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,   0,   0 },    113.92221,   0.00010,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   0,  -1 },    180.83785,   0.00002,     0.016 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   0,   1 },      0.45093,   0.00001,     0.029 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   1,  -1 },      0.66625,   0.00001,     0.021 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -1,  -1 },      0.66832,   0.00007,     0.020 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -1,   1 },      0.70037,   0.00010,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,  -1 },      0.64052,   0.00048,     0.027 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,   1 },      0.67574,   0.00173,     0.097 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   1,  -1 },      0.67873,   0.00052,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   1,   1 },      0.67045,   0.00046,     0.344 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   2,   1 },      0.69569,   0.00006,     0.062 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,   0,   0 },    169.15261,   0.00004,     0.088 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,   1,   0 },    350.12312,   0.00004,     0.540 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -2,  -1 },      0.45335,   0.00005,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   1 },      0.44023,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,  -1 },      0.48807,   0.00040,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   1 },      0.46783,   0.00037,     0.883 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -1 },      0.71705,   0.00030,     0.080 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1 },      0.62054,   0.00037,     0.070 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,  -1 },      0.40573,   0.00028,     1.251 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,   1 },      0.41854,   0.00032,     0.036 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -1 },      0.44281,   0.00002,     0.071 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   2,   1 },      0.39503,   0.00004,     0.024 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,   0 },    350.12312,   0.00025,    12.652 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,  -2 },    170.12708,   0.00002,     0.074 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,  -1 },    180.31830,   0.00007,    11.401 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   0 },    170.12314,   0.00459,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   1 },    180.31836,   0.00007,     0.037 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,   0 },    170.12312,   0.00025,     0.038 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -1,  -1 },      0.32577,   0.00007,     1.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -1,   1 },      0.29459,   0.00011,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   0,  -1 },      0.23922,   0.00077,     0.081 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   0,   1 },      0.48207,   0.00008,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   1,  -1 },      0.18542,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   3,  -1,   0 },    170.12312,   0.00005,     0.040 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   3,   0,   0 },    170.12312,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0 },    247.93651,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   1,  -1,   0 },    254.07509,   0.00002,    11.213 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0 },     74.07509,   0.00038,     0.075 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   1,   0 },     74.07509,   0.00002,     0.038 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,  -1 },    353.68391,   0.00001,     0.027 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   1 },    353.62443,   0.00004,     0.097 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0 },    287.95045,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   1,  -1,   0 },    311.78513,   0.00003,     9.916 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0,   0 },    131.78513,   0.00061,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   1,   1,   0 },    131.78513,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0 },    272.78238,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   1,   0 },    266.25951,   0.00002,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     63.46186,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,   1 },     61.88541,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,  -1 },     67.37533,   0.00014,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1 },     64.11315,   0.00053,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,  -1 },     64.15389,   0.00018,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   1 },     61.69112,   0.00013,     0.340 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   0 },    273.70447,   0.00018,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -2 },    108.00491,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    269.83254,   0.00237,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0 },    109.42328,   0.00186,     0.530 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   0 },    109.15517,   0.00017,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1 },     81.97168,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1 },     90.02255,   0.00017,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    103.35314,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1 },     82.30097,   0.00015,     0.857 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },     94.96612,   0.00020,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    291.96119,   0.00004,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },     92.28543,   0.00021,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   3 },     80.81828,   0.00001,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1 },     87.97448,   0.00011,     1.200 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    282.35340,   0.00003,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   1 },     91.52450,   0.00012,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   1 },     81.31906,   0.00001,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,  -2 },    273.81692,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0 },    111.95783,   0.00777,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -2 },    305.93159,   0.00008,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    264.80325,   0.00005,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    291.98200,   0.14073,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    264.80303,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   2 },    112.55595,   0.00015,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -2 },    283.20849,   0.00006,     4.535 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0 },    291.85752,   0.00776,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   2 },    112.18566,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   2,   0 },    291.96124,   0.00047,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   3,   0 },    291.96124,   0.00003,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,  -1 },      2.68784,   0.00009,     1.063 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   1 },     16.45598,   0.00011,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -3 },    261.47872,   0.00001,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },     45.60552,   0.00040,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    111.96117,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   1 },     37.86858,   0.00004,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   1,  -1 },     42.61536,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,   0 },    291.96122,   0.00151,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,  -2 },    293.13069,   0.00001,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0 },    291.79012,   0.00090,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   1,   0 },    291.96121,   0.00011,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   5,  -2,   0 },    291.96121,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   5,  -1,   0 },    291.96120,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0 },    187.93334,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0 },    110.45407,   0.00010,     0.075 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   0,  -1 },    165.74501,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   0,   1 },    165.92703,   0.00007,     0.097 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,  -1,   0 },    334.18986,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   0,   0 },    334.31222,   0.00031,     0.088 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   1,   0 },    334.18975,   0.00012,     0.521 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -2,   0 },    334.18927,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,  -2 },    333.71621,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,   0 },    334.18759,   0.00074,     6.806 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,  -2 },    331.28563,   0.00015,     0.074 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,   0 },    154.18809,   0.01283,     0.075 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   1,  -2 },    334.92554,   0.00003,     5.360 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   1,   0 },    154.18852,   0.00066,     0.038 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   2,   0 },    154.18964,   0.00004,     0.025 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,  -1,   0 },    154.18980,   0.00014,     0.040 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,   0,   0 },    154.18993,   0.00008,     0.026 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -1,   0,   0 },     55.59056,   0.00002,     0.088 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,  -1,   0 },     55.59221,   0.00004,     5.528 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,  -2 },    219.04612,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,   0 },    235.62801,   0.00077,     0.074 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   1,   0 },    235.58945,   0.00004,     0.037 },
 { {   0,   0,   1,   0,   0,   3,   0,   0,   1,   0,   0 },    139.64195,   0.00002,     0.074 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,  -1 },     11.69898,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,   1 },     11.68445,   0.00008,     0.096 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   1,  -1 },     11.66102,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   1,   1 },     11.66103,   0.00003,     0.350 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,  -1,  -1 },     11.61195,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,  -1,   1 },     11.91431,   0.00001,     0.925 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   0,  -1 },     12.61265,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   0,   1 },     12.61265,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   1,  -1 },     11.91431,   0.00001,     1.337 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   1,   1 },     11.61195,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   1,   0,  -2 },     87.49696,   0.00001,     0.074 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   1,   0,   0 },     87.49696,   0.00001,     0.075 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,   0,  -1 },     10.26566,   0.00003,     0.082 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   1,   0,   0 },    136.66112,   0.00011,     0.075 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   0,  -1 },    183.19586,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   0,   1 },    183.24395,   0.00008,     0.097 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,  -1,  -1 },    182.33777,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,  -1,   1 },    182.33777,   0.00001,     0.845 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   0,  -1 },    181.35578,   0.00002,     0.081 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   0,   1 },    181.35578,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   1,  -1 },    182.33777,   0.00001,     1.177 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   1,   1 },    182.33777,   0.00001,     0.036 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   1,   0,   0 },    258.99694,   0.00027,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   0,   0 },    274.30804,   0.00005,     0.088 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -1,  -2 },    273.97952,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -1,   0 },    274.22479,   0.00010,     6.122 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,  -2 },    274.31746,   0.00014,     0.074 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,   0 },     94.41769,   0.00115,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   1,  -2 },    273.80308,   0.00003,     5.877 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   1,   0 },     95.23592,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   1,  -1 },    245.09095,   0.00001,     0.021 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -1,  -1 },    244.13139,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -1,   1 },    245.16798,   0.00005,     0.043 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,  -1 },    245.27184,   0.00019,     0.027 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   1 },    245.20223,   0.00071,     0.098 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   1,  -1 },    248.17765,   0.00018,     0.042 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   1,   1 },    247.69530,   0.00016,     0.331 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,  -1,   0 },    255.77508,   0.00009,     0.041 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,  -2 },     75.92819,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,   0 },    255.78993,   0.00171,     0.089 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   1,   0 },    255.38210,   0.00028,     0.507 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   2,   0 },    255.50875,   0.00003,     0.066 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -1,  -1 },    232.63954,   0.00010,     0.039 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -1,   1 },    234.85059,   0.00011,     0.799 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,  -1 },    239.22384,   0.00006,     0.081 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   1 },    239.35793,   0.00006,     0.069 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   1,  -1 },    237.97375,   0.00004,     1.090 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   1,   1 },    236.93658,   0.00004,     0.036 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -2,  -2 },     75.76353,   0.00003,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -2,   0 },    255.37651,   0.00008,     0.077 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,  -2 },     75.91670,   0.00023,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,   0 },    255.49958,   0.00171,     5.069 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,  -2 },     77.44064,   0.00077,     0.075 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   0 },     75.50853,   0.03480,     0.074 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   2 },    255.50786,   0.00003,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   1,  -2 },     75.06245,   0.00015,     7.341 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   1,   0 },     75.51377,   0.00210,     0.037 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   2,   0 },     75.51906,   0.00014,     0.025 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,  -1,   1 },     55.90916,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,   0,  -1 },    109.92913,   0.00003,     0.081 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -1,  -2 },    256.37210,   0.00002,     0.524 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -1,   0 },     75.49819,   0.00036,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   0,  -2 },     74.49515,   0.00005,     0.088 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   0,   0 },     75.50575,   0.00023,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   1,   0 },     75.50902,   0.00003,     0.019 },
 { {   0,   0,   1,   0,   2,  -8,   0,   0,   1,   0,   0 },     35.54263,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   1,   0,   0 },    302.60111,   0.00022,     0.075 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   0,  -1 },    190.86072,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   0,   1 },    190.86072,   0.00001,     0.070 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   1,   0,   0 },    209.74791,   0.00003,     0.075 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,   1,   0,   0 },     67.39747,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   0,   1 },     66.91787,   0.00002,     0.029 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   1,  -1 },     66.96098,   0.00001,     0.021 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -1,  -1 },     66.85456,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -1,   1 },     67.00078,   0.00006,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,  -1 },     66.82961,   0.00020,     0.027 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   1 },     66.98236,   0.00074,     0.097 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,  -1 },     66.52674,   0.00026,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,   1 },     66.56632,   0.00022,     0.340 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   2,   1 },     66.96888,   0.00002,     0.062 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,   0,   0 },     59.08097,   0.00007,     0.088 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,   1,   0 },     59.05884,   0.00015,     0.530 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -2,  -1 },     65.75660,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,  -1 },     66.26359,   0.00018,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,   1 },     66.16799,   0.00016,     0.858 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   0,  -1 },     67.06263,   0.00014,     0.080 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   0,   1 },     66.67670,   0.00018,     0.069 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,  -1 },     65.25482,   0.00013,     1.202 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,   1 },     65.23517,   0.00014,     0.036 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   2,   1 },     65.76576,   0.00002,     0.024 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,  -1,   0 },     59.05936,   0.00076,     8.940 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,  -2 },    239.06208,   0.00002,     0.074 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,  -1 },    245.35187,   0.00002,    18.216 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   0 },    239.05935,   0.01398,     0.075 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   1 },    245.35195,   0.00002,     0.037 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   1,   0 },    239.05936,   0.00076,     0.038 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   2,   0 },    239.05936,   0.00005,     0.025 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,  -1,  -1 },     65.90981,   0.00003,     1.062 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,  -1,   1 },     65.83809,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,  -1 },     64.52337,   0.00029,     0.081 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,   1 },     64.30564,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   3,  -1,   0 },    239.05936,   0.00015,     0.040 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   3,   0,   0 },    239.05936,   0.00009,     0.026 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,  -1,   0 },    323.19213,   0.00002,     6.858 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   0,   0 },    143.19213,   0.00037,     0.075 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   1,   0 },    143.19213,   0.00002,     0.038 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,   1,   0,   0 },     27.60512,   0.00013,     0.074 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,   1,   0,   0 },     94.96374,   0.00028,     0.074 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -1,   0,   0 },    258.76429,   0.00002,     0.089 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,   0 },     81.74884,   0.00023,     0.074 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,   0,   0 },    173.32458,   0.00025,     0.074 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,   1,   0,   0 },     70.72105,   0.00013,     0.075 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -1,   0,   0 },    139.57414,   0.00002,     0.092 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   1,   0,   0 },    335.17269,   0.00016,     0.072 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   0,   0 },    237.93377,   0.00006,     0.097 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   1,   0 },    240.81488,   0.00001,     0.339 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   0,   0 },     14.88503,   0.00004,     0.069 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,  -1,  -1 },    132.69757,   0.00001,     0.254 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,  -1,   1 },    132.69757,   0.00001,     0.044 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   0,  -1 },    132.69434,   0.00006,     0.107 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   0,   1 },    132.69149,   0.00002,     0.028 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -1,   0,   0 },    309.17101,   0.00006,     0.074 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   0,   0 },    351.69757,   0.00026,     0.077 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   1,   0,   0 },    352.00390,   0.00009,     0.085 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,   0,  -1 },     15.65056,   0.00002,     0.094 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   0,  -1 },    121.18045,   0.00002,     0.016 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   2,  -1 },    301.90836,   0.00001,     0.027 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,  -1 },    301.09494,   0.00007,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,   1 },    300.61960,   0.00013,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,  -1 },    301.84976,   0.00035,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   1 },    301.09502,   0.00188,     0.087 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,  -1 },    302.88150,   0.00045,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,   1 },    303.00574,   0.00037,     0.551 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   2,   1 },    304.47177,   0.00003,     0.066 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   0 },     21.99559,   0.00020,     0.080 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -2,  -1 },    297.42254,   0.00008,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1 },    297.37241,   0.00004,     0.076 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1 },    298.29632,   0.00081,     0.037 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1 },    297.83506,   0.00073,    24.988 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1 },    303.60068,   0.00077,     0.074 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1 },    303.68918,   0.00077,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -1 },    297.84993,   0.00074,     3.407 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   1 },    298.33006,   0.00082,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1 },    297.21309,   0.00004,     0.077 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,   1 },    297.60821,   0.00009,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,   0 },     24.87299,   0.00024,     0.082 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -2,  -1 },    304.76004,   0.00005,     0.065 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,  -1 },    303.34698,   0.00048,     0.484 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   1 },    303.43099,   0.00057,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,  -3 },    302.88687,   0.00001,     0.064 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,  -1 },    302.46154,   0.00215,     0.089 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   1 },    302.35163,   0.00044,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,  -1 },    301.86388,   0.00015,     0.041 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,   1 },    301.63562,   0.00008,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -2,   1 },    302.43834,   0.00001,     0.028 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -1,  -1 },    302.66207,   0.00003,     0.044 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -1,   1 },    302.66716,   0.00001,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,   0,  -1 },    302.68887,   0.00002,     0.028 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   0,   1 },      0.61461,   0.00004,     0.089 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,  -1,  -1 },    359.05984,   0.00001,     0.038 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,  -1,   1 },    359.05984,   0.00001,     4.793 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0,  -1 },      0.29610,   0.00003,     0.075 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0,   1 },      0.29610,   0.00003,     0.074 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   1,  -1 },    359.05984,   0.00001,     8.008 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   1,   1 },    359.05984,   0.00001,     0.037 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,   0,  -1 },      0.50894,   0.00003,     0.088 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   1,  -1 },    330.78855,   0.00001,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -1,  -1 },    330.80350,   0.00006,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -1,   1 },    330.83187,   0.00010,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,  -1 },    330.74143,   0.00037,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   1 },    330.89367,   0.00163,     0.092 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,  -1 },    330.91605,   0.00047,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,   1 },    330.87904,   0.00041,     0.426 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   2,   1 },    331.09997,   0.00004,     0.064 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   0 },     52.83040,   0.00018,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1 },    329.95847,   0.00005,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1 },    329.73304,   0.00002,     0.079 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1 },    329.82166,   0.00039,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1 },    329.74902,   0.00036,     1.749 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1 },    330.49672,   0.00030,     0.077 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1 },    330.39522,   0.00035,     0.072 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1 },    329.74703,   0.00033,     4.198 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1 },    329.79322,   0.00037,     0.037 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1 },    329.73329,   0.00002,     0.074 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1 },    329.55563,   0.00004,     0.025 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,  -1,   0 },    233.39730,   0.00004,     2.074 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,   0 },     53.39754,   0.00076,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   1,   0 },     53.39730,   0.00004,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -2,  -1 },    330.90088,   0.00002,     0.068 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,  -1 },    330.90624,   0.00024,     0.651 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,   1 },    330.88099,   0.00029,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,  -1 },    330.53889,   0.00117,     0.085 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,   1 },    330.58653,   0.00020,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   1,  -1 },    330.48922,   0.00009,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   1,   1 },    330.59031,   0.00003,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,  -1,  -1 },    358.59973,   0.00001,     0.013 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,  -1,   1 },    358.81077,   0.00002,     0.021 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,  -1 },      0.76811,   0.00005,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,   1 },    359.64897,   0.00011,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   1,  -1 },    359.18856,   0.00010,     0.021 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   1,   1 },    359.01086,   0.00015,     0.046 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   2,  -1 },    358.89533,   0.00007,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   2,   1 },    358.79928,   0.00002,     0.116 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -2,  -1 },    358.78318,   0.00004,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -2,   1 },    358.78504,   0.00003,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,  -1 },    359.11154,   0.00034,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,   1 },    358.97787,   0.00053,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,  -1 },    358.49780,   0.00223,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   1 },    358.62815,   0.00850,     0.096 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   3 },    358.70995,   0.00002,     0.061 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,  -1 },    358.92468,   0.00257,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   1 },    359.13760,   0.00216,     0.348 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   2,  -1 },    177.04898,   0.00002,     0.093 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   2,   1 },    358.84490,   0.00021,     0.062 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   3,   1 },    358.94450,   0.00002,     0.034 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,  -1,  -1 },      0.23070,   0.00001,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,  -1 },      0.61603,   0.00005,     0.040 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,   0 },     98.28521,   0.00013,     0.087 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,   1 },      0.83547,   0.00003,     0.502 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   1,   0 },    279.40897,   0.00005,     0.548 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -3,  -1 },    359.72429,   0.00002,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1 },    359.65262,   0.00022,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1 },    359.77207,   0.00010,     0.082 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -1 },    359.36781,   0.00194,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1 },    359.40999,   0.00159,     0.906 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1 },    359.58777,   0.00141,     0.080 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1 },    359.72965,   0.00187,     0.070 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1 },    359.90978,   0.00154,     1.299 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1 },      0.12140,   0.00160,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1 },      0.62349,   0.00007,     0.071 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1 },    359.94439,   0.00017,     0.024 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   3,   1 },    359.97125,   0.00001,     0.018 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -2,  -1 },    180.43365,   0.00001,     0.038 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,  -1 },      0.40685,   0.00004,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   0 },    279.40935,   0.00033,    20.152 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   1 },      0.21643,   0.00008,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -2 },     89.01744,   0.00003,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -1 },    180.23488,   0.00054,     8.537 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   0 },     99.41307,   0.00615,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   1 },    180.23489,   0.00055,     0.037 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,  -1 },    180.23488,   0.00003,     0.076 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,   0 },     99.40915,   0.00033,     0.038 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,   1 },    180.23489,   0.00007,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   2,   0 },     99.40913,   0.00002,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -2,   1 },    181.56772,   0.00003,     0.079 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,  -1 },    359.51807,   0.00041,     0.996 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,   1 },    359.94524,   0.00059,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -3 },      1.21615,   0.00001,     0.069 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -1 },      0.81383,   0.00404,     0.082 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   1 },      0.40649,   0.00048,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   1,  -1 },      0.29878,   0.00028,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   1,   1 },    359.31592,   0.00009,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -1,   0 },     99.40889,   0.00007,     0.040 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -1,   1 },    180.23070,   0.00001,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,   0,   0 },     99.40870,   0.00004,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -2,   1 },    359.88642,   0.00002,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -1,  -1 },      0.48619,   0.00005,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -1,   1 },      0.39430,   0.00002,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,   0,  -1 },      0.26457,   0.00002,     0.027 },
 { {   0,   0,   2,  -2,   2,   0,   0,   0,   1,   0,   0 },    135.93258,   0.00003,     0.074 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,  -1 },    255.15260,   0.00002,     0.027 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,   1 },    256.82237,   0.00007,     0.102 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   1,  -1 },    257.70648,   0.00001,     0.043 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   1,   1 },    257.70648,   0.00001,     0.293 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -1,   0,   0 },    128.95929,   0.00002,     0.092 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1 },    227.74277,   0.00001,     0.084 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1 },    227.74277,   0.00001,     0.067 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,  -1,   0 },    128.95930,   0.00003,     1.720 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,   0 },    308.96080,   0.00049,     0.072 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   1,   0 },    308.95929,   0.00003,     0.037 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,  -1 },     37.07681,   0.00002,     2.117 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,   1 },     37.07493,   0.00002,     0.038 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,   0,  -1 },     37.60113,   0.00002,     0.078 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   0,  -1 },    209.19048,   0.00002,     0.027 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   0,   1 },    209.15663,   0.00004,     0.103 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   1,   0,   0 },    298.55959,   0.00004,     0.072 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -1,  -1 },     29.46509,   0.00003,     2.760 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -1,   1 },     29.46510,   0.00003,     0.038 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,   0,  -1 },     30.32527,   0.00003,     0.078 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,  -1,  -1 },     54.21729,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,  -1,   1 },     54.21696,   0.00003,     0.522 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -2,  -1 },    234.12563,   0.00006,     0.074 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -2,   1 },    234.12563,   0.00006,     0.075 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -1,  -1 },    234.21681,   0.00017,     5.193 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -1,   1 },    234.21817,   0.00019,     0.038 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,   0,  -1 },    238.50423,   0.00004,     0.077 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,   0,   1 },    234.21724,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,  -1,  -1 },    199.87941,   0.00002,     0.020 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,  -1,   1 },    199.78362,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,  -1 },    199.94224,   0.00011,     0.027 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   1 },    199.78834,   0.00036,     0.104 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   1,  -1 },    199.80173,   0.00012,     0.043 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   1,   1 },    199.83564,   0.00010,     0.278 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -1,   0,   0 },    295.24783,   0.00003,     0.093 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -2,  -1 },    199.90884,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -1,  -1 },    200.22373,   0.00020,     0.040 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -1,   1 },    200.27692,   0.00017,     0.547 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,  -1 },    200.72477,   0.00006,     0.085 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   1 },    200.64509,   0.00007,     0.066 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   1,  -1 },    200.77157,   0.00006,     0.668 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   1,   1 },    200.65510,   0.00007,     0.035 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   1,   0,   0 },    289.14573,   0.00023,     0.071 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -3,  -1 },     19.86091,   0.00001,     0.037 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -2,  -1 },     19.82386,   0.00009,     0.074 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -2,   1 },     19.44019,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,  -1 },     19.95187,   0.00057,     3.597 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   1 },     19.95034,   0.00059,     0.038 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   0,  -1 },     20.58158,   0.00032,     0.077 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   0,   1 },     20.03290,   0.00009,     0.025 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -2,   1 },     19.93807,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -1,  -1 },     19.93997,   0.00002,     0.041 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   0,  -1 },    182.70223,   0.00001,     0.027 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   0,   1 },    182.70214,   0.00001,     0.104 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -1,  -1 },    182.38421,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -1,   1 },    182.38386,   0.00002,     0.527 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   0,  -1 },    182.14314,   0.00001,     0.086 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   0,   1 },    182.14314,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -2,  -1 },      2.29132,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -2,   1 },      2.29132,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -1,  -1 },      2.38370,   0.00013,     4.760 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -1,   1 },      2.38414,   0.00013,     0.038 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,   0,  -1 },      1.22881,   0.00003,     0.077 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -2,   1 },     57.32802,   0.00001,     0.088 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -1,  -1 },    237.10139,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -1,   1 },    237.14735,   0.00003,     0.518 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -3,  -1 },     54.92168,   0.00002,     0.037 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -2,  -1 },     57.39756,   0.00017,     0.074 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -2,   1 },     57.39756,   0.00017,     0.075 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -1,  -1 },     57.15368,   0.00021,     5.677 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -1,   1 },     57.17290,   0.00023,     0.037 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,   0,   1 },     57.14688,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -2,  -1 },     57.32802,   0.00001,     0.088 },
 { {   0,   0,   2,   0,  -4,   4,  -2,   0,   2,  -2,  -1 },    174.85996,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -4,   4,  -2,   0,   2,  -2,   1 },    174.85996,   0.00002,     0.074 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   0,  -1 },    351.22659,   0.00003,     0.028 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   0,   1 },    351.22874,   0.00006,     0.105 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   1,  -1 },    351.22847,   0.00001,     0.043 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   1,   1 },    351.22847,   0.00001,     0.265 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -1,  -1 },    171.21484,   0.00006,     0.040 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -1,   1 },    171.21442,   0.00005,     0.500 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -2,  -1 },    170.22881,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -2,   1 },    170.22881,   0.00002,     0.074 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,  -1 },    351.21424,   0.00029,     9.238 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,   1 },    351.21475,   0.00029,     0.037 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   0,  -1 },    169.53680,   0.00012,     0.076 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   0,   1 },    351.21478,   0.00004,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   0,  -1 },    189.86753,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   0,   1 },    190.14971,   0.00005,     0.029 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   1,  -1 },    190.39016,   0.00002,     0.021 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   1,   1 },    190.33125,   0.00002,     0.048 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   2,  -1 },    190.40630,   0.00001,     0.029 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -2,  -1 },    190.24163,   0.00003,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -2,   1 },    190.10462,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,  -1 },    190.25215,   0.00018,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,   1 },    190.21534,   0.00028,     0.044 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,  -1 },    190.36429,   0.00091,     0.027 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   1 },    190.29094,   0.00253,     0.105 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,  -1 },    190.36957,   0.00077,     0.043 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   1 },    190.40903,   0.00067,     0.271 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   2,   1 },    190.35135,   0.00007,     0.059 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,   0,   0 },    296.39680,   0.00009,     0.094 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -3,   1 },     10.07890,   0.00004,     0.041 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,  -1 },    190.23904,   0.00033,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   1 },     10.01395,   0.00054,     0.088 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,  -1 },    190.17080,   0.00408,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   1 },    190.17837,   0.00355,     0.523 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,  -1 },    190.74070,   0.00035,     0.086 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   1 },    190.34901,   0.00103,     0.066 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,  -3 },     10.77838,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,  -1 },    190.72340,   0.00038,     0.633 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   1 },    190.59120,   0.00045,     0.035 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -1 },    190.72966,   0.00002,     0.067 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   2,   1 },    190.74366,   0.00005,     0.024 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -1,   0 },     97.94235,   0.00005,     1.163 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   0,   0 },    277.90250,   0.00088,     0.071 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   1,   0 },    277.94234,   0.00005,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -4,  -1 },     10.11869,   0.00007,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -4,   1 },     10.24894,   0.00001,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,  -1 },     10.09413,   0.00080,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   1 },     10.54634,   0.00009,     7.152 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,  -3 },    190.04131,   0.00001,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,  -1 },     10.02715,   0.00681,     0.074 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   1 },     10.03367,   0.00510,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   3 },    190.04131,   0.00001,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,  -3 },     10.08414,   0.00005,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,  -1 },     10.08059,   0.01917,     5.163 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   1 },     10.08526,   0.02012,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   3 },    190.08576,   0.00002,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,  -1 },     10.51992,   0.00436,     0.077 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   1 },     10.02284,   0.00251,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,  -1 },     10.51349,   0.00027,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,   1 },     10.06616,   0.00026,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   2,   1 },     10.09682,   0.00002,     0.015 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,  -1 },      9.97761,   0.00011,     0.508 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,   1 },     10.00026,   0.00024,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,  -1 },     10.05280,   0.00075,     0.089 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   1 },     10.06911,   0.00066,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,  -1 },     10.11291,   0.00073,     0.041 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,   1 },     10.08053,   0.00037,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   0,  -1 },     10.21447,   0.00010,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   0,   1 },     10.08568,   0.00006,     0.015 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -3,  -1 },     10.02216,   0.00001,     0.044 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -2,  -1 },     10.08279,   0.00002,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -2,   1 },     10.08279,   0.00001,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   2,   0,   2,  -2,  -1 },    349.75714,   0.00001,     0.074 },
 { {   0,   0,   2,   0,  -3,   0,   2,   0,   2,  -2,   1 },    349.75714,   0.00001,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -2,   1 },    181.29767,   0.00002,     0.088 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -1,  -1 },    181.41215,   0.00006,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -1,   1 },    181.43813,   0.00004,     0.513 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -3,  -1 },    180.40512,   0.00003,     0.037 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -2,  -1 },    181.21817,   0.00022,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -2,   1 },    181.21817,   0.00022,     0.074 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -1,  -1 },      1.44361,   0.00022,     6.260 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -1,   1 },      1.47046,   0.00020,     0.037 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,   0,   1 },      1.43831,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -3,   1 },    181.65950,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -2,  -1 },    181.29767,   0.00002,     0.088 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -2,  -1 },     10.19156,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -2,   1 },     10.19156,   0.00002,     0.074 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -1,  -1 },    191.44254,   0.00005,     7.950 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -1,   1 },    191.44340,   0.00005,     0.037 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -2,  -1 },    165.07684,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -2,   1 },    165.07684,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -1,  -1 },    165.66141,   0.00004,     5.641 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -1,   1 },    165.66223,   0.00004,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   0,  -1 },    179.72118,   0.00008,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   0,   1 },    180.34726,   0.00018,     0.029 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,  -1 },    180.71399,   0.00019,     0.021 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,   1 },    180.79518,   0.00026,     0.048 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   2,  -1 },    180.82698,   0.00009,     0.029 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   2,   1 },    180.90255,   0.00002,     0.130 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -2,  -1 },    181.39000,   0.00004,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -2,   1 },    181.62774,   0.00004,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,  -1 },    180.98650,   0.00040,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,   1 },    181.34434,   0.00040,     0.044 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,  -1 },    181.01420,   0.00391,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   1 },    180.91996,   0.01230,     0.105 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   3 },    182.05125,   0.00003,     0.058 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,  -1 },    180.80883,   0.00380,     0.043 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   1 },    180.76225,   0.00328,     0.265 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   2,   1 },    180.85914,   0.00033,     0.059 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   3,   1 },    180.80272,   0.00004,     0.033 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,  -1 },      0.17907,   0.00007,     0.042 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,   0 },    352.67498,   0.00014,     0.095 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,   1 },      0.17849,   0.00007,     0.347 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   1,   0 },    177.57025,   0.00002,     0.368 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -3,  -1 },      0.06182,   0.00013,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -3,   1 },      0.11588,   0.00007,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,  -1 },    359.99791,   0.00107,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   1 },      0.02069,   0.00074,     0.089 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,  -1 },    359.95929,   0.00778,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   1 },    359.96504,   0.00657,     0.500 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,  -1 },    180.27107,   0.00234,     0.086 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   1 },    181.14686,   0.00057,     0.066 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,  -1 },    180.44257,   0.00199,     0.601 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,   1 },    180.50393,   0.00148,     0.035 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   2,  -1 },    180.36802,   0.00013,     0.067 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   2,   1 },    180.50452,   0.00017,     0.024 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   3,   1 },    180.33302,   0.00001,     0.018 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -1,  -1 },      0.32090,   0.00002,     0.080 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,  -1 },      0.16253,   0.00008,     1.286 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,   0 },    195.94648,   0.00030,     0.070 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,   1 },      0.16253,   0.00008,     0.036 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -4,  -1 },      0.05548,   0.00004,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -4,   1 },      0.07938,   0.00001,     0.077 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,  -1 },      0.01715,   0.00031,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,   1 },    359.75487,   0.00002,     4.462 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,  -3 },    180.10803,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,  -1 },    359.95130,   0.00175,     0.075 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   1 },      0.09559,   0.00623,     0.074 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,  -3 },    180.13950,   0.00018,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,  -1 },    180.11961,   0.05223,     9.142 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   1 },    180.12138,   0.05069,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   3 },      0.12183,   0.00005,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,  -3 },      0.11606,   0.00006,     0.073 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,  -1 },      0.03123,   0.02194,     0.076 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,   1 },    180.14886,   0.00588,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,  -3 },    180.12998,   0.00002,     2.258 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,  -1 },    359.99819,   0.00108,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,   1 },    180.17747,   0.00057,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,  -1 },      0.09161,   0.00008,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,   1 },    180.07715,   0.00004,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -1,  -1 },      0.12183,   0.00003,     0.082 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -1,   1 },      0.12183,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -4,  -1 },      0.09959,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,  -1 },      0.10593,   0.00013,     0.531 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,   1 },      0.09339,   0.00027,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,  -1 },      0.07191,   0.00034,     0.088 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,   1 },    180.12783,   0.00105,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,  -1 },    180.13439,   0.00140,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,   1 },    180.12663,   0.00084,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,  -1 },    181.33580,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,   1 },    180.12364,   0.00015,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   1,  -1 },    180.12183,   0.00002,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   1,   1 },    180.12183,   0.00002,     0.013 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -3,   1 },    180.12302,   0.00002,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -2,  -1 },    180.12198,   0.00004,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -2,   1 },    180.12198,   0.00003,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -1,  -1 },    180.12185,   0.00003,     0.020 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,  -1,  -1 },    198.97353,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,  -1,   1 },    198.97300,   0.00003,     0.523 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -2,  -1 },     18.40895,   0.00005,     0.074 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -2,   1 },     18.40895,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -1,  -1 },     18.97276,   0.00015,     5.133 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -1,   1 },     18.95284,   0.00017,     0.038 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,   0,  -3 },    197.67342,   0.00001,     0.073 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,   0,  -1 },     17.67342,   0.00001,     0.077 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,   0,   1 },     18.97344,   0.00002,     0.025 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -1,  -1 },    349.90706,   0.00002,    10.756 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -1,   1 },    349.90864,   0.00002,     0.037 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   0,   1 },    263.00302,   0.00006,     0.029 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   1,  -1 },    263.57403,   0.00003,     0.021 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -2,  -1 },    264.86120,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -1,  -1 },    264.09885,   0.00012,     0.020 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -1,   1 },    264.38948,   0.00013,     0.044 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,  -1 },    264.57044,   0.00077,     0.028 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   1 },    264.50670,   0.00224,     0.106 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,  -1 },    264.97846,   0.00065,     0.044 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,   1 },    264.97673,   0.00056,     0.259 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   2,   1 },    264.96411,   0.00007,     0.058 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,  -1 },    257.77326,   0.00003,     0.042 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,   0 },    280.68048,   0.00006,     0.096 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,   1 },    257.77326,   0.00002,     0.337 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -2,  -1 },    242.09865,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -1,  -1 },    248.88012,   0.00024,     0.040 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -1,   1 },    248.79983,   0.00019,     0.480 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -1 },    300.08934,   0.00007,     0.087 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1 },    267.76784,   0.00011,     0.065 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   1,  -1 },    271.64276,   0.00014,     0.572 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   1,   1 },    275.81376,   0.00011,     0.035 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   2,   1 },    269.72103,   0.00002,     0.024 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -2,  -1 },     77.77411,   0.00001,     0.039 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,  -1,   0 },     77.65733,   0.00004,     0.972 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   0 },    257.62528,   0.00079,     0.070 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   1,   0 },    257.65732,   0.00004,     0.036 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -2,  -1 },    271.95619,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -2,   1 },    170.90742,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,  -1 },    309.84404,   0.00062,    39.875 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,   1 },    306.65023,   0.00060,     0.037 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,  -1 },    168.55932,   0.00080,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,   1 },    287.81397,   0.00008,     0.025 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   1,  -1 },    165.82536,   0.00004,     0.038 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -2,   1 },    305.56260,   0.00001,     0.026 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -1,  -1 },    305.66785,   0.00002,     0.040 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,  -1 },    186.59722,   0.00003,     0.028 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,   1 },    188.10038,   0.00007,     0.105 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   1,  -1 },    189.09821,   0.00001,     0.043 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   1,   1 },    189.09821,   0.00001,     0.265 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -1,  -1 },      7.66527,   0.00006,     0.040 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -1,   1 },      7.66538,   0.00005,     0.501 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -2,  -1 },      5.64015,   0.00002,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -2,   1 },      5.64015,   0.00002,     0.074 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,  -1 },    187.66543,   0.00030,     9.049 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,   1 },    187.66529,   0.00031,     0.037 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   0,  -1 },      8.53739,   0.00012,     0.076 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   0,   1 },    187.66529,   0.00004,     0.025 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,   0,  -1 },      7.72315,   0.00001,     0.087 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,   0,   1 },      7.72315,   0.00001,     0.065 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -1,  -1 },      6.18631,   0.00004,    13.061 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -1,   1 },      6.18423,   0.00004,     0.037 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,   0,  -1 },    185.48427,   0.00003,     0.076 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   0,  -1 },    273.09063,   0.00004,     0.028 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   0,   1 },    272.88474,   0.00009,     0.106 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   1,  -1 },    272.74270,   0.00002,     0.043 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   1,   1 },    272.74270,   0.00002,     0.260 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -1,  -1 },     92.92599,   0.00003,     0.040 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -1,   1 },     92.94230,   0.00002,     0.484 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   0,  -1 },    274.54809,   0.00001,     0.087 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   0,   1 },    274.54809,   0.00001,     0.065 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   1,  -1 },    272.74794,   0.00001,     0.578 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   1,   1 },    272.74794,   0.00001,     0.035 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   1,   0,   0 },    294.42021,   0.00005,     0.070 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -1,  -1 },    272.87154,   0.00021,    23.464 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -1,   1 },    272.87112,   0.00021,     0.037 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,   0,  -1 },     92.71831,   0.00030,     0.076 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,   0,   1 },    272.87110,   0.00003,     0.025 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,  -1,  -1 },    180.20469,   0.00003,     0.020 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,  -1,   1 },    180.21315,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   0,  -1 },    180.19986,   0.00016,     0.028 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   0,   1 },    180.21017,   0.00050,     0.107 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   1,  -1 },    180.21351,   0.00014,     0.044 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   1,   1 },    180.21144,   0.00014,     0.258 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -1,   0,   0 },      9.51960,   0.00001,     0.096 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -2,  -1 },      0.04701,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -1,  -1 },    359.81760,   0.00006,     0.040 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -1,   1 },    359.72729,   0.00004,     0.476 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,  -1 },    180.07016,   0.00013,     0.087 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   1 },    180.19524,   0.00003,     0.065 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   1,  -1 },    180.12921,   0.00008,     0.566 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   1,   1 },    180.12831,   0.00009,     0.035 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   1,   0,   0 },    189.94508,   0.00006,     0.070 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -2,  -1 },    180.04535,   0.00004,     0.075 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -2,   1 },    359.92220,   0.00002,     0.074 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,  -1 },    180.03032,   0.00093,   115.334 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,   1 },    180.03313,   0.00083,     0.037 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   0,  -1 },      0.00427,   0.00187,     0.075 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   0,   1 },    180.04824,   0.00006,     0.025 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   1,  -1 },      0.00554,   0.00010,     0.038 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -2,   1 },    180.04357,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -1,  -1 },    180.03443,   0.00003,     0.040 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   0,  -1 },    258.50586,   0.00005,     0.028 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   0,   1 },    258.47279,   0.00012,     0.107 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   1,  -1 },    258.49172,   0.00004,     0.044 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   1,   1 },    258.49170,   0.00004,     0.256 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,  -1,  -1 },    258.92679,   0.00001,     0.040 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,  -1,   1 },    258.92679,   0.00001,     0.469 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,  -1 },    258.64725,   0.00003,     0.087 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   1 },    258.92363,   0.00001,     0.065 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   1,  -1 },    258.92679,   0.00001,     0.556 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   1,   1 },    258.92679,   0.00001,     0.035 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   1,   0,   0 },    246.32232,   0.00005,     0.070 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -1,  -1 },    262.42510,   0.00004,    39.561 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -1,   1 },    262.42511,   0.00004,     0.037 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,   0,  -1 },     87.71206,   0.00006,     0.075 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   2,   0,  -1 },    359.98693,   0.00005,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,  -1 },      3.70377,   0.00003,     0.028 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   1 },      6.35626,   0.00007,     0.107 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   1,  -1 },      1.78493,   0.00001,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   1 },      1.78493,   0.00001,     0.254 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,  -1,   0 },    192.09808,   0.00002,     0.042 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0 },    192.30954,   0.00035,     0.096 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   1,   0 },      9.88923,   0.00011,     0.346 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1 },     20.81165,   0.00005,     0.088 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1 },     20.78322,   0.00005,     0.065 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,  -1 },     15.74585,   0.00001,     0.545 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   1 },     18.27385,   0.00003,     0.035 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0 },      7.65548,   0.00094,     0.070 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   1,   0 },      8.31041,   0.00009,     0.036 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,  -2 },    187.37738,   0.00001,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,  -1 },     22.61157,   0.00011,    16.885 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    187.37738,   0.00001,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   1 },     23.04152,   0.00009,     0.037 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -3 },    330.78696,   0.00001,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },    200.47085,   0.00034,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   1,  -1 },    200.47088,   0.00002,     0.038 },
 { {   0,   0,   2,   0,   0,   1,   0,   0,  -1,   0,   0 },    231.25268,   0.00002,     0.097 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   0,  -1 },    139.24498,   0.00001,     0.028 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   0,   1 },    139.58569,   0.00003,     0.108 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -1,   0,   0 },    152.77159,   0.00008,     0.097 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   1,   0,   0 },    333.46323,   0.00003,     0.069 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -3 },    164.91620,   0.00005,     0.074 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -1 },    304.76570,   0.00003,     0.075 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,   0,  -3 },    188.74878,   0.00002,     0.074 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,   0,  -1 },    188.74878,   0.00002,     0.075 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   0,  -1 },    323.29202,   0.00002,     0.028 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   0,   1 },    323.35645,   0.00005,     0.107 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   1,   0,   0 },    316.11910,   0.00002,     0.070 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,   0,  -3 },    127.91840,   0.00002,     0.075 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,   0,  -1 },    307.91840,   0.00002,     0.074 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   1,   0,   0 },      7.79026,   0.00002,     0.067 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   0,  -1 },    209.65421,   0.00002,     0.026 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   0,   1 },    209.51723,   0.00023,     0.081 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   1,   1 },    209.68540,   0.00002,     1.064 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -1,   0,   0 },    188.17382,   0.00012,     0.075 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -1,  -1 },     31.29605,   0.00005,     0.036 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -1,   1 },     31.33869,   0.00005,     1.199 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   0,  -1 },     29.82629,   0.00006,     0.069 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   0,   1 },     29.83622,   0.00008,     0.081 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,  -1 },     29.26126,   0.00015,     0.857 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,   1 },     29.23277,   0.00017,     0.039 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   2,   1 },     28.84658,   0.00002,     0.026 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -2,  -1 },     30.17827,   0.00004,     0.062 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,  -1 },     29.95855,   0.00035,     0.340 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,   1 },     29.96467,   0.00042,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,  -1 },     29.67723,   0.00161,     0.097 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   1 },     29.68462,   0.00043,     0.027 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   1,  -1 },     29.46773,   0.00010,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   1,   1 },     29.63502,   0.00007,     0.020 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,  -1,   1 },     29.73236,   0.00001,     0.021 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   0,  -1 },     29.74023,   0.00003,     0.029 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   0,   1 },    209.72619,   0.00001,     0.016 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   2,  -2 },    144.43917,   0.00002,     0.026 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   2,   0 },    143.71171,   0.00005,     0.088 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   3,  -2 },    144.15728,   0.00003,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,   0 },    322.47039,   0.00014,     0.076 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,  -2 },    317.35975,   0.00002,     0.037 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,   0 },    324.06097,   0.00004,     6.053 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,  -2 },    146.77026,   0.00036,     0.074 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,   0 },    152.93552,   0.00013,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   3,  -2 },    142.30776,   0.00002,     5.942 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   3,   0 },    145.51652,   0.00005,     0.037 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   0,   0 },    324.25535,   0.00003,     0.086 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   1,   0 },    144.15728,   0.00003,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   2,  -2 },    143.71171,   0.00005,     0.088 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   2,   0 },    144.43917,   0.00002,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,  -1 },    272.83512,   0.00006,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,   1 },    271.88097,   0.00031,     0.087 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   1,  -1 },    273.72230,   0.00008,     0.040 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   1,   1 },    274.01449,   0.00007,     0.571 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   0 },    352.48664,   0.00010,     0.080 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -2,  -1 },    268.75012,   0.00001,     0.025 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1 },    269.73874,   0.00016,     0.037 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1 },    269.17390,   0.00014,    42.830 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1 },    274.40327,   0.00016,     0.073 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1 },    274.75238,   0.00014,     0.076 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,  -1 },    269.17390,   0.00014,     2.802 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,   1 },    269.73874,   0.00016,     0.038 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   2,   1 },    268.66692,   0.00001,     0.025 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,   0,   0 },    354.66725,   0.00012,     0.082 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,  -1 },    275.61271,   0.00009,     0.469 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,   1 },    275.14810,   0.00012,     0.040 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,  -1 },    274.30238,   0.00040,     0.090 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,   1 },    274.15539,   0.00009,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   1,   1 },    274.66053,   0.00001,     0.020 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   0,  -1 },    331.15955,   0.00002,     0.026 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   0,   1 },    331.15956,   0.00010,     0.088 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   1,  -1 },    331.15715,   0.00004,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   1,   1 },    331.15715,   0.00004,     0.521 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -1,  -1 },    330.23807,   0.00005,     0.037 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -1,   1 },    329.93793,   0.00004,     6.884 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0,  -1 },    333.36976,   0.00004,     0.074 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0,   1 },    333.36976,   0.00004,     0.075 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   1,  -1 },    329.93793,   0.00004,     5.312 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   1,   1 },    330.23807,   0.00005,     0.038 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,  -1,  -1 },    331.22976,   0.00004,     0.510 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,  -1,   1 },    331.22975,   0.00004,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   0,  -1 },    331.22772,   0.00011,     0.089 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   0,   1 },    331.21428,   0.00002,     0.026 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,  -1,  -1 },    301.64550,   0.00003,     0.020 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,  -1,   1 },    301.80688,   0.00006,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,  -1 },    301.48921,   0.00017,     0.026 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   1 },    301.69199,   0.00078,     0.091 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   1,  -1 },    301.78678,   0.00022,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   1,   1 },    301.77494,   0.00019,     0.438 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   2,   1 },    301.70864,   0.00002,     0.064 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   0 },     22.86769,   0.00012,     0.083 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1 },    300.32048,   0.00003,     0.025 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,  -1 },    300.28186,   0.00020,     0.038 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   1 },    300.11440,   0.00019,     1.967 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1 },    301.71969,   0.00019,     0.076 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1 },    301.68749,   0.00017,     0.073 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1 },    299.98559,   0.00016,     5.720 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,   1 },    300.08935,   0.00017,     0.037 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1 },    120.88570,   0.00004,     0.074 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1 },    300.31343,   0.00003,     0.025 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,  -1,   0 },    203.02177,   0.00002,     1.833 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,   0 },     23.02177,   0.00044,     0.079 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   1,   0 },     23.02177,   0.00002,     0.039 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,  -1 },    301.73622,   0.00011,     0.625 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,   1 },    301.64234,   0.00016,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,  -1 },    301.28548,   0.00058,     0.086 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,   1 },    301.30464,   0.00010,     0.026 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   1,  -1 },    301.35085,   0.00002,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   1,   1 },    301.39044,   0.00001,     0.019 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   0,   1 },    329.51777,   0.00003,     0.028 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   1,  -1 },    328.97931,   0.00002,     0.021 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   1,   1 },    328.92353,   0.00002,     0.046 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   2,  -1 },    328.72849,   0.00001,     0.028 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -1,  -1 },    328.77764,   0.00008,     0.020 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -1,   1 },    328.73648,   0.00014,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,  -1 },    328.89435,   0.00054,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   1 },    328.51862,   0.00206,     0.096 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,  -1 },    328.68509,   0.00062,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,   1 },    328.72836,   0.00055,     0.355 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   2,   1 },    328.57735,   0.00006,     0.062 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,  -1,   0 },     54.63434,   0.00001,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,  -1,   1 },    331.15410,   0.00003,     0.088 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,  -1 },    151.23247,   0.00008,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   0 },     54.43136,   0.00015,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   1 },    151.23559,   0.00008,     0.519 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   1,   0 },    234.80404,   0.00004,     0.568 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1 },    329.85141,   0.00007,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1 },    330.13959,   0.00002,     0.082 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1 },    329.73420,   0.00048,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1 },    329.79356,   0.00042,     0.962 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1 },    328.45971,   0.00034,     0.080 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1 },    329.20361,   0.00045,     0.070 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1 },    330.18443,   0.00036,     1.415 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1 },    330.21171,   0.00038,     0.036 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1 },    330.11200,   0.00002,     0.072 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1 },    330.35206,   0.00005,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -2,  -1 },    331.58612,   0.00004,     0.037 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -2,   0 },     54.80055,   0.00002,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -2,   1 },    331.47826,   0.00001,     6.536 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,  -1 },    331.59085,   0.00027,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   0 },    234.83811,   0.00023,    72.752 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   1 },    331.71766,   0.00024,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -2 },     54.85016,   0.00003,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -1 },    331.22497,   0.00048,     5.540 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },     54.83801,   0.00421,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   1 },    331.24812,   0.00051,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,  -1 },    331.23560,   0.00003,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,   0 },     54.83811,   0.00023,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,   1 },    331.23563,   0.00006,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -2,   1 },    151.56130,   0.00001,     0.080 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -1,  -1 },    330.38621,   0.00015,     0.937 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -1,   1 },    330.78267,   0.00021,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,  -1 },    331.32392,   0.00109,     0.082 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,   1 },    329.42114,   0.00012,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   1,  -1 },    331.21430,   0.00007,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   1,   1 },    330.16624,   0.00001,     0.019 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -2,   1 },    331.14635,   0.00002,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -1,  -1 },    331.15410,   0.00003,     0.088 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -1,   0 },     54.83811,   0.00004,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -1,   1 },    331.23670,   0.00002,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,   0,   0 },     54.83811,   0.00003,     0.026 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   1,  -1 },    187.42289,   0.00001,     0.021 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,  -1,  -1 },    187.00935,   0.00003,     0.020 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,  -1,   1 },    187.57701,   0.00005,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,  -1 },    186.44860,   0.00018,     0.027 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   1 },    187.16795,   0.00062,     0.101 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,  -1 },    186.88310,   0.00019,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,   1 },    187.03959,   0.00016,     0.299 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1 },    183.53429,   0.00001,     0.026 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -1,  -1 },    183.34454,   0.00022,     0.040 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   1 },    183.41321,   0.00020,     0.636 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1 },    184.97872,   0.00012,     0.083 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1 },    184.49240,   0.00013,     0.067 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1 },    182.96538,   0.00010,     0.808 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1 },    182.52295,   0.00010,     0.036 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1 },    183.86174,   0.00001,     0.024 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,  -1,   0 },     97.40417,   0.00002,     1.931 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   0,   0 },    277.40416,   0.00039,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   1,   0 },    277.40416,   0.00002,     0.037 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -2,  -1 },      0.96096,   0.00003,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -2,   1 },      0.96389,   0.00003,     0.077 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,  -1 },      0.75021,   0.00029,     1.866 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,   1 },      0.67740,   0.00028,     0.038 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,  -1 },      0.25862,   0.00024,     0.079 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,   1 },    359.21692,   0.00005,     0.025 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   1,  -1 },    256.76666,   0.00001,     0.021 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,  -1,  -1 },    256.80307,   0.00003,     0.020 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,  -1,   1 },    256.82957,   0.00005,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,  -1 },    256.60098,   0.00021,     0.028 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   1 },    256.74510,   0.00062,     0.107 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   1,  -1 },    256.97523,   0.00016,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   1,   1 },    257.00265,   0.00015,     0.258 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   2,   1 },    256.86601,   0.00001,     0.058 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1 },    256.03349,   0.00008,     0.040 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1 },    256.78840,   0.00009,     0.475 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1 },    254.09050,   0.00007,     0.087 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1 },    253.43640,   0.00006,     0.065 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1 },    256.62108,   0.00006,     0.565 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1 },    255.73033,   0.00003,     0.035 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,  -1,   0 },    127.56923,   0.00002,     0.953 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,   0 },    307.56923,   0.00028,     0.070 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   1,   0 },    307.56923,   0.00002,     0.036 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,  -2 },    127.91441,   0.00001,     0.074 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,  -1 },    258.98099,   0.00010,   241.391 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    129.33141,   0.00007,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   1 },    259.19007,   0.00009,     0.037 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,  -1 },    225.08794,   0.00003,     0.075 },
 { {   0,   0,   3,  -2,   2,   0,   0,   0,   2,   0,  -1 },    264.07786,   0.00002,     0.075 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   1,   0,   0 },    339.04535,   0.00005,     0.067 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   2,   0,  -1 },    210.64188,   0.00004,     0.072 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   0,  -1 },    194.08218,   0.00002,     0.028 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   0,   1 },    193.99756,   0.00003,     0.114 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   1,   0,   0 },    290.76128,   0.00006,     0.067 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -1,  -1 },    199.58752,   0.00005,     1.568 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -1,   1 },    199.58753,   0.00005,     0.036 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   0,  -1 },    200.68587,   0.00026,     0.072 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   0,   1 },     20.63468,   0.00001,     0.025 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,   0,  -1 },    181.11119,   0.00002,     0.072 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,   0,  -1 },    348.35145,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,  -1,  -1 },    182.18358,   0.00002,     0.021 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,  -1 },    182.59909,   0.00009,     0.028 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,   1 },    182.27424,   0.00021,     0.116 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   1,  -1 },    182.49180,   0.00007,     0.045 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   1,   1 },    182.49082,   0.00007,     0.217 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,  -1 },     13.84071,   0.00003,     0.093 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   1 },    187.13242,   0.00001,     0.062 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   1,  -1 },    185.43333,   0.00002,     0.401 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   1,   1 },    185.76159,   0.00004,     0.034 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,   0,   0 },    280.40374,   0.00019,     0.067 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -2,  -1 },    189.35911,   0.00002,     0.080 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -1,  -1 },    189.31783,   0.00019,     1.385 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -1,   1 },    189.64182,   0.00027,     0.036 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   0,  -1 },    190.57604,   0.00146,     0.072 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   0,   1 },      9.97955,   0.00003,     0.024 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   1,  -1 },    190.55362,   0.00008,     0.037 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   1,   1 },     10.08736,   0.00001,     0.018 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -4,  -1 },     14.53105,   0.00002,     0.037 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -3,  -1 },     14.70252,   0.00009,     0.075 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -3,   1 },     14.70252,   0.00009,     0.074 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -2,  -1 },    194.00434,   0.00003,     6.061 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   0,   1 },    165.78724,   0.00002,     0.030 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   1,  -1 },    165.50646,   0.00002,     0.021 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,  -1,  -1 },    166.59168,   0.00005,     0.021 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,  -1,   1 },    165.89585,   0.00004,     0.046 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,  -1 },    166.71890,   0.00029,     0.028 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   1 },    165.91156,   0.00074,     0.117 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   1,  -1 },    165.72776,   0.00023,     0.045 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   1,   1 },    166.06682,   0.00022,     0.213 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   2,   1 },    165.30976,   0.00002,     0.056 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -3,  -1 },     13.60499,   0.00001,     0.020 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -2,  -1 },     38.85573,   0.00001,     0.027 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -2,   1 },    267.72259,   0.00003,     0.097 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -1,  -1 },     72.30010,   0.00018,     0.042 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -1,   1 },     89.17824,   0.00016,     0.343 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,  -1 },      0.74199,   0.00015,     0.094 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   1 },    153.67034,   0.00008,     0.062 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   1,  -1 },    166.42271,   0.00005,     0.388 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   1,   1 },    174.47419,   0.00013,     0.034 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   2,   1 },    174.64706,   0.00002,     0.023 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   1,   0,   0 },    261.76380,   0.00003,     0.066 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -2,  -1 },    178.50700,   0.00007,     0.080 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -2,   1 },     64.86901,   0.00001,     0.069 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,  -1 },    171.07197,   0.00068,     1.240 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,   1 },    173.51013,   0.00097,     0.036 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,  -3 },    180.56150,   0.00003,     0.078 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,  -1 },    180.13318,   0.00518,     0.071 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,   1 },     13.66109,   0.00010,     0.024 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   1,  -1 },    180.20576,   0.00027,     0.037 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   1,   1 },    359.68888,   0.00005,     0.018 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -2,  -1 },     36.53363,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -2,   1 },    178.99928,   0.00002,     0.026 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -1,  -1 },    170.42543,   0.00010,     0.039 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   0,  -1 },    180.56992,   0.00002,     0.026 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   0,   1 },    180.03515,   0.00001,     0.015 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   0,  -1 },    256.75236,   0.00003,     0.017 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   0,   1 },     74.82917,   0.00001,     0.030 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   1,  -1 },     76.71346,   0.00002,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -1,  -1 },     74.43009,   0.00004,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -1,   1 },     70.34492,   0.00002,     0.046 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,  -1 },     75.66236,   0.00039,     0.028 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,   1 },     75.75402,   0.00096,     0.118 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   1,  -1 },     75.56420,   0.00028,     0.045 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   1,   1 },     75.44521,   0.00025,     0.210 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   2,   1 },     76.36607,   0.00002,     0.055 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -2,  -1 },    257.72089,   0.00007,     0.027 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -2,   1 },    257.08282,   0.00005,     0.097 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -1,  -1 },    256.81932,   0.00035,     0.042 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -1,   1 },    258.68478,   0.00027,     0.334 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,  -1 },     87.63921,   0.00005,     0.094 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   1 },    240.47630,   0.00002,     0.062 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   1,  -1 },     71.43023,   0.00006,     0.375 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   1,   1 },     70.77431,   0.00005,     0.034 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,   0,   0 },    259.54684,   0.00017,     0.066 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -2,   1 },    262.13449,   0.00002,     0.069 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -1,  -1 },    269.64559,   0.00022,     1.123 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -1,   1 },    268.81664,   0.00023,     0.036 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   0,  -1 },    321.90583,   0.00023,     0.071 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   0,   1 },    245.91138,   0.00008,     0.024 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,  -2 },     75.57123,   0.00003,     0.075 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,   0 },    262.11741,   0.00004,     0.074 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,  -2 },    257.28555,   0.00002,     6.129 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,   0 },    257.28617,   0.00002,     0.037 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -2,   1 },    257.18505,   0.00001,     0.025 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -1,  -1 },    257.18490,   0.00004,     0.039 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,   0,  -1 },    188.72979,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   0,  -1 },    166.22546,   0.00005,     0.028 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   0,   1 },    165.77733,   0.00010,     0.119 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   1,  -1 },    165.85211,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   1,   1 },    165.85211,   0.00002,     0.206 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   2,   0,  -1 },    272.72768,   0.00003,     0.071 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   0,  -1 },    176.80699,   0.00001,     0.028 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   0,   1 },    176.80731,   0.00001,     0.119 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,  -1,  -1 },    179.90009,   0.00002,     1.045 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,  -1,   1 },    179.90009,   0.00002,     0.036 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   0,  -1 },    180.00734,   0.00012,     0.070 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   0,   1 },    179.99419,   0.00001,     0.024 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   0,  -1 },     76.91583,   0.00002,     0.028 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   0,   1 },     76.94466,   0.00003,     0.119 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   0 },    120.95054,   0.00008,     0.107 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0 },    306.37335,   0.00010,     0.065 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,  -3 },     49.93705,   0.00001,     0.080 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,  -1 },    229.93705,   0.00001,     0.070 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -1,   0,   0 },    293.27850,   0.00008,     0.076 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   0,   1 },    242.33988,   0.00005,     0.086 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -1,   0,   0 },    323.29392,   0.00004,     0.079 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,  -1 },    241.30370,   0.00004,     0.037 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   1 },    240.06656,   0.00003,    11.532 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0,  -1 },    250.15478,   0.00003,     0.073 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0,   1 },    250.15478,   0.00003,     0.076 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,  -1 },    240.06656,   0.00003,     2.380 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,   1 },    241.30370,   0.00004,     0.038 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   1,   0,   0 },    324.98685,   0.00005,     0.083 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -1,  -1 },    247.03934,   0.00002,     0.456 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -1,   1 },    247.03908,   0.00002,     0.041 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,  -1 },    247.06307,   0.00007,     0.090 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,   1 },    246.88748,   0.00002,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   0,  -1 },    279.17335,   0.00003,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   0,   1 },    279.10397,   0.00015,     0.088 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   1,  -1 },    279.17345,   0.00005,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   1,   1 },    279.17345,   0.00005,     0.515 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -1,  -1 },    277.20077,   0.00007,     0.037 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -1,   1 },    276.91325,   0.00006,     5.976 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   0,  -1 },    279.71315,   0.00006,     0.075 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   0,   1 },    279.71315,   0.00006,     0.074 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   1,  -1 },    276.91325,   0.00006,     6.017 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   1,   1 },    277.20077,   0.00007,     0.037 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -1,  -1 },    279.17011,   0.00005,     0.515 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -1,   1 },    279.17011,   0.00005,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   0,  -1 },    279.10099,   0.00015,     0.088 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   0,   1 },    279.17021,   0.00003,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   3,  -1 },    286.78200,   0.00001,     0.021 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,  -1,   1 },    286.78191,   0.00002,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,  -1 },    286.67506,   0.00010,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   1 },    286.71100,   0.00027,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,  -1 },    286.84523,   0.00023,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   1 },    286.87248,   0.00046,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   2,  -1 },    287.08754,   0.00015,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   2,   1 },    286.78265,   0.00004,     0.105 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   0,  -1 },    286.78061,   0.00001,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   0,   1 },    286.78061,   0.00001,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -2,  -1 },    286.31002,   0.00010,     0.015 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -2,   1 },    286.31640,   0.00013,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,  -1 },    286.41978,   0.00070,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   1 },    286.40026,   0.00150,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,  -1 },    286.69718,   0.00453,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   1 },    286.68417,   0.02346,     0.088 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   3 },    286.64097,   0.00009,     0.064 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,  -1 },    287.11261,   0.00618,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   1 },    287.10965,   0.00503,     0.515 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,  -1 },    110.00408,   0.00001,     0.086 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   1 },    287.77781,   0.00037,     0.066 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   3,   1 },    288.23989,   0.00004,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   0,  -1 },    106.70442,   0.00011,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   0,   1 },    106.70524,   0.00012,     0.949 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -3,  -1 },    284.96336,   0.00009,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -3,   1 },    284.43207,   0.00002,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,  -1 },    285.03235,   0.00095,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   1 },    284.90345,   0.00046,     0.076 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,  -1 },    285.10406,   0.00804,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   1 },    284.85089,   0.00710,     6.017 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   3 },    284.85172,   0.00002,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,  -1 },    287.22930,   0.00762,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   1 },    287.22951,   0.00762,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,  -3 },    284.85184,   0.00002,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,  -1 },    284.85097,   0.00710,     5.977 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   1 },    285.10433,   0.00804,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,  -1 },    284.90024,   0.00046,     0.076 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   1 },    285.03268,   0.00095,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   3,  -1 },    284.43244,   0.00002,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   3,   1 },    284.96379,   0.00009,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   0,  -1 },    106.70530,   0.00012,     0.948 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   0,   1 },    106.70450,   0.00011,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -3,  -1 },    288.24107,   0.00004,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,  -1 },    287.77888,   0.00037,     0.066 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   1 },    110.07367,   0.00001,     0.086 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,  -1 },    287.11109,   0.00503,     0.515 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   1 },    287.11427,   0.00619,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -3 },    286.64344,   0.00009,     0.064 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -1 },    286.68658,   0.02348,     0.088 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   1 },    286.69867,   0.00453,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,  -1 },    286.40242,   0.00151,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   1 },    286.42134,   0.00070,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   2,  -1 },    286.31844,   0.00013,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   2,   1 },    286.31169,   0.00010,     0.015 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   0,  -1 },    286.78436,   0.00001,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   0,   1 },    286.78436,   0.00001,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -2,  -1 },    286.78363,   0.00004,     0.105 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -2,   1 },    287.08910,   0.00015,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,  -1 },    286.87472,   0.00046,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,   1 },    286.84690,   0.00023,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,  -1 },    286.71335,   0.00027,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,   1 },    286.67805,   0.00010,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   1,  -1 },    286.78391,   0.00002,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -3,   1 },    286.78362,   0.00001,     0.021 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,  -1,  -1 },    272.68521,   0.00001,     0.020 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,  -1 },    272.66388,   0.00009,     0.026 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   1 },    272.67988,   0.00042,     0.091 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   1,  -1 },    272.74989,   0.00012,     0.041 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   1,   1 },    272.82171,   0.00010,     0.451 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   0 },    353.55877,   0.00006,     0.083 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,  -1 },    271.00538,   0.00012,     0.038 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   1 },    270.80391,   0.00011,     2.247 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1 },    272.59085,   0.00009,     0.076 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1 },    272.51590,   0.00011,     0.073 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,  -1 },    270.64173,   0.00010,     8.973 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,   1 },    270.88111,   0.00011,     0.037 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,   0,   0 },    353.41690,   0.00021,     0.079 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,  -1 },    272.61923,   0.00008,     0.601 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,   1 },    272.53495,   0.00009,     0.040 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,  -1 },    272.13301,   0.00030,     0.086 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,   1 },    272.29995,   0.00005,     0.026 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,  -1,  -1 },    300.52749,   0.00003,     0.020 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,  -1,   1 },    300.50527,   0.00005,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,  -1 },    300.57766,   0.00016,     0.027 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   1 },    300.55061,   0.00064,     0.095 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   1,  -1 },    300.51051,   0.00019,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   1,   1 },    300.46992,   0.00016,     0.364 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   0 },     23.77848,   0.00008,     0.087 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -1 },    301.03870,   0.00001,     0.026 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -1,  -1 },    300.89623,   0.00015,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   1 },    301.02336,   0.00015,     1.024 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1 },    300.21969,   0.00013,     0.079 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1 },    300.59418,   0.00016,     0.070 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1 },    301.23026,   0.00010,     1.555 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   1,   1 },    301.21400,   0.00010,     0.036 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   2,   1 },    301.03055,   0.00001,     0.025 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -2,   0 },     24.02444,   0.00002,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,  -1 },    301.28450,   0.00001,     0.074 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   0 },    204.01481,   0.00009,    12.968 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   1 },    301.28450,   0.00001,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,  -2 },     24.02228,   0.00001,     0.073 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,  -1 },    302.24117,   0.00003,     4.101 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   0 },     24.01488,   0.00171,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   1 },    302.24104,   0.00003,     0.038 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   1,   0 },     24.01480,   0.00009,     0.038 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -1,  -1 },    301.84689,   0.00006,     0.884 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -1,   1 },    298.94253,   0.00010,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,  -1 },    302.03997,   0.00038,     0.082 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,   1 },    301.74710,   0.00005,     0.026 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   1,  -1 },    302.15757,   0.00002,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   3,  -1,   0 },     24.01480,   0.00002,     0.041 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,   1,   0,   0 },    256.79940,   0.00002,     0.075 },
 { {   0,   0,   4,  -6,   3,   0,   0,   0,   1,   0,   0 },     56.80519,   0.00002,     0.074 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,  -1,  -1 },    155.91215,   0.00002,     0.020 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,  -1,   1 },    156.24453,   0.00002,     0.043 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,  -1 },    155.44885,   0.00012,     0.027 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,   1 },    156.03822,   0.00041,     0.100 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   1,  -1 },    155.95854,   0.00013,     0.042 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   1,   1 },    156.13025,   0.00012,     0.305 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -1,   0,   0 },     56.04727,   0.00002,     0.091 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -1,  -1 },    153.30249,   0.00014,     0.039 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   1 },    153.35986,   0.00013,     0.663 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1 },    154.25072,   0.00009,     0.083 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1 },    154.04056,   0.00008,     0.068 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1 },    153.48393,   0.00008,     0.851 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1 },    152.98784,   0.00008,     0.036 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,  -1,   0 },     56.04727,   0.00004,     2.200 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,   0 },    236.04793,   0.00071,     0.073 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   1,   0 },    236.04727,   0.00004,     0.037 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -2,  -1 },    331.69625,   0.00002,     0.072 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -2,   1 },    331.69466,   0.00002,     0.077 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -1,  -1 },    331.39422,   0.00017,     1.669 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -1,   1 },    331.33642,   0.00016,     0.038 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,  -1 },    331.59144,   0.00016,     0.079 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,   1 },    330.99838,   0.00003,     0.025 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,  -1,  -1 },    204.59430,   0.00001,     0.020 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,  -1 },    203.70981,   0.00007,     0.028 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   1 },    205.40255,   0.00023,     0.106 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   1,  -1 },    204.67792,   0.00008,     0.043 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   1,   1 },    204.44128,   0.00007,     0.262 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -1,   0,   0 },     96.84682,   0.00001,     0.095 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -1,   0,   2 },    276.84682,   0.00001,     0.061 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -1 },      2.17095,   0.00002,     0.026 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -1,  -1 },    354.33510,   0.00011,     0.040 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   1 },    355.64672,   0.00011,     0.490 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1 },    189.61665,   0.00007,     0.087 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1 },    194.13956,   0.00003,     0.586 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1 },    192.91271,   0.00006,     0.035 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,   0 },    278.78553,   0.00020,     0.070 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -2,  -1 },    191.36036,   0.00001,     0.075 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -2,   1 },    358.56478,   0.00006,     0.074 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,  -1 },    181.90665,   0.00096,    14.813 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   1 },    182.00791,   0.00089,     0.037 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,  -1 },      0.57928,   0.00105,     0.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,   1 },    182.62004,   0.00008,     0.025 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   1,  -1 },      0.66335,   0.00006,     0.038 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -2,   1 },    181.96063,   0.00002,     0.026 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -1,  -1 },    181.97431,   0.00003,     0.041 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   0,  -1 },     76.19426,   0.00002,     0.028 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   0,   1 },     76.69060,   0.00004,     0.112 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -1,   0,   0 },    128.56316,   0.00002,     0.100 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   1,   0,   0 },    310.14809,   0.00018,     0.068 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,  -1 },     20.93296,   0.00006,     0.073 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   0,  -1 },    154.68866,   0.00001,     0.028 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   0,   1 },    154.68917,   0.00001,     0.119 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   1,   0,   0 },    338.13239,   0.00002,     0.065 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,   0,  -1 },    200.62885,   0.00006,     0.068 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   0,  -1 },    183.49645,   0.00001,     0.029 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   0,   1 },    183.49734,   0.00001,     0.129 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,  -1 },      6.74177,   0.00001,     0.102 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   1 },      6.74177,   0.00001,     0.059 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   1,   0,   0 },    279.54852,   0.00004,     0.063 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -1,  -1 },    190.26607,   0.00005,     0.611 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -1,   1 },    190.26648,   0.00005,     0.035 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   0,  -1 },    190.34114,   0.00021,     0.067 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   0,   1 },     10.28636,   0.00003,     0.024 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   0,  -1 },    166.88722,   0.00004,     0.029 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   0,   1 },    166.75443,   0.00007,     0.131 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   1,  -1 },    166.74151,   0.00002,     0.047 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   1,   1 },    166.74151,   0.00002,     0.178 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,  -1,  -1 },      5.67760,   0.00001,     0.043 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,  -1 },      5.44148,   0.00001,     0.102 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   1,  -1 },    172.79684,   0.00001,     0.286 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   1,   1 },    172.86597,   0.00001,     0.033 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -1,  -1 },    179.01552,   0.00011,     0.581 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -1,   1 },    179.14628,   0.00014,     0.035 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   0,  -1 },    179.83524,   0.00061,     0.067 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   0,   1 },    359.34027,   0.00005,     0.024 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   1,  -1 },    179.78045,   0.00004,     0.035 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   1,   1 },    359.48728,   0.00001,     0.018 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   0,  -1 },     64.13628,   0.00004,     0.029 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   0,   1 },     64.03933,   0.00006,     0.132 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   1,  -1 },     64.05878,   0.00002,     0.047 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   1,   1 },     64.05878,   0.00002,     0.176 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   1,   0,   0 },    259.10915,   0.00003,     0.062 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,  -1,  -1 },     27.72385,   0.00002,     0.554 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,  -1,   1 },     27.72374,   0.00002,     0.035 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,   0,  -1 },    357.85323,   0.00004,     0.066 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,   0,   1 },    209.84084,   0.00001,     0.024 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   0,  -1 },    331.11556,   0.00003,     0.029 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   0,   1 },    331.20380,   0.00005,     0.134 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   1,  -1 },    331.22581,   0.00001,     0.047 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   1,   1 },    331.22581,   0.00001,     0.173 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   0,   0 },    263.61835,   0.00004,     0.076 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   1,   0,   0 },    295.52657,   0.00002,     0.083 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,  -1 },    243.82160,   0.00006,     0.026 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,   1 },    243.68108,   0.00023,     0.090 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,  -1 },    244.12750,   0.00006,     0.041 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,   1 },    244.05580,   0.00006,     0.464 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -1,   0,   0 },    324.40920,   0.00002,     0.082 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,  -1 },    241.55902,   0.00007,     0.038 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   1 },    241.14494,   0.00006,     2.621 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -1 },    244.77103,   0.00007,     0.076 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1 },    244.81396,   0.00007,     0.073 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,  -1 },    241.14494,   0.00006,    20.797 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,   1 },    241.55902,   0.00007,     0.037 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   1,   0,   0 },    324.00025,   0.00009,     0.079 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,  -1 },    243.52480,   0.00005,     0.579 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,   1 },    243.60174,   0.00005,     0.040 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,  -1 },    243.04897,   0.00018,     0.087 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,   1 },    243.27034,   0.00003,     0.026 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,  -1 },    271.97892,   0.00006,     0.027 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   1 },    271.97608,   0.00020,     0.095 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   1,  -1 },    271.98089,   0.00007,     0.041 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   1,   1 },    271.98091,   0.00007,     0.372 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,   0,   0 },    354.13481,   0.00004,     0.086 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -1,  -1 },    271.92907,   0.00006,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   1 },    271.81280,   0.00003,     1.095 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1 },    272.03777,   0.00006,     0.079 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1 },    272.04212,   0.00005,     0.071 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   1,  -1 },    271.81280,   0.00003,     1.725 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   1,   1 },    271.92907,   0.00006,     0.036 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -2,   0 },    354.27675,   0.00004,     0.075 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   0 },    174.30635,   0.00004,     7.118 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   0,   0 },    354.30635,   0.00066,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   1,   0 },    354.30635,   0.00004,     0.038 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -1,  -1 },    273.09771,   0.00003,     0.837 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -1,   1 },    273.09781,   0.00003,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,  -1 },    273.03444,   0.00013,     0.083 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,   1 },    273.14999,   0.00002,     0.026 },
 { {   0,   0,   5,  -8,   2,   0,   0,   0,   1,   0,  -2 },     60.91191,   0.00001,     0.074 },
 { {   0,   0,   5,  -8,   2,   0,   0,   0,   1,   0,   0 },    240.91191,   0.00001,     0.075 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   0,  -1 },      4.55970,   0.00003,     0.016 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   0,   1 },    184.56866,   0.00002,     0.029 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   1,  -1 },    184.44031,   0.00002,     0.021 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -1,  -1 },    184.72649,   0.00007,     0.020 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -1,   1 },    184.51167,   0.00010,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,  -1 },    184.84987,   0.00043,     0.027 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   1 },    184.82589,   0.00157,     0.097 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,  -1 },    185.29548,   0.00042,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   1 },    185.29529,   0.00035,     0.340 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   2,   1 },    185.80933,   0.00004,     0.062 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -2,  -1 },    184.40740,   0.00002,     0.026 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,  -1 },    185.78161,   0.00016,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,   1 },    185.68203,   0.00017,     0.858 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0,  -1 },    188.40541,   0.00008,     0.080 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0,   1 },    188.68026,   0.00007,     0.069 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,  -1 },    191.24297,   0.00004,     1.201 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,   1 },    191.69022,   0.00005,     0.036 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   1,   0,   0 },      8.17354,   0.00012,     0.075 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -1,  -1 },      4.82415,   0.00004,     1.062 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -1,   1 },      3.13758,   0.00002,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   0,  -1 },      1.24120,   0.00022,     0.081 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,  -1,  -1 },    126.45031,   0.00001,     0.020 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,  -1 },    126.41498,   0.00008,     0.027 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   1 },    126.53417,   0.00024,     0.100 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,  -1 },    126.46274,   0.00008,     0.042 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,   1 },    126.42503,   0.00007,     0.311 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -1,   0,   0 },     24.42966,   0.00002,     0.090 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1 },    124.36141,   0.00006,     0.039 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1 },    124.38606,   0.00006,     0.692 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1 },    124.00421,   0.00007,     0.082 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1 },    123.80253,   0.00006,     0.068 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1 },    124.05358,   0.00005,     0.900 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1 },    124.06183,   0.00005,     0.036 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,  -1,   0 },     24.42966,   0.00004,     2.556 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   0,   0 },    204.43017,   0.00068,     0.073 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   1,   0 },    204.42966,   0.00004,     0.037 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,  -1 },    302.07474,   0.00007,     1.509 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,   1 },    302.07463,   0.00007,     0.038 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,  -1 },    303.13187,   0.00009,     0.079 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,   1 },    303.15968,   0.00001,     0.025 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,  -1 },    165.46641,   0.00004,     0.028 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,   1 },    166.52559,   0.00011,     0.105 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   1,  -1 },    165.97331,   0.00004,     0.043 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   1,   1 },    165.97510,   0.00004,     0.267 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -1,   0,   0 },     57.81824,   0.00001,     0.095 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -1,   0,   2 },    237.81824,   0.00001,     0.061 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -2,  -1 },    331.75898,   0.00006,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1 },    331.66657,   0.00005,     0.089 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,  -1 },    331.21790,   0.00037,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1 },    331.31989,   0.00030,     0.506 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1 },    158.37537,   0.00005,     0.086 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1 },    329.44876,   0.00005,     0.066 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1 },    157.84408,   0.00002,     0.609 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1 },    157.26517,   0.00003,     0.035 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,  -1,   0 },     57.26799,   0.00002,     1.084 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   0,   0 },    237.26799,   0.00030,     0.071 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   1,   0 },    237.26799,   0.00002,     0.036 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -3,  -1 },    332.46199,   0.00004,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,  -1 },    332.23497,   0.00023,     0.075 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,   1 },    331.94551,   0.00038,     0.074 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,  -1 },    151.74303,   0.00195,     7.641 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   1 },    151.74010,   0.00190,     0.037 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,  -1 },    331.60307,   0.00058,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,   1 },    151.71444,   0.00022,     0.025 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   1,  -1 },    331.61469,   0.00004,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   1,   1 },    151.74173,   0.00002,     0.019 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -3,   1 },    331.47829,   0.00002,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -2,  -1 },    331.50746,   0.00003,     0.088 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -2,   1 },    151.78660,   0.00003,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -1,  -1 },    151.74195,   0.00007,     0.041 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -1,   1 },    151.74327,   0.00003,     0.019 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,  -1 },    141.05637,   0.00001,     0.028 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,   1 },    133.21127,   0.00003,     0.111 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -1,  -1 },      5.14327,   0.00002,     0.041 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   1 },      7.05245,   0.00001,     0.399 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1 },      1.93102,   0.00002,     0.090 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   1,   1 },    171.14068,   0.00001,     0.034 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   1,   0,   0 },    282.35454,   0.00013,     0.068 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,  -1 },    178.56771,   0.00007,     2.495 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,   1 },    179.55223,   0.00013,     0.037 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,  -1 },    180.91357,   0.00096,     0.073 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,   1 },      0.27708,   0.00002,     0.025 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   1,  -1 },    180.89782,   0.00005,     0.037 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   1,   0,   0 },    309.94998,   0.00011,     0.066 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,   0,  -1 },     26.45893,   0.00004,     0.070 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,   0,  -1 },    190.05169,   0.00004,     0.063 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,   0,   1 },     10.05168,   0.00001,     0.023 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -1,  -1 },    178.91020,   0.00002,     0.379 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -1,   1 },    178.91165,   0.00002,     0.034 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   0,  -1 },    179.15638,   0.00010,     0.063 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   0,   1 },    359.08914,   0.00003,     0.023 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   0,   0 },    234.19970,   0.00003,     0.075 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   0,  -1 },    186.11080,   0.00001,     0.072 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   0,   1 },    186.11080,   0.00001,     0.077 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,  -1 },    214.83277,   0.00003,     0.026 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   1 },    214.75907,   0.00015,     0.090 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   1,  -1 },    214.85234,   0.00005,     0.040 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   1,   1 },    214.85234,   0.00005,     0.478 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,  -1 },    212.31583,   0.00005,     0.038 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,   1 },    211.81821,   0.00004,     3.142 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   0,  -1 },    216.52098,   0.00005,     0.075 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   0,   1 },    216.52098,   0.00005,     0.074 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,  -1 },    211.81821,   0.00004,    65.431 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,   1 },    212.31583,   0.00005,     0.037 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   1,   0,   0 },    298.07763,   0.00003,     0.080 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -1,  -1 },    214.26167,   0.00004,     0.559 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -1,   1 },    214.26172,   0.00004,     0.040 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,  -1 },    214.17682,   0.00012,     0.087 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,   1 },    214.30318,   0.00002,     0.026 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,  -1 },    243.25334,   0.00002,     0.027 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,   1 },    243.25531,   0.00007,     0.094 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   1,  -1 },    243.25420,   0.00003,     0.041 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   1,   1 },    243.25421,   0.00003,     0.381 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -1,   0,   0 },    324.86200,   0.00001,     0.086 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -1,   0,   2 },    324.86200,   0.00001,     0.066 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,  -1 },    242.65243,   0.00002,     0.038 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,   1 },    242.37240,   0.00001,     1.177 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   0,  -1 },    243.67249,   0.00002,     0.078 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   0,   1 },    243.67249,   0.00002,     0.071 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,  -1 },    242.37240,   0.00001,     1.937 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,   1 },    242.65243,   0.00002,     0.037 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,   0,   0 },    324.93469,   0.00026,     0.077 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,   0,  -1 },    243.81577,   0.00003,     0.083 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,  -1 },     97.36561,   0.00004,     0.027 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,   1 },     97.49062,   0.00012,     0.099 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   1,  -1 },     97.48485,   0.00004,     0.042 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   1,   1 },     97.48494,   0.00004,     0.317 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -1,   0,   0 },    175.77572,   0.00001,     0.090 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -1,   0,   2 },    355.77572,   0.00001,     0.064 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -1,  -1 },     95.03541,   0.00004,     0.039 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   1 },     95.24687,   0.00002,     0.724 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1 },     96.11225,   0.00005,     0.082 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1 },     96.11360,   0.00005,     0.068 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1 },     95.24687,   0.00002,     0.954 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   1,   1 },     95.03541,   0.00004,     0.036 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,  -1,   0 },    354.33538,   0.00003,     3.050 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,   0 },    174.33592,   0.00053,     0.074 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   1,   0 },    174.33538,   0.00003,     0.037 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,  -1 },    273.03570,   0.00003,     1.378 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,   1 },    273.03532,   0.00003,     0.038 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,   0,  -1 },    274.88069,   0.00003,     0.080 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,  -1 },    132.10633,   0.00002,     0.027 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,   1 },    132.78090,   0.00005,     0.105 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -2,  -1 },    122.35240,   0.00002,     0.026 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   1 },    302.36430,   0.00004,     0.088 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,  -1 },    122.54500,   0.00027,     0.040 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   1 },    122.56654,   0.00025,     0.523 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1 },    125.89844,   0.00001,     0.086 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,   1 },    123.04855,   0.00005,     0.066 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1 },    127.23506,   0.00001,     0.633 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,   1 },    127.23506,   0.00001,     0.035 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,   0 },    205.88897,   0.00027,     0.071 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -3,  -1 },    301.87158,   0.00005,     0.037 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,  -1 },    302.35157,   0.00049,     0.074 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   1 },    302.31968,   0.00037,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,  -1 },    302.33599,   0.00142,     5.148 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   1 },    302.33692,   0.00150,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,  -1 },    302.40019,   0.00043,     0.077 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,   1 },    302.34852,   0.00020,     0.025 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   1,  -1 },    302.41394,   0.00002,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   1,   1 },    302.33622,   0.00002,     0.019 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -3,   1 },    302.57727,   0.00002,     0.040 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -2,  -1 },    302.36093,   0.00006,     0.089 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -2,   1 },    302.35512,   0.00005,     0.026 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -1,  -1 },    302.33657,   0.00005,     0.041 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -1,   1 },    302.33867,   0.00002,     0.019 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   0,  -1 },     82.72750,   0.00002,     0.028 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   0,   1 },     82.64325,   0.00006,     0.107 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   1,  -1 },     82.66059,   0.00001,     0.044 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   1,   1 },     82.66059,   0.00001,     0.254 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,   0,  -1 },     83.91621,   0.00001,     0.088 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,   0,   1 },     83.91621,   0.00001,     0.065 },
 { {   0,   0,   6,  -7,  -2,   0,   0,   0,   2,   0,  -1 },     97.56390,   0.00002,     0.074 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,  -1 },    123.01217,   0.00001,     0.028 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,   1 },    123.04011,   0.00001,     0.111 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -1,  -1 },    333.92053,   0.00002,     0.041 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1 },    334.91437,   0.00001,     0.409 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1 },    332.58745,   0.00003,     0.090 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   1,   0,   0 },    239.88461,   0.00018,     0.068 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,  -1 },    149.48878,   0.00005,     2.964 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,   1 },    150.68872,   0.00012,     0.037 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,  -1 },    151.69199,   0.00117,     0.074 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   1,  -1 },    151.68611,   0.00006,     0.037 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -1,  -1 },    228.80481,   0.00002,     0.042 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   1 },    228.80483,   0.00001,     0.336 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   1,   0,   0 },    283.25588,   0.00008,     0.066 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,  -1 },    179.90902,   0.00005,     1.151 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,   1 },    179.90903,   0.00005,     0.036 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,  -1 },    180.60983,   0.00024,     0.071 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,   1 },      0.55831,   0.00002,     0.024 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   1,   0,   0 },    309.73158,   0.00007,     0.064 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,   0,  -1 },     28.64065,   0.00003,     0.068 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,   0,  -1 },    178.11715,   0.00002,     0.059 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   0,   1 },    175.61736,   0.00001,     0.081 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   0,   3 },    355.61736,   0.00001,     0.069 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,  -1,  -1 },    355.61736,   0.00001,     0.036 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,  -1,   1 },    355.61736,   0.00001,     1.200 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   0,  -1 },    355.61736,   0.00002,     0.069 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   0,   1 },    355.61736,   0.00002,     0.081 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   1,  -1 },    355.61736,   0.00001,     0.857 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   1,   1 },    355.61736,   0.00001,     0.039 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -1,  -1 },    355.61736,   0.00005,     0.340 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -1,   1 },    355.61736,   0.00005,     0.042 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   0,  -1 },    355.61736,   0.00015,     0.097 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   0,   1 },    355.61736,   0.00004,     0.027 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -1,   0,   0 },    204.39274,   0.00007,     0.075 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   0,  -1 },    258.20514,   0.00004,     0.097 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   0,   1 },    257.64306,   0.00001,     0.027 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,  -1 },    185.84606,   0.00003,     0.026 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   1 },    185.84609,   0.00014,     0.089 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,  -1 },    185.85829,   0.00005,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,   1 },    185.85826,   0.00005,     0.493 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -1,  -1 },    183.22892,   0.00006,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -1,   1 },    182.79535,   0.00005,     3.924 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,  -1 },    186.98280,   0.00005,     0.075 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   1 },    186.98280,   0.00005,     0.074 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   1,  -1 },    182.79535,   0.00005,    12.714 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   1,   1 },    183.22892,   0.00006,     0.037 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -1,  -1 },    185.33581,   0.00004,     0.540 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -1,   1 },    185.33563,   0.00004,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,  -1 },    185.38990,   0.00013,     0.088 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,   1 },    185.40528,   0.00003,     0.026 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   0,   1 },    214.45579,   0.00003,     0.094 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   1,   0,   0 },    295.71317,   0.00010,     0.077 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   0,  -1 },     68.92568,   0.00003,     0.027 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   0,   1 },     68.99195,   0.00008,     0.098 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   1,  -1 },     68.94480,   0.00003,     0.042 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   1,   1 },     68.94489,   0.00003,     0.323 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,  -1 },     66.38955,   0.00002,     0.039 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,   1 },     66.63061,   0.00001,     0.759 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   0,  -1 },     67.69159,   0.00002,     0.081 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   0,   1 },     67.69159,   0.00002,     0.069 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,  -1 },     66.63061,   0.00001,     1.016 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,   1 },     66.38955,   0.00002,     0.036 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,  -1,   0 },    324.57725,   0.00002,     3.781 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,   0 },    144.57791,   0.00038,     0.074 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   1,   0 },    144.57725,   0.00002,     0.037 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -1,  -1 },     93.12670,   0.00004,     0.040 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -1,   1 },     93.12670,   0.00004,     0.541 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   1,   0,   0 },    175.08489,   0.00019,     0.071 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -2,  -1 },    273.62506,   0.00003,     0.074 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -2,   1 },    273.62506,   0.00003,     0.075 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,  -1 },    273.12670,   0.00022,     3.882 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,   1 },    273.12670,   0.00022,     0.038 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,  -1 },    273.47153,   0.00014,     0.077 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,   1 },    273.12670,   0.00003,     0.025 },
 { {   0,   0,   7, -10,   1,   0,   0,   0,   2,  -2,  -1 },    124.26595,   0.00001,     0.074 },
 { {   0,   0,   7, -10,   1,   0,   0,   0,   2,  -2,   1 },    124.26595,   0.00001,     0.075 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,  -1,  -1 },    301.73396,   0.00002,     0.041 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1 },    303.12384,   0.00004,     0.089 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   1,   0,   0 },    206.60643,   0.00016,     0.069 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,  -1 },    119.50567,   0.00003,     3.649 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,   1 },    121.73392,   0.00009,     0.037 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,  -1 },    122.63781,   0.00116,     0.074 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,   1 },    302.28313,   0.00001,     0.025 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   1,  -1 },    122.63544,   0.00006,     0.037 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   1,   0,   0 },    241.33575,   0.00011,     0.066 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -1,  -1 },    151.12432,   0.00004,     1.241 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -1,   1 },    151.22240,   0.00006,     0.036 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,  -1 },    151.49705,   0.00027,     0.071 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,   1 },    331.38645,   0.00002,     0.024 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   1,  -1 },    151.48993,   0.00002,     0.037 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -1 },    177.98594,   0.00001,     0.099 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0,   1 },    177.98594,   0.00001,     0.060 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   1,   0,   0 },    283.88430,   0.00006,     0.064 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,  -1 },    180.07257,   0.00003,     0.748 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,   1 },    180.07257,   0.00003,     0.035 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,  -1 },    180.41766,   0.00010,     0.069 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,   1 },      0.41627,   0.00001,     0.024 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   1,   0,   0 },    309.35199,   0.00004,     0.062 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,   0,  -1 },     29.84111,   0.00002,     0.066 },
 { {   0,   0,   8, -19,   0,   0,   0,   0,  -2,   0,   1 },    208.12327,   0.00001,     0.075 },
 { {   0,   0,   8, -19,   0,   0,   0,   0,  -2,   0,   3 },     28.12327,   0.00001,     0.074 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,  -1,   0,   0 },    354.86158,   0.00008,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   0,   1 },    252.68044,   0.00003,     0.028 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   1,  -1 },    252.68044,   0.00001,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   1,   1 },    252.68044,   0.00004,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   2,  -1 },    252.68044,   0.00001,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,  -1 },    252.68044,   0.00006,     0.019 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,   1 },    252.68044,   0.00014,     0.041 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,  -1 },    252.68044,   0.00043,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,   1 },    252.68044,   0.00228,     0.088 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,  -1 },    252.68044,   0.00074,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,   1 },    252.68044,   0.00064,     0.515 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   2,  -1 },     72.68044,   0.00001,     0.086 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   2,   1 },    252.68044,   0.00007,     0.066 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -2,  -1 },     72.68044,   0.00003,     0.025 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -1,  -1 },     72.68044,   0.00009,     0.037 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -1,   1 },     72.68044,   0.00007,     5.996 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   0,  -1 },     72.68044,   0.00013,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   0,   1 },     72.68044,   0.00013,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   1,  -1 },     72.68044,   0.00007,     5.997 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   1,   1 },     72.68044,   0.00009,     0.037 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   2,   1 },     72.68044,   0.00003,     0.025 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -2,  -1 },    252.68044,   0.00007,     0.066 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -2,   1 },     72.68044,   0.00001,     0.086 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,  -1 },    252.68044,   0.00064,     0.515 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,   1 },    252.68044,   0.00074,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,  -1 },    252.68044,   0.00228,     0.088 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,   1 },    252.68044,   0.00043,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,  -1 },    252.68044,   0.00014,     0.041 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,   1 },    252.68044,   0.00006,     0.019 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -2,   1 },    252.68044,   0.00001,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -1,  -1 },    252.68044,   0.00004,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -1,   1 },    252.68044,   0.00001,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,   0,  -1 },    252.68044,   0.00003,     0.028 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -1,  -1 },    156.85454,   0.00002,     0.019 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -1,   1 },    156.82674,   0.00004,     0.041 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,  -1 },    156.61961,   0.00014,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   1 },    156.64435,   0.00069,     0.089 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,  -1 },    157.35257,   0.00018,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,   1 },    157.33976,   0.00016,     0.509 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -2,  -1 },    154.16377,   0.00003,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -2,   1 },    153.44089,   0.00001,     0.077 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,  -1 },    154.02735,   0.00025,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   1 },    153.75828,   0.00022,     5.222 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,  -1 },    156.78302,   0.00022,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   1 },    156.76491,   0.00022,     0.074 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,  -1 },    153.71382,   0.00022,     7.041 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   1 },    153.98444,   0.00024,     0.037 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   2,  -1 },    153.44710,   0.00001,     0.076 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   2,   1 },    154.16453,   0.00003,     0.025 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,  -1 },    157.20732,   0.00015,     0.522 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,   1 },    157.21869,   0.00018,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,  -1 },    156.49498,   0.00067,     0.088 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,   1 },    156.47816,   0.00014,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   1,  -1 },    156.66446,   0.00004,     0.041 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   1,   1 },    156.69275,   0.00002,     0.019 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1 },    184.33493,   0.00001,     0.078 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   0,   1 },    184.33493,   0.00001,     0.072 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   1,   0,   0 },    266.73651,   0.00003,     0.077 },
 { {   0,   0,   8, -13,   0,  -1,   0,   0,   1,   0,  -2 },    191.19471,   0.00002,     0.075 },
 { {   0,   0,   8, -13,   0,  -1,   0,   0,   1,   0,   0 },    220.33319,   0.00003,     0.075 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,  -1 },     41.64984,   0.00002,     0.027 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,   1 },     41.73968,   0.00006,     0.098 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1 },     38.60894,   0.00001,     0.081 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   0,   1 },     38.60894,   0.00001,     0.069 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   0,  -2 },    111.85263,   0.00001,     0.075 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   0,   0 },    113.38173,   0.00024,     0.074 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   1,   0,   0 },    145.60383,   0.00012,     0.071 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -1,  -1 },    244.18160,   0.00006,     3.115 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -1,   1 },    244.18174,   0.00006,     0.038 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,   0,  -1 },    243.03567,   0.00006,     0.077 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   0,  -1 },    276.03927,   0.00004,     0.089 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   0,   1 },     86.94818,   0.00001,     0.064 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   1,   0,   0 },    175.95572,   0.00011,     0.069 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,  -1 },    275.37946,   0.00003,     4.746 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,   1 },     92.85099,   0.00005,     0.037 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,  -3 },     91.39986,   0.00003,     0.075 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,  -1 },     93.68983,   0.00152,     0.074 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   1,  -1 },     93.69417,   0.00008,     0.037 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   1,   0,   0 },    207.57260,   0.00010,     0.066 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -1,  -1 },    122.39871,   0.00004,     1.347 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -1,   1 },    122.39871,   0.00004,     0.036 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,  -1 },    122.44335,   0.00020,     0.071 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,   1 },    302.41014,   0.00001,     0.024 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   1,   0,   0 },    242.90213,   0.00007,     0.064 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,  -1 },    151.38163,   0.00003,     0.785 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,   1 },    151.38169,   0.00003,     0.036 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,  -1 },    151.41161,   0.00013,     0.069 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,   1 },    331.14074,   0.00001,     0.024 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   1,   0,   0 },    284.29359,   0.00004,     0.062 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,  -1 },    180.06390,   0.00007,     0.066 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,   1 },    180.06381,   0.00001,     0.024 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   1,   0,   0 },    308.78818,   0.00003,     0.060 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,   0,  -1 },     30.57460,   0.00002,     0.064 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   0,  -1 },    174.22418,   0.00001,     0.069 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   0,   1 },    174.22418,   0.00001,     0.081 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,   0,  -1 },    174.22422,   0.00003,     0.097 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,  -1 },    127.72002,   0.00002,     0.026 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,   1 },    127.72006,   0.00009,     0.088 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   1,  -1 },    127.70557,   0.00002,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   1,   1 },    127.70557,   0.00002,     0.526 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,  -1 },    125.74273,   0.00004,     0.037 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,   1 },    125.04059,   0.00004,     7.805 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   0,  -1 },    131.68133,   0.00004,     0.074 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   0,   1 },    131.68133,   0.00004,     0.075 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,  -1 },    125.04059,   0.00004,     4.869 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,   1 },    125.74273,   0.00004,     0.038 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,  -1 },    128.04398,   0.00003,     0.505 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,   1 },    128.04399,   0.00003,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,  -1 },    128.03477,   0.00009,     0.089 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,   1 },    128.03477,   0.00002,     0.026 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   0,  -1 },    149.74351,   0.00004,     0.027 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   0,   1 },    149.74351,   0.00015,     0.097 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   1,  -1 },    149.74351,   0.00005,     0.042 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   1,   1 },    149.74351,   0.00005,     0.340 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,  -1,  -1 },    149.74351,   0.00001,     0.039 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,  -1,   1 },    149.74351,   0.00001,     0.857 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   0,  -1 },    149.74351,   0.00002,     0.081 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   0,   1 },    149.74351,   0.00002,     0.069 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   1,  -1 },    149.74351,   0.00001,     1.200 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   1,   1 },    149.74351,   0.00001,     0.036 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,   0,  -1 },    329.74351,   0.00002,     0.081 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,  -1 },     21.12507,   0.00003,     0.027 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,   1 },     21.87200,   0.00011,     0.097 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   1,  -1 },     21.27029,   0.00004,     0.042 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   1,   1 },     21.27207,   0.00004,     0.337 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,  -1,  -1 },     13.84456,   0.00003,     0.039 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,  -1,   1 },     14.41568,   0.00001,     0.839 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1 },     18.39601,   0.00004,     0.081 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   0,   1 },     18.39911,   0.00004,     0.069 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   1,  -1 },     14.41568,   0.00001,     1.166 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   1,   1 },     13.84456,   0.00003,     0.036 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,  -1,   0 },    267.67903,   0.00005,     7.261 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,   0 },     87.67799,   0.00092,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   1,   0 },     87.67799,   0.00005,     0.038 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,   0,  -1 },    359.67774,   0.00002,     0.081 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   1,   0,   0 },    116.23587,   0.00007,     0.072 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -1,  -1 },    215.88709,   0.00002,     2.602 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -1,   1 },    215.88709,   0.00002,     0.038 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,   0,  -1 },    215.53660,   0.00002,     0.078 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   0,  -1 },     64.59188,   0.00006,     0.088 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   1,   0,   0 },    146.11802,   0.00007,     0.069 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,  -1 },     64.41947,   0.00012,     6.788 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,   1 },     63.92327,   0.00003,     0.037 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,  -3 },     50.69762,   0.00002,     0.074 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,  -1 },    244.58740,   0.00163,     0.075 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   1,  -1 },    244.59212,   0.00009,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   4,  -1,  -1 },    244.59291,   0.00002,     0.040 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   1,   0,   0 },    176.57249,   0.00007,     0.067 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,  -1 },     93.42688,   0.00002,     1.473 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,   1 },     93.42688,   0.00002,     0.036 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,  -1 },     93.43255,   0.00013,     0.072 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   1,   0,   0 },    208.56603,   0.00006,     0.064 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -1,  -1 },    122.36483,   0.00002,     0.826 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -1,   1 },    122.36483,   0.00002,     0.036 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,  -1 },    122.42025,   0.00010,     0.069 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,   1 },    301.99917,   0.00001,     0.024 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   1,   0,   0 },    244.37319,   0.00004,     0.062 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,  -1 },    151.20405,   0.00002,     0.574 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,   1 },    151.20405,   0.00002,     0.035 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,  -1 },    151.11034,   0.00007,     0.067 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,   1 },    331.95115,   0.00001,     0.024 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   1,   0,   0 },    284.51804,   0.00002,     0.060 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,   0,  -1 },    179.47104,   0.00003,     0.064 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,   0,  -5 },    190.06154,   0.00004,     0.075 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,   0,  -3 },    190.06154,   0.00004,     0.075 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   1,   0,   0 },    308.00219,   0.00002,     0.058 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   0,  -1 },     98.02006,   0.00001,     0.074 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   0,   1 },     98.02006,   0.00001,     0.075 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,  -1,  -1 },     71.28735,   0.00001,     0.019 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,  -1,   1 },     71.28735,   0.00002,     0.041 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,  -1 },     71.28735,   0.00007,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   1 },     71.28735,   0.00045,     0.088 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,  -1 },     71.28735,   0.00015,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,   1 },     71.28735,   0.00012,     0.515 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,  -1,  -1 },    251.28720,   0.00001,     0.037 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   1,   1 },    251.28742,   0.00001,     0.037 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,  -1 },     71.28727,   0.00012,     0.515 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,   1 },     71.28727,   0.00015,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,  -1 },     71.28727,   0.00045,     0.088 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,   1 },     71.28727,   0.00007,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   1,  -1 },     71.28727,   0.00002,     0.041 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   1,   1 },     71.28727,   0.00001,     0.019 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   1,   0,   0 },    238.12639,   0.00021,     0.075 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   1,   0,   0 },     87.10481,   0.00004,     0.072 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   1,   0,   0 },    116.63173,   0.00004,     0.069 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,  -1 },     32.38306,   0.00002,    11.910 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,   1 },     32.38303,   0.00002,     0.037 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,  -1 },    215.49135,   0.00018,     0.075 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   1,   0,   0 },    146.54602,   0.00004,     0.067 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,  -1 },     64.43923,   0.00008,     0.072 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   1,   0,   0 },    177.21590,   0.00005,     0.065 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   0,  -1 },     93.38354,   0.00007,     0.069 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   1,   0,   0 },    209.57106,   0.00004,     0.063 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,  -1 },    122.61940,   0.00002,     0.596 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,   1 },    122.61940,   0.00002,     0.035 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,   0,  -1 },    122.27951,   0.00006,     0.067 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   1,   0,   0 },    245.72576,   0.00003,     0.061 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,   0,  -1 },    150.99677,   0.00004,     0.065 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   0,   1 },    128.40061,   0.00002,     0.088 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,   0,  -1 },    128.36940,   0.00002,     0.088 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,   1,   0,   0 },    208.84621,   0.00003,     0.075 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,   0,   1 },    328.35041,   0.00003,     0.097 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   0,  -1 },    328.35044,   0.00001,     0.080 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   0,   1 },    328.35044,   0.00001,     0.069 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   1,   0,   0 },     57.95810,   0.00002,     0.072 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,   0,  -1 },    186.53468,   0.00005,     0.075 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   1,   0,   0 },    116.94588,   0.00002,     0.067 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,  -1 },     35.45803,   0.00004,     0.072 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   1,   0,   0 },    146.99391,   0.00003,     0.065 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,   0,  -1 },     64.39438,   0.00004,     0.070 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   1,   0,   0 },    177.87895,   0.00003,     0.063 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,   0,  -1 },     93.30861,   0.00004,     0.067 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   1,   0,   0 },    210.57313,   0.00003,     0.061 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,   0,  -1 },    122.13677,   0.00003,     0.065 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   1,   0,   0 },    246.94465,   0.00002,     0.059 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   0,  -1 },    150.67908,   0.00004,     0.063 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,   0,  -1 },      6.48707,   0.00002,     0.073 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   1,   0,   0 },    117.27528,   0.00002,     0.065 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,   0,  -1 },     35.41250,   0.00002,     0.070 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   1,   0,   0 },    147.45743,   0.00002,     0.063 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,   0,  -1 },     64.33178,   0.00002,     0.068 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   1,   0,   0 },    178.55504,   0.00002,     0.061 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,   0,  -1 },     93.20470,   0.00003,     0.065 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   1,   0,   0 },    211.55878,   0.00002,     0.059 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   0,  -1 },    121.94675,   0.00004,     0.063 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,   0,  -1 },    150.25421,   0.00002,     0.061 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,   0,  -1 },     64.24843,   0.00002,     0.065 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,   0,  -1 },    121.70393,   0.00002,     0.061 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -2,   1 },     98.66413,   0.00001,     0.088 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -3,  -1 },     98.14017,   0.00002,     0.037 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,  -1 },     98.49615,   0.00016,     0.074 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   1 },     98.49615,   0.00016,     0.075 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,  -1 },     98.69230,   0.00006,     5.899 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,   1 },     98.64945,   0.00008,     0.037 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -2,  -1 },     98.66413,   0.00001,     0.088 },
 { {   0,   0,  17, -28,   0,   0,   0,   0,   2,   0,  -1 },    220.60890,   0.00002,     0.074 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   1,  -1 },    327.06701,   0.00007,     0.075 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   1,   1 },    327.06701,   0.00007,     0.074 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   1,  -1 },     28.79619,   0.00002,     0.075 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   1,   1 },     28.79619,   0.00002,     0.074 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   1,  -1 },    305.75894,   0.00002,     0.074 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   1,   1 },    305.75894,   0.00002,     0.075 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,  -2,   0,   1 },    230.46931,   0.00001,     0.064 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,  -2,   0,   3 },     50.46931,   0.00001,     0.089 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    191.77477,   0.00001,     0.073 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   3 },    191.77477,   0.00001,     0.076 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    237.79278,   0.00013,     0.068 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },     60.60768,   0.00010,     0.100 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -3,   0,   0 },    165.50750,   0.00002,     0.026 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },    165.50730,   0.00003,     0.040 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },    334.48967,   0.00005,     0.079 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },    337.67869,   0.00004,     0.038 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },    337.67630,   0.00004,     1.766 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    165.50702,   0.00017,     0.037 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    165.49094,   0.00308,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   2 },    170.27325,   0.00003,     0.076 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },    345.50681,   0.00017,     2.051 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    159.82704,   0.00003,     0.036 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },    163.53135,   0.00001,     0.828 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    148.04129,   0.00003,     0.068 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    148.04129,   0.00003,     0.083 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    163.53135,   0.00001,     0.649 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   1 },    159.82704,   0.00003,     0.040 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    169.59464,   0.00004,     0.442 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    346.82753,   0.00087,     0.091 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    346.76901,   0.00004,     0.041 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    155.66494,   0.00002,     0.101 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,   1 },    155.66731,   0.00002,     0.027 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,  -1,   0,   0 },    245.43408,   0.00002,     0.073 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,  -1,   0,   2 },     65.43408,   0.00002,     0.076 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   0,   0 },    256.61217,   0.00004,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   1,   0 },    256.61209,   0.00007,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1,   1 },    122.89786,   0.00002,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },    124.39636,   0.00006,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },    122.53014,   0.00051,     0.086 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,  -1 },    207.63508,   0.00006,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,   1 },    234.82796,   0.00006,     0.638 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -2,   0 },    256.61198,   0.00002,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.61246,   0.00039,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   2 },    256.07091,   0.00001,     1.182 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    256.60919,   0.00649,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   2 },    260.38209,   0.00004,     0.071 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -2 },     76.62136,   0.00003,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },     76.62234,   0.00056,     1.951 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,   0 },     76.62282,   0.00011,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,  -1 },    106.94847,   0.00005,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,  -1 },    107.47720,   0.00043,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    256.61993,   0.00011,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   1 },    108.26795,   0.00036,     4.814 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -3 },     90.34100,   0.00001,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    103.32436,   0.00039,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    103.12621,   0.00034,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -1 },    107.79187,   0.00036,     1.848 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     76.62154,   0.00004,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   1 },    107.22385,   0.00041,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2,  -1 },    107.38433,   0.00002,     0.079 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2,   1 },    106.67624,   0.00005,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,   0 },     76.67562,   0.00007,     0.069 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },     76.66930,   0.00065,     0.793 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   2 },     76.44727,   0.00002,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },    256.80537,   0.00006,     0.067 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    256.66200,   0.00753,     0.083 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   2 },     83.95409,   0.00001,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,   0 },    256.67761,   0.00046,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   2,   0 },    256.67148,   0.00002,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,  -1 },    221.16838,   0.00007,     0.432 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   1 },    215.41031,   0.00006,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    234.65740,   0.00029,     0.091 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   1 },    220.60344,   0.00005,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   1,   1 },    253.81284,   0.00001,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },    256.67149,   0.00008,     0.043 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   3,   0,  -2 },     76.68032,   0.00001,     0.101 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   3,   0,   0 },    256.67063,   0.00006,     0.027 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   1,   0,  -2 },     76.96164,   0.00002,     0.072 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   1,   0,   0 },    256.96164,   0.00002,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,  -1 },    179.98999,   0.00010,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,   1 },    179.98689,   0.00025,     0.028 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,  -1 },    179.98930,   0.00024,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,   1 },    179.98560,   0.00039,     0.045 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   2,  -1 },    179.97664,   0.00014,     0.028 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   2,   1 },    179.95684,   0.00003,     0.112 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,  -1 },    179.97414,   0.00001,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   1 },    179.97421,   0.00001,     0.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    179.99828,   0.00011,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -2,   1 },    179.99409,   0.00014,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    179.98981,   0.00074,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   1 },    179.98995,   0.00130,     0.042 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1 },    179.98623,   0.00489,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },    179.98525,   0.02008,     0.094 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   3 },    179.97827,   0.00006,     0.062 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,  -1 },    179.98223,   0.00575,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   1 },    179.98246,   0.00475,     0.390 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   1 },    179.98278,   0.00039,     0.063 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   3,   1 },    179.96935,   0.00004,     0.034 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,  -1 },      0.01961,   0.00001,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },      0.00796,   0.00030,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    168.23617,   0.00011,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },      0.00701,   0.00026,     0.596 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    359.99840,   0.00002,     0.084 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   1 },    359.99840,   0.00002,     0.067 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -3,  -1 },    180.01033,   0.00008,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -3,   1 },    180.01686,   0.00002,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,  -1 },    180.01028,   0.00085,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   1 },    180.01606,   0.00038,     0.080 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.00957,   0.00724,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   1 },    180.01218,   0.00594,     1.262 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   3 },    180.00609,   0.00001,     0.036 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1 },    179.98778,   0.00708,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },    179.99325,   0.00830,     0.071 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -3 },    180.06812,   0.00002,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.01282,   0.00560,     2.180 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   1 },    180.01228,   0.00616,     0.037 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,  -1 },    180.04129,   0.00032,     0.073 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   1 },    180.01627,   0.00074,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   3,  -1 },    180.02246,   0.00002,     0.037 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   3,   1 },    180.01781,   0.00007,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    180.07008,   0.00002,     0.037 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    180.01323,   0.00005,     0.073 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },    206.36891,   0.00005,     3.824 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    179.99934,   0.00002,     0.076 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },    180.02051,   0.00054,     2.335 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },     26.15315,   0.00094,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    180.02061,   0.00054,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   2 },    191.86002,   0.00001,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },    180.02151,   0.00003,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,   0 },     26.36894,   0.00005,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,   1 },    180.02151,   0.00006,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -3,  -1 },    359.73074,   0.00001,     0.036 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,  -1 },    359.98024,   0.00008,     0.069 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.84410,   0.00029,     0.760 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.94627,   0.00066,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -3 },    179.02650,   0.00002,     0.067 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    180.01893,   0.00812,     0.084 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   1 },    179.98207,   0.00114,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,  -1 },    180.01947,   0.00064,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,   1 },    180.00000,   0.00025,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   2,  -1 },    180.04042,   0.00003,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   2,   1 },    180.04361,   0.00003,     0.015 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -1,   1 },    180.01961,   0.00001,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -2,   1 },    179.83442,   0.00001,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -1,  -1 },    180.00336,   0.00007,     0.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -1,   1 },    179.97399,   0.00003,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   0,  -3 },    180.07395,   0.00002,     0.102 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   0,  -1 },    180.00641,   0.00008,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   0,   1 },    179.99504,   0.00003,     0.016 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   1,   0,  -2 },    102.70698,   0.00002,     0.073 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   1,   0,   0 },    282.70698,   0.00002,     0.076 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   1,   0,  -2 },    275.31541,   0.00003,     0.077 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,   1,   0,   0 },     95.31541,   0.00003,     0.072 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   1 },     77.65775,   0.00002,     0.029 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   1,  -1 },     77.39007,   0.00002,     0.021 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0 },    283.32041,   0.00004,     0.028 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   0 },    283.33617,   0.00006,     0.045 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     76.92245,   0.00007,     0.020 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   1 },     77.20926,   0.00007,     0.044 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,  -1 },     77.08769,   0.00045,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   1 },     77.04679,   0.00135,     0.103 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,  -1 },     77.17599,   0.00038,     0.043 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   1 },     76.88878,   0.00032,     0.280 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   1 },     76.03209,   0.00004,     0.059 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -2,   0 },    283.33617,   0.00002,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.33320,   0.00043,     0.042 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.33652,   0.00002,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.33656,   0.00744,     0.093 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   2 },    283.43123,   0.00005,     0.062 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.33026,   0.00037,     0.398 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   0 },     98.73834,   0.00002,     0.063 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,  -1 },     78.26164,   0.00002,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1 },     81.65342,   0.00031,     0.040 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   1 },     81.35161,   0.00027,     0.558 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -3 },    264.39695,   0.00002,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },     81.59078,   0.00014,     0.085 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 },     83.48746,   0.00014,     0.066 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1 },     81.18004,   0.00009,     0.686 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   1 },     83.10876,   0.00008,     0.035 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.33101,   0.00017,     0.080 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,  -2 },    103.33434,   0.00007,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.34042,   0.00117,     1.354 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   2 },    283.33816,   0.00004,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.33177,   0.00026,     0.078 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    103.33779,   0.02736,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   2 },    283.32581,   0.00001,     0.024 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   1,  -2 },    103.33618,   0.00005,     2.469 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   1,   0 },    103.33850,   0.00153,     0.037 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   2,   0 },    103.33785,   0.00009,     0.025 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,  -1 },    265.35431,   0.00006,     0.074 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   1 },    265.32785,   0.00006,     0.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,  -1 },    270.14332,   0.00036,     3.172 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    103.33746,   0.00006,     0.076 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   1 },    270.36532,   0.00035,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },    292.45148,   0.00012,     0.077 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   1 },    278.45824,   0.00004,     0.025 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.33773,   0.00030,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,  -2 },    103.33495,   0.00002,     0.084 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,   0 },    103.33987,   0.00017,     0.026 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   1,   0 },    103.33785,   0.00002,     0.019 },
 { {   0,   1,   1,  -3,   0,   0,   0,   0,   1,   0,   0 },    240.88312,   0.00003,     0.075 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,  -1 },    359.52820,   0.00004,     0.028 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   1 },    356.83322,   0.00009,     0.115 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,  -1 },    357.16913,   0.00002,     0.045 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   1 },    357.16913,   0.00002,     0.219 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,  -1,   0 },    189.43812,   0.00002,     0.043 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    188.76307,   0.00047,     0.103 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0 },      9.70934,   0.00004,     0.285 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1 },     19.41158,   0.00002,     0.042 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   1 },     16.30535,   0.00001,     0.358 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },     17.76394,   0.00005,     0.093 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1 },     17.73657,   0.00005,     0.062 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,  -1 },     16.30535,   0.00001,     0.407 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   1 },     19.41158,   0.00002,     0.034 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,  -1,   0 },    192.12060,   0.00003,     0.575 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0 },     12.07939,   0.00050,     0.067 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   1,   0 },     12.11927,   0.00003,     0.035 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,  -1 },     24.14234,   0.00003,     1.460 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,   1 },     24.13907,   0.00003,     0.036 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },     26.27779,   0.00010,     0.072 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   0 },    122.43020,   0.00006,     0.114 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   1,   0,   0 },    305.05026,   0.00008,     0.063 },
 { {   0,   1,  43, -59,   0,   0,   0,   0,   0,  -1,  -1 },     97.32856,   0.00001,     0.075 },
 { {   0,   1,  43, -59,   0,   0,   0,   0,   0,  -1,   1 },     97.32856,   0.00001,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   2,   1 },    295.93028,   0.00001,     0.088 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   1,  -1 },    300.91642,   0.00002,     0.037 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   2,  -1 },    300.40065,   0.00019,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   2,   1 },    300.40065,   0.00019,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   3,   1 },    300.91642,   0.00002,     0.037 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   2,  -1 },    295.93028,   0.00001,     0.088 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,  -2,   0,   1 },    230.46802,   0.00001,     0.067 },
 { {   0,   2,  -7,   0,   1,   0,   0,   0,  -2,   0,   3 },     50.46802,   0.00001,     0.084 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   0,   1 },    240.48143,   0.00001,     0.071 },
 { {   0,   2,  -6,   0,  -1,   0,   0,   0,  -2,   0,   3 },     60.48143,   0.00001,     0.079 },
 { {   0,   2,  -6,   0,   1,   0,   0,   0,  -2,   0,   1 },     38.69076,   0.00002,     0.072 },
 { {   0,   2,  -6,   0,   1,   0,   0,   0,  -2,   0,   3 },    218.69076,   0.00002,     0.078 },
 { {   0,   2,  -5,   0,  -1,   0,   0,   0,  -2,   0,   1 },     55.78912,   0.00002,     0.076 },
 { {   0,   2,  -5,   0,  -1,   0,   0,   0,  -2,   0,   3 },    235.78912,   0.00002,     0.073 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   1,  -1 },     67.58913,   0.00009,     0.038 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   1,   1 },     67.59142,   0.00009,     5.213 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,  -1 },     68.36040,   0.00003,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,   1 },     68.36040,   0.00003,     0.074 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    235.97460,   0.00023,     0.071 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   1,   1 },    247.58874,   0.00002,     0.040 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },     60.43254,   0.00012,     0.094 },
 { {   0,   2,  -4,   0,  -3,   0,   0,   0,  -1,   0,   0 },    280.30185,   0.00005,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    199.01405,   0.00009,     0.083 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,  -1 },    350.17452,   0.00002,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,   1 },    350.17627,   0.00002,     0.839 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    353.99036,   0.00003,     0.038 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    158.33957,   0.00002,     0.038 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    354.18438,   0.00045,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    158.33745,   0.00002,     3.281 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -2 },    352.64339,   0.00003,     0.037 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },    352.60004,   0.00008,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },    139.64439,   0.00004,     0.036 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },    132.49469,   0.00002,     1.718 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    137.30396,   0.00005,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    136.23733,   0.00005,     0.079 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },     48.61538,   0.00003,     1.092 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,   1 },     84.77215,   0.00004,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },    169.59489,   0.00005,     0.611 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -1,   2 },    349.23457,   0.00001,     0.040 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    349.74070,   0.00065,     0.086 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    349.73725,   0.00003,     0.040 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },     21.34304,   0.00009,     0.372 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },     20.08185,   0.00012,     0.041 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },     17.07285,   0.00043,     0.095 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },     18.94005,   0.00009,     0.027 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   1,   1 },     19.69852,   0.00001,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -6,   2,  -1 },     91.82239,   0.00001,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,  -1,  -1 },    271.81624,   0.00001,     0.013 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,  -1 },    271.03870,   0.00005,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,   1 },    271.14982,   0.00011,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,  -1 },    271.05859,   0.00009,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,   1 },    271.43589,   0.00017,     0.044 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   2,  -1 },    271.11093,   0.00007,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   2,   1 },    270.99008,   0.00002,     0.108 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   0,   0 },    256.70295,   0.00002,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },    256.70295,   0.00003,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    271.10130,   0.00004,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -2,   1 },    271.45062,   0.00005,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    271.11717,   0.00033,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   1 },    270.77063,   0.00061,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },    271.36467,   0.00210,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },    271.40179,   0.00986,     0.090 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   3 },    271.85576,   0.00003,     0.063 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },    271.50762,   0.00272,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },    271.44549,   0.00222,     0.456 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,  -1 },     93.70659,   0.00004,     0.088 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   1 },    272.01782,   0.00018,     0.065 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   3,   1 },    270.52639,   0.00002,     0.035 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.71381,   0.00015,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },     90.90980,   0.00007,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    256.72565,   0.00249,     0.083 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },     90.91076,   0.00006,     0.766 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   2 },    256.89894,   0.00002,     0.068 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },     76.65288,   0.00037,     0.879 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   2,   0 },     76.67810,   0.00004,     0.069 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -3,  -1 },    268.70203,   0.00003,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,  -1 },    268.71478,   0.00034,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,   1 },    268.61058,   0.00016,     0.078 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    268.89056,   0.00288,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     77.80012,   0.00005,     0.077 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },    268.33011,   0.00237,     2.394 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    257.33928,   0.00004,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    271.41979,   0.00330,     0.076 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    257.33923,   0.00004,     3.983 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    271.29933,   0.00350,     0.073 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    268.43061,   0.00234,    11.864 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    256.57651,   0.00026,     0.074 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   1 },    268.79256,   0.00270,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,  -1 },    268.48453,   0.00015,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   1 },    268.93190,   0.00030,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,   1 },    268.57540,   0.00003,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },     76.70607,   0.00017,     0.072 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,  -2 },    256.57442,   0.00003,     0.036 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },     76.58113,   0.00096,     1.572 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   2 },     76.58885,   0.00004,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -2 },    256.55980,   0.00012,     0.070 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    256.57252,   0.01388,     0.079 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   2 },     78.55609,   0.00001,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    256.57874,   0.00082,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   2,   0 },    256.57490,   0.00004,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,  -1 },    272.72991,   0.00007,     0.067 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   1 },     89.43391,   0.00003,     0.084 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },    271.24191,   0.00122,     0.592 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   1 },    270.96940,   0.00152,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -3 },    270.01093,   0.00002,     0.065 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    270.03887,   0.00635,     0.086 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   1 },    270.08604,   0.00112,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,  -1 },    269.88823,   0.00047,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,   1 },    270.10268,   0.00020,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   2,   1 },    270.47081,   0.00001,     0.015 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },    256.57490,   0.00014,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },    256.58752,   0.00011,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -2,   1 },    270.66736,   0.00003,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,  -1 },    270.73360,   0.00012,     0.043 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,   1 },    270.88706,   0.00005,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,  -1 },    270.75159,   0.00009,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,   1 },    271.98009,   0.00002,     0.016 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   0,  -1 },    359.36237,   0.00001,     0.076 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   0,   1 },    359.36237,   0.00001,     0.073 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   2,  -3 },    178.46520,   0.00003,     0.075 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   2,  -1 },    358.46520,   0.00003,     0.074 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   2,   0,  -3 },    178.72030,   0.00001,     0.066 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   2,   0,  -1 },    358.72030,   0.00001,     0.086 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   1,   0,  -2 },    268.61355,   0.00003,     0.071 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   1,   0,   0 },     88.61355,   0.00003,     0.078 },
 { {   0,   2,  -2,  -1,   0,   0,   0,   0,   1,  -2,   0 },    166.28235,   0.00001,     0.074 },
 { {   0,   2,  -2,  -1,   0,   0,   0,   0,   1,  -2,   2 },    166.28235,   0.00001,     0.075 },
 { {   0,   2,  -2,   0,  -3,   0,   0,   0,   1,   0,   0 },    100.30156,   0.00004,     0.075 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   1,   0,  -2 },    257.84348,   0.00002,     0.075 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   1,   0,   0 },     77.84348,   0.00002,     0.074 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   1,   1 },      0.32343,   0.00001,     0.022 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -2,  -1 },      0.36505,   0.00001,     0.011 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -1,  -1 },    359.94325,   0.00003,     0.013 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -1,   1 },      0.19436,   0.00004,     0.021 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,  -1 },    359.83350,   0.00012,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   1 },      0.12602,   0.00029,     0.029 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,  -1 },      0.26137,   0.00031,     0.021 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,   1 },      0.29466,   0.00048,     0.046 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   2,  -1 },      0.32849,   0.00015,     0.028 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   2,   1 },      0.34506,   0.00002,     0.121 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0,  -1 },      0.44803,   0.00002,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0,   3 },    180.41551,   0.00001,     0.227 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,  -1 },      0.28847,   0.00011,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   1 },      0.32439,   0.00011,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,  -1 },      0.27552,   0.00090,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,   1 },      0.30840,   0.00126,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },      0.38835,   0.00643,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },      0.36009,   0.02254,     0.099 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   3 },      0.33819,   0.00007,     0.060 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,  -1 },      0.29448,   0.00695,     0.042 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   1 },      0.26051,   0.00587,     0.313 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,  -1 },    180.70785,   0.00005,     0.096 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   1 },      0.27996,   0.00056,     0.061 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   3,   1 },      0.25472,   0.00006,     0.034 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },      4.08576,   0.00003,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    180.05661,   0.00009,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },      4.03452,   0.00085,     0.090 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    180.05638,   0.00009,     0.434 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },      6.01412,   0.00005,     0.468 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -3,  -1 },      0.36006,   0.00003,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,  -1 },      0.36590,   0.00031,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   1 },      0.26680,   0.00013,     0.084 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,  -1 },      0.50721,   0.00252,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   1 },      0.43726,   0.00216,     0.705 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },      0.28627,   0.00141,     0.082 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },      0.28311,   0.00250,     0.068 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,  -1 },      0.20210,   0.00237,     0.922 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   1 },      0.13293,   0.00245,     0.036 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -1 },      0.13336,   0.00011,     0.070 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,   1 },      0.22452,   0.00028,     0.024 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   3,   1 },      0.20505,   0.00003,     0.018 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,  -2 },    186.65310,   0.00003,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },      5.81245,   0.00047,     2.747 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },    183.00177,   0.00010,     0.076 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    359.94347,   0.00007,     5.069 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    185.88810,   0.00900,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    359.94355,   0.00008,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -3 },    359.65516,   0.00001,     0.075 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -2 },    188.52676,   0.00002,    32.762 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    359.99939,   0.00017,     0.074 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,   0 },    185.96083,   0.00052,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   2,   0 },    185.89453,   0.00003,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -3,  -1 },      0.26707,   0.00004,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,  -1 },    359.90510,   0.00036,     0.072 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   1 },    179.59069,   0.00007,     0.078 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,  -1 },      0.02586,   0.00504,     1.450 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   1 },      0.01635,   0.00603,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,  -3 },    359.99342,   0.00008,     0.070 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },      0.00073,   0.01993,     0.080 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   1 },      0.10530,   0.00198,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,  -1 },      0.02484,   0.00114,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,   1 },      0.15534,   0.00030,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2,  -1 },    359.99026,   0.00006,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2,   1 },      0.07578,   0.00002,     0.015 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },    185.89441,   0.00010,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,   0 },    185.89432,   0.00006,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -2,  -1 },      0.10470,   0.00004,     0.093 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -2,   1 },      0.03295,   0.00014,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,  -1 },    359.99604,   0.00038,     0.042 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,   1 },      0.03391,   0.00013,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.01014,   0.00022,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   0,   1 },      0.01680,   0.00003,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -3,   1 },      0.01900,   0.00001,     0.020 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,   0,  -3 },    346.94304,   0.00001,     0.070 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,   0,  -1 },    166.94304,   0.00001,     0.079 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,   0,  -3 },    170.19984,   0.00004,     0.074 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,   0,  -1 },    350.19984,   0.00004,     0.075 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,   0,  -3 },    349.13734,   0.00002,     0.075 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,   0,  -1 },    169.13734,   0.00002,     0.074 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   1 },    257.00310,   0.00005,     0.030 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   1,  -1 },    257.05968,   0.00002,     0.021 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   0,   0 },    283.28852,   0.00003,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,   0 },    283.21970,   0.00004,     0.046 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    257.10056,   0.00010,     0.020 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   1 },    257.03097,   0.00010,     0.045 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1 },    257.05405,   0.00063,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },    257.08245,   0.00168,     0.110 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,  -1 },    257.07006,   0.00051,     0.044 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   1 },    257.17851,   0.00045,     0.239 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   2,   1 },    257.08246,   0.00006,     0.057 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.25206,   0.00029,     0.043 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.22313,   0.00002,     0.027 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.22744,   0.00497,     0.099 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   2 },    283.21162,   0.00003,     0.060 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.21614,   0.00036,     0.319 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   2,   0 },    103.30969,   0.00003,     0.061 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -2,  -1 },    255.92735,   0.00002,     0.027 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1 },    256.81467,   0.00014,     0.041 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   1 },    256.72408,   0.00012,     0.414 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1 },     83.46685,   0.00004,     0.090 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },    249.60239,   0.00008,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1 },    254.33534,   0.00008,     0.480 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   1 },    252.90927,   0.00007,     0.035 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.31535,   0.00009,     0.084 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,  -2 },    103.23914,   0.00003,     0.039 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.29748,   0.00046,     0.733 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   2 },    283.42597,   0.00002,     0.035 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.03465,   0.00012,     0.082 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    103.29809,   0.01350,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   2 },    103.29811,   0.00001,     0.024 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,  -2 },    103.36000,   0.00002,     0.970 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,   0 },    103.29575,   0.00074,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   2,   0 },    103.29876,   0.00005,     0.024 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,  -1 },    256.48622,   0.00001,     0.077 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,   1 },    256.38998,   0.00001,     0.072 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1 },    238.66965,   0.00004,     3.223 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    283.33795,   0.00007,     0.073 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   1 },    243.14402,   0.00009,     0.037 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    246.06575,   0.00096,     0.074 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   1,  -1 },    245.75948,   0.00005,     0.037 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,  -2,   0 },    283.31716,   0.00002,     0.078 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.29875,   0.00015,     0.038 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,   0,   0 },    103.29223,   0.00009,     0.025 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,  -1 },    117.00935,   0.00003,     0.029 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   1 },    124.92104,   0.00006,     0.124 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   1,  -1 },    124.84177,   0.00001,     0.046 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   1 },    124.84177,   0.00001,     0.193 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    188.42318,   0.00029,     0.110 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },     47.25504,   0.00003,     0.098 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 },     47.32705,   0.00003,     0.060 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0 },     12.83244,   0.00029,     0.064 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,  -1 },     31.21391,   0.00002,     0.763 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,   1 },     31.21973,   0.00002,     0.036 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },     25.97661,   0.00006,     0.069 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   1 },    207.65704,   0.00001,     0.024 },
 { {   0,   2,   1,  -4,   0,   0,   0,   0,   2,   0,  -1 },    104.91035,   0.00002,     0.074 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    123.70431,   0.00005,     0.123 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    306.21736,   0.00007,     0.060 },
 { {   0,   3,  -8,   0,   1,   0,   0,   0,  -2,   0,   1 },     50.74651,   0.00001,     0.070 },
 { {   0,   3,  -8,   0,   1,   0,   0,   0,  -2,   0,   3 },    230.74651,   0.00001,     0.080 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,   0,  -1 },    249.81134,   0.00005,     0.097 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,   0,   1 },    249.73586,   0.00001,     0.027 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   0,   1 },    211.31490,   0.00001,     0.074 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   0,   3 },    101.31364,   0.00002,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    301.66265,   0.00004,     0.074 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   3 },    264.38359,   0.00001,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    328.85422,   0.00002,     0.069 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    148.75438,   0.00002,     0.098 },
 { {   0,   3,  -7,   0,   0,   0,   1,   0,  -2,   0,   1 },    238.38884,   0.00001,     0.074 },
 { {   0,   3,  -7,   0,   0,   0,   1,   0,  -2,   0,   3 },     58.38884,   0.00001,     0.075 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   0,   1 },    177.84774,   0.00005,     0.075 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   0,   3 },    357.84774,   0.00005,     0.074 },
 { {   0,   3,  -7,   2,   0,   0,   0,   0,  -1,   0,   0 },    342.19157,   0.00003,     0.074 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,  -1,  -1 },    147.64087,   0.00003,     0.019 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,  -1,   1 },    147.64086,   0.00006,     0.041 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,  -1 },    147.73032,   0.00018,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,   1 },    147.71499,   0.00087,     0.088 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   1,  -1 },    147.31485,   0.00023,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   1,   1 },    147.33811,   0.00020,     0.514 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   2,   1 },    147.64012,   0.00002,     0.066 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -2,  -1 },    149.13706,   0.00004,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -2,   1 },    149.73551,   0.00001,     0.076 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -1,  -1 },    149.27363,   0.00028,     0.037 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -1,   1 },    149.48462,   0.00026,     5.880 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,  -1 },    146.95199,   0.00029,     0.075 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,   1 },    146.95261,   0.00029,     0.074 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   1,  -1 },    149.48853,   0.00026,     6.118 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   1,   1 },    149.27808,   0.00028,     0.037 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   2,  -1 },    149.73453,   0.00001,     0.076 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   2,   1 },    149.13689,   0.00004,     0.025 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -2,  -1 },    147.65106,   0.00002,     0.066 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -1,  -1 },    147.34786,   0.00020,     0.516 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -1,   1 },    147.32418,   0.00023,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   0,  -1 },    147.72384,   0.00087,     0.088 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   0,   1 },    147.73895,   0.00018,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   1,  -1 },    147.65036,   0.00006,     0.041 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   1,   1 },    147.65035,   0.00003,     0.019 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   0,   0 },    239.01364,   0.00003,     0.026 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   1,   0 },    239.01308,   0.00004,     0.040 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    269.73995,   0.00001,     0.080 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   3 },    269.73995,   0.00001,     0.069 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   0 },    239.01621,   0.00024,     0.037 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   2 },    236.57700,   0.00001,    10.300 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    239.00808,   0.00416,     0.074 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   2 },    248.61063,   0.00006,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },     58.99820,   0.00021,     4.230 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   2 },    240.86754,   0.00002,     0.038 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1,  -1 },     91.69031,   0.00001,     0.036 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1,   1 },     91.69031,   0.00001,     1.045 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },     94.09790,   0.00002,     0.069 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },     93.47097,   0.00003,     0.081 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1,  -1 },     91.69031,   0.00001,     0.775 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1,   1 },     91.69031,   0.00001,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },     59.01286,   0.00004,     0.497 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },     59.93061,   0.00036,     0.089 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -1,  -1 },     90.18841,   0.00007,     0.326 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -1,   1 },     90.18848,   0.00007,     0.042 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },     89.13487,   0.00023,     0.098 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,   1 },     90.18717,   0.00006,     0.027 },
 { {   0,   3,  -6,   0,   0,   2,   0,   0,  -1,   0,   0 },    309.73905,   0.00002,     0.074 },
 { {   0,   3,  -6,   0,   0,   2,   0,   0,  -1,   0,   2 },    129.73905,   0.00002,     0.075 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,   0,   1 },     50.34316,   0.00003,     0.097 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   0,   1 },    344.97708,   0.00003,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   1,  -1 },    345.27889,   0.00001,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   1,   1 },    344.71128,   0.00005,     0.044 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   2,  -1 },    345.74067,   0.00002,     0.027 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    344.88344,   0.00009,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,   1 },    343.94957,   0.00020,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,  -1 },    344.34087,   0.00054,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    344.17213,   0.00295,     0.087 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   3 },    329.44688,   0.00001,     0.065 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,  -1 },    346.26804,   0.00085,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   1 },    346.32649,   0.00073,     0.550 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   2,  -1 },    165.05479,   0.00001,     0.085 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   2,   1 },    346.43595,   0.00007,     0.066 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,  -1,   0 },    349.96107,   0.00003,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    350.11093,   0.00050,     0.080 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   1,   0 },    169.96108,   0.00003,     1.310 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -2,  -1 },    338.11818,   0.00007,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -2,   1 },    337.52058,   0.00003,     0.076 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,  -1 },    339.30132,   0.00069,     0.037 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    167.97289,   0.00006,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   1 },    337.80609,   0.00061,    23.081 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    351.29490,   0.00079,     0.074 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    351.32968,   0.00076,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,  -1 },    337.79894,   0.00062,     3.446 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   1 },    339.76776,   0.00072,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   2,  -1 },    335.74476,   0.00003,     0.077 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   2,   1 },    338.88924,   0.00008,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },    167.69196,   0.00008,     0.990 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    347.07335,   0.00116,     0.082 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },    347.07663,   0.00006,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,  -1 },    347.60455,   0.00008,     0.065 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,   1 },    169.13223,   0.00002,     0.087 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,  -1 },    346.61447,   0.00082,     0.484 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   1 },    346.77834,   0.00099,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,  -3 },    346.67640,   0.00001,     0.064 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    346.40963,   0.00354,     0.089 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   1 },    346.51433,   0.00071,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,  -1 },    345.63046,   0.00023,     0.041 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,   1 },    345.45797,   0.00011,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   2,   1 },    346.49173,   0.00002,     0.015 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -2,   1 },    346.14963,   0.00002,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -1,  -1 },    346.26569,   0.00007,     0.044 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -1,   1 },    346.23156,   0.00002,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   0,  -1 },    346.00759,   0.00004,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   0,   1 },    345.34613,   0.00001,     0.016 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,  -2,   0,   1 },     82.52718,   0.00002,     0.088 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   0,  -1 },     78.49133,   0.00002,     0.074 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   0,   1 },     78.49133,   0.00002,     0.075 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,   0,  -1 },     82.84723,   0.00002,     0.089 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   0,  -1 },    179.00737,   0.00003,     0.026 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   0,   1 },    178.97916,   0.00013,     0.089 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   1,  -1 },    178.97117,   0.00004,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   1,   1 },    178.97117,   0.00004,     0.504 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -1,  -1 },    179.74698,   0.00006,     0.038 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -1,   1 },    180.64376,   0.00005,     4.719 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   0,  -1 },    173.04004,   0.00006,     0.075 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   0,   1 },    173.04004,   0.00006,     0.074 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   1,  -1 },    180.64376,   0.00005,     8.225 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   1,   1 },    179.74698,   0.00006,     0.037 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,  -1,  -1 },    178.96630,   0.00004,     0.528 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,  -1,   1 },    178.96630,   0.00004,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   0,  -1 },    178.95989,   0.00012,     0.088 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   0,   1 },    178.93392,   0.00002,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,  -1 },    271.45178,   0.00004,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,   1 },    270.91583,   0.00011,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,  -1 },    270.70021,   0.00007,     0.021 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,   1 },    270.65819,   0.00012,     0.046 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   2,  -1 },    270.56328,   0.00005,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   0,   0 },     76.61508,   0.00003,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   1,   0 },     76.62239,   0.00003,     0.044 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    270.58536,   0.00004,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    270.65102,   0.00025,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   1 },    270.63981,   0.00043,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },    270.63653,   0.00176,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    270.62094,   0.00683,     0.096 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   3 },    270.73530,   0.00003,     0.061 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,  -1 },    270.60305,   0.00199,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   1 },    270.60551,   0.00168,     0.355 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   2,  -1 },     89.18702,   0.00002,     0.093 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   2,   1 },    270.68298,   0.00016,     0.062 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -2,   1 },    270.16473,   0.00001,     0.041 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },     76.59617,   0.00021,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    270.34735,   0.00023,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },     90.35074,   0.00050,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     76.59330,   0.00357,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },     90.36329,   0.00046,     0.518 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   2 },     76.61728,   0.00003,     0.065 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },    256.80819,   0.00015,     0.567 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },     90.28190,   0.00008,     0.066 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },    256.58316,   0.00002,     0.067 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -3,  -1 },    270.92769,   0.00001,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2,  -1 },    270.87368,   0.00023,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2,   1 },    270.95533,   0.00009,     0.082 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },    271.12289,   0.00184,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },    271.05524,   0.00154,     0.958 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    270.35729,   0.00138,     0.080 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    270.74234,   0.00174,     0.070 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },    270.89913,   0.00135,     1.409 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    256.50120,   0.00002,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   1 },    270.91059,   0.00143,     0.036 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2,  -1 },    270.97943,   0.00009,     0.072 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2,   1 },    271.03382,   0.00016,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   3,   1 },    271.03502,   0.00001,     0.018 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -3,  -1 },    270.11464,   0.00004,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,  -1 },    270.09880,   0.00029,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    255.70857,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   1 },    269.32761,   0.00007,     6.397 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    270.42508,   0.00203,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },     76.30321,   0.00052,    95.796 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    270.27510,   0.00178,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },    255.17320,   0.00009,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    270.28513,   0.00277,     5.644 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    256.31567,   0.01018,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    270.28991,   0.00300,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    270.28193,   0.00017,     0.076 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    256.31126,   0.00055,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   1 },    270.29006,   0.00037,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   2,   0 },    256.31062,   0.00003,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   2,   1 },    270.28188,   0.00003,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -2,  -1 },     87.40869,   0.00004,     0.070 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -2,   1 },     93.42511,   0.00002,     0.080 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },    286.15719,   0.00011,     0.940 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     78.14395,   0.00002,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },    280.11472,   0.00023,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    271.22756,   0.00190,     0.082 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },    272.61633,   0.00027,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,  -1 },    271.14036,   0.00015,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,   1 },    270.65859,   0.00006,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -2,  -1 },    270.66097,   0.00004,     0.513 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -2,   1 },    270.51712,   0.00008,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    270.34231,   0.00027,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,   0 },    256.31056,   0.00011,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,   1 },    270.33004,   0.00014,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,  -1 },    270.29263,   0.00011,     0.041 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,   0 },    256.31052,   0.00006,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,   1 },    270.30094,   0.00006,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -2,   1 },    282.96735,   0.00004,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -1,  -1 },    282.87266,   0.00003,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -1,   1 },    282.91826,   0.00004,     0.020 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   1,   0,  -2 },    202.55955,   0.00002,     0.074 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   1,   0,   0 },     22.55955,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,   0,  -2 },     82.49577,   0.00004,     0.074 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,   0,   0 },    276.04536,   0.00008,     0.075 },
 { {   0,   3,  -4,   0,   3,   0,   0,   0,   1,   0,  -2 },    117.66853,   0.00001,     0.075 },
 { {   0,   3,  -4,   0,   3,   0,   0,   0,   1,   0,   0 },    297.66853,   0.00001,     0.074 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -2,  -1 },    188.73096,   0.00001,     0.075 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -2,   1 },      8.73096,   0.00001,     0.074 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,   0,  -3 },    186.82401,   0.00002,     0.073 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,   0,  -1 },      6.82401,   0.00002,     0.076 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -2,  -1 },    122.06366,   0.00003,     0.075 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -2,   1 },    122.06366,   0.00003,     0.074 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   0,  -1 },    357.83055,   0.00001,     0.016 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   0,   1 },      1.53445,   0.00005,     0.029 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   1,  -1 },      2.67627,   0.00002,     0.021 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   1,   1 },      2.71960,   0.00002,     0.048 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -2,  -1 },      0.07424,   0.00002,     0.016 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -2,   1 },      0.10263,   0.00002,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -1 },      1.21813,   0.00019,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,   1 },      1.01293,   0.00032,     0.044 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },      3.03672,   0.00077,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },      3.12384,   0.00195,     0.106 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },      2.49338,   0.00064,     0.043 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },      2.39000,   0.00056,     0.262 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   3 },    181.98923,   0.00001,     0.033 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   2,  -1 },    181.92325,   0.00001,     0.102 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   2,   1 },      2.49042,   0.00007,     0.059 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    354.76946,   0.00002,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    337.36563,   0.00001,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    355.47882,   0.00056,     0.095 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    347.34502,   0.00002,     0.341 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -3,  -1 },      0.02696,   0.00007,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -3,   1 },      0.22432,   0.00003,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,  -1 },      0.09584,   0.00068,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   1 },      0.09724,   0.00031,     0.089 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },      0.10566,   0.00533,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },      0.13793,   0.00455,     0.489 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    179.92927,   0.00207,     0.087 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },      3.14878,   0.00008,     0.065 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    179.55310,   0.00079,     0.585 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,   1 },    179.15765,   0.00030,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,  -1 },    179.75676,   0.00006,     0.067 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,   1 },    178.88741,   0.00005,     0.024 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },      4.43317,   0.00012,     1.011 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,  -2 },    184.47597,   0.00002,     0.079 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },      0.03022,   0.00009,     1.215 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    184.37576,   0.00216,     0.070 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },      0.03014,   0.00009,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    184.43317,   0.00012,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -4,  -1 },      1.46334,   0.00001,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -3,  -1 },      8.48130,   0.00002,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,  -3 },    179.98616,   0.00001,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,  -1 },    179.50818,   0.00031,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   1 },      0.02854,   0.00227,     0.074 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   3 },    359.98628,   0.00001,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,  -3 },    179.98674,   0.00011,     0.037 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.98734,   0.03094,    15.575 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    190.48784,   0.00004,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.98167,   0.02862,     0.037 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   3 },    359.98095,   0.00003,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -3 },      0.04296,   0.00013,     0.073 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },      0.05680,   0.04216,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   1 },    179.94870,   0.00289,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,  -3 },    180.01042,   0.00003,     2.514 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,  -1 },      0.05578,   0.00225,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,   1 },    179.93958,   0.00025,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,  -1 },      0.05056,   0.00015,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,   1 },    179.90458,   0.00001,     0.015 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,  -2,   1 },    180.01642,   0.00001,     0.039 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },    184.43316,   0.00002,     0.039 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -3,  -1 },    359.41485,   0.00005,     0.544 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -3,   1 },    359.66508,   0.00009,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,  -1 },      0.25378,   0.00003,     0.088 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,   1 },    179.97888,   0.00064,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,  -1 },    179.91387,   0.00048,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,   1 },    179.98119,   0.00048,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.08660,   0.00021,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   0,   1 },    179.96821,   0.00007,     0.015 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   1,  -1 },      0.05064,   0.00003,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -3,   1 },    180.02406,   0.00001,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -2,  -1 },    179.98628,   0.00002,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -2,   1 },    179.98628,   0.00002,     0.016 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   1,   0,  -2 },    274.35140,   0.00001,     0.080 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   1,   0,   0 },     94.35140,   0.00001,     0.070 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,   0,  -3 },    128.94041,   0.00002,     0.074 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,   0,  -1 },    308.94041,   0.00002,     0.075 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -3,   1,   0 },    283.16205,   0.00003,     0.048 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },    256.08652,   0.00007,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },    257.09872,   0.00019,     0.119 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,  -1 },    256.93912,   0.00006,     0.045 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,   1 },    256.81090,   0.00005,     0.208 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.20901,   0.00017,     0.044 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.15858,   0.00002,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.17493,   0.00326,     0.105 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   2 },    283.14382,   0.00002,     0.058 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.21086,   0.00028,     0.266 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -2,  -1 },    256.90992,   0.00003,     0.027 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,  -1 },    256.95577,   0.00027,     0.042 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,   1 },    256.93957,   0.00022,     0.329 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },     81.83455,   0.00009,     0.095 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    224.03787,   0.00001,     0.061 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.32511,   0.00004,     0.089 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.26645,   0.00022,     0.503 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },    102.02028,   0.00008,     0.086 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    103.29245,   0.00833,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   2 },    103.30302,   0.00001,     0.024 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,  -2 },    103.67889,   0.00001,     0.604 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,   0 },    103.28986,   0.00045,     0.035 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   2,   0 },    103.29405,   0.00003,     0.024 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -2,   1 },    257.36565,   0.00002,     0.069 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,  -1 },    256.14070,   0.00017,     1.069 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    283.34259,   0.00002,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,   1 },    256.17548,   0.00018,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    255.48743,   0.00028,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,   1 },    258.45332,   0.00003,     0.024 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -3,   0 },    283.31278,   0.00002,     4.640 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,   0 },    103.31414,   0.00033,     0.074 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -2 },    103.29667,   0.00011,     8.477 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.29233,   0.00022,     0.037 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,   0,  -2 },    103.46494,   0.00002,     0.076 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,   0,   0 },    103.28364,   0.00005,     0.025 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -1,  -1 },    257.04391,   0.00002,     0.039 },
 { {   0,   3,  -1,  -3,   0,   0,   0,   0,   2,   0,  -3 },    338.41275,   0.00001,     0.075 },
 { {   0,   3,  -1,  -3,   0,   0,   0,   0,   2,   0,  -1 },    338.41275,   0.00001,     0.074 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1 },     38.16582,   0.00002,     0.029 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },     38.12074,   0.00002,     0.134 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    188.81502,   0.00016,     0.118 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,  -1 },     29.12132,   0.00003,     0.105 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },     29.11629,   0.00003,     0.058 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },     12.84714,   0.00018,     0.062 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1 },     28.69034,   0.00002,     0.517 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,   1 },     28.69786,   0.00002,     0.035 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },     30.98960,   0.00004,     0.066 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,   1 },    204.36445,   0.00001,     0.024 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    124.71875,   0.00004,     0.133 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    307.12333,   0.00006,     0.058 },
 { {   0,   3,   7, -19,   0,   0,   0,   0,   2,  -2,  -1 },    311.20488,   0.00001,     0.075 },
 { {   0,   3,   7, -19,   0,   0,   0,   0,   2,  -2,   1 },    311.20488,   0.00001,     0.074 },
 { {   0,   4, -15,  14,   0,   0,   0,   0,  -1,   0,   0 },    286.27969,   0.00001,     0.074 },
 { {   0,   4, -15,  14,   0,   0,   0,   0,  -1,   0,   2 },    286.27969,   0.00001,     0.075 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   0,   1 },     51.63411,   0.00001,     0.068 },
 { {   0,   4, -10,   0,   2,   0,   0,   0,  -2,   0,   3 },    231.63411,   0.00001,     0.082 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,  -2,   0,   1 },     60.47421,   0.00001,     0.078 },
 { {   0,   4,  -8,   0,  -1,   0,   0,   0,  -2,   0,   3 },    240.47421,   0.00001,     0.072 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },     51.95076,   0.00002,     0.078 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    327.16288,   0.00004,     0.072 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    149.47290,   0.00002,     0.092 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     61.44196,   0.00011,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },     64.37201,   0.00018,     0.084 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },     84.15427,   0.00003,     0.092 },
 { {   0,   4,  -7,   0,   2,   0,   0,   0,  -1,   0,   0 },    112.12784,   0.00001,     0.079 },
 { {   0,   4,  -7,   0,   2,   0,   0,   0,  -1,   0,   2 },    292.12784,   0.00001,     0.071 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    345.25334,   0.00002,     0.020 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -1,   1 },    345.84737,   0.00004,     0.042 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,  -1 },    344.77231,   0.00017,     0.027 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    345.52850,   0.00065,     0.092 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,  -1 },    345.66957,   0.00019,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,   1 },    345.69320,   0.00014,     0.409 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,  -2 },    168.13207,   0.00001,     0.026 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    329.56998,   0.00002,     0.084 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },    169.36111,   0.00002,     0.720 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -2,  -1 },    344.02688,   0.00003,     0.025 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -2,   1 },    345.98511,   0.00001,     0.079 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,  -1 },    343.16362,   0.00026,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,   1 },    342.90551,   0.00023,     1.495 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    344.81289,   0.00030,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },    344.78310,   0.00029,     0.072 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,  -1 },    342.44705,   0.00018,     2.982 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,   1 },    342.77184,   0.00021,     0.037 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   2,  -1 },    342.82893,   0.00001,     0.074 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   2,   1 },    342.97000,   0.00003,     0.025 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -2,   0 },    170.60782,   0.00005,     0.073 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },    169.32376,   0.00017,     2.598 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },    349.43957,   0.00002,     0.072 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    349.35807,   0.00283,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },    349.36111,   0.00015,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    344.79249,   0.00023,     0.085 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,   1 },    344.81176,   0.00004,     0.026 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   3,  -1,   0 },    349.36111,   0.00003,     0.041 },
 { {   0,   4,  -6,   3,   0,   0,   0,   0,   2,   0,  -3 },    181.39605,   0.00002,     0.074 },
 { {   0,   4,  -6,   3,   0,   0,   0,   0,   2,   0,  -1 },    181.39605,   0.00002,     0.075 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -3,   1,   0 },    256.54618,   0.00002,     0.045 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     93.59887,   0.00003,     0.020 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,  -1,   1 },     93.53882,   0.00002,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,  -1 },     93.69207,   0.00016,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },     93.36233,   0.00055,     0.102 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   3 },    273.52972,   0.00001,     0.059 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,  -1 },     93.48567,   0.00016,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,   1 },     93.44927,   0.00017,     0.290 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.53467,   0.00020,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    256.53671,   0.00346,     0.092 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   2 },    256.51251,   0.00002,     0.063 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,   0 },    256.51319,   0.00002,     0.064 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -3,  -1 },    271.61508,   0.00001,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -2,  -1 },    271.64255,   0.00003,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,  -1 },    271.30052,   0.00023,     0.040 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,   1 },    271.23674,   0.00019,     0.599 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    268.83890,   0.00006,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    264.45876,   0.00002,     0.067 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1,  -1 },     92.65509,   0.00002,     0.749 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,   0 },    256.51519,   0.00015,     0.079 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,  -2 },     76.63171,   0.00007,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },    256.51252,   0.00135,     1.626 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   2 },    256.52039,   0.00003,     0.036 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -2 },     77.08526,   0.00028,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },     76.51440,   0.02884,     0.072 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   2 },    254.58710,   0.00002,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,  -2 },     76.33526,   0.00005,     3.552 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },     76.51320,   0.00164,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   2,   0 },     76.51317,   0.00010,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -3,   1 },    274.25637,   0.00001,     9.509 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -2,  -1 },     91.41293,   0.00012,     0.073 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,  -1 },     91.33236,   0.00117,     2.280 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     78.12030,   0.00005,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   1 },     91.37688,   0.00133,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },     92.10566,   0.00240,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,   1 },     91.65866,   0.00026,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   1,  -1 },     92.14390,   0.00014,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   1,   1 },     92.04106,   0.00002,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,  -1,   0 },     76.51312,   0.00032,     0.039 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,  -2 },     76.26250,   0.00002,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },     76.51321,   0.00018,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   1,   0 },     76.51322,   0.00002,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -2,   1 },     91.41247,   0.00003,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -1,  -1 },     91.78282,   0.00007,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -1,   1 },     91.61508,   0.00001,     0.020 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   1,   0,  -2 },     88.60950,   0.00002,     0.078 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   1,   0,   0 },    268.60950,   0.00002,     0.071 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,   0,  -3 },      6.76846,   0.00002,     0.076 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,   0,  -1 },    186.76846,   0.00002,     0.073 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   1,  -1 },    359.23144,   0.00001,     0.021 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    359.15483,   0.00003,     0.020 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -1,   1 },    359.69197,   0.00005,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },    359.16444,   0.00018,     0.028 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.46875,   0.00055,     0.113 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   1,  -1 },    359.17364,   0.00016,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   1,   1 },    359.13915,   0.00015,     0.225 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   2,   1 },    358.99113,   0.00001,     0.057 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     27.31788,   0.00019,     0.101 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,  -1 },      0.14250,   0.00004,     0.027 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },      0.04588,   0.00043,     0.041 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },      0.06769,   0.00035,     0.375 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    359.86749,   0.00043,     0.092 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    180.17421,   0.00056,     0.063 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.37450,   0.00017,     0.428 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,   1 },    180.20429,   0.00035,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   2,  -1 },    180.03123,   0.00002,     0.064 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   2,   1 },    180.09345,   0.00004,     0.024 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },      0.37121,   0.00001,     0.085 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },     10.53658,   0.00005,     0.619 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },      0.37121,   0.00001,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },    184.37229,   0.00002,     0.084 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },      0.12222,   0.00005,     0.690 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    190.35913,   0.00115,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },      0.12223,   0.00005,     0.035 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    190.33734,   0.00007,     0.036 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -3,  -1 },    180.25995,   0.00002,     0.039 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.02493,   0.00022,     0.079 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,   1 },      0.07246,   0.00007,     0.071 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,  -3 },    180.01449,   0.00002,     0.038 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.02346,   0.00244,     1.782 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.02913,   0.00371,     0.036 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -3 },    180.01109,   0.00014,     0.077 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    180.03860,   0.02359,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },      0.26690,   0.00016,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   1,  -1 },    180.03811,   0.00127,     0.037 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   1,   1 },      0.10044,   0.00003,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   2,  -1 },    180.05096,   0.00008,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   2,   1 },    359.97773,   0.00001,     0.015 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -1,  -3 },    180.06452,   0.00001,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    180.06452,   0.00001,     0.077 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -2,  -1 },    180.02454,   0.00002,     0.083 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -2,   1 },    180.02809,   0.00009,     0.026 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,  -1 },    180.03685,   0.00037,     0.040 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,   1 },    180.04401,   0.00006,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,  -1 },    180.04424,   0.00009,     0.026 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,   1 },      0.02722,   0.00002,     0.015 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   2,   0,  -3 },    166.93582,   0.00001,     0.077 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   2,   0,  -1 },    346.93582,   0.00001,     0.072 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },    283.31277,   0.00002,     0.049 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },    250.61971,   0.00002,     0.029 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },    253.80617,   0.00005,     0.128 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },    253.24184,   0.00001,     0.047 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },    253.24184,   0.00001,     0.184 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.30820,   0.00012,     0.045 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.31593,   0.00220,     0.113 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.30073,   0.00022,     0.228 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    257.09363,   0.00004,     0.043 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },    257.09363,   0.00003,     0.273 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    243.42468,   0.00002,     0.101 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.32585,   0.00003,     0.094 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.34273,   0.00011,     0.382 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.32906,   0.00005,     0.091 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    103.32908,   0.00566,     0.063 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    103.32945,   0.00030,     0.034 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   2,   0 },    103.32912,   0.00002,     0.024 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },    259.41159,   0.00004,     0.640 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,   1 },    259.36043,   0.00004,     0.035 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    259.84577,   0.00018,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -2,   0 },    103.32294,   0.00002,     0.071 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.32912,   0.00006,     0.037 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },    103.33032,   0.00003,     0.025 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },     30.37783,   0.00002,     0.030 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },     30.35857,   0.00002,     0.147 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    187.76359,   0.00011,     0.127 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },     27.14725,   0.00003,     0.112 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },     27.14093,   0.00003,     0.056 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },     10.69364,   0.00012,     0.059 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },     25.92241,   0.00003,     0.063 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,   1 },    206.23751,   0.00001,     0.023 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    125.53119,   0.00003,     0.146 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    307.80650,   0.00005,     0.056 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   2,  -1 },    230.45017,   0.00003,     0.074 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   2,   1 },     50.45017,   0.00003,     0.075 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    324.28314,   0.00002,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    150.97774,   0.00002,     0.087 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    173.67921,   0.00002,     0.096 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     71.78681,   0.00002,     0.019 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -1,   1 },     72.24124,   0.00002,     0.041 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,  -1 },     71.61039,   0.00011,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },     71.94075,   0.00056,     0.089 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,  -1 },     72.36959,   0.00016,     0.040 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,   1 },     72.48088,   0.00011,     0.483 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },     24.63174,   0.00003,     0.082 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -2,  -1 },     69.14175,   0.00003,     0.025 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,  -1 },     69.75879,   0.00021,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,   1 },     69.38887,   0.00018,     3.397 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },     72.18962,   0.00026,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },     72.17923,   0.00026,     0.074 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,  -1 },     69.15925,   0.00016,    25.549 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,   1 },     69.58383,   0.00019,     0.037 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   2,   1 },     69.74644,   0.00003,     0.025 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    247.83902,   0.00002,     1.317 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     67.83828,   0.00034,     0.080 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },     67.83827,   0.00002,     0.039 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },     72.65184,   0.00008,     0.552 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,   1 },     72.31058,   0.00012,     0.040 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },     71.41016,   0.00045,     0.087 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,   1 },     71.27974,   0.00008,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   1,  -1 },     71.64402,   0.00002,     0.040 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   1,   1 },     71.63216,   0.00001,     0.019 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },    272.21933,   0.00002,     0.042 },
 { {   0,   5,  -7,   0,  -1,   0,   0,   0,   1,   0,  -2 },    249.32133,   0.00001,     0.074 },
 { {   0,   5,  -7,   0,  -1,   0,   0,   0,   1,   0,   0 },     69.32133,   0.00001,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    168.76345,   0.00002,     0.020 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,  -1,   1 },    168.43043,   0.00004,     0.043 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,  -1 },    168.93951,   0.00013,     0.027 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    168.53028,   0.00052,     0.098 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,  -1 },    168.47316,   0.00015,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,   1 },    168.48832,   0.00013,     0.326 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1,   0 },    349.19940,   0.00010,     0.041 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,  -2 },    169.47135,   0.00002,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    349.19435,   0.00191,     0.089 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,  -2 },    169.28805,   0.00001,     0.040 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,   0 },    349.32658,   0.00029,     0.496 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   2,   0 },    349.28240,   0.00003,     0.065 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,  -1 },    170.20978,   0.00008,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,   1 },    170.02379,   0.00007,     0.773 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,  -1 },    144.03637,   0.00004,     0.081 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,   1 },    169.80273,   0.00008,     0.069 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,  -1 },    169.39426,   0.00006,     1.041 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,   1 },    169.50196,   0.00006,     0.036 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -2,   0 },    349.23976,   0.00010,     0.077 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,  -2 },    169.53216,   0.00015,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },    349.27805,   0.00188,     4.156 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,  -2 },    170.21814,   0.00052,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    169.28198,   0.03712,     0.074 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   2 },    349.28223,   0.00004,     0.025 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,  -2 },    169.12161,   0.00010,    10.764 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,   0 },    169.28699,   0.00216,     0.037 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   2,   0 },    169.29293,   0.00013,     0.025 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,  -1 },    162.99685,   0.00010,     1.230 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,   1 },    165.70567,   0.00013,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    166.73525,   0.00045,     0.080 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,   1 },    166.16444,   0.00004,     0.025 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   1,  -1 },     83.34235,   0.00001,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -1,  -2 },    349.84962,   0.00001,     0.536 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -1,   0 },    169.27605,   0.00039,     0.040 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   0,  -2 },    166.35397,   0.00003,     0.088 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },    169.28151,   0.00024,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   1,   0 },    169.28248,   0.00003,     0.019 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -1,  -1 },     15.24325,   0.00003,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,   0,  -1 },     15.24326,   0.00003,     0.027 },
 { {   0,   5,  -7,   0,   1,   0,   0,   0,   1,   0,  -2 },    255.03919,   0.00002,     0.075 },
 { {   0,   5,  -7,   0,   1,   0,   0,   0,   1,   0,   0 },     75.03919,   0.00002,     0.074 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   0,  -1 },      9.41215,   0.00002,     0.026 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   0,   1 },      9.41215,   0.00008,     0.088 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   1,  -1 },      9.41266,   0.00002,     0.040 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   1,   1 },      9.41266,   0.00002,     0.515 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -1,  -1 },      6.46078,   0.00003,     0.037 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -1,   1 },      4.23969,   0.00002,     5.965 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   0,  -1 },     13.20448,   0.00005,     0.075 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   0,   1 },     13.20448,   0.00005,     0.074 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   1,  -1 },      4.23969,   0.00002,     6.029 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   1,   1 },      6.46078,   0.00003,     0.037 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,  -1,  -1 },      9.39970,   0.00002,     0.516 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,  -1,   1 },      9.39970,   0.00002,     0.040 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,   0,  -1 },      9.40021,   0.00008,     0.088 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,   0,   1 },      9.40021,   0.00002,     0.026 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   2,   0,  -3 },    278.01987,   0.00002,     0.074 },
 { {   0,   5,  -6,   0,  -1,   0,   0,   0,   2,   0,  -1 },     98.01987,   0.00002,     0.075 },
 { {   0,   5,  -6,   0,   0,  -1,   0,   0,   2,   0,  -3 },      3.57645,   0.00001,     0.074 },
 { {   0,   5,  -6,   0,   0,  -1,   0,   0,   2,   0,  -1 },    183.57645,   0.00001,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,  -1 },     85.53732,   0.00006,     0.028 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },     86.60188,   0.00018,     0.109 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,  -1 },     86.99783,   0.00007,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,   1 },     84.43579,   0.00005,     0.246 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.28596,   0.00008,     0.043 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.93432,   0.00168,     0.098 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },     76.25689,   0.00007,     0.332 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -3,  -1 },     91.84748,   0.00001,     0.020 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -2,  -1 },     91.84748,   0.00003,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,  -1 },     91.77799,   0.00027,     0.041 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,   1 },     91.48900,   0.00014,     0.436 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -3 },    272.35142,   0.00004,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },     92.35437,   0.00207,     0.089 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },     99.33690,   0.00001,     0.064 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,  -1 },     92.26828,   0.00060,     0.510 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,   1 },    273.30189,   0.00003,     0.035 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,  -1 },     92.32777,   0.00005,     0.066 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -2,   0 },    256.41893,   0.00005,     0.083 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,  -2 },     76.22668,   0.00002,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },    256.36758,   0.00036,     0.806 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },     75.65166,   0.00008,     0.081 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },     76.38459,   0.00928,     0.069 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1,  -2 },     76.59988,   0.00001,     1.102 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },     76.38036,   0.00052,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   2,   0 },     76.38213,   0.00003,     0.024 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,  -3 },    272.33852,   0.00005,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,  -1 },     93.66969,   0.00003,     0.077 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,   1 },     93.59356,   0.00003,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -3 },    272.35924,   0.00042,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -1 },     92.53466,   0.00228,     5.352 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    256.36944,   0.00010,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   1 },    271.76845,   0.00098,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -3 },    272.53140,   0.00164,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.30597,   0.06663,     0.074 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,   1 },     93.25189,   0.00003,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -3 },    272.23384,   0.00025,     6.818 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -1 },    272.30644,   0.00398,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,   1 },     92.15030,   0.00002,     0.019 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   2,  -3 },    272.27447,   0.00001,     0.076 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   2,  -1 },    272.30675,   0.00026,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,  -1,   0 },     76.38210,   0.00010,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,   0,   0 },     76.38207,   0.00006,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -2,   1 },    272.06312,   0.00003,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,  -3 },     92.44936,   0.00003,     0.521 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,  -1 },    272.28106,   0.00074,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,   1 },    271.84748,   0.00001,     0.019 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,  -3 },    272.17403,   0.00011,     0.088 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,  -1 },    272.30529,   0.00045,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   1,  -1 },    272.30635,   0.00005,     0.019 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,   2,   0,  -3 },    278.70545,   0.00001,     0.075 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,   2,   0,  -1 },     98.70545,   0.00001,     0.074 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,  -1 },    358.34454,   0.00006,     0.029 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.36171,   0.00013,     0.122 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   1,  -1 },    359.25061,   0.00005,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   1,   1 },    359.33171,   0.00007,     0.197 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },     17.56100,   0.00017,     0.108 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,  -1 },      0.55290,   0.00006,     0.042 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   1 },      0.43959,   0.00009,     0.304 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   3 },    179.98393,   0.00001,     0.033 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    351.79937,   0.00015,     0.097 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    180.16312,   0.00047,     0.060 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,  -1 },    181.54001,   0.00019,     0.338 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,   1 },    180.10526,   0.00027,     0.034 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,  -1 },    180.00715,   0.00002,     0.062 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,   1 },    180.05299,   0.00003,     0.023 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },     10.33519,   0.00003,     0.446 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },      0.30479,   0.00003,     0.482 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    190.26462,   0.00083,     0.065 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },      0.30363,   0.00003,     0.035 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },    190.25963,   0.00005,     0.035 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -2,  -1 },    179.97875,   0.00014,     0.083 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.20192,   0.00128,     0.843 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.03880,   0.00171,     0.036 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,  -3 },    177.50654,   0.00004,     0.081 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    180.06538,   0.00796,     0.069 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   1 },    359.99937,   0.00030,     0.024 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,  -3 },    171.96805,   0.00001,     1.172 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,  -1 },    180.56794,   0.00036,     0.036 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,   1 },      0.02032,   0.00007,     0.018 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   2,  -1 },    180.06376,   0.00003,     0.024 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   2,   1 },    359.96509,   0.00001,     0.015 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -1,  -1 },      0.06488,   0.00010,     0.074 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   0,  -2 },    194.50708,   0.00003,     0.075 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -2,  -1 },    180.08631,   0.00001,     0.079 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -2,   1 },    180.05470,   0.00005,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -1,  -1 },    180.03680,   0.00014,     0.039 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -1,   1 },    180.06549,   0.00001,     0.019 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   0,   1 },      0.01678,   0.00002,     0.015 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.29080,   0.00010,     0.047 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.26644,   0.00002,     0.029 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.29672,   0.00151,     0.121 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.28923,   0.00019,     0.199 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },    257.09947,   0.00002,     0.044 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },     87.19177,   0.00001,     0.108 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.32512,   0.00002,     0.100 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },    283.35431,   0.00004,     0.309 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.32757,   0.00004,     0.097 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    103.32621,   0.00398,     0.061 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    103.32651,   0.00022,     0.034 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },    267.00310,   0.00002,     0.457 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },    267.00225,   0.00002,     0.034 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    262.43833,   0.00013,     0.065 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },     89.81529,   0.00001,     0.024 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.32647,   0.00005,     0.036 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },     29.24776,   0.00002,     0.030 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },     29.24745,   0.00002,     0.162 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    185.74221,   0.00006,     0.138 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },     25.77988,   0.00003,     0.121 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },     25.77359,   0.00003,     0.054 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },      8.06323,   0.00007,     0.057 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },     29.20058,   0.00003,     0.061 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,   1 },    210.11842,   0.00001,     0.023 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    126.19004,   0.00002,     0.160 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    308.32067,   0.00004,     0.054 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,  -2,   0,   1 },     23.13190,   0.00003,     0.075 },
 { {   0,   6, -12,   0,   2,   0,   0,   0,  -2,   0,   3 },    203.13190,   0.00003,     0.074 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   1 },    299.10566,   0.00001,     0.080 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   3 },    119.10566,   0.00001,     0.070 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,  -2,   0,   1 },     50.44615,   0.00001,     0.080 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,  -2,   0,   3 },    230.44615,   0.00001,     0.070 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    148.48146,   0.00002,     0.079 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    149.37872,   0.00003,     0.082 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   0,  -1 },    293.07924,   0.00002,     0.026 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   0,   1 },    293.07924,   0.00009,     0.088 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   1,  -1 },    293.07959,   0.00002,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   1,   1 },    293.07959,   0.00002,     0.514 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -1,  -1 },    291.21176,   0.00004,     0.037 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -1,   1 },    290.17003,   0.00002,     5.787 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   0,  -1 },    291.58394,   0.00003,     0.075 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   0,   1 },    291.58394,   0.00003,     0.074 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   1,  -1 },    290.17003,   0.00002,     6.223 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   1,   1 },    291.21176,   0.00004,     0.037 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,  -1,  -1 },    293.07046,   0.00002,     0.517 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,  -1,   1 },    293.07046,   0.00002,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   0,  -1 },    293.07083,   0.00009,     0.088 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   0,   1 },    293.07083,   0.00002,     0.026 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,  -1 },     68.47053,   0.00002,     0.027 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,   1 },     71.32757,   0.00007,     0.095 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    247.07999,   0.00003,     0.086 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,  -1 },     71.27334,   0.00003,     0.039 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,   1 },     71.34755,   0.00002,     1.087 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0,  -1 },     72.48527,   0.00002,     0.079 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0,   1 },     72.48527,   0.00002,     0.071 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1,  -1 },     71.34755,   0.00002,     1.705 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1,   1 },     71.27334,   0.00003,     0.036 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -2,   0 },     65.91780,   0.00006,     0.075 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },    247.07999,   0.00004,     7.470 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     67.07606,   0.00079,     0.076 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },     67.07998,   0.00004,     0.038 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,  -1,  -1 },    270.18317,   0.00002,     0.842 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,  -1,   1 },    270.18317,   0.00002,     0.039 },
 { {   0,   6,  -9,   0,   3,   0,   0,   0,   1,   0,  -2 },    246.28830,   0.00002,     0.074 },
 { {   0,   6,  -9,   0,   3,   0,   0,   0,   1,   0,   0 },     66.28830,   0.00002,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,  -1,   1 },    347.24565,   0.00002,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },    172.06941,   0.00003,     0.105 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.49513,   0.00028,     0.094 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   1,   0 },    349.11872,   0.00002,     0.379 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -2,  -1 },    347.27873,   0.00004,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -2,   1 },    167.11554,   0.00012,     0.088 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,  -1 },    347.19574,   0.00061,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,   1 },    347.18239,   0.00053,     0.521 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    341.95427,   0.00003,     0.086 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    345.50412,   0.00011,     0.066 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    349.13990,   0.00012,     1.155 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,  -2 },    169.31955,   0.00002,     0.079 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    169.12140,   0.00246,     0.071 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },    169.11871,   0.00014,     0.037 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -4,  -1 },    168.07384,   0.00002,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,  -1 },    168.10964,   0.00015,     0.037 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,   1 },    180.25854,   0.00001,     6.855 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,  -1 },    167.25813,   0.00138,     0.074 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,   1 },    166.94278,   0.00107,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },    167.22739,   0.00329,     5.330 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    170.64101,   0.00002,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   1 },    167.24129,   0.00349,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    168.48459,   0.00090,     0.077 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   1 },    167.22768,   0.00045,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   1,  -1 },    168.91902,   0.00004,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   1,   1 },    167.24632,   0.00004,     0.019 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },    169.11872,   0.00003,     0.039 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -3,  -1 },    166.06406,   0.00002,     0.510 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -3,   1 },    166.55842,   0.00005,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -2,  -1 },    167.13840,   0.00016,     0.089 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -2,   1 },    167.17344,   0.00012,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -1,  -1 },    167.23588,   0.00012,     0.041 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -1,   1 },    167.22479,   0.00007,     0.019 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   2,   0,  -3 },    269.87240,   0.00001,     0.073 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   2,   0,  -1 },     89.87240,   0.00001,     0.076 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0,  -1 },     87.96839,   0.00001,     0.028 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },     89.68861,   0.00004,     0.117 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.24529,   0.00004,     0.044 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    256.25502,   0.00093,     0.104 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1,   0 },     76.26923,   0.00007,     0.275 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,  -1 },    182.09631,   0.00001,     0.042 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,   1 },    141.45458,   0.00001,     0.342 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,  -1 },     92.82532,   0.00005,     0.094 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   1 },    272.47454,   0.00003,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,  -1 },    270.35305,   0.00002,     0.387 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,   1 },    270.98889,   0.00003,     0.034 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -2,   0 },    260.79637,   0.00003,     0.088 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },    256.27798,   0.00013,     0.536 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,  -2 },     76.24699,   0.00005,     0.085 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },     76.28667,   0.00471,     0.066 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   1,   0 },     76.28673,   0.00024,     0.035 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -2,  -1 },    270.41998,   0.00003,     0.080 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,  -1 },    270.03909,   0.00026,     1.231 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    258.23708,   0.00002,     0.070 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   1 },    270.51615,   0.00036,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,  -3 },    278.50242,   0.00001,     0.078 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.16015,   0.00191,     0.071 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,   1 },     92.04030,   0.00004,     0.024 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,  -1 },    272.15664,   0.00011,     0.037 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,   1 },     92.20258,   0.00002,     0.018 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -2,   0 },    256.21662,   0.00011,     0.075 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -1,  -2 },    256.21758,   0.00003,     4.143 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -1,   0 },     76.40522,   0.00002,     0.038 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },     76.28777,   0.00003,     0.025 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,  -1,  -1 },    272.15680,   0.00002,     0.039 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,  -1 },    177.78097,   0.00001,     0.029 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.94430,   0.00003,     0.132 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     15.29401,   0.00015,     0.116 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -3,  -1 },      0.06343,   0.00001,     0.020 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -2,   1 },    179.65507,   0.00001,     0.107 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    182.72312,   0.00003,     0.103 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },    180.18800,   0.00037,     0.058 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.03918,   0.00016,     0.279 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,   1 },    180.07447,   0.00019,     0.033 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   2,  -1 },    179.97555,   0.00001,     0.059 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   2,   1 },    179.99339,   0.00001,     0.023 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },     10.73319,   0.00002,     0.349 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },    194.13618,   0.00001,     0.093 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },      0.01965,   0.00003,     0.370 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    190.60355,   0.00063,     0.062 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },      0.01965,   0.00003,     0.034 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },    190.59296,   0.00004,     0.034 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.03458,   0.00007,     0.087 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.02382,   0.00077,     0.552 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.02376,   0.00098,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,  -3 },    180.05026,   0.00003,     0.085 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    180.02696,   0.00416,     0.066 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   1 },      0.02554,   0.00031,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,  -1 },    180.02477,   0.00018,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,   1 },      0.02550,   0.00009,     0.018 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -2,   1 },    180.02090,   0.00003,     0.025 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -1,  -1 },    180.02408,   0.00008,     0.038 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.03227,   0.00002,     0.025 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   0,   1 },      0.02676,   0.00002,     0.015 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.27725,   0.00007,     0.048 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.28405,   0.00002,     0.029 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.28087,   0.00106,     0.131 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.27865,   0.00016,     0.177 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -2,   0 },    283.32154,   0.00002,     0.107 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.32437,   0.00003,     0.103 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    103.32226,   0.00282,     0.058 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },    103.32263,   0.00016,     0.033 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    264.36313,   0.00009,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,   1 },     92.71680,   0.00001,     0.023 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.32260,   0.00003,     0.035 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },    103.32260,   0.00002,     0.024 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },     27.68258,   0.00002,     0.031 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },     27.68276,   0.00002,     0.180 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    184.29594,   0.00004,     0.151 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },     25.62908,   0.00002,     0.130 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },     25.62250,   0.00002,     0.052 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },      3.30494,   0.00004,     0.055 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },     23.18808,   0.00002,     0.059 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },    212.32580,   0.00001,     0.023 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    126.73128,   0.00002,     0.178 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    308.71140,   0.00004,     0.052 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },     38.65913,   0.00001,     0.077 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   3 },    218.65913,   0.00001,     0.072 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   0,   1 },    231.60911,   0.00001,     0.078 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   0,   3 },     51.60911,   0.00001,     0.071 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   0,  -1 },     57.88299,   0.00001,     0.071 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   0,   1 },     57.88299,   0.00001,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    149.30429,   0.00006,     0.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },     67.06123,   0.00008,     0.091 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,  -1,   0 },     68.37961,   0.00005,     2.034 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,  -2 },    248.44641,   0.00001,     0.076 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    248.38034,   0.00102,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   1,   0 },    248.37961,   0.00006,     0.037 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -1,  -1 },    254.25107,   0.00003,     1.779 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -1,   1 },    254.25287,   0.00003,     0.038 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    253.34345,   0.00005,     0.079 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.37045,   0.00014,     0.100 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,  -1 },    168.55097,   0.00002,     0.041 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,   1 },    168.55083,   0.00002,     0.393 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0,  -1 },    171.32188,   0.00002,     0.091 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },    348.97796,   0.00006,     0.670 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,  -2 },    168.99258,   0.00001,     0.083 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    168.97991,   0.00111,     0.068 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },    168.97796,   0.00006,     0.036 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,  -1 },    347.21785,   0.00006,     2.284 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,   1 },    348.47811,   0.00010,     0.037 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -3 },    174.63991,   0.00001,     0.076 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    350.08404,   0.00083,     0.073 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,   1 },    168.95397,   0.00002,     0.025 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   1,  -1 },    350.05272,   0.00005,     0.037 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,  -1,   0 },    256.09995,   0.00002,     0.045 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    256.11106,   0.00058,     0.111 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   1,   0 },     76.12511,   0.00006,     0.234 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -1,   1 },     92.13155,   0.00001,     0.282 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    263.04725,   0.00001,     0.100 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    269.27250,   0.00004,     0.060 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   1,   1 },    271.23674,   0.00002,     0.033 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    254.69350,   0.00005,     0.401 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,  -2 },     76.33549,   0.00002,     0.090 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     76.14582,   0.00274,     0.064 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },     76.14542,   0.00014,     0.034 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -2,   1 },     92.33880,   0.00001,     0.067 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },    271.67521,   0.00015,     0.695 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -1,   1 },    271.80762,   0.00018,     0.035 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.35437,   0.00087,     0.068 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,   1 },     92.35518,   0.00005,     0.024 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,  -1 },    272.33801,   0.00005,     0.036 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,   1 },     92.30958,   0.00002,     0.018 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },     76.14626,   0.00003,     0.037 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,   0,   0 },     76.14627,   0.00002,     0.025 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,  -1 },    181.41379,   0.00002,     0.030 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    182.16820,   0.00002,     0.144 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.42175,   0.00002,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   3,  -1 },      0.14686,   0.00001,     0.167 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     15.00068,   0.00012,     0.125 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.18719,   0.00002,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,  -1 },    180.94726,   0.00003,     0.111 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,   1 },    180.09303,   0.00028,     0.056 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.06477,   0.00012,     0.237 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   1,   1 },    180.09748,   0.00010,     0.032 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   2,   1 },    180.10255,   0.00002,     0.023 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },     10.75720,   0.00003,     0.286 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    190.76073,   0.00048,     0.060 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   1,   0 },    190.75744,   0.00003,     0.033 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.11885,   0.00005,     0.092 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.03157,   0.00049,     0.410 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.02986,   0.00060,     0.034 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,  -3 },    180.02899,   0.00001,     0.090 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    180.03045,   0.00246,     0.064 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,   1 },    359.99792,   0.00027,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   1,  -1 },    180.03114,   0.00010,     0.035 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   1,   1 },      0.01084,   0.00007,     0.018 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -2,   1 },    179.97021,   0.00002,     0.025 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -1,  -1 },    180.02680,   0.00007,     0.037 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -1,   1 },      0.01955,   0.00002,     0.019 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.01811,   0.00004,     0.025 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.26290,   0.00005,     0.049 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.26710,   0.00074,     0.143 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.26685,   0.00013,     0.160 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.27651,   0.00002,     0.223 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.30148,   0.00003,     0.110 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    103.31672,   0.00201,     0.056 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   1,   0 },    103.31666,   0.00011,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    266.06182,   0.00007,     0.060 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,   1 },     93.50135,   0.00001,     0.023 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.31660,   0.00002,     0.034 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   0,  -1 },     26.83923,   0.00002,     0.031 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },     26.83865,   0.00002,     0.203 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    182.69243,   0.00002,     0.167 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    353.74568,   0.00002,     0.053 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },     22.65134,   0.00002,     0.057 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    309.01239,   0.00003,     0.051 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   1 },    160.56629,   0.00006,     0.081 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    131.57467,   0.00014,     0.075 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1,  -1 },    340.98285,   0.00004,     0.036 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1,   1 },    341.59624,   0.00002,     1.206 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,  -1 },    338.06603,   0.00002,     0.069 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,   1 },    338.37798,   0.00003,     0.080 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   1,  -1 },    339.66505,   0.00005,     0.860 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   1,   1 },    339.46026,   0.00006,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -1,  -1 },    340.94671,   0.00010,     0.341 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -1,   1 },    340.76367,   0.00012,     0.042 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },    340.47335,   0.00042,     0.097 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,   1 },    339.99223,   0.00011,     0.027 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   1,   1 },    340.28422,   0.00001,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   0,  -1 },    236.48131,   0.00003,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   0,   1 },    237.04331,   0.00008,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   1,  -1 },    237.28050,   0.00005,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   1,   1 },    237.69315,   0.00010,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   2,  -1 },    236.71455,   0.00005,     0.027 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   2,   1 },    234.92068,   0.00001,     0.105 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    237.53797,   0.00002,     0.015 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    236.75473,   0.00019,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   1 },    236.63748,   0.00040,     0.041 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,  -1 },    237.40216,   0.00122,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },    237.28901,   0.00626,     0.088 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   3 },    237.99295,   0.00002,     0.064 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,  -1 },    237.97758,   0.00171,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   1 },    238.02034,   0.00140,     0.514 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   2,  -1 },     52.06430,   0.00002,     0.086 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   2,   1 },    241.17882,   0.00010,     0.066 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   3,   1 },    238.63883,   0.00001,     0.035 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },     55.67498,   0.00003,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0,   1 },     55.67473,   0.00003,     0.945 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -3,  -1 },    233.34244,   0.00002,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,  -1 },    233.46495,   0.00022,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,   1 },    232.93762,   0.00010,     0.076 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,  -1 },    233.77003,   0.00187,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   1 },    233.09698,   0.00159,     5.850 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,  -1 },    237.86841,   0.00218,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   1 },    237.86437,   0.00218,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,  -1 },    233.09372,   0.00159,     6.151 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   1 },    233.76930,   0.00187,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,  -1 },    233.22607,   0.00010,     0.076 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   1 },    233.57451,   0.00022,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   3,   1 },    233.46832,   0.00002,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },     57.88331,   0.00003,     0.953 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   0,   1 },     57.88331,   0.00003,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -3,  -1 },    237.45661,   0.00001,     0.035 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -2,  -1 },    239.56601,   0.00010,     0.066 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -2,   1 },     58.84138,   0.00002,     0.086 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,  -1 },    237.96024,   0.00139,     0.516 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   1 },    237.92651,   0.00170,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,  -3 },    237.96285,   0.00002,     0.064 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    237.23099,   0.00623,     0.088 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   1 },    237.22771,   0.00121,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,  -1 },    236.48643,   0.00040,     0.041 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,   1 },    236.41795,   0.00019,     0.019 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   2,   1 },    237.38610,   0.00002,     0.015 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -2,  -1 },    237.38206,   0.00001,     0.105 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -2,   1 },    237.37904,   0.00005,     0.027 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -1,  -1 },    237.64783,   0.00010,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -1,   1 },    237.22620,   0.00005,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   0,  -1 },    236.93172,   0.00008,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   0,   1 },    236.33304,   0.00003,     0.016 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,  -1 },    104.23559,   0.00006,     0.027 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,   1 },    103.40046,   0.00018,     0.097 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,  -1 },     83.14471,   0.00004,     0.042 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,   1 },     74.09872,   0.00004,     0.340 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,  -1,   0 },    333.47607,   0.00002,     0.041 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    332.89054,   0.00011,     0.088 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },    333.47607,   0.00010,     0.529 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -1,  -1 },      3.33282,   0.00004,     0.039 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -1,   1 },     20.80825,   0.00002,     0.854 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   0,  -1 },     10.58326,   0.00003,     0.081 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   0,   1 },     16.11328,   0.00003,     0.069 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   1,  -1 },    350.69418,   0.00003,     1.194 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   1,   1 },    351.53524,   0.00005,     0.036 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -2,   0 },    333.47607,   0.00002,     0.076 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },    333.47607,   0.00060,     8.534 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    153.47650,   0.01107,     0.075 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   1,   0 },    153.47607,   0.00060,     0.038 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   2,   0 },    153.47607,   0.00004,     0.025 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },    327.22565,   0.00016,     0.081 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   0,   1 },    330.41620,   0.00001,     0.026 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   3,  -1,   0 },    153.47607,   0.00012,     0.040 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   3,   0,   0 },    153.47607,   0.00007,     0.026 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },     82.62981,   0.00001,     0.096 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,  -1,   0 },     68.20431,   0.00002,     0.895 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    248.20431,   0.00030,     0.070 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,   1,   0 },    248.20431,   0.00002,     0.036 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,  -1 },     75.54505,   0.00006,    15.770 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,   1 },     76.01579,   0.00004,     0.037 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    254.43704,   0.00035,     0.075 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   1,  -1 },    254.43704,   0.00002,     0.038 },
 { {   0,   8, -11,   0,   0,   3,   0,   0,   2,   0,  -3 },    358.61239,   0.00001,     0.074 },
 { {   0,   8, -11,   0,   0,   3,   0,   0,   2,   0,  -1 },    178.61239,   0.00001,     0.075 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.32546,   0.00009,     0.107 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,  -1,   0 },    348.86721,   0.00004,     0.472 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,  -2 },    168.95727,   0.00001,     0.087 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    168.86548,   0.00064,     0.065 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   1,   0 },    168.86722,   0.00004,     0.035 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -1,  -1 },    350.83742,   0.00005,     0.940 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -1,   1 },    350.83712,   0.00005,     0.036 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    351.09119,   0.00023,     0.070 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,   1 },    170.31095,   0.00002,     0.024 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,  -1,  -1 },    169.99155,   0.00006,     0.074 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.94060,   0.00038,     0.120 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   1,   0 },     75.93432,   0.00003,     0.204 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,  -1 },    270.84446,   0.00001,     0.106 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,   1 },    271.84259,   0.00002,     0.057 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },    255.96258,   0.00001,     0.321 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,  -2 },     75.83797,   0.00002,     0.095 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     75.97687,   0.00171,     0.061 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },     75.97653,   0.00008,     0.034 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.39262,   0.00010,     0.485 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.47603,   0.00012,     0.035 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.55013,   0.00052,     0.065 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,   1 },     93.74093,   0.00007,     0.024 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   1,  -1 },    272.66787,   0.00003,     0.035 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   1,   1 },     93.19257,   0.00001,     0.018 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   3,  -1,   0 },     75.97715,   0.00002,     0.036 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.61752,   0.00002,     0.030 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.19868,   0.00005,     0.158 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },     15.48670,   0.00008,     0.136 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.90989,   0.00002,     0.046 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.59142,   0.00003,     0.193 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    183.56497,   0.00003,     0.119 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    180.46464,   0.00023,     0.054 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   1,  -1 },    179.98470,   0.00011,     0.207 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   1,   1 },    180.10618,   0.00009,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   2,   1 },    179.60945,   0.00001,     0.022 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },     10.97691,   0.00002,     0.243 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    190.97692,   0.00038,     0.058 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },    190.97691,   0.00002,     0.033 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.32633,   0.00003,     0.098 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.04037,   0.00030,     0.326 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.01488,   0.00038,     0.033 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.93498,   0.00158,     0.061 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   1 },    359.99607,   0.00023,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   1,  -1 },    179.88975,   0.00006,     0.034 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   1,   1 },      0.00521,   0.00006,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -2,   1 },    179.84408,   0.00001,     0.024 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -1,  -1 },    179.95119,   0.00002,     0.036 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -1,   1 },      0.02404,   0.00001,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,   0,  -1 },      0.01034,   0.00003,     0.024 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.24678,   0.00004,     0.051 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.25148,   0.00052,     0.157 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.25801,   0.00011,     0.145 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.29061,   0.00005,     0.195 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.30921,   0.00002,     0.118 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    103.30819,   0.00145,     0.054 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   1,   0 },    103.30831,   0.00007,     0.032 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    267.40883,   0.00005,     0.058 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,   1 },     94.45970,   0.00001,     0.023 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   3,  -1,   0 },    103.30802,   0.00002,     0.033 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   0,  -1 },     26.48064,   0.00002,     0.032 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },     23.30102,   0.00001,     0.055 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    309.24747,   0.00003,     0.049 },
 { {   0,   9, -17,   0,   3,   0,   0,   0,  -2,   0,   1 },    146.35661,   0.00001,     0.074 },
 { {   0,   9, -17,   0,   3,   0,   0,   0,  -2,   0,   3 },    326.35661,   0.00001,     0.075 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   0,  -1 },     57.84658,   0.00001,     0.078 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   0,   1 },     57.84658,   0.00001,     0.071 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    229.85316,   0.00002,     0.077 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    333.10097,   0.00007,     0.071 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    249.64846,   0.00015,     0.067 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },     75.07382,   0.00007,     0.072 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.12220,   0.00005,     0.114 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,  -1,   0 },    348.63570,   0.00002,     0.365 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    168.61822,   0.00040,     0.063 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   1,   0 },    168.63636,   0.00002,     0.034 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -1,  -1 },    352.18150,   0.00003,     0.592 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -1,   1 },    352.18150,   0.00003,     0.035 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    352.01705,   0.00012,     0.067 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,   1 },    170.99337,   0.00002,     0.024 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.73972,   0.00025,     0.129 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   1,   0 },     75.75125,   0.00002,     0.181 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,  -1,   1 },    272.19907,   0.00002,     0.208 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   0 },     75.77742,   0.00111,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   1,   0 },     75.77788,   0.00007,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.47979,   0.00007,     0.372 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.45197,   0.00008,     0.034 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.18847,   0.00033,     0.063 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,   1 },     92.96851,   0.00006,     0.023 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   1,  -1 },    272.31106,   0.00002,     0.034 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   1,   1 },     92.71295,   0.00001,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.52985,   0.00002,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.27988,   0.00004,     0.176 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     15.27950,   0.00006,     0.148 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.30951,   0.00003,     0.048 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -1,   1 },    180.17277,   0.00004,     0.173 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.19568,   0.00002,     0.128 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   1 },    180.09327,   0.00017,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.21264,   0.00005,     0.183 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   1,   1 },    180.16658,   0.00007,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },     11.19231,   0.00002,     0.211 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    191.19231,   0.00029,     0.056 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },    191.19231,   0.00002,     0.032 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.00923,   0.00002,     0.105 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.00012,   0.00020,     0.271 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.99778,   0.00025,     0.033 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.99711,   0.00102,     0.059 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   1 },      0.00778,   0.00019,     0.023 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   1,  -1 },    179.99448,   0.00004,     0.033 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   1,   1 },      0.00395,   0.00006,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,   0,   1 },    180.00770,   0.00001,     0.015 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.23243,   0.00002,     0.053 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.23342,   0.00037,     0.174 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.24205,   0.00007,     0.133 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.28195,   0.00005,     0.174 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.29788,   0.00001,     0.127 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    103.29685,   0.00105,     0.053 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   1,   0 },    103.29699,   0.00005,     0.031 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    268.77345,   0.00004,     0.056 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,   1 },     93.33368,   0.00001,     0.022 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },     24.34790,   0.00001,     0.053 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    309.43337,   0.00002,     0.048 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   1,   0,   0 },     52.33272,   0.00003,     0.073 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    332.18217,   0.00003,     0.068 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },    157.21334,   0.00004,     0.074 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    250.30634,   0.00010,     0.064 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.03001,   0.00003,     0.069 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.09236,   0.00003,     0.123 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    168.76312,   0.00027,     0.060 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -1,  -1 },    353.45687,   0.00002,     0.432 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -1,   1 },    353.45801,   0.00002,     0.034 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },    352.91888,   0.00007,     0.064 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,   1 },    171.89652,   0.00002,     0.024 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.50719,   0.00017,     0.141 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   1,   0 },     75.51973,   0.00002,     0.163 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,  -1 },    272.22172,   0.00001,     0.047 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,   1 },    272.22172,   0.00001,     0.184 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,  -1,   0 },     75.55422,   0.00001,     0.229 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   0 },     75.55381,   0.00074,     0.057 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   1,   0 },     75.55397,   0.00005,     0.032 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.34461,   0.00007,     0.302 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.34529,   0.00005,     0.033 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.26173,   0.00022,     0.060 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,   1 },     92.53178,   0.00005,     0.023 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   1,   1 },     92.44705,   0.00001,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.21464,   0.00004,     0.031 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.21416,   0.00004,     0.198 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.12628,   0.00001,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.12628,   0.00001,     0.122 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },     15.35908,   0.00003,     0.164 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.99159,   0.00005,     0.049 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.97965,   0.00002,     0.156 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.59884,   0.00001,     0.140 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,   1 },    180.09341,   0.00013,     0.051 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   1,   1 },    180.03471,   0.00005,     0.030 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    191.38023,   0.00023,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -2,  -1 },    180.20518,   0.00001,     0.112 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.99102,   0.00015,     0.232 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.99163,   0.00018,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.98666,   0.00066,     0.057 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   1 },      0.01613,   0.00015,     0.023 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   1,  -1 },    179.98740,   0.00003,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   1,   1 },    359.99856,   0.00004,     0.017 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,  -1,   1 },    359.99777,   0.00002,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   0,   1 },    179.99097,   0.00002,     0.014 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.21313,   0.00027,     0.196 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.22383,   0.00006,     0.123 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.57905,   0.00005,     0.157 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    103.28294,   0.00076,     0.051 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   1,   0 },    103.28273,   0.00005,     0.030 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },    269.50340,   0.00003,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,   1 },     94.20378,   0.00001,     0.022 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    309.58186,   0.00002,     0.046 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,  -1,   1 },    256.24244,   0.00002,     0.081 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,  -2,  -1 },    256.24244,   0.00001,     0.036 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },     76.24244,   0.00002,     0.039 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -2,   1 },     76.24244,   0.00002,     0.042 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },     76.24244,   0.00004,     0.097 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },     76.24244,   0.00003,     0.027 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,  -1,   1 },    333.30551,   0.00002,     0.028 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   0,  -1 },    333.30551,   0.00002,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   0,   1 },    333.30551,   0.00002,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -3,  -1 },    333.30551,   0.00001,     0.015 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    333.30551,   0.00008,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -2,   1 },    333.30551,   0.00009,     0.041 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    333.30551,   0.00048,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   1 },    333.39853,   0.00086,     0.088 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },    333.30551,   0.00054,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },    153.30551,   0.00002,     0.066 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    153.30551,   0.00001,     0.039 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -4,  -1 },    333.30551,   0.00002,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,  -1 },    333.30551,   0.00025,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,   1 },    333.30551,   0.00004,     0.076 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,  -1 },    333.25817,   0.00274,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   1 },    341.15840,   0.00002,     6.016 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,  -3 },    153.30551,   0.00003,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    333.18883,   0.02492,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },    333.23168,   0.02491,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   3 },    153.30551,   0.00003,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    341.15840,   0.00002,     5.978 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    333.25817,   0.00274,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    333.30551,   0.00004,     0.076 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   1 },    333.30551,   0.00025,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   2,   1 },    333.30551,   0.00002,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    153.30551,   0.00001,     0.039 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -3,  -1 },    153.30551,   0.00002,     0.066 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   1 },    333.30551,   0.00054,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },    333.39853,   0.00086,     0.088 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   1 },    333.30551,   0.00048,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    333.30551,   0.00009,     0.041 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   0,   1 },    333.30551,   0.00008,     0.019 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   1,   1 },    333.30551,   0.00001,     0.015 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -2,  -1 },    333.30551,   0.00002,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -2,   1 },    333.30551,   0.00002,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -1,  -1 },    333.30551,   0.00002,     0.028 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    230.36858,   0.00003,     0.027 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -1,   1 },    230.36858,   0.00004,     0.097 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },    230.36858,   0.00002,     0.042 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,  -2,  -1 },    230.36858,   0.00002,     0.039 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },     50.36858,   0.00001,     0.036 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,  -1,  -1 },     50.36858,   0.00002,     0.081 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    334.42979,   0.00002,     0.066 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    250.89101,   0.00007,     0.062 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.77296,   0.00002,     0.066 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.18552,   0.00003,     0.134 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    168.68850,   0.00019,     0.058 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    353.66021,   0.00005,     0.062 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,   0,   1 },    172.90028,   0.00001,     0.023 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    255.24458,   0.00012,     0.154 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -1,  -1 },    275.48597,   0.00001,     0.048 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -1,   1 },    275.48597,   0.00001,     0.165 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,  -1 },    275.48597,   0.00001,     0.175 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,   1 },    275.48597,   0.00001,     0.031 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },     75.29733,   0.00001,     0.200 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     75.29738,   0.00049,     0.055 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   1,   0 },     75.29756,   0.00003,     0.032 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.49091,   0.00005,     0.254 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.55449,   0.00004,     0.032 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.01473,   0.00016,     0.058 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,   1 },     93.39380,   0.00004,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.18711,   0.00004,     0.032 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.18659,   0.00003,     0.226 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.07998,   0.00001,     0.055 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.07998,   0.00001,     0.113 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },     14.82420,   0.00003,     0.182 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.99217,   0.00004,     0.051 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.96811,   0.00002,     0.142 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   0,   1 },    180.14879,   0.00008,     0.049 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   1,   1 },    180.08343,   0.00002,     0.030 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    191.54425,   0.00018,     0.052 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.97139,   0.00011,     0.202 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.97478,   0.00012,     0.031 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.95856,   0.00045,     0.055 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   1 },      0.04756,   0.00010,     0.022 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   1,   1 },    359.99814,   0.00004,     0.017 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,  -1,   1 },    359.98935,   0.00001,     0.018 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.18758,   0.00019,     0.223 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.20353,   0.00005,     0.114 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.28701,   0.00004,     0.143 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    103.23432,   0.00055,     0.049 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   1,   0 },    103.24011,   0.00004,     0.030 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    270.10953,   0.00002,     0.052 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    251.41110,   0.00005,     0.059 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    348.13089,   0.00002,     0.146 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    168.60573,   0.00014,     0.056 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },    354.32071,   0.00003,     0.059 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   0,   1 },    173.71682,   0.00001,     0.023 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    254.93901,   0.00007,     0.171 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,  -1,   0 },     75.00957,   0.00001,     0.178 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,   0 },     75.00980,   0.00034,     0.053 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   1,   0 },     75.01001,   0.00002,     0.031 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.89622,   0.00003,     0.219 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.89264,   0.00003,     0.032 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.18503,   0.00009,     0.056 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,   1 },     92.55731,   0.00005,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.20833,   0.00003,     0.033 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.20774,   0.00003,     0.263 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.11001,   0.00001,     0.057 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.11001,   0.00001,     0.106 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },     14.40632,   0.00002,     0.206 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.92365,   0.00002,     0.052 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.91449,   0.00002,     0.130 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   0,  -1 },    180.25970,   0.00003,     0.169 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   0,   1 },    180.12790,   0.00005,     0.048 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    191.68764,   0.00014,     0.050 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.94940,   0.00006,     0.180 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.95320,   0.00008,     0.031 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.95331,   0.00029,     0.053 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   1 },    359.96021,   0.00012,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   1,   1 },    359.95690,   0.00003,     0.017 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,  -1,   1 },    359.95074,   0.00001,     0.017 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.16090,   0.00013,     0.259 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.18120,   0.00004,     0.106 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.24466,   0.00004,     0.131 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    103.24650,   0.00040,     0.048 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   1,   0 },    103.24649,   0.00003,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    271.24007,   0.00002,     0.050 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   1,  -1 },    272.38258,   0.00002,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   1,   1 },    272.38258,   0.00002,     0.074 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   1,  -1 },    228.50839,   0.00003,     0.026 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   1,   1 },    228.50839,   0.00005,     0.089 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   2,  -1 },    228.50839,   0.00003,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    228.50839,   0.00002,     0.025 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,  -1 },    228.37325,   0.00014,     0.038 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },     49.08404,   0.00003,     5.228 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,  -1 },    228.50839,   0.00156,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   1 },    228.50839,   0.00156,     0.074 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   1 },    228.56475,   0.00017,     0.037 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   3,   1 },    228.50839,   0.00002,     0.025 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    228.50839,   0.00003,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   1,  -1 },    228.50839,   0.00005,     0.088 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   1,   1 },    228.50839,   0.00003,     0.026 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   1,  -2 },    149.22796,   0.00004,     0.074 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   1,   0 },    149.22796,   0.00004,     0.075 },
 { {   0,  12,  -6,   0,   1,   0,   0,   0,   0,   1,  -3 },    278.68671,   0.00001,     0.075 },
 { {   0,  12,  -6,   0,   1,   0,   0,   0,   0,   1,  -1 },    278.68671,   0.00001,     0.074 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    251.87408,   0.00003,     0.057 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    168.51430,   0.00010,     0.054 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   0,  -1 },    354.93827,   0.00002,     0.057 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   0,   1 },    174.34383,   0.00001,     0.023 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    254.61162,   0.00005,     0.191 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   0,   1 },     92.27626,   0.00002,     0.049 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,   0 },     74.68946,   0.00023,     0.051 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.37926,   0.00002,     0.193 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.37832,   0.00002,     0.031 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },    271.77234,   0.00006,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,   1 },     93.12857,   0.00004,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.16815,   0.00003,     0.033 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.16751,   0.00003,     0.314 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.02448,   0.00001,     0.060 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.02448,   0.00001,     0.099 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },     14.33592,   0.00002,     0.236 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.90872,   0.00003,     0.054 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.90861,   0.00003,     0.121 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   0,  -1 },    180.04894,   0.00003,     0.189 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   0,   1 },    180.46431,   0.00004,     0.046 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    191.81325,   0.00010,     0.049 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -1,  -1 },    180.04763,   0.00006,     0.162 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -1,   1 },    180.05377,   0.00006,     0.030 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.93635,   0.00020,     0.051 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,   1 },    359.92508,   0.00009,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   1,  -1 },    359.92737,   0.00002,     0.031 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   1,   1 },    359.92776,   0.00002,     0.017 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.12732,   0.00009,     0.309 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.15690,   0.00004,     0.100 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.22344,   0.00003,     0.121 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    103.22507,   0.00029,     0.047 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   1,   0 },    103.22506,   0.00002,     0.029 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   1,  -1 },     25.96323,   0.00001,     0.068 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   1,   1 },     25.96323,   0.00001,     0.083 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -1,   1 },    102.34519,   0.00002,     0.074 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -1,   3 },    102.34519,   0.00002,     0.075 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   1,  -1 },    280.36378,   0.00001,     0.073 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   1,   1 },    280.36378,   0.00001,     0.076 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    252.28639,   0.00003,     0.055 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    167.65030,   0.00008,     0.052 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,   0,  -1 },    355.49114,   0.00002,     0.055 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    254.25330,   0.00002,     0.217 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,   0 },     74.33451,   0.00016,     0.050 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.15269,   0.00001,     0.172 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,   1 },    272.15522,   0.00001,     0.031 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,  -1 },    271.58924,   0.00006,     0.052 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,   1 },     93.62177,   0.00003,     0.022 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,  -1 },    179.95542,   0.00001,     0.034 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,   1 },    179.95542,   0.00001,     0.391 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.08830,   0.00001,     0.062 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.08830,   0.00001,     0.093 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   2,  -1 },    359.95542,   0.00001,     0.346 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   2,   1 },    359.95542,   0.00001,     0.042 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.88768,   0.00003,     0.056 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.88759,   0.00003,     0.112 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   0,  -1 },    180.86021,   0.00004,     0.214 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   0,   1 },    181.12860,   0.00003,     0.045 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    191.92346,   0.00008,     0.047 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.90402,   0.00005,     0.147 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.90915,   0.00003,     0.030 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.86612,   0.00014,     0.050 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   0,   1 },    359.99773,   0.00007,     0.021 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   1,   1 },    359.95546,   0.00001,     0.017 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.09809,   0.00006,     0.383 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.13069,   0.00004,     0.094 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.20022,   0.00005,     0.113 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    103.20122,   0.00021,     0.045 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.30988,   0.00001,     0.047 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   0,   1 },    272.30988,   0.00001,     0.021 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   3,  -2 },    352.33008,   0.00001,     0.074 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   3,   0 },    352.33008,   0.00001,     0.075 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   1,  -1 },     26.72280,   0.00003,     0.071 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   1,   1 },     26.72280,   0.00003,     0.079 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   1,  -1 },    281.63014,   0.00001,     0.076 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   1,   1 },    281.63014,   0.00001,     0.073 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   0,  -3 },    320.24046,   0.00001,     0.037 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   1,  -3 },    320.24046,   0.00009,     0.074 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   1,  -1 },    320.24046,   0.00009,     0.075 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   2,  -1 },    320.24046,   0.00001,     0.037 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   1,   0,   0 },    252.65354,   0.00002,     0.053 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    168.30224,   0.00006,     0.051 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,  -3 },    355.78633,   0.00001,     0.123 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,  -1 },    355.78159,   0.00002,     0.053 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,   1 },    355.77726,   0.00001,     0.022 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,   0 },    253.86672,   0.00002,     0.251 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },     99.56760,   0.00001,     0.198 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,   1 },     99.56760,   0.00001,     0.046 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   1,   0,   0 },     73.94293,   0.00011,     0.048 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -1,  -1 },     92.51959,   0.00001,     0.155 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   0,  -1 },    271.72781,   0.00004,     0.051 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   0,   1 },     93.27726,   0.00003,     0.021 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   0,   1 },    359.60386,   0.00001,     0.037 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   1,   1 },    359.60386,   0.00009,     0.075 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   1,   3 },    359.60386,   0.00009,     0.074 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   2,   3 },    359.60386,   0.00001,     0.037 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,  -1 },    179.93935,   0.00001,     0.035 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   1 },    179.93935,   0.00001,     0.518 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.07343,   0.00001,     0.064 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.07343,   0.00001,     0.088 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   2,  -1 },    359.93935,   0.00001,     0.441 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   2,   1 },    359.93935,   0.00001,     0.041 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.02806,   0.00001,     0.058 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -1,   1 },    180.02806,   0.00001,     0.105 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.29580,   0.00003,     0.248 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   0,   1 },    181.67888,   0.00002,     0.044 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   1,  -1 },    359.88005,   0.00001,     0.109 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   1,   1 },    359.88005,   0.00001,     0.028 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    192.02028,   0.00006,     0.046 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.87940,   0.00003,     0.134 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.88250,   0.00002,     0.029 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.83993,   0.00010,     0.048 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   0,   1 },    359.96498,   0.00005,     0.021 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   1,   1 },    359.92529,   0.00001,     0.016 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   1,   0 },    283.38829,   0.00003,     0.074 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   1,   2 },    283.38829,   0.00003,     0.075 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.06720,   0.00005,     0.504 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.10261,   0.00003,     0.089 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.17361,   0.00003,     0.105 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    103.17499,   0.00015,     0.044 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   1,  -1 },     26.32315,   0.00002,     0.026 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   1,   1 },     26.32315,   0.00004,     0.088 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   2,  -1 },     26.36029,   0.00002,     0.040 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     26.32315,   0.00001,     0.025 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,  -1 },     26.26398,   0.00023,     0.037 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },     26.19368,   0.00010,    14.735 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,  -1 },     26.32315,   0.00112,     0.074 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   1 },     26.32315,   0.00112,     0.075 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   2,   1 },     26.28499,   0.00012,     0.038 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   3,   1 },     26.32315,   0.00001,     0.025 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   0,  -1 },    206.19364,   0.00002,     0.490 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   1,  -1 },     26.32315,   0.00004,     0.089 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   1,   1 },     26.32315,   0.00002,     0.026 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   0,  -1 },     99.70101,   0.00001,     0.026 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   0,   1 },     99.70101,   0.00002,     0.088 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   1,  -1 },     99.70101,   0.00001,     0.040 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     99.70101,   0.00007,     0.038 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },     99.70101,   0.00066,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,   1 },     99.70101,   0.00066,     0.074 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,  -2 },    309.25004,   0.00003,     0.074 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },    309.29455,   0.00005,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,   1 },     99.70101,   0.00007,     0.037 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,  -1,   1 },     99.70101,   0.00001,     0.040 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },     99.70101,   0.00002,     0.088 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   0,   1 },     99.70101,   0.00001,     0.026 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -1,  -1 },    231.71999,   0.00004,     0.075 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -1,   1 },    231.71999,   0.00004,     0.074 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -2,  -1 },    207.17736,   0.00003,     0.037 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -1,  -1 },    207.17736,   0.00027,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -1,   1 },    207.17736,   0.00027,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,   0,   1 },    207.17736,   0.00003,     0.037 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   1,   0,   0 },    168.18285,   0.00005,     0.049 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   1,   0,   0 },     73.51241,   0.00008,     0.047 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.94036,   0.00003,     0.049 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   0,   1 },     91.33994,   0.00002,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,  -1 },    179.91456,   0.00001,     0.036 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,   1 },    179.91456,   0.00001,     0.765 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   1,  -1 },    179.86126,   0.00002,     0.067 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   1,   1 },    179.86126,   0.00002,     0.084 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,  -1 },    359.91456,   0.00001,     0.610 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,   1 },    359.91456,   0.00001,     0.040 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.05313,   0.00001,     0.060 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -1,   1 },    180.05313,   0.00001,     0.098 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },    179.71250,   0.00003,     0.293 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   0,   1 },    179.69481,   0.00002,     0.043 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    192.10537,   0.00006,     0.045 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.84867,   0.00002,     0.124 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.85350,   0.00001,     0.029 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.85033,   0.00007,     0.047 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   0,   1 },    359.85880,   0.00004,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   1,   1 },    179.84226,   0.00001,     0.016 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    283.04963,   0.00002,     0.735 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.07275,   0.00004,     0.084 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.14514,   0.00002,     0.099 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    103.14643,   0.00011,     0.043 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   1,  -1 },     26.56171,   0.00003,     0.078 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   1,   1 },     26.56171,   0.00003,     0.072 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    206.35562,   0.00001,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,   1 },    206.35562,   0.00001,     0.077 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   1,  -1 },     47.78167,   0.00004,     0.075 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   1,   1 },     47.78167,   0.00004,     0.074 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   1,   0,   0 },    168.05332,   0.00003,     0.048 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   1,   0,   0 },     71.75214,   0.00005,     0.045 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.34755,   0.00002,     0.048 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   0,   1 },     92.22718,   0.00002,     0.021 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -3,   2,  -1 },      0.24149,   0.00001,     0.075 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -3,   2,   1 },      0.24149,   0.00001,     0.074 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,  -1 },    179.93243,   0.00003,     0.036 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,   1 },    179.93239,   0.00003,     1.468 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.04474,   0.00004,     0.070 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.04474,   0.00004,     0.080 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.92974,   0.00001,     0.062 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.92974,   0.00001,     0.093 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.24264,   0.00002,     0.359 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    192.18016,   0.00005,     0.043 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.82044,   0.00003,     0.046 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   0,   1 },    359.82337,   0.00004,     0.020 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,  -1 },     96.97758,   0.00006,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,   1 },     96.97758,   0.00006,     0.074 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,  -2 },    283.04847,   0.00003,     0.074 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    283.04847,   0.00003,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },    283.01642,   0.00002,     0.071 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },    103.04139,   0.00006,     0.080 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.11437,   0.00002,     0.093 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    103.11559,   0.00008,     0.042 },
 { {   0,  17, -15,   0,  -1,   0,   0,   0,  -1,   1,  -2 },    104.27292,   0.00002,     0.075 },
 { {   0,  17, -15,   0,  -1,   0,   0,   0,  -1,   1,   0 },    284.27292,   0.00002,     0.074 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    206.35980,   0.00002,     0.076 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,   1 },    206.35980,   0.00002,     0.073 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,  -2 },    320.29031,   0.00001,     0.075 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,   0 },    320.29031,   0.00001,     0.074 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -1,  -1 },     26.70587,   0.00003,     0.071 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -1,   1 },     26.70587,   0.00003,     0.078 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   0,  -1 },     27.90536,   0.00002,     1.268 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   0,   1 },     27.90677,   0.00002,     0.038 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   1,   0,   0 },    167.91334,   0.00003,     0.046 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   0,   1 },    270.03812,   0.00007,     0.074 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   0,   3 },    270.03812,   0.00007,     0.075 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   1,   0,   0 },     72.52565,   0.00005,     0.044 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.34085,   0.00002,     0.046 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   0,   1 },     92.22070,   0.00001,     0.021 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   1,   1 },    179.88810,   0.00001,     0.087 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,  -1 },    177.31449,   0.00013,     0.037 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    195.18854,   0.00001,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   1 },    180.15638,   0.00009,    17.998 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   2 },    195.18854,   0.00001,     0.075 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,  -1 },    179.88810,   0.00042,     0.073 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   1 },    179.88810,   0.00042,     0.076 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   2,   1 },    179.70962,   0.00005,     0.038 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.81988,   0.00001,     0.065 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.81988,   0.00001,     0.088 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.25086,   0.00002,     0.462 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   0,   1 },    193.29611,   0.00003,     0.041 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   1,  -1 },    179.88810,   0.00001,     0.090 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   1,   0,   0 },    192.24580,   0.00004,     0.042 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -2,   1 },    231.71713,   0.00002,     0.044 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -1,  -1 },    251.25957,   0.00003,     0.107 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -1,   1 },    250.02667,   0.00003,     0.028 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.78513,   0.00004,     0.044 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   0,   1 },    347.08871,   0.00003,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,  -1,   1 },    309.47954,   0.00001,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,   0,   1 },    309.47954,   0.00001,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    309.47954,   0.00003,     0.019 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -2,   1 },    309.47954,   0.00006,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    309.47954,   0.00013,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,   1 },    309.47954,   0.00042,     0.081 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   0,  -1 },    309.47954,   0.00009,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,  -1 },    129.47954,   0.00001,     0.025 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   0 },    283.01956,   0.00003,     0.038 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    129.47954,   0.00006,     0.037 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   0 },    282.82685,   0.00015,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   2 },    282.86574,   0.00024,     0.074 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,  -2 },    103.00785,   0.00002,     0.037 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    286.16330,   0.00002,     0.074 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },    103.00785,   0.00002,     9.147 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   1 },    286.16330,   0.00002,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   2 },    283.01956,   0.00003,     0.037 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,  -2 },    102.94937,   0.00007,     0.073 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,   0 },    102.98623,   0.00018,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -3,  -1 },    309.47954,   0.00007,     0.024 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -3,   1 },    309.47954,   0.00001,     0.071 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,  -1 },    309.47954,   0.00037,     0.036 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -1,  -1 },    309.47954,   0.00022,     0.069 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -1,   1 },    129.47954,   0.00017,     0.080 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   1 },    129.47954,   0.00047,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   1,  -1 },    129.47954,   0.00001,     0.083 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   1,   1 },    129.47954,   0.00008,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   2,   1 },    129.47954,   0.00001,     0.019 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.08136,   0.00002,     0.088 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    103.08249,   0.00006,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -3,  -1 },    309.47954,   0.00002,     0.062 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   1 },    129.47921,   0.00061,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,  -1 },    129.47935,   0.00103,     0.097 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   1 },    129.47954,   0.00082,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,  -1 },    129.47954,   0.00012,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   1 },    129.47954,   0.00016,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   1,  -1 },    129.47954,   0.00001,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   1,   1 },    129.47954,   0.00002,     0.016 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -3,   1 },    129.47954,   0.00002,     0.028 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -2,  -1 },    129.47954,   0.00004,     0.046 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -2,   1 },    129.47954,   0.00005,     0.021 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -1,  -1 },    129.47954,   0.00004,     0.029 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -1,   1 },    129.47954,   0.00003,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   0,   1 },     26.54261,   0.00001,     0.021 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   1,  -1 },     26.54261,   0.00001,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -2,  -1 },     26.54261,   0.00004,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -2,   1 },     26.54261,   0.00007,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,  -1 },     26.54261,   0.00021,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   1 },     26.54261,   0.00038,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,  -1 },     26.54261,   0.00041,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   1 },     26.54261,   0.00045,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,  -1 },     26.54261,   0.00025,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,   1 },     26.54261,   0.00002,     0.105 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,  -1,   1 },    206.54261,   0.00002,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   0,  -1 },    206.54261,   0.00002,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   0,   1 },    206.54261,   0.00001,     0.096 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -4,  -1 },     26.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -4,   1 },     26.54261,   0.00002,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,  -1 },     26.54261,   0.00026,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   1 },     26.54261,   0.00022,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,  -1 },     26.54268,   0.00210,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   1 },     26.54281,   0.00233,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,  -3 },    206.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     26.54252,   0.01216,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   1 },     26.54240,   0.02181,     0.088 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   3 },    206.54261,   0.00008,     0.064 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -3 },    206.54261,   0.00003,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -1 },     26.54132,   0.01408,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   1 },     26.49709,   0.00038,     0.514 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   3 },    206.54261,   0.00002,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,  -1 },    206.54261,   0.00006,     0.086 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   1 },    206.54895,   0.00048,     0.066 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   2,   1 },    206.54261,   0.00010,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   3,   1 },    206.54261,   0.00001,     0.024 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -2,  -1 },    206.54261,   0.00007,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -2,   1 },    206.54261,   0.00002,     0.082 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    206.54261,   0.00037,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },     26.54261,   0.00001,     0.070 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -5,  -1 },     26.54261,   0.00005,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,  -1 },     26.54309,   0.00055,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,   1 },     26.54261,   0.00011,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,  -3 },    206.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,  -1 },     26.54310,   0.00642,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   1 },     26.54367,   0.00113,     0.076 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -3 },    206.54261,   0.00014,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -1 },     26.54302,   0.06965,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   1 },     26.56543,   0.00068,     5.868 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   3 },    206.54261,   0.00019,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -3 },    206.54261,   0.00065,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -1 },     26.54272,   0.63014,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   1 },     26.54287,   0.63037,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   3 },    206.54261,   0.00065,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -3 },    206.54261,   0.00019,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },    206.48133,   0.00149,     6.132 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   1 },     26.54416,   0.06747,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   3 },    206.54261,   0.00014,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,  -1 },     26.54867,   0.00099,     0.076 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   1 },     26.54463,   0.00616,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   3 },    206.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,  -1 },     26.54261,   0.00011,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   1 },     26.54482,   0.00053,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   3,   1 },     26.54261,   0.00005,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -2,  -1 },     26.54261,   0.00001,     0.070 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   1 },    206.54261,   0.00037,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    206.54261,   0.00002,     0.082 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   0,   1 },    206.54261,   0.00007,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -5,  -1 },    206.54261,   0.00001,     0.024 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -4,  -1 },    206.54261,   0.00010,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,  -1 },    206.54387,   0.00055,     0.066 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,   1 },    206.54261,   0.00006,     0.086 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,  -3 },    206.54261,   0.00002,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,  -1 },     26.31826,   0.00002,     0.516 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   1 },     26.54238,   0.01365,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   3 },    206.54261,   0.00003,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,  -3 },    206.54261,   0.00008,     0.064 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,  -1 },     26.54244,   0.02178,     0.088 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   1 },     26.54282,   0.01205,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   3 },    206.54261,   0.00002,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,  -1 },     26.54398,   0.00225,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   1 },     26.54341,   0.00206,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,  -1 },     26.54261,   0.00022,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   1 },     26.54261,   0.00026,     0.015 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   2,  -1 },     26.54261,   0.00002,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   2,   1 },     26.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -2,  -1 },    206.54261,   0.00001,     0.096 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -2,   1 },    206.54261,   0.00002,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -1,  -1 },    206.54261,   0.00002,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,  -1 },     26.54261,   0.00002,     0.105 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   1 },     26.54261,   0.00025,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,  -1 },     26.54261,   0.00045,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   1 },     26.54261,   0.00041,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,  -1 },     26.54261,   0.00038,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   1 },     26.54261,   0.00021,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   0,  -1 },     26.54261,   0.00007,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   0,   1 },     26.54261,   0.00004,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -3,   1 },     26.54261,   0.00001,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -2,  -1 },     26.54261,   0.00001,     0.021 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -1,  -1 },    283.60568,   0.00003,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -1,   1 },    283.60568,   0.00004,     0.029 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   0,  -1 },    283.60568,   0.00005,     0.021 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   0,   1 },    283.60568,   0.00004,     0.046 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   1,  -1 },    283.60568,   0.00002,     0.028 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -3,  -1 },    283.60568,   0.00002,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -3,   1 },    283.60568,   0.00001,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    283.60568,   0.00016,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   1 },    283.60568,   0.00012,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    283.60568,   0.00082,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   1 },    283.60549,   0.00103,     0.097 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,  -1 },    283.60430,   0.00063,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   1,   1 },    103.60568,   0.00002,     0.062 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -4,  -1 },    283.60568,   0.00001,     0.019 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -3,  -1 },    283.60568,   0.00008,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -3,   1 },    283.60568,   0.00001,     0.083 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,  -1 },    283.60568,   0.00047,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -1,  -1 },    283.60568,   0.00017,     0.081 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -1,   1 },    103.60568,   0.00022,     0.069 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   1 },    103.60568,   0.00037,     0.036 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   1,  -1 },    103.60568,   0.00001,     0.071 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   1,   1 },    103.60568,   0.00007,     0.024 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -2,  -2 },    310.09123,   0.00002,     0.037 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,  -2 },    310.05153,   0.00015,     0.074 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,   0 },    310.11375,   0.00019,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,   1 },    283.60568,   0.00006,     0.037 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    310.09123,   0.00002,     0.038 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,   1 },    283.60568,   0.00001,     0.025 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -2,   1 },    103.60568,   0.00009,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,  -1 },    103.60568,   0.00042,     0.081 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,   1 },    103.60568,   0.00013,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,  -1 },    103.60568,   0.00006,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,   1 },    103.60568,   0.00003,     0.019 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -2,  -1 },    103.60568,   0.00001,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -1,  -1 },    103.60568,   0.00001,     0.027 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    180.66875,   0.00004,     0.028 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -1,   1 },    180.66875,   0.00004,     0.107 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,   0,  -1 },    180.66875,   0.00002,     0.044 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -2,   0 },    313.46714,   0.00002,     0.037 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -1,   0 },    313.46714,   0.00017,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -1,   2 },    313.46714,   0.00017,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,   0,   2 },    313.46714,   0.00002,     0.037 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   1,   0,   0 },    167.76260,   0.00002,     0.045 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   1,   1 },     14.98714,   0.00001,     0.088 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   0,  -1 },     14.98714,   0.00004,     0.037 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,  -1 },     14.98714,   0.00035,     0.074 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,   1 },     14.98714,   0.00035,     0.075 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   2,   1 },     14.98714,   0.00004,     0.037 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   1,  -1 },     14.98714,   0.00001,     0.088 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   1,  -1 },    270.73573,   0.00001,     0.071 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   1,   1 },    270.73573,   0.00001,     0.078 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   1,   0,   0 },     71.96169,   0.00004,     0.043 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.27742,   0.00001,     0.045 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,   1 },     92.27748,   0.00001,     0.020 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,  -1 },    359.53465,   0.00003,     0.038 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.53424,   0.00003,     1.755 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   1,  -1 },    180.12040,   0.00008,     0.077 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   1,   1 },    180.12040,   0.00008,     0.072 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },      0.07775,   0.00001,     0.072 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,   1 },      0.07775,   0.00001,     0.077 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.94690,   0.00002,     0.068 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.94690,   0.00002,     0.083 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.19484,   0.00002,     0.650 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,   1 },    181.19514,   0.00002,     0.040 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   1,   0,   0 },    192.30331,   0.00003,     0.041 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.75335,   0.00003,     0.043 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   0,   1 },    359.75344,   0.00002,     0.020 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   1,  -2 },    102.97251,   0.00003,     0.076 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   1,   0 },    282.97355,   0.00004,     0.073 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.04613,   0.00003,     0.083 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,   0,   0 },    103.04720,   0.00004,     0.040 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -1,  -1 },     26.46717,   0.00002,     0.078 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -1,   1 },     26.46717,   0.00002,     0.071 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   0,  -1 },    271.80227,   0.00002,     0.019 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   0,   1 },    271.80227,   0.00002,     0.041 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   1,  -1 },    271.80227,   0.00009,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   1,   1 },    271.80227,   0.00016,     0.089 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   2,  -1 },    271.65567,   0.00010,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   2,   1 },     91.76362,   0.00001,     0.508 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    271.80227,   0.00005,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,  -1 },    272.00036,   0.00040,     0.038 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   1 },     91.12551,   0.00012,     5.116 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,  -1 },    271.81875,   0.00479,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   1 },    271.81875,   0.00479,     0.074 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,  -1 },    279.24057,   0.00001,     7.245 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   1 },    271.94354,   0.00053,     0.037 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   3,   1 },    271.80227,   0.00005,     0.025 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   0,  -1 },    271.76362,   0.00001,     0.523 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   0,   1 },    271.67718,   0.00013,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   1,  -1 },    271.80227,   0.00016,     0.088 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   1,   1 },    271.80227,   0.00009,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   2,  -1 },    271.80227,   0.00002,     0.041 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   2,   1 },    271.80227,   0.00002,     0.019 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   1,   0,   0 },     71.34646,   0.00003,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,  -2 },    194.02829,   0.00004,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,   0 },    194.02829,   0.00006,     0.088 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   2,  -2 },    194.02829,   0.00004,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,  -2 },    194.02829,   0.00002,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,  -1 },      0.00384,   0.00001,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -2 },    193.88162,   0.00027,     0.037 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },      0.00384,   0.00012,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,   0 },    193.28785,   0.00004,     7.056 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,   1 },      0.00384,   0.00012,     0.074 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -2 },    194.02937,   0.00192,     0.074 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   0 },    193.93985,   0.00222,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   1 },      0.00384,   0.00001,     0.037 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,   0 },    193.98906,   0.00022,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   3,   0 },    194.02829,   0.00002,     0.025 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -1,  -1 },    180.03501,   0.00005,     0.071 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -1,   1 },    180.03501,   0.00005,     0.079 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,  -1 },    181.06591,   0.00003,     1.096 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,   1 },    181.06621,   0.00003,     0.039 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   0,   0 },    193.45771,   0.00006,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   1,  -2 },    194.02829,   0.00006,     0.089 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   1,   0 },    194.02829,   0.00004,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.71573,   0.00002,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   0,   1 },    359.71582,   0.00002,     0.020 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,  -1,  -1 },    116.74397,   0.00002,     0.075 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,  -1,   1 },    116.74397,   0.00002,     0.074 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,  -2 },    282.98937,   0.00009,     0.075 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    282.98195,   0.00004,     0.074 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,  -1,   0 },    103.00875,   0.00004,     0.079 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,   0,   0 },    103.00975,   0.00004,     0.039 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   0,  -3 },    206.45318,   0.00003,     0.075 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },    206.45318,   0.00003,     0.074 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,  -1,  -1 },    100.59376,   0.00002,     0.075 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,  -1,   1 },    100.59376,   0.00002,     0.074 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   0,  -2 },    354.13759,   0.00002,     0.037 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   1,  -2 },    354.13759,   0.00018,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   1,   0 },    354.13759,   0.00018,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   2,   0 },    354.13759,   0.00002,     0.037 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -1,   1 },    166.67445,   0.00002,     0.074 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -1,   3 },    166.67445,   0.00002,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,  -1,   1 },     15.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,  -1,   3 },     15.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   4,   0,   0,   0,   0,   0,  -1,  -1 },     59.68772,   0.00001,     0.075 },
 { {   0,  21, -23,   4,   0,   0,   0,   0,   0,  -1,   1 },     59.68772,   0.00001,     0.074 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   0 },    250.55328,   0.00003,     0.074 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   2 },    250.55328,   0.00003,     0.075 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },     91.99959,   0.00001,     0.073 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,   1 },     91.99959,   0.00001,     0.076 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   1,   0,   0 },     70.67587,   0.00002,     0.041 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,   0,  -1 },    272.25233,   0.00001,     0.043 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,   0,   1 },    272.25233,   0.00001,     0.020 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    179.94393,   0.00003,     0.026 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -1,   1 },    179.94393,   0.00005,     0.087 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.84289,   0.00003,     0.549 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -3,  -1 },    179.94393,   0.00001,     0.025 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -2,  -1 },    180.03436,   0.00016,     0.037 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.94423,   0.00143,     0.074 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.94758,   0.00141,     0.075 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,  -1 },    179.84280,   0.00016,     3.490 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   1 },    179.89331,   0.00032,     0.038 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   1,   1 },    179.94393,   0.00001,     0.025 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -2,   1 },    179.85616,   0.00003,     0.040 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -1,  -1 },    179.94393,   0.00005,     0.089 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -1,   1 },    179.94393,   0.00003,     0.026 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.67604,   0.00001,     0.041 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,   0,   1 },    359.67613,   0.00001,     0.019 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,  -2 },    102.92454,   0.00001,     0.037 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,  -1 },      6.55100,   0.00002,     0.074 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,   1 },      6.55100,   0.00002,     0.075 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,  -2 },    102.92872,   0.00013,     0.074 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,   0 },    102.94561,   0.00027,     0.075 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,  -2 },    103.53899,   0.00002,     2.939 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,   0 },    103.13705,   0.00007,     0.038 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   1,  -1 },    206.04256,   0.00001,     0.075 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   1,   1 },    206.04256,   0.00001,     0.074 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   1,  -1 },    347.10567,   0.00001,     0.076 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   1,   1 },    347.10567,   0.00001,     0.073 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,  -1,  -1 },    272.40772,   0.00003,     0.072 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,  -1,   1 },    272.40772,   0.00003,     0.077 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,  -1,  -1 },    179.79678,   0.00006,     0.077 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,  -1,   1 },    179.79678,   0.00006,     0.072 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,  -1 },      1.61247,   0.00002,     2.950 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,   1 },      1.61299,   0.00002,     0.037 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   1,  -3 },    179.54624,   0.00002,     0.075 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,  -1,  -2 },    102.92071,   0.00001,     0.077 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,  -1,   0 },    282.93198,   0.00002,     0.072 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,   0,   0 },    102.92860,   0.00002,     0.037 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   1,  -1 },     71.69479,   0.00002,     0.074 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   1,   1 },     71.69479,   0.00002,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   0,  -1 },    168.18245,   0.00001,     0.026 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   0,   1 },    168.18245,   0.00002,     0.089 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   1,  -1 },    168.18245,   0.00001,     0.040 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    168.18245,   0.00006,     0.038 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,  -1 },    168.18245,   0.00056,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,   1 },    168.18245,   0.00056,     0.074 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,  -2 },    166.87428,   0.00001,     0.074 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,   0 },    166.87428,   0.00001,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,   1 },    168.18245,   0.00006,     0.037 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,  -1,   1 },    168.18245,   0.00001,     0.040 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,   0,  -1 },    168.18245,   0.00002,     0.088 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,   0,   1 },    168.18245,   0.00001,     0.026 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -2,  -1 },    271.94247,   0.00003,     0.038 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,  -2 },    249.19500,   0.00001,     0.037 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,  -1 },    272.28110,   0.00028,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,   1 },    272.28110,   0.00028,     0.074 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,  -2 },    249.19595,   0.00011,     0.074 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,  -1 },     92.96229,   0.00003,    28.187 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    249.14773,   0.00014,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,   0 },    249.19500,   0.00001,     0.038 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -2,  -2 },    192.75345,   0.00002,     0.038 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,  -2 },    192.74613,   0.00020,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,   0 },    192.00178,   0.00010,     0.074 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -3 },    359.67965,   0.00001,     0.074 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -1 },    359.67965,   0.00001,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,   0 },    192.75345,   0.00002,     0.037 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -2,  -3 },    274.98030,   0.00001,     0.038 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -1,  -3 },    274.98215,   0.00014,     0.075 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -1,  -1 },    275.07045,   0.00010,     0.074 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,   0,  -1 },    274.98030,   0.00001,     0.037 },
 { {   0,  24, -27,  -1,   0,   0,   0,   0,  -2,   1,  -1 },    218.20623,   0.00001,     0.074 },
 { {   0,  24, -27,  -1,   0,   0,   0,   0,  -2,   1,   1 },    218.20623,   0.00001,     0.075 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -1,  -1 },    350.13490,   0.00002,     0.073 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -1,   1 },    350.13490,   0.00002,     0.076 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -3,   1 },    179.43126,   0.00002,     0.088 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -4,  -1 },    179.43054,   0.00005,     0.037 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,  -1 },    179.42857,   0.00044,     0.074 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,   1 },    179.42857,   0.00044,     0.075 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -2,  -1 },    179.54357,   0.00002,     5.436 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -2,   1 },    179.46539,   0.00007,     0.038 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -3,  -1 },    179.43126,   0.00002,     0.089 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   1,  -3 },    356.31599,   0.00002,     0.075 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   1,  -1 },    356.31599,   0.00002,     0.074 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   1,  -4 },     67.08515,   0.00004,     0.074 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   1,  -2 },     67.08515,   0.00004,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -2,  -1 },     75.00074,   0.00002,     0.019 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -2,   1 },     75.00074,   0.00002,     0.041 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     75.00074,   0.00009,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -1,   1 },     75.00074,   0.00016,     0.088 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   0,  -1 },     75.00074,   0.00010,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -3,  -1 },     75.00074,   0.00005,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,  -1 },     75.02413,   0.00052,     0.037 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,  -1 },     74.99747,   0.00464,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   1 },     74.99747,   0.00464,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,  -1 },    252.96025,   0.00003,     6.293 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   1 },     75.13365,   0.00048,     0.037 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   1,   1 },     75.00074,   0.00005,     0.025 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -2,   1 },     75.00074,   0.00010,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -1,  -1 },     75.00074,   0.00016,     0.088 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -1,   1 },     75.00074,   0.00009,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,   0,  -1 },     75.00074,   0.00002,     0.041 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,   0,   1 },     75.00074,   0.00002,     0.019 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -2,  -2 },    166.27827,   0.00002,     0.037 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -1,  -2 },    166.26145,   0.00014,     0.074 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -1,   0 },    166.28348,   0.00018,     0.075 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   0,   0 },    166.27827,   0.00002,     0.038 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -1,  -3 },    272.16400,   0.00002,     0.074 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -1,  -1 },    272.16496,   0.00004,     0.075 },
 { {   0,  27, -20, -20,   0,   0,   0,   0,   0,  -1,  -1 },    329.81539,   0.00001,     0.075 },
 { {   0,  27, -20, -20,   0,   0,   0,   0,   0,  -1,   1 },    329.81539,   0.00001,     0.075 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -2,  -1 },    335.62831,   0.00002,     0.074 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -2,   1 },    335.62831,   0.00002,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -2,  -2 },    162.70535,   0.00001,     0.037 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -1,  -2 },    162.70535,   0.00010,     0.074 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -1,   0 },    162.70535,   0.00010,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,   0,   0 },    162.70535,   0.00001,     0.037 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   2,  -2 },    140.41562,   0.00001,     0.075 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   2,   0 },    140.41562,   0.00001,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,   0,  -2 },    113.91542,   0.00006,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,   0,   0 },    113.91542,   0.00006,     0.075 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   0,  -1 },    307.19129,   0.00003,     0.075 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   0,   1 },    307.19129,   0.00003,     0.074 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,   0,  -1 },     16.24881,   0.00008,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,   0,   1 },     16.24881,   0.00008,     0.075 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   0,  -1 },     91.32596,   0.00002,     0.074 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   0,   1 },     91.32596,   0.00002,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,  -1,  -4 },    281.55562,   0.00003,     0.037 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -4 },    281.55562,   0.00030,     0.074 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -2 },    281.55562,   0.00030,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   1,  -2 },    281.55562,   0.00003,     0.037 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,   0,  -1 },    296.45270,   0.00008,     0.075 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,   0,   1 },    296.45270,   0.00008,     0.075 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,   0,  -1 },    259.65428,   0.00003,     0.074 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,   0,   1 },    259.65428,   0.00003,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   1,   1 },     77.18279,   0.00004,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   1,   3 },     77.18279,   0.00004,     0.074 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -2,  -1 },    114.25737,   0.00004,     0.074 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -2,   1 },    114.25737,   0.00004,     0.075 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,   0,   1 },    126.93434,   0.00002,     0.088 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,   0,  -1 },    126.93434,   0.00002,     0.088 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },    209.54963,   0.00002,     0.026 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    209.54929,   0.00008,     0.090 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,  -1 },    209.52702,   0.00002,     0.040 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   1 },    209.52702,   0.00002,     0.478 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },    211.49028,   0.00004,     0.038 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },    212.09398,   0.00002,     3.136 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    207.91779,   0.00003,     0.075 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    207.91779,   0.00003,     0.074 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },    212.09398,   0.00002,    68.187 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,   1 },    211.49028,   0.00004,     0.037 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,  -1 },    210.64674,   0.00002,     0.559 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   1 },    210.64688,   0.00002,     0.040 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    210.27036,   0.00007,     0.087 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   1 },    210.35237,   0.00001,     0.026 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },    136.63470,   0.00002,     0.027 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },    136.64189,   0.00007,     0.098 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },    136.67635,   0.00002,     0.042 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },    136.67634,   0.00002,     0.323 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    109.58855,   0.00002,     0.089 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    139.76983,   0.00001,     0.081 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    139.76983,   0.00001,     0.069 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },    109.58933,   0.00003,     3.772 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    289.59467,   0.00063,     0.074 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    289.58802,   0.00003,     0.037 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    138.57012,   0.00003,     0.080 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    258.01574,   0.00004,     0.089 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },     34.65279,   0.00008,     0.069 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,  -1 },    257.00902,   0.00003,     4.732 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,   1 },     75.75636,   0.00006,     0.037 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   0,  -3 },     78.18480,   0.00002,     0.075 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.19897,   0.00161,     0.074 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   1,  -1 },     76.19612,   0.00009,     0.037 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },     53.92658,   0.00003,     0.064 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    177.86857,   0.00002,     0.069 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    316.26922,   0.00001,     0.122 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    135.94127,   0.00003,     0.061 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -2 },      4.34116,   0.00001,     0.037 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   1,  -2 },      4.34116,   0.00013,     0.074 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },      4.34116,   0.00013,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   2,   0 },      4.34116,   0.00001,     0.037 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -3,  -1 },     27.44429,   0.00003,     0.074 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -3,   1 },     27.44429,   0.00003,     0.075 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },     90.18924,   0.00002,     0.074 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    303.94220,   0.00002,     0.075 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    180.16141,   0.00001,     0.140 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    180.16141,   0.00001,     0.051 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -2,   1 },    179.90096,   0.00001,     0.056 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    179.88907,   0.00003,     0.057 },
 { {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    132.26122,   0.00002,     0.051 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   1,   0 },    312.75570,   0.00002,     0.075 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   1,   2 },    312.75570,   0.00002,     0.074 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   1,  -1 },     96.56536,   0.00002,     0.026 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   1,   1 },     96.56536,   0.00004,     0.088 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   2,  -1 },     96.56536,   0.00003,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     96.56536,   0.00001,     0.025 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1 },     96.40479,   0.00018,     0.037 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },     95.70877,   0.00003,     7.078 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,  -1 },     96.56536,   0.00131,     0.074 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   1 },     96.56536,   0.00131,     0.075 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   1 },     96.52238,   0.00014,     0.038 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   1 },     96.56536,   0.00001,     0.025 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },     96.56536,   0.00003,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,  -1 },     96.56536,   0.00004,     0.089 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   1 },     96.56536,   0.00002,     0.026 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   1,  -1 },    353.96206,   0.00007,     0.074 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   1,   1 },    353.96206,   0.00007,     0.075 },
 { {   3,   2,  -6,   0,   0,   0,   0,   0,  -4,   3,  -1 },    239.57624,   0.00001,     0.075 },
 { {   3,   2,  -6,   0,   0,   0,   0,   0,  -4,   3,   1 },    239.57624,   0.00001,     0.075 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   2,  -1 },     19.56685,   0.00003,     0.075 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   2,   1 },     19.56685,   0.00003,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   1,  -1 },     30.68945,   0.00007,     0.074 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   1,   1 },     30.68945,   0.00007,     0.075 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   0,  -1 },     73.47726,   0.00001,     0.038 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   1,  -1 },     73.47726,   0.00013,     0.075 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   1,   1 },     73.47726,   0.00013,     0.074 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   2,   1 },     73.47726,   0.00001,     0.037 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   1,  -2 },     24.76162,   0.00005,     0.075 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },     24.57741,   0.00003,     0.074 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },    183.02988,   0.00002,     0.071 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },    183.02988,   0.00002,     0.078 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },     99.36455,   0.00003,     0.077 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },     99.36455,   0.00003,     0.073 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -3 },    100.63105,   0.00002,     0.038 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   1,  -3 },    100.63014,   0.00019,     0.075 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    100.68226,   0.00016,     0.074 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   2,  -1 },    100.63105,   0.00002,     0.037 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   1,  -1 },    303.08124,   0.00002,     0.075 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   1,   1 },    303.08124,   0.00002,     0.075 },
 { {   6,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,  -1 },    231.51527,   0.00001,     0.075 },
 { {   6,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,   1 },    231.51527,   0.00001,     0.074 },
 { {   6,   0,   0,   0,   0,   0,   0,   0,  -2,   0,  -1 },    235.02094,   0.00001,     0.075 },
 { {   6,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   1 },    235.02094,   0.00001,     0.074 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },    138.00381,   0.00001,     0.076 },
 { {   7,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    138.00381,   0.00001,     0.073 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 6631> g_ELP12
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     90.00000,   0.02045, 99999.999 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 },    270.01562,   0.00037,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },     89.99679,   0.00010,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    269.93968,   0.00720,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    269.94077,   0.00028,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    270.00276,   0.00004,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    270.01112,   0.00025,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },     90.01711,   0.00343,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },     90.01715,   0.00021,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.08025,   0.00203,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2 },     90.14419,   0.00007,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.10502,   0.03074,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    270.07659,   0.01960,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2 },     90.02518,   0.00004,     0.090 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    270.11156,   0.00124,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    269.93283,   0.00010,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.03445,   0.00007,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 },    270.04164,   0.00010,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0 },    270.12246,   0.00034,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.18153,   0.00047,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0 },    270.13595,   0.00018,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0 },     49.83798,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   1,   0 },    171.18864,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },    351.05178,   0.00029,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0 },    171.05178,   0.00029,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0 },    311.68508,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0 },    224.65041,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   0 },    243.55642,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },     63.58419,   0.00014,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },    243.58419,   0.00014,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },     63.55642,   0.00002,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   0,   0 },    245.69512,   0.00012,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   1,   0 },    348.85483,   0.00015,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   2,   0 },    192.20133,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -2,   0 },    192.00892,   0.00005,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -1,   0 },    191.14709,   0.00115,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   1,   0 },     11.14709,   0.00115,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0 },     12.00892,   0.00005,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,  -2,   0 },     12.20132,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,  -1,   0 },    168.93620,   0.00015,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,  -2,   2,   0,   0 },     65.99758,   0.00012,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0 },    277.86427,   0.00022,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0 },    258.86594,   0.00021,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0 },    327.31563,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0 },    331.47933,   0.00042,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0 },    151.55912,   0.00043,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   0,   2,   0 },    147.28441,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -1,   0 },     80.93914,   0.00021,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0 },     99.52340,   0.00022,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0 },     71.93169,   0.00010,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   2,   0,   0,   1,   0 },    251.93169,   0.00010,     0.075 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   0,  -1,   0 },     35.75292,   0.00007,     0.075 },
 { {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   1,   0 },    215.75292,   0.00007,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0 },    357.85025,   0.00005,     0.020 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   1,   0 },    357.83516,   0.00016,     0.028 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -4,   2,   0 },    357.71244,   0.00005,     0.044 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,  -1,   0 },    356.32430,   0.00036,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0 },    357.69043,   0.00487,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   0 },    359.14932,   0.00469,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   2,   0 },    182.84467,   0.00022,     0.553 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0 },    177.44541,   0.00011,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0 },    350.50556,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0 },    350.84334,   0.00043,     0.038 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },    350.76811,   0.00785,     0.076 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   2 },    171.50986,   0.00003,     0.073 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },    168.62060,   0.00008,    29.457 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -2 },    351.50986,   0.00003,     0.074 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0 },    170.73819,   0.00786,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   2,   0 },    170.80556,   0.00043,     0.038 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   3,   0 },    170.40041,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0 },    357.70167,   0.00010,     0.081 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -2,   0 },      2.33327,   0.00022,     0.575 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,   0 },    178.83645,   0.00467,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0 },    176.90427,   0.00495,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   1,   0 },    176.20134,   0.00035,     0.026 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,  -2,   0 },    177.48624,   0.00005,     0.043 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,  -1,   0 },    177.47712,   0.00015,     0.028 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   4,   0,   0 },    177.09498,   0.00005,     0.020 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0 },    260.99447,   0.00027,     0.041 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,  -2,   1,   0 },    252.27147,   0.00039,     0.088 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0 },     55.05636,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0 },     56.81895,   0.00047,     0.076 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,   0 },    236.96045,   0.00048,     0.075 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   2,   0 },    235.00835,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,  -1,   0 },     73.32384,   0.00032,     0.087 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   2,   0,   0 },     89.32108,   0.00017,     0.040 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,  -2,   0,   0 },    146.54862,   0.00005,     0.041 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   0,  -1,   0 },    349.33632,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   0,   3,   0,   0,   0,   1,   0 },    169.33632,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,  -2,   0,   0 },    290.35357,   0.00020,     0.040 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,  -2,   1,   0 },    282.43906,   0.00011,     0.086 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,  -1,   0 },    334.87634,   0.00011,     0.075 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,   1,   0 },    154.87634,   0.00011,     0.076 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   2,  -1,   0 },     96.07022,   0.00012,     0.088 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   2,   0,   0 },    103.55451,   0.00023,     0.041 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,  -2,   0,   0 },    224.88844,   0.00010,     0.040 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,  -2,   1,   0 },    211.80743,   0.00006,     0.087 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   0,  -1,   0 },    325.51258,   0.00009,     0.075 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   0,   1,   0 },    145.51258,   0.00009,     0.076 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   2,  -1,   0 },     28.84273,   0.00006,     0.087 },
 { {   0,   0,   0,   0,   1,  -3,   0,   0,   2,   0,   0 },     41.88420,   0.00010,     0.040 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   0,   0 },     82.29538,   0.00032,     0.040 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,  -2,   1,   0 },    100.92564,   0.00030,     0.087 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -2,   0 },     28.29518,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,  -1,   0 },     20.76014,   0.00057,     0.076 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   1,   0 },    200.68661,   0.00057,     0.075 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   0,   2,   0 },    207.63543,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,  -1,   0 },    278.57334,   0.00030,     0.087 },
 { {   0,   0,   0,   0,   1,  -2,   0,   0,   2,   0,   0 },    260.36849,   0.00032,     0.040 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   0,   0 },    267.14581,   0.00022,     0.041 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,  -2,   1,   0 },    267.28526,   0.00022,     0.087 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,  -1,   0 },    265.68721,   0.00011,     0.076 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   0,   1,   0 },     85.57195,   0.00011,     0.075 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,  -1,   0 },     86.96694,   0.00022,     0.087 },
 { {   0,   0,   0,   0,   1,  -1,   0,   0,   2,   0,   0 },     86.87333,   0.00021,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -6,   1,   0 },     87.02777,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,  -1,   0 },     78.72463,   0.00005,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   0,   0 },     90.37129,   0.00036,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   1,   0 },     91.36642,   0.00095,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -4,   2,   0 },     91.76254,   0.00052,     0.044 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 },     72.24783,   0.00015,     0.027 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -3,   1,   0 },    273.84937,   0.00007,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -2,   0 },     80.23644,   0.00032,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   0 },     81.83990,   0.00255,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   2 },    265.19994,   0.00004,     0.090 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 },     85.30070,   0.04106,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   2 },    265.49263,   0.00018,     0.456 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   0 },     89.84062,   0.04609,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   2 },    270.25371,   0.00011,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   2,   0 },    276.15644,   0.00290,     0.538 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   3,   0 },    288.90844,   0.00014,     0.066 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   0 },    299.54440,   0.00006,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   1 },    168.48339,   0.00005,     0.083 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,  -1 },    348.86944,   0.00006,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 },    281.93025,   0.00082,     0.081 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,  -1 },    348.53595,   0.00008,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   0 },     91.01809,   0.00004,     1.030 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   1 },    348.48339,   0.00005,     0.069 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -3,   0 },     13.52337,   0.00014,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 },     21.69425,   0.00195,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },     24.97221,   0.03716,     0.076 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   2 },    200.95689,   0.00013,     0.073 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },    195.06565,   0.00060,    11.862 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   2 },     53.65397,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,  -2 },     25.35208,   0.00016,     0.074 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 },    203.58505,   0.03741,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,   0 },    200.33164,   0.00200,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   3,   0 },    197.77664,   0.00012,     0.025 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,  -1 },    155.62597,   0.00006,     0.070 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0 },    283.97780,   0.00003,     1.246 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   1 },    168.55283,   0.00009,     0.079 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },     88.34629,   0.00061,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   1 },    168.56374,   0.00006,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   1,  -1 },      6.04259,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   1,   0 },     84.06783,   0.00004,     0.039 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -3,   0 },    113.13046,   0.00014,     0.067 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },     96.67799,   0.00266,     0.592 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,  -2 },     88.54743,   0.00011,     0.065 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    266.09561,   0.04282,     0.086 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,  -2 },     80.59674,   0.00016,     0.494 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    256.15478,   0.03539,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,  -2 },     80.04295,   0.00004,     0.089 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,   0 },    247.09364,   0.00226,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   2,   0 },    256.49361,   0.00028,     0.019 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,  -1,  -1 },    268.56435,   0.00002,     0.095 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,  -1,   0 },     78.94953,   0.00006,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,   0,  -1 },     88.41128,   0.00003,     0.042 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   3,   0,   0 },    256.72384,   0.00013,     0.027 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -2,   0 },    267.49659,   0.00047,     0.043 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,  -1,   0 },    257.83354,   0.00082,     0.028 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   0,   0 },    255.05429,   0.00031,     0.020 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   4,   1,   0 },    266.63962,   0.00004,     0.016 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   6,  -1,   0 },    254.76070,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,  -2,   0,   0 },    191.18291,   0.00007,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,  -2,   1,   0 },    191.18292,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   0,  -1,   0 },    198.51356,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   0,   1,   0 },     18.51356,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   2,  -1,   0 },     10.40235,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   0,   0,   2,   0,   0 },     10.40235,   0.00006,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,  -2,   0,   0 },    108.65301,   0.00006,     0.040 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,  -2,   1,   0 },    270.04391,   0.00009,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -2,   0 },     90.04391,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,  -1,   0 },     89.84776,   0.00059,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   1,   0 },    269.84776,   0.00059,     0.075 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   0,   2,   0 },    270.04391,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   2,  -1,   0 },     90.04391,   0.00009,     0.087 },
 { {   0,   0,   0,   0,   2,  -6,   3,   0,   2,   0,   0 },    288.33086,   0.00006,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   0,   0 },    331.45717,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   1,   0 },    282.32247,   0.00013,     0.028 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -4,   2,   0 },    265.44868,   0.00012,     0.044 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -2,   0 },     74.69138,   0.00007,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,  -1,   0 },     62.26562,   0.00052,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   0 },    350.23093,   0.00192,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   0 },    268.64593,   0.01128,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   2 },     75.96095,   0.00002,     0.065 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   2,   0 },     82.90966,   0.00151,     0.564 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   3,   0 },     77.45229,   0.00012,     0.067 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -1,   0,   0 },    150.02602,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -1,   1,   0 },    255.96095,   0.00002,     1.129 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -3,   0 },     78.95434,   0.00022,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   0 },     76.47523,   0.00344,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   2 },     75.96095,   0.00002,     3.009 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   0 },     76.55968,   0.06312,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   2 },    256.06876,   0.00025,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,  -2 },     76.06876,   0.00025,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   0 },    256.56163,   0.06313,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,  -2 },    255.96095,   0.00002,     2.988 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   0 },    256.47665,   0.00344,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   3,   0 },    258.96570,   0.00022,     0.025 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,  -1,   0 },     75.96095,   0.00002,     1.126 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   1,   0,   0 },    333.50322,   0.00005,     0.081 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -3,   0 },    257.47361,   0.00012,     0.067 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -2,   0 },    262.87097,   0.00151,     0.563 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,  -2 },    255.96095,   0.00002,     0.065 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   0 },     88.52073,   0.01131,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,   0 },    165.60693,   0.00192,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   1,   0 },    243.47733,   0.00052,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   2,   0 },    254.55349,   0.00007,     0.020 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -2,   0 },     85.38961,   0.00012,     0.044 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,  -1,   0 },    101.89847,   0.00013,     0.028 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   4,   0,   0 },    149.21458,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,   0,   0 },    340.78551,   0.00011,     0.040 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,  -2,   1,   0 },    340.78304,   0.00015,     0.087 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,  -1,   0 },    349.39476,   0.00007,     0.076 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   1,   0 },    169.39476,   0.00007,     0.075 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,  -1,   0 },    161.03744,   0.00015,     0.087 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   2,   0,   0 },    161.03215,   0.00010,     0.040 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,   0,   0 },     30.40427,   0.00017,     0.041 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,  -2,   1,   0 },     30.33370,   0.00015,     0.088 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,  -1,   0 },     30.86788,   0.00018,     0.076 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,   1,   0 },    210.86434,   0.00018,     0.075 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,  -1,   0 },    210.28614,   0.00015,     0.087 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   2,   0,   0 },    210.36994,   0.00015,     0.040 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   0,   0 },     89.18166,   0.00038,     0.041 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,  -2,   1,   0 },     89.44582,   0.00035,     0.088 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,  -1,   0 },     86.29804,   0.00018,     0.076 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   0,   1,   0 },    266.00692,   0.00017,     0.075 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,  -1,   0 },    268.81609,   0.00037,     0.086 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   0,   0 },    268.65381,   0.00032,     0.040 },
 { {   0,   0,   0,   0,   2,  -2,   0,   0,   2,   1,   0 },     88.65393,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -4,   1,   0 },    218.75673,   0.00002,     0.028 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -4,   2,   0 },    244.93306,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,  -1,   0 },    102.58522,   0.00011,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 },    110.38163,   0.00082,     0.041 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,  -1 },    222.88991,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,   0 },    128.01880,   0.00075,     0.088 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   2,   0 },    340.32352,   0.00005,     0.515 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0 },    231.21443,   0.00008,     0.082 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 },     49.26596,   0.00022,     0.038 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0 },     47.74741,   0.00386,     0.076 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1 },    223.17132,   0.00033,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 },    231.53843,   0.00015,     5.931 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   0 },    227.28006,   0.00397,     0.074 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1 },     43.67192,   0.00013,     0.075 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   2,   0 },    229.40522,   0.00023,     0.037 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -2,   0 },    138.90462,   0.00004,     0.623 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -1,   0 },    262.13546,   0.00103,     0.086 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   0 },    243.60175,   0.00154,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   1,  -1 },     42.88991,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   1,   0 },    241.13201,   0.00022,     0.026 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   4,  -1,   0 },    248.47716,   0.00003,     0.027 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,  -2,   0 },    209.78860,   0.00002,     0.697 },
 { {   0,   0,   0,   0,   2,   5,   0,   0,   2,   0,   0 },     29.78860,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   0,   0 },    174.97542,   0.00028,     0.041 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,  -2,   1,   0 },    174.44718,   0.00026,     0.088 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,  -1,   0 },    178.38081,   0.00024,     0.076 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   1,   0 },    358.33407,   0.00024,     0.075 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,  -1,   0 },    354.23761,   0.00026,     0.086 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   0 },    354.80209,   0.00025,     0.040 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,  -2,   0,   0 },    210.93600,   0.00006,     0.041 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,  -2,   1,   0 },    210.93600,   0.00003,     0.088 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   0,  -1,   0 },    212.74231,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   0,   1,   0 },     32.74231,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   2,  -1,   0 },     30.50365,   0.00003,     0.086 },
 { {   0,   0,   0,   0,   3,  -4,   0,   0,   2,   0,   0 },     30.50365,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,  -2,   0,   0 },     72.07391,   0.00007,     0.041 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,  -2,   1,   0 },     72.07391,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   0,  -1,   0 },     80.34054,   0.00002,     0.076 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   0,   1,   0 },    260.34054,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   2,  -1,   0 },    247.34517,   0.00004,     0.086 },
 { {   0,   0,   0,   0,   3,  -3,   0,   0,   2,   0,   0 },    247.34516,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,  -2,   0,   0 },     86.81054,   0.00007,     0.041 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   0,  -1,   0 },     65.80159,   0.00004,     0.077 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   0,   1,   0 },    245.80159,   0.00004,     0.074 },
 { {   0,   0,   0,   0,   3,  -2,   0,   0,   2,   0,   0 },    257.94802,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -2,   0,   0 },    198.50316,   0.00017,     0.041 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,  -2,   1,   0 },    162.37366,   0.00004,     0.089 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,  -1,   0 },     43.76416,   0.00029,     0.077 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   1,   0 },    225.29075,   0.00034,     0.074 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   2,  -1,   0 },    343.47058,   0.00003,     0.085 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   2,   0,   0 },    231.59346,   0.00016,     0.040 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,  -2,   0,   0 },    129.19358,   0.00004,     0.040 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,  -2,   1,   0 },     87.39404,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,  -1,   0 },    268.52240,   0.00024,     0.075 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   0,   1,   0 },     88.52240,   0.00024,     0.075 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   2,  -1,   0 },    267.39404,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   4, -10,   0,   0,   2,   0,   0 },    307.90135,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,  -2,   0,   0 },    164.66964,   0.00005,     0.041 },
 { {   0,   0,   0,   0,   4,  -5,   0,   0,   2,   0,   0 },    351.33331,   0.00004,     0.040 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,  -2,   0 },     50.02187,   0.00002,     0.695 },
 { {   0,   0,   0,   0,   6,  -5,   0,   0,   2,   0,   0 },    230.02187,   0.00002,     0.040 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -4,   1,   0 },    128.92376,   0.00004,     0.028 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -1,   0 },    128.70840,   0.00014,     0.027 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0 },    130.22468,   0.00101,     0.041 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0 },    128.10800,   0.00099,     0.091 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0 },    308.44618,   0.00008,     0.434 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0 },      5.75619,   0.00003,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },    122.41498,   0.00010,     1.881 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0 },    227.62359,   0.00003,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1 },    218.61361,   0.00002,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1 },    218.70847,   0.00011,     0.077 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0 },    299.72391,   0.00004,     0.078 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1 },     38.61361,   0.00002,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -2,   0 },    119.37184,   0.00004,     0.805 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1,   0 },    301.61006,   0.00064,     0.083 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,   0 },    294.09327,   0.00045,     0.040 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   1,   0 },    300.14567,   0.00004,     0.026 },
 { {   0,   0,   0,   2,   0,  -2,   0,   0,  -2,   0,   0 },    117.14722,   0.00007,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,   0,   0 },    214.85823,   0.00005,     0.021 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -4,   1,   0 },    215.36818,   0.00008,     0.028 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -1,   0 },    219.01822,   0.00028,     0.027 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0 },    223.31948,   0.00277,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   1,   0 },    225.04933,   0.00155,     0.096 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   2,   0 },     52.31701,   0.00008,     0.352 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,  -1,  -1 },    253.73806,   0.00014,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,  -1 },    253.72257,   0.00094,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   1 },    253.75180,   0.00004,     0.472 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1,  -1 },     73.79833,   0.00060,     0.087 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   1,   1 },    253.75180,   0.00003,     0.065 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   2,  -1 },    253.76077,   0.00016,     0.561 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0 },    238.77993,   0.00005,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0 },    250.09525,   0.00081,     0.082 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },     85.51993,   0.00027,     0.940 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,   0 },    126.24779,   0.00042,     0.070 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -3,  -1 },    253.76077,   0.00002,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -2,  -1 },    253.74835,   0.00055,     0.038 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -1 },    253.74835,   0.01008,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,   1 },    253.63608,   0.00034,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -3 },    253.70448,   0.00004,     0.074 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -1 },     73.74835,   0.01008,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   2,  -1 },     73.74835,   0.00055,     0.038 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   3,  -1 },     73.76077,   0.00002,     0.025 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1,   0 },    242.52033,   0.00007,     0.080 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,   0 },    191.65634,   0.00023,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   1,   0 },    208.50267,   0.00004,     0.026 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -2,  -1 },     73.76077,   0.00016,     0.567 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,  -1,  -1 },    253.79265,   0.00068,     0.087 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   0,  -1 },     73.72138,   0.00089,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   3,   1,  -1 },     73.73806,   0.00014,     0.026 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   0,   0 },     56.65462,   0.00006,     0.043 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -1,   0 },      0.42726,   0.00010,     0.076 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,  -2,   0,   0 },    126.93739,   0.00010,     0.044 },
 { {   0,   0,   1,  -8,   3,   0,   0,   0,   2,   0,   0 },    145.64029,   0.00006,     0.046 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   0,   0 },    314.15783,   0.00009,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   1,   0 },    256.06892,   0.00003,     0.079 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,  -1,   1 },     75.16456,   0.00005,     0.075 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   1,   1 },    255.16456,   0.00005,     0.076 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0 },     77.33948,   0.00006,     0.070 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0 },    113.26843,   0.00006,     0.888 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0 },    298.85069,   0.00024,     0.082 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -2,   0 },    131.39379,   0.00003,     0.345 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -1,   0 },    312.30028,   0.00060,     0.097 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,   0 },    310.93336,   0.00103,     0.042 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   1,   0 },    310.66572,   0.00010,     0.027 },
 { {   0,   0,   1,  -4,   2,   0,   0,   0,   2,   0,   0 },     25.72613,   0.00005,     0.042 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   0,   0 },    297.21645,   0.00011,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   1,   0 },    216.65599,   0.00012,     0.083 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   1,  -1 },    291.28465,   0.00008,     0.076 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0 },     11.72788,   0.00025,     0.072 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0 },     54.95467,   0.00007,     1.681 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0 },    200.97457,   0.00019,     0.079 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,  -1,   0 },    353.16591,   0.00026,     0.092 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,   0 },    341.96095,   0.00042,     0.041 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   1,   0 },    352.10459,   0.00003,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,  -1,   0 },     48.65717,   0.00003,     0.016 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   0,   0 },     49.49209,   0.00032,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   1,   0 },     49.41380,   0.00059,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -4,   2,   0 },     51.45272,   0.00029,     0.043 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   0,   0 },     47.93438,   0.00009,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   0,   1 },      3.34108,   0.00006,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   1,   0 },    227.98847,   0.00004,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -3,   1,   1 },    183.26775,   0.00003,     0.095 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -2,   0 },     47.77025,   0.00025,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   0 },     46.46016,   0.00215,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   2 },    228.63299,   0.00003,     0.089 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   0 },     48.21733,   0.02862,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   2 },    228.85620,   0.00011,     0.489 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   0 },     50.88279,   0.02591,     0.087 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   2 },    231.77272,   0.00006,     0.065 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   2,   0 },    236.47500,   0.00099,     0.585 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   3,   0 },    248.45876,   0.00004,     0.067 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,  -1,   1 },      3.31472,   0.00006,     0.082 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   0 },    229.56886,   0.00047,     0.080 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   1,   1 },      3.31472,   0.00005,     0.070 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   0 },     42.99992,   0.00019,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0 },     43.27565,   0.00308,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   2 },     42.29837,   0.00002,     3.702 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0 },     43.26801,   0.05857,     0.075 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2 },    222.66289,   0.00022,     0.074 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2 },    227.81814,   0.00004,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },     47.28992,   0.00154,    15.781 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2 },    229.84124,   0.00004,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2 },     42.64390,   0.00020,     0.073 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0 },    223.46741,   0.05896,     0.076 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -2 },    223.47080,   0.00002,     2.520 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0 },    223.17234,   0.00309,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   3,   0 },    223.15082,   0.00019,     0.025 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,   0 },     51.79641,   0.00054,     0.081 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   1,   0 },     46.02609,   0.00005,     0.039 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -3,   0 },     64.46189,   0.00005,     0.066 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -2,   0 },     56.06465,   0.00124,     0.544 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,  -2 },     52.31357,   0.00006,     0.065 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   0 },    231.97848,   0.02841,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,  -2 },     50.25675,   0.00012,     0.461 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   0 },    230.25486,   0.03123,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,  -2 },     50.05313,   0.00003,     0.090 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,   0 },    229.04338,   0.00226,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,   0 },    228.76174,   0.00030,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   3,  -1,   0 },     50.79243,   0.00005,     0.042 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   3,   0,   0 },    230.88134,   0.00010,     0.027 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -2,   0 },    232.03970,   0.00032,     0.044 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,  -1,   0 },    230.80536,   0.00065,     0.028 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   0,   0 },    228.06391,   0.00028,     0.020 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   4,   1,   0 },    229.33776,   0.00003,     0.016 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   0,   0 },    123.61076,   0.00026,     0.040 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,  -2,   1,   0 },    125.08368,   0.00024,     0.087 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,  -1,   0 },    117.19533,   0.00032,     0.075 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   0,   1,   0 },    297.18222,   0.00032,     0.075 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,  -1,   0 },    305.03185,   0.00024,     0.087 },
 { {   0,   0,   1,  -2,   0,   2,   0,   0,   2,   0,   0 },    303.56422,   0.00026,     0.040 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,  -2,   0,   0 },    285.15893,   0.00007,     0.040 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,  -2,   1,   0 },    285.15893,   0.00004,     0.087 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,  -1,   0 },    281.93895,   0.00004,     0.076 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   0,   1,   0 },    101.93895,   0.00004,     0.075 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   2,  -1,   0 },    103.61160,   0.00005,     0.087 },
 { {   0,   0,   1,  -2,   1,   0,   0,   0,   2,   0,   0 },    103.66881,   0.00007,     0.040 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   0,   0 },     90.69725,   0.00006,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   1,   0 },     90.64241,   0.00020,     0.028 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -4,   2,   0 },     90.65247,   0.00007,     0.044 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,  -1,   0 },     90.62425,   0.00035,     0.027 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   0 },     90.67958,   0.00564,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   2 },    270.64810,   0.00004,     0.388 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   0 },     90.67899,   0.00535,     0.091 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   2,   0 },    270.70295,   0.00034,     0.446 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   0 },    270.46955,   0.00012,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0 },     90.53306,   0.00018,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0 },     90.52550,   0.00302,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0 },    269.17838,   0.00035,     2.135 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0 },    270.44677,   0.00274,     0.073 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0 },    270.39238,   0.00013,     0.037 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,  -3,   1 },      4.41151,   0.00002,     0.074 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,  -1,   1 },    184.41151,   0.00002,     0.077 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0 },    270.23302,   0.00009,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -2,   0 },     90.19296,   0.00018,     0.766 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -1,   0 },    270.36164,   0.00368,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0 },    270.30126,   0.00272,     0.040 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   1,   0 },    270.28706,   0.00034,     0.026 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,  -2,   0 },    270.47909,   0.00004,     0.043 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,  -1,   0 },    270.28776,   0.00005,     0.027 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   4,   0,   0 },    270.31689,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,  -2,   0,   0 },     57.39648,   0.00010,     0.041 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,  -2,   1,   0 },     57.39451,   0.00004,     0.092 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   0,  -1,   0 },     53.41315,   0.00006,     0.079 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   0,   1,   0 },    233.41315,   0.00006,     0.072 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   2,  -1,   0 },    235.92462,   0.00003,     0.083 },
 { {   0,   0,   1,   0,  -5,   0,   0,   0,   2,   0,   0 },    235.92462,   0.00003,     0.040 },
 { {   0,   0,   1,   0,  -5,   2,   0,   0,  -2,   0,   0 },     42.39744,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,  -2,   0,   0 },    273.85045,   0.00008,     0.042 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   0,  -1,   0 },    272.85644,   0.00003,     0.080 },
 { {   0,   0,   1,   0,  -5,   5,   0,   0,   0,   1,   0 },     92.85644,   0.00003,     0.071 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,  -1,   0 },     47.45200,   0.00007,     0.027 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   0,   0 },     47.92614,   0.00052,     0.042 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,  -2,   1,   0 },     47.45554,   0.00036,     0.092 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -2,   0 },     43.66100,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,  -1,   0 },     43.23139,   0.00052,     0.079 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 },    226.86622,   0.00011,     1.509 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   1,   0 },    222.97252,   0.00048,     0.072 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   2,   0 },    223.72619,   0.00002,     0.037 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   1,  -1,   1 },    211.20460,   0.00011,     0.076 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,  -1,   0 },    243.67024,   0.00009,     0.082 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   2,   0,   0 },    226.08335,   0.00017,     0.039 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,  -2,   0,   0 },     43.23834,   0.00012,     0.042 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,  -2,   1,   0 },     43.23834,   0.00005,     0.093 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,  -1,   0 },     38.34030,   0.00005,     0.080 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,   0,   0 },    221.90391,   0.00003,     1.369 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   0,   1,   0 },    218.34030,   0.00005,     0.071 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   1,  -1,   1 },     11.05460,   0.00006,     0.075 },
 { {   0,   0,   1,   0,  -4,   2,   0,   0,   2,  -1,   0 },    224.73021,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,  -2,   0,   0 },     96.00873,   0.00007,     0.042 },
 { {   0,   0,   1,   0,  -4,   3,   0,   0,   2,  -1,   0 },     70.08400,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -4,   4,   0,   0,  -2,   0,   0 },    329.61590,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,  -2,   0,   0 },    207.63882,   0.00020,     0.042 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,  -2,   1,   0 },    207.63886,   0.00007,     0.094 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,  -1,   0 },    201.37049,   0.00010,     0.080 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,   1,   0 },     21.37049,   0.00010,     0.071 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,  -2,   0 },     28.76570,   0.00002,     1.062 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,  -1,   0 },     12.11278,   0.00005,     0.081 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   2,   0,   0 },     28.76570,   0.00002,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   0,   0 },     39.30947,   0.00006,     0.021 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   1,   0 },     39.29864,   0.00020,     0.028 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -4,   2,   0 },     39.92868,   0.00004,     0.045 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -2,   0 },     38.80969,   0.00002,     0.020 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,  -1,   0 },     38.16764,   0.00036,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },     39.55402,   0.00533,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   2 },    219.27894,   0.00004,     0.350 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,   0 },     39.48749,   0.00439,     0.093 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   2,   0 },    219.03619,   0.00026,     0.397 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,  -2,   1 },     20.65730,   0.00005,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,  -1,   1 },     20.64762,   0.00041,     0.087 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,  -1 },     21.73568,   0.00003,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,   0 },    218.38837,   0.00014,     0.086 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -1,   0,   1 },    200.66973,   0.00006,     0.555 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -2,   0 },     35.47008,   0.00026,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,   0 },     36.60085,   0.00407,     0.080 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 },    214.84856,   0.00126,     1.339 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   0 },    214.91711,   0.00400,     0.071 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   2,   0 },    214.60802,   0.00020,     0.037 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -3,   1 },     20.62046,   0.00007,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -2,   1 },    200.66035,   0.00013,    37.265 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,  -1 },    201.02527,   0.00015,     0.074 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,  -1,   1 },     20.66609,   0.00226,     0.075 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,   0 },    217.99373,   0.00029,     0.076 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   0,   1 },     20.52414,   0.00008,     0.038 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   1,   1,  -1 },     21.02527,   0.00015,     0.077 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -3,   0 },     24.66197,   0.00004,     0.070 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -2,   0 },    184.77724,   0.00006,     0.974 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,   0 },    358.88438,   0.00131,     0.082 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,   0 },    224.47052,   0.00122,     0.039 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   1,   0 },    218.55783,   0.00019,     0.026 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -3,   0 },     23.01739,   0.00015,     0.096 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -2,   0 },     22.29867,   0.00046,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,  -1,   0 },     20.17381,   0.00024,     0.027 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   4,   0,   0 },     26.80943,   0.00004,     0.020 },
 { {   0,   0,   1,   0,  -3,   1,   0,   0,  -2,   0,   0 },    220.31710,   0.00003,     0.042 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,  -2,   0,   0 },     86.28552,   0.00015,     0.042 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,  -2,   1,   0 },     86.28551,   0.00013,     0.094 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,  -1,   0 },     85.28036,   0.00013,     0.080 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,   0,   0 },    264.66062,   0.00002,     1.227 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   0,   1,   0 },    265.28036,   0.00013,     0.071 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,  -3,   0 },     90.63177,   0.00002,     0.070 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,  -1,   0 },    267.17054,   0.00007,     0.081 },
 { {   0,   0,   1,   0,  -3,   2,   0,   0,   2,   0,   0 },    266.79717,   0.00008,     0.039 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,  -2,   0,   0 },    150.48779,   0.00013,     0.042 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,  -2,   1,   0 },    150.48779,   0.00005,     0.094 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,  -1,   0 },    150.37029,   0.00004,     0.081 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,   0,   0 },    330.23020,   0.00003,     1.178 },
 { {   0,   0,   1,   0,  -3,   3,   0,   0,   0,   1,   0 },    330.37029,   0.00004,     0.071 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,  -1,   0 },    262.54480,   0.00005,     0.027 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   0,   0 },    262.55519,   0.00049,     0.042 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   1,   0 },    262.44040,   0.00027,     0.095 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,  -1,   0 },    262.82124,   0.00033,     0.081 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   0,   0 },     82.49786,   0.00011,     1.091 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   1,   0 },     82.89465,   0.00026,     0.071 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   1,   0,   0 },    262.58186,   0.00016,     0.075 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,  -1,   0 },     80.74165,   0.00007,     0.081 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   2,   0,   0 },     82.46904,   0.00017,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,  -1,   0 },     35.66954,   0.00007,     0.016 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   0,   0 },     35.94628,   0.00036,     0.021 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   0,   2 },     36.21385,   0.00003,     0.046 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   1,   0 },     36.85525,   0.00081,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -4,   2,   0 },     35.70036,   0.00042,     0.045 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   0,   0 },     36.69102,   0.00013,     0.028 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   0,   1 },    190.50464,   0.00002,     0.044 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   1,   0 },    216.64110,   0.00007,     0.043 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -3,   1,   1 },     10.56776,   0.00002,     0.104 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -2,   0 },     34.50986,   0.00021,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,   0 },     36.14599,   0.00271,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },     36.02580,   0.04322,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   2 },    215.72866,   0.00019,     0.340 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },     36.01309,   0.03426,     0.094 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   2 },    215.51000,   0.00011,     0.062 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   2,   0 },    216.68895,   0.00214,     0.384 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   3,   0 },    218.07668,   0.00011,     0.063 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -2,   1 },     10.52403,   0.00004,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   0 },    216.06885,   0.00007,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,  -1,   1 },     10.51669,   0.00058,     0.088 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0 },    215.72767,   0.00177,     0.087 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   1 },     10.49585,   0.00006,     0.530 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,  -1 },     10.54382,   0.00031,     0.085 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,   0 },     35.94213,   0.00020,     0.582 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   1,   1 },     10.55754,   0.00015,     0.066 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   2,  -1 },    190.51854,   0.00002,     0.644 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -3,   0 },     30.00233,   0.00012,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   0 },     29.41445,   0.00178,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   0 },     31.07972,   0.03351,     0.080 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   2 },    210.25810,   0.00008,     0.069 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -2 },     32.13632,   0.00009,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },    210.97968,   0.01002,     1.203 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   2 },     34.68286,   0.00010,     0.036 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,  -2 },     31.40277,   0.00006,     0.078 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   0 },    211.44985,   0.02756,     0.071 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   2,   0 },    211.32242,   0.00149,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   3,   0 },    215.28757,   0.00007,     0.025 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,  -1 },    190.52317,   0.00008,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   0 },     38.83535,   0.00018,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   1 },    190.52411,   0.00025,     8.998 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,  -1 },    190.52322,   0.00146,     0.074 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   0 },     33.53642,   0.00039,    17.980 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   1 },     10.52560,   0.00393,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -1 },    190.52643,   0.00028,     4.497 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },    216.21847,   0.00641,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,  -1 },     10.52401,   0.00151,     0.077 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   0 },    216.09015,   0.00034,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   2,  -1 },     10.52321,   0.00008,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -3,   0 },    357.66711,   0.00003,     0.070 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -2,   0 },     42.25914,   0.00033,     1.061 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,  -2 },     44.61502,   0.00003,     0.069 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },    211.32540,   0.01569,     0.081 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },    214.95659,   0.01035,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   1,   0 },    213.54295,   0.00087,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   2,   0 },    215.06090,   0.00007,     0.019 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,  -1 },     10.51848,   0.00002,     0.501 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,   0 },    215.50376,   0.00002,     0.088 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -2,   1 },     10.52411,   0.00004,     0.040 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,  -1 },    190.48123,   0.00008,     0.089 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,   0 },    215.84452,   0.00011,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,  -1,   1 },     10.52411,   0.00004,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,  -1 },     10.53927,   0.00014,     0.041 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   3,   0,   0 },    215.73448,   0.00010,     0.026 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -3,   0 },     13.91514,   0.00015,     0.095 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -2,   0 },    199.83097,   0.00037,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,  -1,   0 },    201.05210,   0.00093,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   4,   0,   0 },    204.41156,   0.00021,     0.020 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   6,  -2,   0 },    191.21440,   0.00003,     0.021 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,  -2,   0,   0 },    272.97698,   0.00011,     0.042 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,  -2,   1,   0 },    272.97698,   0.00004,     0.094 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   0,  -1,   0 },    272.79962,   0.00003,     0.081 },
 { {   0,   0,   1,   0,  -2,   1,   0,   0,   0,   1,   0 },     92.79962,   0.00003,     0.071 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,  -2,   0,   0 },    118.72600,   0.00010,     0.042 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,  -2,   1,   0 },    118.72600,   0.00009,     0.094 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   0,  -1,   0 },    112.40980,   0.00008,     0.081 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   0,   1,   0 },    292.40980,   0.00008,     0.071 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   1,   0,   0 },    101.06371,   0.00007,     0.075 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   2,  -3,   0 },    278.86332,   0.00002,     0.071 },
 { {   0,   0,   1,   0,  -2,   2,   0,   0,   2,  -1,   0 },    278.86332,   0.00002,     0.081 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,  -2,   0,   0 },     78.12065,   0.00012,     0.042 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,  -2,   1,   0 },     78.12065,   0.00004,     0.095 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,  -1,   0 },     78.58619,   0.00003,     0.081 },
 { {   0,   0,   1,   0,  -2,   4,   0,   0,   0,   1,   0 },    258.58619,   0.00003,     0.070 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,  -1,   0 },    224.48834,   0.00004,     0.027 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   0,   0 },    284.74502,   0.00036,     0.042 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   1,   0 },    341.54281,   0.00047,     0.095 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   2,   0 },    181.37118,   0.00005,     0.360 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -2,   0 },    180.15314,   0.00004,     0.039 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,  -1,   0 },    199.77547,   0.00037,     0.082 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   0,   0 },    127.23385,   0.00007,     0.999 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   1,   0 },    164.98779,   0.00036,     0.070 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   2,   0 },    181.10212,   0.00003,     0.036 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,  -2,   0 },    250.19627,   0.00002,     1.294 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   2,  -1,   0 },    346.84455,   0.00007,     0.080 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,  -2,   0,   0 },     48.20820,   0.00008,     0.042 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,  -2,   1,   0 },     48.20777,   0.00003,     0.093 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   0,  -1,   0 },     45.89734,   0.00003,     0.080 },
 { {   0,   0,   1,   0,  -1,  -5,   0,   0,   0,   1,   0 },    225.89734,   0.00003,     0.071 },
 { {   0,   0,   1,   0,  -1,  -2,   0,   0,  -2,   0,   0 },    197.01946,   0.00005,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   1,   0 },     91.29682,   0.00005,     0.017 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -6,   2,   0 },     91.57135,   0.00011,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -3,   0 },     91.89589,   0.00002,     0.011 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -2,   0 },    272.07804,   0.00002,     0.013 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,  -1,   0 },     90.34095,   0.00029,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,   0 },     93.03367,   0.00159,     0.021 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   0,   2 },     91.78676,   0.00013,     0.046 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,   0 },     92.53414,   0.00401,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   1,   2 },     91.80013,   0.00008,     0.118 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   2,   0 },     92.35977,   0.00181,     0.045 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -4,   3,   0 },     91.07113,   0.00006,     0.114 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,  -1,   0 },     92.08174,   0.00006,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   0,   0 },     91.67754,   0.00019,     0.028 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   1,   0 },    272.43881,   0.00036,     0.044 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -3,   2,   0 },    272.42820,   0.00003,     0.103 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -3,   0 },     91.68026,   0.00006,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -2,   0 },     91.54481,   0.00077,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,   0 },     92.06684,   0.01279,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,  -1,   2 },    271.76354,   0.00012,     0.098 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },     91.79862,   0.20497,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   2 },    271.75850,   0.00075,     0.331 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,  -2 },    271.74358,   0.00008,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,   0 },     91.74578,   0.15751,     0.095 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   1,   2 },    271.64003,   0.00040,     0.061 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,  -2 },    271.65165,   0.00003,     0.041 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,   0 },    271.66304,   0.01034,     0.372 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   2,   2 },    271.66200,   0.00003,     0.034 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   3,   0 },    272.12848,   0.00056,     0.063 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   4,   0 },    271.79696,   0.00003,     0.034 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -2,   0 },     90.69167,   0.00005,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0 },     90.52859,   0.00066,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,  -1,   1 },      6.69191,   0.00004,     0.089 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,  -1 },      5.80078,   0.00002,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0 },     90.58108,   0.00793,     0.087 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,  -1 },      6.53228,   0.00002,     0.086 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   0 },    270.61178,   0.00121,     0.555 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -1,   2,   0 },    270.34958,   0.00006,     0.066 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -3,   0 },     88.95597,   0.00042,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,   0 },     91.25048,   0.00682,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -2,   2 },     91.25718,   0.00004,     0.801 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,  -2 },     91.62354,   0.00002,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   0 },     91.22751,   0.12666,     0.081 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,   2 },    270.77724,   0.00036,     0.069 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -2 },     90.88150,   0.00009,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },    271.22625,   0.04277,     1.092 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2 },     90.32807,   0.00012,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,  -2 },     88.53725,   0.00017,     0.079 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   0 },    271.06726,   0.10504,     0.071 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,  -2 },    271.28916,   0.00004,     1.718 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   0 },    271.01187,   0.00569,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   3,   0 },    271.06018,   0.00032,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -3,   0 },    270.85726,   0.00003,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -2,   0 },    270.85724,   0.00052,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,  -1 },    186.64287,   0.00019,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   0 },    270.69831,   0.00340,    34.860 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,  -1,   1 },      6.64800,   0.00019,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -2 },    270.71019,   0.00024,     0.074 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   0 },     90.70888,   0.06209,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   1 },    187.38991,   0.00004,     0.037 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,  -1 },      6.64287,   0.00019,     0.076 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   1,   0 },     90.71613,   0.00343,     0.038 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,   0 },     90.71019,   0.00021,     0.025 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    273.14810,   0.00006,     0.071 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -2,  -2 },    271.62242,   0.00002,     0.036 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -2,   0 },     89.53913,   0.00099,     1.165 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,  -2 },     91.05326,   0.00012,     0.069 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    270.55045,   0.04366,     0.081 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,  -2 },     91.01771,   0.00012,     0.839 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },    270.46442,   0.03875,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,  -2 },     87.14531,   0.00003,     0.083 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   1,   0 },    269.97364,   0.00319,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   2,   2,   0 },    271.11594,   0.00032,     0.019 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -3,   0 },     90.74379,   0.00004,     0.573 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -2,   0 },    270.72130,   0.00051,     0.087 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,  -1 },    186.53228,   0.00002,     0.088 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,  -1,   0 },     90.61843,   0.00035,     0.040 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,  -1 },      6.79214,   0.00002,     0.041 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   0,   0 },     90.93586,   0.00039,     0.026 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   3,   1,   0 },     90.71016,   0.00007,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -3,   0 },     91.44863,   0.00002,     0.094 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -2,   0 },    270.88771,   0.00044,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,  -2 },    270.24845,   0.00002,     0.097 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,  -1,   0 },    270.71172,   0.00082,     0.027 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,  -2 },    270.24713,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   0,   0 },    271.48811,   0.00044,     0.020 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   1,   0 },    270.19324,   0.00006,     0.016 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   4,   2,   0 },     90.54069,   0.00002,     0.013 },
 { {   0,   0,   1,   0,  -1,   1,   0,   0,  -2,   0,   0 },    315.45822,   0.00007,     0.042 },
 { {   0,   0,   1,   0,  -1,   2,   0,   0,  -2,   0,   0 },    140.24296,   0.00006,     0.042 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   0,   0 },     44.82468,   0.00024,     0.042 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   1,   0 },     45.26176,   0.00014,     0.094 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,  -1,   0 },     41.28547,   0.00019,     0.080 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   0,   0 },    220.70699,   0.00003,     1.204 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   1,   0 },    220.60813,   0.00016,     0.071 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   1,   0,   0 },    219.96769,   0.00003,     0.076 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   2,  -1,   0 },    204.92664,   0.00007,     0.081 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   2,   0,   0 },    224.84796,   0.00008,     0.039 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,  -2,   0,   0 },    198.78024,   0.00009,     0.042 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,  -2,   1,   0 },    198.78018,   0.00003,     0.094 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   0,  -1,   0 },    199.31978,   0.00002,     0.081 },
 { {   0,   0,   1,   0,   0,  -4,   0,   0,   0,   1,   0 },     19.31978,   0.00002,     0.071 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,  -1,   0 },    106.79985,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   0,   0 },    106.61589,   0.00044,     0.042 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   1,   0 },    107.07208,   0.00024,     0.094 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -1,   0,   0 },    107.01472,   0.00006,     0.087 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,  -1,   0 },    106.82618,   0.00033,     0.081 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0 },    287.20441,   0.00010,     1.113 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   1,   0 },    286.78334,   0.00027,     0.071 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,  -1 },    296.53066,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,  -1,   1 },    116.53071,   0.00006,     0.074 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   0,   0 },    107.01742,   0.00034,     0.075 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   1,   1,  -1 },    116.53066,   0.00004,     0.076 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   2,  -1,   0 },    287.11865,   0.00007,     0.081 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   2,   0,   0 },    286.99449,   0.00014,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   0,   0 },     16.17964,   0.00003,     0.021 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   1,   0 },     16.08012,   0.00007,     0.028 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -4,   2,   0 },     15.77198,   0.00002,     0.045 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,  -1,   0 },     15.96450,   0.00022,     0.027 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,   0 },     15.44598,   0.00301,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   1,   0 },     15.92074,   0.00229,     0.095 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   2,   0 },    196.03443,   0.00014,     0.370 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,  -1,   1 },     23.92219,   0.00005,     0.089 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,  -1 },     23.92206,   0.00004,     0.040 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,   0 },     16.30358,   0.00006,     0.087 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   0,   1 },     23.92220,   0.00003,     0.504 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   1,  -1 },     23.92208,   0.00004,     0.086 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -1,   1,   1 },     23.92221,   0.00002,     0.066 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -2,   0 },     15.60886,   0.00013,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,  -1,   0 },     15.63019,   0.00200,     0.081 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0 },    196.79617,   0.00072,     1.073 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   1,   0 },    195.68549,   0.00180,     0.070 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   2,   0 },    195.63199,   0.00009,     0.036 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,  -1 },    203.92217,   0.00036,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   0 },    196.20551,   0.00003,    22.176 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,  -1,   1 },     23.92219,   0.00028,     0.074 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,  -1 },    203.92222,   0.00004,     8.220 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   0 },     16.27296,   0.00061,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   1 },    203.92230,   0.00007,     0.037 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,  -1 },     23.92217,   0.00036,     0.076 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   1,   0 },     16.27296,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -2,   0 },    195.86299,   0.00004,     1.188 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,  -1,   0 },    196.67405,   0.00044,     0.081 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   0,   0 },    196.17179,   0.00050,     0.039 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   2,   1,   0 },    196.50921,   0.00003,     0.026 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,  -1,  -1 },    203.92209,   0.00002,     0.088 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   3,   0,  -1 },     23.92222,   0.00004,     0.041 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -6,   2,   0 },     90.69105,   0.00002,     0.021 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   0,   0 },     90.27659,   0.00013,     0.021 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   1,   0 },     90.67478,   0.00037,     0.028 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -4,   2,   0 },     90.64961,   0.00009,     0.045 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -3,   0,   0 },     90.84442,   0.00004,     0.028 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -2,   0 },     90.68072,   0.00005,     0.020 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,  -1,   0 },     90.63667,   0.00072,     0.027 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,   0 },     90.74879,   0.01179,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   0,   2 },    270.69071,   0.00003,     0.325 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   1,   0 },     90.68998,   0.00895,     0.095 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   2,   0 },    270.73415,   0.00058,     0.365 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -2,   3,   0 },    270.71592,   0.00003,     0.063 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,  -1,   0,   0 },     89.76013,   0.00008,     0.088 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   0 },     90.46546,   0.00041,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   0 },     90.45410,   0.00761,     0.081 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   2 },    270.46461,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0 },    270.42037,   0.00274,     1.035 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,  -2 },     90.46461,   0.00002,     0.079 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   1,   0 },    270.39594,   0.00637,     0.070 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   2,   0 },    270.38115,   0.00031,     0.036 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -2,   0 },    271.80192,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,  -1 },    260.10657,   0.00007,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,   0 },    270.29648,   0.00008,    12.652 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,  -1,   1 },     80.18593,   0.00003,     0.074 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   0 },     90.31837,   0.00144,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,  -1 },     80.10657,   0.00007,     0.076 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   1,   0 },     90.31837,   0.00008,     0.038 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -2,   0 },     89.06595,   0.00002,     1.238 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,  -1,   0 },    270.25501,   0.00197,     0.080 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   0,   0 },    270.07998,   0.00219,     0.039 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   2,   1,   0 },    270.33045,   0.00026,     0.026 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   4,  -2,   0 },    270.38304,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   4,  -1,   0 },    270.18781,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   0,   0,  -3,   0,  -2,   0,   0 },    182.89549,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   0,   0 },      6.99875,   0.00015,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   1,   0 },      6.99875,   0.00013,     0.095 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,  -1,   0 },      7.01513,   0.00013,     0.081 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0 },    187.00400,   0.00003,     1.024 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0 },    187.01513,   0.00013,     0.070 },
 { {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0 },    186.82619,   0.00005,     0.039 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0 },     83.67648,   0.00031,     0.042 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   1,   0 },     83.42781,   0.00017,     0.095 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0 },     85.08740,   0.00020,     0.082 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0 },    266.45524,   0.00004,     1.012 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0 },    265.40732,   0.00016,     0.070 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -1,   0 },    279.91543,   0.00003,     0.080 },
 { {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0 },    269.01676,   0.00006,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   0,   0 },     13.59447,   0.00022,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   1,   0 },     13.59447,   0.00008,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,  -1,   0 },     13.58574,   0.00008,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0 },    193.58574,   0.00008,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0 },     91.69966,   0.00022,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -2,   1,   0 },     91.69966,   0.00008,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0 },     91.37850,   0.00008,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0 },    271.37850,   0.00008,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   0 },    158.46671,   0.00005,     0.021 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   1,   0 },    155.50381,   0.00011,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   2,   0 },    156.70181,   0.00003,     0.046 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    159.54341,   0.00031,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.83806,   0.00377,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    153.96299,   0.00295,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },    327.35807,   0.00018,     0.360 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0 },      3.72070,   0.00003,     0.062 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   1 },      5.80020,   0.00006,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    176.19382,   0.00007,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    182.88601,   0.00008,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    181.57498,   0.00017,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    179.67637,   0.00032,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   1 },    188.91140,   0.00011,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0 },    180.15013,   0.00017,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.63649,   0.00301,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },     11.13590,   0.00008,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },      5.67716,   0.00126,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },     13.35181,   0.00007,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    358.64832,   0.00260,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0 },    359.15168,   0.00014,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    355.69246,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   1 },     16.72666,   0.00003,     3.575 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,  -1 },      4.99466,   0.00090,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0 },    354.88227,   0.00006,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   1 },    186.15699,   0.00018,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    172.91244,   0.00009,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    174.80301,   0.00110,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    187.97450,   0.00012,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -1 },    183.72320,   0.00089,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0 },    174.80301,   0.00006,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   2,  -1 },    175.44886,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },     60.06007,   0.00013,     1.292 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    287.84670,   0.00210,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },      8.01397,   0.00003,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    314.08649,   0.00136,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   1,   0 },    319.86989,   0.00013,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,  -1 },      8.00606,   0.00020,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,  -1 },      1.58522,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -2,   0 },    307.30162,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -1,   0 },    307.39051,   0.00003,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   1,  -2,  -2,   0,   0 },    253.79136,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0 },    173.74899,   0.00005,     0.042 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,  -1,   0 },    255.71137,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   0,   0 },    255.32373,   0.00054,     0.042 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -2,   1,   0 },    257.46336,   0.00021,     0.096 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   0,  -1 },     64.27632,   0.00007,     0.040 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,  -1,   1,  -1 },    244.06069,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,  -1,   0 },    254.42443,   0.00017,     0.082 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0 },     75.11987,   0.00009,     0.967 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,   1,   0 },     81.04119,   0.00007,     0.070 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -2,  -1 },     64.14483,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,  -1 },     64.18174,   0.00091,     0.075 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,  -1,   1 },    244.21352,   0.00008,     0.074 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   1,  -1 },    244.18174,   0.00091,     0.076 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   2,  -1 },    244.14483,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   2,  -1,   0 },    256.02318,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   2,   0,   0 },    254.51034,   0.00005,     0.039 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,  -1,  -1 },     64.06069,   0.00007,     0.087 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   3,   0,  -1 },    244.27632,   0.00007,     0.040 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -2,   0,   0 },    153.28849,   0.00005,     0.042 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,  -1,   1,  -1 },    146.92754,   0.00002,     0.087 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,  -1,  -1 },    326.72933,   0.00015,     0.076 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   1,  -1 },    146.72933,   0.00015,     0.075 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   3,  -1,  -1 },    326.92754,   0.00002,     0.087 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,  -1,   0 },    101.68972,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,   0 },    101.69157,   0.00054,     0.042 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   1,   0 },    101.59299,   0.00031,     0.095 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,  -1,   0 },    101.72335,   0.00036,     0.081 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   0,   0 },    281.13632,   0.00012,     1.093 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   1,   0 },    281.73959,   0.00029,     0.071 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   1,   0,   0 },    101.20783,   0.00017,     0.075 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,  -1,   0 },    278.45194,   0.00009,     0.081 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   2,   0,   0 },    280.34323,   0.00017,     0.039 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,  -2,   0,   0 },     38.90079,   0.00010,     0.042 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,  -2,   1,   0 },     38.90079,   0.00008,     0.095 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   0,  -1,   0 },     38.30631,   0.00008,     0.081 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   0,   1,   0 },    218.30631,   0.00008,     0.070 },
 { {   0,   0,   1,   0,   1,  -3,   0,   0,   2,   0,   0 },    215.22850,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,  -1,   0 },    273.22667,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   0,   0 },    273.14827,   0.00054,     0.042 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,  -2,   1,   0 },    273.33414,   0.00027,     0.096 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,  -1,   0 },    271.99854,   0.00035,     0.082 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   0,   0 },     91.35868,   0.00012,     0.984 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   0,   1,   0 },     91.72796,   0.00029,     0.070 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   2,  -1,   0 },     81.44023,   0.00003,     0.080 },
 { {   0,   0,   1,   0,   1,  -2,   0,   0,   2,   0,   0 },     89.15559,   0.00013,     0.039 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -2,   0,   0 },    156.36315,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   0,  -1 },      3.73523,   0.00007,     0.040 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,  -1,   1,  -1 },    184.19933,   0.00005,     0.087 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -2,  -1 },      3.99044,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,  -1,  -1 },      3.91123,   0.00080,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   1,  -1 },    183.91123,   0.00080,     0.075 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   2,  -1 },    183.99044,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   3,  -1,  -1 },      4.19933,   0.00005,     0.087 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   3,   0,  -1 },    183.73523,   0.00007,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,  -1,   0 },    157.99711,   0.00002,     0.016 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   0,   0 },    336.46711,   0.00009,     0.021 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   1,   0 },    338.31018,   0.00025,     0.028 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -4,   2,   0 },    336.37335,   0.00006,     0.046 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -2,   0 },    334.90207,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,  -1,   0 },    334.07808,   0.00036,     0.027 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   0 },    334.67637,   0.00500,     0.042 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   2 },    155.11045,   0.00003,     0.313 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   1,   0 },    337.41182,   0.00348,     0.096 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   2,   0 },    162.51848,   0.00022,     0.350 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,  -1,  -1 },    165.53653,   0.00005,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,  -1 },    165.54455,   0.00044,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   0,   1 },    165.50862,   0.00004,     0.468 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   1,  -1 },    345.82368,   0.00021,     0.087 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   1,   1 },    165.50863,   0.00003,     0.065 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -1,   2,  -1 },    165.42676,   0.00006,     0.554 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -2,   0 },    318.13269,   0.00011,     0.039 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,  -1,   0 },    317.54448,   0.00165,     0.082 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0 },    145.72110,   0.00057,     0.922 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   1,   0 },    149.70040,   0.00104,     0.070 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   2,   0 },    156.09773,   0.00004,     0.036 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -2,  -1 },    165.57727,   0.00021,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,  -1 },    165.57727,   0.00393,     0.076 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,  -1,   1 },    165.69217,   0.00043,     0.073 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   0 },    308.92839,   0.00007,     0.074 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   1 },    345.01278,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   1,  -1 },    345.57727,   0.00393,     0.075 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   2,  -1 },    345.57727,   0.00021,     0.038 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,  -2,   0 },    140.32880,   0.00004,     1.450 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },    332.71833,   0.00011,     0.080 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,   0,   0 },      9.06968,   0.00008,     0.039 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   2,   1,   0 },    336.39100,   0.00006,     0.026 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -2,  -1 },    345.62742,   0.00006,     0.574 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,  -1,  -1 },    165.73468,   0.00027,     0.087 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   0,  -1 },    345.47501,   0.00034,     0.040 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   3,   1,  -1 },    345.53653,   0.00005,     0.026 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,  -2,   0,   0 },    203.76645,   0.00008,     0.042 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   0,  -1,   0 },    203.70714,   0.00002,     0.081 },
 { {   0,   0,   1,   0,   2,  -7,   0,   0,   0,   1,   0 },     23.70714,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,  -2,   0,   0 },    281.22542,   0.00015,     0.042 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,  -2,   1,   0 },    281.22542,   0.00005,     0.095 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,  -1,   0 },    281.18315,   0.00004,     0.081 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   0,   0 },    101.12774,   0.00004,     1.036 },
 { {   0,   0,   1,   0,   2,  -6,   0,   0,   0,   1,   0 },    101.18315,   0.00004,     0.070 },
 { {   0,   0,   1,   0,   2,  -6,   3,   0,  -2,   0,   0 },    347.32753,   0.00005,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   0,   0 },    155.69358,   0.00005,     0.021 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   1,   0 },    156.92934,   0.00017,     0.028 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -4,   2,   0 },    155.61946,   0.00005,     0.046 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -2,   0 },    156.88348,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,  -1,   0 },    157.28320,   0.00028,     0.027 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   0 },    157.19506,   0.00520,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   2 },    336.96168,   0.00004,     0.322 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,   0 },    156.58650,   0.00429,     0.095 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   2,   0 },    336.42606,   0.00030,     0.361 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -1,   0,  -1 },    149.05563,   0.00004,     0.040 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -2,   0 },    156.51051,   0.00017,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,   0 },    155.85968,   0.00293,     0.082 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   0,   0 },    335.52942,   0.00121,     1.001 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,   0 },    335.34682,   0.00306,     0.070 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   2,   0 },    334.98094,   0.00016,     0.036 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   0 },    155.35196,   0.00044,     0.075 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   1 },    329.05606,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,  -1,   0 },    336.61595,   0.00045,     0.080 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,   0 },    333.06619,   0.00079,     0.039 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   1,   0 },    334.58985,   0.00011,     0.026 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,  -2,   0,   0 },     13.38647,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,  -1,  -1 },     52.38408,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   2,  -4,   0,   0,   1,   1,  -1 },    232.38408,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   2,  -3,   0,   0,  -2,   0,   0 },    273.77514,   0.00007,     0.042 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,  -2,   0,   0 },    330.32391,   0.00010,     0.042 },
 { {   0,   0,   1,   0,   2,  -2,   0,   0,  -2,   1,   0 },    330.32391,   0.00003,     0.096 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -2,   0,   0 },    338.30073,   0.00023,     0.042 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,  -2,   1,   0 },    352.95657,   0.00005,     0.097 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   0,  -1,   0 },    316.86425,   0.00007,     0.083 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   0,   1,   0 },    149.60323,   0.00003,     0.069 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,  -2,   0,   0 },     59.05094,   0.00006,     0.042 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,  -1,  -1 },    260.12392,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   3,  -5,   0,   0,   1,   1,  -1 },     80.12392,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   4, -10,   0,   0,  -2,   0,   0 },    348.14034,   0.00007,     0.042 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0 },      4.12870,   0.00027,     0.043 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,   1,   0 },      7.34054,   0.00007,     0.101 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0 },    340.92234,   0.00012,     0.085 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0 },    155.88880,   0.00004,     0.653 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   1,   0 },    166.44133,   0.00008,     0.068 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -1,   0 },    330.93604,   0.00025,     0.077 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,   0,   0 },    328.65531,   0.00003,     0.038 },
 { {   0,   0,   1,   2,  -3,   0,   0,   0,   2,  -1,   0 },    154.51699,   0.00004,     0.075 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   0,   0 },    127.20384,   0.00016,     0.044 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   1,   0 },    136.23804,   0.00005,     0.106 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   0,  -1 },    150.78506,   0.00006,     0.042 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -1,   1,  -1 },    330.85840,   0.00003,     0.095 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,  -1,  -1 },    150.81142,   0.00006,     0.082 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   1,  -1 },    150.81142,   0.00005,     0.070 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   0,  -1,   0 },    223.01858,   0.00003,     0.066 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   0,   0,   0 },    222.70909,   0.00004,     0.500 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   0,   1,   0 },     42.48240,   0.00007,     0.089 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,  -1,   0 },     43.03684,   0.00018,     0.105 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   0,   0 },     42.64012,   0.00072,     0.044 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   2,   1,   0 },     42.69438,   0.00006,     0.028 },
 { {   0,   0,   2,  -8,   3,   0,   0,   0,   4,   0,   0 },    222.68513,   0.00004,     0.021 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   1,   0 },    148.21025,   0.00002,     0.079 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   3,   0 },    328.21025,   0.00002,     0.072 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,  -1,   0 },     99.12668,   0.00003,     0.069 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,   1,   0 },    279.12668,   0.00003,     0.083 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,  -1,   0 },    287.03533,   0.00008,     0.097 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,   0,   0 },    285.99615,   0.00016,     0.042 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   2,  -1,   0 },    343.92457,   0.00003,     0.096 },
 { {   0,   0,   2,  -6,   2,   0,   0,   0,   2,   0,   0 },    343.92457,   0.00010,     0.042 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   0,   0 },    323.16393,   0.00006,     0.039 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   1,   0 },    186.56291,   0.00002,     0.082 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   1,  -1 },    261.96109,   0.00008,     0.076 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0 },      4.01895,   0.00018,     0.072 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0 },     18.09243,   0.00004,     1.519 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0 },    185.98450,   0.00018,     0.079 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,  -1,   0 },    270.99668,   0.00009,     0.092 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,   0,   0 },    272.96353,   0.00016,     0.042 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   0,   0 },     30.15596,   0.00008,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   1,   0 },     31.85300,   0.00027,     0.028 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -4,   2,   0 },     33.84388,   0.00006,     0.043 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -3,   0,   0 },     30.69936,   0.00003,     0.027 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -2,   0 },     32.27718,   0.00003,     0.019 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,   0 },     30.96842,   0.00062,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   0 },     30.56336,   0.00861,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,   0 },     32.46000,   0.00834,     0.086 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   2,   0 },    215.09648,   0.00035,     0.607 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   0 },    211.94270,   0.00012,     0.080 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -3,   0 },     27.75206,   0.00005,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0 },     27.84663,   0.00083,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0 },     27.79153,   0.01596,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2 },    207.74734,   0.00006,     0.074 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0 },     29.98443,   0.00086,     7.890 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,  -2 },     27.74734,   0.00006,     0.073 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0 },    208.07736,   0.01622,     0.076 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0 },    207.79522,   0.00086,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   3,   0 },    207.61107,   0.00005,     0.025 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,   0 },     22.48236,   0.00028,     0.082 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -2,   0 },     35.69339,   0.00050,     0.526 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,  -2 },     32.14626,   0.00002,     0.065 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   0 },    213.40837,   0.01021,     0.088 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,  -2 },     32.65700,   0.00003,     0.448 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   0 },    212.72441,   0.01097,     0.041 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,  -2 },     32.82087,   0.00002,     0.091 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,   0 },    211.63225,   0.00077,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   2,   0 },    212.74301,   0.00006,     0.020 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   3,   0,   0 },    213.30441,   0.00004,     0.027 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -2,   0 },    213.94064,   0.00008,     0.044 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,  -1,   0 },    212.82677,   0.00021,     0.028 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   4,   0,   0 },    212.68002,   0.00011,     0.020 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,  -2,   0,   0 },    245.38481,   0.00015,     0.040 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,  -2,   1,   0 },    245.84589,   0.00007,     0.087 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   0,  -1,   0 },    243.66681,   0.00013,     0.075 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   0,   1,   0 },     63.66681,   0.00013,     0.076 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   2,  -1,   0 },     66.23625,   0.00007,     0.087 },
 { {   0,   0,   2,  -4,   1,   0,   0,   0,   2,   0,   0 },     65.59135,   0.00016,     0.040 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   0,   0 },     90.55363,   0.00024,     0.040 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,  -2,   1,   0 },     90.96230,   0.00020,     0.087 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,  -1,   0 },     89.32228,   0.00034,     0.076 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   0,   1,   0 },    269.13527,   0.00030,     0.075 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,  -1,   0 },    270.81488,   0.00017,     0.087 },
 { {   0,   0,   2,  -4,   2,   0,   0,   0,   2,   0,   0 },    270.45502,   0.00022,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   0,   0 },     60.87501,   0.00009,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   1,   0 },     60.85578,   0.00030,     0.028 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -4,   2,   0 },     60.72542,   0.00008,     0.044 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -3,   0,   0 },     61.11865,   0.00004,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -2,   0 },     62.35818,   0.00004,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -1,   0 },     60.66542,   0.00057,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   0 },     61.05210,   0.00911,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   2 },    240.80293,   0.00002,     0.398 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,   0 },     60.94047,   0.00844,     0.090 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   2,   0 },    241.17101,   0.00050,     0.459 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   3,   0 },    240.87435,   0.00003,     0.065 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   0 },    240.67447,   0.00017,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0 },     59.78571,   0.00040,     0.038 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0 },     59.83434,   0.00768,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   2 },    239.79971,   0.00002,     0.071 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0 },    240.32130,   0.00120,     2.470 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0 },    239.73941,   0.00702,     0.073 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0 },    239.73125,   0.00037,     0.037 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,  -1,   1 },    143.38522,   0.00006,     0.077 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,   0 },    240.92253,   0.00008,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -2,   0 },     61.10884,   0.00026,     0.730 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    241.13613,   0.00536,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,  -2 },     60.59385,   0.00003,     0.587 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,   0 },    240.55894,   0.00458,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   1,   0 },    240.46332,   0.00035,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,  -2,   0 },    240.80858,   0.00006,     0.043 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,  -1,   0 },    240.62278,   0.00017,     0.027 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   4,   0,   0 },    240.78584,   0.00005,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,  -1,   0 },     91.02514,   0.00007,     0.016 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,   0 },     86.92302,   0.00043,     0.021 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   0,   2 },     88.77042,   0.00005,     0.046 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   1,   0 },     87.32136,   0.00106,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -4,   2,   0 },     87.94038,   0.00050,     0.045 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -3,   0,   0 },     88.17213,   0.00018,     0.028 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -3,   1,   0 },    268.26192,   0.00009,     0.044 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -2,   0 },     89.10137,   0.00026,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,  -1,   0 },     87.97493,   0.00356,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,  -2 },    268.59973,   0.00002,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   0 },     88.57600,   0.05723,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   2 },    268.89112,   0.00025,     0.329 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,  -2 },    268.81148,   0.00002,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   0 },     88.87113,   0.04396,     0.095 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   2 },    269.00534,   0.00013,     0.061 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   2,   0 },    269.15356,   0.00287,     0.369 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   3,   0 },    268.30078,   0.00013,     0.063 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,  -1,   0 },     90.57349,   0.00006,     0.041 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   0,   0 },     90.37955,   0.00111,     0.087 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -1,   1,   0 },    270.34040,   0.00020,     0.548 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0 },     89.61458,   0.00013,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0 },     89.57018,   0.00183,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0 },     89.53531,   0.03726,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   2 },    270.69668,   0.00008,     0.069 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2 },     90.05491,   0.00012,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0 },    269.74341,   0.01329,     1.068 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2 },     89.66223,   0.00013,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -2 },     92.09324,   0.00007,     0.079 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0 },    270.14383,   0.03116,     0.070 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0 },    270.26800,   0.00174,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0 },    269.96723,   0.00009,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -4,   0 },    270.23891,   0.00002,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -2,   0 },    270.49158,   0.00017,     0.076 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,  -1 },    189.34031,   0.00013,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   0 },    270.23070,   0.00063,    20.152 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,  -1,   1 },      9.57005,   0.00009,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -2 },    270.23488,   0.00004,     0.074 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   0 },     90.27482,   0.01112,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,  -1 },      9.34031,   0.00013,     0.076 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   1,   0 },     90.23489,   0.00062,     0.038 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   2,   0 },     90.23891,   0.00002,     0.025 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -3,   0 },    272.26882,   0.00002,     0.071 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -2,   0 },     92.25252,   0.00017,     1.194 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,  -2 },     89.49930,   0.00003,     0.069 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,   0 },    270.49242,   0.01117,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -2 },     89.79128,   0.00002,     0.854 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   0 },    270.95781,   0.01092,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   1,   0 },    271.13134,   0.00090,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   2,   0 },    269.68448,   0.00008,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -2,   0 },    270.43176,   0.00004,     0.087 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,  -1,   0 },     90.21873,   0.00009,     0.040 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   3,   0,   0 },     89.39234,   0.00005,     0.026 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -2,   0 },    269.86234,   0.00010,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -1,   0 },    270.23299,   0.00033,     0.027 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,   0,   0 },    270.45035,   0.00010,     0.020 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,   1,   0 },     90.71428,   0.00003,     0.016 },
 { {   0,   0,   2,  -2,   0,   2,   0,   0,  -2,   0,   0 },     20.92236,   0.00004,     0.042 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,  -1,   0 },    346.28351,   0.00005,     0.027 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,   0 },    344.12490,   0.00065,     0.043 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   1,   0 },    348.86065,   0.00019,     0.100 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0 },    325.02762,   0.00033,     0.085 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0 },    139.87725,   0.00017,     0.681 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0 },    144.69868,   0.00019,     0.068 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,   0 },    308.60120,   0.00049,     0.077 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,   0,   0 },    304.15006,   0.00008,     0.038 },
 { {   0,   0,   2,   0,  -7,  10,   0,   0,   2,  -3,   0 },    337.91736,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -7,  10,   0,   0,   2,  -1,   0 },    337.91736,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,  -2,   0,   0 },    308.52001,   0.00008,     0.043 },
 { {   0,   0,   2,   0,  -6,   0,   0,   0,   2,  -1,   0 },    311.06395,   0.00005,     0.077 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,  -2,   0,   0 },    170.01952,   0.00009,     0.043 },
 { {   0,   0,   2,   0,  -6,   5,   0,   0,   2,  -1,   0 },    169.07267,   0.00018,     0.076 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   0,   0 },    299.14380,   0.00034,     0.043 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,  -2,   1,   0 },    299.00364,   0.00014,     0.101 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,  -1,   0 },    299.97844,   0.00029,     0.086 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 },    120.61470,   0.00012,     0.634 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,   1,   0 },    120.60147,   0.00014,     0.067 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -2,   0 },    119.58226,   0.00004,     5.114 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,  -1,   0 },    299.49589,   0.00071,     0.077 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   2,   0,   0 },    300.05941,   0.00012,     0.038 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,  -2,   0,   0 },    280.25919,   0.00013,     0.043 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   0,  -1,   0 },    281.19971,   0.00006,     0.086 },
 { {   0,   0,   2,   0,  -5,   2,   0,   0,   2,  -1,   0 },    281.48548,   0.00022,     0.076 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,  -2,   0,   0 },    304.00700,   0.00007,     0.043 },
 { {   0,   0,   2,   0,  -5,   3,   0,   0,   2,  -1,   0 },    302.10662,   0.00015,     0.076 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,  -2,   0,   0 },    154.01451,   0.00004,     0.043 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   2,  -3,   0 },    163.79644,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -5,   4,   0,   0,   2,  -1,   0 },    148.28505,   0.00015,     0.076 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,  -2,   0,   0 },    146.73169,   0.00019,     0.043 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,  -2,   1,   0 },    146.74021,   0.00009,     0.103 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,  -1,   0 },    144.46457,   0.00069,     0.087 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,   0,   0 },    324.76592,   0.00011,     0.572 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,   1,   0 },    326.54293,   0.00007,     0.067 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -3,   0 },    336.91498,   0.00004,     0.075 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -2,   0 },    323.75442,   0.00019,    38.763 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -1,   0 },    144.21774,   0.00377,     0.076 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,   0,   0 },    144.47008,   0.00022,     0.038 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -3,   0 },    141.98697,   0.00004,     0.087 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -2,   0 },    144.15554,   0.00008,     0.040 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   4,  -1,   0 },    144.20706,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -4,   0,   0 },    289.88818,   0.00003,     0.021 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -4,   1,   0 },    289.81956,   0.00007,     0.029 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    289.84226,   0.00022,     0.028 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   0 },    289.77206,   0.00301,     0.043 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   1,   0 },    289.80253,   0.00176,     0.102 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   2,   0 },    109.65110,   0.00012,     0.291 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -2,   0 },    290.31309,   0.00017,     0.040 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -1,   0 },    290.27869,   0.00337,     0.086 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 },    111.55859,   0.00105,     0.601 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   1,   0 },    110.94492,   0.00113,     0.067 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   2,   0 },    110.78655,   0.00007,     0.035 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   1,   0,   0 },    291.43218,   0.00004,     0.071 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -3,   0 },    109.92683,   0.00011,     0.075 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -2,   0 },    110.09574,   0.00069,     8.990 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,  -2 },    109.94000,   0.00004,     0.073 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   0 },    289.76123,   0.01191,     0.076 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   0,   0 },    290.46739,   0.00089,     0.038 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,   1,   0 },    290.15906,   0.00008,     0.025 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -3,   0 },    289.76512,   0.00003,     0.088 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -2,   0 },    289.81564,   0.00016,     0.041 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   4,  -1,   0 },    289.93987,   0.00010,     0.026 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,  -2,   0,   0 },    105.30412,   0.00004,     0.043 },
 { {   0,   0,   2,   0,  -4,   1,   0,   0,   2,  -1,   0 },     99.72049,   0.00005,     0.076 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   0,   0 },    272.70327,   0.00025,     0.043 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,  -2,   1,   0 },    272.71380,   0.00011,     0.103 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -2,   0 },    272.22422,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,  -1,   0 },    272.40722,   0.00058,     0.087 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   0,   0 },     92.37979,   0.00010,     0.578 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   0,   1,   0 },     92.52149,   0.00010,     0.067 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -2,   0 },     92.26469,   0.00015,    23.075 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,  -1,   0 },    272.38168,   0.00279,     0.076 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   2,   0,   0 },    272.05635,   0.00020,     0.038 },
 { {   0,   0,   2,   0,  -4,   2,   0,   0,   4,  -2,   0 },    272.23176,   0.00003,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,  -2,   0,   0 },    329.16234,   0.00014,     0.044 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,  -2,   1,   0 },    329.16234,   0.00004,     0.103 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -2,   0 },    147.29946,   0.00007,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,  -1,   0 },    327.18307,   0.00068,     0.087 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,   0,   0 },    147.62167,   0.00013,     0.567 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   0,   1,   0 },    148.91114,   0.00002,     0.067 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -3,   0 },    142.68724,   0.00021,     0.075 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -2,   0 },    147.22841,   0.00023,   106.494 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -1,   0 },    327.16334,   0.00457,     0.075 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,   0,   0 },    327.31129,   0.00026,     0.038 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -3,   0 },    328.92931,   0.00011,     0.087 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -2,   0 },    327.34266,   0.00016,     0.040 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   4,  -1,   0 },    327.16426,   0.00004,     0.026 },
 { {   0,   0,   2,   0,  -4,   4,  -2,   0,   2,  -3,   0 },    264.85996,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -4,   4,  -2,   0,   2,  -1,   0 },     84.85996,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,  -2,   0,   0 },    207.58238,   0.00003,     0.044 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   0,  -1,   0 },     26.93960,   0.00003,     0.087 },
 { {   0,   0,   2,   0,  -4,   4,   0,   0,   2,  -1,   0 },     26.93914,   0.00019,     0.075 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,  -1,   0 },     81.22808,   0.00005,     0.028 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   0,   0 },     81.22968,   0.00054,     0.044 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   1,   0 },     81.23029,   0.00019,     0.104 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -2,   0 },    260.99340,   0.00008,     0.041 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,  -1,   0 },    261.20044,   0.00084,     0.088 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,   1,   0 },    261.28381,   0.00017,     0.066 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -3,   0 },    236.43696,   0.00005,     0.076 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -2,   0 },     81.56843,   0.00034,    17.093 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,   0 },    261.21891,   0.00585,     0.075 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,   0,   0 },    265.07828,   0.00010,     0.038 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   4,  -2,   0 },    260.30516,   0.00004,     0.040 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   4,  -1,   0 },    261.20186,   0.00005,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,  -1,   0 },    280.01699,   0.00003,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   0,   0 },    279.42509,   0.00021,     0.021 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   1,   0 },    280.41618,   0.00041,     0.029 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -4,   2,   0 },    280.48557,   0.00012,     0.047 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -3,   0,   0 },    280.20972,   0.00007,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -3,   0 },    280.10241,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -2,   0 },    280.26082,   0.00017,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    280.29445,   0.00192,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,  -1,   2 },    100.23544,   0.00002,     0.106 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   0 },    280.17755,   0.02206,     0.043 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   2 },    100.36288,   0.00009,     0.259 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   0 },    280.34145,   0.01241,     0.103 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   2 },    100.57907,   0.00002,     0.058 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   2,   0 },    100.31340,   0.00090,     0.284 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   3,   0 },    100.38566,   0.00006,     0.060 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,  -1,   0 },    100.08580,   0.00006,     0.042 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -1,   0,   0 },    101.53081,   0.00015,     0.094 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -3,   0 },     99.64991,   0.00004,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   0 },    281.28938,   0.00042,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   2 },    280.08904,   0.00003,     0.481 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   0 },    280.16516,   0.07725,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   2 },    100.17114,   0.00019,     0.065 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,  -2 },    280.82510,   0.00004,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 },    100.18878,   0.01480,     0.572 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,  -2 },    280.69419,   0.00003,     0.084 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   0 },    100.77083,   0.00819,     0.067 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   2,   0 },    100.63964,   0.00043,     0.035 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -2,   0 },    280.65987,   0.00004,     0.081 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,  -1,   0 },    280.08584,   0.00016,     1.163 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   1,   0,   0 },    278.19427,   0.00039,     0.071 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -5,   0 },    100.15203,   0.00003,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -4,   0 },    100.15223,   0.00049,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   0 },    100.14906,   0.00895,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   2 },     99.82344,   0.00002,     0.074 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,  -2 },    100.08279,   0.00008,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   0 },    100.09665,   0.02341,    37.133 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   2 },    100.08279,   0.00008,     0.037 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,  -2 },    100.07396,   0.00145,     0.073 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   0 },    280.06924,   0.41011,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,  -2 },    280.08586,   0.00017,     2.774 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   0 },    280.38232,   0.02460,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,  -2 },    100.38460,   0.00006,     0.078 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   1,   0 },    279.99843,   0.00147,     0.025 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   2,   0 },    280.03419,   0.00007,     0.019 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   3,  -2,   0 },    100.01302,   0.00011,     0.081 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   3,  -1,   0 },    100.08583,   0.00012,     0.039 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -4,   0 },     99.86235,   0.00013,     0.555 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,   0 },    279.99299,   0.00345,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,  -2 },    280.08279,   0.00003,     0.469 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   0 },    280.06118,   0.00804,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,  -2 },    100.08587,   0.00008,     0.090 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,   0 },    280.08378,   0.00391,     0.026 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   0,   0 },    278.10214,   0.00049,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,   1,   0 },    280.08579,   0.00004,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -4,   0 },    279.99181,   0.00004,     0.044 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -3,   0 },    280.04413,   0.00013,     0.028 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -2,   0 },    280.07115,   0.00011,     0.020 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   6,  -1,   0 },    280.08279,   0.00002,     0.016 },
 { {   0,   0,   2,   0,  -3,   0,   2,   0,   2,  -1,   0 },    259.78404,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,  -2,   0,   0 },     93.60212,   0.00013,     0.044 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -2,   0 },    271.30737,   0.00021,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,  -1,   0 },    271.42391,   0.00098,     0.087 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   0,   0,   0 },     91.43828,   0.00012,     0.561 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -3,   0 },    269.80514,   0.00033,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -2,   0 },     91.34035,   0.00027,   142.500 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,  -1,   0 },    271.45648,   0.00442,     0.075 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   2,   0,   0 },    271.45648,   0.00024,     0.038 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -3,   0 },     91.91425,   0.00014,     0.087 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -2,   0 },     91.27988,   0.00011,     0.040 },
 { {   0,   0,   2,   0,  -3,   1,   0,   0,   4,  -1,   0 },    271.43169,   0.00004,     0.026 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,  -2,   0,   0 },    292.89251,   0.00015,     0.044 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,  -2,   1,   0 },    292.89251,   0.00004,     0.103 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,  -1,   0 },    100.84556,   0.00015,     0.087 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,   0,   0 },    111.12657,   0.00004,     0.551 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   0,   1,   0 },    105.51986,   0.00002,     0.066 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -3,   0 },     91.20864,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -2,   0 },    281.14204,   0.00006,    24.411 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,  -1,   0 },    101.46582,   0.00097,     0.075 },
 { {   0,   0,   2,   0,  -3,   2,   0,   0,   2,   0,   0 },    101.44347,   0.00005,     0.038 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,  -2,   0,   0 },    166.31486,   0.00009,     0.044 },
 { {   0,   0,   2,   0,  -3,   5,   0,   0,   2,  -1,   0 },     40.22274,   0.00004,     0.075 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,  -2,   0,   0 },    298.35504,   0.00006,     0.043 },
 { {   0,   0,   2,   0,  -2,  -5,   0,   0,   2,  -1,   0 },    299.41191,   0.00010,     0.076 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   0,  -2,   0 },    255.35711,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   0,  -1,   0 },     75.66243,   0.00014,     0.087 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -2,   0 },    255.39155,   0.00005,    95.036 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,  -1,   0 },     75.63628,   0.00084,     0.076 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   2,   0,   0 },     75.66230,   0.00004,     0.038 },
 { {   0,   0,   2,   0,  -2,  -2,   0,   0,   4,  -2,   0 },     75.35711,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   2,  -3,   0 },    250.43129,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -2,  -1,   0,   0,   2,  -1,   0 },     78.08130,   0.00017,     0.075 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,  -1,   0 },    268.87707,   0.00013,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   0,   0 },    272.01435,   0.00084,     0.021 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   0,   2 },    270.87980,   0.00005,     0.048 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   1,   0 },    271.72113,   0.00199,     0.029 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   2,   0 },    271.24497,   0.00076,     0.047 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -4,   3,   0 },    270.57370,   0.00003,     0.127 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,   0,   0 },    271.53007,   0.00005,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -3,   1,   0 },     91.14025,   0.00016,     0.045 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -3,   0 },    271.43901,   0.00003,     0.016 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -2,   0 },    271.74095,   0.00015,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    271.52568,   0.00461,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    270.93928,   0.10503,     0.044 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   2 },     90.84758,   0.00041,     0.254 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,  -2 },     90.80636,   0.00002,     0.027 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   0 },    270.86062,   0.06082,     0.104 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   2 },     90.78204,   0.00019,     0.058 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   2,   0 },     90.79876,   0.00451,     0.277 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   3,   0 },     90.96934,   0.00025,     0.059 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,  -1,   0 },    268.97731,   0.00004,     0.042 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,  -1,   1 },    267.58718,   0.00003,     0.096 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   0,   0 },     91.75790,   0.00069,     0.095 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -1,   1,   0 },    270.56334,   0.00011,     0.368 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -4,   0 },     90.12213,   0.00006,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -3,   0 },     89.98423,   0.00070,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.99665,   0.01160,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -2,   2 },     90.10499,   0.00008,     0.462 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,  -2 },    270.12198,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   0 },     89.97863,   0.14450,     0.088 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   2 },    270.08726,   0.00034,     0.065 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,  -2 },    270.30751,   0.00033,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 },     94.02089,   0.00200,     0.546 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   2 },    270.38512,   0.00034,     0.035 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,  -2 },     90.00276,   0.00004,     0.085 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   1,   0 },     90.39429,   0.03280,     0.066 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   2,   0 },     90.34022,   0.00184,     0.035 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   3,   0 },     90.38811,   0.00011,     0.024 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -2,   0 },    270.12949,   0.00014,     0.081 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,  -1,   0 },     90.12884,   0.00053,     1.059 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,   0 },    270.68035,   0.00135,     0.070 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   1,   0 },    270.16253,   0.00009,     0.036 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -4,   0 },     90.04770,   0.00023,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,   0 },     90.05783,   0.00431,     0.076 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -3,   2 },     90.11651,   0.00003,     0.073 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,  -2 },    270.12720,   0.00025,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   0 },    270.11778,   0.06090,    17.430 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   2 },    270.15384,   0.00024,     0.037 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,  -2 },    270.12077,   0.00450,     0.074 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   0 },     90.11969,   1.05870,     0.075 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,  -2 },     90.12800,   0.00061,     3.621 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   0,   0 },     90.12570,   0.04350,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,  -2 },     90.11283,   0.00042,     0.077 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   1,   0 },     90.22112,   0.00285,     0.025 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,  -2 },     90.11362,   0.00003,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,   2,   0 },     89.89788,   0.00015,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -2,   0 },    270.19629,   0.00003,     0.080 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,  -1,   0 },    270.12183,   0.00033,     0.039 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   3,   0,   0 },    270.12183,   0.00004,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -4,   0 },     90.10147,   0.00018,     0.583 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,   0 },    270.09213,   0.00370,     0.087 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -3,   2 },    270.12198,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,  -2 },     90.12198,   0.00009,     0.488 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -2,   0 },     90.13826,   0.00717,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,  -2 },    270.12521,   0.00028,     0.089 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,  -1,   0 },     90.12662,   0.00887,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,  -2 },    270.11732,   0.00002,     0.041 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   0,   0 },     90.12249,   0.00089,     0.019 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   1,   0 },     90.13996,   0.00006,     0.015 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   4,   2,   0 },    270.11810,   0.00002,     0.013 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -4,   0 },    270.09093,   0.00004,     0.043 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -3,   0 },     90.16908,   0.00005,     0.028 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -2,   0 },     90.12878,   0.00019,     0.020 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   6,  -1,   0 },     90.12186,   0.00009,     0.016 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,  -2,   0,   0 },    131.73264,   0.00003,     0.044 },
 { {   0,   0,   2,   0,  -2,   1,   0,   0,   2,  -1,   0 },    275.99465,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,  -2,   0,   0 },     13.51314,   0.00009,     0.044 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   0,  -1,   0 },    190.95412,   0.00004,     0.088 },
 { {   0,   0,   2,   0,  -2,   2,   0,   0,   2,  -1,   0 },    190.95419,   0.00022,     0.075 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,  -2,   0,   0 },    259.15461,   0.00010,     0.044 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   0,  -1,   0 },     80.93398,   0.00004,     0.088 },
 { {   0,   0,   2,   0,  -2,   3,   0,   0,   2,  -1,   0 },     80.93398,   0.00025,     0.075 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,  -2,   0,   0 },    197.13677,   0.00005,     0.044 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,   2,  -3,   0 },    230.76610,   0.00003,     0.077 },
 { {   0,   0,   2,   0,  -2,   5,   0,   0,   2,  -1,   0 },     50.76610,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,  -2,   0,   0 },    288.78321,   0.00015,     0.043 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,  -2,   1,   0 },    288.77476,   0.00008,     0.103 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,  -1,   0 },    288.98969,   0.00062,     0.087 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,   0,   0 },    109.04077,   0.00010,     0.573 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,   1,   0 },    109.02530,   0.00007,     0.067 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -3,   0 },    140.81269,   0.00002,     0.075 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -2,   0 },    108.70392,   0.00017,    35.635 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -1,   0 },    288.96282,   0.00333,     0.076 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,   0,   0 },    288.98736,   0.00020,     0.038 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   4,  -3,   0 },    285.19557,   0.00004,     0.087 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   4,  -2,   0 },    288.70580,   0.00007,     0.040 },
 { {   0,   0,   2,   0,  -1,  -4,   0,   0,   2,  -1,   0 },     66.42941,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,  -2,   0,   0 },     81.91841,   0.00011,     0.044 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   0,  -1,   0 },    259.90902,   0.00007,     0.088 },
 { {   0,   0,   2,   0,  -1,  -2,   0,   0,   2,  -1,   0 },    259.90876,   0.00042,     0.075 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   2,  -3,   0 },     82.41457,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -1,  -1,   0,   0,   2,  -1,   0 },     82.41457,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,  -2,   0 },    171.95360,   0.00002,     0.014 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   0,   0 },    351.40632,   0.00035,     0.021 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   1,   0 },    352.46346,   0.00061,     0.029 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -4,   2,   0 },    352.92315,   0.00015,     0.048 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -3,   0,   0 },    353.40486,   0.00009,     0.028 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -3,   1,   0 },    173.42347,   0.00004,     0.046 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -2,   0 },    355.39289,   0.00010,     0.020 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,  -1,   0 },    353.63109,   0.00139,     0.028 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    354.46019,   0.02010,     0.044 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   2 },    175.28329,   0.00004,     0.249 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   1,   0 },    354.87177,   0.01039,     0.105 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   2,   0 },    175.31949,   0.00069,     0.271 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   3,   0 },    174.51282,   0.00005,     0.059 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,  -1,   0 },    347.76530,   0.00005,     0.042 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   0,   0 },    344.90379,   0.00042,     0.096 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -1,   1,   0 },    167.77326,   0.00006,     0.357 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -3,   0 },    180.72065,   0.00003,     0.026 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -2,   0 },    332.82533,   0.00019,     0.041 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -1,   0 },    336.36877,   0.00376,     0.088 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 },    174.72897,   0.00209,     0.522 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   1,   0 },    182.78572,   0.00244,     0.066 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   2,   0 },    190.47305,   0.00008,     0.035 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   3,   0 },    354.87744,   0.00002,     0.024 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   0 },    163.22531,   0.00031,     0.070 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   1 },    169.55293,   0.00003,     0.036 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   1,   0 },    167.77326,   0.00004,     0.036 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -3,   0 },      8.55757,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -2,   0 },     41.89622,   0.00063,     7.058 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,  -2 },     44.37870,   0.00006,     0.074 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    217.41823,   0.01224,     0.075 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,  -1 },    350.55577,   0.00004,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   0,   0 },    188.96153,   0.00035,     0.038 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,   1,   0 },    165.71139,   0.00007,     0.025 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -2,   0 },    215.54693,   0.00013,     0.040 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   4,  -1,   0 },    215.66230,   0.00011,     0.026 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,  -1,   0 },    277.63854,   0.00005,     0.028 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,   0 },    276.63130,   0.00057,     0.044 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   1,   0 },    279.50950,   0.00020,     0.104 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -2,   0 },     97.79021,   0.00009,     0.041 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,  -1,   0 },     97.43839,   0.00090,     0.088 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,   1,   0 },     98.43044,   0.00017,     0.066 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -3,   0 },    107.14656,   0.00003,     0.076 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -2,   0 },    277.65244,   0.00036,    17.781 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,   0 },     97.65564,   0.00625,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,   0,   0 },     95.28626,   0.00012,     0.038 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   4,  -2,   0 },     97.68186,   0.00004,     0.040 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   4,  -1,   0 },     97.66867,   0.00006,     0.026 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,  -2,   0,   0 },     95.37315,   0.00021,     0.044 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,  -2,   1,   0 },     95.37314,   0.00005,     0.104 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,  -1,   0 },    276.40226,   0.00011,     0.088 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   0,   1,   0 },    275.41316,   0.00003,     0.066 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -2,   0 },     96.91846,   0.00004,    11.088 },
 { {   0,   0,   2,   0,   0,  -4,   0,   0,   2,  -1,   0 },    276.18406,   0.00079,     0.075 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,  -1,   0 },      2.90089,   0.00010,     0.028 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   0,   0 },      3.29120,   0.00085,     0.044 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   1,   0 },      2.68475,   0.00028,     0.104 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -3,   0 },    183.02011,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -2,   0 },    183.01149,   0.00004,     0.041 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,  -1,   0 },    182.91964,   0.00046,     0.088 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   0,   0 },    181.98573,   0.00012,     0.527 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   1,   0 },    182.84337,   0.00028,     0.066 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -2,   0 },      3.02007,   0.00025,     8.056 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,  -1,   0 },    182.87243,   0.00427,     0.075 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   2,   0,   0 },    183.10835,   0.00009,     0.038 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   4,  -2,   0 },    183.02021,   0.00004,     0.040 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   4,  -1,   0 },    183.02011,   0.00002,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,   0,   0 },    270.20145,   0.00005,     0.021 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -4,   1,   0 },    270.21038,   0.00010,     0.029 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,  -1,   0 },    270.20818,   0.00026,     0.028 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   0,   0 },    270.23180,   0.00442,     0.044 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   1,   0 },    270.21118,   0.00238,     0.105 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,  -2,   2,   0 },     90.22014,   0.00019,     0.270 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   0 },     89.85920,   0.00006,     0.041 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,  -1,   0 },     89.88885,   0.00128,     0.088 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0 },     90.12849,   0.00063,     0.518 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   1,   0 },     90.10920,   0.00143,     0.066 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   2,   0 },     90.00392,   0.00007,     0.035 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   1,  -2,   0 },     89.84937,   0.00002,     0.082 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   1,   0,   0 },    269.84937,   0.00002,     0.070 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -3,   0 },    265.64563,   0.00004,     0.076 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -2,   0 },    270.04086,   0.00096,     6.326 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,  -2 },    270.02339,   0.00011,     0.074 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,  -1,   0 },     90.03369,   0.01755,     0.075 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   0,   0 },     90.28889,   0.00014,     0.037 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   1,  -2 },     90.00555,   0.00004,     0.076 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   2,   1,   0 },    269.98381,   0.00006,     0.025 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -2,   0 },     90.04353,   0.00018,     0.040 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,  -1,   0 },     90.03822,   0.00014,     0.026 },
 { {   0,   0,   2,   0,   0,  -2,   0,   0,   4,   0,   0 },     90.00555,   0.00004,     0.019 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -4,   0,   0 },    167.93527,   0.00003,     0.021 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -4,   1,   0 },    348.46290,   0.00002,     0.029 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,  -1,   0 },    348.47691,   0.00013,     0.028 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   0,   0 },    348.38020,   0.00130,     0.044 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   1,   0 },    347.80914,   0.00059,     0.105 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,  -2,   2,   0 },    168.42971,   0.00002,     0.267 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,  -1,   0 },    345.32168,   0.00014,     0.089 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0 },    169.53256,   0.00012,     0.509 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   1,   0 },    169.55055,   0.00015,     0.066 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -2,   0 },    352.83373,   0.00003,     5.207 },
 { {   0,   0,   2,   0,   0,  -1,   0,   0,   2,  -1,   0 },    172.62023,   0.00072,     0.074 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,  -2,   0,   0 },    268.79845,   0.00016,     0.044 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,  -1,   0 },     91.76912,   0.00003,     0.089 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0 },     88.97662,   0.00004,     0.506 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   1,   0 },     86.84794,   0.00002,     0.066 },
 { {   0,   0,   2,   0,   0,   0,  -2,   0,   2,  -1,   0 },     89.76064,   0.00026,     0.074 },
 { {   0,   0,   2,   0,   0,   0,  -1,   0,  -2,   0,   0 },    339.05605,   0.00005,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,  -2,  -2,   0,   0 },    270.21905,   0.00006,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,  -2,   2,  -1,   0 },     90.01898,   0.00006,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,  -1,   0 },     93.72112,   0.00004,     0.028 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },     91.45986,   0.00073,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   0 },     93.23130,   0.00022,     0.105 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   2,   0 },    287.63860,   0.00003,     0.265 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    100.39729,   0.00007,     0.042 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0 },    293.87729,   0.00004,     0.096 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0 },    295.59041,   0.00003,     0.041 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },     84.83892,   0.00009,     0.089 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    291.21906,   0.00043,     0.500 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    289.57459,   0.00037,     0.066 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0 },    114.65148,   0.00008,     0.070 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    302.41540,   0.00002,     0.077 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    115.82026,   0.00011,     4.425 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    293.17703,   0.00193,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },    102.71182,   0.00007,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    300.70623,   0.00011,     0.037 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   4,  -2,   0 },    290.55171,   0.00002,     0.040 },
 { {   0,   0,   2,   0,   0,   1,   0,   0,  -2,   0,   0 },    153.29324,   0.00006,     0.044 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,  -2,   0,   0 },     16.81842,   0.00007,     0.044 },
 { {   0,   0,   2,   0,   1,  -5,   0,   0,   2,  -1,   0 },    288.62498,   0.00008,     0.075 },
 { {   0,   0,   2,   0,   1,  -2,   0,   0,  -2,   0,   0 },    169.24618,   0.00009,     0.044 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   0,   0 },    229.96498,   0.00038,     0.044 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -2,   1,   0 },    228.94059,   0.00010,     0.106 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -1,   0,  -1 },     62.53742,   0.00003,     0.042 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,  -1,   1,  -1 },    242.69049,   0.00002,     0.096 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   0,  -1,   0 },    225.32724,   0.00003,     0.090 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   0,   1,   0 },     45.32724,   0.00003,     0.065 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,  -1,   0 },    342.33310,   0.00007,     0.074 },
 { {   0,   0,   2,   0,   1,   2,   0,   0,   2,  -3,   0 },    225.69703,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   1,   2,   0,   0,   2,  -1,   0 },     45.69703,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   2,  -8,   0,   0,  -2,   0,   0 },    191.76991,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   2,  -8,   0,   0,   2,  -1,   0 },     11.40931,   0.00007,     0.075 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,  -2,   0,   0 },     98.27321,   0.00012,     0.044 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   0,  -1,   0 },    278.73493,   0.00005,     0.088 },
 { {   0,   0,   2,   0,   2,  -7,   0,   0,   2,  -1,   0 },    278.73498,   0.00026,     0.075 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,  -1,   0 },     53.33047,   0.00004,     0.028 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   0,   0 },     53.42467,   0.00053,     0.044 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   1,   0 },     53.04254,   0.00013,     0.105 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,  -1,   0 },     51.62762,   0.00010,     0.089 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,   0,   0 },    232.12117,   0.00003,     0.500 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,   1,   0 },    232.89209,   0.00007,     0.066 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   2,  -3,   0 },     78.08174,   0.00003,     0.077 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   2,  -1,   0 },    238.39124,   0.00015,     0.074 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,  -3,   0 },    218.32362,   0.00003,     0.078 },
 { {   0,   0,   2,   0,   2,   0,   0,   0,   2,  -1,   0 },    218.32362,   0.00003,     0.073 },
 { {   0,   0,   2,   0,   3,  -1,   0,   0,   2,  -3,   0 },    216.93963,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   3,  -1,   0,   0,   2,  -1,   0 },     36.93963,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -3,   0,   0,   2,  -3,   0 },    338.54160,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   4,  -3,   0,   0,   2,  -1,   0 },    158.54160,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -3,   0 },    217.33748,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -2,   0 },    217.13180,   0.00002,     2.019 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,  -1,   0 },     37.33748,   0.00002,     0.073 },
 { {   0,   0,   2,   0,   4,  -2,   0,   0,   2,   0,   0 },     37.13180,   0.00002,     0.037 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,  -3,   0 },    354.91699,   0.00002,     0.078 },
 { {   0,   0,   2,   0,   5,  -5,   0,   0,   2,  -1,   0 },    174.91699,   0.00002,     0.073 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   1,   0 },    297.62016,   0.00008,     0.075 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,   0,   0 },    259.69752,   0.00007,     0.043 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   2,  -1,   0 },    291.99916,   0.00005,     0.095 },
 { {   0,   0,   3,  -8,   1,   5,   0,   0,   2,   0,   0 },    291.99916,   0.00013,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -4,   1,   0 },    299.71833,   0.00002,     0.027 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    299.38355,   0.00013,     0.026 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   0,   0 },    299.50407,   0.00068,     0.039 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   1,   0 },    300.32328,   0.00036,     0.080 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -2,   2,   0 },    119.79138,   0.00004,     1.293 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,  -1,   0,   0 },    118.93912,   0.00004,     0.075 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -1,   0 },    121.38954,   0.00094,     0.070 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   0,   0 },    119.77897,   0.00087,     0.999 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,   0 },    299.19915,   0.00313,     0.082 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   2,   0 },    298.83562,   0.00020,     0.039 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -2,   0 },    120.31291,   0.00042,     0.360 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,   0 },    299.96210,   0.00716,     0.095 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,  -2 },    119.74834,   0.00002,     0.322 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   0 },    299.68080,   0.01116,     0.042 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   1,   0 },    299.48179,   0.00084,     0.027 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   2,   0 },    299.74849,   0.00005,     0.020 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   3,   0,   0 },    299.74023,   0.00005,     0.028 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,  -2,   0 },    300.15226,   0.00011,     0.046 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,  -1,   0 },    299.79781,   0.00041,     0.028 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   0,   0 },    299.67886,   0.00020,     0.021 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   4,   1,   0 },    119.72773,   0.00003,     0.016 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   1,  -1 },    234.43917,   0.00002,     0.026 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   2,  -1 },    234.49816,   0.00024,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -3,   3,  -1 },    234.40807,   0.00018,     0.087 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,  -1 },     52.16032,   0.00004,     0.038 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,  -1 },     52.16031,   0.00065,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,  -1 },    233.43668,   0.00008,   650.233 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   3,  -1 },     54.47906,   0.00079,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   4,  -1 },     54.47906,   0.00004,     0.038 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0 },    346.27309,   0.00007,     0.072 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   1,   0 },    166.27309,   0.00007,     0.080 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   0,  -1 },    233.49493,   0.00004,     0.563 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   1,  -1 },     53.84384,   0.00044,     0.087 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   2,  -1 },     54.38639,   0.00027,     0.040 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   1,   3,  -1 },     54.43917,   0.00002,     0.026 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -1,   0 },    259.61647,   0.00007,     0.093 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,   0,   0 },    250.19606,   0.00017,     0.042 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -4,   1,   0 },      1.99261,   0.00004,     0.027 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,  -1,   0 },      1.85186,   0.00018,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,   0 },      0.98764,   0.00142,     0.040 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   1,   0 },      3.91653,   0.00131,     0.086 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   2,   0 },    188.01529,   0.00004,     0.631 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0 },    358.67166,   0.00015,     0.037 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0 },    358.87151,   0.00283,     0.074 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0 },      1.37781,   0.00026,     5.260 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0 },    179.24562,   0.00286,     0.077 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0 },    179.36515,   0.00017,     0.038 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,  -2,   0 },    183.33445,   0.00002,     0.070 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   1,   0,   0 },      3.33445,   0.00002,     0.082 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -2,   0 },      8.21187,   0.00008,     0.509 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,   0 },    185.12586,   0.00182,     0.089 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,   0 },    184.60656,   0.00208,     0.041 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   1,   0 },    183.99609,   0.00026,     0.026 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   4,  -1,   0 },    184.71158,   0.00005,     0.028 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,  -1,   0 },     61.15956,   0.00006,     0.026 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   0,   0 },     61.07434,   0.00057,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,  -2,   1,   0 },     61.41504,   0.00046,     0.087 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -2,   0 },     59.91923,   0.00004,     0.038 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,  -1,   0 },     59.95810,   0.00085,     0.075 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   1,   0 },    239.96882,   0.00085,     0.076 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   2,   0 },    239.91579,   0.00004,     0.038 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,  -1,   0 },    241.49692,   0.00047,     0.087 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   0,   0 },    241.12627,   0.00058,     0.040 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   2,   1,   0 },    241.22454,   0.00006,     0.026 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,  -2,   0,   0 },    217.66330,   0.00003,     0.040 },
 { {   0,   0,   3,  -5,  -4,   0,   0,   0,   2,   0,   0 },     37.77034,   0.00003,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   0,   0 },     31.55381,   0.00004,     0.020 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   1,   0 },     31.70196,   0.00016,     0.028 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -4,   2,   0 },     31.58121,   0.00005,     0.044 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,  -1,   0 },     31.42905,   0.00028,     0.027 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   0 },     31.91779,   0.00421,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   2 },    211.74537,   0.00003,     0.408 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   1,   0 },     31.79071,   0.00398,     0.090 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   2,   0 },    212.17563,   0.00024,     0.473 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   0 },    211.46003,   0.00009,     0.083 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0 },     30.17584,   0.00021,     0.038 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0 },     30.10889,   0.00365,     0.077 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0 },    210.85854,   0.00051,     2.928 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -2 },    208.46614,   0.00016,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0 },    209.99527,   0.00344,     0.074 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0 },    209.91273,   0.00018,     0.037 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -2,   0 },     32.21021,   0.00013,     0.698 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,   0 },    211.63208,   0.00278,     0.085 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,   0 },    211.29992,   0.00243,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   1,   0 },    211.20385,   0.00022,     0.026 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,  -2,   0 },    211.43748,   0.00003,     0.043 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,  -1,   0 },    211.37779,   0.00004,     0.027 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,   0,   0 },    211.66578,   0.00003,     0.020 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,  -1,   0 },    237.30252,   0.00003,     0.016 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   0,   0 },     62.84030,   0.00015,     0.021 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   1,   0 },     58.62691,   0.00029,     0.028 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -4,   2,   0 },     58.80006,   0.00009,     0.045 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -3,   0,   0 },     57.69370,   0.00005,     0.028 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -2,   0 },     58.64976,   0.00008,     0.020 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,  -1,   0 },     58.91676,   0.00088,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   0 },     58.04506,   0.01354,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   2 },    238.65648,   0.00003,     0.335 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,   0 },     58.65944,   0.01067,     0.094 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   2,   0 },    239.22193,   0.00070,     0.378 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   3,   0 },    238.65685,   0.00005,     0.063 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,  -1,   0 },     61.28584,   0.00011,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,  -1,   1 },    324.81545,   0.00004,     0.088 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   0 },    241.07707,   0.00179,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   1,   0 },     61.23562,   0.00023,     0.568 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0 },     60.00118,   0.00044,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0 },     59.68541,   0.00879,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2 },    240.01765,   0.00004,     0.069 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0 },    240.23087,   0.00299,     1.145 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2 },     60.01765,   0.00004,     0.079 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0 },    240.65577,   0.00749,     0.071 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0 },    240.39135,   0.00038,     0.037 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0 },    238.82087,   0.00003,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -3,   0 },     61.55196,   0.00003,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -2,   0 },     61.55195,   0.00057,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,  -1 },    144.79510,   0.00011,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   0 },     61.21845,   0.00056,    72.752 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   1 },    324.81093,   0.00018,     0.075 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -2 },     61.23560,   0.00004,     0.074 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },    241.23208,   0.01040,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,  -1 },    324.79510,   0.00011,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   1,   0 },    241.23688,   0.00056,     0.038 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   2,   0 },    241.23563,   0.00003,     0.025 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -2,   0 },     62.55230,   0.00011,     1.110 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -1,   0 },    240.17446,   0.00406,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,   0 },    242.40048,   0.00331,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   1,   0 },    239.83773,   0.00027,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -2,   0 },    241.05693,   0.00015,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,  -1,   0 },    241.24221,   0.00030,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   3,   0,   0 },    241.23771,   0.00010,     0.026 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,  -2,   0 },    240.42115,   0.00004,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,  -1,   0 },    240.94209,   0.00008,     0.027 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   4,   0,   0 },    244.37044,   0.00003,     0.020 },
 { {   0,   0,   3,  -4,   2,   0,   0,   0,  -2,   0,   0 },    326.72569,   0.00006,     0.042 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   0,   0 },    273.77417,   0.00005,     0.021 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   1,   0 },    277.24891,   0.00011,     0.029 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -4,   2,   0 },    276.61853,   0.00004,     0.046 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,  -1,   0 },    277.00587,   0.00026,     0.027 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   0 },    278.14234,   0.00500,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,   0 },    277.42238,   0.00315,     0.099 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   2,   0 },     98.14127,   0.00023,     0.315 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   0,   0 },     94.34579,   0.00003,     0.091 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   2,   0 },    274.34579,   0.00003,     0.064 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0 },    272.95349,   0.00023,     0.040 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0 },    273.13015,   0.00384,     0.084 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0 },     92.46233,   0.00158,     0.712 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0 },     92.53777,   0.00226,     0.068 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0 },     92.57021,   0.00012,     0.036 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,  -2,   0 },     96.62959,   0.00003,     0.079 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   1,   0,   0 },    276.62959,   0.00003,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -3,   0 },     86.11248,   0.00003,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -2,   0 },     91.17799,   0.00035,     2.710 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,   0 },    270.94710,   0.00558,     0.078 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,   0 },    273.55351,   0.00050,     0.038 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   1,   0 },    270.02810,   0.00009,     0.025 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   4,  -2,   0 },    270.55261,   0.00005,     0.041 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   4,  -1,   0 },    270.59037,   0.00005,     0.027 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,  -1,   0 },    166.44481,   0.00002,     0.016 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   0,   0 },    347.06829,   0.00011,     0.021 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   1,   0 },    346.40159,   0.00025,     0.029 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -4,   2,   0 },    346.60038,   0.00006,     0.048 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,  -1,   0 },    346.55801,   0.00041,     0.028 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   0 },    346.54192,   0.00555,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   2 },    166.76605,   0.00004,     0.247 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   1,   0 },    346.96906,   0.00284,     0.105 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   2,   0 },    166.92353,   0.00021,     0.269 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -1,   0,   0 },    347.29796,   0.00008,     0.096 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0 },    346.42426,   0.00008,     0.041 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0 },    345.78426,   0.00114,     0.088 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0 },    165.27609,   0.00056,     0.516 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0 },    166.12079,   0.00075,     0.066 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0 },    167.07462,   0.00002,     0.035 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,   0 },    167.29793,   0.00006,     0.070 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },    349.67449,   0.00009,     6.150 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    169.06076,   0.00186,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,  -1 },     40.75854,   0.00019,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },    166.44360,   0.00010,     0.037 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   4,  -2,   0 },    169.09179,   0.00002,     0.040 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,  -2,   0,   0 },    248.92544,   0.00007,     0.045 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,  -2,   0,   0 },    294.45128,   0.00008,     0.045 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   2,  -1,   0 },    120.63950,   0.00013,     0.072 },
 { {   0,   0,   3,   0,  -6,   0,   0,   0,   2,   0,   0 },    301.23830,   0.00007,     0.037 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,  -2,   0,   0 },    122.00638,   0.00004,     0.045 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,  -2,   0 },    132.99596,   0.00002,     1.124 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,  -1,   0 },    316.73342,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -6,   5,   0,   0,   2,   0,   0 },    132.99596,   0.00002,     0.036 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   0,   0 },    284.06705,   0.00035,     0.045 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,  -2,   1,   0 },    283.21716,   0.00007,     0.112 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,  -1,   0 },    114.84089,   0.00004,     0.094 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 },    108.13940,   0.00009,     0.388 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   0,   1,   0 },    108.10523,   0.00015,     0.063 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -2,   0 },    289.59416,   0.00006,     1.243 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,  -1,   0 },    109.63477,   0.00107,     0.071 },
 { {   0,   0,   3,   0,  -5,   0,   0,   0,   2,   0,   0 },    291.03408,   0.00026,     0.037 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,  -2,   0,   0 },    256.25768,   0.00004,     0.045 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,  -2,   0 },    271.17200,   0.00002,     1.146 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,  -1,   0 },     91.19603,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -5,   2,   0,   0,   2,   0,   0 },    271.17200,   0.00002,     0.037 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,  -2,   0,   0 },     73.13987,   0.00014,     0.045 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   0,   0,   0 },    254.32874,   0.00004,     0.364 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,  -3,   0 },     56.67752,   0.00002,     0.081 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,  -1,   0 },    258.77768,   0.00020,     0.070 },
 { {   0,   0,   3,   0,  -5,   5,   0,   0,   2,   0,   0 },     78.24812,   0.00013,     0.036 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -4,   1,   0 },    272.11996,   0.00006,     0.030 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    272.29956,   0.00022,     0.028 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,   0 },    271.28574,   0.00207,     0.045 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   1,   0 },    272.00058,   0.00087,     0.113 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   2,   0 },     91.54332,   0.00007,     0.225 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,  -2,   0 },     98.88077,   0.00002,     0.042 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,  -1,   0 },    122.47013,   0.00030,     0.094 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 },     96.19754,   0.00027,     0.376 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   1,   0 },     97.76390,   0.00079,     0.063 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   2,   0 },    100.46329,   0.00004,     0.034 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,  -3,   0 },    104.46989,   0.00006,     0.040 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   1,  -2,   0 },    104.13639,   0.00014,     0.087 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -2,   0 },    279.86202,   0.00030,     1.125 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -1,   0 },     99.16656,   0.00584,     0.071 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   0,   0 },    280.75815,   0.00151,     0.036 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,   1,   0 },    280.30072,   0.00030,     0.025 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -4,   0 },    104.64024,   0.00013,     0.075 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   3,  -2,   0 },    103.70018,   0.00035,     0.075 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   4,  -2,   0 },    101.24848,   0.00006,     0.039 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   4,   0,   0 },    276.31953,   0.00003,     0.019 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   5,  -4,   0 },    284.40293,   0.00006,     0.087 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   5,  -3,   0 },    284.46989,   0.00006,     0.040 },
 { {   0,   0,   3,   0,  -4,   2,   0,   0,  -2,   0,   0 },    188.08692,   0.00004,     0.045 },
 { {   0,   0,   3,   0,  -4,   3,   0,   0,   0,  -1,   0 },    224.21009,   0.00003,     0.095 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,  -2,   0,   0 },    336.13706,   0.00010,     0.046 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,   0,  -1,   0 },    158.27774,   0.00004,     0.096 },
 { {   0,   0,   3,   0,  -4,   5,   0,   0,   2,  -1,   0 },    338.27676,   0.00003,     0.070 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,  -1,   0 },     75.25204,   0.00002,     0.017 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   0,   0 },    276.81411,   0.00008,     0.021 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   1,   0 },    250.72198,   0.00022,     0.030 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -4,   2,   0 },    257.46454,   0.00003,     0.049 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    245.56841,   0.00045,     0.028 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   0 },    253.47618,   0.00790,     0.045 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   2 },     75.94179,   0.00002,     0.206 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   1,   0 },    255.02396,   0.00338,     0.114 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   2,   0 },     73.99009,   0.00027,     0.221 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -1,   0,   0 },     78.30192,   0.00005,     0.104 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -2,   0 },    349.91617,   0.00006,     0.042 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -1,   0 },    167.28648,   0.00326,     0.095 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 },     63.09353,   0.00131,     0.364 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   1,   0 },     85.30652,   0.00272,     0.062 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,   2,   0 },     85.02244,   0.00011,     0.034 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   1,   0,   0 },    266.29867,   0.00010,     0.066 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -3,   0 },    302.00937,   0.00010,     0.081 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -2,   0 },    274.12247,   0.00106,     1.028 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,  -2 },    269.50170,   0.00007,     0.079 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,  -1,   0 },     83.90501,   0.02139,     0.070 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,  -2 },     89.90601,   0.00006,     1.564 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   0,   0 },    272.90661,   0.00552,     0.036 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   1,   0 },    270.26695,   0.00032,     0.025 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   2,   2,   0 },    269.85605,   0.00006,     0.019 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   3,  -4,   0 },    270.09834,   0.00002,     0.076 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   3,  -2,   0 },     90.09834,   0.00002,     0.075 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -3,   0 },    340.76533,   0.00003,     0.080 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -2,   0 },     67.18707,   0.00024,     0.039 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,  -1,   0 },    334.30898,   0.00006,     0.026 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   0,   0 },    269.68826,   0.00009,     0.019 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   4,   1,   0 },     90.09147,   0.00003,     0.015 },
 { {   0,   0,   3,   0,  -3,   1,   0,   0,   0,  -1,   0 },    168.50132,   0.00004,     0.095 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,  -2,   0,   0 },    280.40449,   0.00004,     0.045 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,  -2,   0 },    288.85118,   0.00002,     1.127 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,  -1,   0 },    109.10326,   0.00005,     0.071 },
 { {   0,   0,   3,   0,  -2,  -5,   0,   0,   2,   0,   0 },    288.85118,   0.00002,     0.036 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,  -1,   0 },    347.02192,   0.00002,     0.017 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   0,   0 },    166.33651,   0.00020,     0.021 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   1,   0 },    166.60102,   0.00041,     0.030 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -4,   2,   0 },    166.02393,   0.00007,     0.050 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -3,   0,   0 },    166.28782,   0.00004,     0.029 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,  -1,   0 },    165.77667,   0.00052,     0.028 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,   0 },    165.72567,   0.01046,     0.046 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   0,   2 },    345.54131,   0.00002,     0.202 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   1,   0 },    165.70649,   0.00433,     0.116 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -2,   2,   0 },    346.43893,   0.00038,     0.217 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -1,   0,   0 },    345.19857,   0.00002,     0.105 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,  -1,   2,   0 },    345.19857,   0.00002,     0.059 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -3,   0 },    347.10388,   0.00006,     0.027 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -2,   0 },    346.92345,   0.00066,     0.042 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,  -1,   0 },    346.78355,   0.00590,     0.096 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 },    320.55235,   0.00011,     0.353 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   1,   0 },    341.36864,   0.00110,     0.062 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   0,   2,   0 },    335.76316,   0.00003,     0.034 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,  -2,  -1 },    165.99754,   0.00002,     0.040 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   1,  -1,  -1 },    167.28632,   0.00008,     0.087 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -3,   0 },    350.15775,   0.00003,     0.082 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -2,   0 },    155.51719,   0.00027,     0.946 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,  -1,   0 },    180.84797,   0.00430,     0.070 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   0,   0 },    130.89531,   0.00048,     0.036 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   2,   1,   0 },    141.57165,   0.00009,     0.024 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,   0 },    347.18505,   0.00003,     0.074 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,  -1 },    167.55602,   0.00039,     0.075 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,   0 },    347.18490,   0.00005,     0.037 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -2,   0 },    173.98251,   0.00008,     0.039 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   4,  -1,   0 },    167.18491,   0.00015,     0.026 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   5,  -2,  -1 },    345.99754,   0.00002,     0.040 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,  -2,   0,   0 },    262.26728,   0.00012,     0.045 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,  -3,   0 },    280.08805,   0.00002,     0.081 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,  -1,   0 },     98.61983,   0.00019,     0.070 },
 { {   0,   0,   3,   0,  -1,  -5,   0,   0,   2,   0,   0 },    278.61040,   0.00013,     0.036 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -4,   0,   0 },     70.39704,   0.00005,     0.021 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -4,   1,   0 },    253.93623,   0.00004,     0.030 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,  -1,   0 },    255.78271,   0.00012,     0.029 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    254.67413,   0.00134,     0.046 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   1,   0 },    255.47108,   0.00044,     0.117 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   2,   0 },     75.26443,   0.00003,     0.213 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,  -1,   0 },    189.51682,   0.00005,     0.097 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 },     95.86851,   0.00002,     0.343 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   0,   1,   0 },    103.28199,   0.00003,     0.062 },
 { {   0,   0,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    306.51296,   0.00003,     0.069 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,  -2,   0,   0 },    168.66934,   0.00011,     0.046 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   0,  -1,   0 },    354.72839,   0.00005,     0.096 },
 { {   0,   0,   3,   0,   0,  -5,   0,   0,   2,  -1,   0 },    174.72864,   0.00003,     0.070 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,  -2,   0,   0 },      0.56299,   0.00011,     0.046 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   2,  -1,   0 },    182.72768,   0.00013,     0.070 },
 { {   0,   0,   3,   0,   0,  -4,   0,   0,   2,   0,   0 },      2.72768,   0.00011,     0.036 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   0,   0 },    266.80542,   0.00023,     0.046 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,  -2,   1,   0 },    266.80694,   0.00008,     0.117 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,  -1,   0 },     85.28807,   0.00005,     0.097 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,   0,   0 },     87.96577,   0.00005,     0.345 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   0,   1,   0 },     88.58528,   0.00007,     0.062 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,  -1,   0 },     89.90009,   0.00050,     0.070 },
 { {   0,   0,   3,   0,   0,  -3,   0,   0,   2,   0,   0 },    269.91817,   0.00020,     0.036 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,  -1,   0 },    166.91555,   0.00002,     0.029 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   0,   0 },    166.97715,   0.00046,     0.046 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,  -2,   1,   0 },    166.95952,   0.00010,     0.117 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,  -1,   0 },    347.22500,   0.00008,     0.097 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,   0,   0 },    346.65263,   0.00003,     0.341 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   0,   1,   0 },    346.86595,   0.00005,     0.062 },
 { {   0,   0,   3,   0,   0,  -2,   0,   0,   2,  -1,   0 },    167.09738,   0.00009,     0.069 },
 { {   0,   0,   3,   0,   0,  -1,   0,   0,  -2,   0,   0 },    245.96482,   0.00014,     0.046 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },      3.75761,   0.00013,     0.046 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    200.05228,   0.00004,     0.333 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -3,   0 },    306.57569,   0.00003,     0.083 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },    129.15159,   0.00002,     0.816 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    213.91287,   0.00004,     0.069 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     29.02837,   0.00003,     0.036 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,  -2,   0,   0 },    125.66827,   0.00003,     0.046 },
 { {   0,   0,   3,   0,   2,  -5,   0,   0,  -2,   0,   0 },    310.27226,   0.00005,     0.046 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   1,   0 },     62.22565,   0.00004,     0.075 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -1,   0 },    237.80193,   0.00004,     0.093 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,   0,   0 },    231.63310,   0.00011,     0.042 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   0,   0 },    332.96858,   0.00027,     0.040 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   1,   0 },    335.80058,   0.00022,     0.085 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0 },    329.86291,   0.00003,     0.037 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0 },    330.14225,   0.00064,     0.074 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0 },    321.05046,   0.00007,     3.945 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,   0 },    150.59561,   0.00066,     0.077 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   2,   0 },    150.83935,   0.00003,     0.038 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -1,   0 },    158.63925,   0.00032,     0.089 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,   0 },    156.40293,   0.00048,     0.041 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   1,   0 },    156.99093,   0.00004,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,  -1,   0 },      9.17335,   0.00009,     0.026 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   0,   0 },      9.04468,   0.00079,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   1,   0 },      9.72352,   0.00063,     0.087 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,  -2,   2,   0 },    191.75096,   0.00002,     0.564 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -2,   0 },      7.01677,   0.00007,     0.038 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,  -1,   0 },      6.80673,   0.00114,     0.075 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   1,   0 },    186.80616,   0.00114,     0.075 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   0,   2,   0 },    187.01688,   0.00007,     0.038 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -2,   0 },     11.74668,   0.00002,     0.564 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,  -1,   0 },    189.71983,   0.00063,     0.087 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   0,   0 },    189.04176,   0.00079,     0.040 },
 { {   0,   0,   4,  -8,   1,   5,   0,   0,   2,   1,   0 },    189.17021,   0.00009,     0.026 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,  -2,   0,   0 },     31.45403,   0.00003,     0.040 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   0,  -1,   0 },     31.82590,   0.00004,     0.075 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   0,   1,   0 },    211.82590,   0.00004,     0.076 },
 { {   0,   0,   4,  -8,   2,   0,   0,   0,   2,   0,   0 },    211.64415,   0.00004,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   0,   0 },    196.78240,   0.00002,     0.013 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   1,   0 },     16.78080,   0.00003,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -6,   2,   0 },     16.78140,   0.00009,     0.021 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,  -1,   0 },     16.69789,   0.00015,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   0 },     16.68924,   0.00090,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   2 },     16.78090,   0.00008,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   0 },     16.87866,   0.00237,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   2 },     16.78106,   0.00006,     0.107 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   2,   0 },     17.13564,   0.00123,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   3,   0 },     16.78424,   0.00003,     0.103 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,  -1,   0 },     16.78041,   0.00002,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   0,   0 },     16.77901,   0.00006,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   1,   0 },    196.93696,   0.00025,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -3,   2,   0 },    196.78062,   0.00003,     0.094 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -3,   0 },     16.78186,   0.00003,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -2,   0 },     16.17545,   0.00053,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   0 },     16.40754,   0.00775,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   2 },    196.23936,   0.00009,     0.090 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   0 },     16.68897,   0.11216,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   2 },    196.69295,   0.00045,     0.475 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,  -2 },    196.78143,   0.00004,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   0 },     17.11263,   0.11251,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   2 },    197.05263,   0.00031,     0.065 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,  -2 },    196.78219,   0.00002,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   0 },    197.80892,   0.00565,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   2 },    196.78199,   0.00002,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   3,   0 },    199.13197,   0.00023,     0.067 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,  -1,   0 },    196.27160,   0.00019,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   0,   0 },    196.68183,   0.00187,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   1,   0 },     17.71680,   0.00011,     1.128 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   2,   0 },     18.37089,   0.00003,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -4,   0 },     14.33372,   0.00002,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -3,   0 },     15.02906,   0.00054,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   0 },     14.99238,   0.00862,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   2 },     14.94406,   0.00006,     3.003 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   0 },     14.98841,   0.15780,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   2 },    194.80905,   0.00056,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,  -2 },    196.76200,   0.00003,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   0 },     16.42476,   0.00002,  1783.359 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   2 },     16.78698,   0.00003,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,  -2 },     14.80018,   0.00056,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   0 },    194.98833,   0.15780,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,  -2 },    194.94397,   0.00006,     2.993 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   0 },    194.99256,   0.00862,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   3,   0 },    195.02965,   0.00054,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   4,   0 },    194.33416,   0.00002,     0.019 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,  -2,   0 },    198.37454,   0.00003,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,  -1,   0 },    197.71508,   0.00011,     1.127 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   0,   0 },     16.68175,   0.00188,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   1,   0 },     16.27256,   0.00019,     0.039 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -3,   0 },     19.13184,   0.00023,     0.067 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,  -2 },     16.78360,   0.00002,     0.035 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   0 },     17.81079,   0.00566,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   2 },     16.78362,   0.00002,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,  -2 },     17.05382,   0.00031,     0.065 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   0 },    197.11421,   0.11256,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   2 },     16.78391,   0.00004,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -2 },     16.69518,   0.00045,     0.474 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   0 },    196.69224,   0.11226,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,  -2 },     16.24251,   0.00009,     0.090 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   0 },    196.40911,   0.00776,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   2,   0 },    196.17683,   0.00053,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   3,   0 },    196.78393,   0.00003,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,  -2,   0 },     16.78432,   0.00003,     0.094 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,  -1,   0 },     16.94044,   0.00025,     0.042 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   0,   0 },    196.78545,   0.00006,     0.027 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   3,   1,   0 },    196.78456,   0.00002,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -3,   0 },    196.78372,   0.00003,     0.103 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -2,   0 },    197.13691,   0.00123,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,  -2 },    196.78401,   0.00006,     0.107 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,   0 },    196.88043,   0.00237,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,  -2 },    196.78391,   0.00008,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,   0 },    196.69361,   0.00090,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   1,   0 },    196.70060,   0.00015,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -2,   0 },    196.78383,   0.00009,     0.021 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,  -1,   0 },    196.78389,   0.00003,     0.016 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   6,   0,   0 },     16.78393,   0.00002,     0.013 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,  -2,   0,   0 },    151.30125,   0.00018,     0.040 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,  -2,   1,   0 },    151.30125,   0.00011,     0.087 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   2,  -1,   0 },    331.30037,   0.00011,     0.087 },
 { {   0,   0,   4,  -8,   5,  -5,   0,   0,   2,   0,   0 },    331.30037,   0.00018,     0.040 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   1,   0 },      2.77026,   0.00005,     0.028 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   2,   0 },      2.55259,   0.00002,     0.044 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,  -1,   0 },      2.76104,   0.00024,     0.027 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   0 },      2.68527,   0.00214,     0.041 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   1,   0 },      2.82081,   0.00201,     0.089 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   2,   0 },    183.34893,   0.00010,     0.487 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   0 },    182.39874,   0.00002,     0.083 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   2,   0 },      2.39874,   0.00002,     0.069 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0 },      0.69648,   0.00011,     0.038 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0 },      0.44473,   0.00187,     0.077 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0 },    181.45164,   0.00024,     3.594 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0 },    180.38628,   0.00184,     0.074 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0 },    180.32488,   0.00010,     0.037 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -2,   0 },      3.34072,   0.00007,     0.669 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,   0 },    182.61029,   0.00151,     0.085 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,   0 },    182.26694,   0.00144,     0.040 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   1,   0 },    182.22254,   0.00014,     0.026 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   4,  -1,   0 },    182.23268,   0.00004,     0.027 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   0,   0 },     30.53977,   0.00004,     0.021 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   1,   0 },     30.52555,   0.00009,     0.028 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -4,   2,   0 },     30.51520,   0.00004,     0.045 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,  -1,   0 },     30.60179,   0.00026,     0.027 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   0 },     30.36990,   0.00408,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   2 },    210.51555,   0.00003,     0.343 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   1,   0 },     30.68482,   0.00329,     0.094 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   2,   0 },    210.40165,   0.00022,     0.387 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,  -1,   1 },    294.00194,   0.00004,     0.088 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   0 },    211.64874,   0.00016,     0.087 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0 },     30.92718,   0.00015,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0 },     30.21898,   0.00263,     0.080 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0 },    212.27312,   0.00084,     1.235 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0 },    211.16886,   0.00226,     0.071 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0 },    211.27360,   0.00011,     0.037 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,  -1 },    113.99072,   0.00005,     0.074 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   0 },     32.28864,   0.00003,    12.968 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   1 },    294.00194,   0.00020,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   0 },    212.24103,   0.00055,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   1,  -1 },    293.99072,   0.00005,     0.077 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   1,   0 },    212.24103,   0.00003,     0.038 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -2,   0 },     32.46003,   0.00006,     1.037 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -1,   0 },    205.77896,   0.00175,     0.081 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,   0 },    210.90504,   0.00118,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   1,   0 },    211.89688,   0.00015,     0.026 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,  -2,   0 },    204.53826,   0.00002,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   4,  -1,   0 },    191.99394,   0.00006,     0.027 },
 { {   0,   0,   4,  -6,   2,   0,   0,   0,  -2,   0,   0 },     11.53588,   0.00007,     0.042 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -4,   0,   0 },    240.76638,   0.00003,     0.021 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -4,   1,   0 },    246.13743,   0.00007,     0.029 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,  -1,   0 },    246.09401,   0.00021,     0.027 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,   0 },    246.91630,   0.00315,     0.043 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   1,   0 },    245.69277,   0.00198,     0.099 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   2,   0 },     66.67528,   0.00014,     0.321 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0 },    243.20778,   0.00015,     0.040 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0 },    243.24038,   0.00233,     0.084 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0 },     62.67772,   0.00098,     0.745 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0 },     62.68388,   0.00151,     0.069 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0 },     62.69418,   0.00008,     0.036 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -2,   0 },     61.64961,   0.00021,     2.313 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -1,   0 },    241.34112,   0.00327,     0.078 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,   0 },    242.42090,   0.00028,     0.038 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   1,   0 },    241.47924,   0.00006,     0.025 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   4,  -2,   0 },    241.19540,   0.00003,     0.041 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -4,   1,   0 },    295.60175,   0.00005,     0.029 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    295.40057,   0.00023,     0.028 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   0 },    296.60461,   0.00224,     0.044 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   1,   0 },    295.89641,   0.00115,     0.104 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   2,   0 },    117.21833,   0.00008,     0.274 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -1,   0,   0 },    138.29869,   0.00007,     0.095 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0 },     87.68018,   0.00017,     0.041 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0 },     87.71522,   0.00247,     0.088 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0 },    123.57347,   0.00014,     0.534 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0 },     99.30845,   0.00070,     0.066 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0 },    101.33446,   0.00004,     0.035 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,   0 },    318.29859,   0.00005,     0.070 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -3,   0 },     73.66930,   0.00002,     0.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -2,   0 },    271.74291,   0.00106,    10.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,  -2 },    271.66723,   0.00009,     0.074 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   0 },     91.92022,   0.01882,     0.075 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,   0 },     93.31757,   0.00047,     0.038 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   1,  -2 },     90.64281,   0.00002,     0.077 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   1,   0 },    266.53177,   0.00002,     0.025 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -2,   0 },     92.30378,   0.00016,     0.040 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -1,   0 },     91.97359,   0.00017,     0.026 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,  -1,   0 },    166.18926,   0.00002,     0.028 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   0,   0 },    167.57534,   0.00051,     0.045 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   1,   0 },    166.59005,   0.00014,     0.110 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0 },    181.26500,   0.00008,     0.092 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0 },    340.98965,   0.00007,     0.416 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0 },    341.48598,   0.00008,     0.064 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -1,   0 },    308.90246,   0.00025,     0.072 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,   0 },    112.58185,   0.00009,     0.037 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   0,   0 },    244.49998,   0.00037,     0.046 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   1,   0 },    246.55194,   0.00008,     0.117 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0 },    247.42523,   0.00003,     0.097 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0 },     67.42523,   0.00003,     0.062 },
 { {   0,   0,   4,   0,  -7,   0,   0,   0,   2,  -1,   0 },    121.31820,   0.00004,     0.067 },
 { {   0,   0,   4,   0,  -7,   0,   0,   0,   2,   0,   0 },    300.78170,   0.00004,     0.036 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,  -2,   0,   0 },    285.38592,   0.00011,     0.047 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   0,   1,   0 },    109.36249,   0.00003,     0.060 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,  -1,   0 },    110.62698,   0.00024,     0.067 },
 { {   0,   0,   4,   0,  -6,   0,   0,   0,   2,   0,   0 },    290.56059,   0.00014,     0.035 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,  -3,   0 },    258.16613,   0.00002,     0.088 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,  -2,   0 },     78.30475,   0.00003,     0.529 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,  -1,   0 },    258.16613,   0.00002,     0.066 },
 { {   0,   0,   4,   0,  -6,   5,   0,   0,   2,   0,   0 },     78.30475,   0.00003,     0.035 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   0,   0 },    273.48030,   0.00031,     0.047 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,  -2,   1,   0 },    272.36606,   0.00006,     0.127 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,  -1,   0 },    101.35289,   0.00006,     0.103 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 },     96.29685,   0.00006,     0.279 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   0,   1,   0 },     99.97299,   0.00020,     0.059 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -2,   0 },    280.20576,   0.00007,     0.554 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,  -1,   0 },    100.28371,   0.00109,     0.066 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   0,   0 },    280.34272,   0.00059,     0.035 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   2,   1,   0 },    280.31700,   0.00010,     0.024 },
 { {   0,   0,   4,   0,  -5,   0,   0,   0,   4,   0,   0 },    100.33379,   0.00002,     0.019 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,  -1,   0 },    256.75512,   0.00009,     0.029 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   0,   0 },    256.62221,   0.00090,     0.047 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   1,   0 },    256.22312,   0.00028,     0.128 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,  -2,   2,   0 },     80.50037,   0.00002,     0.184 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,  -1,   0 },     81.94212,   0.00019,     0.104 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 },     41.42265,   0.00003,     0.273 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   1,   0 },     86.82147,   0.00049,     0.059 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   0,   2,   0 },     86.25688,   0.00003,     0.033 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -2,   0 },    269.10451,   0.00017,     0.529 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,  -1,   0 },     89.80532,   0.00325,     0.066 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   0,   0 },    269.68991,   0.00189,     0.035 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   2,   1,   0 },    269.57347,   0.00018,     0.024 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,  -2,   0 },     89.25698,   0.00004,     0.038 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,  -1,   0 },    269.63544,   0.00005,     0.025 },
 { {   0,   0,   4,   0,  -4,   0,   0,   0,   4,   0,   0 },    269.45704,   0.00003,     0.019 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -4,   0,   0 },    333.36120,   0.00004,     0.022 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,  -1,   0 },    154.05003,   0.00008,     0.029 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   0,   0 },    153.85664,   0.00090,     0.048 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   1,   0 },    153.99377,   0.00023,     0.129 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,  -2,   2,   0 },    333.76621,   0.00002,     0.181 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,  -1,   0 },     26.36430,   0.00012,     0.105 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 },    333.49274,   0.00008,     0.267 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   0,   1,   0 },    329.14238,   0.00010,     0.059 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,  -1,   0 },    275.52394,   0.00026,     0.066 },
 { {   0,   0,   4,   0,  -3,   0,   0,   0,   2,   0,   0 },     81.02757,   0.00015,     0.035 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -3,   0 },     98.51751,   0.00002,     0.088 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -2,   0 },    278.51751,   0.00003,     0.530 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,  -1,   0 },     98.51751,   0.00002,     0.066 },
 { {   0,   0,   4,   0,  -2,  -5,   0,   0,   2,   0,   0 },    278.51751,   0.00003,     0.035 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -4,   0,   0 },    242.81718,   0.00004,     0.022 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,  -1,   0 },     61.16850,   0.00005,     0.029 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   0,   0 },     61.49626,   0.00070,     0.048 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,  -2,   1,   0 },     60.01399,   0.00007,     0.131 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,  -2,   0 },    244.13395,   0.00002,     0.044 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,  -1,   0 },    245.00457,   0.00022,     0.106 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 },     69.88813,   0.00003,     0.261 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   0,   1,   0 },    239.93981,   0.00004,     0.059 },
 { {   0,   0,   4,   0,  -2,   0,   0,   0,   2,  -1,   0 },     65.72173,   0.00004,     0.065 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,  -2,   0,   0 },    155.00857,   0.00010,     0.048 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    227.91238,   0.00003,     0.065 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   2,   0,   0 },     81.42906,   0.00006,     0.035 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,   2,  -1,   0 },     89.88721,   0.00005,     0.065 },
 { {   0,   0,   4,   0,   0,  -4,   0,   0,   2,   0,   0 },    269.88721,   0.00007,     0.035 },
 { {   0,   0,   4,   0,   0,  -3,   0,   0,  -2,   0,   0 },    164.37688,   0.00004,     0.048 },
 { {   0,   0,   4,   0,   0,  -2,   0,   0,  -2,   0,   0 },     63.83503,   0.00004,     0.048 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -3,   0 },     36.54187,   0.00003,     0.091 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    218.11929,   0.00005,     0.449 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },    216.54187,   0.00003,     0.065 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,   0 },     38.07239,   0.00004,     0.035 },
 { {   0,   0,   4,   0,   1,   0,   0,   0,   2,  -1,   0 },     29.69063,   0.00003,     0.064 },
 { {   0,   0,   4,   0,   1,   0,   0,   0,   2,   0,   0 },    207.34397,   0.00006,     0.035 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,  -1,   0 },    216.92615,   0.00003,     0.094 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,   0,   0 },    216.92615,   0.00009,     0.042 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   0,   0 },    302.40289,   0.00010,     0.040 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   1,   0 },    302.16840,   0.00005,     0.085 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0 },    302.97555,   0.00018,     0.074 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   1,   0 },    122.97555,   0.00018,     0.077 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   2,  -1,   0 },    132.58440,   0.00007,     0.090 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   2,   0,   0 },    129.67365,   0.00022,     0.041 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -4,   1,   0 },    333.83555,   0.00004,     0.028 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,  -1,   0 },    333.83203,   0.00013,     0.026 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,   0 },    333.73403,   0.00120,     0.041 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,   0 },    334.04295,   0.00113,     0.089 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   2,   0 },    154.77639,   0.00006,     0.503 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0 },    331.03706,   0.00007,     0.038 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0 },    331.00559,   0.00113,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0 },    147.58020,   0.00007,     4.655 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0 },    150.84262,   0.00109,     0.074 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0 },    151.31109,   0.00007,     0.037 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -2,   0 },    334.32595,   0.00004,     0.641 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,   0 },    153.50799,   0.00091,     0.085 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,   0 },    153.19388,   0.00087,     0.040 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   1,   0 },    153.20391,   0.00009,     0.026 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   4,   0,   0 },    333.15886,   0.00003,     0.020 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -4,   1,   0 },      1.97702,   0.00005,     0.028 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,  -1,   0 },      1.97658,   0.00016,     0.027 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   0 },      1.98044,   0.00138,     0.042 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   1,   0 },      1.99935,   0.00110,     0.093 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   2,   0 },    181.86884,   0.00008,     0.397 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,  -1,   1 },    264.29635,   0.00006,     0.087 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0 },      1.80469,   0.00003,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0 },      3.62353,   0.00074,     0.080 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0 },    182.68980,   0.00022,     1.340 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0 },    181.86057,   0.00072,     0.071 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0 },    181.81280,   0.00004,     0.037 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   1 },    264.29636,   0.00033,     0.075 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   0,   0 },    182.81861,   0.00011,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -2,   0 },      3.07649,   0.00004,     0.973 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -1,   0 },    199.62949,   0.00048,     0.082 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,   0 },    183.10092,   0.00059,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   4,  -2,   0 },    315.22875,   0.00003,     0.042 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,  -2,   0,   0 },    264.66766,   0.00014,     0.042 },
 { {   0,   0,   5,  -8,   1,   5,   0,   0,  -2,   1,   0 },    264.66766,   0.00005,     0.095 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,  -1,   0 },     94.31858,   0.00003,     0.016 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   0,   0 },    274.39128,   0.00020,     0.021 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   1,   0 },    274.55822,   0.00053,     0.028 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -4,   2,   0 },    275.01742,   0.00011,     0.046 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -3,   0,   0 },    274.68618,   0.00005,     0.028 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -2,   0 },    275.12707,   0.00004,     0.020 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    274.29691,   0.00079,     0.027 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   0 },    274.81655,   0.01105,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   2 },     95.05514,   0.00002,     0.322 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   0 },    275.26126,   0.00710,     0.095 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   2,   0 },     95.76027,   0.00042,     0.361 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -2,   0 },    274.93650,   0.00020,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,   0 },    275.62980,   0.00307,     0.082 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0,   0 },     96.98288,   0.00087,     1.001 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,   0 },    102.12254,   0.00090,     0.070 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   2,   0 },    248.94681,   0.00002,     0.036 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   1,   0,   0 },     93.06527,   0.00004,     0.075 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -2,   0 },     97.03549,   0.00004,     1.291 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,  -1,   0 },    271.28214,   0.00039,     0.080 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   0,   0 },    271.31016,   0.00069,     0.039 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   2,   1,   0 },    273.39002,   0.00014,     0.026 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   4,  -1,   0 },    273.84698,   0.00002,     0.027 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -4,   1,   0 },    216.58567,   0.00004,     0.029 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,  -1,   0 },    216.56943,   0.00019,     0.027 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   0 },    216.47737,   0.00174,     0.043 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,   0 },    216.58810,   0.00112,     0.098 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   2,   0 },     37.02203,   0.00007,     0.328 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0 },    214.61214,   0.00007,     0.040 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0 },    213.56898,   0.00128,     0.083 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0 },     31.65637,   0.00048,     0.782 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0 },     33.58386,   0.00088,     0.069 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0 },     33.02367,   0.00004,     0.036 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -2,   0 },     33.08122,   0.00010,     2.017 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,   0 },    212.21648,   0.00153,     0.078 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,   0 },    215.51078,   0.00010,     0.038 },
 { {   0,   0,   5,  -6,  -1,   0,   0,   0,   2,  -3,   0 },     18.37437,   0.00002,     0.075 },
 { {   0,   0,   5,  -6,  -1,   0,   0,   0,   2,  -1,   0 },    198.37437,   0.00002,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -4,   0,   0 },     80.29851,   0.00004,     0.021 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,  -1,   0 },    280.52658,   0.00004,     0.028 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,   0 },    259.07552,   0.00098,     0.044 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   1,   0 },    255.88161,   0.00053,     0.103 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   2,   0 },     77.22579,   0.00002,     0.279 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -3,   0 },     61.80961,   0.00004,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0 },     61.53403,   0.00062,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0 },     61.26719,   0.00679,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0 },    239.21047,   0.00059,     0.553 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0 },     66.58733,   0.00040,     0.066 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0 },     67.48974,   0.00002,     0.035 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -4,   0 },     62.84688,   0.00002,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -3,   0 },     62.84714,   0.00038,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,   0 },    241.75140,   0.00228,    27.873 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,  -2 },    241.74175,   0.00015,     0.074 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   0 },     61.73416,   0.03974,     0.075 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,   0 },     61.84913,   0.00173,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   1,   0 },     61.79322,   0.00007,     0.025 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -3,   0 },    241.35211,   0.00024,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -2,   0 },     61.86449,   0.00016,     0.040 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,  -1,   0 },     61.74644,   0.00034,     0.026 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   4,   0,   0 },     61.74218,   0.00004,     0.019 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,   0 },    222.63169,   0.00030,     0.045 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   1,   0 },    216.33039,   0.00007,     0.109 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0 },     88.75219,   0.00003,     0.041 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0 },    100.56387,   0.00020,     0.092 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0 },     79.74761,   0.00017,     0.427 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0 },     81.61395,   0.00027,     0.064 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -2,   0 },    269.57913,   0.00015,     1.762 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,   0 },     89.68887,   0.00265,     0.072 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,   0 },    271.46138,   0.00034,     0.037 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   1,   0 },    270.46933,   0.00005,     0.025 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   4,  -2,   0 },     89.54427,   0.00003,     0.040 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   0,   0 },    176.58665,   0.00031,     0.046 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   1,   0 },    177.18236,   0.00005,     0.116 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0 },    340.90388,   0.00013,     0.096 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0 },    318.24162,   0.00004,     0.348 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0 },    337.96159,   0.00004,     0.062 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -1,   0 },    271.91260,   0.00014,     0.070 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,   0,   0 },    116.46504,   0.00015,     0.036 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,  -2,   0,   0 },     57.58329,   0.00005,     0.047 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   0,   0 },    142.39727,   0.00003,     0.048 },
 { {   0,   0,   5,   0,  -7,   0,   0,   0,   2,  -1,   0 },    110.50710,   0.00007,     0.063 },
 { {   0,   0,   5,   0,  -7,   0,   0,   0,   2,   0,   0 },    290.50710,   0.00012,     0.034 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,  -2,   0,   0 },    272.60645,   0.00009,     0.049 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   0,   1,   0 },    100.05169,   0.00003,     0.056 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,  -1,   0 },    100.05168,   0.00028,     0.063 },
 { {   0,   0,   5,   0,  -6,   0,   0,   0,   2,   0,   0 },    280.05168,   0.00022,     0.034 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,  -2,   0,   0 },    254.95317,   0.00021,     0.050 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,  -1,   0 },     95.63813,   0.00003,     0.115 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,   0,   0 },     64.16803,   0.00002,     0.218 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   0,   1,   0 },     86.88951,   0.00009,     0.056 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -2,   0 },    269.66650,   0.00003,     0.357 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,  -1,   0 },     88.99257,   0.00072,     0.062 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   0,   0 },    269.21068,   0.00051,     0.034 },
 { {   0,   0,   5,   0,  -5,   0,   0,   0,   2,   1,   0 },    269.12470,   0.00009,     0.023 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,  -2,   0,   0 },    152.75816,   0.00017,     0.050 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,  -1,   0 },    145.05633,   0.00002,     0.116 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,   0,   0 },    292.79045,   0.00003,     0.214 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   0,   1,   0 },    325.05633,   0.00002,     0.056 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   2,  -1,   0 },    299.31292,   0.00009,     0.062 },
 { {   0,   0,   5,   0,  -4,   0,   0,   0,   2,   0,   0 },    111.08564,   0.00015,     0.034 },
 { {   0,   0,   5,   0,  -3,   0,   0,   0,  -2,   0,   0 },     49.47699,   0.00009,     0.050 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,  -2,   0,   0 },    317.41719,   0.00005,     0.050 },
 { {   0,   0,   5,   0,  -1,   0,   0,   0,   2,  -2,   0 },     39.97636,   0.00004,     0.318 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,  -2,   0 },    266.94941,   0.00002,     0.046 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0 },    266.94941,   0.00002,     0.200 },
 { {   0,   0,   6, -16,   4,   5,   0,   0,   2,   0,   0 },      8.55429,   0.00012,     0.044 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   2,  -1,   0 },    193.39219,   0.00004,     0.095 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   2,   0,   0 },    193.39219,   0.00010,     0.042 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,  -1,   0 },    273.42700,   0.00003,     0.073 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   1,   0 },     93.42700,   0.00003,     0.078 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   2,   0,   0 },    103.94016,   0.00005,     0.041 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,  -1,   0 },    304.84436,   0.00009,     0.026 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   0 },    304.71099,   0.00080,     0.041 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   1,   0 },    305.27722,   0.00077,     0.088 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   2,   0 },    126.21801,   0.00004,     0.519 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0 },    303.24446,   0.00005,     0.038 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,   0 },    301.66206,   0.00083,     0.076 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   0,   0 },    121.59379,   0.00006,     6.602 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,   0 },    121.59391,   0.00084,     0.075 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   2,   0 },    122.02965,   0.00005,     0.038 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -2,   0 },    307.15005,   0.00002,     0.616 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -1,   0 },    124.78383,   0.00054,     0.086 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,   0 },    124.21426,   0.00066,     0.040 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   1,   0 },    124.25874,   0.00007,     0.026 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,  -1,   0 },    333.25453,   0.00004,     0.027 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,   0 },    333.24590,   0.00046,     0.042 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   1,   0 },    333.36463,   0.00030,     0.093 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,   0 },    332.30136,   0.00028,     0.080 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   0,   0 },    153.51573,   0.00010,     1.464 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,   0 },    152.37760,   0.00026,     0.072 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,  -1,   1 },     54.93216,   0.00010,     0.076 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,  -1,   0 },    154.12411,   0.00026,     0.082 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,   0,   0 },    154.10098,   0.00020,     0.039 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -4,   0,   0 },      7.55552,   0.00004,     0.021 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,  -1,   0 },    187.47257,   0.00011,     0.027 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,   0 },    187.40297,   0.00093,     0.043 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   1,   0 },    187.65176,   0.00064,     0.097 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   2,   0 },      6.36915,   0.00002,     0.335 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0 },    185.05892,   0.00002,     0.040 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0 },    184.62852,   0.00066,     0.083 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0 },      3.84451,   0.00027,     0.823 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0 },      4.53034,   0.00048,     0.069 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   2,   0 },      5.18377,   0.00002,     0.036 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,   0 },    183.23004,   0.00065,     0.079 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,  -1,   0 },    216.83928,   0.00009,     0.028 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,   0 },    221.16005,   0.00053,     0.043 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   1,   0 },    224.48756,   0.00014,     0.103 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0 },    212.60262,   0.00534,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0 },     33.03286,   0.00089,     0.573 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0 },     35.33806,   0.00034,     0.067 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -4,   0 },     31.52312,   0.00004,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -3,   0 },     31.52311,   0.00064,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   0 },     32.35522,   0.00169,    36.376 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,  -2 },     32.33625,   0.00011,     0.073 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   0 },    212.34364,   0.03048,     0.076 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,   0 },    212.34170,   0.00191,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   1,   0 },    212.37052,   0.00013,     0.025 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -3,   0 },    212.74059,   0.00025,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -2,   0 },    212.35542,   0.00059,     0.040 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,  -1,   0 },    212.33920,   0.00028,     0.026 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   4,   0,   0 },    212.33694,   0.00003,     0.020 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -4,   0,   0 },    351.95983,   0.00005,     0.021 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,  -1,   0 },    172.67893,   0.00005,     0.028 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   0,   0 },    172.40628,   0.00076,     0.044 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,  -2,   1,   0 },    173.30783,   0.00016,     0.105 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,  -1,   0 },    175.04634,   0.00004,     0.089 },
 { {   0,   0,   6,  -8,   3,   0,   0,   0,   0,   0,   0 },    352.71055,   0.00004,     0.500 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,   0 },    212.91765,   0.00024,     0.045 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   1,   0 },    204.62187,   0.00007,     0.109 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0 },     60.31727,   0.00003,     0.041 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0 },     66.51764,   0.00023,     0.091 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0 },     53.88842,   0.00012,     0.439 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0 },     55.27712,   0.00024,     0.064 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -2,   0 },    240.76173,   0.00015,     1.983 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,   0 },     60.84473,   0.00263,     0.073 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,   0 },    242.15676,   0.00028,     0.037 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   1,   0 },    241.42762,   0.00006,     0.025 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   4,  -2,   0 },     60.74194,   0.00003,     0.040 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   0,   0 },    239.65697,   0.00025,     0.046 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   1,   0 },    210.62283,   0.00002,     0.115 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0 },    318.83076,   0.00004,     0.042 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0 },    325.38905,   0.00026,     0.096 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0 },     96.29418,   0.00016,     0.356 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0 },     85.35425,   0.00015,     0.062 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -2,   0 },    269.91671,   0.00006,     0.965 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,   0 },     96.87273,   0.00129,     0.070 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,   0 },    270.47805,   0.00035,     0.036 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   1,   0 },    270.58258,   0.00004,     0.025 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   3,  -1,  -1 },     13.22059,   0.00005,     0.076 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,  -2,   0,   0 },    130.94289,   0.00009,     0.047 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,  -1,   0 },    299.85191,   0.00016,     0.067 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,   0,   0 },    118.64485,   0.00018,     0.036 },
 { {   0,   0,   6,   0,  -8,   0,   0,   0,   2,   0,   0 },    290.24811,   0.00004,     0.033 },
 { {   0,   0,   6,   0,  -7,   0,   0,   0,   2,  -1,   0 },     99.58455,   0.00006,     0.059 },
 { {   0,   0,   6,   0,  -7,   0,   0,   0,   2,   0,   0 },    279.58455,   0.00015,     0.033 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,  -2,   0,   0 },    252.13956,   0.00003,     0.052 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,  -1,   0 },     88.11714,   0.00013,     0.059 },
 { {   0,   0,   6,   0,  -6,   0,   0,   0,   2,   0,   0 },    268.11714,   0.00013,     0.033 },
 { {   0,   0,   6,   0,  -5,   0,   0,   0,   2,   0,   0 },    113.28438,   0.00005,     0.033 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   0,   0 },    265.61736,   0.00009,     0.039 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,  -2,   1,   0 },    265.61736,   0.00006,     0.080 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,  -1,   0 },     85.61736,   0.00016,     0.070 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   0,   0 },     85.61736,   0.00011,     1.000 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   0,   1,   0 },    265.61736,   0.00015,     0.082 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -2,   0 },     85.61736,   0.00002,     0.360 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,  -1,   0 },    265.61736,   0.00082,     0.095 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   0,   0 },    265.61736,   0.00114,     0.042 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   2,   1,   0 },    265.61736,   0.00012,     0.027 },
 { {   0,   0,   7, -16,   4,   5,   0,   0,   4,  -1,   0 },    265.61736,   0.00002,     0.028 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -2,   0,   0 },    169.79162,   0.00003,     0.039 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,  -1,   0 },    346.48222,   0.00005,     0.070 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   0,   0 },    349.50869,   0.00005,     1.025 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    166.49709,   0.00009,     0.081 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,  -1,   0 },    167.83108,   0.00015,     0.095 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   0,   0 },    168.01628,   0.00033,     0.042 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,  -2,   0,   0 },    206.21464,   0.00011,     0.040 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,  -2,   1,   0 },    206.77187,   0.00005,     0.087 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   0,  -1,   0 },    205.10657,   0.00012,     0.075 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   0,   1,   0 },     25.10657,   0.00012,     0.075 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   2,  -1,   0 },     26.76484,   0.00005,     0.087 },
 { {   0,   0,   7, -13,  -1,   0,   0,   0,   2,   0,   0 },     26.20735,   0.00011,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,  -1,   0 },    275.84608,   0.00009,     0.026 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   0 },    275.68420,   0.00078,     0.041 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,   0 },    276.47581,   0.00062,     0.088 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   2,   0 },     98.54645,   0.00002,     0.537 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0 },    272.48944,   0.00004,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -1,   0 },    272.72889,   0.00096,     0.076 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   0 },     93.75880,   0.00004,    11.350 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   1,   0 },     92.62789,   0.00096,     0.075 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   2,   0 },     92.51698,   0.00004,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -2,   0 },    277.84566,   0.00003,     0.593 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -1,   0 },     95.88929,   0.00057,     0.086 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,   0 },     95.25987,   0.00070,     0.040 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   1,   0 },     95.39371,   0.00007,     0.026 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   0,   0 },    304.45560,   0.00017,     0.041 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   1,   0 },    304.54040,   0.00012,     0.092 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,  -1,   0 },    303.75284,   0.00012,     0.079 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   0,   0 },    124.03851,   0.00002,     1.613 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   1,   0 },    123.75284,   0.00012,     0.072 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   2,  -1,   0 },    125.22277,   0.00010,     0.083 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   2,   0,   0 },    124.93262,   0.00014,     0.039 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,  -1,   0 },    158.98203,   0.00007,     0.027 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   0,   0 },    158.87784,   0.00055,     0.042 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   1,   0 },    159.13621,   0.00029,     0.097 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,   0 },    156.26620,   0.00040,     0.083 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   0,   0 },    334.99705,   0.00016,     0.868 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,   0 },    335.96309,   0.00030,     0.069 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,  -1,   0 },    154.62553,   0.00025,     0.079 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,   0,   0 },    158.69219,   0.00004,     0.039 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   0,   0 },    191.08491,   0.00022,     0.043 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   1,   0 },    192.51324,   0.00008,     0.102 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0 },    183.05836,   0.00005,     0.040 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -1,   0 },    183.68952,   0.00096,     0.086 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   0,   0 },      4.33329,   0.00020,     0.594 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   1,   0 },      6.59053,   0.00016,     0.067 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -3,   0 },      3.46326,   0.00004,     0.075 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -2,   0 },      3.12571,   0.00026,    11.006 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,   0 },    183.14660,   0.00463,     0.076 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,   0 },    183.23292,   0.00036,     0.038 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   4,  -2,   0 },    183.12625,   0.00005,     0.041 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   4,  -1,   0 },    183.12666,   0.00004,     0.026 },
 { {   0,   0,   7, -10,   1,   0,   0,   0,   2,  -1,   0 },     56.09007,   0.00006,     0.075 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   0,   0 },    181.93757,   0.00024,     0.044 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   1,   0 },    171.84221,   0.00003,     0.108 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0 },     35.93413,   0.00018,     0.091 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0 },     24.49076,   0.00006,     0.452 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0 },     26.50549,   0.00015,     0.065 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -2,   0 },    211.65876,   0.00010,     2.269 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,   0 },     31.80346,   0.00185,     0.073 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,   0 },    212.63440,   0.00020,     0.037 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   1,   0 },    212.37421,   0.00003,     0.025 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   0,   0 },    225.90354,   0.00023,     0.045 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   1,   0 },    212.82178,   0.00002,     0.114 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0 },     99.97252,   0.00026,     0.095 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0 },     46.85542,   0.00011,     0.364 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0 },     59.16502,   0.00015,     0.062 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -2,   0 },    241.22517,   0.00007,     1.028 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -1,   0 },     55.57242,   0.00111,     0.070 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,   0 },    241.50305,   0.00031,     0.036 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   1,   0 },    241.44329,   0.00006,     0.025 },
 { {   0,   0,   7,  -8,   3,   0,   0,   0,  -2,   0,   0 },     69.93002,   0.00006,     0.046 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,  -2,   0,   0 },    257.08176,   0.00009,     0.047 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0 },     95.02876,   0.00003,     0.100 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0 },     84.58471,   0.00009,     0.305 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0 },     87.80809,   0.00007,     0.060 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,   0 },     90.06018,   0.00054,     0.068 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,   0 },    270.41632,   0.00027,     0.036 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   1,   0 },     90.41742,   0.00002,     0.024 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   0,   0 },    133.14318,   0.00005,     0.048 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -1,   0 },    299.84573,   0.00012,     0.065 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,   0,   0 },    119.84432,   0.00018,     0.035 },
 { {   0,   0,   7,   0,  -8,   0,   0,   0,   2,   0,   0 },    278.93110,   0.00004,     0.032 },
 { {   0,   0,   7,   0,  -7,   0,   0,   0,   2,   0,   0 },    266.69843,   0.00007,     0.032 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   2,  -1,   0 },    142.43473,   0.00003,     0.096 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   2,   0,   0 },    142.43473,   0.00009,     0.042 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   0,   0 },    342.68044,   0.00012,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   1,   0 },    342.68044,   0.00024,     0.028 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -4,   2,   0 },    342.68044,   0.00012,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -3,   0,   0 },    342.68044,   0.00005,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -2,   0 },    342.68044,   0.00005,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,   0 },    342.68044,   0.00061,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,  -1,   2 },    162.68044,   0.00002,     0.090 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,   0 },    342.68044,   0.01073,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   0,   2 },    162.68044,   0.00003,     0.474 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,   0 },    342.68044,   0.01377,     0.087 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   1,   2 },    162.68044,   0.00002,     0.065 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   2,   0 },    162.68044,   0.00093,     0.564 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -2,   3,   0 },    162.68044,   0.00005,     0.067 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,  -1,   0,   0 },    162.68044,   0.00019,     0.081 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -2,   0 },    162.68044,   0.00014,     0.038 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,  -1,   0 },    162.68044,   0.00181,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   1,   0 },    342.68044,   0.00181,     0.075 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   0,   2,   0 },    342.68044,   0.00014,     0.038 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   1,   0,   0 },    342.68044,   0.00019,     0.081 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -3,   0 },    342.68044,   0.00005,     0.067 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -2,   0 },    342.68044,   0.00093,     0.564 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,  -2 },    342.68044,   0.00002,     0.065 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,  -1,   0 },    162.68044,   0.01377,     0.087 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,  -2 },    342.68044,   0.00003,     0.474 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   0,   0 },    162.68044,   0.01073,     0.040 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,  -2 },    342.68044,   0.00002,     0.090 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   1,   0 },    162.68044,   0.00061,     0.026 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   2,   2,   0 },    162.68044,   0.00005,     0.020 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   3,   0,   0 },    162.68044,   0.00005,     0.027 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -2,   0 },    162.68044,   0.00012,     0.044 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,  -1,   0 },    162.68044,   0.00024,     0.028 },
 { {   0,   0,   8, -16,   4,   5,   0,   0,   4,   0,   0 },    162.68044,   0.00012,     0.020 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,  -2,   0,   0 },    351.02476,   0.00014,     0.040 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,  -2,   1,   0 },    351.02476,   0.00008,     0.087 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,   2,  -1,   0 },    171.02476,   0.00008,     0.087 },
 { {   0,   0,   8, -16,   6,   0,   0,   0,   2,   0,   0 },    171.02476,   0.00014,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   0,   0 },    246.81154,   0.00003,     0.020 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   1,   0 },    246.82549,   0.00009,     0.028 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -4,   2,   0 },    246.82434,   0.00003,     0.044 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -1,   0 },    246.25345,   0.00026,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   0 },    246.73187,   0.00331,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,   0 },    247.36727,   0.00330,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   2,   0 },     68.52235,   0.00017,     0.556 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -1,   0,   0 },     66.69409,   0.00007,     0.081 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0 },    243.78938,   0.00025,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   0 },    243.68328,   0.00452,     0.076 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   0 },     61.77892,   0.00005,    40.427 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   0 },     63.66033,   0.00451,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   2,   0 },     63.74042,   0.00025,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   1,   0,   0 },    246.77544,   0.00007,     0.081 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -2,   0 },    248.39903,   0.00016,     0.572 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,   0 },     67.23320,   0.00319,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,   0 },     66.56846,   0.00317,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   1,   0 },     66.11032,   0.00025,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,  -2,   0 },     66.69363,   0.00003,     0.044 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,  -1,   0 },     66.66914,   0.00008,     0.028 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   4,   0,   0 },     66.64989,   0.00003,     0.020 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   0,   0 },    275.61440,   0.00013,     0.041 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   1,   0 },    275.61431,   0.00005,     0.092 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0 },    271.99356,   0.00004,     0.079 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   1,   0 },     91.99356,   0.00004,     0.072 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   2,  -1,   0 },     94.99472,   0.00006,     0.083 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   2,   0,   0 },     95.89961,   0.00006,     0.040 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,  -1,   0 },    131.70782,   0.00004,     0.027 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0 },    131.57589,   0.00041,     0.042 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   1,   0 },    132.16977,   0.00020,     0.096 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0 },    128.32928,   0.00026,     0.082 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   0,   0 },    306.04008,   0.00010,     0.919 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   1,   0 },    307.57830,   0.00021,     0.070 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,  -1,   0 },    126.36973,   0.00010,     0.080 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,   0,   0 },    307.63577,   0.00002,     0.039 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   0,   0 },    160.66527,   0.00018,     0.043 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   1,   0 },    160.66535,   0.00005,     0.101 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,  -1,   0 },    154.87028,   0.00030,     0.086 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   0,   0 },    335.11341,   0.00009,     0.617 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   1,   0 },    336.93573,   0.00007,     0.067 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -2,   0 },    334.13182,   0.00007,     6.484 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -1,   0 },    154.18175,   0.00126,     0.076 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,   0,   0 },    154.39876,   0.00022,     0.038 },
 { {   0,   0,   8, -12,   2,   0,   0,   0,   2,  -1,   0 },     39.28761,   0.00006,     0.075 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   0,   0 },    141.69781,   0.00011,     0.044 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,  -1,   0 },      5.62958,   0.00014,     0.090 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   0,   0 },    348.87231,   0.00006,     0.465 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   1,   0 },    356.34026,   0.00006,     0.065 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -2,   0 },    182.74584,   0.00006,     2.649 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,  -2 },    183.55982,   0.00006,     0.076 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,   0 },      2.85101,   0.00113,     0.073 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,   0 },    184.15133,   0.00005,     0.037 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   1,  -2 },      3.55982,   0.00006,     0.075 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   0,   0 },    200.61546,   0.00014,     0.045 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0 },     52.71594,   0.00008,     0.095 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   0,   0 },     28.00147,   0.00010,     0.373 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   1,   0 },     30.64478,   0.00010,     0.063 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -2,   0 },    212.47982,   0.00004,     1.100 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -1,   0 },     32.42678,   0.00084,     0.071 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,   0 },    212.41433,   0.00022,     0.036 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   0,   0 },    235.37200,   0.00012,     0.046 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0 },     64.75496,   0.00003,     0.100 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0 },     58.63505,   0.00010,     0.311 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0 },     60.43728,   0.00009,     0.061 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -2,   0 },    243.40024,   0.00003,     0.694 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,   0 },     61.38117,   0.00071,     0.068 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,   0 },    241.13759,   0.00027,     0.036 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   1,   0 },    241.26449,   0.00003,     0.024 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,  -2,   0,   0 },    249.70001,   0.00004,     0.048 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0 },     89.54629,   0.00002,     0.105 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0 },     90.06531,   0.00003,     0.059 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -1,   0 },     89.86017,   0.00030,     0.066 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,   0 },    270.06242,   0.00018,     0.035 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,  -1,   0 },    300.57836,   0.00010,     0.063 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,   0,   0 },    120.57721,   0.00017,     0.034 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,  -1,   0 },    264.22414,   0.00004,     0.070 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   0,   0 },    264.22427,   0.00003,     1.000 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   0,   1,   0 },     84.22414,   0.00004,     0.082 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,  -1,   0 },     84.22422,   0.00013,     0.095 },
 { {   0,   0,   9, -19,   0,   3,   0,   0,   2,   0,   0 },     84.22421,   0.00024,     0.042 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,  -1,   0 },    217.72004,   0.00004,     0.026 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,   0 },    217.56500,   0.00046,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   1,   0 },    218.35070,   0.00036,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -2,   0 },    214.78686,   0.00003,     0.038 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,   0 },    215.09222,   0.00071,     0.075 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,   0 },     35.14432,   0.00071,     0.076 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   2,   0 },     34.79678,   0.00003,     0.038 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -2,   0 },    218.04748,   0.00003,     0.552 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,   0 },     38.73228,   0.00038,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,   0 },     37.85083,   0.00048,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   1,   0 },     38.03477,   0.00004,     0.026 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   0,   0 },    246.74992,   0.00005,     0.041 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,  -1,   0 },    246.61556,   0.00002,     0.078 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   1,   0 },     66.61556,   0.00002,     0.073 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -4,   1,   0 },    239.74351,   0.00004,     0.028 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,  -1,   0 },    239.74351,   0.00012,     0.027 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   0,   0 },    239.74351,   0.00117,     0.042 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   1,   0 },    239.74351,   0.00082,     0.095 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,  -2,   2,   0 },     59.74351,   0.00006,     0.360 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,  -1,   0 },    239.74351,   0.00017,     0.082 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   0,   0 },     59.74351,   0.00009,     1.000 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   0,   1,   0 },     59.74351,   0.00016,     0.070 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,  -1,   0 },    239.74351,   0.00006,     0.080 },
 { {   0,   0,   9, -16,   4,   5,   0,   0,   2,   0,   0 },    239.74351,   0.00009,     0.039 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -4,   0,   0 },    295.21592,   0.00003,     0.021 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,  -1,   0 },    111.75303,   0.00009,     0.027 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,   0 },    113.90049,   0.00083,     0.042 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   1,   0 },    110.15866,   0.00055,     0.096 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,  -1,   0 },    105.38908,   0.00039,     0.082 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   0,   0 },    284.58169,   0.00014,     0.976 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   1,   0 },    279.36856,   0.00029,     0.070 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,  -1,  -1 },    358.08074,   0.00002,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,   0 },    101.62530,   0.00004,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   1,  -1 },    178.08074,   0.00002,     0.076 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,   0,   0 },    260.92840,   0.00009,     0.039 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,  -2,   0,   0 },    130.52604,   0.00008,     0.043 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,  -1,   0 },    126.35735,   0.00009,     0.085 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,   1,   0 },    307.81486,   0.00002,     0.068 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -1,   0 },    125.88709,   0.00040,     0.077 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,   0,   0 },    125.54628,   0.00007,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,  -2,   0,   0 },     88.38527,   0.00006,     0.044 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,  -1,   0 },    335.88395,   0.00009,     0.090 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   0,   0 },    334.50392,   0.00003,     0.479 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   1,  -2 },    334.61893,   0.00003,     0.087 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   1,   0 },    327.26965,   0.00003,     0.065 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -2,   0 },    154.05130,   0.00004,     3.184 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,  -2 },    154.65635,   0.00020,     0.075 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,   0 },    333.92324,   0.00066,     0.074 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,   0 },    154.63393,   0.00012,     0.037 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   1,  -2 },    334.65635,   0.00020,     0.076 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   4,  -1,  -2 },    154.61893,   0.00003,     0.087 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,  -2,   0,   0 },    171.71335,   0.00007,     0.045 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,  -1,   0 },      3.48317,   0.00005,     0.094 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,   1,   0 },      3.35662,   0.00004,     0.063 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,   0 },      3.42688,   0.00049,     0.071 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,   0 },    183.42159,   0.00011,     0.037 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,  -2,   0,   0 },    208.25968,   0.00007,     0.046 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0 },     33.30017,   0.00003,     0.099 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   0,   0 },     31.02479,   0.00006,     0.317 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   1,   0 },     31.83471,   0.00006,     0.061 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -2,   0 },    214.98186,   0.00003,     0.726 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -1,   0 },     32.36479,   0.00051,     0.068 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,   0 },    212.01515,   0.00022,     0.036 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,  -2,   0,   0 },    233.60310,   0.00007,     0.047 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0 },     61.70370,   0.00002,     0.104 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   0,   0 },     58.56540,   0.00006,     0.272 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   1,   0 },     60.98344,   0.00005,     0.059 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,   0 },     61.22736,   0.00043,     0.066 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,   0 },    241.92578,   0.00024,     0.035 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,  -1,   0 },     89.46974,   0.00016,     0.064 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,   0,   0 },    269.46837,   0.00008,     0.035 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,  -1,  -4 },     94.41056,   0.00007,     0.075 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   4,   1,  -4 },    274.41056,   0.00007,     0.075 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,  -1,   0 },    301.02708,   0.00007,     0.062 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,   0,   0 },    121.02611,   0.00014,     0.034 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   0,   0 },    188.56215,   0.00004,     0.040 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,  -1,   0 },    188.42399,   0.00003,     0.075 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   1,   0 },      8.42399,   0.00003,     0.076 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   2,   0,   0 },      9.59175,   0.00005,     0.041 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   0,   0 },    161.28735,   0.00002,     0.020 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   1,   0 },    161.28735,   0.00004,     0.028 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -4,   2,   0 },    161.28735,   0.00002,     0.044 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,  -1,   0 },    161.28735,   0.00016,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   0 },    161.28735,   0.00214,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,   0 },    161.28735,   0.00269,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   2,   0 },    341.28735,   0.00017,     0.564 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -1,   0,   0 },    341.28739,   0.00005,     0.081 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,  -2,   0 },    341.28728,   0.00003,     0.038 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,  -1,   0 },    341.28728,   0.00046,     0.075 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   1,   0 },    161.28735,   0.00046,     0.075 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   0,   2,   0 },    161.28734,   0.00003,     0.038 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   1,   0,   0 },    161.28723,   0.00005,     0.081 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -2,   0 },    161.28727,   0.00017,     0.564 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,   0 },    341.28727,   0.00269,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,   0 },    341.28727,   0.00215,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   1,   0 },    341.28727,   0.00016,     0.026 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,  -2,   0 },    341.28727,   0.00002,     0.044 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,  -1,   0 },    341.28727,   0.00004,     0.028 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   4,   0,   0 },    341.28727,   0.00002,     0.020 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   0,   0 },    232.20069,   0.00010,     0.042 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   1,   0 },    232.20071,   0.00008,     0.095 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,  -1,   0 },    240.94120,   0.00009,     0.081 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   0,   0 },     63.47744,   0.00002,     1.040 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   1,   0 },     60.94120,   0.00009,     0.070 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,  -3,   0 },    247.47889,   0.00003,     0.071 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,  -1,   0 },     67.47889,   0.00003,     0.080 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,   0,   0 },     70.54402,   0.00004,     0.039 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,  -2,   0,   0 },    100.77046,   0.00004,     0.043 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,  -1,   0 },     96.62790,   0.00013,     0.077 },
 { {   0,   0,  10, -16,   4,   5,   0,   0,  -2,   0,   0 },    136.80658,   0.00012,     0.044 },
 { {   0,   0,  10, -16,   4,   5,   0,   0,  -2,   1,   0 },    136.80658,   0.00003,     0.105 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,  -2,   0,   0 },     19.09643,   0.00011,     0.044 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   0,  -1,   0 },    302.38303,   0.00006,     0.089 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,   0 },    302.38303,   0.00036,     0.074 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,   0 },    125.50343,   0.00007,     0.037 },
 { {   0,   0,  10, -15,   1,   0,   0,   0,   2,  -1,  -2 },    339.58698,   0.00002,     0.075 },
 { {   0,   0,  10, -15,   1,   0,   0,   0,   2,   1,  -2 },    159.58698,   0.00002,     0.075 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,  -2,   0,   0 },    143.19056,   0.00003,     0.045 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,  -1,   0 },    334.42345,   0.00003,     0.093 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,   0,   0 },    334.44171,   0.00003,     0.391 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,   1,   0 },    334.47172,   0.00002,     0.063 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -1,   0 },    334.44341,   0.00027,     0.071 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,   0 },    154.43949,   0.00016,     0.037 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,  -2,   0,   0 },    181.56385,   0.00004,     0.046 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,  -1,   0 },      3.47546,   0.00002,     0.098 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,   1,   0 },      3.30174,   0.00003,     0.061 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,  -1,   0 },      3.37646,   0.00030,     0.069 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   0,   0 },    183.37089,   0.00010,     0.036 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   0,   0 },    207.58611,   0.00004,     0.047 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,   1,   0 },     32.11411,   0.00004,     0.059 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,   0 },     32.61940,   0.00032,     0.066 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,   0,   0 },    212.27918,   0.00019,     0.035 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,  -2,   0,   0 },    229.98306,   0.00003,     0.049 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,   1,   0 },     60.74303,   0.00003,     0.057 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -1,   0 },     61.61505,   0.00026,     0.064 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,   0,   0 },    240.99605,   0.00018,     0.035 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   2,  -1,   0 },     88.48540,   0.00008,     0.062 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   2,   0,   0 },    268.48593,   0.00016,     0.034 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,  -1,   0 },    301.27209,   0.00005,     0.060 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,   0,   0 },    121.27124,   0.00011,     0.033 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   0,   0 },    218.08636,   0.00011,     0.040 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   1,   0 },    219.95674,   0.00011,     0.087 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,  -1,   0 },    213.88269,   0.00024,     0.075 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,   1,   0 },     33.88269,   0.00024,     0.075 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,  -1,   0 },     39.92187,   0.00011,     0.087 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,   0,   0 },     38.05797,   0.00011,     0.040 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   0,   0 },    213.83687,   0.00004,     0.042 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,   0,   0 },     58.35041,   0.00023,     0.042 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,  -2,   1,   0 },     58.35041,   0.00013,     0.095 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,  -1,   0 },     58.35047,   0.00004,     0.082 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   0,   0 },    238.35036,   0.00003,     1.000 },
 { {   0,   0,  11, -19,   0,   3,   0,   0,   0,   1,   0 },    238.35047,   0.00004,     0.070 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -1,   0 },     67.70399,   0.00004,     0.077 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   0,   0 },    120.05598,   0.00003,     0.044 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   0,  -1,   0 },    276.34896,   0.00004,     0.089 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,  -1,   0 },    276.34907,   0.00022,     0.074 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,  -1,   0 },    305.47257,   0.00014,     0.072 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,   0 },    125.46251,   0.00008,     0.037 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   0,   0,   0 },    334.37327,   0.00003,     0.331 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,  -1,   0 },    334.39571,   0.00017,     0.069 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,   0,   0 },    154.39362,   0.00015,     0.036 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   0,   1,   0 },      3.21488,   0.00002,     0.059 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -1,   0 },      3.30044,   0.00021,     0.066 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,   0,   0 },    183.29446,   0.00008,     0.035 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   0,   1,   0 },     32.13700,   0.00002,     0.058 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,  -1,   0 },     32.13649,   0.00022,     0.064 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,   0,   0 },    212.13639,   0.00015,     0.035 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0 },    247.86912,   0.00002,     0.116 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   0,   1,   0 },     67.86912,   0.00002,     0.056 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,  -1,   0 },     60.67863,   0.00016,     0.062 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   0,   0 },    240.67840,   0.00013,     0.034 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   2,  -1,   0 },     86.74821,   0.00004,     0.060 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   2,   0,   0 },    266.74895,   0.00009,     0.033 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,  -1,   0 },    301.35081,   0.00004,     0.058 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,   0,   0 },    121.35005,   0.00009,     0.033 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,   2,  -1,   0 },    247.62406,   0.00014,     0.075 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -1,   0 },    275.68777,   0.00007,     0.072 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,  -1,   0 },    305.42273,   0.00009,     0.069 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,   0,   0 },    125.41536,   0.00008,     0.036 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   0,   0,   0 },    334.29212,   0.00003,     0.286 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -1,   0 },    334.33095,   0.00012,     0.067 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,   0,   0 },    154.33025,   0.00014,     0.035 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   0,  -1,   0 },    183.10101,   0.00002,     0.108 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   0,   1,   0 },      3.10101,   0.00002,     0.058 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -1,   0 },      3.19566,   0.00015,     0.064 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,   0,   0 },    183.58896,   0.00023,     0.035 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,  -1,   0 },     31.94652,   0.00015,     0.062 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   0,   0 },    211.94640,   0.00012,     0.034 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,  -1,   0 },     60.25368,   0.00010,     0.060 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,   0,   0 },    240.25305,   0.00006,     0.034 },
 { {   0,   0,  12, -12,   0,   0,   0,   0,   2,   0,   0 },    263.14169,   0.00004,     0.033 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,   0,   0 },    121.27726,   0.00007,     0.032 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -1,   0 },    218.26402,   0.00012,     0.075 },
 { {   0,   0,  13, -20,   0,   0,   0,   0,   2,  -1,   0 },    246.06123,   0.00004,     0.072 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,   2,  -1,   0 },    276.45732,   0.00004,     0.070 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,   2,   0,   0 },     96.43789,   0.00004,     0.036 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,  -1,   0 },    305.36225,   0.00006,     0.067 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,   0,   0 },    125.35695,   0.00007,     0.036 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,  -1,   0 },    334.24592,   0.00008,     0.065 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,   0,   0 },    154.24631,   0.00013,     0.035 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   0,   0,   0 },      2.94184,   0.00004,     0.223 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,  -1,   0 },      3.06893,   0.00010,     0.063 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,   0,   0 },    183.06897,   0.00019,     0.034 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,  -1,   0 },     31.70367,   0.00010,     0.061 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,   0,   0 },    211.70336,   0.00006,     0.034 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   0,   0,   0 },     59.36730,   0.00003,     0.180 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,  -1,   0 },     59.69149,   0.00006,     0.059 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,   0,   0 },    239.69168,   0.00015,     0.033 },
 { {   0,   0,  13, -12,   0,   0,   0,   0,   2,   0,   0 },    121.05193,   0.00005,     0.032 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -2,   0 },    188.66413,   0.00011,     0.040 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   0,  -1,   0 },      8.67264,   0.00012,     0.087 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -3,   0 },    187.86185,   0.00023,     0.075 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   0 },    188.81636,   0.00007,   362.998 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,   0 },      8.66809,   0.00150,     0.075 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,   0,   0 },      8.66809,   0.00008,     0.038 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -3,   0 },      8.98748,   0.00010,     0.087 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   4,  -2,   0 },      8.66413,   0.00011,     0.040 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,  -1,   0 },    305.28760,   0.00005,     0.065 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,   0,   0 },    125.28392,   0.00007,     0.035 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,  -1,   0 },    335.52842,   0.00006,     0.063 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,   0,   0 },    154.81188,   0.00011,     0.034 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,  -1,   0 },      2.89875,   0.00007,     0.061 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,   0,   0 },    182.89880,   0.00016,     0.034 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   2,  -1,   0 },     31.40073,   0.00007,     0.059 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   2,   0,   0 },    211.40081,   0.00016,     0.033 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,   2,  -1,   0 },     58.94522,   0.00004,     0.057 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,   2,   0,   0 },    238.94545,   0.00010,     0.033 },
 { {   0,   0,  14, -13,   0,   0,   0,   0,   2,   0,   0 },    120.65860,   0.00003,     0.031 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,  -1,   0 },    305.19391,   0.00003,     0.063 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,   0,   0 },    125.19393,   0.00006,     0.034 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,  -1,   0 },    334.01038,   0.00005,     0.061 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,   0,   0 },    154.01041,   0.00010,     0.034 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   2,  -1,   0 },      2.69086,   0.00005,     0.059 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   2,   0,   0 },    182.69091,   0.00012,     0.033 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   2,  -1,   0 },     31.02820,   0.00004,     0.057 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   2,   0,   0 },    211.02830,   0.00012,     0.033 },
 { {   0,   0,  15, -16,   0,   0,   0,   0,   2,   0,   0 },    237.93740,   0.00006,     0.032 },
 { {   0,   0,  16, -21,   0,   0,   0,   0,   2,   0,   0 },    125.08560,   0.00005,     0.034 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,  -1,   0 },    333.85177,   0.00003,     0.059 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,   0,   0 },    153.85181,   0.00008,     0.033 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   2,  -1,   0 },      2.44152,   0.00004,     0.058 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   2,   0,   0 },    182.44158,   0.00010,     0.033 },
 { {   0,   0,  16, -18,   0,   0,   0,   0,   2,  -1,   0 },     30.57340,   0.00003,     0.056 },
 { {   0,   0,  16, -18,   0,   0,   0,   0,   2,   0,   0 },    210.57352,   0.00008,     0.032 },
 { {   0,   0,  16, -17,   0,   0,   0,   0,   2,   0,   0 },    236.52838,   0.00004,     0.032 },
 { {   0,   0,  17, -22,   0,   0,   0,   0,   2,   0,   0 },    124.95661,   0.00005,     0.033 },
 { {   0,   0,  17, -21,   0,   0,   0,   0,   2,   0,   0 },    153.66404,   0.00007,     0.033 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,   2,   0,   0 },    182.14627,   0.00007,     0.032 },
 { {   0,   0,  17, -19,   0,   0,   0,   0,   2,   0,   0 },    210.01898,   0.00006,     0.032 },
 { {   0,   0,  18, -23,   0,   0,   0,   0,   2,   0,   0 },    124.80547,   0.00004,     0.033 },
 { {   0,   0,  18, -22,   0,   0,   0,   0,   2,   0,   0 },    153.44478,   0.00005,     0.032 },
 { {   0,   0,  18, -21,   0,   0,   0,   0,   2,   0,   0 },    181.79936,   0.00005,     0.032 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,   2,   0,   0 },    209.33999,   0.00004,     0.031 },
 { {   0,   0,  19, -24,   0,   0,   0,   0,   2,   0,   0 },    124.63064,   0.00003,     0.032 },
 { {   0,   0,  19, -23,   0,   0,   0,   0,   2,   0,   0 },    153.19136,   0.00004,     0.032 },
 { {   0,   0,  19, -22,   0,   0,   0,   0,   2,   0,   0 },    181.39377,   0.00004,     0.031 },
 { {   0,   0,  20, -24,   0,   0,   0,   0,   2,   0,   0 },    152.90064,   0.00003,     0.031 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -4,   1,   0 },     57.06701,   0.00002,     0.040 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   0,   0 },     57.06701,   0.00008,     0.076 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   2,   0 },    237.06701,   0.00008,     0.075 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,   0,   1,   0 },    237.06701,   0.00002,     0.040 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   0,   0 },    118.79619,   0.00002,     0.076 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   2,   0 },    298.79619,   0.00002,     0.075 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   0,   0 },     35.75894,   0.00003,     0.075 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   2,   0 },    215.75894,   0.00003,     0.076 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,  -2,   0,   0 },    235.52467,   0.00011,     0.040 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,  -2,   1,   0 },    235.52467,   0.00007,     0.087 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   0,  -1,   0 },    234.84768,   0.00010,     0.075 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   0,   1,   0 },     54.84768,   0.00010,     0.075 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   2,  -1,   0 },     55.52477,   0.00007,     0.087 },
 { {   0,   1,  -8,  12,   0,   0,   0,   0,   2,   0,   0 },     55.52477,   0.00011,     0.040 },
 { {   0,   1,  -7,   0,   2,   0,   0,   0,  -2,   0,   0 },    321.68138,   0.00004,     0.033 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.14466,   0.00003,     0.034 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    252.14466,   0.00003,     0.385 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    248.47477,   0.00004,     0.229 },
 { {   0,   1,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },    141.21264,   0.00004,     0.034 },
 { {   0,   1,  -6,   0,   2,   0,   0,   0,  -2,   0,   0 },     21.78204,   0.00003,     0.035 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    142.19356,   0.00002,     0.036 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    322.02408,   0.00003,     0.067 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    322.02436,   0.00004,     0.625 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    142.02408,   0.00003,     0.086 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -2,   1,   0 },     84.32842,   0.00008,     0.075 },
 { {   0,   1,  -5,   3,   0,   0,   0,   0,  -2,   3,   0 },    283.96996,   0.00005,     0.075 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,  -2,   0,   0 },    144.54091,   0.00005,     0.037 },
 { {   0,   1,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0 },    152.34108,   0.00009,     0.072 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    157.13951,   0.00003,     0.037 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    339.08501,   0.00007,     0.072 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },     59.65308,   0.00002,     1.665 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    268.46578,   0.00003,     0.079 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    340.13853,   0.00004,     0.421 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    152.49345,   0.00011,     0.045 },
 { {   0,   1,  -4,   0,   2,   0,   0,   0,   2,  -2,   0 },    180.68012,   0.00002,     0.251 },
 { {   0,   1,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },    180.68012,   0.00002,     0.044 },
 { {   0,   1,  -4,   1,   0,   0,   0,   0,  -2,   1,   0 },     27.13144,   0.00003,     0.075 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   0,  -1,   0 },    119.73664,   0.00003,     0.074 },
 { {   0,   1,  -4,   4,   0,   0,   0,   0,   0,   1,   0 },    299.73664,   0.00003,     0.077 },
 { {   0,   1,  -3,   0,  -2,   0,   0,   0,  -2,   1,   0 },    320.08268,   0.00006,     0.077 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -3,   1,   1 },     76.98910,   0.00002,     0.085 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     47.07604,   0.00003,     0.038 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     68.02663,   0.00095,     0.078 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    236.58668,   0.00004,     2.503 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    265.64823,   0.00004,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    256.46533,   0.00016,     0.074 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },     66.22743,   0.00015,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },     76.53189,   0.00006,     0.072 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },     76.46533,   0.00016,     0.077 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    248.23147,   0.00046,     0.068 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    246.38557,   0.00037,     0.728 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     73.31671,   0.00068,     0.084 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    256.98910,   0.00002,     0.090 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    246.96616,   0.00005,     0.091 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     79.36906,   0.00008,     0.041 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    253.86758,   0.00003,     0.318 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     66.47648,   0.00012,     0.099 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     67.88464,   0.00027,     0.043 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,   2,  -2,   0 },    179.35628,   0.00003,     0.326 },
 { {   0,   1,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    359.35628,   0.00003,     0.043 },
 { {   0,   1,  -3,   0,   2,   0,   0,   0,  -2,   1,   0 },    268.66527,   0.00004,     0.079 },
 { {   0,   1,  -3,   0,   3,   0,   0,   0,   2,   0,   0 },     80.75343,   0.00005,     0.042 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,  -3,   0 },     90.20649,   0.00002,     0.065 },
 { {   0,   1,  -3,   2,   0,   0,   0,   0,   2,  -1,   0 },    270.20649,   0.00002,     0.090 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,  -2,   1,   0 },    294.02656,   0.00002,     0.084 },
 { {   0,   1,  -2,   0,  -1,   0,   0,   0,  -2,   3,   0 },    294.02656,   0.00002,     0.069 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },     12.32092,   0.00003,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },    214.28915,   0.00005,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   0,   1 },    346.60205,   0.00002,     0.042 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   1,   1 },    166.65797,   0.00004,     0.093 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    208.03638,   0.00035,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    210.80407,   0.00222,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },    166.61446,   0.00005,     0.086 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    295.03789,   0.00098,     0.084 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.52241,   0.00014,     0.715 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },    183.89732,   0.00003,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    346.64829,   0.00039,     0.080 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    347.24798,   0.00003,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    178.42271,   0.00029,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    165.35536,   0.00012,     1.472 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.62057,   0.00078,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.64311,   0.00032,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,  -1 },    346.62296,   0.00004,     2.892 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    195.90991,   0.00049,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    195.74619,   0.00871,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    211.35320,   0.00005,     0.076 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    166.33199,   0.00014,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    192.87158,   0.00131,     2.670 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    166.61446,   0.00031,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -2 },    198.91975,   0.00002,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     16.08502,   0.00775,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },     16.86358,   0.00041,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   3,   0 },     16.79550,   0.00002,     0.025 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    346.67897,   0.00026,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    346.67048,   0.00050,     0.082 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.67240,   0.00027,     0.700 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    190.50914,   0.00030,     0.083 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    357.80813,   0.00008,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    166.66031,   0.00038,     0.085 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    331.73059,   0.00014,     0.465 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    139.04283,   0.00141,     0.090 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    149.71520,   0.00177,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },     87.45056,   0.00007,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   3,   0,  -1 },    166.71552,   0.00003,     0.043 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },      3.32658,   0.00002,     0.044 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    165.47003,   0.00010,     0.028 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },     16.36692,   0.00005,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   1,   0 },    357.84108,   0.00002,     0.016 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,  -2,   0,   0 },    265.17801,   0.00009,     0.040 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,  -2,   1,   0 },    268.13340,   0.00016,     0.085 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,  -1,   0 },     86.99000,   0.00004,     0.074 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0 },     88.75061,   0.00016,     3.446 },
 { {   0,   1,  -2,   0,   1,   0,   0,   0,   0,   1,   0 },    266.99000,   0.00004,     0.077 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   0,  -1,   0 },    150.84591,   0.00007,     0.074 },
 { {   0,   1,  -2,   0,   2,   0,   0,   0,   0,   1,   0 },    330.84591,   0.00007,     0.077 },
 { {   0,   1,  -2,   1,   0,   0,   0,   0,   0,  -1,   0 },    258.88743,   0.00002,     0.076 },
 { {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   1,   0 },     78.88743,   0.00002,     0.075 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   0,   0 },    310.94154,   0.00003,     0.042 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,  -2,   2,   0 },    130.94154,   0.00003,     0.406 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   0,  -1,   0 },    312.50139,   0.00002,     0.080 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   0,   1,   0 },    132.50139,   0.00002,     0.072 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -3,   0 },    134.75552,   0.00003,     0.070 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -2,   0 },    134.75553,   0.00003,     0.922 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,  -1,   0 },    314.75552,   0.00003,     0.082 },
 { {   0,   1,  -2,   2,   0,   0,   0,   0,   2,   0,   0 },    314.75553,   0.00003,     0.039 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,  -3,   0 },    326.45162,   0.00003,     0.072 },
 { {   0,   1,  -2,   3,   0,   0,   0,   0,   2,  -1,   0 },    146.45162,   0.00003,     0.079 },
 { {   0,   1,  -2,   4,   0,   0,   0,   0,   2,  -3,   0 },    333.63946,   0.00002,     0.075 },
 { {   0,   1,  -2,   4,   0,   0,   0,   0,   2,  -1,   0 },    153.63946,   0.00002,     0.076 },
 { {   0,   1,  -1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    114.95662,   0.00002,     0.041 },
 { {   0,   1,  -1,   0,  -2,   0,   0,   0,  -2,   2,   0 },    114.95662,   0.00002,     0.448 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    306.45813,   0.00007,     0.041 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,  -2,   2,   0 },      1.98328,   0.00002,     0.432 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    283.92591,   0.00002,     0.069 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    283.92591,   0.00003,     0.811 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,  -1,   0 },    103.92591,   0.00002,     0.083 },
 { {   0,   1,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },    103.92591,   0.00003,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   1,   0 },    269.98345,   0.00003,     0.017 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -6,   2,   0 },    269.97961,   0.00002,     0.021 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,  -1,   0 },    269.98823,   0.00017,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    269.99266,   0.00098,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   0,   2 },    269.97991,   0.00007,     0.045 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    269.99144,   0.00253,     0.028 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,   2 },    269.98089,   0.00003,     0.114 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },    269.96808,   0.00112,     0.045 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   3,   0 },    269.70812,   0.00006,     0.110 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   3,   2 },    269.98533,   0.00002,     0.056 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,  -1,   0 },    269.97043,   0.00002,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   0 },    269.95074,   0.00010,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   1,   0 },     89.97544,   0.00025,     0.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -3,   0 },    270.00857,   0.00003,     0.016 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    270.01878,   0.00061,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    269.98945,   0.00831,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.98523,   0.12362,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   2 },     89.98474,   0.00046,     0.366 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,  -2 },     89.98361,   0.00002,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    269.98452,   0.10060,     0.092 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   2 },     89.98480,   0.00031,     0.063 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,  -2 },     89.96821,   0.00002,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.99045,   0.00558,     0.417 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   2 },     89.99192,   0.00003,     0.034 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },     89.93761,   0.00016,     0.064 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   0 },     90.00215,   0.00038,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },     90.00945,   0.00511,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0 },    270.01103,   0.00046,     0.661 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0 },    269.59737,   0.00005,     0.068 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -3,   0 },    270.00822,   0.00049,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    270.00986,   0.00739,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   2 },    270.01334,   0.00004,     1.043 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.00993,   0.13811,     0.079 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   2 },     90.01180,   0.00048,     0.070 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -2 },    270.00132,   0.00027,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     90.04409,   0.03740,     1.599 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   2 },    270.02548,   0.00036,     0.036 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -2 },    270.08577,   0.00030,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     90.01606,   0.11998,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,  -2 },     90.00939,   0.00004,     3.425 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },     90.01663,   0.00652,     0.037 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   3,   0 },     90.01827,   0.00038,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0 },    270.10506,   0.00022,     0.074 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    126.18032,   0.00002,     0.073 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },    270.01733,   0.00070,     3.824 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    140.92028,   0.00012,     0.076 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -2 },    270.02151,   0.00004,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },     90.02094,   0.01093,     0.077 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },    306.18032,   0.00002,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,   0 },     90.02179,   0.00055,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   2,   0 },    270.18257,   0.00002,     0.025 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },     87.90163,   0.00002,     0.069 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },     89.93520,   0.00097,     0.871 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -2 },    269.81837,   0.00002,     0.068 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.88125,   0.01042,     0.083 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -2 },    269.99276,   0.00006,     0.675 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     90.00927,   0.02747,     0.039 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,  -2 },    270.03854,   0.00003,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   1,   0 },     90.00126,   0.00260,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   2,   0 },     90.07147,   0.00035,     0.019 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.01788,   0.00014,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   3,   0,   0 },     90.02064,   0.00015,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -3,   0 },    270.24942,   0.00003,     0.097 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.67085,   0.00009,     0.042 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,  -1,   0 },     89.97343,   0.00043,     0.027 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   0,   0 },     89.92755,   0.00009,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   4,   1,   0 },    270.03420,   0.00002,     0.016 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.30700,   0.00006,     0.042 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   0,  -1,   0 },     79.06361,   0.00007,     0.080 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   0,   1,   0 },    259.06361,   0.00007,     0.072 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   1,  -1,   1 },      1.85473,   0.00008,     0.075 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,  -2,   0 },     51.66988,   0.00003,     0.940 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,  -1,   0 },    234.42101,   0.00005,     0.082 },
 { {   0,   1,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },    231.66988,   0.00003,     0.039 },
 { {   0,   1,  -1,   2,   0,   0,   0,   0,   2,  -3,   0 },    301.63681,   0.00002,     0.075 },
 { {   0,   1,  -1,   2,   0,   0,   0,   0,   2,  -1,   0 },    141.86986,   0.00005,     0.076 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   0,  -1,   0 },     51.56224,   0.00002,     0.079 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0 },    231.56224,   0.00002,     0.072 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -3,   0 },    229.61750,   0.00003,     0.069 },
 { {   0,   1,   0,  -2,   0,   0,   0,   0,   2,  -1,   0 },     49.61750,   0.00003,     0.083 },
 { {   0,   1,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },     32.71877,   0.00007,     0.043 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -2,   0 },    270.99083,   0.00003,     0.040 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0 },     90.45420,   0.00004,     0.085 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },    270.99083,   0.00003,     0.649 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0 },    270.45420,   0.00004,     0.068 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -3,   0 },     93.60398,   0.00003,     0.074 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -2,   0 },     97.02583,   0.00002,     4.296 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,  -1,   0 },     93.60398,   0.00003,     0.077 },
 { {   0,   1,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },    277.02583,   0.00002,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -6,   2,   0 },    167.04533,   0.00002,     0.022 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,  -1,   0 },    346.15843,   0.00003,     0.016 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },    168.09008,   0.00028,     0.021 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },    166.91904,   0.00056,     0.029 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   2,   0 },    168.08896,   0.00010,     0.047 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0 },    166.36328,   0.00004,     0.028 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   1 },    193.33675,   0.00002,     0.045 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   1 },    193.33862,   0.00003,     0.114 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -2,   0 },    166.71247,   0.00004,     0.020 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },    167.39458,   0.00088,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    167.02994,   0.01143,     0.043 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   2 },    348.25678,   0.00003,     0.268 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    167.04650,   0.00614,     0.101 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    344.88604,   0.00037,     0.294 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   3,   0 },    359.14862,   0.00004,     0.060 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    193.33651,   0.00002,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1 },     13.34124,   0.00009,     0.095 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.35162,   0.00056,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    350.10403,   0.00015,     0.093 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.33574,   0.00043,     0.373 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,  -1 },    192.39529,   0.00030,     0.092 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   1 },    193.33870,   0.00024,     0.063 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,  -1 },    193.33552,   0.00002,     0.426 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0 },    170.77577,   0.00033,     0.040 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    171.66401,   0.00549,     0.086 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },     12.78261,   0.00009,     0.085 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    352.61779,   0.00196,     0.615 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 },     13.33745,   0.00004,     0.066 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    351.52136,   0.00196,     0.067 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    355.23806,   0.00006,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,  -1 },    193.33518,   0.00007,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   1 },     13.42428,   0.00005,     1.105 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.33502,   0.00124,     0.079 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.33765,   0.00103,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.33561,   0.00040,     1.749 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    180.55653,   0.00017,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   1 },     13.38631,   0.00019,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.37643,   0.00119,     0.072 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   2,  -1 },     13.33503,   0.00007,     0.037 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },      0.05436,   0.00004,     0.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    358.22740,   0.00053,     6.734 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.19267,   0.00744,     0.076 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },     13.33748,   0.00019,     0.077 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    184.73363,   0.00049,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    350.34221,   0.00004,     0.025 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -1,  -1 },    193.32661,   0.00016,     0.083 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,  -1 },     13.33431,   0.00015,     0.040 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.66201,   0.00008,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   4,  -1,   0 },    180.79425,   0.00006,     0.026 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,  -3,   0 },     15.84044,   0.00002,     0.075 },
 { {   0,   1,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    212.40392,   0.00006,     0.076 },
 { {   0,   1,   0,   0,   2,   0,   0,   0,   2,  -3,   0 },    224.44695,   0.00003,     0.076 },
 { {   0,   1,   0,   0,   2,   0,   0,   0,   2,  -1,   0 },     44.44695,   0.00003,     0.075 },
 { {   0,   1,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },    281.36261,   0.00005,     0.045 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },    270.05463,   0.00003,     0.111 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,  -2,   3,   0 },    270.05463,   0.00003,     0.057 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   0,  -2,   0 },    269.97189,   0.00002,     0.042 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },     89.97189,   0.00002,     0.407 },
 { {   0,   1,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },     90.14226,   0.00005,     0.072 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.65294,   0.00003,     0.045 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,  -2,   2,   0 },      0.65294,   0.00003,     0.232 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,   0,  -2,   0 },     75.84756,   0.00002,     0.042 },
 { {   0,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },     75.84756,   0.00002,     0.394 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -4,   0,   0 },    240.64683,   0.00004,     0.021 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,  -1,   0 },     84.82680,   0.00009,     0.028 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },     83.93996,   0.00108,     0.045 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },     90.02783,   0.00023,     0.113 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },    273.02751,   0.00004,     0.227 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,  -1 },     99.96781,   0.00006,     0.043 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    101.10373,   0.00026,     0.094 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    295.36563,   0.00049,     0.381 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    291.84656,   0.00030,     0.063 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    115.19085,   0.00005,     0.067 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    105.86186,   0.00003,     1.174 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    292.71669,   0.00067,     0.071 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    123.54008,   0.00007,     0.037 },
 { {   0,   1,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.30771,   0.00002,     0.047 },
 { {   0,   1,   2,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.30771,   0.00002,     0.191 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },      5.06853,   0.00015,     0.047 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    201.03378,   0.00004,     0.276 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    330.58304,   0.00004,     0.088 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    135.09643,   0.00004,     0.540 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    150.58304,   0.00004,     0.066 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,   0 },      2.77021,   0.00004,     0.035 },
 { {   0,   1,   2,   0,   1,   0,   0,   0,   2,   0,   0 },    218.85954,   0.00008,     0.035 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -3,   0 },     50.59981,   0.00003,     0.095 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -2,   0 },     31.02798,   0.00004,     0.361 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    230.59981,   0.00003,     0.062 },
 { {   0,   1,   3,   0,  -1,   0,   0,   0,   2,   0,   0 },    211.02798,   0.00004,     0.034 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    218.21312,   0.00002,     0.116 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   1,   0 },     38.21312,   0.00002,     0.056 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },    218.14134,   0.00003,     0.351 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   2,   0 },     38.21311,   0.00002,     0.018 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -3,   0 },     39.24732,   0.00002,     0.105 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -2,   0 },    218.78330,   0.00002,     0.265 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    219.24732,   0.00002,     0.059 },
 { {   0,   1,   4,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.78330,   0.00002,     0.033 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    297.61469,   0.00004,     0.260 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,  -2,   0 },    218.18457,   0.00002,     0.264 },
 { {   0,   1,   5,  -2,   0,   0,   0,   0,   2,   0,   0 },     38.18457,   0.00002,     0.033 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -3,   0 },     38.35760,   0.00002,     0.117 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -2,   0 },     38.31457,   0.00002,     0.214 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,  -1,   0 },     38.35760,   0.00002,     0.056 },
 { {   0,   1,   5,   0,  -2,   0,   0,   0,   2,   0,   0 },    218.31457,   0.00002,     0.032 },
 { {   0,   1,   6, -11,   0,   0,   0,   0,   2,  -3,   0 },    232.27977,   0.00002,     0.075 },
 { {   0,   1,   6, -11,   0,   0,   0,   0,   2,  -1,   0 },     52.27977,   0.00002,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   2,   0 },     25.93028,   0.00012,     0.040 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -4,   3,   0 },     24.21505,   0.00011,     0.087 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   0,   0 },     25.93028,   0.00002,     0.038 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   1,   0 },     20.93340,   0.00026,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   3,   0 },    231.39744,   0.00014,     0.075 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,  -2,   4,   0 },    205.93028,   0.00002,     0.038 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   1,   0 },    204.21505,   0.00011,     0.087 },
 { {   0,   2, -13,  15,   0,   0,   0,   0,   0,   2,   0 },    205.93028,   0.00012,     0.040 },
 { {   0,   2,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },    321.67691,   0.00003,     0.034 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   1,   0 },    141.77835,   0.00003,     0.066 },
 { {   0,   2,  -8,   2,   0,   0,   0,   0,  -2,   3,   0 },    321.77835,   0.00003,     0.088 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,  -2,   0,   0 },    353.80525,   0.00004,     0.040 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   0,  -1,   0 },    351.44265,   0.00003,     0.076 },
 { {   0,   2,  -8,   9,   0,   0,   0,   0,   0,   1,   0 },    171.44265,   0.00003,     0.075 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    232.94180,   0.00004,     0.066 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },     52.94180,   0.00004,     0.089 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    237.66985,   0.00002,     0.267 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },     57.66985,   0.00002,     0.044 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    239.12496,   0.00004,     0.129 },
 { {   0,   2,  -7,   0,   2,   0,   0,   0,  -2,   0,   0 },     21.77928,   0.00003,     0.035 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,  -2,   0,   0 },    324.68596,   0.00011,     0.040 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,  -2,   1,   0 },    324.68596,   0.00007,     0.087 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   0,  -1,   0 },    327.22665,   0.00009,     0.075 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   0,   1,   0 },    147.22665,   0.00009,     0.076 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   2,  -1,   0 },    144.59503,   0.00007,     0.087 },
 { {   0,   2,  -7,   7,   0,   0,   0,   0,   2,   0,   0 },    144.59503,   0.00012,     0.040 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.89069,   0.00007,     0.036 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    346.62574,   0.00005,     0.070 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.87141,   0.00009,     1.026 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    147.36296,   0.00008,     0.081 },
 { {   0,   2,  -5,   0,  -2,   0,   0,   0,  -2,   1,   0 },     71.49481,   0.00006,     0.075 },
 { {   0,   2,  -5,   0,  -1,   0,   0,   0,  -2,   1,   0 },    241.73792,   0.00004,     0.075 },
 { {   0,   2,  -5,   0,   0,  -1,   0,   0,  -2,   1,   0 },    323.89110,   0.00003,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,  -2,   0,  -2,   1,   0 },     68.08121,   0.00004,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    158.28312,   0.00004,     0.040 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    157.64937,   0.00013,     0.038 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    157.60981,   0.00184,     0.076 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    349.40745,   0.00010,    39.899 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    324.71178,   0.00011,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    307.67392,   0.00004,     0.067 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    338.36481,   0.00005,     0.572 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    154.64845,   0.00035,     0.087 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     34.10980,   0.00008,     0.103 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     75.36049,   0.00017,     0.044 },
 { {   0,   2,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },    150.01488,   0.00005,     0.076 },
 { {   0,   2,  -4,   0,  -2,   0,   0,   0,  -2,   1,   0 },    141.84137,   0.00003,     0.081 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    281.19445,   0.00006,     0.026 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    283.89549,   0.00035,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     57.55243,   0.00028,     0.082 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    105.29354,   0.00004,     0.976 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },     80.49890,   0.00004,     0.077 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    248.33973,   0.00049,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },     82.64339,   0.00053,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    260.49890,   0.00004,     0.074 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    262.10469,   0.00005,     0.037 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    246.67752,   0.00083,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    206.63482,   0.00023,     1.335 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     30.53932,   0.00056,     0.080 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },     40.41991,   0.00002,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    244.09084,   0.00013,     0.086 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     79.78882,   0.00005,     0.040 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     82.64339,   0.00009,     0.087 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    117.98762,   0.00010,     0.396 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    289.58610,   0.00190,     0.093 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    288.63536,   0.00288,     0.042 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    289.56106,   0.00024,     0.027 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },    288.23514,   0.00006,     0.028 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },     42.44711,   0.00002,     0.021 },
 { {   0,   2,  -4,   0,   1,   0,   0,   0,   2,   0,   0 },    359.40886,   0.00006,     0.042 },
 { {   0,   2,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },     89.84936,   0.00005,     0.041 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,  -2,   0,   0 },     88.13895,   0.00002,     0.041 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,  -2,   2,   0 },    268.13895,   0.00002,     0.479 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   0,  -1,   0 },     89.10399,   0.00004,     0.077 },
 { {   0,   2,  -4,   2,   0,   0,   0,   0,   0,   1,   0 },    269.10399,   0.00004,     0.074 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,   2,  -3,   0 },    151.47355,   0.00003,     0.073 },
 { {   0,   2,  -4,   4,   0,   0,   0,   0,   2,  -1,   0 },    331.47355,   0.00003,     0.078 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,  -2,   0,   0 },     30.94478,   0.00004,     0.040 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   0,  -1,   0 },    321.88655,   0.00004,     0.075 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   0,   1,   0 },    141.88655,   0.00004,     0.075 },
 { {   0,   2,  -3,   0,  -3,   0,   0,   0,   2,   0,   0 },    210.91986,   0.00004,     0.040 },
 { {   0,   2,  -3,   0,  -1,   0,   0,   0,   0,  -1,   0 },     31.02922,   0.00003,     0.076 },
 { {   0,   2,  -3,   0,  -1,   0,   0,   0,   0,   1,   0 },    211.02922,   0.00003,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,  -1,   0 },      0.48390,   0.00005,     0.016 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },      2.17464,   0.00051,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,   2 },      1.49095,   0.00004,     0.045 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },      1.97999,   0.00117,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   2,   0 },      1.25075,   0.00054,     0.044 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   0,   0 },      1.79899,   0.00017,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   1,   0 },    181.49759,   0.00008,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   1,   1 },    166.64586,   0.00002,     0.098 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -3,   2,   0 },      1.32632,   0.00002,     0.097 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.67624,   0.00040,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },      1.40842,   0.00359,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.43434,   0.05247,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   2 },    181.16484,   0.00022,     0.424 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,  -2 },    181.32093,   0.00002,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },      1.58991,   0.04883,     0.089 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   2 },    181.01900,   0.00015,     0.064 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    181.94640,   0.00270,     0.494 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    181.79112,   0.00017,     0.065 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    180.74586,   0.00006,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    346.62544,   0.00032,     0.084 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    346.41400,   0.00006,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    180.69774,   0.00117,     0.083 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    165.94964,   0.00010,     0.766 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.66681,   0.00023,     0.081 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.73707,   0.00006,     0.879 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.66462,   0.00015,     0.069 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },    358.48038,   0.00017,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    358.49358,   0.00283,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },    347.33928,   0.00004,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    358.71194,   0.05507,     0.077 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    178.41919,   0.00019,     0.072 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    357.82816,   0.00003,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    347.32345,   0.00082,     0.076 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    179.96485,   0.00638,     3.983 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    346.57767,   0.00015,     0.073 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   2 },      0.16244,   0.00008,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },    358.41919,   0.00019,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },    167.32478,   0.00005,    11.864 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    178.55815,   0.05164,     0.074 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    178.54368,   0.00270,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,   0 },    178.64442,   0.00016,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   1 },    174.32629,   0.00006,     2.131 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    346.57874,   0.00045,     0.071 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    346.58913,   0.00108,     0.078 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.57790,   0.00027,     1.246 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },     47.47691,   0.00005,     0.079 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },    348.87169,   0.00018,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    166.58588,   0.00049,     0.080 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },      1.96927,   0.00006,     0.068 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },      2.02074,   0.00129,     0.657 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,  -2 },      0.79708,   0.00007,     0.066 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    181.01795,   0.02834,     0.085 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },      0.30837,   0.00009,     0.539 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    347.47866,   0.00019,     0.086 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    179.99887,   0.02704,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },      0.71959,   0.00004,     0.088 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    179.13219,   0.00190,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    179.99514,   0.00022,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },      0.14658,   0.00004,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0,  -1 },    166.56436,   0.00006,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },    179.86571,   0.00010,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },    179.95743,   0.00022,     0.043 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    176.86416,   0.00045,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    178.13301,   0.00027,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   1,   0 },    181.25840,   0.00004,     0.016 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,  -2,   0,   0 },    259.87033,   0.00004,     0.041 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,  -2,   1,   0 },     93.02982,   0.00005,     0.090 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   0,   0 },     88.75931,   0.00013,     2.982 },
 { {   0,   2,  -3,   0,   1,   0,   0,   0,   0,   2,   0 },     88.71744,   0.00002,     0.037 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,  -2,   0 },    150.19491,   0.00002,     0.738 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,  -1,   0 },    330.30983,   0.00006,     0.084 },
 { {   0,   2,  -3,   0,   2,   0,   0,   0,   2,   0,   0 },    330.19491,   0.00002,     0.040 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   0,   0 },    130.23663,   0.00003,     0.043 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,  -2,   2,   0 },    310.23663,   0.00003,     0.324 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,   2,  -2,   0 },    134.71905,   0.00003,     2.176 },
 { {   0,   2,  -3,   2,   0,   0,   0,   0,   2,   0,   0 },    314.71905,   0.00003,     0.038 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,  -3,   0 },    323.51469,   0.00004,     0.076 },
 { {   0,   2,  -3,   3,   0,   0,   0,   0,   2,  -1,   0 },    143.51469,   0.00004,     0.075 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    342.11227,   0.00006,     0.042 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },    279.73027,   0.00004,     0.340 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,  -2,   0 },    283.99618,   0.00003,     1.647 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    102.20545,   0.00005,     0.079 },
 { {   0,   2,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },    103.99618,   0.00003,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   1,   0 },     90.38097,   0.00004,     0.017 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -6,   2,   0 },     90.31470,   0.00013,     0.022 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,  -1,   0 },     89.65838,   0.00024,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },     90.79310,   0.00124,     0.021 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   2 },     90.30640,   0.00011,     0.047 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },     90.73558,   0.00304,     0.029 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   1,   2 },     90.32368,   0.00006,     0.123 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   2,   0 },     90.54637,   0.00131,     0.046 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   3,   0 },     89.97710,   0.00003,     0.118 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,  -1,   0 },     90.41551,   0.00003,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   0,   0 },     90.91553,   0.00005,     0.028 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -3,   1,   0 },    270.39757,   0.00031,     0.044 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -3,   0 },     90.31084,   0.00007,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },     90.26537,   0.00054,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },     90.48451,   0.00978,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -1,   2 },    270.31046,   0.00007,     0.101 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     90.36386,   0.16644,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   2 },    270.33183,   0.00059,     0.298 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,  -2 },    270.30534,   0.00007,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     90.31081,   0.11617,     0.098 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   2 },    270.30119,   0.00027,     0.060 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,  -2 },    270.34204,   0.00002,     0.042 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    270.21793,   0.00819,     0.331 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   2 },    270.29822,   0.00002,     0.033 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },    270.24144,   0.00055,     0.061 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,  -1,   0 },    270.14422,   0.00014,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    278.28529,   0.00011,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    270.05775,   0.00183,     0.090 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    101.97736,   0.00002,     0.089 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },     90.10439,   0.00020,     0.468 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   2,   0 },     89.93683,   0.00003,     0.065 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.37888,   0.00016,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.50636,   0.00262,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   2 },     90.25733,   0.00002,     0.631 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.39176,   0.04919,     0.083 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    270.29250,   0.00005,     0.067 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,  -2 },    270.14757,   0.00004,     0.039 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    270.23382,   0.02611,     0.799 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   2 },    269.80576,   0.00003,     0.036 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,  -2 },     89.83579,   0.00012,     0.081 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    270.14292,   0.04924,     0.069 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -2 },    270.26038,   0.00002,     1.090 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },    270.14037,   0.00265,     0.036 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   3,  -2 },     89.74927,   0.00002,     0.071 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   3,   0 },    270.22144,   0.00014,     0.024 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    277.48286,   0.00048,     0.077 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },     89.98018,   0.00007,     2.747 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    275.61116,   0.00022,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },     89.98947,   0.00031,     0.076 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    269.96744,   0.00200,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },     96.13424,   0.00007,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },     97.48286,   0.00048,     0.074 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,   0 },    269.95178,   0.00010,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   2,  -2 },    270.03903,   0.00002,     0.075 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },     90.70477,   0.00043,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,  -2 },     90.01900,   0.00003,     0.037 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },     89.90470,   0.00636,     1.912 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   2 },     90.08303,   0.00003,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,  -2 },     90.02556,   0.00033,     0.071 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    269.99692,   0.12047,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.11734,   0.00005,     1.168 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    269.97026,   0.04270,     0.038 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,  -2 },     89.96682,   0.00011,     0.081 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    270.14562,   0.00281,     0.025 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },    270.13057,   0.00030,     0.019 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },    281.97736,   0.00002,     0.086 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,   0 },     89.99993,   0.00009,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,  -2 },     89.98947,   0.00006,     0.087 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,  -1 },     95.61233,   0.00005,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   3,   0,   0 },    269.99421,   0.00023,     0.026 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    270.07161,   0.00127,     0.041 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,  -2 },    269.98140,   0.00006,     0.094 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.01542,   0.00163,     0.027 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   0,  -2 },    269.98859,   0.00003,     0.042 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    270.02050,   0.00051,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   4,   1,   0 },    270.07634,   0.00004,     0.016 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -2,   0 },    270.04594,   0.00003,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   6,  -1,   0 },    270.00792,   0.00002,     0.016 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.48011,   0.00007,     0.043 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   1,   0 },    262.22066,   0.00003,     0.099 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,  -2,   3,   0 },    262.22066,   0.00003,     0.061 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,  -1,   0 },    246.37806,   0.00007,     0.078 },
 { {   0,   2,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    226.89368,   0.00004,     0.038 },
 { {   0,   2,  -2,   1,   0,   0,   0,   0,   2,  -3,   0 },    316.01931,   0.00002,     0.075 },
 { {   0,   2,  -2,   1,   0,   0,   0,   0,   2,  -1,   0 },    136.01931,   0.00002,     0.075 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,  -2,   0,   0 },    225.09889,   0.00004,     0.042 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,  -2,   2,   0 },     45.09889,   0.00004,     0.338 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,   2,  -2,   0 },    229.61280,   0.00003,     1.706 },
 { {   0,   2,  -1,  -2,   0,   0,   0,   0,   2,   0,   0 },     49.61280,   0.00003,     0.039 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    207.13760,   0.00007,     0.044 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,  -3,   0 },     51.18568,   0.00002,     0.077 },
 { {   0,   2,  -1,   0,  -2,   0,   0,   0,   2,  -1,   0 },     51.43982,   0.00005,     0.074 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },     91.27905,   0.00018,     0.044 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,  -2,   1,   0 },    271.35098,   0.00006,     0.107 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,  -1,   0 },    271.46864,   0.00003,     0.090 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 },     90.95380,   0.00015,     0.462 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   1,   0 },     91.46864,   0.00003,     0.065 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,  -1,   0 },     87.54832,   0.00015,     0.073 },
 { {   0,   2,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },    274.04211,   0.00010,     0.037 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,  -1,   0 },    167.42538,   0.00009,     0.017 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    346.61338,   0.00028,     0.021 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    347.17767,   0.00056,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },    346.36982,   0.00013,     0.048 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   0,   0 },    347.41390,   0.00007,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   0,   1 },    193.23116,   0.00002,     0.047 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,   0 },    166.95772,   0.00003,     0.046 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -3,   1,   1 },    193.25308,   0.00002,     0.122 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    347.19642,   0.00008,     0.020 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    347.02947,   0.00113,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    347.11886,   0.01639,     0.044 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   2 },    166.47500,   0.00003,     0.229 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.24267,   0.00794,     0.108 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    168.24772,   0.00063,     0.248 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },    161.80849,   0.00002,     0.058 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    193.23274,   0.00002,     0.027 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1 },     13.17842,   0.00008,     0.100 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.17875,   0.00044,     0.042 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.22202,   0.00037,     0.303 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    190.93840,   0.00019,     0.097 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   1 },    193.23906,   0.00017,     0.060 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    345.74194,   0.00007,     0.041 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    346.90076,   0.00211,     0.091 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    162.32916,   0.00139,     0.444 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    163.76150,   0.00159,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },    161.50598,   0.00005,     0.035 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    193.27937,   0.00002,     0.040 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.28763,   0.00057,     0.083 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.30035,   0.00059,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.30460,   0.00041,     0.835 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    193.04972,   0.00004,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },     13.27734,   0.00023,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.43234,   0.00054,     0.069 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    347.73509,   0.00007,     0.078 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    339.94274,   0.00015,     2.096 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    153.68564,   0.00199,     0.073 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    193.33796,   0.00022,     0.074 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    337.46444,   0.00025,     0.037 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   1,   0 },    348.48623,   0.00003,     0.025 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1 },    193.33015,   0.00019,     0.079 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   3,   0,  -1 },     13.30683,   0.00010,     0.039 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   4,  -2,   0 },    148.99258,   0.00002,     0.040 },
 { {   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -3,   0 },    337.54092,   0.00003,     0.075 },
 { {   0,   2,   0,  -3,   0,   0,   0,   0,   2,  -1,   0 },    157.54092,   0.00003,     0.076 },
 { {   0,   2,   0,   0,  -3,   0,   0,   0,  -2,   0,   0 },     99.16578,   0.00005,     0.046 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   1,   0 },     90.11781,   0.00007,     0.119 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,  -2,   3,   0 },    270.06360,   0.00002,     0.055 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,  -2,   0 },     90.06162,   0.00002,     0.043 },
 { {   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0 },    270.00907,   0.00008,     0.324 },
 { {   0,   2,   0,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.56956,   0.00003,     0.046 },
 { {   0,   2,   0,   0,  -1,   0,   0,   0,  -2,   2,   0 },      0.56956,   0.00003,     0.202 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },     65.49171,   0.00007,     0.022 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },    225.49238,   0.00008,     0.029 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    220.72242,   0.00087,     0.047 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    231.00581,   0.00031,     0.121 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    344.33387,   0.00002,     0.199 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },     99.19715,   0.00005,     0.044 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    129.65186,   0.00014,     0.100 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    301.38492,   0.00047,     0.308 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    303.34850,   0.00022,     0.061 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    115.57509,   0.00004,     0.064 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    300.39166,   0.00043,     0.068 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    116.60908,   0.00014,     0.036 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,  -2,   0,   0 },    109.93267,   0.00011,     0.040 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,  -2,   1,   0 },    109.93267,   0.00007,     0.087 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   2,  -1,   0 },    290.00215,   0.00006,     0.087 },
 { {   0,   2,   1,  -8,   0,   0,   0,   0,   2,   0,   0 },    290.00215,   0.00011,     0.040 },
 { {   0,   2,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.20818,   0.00002,     0.048 },
 { {   0,   2,   1,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.20818,   0.00002,     0.171 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.14120,   0.00009,     0.049 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    212.76217,   0.00005,     0.235 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   2,   0 },     15.32187,   0.00002,     0.033 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,  -3,   0 },    217.25112,   0.00003,     0.094 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },    217.25112,   0.00003,     0.063 },
 { {   0,   2,   1,   0,   1,   0,   0,   0,   2,   0,   0 },     37.73304,   0.00006,     0.034 },
 { {   0,   2,   2,   0,  -1,   0,   0,   0,   2,   0,   0 },     67.51252,   0.00005,     0.033 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    218.24391,   0.00005,     0.125 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   1,   0 },     38.24391,   0.00005,     0.054 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    216.07098,   0.00005,     0.288 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    218.27481,   0.00002,     0.060 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   1,   0 },     38.27481,   0.00002,     0.023 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   2,   0 },     38.24391,   0.00006,     0.018 },
 { {   0,   2,   2,   0,   1,   0,   0,   0,   2,   0,   0 },    209.69084,   0.00005,     0.033 },
 { {   0,   2,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },    174.58108,   0.00002,     0.293 },
 { {   0,   2,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },    354.58108,   0.00002,     0.033 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   0,   0,   0 },    219.53134,   0.00003,     0.162 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    218.78311,   0.00002,     0.228 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,  -1,   0 },     39.48082,   0.00008,     0.057 },
 { {   0,   2,   3,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.78311,   0.00002,     0.032 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.09665,   0.00004,     0.054 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.09665,   0.00004,     0.125 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    299.69225,   0.00002,     0.143 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    119.69225,   0.00002,     0.051 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,  -2,   0 },     38.22966,   0.00004,     0.227 },
 { {   0,   2,   4,  -2,   0,   0,   0,   0,   2,   0,   0 },    218.22966,   0.00004,     0.032 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -3,   0 },    218.35803,   0.00002,     0.126 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -2,   0 },     38.31438,   0.00002,     0.188 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,  -1,   0 },     38.35803,   0.00002,     0.054 },
 { {   0,   2,   4,   0,  -2,   0,   0,   0,   2,   0,   0 },    218.31438,   0.00002,     0.031 },
 { {   0,   3,  -9,   0,   2,   0,   0,   0,  -2,   0,   0 },    321.67057,   0.00003,     0.035 },
 { {   0,   3,  -9,   4,   0,   0,   0,   0,   2,   0,   0 },    262.38977,   0.00004,     0.044 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     25.57879,   0.00002,     0.036 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    205.57879,   0.00002,     0.068 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    205.57879,   0.00002,     0.741 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },     25.57879,   0.00002,     0.084 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },     60.82972,   0.00002,     0.055 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    240.82972,   0.00002,     0.120 },
 { {   0,   3,  -8,   0,   2,   0,   0,   0,  -2,   0,   0 },     21.77482,   0.00003,     0.036 },
 { {   0,   3,  -8,   0,   2,   0,   0,   0,  -2,   1,   0 },    201.89203,   0.00004,     0.069 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,  -2,   0,   0 },    160.57779,   0.00004,     0.039 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,  -1,   0 },    334.10689,   0.00003,     0.070 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,   0,   0 },    336.61507,   0.00003,     1.003 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   0,   1,   0 },    158.85593,   0.00009,     0.082 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,  -1,   0 },    159.51861,   0.00021,     0.095 },
 { {   0,   3,  -8,   4,   0,   0,   0,   0,   2,   0,   0 },    159.97967,   0.00044,     0.042 },
 { {   0,   3,  -7,   0,  -1,   0,   0,   0,  -2,   1,   0 },    357.24677,   0.00007,     0.073 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,  -1,   2 },    140.72777,   0.00003,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.85791,   0.00006,     0.037 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },      4.11689,   0.00003,     0.074 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   2 },    320.72777,   0.00003,     0.076 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.85791,   0.00006,     2.862 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    184.11689,   0.00003,     0.077 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    321.76503,   0.00002,     0.035 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    298.10099,   0.00002,     0.065 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    247.64353,   0.00002,     0.471 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    118.10099,   0.00002,     0.090 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    100.65158,   0.00004,     0.107 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    105.11537,   0.00017,     0.044 },
 { {   0,   3,  -7,   0,   0,   1,   0,   0,  -2,   1,   0 },    229.42666,   0.00002,     0.074 },
 { {   0,   3,  -7,   0,   0,   1,   0,   0,  -2,   3,   0 },    229.42666,   0.00002,     0.077 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   1,   0 },     92.61949,   0.00003,     0.074 },
 { {   0,   3,  -7,   0,   1,   0,   0,   0,  -2,   3,   0 },    272.61949,   0.00003,     0.077 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   0,   0 },    237.64087,   0.00004,     0.020 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   1,   0 },    237.64076,   0.00014,     0.028 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -4,   2,   0 },    237.64052,   0.00004,     0.044 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,  -1,   0 },    237.95767,   0.00030,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,   0 },    237.70631,   0.00419,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   0,   2 },     57.64086,   0.00003,     0.474 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   1,   0 },    237.31955,   0.00429,     0.087 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -2,   2,   0 },     56.71528,   0.00022,     0.563 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,  -1,   0,   0 },     57.64797,   0.00009,     0.081 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -2,   0 },    239.42267,   0.00030,     0.038 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,  -1,   0 },    239.43600,   0.00541,     0.075 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   1,   0 },     59.43866,   0.00540,     0.075 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   2,   0 },     59.42667,   0.00030,     0.038 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   1,   0,   0 },    237.64318,   0.00009,     0.081 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -2,   0 },    236.72940,   0.00022,     0.565 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,  -1,   0 },     57.32929,   0.00426,     0.087 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   0,   0 },     57.71522,   0.00416,     0.040 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   2,   1,   0 },     57.96650,   0.00030,     0.026 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,  -2,   0 },     57.65077,   0.00004,     0.044 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,  -1,   0 },     57.65047,   0.00013,     0.028 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   4,   0,   0 },     57.65034,   0.00004,     0.020 },
 { {   0,   3,  -6,   0,  -2,   0,   0,   0,  -2,   1,   0 },     94.67010,   0.00003,     0.078 },
 { {   0,   3,  -6,   0,  -2,   0,   0,   0,  -2,   3,   0 },     94.67010,   0.00003,     0.073 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -2,   1,   0 },    326.76955,   0.00008,     0.079 },
 { {   0,   3,  -6,   0,  -1,   0,   0,   0,  -2,   3,   0 },    142.46495,   0.00002,     0.072 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   0,   1 },    328.81434,   0.00003,     0.040 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -3,   1,   1 },    149.66610,   0.00002,     0.087 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    355.74694,   0.00010,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    336.40820,   0.00002,     0.079 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },     73.60428,   0.00004,     0.072 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },    329.13492,   0.00032,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },    148.99031,   0.00005,     0.073 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    149.13492,   0.00032,     0.076 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    195.00494,   0.00010,     0.070 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    181.26780,   0.00008,     0.890 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },      1.73846,   0.00027,     0.082 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    329.66610,   0.00002,     0.088 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    148.81434,   0.00003,     0.041 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.32805,   0.00008,     0.345 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },      0.31777,   0.00112,     0.097 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },      1.11838,   0.00166,     0.042 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    359.30476,   0.00018,     0.027 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    359.42521,   0.00007,     0.028 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },    179.90922,   0.00004,     0.021 },
 { {   0,   3,  -6,   0,   2,   0,   0,   0,   2,   0,   0 },    198.72423,   0.00007,     0.042 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,  -2,   0,   0 },    184.49196,   0.00004,     0.040 },
 { {   0,   3,  -6,   2,   0,   0,   0,   0,   2,   0,   0 },      4.68674,   0.00005,     0.041 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,   0,   0 },    138.99979,   0.00033,     0.042 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,  -2,   1,   0 },    140.29115,   0.00017,     0.095 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   0,  -1,   0 },    171.28692,   0.00002,     0.082 },
 { {   0,   3,  -6,   4,   0,   0,   0,   0,   2,   0,   0 },    130.74568,   0.00007,     0.039 },
 { {   0,   3,  -5,  -3,   0,   0,   0,   0,  -2,   1,   0 },    188.35678,   0.00003,     0.076 },
 { {   0,   3,  -5,  -3,   0,   0,   0,   0,  -2,   3,   0 },      8.35678,   0.00003,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },     75.66706,   0.00014,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },     75.49011,   0.00029,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },     77.92021,   0.00013,     0.043 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -3,   0,   0 },     73.87917,   0.00005,     0.027 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -2,   0 },     75.44986,   0.00008,     0.019 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },     72.26375,   0.00085,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,  -1,   2 },    262.99771,   0.00004,     0.089 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.03146,   0.01339,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   2 },    255.01941,   0.00003,     0.504 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     75.87198,   0.01558,     0.086 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   2 },    242.71952,   0.00002,     0.066 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    256.28581,   0.00087,     0.606 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    256.82644,   0.00005,     0.067 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    243.93529,   0.00023,     0.080 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },     63.90660,   0.00003,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },     68.76802,   0.00069,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     67.83981,   0.01314,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   2 },    247.85171,   0.00005,     0.074 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     74.14930,   0.00076,     8.102 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },     78.40922,   0.00017,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,  -2 },     67.85171,   0.00005,     0.073 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    248.71735,   0.01375,     0.076 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    247.26710,   0.00065,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   3,   0 },    242.70375,   0.00003,     0.025 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },     76.96409,   0.00006,     0.081 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    310.49272,   0.00020,     0.082 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },     76.96407,   0.00005,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },    283.21881,   0.00004,     0.039 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },     78.91672,   0.00006,     0.066 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     77.65758,   0.00107,     0.527 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,  -2 },     75.75537,   0.00003,     0.065 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    256.97761,   0.01830,     0.088 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },     76.41168,   0.00007,     0.448 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    256.54159,   0.01777,     0.041 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,  -2 },     76.31553,   0.00004,     0.091 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    255.82186,   0.00114,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    256.72826,   0.00011,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -2,   0 },    282.97958,   0.00005,     0.096 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -1,   0 },    283.27586,   0.00013,     0.042 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },    265.23111,   0.00011,     0.027 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },    257.62900,   0.00015,     0.044 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },    256.59816,   0.00040,     0.028 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    254.19051,   0.00013,     0.020 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   4,   1,   0 },    255.76536,   0.00003,     0.016 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   0,  -1,   0 },    269.67378,   0.00003,     0.075 },
 { {   0,   3,  -5,   0,   0,   2,   0,   0,   0,   1,   0 },     89.67378,   0.00003,     0.076 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,  -2,   0,   0 },    177.11605,   0.00004,     0.040 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   0,  -1,   0 },    177.32583,   0.00004,     0.075 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   0,   1,   0 },    357.32583,   0.00004,     0.076 },
 { {   0,   3,  -5,   0,   0,   3,   0,   0,   2,   0,   0 },    357.12420,   0.00004,     0.040 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },    172.52717,   0.00023,     0.040 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,  -2,   1,   0 },    172.38205,   0.00011,     0.087 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,  -1,   0 },    172.45129,   0.00023,     0.075 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   0,   1,   0 },    352.45129,   0.00023,     0.076 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },    353.46404,   0.00013,     0.087 },
 { {   0,   3,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },    355.11470,   0.00016,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,  -1,   0 },    268.98598,   0.00008,     0.026 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   0,   0 },    268.99423,   0.00070,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   1,   0 },    268.55068,   0.00057,     0.087 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,  -2,   2,   0 },     88.97077,   0.00004,     0.550 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -2,   0 },    270.82104,   0.00004,     0.038 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,  -1,   0 },    270.31040,   0.00097,     0.076 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   1,   0 },     90.31883,   0.00097,     0.075 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   0,   2,   0 },     90.73039,   0.00004,     0.038 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,  -1,   0 },     88.50760,   0.00052,     0.087 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   0,   0 },     88.93283,   0.00063,     0.040 },
 { {   0,   3,  -5,   0,   2,   0,   0,   0,   2,   1,   0 },     88.94936,   0.00005,     0.026 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,  -2,   0,   0 },    260.28352,   0.00003,     0.041 },
 { {   0,   3,  -5,   0,   3,   0,   0,   0,   2,  -1,   0 },     80.19941,   0.00004,     0.086 },
 { {   0,   3,  -5,   2,   0,   0,   0,   0,  -2,   0,   0 },     89.16063,   0.00005,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -6,   1,   0 },      0.60846,   0.00002,     0.017 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,  -1,   0 },      1.37821,   0.00005,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },      2.49393,   0.00034,     0.021 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },      0.58508,   0.00087,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   2,   0 },      0.45971,   0.00038,     0.045 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   0,   0 },      0.44435,   0.00007,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -3,   1,   0 },    180.32414,   0.00018,     0.044 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -2,   0 },      0.60134,   0.00036,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },      0.62531,   0.00288,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   2 },    180.58532,   0.00003,     0.097 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },      0.54004,   0.04556,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   2 },    180.60435,   0.00019,     0.335 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.61150,   0.03489,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   2 },    180.65036,   0.00009,     0.062 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    180.66879,   0.00223,     0.377 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    180.68848,   0.00012,     0.063 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -2,   0 },      0.21385,   0.00008,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },      0.33403,   0.00089,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    166.62822,   0.00009,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    346.58129,   0.00016,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    180.34595,   0.00957,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    346.63162,   0.00013,     0.518 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    346.63601,   0.00019,     0.086 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   0 },      0.30879,   0.00133,     0.567 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    346.62822,   0.00009,     0.066 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },      0.28719,   0.00012,     0.067 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -3,   0 },      1.63869,   0.00011,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },      1.40683,   0.00180,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },      1.19151,   0.03488,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   2 },    180.93437,   0.00011,     0.069 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -2 },      0.84529,   0.00012,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    346.48346,   0.00007,     0.080 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    181.01987,   0.01145,     1.141 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    346.31181,   0.00004,     0.070 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   2 },      0.91617,   0.00008,     0.036 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,  -2 },      1.99373,   0.00006,     0.079 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    181.14487,   0.02778,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    180.93407,   0.00150,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   3,   0 },    181.04300,   0.00009,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -3,   0 },    359.95433,   0.00021,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    359.95433,   0.00380,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    167.62943,   0.00005,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.31013,   0.00319,    95.796 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    165.48379,   0.00008,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },      0.28240,   0.00021,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.62339,   0.00005,     5.644 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    180.28888,   0.06007,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    354.57477,   0.00005,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    347.62943,   0.00005,     0.076 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    180.28757,   0.00327,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   2,   0 },    180.28188,   0.00018,     0.025 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    245.60229,   0.00004,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.11626,   0.00037,     1.114 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,  -2 },    181.61142,   0.00003,     0.069 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    192.87947,   0.00362,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    346.52829,   0.00006,     0.082 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    182.04149,   0.00566,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,  -2 },      1.61142,   0.00003,     0.083 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    180.15883,   0.00057,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    181.32049,   0.00005,     0.019 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -3,   0 },      1.35506,   0.00003,     0.560 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -2,   0 },    180.56971,   0.00119,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },    166.64546,   0.00006,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,  -1,   0 },    180.33672,   0.00205,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,  -1 },    166.47713,   0.00004,     0.041 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   0,   0 },    180.29719,   0.00061,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   3,   1,   0 },    180.28300,   0.00006,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -3,   0 },      4.17906,   0.00005,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },    193.48229,   0.00024,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },    189.81414,   0.00048,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    185.49621,   0.00007,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   1,   0 },      1.69537,   0.00002,     0.016 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   5,  -2,   0 },    180.40960,   0.00004,     0.028 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   0,   0 },     80.00000,   0.00011,     0.042 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,  -2,   1,   0 },     85.61226,   0.00007,     0.095 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,  -1,   0 },     86.76213,   0.00002,     0.081 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   0,   0 },     88.68872,   0.00003,     1.041 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   1,   0 },    266.76213,   0.00002,     0.070 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   0,   2,   0 },     88.68872,   0.00003,     0.036 },
 { {   0,   3,  -4,   0,   1,   0,   0,   0,   2,  -1,   0 },    237.78864,   0.00004,     0.080 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,  -2,   0,   0 },    173.49065,   0.00008,     0.042 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,  -1,  -1 },    320.96169,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   1,   1,  -1 },    140.96169,   0.00002,     0.075 },
 { {   0,   3,  -4,   0,   2,   0,   0,   0,   2,   0,   0 },    330.25736,   0.00005,     0.039 },
 { {   0,   3,  -4,   2,   0,   0,   0,   0,   2,  -3,   0 },    201.43075,   0.00002,     0.076 },
 { {   0,   3,  -4,   2,   0,   0,   0,   0,   2,  -1,   0 },     21.43075,   0.00002,     0.075 },
 { {   0,   3,  -3,   0,  -2,   0,   0,   0,   2,  -1,   0 },    300.30906,   0.00003,     0.076 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    305.94531,   0.00007,     0.044 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    182.19971,   0.00003,     0.280 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -3,   0 },     13.16266,   0.00002,     0.076 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -2,   0 },    284.04550,   0.00002,    54.802 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,  -1,   0 },    172.56988,   0.00008,     0.075 },
 { {   0,   3,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },    104.04550,   0.00002,     0.038 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -3,   0 },    212.06366,   0.00004,     0.075 },
 { {   0,   3,  -3,   0,   0,  -3,   0,   0,   2,  -1,   0 },     32.06366,   0.00004,     0.075 },
 { {   0,   3,  -3,   0,   0,  -1,   0,   0,   2,  -1,   0 },      2.79090,   0.00006,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },     97.66787,   0.00014,     0.021 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },     99.29786,   0.00019,     0.029 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -4,   2,   0 },     93.32847,   0.00007,     0.047 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -3,   0,   0 },     93.93856,   0.00005,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },     91.14277,   0.00022,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },     92.09005,   0.00186,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,  -1,   2 },    272.08724,   0.00002,     0.108 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.98757,   0.01729,     0.044 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     92.71917,   0.00987,     0.104 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   1,   2 },     92.08724,   0.00002,     0.058 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    271.78685,   0.00096,     0.274 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    271.02632,   0.00010,     0.059 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    265.10711,   0.00019,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    268.57890,   0.00006,     0.042 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },     65.18792,   0.00019,     0.095 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -4,   0 },     90.14305,   0.00002,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.04474,   0.00043,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.14029,   0.00658,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   2 },     89.98628,   0.00005,     0.452 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.10540,   0.09932,     0.088 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    270.12239,   0.00023,     0.065 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    269.46556,   0.00004,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    270.57562,   0.00737,     0.533 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   2 },    269.90191,   0.00005,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },     90.20675,   0.00019,     0.085 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     89.17732,   0.00707,     0.066 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },     89.62631,   0.00047,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },    264.89913,   0.00006,     0.082 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    272.09030,   0.00009,     0.080 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },     89.99747,   0.00037,     1.011 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    276.58995,   0.00005,     0.069 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    268.34447,   0.00225,     0.070 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     96.59464,   0.00006,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },     92.09030,   0.00009,     0.071 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,   0 },    270.02366,   0.00012,     0.036 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -4,   0 },     95.82791,   0.00003,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },     95.35975,   0.00066,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,  -2 },    269.99509,   0.00017,     0.037 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    269.99573,   0.03452,     9.751 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   2 },    269.98152,   0.00013,     0.037 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,  -2 },    269.98820,   0.00296,     0.074 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.98187,   0.59827,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },     89.99076,   0.00032,     4.330 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    101.31859,   0.00014,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     89.85303,   0.01139,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },     90.01467,   0.00065,     0.077 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },     89.75324,   0.00095,     0.025 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,  -2 },     90.02057,   0.00004,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    270.21384,   0.00006,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,  -1,   0 },    269.94140,   0.00013,     0.039 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   3,   0,   0 },    270.06278,   0.00019,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -4,   0 },     88.80426,   0.00007,     0.598 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -3,   0 },    269.46716,   0.00123,     0.086 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,  -2 },     89.98628,   0.00005,     0.499 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.96794,   0.00497,     0.040 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,  -2 },    269.99877,   0.00025,     0.089 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },     89.95865,   0.00490,     0.026 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },     89.96719,   0.00035,     0.019 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -3,   0 },     90.06342,   0.00007,     0.028 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -2,   0 },     89.98236,   0.00012,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   6,  -1,   0 },     89.98197,   0.00005,     0.016 },
 { {   0,   3,  -3,   0,   0,   1,   0,   0,   2,  -3,   0 },    177.68777,   0.00002,     0.076 },
 { {   0,   3,  -3,   0,   0,   1,   0,   0,   2,  -1,   0 },    357.68777,   0.00002,     0.075 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,  -2,   0,   0 },     60.11801,   0.00007,     0.044 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,  -3,   0 },    300.29140,   0.00002,     0.077 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,  -1,   0 },    120.29140,   0.00002,     0.074 },
 { {   0,   3,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    242.07525,   0.00003,     0.037 },
 { {   0,   3,  -3,   0,   2,   0,   0,   0,   2,  -3,   0 },     85.17090,   0.00002,     0.077 },
 { {   0,   3,  -3,   0,   2,   0,   0,   0,   2,  -1,   0 },     85.17090,   0.00002,     0.074 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,  -3,   0 },     33.51344,   0.00003,     0.073 },
 { {   0,   3,  -2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    213.51344,   0.00003,     0.078 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   2,  -2,   0 },    229.60740,   0.00003,    25.521 },
 { {   0,   3,  -2,  -2,   0,   0,   0,   0,   2,   0,   0 },     49.60740,   0.00003,     0.038 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    179.59678,   0.00002,     0.045 },
 { {   0,   3,  -2,   0,  -2,   0,   0,   0,  -2,   2,   0 },    179.59678,   0.00002,     0.223 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },     91.30546,   0.00020,     0.046 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },    271.48892,   0.00003,     0.219 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0 },     90.93955,   0.00010,     0.358 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   2,  -1,   0 },     94.78524,   0.00009,     0.070 },
 { {   0,   3,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },    274.00432,   0.00009,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -4,   0,   0 },    170.05798,   0.00005,     0.021 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -4,   1,   0 },    346.71326,   0.00004,     0.030 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    347.04411,   0.00032,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    347.46870,   0.00203,     0.046 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.44519,   0.00086,     0.116 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    165.34323,   0.00004,     0.215 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    193.18326,   0.00003,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },     13.21433,   0.00006,     0.107 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.09292,   0.00036,     0.044 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.26532,   0.00027,     0.254 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.37838,   0.00006,     0.104 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },    193.25977,   0.00011,     0.058 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },    347.15235,   0.00002,     0.027 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    346.76480,   0.00034,     0.042 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    346.67057,   0.00421,     0.096 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    158.26956,   0.00051,     0.348 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     25.50823,   0.00006,     0.062 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },     13.16782,   0.00006,     0.087 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.30998,   0.00034,     0.065 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.36425,   0.00034,     0.549 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },     13.26624,   0.00021,     0.035 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.30739,   0.00034,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -4,   0 },    162.63289,   0.00002,     0.039 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    176.07523,   0.00004,     0.082 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    168.20342,   0.00005,     0.907 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    166.12093,   0.00364,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    193.34109,   0.00008,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    341.91350,   0.00010,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    344.12119,   0.00004,     0.024 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,  -1 },    193.28511,   0.00013,    20.498 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },     13.28753,   0.00222,     0.075 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,   0,  -1 },     13.29293,   0.00016,     0.038 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    166.71469,   0.00007,     0.039 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    167.04427,   0.00008,     0.026 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   5,  -2,  -1 },     13.28731,   0.00002,     0.040 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   0,   0 },     90.49502,   0.00002,     0.046 },
 { {   0,   3,  -1,  -2,   0,   0,   0,   0,  -2,   2,   0 },    270.49502,   0.00002,     0.218 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.98955,   0.00003,     0.044 },
 { {   0,   3,  -1,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.98955,   0.00003,     0.270 },
 { {   0,   3,  -1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.40749,   0.00003,     0.048 },
 { {   0,   3,  -1,   0,  -1,   0,   0,   0,  -2,   2,   0 },      0.40749,   0.00003,     0.180 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    132.19946,   0.00003,     0.029 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    158.59623,   0.00024,     0.048 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    134.26982,   0.00004,     0.131 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },     98.19776,   0.00004,     0.046 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.67249,   0.00012,     0.107 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    294.99561,   0.00052,     0.258 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    295.78164,   0.00020,     0.058 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    117.97900,   0.00003,     0.062 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    255.92968,   0.00002,     0.090 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    303.30416,   0.00034,     0.065 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    116.31525,   0.00016,     0.035 },
 { {   0,   3,   0,   0,  -2,   0,   0,   0,  -2,   0,   0 },    359.02651,   0.00002,     0.050 },
 { {   0,   3,   0,   0,  -2,   0,   0,   0,  -2,   2,   0 },    359.02651,   0.00002,     0.154 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },     22.27467,   0.00009,     0.050 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    202.18343,   0.00003,     0.205 },
 { {   0,   3,   0,   0,   1,   0,   0,   0,   2,   0,   0 },     35.38836,   0.00004,     0.034 },
 { {   0,   3,   1,   0,  -1,   0,   0,   0,   2,  -2,   0 },    242.57764,   0.00002,     0.249 },
 { {   0,   3,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },     62.57764,   0.00002,     0.033 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },     38.32687,   0.00002,     0.136 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    218.32687,   0.00002,     0.052 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    216.20660,   0.00005,     0.244 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     34.67343,   0.00003,     0.033 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   2,   0 },     38.21162,   0.00002,     0.018 },
 { {   0,   3,   1,   0,   1,   0,   0,   0,   2,  -1,   0 },    209.51773,   0.00003,     0.057 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.08098,   0.00005,     0.056 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.08098,   0.00005,     0.116 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    114.78509,   0.00003,     0.145 },
 { {   0,   3,   5, -15,   0,   0,   0,   0,   2,  -3,   0 },    246.70371,   0.00003,     0.076 },
 { {   0,   3,   5, -15,   0,   0,   0,   0,   2,  -1,   0 },     66.70371,   0.00003,     0.075 },
 { {   0,   4, -10,   3,   0,   0,   0,   0,  -2,   1,   0 },     62.55143,   0.00002,     0.075 },
 { {   0,   4, -10,   3,   0,   0,   0,   0,  -2,   3,   0 },    242.55143,   0.00002,     0.076 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     35.92444,   0.00004,     0.037 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    127.03815,   0.00003,     0.072 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    215.92444,   0.00004,     1.381 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    307.03815,   0.00003,     0.080 },
 { {   0,   4,  -8,   0,   0,  -1,   0,   0,  -2,   1,   0 },     54.31545,   0.00006,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.77386,   0.00003,     0.038 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    214.69537,   0.00002,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.77386,   0.00003,     3.622 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },     34.69537,   0.00002,     0.074 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    321.81112,   0.00003,     0.036 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    141.81112,   0.00003,     0.668 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },    132.07381,   0.00002,     0.038 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,  -2,   1,   0 },    270.30031,   0.00009,     0.078 },
 { {   0,   4,  -8,   0,   1,   0,   0,   0,  -2,   2,   0 },    312.07381,   0.00002,     2.775 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    112.21405,   0.00006,     0.040 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    146.84751,   0.00008,     0.083 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    344.18890,   0.00010,     0.073 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    157.95616,   0.00007,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },      2.72881,   0.00011,     0.091 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     14.89368,   0.00008,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },     75.16771,   0.00004,     0.020 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   1,   0 },     74.75645,   0.00008,     0.028 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },     80.06348,   0.00003,     0.045 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },     76.57425,   0.00027,     0.027 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     75.38155,   0.00393,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },     75.61732,   0.00316,     0.091 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    251.94183,   0.00017,     0.439 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },     79.23845,   0.00005,     0.085 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    252.19468,   0.00015,     0.084 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },     75.40468,   0.00029,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     72.68050,   0.00447,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    253.74084,   0.00095,     1.992 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    252.62772,   0.00413,     0.073 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    250.35732,   0.00020,     0.037 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },     79.21336,   0.00008,     0.072 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },     79.23846,   0.00027,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    250.44634,   0.00024,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },     79.19328,   0.00007,     0.038 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },    259.21336,   0.00008,     0.079 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    251.25825,   0.00008,     0.786 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    140.18286,   0.00006,     0.083 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    252.86840,   0.00078,     0.040 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    254.44645,   0.00013,     0.026 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },     75.19751,   0.00005,     0.020 },
 { {   0,   4,  -6,   0,   1,   0,   0,   0,   2,  -2,   0 },    359.54365,   0.00003,     0.842 },
 { {   0,   4,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },    359.54365,   0.00003,     0.039 },
 { {   0,   4,  -6,   0,   2,   0,   0,   0,  -2,   0,   0 },    269.87138,   0.00005,     0.042 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   2,  -3,   0 },    218.67202,   0.00002,     0.074 },
 { {   0,   4,  -6,   2,   0,   0,   0,   0,   2,  -1,   0 },     38.67202,   0.00002,     0.077 },
 { {   0,   4,  -5,   0,  -1,   0,   0,   0,   2,  -1,   0 },      8.18993,   0.00004,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    180.25913,   0.00003,     0.021 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },    183.40934,   0.00010,     0.029 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    182.56671,   0.00003,     0.047 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    183.93937,   0.00021,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    184.28906,   0.00414,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    182.93252,   0.00289,     0.100 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },      4.15949,   0.00017,     0.305 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    181.04546,   0.00004,     0.060 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },    166.52770,   0.00042,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    358.59584,   0.00005,     0.092 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    165.87667,   0.00019,     0.391 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.54760,   0.00010,     0.091 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.53829,   0.00011,     0.063 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,  -1 },    166.38043,   0.00002,     0.450 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },      2.86448,   0.00002,     0.026 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },      1.46847,   0.00022,     0.040 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },      0.78529,   0.00444,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    346.44170,   0.00004,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    178.51551,   0.00046,     0.666 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    175.96391,   0.00035,     0.068 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,  -1 },    166.51343,   0.00007,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -2,   1 },    346.48615,   0.00004,     1.279 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.51350,   0.00121,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    166.51855,   0.00095,     0.071 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.51639,   0.00033,     2.231 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    358.59602,   0.00030,     0.072 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    349.66498,   0.00023,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    346.51350,   0.00121,     0.073 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   2,  -1 },    346.51350,   0.00007,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    182.11004,   0.00159,     3.678 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,  -2 },    181.43363,   0.00008,     0.072 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },      1.37914,   0.02671,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    346.46312,   0.00016,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },      2.18950,   0.00447,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   1,  -2 },    182.25678,   0.00002,     0.079 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },      2.66273,   0.00025,     0.025 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,  -1,  -1 },    166.51545,   0.00012,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   3,   0,  -1 },    346.51131,   0.00012,     0.040 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -3,   0 },      3.96732,   0.00005,     0.089 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },    358.81547,   0.00025,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },      1.16743,   0.00028,     0.027 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },      7.20549,   0.00006,     0.020 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   6,  -2,   0 },    184.61369,   0.00002,     0.020 },
 { {   0,   4,  -5,   0,   0,   1,   0,   0,   2,  -1,   0 },    269.71691,   0.00004,     0.077 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },    277.37805,   0.00011,     0.043 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,  -1,   0 },    268.00947,   0.00002,     0.086 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,   0,   0 },     88.85174,   0.00005,     0.630 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   0,   1,   0 },     88.00947,   0.00002,     0.067 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,  -3,   0 },     89.84212,   0.00003,     0.074 },
 { {   0,   4,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },    359.73782,   0.00005,     0.077 },
 { {   0,   4,  -5,   0,   2,   0,   0,   0,   2,   0,   0 },    330.29485,   0.00004,     0.038 },
 { {   0,   4,  -5,   2,   0,   0,   0,   0,   2,  -3,   0 },    314.65763,   0.00004,     0.080 },
 { {   0,   4,  -5,   2,   0,   0,   0,   0,   2,  -1,   0 },    134.65763,   0.00004,     0.071 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,  -2,   0,   0 },    283.10910,   0.00005,     0.045 },
 { {   0,   4,  -4,   0,  -1,   0,   0,   0,   2,  -1,   0 },    282.64858,   0.00006,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -6,   2,   0 },    269.12126,   0.00002,     0.022 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,  -1,   0 },    269.55058,   0.00003,     0.017 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },     90.04815,   0.00007,     0.021 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },     89.50002,   0.00022,     0.030 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -2,   0 },     89.68196,   0.00002,     0.021 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },     89.11658,   0.00037,     0.028 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     88.98050,   0.00534,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     88.89580,   0.00265,     0.111 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.06736,   0.00035,     0.234 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    268.59522,   0.00004,     0.057 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     92.44071,   0.00004,     0.101 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.26910,   0.00003,     0.027 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.37485,   0.00037,     0.042 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.27940,   0.00595,     0.093 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   2 },     90.09232,   0.00002,     0.062 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,  -2 },    270.05376,   0.00012,     0.041 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     90.14325,   0.00627,     0.400 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   2 },    270.03963,   0.00010,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,  -2 },    270.11425,   0.00012,     0.090 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     90.06717,   0.00732,     0.063 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },     90.05766,   0.00047,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   3,   0 },     90.35908,   0.00002,     0.024 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },     91.14020,   0.00007,     0.086 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    279.95183,   0.00004,     0.085 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },     90.03912,   0.00007,     0.619 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    270.08806,   0.00123,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     99.83600,   0.00006,     0.035 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     99.95183,   0.00004,     0.068 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },    270.10075,   0.00007,     0.036 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    270.39945,   0.00021,     0.080 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.01213,   0.00419,     1.374 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,  -2 },    269.86310,   0.00031,     0.078 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.99224,   0.07945,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.04232,   0.00006,     2.535 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    270.06185,   0.01476,     0.037 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    270.21115,   0.00070,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    269.95549,   0.00016,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -2,   0 },     90.10242,   0.00035,     0.076 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.06459,   0.00008,     0.038 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,   0,  -2 },     90.06452,   0.00004,     0.078 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   3,   0,   0 },    270.06454,   0.00023,     0.025 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.04943,   0.00084,     0.039 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,  -2 },     90.19917,   0.00005,     0.084 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },     90.12249,   0.00043,     0.026 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,  -2 },     90.01679,   0.00006,     0.040 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    270.01294,   0.00016,     0.019 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   4,   1,   0 },    270.21122,   0.00004,     0.015 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,  -2,   0,   0 },     75.71920,   0.00005,     0.045 },
 { {   0,   4,  -4,   0,   1,   0,   0,   0,   2,  -1,   0 },     50.45337,   0.00005,     0.071 },
 { {   0,   4,  -3,   0,  -2,   0,   0,   0,  -2,   0,   0 },    178.12706,   0.00002,     0.047 },
 { {   0,   4,  -3,   0,  -2,   0,   0,   0,  -2,   2,   0 },    358.12706,   0.00002,     0.196 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.39435,   0.00009,     0.293 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   0,   2,   0 },     91.13891,   0.00002,     0.033 },
 { {   0,   4,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },    273.96489,   0.00008,     0.036 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },    166.15224,   0.00004,     0.022 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    343.41837,   0.00005,     0.029 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    343.51870,   0.00062,     0.047 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    344.63919,   0.00027,     0.125 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    196.63623,   0.00002,     0.190 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },     13.30117,   0.00004,     0.115 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    192.89991,   0.00029,     0.045 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.29766,   0.00022,     0.220 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.29968,   0.00007,     0.111 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    193.31284,   0.00008,     0.056 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    346.86576,   0.00003,     0.043 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    345.05693,   0.00039,     0.103 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     34.85071,   0.00004,     0.286 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    352.82291,   0.00007,     0.060 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.32328,   0.00023,     0.093 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.32682,   0.00026,     0.062 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.38167,   0.00037,     0.408 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     13.32711,   0.00020,     0.034 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.32328,   0.00023,     0.064 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    170.89445,   0.00003,     0.087 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    347.10820,   0.00003,     0.579 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    169.26049,   0.00094,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    193.34634,   0.00004,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    350.37239,   0.00042,     0.035 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    347.99278,   0.00006,     0.024 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -1,  -1 },     13.32467,   0.00012,     0.072 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,   0,  -1 },     13.32938,   0.00003,     0.037 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    180.16528,   0.00002,     0.049 },
 { {   0,   4,  -2,   0,  -1,   0,   0,   0,  -2,   2,   0 },      0.16528,   0.00002,     0.162 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    121.80694,   0.00003,     0.030 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    130.24491,   0.00023,     0.049 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    116.26584,   0.00011,     0.114 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    295.21657,   0.00049,     0.222 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    296.21226,   0.00016,     0.056 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    296.11163,   0.00030,     0.063 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    116.22628,   0.00017,     0.034 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     19.69050,   0.00008,     0.052 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    151.11331,   0.00002,     0.268 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    331.11331,   0.00002,     0.033 },
 { {   0,   4,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },     35.38594,   0.00003,     0.033 },
 { {   0,   4,   0,   0,  -1,   0,   0,   0,   2,  -2,   0 },    237.09829,   0.00003,     0.215 },
 { {   0,   4,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },     57.09829,   0.00003,     0.032 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     38.21204,   0.00003,     0.154 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    218.21204,   0.00003,     0.030 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    217.98852,   0.00004,     0.212 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,   0,   0 },     37.98852,   0.00004,     0.032 },
 { {   0,   4,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },     27.34016,   0.00002,     0.208 },
 { {   0,   4,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    207.34016,   0.00002,     0.032 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,  -3,   0 },    219.24544,   0.00002,     0.131 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,  -1,   0 },     39.24544,   0.00002,     0.053 },
 { {   0,   4,   1,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.78395,   0.00004,     0.031 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.06668,   0.00004,     0.058 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.06668,   0.00004,     0.108 },
 { {   0,   5, -13,   6,   0,   0,   0,   0,  -2,   1,   0 },     20.07513,   0.00004,     0.076 },
 { {   0,   5, -13,   6,   0,   0,   0,   0,  -2,   3,   0 },    200.07513,   0.00004,     0.075 },
 { {   0,   5, -11,   0,   1,   0,   0,   0,  -2,   1,   0 },    231.90583,   0.00003,     0.070 },
 { {   0,   5, -11,   0,   2,   0,   0,   0,  -2,   1,   0 },    141.65192,   0.00003,     0.070 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.76932,   0.00004,     0.038 },
 { {   0,   5, -11,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.76932,   0.00004,    28.573 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     33.07280,   0.00004,     0.038 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    213.07280,   0.00004,    10.166 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    347.55874,   0.00002,     0.274 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    167.55874,   0.00002,     0.044 },
 { {   0,   5, -10,   0,   0,   3,   0,   0,  -2,   1,   0 },    133.35767,   0.00002,     0.075 },
 { {   0,   5, -10,   0,   0,   3,   0,   0,  -2,   3,   0 },    313.35767,   0.00002,     0.075 },
 { {   0,   5, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },     37.75030,   0.00005,     0.075 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    330.28623,   0.00002,     0.039 },
 { {   0,   5,  -9,   0,  -1,   0,   0,   0,  -2,   2,   0 },    150.28623,   0.00002,     1.224 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    138.60764,   0.00004,     0.039 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    173.33113,   0.00003,     0.081 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    318.60326,   0.00004,     1.109 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    353.33113,   0.00003,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     38.59495,   0.00011,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   2,   0 },    129.24127,   0.00003,     0.075 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    307.49901,   0.00003,     0.037 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    260.15031,   0.00002,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    211.49853,   0.00003,     1.146 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },     81.45849,   0.00006,     0.081 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     83.88376,   0.00012,     0.094 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },     85.43401,   0.00021,     0.042 },
 { {   0,   5,  -9,   5,   0,   0,   0,   0,   2,  -3,   0 },    289.05792,   0.00002,     0.076 },
 { {   0,   5,  -9,   5,   0,   0,   0,   0,   2,  -1,   0 },    109.05792,   0.00002,     0.075 },
 { {   0,   5,  -8,   0,  -2,   0,   0,   0,   0,  -1,   0 },    108.04124,   0.00003,     0.075 },
 { {   0,   5,  -8,   0,  -2,   0,   0,   0,   0,   1,   0 },    288.04124,   0.00003,     0.076 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    170.31563,   0.00003,     0.020 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    162.27456,   0.00005,     0.028 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    155.13571,   0.00002,     0.044 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    161.45440,   0.00024,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.82780,   0.00290,     0.041 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    162.40088,   0.00265,     0.088 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    331.94560,   0.00012,     0.526 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    341.68149,   0.00006,     0.082 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    157.95842,   0.00018,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    159.40398,   0.00370,     0.076 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    333.76431,   0.00006,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    339.35204,   0.00019,     7.836 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    339.32620,   0.00360,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    339.23614,   0.00019,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    163.26260,   0.00008,     0.607 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    342.40205,   0.00204,     0.086 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    341.39875,   0.00198,     0.040 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },    341.29965,   0.00028,     0.026 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },    341.98319,   0.00002,     0.043 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },    341.63476,   0.00005,     0.028 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   0,  -1,   0 },    181.99814,   0.00003,     0.077 },
 { {   0,   5,  -8,   0,   2,   0,   0,   0,   0,   1,   0 },      1.99814,   0.00003,     0.074 },
 { {   0,   5,  -8,   3,   0,   0,   0,   0,   2,  -3,   0 },     86.82216,   0.00003,     0.075 },
 { {   0,   5,  -8,   3,   0,   0,   0,   0,   2,  -1,   0 },    266.82216,   0.00003,     0.076 },
 { {   0,   5,  -7,  -4,   0,   0,   0,   0,   2,   0,   0 },     22.33521,   0.00007,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },    257.92007,   0.00004,     0.021 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   1,   0 },    259.38719,   0.00008,     0.028 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },    251.74802,   0.00003,     0.046 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    256.69154,   0.00023,     0.027 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    259.22378,   0.00353,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    258.44704,   0.00271,     0.097 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     83.03640,   0.00017,     0.345 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    259.26743,   0.00002,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,  -1,   1 },    261.12219,   0.00005,     0.090 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    259.21958,   0.00035,     0.041 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     65.28731,   0.00003,     0.089 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    259.28236,   0.00004,     0.458 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },     79.82225,   0.00008,     0.088 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   1,   1 },    259.28236,   0.00003,     0.065 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   2,  -1 },    259.42981,   0.00004,     0.541 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    245.36292,   0.00005,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    260.94484,   0.00109,     0.082 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     78.27389,   0.00063,     0.887 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     79.21912,   0.00125,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },     79.62064,   0.00006,     0.036 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -2,  -1 },    259.37065,   0.00014,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    259.37065,   0.00250,     0.076 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    260.16047,   0.00060,     0.073 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },     65.28786,   0.00020,     0.074 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },     79.24342,   0.00010,     0.037 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },     79.37065,   0.00250,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   2,  -1 },     79.37065,   0.00014,     0.038 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    253.94524,   0.00013,     1.547 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,  -2 },    285.25987,   0.00002,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     74.09033,   0.00228,     0.079 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     77.93727,   0.00115,     0.039 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   1,  -2 },    285.24922,   0.00004,     0.081 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },     76.57274,   0.00010,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -2,  -1 },     79.43025,   0.00004,     0.588 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,  -1,  -1 },    259.56265,   0.00016,     0.087 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   0,  -1 },     79.26731,   0.00024,     0.040 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   3,   1,  -1 },     79.26743,   0.00002,     0.026 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },     76.31182,   0.00003,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },     76.89919,   0.00003,     0.027 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,   0,  -2 },    285.23270,   0.00005,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },     76.76787,   0.00004,     0.020 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,  -1,   0 },     89.85921,   0.00003,     0.084 },
 { {   0,   5,  -7,   0,   2,   0,   0,   0,   0,   1,   0 },    269.85921,   0.00003,     0.069 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -3,   0 },     67.23945,   0.00002,     0.078 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -2,   0 },     90.18132,   0.00003,     2.398 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,  -1,   0 },    247.23945,   0.00002,     0.073 },
 { {   0,   5,  -7,   2,   0,   0,   0,   0,   2,   0,   0 },    270.18132,   0.00003,     0.037 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   0,   0 },     99.36607,   0.00041,     0.040 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,  -2,   1,   0 },    100.36432,   0.00035,     0.087 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -2,   0 },     94.04642,   0.00002,     0.038 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,  -1,   0 },     95.11868,   0.00047,     0.075 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   1,   0 },    275.11576,   0.00047,     0.075 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   0,   2,   0 },    274.04734,   0.00002,     0.038 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,  -1,   0 },    280.35139,   0.00035,     0.087 },
 { {   0,   5,  -6,  -4,   0,   0,   0,   0,   2,   0,   0 },    279.35540,   0.00041,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -4,   1,   0 },    175.31106,   0.00005,     0.029 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    176.39718,   0.00019,     0.028 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    175.08840,   0.00144,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    175.24394,   0.00076,     0.107 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    193.20023,   0.00002,     0.058 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },    165.60847,   0.00026,     0.042 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    182.26221,   0.00004,     0.098 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    165.14361,   0.00009,     0.314 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },    166.43824,   0.00004,     0.096 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.42769,   0.00004,     0.061 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    180.95678,   0.00016,     0.041 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,  -2 },      2.30503,   0.00010,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    181.47610,   0.00293,     0.090 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -2 },      2.29599,   0.00080,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    166.36944,   0.00006,     0.089 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,  -2 },    182.39585,   0.00020,     0.087 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    182.50048,   0.00095,     0.065 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,  -2 },      2.32294,   0.00010,     0.557 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    182.56370,   0.00003,     0.035 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.37570,   0.00036,     0.082 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    166.40094,   0.00037,     0.068 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.43686,   0.00025,     0.931 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },      1.04499,   0.00017,     0.069 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    349.80391,   0.00018,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },    346.37570,   0.00036,     0.070 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,  -2 },      2.31229,   0.00037,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },      1.26851,   0.00120,     2.828 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },      2.30627,   0.00690,     0.076 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    181.90502,   0.02110,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    166.36944,   0.00034,     0.074 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },      3.51058,   0.00150,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },    182.31229,   0.00682,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },      5.66411,   0.00008,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   2,  -2 },    182.31229,   0.00037,     0.038 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,  -1,  -1 },    172.82164,   0.00010,     0.080 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,   0,  -1 },    346.39039,   0.00004,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   3,   0,   0 },      2.45751,   0.00004,     0.025 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -2,  -2 },    182.32294,   0.00010,     0.570 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },    183.24175,   0.00024,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,  -2 },      2.35792,   0.00033,     0.087 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    179.64969,   0.00011,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,  -2 },    182.29552,   0.00072,     0.040 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },      6.27440,   0.00005,     0.019 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   4,   1,  -2 },    182.30503,   0.00010,     0.026 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   6,  -3,   0 },    181.84748,   0.00002,     0.027 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },    279.25430,   0.00010,     0.044 },
 { {   0,   5,  -6,   0,   1,   0,   0,   0,   0,   0,   0 },     88.84540,   0.00004,     0.452 },
 { {   0,   5,  -5,  -4,   0,   0,   0,   0,  -2,   0,   0 },    351.37065,   0.00006,     0.042 },
 { {   0,   5,  -5,   0,  -1,   0,   0,   0,  -2,   0,   0 },    281.39709,   0.00005,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },     89.23584,   0.00003,     0.022 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },     88.75400,   0.00005,     0.030 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    268.83722,   0.00005,     0.050 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -3,   2,   0 },     89.99517,   0.00002,     0.134 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     88.05507,   0.00112,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },     86.26411,   0.00052,     0.120 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.20898,   0.00019,     0.204 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    265.60697,   0.00002,     0.055 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },     89.88147,   0.00008,     0.108 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.46374,   0.00003,     0.027 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     87.52888,   0.00079,     0.099 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,  -2 },    265.64350,   0.00012,     0.042 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     90.06946,   0.00626,     0.320 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   2 },    270.00907,   0.00008,     0.033 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },     90.03938,   0.00539,     0.061 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },     90.03687,   0.00039,     0.034 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },     90.28601,   0.00005,     0.446 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    270.22943,   0.00080,     0.065 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    100.14521,   0.00006,     0.035 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   1,   0 },    270.22121,   0.00005,     0.035 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },    261.84369,   0.00003,     0.084 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    269.99768,   0.00186,     0.739 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,  -2 },    267.19386,   0.00016,     0.082 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.06831,   0.03760,     0.068 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.03357,   0.00006,     0.980 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    270.05766,   0.01439,     0.036 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,  -2 },    266.58987,   0.00006,     0.070 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    270.01998,   0.00075,     0.024 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    269.90025,   0.00012,     0.018 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -2,   0 },    270.14383,   0.00012,     0.073 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.06639,   0.00008,     0.037 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   0,  -2 },    270.06353,   0.00017,     0.075 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   0,   0 },    270.06634,   0.00025,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   3,   2,   0 },    270.06633,   0.00002,     0.015 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.06048,   0.00044,     0.038 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,  -1,   0 },     90.08229,   0.00010,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    270.00299,   0.00019,     0.019 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   4,   1,   0 },    270.01967,   0.00004,     0.015 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,  -2,   0,   0 },     75.89672,   0.00005,     0.046 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,  -2,   0 },    256.28960,   0.00002,     0.695 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,  -1,   0 },    256.28964,   0.00004,     0.068 },
 { {   0,   5,  -5,   0,   1,   0,   0,   0,   2,   0,   0 },     76.28960,   0.00002,     0.036 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0 },    271.48661,   0.00002,     0.134 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,  -2,   3,   0 },    271.48661,   0.00002,     0.052 },
 { {   0,   5,  -4,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.47096,   0.00007,     0.247 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.94720,   0.00003,     0.049 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.99479,   0.00008,     0.136 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },     13.30982,   0.00004,     0.124 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.29703,   0.00021,     0.046 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.28919,   0.00018,     0.193 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    193.29020,   0.00005,     0.119 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    193.30982,   0.00004,     0.054 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    344.75629,   0.00013,     0.110 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },      3.11181,   0.00007,     0.242 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    340.32141,   0.00007,     0.058 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.32510,   0.00014,     0.098 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.32539,   0.00019,     0.060 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.32431,   0.00035,     0.325 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     13.32478,   0.00017,     0.033 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.32510,   0.00014,     0.061 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    355.78779,   0.00002,     0.425 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    173.44640,   0.00068,     0.064 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    352.66051,   0.00050,     0.035 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    351.61054,   0.00006,     0.024 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   3,  -1,  -1 },     13.32235,   0.00006,     0.069 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   3,   0,  -1 },     13.32747,   0.00002,     0.036 },
 { {   0,   5,  -3,   0,  -1,   0,   0,   0,  -2,   0,   0 },    179.84334,   0.00002,     0.051 },
 { {   0,   5,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0 },    179.84334,   0.00002,     0.147 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    123.76040,   0.00023,     0.051 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    114.63914,   0.00004,     0.157 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    115.28184,   0.00008,     0.123 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    293.15605,   0.00047,     0.195 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    296.26613,   0.00013,     0.054 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    272.73111,   0.00002,     0.101 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    298.00762,   0.00026,     0.060 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    121.91608,   0.00017,     0.033 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     19.61503,   0.00002,     0.054 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },     19.61503,   0.00002,     0.127 },
 { {   0,   5,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },     35.38516,   0.00003,     0.032 },
 { {   0,   5,  -1,   0,  -1,   0,   0,   0,   2,   0,   0 },     55.66043,   0.00006,     0.031 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,   2,   0 },     38.21247,   0.00002,     0.017 },
 { {   0,   5,  -1,   0,   1,   0,   0,   0,   2,  -2,   0 },     28.62264,   0.00004,     0.184 },
 { {   0,   5,  -1,   0,   1,   0,   0,   0,   2,   0,   0 },    208.62264,   0.00004,     0.031 },
 { {   0,   5,   0,   0,  -1,   0,   0,   0,   2,   0,   0 },     38.78376,   0.00003,     0.031 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.05032,   0.00004,     0.060 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.05032,   0.00004,     0.101 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    309.60542,   0.00002,     0.037 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    129.60542,   0.00002,     2.115 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.76222,   0.00003,     0.039 },
 { {   0,   6, -12,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.76222,   0.00003,     1.693 },
 { {   0,   6, -11,   0,  -1,   0,   0,   0,  -2,   1,   0 },     62.71339,   0.00002,     0.078 },
 { {   0,   6, -11,   0,  -1,   0,   0,   0,  -2,   3,   0 },     62.71339,   0.00002,     0.073 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     68.22018,   0.00003,     0.038 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    287.21763,   0.00004,     0.079 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    248.22018,   0.00003,     1.897 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    107.21763,   0.00004,     0.073 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,   0,  -1,   0 },    140.43674,   0.00003,     0.069 },
 { {   0,   6, -11,   0,   1,   0,   0,   0,   0,   1,   0 },    320.43674,   0.00003,     0.083 },
 { {   0,   6, -11,   0,   2,   0,   0,   0,  -2,   1,   0 },     21.95740,   0.00003,     0.080 },
 { {   0,   6, -11,   0,   2,   0,   0,   0,  -2,   3,   0 },    201.95740,   0.00003,     0.072 },
 { {   0,   6, -11,   0,   3,   0,   0,   0,   2,  -1,   0 },    305.90102,   0.00003,     0.095 },
 { {   0,   6, -11,   0,   3,   0,   0,   0,   2,   0,   0 },    305.90102,   0.00008,     0.042 },
 { {   0,   6, -10,  -3,   0,   0,   0,   0,  -2,   1,   0 },    176.49969,   0.00002,     0.075 },
 { {   0,   6, -10,  -3,   0,   0,   0,   0,  -2,   3,   0 },    356.49969,   0.00002,     0.076 },
 { {   0,   6, -10,   0,  -1,   0,   0,   0,  -2,   1,   0 },    331.62613,   0.00006,     0.085 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,  -2,   0 },    141.70494,   0.00002,     0.019 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    233.03879,   0.00007,     0.040 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    225.52828,   0.00006,     0.085 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    325.93503,   0.00003,     0.655 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    325.14385,   0.00003,     0.068 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.19644,   0.00005,     0.074 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    108.19644,   0.00005,     0.077 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },     65.70210,   0.00007,     0.089 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   2,   0,   0 },     67.67815,   0.00013,     0.041 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,  -2,   1,   0 },    129.00000,   0.00003,     0.086 },
 { {   0,   6, -10,   0,   1,   0,   0,   0,  -2,   3,   0 },    129.00000,   0.00003,     0.067 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,  -1,   0 },     23.07924,   0.00004,     0.026 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   0,   0 },     23.02684,   0.00047,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,  -2,   1,   0 },     23.74612,   0.00038,     0.087 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -2,   0 },     19.90805,   0.00003,     0.038 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,  -1,   0 },     20.53295,   0.00063,     0.075 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   1,   0 },    200.52912,   0.00063,     0.075 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   0,   2,   0 },    199.90711,   0.00003,     0.038 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,  -1,   0 },    203.74093,   0.00037,     0.087 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   0,   0 },    203.01871,   0.00047,     0.040 },
 { {   0,   6, -10,   0,   3,   0,   0,   0,   2,   1,   0 },    203.07083,   0.00004,     0.026 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    160.25515,   0.00004,     0.027 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    157.37067,   0.00044,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    162.96082,   0.00025,     0.093 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,  -1,   1 },    336.55187,   0.00005,     0.087 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    157.56168,   0.00039,     0.080 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    338.88164,   0.00013,     1.328 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    341.12762,   0.00041,     0.071 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },    336.55187,   0.00028,     0.075 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    340.05078,   0.00014,     0.076 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    175.52352,   0.00056,     0.082 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    178.70545,   0.00005,     0.039 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -3,   0 },    179.32138,   0.00003,     0.096 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    180.09016,   0.00009,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },    180.18063,   0.00004,     0.027 },
 { {   0,   6,  -9,   2,   0,   0,   0,   0,   2,  -3,   0 },     91.52469,   0.00003,     0.076 },
 { {   0,   6,  -9,   2,   0,   0,   0,   0,   2,  -1,   0 },    271.52469,   0.00003,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },     77.24301,   0.00012,     0.028 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    318.07271,   0.00005,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    263.81945,   0.00015,     0.103 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -3,   0 },    258.18763,   0.00002,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    259.31839,   0.00015,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     77.13663,   0.01165,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    256.86583,   0.00176,     0.570 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    253.64404,   0.00037,     0.067 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },    258.98609,   0.00009,     0.080 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    258.98829,   0.00006,     0.070 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     72.24178,   0.00009,     0.071 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },     78.98828,   0.00006,     0.036 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },     78.98609,   0.00009,     0.072 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -4,   0 },    258.84950,   0.00010,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    258.84554,   0.00183,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    257.19270,   0.00391,    47.898 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,  -2 },    257.24624,   0.00026,     0.073 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     77.22087,   0.07054,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },     79.10173,   0.00007,     0.077 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     77.62249,   0.00439,     0.038 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },     77.31751,   0.00028,     0.025 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -4,   0 },    254.51529,   0.00003,     0.557 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -3,   0 },     76.37024,   0.00074,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },     77.15486,   0.00154,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },     77.23009,   0.00067,     0.026 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },     77.24443,   0.00007,     0.020 },
 { {   0,   6,  -8,   0,   1,   0,   0,   0,   2,  -1,   0 },    154.22629,   0.00012,     0.075 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   2,  -3,   0 },     84.65748,   0.00002,     0.076 },
 { {   0,   6,  -8,   0,   2,   0,   0,   0,   2,  -1,   0 },    264.65748,   0.00002,     0.075 },
 { {   0,   6,  -8,   2,   0,   0,   0,   0,   2,  -2,   0 },     90.17123,   0.00002,     0.959 },
 { {   0,   6,  -8,   2,   0,   0,   0,   0,   2,   0,   0 },    270.17123,   0.00002,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    179.73326,   0.00034,     0.045 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    174.37287,   0.00009,     0.115 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    188.17738,   0.00003,     0.221 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,  -1,   1 },    346.30780,   0.00003,     0.106 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    166.26861,   0.00017,     0.043 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    166.29517,   0.00007,     0.263 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1,   1 },    166.30780,   0.00003,     0.059 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    170.98071,   0.00006,     0.042 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    236.35341,   0.00028,     0.095 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    177.52829,   0.00059,     0.363 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    181.88327,   0.00094,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    181.64035,   0.00002,     0.034 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.25252,   0.00030,     0.087 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    168.12012,   0.00021,     0.065 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.29555,   0.00024,     0.588 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },      1.97207,   0.00013,     0.066 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    346.29617,   0.00010,     0.035 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },    346.27602,   0.00018,     0.067 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    359.69128,   0.00037,     1.021 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    180.87819,   0.00794,     0.070 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,  -1 },    166.24924,   0.00007,     0.071 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },      3.55773,   0.00210,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },      2.15314,   0.00013,     0.025 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -2,  -1 },    346.21648,   0.00004,    13.397 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -2,   0 },      2.30487,   0.00011,     0.075 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -1,  -1 },    166.21948,   0.00070,     0.076 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },      2.44443,   0.00004,     0.025 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },    183.38839,   0.00009,     0.039 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },    356.36602,   0.00002,     0.019 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,  -2,   0,   0 },    280.41729,   0.00009,     0.046 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,  -2,   0 },     88.83709,   0.00002,     0.042 },
 { {   0,   6,  -7,   0,   1,   0,   0,   0,   0,   0,   0 },     88.83709,   0.00002,     0.352 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,  -2,   0,   0 },    279.52993,   0.00005,     0.047 },
 { {   0,   6,  -6,   0,  -1,   0,   0,   0,   2,   0,   0 },    276.66377,   0.00004,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.62701,   0.00011,     0.048 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    273.94590,   0.00005,     0.129 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    273.35411,   0.00012,     0.181 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     89.94384,   0.00006,     0.116 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    268.15115,   0.00004,     0.044 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    271.07852,   0.00058,     0.105 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,  -2 },    269.99202,   0.00006,     0.043 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     90.09504,   0.00527,     0.266 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     90.01085,   0.00382,     0.059 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },     90.00205,   0.00030,     0.033 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    270.01950,   0.00056,     0.062 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    100.45156,   0.00006,     0.034 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.02147,   0.00102,     0.505 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },    270.01924,   0.00007,     0.086 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.04177,   0.02196,     0.066 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.03307,   0.00005,     0.608 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    270.06389,   0.01237,     0.035 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },    270.02986,   0.00003,     0.067 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    270.02880,   0.00068,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },    270.01732,   0.00010,     0.018 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,  -2,   0 },    270.06624,   0.00005,     0.070 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.06572,   0.00007,     0.036 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   3,   0,   0 },    270.06569,   0.00022,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.02028,   0.00025,     0.037 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.04136,   0.00002,     0.025 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   0,   0 },    270.02513,   0.00020,     0.019 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   4,   2,   0 },    270.03305,   0.00002,     0.013 },
 { {   0,   6,  -6,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.05912,   0.00005,     0.048 },
 { {   0,   6,  -6,   0,   1,   0,   0,   0,   2,   0,   0 },     76.33915,   0.00003,     0.035 },
 { {   0,   6,  -5,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.50669,   0.00008,     0.214 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    177.54722,   0.00010,     0.050 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    345.15050,   0.00006,     0.149 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    184.25683,   0.00003,     0.153 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.28005,   0.00016,     0.048 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.27807,   0.00014,     0.172 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    337.30235,   0.00006,     0.118 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    354.25742,   0.00006,     0.210 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    339.58025,   0.00003,     0.056 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.32174,   0.00010,     0.105 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.32154,   0.00011,     0.058 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.32130,   0.00031,     0.270 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },     13.32123,   0.00014,     0.033 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.32174,   0.00010,     0.059 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    176.15627,   0.00055,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    353.84126,   0.00049,     0.034 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    354.49982,   0.00007,     0.023 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    120.18639,   0.00020,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    115.43968,   0.00006,     0.133 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    294.20019,   0.00041,     0.174 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    296.30756,   0.00010,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.89921,   0.00003,     0.251 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    298.00674,   0.00019,     0.058 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    124.07571,   0.00017,     0.033 },
 { {   0,   6,  -2,   0,  -1,   0,   0,   0,   2,   0,   0 },     63.60897,   0.00006,     0.031 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    218.23943,   0.00003,     0.182 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     38.23943,   0.00003,     0.048 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    218.21227,   0.00002,     0.031 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },     38.21227,   0.00002,     0.017 },
 { {   0,   6,  -2,   0,   1,   0,   0,   0,   2,   0,   0 },    208.63880,   0.00007,     0.031 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.03320,   0.00003,     0.063 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.03320,   0.00003,     0.095 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    306.98408,   0.00003,     0.143 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    126.98408,   0.00003,     0.030 },
 { {   0,   6,   0,  -2,   0,   0,   0,   0,   2,  -2,   0 },     38.30059,   0.00002,     0.145 },
 { {   0,   6,   0,  -2,   0,   0,   0,   0,   2,   0,   0 },    218.30059,   0.00002,     0.030 },
 { {   0,   7, -19,  11,   0,   0,   0,   0,  -2,   1,   0 },    248.55636,   0.00004,     0.075 },
 { {   0,   7, -19,  11,   0,   0,   0,   0,  -2,   3,   0 },     68.55636,   0.00004,     0.075 },
 { {   0,   7, -14,   5,   0,   0,   0,   0,   0,  -1,   0 },     19.17636,   0.00003,     0.076 },
 { {   0,   7, -14,   5,   0,   0,   0,   0,   0,   1,   0 },    199.17636,   0.00003,     0.075 },
 { {   0,   7, -13,   0,  -1,   0,   0,   0,  -2,   1,   0 },    141.98380,   0.00007,     0.076 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    309.96242,   0.00002,     0.038 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },     23.69386,   0.00002,     0.076 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    129.96242,   0.00002,     6.550 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    353.67546,   0.00003,     0.075 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   0,   0 },    141.57778,   0.00002,     0.038 },
 { {   0,   7, -13,   0,   2,   0,   0,   0,  -2,   2,   0 },    321.57778,   0.00002,     3.113 },
 { {   0,   7, -13,   0,   3,   0,   0,   0,  -2,   1,   0 },    131.58668,   0.00003,     0.078 },
 { {   0,   7, -13,   0,   3,   0,   0,   0,  -2,   3,   0 },    311.58668,   0.00003,     0.073 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.75298,   0.00003,     0.040 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   1,   0 },    321.75298,   0.00003,     0.083 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.75298,   0.00003,     0.822 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,  -2,   3,   0 },    141.75298,   0.00003,     0.069 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,   0,  -1,   0 },    141.75298,   0.00002,     0.072 },
 { {   0,   7, -13,   2,   0,   0,   0,   0,   0,   1,   0 },    321.75298,   0.00002,     0.079 },
 { {   0,   7, -13,   3,   0,   0,   0,   0,   0,  -1,   0 },    110.15736,   0.00004,     0.075 },
 { {   0,   7, -13,   3,   0,   0,   0,   0,   0,   1,   0 },    290.15736,   0.00004,     0.076 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     67.04817,   0.00003,     0.039 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     83.34211,   0.00005,     0.083 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    247.04817,   0.00003,     0.868 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    263.34211,   0.00005,     0.069 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    235.15313,   0.00012,     0.041 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    213.33484,   0.00006,     0.090 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    328.16473,   0.00002,     0.465 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    322.02122,   0.00003,     0.065 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    285.57553,   0.00002,     0.038 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    220.28157,   0.00010,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    105.57553,   0.00002,     2.641 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   1,   0 },     40.28157,   0.00010,     0.073 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },     66.99206,   0.00004,     0.084 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     63.81187,   0.00004,     0.040 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },    128.95830,   0.00004,     0.041 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },    308.95830,   0.00004,     0.447 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,   0,  -1,   0 },    128.92048,   0.00003,     0.078 },
 { {   0,   7, -11,   0,   1,   0,   0,   0,   0,   1,   0 },    308.92048,   0.00003,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    354.10266,   0.00022,     0.043 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    344.31894,   0.00006,     0.099 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    164.25330,   0.00010,     0.084 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    343.61803,   0.00003,     2.529 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    164.25301,   0.00057,     0.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    163.58028,   0.00021,     0.038 },
 { {   0,   7, -10,   0,   3,   0,   0,   0,   2,  -3,   0 },    171.50309,   0.00002,     0.075 },
 { {   0,   7, -10,   0,   3,   0,   0,   0,   2,  -1,   0 },    124.08513,   0.00005,     0.076 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    259.38740,   0.00013,     0.045 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.04781,   0.00003,     0.110 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    261.31826,   0.00021,     0.092 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    247.62530,   0.00012,     0.420 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    255.69081,   0.00021,     0.064 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },    258.92485,   0.00004,     0.084 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     73.25073,   0.00005,     0.068 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },     78.93860,   0.00005,     0.035 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },     78.92485,   0.00004,     0.069 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },     81.17194,   0.00006,     0.079 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },     79.63565,   0.00011,     1.654 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    258.56750,   0.00235,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },     82.88021,   0.00037,     0.037 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },     79.29444,   0.00005,     0.025 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    257.87237,   0.00003,     0.039 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,  -3,   0 },    359.89632,   0.00002,     0.080 },
 { {   0,   7,  -9,   0,   1,   0,   0,   0,   2,  -1,   0 },    179.89632,   0.00002,     0.072 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },      5.05442,   0.00003,     0.047 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    181.84078,   0.00008,     0.123 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },    166.13057,   0.00011,     0.045 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    166.14564,   0.00005,     0.226 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    182.41518,   0.00016,     0.101 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    181.16778,   0.00056,     0.296 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    182.03723,   0.00075,     0.060 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },    166.14314,   0.00011,     0.091 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    166.15114,   0.00010,     0.063 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    346.15028,   0.00020,     0.430 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },      1.93450,   0.00009,     0.064 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    346.15117,   0.00007,     0.034 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },    346.14314,   0.00011,     0.064 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },      1.30967,   0.00003,     0.086 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },      1.60736,   0.00018,     0.623 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.05735,   0.00431,     0.067 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,  -1 },    168.11215,   0.00004,     0.068 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },      3.06151,   0.00211,     0.036 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },      2.29066,   0.00012,     0.024 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,  -1,  -1 },    353.82226,   0.00008,     0.072 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   3,   0,   0 },      2.43806,   0.00004,     0.025 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },    181.85233,   0.00005,     0.038 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },    357.15838,   0.00002,     0.019 },
 { {   0,   7,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },    281.58450,   0.00009,     0.047 },
 { {   0,   7,  -7,   0,  -1,   0,   0,   0,  -2,   0,   0 },    277.54166,   0.00005,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.93979,   0.00003,     0.030 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.84112,   0.00058,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.30773,   0.00009,     0.141 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.86345,   0.00015,     0.163 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     89.94656,   0.00005,     0.125 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    269.75419,   0.00005,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.56717,   0.00076,     0.113 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,  -2 },    269.98320,   0.00005,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     90.04790,   0.00430,     0.228 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     90.04125,   0.00268,     0.057 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },     90.16610,   0.00021,     0.032 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    270.01131,   0.00032,     0.060 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    100.73027,   0.00006,     0.033 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.19473,   0.00063,     0.384 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,  -2 },    270.02690,   0.00008,     0.091 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.02547,   0.01390,     0.063 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.02303,   0.00011,     0.440 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    270.02081,   0.01013,     0.034 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   1,  -2 },     90.02900,   0.00002,     0.064 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00815,   0.00060,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },    270.00277,   0.00008,     0.018 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.06255,   0.00007,     0.035 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   3,   0,   0 },    270.06251,   0.00021,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.92572,   0.00018,     0.037 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.99295,   0.00006,     0.025 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },    270.01117,   0.00015,     0.019 },
 { {   0,   7,  -7,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.20478,   0.00005,     0.049 },
 { {   0,   7,  -6,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.50885,   0.00007,     0.189 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    198.05290,   0.00002,     0.052 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.10478,   0.00002,     0.164 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    198.30215,   0.00002,     0.140 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.26681,   0.00026,     0.049 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.47585,   0.00009,     0.155 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    330.05351,   0.00003,     0.127 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.31572,   0.00008,     0.112 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.31541,   0.00008,     0.056 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.31548,   0.00028,     0.231 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },     13.31512,   0.00011,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.31572,   0.00008,     0.057 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    178.10236,   0.00044,     0.059 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    355.18523,   0.00045,     0.033 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    356.30321,   0.00006,     0.023 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.98422,   0.00005,     0.049 },
 { {   0,   7,  -5,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.98422,   0.00005,     0.161 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    117.18965,   0.00031,     0.054 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    116.55062,   0.00005,     0.145 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    294.28194,   0.00035,     0.157 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    296.32215,   0.00008,     0.051 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     41.92710,   0.00002,     0.217 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.96122,   0.00016,     0.056 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    123.89896,   0.00016,     0.032 },
 { {   0,   7,  -4,   0,   1,   0,   0,   0,   2,  -2,   0 },     39.10592,   0.00002,     0.176 },
 { {   0,   7,  -4,   0,   1,   0,   0,   0,   2,   0,   0 },    219.10592,   0.00002,     0.031 },
 { {   0,   7,  -3,   0,  -1,   0,   0,   0,   2,  -2,   0 },     54.70219,   0.00003,     0.153 },
 { {   0,   7,  -3,   0,  -1,   0,   0,   0,   2,   0,   0 },     54.70219,   0.00003,     0.030 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    218.23584,   0.00003,     0.205 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     38.23584,   0.00003,     0.046 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    218.02587,   0.00002,     0.151 },
 { {   0,   7,  -3,   0,   1,   0,   0,   0,   2,   0,   0 },    208.65066,   0.00006,     0.030 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.01220,   0.00004,     0.065 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    212.01220,   0.00004,     0.090 },
 { {   0,   7,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    236.63392,   0.00004,     0.075 },
 { {   0,   7,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },     56.63392,   0.00004,     0.076 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     19.62237,   0.00003,     0.035 },
 { {   0,   8, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    199.62237,   0.00003,     0.450 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    216.04549,   0.00003,     0.074 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   3,   0 },     36.04549,   0.00003,     0.077 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    193.76468,   0.00002,     0.066 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     13.76468,   0.00002,     0.089 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    353.70382,   0.00007,     0.105 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    353.49474,   0.00018,     0.044 },
 { {   0,   8, -15,   0,   2,   0,   0,   0,  -2,   1,   0 },    234.67656,   0.00003,     0.075 },
 { {   0,   8, -15,   0,   3,   0,   0,   0,  -2,   1,   0 },     41.04069,   0.00002,     0.076 },
 { {   0,   8, -15,   0,   3,   0,   0,   0,  -2,   3,   0 },    221.04069,   0.00002,     0.075 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.39603,   0.00004,     0.026 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    251.79220,   0.00022,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    251.02859,   0.00007,     0.080 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },     72.87196,   0.00024,     0.070 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     77.61991,   0.00023,     1.004 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    248.87102,   0.00068,     0.082 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   2,   0 },    248.42028,   0.00004,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },     71.51680,   0.00011,     0.361 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.78371,   0.00192,     0.095 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    249.77975,   0.00302,     0.042 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    250.03388,   0.00029,     0.027 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,  -2,   0 },    250.23773,   0.00003,     0.046 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,  -1,   0 },    250.42638,   0.00009,     0.028 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   4,   0,   0 },    250.08534,   0.00006,     0.021 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.74157,   0.00003,     0.041 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.74157,   0.00003,     0.543 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,   0,  -1,   0 },    132.04044,   0.00002,     0.076 },
 { {   0,   8, -14,   2,   0,   0,   0,   0,   0,   1,   0 },    312.04044,   0.00002,     0.075 },
 { {   0,   8, -13,   0,  -1,   0,   0,   0,   0,  -1,   0 },    241.61017,   0.00002,     0.075 },
 { {   0,   8, -13,   0,  -1,   0,   0,   0,   0,   1,   0 },     61.61017,   0.00002,     0.076 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,  -1,   0 },    322.16926,   0.00003,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   0,   0 },    324.94501,   0.00025,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   1,   0 },    327.63996,   0.00066,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    326.23761,   0.00034,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -3,   0,   0 },    327.99306,   0.00010,     0.027 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -3,   1,   0 },    147.89662,   0.00005,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -2,   0 },    326.88692,   0.00027,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    326.26434,   0.00201,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   2 },    147.88744,   0.00003,     0.090 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    327.34466,   0.03005,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   2 },    147.21790,   0.00012,     0.474 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    328.08476,   0.03125,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   2 },    147.20251,   0.00008,     0.065 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    151.18584,   0.00172,     0.562 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    149.88967,   0.00012,     0.067 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,  -1,   0 },    145.67501,   0.00003,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    145.35416,   0.00053,     0.081 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -1,   1,   0 },    325.58226,   0.00003,     1.122 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -3,   0 },    324.12768,   0.00011,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    322.92106,   0.00184,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    323.44138,   0.03577,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   2 },    143.24613,   0.00013,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    155.20447,   0.00008,   238.919 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,  -2 },    323.24613,   0.00013,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    143.43350,   0.03574,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    143.21180,   0.00183,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   3,   0 },    143.07687,   0.00011,     0.025 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,  -1,   0 },    147.88398,   0.00003,     1.133 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    328.02937,   0.00053,     0.081 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   1,   1,   0 },    327.88331,   0.00003,     0.039 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },    329.79237,   0.00012,     0.067 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    329.10687,   0.00168,     0.565 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,  -2 },    327.19653,   0.00008,     0.065 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    148.04700,   0.03103,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,  -2 },    327.08222,   0.00012,     0.475 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    147.21064,   0.02975,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,  -2 },    327.85242,   0.00003,     0.090 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,   0 },    146.29709,   0.00199,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   2,   0 },    146.82860,   0.00027,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   3,  -1,   0 },    327.84754,   0.00005,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   3,   0,   0 },    147.94074,   0.00010,     0.027 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -2,   0 },    147.95411,   0.00034,     0.044 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,  -1,   0 },    147.45242,   0.00065,     0.028 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   0,   0 },    144.76663,   0.00025,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   4,   1,   0 },    146.41033,   0.00003,     0.016 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,  -1,   0 },    250.86795,   0.00003,     0.076 },
 { {   0,   8, -13,   0,   1,   0,   0,   0,   0,   1,   0 },     70.86795,   0.00003,     0.075 },
 { {   0,   8, -12,  -2,   0,   0,   0,   0,   0,  -1,   0 },     12.32972,   0.00002,     0.075 },
 { {   0,   8, -12,  -2,   0,   0,   0,   0,   0,   1,   0 },    192.32972,   0.00002,     0.076 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    331.63038,   0.00003,     0.042 },
 { {   0,   8, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },    151.63038,   0.00003,     0.371 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -4,   0,   0 },    234.02951,   0.00002,     0.021 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -4,   1,   0 },    206.69608,   0.00008,     0.028 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    199.57518,   0.00020,     0.027 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    190.18965,   0.00131,     0.042 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    182.87260,   0.00074,     0.095 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    303.35822,   0.00003,     0.360 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },     68.44694,   0.00002,     0.039 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },     89.38750,   0.00063,     0.082 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    256.04686,   0.00027,     0.996 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    250.61840,   0.00083,     0.070 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    251.91705,   0.00004,     0.036 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,  -1,  -1 },     63.46151,   0.00006,     0.075 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     67.30464,   0.00014,     0.075 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   1,  -1 },    243.46151,   0.00006,     0.076 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    238.99254,   0.00032,     0.080 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    239.29732,   0.00050,     0.039 },
 { {   0,   8, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },    129.13031,   0.00005,     0.042 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     25.67120,   0.00015,     0.044 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     15.11923,   0.00004,     0.106 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    231.87292,   0.00002,     0.265 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    346.01552,   0.00016,     0.089 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    138.51872,   0.00002,     0.499 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    166.42489,   0.00005,     4.345 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    346.01570,   0.00093,     0.074 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    336.35744,   0.00004,     0.075 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    163.80300,   0.00013,     0.037 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    258.44358,   0.00002,     0.118 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    274.19922,   0.00004,     0.098 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    254.24549,   0.00020,     0.333 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    257.24314,   0.00016,     0.062 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,   1 },     78.88488,   0.00004,     0.035 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },     75.84735,   0.00002,     0.083 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },     84.13129,   0.00006,     0.813 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    260.88220,   0.00111,     0.069 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,   0 },     83.16739,   0.00038,     0.036 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   1,   0 },     80.55494,   0.00004,     0.024 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,  -1,  -1 },     78.88498,   0.00004,     0.074 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   3,   0,  -2 },     79.94439,   0.00011,     0.075 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },      0.47126,   0.00012,     0.048 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    182.35171,   0.00008,     0.134 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    165.95457,   0.00009,     0.046 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    185.17333,   0.00003,     0.108 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    180.50142,   0.00051,     0.250 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    182.28544,   0.00057,     0.058 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    193.64267,   0.00002,     0.033 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   3,   0 },    183.02019,   0.00002,     0.023 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,  -1 },    165.95227,   0.00007,     0.097 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },    165.98002,   0.00007,     0.060 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },    345.97891,   0.00016,     0.339 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },      2.11200,   0.00007,     0.061 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    345.98001,   0.00017,     0.034 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   1,  -1 },    345.95227,   0.00007,     0.062 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },     13.10183,   0.00006,     0.091 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    346.03675,   0.00012,     0.448 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.66942,   0.00286,     0.065 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },      4.02313,   0.00189,     0.035 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },      3.21028,   0.00014,     0.024 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   3,   0,   0 },      2.43499,   0.00004,     0.024 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },    191.45431,   0.00005,     0.037 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.44090,   0.00005,     0.025 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },      6.07232,   0.00004,     0.019 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },    282.74983,   0.00008,     0.048 },
 { {   0,   8,  -9,   0,   1,   0,   0,   0,   2,  -1,   0 },     88.69890,   0.00005,     0.064 },
 { {   0,   8,  -8,   0,  -1,   0,   0,   0,  -2,   0,   0 },    275.45992,   0.00005,     0.050 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.59394,   0.00003,     0.030 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.44366,   0.00061,     0.051 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.04455,   0.00034,     0.154 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    270.03820,   0.00012,     0.148 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },     89.93565,   0.00003,     0.136 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.23019,   0.00004,     0.046 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.99913,   0.00070,     0.121 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     90.06702,   0.00337,     0.200 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },     90.01596,   0.00192,     0.055 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },     89.24941,   0.00014,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    270.01152,   0.00023,     0.058 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    100.97656,   0.00005,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   2,   0 },    270.05078,   0.00002,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    272.05043,   0.00003,     0.100 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.49627,   0.00041,     0.310 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,  -2 },    270.01134,   0.00004,     0.097 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.01074,   0.00915,     0.061 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.01810,   0.00009,     0.345 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    270.02345,   0.00819,     0.034 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00649,   0.00054,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   2,   0 },    270.00553,   0.00008,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.05685,   0.00006,     0.035 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   3,   0,   0 },    270.05682,   0.00019,     0.024 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.67241,   0.00010,     0.036 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.02689,   0.00006,     0.024 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },    269.99968,   0.00013,     0.018 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   4,   1,   0 },    270.01622,   0.00002,     0.015 },
 { {   0,   8,  -8,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.33389,   0.00004,     0.051 },
 { {   0,   8,  -7,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.50798,   0.00007,     0.169 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    177.72934,   0.00002,     0.053 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    177.91659,   0.00002,     0.129 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.23733,   0.00024,     0.051 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.24871,   0.00007,     0.142 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    325.99290,   0.00004,     0.138 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    337.54744,   0.00003,     0.167 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.30707,   0.00006,     0.120 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.30668,   0.00006,     0.054 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.42223,   0.00022,     0.202 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },     13.30624,   0.00007,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.30707,   0.00006,     0.055 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.50040,   0.00038,     0.057 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    355.97249,   0.00042,     0.033 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },    357.45839,   0.00004,     0.023 },
 { {   0,   8,  -6,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.88680,   0.00002,     0.051 },
 { {   0,   8,  -6,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.88680,   0.00002,     0.146 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    116.77074,   0.00028,     0.056 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    116.29590,   0.00006,     0.160 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    294.56476,   0.00031,     0.143 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    296.29590,   0.00006,     0.049 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    293.82232,   0.00014,     0.054 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    120.33900,   0.00017,     0.032 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     38.20821,   0.00002,     0.111 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    218.20821,   0.00002,     0.028 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    218.06931,   0.00002,     0.049 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },     38.06931,   0.00002,     0.021 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    149.19834,   0.00003,     0.073 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    329.19834,   0.00003,     0.078 },
 { {   0,   9, -15,   0,   2,   0,   0,   0,  -2,   1,   0 },    141.58703,   0.00006,     0.086 },
 { {   0,   9, -15,   0,   3,   0,   0,   0,   0,  -1,   0 },    311.57463,   0.00003,     0.075 },
 { {   0,   9, -15,   0,   3,   0,   0,   0,   0,   1,   0 },    131.57463,   0.00003,     0.076 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.72800,   0.00002,     0.042 },
 { {   0,   9, -15,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.72800,   0.00002,     0.405 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     22.96262,   0.00007,     0.041 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    202.96262,   0.00007,     0.416 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   2,  -3,   0 },    147.87953,   0.00003,     0.069 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    327.87953,   0.00003,     0.083 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,  -2,   1,   0 },    320.38725,   0.00002,     0.093 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,  -2,   3,   0 },    320.38725,   0.00002,     0.064 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,   0,  -1,   0 },    320.38725,   0.00003,     0.080 },
 { {   0,   9, -14,   0,   1,   0,   0,   0,   0,   1,   0 },    140.38725,   0.00003,     0.072 },
 { {   0,   9, -13,  -3,   0,   0,   0,   0,   0,  -1,   0 },    356.56794,   0.00003,     0.076 },
 { {   0,   9, -13,  -3,   0,   0,   0,   0,   0,   1,   0 },    176.56794,   0.00003,     0.075 },
 { {   0,   9, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    331.62732,   0.00006,     0.043 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    131.95171,   0.00002,     0.043 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    321.77865,   0.00003,     0.102 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    311.95171,   0.00002,     0.294 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    141.77865,   0.00003,     0.060 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    276.67837,   0.00005,     0.086 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    141.77864,   0.00002,     0.067 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },    151.18213,   0.00004,     0.075 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.44120,   0.00023,     0.076 },
 { {   0,   9, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    128.87919,   0.00006,     0.043 },
 { {   0,   9, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    143.65273,   0.00006,     0.045 },
 { {   0,   9, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },    320.81335,   0.00002,     0.232 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     30.04424,   0.00003,     0.045 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    210.04424,   0.00003,     0.227 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    345.25518,   0.00005,     0.094 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    345.25519,   0.00026,     0.071 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    165.21558,   0.00017,     0.037 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.93359,   0.00002,     0.047 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    251.99282,   0.00002,     0.185 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    256.38103,   0.00018,     0.276 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    258.52994,   0.00012,     0.059 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,   1 },     78.82543,   0.00004,     0.034 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },     82.17311,   0.00004,     0.539 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    262.20512,   0.00072,     0.066 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     81.70424,   0.00034,     0.035 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   1,   0 },     80.98663,   0.00002,     0.024 },
 { {   0,   9, -11,   0,   2,   0,   0,   0,   2,  -2,   0 },    269.82719,   0.00002,     0.494 },
 { {   0,   9, -11,   0,   2,   0,   0,   0,   2,   0,   0 },     89.82719,   0.00002,     0.035 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    178.53187,   0.00004,     0.146 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    188.63621,   0.00002,     0.051 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    165.75137,   0.00007,     0.047 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    180.64879,   0.00046,     0.216 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    182.16002,   0.00040,     0.056 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    182.30214,   0.00003,     0.032 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    173.44531,   0.00005,     0.103 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,  -1,   1 },    165.78169,   0.00004,     0.058 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    345.78035,   0.00014,     0.280 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   0 },      2.01732,   0.00005,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   1 },    345.78168,   0.00014,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   1,  -1 },    353.44531,   0.00005,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },      8.48278,   0.00002,     0.096 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    346.56540,   0.00007,     0.350 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.63706,   0.00199,     0.062 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,   0 },      3.54067,   0.00171,     0.034 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   1,   0 },      2.69849,   0.00012,     0.023 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   3,   0,   0 },      2.43334,   0.00004,     0.024 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,  -2,   0 },    189.15541,   0.00003,     0.036 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.24309,   0.00004,     0.024 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   4,   0,   0 },      6.86408,   0.00004,     0.018 },
 { {   0,   9, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },    288.79742,   0.00006,     0.050 },
 { {   0,   9, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },    268.36767,   0.00002,     0.154 },
 { {   0,   9,  -9,   0,  -1,   0,   0,   0,  -2,   0,   0 },    273.30738,   0.00005,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.51876,   0.00003,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.50129,   0.00039,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    270.51964,   0.00023,     0.171 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.79877,   0.00011,     0.135 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     89.97257,   0.00005,     0.148 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.24008,   0.00060,     0.131 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     90.07247,   0.00262,     0.178 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },     90.11259,   0.00137,     0.053 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },     90.80706,   0.00011,     0.031 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    270.11740,   0.00016,     0.056 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    101.19200,   0.00004,     0.032 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.01150,   0.00028,     0.259 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.99473,   0.00614,     0.058 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.00590,   0.00007,     0.284 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    270.00988,   0.00649,     0.033 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00448,   0.00041,     0.023 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },    269.98227,   0.00004,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.04864,   0.00005,     0.034 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   3,   0,   0 },    270.04861,   0.00017,     0.023 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.00259,   0.00007,     0.035 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.99623,   0.00003,     0.024 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },    270.00107,   0.00010,     0.018 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   4,   1,   0 },     90.00652,   0.00002,     0.015 },
 { {   0,   9,  -9,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.44744,   0.00004,     0.053 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,   0,  -2,   0 },    270.96308,   0.00002,     0.050 },
 { {   0,   9,  -8,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.36800,   0.00008,     0.153 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.24213,   0.00021,     0.052 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.20695,   0.00004,     0.130 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -1,  -1 },    193.29582,   0.00004,     0.130 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.29537,   0.00005,     0.052 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.30498,   0.00018,     0.179 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },     13.29469,   0.00004,     0.031 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   1,  -1 },     13.29582,   0.00004,     0.053 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.35712,   0.00034,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    356.29002,   0.00039,     0.032 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },    358.08804,   0.00004,     0.022 },
 { {   0,   9,  -7,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.88545,   0.00005,     0.134 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    115.92787,   0.00025,     0.058 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    102.69648,   0.00005,     0.178 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    296.27410,   0.00026,     0.131 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    282.69648,   0.00005,     0.048 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.84587,   0.00011,     0.052 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    118.84164,   0.00014,     0.031 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    218.00548,   0.00002,     0.127 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },     38.22370,   0.00003,     0.016 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.96843,   0.00002,     0.043 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    254.96843,   0.00002,     0.330 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },    327.88051,   0.00003,     0.073 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    147.88051,   0.00003,     0.079 },
 { {   0,  10, -15,   0,   1,   0,   0,   0,  -2,   1,   0 },    140.39233,   0.00003,     0.099 },
 { {   0,  10, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    331.62105,   0.00006,     0.044 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    144.90483,   0.00004,     0.044 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    324.90483,   0.00004,     0.248 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     65.56117,   0.00007,     0.073 },
 { {   0,  10, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    128.84074,   0.00003,     0.045 },
 { {   0,  10, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },    128.84074,   0.00003,     0.243 },
 { {   0,  10, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    143.63889,   0.00005,     0.046 },
 { {   0,  10, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    140.80797,   0.00002,     0.202 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    346.34575,   0.00014,     0.068 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    165.98845,   0.00015,     0.036 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     76.90387,   0.00003,     0.049 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    256.92745,   0.00003,     0.166 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    255.99592,   0.00016,     0.235 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    259.54870,   0.00009,     0.057 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.39900,   0.00052,     0.064 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,   0 },     82.62131,   0.00032,     0.034 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   2,  -2,   0 },    269.80460,   0.00003,     0.377 },
 { {   0,  10, -12,   0,   2,   0,   0,   0,   2,   0,   0 },     89.80460,   0.00003,     0.034 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.72348,   0.00003,     0.051 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    178.36822,   0.00003,     0.161 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    188.04601,   0.00002,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    165.51987,   0.00006,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },      0.99907,   0.00002,     0.125 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    181.53872,   0.00040,     0.190 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    182.21456,   0.00032,     0.054 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    345.55317,   0.00009,     0.238 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   1 },    345.55498,   0.00010,     0.032 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },     18.45556,   0.00002,     0.102 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },     11.13249,   0.00003,     0.287 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.57587,   0.00143,     0.060 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,   0 },      3.19670,   0.00149,     0.033 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   1,   0 },      2.36050,   0.00026,     0.023 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   3,   0,   0 },      2.43205,   0.00004,     0.024 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.41298,   0.00006,     0.024 },
 { {   0,  10, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },    305.60864,   0.00004,     0.052 },
 { {   0,  10, -11,   0,   1,   0,   0,   0,   2,   0,   0 },    268.67655,   0.00005,     0.033 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,  -2,   0,   0 },    271.88600,   0.00003,     0.054 },
 { {   0,  10, -10,   0,  -1,   0,   0,   0,  -2,   2,   0 },     91.88600,   0.00003,     0.126 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.21468,   0.00005,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.31134,   0.00056,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.84712,   0.00011,     0.125 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.99398,   0.00045,     0.143 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     89.94661,   0.00212,     0.160 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   1,   0 },     89.99450,   0.00098,     0.051 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   2,   0 },     91.60151,   0.00003,     0.031 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    270.11943,   0.00011,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   1 },    101.37994,   0.00004,     0.031 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    271.68827,   0.00002,     0.114 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.21532,   0.00018,     0.223 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.98767,   0.00420,     0.056 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,  -2 },     89.99392,   0.00006,     0.241 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    270.00251,   0.00504,     0.032 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00227,   0.00032,     0.023 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   2,   0 },    269.99442,   0.00006,     0.017 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.03794,   0.00005,     0.033 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   3,   0,   0 },    270.03791,   0.00014,     0.023 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.77488,   0.00005,     0.034 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,  -1,   0 },    270.00690,   0.00006,     0.024 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   0,   0 },    269.99969,   0.00008,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   4,   1,   0 },    269.99110,   0.00002,     0.014 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.54674,   0.00002,     0.054 },
 { {   0,  10, -10,   0,   1,   0,   0,   0,  -2,   2,   0 },    256.54674,   0.00002,     0.123 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,   0,  -2,   0 },    270.95286,   0.00002,     0.052 },
 { {   0,  10,  -9,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.35750,   0.00008,     0.139 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.21858,   0.00017,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.21228,   0.00005,     0.120 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.28560,   0.00018,     0.161 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },     13.28165,   0.00015,     0.030 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.31672,   0.00023,     0.053 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    357.15252,   0.00033,     0.031 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.85724,   0.00002,     0.054 },
 { {   0,  10,  -8,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.85724,   0.00002,     0.124 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    115.90197,   0.00023,     0.061 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    296.63199,   0.00023,     0.121 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.82103,   0.00009,     0.051 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    115.83612,   0.00012,     0.030 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     38.21488,   0.00003,     0.098 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    218.21488,   0.00003,     0.027 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    218.19741,   0.00002,     0.209 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     38.19741,   0.00002,     0.046 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    218.64926,   0.00002,     0.029 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },     38.64926,   0.00002,     0.016 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.17937,   0.00003,     0.044 },
 { {   0,  10,  -4,   0,  -2,   0,   0,   0,   2,  -2,   0 },     38.31190,   0.00002,     0.097 },
 { {   0,  10,  -4,   0,  -2,   0,   0,   0,   2,   0,   0 },    218.31190,   0.00002,     0.027 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    346.24244,   0.00008,     0.039 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    346.24244,   0.00003,     0.080 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0 },    166.24244,   0.00002,     0.075 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    346.24244,   0.00014,     0.070 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    346.24244,   0.00018,     0.082 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    346.24244,   0.00002,     0.039 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    346.24244,   0.00021,     0.095 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    346.24244,   0.00056,     0.042 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    346.24244,   0.00005,     0.027 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,  -1,   0 },     63.30551,   0.00006,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },     63.30551,   0.00014,     0.028 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },     63.30551,   0.00006,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -3,   0 },     63.30551,   0.00006,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },     63.30551,   0.00070,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },     63.30551,   0.00807,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     63.58019,   0.00502,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },     63.30551,   0.00002,     0.067 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -1,  -1,   0 },    243.30551,   0.00015,     0.081 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -4,   0 },     63.30551,   0.00010,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },     63.30551,   0.00156,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },     63.10623,   0.02854,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   2 },     63.30551,   0.00011,     0.074 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    248.46207,   0.00028,  1911.695 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    243.30551,   0.00011,     0.074 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    243.03414,   0.02873,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    243.30551,   0.00156,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    243.30551,   0.00010,     0.025 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },     63.30551,   0.00015,     0.081 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -4,   0 },    243.30551,   0.00002,     0.067 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    243.50470,   0.00505,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    243.30551,   0.00807,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    243.30551,   0.00070,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    243.30551,   0.00006,     0.020 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -3,   0 },    243.30551,   0.00006,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -2,   0 },    243.30551,   0.00014,     0.028 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    243.30551,   0.00006,     0.020 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    320.36858,   0.00005,     0.027 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    320.36858,   0.00056,     0.042 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    320.36858,   0.00021,     0.095 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,  -3,   0 },    320.36858,   0.00002,     0.039 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    320.36858,   0.00018,     0.082 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    320.36858,   0.00014,     0.070 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },    140.36858,   0.00002,     0.075 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    320.36858,   0.00003,     0.080 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    320.36858,   0.00008,     0.039 },
 { {   0,  10,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    217.43165,   0.00003,     0.044 },
 { {   0,  11, -17,   0,   2,   0,   0,   0,  -2,   0,   0 },    141.55062,   0.00003,     0.042 },
 { {   0,  11, -17,   0,   2,   0,   0,   0,  -2,   2,   0 },    321.55062,   0.00003,     0.354 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     23.37460,   0.00002,     0.044 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    203.37460,   0.00002,     0.274 },
 { {   0,  11, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    319.94597,   0.00002,     0.044 },
 { {   0,  11, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },    139.94597,   0.00002,     0.267 },
 { {   0,  11, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    331.61175,   0.00002,     0.046 },
 { {   0,  11, -15,   0,  -1,   0,   0,   0,  -2,   2,   0 },    331.61175,   0.00002,     0.219 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    146.30352,   0.00002,     0.046 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    326.30352,   0.00002,     0.215 },
 { {   0,  11, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    145.45174,   0.00003,     0.048 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    345.72015,   0.00009,     0.065 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    166.72559,   0.00014,     0.035 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     79.44623,   0.00004,     0.050 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    259.44623,   0.00004,     0.150 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    259.42494,   0.00015,     0.205 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    261.66717,   0.00006,     0.055 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.26130,   0.00037,     0.061 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,   0,   0 },     83.23731,   0.00031,     0.034 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },      3.57831,   0.00004,     0.053 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    185.48597,   0.00004,     0.179 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },      3.57335,   0.00002,     0.131 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    165.25884,   0.00005,     0.050 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },     26.15659,   0.00003,     0.136 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    181.43826,   0.00039,     0.170 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    185.28103,   0.00025,     0.052 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    345.29620,   0.00006,     0.207 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,   1 },    345.29859,   0.00008,     0.032 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.58350,   0.00101,     0.058 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,   0 },      3.66116,   0.00124,     0.033 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   1,   0 },      2.55278,   0.00022,     0.023 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.32186,   0.00005,     0.024 },
 { {   0,  11, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },    307.26958,   0.00004,     0.053 },
 { {   0,  11, -12,   0,   1,   0,   0,   0,   2,   0,   0 },    268.66162,   0.00004,     0.033 },
 { {   0,  11, -11,   0,  -1,   0,   0,   0,  -2,   0,   0 },    270.05004,   0.00005,     0.056 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.02157,   0.00002,     0.032 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.76873,   0.00022,     0.056 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     85.53310,   0.00003,     0.217 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.37837,   0.00007,     0.116 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    275.44113,   0.00003,     0.046 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },     93.36898,   0.00002,     0.051 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.98592,   0.00036,     0.157 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     89.95190,   0.00161,     0.145 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   1,   0 },     89.70180,   0.00074,     0.050 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   2,   0 },     89.96924,   0.00005,     0.030 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    269.98421,   0.00008,     0.052 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -3,   0 },    270.95419,   0.00002,     0.123 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.60990,   0.00013,     0.196 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.97284,   0.00285,     0.054 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,  -2 },     89.98107,   0.00005,     0.210 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    269.99574,   0.00390,     0.032 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00096,   0.00031,     0.022 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   2,   0 },    269.99157,   0.00003,     0.017 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   3,   0,   0 },    270.02476,   0.00012,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.51602,   0.00004,     0.034 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.99116,   0.00012,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   4,   0,   0 },    270.00259,   0.00006,     0.018 },
 { {   0,  11, -11,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.63312,   0.00002,     0.056 },
 { {   0,  11, -11,   0,   1,   0,   0,   0,  -2,   2,   0 },    256.63312,   0.00002,     0.114 },
 { {   0,  11, -10,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.39465,   0.00005,     0.128 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.18979,   0.00013,     0.056 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.19001,   0.00004,     0.112 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    214.29565,   0.00002,     0.029 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.26403,   0.00017,     0.146 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,   1 },     13.26520,   0.00012,     0.030 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.41067,   0.00015,     0.052 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    357.87896,   0.00027,     0.031 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   1,   0 },    180.56379,   0.00002,     0.022 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,   0,  -2,   0 },    269.83969,   0.00002,     0.056 },
 { {   0,  11,  -9,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.83969,   0.00002,     0.115 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    115.78297,   0.00020,     0.063 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    298.44974,   0.00020,     0.113 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.79942,   0.00007,     0.049 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    115.81271,   0.00010,     0.030 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     38.62174,   0.00002,     0.092 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    218.62174,   0.00002,     0.027 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    218.62174,   0.00003,     0.241 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     38.62174,   0.00003,     0.045 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    211.89966,   0.00002,     0.078 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },     31.89966,   0.00002,     0.073 },
 { {   0,  12, -18,   0,   2,   0,   0,   0,  -2,   0,   0 },    141.52944,   0.00006,     0.043 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -3,   0 },    147.87334,   0.00003,     0.080 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    327.87334,   0.00003,     0.071 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    144.04978,   0.00004,     0.047 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    324.04978,   0.00004,     0.189 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    346.72673,   0.00005,     0.063 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    167.35472,   0.00012,     0.034 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.39434,   0.00005,     0.052 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    243.47031,   0.00002,     0.129 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    257.57645,   0.00012,     0.182 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    277.40571,   0.00003,     0.053 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    263.95812,   0.00028,     0.059 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   0,   0 },     83.92927,   0.00026,     0.033 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    264.28856,   0.00002,     0.023 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.82784,   0.00004,     0.055 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    185.59241,   0.00003,     0.201 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },      1.88101,   0.00002,     0.121 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    164.96689,   0.00004,     0.052 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    181.50126,   0.00002,     0.050 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },     11.47595,   0.00006,     0.148 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    181.65137,   0.00034,     0.154 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    184.74561,   0.00024,     0.051 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    345.00854,   0.00005,     0.183 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,   1 },    345.01106,   0.00006,     0.031 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.61675,   0.00074,     0.056 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,   0 },      3.37858,   0.00099,     0.032 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   1,   0 },      2.39067,   0.00019,     0.022 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.40687,   0.00006,     0.023 },
 { {   0,  12, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },    308.84890,   0.00004,     0.055 },
 { {   0,  12, -12,   0,  -1,   0,   0,   0,  -2,   0,   0 },    268.19969,   0.00003,     0.058 },
 { {   0,  12, -12,   0,  -1,   0,   0,   0,  -2,   2,   0 },     88.19969,   0.00003,     0.109 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,  -2,   0 },     90.03288,   0.00002,     0.023 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.61685,   0.00026,     0.058 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     89.95656,   0.00012,     0.252 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.45506,   0.00007,     0.108 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },     89.99944,   0.00004,     0.044 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },     92.89323,   0.00002,     0.053 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.99826,   0.00016,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     89.95169,   0.00128,     0.133 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   1,   0 },     89.95888,   0.00057,     0.048 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   2,   0 },     89.98760,   0.00004,     0.029 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    269.97359,   0.00006,     0.050 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    269.54628,   0.00008,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.95333,   0.00196,     0.053 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    269.94938,   0.00299,     0.031 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   1,   0 },    269.95843,   0.00024,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   3,   0,   0 },    270.00919,   0.00010,     0.022 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.08103,   0.00003,     0.033 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.95404,   0.00011,     0.023 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   4,   0,   0 },    269.96103,   0.00005,     0.018 },
 { {   0,  12, -12,   0,   1,   0,   0,   0,  -2,   0,   0 },     77.64726,   0.00003,     0.058 },
 { {   0,  12, -12,   0,   1,   0,   0,   0,  -2,   2,   0 },    257.64726,   0.00003,     0.107 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,   0,  -2,   0 },     91.20533,   0.00002,     0.055 },
 { {   0,  12, -11,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.25666,   0.00007,     0.119 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,  -2,   1 },    193.16568,   0.00002,     0.059 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.16542,   0.00011,     0.058 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.16568,   0.00002,     0.105 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,   2,   0 },      3.91041,   0.00003,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.24477,   0.00011,     0.134 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,   1 },     13.24626,   0.00009,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.25594,   0.00013,     0.050 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    357.78186,   0.00023,     0.030 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,   1,   0 },    181.25967,   0.00002,     0.022 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,  -2,   0,   0 },     90.24735,   0.00002,     0.065 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,  -2,   2,   0 },    270.24735,   0.00002,     0.090 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,   0,  -2,   0 },    269.81983,   0.00002,     0.058 },
 { {   0,  12, -10,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.81983,   0.00002,     0.107 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    115.85757,   0.00020,     0.066 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    293.12541,   0.00017,     0.105 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.77982,   0.00006,     0.048 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    121.09891,   0.00008,     0.029 },
 { {   0,  12,  -9,   0,  -2,   0,   0,   0,  -2,   0,   0 },    176.02425,   0.00002,     0.069 },
 { {   0,  12,  -9,   0,  -2,   0,   0,   0,  -2,   2,   0 },    176.02425,   0.00002,     0.083 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    241.44532,   0.00003,     0.042 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   0,   0 },      2.38258,   0.00002,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,  -2,   0,  -2,   2,   0 },    182.38258,   0.00002,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    318.50839,   0.00004,     0.026 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    318.50839,   0.00050,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    318.38918,   0.00030,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },    318.15526,   0.00006,     0.038 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    318.15526,   0.00110,     0.076 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    318.99472,   0.00018,    40.749 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    138.61895,   0.00184,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    138.61895,   0.00010,     0.038 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    138.66677,   0.00043,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    138.50839,   0.00050,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    138.50839,   0.00004,     0.026 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    218.19132,   0.00003,     0.284 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     38.19132,   0.00003,     0.044 },
 { {   0,  12,  -7,  -2,   0,   0,   0,   0,  -2,   0,   0 },     18.13049,   0.00002,     0.075 },
 { {   0,  12,  -7,  -2,   0,   0,   0,   0,  -2,   2,   0 },    198.13049,   0.00002,     0.076 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -4,   1,   0 },    215.57146,   0.00003,     0.042 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    239.22796,   0.00004,     0.075 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   2,  -1 },     59.22796,   0.00004,     0.076 },
 { {   0,  12,  -6,   0,  -3,   0,   0,   0,   0,  -2,   0 },     98.30745,   0.00002,     0.075 },
 { {   0,  12,  -6,   0,  -3,   0,   0,   0,   0,   0,   0 },     98.30745,   0.00002,     0.075 },
 { {   0,  13, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    110.74787,   0.00003,     0.075 },
 { {   0,  13, -23,   0,   0,   0,   0,   0,  -2,   3,   0 },    290.74787,   0.00003,     0.076 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },     19.61015,   0.00004,     0.081 },
 { {   0,  13, -22,   0,   0,   0,   0,   0,  -2,   3,   0 },    199.61015,   0.00004,     0.071 },
 { {   0,  13, -19,   0,   2,   0,   0,   0,  -2,   0,   0 },    141.50629,   0.00003,     0.045 },
 { {   0,  13, -19,   0,   2,   0,   0,   0,  -2,   2,   0 },    141.50629,   0.00003,     0.245 },
 { {   0,  13, -18,   0,   1,   0,   0,   0,  -2,   0,   0 },    320.29745,   0.00004,     0.046 },
 { {   0,  13, -18,   0,   1,   0,   0,   0,  -2,   2,   0 },    140.29745,   0.00004,     0.200 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    144.03676,   0.00004,     0.049 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    324.03676,   0.00004,     0.169 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    168.02723,   0.00011,     0.033 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     81.31825,   0.00006,     0.054 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    243.30539,   0.00002,     0.126 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    258.28014,   0.00010,     0.163 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    278.16611,   0.00002,     0.052 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    264.58859,   0.00020,     0.057 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     84.52325,   0.00017,     0.032 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   2,   0 },     84.62519,   0.00003,     0.017 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },      3.14315,   0.00003,     0.057 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    185.67800,   0.00002,     0.230 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },      2.00874,   0.00002,     0.112 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },     16.55896,   0.00003,     0.163 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    181.69464,   0.00031,     0.140 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   1,   0 },    185.13621,   0.00017,     0.049 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    344.68816,   0.00005,     0.165 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,   1 },    344.69077,   0.00005,     0.030 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.24051,   0.00053,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,   0 },      3.37266,   0.00081,     0.031 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   1,   0 },      2.56889,   0.00015,     0.022 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.40433,   0.00005,     0.023 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   4,   0,   0 },      2.40061,   0.00006,     0.018 },
 { {   0,  13, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },    310.34918,   0.00003,     0.057 },
 { {   0,  13, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    266.34235,   0.00003,     0.060 },
 { {   0,  13, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },     86.34235,   0.00003,     0.102 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,  -2,   0 },     90.01877,   0.00002,     0.023 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.08221,   0.00040,     0.060 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },     89.92657,   0.00007,     0.299 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.76522,   0.00009,     0.101 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },     79.77560,   0.00003,     0.054 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.73923,   0.00013,     0.195 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     90.32297,   0.00102,     0.123 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   1,   0 },     89.96932,   0.00054,     0.047 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    269.96373,   0.00004,     0.049 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    270.56023,   0.00005,     0.157 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.04208,   0.00140,     0.051 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    269.91126,   0.00228,     0.030 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   1,   0 },    269.92914,   0.00020,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   3,   0,   0 },    269.99125,   0.00009,     0.022 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   4,  -2,   0 },     90.27398,   0.00002,     0.032 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.92660,   0.00007,     0.023 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   4,   0,   0 },    269.93372,   0.00004,     0.017 },
 { {   0,  13, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },     77.69253,   0.00003,     0.061 },
 { {   0,  13, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },    257.69253,   0.00003,     0.100 },
 { {   0,  13, -12,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.26145,   0.00004,     0.111 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    348.02223,   0.00003,     0.064 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    168.02223,   0.00003,     0.092 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.13910,   0.00010,     0.060 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    148.78549,   0.00002,     0.110 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.22348,   0.00010,     0.124 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,   1 },     13.22485,   0.00007,     0.029 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.63631,   0.00011,     0.049 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    358.29944,   0.00019,     0.030 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,  -2,   0,   0 },     90.23184,   0.00002,     0.068 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,  -2,   2,   0 },    270.23184,   0.00002,     0.085 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,   0,  -2,   0 },    269.79797,   0.00002,     0.060 },
 { {   0,  13, -11,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.79797,   0.00002,     0.100 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    116.72364,   0.00023,     0.068 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    112.32909,   0.00003,     0.084 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    292.84773,   0.00016,     0.099 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.76134,   0.00005,     0.046 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    115.78873,   0.00004,     0.029 },
 { {   0,  13, -10,   0,  -2,   0,   0,   0,  -2,   0,   0 },    346.53163,   0.00002,     0.073 },
 { {   0,  13, -10,   0,  -2,   0,   0,   0,  -2,   2,   0 },    346.53163,   0.00002,     0.079 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     15.89599,   0.00005,     0.073 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    195.89599,   0.00005,     0.077 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    129.41831,   0.00002,     0.107 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    309.41831,   0.00002,     0.028 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },     44.57238,   0.00005,     0.075 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    224.57238,   0.00005,     0.076 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -3,   0 },    159.90906,   0.00003,     0.075 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },    339.90906,   0.00003,     0.076 },
 { {   0,  14, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    320.27061,   0.00006,     0.048 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.79625,   0.00004,     0.050 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.79625,   0.00004,     0.153 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    168.66298,   0.00009,     0.033 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.76018,   0.00005,     0.056 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    265.76018,   0.00005,     0.117 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    100.22086,   0.00003,     0.154 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    258.93118,   0.00009,     0.148 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    280.22086,   0.00003,     0.050 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    265.14254,   0.00015,     0.055 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,   0,   0 },     85.05687,   0.00012,     0.032 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },      3.39121,   0.00003,     0.059 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },      1.99267,   0.00002,     0.105 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    181.31615,   0.00028,     0.129 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    185.30060,   0.00013,     0.048 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    344.33324,   0.00004,     0.149 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.18465,   0.00039,     0.052 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,   0 },      4.40870,   0.00066,     0.031 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   1,   0 },      2.71083,   0.00013,     0.022 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   4,  -1,   0 },      2.31339,   0.00002,     0.023 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   4,   0,   0 },      2.31339,   0.00007,     0.017 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    304.22987,   0.00002,     0.059 },
 { {   0,  14, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },    124.22987,   0.00002,     0.104 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    288.72126,   0.00003,     0.062 },
 { {   0,  14, -14,   0,  -1,   0,   0,   0,  -2,   2,   0 },    108.72126,   0.00003,     0.096 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.94565,   0.00034,     0.063 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },     89.90378,   0.00005,     0.367 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.75905,   0.00009,     0.095 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.97160,   0.00002,     0.056 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.29507,   0.00010,     0.222 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     90.64332,   0.00084,     0.114 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   1,   0 },     90.00377,   0.00040,     0.045 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    265.84290,   0.00004,     0.143 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.90786,   0.00097,     0.049 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    269.95269,   0.00177,     0.030 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   2,   1,   0 },    269.98582,   0.00009,     0.021 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   3,   0,   0 },    269.97097,   0.00007,     0.022 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.90513,   0.00005,     0.022 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   4,   0,   0 },    269.95173,   0.00003,     0.017 },
 { {   0,  14, -14,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.82704,   0.00002,     0.063 },
 { {   0,  14, -14,   0,   1,   0,   0,   0,  -2,   2,   0 },    256.82704,   0.00002,     0.094 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,   0,  -2,   0 },     91.24274,   0.00002,     0.059 },
 { {   0,  14, -13,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.24274,   0.00002,     0.103 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    356.37148,   0.00002,     0.067 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    176.37148,   0.00002,     0.087 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.11089,   0.00009,     0.062 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.19974,   0.00009,     0.115 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,   1 },     13.20101,   0.00006,     0.028 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    178.86557,   0.00008,     0.047 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   0,   0 },      1.44777,   0.00012,     0.029 },
 { {   0,  14, -12,   0,  -3,   0,   0,   0,  -2,   0,   0 },    100.44918,   0.00004,     0.070 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,  -2,   0,   0 },     90.28475,   0.00002,     0.071 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,  -2,   2,   0 },    270.28475,   0.00002,     0.081 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,   0,  -2,   0 },    269.77415,   0.00002,     0.063 },
 { {   0,  14, -12,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.77415,   0.00002,     0.094 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    117.78642,   0.00034,     0.072 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    104.12320,   0.00004,     0.080 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    295.94367,   0.00014,     0.093 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.74327,   0.00004,     0.045 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    115.73966,   0.00016,     0.028 },
 { {   0,  14, -11,   0,  -3,   0,   0,   0,  -2,   0,   0 },    198.63900,   0.00007,     0.076 },
 { {   0,  14, -11,   0,  -3,   0,   0,   0,  -2,   2,   0 },    238.35250,   0.00002,     0.075 },
 { {   0,  14, -11,   0,  -2,   0,   0,   0,  -2,   0,   0 },    161.86776,   0.00012,     0.076 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    195.10743,   0.00004,     0.077 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },     15.10743,   0.00004,     0.074 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    129.58528,   0.00003,     0.100 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    309.58528,   0.00003,     0.027 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },     19.29517,   0.00004,     0.073 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    204.19758,   0.00005,     0.078 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,  -2,   1,  -2 },     50.24046,   0.00003,     0.040 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   0,  -2 },     30.63675,   0.00010,     0.075 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   0,   2,  -2 },    228.80077,   0.00010,     0.075 },
 { {   0,  14,  -9,   0,  -2,   0,   0,   0,   2,   1,  -2 },    230.24046,   0.00003,     0.040 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    252.67673,   0.00002,     0.049 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },     72.67673,   0.00002,     0.162 },
 { {   0,  15, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    320.24202,   0.00006,     0.049 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.76873,   0.00003,     0.052 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.76873,   0.00003,     0.140 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    169.26357,   0.00008,     0.032 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     87.89375,   0.00003,     0.058 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    267.89375,   0.00003,     0.109 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   0,   2,   0 },     83.76933,   0.00002,     0.030 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    265.64088,   0.00012,     0.053 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,   0 },     85.55384,   0.00009,     0.031 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.03047,   0.00006,     0.061 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.11147,   0.00002,     0.098 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    181.53489,   0.00024,     0.119 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    182.30946,   0.00007,     0.046 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.30102,   0.00046,     0.050 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   0,   0 },      3.56744,   0.00049,     0.030 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   1,   0 },      2.59680,   0.00010,     0.022 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   4,   0,   0 },      2.30949,   0.00006,     0.017 },
 { {   0,  15, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },    305.40189,   0.00002,     0.061 },
 { {   0,  15, -16,   0,   1,   0,   0,   0,  -2,   2,   0 },    125.40189,   0.00002,     0.098 },
 { {   0,  15, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    291.18963,   0.00005,     0.065 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -6,   1,   2 },     89.60386,   0.00003,     0.040 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   0,   2 },     89.60386,   0.00011,     0.075 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   2,   2 },    269.60386,   0.00011,     0.075 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.97688,   0.00033,     0.065 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     90.90210,   0.00003,     0.476 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   2 },    269.60386,   0.00003,     0.040 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.76413,   0.00010,     0.090 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.73567,   0.00002,     0.033 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    273.68656,   0.00006,     0.258 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     90.68906,   0.00070,     0.107 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     91.06592,   0.00019,     0.044 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    272.69269,   0.00002,     0.131 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.84521,   0.00072,     0.048 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    269.93155,   0.00131,     0.029 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   2,   1,   0 },    269.92872,   0.00010,     0.021 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   3,   0,   0 },    269.94839,   0.00006,     0.022 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.91737,   0.00004,     0.022 },
 { {   0,  15, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.87338,   0.00004,     0.065 },
 { {   0,  15, -14,   0,  -2,   0,   0,   0,  -2,   0,   0 },    184.75609,   0.00005,     0.069 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,   0,  -2,   0 },     91.22032,   0.00002,     0.062 },
 { {   0,  15, -14,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.22032,   0.00002,     0.097 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   0,   1 },     13.38829,   0.00004,     0.075 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   2,   1 },    193.38829,   0.00004,     0.076 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.08090,   0.00008,     0.065 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   0,   1,   0 },     39.26008,   0.00002,     0.042 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.17362,   0.00008,     0.107 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,   1 },     13.17480,   0.00004,     0.028 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.39588,   0.00006,     0.046 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,   0,   0 },      3.42666,   0.00006,     0.028 },
 { {   0,  15, -13,   0,  -3,   0,   0,   0,  -2,   0,   0 },    143.02265,   0.00004,     0.073 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,  -2,   0,   0 },     89.17801,   0.00003,     0.074 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,  -2,   2,   0 },    269.17801,   0.00003,     0.077 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,   0,  -2,   0 },     89.74833,   0.00002,     0.065 },
 { {   0,  15, -13,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.74833,   0.00002,     0.089 },
 { {   0,  15, -13,   0,  -1,   0,   0,   0,  -2,   0,   0 },    233.48736,   0.00004,     0.074 },
 { {   0,  15, -13,   0,  -1,   0,   0,   0,  -2,   2,   0 },    322.62167,   0.00003,     0.076 },
 { {   0,  15, -13,   0,   0,  -2,   0,   0,  -2,   0,   0 },     90.10078,   0.00002,     0.075 },
 { {   0,  15, -13,   0,   0,  -2,   0,   0,  -2,   2,   0 },    270.10078,   0.00002,     0.076 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   0,   0 },    116.32315,   0.00003,     0.026 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   1,   0 },    116.31840,   0.00038,     0.040 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    116.39647,   0.00023,     0.086 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    116.24191,   0.00019,     0.038 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    116.23870,   0.00338,     0.075 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    104.78660,   0.00010,    10.112 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    296.24571,   0.00124,     0.076 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    296.24571,   0.00007,     0.038 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    296.19356,   0.00005,     0.534 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    298.92560,   0.00004,     0.088 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    296.32850,   0.00034,     0.041 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    296.32315,   0.00003,     0.026 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    115.72193,   0.00013,     0.028 },
 { {   0,  15, -13,   0,   1,   0,   0,   0,  -2,   0,   0 },     24.76390,   0.00003,     0.075 },
 { {   0,  15, -13,   0,   1,   0,   0,   0,  -2,   2,   0 },     24.76390,   0.00003,     0.076 },
 { {   0,  15, -12,  -4,   0,   0,   0,   0,  -2,   0,   0 },    240.02814,   0.00002,     0.069 },
 { {   0,  15, -12,  -4,   0,   0,   0,   0,  -2,   2,   0 },     60.02814,   0.00002,     0.083 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -4,   1,   0 },     13.38622,   0.00002,     0.042 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   0,   0 },    189.70101,   0.00021,     0.040 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -3,   1,   0 },    189.70101,   0.00013,     0.087 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -2,   0 },    189.70101,   0.00004,     0.038 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,  -1,   0 },    189.70101,   0.00074,     0.076 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },     39.31611,   0.00011,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },      9.70101,   0.00074,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   2,  -1 },    219.25129,   0.00003,     0.076 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   2,   0 },      9.70101,   0.00004,     0.038 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,  -1,   0 },      9.70101,   0.00013,     0.087 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   1,   0,   0 },      9.70101,   0.00021,     0.040 },
 { {   0,  15, -11,   0,  -1,   0,   0,   0,   0,  -2,   0 },    234.73659,   0.00003,     0.076 },
 { {   0,  15, -11,   0,  -1,   0,   0,   0,   0,   0,   0 },    234.73659,   0.00003,     0.075 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    321.69845,   0.00004,     0.076 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    320.59228,   0.00003,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,  -2,  -1,   0 },    297.17736,   0.00009,     0.040 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,  -2,   0,   0 },    306.91471,   0.00008,     0.087 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,  -2,   0 },    276.80597,   0.00021,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   0,   0,   0 },     96.80597,   0.00021,     0.075 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   2,  -2,   0 },    126.91471,   0.00008,     0.087 },
 { {   0,  15,  -9,  -4,   0,   0,   0,   0,   2,  -1,   0 },    117.17736,   0.00009,     0.040 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    278.35933,   0.00003,     0.051 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },     98.35933,   0.00003,     0.147 },
 { {   0,  16, -21,   0,   1,   0,   0,   0,  -2,   0,   0 },    320.21169,   0.00005,     0.051 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.74478,   0.00003,     0.053 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.74478,   0.00003,     0.128 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    169.81108,   0.00005,     0.032 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.05079,   0.00005,     0.060 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   0,   2,   0 },     84.33492,   0.00002,     0.029 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    266.09090,   0.00009,     0.051 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,   0,   0 },     86.00275,   0.00007,     0.031 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.93722,   0.00007,     0.064 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    182.32186,   0.00023,     0.111 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,   1,   0 },    182.30386,   0.00005,     0.045 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.15735,   0.00031,     0.049 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   0,   0 },      1.31553,   0.00038,     0.030 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   1,   0 },      1.82220,   0.00005,     0.021 },
 { {   0,  16, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },    295.01426,   0.00004,     0.068 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.92977,   0.00037,     0.068 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },     89.85330,   0.00002,     0.679 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.27998,   0.00007,     0.085 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.27246,   0.00005,     0.061 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.83804,   0.00002,     0.308 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     90.93041,   0.00056,     0.100 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,   1,   0 },     89.85040,   0.00013,     0.043 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.85172,   0.00051,     0.047 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    269.83503,   0.00097,     0.029 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   2,   1,   0 },    269.85433,   0.00023,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   3,   0,   0 },    269.92357,   0.00005,     0.021 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.84804,   0.00007,     0.022 },
 { {   0,  16, -16,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.91202,   0.00005,     0.068 },
 { {   0,  16, -15,  -2,   0,   0,   0,   0,  -2,   0,   0 },      2.70001,   0.00002,     0.068 },
 { {   0,  16, -15,  -2,   0,   0,   0,   0,  -2,   2,   0 },    182.70001,   0.00002,     0.085 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,  -2,   0,   0 },    268.30474,   0.00003,     0.072 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,  -2,   2,   0 },     88.30474,   0.00003,     0.079 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,   0,  -2,   0 },     91.19577,   0.00002,     0.064 },
 { {   0,  16, -15,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.19577,   0.00002,     0.092 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    234.96929,   0.00004,     0.073 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    165.98703,   0.00003,     0.078 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.04922,   0.00009,     0.068 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   0,   0,   0 },      3.73326,   0.00004,     0.091 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.14515,   0.00008,     0.100 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },    179.85732,   0.00002,     0.249 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    179.85732,   0.00002,     0.044 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,   0,   0 },      3.73113,   0.00004,     0.028 },
 { {   0,  16, -15,   0,   1,   0,   0,   0,  -2,   0,   0 },    136.29914,   0.00002,     0.073 },
 { {   0,  16, -15,   0,   1,   0,   0,   0,  -2,   2,   0 },    136.29914,   0.00002,     0.078 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -2,   0,   0 },    175.33679,   0.00005,     0.075 },
 { {   0,  16, -14,  -1,   0,   0,   0,   0,  -2,   2,   0 },    355.33679,   0.00005,     0.075 },
 { {   0,  16, -14,   0,  -3,   0,   0,   0,  -2,   0,   0 },    317.51174,   0.00002,     0.077 },
 { {   0,  16, -14,   0,  -3,   0,   0,   0,  -2,   2,   0 },    137.51174,   0.00002,     0.074 },
 { {   0,  16, -14,   0,  -2,   0,   0,   0,   0,   0,   0 },    269.69778,   0.00005,     0.084 },
 { {   0,  16, -14,   0,  -1,   0,   0,   0,  -2,   0,   0 },    172.31878,   0.00002,     0.078 },
 { {   0,  16, -14,   0,  -1,   0,   0,   0,  -2,   2,   0 },    352.31878,   0.00002,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    297.97938,   0.00026,     0.079 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    294.30146,   0.00009,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.71961,   0.00002,     0.069 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    295.75035,   0.00026,     0.083 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    115.70363,   0.00010,     0.027 },
 { {   0,  16, -13,  -3,   0,   0,   0,   0,  -2,   0,   0 },    281.20347,   0.00003,     0.075 },
 { {   0,  16, -13,  -3,   0,   0,   0,   0,  -2,   2,   0 },    101.20347,   0.00003,     0.076 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    311.43881,   0.00002,     0.074 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    145.77825,   0.00006,     0.077 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     17.05832,   0.00003,     0.052 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    197.05832,   0.00003,     0.135 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   0,   0 },    137.78167,   0.00004,     0.075 },
 { {   0,  17, -22,  11,   0,   0,   0,   0,  -2,   2,   0 },    317.78167,   0.00004,     0.075 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.25599,   0.00003,     0.062 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.25599,   0.00003,     0.096 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     86.40968,   0.00006,     0.030 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.89034,   0.00007,     0.066 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.19165,   0.00002,     0.088 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    182.29488,   0.00002,     0.276 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    182.20487,   0.00018,     0.104 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    182.29870,   0.00002,     0.044 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.29094,   0.00024,     0.047 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   0,   0 },      2.29668,   0.00034,     0.029 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   1,   0 },      2.26197,   0.00004,     0.021 },
 { {   0,  17, -17,   0,  -1,   0,   0,   0,  -2,   0,   0 },    289.69564,   0.00006,     0.070 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.75281,   0.00054,     0.071 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.40853,   0.00003,     0.081 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },     90.37198,   0.00004,     0.063 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.82034,   0.00002,     0.381 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     91.26352,   0.00044,     0.094 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   1,   0 },     89.82190,   0.00012,     0.042 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.00802,   0.00004,     0.113 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.82215,   0.00038,     0.045 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    269.82178,   0.00075,     0.028 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   2,   1,   0 },    269.82256,   0.00017,     0.021 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   3,  -2,   0 },    269.89654,   0.00002,     0.047 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   3,   0,   0 },    269.89654,   0.00002,     0.021 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   4,  -2,   0 },     89.83113,   0.00002,     0.030 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   4,  -1,   0 },    269.82162,   0.00005,     0.021 },
 { {   0,  17, -17,   0,   1,   0,   0,   0,  -2,   0,   0 },     76.94364,   0.00004,     0.071 },
 { {   0,  17, -17,   0,   1,   0,   0,   0,  -2,   2,   0 },    256.94364,   0.00004,     0.080 },
 { {   0,  17, -16,   0,  -2,   0,   0,   0,  -2,   1,   0 },    205.59864,   0.00021,    60.783 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   0,   0 },    153.32530,   0.00002,     0.076 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   1,   0 },     91.15668,   0.00019,    14.738 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,  -2,   2,   0 },    153.32530,   0.00002,     0.075 },
 { {   0,  17, -16,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.27318,   0.00005,     0.087 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     12.38691,   0.00007,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,  -1 },     13.04847,   0.00004,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },      5.74994,   0.00011,     0.075 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   3,  -1 },    193.04847,   0.00004,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },    193.04131,   0.00015,     0.071 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.01642,   0.00006,     0.079 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.11439,   0.00007,     0.095 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   2,   0,   0 },      3.98824,   0.00019,     0.027 },
 { {   0,  17, -15,  -3,   0,   0,   0,   0,  -2,   0,   0 },     58.79301,   0.00002,     0.073 },
 { {   0,  17, -15,  -3,   0,   0,   0,   0,  -2,   2,   0 },    238.79301,   0.00002,     0.078 },
 { {   0,  17, -15,  -2,   0,   0,   0,   0,  -2,   0,   0 },     92.88709,   0.00003,     0.076 },
 { {   0,  17, -15,  -2,   0,   0,   0,   0,  -2,   2,   0 },    272.88709,   0.00003,     0.075 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    294.56241,   0.00013,     0.083 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,  -1,   0 },    296.35980,   0.00002,     0.077 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,   0 },    116.35980,   0.00002,     0.074 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },    120.35201,   0.00005,     0.072 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    297.79947,   0.00038,     0.079 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    115.68445,   0.00008,     0.027 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -2,   0,   0 },     60.36572,   0.00005,     0.076 },
 { {   0,  17, -14,  -4,   0,   0,   0,   0,  -2,   2,   0 },    240.36572,   0.00005,     0.075 },
 { {   0,  17, -14,   0,  -3,   0,   0,   0,   0,  -2,   0 },    267.53133,   0.00002,     0.076 },
 { {   0,  17, -14,   0,  -3,   0,   0,   0,   0,   0,   0 },     87.53133,   0.00002,     0.075 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    325.92588,   0.00004,     0.054 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    145.92588,   0.00004,     0.124 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.79889,   0.00004,     0.065 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.79889,   0.00004,     0.091 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     86.77974,   0.00005,     0.029 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -5,   0,   2 },      0.03812,   0.00002,     0.040 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,  -1,   2 },      0.03812,   0.00008,     0.075 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   1,   2 },    180.03812,   0.00008,     0.075 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },      1.78864,   0.00007,     0.069 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },      1.78864,   0.00007,     0.083 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,   0,   2 },    180.03812,   0.00002,     0.040 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    182.24239,   0.00016,     0.097 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },    189.20356,   0.00021,     0.046 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   0,   0 },      2.28785,   0.00026,     0.029 },
 { {   0,  18, -19,   0,   1,   0,   0,   0,  -2,   0,   0 },    299.72865,   0.00005,     0.069 },
 { {   0,  18, -19,   2,   0,   0,   0,   0,  -2,   0,   0 },    128.55896,   0.00002,     0.075 },
 { {   0,  18, -19,   2,   0,   0,   0,   0,  -2,   2,   0 },    308.55896,   0.00002,     0.076 },
 { {   0,  18, -19,   3,   0,   0,   0,   0,  -2,   0,   0 },    146.14540,   0.00002,     0.078 },
 { {   0,  18, -19,   3,   0,   0,   0,   0,  -2,   2,   0 },    326.14540,   0.00002,     0.073 },
 { {   0,  18, -18,   0,  -1,   0,   0,   0,  -2,   0,   0 },    322.47555,   0.00004,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   1,   0 },    269.88810,   0.00013,     0.040 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -4,   2,   0 },    270.29328,   0.00008,     0.085 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.10077,   0.00013,     0.037 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    268.34739,   0.00216,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     90.26071,   0.00010,     4.498 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.54546,   0.00051,     0.077 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   3,   0 },     89.54531,   0.00003,     0.038 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    121.94987,   0.00006,     0.066 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.13594,   0.00005,     0.501 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    101.19236,   0.00035,     0.089 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   1,   0 },     89.84790,   0.00023,     0.041 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -2,   0 },    141.52400,   0.00017,     0.105 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    128.25168,   0.00087,     0.044 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    263.71709,   0.00051,     0.028 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   2,   1,   0 },    269.79007,   0.00013,     0.020 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,  -2,   0 },    142.58377,   0.00003,     0.029 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   4,   0,   0 },    269.78889,   0.00007,     0.016 },
 { {   0,  18, -18,   0,   2,   0,   0,   0,  -2,   0,   0 },     99.68971,   0.00002,     0.075 },
 { {   0,  18, -18,   0,   2,   0,   0,   0,  -2,   2,   0 },     99.68971,   0.00002,     0.076 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,   0,   0 },    199.59090,   0.00003,     0.074 },
 { {   0,  18, -17,  -2,   0,   0,   0,   0,  -2,   2,   0 },     19.59090,   0.00003,     0.077 },
 { {   0,  18, -17,   0,  -1,   0,   0,   0,   0,   0,   0 },     91.34807,   0.00006,     0.082 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,  -1,   0 },     39.47954,   0.00003,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -4,   0,   0 },     39.47954,   0.00006,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,  -1,   1 },     13.01956,   0.00008,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   0,   1 },     12.97681,   0.00010,     0.088 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -3,   1,  -1 },    192.94981,   0.00002,     0.040 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -3,   0 },     39.47954,   0.00003,     0.019 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -2,   0 },     39.47954,   0.00027,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },     39.47954,   0.00213,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     39.47684,   0.00088,     0.080 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   0 },    219.47954,   0.00006,     0.038 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   1 },     13.01935,   0.00028,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   0 },    220.42581,   0.00055,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    192.99765,   0.00044,     0.074 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,   0 },    196.16330,   0.00002,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },     12.94999,   0.00008,     0.077 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -4,   0 },     39.47954,   0.00004,     0.025 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -3,   0 },     39.47847,   0.00041,     0.036 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },     39.48016,   0.00365,     0.070 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     39.48253,   0.00423,     0.082 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   1,   0 },     39.48066,   0.00047,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   2,   0 },     39.47954,   0.00004,     0.026 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -2,   1 },    193.01975,   0.00005,     0.087 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,   1 },    193.01956,   0.00008,     0.040 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   1,  -1 },     12.94981,   0.00002,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },     39.47815,   0.00521,     0.095 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },     39.47959,   0.01411,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   0 },     36.06766,   0.00144,     0.027 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   1,   0 },     39.47954,   0.00012,     0.020 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -3,   0 },     39.47954,   0.00013,     0.046 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -2,   0 },     39.47954,   0.00041,     0.028 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,  -1,   0 },     39.47954,   0.00022,     0.021 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   4,   0,   0 },     39.47954,   0.00004,     0.016 },
 { {   0,  18, -16,  -3,   0,   0,   0,   0,  -2,   0,   0 },    239.78381,   0.00002,     0.076 },
 { {   0,  18, -16,  -3,   0,   0,   0,   0,  -2,   2,   0 },     59.78381,   0.00002,     0.074 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,  -2,   0,   0 },    269.96849,   0.00003,     0.086 },
 { {   0,  18, -16,   0,  -2,   0,   0,   0,  -2,   2,   0 },     89.96849,   0.00003,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   0,   0 },    116.54261,   0.00005,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   1,   0 },    116.54261,   0.00006,     0.021 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -6,   2,   0 },    116.54261,   0.00002,     0.029 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -3,   0 },    116.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -2,   0 },    116.54261,   0.00025,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   0 },    116.54261,   0.00161,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   2 },    116.54261,   0.00004,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   0 },    116.54163,   0.00368,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },    116.53993,   0.00158,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,  -1,   0 },    116.54261,   0.00015,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   0,   0 },    296.54261,   0.00023,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -3,   1,   0 },    296.54261,   0.00003,     0.094 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -4,   0 },    116.54261,   0.00012,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   0 },    116.54288,   0.00145,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   2 },    296.54261,   0.00002,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    116.54276,   0.01777,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   2 },    296.54261,   0.00014,     0.090 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    116.54222,   0.20473,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -2 },    296.54261,   0.00006,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    116.53978,   0.13477,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   2 },    116.54261,   0.00030,     0.065 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,  -2 },    296.54261,   0.00005,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    296.49083,   0.00093,     0.563 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   2 },    116.54261,   0.00003,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.47764,   0.00044,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   3,   0 },    116.54261,   0.00005,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -3,   0 },    296.54261,   0.00004,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -2,   0 },    296.54261,   0.00044,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },    296.54261,   0.00376,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   2 },    296.54261,   0.00003,     0.069 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },    296.54261,   0.00006,     0.071 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -5,   0 },    116.54261,   0.00015,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,   0 },    116.54324,   0.00242,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   0 },    116.54324,   0.03968,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    116.54311,   0.72783,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   2 },    116.54234,   0.00267,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -2 },    116.54261,   0.00022,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    296.00461,   0.00724,   273.043 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   2 },    296.54261,   0.00022,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -2 },    296.54023,   0.00284,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    296.54574,   0.68256,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    296.54559,   0.03722,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   0 },    296.54559,   0.00227,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   3,   0 },    296.54261,   0.00015,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -3,   0 },    116.54261,   0.00006,     0.071 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,  -2 },    116.54261,   0.00003,     0.069 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   0 },    116.54261,   0.00376,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    116.54261,   0.00044,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   1,   0 },    116.54261,   0.00004,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -5,   0 },    296.54261,   0.00005,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -4,   0 },    296.54449,   0.00051,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,  -2 },    296.54261,   0.00003,     0.035 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,   0 },    296.62465,   0.00013,     0.565 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -3,   2 },    116.54261,   0.00005,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,  -2 },    296.54261,   0.00030,     0.065 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },    296.54238,   0.12671,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   2 },    116.54261,   0.00006,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    296.54299,   0.20367,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,  -2 },    116.54261,   0.00014,     0.090 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    296.54709,   0.01727,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,  -2 },    116.54261,   0.00002,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   0 },    296.54411,   0.00140,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   2,   0 },    296.54261,   0.00012,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -3,   0 },    116.54261,   0.00003,     0.094 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -2,   0 },    116.54261,   0.00023,     0.042 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   3,  -1,   0 },    296.54261,   0.00015,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   0 },    296.54196,   0.00148,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   0 },    296.54244,   0.00359,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,  -2 },    296.54261,   0.00004,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },    296.54261,   0.00161,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   0,   0 },    296.54261,   0.00025,     0.016 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,   1,   0 },    296.54261,   0.00003,     0.013 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -4,   0 },    296.54261,   0.00002,     0.029 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -3,   0 },    296.54261,   0.00006,     0.021 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   6,  -2,   0 },    296.54261,   0.00005,     0.016 },
 { {   0,  18, -15,  -4,   0,   0,   0,   0,  -2,   0,   0 },     59.93569,   0.00002,     0.079 },
 { {   0,  18, -15,  -4,   0,   0,   0,   0,  -2,   2,   0 },    239.93569,   0.00002,     0.072 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -2,   0 },     13.60568,   0.00004,     0.016 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,  -1,   0 },     13.60568,   0.00022,     0.021 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   0,   0 },     13.60568,   0.00041,     0.028 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -4,   1,   0 },     13.60568,   0.00013,     0.046 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -3,   0 },     13.60568,   0.00012,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   0 },     13.60528,   0.00135,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },     13.60562,   0.01416,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     13.60300,   0.00554,     0.095 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    193.55391,   0.00004,     0.360 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -1,  -1,  -1 },     40.09123,   0.00005,     0.040 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },     40.09123,   0.00003,     0.087 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -4,   0 },     13.60568,   0.00004,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -3,   0 },     13.60462,   0.00050,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },     13.60631,   0.00451,     0.082 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     13.60868,   0.00342,     0.070 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     13.60681,   0.00038,     0.036 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   2,   0 },     13.60568,   0.00004,     0.025 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -2,  -1 },     40.09098,   0.00017,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,   0 },    193.60568,   0.00058,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    220.11089,   0.00029,     0.076 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    193.60568,   0.00006,     0.038 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },     13.60505,   0.00083,     0.080 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },     13.60568,   0.00213,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     13.60568,   0.00027,     0.026 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   1,   0 },     13.60568,   0.00003,     0.019 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -2,  -1 },    220.09123,   0.00003,     0.087 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   3,  -1,  -1 },    220.09123,   0.00005,     0.041 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -2,   0 },     13.60568,   0.00006,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   4,  -1,   0 },     13.60568,   0.00003,     0.020 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -4,   0,   0 },    270.66875,   0.00003,     0.029 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -2,   0 },    270.66875,   0.00007,     0.028 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    270.66875,   0.00068,     0.044 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.66875,   0.00017,     0.106 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    270.66875,   0.00007,     0.089 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    270.66875,   0.00004,     0.066 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    296.93915,   0.00007,     0.074 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    218.07255,   0.00004,     0.025 },
 { {   0,  18, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    167.73182,   0.00003,     0.046 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.51153,   0.00005,     0.056 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.51153,   0.00005,     0.115 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -3,  -1,   1 },     43.46714,   0.00005,     0.040 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -3,   0,   1 },     43.46714,   0.00003,     0.087 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,  -2,   1 },     43.46714,   0.00019,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,  -1,   0,   1 },    223.46714,   0.00019,     0.075 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,   1,  -2,   1 },    223.46714,   0.00003,     0.087 },
 { {   0,  19, -24,  10,   0,   0,   0,   0,   1,  -1,   1 },    223.46714,   0.00005,     0.040 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     96.70497,   0.00003,     0.067 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    276.70497,   0.00003,     0.086 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    261.37040,   0.00004,     0.101 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   2,   0,   0 },     87.25431,   0.00017,     0.029 },
 { {   0,  19, -21,   2,   0,   0,   0,   0,  -2,   0,   0 },    269.38441,   0.00003,     0.073 },
 { {   0,  19, -21,   2,   0,   0,   0,   0,  -2,   2,   0 },     89.38441,   0.00003,     0.079 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   1,   0 },    104.98714,   0.00011,     0.040 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -4,   2,   0 },    104.98714,   0.00007,     0.087 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,  -1,   0 },    104.98714,   0.00002,     0.038 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   0,   0 },    105.30223,   0.00045,     0.075 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   1,   0 },     59.71256,   0.00049,   573.318 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   2,   0 },    285.30223,   0.00045,     0.075 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,  -2,   3,   0 },    284.98714,   0.00002,     0.038 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   0,   0 },    284.98714,   0.00007,     0.087 },
 { {   0,  19, -21,   3,   0,   0,   0,   0,   0,   1,   0 },    284.98714,   0.00011,     0.040 },
 { {   0,  19, -20,   0,  -1,   0,   0,   0,  -2,   0,   0 },     83.12753,   0.00002,     0.072 },
 { {   0,  19, -20,   0,  -1,   0,   0,   0,  -2,   2,   0 },    263.12753,   0.00002,     0.080 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -4,   2,   0 },      1.81210,   0.00002,     0.083 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.05015,   0.00019,     0.072 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.12723,   0.00003,     0.079 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    185.45699,   0.00016,     0.092 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.27747,   0.00011,     0.045 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,   0 },      2.27749,   0.00020,     0.028 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   1,   0 },    182.27745,   0.00002,     0.020 },
 { {   0,  19, -20,   0,   1,   0,   0,   0,  -2,   0,   0 },    221.06144,   0.00007,     0.073 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -4,   1,   0 },    270.12040,   0.00003,     0.041 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.67722,   0.00065,     0.078 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    269.39214,   0.00004,     2.480 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.18904,   0.00012,     0.073 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,   0 },     90.77308,   0.00070,     0.084 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   1,   0 },     89.88177,   0.00008,     0.040 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.75342,   0.00034,     0.043 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    269.75277,   0.00041,     0.027 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   2,   1,   0 },    269.75341,   0.00010,     0.020 },
 { {   0,  19, -18,  -3,   0,   0,   0,   0,  -2,   0,   0 },    193.51920,   0.00002,     0.074 },
 { {   0,  19, -18,  -3,   0,   0,   0,   0,  -2,   2,   0 },     13.51920,   0.00002,     0.077 },
 { {   0,  19, -18,  -2,   0,   0,   0,   0,  -2,   0,   0 },     59.18483,   0.00002,     0.077 },
 { {   0,  19, -18,  -2,   0,   0,   0,   0,  -2,   2,   0 },    239.18483,   0.00002,     0.074 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   0,  -1 },     12.97331,   0.00018,     0.077 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   2,  -1 },     12.97192,   0.00003,     0.074 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.02364,   0.00012,     0.085 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   2,   0,   0 },      4.37450,   0.00012,     0.027 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    117.67174,   0.00003,     0.092 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    117.78998,   0.00018,     0.072 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    115.64259,   0.00005,     0.026 },
 { {   0,  19, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },     38.04590,   0.00003,     0.071 },
 { {   0,  19, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    218.04590,   0.00003,     0.025 },
 { {   0,  20, -29,  15,   0,   0,   0,   0,  -2,   0,   0 },    105.07727,   0.00002,     0.075 },
 { {   0,  20, -29,  15,   0,   0,   0,   0,  -2,   2,   0 },    285.07727,   0.00002,     0.075 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.64000,   0.00004,     0.058 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.64000,   0.00004,     0.108 },
 { {   0,  20, -25,   0,   1,   0,   0,   0,  -2,   0,   0 },    140.11869,   0.00003,     0.058 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    273.26790,   0.00008,     0.095 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   1,   0 },    284.73920,   0.00011,     0.042 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   2,   0,   0 },     87.56015,   0.00014,     0.028 },
 { {   0,  20, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },    359.80161,   0.00002,     0.071 },
 { {   0,  20, -22,   0,   1,   0,   0,   0,  -2,   2,   0 },    359.80161,   0.00002,     0.081 },
 { {   0,  20, -21,   0,  -1,   0,   0,   0,  -2,   0,   0 },     29.92733,   0.00002,     0.075 },
 { {   0,  20, -21,   0,  -1,   0,   0,   0,  -2,   2,   0 },    209.92733,   0.00002,     0.076 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -6,   2,   0 },      1.80227,   0.00003,     0.028 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   0,   0 },      1.80227,   0.00013,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   1,   0 },      1.80295,   0.00150,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   2,   0 },      1.50349,   0.00092,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -3,   1,   0 },    181.80227,   0.00003,     0.081 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,  -1,   0 },      2.36161,   0.00016,     0.038 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.36161,   0.00296,     0.076 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   2 },      1.80227,   0.00002,     0.073 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    181.76360,   0.00026,    34.812 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,  -2 },    181.80227,   0.00002,     0.074 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    182.07980,   0.00560,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   3,   0 },    182.07980,   0.00031,     0.038 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   1,   0 },      1.80227,   0.00003,     0.081 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },      1.12364,   0.00006,     0.573 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    181.91315,   0.00156,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   1,   0 },    181.80161,   0.00155,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   2,   0 },    181.79986,   0.00013,     0.026 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   1,   1,  -1 },    206.96521,   0.00004,     0.042 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.26563,   0.00010,     0.043 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,   0,   0 },      2.47814,   0.00006,     0.028 },
 { {   0,  20, -21,   0,   1,   0,   0,   0,  -2,   0,   0 },     50.28619,   0.00002,     0.076 },
 { {   0,  20, -21,   0,   1,   0,   0,   0,  -2,   2,   0 },     50.28619,   0.00002,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -4,   1,   0 },    258.86534,   0.00011,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -4,   2,   0 },    248.52825,   0.00002,     0.096 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   0,  -1 },    284.02829,   0.00005,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   0,   0 },     90.00384,   0.00004,     0.041 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,  -1 },    284.02829,   0.00061,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   1,   0 },     90.00384,   0.00002,     0.088 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -3,   2,  -1 },    284.06008,   0.00038,     0.087 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     91.84021,   0.00027,     0.082 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    283.80698,   0.00017,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,  -1,   0 },     90.00384,   0.00013,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },    283.80698,   0.00314,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    103.36669,   0.00005,    39.938 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   0 },    270.00384,   0.00013,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,  -1 },    103.99725,   0.00219,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   3,  -1 },    103.99725,   0.00012,     0.038 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -3,   0 },     90.96311,   0.00002,     0.037 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    271.51477,   0.00003,     0.071 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.96311,   0.00004,     1.342 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,   0 },     91.31129,   0.00081,     0.080 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,  -1,   0 },    270.00384,   0.00002,     0.086 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },    104.60609,   0.00021,     0.087 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   0,   0 },    270.00384,   0.00004,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   1,  -1 },    104.02829,   0.00061,     0.040 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   1,   2,  -1 },    104.02829,   0.00005,     0.026 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.71581,   0.00021,     0.042 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   0,   0 },    269.71583,   0.00034,     0.027 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   2,   1,   0 },    269.71580,   0.00004,     0.020 },
 { {   0,  20, -19,  -3,   0,   0,   0,   0,  -2,   0,   0 },    213.99606,   0.00003,     0.078 },
 { {   0,  20, -19,  -3,   0,   0,   0,   0,  -2,   2,   0 },     33.99606,   0.00003,     0.073 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,  -2,   0 },    197.85560,   0.00006,     0.075 },
 { {   0,  20, -19,   0,  -3,   0,   0,   0,   0,   0,   0 },     71.21855,   0.00004,     0.075 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -3,   1,  -1 },    181.09136,   0.00004,     0.042 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   0,  -1 },     12.98935,   0.00003,     0.041 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,  -1,  -1 },     12.98935,   0.00010,     0.076 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   1,  -1 },    192.98935,   0.00010,     0.075 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.00875,   0.00012,     0.080 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,   0,  -1 },    192.98935,   0.00003,     0.040 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   2,   0,   0 },      2.98842,   0.00008,     0.026 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    110.96554,   0.00009,     0.069 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,  -1,  -2 },    296.45318,   0.00003,     0.076 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,   1,  -2 },    116.45318,   0.00003,     0.075 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    115.61960,   0.00004,     0.025 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,  -2,   0 },    190.59376,   0.00003,     0.075 },
 { {   0,  20, -15,  -8,   0,   0,   0,   0,   0,   0,   0 },     10.59376,   0.00003,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -3,   1,  -1 },     84.13759,   0.00006,     0.040 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -3,   2,  -1 },     84.13759,   0.00004,     0.087 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   0,  -1 },     84.13759,   0.00021,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,  -1,   2,  -1 },    264.13759,   0.00021,     0.075 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,   1,   0,  -1 },    264.13759,   0.00004,     0.087 },
 { {   0,  20, -12, -15,   0,   0,   0,   0,   1,   1,  -1 },    264.13759,   0.00006,     0.040 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.74721,   0.00004,     0.060 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.74721,   0.00004,     0.101 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -2,   2 },    256.67445,   0.00003,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },     65.36950,   0.00003,     0.074 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   0,   2 },     76.67445,   0.00003,     0.076 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     65.36950,   0.00003,     0.077 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    262.19191,   0.00004,     0.090 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   2,   0,   0 },     87.84066,   0.00011,     0.028 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,  -2,   2 },    105.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   0,   0,   1,   0,  -2,   0,   2 },    285.43642,   0.00004,     0.075 },
 { {   0,  21, -23,   0,   1,   0,   0,   0,  -2,   0,   0 },      3.66778,   0.00006,     0.074 },
 { {   0,  21, -23,   0,   1,   0,   0,   0,  -2,   2,   0 },    359.93339,   0.00002,     0.077 },
 { {   0,  21, -23,   0,   4,   0,   0,   0,  -2,   0,   0 },     70.82882,   0.00002,     0.075 },
 { {   0,  21, -23,   0,   4,   0,   0,   0,  -2,   2,   0 },     70.82882,   0.00002,     0.075 },
 { {   0,  21, -23,   2,   0,   0,   0,   0,  -2,   0,   0 },     66.46116,   0.00002,     0.080 },
 { {   0,  21, -23,   2,   0,   0,   0,   0,  -2,   2,   0 },    246.46116,   0.00002,     0.072 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    181.78587,   0.00006,     0.079 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    181.78587,   0.00006,     0.072 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -2,   1 },    340.55278,   0.00003,     0.075 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },    160.61512,   0.00003,     0.073 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,   1 },    160.57042,   0.00009,     0.076 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    184.75461,   0.00016,     0.082 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -3,   0 },    182.25031,   0.00002,     0.346 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -1,   0 },    186.04457,   0.00009,     0.042 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,   0,   0 },      2.25240,   0.00006,     0.027 },
 { {   0,  21, -22,   0,   1,   0,   0,   0,  -2,   0,   0 },    111.40410,   0.00004,     0.080 },
 { {   0,  21, -21,  -3,   0,   0,   0,   0,  -2,   0,   0 },    103.84813,   0.00003,     0.076 },
 { {   0,  21, -21,  -3,   0,   0,   0,   0,  -2,   2,   0 },    283.84813,   0.00003,     0.075 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,  -2,   0,   0 },    192.26006,   0.00002,     0.086 },
 { {   0,  21, -21,   0,  -1,   0,   0,   0,  -2,   2,   0 },     12.26006,   0.00002,     0.067 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -2,   0 },    269.94393,   0.00004,     0.026 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,  -1,   0 },    269.95842,   0.00043,     0.040 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     90.02524,   0.00035,     0.086 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    269.84304,   0.00008,     0.605 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.64813,   0.00005,     0.067 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -3,   0 },    270.12824,   0.00009,     0.038 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },    270.12824,   0.00158,     0.075 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.77958,   0.00055,     8.350 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   0 },     89.87628,   0.00495,     0.076 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   1,   0 },     89.87653,   0.00027,     0.038 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },     89.77335,   0.00030,     0.088 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.86589,   0.00067,     0.041 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   2,   0,   0 },    269.62472,   0.00021,     0.026 },
 { {   0,  21, -20,   0,  -1,   0,   0,   0,   0,  -2,   0 },    271.04329,   0.00002,     0.080 },
 { {   0,  21, -20,   0,  -1,   0,   0,   0,   0,   0,   0 },    271.04329,   0.00002,     0.071 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -2,  -1,   0 },    167.00700,   0.00003,     0.042 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    192.93263,   0.00004,     0.040 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },     13.86569,   0.00005,     0.086 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,  -1 },    192.91688,   0.00015,     0.074 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },     13.37697,   0.00058,     0.076 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   2,   0,   0 },      1.39455,   0.00006,     0.026 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   3,  -2,  -1 },     12.94780,   0.00003,     0.088 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   3,  -1,  -1 },     12.93263,   0.00004,     0.041 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    116.46642,   0.00004,     0.066 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     96.67194,   0.00004,     0.025 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   0,   0 },     66.34051,   0.00004,     0.075 },
 { {   0,  22, -28,   7,   0,   0,   0,   0,  -2,   2,   0 },    246.34051,   0.00004,     0.075 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.54734,   0.00003,     0.063 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.54734,   0.00003,     0.095 },
 { {   0,  22, -24,   0,  -2,   0,   0,   0,  -2,   0,   0 },    135.10365,   0.00005,     0.076 },
 { {   0,  22, -24,   0,  -2,   0,   0,   0,  -2,   2,   0 },    306.92034,   0.00002,     0.075 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.06757,   0.00003,     0.077 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    244.06757,   0.00003,     0.074 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    262.56528,   0.00004,     0.085 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   2,   0,   0 },     88.09852,   0.00009,     0.027 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    193.80658,   0.00004,     0.084 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     26.79422,   0.00002,     0.069 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    339.88668,   0.00004,     0.077 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    182.13689,   0.00020,     0.078 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   2,  -2,   0 },    182.05044,   0.00002,     0.091 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   2,   0,   0 },      2.23773,   0.00005,     0.027 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    258.16813,   0.00010,     0.091 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },     87.04646,   0.00004,     0.064 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,  -2,   0 },    269.70942,   0.00011,     0.078 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    271.91983,   0.00042,     0.073 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   1,   0,  -1 },    102.43394,   0.00005,     0.079 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.63439,   0.00009,     0.040 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   2,   0,   0 },    269.63441,   0.00018,     0.026 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   2,   1,   0 },     89.63436,   0.00002,     0.019 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.92880,   0.00010,     0.073 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },      1.58462,   0.00002,     0.077 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   2,   0,   0 },      1.58462,   0.00002,     0.025 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    326.64860,   0.00004,     0.065 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.64860,   0.00004,     0.090 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    141.93070,   0.00002,     0.070 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    321.93070,   0.00002,     0.082 },
 { {   0,  23, -27,   2,   0,   0,   0,   0,  -2,   0,   0 },    118.22856,   0.00004,     0.075 },
 { {   0,  23, -27,   2,   0,   0,   0,   0,  -2,   2,   0 },    298.22856,   0.00004,     0.076 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    155.71742,   0.00007,     0.075 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    355.28863,   0.00004,     0.076 },
 { {   0,  23, -26,   0,   1,   0,   0,   0,  -2,   0,   0 },      2.14684,   0.00007,     0.075 },
 { {   0,  23, -26,   0,   1,   0,   0,   0,  -2,   2,   0 },     75.26576,   0.00003,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   0,   0 },    258.18245,   0.00018,     0.040 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -3,   1,   0 },    258.18245,   0.00011,     0.087 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -2,   0 },    258.18245,   0.00003,     0.038 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -1,   0 },    258.18245,   0.00063,     0.076 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,   0 },     78.18245,   0.00063,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   2,   0 },     78.18245,   0.00003,     0.038 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    262.91651,   0.00006,     0.081 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,  -1,   0 },     78.18245,   0.00011,     0.087 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   1,   0,   0 },     78.18245,   0.00018,     0.040 },
 { {   0,  23, -24,   0,  -1,   0,   0,   0,   0,  -2,   0 },    295.65284,   0.00002,     0.076 },
 { {   0,  23, -24,   0,  -1,   0,   0,   0,   0,   0,   0 },     73.84038,   0.00006,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,  -1,   0 },      2.28110,   0.00009,     0.041 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   0,  -1 },    339.19500,   0.00004,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.75017,   0.00018,     0.088 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   1,  -1 },    339.19500,   0.00002,     0.087 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },      1.61369,   0.00033,     0.076 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,  -1 },    339.19500,   0.00013,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    182.64547,   0.00004,     7.618 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,   0 },      3.40918,   0.00042,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,  -1 },    159.19500,   0.00013,     0.076 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   1,   0 },      2.96327,   0.00004,     0.038 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -2,   0 },    182.99323,   0.00005,     0.086 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -1,  -1 },    159.19500,   0.00002,     0.087 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -1,   0 },    182.28110,   0.00009,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,   0,  -1 },    159.19500,   0.00004,     0.040 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,   0,   0 },      2.22171,   0.00004,     0.026 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    259.15905,   0.00003,     0.097 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     79.15905,   0.00003,     0.062 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    282.75345,   0.00006,     0.041 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    283.22725,   0.00009,     0.088 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    270.48879,   0.00016,     0.070 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   2,   0 },     89.58666,   0.00002,     0.024 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -2,  -1 },    282.75599,   0.00022,     0.076 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -1 },    286.15031,   0.00007,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.53261,   0.00007,     0.039 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   2,   0,   0 },    269.59075,   0.00006,     0.026 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   3,  -2,  -1 },    102.75142,   0.00004,     0.087 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   3,  -1,  -1 },    102.75345,   0.00006,     0.040 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   0,  -1,  -2 },      4.98030,   0.00004,     0.040 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   0,   0,  -2 },      4.98030,   0.00003,     0.087 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   1,   0,  -1 },    192.88414,   0.00004,     0.070 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -2,  -2 },      4.98214,   0.00015,     0.076 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,   0,  -2 },    186.69003,   0.00010,     0.075 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   4,  -2,  -2 },    184.98030,   0.00003,     0.087 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   4,  -1,  -2 },    184.98030,   0.00004,     0.040 },
 { {   0,  24, -29,   0,   1,   0,   0,   0,  -2,   0,   0 },    139.95990,   0.00003,     0.068 },
 { {   0,  24, -28,   0,   1,   0,   0,   0,  -2,   0,   0 },    128.24715,   0.00006,     0.073 },
 { {   0,  24, -27,   0,  -1,   0,   0,   0,  -2,   0,   0 },    324.92349,   0.00003,     0.078 },
 { {   0,  24, -26,  -3,   0,   0,   0,   0,  -2,   0,   0 },    299.67391,   0.00002,     0.075 },
 { {   0,  24, -26,  -3,   0,   0,   0,   0,  -2,   2,   0 },    119.67391,   0.00002,     0.076 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },     82.75728,   0.00004,     0.074 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    261.57166,   0.00015,     0.077 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,   0,   0 },      2.11520,   0.00012,     0.071 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,   0,  -1 },    338.28289,   0.00005,     0.077 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   2,   0,   0 },      2.20431,   0.00017,     0.026 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    269.41831,   0.00002,     0.103 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },     89.41831,   0.00002,     0.060 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -3,   0 },    269.43126,   0.00015,     0.040 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    269.68129,   0.00016,     0.067 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -5,   0 },    269.43126,   0.00003,     0.038 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -4,   0 },    269.42984,   0.00051,     0.075 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -2,   0 },     89.48359,   0.00097,     0.076 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.43126,   0.00003,     0.038 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,   0,   0 },    269.54510,   0.00005,     0.025 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -4,   0 },     89.43263,   0.00009,     0.087 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   4,  -3,   0 },     89.43126,   0.00015,     0.040 },
 { {   0,  25, -30,   0,   1,   0,   0,   0,  -2,   0,   0 },    319.86582,   0.00005,     0.071 },
 { {   0,  25, -29,   0,  -2,   0,   0,   0,  -2,   0,   0 },     42.15225,   0.00006,     0.075 },
 { {   0,  25, -29,   0,  -1,   0,   0,   0,  -2,   0,   0 },    290.15784,   0.00002,     0.076 },
 { {   0,  25, -29,   0,  -1,   0,   0,   0,  -2,   2,   0 },    110.15784,   0.00002,     0.075 },
 { {   0,  25, -27,  -4,   0,   0,   0,   0,  -2,   0,   0 },    263.13905,   0.00002,     0.076 },
 { {   0,  25, -27,  -4,   0,   0,   0,   0,  -2,   2,   0 },     83.13905,   0.00002,     0.075 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,  -2,   0 },    269.51321,   0.00004,     0.078 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   0,   0 },     87.21272,   0.00009,     0.073 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    181.62336,   0.00002,     0.099 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },      1.62336,   0.00002,     0.061 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,   0,   0 },      2.03489,   0.00007,     0.068 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   0,  -3 },    157.08515,   0.00005,     0.075 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   2,  -3 },    337.08515,   0.00005,     0.075 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   2,   0,   0 },      2.18568,   0.00013,     0.025 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.55183,   0.00002,     0.092 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    269.64465,   0.00011,     0.064 },
 { {   0,  25, -25,   0,   0,   0,   0,   0,   2,   0,   0 },    269.49756,   0.00003,     0.025 },
 { {   0,  26, -32,   2,   0,   0,   0,   0,  -2,   0,   0 },    321.19979,   0.00004,     0.075 },
 { {   0,  26, -32,   2,   0,   0,   0,   0,  -2,   2,   0 },    141.19979,   0.00004,     0.076 },
 { {   0,  26, -31,   0,   1,   0,   0,   0,  -2,   0,   0 },    199.35886,   0.00007,     0.075 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   0,  -2,   0 },     87.93767,   0.00003,     0.070 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   0,   0,   0 },     87.93767,   0.00003,     0.082 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   2,  -2,   0 },     87.93767,   0.00004,     0.095 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,   2,  -1,   0 },     87.93767,   0.00010,     0.042 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -4,   0,   0 },    165.00074,   0.00003,     0.028 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -2,   0 },    165.00074,   0.00013,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,  -1,   0 },    165.00074,   0.00150,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    164.77845,   0.00106,     0.087 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -1,  -1,   0 },    345.00074,   0.00003,     0.081 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -3,   0 },    165.04743,   0.00029,     0.038 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   0 },    165.04743,   0.00535,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   2 },    165.00074,   0.00002,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   0 },    343.41852,   0.00004,   127.422 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,  -2 },    345.00074,   0.00002,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   0 },    345.28498,   0.00467,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   1,   0 },    345.28498,   0.00025,     0.038 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   1,  -1,   0 },    165.00074,   0.00003,     0.081 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -2,   0 },    344.95342,   0.00094,     0.087 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,  -1,   0 },    345.00074,   0.00150,     0.040 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   2,   0,   0 },    345.00074,   0.00013,     0.026 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   4,  -2,   0 },    345.00074,   0.00003,     0.028 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -2,  -1,   0 },     62.06381,   0.00010,     0.042 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.06381,   0.00004,     0.095 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -1,  -1,  -1 },    256.27827,   0.00005,     0.040 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,  -1,   0,  -1 },    256.27827,   0.00003,     0.087 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,  -2,   0 },     62.03821,   0.00003,     0.082 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,   0,   0 },     69.82097,   0.00006,     0.070 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -2,  -1 },    256.26445,   0.00016,     0.075 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.30774,   0.00028,     0.076 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   3,  -2,  -1 },     76.27827,   0.00003,     0.087 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   3,  -1,  -1 },     76.27827,   0.00005,     0.041 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,  -2,  -2 },      2.16399,   0.00002,     0.075 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,   0,  -2 },    182.16399,   0.00002,     0.076 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,   2,   0,   0 },      2.16580,   0.00010,     0.025 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.65360,   0.00002,     0.118 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.65360,   0.00002,     0.055 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },     89.69993,   0.00002,     0.098 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    269.69993,   0.00002,     0.061 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,   2,   0,   0 },    269.44805,   0.00018,     0.024 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    180.37860,   0.00002,     0.078 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   2,   0 },     27.21860,   0.00005,     0.073 },
 { {   0,  27, -32,   0,   1,   0,   0,   0,  -2,   0,   0 },    139.77280,   0.00002,     0.078 },
 { {   0,  27, -32,   0,   1,   0,   0,   0,  -2,   2,   0 },    139.77280,   0.00002,     0.073 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,   2,  -2,   0 },      2.14260,   0.00002,     0.071 },
 { {   0,  27, -28,   0,   0,   0,   0,   0,   2,   0,   0 },      2.14540,   0.00006,     0.025 },
 { {   0,  27, -27,   0,   0,   0,   0,   0,   2,   0,   0 },    269.39685,   0.00013,     0.024 },
 { {   0,  28, -35,   2,   0,   0,   0,   0,  -2,   0,   0 },    233.30848,   0.00002,     0.076 },
 { {   0,  28, -35,   2,   0,   0,   0,   0,  -2,   2,   0 },     53.30848,   0.00002,     0.075 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    152.47435,   0.00007,     0.076 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   2,   0 },    332.47435,   0.00007,     0.075 },
 { {   0,  28, -34,   0,   1,   0,   0,   0,  -2,   0,   0 },    231.54621,   0.00002,     0.076 },
 { {   0,  28, -34,   0,   1,   0,   0,   0,  -2,   2,   0 },     51.54621,   0.00002,     0.075 },
 { {   0,  28, -34,   0,   2,   0,   0,   0,  -2,   0,   0 },    320.94938,   0.00007,     0.077 },
 { {   0,  28, -28,   0,   0,   0,   0,   0,   2,   0,   0 },    269.34388,   0.00010,     0.024 },
 { {   0,  29, -36,   0,   4,   0,   0,   0,  -2,   0,   0 },    316.91482,   0.00003,     0.075 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,  -2,   0 },    327.10944,   0.00005,     0.075 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,   0,   0 },    147.10944,   0.00005,     0.076 },
 { {   0,  29, -34,   0,   1,   0,   0,   0,   0,  -2,   0 },    319.37312,   0.00002,     0.075 },
 { {   0,  29, -34,   0,   1,   0,   0,   0,   0,   0,   0 },    139.37312,   0.00002,     0.076 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -3,   0 },     65.62831,   0.00002,     0.075 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -1,   0 },    245.62831,   0.00002,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,  -1,  -1,  -1 },    252.70535,   0.00003,     0.040 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,  -1,   0,  -1 },    252.70535,   0.00002,     0.087 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,  -2,  -1 },    252.70535,   0.00011,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   1,   0,  -1 },     72.70535,   0.00011,     0.075 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   3,  -2,  -1 },     72.70535,   0.00002,     0.087 },
 { {   0,  29, -33,   0,   2,   0,   0,   0,   3,  -1,  -1 },     72.70535,   0.00003,     0.040 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -3,   0,  -1 },    203.91542,   0.00002,     0.040 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,  -1,  -1 },    203.91542,   0.00007,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,  -1,   1,  -1 },     23.91542,   0.00007,     0.075 },
 { {   0,  35, -31,   0,   0,  -3,   0,   0,   1,   0,  -1 },     23.91542,   0.00002,     0.040 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,  -1,   0 },     37.19129,   0.00003,     0.076 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   1,   0 },    217.19129,   0.00003,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -4,   0,   0 },    106.24881,   0.00003,     0.040 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,  -1,   0 },    106.24881,   0.00009,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,  -2,   1,   0 },    286.24881,   0.00009,     0.075 },
 { {   0,  39, -36,  -5,   0,   0,   0,   0,   0,   0,   0 },    286.24881,   0.00003,     0.040 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,  -1,   0 },    181.32596,   0.00002,     0.075 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   1,   0 },      1.32596,   0.00002,     0.076 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,  -1,   0,  -3 },     11.55562,   0.00009,     0.040 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,  -1,   1,  -3 },     11.55562,   0.00006,     0.087 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,  -1,  -3 },     11.55562,   0.00033,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   1,  -3 },    191.55562,   0.00033,     0.075 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   3,  -1,  -3 },    191.55562,   0.00006,     0.087 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   3,   0,  -3 },    191.55562,   0.00009,     0.040 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -4,   0,   0 },     26.45270,   0.00003,     0.040 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,  -1,   0 },     26.45270,   0.00009,     0.075 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,  -2,   1,   0 },    206.45270,   0.00009,     0.075 },
 { {   0,  44, -42,  -9,   0,   0,   0,   0,   0,   0,   0 },    206.45270,   0.00003,     0.040 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,  -1,   0 },    349.65428,   0.00003,     0.075 },
 { {   0,  47, -52,   0,   4,   0,   0,   0,  -2,   1,   0 },    169.65428,   0.00003,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   0,   2 },    167.18279,   0.00004,     0.075 },
 { {   1, -28,  26,   0,   0,   0,   0,   0,  -2,   2,   2 },    347.18279,   0.00004,     0.075 },
 { {   1, -25,  25,   0,   0,   0,   0,   0,   2,  -2,   0 },    162.86625,   0.00003,     0.075 },
 { {   1, -25,  25,   0,   0,   0,   0,   0,   2,   0,   0 },    342.86625,   0.00003,     0.075 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -3,   0 },    204.25737,   0.00005,     0.075 },
 { {   1,  -7,   9,   0,   0,   0,   0,   0,   2,  -1,   0 },     24.25737,   0.00005,     0.075 },
 { {   1,  -6,   4,   0,   0,   0,   0,   0,  -2,   1,   0 },    166.90106,   0.00002,     0.076 },
 { {   1,  -6,   4,   0,   0,   0,   0,   0,  -2,   3,   0 },    346.90106,   0.00002,     0.074 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,  -2,   0,   0 },     81.70769,   0.00003,     0.040 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   0,  -1,   0 },     78.53644,   0.00003,     0.076 },
 { {   1,  -5,   4,   0,   0,   0,   0,   0,   0,   1,   0 },    258.53644,   0.00003,     0.075 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,   0,   0 },    216.93434,   0.00009,     0.040 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,  -2,   1,   0 },    216.93434,   0.00013,     0.087 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,  -1,   0 },     36.93434,   0.00013,     0.087 },
 { {   1,   0, -10,  11,   0,   0,   0,   0,   2,   0,   0 },     36.93434,   0.00009,     0.040 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    218.87340,   0.00003,     0.075 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },     38.87340,   0.00003,     0.076 },
 { {   1,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    230.55205,   0.00005,     0.042 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    299.54950,   0.00004,     0.026 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    299.64234,   0.00044,     0.041 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    298.90242,   0.00032,     0.088 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    300.60085,   0.00003,     0.038 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    301.82419,   0.00063,     0.076 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    123.25105,   0.00004,     6.575 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    121.98001,   0.00062,     0.075 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    122.20501,   0.00003,     0.038 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    120.19733,   0.00029,     0.086 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    120.44267,   0.00033,     0.040 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    226.64050,   0.00006,     0.027 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    226.64537,   0.00049,     0.042 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    226.65645,   0.00025,     0.097 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    226.85471,   0.00022,     0.083 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     47.28221,   0.00010,     0.868 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     47.14394,   0.00020,     0.069 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },     19.96838,   0.00002,     0.076 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    199.96838,   0.00002,     0.075 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     48.61684,   0.00015,     0.079 },
 { {   1,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     48.59945,   0.00016,     0.039 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.38131,   0.00021,     0.044 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    357.20559,   0.00005,     0.107 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    347.31603,   0.00017,     0.090 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    352.12411,   0.00006,     0.465 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    337.91554,   0.00003,     0.065 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    165.67983,   0.00007,     2.645 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,  -2 },    167.59292,   0.00007,     0.076 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    345.75638,   0.00115,     0.073 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    174.36265,   0.00012,     0.037 },
 { {   1,   0,  -2,   0,   0,   0,   0,   0,   2,   1,  -2 },    347.59292,   0.00007,     0.075 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     94.87168,   0.00009,     0.046 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    268.28031,   0.00003,     0.099 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     88.66698,   0.00011,     0.317 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     88.28031,   0.00003,     0.061 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },     88.12267,   0.00013,     0.068 },
 { {   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    268.03358,   0.00013,     0.036 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },     22.07596,   0.00007,     0.049 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    200.78829,   0.00003,     0.241 },
 { {   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },     11.64758,   0.00005,     0.035 },
 { {   1,   1,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },     72.10263,   0.00003,     0.075 },
 { {   1,   1,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    252.10263,   0.00003,     0.075 },
 { {   1,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    166.83770,   0.00002,     0.076 },
 { {   1,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    346.83770,   0.00002,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -3,   1,  -1 },     94.34116,   0.00004,     0.040 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -3,   2,  -1 },     94.34116,   0.00003,     0.087 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },     92.38969,   0.00018,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },    272.38969,   0.00018,     0.075 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    274.34116,   0.00003,     0.087 },
 { {   1,  15, -16,   0,   0,   0,   0,   0,   1,   1,  -1 },    274.34116,   0.00004,     0.040 },
 { {   2, -30,  29,   0,   0,   0,   0,   0,   2,  -2,   0 },     81.36750,   0.00002,     0.076 },
 { {   2, -30,  29,   0,   0,   0,   0,   0,   2,   0,   0 },    261.36750,   0.00002,     0.075 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -4,   0 },    117.44429,   0.00003,     0.075 },
 { {   2, -24,  21,   0,   0,   0,   0,   0,   4,  -2,   0 },    297.44429,   0.00003,     0.075 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },     44.45685,   0.00003,     0.076 },
 { {   2,  -4,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    224.45685,   0.00003,     0.075 },
 { {   2,  -2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },      3.69942,   0.00003,     0.074 },
 { {   2,  -2,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    183.69942,   0.00003,     0.076 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    203.79542,   0.00014,     0.075 },
 { {   2,  -2,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },     23.79542,   0.00014,     0.075 },
 { {   2,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    230.50891,   0.00004,     0.063 },
 { {   2,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     50.51950,   0.00007,     0.034 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    341.81416,   0.00006,     0.189 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    340.90166,   0.00009,     0.060 },
 { {   2,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    160.90172,   0.00006,     0.033 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    270.22536,   0.00013,     0.054 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    270.04805,   0.00003,     0.144 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     90.38851,   0.00018,     0.159 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     90.04805,   0.00003,     0.051 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.89230,   0.00015,     0.056 },
 { {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    269.91129,   0.00017,     0.032 },
 { {   2,  13, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    129.86755,   0.00002,     0.075 },
 { {   2,  13, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    309.86755,   0.00002,     0.076 },
 { {   3,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     51.06433,   0.00004,     0.031 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    162.96838,   0.00004,     0.061 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    342.27346,   0.00004,     0.118 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    341.19809,   0.00003,     0.050 },
 { {   3,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    161.19807,   0.00011,     0.030 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.31095,   0.00012,     0.065 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     91.37743,   0.00014,     0.106 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.44552,   0.00005,     0.048 },
 { {   3,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    270.44548,   0.00018,     0.029 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   0,   1 },     42.75570,   0.00002,     0.076 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -3,   2,   1 },    222.75570,   0.00002,     0.075 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    185.68887,   0.00011,     0.070 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },      8.27120,   0.00005,     0.096 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    109.50229,   0.00003,     0.042 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    193.75221,   0.00005,     0.028 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    186.56536,   0.00004,     0.026 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    186.56536,   0.00042,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -4,   2,   0 },    186.65436,   0.00026,     0.087 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    186.29492,   0.00012,     0.038 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    186.29492,   0.00216,     0.075 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },      5.86444,   0.00004,    39.271 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },      6.48023,   0.00151,     0.076 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },      6.48023,   0.00008,     0.038 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },      7.29777,   0.00014,     0.087 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },      6.56536,   0.00042,     0.040 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },      6.56536,   0.00004,     0.026 },
 { {   3,   0,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },     83.62843,   0.00003,     0.042 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -4,   1,   0 },     83.96206,   0.00002,     0.040 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   0,   0 },     83.96206,   0.00008,     0.075 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,  -2,   2,   0 },    263.96206,   0.00008,     0.075 },
 { {   3,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    263.96206,   0.00002,     0.040 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   1,   0 },    100.24706,   0.00010,     0.075 },
 { {   4, -15,   6,   0,   0,   0,   0,   0,  -2,   3,   0 },    280.24706,   0.00010,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -4,   1,   0 },    120.68945,   0.00002,     0.040 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },    120.68945,   0.00008,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    300.68945,   0.00008,     0.075 },
 { {   4,  -5,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    300.68945,   0.00002,     0.040 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },    163.47726,   0.00004,     0.041 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    163.47059,   0.00002,     0.088 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    343.74934,   0.00018,     0.076 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    343.48383,   0.00015,     0.075 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    342.73085,   0.00013,     0.086 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    343.47726,   0.00004,     0.040 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    161.68534,   0.00005,     0.028 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.98944,   0.00008,     0.082 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,  -1,   2,  -1 },    294.76028,   0.00006,     0.075 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     92.68180,   0.00018,     0.079 },
 { {   4,   0,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    271.50632,   0.00007,     0.027 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    189.34229,   0.00004,     0.090 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   1,  -2 },    190.63105,   0.00006,     0.040 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   2,  -2 },    190.63105,   0.00004,     0.087 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    189.28315,   0.00004,     0.078 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2 },    190.77267,   0.00014,     0.076 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    189.33814,   0.00020,     0.073 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   2,  -2 },     10.73601,   0.00022,     0.075 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },     10.63105,   0.00004,     0.087 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    193.54097,   0.00004,     0.026 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   2,   1,  -2 },     10.63105,   0.00006,     0.040 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   0,   0 },     33.08124,   0.00003,     0.075 },
 { {   5,   0, -13,   7,   0,   0,   0,   0,  -2,   2,   0 },    213.08124,   0.00003,     0.075 },
 { {   5,   0,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     52.90780,   0.00003,     0.073 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 4384> g_ELP13
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    270.00000,   0.00011, 99999.999 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    277.11719,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },      3.01272,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   1 },    248.93951,   0.00001,     1.388 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },     70.48240,   0.00006,     0.949 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1 },     68.93951,   0.00001,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    187.47509,   0.00002,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },     79.95550,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    150.06581,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },     16.67961,   0.00007,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0 },     86.25753,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0 },    144.61213,   0.00006,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   0 },    142.52892,   0.00007,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },    152.91482,   0.00009,     0.076 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },     19.39366,   0.00004,    29.457 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0 },    152.91482,   0.00009,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,   0 },    142.40817,   0.00007,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0 },    144.56735,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 },     59.80021,   0.00004,    14.729 },
 { {   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 },    308.78241,   0.00002,    11.705 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,  -1,   0 },    300.69188,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 },    300.69183,   0.00018,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   0 },    307.21040,   0.00035,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   2,   0 },    300.69188,   0.00001,     0.538 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 },    223.33584,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },    228.05341,   0.00018,     0.076 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },    334.24847,   0.00063,    11.862 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 },    225.54658,   0.00018,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   2,   0 },    223.62764,   0.00001,     0.038 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,  -1 },      4.05502,   0.00002,     1.032 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -2,   0 },    300.72259,   0.00001,     0.592 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    307.69915,   0.00032,     0.086 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    300.72275,   0.00017,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   1,   0 },    300.72267,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   0 },    330.24353,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   0 },    276.71860,   0.00032,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   2,   0 },    259.08333,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   0 },     79.17198,   0.00009,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   0 },     79.70328,   0.00131,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 },     75.69874,   0.00293,   883.328 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   0 },     79.70215,   0.00131,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   0 },     79.17231,   0.00009,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -2,   0 },    259.08146,   0.00002,     0.563 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   0 },    276.56701,   0.00032,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,   0 },    328.37435,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 },    354.94068,   0.00003,    30.473 },
 { {   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 },     37.88628,   0.00001,    14.978 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 },    315.92589,   0.00002,     0.041 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   1,   0 },    344.07043,   0.00002,     0.088 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0 },    236.49931,   0.00005,     0.076 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 },     75.13795,   0.00007,     5.931 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   0 },    236.49931,   0.00005,     0.074 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,  -1,   0 },    319.61726,   0.00002,     0.086 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   0 },    286.40217,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 },    188.04062,   0.00005,    12.023 },
 { {   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 },    265.05937,   0.00001,     3.954 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   0 },    302.93942,   0.00002,     0.020 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0 },    139.85925,   0.00001,     0.041 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   1,   0 },    303.08323,   0.00008,     0.091 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0 },    297.30874,   0.00002,     0.434 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   3,   0 },    122.69884,   0.00002,     0.064 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1 },    270.93278,   0.00002,     0.631 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0 },     57.15141,   0.00001,     0.038 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0 },    126.81789,   0.00005,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },    283.03210,   0.00005,     1.881 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0 },    125.91783,   0.00007,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   2,   0 },    119.85392,   0.00001,     0.037 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1 },     45.61218,   0.00004,     1.915 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -2,   0 },    240.34411,   0.00001,     0.805 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,  -1,   0 },    300.35413,   0.00003,     0.083 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   2,   0,   0 },     68.02734,   0.00001,     0.040 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,  -2,   0 },    152.02796,   0.00001,     0.020 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0 },    105.91603,   0.00002,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0 },    298.95146,   0.00002,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0 },    145.34707,   0.00003,     0.082 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },     14.57036,   0.00007,     0.940 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,   0 },    147.25478,   0.00004,     0.070 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -1 },     35.53110,   0.00002,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,  -1 },    352.59690,   0.00016,   104.137 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -1 },     35.53110,   0.00002,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -3,   0 },    132.43105,   0.00002,     0.072 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -2,   0 },    324.37576,   0.00005,     1.407 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1,   0 },    150.24241,   0.00003,     0.080 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,   0 },    152.35373,   0.00006,     0.039 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   1,   0 },      0.95965,   0.00003,     0.101 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   3,   0 },    180.95965,   0.00003,     0.060 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0 },    357.45076,   0.00004,     0.086 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 },    193.73556,   0.00005,     0.627 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   1,   0,  -1 },    327.00435,   0.00001,     1.847 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -3,   0 },     30.49535,   0.00003,     0.074 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -2,   0 },      6.35957,   0.00006,     5.588 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -1,   0 },    177.45046,   0.00021,     0.076 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,   0,   0 },    354.78056,   0.00004,     0.038 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   0,  -1,   0 },     42.64415,   0.00002,     0.090 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0 },    222.64415,   0.00003,     0.470 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   0,   1,   0 },     42.64415,   0.00002,     0.065 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,  -3,   0 },     32.20669,   0.00001,     0.078 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,  -1,  -2 },     42.42850,   0.00004,     0.076 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,  -1,   0 },    194.96757,   0.00005,     0.073 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   0,  -2 },     52.93399,   0.00010,    52.068 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   0,   0 },    250.62218,   0.00003,     0.037 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   1,  -2 },     42.42850,   0.00004,     0.075 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,  -1,   0 },    251.39928,   0.00003,     0.094 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0 },     71.35046,   0.00006,     0.376 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,   1,   0 },    251.39926,   0.00004,     0.063 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,  -3,   0 },     68.86225,   0.00002,     0.081 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,  -1,   0 },     70.72838,   0.00008,     0.071 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,   0,  -2 },     71.39920,   0.00001,     1.815 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,   0,   0 },     60.78440,   0.00005,     0.037 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   3,  -3,   0 },     71.26340,   0.00001,   383.916 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,  -1,   0 },    280.22737,   0.00001,     0.099 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,   0,   0 },    100.06868,   0.00005,     0.313 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,   1,   0 },    280.22737,   0.00001,     0.061 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,   2,   0 },    280.77506,   0.00001,     0.034 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,  -3,   0 },    275.27611,   0.00001,     0.084 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,  -2,   0 },     97.94810,   0.00002,     0.706 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,  -1,   0 },     99.50822,   0.00010,     0.068 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,   0,   0 },    278.72105,   0.00007,     0.036 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   0,  -1,   0 },    129.65557,   0.00001,     0.105 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   0,   0,   0 },    309.65557,   0.00002,     0.269 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   0,   1,   0 },    129.65557,   0.00001,     0.059 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   0,   2,   0 },    309.65557,   0.00002,     0.033 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   2,  -1,   0 },    309.28128,   0.00002,     0.066 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   2,   0,   0 },    129.74506,   0.00002,     0.035 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   2,   2,   0 },    309.65557,   0.00001,     0.018 },
 { {   0,   0,   0,   8,   0,   0,   0,   0,   2,  -1,   0 },    338.39798,   0.00004,     0.064 },
 { {   0,   0,   0,   8,   0,   0,   0,   0,   2,   0,   0 },    338.39798,   0.00002,     0.034 },
 { {   0,   0,   1, -10,   0,   0,   0,   0,  -2,   0,   0 },    352.43647,   0.00003,     0.034 },
 { {   0,   0,   1, -10,   0,   0,   0,   0,  -2,   1,   0 },    352.27071,   0.00003,     0.063 },
 { {   0,   0,   1, -10,   0,   0,   0,   0,  -2,   2,   0 },    172.54119,   0.00002,     0.393 },
 { {   0,   0,   1,  -9,   0,   0,   0,   0,  -2,   0,   0 },    201.55034,   0.00002,     0.035 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,  -2,  -2,   0 },    230.30250,   0.00001,     0.018 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,  -2,   0,   0 },     50.41425,   0.00003,     0.036 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,  -2,   1,   0 },    230.40469,   0.00003,     0.068 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,   0,  -2,   0 },    230.30250,   0.00002,     0.034 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,   0,  -1,   0 },     50.30250,   0.00002,     0.061 },
 { {   0,   0,   1,  -8,   0,   0,   0,   0,   0,   1,   0 },     50.30250,   0.00002,     0.100 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   0,   0 },    260.41432,   0.00006,     0.036 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   0,   2 },     79.41346,   0.00001,     1.628 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   1,   0 },     78.82597,   0.00004,     0.070 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   2,   0 },     79.41346,   0.00002,     1.055 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   3,   0 },     77.16219,   0.00001,     0.081 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   0 },    259.18493,   0.00001,     0.034 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,  -1,   0 },    259.46790,   0.00003,     0.063 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,   0,   0 },     79.54130,   0.00004,     0.367 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,   1,   0 },    259.46790,   0.00003,     0.095 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,  -2,   0,   0 },    108.46861,   0.00003,     0.037 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,  -2,   0,   2 },    159.82732,   0.00001,    12.111 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,  -2,   1,   0 },    113.71720,   0.00004,     0.073 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,  -2,   3,   0 },    123.29377,   0.00002,     0.078 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,  -1,   0 },    288.49967,   0.00003,     0.065 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,   0,   0 },    108.51420,   0.00005,     0.457 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,   1,   0 },    288.49967,   0.00003,     0.090 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -4,   3,   0 },    324.09706,   0.00001,     0.088 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   0,   0 },    318.13518,   0.00005,     0.038 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   1,   0 },    328.80329,   0.00012,     0.076 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   3,   0 },    124.59770,   0.00004,     0.075 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -1,   0,   1 },    234.96480,   0.00001,     1.654 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0 },    334.95605,   0.00006,     0.603 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   0,   1,   0 },    141.39948,   0.00002,     0.086 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   2,  -3,   0 },    330.11867,   0.00001,     0.060 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   2,  -1,   0 },    150.11867,   0.00001,     0.102 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,   2,   0,   0 },    330.20189,   0.00003,     0.043 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -4,   2,   0 },    357.95333,   0.00001,     0.042 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   0,   0 },    164.87700,   0.00011,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   1,   0 },    184.46906,   0.00004,     0.079 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   2,   0 },    193.26270,   0.00002,     1.545 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   1 },    251.39600,   0.00005,    13.704 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0 },    270.33530,   0.00001,     0.888 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   2,   0 },    328.81415,   0.00001,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,  -2,   0 },    174.20052,   0.00001,     0.345 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,   2,   0,   0 },     74.14706,   0.00001,     0.042 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   0,   0 },     32.50234,   0.00014,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   1,   0 },     35.52063,   0.00006,     0.083 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   2,   0 },     42.12003,   0.00002,     0.848 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,   1 },     70.25463,   0.00003,     2.180 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0 },     28.24731,   0.00011,     0.072 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0 },    336.70157,   0.00007,     1.681 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0 },     27.21144,   0.00008,     0.079 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0 },    208.61954,   0.00002,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,  -1,   0 },    208.24731,   0.00002,     0.092 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,   0 },     29.92875,   0.00006,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -2,   0 },    244.87714,   0.00001,     0.019 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,  -1,   0 },     53.07731,   0.00002,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   0 },     54.17379,   0.00020,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   0 },     59.52204,   0.00030,     0.087 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   3,   0 },     64.01269,   0.00002,     0.067 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   1 },     58.74954,   0.00002,     1.010 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0 },     29.75239,   0.00003,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0 },     43.10853,   0.00057,     0.075 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },     57.64503,   0.00235,    15.781 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0 },     43.40532,   0.00059,     0.076 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0 },     45.51386,   0.00005,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,  -1 },     60.69304,   0.00003,     0.895 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   0 },     57.99304,   0.00035,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   0 },     43.73973,   0.00007,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   1,   0 },     51.09044,   0.00002,     0.026 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,   0 },    237.07553,   0.00002,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   0 },    259.94498,   0.00003,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   0 },     92.91490,   0.00014,     0.091 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   2,   0 },     93.10758,   0.00002,     0.446 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   1 },    241.92486,   0.00001,     0.657 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0 },    271.86978,   0.00003,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0 },    252.57661,   0.00011,     2.135 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0 },    242.56972,   0.00002,     0.073 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0 },    291.53484,   0.00001,     0.037 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,  -1 },     13.04117,   0.00004,     1.708 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -2,   0 },    266.19431,   0.00001,     0.766 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0 },     87.23744,   0.00012,     0.040 },
 { {   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 },     25.58242,   0.00004,     1.509 },
 { {   0,   0,   1,   0,  -4,   5,   0,   0,   0,   0,   0 },     10.41090,   0.00001,     1.201 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },     32.65208,   0.00006,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,   0 },     32.65206,   0.00007,     0.093 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,   0 },     30.96402,   0.00005,     0.080 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 },     30.47728,   0.00026,     1.339 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   0 },     30.96402,   0.00005,     0.071 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,   0 },     59.94915,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,   0 },     33.68524,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,  -2,   0,   0 },     75.82346,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -3,   5,   0,   0,   0,   0,   0 },     75.68757,   0.00006,     1.091 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,  -1,   0 },     43.94931,   0.00003,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },     43.91263,   0.00026,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },     44.09499,   0.00036,     0.094 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   2,   0 },     43.94746,   0.00001,     0.384 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   1 },    337.44473,   0.00001,     0.530 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   0 },     42.29667,   0.00001,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   0 },     42.82629,   0.00027,     0.080 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },     43.20942,   0.00108,     1.203 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   0 },     42.15331,   0.00019,     0.071 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   2,   0 },     42.38233,   0.00001,     0.037 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   0 },     44.42729,   0.00002,    17.980 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   1 },    337.45446,   0.00003,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,  -1 },    337.45581,   0.00007,     4.497 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },     44.69475,   0.00004,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },     42.50852,   0.00012,     0.081 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },     42.71786,   0.00005,     0.039 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   0,   0 },    120.76681,   0.00001,     0.042 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,  -2,   1,   0 },    177.48175,   0.00001,     0.095 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   0,   0,   0 },    104.54800,   0.00002,     0.999 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    112.45837,   0.00002,     0.042 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },     40.87591,   0.00003,     1.092 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,  -1 },    169.72312,   0.00002,     7.243 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,  -2,   0,   0 },     50.85586,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,  -5,   0,   0,   0,   0,   0 },     49.08444,   0.00004,     1.204 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,  -2,   0,   0 },    136.15438,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0 },    136.36661,   0.00003,     1.113 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,  -2,   0,   0 },     36.68248,   0.00001,     0.042 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0 },     37.08985,   0.00004,     1.073 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,  -1 },     14.88167,   0.00003,     8.220 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0 },    256.60856,   0.00001,     1.035 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   0 },    345.87698,   0.00002,     0.021 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   1,   0 },    341.53692,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    303.06280,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.26829,   0.00009,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },     14.97670,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    134.61183,   0.00011,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },    135.68030,   0.00004,     0.360 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   1 },    192.41737,   0.00002,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    234.35923,   0.00004,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    256.25126,   0.00003,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,  -1 },     29.14816,   0.00004,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   1 },    192.41737,   0.00002,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    133.39625,   0.00010,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    302.13678,   0.00017,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     91.56307,   0.00007,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0 },    241.25638,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -3,   1 },    173.92535,   0.00002,     0.077 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   1 },    318.86811,   0.00001,     3.575 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    127.39075,   0.00008,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   1 },     96.39459,   0.00005,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },     18.10613,   0.00008,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    218.35007,   0.00004,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -1 },    127.39075,   0.00008,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    351.56040,   0.00001,     0.071 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    330.94655,   0.00006,     1.292 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    316.38801,   0.00009,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -2 },    161.06505,   0.00001,     0.903 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     28.09397,   0.00009,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   1,   0 },    161.06505,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   2,   0 },    156.24889,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -1,  -1 },    307.39075,   0.00002,     0.088 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0 },     41.59423,   0.00001,     0.967 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,  -1 },    160.86334,   0.00004,    50.462 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,  -1 },    131.67040,   0.00009,    70.769 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,   0 },    108.35193,   0.00002,     0.042 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,   0,   0,   0 },    107.32090,   0.00006,     1.093 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,  -1 },    347.10492,   0.00007,   293.315 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   0 },    173.94446,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   1,   0 },    173.94446,   0.00002,     0.096 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0 },    158.22083,   0.00007,     0.922 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,  -1 },    165.37557,   0.00011,    32.746 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   0 },    164.34026,   0.00008,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,   0 },    163.70582,   0.00012,     0.095 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,   0 },    164.33778,   0.00007,     0.082 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   0,   0 },    162.46050,   0.00037,     1.001 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,   0 },    163.60736,   0.00008,     0.070 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   2,   0,   0 },    160.77928,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,  -1 },    141.08380,   0.00007,     8.708 },
 { {   0,   0,   1,   0,   4,   0,   0,   0,   2,  -3,   0 },     67.72502,   0.00002,     0.073 },
 { {   0,   0,   1,   0,   4,   0,   0,   0,   2,  -1,   0 },     67.72502,   0.00002,     0.078 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0 },    152.82704,   0.00003,     0.043 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0 },    266.15523,   0.00001,     0.040 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0 },    332.71745,   0.00005,     0.085 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0 },    269.60407,   0.00003,     0.653 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   1,   0 },    331.79812,   0.00004,     0.068 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   1,   0,  -1 },    284.63096,   0.00003,     2.092 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -3,   0 },    307.89695,   0.00003,     0.074 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -2,   0 },    318.85007,   0.00003,     4.127 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -1,   0 },    164.64149,   0.00008,     0.077 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,   0,   0 },    313.68762,   0.00003,     0.038 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   0,   0 },    355.87353,   0.00001,     0.044 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,  -2,   2,   0 },    355.87353,   0.00001,     0.261 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   0,  -2,   0 },    180.52867,   0.00001,     0.041 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   0,  -1,   0 },    180.24927,   0.00001,     0.089 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0 },    194.77157,   0.00007,     0.485 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,  -3,   0 },    182.73767,   0.00002,     0.077 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,  -2,   0 },     27.01593,   0.00006,     3.456 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,  -1,   0 },    359.67667,   0.00010,     0.074 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,   0,  -2 },    342.58937,   0.00004,    22.644 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,   0,   0 },    221.12871,   0.00003,     0.037 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   4,  -3,   0 },      0.24927,   0.00001,     0.085 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   0,  -1,   0 },     43.59002,   0.00001,     0.094 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   0,   0,   0 },    222.36003,   0.00003,     0.385 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   0,   1,   0 },     42.36003,   0.00002,     0.063 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -3,   0 },     47.35774,   0.00001,     0.080 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -2,   0 },     42.36003,   0.00001,     1.218 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -1,   0 },     38.66278,   0.00005,     0.071 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,   0,  -2 },     42.36003,   0.00001,     2.051 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,   0,   0 },    219.35598,   0.00003,     0.037 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,   2,   0 },    222.36003,   0.00001,     0.019 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,  -1,   0 },    251.26742,   0.00003,     0.099 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,   0,   0 },     67.91628,   0.00011,     0.320 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,   1,   0 },    251.28008,   0.00004,     0.061 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   2,  -3,   0 },     73.17035,   0.00002,     0.084 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   2,  -1,   0 },     71.66332,   0.00008,     0.068 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   0,   0,   0 },    100.19570,   0.00008,     0.273 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   0,   2,   0 },    280.73402,   0.00001,     0.033 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,  -3,   0 },    280.34046,   0.00001,     0.088 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,  -2,   0 },     98.46457,   0.00005,     0.531 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,  -1,   0 },    100.46497,   0.00001,     0.066 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,   0,   0 },    278.76815,   0.00006,     0.035 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,   2,   0 },    100.49671,   0.00002,     0.018 },
 { {   0,   0,   1,   6,   0,   0,   0,   0,   0,   0,   0 },    309.60011,   0.00003,     0.239 },
 { {   0,   0,   1,   6,   0,   0,   0,   0,   2,  -1,   0 },    309.60011,   0.00001,     0.064 },
 { {   0,   0,   1,   7,   0,   0,   0,   0,   2,   0,   0 },    338.32567,   0.00002,     0.034 },
 { {   0,   0,   1,   8,   0,   0,   0,   0,   2,   0,   0 },    187.40272,   0.00001,     0.033 },
 { {   0,   0,   2, -12,   0,   0,   0,   0,  -2,   0,   0 },    143.25538,   0.00001,     0.034 },
 { {   0,   0,   2, -12,   0,   0,   0,   0,  -2,   2,   0 },    323.25538,   0.00001,     0.384 },
 { {   0,   0,   2, -11,   0,   0,   0,   0,  -2,   0,   0 },    352.33044,   0.00006,     0.035 },
 { {   0,   0,   2, -11,   0,   0,   0,   0,  -2,   2,   0 },    171.98487,   0.00001,     0.482 },
 { {   0,   0,   2, -10,   0,   0,   0,   0,  -2,   0,   0 },    201.29161,   0.00003,     0.036 },
 { {   0,   0,   2, -10,   0,   0,   0,   0,   0,   0,   0 },    201.19116,   0.00001,     0.301 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,  -2,  -2,   0 },    230.25929,   0.00001,     0.019 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,  -2,   2,   0 },     50.33268,   0.00002,     0.989 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,   0,  -2,   0 },    230.25929,   0.00002,     0.034 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,   0,  -1,   0 },     50.25929,   0.00002,     0.062 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,   0,   0,   0 },     50.09523,   0.00001,     0.359 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,   0,   1,   0 },     50.25929,   0.00002,     0.096 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   0,   0 },    260.14389,   0.00007,     0.037 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   0,   2 },    259.37664,   0.00003,     6.852 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   1,   0 },     79.37665,   0.00003,     0.073 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   0 },    259.39286,   0.00002,     0.035 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,  -1,   0 },    259.37664,   0.00003,     0.064 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,   0,   0 },     84.63228,   0.00008,     0.444 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,   1,   0 },    259.37664,   0.00003,     0.091 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   0,   0 },    108.79025,   0.00005,     0.038 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   0,   2 },    288.28104,   0.00001,     2.592 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   1,   0 },    295.08151,   0.00016,     0.076 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   2,   0 },    317.33501,   0.00005,    19.147 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   3,   0 },    103.39579,   0.00003,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   0 },    288.28104,   0.00002,     0.035 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   0,  -1,   0 },    288.30493,   0.00004,     0.067 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   0,   0,   0 },    293.72162,   0.00005,     0.581 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   2,  -2,   0 },    301.07966,   0.00002,     0.286 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   2,   0,   0 },    121.03113,   0.00001,     0.043 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   0,   0 },    317.22973,   0.00008,     0.039 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   1,   0 },    141.13474,   0.00001,     0.079 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   3,   0 },    141.13474,   0.00001,     0.072 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -1,   0,   1 },    165.42097,   0.00003,     7.335 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0,   0 },    332.88235,   0.00004,     0.840 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,   2,   0,   0 },    329.94602,   0.00004,     0.042 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -4,   2,   0 },    177.96133,   0.00001,     0.042 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -4,   3,   0 },    177.97122,   0.00001,     0.096 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   0,   0 },    180.75768,   0.00008,     0.039 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   1,   0 },    180.78341,   0.00012,     0.082 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   2,   0 },    344.65033,   0.00002,     0.896 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   3,   0 },    344.50089,   0.00002,     0.070 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   0,   1 },     31.74822,   0.00002,     2.529 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   1,  -1 },    200.99808,   0.00005,     0.076 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   3,  -1 },     28.43852,   0.00002,     0.075 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0 },    178.63366,   0.00002,     0.037 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0 },      5.18862,   0.00005,     0.072 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0 },    161.58381,   0.00002,     1.519 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0 },      3.62272,   0.00006,     0.079 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,  -2,   0 },    174.19052,   0.00001,     0.411 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   2,   0,   0 },    354.19052,   0.00001,     0.042 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,  -1,   0 },     29.53789,   0.00002,     0.026 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   0 },     32.06495,   0.00021,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,   0 },     27.15722,   0.00014,     0.086 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   2,   0 },    176.05224,   0.00001,     0.607 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   3,   0 },    212.38163,   0.00003,     0.067 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   1 },    107.78226,   0.00002,     1.079 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0 },     20.11931,   0.00001,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0 },     21.96217,   0.00023,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0 },     27.88777,   0.00143,     7.890 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0 },     22.32573,   0.00023,     0.076 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0 },    212.74096,   0.00002,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   0 },     29.52490,   0.00024,     0.088 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   0 },     26.90851,   0.00013,     0.041 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   1,   0 },     27.35939,   0.00001,     0.026 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,  -2,   0 },    244.54368,   0.00001,     0.020 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   0 },     56.72033,   0.00006,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,   0 },     52.26406,   0.00006,     0.090 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   2,   0 },     59.15026,   0.00001,     0.459 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   1 },     56.86728,   0.00002,     0.686 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0 },     64.86936,   0.00009,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0 },     62.19489,   0.00031,     2.470 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0 },     67.47174,   0.00006,     0.073 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -3,   0 },     52.19123,   0.00001,     0.068 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,   0 },     44.95746,   0.00001,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,   0 },    240.11623,   0.00005,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   2,   0 },    236.84984,   0.00003,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   0 },    274.30328,   0.00009,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   1,   0 },     88.99618,   0.00006,     0.095 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   2,   0 },     93.29190,   0.00002,     0.369 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0 },    286.68567,   0.00002,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0 },    266.58059,   0.00005,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0 },    216.68098,   0.00005,     1.068 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0 },    264.14315,   0.00004,     0.070 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0 },     86.93239,   0.00002,     0.036 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,  -1 },    262.15176,   0.00003,     8.537 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   0 },    266.07959,   0.00003,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -3,   0 },     90.39785,   0.00002,     0.071 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -2,   0 },     86.94869,   0.00003,     1.194 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -1,   0 },    260.63845,   0.00001,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   0 },     91.13093,   0.00002,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   4,  -2,   0 },    266.61500,   0.00001,     0.042 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,   0 },    301.55063,   0.00003,     0.043 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   2,   0 },    302.55912,   0.00003,     0.308 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0 },    114.61244,   0.00002,     0.085 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0 },    275.88151,   0.00003,     0.681 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0 },    114.61244,   0.00002,     0.068 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0 },    114.92635,   0.00001,     0.036 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,  -1 },    218.12144,   0.00004,     2.412 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -3,   0 },    320.70582,   0.00001,     0.074 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -2,   0 },    296.61057,   0.00006,     3.271 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,   0 },    299.19743,   0.00005,     0.077 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,   0,   0 },    105.97857,   0.00007,     0.038 },
 { {   0,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 },    261.05035,   0.00002,     0.634 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   0,   0,   0 },    313.23121,   0.00001,     0.572 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -2,   0 },    132.81161,   0.00002,    38.763 },
 { {   0,   0,   2,   0,  -5,   5,   0,   0,   2,  -1,   0 },    133.12360,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   0 },    251.45657,   0.00003,     0.043 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   1,   0 },    251.45629,   0.00003,     0.102 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,  -1,   0 },    249.64464,   0.00003,     0.086 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 },    251.14760,   0.00007,     0.601 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   0 },     69.64464,   0.00013,     0.076 },
 { {   0,   0,   2,   0,  -4,   3,   0,   0,   2,  -2,   0 },    320.98065,   0.00002,   106.494 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,  -2,   0,   0 },    254.34594,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   0,   0,   0 },    254.28797,   0.00004,     0.546 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -2,   0 },     74.60049,   0.00001,    17.093 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,   0 },    254.54674,   0.00005,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   0 },    243.08832,   0.00009,     0.043 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   0 },    243.08832,   0.00010,     0.103 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -2,   0 },    239.23033,   0.00001,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   0 },    239.68341,   0.00052,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 },    240.99668,   0.00028,     0.572 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   0 },    242.23723,   0.00005,     0.067 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   0 },     60.43108,   0.00006,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   0 },     59.31457,   0.00157,    37.133 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   0 },     59.33774,   0.00244,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   0 },     59.33860,   0.00016,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -3,   0 },     58.71742,   0.00002,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   0 },     59.31176,   0.00005,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -1,   0 },     59.33580,   0.00002,     0.026 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },    310.39487,   0.00003,     0.044 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   1,   0 },    310.39487,   0.00003,     0.104 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   0 },    176.54929,   0.00002,     0.088 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 },    326.97840,   0.00004,     0.546 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -2,   0 },    111.11498,   0.00002,    17.430 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   0 },    356.54914,   0.00011,     0.075 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   0,   0,   0 },    290.90162,   0.00001,     0.573 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -2,   0 },    110.80632,   0.00002,    35.635 },
 { {   0,   0,   2,   0,  -1,  -5,   0,   0,   2,  -1,   0 },    110.81494,   0.00003,     0.076 },
 { {   0,   0,   2,   0,  -1,  -4,   0,   0,   2,  -2,   0 },    358.77389,   0.00002,   169.910 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,  -2,   0,   0 },    300.55385,   0.00002,     0.044 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 },    295.79489,   0.00002,     0.522 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,  -1 },    184.18359,   0.00002,     1.160 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -3,   0 },     50.69297,   0.00002,     0.076 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   2,  -1,   0 },    304.94416,   0.00003,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,   0 },    284.91052,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   0,   0,   0 },    284.34066,   0.00004,     0.546 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,   0 },    284.69684,   0.00006,     0.075 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,  -2,   0,   0 },     34.99159,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   0,  -3,   0,   0,   0,   0,   0 },     35.02135,   0.00002,     0.527 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,  -2,   0 },     50.01487,   0.00003,     0.020 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },     82.81407,   0.00004,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   0 },    331.60891,   0.00003,     0.105 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   1 },    246.74124,   0.00004,     0.327 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0 },    294.88030,   0.00002,     0.041 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    131.27868,   0.00004,     0.089 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    148.35154,   0.00018,     0.500 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0 },     77.06584,   0.00003,     0.066 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   2,   0 },    247.63825,   0.00001,     0.035 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,  -1 },    317.71087,   0.00003,     1.057 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    354.21017,   0.00003,     0.077 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    317.39646,   0.00008,     4.425 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },      2.83711,   0.00017,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    312.80638,   0.00020,     0.037 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   1,   0 },    312.90814,   0.00002,     0.025 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -2 },     77.60640,   0.00002,    43.055 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,  -2,   0,   0 },     60.54610,   0.00002,     0.044 },
 { {   0,   0,   2,   0,   2,  -5,   0,   0,   0,   0,   0 },     59.66652,   0.00001,     0.500 },
 { {   0,   0,   2,   0,   3,   0,   0,   0,   2,  -1,   0 },    247.75907,   0.00004,     0.073 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,  -2,   1,   0 },      0.06117,   0.00001,     0.112 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   0,  -1,   0 },     13.32593,   0.00001,     0.093 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0 },    187.65676,   0.00006,     0.395 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   0,   1,   0 },     13.32593,   0.00001,     0.063 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -3,   0 },    172.86138,   0.00001,     0.080 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -1,   0 },    182.33364,   0.00006,     0.071 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,   0,   0 },    188.44995,   0.00009,     0.037 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,  -2,   0,   0 },     28.05557,   0.00002,     0.046 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,  -2,   2,   0 },    208.05557,   0.00002,     0.207 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   0,  -1,   0 },    220.50233,   0.00002,     0.098 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   0,   0,   0 },     46.04182,   0.00003,     0.326 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   0,   2,   0 },    221.99679,   0.00001,     0.034 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,  -2,   0 },     42.17129,   0.00001,     0.776 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,  -1,   0 },     41.14905,   0.00014,     0.069 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,   0,   0 },    220.81605,   0.00010,     0.036 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   0,  -1,   0 },    250.95342,   0.00002,     0.104 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   0,   0,   0 },     91.28349,   0.00003,     0.278 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   0,   1,   0 },    251.08532,   0.00003,     0.059 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   2,  -1,   0 },     70.56342,   0.00007,     0.066 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   2,   0,   0 },    252.89429,   0.00003,     0.035 },
 { {   0,   0,   2,   4,   0,   0,   0,   0,   0,   0,   0 },    100.33227,   0.00004,     0.242 },
 { {   0,   0,   2,   4,   0,   0,   0,   0,   0,   2,   0 },    280.13752,   0.00002,     0.033 },
 { {   0,   0,   2,   4,   0,   0,   0,   0,   2,   2,   0 },    100.45679,   0.00002,     0.018 },
 { {   0,   0,   2,   5,   0,   0,   0,   0,   0,   0,   0 },    309.51071,   0.00004,     0.215 },
 { {   0,   0,   2,   5,   0,   0,   0,   0,   2,  -1,   0 },    309.51071,   0.00001,     0.062 },
 { {   0,   0,   2,   6,   0,   0,   0,   0,   2,  -3,   0 },    158.39827,   0.00001,     0.102 },
 { {   0,   0,   2,   6,   0,   0,   0,   0,   2,  -1,   0 },    158.39827,   0.00001,     0.060 },
 { {   0,   0,   2,   6,   0,   0,   0,   0,   2,   0,   0 },    338.21394,   0.00002,     0.033 },
 { {   0,   0,   2,   7,   0,   0,   0,   0,   2,   0,   0 },      7.11782,   0.00002,     0.033 },
 { {   0,   0,   3, -13,   0,   0,   0,   0,  -2,   0,   0 },    143.15120,   0.00003,     0.035 },
 { {   0,   0,   3, -12,   0,   0,   0,   0,  -2,   0,   0 },    352.20688,   0.00008,     0.036 },
 { {   0,   0,   3, -12,   0,   0,   0,   0,  -2,   1,   0 },    172.42642,   0.00003,     0.067 },
 { {   0,   0,   3, -11,   0,   0,   0,   0,  -2,   0,   0 },    201.23300,   0.00003,     0.036 },
 { {   0,   0,   3, -11,   0,   0,   0,   0,  -2,   1,   0 },    201.49178,   0.00002,     0.070 },
 { {   0,   0,   3, -11,   0,   0,   0,   0,   0,   0,   0 },    201.14616,   0.00001,     0.351 },
 { {   0,   0,   3, -10,   0,   0,   0,   0,  -2,   0,   0 },     50.62099,   0.00001,     0.037 },
 { {   0,   0,   3, -10,   0,   0,   0,   0,  -2,   0,   2 },    230.31007,   0.00001,     4.778 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   0,   0 },    267.22623,   0.00007,     0.038 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   0,   2 },    259.22934,   0.00002,     3.102 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   1,   0 },     60.12555,   0.00002,     0.075 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   2,   0 },      8.89270,   0.00003,    89.761 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,  -1,   0 },    259.23133,   0.00003,     0.066 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,   0,   0 },    248.19879,   0.00001,     0.560 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,   1,   0 },    259.23133,   0.00003,     0.087 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   0,   0 },    130.52072,   0.00002,     0.038 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   1,   0 },    289.48661,   0.00005,     0.079 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   2,   0 },    288.04747,   0.00002,     1.921 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   3,   0 },    289.70917,   0.00004,     0.073 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -1,   0,   1 },    103.56677,   0.00002,     5.007 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,  -1,   0 },    288.23198,   0.00005,     0.069 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,   0,   0 },    282.26013,   0.00010,     0.798 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,   1,   0 },    287.93278,   0.00004,     0.083 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,  -1,   0 },    108.23198,   0.00001,     0.098 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,   0,   0 },    120.70798,   0.00005,     0.043 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   0,   0 },    105.00688,   0.00012,     0.999 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,   0 },    104.66914,   0.00004,     0.082 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,   0 },    105.04585,   0.00012,     0.095 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   0 },    104.95275,   0.00010,     0.042 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   0,   0 },    159.67196,   0.00008,     0.039 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   1,   0 },    151.50178,   0.00012,     0.082 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   2,   0 },      1.12049,   0.00002,     0.950 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   3,   0 },    177.60857,   0.00001,     0.070 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,   1 },    342.97850,   0.00001,     3.012 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,  -1 },    183.57380,   0.00006,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,  -1 },    289.29003,   0.00030,   650.233 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0 },    318.18261,   0.00003,     0.072 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0,   0 },    306.52929,   0.00003,     1.385 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   1,   0 },    323.95632,   0.00006,     0.080 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -3,   0 },    329.52324,   0.00003,     0.063 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -1,   0 },    149.59264,   0.00005,     0.093 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,   0,   0 },    329.98792,   0.00004,     0.042 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   0,   0 },    214.47477,   0.00001,     0.040 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   1,   0 },     12.90321,   0.00002,     0.086 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -2,   2,   0 },    214.47477,   0.00001,     0.631 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   1 },     22.38730,   0.00003,     1.158 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0 },    147.68456,   0.00001,     0.037 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0 },    349.81405,   0.00015,     0.074 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0 },    357.99442,   0.00045,     5.260 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0 },    349.81405,   0.00015,     0.077 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,   0 },     23.78522,   0.00001,     0.089 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,   0 },    359.68509,   0.00005,     0.041 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   2,   0 },    178.65218,   0.00001,     0.020 },
 { {   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0,   0 },     55.23630,   0.00004,    46.520 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   0 },     29.83620,   0.00017,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   1,   0 },    199.73780,   0.00001,     0.090 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   1 },     24.26204,   0.00002,     0.717 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0 },     27.81879,   0.00001,     0.038 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0 },     52.63547,   0.00005,     0.077 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0 },     32.63386,   0.00021,     2.928 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -2 },    203.99216,   0.00004,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0 },     53.10919,   0.00005,     0.074 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2 },     29.60034,   0.00005,   123.995 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0 },    110.05993,   0.00001,     0.037 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -2 },    211.81735,   0.00002,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,  -1 },    234.94500,   0.00002,     1.404 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -3,   0 },    206.15626,   0.00003,     0.068 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -2,   0 },    207.75557,   0.00001,     0.698 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,   0 },     22.59276,   0.00005,     0.085 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,   0 },     26.77203,   0.00007,     0.040 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   4,  -3,   0 },     26.78999,   0.00001,     0.100 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   0 },     47.03307,   0.00007,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,   0 },     49.81689,   0.00010,     0.094 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   3,   0 },     59.99841,   0.00001,     0.063 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0 },     85.88044,   0.00004,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0 },     56.51311,   0.00005,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0 },     59.52230,   0.00020,     1.145 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,  -1,   0 },     47.68292,   0.00010,    72.752 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,  -1 },    243.80530,   0.00002,     5.540 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },     56.28075,   0.00004,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -3,   0 },     95.33721,   0.00001,     0.071 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,  -2,   0 },    236.88426,   0.00002,     1.110 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,   0 },    246.57573,   0.00005,     0.039 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   0 },    273.19126,   0.00009,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,   0 },     93.00443,   0.00004,     0.099 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0 },    253.95451,   0.00003,     0.040 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0 },    307.74518,   0.00003,     0.712 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0 },     96.81662,   0.00001,     0.068 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0 },    265.89225,   0.00002,     0.036 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -3,   0 },     95.26276,   0.00004,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -2,   0 },    265.92080,   0.00005,     2.710 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,   0 },    260.80368,   0.00009,     0.078 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,   0 },    103.30263,   0.00004,     0.038 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   0 },    296.83446,   0.00003,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0 },    288.47673,   0.00004,     0.516 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0 },    305.58739,   0.00003,     0.066 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0 },     49.49890,   0.00002,     0.035 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,  -1 },    274.99372,   0.00001,     1.133 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -3,   0 },    133.94177,   0.00001,     0.076 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -2,   0 },    128.33238,   0.00007,     6.150 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,  -1 },     21.91190,   0.00001,   241.391 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    308.10314,   0.00012,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,  -2 },    294.58170,   0.00001,     5.851 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },    118.74589,   0.00013,     0.037 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,  -2,   1,   0 },    331.12600,   0.00003,     0.111 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0 },    173.72547,   0.00002,     0.405 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -3,   0 },    249.81661,   0.00001,     0.080 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -2,   0 },    333.16467,   0.00003,     1.440 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -1,   0 },    131.02930,   0.00004,     0.072 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,   0,   0 },    339.72842,   0.00012,     0.037 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,  -2,   0,   0 },    224.98950,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 },    231.69834,   0.00002,     0.376 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -3,   0 },     60.52726,   0.00001,     0.081 },
 { {   0,   0,   3,   0,  -4,   0,   0,   0,   2,  -1,   0 },    240.52726,   0.00001,     0.071 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,  -2,   0,   0 },    137.35840,   0.00002,     0.045 },
 { {   0,   0,   3,   0,  -3,   0,   0,   0,   0,  -1,   0 },    136.40163,   0.00002,     0.095 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -2,  -1 },    193.67885,   0.00027,   277.143 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,  -1 },     14.85537,   0.00006,     0.075 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   1,   0 },    357.65517,   0.00002,     0.118 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   3,   0 },    177.65517,   0.00001,     0.055 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -2,   0 },    177.65517,   0.00001,     0.043 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },     21.94975,   0.00003,     0.098 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    176.53243,   0.00009,     0.333 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   1,   0 },     59.47798,   0.00003,     0.062 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -3,   0 },    191.19058,   0.00001,     0.083 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },     33.29375,   0.00002,     0.816 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    187.32076,   0.00005,     0.069 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,  -2 },    323.97184,   0.00002,     1.121 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     45.96759,   0.00003,     0.036 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   1,   0 },    147.18724,   0.00001,     0.024 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0 },     36.29038,   0.00007,     0.283 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   0,   2,   0 },    222.10586,   0.00003,     0.033 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   2,  -3,   0 },    221.22403,   0.00001,     0.087 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   2,  -1,   0 },    219.93362,   0.00006,     0.067 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   2,   0,   0 },    219.79645,   0.00009,     0.035 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   0,   0,   0 },    201.25050,   0.00003,     0.246 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,  -2,   0 },    252.19107,   0.00003,     0.437 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,  -1,   0 },    251.01817,   0.00002,     0.064 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,   0,   0 },    252.26165,   0.00004,     0.035 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   0,  -1,   0 },    100.39357,   0.00001,     0.115 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   0,   0,   0 },    100.28457,   0.00002,     0.218 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   0,   1,   0 },    100.39357,   0.00001,     0.056 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,  -1,   0 },    100.09760,   0.00002,     0.062 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,   0,   0 },    100.16217,   0.00003,     0.034 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,   2,   0 },    100.39357,   0.00001,     0.018 },
 { {   0,   0,   3,   4,   0,   0,   0,   0,   2,  -1,   0 },    309.19641,   0.00005,     0.060 },
 { {   0,   0,   3,   5,   0,   0,   0,   0,   2,   0,   0 },    158.05995,   0.00005,     0.033 },
 { {   0,   0,   4, -14,   0,   0,   0,   0,  -2,   0,   0 },    143.07536,   0.00001,     0.035 },
 { {   0,   0,   4, -14,   0,   0,   0,   0,  -2,   1,   0 },    323.07536,   0.00002,     0.067 },
 { {   0,   0,   4, -14,   0,   0,   0,   0,  -2,   2,   0 },    143.07536,   0.00001,     0.599 },
 { {   0,   0,   4, -13,   0,   0,   0,   0,  -2,   0,   0 },    352.13434,   0.00005,     0.036 },
 { {   0,   0,   4, -13,   0,   0,   0,   0,  -2,   2,   0 },    352.16306,   0.00003,     0.879 },
 { {   0,   0,   4, -12,   0,   0,   0,   0,  -2,   0,   0 },    201.17149,   0.00003,     0.037 },
 { {   0,   0,   4, -12,   0,   0,   0,   0,  -2,   0,   2 },     21.09777,   0.00001,     3.667 },
 { {   0,   0,   4, -12,   0,   0,   0,   0,  -2,   2,   0 },     21.09777,   0.00001,     1.650 },
 { {   0,   0,   4, -12,   0,   0,   0,   0,   0,  -1,   0 },     21.09777,   0.00001,     0.064 },
 { {   0,   0,   4, -12,   0,   0,   0,   0,   0,   1,   0 },     21.09777,   0.00001,     0.092 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   0,   0 },     50.45255,   0.00003,     0.038 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   0,   2 },    230.13026,   0.00002,     3.861 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   1,   0 },    237.07205,   0.00008,     0.075 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   2,   0 },     48.75706,   0.00006,    13.421 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   3,   0 },    166.45804,   0.00001,     0.076 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,   0,  -1,   0 },    230.17678,   0.00003,     0.066 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,   0,   1,   0 },    230.17678,   0.00003,     0.088 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,  -2,   0 },    259.08297,   0.00001,     0.019 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,   0,   0 },    268.70427,   0.00006,     0.038 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,   1,   0 },    259.08297,   0.00002,     0.078 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,   2,   0 },    143.42760,   0.00001,     2.187 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,  -1,   0 },    259.11509,   0.00004,     0.069 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,   0,   0 },    263.39263,   0.00005,     0.759 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,   1,   0 },    259.11509,   0.00004,     0.084 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   2,   0,   0 },     91.86882,   0.00002,     0.043 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -2,   0,   0 },    107.45379,   0.00002,     0.039 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -1,   0,   1 },    325.53234,   0.00003,     3.723 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   0,   0,   0 },    291.67780,   0.00010,     1.274 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -3,   0 },    300.65416,   0.00001,     0.063 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -2,   0 },    120.50641,   0.00003,     0.391 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -1,   0 },    120.65416,   0.00001,     0.093 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,   0,   0 },    120.50641,   0.00003,     0.042 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   0,   0 },    146.95106,   0.00015,     0.040 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   1,   0 },    146.75328,   0.00005,     0.085 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   2,   0 },    210.19694,   0.00001,     0.658 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   3,   0 },    324.06782,   0.00003,     0.068 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0 },    331.55717,   0.00009,     0.074 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0 },    326.76497,   0.00020,     3.945 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,   0 },    331.55717,   0.00009,     0.077 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   2,   0 },    329.56744,   0.00001,     0.038 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,  -1,   0 },    150.64864,   0.00004,     0.089 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,   0 },    331.16955,   0.00003,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   0,   0 },      1.93148,   0.00002,     0.020 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   2,   0 },      1.93144,   0.00002,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   0 },      1.87056,   0.00011,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   0 },      1.97038,   0.00107,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   2 },      1.93148,   0.00003,     0.475 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   0 },      2.12715,   0.00167,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   0 },      1.93132,   0.00007,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   0,   0 },    181.93182,   0.00001,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -1,   2,   0 },    181.93182,   0.00001,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   0 },      1.25858,   0.00012,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   0 },      1.32715,   0.00202,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   0 },      2.27463,   0.01056,  1783.359 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   0 },      1.32745,   0.00202,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   0 },      1.25889,   0.00012,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,  -2,   0 },    181.93233,   0.00001,     0.071 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   1,   0,   0 },    181.93233,   0.00001,     0.081 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   0 },      1.93283,   0.00007,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   0 },      2.12840,   0.00168,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -2 },      1.93267,   0.00003,     0.474 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   0 },      1.97140,   0.00107,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   0 },      1.87151,   0.00011,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -2,   0 },      1.93272,   0.00002,     0.044 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,   0,   0 },      1.93267,   0.00002,     0.020 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -4,   2,   0 },    358.70257,   0.00001,     0.044 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   0 },      2.86841,   0.00003,     0.041 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   1,   0 },    353.78927,   0.00003,     0.089 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   2,   0 },    178.87609,   0.00003,     0.487 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0 },    147.64076,   0.00001,     0.038 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0 },    359.73937,   0.00006,     0.077 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0 },      7.37619,   0.00026,     3.594 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0 },    359.73937,   0.00006,     0.074 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0 },     36.67531,   0.00002,     0.037 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -3,   0 },    179.15460,   0.00002,     0.068 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -2,   0 },      5.86641,   0.00001,     0.669 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,   0 },    358.67964,   0.00003,     0.085 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,   0 },    355.62847,   0.00003,     0.040 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   2,   0 },    178.59671,   0.00001,     0.019 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   0 },     30.59698,   0.00017,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   1,   0 },    352.76144,   0.00003,     0.094 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   2,   0 },     33.03988,   0.00001,     0.387 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   3,   0 },     90.41796,   0.00001,     0.063 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0 },     25.37719,   0.00001,     0.080 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0 },     21.76312,   0.00009,     1.235 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -3,   1 },     80.38363,   0.00001,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,  -1,   1 },    260.38363,   0.00001,     0.075 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,  -1 },     40.40567,   0.00001,     4.101 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   0 },     27.71847,   0.00003,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -2,   0 },    212.03690,   0.00001,     1.037 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,  -1,   0 },    206.39785,   0.00002,     0.081 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,   0 },     27.06190,   0.00010,     0.039 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   2,   0 },     27.71744,   0.00002,     0.019 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   0,   0 },     52.12984,   0.00002,     0.043 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -2,   1,   0 },    247.83196,   0.00002,     0.099 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0 },     66.65496,   0.00001,     0.040 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0 },    239.96499,   0.00002,     0.084 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0 },    242.84546,   0.00017,     0.745 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0 },    238.63301,   0.00003,     0.069 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0 },    213.03374,   0.00001,     0.036 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,  -1 },    142.70293,   0.00002,     3.475 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -3,   0 },    239.19717,   0.00005,     0.073 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -2,   0 },    237.91376,   0.00005,     2.313 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -1,   0 },    236.82050,   0.00005,     0.078 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,   0 },    234.83146,   0.00004,     0.038 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   2,   0 },     56.85308,   0.00002,     0.019 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   0 },    281.63410,   0.00007,     0.044 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   2,   0 },     86.59059,   0.00001,     0.274 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0 },    295.98026,   0.00003,     0.041 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0 },    276.79312,   0.00008,     0.534 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,  -1 },    227.52682,   0.00001,     1.220 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -3,   0 },    111.24471,   0.00003,     0.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -2,   0 },    277.13350,   0.00008,    10.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   0 },    267.14323,   0.00013,     0.075 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,  -2 },    265.60813,   0.00002,     4.269 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,   0 },     85.13241,   0.00003,     0.038 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   4,  -3,   0 },    272.27232,   0.00001,     0.086 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,  -2,   0,   0 },    299.53021,   0.00003,     0.045 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0 },    303.00236,   0.00002,     0.041 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0 },    302.52133,   0.00004,     0.092 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0 },    176.40117,   0.00002,     0.416 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0 },    302.53869,   0.00003,     0.064 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0 },    147.71024,   0.00003,     0.035 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   1,   0,  -1 },    303.81104,   0.00003,     0.740 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -3,   0 },     40.07085,   0.00002,     0.079 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -2,   0 },    295.10924,   0.00002,     1.585 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -1,   0 },    220.07085,   0.00002,     0.072 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,  -2,   0,   0 },    330.70292,   0.00002,     0.046 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0 },    333.33609,   0.00001,     0.097 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0 },    159.66135,   0.00005,     0.341 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -3,   0 },    158.00720,   0.00003,     0.083 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -2,   0 },    349.46373,   0.00001,     0.860 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -1,   0 },    151.40553,   0.00007,     0.069 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,   0,   0 },    333.19731,   0.00012,     0.036 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   4,  -3,   0 },    333.33609,   0.00001,     0.079 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   0,  -1,   0 },     12.94082,   0.00001,     0.102 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0 },    181.98217,   0.00006,     0.288 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   0,   1,   0 },     12.94082,   0.00001,     0.060 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   2,  -3,   0 },     10.92161,   0.00001,     0.086 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   2,  -2,   0 },     10.92161,   0.00001,     0.590 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   2,  -1,   0 },    192.04271,   0.00003,     0.067 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   1,   0,  -1 },    322.98319,   0.00002,     0.349 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.50175,   0.00003,     0.048 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,  -2,   2,   0 },     31.06548,   0.00001,     0.173 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0 },     23.44561,   0.00011,     0.250 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   0,   2,   0 },    210.86789,   0.00005,     0.033 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -3,   0 },    109.97495,   0.00001,     0.091 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },     23.03359,   0.00013,     0.065 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    105.11173,   0.00006,     0.035 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   2,   0 },     20.41579,   0.00002,     0.018 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   0,  -1,   0 },     71.19172,   0.00002,     0.115 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   0,   0,   0 },     71.53393,   0.00004,     0.221 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   2,  -1,   0 },    250.99133,   0.00001,     0.062 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   2,   0,   0 },    250.95112,   0.00001,     0.034 },
 { {   0,   0,   4,   2,   0,   0,   0,   0,   0,  -1,   0 },    100.28024,   0.00001,     0.122 },
 { {   0,   0,   4,   2,   0,   0,   0,   0,   0,   1,   0 },    100.28024,   0.00001,     0.055 },
 { {   0,   0,   4,   2,   0,   0,   0,   0,   2,  -1,   0 },    280.28024,   0.00001,     0.060 },
 { {   0,   0,   4,   2,   0,   0,   0,   0,   2,   0,   0 },    100.28024,   0.00002,     0.034 },
 { {   0,   0,   5, -15,   0,   0,   0,   0,  -2,   2,   0 },    143.01562,   0.00002,     0.832 },
 { {   0,   0,   5, -14,   0,   0,   0,   0,  -2,   0,   0 },    352.10280,   0.00004,     0.037 },
 { {   0,   0,   5, -14,   0,   0,   0,   0,  -2,   2,   0 },    352.10280,   0.00004,     1.494 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   0,   0 },    201.10046,   0.00003,     0.038 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   0,   2 },    201.03874,   0.00001,     5.112 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   1,   0 },    197.87401,   0.00003,     0.075 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   3,   0 },     16.44235,   0.00002,     0.076 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,   0,  -1,   0 },     21.03874,   0.00001,     0.066 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,   0,   1,   0 },     21.03874,   0.00001,     0.088 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,  -2,   1,   0 },    232.77012,   0.00004,     0.078 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,  -2,   2,   0 },     50.25016,   0.00002,     2.539 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,   0,  -1,   0 },    230.06752,   0.00003,     0.068 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,   0,   0,   0 },     66.21921,   0.00003,     0.725 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,   0,   1,   0 },    230.06752,   0.00003,     0.084 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -2,   0,   0 },    273.06313,   0.00004,     0.039 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -2,   2,   0 },    101.36250,   0.00003,     1.081 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   0,   0 },     78.89469,   0.00003,     0.076 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   0,   1 },    356.51157,   0.00001,     4.872 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   1,   0 },     78.89290,   0.00001,    25.987 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   0,  -1,   0 },    259.36738,   0.00002,     0.071 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   0,   0,   0 },    264.77180,   0.00007,     1.179 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   0,   1,   0 },    259.36738,   0.00002,     0.081 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,  -2,   0 },     91.65610,   0.00001,     0.381 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,   2,   0,   0 },     91.65610,   0.00001,     0.042 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   0,   0 },    286.28470,   0.00005,     0.040 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   1,   0 },    287.28788,   0.00002,     0.085 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,  -2,   2,   0 },    283.37866,   0.00001,     0.686 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,  -2,   0 },    107.28788,   0.00001,     0.037 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0 },    124.31783,   0.00003,     0.074 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   0,   0 },    264.31520,   0.00006,     3.156 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   1,   0 },    124.31783,   0.00003,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   0,   0 },    143.29495,   0.00006,     0.041 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,   0 },    326.55513,   0.00008,     0.089 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   3,   0 },    133.57699,   0.00001,     0.066 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0 },    142.56340,   0.00007,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0 },    342.32447,   0.00018,     4.655 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0 },    142.56340,   0.00007,     0.074 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   2,   0 },    329.49512,   0.00001,     0.037 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,   0 },    326.23040,   0.00003,     0.085 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,   0 },    329.69876,   0.00004,     0.040 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   0 },      5.38431,   0.00004,     0.042 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   2,   0 },    165.10762,   0.00001,     0.397 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0 },      3.15965,   0.00019,     1.340 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -3,   1 },    101.51607,   0.00002,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -2,   1 },    118.34073,   0.00003,    38.061 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   1 },    231.88326,   0.00004,     0.075 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,   0,  -1 },    298.34055,   0.00003,     3.255 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -2,   0 },    178.55479,   0.00001,     0.973 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,  -1,   0 },    178.39567,   0.00003,     0.082 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,   0 },    358.55479,   0.00001,     0.039 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   2,   0 },    178.55479,   0.00001,     0.019 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   0 },    259.64233,   0.00010,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   0 },    259.85523,   0.00012,     0.095 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,   0 },    259.81893,   0.00004,     0.082 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0,   0 },    261.17260,   0.00012,     1.001 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,   0 },    260.72877,   0.00002,     0.070 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   0 },     33.56153,   0.00009,     0.043 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,   0 },    215.43499,   0.00008,     0.098 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   2,   0 },     30.67198,   0.00004,     0.328 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   3,   0 },    215.58513,   0.00001,     0.061 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0 },    207.52832,   0.00020,     0.782 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0 },    207.66575,   0.00001,     0.069 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0 },    207.28620,   0.00002,     0.036 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   0,  -1 },    161.22630,   0.00001,     4.456 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -3,   0 },    268.06064,   0.00001,     0.073 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -2,   0 },    200.71945,   0.00002,     2.017 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,   0 },    218.72430,   0.00009,     0.078 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,   0 },     26.61575,   0.00009,     0.038 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   4,  -2,   0 },    207.88875,   0.00001,     0.041 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,  -2,   0 },    244.40360,   0.00001,     0.020 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,   0 },     64.69611,   0.00003,     0.044 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   1,   0 },    244.91498,   0.00003,     0.103 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   3,   0 },    243.81196,   0.00001,     0.059 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0 },     68.41747,   0.00007,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0 },    242.27363,   0.00013,     0.553 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0 },    242.32784,   0.00001,     0.066 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   0,  -1 },    236.36559,   0.00003,     1.323 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -3,   0 },    182.05246,   0.00002,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,   0 },     66.47192,   0.00019,    27.873 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   0 },    246.46578,   0.00041,     0.075 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,  -2 },    236.70562,   0.00002,     3.360 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,   0 },    236.15567,   0.00005,     0.038 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,  -2,   0 },     93.42860,   0.00001,     0.020 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,   0 },     92.26078,   0.00004,     0.045 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   2,   0 },     92.10141,   0.00004,     0.243 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0 },    290.49954,   0.00003,     0.041 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0 },    277.83917,   0.00003,     0.092 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0 },     52.86385,   0.00003,     0.427 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0 },    281.10119,   0.00002,     0.064 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0 },     85.22804,   0.00001,     0.035 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   1,   0,  -1 },    272.83707,   0.00003,     0.777 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -3,   0 },    273.01157,   0.00002,     0.079 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,   0 },    270.26662,   0.00004,     0.072 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,  -2 },     85.20941,   0.00001,     4.273 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,   0,   0 },     87.09084,   0.00007,     0.037 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   0,   0 },    122.22801,   0.00002,     0.046 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   1,   0 },    302.30607,   0.00006,     0.116 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,  -1,   0 },    302.34491,   0.00002,     0.096 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0 },    124.32586,   0.00008,     0.348 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -3,   0 },    315.08906,   0.00002,     0.082 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -2,   0 },    134.15316,   0.00001,     0.910 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -1,   0 },    122.34473,   0.00012,     0.070 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,   0,   0 },    301.89835,   0.00006,     0.036 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,  -2,   0,   0 },    325.89748,   0.00002,     0.047 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0 },     84.44459,   0.00002,     0.043 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0 },    340.06329,   0.00002,     0.102 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0 },    184.22073,   0.00003,     0.294 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   2,  -3,   0 },    183.53226,   0.00001,     0.086 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   2,  -2,   0 },     40.51555,   0.00001,     0.613 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   2,  -1,   0 },    160.06349,   0.00011,     0.067 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   2,   0,   0 },    172.13504,   0.00007,     0.036 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   0,   0 },    356.86751,   0.00001,     0.048 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   2,   0 },    357.28692,   0.00002,     0.175 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   0,  -1,   0 },     12.85143,   0.00001,     0.107 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0,   0 },    191.25170,   0.00006,     0.254 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   0,   1,   0 },     12.85143,   0.00001,     0.058 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   2,  -2,   0 },     10.26692,   0.00001,     0.462 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   2,  -1,   0 },    191.92163,   0.00005,     0.065 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,   2,   0,   0 },    190.90373,   0.00005,     0.035 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   0,  -1,   0 },     41.95482,   0.00002,     0.114 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0 },     44.56689,   0.00004,     0.224 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   0,   1,   0 },     41.95482,   0.00002,     0.056 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   0,   2,   0 },    221.59809,   0.00002,     0.032 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   2,  -3,   0 },    221.51400,   0.00003,     0.095 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   2,  -1,   0 },     41.86654,   0.00003,     0.063 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   2,   0,   0 },    294.08765,   0.00001,     0.034 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   2,   2,   0 },     41.95482,   0.00002,     0.018 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,  -2,   0 },    109.78920,   0.00002,     0.046 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0 },     78.32317,   0.00009,     0.200 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   1,   0 },    280.24748,   0.00003,     0.055 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   2,   0 },    251.32894,   0.00001,     0.032 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   3,   0 },    109.78920,   0.00002,     0.022 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -3,   0 },     71.35877,   0.00002,     0.100 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -2,   0 },    294.12926,   0.00001,     0.310 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -1,   0 },    115.65543,   0.00003,     0.061 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   0,  -2 },    108.01509,   0.00001,     0.346 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   0,   0 },    259.85727,   0.00016,     0.034 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   1,   0 },    108.01510,   0.00002,     0.023 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   2,   0 },    103.54531,   0.00002,     0.018 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   3,   0 },    109.78920,   0.00001,     0.014 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   4,  -1,   0 },    289.78920,   0.00001,     0.024 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   4,   0,   0 },    108.01510,   0.00004,     0.018 },
 { {   0,   0,   5,   1,   0,   0,   0,   0,   2,  -2,   0 },     95.49119,   0.00001,     0.266 },
 { {   0,   0,   5,   1,   0,   0,   0,   0,   2,   2,   0 },    280.09151,   0.00001,     0.018 },
 { {   0,   0,   6, -16,   0,   0,   0,   0,  -2,   0,   0 },    322.96279,   0.00001,     0.037 },
 { {   0,   0,   6, -16,   0,   0,   0,   0,  -2,   1,   0 },    143.05494,   0.00007,     0.071 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   0,   0 },    352.01632,   0.00005,     0.037 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   0,   2 },    171.96508,   0.00001,     7.561 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   1,   0 },    176.22258,   0.00002,     0.074 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   2,   0 },    352.03847,   0.00007,     4.969 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   3,   0 },    356.22258,   0.00002,     0.077 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,   0,   0,   0 },    171.96508,   0.00001,     0.506 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   0,   0 },    201.01489,   0.00007,     0.038 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   1,   0 },    229.65487,   0.00004,     0.077 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   3,   0 },    205.64693,   0.00004,     0.074 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,   0,  -1,   0 },     20.96378,   0.00002,     0.068 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,   0,   0,   0 },     17.61008,   0.00002,     0.693 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,   0,   1,   0 },     20.96378,   0.00002,     0.085 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -2,   0,   0 },    229.78411,   0.00002,     0.039 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -2,   1,   0 },     50.30503,   0.00001,     0.081 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   0,   0 },    229.90397,   0.00004,     0.075 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   1,   0 },    229.93137,   0.00002,    40.183 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,  -1,   0 },    229.90672,   0.00004,     0.071 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,   0,   0 },    229.90397,   0.00002,     1.097 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,   1,   0 },    229.90672,   0.00004,     0.081 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,  -2,   0,   0 },    258.66018,   0.00004,     0.040 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,  -2,   1,   0 },    258.65376,   0.00003,     0.084 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,  -2,   2,   0 },     78.61458,   0.00002,     0.717 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,  -1,   0 },     81.40000,   0.00002,     0.073 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   0,   0 },    265.66955,   0.00013,     2.630 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   1,   0 },     81.40000,   0.00002,     0.078 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   0 },    291.36972,   0.00009,     0.041 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   1,   0 },     75.12304,   0.00002,     0.088 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,   0 },    105.85126,   0.00001,     0.076 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   0,   0 },    314.62812,   0.00017,     6.602 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,   0 },    105.85126,   0.00001,     0.075 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,  -3,   0 },    300.26357,   0.00003,     0.067 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,   0 },    300.08666,   0.00002,     0.040 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,   0 },    329.01006,   0.00003,     0.042 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,   0 },    355.18926,   0.00002,     0.080 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   0,   0 },      4.67775,   0.00005,     1.464 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,   0 },    355.18926,   0.00002,     0.072 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   2,   0 },    329.44344,   0.00001,     0.037 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,  -1,   0 },    332.15268,   0.00001,     0.082 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   1,   0 },    186.26648,   0.00003,     0.097 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   2,   0 },    172.80136,   0.00002,     0.335 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0 },    178.95615,   0.00002,     0.040 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0 },    170.43257,   0.00005,     0.823 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -3,   0 },      5.18166,   0.00004,     0.072 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -2,   0 },    178.51158,   0.00001,     1.788 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,   0 },    185.18166,   0.00004,     0.079 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,   2,   0 },    178.51158,   0.00001,     0.019 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -4,   2,   0 },    211.89790,   0.00001,     0.047 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,   0 },     31.37040,   0.00008,     0.043 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   2,   0 },     34.01649,   0.00003,     0.284 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0 },    229.73551,   0.00008,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0 },    227.92624,   0.00004,     0.573 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0 },     25.99006,   0.00003,     0.067 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0 },     27.80036,   0.00001,     0.035 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,  -1 },    148.89543,   0.00002,     1.444 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -3,   0 },    242.89069,   0.00002,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   0 },     35.56634,   0.00034,    36.376 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   0 },     44.63031,   0.00053,     0.076 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,   0 },     37.09845,   0.00011,     0.038 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,  -2,   0 },    244.32951,   0.00001,     0.020 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,   0 },     65.98367,   0.00002,     0.045 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   1,   0 },    244.49395,   0.00009,     0.109 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   2,   0 },     63.26509,   0.00002,     0.247 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   3,   0 },    244.51931,   0.00001,     0.058 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0 },     54.12300,   0.00002,     0.091 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0 },    256.67098,   0.00005,     0.439 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0 },     54.12300,   0.00002,     0.064 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0 },    236.53332,   0.00002,     0.035 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -3,   0 },     61.75396,   0.00001,     0.078 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,  -1,   0 },    234.12007,   0.00002,     0.073 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,  -2 },     56.53013,   0.00003,     5.859 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,   0 },     60.02614,   0.00003,     0.037 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   0,   0 },     96.06744,   0.00004,     0.046 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   1,   0 },    273.42303,   0.00006,     0.115 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   2,   0 },     88.17725,   0.00001,     0.218 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0 },    351.49579,   0.00002,     0.042 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0 },     23.67153,   0.00003,     0.356 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -3,   0 },     78.66681,   0.00003,     0.082 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,   0 },    258.66681,   0.00003,     0.070 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,   0 },     86.60540,   0.00007,     0.036 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   3,  -2,  -1 },    115.15601,   0.00003,    55.903 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,  -2,   1,   0 },    301.92558,   0.00003,     0.123 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0 },    343.48744,   0.00002,     0.043 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0 },    127.90865,   0.00006,     0.299 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,  -1,   0 },    113.37957,   0.00006,     0.067 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,   0,   0 },    300.36300,   0.00010,     0.036 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,  -2,   0,   0 },    148.48299,   0.00002,     0.048 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   0,  -1,   0 },    342.19693,   0.00002,     0.107 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0 },    155.26932,   0.00002,     0.258 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   2,  -2,   0 },    342.86018,   0.00002,     0.476 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   2,  -1,   0 },    162.19693,   0.00010,     0.065 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   2,   0,   0 },    162.39902,   0.00012,     0.035 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0,   0 },    248.96803,   0.00001,     0.227 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,  -3,   0 },    191.34413,   0.00003,     0.094 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,  -2,   0 },     10.21977,   0.00001,     0.380 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,  -1,   0 },    191.71445,   0.00008,     0.063 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,   0,   0 },    190.94002,   0.00005,     0.034 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   0,  -1,   0 },     41.91669,   0.00001,     0.120 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   0,   0,   0 },     41.84446,   0.00004,     0.203 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   0,   1,   0 },     41.91669,   0.00001,     0.055 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   0,   2,   0 },    221.56601,   0.00002,     0.032 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -3,   0 },    221.32781,   0.00001,     0.099 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -2,   0 },    221.32204,   0.00002,     0.316 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -1,   0 },    221.25671,   0.00002,     0.061 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,   0,   0 },    223.99963,   0.00003,     0.034 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,   2,   0 },     41.91669,   0.00002,     0.018 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   0,  -1,   0 },     71.05233,   0.00001,     0.128 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   0 },     70.77976,   0.00002,     0.183 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   0,   1,   0 },     71.05233,   0.00001,     0.053 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   0,   2,   0 },    250.77976,   0.00002,     0.031 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   2,  -1,   0 },     71.30502,   0.00001,     0.059 },
 { {   0,   0,   6,  -1,   0,   0,   0,   0,   2,   0,   0 },    250.45440,   0.00001,     0.033 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   0,   0,   0 },    226.15228,   0.00005,     0.167 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   0,   1,   0 },     18.96065,   0.00001,     0.052 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,  -1,   0 },    305.05365,   0.00003,     0.057 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,   0,   0 },    237.92029,   0.00004,     0.033 },
 { {   0,   0,   7, -18,   0,   0,   0,   0,  -2,   1,   0 },    293.92709,   0.00003,     0.071 },
 { {   0,   0,   7, -17,   0,   0,   0,   0,  -2,   0,   0 },    322.90996,   0.00001,     0.037 },
 { {   0,   0,   7, -17,   0,   0,   0,   0,  -2,   0,   2 },    142.90996,   0.00001,    14.514 },
 { {   0,   0,   7, -17,   0,   0,   0,   0,  -2,   1,   0 },    143.65072,   0.00001,     0.074 },
 { {   0,   0,   7, -17,   0,   0,   0,   0,  -2,   3,   0 },    143.65072,   0.00001,     0.077 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   0,   0 },    351.95396,   0.00009,     0.038 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   1,   0 },    173.37128,   0.00004,     0.077 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   2,   0 },    171.90196,   0.00001,     3.744 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   3,   0 },    353.37128,   0.00004,     0.074 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,   0,   0,   0 },    171.90196,   0.00001,     0.664 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -2,   0,   0 },    200.91087,   0.00006,     0.039 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -2,   1,   0 },     21.34914,   0.00004,     0.080 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,  -1,   0 },    200.91086,   0.00003,     0.070 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   0,   0 },      3.44334,   0.00005,     1.025 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    200.91086,   0.00003,     0.081 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   2,   0,   0 },    343.98603,   0.00002,     0.042 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,  -2,   0,   0 },    229.71374,   0.00002,     0.040 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,  -2,   1,   0 },     51.24984,   0.00001,     0.084 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,   0,  -2,   0 },     49.74381,   0.00001,     0.037 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,   0,   0,   0 },    238.14374,   0.00005,     2.254 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   0 },    264.77244,   0.00007,     0.041 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,   0 },     40.11904,   0.00002,     0.088 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0 },     78.35972,   0.00001,     0.038 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   0 },    274.01288,   0.00033,    11.350 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -3,   0 },    271.14837,   0.00002,     0.067 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,  -1,   0 },    271.14837,   0.00002,     0.086 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,   0 },    271.14837,   0.00003,     0.040 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   0,   0 },     96.03416,   0.00001,     0.041 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,  -2,   1,   0 },    286.13395,   0.00003,     0.092 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,  -1,   0 },    292.04925,   0.00002,     0.079 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   0,   0 },    323.09149,   0.00002,     1.613 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   0,   1,   0 },    292.04925,   0.00002,     0.072 },
 { {   0,   0,   7, -12,   0,   0,   0,   0,   2,   0,   0 },    300.20208,   0.00003,     0.039 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   2,   0 },    357.69703,   0.00001,     0.342 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,   0 },    329.21739,   0.00002,     0.083 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   0,   0 },    136.22960,   0.00008,     0.868 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,   0 },    329.21739,   0.00002,     0.069 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   2,   0 },    329.39845,   0.00001,     0.036 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,  -1 },     37.84249,   0.00004,    10.235 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,  -2,   0 },    149.73575,   0.00002,     1.606 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,  -1,   0 },    149.39845,   0.00002,     0.079 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,   0,   0 },    329.49092,   0.00003,     0.039 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   0,   0 },      6.37592,   0.00003,     0.043 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   1,   0 },    186.21769,   0.00002,     0.102 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,  -2,   2,   0 },    184.04133,   0.00003,     0.289 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   0,   0 },    144.50531,   0.00002,     0.594 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   1,   0 },    358.43847,   0.00002,     0.067 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   2,   0 },    178.45663,   0.00002,     0.035 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -2,   0 },    201.42022,   0.00002,    11.006 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,   0 },    336.60920,   0.00012,     0.076 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   0,  -2 },    178.45663,   0.00001,     2.356 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,   2,   0 },    178.45663,   0.00001,     0.019 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   0,   0 },     41.94811,   0.00006,     0.044 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   2,   0 },     32.72651,   0.00001,     0.251 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0 },     18.17482,   0.00001,     0.091 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0 },    183.56361,   0.00010,     0.452 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0 },     18.17482,   0.00001,     0.065 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0 },    207.69947,   0.00001,     0.035 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -3,   0 },    214.46795,   0.00001,     0.078 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -2,   0 },     32.05125,   0.00001,     2.269 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,  -1,   0 },    204.64020,   0.00003,     0.073 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,  -2 },    207.88431,   0.00003,     9.319 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,   0 },     26.89810,   0.00007,     0.037 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   0,   0 },     76.19702,   0.00003,     0.045 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   1,   0 },    244.39084,   0.00005,     0.114 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   2,   0 },     44.98657,   0.00001,     0.221 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   3,   0 },     64.39084,   0.00003,     0.056 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0 },    242.85554,   0.00010,     0.364 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,  -1,   0 },    229.82002,   0.00006,     0.070 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,   0 },     56.69732,   0.00001,     0.036 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,  -2,   0,   0 },     92.70382,   0.00004,     0.047 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0 },     93.87102,   0.00002,     0.043 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0 },    273.09215,   0.00001,     0.100 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0 },    273.09215,   0.00001,     0.060 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0 },    225.94167,   0.00001,     0.034 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -2,   0 },    262.85000,   0.00002,     0.665 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,   0 },     86.35643,   0.00009,     0.068 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   0,   0 },    296.90851,   0.00001,     0.048 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   1,   0 },    301.48713,   0.00001,     0.130 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   2,   0 },    116.90851,   0.00001,     0.179 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,  -2,   3,   0 },    301.48713,   0.00001,     0.053 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0 },    135.16088,   0.00002,     0.262 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -3,   0 },     70.03146,   0.00001,     0.089 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -2,   0 },    306.87615,   0.00002,     0.491 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -1,   0 },    300.33864,   0.00004,     0.065 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,   0,   0 },    304.73469,   0.00019,     0.035 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   0,  -1,   0 },    342.43456,   0.00002,     0.112 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0,   0 },    155.45395,   0.00001,     0.230 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   2,  -2,   0 },    342.27043,   0.00003,     0.389 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   2,  -1,   0 },    162.43456,   0.00011,     0.063 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   2,   0,   0 },    162.75497,   0.00009,     0.034 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0,   0 },    192.75404,   0.00004,     0.205 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   2,  -3,   0 },      7.24584,   0.00001,     0.098 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   2,  -1,   0 },    191.48971,   0.00006,     0.061 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   2,   0,   0 },    191.67176,   0.00007,     0.034 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   0,  -1,   0 },     41.86500,   0.00001,     0.127 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   0,   1,   0 },     41.86500,   0.00001,     0.054 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   2,  -1,   0 },    221.25672,   0.00004,     0.059 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   2,   0,   0 },    220.62471,   0.00007,     0.033 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   2,   2,   0 },     41.86500,   0.00001,     0.018 },
 { {   0,   0,   7,  -2,   0,   0,   0,   0,   0,  -1,   0 },     70.92015,   0.00001,     0.137 },
 { {   0,   0,   7,  -2,   0,   0,   0,   0,   0,   1,   0 },     70.92015,   0.00001,     0.052 },
 { {   0,   0,   7,  -2,   0,   0,   0,   0,   2,  -1,   0 },    250.92015,   0.00001,     0.057 },
 { {   0,   0,   7,  -2,   0,   0,   0,   0,   2,   2,   0 },     70.92015,   0.00001,     0.017 },
 { {   0,   0,   7,  -1,   0,   0,   0,   0,   2,   0,   0 },    279.44822,   0.00002,     0.032 },
 { {   0,   0,   7,   0,   0,   0,   0,   0,   2,   0,   0 },    303.09810,   0.00002,     0.032 },
 { {   0,   0,   8, -18,   0,   0,   0,   0,  -2,   0,   0 },    322.85202,   0.00001,     0.038 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,  -2,   0,   0 },    351.86966,   0.00008,     0.039 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   0,  -1,   0 },    351.82515,   0.00001,     0.070 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,   0,   1,   0 },    351.82515,   0.00001,     0.082 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,  -2,   0,   0 },    200.78520,   0.00006,     0.040 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,   0,  -1,   0 },    200.53528,   0.00002,     0.073 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,   0,   1,   0 },    200.53528,   0.00002,     0.078 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,  -1,   0 },    243.19478,   0.00003,     0.026 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   0 },    239.62675,   0.00019,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,   0 },    248.11397,   0.00019,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0 },    234.69162,   0.00002,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   0 },    235.92541,   0.00028,     0.076 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   0 },    240.85911,   0.00143,    40.427 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   0 },    235.55508,   0.00030,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   2,   0 },    234.70582,   0.00002,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,   0 },    243.74380,   0.00023,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,   0 },    242.15468,   0.00014,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   1,   0 },    242.31946,   0.00001,     0.026 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   0,   0 },    258.00897,   0.00006,     0.041 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   1,   0 },     78.00898,   0.00002,     0.092 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   3,   0 },    258.00898,   0.00002,     0.064 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0 },     73.30389,   0.00004,     0.079 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   0,   0 },     86.42663,   0.00004,     1.797 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   1,   0 },     73.30389,   0.00004,     0.072 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   2,   0,   0 },    271.32764,   0.00001,     0.040 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0 },    106.04519,   0.00011,     0.042 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   0,   0,   0 },    119.92610,   0.00013,     0.919 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   0,  -1 },     23.25220,   0.00005,    29.122 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,  -3,   0 },    121.09335,   0.00002,     0.072 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,  -1,   0 },    301.09335,   0.00002,     0.080 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   2,   0,   0 },    300.15719,   0.00003,     0.039 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   0,   0 },    155.22241,   0.00003,     0.043 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   2,   0 },    157.40773,   0.00001,     0.295 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   0,   0 },    152.06526,   0.00005,     0.617 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   0,   2,   0 },    329.35005,   0.00001,     0.036 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   1,   0,  -1 },      1.35672,   0.00002,     1.767 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -3,   0 },    352.06795,   0.00002,     0.075 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,  -1,   0 },    163.66727,   0.00003,     0.076 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,   0,   0 },    329.25673,   0.00003,     0.038 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   0,   0 },     33.09756,   0.00003,     0.044 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   1,   0 },    186.16433,   0.00001,     0.107 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,  -2,   2,   0 },    148.37367,   0.00002,     0.255 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   0,   0,   0 },    180.61065,   0.00005,     0.465 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -3,   0 },      5.40596,   0.00003,     0.078 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,   0 },     14.86300,   0.00002,     0.073 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,   0 },    358.70483,   0.00003,     0.037 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   0,   0 },     39.88107,   0.00005,     0.045 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   1,   0 },    215.16304,   0.00001,     0.114 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   2,   0 },    185.86515,   0.00001,     0.224 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   3,   0 },    215.16304,   0.00001,     0.056 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0 },     27.33525,   0.00002,     0.095 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   0,   0 },    211.85246,   0.00009,     0.373 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   1,   0 },     27.33525,   0.00002,     0.063 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -2,   0 },    207.99991,   0.00002,     1.100 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -1,   0 },    207.33525,   0.00003,     0.071 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,   0 },     27.50610,   0.00005,     0.036 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   2,   0 },    207.33525,   0.00001,     0.019 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   0,   0 },     50.18980,   0.00003,     0.046 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   1,   0 },    244.25174,   0.00003,     0.121 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   2,   0 },    357.10567,   0.00001,     0.200 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0 },    220.68000,   0.00011,     0.311 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -3,   0 },     54.16642,   0.00002,     0.085 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -2,   0 },     56.76959,   0.00001,     0.694 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,   0 },    234.78025,   0.00004,     0.068 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,   0 },     54.03905,   0.00006,     0.036 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,  -2,   0,   0 },     91.68772,   0.00003,     0.048 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,  -2,   1,   0 },    272.93986,   0.00003,     0.129 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0 },    272.77659,   0.00003,     0.044 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0 },    238.29582,   0.00002,     0.267 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -2,   0 },     77.57857,   0.00003,     0.507 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -1,   0 },     72.75951,   0.00004,     0.066 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,   0 },    102.78852,   0.00001,     0.035 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,  -2,   0,   0 },    121.72107,   0.00002,     0.049 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,  -2,   1,   0 },    300.83860,   0.00002,     0.139 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   0 },    301.13082,   0.00001,     0.045 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,  -2,   0 },     86.45554,   0.00002,     0.399 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,  -1,   0 },    314.63355,   0.00003,     0.063 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,   0,   0 },    307.89199,   0.00013,     0.034 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,  -2,   0,   0 },    327.44806,   0.00001,     0.050 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   0 },    252.45551,   0.00001,     0.046 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0,   0 },    352.04658,   0.00002,     0.208 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -3,   0 },    343.56115,   0.00002,     0.098 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -2,   0 },    341.91461,   0.00006,     0.329 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -1,   0 },    163.42903,   0.00005,     0.061 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,   0,   0 },    163.82742,   0.00004,     0.034 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0,   0 },    192.71083,   0.00004,     0.187 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   2,  -2,   0 },     32.53055,   0.00001,     0.280 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   2,  -1,   0 },    192.17390,   0.00003,     0.059 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   2,   0,   0 },    197.30565,   0.00007,     0.033 },
 { {   0,   0,   8,  -4,   0,   0,   0,   0,   2,  -1,   0 },    221.78899,   0.00001,     0.058 },
 { {   0,   0,   8,  -4,   0,   0,   0,   0,   2,   0,   0 },    220.67715,   0.00007,     0.033 },
 { {   0,   0,   8,  -3,   0,   0,   0,   0,   0,   0,   0 },    250.73258,   0.00001,     0.156 },
 { {   0,   0,   8,  -3,   0,   0,   0,   0,   2,   0,   0 },     70.73258,   0.00001,     0.032 },
 { {   0,   0,   8,  -3,   0,   0,   0,   0,   2,   2,   0 },    250.73258,   0.00001,     0.017 },
 { {   0,   0,   8,  -2,   0,   0,   0,   0,   2,  -3,   0 },     98.87261,   0.00001,     0.124 },
 { {   0,   0,   8,  -2,   0,   0,   0,   0,   2,  -1,   0 },     98.87261,   0.00001,     0.054 },
 { {   0,   0,   9, -21,   0,   0,   0,   0,  -2,   1,   0 },    264.84859,   0.00003,     0.073 },
 { {   0,   0,   9, -19,   0,   0,   0,   0,  -2,   1,   0 },    322.84927,   0.00007,     0.079 },
 { {   0,   0,   9, -19,   0,   0,   0,   0,  -2,   2,   0 },    142.78505,   0.00002,     1.488 },
 { {   0,   0,   9, -19,   0,   0,   0,   0,   0,   0,   0 },    142.78505,   0.00001,     0.907 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,  -2,   0,   0 },    351.76955,   0.00008,     0.040 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,  -2,   1,   0 },    351.77777,   0.00006,     0.083 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,   0,  -1,   0 },    171.77776,   0.00003,     0.072 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,   0,   0,   0 },    348.50560,   0.00003,     1.753 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,   0,   1,   0 },    171.77776,   0.00003,     0.079 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,   0 },    213.33163,   0.00006,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   1,   0 },      1.00679,   0.00002,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,   0 },    178.10837,   0.00003,     0.075 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   0,   0 },    208.79669,   0.00023,    25.886 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,   0 },    178.10837,   0.00003,     0.076 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,   0 },    213.59995,   0.00004,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,   0 },    213.60002,   0.00003,     0.040 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   0,   0 },    228.41978,   0.00002,     0.041 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   3,   0 },    229.34023,   0.00003,     0.064 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,  -1,   0 },     49.36747,   0.00004,     0.078 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   0,   0 },     60.32886,   0.00005,     2.028 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   1,   0 },     49.36747,   0.00004,     0.073 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,   0 },    144.44046,   0.00002,     0.042 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   1,   0 },    109.88546,   0.00002,     0.096 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   0,   0,   0 },     97.28076,   0.00013,     0.976 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,  -1 },    328.31005,   0.00003,    34.449 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,  -1,   0 },     91.13400,   0.00005,     0.080 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,   0,   0 },    271.26790,   0.00001,     0.039 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,  -2,   0,   0 },    104.36671,   0.00001,     0.043 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   0,   0,   0 },    104.62985,   0.00007,     0.643 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -2,   0 },    120.16255,   0.00005,     4.596 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,  -2,   0,   0 },    178.34768,   0.00003,     0.044 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,  -2,   2,   0 },    214.40230,   0.00001,     0.259 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,  -1,   0 },    329.29102,   0.00001,     0.090 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   0,   0 },    144.36211,   0.00004,     0.479 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   0,   1,   0 },    329.29102,   0.00001,     0.065 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -3,   0 },    176.57860,   0.00003,     0.077 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -2,   0 },    149.21343,   0.00002,     3.184 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,  -2 },    330.45159,   0.00001,     0.075 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,   0 },    323.11371,   0.00004,     0.074 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,  -2 },    346.76471,   0.00005,    51.478 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,   0 },    329.21343,   0.00002,     0.037 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   1,  -2 },    330.45159,   0.00001,     0.076 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,  -2,   0,   0 },      8.07392,   0.00003,     0.045 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,  -2,   1,   0 },    186.10121,   0.00001,     0.113 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,  -1,   0 },    358.28385,   0.00001,     0.094 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,   0,   0 },    172.97575,   0.00004,     0.382 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   0,   1,   0 },    358.28385,   0.00001,     0.063 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -3,   0 },    358.50631,   0.00004,     0.081 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -2,   0 },    358.94385,   0.00001,     1.182 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,   0 },    178.41871,   0.00006,     0.071 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,   0 },    358.65975,   0.00003,     0.037 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,  -2,   0,   0 },     36.94289,   0.00003,     0.046 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,  -2,   1,   0 },    215.06817,   0.00002,     0.120 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0 },     27.14518,   0.00001,     0.099 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   0,   0 },    210.98703,   0.00005,     0.317 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   0,   1,   0 },     27.14518,   0.00001,     0.061 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,  -1,   0 },    207.12576,   0.00005,     0.068 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,   0 },     27.12022,   0.00005,     0.036 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,  -2,   0,   0 },     61.74219,   0.00008,     0.047 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,  -2,   1,   0 },    243.94307,   0.00003,     0.128 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   0,   0 },    218.30508,   0.00007,     0.272 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -3,   0 },     55.61777,   0.00003,     0.088 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,   0 },     55.61777,   0.00003,     0.066 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,   0,   0 },     55.58650,   0.00005,     0.035 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,  -2,   0,   0 },     91.09089,   0.00003,     0.049 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,  -2,   1,   0 },    272.55897,   0.00001,     0.138 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,  -2,   3,   0 },     92.55897,   0.00001,     0.052 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0 },    194.54583,   0.00001,     0.045 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0 },    297.55210,   0.00002,     0.237 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   2,   0,   0 },    244.16479,   0.00002,     0.035 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   0 },     41.36526,   0.00002,     0.046 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0 },    104.25346,   0.00004,     0.211 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,  -1,   0 },    314.57008,   0.00003,     0.062 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,   0,   0 },    313.30315,   0.00012,     0.034 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0,   0 },    208.00648,   0.00003,     0.189 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,  -3,   0 },    343.28351,   0.00001,     0.103 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,  -2,   0 },    341.82323,   0.00003,     0.285 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,  -1,   0 },    163.28351,   0.00001,     0.060 },
 { {   0,   0,   9,  -6,   0,   0,   0,   0,   0,   0,   0 },    192.65589,   0.00004,     0.172 },
 { {   0,   0,   9,  -6,   0,   0,   0,   0,   2,  -1,   0 },    192.56427,   0.00008,     0.058 },
 { {   0,   0,   9,  -6,   0,   0,   0,   0,   2,   0,   0 },    193.19003,   0.00006,     0.033 },
 { {   0,   0,   9,  -5,   0,   0,   0,   0,   0,   0,   0 },    221.68088,   0.00001,     0.158 },
 { {   0,   0,   9,  -5,   0,   0,   0,   0,   2,  -2,   0 },     41.11896,   0.00002,     0.219 },
 { {   0,   0,   9,  -5,   0,   0,   0,   0,   2,   0,   0 },     41.51987,   0.00002,     0.032 },
 { {   0,   0,   9,  -5,   0,   0,   0,   0,   2,   2,   0 },    221.68088,   0.00001,     0.017 },
 { {   0,   0,   9,  -4,   0,   0,   0,   0,   0,   0,   0 },    250.48029,   0.00002,     0.145 },
 { {   0,   0,   9,  -4,   0,   0,   0,   0,   2,  -3,   0 },     70.37340,   0.00002,     0.123 },
 { {   0,   0,   9,  -4,   0,   0,   0,   0,   2,  -1,   0 },    250.37340,   0.00002,     0.054 },
 { {   0,   0,  10, -22,   0,   0,   0,   0,  -2,   1,   0 },     89.72799,   0.00001,     0.076 },
 { {   0,   0,  10, -22,   0,   0,   0,   0,  -2,   2,   0 },    264.87723,   0.00003,    12.993 },
 { {   0,   0,  10, -22,   0,   0,   0,   0,  -2,   3,   0 },    269.72799,   0.00001,     0.075 },
 { {   0,   0,  10, -20,   0,   0,   0,   0,  -2,   1,   0 },    142.70606,   0.00002,     0.083 },
 { {   0,   0,  10, -20,   0,   0,   0,   0,  -2,   2,   0 },    142.70606,   0.00002,     0.877 },
 { {   0,   0,  10, -20,   0,   0,   0,   0,  -2,   3,   0 },    142.70606,   0.00002,     0.069 },
 { {   0,   0,  10, -20,   0,   0,   0,   0,   0,  -1,   0 },    322.70606,   0.00001,     0.072 },
 { {   0,   0,  10, -20,   0,   0,   0,   0,   0,   1,   0 },    322.70606,   0.00001,     0.079 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   0,   0 },    351.64538,   0.00009,     0.040 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   1,   0 },    171.61898,   0.00003,     0.086 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,  -1,   0 },    354.88051,   0.00002,     0.075 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   0,   0 },    191.58158,   0.00010,     9.804 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,   0,   1,   0 },    354.88051,   0.00002,     0.076 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   0 },    162.29779,   0.00005,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,   0 },    162.29779,   0.00009,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,   0 },    162.29779,   0.00009,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,   0 },    162.29779,   0.00005,     0.040 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,  -2,   0,   0 },     20.45814,   0.00008,     0.041 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,  -2,   1,   0 },    200.43532,   0.00003,     0.090 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,   0,  -1,   0 },     20.42974,   0.00004,     0.078 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,   0,   1,   0 },     20.42974,   0.00004,     0.073 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   0,   0 },     50.09670,   0.00003,     0.042 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   1,   0 },    229.03296,   0.00003,     0.095 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   2,   0 },     49.84020,   0.00002,     0.366 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,  -1,   0 },    229.03296,   0.00002,     0.081 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   0,   0 },    238.55741,   0.00003,     1.040 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   0,   1,   0 },    229.03296,   0.00002,     0.070 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   2,  -1,   0 },     62.23825,   0.00003,     0.080 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,  -2,   0,   0 },     77.05128,   0.00010,     0.043 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   0,   0,   0 },     86.22118,   0.00005,     0.670 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,  -3,   0 },    276.10278,   0.00001,     0.074 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,  -1,   0 },     96.10278,   0.00001,     0.077 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,   2,   0,   0 },    271.21507,   0.00001,     0.038 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   0,   0,   0 },    117.99573,   0.00006,     0.494 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -3,   0 },    121.20740,   0.00002,     0.077 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -2,   0 },    120.02952,   0.00001,     3.989 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,   0 },    119.39007,   0.00006,     0.074 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,  -2 },    300.12408,   0.00002,    12.078 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,   0 },    120.01013,   0.00001,     0.037 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,  -2,   2,   0 },    327.52299,   0.00001,     0.231 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   0,   0,   0 },    127.42161,   0.00005,     0.391 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -3,   0 },    149.86800,   0.00001,     0.080 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -1,   0 },    149.66463,   0.00004,     0.071 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,   0 },    329.15213,   0.00005,     0.037 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,  -2,   0,   0 },      7.50547,   0.00003,     0.046 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,  -2,   1,   0 },    186.02440,   0.00001,     0.119 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,   0,   0 },    158.47727,   0.00006,     0.324 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   0,   2,   0 },    358.15625,   0.00001,     0.034 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,  -3,   0 },    178.61089,   0.00002,     0.084 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,  -1,   0 },    178.43955,   0.00003,     0.069 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   0,   0 },    358.41323,   0.00006,     0.036 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   0,   0 },     34.92055,   0.00003,     0.047 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   2,   0 },    216.58756,   0.00002,     0.185 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,  -2,   0 },     34.95022,   0.00001,     0.044 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   0,   0,   0 },    203.09128,   0.00005,     0.276 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,   0 },    207.31653,   0.00004,     0.066 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,   0,   0 },     26.50124,   0.00003,     0.035 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,  -2,   0,   0 },     63.16801,   0.00004,     0.049 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,  -2,   1,   0 },    243.85993,   0.00003,     0.136 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,   0,   0 },    203.59598,   0.00005,     0.241 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -3,   0 },    234.92240,   0.00003,     0.092 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -2,   0 },    235.12835,   0.00003,     0.421 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -1,   0 },     54.92240,   0.00003,     0.064 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0 },    272.14503,   0.00001,     0.046 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,   0,   0 },    270.99130,   0.00003,     0.214 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   2,   0,   0 },    105.52689,   0.00003,     0.034 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0,   0 },    133.17263,   0.00002,     0.192 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,  -1,   0 },    314.51877,   0.00002,     0.060 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,   0,   0 },    313.31186,   0.00012,     0.033 },
 { {   0,   0,  10,  -8,   0,   0,   0,   0,  -2,   0,   0 },    143.73263,   0.00002,     0.053 },
 { {   0,   0,  10,  -8,   0,   0,   0,   0,   2,   0,   0 },    341.66221,   0.00004,     0.033 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   0,   0,   0 },    192.58179,   0.00002,     0.159 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   2,  -2,   0 },     11.59232,   0.00001,     0.222 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   2,  -1,   0 },    192.37284,   0.00005,     0.056 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   2,   0,   0 },      7.44704,   0.00002,     0.032 },
 { {   0,   0,  10,  -6,   0,   0,   0,   0,   0,   0,   0 },    221.53450,   0.00003,     0.147 },
 { {   0,   0,  10,  -6,   0,   0,   0,   0,   2,  -1,   0 },    221.32170,   0.00004,     0.055 },
 { {   0,   0,  10,  -5,   0,   0,   0,   0,   2,   0,   0 },    250.03639,   0.00003,     0.031 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,  -2,   0,   0 },    142.61271,   0.00003,     0.040 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,  -2,   1,   0 },    142.61271,   0.00001,     0.086 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,  -2,   3,   0 },    322.61271,   0.00001,     0.067 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,  -1,   0 },    142.41752,   0.00002,     0.075 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,   0,   0 },     70.89216,   0.00002,     6.047 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,   1,   0 },    142.41752,   0.00002,     0.076 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,  -2,   0,   0 },    213.45127,   0.00002,     0.040 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   0,   0,   0 },    214.36969,   0.00008,   308.674 },
 { {   0,   0,  11, -21,   2,   0,   0,   0,   2,   0,   0 },    213.45108,   0.00002,     0.040 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,  -2,   0,   0 },    351.50763,   0.00008,     0.041 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,  -2,   1,   0 },    171.48511,   0.00002,     0.090 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,  -1,   0 },    351.51838,   0.00003,     0.078 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,   0,   0 },    169.40223,   0.00002,     2.730 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,   1,   0 },    351.51838,   0.00003,     0.073 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   0,   0 },     20.24570,   0.00010,     0.042 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   1,   0 },    200.23278,   0.00002,     0.094 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,   0,  -1,   0 },    200.23278,   0.00001,     0.081 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,   0,   1,   0 },    200.23278,   0.00001,     0.071 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,  -2,   0,   0 },     48.71350,   0.00010,     0.043 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   0,   0,   0 },     48.70627,   0.00001,     0.699 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -3,   0 },     62.05105,   0.00003,     0.074 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -1,   0 },     62.05105,   0.00003,     0.077 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   0,   0 },     76.55294,   0.00006,     0.044 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   2,   0 },    256.12267,   0.00001,     0.268 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   0,   0,   0 },     91.24243,   0.00003,     0.510 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,   0,   0 },    271.16225,   0.00001,     0.037 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,  -2,   0,   0 },    128.01427,   0.00003,     0.045 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   0,   0,   0 },    123.39082,   0.00008,     0.401 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,   0 },    120.04430,   0.00004,     0.037 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,  -2,   2,   0 },    157.14602,   0.00003,     0.208 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   0,   0,   0 },    148.62321,   0.00007,     0.331 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,  -2,   0 },    148.45552,   0.00001,     0.800 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,   2,   0,   0 },    329.64567,   0.00001,     0.036 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,  -2,   0,   0 },    351.98247,   0.00004,     0.047 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,  -2,   1,   0 },    185.93089,   0.00001,     0.126 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,  -2,   2,   0 },    327.86915,   0.00002,     0.188 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   0,   0,   0 },    173.78452,   0.00004,     0.281 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -3,   0 },    357.97054,   0.00003,     0.087 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -2,   0 },    178.47567,   0.00002,     0.561 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -1,   0 },    357.97054,   0.00003,     0.066 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,   0,   0 },    358.09281,   0.00007,     0.035 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,  -2,   0,   0 },     31.40935,   0.00005,     0.048 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,  -2,   1,   0 },    214.80650,   0.00001,     0.135 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,  -2,   2,   0 },     33.65491,   0.00001,     0.171 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,  -1,   0 },     26.60646,   0.00005,     0.064 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,   0,   0 },     26.60645,   0.00005,     0.035 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,  -2,   0,   0 },     62.47145,   0.00003,     0.050 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,  -2,   1,   0 },    243.59998,   0.00003,     0.146 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   0,   0,   0 },    234.07361,   0.00004,     0.216 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,  -2,   0 },     56.48683,   0.00001,     0.351 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,  -1,   0 },     54.38623,   0.00003,     0.062 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   0,   0 },    233.47371,   0.00002,     0.034 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,  -2,   1,   0 },    271.72320,   0.00002,     0.158 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   0,   0,   0 },    145.39231,   0.00002,     0.194 },
 { {   0,   0,  11, -11,   0,   0,   0,   0,   2,   0,   0 },    105.46716,   0.00003,     0.033 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,  -3,   0 },    134.46997,   0.00001,     0.107 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,  -2,   0 },    184.40724,   0.00001,     0.256 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,  -1,   0 },    314.46997,   0.00001,     0.058 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,   0,   0 },    318.94795,   0.00010,     0.033 },
 { {   0,   0,  11,  -9,   0,   0,   0,   0,   2,  -3,   0 },    343.49028,   0.00002,     0.113 },
 { {   0,   0,  11,  -9,   0,   0,   0,   0,   2,  -1,   0 },    343.49028,   0.00002,     0.057 },
 { {   0,   0,  11,  -9,   0,   0,   0,   0,   2,   0,   0 },    350.70471,   0.00005,     0.032 },
 { {   0,   0,  11,  -8,   0,   0,   0,   0,   2,  -1,   0 },    192.48310,   0.00001,     0.055 },
 { {   0,   0,  11,  -7,   0,   0,   0,   0,   0,   0,   0 },    221.34444,   0.00001,     0.137 },
 { {   0,   0,  11,  -7,   0,   0,   0,   0,   0,   2,   0 },     41.34444,   0.00001,     0.030 },
 { {   0,   0,  11,  -7,   0,   0,   0,   0,   2,  -1,   0 },    221.34444,   0.00001,     0.053 },
 { {   0,   0,  11,  -6,   0,   0,   0,   0,   2,  -3,   0 },    249.73942,   0.00001,     0.139 },
 { {   0,   0,  11,  -6,   0,   0,   0,   0,   2,  -1,   0 },     69.73942,   0.00001,     0.052 },
 { {   0,   0,  11,  -6,   0,   0,   0,   0,   2,   0,   0 },     69.61688,   0.00002,     0.031 },
 { {   0,   0,  12, -26,   0,   0,   0,   0,  -2,   1,   0 },     26.86990,   0.00006,     0.075 },
 { {   0,   0,  12, -23,   0,   0,   0,   0,   0,  -1,   0 },    293.60118,   0.00002,     0.074 },
 { {   0,   0,  12, -23,   0,   0,   0,   0,   0,   0,   0 },    293.49537,   0.00003,     4.372 },
 { {   0,   0,  12, -23,   0,   0,   0,   0,   0,   1,   0 },    293.60118,   0.00002,     0.077 },
 { {   0,   0,  12, -22,   0,   0,   0,   0,  -2,   0,   0 },    142.50304,   0.00003,     0.041 },
 { {   0,   0,  12, -22,   0,   0,   0,   0,   0,   0,   0 },    138.87961,   0.00003,     3.301 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,  -2,   0,   0 },    351.34280,   0.00005,     0.042 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,  -2,   2,   0 },    171.35654,   0.00002,     0.383 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,   0,   0,   0 },    351.32818,   0.00001,     1.198 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,  -2,   0,   0 },     20.00834,   0.00009,     0.043 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,   0,   0,   0 },     19.99716,   0.00001,     0.732 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,   2,  -1,   0 },     33.16046,   0.00003,     0.078 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,  -2,   0,   0 },     48.33511,   0.00008,     0.044 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,   2,  -2,   0 },     61.96825,   0.00003,     8.066 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,  -2,   0,   0 },    255.76084,   0.00002,     0.045 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,  -2,   2,   0 },    255.76084,   0.00002,     0.238 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -3,   0 },    271.01129,   0.00003,     0.079 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -2,   0 },     91.10430,   0.00002,     1.525 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -1,   0 },    271.01129,   0.00003,     0.072 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,  -2,   0,   0 },    127.95213,   0.00003,     0.046 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   0,   0,   0 },    119.84746,   0.00003,     0.338 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,  -1,   0 },    300.00937,   0.00005,     0.069 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,   0,   0 },    119.98390,   0.00003,     0.036 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,  -2,   0,   0 },    157.63180,   0.00002,     0.047 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,  -2,   2,   0 },    156.71986,   0.00002,     0.190 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   0,   0,   0 },    148.55244,   0.00004,     0.286 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -2,   0 },    149.21991,   0.00002,     0.582 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -1,   0 },    328.96065,   0.00005,     0.067 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,   0,   0 },    148.70213,   0.00002,     0.035 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,  -2,   0,   0 },      3.24294,   0.00007,     0.048 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,  -2,   1,   0 },    185.81823,   0.00001,     0.134 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -2,   0 },    177.78052,   0.00004,     0.444 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -1,   0 },    357.78052,   0.00005,     0.064 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,   0,   0 },    177.78053,   0.00001,     0.035 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,  -2,   0,   0 },     34.46585,   0.00009,     0.050 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,  -2,   1,   0 },    214.63457,   0.00001,     0.144 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,  -3,   0 },     26.25626,   0.00002,     0.095 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,  -2,   0 },    207.29193,   0.00001,     0.359 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,  -1,   0 },     26.25626,   0.00002,     0.062 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   0,   0 },    205.30844,   0.00001,     0.034 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,  -2,   0,   0 },     61.99429,   0.00002,     0.051 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,  -2,   1,   0 },    243.29183,   0.00002,     0.156 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   0,   0,   0 },     61.60276,   0.00002,     0.197 },
 { {   0,   0,  12, -13,   0,   0,   0,   0,   2,  -2,   0 },    233.59780,   0.00002,     0.302 },
 { {   0,   0,  12, -12,   0,   0,   0,   0,  -2,   0,   0 },     91.14695,   0.00002,     0.052 },
 { {   0,   0,  12, -12,   0,   0,   0,   0,   0,   0,   0 },     89.32543,   0.00001,     0.178 },
 { {   0,   0,  12, -12,   0,   0,   0,   0,   2,   0,   0 },    105.41432,   0.00003,     0.033 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,  -2,   0 },    134.28989,   0.00002,     0.228 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,   0,   0 },    314.22099,   0.00007,     0.032 },
 { {   0,   0,  12, -10,   0,   0,   0,   0,   2,  -1,   0 },    343.41532,   0.00003,     0.055 },
 { {   0,   0,  12,  -9,   0,   0,   0,   0,   2,   0,   0 },     11.24539,   0.00002,     0.031 },
 { {   0,   0,  12,  -8,   0,   0,   0,   0,   2,  -3,   0 },     41.10530,   0.00001,     0.137 },
 { {   0,   0,  12,  -8,   0,   0,   0,   0,   2,  -1,   0 },     41.10530,   0.00001,     0.052 },
 { {   0,   0,  12,  -7,   0,   0,   0,   0,   2,   0,   0 },    249.38024,   0.00002,     0.030 },
 { {   0,   0,  13, -24,   0,   0,   0,   0,   0,  -1,   0 },    113.50737,   0.00002,     0.077 },
 { {   0,   0,  13, -24,   0,   0,   0,   0,   0,   0,   0 },    113.37505,   0.00002,     4.174 },
 { {   0,   0,  13, -24,   0,   0,   0,   0,   0,   1,   0 },    113.50737,   0.00002,     0.074 },
 { {   0,   0,  13, -23,   0,   0,   0,   0,  -2,   0,   0 },    142.37542,   0.00003,     0.042 },
 { {   0,   0,  13, -22,   0,   0,   0,   0,  -2,   2,   0 },    171.14641,   0.00002,     0.325 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,  -2,   0,   0 },     19.73354,   0.00005,     0.044 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,  -2,   2,   0 },    199.74060,   0.00003,     0.277 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -3,   0 },    209.68501,   0.00003,     0.076 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -1,   0 },    209.68501,   0.00003,     0.075 },
 { {   0,   0,  13, -20,   0,   0,   0,   0,  -2,   0,   0 },     47.88817,   0.00001,     0.045 },
 { {   0,   0,  13, -20,   0,   0,   0,   0,   0,   0,   0 },    242.41534,   0.00003,     0.423 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,  -2,   0,   0 },    254.91384,   0.00003,     0.046 },
 { {   0,   0,  13, -19,   0,   0,   0,   0,   2,   0,   0 },     91.03733,   0.00003,     0.036 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,  -2,   0,   0 },    127.88164,   0.00003,     0.047 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,   0,   0 },    119.81787,   0.00001,     0.036 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,  -2,   0,   0 },    157.08151,   0.00003,     0.048 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,  -2,   2,   0 },    156.99107,   0.00001,     0.174 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   0,   0,   0 },    153.60517,   0.00006,     0.252 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,   0,   0 },    148.80073,   0.00002,     0.035 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,  -2,   0,   0 },      4.95771,   0.00004,     0.049 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,  -2,   1,   0 },    185.68436,   0.00001,     0.143 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   0,   0,   0 },    241.50751,   0.00002,     0.223 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,  -1,   0 },    357.56425,   0.00003,     0.063 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,   0,   0 },    177.53817,   0.00002,     0.034 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,  -2,   0,   0 },     34.15123,   0.00009,     0.051 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,  -2,   2,   0 },     33.67462,   0.00001,     0.147 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   0,   0,   0 },    201.53923,   0.00003,     0.199 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,  -1,   0 },     25.83054,   0.00002,     0.061 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,   0,   0 },    205.81689,   0.00003,     0.034 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,  -2,   0,   0 },     62.93046,   0.00005,     0.052 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,  -2,   1,   0 },    242.90552,   0.00001,     0.169 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   0,   0,   0 },     60.17256,   0.00001,     0.180 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,  -2,   0 },     52.61899,   0.00001,     0.264 },
 { {   0,   0,  13, -14,   0,   0,   0,   0,   2,   0,   0 },    232.61899,   0.00001,     0.033 },
 { {   0,   0,  13, -13,   0,   0,   0,   0,   0,   0,   0 },    273.12160,   0.00003,     0.164 },
 { {   0,   0,  13, -13,   0,   0,   0,   0,   2,   0,   0 },    105.36150,   0.00003,     0.032 },
 { {   0,   0,  13, -12,   0,   0,   0,   0,   2,   0,   0 },    314.13101,   0.00004,     0.032 },
 { {   0,   0,  13, -11,   0,   0,   0,   0,   2,  -3,   0 },    163.32045,   0.00001,     0.127 },
 { {   0,   0,  13, -11,   0,   0,   0,   0,   2,  -2,   0 },    162.64597,   0.00002,     0.186 },
 { {   0,   0,  13, -11,   0,   0,   0,   0,   2,  -1,   0 },    343.32045,   0.00001,     0.054 },
 { {   0,   0,  13, -10,   0,   0,   0,   0,   2,  -1,   0 },     12.19535,   0.00002,     0.052 },
 { {   0,   0,  14, -26,   0,   0,   0,   0,   0,   0,   0 },    264.50692,   0.00002,     5.675 },
 { {   0,   0,  14, -24,   0,   0,   0,   0,  -2,   0,   0 },    142.22837,   0.00003,     0.043 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,  -2,   0,   0 },    170.93804,   0.00003,     0.044 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   0 },    187.22667,   0.00057,   362.998 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,   0 },    136.53714,   0.00004,     0.075 },
 { {   0,   0,  14, -22,   0,   0,   0,   0,  -2,   0,   0 },    199.41552,   0.00002,     0.045 },
 { {   0,   0,  14, -22,   0,   0,   0,   0,  -2,   2,   0 },    199.41552,   0.00002,     0.245 },
 { {   0,   0,  14, -22,   0,   0,   0,   0,   2,  -3,   0 },     33.04273,   0.00001,     0.079 },
 { {   0,   0,  14, -22,   0,   0,   0,   0,   2,  -1,   0 },    213.04273,   0.00001,     0.073 },
 { {   0,   0,  14, -21,   0,   0,   0,   0,  -2,   0,   0 },    227.38286,   0.00003,     0.046 },
 { {   0,   0,  14, -20,   0,   0,   0,   0,   0,   0,   0 },    271.32180,   0.00003,     0.297 },
 { {   0,   0,  14, -20,   0,   0,   0,   0,   2,  -1,   0 },    270.95835,   0.00002,     0.067 },
 { {   0,   0,  14, -20,   0,   0,   0,   0,   2,   0,   0 },     90.95835,   0.00003,     0.036 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,  -2,   0,   0 },    127.80044,   0.00003,     0.048 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,  -1,   0 },    299.87126,   0.00003,     0.065 },
 { {   0,   0,  14, -19,   0,   0,   0,   0,   2,   0,   0 },    299.87126,   0.00001,     0.035 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,  -2,   0,   0 },    156.93801,   0.00004,     0.049 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   0,   0,   0 },    130.84516,   0.00003,     0.226 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,  -1,   0 },    328.68760,   0.00003,     0.063 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,   0,   0 },    148.60426,   0.00002,     0.034 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,  -2,   0,   0 },      4.77144,   0.00002,     0.051 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,  -2,   1,   0 },    185.52743,   0.00001,     0.153 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,  -2,   2,   0 },    184.77144,   0.00002,     0.148 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   0,   0,   0 },    183.58213,   0.00006,     0.202 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,  -1,   0 },    357.28525,   0.00002,     0.061 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,   0,   0 },    177.24573,   0.00001,     0.034 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,  -2,   0,   0 },     33.46544,   0.00003,     0.052 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,  -2,   2,   0 },    213.04476,   0.00001,     0.138 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   0,   0,   0 },    296.87962,   0.00002,     0.182 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,   2,  -2,   0 },    205.32445,   0.00002,     0.269 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,  -2,   0,   0 },     62.37959,   0.00001,     0.053 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,  -2,   2,   0 },     62.37959,   0.00001,     0.128 },
 { {   0,   0,  14, -15,   0,   0,   0,   0,   0,   0,   0 },    269.64243,   0.00004,     0.166 },
 { {   0,   0,  14, -14,   0,   0,   0,   0,   0,   0,   0 },    233.26056,   0.00002,     0.153 },
 { {   0,   0,  14, -14,   0,   0,   0,   0,   2,   0,   0 },    105.30355,   0.00003,     0.032 },
 { {   0,   0,  14, -13,   0,   0,   0,   0,   2,  -2,   0 },    133.99431,   0.00002,     0.188 },
 { {   0,   0,  14, -13,   0,   0,   0,   0,   2,   0,   0 },    313.99431,   0.00002,     0.031 },
 { {   0,   0,  14, -12,   0,   0,   0,   0,   2,  -2,   0 },    343.09262,   0.00003,     0.171 },
 { {   0,   0,  14, -12,   0,   0,   0,   0,   2,   0,   0 },    343.22282,   0.00002,     0.031 },
 { {   0,   0,  14, -11,   0,   0,   0,   0,   2,   0,   0 },    191.83617,   0.00002,     0.030 },
 { {   0,   0,  14,  -9,   0,   0,   0,   0,   2,   0,   0 },     67.79706,   0.00001,     0.029 },
 { {   0,   0,  15, -28,   0,   0,   0,   0,   0,   0,   0 },    235.50622,   0.00001,     8.862 },
 { {   0,   0,  15, -25,   0,   0,   0,   0,  -2,   0,   0 },    142.06053,   0.00002,     0.043 },
 { {   0,   0,  15, -24,   0,   0,   0,   0,  -2,   0,   0 },    170.70121,   0.00003,     0.044 },
 { {   0,   0,  15, -23,   0,   0,   0,   0,  -2,   0,   0 },    199.06296,   0.00003,     0.046 },
 { {   0,   0,  15, -22,   0,   0,   0,   0,  -2,   0,   0 },    226.80119,   0.00002,     0.047 },
 { {   0,   0,  15, -21,   0,   0,   0,   0,   2,   0,   0 },     90.86499,   0.00002,     0.035 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,  -2,   0,   0 },    127.70662,   0.00003,     0.049 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,   0,   0 },    119.73739,   0.00002,     0.034 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,  -2,   0,   0 },    156.71998,   0.00006,     0.050 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,  -2,   0 },    148.48507,   0.00002,     0.320 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,  -2,   1,   0 },    185.34567,   0.00001,     0.165 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   0,   0,   0 },    185.06884,   0.00004,     0.184 },
 { {   0,   0,  15, -18,   0,   0,   0,   0,   2,  -2,   0 },    176.95855,   0.00002,     0.274 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,  -2,   0,   0 },     33.53673,   0.00002,     0.053 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   0,   0,   0 },    213.83706,   0.00004,     0.168 },
 { {   0,   0,  15, -17,   0,   0,   0,   0,   2,   0,   0 },    204.72456,   0.00002,     0.033 },
 { {   0,   0,  15, -16,   0,   0,   0,   0,  -2,   0,   0 },     62.07678,   0.00002,     0.055 },
 { {   0,   0,  15, -16,   0,   0,   0,   0,   0,   0,   0 },    242.33797,   0.00004,     0.154 },
 { {   0,   0,  15, -15,   0,   0,   0,   0,   2,   0,   0 },    105.23658,   0.00002,     0.031 },
 { {   0,   0,  15, -13,   0,   0,   0,   0,   2,   0,   0 },    342.89244,   0.00005,     0.030 },
 { {   0,   0,  15, -12,   0,   0,   0,   0,   2,  -2,   0 },     11.64001,   0.00002,     0.146 },
 { {   0,   0,  15, -12,   0,   0,   0,   0,   2,   0,   0 },     11.68405,   0.00003,     0.030 },
 { {   0,   0,  16, -30,   0,   0,   0,   0,   0,   0,   0 },    206.50763,   0.00001,    20.213 },
 { {   0,   0,  16, -25,   0,   0,   0,   0,  -2,   0,   0 },    170.43390,   0.00002,     0.045 },
 { {   0,   0,  16, -24,   0,   0,   0,   0,  -2,   0,   0 },    198.66394,   0.00002,     0.047 },
 { {   0,   0,  16, -22,   0,   0,   0,   0,   2,  -2,   0 },    270.75532,   0.00001,     0.395 },
 { {   0,   0,  16, -22,   0,   0,   0,   0,   2,   0,   0 },     90.75532,   0.00001,     0.034 },
 { {   0,   0,  16, -21,   0,   0,   0,   0,  -2,   0,   0 },    127.59864,   0.00003,     0.050 },
 { {   0,   0,  16, -21,   0,   0,   0,   0,   2,   0,   0 },    119.58046,   0.00003,     0.034 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,  -2,   0,   0 },    156.56447,   0.00003,     0.052 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,  -2,   2,   0 },    336.56447,   0.00003,     0.140 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,   0,   0 },    148.24944,   0.00002,     0.033 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,  -2,   0,   0 },      5.13730,   0.00002,     0.053 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,  -2,   1,   0 },    185.13730,   0.00001,     0.179 },
 { {   0,   0,  16, -19,   0,   0,   0,   0,   2,   0,   0 },    176.57903,   0.00003,     0.033 },
 { {   0,   0,  16, -17,   0,   0,   0,   0,  -2,   0,   0 },     61.63920,   0.00002,     0.056 },
 { {   0,   0,  16, -16,   0,   0,   0,   0,   2,  -2,   0 },    285.15760,   0.00001,     0.175 },
 { {   0,   0,  16, -16,   0,   0,   0,   0,   2,   0,   0 },    105.15760,   0.00001,     0.031 },
 { {   0,   0,  16, -15,   0,   0,   0,   0,   2,   0,   0 },    314.07051,   0.00001,     0.031 },
 { {   0,   0,  16, -12,   0,   0,   0,   0,   2,   0,   0 },     39.52371,   0.00001,     0.029 },
 { {   0,   0,  17, -32,   0,   0,   0,   0,   0,   0,   0 },    177.43281,   0.00003,    71.966 },
 { {   0,   0,  17, -30,   0,   0,   0,   0,   1,   0,  -1 },    127.79195,   0.00004,   232.951 },
 { {   0,   0,  17, -26,   0,   0,   0,   0,  -2,   0,   0 },    170.13386,   0.00001,     0.046 },
 { {   0,   0,  17, -26,   0,   0,   0,   0,  -2,   2,   0 },    350.13386,   0.00001,     0.202 },
 { {   0,   0,  17, -22,   0,   0,   0,   0,  -2,   0,   0 },    127.47514,   0.00003,     0.051 },
 { {   0,   0,  17, -22,   0,   0,   0,   0,   2,   0,   0 },    119.39870,   0.00002,     0.033 },
 { {   0,   0,  17, -21,   0,   0,   0,   0,   2,   0,   0 },    147.97804,   0.00003,     0.033 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,  -2,   0,   0 },    184.98776,   0.00003,     0.054 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,  -2,   1,   0 },    184.90046,   0.00001,     0.195 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,  -2,   2,   0 },    184.98776,   0.00003,     0.123 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,   2,  -2,   0 },    356.14045,   0.00001,     0.218 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,   2,   0,   0 },    176.14045,   0.00001,     0.032 },
 { {   0,   0,  17, -19,   0,   0,   0,   0,  -2,   0,   0 },     33.26221,   0.00001,     0.056 },
 { {   0,   0,  17, -16,   0,   0,   0,   0,   2,  -2,   0 },    133.93665,   0.00001,     0.149 },
 { {   0,   0,  17, -15,   0,   0,   0,   0,   2,   0,   0 },    342.68432,   0.00001,     0.030 },
 { {   0,   0,  17, -11,   0,   0,   0,   0,  -2,  -1,   2 },    255.25925,   0.00001,   307.754 },
 { {   0,   0,  18, -23,   0,   0,   0,   0,  -2,   0,   0 },    127.33491,   0.00002,     0.053 },
 { {   0,   0,  18, -21,   0,   0,   0,   0,  -2,   0,   0 },    184.70038,   0.00005,     0.056 },
 { {   0,   0,  18, -21,   0,   0,   0,   0,  -2,   1,   0 },    184.63315,   0.00001,     0.215 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,  -2,   0,   0 },    212.83122,   0.00003,     0.058 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,  -2,   2,   0 },    212.83122,   0.00003,     0.109 },
 { {   0,   0,  18, -17,   0,   0,   0,   0,   2,   0,   0 },    133.77971,   0.00002,     0.030 },
 { {   0,   0,  18, -16,   0,   0,   0,   0,   2,  -2,   0 },    162.44870,   0.00001,     0.130 },
 { {   0,   0,  18, -16,   0,   0,   0,   0,   2,   0,   0 },    162.44870,   0.00001,     0.029 },
 { {   0,   0,  18, -15,   0,   0,   0,   0,   2,   0,   0 },    190.77828,   0.00002,     0.029 },
 { {   0,   0,  19, -24,   0,   0,   0,   0,  -2,   0,   0 },    127.17685,   0.00001,     0.054 },
 { {   0,   0,  19, -24,   0,   0,   0,   0,  -2,   2,   0 },    307.17685,   0.00001,     0.125 },
 { {   0,   0,  19, -21,   0,   0,   0,   0,  -2,   0,   0 },    212.33767,   0.00002,     0.059 },
 { {   0,   0,  19, -21,   0,   0,   0,   0,  -2,   2,   0 },     32.33767,   0.00002,     0.104 },
 { {   0,   0,  20, -25,   0,   0,   0,   0,  -2,   0,   0 },    126.99989,   0.00001,     0.055 },
 { {   0,   0,  20, -25,   0,   0,   0,   0,  -2,   2,   0 },    306.99989,   0.00001,     0.118 },
 { {   0,   0,  20, -18,   0,   0,   0,   0,  -3,   0,   1 },    208.41212,   0.00001,     0.075 },
 { {   0,   0,  20, -18,   0,   0,   0,   0,  -3,   1,   1 },    208.41212,   0.00020,   584.305 },
 { {   0,   0,  20, -18,   0,   0,   0,   0,  -3,   2,   1 },    208.41212,   0.00001,     0.075 },
 { {   0,   0,  20, -16,   0,   0,   0,   0,  -2,   1,   0 },    273.88190,   0.00002,    88.385 },
 { {   0,   0,  21, -24,   0,   0,   0,   0,  -2,   1,   0 },    183.62426,   0.00001,     0.308 },
 { {   0,   0,  21, -18,   0,   0,   0,   0,  -2,   1,   0 },      2.12334,   0.00006,    19.211 },
 { {   0,   0,  22, -26,   0,   0,   0,   0,  -2,   0,   0 },    155.05940,   0.00001,     0.060 },
 { {   0,   0,  22, -26,   0,   0,   0,   0,  -2,   2,   0 },    155.05940,   0.00001,     0.101 },
 { {   0,   0,  22, -25,   0,   0,   0,   0,  -2,   1,   0 },    183.20907,   0.00001,     0.361 },
 { {   0,   0,  22, -20,   0,   0,   0,   0,  -2,   1,   0 },    158.94815,   0.00004,     8.664 },
 { {   0,   0,  23, -27,   0,   0,   0,   0,  -2,   0,   0 },    154.73587,   0.00002,     0.062 },
 { {   0,   0,  23, -23,   0,   0,   0,   0,  -2,   1,   0 },     72.25049,   0.00001,     1.408 },
 { {   0,   0,  23, -22,   0,   0,   0,   0,  -2,   1,   0 },    309.15553,   0.00002,     5.593 },
 { {   0,   0,  23, -20,   0,   0,   0,   0,  -1,   0,   0 },      2.10578,   0.00002,   411.317 },
 { {   0,   0,  24, -28,   0,   0,   0,   0,  -2,   0,   0 },    154.37991,   0.00001,     0.064 },
 { {   0,   0,  24, -28,   0,   0,   0,   0,  -2,   2,   0 },    334.37991,   0.00001,     0.092 },
 { {   0,   0,  24, -27,   0,   0,   0,   0,  -2,   0,   0 },    182.23674,   0.00001,     0.066 },
 { {   0,   0,  24, -27,   0,   0,   0,   0,  -2,   2,   0 },    182.23674,   0.00001,     0.088 },
 { {   0,   0,  24, -24,   0,   0,   0,   0,  -2,   1,   0 },    238.97560,   0.00003,     4.130 },
 { {   0,   0,  24, -23,   0,   0,   0,   0,  -2,   1,   0 },    132.62908,   0.00002,     3.454 },
 { {   0,   0,  25, -29,   0,   0,   0,   0,  -2,   1,   0 },    153.98902,   0.00001,     0.526 },
 { {   0,   0,  25, -28,   0,   0,   0,   0,  -2,   0,   0 },    181.66892,   0.00001,     0.068 },
 { {   0,   0,  25, -28,   0,   0,   0,   0,  -2,   2,   0 },    181.66892,   0.00001,     0.084 },
 { {   0,   0,  25, -26,   0,   0,   0,   0,  -2,   1,   0 },    230.75604,   0.00002,     3.273 },
 { {   0,   0,  25, -25,   0,   0,   0,   0,  -2,   1,   0 },    225.88579,   0.00002,     4.422 },
 { {   0,   0,  25, -22,   0,   0,   0,   0,   0,  -1,   0 },      2.97619,   0.00003,    21.191 },
 { {   0,   0,  26, -27,   0,   0,   0,   0,  -2,   1,   0 },    226.39043,   0.00002,     6.143 },
 { {   0,   0,  26, -24,   0,   0,   0,   0,   0,  -1,   0 },    162.60953,   0.00006,    61.818 },
 { {   0,   0,  26, -20,   0,   0,   0,   0,   2,  -1,  -2 },    110.28992,   0.00005,   330.156 },
 { {   0,   0,  27, -32,   0,   0,   0,   0,  -2,   1,   0 },    125.13390,   0.00001,     0.669 },
 { {   0,   0,  27, -31,   0,   0,   0,   0,  -2,   0,   0 },    153.09018,   0.00001,     0.070 },
 { {   0,   0,  27, -31,   0,   0,   0,   0,  -2,   2,   0 },    153.09018,   0.00001,     0.081 },
 { {   0,   0,  27, -30,   0,   0,   0,   0,  -2,   1,   0 },    356.80000,   0.00001,     2.313 },
 { {   0,   0,  27, -26,   0,   0,   0,   0,   0,  -1,   0 },    352.51668,   0.00001,    12.572 },
 { {   0,   0,  28, -32,   0,   0,   0,   0,  -2,   1,   0 },    331.74936,   0.00003,     2.018 },
 { {   0,   0,  28, -31,   0,   0,   0,   0,  -2,   1,   0 },    163.29991,   0.00005,    27.739 },
 { {   0,   0,  28, -29,   0,   0,   0,   0,  -1,   1,  -1 },    310.99597,   0.00002,    21.904 },
 { {   0,   0,  28, -28,   0,   0,   0,   0,   0,  -1,   0 },    283.32450,   0.00002,     6.997 },
 { {   0,   0,  29, -35,   0,   0,   0,   0,  -2,   1,   0 },     96.22138,   0.00001,     0.917 },
 { {   0,   0,  29, -34,   0,   0,   0,   0,  -2,   0,   0 },    124.35509,   0.00001,     0.072 },
 { {   0,   0,  29, -34,   0,   0,   0,   0,  -2,   1,   0 },    302.76516,   0.00001,     1.789 },
 { {   0,   0,  29, -34,   0,   0,   0,   0,  -2,   2,   0 },    124.35509,   0.00001,     0.079 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   0,   0 },    335.12452,   0.00002,     0.075 },
 { {   0,   0,  29, -33,   0,   0,   0,   0,  -2,   1,   0 },    164.61805,   0.00010,    36.607 },
 { {   0,   0,  29, -31,   0,   0,   0,   0,  -1,   1,  -1 },    325.10220,   0.00004,    56.451 },
 { {   0,   0,  29, -29,   0,   0,   0,   0,   0,   1,  -2 },    276.74304,   0.00004,   123.278 },
 { {   0,   0,  30, -35,   0,   0,   0,   0,  -2,   0,   0 },    302.14790,   0.00003,     0.075 },
 { {   0,   0,  30, -35,   0,   0,   0,   0,  -2,   1,   0 },    100.84631,   0.00004,    11.027 },
 { {   0,   0,  30, -32,   0,   0,   0,   0,   0,  -1,   0 },    203.30541,   0.00002,     3.709 },
 { {   0,   0,  31, -37,   0,   0,   0,   0,  -2,   1,   0 },     87.28854,   0.00007,     6.491 },
 { {   0,   0,  31, -36,   0,   0,   0,   0,  -2,   0,   0 },    303.43665,   0.00001,     0.078 },
 { {   0,   0,  31, -36,   0,   0,   0,   0,  -2,   1,   0 },    301.50538,   0.00001,     2.648 },
 { {   0,   0,  31, -36,   0,   0,   0,   0,  -2,   2,   0 },    123.43665,   0.00001,     0.073 },
 { {   0,   0,  31, -35,   0,   0,   0,   0,  -1,   0,   0 },    139.98011,   0.00001,    44.834 },
 { {   0,   0,  31, -34,   0,   0,   0,   0,   0,  -1,   0 },    283.09021,   0.00001,     3.003 },
 { {   0,   0,  32, -39,   0,   0,   0,   0,  -2,   0,   0 },     66.96841,   0.00001,     0.074 },
 { {   0,   0,  32, -39,   0,   0,   0,   0,  -2,   1,   0 },    246.30668,   0.00002,     4.599 },
 { {   0,   0,  32, -39,   0,   0,   0,   0,  -2,   2,   0 },     66.96841,   0.00001,     0.077 },
 { {   0,   0,  32, -38,   0,   0,   0,   0,  -2,   1,   0 },    270.87146,   0.00003,     3.182 },
 { {   0,   0,  33, -40,   0,   0,   0,   0,  -2,   1,   0 },    247.00413,   0.00004,     3.986 },
 { {   0,   0,  33, -37,   0,   0,   0,   0,   0,  -1,   0 },      3.93918,   0.00001,    13.903 },
 { {   0,   0,  34, -43,   0,   0,   0,   0,  -2,  -1,   2 },    348.90192,   0.00002,    93.931 },
 { {   0,   0,  34, -43,   0,   0,   0,   0,  -2,   1,   0 },      9.64559,   0.00002,     2.906 },
 { {   0,   0,  34, -42,   0,   0,   0,   0,  -2,   1,   0 },    216.14231,   0.00006,     5.333 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,  -2,   1 },    145.60826,   0.00001,     0.075 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,  -1,   1 },    145.60826,   0.00023,   958.407 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,   0,   1 },    145.60826,   0.00001,     0.075 },
 { {   0,   0,  34, -39,   0,   0,   0,   0,   0,  -1,   0 },    321.59415,   0.00006,   116.831 },
 { {   0,   0,  34, -39,   0,   0,   0,   0,   0,   0,   0 },    121.94508,   0.00002,     0.075 },
 { {   0,   0,  34, -37,   0,   0,   0,   0,   1,  -1,  -1 },    192.00736,   0.00001,    55.060 },
 { {   0,   0,  35, -44,   0,   0,   0,   0,  -2,   1,   0 },    146.09153,   0.00001,     8.055 },
 { {   0,   0,  35, -41,   0,   0,   0,   0,   0,  -1,   0 },    291.69480,   0.00001,    18.245 },
 { {   0,   0,  36, -46,   0,   0,   0,   0,  -2,   1,   0 },    330.91251,   0.00002,    16.451 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   0,   0 },    130.62847,   0.00003,     0.075 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   1,   0 },    338.03715,   0.00009,   387.400 },
 { {   0,   0,  37, -46,   0,   0,   0,   0,  -1,   1,  -1 },    339.49277,   0.00002,   142.421 },
 { {   0,   0,  37, -44,   0,   0,   0,   0,   0,   1,  -2 },     57.34447,   0.00002,    60.153 },
 { {   0,   0,  38, -50,   0,   0,   0,   0,  -2,   1,   0 },     97.76801,   0.00004,    15.163 },
 { {   0,   0,  39, -52,   0,   0,   0,   0,  -2,   1,   0 },    230.78360,   0.00004,     7.733 },
 { {   0,   0,  40, -54,   0,   0,   0,   0,  -2,   1,   0 },     20.79525,   0.00001,     5.190 },
 { {   0,   0,  42, -54,   0,   0,   0,   0,   0,  -1,   0 },    246.21724,   0.00004,    30.034 },
 { {   0,   0,  43, -56,   0,   0,   0,   0,   0,  -1,   0 },    228.92027,   0.00002,    33.253 },
 { {   0,   0,  44, -61,   0,   0,   0,   0,  -2,   1,   0 },    295.18201,   0.00002,    11.693 },
 { {   0,   0,  45, -63,   0,   0,   0,   0,  -2,   1,   0 },    104.45353,   0.00001,    45.137 },
 { {   0,   0,  46, -65,   0,   0,   0,   0,  -2,   1,   0 },     57.93969,   0.00003,    24.264 },
 { {   0,   0,  51, -71,   0,   0,   0,   0,   0,  -1,   0 },     68.68964,   0.00001,   187.401 },
 { {   0,   1,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    237.93464,   0.00002,     0.032 },
 { {   0,   1,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     55.12760,   0.00002,     0.056 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    159.04683,   0.00001,     0.031 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.06990,   0.00001,     0.054 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    340.20497,   0.00006,     0.186 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },     70.47271,   0.00006,     0.019 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -4,   1,   0 },    250.21061,   0.00001,     0.025 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,  -2,   0 },     76.45194,   0.00002,     0.018 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },     70.47271,   0.00003,     0.024 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    264.94565,   0.00004,     0.034 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   0,   2 },     70.47271,   0.00003,     0.441 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },      9.14843,   0.00001,     0.063 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },     98.35052,   0.00004,     0.385 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,  -3,   0 },     70.21061,   0.00001,     0.023 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    251.98361,   0.00001,     0.057 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     70.21061,   0.00003,     0.229 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    256.45190,   0.00002,     0.113 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     18.08649,   0.00001,     0.036 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    190.68014,   0.00007,     0.067 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },     61.73109,   0.00001,     0.625 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    339.42181,   0.00002,     0.033 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    161.15086,   0.00005,     0.296 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    268.95182,   0.00002,     0.047 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },     32.81257,   0.00001,     0.025 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    144.92466,   0.00002,     0.037 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   2 },    177.26229,   0.00003,     3.746 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     36.05892,   0.00008,     0.072 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },     29.12875,   0.00001,     0.079 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    167.94221,   0.00002,     0.035 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     92.27087,   0.00002,     0.064 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    282.33974,   0.00003,     0.421 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    172.05046,   0.00002,     0.092 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },      3.37989,   0.00001,     0.041 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    348.00750,   0.00001,     0.241 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    183.37989,   0.00004,     0.110 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    348.00750,   0.00001,     0.045 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },      4.71253,   0.00003,     0.038 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    339.10825,   0.00007,     0.078 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },     81.70639,   0.00001,     2.503 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   3,   0 },    241.38144,   0.00002,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },    298.16297,   0.00004,     0.074 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },      3.76314,   0.00004,     3.118 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    298.16297,   0.00004,     0.077 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    241.21795,   0.00005,     0.068 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     60.31176,   0.00012,     0.728 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    231.96104,   0.00005,     0.084 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    295.29056,   0.00004,     0.412 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },     69.64038,   0.00003,     0.061 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    253.24131,   0.00002,     0.099 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    108.21808,   0.00003,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    253.07077,   0.00001,     0.019 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    222.04182,   0.00001,     0.026 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    138.24040,   0.00009,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    272.64309,   0.00006,     0.084 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    202.21862,   0.00005,     0.715 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -2,   1 },    304.16084,   0.00002,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    180.38629,   0.00002,     0.080 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },     73.17937,   0.00002,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },     13.94859,   0.00005,     1.472 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },     18.17852,   0.00006,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },    180.38629,   0.00002,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,  -1 },      3.98463,   0.00003,     2.892 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,   1 },      3.98463,   0.00002,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -3,   2 },    199.50609,   0.00002,     0.075 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    359.10606,   0.00001,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   2 },    191.66664,   0.00003,    24.413 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    300.36488,   0.00006,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    191.55300,   0.00007,     0.076 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    248.84996,   0.00027,     2.670 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    231.62449,   0.00003,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,  -1 },    184.00908,   0.00001,     0.036 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,   1 },     77.48486,   0.00002,     0.913 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    203.70408,   0.00004,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },     41.86021,   0.00006,     0.082 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    344.36281,   0.00005,     0.700 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    188.42332,   0.00002,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    203.70408,   0.00004,     0.085 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   2,  -1 },    180.11553,   0.00003,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    188.83984,   0.00002,     0.465 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     26.45087,   0.00006,     0.090 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },     18.11614,   0.00002,     0.403 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     10.04136,   0.00002,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   1,   0 },    198.18672,   0.00005,     0.027 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },     15.65681,   0.00002,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   3,   0 },     18.46289,   0.00001,     0.016 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    198.11614,   0.00003,     0.044 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -1,   0 },    198.46289,   0.00001,     0.028 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    198.11614,   0.00003,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   3,   0 },    268.78542,   0.00001,     0.110 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    147.67739,   0.00007,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    271.26418,   0.00011,     0.092 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    283.90058,   0.00002,     0.417 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },     75.17521,   0.00001,     0.064 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    111.05739,   0.00001,     0.040 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    193.51503,   0.00001,     0.596 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,  -1 },    291.05739,   0.00001,     0.743 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    112.71672,   0.00004,     0.079 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    321.66506,   0.00009,     1.599 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     97.46075,   0.00005,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   1 },     78.49336,   0.00003,    10.557 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    248.38540,   0.00001,     0.073 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },    305.58851,   0.00002,     0.076 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },     90.42922,   0.00011,     2.335 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },     68.97006,   0.00001,     0.038 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },    248.38540,   0.00001,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    287.45829,   0.00001,     0.871 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    291.10129,   0.00014,     0.083 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     36.83157,   0.00003,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },    159.60029,   0.00006,     0.021 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0 },    159.76262,   0.00001,     0.029 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   0 },    159.60029,   0.00003,     0.027 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    330.26191,   0.00004,     0.043 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   2 },    339.76262,   0.00001,     0.268 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },     20.42747,   0.00004,     0.101 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    302.51515,   0.00006,     0.294 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   3,   0 },    355.54309,   0.00002,     0.060 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1 },    359.88529,   0.00003,     0.095 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },    154.90262,   0.00004,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   1 },    359.88529,   0.00003,     0.063 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   1 },    359.88529,   0.00002,     0.034 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    339.87950,   0.00011,     0.086 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    283.95544,   0.00005,     0.615 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    133.28751,   0.00009,     0.067 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    227.46391,   0.00001,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,  -1 },    185.41196,   0.00001,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },     75.04338,   0.00003,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    211.05127,   0.00003,     1.749 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   1 },    241.34006,   0.00004,     0.036 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   2,  -1 },    158.11077,   0.00002,     0.037 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    195.44819,   0.00002,     0.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    243.63467,   0.00006,     6.734 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    183.68179,   0.00015,     0.076 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -2 },     52.11306,   0.00001,     2.075 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    101.47110,   0.00002,     0.038 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    161.06485,   0.00001,     0.025 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,  -2,   0 },     71.87100,   0.00003,     0.021 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.51257,   0.00010,     0.045 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    282.18411,   0.00012,     0.113 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    244.77709,   0.00003,     0.272 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -2,   0 },    270.17206,   0.00002,     0.042 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    278.99684,   0.00002,     0.094 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    145.62027,   0.00005,     0.381 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    278.99684,   0.00002,     0.063 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    204.00500,   0.00001,     0.636 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    352.98901,   0.00002,     1.174 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },     83.25837,   0.00004,     0.071 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    316.79985,   0.00002,     0.104 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0 },     73.23439,   0.00003,     0.276 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    192.49446,   0.00001,     0.059 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   0,   2,   0 },    123.06770,   0.00003,     0.033 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },     35.56497,   0.00002,     0.540 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    261.41376,   0.00006,     0.066 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    337.62855,   0.00007,     0.035 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   2,   0 },      5.55511,   0.00003,     0.018 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   4,   0,   0 },    173.47129,   0.00001,     0.019 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,  -2,   0,   0 },    271.88307,   0.00001,     0.050 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,  -2,   2,   0 },    271.88307,   0.00001,     0.156 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     20.45769,   0.00009,     0.216 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    200.62169,   0.00001,     0.056 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -3,   0 },    194.63139,   0.00001,     0.096 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },     58.64320,   0.00001,     0.351 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    336.69242,   0.00004,     0.062 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     52.01281,   0.00003,     0.034 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0 },    288.01491,   0.00002,     0.178 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   0,   1,   0 },    288.01491,   0.00003,     0.053 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   0,   3,   0 },    108.01491,   0.00001,     0.022 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },    103.54401,   0.00002,     0.058 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   0,  -2 },    108.01491,   0.00001,     0.284 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    273.58577,   0.00005,     0.033 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   1,   0 },    108.01491,   0.00001,     0.023 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   2,   0 },    283.54402,   0.00002,     0.018 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,  -1,   0 },    198.95755,   0.00001,     0.151 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,   0,   0 },    218.65009,   0.00004,     0.151 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,   1,   0 },    198.95755,   0.00001,     0.050 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   0,   2,   0 },     17.18636,   0.00002,     0.030 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   2,  -1,   0 },     67.70491,   0.00003,     0.055 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   2,   0,   0 },     81.95901,   0.00005,     0.032 },
 { {   0,   1,   6,   0,   0,   0,   0,   0,   2,  -2,   0 },    126.18434,   0.00003,     0.171 },
 { {   0,   1,   6,   0,   0,   0,   0,   0,   2,   0,   0 },    128.15219,   0.00002,     0.031 },
 { {   0,   2,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    236.75680,   0.00003,     0.033 },
 { {   0,   2,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     56.16034,   0.00003,     0.058 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.51105,   0.00003,     0.032 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.11980,   0.00001,     0.056 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    341.51105,   0.00003,     0.211 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },     70.42918,   0.00006,     0.019 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -4,   1,   0 },    250.20932,   0.00001,     0.025 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },     70.42918,   0.00003,     0.024 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    232.98130,   0.00002,     0.035 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     65.16371,   0.00004,     0.066 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     70.20932,   0.00001,     0.506 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     70.20932,   0.00002,     0.267 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    256.44744,   0.00002,     0.105 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    241.13389,   0.00003,     0.070 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    159.01699,   0.00001,     1.026 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    191.34340,   0.00001,     0.081 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    190.35234,   0.00003,     0.364 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    269.46614,   0.00002,     0.045 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.74835,   0.00007,     0.038 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   2 },    328.89788,   0.00002,     2.789 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    165.98086,   0.00006,    39.899 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    281.38693,   0.00002,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },     41.22923,   0.00004,     1.438 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     41.27098,   0.00002,     0.067 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    347.31228,   0.00006,     0.572 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },    184.41262,   0.00002,     0.060 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    349.93304,   0.00002,     0.284 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    184.41262,   0.00004,     0.103 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    349.93304,   0.00002,     0.044 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    347.36166,   0.00003,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    265.39157,   0.00007,     0.082 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },     42.61552,   0.00001,     0.976 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    307.97442,   0.00001,     0.070 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    120.03600,   0.00003,     0.077 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },     94.89382,   0.00001,     0.038 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    180.06441,   0.00002,     3.281 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },     30.86964,   0.00002,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    120.03600,   0.00003,     0.074 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,  -1 },     61.07540,   0.00004,    34.842 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   3,  -1 },     68.72621,   0.00004,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    339.70947,   0.00004,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     57.53385,   0.00014,     1.335 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    339.70947,   0.00004,     0.080 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    250.11517,   0.00004,     0.063 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    214.41864,   0.00003,     0.093 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    227.69362,   0.00004,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   0,   0 },     37.76246,   0.00001,     0.020 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -4,   1,   0 },     28.39318,   0.00002,     0.028 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },     20.61094,   0.00005,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     23.49623,   0.00032,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     10.82813,   0.00051,     0.089 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },      4.72372,   0.00006,     0.494 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },     67.08162,   0.00002,     0.084 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },     26.56720,   0.00006,     0.039 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    156.40311,   0.00003,     0.766 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },     67.08162,   0.00002,     0.069 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },      4.78686,   0.00004,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     19.07630,   0.00061,     0.077 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     17.46813,   0.00353,     3.983 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -2 },    184.10687,   0.00002,     0.075 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     19.76925,   0.00050,     0.074 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },     17.57116,   0.00004,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,  -2 },    184.10687,   0.00002,     0.076 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,   0 },     18.46160,   0.00002,     0.025 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,  -1 },    183.79418,   0.00001,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },     18.46160,   0.00001,     0.072 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   1 },    163.80157,   0.00002,     2.131 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },    342.69259,   0.00004,     0.071 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    196.13750,   0.00003,     1.246 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    198.46160,   0.00001,     0.079 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     95.50990,   0.00003,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    342.69259,   0.00004,     0.080 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },     45.96315,   0.00002,     0.068 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },     97.44246,   0.00002,     0.657 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     16.76641,   0.00034,     0.085 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },     18.17297,   0.00003,     0.539 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     17.09378,   0.00018,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    198.52933,   0.00002,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    195.09805,   0.00002,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,  -1,   0 },    198.46160,   0.00001,     0.027 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    198.17297,   0.00005,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    253.29625,   0.00003,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    237.28132,   0.00008,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    285.54060,   0.00003,     0.098 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   2,   0 },    350.35016,   0.00001,     0.331 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },     32.68511,   0.00003,     0.434 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    122.18780,   0.00002,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    133.43811,   0.00002,     0.083 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    280.49517,   0.00008,     0.799 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -3,   1 },     65.85423,   0.00001,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },     59.56759,   0.00002,     0.077 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    245.85423,   0.00001,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    342.41017,   0.00003,     5.069 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },     59.56759,   0.00002,     0.074 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    296.40014,   0.00004,     1.912 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    308.75757,   0.00011,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     27.43604,   0.00006,     0.038 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },    342.09921,   0.00006,     0.021 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   1,   0 },    341.53776,   0.00001,     0.029 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0 },    341.43783,   0.00009,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     16.29036,   0.00008,     0.044 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   2 },    161.53776,   0.00001,     0.229 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    151.19064,   0.00004,     0.108 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    145.98506,   0.00002,     0.248 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   3,   0 },    140.00958,   0.00002,     0.058 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,  -1 },    230.71432,   0.00005,     0.042 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    359.88512,   0.00005,     0.303 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    179.88511,   0.00005,     0.097 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   2,   1 },    359.88510,   0.00002,     0.034 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },     82.19246,   0.00006,     0.091 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    261.81283,   0.00017,     0.444 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     93.69296,   0.00005,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,  -1 },    277.28671,   0.00001,     0.040 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,   1 },     59.21660,   0.00001,     0.653 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    205.75512,   0.00002,     0.083 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1 },     45.56339,   0.00003,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },    218.94335,   0.00009,     0.835 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    277.72548,   0.00002,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   1,  -1 },    205.75512,   0.00002,     0.069 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },    343.86597,   0.00002,     0.078 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },    110.76544,   0.00003,     2.096 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.06938,   0.00005,     0.073 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -2 },     48.25209,   0.00006,     6.968 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    279.71870,   0.00002,     0.037 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.81102,   0.00003,     0.047 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },    130.12067,   0.00002,     0.121 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },     62.93729,   0.00003,     0.100 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    130.45832,   0.00005,     0.308 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     71.80200,   0.00003,     0.061 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    270.97820,   0.00001,     0.034 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    272.55475,   0.00003,     0.455 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    265.85384,   0.00003,     0.085 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },     65.93661,   0.00001,     0.677 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    189.25373,   0.00007,     0.068 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    347.44591,   0.00005,     0.036 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },      1.30914,   0.00003,     0.138 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },      8.41556,   0.00004,     0.111 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    184.81561,   0.00004,     0.235 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     49.83667,   0.00004,     0.057 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   2,   0 },    164.48739,   0.00001,     0.033 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   1,   0,  -1 },    321.51074,   0.00001,     0.313 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },     74.26774,   0.00002,     0.093 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },     78.87845,   0.00001,     0.404 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    208.97776,   0.00008,     0.064 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    353.48209,   0.00006,     0.034 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   1,   0 },    147.18808,   0.00001,     0.024 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },    272.18342,   0.00001,     0.051 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,  -2,   2,   0 },    272.18342,   0.00001,     0.142 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0 },     20.36467,   0.00009,     0.190 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    200.61859,   0.00001,     0.054 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },     61.49754,   0.00003,     0.102 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    168.66316,   0.00003,     0.288 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    327.66206,   0.00002,     0.060 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    200.87584,   0.00002,     0.033 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    109.79004,   0.00001,     0.143 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    109.79004,   0.00004,     0.160 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    109.71005,   0.00006,     0.051 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,  -2,   0 },    103.22966,   0.00005,     0.223 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },    292.22983,   0.00004,     0.056 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,  -2 },    289.79004,   0.00001,     0.241 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    278.00460,   0.00004,     0.032 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   1,   0 },    289.47871,   0.00008,     0.023 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   2,   0 },    283.54384,   0.00002,     0.017 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   4,  -2,   0 },    291.87851,   0.00001,     0.034 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   4,   0,   0 },    111.87851,   0.00001,     0.018 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    224.76610,   0.00004,     0.183 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    238.17762,   0.00004,     0.031 },
 { {   0,   2,   5,   0,   0,   0,   0,   0,   2,  -2,   0 },    125.14479,   0.00002,     0.154 },
 { {   0,   2,   5,   0,   0,   0,   0,   0,   2,   0,   0 },    125.14479,   0.00002,     0.030 },
 { {   0,   2,  10,   0,   0,   0,   0,   0,   0,  -1,   0 },     74.70866,   0.00003,   212.377 },
 { {   0,   3, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    237.71139,   0.00003,     0.033 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    339.41076,   0.00006,     0.064 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    315.17239,   0.00004,     0.092 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.52631,   0.00003,     0.033 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    162.23374,   0.00003,     0.058 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    326.41346,   0.00004,     0.243 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    158.14903,   0.00002,     0.110 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },     70.96943,   0.00004,     0.019 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    250.20656,   0.00001,     0.025 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    248.22796,   0.00001,     0.038 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,  -1,   0 },     71.97639,   0.00001,     0.024 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     67.78814,   0.00003,     0.068 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    244.27182,   0.00003,     0.741 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },     56.98550,   0.00001,     0.084 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },     76.54934,   0.00001,     0.034 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    251.89382,   0.00002,     0.061 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     73.45400,   0.00002,     0.320 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    256.44111,   0.00002,     0.099 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    250.20656,   0.00001,     0.043 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,  -2,   0 },    338.75488,   0.00001,     0.019 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,  -1,   2 },    137.63746,   0.00004,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.36872,   0.00002,     0.037 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.80622,   0.00001,     0.074 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   2 },    137.63746,   0.00004,     0.076 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   3,   0 },    277.72940,   0.00001,     0.077 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    338.75488,   0.00002,     0.035 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    330.17014,   0.00002,     0.065 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    216.11216,   0.00002,     0.471 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    330.17014,   0.00002,     0.090 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.02831,   0.00001,     0.257 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    270.02831,   0.00001,     0.044 },
 { {   0,   3,  -7,   4,   0,   0,   0,   0,   0,   0,   0 },     56.72250,   0.00001,   302.434 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    118.17891,   0.00005,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   2 },     32.80852,   0.00001,     1.016 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    177.14859,   0.00002,     0.079 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },      6.70688,   0.00005,     1.537 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    171.35881,   0.00001,     0.072 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },     37.19607,   0.00005,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },    236.98488,   0.00002,     0.037 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },     54.25868,   0.00021,    14.355 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },     37.19607,   0.00005,     0.076 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    279.38394,   0.00001,     0.036 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     47.41658,   0.00004,     0.070 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },     15.39028,   0.00006,     0.890 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },     56.13301,   0.00001,     0.082 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    351.79727,   0.00001,     0.345 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    185.42097,   0.00002,     0.097 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    345.06497,   0.00006,     0.042 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    223.81127,   0.00004,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    199.48096,   0.00015,     0.086 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },      5.16921,   0.00002,     0.606 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },     87.30413,   0.00001,     1.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    221.45681,   0.00001,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    219.07659,   0.00017,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    169.13282,   0.00049,     8.102 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    215.86401,   0.00018,     0.076 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    215.44316,   0.00003,     0.038 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },      6.09885,   0.00001,     0.527 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    190.50865,   0.00018,     0.088 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    209.72852,   0.00010,     0.041 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   1,   0 },    195.22722,   0.00002,     0.026 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    108.72659,   0.00003,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   0,   0 },    199.10360,   0.00002,     0.021 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -4,   1,   0 },     21.01910,   0.00001,     0.028 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },     24.70838,   0.00004,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     35.03037,   0.00035,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.36084,   0.00031,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    162.91385,   0.00001,     0.088 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -1 },    335.92175,   0.00003,     0.040 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    190.89409,   0.00007,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    155.99394,   0.00004,     0.518 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    210.39961,   0.00003,     0.086 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },    162.91385,   0.00001,     0.066 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },      9.39327,   0.00002,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     13.81478,   0.00038,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     18.08492,   0.00157,     1.141 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     16.61527,   0.00028,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },      8.36303,   0.00002,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },     25.73822,   0.00004,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },    108.36239,   0.00002,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },     14.20298,   0.00057,    95.796 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },    135.93646,   0.00004,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    174.05670,   0.00012,     5.644 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },     11.16568,   0.00038,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    138.90717,   0.00005,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,  -1 },    108.36239,   0.00002,     0.076 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   1,   0 },     10.89455,   0.00002,     0.038 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    356.88165,   0.00001,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    165.59188,   0.00002,     1.114 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     55.36802,   0.00004,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    150.42610,   0.00003,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    198.22105,   0.00002,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,  -1,   0 },    198.45884,   0.00001,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    198.22105,   0.00004,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.41607,   0.00005,     0.044 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   1 },    301.61843,   0.00005,     0.341 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    102.67689,   0.00003,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     90.81909,   0.00003,     0.088 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    289.53952,   0.00019,     0.533 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    167.27829,   0.00002,     0.066 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    288.50382,   0.00001,     0.035 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    260.53596,   0.00001,     1.215 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    194.12291,   0.00002,     0.076 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    134.81127,   0.00004,     9.751 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    247.82718,   0.00018,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -2 },    287.04145,   0.00002,     4.330 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     21.36001,   0.00008,     0.038 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    287.00716,   0.00001,     0.019 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    341.15854,   0.00006,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     99.62915,   0.00002,     0.046 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   2 },    161.53757,   0.00001,     0.201 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    144.11820,   0.00006,     0.116 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },    172.77469,   0.00001,     0.056 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    337.73451,   0.00004,     0.044 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },    313.28069,   0.00002,     0.254 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    179.88382,   0.00002,     0.104 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    322.14856,   0.00010,     0.096 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    258.75436,   0.00015,     0.348 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    227.39335,   0.00001,     0.062 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -3,   1 },    158.54726,   0.00001,     0.090 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -2,   1 },    346.43717,   0.00002,     0.464 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    324.11039,   0.00002,     0.087 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    139.21760,   0.00003,     0.065 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    246.19949,   0.00007,     0.549 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    109.05388,   0.00002,     0.035 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    324.11039,   0.00002,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    175.38472,   0.00002,     0.082 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    206.02217,   0.00005,     0.907 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    317.37626,   0.00005,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -2 },    338.35561,   0.00002,     1.300 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    117.81178,   0.00002,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,  -1 },    347.72281,   0.00002,    20.498 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },     65.37118,   0.00004,     0.075 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.72482,   0.00004,     0.048 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    311.60466,   0.00003,     0.131 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    200.95592,   0.00005,     0.258 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },    237.57842,   0.00001,     0.354 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },     74.30925,   0.00001,     0.476 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    171.50724,   0.00005,     0.065 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     38.03715,   0.00006,     0.035 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },      2.20065,   0.00003,     0.151 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },     19.39753,   0.00001,     0.119 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    194.48418,   0.00005,     0.205 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     70.59666,   0.00001,     0.055 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    178.61147,   0.00010,     0.061 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    137.06954,   0.00003,     0.034 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    221.75688,   0.00001,     0.018 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },     20.64914,   0.00002,     0.136 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     23.98428,   0.00007,     0.170 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    260.88870,   0.00001,     0.244 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },     20.35055,   0.00003,     0.058 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     34.40026,   0.00001,     0.033 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    108.06780,   0.00003,     0.157 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    103.46985,   0.00008,     0.145 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    109.19685,   0.00002,     0.050 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   3,   0 },    109.78985,   0.00001,     0.021 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    108.01575,   0.00002,     0.196 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },     99.93135,   0.00001,     0.054 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   0,  -2 },    289.78985,   0.00001,     0.210 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    283.54426,   0.00002,     0.032 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   1,   0 },    108.01575,   0.00002,     0.022 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   2,   0 },    103.54426,   0.00002,     0.017 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   4,  -2,   0 },    291.73443,   0.00001,     0.034 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   4,  -1,   0 },    289.78985,   0.00001,     0.023 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   4,   0,   0 },    108.82108,   0.00005,     0.018 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     18.78524,   0.00001,     0.127 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     25.55689,   0.00002,     0.031 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   2,   2,   0 },    198.78524,   0.00001,     0.017 },
 { {   0,   3,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    126.71802,   0.00005,     0.030 },
 { {   0,   4, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    238.62640,   0.00003,     0.034 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    175.03590,   0.00001,     0.035 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    355.46517,   0.00004,     0.067 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    311.63064,   0.00002,     0.580 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    321.47167,   0.00002,     0.087 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.54758,   0.00003,     0.033 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.45846,   0.00004,     0.060 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    325.05327,   0.00004,     0.286 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    161.54758,   0.00003,     0.102 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -4,   0,   0 },     70.35089,   0.00003,     0.019 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -4,   2,   0 },    250.35089,   0.00003,     0.039 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.20209,   0.00002,     0.025 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,  -1,   2 },     69.93955,   0.00001,     0.073 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     59.56273,   0.00002,     0.037 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   0,   2 },     70.35089,   0.00003,     2.562 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     18.23406,   0.00003,     0.072 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   1,   2 },     69.93955,   0.00001,     0.078 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    254.87257,   0.00003,     1.381 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    200.04939,   0.00001,     0.080 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },     55.47122,   0.00002,     0.034 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     20.34511,   0.00003,     0.400 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },     70.20209,   0.00001,     0.057 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.20209,   0.00001,     0.111 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    332.93618,   0.00006,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    124.20841,   0.00006,     3.622 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.46061,   0.00003,     0.068 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     89.24707,   0.00005,     0.668 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    327.46061,   0.00003,     0.085 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.62535,   0.00001,     0.306 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    270.62535,   0.00001,     0.043 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    127.71539,   0.00005,     0.040 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     69.56040,   0.00009,     0.083 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     19.48543,   0.00004,     0.784 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    224.88769,   0.00003,     1.799 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    279.44566,   0.00001,     0.037 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    227.17211,   0.00002,     0.073 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     79.11937,   0.00004,     2.008 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    224.59384,   0.00004,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     31.29656,   0.00001,     0.644 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    170.40674,   0.00002,     0.440 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    189.89143,   0.00004,     0.091 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    358.70886,   0.00004,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    210.64496,   0.00003,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    157.43085,   0.00003,     0.091 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    255.59596,   0.00003,     0.439 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    173.32575,   0.00001,     0.064 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    180.13366,   0.00001,     0.643 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    249.88891,   0.00009,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    211.78428,   0.00021,     1.992 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    249.88891,   0.00009,     0.073 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    216.64922,   0.00005,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    110.56874,   0.00001,     0.786 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    107.26761,   0.00004,     0.083 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },    108.49190,   0.00003,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },      6.16459,   0.00001,     0.047 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    136.27517,   0.00005,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    216.26225,   0.00007,     0.100 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    156.21844,   0.00002,     0.305 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    214.19694,   0.00001,     0.060 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },    324.41426,   0.00004,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    137.51978,   0.00002,     0.391 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,  -1 },    303.80496,   0.00001,     0.450 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   3,  -1 },    160.47482,   0.00001,     0.065 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    332.99258,   0.00010,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    188.19895,   0.00013,     0.666 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    312.38622,   0.00005,     0.068 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    177.24595,   0.00002,     0.036 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    246.72078,   0.00002,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    321.13691,   0.00007,     0.071 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    353.58967,   0.00011,     2.231 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    157.20532,   0.00007,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    246.72078,   0.00002,     0.073 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   2,  -1 },      0.11129,   0.00002,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    180.74996,   0.00004,     3.678 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    174.17457,   0.00028,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,  -2 },    199.01554,   0.00002,     1.652 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    179.45620,   0.00009,     0.038 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },     18.26096,   0.00002,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    198.26096,   0.00002,     0.020 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     96.99257,   0.00010,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    258.34245,   0.00002,     0.111 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },    103.74635,   0.00002,     0.234 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   1 },    275.73663,   0.00003,     0.281 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },     91.35972,   0.00003,     0.042 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    241.05309,   0.00002,     0.093 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    292.63867,   0.00015,     0.400 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    241.05309,   0.00002,     0.063 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },    255.53447,   0.00001,     0.034 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    354.78341,   0.00003,     0.690 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    249.20274,   0.00008,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -2 },    106.67067,   0.00001,     2.535 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     47.87601,   0.00009,     0.037 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,  -3,   0 },    161.53627,   0.00001,     0.016 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.38978,   0.00004,     0.029 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     17.84022,   0.00004,     0.047 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    148.84117,   0.00007,     0.125 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    162.68956,   0.00003,     0.190 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,  -1 },    227.54902,   0.00005,     0.045 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    291.93713,   0.00004,     0.103 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    275.38693,   0.00011,     0.286 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    260.00909,   0.00003,     0.060 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },     88.82208,   0.00001,     0.033 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,  -2,   1 },    247.83106,   0.00001,     0.359 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    272.54025,   0.00005,     0.408 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     34.80233,   0.00001,     0.034 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    206.79385,   0.00001,     0.087 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    341.05634,   0.00001,     0.579 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    332.24674,   0.00003,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    194.39228,   0.00004,     0.035 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },     22.40352,   0.00002,     0.018 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -3,  -1 },    176.01454,   0.00001,     0.079 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   3,  -1,  -1 },    176.01454,   0.00001,     0.072 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },     75.76835,   0.00003,     0.049 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },      5.81806,   0.00004,     0.143 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   3,   0 },    114.49389,   0.00001,     0.051 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.34395,   0.00001,     0.114 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    189.45465,   0.00002,     0.222 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     70.34395,   0.00001,     0.056 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },     91.20893,   0.00001,     0.032 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    226.87621,   0.00004,     0.290 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },     56.25997,   0.00003,     0.095 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },     45.75332,   0.00004,     0.367 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     56.25997,   0.00003,     0.063 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     49.34236,   0.00009,     0.034 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    216.90685,   0.00001,     0.036 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    202.61651,   0.00002,     0.052 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    332.51975,   0.00002,     0.129 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    225.61677,   0.00005,     0.182 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,  -3,   0 },     93.00243,   0.00001,     0.105 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    191.34433,   0.00005,     0.059 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     16.00194,   0.00004,     0.033 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },     20.69904,   0.00001,     0.148 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     23.11011,   0.00006,     0.154 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    202.74043,   0.00001,     0.030 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    235.48616,   0.00001,     0.212 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     20.34617,   0.00003,     0.056 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,   0,   0 },     44.61452,   0.00005,     0.032 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    109.78855,   0.00001,     0.174 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    103.29433,   0.00002,     0.133 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    109.78855,   0.00003,     0.048 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },    108.01556,   0.00002,     0.175 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    283.54469,   0.00002,     0.031 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   1,   0 },    108.01556,   0.00002,     0.022 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   2,   0 },    283.54469,   0.00002,     0.017 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   4,   0,   0 },    108.01556,   0.00004,     0.018 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    213.43590,   0.00003,     0.153 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },     33.43590,   0.00003,     0.050 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,   0,   0 },     52.49807,   0.00003,     0.030 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,   2,   0 },     18.59623,   0.00001,     0.017 },
 { {   0,   4,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    127.58869,   0.00005,     0.029 },
 { {   0,   4,   5,   0,   0,   0,   0,   0,  -2,   1,   0 },    171.49247,   0.00005,    49.129 },
 { {   0,   5, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    239.49283,   0.00003,     0.035 },
 { {   0,   5, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    240.32570,   0.00004,     0.065 },
 { {   0,   5, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },     56.03796,   0.00001,     0.090 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    215.93422,   0.00002,     0.036 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   0,   2 },    341.30282,   0.00001,     1.307 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    342.18853,   0.00005,     0.070 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.50170,   0.00005,     0.062 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    341.57156,   0.00007,     0.348 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    161.57156,   0.00004,     0.096 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.19575,   0.00001,     0.025 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     67.81255,   0.00004,     0.075 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    216.62659,   0.00002,    10.166 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    196.29099,   0.00001,     0.076 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   0,  -3,   0 },    250.19575,   0.00001,     0.024 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.19575,   0.00001,     0.066 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     54.68545,   0.00004,     0.534 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },     70.19575,   0.00001,     0.059 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.19575,   0.00001,     0.104 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    226.25502,   0.00002,     0.039 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     39.51297,   0.00001,     0.081 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    130.52990,   0.00002,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    316.48616,   0.00006,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },    283.79515,   0.00003,     5.513 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   1,   0 },    275.01817,   0.00002,    68.379 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    304.18855,   0.00004,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    111.72163,   0.00001,     1.146 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    304.18855,   0.00004,     0.081 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,  -2,   0 },    263.23769,   0.00001,     0.020 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    120.87102,   0.00006,     0.041 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    222.34288,   0.00002,     0.088 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },     67.71088,   0.00002,     0.526 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    137.37991,   0.00001,     0.066 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    212.57091,   0.00002,     0.846 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    279.49445,   0.00001,     0.038 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    318.17443,   0.00004,     0.076 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    277.45356,   0.00018,     7.836 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    282.46665,   0.00004,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    115.47477,   0.00002,     1.080 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    198.47444,   0.00002,     0.067 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    194.99750,   0.00004,     0.086 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    348.75528,   0.00005,     0.040 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    307.15247,   0.00003,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    109.29271,   0.00005,     0.097 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    189.33649,   0.00001,     0.345 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    140.49138,   0.00003,     0.458 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },     86.92522,   0.00008,     0.082 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    341.88306,   0.00006,     0.887 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     90.38018,   0.00009,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },     62.58660,   0.00001,     0.036 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -3,   1 },    271.14606,   0.00003,     0.078 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },     45.40023,   0.00004,     0.076 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    271.14606,   0.00003,     0.073 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.22541,   0.00001,    13.541 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },     45.40023,   0.00004,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    111.23040,   0.00001,     0.072 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    227.76082,   0.00001,     1.547 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    198.54660,   0.00004,     0.079 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    137.59818,   0.00002,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,  -2,   0 },    261.99828,   0.00002,     0.020 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     99.02006,   0.00002,     0.044 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    260.63308,   0.00008,     0.107 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    207.04874,   0.00001,     0.256 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    311.25820,   0.00001,     0.058 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },    322.97202,   0.00004,     0.042 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },     13.29165,   0.00002,     0.314 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   1,  -1 },    201.88564,   0.00001,     0.096 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   2,  -1 },    303.67119,   0.00001,     0.351 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },     97.59430,   0.00001,     0.041 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    214.43229,   0.00004,     0.090 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    249.37020,   0.00007,     0.470 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    242.27771,   0.00004,     0.065 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   3,   0 },    198.44804,   0.00001,     0.024 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,  -2,  -1 },    160.52613,   0.00001,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },     16.51362,   0.00006,     0.931 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    145.11189,   0.00005,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    101.52469,   0.00001,     0.078 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },     33.85669,   0.00003,     2.828 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },     16.75491,   0.00007,     0.076 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.37464,   0.00016,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -2 },     16.37097,   0.00043,    49.790 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    187.26599,   0.00009,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },     16.75491,   0.00007,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,   0 },    198.44804,   0.00001,     0.025 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    252.50602,   0.00001,     0.029 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     97.40159,   0.00005,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    255.43248,   0.00002,     0.120 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.20773,   0.00001,     0.204 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    301.16653,   0.00002,     0.239 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },    169.59331,   0.00002,     0.043 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    253.18809,   0.00006,     0.099 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    314.57609,   0.00007,     0.320 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    253.05765,   0.00003,     0.061 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    106.22954,   0.00001,     0.034 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    343.40876,   0.00003,     0.482 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    194.44198,   0.00004,     0.068 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     59.98132,   0.00009,     0.036 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.72733,   0.00003,     0.030 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.34737,   0.00003,     0.049 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    145.90588,   0.00004,     0.136 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   3,   0 },    332.24381,   0.00002,     0.052 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },     93.39796,   0.00002,     0.045 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    341.47966,   0.00005,     0.110 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    243.13196,   0.00004,     0.242 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    354.93214,   0.00003,     0.058 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },     93.04228,   0.00001,     0.033 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    275.04686,   0.00005,     0.325 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },     23.80909,   0.00002,     0.092 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    171.43451,   0.00002,     0.425 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    119.69202,   0.00005,     0.064 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    176.00826,   0.00004,     0.035 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    265.02683,   0.00001,     0.018 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    105.48003,   0.00003,     0.051 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    296.68280,   0.00004,     0.157 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    162.53848,   0.00006,     0.195 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    251.38158,   0.00001,     0.032 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    243.24748,   0.00002,     0.245 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     55.74031,   0.00003,     0.060 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     45.75242,   0.00009,     0.033 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    202.06432,   0.00002,     0.054 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },    326.28151,   0.00002,     0.140 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    217.28190,   0.00004,     0.163 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    169.87274,   0.00002,     0.112 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    115.21656,   0.00006,     0.057 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },      7.16433,   0.00002,     0.032 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   2,   0 },    277.37766,   0.00001,     0.017 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },     20.75830,   0.00001,     0.163 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     19.75249,   0.00005,     0.140 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,   2,   0 },    202.81588,   0.00001,     0.030 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,  -2,   0 },     60.25714,   0.00001,     0.187 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },     20.35830,   0.00003,     0.054 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     45.90923,   0.00004,     0.031 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   0,  -2,   0 },    108.11458,   0.00002,     0.054 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    109.78580,   0.00001,     0.195 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    109.78580,   0.00004,     0.047 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    272.95738,   0.00003,     0.030 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    103.54451,   0.00002,     0.017 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   4,  -2,   0 },    288.01426,   0.00002,     0.032 },
 { {   0,   5,   0,   0,   0,   0,   0,   0,   4,   0,   0 },    108.01426,   0.00002,     0.017 },
 { {   0,   5,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     18.33412,   0.00001,     0.110 },
 { {   0,   5,   1,   0,   0,   0,   0,   0,   0,   2,   0 },     18.33412,   0.00001,     0.028 },
 { {   0,   5,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     44.88764,   0.00006,     0.030 },
 { {   0,   5,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    307.82590,   0.00002,     0.120 },
 { {   0,   5,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    127.82590,   0.00002,     0.029 },
 { {   0,   6, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    240.30579,   0.00003,     0.036 },
 { {   0,   6, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    240.30579,   0.00004,     0.068 },
 { {   0,   6, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    240.30579,   0.00004,     0.085 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    215.01208,   0.00002,     0.037 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    341.59613,   0.00003,     2.115 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    352.05811,   0.00002,     0.078 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.36141,   0.00001,     0.065 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    346.34745,   0.00006,     0.445 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -4,   0,   0 },    250.18744,   0.00002,     0.020 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -4,   2,   0 },    250.18744,   0.00002,     0.041 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.24113,   0.00007,     0.025 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.55192,   0.00001,     0.038 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    198.96842,   0.00005,     0.079 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },     70.18744,   0.00003,     0.073 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     50.72055,   0.00003,     0.802 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.18744,   0.00002,     0.098 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    294.11999,   0.00003,     0.040 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     21.14771,   0.00004,     0.085 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    102.93139,   0.00002,     0.068 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    157.51497,   0.00001,     0.037 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.46567,   0.00002,     0.074 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     96.49542,   0.00003,     4.051 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    269.46567,   0.00002,     0.077 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,  -2,   0 },    318.52018,   0.00001,     0.020 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    108.30744,   0.00008,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    202.50162,   0.00003,     0.093 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },      6.30122,   0.00001,     0.396 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    279.53259,   0.00001,     0.039 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },      1.89751,   0.00002,     0.080 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    285.66350,   0.00008,     1.328 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    306.98140,   0.00001,     0.071 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -3,   1 },      7.10310,   0.00001,     0.076 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -2,   1 },    171.60205,   0.00003,    30.407 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },     31.70023,   0.00007,     0.075 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },      4.10716,   0.00003,     3.326 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },      7.63793,   0.00002,     0.039 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    245.41075,   0.00005,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    111.43511,   0.00002,     0.103 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   3,   0 },    292.79977,   0.00001,     0.060 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    177.49243,   0.00002,     0.356 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },     63.72860,   0.00002,     0.040 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    203.97112,   0.00010,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    294.93713,   0.00003,     0.570 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },     27.73581,   0.00005,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    359.48088,   0.00042,    47.898 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     16.69061,   0.00051,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },    109.30345,   0.00002,     2.822 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     50.36638,   0.00002,     0.038 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    329.28293,   0.00003,     0.045 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    277.51667,   0.00009,     0.115 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },    289.83189,   0.00004,     0.043 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    211.24322,   0.00001,     0.263 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    201.88117,   0.00003,     0.102 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   3,  -1 },    201.88117,   0.00001,     0.060 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    148.78642,   0.00001,     0.042 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },     68.14138,   0.00003,     0.095 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     31.63278,   0.00007,     0.363 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },     54.97616,   0.00003,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -2,  -1 },     21.88117,   0.00001,     0.040 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    345.60965,   0.00002,     0.087 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    180.10049,   0.00005,     0.065 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },      5.54990,   0.00002,     0.588 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    173.15578,   0.00005,     0.035 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },    345.60965,   0.00002,     0.067 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },     85.40438,   0.00001,     0.081 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },     10.60053,   0.00005,     1.021 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     15.08477,   0.00010,     0.070 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    164.93437,   0.00002,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   3,  -1,  -1 },    229.52247,   0.00002,     0.076 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,  -2,   0 },    198.43972,   0.00002,     0.039 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   4,   0,   0 },    198.43972,   0.00002,     0.019 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    252.56828,   0.00001,     0.029 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     88.26182,   0.00007,     0.048 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    255.22581,   0.00002,     0.129 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },     62.11659,   0.00001,     0.181 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   1 },    331.20180,   0.00001,     0.208 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    207.27743,   0.00002,     0.044 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    258.40784,   0.00009,     0.105 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    327.80851,   0.00003,     0.266 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    253.69110,   0.00003,     0.059 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    215.23510,   0.00001,     0.033 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    245.74358,   0.00002,     0.370 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    229.85587,   0.00002,     0.505 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     94.37305,   0.00011,     0.066 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     74.04347,   0.00009,     0.035 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },    342.75996,   0.00003,     0.022 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -4,   1,   0 },    341.52905,   0.00001,     0.032 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    342.75996,   0.00003,     0.055 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.75996,   0.00006,     0.030 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.91892,   0.00004,     0.050 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    146.78335,   0.00004,     0.149 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,   1 },    174.74924,   0.00001,     0.172 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    332.00410,   0.00008,     0.118 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    178.77357,   0.00004,     0.210 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    340.84561,   0.00006,     0.056 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    202.10221,   0.00002,     0.032 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -2,  -1 },     97.27222,   0.00001,     0.044 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    176.00731,   0.00001,     0.105 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    356.00731,   0.00002,     0.058 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    326.98458,   0.00004,     0.270 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },    326.89938,   0.00001,     0.033 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    176.00731,   0.00001,     0.059 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    146.07546,   0.00005,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     16.50948,   0.00003,     0.034 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    341.11066,   0.00002,     0.018 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    106.49661,   0.00005,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    296.23561,   0.00005,     0.175 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    203.17572,   0.00005,     0.174 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    261.29030,   0.00001,     0.108 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     65.33369,   0.00004,     0.058 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     17.33361,   0.00004,     0.033 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    190.90507,   0.00005,     0.056 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.44602,   0.00003,     0.154 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    227.02961,   0.00001,     0.148 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    319.77697,   0.00002,     0.050 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    195.58997,   0.00002,     0.030 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    277.72539,   0.00001,     0.121 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    136.74451,   0.00003,     0.055 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },      3.57627,   0.00004,     0.032 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     20.82057,   0.00001,     0.182 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     18.93432,   0.00004,     0.129 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },     20.82057,   0.00001,     0.048 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    319.74101,   0.00001,     0.167 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },     20.38003,   0.00003,     0.052 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     55.75275,   0.00005,     0.031 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   0,  -2,   0 },    108.11067,   0.00002,     0.056 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    108.76119,   0.00003,     0.222 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    106.47422,   0.00001,     0.045 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    327.36545,   0.00001,     0.030 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,   1,   0 },    289.78133,   0.00001,     0.021 },
 { {   0,   6,  -1,   0,   0,   0,   0,   0,   2,   2,   0 },    103.54322,   0.00002,     0.017 },
 { {   0,   7, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    243.22762,   0.00002,     0.037 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    214.52157,   0.00002,     0.038 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.61999,   0.00003,     0.076 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    140.50903,   0.00001,     6.550 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.41597,   0.00001,     0.067 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    345.98409,   0.00005,     0.617 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -4,   0,   0 },    250.17709,   0.00003,     0.020 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.21747,   0.00006,     0.026 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    239.69790,   0.00002,     0.039 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    243.82396,   0.00005,     0.083 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    250.17709,   0.00003,     0.077 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },     63.63303,   0.00002,     0.072 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     41.24613,   0.00002,     1.609 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   1,   0 },     63.63303,   0.00002,     0.079 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    250.17709,   0.00003,     0.092 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    294.01403,   0.00005,     0.041 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    123.28536,   0.00007,     0.090 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    271.82156,   0.00002,     0.065 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    120.76434,   0.00002,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    356.91065,   0.00002,     2.641 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    120.14109,   0.00001,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,  -2,   0 },    251.22528,   0.00002,     0.020 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    113.70379,   0.00008,     0.043 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     41.68021,   0.00001,     0.099 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    210.54627,   0.00001,     0.317 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    340.57936,   0.00002,     0.084 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    343.05074,   0.00004,     0.725 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    340.57936,   0.00002,     0.068 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    177.77218,   0.00004,     3.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    124.52135,   0.00002,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    304.52135,   0.00002,     0.078 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   2,   0,   0 },      8.49673,   0.00002,     0.038 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    246.56466,   0.00004,     0.045 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    110.02782,   0.00001,     0.110 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    299.94504,   0.00001,     0.241 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    292.98190,   0.00001,     0.043 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },    115.56672,   0.00003,     0.291 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },    292.98190,   0.00001,     0.323 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },     62.75233,   0.00002,     0.041 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    289.39993,   0.00003,     0.092 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    118.96407,   0.00008,     0.420 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    289.22361,   0.00002,     0.064 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    104.25194,   0.00001,     0.035 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },     93.25988,   0.00005,     0.755 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    291.34493,   0.00001,     0.079 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    302.72674,   0.00003,     1.654 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     50.41077,   0.00003,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    289.48162,   0.00004,     3.688 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    258.76169,   0.00001,     0.037 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },    289.66826,   0.00002,     0.019 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    335.16836,   0.00004,     0.047 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    279.16556,   0.00006,     0.123 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },    274.74880,   0.00003,     0.045 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    125.46726,   0.00002,     0.226 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   1,  -1 },    201.87483,   0.00002,     0.109 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   2,   1 },     21.87483,   0.00001,     0.032 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   3,  -1 },     21.87483,   0.00001,     0.058 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    149.09965,   0.00001,     0.043 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     22.59576,   0.00007,     0.101 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     40.22952,   0.00001,     0.296 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },     29.22520,   0.00007,     0.060 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    331.85421,   0.00001,     0.430 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    180.09218,   0.00002,     0.034 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    321.21875,   0.00002,     0.086 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },     11.72041,   0.00002,     0.623 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    351.47761,   0.00004,     0.067 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,  -2 },    198.33889,   0.00002,     0.787 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },     11.72041,   0.00001,     0.036 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },     18.23490,   0.00001,     0.024 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   4,   0,   0 },    198.42938,   0.00003,     0.019 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     72.01001,   0.00006,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.30297,   0.00001,     0.141 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    120.97839,   0.00002,     0.163 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   1 },    319.02551,   0.00001,     0.184 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    241.74020,   0.00003,     0.045 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.71187,   0.00006,     0.113 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    140.57312,   0.00002,     0.228 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },     76.38331,   0.00003,     0.301 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    242.07888,   0.00003,     0.384 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     86.02335,   0.00010,     0.063 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },    342.78735,   0.00003,     0.023 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -4,   1,   0 },    341.52271,   0.00001,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },    342.78735,   0.00003,     0.057 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,  -1,   0 },    342.78735,   0.00005,     0.031 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.76990,   0.00005,     0.052 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    150.60818,   0.00004,     0.164 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    121.88496,   0.00001,     0.140 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    347.56452,   0.00009,     0.127 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    144.14796,   0.00008,     0.186 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },      2.18039,   0.00006,     0.054 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    145.16053,   0.00002,     0.031 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,  -1 },    338.44804,   0.00002,     0.231 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    319.18600,   0.00004,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   2,  -1 },    176.00097,   0.00001,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    291.99769,   0.00001,     0.104 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    223.33454,   0.00001,     0.277 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    160.03735,   0.00012,     0.059 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     13.13581,   0.00005,     0.033 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },    161.27242,   0.00001,     0.018 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },     97.01960,   0.00006,     0.054 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    305.73577,   0.00004,     0.196 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    249.33293,   0.00001,     0.047 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },     85.60022,   0.00002,     0.050 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.33293,   0.00001,     0.145 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    143.40048,   0.00003,     0.157 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    113.15165,   0.00001,     0.188 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     12.18558,   0.00003,     0.032 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    191.23788,   0.00006,     0.058 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.19770,   0.00003,     0.170 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },     33.84883,   0.00002,     0.136 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    327.19770,   0.00003,     0.048 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,   2,   0 },     97.27399,   0.00001,     0.030 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },     77.19810,   0.00002,     0.178 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    139.24943,   0.00003,     0.053 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     61.50093,   0.00004,     0.031 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    326.94990,   0.00002,     0.017 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    272.39330,   0.00001,     0.322 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     19.00945,   0.00002,     0.119 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    199.00945,   0.00002,     0.029 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    108.10588,   0.00002,     0.058 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    103.40582,   0.00001,     0.107 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },    103.54044,   0.00002,     0.028 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    289.77499,   0.00001,     0.132 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     66.58643,   0.00002,     0.029 },
 { {   0,   7,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },     68.62918,   0.00001,     0.028 },
 { {   0,   7,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    212.10352,   0.00001,   100.849 },
 { {   0,   7,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    147.99296,   0.00001,     8.136 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     60.43407,   0.00002,     0.037 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    104.43214,   0.00002,     0.074 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     61.76665,   0.00007,     4.509 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   3,   0 },    189.85346,   0.00001,     0.077 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    241.57671,   0.00002,     0.501 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,  -2,   0 },    341.46602,   0.00002,     0.019 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    182.31535,   0.00002,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    319.56680,   0.00005,     0.080 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    146.86951,   0.00002,     1.285 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.46602,   0.00001,     0.070 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    308.89554,   0.00004,     1.004 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    195.66078,   0.00004,     0.095 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    195.66078,   0.00003,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -4,   0,   0 },    250.16468,   0.00003,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,  -1,   0 },    163.75831,   0.00002,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     95.73760,   0.00029,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   2 },     70.16468,   0.00001,     0.474 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },     96.66542,   0.00048,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    220.99430,   0.00002,     0.562 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },     64.71851,   0.00001,     0.067 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    103.77830,   0.00003,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },     84.37067,   0.00055,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     93.28221,   0.00320,   238.919 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },     84.42824,   0.00055,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   0 },     85.02476,   0.00004,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },     92.75224,   0.00002,     0.565 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     96.01633,   0.00049,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   0 },     92.98384,   0.00031,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   1,   0 },     92.75224,   0.00005,     0.026 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    280.59105,   0.00004,     0.042 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     14.74659,   0.00003,     0.095 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    247.33696,   0.00001,     0.360 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    281.10348,   0.00001,     0.062 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    227.46267,   0.00008,     0.996 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    260.01426,   0.00003,    20.170 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,  -3,   0 },     93.14596,   0.00001,     0.071 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    273.14596,   0.00001,     0.080 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.64540,   0.00007,     0.044 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     23.18701,   0.00002,     0.106 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    211.41629,   0.00001,     0.265 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    192.80198,   0.00001,     0.327 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    322.33087,   0.00002,     0.089 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    342.09259,   0.00004,     0.499 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    322.33087,   0.00002,     0.066 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    312.70216,   0.00002,     1.052 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    193.69961,   0.00004,     4.345 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },      6.25724,   0.00006,     0.074 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    111.06907,   0.00002,     0.118 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.16853,   0.00002,     0.209 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.04049,   0.00001,     0.044 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },    121.57093,   0.00002,     0.246 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   2,  -1 },    293.04049,   0.00001,     0.269 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },     36.71067,   0.00001,     0.043 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.81506,   0.00003,     0.098 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    114.65248,   0.00012,     0.333 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    288.81506,   0.00003,     0.062 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   2,   0 },    109.45588,   0.00002,     0.034 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    213.46606,   0.00003,     0.513 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },     88.10709,   0.00002,     0.083 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    100.75266,   0.00004,     0.069 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    280.96565,   0.00004,     0.036 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.69535,   0.00003,     0.048 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    285.45153,   0.00004,     0.134 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    243.60924,   0.00002,     0.053 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    253.17003,   0.00004,     0.046 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,   1 },     96.85549,   0.00002,     0.198 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },    192.78934,   0.00002,     0.212 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    116.02131,   0.00002,     0.044 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },      1.90522,   0.00006,     0.108 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    263.58744,   0.00003,     0.250 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },      9.37421,   0.00006,     0.058 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },      0.08183,   0.00006,     0.034 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    196.11462,   0.00004,     0.448 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.35314,   0.00002,     0.065 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    198.41696,   0.00001,     0.527 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    192.41797,   0.00001,     0.035 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   4,   0,   0 },    198.41696,   0.00002,     0.019 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,  -2,   0 },    253.88068,   0.00003,     0.022 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.53480,   0.00002,     0.051 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    268.11263,   0.00003,     0.154 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.32217,   0.00001,     0.148 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    171.04574,   0.00001,     0.046 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    253.88068,   0.00003,     0.121 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     71.91386,   0.00005,     0.200 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },     84.14181,   0.00003,     0.032 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },     66.28690,   0.00002,     0.253 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },     89.76321,   0.00005,     0.061 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    282.01294,   0.00005,     0.034 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },    342.80959,   0.00003,     0.023 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },    162.80959,   0.00003,     0.059 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,  -2,   0 },    161.36926,   0.00002,     0.022 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,  -1,   0 },    338.14435,   0.00001,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     54.40875,   0.00004,     0.053 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    108.77992,   0.00002,     0.183 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },    185.60874,   0.00002,     0.129 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    330.27522,   0.00003,     0.138 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    125.65557,   0.00014,     0.167 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    353.03858,   0.00001,     0.052 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    155.69549,   0.00002,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,  -1 },    250.98163,   0.00002,     0.202 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    305.38966,   0.00005,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    162.59300,   0.00004,     0.057 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    311.35408,   0.00006,     0.033 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },    161.01892,   0.00001,     0.017 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     88.95861,   0.00004,     0.056 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    294.52857,   0.00003,     0.223 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },     67.75555,   0.00002,     0.051 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    109.36743,   0.00003,     0.143 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     17.73199,   0.00002,     0.032 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    196.04949,   0.00004,     0.060 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    179.32997,   0.00002,     0.102 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },      3.88887,   0.00001,     0.054 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    113.25555,   0.00002,     0.125 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    324.71618,   0.00002,     0.029 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -3,   0 },    191.86153,   0.00001,     0.142 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     86.52229,   0.00002,     0.161 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    167.40108,   0.00001,     0.051 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     39.93882,   0.00004,     0.031 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    358.67618,   0.00001,     0.017 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    200.94065,   0.00001,     0.111 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    199.10777,   0.00002,     0.166 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    204.99398,   0.00003,     0.138 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    199.10777,   0.00002,     0.049 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    204.99398,   0.00003,     0.030 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    108.09987,   0.00002,     0.061 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    337.52115,   0.00001,     0.100 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    283.53598,   0.00002,     0.027 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    294.22464,   0.00001,     0.047 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    288.00070,   0.00002,     0.021 },
 { {   0,   8,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     62.11525,   0.00002,     2.094 },
 { {   0,   8,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     75.12929,   0.00002,     0.028 },
 { {   0,   8,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    132.14115,   0.00001,     0.027 },
 { {   0,   8,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    220.43667,   0.00001,     3.975 },
 { {   0,   9, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    145.51722,   0.00003,     0.072 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.33393,   0.00005,     0.038 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },     62.41707,   0.00004,     2.477 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.66272,   0.00006,     0.084 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    176.51281,   0.00002,     0.712 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    170.24336,   0.00004,     0.073 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    341.66272,   0.00006,     2.701 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   1,   0 },    173.11796,   0.00003,     0.078 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -4,   0,   0 },    250.15019,   0.00002,     0.020 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    250.18131,   0.00003,     0.027 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.71042,   0.00003,     0.041 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    250.18131,   0.00003,     0.092 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    239.70874,   0.00003,     0.039 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.15019,   0.00002,     0.079 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    174.75724,   0.00012,     0.043 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.89511,   0.00001,     0.294 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    335.95109,   0.00002,     0.614 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },     72.89070,   0.00002,     0.075 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    273.76951,   0.00001,     6.930 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     72.89070,   0.00002,     0.076 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    100.78975,   0.00006,     0.045 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    348.07587,   0.00005,     0.113 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    208.98798,   0.00002,     0.227 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    351.05224,   0.00002,     0.042 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },      4.22108,   0.00003,     0.094 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    287.60479,   0.00005,     0.380 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,   1,   0 },      4.22108,   0.00003,     0.063 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    171.40601,   0.00002,     0.635 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },     36.68804,   0.00002,     0.071 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    189.92488,   0.00004,     0.037 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    110.04220,   0.00004,     0.047 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    111.68949,   0.00001,     0.127 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    315.31241,   0.00001,     0.185 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.09505,   0.00001,     0.045 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    251.95531,   0.00001,     0.214 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   2,  -1 },    293.09505,   0.00001,     0.230 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },     72.87826,   0.00001,     0.044 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    289.76367,   0.00002,     0.104 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    120.84325,   0.00009,     0.276 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    289.76367,   0.00002,     0.059 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    100.85526,   0.00002,     0.033 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    212.93145,   0.00002,     0.388 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.65491,   0.00003,     0.066 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    289.76367,   0.00003,     0.035 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.36999,   0.00003,     0.050 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    326.07792,   0.00003,     0.146 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    231.03441,   0.00005,     0.047 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },     96.84901,   0.00001,     0.176 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    154.78560,   0.00002,     0.046 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },      1.50885,   0.00001,     0.116 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    220.68406,   0.00003,     0.216 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   0,   1,   0 },     18.36241,   0.00002,     0.056 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },     17.29788,   0.00002,     0.280 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   1 },      0.06942,   0.00005,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    192.96270,   0.00001,     0.034 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,  -2,   0 },    253.96521,   0.00003,     0.022 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    292.33276,   0.00002,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    253.96521,   0.00006,     0.171 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    176.39039,   0.00001,     0.048 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    256.03440,   0.00005,     0.131 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    119.54630,   0.00005,     0.178 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    252.74293,   0.00001,     0.053 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,  -1 },     53.57695,   0.00002,     0.219 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },     69.78367,   0.00001,     0.259 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    310.81449,   0.00002,     0.033 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     46.48292,   0.00007,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.21145,   0.00005,     0.206 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,   1 },    277.19953,   0.00001,     0.130 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,  -3,   0 },    161.50405,   0.00001,     0.030 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    227.96158,   0.00001,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    146.51029,   0.00009,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    182.01903,   0.00002,     0.050 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,  -2,   1 },    111.37315,   0.00002,     0.169 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,  -1 },    250.97515,   0.00001,     0.179 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,   1 },    315.01185,   0.00002,     0.031 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    269.98340,   0.00001,     0.119 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    172.62174,   0.00001,     0.206 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    136.51792,   0.00003,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    239.35804,   0.00002,     0.032 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.93657,   0.00003,     0.058 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    328.20964,   0.00002,     0.260 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    231.69244,   0.00001,     0.053 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },     99.61314,   0.00006,     0.131 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    334.76404,   0.00002,     0.031 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    196.97812,   0.00005,     0.062 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },      4.36331,   0.00001,     0.056 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    140.85362,   0.00003,     0.116 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    147.41609,   0.00002,     0.028 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    248.58871,   0.00001,     0.146 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     13.93276,   0.00003,     0.050 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     56.36070,   0.00009,     0.030 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    272.81445,   0.00002,     0.540 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    200.99521,   0.00001,     0.104 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    199.16636,   0.00004,     0.047 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    200.85481,   0.00007,     0.029 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    207.18175,   0.00003,     1.174 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    288.09764,   0.00002,     0.063 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    297.32613,   0.00004,     0.094 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    283.52966,   0.00004,     0.045 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   2,   1,   0 },    289.75633,   0.00001,     0.021 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   4,  -2,   0 },    289.75633,   0.00001,     0.030 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    289.75633,   0.00001,     0.017 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     45.90304,   0.00001,     6.758 },
 { {   0,   9,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     30.74131,   0.00002,     0.027 },
 { {   0,   9,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    312.61559,   0.00002,     0.027 },
 { {   0,   9,  -1,   0,   0,   0,   0,   0,   0,   1,  -2 },     40.14738,   0.00001,    24.729 },
 { {   0,  10, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    272.53034,   0.00003,     0.076 },
 { {   0,  10, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    215.93981,   0.00001,    34.189 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.97437,   0.00008,     0.039 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },     63.01766,   0.00004,     0.082 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   3,   0 },     63.01766,   0.00004,     0.070 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     63.00497,   0.00003,     1.343 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.81215,   0.00007,     0.089 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    355.94908,   0.00002,     0.493 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    161.68088,   0.00003,     0.038 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    353.08485,   0.00004,     0.077 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    312.00297,   0.00002,     3.918 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    353.08485,   0.00004,     0.074 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    245.14457,   0.00003,     0.043 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     76.33979,   0.00004,     0.098 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     70.13362,   0.00001,     0.330 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    250.26473,   0.00002,     0.797 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     76.33971,   0.00002,     0.069 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },     70.14825,   0.00004,     0.073 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    250.13362,   0.00001,     1.928 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },     70.14825,   0.00004,     0.079 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    163.21491,   0.00011,     0.044 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     18.34875,   0.00001,     0.248 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     94.37998,   0.00003,     0.073 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     93.73900,   0.00006,     0.047 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    359.86021,   0.00008,     0.122 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    213.01963,   0.00001,     0.199 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },     83.37638,   0.00001,     0.043 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },     44.63443,   0.00001,     0.100 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    304.40790,   0.00002,     0.307 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,   1,   0 },     44.63443,   0.00001,     0.061 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },      8.29596,   0.00001,     0.085 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     17.90175,   0.00004,     0.068 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    190.59041,   0.00003,     0.036 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     66.36161,   0.00005,     0.049 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    111.34951,   0.00001,     0.138 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.14510,   0.00002,     0.046 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    308.86889,   0.00001,     0.045 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    289.80443,   0.00001,     0.111 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   0,   0 },     94.84945,   0.00012,     0.235 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    289.80443,   0.00001,     0.057 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    111.63748,   0.00001,     0.033 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.80443,   0.00001,     0.064 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    289.15564,   0.00001,     0.034 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,  -2,   0 },     18.03720,   0.00001,     0.022 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     49.09938,   0.00001,     0.051 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    305.90237,   0.00004,     0.161 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    277.20648,   0.00001,     0.142 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    230.64435,   0.00004,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,   1 },    108.65325,   0.00002,     0.159 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    198.36714,   0.00003,     0.125 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    217.92201,   0.00009,     0.190 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    198.36715,   0.00001,     0.054 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    240.90307,   0.00001,     0.031 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,  -2,   1 },    180.05492,   0.00002,     0.221 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    162.24125,   0.00002,     0.238 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   1 },      0.05492,   0.00002,     0.032 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    198.38591,   0.00001,     0.287 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.38591,   0.00002,     0.060 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    205.25292,   0.00001,     0.033 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   1,   0 },     18.38591,   0.00002,     0.023 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   2,   0 },    198.36714,   0.00003,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.50276,   0.00003,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    254.04259,   0.00003,     0.191 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    103.83402,   0.00003,     0.125 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },     74.04259,   0.00003,     0.047 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },     96.35059,   0.00002,     0.049 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.43068,   0.00004,     0.143 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    143.54604,   0.00005,     0.160 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },     24.82758,   0.00002,     0.192 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    176.43697,   0.00001,     0.223 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    287.55278,   0.00005,     0.032 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,  -2,   0 },    341.22490,   0.00002,     0.023 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     35.23783,   0.00008,     0.057 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     56.14499,   0.00003,     0.237 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },     94.50989,   0.00002,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },     94.50989,   0.00002,     0.125 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,  -3,   0 },    341.49164,   0.00001,     0.031 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    273.27857,   0.00001,     0.052 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },    267.36418,   0.00001,     0.167 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    137.77659,   0.00004,     0.138 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    178.84423,   0.00002,     0.049 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },     18.64391,   0.00001,     0.030 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    355.96990,   0.00001,     0.030 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    343.45778,   0.00001,     0.182 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     40.91432,   0.00002,     0.053 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    313.51076,   0.00003,     0.031 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.96185,   0.00003,     0.061 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    326.09183,   0.00002,     0.310 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },     95.47558,   0.00007,     0.121 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,  -2,   0 },    189.30543,   0.00001,     0.154 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    278.50426,   0.00001,     0.030 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    197.90516,   0.00005,     0.065 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    349.75954,   0.00002,     0.449 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },      4.78575,   0.00001,     0.058 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    150.92332,   0.00004,     0.108 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    147.66134,   0.00002,     0.173 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    284.21724,   0.00002,     0.134 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },    170.26123,   0.00001,     0.017 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,  -3,   0,   2 },    163.32775,   0.00002,   167.135 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    279.17570,   0.00001,     0.069 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    279.17570,   0.00001,     0.083 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },     17.69900,   0.00002,     0.118 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    201.90741,   0.00003,     0.029 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },     21.04526,   0.00001,     0.016 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    265.00165,   0.00001,     0.074 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    192.77808,   0.00002,     4.415 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    265.00165,   0.00001,     0.077 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    287.98204,   0.00002,     0.041 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    106.34129,   0.00001,     0.028 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    106.34129,   0.00001,     0.016 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   4,  -2,   0 },    109.74392,   0.00001,     0.029 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   4,   0,   0 },    289.74392,   0.00001,     0.016 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    196.19048,   0.00001,     8.810 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    163.59965,   0.00001,     0.042 },
 { {   0,  10,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     65.32354,   0.00001,     0.027 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -2,   0 },     60.66272,   0.00002,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },     60.66272,   0.00017,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     60.99929,   0.00019,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },     60.66272,   0.00005,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },     59.60862,   0.00079,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,  -2 },    240.66272,   0.00003,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     60.66272,   0.01443,  1911.695 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   2 },    240.66272,   0.00003,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },     60.17999,   0.00065,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },     60.66272,   0.00005,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },     65.13700,   0.00017,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     60.66272,   0.00017,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     30.27657,   0.00002,     0.026 },
 { {   0,  10,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0 },    317.72579,   0.00001,     0.042 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     63.53349,   0.00008,     0.040 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    104.01711,   0.00001,     0.075 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    250.83791,   0.00002,     8.386 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    104.01711,   0.00001,     0.076 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    340.98594,   0.00001,     0.042 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.69670,   0.00005,     0.094 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    163.39481,   0.00001,     0.039 },
 { {   0,  11, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    346.71095,   0.00003,     1.135 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },     70.11468,   0.00004,     0.020 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },     70.11500,   0.00002,     0.028 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     82.88005,   0.00001,     0.044 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    250.11370,   0.00003,     0.088 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    254.19751,   0.00002,     0.532 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    114.32049,   0.00004,     0.075 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    146.75647,   0.00008,     0.046 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     40.72515,   0.00002,     0.215 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    159.11193,   0.00003,     0.048 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },     21.99587,   0.00007,     0.132 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    259.67924,   0.00004,     0.177 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   3,   0 },      1.91340,   0.00001,     0.053 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    182.88332,   0.00002,     0.258 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },     11.31410,   0.00002,     0.475 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     11.60538,   0.00003,     0.065 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     17.55743,   0.00005,     0.050 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    309.74978,   0.00003,     0.150 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.19047,   0.00002,     0.048 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    288.75921,   0.00001,     0.046 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     88.97096,   0.00005,     0.205 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },     91.45688,   0.00003,     0.261 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    278.45879,   0.00002,     0.322 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.84073,   0.00002,     0.061 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.75298,   0.00002,     0.053 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    349.42450,   0.00002,     0.179 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,  -2,   1 },     21.82927,   0.00002,     0.051 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    230.29914,   0.00004,     0.050 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,   1 },     38.21934,   0.00004,     0.144 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },    201.82927,   0.00001,     0.150 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    131.56755,   0.00002,     0.048 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    170.84455,   0.00002,     0.170 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    198.36769,   0.00002,     0.052 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   2,   0 },     24.03746,   0.00002,     0.031 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    118.33011,   0.00002,     0.207 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -3,   0 },     18.36443,   0.00001,     0.109 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.36682,   0.00005,     0.058 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    202.35226,   0.00003,     0.033 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   1,   0 },     18.36728,   0.00001,     0.023 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    211.34584,   0.00002,     0.056 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.86218,   0.00001,     0.116 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },     68.41457,   0.00003,     0.051 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    257.74869,   0.00004,     0.157 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     58.43805,   0.00005,     0.145 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    265.84699,   0.00006,     0.032 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     33.40884,   0.00010,     0.059 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     77.65300,   0.00002,     0.278 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    151.78067,   0.00003,     0.056 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   1 },    329.07506,   0.00002,     0.112 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   2,  -1 },     94.27845,   0.00002,     0.116 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.02716,   0.00001,     0.186 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    153.04856,   0.00001,     0.127 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,  -2,   1 },    175.95541,   0.00002,     0.140 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,  -1 },    342.14062,   0.00002,     0.146 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    161.22266,   0.00001,     0.140 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },    125.36200,   0.00001,     0.164 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    341.04451,   0.00004,     0.052 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    307.99238,   0.00004,     0.031 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     93.68574,   0.00004,     0.063 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    335.16625,   0.00002,     0.384 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.97214,   0.00001,     0.057 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     91.24895,   0.00007,     0.113 },
 { {   0,  11,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },     57.65859,   0.00002,     0.141 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    209.43330,   0.00003,     0.067 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },      5.16191,   0.00001,     0.060 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    127.52199,   0.00001,     0.101 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    338.92857,   0.00002,     0.027 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    357.95513,   0.00001,     0.194 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    330.33781,   0.00002,     0.047 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    223.75958,   0.00002,     0.029 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    148.91896,   0.00003,     0.072 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     28.95443,   0.00001,     0.079 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    201.09063,   0.00001,     0.092 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },    201.09063,   0.00001,     0.027 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     21.09063,   0.00001,     0.028 },
 { {   0,  11,  -7,   0,   0,   0,   0,   0,   2,   2,   0 },    201.09063,   0.00001,     0.016 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },     95.18049,   0.00001,     0.078 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },     37.42702,   0.00002,     2.506 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    275.18049,   0.00001,     0.073 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    286.31702,   0.00001,     0.084 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },    286.31702,   0.00001,     0.026 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    287.96963,   0.00002,     0.043 },
 { {   0,  11,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    109.80192,   0.00001,     2.667 },
 { {   0,  11,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    244.63302,   0.00002,     0.027 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.02145,   0.00009,     0.041 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    243.88986,   0.00002,     0.091 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    244.05076,   0.00003,     1.975 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,  -2,   0 },    161.62052,   0.00001,     0.020 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.62052,   0.00001,     0.043 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.71015,   0.00012,     0.100 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    341.71015,   0.00003,     0.085 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    159.31109,   0.00002,     0.664 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    341.71015,   0.00003,     0.068 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    341.62052,   0.00003,     0.077 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },     70.09434,   0.00001,     0.028 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.83969,   0.00002,     0.045 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.96973,   0.00002,     0.111 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    250.07034,   0.00002,     0.093 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },     70.09434,   0.00003,     0.071 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     11.82009,   0.00003,     0.047 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    358.87362,   0.00002,     0.189 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    196.77230,   0.00008,     0.049 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },      8.97653,   0.00005,     0.143 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    200.60017,   0.00002,     0.159 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    294.72255,   0.00001,     0.222 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     12.14215,   0.00001,     0.034 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    355.59595,   0.00006,     0.052 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     14.50914,   0.00002,     0.137 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.23122,   0.00002,     0.049 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    147.19361,   0.00001,     0.048 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     97.33680,   0.00006,     0.182 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },     91.49318,   0.00002,     0.225 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    289.71400,   0.00004,     0.059 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    289.97061,   0.00004,     0.033 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     66.48922,   0.00002,     0.055 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    292.93299,   0.00001,     0.201 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    214.67014,   0.00003,     0.052 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,   1 },     19.21202,   0.00004,     0.132 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   2,  -1 },     93.68767,   0.00001,     0.138 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    133.17763,   0.00002,     0.050 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    358.73017,   0.00014,     0.154 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   1,   0 },     18.34662,   0.00001,     0.051 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   2,   0 },      1.96284,   0.00001,     0.030 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    118.13575,   0.00002,     0.183 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    192.59740,   0.00002,     0.211 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.36437,   0.00002,     0.056 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    197.44520,   0.00004,     0.032 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    225.06930,   0.00002,     0.058 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },     91.84249,   0.00001,     0.108 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    258.08999,   0.00004,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    340.67614,   0.00003,     0.133 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },     72.67638,   0.00001,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    252.53313,   0.00004,     0.031 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -4,   0,   0 },    341.46058,   0.00001,     0.024 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -4,   2,   0 },    287.70221,   0.00001,     0.069 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    341.46058,   0.00001,     0.034 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     45.53295,   0.00007,     0.062 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },     52.00779,   0.00002,     0.336 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },     27.26115,   0.00004,     0.097 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   3,   0 },    338.04125,   0.00001,     0.042 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    201.24076,   0.00002,     0.058 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    306.70028,   0.00002,     0.211 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    131.46850,   0.00001,     0.118 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,  -2,  -1 },    175.93884,   0.00002,     0.052 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,  -1 },    238.97756,   0.00001,     0.134 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,   1 },    175.93884,   0.00003,     0.029 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    342.40333,   0.00003,     0.030 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     97.72013,   0.00005,     0.066 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    343.40580,   0.00002,     0.506 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    279.08705,   0.00003,     0.059 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    153.41561,   0.00002,     0.105 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    221.18607,   0.00002,     0.029 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,  -3,   1,   1 },     23.21869,   0.00001,    12.771 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    304.26211,   0.00002,     1.025 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },      5.49702,   0.00001,     0.364 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    147.08508,   0.00001,     0.221 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },    312.53669,   0.00001,     0.115 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    342.83562,   0.00003,     0.045 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    306.31144,   0.00003,     0.028 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   1,   0 },    321.11523,   0.00001,     0.040 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    356.08875,   0.00002,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },      2.93100,   0.00010,     0.076 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    321.11523,   0.00122,    40.749 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    294.96594,   0.00007,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    232.88272,   0.00003,     0.087 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    321.11523,   0.00001,     0.040 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    194.41448,   0.00001,     0.976 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    305.69001,   0.00002,     7.271 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,  -1,   1,  -1 },    304.55423,   0.00004,    34.218 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.58277,   0.00001,     0.042 },
 { {   0,  12,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },    109.71286,   0.00001,     0.020 },
 { {   0,  12,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    186.77593,   0.00001,     0.077 },
 { {   0,  12,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    186.77593,   0.00001,     0.074 },
 { {   0,  12,  -6,   0,   0,   0,   0,   0,   0,   1,  -2 },    224.75792,   0.00002,    12.050 },
 { {   0,  12,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },     13.46275,   0.00001,     0.085 },
 { {   0,  12,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    273.02862,   0.00002,     0.026 },
 { {   0,  13, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.85485,   0.00002,     0.076 },
 { {   0,  13, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    328.09594,   0.00003,     7.587 },
 { {   0,  13, -21,   0,   0,   0,   0,   0,   0,   1,   0 },    327.85485,   0.00002,     0.075 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.50109,   0.00008,     0.042 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,   0,   0,   0 },     64.38468,   0.00001,     0.884 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.72125,   0.00008,     0.107 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },    161.64860,   0.00002,     0.041 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.64860,   0.00002,     0.090 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    161.66556,   0.00004,     0.469 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,   0,   1,   0 },    161.64860,   0.00002,     0.065 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,  -1,   0 },     70.07168,   0.00001,     0.029 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     66.02483,   0.00001,     0.204 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    250.03967,   0.00005,     0.099 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    250.04560,   0.00004,     0.061 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.37184,   0.00007,     0.049 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },     11.69996,   0.00001,     0.169 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    190.32632,   0.00006,     0.051 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },     10.54617,   0.00004,     0.157 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },     32.90866,   0.00003,     0.047 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     32.90866,   0.00003,     0.195 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    192.63697,   0.00002,     0.033 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.49940,   0.00006,     0.054 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },      5.17677,   0.00001,     0.126 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.26753,   0.00001,     0.051 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -1,   0,   1 },    113.26753,   0.00001,     0.139 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -1,   2,  -1 },    113.26753,   0.00001,     0.146 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     66.52676,   0.00004,     0.163 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.90088,   0.00001,     0.057 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    289.80681,   0.00002,     0.032 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    124.24864,   0.00005,     0.057 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    293.44539,   0.00001,     0.230 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.82096,   0.00003,     0.112 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.79408,   0.00003,     0.122 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   2,  -1 },     49.71336,   0.00002,     0.127 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    185.53145,   0.00002,     0.052 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    212.98062,   0.00004,     0.163 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    355.67767,   0.00007,     0.140 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.34417,   0.00004,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    190.91276,   0.00003,     0.031 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    224.83018,   0.00002,     0.060 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    268.34549,   0.00003,     0.299 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    256.62347,   0.00006,     0.195 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    313.43477,   0.00003,     0.123 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    267.39367,   0.00001,     0.141 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },     72.50284,   0.00001,     0.157 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    252.50284,   0.00001,     0.030 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -4,   0,   0 },    341.44196,   0.00001,     0.025 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -4,   2,   0 },    161.44196,   0.00001,     0.072 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    162.85225,   0.00003,     0.035 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     58.21126,   0.00004,     0.064 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    126.56403,   0.00002,     0.426 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    200.61369,   0.00001,     0.092 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    155.16346,   0.00003,     0.060 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },    148.07783,   0.00002,     0.294 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   2,  -1 },    273.90482,   0.00002,     0.102 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   0,   0,   0 },      3.79918,   0.00003,     0.110 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,  -1 },    223.88203,   0.00003,     0.124 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    343.00418,   0.00004,     0.030 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.92925,   0.00005,     0.068 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    295.35037,   0.00001,     0.061 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    294.51651,   0.00001,     0.320 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    295.35037,   0.00001,     0.099 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    250.11930,   0.00005,     0.029 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -2,   2 },    187.34580,   0.00003,     0.075 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    288.21859,   0.00004,     0.073 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   2 },    187.34580,   0.00003,     0.076 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    181.90381,   0.00001,     0.077 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    216.84738,   0.00002,     0.065 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     36.84738,   0.00002,     0.471 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   0,   0,   0 },     36.84738,   0.00002,     0.090 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    238.02567,   0.00002,     0.028 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    197.03743,   0.00002,     1.538 },
 { {   0,  13,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },     21.16769,   0.00001,     0.042 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     50.04307,   0.00004,     8.068 },
 { {   0,  13,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    283.48411,   0.00002,     0.026 },
 { {   0,  13,  -6,   0,   0,   0,   0,   0,   2,  -3,   0 },    134.04802,   0.00001,     9.801 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     64.94371,   0.00008,     0.044 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -3,   0 },    277.02633,   0.00001,     0.075 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -2,   0 },    265.83931,   0.00002,    59.908 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,   2,  -1,   0 },     97.02633,   0.00001,     0.076 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.73668,   0.00001,     0.045 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    317.73668,   0.00001,     0.221 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,  -2,   3,   0 },    161.67292,   0.00002,     0.056 },
 { {   0,  14, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    161.67292,   0.00002,     0.363 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,  -2,   0 },    250.00977,   0.00002,     0.021 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.95309,   0.00005,     0.129 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   3,   0 },    250.00977,   0.00004,     0.053 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.04704,   0.00001,     0.105 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,   0,   0,   0 },     69.85407,   0.00003,     0.266 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    351.21766,   0.00007,     0.050 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    312.30919,   0.00002,     0.153 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    187.77296,   0.00007,     0.053 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },     11.14021,   0.00004,     0.175 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    201.32164,   0.00001,     0.048 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     23.49714,   0.00005,     0.174 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    193.09339,   0.00003,     0.033 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    357.91093,   0.00005,     0.056 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },      6.68067,   0.00001,     0.117 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.29960,   0.00001,     0.053 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },    113.29960,   0.00001,     0.134 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    122.13412,   0.00008,     0.148 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    289.92521,   0.00003,     0.055 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    289.85134,   0.00002,     0.032 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    113.79273,   0.00004,     0.059 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    169.87578,   0.00001,     0.105 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },     21.77342,   0.00003,     0.055 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.77342,   0.00003,     0.113 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    221.94686,   0.00002,     0.182 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   0,   0,   0 },     11.05208,   0.00004,     0.129 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    359.97641,   0.00003,     0.149 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    198.36905,   0.00003,     0.052 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    198.47119,   0.00002,     0.031 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,  -2,   0 },     72.94748,   0.00001,     0.024 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    243.92250,   0.00003,     0.063 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    115.21989,   0.00001,     0.367 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    109.31182,   0.00001,     0.095 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    252.32206,   0.00005,     0.222 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,   0,   0,   0 },     74.67376,   0.00001,     0.114 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -4,   2,   0 },    157.99408,   0.00002,     0.076 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -4,   3,   0 },     66.05714,   0.00003,    20.160 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -3,   1,   1 },    201.20556,   0.00003,     1.491 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     67.53060,   0.00004,     0.067 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },      2.94034,   0.00003,     0.580 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },     75.21997,   0.00001,     0.087 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,  -1 },    174.46098,   0.00002,     0.062 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,   1 },    201.20557,   0.00002,     0.092 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },    147.64116,   0.00003,     0.360 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   2,  -1 },    275.19387,   0.00001,     0.095 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    325.83623,   0.00002,     0.286 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    255.22814,   0.00002,     0.102 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    161.42130,   0.00001,     0.043 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    124.07940,   0.00001,     0.028 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,  -1 },    215.30444,   0.00002,     0.115 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    301.65522,   0.00003,     0.029 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -3,   3,  -1 },    233.50354,   0.00004,   240.318 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    147.48567,   0.00002,     0.072 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     49.41572,   0.00005,     1.380 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    249.24390,   0.00001,     0.080 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },    294.96166,   0.00002,     0.063 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    294.96166,   0.00001,     0.400 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    294.96166,   0.00002,     0.093 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },     70.21118,   0.00002,     0.111 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    250.21118,   0.00002,     0.028 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    155.84219,   0.00009,     0.077 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },      9.66514,   0.00003,     3.629 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -1,  -1,   1 },     70.65590,   0.00001,     2.249 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -1,   1,  -1 },    285.16233,   0.00001,     1.285 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     48.64572,   0.00004,     0.085 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   0,   2,   0 },    131.99485,   0.00001,     0.026 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    298.35569,   0.00001,     0.100 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    331.54163,   0.00003,     0.027 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     89.19384,   0.00001,     0.083 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    269.19384,   0.00001,     0.069 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,  -1,   1,  -1 },    252.17204,   0.00002,     4.508 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    223.54555,   0.00003,     2.006 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,  -2,   0 },     17.75488,   0.00001,     0.091 },
 { {   0,  14, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    199.57841,   0.00002,     0.027 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    287.91995,   0.00002,     0.073 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    287.91995,   0.00002,     0.025 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.67358,   0.00001,     0.040 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    107.91994,   0.00002,     0.026 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   2,   2,   0 },    287.91994,   0.00002,     0.015 },
 { {   0,  15, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     65.35110,   0.00004,     0.045 },
 { {   0,  15, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    245.39172,   0.00002,     0.241 },
 { {   0,  15, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    342.07339,   0.00001,     0.047 },
 { {   0,  15, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    342.07339,   0.00001,     0.194 },
 { {   0,  15, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    161.69373,   0.00002,     0.296 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,  -2,   0 },    250.02046,   0.00002,     0.021 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    250.02046,   0.00002,     0.049 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.92201,   0.00007,     0.141 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.02046,   0.00003,     0.113 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   0,   0,   0 },     69.80516,   0.00002,     0.228 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,   0,   1,   0 },     70.02046,   0.00002,     0.057 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    354.24610,   0.00006,     0.052 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    148.11354,   0.00001,     0.140 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    212.15530,   0.00008,     0.054 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },      2.04363,   0.00001,     0.196 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    223.96242,   0.00001,     0.050 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     43.96242,   0.00001,     0.157 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   2,  -2,   0 },     13.51484,   0.00001,     0.217 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    193.51484,   0.00001,     0.032 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    347.09984,   0.00005,     0.058 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    315.65236,   0.00001,     0.109 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.32768,   0.00001,     0.054 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },    113.32768,   0.00001,     0.123 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    107.86663,   0.00007,     0.135 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    289.94601,   0.00002,     0.053 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    289.89063,   0.00002,     0.031 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    134.57026,   0.00002,     0.061 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },     21.75075,   0.00001,     0.057 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },     28.24154,   0.00003,     0.106 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    141.15618,   0.00001,     0.055 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     14.05338,   0.00006,     0.119 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    359.95177,   0.00002,     0.136 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    198.33374,   0.00004,     0.050 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    198.12170,   0.00004,     0.030 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -4,   1,   2 },     83.37817,   0.00006,   115.724 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,  -2,   0 },    252.97555,   0.00001,     0.024 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    251.97366,   0.00008,     0.065 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    257.25714,   0.00004,     0.476 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.09783,   0.00001,     0.090 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    254.33225,   0.00002,     0.258 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -3,   1,   1 },    280.97579,   0.00002,    22.161 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },      7.28245,   0.00003,     0.070 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    111.58613,   0.00003,     0.910 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     14.15222,   0.00002,     0.082 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },     93.61347,   0.00002,     0.065 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   2,  -1 },     93.61347,   0.00002,     0.090 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    299.25995,   0.00001,     0.062 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,  -1 },    175.87689,   0.00003,     0.107 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    277.53855,   0.00002,     0.028 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     86.72622,   0.00008,     0.075 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    108.32836,   0.00052,    10.112 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    131.53440,   0.00003,     0.076 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },     91.23261,   0.00003,     0.868 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },     84.60142,   0.00002,     0.534 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    225.18267,   0.00003,     1.110 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    355.14718,   0.00005,     0.075 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    276.35362,   0.00004,    70.915 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   1,  -1 },     23.97247,   0.00002,     6.551 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    162.96784,   0.00004,     1.146 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    155.00624,   0.00001,     0.094 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    299.85502,   0.00003,     0.027 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    148.98844,   0.00001,     0.076 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    313.50814,   0.00005,     7.868 },
 { {   0,  15, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     64.69365,   0.00002,     0.075 },
 { {   0,  15, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.07433,   0.00002,     0.887 },
 { {   0,  15, -10,   0,   0,   0,   0,   0,   1,  -1,  -1 },    346.61309,   0.00002,    13.638 },
 { {   0,  15,  -9,   0,   0,   0,   0,   0,   2,  -2,  -2 },     62.72951,   0.00001,     0.076 },
 { {   0,  15,  -9,   0,   0,   0,   0,   0,   2,   0,  -2 },    242.72951,   0.00001,     0.075 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.97569,   0.00002,     0.048 },
 { {   0,  16, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    161.71124,   0.00003,     0.249 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.99199,   0.00003,     0.051 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.99199,   0.00001,     0.121 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    249.94431,   0.00009,     0.200 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     28.69950,   0.00005,     0.053 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },     86.22844,   0.00001,     0.128 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    190.79817,   0.00006,     0.056 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     22.27924,   0.00004,     0.224 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.01405,   0.00005,     0.060 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    313.48045,   0.00001,     0.288 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     83.67000,   0.00004,     0.125 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    109.89610,   0.00002,     0.031 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     87.32649,   0.00003,     0.064 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    175.58567,   0.00003,     0.093 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },     21.72611,   0.00001,     0.060 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.72611,   0.00004,     0.099 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },     25.13442,   0.00001,     0.057 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },      9.11105,   0.00001,     0.235 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    347.02371,   0.00005,     0.111 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,  -3,   0 },    198.24427,   0.00002,     0.166 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    198.17878,   0.00004,     0.030 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.01172,   0.00010,     0.068 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    178.90284,   0.00003,     0.679 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    294.47792,   0.00002,     0.085 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },    112.73917,   0.00002,     0.063 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },    292.73917,   0.00002,     0.093 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    254.37348,   0.00002,     0.308 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -3,   1,   1 },    201.16224,   0.00002,     1.723 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,  -1,   2 },    285.47789,   0.00003,     7.151 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    131.67568,   0.00002,     0.073 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   2 },    157.79594,   0.00002,     0.076 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },     94.90626,   0.00002,     2.113 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    311.67568,   0.00002,     0.078 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    119.58745,   0.00001,     0.445 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    355.50099,   0.00005,     0.091 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    175.85226,   0.00004,     0.100 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    166.83604,   0.00005,     0.028 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    109.05479,   0.00005,     0.079 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },     79.73715,   0.00002,     1.899 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    245.50365,   0.00001,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,  -1,   1 },     61.49610,   0.00003,     5.165 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },     91.20994,   0.00006,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    107.64813,   0.00001,     2.775 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,   1 },    271.20994,   0.00005,     0.077 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   1,  -1 },     75.17412,   0.00002,     1.897 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,  -2,   0 },    104.96731,   0.00002,     0.069 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    284.96731,   0.00002,     0.083 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,  -1,   1,  -1 },    279.80927,   0.00002,     2.115 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    312.45183,   0.00003,     4.042 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    167.77736,   0.00010,     0.077 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    335.41259,   0.00003,     0.089 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    148.79293,   0.00002,     0.026 },
 { {   0,  16, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    103.42826,   0.00002,     0.076 },
 { {   0,  16, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    105.46525,   0.00001,     0.025 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.69462,   0.00005,     0.050 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.76656,   0.00001,     0.146 },
 { {   0,  17, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    161.72565,   0.00002,     0.216 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.36729,   0.00002,     0.052 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },     70.22350,   0.00002,     0.171 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },     76.14107,   0.00002,     0.135 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.96167,   0.00001,     0.131 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    249.90508,   0.00008,     0.178 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     29.88169,   0.00005,     0.055 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },     86.54407,   0.00001,     0.119 },
 { {   0,  17, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    187.84348,   0.00006,     0.058 },
 { {   0,  17, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },      9.70751,   0.00003,     0.260 },
 { {   0,  17, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    353.59632,   0.00001,     0.106 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    336.38158,   0.00004,     0.062 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     39.21386,   0.00002,     0.351 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    108.91558,   0.00002,     0.096 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.67381,   0.00001,     0.056 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    108.58938,   0.00003,     0.116 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    109.94692,   0.00003,     0.030 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     17.37957,   0.00003,     0.066 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    326.06513,   0.00004,     0.541 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.69954,   0.00002,     0.094 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },      9.30195,   0.00001,     0.276 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    351.74890,   0.00004,     0.104 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,  -1,   0 },    198.21395,   0.00002,     0.047 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   2,   0,   0 },    198.29815,   0.00002,     0.029 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.58665,   0.00005,     0.071 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    133.06843,   0.00004,     1.179 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    112.76350,   0.00002,     0.066 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },    112.76350,   0.00002,     0.088 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    254.41001,   0.00003,     0.381 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     94.40739,   0.00001,     0.094 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    162.81868,   0.00002,     0.025 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    342.81868,   0.00002,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },    236.24807,   0.00004,     6.572 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   2,  -1 },     21.30980,   0.00001,    68.493 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   3,   0 },    161.34742,   0.00001,     0.038 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -2,   1 },    326.51983,   0.00003,     0.072 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -1,   1 },    175.78198,   0.00001,     1.760 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },    326.51983,   0.00003,     0.079 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   1,  -1 },     51.75136,   0.00002,     1.109 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },     91.03399,   0.00002,     0.617 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     58.19037,   0.00003,     0.086 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   1,   0,  -1 },    175.82568,   0.00003,     0.095 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },    340.81591,   0.00001,     0.101 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    160.81591,   0.00001,     0.027 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    216.61300,   0.00001,     0.083 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,  -1 },    237.58449,   0.00003,     0.076 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    189.28943,   0.00001,     3.772 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,  -1,   1,  -1 },    157.91220,   0.00003,    10.164 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },    283.05576,   0.00001,     0.072 },
 { {   0,  17, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    283.05576,   0.00001,     0.079 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,  -1,   0 },    269.64871,   0.00002,     2.645 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   0,   1,  -2 },    315.29266,   0.00002,    22.409 },
 { {   0,  17, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    148.40343,   0.00002,     0.084 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    109.59970,   0.00001,     0.064 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    109.59970,   0.00001,     0.024 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   2,  -3,   0 },    289.59970,   0.00001,     1.655 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    283.40560,   0.00002,     0.072 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   2,  -1,  -2 },    283.40560,   0.00002,     3.695 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   2,  -1,   0 },    289.59970,   0.00001,     0.037 },
 { {   0,  17, -12,   0,   0,   0,   0,   0,   2,   2,   0 },    109.59970,   0.00001,     0.015 },
 { {   0,  18, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    246.34666,   0.00002,     0.049 },
 { {   0,  18, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    246.34666,   0.00002,     0.166 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.73558,   0.00005,     0.051 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.75059,   0.00001,     0.161 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.73715,   0.00001,     0.047 },
 { {   0,  18, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    161.73715,   0.00001,     0.190 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.88158,   0.00005,     0.054 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.69417,   0.00004,     0.191 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   3,   0 },    249.92953,   0.00001,     0.047 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.92953,   0.00001,     0.143 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    249.85589,   0.00006,     0.160 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },     41.99936,   0.00003,     0.057 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    124.33912,   0.00001,     0.237 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    124.73778,   0.00002,     0.111 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    190.72916,   0.00006,     0.061 },
 { {   0,  18, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },     10.68111,   0.00002,     0.310 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    338.38934,   0.00004,     0.065 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    330.18205,   0.00002,     0.450 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },    109.04775,   0.00002,     0.091 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.68822,   0.00001,     0.058 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   0,   0,   0 },     98.68822,   0.00001,     0.108 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,   2,   0,   0 },    289.98944,   0.00002,     0.029 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   0,   1 },     21.67107,   0.00001,     0.074 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   0,   2 },    206.91373,   0.00017,   556.277 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   1,   1 },     97.56858,   0.00002,     5.933 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -3,   2,   1 },    201.67107,   0.00001,     0.076 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    359.50207,   0.00003,     0.069 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    310.81194,   0.00003,     0.818 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,  -1,   1 },     21.67107,   0.00001,     0.515 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.67107,   0.00003,     0.088 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,  -2,   0 },    210.11530,   0.00001,     0.062 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   0,   0,   0 },      6.33690,   0.00004,     0.097 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,   2,   0,   0 },    198.27630,   0.00003,     0.029 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,  -2,   2 },    253.03819,   0.00001,     0.076 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    142.66207,   0.00004,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   2 },    253.03819,   0.00001,     0.075 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    244.71313,   0.00002,     4.498 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     97.78948,   0.00002,     0.077 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -1,   1,  -1 },    240.75267,   0.00002,     0.784 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    252.23333,   0.00002,     0.066 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },     63.52135,   0.00003,     0.501 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    289.16049,   0.00001,     0.089 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },     37.50243,   0.00003,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    231.66485,   0.00005,     0.080 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    132.53322,   0.00002,     1.286 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   1 },      2.41180,   0.00001,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   0 },    217.50243,   0.00001,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -1,   1 },     16.93926,   0.00005,    17.413 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    122.40657,   0.00001,     0.074 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },    282.85972,   0.00003,     9.147 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   1 },    166.01974,   0.00004,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   1,  -1 },    178.81262,   0.00005,     3.622 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },    122.40657,   0.00001,     0.077 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },     37.40480,   0.00007,     0.070 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    146.93678,   0.00002,     1.004 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    202.44737,   0.00007,     0.082 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   1,   0 },    217.50243,   0.00001,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,  -1,  -1 },    175.79721,   0.00001,     0.488 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   0,  -1 },    175.79721,   0.00003,     0.089 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    217.50243,   0.00013,     0.095 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    217.50243,   0.00020,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    217.50243,   0.00003,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   0 },    114.56550,   0.00003,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   0 },    114.56550,   0.00007,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },    114.56550,   0.00004,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   0 },    114.56550,   0.00004,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    114.56550,   0.00035,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,  -2 },    294.56550,   0.00001,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    114.55114,   0.00293,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -2 },    294.56550,   0.00002,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.55967,   0.00298,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },     65.48676,   0.00002,     0.563 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.90658,   0.00001,     0.067 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -2,   0 },    294.56550,   0.00001,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },    294.56550,   0.00008,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,   0 },    114.56550,   0.00007,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   0 },    114.55394,   0.00096,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -2 },    294.56550,   0.00008,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    114.55388,   0.01410,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   2 },    294.56550,   0.00002,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -2 },    294.56550,   0.00051,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    114.56550,   0.25425,   273.043 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   2 },    294.56550,   0.00051,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -2 },    294.56550,   0.00002,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    114.79544,   0.01326,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   2 },    294.56550,   0.00008,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    114.79537,   0.00090,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   0 },    114.56550,   0.00007,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   0 },    294.56550,   0.00008,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,   0,   0 },    294.56550,   0.00001,     0.039 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },    114.57740,   0.00280,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   2 },    294.56550,   0.00002,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    114.57998,   0.00291,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   2 },    294.56550,   0.00001,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    114.56550,   0.00035,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   0 },    114.56550,   0.00004,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   0 },    114.56550,   0.00004,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   0 },    114.56550,   0.00007,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },    114.56550,   0.00003,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   0 },     11.62857,   0.00003,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },     11.62857,   0.00020,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     12.54456,   0.00010,     0.095 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    188.41031,   0.00003,     0.062 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -3,   0 },     11.62857,   0.00001,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },     11.62857,   0.00009,     0.082 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    220.17459,   0.00001,     0.996 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    191.62857,   0.00007,     0.070 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,  -1 },     22.72695,   0.00014,    19.959 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,  -1,   0 },     11.62857,   0.00001,     0.075 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },    134.23070,   0.00004,     0.076 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -2,   0 },    191.76310,   0.00002,     0.080 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    191.62857,   0.00003,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,   0 },    155.90294,   0.00004,     0.026 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    199.49867,   0.00002,     0.074 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,  -1,  -2 },     20.71545,   0.00001,     9.628 },
 { {   0,  18, -14,   0,   0,   0,   0,   0,   2,   0,   0 },     19.49867,   0.00002,     0.025 },
 { {   0,  18, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    107.82541,   0.00002,     0.024 },
 { {   0,  19, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    246.65655,   0.00003,     0.050 },
 { {   0,  19, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.74206,   0.00005,     0.053 },
 { {   0,  19, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.73870,   0.00001,     0.179 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },     76.06814,   0.00003,     0.056 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.68092,   0.00004,     0.218 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },     69.81455,   0.00003,     0.051 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    249.81455,   0.00003,     0.145 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.09746,   0.00004,     0.059 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    124.34789,   0.00001,     0.278 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    161.56852,   0.00002,     0.104 },
 { {   0,  19, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    201.00228,   0.00005,     0.063 },
 { {   0,  19, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    344.88870,   0.00002,     0.385 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    356.49200,   0.00005,     0.067 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    135.34120,   0.00002,     0.626 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    257.55205,   0.00002,     0.086 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.69973,   0.00001,     0.060 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   0,   0,   0 },     98.69973,   0.00001,     0.101 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,   2,   0,   0 },    109.99821,   0.00002,     0.029 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -3,   1,   1 },     21.64074,   0.00002,     2.188 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    338.29818,   0.00002,     0.072 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    253.09434,   0.00004,     1.676 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -1,   0,   1 },     21.64074,   0.00003,     0.084 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    228.46807,   0.00002,     0.606 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.43673,   0.00002,     0.064 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    341.55825,   0.00002,     0.422 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   0,   0 },      1.77047,   0.00005,     0.092 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,   0 },     18.33043,   0.00002,     0.028 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    314.66815,   0.00009,     0.078 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    153.25719,   0.00005,     2.480 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    120.53283,   0.00001,     0.073 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,  -1,   1 },    251.83642,   0.00002,     3.154 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },    240.39388,   0.00004,     0.072 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   1,  -1 },     87.15954,   0.00002,     1.537 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    292.94797,   0.00003,     0.730 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,   0 },     75.68481,   0.00004,     0.084 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -4,   2,   0 },    240.17166,   0.00001,     0.099 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.91061,   0.00008,     0.084 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     89.31352,   0.00003,     0.713 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     34.30042,   0.00003,     0.068 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   0,  -1 },    355.68739,   0.00001,     0.077 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   1,  -1 },    100.85421,   0.00002,     2.862 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -1,   2,  -1 },    175.68739,   0.00001,     0.074 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },     27.26966,   0.00002,     0.073 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    188.17760,   0.00004,     2.697 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,   0,   0 },    107.83359,   0.00004,     0.078 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,   1,  -2 },    156.50375,   0.00001,     1.420 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,  -1,  -1 },    289.17288,   0.00002,     0.702 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   1,   0,  -1 },    175.76688,   0.00003,     0.085 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    116.47472,   0.00005,     0.092 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    296.46987,   0.00002,     0.416 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.40503,   0.00003,     0.079 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -1,   0 },    110.49147,   0.00002,     1.589 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,   0,   0 },     98.40503,   0.00003,     0.072 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   1,  -1,  -1 },    251.57088,   0.00002,     2.355 },
 { {   0,  19, -16,   0,   0,   0,   0,   0,  -2,   1,   0 },      7.01975,   0.00001,     0.294 },
 { {   0,  19, -14,   0,   0,   0,   0,   0,   2,  -2,   0 },    286.11208,   0.00001,     0.066 },
 { {   0,  20, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    246.94442,   0.00003,     0.052 },
 { {   0,  20, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.73861,   0.00002,     0.055 },
 { {   0,  20, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.72913,   0.00002,     0.201 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },     76.02921,   0.00003,     0.058 },
 { {   0,  20, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.60937,   0.00004,     0.252 },
 { {   0,  20, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.53344,   0.00004,     0.062 },
 { {   0,  20, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    161.76613,   0.00002,     0.097 },
 { {   0,  20, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    218.10786,   0.00003,     0.066 },
 { {   0,  20, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    139.35967,   0.00001,     0.507 },
 { {   0,  20, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     45.92023,   0.00001,     0.089 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -3,   1,   1 },    102.53096,   0.00004,    12.123 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    135.85091,   0.00002,     0.070 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    315.85091,   0.00002,     0.081 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.41623,   0.00001,     0.065 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -1,   1,  -1 },    293.41623,   0.00001,     0.494 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -1,   2,  -1 },    113.41623,   0.00001,     0.089 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.70850,   0.00001,     0.062 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    278.48022,   0.00001,     0.364 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   0,   0,   0 },     98.70850,   0.00001,     0.095 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,   2,   0,   0 },    110.00440,   0.00003,     0.028 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   1,   0 },     18.18419,   0.00001,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -4,   2,   0 },    340.31558,   0.00002,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.23119,   0.00014,     0.076 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },     18.18419,   0.00118,    34.812 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },     52.69741,   0.00008,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,  -1,   1 },    107.24334,   0.00002,     1.446 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   0,  -1 },    228.30622,   0.00004,     0.070 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   1,  -1 },    277.74300,   0.00001,     0.976 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    326.30986,   0.00001,     0.573 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    348.79911,   0.00005,     0.087 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   0,   1,   0 },     18.18419,   0.00001,     0.040 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,   0,   0 },     18.26399,   0.00003,     0.028 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     43.90422,   0.00001,     0.082 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    200.88174,   0.00004,     0.972 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },    283.38152,   0.00007,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,   0 },     93.62147,   0.00001,     7.056 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    297.35671,   0.00071,    39.938 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,  -1 },    297.91452,   0.00004,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    141.58314,   0.00001,     0.071 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    299.84282,   0.00003,     0.080 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    260.80620,   0.00003,     0.089 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     99.77311,   0.00002,     0.493 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },    145.90271,   0.00002,     0.065 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,  -1,   1 },    201.05223,   0.00003,     0.764 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },    325.60871,   0.00005,     0.081 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    206.35429,   0.00003,     3.926 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,  -1 },    108.19655,   0.00002,    11.370 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    334.54257,   0.00006,     0.074 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   1,  -2 },    294.24945,   0.00002,    12.690 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   1,  -1,  -1 },     21.76948,   0.00002,     1.251 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.34455,   0.00001,     0.098 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },    294.34455,   0.00001,     0.330 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    241.81669,   0.00001,     0.083 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     61.81669,   0.00001,     0.069 },
 { {   0,  20, -18,   0,   0,   0,   0,   0,   1,  -1,  -1 },    161.98184,   0.00001,     4.977 },
 { {   0,  20, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    187.15323,   0.00001,     0.108 },
 { {   0,  20, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    187.15323,   0.00001,     0.058 },
 { {   0,  21, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    247.21213,   0.00003,     0.054 },
 { {   0,  21, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.75588,   0.00002,     0.057 },
 { {   0,  21, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.72069,   0.00002,     0.230 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.44575,   0.00005,     0.060 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.52962,   0.00004,     0.299 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    255.98870,   0.00001,     0.101 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,   0,   0,   0 },     69.82269,   0.00003,     0.123 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },     63.08945,   0.00003,     0.064 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },     94.46378,   0.00002,     0.426 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    133.95643,   0.00003,     0.092 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    205.82614,   0.00005,     0.068 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },     84.43295,   0.00003,     0.743 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },    136.74142,   0.00001,     0.084 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,  -1,   2 },    103.95880,   0.00007,    81.722 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    162.13858,   0.00002,     0.074 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },     25.12955,   0.00005,     2.892 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    258.31939,   0.00001,     0.077 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -1,  -1,   1 },    113.42501,   0.00001,     0.938 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    293.42501,   0.00003,     0.068 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },     98.71469,   0.00001,     0.472 },
 { {   0,  21, -23,   0,   0,   0,   0,   0,   0,   0,   0 },     98.71469,   0.00003,     0.090 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    169.79050,   0.00006,     0.079 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    323.97891,   0.00004,     1.528 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    217.15330,   0.00002,     0.072 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   1 },    193.19636,   0.00006,    15.151 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },    283.76929,   0.00004,     0.073 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,   0 },    153.51456,   0.00002,     4.296 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   2,  -1 },    299.01504,   0.00001,     0.078 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,  -1,   0 },    225.55451,   0.00002,     0.893 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    327.44265,   0.00003,     0.082 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,   2,  -2,   0 },     18.07497,   0.00002,     0.097 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    251.50044,   0.00002,     0.086 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,  -1,   1,  -1 },     54.27069,   0.00001,     1.665 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,  -1,   0 },    279.48476,   0.00007,     8.350 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    245.73357,   0.00004,     0.076 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,  -1,   1,  -1 },    145.33983,   0.00002,     0.625 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    121.54208,   0.00001,     0.081 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   0,  -1,   0 },    119.84758,   0.00002,     1.136 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    301.54208,   0.00001,     0.071 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   0,   1,  -2 },    156.43770,   0.00001,     1.829 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -2,   0 },    215.53700,   0.00007,   147.570 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,  -1 },    197.35029,   0.00003,     5.763 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },    153.47619,   0.00005,     0.076 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.31242,   0.00004,     0.104 },
 { {   0,  21, -19,   0,   0,   0,   0,   0,   0,  -1,   0 },    225.38397,   0.00001,     0.532 },
 { {   0,  21, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    201.32517,   0.00001,     0.065 },
 { {   0,  21, -17,   0,   0,   0,   0,   0,   4,  -3,  -2 },     17.85044,   0.00003,   759.662 },
 { {   0,  21, -16,   0,   0,   0,   0,   0,   2,   0,   0 },    107.73447,   0.00002,     0.023 },
 { {   0,  22, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    247.46138,   0.00003,     0.056 },
 { {   0,  22, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.75733,   0.00003,     0.059 },
 { {   0,  22, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.72645,   0.00003,     0.269 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },     70.71411,   0.00009,     0.063 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.68333,   0.00003,     0.368 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    255.94666,   0.00001,     0.095 },
 { {   0,  22, -27,   0,   0,   0,   0,   0,   0,   0,   0 },     69.78376,   0.00003,     0.114 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     31.81554,   0.00003,     0.067 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },     62.40300,   0.00001,     0.581 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    150.00742,   0.00001,     0.087 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.08813,   0.00006,     0.072 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    219.06785,   0.00002,     1.388 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    187.17164,   0.00001,     0.080 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },      6.85543,   0.00004,     3.574 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,  -2,   1 },    293.43119,   0.00001,     0.073 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,  -1,   1 },    141.30745,   0.00002,     2.270 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,   0,   1 },    113.43119,   0.00001,     0.078 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,  -1,   1,  -1 },    321.30870,   0.00002,     1.292 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    100.19737,   0.00001,     0.669 },
 { {   0,  22, -24,   0,   0,   0,   0,   0,   0,   0,   0 },     98.71845,   0.00003,     0.085 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -2,   1,   0 },    284.77466,   0.00001,     0.781 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,  -2,   1 },    201.53907,   0.00002,     0.079 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },     67.17037,   0.00005,     0.077 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,   1 },    201.53907,   0.00002,     0.072 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   1,  -1 },     70.40947,   0.00003,     4.425 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   2,  -1 },      5.70917,   0.00002,     0.074 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,  -2,   0 },    113.72516,   0.00001,     0.073 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    350.18130,   0.00005,     2.023 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    184.50099,   0.00002,     0.078 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,   2,   0,   0 },     18.03604,   0.00002,     0.027 },
 { {   0,  22, -22,   0,   0,   0,   0,   0,   1,  -1,  -1 },     82.07121,   0.00004,     1.825 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },    154.93166,   0.00002,     0.085 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    159.81300,   0.00007,     0.068 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,  -2,  -1 },    288.77819,   0.00002,     0.078 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   1,   0,  -1 },    108.77819,   0.00002,     0.073 },
 { {   0,  22, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.27851,   0.00001,     0.111 },
 { {   0,  23, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    247.69374,   0.00003,     0.058 },
 { {   0,  23, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.75660,   0.00003,     0.061 },
 { {   0,  23, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.71787,   0.00004,     0.323 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },     70.84604,   0.00007,     0.065 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },    243.66350,   0.00001,     0.477 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    255.90314,   0.00001,     0.090 },
 { {   0,  23, -28,   0,   0,   0,   0,   0,   0,   0,   0 },     69.74324,   0.00003,     0.107 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },     65.24185,   0.00002,     0.070 },
 { {   0,  23, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },    189.39595,   0.00002,     0.914 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },     79.62777,   0.00002,     0.075 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },     96.00394,   0.00004,    10.559 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    259.62777,   0.00002,     0.076 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    216.64096,   0.00004,     0.081 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    120.89765,   0.00003,     1.105 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },     60.63037,   0.00002,     0.071 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,  -1,   1 },    121.24690,   0.00002,     5.404 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,  -1 },    354.08940,   0.00005,     0.075 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,   0 },    131.74040,   0.00006,    54.684 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   1,  -1 },    247.80359,   0.00004,     6.735 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   0,  -2,   0 },    296.40668,   0.00002,     0.071 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },     22.79295,   0.00002,     1.151 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    333.53242,   0.00001,     0.525 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,  -1,  -1,   1 },     21.50176,   0.00001,     0.844 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    276.20069,   0.00001,     0.076 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },     29.57980,   0.00011,     7.618 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,  -1 },    176.15948,   0.00003,    28.187 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    187.38054,   0.00008,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   1,  -1,  -1 },    201.50176,   0.00001,     1.084 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,  -2,   0 },    197.99552,   0.00001,     0.086 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   2,   0,   0 },     17.99552,   0.00001,     0.026 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    252.63329,   0.00002,     0.097 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    252.63329,   0.00002,     0.062 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,  -2,   0 },    249.60783,   0.00001,     0.082 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },     64.81111,   0.00003,     0.884 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    249.60783,   0.00001,     0.070 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,  -1 },    281.56958,   0.00011,    12.901 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -1 },    293.56334,   0.00004,     0.075 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.67254,   0.00002,     0.107 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,  -2,   1,   0 },    340.67253,   0.00002,     0.256 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   1,  -1,  -1 },    175.62790,   0.00002,     0.928 },
 { {   0,  23, -22,   0,   0,   0,   0,   0,   2,  -1,  -2 },    178.40229,   0.00008,    42.111 },
 { {   0,  23, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.24288,   0.00001,     0.120 },
 { {   0,  24, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    247.91049,   0.00003,     0.060 },
 { {   0,  24, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.75379,   0.00001,     0.064 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },     73.71205,   0.00005,     0.068 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    249.58987,   0.00003,     0.681 },
 { {   0,  24, -29,   0,   0,   0,   0,   0,   0,   0,   0 },     69.70120,   0.00003,     0.100 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    169.16347,   0.00002,     0.073 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },    107.96111,   0.00007,     2.131 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    349.16347,   0.00002,     0.078 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    276.81221,   0.00006,     0.079 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,  -2,   1,   0 },    207.80373,   0.00003,     1.884 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,  -1,  -1,   1 },    278.49242,   0.00002,     5.279 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },     11.93342,   0.00003,     0.804 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },     94.81630,   0.00002,     0.068 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -1,  -1,   1 },    113.43641,   0.00001,     1.234 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,  -1,   1,  -1 },    339.47748,   0.00002,     2.096 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },    185.25086,   0.00001,     0.074 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },     43.46344,   0.00003,     4.112 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    352.20165,   0.00003,     0.077 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    167.73993,   0.00001,     0.093 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.73993,   0.00001,     0.395 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    167.73993,   0.00001,     0.063 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,  -1,   0,  -1 },     21.46283,   0.00003,     0.085 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,  -1,   0 },    278.08698,   0.00001,     1.321 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    161.56111,   0.00003,     0.071 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,  -3,   1 },    201.46283,   0.00001,    27.360 },
 { {   0,  24, -25,   0,   0,   0,   0,   0,   1,  -2,  -1 },    179.51335,   0.00002,     0.074 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },     73.08429,   0.00001,     0.283 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    237.98632,   0.00001,     0.087 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,   0,   0 },     78.59599,   0.00004,     0.067 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   1,  -1,  -1 },    267.56255,   0.00001,     1.422 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   2,  -3,   0 },    137.84918,   0.00004,    58.088 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.63022,   0.00002,     0.115 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },    160.63022,   0.00002,     0.056 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,   0,  -1,   0 },    341.11071,   0.00001,     0.363 },
 { {   0,  25, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    248.11300,   0.00003,     0.062 },
 { {   0,  25, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.74902,   0.00001,     0.066 },
 { {   0,  25, -31,   0,   0,   0,   0,   0,  -2,   1,   0 },    161.74902,   0.00001,     0.542 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },     73.89974,   0.00003,     0.071 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   2,   0 },     75.87580,   0.00002,     0.081 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.65768,   0.00001,     0.382 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,   0,   0,   0 },     69.65768,   0.00003,     0.094 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    102.21071,   0.00006,     0.076 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },    245.91746,   0.00005,     6.396 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.74943,   0.00001,     0.075 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.12535,   0.00003,     0.083 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -2,   1,   0 },    244.87167,   0.00001,     0.865 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    278.91717,   0.00001,     0.069 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -1,   0,  -1 },     38.62733,   0.00002,     0.076 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -1,   1,  -1 },     75.85706,   0.00002,     9.749 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,  -1,   2,  -1 },     38.62733,   0.00002,     0.075 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,  -2,   0 },    212.52555,   0.00001,     0.072 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,  -1,   0 },    217.54486,   0.00003,     1.619 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,   0,   0 },     39.64846,   0.00003,     0.079 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,  -1,   1,  -1 },    293.43567,   0.00001,     0.907 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,  -1,   0 },    146.53298,   0.00001,     2.616 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   0,   0 },    258.19868,   0.00002,     0.073 },
 { {   0,  25, -27,   0,   0,   0,   0,   0,   0,   1,  -2 },    247.09837,   0.00003,    20.488 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    321.46872,   0.00001,     0.317 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,  -1,   0 },    327.23588,   0.00001,     0.723 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    186.66377,   0.00003,     0.068 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,  -1,  -1 },    158.59275,   0.00003,     3.044 },
 { {   0,  25, -26,   0,   0,   0,   0,   0,   1,   1,  -3 },      6.76435,   0.00006,   201.839 },
 { {   0,  25, -24,   0,   0,   0,   0,   0,  -2,   1,   0 },    163.91520,   0.00001,     0.194 },
 { {   0,  25, -24,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.07020,   0.00001,     0.296 },
 { {   0,  25, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    353.83036,   0.00003,     0.060 },
 { {   0,  26, -33,   0,   0,   0,   0,   0,  -2,   0,   0 },    248.30224,   0.00003,     0.065 },
 { {   0,  26, -33,   0,   0,   0,   0,   0,  -2,   1,   0 },    248.30224,   0.00001,     0.451 },
 { {   0,  26, -32,   0,   0,   0,   0,   0,  -3,   0,   2 },    196.68313,   0.00004,   418.786 },
 { {   0,  26, -32,   0,   0,   0,   0,   0,  -2,   1,   0 },    161.74236,   0.00001,     0.821 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    211.12417,   0.00004,     0.074 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.12969,   0.00002,     0.077 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.61273,   0.00001,     0.502 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,   0,   0,   0 },     69.61273,   0.00003,     0.089 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    265.80269,   0.00003,     0.080 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   1,   0 },    191.43382,   0.00002,     1.279 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -2,   2,   0 },    341.76300,   0.00002,     0.071 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -1,  -1,   1 },    323.75564,   0.00003,    16.230 },
 { {   0,  26, -30,   0,   0,   0,   0,   0,  -1,   0,   0 },    229.78800,   0.00002,     9.511 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    197.30346,   0.00003,     0.087 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   1,   0 },     20.52337,   0.00002,     0.561 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.55982,   0.00002,     0.067 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   0 },    233.18812,   0.00003,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   0 },    238.39622,   0.00079,   127.422 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   0 },    294.92199,   0.00003,     0.075 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,  -2,   0 },    278.47046,   0.00001,     0.082 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,  -1,   0 },    100.91851,   0.00001,     0.992 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   0,   0,   0 },     98.47046,   0.00001,     0.070 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,  -1,  -1 },    340.60035,   0.00005,    21.779 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   0,  -1 },    199.48575,   0.00004,     0.076 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },    142.20595,   0.00002,     0.106 },
 { {   0,  26, -27,   0,   0,   0,   0,   0,  -2,   2,   0 },    142.20595,   0.00002,     0.059 },
 { {   0,  26, -26,   0,   0,   0,   0,   0,  -2,   1,   0 },    253.08075,   0.00001,     0.209 },
 { {   0,  26, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    157.61201,   0.00001,     0.134 },
 { {   0,  26, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    157.61201,   0.00001,     0.053 },
 { {   0,  26, -25,   0,   0,   0,   0,   0,   0,  -2,   0 },    161.02815,   0.00001,     0.108 },
 { {   0,  26, -25,   0,   0,   0,   0,   0,   0,   0,   0 },    341.02815,   0.00001,     0.058 },
 { {   0,  27, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    248.47930,   0.00003,     0.067 },
 { {   0,  27, -33,   0,   0,   0,   0,   0,  -2,   1,   0 },    334.06232,   0.00004,     1.687 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.00479,   0.00001,     0.078 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   1,   0 },    288.09360,   0.00002,     2.455 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   2,   0 },    211.10728,   0.00002,     0.073 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,   0,  -1,   0 },    249.44103,   0.00003,     0.732 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,   0,   0,   0 },     69.56638,   0.00003,     0.084 },
 { {   0,  27, -31,   0,   0,   0,   0,   0,   0,  -1,   0 },    325.58795,   0.00001,     2.728 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    143.61082,   0.00003,     0.076 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   1,   0 },     93.25798,   0.00002,    30.385 },
 { {   0,  28, -34,   0,   0,   0,   0,   0,  -2,   2,   0 },     13.06906,   0.00001,     0.075 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,  -1,   1,  -1 },    273.17053,   0.00003,    47.954 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,  -2,   0 },     69.38922,   0.00002,     0.071 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,   0,   0 },    249.38922,   0.00002,     0.080 },
 { {   0,  28, -32,   0,   0,   0,   0,   0,   0,  -1,   0 },    252.58935,   0.00005,     3.863 },
 { {   0,  28, -31,   0,   0,   0,   0,   0,   1,  -1,  -1 },    155.14150,   0.00001,     4.875 },
 { {   0,  28, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    143.62657,   0.00001,     0.122 },
 { {   0,  29, -36,   0,   0,   0,   0,   0,  -2,  -1,   2 },    134.85148,   0.00002,   124.208 },
 { {   0,  29, -35,   0,   0,   0,   0,   0,  -2,   1,   0 },    341.71187,   0.00001,     1.519 },
 { {   0,  29, -35,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.71187,   0.00001,     0.896 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },     68.52736,   0.00005,     0.086 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,  -2,   2,   0 },     71.20539,   0.00002,     0.067 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,  -2,   0 },    255.33039,   0.00001,     0.075 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,   0,  -1,   0 },    326.35535,   0.00001,     8.652 },
 { {   0,  29, -33,   0,   0,   0,   0,   0,   1,  -2,   0 },     31.62104,   0.00002,   385.960 },
 { {   0,  29, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },     99.70016,   0.00001,     0.132 },
 { {   0,  30, -38,   0,   0,   0,   0,   0,  -2,   1,   0 },    334.46690,   0.00002,     1.397 },
 { {   0,  30, -37,   0,   0,   0,   0,   0,  -2,   1,   0 },     23.44040,   0.00001,     3.522 },
 { {   0,  30, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.61633,   0.00006,     0.084 },
 { {   0,  30, -36,   0,   0,   0,   0,   0,   0,   0,   0 },    161.69847,   0.00003,     0.078 },
 { {   0,  30, -35,   0,   0,   0,   0,   0,   0,  -2,   0 },    251.18451,   0.00001,     0.078 },
 { {   0,  30, -35,   0,   0,   0,   0,   0,   0,   0,   0 },    248.72888,   0.00003,     0.073 },
 { {   0,  30, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    159.97196,   0.00001,     0.143 },
 { {   0,  30, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    256.67384,   0.00005,    35.458 },
 { {   0,  31, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.72789,   0.00004,     0.075 },
 { {   0,  31, -39,   0,   0,   0,   0,   0,  -2,   2,   0 },    154.72789,   0.00004,     0.076 },
 { {   0,  31, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },     69.08241,   0.00002,     0.081 },
 { {   0,  31, -38,   0,   0,   0,   0,   0,  -2,   1,   0 },     69.04196,   0.00003,     1.100 },
 { {   0,  31, -38,   0,   0,   0,   0,   0,  -1,   0,   0 },    196.06844,   0.00001,    44.499 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.66261,   0.00003,     0.088 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,  -2,   1,   0 },    161.68351,   0.00001,     0.524 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,  -2,   0 },    161.59648,   0.00004,     0.076 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,  -1,   0 },    328.39340,   0.00006,     7.382 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,   0,   0 },    161.59648,   0.00004,     0.075 },
 { {   0,  31, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.36808,   0.00002,     0.097 },
 { {   0,  31, -36,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.36808,   0.00001,     0.881 },
 { {   0,  31, -35,   0,   0,   0,   0,   0,   2,  -1,  -2 },    102.97481,   0.00001,    35.801 },
 { {   0,  31, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    189.42635,   0.00002,     0.426 },
 { {   0,  32, -38,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.66711,   0.00001,     1.314 },
 { {   0,  32, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    113.85268,   0.00001,     0.368 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   1,  -1 },    163.87403,   0.00001,     0.075 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   2,  -1 },    163.87403,   0.00022,  2005.078 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -3,   3,  -1 },    163.87403,   0.00001,     0.075 },
 { {   0,  32, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    320.91856,   0.00001,     1.387 },
 { {   0,  33, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.40936,   0.00001,     0.196 },
 { {   0,  33, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    113.80238,   0.00001,     0.477 },
 { {   0,  33, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    103.20905,   0.00001,    10.501 },
 { {   0,  34, -42,   0,   0,   0,   0,   0,   0,  -1,   0 },    193.74896,   0.00003,    83.102 },
 { {   0,  34, -42,   0,   0,   0,   0,   0,   0,   0,   0 },    335.43222,   0.00007,     0.075 },
 { {   0,  34, -35,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.53732,   0.00001,     0.224 },
 { {   0,  34, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.02003,   0.00003,     0.680 },
 { {   0,  34, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    320.94296,   0.00001,     1.886 },
 { {   0,  35, -34,   0,   0,   0,   0,   0,  -4,   2,   0 },    340.09522,   0.00002,    14.253 },
 { {   0,  35, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    140.14577,   0.00002,     6.418 },
 { {   0,  35, -31,   0,   0,   0,   0,   0,  -1,   0,  -1 },    117.69977,   0.00003,     9.799 },
 { {   0,  36, -47,   0,   0,   0,   0,   0,  -2,   1,   0 },     49.66473,   0.00003,    26.957 },
 { {   0,  36, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    252.43941,   0.00002,     0.451 },
 { {   0,  36, -35,   0,   0,   0,   0,   0,  -2,   0,   0 },    222.43805,   0.00002,     0.821 },
 { {   0,  36, -32,   0,   0,   0,   0,   0,   0,  -2,   0 },    332.69353,   0.00002,   136.522 },
 { {   0,  37, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },     45.61113,   0.00001,     0.386 },
 { {   0,  37, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    261.09169,   0.00003,     1.686 },
 { {   0,  38, -38,   0,   0,   0,   0,   0,  -3,   0,   1 },     56.25155,   0.00003,    11.607 },
 { {   0,  38, -37,   0,   0,   0,   0,   0,  -2,   0,   0 },     28.98772,   0.00008,    30.876 },
 { {   0,  38, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    113.53447,   0.00001,     0.969 },
 { {   0,  38, -36,   0,   0,   0,   0,   0,  -1,   0,  -1 },    251.37552,   0.00002,    46.780 },
 { {   0,  39, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },     32.03176,   0.00002,     2.923 },
 { {   0,  39, -38,   0,   0,   0,   0,   0,  -2,   0,   0 },    232.35240,   0.00004,     1.520 },
 { {   0,  39, -38,   0,   0,   0,   0,   0,  -1,   0,  -1 },    271.71037,   0.00002,     2.526 },
 { {   0,  40, -41,   0,   0,   0,   0,   0,  -3,   2,  -1 },    172.39781,   0.00004,   271.239 },
 { {   0,  40, -41,   0,   0,   0,   0,   0,  -2,   0,   0 },     96.62249,   0.00001,     1.396 },
 { {   0,  40, -40,   0,   0,   0,   0,   0,  -2,   0,   0 },      5.87933,   0.00003,     3.528 },
 { {   0,  40, -39,   0,   0,   0,   0,   0,  -1,   0,  -1 },    228.48820,   0.00001,     4.354 },
 { {   0,  41, -43,   0,   0,   0,   0,   0,  -3,   0,   1 },    181.02355,   0.00002,    26.827 },
 { {   0,  41, -42,   0,   0,   0,   0,   0,  -2,   0,   0 },    132.08486,   0.00005,    10.984 },
 { {   0,  41, -41,   0,   0,   0,   0,   0,  -1,   0,  -1 },    250.77969,   0.00002,     6.906 },
 { {   0,  42, -44,   0,   0,   0,   0,   0,  -2,   0,   0 },    124.21148,   0.00001,     2.148 },
 { {   0,  42, -43,   0,   0,   0,   0,   0,  -2,   0,   0 },     96.55005,   0.00001,     1.871 },
 { {   0,  42, -43,   0,   0,   0,   0,   0,  -1,   0,  -1 },     20.53949,   0.00002,     1.926 },
 { {   0,  42, -42,   0,   0,   0,   0,   0,  -1,   0,  -1 },    292.60330,   0.00002,     2.080 },
 { {   0,  43, -46,   0,   0,   0,   0,   0,  -2,   0,   0 },    211.27677,   0.00001,     1.191 },
 { {   0,  43, -45,   0,   0,   0,   0,   0,  -2,   0,   0 },    264.56372,   0.00004,     6.250 },
 { {   0,  43, -44,   0,   0,   0,   0,   0,  -2,   0,   0 },    132.50076,   0.00002,     0.862 },
 { {   0,  43, -44,   0,   0,   0,   0,   0,  -1,   0,  -1 },    343.39930,   0.00002,     9.413 },
 { {   0,  43, -42,   0,   0,   0,   0,   0,   1,   0,  -3 },    175.49236,   0.00007,   773.987 },
 { {   0,  44, -47,   0,   0,   0,   0,   0,  -2,   0,   0 },     19.57494,   0.00002,     4.663 },
 { {   0,  44, -45,   0,   0,   0,   0,   0,   0,  -2,   0 },    234.69691,   0.00003,    86.878 },
 { {   0,  45, -49,   0,   0,   0,   0,   0,  -2,   0,   0 },    304.06344,   0.00002,     1.698 },
 { {   0,  45, -48,   0,   0,   0,   0,   0,  -2,   0,   0 },    155.22993,   0.00002,     2.433 },
 { {   0,  45, -47,   0,   0,   0,   0,   0,  -1,   0,  -1 },    113.07128,   0.00001,     2.799 },
 { {   0,  46, -51,   0,   0,   0,   0,   0,  -2,   0,   0 },    248.47942,   0.00002,     1.038 },
 { {   0,  46, -50,   0,   0,   0,   0,   0,  -2,   0,   0 },     74.82605,   0.00004,    27.342 },
 { {   0,  47, -52,   0,   0,   0,   0,   0,  -2,   0,   0 },     66.66755,   0.00001,     2.959 },
 { {   0,  48, -54,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.23629,   0.00002,     1.404 },
 { {   0,  48, -53,   0,   0,   0,   0,   0,  -2,   0,   0 },     18.59116,   0.00001,     3.477 },
 { {   0,  49, -55,   0,   0,   0,   0,   0,  -2,   0,   0 },    159.16878,   0.00003,    11.513 },
 { {   0,  50, -57,   0,   0,   0,   0,   0,  -2,   0,   0 },     70.62820,   0.00001,     2.168 },
 { {   0,  50, -56,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.19301,   0.00002,     1.856 },
 { {   0,  51, -58,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.61950,   0.00001,     6.091 },
 { {   0,  51, -55,   0,   0,   0,   0,   0,   1,   0,  -3 },     88.13621,   0.00002,   345.601 },
 { {   0,  52, -60,   0,   0,   0,   0,   0,  -2,   0,   0 },    158.61062,   0.00002,     4.755 },
 { {   0,  52, -58,   0,   0,   0,   0,   0,   0,  -2,   0 },    341.02755,   0.00003,    63.711 },
 { {   1,   0,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    191.14897,   0.00002,     0.075 },
 { {   1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    287.01148,   0.00002,     6.575 },
 { {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     42.97331,   0.00001,     0.975 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    320.84620,   0.00005,     0.075 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    218.71494,   0.00023,    39.271 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0 },    218.36976,   0.00001,     0.076 },
 { {   4,   0,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    123.75027,   0.00004,     7.897 },
 { {   4,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     43.57579,   0.00001,     2.836 },
 { {   5,   0,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.23702,   0.00001,     3.588 },
 { {   7,   0,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    142.84581,   0.00002,     1.486 },
 { {   8,   0,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },      7.63995,   0.00001,     1.920 },
 { {   8,   0,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    114.91093,   0.00003,     2.087 },
 { {   9,   0, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    322.07377,   0.00001,     2.712 },
 { {  13,   0, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    214.04379,   0.00001,    62.252 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 833> g_ELP14
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },     73.43578,   0.00010,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },    277.11719,   0.00005,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   1 },    300.69188,   0.00004,     0.089 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1 },    354.55159,   0.00001,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   1 },    354.55159,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },      4.02418,   0.00006,     0.080 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,  -1 },    300.72270,   0.00004,     0.088 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,  -1 },    259.58132,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,  -1 },     78.69282,   0.00006,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   1 },     79.58132,   0.00005,     6.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,  -1 },    270.67200,   0.00002,     0.074 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   1 },    270.67200,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,  -1 },     79.58132,   0.00005,     5.956 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   1 },     78.69282,   0.00006,     0.037 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   1 },    259.58132,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0 },     55.07650,   0.00002,     0.080 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1 },    302.93944,   0.00003,     0.093 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },    294.37462,   0.00005,     0.084 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2 },    114.37462,   0.00005,     0.067 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1 },    120.03140,   0.00003,     0.078 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1 },    120.03140,   0.00003,     0.072 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2 },    230.49656,   0.00003,     0.072 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0 },     12.94819,   0.00004,     0.078 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0 },    159.71687,   0.00002,     0.088 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   0,   0 },     15.74606,   0.00009,     0.074 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   1,   0,  -2 },     72.48311,   0.00003,     0.078 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   1,   0,   0 },    338.59946,   0.00004,     0.072 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,   0,  -1 },    193.96228,   0.00003,     0.077 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   0,  -3 },    102.78502,   0.00001,     0.075 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   0,  -1 },     43.30931,   0.00007,     0.074 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,   0,  -1 },    251.26340,   0.00002,     0.072 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,   0,  -1 },    280.46229,   0.00006,     0.069 },
 { {   0,   0,   0,   7,   0,   0,   0,   0,   2,   0,  -1 },    129.65557,   0.00002,     0.067 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,  -2,   0,   1 },    114.24115,   0.00003,     0.074 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   0,   1 },    135.82277,   0.00005,     0.077 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   0,   3 },    141.67736,   0.00001,     0.072 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -1,   0,   0 },    166.19613,   0.00005,     0.071 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -1,   0,   2 },     15.27608,   0.00002,     0.078 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   0 },     83.24752,   0.00001,     0.074 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -1,   0,   2 },     83.24752,   0.00001,     0.075 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -1,   0,   0 },    109.85138,   0.00005,     0.077 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   1,   0,  -2 },    320.56658,   0.00001,     0.066 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   1,   0,   0 },    320.56658,   0.00001,     0.085 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   1 },     52.92886,   0.00004,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   0 },    128.11846,   0.00007,     0.080 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1 },     46.38324,   0.00003,     0.037 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1 },     40.62315,   0.00002,     9.672 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1 },     40.62315,   0.00002,     4.346 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1 },     46.38324,   0.00003,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   1,   0,   0 },    351.37034,   0.00004,     0.081 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,  -1 },     54.18332,   0.00002,     0.499 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   1 },     54.18332,   0.00002,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,  -1 },     54.14347,   0.00006,     0.089 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   1 },    273.70254,   0.00002,     0.092 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   0 },    122.13777,   0.00007,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   2 },    319.35503,   0.00002,     0.067 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1 },     89.75374,   0.00004,     0.077 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1 },     89.75374,   0.00004,     0.072 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0 },     37.68611,   0.00002,     0.078 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,  -3 },     86.19431,   0.00002,     0.066 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,  -1 },    266.19431,   0.00002,     0.085 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,  -1 },     30.47728,   0.00001,     0.079 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   1 },     30.47728,   0.00001,     0.071 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,  -1 },     43.95014,   0.00001,     0.027 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   1 },     43.94826,   0.00005,     0.095 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0 },    342.04010,   0.00005,     0.087 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -1,   0 },    153.54295,   0.00002,    17.980 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },    333.54238,   0.00032,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   1,   0 },    333.54238,   0.00002,     0.038 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,  -1 },     42.81746,   0.00002,     0.082 },
 { {   0,   0,   1,   0,  -2,   5,   0,   0,   1,   0,   0 },    141.26330,   0.00003,     0.075 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,   1,   0,   0 },    164.86231,   0.00012,     0.075 },
 { {   0,   0,   1,   0,   0,  -2,   0,   0,   1,   0,   0 },     14.88171,   0.00014,     0.075 },
 { {   0,   0,   1,   0,   0,  -1,   0,   0,   1,   0,   0 },    273.50613,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1 },    183.12543,   0.00004,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -2 },    101.83120,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    302.39044,   0.00005,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   2 },    178.30350,   0.00001,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.30072,   0.00004,     0.530 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },    120.79923,   0.00006,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },    120.79923,   0.00006,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0 },    138.64316,   0.00003,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0 },    189.57541,   0.00007,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0 },     11.13687,   0.00132,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0 },      9.57594,   0.00007,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -3 },     32.97163,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },    281.31434,   0.00006,     0.081 },
 { {   0,   0,   1,   0,   0,   1,   0,   0,   1,   0,   0 },    340.86229,   0.00012,     0.075 },
 { {   0,   0,   1,   0,   0,   2,   0,   0,   1,   0,   0 },    136.95201,   0.00006,     0.074 },
 { {   0,   0,   1,   0,   1,  -1,   0,   0,   1,   0,   0 },    167.07365,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,   1,   0,   0 },    177.16017,   0.00014,     0.074 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   1,   0,   0 },    175.97986,   0.00008,     0.075 },
 { {   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,   0 },    141.86655,   0.00017,     0.074 },
 { {   0,   0,   1,   0,   3,   0,   0,   0,   1,   0,   0 },    140.11131,   0.00002,     0.073 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1 },    151.53581,   0.00002,     0.084 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   1 },    151.53581,   0.00002,     0.067 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   1,   0,   0 },    254.26327,   0.00003,     0.072 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   0,  -2 },    178.98604,   0.00002,     0.081 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   1,   0,   0 },    358.98604,   0.00002,     0.069 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,   0,  -3 },    195.51964,   0.00002,     0.074 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,   0,  -1 },    183.42468,   0.00009,     0.075 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   2,   0,  -3 },     71.26741,   0.00004,     0.081 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   2,   0,  -1 },    251.26741,   0.00004,     0.069 },
 { {   0,   0,   2,  -9,   0,   0,   0,   0,  -2,   0,   1 },    230.33298,   0.00006,     0.071 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   0,   1 },     80.55793,   0.00003,     0.074 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   0,   3 },    260.55793,   0.00003,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   0,   1 },    108.19444,   0.00002,     0.077 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   2,  -1 },    108.28104,   0.00001,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   2,   1 },    288.28104,   0.00001,     0.074 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   0,   1 },    316.42107,   0.00001,     0.080 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -2,   0,   3 },    316.42107,   0.00001,     0.070 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -1,   0,   0 },    247.70960,   0.00002,     0.074 },
 { {   0,   0,   2,  -6,   0,   0,   0,   0,  -1,   0,   2 },    278.11075,   0.00001,     0.075 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   0,   0 },     47.55273,   0.00004,     0.077 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   0,   2 },    189.55981,   0.00002,     0.072 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   1 },     27.59950,   0.00002,     0.087 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -1,   0,   0 },     56.02325,   0.00005,     0.080 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1 },    189.14711,   0.00007,     0.074 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1 },    189.14711,   0.00007,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,  -2 },    230.83983,   0.00001,     0.068 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   1,   0,   0 },    230.83983,   0.00001,     0.082 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   0 },     99.15862,   0.00003,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -1,   0,   2 },     64.63208,   0.00002,     0.067 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1 },     57.09551,   0.00003,     0.077 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1 },     57.09551,   0.00003,     0.072 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   1,   0,   0 },    290.44512,   0.00005,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,  -3 },    236.99291,   0.00002,     0.066 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,  -1 },    236.76720,   0.00004,     0.085 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   0 },    202.60402,   0.00005,     0.075 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -3 },    266.07243,   0.00002,     0.069 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,  -1 },    266.07243,   0.00002,     0.082 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -1,   0,   0 },    119.18610,   0.00002,     0.092 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -1,   0,   2 },    299.18610,   0.00002,     0.063 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   1,   0,   0 },    194.06866,   0.00005,     0.072 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,   0,  -1 },    297.29273,   0.00003,     0.078 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,  -1 },    239.33862,   0.00002,     0.040 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   1 },    239.33856,   0.00002,     0.523 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,  -1 },    241.35498,   0.00001,     0.086 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   1 },    241.35498,   0.00001,     0.066 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,  -1 },     59.28916,   0.00003,     0.074 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   1 },     59.28916,   0.00003,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,  -1 },     59.33853,   0.00010,     5.163 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   1 },     59.33697,   0.00011,     0.038 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   1,   0,   0 },    123.55044,   0.00002,     0.070 },
 { {   0,   0,   2,   0,  -1,   0,   0,   0,   1,   0,   0 },    185.34963,   0.00007,     0.070 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1 },    245.92350,   0.00002,     0.088 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1 },    245.92350,   0.00002,     0.065 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0 },    301.04803,   0.00006,     0.070 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -3 },    154.15550,   0.00003,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },    342.92927,   0.00006,     0.075 },
 { {   0,   0,   2,   0,   0,   3,   0,   0,   2,   0,  -3 },     92.98042,   0.00004,     0.074 },
 { {   0,   0,   2,   0,   0,   3,   0,   0,   2,   0,  -1 },    272.98042,   0.00004,     0.075 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   1,   0,   0 },    140.18799,   0.00003,     0.069 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -3 },    261.44139,   0.00002,     0.074 },
 { {   0,   0,   2,   0,   1,   0,   0,   0,   2,   0,  -1 },    261.44139,   0.00002,     0.075 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   1,   0,  -2 },     93.49857,   0.00002,     0.084 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   1,   0,   0 },    273.49857,   0.00002,     0.067 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,   0,  -1 },    191.78434,   0.00009,     0.072 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,   0,  -1 },    222.10175,   0.00006,     0.070 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   2,   0,  -1 },    250.95341,   0.00003,     0.067 },
 { {   0,   0,   2,   4,   0,   0,   0,   0,   2,   0,  -3 },    280.45679,   0.00001,     0.088 },
 { {   0,   0,   2,   4,   0,   0,   0,   0,   2,   0,  -1 },    100.45679,   0.00001,     0.065 },
 { {   0,   0,   3, -11,   0,   0,   0,   0,  -2,   0,   1 },     21.14616,   0.00001,     0.071 },
 { {   0,   0,   3, -11,   0,   0,   0,   0,  -2,   0,   3 },    201.14616,   0.00001,     0.079 },
 { {   0,   0,   3, -10,   0,   0,   0,   0,  -2,   0,   1 },    230.62099,   0.00001,     0.073 },
 { {   0,   0,   3, -10,   0,   0,   0,   0,  -2,   0,   3 },     50.62099,   0.00001,     0.076 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   0,   1 },     79.37877,   0.00002,     0.076 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   2,   1 },    258.93633,   0.00003,     0.075 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   0,   1 },    108.02555,   0.00002,     0.080 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   0,   3 },    108.02555,   0.00002,     0.070 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -1,   0,   0 },    163.80382,   0.00003,     0.073 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   0,   0 },      3.92560,   0.00004,     0.076 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   2,   0 },    339.85551,   0.00005,     0.075 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   0 },     39.74888,   0.00003,     0.080 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,  -1,   0,   2 },    176.16203,   0.00001,     0.070 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -1 },    225.75989,   0.00003,     0.073 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   1 },    225.75989,   0.00003,     0.076 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -1,   0,   0 },    108.64315,   0.00003,     0.083 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1 },     28.91967,   0.00003,     0.076 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1 },     28.91967,   0.00003,     0.073 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,  -2 },    154.51882,   0.00002,     0.071 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   1,   0,   0 },    317.77044,   0.00006,     0.079 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   0 },     43.89753,   0.00004,     0.087 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -1,   0,   2 },    303.72594,   0.00001,     0.065 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },    287.58993,   0.00006,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,  -1 },    236.88424,   0.00005,     0.082 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   0,   0 },    111.97978,   0.00001,     0.091 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -1,   0,   2 },    111.97978,   0.00001,     0.063 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,  -3 },    266.58804,   0.00001,     0.071 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,  -1 },    265.72368,   0.00003,     0.079 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1 },    122.28119,   0.00002,     0.087 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1 },    122.28119,   0.00002,     0.065 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   1,   0,   0 },    277.84284,   0.00005,     0.070 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,  -1 },    117.57588,   0.00003,     0.075 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   1,   0,  -2 },     97.15537,   0.00002,     0.083 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   1,   0,   0 },    316.83723,   0.00003,     0.067 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,   0,  -3 },    140.89132,   0.00004,     0.077 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,   0,  -1 },     74.89159,   0.00002,     0.073 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,  -2 },    157.92807,   0.00003,     0.087 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0 },    153.81155,   0.00005,     0.065 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,  -1 },     14.52258,   0.00004,     0.070 },
 { {   0,   0,   3,   0,   1,   0,   0,   0,   1,   0,   0 },    321.69091,   0.00002,     0.065 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,   0,  -1 },    250.90067,   0.00002,     0.065 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,   0,  -3 },    100.39357,   0.00001,     0.091 },
 { {   0,   0,   3,   3,   0,   0,   0,   0,   2,   0,  -1 },    100.39357,   0.00001,     0.063 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   0,   1 },     50.58788,   0.00004,     0.076 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,   0,   1 },     79.10046,   0.00005,     0.079 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -1,   0,   0 },    167.75209,   0.00003,     0.073 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -2,   0,   1 },    287.70170,   0.00001,     0.083 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -2,   0,   3 },    287.70170,   0.00001,     0.068 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,  -1,   0,   0 },    309.73027,   0.00002,     0.076 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -1,   0,   0 },    339.76684,   0.00006,     0.079 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -1,   0,   2 },    217.75755,   0.00002,     0.070 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,  -1 },      1.93148,   0.00005,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   1 },      1.95733,   0.00027,     0.088 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,  -1 },      2.12896,   0.00008,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   1 },      2.07541,   0.00007,     0.515 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,  -1 },    181.93153,   0.00001,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   1 },    181.93153,   0.00001,     0.076 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,  -1 },      1.47685,   0.00011,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   1 },      1.28718,   0.00009,     6.017 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,  -1 },      3.09929,   0.00010,     0.074 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   0,   1 },      3.09924,   0.00010,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,  -1 },      1.28718,   0.00009,     5.977 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   1 },      1.47685,   0.00011,     0.037 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,  -1 },    181.93262,   0.00001,     0.076 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   1 },    181.93262,   0.00001,     0.025 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,  -1 },      2.07675,   0.00007,     0.515 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   1 },      2.13013,   0.00008,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,  -1 },      1.95837,   0.00027,     0.088 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   1 },      1.93267,   0.00005,     0.026 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   0 },     13.12435,   0.00005,     0.083 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -1,   0,   2 },    224.04613,   0.00002,     0.068 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1 },      2.63717,   0.00001,     0.076 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1 },      2.63717,   0.00001,     0.073 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,   0,  -2 },    232.81948,   0.00003,     0.070 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   1,   0,   0 },    232.81948,   0.00003,     0.079 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   0 },     19.25044,   0.00005,     0.087 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -1,   0,   2 },    252.93377,   0.00002,     0.065 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   1,   0,   0 },    286.61681,   0.00006,     0.076 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,  -3 },     27.46764,   0.00001,     0.068 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,  -1 },    207.59345,   0.00005,     0.082 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -1,   0,   0 },     86.49864,   0.00002,     0.091 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,  -1,   0,   2 },     86.49864,   0.00002,     0.063 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,  -2 },    336.17647,   0.00001,     0.076 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   1,   0,   0 },    132.45280,   0.00005,     0.073 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,  -3 },     56.78705,   0.00003,     0.070 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,  -2 },     92.87911,   0.00001,     0.079 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,   0 },    204.11469,   0.00007,     0.070 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,   0,  -1 },     74.84167,   0.00003,     0.076 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   1,   0,   0 },    267.50416,   0.00002,     0.068 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,  -3 },    312.80713,   0.00002,     0.076 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,  -1 },    111.41408,   0.00006,     0.073 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   1,   0,  -2 },    305.58033,   0.00002,     0.087 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   1,   0,   0 },    305.58033,   0.00002,     0.065 },
 { {   0,   0,   4,  -1,   0,   0,   0,   0,   2,   0,  -1 },     12.94082,   0.00002,     0.068 },
 { {   0,   0,   4,   0,  -1,   0,   0,   0,   1,   0,   0 },    322.98309,   0.00005,     0.061 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,  -3 },    200.71802,   0.00002,     0.087 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,  -1 },     34.74182,   0.00004,     0.065 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   2,   0,  -1 },     71.19172,   0.00003,     0.063 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,  -2,   0,   1 },     50.28506,   0.00001,     0.079 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,  -2,   0,   3 },     49.96695,   0.00003,     0.071 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -2,   0,   1 },    258.89290,   0.00002,     0.082 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   0,   0 },    282.98694,   0.00004,     0.076 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -1,   0,   2 },    169.52350,   0.00002,     0.073 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -1 },    179.97120,   0.00002,     0.076 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1 },    179.97120,   0.00002,     0.073 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,   0,   0 },     28.28932,   0.00001,     0.086 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -1,   0,   2 },     28.28932,   0.00001,     0.066 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   2,   0,  -1 },    178.55479,   0.00003,     0.083 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -1,   0,   0 },     57.36748,   0.00001,     0.090 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -1,   0,   2 },     57.36748,   0.00001,     0.063 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   1,   0,   0 },    121.89060,   0.00005,     0.073 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,  -3 },     27.59826,   0.00003,     0.070 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,  -1 },    207.59826,   0.00003,     0.079 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   1,   0,   0 },    257.56063,   0.00003,     0.071 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -2,  -1 },     59.88418,   0.00003,     0.075 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   1 },    247.55824,   0.00002,     0.037 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,  -3 },    228.81184,   0.00001,     0.073 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,   0,  -1 },     54.68957,   0.00004,     0.076 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   1,   0,   0 },    283.59223,   0.00004,     0.068 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   1,   0,   0 },    346.06156,   0.00005,     0.066 },
 { {   0,   0,   5,  -1,   0,   0,   0,   0,   2,   0,  -1 },     41.95482,   0.00002,     0.063 },
 { {   0,   0,   5,   0,  -2,   0,   0,   0,   1,   0,   0 },    142.07390,   0.00002,     0.058 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   0,  -1 },    101.31261,   0.00006,     0.061 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   0,   1 },    200.96378,   0.00003,     0.078 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -2,   0,   1 },    229.90397,   0.00002,     0.082 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   0,   0 },    311.02079,   0.00002,     0.075 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   0,   2 },    311.02079,   0.00002,     0.074 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   1,   0,   0 },    204.89847,   0.00008,     0.077 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,  -2 },    208.94928,   0.00001,     0.075 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   1,   0,   0 },     69.70837,   0.00005,     0.074 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,   0,  -3 },    178.41879,   0.00002,     0.070 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,   0,  -1 },    178.41879,   0.00002,     0.080 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,  -2 },     35.75749,   0.00001,     0.079 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   1,   0,   0 },    238.38250,   0.00004,     0.071 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   1 },     31.63186,   0.00003,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,  -1 },     43.90650,   0.00002,     5.148 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   1 },     43.90477,   0.00002,     0.038 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,  -1 },     27.84415,   0.00002,     0.077 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   1,   0,   0 },    243.18589,   0.00005,     0.068 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,  -3 },    235.65112,   0.00001,     0.075 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,  -1 },    235.65112,   0.00001,     0.074 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   1,   0,  -2 },     48.25718,   0.00002,     0.086 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   1,   0,   0 },    290.95637,   0.00005,     0.066 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,   0,  -1 },     84.67335,   0.00003,     0.071 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   1,   0,  -2 },    280.28298,   0.00001,     0.090 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   1,   0,   0 },    354.01903,   0.00002,     0.064 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -1,   0,   0 },    298.96411,   0.00002,     0.075 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,  -1 },    274.50791,   0.00001,     0.075 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   0,   0,   1 },    274.50791,   0.00001,     0.074 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,  -2 },    154.45282,   0.00001,     0.075 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   1,   0,   0 },     38.76308,   0.00004,     0.074 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   2,   0,  -1 },    149.39845,   0.00002,     0.080 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   1,   0,   0 },     62.92978,   0.00002,     0.071 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   1,   0,  -2 },    148.41736,   0.00002,     0.082 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   1,   0,   0 },    148.41736,   0.00002,     0.069 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   1,   0,  -2 },    273.96530,   0.00001,     0.085 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   1,   0,   0 },    273.96530,   0.00001,     0.066 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   2,   0,  -1 },     56.28506,   0.00005,     0.071 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   1,   0,  -2 },    303.04347,   0.00001,     0.089 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   1,   0,   0 },    303.04347,   0.00001,     0.064 },
 { {   0,   0,   8, -19,   0,   0,   0,   0,  -2,   0,   1 },    286.85088,   0.00003,     0.075 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,  -1,   0,   0 },    231.25942,   0.00003,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   1 },    242.58701,   0.00003,     0.089 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,  -1 },    234.90043,   0.00001,     0.038 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   1 },    234.90043,   0.00001,     5.222 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,  -1 },     52.02035,   0.00004,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   1 },     52.02035,   0.00004,     0.074 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,  -1 },    234.90043,   0.00001,     7.041 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   1 },    234.90043,   0.00001,     0.037 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,  -1 },    242.31946,   0.00002,     0.088 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,   1,   0,   0 },    163.49481,   0.00002,     0.074 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   1,   0,   0 },     68.70866,   0.00002,     0.071 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,   2,   0,  -1 },    149.25672,   0.00004,     0.077 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,  -3 },      6.36668,   0.00001,     0.075 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,   0,  -1 },      5.07385,   0.00004,     0.074 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   1,   0,  -2 },     59.34192,   0.00002,     0.085 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   1,   0,   0 },     59.34192,   0.00002,     0.066 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,  -1 },     27.31373,   0.00006,     0.071 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,  -1 },     55.95398,   0.00002,     0.069 },
 { {   0,   0,   9, -21,   0,   0,   0,   0,  -2,   0,   1 },    264.84859,   0.00001,     0.074 },
 { {   0,   0,   9, -21,   0,   0,   0,   0,  -2,   0,   3 },     84.84859,   0.00001,     0.075 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,  -2 },    300.58258,   0.00001,     0.074 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   1,   0,   0 },    187.49709,   0.00010,     0.075 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   1,   0,  -2 },    152.27738,   0.00002,     0.077 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   1,   0,   0 },    332.27738,   0.00002,     0.072 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,   0,  -1 },    120.27072,   0.00002,     0.078 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,   0,  -1 },    179.35949,   0.00003,     0.075 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,  -1 },    358.23925,   0.00005,     0.072 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,   2,   0,  -1 },     27.14518,   0.00003,     0.069 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   1,   0,  -2 },    151.80196,   0.00003,     0.074 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,   1,   0,   0 },    151.80196,   0.00003,     0.075 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,  -1 },    300.65275,   0.00003,     0.075 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,   0,  -1 },    329.15213,   0.00004,     0.072 },
 { {   0,   0,  10, -13,   0,   0,   0,   0,   2,   0,  -1 },    358.15625,   0.00003,     0.069 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,   0,  -3 },    270.92702,   0.00001,     0.074 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,   2,   0,  -1 },     91.02945,   0.00005,     0.075 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,  -3 },    120.15424,   0.00001,     0.077 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,  -1 },    300.15424,   0.00001,     0.072 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,  -1 },    190.68252,   0.00001,     0.074 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -2,   1 },    190.68252,   0.00001,     0.075 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,  -1,   0 },    145.60826,   0.00001,     0.074 },
 { {   0,   0,  34, -41,   0,   0,   0,   0,  -1,  -1,   2 },    145.60826,   0.00001,     0.075 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    341.06990,   0.00002,     0.060 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    339.26429,   0.00002,     0.068 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   3 },    339.26429,   0.00002,     0.082 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },    270.21891,   0.00003,     0.073 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   3 },     90.21891,   0.00003,     0.076 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     12.48068,   0.00006,     0.068 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   2 },    240.17034,   0.00001,     0.083 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0,  -2 },      6.67687,   0.00002,     0.059 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },      6.67687,   0.00002,     0.100 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    301.36251,   0.00008,     0.073 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   2 },    262.06705,   0.00004,     0.076 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,  -3 },    109.32219,   0.00001,     0.059 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    109.32219,   0.00001,     0.101 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },    305.17513,   0.00004,     0.086 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },     83.11622,   0.00005,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.87295,   0.00001,     0.038 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   0 },    357.18813,   0.00003,     1.951 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,   0 },    267.66778,   0.00003,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    246.36075,   0.00001,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    246.36075,   0.00001,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.11554,   0.00003,     0.793 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   2 },    180.11554,   0.00003,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },     15.18784,   0.00002,     0.067 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    199.89038,   0.00010,     0.083 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   2 },      8.41047,   0.00004,     0.026 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },    252.36941,   0.00002,     0.094 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },     85.39922,   0.00002,     0.085 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   2 },    192.34206,   0.00002,     0.066 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1 },     37.30805,   0.00001,     0.078 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1 },     37.30805,   0.00001,     0.071 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -2 },    104.82745,   0.00002,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    104.82745,   0.00002,     0.077 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },    341.82985,   0.00007,     0.093 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },    117.92618,   0.00007,     0.085 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 },    117.92618,   0.00007,     0.066 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    158.11077,   0.00001,     0.080 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   2 },    158.11077,   0.00001,     0.070 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    338.11077,   0.00003,     1.354 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -2 },     89.48321,   0.00002,     0.078 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    202.69423,   0.00019,     0.071 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -3 },    239.75888,   0.00002,     0.072 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },    342.20332,   0.00001,     0.077 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },    254.64264,   0.00005,     0.103 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   2 },      2.60660,   0.00002,     0.058 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,  -2 },    346.46820,   0.00002,     0.084 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    239.52536,   0.00002,     0.067 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },     66.42887,   0.00002,     0.072 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   1,   0,   0 },    158.74254,   0.00005,     0.063 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },    309.00828,   0.00004,     0.067 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   0,  -1 },     20.62169,   0.00002,     0.063 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },    341.11980,   0.00001,     0.062 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   0,   3 },    161.11980,   0.00001,     0.093 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    256.44748,   0.00004,     0.066 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    302.86711,   0.00005,     0.071 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    112.42360,   0.00002,     0.077 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   3 },    112.42360,   0.00002,     0.073 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,  -1 },    167.17598,   0.00003,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   2,   1 },    302.78975,   0.00002,     0.074 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },     47.36714,   0.00007,     0.071 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0,  -2 },    165.22731,   0.00002,     0.062 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },     19.06682,   0.00003,     0.094 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    136.78260,   0.00007,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0,   2 },     87.32884,   0.00001,     0.073 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,  -2 },    186.11399,   0.00001,     0.074 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   2,   0 },    100.40238,   0.00003,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    285.47156,   0.00003,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    285.47156,   0.00003,     0.079 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,  -3 },    289.37209,   0.00001,     0.061 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    109.37209,   0.00001,     0.095 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },     23.26965,   0.00002,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   1 },     34.58656,   0.00014,     0.090 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,  -1 },     12.78501,   0.00002,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   1 },     12.78513,   0.00002,     0.456 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    106.78894,   0.00006,     0.083 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },     17.56995,   0.00003,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },     17.78322,   0.00002,     2.394 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },     48.82838,   0.00009,     0.076 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },     48.82838,   0.00009,     0.073 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,  -1 },     17.78322,   0.00002,    11.864 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   1 },     17.56995,   0.00003,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   0 },      0.01545,   0.00005,     0.072 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.11535,   0.00005,     1.572 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -2 },    334.10870,   0.00001,     0.070 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    194.30034,   0.00014,     0.079 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -3 },    195.14397,   0.00001,     0.065 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },     17.75499,   0.00004,     0.086 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    355.86925,   0.00003,     0.090 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    311.60404,   0.00021,     0.073 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    284.37092,   0.00007,     0.080 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1 },    161.53776,   0.00001,     0.028 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   1 },    161.71334,   0.00002,     0.110 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   3 },    340.07907,   0.00003,     0.056 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,  -2 },    106.37616,   0.00001,     0.027 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    322.46316,   0.00004,     0.099 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   0 },    179.88511,   0.00006,     0.319 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,  -2 },    152.98953,   0.00001,     0.082 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },    220.27956,   0.00009,     0.068 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   2 },     59.20958,   0.00001,     0.024 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    165.31240,   0.00005,     0.074 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,  -2 },    114.55252,   0.00001,     0.089 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    239.28758,   0.00004,     0.064 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   1,   0,   0 },    321.51064,   0.00004,     0.060 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,  -3 },      8.44702,   0.00003,     0.089 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,  -1 },      8.44702,   0.00003,     0.064 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   0,  -1 },     20.61859,   0.00002,     0.060 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,  -3 },    288.01533,   0.00004,     0.108 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,  -1 },    108.01533,   0.00004,     0.057 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    286.95899,   0.00001,     0.069 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0,   2 },    286.95899,   0.00001,     0.081 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    359.88250,   0.00003,     0.080 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   0 },     56.98331,   0.00002,     0.037 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     56.92022,   0.00041,     0.074 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0,   2 },      2.21551,   0.00001,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   0 },    236.98886,   0.00002,     4.230 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },     50.46389,   0.00002,     0.089 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    202.22287,   0.00002,     0.087 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    153.84684,   0.00004,     0.074 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    153.84684,   0.00004,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    198.15991,   0.00002,     0.089 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,  -1 },     28.06134,   0.00002,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   1 },     21.98394,   0.00008,     0.096 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,  -2 },    303.80888,   0.00001,     0.026 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },     91.29083,   0.00006,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,  -1 },      8.45105,   0.00001,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   1 },      8.45105,   0.00001,     0.958 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    359.12453,   0.00002,     0.080 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    359.12453,   0.00002,     0.070 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,  -1 },      8.45105,   0.00001,     1.409 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   1 },      8.45105,   0.00001,     0.036 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    301.81571,   0.00005,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1 },     14.20298,   0.00003,     0.074 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.11405,   0.00004,    95.796 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },     14.20298,   0.00003,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    182.49682,   0.00025,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },     16.34268,   0.00002,     0.082 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },     51.12627,   0.00003,     0.095 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    246.72715,   0.00003,     0.070 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },    358.05863,   0.00004,     0.076 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,  -1 },    161.53757,   0.00001,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   1 },    161.53757,   0.00001,     0.119 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -2 },     92.85820,   0.00002,     0.028 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    215.64094,   0.00006,     0.105 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,  -1 },    171.66753,   0.00004,     0.095 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   1 },    171.66753,   0.00004,     0.061 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -2 },    171.60220,   0.00002,     0.086 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    227.62327,   0.00010,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    125.37021,   0.00003,     0.070 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -2,   0 },    318.89823,   0.00003,     0.074 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   0 },    245.62891,   0.00007,     0.118 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   0,  -3 },    147.18331,   0.00002,     0.094 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },    325.57499,   0.00004,     0.062 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   0,   1 },    161.54758,   0.00006,     0.068 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   0,   1 },     59.56273,   0.00002,     0.072 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },    147.44479,   0.00005,     0.078 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    109.15168,   0.00005,     0.072 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0,   2 },    100.22492,   0.00001,     0.077 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0,   1 },    339.04076,   0.00005,     0.085 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    237.03012,   0.00007,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    129.54714,   0.00003,     0.084 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },     90.92395,   0.00001,     0.084 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -1,   0,   2 },     90.92395,   0.00001,     0.067 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    286.77016,   0.00005,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },    286.77016,   0.00005,     0.072 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },    108.72676,   0.00005,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },     75.66912,   0.00006,     0.092 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    207.76838,   0.00003,     0.084 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    207.76838,   0.00003,     0.067 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -2 },    290.15139,   0.00004,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    352.67276,   0.00037,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    316.24311,   0.00002,     0.101 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    245.05009,   0.00002,     0.067 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -3 },    263.37816,   0.00002,     0.077 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },     76.69643,   0.00006,     0.072 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,  -2 },     97.29334,   0.00002,     0.028 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    253.44398,   0.00004,     0.113 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,  -1 },    356.31610,   0.00001,     0.101 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },    356.31610,   0.00001,     0.059 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    201.21272,   0.00011,     0.063 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,   2 },    345.41604,   0.00002,     0.023 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,  -1 },     94.17629,   0.00003,     0.067 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -1,   0,   0 },    246.37985,   0.00006,     0.127 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   1,   0,   0 },    262.16243,   0.00003,     0.059 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,   0,  -1 },    333.55374,   0.00004,     0.059 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   0,   1 },    241.31053,   0.00003,     0.076 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    101.71877,   0.00001,     0.070 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0,   2 },    281.71877,   0.00001,     0.079 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -2,   0,   1 },    338.00597,   0.00002,     0.082 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    296.12277,   0.00006,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   2 },    158.08552,   0.00002,     0.074 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   1 },     36.89673,   0.00004,     0.089 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   3 },    212.80934,   0.00001,     0.064 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    224.53605,   0.00008,     0.082 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -1,   0,   2 },     11.61186,   0.00003,     0.068 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    338.21217,   0.00003,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    338.21217,   0.00003,     0.074 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     28.55529,   0.00007,     0.080 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    180.51603,   0.00003,     0.089 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },     46.70781,   0.00011,     0.074 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    138.55208,   0.00006,     0.098 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,   2 },    359.94241,   0.00002,     0.060 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,  -1 },    194.19281,   0.00006,     0.089 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   1 },    193.55447,   0.00004,     0.064 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,  -2 },    155.05681,   0.00002,     0.081 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    350.54565,   0.00012,     0.069 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,  -1 },    198.44804,   0.00001,     0.077 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -2,   1 },     18.44804,   0.00001,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -1 },    195.58051,   0.00003,     5.352 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -3 },    341.33544,   0.00002,     0.075 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,  -1 },     15.60353,   0.00057,     0.074 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -1 },     15.57659,   0.00003,     0.037 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    296.78525,   0.00002,     0.108 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    348.39563,   0.00006,     0.065 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    106.25825,   0.00003,     0.069 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,  -1 },    356.85170,   0.00001,     0.108 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0,   1 },    356.85170,   0.00001,     0.057 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,   0 },    176.01178,   0.00001,     0.309 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,   2 },    356.01178,   0.00001,     0.033 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    178.06791,   0.00006,     0.061 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,  -1 },    358.80709,   0.00002,     0.065 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -1,   0,   0 },    247.29410,   0.00005,     0.138 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   1,   0,   0 },    261.05772,   0.00004,     0.057 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   0,  -3 },    338.48825,   0.00002,     0.106 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    338.48825,   0.00002,     0.057 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   1 },    292.10514,   0.00001,     0.074 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   3 },    195.25208,   0.00002,     0.075 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   1 },     76.41008,   0.00002,     0.080 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   0,   3 },    256.41008,   0.00002,     0.070 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    210.04133,   0.00005,     0.086 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -1,   0,   2 },    200.69440,   0.00001,     0.066 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    350.98531,   0.00002,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    126.34587,   0.00005,     0.094 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,  -1 },     14.81053,   0.00002,     5.330 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   1 },     14.81594,   0.00002,     0.038 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,  -2 },    192.83451,   0.00002,     0.027 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },     74.47543,   0.00003,     0.104 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,  -1 },     38.85257,   0.00002,     0.094 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   1 },     38.85257,   0.00002,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    335.64739,   0.00001,     0.066 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,   0,   2 },    299.92180,   0.00001,     0.024 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -1,   0,   0 },    276.39064,   0.00003,     0.116 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    350.37576,   0.00006,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   1 },    162.75996,   0.00003,     0.152 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   3 },    342.75996,   0.00003,     0.049 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,  -2 },    275.97201,   0.00002,     0.029 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -1,   0,   0 },    275.97201,   0.00002,     0.131 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,  -1 },    299.36815,   0.00001,     0.115 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   0,   1 },    299.36815,   0.00001,     0.055 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,  -1,   0 },    176.00731,   0.00002,     0.259 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    124.72593,   0.00003,     0.058 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    341.04169,   0.00002,     0.062 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    228.46448,   0.00007,     0.055 },
 { {   0,   7, -14,   0,   0,   0,   0,   0,  -2,   0,   1 },     60.71177,   0.00003,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    129.66650,   0.00006,     0.100 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -3 },     86.90434,   0.00001,     0.076 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,  -1 },     86.90434,   0.00001,     0.073 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },     51.99043,   0.00004,     0.111 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,  -2 },    215.06507,   0.00002,     0.090 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },     69.24325,   0.00003,     0.064 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   0 },    305.25940,   0.00005,     0.125 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -1,   0,   2 },    108.21481,   0.00003,     0.053 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    286.90215,   0.00005,     0.060 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   1 },    162.78735,   0.00002,     0.168 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   3 },    162.78735,   0.00002,     0.048 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    133.48609,   0.00003,     0.056 },
 { {   0,   7,  -5,   0,   0,   0,   0,   0,   1,   0,   0 },    238.67023,   0.00003,     0.053 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,   0,  -3 },    108.00704,   0.00002,     0.163 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,   0,  -1 },    108.00704,   0.00002,     0.048 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0,   1 },     43.42682,   0.00002,     0.075 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0,   3 },    223.66690,   0.00001,     0.074 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   1 },    341.64235,   0.00005,     0.081 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,  -1 },     94.87458,   0.00001,     0.026 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   1 },     93.18476,   0.00007,     0.088 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,  -1 },     87.10337,   0.00003,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   1 },     81.68583,   0.00002,     5.850 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,  -1 },     82.04802,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   1 },     82.04802,   0.00003,     0.074 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,  -1 },     81.68583,   0.00002,     6.151 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   1 },     87.10337,   0.00003,     0.037 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },     92.75224,   0.00007,     0.088 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   1 },     92.75224,   0.00002,     0.026 },
 { {   0,   8, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     26.06739,   0.00013,     0.075 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    220.33590,   0.00002,     0.096 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    306.51920,   0.00003,     0.070 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,  -3 },      8.67488,   0.00001,     0.074 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },      8.67488,   0.00001,     0.075 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    166.89719,   0.00004,     0.107 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   1,   0,   0 },     91.95159,   0.00006,     0.065 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,  -3 },    109.48733,   0.00003,     0.080 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    289.48733,   0.00003,     0.070 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,  -2 },     93.76917,   0.00001,     0.028 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     37.96173,   0.00003,     0.120 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     82.87429,   0.00001,     0.061 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   2 },      0.08183,   0.00003,     0.023 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    313.53217,   0.00004,     0.136 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -1,   0,   2 },    276.00405,   0.00002,     0.051 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   0 },    123.78089,   0.00004,     0.054 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   1,   0,   0 },    227.80786,   0.00002,     0.052 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   0,   1 },    161.51139,   0.00003,     0.086 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   0,   1 },    250.18131,   0.00007,     0.094 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   0,  -1 },     70.15019,   0.00001,     0.078 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,   0,   0,   1 },     70.15019,   0.00001,     0.071 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   1,   0,  -2 },    352.41107,   0.00002,     0.084 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    238.99471,   0.00002,     0.067 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },     10.39083,   0.00002,     0.072 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },    130.82205,   0.00005,     0.114 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,  -1,   0,   2 },    245.59469,   0.00001,     0.055 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,  -2 },    349.21850,   0.00002,     0.092 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    169.21850,   0.00002,     0.063 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,  -3 },    109.76367,   0.00002,     0.084 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    289.76367,   0.00002,     0.067 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },     70.58259,   0.00005,     0.129 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    180.06941,   0.00005,     0.059 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,  -3 },     18.36241,   0.00003,     0.092 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,  -1 },    198.36241,   0.00003,     0.063 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,  -2 },    207.58176,   0.00001,     0.113 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },     82.24167,   0.00006,     0.056 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -1,   0,   0 },    330.28261,   0.00004,     0.174 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0,  -1 },    336.71918,   0.00002,     0.147 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0,   1 },    336.71918,   0.00002,     0.050 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,  -2 },     70.97515,   0.00002,     0.127 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   1,   0,   0 },    143.75345,   0.00002,     0.053 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   1 },    250.14821,   0.00002,     0.099 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   3 },    250.14821,   0.00002,     0.060 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    193.81135,   0.00001,     0.110 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -1,   0,   2 },    193.81135,   0.00001,     0.056 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    172.04328,   0.00004,     0.064 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    118.03576,   0.00003,     0.123 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    212.47124,   0.00004,     0.060 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,   0 },     96.88089,   0.00003,     0.141 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    196.51012,   0.00002,     0.057 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,  -1 },    198.36714,   0.00004,     0.060 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   1,   0,   0 },     52.35202,   0.00003,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   0,   1 },    161.49164,   0.00001,     0.246 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   0,   3 },    161.49164,   0.00001,     0.044 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,  -2 },    329.64858,   0.00002,     0.031 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },    329.64858,   0.00002,     0.196 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   0 },    224.10279,   0.00002,     0.051 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   1,   0,   2 },    111.03375,   0.00002,     0.022 },
 { {   0,  10,  -5,   0,   0,   0,   0,   0,   2,   0,  -1 },    289.74392,   0.00002,     0.044 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     60.66272,   0.00001,     0.026 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   1 },     60.66272,   0.00002,     0.088 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,  -1 },     60.66272,   0.00001,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,  -1 },     60.66272,   0.00007,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,  -1 },     60.19125,   0.00064,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   1 },     60.19125,   0.00064,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   1 },     60.66272,   0.00007,     0.037 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   1 },     60.66272,   0.00001,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,  -1 },     60.66272,   0.00002,     0.088 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   1 },     60.66272,   0.00001,     0.026 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   0,   1 },     70.11500,   0.00003,     0.106 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    212.07367,   0.00003,     0.058 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    104.39021,   0.00003,     0.154 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   1,  -2 },     21.82927,   0.00001,     0.050 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.82927,   0.00001,     0.148 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,  -1 },    198.36728,   0.00003,     0.058 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   1,   0,   0 },     51.88458,   0.00004,     0.052 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   0 },    274.27847,   0.00002,     0.223 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,   2 },     94.27847,   0.00002,     0.045 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.25525,   0.00002,     0.055 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.25525,   0.00002,     0.114 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,  -2 },    251.06487,   0.00001,     0.152 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   1,   0,   0 },    251.06487,   0.00001,     0.049 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    108.26053,   0.00002,     0.171 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   0,   2 },    108.26053,   0.00002,     0.048 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.81270,   0.00003,     0.135 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    124.36507,   0.00004,     0.053 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,  -1 },    198.34662,   0.00002,     0.056 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    292.60196,   0.00002,     0.206 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,  -1,   0,   2 },    292.60196,   0.00002,     0.045 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   1,   0,   0 },     24.02760,   0.00004,     0.050 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.24076,   0.00002,     0.057 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.24076,   0.00002,     0.106 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,   1,   0,   0 },    318.84943,   0.00003,     0.048 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,  -1 },    321.11523,   0.00005,     0.075 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   1,   1 },    321.11523,   0.00005,     0.074 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,   1,   0,   0 },     91.52526,   0.00005,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },     49.71215,   0.00004,     0.191 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.79408,   0.00002,     0.125 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    117.96518,   0.00004,     0.051 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,   0 },    328.07785,   0.00004,     0.309 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.22420,   0.00002,     0.059 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.22420,   0.00002,     0.100 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,   1,   0,   0 },    347.82427,   0.00005,     0.047 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   1,   0,  -2 },     91.55333,   0.00002,     0.130 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,   1,   0,   0 },     91.55333,   0.00002,     0.052 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },     72.75218,   0.00002,     0.217 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   0,   2 },     87.70592,   0.00001,     0.045 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.77342,   0.00002,     0.116 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,  -2 },    117.81389,   0.00001,     0.149 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    117.81389,   0.00001,     0.050 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,  -2 },     21.20557,   0.00002,     0.034 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   0,   0 },    201.20557,   0.00002,     0.383 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.20557,   0.00002,     0.062 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.20557,   0.00002,     0.094 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,  -2 },    112.94670,   0.00001,     0.212 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,   1,   0,   0 },    323.54292,   0.00003,     0.045 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },    113.32768,   0.00002,     0.201 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,  -2 },    201.75075,   0.00001,     0.032 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   0,   0 },     21.75075,   0.00001,     0.251 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.75075,   0.00002,     0.108 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    267.45382,   0.00002,     0.046 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,   0 },    201.18491,   0.00004,     0.504 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.18491,   0.00002,     0.064 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   1,   0 },     21.18491,   0.00002,     0.089 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   1,   0,   0 },    289.81896,   0.00004,     0.044 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,  -1 },    108.32836,   0.00002,     0.074 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   1,   1 },    108.32836,   0.00002,     0.075 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -1,   0,   0 },     21.72611,   0.00003,     0.298 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    359.92520,   0.00005,     0.047 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,  -1,   1 },    342.83188,   0.00002,     0.074 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,  -1,   3 },    162.83188,   0.00002,     0.075 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -1,   0,   0 },    201.16225,   0.00004,     0.735 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,  -2 },    289.63805,   0.00002,     0.288 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   1,   0,   0 },    289.63805,   0.00002,     0.043 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -1,   0,   0 },     21.69954,   0.00002,     0.366 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   1,   0,  -2 },    359.89673,   0.00002,     0.207 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,   1,   0,   0 },    359.89673,   0.00002,     0.045 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },    201.13761,   0.00002,     0.071 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -1,   2 },    201.13761,   0.00002,     0.078 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   1,  -2 },    146.51981,   0.00002,     0.070 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   1,   0 },    146.51981,   0.00002,     0.080 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -1,   0,   0 },     21.67107,   0.00002,     0.474 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   1,   0,   0 },    175.79721,   0.00002,     0.041 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   1 },    217.50243,   0.00001,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,  -1 },    217.50243,   0.00002,     0.097 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   1 },    217.50243,   0.00001,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,  -1 },    114.56550,   0.00004,     0.019 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   1 },    114.56550,   0.00004,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,  -1 },    114.56550,   0.00022,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   1 },    114.56550,   0.00038,     0.088 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,  -1 },    114.56550,   0.00025,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,  -1 },    114.56550,   0.00011,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   1 },    114.56550,   0.00002,     0.076 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,  -1 },    114.55972,   0.00125,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -3 },    294.56550,   0.00001,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,  -1 },    114.50884,   0.01126,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   1 },    114.62374,   0.01123,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   3 },    294.56550,   0.00001,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -1 },    290.17387,   0.00003,     6.132 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   1 },    114.67748,   0.00121,     0.037 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,  -1 },    114.56550,   0.00002,     0.076 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   1 },    114.56550,   0.00011,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   1 },    114.56550,   0.00025,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,  -1 },    114.56550,   0.00038,     0.088 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   1 },    114.56550,   0.00022,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,  -1 },    114.56550,   0.00004,     0.041 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   1 },    114.56550,   0.00004,     0.019 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,  -1 },     11.62857,   0.00001,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   1 },     11.62857,   0.00002,     0.097 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,  -1 },     11.62857,   0.00001,     0.042 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.64074,   0.00002,     0.085 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,  -1 },     18.18419,   0.00005,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   1 },     18.18419,   0.00005,     0.074 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.60861,   0.00003,     0.081 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,  -2 },    296.75407,   0.00003,     0.074 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   1,   0 },    317.44238,   0.00005,     0.075 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   1,  -2 },    201.05224,   0.00002,     0.080 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   1,   0 },    201.05224,   0.00002,     0.069 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   0 },     48.15158,   0.00002,     0.074 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,  -1,   2 },    228.15158,   0.00002,     0.075 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,  -2 },    288.90331,   0.00002,     0.074 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,  -1,   0 },    288.90331,   0.00002,     0.075 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   1,   0 },     83.05036,   0.00002,     0.073 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,  -2 },    267.56110,   0.00002,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,  -1,   0 },     87.56110,   0.00002,     0.074 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,  -1 },    238.39622,   0.00004,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -1,   1 },    238.39622,   0.00004,     0.074 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,  -1 },    218.71494,   0.00001,     0.074 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   1 },    218.71494,   0.00001,     0.075 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 1715> g_ELP15
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     90.00000,   0.00003, 99999.999 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    277.41835,   0.00002,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    177.30367,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     60.07894,   0.00004,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },      0.78026,   0.00043,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0 },    234.61213,   0.00011,     0.040 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,  -2,   1,   0 },    232.10129,   0.00005,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },    242.91482,   0.00008,     0.076 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0 },     62.91482,   0.00008,     0.075 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,  -1,   0 },     51.99997,   0.00005,     0.087 },
 { {   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0 },     54.56736,   0.00010,     0.040 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 },     30.69188,   0.00022,     0.041 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,  -2,   1,   0 },     38.28883,   0.00024,     0.088 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },    312.64300,   0.00016,     0.076 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 },    132.02726,   0.00016,     0.075 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0 },    218.66027,   0.00023,     0.086 },
 { {   0,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0 },    210.72265,   0.00019,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   0,   0 },     42.64159,   0.00014,     0.040 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   1,   0 },      5.31682,   0.00024,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,  -2,   2,   0 },    169.38529,   0.00002,     0.564 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -2,   0 },    169.69306,   0.00006,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,  -1,   0 },    169.08218,   0.00120,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   1,   0 },    349.08218,   0.00120,     0.075 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   0,   2,   0 },    349.69343,   0.00006,     0.038 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -2,   0 },    349.38460,   0.00002,     0.563 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,  -1,   0 },    185.18801,   0.00025,     0.087 },
 { {   0,   0,   0,   0,   2,  -5,   0,   0,   2,   0,   0 },    221.32288,   0.00014,     0.040 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 },     45.92589,   0.00004,     0.041 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,  -1,   0 },    326.49931,   0.00005,     0.076 },
 { {   0,   0,   0,   0,   2,   0,   0,   0,   0,   1,   0 },    146.49931,   0.00005,     0.074 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,  -2,   0 },     32.93942,   0.00002,     0.020 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0 },    213.43884,   0.00016,     0.041 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0 },    209.49782,   0.00002,     0.434 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0 },    215.91783,   0.00007,     0.079 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },    217.85540,   0.00014,     1.881 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0 },     35.91783,   0.00007,     0.073 },
 { {   0,   0,   0,   1,   0,   0,   0,   0,   0,   2,   0 },     37.84227,   0.00002,     0.037 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0 },    155.59263,   0.00003,     0.042 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0 },     61.85396,   0.00003,     0.039 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0 },    237.25478,   0.00004,     0.082 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },    241.85396,   0.00003,     0.940 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   0,   1,   0 },     57.25478,   0.00004,     0.070 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,  -1,  -1 },    125.53110,   0.00002,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   1,   1,  -1 },    305.53110,   0.00002,     0.075 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -3,   0 },    235.08924,   0.00003,     0.072 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -2,   0 },     59.45956,   0.00004,     1.407 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,  -1,   0 },     55.08924,   0.00003,     0.080 },
 { {   0,   0,   0,   2,   0,   0,   0,   0,   2,   0,   0 },     59.45956,   0.00004,     0.039 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   1,   0 },     90.95965,   0.00002,     0.101 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,  -2,   3,   0 },     90.95965,   0.00002,     0.060 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0 },     90.76455,   0.00002,     0.086 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -3,   0 },    120.70688,   0.00002,     0.074 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,  -1,   0 },     87.44760,   0.00020,     0.076 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   2,   0,   0 },    272.90439,   0.00008,     0.038 },
 { {   0,   0,   0,   3,   0,   0,   0,   0,   4,  -3,   0 },     90.76455,   0.00002,     0.088 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,  -1,  -2 },    132.42850,   0.00003,     0.076 },
 { {   0,   0,   0,   4,   0,   0,   0,   0,   2,   1,  -2 },    312.42850,   0.00003,     0.075 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,  -1,   0 },    341.39928,   0.00003,     0.094 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0 },    161.35046,   0.00005,     0.376 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   0,   1,   0 },    161.39928,   0.00003,     0.063 },
 { {   0,   0,   0,   5,   0,   0,   0,   0,   2,   0,   0 },    341.39920,   0.00015,     0.037 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,   0,   0 },    190.46229,   0.00002,     0.313 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   0,   2,   0 },    190.46229,   0.00002,     0.034 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,  -1,   0 },      9.13843,   0.00007,     0.068 },
 { {   0,   0,   0,   6,   0,   0,   0,   0,   2,   0,   0 },    189.13714,   0.00007,     0.036 },
 { {   0,   0,   0,   8,   0,   0,   0,   0,   2,  -1,   0 },    248.39798,   0.00003,     0.064 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,  -2,   0,   0 },    350.12583,   0.00006,     0.036 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   0 },    349.46162,   0.00002,     0.034 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,  -1,   0 },    349.46790,   0.00003,     0.063 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,   0,   0 },    349.46162,   0.00002,     0.367 },
 { {   0,   0,   1,  -7,   0,   0,   0,   0,   0,   1,   0 },    169.46790,   0.00003,     0.095 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,  -1,   0 },     18.49967,   0.00003,     0.065 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,   0,   0 },     18.51420,   0.00005,     0.457 },
 { {   0,   0,   1,  -6,   0,   0,   0,   0,   0,   1,   0 },    198.49967,   0.00003,     0.090 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   1,   0 },     61.57506,   0.00009,     0.076 },
 { {   0,   0,   1,  -5,   0,   0,   0,   0,  -2,   3,   0 },     34.54525,   0.00004,     0.075 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   0,   0 },    264.84926,   0.00010,     0.039 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   1,   0 },    274.46906,   0.00004,     0.079 },
 { {   0,   0,   1,  -4,   0,   0,   0,   0,  -2,   2,   0 },    103.19375,   0.00002,     1.545 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   0,   0 },    120.58601,   0.00006,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   1,   0 },    125.19785,   0.00006,     0.083 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,  -2,   2,   0 },    312.09402,   0.00002,     0.848 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0 },    118.24731,   0.00010,     0.072 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0 },    298.61954,   0.00002,     1.681 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0 },    298.24731,   0.00010,     0.079 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0 },    118.61954,   0.00002,     0.039 },
 { {   0,   0,   1,  -3,   0,   0,   0,   0,   2,   0,   0 },    298.31785,   0.00025,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   0,   0 },    145.30624,   0.00024,     0.040 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   1,   0 },    150.81802,   0.00024,     0.087 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,  -2,   3,   0 },    333.89907,   0.00002,     0.067 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0 },    132.98447,   0.00052,     0.075 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },    148.93394,   0.00009,    15.781 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0 },    312.92929,   0.00052,     0.076 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0 },    310.17203,   0.00002,     0.038 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,  -1,   0 },    328.38867,   0.00028,     0.088 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   0,   0 },    322.62834,   0.00022,     0.041 },
 { {   0,   0,   1,  -2,   0,   0,   0,   0,   2,   2,   0 },    147.06940,   0.00002,     0.020 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   0 },    183.70254,   0.00020,     0.041 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,  -2,   1,   0 },    182.81901,   0.00010,     0.091 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0 },      1.86978,   0.00003,     0.038 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0 },      4.86928,   0.00003,     2.135 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,  -1,   0 },    356.92797,   0.00005,     0.084 },
 { {   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0 },    356.36978,   0.00023,     0.040 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   0,   0 },    122.65203,   0.00011,     0.042 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,  -2,   1,   0 },    122.65206,   0.00004,     0.093 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,  -1,   0 },    121.15535,   0.00003,     0.080 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   0,   1,   0 },    301.15535,   0.00003,     0.071 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,  -1,   0 },    303.68524,   0.00003,     0.082 },
 { {   0,   0,   1,   0,  -3,   0,   0,   0,   2,   0,   0 },    303.68524,   0.00003,     0.039 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   0,   0 },    133.91660,   0.00035,     0.042 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,  -2,   1,   0 },    134.21033,   0.00019,     0.094 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   0 },    132.58756,   0.00023,     0.080 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 },    314.04701,   0.00004,     1.203 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   0,   1,   0 },    312.28804,   0.00019,     0.071 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,  -2,   0 },    314.37379,   0.00002,     0.075 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   1,   0,   0 },    314.37379,   0.00002,     0.076 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,  -1,   0 },    312.30519,   0.00008,     0.081 },
 { {   0,   0,   1,   0,  -2,   0,   0,   0,   2,   0,   0 },    312.79413,   0.00014,     0.039 },
 { {   0,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   0 },    202.45835,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,  -2,   0 },     33.06280,   0.00002,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    138.15761,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },    219.69711,   0.00009,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0 },     51.91674,   0.00004,     0.360 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,  -1,   1 },    282.41737,   0.00002,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },    111.30072,   0.00005,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    144.35924,   0.00004,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   1 },    102.41737,   0.00002,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    208.03499,   0.00009,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     28.81897,   0.00010,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     28.03499,   0.00009,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0 },    151.42896,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -3,   1 },    263.82465,   0.00002,     0.077 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,  -1 },    217.39075,   0.00008,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   1 },      6.41372,   0.00005,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   1 },    232.92205,   0.00005,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -1 },     37.39075,   0.00008,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    106.74943,   0.00002,     0.071 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0 },     85.59562,   0.00002,     1.292 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    253.34103,   0.00003,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    265.59562,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   1,  -5,   0,   0,  -2,   0,   0 },    198.35193,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   0 },    263.94446,   0.00006,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   0,   0 },    254.34026,   0.00020,     0.042 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,  -2,   1,   0 },    254.34026,   0.00007,     0.095 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,  -1,   0 },    253.60736,   0.00007,     0.082 },
 { {   0,   0,   1,   0,   2,  -5,   0,   0,   0,   1,   0 },     73.60736,   0.00007,     0.070 },
 { {   0,   0,   1,   0,   4,   0,   0,   0,   2,  -3,   0 },    157.72502,   0.00002,     0.073 },
 { {   0,   0,   1,   0,   4,   0,   0,   0,   2,  -1,   0 },    337.72502,   0.00002,     0.078 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0 },     61.79812,   0.00004,     0.085 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0 },    246.08904,   0.00010,     0.653 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   0,   1,   0 },    241.79812,   0.00004,     0.068 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -3,   0 },     37.81595,   0.00002,     0.074 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,  -1,   0 },     76.05857,   0.00007,     0.077 },
 { {   0,   0,   1,   1,   0,   0,   0,   0,   2,   0,   0 },    231.57215,   0.00006,     0.038 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,  -3,   0 },    272.74814,   0.00002,     0.077 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,  -1,   0 },    269.67612,   0.00010,     0.074 },
 { {   0,   0,   1,   2,   0,   0,   0,   0,   2,   0,   0 },    269.17596,   0.00006,     0.037 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   0,   0,   0 },    312.36003,   0.00003,     0.385 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,  -1,   0 },    309.45560,   0.00006,     0.071 },
 { {   0,   0,   1,   3,   0,   0,   0,   0,   2,   0,   0 },    129.72421,   0.00003,     0.037 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,  -1,   0 },    341.26742,   0.00003,     0.099 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,   0,   0 },    161.24923,   0.00006,     0.320 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   0,   1,   0 },    161.26742,   0.00003,     0.061 },
 { {   0,   0,   1,   4,   0,   0,   0,   0,   2,   0,   0 },    341.31259,   0.00020,     0.036 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   0,   0,   0 },    190.40401,   0.00002,     0.273 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   0,   2,   0 },    190.40401,   0.00002,     0.033 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,  -2,   0 },    188.46457,   0.00004,     0.531 },
 { {   0,   0,   1,   5,   0,   0,   0,   0,   2,   0,   0 },    188.46457,   0.00004,     0.035 },
 { {   0,   0,   2, -11,   0,   0,   0,   0,  -2,   0,   0 },     82.36851,   0.00004,     0.035 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,  -2,   0,   0 },    350.02180,   0.00006,     0.037 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   0 },    349.39286,   0.00002,     0.035 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,  -1,   0 },    349.37664,   0.00003,     0.064 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,   0,   0 },    349.39286,   0.00002,     0.444 },
 { {   0,   0,   2,  -8,   0,   0,   0,   0,   0,   1,   0 },    169.37664,   0.00003,     0.091 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   1,   0 },     25.08304,   0.00015,     0.076 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,  -2,   3,   0 },     13.34152,   0.00003,     0.075 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   0,  -1,   0 },     18.30493,   0.00003,     0.067 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   2,  -2,   0 },     31.07966,   0.00002,     0.286 },
 { {   0,   0,   2,  -7,   0,   0,   0,   0,   2,   0,   0 },     31.07966,   0.00002,     0.043 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   0,   0 },    270.39414,   0.00010,     0.039 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   1,   0 },    270.40285,   0.00010,     0.082 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   2,   0 },    254.58383,   0.00002,     0.896 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -2,   3,   0 },    254.43292,   0.00002,     0.070 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   1,  -1 },    292.98114,   0.00003,     0.076 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,  -1,   3,  -1 },    292.98114,   0.00003,     0.075 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,  -1,   0 },     95.18862,   0.00004,     0.072 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0 },    268.66158,   0.00004,     1.519 },
 { {   0,   0,   2,  -5,   0,   0,   0,   0,   0,   1,   0 },    275.18862,   0.00004,     0.079 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   0,   0 },    121.03711,   0.00015,     0.040 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   1,   0 },    118.34849,   0.00009,     0.086 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,  -2,   3,   0 },    123.35297,   0.00002,     0.067 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   0 },    111.84063,   0.00020,     0.075 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   1,   0 },    292.75097,   0.00024,     0.076 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0 },    118.14203,   0.00003,     0.038 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,  -1,   0 },    299.77484,   0.00016,     0.088 },
 { {   0,   0,   2,  -4,   0,   0,   0,   0,   2,   0,   0 },    297.86641,   0.00037,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   0,   0 },    148.68625,   0.00018,     0.041 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,  -2,   1,   0 },    148.75622,   0.00009,     0.090 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,  -1,   0 },    154.13025,   0.00009,     0.078 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0 },    326.39583,   0.00004,     2.470 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0 },    334.13025,   0.00009,     0.073 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,  -1,   0 },    326.06433,   0.00006,     0.084 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   0,   0 },    326.72449,   0.00026,     0.040 },
 { {   0,   0,   2,  -3,   0,   0,   0,   0,   2,   2,   0 },    146.84984,   0.00003,     0.019 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,  -2,   0,   0 },    180.42221,   0.00018,     0.042 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0 },      2.85782,   0.00003,     0.039 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0 },    354.14315,   0.00004,     0.081 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0 },    182.85782,   0.00003,     1.068 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0 },    174.14315,   0.00004,     0.070 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,  -2,   0 },    176.61972,   0.00004,     1.194 },
 { {   0,   0,   2,  -2,   0,   0,   0,   0,   2,   0,   0 },    356.61268,   0.00008,     0.039 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   0,   0 },    212.77391,   0.00024,     0.043 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,  -2,   2,   0 },    212.55912,   0.00003,     0.308 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -3,   0 },     34.80258,   0.00002,     0.074 },
 { {   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,   0 },    214.80258,   0.00002,     0.077 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,  -2,   0,   0 },    341.45658,   0.00006,     0.043 },
 { {   0,   0,   2,   0,  -4,   0,   0,   0,   2,  -1,   0 },    339.64463,   0.00012,     0.076 },
 { {   0,   0,   2,   0,  -4,   5,   0,   0,   2,  -1,   0 },    164.54674,   0.00004,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   0,   0 },    333.08832,   0.00021,     0.043 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,  -2,   1,   0 },    333.08832,   0.00005,     0.103 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,  -1,   0 },    329.60257,   0.00040,     0.087 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 },    149.87128,   0.00007,     0.572 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   0,   1,   0 },    152.56632,   0.00003,     0.067 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -3,   0 },    150.39733,   0.00006,     0.075 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -2,   0 },    149.31013,   0.00012,    37.133 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,  -1,   0 },    329.33777,   0.00225,     0.076 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   2,   0,   0 },    329.33862,   0.00012,     0.038 },
 { {   0,   0,   2,   0,  -3,   0,   0,   0,   4,  -2,   0 },    329.31238,   0.00004,     0.040 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,  -2,   0,   0 },     40.39487,   0.00006,     0.044 },
 { {   0,   0,   2,   0,  -2,   0,   0,   0,   2,  -1,   0 },    266.54766,   0.00010,     0.075 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,  -2,   0,   0 },     14.91052,   0.00003,     0.044 },
 { {   0,   0,   2,   0,   0,  -5,   0,   0,   2,  -1,   0 },    194.69684,   0.00005,     0.075 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    140.01487,   0.00003,     0.020 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0 },    155.24380,   0.00011,     0.044 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0 },    167.06584,   0.00003,     0.089 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    224.44608,   0.00008,     0.500 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0 },    347.06584,   0.00003,     0.066 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },     85.03114,   0.00002,     0.077 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    272.83613,   0.00015,     0.074 },
 { {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    222.90879,   0.00023,     0.037 },
 { {   0,   0,   2,   0,   3,   0,   0,   0,   2,  -1,   0 },    157.75907,   0.00004,     0.073 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -3,   0 },    268.36184,   0.00003,     0.080 },
 { {   0,   0,   2,   1,   0,   0,   0,   0,   2,  -1,   0 },     88.36184,   0.00003,     0.071 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,  -2,   0,   0 },    118.05557,   0.00002,     0.046 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,  -2,   2,   0 },    118.05557,   0.00002,     0.207 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,  -1,   0 },    311.08981,   0.00012,     0.069 },
 { {   0,   0,   2,   2,   0,   0,   0,   0,   2,   0,   0 },    131.09255,   0.00012,     0.036 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   0,   0,   0 },    160.92767,   0.00003,     0.278 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   2,  -1,   0 },    340.56342,   0.00006,     0.066 },
 { {   0,   0,   2,   3,   0,   0,   0,   0,   2,   0,   0 },    341.16450,   0.00019,     0.035 },
 { {   0,   0,   3, -12,   0,   0,   0,   0,  -2,   0,   0 },     82.22560,   0.00006,     0.036 },
 { {   0,   0,   3, -10,   0,   0,   0,   0,   0,   0,   0 },    140.20435,   0.00003,     0.432 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,  -2,   0,   0 },    357.56385,   0.00006,     0.038 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,  -1,   0 },    349.23133,   0.00003,     0.066 },
 { {   0,   0,   3,  -9,   0,   0,   0,   0,   0,   1,   0 },    169.23133,   0.00003,     0.087 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   1,   0 },     19.70917,   0.00004,     0.079 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,  -2,   3,   0 },    199.70917,   0.00004,     0.073 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,  -1,   0 },     18.23198,   0.00005,     0.069 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   0,   1,   0 },    198.23198,   0.00005,     0.083 },
 { {   0,   0,   3,  -8,   0,   0,   0,   0,   2,   0,   0 },     30.73643,   0.00005,     0.043 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,  -1,   0 },    194.86772,   0.00002,     0.070 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   0,   1,   0 },     14.86772,   0.00002,     0.082 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,  -1,   0 },     14.92789,   0.00008,     0.095 },
 { {   0,   0,   3,  -8,   3,   0,   0,   0,   2,   0,   0 },     14.92789,   0.00022,     0.042 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   0,   0 },    239.76196,   0.00010,     0.039 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -2,   1,   0 },    241.50424,   0.00011,     0.082 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,  -1,   1,  -1 },    273.58024,   0.00005,     0.075 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,  -1,   0 },     48.18261,   0.00003,     0.072 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   0,   1,   0 },    228.18261,   0.00003,     0.080 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -3,   0 },     59.52324,   0.00002,     0.063 },
 { {   0,   0,   3,  -7,   0,   0,   0,   0,   2,  -1,   0 },     59.52324,   0.00002,     0.093 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   0 },     79.81405,   0.00014,     0.074 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0 },    268.46852,   0.00004,     5.260 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   0,   1,   0 },    259.81405,   0.00014,     0.077 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,  -1,   0 },    279.60042,   0.00002,     0.089 },
 { {   0,   0,   3,  -6,   0,   0,   0,   0,   2,   0,   0 },    269.95106,   0.00008,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,  -2,   0,   0 },    120.89323,   0.00017,     0.041 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   0 },    138.08497,   0.00006,     0.077 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,  -2 },    296.57424,   0.00003,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   1,   0 },    318.08497,   0.00006,     0.074 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   0,   3,  -2 },    116.57424,   0.00003,     0.075 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -3,   0 },    296.15440,   0.00003,     0.068 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,  -1,   0 },    297.02083,   0.00007,     0.085 },
 { {   0,   0,   3,  -5,   0,   0,   0,   0,   2,   0,   0 },    297.51520,   0.00030,     0.040 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   0,   0 },    144.72942,   0.00024,     0.042 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,  -2,   1,   0 },    145.35659,   0.00010,     0.094 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0 },    163.74528,   0.00002,     0.039 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0 },    147.48259,   0.00004,     0.081 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0 },    327.48259,   0.00004,     0.071 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   1,   0,   0 },    326.28053,   0.00003,     0.076 },
 { {   0,   0,   3,  -4,   0,   0,   0,   0,   2,   0,   0 },    327.20572,   0.00025,     0.039 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   0,   0 },    183.61272,   0.00019,     0.043 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,  -2,   1,   0 },    183.00443,   0.00004,     0.099 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0 },      2.78273,   0.00003,     0.040 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0 },    182.78273,   0.00003,     0.712 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -3,   0 },    174.60099,   0.00004,     0.073 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,  -1,   0 },    174.60099,   0.00004,     0.078 },
 { {   0,   0,   3,  -3,   0,   0,   0,   0,   2,   0,   0 },    356.12762,   0.00004,     0.038 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,  -2,   0,   0 },    212.66786,   0.00019,     0.044 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0 },     35.58739,   0.00003,     0.088 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0 },    215.58739,   0.00003,     0.066 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,  -1,   0 },    218.10270,   0.00011,     0.075 },
 { {   0,   0,   3,  -2,   0,   0,   0,   0,   2,   0,   0 },     32.17338,   0.00005,     0.037 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0 },    241.36396,   0.00004,     0.405 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -2,   0 },     63.17556,   0.00003,     1.440 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,  -1,   0 },     41.03896,   0.00004,     0.072 },
 { {   0,   0,   3,  -1,   0,   0,   0,   0,   2,   0,   0 },    246.02196,   0.00013,     0.037 },
 { {   0,   0,   3,   0,  -2,   0,   0,   0,   3,  -1,  -1 },    284.63191,   0.00007,     0.075 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0 },    111.94975,   0.00003,     0.098 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    291.94975,   0.00006,     0.333 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   0,   1,   0 },    291.94975,   0.00003,     0.062 },
 { {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    229.83213,   0.00028,     0.036 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0 },    131.98969,   0.00002,     0.283 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   0,   2,   0 },    131.98969,   0.00002,     0.033 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   2,  -1,   0 },    130.19671,   0.00007,     0.067 },
 { {   0,   0,   3,   1,   0,   0,   0,   0,   2,   0,   0 },    130.31514,   0.00006,     0.035 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,  -2,   0 },    342.19107,   0.00003,     0.437 },
 { {   0,   0,   3,   2,   0,   0,   0,   0,   2,   0,   0 },    162.19107,   0.00003,     0.035 },
 { {   0,   0,   3,   4,   0,   0,   0,   0,   2,  -1,   0 },    219.15153,   0.00004,     0.060 },
 { {   0,   0,   3,   5,   0,   0,   0,   0,   2,   0,   0 },     68.03909,   0.00004,     0.033 },
 { {   0,   0,   4, -13,   0,   0,   0,   0,  -2,   0,   0 },     82.16306,   0.00003,     0.036 },
 { {   0,   0,   4, -13,   0,   0,   0,   0,  -2,   2,   0 },    262.16306,   0.00003,     0.879 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,  -2,   1,   0 },    330.98771,   0.00005,     0.075 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,   0,  -1,   0 },    320.17678,   0.00002,     0.066 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,   0,   0,   0 },    140.13026,   0.00003,     0.541 },
 { {   0,   0,   4, -11,   0,   0,   0,   0,   0,   1,   0 },    140.17678,   0.00002,     0.088 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,  -2,   0,   0 },      1.82178,   0.00004,     0.038 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,  -1,   0 },    349.11509,   0.00004,     0.069 },
 { {   0,   0,   4, -10,   0,   0,   0,   0,   0,   1,   0 },    169.11509,   0.00004,     0.084 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,  -2,   0 },    210.50641,   0.00003,     0.391 },
 { {   0,   0,   4,  -9,   0,   0,   0,   0,   2,   0,   0 },     30.50641,   0.00003,     0.042 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   0,   0 },    237.92978,   0.00012,     0.040 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   1,   0 },    234.06782,   0.00003,     0.085 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,  -2,   3,   0 },    234.06782,   0.00003,     0.068 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,  -1,   0 },     62.13010,   0.00008,     0.074 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   0,   1,   0 },    242.13010,   0.00008,     0.077 },
 { {   0,   0,   4,  -8,   0,   0,   0,   0,   2,   0,   0 },    241.46506,   0.00004,     0.041 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -4,   1,   0 },     91.93148,   0.00004,     0.028 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,  -1,   0 },     91.93148,   0.00015,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   0,   0 },     91.97023,   0.00134,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   1,   0 },     92.13235,   0.00131,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,  -2,   2,   0 },    272.44298,   0.00006,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -2,   0 },     91.35393,   0.00011,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,  -1,   0 },     91.29765,   0.00178,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   1,   0 },    271.29785,   0.00178,     0.075 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   0,   2,   0 },    271.35396,   0.00011,     0.038 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -2,   0 },     92.44441,   0.00006,     0.564 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,  -1,   0 },    272.13364,   0.00132,     0.087 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   0,   0 },    271.97125,   0.00134,     0.040 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   2,   1,   0 },    271.93267,   0.00015,     0.026 },
 { {   0,   0,   4,  -8,   3,   0,   0,   0,   4,  -1,   0 },    271.93267,   0.00004,     0.028 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   0,   0 },     88.70257,   0.00005,     0.041 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,  -2,   2,   0 },     88.70257,   0.00005,     0.487 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,  -1,   0 },     88.81699,   0.00004,     0.077 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0,   0 },    268.45556,   0.00002,     3.594 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   1,   0 },    268.81699,   0.00004,     0.074 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   0,   2,   0 },    268.45556,   0.00002,     0.037 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -2,   0 },     88.27910,   0.00003,     0.669 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,  -1,   0 },    268.56366,   0.00006,     0.085 },
 { {   0,   0,   4,  -7,   0,   0,   0,   0,   2,   0,   0 },    268.27910,   0.00003,     0.040 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,  -2,   0,   0 },    118.49179,   0.00018,     0.042 },
 { {   0,   0,   4,  -6,   0,   0,   0,   0,   2,   0,   0 },    297.58368,   0.00028,     0.039 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0 },    156.05551,   0.00005,     0.745 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0 },    147.49549,   0.00002,     0.036 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,  -3,   0 },    329.20033,   0.00004,     0.073 },
 { {   0,   0,   4,  -5,   0,   0,   0,   0,   2,   0,   0 },    326.85308,   0.00018,     0.038 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,  -2,   0,   0 },    182.04215,   0.00016,     0.044 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0 },     25.98026,   0.00002,     0.041 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0 },    205.98026,   0.00002,     0.534 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -3,   0 },    201.31670,   0.00002,     0.076 },
 { {   0,   0,   4,  -4,   0,   0,   0,   0,   2,  -1,   0 },    177.58080,   0.00009,     0.075 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0 },     33.00236,   0.00002,     0.041 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,  -1,   0 },     32.53869,   0.00003,     0.092 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0 },    225.90963,   0.00005,     0.416 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0 },    212.53869,   0.00003,     0.064 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0 },     57.71024,   0.00002,     0.035 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,  -2,   0 },     19.47975,   0.00003,     1.585 },
 { {   0,   0,   4,  -3,   0,   0,   0,   0,   2,   0,   0 },    199.47975,   0.00003,     0.037 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0 },    240.99347,   0.00003,     0.341 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -3,   0 },    247.99473,   0.00003,     0.083 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,  -1,   0 },     61.40767,   0.00007,     0.069 },
 { {   0,   0,   4,  -2,   0,   0,   0,   0,   2,   0,   0 },    244.47778,   0.00013,     0.036 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0 },    120.86789,   0.00005,     0.250 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   0,   2,   0 },    120.86789,   0.00005,     0.033 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -1,   0 },    287.54557,   0.00007,     0.065 },
 { {   0,   0,   4,   0,   0,   0,   0,   0,   2,   0,   0 },     84.75748,   0.00011,     0.035 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   0,  -1,   0 },    161.19172,   0.00002,     0.115 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   0,   1,   0 },    341.19172,   0.00002,     0.056 },
 { {   0,   0,   4,   1,   0,   0,   0,   0,   2,   0,   0 },    341.35196,   0.00014,     0.034 },
 { {   0,   0,   5, -14,   0,   0,   0,   0,  -2,   0,   0 },     82.10280,   0.00003,     0.037 },
 { {   0,   0,   5, -14,   0,   0,   0,   0,  -2,   2,   0 },    262.10280,   0.00003,     1.494 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   1,   0 },    286.44235,   0.00002,     0.075 },
 { {   0,   0,   5, -13,   0,   0,   0,   0,  -2,   3,   0 },    286.44235,   0.00002,     0.076 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,   0,  -1,   0 },    320.06752,   0.00003,     0.068 },
 { {   0,   0,   5, -12,   0,   0,   0,   0,   0,   1,   0 },    140.06752,   0.00003,     0.084 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -2,   0,   0 },     11.36250,   0.00002,     0.039 },
 { {   0,   0,   5, -11,   0,   0,   0,   0,  -2,   2,   0 },     11.36250,   0.00002,     1.081 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,  -1,   0 },    214.31783,   0.00002,     0.074 },
 { {   0,   0,   5, -10,   0,   0,   0,   0,   0,   1,   0 },     34.31783,   0.00002,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,  -2,   1,   0 },     56.15346,   0.00005,     0.089 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0 },    235.03912,   0.00007,     0.077 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   0,   1,   0 },     55.03912,   0.00007,     0.074 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,  -1,   0 },    239.12193,   0.00003,     0.085 },
 { {   0,   0,   5,  -9,   0,   0,   0,   0,   2,   0,   0 },    239.47007,   0.00006,     0.040 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,  -2,   0,   0 },     90.68481,   0.00005,     0.042 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -3,   1 },    191.58743,   0.00002,     0.076 },
 { {   0,   0,   5,  -8,   0,   0,   0,   0,   1,  -1,   1 },    141.86679,   0.00004,     0.075 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   0,   0 },    349.65845,   0.00022,     0.042 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,  -2,   1,   0 },    349.65845,   0.00008,     0.095 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,  -1,   0 },    350.72877,   0.00002,     0.082 },
 { {   0,   0,   5,  -8,   3,   0,   0,   0,   0,   1,   0 },    170.72877,   0.00002,     0.070 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   0,   0 },    121.52645,   0.00003,     0.043 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   1,   0 },    305.45451,   0.00006,     0.098 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,  -2,   2,   0 },    301.52645,   0.00003,     0.328 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0 },    298.01340,   0.00006,     0.782 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0 },    117.28791,   0.00002,     0.036 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -2,   0 },    120.98317,   0.00003,     2.017 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,  -1,   0 },    139.45147,   0.00004,     0.078 },
 { {   0,   0,   5,  -7,   0,   0,   0,   0,   2,   0,   0 },    297.29054,   0.00020,     0.038 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,  -2,   0,   0 },    146.37521,   0.00006,     0.044 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0 },    157.55629,   0.00006,     0.087 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0 },    164.14249,   0.00002,     0.553 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -3,   0 },    293.13817,   0.00003,     0.076 },
 { {   0,   0,   5,  -6,   0,   0,   0,   0,   2,  -1,   0 },    157.55760,   0.00035,     0.075 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   0,   0 },    182.10141,   0.00004,     0.045 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,  -2,   2,   0 },      2.10141,   0.00004,     0.243 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0 },     20.49954,   0.00003,     0.041 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0 },    193.38755,   0.00005,     0.427 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0 },      2.98149,   0.00002,     0.035 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -3,   0 },      3.01157,   0.00002,     0.079 },
 { {   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -1,   0 },    183.01157,   0.00002,     0.072 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,  -2,   1,   0 },     32.30607,   0.00005,     0.116 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0,   0 },    212.62444,   0.00007,     0.348 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -3,   0 },     45.14726,   0.00002,     0.082 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,  -1,   0 },     32.34303,   0.00011,     0.070 },
 { {   0,   0,   5,  -4,   0,   0,   0,   0,   2,   0,   0 },    204.70881,   0.00008,     0.036 },
 { {   0,   0,   5,  -3,   0,   0,   0,   0,   2,  -1,   0 },     70.06532,   0.00010,     0.067 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   0,   0 },     87.28692,   0.00002,     0.048 },
 { {   0,   0,   5,  -2,   0,   0,   0,   0,  -2,   2,   0 },    267.28692,   0.00002,     0.175 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0 },    160.94484,   0.00004,     0.200 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -3,   0 },    160.95666,   0.00003,     0.100 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,  -1,   0 },     25.37823,   0.00017,     0.061 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   2,   0,   0 },     14.94755,   0.00045,     0.034 },
 { {   0,   0,   5,   0,   0,   0,   0,   0,   4,   0,   0 },     18.01510,   0.00004,     0.018 },
 { {   0,   0,   6, -16,   0,   0,   0,   0,  -2,   1,   0 },    233.05494,   0.00006,     0.071 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   0,   0 },     82.03847,   0.00003,     0.037 },
 { {   0,   0,   6, -15,   0,   0,   0,   0,  -2,   2,   0 },    262.03847,   0.00003,     4.969 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   1,   0 },    310.75892,   0.00003,     0.077 },
 { {   0,   0,   6, -14,   0,   0,   0,   0,  -2,   3,   0 },    130.75892,   0.00003,     0.074 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,  -1,   0,   0 },    319.90397,   0.00003,     0.075 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,  -1,   0 },    319.90672,   0.00004,     0.071 },
 { {   0,   0,   6, -13,   0,   0,   0,   0,   0,   1,   0 },    139.90672,   0.00004,     0.081 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,  -1,   0 },    171.40000,   0.00002,     0.073 },
 { {   0,   0,   6, -12,   0,   0,   0,   0,   0,   1,   0 },    351.40000,   0.00002,     0.078 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,  -2,   0,   0 },     30.57102,   0.00006,     0.041 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,  -1,   0 },    208.26881,   0.00002,     0.076 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   0,   1,   0 },     28.26881,   0.00002,     0.075 },
 { {   0,   0,   6, -11,   0,   0,   0,   0,   2,   0,   0 },    210.08666,   0.00004,     0.040 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,  -2,   0,   0 },     59.01006,   0.00003,     0.042 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,  -1,   0 },     85.18926,   0.00002,     0.080 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   0,   1,   0 },    265.18926,   0.00002,     0.072 },
 { {   0,   0,   6, -10,   0,   0,   0,   0,   2,  -1,   0 },    240.27224,   0.00003,     0.082 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   0,   0 },    267.63099,   0.00003,     0.043 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,  -2,   2,   0 },     87.63099,   0.00003,     0.335 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -3,   0 },     95.18166,   0.00003,     0.072 },
 { {   0,   0,   6,  -9,   0,   0,   0,   0,   2,  -1,   0 },     95.18166,   0.00003,     0.079 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   0,   0 },    120.88966,   0.00006,     0.043 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,  -2,   2,   0 },    304.01135,   0.00003,     0.284 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,  -1,   0 },    320.90239,   0.00006,     0.087 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0,   0 },    297.54139,   0.00002,     0.573 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   1,   0 },    295.99006,   0.00003,     0.067 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0 },    297.54139,   0.00002,     0.035 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -3,   0 },    333.24107,   0.00002,     0.075 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -2,   0 },    120.03157,   0.00004,    36.376 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,  -1,   0 },    313.69552,   0.00051,     0.076 },
 { {   0,   0,   6,  -8,   0,   0,   0,   0,   2,   0,   0 },    305.35836,   0.00008,     0.038 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   0,   0 },    152.66733,   0.00005,     0.045 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,  -2,   1,   0 },    334.49237,   0.00005,     0.109 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0 },    157.63335,   0.00004,     0.439 },
 { {   0,   0,   6,  -7,   0,   0,   0,   0,   2,   0,   0 },    326.97772,   0.00003,     0.037 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   0,   0 },    183.95935,   0.00002,     0.046 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   1,   0 },      3.42303,   0.00006,     0.115 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,  -2,   2,   0 },      3.95935,   0.00002,     0.218 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0 },     81.40245,   0.00002,     0.042 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0,   0 },    196.34827,   0.00009,     0.356 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -3,   0 },    168.66681,   0.00003,     0.082 },
 { {   0,   0,   6,  -6,   0,   0,   0,   0,   2,  -1,   0 },    168.66681,   0.00003,     0.070 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,  -1,   0 },     23.37705,   0.00006,     0.067 },
 { {   0,   0,   6,  -5,   0,   0,   0,   0,   2,   0,   0 },    212.41554,   0.00010,     0.036 },
 { {   0,   0,   6,  -4,   0,   0,   0,   0,   2,  -1,   0 },     72.19693,   0.00009,     0.065 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,  -3,   0 },    281.22234,   0.00004,     0.094 },
 { {   0,   0,   6,  -3,   0,   0,   0,   0,   2,  -1,   0 },    101.22234,   0.00004,     0.063 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -3,   0 },    311.38822,   0.00002,     0.099 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -2,   0 },    311.32450,   0.00002,     0.316 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,  -1,   0 },    131.38822,   0.00002,     0.061 },
 { {   0,   0,   6,  -2,   0,   0,   0,   0,   2,   0,   0 },    132.43720,   0.00007,     0.034 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   0,   0,   0 },    316.15228,   0.00004,     0.167 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,  -1,   0 },    215.05365,   0.00003,     0.057 },
 { {   0,   0,   6,   0,   0,   0,   0,   0,   2,   0,   0 },    108.96065,   0.00019,     0.033 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   0,   0 },     81.96522,   0.00007,     0.038 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   1,   0 },    263.37128,   0.00003,     0.077 },
 { {   0,   0,   7, -16,   0,   0,   0,   0,  -2,   3,   0 },    263.37128,   0.00003,     0.074 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,  -2,   1,   0 },    111.34914,   0.00004,     0.080 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,  -1,   0 },    290.91086,   0.00002,     0.070 },
 { {   0,   0,   7, -15,   0,   0,   0,   0,   0,   1,   0 },    110.91086,   0.00002,     0.081 },
 { {   0,   0,   7, -14,   0,   0,   0,   0,  -2,   1,   0 },    140.22339,   0.00003,     0.084 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   0,   0 },      3.30880,   0.00006,     0.041 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,  -2,   1,   0 },    153.20184,   0.00003,     0.088 },
 { {   0,   0,   7, -13,   0,   0,   0,   0,   2,   0,   0 },    181.14837,   0.00006,     0.040 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,  -2,   0,   0 },    250.85544,   0.00004,     0.042 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,  -1,   0 },     59.21739,   0.00002,     0.083 },
 { {   0,   0,   7, -11,   0,   0,   0,   0,   0,   1,   0 },    239.21739,   0.00002,     0.069 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,  -1,   0 },     88.43847,   0.00002,     0.086 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   0,   1,   0 },    268.43847,   0.00002,     0.067 },
 { {   0,   0,   7, -10,   0,   0,   0,   0,   2,  -1,   0 },    248.98311,   0.00012,     0.076 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,  -2,   0,   0 },    123.57574,   0.00010,     0.044 },
 { {   0,   0,   7,  -9,   0,   0,   0,   0,   2,   0,   0 },    297.04740,   0.00007,     0.037 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   0,   0 },    158.72999,   0.00006,     0.045 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   1,   0 },    334.39084,   0.00002,     0.114 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,  -2,   3,   0 },    334.39084,   0.00002,     0.056 },
 { {   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0 },    156.86218,   0.00004,     0.364 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,  -2,   0,   0 },    182.52292,   0.00006,     0.047 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -2,   0 },    358.32539,   0.00003,     0.665 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,  -1,   0 },    356.35643,   0.00008,     0.068 },
 { {   0,   0,   7,  -7,   0,   0,   0,   0,   2,   0,   0 },    178.32539,   0.00003,     0.036 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0 },    212.98686,   0.00006,     0.262 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -2,   0 },     36.87128,   0.00002,     0.491 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,  -1,   0 },    210.34814,   0.00003,     0.065 },
 { {   0,   0,   7,  -6,   0,   0,   0,   0,   2,   0,   0 },    216.06697,   0.00014,     0.035 },
 { {   0,   0,   7,  -5,   0,   0,   0,   0,   2,  -1,   0 },     72.43456,   0.00010,     0.063 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   2,  -1,   0 },    101.22821,   0.00005,     0.061 },
 { {   0,   0,   7,  -4,   0,   0,   0,   0,   2,   0,   0 },    101.42138,   0.00004,     0.034 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   2,  -1,   0 },    131.25672,   0.00004,     0.059 },
 { {   0,   0,   7,  -3,   0,   0,   0,   0,   2,   0,   0 },    131.52854,   0.00007,     0.033 },
 { {   0,   0,   8, -17,   0,   0,   0,   0,  -2,   0,   0 },     81.87930,   0.00006,     0.039 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,   0,  -1,   0 },    290.53528,   0.00002,     0.073 },
 { {   0,   0,   8, -16,   0,   0,   0,   0,   0,   1,   0 },    110.53528,   0.00002,     0.078 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   0,   0 },    334.14506,   0.00016,     0.040 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,  -2,   1,   0 },    344.80789,   0.00009,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,  -1,   0 },    325.67652,   0.00025,     0.076 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   0,   1,   0 },    146.44142,   0.00029,     0.075 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,  -1,   0 },    154.18238,   0.00015,     0.087 },
 { {   0,   0,   8, -15,   0,   0,   0,   0,   2,   0,   0 },    152.15763,   0.00018,     0.040 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   1,   0 },    168.00898,   0.00002,     0.092 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,  -2,   3,   0 },    168.00898,   0.00002,     0.064 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,  -1,   0 },    163.30389,   0.00003,     0.079 },
 { {   0,   0,   8, -14,   0,   0,   0,   0,   0,   1,   0 },    343.30389,   0.00003,     0.072 },
 { {   0,   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0 },    200.37241,   0.00009,     0.042 },
 { {   0,   0,   8, -12,   0,   0,   0,   0,  -2,   0,   0 },    246.10373,   0.00005,     0.043 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -3,   0 },     95.50606,   0.00004,     0.078 },
 { {   0,   0,   8, -11,   0,   0,   0,   0,   2,  -1,   0 },    275.50606,   0.00004,     0.073 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,  -2,   0,   0 },    126.72549,   0.00009,     0.045 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   0,   0,   0 },    297.33525,   0.00003,     0.373 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,  -2,   0 },    297.64369,   0.00004,     1.100 },
 { {   0,   0,   8, -10,   0,   0,   0,   0,   2,   0,   0 },    297.64369,   0.00004,     0.036 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,  -2,   0,   0 },    147.58484,   0.00007,     0.046 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0 },    153.87976,   0.00005,     0.311 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -3,   0 },    144.16642,   0.00002,     0.085 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,  -1,   0 },    144.16642,   0.00002,     0.068 },
 { {   0,   0,   8,  -9,   0,   0,   0,   0,   2,   0,   0 },    323.51205,   0.00004,     0.036 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0 },      2.77659,   0.00003,     0.044 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0,   0 },    182.77659,   0.00003,     0.267 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -3,   0 },    335.80897,   0.00002,     0.089 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -2,   0 },    167.57857,   0.00002,     0.507 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,  -1,   0 },    335.80897,   0.00002,     0.066 },
 { {   0,   0,   8,  -8,   0,   0,   0,   0,   2,   0,   0 },    167.57857,   0.00002,     0.035 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,  -2,   0,   0 },    211.42201,   0.00003,     0.049 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0 },    211.47995,   0.00005,     0.234 },
 { {   0,   0,   8,  -7,   0,   0,   0,   0,   2,   0,   0 },    217.23790,   0.00011,     0.034 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -3,   0 },     73.46779,   0.00003,     0.098 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -2,   0 },     71.91424,   0.00006,     0.329 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,  -1,   0 },     73.46779,   0.00003,     0.061 },
 { {   0,   0,   8,  -6,   0,   0,   0,   0,   2,   0,   0 },    250.89433,   0.00003,     0.034 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   2,  -2,   0 },    115.08905,   0.00002,     0.280 },
 { {   0,   0,   8,  -5,   0,   0,   0,   0,   2,   0,   0 },    115.08905,   0.00002,     0.033 },
 { {   0,   0,   8,  -4,   0,   0,   0,   0,   2,   0,   0 },    130.83745,   0.00006,     0.033 },
 { {   0,   0,   9, -19,   0,   0,   0,   0,  -2,   1,   0 },     52.84927,   0.00006,     0.079 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,  -2,   0,   0 },     81.77777,   0.00006,     0.040 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,  -2,   1,   0 },     81.77777,   0.00005,     0.083 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,   0,  -1,   0 },    261.77776,   0.00002,     0.072 },
 { {   0,   0,   9, -18,   0,   0,   0,   0,   0,   1,   0 },     81.77776,   0.00002,     0.079 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,  -2,   0,   0 },    303.33163,   0.00005,     0.040 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,  -1,   0 },    252.12383,   0.00002,     0.075 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   0,   1,   0 },     72.12383,   0.00002,     0.076 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,  -1,   0 },    123.60002,   0.00003,     0.087 },
 { {   0,   0,   9, -17,   0,   0,   0,   0,   2,   0,   0 },    123.60002,   0.00006,     0.040 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   0,   0 },    139.89309,   0.00003,     0.041 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   1,   0 },    139.32434,   0.00002,     0.091 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,  -2,   3,   0 },    139.32434,   0.00002,     0.064 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,  -1,   0 },    139.36747,   0.00004,     0.078 },
 { {   0,   0,   9, -16,   0,   0,   0,   0,   0,   1,   0 },    319.36747,   0.00004,     0.073 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,  -2,   0,   0 },    199.88545,   0.00008,     0.042 },
 { {   0,   0,   9, -15,   0,   0,   0,   0,   2,  -1,   0 },      1.13400,   0.00004,     0.080 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,  -2,   0 },    210.16255,   0.00003,     4.596 },
 { {   0,   0,   9, -14,   0,   0,   0,   0,   2,   0,   0 },     30.16255,   0.00003,     0.038 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,  -2,   0,   0 },    257.86088,   0.00005,     0.044 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -3,   0 },    245.90078,   0.00003,     0.077 },
 { {   0,   0,   9, -13,   0,   0,   0,   0,   2,  -1,   0 },    245.90078,   0.00003,     0.074 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -3,   0 },     88.50631,   0.00004,     0.081 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -2,   0 },     88.73130,   0.00002,     1.182 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,  -1,   0 },     88.50631,   0.00004,     0.071 },
 { {   0,   0,   9, -12,   0,   0,   0,   0,   2,   0,   0 },    268.73130,   0.00002,     0.037 },
 { {   0,   0,   9, -11,   0,   0,   0,   0,  -2,   0,   0 },    125.59173,   0.00006,     0.046 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,  -2,   0,   0 },    152.18383,   0.00011,     0.047 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   0,   0,   0 },    153.74613,   0.00004,     0.272 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -3,   0 },    145.61777,   0.00003,     0.088 },
 { {   0,   0,   9, -10,   0,   0,   0,   0,   2,  -1,   0 },    325.61777,   0.00003,     0.066 },
 { {   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0 },    194.38718,   0.00005,     0.237 },
 { {   0,   0,   9,  -8,   0,   0,   0,   0,   2,   0,   0 },    223.16101,   0.00010,     0.034 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,  -2,   0 },     71.71953,   0.00004,     0.285 },
 { {   0,   0,   9,  -7,   0,   0,   0,   0,   2,   0,   0 },    251.71953,   0.00004,     0.033 },
 { {   0,   0,   9,  -6,   0,   0,   0,   0,   2,  -1,   0 },    102.54435,   0.00006,     0.058 },
 { {   0,   0,  10, -19,   0,   0,   0,   0,  -2,   0,   0 },     81.65816,   0.00006,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   0,   0 },    252.29779,   0.00010,     0.040 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,  -2,   1,   0 },    252.29779,   0.00006,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,  -1,   0 },     72.29779,   0.00006,     0.087 },
 { {   0,   0,  10, -19,   0,   3,   0,   0,   2,   0,   0 },     72.29779,   0.00010,     0.040 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,  -2,   0,   0 },    110.45814,   0.00007,     0.041 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,  -2,   1,   0 },    290.43532,   0.00003,     0.090 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,   0,  -1,   0 },    110.42974,   0.00004,     0.078 },
 { {   0,   0,  10, -18,   0,   0,   0,   0,   0,   1,   0 },    290.42974,   0.00004,     0.073 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   0,   0 },    139.77973,   0.00003,     0.042 },
 { {   0,   0,  10, -17,   0,   0,   0,   0,  -2,   2,   0 },    319.77973,   0.00003,     0.366 },
 { {   0,   0,  10, -16,   0,   0,   0,   0,  -2,   0,   0 },    167.04185,   0.00006,     0.043 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -3,   0 },    211.20299,   0.00002,     0.077 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -2,   0 },    210.12408,   0.00003,     3.989 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,  -1,   0 },     29.39053,   0.00006,     0.074 },
 { {   0,   0,  10, -15,   0,   0,   0,   0,   2,   0,   0 },     30.12408,   0.00003,     0.037 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,  -2,   0,   0 },    245.96619,   0.00002,     0.045 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,  -2,   2,   0 },    245.96619,   0.00002,     0.231 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -3,   0 },    239.71125,   0.00002,     0.080 },
 { {   0,   0,  10, -14,   0,   0,   0,   0,   2,  -1,   0 },     59.71125,   0.00002,     0.071 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   0,   0 },    125.32187,   0.00003,     0.047 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,  -2,   2,   0 },    125.32187,   0.00003,     0.185 },
 { {   0,   0,  10, -12,   0,   0,   0,   0,   2,  -1,   0 },    117.45470,   0.00003,     0.066 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,  -2,   0,   0 },    153.16781,   0.00004,     0.049 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   0,   0,   0 },    146.38724,   0.00004,     0.241 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -3,   0 },    324.92240,   0.00003,     0.092 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -2,   0 },    325.12835,   0.00003,     0.421 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,  -1,   0 },    324.92240,   0.00003,     0.064 },
 { {   0,   0,  10, -11,   0,   0,   0,   0,   2,   0,   0 },    145.12835,   0.00003,     0.035 },
 { {   0,   0,  10, -10,   0,   0,   0,   0,   0,   0,   0 },    182.25034,   0.00005,     0.214 },
 { {   0,   0,  10,  -9,   0,   0,   0,   0,   2,   0,   0 },    223.17263,   0.00010,     0.033 },
 { {   0,   0,  10,  -8,   0,   0,   0,   0,   2,   0,   0 },    251.74539,   0.00007,     0.033 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   2,  -2,   0 },     99.91806,   0.00003,     0.222 },
 { {   0,   0,  10,  -7,   0,   0,   0,   0,   2,   0,   0 },    279.91806,   0.00003,     0.032 },
 { {   0,   0,  10,  -6,   0,   0,   0,   0,   2,  -1,   0 },    131.32170,   0.00003,     0.055 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,  -2,   0,   0 },    232.61271,   0.00003,     0.040 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,  -1,   0 },    232.41752,   0.00002,     0.075 },
 { {   0,   0,  11, -21,   0,   0,   0,   0,   0,   1,   0 },     52.41752,   0.00002,     0.076 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,  -2,   0,   0 },     81.51839,   0.00005,     0.041 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,  -1,   0 },     81.51838,   0.00003,     0.078 },
 { {   0,   0,  11, -20,   0,   0,   0,   0,   0,   1,   0 },    261.51838,   0.00003,     0.073 },
 { {   0,   0,  11, -19,   0,   0,   0,   0,  -2,   0,   0 },    110.25177,   0.00006,     0.042 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,  -2,   0,   0 },    138.71684,   0.00007,     0.043 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -3,   0 },    152.05105,   0.00003,     0.074 },
 { {   0,   0,  11, -18,   0,   0,   0,   0,   2,  -1,   0 },    332.05105,   0.00003,     0.077 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   0,   0 },    166.49522,   0.00002,     0.044 },
 { {   0,   0,  11, -17,   0,   0,   0,   0,  -2,   2,   0 },    166.49522,   0.00002,     0.268 },
 { {   0,   0,  11, -16,   0,   0,   0,   0,   2,   0,   0 },     30.07435,   0.00005,     0.037 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,  -2,   0,   0 },    247.14602,   0.00003,     0.046 },
 { {   0,   0,  11, -15,   0,   0,   0,   0,  -2,   2,   0 },     67.14602,   0.00003,     0.208 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -3,   0 },     87.97054,   0.00003,     0.087 },
 { {   0,   0,  11, -14,   0,   0,   0,   0,   2,  -1,   0 },    267.97054,   0.00003,     0.066 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,  -2,   0,   0 },    121.40976,   0.00004,     0.048 },
 { {   0,   0,  11, -13,   0,   0,   0,   0,   2,  -1,   0 },    296.60646,   0.00004,     0.064 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,  -2,   0 },    144.52383,   0.00002,     0.351 },
 { {   0,   0,  11, -12,   0,   0,   0,   0,   2,   0,   0 },    144.52383,   0.00002,     0.034 },
 { {   0,   0,  11, -10,   0,   0,   0,   0,   2,   0,   0 },    229.56226,   0.00008,     0.033 },
 { {   0,   0,  11,  -9,   0,   0,   0,   0,   2,   0,   0 },    258.92672,   0.00006,     0.032 },
 { {   0,   0,  11,  -8,   0,   0,   0,   0,   2,   0,   0 },    281.78785,   0.00005,     0.032 },
 { {   0,   0,  12, -26,   0,   0,   0,   0,  -2,   1,   0 },    116.86990,   0.00005,     0.075 },
 { {   0,   0,  12, -22,   0,   0,   0,   0,  -2,   0,   0 },    232.50304,   0.00003,     0.041 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,  -2,   0,   0 },     81.35654,   0.00002,     0.042 },
 { {   0,   0,  12, -21,   0,   0,   0,   0,  -2,   2,   0 },     81.35654,   0.00002,     0.383 },
 { {   0,   0,  12, -20,   0,   0,   0,   0,  -2,   0,   0 },    110.01350,   0.00006,     0.043 },
 { {   0,   0,  12, -19,   0,   0,   0,   0,  -2,   0,   0 },    138.33888,   0.00005,     0.044 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -3,   0 },      1.01129,   0.00003,     0.079 },
 { {   0,   0,  12, -18,   0,   0,   0,   0,   2,  -1,   0 },    181.01129,   0.00003,     0.072 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,  -1,   0 },    210.00937,   0.00005,     0.069 },
 { {   0,   0,  12, -17,   0,   0,   0,   0,   2,   0,   0 },     30.00937,   0.00004,     0.036 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,  -2,   0,   0 },    247.06576,   0.00003,     0.047 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,  -2,   2,   0 },     67.06576,   0.00003,     0.190 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -2,   0 },    238.96065,   0.00003,     0.582 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,  -1,   0 },    238.96065,   0.00005,     0.067 },
 { {   0,   0,  12, -16,   0,   0,   0,   0,   2,   0,   0 },     58.96065,   0.00003,     0.035 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,  -2,   0,   0 },     93.24294,   0.00006,     0.048 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -2,   0 },    267.78052,   0.00003,     0.444 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,  -1,   0 },    267.78052,   0.00004,     0.064 },
 { {   0,   0,  12, -15,   0,   0,   0,   0,   2,   0,   0 },     87.78052,   0.00003,     0.035 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,  -2,   0,   0 },    123.87763,   0.00003,     0.050 },
 { {   0,   0,  12, -14,   0,   0,   0,   0,   2,   0,   0 },    116.02189,   0.00004,     0.034 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,  -2,   0 },    224.29011,   0.00002,     0.228 },
 { {   0,   0,  12, -11,   0,   0,   0,   0,   2,   0,   0 },    224.17999,   0.00005,     0.032 },
 { {   0,   0,  12, -10,   0,   0,   0,   0,   2,   0,   0 },    252.78022,   0.00004,     0.032 },
 { {   0,   0,  12,  -9,   0,   0,   0,   0,   2,   0,   0 },    281.61329,   0.00004,     0.031 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,  -2,   0,   0 },    109.74060,   0.00002,     0.044 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,  -2,   2,   0 },    109.74060,   0.00002,     0.277 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -3,   0 },    299.68501,   0.00003,     0.076 },
 { {   0,   0,  13, -21,   0,   0,   0,   0,   2,  -1,   0 },    119.68501,   0.00003,     0.075 },
 { {   0,   0,  13, -18,   0,   0,   0,   0,   2,   0,   0 },     29.92614,   0.00004,     0.036 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,  -2,   0,   0 },    246.96487,   0.00005,     0.048 },
 { {   0,   0,  13, -17,   0,   0,   0,   0,   2,   0,   0 },     58.82488,   0.00005,     0.035 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,  -2,   0,   0 },     94.95771,   0.00004,     0.049 },
 { {   0,   0,  13, -16,   0,   0,   0,   0,   2,   0,   0 },     87.54893,   0.00005,     0.034 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,  -2,   0,   0 },    123.61645,   0.00002,     0.051 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,  -2,   2,   0 },    303.61645,   0.00002,     0.147 },
 { {   0,   0,  13, -15,   0,   0,   0,   0,   2,   0,   0 },    115.82057,   0.00004,     0.034 },
 { {   0,   0,  13, -12,   0,   0,   0,   0,   2,   0,   0 },    224.05743,   0.00003,     0.032 },
 { {   0,   0,  14, -23,   0,   0,   0,   0,   2,  -1,   0 },     46.53989,   0.00004,     0.075 },
 { {   0,   0,  14, -21,   0,   0,   0,   0,  -2,   0,   0 },    317.38286,   0.00003,     0.046 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,  -2,   0,   0 },    246.85441,   0.00005,     0.049 },
 { {   0,   0,  14, -18,   0,   0,   0,   0,   2,   0,   0 },     58.65925,   0.00004,     0.034 },
 { {   0,   0,  14, -17,   0,   0,   0,   0,   2,   0,   0 },     87.27202,   0.00004,     0.034 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,  -2,   0,   0 },    123.34996,   0.00002,     0.052 },
 { {   0,   0,  14, -16,   0,   0,   0,   0,  -2,   2,   0 },    123.34996,   0.00002,     0.138 },
 { {   0,   0,  14, -12,   0,   0,   0,   0,   2,  -2,   0 },     73.09262,   0.00003,     0.171 },
 { {   0,   0,  14, -12,   0,   0,   0,   0,   2,   0,   0 },    253.09262,   0.00003,     0.031 },
 { {   0,   0,  15, -20,   0,   0,   0,   0,   2,   0,   0 },     29.73739,   0.00003,     0.034 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,  -2,   0,   0 },    246.71998,   0.00005,     0.050 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,  -2,   0 },    238.48507,   0.00002,     0.320 },
 { {   0,   0,  15, -19,   0,   0,   0,   0,   2,   0,   0 },     58.48507,   0.00002,     0.034 },
 { {   0,   0,  15, -13,   0,   0,   0,   0,   2,   0,   0 },    252.84332,   0.00004,     0.030 },
 { {   0,   0,  15, -12,   0,   0,   0,   0,   2,  -2,   0 },    101.64001,   0.00002,     0.146 },
 { {   0,   0,  15, -12,   0,   0,   0,   0,   2,   0,   0 },    281.64001,   0.00002,     0.030 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,  -2,   0,   0 },    246.56447,   0.00002,     0.052 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,  -2,   2,   0 },    246.56447,   0.00002,     0.140 },
 { {   0,   0,  16, -20,   0,   0,   0,   0,   2,   0,   0 },     58.24944,   0.00003,     0.033 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,  -2,   0,   0 },    274.98776,   0.00003,     0.054 },
 { {   0,   0,  17, -20,   0,   0,   0,   0,  -2,   2,   0 },     94.98776,   0.00003,     0.123 },
 { {   0,   0,  18, -21,   0,   0,   0,   0,  -2,   0,   0 },    274.70038,   0.00005,     0.056 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,  -2,   0,   0 },    302.83122,   0.00003,     0.058 },
 { {   0,   0,  18, -20,   0,   0,   0,   0,  -2,   2,   0 },    122.83122,   0.00003,     0.109 },
 { {   0,   0,  19, -21,   0,   0,   0,   0,  -2,   0,   0 },    302.33767,   0.00002,     0.059 },
 { {   0,   0,  19, -21,   0,   0,   0,   0,  -2,   2,   0 },    302.33767,   0.00002,     0.104 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   0,   0 },    222.29550,   0.00002,     0.075 },
 { {   0,   0,  37, -48,   0,   0,   0,   0,  -2,   2,   0 },     42.29550,   0.00002,     0.075 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.06990,   0.00020,     0.033 },
 { {   0,   1,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    252.04854,   0.00005,     0.186 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },    160.47271,   0.00006,     0.019 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    160.21061,   0.00019,     0.063 },
 { {   0,   1,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },      8.35958,   0.00004,     0.385 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    106.09845,   0.00005,     0.036 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    295.97335,   0.00005,     0.067 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,  -2,   0 },     69.42181,   0.00002,     0.033 },
 { {   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     69.42181,   0.00002,     0.296 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    302.73818,   0.00021,     0.037 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    141.85538,   0.00006,     0.072 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    257.94221,   0.00002,     0.035 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    244.41935,   0.00002,     0.064 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    257.94221,   0.00002,     0.421 },
 { {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },     64.41935,   0.00002,     0.092 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },     61.65773,   0.00007,     0.078 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,  -1,   1 },     28.16297,   0.00003,     0.074 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   1,   1 },    208.16297,   0.00003,     0.077 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    331.21795,   0.00004,     0.068 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    151.21795,   0.00004,     0.084 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    159.64038,   0.00003,     0.061 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    159.64038,   0.00003,     0.099 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },     17.44240,   0.00008,     0.043 },
 { {   0,   1,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },     18.21808,   0.00003,     0.020 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    115.70713,   0.00044,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },     75.68938,   0.00004,     0.084 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1 },    270.38629,   0.00002,     0.080 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   1 },     67.77681,   0.00003,     1.472 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,  -1 },    115.95776,   0.00004,     0.077 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   1,   1 },     90.38629,   0.00002,     0.072 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   2,   1 },    273.98463,   0.00002,     0.037 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -3,   2 },    289.48808,   0.00002,     0.075 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     42.94839,   0.00002,     0.073 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   2 },    101.55514,   0.00006,     0.076 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    179.29612,   0.00003,     2.670 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    222.94839,   0.00002,     0.078 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -2,   1 },    167.48486,   0.00002,     0.913 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },    293.70408,   0.00003,     0.068 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },    311.86021,   0.00006,     0.082 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    270.58118,   0.00008,     0.700 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   1 },    347.48486,   0.00002,     0.039 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    113.70408,   0.00003,     0.085 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   2,  -1 },     90.11553,   0.00003,     0.040 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    295.60318,   0.00004,     0.090 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    281.29447,   0.00003,     0.041 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,  -2,   0 },    288.11614,   0.00002,     0.044 },
 { {   0,   1,  -2,   0,   0,   0,   0,   0,   4,   0,   0 },    108.11614,   0.00002,     0.020 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    171.53584,   0.00019,     0.041 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },      0.24826,   0.00009,     0.092 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    191.61513,   0.00005,     0.079 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },     18.30833,   0.00009,     1.599 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },     11.61513,   0.00005,     0.072 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },    207.30913,   0.00006,     0.083 },
 { {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    218.78424,   0.00004,     0.039 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   0 },    249.60029,   0.00005,     0.021 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    108.64402,   0.00003,     0.043 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0 },     73.20048,   0.00026,     0.101 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.22854,   0.00003,     0.294 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   3,   0 },    281.60293,   0.00003,     0.060 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1 },     89.88529,   0.00002,     0.095 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1 },    190.66809,   0.00005,     0.041 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },    269.88529,   0.00002,     0.373 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   1 },    269.88529,   0.00002,     0.063 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   1 },    269.88529,   0.00002,     0.034 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    195.58666,   0.00004,     0.086 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     22.00317,   0.00003,     0.615 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     15.58666,   0.00004,     0.067 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.10629,   0.00013,     0.076 },
 { {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    239.68688,   0.00027,     0.038 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,  -2,   0 },    161.87100,   0.00002,     0.021 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0 },    162.58952,   0.00008,     0.045 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0 },      8.99684,   0.00004,     0.113 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    189.74282,   0.00009,     0.381 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    352.27463,   0.00002,     0.081 },
 { {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    352.27463,   0.00002,     0.071 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    181.31338,   0.00006,     0.066 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    266.09032,   0.00018,     0.035 },
 { {   0,   1,   2,   0,   0,   0,   0,   0,   2,   2,   0 },    275.52492,   0.00003,     0.018 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0 },    110.34109,   0.00005,     0.216 },
 { {   0,   1,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    302.38270,   0.00020,     0.034 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   0,   0 },     13.54402,   0.00002,     0.033 },
 { {   0,   1,   4,   0,   0,   0,   0,   0,   2,   2,   0 },    193.54402,   0.00002,     0.018 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   2,  -1,   0 },    317.24913,   0.00004,     0.055 },
 { {   0,   1,   5,   0,   0,   0,   0,   0,   2,   0,   0 },    326.77635,   0.00010,     0.032 },
 { {   0,   1,   6,   0,   0,   0,   0,   0,   2,  -2,   0 },    216.18434,   0.00003,     0.171 },
 { {   0,   1,   6,   0,   0,   0,   0,   0,   2,   0,   0 },     36.18434,   0.00003,     0.031 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.11980,   0.00021,     0.034 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },     71.51105,   0.00003,     0.032 },
 { {   0,   2,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    251.51105,   0.00003,     0.211 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },    160.42918,   0.00005,     0.019 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    160.20932,   0.00018,     0.066 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    346.44744,   0.00002,     0.059 },
 { {   0,   2,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    166.44744,   0.00002,     0.105 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.90597,   0.00003,     0.036 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    319.72141,   0.00002,     0.070 },
 { {   0,   2,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },    139.72141,   0.00002,     0.081 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    303.17533,   0.00019,     0.038 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    254.25467,   0.00003,     0.076 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    191.32818,   0.00002,     0.075 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    238.89788,   0.00004,     0.572 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     79.93304,   0.00002,     0.284 },
 { {   0,   2,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    259.93304,   0.00002,     0.044 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },     19.63488,   0.00004,     0.039 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    355.39585,   0.00007,     0.082 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,  -1,   1 },    210.03600,   0.00002,     0.077 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,  -1 },    147.07577,   0.00002,     0.075 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   1,   1 },     30.03600,   0.00002,     0.074 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,  -1,   3,  -1 },    327.07577,   0.00002,     0.076 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     69.70947,   0.00003,     0.071 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    249.70947,   0.00003,     0.080 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    324.34784,   0.00003,     0.063 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    144.34784,   0.00003,     0.093 },
 { {   0,   2,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     36.94740,   0.00014,     0.042 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    115.64869,   0.00007,     0.026 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },     21.43923,   0.00030,     0.041 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    102.76193,   0.00034,     0.089 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   0 },    229.80834,   0.00002,     0.494 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },     84.84427,   0.00002,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    108.27203,   0.00051,     0.077 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    214.01564,   0.00006,     3.983 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    288.36269,   0.00047,     0.074 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },    287.73117,   0.00002,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,  -1 },    261.60283,   0.00002,     0.037 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -2,   1 },    253.71954,   0.00002,     2.131 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,  -1,  -1 },     72.69259,   0.00004,     0.071 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    267.45758,   0.00007,     1.246 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   0,   1 },     15.05776,   0.00004,     0.038 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1 },    252.69259,   0.00004,     0.080 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    132.23962,   0.00002,     0.068 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    231.33143,   0.00002,     0.657 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    287.19142,   0.00046,     0.085 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    287.34107,   0.00019,     0.040 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   2,   2,   0 },    107.01968,   0.00002,     0.019 },
 { {   0,   2,  -3,   0,   0,   0,   0,   0,   4,   0,   0 },    108.17297,   0.00005,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,  -2,   0 },    343.29625,   0.00003,     0.020 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    143.88984,   0.00010,     0.043 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   0 },    196.29491,   0.00003,     0.040 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },     47.68262,   0.00002,     0.799 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },     79.60530,   0.00003,     1.912 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    223.43583,   0.00009,     0.079 },
 { {   0,   2,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    259.60530,   0.00003,     0.038 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   0 },     72.09921,   0.00005,     0.021 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },    234.93527,   0.00022,     0.044 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0 },    250.51329,   0.00035,     0.108 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   1 },    269.88512,   0.00006,     0.303 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1 },    269.88511,   0.00005,     0.097 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0 },    129.92718,   0.00006,     0.091 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },      3.34213,   0.00006,     0.444 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0 },    349.39992,   0.00005,     0.064 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,   1 },    155.22381,   0.00002,     0.653 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   1 },    155.22381,   0.00002,     0.036 },
 { {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    229.17431,   0.00021,     0.037 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0 },    162.74379,   0.00005,     0.047 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },    161.80200,   0.00003,     0.100 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 },     32.24349,   0.00004,     0.308 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    341.80200,   0.00003,     0.061 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },    355.75507,   0.00003,     0.085 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     99.27309,   0.00006,     0.068 },
 { {   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    262.22109,   0.00007,     0.036 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },     98.41556,   0.00004,     0.111 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    305.28945,   0.00004,     0.235 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    278.41556,   0.00004,     0.057 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -3,   0 },    164.26774,   0.00002,     0.093 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -1,   0 },    164.26774,   0.00002,     0.064 },
 { {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    203.47164,   0.00026,     0.034 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    110.29190,   0.00005,     0.190 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    151.49754,   0.00002,     0.102 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -2,   0 },    225.02914,   0.00002,     0.288 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    151.49754,   0.00002,     0.060 },
 { {   0,   2,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    297.12999,   0.00018,     0.033 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0 },     19.79004,   0.00003,     0.160 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,  -1,   0 },     19.69575,   0.00023,     0.056 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     13.54384,   0.00002,     0.032 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   1,   0 },    199.38414,   0.00005,     0.023 },
 { {   0,   2,   3,   0,   0,   0,   0,   0,   2,   2,   0 },    193.54384,   0.00002,     0.017 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,  -2,   0 },    323.62010,   0.00003,     0.183 },
 { {   0,   2,   4,   0,   0,   0,   0,   0,   2,   0,   0 },    143.62010,   0.00003,     0.031 },
 { {   0,   2,   5,   0,   0,   0,   0,   0,   2,  -2,   0 },    215.14479,   0.00002,     0.154 },
 { {   0,   2,   5,   0,   0,   0,   0,   0,   2,   0,   0 },     35.14479,   0.00002,     0.030 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.17906,   0.00021,     0.035 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },     57.30083,   0.00003,     0.064 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,  -2,   3,   0 },    237.30083,   0.00003,     0.092 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },     71.52631,   0.00003,     0.033 },
 { {   0,   3,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    251.52631,   0.00003,     0.243 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -4,   0,   0 },    160.38872,   0.00003,     0.019 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -4,   2,   0 },    160.38872,   0.00003,     0.038 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.76366,   0.00015,     0.036 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    160.20656,   0.00018,     0.068 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    346.44111,   0.00002,     0.061 },
 { {   0,   3,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    166.44111,   0.00002,     0.099 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,  -1,   2 },    227.63746,   0.00004,     0.075 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    137.60758,   0.00003,     0.037 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,  -2,   1,   2 },     47.63746,   0.00004,     0.076 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },     60.17014,   0.00002,     0.065 },
 { {   0,   3,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    240.17014,   0.00002,     0.090 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    319.07613,   0.00018,     0.039 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    278.62269,   0.00004,     0.079 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    335.61028,   0.00002,     1.537 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,  -1,   1 },    122.00066,   0.00003,     0.075 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,  -1,   1,   1 },    302.00066,   0.00003,     0.076 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },     57.27848,   0.00002,     0.036 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    237.27848,   0.00002,     0.890 },
 { {   0,   3,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    238.46794,   0.00005,     0.042 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    301.03926,   0.00017,     0.040 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    289.43433,   0.00011,     0.086 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    305.13574,   0.00018,     0.075 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    305.81156,   0.00003,     8.102 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    125.13574,   0.00018,     0.076 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },     63.85440,   0.00002,     0.527 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     98.65962,   0.00012,     0.088 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     90.35427,   0.00013,     0.041 },
 { {   0,   3,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },     42.79754,   0.00003,     0.020 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,  -1,   0 },    118.11521,   0.00002,     0.027 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    108.97703,   0.00039,     0.042 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },     92.45517,   0.00018,     0.094 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,  -1,   0,   0 },    280.89567,   0.00006,     0.087 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },    101.64024,   0.00030,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    203.43612,   0.00005,     1.141 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    282.58781,   0.00026,     0.071 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -2,   0 },    115.61755,   0.00003,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,  -1,   1 },     45.93821,   0.00004,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    270.11405,   0.00004,     5.644 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   0 },    281.16673,   0.00035,     0.075 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },     31.90229,   0.00005,     0.037 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    291.46727,   0.00019,     0.081 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    290.13498,   0.00002,     0.039 },
 { {   0,   3,  -4,   0,   0,   0,   0,   0,   4,   0,   0 },    108.22105,   0.00004,     0.020 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.03935,   0.00015,     0.044 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    192.93214,   0.00004,     0.041 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    154.38632,   0.00015,     0.075 },
 { {   0,   3,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    256.41968,   0.00005,     0.038 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    243.12209,   0.00053,     0.046 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    251.59572,   0.00045,     0.116 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -1 },    106.71680,   0.00007,     0.044 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     56.60987,   0.00004,     0.096 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    335.87261,   0.00006,     0.348 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    236.60987,   0.00004,     0.062 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -3,   1 },    234.11232,   0.00002,     0.090 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,  -1 },     54.11039,   0.00002,     0.087 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,  -1,   1 },     54.11232,   0.00002,     0.065 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   0,  -1 },    204.99035,   0.00004,     0.549 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   1,   1,  -1 },    234.11039,   0.00002,     0.066 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    308.41420,   0.00004,     0.907 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    243.65792,   0.00040,     0.036 },
 { {   0,   3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1 },    335.37157,   0.00004,     0.075 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     83.72073,   0.00006,     0.048 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,  -1,   0 },     81.50723,   0.00005,     0.065 },
 { {   0,   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    291.04037,   0.00006,     0.035 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    289.70802,   0.00003,     0.205 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     76.76014,   0.00006,     0.061 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    236.19470,   0.00045,     0.034 },
 { {   0,   3,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    108.89549,   0.00002,     0.018 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    110.26493,   0.00006,     0.170 },
 { {   0,   3,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    290.64914,   0.00020,     0.033 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0 },    193.54423,   0.00005,     0.145 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   0,   0 },     16.42646,   0.00022,     0.032 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   2,   2,   0 },     13.54426,   0.00002,     0.017 },
 { {   0,   3,   2,   0,   0,   0,   0,   0,   4,   0,   0 },     18.82059,   0.00005,     0.018 },
 { {   0,   3,   3,   0,   0,   0,   0,   0,   2,   0,   0 },    282.71220,   0.00004,     0.031 },
 { {   0,   3,   4,   0,   0,   0,   0,   0,   2,   0,   0 },     36.10807,   0.00004,     0.030 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.24133,   0.00021,     0.035 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },     77.60681,   0.00002,     0.067 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    257.60681,   0.00002,     0.087 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },     71.54758,   0.00003,     0.033 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    251.54758,   0.00003,     0.060 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    251.54758,   0.00003,     0.286 },
 { {   0,   4, -10,   0,   0,   0,   0,   0,   0,   1,   0 },     71.54758,   0.00003,     0.102 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -4,   0,   0 },    160.35089,   0.00002,     0.019 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -4,   2,   0 },    160.35089,   0.00002,     0.039 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    161.49181,   0.00014,     0.037 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    145.47122,   0.00002,     0.034 },
 { {   0,   4,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    325.47122,   0.00002,     0.400 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     57.82212,   0.00003,     0.077 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     57.46061,   0.00003,     0.068 },
 { {   0,   4,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    237.46061,   0.00003,     0.085 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    307.18581,   0.00040,     0.040 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    182.39234,   0.00006,     0.083 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    314.59384,   0.00004,     0.073 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    134.59384,   0.00004,     0.078 },
 { {   0,   4,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    106.01195,   0.00004,     0.091 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    324.32159,   0.00007,     0.041 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    274.68624,   0.00002,     0.091 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   2,   0 },    165.55503,   0.00003,     0.439 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,  -2,   3,   0 },     94.68624,   0.00002,     0.064 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    344.56756,   0.00008,     0.078 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    321.43840,   0.00004,     1.992 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    164.56756,   0.00008,     0.073 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   1,  -1,   1 },    126.64922,   0.00004,     0.077 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    197.14915,   0.00020,     0.040 },
 { {   0,   4,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },     18.49190,   0.00003,     0.019 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    288.84054,   0.00022,     0.043 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    293.86532,   0.00002,     0.100 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },     66.32927,   0.00002,     0.305 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -2,   3,   0 },    113.86532,   0.00002,     0.060 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   0,  -1 },     20.76355,   0.00002,     0.041 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,  -1,   2,  -1 },    200.76355,   0.00002,     0.450 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     54.42127,   0.00008,     0.085 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    187.40576,   0.00005,     0.666 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    211.27460,   0.00005,     0.068 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },     82.69047,   0.00003,     0.036 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,  -1 },    336.72078,   0.00002,     0.078 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,  -1,   1 },    231.13688,   0.00006,     0.071 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,  -1 },    270.11129,   0.00002,     2.231 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   0,   1 },     31.10125,   0.00004,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   1,  -1 },    156.72078,   0.00002,     0.073 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   1,   2,  -1 },    270.11129,   0.00002,     0.037 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     38.07690,   0.00010,     0.077 },
 { {   0,   4,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    319.14573,   0.00006,     0.038 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    344.26229,   0.00010,     0.045 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,  -2,   2,   0 },     13.79629,   0.00002,     0.234 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },    181.33161,   0.00003,     0.042 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    165.23652,   0.00002,     0.400 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    143.51507,   0.00006,     0.072 },
 { {   0,   4,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    273.91409,   0.00004,     0.037 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    241.26904,   0.00026,     0.047 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,  -2,   1,   0 },    251.88601,   0.00046,     0.125 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },      8.03469,   0.00002,     0.103 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },      1.94680,   0.00003,     0.286 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    188.03469,   0.00002,     0.060 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   0,   2,   0 },      1.94680,   0.00003,     0.033 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   1,   0,  -1 },    204.10218,   0.00004,     0.408 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -3,   0 },    291.90102,   0.00002,     0.087 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    291.90102,   0.00002,     0.067 },
 { {   0,   4,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    234.09308,   0.00019,     0.035 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   0,   0 },    339.38400,   0.00014,     0.049 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,  -2,   1,   0 },    151.72528,   0.00004,     0.143 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -3,   0 },    146.25997,   0.00003,     0.095 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -2,   0 },    135.73126,   0.00003,     0.367 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    326.25997,   0.00003,     0.063 },
 { {   0,   4,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    302.92725,   0.00007,     0.034 },
 { {   0,   4,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    243.88387,   0.00028,     0.033 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    109.71889,   0.00004,     0.154 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },    323.65364,   0.00002,     0.212 },
 { {   0,   4,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    293.71425,   0.00021,     0.032 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   0,   0 },     13.54469,   0.00002,     0.031 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   2,   2,   0 },    193.54469,   0.00002,     0.017 },
 { {   0,   4,   1,   0,   0,   0,   0,   0,   4,   0,   0 },     18.01556,   0.00003,     0.018 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,  -3,   0 },    303.43590,   0.00002,     0.153 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,  -1,   0 },    303.43590,   0.00002,     0.050 },
 { {   0,   4,   2,   0,   0,   0,   0,   0,   2,   0,   0 },    312.19534,   0.00005,     0.030 },
 { {   0,   4,   3,   0,   0,   0,   0,   0,   2,   0,   0 },     37.00480,   0.00003,     0.029 },
 { {   0,   5, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    330.32651,   0.00004,     0.065 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.30282,   0.00021,     0.036 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    251.57156,   0.00003,     0.062 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    251.57156,   0.00006,     0.348 },
 { {   0,   5, -11,   0,   0,   0,   0,   0,   0,   1,   0 },     71.57156,   0.00003,     0.096 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    148.24389,   0.00002,     0.075 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    148.24389,   0.00002,     0.076 },
 { {   0,   5, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    324.68538,   0.00004,     0.534 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,  -1,   0,   0 },     46.48635,   0.00006,     0.076 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     34.18855,   0.00003,     0.071 },
 { {   0,   5,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    214.18855,   0.00003,     0.081 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    306.98632,   0.00036,     0.041 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },    322.47304,   0.00004,     0.088 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },     12.46665,   0.00004,     0.076 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    192.46665,   0.00004,     0.075 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    295.40906,   0.00003,     0.067 },
 { {   0,   5,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    115.40906,   0.00003,     0.086 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },     47.47098,   0.00008,     0.042 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },    192.64667,   0.00005,     0.097 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    176.92522,   0.00007,     0.082 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    332.23467,   0.00003,     0.887 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    356.92522,   0.00007,     0.070 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -3,   1 },      1.14606,   0.00003,     0.078 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,  -1 },    135.40023,   0.00003,     0.076 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },    181.14606,   0.00003,     0.073 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   1,   1,  -1 },    315.40023,   0.00003,     0.075 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -3,   0 },    251.67728,   0.00002,     0.072 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },     71.67728,   0.00002,     0.079 },
 { {   0,   5,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    199.55511,   0.00022,     0.039 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -2,   1,   0 },    347.53401,   0.00005,     0.107 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   0,  -1 },     20.25865,   0.00002,     0.042 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,  -1,   2,  -1 },    200.25865,   0.00002,     0.351 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    190.06846,   0.00003,     0.041 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    353.41196,   0.00003,     0.090 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    241.37209,   0.00007,     0.470 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    173.41196,   0.00003,     0.065 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },     30.45206,   0.00003,     0.036 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,  -2 },    106.75491,   0.00007,     0.076 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },    270.89587,   0.00017,     0.073 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     94.85443,   0.00007,     0.037 },
 { {   0,   5,  -6,   0,   0,   0,   0,   0,   2,   1,  -2 },    286.75491,   0.00007,     0.075 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    345.57480,   0.00011,     0.046 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },    343.05765,   0.00002,     0.099 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },     29.71886,   0.00002,     0.320 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    163.05765,   0.00002,     0.061 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },     29.71886,   0.00002,     0.034 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     87.72857,   0.00004,     0.068 },
 { {   0,   5,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    267.72919,   0.00004,     0.036 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.94383,   0.00027,     0.049 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,  -2,   1,   0 },    252.36498,   0.00043,     0.136 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     91.70947,   0.00003,     0.110 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    271.70947,   0.00003,     0.058 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,  -1 },    203.36223,   0.00004,     0.325 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   1,   0,   1 },    216.05989,   0.00003,     0.033 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    113.94537,   0.00002,     0.092 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },     10.86309,   0.00006,     0.064 },
 { {   0,   5,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    231.97290,   0.00016,     0.035 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    164.26103,   0.00004,     0.051 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    325.74031,   0.00003,     0.060 },
 { {   0,   5,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    288.43231,   0.00005,     0.033 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   0,  -1,   0 },     56.28151,   0.00002,     0.140 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   0,   1,   0 },    236.28151,   0.00002,     0.052 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,  -1,   0 },    358.67182,   0.00005,     0.057 },
 { {   0,   5,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    254.27161,   0.00040,     0.032 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0 },    109.75194,   0.00004,     0.140 },
 { {   0,   5,  -1,   0,   0,   0,   0,   0,   2,   0,   0 },    293.71650,   0.00021,     0.031 },
 { {   0,   5,   1,   0,   0,   0,   0,   0,   2,   0,   0 },    313.65644,   0.00004,     0.030 },
 { {   0,   6, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    330.30579,   0.00003,     0.068 },
 { {   0,   6, -13,   0,   0,   0,   0,   0,  -2,   3,   0 },    150.30579,   0.00003,     0.085 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.36141,   0.00020,     0.037 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },    262.05811,   0.00002,     0.073 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,  -2,   3,   0 },    262.05811,   0.00002,     0.078 },
 { {   0,   6, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    251.59613,   0.00006,     0.445 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,  -1,   0 },    340.28069,   0.00004,     0.025 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    217.14879,   0.00004,     0.079 },
 { {   0,   6, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    323.84167,   0.00003,     0.802 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    210.19227,   0.00002,     0.040 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    158.29838,   0.00002,     0.085 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },    210.19227,   0.00002,     0.655 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    338.29838,   0.00002,     0.068 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },    359.46567,   0.00002,     0.074 },
 { {   0,   6, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    179.46567,   0.00002,     0.077 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    282.14796,   0.00015,     0.042 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    298.30598,   0.00004,     0.093 },
 { {   0,   6,  -9,   0,   0,   0,   0,   0,   1,  -1,   1 },    301.69719,   0.00006,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    353.07356,   0.00005,     0.044 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    284.81724,   0.00008,     0.087 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },    117.61876,   0.00004,     0.075 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    284.13153,   0.00047,     0.076 },
 { {   0,   6,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    199.61369,   0.00020,     0.038 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -2,   1,   0 },     11.53131,   0.00006,     0.115 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,  -1,   0,  -1 },     19.83189,   0.00004,     0.043 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -2,   0 },    213.53028,   0.00002,     0.042 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -1,   0 },    144.97616,   0.00003,     0.095 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    213.53028,   0.00002,     0.363 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   0,   1,   0 },    324.97616,   0.00003,     0.062 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   1,  -1,   1 },     90.10049,   0.00005,     0.065 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },     91.44164,   0.00003,     1.021 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    281.92135,   0.00006,     0.070 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },     91.44164,   0.00003,     0.036 },
 { {   0,   6,  -7,   0,   0,   0,   0,   0,   2,   1,   0 },    288.31918,   0.00003,     0.025 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.33202,   0.00010,     0.048 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },    343.69110,   0.00003,     0.105 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    163.69110,   0.00003,     0.059 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },      4.37305,   0.00010,     0.066 },
 { {   0,   6,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    236.22722,   0.00003,     0.035 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -4,   0,   0 },     72.75996,   0.00003,     0.022 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -4,   2,   0 },    252.75996,   0.00003,     0.055 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.34440,   0.00026,     0.050 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,  -2,   1,   0 },    252.75996,   0.00041,     0.149 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.84561,   0.00005,     0.118 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   0,   1,   0 },    250.84561,   0.00005,     0.056 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },     30.61675,   0.00003,     0.062 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     68.02265,   0.00002,     0.034 },
 { {   0,   6,  -5,   0,   0,   0,   0,   0,   2,   2,   0 },    248.02265,   0.00002,     0.018 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    165.43651,   0.00004,     0.053 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    339.14622,   0.00002,     0.108 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    339.14622,   0.00002,     0.058 },
 { {   0,   6,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    239.73536,   0.00003,     0.033 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    280.90444,   0.00005,     0.056 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },     57.44602,   0.00003,     0.154 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    237.44602,   0.00003,     0.050 },
 { {   0,   6,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    255.13440,   0.00034,     0.032 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    108.93433,   0.00003,     0.129 },
 { {   0,   6,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },    299.25042,   0.00020,     0.031 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.41597,   0.00019,     0.038 },
 { {   0,   7, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    251.61999,   0.00006,     0.617 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    340.24721,   0.00003,     0.026 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    333.46813,   0.00003,     0.039 },
 { {   0,   7, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    323.04839,   0.00003,     1.609 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    219.72951,   0.00004,     0.041 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,  -2,   1,   0 },    217.67704,   0.00007,     0.090 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    210.76434,   0.00002,     0.078 },
 { {   0,   7, -11,   0,   0,   0,   0,   0,   0,   1,   0 },     30.76434,   0.00002,     0.073 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,  -2,   0 },    305.98575,   0.00003,     0.020 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    276.39724,   0.00014,     0.043 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     70.57936,   0.00002,     0.084 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    233.02481,   0.00003,     0.725 },
 { {   0,   7, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    250.57936,   0.00002,     0.068 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },     16.74038,   0.00003,     0.045 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,  -2,   2,   0 },    196.74038,   0.00003,     0.241 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     19.39993,   0.00003,     0.092 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },      8.59703,   0.00005,     0.420 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    199.39993,   0.00003,     0.064 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },     53.46483,   0.00002,     0.079 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    233.46483,   0.00002,     0.072 },
 { {   0,   7,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    199.66826,   0.00017,     0.037 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    293.34094,   0.00017,     0.047 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    252.22899,   0.00002,     0.194 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,  -1,   0,  -1 },     19.46886,   0.00003,     0.045 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -2,   0 },    200.07978,   0.00003,     0.043 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    126.49272,   0.00004,     0.101 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    200.07978,   0.00003,     0.296 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },    306.49272,   0.00004,     0.060 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,  -3,   0 },     72.86869,   0.00002,     0.086 },
 { {   0,   7,  -8,   0,   0,   0,   0,   0,   2,   1,   0 },    288.33889,   0.00002,     0.024 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    352.79363,   0.00008,     0.049 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,  -2,   2,   0 },     10.01050,   0.00003,     0.163 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -2,   0 },    332.18242,   0.00003,     0.384 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,  -1,   0 },    356.02184,   0.00010,     0.063 },
 { {   0,   7,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    192.01434,   0.00009,     0.034 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -4,   0,   0 },     72.78735,   0.00003,     0.023 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -4,   2,   0 },    252.78735,   0.00003,     0.057 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,  -2,   0,   0 },    241.90536,   0.00025,     0.052 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,  -1,   0 },     95.09993,   0.00004,     0.127 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    252.50498,   0.00005,     0.186 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   1,   0 },    275.09993,   0.00004,     0.054 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   0,   2,   0 },     55.19670,   0.00002,     0.031 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   1,   0,   1 },    229.18600,   0.00004,     0.032 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,  -1,   0 },     60.43486,   0.00008,     0.059 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    215.07533,   0.00002,     0.033 },
 { {   0,   7,  -6,   0,   0,   0,   0,   0,   2,   2,   0 },     71.14125,   0.00002,     0.018 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,  -2,   0,   0 },    281.23788,   0.00006,     0.058 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,  -1,   0 },     57.19770,   0.00002,     0.170 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   0,   1,   0 },    237.19770,   0.00002,     0.048 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    179.81728,   0.00003,     0.178 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },     19.32998,   0.00005,     0.031 },
 { {   0,   7,  -4,   0,   0,   0,   0,   0,   2,   2,   0 },    236.94990,   0.00002,     0.017 },
 { {   0,   7,  -3,   0,   0,   0,   0,   0,   2,   0,   0 },    290.88206,   0.00017,     0.030 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,   0,   0 },    195.71988,   0.00004,     0.107 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   0,   2,   0 },     13.54044,   0.00002,     0.028 },
 { {   0,   7,  -2,   0,   0,   0,   0,   0,   2,   0,   0 },     16.43122,   0.00015,     0.029 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    151.76665,   0.00003,     0.037 },
 { {   0,   8, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    331.76665,   0.00003,     4.509 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     70.66357,   0.00015,     0.039 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },     77.79756,   0.00002,     1.285 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    251.64234,   0.00005,     1.004 },
 { {   0,   8, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    105.66078,   0.00007,     0.042 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    185.32873,   0.00037,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    183.94690,   0.00034,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    173.96053,   0.00048,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    325.61721,   0.00002,   238.919 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    354.01642,   0.00048,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    351.29952,   0.00002,     0.038 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },      7.10012,   0.00031,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   2,   0,   0 },      2.84344,   0.00039,     0.040 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    185.72721,   0.00007,     0.044 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    232.33087,   0.00003,     0.499 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    279.32480,   0.00003,     4.345 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,  -1,   0 },    276.25836,   0.00006,     0.074 },
 { {   0,   8, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     99.32480,   0.00003,     0.037 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     62.35257,   0.00002,     0.046 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },     62.35257,   0.00002,     0.209 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,  -1,   0 },     18.81506,   0.00003,     0.098 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    353.83062,   0.00002,     0.333 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   1,   0 },    198.81506,   0.00003,     0.062 },
 { {   0,   8, -10,   0,   0,   0,   0,   0,   0,   2,   0 },     17.59591,   0.00003,     0.034 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    288.84670,   0.00017,     0.048 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    321.47924,   0.00002,     0.046 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },    141.47924,   0.00002,     0.212 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -2,   0 },    200.09856,   0.00002,     0.044 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,  -1,   0 },     99.12852,   0.00004,     0.108 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },    249.14724,   0.00005,     0.250 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   0,   1,   0 },    279.12852,   0.00004,     0.058 },
 { {   0,   8,  -9,   0,   0,   0,   0,   0,   1,   0,   1 },    270.08183,   0.00006,     0.034 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,  -2,   0 },    343.88068,   0.00003,     0.022 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    348.87327,   0.00004,     0.051 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    355.51068,   0.00002,     0.148 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    359.76321,   0.00004,     0.061 },
 { {   0,   8,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    184.31246,   0.00009,     0.034 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -4,   0,   0 },     72.80959,   0.00002,     0.023 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -4,   2,   0 },     72.80959,   0.00002,     0.059 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,  -2,   0,   0 },    118.08016,   0.00006,     0.053 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    235.95131,   0.00009,     0.167 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   0,   2,   0 },     65.66046,   0.00002,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   1,   0,   1 },    201.76139,   0.00004,     0.031 },
 { {   0,   8,  -7,   0,   0,   0,   0,   0,   2,   0,   0 },    210.42976,   0.00007,     0.033 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,  -2,   0 },    165.56364,   0.00003,     0.051 },
 { {   0,   8,  -6,   0,   0,   0,   0,   0,   0,   0,   0 },    345.56364,   0.00003,     0.143 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,  -2,   0,   0 },    280.85483,   0.00003,     0.060 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,  -2,   2,   0 },    100.85483,   0.00003,     0.102 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,   0,   0 },    234.71618,   0.00002,     0.125 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   0,   2,   0 },    234.71618,   0.00002,     0.029 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,  -2,   0 },    176.52229,   0.00002,     0.161 },
 { {   0,   8,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },     42.26042,   0.00005,     0.031 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -3,   0 },    289.10777,   0.00002,     0.166 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -2,   0 },    294.99398,   0.00002,     0.138 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.10777,   0.00002,     0.049 },
 { {   0,   8,  -4,   0,   0,   0,   0,   0,   2,   0,   0 },    114.99398,   0.00002,     0.030 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    195.91388,   0.00004,     0.100 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },     18.00070,   0.00002,     0.047 },
 { {   0,   8,  -3,   0,   0,   0,   0,   0,   2,   1,   0 },    198.00070,   0.00002,     0.021 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    152.41707,   0.00003,     0.038 },
 { {   0,   9, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    332.41707,   0.00003,     2.477 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    266.51281,   0.00002,     0.040 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     86.51281,   0.00002,     0.712 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,  -1,   0 },    263.11796,   0.00003,     0.073 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    251.66272,   0.00005,     2.701 },
 { {   0,   9, -15,   0,   0,   0,   0,   0,   0,   1,   0 },     83.11796,   0.00003,     0.078 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,  -1,   0 },    340.18131,   0.00002,     0.027 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    334.71042,   0.00003,     0.041 },
 { {   0,   9, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    160.18131,   0.00002,     0.092 },
 { {   0,   9, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    245.95360,   0.00008,     0.043 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    206.45037,   0.00006,     0.045 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     51.73111,   0.00003,     0.113 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    118.85852,   0.00002,     0.227 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -2,   0 },     67.56655,   0.00003,     0.042 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },     94.22108,   0.00002,     0.094 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    247.56655,   0.00003,     0.380 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    274.22108,   0.00002,     0.063 },
 { {   0,   9, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    100.05539,   0.00005,     0.037 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   0,   0 },     13.09662,   0.00003,     0.276 },
 { {   0,   9, -11,   0,   0,   0,   0,   0,   0,   2,   0 },     10.85526,   0.00002,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    321.03441,   0.00004,     0.047 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   1,   0,   1 },    270.06942,   0.00005,     0.033 },
 { {   0,   9, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    108.36241,   0.00041,     0.034 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,  -2,   0 },    343.96521,   0.00002,     0.022 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    355.02140,   0.00006,     0.052 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,  -2,   1,   0 },    343.96521,   0.00006,     0.171 },
 { {   0,   9,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    190.06176,   0.00007,     0.033 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },    112.67588,   0.00010,     0.055 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,  -2,   1,   0 },     71.29061,   0.00003,     0.206 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,  -1,   0 },    268.33270,   0.00002,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    239.05619,   0.00005,     0.151 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   0,   1,   0 },     88.33270,   0.00002,     0.050 },
 { {   0,   9,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    201.68370,   0.00005,     0.032 },
 { {   0,   9,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    349.63369,   0.00005,     0.131 },
 { {   0,   9,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     35.41598,   0.00008,     0.030 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,  -1,   0 },    109.16636,   0.00004,     0.047 },
 { {   0,   9,  -5,   0,   0,   0,   0,   0,   2,   0,   0 },    110.85481,   0.00006,     0.029 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   0,  -2,   0 },     18.09764,   0.00002,     0.063 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   0,   0,   0 },    198.09764,   0.00002,     0.094 },
 { {   0,   9,  -4,   0,   0,   0,   0,   0,   2,  -1,   0 },    193.52966,   0.00004,     0.045 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    153.01766,   0.00006,     0.039 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },    153.01766,   0.00003,     0.082 },
 { {   0,  10, -17,   0,   0,   0,   0,   0,  -2,   3,   0 },    333.01766,   0.00003,     0.070 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    265.94908,   0.00002,     0.041 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    265.94908,   0.00002,     0.493 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    251.68088,   0.00002,     0.038 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },     83.08485,   0.00004,     0.077 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    251.68088,   0.00002,     3.918 },
 { {   0,  10, -16,   0,   0,   0,   0,   0,   0,   1,   0 },    263.08485,   0.00004,     0.074 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.43644,   0.00052,     0.043 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,  -2,   1,   0 },    166.33979,   0.00003,     0.098 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -3,   0 },    160.14825,   0.00004,     0.073 },
 { {   0,  10, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    340.14825,   0.00004,     0.079 },
 { {   0,  10, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    241.69734,   0.00007,     0.044 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    216.07293,   0.00005,     0.047 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },     93.66456,   0.00005,     0.122 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    259.50215,   0.00005,     0.307 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -3,   0 },    105.64638,   0.00002,     0.085 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,  -1,   0 },    285.64638,   0.00002,     0.068 },
 { {   0,  10, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    100.71106,   0.00004,     0.036 },
 { {   0,  10, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    167.59915,   0.00004,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,  -1,   0,  -1 },    320.64435,   0.00004,     0.049 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.36714,   0.00003,     0.125 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    312.69421,   0.00004,     0.190 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   0,   1,   0 },    108.36714,   0.00003,     0.054 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,  -2,   1 },    270.05492,   0.00002,     0.221 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   1,   0,   1 },    270.05492,   0.00002,     0.032 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    108.36714,   0.00030,     0.033 },
 { {   0,  10, -11,   0,   0,   0,   0,   0,   2,   2,   0 },    108.36714,   0.00003,     0.018 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },      0.99848,   0.00004,     0.054 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   1,   0 },    344.04259,   0.00002,     0.191 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   2,   0 },      0.99848,   0.00004,     0.125 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,  -2,   3,   0 },    344.04259,   0.00002,     0.047 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,  -2,   0 },    186.35059,   0.00002,     0.049 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   0,   0,   0 },    186.35059,   0.00002,     0.160 },
 { {   0,  10, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    187.72468,   0.00008,     0.032 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -2,   0,   0 },    128.24552,   0.00009,     0.057 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   0,  -1 },    184.50989,   0.00002,     0.054 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,  -1,   2,  -1 },      4.50989,   0.00002,     0.125 },
 { {   0,  10,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    211.75282,   0.00006,     0.031 },
 { {   0,  10,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    355.76036,   0.00004,     0.121 },
 { {   0,  10,  -7,   0,   0,   0,   0,   0,   0,   0,   0 },    248.87951,   0.00005,     0.108 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,  -2,   0 },    110.50135,   0.00003,     0.118 },
 { {   0,  10,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },    110.50135,   0.00003,     0.029 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,  -1,   0 },    150.66272,   0.00021,     0.040 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,  -2,   0,   0 },    150.66272,   0.00013,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -3,   0 },    150.66272,   0.00004,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -2,   0 },    149.60900,   0.00073,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,  -1,   0 },    225.68043,   0.00030,  1911.695 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   0 },    330.17945,   0.00060,     0.075 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   0,   1,   0 },    330.66272,   0.00004,     0.038 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -2,   0 },    330.66272,   0.00013,     0.087 },
 { {   0,  10,  -3,   0,   0,   0,   0,   0,   2,  -1,   0 },    330.66272,   0.00021,     0.040 },
 { {   0,  11, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    153.57175,   0.00006,     0.040 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    160.11468,   0.00004,     0.020 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.11468,   0.00043,     0.044 },
 { {   0,  11, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },     24.32056,   0.00003,     0.075 },
 { {   0,  11, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    215.24606,   0.00006,     0.046 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    249.16598,   0.00003,     0.048 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   1,   0 },    130.25222,   0.00004,     0.132 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    169.64308,   0.00004,     0.177 },
 { {   0,  11, -14,   0,   0,   0,   0,   0,   0,   0,   0 },    271.40303,   0.00004,     0.258 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    109.23253,   0.00003,     0.050 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },    219.78808,   0.00003,     0.150 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,  -2,   0 },    350.21056,   0.00002,     0.046 },
 { {   0,  11, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    350.21056,   0.00002,     0.205 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    320.29914,   0.00004,     0.050 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,  -1,   0 },    288.36769,   0.00002,     0.136 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   0,   0 },    265.62925,   0.00003,     0.170 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   1,   0 },    108.36769,   0.00002,     0.052 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   0,   2,   0 },    294.03746,   0.00002,     0.031 },
 { {   0,  11, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    288.36769,   0.00005,     0.033 },
 { {   0,  11, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    173.71489,   0.00006,     0.032 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },    128.47738,   0.00009,     0.059 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   0,  -1 },    184.27845,   0.00002,     0.056 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,  -1,   2,  -1 },      4.27845,   0.00002,     0.116 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -3,   0 },    251.08531,   0.00002,     0.140 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,  -1,   0 },    251.08531,   0.00002,     0.052 },
 { {   0,  11, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    218.43504,   0.00005,     0.031 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   0,   0,   0 },    248.92857,   0.00002,     0.101 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   0,   2,   0 },    248.92857,   0.00002,     0.027 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,  -1,   0 },    244.55096,   0.00004,     0.047 },
 { {   0,  11,  -8,   0,   0,   0,   0,   0,   2,   0,   0 },    150.70057,   0.00003,     0.029 },
 { {   0,  11,  -6,   0,   0,   0,   0,   0,   2,   0,   0 },     17.96963,   0.00033,     0.027 },
 { {   0,  12, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.08282,   0.00006,     0.041 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,  -2,   1,   0 },     71.71015,   0.00007,     0.100 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },     71.71015,   0.00003,     0.085 },
 { {   0,  12, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    251.71015,   0.00003,     0.068 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    160.08053,   0.00033,     0.045 },
 { {   0,  12, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    340.09434,   0.00003,     0.071 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    175.24378,   0.00004,     0.047 },
 { {   0,  12, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    268.96262,   0.00002,     0.189 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    296.63525,   0.00003,     0.049 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    116.63525,   0.00003,     0.159 },
 { {   0,  12, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    271.42757,   0.00004,     0.222 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     86.85976,   0.00003,     0.052 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    266.85976,   0.00003,     0.137 },
 { {   0,  12, -14,   0,   0,   0,   0,   0,   2,  -1,   0 },    199.75658,   0.00004,     0.059 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    132.38533,   0.00002,     0.154 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   0,   2,   0 },    277.83874,   0.00003,     0.030 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,  -2,   0 },    282.59740,   0.00002,     0.211 },
 { {   0,  12, -13,   0,   0,   0,   0,   0,   2,   0,   0 },    107.79358,   0.00023,     0.032 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    162.56529,   0.00002,     0.174 },
 { {   0,  12, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    162.56529,   0.00002,     0.031 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    128.55369,   0.00005,     0.062 },
 { {   0,  12, -11,   0,   0,   0,   0,   0,  -2,   2,   0 },    308.55369,   0.00005,     0.097 },
 { {   0,  12, -10,   0,   0,   0,   0,   0,   2,   0,   0 },    148.26719,   0.00004,     0.029 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -3,   0 },    248.96811,   0.00002,     0.221 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -2,   0 },      9.65392,   0.00002,     0.115 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,  -1,   0 },    248.96811,   0.00002,     0.045 },
 { {   0,  12,  -9,   0,   0,   0,   0,   0,   2,   0,   0 },    189.65392,   0.00002,     0.028 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.89965,   0.00009,     0.076 },
 { {   0,  12,  -8,   0,   0,   0,   0,   0,  -2,   2,   0 },    205.02132,   0.00007,     0.075 },
 { {   0,  13, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.55432,   0.00005,     0.042 },
 { {   0,  13, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },     71.72125,   0.00005,     0.107 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    340.04560,   0.00005,     0.046 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   0,  -1,   0 },    340.04560,   0.00003,     0.099 },
 { {   0,  13, -18,   0,   0,   0,   0,   0,   0,   1,   0 },    160.04560,   0.00003,     0.061 },
 { {   0,  13, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    112.24974,   0.00005,     0.049 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    299.17423,   0.00003,     0.051 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    122.90866,   0.00002,     0.047 },
 { {   0,  13, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    302.90866,   0.00002,     0.195 },
 { {   0,  13, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },     86.53804,   0.00005,     0.054 },
 { {   0,  13, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    107.44398,   0.00024,     0.031 },
 { {   0,  13, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    165.24288,   0.00015,     0.060 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,  -1,   0 },    252.85225,   0.00003,     0.035 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   0,   0 },    239.06836,   0.00014,     0.064 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   1,   0 },     72.85225,   0.00003,     0.426 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -2,   2,   0 },    139.99228,   0.00002,     0.092 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },    183.90482,   0.00002,     0.060 },
 { {   0,  13, -12,   0,   0,   0,   0,   0,  -1,   2,  -1 },    183.90482,   0.00002,     0.102 },
 { {   0,  13, -11,   0,   0,   0,   0,   0,   2,   0,   0 },    160.11930,   0.00005,     0.029 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,  -2,   2 },    277.34580,   0.00003,     0.075 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   0 },     45.74307,   0.00005,     0.073 },
 { {   0,  13, -10,   0,   0,   0,   0,   0,  -2,   0,   2 },     97.34580,   0.00003,     0.076 },
 { {   0,  14, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    154.98959,   0.00005,     0.044 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,  -2,   0 },    340.00977,   0.00002,     0.021 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    340.04400,   0.00027,     0.048 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   1,   0 },    340.00977,   0.00004,     0.129 },
 { {   0,  14, -19,   0,   0,   0,   0,   0,  -2,   3,   0 },    160.00977,   0.00004,     0.053 },
 { {   0,  14, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     97.34976,   0.00006,     0.050 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    298.68478,   0.00004,     0.053 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    302.88466,   0.00002,     0.048 },
 { {   0,  14, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    302.88466,   0.00002,     0.174 },
 { {   0,  14, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },     87.91276,   0.00005,     0.056 },
 { {   0,  14, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    108.34717,   0.00005,     0.052 },
 { {   0,  14, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },    338.78730,   0.00005,     0.063 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    134.10769,   0.00004,     0.067 },
 { {   0,  14, -13,   0,   0,   0,   0,   0,  -2,   1,   0 },    252.84885,   0.00041,     0.580 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,  -2,   0 },    160.21118,   0.00002,     0.111 },
 { {   0,  14, -12,   0,   0,   0,   0,   0,   2,   0,   0 },    160.21118,   0.00002,     0.028 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,  -2,   0,   0 },    234.75876,   0.00006,     0.077 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   0,  -2,   0 },    125.18699,   0.00002,     0.068 },
 { {   0,  14, -11,   0,   0,   0,   0,   0,   0,   0,   0 },    305.18699,   0.00002,     0.085 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,   0,   0 },     17.91995,   0.00002,     0.073 },
 { {   0,  14,  -9,   0,   0,   0,   0,   0,   0,   2,   0 },    197.91995,   0.00002,     0.025 },
 { {   0,  15, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    155.39172,   0.00002,     0.045 },
 { {   0,  15, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    155.39172,   0.00002,     0.241 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    340.02046,   0.00019,     0.049 },
 { {   0,  15, -20,   0,   0,   0,   0,   0,  -2,   1,   0 },    339.97294,   0.00005,     0.141 },
 { {   0,  15, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     92.71295,   0.00005,     0.052 },
 { {   0,  15, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    312.13866,   0.00004,     0.054 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.24196,   0.00003,     0.058 },
 { {   0,  15, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },    251.24196,   0.00003,     0.109 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   0,   0,   0 },    103.47443,   0.00005,     0.119 },
 { {   0,  15, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },    108.33374,   0.00004,     0.050 },
 { {   0,  15, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    341.95366,   0.00007,     0.065 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   0,   0 },     95.25011,   0.00003,     0.070 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -2,   2,   0 },    275.25011,   0.00003,     0.082 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    183.61347,   0.00002,     0.065 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,  -1,   2,  -1 },      3.61347,   0.00002,     0.090 },
 { {   0,  15, -14,   0,   0,   0,   0,   0,   2,   0,   0 },    113.25401,   0.00003,     0.028 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   0,   0 },    165.17956,   0.00004,     0.075 },
 { {   0,  15, -13,   0,   0,   0,   0,   0,  -2,   2,   0 },     68.64291,   0.00003,     0.076 },
 { {   0,  15, -12,   0,   0,   0,   0,   0,  -1,   0,  -1 },     85.14718,   0.00004,     0.075 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    339.99199,   0.00017,     0.051 },
 { {   0,  16, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    159.93506,   0.00007,     0.200 },
 { {   0,  16, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },    118.71101,   0.00004,     0.053 },
 { {   0,  16, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    299.10866,   0.00004,     0.056 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     94.66008,   0.00002,     0.060 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    274.66008,   0.00002,     0.102 },
 { {   0,  16, -18,   0,   0,   0,   0,   0,   2,   0,   0 },     19.92510,   0.00005,     0.031 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.58567,   0.00002,     0.064 },
 { {   0,  16, -17,   0,   0,   0,   0,   0,  -2,   2,   0 },     85.58567,   0.00002,     0.093 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    338.32582,   0.00008,     0.068 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -2,   2,   0 },    204.34708,   0.00002,     0.085 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -1,   0,  -1 },    202.73917,   0.00002,     0.063 },
 { {   0,  16, -16,   0,   0,   0,   0,   0,  -1,   2,  -1 },    202.73917,   0.00002,     0.093 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    221.67568,   0.00002,     0.073 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },    221.67568,   0.00002,     0.078 },
 { {   0,  16, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     75.69663,   0.00005,     0.028 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,  -1 },    181.20994,   0.00005,     0.073 },
 { {   0,  16, -14,   0,   0,   0,   0,   0,  -1,   0,   1 },    181.20994,   0.00005,     0.077 },
 { {   0,  16, -13,   0,   0,   0,   0,   0,   0,   0,   0 },     77.77600,   0.00009,     0.077 },
 { {   0,  16, -11,   0,   0,   0,   0,   0,   2,  -2,   0 },    193.42826,   0.00002,     0.076 },
 { {   0,  16, -11,   0,   0,   0,   0,   0,   2,   0,   0 },     13.42826,   0.00002,     0.025 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    166.14107,   0.00002,     0.052 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    346.14107,   0.00002,     0.135 },
 { {   0,  17, -22,   0,   0,   0,   0,   0,   0,   0,   0 },    159.89605,   0.00006,     0.178 },
 { {   0,  17, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    119.89321,   0.00004,     0.055 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     59.83316,   0.00002,     0.062 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,  -2,   2,   0 },     59.83316,   0.00002,     0.096 },
 { {   0,  17, -19,   0,   0,   0,   0,   0,   2,   0,   0 },     19.95514,   0.00004,     0.030 },
 { {   0,  17, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },     85.14548,   0.00005,     0.066 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.58760,   0.00004,     0.071 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   0,  -1 },    202.76350,   0.00002,     0.066 },
 { {   0,  17, -17,   0,   0,   0,   0,   0,  -1,   2,  -1 },     22.76350,   0.00002,     0.088 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    252.81868,   0.00002,     0.025 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    252.81868,   0.00002,     0.076 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,  -2,   1 },     56.51983,   0.00003,     0.072 },
 { {   0,  17, -16,   0,   0,   0,   0,   0,  -1,   0,   1 },    236.51983,   0.00003,     0.079 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    166.10545,   0.00004,     0.054 },
 { {   0,  18, -23,   0,   0,   0,   0,   0,   0,   0,   0 },    159.85589,   0.00005,     0.160 },
 { {   0,  18, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    132.02196,   0.00003,     0.057 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   0,   0 },     60.18205,   0.00002,     0.065 },
 { {   0,  18, -20,   0,   0,   0,   0,   0,  -2,   2,   0 },     60.18205,   0.00002,     0.091 },
 { {   0,  18, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     71.01268,   0.00004,     0.069 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    215.62982,   0.00002,     0.074 },
 { {   0,  18, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },     35.62982,   0.00002,     0.077 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,  -1,   0 },    127.50243,   0.00004,     0.039 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    321.29291,   0.00005,     0.080 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -2,   1,   0 },     72.80257,   0.00038,     1.286 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,  -2,   1 },     79.59785,   0.00002,     0.076 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,  -1,   0,   1 },     79.59785,   0.00002,     0.075 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },    127.50243,   0.00006,     0.070 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   0,   0,   0 },    127.50243,   0.00008,     0.082 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -2,   0 },    127.50243,   0.00009,     0.095 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,  -1,   0 },    127.50243,   0.00025,     0.042 },
 { {   0,  18, -17,   0,   0,   0,   0,   0,   2,   0,   0 },    127.50243,   0.00002,     0.027 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,  -1,   0 },    204.56550,   0.00003,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   0,   0 },    204.56550,   0.00006,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -4,   1,   0 },    204.56550,   0.00003,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -3,   0 },    204.56550,   0.00003,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -2,   0 },    204.56550,   0.00032,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,  -1,   0 },    204.56550,   0.00364,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -2,   0,   0 },    204.58914,   0.00241,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,  -1,  -1,   0 },     24.56550,   0.00007,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -4,   0 },    204.56550,   0.00004,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -3,   0 },    204.45391,   0.00072,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   0 },    204.55406,   0.01302,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -2,   2 },    204.56550,   0.00005,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,  -1,   0 },    351.08799,   0.00055,   273.043 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,  -2 },     24.56550,   0.00005,     0.074 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   0,   0 },     24.79524,   0.01225,     0.075 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   1,   0 },     24.68404,   0.00068,     0.038 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   0,   2,   0 },     24.56550,   0.00004,     0.025 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   1,  -1,   0 },    204.56550,   0.00007,     0.081 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -2,   0 },     24.57709,   0.00226,     0.087 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,  -1,   0 },     24.56550,   0.00364,     0.040 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   0,   0 },     24.56550,   0.00032,     0.026 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   2,   1,   0 },     24.56550,   0.00003,     0.020 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -3,   0 },     24.56550,   0.00003,     0.044 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -2,   0 },     24.56550,   0.00006,     0.028 },
 { {   0,  18, -16,   0,   0,   0,   0,   0,   4,  -1,   0 },     24.56550,   0.00003,     0.020 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -2,   0 },    101.62857,   0.00002,     0.027 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,  -1,   0 },    101.62857,   0.00025,     0.042 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   0,   0 },    102.85686,   0.00007,     0.095 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,  -2,   2,   0 },     98.41031,   0.00003,     0.062 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,  -2,   0 },    101.62857,   0.00008,     0.082 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   0,   0,   0 },    101.62857,   0.00006,     0.070 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   1,   0,  -1 },     44.23071,   0.00004,     0.076 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,  -1,   0 },    101.62857,   0.00004,     0.039 },
 { {   0,  18, -15,   0,   0,   0,   0,   0,   2,   0,   0 },     69.06073,   0.00004,     0.026 },
 { {   0,  19, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    336.65655,   0.00003,     0.050 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    166.06814,   0.00003,     0.056 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    159.81455,   0.00002,     0.051 },
 { {   0,  19, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    159.81455,   0.00002,     0.145 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    159.11865,   0.00004,     0.059 },
 { {   0,  19, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     71.47333,   0.00002,     0.104 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },     86.47166,   0.00005,     0.067 },
 { {   0,  19, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    167.64907,   0.00002,     0.086 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },     71.43673,   0.00002,     0.064 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    251.43673,   0.00002,     0.092 },
 { {   0,  19, -20,   0,   0,   0,   0,   0,   2,   0,   0 },    288.25192,   0.00005,     0.028 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },     55.14940,   0.00008,     0.078 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },    330.39388,   0.00004,     0.072 },
 { {   0,  19, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    333.79318,   0.00006,     0.084 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -2,   0,   0 },    339.92076,   0.00007,     0.084 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,  -2,   2,   0 },    304.23278,   0.00003,     0.068 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,  -2,   0 },    117.35099,   0.00002,     0.073 },
 { {   0,  19, -18,   0,   0,   0,   0,   0,   0,   0,   0 },     29.22150,   0.00004,     0.078 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,  -2,   0,   0 },    206.47472,   0.00005,     0.092 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,  -2,   0 },      8.40503,   0.00002,     0.079 },
 { {   0,  19, -17,   0,   0,   0,   0,   0,   0,   0,   0 },      8.40503,   0.00002,     0.072 },
 { {   0,  20, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    159.55412,   0.00004,     0.062 },
 { {   0,  20, -24,   0,   0,   0,   0,   0,  -2,   2,   0 },     71.66850,   0.00002,     0.097 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    225.85091,   0.00002,     0.070 },
 { {   0,  20, -22,   0,   0,   0,   0,   0,  -2,   2,   0 },    225.85091,   0.00002,     0.081 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   0,   0 },    204.20741,   0.00013,     0.076 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   1,   0 },    231.73897,   0.00033,    34.812 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -2,   2,   0 },    322.76307,   0.00008,     0.075 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,  -1,   0,  -1 },    318.30622,   0.00004,     0.070 },
 { {   0,  20, -21,   0,   0,   0,   0,   0,   2,   0,   0 },    288.22516,   0.00004,     0.028 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   0,  -1 },     27.35671,   0.00004,     0.075 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,  -1,   2,  -1 },    207.35671,   0.00004,     0.076 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,  -2,   0 },    216.47465,   0.00002,     0.071 },
 { {   0,  20, -20,   0,   0,   0,   0,   0,   0,   0,   0 },    216.47465,   0.00002,     0.080 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -2,   0,   0 },    350.78781,   0.00003,     0.089 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,  -1,   0,  -1 },     55.60871,   0.00005,     0.081 },
 { {   0,  20, -19,   0,   0,   0,   0,   0,   0,   0,   0 },    238.77139,   0.00005,     0.074 },
 { {   0,  21, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    339.82269,   0.00003,     0.123 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    153.12422,   0.00003,     0.064 },
 { {   0,  21, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },     43.90455,   0.00003,     0.092 },
 { {   0,  21, -24,   0,   0,   0,   0,   0,  -2,   0,   0 },    291.38573,   0.00003,     0.068 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -2,   0,   0 },    265.71080,   0.00005,     0.079 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   0,  -1 },     17.14749,   0.00002,     0.073 },
 { {   0,  21, -22,   0,   0,   0,   0,   0,  -1,   2,  -1 },    197.14749,   0.00002,     0.078 },
 { {   0,  21, -21,   0,   0,   0,   0,   0,   0,   0,   0 },    155.73792,   0.00004,     0.076 },
 { {   0,  21, -20,   0,   0,   0,   0,   0,   1,   0,  -1 },     63.47623,   0.00004,     0.076 },
 { {   0,  22, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.82903,   0.00003,     0.067 },
 { {   0,  22, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    182.09559,   0.00006,     0.072 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   0,  -1 },    157.15316,   0.00005,     0.077 },
 { {   0,  22, -23,   0,   0,   0,   0,   0,  -1,   2,  -1 },    275.79970,   0.00002,     0.074 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   0,  -2,   0 },    244.36297,   0.00003,     0.085 },
 { {   0,  22, -21,   0,   0,   0,   0,   0,   0,   0,   0 },     64.36297,   0.00003,     0.068 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   0,   0 },    169.62777,   0.00002,     0.075 },
 { {   0,  23, -26,   0,   0,   0,   0,   0,  -2,   2,   0 },    169.62777,   0.00002,     0.076 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   0,   0 },    313.33115,   0.00003,     0.081 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -2,   2,   0 },    313.33115,   0.00003,     0.071 },
 { {   0,  23, -25,   0,   0,   0,   0,   0,  -1,   0,  -1 },     84.08940,   0.00005,     0.075 },
 { {   0,  23, -24,   0,   0,   0,   0,   0,   0,   0,   0 },     97.38663,   0.00007,     0.075 },
 { {   0,  23, -23,   0,   0,   0,   0,   0,   1,   0,  -1 },    203.56270,   0.00004,     0.075 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   0,   0 },    259.16347,   0.00002,     0.073 },
 { {   0,  24, -28,   0,   0,   0,   0,   0,  -2,   2,   0 },    259.16347,   0.00002,     0.078 },
 { {   0,  24, -27,   0,   0,   0,   0,   0,  -2,   0,   0 },      6.81221,   0.00005,     0.079 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,  -2,   0 },    265.30916,   0.00002,     0.074 },
 { {   0,  24, -26,   0,   0,   0,   0,   0,   0,   0,   0 },    265.30916,   0.00002,     0.077 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,  -2,   0 },    344.48060,   0.00003,     0.087 },
 { {   0,  24, -24,   0,   0,   0,   0,   0,   0,   0,   0 },    344.48060,   0.00003,     0.067 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,  -2,   0,   0 },    250.63022,   0.00002,     0.115 },
 { {   0,  24, -23,   0,   0,   0,   0,   0,  -2,   2,   0 },     70.63022,   0.00002,     0.056 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   0,   0 },    165.87580,   0.00002,     0.071 },
 { {   0,  25, -30,   0,   0,   0,   0,   0,  -2,   2,   0 },    345.87580,   0.00002,     0.081 },
 { {   0,  25, -29,   0,   0,   0,   0,   0,  -2,   0,   0 },    192.20447,   0.00005,     0.076 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,  -2,   0 },    307.92315,   0.00002,     0.072 },
 { {   0,  25, -28,   0,   0,   0,   0,   0,   0,   0,   0 },    307.92315,   0.00002,     0.079 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   0,   0 },    276.43390,   0.00003,     0.074 },
 { {   0,  26, -31,   0,   0,   0,   0,   0,  -2,   2,   0 },    178.06691,   0.00002,     0.077 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,  -2,   0 },    352.00671,   0.00003,     0.075 },
 { {   0,  26, -29,   0,   0,   0,   0,   0,   0,   0,   0 },    172.00671,   0.00003,     0.075 },
 { {   0,  26, -28,   0,   0,   0,   0,   0,   1,   0,  -1 },    109.48589,   0.00004,     0.076 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   0,   0 },    121.10728,   0.00002,     0.078 },
 { {   0,  27, -32,   0,   0,   0,   0,   0,  -2,   2,   0 },    121.10728,   0.00002,     0.073 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,  -2,   0 },    159.38922,   0.00002,     0.071 },
 { {   0,  28, -33,   0,   0,   0,   0,   0,   0,   0,   0 },    159.38922,   0.00002,     0.080 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,  -2,   0,   0 },    158.70547,   0.00006,     0.086 },
 { {   0,  29, -34,   0,   0,   0,   0,   0,  -2,   2,   0 },    341.19960,   0.00002,     0.067 },
 { {   0,  30, -36,   0,   0,   0,   0,   0,  -2,   0,   0 },    251.61633,   0.00005,     0.084 },
 { {   0,  31, -39,   0,   0,   0,   0,   0,  -2,   0,   0 },    244.72789,   0.00004,     0.075 },
 { {   0,  31, -39,   0,   0,   0,   0,   0,  -2,   2,   0 },     64.72789,   0.00004,     0.076 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,  -2,   0 },    251.59648,   0.00004,     0.076 },
 { {   0,  31, -37,   0,   0,   0,   0,   0,   0,   0,   0 },     71.59648,   0.00004,     0.075 },
 { {   0,  34, -42,   0,   0,   0,   0,   0,   0,   0,   0 },    245.43224,   0.00006,     0.075 },
 { {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0 },     63.34872,   0.00005,     0.075 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 170> g_ELP16
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 },    180.00031,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },    359.99968,   0.00012,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    180.00567,   0.00963,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   2 },      0.00000,   0.00002,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    180.00006,   0.00065,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0 },    179.99986,   0.00005,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    153.73567,   0.00009,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    154.27480,   0.00135,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    155.76630,   0.00661,     1.000 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },    154.62845,   0.00120,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0 },    153.96248,   0.00008,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },    336.25846,   0.00002,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    154.00745,   0.00008,     0.500 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    155.89560,   0.00001,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0 },     24.15517,   0.00001,     0.096 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   0 },     24.15511,   0.00003,     0.530 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },     24.08286,   0.00031,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0 },     23.35231,   0.00002,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },      0.00000,   0.00001,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },    346.96809,   0.00002,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,   0 },    336.31637,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0 },    165.16990,   0.00018,     8.850 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0 },    335.78680,   0.00159,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   0 },    335.84492,   0.00011,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0 },    210.91539,   0.00003,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   0 },     40.76689,   0.00017,     0.095 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0 },    198.78197,   0.00017,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,   0 },    182.66055,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0 },    180.00152,   0.00002,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0 },    179.99919,   0.00018,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -2 },    359.99976,   0.00002,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },    180.07400,   0.00181,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2 },    359.99982,   0.00015,     0.474 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },    179.70982,   0.00043,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },    179.99999,   0.00011,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,   0 },    350.80827,   0.00001,     1.292 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   0 },    335.90413,   0.00056,     0.080 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -2 },    335.92813,   0.00001,     0.903 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   0 },    155.91133,   0.00007,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   0 },    155.90587,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -2,   0 },     11.98593,   0.00002,     4.425 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },    156.00589,   0.00024,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,  -2 },    180.00000,   0.00004,     9.300 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,   0 },    156.00396,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -1,   0 },    334.71131,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,   0 },    335.83848,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0 },    180.00003,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,   0 },    180.00021,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,  -1,   1 },    275.13149,   0.00001,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   1,  -1 },    275.13100,   0.00006,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   1,   1 },     95.13240,   0.00004,     0.106 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   2,  -1 },    275.13354,   0.00010,     0.043 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0 },     77.00007,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   1 },    275.13228,   0.00004,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0 },     77.05643,   0.00004,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,  -1,   1 },     95.13228,   0.00001,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,  -1 },     95.12539,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   1 },    275.13214,   0.00023,     0.949 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   1,  -1 },     95.13827,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   2,  -1 },    275.13355,   0.00002,     1.388 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -2,  -1 },     95.13543,   0.00008,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -2,   1 },    275.13140,   0.00011,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,  -1 },     95.13608,   0.00053,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   1 },    275.13151,   0.00146,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,  -1 },     95.14482,   0.00080,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1 },     95.13222,   0.00851,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   3 },    275.13231,   0.00002,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,  -1 },    275.13348,   0.00772,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   1 },     95.13089,   0.00006,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2,  -1 },    275.13413,   0.00094,     0.581 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2,   1 },    275.13215,   0.00008,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   3,  -1 },    275.13430,   0.00010,     0.067 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   3,   1 },    275.13228,   0.00001,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,  -1 },     95.13568,   0.00004,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,   1 },    275.13157,   0.00008,     0.098 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,  -1 },     95.13903,   0.00010,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   1 },     95.13223,   0.00033,     0.327 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   1,  -1 },    275.13356,   0.00033,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   2,  -1 },    275.13422,   0.00004,     0.368 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,  -1 },    275.13807,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },     76.15482,   0.00003,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },     95.13038,   0.00015,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },     72.38269,   0.00009,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },    275.13220,   0.00009,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1 },     95.13339,   0.00007,     1.200 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },     75.53411,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,  -1 },     95.13373,   0.00001,     0.071 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -2,  -1 },    275.13501,   0.00003,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,  -1 },    275.13540,   0.00022,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,   1 },     95.12714,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,  -1 },     95.13243,   0.00041,     0.949 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,   1 },     95.13161,   0.00001,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1,  -1 },     95.13406,   0.00052,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1,   1 },     95.13228,   0.00005,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   2,  -1 },     95.13425,   0.00006,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -4,  -1 },     95.13456,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -3,  -1 },     95.13463,   0.00029,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -3,   1 },    275.13107,   0.00002,     0.077 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -3 },    275.13460,   0.00003,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -1 },     95.13465,   0.00410,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,   1 },    275.13112,   0.00042,     3.575 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -3 },    275.13461,   0.00018,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -1 },     95.13459,   0.06174,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   0 },     77.02181,   0.00001,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   1 },     95.13157,   0.00858,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   3 },    275.13186,   0.00003,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -3 },    275.13472,   0.00047,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -1 },     95.13523,   0.28938,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   1 },     95.13225,   0.06154,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   3 },    275.13228,   0.00013,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -3 },     95.13641,   0.00003,     0.073 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -1 },     95.13453,   0.06245,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   1 },     95.13221,   0.00661,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   3 },    275.13228,   0.00003,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,  -1 },     95.13449,   0.00432,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,   1 },     95.13217,   0.00057,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   3,  -1 },     95.13450,   0.00031,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   3,   1 },     95.13220,   0.00005,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   4,  -1 },     95.13456,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -2,  -1 },     95.13498,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1,  -1 },     95.13560,   0.00028,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,  -1 },    275.13212,   0.00046,     1.057 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,   1 },    275.13248,   0.00006,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   1,  -1 },    275.13400,   0.00039,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   1,   1 },    275.13228,   0.00004,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   2,  -1 },    275.13424,   0.00005,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,  -1 },    275.13319,   0.00031,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,   1 },    275.13228,   0.00004,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   1,  -1 },    275.13381,   0.00005,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   1,   0,  -1 },     95.13229,   0.00005,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -2,   0,  -1 },     95.13429,   0.00002,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -2,  -1 },    275.13445,   0.00003,     0.354 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -1,  -1 },    275.13432,   0.00022,     0.096 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -1,   1 },     95.13228,   0.00006,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   0,  -1 },     95.13417,   0.00036,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   0,   1 },     95.13228,   0.00006,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   1,  -1 },     95.13427,   0.00008,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   1,   1 },     95.13228,   0.00001,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -3,  -1 },    275.13456,   0.00008,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,  -1 },    275.13455,   0.00079,     0.547 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,   1 },     95.13166,   0.00011,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,  -1 },    275.13441,   0.00465,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   1 },     95.13223,   0.00143,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -3 },     95.13505,   0.00003,     0.463 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -1 },     95.13395,   0.00582,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   1 },     95.13223,   0.00084,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,  -3 },    275.13558,   0.00001,     0.090 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,  -1 },     95.13416,   0.00106,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,   1 },     95.13218,   0.00014,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   2,  -1 },     95.13424,   0.00012,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   2,   1 },     95.13228,   0.00002,     0.013 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   3,  -1 },     95.13428,   0.00001,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,  -1,  -1 },     95.13319,   0.00007,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,  -1,   1 },    275.13228,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,   0,  -1 },    275.13349,   0.00008,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,   1,  -1 },    275.13385,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -2,  -1 },    275.13439,   0.00006,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -2,   1 },     95.13228,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -1,  -1 },     95.13346,   0.00007,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -1,   1 },     95.13228,   0.00003,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   0,  -1 },     95.13387,   0.00007,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   0,   1 },     95.13228,   0.00001,     0.013 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   1,  -1 },     95.13402,   0.00002,     0.016 },
 { {   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0,   0 },    234.68628,   0.00001,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,  -2,   0,   1,   0 },    234.68866,   0.00002,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    234.67572,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   0,   0 },    234.68884,   0.00014,   238.919 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    234.67574,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   2,   0,  -1,   0 },    234.68887,   0.00002,     0.087 },
 { {   0,   8, -13,   0,   0,   0,   0,   2,   0,   0,   0 },    234.68643,   0.00001,     0.040 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 150> g_ELP17
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 },      0.00002,   0.00068,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1 },    180.00000,   0.00041,     5.997 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1 },    180.00001,   0.00039,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1 },    180.00134,   0.00001,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   1 },    180.00003,   0.00005,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1 },    155.90141,   0.00007,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },    155.89388,   0.00006,     0.857 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },    155.96862,   0.00006,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1 },    155.96726,   0.00006,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1 },    155.89029,   0.00005,     1.200 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   1 },    155.89864,   0.00006,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,  -1 },     24.15518,   0.00001,     0.487 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   1 },     24.15503,   0.00001,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   1 },    168.91635,   0.00001,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,  -1 },    335.84475,   0.00007,    18.600 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   1 },    335.84494,   0.00007,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   1 },     24.07078,   0.00001,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,  -1 },     25.70742,   0.00006,     0.097 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -1 },    180.00014,   0.00001,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -1 },    179.99989,   0.00011,     0.515 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   1 },    180.00001,   0.00007,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1 },    359.99976,   0.00057,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   1 },    179.99998,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,  -1 },    359.99956,   0.00002,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,  -1 },    335.90294,   0.00002,     1.063 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   1 },    335.89944,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -1 },    335.90598,   0.00006,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,  -1 },    156.00424,   0.00001,    16.885 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   1 },    156.00393,   0.00001,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,  -1 },    335.58947,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,  -1,   1,   0 },    275.13228,   0.00001,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,  -1,   0 },     95.13228,   0.00004,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,   0 },     95.13231,   0.00029,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,   2 },     95.13225,   0.00001,     0.108 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   1,   0 },     95.13223,   0.00050,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   2,   0 },     95.13217,   0.00004,     0.104 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   1,   0,   0 },     95.13228,   0.00003,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   1,   1,   0 },     95.13228,   0.00004,     0.046 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0 },    275.13228,   0.00003,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   1 },     77.05290,   0.00003,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0 },    275.13228,   0.00002,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0 },    275.13105,   0.00002,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,  -1,   0 },    275.13228,   0.00006,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   0 },    275.13221,   0.00096,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   2 },    275.13232,   0.00001,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   1,   0 },    275.13228,   0.00006,     1.127 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -3,   0 },     95.13228,   0.00001,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -2,   0 },     95.13251,   0.00017,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,  -2 },     95.13530,   0.00006,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0 },     95.13236,   0.00264,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   2 },    275.13175,   0.00008,     0.450 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,  -2 },     95.13684,   0.00014,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0 },     95.13226,   0.04834,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   2 },     95.13225,   0.00050,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,  -2 },    275.13286,   0.00024,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0 },     95.13224,   0.01209,     0.530 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   2 },     95.13229,   0.00007,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2,   0 },     95.13216,   0.00108,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   3,   0 },     95.13204,   0.00010,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   4,   0 },     95.13228,   0.00001,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,   0 },     95.13239,   0.00014,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,  -2 },     95.13641,   0.00001,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0 },     95.13228,   0.00227,     0.096 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   2 },     95.13230,   0.00002,     0.061 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   1,  -2 },    275.13287,   0.00001,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   1,   0 },     95.13223,   0.00054,     0.346 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   2,   0 },     95.13217,   0.00005,     0.062 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   2,   0,   0 },     95.13228,   0.00008,     0.107 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   2,   1,   0 },     95.13228,   0.00002,     0.257 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0 },     95.13228,   0.00006, 99999.999 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    275.13228,   0.00004,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },    257.75566,   0.00002,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 },    275.13228,   0.00035,     1.000 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },    257.75566,   0.00002,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    275.13228,   0.00003,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -2,   0 },    275.13228,   0.00002,     0.071 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,  -2 },    275.13552,   0.00001,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,   0 },    275.13237,   0.00038,     1.127 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,  -2 },     95.13134,   0.00001,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,   0 },    275.13175,   0.00044,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1,  -2 },     95.13401,   0.00002,     0.819 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1,   0 },    275.13214,   0.00048,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   2,   0 },    275.13216,   0.00006,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -3,  -2 },     95.13461,   0.00003,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -3,   0 },     95.13212,   0.00015,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -2 },     95.13464,   0.00037,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0 },     95.13209,   0.00294,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,   2 },    275.13183,   0.00004,     0.073 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -2 },     95.13467,   0.00325,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   0 },     95.13234,   0.07755,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   2 },     95.13192,   0.00075,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -4 },    275.13485,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -2 },     95.13510,   0.00917,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0 },    275.13226,   1.37497,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   2 },     95.13223,   0.00142,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -2 },     95.13472,   0.00198,     4.535 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   0 },    275.13217,   0.07245,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   2 },     95.13223,   0.00023,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,  -2 },     95.13490,   0.00013,     0.077 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,   0 },    275.13207,   0.00427,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,   2 },     95.13228,   0.00003,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   3,  -2 },     95.13477,   0.00001,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   3,   0 },    275.13197,   0.00026,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   4,   0 },    275.13203,   0.00002,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -2,   0 },     95.13228,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1,  -2 },     95.13550,   0.00002,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1,   0 },     95.13231,   0.00044,     0.898 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0 },     95.13293,   0.00037,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   1,  -2 },    275.13410,   0.00001,     1.283 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   1,   0 },     95.13220,   0.00037,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   2,   0 },     95.13216,   0.00004,     0.024 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,  -1,   0 },    275.13228,   0.00001,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,  -1 },     77.14857,   0.00001,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0 },     95.13223,   0.00039,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   1,   0 },     95.13228,   0.00005,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   1,   0,   0 },    275.13228,   0.00006,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -2,  -1,   0 },    275.13228,   0.00002,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -2,   0,   0 },    275.13228,   0.00003,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -1,  -2 },    275.13446,   0.00001,     0.334 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -1,   0 },    275.13228,   0.00067,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   0,  -2 },     95.13544,   0.00003,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   0,   0 },    275.13220,   0.00058,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   1,   0 },    275.13210,   0.00008,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -3,  -2 },    275.13456,   0.00001,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -3,   0 },    275.13228,   0.00002,     0.602 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,  -2 },    275.13457,   0.00007,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,   0 },     95.13267,   0.00012,     0.086 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,   2 },     95.13191,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,  -2 },    275.13456,   0.00026,     0.501 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   0 },    275.13228,   0.01504,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   2 },     95.13228,   0.00006,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -2 },     95.13571,   0.00050,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   0 },    275.13220,   0.00849,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   2 },     95.13228,   0.00003,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,  -2 },     95.13446,   0.00014,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,   0 },    275.13210,   0.00105,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   2,  -2 },     95.13433,   0.00002,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   2,   0 },    275.13200,   0.00010,     0.015 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   3,   0 },    275.13228,   0.00001,     0.013 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,  -1,   0 },     95.13234,   0.00010,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,   0,   0 },     95.13220,   0.00009,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,   1,   0 },     95.13228,   0.00002,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,   4,   0,  -1,   0 },     95.13228,   0.00002,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,  -1,  -2,   0 },    275.13228,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,  -1,  -1,   0 },    275.13228,   0.00003,     0.021 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,  -1,   0,   0 },    275.13228,   0.00001,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -2,   0 },    275.13231,   0.00018,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -1,   0 },    275.13223,   0.00022,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   0,   0 },    275.13214,   0.00008,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   1,   0 },    275.13228,   0.00002,     0.013 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 114> g_ELP18
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    270.00000,   0.02702, 99999.999 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 },    270.00030,   0.00004,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },     89.99962,   0.00010,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },     89.99752,   0.01035,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },     90.00007,   0.00053,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0 },     89.99979,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },    245.88710,   0.00007,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },    244.23379,   0.00126,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },     64.42940,   0.00112,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0 },     65.90169,   0.00006,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   0 },    114.15507,   0.00003,     0.530 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },    294.14183,   0.00025,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },    261.05217,   0.00002,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,   0 },     66.29236,   0.00003,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0 },     65.77739,   0.00008,     8.850 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0 },    245.78631,   0.00147,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   0 },    245.84495,   0.00008,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0 },    114.06475,   0.00003,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   0 },    301.71251,   0.00029,     0.095 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0 },    105.06390,   0.00022,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0 },    270.00021,   0.00022,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -2 },     89.99975,   0.00002,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },     90.09472,   0.00166,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2 },     89.99974,   0.00002,     0.474 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },     89.73046,   0.00057,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },     90.00003,   0.00007,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,   0 },     65.87675,   0.00004,     1.292 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   0 },    246.39267,   0.00051,     0.080 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   0 },     63.77987,   0.00010,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },     66.00391,   0.00023,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0 },     89.99926,   0.00002,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,   0 },     90.00001,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,  -1 },    185.13603,   0.00003,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,   1 },    185.13216,   0.00008,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   1,  -1 },      5.13227,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   1,   1 },    185.13251,   0.00004,     0.106 },
 { {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   2,  -1 },      5.13352,   0.00007,     0.043 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0 },    166.99439,   0.00004,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   1 },      5.13228,   0.00003,     0.097 },
 { {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0 },    167.07345,   0.00002,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,  -1 },      5.13666,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   1 },    185.13228,   0.00002,     0.949 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -2,  -1 },    185.13470,   0.00007,     0.019 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -2,   1 },      5.13142,   0.00007,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,  -1 },    185.13498,   0.00066,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   1 },      5.13138,   0.00101,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,  -1 },    185.13603,   0.00328,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1 },      5.13237,   0.00149,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,  -1 },      5.13343,   0.00648,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   1 },      5.13217,   0.00133,     0.065 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2,  -1 },    185.13414,   0.00110,     0.581 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   2,   1 },      5.13228,   0.00007,     0.035 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   3,  -1 },    185.13434,   0.00010,     0.067 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,  -1 },    185.13505,   0.00005,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,   1 },      5.13142,   0.00005,     0.098 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,  -1 },    185.13610,   0.00022,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   1 },      5.13228,   0.00010,     0.327 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   1,  -1 },      5.13354,   0.00025,     0.095 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   1,   1 },      5.13228,   0.00005,     0.061 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   2,  -1 },    185.13433,   0.00004,     0.368 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0 },    165.96905,   0.00002,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1 },    185.13043,   0.00014,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1 },    185.13228,   0.00012,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1 },      5.13334,   0.00003,     1.200 },
 { {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0 },    345.96905,   0.00002,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -2,  -1 },      5.13493,   0.00003,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,  -1 },      5.13544,   0.00021,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,  -1,   1 },    185.13386,   0.00002,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,  -1 },    185.13227,   0.00004,     0.949 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,   1 },    185.13228,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   1,  -1 },      5.13404,   0.00046,     0.082 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,  -1,   2,  -1 },      5.13427,   0.00004,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -3,  -1 },    185.13456,   0.00019,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -1 },    185.13459,   0.00311,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,   1 },    185.13193,   0.00082,     3.575 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -1 },    185.13459,   0.05703,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   1 },      5.13190,   0.01447,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -1 },    185.12989,   0.00183,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   1 },    185.13309,   0.00035,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -3 },    185.13504,   0.00010,     0.073 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -1 },      5.13453,   0.05765,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,  -1 },      5.13453,   0.00314,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   3,  -1 },      5.13456,   0.00019,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -2,  -1 },    185.13494,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1,  -1 },    185.13559,   0.00025,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,  -1,   1 },      5.13304,   0.00004,     0.069 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,  -1 },    185.13227,   0.00004,     1.057 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   0,   1 },    185.13228,   0.00002,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   1,  -1 },    185.13402,   0.00037,     0.070 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   1,   2,  -1 },    185.13428,   0.00004,     0.036 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,  -1 },    185.13319,   0.00027,     0.081 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   0,   1,  -1 },    185.13387,   0.00003,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   2,   1,   0,  -1 },      5.13229,   0.00004,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -2,  -1 },      5.13433,   0.00004,     0.354 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,  -1,  -1 },    185.13438,   0.00015,     0.096 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   0,  -1 },      5.13469,   0.00035,     0.042 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,  -1,   1,  -1 },      5.13457,   0.00006,     0.027 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -3,  -1 },      5.13451,   0.00009,     0.066 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,  -1 },      5.13456,   0.00091,     0.547 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -2,   1 },      5.13187,   0.00014,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,  -1 },    185.13456,   0.00351,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   1 },      5.13183,   0.00011,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -1 },      5.13469,   0.00503,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   1,  -1 },      5.13462,   0.00076,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   2,  -1 },      5.13470,   0.00007,     0.020 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,  -1,  -1 },      5.13321,   0.00005,     0.080 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   1,   0,  -1 },    185.13451,   0.00006,     0.039 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -2,  -1 },    185.13445,   0.00004,     0.044 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,  -1,  -1 },      5.13227,   0.00002,     0.028 },
 { {   0,   0,   1,   0,   0,   0,   0,   5,   0,   0,  -1 },      5.13472,   0.00004,     0.020 },
 { {   0,   8, -13,   0,   0,   0,   0,  -2,   0,   0,   0 },    324.68594,   0.00003,     0.040 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,  -1,   0 },    324.67709,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   0,   0,   1,   0 },    144.67709,   0.00003,     0.075 },
 { {   0,   8, -13,   0,   0,   0,   0,   2,   0,   0,   0 },    144.68611,   0.00003,     0.040 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 226> g_ELP19
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 },    180.00000,   0.00002,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },      0.00000,   0.00011,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },      0.00000,   0.00097,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },      0.00000,   0.00035,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0 },      0.00000,   0.00004,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -4,   0 },      0.00000,   0.00012,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,  -2 },    180.00000,   0.00002,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0 },      0.00000,   0.00169,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,  -2 },    180.00000,   0.00013,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },      0.00000,   0.02438,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   2 },    180.00000,   0.00006,     0.750 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -2 },    180.00000,   0.00076,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },      0.00000,   0.37115,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   2 },      0.00000,   0.00020,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2 },    180.00000,   0.00019,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },      0.00000,   1.67680,     1.000 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   2 },    180.00000,   0.00104,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2 },      0.00000,   0.00021,     0.079 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },      0.00000,   0.27560,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   2 },    180.00000,   0.00066,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,  -2 },    180.00000,   0.00007,     1.500 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0 },      0.00000,   0.01923,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   2 },    180.00000,   0.00011,     0.018 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   0 },      0.00000,   0.00138,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   2 },    180.00000,   0.00001,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   4,   0 },      0.00000,   0.00010,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },      0.00000,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },      0.00000,   0.00097,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2 },    180.00000,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },      0.00000,   0.01293,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2 },      0.00000,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },      0.00000,   0.03747,     0.500 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   2 },    180.00000,   0.00003,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   0 },      0.00000,   0.00585,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   2 },    180.00000,   0.00001,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0 },      0.00000,   0.00033,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   3,   0 },      0.00000,   0.00002,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0 },      0.00000,   0.00003,     0.043 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },      0.00000,   0.00039,     0.098 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 },      0.00000,   0.00076,     0.333 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   1,   0 },      0.00000,   0.00014,     0.062 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0 },      0.00000,   0.00001,     0.108 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0 },      0.00000,   0.00001,     0.250 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 },    180.00000,   0.00002,     0.107 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,  -1,   0 },    180.00000,   0.00001,     0.346 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0 },    180.00000,   0.00022,     0.096 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -2,   0 },      0.00000,   0.00002,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   0 },      0.00000,   0.00034,     0.530 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,  -2 },      0.00000,   0.00003,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },      0.00000,   0.00141,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0 },      0.00000,   0.00030,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   2,   0 },      0.00000,   0.00004,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0 },      0.00000,   0.00002,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },      0.00000,   0.00004,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },      0.00000,   0.00011,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0 },    180.00000,   0.00003,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -3,   0 },    180.00000,   0.00006,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,   0 },    180.00000,   0.00088,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0 },    180.00000,   0.00270,     8.850 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0 },    180.00000,   0.04516,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   2 },      0.00000,   0.00009,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   0 },    180.00000,   0.00317,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   2 },      0.00000,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   2,   0 },    180.00000,   0.00023,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   3,   0 },    180.00000,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,  -2,   0 },    180.00000,   0.00002,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0 },      0.00000,   0.00002,     0.898 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0 },      0.00000,   0.00020,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   1,   0 },      0.00000,   0.00004,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,  -1,   0 },    180.00000,   0.00008,     0.134 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0 },    180.00000,   0.00013,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,   1,   0 },    180.00000,   0.00001,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -2,   0 },    180.00000,   0.00007,     0.209 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -1,   0 },    180.00000,   0.00186,     0.118 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,  -2 },    180.00000,   0.00001,     0.196 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0 },    180.00000,   0.00256,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   1,   0 },    180.00000,   0.00025,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   2,   0 },    180.00000,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -3,   0 },    180.00000,   0.00008,     0.059 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -2,   0 },    180.00000,   0.00148,     0.265 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,  -2 },      0.00000,   0.00002,     0.058 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,   0 },    180.00000,   0.03707,     0.105 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,   2 },      0.00000,   0.00008,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,  -2 },    180.00000,   0.00032,     0.243 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0 },    180.00000,   0.04048,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   2 },      0.00000,   0.00009,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,  -2 },      0.00000,   0.00008,     0.109 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   0 },    180.00000,   0.00378,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   2 },      0.00000,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   2,   0 },    180.00000,   0.00032,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   3,   0 },    180.00000,   0.00003,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -4,   0 },    180.00000,   0.00007,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -3,   0 },    180.00000,   0.00120,     0.062 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,  -2 },      0.00000,   0.00007,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,   0 },    180.00000,   0.02165,     0.360 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,   2 },      0.00000,   0.00005,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,  -2 },      0.00000,   0.00015,     0.061 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   0 },    180.00000,   0.51642,     0.095 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   2 },      0.00000,   0.00107,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,  -2 },    180.00000,   0.00539,     0.322 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0 },    180.00000,   0.41383,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   2 },      0.00000,   0.00096,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,  -2 },      0.00000,   0.00094,     0.099 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,   0 },    180.00000,   0.03649,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,   2 },      0.00000,   0.00019,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,  -2 },      0.00000,   0.00007,     0.043 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,   0 },    180.00000,   0.00295,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,   2 },      0.00000,   0.00003,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   3,   0 },    180.00000,   0.00023,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   4,   0 },    180.00000,   0.00002,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0 },    180.00000,   0.00003,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0 },    180.00000,   0.00031,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },      0.00000,   0.00819,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   2 },    180.00000,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2 },      0.00000,   0.00029,     0.474 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },      0.00000,   0.01072,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   2 },    180.00000,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,  -2 },    180.00000,   0.00005,     0.090 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },      0.00000,   0.00128,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   2,   0 },      0.00000,   0.00013,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   3,   0 },      0.00000,   0.00001,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -4,   0 },    180.00000,   0.00005,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -3,   0 },    180.00000,   0.00045,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,   0 },    180.00000,   0.00631,     1.292 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   0 },      0.00000,   0.07118,     0.080 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   2 },    180.00000,   0.00017,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -2 },      0.00000,   0.00362,     0.903 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   0 },      0.00000,   0.06128,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   2 },    180.00000,   0.00016,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,  -2 },    180.00000,   0.00021,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   0 },      0.00000,   0.00734,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   2 },    180.00000,   0.00004,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   2,  -2 },    180.00000,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   2,   0 },      0.00000,   0.00073,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   3,   0 },      0.00000,   0.00007,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -3,   0 },      0.00000,   0.00005,     0.077 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -2,   0 },      0.00000,   0.00126,     4.425 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,  -2 },      0.00000,   0.00003,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },      0.00000,   0.01270,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   2 },    180.00000,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,  -2 },      0.00000,   0.00012,     9.300 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,   0 },      0.00000,   0.00094,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   1,   0 },      0.00000,   0.00007,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   3,  -1,   0 },      0.00000,   0.00001,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -2,  -2,   0 },      0.00000,   0.00002,     0.116 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -2,  -1,   0 },      0.00000,   0.00006,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0 },    180.00000,   0.00003,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -2,   0 },      0.00000,   0.00021,     0.104 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0 },      0.00000,   0.00058,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,  -2 },      0.00000,   0.00005,     0.108 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0 },    180.00000,   0.00017,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0 },    180.00000,   0.00007,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 },    180.00000,   0.00004,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -3,   0 },      0.00000,   0.00002,     0.602 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -2,   0 },      0.00000,   0.00011,     0.086 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -1,   0 },    180.00000,   0.00069,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,  -2 },    180.00000,   0.00002,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,   0 },    180.00000,   0.00038,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   1,   0 },    180.00000,   0.00006,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -2,   0 },    180.00000,   0.00006,     0.050 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -1,   0 },    180.00000,   0.00013,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0 },    180.00000,   0.00007,     0.022 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,   1,   0 },    180.00000,   0.00001,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -3,   0 },    180.00000,   0.00003,     0.130 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -2,   0 },    180.00000,   0.00079,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -1,   0 },    180.00000,   0.00155,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0 },    180.00000,   0.00076,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   1,   0 },    180.00000,   0.00012,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   2,   0 },    180.00000,   0.00001,     0.014 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -3,   0 },    180.00000,   0.00025,     0.115 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -2,   0 },    180.00000,   0.00686,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -2,   2 },      0.00000,   0.00004,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,  -2 },    180.00000,   0.00006,     0.119 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   0 },    180.00000,   0.01097,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   2 },      0.00000,   0.00006,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,  -2 },    180.00000,   0.00002,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0 },    180.00000,   0.00469,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   2 },      0.00000,   0.00003,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1,  -2 },      0.00000,   0.00002,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1,   0 },    180.00000,   0.00071,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   2,   0 },    180.00000,   0.00008,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0 },      0.00000,   0.00018,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,   0 },      0.00000,   0.00041,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0 },      0.00000,   0.00022,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   1,   0 },      0.00000,   0.00004,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -3,   0 },    180.00000,   0.00007,     0.093 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -2,   0 },      0.00000,   0.00090,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,  -2 },      0.00000,   0.00005,     0.096 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,   0 },      0.00000,   0.00160,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   0,   0 },      0.00000,   0.00073,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   1,   0 },      0.00000,   0.00013,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   2,   0 },      0.00000,   0.00002,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -3,   0 },      0.00000,   0.00002,     0.085 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -2,   0 },      0.00000,   0.00018,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -1,   0 },      0.00000,   0.00011,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,   0,   0 },      0.00000,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   5,  -1,  -2,   0 },      0.00000,   0.00002,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   5,   1,  -1,   0 },    180.00000,   0.00001,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -3,   0 },    180.00000,   0.00002,     0.031 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -2,   0 },    180.00000,   0.00004,     0.022 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -1,   0 },    180.00000,   0.00004,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,   0,   0 },    180.00000,   0.00001,     0.014 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -3,   0 },    180.00000,   0.00010,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -2,   0 },    180.00000,   0.00023,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -1,   0 },    180.00000,   0.00018,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,   0,   0 },    180.00000,   0.00006,     0.014 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,   1,   0 },    180.00000,   0.00001,     0.012 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   0,  -2,   0 },      0.00000,   0.00001,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   0,  -1,   0 },      0.00000,   0.00001,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   1,  -3,   0 },      0.00000,   0.00001,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   1,  -2,   0 },      0.00000,   0.00003,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   1,  -1,   0 },      0.00000,   0.00003,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   1 },     14.81311,   0.00001,     0.089 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1 },    194.81311,   0.00007,     0.487 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,  -1 },     14.81311,   0.00006,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -1 },    194.81311,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -1 },    194.81311,   0.00053,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   1 },    194.81311,   0.00007,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -1 },    194.81311,   0.00247,    18.600 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   1 },    194.81311,   0.00052,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -1 },    194.81311,   0.00053,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   1 },    194.81311,   0.00006,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,  -1 },    194.81311,   0.00004,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,  -1 },     14.81311,   0.00004,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   1 },    194.81311,   0.00001,     0.026 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -1 },    194.81311,   0.00005,     0.041 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 188> g_ELP20
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 },      0.00000,   0.00007,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1 },      0.00000,   0.00008,     5.997 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1 },      0.00000,   0.00008,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1 },      0.00000,   0.00005,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   1 },      0.00000,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -4,  -1 },      0.00000,   0.00002,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,  -1 },      0.00000,   0.00019,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   1 },      0.00000,   0.00006,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,  -1 },      0.00000,   0.00200,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   1 },      0.00000,   0.00076,     0.083 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -3 },    180.00000,   0.00002,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1 },      0.00000,   0.01696,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1 },      0.00000,   0.01419,     0.857 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   3 },      0.00000,   0.00002,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1 },      0.00000,   0.01215,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1 },      0.00000,   0.01623,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   3 },    180.00000,   0.00003,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -3 },      0.00000,   0.00002,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1 },      0.00000,   0.01279,     1.200 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   1 },      0.00000,   0.01338,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   3 },    180.00000,   0.00002,     0.018 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1 },      0.00000,   0.00079,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   1 },      0.00000,   0.00161,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,  -1 },      0.00000,   0.00006,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   1 },      0.00000,   0.00016,     0.018 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   4,   1 },      0.00000,   0.00001,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1 },      0.00000,   0.00008,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1 },      0.00000,   0.00003,     0.090 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1 },      0.00000,   0.00057,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1 },      0.00000,   0.00048,     0.462 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1 },      0.00000,   0.00009,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   1 },      0.00000,   0.00029,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1 },      0.00000,   0.00031,     0.545 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   1 },      0.00000,   0.00028,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1 },      0.00000,   0.00001,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   1 },      0.00000,   0.00003,     0.024 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1 },      0.00000,   0.00002,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   1 },      0.00000,   0.00001,     0.316 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,  -1 },    180.00000,   0.00001,     0.327 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   1 },    180.00000,   0.00001,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,  -1 },    180.00000,   0.00003,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,  -1 },      0.00000,   0.00003,     0.487 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   1 },      0.00000,   0.00005,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,  -1 },      0.00000,   0.00005,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   1 },      0.00000,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,  -1 },    180.00000,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,   1 },    180.00000,   0.00004,     3.575 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,  -1 },    180.00000,   0.00002,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   1 },      0.00000,   0.00014,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,  -1 },    180.00000,   0.00202,    18.600 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   1 },    180.00000,   0.00202,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,  -1 },    180.00000,   0.00009,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   1 },    180.00000,   0.00026,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   2,   1 },    180.00000,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   1 },      0.00000,   0.00001,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,  -1 },    180.00000,   0.00001,     0.137 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -1,  -1 },    180.00000,   0.00007,     0.202 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -1,   1 },    180.00000,   0.00008,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,  -1 },    180.00000,   0.00028,     0.120 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   1 },    180.00000,   0.00012,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   1,  -1 },    180.00000,   0.00002,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   1,   1 },    180.00000,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -2,  -1 },    180.00000,   0.00011,     0.058 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,  -1 },    180.00000,   0.00135,     0.254 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,   1 },    180.00000,   0.00158,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,  -3 },    180.00000,   0.00001,     0.057 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,  -1 },    180.00000,   0.00546,     0.107 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   1 },    180.00000,   0.00193,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,  -1 },    180.00000,   0.00035,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   1 },    180.00000,   0.00029,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   2,  -1 },    180.00000,   0.00003,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   2,   1 },    180.00000,   0.00003,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -4,  -1 },    180.00000,   0.00001,     0.023 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -3,  -1 },    180.00000,   0.00014,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -3,   1 },    180.00000,   0.00002,     0.384 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,  -1 },    180.00000,   0.00164,     0.062 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,   1 },      0.00000,   0.00016,     0.094 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,  -3 },    180.00000,   0.00003,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,  -1 },    180.00000,   0.01869,     0.340 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   1 },    180.00000,   0.02229,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   3 },      0.00000,   0.00003,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,  -3 },    180.00000,   0.00022,     0.060 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,  -1 },    180.00000,   0.07430,     0.097 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   1 },    180.00000,   0.01999,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   3 },      0.00000,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,  -3 },      0.00000,   0.00004,     0.305 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,  -1 },    180.00000,   0.00443,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,   1 },    180.00000,   0.00285,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,  -1 },    180.00000,   0.00032,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,   1 },    180.00000,   0.00031,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   3,  -1 },    180.00000,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   3,   1 },    180.00000,   0.00003,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -1 },      0.00000,   0.00026,     0.515 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   1 },      0.00000,   0.00037,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -3 },      0.00000,   0.00001,     0.064 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1 },      0.00000,   0.00295,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   1 },      0.00000,   0.00053,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,  -1 },      0.00000,   0.00022,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   1 },      0.00000,   0.00010,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   2,  -1 },      0.00000,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   2,   1 },      0.00000,   0.00001,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -3,  -1 },    180.00000,   0.00005,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,  -1 },    180.00000,   0.00020,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,   1 },    180.00000,   0.00014,     0.079 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,  -1 },      0.00000,   0.00198,     1.063 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   1 },      0.00000,   0.00330,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -3 },      0.00000,   0.00014,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -1 },      0.00000,   0.03043,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   1 },      0.00000,   0.00318,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,  -3 },    180.00000,   0.00001,     0.785 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,  -1 },      0.00000,   0.00207,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   1 },      0.00000,   0.00060,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   2,  -1 },      0.00000,   0.00016,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   2,   1 },      0.00000,   0.00008,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   3,  -1 },      0.00000,   0.00001,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -2,  -1 },      0.00000,   0.00006,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -2,   1 },      0.00000,   0.00003,     0.073 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,  -1 },      0.00000,   0.00053,    16.885 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   1 },      0.00000,   0.00057,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,  -1 },      0.00000,   0.00068,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,   1 },      0.00000,   0.00007,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   1,  -1 },      0.00000,   0.00003,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,  -1 },      0.00000,   0.00005,     0.106 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   1 },      0.00000,   0.00003,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,  -1 },      0.00000,   0.00006,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -1,  -1 },    180.00000,   0.00003,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -1,   1 },    180.00000,   0.00005,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,  -1 },    180.00000,   0.00009,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,   1 },    180.00000,   0.00003,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   1,  -1 },    180.00000,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -1,  -1 },    180.00000,   0.00001,     0.051 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -1,   1 },    180.00000,   0.00001,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,   0,  -1 },    180.00000,   0.00001,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -2,  -1 },    180.00000,   0.00001,     0.133 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -2,   1 },    180.00000,   0.00006,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -1,  -1 },    180.00000,   0.00017,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -1,   1 },    180.00000,   0.00012,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,  -1 },    180.00000,   0.00015,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   1 },    180.00000,   0.00006,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   1,  -1 },    180.00000,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   1,   1 },    180.00000,   0.00001,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -2,  -1 },    180.00000,   0.00013,     0.117 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -2,   1 },    180.00000,   0.00054,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,  -1 },    180.00000,   0.00149,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   1 },    180.00000,   0.00085,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,  -3 },    180.00000,   0.00001,     0.122 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,  -1 },    180.00000,   0.00104,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   1 },    180.00000,   0.00040,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1,  -1 },    180.00000,   0.00014,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1,   1 },    180.00000,   0.00008,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   2,  -1 },    180.00000,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   2,   1 },    180.00000,   0.00001,     0.011 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2,   1 },      0.00000,   0.00001,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,  -1 },      0.00000,   0.00008,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,   1 },      0.00000,   0.00003,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,  -1 },      0.00000,   0.00007,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   1 },      0.00000,   0.00002,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   1,  -1 },      0.00000,   0.00001,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -2,   1 },      0.00000,   0.00007,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,  -1 },      0.00000,   0.00043,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,   1 },      0.00000,   0.00013,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   0,  -1 },      0.00000,   0.00028,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   0,   1 },      0.00000,   0.00006,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   1,  -1 },      0.00000,   0.00004,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   1,   1 },      0.00000,   0.00001,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -2,   1 },      0.00000,   0.00001,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -1,  -1 },      0.00000,   0.00002,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -3,   1 },    180.00000,   0.00001,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -2,  -1 },    180.00000,   0.00003,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -2,   1 },    180.00000,   0.00002,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -1,  -1 },    180.00000,   0.00004,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -1,   1 },    180.00000,   0.00002,     0.014 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,   0,  -1 },    180.00000,   0.00002,     0.017 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0 },    194.81311,   0.00002,     0.041 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0 },    194.81311,   0.00041,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0 },    194.81311,   0.00010,     0.530 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0 },    194.81311,   0.00002,     0.096 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0 },    194.81311,   0.00002,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -2 },    194.81311,   0.00003,     0.037 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   0 },    194.81311,   0.00066,     8.850 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -2 },    194.81311,   0.00008,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0 },     14.81311,   0.01172,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   2 },    194.81311,   0.00001,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -2 },    194.81311,   0.00002,     4.535 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,   0 },     14.81311,   0.00062,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,   0 },     14.81311,   0.00004,     0.025 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   0 },     14.81311,   0.00013,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   0 },     14.81311,   0.00007,     0.026 }
} };

constexpr array<ELP2000PlanetPertCoefficient, 169> g_ELP21
{ {
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },    270.00000,   0.00149, 99999.999 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2 },     90.00000,   0.00010,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 },    270.00000,   0.00174,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 },    270.00000,   0.00029,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0 },    270.00000,   0.00003,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -4,   0 },     90.00000,   0.00007,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0 },     90.00000,   0.00106,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 },     90.00000,   0.01764,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   2 },     90.00000,   0.00013,     0.750 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -2 },     90.00000,   0.00002,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0 },     90.00000,   0.32654,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   2 },    270.00000,   0.00084,     0.069 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2 },     90.00000,   0.00047,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },    270.00000,   0.12302,     1.000 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   2 },     90.00000,   0.00040,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2 },     90.00000,   0.00062,     0.079 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0 },    270.00000,   0.26396,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,  -2 },    270.00000,   0.00008,     1.500 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0 },    270.00000,   0.01449,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   0 },    270.00000,   0.00089,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   1,   4,   0 },    270.00000,   0.00006,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0 },     90.00000,   0.00005,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 },     90.00000,   0.00069,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },     90.00000,   0.01066,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2 },    270.00000,   0.00002,     0.064 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2 },     90.00000,   0.00003,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 },    270.00000,   0.00536,     0.500 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   0 },    270.00000,   0.00587,     0.066 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0 },    270.00000,   0.00026,     0.035 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0 },     90.00000,   0.00002,     0.043 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0 },     90.00000,   0.00029,     0.098 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 },    270.00000,   0.00016,     0.333 },
 { {   0,   0,   0,   0,   0,   0,   0,   0,   3,   1,   0 },    270.00000,   0.00014,     0.062 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,  -1,   0 },    270.00000,   0.00004,     0.346 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0 },     90.00000,   0.00015,     0.096 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   0 },    270.00000,   0.00029,     0.530 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,  -2 },     90.00000,   0.00003,     0.065 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },    270.00000,   0.00126,     0.088 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0 },    270.00000,   0.00028,     0.041 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,  -1,   2,   0 },    270.00000,   0.00003,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0 },     90.00000,   0.00002,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },     90.00000,   0.00004,     1.127 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },    270.00000,   0.00008,     0.081 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0 },     90.00000,   0.00002,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -3,   0 },    270.00000,   0.00004,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -2,   0 },    270.00000,   0.00084,     0.076 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0 },    270.00000,   0.00214,     8.850 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,  -2 },    270.00000,   0.00008,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0 },     90.00000,   0.04194,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   0 },     90.00000,   0.00235,     0.038 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   1,   2,   0 },     90.00000,   0.00015,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0 },    270.00000,   0.00002,     0.898 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0 },    270.00000,   0.00019,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   1,   2,   1,   0 },    270.00000,   0.00003,     0.036 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,  -1,   0 },     90.00000,   0.00004,     0.134 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0 },     90.00000,   0.00016,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -2,   0 },    270.00000,   0.00008,     0.209 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,  -1,   0 },     90.00000,   0.00112,     0.118 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0 },     90.00000,   0.00310,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   1,   0 },     90.00000,   0.00022,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -3,   0 },    270.00000,   0.00008,     0.059 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -2,   0 },    270.00000,   0.00173,     0.265 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,  -2 },    270.00000,   0.00005,     0.058 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,   0 },     90.00000,   0.02490,     0.105 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,  -2 },    270.00000,   0.00014,     0.243 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0 },     90.00000,   0.04970,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,  -2 },    270.00000,   0.00004,     0.109 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   1,   0 },     90.00000,   0.00331,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   2,   0 },     90.00000,   0.00023,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -4,   0 },    270.00000,   0.00006,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -3,   0 },    270.00000,   0.00125,     0.062 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,  -2 },    270.00000,   0.00004,     0.034 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,   0 },    270.00000,   0.02529,     0.360 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -2,   2 },    270.00000,   0.00008,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,  -2 },    270.00000,   0.00081,     0.061 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   0 },     90.00000,   0.38245,     0.095 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,  -1,   2 },    270.00000,   0.00009,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,  -2 },    270.00000,   0.00165,     0.322 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0 },     90.00000,   0.51395,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,  -2 },    270.00000,   0.00053,     0.099 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   1,   0 },     90.00000,   0.03222,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,  -2 },    270.00000,   0.00005,     0.043 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   2,   0 },     90.00000,   0.00213,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   3,   0 },     90.00000,   0.00015,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0 },    270.00000,   0.00002,     0.067 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0 },     90.00000,   0.00004,     0.564 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0 },    270.00000,   0.00633,     0.087 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2 },     90.00000,   0.00005,     0.474 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },    270.00000,   0.01356,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,  -2 },     90.00000,   0.00003,     0.090 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,   0 },    270.00000,   0.00114,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   0,   2,   0 },    270.00000,   0.00009,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -4,   0 },    270.00000,   0.00004,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -3,   0 },    270.00000,   0.00043,     0.071 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -2,   0 },    270.00000,   0.00039,     1.292 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,  -2 },     90.00000,   0.00012,     0.070 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   0 },    270.00000,   0.06068,     0.080 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,  -2 },     90.00000,   0.00034,     0.903 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   0,   0 },    270.00000,   0.07754,     0.039 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,  -2 },     90.00000,   0.00014,     0.082 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   0 },    270.00000,   0.00658,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   2,   0 },    270.00000,   0.00053,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   1,   3,   0 },    270.00000,   0.00004,     0.015 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -3,   0 },     90.00000,   0.00005,     0.077 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -2,   0 },     90.00000,   0.00055,     4.425 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,  -2 },     90.00000,   0.00005,     0.075 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0 },    270.00000,   0.01186,     0.074 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,   0 },    270.00000,   0.00074,     0.037 },
 { {   0,   0,   0,   0,   0,   0,   0,   2,   2,   1,   0 },    270.00000,   0.00005,     0.025 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -2,  -1,   0 },    270.00000,   0.00007,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0 },     90.00000,   0.00005,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -2,   0 },    270.00000,   0.00013,     0.104 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0 },    270.00000,   0.00064,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,  -2 },    270.00000,   0.00004,     0.108 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0 },     90.00000,   0.00040,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,  -1,   1,   0 },     90.00000,   0.00002,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0 },     90.00000,   0.00006,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 },     90.00000,   0.00003,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -2,   0 },    270.00000,   0.00010,     0.086 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,  -1,   0 },     90.00000,   0.00053,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,  -2 },     90.00000,   0.00002,     0.089 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   0,   0 },     90.00000,   0.00027,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   3,   1,   1,   0 },     90.00000,   0.00004,     0.019 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -2,   0 },     90.00000,   0.00004,     0.050 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -1,   0 },     90.00000,   0.00012,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0 },     90.00000,   0.00006,     0.022 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -2,   0 },     90.00000,   0.00054,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,  -1,   0 },     90.00000,   0.00140,     0.029 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,  -2 },     90.00000,   0.00002,     0.048 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0 },     90.00000,   0.00064,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -2,   1,   0 },     90.00000,   0.00009,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -3,   0 },     90.00000,   0.00011,     0.115 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -2,   0 },     90.00000,   0.00476,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,  -2 },     90.00000,   0.00007,     0.119 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   0 },     90.00000,   0.00993,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,  -2 },     90.00000,   0.00014,     0.046 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0 },     90.00000,   0.00394,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   1,   0 },     90.00000,   0.00051,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,  -1,   2,   0 },     90.00000,   0.00005,     0.013 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0 },    270.00000,   0.00012,     0.044 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,  -1,   0 },    270.00000,   0.00038,     0.028 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0 },    270.00000,   0.00019,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   0,   1,   0 },    270.00000,   0.00003,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -3,   0 },     90.00000,   0.00006,     0.093 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -2,   0 },    270.00000,   0.00060,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,  -2 },    270.00000,   0.00004,     0.096 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,  -1,   0 },    270.00000,   0.00146,     0.027 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   0,  -2 },    270.00000,   0.00005,     0.042 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   0,   0 },    270.00000,   0.00061,     0.020 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   1,   1,   0 },    270.00000,   0.00009,     0.016 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -2,   0 },    270.00000,   0.00013,     0.040 },
 { {   0,   0,   0,   0,   0,   0,   0,   4,   2,  -1,   0 },    270.00000,   0.00009,     0.026 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -2,   0 },     90.00000,   0.00003,     0.022 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -1,   0 },     90.00000,   0.00003,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -3,   0 },     90.00000,   0.00006,     0.030 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -2,   0 },     90.00000,   0.00017,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,  -1,   0 },     90.00000,   0.00013,     0.017 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,  -1,   0,   0 },     90.00000,   0.00004,     0.014 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   1,  -2,   0 },    270.00000,   0.00002,     0.021 },
 { {   0,   0,   0,   0,   0,   0,   0,   6,   1,  -1,   0 },    270.00000,   0.00002,     0.016 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,  -1 },    284.81311,   0.00003,     0.040 },
 { {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,  -1 },    104.81311,   0.00005,     0.087 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -2,  -1 },    284.81311,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,  -1 },    284.81311,   0.00049,     0.075 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,  -1,   1 },    104.81311,   0.00012,     0.074 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -1 },    104.81311,   0.00049,     0.076 },
 { {   0,   0,   1,   0,   0,   0,   0,   1,   0,   2,  -1 },    104.81311,   0.00003,     0.038 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,  -1,  -1 },    284.81311,   0.00003,     0.088 },
 { {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,  -1 },    104.81311,   0.00004,     0.041 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 3> g_ELP22
{ {
 {  0, {  1,  1, -1, -1 },    192.93665,   0.00004,     0.075 },
 {  0, {  1,  1,  0, -1 },    192.93665,   0.00082,    18.600 },
 {  0, {  1,  1,  1, -1 },    192.93665,   0.00004,     0.076 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 2> g_ELP23
{ {
 {  0, {  1,  1,  0, -2 },    192.93663,   0.00004,     0.074 },
 {  0, {  1,  1,  0,  0 },    192.93664,   0.00004,     0.075 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 2> g_ELP24
{ {
 {  0, {  1,  1, -1, -1 },    282.93665,   0.00004,     0.075 },
 {  0, {  1,  1,  1, -1 },    102.93665,   0.00004,     0.076 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 6> g_ELP25
{ {
 {  0, {  0,  0,  1,  0 },      0.00000,   0.00058,     0.075 },
 {  0, {  0,  0,  2,  0 },      0.00000,   0.00004,     0.038 },
 {  0, {  2,  0, -2,  0 },      0.00000,   0.00002,     0.564 },
 {  0, {  2,  0, -1,  0 },      0.00000,   0.00021,     0.087 },
 {  0, {  2,  0,  0,  0 },      0.00000,   0.00009,     0.040 },
 {  0, {  2,  0,  1,  0 },      0.00000,   0.00001,     0.026 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 4> g_ELP26
{ {
 {  0, {  0,  0,  0,  1 },    180.00000,   0.00005,     0.075 },
 {  0, {  0,  0,  1, -1 },      0.00000,   0.00003,     5.997 },
 {  0, {  0,  0,  1,  1 },      0.00000,   0.00003,     0.037 },
 {  0, {  2,  0,  0, -1 },      0.00000,   0.00001,     0.088 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 5> g_ELP27
{ {
 {  0, {  0,  0,  0,  0 },     90.00000,   0.00356, 99999.999 },
 {  0, {  0,  0,  1,  0 },    270.00000,   0.00072,     0.075 },
 {  0, {  0,  0,  2,  0 },    270.00000,   0.00003,     0.038 },
 {  0, {  2,  0, -1,  0 },    270.00000,   0.00019,     0.087 },
 {  0, {  2,  0,  0,  0 },    270.00000,   0.00013,     0.040 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 20> g_ELP28
{ {
 {  0, {  0,  0,  0,  1 },    303.96185,   0.00004,     0.075 },
 {  0, {  0,  0,  1, -1 },    259.88393,   0.00016,     5.997 },
 {  0, {  0,  0,  2, -2 },      0.43020,   0.00040,     2.998 },
 {  0, {  0,  0,  3, -2 },      0.43379,   0.00002,     0.077 },
 {  0, {  0,  1, -1,  0 },    359.99858,   0.00014,     0.082 },
 {  0, {  0,  1,  0,  0 },    359.99982,   0.00223,     1.000 },
 {  0, {  0,  1,  1,  0 },    359.99961,   0.00014,     0.070 },
 {  0, {  1,  0, -1,  0 },    359.99331,   0.00009,     1.127 },
 {  0, {  1,  0,  0,  0 },    359.99537,   0.00001,     0.081 },
 {  0, {  1,  1, -1,  0 },      0.06418,   0.00003,     8.850 },
 {  0, {  2, -1, -1,  0 },    180.00095,   0.00004,     0.095 },
 {  0, {  2, -1,  0,  0 },    180.00014,   0.00003,     0.042 },
 {  0, {  2,  0, -3,  0 },    179.98126,   0.00001,     0.067 },
 {  0, {  2,  0, -2,  0 },    179.98366,   0.00025,     0.564 },
 {  0, {  2,  0, -1,  0 },    179.99638,   0.00014,     0.087 },
 {  0, {  2,  0,  0, -2 },    179.95864,   0.00003,     0.474 },
 {  0, {  2,  0,  0,  0 },    179.99904,   0.00002,     0.040 },
 {  0, {  2,  1, -2,  0 },    179.99184,   0.00002,     1.292 },
 {  0, {  2,  1, -1,  0 },      0.00313,   0.00002,     0.080 },
 {  0, {  2,  1,  0,  0 },    359.99965,   0.00002,     0.039 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 12> g_ELP29
{ {
 {  0, {  0,  0,  1, -1 },      0.02199,   0.00003,     5.997 },
 {  0, {  0,  0,  1,  0 },    245.99067,   0.00001,     0.075 },
 {  0, {  0,  0,  1,  1 },      0.00530,   0.00001,     0.037 },
 {  0, {  0,  0,  2, -3 },      0.42283,   0.00002,     0.073 },
 {  0, {  0,  0,  2, -1 },      0.74505,   0.00001,     0.076 },
 {  0, {  0,  1, -1, -1 },    359.99982,   0.00001,     0.039 },
 {  0, {  0,  1,  0, -1 },    359.99982,   0.00010,     0.081 },
 {  0, {  0,  1,  0,  1 },    359.99982,   0.00010,     0.069 },
 {  0, {  0,  1,  1,  1 },    359.99982,   0.00001,     0.036 },
 {  0, {  2,  0, -2, -1 },    179.98356,   0.00001,     0.066 },
 {  0, {  2,  0, -2,  1 },    179.98353,   0.00001,     0.086 },
 {  0, {  2,  0,  0, -1 },    179.99478,   0.00005,     0.088 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 14> g_ELP30
{ {
 {  0, {  0,  0,  0,  0 },     90.00000,   0.00130, 99999.999 },
 {  0, {  0,  0,  0,  1 },    213.95720,   0.00003,     0.075 },
 {  0, {  0,  0,  0,  2 },    270.03745,   0.00002,     0.037 },
 {  0, {  0,  0,  1,  0 },     90.07597,   0.00004,     0.075 },
 {  0, {  0,  0,  3, -2 },    270.43429,   0.00002,     0.077 },
 {  0, {  0,  1, -1,  0 },     89.99919,   0.00013,     0.082 },
 {  0, {  0,  1,  0,  0 },    270.00007,   0.00022,     1.000 },
 {  0, {  0,  1,  1,  0 },    269.99903,   0.00011,     0.070 },
 {  0, {  2, -1, -1,  0 },     89.99815,   0.00002,     0.095 },
 {  0, {  2, -1,  0,  0 },     90.00052,   0.00003,     0.042 },
 {  0, {  2,  0, -2,  0 },    269.98585,   0.00005,     0.564 },
 {  0, {  2,  0, -1,  0 },     89.99863,   0.00013,     0.087 },
 {  0, {  2,  1, -1,  0 },    269.99982,   0.00002,     0.080 },
 {  0, {  2,  1,  0,  0 },    269.99982,   0.00003,     0.039 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 11> g_ELP31
{ {
 {  0, {  0,  1, -1,  0 },    179.93473,   0.00006,     0.082 },
 {  0, {  0,  1,  0,  0 },    179.98532,   0.00081,     1.000 },
 {  0, {  0,  1,  1,  0 },    179.96323,   0.00005,     0.070 },
 {  0, {  1,  0,  0,  0 },      0.00001,   0.00013,     0.081 },
 {  0, {  1,  1,  0,  0 },    180.02282,   0.00001,     0.075 },
 {  0, {  2, -1, -1,  0 },      0.02264,   0.00002,     0.095 },
 {  0, {  2,  0, -1,  0 },    359.98826,   0.00002,     0.087 },
 {  0, {  2,  0,  0,  0 },    180.00019,   0.00055,     0.040 },
 {  0, {  2,  0,  1,  0 },    180.00017,   0.00006,     0.026 },
 {  0, {  2,  1, -1,  0 },    180.74954,   0.00001,     0.080 },
 {  0, {  4,  0, -1,  0 },    180.00035,   0.00001,     0.028 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 4> g_ELP32
{ {
 {  0, {  0,  1,  0, -1 },    179.99803,   0.00004,     0.081 },
 {  0, {  0,  1,  0,  1 },    179.99798,   0.00004,     0.069 },
 {  0, {  2,  0,  0, -1 },    359.99810,   0.00002,     0.088 },
 {  0, {  2,  0,  0,  1 },    180.00026,   0.00002,     0.026 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 10> g_ELP33
{ {
 {  0, {  0,  0,  0,  0 },    270.00000,   0.00828, 99999.999 },
 {  0, {  0,  0,  1,  0 },     89.99994,   0.00043,     0.075 },
 {  0, {  0,  1, -1,  0 },    269.93292,   0.00005,     0.082 },
 {  0, {  0,  1,  0,  0 },    270.00908,   0.00009,     1.000 },
 {  0, {  0,  1,  1,  0 },     89.95765,   0.00005,     0.070 },
 {  0, {  1,  0,  0,  0 },    270.00002,   0.00006,     0.081 },
 {  0, {  2, -1,  0,  0 },     89.97071,   0.00002,     0.042 },
 {  0, {  2,  0, -1,  0 },    269.99367,   0.00003,     0.087 },
 {  0, {  2,  0,  0,  0 },     90.00014,   0.00106,     0.040 },
 {  0, {  2,  0,  1,  0 },     90.00010,   0.00008,     0.026 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 28> g_ELP34
{ {
 {  0, {  0,  1, -2,  0 },      0.00000,   0.00007,     0.039 },
 {  0, {  0,  1, -1,  0 },      0.00000,   0.00108,     0.082 },
 {  0, {  0,  1,  0,  0 },      0.00000,   0.00487,     1.000 },
 {  0, {  0,  1,  1,  0 },      0.00000,   0.00080,     0.070 },
 {  0, {  0,  1,  2,  0 },      0.00000,   0.00006,     0.036 },
 {  0, {  0,  2, -1,  0 },      0.00000,   0.00004,     0.089 },
 {  0, {  0,  2,  0,  0 },      0.00000,   0.00011,     0.500 },
 {  0, {  0,  2,  1,  0 },      0.00000,   0.00002,     0.066 },
 {  0, {  1,  1,  0,  0 },    180.00000,   0.00013,     0.075 },
 {  0, {  2, -2, -1,  0 },    180.00000,   0.00011,     0.105 },
 {  0, {  2, -2,  0,  0 },    180.00000,   0.00012,     0.044 },
 {  0, {  2, -2,  1,  0 },    180.00000,   0.00001,     0.028 },
 {  0, {  2, -1, -2,  0 },    180.00000,   0.00006,     0.360 },
 {  0, {  2, -1, -1,  0 },    180.00000,   0.00150,     0.095 },
 {  0, {  2, -1,  0, -2 },    180.00000,   0.00002,     0.322 },
 {  0, {  2, -1,  0,  0 },    180.00000,   0.00120,     0.042 },
 {  0, {  2, -1,  1,  0 },    180.00000,   0.00011,     0.027 },
 {  0, {  2,  0, -1,  0 },      0.00000,   0.00002,     0.087 },
 {  0, {  2,  0,  0,  0 },      0.00000,   0.00003,     0.040 },
 {  0, {  2,  1, -2,  0 },    180.00000,   0.00002,     1.292 },
 {  0, {  2,  1, -1,  0 },      0.00000,   0.00021,     0.080 },
 {  0, {  2,  1,  0, -2 },      0.00000,   0.00001,     0.903 },
 {  0, {  2,  1,  0,  0 },      0.00000,   0.00018,     0.039 },
 {  0, {  2,  1,  1,  0 },      0.00000,   0.00002,     0.026 },
 {  0, {  2,  2, -1,  0 },      0.00000,   0.00004,     0.074 },
 {  0, {  4, -1, -2,  0 },    180.00000,   0.00002,     0.046 },
 {  0, {  4, -1, -1,  0 },    180.00000,   0.00003,     0.028 },
 {  0, {  4, -1,  0,  0 },    180.00000,   0.00001,     0.021 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 13> g_ELP35
{ {
 {  0, {  0,  1, -1, -1 },      0.00000,   0.00005,     0.039 },
 {  0, {  0,  1, -1,  1 },      0.00000,   0.00004,     0.857 },
 {  0, {  0,  1,  0, -1 },      0.00000,   0.00004,     0.081 },
 {  0, {  0,  1,  0,  1 },      0.00000,   0.00005,     0.069 },
 {  0, {  0,  1,  1, -1 },      0.00000,   0.00004,     1.200 },
 {  0, {  0,  1,  1,  1 },      0.00000,   0.00004,     0.036 },
 {  0, {  2, -2,  0, -1 },    180.00000,   0.00002,     0.107 },
 {  0, {  2, -1, -1, -1 },    180.00000,   0.00005,     0.340 },
 {  0, {  2, -1, -1,  1 },    180.00000,   0.00006,     0.042 },
 {  0, {  2, -1,  0, -1 },    180.00000,   0.00022,     0.097 },
 {  0, {  2, -1,  0,  1 },    180.00000,   0.00006,     0.027 },
 {  0, {  2, -1,  1, -1 },    180.00000,   0.00001,     0.042 },
 {  0, {  2,  1,  0, -1 },      0.00000,   0.00009,     0.081 }
} };

constexpr array<ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient, 19> g_ELP36
{ {
 {  0, {  0,  1, -2,  0 },     90.00000,   0.00005,     0.039 },
 {  0, {  0,  1, -1,  0 },     90.00000,   0.00095,     0.082 },
 {  0, {  0,  1,  0,  0 },    270.00000,   0.00036,     1.000 },
 {  0, {  0,  1,  1,  0 },    270.00000,   0.00077,     0.070 },
 {  0, {  0,  1,  2,  0 },    270.00000,   0.00004,     0.036 },
 {  0, {  0,  2, -1,  0 },     90.00000,   0.00003,     0.089 },
 {  0, {  1,  1,  0,  0 },     90.00000,   0.00012,     0.075 },
 {  0, {  2, -2, -1,  0 },     90.00000,   0.00007,     0.105 },
 {  0, {  2, -2,  0,  0 },     90.00000,   0.00014,     0.044 },
 {  0, {  2, -1, -2,  0 },    270.00000,   0.00007,     0.360 },
 {  0, {  2, -1, -1,  0 },     90.00000,   0.00111,     0.095 },
 {  0, {  2, -1,  0,  0 },     90.00000,   0.00149,     0.042 },
 {  0, {  2, -1,  1,  0 },     90.00000,   0.00009,     0.027 },
 {  0, {  2,  0,  0,  0 },    270.00000,   0.00004,     0.040 },
 {  0, {  2,  1, -1,  0 },    270.00000,   0.00018,     0.080 },
 {  0, {  2,  1,  0,  0 },    270.00000,   0.00023,     0.039 },
 {  0, {  2,  1,  1,  0 },    270.00000,   0.00002,     0.026 },
 {  0, {  2,  2, -1,  0 },    270.00000,   0.00003,     0.074 },
 {  0, {  4, -1, -1,  0 },     90.00000,   0.00003,     0.028 }
} };


////////////////////////////// Implementation /////////////////////////////////

double CAAELP2000::MoonMeanMeanLongitude(const double* pT, int nTSize) noexcept //Aka W1
{
  //Validate our parameters
  assert(pT);
  assert((nTSize == 5) || (nTSize == 2));
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  double fValue = g_W[0] + g_W[3] * pT[1];
  if (nTSize == 5)
    fValue += g_W[6] * pT[2] + g_W[9] * pT[3] + g_W[12] * pT[4];
  return fValue;
}

double CAAELP2000::MoonMeanMeanLongitude(double JD) noexcept //Aka W1
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MoonMeanMeanLongitude(t.data(), 5);
}

double CAAELP2000::MeanLongitudeLunarPerigee(const double* pT, int nTSize) noexcept //Aka W2
{
  //Validate our parameters
  assert(pT);
  assert((nTSize == 5) || (nTSize == 2));
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  double fValue = g_W[1] + g_W[4] * pT[1];
  if (nTSize == 5)
    fValue += g_W[7] * pT[2] + g_W[10] * pT[3] + g_W[13] * pT[4];
  return fValue;
}

double CAAELP2000::MeanLongitudeLunarPerigee(double JD) noexcept //Aka W2
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MeanLongitudeLunarPerigee(t.data(), 5);
}

double CAAELP2000::MeanLongitudeLunarAscendingNode(const double* pT, int nTSize) noexcept //Aka W3
{
  //Validate our parameters
  assert(pT);
  assert((nTSize == 5) || (nTSize == 2));
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  double fValue = g_W[2] + g_W[5] * pT[1];
  if (nTSize == 5)
    fValue += g_W[8] * pT[2] + g_W[11] * pT[3] + g_W[14] * pT[4];
  return fValue;
}

double CAAELP2000::MeanLongitudeLunarAscendingNode(double JD) noexcept //Aka W3
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MeanLongitudeLunarAscendingNode(t.data(), 5);
}

double CAAELP2000::MeanHeliocentricMeanLongitudeEarthMoonBarycentre(const double* pT, int nTSize) noexcept //Aka T
{
  //Validate our parameters
  assert(pT);
  assert((nTSize == 5) || (nTSize == 2));
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  double fValue = g_EARTH[0] + g_EARTH[1] * pT[1];
  if (nTSize == 5)
    fValue += g_EARTH[2] * pT[2] + g_EARTH[3] * pT[3] + g_EARTH[4] * pT[4];
  return fValue;
}

double CAAELP2000::MeanHeliocentricMeanLongitudeEarthMoonBarycentre(double JD) noexcept //Aka T
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MeanHeliocentricMeanLongitudeEarthMoonBarycentre(t.data(), 5);
}

double CAAELP2000::MeanLongitudeOfPerilhelionOfEarthMoonBarycentre(const double* pT, int nTSize) noexcept //Aka Omega'
{
  //Validate our parameters
  assert(pT);
  assert((nTSize == 5) || (nTSize == 2));
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  double fValue = g_PERI[0] + g_PERI[1] * pT[1];
  if (nTSize == 5)
    fValue += g_PERI[2] * pT[2] + g_PERI[3] * pT[3];

  return fValue;
}

double CAAELP2000::MeanLongitudeOfPerilhelionOfEarthMoonBarycentre(double JD) noexcept //Aka Omega'
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MeanLongitudeOfPerilhelionOfEarthMoonBarycentre(t.data(), 5);
}

double CAAELP2000::MoonMeanSolarElongation(const double* pT, int nTSize) noexcept //Aka D
{
  //Implement D in terms of W1 and T
  return MoonMeanMeanLongitude(pT, nTSize) - MeanHeliocentricMeanLongitudeEarthMoonBarycentre(pT, nTSize) + CAACoordinateTransformation::PI();
}

double CAAELP2000::MoonMeanSolarElongation(double JD) noexcept //Aka D
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MoonMeanSolarElongation(t.data(), 5);
}

double CAAELP2000::SunMeanAnomaly(const double* pT, int nTSize) noexcept //Aka l'
{
  //Implement l' in terms of T and Omega'
  return MeanHeliocentricMeanLongitudeEarthMoonBarycentre(pT, nTSize) - MeanLongitudeOfPerilhelionOfEarthMoonBarycentre(pT, nTSize);
}

double CAAELP2000::SunMeanAnomaly(double JD) noexcept //Aka l'
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return SunMeanAnomaly(t.data(), 5);
}

double CAAELP2000::MoonMeanAnomaly(const double* pT, int nTSize) noexcept //Aka l
{
  //Implement L in terms of W1 and W2
  return MoonMeanMeanLongitude(pT, nTSize) - MeanLongitudeLunarPerigee(pT, nTSize);
}

double CAAELP2000::MoonMeanAnomaly(double JD) noexcept //Aka l
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MoonMeanAnomaly(t.data(), 5);
}

double CAAELP2000::MoonMeanArgumentOfLatitude(const double* pT, int nTSize) noexcept //Aka F
{
  //Implement F in terms of W1 and W3
  return MoonMeanMeanLongitude(pT, nTSize) - MeanLongitudeLunarAscendingNode(pT, nTSize);
}

double CAAELP2000::MoonMeanArgumentOfLatitude(double JD) noexcept //Aka F
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return MoonMeanArgumentOfLatitude(t.data(), 5);
}

double CAAELP2000::MercuryMeanLongitude(double T) noexcept
{
  return g_P[0][0] + g_P[0][1]*T;
}

double CAAELP2000::VenusMeanLongitude(double T) noexcept
{
  return g_P[1][0] + g_P[1][1]*T;
}

double CAAELP2000::MarsMeanLongitude(double T) noexcept
{
  return g_P[3][0] + g_P[3][1]*T;
}

double CAAELP2000::JupiterMeanLongitude(double T) noexcept
{
  return g_P[4][0] + g_P[4][1]*T;
}

double CAAELP2000::SaturnMeanLongitude(double T) noexcept
{
  return g_P[5][0] + g_P[5][1]*T;
}

double CAAELP2000::UranusMeanLongitude(double T) noexcept
{
  return g_P[6][0] + g_P[6][1]*T;
}

double CAAELP2000::NeptuneMeanLongitude(double T) noexcept
{
  return g_P[7][0] + g_P[7][1]*T;
}

//Handle the main problem calculation (Longitude & Latitude)
double CAAELP2000::Accumulate(const ELP2000MainProblemCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF) noexcept
{
  //Validate our parameters
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double tgv = pCoefficients[j].m_B[0] + (DTASM * pCoefficients[j].m_B[4]);
    const double x = pCoefficients[j].m_A + (tgv * (g_DELNP - AM * g_DELNU)) + (pCoefficients[j].m_B[1] * g_DELG) + (pCoefficients[j].m_B[2] * g_DELE) + (pCoefficients[j].m_B[3] * g_DELEP);
    const double y = (fD * pCoefficients[j].m_I[0]) + (fldash * pCoefficients[j].m_I[1]) + (fl * pCoefficients[j].m_I[2]) + (fF * pCoefficients[j].m_I[3]);
    fResult += (x * sin(y));
  }

  return fResult;
}

double CAAELP2000::Accumulate_2(const ELP2000MainProblemCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF) noexcept
{
  //Validate our parameters
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double tgv = pCoefficients[j].m_B[0] + (DTASM * pCoefficients[j].m_B[4]);
    double A = pCoefficients[j].m_A;
    A -= (A * 2.0 * g_DELNU / 3.0);
    const double x = A + (tgv * (g_DELNP - AM * g_DELNU)) + (pCoefficients[j].m_B[1] * g_DELG) + (pCoefficients[j].m_B[2] * g_DELE) + (pCoefficients[j].m_B[3] * g_DELEP);
    const double y = (fD * pCoefficients[j].m_I[0]) + (fldash * pCoefficients[j].m_I[1]) + (fl * pCoefficients[j].m_I[2]) + (fF * pCoefficients[j].m_I[3]);
    fResult += (x * cos(y));
  }

  return fResult;
}


//Handle the Earth figure perturbations, Tidal Effects, Moon figure & Relativistic perturbations calculation
double CAAELP2000::Accumulate(const double* pT, int nTSize, const ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF, bool bI1isZero) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    if (bI1isZero)
      assert(pCoefficients[j].m_IZ == 0);

    double y = (fD * pCoefficients[j].m_I[0]) + (fldash * pCoefficients[j].m_I[1]) +
               (fl * pCoefficients[j].m_I[2]) + (fF * pCoefficients[j].m_I[3]) +
               CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_O);
    if (!bI1isZero)
      y += (pCoefficients[j].m_IZ * g_ZETA[0] * pT[0]) + (pCoefficients[j].m_IZ * g_ZETA[1] * pT[1]);
    fResult += (pCoefficients[j].m_A * sin(y));
  }

  return fResult;
}

//Handle the Earth figure perturbations & Tidal Effects /t calculation
double CAAELP2000::Accumulate_2(const double* pT, int nTSize, const ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF, bool bI1isZero) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    if (bI1isZero)
      assert(pCoefficients[j].m_IZ == 0);

    double y = (fD * pCoefficients[j].m_I[0]) + (fldash * pCoefficients[j].m_I[1]) +
               (fl * pCoefficients[j].m_I[2]) + (fF * pCoefficients[j].m_I[3]) +
               CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_O);
    if (!bI1isZero)
      y += (pCoefficients[j].m_IZ * g_ZETA[0] * pT[0]) + (pCoefficients[j].m_IZ * g_ZETA[1] * pT[1]);
    fResult += pCoefficients[j].m_A * pT[1] * sin(y);
  }

  return fResult;
}

//Handle the Planetary perturbations Table 1 calculation
double CAAELP2000::AccumulateTable1(const ELP2000PlanetPertCoefficient* pCoefficients, size_t nCoefficients, double fD, double fl, double fF, double fMe, double fV, double fT, double fMa, double fJ, double fS, double fU, double fN) noexcept
{
  //Validate our parameters
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double y = (fMe * pCoefficients[j].m_ip[0]) +
                     (fV * pCoefficients[j].m_ip[1]) +
                     (fT * pCoefficients[j].m_ip[2]) +
                     (fMa * pCoefficients[j].m_ip[3]) +
                     (fJ * pCoefficients[j].m_ip[4]) +
                     (fS * pCoefficients[j].m_ip[5]) +
                     (fU * pCoefficients[j].m_ip[6]) +
                     (fN * pCoefficients[j].m_ip[7]) +
                     (fD * pCoefficients[j].m_ip[8]) +
                     (fl * pCoefficients[j].m_ip[9]) +
                     (fF * pCoefficients[j].m_ip[10]) +
                     CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_theta);
    fResult += (pCoefficients[j].m_O * sin(y));
  }

  return fResult;
}

//Handle the Planetary perturbations Table 1 /t calculation
double CAAELP2000::AccumulateTable1_2(const double* pT, int nTSize, const ELP2000PlanetPertCoefficient* pCoefficients, size_t nCoefficients, double fD, double fl, double fF, double fMe, double fV, double fT, double fMa, double fJ, double fS, double fU, double fN) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double y = (fMe * pCoefficients[j].m_ip[0]) +
                     (fV * pCoefficients[j].m_ip[1]) +
                     (fT * pCoefficients[j].m_ip[2]) +
                     (fMa * pCoefficients[j].m_ip[3]) +
                     (fJ * pCoefficients[j].m_ip[4]) +
                     (fS * pCoefficients[j].m_ip[5]) +
                     (fU * pCoefficients[j].m_ip[6]) +
                     (fN * pCoefficients[j].m_ip[7]) +
                     (fD * pCoefficients[j].m_ip[8]) +
                     (fl * pCoefficients[j].m_ip[9]) +
                     (fF * pCoefficients[j].m_ip[10]) +
                     CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_theta);
    fResult += (pCoefficients[j].m_O * pT[1] * sin(y));
  }

  return fResult;
}

//Handle the Planetary perturbations Table 2 calculation
double CAAELP2000::AccumulateTable2(const ELP2000PlanetPertCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF, double fMe, double fV, double fT, double fMa, double fJ, double fS, double fU) noexcept
{
  //Validate our parameters
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double y = (fMe * pCoefficients[j].m_ip[0]) +
                     (fV * pCoefficients[j].m_ip[1]) +
                     (fT * pCoefficients[j].m_ip[2]) +
                     (fMa * pCoefficients[j].m_ip[3]) +
                     (fJ * pCoefficients[j].m_ip[4]) +
                     (fS * pCoefficients[j].m_ip[5]) +
                     (fU * pCoefficients[j].m_ip[6]) +
                     (fD * pCoefficients[j].m_ip[7]) +
                     (fldash * pCoefficients[j].m_ip[8]) +
                     (fl * pCoefficients[j].m_ip[9]) +
                     (fF * pCoefficients[j].m_ip[10]) +
                     CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_theta);
    fResult += (pCoefficients[j].m_O * sin(y));
  }

  return fResult;
}

//Handle the Planetary perturbations Table 2 /t calculation
double CAAELP2000::AccumulateTable2_2(const double* pT, int nTSize, const ELP2000PlanetPertCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF, double fMe, double fV, double fT, double fMa, double fJ, double fS, double fU) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  //What will be the return value from this function
  double fResult = 0;

  //Accumulate the result
  for (size_t j=0; j<nCoefficients; j++)
  {
    const double y = (fMe * pCoefficients[j].m_ip[0]) +
                     (fV * pCoefficients[j].m_ip[1]) +
                     (fT * pCoefficients[j].m_ip[2]) +
                     (fMa * pCoefficients[j].m_ip[3]) +
                     (fJ * pCoefficients[j].m_ip[4]) +
                     (fS * pCoefficients[j].m_ip[5]) +
                     (fU * pCoefficients[j].m_ip[6]) +
                     (fD * pCoefficients[j].m_ip[7]) +
                     (fldash * pCoefficients[j].m_ip[8]) +
                     (fl * pCoefficients[j].m_ip[9]) +
                     (fF * pCoefficients[j].m_ip[10]) +
                     CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_theta);
    fResult += pCoefficients[j].m_O * pT[1] * sin(y);
  }

  return fResult;
}

//Handle the Planetary perturbations (solar eccentricity) /t squared calculation
double CAAELP2000::Accumulate_3(const double* pT, int nTSize, const ELP2000EarthTidalMoonRelativisticSolarEccentricityCoefficient* pCoefficients, size_t nCoefficients, double fD, double fldash, double fl, double fF) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);
  assert(pCoefficients);
  assert(nCoefficients);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pCoefficients);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  //What will be the return value from this function
  double fResult = 0;

  for (size_t j=0; j<nCoefficients; j++)
  {
    assert(pCoefficients[j].m_IZ == 0);

    const double y = (fD * pCoefficients[j].m_I[0]) +
                     (fldash * pCoefficients[j].m_I[1]) +
                     (fl * pCoefficients[j].m_I[2]) +
                     (fF * pCoefficients[j].m_I[3]) +
                     CAACoordinateTransformation::DegreesToRadians(pCoefficients[j].m_O);
    fResult += (pCoefficients[j].m_A * pT[2] * sin(y));
  }

  return fResult;
}

#ifdef _MSC_VER
#pragma warning(suppress : 26429)
#endif //#ifdef _MSC_VER
double CAAELP2000::EclipticLongitude(const double* pT, int nTSize) noexcept
{
  //Validate our parameters
  assert(pT != nullptr);
  assert(nTSize >= 2);

  //Compute the delaney arguments for the specified time
  const double fD = MoonMeanSolarElongation(pT, nTSize);
  const double fldash = SunMeanAnomaly(pT, nTSize);
  const double fl = MoonMeanAnomaly(pT, nTSize);
  const double fF = MoonMeanArgumentOfLatitude(pT, nTSize);
  const double fD2 = MoonMeanSolarElongation(pT, 2);
  const double fldash2 = SunMeanAnomaly(pT, 2);
  const double fl2 = MoonMeanAnomaly(pT, 2);
  const double fF2 = MoonMeanArgumentOfLatitude(pT, 2);

  //Compute the planet mean longitudes for the specified time
  const double fMe = MercuryMeanLongitude(pT[1]);
  const double fV = VenusMeanLongitude(pT[1]);
  const double fT = MeanHeliocentricMeanLongitudeEarthMoonBarycentre(pT, 2);
  const double fMa = MarsMeanLongitude(pT[1]);
  const double fJ = JupiterMeanLongitude(pT[1]);
  const double fS = SaturnMeanLongitude(pT[1]);
  const double fU = UranusMeanLongitude(pT[1]);
  const double fN = NeptuneMeanLongitude(pT[1]);

  //Calculate the Longitude
  const double A = Accumulate(g_ELP1.data(), g_ELP1.size(), fD, fldash, fl, fF) +
                   Accumulate(pT, nTSize, g_ELP4.data(), g_ELP4.size(), fD2, fldash2, fl2, fF2, false) +
                   Accumulate_2(pT, nTSize, g_ELP7.data(), g_ELP7.size(), fD2, fldash2, fl2, fF2, false) +
                   AccumulateTable1(g_ELP10.data(), g_ELP10.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                   AccumulateTable1_2(pT, nTSize, g_ELP13.data(), g_ELP13.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                   AccumulateTable2(g_ELP16.data(), g_ELP16.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                   AccumulateTable2_2(pT, nTSize, g_ELP19.data(), g_ELP19.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                   Accumulate(pT, nTSize, g_ELP22.data(), g_ELP22.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate_2(pT, nTSize, g_ELP25.data(), g_ELP25.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate(pT, nTSize, g_ELP28.data(), g_ELP28.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate(pT, nTSize, g_ELP31.data(), g_ELP31.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate_3(pT, nTSize, g_ELP34.data(), g_ELP34.size(), fD2, fldash2, fl2, fF2);
  return CAACoordinateTransformation::MapTo0To360Range(A/3600.0 + CAACoordinateTransformation::RadiansToDegrees(MoonMeanMeanLongitude(pT, nTSize)));
}

double CAAELP2000::EclipticLongitude(double JD) noexcept
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return EclipticLongitude(t.data(), 5);
}

#ifdef _MSC_VER
#pragma warning(suppress : 26429)
#endif //#ifdef _MSC_VER
double CAAELP2000::EclipticLatitude(const double* pT, int nTSize) noexcept
{
  //Validate our parameters
  assert(pT != nullptr);
  assert(nTSize >= 2);

  //Compute the delaney arguments for the specified time
  const double fD = MoonMeanSolarElongation(pT, nTSize);
  const double fldash = SunMeanAnomaly(pT, nTSize);
  const double fl = MoonMeanAnomaly(pT, nTSize);
  const double fF = MoonMeanArgumentOfLatitude(pT, nTSize);
  const double fD2 = MoonMeanSolarElongation(pT, 2);
  const double fldash2 = SunMeanAnomaly(pT, 2);
  const double fl2 = MoonMeanAnomaly(pT, 2);
  const double fF2 = MoonMeanArgumentOfLatitude(pT, 2);

  //Compute the planet mean longitudes for the specified time
  const double fMe = MercuryMeanLongitude(pT[1]);
  const double fV = VenusMeanLongitude(pT[1]);
  const double fT = MeanHeliocentricMeanLongitudeEarthMoonBarycentre(pT, 2);
  const double fMa = MarsMeanLongitude(pT[1]);
  const double fJ = JupiterMeanLongitude(pT[1]);
  const double fS = SaturnMeanLongitude(pT[1]);
  const double fU = UranusMeanLongitude(pT[1]);
  const double fN = NeptuneMeanLongitude(pT[1]);

  //Calculate the Longitude
  const double B = Accumulate(g_ELP2.data(), g_ELP2.size(), fD, fldash, fl, fF) +
                   Accumulate(pT, nTSize, g_ELP5.data(), g_ELP5.size(), fD2, fldash2, fl2, fF2, false) +
                   Accumulate_2(pT, nTSize, g_ELP8.data(), g_ELP8.size(), fD2, fldash2, fl2, fF2, false) +
                   AccumulateTable1(g_ELP11.data(), g_ELP11.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                   AccumulateTable1_2(pT, nTSize, g_ELP14.data(), g_ELP14.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                   AccumulateTable2(g_ELP17.data(), g_ELP17.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                   AccumulateTable2_2(pT, nTSize, g_ELP20.data(), g_ELP20.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                   Accumulate(pT, nTSize, g_ELP23.data(), g_ELP23.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate_2(pT, nTSize, g_ELP26.data(), g_ELP26.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate(pT, nTSize, g_ELP29.data(), g_ELP29.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate(pT, nTSize, g_ELP32.data(), g_ELP32.size(), fD2, fldash2, fl2, fF2, true) +
                   Accumulate_3(pT, nTSize, g_ELP35.data(), g_ELP35.size(), fD2, fldash2, fl2, fF2);
  return CAACoordinateTransformation::MapToMinus90To90Range(B/3600.0);
}

double CAAELP2000::EclipticLatitude(double JD) noexcept
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return EclipticLatitude(t.data(), 5);
}

#ifdef _MSC_VER
#pragma warning(suppress : 26429)
#endif //#ifdef _MSC_VER
double CAAELP2000::RadiusVector(const double* pT, int nTSize) noexcept
{
  //Validate our parameters
  assert(pT != nullptr);
  assert(nTSize >= 2);

  //Compute the delaney arguments for the specified time
  const double fD = MoonMeanSolarElongation(pT, nTSize);
  const double fldash = SunMeanAnomaly(pT, nTSize);
  const double fl = MoonMeanAnomaly(pT, nTSize);
  const double fF = MoonMeanArgumentOfLatitude(pT, nTSize);
  const double fD2 = MoonMeanSolarElongation(pT, 2);
  const double fldash2 = SunMeanAnomaly(pT, 2);
  const double fl2 = MoonMeanAnomaly(pT, 2);
  const double fF2 = MoonMeanArgumentOfLatitude(pT, 2);

  //Compute the planet mean longitudes for the specified time
  const double fMe = MercuryMeanLongitude(pT[1]);
  const double fV = VenusMeanLongitude(pT[1]);
  const double fT = MeanHeliocentricMeanLongitudeEarthMoonBarycentre(pT, 2);
  const double fMa = MarsMeanLongitude(pT[1]);
  const double fJ = JupiterMeanLongitude(pT[1]);
  const double fS = SaturnMeanLongitude(pT[1]);
  const double fU = UranusMeanLongitude(pT[1]);
  const double fN = NeptuneMeanLongitude(pT[1]);

  //Calculate the Longitude
  const double fValue = Accumulate_2(g_ELP3.data(), g_ELP3.size(), fD, fldash, fl, fF) +
                        Accumulate(pT, nTSize, g_ELP6.data(), g_ELP6.size(), fD2, fldash2, fl2, fF2, false) +
                        Accumulate_2(pT, nTSize, g_ELP9.data(), g_ELP9.size(), fD2, fldash2, fl2, fF2, false) +
                        AccumulateTable1(g_ELP12.data(), g_ELP12.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                        AccumulateTable1_2(pT, nTSize, g_ELP15.data(), g_ELP15.size(), fD2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU, fN) +
                        AccumulateTable2(g_ELP18.data(), g_ELP18.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                        AccumulateTable2_2(pT, nTSize, g_ELP21.data(), g_ELP21.size(), fD2, fldash2, fl2, fF2, fMe, fV, fT, fMa, fJ, fS, fU) +
                        Accumulate(pT, nTSize, g_ELP24.data(), g_ELP24.size(), fD2, fldash2, fl2, fF2, true) +
                        Accumulate_2(pT, nTSize, g_ELP27.data(), g_ELP27.size(), fD2, fldash2, fl2, fF2, true) +
                        Accumulate(pT, nTSize, g_ELP30.data(), g_ELP30.size(), fD2, fldash2, fl2, fF2, true) +
                        Accumulate(pT, nTSize, g_ELP33.data(), g_ELP33.size(), fD2, fldash2, fl2, fF2, true) +
                        Accumulate_3(pT, nTSize, g_ELP36.data(), g_ELP36.size(), fD2, fldash2, fl2, fF2);
  return fValue * 384747.9806448954 / 384747.9806743165;
}

double CAAELP2000::RadiusVector(double JD) noexcept
{
  //Calculate Julian centuries
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  return RadiusVector(t.data(), 5);
}

CAA3DCoordinate CAAELP2000::EclipticRectangularCoordinates(double JD) noexcept
{
  double fLongitude = EclipticLongitude(JD);
  double fLatitude = EclipticLatitude(JD);
  const double fR = RadiusVector(JD);

  CAA3DCoordinate value;
  fLongitude = CAACoordinateTransformation::DegreesToRadians(fLongitude);
  fLatitude = CAACoordinateTransformation::DegreesToRadians(fLatitude);
  const double fCosLat = cos(fLatitude);
  value.X = fR * cos(fLongitude) * fCosLat;
  value.Y = fR * sin(fLongitude) * fCosLat;
  value.Z = fR * sin(fLatitude);

  return value;
}

CAA3DCoordinate CAAELP2000::EclipticRectangularCoordinatesJ2000(double JD) noexcept
{
  array<double, 5> t{};
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];
  const double P = (1.0180391e-5   + 4.7020439e-7*t[1] + -5.417367e-10*t[2] + -2.507948e-12*t[3] + 4.63486e-15*t[4]) * t[1];
  const double Q = (-1.13469002e-4 + 1.2372674e-7*t[1] + 1.265417e-9*t[2]   + -1.371808e-12*t[3] + -3.20334e-15*t[4]) * t[1];
  const double TwoP = 2*P;
  const double P2 = P*P;
  const double Q2 = Q*Q;
  const double OneMinus2P2 = 1 - 2*P2;
  const double TwoPQ = TwoP*Q;
  const double Twosqrt1MinusPart = 2*sqrt(1 - P2 - Q2);

  const CAA3DCoordinate Ecliptic = EclipticRectangularCoordinates(JD);

  CAA3DCoordinate J2000;
  J2000.X = (OneMinus2P2*Ecliptic.X) + (TwoPQ*Ecliptic.Y) + (P*Twosqrt1MinusPart*Ecliptic.Z);
  J2000.Y = (TwoPQ*Ecliptic.X) + ((1 - 2 * Q2)*Ecliptic.Y) - (Q*Twosqrt1MinusPart*Ecliptic.Z);
  J2000.Z = (-P*Twosqrt1MinusPart*Ecliptic.X) + (Q*Twosqrt1MinusPart*Ecliptic.Y) + ((1-2*P2-2*Q2)*Ecliptic.Z);

  return J2000;
}

CAA3DCoordinate CAAELP2000::EquatorialRectangularCoordinatesFK5(double JD) noexcept
{
  const CAA3DCoordinate J2000 = EclipticRectangularCoordinatesJ2000(JD);

  CAA3DCoordinate FK5;
  FK5.X = J2000.X + (0.000000437913*J2000.Y) - (0.000000189859*J2000.Z);
  FK5.Y = (-0.000000477299*J2000.X) + (0.917482137607*J2000.Y) - (0.397776981701*J2000.Z);
  FK5.Z = (0.397776981701*J2000.Y) + (0.917482137607*J2000.Z);

  return FK5;
}
