/*
Module : AAELPMPP2.cpp
Purpose: Implementation for the algorithms for ELP/MPP02
Created: PJN / 30-07-2017
History: PJN / 30-07-2017 1. Initial public release.
         PJN / 01-08-2017 1. Lookup tables in AAELPMPP02.cpp now consistently use lowercase "e" when declaring values
                          in lookup tables.
         PJN / 08-06-2019 1. Updated the code to clean compile on VC 2019
         PJN / 18-08-2019 1. Fixed some further compiler warnings when using VC 2019 Preview v16.3.0 Preview 2.0
         PJN / 18-04-2020 1. Reworked C arrays to use std::array
         PJN / 29-04-2020 1. Fixed a compilation issue on GCC where size_t was undefined in various modules. Thanks
                          to Bert Devlieghe for reporting this bug.
         PJN / 23-11-2020 1. Reworked CAAELPMPP02::EclipticLatitude & CAAELPMPP02::RadiusVector methods to use 
                          std::array::size instead of sizeof.

Copyright (c) 2017 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "AAELPMPP02.h"
#include "AACoordinateTransformation.h"
#include "AA3DCoordinate.h"
#include "AADefines.h"
#include <cmath>
#include <cassert>
#include <cstddef>
#include <array>
using namespace std;


////////////////////////////// Macros / Defines ///////////////////////////////

#ifdef _MSC_VER
#pragma warning(disable : 26446 26481 26482 26485)
#endif //#ifdef _MSC_VER

constexpr array<array<double, 5>, 3> g_W_Nominal
{ {
  { 3.8103440873402215,     //CAACoordinateTransformation::DegreesToRadians(218 + 18/60.0 + 59.95571/3600.0 - 0.0708/3600.0)
    8399.6847317739157,     //CAACoordinateTransformation::DegreesToRadians(1732559343.73604 / 3600.0),
   -3.3008054664661645e-05, //CAACoordinateTransformation::DegreesToRadians(-6.8084 / 3600.0)
    3.2017095500473753e-08, //CAACoordinateTransformation::DegreesToRadians(0.006604 / 3600.0),
   -1.5363745554361199e-10  //CAACoordinateTransformation::DegreesToRadians(-0.00003169 / 3600.0),
  },
  {
    1.4547885323225087,     //CAACoordinateTransformation::DegreesToRadians(83 + 21/60.0 + 11.67475/3600.0)
    70.993305079674201,     //CAACoordinateTransformation::DegreesToRadians(14643420.3171 / 3600.0),
   -0.00018550425880294177, //CAACoordinateTransformation::DegreesToRadians(-38.263 / 3600.0),
   -2.1839401892941267e-07, //CAACoordinateTransformation::DegreesToRadians(-0.045047 / 3600.0),
    1.0327016221314225e-09  //CAACoordinateTransformation::DegreesToRadians(0.00021301 / 3600.0)
  },
  {
    2.1824391972168398,     //CAACoordinateTransformation::DegreesToRadians(125 + 2/60.0 + 40.39816/3600.0),
   -33.781427210382816,     //CAACoordinateTransformation::DegreesToRadians(-6967919.5383 / 3600.0)
    3.0829301981755397e-05, //CAACoordinateTransformation::DegreesToRadians(6.359 / 3600.0),
    3.6967043184602122e-08, //CAACoordinateTransformation::DegreesToRadians(0.007625 / 3600.0),
   -1.7385418604587962e-10  //CAACoordinateTransformation::DegreesToRadians(-0.00003586 / 3600.0)
  }
} };

constexpr array<array<double, 5>, 3> g_W_LLR
{ {
  {
    3.8103439203219081,
    8399.6847302074329,
   -3.3189520425500942e-05,
    3.2017095500473753e-08,
   -1.5363745554361197e-10
  },
  {
    1.4547893480700087,
    70.99330548230661,
   -0.00018550474361662286,
   -2.1839401892941265e-07,
    1.0327016221314225e-09
  },
  {
    2.1824386755573193,
   -33.781427433053146,
    3.082930198175539e-05,
    3.6967043184602116e-08,
   -1.738541860458796e-10
  }
} };

constexpr array<array<double, 5>, 3> g_W_DE405
{ {
  {
    3.8103440908308803,
    8399.68473007193,
   -3.3189520425500942e-05,
    3.2017095500473753e-08,
   -1.5363745554361197e-10
  },
  {
    1.4547895404440776,
    70.993305448479248,
   -0.00018550474361662286,
   -2.1839401892941265e-07,
    1.0327016221314225e-09
  },
  {
    2.1824388474237688,
   -33.781427419672326,
    3.082930198175539e-05,
    3.6967043184602116e-08,
   -1.738541860458796e-10
  }
} };

constexpr array<array<double, 5>, 3> g_W_DE406
{ {
  {
    3.8103440908308803,
    8399.68473007193,
   -3.3189520425500942e-05,
    3.1102494491060616e-08,
   -2.0328237648922842e-10
  },
  {
    1.4547895404440776,
    70.993305448479248,
   -0.00018548192818782712,
   -2.1961637966359414e-07,
    1.0327016221314225e-09
  },
  {
    2.1824388474237688,
   -33.781427419672326,
    3.0816644950982666e-05,
    3.6447710769397583e-08,
   -1.738541860458796e-10
  }
} };

constexpr array<double, 5> g_ZETA_Nominal
{ {
  3.8103440873402215,
  8399.7091120695241,
 -3.3008054664661645e-05,
  3.2017095500473753e-08,
 -1.5363745554361199e-10
} };

constexpr array<double, 5> g_ZETA_LLR
{ {
  3.8103439203219081,
  8399.7091105030413,
 -3.3189520425500942e-05,
  3.2017095500473753e-08,
 -1.5363745554361197e-10
} };

constexpr array<double, 5> g_ZETA_DE405
{ {
  3.8103440908308803,
  8399.7091103675375,
 -3.3189520425500942e-05,
  3.2017095500473753e-08,
 -1.5363745554361197e-10
} };

constexpr array<double, 5> g_ZETA_DE406
{ {
  3.8103440908308803,
  8399.7091103675375,
 -3.3189520425500942e-05,
  3.1102494491060616e-08,
 -2.0328237648922842e-10
} };

constexpr array<double, 5> g_EARTH_Nominal
{ {
  1.7534699468639550,     //CAACoordinateTransformation::DegreesToRadians(100 + 27/60.0 + 59.13885/3600.0),
  628.30758504554331,     //CAACoordinateTransformation::DegreesToRadians(129597742.293 / 3600.0),
 -9.7932363584126268e-08, //CAACoordinateTransformation::DegreesToRadians(-0.0202 / 3600.0),
  4.3633231299858238e-11, //CAACoordinateTransformation::DegreesToRadians(0.000009 / 3600.0),
  7.2722052166430391e-13  //CAACoordinateTransformation::DegreesToRadians(0.00000015 / 3600.0)
} };

constexpr array<double, 5> g_EARTH_LLR
{ {
  1.753469752356706,
  628.30758511545343,
 -9.7932363584126268e-08,
  4.3633231299858238e-11,
  7.2722052166430391e-13
} };

constexpr array<double, 5> g_EARTH_DE405
{ {
  1.7534699452640696,
  628.30758508103156,
 -9.7932363584126268e-08,
  4.3633231299858238e-11,
  7.2722052166430391e-13
} };

constexpr const array<double, 5>& g_EARTH_DE406 = g_EARTH_DE405;

constexpr array<double, 5> g_PERI_Nominal
{ {
  1.7965956117694057,     //CAACoordinateTransformation::DegreesToRadians(102 + 56/60.0 + 14.445766/3600.0),
  0.0056298669711442699,  //CAACoordinateTransformation::DegreesToRadians(1161.24342 / 3600.0),
  2.5659491293243858e-06, //CAACoordinateTransformation::DegreesToRadians(0.529265 / 3600.0),
 -5.7275888286280579e-10, //CAACoordinateTransformation::DegreesToRadians(-0.00011814 / 3600.0),
  5.5166948773454099e-11  //CAACoordinateTransformation::DegreesToRadians(0.000011379 / 3600.0)
} };

constexpr array<double, 5> g_PERI_LLR
{ {
  1.7965954341045842,
  0.00562986697114427,
  2.5659491293243853e-06,
 -5.7275888286280579e-10,
  5.51669487734541e-11
} };

constexpr array<double, 5> g_PERI_DE405
{ {
  1.7965956331206,
  0.00562986697114427,
  2.5659491293243853e-06,
 -5.7275888286280579e-10,
  5.51669487734541e-11
} };

constexpr array<double, 5> g_PERI_DE406
{ {
  1.7965956331206,
  0.00562986697114427,
  2.5659491293243853e-06,
 -5.7275888286280579e-10,
  5.51669487734541e-11
} };

constexpr array<array<double, 2>, 8> g_P
{ {
  { 4.4026086342191189,  2608.7903141569259 }, //{ CAACoordinateTransformation::DegreesToRadians(252 + 15/60.0 + 3.216919/3600.0),  CAACoordinateTransformation::DegreesToRadians(538101628.68888 / 3600.0) },
  { 3.1761344571511088,  1021.3285547385308 }, //{ CAACoordinateTransformation::DegreesToRadians(181 + 58/60.0 + 44.758419/3600.0), CAACoordinateTransformation::DegreesToRadians(210664136.45777 / 3600.0) },
  { 1.7534699468639550,  628.30758504554331 }, //{ CAACoordinateTransformation::DegreesToRadians(100 + 27/60.0 + 59.13885/3600.0),  CAACoordinateTransformation::DegreesToRadians(129597742.293 / 3600.0)   },
  { 6.2034995986912955,  334.06124347172772 }, //{ CAACoordinateTransformation::DegreesToRadians(355 + 26/60.0 + 3.642778/3600.0),  CAACoordinateTransformation::DegreesToRadians(68905077.65936 / 3600.0)  },
  { 0.59954667808842876, 52.969097211191368 }, //{ CAACoordinateTransformation::DegreesToRadians(34  + 21/60.0 + 5.379392/3600.0),  CAACoordinateTransformation::DegreesToRadians(10925660.57335 / 3600.0)  },
  { 0.87401678344988065, 21.329907977851814 }, //{ CAACoordinateTransformation::DegreesToRadians(50  + 4/60.0  + 38.902495/3600.0), CAACoordinateTransformation::DegreesToRadians(4399609.33632 / 3600.0)   },
  { 5.4812276258103312,  7.4781665690567305 }, //{ CAACoordinateTransformation::DegreesToRadians(314 + 3/60.0  + 4.354234/3600.0),  CAACoordinateTransformation::DegreesToRadians(1542482.57845 / 3600.0)   },
  { 5.3118941049906994,  3.8132918131353417 }  //{ CAACoordinateTransformation::DegreesToRadians(304 + 20/60.0 + 56.808371/3600.0), CAACoordinateTransformation::DegreesToRadians(786547.89700 / 3600.0)    }
} };

struct AAPLUS_EXT_CLASS ELPMPP02MainProblemNominalCoefficient
{
  array<int, 4> m_I;
};

constexpr array<ELPMPP02MainProblemNominalCoefficient, 1023> g_MAIN_S1
{ {
 {  0,  2,  0,  0 },
 {  0,  4,  0,  0 },
 {  0,  6,  0,  0 },
 {  0, -6,  1,  0 },
 {  0, -4,  1,  0 },
 {  0, -2,  1,  0 },
 {  0,  0,  1,  0 },
 {  0,  2,  1,  0 },
 {  0,  4,  1,  0 },
 {  0,  6,  1,  0 },
 {  0, -4,  2,  0 },
 {  0, -2,  2,  0 },
 {  0,  0,  2,  0 },
 {  0,  2,  2,  0 },
 {  0,  4,  2,  0 },
 {  0,  6,  2,  0 },
 {  0, -4,  3,  0 },
 {  0, -2,  3,  0 },
 {  0,  0,  3,  0 },
 {  0,  2,  3,  0 },
 {  0,  4,  3,  0 },
 {  0, -2,  4,  0 },
 {  0,  0,  4,  0 },
 {  0,  2,  4,  0 },
 {  0,  4,  4,  0 },
 {  0, -2,  5,  0 },
 {  0,  0,  5,  0 },
 {  0,  2,  5,  0 },
 {  0,  4,  5,  0 },
 {  0, -2,  6,  0 },
 {  0,  0,  6,  0 },
 {  0,  2,  6,  0 },
 {  0,  0,  7,  0 },
 {  0,  2,  7,  0 },
 {  0,  0,  8,  0 },
 {  0,  0, -7,  1 },
 {  0,  0, -6,  1 },
 {  0, -2, -5,  1 },
 {  0,  0, -5,  1 },
 {  0, -2, -4,  1 },
 {  0,  0, -4,  1 },
 {  0,  2, -4,  1 },
 {  0, -4, -3,  1 },
 {  0, -2, -3,  1 },
 {  0,  0, -3,  1 },
 {  0,  2, -3,  1 },
 {  0, -4, -2,  1 },
 {  0, -2, -2,  1 },
 {  0,  0, -2,  1 },
 {  0,  2, -2,  1 },
 {  0,  4, -2,  1 },
 {  0, -4, -1,  1 },
 {  0, -2, -1,  1 },
 {  0,  0, -1,  1 },
 {  0,  2, -1,  1 },
 {  0,  4, -1,  1 },
 {  0, -4,  0,  1 },
 {  0, -2,  0,  1 },
 {  0,  0,  0,  1 },
 {  0,  2,  0,  1 },
 {  0,  4,  0,  1 },
 {  0, -4,  1,  1 },
 {  0, -2,  1,  1 },
 {  0,  0,  1,  1 },
 {  0,  2,  1,  1 },
 {  0,  4,  1,  1 },
 {  0, -4,  2,  1 },
 {  0, -2,  2,  1 },
 {  0,  0,  2,  1 },
 {  0,  2,  2,  1 },
 {  0,  4,  2,  1 },
 {  0, -2,  3,  1 },
 {  0,  0,  3,  1 },
 {  0,  2,  3,  1 },
 {  0,  4,  3,  1 },
 {  0, -2,  4,  1 },
 {  0,  0,  4,  1 },
 {  0,  2,  4,  1 },
 {  0, -2,  5,  1 },
 {  0,  0,  5,  1 },
 {  0,  2,  5,  1 },
 {  0,  0,  6,  1 },
 {  0,  0,  7,  1 },
 {  0,  0, -5,  2 },
 {  0, -2, -4,  2 },
 {  0,  0, -4,  2 },
 {  0, -2, -3,  2 },
 {  0,  0, -3,  2 },
 {  0,  2, -3,  2 },
 {  0, -2, -2,  2 },
 {  0,  0, -2,  2 },
 {  0,  2, -2,  2 },
 {  0, -2, -1,  2 },
 {  0,  0, -1,  2 },
 {  0,  2, -1,  2 },
 {  0, -2,  0,  2 },
 {  0,  0,  0,  2 },
 {  0,  2,  0,  2 },
 {  0,  4,  0,  2 },
 {  0, -2,  1,  2 },
 {  0,  0,  1,  2 },
 {  0,  2,  1,  2 },
 {  0, -2,  2,  2 },
 {  0,  0,  2,  2 },
 {  0,  2,  2,  2 },
 {  0, -2,  3,  2 },
 {  0,  0,  3,  2 },
 {  0,  2,  3,  2 },
 {  0,  0,  4,  2 },
 {  0,  0, -4,  3 },
 {  0,  0, -3,  3 },
 {  0, -2, -2,  3 },
 {  0,  0, -2,  3 },
 {  0,  2, -2,  3 },
 {  0, -2, -1,  3 },
 {  0,  0, -1,  3 },
 {  0,  2, -1,  3 },
 {  0, -2,  0,  3 },
 {  0,  0,  0,  3 },
 {  0,  2,  0,  3 },
 {  0,  0,  1,  3 },
 {  0,  2,  1,  3 },
 {  0,  0,  2,  3 },
 {  0,  0,  3,  3 },
 {  0,  0, -2,  4 },
 {  0,  0, -1,  4 },
 {  0,  0,  0,  4 },
 {  0,  0,  1,  4 },
 {  0,  0,  2,  4 },
 {  0,  0, -1,  5 },
 {  1,  0,  0, -4 },
 {  1,  0, -2, -3 },
 {  1,  0, -1, -3 },
 {  1,  0,  0, -3 },
 {  1,  0,  1, -3 },
 {  1,  0,  2, -3 },
 {  1,  0, -3, -2 },
 {  1,  0, -2, -2 },
 {  1, -2, -1, -2 },
 {  1,  0, -1, -2 },
 {  1, -2,  0, -2 },
 {  1,  0,  0, -2 },
 {  1,  2,  0, -2 },
 {  1, -2,  1, -2 },
 {  1,  0,  1, -2 },
 {  1,  0,  2, -2 },
 {  1,  0,  3, -2 },
 {  1,  0, -5, -1 },
 {  1,  0, -4, -1 },
 {  1, -2, -3, -1 },
 {  1,  0, -3, -1 },
 {  1,  2, -3, -1 },
 {  1, -2, -2, -1 },
 {  1,  0, -2, -1 },
 {  1,  2, -2, -1 },
 {  1, -2, -1, -1 },
 {  1,  0, -1, -1 },
 {  1,  2, -1, -1 },
 {  1, -4,  0, -1 },
 {  1, -2,  0, -1 },
 {  1,  0,  0, -1 },
 {  1,  2,  0, -1 },
 {  1, -2,  1, -1 },
 {  1,  0,  1, -1 },
 {  1,  2,  1, -1 },
 {  1, -2,  2, -1 },
 {  1,  0,  2, -1 },
 {  1,  2,  2, -1 },
 {  1, -2,  3, -1 },
 {  1,  0,  3, -1 },
 {  1,  2,  3, -1 },
 {  1,  0,  4, -1 },
 {  1,  0,  5, -1 },
 {  1,  0, -6,  0 },
 {  1, -2, -5,  0 },
 {  1,  0, -5,  0 },
 {  1, -2, -4,  0 },
 {  1,  0, -4,  0 },
 {  1,  2, -4,  0 },
 {  1, -2, -3,  0 },
 {  1,  0, -3,  0 },
 {  1,  2, -3,  0 },
 {  1, -4, -2,  0 },
 {  1, -2, -2,  0 },
 {  1,  0, -2,  0 },
 {  1,  2, -2,  0 },
 {  1, -4, -1,  0 },
 {  1, -2, -1,  0 },
 {  1,  0, -1,  0 },
 {  1,  2, -1,  0 },
 {  1,  4, -1,  0 },
 {  1, -4,  0,  0 },
 {  1, -2,  0,  0 },
 {  1,  0,  0,  0 },
 {  1,  2,  0,  0 },
 {  1,  4,  0,  0 },
 {  1, -4,  1,  0 },
 {  1, -2,  1,  0 },
 {  1,  0,  1,  0 },
 {  1,  2,  1,  0 },
 {  1,  4,  1,  0 },
 {  1, -2,  2,  0 },
 {  1,  0,  2,  0 },
 {  1,  2,  2,  0 },
 {  1,  4,  2,  0 },
 {  1, -2,  3,  0 },
 {  1,  0,  3,  0 },
 {  1,  2,  3,  0 },
 {  1,  0,  4,  0 },
 {  1,  2,  4,  0 },
 {  1,  0,  5,  0 },
 {  1,  0,  6,  0 },
 {  1,  0, -5,  1 },
 {  1, -2, -4,  1 },
 {  1,  0, -4,  1 },
 {  1, -2, -3,  1 },
 {  1,  0, -3,  1 },
 {  1,  2, -3,  1 },
 {  1, -2, -2,  1 },
 {  1,  0, -2,  1 },
 {  1,  2, -2,  1 },
 {  1, -2, -1,  1 },
 {  1,  0, -1,  1 },
 {  1,  2, -1,  1 },
 {  1, -2,  0,  1 },
 {  1,  0,  0,  1 },
 {  1,  2,  0,  1 },
 {  1,  4,  0,  1 },
 {  1, -2,  1,  1 },
 {  1,  0,  1,  1 },
 {  1,  2,  1,  1 },
 {  1,  4,  1,  1 },
 {  1, -2,  2,  1 },
 {  1,  0,  2,  1 },
 {  1,  2,  2,  1 },
 {  1, -2,  3,  1 },
 {  1,  0,  3,  1 },
 {  1,  2,  3,  1 },
 {  1,  0,  4,  1 },
 {  1,  0,  5,  1 },
 {  1,  0, -4,  2 },
 {  1,  0, -3,  2 },
 {  1,  0, -2,  2 },
 {  1,  0, -1,  2 },
 {  1,  2, -1,  2 },
 {  1, -2,  0,  2 },
 {  1,  0,  0,  2 },
 {  1,  2,  0,  2 },
 {  1, -2,  1,  2 },
 {  1,  0,  1,  2 },
 {  1,  2,  1,  2 },
 {  1,  0,  2,  2 },
 {  1,  2,  2,  2 },
 {  1,  0,  3,  2 },
 {  1,  0,  4,  2 },
 {  1,  0, -2,  3 },
 {  1,  0, -1,  3 },
 {  1, -2,  0,  3 },
 {  1,  0,  0,  3 },
 {  1,  0,  1,  3 },
 {  1,  0,  0,  4 },
 {  2,  0,  0, -6 },
 {  2,  0, -2, -5 },
 {  2,  0, -1, -5 },
 {  2,  0,  0, -5 },
 {  2,  0,  1, -5 },
 {  2,  0, -3, -4 },
 {  2,  0, -2, -4 },
 {  2,  0, -1, -4 },
 {  2,  2, -1, -4 },
 {  2, -2,  0, -4 },
 {  2,  0,  0, -4 },
 {  2,  2,  0, -4 },
 {  2, -2,  1, -4 },
 {  2,  0,  1, -4 },
 {  2,  0,  2, -4 },
 {  2,  0,  3, -4 },
 {  2,  0, -4, -3 },
 {  2,  0, -3, -3 },
 {  2, -2, -2, -3 },
 {  2,  0, -2, -3 },
 {  2,  2, -2, -3 },
 {  2, -2, -1, -3 },
 {  2,  0, -1, -3 },
 {  2,  2, -1, -3 },
 {  2, -2,  0, -3 },
 {  2,  0,  0, -3 },
 {  2,  2,  0, -3 },
 {  2, -2,  1, -3 },
 {  2,  0,  1, -3 },
 {  2,  2,  1, -3 },
 {  2, -2,  2, -3 },
 {  2,  0,  2, -3 },
 {  2,  2,  2, -3 },
 {  2,  0,  3, -3 },
 {  2,  0,  4, -3 },
 {  2,  0, -5, -2 },
 {  2, -2, -4, -2 },
 {  2,  0, -4, -2 },
 {  2, -2, -3, -2 },
 {  2,  0, -3, -2 },
 {  2, -2, -2, -2 },
 {  2,  0, -2, -2 },
 {  2,  2, -2, -2 },
 {  2, -2, -1, -2 },
 {  2,  0, -1, -2 },
 {  2,  2, -1, -2 },
 {  2,  4, -1, -2 },
 {  2, -4,  0, -2 },
 {  2, -2,  0, -2 },
 {  2,  0,  0, -2 },
 {  2,  2,  0, -2 },
 {  2,  4,  0, -2 },
 {  2, -4,  1, -2 },
 {  2, -2,  1, -2 },
 {  2,  0,  1, -2 },
 {  2,  2,  1, -2 },
 {  2,  4,  1, -2 },
 {  2, -2,  2, -2 },
 {  2,  0,  2, -2 },
 {  2,  2,  2, -2 },
 {  2, -2,  3, -2 },
 {  2,  0,  3, -2 },
 {  2,  2,  3, -2 },
 {  2,  0,  4, -2 },
 {  2,  0,  5, -2 },
 {  2,  0, -6, -1 },
 {  2, -2, -5, -1 },
 {  2,  0, -5, -1 },
 {  2, -2, -4, -1 },
 {  2,  0, -4, -1 },
 {  2,  2, -4, -1 },
 {  2, -4, -3, -1 },
 {  2, -2, -3, -1 },
 {  2,  0, -3, -1 },
 {  2,  2, -3, -1 },
 {  2, -4, -2, -1 },
 {  2, -2, -2, -1 },
 {  2,  0, -2, -1 },
 {  2,  2, -2, -1 },
 {  2,  4, -2, -1 },
 {  2, -4, -1, -1 },
 {  2, -2, -1, -1 },
 {  2,  0, -1, -1 },
 {  2,  2, -1, -1 },
 {  2,  4, -1, -1 },
 {  2, -4,  0, -1 },
 {  2, -2,  0, -1 },
 {  2,  0,  0, -1 },
 {  2,  2,  0, -1 },
 {  2,  4,  0, -1 },
 {  2, -4,  1, -1 },
 {  2, -2,  1, -1 },
 {  2,  0,  1, -1 },
 {  2,  2,  1, -1 },
 {  2,  4,  1, -1 },
 {  2, -4,  2, -1 },
 {  2, -2,  2, -1 },
 {  2,  0,  2, -1 },
 {  2,  2,  2, -1 },
 {  2,  4,  2, -1 },
 {  2, -2,  3, -1 },
 {  2,  0,  3, -1 },
 {  2,  2,  3, -1 },
 {  2, -2,  4, -1 },
 {  2,  0,  4, -1 },
 {  2,  2,  4, -1 },
 {  2, -2,  5, -1 },
 {  2,  0,  5, -1 },
 {  2,  2,  5, -1 },
 {  2,  0,  6, -1 },
 {  2,  0, -8,  0 },
 {  2,  0, -7,  0 },
 {  2, -2, -6,  0 },
 {  2,  0, -6,  0 },
 {  2, -2, -5,  0 },
 {  2,  0, -5,  0 },
 {  2,  2, -5,  0 },
 {  2, -4, -4,  0 },
 {  2, -2, -4,  0 },
 {  2,  0, -4,  0 },
 {  2,  2, -4,  0 },
 {  2, -4, -3,  0 },
 {  2, -2, -3,  0 },
 {  2,  0, -3,  0 },
 {  2,  2, -3,  0 },
 {  2,  4, -3,  0 },
 {  2, -4, -2,  0 },
 {  2, -2, -2,  0 },
 {  2,  0, -2,  0 },
 {  2,  2, -2,  0 },
 {  2,  4, -2,  0 },
 {  2, -6, -1,  0 },
 {  2, -4, -1,  0 },
 {  2, -2, -1,  0 },
 {  2,  0, -1,  0 },
 {  2,  2, -1,  0 },
 {  2,  4, -1,  0 },
 {  2,  6, -1,  0 },
 {  2, -6,  0,  0 },
 {  2, -4,  0,  0 },
 {  2, -2,  0,  0 },
 {  2,  0,  0,  0 },
 {  2,  2,  0,  0 },
 {  2,  4,  0,  0 },
 {  2,  6,  0,  0 },
 {  2, -6,  1,  0 },
 {  2, -4,  1,  0 },
 {  2, -2,  1,  0 },
 {  2,  0,  1,  0 },
 {  2,  2,  1,  0 },
 {  2,  4,  1,  0 },
 {  2,  6,  1,  0 },
 {  2, -4,  2,  0 },
 {  2, -2,  2,  0 },
 {  2,  0,  2,  0 },
 {  2,  2,  2,  0 },
 {  2,  4,  2,  0 },
 {  2, -4,  3,  0 },
 {  2, -2,  3,  0 },
 {  2,  0,  3,  0 },
 {  2,  2,  3,  0 },
 {  2,  4,  3,  0 },
 {  2, -2,  4,  0 },
 {  2,  0,  4,  0 },
 {  2,  2,  4,  0 },
 {  2,  4,  4,  0 },
 {  2, -2,  5,  0 },
 {  2,  0,  5,  0 },
 {  2,  2,  5,  0 },
 {  2, -2,  6,  0 },
 {  2,  0,  6,  0 },
 {  2,  2,  6,  0 },
 {  2,  0,  7,  0 },
 {  2,  0, -7,  1 },
 {  2,  0, -6,  1 },
 {  2, -2, -5,  1 },
 {  2,  0, -5,  1 },
 {  2, -2, -4,  1 },
 {  2,  0, -4,  1 },
 {  2,  2, -4,  1 },
 {  2, -2, -3,  1 },
 {  2,  0, -3,  1 },
 {  2,  2, -3,  1 },
 {  2, -4, -2,  1 },
 {  2, -2, -2,  1 },
 {  2,  0, -2,  1 },
 {  2,  2, -2,  1 },
 {  2, -4, -1,  1 },
 {  2, -2, -1,  1 },
 {  2,  0, -1,  1 },
 {  2,  2, -1,  1 },
 {  2,  4, -1,  1 },
 {  2, -4,  0,  1 },
 {  2, -2,  0,  1 },
 {  2,  0,  0,  1 },
 {  2,  2,  0,  1 },
 {  2,  4,  0,  1 },
 {  2, -4,  1,  1 },
 {  2, -2,  1,  1 },
 {  2,  0,  1,  1 },
 {  2,  2,  1,  1 },
 {  2,  4,  1,  1 },
 {  2, -2,  2,  1 },
 {  2,  0,  2,  1 },
 {  2,  2,  2,  1 },
 {  2,  4,  2,  1 },
 {  2, -2,  3,  1 },
 {  2,  0,  3,  1 },
 {  2,  2,  3,  1 },
 {  2, -2,  4,  1 },
 {  2,  0,  4,  1 },
 {  2,  2,  4,  1 },
 {  2,  0,  5,  1 },
 {  2,  0,  6,  1 },
 {  2,  0, -5,  2 },
 {  2,  0, -4,  2 },
 {  2, -2, -3,  2 },
 {  2,  0, -3,  2 },
 {  2,  2, -3,  2 },
 {  2, -2, -2,  2 },
 {  2,  0, -2,  2 },
 {  2,  2, -2,  2 },
 {  2, -4, -1,  2 },
 {  2, -2, -1,  2 },
 {  2,  0, -1,  2 },
 {  2,  2, -1,  2 },
 {  2,  4, -1,  2 },
 {  2, -4,  0,  2 },
 {  2, -2,  0,  2 },
 {  2,  0,  0,  2 },
 {  2,  2,  0,  2 },
 {  2, -2,  1,  2 },
 {  2,  0,  1,  2 },
 {  2,  2,  1,  2 },
 {  2, -2,  2,  2 },
 {  2,  0,  2,  2 },
 {  2, -2,  3,  2 },
 {  2,  0, -3,  3 },
 {  2,  0, -2,  3 },
 {  2, -2, -1,  3 },
 {  2,  0, -1,  3 },
 {  2, -2,  0,  3 },
 {  2,  0,  0,  3 },
 {  2, -2,  1,  3 },
 {  2,  0,  1,  3 },
 {  2,  0, -2,  4 },
 {  2,  0, -1,  4 },
 {  3,  0, -1, -4 },
 {  3,  0,  0, -4 },
 {  3,  0, -2, -3 },
 {  3,  0, -1, -3 },
 {  3, -2,  0, -3 },
 {  3,  0,  0, -3 },
 {  3,  0,  1, -3 },
 {  3,  0, -3, -2 },
 {  3,  0, -2, -2 },
 {  3, -2, -1, -2 },
 {  3,  0, -1, -2 },
 {  3,  2, -1, -2 },
 {  3, -2,  0, -2 },
 {  3,  0,  0, -2 },
 {  3, -2,  1, -2 },
 {  3,  0,  1, -2 },
 {  3,  0, -4, -1 },
 {  3,  0, -3, -1 },
 {  3,  0, -2, -1 },
 {  3,  2, -2, -1 },
 {  3, -2, -1, -1 },
 {  3,  0, -1, -1 },
 {  3,  2, -1, -1 },
 {  3, -4,  0, -1 },
 {  3, -2,  0, -1 },
 {  3,  0,  0, -1 },
 {  3,  2,  0, -1 },
 {  3, -2,  1, -1 },
 {  3,  0,  1, -1 },
 {  3,  2,  1, -1 },
 {  3, -2,  2, -1 },
 {  3,  0,  2, -1 },
 {  3,  2,  2, -1 },
 {  3,  0,  3, -1 },
 {  3,  0, -5,  0 },
 {  3, -2, -4,  0 },
 {  3,  0, -4,  0 },
 {  3, -2, -3,  0 },
 {  3,  0, -3,  0 },
 {  3,  2, -3,  0 },
 {  3, -2, -2,  0 },
 {  3,  0, -2,  0 },
 {  3,  2, -2,  0 },
 {  3,  4, -2,  0 },
 {  3, -4, -1,  0 },
 {  3, -2, -1,  0 },
 {  3,  0, -1,  0 },
 {  3,  2, -1,  0 },
 {  3,  4, -1,  0 },
 {  3, -4,  0,  0 },
 {  3, -2,  0,  0 },
 {  3,  0,  0,  0 },
 {  3,  2,  0,  0 },
 {  3,  4,  0,  0 },
 {  3, -4,  1,  0 },
 {  3, -2,  1,  0 },
 {  3,  0,  1,  0 },
 {  3,  2,  1,  0 },
 {  3, -2,  2,  0 },
 {  3,  0,  2,  0 },
 {  3,  2,  2,  0 },
 {  3, -2,  3,  0 },
 {  3,  0,  3,  0 },
 {  3,  2,  3,  0 },
 {  3,  0,  4,  0 },
 {  3,  0, -5,  1 },
 {  3,  0, -4,  1 },
 {  3, -2, -3,  1 },
 {  3,  0, -3,  1 },
 {  3,  2, -3,  1 },
 {  3, -2, -2,  1 },
 {  3,  0, -2,  1 },
 {  3,  2, -2,  1 },
 {  3, -2, -1,  1 },
 {  3,  0, -1,  1 },
 {  3,  2, -1,  1 },
 {  3, -2,  0,  1 },
 {  3,  0,  0,  1 },
 {  3,  2,  0,  1 },
 {  3, -2,  1,  1 },
 {  3,  0,  1,  1 },
 {  3,  2,  1,  1 },
 {  3, -2,  2,  1 },
 {  3,  0,  2,  1 },
 {  3,  2,  2,  1 },
 {  3,  0,  3,  1 },
 {  3,  0,  4,  1 },
 {  3,  0, -4,  2 },
 {  3,  0, -3,  2 },
 {  3,  0, -2,  2 },
 {  3, -2, -1,  2 },
 {  3,  0, -1,  2 },
 {  3, -2,  0,  2 },
 {  3,  0,  0,  2 },
 {  3,  2,  0,  2 },
 {  3,  0,  1,  2 },
 {  3,  0,  2,  2 },
 {  3,  0, -3,  3 },
 {  3, -2, -2,  3 },
 {  3,  0, -2,  3 },
 {  3,  0, -1,  3 },
 {  3, -2,  0,  3 },
 {  3,  0,  0,  3 },
 {  4,  0, -2, -5 },
 {  4,  0, -1, -5 },
 {  4,  0,  0, -5 },
 {  4,  0, -3, -4 },
 {  4,  0, -2, -4 },
 {  4,  0, -1, -4 },
 {  4,  0,  0, -4 },
 {  4,  0,  1, -4 },
 {  4,  0,  2, -4 },
 {  4,  0, -3, -3 },
 {  4,  0, -2, -3 },
 {  4,  2, -2, -3 },
 {  4, -2, -1, -3 },
 {  4,  0, -1, -3 },
 {  4,  2, -1, -3 },
 {  4, -2,  0, -3 },
 {  4,  0,  0, -3 },
 {  4,  2,  0, -3 },
 {  4, -2,  1, -3 },
 {  4,  0,  1, -3 },
 {  4,  2,  1, -3 },
 {  4,  0,  2, -3 },
 {  4,  0,  3, -3 },
 {  4,  0, -5, -2 },
 {  4,  0, -4, -2 },
 {  4,  0, -3, -2 },
 {  4,  2, -3, -2 },
 {  4, -2, -2, -2 },
 {  4,  0, -2, -2 },
 {  4,  2, -2, -2 },
 {  4, -2, -1, -2 },
 {  4,  0, -1, -2 },
 {  4,  2, -1, -2 },
 {  4, -2,  0, -2 },
 {  4,  0,  0, -2 },
 {  4,  2,  0, -2 },
 {  4, -2,  1, -2 },
 {  4,  0,  1, -2 },
 {  4,  2,  1, -2 },
 {  4, -2,  2, -2 },
 {  4,  0,  2, -2 },
 {  4,  2,  2, -2 },
 {  4, -2,  3, -2 },
 {  4,  0,  3, -2 },
 {  4,  0,  4, -2 },
 {  4,  0, -6, -1 },
 {  4,  0, -5, -1 },
 {  4, -2, -4, -1 },
 {  4,  0, -4, -1 },
 {  4, -2, -3, -1 },
 {  4,  0, -3, -1 },
 {  4,  2, -3, -1 },
 {  4, -4, -2, -1 },
 {  4, -2, -2, -1 },
 {  4,  0, -2, -1 },
 {  4,  2, -2, -1 },
 {  4,  4, -2, -1 },
 {  4, -4, -1, -1 },
 {  4, -2, -1, -1 },
 {  4,  0, -1, -1 },
 {  4,  2, -1, -1 },
 {  4,  4, -1, -1 },
 {  4, -4,  0, -1 },
 {  4, -2,  0, -1 },
 {  4,  0,  0, -1 },
 {  4,  2,  0, -1 },
 {  4,  4,  0, -1 },
 {  4, -4,  1, -1 },
 {  4, -2,  1, -1 },
 {  4,  0,  1, -1 },
 {  4,  2,  1, -1 },
 {  4,  4,  1, -1 },
 {  4, -2,  2, -1 },
 {  4,  0,  2, -1 },
 {  4,  2,  2, -1 },
 {  4, -2,  3, -1 },
 {  4,  0,  3, -1 },
 {  4,  2,  3, -1 },
 {  4, -2,  4, -1 },
 {  4,  0,  4, -1 },
 {  4,  0,  5, -1 },
 {  4,  0, -7,  0 },
 {  4, -2, -6,  0 },
 {  4,  0, -6,  0 },
 {  4, -2, -5,  0 },
 {  4,  0, -5,  0 },
 {  4, -2, -4,  0 },
 {  4,  0, -4,  0 },
 {  4,  2, -4,  0 },
 {  4, -4, -3,  0 },
 {  4, -2, -3,  0 },
 {  4,  0, -3,  0 },
 {  4,  2, -3,  0 },
 {  4,  4, -3,  0 },
 {  4, -4, -2,  0 },
 {  4, -2, -2,  0 },
 {  4,  0, -2,  0 },
 {  4,  2, -2,  0 },
 {  4,  4, -2,  0 },
 {  4, -4, -1,  0 },
 {  4, -2, -1,  0 },
 {  4,  0, -1,  0 },
 {  4,  2, -1,  0 },
 {  4,  4, -1,  0 },
 {  4, -4,  0,  0 },
 {  4, -2,  0,  0 },
 {  4,  0,  0,  0 },
 {  4,  2,  0,  0 },
 {  4,  4,  0,  0 },
 {  4, -4,  1,  0 },
 {  4, -2,  1,  0 },
 {  4,  0,  1,  0 },
 {  4,  2,  1,  0 },
 {  4,  4,  1,  0 },
 {  4, -4,  2,  0 },
 {  4, -2,  2,  0 },
 {  4,  0,  2,  0 },
 {  4,  2,  2,  0 },
 {  4,  4,  2,  0 },
 {  4, -2,  3,  0 },
 {  4,  0,  3,  0 },
 {  4,  2,  3,  0 },
 {  4, -2,  4,  0 },
 {  4,  0,  4,  0 },
 {  4,  2,  4,  0 },
 {  4,  0,  5,  0 },
 {  4,  0,  6,  0 },
 {  4,  0, -6,  1 },
 {  4,  0, -5,  1 },
 {  4,  0, -4,  1 },
 {  4, -2, -3,  1 },
 {  4,  0, -3,  1 },
 {  4,  2, -3,  1 },
 {  4, -2, -2,  1 },
 {  4,  0, -2,  1 },
 {  4,  2, -2,  1 },
 {  4, -2, -1,  1 },
 {  4,  0, -1,  1 },
 {  4,  2, -1,  1 },
 {  4,  4, -1,  1 },
 {  4, -2,  0,  1 },
 {  4,  0,  0,  1 },
 {  4,  2,  0,  1 },
 {  4, -4,  1,  1 },
 {  4, -2,  1,  1 },
 {  4,  0,  1,  1 },
 {  4,  2,  1,  1 },
 {  4, -2,  2,  1 },
 {  4,  0,  2,  1 },
 {  4,  2,  2,  1 },
 {  4, -2,  3,  1 },
 {  4,  0,  3,  1 },
 {  4,  2,  3,  1 },
 {  4,  0,  4,  1 },
 {  4,  0, -4,  2 },
 {  4,  0, -3,  2 },
 {  4,  2, -3,  2 },
 {  4, -2, -2,  2 },
 {  4,  0, -2,  2 },
 {  4,  2, -2,  2 },
 {  4, -2, -1,  2 },
 {  4,  0, -1,  2 },
 {  4,  2, -1,  2 },
 {  4, -2,  0,  2 },
 {  4,  0,  0,  2 },
 {  4,  2,  0,  2 },
 {  4, -2,  1,  2 },
 {  4,  0,  1,  2 },
 {  4,  0,  2,  2 },
 {  4,  0, -2,  3 },
 {  4,  0, -1,  3 },
 {  4,  0,  0,  3 },
 {  4, -2, -2,  4 },
 {  5,  0, -2, -3 },
 {  5,  0,  0, -3 },
 {  5,  0, -3, -2 },
 {  5,  0, -2, -2 },
 {  5, -2, -1, -2 },
 {  5,  0, -1, -2 },
 {  5, -2,  0, -2 },
 {  5,  0,  0, -2 },
 {  5,  0,  1, -2 },
 {  5,  0, -4, -1 },
 {  5,  0, -3, -1 },
 {  5,  2, -3, -1 },
 {  5, -2, -2, -1 },
 {  5,  0, -2, -1 },
 {  5,  2, -2, -1 },
 {  5, -2, -1, -1 },
 {  5,  0, -1, -1 },
 {  5,  2, -1, -1 },
 {  5, -2,  0, -1 },
 {  5,  0,  0, -1 },
 {  5, -2,  1, -1 },
 {  5,  0,  1, -1 },
 {  5,  0, -4,  0 },
 {  5,  0, -3,  0 },
 {  5,  2, -3,  0 },
 {  5, -2, -2,  0 },
 {  5,  0, -2,  0 },
 {  5,  2, -2,  0 },
 {  5, -2, -1,  0 },
 {  5,  0, -1,  0 },
 {  5,  2, -1,  0 },
 {  5, -4,  0,  0 },
 {  5, -2,  0,  0 },
 {  5,  0,  0,  0 },
 {  5,  2,  0,  0 },
 {  5, -2,  1,  0 },
 {  5,  0,  1,  0 },
 {  5,  2,  1,  0 },
 {  5, -2,  2,  0 },
 {  5,  0,  2,  0 },
 {  5,  0, -4,  1 },
 {  5,  0, -3,  1 },
 {  5, -2, -2,  1 },
 {  5,  0, -2,  1 },
 {  5,  2, -2,  1 },
 {  5, -2, -1,  1 },
 {  5,  0, -1,  1 },
 {  5,  2, -1,  1 },
 {  5, -2,  0,  1 },
 {  5,  0,  0,  1 },
 {  5,  2,  0,  1 },
 {  5,  0,  1,  1 },
 {  5,  0,  2,  1 },
 {  5,  0, -3,  2 },
 {  5,  0, -2,  2 },
 {  5,  0, -1,  2 },
 {  5,  0,  0,  2 },
 {  5,  0,  1,  2 },
 {  5, -2, -2,  3 },
 {  6,  0, -3, -4 },
 {  6,  0, -2, -4 },
 {  6,  0, -1, -4 },
 {  6,  0,  0, -4 },
 {  6,  0, -3, -3 },
 {  6,  0, -2, -3 },
 {  6,  0, -1, -3 },
 {  6,  0,  0, -3 },
 {  6,  0,  1, -3 },
 {  6,  0, -4, -2 },
 {  6,  0, -3, -2 },
 {  6,  2, -3, -2 },
 {  6, -2, -2, -2 },
 {  6,  0, -2, -2 },
 {  6,  2, -2, -2 },
 {  6, -2, -1, -2 },
 {  6,  0, -1, -2 },
 {  6,  2, -1, -2 },
 {  6, -2,  0, -2 },
 {  6,  0,  0, -2 },
 {  6,  2,  0, -2 },
 {  6, -2,  1, -2 },
 {  6,  0,  1, -2 },
 {  6,  0,  2, -2 },
 {  6,  0,  3, -2 },
 {  6,  0, -5, -1 },
 {  6,  0, -4, -1 },
 {  6,  2, -4, -1 },
 {  6, -2, -3, -1 },
 {  6,  0, -3, -1 },
 {  6,  2, -3, -1 },
 {  6, -2, -2, -1 },
 {  6,  0, -2, -1 },
 {  6,  2, -2, -1 },
 {  6, -2, -1, -1 },
 {  6,  0, -1, -1 },
 {  6,  2, -1, -1 },
 {  6, -2,  0, -1 },
 {  6,  0,  0, -1 },
 {  6,  2,  0, -1 },
 {  6, -2,  1, -1 },
 {  6,  0,  1, -1 },
 {  6,  2,  1, -1 },
 {  6, -2,  2, -1 },
 {  6,  0,  2, -1 },
 {  6,  2,  2, -1 },
 {  6,  0,  3, -1 },
 {  6,  0, -5,  0 },
 {  6,  0, -4,  0 },
 {  6,  2, -4,  0 },
 {  6, -2, -3,  0 },
 {  6,  0, -3,  0 },
 {  6,  2, -3,  0 },
 {  6,  4, -3,  0 },
 {  6, -2, -2,  0 },
 {  6,  0, -2,  0 },
 {  6,  2, -2,  0 },
 {  6,  4, -2,  0 },
 {  6, -2, -1,  0 },
 {  6,  0, -1,  0 },
 {  6,  2, -1,  0 },
 {  6,  4, -1,  0 },
 {  6, -4,  0,  0 },
 {  6, -2,  0,  0 },
 {  6,  0,  0,  0 },
 {  6,  2,  0,  0 },
 {  6, -4,  1,  0 },
 {  6, -2,  1,  0 },
 {  6,  0,  1,  0 },
 {  6,  2,  1,  0 },
 {  6, -2,  2,  0 },
 {  6,  0,  2,  0 },
 {  6,  2,  2,  0 },
 {  6, -2,  3,  0 },
 {  6,  0,  3,  0 },
 {  6,  0,  4,  0 },
 {  6,  0, -5,  1 },
 {  6,  0, -4,  1 },
 {  6,  0, -3,  1 },
 {  6,  2, -3,  1 },
 {  6, -2, -2,  1 },
 {  6,  0, -2,  1 },
 {  6,  2, -2,  1 },
 {  6, -2, -1,  1 },
 {  6,  0, -1,  1 },
 {  6,  2, -1,  1 },
 {  6, -2,  0,  1 },
 {  6,  0,  0,  1 },
 {  6,  2,  0,  1 },
 {  6, -2,  1,  1 },
 {  6,  0,  1,  1 },
 {  6,  2,  1,  1 },
 {  6,  0,  2,  1 },
 {  6,  0,  3,  1 },
 {  6,  0, -4,  2 },
 {  6,  0, -3,  2 },
 {  6, -2, -2,  2 },
 {  6,  0, -2,  2 },
 {  6, -2, -1,  2 },
 {  6,  0, -1,  2 },
 {  6,  0,  0,  2 },
 {  6,  0, -2,  3 },
 {  7,  0, -3, -2 },
 {  7,  0, -2, -2 },
 {  7,  0, -4, -1 },
 {  7,  0, -3, -1 },
 {  7,  0, -2, -1 },
 {  7,  0,  0, -1 },
 {  7,  0, -5,  0 },
 {  7,  0, -4,  0 },
 {  7,  0, -3,  0 },
 {  7,  2, -3,  0 },
 {  7, -2, -2,  0 },
 {  7,  0, -2,  0 },
 {  7,  2, -2,  0 },
 {  7, -2, -1,  0 },
 {  7,  0, -1,  0 },
 {  7, -2,  0,  0 },
 {  7,  0,  0,  0 },
 {  7,  0, -3,  1 },
 {  7,  0, -2,  1 },
 {  7,  0, -1,  1 },
 {  7,  0,  0,  1 },
 {  8,  0, -3, -3 },
 {  8,  0, -2, -3 },
 {  8,  0, -1, -3 },
 {  8,  0, -4, -2 },
 {  8,  0, -3, -2 },
 {  8,  0, -2, -2 },
 {  8,  0, -1, -2 },
 {  8,  0,  0, -2 },
 {  8,  0,  1, -2 },
 {  8,  0, -5, -1 },
 {  8,  0, -4, -1 },
 {  8,  0, -3, -1 },
 {  8,  2, -3, -1 },
 {  8,  0, -2, -1 },
 {  8,  2, -2, -1 },
 {  8,  0, -1, -1 },
 {  8,  2, -1, -1 },
 {  8,  0,  0, -1 },
 {  8,  0,  1, -1 },
 {  8,  0,  2, -1 },
 {  8,  0, -5,  0 },
 {  8,  0, -4,  0 },
 {  8,  2, -4,  0 },
 {  8, -2, -3,  0 },
 {  8,  0, -3,  0 },
 {  8,  2, -3,  0 },
 {  8, -2, -2,  0 },
 {  8,  0, -2,  0 },
 {  8,  2, -2,  0 },
 {  8, -2, -1,  0 },
 {  8,  0, -1,  0 },
 {  8,  2, -1,  0 },
 {  8, -2,  0,  0 },
 {  8,  0,  0,  0 },
 {  8,  2,  0,  0 },
 {  8, -2,  1,  0 },
 {  8,  0,  1,  0 },
 {  8,  0,  2,  0 },
 {  8,  0, -4,  1 },
 {  8,  0, -3,  1 },
 {  8,  0, -2,  1 },
 {  8,  0, -1,  1 },
 {  8,  0,  0,  1 },
 {  8,  0,  1,  1 },
 {  8,  0, -3,  2 },
 {  9,  0, -4,  0 },
 {  9,  0, -3,  0 },
 { 10,  0, -4, -1 },
 { 10,  0, -3, -1 },
 { 10,  0, -2, -1 },
 { 10,  0, -1, -1 },
 { 10,  0, -5,  0 },
 { 10,  0, -4,  0 },
 { 10,  0, -3,  0 },
 { 10,  0, -2,  0 },
 { 10,  0, -1,  0 },
 { 10,  0,  0,  0 }
} };

constexpr array<double, 1023> g_MAIN_S1_Nominal
{ {
    -411.60287,
       0.42034,
      -0.00059,
       0.00016,
      -0.08019,
      39.53393,
   22639.55000,
     -45.10032,
       0.09092,
      -0.00019,
       0.00068,
      -1.37259,
     769.02326,
      -3.99767,
       0.01297,
      -0.00004,
       0.00009,
      -0.05684,
      36.12364,
      -0.32866,
       0.00153,
      -0.00354,
       1.93367,
      -0.02602,
       0.00016,
      -0.00023,
       0.11100,
      -0.00201,
       0.00002,
      -0.00002,
       0.00665,
      -0.00015,
       0.00041,
      -0.00001,
       0.00003,
      -0.00002,
      -0.00024,
       0.00006,
      -0.00332,
       0.00064,
      -0.04702,
       0.00007,
      -0.00003,
       0.00623,
      -0.66940,
       0.00106,
      -0.00017,
       0.05165,
      -9.67938,
       0.02429,
      -0.00004,
      -0.00063,
       0.30157,
    -147.32654,
      -0.08049,
       0.00013,
      -0.00023,
       0.07501,
    -666.44186,
       0.41342,
      -0.00083,
       0.00011,
      -0.08282,
    -109.38419,
       0.26338,
      -0.00062,
      -0.00003,
       0.02677,
      -7.63041,
       0.04338,
      -0.00015,
       0.00166,
      -0.54594,
       0.00526,
      -0.00003,
       0.00014,
      -0.03947,
       0.00055,
       0.00001,
      -0.00287,
       0.00005,
      -0.00021,
      -0.00002,
      -0.00009,
       0.00002,
      -0.00119,
       0.00014,
      -0.01514,
       0.00003,
       0.00100,
      -0.19311,
       0.00060,
       0.00473,
      -2.56813,
       0.00013,
      -0.00355,
      -7.44804,
       0.00519,
      -0.00001,
      -0.00040,
      -1.16177,
       0.00270,
       0.00040,
      -0.06513,
       0.00035,
       0.00002,
      -0.00341,
       0.00003,
      -0.00015,
      -0.00003,
      -0.00035,
       0.00002,
      -0.00416,
       0.00002,
       0.00008,
      -0.05142,
       0.00003,
      -0.00016,
      -0.10136,
       0.00007,
      -0.01796,
       0.00004,
      -0.00095,
      -0.00005,
      -0.00009,
      -0.00108,
      -0.00128,
      -0.00032,
      -0.00002,
      -0.00002,
       0.00011,
       0.00002,
       0.00020,
       0.00254,
       0.00007,
      -0.00001,
       0.00005,
       0.00025,
      -0.00004,
       0.00229,
      -0.00032,
       0.04304,
      -0.00007,
      -0.00008,
      -0.00033,
      -0.00037,
      -0.00005,
       0.00002,
       0.00014,
      -0.00002,
       0.00069,
       0.00002,
      -0.00011,
      -0.00694,
       0.00013,
      -0.00100,
      -0.13636,
       0.00013,
       0.00002,
      -0.01281,
      -0.55694,
       0.00155,
      -0.00212,
      -0.12073,
       0.00060,
       0.00012,
      -0.01483,
       0.00012,
       0.00001,
      -0.00149,
       0.00002,
      -0.00014,
      -0.00001,
      -0.00005,
       0.00001,
      -0.00072,
       0.00009,
      -0.00955,
       0.00002,
       0.00069,
      -0.12812,
       0.00038,
       0.00001,
       0.00289,
      -1.75296,
      -0.00142,
       0.00019,
      -0.02379,
     -18.58467,
       0.01598,
      -0.00011,
       0.00123,
      -0.58332,
    -124.98806,
       0.25429,
      -0.00054,
       0.00011,
      -0.04012,
      -8.45308,
       0.04450,
      -0.00015,
      -0.00077,
      -0.58399,
       0.00540,
      -0.00003,
      -0.00006,
      -0.04074,
       0.00056,
      -0.00285,
       0.00005,
      -0.00020,
      -0.00001,
       0.00011,
      -0.00001,
       0.00155,
      -0.00012,
       0.02296,
      -0.00006,
      -0.00071,
       0.34966,
       0.00063,
      -0.00008,
       1.07773,
       0.00000,
       0.00017,
      17.95512,
      -0.03587,
       0.00007,
      -0.00157,
       1.26186,
      -0.00659,
       0.00002,
       0.00028,
       0.09205,
      -0.00084,
       0.00002,
       0.00679,
      -0.00009,
       0.00050,
       0.00004,
       0.00004,
       0.00038,
       0.00361,
      -0.00300,
       0.00001,
       0.00003,
      -0.03968,
       0.00011,
       0.00002,
      -0.00849,
       0.00005,
      -0.00102,
       0.00001,
      -0.00011,
      -0.00001,
       0.00007,
      -0.00011,
       0.00000,
      -0.00056,
      -0.00009,
      -0.00001,
       0.00002,
       0.00001,
       0.00027,
       0.00050,
       0.00005,
       0.00002,
       0.00032,
       0.00812,
      -0.00002,
       0.00004,
       0.01331,
      -0.00003,
      -0.00002,
       0.00133,
       0.00012,
       0.00001,
       0.00003,
       0.00049,
      -0.00004,
       0.00965,
      -0.00002,
      -0.00021,
       0.24697,
      -0.00052,
       0.00163,
       0.33983,
      -0.00075,
      -0.00065,
       0.03300,
      -0.00016,
      -0.00005,
       0.00295,
      -0.00003,
       0.00026,
       0.00002,
       0.00005,
      -0.00001,
       0.00088,
      -0.00013,
       0.01544,
      -0.00111,
       0.29422,
      -0.00072,
      -0.00434,
       7.37173,
      -0.01541,
       0.00003,
      -0.00008,
       0.06314,
       8.05076,
      -0.01816,
       0.00004,
       0.00001,
      -0.01644,
       0.75173,
      -0.00378,
       0.00001,
      -0.00133,
       0.06439,
      -0.00055,
      -0.00011,
       0.00538,
      -0.00007,
       0.00044,
       0.00004,
       0.00010,
      -0.00003,
       0.00178,
      -0.00036,
       0.02949,
      -0.00002,
       0.00001,
      -0.00365,
       0.47842,
       0.00003,
       0.00005,
      -0.02804,
       8.60582,
      -0.02157,
       0.00012,
      -0.00015,
      -0.06103,
     205.44315,
      -0.42624,
       0.00093,
      -0.00286,
       2.14619,
     164.73458,
      -0.38215,
       0.00089,
       0.00018,
      -0.37453,
      14.53078,
      -0.07373,
       0.00025,
       0.00001,
      -0.02887,
       1.17704,
      -0.01007,
       0.00005,
      -0.00232,
       0.09320,
      -0.00117,
      -0.00019,
       0.00730,
      -0.00012,
      -0.00001,
       0.00057,
      -0.00001,
       0.00004,
       0.00003,
       0.00036,
      -0.00011,
       0.00497,
      -0.00111,
       0.06850,
      -0.00007,
       0.00005,
      -0.01077,
       0.94827,
      -0.00054,
       0.00024,
      -0.09154,
      13.19400,
       0.00467,
      -0.00003,
       0.00037,
      -0.56065,
     211.65487,
      -0.53572,
       0.00282,
       0.00003,
      -0.00850,
       0.17903,
    4586.43061,
      -9.36601,
       0.01993,
      -0.00004,
       0.00015,
      -0.07518,
      55.17801,
    2369.91227,
      -5.74170,
       0.01372,
      -0.00003,
       0.00001,
       0.00056,
      -6.38325,
     191.95575,
      -0.99023,
       0.00344,
      -0.00001,
       0.00008,
      -0.45380,
      14.37964,
      -0.12386,
       0.00059,
       0.00002,
      -0.03336,
       1.05949,
      -0.01332,
       0.00008,
      -0.00247,
       0.07765,
      -0.00131,
       0.00001,
      -0.00018,
       0.00568,
      -0.00012,
      -0.00001,
       0.00041,
      -0.00001,
       0.00003,
       0.00001,
       0.00015,
      -0.00003,
       0.00173,
      -0.00020,
       0.01819,
      -0.00003,
      -0.00116,
       0.17624,
      -0.00024,
       0.00002,
      -0.00346,
       2.48897,
      -0.00001,
       0.00017,
      -0.00228,
     -28.39810,
       0.06742,
      -0.00016,
       0.00198,
      -1.43724,
     -24.35910,
       0.06541,
      -0.00017,
       0.00003,
       0.08472,
      -2.91464,
       0.01605,
      -0.00006,
       0.00889,
      -0.29023,
       0.00262,
      -0.00001,
       0.00085,
      -0.02676,
       0.00035,
       0.00008,
      -0.00236,
       0.00004,
      -0.00020,
      -0.00002,
      -0.00002,
      -0.00049,
       0.00004,
      -0.00944,
       0.00002,
       0.00006,
      -0.24990,
       0.00060,
       0.00002,
      -0.00654,
      -2.52138,
       0.00515,
      -0.00001,
       0.00004,
      -0.02349,
      -0.18576,
       0.00096,
      -0.00200,
      -0.01343,
       0.00012,
      -0.00014,
      -0.00066,
      -0.00001,
      -0.00009,
      -0.00093,
      -0.00003,
      -0.00188,
       0.00007,
       0.00038,
       0.00002,
       0.00004,
      -0.00001,
      -0.00001,
      -0.00002,
       0.00002,
      -0.00017,
      -0.00053,
      -0.00005,
       0.00030,
       0.00003,
      -0.00014,
      -0.00399,
      -0.00008,
      -0.01182,
       0.00003,
      -0.00102,
       0.00518,
      -0.00004,
       0.00037,
      -0.00013,
      -0.00297,
      -0.08311,
       0.00020,
      -0.00189,
      -0.23141,
       0.00075,
       0.00002,
      -0.01948,
       0.06507,
       0.00020,
      -0.00084,
       0.00330,
       0.00005,
      -0.00004,
      -0.00011,
       0.00001,
      -0.00005,
      -0.00005,
       0.00001,
      -0.00174,
       0.00011,
      -0.04442,
       0.00011,
      -0.00028,
      -1.22412,
       0.00298,
      -0.00001,
       0.00006,
      -0.02979,
      -3.20968,
       0.01215,
      -0.00004,
       0.00025,
      -0.25304,
       0.40423,
       0.00375,
      -0.00002,
       0.00005,
      -0.01062,
       0.00037,
       0.00083,
      -0.00051,
      -0.00389,
       0.00014,
      -0.00002,
      -0.00062,
       0.00002,
      -0.00007,
      -0.00006,
      -0.00059,
       0.00002,
      -0.00623,
       0.00002,
       0.00001,
      -0.04189,
       0.00007,
       0.00116,
       0.27377,
      -0.00138,
       0.00891,
       0.14990,
      -0.00096,
       0.00020,
       0.02244,
      -0.00022,
      -0.00001,
       0.00256,
      -0.00004,
       0.00026,
       0.00002,
      -0.00002,
      -0.00020,
      -0.00036,
      -0.00001,
      -0.00127,
      -0.00009,
      -0.00191,
       0.00001,
      -0.00039,
      -0.00006,
       0.00000,
       0.00000,
       0.00004,
      -0.00014,
      -0.00002,
      -0.00002,
       0.00001,
       0.00004,
       0.00003,
       0.00001,
       0.00033,
       0.00086,
       0.00054,
       0.00009,
       0.00001,
       0.00027,
       0.00755,
      -0.00004,
       0.00004,
       0.01741,
      -0.00009,
       0.00007,
       0.00971,
      -0.00006,
      -0.00003,
       0.00160,
      -0.00001,
       0.00020,
       0.00002,
      -0.00002,
      -0.00002,
       0.00560,
      -0.00004,
       0.00002,
       0.15781,
      -0.00090,
       0.00108,
       0.30874,
      -0.00160,
       0.00091,
       0.15227,
      -0.00089,
      -0.00052,
       0.02411,
      -0.00021,
      -0.00009,
       0.00290,
      -0.00004,
      -0.00001,
       0.00031,
       0.00003,
      -0.00005,
      -0.00034,
       0.00004,
      -0.00020,
       0.00007,
       0.09982,
      -0.00078,
       0.00001,
       0.00043,
       2.73198,
      -0.01533,
       0.00006,
       0.00002,
       0.02289,
       4.37416,
      -0.02273,
       0.00008,
      -0.00001,
       0.00666,
       1.87083,
      -0.01114,
       0.00004,
      -0.00006,
      -0.00732,
       0.28251,
      -0.00251,
       0.00001,
      -0.00121,
       0.03274,
      -0.00041,
      -0.00014,
       0.00336,
      -0.00006,
      -0.00001,
       0.00032,
       0.00003,
      -0.00007,
       0.00001,
      -0.00062,
       0.00010,
      -0.00431,
       0.00042,
      -0.00070,
      -0.00002,
       0.00002,
       0.00023,
       1.18682,
      -0.00934,
       0.00007,
       0.00013,
       0.00662,
      30.77247,
      -0.16977,
       0.00061,
       0.00017,
       0.30086,
      38.42974,
      -0.20134,
       0.00072,
      -0.00003,
      -0.02391,
      13.89903,
      -0.08500,
       0.00035,
      -0.00061,
      -0.06601,
       1.97772,
      -0.01781,
       0.00009,
       0.00001,
      -0.00980,
       0.21853,
      -0.00274,
       0.00002,
      -0.00109,
       0.02149,
      -0.00036,
      -0.00011,
       0.00198,
      -0.00004,
       0.00017,
       0.00001,
      -0.00001,
      -0.00004,
       0.00064,
      -0.00004,
       0.02682,
      -0.00003,
      -0.00039,
      -0.35759,
       0.00220,
      -0.01875,
      -0.63523,
       0.00353,
      -0.00001,
       0.00051,
      -0.28911,
       0.00186,
       0.00004,
       0.00183,
      -0.05070,
       0.00048,
       0.00032,
      -0.00668,
       0.00009,
       0.00004,
      -0.00077,
       0.00001,
      -0.00008,
      -0.00005,
      -0.00369,
       0.00002,
       0.00008,
      -0.03514,
       0.00019,
       0.00090,
      -0.02171,
       0.00012,
       0.00001,
      -0.00208,
       0.00002,
      -0.00002,
      -0.00008,
       0.00001,
       0.00020,
       0.00023,
       0.00005,
       0.00000,
      -0.00003,
       0.00001,
      -0.00013,
      -0.00052,
      -0.00003,
      -0.00011,
      -0.00002,
       0.00012,
       0.00002,
      -0.00007,
      -0.00189,
       0.00001,
      -0.00003,
      -0.00693,
       0.00004,
      -0.00040,
      -0.00154,
       0.00003,
      -0.00021,
       0.00102,
      -0.00003,
       0.00014,
      -0.00059,
      -0.01686,
       0.00010,
      -0.00031,
      -0.05725,
       0.00039,
      -0.00335,
      -0.01350,
       0.00024,
      -0.00002,
      -0.00137,
       0.00439,
       0.00006,
      -0.00019,
       0.00050,
       0.00001,
      -0.00002,
       0.00001,
      -0.00006,
      -0.00046,
       0.00001,
       0.00392,
      -0.00004,
       0.00016,
       0.00475,
      -0.00005,
       0.00005,
       0.00181,
      -0.00002,
       0.00035,
       0.00005,
       0.00001,
      -0.00001,
      -0.00008,
      -0.00004,
      -0.00001,
       0.00000,
       0.00001,
       0.00004,
       0.00004,
       0.00002,
       0.00020,
       0.00059,
       0.00059,
       0.00024,
       0.00005,
       0.00014,
       0.00312,
      -0.00003,
       0.00003,
       0.00836,
      -0.00007,
       0.00003,
       0.00744,
      -0.00007,
      -0.00002,
       0.00285,
      -0.00003,
      -0.00002,
       0.00056,
       0.00008,
       0.00001,
       0.00002,
       0.00176,
      -0.00002,
       0.00002,
       0.03900,
      -0.00039,
       0.00044,
       0.09033,
      -0.00080,
       0.00018,
       0.07142,
      -0.00064,
      -0.00025,
       0.02510,
      -0.00026,
      -0.00016,
       0.00477,
      -0.00006,
      -0.00003,
       0.00069,
      -0.00001,
       0.00009,
       0.00020,
       0.01376,
      -0.00017,
       0.00018,
       0.29255,
      -0.00287,
       0.00002,
       0.00364,
       0.57156,
      -0.00509,
       0.00003,
       0.00007,
       0.39451,
      -0.00359,
       0.00002,
      -0.00008,
      -0.00214,
       0.12616,
      -0.00132,
      -0.00001,
      -0.00094,
       0.02285,
      -0.00031,
      -0.00018,
       0.00318,
      -0.00005,
      -0.00003,
       0.00038,
       0.00004,
       0.00001,
       0.00028,
      -0.00478,
       0.00005,
      -0.00026,
      -0.01282,
       0.00012,
      -0.00007,
      -0.01064,
       0.00010,
       0.00008,
      -0.00396,
       0.00004,
       0.00004,
      -0.00083,
       0.00001,
      -0.00013,
      -0.00002,
      -0.00006,
      -0.00051,
       0.00001,
      -0.00061,
       0.00001,
      -0.00021,
      -0.00002,
       0.00001,
      -0.00002,
      -0.00001,
      -0.00004,
      -0.00016,
      -0.00012,
       0.00002,
      -0.00001,
      -0.00025,
      -0.00098,
       0.00001,
      -0.00005,
      -0.00068,
       0.00001,
      -0.00004,
      -0.00005,
      -0.00001,
       0.00005,
       0.00006,
       0.00011,
       0.00007,
       0.00003,
       0.00002,
       0.00002,
       0.00002,
       0.00006,
       0.00019,
       0.00025,
       0.00015,
       0.00005,
       0.00001,
       0.00003,
       0.00058,
       0.00170,
      -0.00002,
       0.00196,
      -0.00003,
       0.00112,
      -0.00001,
       0.00037,
       0.00008,
       0.00001,
       0.00018,
       0.00324,
      -0.00005,
       0.00005,
       0.00840,
      -0.00011,
       0.00001,
       0.00866,
      -0.00011,
      -0.00005,
       0.00455,
      -0.00006,
      -0.00004,
       0.00139,
      -0.00002,
      -0.00001,
       0.00029,
       0.00005,
      -0.00007,
      -0.00024,
      -0.00029,
      -0.00017,
      -0.00006,
      -0.00001,
      -0.00001,
      -0.00002,
      -0.00002,
       0.00003,
       0.00005,
       0.00004,
       0.00002,
       0.00004,
       0.00012,
       0.00017,
       0.00013,
       0.00006,
       0.00002
} };

constexpr array<double, 1023> g_MAIN_S1_LLR
{ {
   -4.1159573394034464e+002,
    4.2032545247910946e-001,
   -5.8996898355993734e-004,
    1.5999224588998434e-004,
   -8.0187327780507925e-002,
    3.9533300966452579e+001,
    2.2639585879999784e+004,
   -4.5099609183190310e+001,
    9.0916994547684135e-002,
   -1.8999224588998434e-004,
    6.7997934422036981e-004,
   -1.3725700882644198e+000,
    7.6902571867111135e+002,
   -3.9976133688320399e+000,
    1.2969592010282917e-002,
   -4.0000000000000003e-005,
    9.0000000000000006e-005,
   -5.6839244828719195e-002,
    3.6123814129090093e+001,
   -3.2865586543214154e-001,
    1.5299526792758415e-003,
   -3.5399567052666744e-003,
    1.9336824738488418e+000,
   -2.6019715516762902e-002,
    1.5999699270073610e-004,
   -2.2999786245648000e-004,
    1.1100089397044806e-001,
   -2.0099807621083198e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
    6.6500634929455380e-003,
   -1.5000173951148780e-004,
    4.1000434877871956e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.3999647491972467e-004,
    6.0000000000000002e-005,
   -3.3199034030053608e-003,
    6.3996817983555138e-004,
   -4.7018579161268337e-002,
    6.9999999999999994e-005,
   -3.0000000000000001e-005,
    6.2296940378636818e-003,
   -6.6937872520878050e-001,
    1.0599428920506418e-003,
   -1.6998611154247728e-004,
    5.1647380979449500e-002,
   -9.6790567959049891e+000,
    2.4288759165808704e-002,
   -4.0000000000000003e-005,
   -6.2995306964755220e-004,
    3.0155423091948980e-001,
   -1.4732138419770178e+002,
   -8.0485756779276163e-002,
    1.2999612294499217e-004,
   -2.2998611193059370e-004,
    7.5006159646341913e-002,
   -6.6641753993238694e+002,
    4.1339790248083047e-001,
   -8.2994218887740977e-004,
    1.0999612294499217e-004,
   -8.2815647886971808e-002,
   -1.0938036372441518e+002,
    2.6336624054187213e-001,
   -6.1995307003566858e-004,
   -3.0000000000000001e-005,
    2.6768637581412277e-002,
   -7.6301553320624675e+000,
    4.3377800674324000e-002,
   -1.4999612294499216e-004,
    1.6599154655578634e-003,
   -5.4592266032453884e-001,
    5.2597397476500021e-003,
   -3.0000000000000001e-005,
    1.4000086975574390e-004,
   -3.9468809119166517e-002,
    5.4997732148232231e-004,
    1.0000000000000001e-005,
   -2.8699183274130229e-003,
    5.0000000000000002e-005,
   -2.0999560516398076e-004,
   -2.0000000000000002e-005,
   -8.9993865652492929e-005,
    2.0000000000000002e-005,
   -1.1899219453841306e-003,
    1.3998773130498583e-004,
   -1.5138965730477152e-002,
    3.0000000000000001e-005,
    9.9991435746851027e-004,
   -1.9309647372217842e-001,
    5.9994716713768193e-004,
    4.7295767471812302e-003,
   -2.5679459329852983e+000,
    1.2998772936440373e-004,
   -3.5496785749220628e-003,
   -7.4474952294939296e+000,
    5.1895322388652177e-003,
   -1.0000000000000001e-005,
   -3.9996157725738035e-004,
   -1.1616866904843302e+000,
    2.6997615238340497e-003,
    3.9996331288770390e-004,
   -6.5125430804016088e-002,
    3.4996857825569992e-004,
    2.0000000000000002e-005,
   -3.4097650171113440e-003,
    3.0000000000000001e-005,
   -1.4998860106072974e-004,
   -3.0000000000000001e-005,
   -3.4996493303832892e-004,
    2.0000000000000002e-005,
   -4.1595529644513803e-003,
    2.0000000000000002e-005,
    7.9993865652492929e-005,
   -5.1414425180471221e-002,
    3.0000000000000001e-005,
   -1.5998159695747877e-004,
   -1.0134887457024104e-001,
    6.9993865652492917e-005,
   -1.7958047296051272e-002,
    4.0000000000000003e-005,
   -9.4989832497149494e-004,
   -5.0000000000000002e-005,
   -8.9987731304985851e-005,
   -1.0798422449238381e-003,
   -1.2798159719034862e-003,
   -3.1995092521994341e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0998159692189333e-004,
    2.0000000000000002e-005,
    1.9997546253880082e-004,
    2.5397239523170926e-003,
    6.9993865625803844e-005,
   -1.0000000000000001e-005,
    4.9999999982207283e-005,
    2.4998246662425908e-004,
   -3.9999999982207283e-005,
    2.2898387182755538e-003,
   -3.1997158544820760e-004,
    4.3036852149411435e-002,
   -6.9999999973310921e-005,
   -7.9999999973310934e-005,
   -3.2997632604020010e-004,
   -3.6997720083256121e-004,
   -4.9999999982207283e-005,
    2.0000000000000002e-005,
    1.4000086971126209e-004,
   -2.0000000000000002e-005,
    6.8997894061651279e-004,
    2.0000000000000002e-005,
   -1.0999999996441456e-004,
   -6.9397623496818854e-003,
    1.2999999995551818e-004,
   -9.9994942485945619e-004,
   -1.3635527134034794e-001,
    1.2999999995551818e-004,
    2.0000000000000002e-005,
   -1.2809306653630607e-002,
   -5.5692003453874528e-001,
    1.5499176474605520e-003,
   -2.1198888374956821e-003,
   -1.2072586850794553e-001,
    5.9996469983172656e-004,
    1.1999999995551821e-004,
   -1.4829514347480970e-002,
    1.1999999995551821e-004,
    1.0000000000000001e-005,
   -1.4899517456855838e-003,
    2.0000000000000002e-005,
   -1.4000086971126209e-004,
   -1.0000000000000001e-005,
   -4.9999999982207283e-005,
    1.0000000000000001e-005,
   -7.2000521750913268e-004,
    8.9999999964414566e-005,
   -9.5500647457892711e-003,
    2.0000000000000002e-005,
    6.8999184684151752e-004,
   -1.2812063098189327e-001,
    3.7999485270396850e-004,
    1.0000000000000001e-005,
    2.8899630362230668e-003,
   -1.7529659229032846e+000,
   -1.4199772108449472e-003,
    1.8999224582770983e-004,
   -2.3789613927173651e-002,
   -1.8584706832567541e+001,
    1.5979758557932412e-002,
   -1.0999999996441456e-004,
    1.2299573515797723e-003,
   -5.8330993741777959e-001,
   -1.2498811853267014e+002,
    2.5428571489518664e-001,
   -5.3998061415002090e-004,
    1.0999999996441456e-004,
   -4.0119364730095274e-002,
   -8.4530981740238644e+000,
    4.4499320857384859e-002,
   -1.4999612289161402e-004,
   -7.6999097899966673e-004,
   -5.8399224155828255e-001,
    5.3999243392690577e-003,
   -2.9999999991103641e-005,
   -5.9999999982207282e-005,
   -4.0740230310782437e-002,
    5.5999485457337969e-004,
   -2.8500182619425442e-003,
    4.9999999982207283e-005,
   -2.0000173944031695e-004,
   -1.0000000000000001e-005,
    1.1000086972015847e-004,
   -1.0000000000000001e-005,
    1.5499543568756049e-003,
   -1.1999999995551821e-004,
    2.2959263029679815e-002,
   -6.0000000758440126e-005,
   -7.0996644047197814e-004,
    3.4964830675722081e-001,
    6.2996557073402700e-004,
   -7.9999999973310934e-005,
    1.0776929071813304e+000,
   -1.1643492701065334e-012,
    1.7000608706358361e-004,
    1.7954467429825353e+001,
   -3.5868072876963038e-002,
    6.9999999973310921e-005,
   -1.5699450438443898e-003,
    1.2618161650053286e+000,
   -6.5896596303992002e-003,
    2.0000000000000002e-005,
    2.7998385298776872e-004,
    9.2046948009017429e-002,
   -8.3995642901608524e-004,
    2.0000000000000002e-005,
    6.7897815393456063e-003,
   -8.9999999964414566e-005,
    4.9998507580252722e-004,
    3.9999999991103643e-005,
    3.9999999991103643e-005,
    3.7997106686427421e-004,
    3.6097475224265130e-003,
   -2.9997835042503917e-003,
    1.0000000000000001e-005,
    2.9993865643596560e-005,
   -3.9677107744042192e-002,
    1.0999386561690749e-004,
    2.0000000388116424e-005,
   -8.4893890732562621e-003,
    4.9999999982207283e-005,
   -1.0199298657208036e-003,
    1.0000000000000001e-005,
   -1.0999386561690749e-004,
   -1.0000000000000001e-005,
    6.9993865625803844e-005,
   -1.0998773126940041e-004,
   -7.5226604952611881e-009,
   -5.5993865594109284e-004,
   -8.9987731278296778e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.6995092521994339e-004,
    4.9990798401116103e-004,
    4.9993865652492925e-005,
    2.0000000000000002e-005,
    3.1995179458757088e-004,
    8.1188229736049913e-003,
   -2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.3308050994493486e-002,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.3298063078183069e-003,
    1.1998159695747877e-004,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    4.8994739975155165e-004,
   -4.0000000000000003e-005,
    9.6489691892412861e-003,
   -2.0000000000000002e-005,
   -2.0997546260997170e-004,
    2.4694321708693220e-001,
   -5.1993703637430422e-004,
    1.6297949645372397e-003,
    3.3979270311502285e-001,
   -7.4990862115303523e-004,
   -6.4991562564440265e-004,
    3.2996434438908137e-002,
   -1.5998159695747877e-004,
   -4.9993865652492925e-005,
    2.9496844414616777e-003,
   -3.0000000000000001e-005,
    2.5997019763009208e-004,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
    8.7994387467127627e-004,
   -1.2998773130498583e-004,
    1.5438938193938302e-002,
   -1.1099043460659954e-003,
    2.9419934557448169e-001,
   -7.1993576897464459e-004,
   -4.3396143001681060e-003,
    7.3712011460145677e+000,
   -1.5408628973371889e-002,
    3.0000000000000001e-005,
   -8.0000000000000007e-005,
    6.3134271742407960e-002,
    8.0501723580323876e+000,
   -1.8158365218141039e-002,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -1.6438530709394418e-002,
    7.5167636331432242e-001,
   -3.7796658667951133e-003,
    1.0000000000000001e-005,
   -1.3298829357242100e-003,
    6.4385513274555953e-002,
   -5.4995104419268978e-004,
   -1.0999386565249293e-004,
    5.3796343552609304e-003,
   -6.9999999999999994e-005,
    4.3997193675346347e-004,
    4.0000000000000003e-005,
    1.0000086975574391e-004,
   -3.0000000000000001e-005,
    1.7799464378812082e-003,
   -3.5998258646220195e-004,
    2.9489101202063411e-002,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -3.6498204034846661e-003,
    4.7840473082119511e-001,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
   -2.8038574353863838e-002,
    8.6055317947092895e+000,
   -2.1568901297389149e-002,
    1.1999612294499217e-004,
   -1.4999612294499216e-004,
   -6.1026812375840983e-002,
    2.0543595303766264e+002,
   -4.2621768879515037e-001,
    9.2993391621014993e-004,
   -2.8597949118859432e-003,
    2.1460741255442124e+000,
    1.6472861853103672e+002,
   -3.8212969058735435e-001,
    8.8993918041379661e-004,
    1.7998611154247728e-004,
   -3.7451033600200928e-001,
    1.4530277885386322e+001,
   -7.3726190818842491e-002,
    2.4997997719497018e-004,
    1.0000000000000001e-005,
   -2.8868533048228333e-002,
    1.1770012792645836e+000,
   -1.0069497864886401e-002,
    5.0000000000000002e-005,
   -2.3198866534689111e-003,
    9.3197090253788417e-002,
   -1.1699437610301531e-003,
   -1.8999473540823685e-004,
    7.2997779147964603e-003,
   -1.2000000000000000e-004,
   -1.0000000000000001e-005,
    5.6998594495996549e-004,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    3.6000434877871959e-004,
   -1.1000086975574391e-004,
    4.9700469629290062e-003,
   -1.1099893115061672e-003,
    6.8500554304412931e-002,
   -6.9999999999999994e-005,
    5.0000000000000002e-005,
   -1.0769882602748599e-002,
    9.4827608962253374e-001,
   -5.3999184785795211e-004,
    2.3999311564572824e-004,
   -9.1538855566803426e-002,
    1.3194063561797675e+001,
    4.6699433182892727e-003,
   -3.0000000000000001e-005,
    3.6999010718211507e-004,
   -5.6064210406550075e-001,
    2.1165555237566161e+002,
   -5.3571236892975049e-001,
    2.8199117704060103e-003,
    3.0000000000000001e-005,
   -8.4997175203998208e-003,
    1.7902697687415575e-001,
    4.5864383203391226e+003,
   -9.3658634662895146e+000,
    1.9929343790710426e-002,
   -4.0000000000000003e-005,
    1.4999224588998431e-004,
   -7.5177389373332126e-002,
    5.5177057781935460e+001,
    2.3699139356645105e+003,
   -5.7416071078822393e+000,
    1.3719545250575591e-002,
   -3.0000000000000001e-005,
    1.0000000000000001e-005,
    5.5998148059954039e-004,
   -6.3831490688693977e+000,
    1.9195619729415324e+002,
   -9.9021535743421063e-001,
    3.4398885053591485e-003,
   -1.0000000000000001e-005,
    8.0000000000000007e-005,
   -4.5379355183867159e-001,
    1.4379697414390638e+001,
   -1.2385836450364340e-001,
    5.8998322321595980e-004,
    2.0000000000000002e-005,
   -3.3359573716410885e-002,
    1.0594960024585396e+000,
   -1.3319840819853743e-002,
    8.0000000000000007e-005,
   -2.4699730068339547e-003,
    7.7650572853363117e-002,
   -1.3099854336749265e-003,
    1.0000000000000001e-005,
   -1.8000173951148783e-004,
    5.6800469606003087e-003,
   -1.2000086975574391e-004,
   -1.0000000000000001e-005,
    4.1000434839060315e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    1.5000173912337137e-004,
   -3.0000000000000001e-005,
    1.7299525679594348e-003,
   -1.9999473463200402e-004,
    1.8189447730470484e-002,
   -3.0012270247479851e-005,
   -1.1599428889457104e-003,
    1.7623440184335515e-001,
   -2.3998559376146584e-004,
    2.0000000000000002e-005,
   -3.4598243130128984e-003,
    2.4888871030915660e+000,
   -9.9999868040416062e-006,
    1.6998611154247728e-004,
   -2.2798819841528895e-003,
   -2.8397100818380377e+001,
    6.7416467629333704e-002,
   -1.5999612294499219e-004,
    1.9798565996754247e-003,
   -1.4371623834197103e+000,
   -2.4358228257667186e+001,
    6.5406557807381582e-002,
   -1.6998611154247728e-004,
    3.0000000000000001e-005,
    8.4715547794242618e-002,
   -2.9145403611714258e+000,
    1.6049176218736076e-002,
   -6.0000000000000002e-005,
    8.8895484966045463e-003,
   -2.9022054660535113e-001,
    2.6198687662668196e-003,
   -1.0000000000000001e-005,
    8.4995816843303634e-004,
   -2.6759174399409124e-002,
    3.4998171631834158e-004,
    8.0000000000000007e-005,
   -2.3599297708571275e-003,
    4.0000000000000003e-005,
   -1.9999560516398076e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -4.8996667293793312e-004,
    4.0000000000000003e-005,
   -9.4393467713475929e-003,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
   -2.4988251297490727e-001,
    5.9994716713768193e-004,
    2.0000000000000002e-005,
   -6.5394169047861220e-003,
   -2.5211989535398205e+000,
    5.1495391891488414e-003,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
   -2.3487921490233606e-002,
   -1.8574704319388002e-001,
    9.5991962206027335e-004,
   -1.9998317659648854e-003,
   -1.3429085736074225e-002,
    1.1999386565249293e-004,
   -1.3998773130498583e-004,
   -6.5995614336629047e-004,
   -1.0000000000000001e-005,
   -8.9987731304985851e-005,
   -9.2989832303091286e-004,
   -3.0000000000000001e-005,
   -1.8797940367945940e-003,
    6.9987731304985840e-005,
    3.7995879791458901e-004,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
   -1.6998159689520425e-004,
   -5.2994478990937995e-004,
   -4.9999999982207283e-005,
    2.9996932815570831e-004,
    2.9999999991103641e-005,
   -1.3998773125160769e-004,
   -3.9897185923957190e-003,
   -7.9999999973310934e-005,
   -1.1819155037885467e-002,
    2.9999999991103641e-005,
   -1.0199070014227888e-003,
    5.1796153179470917e-003,
   -3.9999999982207283e-005,
    3.6997546247652629e-004,
   -1.2999999995551818e-004,
   -2.9699034944760988e-003,
   -8.3107217151914367e-002,
    1.9999386558132204e-004,
   -1.8899041135789240e-003,
   -2.3140200974508465e-001,
    7.4995943518658525e-004,
    2.0000000000000002e-005,
   -1.9478951091972996e-002,
    6.5067436698430528e-002,
    1.9999386480508920e-004,
   -8.3995468950459737e-004,
    3.2998552073227848e-003,
    4.9999999982207283e-005,
   -3.9999999982207283e-005,
   -1.1000086777957636e-004,
    1.0000000000000001e-005,
   -4.9999999982207283e-005,
   -4.9999999982207283e-005,
    1.0000000000000001e-005,
   -1.7400113062197183e-003,
    1.0999999996441456e-004,
   -4.4420192896920646e-002,
    1.0999999996441456e-004,
   -2.7999612362336504e-004,
   -1.2241239597244933e+000,
    2.9799591618492887e-003,
   -1.0000000000000001e-005,
    5.9999999982207282e-005,
   -2.9789531293046709e-002,
   -3.2096876062154860e+000,
    1.2149816237149513e-002,
   -3.9999999991103643e-005,
    2.4999224580102075e-004,
   -2.5303567070897554e-001,
    4.0422723984270315e-001,
    3.7499475238193950e-003,
   -2.0000000000000002e-005,
    4.9999999982207283e-005,
   -1.0619833428403244e-002,
    3.6963186019266567e-004,
    8.2999097548882618e-004,
   -5.0999398522354493e-004,
   -3.8900503698152153e-003,
    1.3999999995551818e-004,
   -2.0000000000000002e-005,
   -6.2000347686888091e-004,
    2.0000000000000002e-005,
   -6.9999999973310921e-005,
   -5.9999999982207282e-005,
   -5.8997894065209832e-004,
    2.0000000000000002e-005,
   -6.2298025870411193e-003,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
   -4.1888592249990426e-002,
    6.9999999973310921e-005,
    1.1599394126273301e-003,
    2.7376051097820830e-001,
   -1.3799302709093882e-003,
    8.9095189191142985e-003,
    1.4989475424557244e-001,
   -9.5994855395715561e-004,
    1.9999386558132204e-004,
    2.2439244786308304e-002,
   -2.1999386557242567e-004,
   -1.0000000000000001e-005,
    2.5599175481060099e-003,
   -3.9999999991103643e-005,
    2.5998947072751005e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.9998860098955888e-004,
   -3.5997632991246803e-004,
   -1.0000000000000001e-005,
   -1.2699105963355189e-003,
   -8.9987731269400412e-005,
   -1.9098606469199124e-003,
    1.0000000000000001e-005,
   -3.8996932774090282e-004,
   -5.9999999982207282e-005,
    1.2268721703235996e-008,
   -6.4850867194194977e-009,
    3.9999999991103643e-005,
   -1.3998773126050403e-004,
   -1.9985019202383471e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    3.9993865652492926e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    3.2995179419945440e-004,
    8.5987291627325705e-004,
    5.3992025192994223e-004,
    8.9987731304985851e-005,
    1.0000000000000001e-005,
    2.6997019763009211e-004,
    7.5491962034395328e-003,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.7408121234192286e-002,
   -8.9987731304985851e-005,
    6.9993865652492917e-005,
    9.7089421656193827e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    1.5998255786759898e-003,
   -1.0000000000000001e-005,
    1.9997546222185527e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    5.5996159506659432e-003,
   -4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.5779894934692521e-001,
   -8.9991962167215679e-004,
    1.0799026077188566e-003,
    3.0871801726609527e-001,
   -1.5998597839303550e-003,
    9.0991527172908809e-004,
    1.5225907067252273e-001,
   -8.8991875152829648e-004,
   -5.1995630878445289e-004,
    2.4108302281639689e-002,
   -2.0998773091686942e-004,
   -8.9993865652492929e-005,
    2.8997965968995945e-003,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
    3.0997720134522664e-004,
    3.0000000000000001e-005,
   -5.0000000000000002e-005,
   -3.3999034018410119e-004,
    4.0000000000000003e-005,
   -1.9999473540823685e-004,
    6.9999999999999994e-005,
    9.9816821877271866e-002,
   -7.7995729790105968e-004,
    1.0000000000000001e-005,
    4.2997558235895091e-004,
    2.7318889692915320e+000,
   -1.5329221553467351e-002,
    6.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2888805918616802e-002,
    4.3740094938716600e+000,
   -2.2728826686090872e-002,
    8.0000000000000007e-005,
   -1.0000000000000001e-005,
    6.6596364444854103e-003,
    1.8707647458517696e+000,
   -1.1139425136911808e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -7.3196177906305606e-003,
    2.8250054434222499e-001,
   -2.5098722894065129e-003,
    1.0000000000000001e-005,
   -1.2099420211305488e-003,
    3.2738956578608140e-002,
   -4.0998171554210871e-004,
   -1.3999999999999999e-004,
    3.3598955697700362e-003,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
    3.1999033979598473e-004,
    3.0000000000000001e-005,
   -6.9999999999999994e-005,
    1.0000000000000001e-005,
   -6.2000608712585816e-004,
    1.0000000000000001e-004,
   -4.3100339134879160e-003,
    4.1999485438098320e-004,
   -7.0000434994306872e-004,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2999699192450325e-004,
    1.1868256998194016e+000,
   -9.3398815099772336e-003,
    6.9999999999999994e-005,
    1.2999612294499217e-004,
    6.6199084360495791e-003,
    3.0772575081204977e+001,
   -1.6976763932382027e-001,
    6.0998235346021586e-004,
    1.6999612255687576e-004,
    3.0085540153188278e-001,
    3.8429834552033149e+001,
   -2.0133704790524942e-001,
    7.1997847601709165e-004,
   -3.0000000000000001e-005,
   -2.3909624008531132e-002,
    1.3899059643458889e+001,
   -8.4998751180535084e-002,
    3.4998836805874361e-004,
   -6.0998148448070469e-004,
   -6.6009003858817805e-002,
    1.9777270053585612e+000,
   -1.7809756222952739e-002,
    9.0000000000000006e-005,
    1.0000000000000001e-005,
   -9.7998650559810805e-003,
    2.1853112449899584e-001,
   -2.7399656460985744e-003,
    2.0000000000000002e-005,
   -1.0899867014627026e-003,
    2.1490147755081266e-002,
   -3.5999398462523933e-004,
   -1.1000000000000000e-004,
    1.9800139118226218e-003,
   -4.0000000000000003e-005,
    1.7000173912337140e-004,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
    6.3997980905999277e-004,
   -4.0000000000000003e-005,
    2.6819163868398235e-002,
   -2.9999999611883578e-005,
   -3.8998084617448124e-004,
   -3.5757788979026356e-001,
    2.1998884421426303e-003,
   -1.8749021572572678e-002,
   -6.3520815370989836e-001,
    3.5298208471858813e-003,
   -1.0000000000000001e-005,
    5.0996783912509152e-004,
   -2.8910002713341593e-001,
    1.8599058510230314e-003,
    4.0000000000000003e-005,
    1.8299049835959860e-003,
   -5.0698322513450050e-002,
    4.7997558119460167e-004,
    3.1998472361760555e-004,
   -6.6797901960926487e-003,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
   -7.6997367510060202e-004,
    1.0000000000000001e-005,
   -8.0000000000000007e-005,
   -5.0000000000000002e-005,
   -3.6897474745097728e-003,
    2.0000000000000002e-005,
    8.0000000000000007e-005,
   -3.5137537313124745e-002,
    1.8998773091686942e-004,
    8.9991875114018010e-004,
   -2.1708466218293903e-002,
    1.1999386565249293e-004,
    1.0000000000000001e-005,
   -2.0798570721713870e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -7.9994735408236837e-005,
    1.0000000000000001e-005,
    1.9997546260997170e-004,
    2.2997546222185527e-004,
    5.0000000000000002e-005,
   -2.4536560417312939e-008,
   -2.9999999991103641e-005,
    1.0000000000000001e-005,
   -1.2998773126050403e-004,
   -5.1996493208416889e-004,
   -2.9999999991103641e-005,
   -1.0999386484067465e-004,
   -2.0000000000000002e-005,
    1.1999386560801113e-004,
    2.0000000000000002e-005,
   -6.9999999982207274e-005,
   -1.8899412159804957e-003,
    1.0000000000000001e-005,
   -2.9999999991103641e-005,
   -6.9297710874748775e-003,
    3.9999999982207283e-005,
   -3.9997997627639562e-004,
   -1.5399556172891688e-003,
    2.9999999991103641e-005,
   -2.0999386558132204e-004,
    1.0199614536594989e-003,
   -2.9999999991103641e-005,
    1.3999999995551818e-004,
   -5.9000347765401017e-004,
   -1.6860079983748045e-002,
    9.9999999964414565e-005,
   -3.0999699259397977e-004,
   -5.7250247274650445e-002,
    3.8999398449179391e-004,
   -3.3499470572835388e-003,
   -1.3500127298342856e-002,
    2.3999612208869210e-004,
   -2.0000000000000002e-005,
   -1.3699767348697473e-003,
    4.3899543106083686e-003,
    5.9999999982207282e-005,
   -1.8999999993772550e-004,
    4.9999826147493424e-004,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -5.9999999982207282e-005,
   -4.5998333436825001e-004,
    1.0000000000000001e-005,
    3.9198719503338525e-003,
   -3.9999999991103643e-005,
    1.5999999994662187e-004,
    4.7498425414331160e-003,
   -4.9999999982207283e-005,
    4.9999999982207283e-005,
    1.8099368660641159e-003,
   -2.0000000000000002e-005,
    3.4998860015994786e-004,
    4.9999999982207283e-005,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -7.9993865625803856e-005,
   -3.9999999982207283e-005,
   -1.0000000000000001e-005,
   -1.7792718719392219e-014,
    1.0000000000000001e-005,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.9998246632510624e-004,
    5.8993426013644425e-004,
    5.8993425974832785e-004,
    2.3997546183373886e-004,
    5.0000000000000002e-005,
    1.3998773130498583e-004,
    3.1197878076156724e-003,
   -3.0000000000000001e-005,
    3.0000000000000001e-005,
    8.3594152442899722e-003,
   -6.9999999999999994e-005,
    3.0000000000000001e-005,
    7.4394761613223986e-003,
   -6.9999999999999994e-005,
   -2.0000000000000002e-005,
    2.8497975103839398e-003,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    5.5995879713835613e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.7599438247328124e-003,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    3.8998764490379520e-002,
   -3.8998171593022512e-004,
    4.3997558197083453e-004,
    9.0327054369161802e-002,
   -7.9995642736908294e-004,
    1.7999386487626008e-004,
    7.1417617507999423e-002,
   -6.3996556901585390e-004,
   -2.4999085835322900e-004,
    2.5099162687542434e-002,
   -2.5998385347374511e-004,
   -1.6000000000000001e-004,
    4.7698477546273101e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    6.8997893930424880e-004,
   -1.0000000000000001e-005,
    9.0000000000000006e-005,
    2.0000173912337139e-004,
    1.3760086941420145e-002,
   -1.7000086975574392e-004,
    1.8000000000000001e-004,
    2.9255151717808109e-001,
   -2.8699635081669379e-003,
    2.0000000000000002e-005,
    3.6399492688990319e-003,
    5.7156244661304834e-001,
   -5.0899338262202351e-003,
    3.0000000000000001e-005,
    6.9997387627836866e-005,
    3.9451146453452324e-001,
   -3.5899492634654017e-003,
    2.0000000000000002e-005,
   -8.0000000000000007e-005,
   -2.1399673150233139e-003,
    1.2616048387604764e-001,
   -1.3199810821793918e-003,
   -1.0000000000000001e-005,
   -9.3998709910661819e-004,
    2.2850114603891285e-002,
   -3.0999699192450319e-004,
   -1.7999999961188358e-004,
    3.1800156451242470e-003,
   -5.0000000000000002e-005,
   -3.0000000000000001e-005,
    3.8000260810288250e-004,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
    2.7998946887589153e-004,
   -4.7798447195122190e-003,
    5.0000000000000002e-005,
   -2.5998385386186157e-004,
   -1.2819580618409180e-002,
    1.2000000000000000e-004,
   -6.9999999611883575e-005,
   -1.0639642962437685e-002,
    1.0000000000000001e-004,
    8.0000000000000007e-005,
   -3.9598670777521560e-003,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
   -8.2997193481288139e-004,
    1.0000000000000001e-005,
   -1.2999999961188356e-004,
   -2.0000000000000002e-005,
   -6.0000000000000002e-005,
   -5.0996580162972357e-004,
    1.0000000000000001e-005,
   -6.0995966650598363e-004,
    1.0000000000000001e-005,
   -2.0998159618124593e-004,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.9999999991103643e-005,
   -1.5999386521099834e-004,
   -1.1999999956740178e-004,
    2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.5000173864629140e-004,
   -9.8000521547958680e-004,
    1.0000000000000001e-005,
   -4.9999999982207283e-005,
   -6.8000260631911153e-004,
    1.0000000000000001e-005,
   -3.9999999982207283e-005,
   -4.9999999205974438e-005,
   -1.0000000000000001e-005,
    4.9999999982207283e-005,
    5.9999999982207282e-005,
    1.0999999996441456e-004,
    6.9999999973310921e-005,
    2.9999999991103641e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
    1.8998860028449693e-004,
    2.4998246554887342e-004,
    1.4998773052875300e-004,
    5.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    5.7998507442798871e-004,
    1.6999473482606220e-003,
   -2.0000000000000002e-005,
    1.9599342082573983e-003,
   -3.0000000000000001e-005,
    1.1199605339330290e-003,
   -1.0000000000000001e-005,
    3.6998946926400801e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    1.8000173912337139e-004,
    3.2400208644349430e-003,
   -5.0000000000000002e-005,
    5.0000000000000002e-005,
    8.4000464671420763e-003,
   -1.1000000000000000e-004,
    1.0000000000000001e-005,
    8.6600425121261086e-003,
   -1.0999999961188359e-004,
   -5.0000000000000002e-005,
    4.5500195144169264e-003,
   -6.0000000000000002e-005,
   -4.0000000000000003e-005,
    1.3900052127127172e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9000173834713855e-004,
    5.0000000000000002e-005,
   -6.9999999999999994e-005,
   -2.3999473502012042e-004,
   -2.8998860028449689e-004,
   -1.6999386487626008e-004,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.1999999961188358e-004,
    1.7000086897951108e-004,
    1.2999999922376715e-004,
    6.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<double, 1023> g_MAIN_S1_DE405
{ {
   -4.1159574817515249e+002,
    4.2032548138143505e-001,
   -5.8996904561608857e-004,
    1.5999226140402215e-004,
   -8.0187332604691075e-002,
    3.9533301970600597e+001,
    2.2639585659999782e+004,
   -4.5099610307254984e+001,
    9.0916999939499477e-002,
   -1.8999226140402215e-004,
    6.7997937456124253e-004,
   -1.3725701099805798e+000,
    7.6902570369870193e+002,
   -3.9976134296908525e+000,
    1.2969592654144397e-002,
   -4.0000000000000003e-005,
    9.0000000000000006e-005,
   -5.6839245216214383e-002,
    3.6123813073253892e+001,
   -3.2865586724190660e-001,
    1.5299527452105023e-003,
   -3.5399566985949336e-003,
    1.9336823984636129e+000,
   -2.6019715403432151e-002,
    1.5999699512480450e-004,
   -2.2999785954759791e-004,
    1.1100088855380498e-001,
   -2.0099807359283814e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
    6.6500631036979776e-003,
   -1.5000172884558683e-004,
    4.1000432211396706e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.3999645795124582e-004,
    6.0000000000000002e-005,
   -3.3199032234219098e-003,
    6.3996817253443345e-004,
   -4.7018577037686374e-002,
    6.9999999999999994e-005,
   -3.0000000000000001e-005,
    6.2296940374589041e-003,
   -6.6937870149957934e-001,
    1.0599428919928163e-003,
   -1.6998612608688771e-004,
    5.1647381444881534e-002,
   -9.6790565471940102e+000,
    2.4288759392986033e-002,
   -4.0000000000000003e-005,
   -6.2995310694929289e-004,
    3.0155423650436874e-001,
   -1.4732138184362128e+002,
   -8.0485758387155673e-002,
    1.2999613070201104e-004,
   -2.2998612650391686e-004,
    7.5006161138150423e-002,
   -6.6641753555245816e+002,
    4.1339791352394223e-001,
   -8.2994223832840526e-004,
    1.0999613070201107e-004,
   -8.2815649505433631e-002,
   -1.0938036196605900e+002,
    2.6336624535315561e-001,
   -6.1995310736632202e-004,
   -3.0000000000000001e-005,
    2.6768637836637570e-002,
   -7.6301551348229371e+000,
    4.3377801050621549e-002,
   -1.4999613070201104e-004,
    1.6599154697477100e-003,
   -5.4592264087203335e-001,
    5.2597397400886694e-003,
   -3.0000000000000001e-005,
    1.4000086442279340e-004,
   -3.9468807322558065e-002,
    5.4997731272676332e-004,
    1.0000000000000001e-005,
   -2.8699181720991682e-003,
    5.0000000000000002e-005,
   -2.0999559352845242e-004,
   -2.0000000000000002e-005,
   -8.9993864682865569e-005,
    2.0000000000000002e-005,
   -1.1899218836971420e-003,
    1.3998772936573111e-004,
   -1.5138965103137052e-002,
    3.0000000000000001e-005,
    9.9991435544251718e-004,
   -1.9309646760617930e-001,
    5.9994717289762057e-004,
    4.7295768064663599e-003,
   -2.5679458767733747e+000,
    1.2998772728058534e-004,
   -3.5496786682241857e-003,
   -7.4474951357555064e+000,
    5.1895323465805943e-003,
   -1.0000000000000001e-005,
   -3.9996158815458320e-004,
   -1.1616866651267410e+000,
    2.6997615591387108e-003,
    3.9996331282987846e-004,
   -6.5125428755333434e-002,
    3.4996858455827779e-004,
    2.0000000000000002e-005,
   -3.4097648760305628e-003,
    3.0000000000000001e-005,
   -1.4998859378852452e-004,
   -3.0000000000000001e-005,
   -3.4996491652575104e-004,
    2.0000000000000002e-005,
   -4.1595528113013806e-003,
    2.0000000000000002e-005,
    7.9993864682865570e-005,
   -5.1414423754705015e-002,
    3.0000000000000001e-005,
   -1.5998159404859670e-004,
   -1.0134887267215567e-001,
    6.9993864682865557e-005,
   -1.7958046806006885e-002,
    4.0000000000000003e-005,
   -9.4989829246006562e-004,
   -5.0000000000000002e-005,
   -8.9987729365731119e-005,
   -1.0798422089319747e-003,
   -1.2798159429881417e-003,
   -3.1995091746292448e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0998159401036031e-004,
    2.0000000000000002e-005,
    1.9997545865498950e-004,
    2.5397239085315128e-003,
    6.9993864654188274e-005,
   -1.0000000000000001e-005,
    4.9999999980881818e-005,
    2.4998245837579912e-004,
   -3.9999999980881819e-005,
    2.2898386651984031e-003,
   -3.1997158931876420e-004,
    4.3036851617944766e-002,
   -6.9999999971322711e-005,
   -7.9999999971322724e-005,
   -3.2997631635752091e-004,
   -3.6997718619213426e-004,
   -4.9999999980881818e-005,
    2.0000000000000002e-005,
    1.4000086437499792e-004,
   -2.0000000000000002e-005,
    6.8997891533048204e-004,
    2.0000000000000002e-005,
   -1.0999999996176364e-004,
   -6.9397621853979179e-003,
    1.2999999995220451e-004,
   -9.9994943937801984e-004,
   -1.3635526910265641e-001,
    1.2999999995220451e-004,
    2.0000000000000002e-005,
   -1.2809307020413470e-002,
   -5.5692002982416444e-001,
    1.5499176927330593e-003,
   -2.1198888732590626e-003,
   -1.2072586632235355e-001,
    5.9996471378934212e-004,
    1.1999999995220454e-004,
   -1.4829513936803736e-002,
    1.1999999995220454e-004,
    1.0000000000000001e-005,
   -1.4899516940545059e-003,
    2.0000000000000002e-005,
   -1.4000086437499792e-004,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
    1.0000000000000001e-005,
   -7.2000518543504765e-004,
    8.9999999961763637e-005,
   -9.5500643636702888e-003,
    2.0000000000000002e-005,
    6.8999184870505300e-004,
   -1.2812062711325589e-001,
    3.7999485203640013e-004,
    1.0000000000000001e-005,
    2.8899630701658338e-003,
   -1.7529658870917373e+000,
   -1.4199772385874402e-003,
    1.8999226133710850e-004,
   -2.3789614548438715e-002,
   -1.8584706631265689e+001,
    1.5979758931841438e-002,
   -1.0999999996176364e-004,
    1.2299574368462561e-003,
   -5.8330995747445824e-001,
   -1.2498811836918333e+002,
    2.5428572335145844e-001,
   -5.3998065289228517e-004,
    1.0999999996176364e-004,
   -4.0119365740716291e-002,
   -8.4530980809119285e+000,
    4.4499321910298706e-002,
   -1.4999613064465648e-004,
   -7.6999098633872805e-004,
   -5.8399222946181695e-001,
    5.3999244182923751e-003,
   -2.9999999990440909e-005,
   -5.9999999980881817e-005,
   -4.0740229068206964e-002,
    5.5999485404507313e-004,
   -2.8500181497324566e-003,
    4.9999999980881818e-005,
   -2.0000172876911412e-004,
   -1.0000000000000001e-005,
    1.1000086438455706e-004,
   -1.0000000000000001e-005,
    1.5499542893889138e-003,
   -1.1999999995220454e-004,
    2.2959262222460274e-002,
   -6.0000000814940131e-005,
   -7.0996644384778017e-004,
    3.4964829777384171e-001,
    6.2996557944410491e-004,
   -7.9999999971322724e-005,
    1.0776928885449957e+000,
   -1.2510874740473208e-012,
    1.7000604964155280e-004,
    1.7954467304506039e+001,
   -3.5868073866512876e-002,
    6.9999999971322711e-005,
   -1.5699450477570172e-003,
    1.2618161438829740e+000,
   -6.5896597480392645e-003,
    2.0000000000000002e-005,
    2.7998385871150466e-004,
    9.2046945572244729e-002,
   -8.3995643917530224e-004,
    2.0000000000000002e-005,
    6.7897812967059792e-003,
   -8.9999999961763637e-005,
    4.9998505154858841e-004,
    3.9999999990440911e-005,
    3.9999999990440911e-005,
    3.7997105128246991e-004,
    3.6097474108487937e-003,
   -2.9997834420817477e-003,
    1.0000000000000001e-005,
    2.9993864673306469e-005,
   -3.9677107212324189e-002,
    1.0999386464462920e-004,
    2.0000000417029158e-005,
   -8.4893888830128075e-003,
    4.9999999980881818e-005,
   -1.0199298326691382e-003,
    1.0000000000000001e-005,
   -1.0999386464462920e-004,
   -1.0000000000000001e-005,
    6.9993864654188274e-005,
   -1.0998772932749475e-004,
   -7.5377475448448789e-009,
   -5.5993864620132634e-004,
   -8.9987729337053836e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.6995091746292446e-004,
    4.9990796940892505e-004,
    4.9993864682865559e-005,
    2.0000000000000002e-005,
    3.1995178146868876e-004,
    8.1188227048091203e-003,
   -2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.3308050673809144e-002,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.3298062606179033e-003,
    1.1998159404859668e-004,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    4.8994737499714110e-004,
   -4.0000000000000003e-005,
    9.6489688340907539e-003,
   -2.0000000000000002e-005,
   -2.0997545873146224e-004,
    2.4694321033747729e-001,
   -5.1993704313278298e-004,
    1.6297949907171786e-003,
    3.3979269667574069e-001,
   -7.4990863173219800e-004,
   -6.4991563188915497e-004,
    3.2996433487961413e-002,
   -1.5998159404859670e-004,
   -4.9993864682865559e-005,
    2.9496843310041938e-003,
   -3.0000000000000001e-005,
    2.5997018742009203e-004,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
    8.7994383294838702e-004,
   -1.2998772936573112e-004,
    1.5438937565750930e-002,
   -1.1099043508273937e-003,
    2.9419933639500490e-001,
   -7.1993576752020353e-004,
   -4.3396143538889835e-003,
    7.3712009873208606e+000,
   -1.5408629174585074e-002,
    3.0000000000000001e-005,
   -8.0000000000000007e-005,
    6.3134273217907197e-002,
    8.0501722501471225e+000,
   -1.8158365570262437e-002,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -1.6438530931310029e-002,
    7.5167634592978361e-001,
   -3.7796659066298708e-003,
    1.0000000000000001e-005,
   -1.3298829400007950e-003,
    6.4385511148316116e-002,
   -5.4995104219560966e-004,
   -1.0999386468286557e-004,
    5.3796341272183213e-003,
   -6.9999999999999994e-005,
    4.3997191584864979e-004,
    4.0000000000000003e-005,
    1.0000086442279343e-004,
   -3.0000000000000001e-005,
    1.7799463418591865e-003,
   -3.5998258403813351e-004,
    2.9489099882669595e-002,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -3.6498203993748546e-003,
    4.7840471403548940e-001,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
   -2.8038574609160361e-002,
    8.6055315759708346e+000,
   -2.1568901502827798e-002,
    1.1999613070201107e-004,
   -1.4999613070201104e-004,
   -6.1026813470894330e-002,
    2.0543594980827186e+002,
   -4.2621769681760546e-001,
    9.2993396172481044e-004,
   -2.8597950970269438e-003,
    2.1460741878253438e+000,
    1.6472861720988391e+002,
   -3.8212969982592837e-001,
    8.8993923220212219e-004,
    1.7998612608688771e-004,
   -3.7451034323166538e-001,
    1.4530277628654355e+001,
   -7.3726191969978747e-002,
    2.4997999076975324e-004,
    1.0000000000000001e-005,
   -2.8868533323010842e-002,
    1.1770012469305144e+000,
   -1.0069497923021724e-002,
    5.0000000000000002e-005,
   -2.3198866479913973e-003,
    9.3197086776599547e-002,
   -1.1699437555815517e-003,
   -1.8999472910565898e-004,
    7.2997775780399357e-003,
   -1.2000000000000000e-004,
   -1.0000000000000001e-005,
    5.6998591532850544e-004,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    3.6000432211396710e-004,
   -1.1000086442279343e-004,
    4.9700466746605528e-003,
   -1.1099892969039314e-003,
    6.8500550949332994e-002,
   -6.9999999999999994e-005,
    5.0000000000000002e-005,
   -1.0769882553088511e-002,
    9.4827605247099767e-001,
   -5.3999184979720686e-004,
    2.3999312582681555e-004,
   -9.1538856043092448e-002,
    1.3194063173790582e+001,
    4.6699433443401547e-003,
   -3.0000000000000001e-005,
    3.6999011970053258e-004,
   -5.6064211237340766e-001,
    2.1165554821401352e+002,
   -5.3571237723129372e-001,
    2.8199119109441522e-003,
    3.0000000000000001e-005,
   -8.4997180352719501e-003,
    1.7902698262118480e-001,
    4.5864382748764037e+003,
   -9.3658636971804885e+000,
    1.9929344962433863e-002,
   -4.0000000000000003e-005,
    1.4999226140402212e-004,
   -7.5177394616022630e-002,
    5.5177059700241529e+001,
    2.3699139285772671e+003,
   -5.7416072726469185e+000,
    1.3719546083927990e-002,
   -3.0000000000000001e-005,
    1.0000000000000001e-005,
    5.5998151376255711e-004,
   -6.3831492299232613e+000,
    1.9195619489527533e+002,
   -9.9021537775935731e-001,
    3.4398886922370146e-003,
   -1.0000000000000001e-005,
    8.0000000000000007e-005,
   -4.5379355888279105e-001,
    1.4379697094884525e+001,
   -1.2385836590823918e-001,
    5.8998324594437726e-004,
    2.0000000000000002e-005,
   -3.3359573910778852e-002,
    1.0594959685741079e+000,
   -1.3319840849812047e-002,
    8.0000000000000007e-005,
   -2.4699729960813159e-003,
    7.7650569609901368e-002,
   -1.3099854267718840e-003,
    1.0000000000000001e-005,
   -1.8000172884558686e-004,
    5.6800466721583787e-003,
   -1.2000086442279343e-004,
   -1.0000000000000001e-005,
    4.1000432169693791e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    1.5000172842855768e-004,
   -3.0000000000000001e-005,
    1.7299524725890004e-003,
   -1.9999472827160066e-004,
    1.8189446900386926e-002,
   -3.0012272302385518e-005,
   -1.1599428886565832e-003,
    1.7623439547424485e-001,
   -2.3998558891332903e-004,
    2.0000000000000002e-005,
   -3.4598243384344031e-003,
    2.4888870370572809e+000,
   -9.9999858210086288e-006,
    1.6998612608688771e-004,
   -2.2798820324075567e-003,
   -2.8397100387825329e+001,
    6.7416468914439406e-002,
   -1.5999613070201107e-004,
    1.9798567275794987e-003,
   -1.4371624246554999e+000,
   -2.4358228023531897e+001,
    6.5406559242570161e-002,
   -1.6998612608688771e-004,
    3.0000000000000001e-005,
    8.4715549444192628e-002,
   -2.9145403053891537e+000,
    1.6049176443184703e-002,
   -6.0000000000000002e-005,
    8.8895485793052953e-003,
   -2.9022053823049210e-001,
    2.6198687771284064e-003,
   -1.0000000000000001e-005,
    8.4995816791930991e-004,
   -2.6759173365087850e-002,
    3.4998171919831095e-004,
    8.0000000000000007e-005,
   -2.3599296610001077e-003,
    4.0000000000000003e-005,
   -1.9999559352845242e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -4.8996664578836700e-004,
    4.0000000000000003e-005,
   -9.4393463774119675e-003,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
   -2.4988250505490764e-001,
    5.9994717289762057e-004,
    2.0000000000000002e-005,
   -6.5394170054182757e-003,
   -2.5211988988567757e+000,
    5.1495392536223573e-003,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
   -2.3487922182196537e-002,
   -1.8574703726238279e-001,
    9.5991962633685819e-004,
   -1.9998317976156355e-003,
   -1.3429085176727666e-002,
    1.1999386468286557e-004,
   -1.3998772936573111e-004,
   -6.5995610358265586e-004,
   -1.0000000000000001e-005,
   -8.9987729365731119e-005,
   -9.2989829037491993e-004,
   -3.0000000000000001e-005,
   -1.8797939874814595e-003,
    6.9987729365731107e-005,
    3.7995878037455833e-004,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
   -1.6998159398168305e-004,
   -5.2994478111099074e-004,
   -4.9999999980881818e-005,
    2.9996932329961866e-004,
    2.9999999990440909e-005,
   -1.3998772930837655e-004,
   -3.9897184654674967e-003,
   -7.9999999971322724e-005,
   -1.1819154776759652e-002,
    2.9999999990440909e-005,
   -1.0199070284880199e-003,
    5.1796152842171735e-003,
   -3.9999999980881819e-005,
    3.6997545858807586e-004,
   -1.2999999995220451e-004,
   -2.9699033914452450e-003,
   -8.3107215024644673e-002,
    1.9999386460639284e-004,
   -1.8899041488641178e-003,
   -2.3140200583971432e-001,
    7.4995944283764667e-004,
    2.0000000000000002e-005,
   -1.9478951652872143e-002,
    6.5067436994949851e-002,
    1.9999386377233451e-004,
   -8.3995471032971536e-004,
    3.2998552375344079e-003,
    4.9999999980881818e-005,
   -3.9999999980881819e-005,
   -1.1000086229941127e-004,
    1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -4.9999999980881818e-005,
    1.0000000000000001e-005,
   -1.7400112368462959e-003,
    1.0999999996176364e-004,
   -4.4420191594329188e-002,
    1.0999999996176364e-004,
   -2.7999613143091934e-004,
   -1.2241239354170774e+000,
    2.9799592037488124e-003,
   -1.0000000000000001e-005,
    5.9999999980881817e-005,
   -2.9789532034855189e-002,
   -3.2096875702680525e+000,
    1.2149816513247130e-002,
   -3.9999999990440911e-005,
    2.4999226130843115e-004,
   -2.5303567949529321e-001,
    4.0422724909322250e-001,
    3.7499475884785771e-003,
   -2.0000000000000002e-005,
    4.9999999980881818e-005,
   -1.0619833696571750e-002,
    3.6963285764371392e-004,
    8.2999098256634747e-004,
   -5.0999399005842710e-004,
   -3.8900502005239719e-003,
    1.3999999995220451e-004,
   -2.0000000000000002e-005,
   -6.2000345537660962e-004,
    2.0000000000000002e-005,
   -6.9999999971322711e-005,
   -5.9999999980881817e-005,
   -5.8997891536871844e-004,
    2.0000000000000002e-005,
   -6.2298023688168265e-003,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
   -4.1888591181851023e-002,
    6.9999999971322711e-005,
    1.1599394338714837e-003,
    2.7376050640024469e-001,
   -1.3799302936026807e-003,
    8.9095191827036693e-003,
    1.4989475229298219e-001,
   -9.5994857372325713e-004,
    1.9999386460639284e-004,
    2.2439244308103623e-002,
   -2.1999386459683372e-004,
   -1.0000000000000001e-005,
    2.5599174746015686e-003,
   -3.9999999990440911e-005,
    2.5998945811572696e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.9998859371205181e-004,
   -3.5997632051825340e-004,
   -1.0000000000000001e-005,
   -1.2699105659621721e-003,
   -8.9987729327494750e-005,
   -1.9098606137589759e-003,
    1.0000000000000001e-005,
   -3.8996932285391227e-004,
   -5.9999999980881817e-005,
    1.2270662946163423e-008,
   -6.4612388089665551e-009,
    3.9999999990440911e-005,
   -1.3998772931793563e-004,
   -1.9984914225727413e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    3.9993864682865560e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    3.2995178105165959e-004,
    8.5987288510061782e-004,
    5.3992023920913562e-004,
    8.9987729365731119e-005,
    1.0000000000000001e-005,
    2.6997018742009205e-004,
    7.5491959222631043e-003,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.7408120729120777e-002,
   -8.9987729365731119e-005,
    6.9993864682865557e-005,
    9.7089419363624394e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    1.5998255289069543e-003,
   -1.0000000000000001e-005,
    1.9997545831443309e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    5.5996157195988012e-003,
   -4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.5779894437490621e-001,
   -8.9991962591982899e-004,
    1.0799026232328945e-003,
    3.0871801000134386e-001,
   -1.5998597982434637e-003,
    9.0991530255477438e-004,
    1.5225906774649181e-001,
   -8.8991876108000648e-004,
   -5.1995631308995058e-004,
    2.4108301622573006e-002,
   -2.0998772894870199e-004,
   -8.9993864682865569e-005,
    2.8997964930560258e-003,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
    3.0997718674299077e-004,
    3.0000000000000001e-005,
   -5.0000000000000002e-005,
   -3.3999032221708226e-004,
    4.0000000000000003e-005,
   -1.9999472910565898e-004,
    6.9999999999999994e-005,
    9.9816818341372293e-002,
   -7.7995730266245831e-004,
    1.0000000000000001e-005,
    4.2997558429820561e-004,
    2.7318888991130463e+000,
   -1.5329221686054887e-002,
    6.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2888806358530470e-002,
    4.3740094149881275e+000,
   -2.2728827028039987e-002,
    8.0000000000000007e-005,
   -1.0000000000000001e-005,
    6.6596367501723175e-003,
    1.8707647192162216e+000,
   -1.1139425328502053e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -7.3196179071459144e-003,
    2.8250053804221126e-001,
   -2.5098723170920148e-003,
    1.0000000000000001e-005,
   -1.2099420263189354e-003,
    3.2738955542451958e-002,
   -4.0998171836425265e-004,
   -1.3999999999999999e-004,
    3.3598954377783608e-003,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
    3.1999032180005311e-004,
    3.0000000000000001e-005,
   -6.9999999999999994e-005,
    1.0000000000000001e-005,
   -6.2000604970846650e-004,
    1.0000000000000001e-004,
   -4.3100337049824184e-003,
    4.1999485383834408e-004,
   -7.0000432336505452e-004,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2999699429074617e-004,
    1.1868256646571000e+000,
   -9.3398815629530830e-003,
    6.9999999999999994e-005,
    1.2999613070201104e-004,
    6.6199085327521001e-003,
    3.0772574469538199e+001,
   -1.6976764181870591e-001,
    6.0998238152158387e-004,
    1.6999613028498192e-004,
    3.0085540903233987e-001,
    3.8429834065082709e+001,
   -2.0133705199593430e-001,
    7.1997851180656579e-004,
   -3.0000000000000001e-005,
   -2.3909623818604519e-002,
    1.3899059511228872e+001,
   -8.4998752997522520e-002,
    3.4998839127197491e-004,
   -6.0998151793284874e-004,
   -6.6009005303252169e-002,
    1.9777269709780243e+000,
   -1.7809756481404950e-002,
    9.0000000000000006e-005,
    1.0000000000000001e-005,
   -9.7998651776269961e-003,
    2.1853111872294201e-001,
   -2.7399656640878030e-003,
    2.0000000000000002e-005,
   -1.0899867028014926e-003,
    2.1490146986460099e-002,
   -3.5999398941555070e-004,
   -1.1000000000000000e-004,
    1.9800138261773740e-003,
   -4.0000000000000003e-005,
    1.7000172842855771e-004,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
    6.3997977834325448e-004,
   -4.0000000000000003e-005,
    2.6819162891074156e-002,
   -2.9999999582970844e-005,
   -3.8998085435848836e-004,
   -3.5757788086640147e-001,
    2.1998884645307976e-003,
   -1.8749021919187647e-002,
   -6.3520814175288576e-001,
    3.5298208949222213e-003,
   -1.0000000000000001e-005,
    5.0996785738860332e-004,
   -2.8910002243214861e-001,
    1.8599058754883145e-003,
    4.0000000000000003e-005,
    1.8299050135676957e-003,
   -5.0698321291315837e-002,
    4.7997558304711818e-004,
    3.1998472407350649e-004,
   -6.6797899699247648e-003,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
   -7.6997364344314910e-004,
    1.0000000000000001e-005,
   -8.0000000000000007e-005,
   -5.0000000000000002e-005,
   -3.6897473244932536e-003,
    2.0000000000000002e-005,
    8.0000000000000007e-005,
   -3.5137536195548776e-002,
    1.8998772894870199e-004,
    8.9991876066297736e-004,
   -2.1708465634576395e-002,
    1.1999386468286557e-004,
    1.0000000000000001e-005,
   -2.0798569834726932e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -7.9994729105658979e-005,
    1.0000000000000001e-005,
    1.9997545873146224e-004,
    2.2997545831443308e-004,
    5.0000000000000002e-005,
   -2.4540377124886545e-008,
   -2.9999999990440909e-005,
    1.0000000000000001e-005,
   -1.2998772931793563e-004,
   -5.1996491550051073e-004,
   -2.9999999990440909e-005,
   -1.0999386381057087e-004,
   -2.0000000000000002e-005,
    1.1999386463507010e-004,
    2.0000000000000002e-005,
   -6.9999999980881810e-005,
   -1.8899411517054080e-003,
    1.0000000000000001e-005,
   -2.9999999990440909e-005,
   -6.9297709077279961e-003,
    3.9999999980881819e-005,
   -3.9997998978274943e-004,
   -1.5399555574949681e-003,
    2.9999999990440909e-005,
   -2.0999386460639284e-004,
    1.0199614584522137e-003,
   -2.9999999990440909e-005,
    1.3999999995220451e-004,
   -5.9000345622022704e-004,
   -1.6860079490600130e-002,
    9.9999999961763637e-005,
   -3.0999699501009536e-004,
   -5.7250246051946750e-002,
    3.8999398927216423e-004,
   -3.3499471412369239e-003,
   -1.3500126888151384e-002,
    2.3999612978192091e-004,
   -2.0000000000000002e-005,
   -1.3699767812032594e-003,
    4.3899544269948888e-003,
    5.9999999980881817e-005,
   -1.8999999993308636e-004,
    4.9999827221431877e-004,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -5.9999999980881817e-005,
   -4.5998332063697402e-004,
    1.0000000000000001e-005,
    3.9198718475904771e-003,
   -3.9999999990440911e-005,
    1.5999999994264545e-004,
    4.7498424417146134e-003,
   -4.9999999980881818e-005,
    4.9999999980881818e-005,
    1.8099368283690308e-003,
   -2.0000000000000002e-005,
    3.4998859282063896e-004,
    4.9999999980881818e-005,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -7.9993864654188287e-005,
   -3.9999999980881819e-005,
   -1.0000000000000001e-005,
   -1.9118187377780694e-014,
    1.0000000000000001e-005,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.9998245805436096e-004,
    5.8993423868899139e-004,
    5.8993423827196224e-004,
    2.3997545789740391e-004,
    5.0000000000000002e-005,
    1.3998772936573111e-004,
    3.1197876805299594e-003,
   -3.0000000000000001e-005,
    3.0000000000000001e-005,
    8.3594149696340162e-003,
   -6.9999999999999994e-005,
    3.0000000000000001e-005,
    7.4394759535129076e-003,
   -6.9999999999999994e-005,
   -2.0000000000000002e-005,
    2.8497974393342251e-003,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    5.5995877954049991e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.7599437444205150e-003,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    3.8998763101098795e-002,
   -3.8998171878128180e-004,
    4.3997558388117648e-004,
    9.0327051829633215e-002,
   -7.9995643740560663e-004,
    1.7999386384880725e-004,
    7.1417615893261940e-002,
   -6.3996557759793654e-004,
   -2.4999085980767006e-004,
    2.5099162146062979e-002,
   -2.5998385923368386e-004,
   -1.6000000000000001e-004,
    4.7698476225132812e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    6.8997891392046096e-004,
   -1.0000000000000001e-005,
    9.0000000000000006e-005,
    2.0000172842855771e-004,
    1.3760086405580775e-002,
   -1.7000086442279342e-004,
    1.8000000000000001e-004,
    2.9255150843605804e-001,
   -2.8699635232183724e-003,
    2.0000000000000002e-005,
    3.6399493233071124e-003,
    5.7156243369584225e-001,
   -5.0899338785667068e-003,
    3.0000000000000001e-005,
    6.9997403395386472e-005,
    3.9451145761016659e-001,
   -3.5899493174687039e-003,
    2.0000000000000002e-005,
   -8.0000000000000007e-005,
   -2.1399673550604606e-003,
    1.2616048172676975e-001,
   -1.3199811017387127e-003,
   -1.0000000000000001e-005,
   -9.3998711399127875e-004,
    2.2850114068409776e-002,
   -3.0999699429074619e-004,
   -1.7999999958297086e-004,
    3.1800155483504944e-003,
   -5.0000000000000002e-005,
   -3.0000000000000001e-005,
    3.8000259201729283e-004,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
    2.7998945612617212e-004,
   -4.7798445528808757e-003,
    5.0000000000000002e-005,
   -2.5998385965071301e-004,
   -1.2819580249561514e-002,
    1.2000000000000000e-004,
   -6.9999999582970830e-005,
   -1.0639642712304840e-002,
    1.0000000000000001e-004,
    8.0000000000000007e-005,
   -3.9598669818323795e-003,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
   -8.2997191376350387e-004,
    1.0000000000000001e-005,
   -1.2999999958297084e-004,
   -2.0000000000000002e-005,
   -6.0000000000000002e-005,
   -5.0996577969745702e-004,
    1.0000000000000001e-005,
   -6.0995964354626422e-004,
    1.0000000000000001e-005,
   -2.0998159321453837e-004,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.9999999990440911e-005,
   -1.5999386420848186e-004,
   -1.1999999953517537e-004,
    2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.5000172791593756e-004,
   -9.8000518325431097e-004,
    1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -6.8000259010063977e-004,
    1.0000000000000001e-005,
   -3.9999999980881819e-005,
   -4.9999999146823498e-005,
   -1.0000000000000001e-005,
    4.9999999980881818e-005,
    5.9999999980881817e-005,
    1.0999999996176364e-004,
    6.9999999971322711e-005,
    2.9999999990440909e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
    1.8998859295446622e-004,
    2.4998245722030263e-004,
    1.4998772853167279e-004,
    5.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    5.7998505007165373e-004,
    1.6999472848011523e-003,
   -2.0000000000000002e-005,
    1.9599341480759733e-003,
   -3.0000000000000001e-005,
    1.1199605054668827e-003,
   -1.0000000000000001e-005,
    3.6998945654320129e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    1.8000172842855771e-004,
    3.2400207357213130e-003,
   -5.0000000000000002e-005,
    5.0000000000000002e-005,
    8.4000461959887664e-003,
   -1.1000000000000000e-004,
    1.0000000000000001e-005,
    8.6600422808942937e-003,
   -1.0999999958297084e-004,
   -5.0000000000000002e-005,
    4.5500194093078998e-003,
   -6.0000000000000002e-005,
   -4.0000000000000003e-005,
    1.3900051802813232e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9000172759449937e-004,
    5.0000000000000002e-005,
   -6.9999999999999994e-005,
   -2.3999472868862983e-004,
   -2.8998859295446626e-004,
   -1.6999386384880725e-004,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.1999999958297084e-004,
    1.7000086358873509e-004,
    1.2999999916594166e-004,
    6.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<double, 1023> g_MAIN_S1_DE406
{ {
   -4.1159574817515249e+002,
    4.2032548138143505e-001,
   -5.8996904561608857e-004,
    1.5999226140402215e-004,
   -8.0187332604691075e-002,
    3.9533301970600597e+001,
    2.2639585659999782e+004,
   -4.5099610307254984e+001,
    9.0916999939499477e-002,
   -1.8999226140402215e-004,
    6.7997937456124253e-004,
   -1.3725701099805798e+000,
    7.6902570369870193e+002,
   -3.9976134296908525e+000,
    1.2969592654144397e-002,
   -4.0000000000000003e-005,
    9.0000000000000006e-005,
   -5.6839245216214383e-002,
    3.6123813073253892e+001,
   -3.2865586724190660e-001,
    1.5299527452105023e-003,
   -3.5399566985949336e-003,
    1.9336823984636129e+000,
   -2.6019715403432151e-002,
    1.5999699512480450e-004,
   -2.2999785954759791e-004,
    1.1100088855380498e-001,
   -2.0099807359283814e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
    6.6500631036979776e-003,
   -1.5000172884558683e-004,
    4.1000432211396706e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.3999645795124582e-004,
    6.0000000000000002e-005,
   -3.3199032234219098e-003,
    6.3996817253443345e-004,
   -4.7018577037686374e-002,
    6.9999999999999994e-005,
   -3.0000000000000001e-005,
    6.2296940374589041e-003,
   -6.6937870149957934e-001,
    1.0599428919928163e-003,
   -1.6998612608688771e-004,
    5.1647381444881534e-002,
   -9.6790565471940102e+000,
    2.4288759392986033e-002,
   -4.0000000000000003e-005,
   -6.2995310694929289e-004,
    3.0155423650436874e-001,
   -1.4732138184362128e+002,
   -8.0485758387155673e-002,
    1.2999613070201104e-004,
   -2.2998612650391686e-004,
    7.5006161138150423e-002,
   -6.6641753555245816e+002,
    4.1339791352394223e-001,
   -8.2994223832840526e-004,
    1.0999613070201107e-004,
   -8.2815649505433631e-002,
   -1.0938036196605900e+002,
    2.6336624535315561e-001,
   -6.1995310736632202e-004,
   -3.0000000000000001e-005,
    2.6768637836637570e-002,
   -7.6301551348229371e+000,
    4.3377801050621549e-002,
   -1.4999613070201104e-004,
    1.6599154697477100e-003,
   -5.4592264087203335e-001,
    5.2597397400886694e-003,
   -3.0000000000000001e-005,
    1.4000086442279340e-004,
   -3.9468807322558065e-002,
    5.4997731272676332e-004,
    1.0000000000000001e-005,
   -2.8699181720991682e-003,
    5.0000000000000002e-005,
   -2.0999559352845242e-004,
   -2.0000000000000002e-005,
   -8.9993864682865569e-005,
    2.0000000000000002e-005,
   -1.1899218836971420e-003,
    1.3998772936573111e-004,
   -1.5138965103137052e-002,
    3.0000000000000001e-005,
    9.9991435544251718e-004,
   -1.9309646760617930e-001,
    5.9994717289762057e-004,
    4.7295768064663599e-003,
   -2.5679458767733747e+000,
    1.2998772728058534e-004,
   -3.5496786682241857e-003,
   -7.4474951357555064e+000,
    5.1895323465805943e-003,
   -1.0000000000000001e-005,
   -3.9996158815458320e-004,
   -1.1616866651267410e+000,
    2.6997615591387108e-003,
    3.9996331282987846e-004,
   -6.5125428755333434e-002,
    3.4996858455827779e-004,
    2.0000000000000002e-005,
   -3.4097648760305628e-003,
    3.0000000000000001e-005,
   -1.4998859378852452e-004,
   -3.0000000000000001e-005,
   -3.4996491652575104e-004,
    2.0000000000000002e-005,
   -4.1595528113013806e-003,
    2.0000000000000002e-005,
    7.9993864682865570e-005,
   -5.1414423754705015e-002,
    3.0000000000000001e-005,
   -1.5998159404859670e-004,
   -1.0134887267215567e-001,
    6.9993864682865557e-005,
   -1.7958046806006885e-002,
    4.0000000000000003e-005,
   -9.4989829246006562e-004,
   -5.0000000000000002e-005,
   -8.9987729365731119e-005,
   -1.0798422089319747e-003,
   -1.2798159429881417e-003,
   -3.1995091746292448e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0998159401036031e-004,
    2.0000000000000002e-005,
    1.9997545865498950e-004,
    2.5397239085315128e-003,
    6.9993864654188274e-005,
   -1.0000000000000001e-005,
    4.9999999980881818e-005,
    2.4998245837579912e-004,
   -3.9999999980881819e-005,
    2.2898386651984031e-003,
   -3.1997158931876420e-004,
    4.3036851617944766e-002,
   -6.9999999971322711e-005,
   -7.9999999971322724e-005,
   -3.2997631635752091e-004,
   -3.6997718619213426e-004,
   -4.9999999980881818e-005,
    2.0000000000000002e-005,
    1.4000086437499792e-004,
   -2.0000000000000002e-005,
    6.8997891533048204e-004,
    2.0000000000000002e-005,
   -1.0999999996176364e-004,
   -6.9397621853979179e-003,
    1.2999999995220451e-004,
   -9.9994943937801984e-004,
   -1.3635526910265641e-001,
    1.2999999995220451e-004,
    2.0000000000000002e-005,
   -1.2809307020413470e-002,
   -5.5692002982416444e-001,
    1.5499176927330593e-003,
   -2.1198888732590626e-003,
   -1.2072586632235355e-001,
    5.9996471378934212e-004,
    1.1999999995220454e-004,
   -1.4829513936803736e-002,
    1.1999999995220454e-004,
    1.0000000000000001e-005,
   -1.4899516940545059e-003,
    2.0000000000000002e-005,
   -1.4000086437499792e-004,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
    1.0000000000000001e-005,
   -7.2000518543504765e-004,
    8.9999999961763637e-005,
   -9.5500643636702888e-003,
    2.0000000000000002e-005,
    6.8999184870505300e-004,
   -1.2812062711325589e-001,
    3.7999485203640013e-004,
    1.0000000000000001e-005,
    2.8899630701658338e-003,
   -1.7529658870917373e+000,
   -1.4199772385874402e-003,
    1.8999226133710850e-004,
   -2.3789614548438715e-002,
   -1.8584706631265689e+001,
    1.5979758931841438e-002,
   -1.0999999996176364e-004,
    1.2299574368462561e-003,
   -5.8330995747445824e-001,
   -1.2498811836918333e+002,
    2.5428572335145844e-001,
   -5.3998065289228517e-004,
    1.0999999996176364e-004,
   -4.0119365740716291e-002,
   -8.4530980809119285e+000,
    4.4499321910298706e-002,
   -1.4999613064465648e-004,
   -7.6999098633872805e-004,
   -5.8399222946181695e-001,
    5.3999244182923751e-003,
   -2.9999999990440909e-005,
   -5.9999999980881817e-005,
   -4.0740229068206964e-002,
    5.5999485404507313e-004,
   -2.8500181497324566e-003,
    4.9999999980881818e-005,
   -2.0000172876911412e-004,
   -1.0000000000000001e-005,
    1.1000086438455706e-004,
   -1.0000000000000001e-005,
    1.5499542893889138e-003,
   -1.1999999995220454e-004,
    2.2959262222460274e-002,
   -6.0000000814940131e-005,
   -7.0996644384778017e-004,
    3.4964829777384171e-001,
    6.2996557944410491e-004,
   -7.9999999971322724e-005,
    1.0776928885449957e+000,
   -1.2510874740473208e-012,
    1.7000604964155280e-004,
    1.7954467304506039e+001,
   -3.5868073866512876e-002,
    6.9999999971322711e-005,
   -1.5699450477570172e-003,
    1.2618161438829740e+000,
   -6.5896597480392645e-003,
    2.0000000000000002e-005,
    2.7998385871150466e-004,
    9.2046945572244729e-002,
   -8.3995643917530224e-004,
    2.0000000000000002e-005,
    6.7897812967059792e-003,
   -8.9999999961763637e-005,
    4.9998505154858841e-004,
    3.9999999990440911e-005,
    3.9999999990440911e-005,
    3.7997105128246991e-004,
    3.6097474108487937e-003,
   -2.9997834420817477e-003,
    1.0000000000000001e-005,
    2.9993864673306469e-005,
   -3.9677107212324189e-002,
    1.0999386464462920e-004,
    2.0000000417029158e-005,
   -8.4893888830128075e-003,
    4.9999999980881818e-005,
   -1.0199298326691382e-003,
    1.0000000000000001e-005,
   -1.0999386464462920e-004,
   -1.0000000000000001e-005,
    6.9993864654188274e-005,
   -1.0998772932749475e-004,
   -7.5377475448448789e-009,
   -5.5993864620132634e-004,
   -8.9987729337053836e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.6995091746292446e-004,
    4.9990796940892505e-004,
    4.9993864682865559e-005,
    2.0000000000000002e-005,
    3.1995178146868876e-004,
    8.1188227048091203e-003,
   -2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.3308050673809144e-002,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.3298062606179033e-003,
    1.1998159404859668e-004,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    4.8994737499714110e-004,
   -4.0000000000000003e-005,
    9.6489688340907539e-003,
   -2.0000000000000002e-005,
   -2.0997545873146224e-004,
    2.4694321033747729e-001,
   -5.1993704313278298e-004,
    1.6297949907171786e-003,
    3.3979269667574069e-001,
   -7.4990863173219800e-004,
   -6.4991563188915497e-004,
    3.2996433487961413e-002,
   -1.5998159404859670e-004,
   -4.9993864682865559e-005,
    2.9496843310041938e-003,
   -3.0000000000000001e-005,
    2.5997018742009203e-004,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
    8.7994383294838702e-004,
   -1.2998772936573112e-004,
    1.5438937565750930e-002,
   -1.1099043508273937e-003,
    2.9419933639500490e-001,
   -7.1993576752020353e-004,
   -4.3396143538889835e-003,
    7.3712009873208606e+000,
   -1.5408629174585074e-002,
    3.0000000000000001e-005,
   -8.0000000000000007e-005,
    6.3134273217907197e-002,
    8.0501722501471225e+000,
   -1.8158365570262437e-002,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -1.6438530931310029e-002,
    7.5167634592978361e-001,
   -3.7796659066298708e-003,
    1.0000000000000001e-005,
   -1.3298829400007950e-003,
    6.4385511148316116e-002,
   -5.4995104219560966e-004,
   -1.0999386468286557e-004,
    5.3796341272183213e-003,
   -6.9999999999999994e-005,
    4.3997191584864979e-004,
    4.0000000000000003e-005,
    1.0000086442279343e-004,
   -3.0000000000000001e-005,
    1.7799463418591865e-003,
   -3.5998258403813351e-004,
    2.9489099882669595e-002,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -3.6498203993748546e-003,
    4.7840471403548940e-001,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
   -2.8038574609160361e-002,
    8.6055315759708346e+000,
   -2.1568901502827798e-002,
    1.1999613070201107e-004,
   -1.4999613070201104e-004,
   -6.1026813470894330e-002,
    2.0543594980827186e+002,
   -4.2621769681760546e-001,
    9.2993396172481044e-004,
   -2.8597950970269438e-003,
    2.1460741878253438e+000,
    1.6472861720988391e+002,
   -3.8212969982592837e-001,
    8.8993923220212219e-004,
    1.7998612608688771e-004,
   -3.7451034323166538e-001,
    1.4530277628654355e+001,
   -7.3726191969978747e-002,
    2.4997999076975324e-004,
    1.0000000000000001e-005,
   -2.8868533323010842e-002,
    1.1770012469305144e+000,
   -1.0069497923021724e-002,
    5.0000000000000002e-005,
   -2.3198866479913973e-003,
    9.3197086776599547e-002,
   -1.1699437555815517e-003,
   -1.8999472910565898e-004,
    7.2997775780399357e-003,
   -1.2000000000000000e-004,
   -1.0000000000000001e-005,
    5.6998591532850544e-004,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    3.6000432211396710e-004,
   -1.1000086442279343e-004,
    4.9700466746605528e-003,
   -1.1099892969039314e-003,
    6.8500550949332994e-002,
   -6.9999999999999994e-005,
    5.0000000000000002e-005,
   -1.0769882553088511e-002,
    9.4827605247099767e-001,
   -5.3999184979720686e-004,
    2.3999312582681555e-004,
   -9.1538856043092448e-002,
    1.3194063173790582e+001,
    4.6699433443401547e-003,
   -3.0000000000000001e-005,
    3.6999011970053258e-004,
   -5.6064211237340766e-001,
    2.1165554821401352e+002,
   -5.3571237723129372e-001,
    2.8199119109441522e-003,
    3.0000000000000001e-005,
   -8.4997180352719501e-003,
    1.7902698262118480e-001,
    4.5864382748764037e+003,
   -9.3658636971804885e+000,
    1.9929344962433863e-002,
   -4.0000000000000003e-005,
    1.4999226140402212e-004,
   -7.5177394616022630e-002,
    5.5177059700241529e+001,
    2.3699139285772671e+003,
   -5.7416072726469185e+000,
    1.3719546083927990e-002,
   -3.0000000000000001e-005,
    1.0000000000000001e-005,
    5.5998151376255711e-004,
   -6.3831492299232613e+000,
    1.9195619489527533e+002,
   -9.9021537775935731e-001,
    3.4398886922370146e-003,
   -1.0000000000000001e-005,
    8.0000000000000007e-005,
   -4.5379355888279105e-001,
    1.4379697094884525e+001,
   -1.2385836590823918e-001,
    5.8998324594437726e-004,
    2.0000000000000002e-005,
   -3.3359573910778852e-002,
    1.0594959685741079e+000,
   -1.3319840849812047e-002,
    8.0000000000000007e-005,
   -2.4699729960813159e-003,
    7.7650569609901368e-002,
   -1.3099854267718840e-003,
    1.0000000000000001e-005,
   -1.8000172884558686e-004,
    5.6800466721583787e-003,
   -1.2000086442279343e-004,
   -1.0000000000000001e-005,
    4.1000432169693791e-004,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    1.5000172842855768e-004,
   -3.0000000000000001e-005,
    1.7299524725890004e-003,
   -1.9999472827160066e-004,
    1.8189446900386926e-002,
   -3.0012272302385518e-005,
   -1.1599428886565832e-003,
    1.7623439547424485e-001,
   -2.3998558891332903e-004,
    2.0000000000000002e-005,
   -3.4598243384344031e-003,
    2.4888870370572809e+000,
   -9.9999858210086288e-006,
    1.6998612608688771e-004,
   -2.2798820324075567e-003,
   -2.8397100387825329e+001,
    6.7416468914439406e-002,
   -1.5999613070201107e-004,
    1.9798567275794987e-003,
   -1.4371624246554999e+000,
   -2.4358228023531897e+001,
    6.5406559242570161e-002,
   -1.6998612608688771e-004,
    3.0000000000000001e-005,
    8.4715549444192628e-002,
   -2.9145403053891537e+000,
    1.6049176443184703e-002,
   -6.0000000000000002e-005,
    8.8895485793052953e-003,
   -2.9022053823049210e-001,
    2.6198687771284064e-003,
   -1.0000000000000001e-005,
    8.4995816791930991e-004,
   -2.6759173365087850e-002,
    3.4998171919831095e-004,
    8.0000000000000007e-005,
   -2.3599296610001077e-003,
    4.0000000000000003e-005,
   -1.9999559352845242e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -4.8996664578836700e-004,
    4.0000000000000003e-005,
   -9.4393463774119675e-003,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
   -2.4988250505490764e-001,
    5.9994717289762057e-004,
    2.0000000000000002e-005,
   -6.5394170054182757e-003,
   -2.5211988988567757e+000,
    5.1495392536223573e-003,
   -1.0000000000000001e-005,
    4.0000000000000003e-005,
   -2.3487922182196537e-002,
   -1.8574703726238279e-001,
    9.5991962633685819e-004,
   -1.9998317976156355e-003,
   -1.3429085176727666e-002,
    1.1999386468286557e-004,
   -1.3998772936573111e-004,
   -6.5995610358265586e-004,
   -1.0000000000000001e-005,
   -8.9987729365731119e-005,
   -9.2989829037491993e-004,
   -3.0000000000000001e-005,
   -1.8797939874814595e-003,
    6.9987729365731107e-005,
    3.7995878037455833e-004,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
   -1.6998159398168305e-004,
   -5.2994478111099074e-004,
   -4.9999999980881818e-005,
    2.9996932329961866e-004,
    2.9999999990440909e-005,
   -1.3998772930837655e-004,
   -3.9897184654674967e-003,
   -7.9999999971322724e-005,
   -1.1819154776759652e-002,
    2.9999999990440909e-005,
   -1.0199070284880199e-003,
    5.1796152842171735e-003,
   -3.9999999980881819e-005,
    3.6997545858807586e-004,
   -1.2999999995220451e-004,
   -2.9699033914452450e-003,
   -8.3107215024644673e-002,
    1.9999386460639284e-004,
   -1.8899041488641178e-003,
   -2.3140200583971432e-001,
    7.4995944283764667e-004,
    2.0000000000000002e-005,
   -1.9478951652872143e-002,
    6.5067436994949851e-002,
    1.9999386377233451e-004,
   -8.3995471032971536e-004,
    3.2998552375344079e-003,
    4.9999999980881818e-005,
   -3.9999999980881819e-005,
   -1.1000086229941127e-004,
    1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -4.9999999980881818e-005,
    1.0000000000000001e-005,
   -1.7400112368462959e-003,
    1.0999999996176364e-004,
   -4.4420191594329188e-002,
    1.0999999996176364e-004,
   -2.7999613143091934e-004,
   -1.2241239354170774e+000,
    2.9799592037488124e-003,
   -1.0000000000000001e-005,
    5.9999999980881817e-005,
   -2.9789532034855189e-002,
   -3.2096875702680525e+000,
    1.2149816513247130e-002,
   -3.9999999990440911e-005,
    2.4999226130843115e-004,
   -2.5303567949529321e-001,
    4.0422724909322250e-001,
    3.7499475884785771e-003,
   -2.0000000000000002e-005,
    4.9999999980881818e-005,
   -1.0619833696571750e-002,
    3.6963285764371392e-004,
    8.2999098256634747e-004,
   -5.0999399005842710e-004,
   -3.8900502005239719e-003,
    1.3999999995220451e-004,
   -2.0000000000000002e-005,
   -6.2000345537660962e-004,
    2.0000000000000002e-005,
   -6.9999999971322711e-005,
   -5.9999999980881817e-005,
   -5.8997891536871844e-004,
    2.0000000000000002e-005,
   -6.2298023688168265e-003,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
   -4.1888591181851023e-002,
    6.9999999971322711e-005,
    1.1599394338714837e-003,
    2.7376050640024469e-001,
   -1.3799302936026807e-003,
    8.9095191827036693e-003,
    1.4989475229298219e-001,
   -9.5994857372325713e-004,
    1.9999386460639284e-004,
    2.2439244308103623e-002,
   -2.1999386459683372e-004,
   -1.0000000000000001e-005,
    2.5599174746015686e-003,
   -3.9999999990440911e-005,
    2.5998945811572696e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.9998859371205181e-004,
   -3.5997632051825340e-004,
   -1.0000000000000001e-005,
   -1.2699105659621721e-003,
   -8.9987729327494750e-005,
   -1.9098606137589759e-003,
    1.0000000000000001e-005,
   -3.8996932285391227e-004,
   -5.9999999980881817e-005,
    1.2270662946163423e-008,
   -6.4612388089665551e-009,
    3.9999999990440911e-005,
   -1.3998772931793563e-004,
   -1.9984914225727413e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    3.9993864682865560e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    3.2995178105165959e-004,
    8.5987288510061782e-004,
    5.3992023920913562e-004,
    8.9987729365731119e-005,
    1.0000000000000001e-005,
    2.6997018742009205e-004,
    7.5491959222631043e-003,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.7408120729120777e-002,
   -8.9987729365731119e-005,
    6.9993864682865557e-005,
    9.7089419363624394e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    1.5998255289069543e-003,
   -1.0000000000000001e-005,
    1.9997545831443309e-004,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    5.5996157195988012e-003,
   -4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.5779894437490621e-001,
   -8.9991962591982899e-004,
    1.0799026232328945e-003,
    3.0871801000134386e-001,
   -1.5998597982434637e-003,
    9.0991530255477438e-004,
    1.5225906774649181e-001,
   -8.8991876108000648e-004,
   -5.1995631308995058e-004,
    2.4108301622573006e-002,
   -2.0998772894870199e-004,
   -8.9993864682865569e-005,
    2.8997964930560258e-003,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
    3.0997718674299077e-004,
    3.0000000000000001e-005,
   -5.0000000000000002e-005,
   -3.3999032221708226e-004,
    4.0000000000000003e-005,
   -1.9999472910565898e-004,
    6.9999999999999994e-005,
    9.9816818341372293e-002,
   -7.7995730266245831e-004,
    1.0000000000000001e-005,
    4.2997558429820561e-004,
    2.7318888991130463e+000,
   -1.5329221686054887e-002,
    6.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2888806358530470e-002,
    4.3740094149881275e+000,
   -2.2728827028039987e-002,
    8.0000000000000007e-005,
   -1.0000000000000001e-005,
    6.6596367501723175e-003,
    1.8707647192162216e+000,
   -1.1139425328502053e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -7.3196179071459144e-003,
    2.8250053804221126e-001,
   -2.5098723170920148e-003,
    1.0000000000000001e-005,
   -1.2099420263189354e-003,
    3.2738955542451958e-002,
   -4.0998171836425265e-004,
   -1.3999999999999999e-004,
    3.3598954377783608e-003,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
    3.1999032180005311e-004,
    3.0000000000000001e-005,
   -6.9999999999999994e-005,
    1.0000000000000001e-005,
   -6.2000604970846650e-004,
    1.0000000000000001e-004,
   -4.3100337049824184e-003,
    4.1999485383834408e-004,
   -7.0000432336505452e-004,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.2999699429074617e-004,
    1.1868256646571000e+000,
   -9.3398815629530830e-003,
    6.9999999999999994e-005,
    1.2999613070201104e-004,
    6.6199085327521001e-003,
    3.0772574469538199e+001,
   -1.6976764181870591e-001,
    6.0998238152158387e-004,
    1.6999613028498192e-004,
    3.0085540903233987e-001,
    3.8429834065082709e+001,
   -2.0133705199593430e-001,
    7.1997851180656579e-004,
   -3.0000000000000001e-005,
   -2.3909623818604519e-002,
    1.3899059511228872e+001,
   -8.4998752997522520e-002,
    3.4998839127197491e-004,
   -6.0998151793284874e-004,
   -6.6009005303252169e-002,
    1.9777269709780243e+000,
   -1.7809756481404950e-002,
    9.0000000000000006e-005,
    1.0000000000000001e-005,
   -9.7998651776269961e-003,
    2.1853111872294201e-001,
   -2.7399656640878030e-003,
    2.0000000000000002e-005,
   -1.0899867028014926e-003,
    2.1490146986460099e-002,
   -3.5999398941555070e-004,
   -1.1000000000000000e-004,
    1.9800138261773740e-003,
   -4.0000000000000003e-005,
    1.7000172842855771e-004,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
    6.3997977834325448e-004,
   -4.0000000000000003e-005,
    2.6819162891074156e-002,
   -2.9999999582970844e-005,
   -3.8998085435848836e-004,
   -3.5757788086640147e-001,
    2.1998884645307976e-003,
   -1.8749021919187647e-002,
   -6.3520814175288576e-001,
    3.5298208949222213e-003,
   -1.0000000000000001e-005,
    5.0996785738860332e-004,
   -2.8910002243214861e-001,
    1.8599058754883145e-003,
    4.0000000000000003e-005,
    1.8299050135676957e-003,
   -5.0698321291315837e-002,
    4.7997558304711818e-004,
    3.1998472407350649e-004,
   -6.6797899699247648e-003,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
   -7.6997364344314910e-004,
    1.0000000000000001e-005,
   -8.0000000000000007e-005,
   -5.0000000000000002e-005,
   -3.6897473244932536e-003,
    2.0000000000000002e-005,
    8.0000000000000007e-005,
   -3.5137536195548776e-002,
    1.8998772894870199e-004,
    8.9991876066297736e-004,
   -2.1708465634576395e-002,
    1.1999386468286557e-004,
    1.0000000000000001e-005,
   -2.0798569834726932e-003,
    2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -7.9994729105658979e-005,
    1.0000000000000001e-005,
    1.9997545873146224e-004,
    2.2997545831443308e-004,
    5.0000000000000002e-005,
   -2.4540377124886545e-008,
   -2.9999999990440909e-005,
    1.0000000000000001e-005,
   -1.2998772931793563e-004,
   -5.1996491550051073e-004,
   -2.9999999990440909e-005,
   -1.0999386381057087e-004,
   -2.0000000000000002e-005,
    1.1999386463507010e-004,
    2.0000000000000002e-005,
   -6.9999999980881810e-005,
   -1.8899411517054080e-003,
    1.0000000000000001e-005,
   -2.9999999990440909e-005,
   -6.9297709077279961e-003,
    3.9999999980881819e-005,
   -3.9997998978274943e-004,
   -1.5399555574949681e-003,
    2.9999999990440909e-005,
   -2.0999386460639284e-004,
    1.0199614584522137e-003,
   -2.9999999990440909e-005,
    1.3999999995220451e-004,
   -5.9000345622022704e-004,
   -1.6860079490600130e-002,
    9.9999999961763637e-005,
   -3.0999699501009536e-004,
   -5.7250246051946750e-002,
    3.8999398927216423e-004,
   -3.3499471412369239e-003,
   -1.3500126888151384e-002,
    2.3999612978192091e-004,
   -2.0000000000000002e-005,
   -1.3699767812032594e-003,
    4.3899544269948888e-003,
    5.9999999980881817e-005,
   -1.8999999993308636e-004,
    4.9999827221431877e-004,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -5.9999999980881817e-005,
   -4.5998332063697402e-004,
    1.0000000000000001e-005,
    3.9198718475904771e-003,
   -3.9999999990440911e-005,
    1.5999999994264545e-004,
    4.7498424417146134e-003,
   -4.9999999980881818e-005,
    4.9999999980881818e-005,
    1.8099368283690308e-003,
   -2.0000000000000002e-005,
    3.4998859282063896e-004,
    4.9999999980881818e-005,
    1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -7.9993864654188287e-005,
   -3.9999999980881819e-005,
   -1.0000000000000001e-005,
   -1.9118187377780694e-014,
    1.0000000000000001e-005,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.9998245805436096e-004,
    5.8993423868899139e-004,
    5.8993423827196224e-004,
    2.3997545789740391e-004,
    5.0000000000000002e-005,
    1.3998772936573111e-004,
    3.1197876805299594e-003,
   -3.0000000000000001e-005,
    3.0000000000000001e-005,
    8.3594149696340162e-003,
   -6.9999999999999994e-005,
    3.0000000000000001e-005,
    7.4394759535129076e-003,
   -6.9999999999999994e-005,
   -2.0000000000000002e-005,
    2.8497974393342251e-003,
   -3.0000000000000001e-005,
   -2.0000000000000002e-005,
    5.5995877954049991e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.7599437444205150e-003,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    3.8998763101098795e-002,
   -3.8998171878128180e-004,
    4.3997558388117648e-004,
    9.0327051829633215e-002,
   -7.9995643740560663e-004,
    1.7999386384880725e-004,
    7.1417615893261940e-002,
   -6.3996557759793654e-004,
   -2.4999085980767006e-004,
    2.5099162146062979e-002,
   -2.5998385923368386e-004,
   -1.6000000000000001e-004,
    4.7698476225132812e-003,
   -6.0000000000000002e-005,
   -3.0000000000000001e-005,
    6.8997891392046096e-004,
   -1.0000000000000001e-005,
    9.0000000000000006e-005,
    2.0000172842855771e-004,
    1.3760086405580775e-002,
   -1.7000086442279342e-004,
    1.8000000000000001e-004,
    2.9255150843605804e-001,
   -2.8699635232183724e-003,
    2.0000000000000002e-005,
    3.6399493233071124e-003,
    5.7156243369584225e-001,
   -5.0899338785667068e-003,
    3.0000000000000001e-005,
    6.9997403395386472e-005,
    3.9451145761016659e-001,
   -3.5899493174687039e-003,
    2.0000000000000002e-005,
   -8.0000000000000007e-005,
   -2.1399673550604606e-003,
    1.2616048172676975e-001,
   -1.3199811017387127e-003,
   -1.0000000000000001e-005,
   -9.3998711399127875e-004,
    2.2850114068409776e-002,
   -3.0999699429074619e-004,
   -1.7999999958297086e-004,
    3.1800155483504944e-003,
   -5.0000000000000002e-005,
   -3.0000000000000001e-005,
    3.8000259201729283e-004,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
    2.7998945612617212e-004,
   -4.7798445528808757e-003,
    5.0000000000000002e-005,
   -2.5998385965071301e-004,
   -1.2819580249561514e-002,
    1.2000000000000000e-004,
   -6.9999999582970830e-005,
   -1.0639642712304840e-002,
    1.0000000000000001e-004,
    8.0000000000000007e-005,
   -3.9598669818323795e-003,
    4.0000000000000003e-005,
    4.0000000000000003e-005,
   -8.2997191376350387e-004,
    1.0000000000000001e-005,
   -1.2999999958297084e-004,
   -2.0000000000000002e-005,
   -6.0000000000000002e-005,
   -5.0996577969745702e-004,
    1.0000000000000001e-005,
   -6.0995964354626422e-004,
    1.0000000000000001e-005,
   -2.0998159321453837e-004,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.9999999990440911e-005,
   -1.5999386420848186e-004,
   -1.1999999953517537e-004,
    2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.5000172791593756e-004,
   -9.8000518325431097e-004,
    1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -6.8000259010063977e-004,
    1.0000000000000001e-005,
   -3.9999999980881819e-005,
   -4.9999999146823498e-005,
   -1.0000000000000001e-005,
    4.9999999980881818e-005,
    5.9999999980881817e-005,
    1.0999999996176364e-004,
    6.9999999971322711e-005,
    2.9999999990440909e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    6.0000000000000002e-005,
    1.8998859295446622e-004,
    2.4998245722030263e-004,
    1.4998772853167279e-004,
    5.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
    5.7998505007165373e-004,
    1.6999472848011523e-003,
   -2.0000000000000002e-005,
    1.9599341480759733e-003,
   -3.0000000000000001e-005,
    1.1199605054668827e-003,
   -1.0000000000000001e-005,
    3.6998945654320129e-004,
    8.0000000000000007e-005,
    1.0000000000000001e-005,
    1.8000172842855771e-004,
    3.2400207357213130e-003,
   -5.0000000000000002e-005,
    5.0000000000000002e-005,
    8.4000461959887664e-003,
   -1.1000000000000000e-004,
    1.0000000000000001e-005,
    8.6600422808942937e-003,
   -1.0999999958297084e-004,
   -5.0000000000000002e-005,
    4.5500194093078998e-003,
   -6.0000000000000002e-005,
   -4.0000000000000003e-005,
    1.3900051802813232e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9000172759449937e-004,
    5.0000000000000002e-005,
   -6.9999999999999994e-005,
   -2.3999472868862983e-004,
   -2.8998859295446626e-004,
   -1.6999386384880725e-004,
   -6.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    3.0000000000000001e-005,
    5.0000000000000002e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    1.1999999958297084e-004,
    1.7000086358873509e-004,
    1.2999999916594166e-004,
    6.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<ELPMPP02MainProblemNominalCoefficient, 918> g_MAIN_S2
{ {
 {  0,  1,  0,  0 },
 {  0,  3,  0,  0 },
 {  0,  5,  0,  0 },
 {  0, -5,  1,  0 },
 {  0, -3,  1,  0 },
 {  0, -1,  1,  0 },
 {  0,  1,  1,  0 },
 {  0,  3,  1,  0 },
 {  0,  5,  1,  0 },
 {  0, -5,  2,  0 },
 {  0, -3,  2,  0 },
 {  0, -1,  2,  0 },
 {  0,  1,  2,  0 },
 {  0,  3,  2,  0 },
 {  0,  5,  2,  0 },
 {  0, -5,  3,  0 },
 {  0, -3,  3,  0 },
 {  0, -1,  3,  0 },
 {  0,  1,  3,  0 },
 {  0,  3,  3,  0 },
 {  0,  5,  3,  0 },
 {  0, -3,  4,  0 },
 {  0, -1,  4,  0 },
 {  0,  1,  4,  0 },
 {  0,  3,  4,  0 },
 {  0, -1,  5,  0 },
 {  0,  1,  5,  0 },
 {  0,  3,  5,  0 },
 {  0, -1,  6,  0 },
 {  0,  1,  6,  0 },
 {  0, -1,  7,  0 },
 {  0,  1,  7,  0 },
 {  0, -1, -6,  1 },
 {  0,  1, -6,  1 },
 {  0, -1, -5,  1 },
 {  0,  1, -5,  1 },
 {  0, -3, -4,  1 },
 {  0, -1, -4,  1 },
 {  0,  1, -4,  1 },
 {  0, -3, -3,  1 },
 {  0, -1, -3,  1 },
 {  0,  1, -3,  1 },
 {  0,  3, -3,  1 },
 {  0, -3, -2,  1 },
 {  0, -1, -2,  1 },
 {  0,  1, -2,  1 },
 {  0,  3, -2,  1 },
 {  0, -5, -1,  1 },
 {  0, -3, -1,  1 },
 {  0, -1, -1,  1 },
 {  0,  1, -1,  1 },
 {  0,  3, -1,  1 },
 {  0, -3,  0,  1 },
 {  0, -1,  0,  1 },
 {  0,  1,  0,  1 },
 {  0,  3,  0,  1 },
 {  0,  5,  0,  1 },
 {  0, -3,  1,  1 },
 {  0, -1,  1,  1 },
 {  0,  1,  1,  1 },
 {  0,  3,  1,  1 },
 {  0,  5,  1,  1 },
 {  0, -3,  2,  1 },
 {  0, -1,  2,  1 },
 {  0,  1,  2,  1 },
 {  0,  3,  2,  1 },
 {  0, -3,  3,  1 },
 {  0, -1,  3,  1 },
 {  0,  1,  3,  1 },
 {  0,  3,  3,  1 },
 {  0, -1,  4,  1 },
 {  0,  1,  4,  1 },
 {  0,  3,  4,  1 },
 {  0, -1,  5,  1 },
 {  0,  1,  5,  1 },
 {  0, -1,  6,  1 },
 {  0,  1,  6,  1 },
 {  0, -1, -5,  2 },
 {  0, -1, -4,  2 },
 {  0,  1, -4,  2 },
 {  0, -1, -3,  2 },
 {  0,  1, -3,  2 },
 {  0, -3, -2,  2 },
 {  0, -1, -2,  2 },
 {  0,  1, -2,  2 },
 {  0,  3, -2,  2 },
 {  0, -3, -1,  2 },
 {  0, -1, -1,  2 },
 {  0,  1, -1,  2 },
 {  0,  3, -1,  2 },
 {  0, -3,  0,  2 },
 {  0, -1,  0,  2 },
 {  0,  1,  0,  2 },
 {  0,  3,  0,  2 },
 {  0, -3,  1,  2 },
 {  0, -1,  1,  2 },
 {  0,  1,  1,  2 },
 {  0,  3,  1,  2 },
 {  0, -3,  2,  2 },
 {  0, -1,  2,  2 },
 {  0,  1,  2,  2 },
 {  0,  3,  2,  2 },
 {  0, -1,  3,  2 },
 {  0,  1,  3,  2 },
 {  0,  1,  4,  2 },
 {  0, -1, -3,  3 },
 {  0,  1, -3,  3 },
 {  0, -1, -2,  3 },
 {  0,  1, -2,  3 },
 {  0, -1, -1,  3 },
 {  0,  1, -1,  3 },
 {  0, -1,  0,  3 },
 {  0,  1,  0,  3 },
 {  0, -1,  1,  3 },
 {  0,  1,  1,  3 },
 {  0, -1,  2,  3 },
 {  0,  1,  2,  3 },
 {  0, -1, -1,  4 },
 {  0,  1, -1,  4 },
 {  0, -1,  0,  4 },
 {  0,  1,  0,  4 },
 {  0, -1,  1,  4 },
 {  0,  1,  1,  4 },
 {  1, -1, -1, -3 },
 {  1, -1,  0, -3 },
 {  1,  1,  0, -3 },
 {  1, -1,  1, -3 },
 {  1, -1, -2, -2 },
 {  1,  1, -2, -2 },
 {  1, -1, -1, -2 },
 {  1,  1, -1, -2 },
 {  1, -1,  0, -2 },
 {  1,  1,  0, -2 },
 {  1, -1,  1, -2 },
 {  1,  1,  1, -2 },
 {  1, -1,  2, -2 },
 {  1,  1,  2, -2 },
 {  1, -1, -4, -1 },
 {  1, -1, -3, -1 },
 {  1,  1, -3, -1 },
 {  1, -1, -2, -1 },
 {  1,  1, -2, -1 },
 {  1, -3, -1, -1 },
 {  1, -1, -1, -1 },
 {  1,  1, -1, -1 },
 {  1, -3,  0, -1 },
 {  1, -1,  0, -1 },
 {  1,  1,  0, -1 },
 {  1,  3,  0, -1 },
 {  1, -3,  1, -1 },
 {  1, -1,  1, -1 },
 {  1,  1,  1, -1 },
 {  1,  3,  1, -1 },
 {  1, -1,  2, -1 },
 {  1,  1,  2, -1 },
 {  1, -1,  3, -1 },
 {  1,  1,  3, -1 },
 {  1, -1,  4, -1 },
 {  1,  1,  4, -1 },
 {  1, -1, -5,  0 },
 {  1,  1, -5,  0 },
 {  1, -1, -4,  0 },
 {  1,  1, -4,  0 },
 {  1, -1, -3,  0 },
 {  1,  1, -3,  0 },
 {  1,  3, -3,  0 },
 {  1, -3, -2,  0 },
 {  1, -1, -2,  0 },
 {  1,  1, -2,  0 },
 {  1,  3, -2,  0 },
 {  1, -3, -1,  0 },
 {  1, -1, -1,  0 },
 {  1,  1, -1,  0 },
 {  1,  3, -1,  0 },
 {  1, -5,  0,  0 },
 {  1, -3,  0,  0 },
 {  1, -1,  0,  0 },
 {  1,  1,  0,  0 },
 {  1,  3,  0,  0 },
 {  1,  5,  0,  0 },
 {  1, -5,  1,  0 },
 {  1, -3,  1,  0 },
 {  1, -1,  1,  0 },
 {  1,  1,  1,  0 },
 {  1,  3,  1,  0 },
 {  1, -1,  2,  0 },
 {  1,  1,  2,  0 },
 {  1,  3,  2,  0 },
 {  1, -1,  3,  0 },
 {  1,  1,  3,  0 },
 {  1,  3,  3,  0 },
 {  1, -1,  4,  0 },
 {  1,  1,  4,  0 },
 {  1, -1,  5,  0 },
 {  1,  1,  5,  0 },
 {  1, -1, -5,  1 },
 {  1, -1, -4,  1 },
 {  1,  1, -4,  1 },
 {  1, -1, -3,  1 },
 {  1,  1, -3,  1 },
 {  1, -3, -2,  1 },
 {  1, -1, -2,  1 },
 {  1,  1, -2,  1 },
 {  1,  3, -2,  1 },
 {  1, -3, -1,  1 },
 {  1, -1, -1,  1 },
 {  1,  1, -1,  1 },
 {  1,  3, -1,  1 },
 {  1, -3,  0,  1 },
 {  1, -1,  0,  1 },
 {  1,  1,  0,  1 },
 {  1,  3,  0,  1 },
 {  1, -3,  1,  1 },
 {  1, -1,  1,  1 },
 {  1,  1,  1,  1 },
 {  1,  3,  1,  1 },
 {  1, -3,  2,  1 },
 {  1, -1,  2,  1 },
 {  1,  1,  2,  1 },
 {  1,  3,  2,  1 },
 {  1, -1,  3,  1 },
 {  1,  1,  3,  1 },
 {  1, -1,  4,  1 },
 {  1,  1,  4,  1 },
 {  1, -1, -3,  2 },
 {  1,  1, -3,  2 },
 {  1, -1, -2,  2 },
 {  1,  1, -2,  2 },
 {  1, -1, -1,  2 },
 {  1,  1, -1,  2 },
 {  1, -1,  0,  2 },
 {  1,  1,  0,  2 },
 {  1, -1,  1,  2 },
 {  1,  1,  1,  2 },
 {  1, -1,  2,  2 },
 {  1,  1,  2,  2 },
 {  1,  1,  3,  2 },
 {  1, -1,  0,  3 },
 {  1,  1,  0,  3 },
 {  2, -1, -1, -5 },
 {  2,  1, -1, -5 },
 {  2, -1,  0, -5 },
 {  2,  1,  0, -5 },
 {  2, -1, -2, -4 },
 {  2, -1, -1, -4 },
 {  2,  1, -1, -4 },
 {  2, -1,  0, -4 },
 {  2,  1,  0, -4 },
 {  2, -1,  1, -4 },
 {  2,  1,  1, -4 },
 {  2,  1,  2, -4 },
 {  2, -1, -3, -3 },
 {  2, -1, -2, -3 },
 {  2,  1, -2, -3 },
 {  2, -1, -1, -3 },
 {  2,  1, -1, -3 },
 {  2,  3, -1, -3 },
 {  2, -3,  0, -3 },
 {  2, -1,  0, -3 },
 {  2,  1,  0, -3 },
 {  2,  3,  0, -3 },
 {  2, -3,  1, -3 },
 {  2, -1,  1, -3 },
 {  2,  1,  1, -3 },
 {  2, -1,  2, -3 },
 {  2,  1,  2, -3 },
 {  2, -1,  3, -3 },
 {  2,  1,  3, -3 },
 {  2, -1, -4, -2 },
 {  2,  1, -4, -2 },
 {  2, -1, -3, -2 },
 {  2,  1, -3, -2 },
 {  2, -1, -2, -2 },
 {  2,  1, -2, -2 },
 {  2,  3, -2, -2 },
 {  2, -3, -1, -2 },
 {  2, -1, -1, -2 },
 {  2,  1, -1, -2 },
 {  2,  3, -1, -2 },
 {  2, -3,  0, -2 },
 {  2, -1,  0, -2 },
 {  2,  1,  0, -2 },
 {  2,  3,  0, -2 },
 {  2, -3,  1, -2 },
 {  2, -1,  1, -2 },
 {  2,  1,  1, -2 },
 {  2,  3,  1, -2 },
 {  2, -3,  2, -2 },
 {  2, -1,  2, -2 },
 {  2,  1,  2, -2 },
 {  2,  3,  2, -2 },
 {  2, -1,  3, -2 },
 {  2,  1,  3, -2 },
 {  2, -1,  4, -2 },
 {  2,  1,  4, -2 },
 {  2, -1, -6, -1 },
 {  2, -1, -5, -1 },
 {  2,  1, -5, -1 },
 {  2, -1, -4, -1 },
 {  2,  1, -4, -1 },
 {  2, -3, -3, -1 },
 {  2, -1, -3, -1 },
 {  2,  1, -3, -1 },
 {  2,  3, -3, -1 },
 {  2, -3, -2, -1 },
 {  2, -1, -2, -1 },
 {  2,  1, -2, -1 },
 {  2,  3, -2, -1 },
 {  2, -5, -1, -1 },
 {  2, -3, -1, -1 },
 {  2, -1, -1, -1 },
 {  2,  1, -1, -1 },
 {  2,  3, -1, -1 },
 {  2,  5, -1, -1 },
 {  2, -5,  0, -1 },
 {  2, -3,  0, -1 },
 {  2, -1,  0, -1 },
 {  2,  1,  0, -1 },
 {  2,  3,  0, -1 },
 {  2,  5,  0, -1 },
 {  2, -5,  1, -1 },
 {  2, -3,  1, -1 },
 {  2, -1,  1, -1 },
 {  2,  1,  1, -1 },
 {  2,  3,  1, -1 },
 {  2, -3,  2, -1 },
 {  2, -1,  2, -1 },
 {  2,  1,  2, -1 },
 {  2,  3,  2, -1 },
 {  2, -3,  3, -1 },
 {  2, -1,  3, -1 },
 {  2,  1,  3, -1 },
 {  2,  3,  3, -1 },
 {  2, -1,  4, -1 },
 {  2,  1,  4, -1 },
 {  2, -1,  5, -1 },
 {  2,  1,  5, -1 },
 {  2, -1, -7,  0 },
 {  2,  1, -7,  0 },
 {  2, -1, -6,  0 },
 {  2,  1, -6,  0 },
 {  2, -3, -5,  0 },
 {  2, -1, -5,  0 },
 {  2,  1, -5,  0 },
 {  2, -3, -4,  0 },
 {  2, -1, -4,  0 },
 {  2,  1, -4,  0 },
 {  2,  3, -4,  0 },
 {  2, -3, -3,  0 },
 {  2, -1, -3,  0 },
 {  2,  1, -3,  0 },
 {  2,  3, -3,  0 },
 {  2,  5, -3,  0 },
 {  2, -5, -2,  0 },
 {  2, -3, -2,  0 },
 {  2, -1, -2,  0 },
 {  2,  1, -2,  0 },
 {  2,  3, -2,  0 },
 {  2,  5, -2,  0 },
 {  2, -5, -1,  0 },
 {  2, -3, -1,  0 },
 {  2, -1, -1,  0 },
 {  2,  1, -1,  0 },
 {  2,  3, -1,  0 },
 {  2,  5, -1,  0 },
 {  2, -5,  0,  0 },
 {  2, -3,  0,  0 },
 {  2, -1,  0,  0 },
 {  2,  1,  0,  0 },
 {  2,  3,  0,  0 },
 {  2,  5,  0,  0 },
 {  2, -5,  1,  0 },
 {  2, -3,  1,  0 },
 {  2, -1,  1,  0 },
 {  2,  1,  1,  0 },
 {  2,  3,  1,  0 },
 {  2,  5,  1,  0 },
 {  2, -5,  2,  0 },
 {  2, -3,  2,  0 },
 {  2, -1,  2,  0 },
 {  2,  1,  2,  0 },
 {  2,  3,  2,  0 },
 {  2,  5,  2,  0 },
 {  2, -3,  3,  0 },
 {  2, -1,  3,  0 },
 {  2,  1,  3,  0 },
 {  2,  3,  3,  0 },
 {  2, -3,  4,  0 },
 {  2, -1,  4,  0 },
 {  2,  1,  4,  0 },
 {  2,  3,  4,  0 },
 {  2, -1,  5,  0 },
 {  2,  1,  5,  0 },
 {  2, -1,  6,  0 },
 {  2,  1,  6,  0 },
 {  2, -1, -6,  1 },
 {  2, -1, -5,  1 },
 {  2,  1, -5,  1 },
 {  2, -1, -4,  1 },
 {  2,  1, -4,  1 },
 {  2, -3, -3,  1 },
 {  2, -1, -3,  1 },
 {  2,  1, -3,  1 },
 {  2, -3, -2,  1 },
 {  2, -1, -2,  1 },
 {  2,  1, -2,  1 },
 {  2,  3, -2,  1 },
 {  2, -3, -1,  1 },
 {  2, -1, -1,  1 },
 {  2,  1, -1,  1 },
 {  2,  3, -1,  1 },
 {  2, -5,  0,  1 },
 {  2, -3,  0,  1 },
 {  2, -1,  0,  1 },
 {  2,  1,  0,  1 },
 {  2,  3,  0,  1 },
 {  2, -5,  1,  1 },
 {  2, -3,  1,  1 },
 {  2, -1,  1,  1 },
 {  2,  1,  1,  1 },
 {  2,  3,  1,  1 },
 {  2, -3,  2,  1 },
 {  2, -1,  2,  1 },
 {  2,  1,  2,  1 },
 {  2,  3,  2,  1 },
 {  2, -3,  3,  1 },
 {  2, -1,  3,  1 },
 {  2,  1,  3,  1 },
 {  2,  3,  3,  1 },
 {  2, -1,  4,  1 },
 {  2,  1,  4,  1 },
 {  2, -1,  5,  1 },
 {  2,  1,  5,  1 },
 {  2, -1, -4,  2 },
 {  2,  1, -4,  2 },
 {  2, -1, -3,  2 },
 {  2,  1, -3,  2 },
 {  2, -3, -2,  2 },
 {  2, -1, -2,  2 },
 {  2,  1, -2,  2 },
 {  2,  3, -2,  2 },
 {  2, -3, -1,  2 },
 {  2, -1, -1,  2 },
 {  2,  1, -1,  2 },
 {  2,  3, -1,  2 },
 {  2, -3,  0,  2 },
 {  2, -1,  0,  2 },
 {  2,  1,  0,  2 },
 {  2,  3,  0,  2 },
 {  2, -3,  1,  2 },
 {  2, -1,  1,  2 },
 {  2,  1,  1,  2 },
 {  2, -1,  2,  2 },
 {  2,  1,  2,  2 },
 {  2, -1, -2,  3 },
 {  2,  1, -2,  3 },
 {  2, -1, -1,  3 },
 {  2,  1, -1,  3 },
 {  2, -1,  0,  3 },
 {  2,  1,  0,  3 },
 {  2, -1,  1,  3 },
 {  3, -1, -1, -3 },
 {  3,  1, -1, -3 },
 {  3, -1,  0, -3 },
 {  3, -1, -3, -2 },
 {  3, -1, -2, -2 },
 {  3,  1, -2, -2 },
 {  3, -1, -1, -2 },
 {  3,  1, -1, -2 },
 {  3, -3,  0, -2 },
 {  3, -1,  0, -2 },
 {  3,  1,  0, -2 },
 {  3, -1,  1, -2 },
 {  3, -1,  2, -2 },
 {  3, -1, -4, -1 },
 {  3, -1, -3, -1 },
 {  3, -1, -2, -1 },
 {  3,  1, -2, -1 },
 {  3, -3, -1, -1 },
 {  3, -1, -1, -1 },
 {  3,  1, -1, -1 },
 {  3,  3, -1, -1 },
 {  3, -3,  0, -1 },
 {  3, -1,  0, -1 },
 {  3,  1,  0, -1 },
 {  3,  3,  0, -1 },
 {  3, -1,  1, -1 },
 {  3,  1,  1, -1 },
 {  3, -1,  2, -1 },
 {  3,  1,  2, -1 },
 {  3, -1,  3, -1 },
 {  3,  1,  3, -1 },
 {  3, -1, -4,  0 },
 {  3,  1, -4,  0 },
 {  3, -1, -3,  0 },
 {  3,  1, -3,  0 },
 {  3,  3, -3,  0 },
 {  3, -3, -2,  0 },
 {  3, -1, -2,  0 },
 {  3,  1, -2,  0 },
 {  3,  3, -2,  0 },
 {  3, -3, -1,  0 },
 {  3, -1, -1,  0 },
 {  3,  1, -1,  0 },
 {  3,  3, -1,  0 },
 {  3, -3,  0,  0 },
 {  3, -1,  0,  0 },
 {  3,  1,  0,  0 },
 {  3,  3,  0,  0 },
 {  3, -3,  1,  0 },
 {  3, -1,  1,  0 },
 {  3,  1,  1,  0 },
 {  3,  3,  1,  0 },
 {  3, -1,  2,  0 },
 {  3,  1,  2,  0 },
 {  3, -1,  3,  0 },
 {  3,  1,  3,  0 },
 {  3, -1,  4,  0 },
 {  3,  1,  4,  0 },
 {  3, -1, -4,  1 },
 {  3,  1, -4,  1 },
 {  3, -1, -3,  1 },
 {  3,  1, -3,  1 },
 {  3, -1, -2,  1 },
 {  3,  1, -2,  1 },
 {  3, -3, -1,  1 },
 {  3, -1, -1,  1 },
 {  3,  1, -1,  1 },
 {  3,  3, -1,  1 },
 {  3, -3,  0,  1 },
 {  3, -1,  0,  1 },
 {  3,  1,  0,  1 },
 {  3,  3,  0,  1 },
 {  3, -1,  1,  1 },
 {  3,  1,  1,  1 },
 {  3, -1,  2,  1 },
 {  3,  1,  2,  1 },
 {  3, -1,  3,  1 },
 {  3,  1,  3,  1 },
 {  3, -1, -2,  2 },
 {  3,  1, -2,  2 },
 {  3, -1, -1,  2 },
 {  3,  1, -1,  2 },
 {  3, -1,  0,  2 },
 {  3,  1,  0,  2 },
 {  3, -1,  1,  2 },
 {  3,  1,  1,  2 },
 {  3, -1,  2,  2 },
 {  3, -3, -1,  3 },
 {  3, -1, -1,  3 },
 {  3,  1, -1,  3 },
 {  4,  1, -2, -4 },
 {  4, -1, -1, -4 },
 {  4,  1, -1, -4 },
 {  4, -1,  0, -4 },
 {  4,  1,  0, -4 },
 {  4, -1,  1, -4 },
 {  4, -1, -2, -3 },
 {  4,  1, -2, -3 },
 {  4, -1, -1, -3 },
 {  4,  1, -1, -3 },
 {  4, -3,  0, -3 },
 {  4, -1,  0, -3 },
 {  4,  1,  0, -3 },
 {  4, -1,  1, -3 },
 {  4,  1,  1, -3 },
 {  4, -1,  2, -3 },
 {  4,  1,  2, -3 },
 {  4, -1, -4, -2 },
 {  4, -1, -3, -2 },
 {  4,  1, -3, -2 },
 {  4, -3, -2, -2 },
 {  4, -1, -2, -2 },
 {  4,  1, -2, -2 },
 {  4,  3, -2, -2 },
 {  4, -3, -1, -2 },
 {  4, -1, -1, -2 },
 {  4,  1, -1, -2 },
 {  4,  3, -1, -2 },
 {  4, -3,  0, -2 },
 {  4, -1,  0, -2 },
 {  4,  1,  0, -2 },
 {  4,  3,  0, -2 },
 {  4, -3,  1, -2 },
 {  4, -1,  1, -2 },
 {  4,  1,  1, -2 },
 {  4, -1,  2, -2 },
 {  4,  1,  2, -2 },
 {  4, -1,  3, -2 },
 {  4,  1,  3, -2 },
 {  4, -1, -6, -1 },
 {  4, -1, -5, -1 },
 {  4, -1, -4, -1 },
 {  4,  1, -4, -1 },
 {  4, -3, -3, -1 },
 {  4, -1, -3, -1 },
 {  4,  1, -3, -1 },
 {  4,  3, -3, -1 },
 {  4, -3, -2, -1 },
 {  4, -1, -2, -1 },
 {  4,  1, -2, -1 },
 {  4,  3, -2, -1 },
 {  4, -3, -1, -1 },
 {  4, -1, -1, -1 },
 {  4,  1, -1, -1 },
 {  4,  3, -1, -1 },
 {  4, -5,  0, -1 },
 {  4, -3,  0, -1 },
 {  4, -1,  0, -1 },
 {  4,  1,  0, -1 },
 {  4,  3,  0, -1 },
 {  4, -3,  1, -1 },
 {  4, -1,  1, -1 },
 {  4,  1,  1, -1 },
 {  4,  3,  1, -1 },
 {  4, -3,  2, -1 },
 {  4, -1,  2, -1 },
 {  4,  1,  2, -1 },
 {  4,  3,  2, -1 },
 {  4, -3,  3, -1 },
 {  4, -1,  3, -1 },
 {  4,  1,  3, -1 },
 {  4, -1,  4, -1 },
 {  4,  1,  4, -1 },
 {  4, -1, -7,  0 },
 {  4, -1, -6,  0 },
 {  4,  1, -6,  0 },
 {  4, -1, -5,  0 },
 {  4, -3, -4,  0 },
 {  4, -1, -4,  0 },
 {  4,  1, -4,  0 },
 {  4,  3, -4,  0 },
 {  4, -3, -3,  0 },
 {  4, -1, -3,  0 },
 {  4,  1, -3,  0 },
 {  4,  3, -3,  0 },
 {  4, -3, -2,  0 },
 {  4, -1, -2,  0 },
 {  4,  1, -2,  0 },
 {  4,  3, -2,  0 },
 {  4,  5, -2,  0 },
 {  4, -5, -1,  0 },
 {  4, -3, -1,  0 },
 {  4, -1, -1,  0 },
 {  4,  1, -1,  0 },
 {  4,  3, -1,  0 },
 {  4,  5, -1,  0 },
 {  4, -5,  0,  0 },
 {  4, -3,  0,  0 },
 {  4, -1,  0,  0 },
 {  4,  1,  0,  0 },
 {  4,  3,  0,  0 },
 {  4, -3,  1,  0 },
 {  4, -1,  1,  0 },
 {  4,  1,  1,  0 },
 {  4,  3,  1,  0 },
 {  4, -3,  2,  0 },
 {  4, -1,  2,  0 },
 {  4,  1,  2,  0 },
 {  4,  3,  2,  0 },
 {  4, -3,  3,  0 },
 {  4, -1,  3,  0 },
 {  4,  1,  3,  0 },
 {  4,  3,  3,  0 },
 {  4, -1,  4,  0 },
 {  4,  1,  4,  0 },
 {  4, -1,  5,  0 },
 {  4,  1,  5,  0 },
 {  4, -1, -5,  1 },
 {  4, -1, -4,  1 },
 {  4,  1, -4,  1 },
 {  4, -1, -3,  1 },
 {  4,  1, -3,  1 },
 {  4, -3, -2,  1 },
 {  4, -1, -2,  1 },
 {  4,  1, -2,  1 },
 {  4,  3, -2,  1 },
 {  4, -3, -1,  1 },
 {  4, -1, -1,  1 },
 {  4,  1, -1,  1 },
 {  4,  3, -1,  1 },
 {  4, -3,  0,  1 },
 {  4, -1,  0,  1 },
 {  4,  1,  0,  1 },
 {  4,  3,  0,  1 },
 {  4, -3,  1,  1 },
 {  4, -1,  1,  1 },
 {  4,  1,  1,  1 },
 {  4,  3,  1,  1 },
 {  4, -3,  2,  1 },
 {  4, -1,  2,  1 },
 {  4,  1,  2,  1 },
 {  4, -1,  3,  1 },
 {  4,  1,  3,  1 },
 {  4, -1,  4,  1 },
 {  4,  1,  4,  1 },
 {  4, -1, -3,  2 },
 {  4,  1, -3,  2 },
 {  4, -3, -2,  2 },
 {  4, -1, -2,  2 },
 {  4,  1, -2,  2 },
 {  4, -3, -1,  2 },
 {  4, -1, -1,  2 },
 {  4,  1, -1,  2 },
 {  4, -1,  0,  2 },
 {  4,  1,  0,  2 },
 {  4, -1,  1,  2 },
 {  4,  1,  1,  2 },
 {  4,  1, -2,  3 },
 {  4, -1, -1,  3 },
 {  4,  1, -1,  3 },
 {  4, -1,  0,  3 },
 {  5, -1, -2, -2 },
 {  5,  1, -2, -2 },
 {  5, -1, -1, -2 },
 {  5,  1, -1, -2 },
 {  5, -1,  0, -2 },
 {  5, -1, -3, -1 },
 {  5,  1, -3, -1 },
 {  5, -1, -2, -1 },
 {  5,  1, -2, -1 },
 {  5, -3, -1, -1 },
 {  5, -1, -1, -1 },
 {  5,  1, -1, -1 },
 {  5, -3,  0, -1 },
 {  5, -1,  0, -1 },
 {  5, -1,  1, -1 },
 {  5,  1, -4,  0 },
 {  5, -1, -3,  0 },
 {  5,  1, -3,  0 },
 {  5, -1, -2,  0 },
 {  5,  1, -2,  0 },
 {  5,  3, -2,  0 },
 {  5, -3, -1,  0 },
 {  5, -1, -1,  0 },
 {  5,  1, -1,  0 },
 {  5,  3, -1,  0 },
 {  5, -3,  0,  0 },
 {  5, -1,  0,  0 },
 {  5,  1,  0,  0 },
 {  5, -3,  1,  0 },
 {  5, -1,  1,  0 },
 {  5,  1,  1,  0 },
 {  5, -1,  2,  0 },
 {  5,  1,  2,  0 },
 {  5, -1, -3,  1 },
 {  5,  1, -3,  1 },
 {  5, -1, -2,  1 },
 {  5,  1, -2,  1 },
 {  5, -1, -1,  1 },
 {  5,  1, -1,  1 },
 {  5, -3,  0,  1 },
 {  5, -1,  0,  1 },
 {  5,  1,  0,  1 },
 {  5, -1,  1,  1 },
 {  5,  1,  1,  1 },
 {  5, -1,  2,  1 },
 {  5, -1, -1,  2 },
 {  5, -1,  0,  2 },
 {  6,  1, -3, -3 },
 {  6, -1, -2, -3 },
 {  6,  1, -2, -3 },
 {  6, -1, -1, -3 },
 {  6,  1, -1, -3 },
 {  6, -1,  0, -3 },
 {  6,  1,  0, -3 },
 {  6, -1,  1, -3 },
 {  6, -1, -3, -2 },
 {  6,  1, -3, -2 },
 {  6, -1, -2, -2 },
 {  6,  1, -2, -2 },
 {  6, -3, -1, -2 },
 {  6, -1, -1, -2 },
 {  6,  1, -1, -2 },
 {  6, -1,  0, -2 },
 {  6,  1,  0, -2 },
 {  6, -1,  1, -2 },
 {  6,  1,  1, -2 },
 {  6, -1,  2, -2 },
 {  6,  1,  2, -2 },
 {  6, -1, -4, -1 },
 {  6,  1, -4, -1 },
 {  6, -1, -3, -1 },
 {  6,  1, -3, -1 },
 {  6,  3, -3, -1 },
 {  6, -1, -2, -1 },
 {  6,  1, -2, -1 },
 {  6,  3, -2, -1 },
 {  6, -3, -1, -1 },
 {  6, -1, -1, -1 },
 {  6,  1, -1, -1 },
 {  6,  3, -1, -1 },
 {  6, -3,  0, -1 },
 {  6, -1,  0, -1 },
 {  6,  1,  0, -1 },
 {  6,  3,  0, -1 },
 {  6, -1,  1, -1 },
 {  6,  1,  1, -1 },
 {  6, -1,  2, -1 },
 {  6,  1,  2, -1 },
 {  6, -1,  3, -1 },
 {  6,  1,  3, -1 },
 {  6, -1, -4,  0 },
 {  6,  1, -4,  0 },
 {  6,  3, -4,  0 },
 {  6, -1, -3,  0 },
 {  6,  1, -3,  0 },
 {  6,  3, -3,  0 },
 {  6, -3, -2,  0 },
 {  6, -1, -2,  0 },
 {  6,  1, -2,  0 },
 {  6,  3, -2,  0 },
 {  6, -3, -1,  0 },
 {  6, -1, -1,  0 },
 {  6,  1, -1,  0 },
 {  6,  3, -1,  0 },
 {  6, -3,  0,  0 },
 {  6, -1,  0,  0 },
 {  6,  1,  0,  0 },
 {  6,  3,  0,  0 },
 {  6, -3,  1,  0 },
 {  6, -1,  1,  0 },
 {  6,  1,  1,  0 },
 {  6,  3,  1,  0 },
 {  6, -3,  2,  0 },
 {  6, -1,  2,  0 },
 {  6,  1,  2,  0 },
 {  6, -1,  3,  0 },
 {  6,  1,  3,  0 },
 {  6, -1,  4,  0 },
 {  6,  1, -4,  1 },
 {  6, -1, -3,  1 },
 {  6,  1, -3,  1 },
 {  6, -1, -2,  1 },
 {  6,  1, -2,  1 },
 {  6, -3, -1,  1 },
 {  6, -1, -1,  1 },
 {  6,  1, -1,  1 },
 {  6, -3,  0,  1 },
 {  6, -1,  0,  1 },
 {  6,  1,  0,  1 },
 {  6, -1,  1,  1 },
 {  6,  1,  1,  1 },
 {  6, -1,  2,  1 },
 {  6,  1,  2,  1 },
 {  6, -1, -3,  2 },
 {  6,  1, -3,  2 },
 {  6, -1, -2,  2 },
 {  6,  1, -2,  2 },
 {  6, -1, -1,  2 },
 {  6,  1, -1,  2 },
 {  7, -1, -3, -1 },
 {  7,  1, -3, -1 },
 {  7, -1, -2, -1 },
 {  7,  1, -2, -1 },
 {  7, -1, -1, -1 },
 {  7,  1, -4,  0 },
 {  7, -1, -3,  0 },
 {  7,  1, -3,  0 },
 {  7, -1, -2,  0 },
 {  7,  1, -2,  0 },
 {  7, -3, -1,  0 },
 {  7, -1, -1,  0 },
 {  7,  1, -1,  0 },
 {  7, -1,  0,  0 },
 {  7, -1, -2,  1 },
 {  7,  1, -2,  1 },
 {  7, -1, -1,  1 },
 {  7,  1, -1,  1 },
 {  8, -1, -3, -2 },
 {  8,  1, -3, -2 },
 {  8, -1, -2, -2 },
 {  8,  1, -2, -2 },
 {  8, -1, -1, -2 },
 {  8,  1, -1, -2 },
 {  8, -1,  0, -2 },
 {  8, -1, -4, -1 },
 {  8,  1, -4, -1 },
 {  8, -1, -3, -1 },
 {  8,  1, -3, -1 },
 {  8, -1, -2, -1 },
 {  8,  1, -2, -1 },
 {  8, -1, -1, -1 },
 {  8,  1, -1, -1 },
 {  8, -1,  0, -1 },
 {  8,  1,  0, -1 },
 {  8, -1,  1, -1 },
 {  8,  1,  1, -1 },
 {  8,  1, -5,  0 },
 {  8, -1, -4,  0 },
 {  8,  1, -4,  0 },
 {  8, -1, -3,  0 },
 {  8,  1, -3,  0 },
 {  8, -3, -2,  0 },
 {  8, -1, -2,  0 },
 {  8,  1, -2,  0 },
 {  8, -3, -1,  0 },
 {  8, -1, -1,  0 },
 {  8,  1, -1,  0 },
 {  8, -1,  0,  0 },
 {  8,  1,  0,  0 },
 {  8, -1,  1,  0 },
 {  8,  1,  1,  0 },
 {  8, -1,  2,  0 },
 {  8, -1, -3,  1 },
 {  8,  1, -3,  1 },
 {  8, -1, -2,  1 },
 {  8,  1, -2,  1 },
 {  8, -1, -1,  1 },
 {  8,  1, -1,  1 },
 {  8, -1,  0,  1 },
 { 10, -1, -4,  0 },
 { 10,  1, -4,  0 },
 { 10, -1, -3,  0 },
 { 10,  1, -3,  0 },
 { 10, -1, -2,  0 },
 { 10,  1, -2,  0 },
 { 10, -1, -1,  0 }
} };

constexpr array<double, 918> g_MAIN_S2_Nominal
{ {
   18461.40000,
      -6.29664,
       0.00592,
      -0.00236,
       2.79871,
     999.70079,
    1010.17430,
      -1.01941,
       0.00156,
      -0.00044,
      -0.13035,
      31.75985,
      61.91229,
      -0.11787,
       0.00026,
       0.00003,
      -0.00146,
       1.58131,
       3.98407,
      -0.01181,
       0.00004,
      -0.00006,
       0.09157,
       0.26325,
      -0.00109,
       0.00566,
       0.01768,
      -0.00010,
       0.00036,
       0.00120,
       0.00002,
       0.00008,
      -0.00004,
      -0.00001,
      -0.00053,
      -0.00014,
       0.00003,
      -0.00646,
      -0.00186,
       0.00023,
      -0.07479,
      -0.02365,
       0.00002,
       0.00154,
      -0.79322,
      -0.30129,
       0.00191,
      -0.00001,
       0.00679,
      -6.73173,
      -5.63260,
      -0.00911,
       0.00010,
      -4.83983,
      -6.46036,
       0.01157,
      -0.00002,
      -0.00703,
      -5.07614,
      -5.31151,
       0.00668,
      -0.00001,
       0.00141,
      -0.31292,
      -0.63884,
       0.00135,
       0.00003,
      -0.02419,
      -0.06176,
       0.00020,
      -0.00191,
      -0.00548,
       0.00002,
      -0.00015,
      -0.00046,
      -0.00001,
      -0.00004,
      -0.00001,
      -0.00016,
      -0.00004,
      -0.00169,
      -0.00049,
       0.00003,
      -0.01571,
      -0.00536,
       0.00004,
       0.00010,
      -0.11335,
      -0.09511,
      -0.00009,
      -0.00014,
      -0.01801,
      -0.05729,
       0.00016,
      -0.00007,
      -0.06187,
      -0.05504,
       0.00007,
       0.00001,
      -0.00284,
      -0.00528,
       0.00001,
      -0.00016,
      -0.00037,
      -0.00002,
      -0.00004,
      -0.00001,
      -0.00033,
      -0.00011,
      -0.00216,
      -0.00190,
       0.00066,
      -0.00031,
      -0.00109,
      -0.00083,
      -0.00004,
      -0.00008,
      -0.00004,
      -0.00004,
       0.00004,
       0.00001,
      -0.00002,
      -0.00001,
       0.00003,
       0.00011,
       0.00012,
      -0.00002,
       0.00007,
      -0.00003,
       0.00064,
      -0.00001,
       0.00204,
       0.00214,
      -0.00060,
       0.00005,
      -0.00006,
      -0.00003,
       0.00003,
       0.00018,
      -0.00004,
       0.00096,
      -0.00137,
      -0.00003,
       0.01031,
       0.00080,
      -0.00052,
      -0.01346,
      -0.01829,
       0.00006,
      -0.00008,
      -0.02012,
      -0.00845,
       0.00002,
      -0.00175,
      -0.00146,
      -0.00015,
      -0.00018,
      -0.00001,
      -0.00002,
      -0.00011,
      -0.00004,
      -0.00122,
      -0.00043,
      -0.01255,
      -0.00537,
       0.00002,
      -0.00025,
      -0.10964,
      -0.07846,
      -0.00020,
      -0.00369,
      -0.42989,
       0.13928,
       0.00234,
       0.00003,
      -0.03226,
      -4.80578,
      -5.36844,
       0.00645,
      -0.00001,
       0.00001,
      -0.00019,
      -0.58893,
      -0.66741,
       0.00142,
      -0.03636,
      -0.06383,
       0.00021,
      -0.00247,
      -0.00553,
       0.00002,
      -0.00017,
      -0.00045,
      -0.00001,
      -0.00004,
       0.00001,
       0.00017,
       0.00007,
       0.00186,
       0.00093,
      -0.00001,
       0.01597,
       0.01680,
       0.00006,
       0.00007,
       0.00635,
      -0.05590,
      -0.00021,
       0.00080,
       0.80426,
       0.80263,
      -0.00083,
      -0.00021,
       0.03465,
       0.10176,
      -0.00020,
       0.00001,
       0.00317,
       0.01016,
      -0.00003,
       0.00027,
       0.00093,
       0.00002,
       0.00008,
       0.00003,
       0.00001,
       0.00017,
       0.00016,
      -0.00008,
      -0.00039,
      -0.00170,
      -0.00204,
      -0.00022,
      -0.00072,
      -0.00003,
      -0.00012,
      -0.00001,
      -0.00003,
      -0.00003,
       0.00001,
       0.00001,
       0.00004,
       0.00002,
       0.00003,
       0.00031,
       0.00034,
       0.00118,
       0.00062,
       0.00008,
       0.00010,
       0.00001,
       0.00006,
       0.00077,
      -0.00005,
       0.00914,
       0.01042,
      -0.00002,
       0.00008,
       0.03647,
       0.01603,
      -0.00002,
      -0.00003,
       0.00240,
       0.00256,
       0.00021,
       0.00031,
       0.00002,
       0.00003,
       0.00012,
       0.00002,
       0.00178,
       0.00030,
       0.02285,
      -0.00192,
      -0.00009,
       0.00024,
       0.26865,
       0.31474,
      -0.00045,
       0.00273,
       1.08587,
       0.38353,
      -0.00044,
      -0.00072,
       0.06915,
       0.05848,
      -0.00011,
      -0.00003,
       0.00555,
       0.00677,
      -0.00002,
       0.00046,
       0.00070,
       0.00004,
       0.00007,
       0.00002,
       0.00029,
       0.00006,
       0.00416,
       0.00080,
      -0.00004,
       0.05502,
       0.00935,
       0.00007,
       0.00033,
       0.65025,
      -0.06208,
      -0.00267,
      -0.00001,
       0.01034,
       7.43488,
       8.86853,
      -0.01177,
       0.00002,
      -0.00009,
       0.08815,
      29.57794,
       7.95891,
      -0.00944,
       0.00002,
      -0.00003,
      -0.01669,
       1.76606,
       1.13466,
      -0.00224,
      -0.00061,
       0.12897,
       0.12387,
      -0.00036,
      -0.00006,
       0.00985,
       0.01211,
      -0.00005,
       0.00076,
       0.00111,
       0.00006,
       0.00010,
       0.00007,
       0.00002,
       0.00092,
       0.00020,
      -0.00003,
       0.01127,
       0.00230,
      -0.00014,
       0.13381,
       0.02496,
      -0.00002,
       0.00023,
       1.51564,
       0.25408,
       0.00232,
       0.00002,
      -0.00005,
       0.02045,
      15.56635,
      -1.62443,
      -0.06561,
       0.00012,
      -0.00046,
       0.32907,
     166.57528,
     199.48515,
      -0.24484,
       0.00038,
      -0.00221,
       2.18637,
     623.65783,
     117.26161,
      -0.14453,
       0.00026,
      -0.00075,
      -0.29116,
      33.35743,
      15.12165,
      -0.03038,
       0.00007,
       0.00008,
      -0.00995,
       2.14618,
       1.51976,
      -0.00451,
       0.00001,
      -0.00093,
       0.14642,
       0.13795,
      -0.00056,
      -0.00008,
       0.01027,
       0.01186,
      -0.00006,
       0.00073,
       0.00099,
       0.00005,
       0.00008,
       0.00003,
       0.00028,
       0.00005,
       0.00247,
       0.00034,
      -0.00002,
       0.01818,
       0.00162,
      -0.00015,
       0.07913,
       0.05429,
       0.00017,
      -0.00342,
      -0.79105,
      -1.31788,
       0.00192,
       0.00005,
      -0.05457,
     -12.09470,
      -1.26433,
       0.00177,
       0.00002,
       0.00465,
      -0.82275,
      -0.23702,
       0.00052,
       0.00021,
      -0.06283,
      -0.03142,
       0.00010,
       0.00002,
      -0.00496,
      -0.00355,
       0.00002,
      -0.00040,
      -0.00037,
      -0.00003,
      -0.00004,
      -0.00007,
      -0.00001,
      -0.00102,
      -0.00003,
      -0.00004,
      -0.01262,
      -0.00504,
       0.00004,
      -0.00046,
      -0.10535,
      -0.11330,
       0.00012,
      -0.00103,
      -0.13415,
      -0.01482,
       0.00003,
      -0.00006,
      -0.00617,
      -0.00145,
      -0.00025,
      -0.00009,
      -0.00006,
      -0.00002,
      -0.00009,
      -0.00005,
       0.00037,
       0.00004,
       0.00006,
      -0.00004,
      -0.00003,
      -0.00006,
      -0.00001,
      -0.00014,
      -0.00016,
      -0.00101,
      -0.00063,
      -0.00004,
      -0.00131,
       0.00008,
      -0.00014,
      -0.00001,
      -0.00001,
      -0.00022,
      -0.00282,
      -0.00329,
      -0.00009,
      -0.02104,
      -0.01356,
       0.00002,
      -0.00076,
      -0.02572,
      -0.00042,
       0.00001,
      -0.00275,
      -0.00030,
      -0.00027,
      -0.00007,
      -0.00002,
      -0.00001,
      -0.00016,
      -0.00003,
      -0.00287,
       0.00015,
       0.00001,
      -0.00008,
      -0.03941,
      -0.04852,
       0.00016,
      -0.00128,
      -0.30517,
      -0.20593,
       0.00038,
      -0.01009,
      -0.35183,
      -0.02840,
       0.00016,
       0.00002,
      -0.03611,
      -0.00658,
       0.00004,
      -0.00322,
      -0.00108,
      -0.00027,
      -0.00014,
      -0.00002,
      -0.00002,
      -0.00007,
      -0.00001,
      -0.00044,
       0.00004,
      -0.00152,
      -0.00261,
       0.00003,
       0.01321,
       0.02083,
      -0.00005,
       0.00031,
       0.03436,
       0.01351,
      -0.00003,
       0.00426,
       0.00258,
       0.00044,
       0.00035,
       0.00004,
       0.00004,
      -0.00002,
      -0.00001,
      -0.00005,
      -0.00010,
      -0.00056,
      -0.00016,
      -0.00009,
      -0.00004,
      -0.00001,
       0.00001,
       0.00000,
      -0.00001,
       0.00003,
       0.00007,
       0.00007,
       0.00008,
       0.00004,
       0.00001,
       0.00013,
       0.00059,
       0.00165,
       0.00134,
       0.00001,
       0.00161,
       0.00080,
       0.00024,
       0.00017,
       0.00003,
       0.00003,
      -0.00003,
      -0.00006,
       0.00011,
      -0.00002,
       0.00283,
       0.01230,
      -0.00004,
      -0.00005,
       0.03462,
       0.02380,
      -0.00005,
       0.00029,
       0.02899,
       0.01270,
      -0.00003,
      -0.00004,
       0.00417,
       0.00255,
       0.00047,
       0.00037,
       0.00005,
       0.00005,
      -0.00001,
      -0.00009,
      -0.00058,
       0.00002,
      -0.00005,
      -0.00090,
       0.00191,
      -0.00011,
      -0.00033,
       0.05251,
       0.21376,
      -0.00060,
      -0.00065,
       0.59580,
       0.33882,
      -0.00075,
      -0.00001,
       0.00534,
       0.41496,
       0.15791,
      -0.00037,
      -0.00062,
       0.05686,
       0.03009,
      -0.00009,
      -0.00010,
       0.00614,
       0.00421,
      -0.00002,
      -0.00001,
       0.00060,
       0.00050,
       0.00006,
       0.00005,
      -0.00002,
      -0.00015,
      -0.00001,
      -0.00116,
      -0.00007,
      -0.00680,
       0.00029,
       0.00003,
      -0.00066,
      -0.00771,
       0.02174,
      -0.00127,
      -0.00390,
       0.63371,
       2.41389,
      -0.00632,
       0.00001,
      -0.00004,
      -0.00554,
       6.57962,
       2.99850,
      -0.00664,
       0.00002,
      -0.00013,
       0.06257,
       3.67449,
       1.19188,
      -0.00288,
      -0.00552,
       0.47338,
       0.21259,
      -0.00068,
      -0.00090,
       0.04834,
       0.02828,
      -0.00012,
      -0.00010,
       0.00449,
       0.00323,
      -0.00002,
       0.00040,
       0.00034,
       0.00003,
       0.00003,
      -0.00001,
      -0.00001,
      -0.00001,
       0.00063,
       0.00192,
       0.00003,
      -0.00203,
      -0.02957,
       0.00009,
      -0.00015,
      -0.17191,
      -0.05097,
       0.00012,
      -0.00304,
      -0.11308,
      -0.02549,
       0.00007,
       0.00011,
      -0.01692,
      -0.00558,
       0.00002,
       0.00003,
      -0.00196,
      -0.00088,
      -0.00020,
      -0.00012,
      -0.00002,
      -0.00001,
      -0.00003,
      -0.00020,
       0.00001,
      -0.00114,
      -0.00277,
       0.00005,
      -0.00438,
      -0.00179,
      -0.00056,
      -0.00025,
      -0.00001,
      -0.00002,
       0.00002,
       0.00008,
       0.00002,
       0.00002,
      -0.00004,
      -0.00004,
      -0.00010,
      -0.00002,
      -0.00002,
      -0.00004,
      -0.00014,
      -0.00057,
      -0.00061,
      -0.00001,
      -0.00127,
      -0.00029,
      -0.00005,
      -0.00032,
      -0.00006,
      -0.00001,
      -0.00033,
      -0.00124,
      -0.00505,
      -0.00519,
       0.00002,
      -0.00011,
      -0.01049,
      -0.00245,
       0.00001,
      -0.00041,
      -0.00254,
      -0.00027,
      -0.00003,
      -0.00045,
      -0.00006,
      -0.00006,
      -0.00001,
      -0.00001,
      -0.00005,
       0.00014,
       0.00041,
       0.00097,
       0.00053,
       0.00003,
       0.00051,
       0.00023,
       0.00010,
       0.00005,
       0.00001,
      -0.00002,
      -0.00001,
       0.00002,
       0.00005,
       0.00006,
       0.00010,
       0.00006,
       0.00005,
       0.00003,
       0.00001,
       0.00007,
       0.00033,
       0.00088,
       0.00087,
       0.00001,
       0.00137,
       0.00078,
       0.00067,
       0.00032,
       0.00013,
       0.00007,
       0.00002,
       0.00001,
       0.00002,
       0.00009,
       0.00089,
       0.00415,
      -0.00002,
       0.01091,
       0.00939,
      -0.00003,
       0.00014,
       0.01486,
       0.00757,
      -0.00003,
       0.00010,
       0.00652,
       0.00288,
      -0.00001,
       0.00122,
       0.00063,
       0.00017,
       0.00010,
       0.00002,
       0.00001,
       0.00014,
       0.00072,
      -0.00002,
       0.00702,
       0.03118,
      -0.00014,
       0.00007,
       0.08096,
       0.05963,
      -0.00021,
       0.00107,
       0.09403,
       0.04217,
      -0.00014,
       0.00060,
       0.03674,
       0.01465,
      -0.00005,
      -0.00005,
       0.00654,
       0.00305,
      -0.00001,
      -0.00002,
       0.00088,
       0.00048,
       0.00010,
       0.00006,
       0.00001,
       0.00003,
      -0.00003,
      -0.00054,
      -0.00256,
      -0.00137,
      -0.00007,
      -0.00350,
      -0.00116,
      -0.00004,
      -0.00153,
      -0.00047,
      -0.00031,
      -0.00011,
      -0.00005,
      -0.00002,
      -0.00002,
      -0.00005,
      -0.00009,
      -0.00006,
      -0.00005,
      -0.00002,
      -0.00001,
      -0.00002,
      -0.00004,
      -0.00002,
      -0.00002,
      -0.00003,
      -0.00009,
      -0.00011,
      -0.00024,
      -0.00010,
      -0.00001,
      -0.00013,
      -0.00003,
      -0.00003,
       0.00002,
       0.00001,
       0.00002,
       0.00001,
       0.00002,
       0.00002,
       0.00004,
       0.00003,
       0.00004,
       0.00002,
       0.00001,
       0.00002,
       0.00007,
       0.00019,
       0.00022,
       0.00038,
       0.00025,
       0.00029,
       0.00015,
       0.00011,
       0.00005,
       0.00002,
       0.00001,
       0.00002,
       0.00009,
       0.00042,
       0.00107,
       0.00108,
       0.00002,
       0.00189,
       0.00111,
       0.00002,
       0.00130,
       0.00060,
       0.00045,
       0.00020,
       0.00009,
       0.00004,
       0.00001,
      -0.00004,
      -0.00003,
      -0.00008,
      -0.00004,
      -0.00006,
      -0.00002,
      -0.00002,
       0.00001,
       0.00002,
       0.00003,
       0.00003,
       0.00003,
       0.00002,
       0.00002
} };

constexpr array<double, 918> g_MAIN_S2_LLR
{ {
    1.8461240060000968e+004,
   -6.2964772977841825e+000,
    5.9197458538809708e-003,
   -2.3599035461247487e-003,
    2.7986413324884101e+000,
    9.9969365545185053e+002,
    1.0101671484113945e+003,
   -1.0193852023009262e+000,
    1.5599366993320986e-003,
   -4.3998235423644866e-004,
   -1.3034701080366828e-001,
    3.1759670884809800e+001,
    6.1911950388791020e+001,
   -1.1786731681604115e-001,
    2.5998923859072039e-004,
    3.0000000000000001e-005,
   -1.4599681890483385e-003,
    1.5813035948086931e+000,
    3.9840545271942500e+000,
   -1.1809750247165851e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
    9.1569770719508434e-002,
    2.6324939977672279e-001,
   -1.0899798186608965e-003,
    5.6599939577684196e-003,
    1.7679988827081931e-002,
   -1.0000000000000001e-004,
    3.6000347902297565e-004,
    1.1999996756596471e-003,
    2.0000000000000002e-005,
    8.0000869755743914e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -5.2998206829307402e-004,
   -1.4000086975574390e-004,
    3.0000000000000001e-005,
   -6.4597473547634478e-003,
   -1.8599291897615727e-003,
    2.2999085835322900e-004,
   -7.4786974647102392e-002,
   -2.3649035623645471e-002,
    2.0000000000000002e-005,
    1.5399112376924722e-003,
   -7.9318664391493243e-001,
   -3.0127728954834437e-001,
    1.9098820720763222e-003,
   -1.0000000000000001e-005,
    6.7895861954236419e-003,
   -6.7314362777559564e+000,
   -5.6323537674990405e+000,
   -9.1094391491054628e-003,
    1.0000522513244267e-004,
   -4.8396143029776244e+000,
   -6.4600720929713935e+000,
    1.1569281673480687e-002,
   -2.0000000000000002e-005,
   -7.0295714410915392e-003,
   -5.0759179266592636e+000,
   -5.3112784086527398e+000,
    6.6795914615678711e-003,
   -1.0000000000000001e-005,
    1.4099203789511268e-003,
   -3.1290679708702374e-001,
   -6.3881314739994810e-001,
    1.3499203797273597e-003,
    3.0000000000000001e-005,
   -2.4189016822182798e-002,
   -6.1757502558210935e-002,
    1.9999085835322900e-004,
   -1.9099291893734563e-003,
   -5.4797858278326480e-003,
    2.0000000000000002e-005,
   -1.5000086975574389e-004,
   -4.5998119892544658e-004,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.5998860106072977e-004,
   -4.0000000000000003e-005,
   -1.6898696306909403e-003,
   -4.8996192573906496e-004,
    3.0000000000000001e-005,
   -1.5708764479858101e-002,
   -5.3595782675569343e-003,
    4.0000000000000003e-005,
    9.9993865652492928e-005,
   -1.1334089571385463e-001,
   -9.5102356895458595e-002,
   -8.9993865652492929e-005,
   -1.3998773130498583e-004,
   -1.8008571604879082e-002,
   -5.7285376738349575e-002,
    1.5998385424997801e-004,
   -6.9999999999999994e-005,
   -6.1865021524293226e-002,
   -5.5035578764496822e-002,
    6.9999999999999994e-005,
    1.0000000000000001e-005,
   -2.8397768666359189e-003,
   -5.2795835344773955e-003,
    1.0000000000000001e-005,
   -1.5998773130498586e-004,
   -3.6997720212145950e-004,
   -2.0000000000000002e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -3.2996406328258504e-004,
   -1.0998773130498585e-004,
   -2.1597448514783452e-003,
   -1.8997785305151446e-003,
    6.5991862945061870e-004,
   -3.0996319701988892e-004,
   -1.0898774320226049e-003,
   -8.2990196980074955e-004,
   -4.0000000000000003e-005,
   -7.9993865652492929e-005,
   -3.9993865652492926e-005,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9999999991103641e-005,
    1.0998773126940041e-004,
    1.1998773126050406e-004,
   -2.0000000000000002e-005,
    6.9999999973310921e-005,
   -2.9999999991103641e-005,
    6.3994791769827054e-004,
   -1.0000000000000001e-005,
    2.0398333889057457e-003,
    2.1398211209513792e-003,
   -5.9995405167545383e-004,
    4.9999999982207283e-005,
   -5.9999999973310922e-005,
   -2.9999999991103641e-005,
    2.9999999991103641e-005,
    1.7999473534596234e-004,
   -3.9999999991103643e-005,
    9.5995891849875222e-004,
   -1.3699436411977830e-003,
   -2.9999999991103641e-005,
    1.0309547887020005e-002,
    7.9995769526918837e-004,
   -5.1996996561452802e-004,
   -1.3459392256837954e-002,
   -1.8289191079292451e-002,
    5.9999999982207282e-005,
   -7.9999999973310934e-005,
   -2.0119122438275325e-002,
   -8.4496360037975109e-003,
    2.0000000000000002e-005,
   -1.7499283646819433e-003,
   -1.4599383750179688e-003,
   -1.4999999994662184e-004,
   -1.7999473534596234e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.1000086972015847e-004,
   -3.9999999991103643e-005,
   -1.2199961946660772e-003,
   -4.3000260911599361e-004,
   -1.2549955313959388e-002,
   -5.3699786917397428e-003,
    2.0000000000000002e-005,
   -2.4999224580102075e-004,
   -1.0963943472731864e-001,
   -7.8459564258628559e-002,
   -1.9999612287382131e-004,
   -3.6899091590814701e-003,
   -4.2988727861871234e-001,
    1.3927888273461300e-001,
    2.3399414433926996e-003,
    2.9999999991103641e-005,
   -3.2259166406104510e-002,
   -4.8057401256423811e+000,
   -5.3683960751123045e+000,
    6.4498350180914191e-003,
   -1.0000000000000001e-005,
    1.0000000000000001e-005,
   -1.8999612327083411e-004,
   -5.8892588698557768e-001,
   -6.6740564069239361e-001,
    1.4199677141185122e-003,
   -3.6359802477709945e-002,
   -6.3829693184392300e-002,
    2.0999612287382131e-004,
   -2.4699927896750925e-003,
   -5.5299795618302255e-003,
    2.0000000000000002e-005,
   -1.7000086969346941e-004,
   -4.5000260910709724e-004,
   -1.0000000000000001e-005,
   -3.9999999991103643e-005,
    1.0000000000000001e-005,
    1.6999473534596234e-004,
    6.9999999973310921e-005,
    1.8599257100980593e-003,
    9.2995978826339244e-004,
   -1.0000000000000001e-005,
    1.5969326708015971e-002,
    1.6799286143175864e-002,
    5.9999999973310922e-005,
    6.9999999973310921e-005,
    6.3497694317794888e-003,
   -5.5897523059121990e-002,
   -2.0998998852631422e-004,
    7.9994994271163835e-004,
    8.0422378129851813e-001,
    8.0259386654890696e-001,
   -8.2994606564773413e-004,
   -2.0998998891443065e-004,
    3.4648501148112923e-002,
    1.0175558253112338e-001,
   -1.9998998852631422e-004,
    1.0000000000000001e-005,
    3.1698667439869852e-003,
    1.0159573555631754e-002,
   -2.9999999991103641e-005,
    2.6998947071861371e-004,
    9.2995978826339244e-004,
    2.0000000000000002e-005,
    7.9999999973310934e-005,
    2.9999999991103641e-005,
    1.0000000000000001e-005,
    1.6998773125160771e-004,
    1.5998773125160771e-004,
   -7.9993865625803856e-005,
   -3.8996932735278636e-004,
   -1.6998618030783178e-003,
   -2.0398372655726373e-003,
   -2.1998159687741152e-004,
   -7.1993790588095014e-004,
   -2.9999999991103641e-005,
   -1.1999386560801113e-004,
   -1.0000000000000001e-005,
   -2.9999999991103641e-005,
   -2.9999999991103641e-005,
    1.0000000000000001e-005,
    1.0000000000000001e-005,
    3.9993865652492926e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0995705956745045e-004,
    3.3995092521994341e-004,
    1.1798166059404291e-003,
    6.1990410656803674e-004,
    7.9987731304985852e-005,
    9.9987731304985851e-005,
    1.0000000000000001e-005,
    5.9993865652492924e-005,
    7.6990897351588415e-004,
   -5.0000000000000002e-005,
    9.1389312020534186e-003,
    1.0418780220407221e-002,
   -2.0000000000000002e-005,
    7.9993865652492929e-005,
    3.6465675799467634e-002,
    1.6028102805463934e-002,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
    2.3997211838440607e-003,
    2.5596989021940664e-003,
    2.0997546260997170e-004,
    3.0996406328258504e-004,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    1.1998773130498585e-004,
    2.0000000000000002e-005,
    1.7798643668754098e-003,
    2.9997720212145948e-004,
    2.2848199762481741e-002,
   -1.9198494900775132e-003,
   -8.9993865652492929e-005,
    2.3997384284746312e-004,
    2.6862839641005326e-001,
    3.1471470172070520e-001,
   -4.4995769709744109e-004,
    2.7297277783777758e-003,
    1.0857809713365898e+000,
    3.8349873471131041e-001,
   -4.3995769670932472e-004,
   -7.1992627457690827e-004,
    6.9144455331716245e-002,
    5.8475325330337197e-002,
   -1.0999386565249293e-004,
   -3.0000000000000001e-005,
    5.5495668670651989e-003,
    6.7694684349527507e-003,
   -2.0000000000000002e-005,
    4.5996580279407287e-004,
    6.9994664896855414e-004,
    4.0000000000000003e-005,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    2.8999034057221757e-004,
    6.0000000000000002e-005,
    4.1598378381608710e-003,
    7.9996679268691247e-004,
   -4.0000000000000003e-005,
    5.5017766679959598e-002,
    9.3496194112699816e-003,
    6.9999999999999994e-005,
    3.2998084733883051e-004,
    6.5022258581976156e-001,
   -6.2077378785298058e-002,
   -2.6698407598428355e-003,
   -1.0000000000000001e-005,
    1.0339366937177970e-002,
    7.4345550490542491e+000,
    8.8681426389130262e+000,
   -1.1769285173540062e-002,
    2.0000000000000002e-005,
   -9.0000000000000006e-005,
    8.8144468391549075e-002,
    2.9576600341313640e+001,
    7.9585541680948806e+000,
   -9.4394217785200331e-003,
    2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -1.6688979285872985e-002,
    1.7659831729967461e+000,
    1.1346109855797797e+000,
   -2.2398673134749387e-003,
   -6.0996169351331178e-004,
    1.2896461784123031e-001,
    1.2386484200071171e-001,
   -3.5998171631834161e-004,
   -6.0000000000000002e-005,
    9.8496044825930381e-003,
    1.2109518389374237e-002,
   -5.0000000000000002e-005,
    7.5996679191067966e-004,
    1.1099562619509206e-003,
    6.0000000000000002e-005,
    1.0000000000000001e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    9.2000094267119070e-004,
    2.0000173951148783e-004,
   -3.0000000000000001e-005,
    1.1269992649539750e-002,
    2.2999988784312333e-003,
   -1.4000086975574390e-004,
    1.3380970029605976e-001,
    2.4959941165328410e-002,
   -2.0000000000000002e-005,
    2.2999612372122502e-004,
    1.5156341378557210e+000,
    2.5407898124276790e-001,
    2.3199492720039632e-003,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
    2.0449529722484837e-002,
    1.5566265447597040e+001,
   -1.6244211538357358e+000,
   -6.5608497190723633e-002,
    1.1999612294499217e-004,
   -4.5998061472496085e-004,
    3.2906196151824524e-001,
    1.6657411527804530e+002,
    1.9948375963301740e+002,
   -2.4483408725828171e-001,
    3.7998449177996868e-004,
   -2.2099030728485714e-003,
    2.1863131782292418e+000,
    6.2365247461499450e+002,
    1.1726069509568451e+002,
   -1.4452645294360378e-001,
    2.5998836883497645e-004,
   -7.4996985331568131e-004,
   -2.9115288606638323e-001,
    3.3357202739468264e+001,
    1.5121554269599795e+001,
   -3.0379290732301105e-002,
    6.9999999999999994e-005,
    8.0000000000000007e-005,
   -9.9497676668088512e-003,
    2.1461691684088535e+000,
    1.5197527792844909e+000,
   -4.5099010788072465e-003,
    1.0000000000000001e-005,
   -9.2998108644867263e-004,
    1.4641952065136907e-001,
    1.3794956624262231e-001,
   -5.5998710027096749e-004,
   -8.0000000000000007e-005,
    1.0269980800016502e-002,
    1.1859978331534651e-002,
   -6.0000000000000002e-005,
    7.2999833379207529e-004,
    9.9000007252733027e-004,
    5.0000000000000002e-005,
    8.0000000000000007e-005,
    3.0000000000000001e-005,
    2.7999033979598468e-004,
    5.0000000000000002e-005,
    2.4699042477415332e-003,
    3.3999033940786832e-004,
   -2.0000000000000002e-005,
    1.8179263934687131e-002,
    1.6199348465636217e-003,
   -1.4999999999999999e-004,
    7.9126693357347011e-002,
    5.4287721682792306e-002,
    1.6999612410934146e-004,
   -3.4197933612597239e-003,
   -7.9101527799402760e-001,
   -1.3178223061528571e+000,
    1.9198856006496451e-003,
    5.0000000000000002e-005,
   -5.4566578036660519e-002,
   -1.2094151145054148e+001,
   -1.2642738981658888e+000,
    1.7698886052320943e-003,
    2.0000000000000002e-005,
    4.6497158765900206e-003,
   -8.2271407173140498e-001,
   -2.3700982134666662e-001,
    5.1997170452771019e-004,
    2.0998998859748510e-004,
   -6.2827365794592221e-002,
   -3.1418700872567075e-002,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
   -4.9597976700850172e-003,
   -3.5498584216992809e-003,
    2.0000000000000002e-005,
   -3.9999034018410118e-004,
   -3.6999033979598470e-004,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -6.9999999999999994e-005,
   -1.0000000000000001e-005,
   -1.0199238510900136e-003,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.2619005146508959e-002,
   -5.0396058064244801e-003,
    4.0000000000000003e-005,
   -4.5995156197370119e-004,
   -1.0534151720387071e-001,
   -1.1329088105326769e-001,
    1.1999386565249293e-004,
   -1.0299013852121475e-003,
   -1.3413898599894700e-001,
   -1.4818830097282552e-002,
    3.0000000000000001e-005,
   -6.0000000000000002e-005,
   -6.1695030832341060e-003,
   -1.4498910410327254e-003,
   -2.4998246671322269e-004,
   -8.9993865652492929e-005,
   -5.9993865652492924e-005,
   -2.0000000000000002e-005,
   -8.9987731304985851e-005,
   -5.0000000000000002e-005,
    3.6995705917933404e-004,
    4.0000000000000003e-005,
    5.9993865652492924e-005,
   -3.9999999982207283e-005,
   -2.9999999991103641e-005,
   -5.9999999982207282e-005,
   -1.0000000000000001e-005,
   -1.3998773125160769e-004,
   -1.5998773125160771e-004,
   -1.0099203717200799e-003,
   -6.2994791653392130e-004,
   -3.9999999982207283e-005,
   -1.3098902168805669e-003,
    7.9999999973310934e-005,
   -1.3998773125160769e-004,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.1999473533706597e-004,
   -2.8198811476243570e-003,
   -3.2898606068781813e-003,
   -8.9999999964414566e-005,
   -2.1039078743255363e-002,
   -1.3559414140626998e-002,
    2.0000000000000002e-005,
   -7.5994994233241826e-004,
   -2.5718834864441851e-002,
   -4.1999642134587118e-004,
    1.0000000000000001e-005,
   -2.7498838014231117e-003,
   -2.9998946876913523e-004,
   -2.6998860058364974e-004,
   -6.9999999973310921e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.6000086970236577e-004,
   -2.9999999991103641e-005,
   -2.8699906512129978e-003,
    1.4999999994662184e-004,
    1.0000000000000001e-005,
   -7.9999999973310934e-005,
   -3.9409787893257976e-002,
   -4.8519742921784299e-002,
    1.5999612289161404e-004,
   -1.2799668451834831e-003,
   -3.0516792782752111e-001,
   -2.0592872345902902e-001,
    3.7999224498030612e-004,
   -1.0089736351726580e-002,
   -3.5182706067923597e-001,
   -2.8399990406986774e-002,
    1.5999612289161404e-004,
    2.0000000000000002e-005,
   -3.6109761777764510e-002,
   -6.5799909429395060e-003,
    3.9999999982207283e-005,
   -3.2199841616701742e-003,
   -1.0799974609458692e-003,
   -2.7000173902551143e-004,
   -1.3999999995551818e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -6.9999999973310921e-005,
   -1.0000000000000001e-005,
   -4.3998333554149565e-004,
    3.9999999991103643e-005,
   -1.5199383742239435e-003,
   -2.6098864103467089e-003,
    2.9999999991103641e-005,
    1.3209424682345905e-002,
    2.0829098772341424e-002,
   -4.9999999982207283e-005,
    3.0997997708821388e-004,
    3.4358451715049691e-002,
    1.3509405449297216e-002,
   -2.9999999991103641e-005,
    4.2598195706491041e-003,
    2.5798946838079947e-003,
    4.3998333631772846e-004,
    3.4998947030380825e-004,
    3.9999999982207283e-005,
    3.9999999982207283e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -4.9999999982207283e-005,
   -9.9993865616907488e-005,
   -5.5995318231672256e-004,
   -1.5998773125160771e-004,
   -8.9993865625803855e-005,
   -3.9999999982207283e-005,
   -1.0000000000000001e-005,
    1.0004394809330180e-005,
    1.1399318490311418e-008,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    6.9987731304985840e-005,
    6.9987731304985840e-005,
    7.9987731304985852e-005,
    3.9993865652492926e-005,
    1.0000000000000001e-005,
    1.2998773130498583e-004,
    5.8993038346955289e-004,
    1.6498030679051731e-003,
    1.3398437510452233e-003,
    1.0000000000000001e-005,
    1.6098065925973320e-003,
    7.9990723284004703e-004,
    2.3997546222185527e-004,
    1.6998159656936234e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -6.0000000000000002e-005,
    1.0999386565249293e-004,
   -2.0000000000000002e-005,
    2.8297768631428710e-003,
    1.2299032690033903e-002,
   -4.0000000000000003e-005,
   -5.0000000000000002e-005,
    3.4617218330952798e-002,
    2.3798096152639951e-002,
   -5.0000000000000002e-005,
    2.8996770849995599e-004,
    2.8987641108066016e-002,
    1.2698980263147393e-002,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
    4.1696670785365698e-003,
    2.5497996586927129e-003,
    4.6996105520708822e-004,
    3.6997720095711023e-004,
    5.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
   -9.0000000000000006e-005,
   -5.7998119814921375e-004,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9996592176681920e-004,
    1.9099257068574329e-003,
   -1.1000000000000000e-004,
   -3.2998084656259770e-004,
    5.2507790992369532e-002,
    2.1375103006616211e-001,
   -5.9996169312519535e-004,
   -6.4996082259321847e-004,
    5.9577402963628567e-001,
    3.3880542150545523e-001,
   -7.4995781529395462e-004,
   -1.0000000000000001e-005,
    5.3396641218539783e-003,
    4.1494151238801508e-001,
    1.5790314258793667e-001,
   -3.6997997641873730e-004,
   -6.1996082336945142e-004,
    5.6857563148343747e-002,
    3.0088730759479598e-002,
   -9.0000000000000006e-005,
   -1.0000000000000001e-004,
    6.1397443596772410e-003,
    4.2098291328411036e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    5.9997506341359023e-004,
    4.9998032800535348e-004,
    6.0000000000000002e-005,
    5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.5000173951148780e-004,
   -1.0000000000000001e-005,
   -1.1599979342075775e-003,
   -6.9999999999999994e-005,
   -6.7999853210948171e-003,
    2.9000173951148782e-004,
    3.0000000000000001e-005,
   -6.5998797002671152e-004,
   -7.7099706085507350e-003,
    2.1739908960198007e-002,
   -1.2699711932426714e-003,
   -3.8999113677165810e-003,
    6.3370660783864674e-001,
    2.4138773743315887e+000,
   -6.3198580285127315e-003,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -5.5398652333261906e-003,
    6.5795749896877664e+000,
    2.9984814500368659e+000,
   -6.6398459229071805e-003,
    2.0000000000000002e-005,
   -1.2999612294499217e-004,
    6.2568390379347888e-002,
    3.6744619049828269e+000,
    1.1918723349900475e+000,
   -2.8799332881790565e-003,
   -5.5198622334011382e-003,
    4.7337715948955844e-001,
    2.1258889481278204e-001,
   -6.7998709949473467e-004,
   -8.9997934616095187e-004,
    4.8339796415044024e-002,
    2.8279898739832832e-002,
   -1.2000000000000000e-004,
   -1.0000000000000001e-004,
    4.4899885821476871e-003,
    3.2299911184994713e-003,
   -2.0000000000000002e-005,
    4.0000260849099890e-004,
    3.4000260849099891e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
    6.2997419288161351e-004,
    1.9199257041406178e-003,
    3.0000000000000001e-005,
   -2.0299139827234633e-003,
   -2.9568742278036842e-002,
    9.0000000000000006e-005,
   -1.4999999961188355e-004,
   -1.7190245161696016e-001,
   -5.0967800333467587e-002,
    1.2000000000000000e-004,
   -3.0398120371516683e-003,
   -1.1307496074008441e-001,
   -2.5488897262257004e-002,
    6.9999999999999994e-005,
    1.1000000000000000e-004,
   -1.6919273478946544e-002,
   -5.5797675548115806e-003,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
   -1.9599178337746711e-003,
   -8.7996505123484245e-004,
   -1.9999473502012042e-004,
   -1.2000000000000000e-004,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.0000000000000001e-005,
   -1.9998860067261333e-004,
    1.0000000000000001e-005,
   -1.1399098413310458e-003,
   -2.7697829955497955e-003,
    5.0000000000000002e-005,
   -4.3796530687776024e-003,
   -1.7898600146036424e-003,
   -5.5995492085958117e-004,
   -2.4998246593698983e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    7.9987731304985852e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
   -3.9999999982207283e-005,
   -3.9999999982207283e-005,
   -9.9993865616907488e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.9999999991103643e-005,
   -1.3999999995551818e-004,
   -5.6997945844200601e-004,
   -6.0997332330047329e-004,
   -1.0000000000000001e-005,
   -1.2699419001605187e-003,
   -2.8998859979852051e-004,
   -4.9999999982207283e-005,
   -3.1998773003388028e-004,
   -5.9999999982207282e-005,
   -1.0000000000000001e-005,
   -3.3000173900771869e-004,
   -1.2399944547575766e-003,
   -5.0499738678599160e-003,
   -5.1899747344573508e-003,
    2.0000000000000002e-005,
   -1.0999999996441456e-004,
   -1.0489932195065683e-002,
   -2.4499884335310571e-003,
    1.0000000000000001e-005,
   -4.0998836830451830e-004,
   -2.5399828177052092e-003,
   -2.7000260723768606e-004,
   -2.9999999991103641e-005,
   -4.4999786114089258e-004,
   -5.9999999594090863e-005,
   -5.9999999982207282e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.9999999982207283e-005,
    1.3999999994662184e-004,
    4.0998947028601551e-004,
    9.6995717742590244e-004,
    5.2997858830704193e-004,
    3.0000000000000001e-005,
    5.0997771894831093e-004,
    2.2999386518430927e-004,
    9.9999999964414565e-005,
    4.9999999982207283e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
    5.9993865652492924e-005,
    9.9987731304985851e-005,
    5.9993865652492924e-005,
    5.0000000000000002e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    6.9999999999999994e-005,
    3.2997720134522664e-004,
    8.7993350974156315e-004,
    8.6993437910919070e-004,
    1.0000000000000001e-005,
    1.3698928250490470e-003,
    7.7993877394520988e-004,
    6.6994791598009738e-004,
    3.1997546144562245e-004,
    1.2998773130498583e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    8.8996592099058625e-004,
    4.1498317428845686e-003,
   -2.0000000000000002e-005,
    1.0909542985422739e-002,
    9.3896119613182654e-003,
   -3.0000000000000001e-005,
    1.3999999999999999e-004,
    1.4859362638795979e-002,
    7.5696827727210420e-003,
   -3.0000000000000001e-005,
    1.0000000000000001e-004,
    6.5197169005045694e-003,
    2.8798802722492340e-003,
   -1.0000000000000001e-005,
    1.2199489039323547e-003,
    6.2997419249349710e-004,
    1.6999386526437649e-004,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    1.4000086975574390e-004,
    7.1999659272812178e-004,
   -2.0000000000000002e-005,
    7.0199748793684938e-003,
    3.1179885019795069e-002,
   -1.3999999999999999e-004,
    6.9999999999999994e-005,
    8.0959591305152970e-002,
    5.9629735945832305e-002,
   -2.0999612255687575e-004,
    1.0699745985739673e-003,
    9.4029434590640532e-002,
    4.2169792378650837e-002,
   -1.3999999961188356e-004,
    5.9998449022750293e-004,
    3.6739771632229627e-002,
    1.4649926777976375e-002,
   -5.0000000000000002e-005,
   -5.0000000000000002e-005,
    6.5399648380029011e-003,
    3.0499880347917134e-003,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    8.7999572258426148e-004,
    4.7999873065975821e-004,
    1.0000000000000001e-004,
    6.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -5.3998032916970269e-004,
   -2.5598925471541109e-003,
   -1.3699445106488011e-003,
   -6.9999999999999994e-005,
   -3.4998471129340280e-003,
   -1.1599527817635954e-003,
   -4.0000000000000003e-005,
   -1.5299357650934802e-003,
   -4.6998333491650088e-004,
   -3.0998773052875295e-004,
   -1.1000000000000000e-004,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9993865652492929e-005,
   -6.0000000000000002e-005,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -3.9999999982207283e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -8.9999999973310933e-005,
   -1.0999999996441456e-004,
   -2.3999999914369992e-004,
   -9.9999999964414565e-005,
   -1.0000000000000001e-005,
   -1.2999999956740178e-004,
   -2.9999999991103641e-005,
   -2.9999999991103641e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    6.9999999999999994e-005,
    1.8999473502012042e-004,
    2.1999473463200401e-004,
    3.7998946965212444e-004,
    2.4999473463200396e-004,
    2.8998773014063655e-004,
    1.4999999922376715e-004,
    1.0999999961188359e-004,
    5.0000000000000002e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    4.2000260810288250e-004,
    1.0699974613195163e-003,
    1.0799974609313999e-003,
    2.0000000000000002e-005,
    1.8899914440041736e-003,
    1.1099965903994234e-003,
    2.0000000000000002e-005,
    1.2999918432005551e-003,
    5.9999872988352528e-004,
    4.5000086820327825e-004,
    1.9999999922376717e-004,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -3.0000000000000001e-005,
   -8.0000000000000007e-005,
   -4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<double, 918> g_MAIN_S2_DE405
{ {
    1.8461240380000967e+004,
   -6.2964776186139018e+000,
    5.9197463551783173e-003,
   -2.3599037212003134e-003,
    2.7986414518273106e+000,
    9.9969366315481682e+002,
    1.0101671561153941e+003,
   -1.0193852447752909e+000,
    1.5599368152025683e-003,
   -4.3998238235564216e-004,
   -1.3034701506935700e-001,
    3.1759670828363294e+001,
    6.1911950259640307e+001,
   -1.1786732060032996e-001,
    2.5998925652882658e-004,
    3.0000000000000001e-005,
   -1.4599682240705745e-003,
    1.5813035766718755e+000,
    3.9840544801889850e+000,
   -1.1809750513357475e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
    9.1569768780338345e-002,
    2.6324939411177511e-001,
   -1.0899798302964248e-003,
    5.6599937867737541e-003,
    1.7679988271116290e-002,
   -1.0000000000000001e-004,
    3.6000345769117371e-004,
    1.1999996296023475e-003,
    2.0000000000000002e-005,
    8.0000864422793429e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -5.2998204644754570e-004,
   -1.4000086442279340e-004,
    3.0000000000000001e-005,
   -6.4597471744927944e-003,
   -1.8599291354046150e-003,
    2.2999085980767006e-004,
   -7.4786973296373349e-002,
   -2.3649035208117827e-002,
    2.0000000000000002e-005,
    1.5399112744515736e-003,
   -7.9318663729777539e-001,
   -3.0127728711866791e-001,
    1.9098821237779097e-003,
   -1.0000000000000001e-005,
    6.7895864296686854e-003,
   -6.7314362866293180e+000,
   -5.6323537756655035e+000,
   -9.1094394883815977e-003,
    1.0000519362625620e-004,
   -4.8396143488170438e+000,
   -6.4600721505279193e+000,
    1.1569282182361574e-002,
   -2.0000000000000002e-005,
   -7.0295716955030710e-003,
   -5.0759179342248535e+000,
   -5.3112784145560363e+000,
    6.6795917022311434e-003,
   -1.0000000000000001e-005,
    1.4099204142268746e-003,
   -3.1290679453673831e-001,
   -6.3881314193420591e-001,
    1.3499204150609330e-003,
    3.0000000000000001e-005,
   -2.4189016390142579e-002,
   -6.1757501429886931e-002,
    1.9999085980767006e-004,
   -1.9099291349875857e-003,
   -5.4797856752831116e-003,
    2.0000000000000002e-005,
   -1.5000086442279340e-004,
   -4.5998118244178145e-004,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.5998859378852455e-004,
   -4.0000000000000003e-005,
   -1.6898695941853505e-003,
   -4.8996191165055552e-004,
    3.0000000000000001e-005,
   -1.5708764255911896e-002,
   -5.3595781915790472e-003,
    4.0000000000000003e-005,
    9.9993864682865568e-005,
   -1.1334089519200992e-001,
   -9.5102356472747202e-002,
   -8.9993864682865569e-005,
   -1.3998772936573111e-004,
   -1.8008571587748830e-002,
   -5.7285376796267398e-002,
    1.5998386006774219e-004,
   -6.9999999999999994e-005,
   -6.1865021262157795e-002,
   -5.5035578507142892e-002,
    6.9999999999999994e-005,
    1.0000000000000001e-005,
   -2.8397768268233711e-003,
   -5.2795834649755633e-003,
    1.0000000000000001e-005,
   -1.5998772936573114e-004,
   -3.6997718757704907e-004,
   -2.0000000000000002e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -3.2996405210295761e-004,
   -1.0998772936573112e-004,
   -2.1597448309137817e-003,
   -1.8997785123746494e-003,
    6.5991863301108811e-004,
   -3.0996319143342663e-004,
   -1.0898774183899964e-003,
   -8.2990196007556318e-004,
   -4.0000000000000003e-005,
   -7.9993864682865570e-005,
   -3.9993864682865560e-005,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9999999990440909e-005,
    1.0998772932749475e-004,
    1.1998772931793565e-004,
   -2.0000000000000002e-005,
    6.9999999971322711e-005,
   -2.9999999990440909e-005,
    6.3994791234875165e-004,
   -1.0000000000000001e-005,
    2.0398334043096023e-003,
    2.1398211344711548e-003,
   -5.9995404726797500e-004,
    4.9999999980881818e-005,
   -5.9999999971322719e-005,
   -2.9999999990440909e-005,
    2.9999999990440909e-005,
    1.7999472903874533e-004,
   -3.9999999990440911e-005,
    9.5995890683870267e-004,
   -1.3699436299252053e-003,
   -2.9999999990440909e-005,
    1.0309547898187386e-002,
    7.9995771355591250e-004,
   -5.1996998596344794e-004,
   -1.3459392383327749e-002,
   -1.8289191222779496e-002,
    5.9999999980881817e-005,
   -7.9999999971322724e-005,
   -2.0119122466600518e-002,
   -8.4496359940365653e-003,
    2.0000000000000002e-005,
   -1.7499283538553252e-003,
   -1.4599383573245114e-003,
   -1.4999999994264542e-004,
   -1.7999472903874533e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.1000086438455706e-004,
   -3.9999999990440911e-005,
   -1.2199961697937801e-003,
   -4.3000259310587563e-004,
   -1.2549955154032127e-002,
   -5.3699786296829404e-003,
    2.0000000000000002e-005,
   -2.4999226130843115e-004,
   -1.0963943437153734e-001,
   -7.8459564051565581e-002,
   -1.9999613062553833e-004,
   -3.6899093154652046e-003,
   -4.2988728107179575e-001,
    1.3927888402397687e-001,
    2.3399415459683273e-003,
    2.9999999990440909e-005,
   -3.2259168071847082e-002,
   -4.8057402037493979e+000,
   -5.3683961612559834e+000,
    6.4498353403772933e-003,
   -1.0000000000000001e-005,
    1.0000000000000001e-005,
   -1.8999613105212656e-004,
   -5.8892589129933504e-001,
   -6.6740564496553167e-001,
    1.4199677678096423e-003,
   -3.6359802389319490e-002,
   -6.3829692974737604e-002,
    2.0999613062553833e-004,
   -2.4699927536774319e-003,
   -5.5299794944654096e-003,
    2.0000000000000002e-005,
   -1.7000086435587977e-004,
   -4.5000259309631654e-004,
   -1.0000000000000001e-005,
   -3.9999999990440911e-005,
    1.0000000000000001e-005,
    1.6999472903874533e-004,
    6.9999999971322711e-005,
    1.8599256770251867e-003,
    9.2995977127105522e-004,
   -1.0000000000000001e-005,
    1.5969326570793512e-002,
    1.6799286013602480e-002,
    5.9999999971322719e-005,
    6.9999999971322711e-005,
    6.3497693454486765e-003,
   -5.5897523214725221e-002,
   -2.0998999530840389e-004,
    7.9994997662805132e-004,
    8.0422378961740737e-001,
    8.0259387492275991e-001,
   -8.2994610732050320e-004,
   -2.0998999572543304e-004,
    3.4648501149383414e-002,
    1.0175558259351709e-001,
   -1.9998999530840389e-004,
    1.0000000000000001e-005,
    3.1698667157862757e-003,
    1.0159573464138262e-002,
   -2.9999999990440909e-005,
    2.6998945810616790e-004,
    9.2995977127105522e-004,
    2.0000000000000002e-005,
    7.9999999971322724e-005,
    2.9999999990440909e-005,
    1.0000000000000001e-005,
    1.6998772930837657e-004,
    1.5998772930837657e-004,
   -7.9993864654188287e-005,
   -3.8996932243688307e-004,
   -1.6998618145833678e-003,
   -2.0398372731905618e-003,
   -2.1998159396256485e-004,
   -7.1993790728792190e-004,
   -2.9999999990440909e-005,
   -1.1999386463507010e-004,
   -1.0000000000000001e-005,
   -2.9999999990440909e-005,
   -2.9999999990440909e-005,
    1.0000000000000001e-005,
    1.0000000000000001e-005,
    3.9993864682865560e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0995705278005889e-004,
    3.3995091746292448e-004,
    1.1798166019751813e-003,
    6.1990409969390700e-004,
    7.9987729365731119e-005,
    9.9987729365731118e-005,
    1.0000000000000001e-005,
    5.9993864682865558e-005,
    7.6990895939846185e-004,
   -5.0000000000000002e-005,
    9.1389311067919324e-003,
    1.0418780115449460e-002,
   -2.0000000000000002e-005,
    7.9993864682865570e-005,
    3.6465675780352910e-002,
    1.6028102763916620e-002,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
    2.3997211540680375e-003,
    2.5596988771505289e-003,
    2.0997545873146224e-004,
    3.0996405210295761e-004,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    1.1998772936573112e-004,
    2.0000000000000002e-005,
    1.7798643241250674e-003,
    2.9997718757704904e-004,
    2.2848199449214169e-002,
   -1.9198494614735061e-003,
   -8.9993864682865569e-005,
    2.3997385545261880e-004,
    2.6862839528785964e-001,
    3.1471470039065103e-001,
   -4.4995771552036097e-004,
    2.7297278859196752e-003,
    1.0857809772361875e+000,
    3.8349873586206878e-001,
   -4.3995771510333184e-004,
   -7.1992629924458047e-004,
    6.9144454997641505e-002,
    5.8475324967037876e-002,
   -1.0999386468286557e-004,
   -3.0000000000000001e-005,
    5.5495667836994549e-003,
    6.7694683279979108e-003,
   -2.0000000000000002e-005,
    4.5996578094854450e-004,
    6.9994663530703279e-004,
    4.0000000000000003e-005,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    2.8999032263411138e-004,
    6.0000000000000002e-005,
    4.1598377235713664e-003,
    7.9996677135511047e-004,
   -4.0000000000000003e-005,
    5.5017765706985064e-002,
    9.3496192475408194e-003,
    6.9999999999999994e-005,
    3.2998085560957582e-004,
    6.5022258058474314e-001,
   -6.2077378299581383e-002,
   -2.6698408305388952e-003,
   -1.0000000000000001e-005,
    1.0339367321461741e-002,
    7.4345550611201423e+000,
    8.8681426530612608e+000,
   -1.1769285589814833e-002,
    2.0000000000000002e-005,
   -9.0000000000000006e-005,
    8.8144472523691508e-002,
    2.9576600674946288e+001,
    7.9585542335066863e+000,
   -9.4394221534410804e-003,
    2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -1.6688979891507517e-002,
    1.7659831748780288e+000,
    1.1346109846345061e+000,
   -2.2398673850249746e-003,
   -6.0996170996806417e-004,
    1.2896461667223055e-001,
    1.2386484072228615e-001,
   -3.5998171919831098e-004,
   -6.0000000000000002e-005,
    9.8496042942762367e-003,
    1.2109518146664864e-002,
   -5.0000000000000002e-005,
    7.5996677052105218e-004,
    1.1099562278983118e-003,
    6.0000000000000002e-005,
    1.0000000000000001e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    9.2000090479789809e-004,
    2.0000172884558686e-004,
   -3.0000000000000001e-005,
    1.1269992293506328e-002,
    2.2999988051376299e-003,
   -1.4000086442279340e-004,
    1.3380969738450180e-001,
    2.4959940627996887e-002,
   -2.0000000000000002e-005,
    2.2999613153606939e-004,
    1.5156341196347716e+000,
    2.5407897826836789e-001,
    2.3199493266433456e-003,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
    2.0449530411600761e-002,
    1.5566265411600057e+001,
   -1.6244211504417858e+000,
   -6.5608499340100476e-002,
    1.1999613070201107e-004,
   -4.5998065351005534e-004,
    3.2906197559064598e-001,
    1.6657411651352464e+002,
    1.9948376111198255e+002,
   -2.4483409732634209e-001,
    3.7998452280804429e-004,
   -2.2099032667162182e-003,
    2.1863132930572471e+000,
    6.2365248530516499e+002,
    1.1726069661889859e+002,
   -1.4452645931932776e-001,
    2.5998839210603319e-004,
   -7.4996991003888199e-004,
   -2.9115289842970743e-001,
    3.3357202973080071e+001,
    1.5121554333970961e+001,
   -3.0379291843547283e-002,
    6.9999999999999994e-005,
    8.0000000000000007e-005,
   -9.9497680019529668e-003,
    2.1461691617920584e+000,
    1.5197527715089005e+000,
   -4.5099012045118504e-003,
    1.0000000000000001e-005,
   -9.2998110632603356e-004,
    1.4641951872491479e-001,
    1.3794956422367508e-001,
   -5.5998711524236618e-004,
   -8.0000000000000007e-005,
    1.0269980564904076e-002,
    1.1859978046964528e-002,
   -6.0000000000000002e-005,
    7.2999831194654687e-004,
    9.9000003995807525e-004,
    5.0000000000000002e-005,
    8.0000000000000007e-005,
    3.0000000000000001e-005,
    2.7999032180005306e-004,
    5.0000000000000002e-005,
    2.4699041754397662e-003,
    3.3999032138302396e-004,
   -2.0000000000000002e-005,
    1.8179263590004310e-002,
    1.6199348116503337e-003,
   -1.4999999999999999e-004,
    7.9126692560293324e-002,
    5.4287721206722483e-002,
    1.6999613195309854e-004,
   -3.4197934884254719e-003,
   -7.9101527986961184e-001,
   -1.3178223087945187e+000,
    1.9198856695799263e-003,
    5.0000000000000002e-005,
   -5.4566580553490514e-002,
   -1.2094151282860921e+001,
   -1.2642739063331361e+000,
    1.7698886715359177e-003,
    2.0000000000000002e-005,
    4.6497160418536602e-003,
   -8.2271407280647579e-001,
   -2.3700982081636493e-001,
    5.1997171416615833e-004,
    2.0998999538487663e-004,
   -6.2827365240563343e-002,
   -3.1418700507997430e-002,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
   -4.9597975766249259e-003,
   -3.5498583509965191e-003,
    2.0000000000000002e-005,
   -3.9999032221708225e-004,
   -3.6999032180005308e-004,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -6.9999999999999994e-005,
   -1.0000000000000001e-005,
   -1.0199238228840819e-003,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.2619004970617135e-002,
   -5.0396057314673432e-003,
    4.0000000000000003e-005,
   -4.5995157936916826e-004,
   -1.0534151676616779e-001,
   -1.1329088055547008e-001,
    1.1999386468286557e-004,
   -1.0299014307268081e-003,
   -1.3413898665326490e-001,
   -1.4818829888765634e-002,
    3.0000000000000001e-005,
   -6.0000000000000002e-005,
   -6.1695030505202874e-003,
   -1.4498910050119490e-003,
   -2.4998245847139008e-004,
   -8.9993864682865569e-005,
   -5.9993864682865558e-005,
   -2.0000000000000002e-005,
   -8.9987729365731119e-005,
   -5.0000000000000002e-005,
    3.6995705236302973e-004,
    4.0000000000000003e-005,
    5.9993864682865558e-005,
   -3.9999999980881819e-005,
   -2.9999999990440909e-005,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -1.3998772930837655e-004,
   -1.5998772930837657e-004,
   -1.0099203648201035e-003,
   -6.2994791109766417e-004,
   -3.9999999980881819e-005,
   -1.3098902244592227e-003,
    7.9999999971322724e-005,
   -1.3998772930837655e-004,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.1999472902918624e-004,
   -2.8198811240780107e-003,
   -3.2898605775021415e-003,
   -8.9999999961763637e-005,
   -2.1039078775602640e-002,
   -1.3559414141111606e-002,
    2.0000000000000002e-005,
   -7.5994997622058126e-004,
   -2.5718835140798816e-002,
   -4.1999636551716586e-004,
    1.0000000000000001e-005,
   -2.7498838000645318e-003,
   -2.9998945601146298e-004,
   -2.6998859327590445e-004,
   -6.9999999971322711e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.6000086436543886e-004,
   -2.9999999990440909e-005,
   -2.8699906122380261e-003,
    1.4999999994264542e-004,
    1.0000000000000001e-005,
   -7.9999999971322724e-005,
   -3.9409787796614117e-002,
   -4.8519742795838566e-002,
    1.5999613064465650e-004,
   -1.2799669042687029e-003,
   -3.0516793001571274e-001,
   -2.0592872468806117e-001,
    3.7999226042657739e-004,
   -1.0089736878568208e-002,
   -3.5182706645119044e-001,
   -2.8399990057422268e-002,
    1.5999613064465650e-004,
    2.0000000000000002e-005,
   -3.6109762019961189e-002,
   -6.5799908815186497e-003,
    3.9999999980881819e-005,
   -3.2199841462111555e-003,
   -1.0799974442030674e-003,
   -2.7000172832340763e-004,
   -1.3999999995220451e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -6.9999999971322711e-005,
   -1.0000000000000001e-005,
   -4.3998332189762061e-004,
    3.9999999990440911e-005,
   -1.5199383564713349e-003,
   -2.6098863929636786e-003,
    2.9999999990440909e-005,
    1.3209424696369491e-002,
    2.0829098791581825e-002,
   -4.9999999980881818e-005,
    3.0997999065504410e-004,
    3.4358452062718502e-002,
    1.3509405503575257e-002,
   -2.9999999990440909e-005,
    4.2598195678548072e-003,
    2.5798946704204497e-003,
    4.3998332273167891e-004,
    3.4998945766046149e-004,
    3.9999999980881819e-005,
    3.9999999980881819e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -9.9993864644629200e-005,
   -5.5995318327176980e-004,
   -1.5998772930837657e-004,
   -8.9993864654188286e-005,
   -3.9999999980881819e-005,
   -1.0000000000000001e-005,
    1.0004406442870324e-005,
    1.1406618945528380e-008,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    6.9987729365731107e-005,
    6.9987729365731107e-005,
    7.9987729365731119e-005,
    3.9993864682865560e-005,
    1.0000000000000001e-005,
    1.2998772936573112e-004,
    5.8993036980803165e-004,
    1.6498030536431872e-003,
    1.3398437348439828e-003,
    1.0000000000000001e-005,
    1.6098065952749124e-003,
    7.9990722930178761e-004,
    2.3997545831443308e-004,
    1.6998159363156755e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -6.0000000000000002e-005,
    1.0999386468286557e-004,
   -2.0000000000000002e-005,
    2.8297768230701090e-003,
    1.2299032515509782e-002,
   -4.0000000000000003e-005,
   -5.0000000000000002e-005,
    3.4617218176446002e-002,
    2.3798095993400669e-002,
   -5.0000000000000002e-005,
    2.8996772013548436e-004,
    2.8987641176711373e-002,
    1.2698980216260428e-002,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
    4.1696670459316987e-003,
    2.5497996329908851e-003,
    4.6996104639370379e-004,
    3.6997718632596162e-004,
    5.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
   -9.0000000000000006e-005,
   -5.7998118160772314e-004,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9996590568122952e-004,
    1.9099256735431499e-003,
   -1.1000000000000000e-004,
   -3.2998085477551752e-004,
    5.2507790558040605e-002,
    2.1375102827470732e-001,
   -5.9996170955103499e-004,
   -6.4996084429418328e-004,
    5.9577403047192357e-001,
    3.3880542107838019e-001,
   -7.4995783941898776e-004,
   -1.0000000000000001e-005,
    5.3396643697027181e-003,
    4.1494151556630332e-001,
    1.5790314283804993e-001,
   -3.6997998993569493e-004,
   -6.1996084512824161e-004,
    5.6857563054754201e-002,
    3.0088730588867105e-002,
   -9.0000000000000006e-005,
   -1.0000000000000001e-004,
    6.1397442904511292e-003,
    4.2098290710028494e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    5.9997504587355950e-004,
    4.9998031676790056e-004,
    6.0000000000000002e-005,
    5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.5000172884558683e-004,
   -1.0000000000000001e-005,
   -1.1599978986716148e-003,
   -6.9999999999999994e-005,
   -6.7999851699930202e-003,
    2.9000172884558682e-004,
    3.0000000000000001e-005,
   -6.5998797966515966e-004,
   -7.7099705155976653e-003,
    2.1739908684458877e-002,
   -1.2699712256095371e-003,
   -3.8999114943329871e-003,
    6.3370660627406994e-001,
    2.4138773681885621e+000,
   -6.3198582256228071e-003,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -5.5398654660233524e-003,
    6.5795750374310860e+000,
    2.9984814622084768e+000,
   -6.6398461619936585e-003,
    2.0000000000000002e-005,
   -1.2999613070201104e-004,
    6.2568393649412357e-002,
    3.6744619522344233e+000,
    1.1918723424246769e+000,
   -2.8799333924651202e-003,
   -5.5198624690717107e-003,
    4.7337716127890950e-001,
    2.1258889463466799e-001,
   -6.7998711440830798e-004,
   -8.9997937664638831e-004,
    4.8339796122932280e-002,
    2.8279898471952968e-002,
   -1.2000000000000000e-004,
   -1.0000000000000001e-004,
    4.4899885073929387e-003,
    3.2299910602862142e-003,
   -2.0000000000000002e-005,
    4.0000259243432197e-004,
    3.4000259243432198e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
    6.2997418061670784e-004,
    1.9199256706239458e-003,
    3.0000000000000001e-005,
   -2.0299139688574517e-003,
   -2.9568742050766070e-002,
    9.0000000000000006e-005,
   -1.4999999958297084e-004,
   -1.7190245191781614e-001,
   -5.0967800232571393e-002,
    1.2000000000000000e-004,
   -3.0398121745785389e-003,
   -1.1307496152169749e-001,
   -2.5488897251826612e-002,
    6.9999999999999994e-005,
    1.1000000000000000e-004,
   -1.6919273441755853e-002,
   -5.5797675136468380e-003,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
   -1.9599178102144865e-003,
   -8.7996504042437784e-004,
   -1.9999472868862983e-004,
   -1.2000000000000000e-004,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.0000000000000001e-005,
   -1.9998859337149540e-004,
    1.0000000000000001e-005,
   -1.1399098217361094e-003,
   -2.7697829563020976e-003,
    5.0000000000000002e-005,
   -4.3796530447837269e-003,
   -1.7898599982578381e-003,
   -5.5995491107656932e-004,
   -2.4998245763733178e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    7.9987729365731119e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
   -3.9999999980881819e-005,
   -3.9999999980881819e-005,
   -9.9993864644629200e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.9999999990440911e-005,
   -1.3999999995220451e-004,
   -5.6997945255183615e-004,
   -6.0997331638152524e-004,
   -1.0000000000000001e-005,
   -1.2699418994401792e-003,
   -2.8998859243228707e-004,
   -4.9999999980881818e-005,
   -3.1998772799993457e-004,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -3.3000172830428945e-004,
   -1.2399944405216048e-003,
   -5.0499738511240021e-003,
   -5.1899747121532079e-003,
    2.0000000000000002e-005,
   -1.0999999996176364e-004,
   -1.0489932263047170e-002,
   -2.4499884236494698e-003,
    1.0000000000000001e-005,
   -4.0998839153605853e-004,
   -2.5399828263017159e-003,
   -2.7000259108764355e-004,
   -2.9999999990440909e-005,
   -4.4999785813400585e-004,
   -5.9999999563852654e-005,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
    1.3999999994264542e-004,
    4.0998945764134326e-004,
    9.6995717631544024e-004,
    5.2997858772157265e-004,
    3.0000000000000001e-005,
    5.0997772372536750e-004,
    2.2999386417980457e-004,
    9.9999999961763637e-005,
    4.9999999980881818e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
    5.9993864682865558e-005,
    9.9987729365731118e-005,
    5.9993864682865558e-005,
    5.0000000000000002e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    6.9999999999999994e-005,
    3.2997718674299077e-004,
    8.7993349941591222e-004,
    8.6993436342167648e-004,
    1.0000000000000001e-005,
    1.3698928165469999e-003,
    7.7993876989322403e-004,
    6.6994791050258322e-004,
    3.1997545748037483e-004,
    1.2998772936573112e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    8.8996590484717120e-004,
    4.1498316651052873e-003,
   -2.0000000000000002e-005,
    1.0909542893541542e-002,
    9.3896118576569775e-003,
   -3.0000000000000001e-005,
    1.3999999999999999e-004,
    1.4859362620021506e-002,
    7.5696827235034493e-003,
   -3.0000000000000001e-005,
    1.0000000000000001e-004,
    6.5197169059330613e-003,
    2.8798802536172223e-003,
   -1.0000000000000001e-005,
    1.2199488988240033e-003,
    6.2997418019967869e-004,
    1.6999386426583643e-004,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    1.4000086442279340e-004,
    7.1999658143284348e-004,
   -2.0000000000000002e-005,
    7.0199747919151498e-003,
    3.1179884630220266e-002,
   -1.3999999999999999e-004,
    6.9999999999999994e-005,
    8.0959591085278074e-002,
    5.9629735618255422e-002,
   -2.0999613028498191e-004,
    1.0699746420915477e-003,
    9.4029434978090118e-002,
    4.2169792332154843e-002,
   -1.3999999958297084e-004,
    5.9998452113992756e-004,
    3.6739771859050800e-002,
    1.4649926765324298e-002,
   -5.0000000000000002e-005,
   -5.0000000000000002e-005,
    6.5399648279462685e-003,
    3.0499880112826523e-003,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    8.7999571659302099e-004,
    4.7999872230227467e-004,
    1.0000000000000001e-004,
    6.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -5.3998031801898801e-004,
   -2.5598925309239574e-003,
   -1.3699444940205722e-003,
   -6.9999999999999994e-005,
   -3.4998471114150585e-003,
   -1.1599527689560507e-003,
   -4.0000000000000003e-005,
   -1.5299357633499118e-003,
   -4.6998332122606684e-004,
   -3.0998772853167283e-004,
   -1.1000000000000000e-004,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9993864682865569e-005,
   -6.0000000000000002e-005,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -3.9999999980881819e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -8.9999999971322723e-005,
   -1.0999999996176364e-004,
   -2.3999999907990986e-004,
   -9.9999999961763637e-005,
   -1.0000000000000001e-005,
   -1.2999999953517536e-004,
   -2.9999999990440909e-005,
   -2.9999999990440909e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    6.9999999999999994e-005,
    1.8999472868862983e-004,
    2.1999472827160065e-004,
    3.7998945696023052e-004,
    2.4999472827160070e-004,
    2.8998772811464368e-004,
    1.4999999916594166e-004,
    1.0999999958297084e-004,
    5.0000000000000002e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    4.2000259201729282e-004,
    1.0699974446045494e-003,
    1.0799974441875202e-003,
    2.0000000000000002e-005,
    1.8899914319349543e-003,
    1.1099965789306686e-003,
    2.0000000000000002e-005,
    1.2999918447928571e-003,
    5.9999872146821636e-004,
    4.5000086275467677e-004,
    1.9999999916594168e-004,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -3.0000000000000001e-005,
   -8.0000000000000007e-005,
   -4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<double, 918> g_MAIN_S2_DE406
{ {
    1.8461240380000967e+004,
   -6.2964776186139018e+000,
    5.9197463551783173e-003,
   -2.3599037212003134e-003,
    2.7986414518273106e+000,
    9.9969366315481682e+002,
    1.0101671561153941e+003,
   -1.0193852447752909e+000,
    1.5599368152025683e-003,
   -4.3998238235564216e-004,
   -1.3034701506935700e-001,
    3.1759670828363294e+001,
    6.1911950259640307e+001,
   -1.1786732060032996e-001,
    2.5998925652882658e-004,
    3.0000000000000001e-005,
   -1.4599682240705745e-003,
    1.5813035766718755e+000,
    3.9840544801889850e+000,
   -1.1809750513357475e-002,
    4.0000000000000003e-005,
   -6.0000000000000002e-005,
    9.1569768780338345e-002,
    2.6324939411177511e-001,
   -1.0899798302964248e-003,
    5.6599937867737541e-003,
    1.7679988271116290e-002,
   -1.0000000000000001e-004,
    3.6000345769117371e-004,
    1.1999996296023475e-003,
    2.0000000000000002e-005,
    8.0000864422793429e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -5.2998204644754570e-004,
   -1.4000086442279340e-004,
    3.0000000000000001e-005,
   -6.4597471744927944e-003,
   -1.8599291354046150e-003,
    2.2999085980767006e-004,
   -7.4786973296373349e-002,
   -2.3649035208117827e-002,
    2.0000000000000002e-005,
    1.5399112744515736e-003,
   -7.9318663729777539e-001,
   -3.0127728711866791e-001,
    1.9098821237779097e-003,
   -1.0000000000000001e-005,
    6.7895864296686854e-003,
   -6.7314362866293180e+000,
   -5.6323537756655035e+000,
   -9.1094394883815977e-003,
    1.0000519362625620e-004,
   -4.8396143488170438e+000,
   -6.4600721505279193e+000,
    1.1569282182361574e-002,
   -2.0000000000000002e-005,
   -7.0295716955030710e-003,
   -5.0759179342248535e+000,
   -5.3112784145560363e+000,
    6.6795917022311434e-003,
   -1.0000000000000001e-005,
    1.4099204142268746e-003,
   -3.1290679453673831e-001,
   -6.3881314193420591e-001,
    1.3499204150609330e-003,
    3.0000000000000001e-005,
   -2.4189016390142579e-002,
   -6.1757501429886931e-002,
    1.9999085980767006e-004,
   -1.9099291349875857e-003,
   -5.4797856752831116e-003,
    2.0000000000000002e-005,
   -1.5000086442279340e-004,
   -4.5998118244178145e-004,
   -1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -1.5998859378852455e-004,
   -4.0000000000000003e-005,
   -1.6898695941853505e-003,
   -4.8996191165055552e-004,
    3.0000000000000001e-005,
   -1.5708764255911896e-002,
   -5.3595781915790472e-003,
    4.0000000000000003e-005,
    9.9993864682865568e-005,
   -1.1334089519200992e-001,
   -9.5102356472747202e-002,
   -8.9993864682865569e-005,
   -1.3998772936573111e-004,
   -1.8008571587748830e-002,
   -5.7285376796267398e-002,
    1.5998386006774219e-004,
   -6.9999999999999994e-005,
   -6.1865021262157795e-002,
   -5.5035578507142892e-002,
    6.9999999999999994e-005,
    1.0000000000000001e-005,
   -2.8397768268233711e-003,
   -5.2795834649755633e-003,
    1.0000000000000001e-005,
   -1.5998772936573114e-004,
   -3.6997718757704907e-004,
   -2.0000000000000002e-005,
   -4.0000000000000003e-005,
   -1.0000000000000001e-005,
   -3.2996405210295761e-004,
   -1.0998772936573112e-004,
   -2.1597448309137817e-003,
   -1.8997785123746494e-003,
    6.5991863301108811e-004,
   -3.0996319143342663e-004,
   -1.0898774183899964e-003,
   -8.2990196007556318e-004,
   -4.0000000000000003e-005,
   -7.9993864682865570e-005,
   -3.9993864682865560e-005,
   -4.0000000000000003e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.9999999990440909e-005,
    1.0998772932749475e-004,
    1.1998772931793565e-004,
   -2.0000000000000002e-005,
    6.9999999971322711e-005,
   -2.9999999990440909e-005,
    6.3994791234875165e-004,
   -1.0000000000000001e-005,
    2.0398334043096023e-003,
    2.1398211344711548e-003,
   -5.9995404726797500e-004,
    4.9999999980881818e-005,
   -5.9999999971322719e-005,
   -2.9999999990440909e-005,
    2.9999999990440909e-005,
    1.7999472903874533e-004,
   -3.9999999990440911e-005,
    9.5995890683870267e-004,
   -1.3699436299252053e-003,
   -2.9999999990440909e-005,
    1.0309547898187386e-002,
    7.9995771355591250e-004,
   -5.1996998596344794e-004,
   -1.3459392383327749e-002,
   -1.8289191222779496e-002,
    5.9999999980881817e-005,
   -7.9999999971322724e-005,
   -2.0119122466600518e-002,
   -8.4496359940365653e-003,
    2.0000000000000002e-005,
   -1.7499283538553252e-003,
   -1.4599383573245114e-003,
   -1.4999999994264542e-004,
   -1.7999472903874533e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.1000086438455706e-004,
   -3.9999999990440911e-005,
   -1.2199961697937801e-003,
   -4.3000259310587563e-004,
   -1.2549955154032127e-002,
   -5.3699786296829404e-003,
    2.0000000000000002e-005,
   -2.4999226130843115e-004,
   -1.0963943437153734e-001,
   -7.8459564051565581e-002,
   -1.9999613062553833e-004,
   -3.6899093154652046e-003,
   -4.2988728107179575e-001,
    1.3927888402397687e-001,
    2.3399415459683273e-003,
    2.9999999990440909e-005,
   -3.2259168071847082e-002,
   -4.8057402037493979e+000,
   -5.3683961612559834e+000,
    6.4498353403772933e-003,
   -1.0000000000000001e-005,
    1.0000000000000001e-005,
   -1.8999613105212656e-004,
   -5.8892589129933504e-001,
   -6.6740564496553167e-001,
    1.4199677678096423e-003,
   -3.6359802389319490e-002,
   -6.3829692974737604e-002,
    2.0999613062553833e-004,
   -2.4699927536774319e-003,
   -5.5299794944654096e-003,
    2.0000000000000002e-005,
   -1.7000086435587977e-004,
   -4.5000259309631654e-004,
   -1.0000000000000001e-005,
   -3.9999999990440911e-005,
    1.0000000000000001e-005,
    1.6999472903874533e-004,
    6.9999999971322711e-005,
    1.8599256770251867e-003,
    9.2995977127105522e-004,
   -1.0000000000000001e-005,
    1.5969326570793512e-002,
    1.6799286013602480e-002,
    5.9999999971322719e-005,
    6.9999999971322711e-005,
    6.3497693454486765e-003,
   -5.5897523214725221e-002,
   -2.0998999530840389e-004,
    7.9994997662805132e-004,
    8.0422378961740737e-001,
    8.0259387492275991e-001,
   -8.2994610732050320e-004,
   -2.0998999572543304e-004,
    3.4648501149383414e-002,
    1.0175558259351709e-001,
   -1.9998999530840389e-004,
    1.0000000000000001e-005,
    3.1698667157862757e-003,
    1.0159573464138262e-002,
   -2.9999999990440909e-005,
    2.6998945810616790e-004,
    9.2995977127105522e-004,
    2.0000000000000002e-005,
    7.9999999971322724e-005,
    2.9999999990440909e-005,
    1.0000000000000001e-005,
    1.6998772930837657e-004,
    1.5998772930837657e-004,
   -7.9993864654188287e-005,
   -3.8996932243688307e-004,
   -1.6998618145833678e-003,
   -2.0398372731905618e-003,
   -2.1998159396256485e-004,
   -7.1993790728792190e-004,
   -2.9999999990440909e-005,
   -1.1999386463507010e-004,
   -1.0000000000000001e-005,
   -2.9999999990440909e-005,
   -2.9999999990440909e-005,
    1.0000000000000001e-005,
    1.0000000000000001e-005,
    3.9993864682865560e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0995705278005889e-004,
    3.3995091746292448e-004,
    1.1798166019751813e-003,
    6.1990409969390700e-004,
    7.9987729365731119e-005,
    9.9987729365731118e-005,
    1.0000000000000001e-005,
    5.9993864682865558e-005,
    7.6990895939846185e-004,
   -5.0000000000000002e-005,
    9.1389311067919324e-003,
    1.0418780115449460e-002,
   -2.0000000000000002e-005,
    7.9993864682865570e-005,
    3.6465675780352910e-002,
    1.6028102763916620e-002,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
    2.3997211540680375e-003,
    2.5596988771505289e-003,
    2.0997545873146224e-004,
    3.0996405210295761e-004,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    1.1998772936573112e-004,
    2.0000000000000002e-005,
    1.7798643241250674e-003,
    2.9997718757704904e-004,
    2.2848199449214169e-002,
   -1.9198494614735061e-003,
   -8.9993864682865569e-005,
    2.3997385545261880e-004,
    2.6862839528785964e-001,
    3.1471470039065103e-001,
   -4.4995771552036097e-004,
    2.7297278859196752e-003,
    1.0857809772361875e+000,
    3.8349873586206878e-001,
   -4.3995771510333184e-004,
   -7.1992629924458047e-004,
    6.9144454997641505e-002,
    5.8475324967037876e-002,
   -1.0999386468286557e-004,
   -3.0000000000000001e-005,
    5.5495667836994549e-003,
    6.7694683279979108e-003,
   -2.0000000000000002e-005,
    4.5996578094854450e-004,
    6.9994663530703279e-004,
    4.0000000000000003e-005,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    2.8999032263411138e-004,
    6.0000000000000002e-005,
    4.1598377235713664e-003,
    7.9996677135511047e-004,
   -4.0000000000000003e-005,
    5.5017765706985064e-002,
    9.3496192475408194e-003,
    6.9999999999999994e-005,
    3.2998085560957582e-004,
    6.5022258058474314e-001,
   -6.2077378299581383e-002,
   -2.6698408305388952e-003,
   -1.0000000000000001e-005,
    1.0339367321461741e-002,
    7.4345550611201423e+000,
    8.8681426530612608e+000,
   -1.1769285589814833e-002,
    2.0000000000000002e-005,
   -9.0000000000000006e-005,
    8.8144472523691508e-002,
    2.9576600674946288e+001,
    7.9585542335066863e+000,
   -9.4394221534410804e-003,
    2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -1.6688979891507517e-002,
    1.7659831748780288e+000,
    1.1346109846345061e+000,
   -2.2398673850249746e-003,
   -6.0996170996806417e-004,
    1.2896461667223055e-001,
    1.2386484072228615e-001,
   -3.5998171919831098e-004,
   -6.0000000000000002e-005,
    9.8496042942762367e-003,
    1.2109518146664864e-002,
   -5.0000000000000002e-005,
    7.5996677052105218e-004,
    1.1099562278983118e-003,
    6.0000000000000002e-005,
    1.0000000000000001e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    9.2000090479789809e-004,
    2.0000172884558686e-004,
   -3.0000000000000001e-005,
    1.1269992293506328e-002,
    2.2999988051376299e-003,
   -1.4000086442279340e-004,
    1.3380969738450180e-001,
    2.4959940627996887e-002,
   -2.0000000000000002e-005,
    2.2999613153606939e-004,
    1.5156341196347716e+000,
    2.5407897826836789e-001,
    2.3199493266433456e-003,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
    2.0449530411600761e-002,
    1.5566265411600057e+001,
   -1.6244211504417858e+000,
   -6.5608499340100476e-002,
    1.1999613070201107e-004,
   -4.5998065351005534e-004,
    3.2906197559064598e-001,
    1.6657411651352464e+002,
    1.9948376111198255e+002,
   -2.4483409732634209e-001,
    3.7998452280804429e-004,
   -2.2099032667162182e-003,
    2.1863132930572471e+000,
    6.2365248530516499e+002,
    1.1726069661889859e+002,
   -1.4452645931932776e-001,
    2.5998839210603319e-004,
   -7.4996991003888199e-004,
   -2.9115289842970743e-001,
    3.3357202973080071e+001,
    1.5121554333970961e+001,
   -3.0379291843547283e-002,
    6.9999999999999994e-005,
    8.0000000000000007e-005,
   -9.9497680019529668e-003,
    2.1461691617920584e+000,
    1.5197527715089005e+000,
   -4.5099012045118504e-003,
    1.0000000000000001e-005,
   -9.2998110632603356e-004,
    1.4641951872491479e-001,
    1.3794956422367508e-001,
   -5.5998711524236618e-004,
   -8.0000000000000007e-005,
    1.0269980564904076e-002,
    1.1859978046964528e-002,
   -6.0000000000000002e-005,
    7.2999831194654687e-004,
    9.9000003995807525e-004,
    5.0000000000000002e-005,
    8.0000000000000007e-005,
    3.0000000000000001e-005,
    2.7999032180005306e-004,
    5.0000000000000002e-005,
    2.4699041754397662e-003,
    3.3999032138302396e-004,
   -2.0000000000000002e-005,
    1.8179263590004310e-002,
    1.6199348116503337e-003,
   -1.4999999999999999e-004,
    7.9126692560293324e-002,
    5.4287721206722483e-002,
    1.6999613195309854e-004,
   -3.4197934884254719e-003,
   -7.9101527986961184e-001,
   -1.3178223087945187e+000,
    1.9198856695799263e-003,
    5.0000000000000002e-005,
   -5.4566580553490514e-002,
   -1.2094151282860921e+001,
   -1.2642739063331361e+000,
    1.7698886715359177e-003,
    2.0000000000000002e-005,
    4.6497160418536602e-003,
   -8.2271407280647579e-001,
   -2.3700982081636493e-001,
    5.1997171416615833e-004,
    2.0998999538487663e-004,
   -6.2827365240563343e-002,
   -3.1418700507997430e-002,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
   -4.9597975766249259e-003,
   -3.5498583509965191e-003,
    2.0000000000000002e-005,
   -3.9999032221708225e-004,
   -3.6999032180005308e-004,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -6.9999999999999994e-005,
   -1.0000000000000001e-005,
   -1.0199238228840819e-003,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
   -1.2619004970617135e-002,
   -5.0396057314673432e-003,
    4.0000000000000003e-005,
   -4.5995157936916826e-004,
   -1.0534151676616779e-001,
   -1.1329088055547008e-001,
    1.1999386468286557e-004,
   -1.0299014307268081e-003,
   -1.3413898665326490e-001,
   -1.4818829888765634e-002,
    3.0000000000000001e-005,
   -6.0000000000000002e-005,
   -6.1695030505202874e-003,
   -1.4498910050119490e-003,
   -2.4998245847139008e-004,
   -8.9993864682865569e-005,
   -5.9993864682865558e-005,
   -2.0000000000000002e-005,
   -8.9987729365731119e-005,
   -5.0000000000000002e-005,
    3.6995705236302973e-004,
    4.0000000000000003e-005,
    5.9993864682865558e-005,
   -3.9999999980881819e-005,
   -2.9999999990440909e-005,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -1.3998772930837655e-004,
   -1.5998772930837657e-004,
   -1.0099203648201035e-003,
   -6.2994791109766417e-004,
   -3.9999999980881819e-005,
   -1.3098902244592227e-003,
    7.9999999971322724e-005,
   -1.3998772930837655e-004,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -2.1999472902918624e-004,
   -2.8198811240780107e-003,
   -3.2898605775021415e-003,
   -8.9999999961763637e-005,
   -2.1039078775602640e-002,
   -1.3559414141111606e-002,
    2.0000000000000002e-005,
   -7.5994997622058126e-004,
   -2.5718835140798816e-002,
   -4.1999636551716586e-004,
    1.0000000000000001e-005,
   -2.7498838000645318e-003,
   -2.9998945601146298e-004,
   -2.6998859327590445e-004,
   -6.9999999971322711e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -1.6000086436543886e-004,
   -2.9999999990440909e-005,
   -2.8699906122380261e-003,
    1.4999999994264542e-004,
    1.0000000000000001e-005,
   -7.9999999971322724e-005,
   -3.9409787796614117e-002,
   -4.8519742795838566e-002,
    1.5999613064465650e-004,
   -1.2799669042687029e-003,
   -3.0516793001571274e-001,
   -2.0592872468806117e-001,
    3.7999226042657739e-004,
   -1.0089736878568208e-002,
   -3.5182706645119044e-001,
   -2.8399990057422268e-002,
    1.5999613064465650e-004,
    2.0000000000000002e-005,
   -3.6109762019961189e-002,
   -6.5799908815186497e-003,
    3.9999999980881819e-005,
   -3.2199841462111555e-003,
   -1.0799974442030674e-003,
   -2.7000172832340763e-004,
   -1.3999999995220451e-004,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -6.9999999971322711e-005,
   -1.0000000000000001e-005,
   -4.3998332189762061e-004,
    3.9999999990440911e-005,
   -1.5199383564713349e-003,
   -2.6098863929636786e-003,
    2.9999999990440909e-005,
    1.3209424696369491e-002,
    2.0829098791581825e-002,
   -4.9999999980881818e-005,
    3.0997999065504410e-004,
    3.4358452062718502e-002,
    1.3509405503575257e-002,
   -2.9999999990440909e-005,
    4.2598195678548072e-003,
    2.5798946704204497e-003,
    4.3998332273167891e-004,
    3.4998945766046149e-004,
    3.9999999980881819e-005,
    3.9999999980881819e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
   -9.9993864644629200e-005,
   -5.5995318327176980e-004,
   -1.5998772930837657e-004,
   -8.9993864654188286e-005,
   -3.9999999980881819e-005,
   -1.0000000000000001e-005,
    1.0004406442870324e-005,
    1.1406618945528380e-008,
   -1.0000000000000001e-005,
    3.0000000000000001e-005,
    6.9987729365731107e-005,
    6.9987729365731107e-005,
    7.9987729365731119e-005,
    3.9993864682865560e-005,
    1.0000000000000001e-005,
    1.2998772936573112e-004,
    5.8993036980803165e-004,
    1.6498030536431872e-003,
    1.3398437348439828e-003,
    1.0000000000000001e-005,
    1.6098065952749124e-003,
    7.9990722930178761e-004,
    2.3997545831443308e-004,
    1.6998159363156755e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -6.0000000000000002e-005,
    1.0999386468286557e-004,
   -2.0000000000000002e-005,
    2.8297768230701090e-003,
    1.2299032515509782e-002,
   -4.0000000000000003e-005,
   -5.0000000000000002e-005,
    3.4617218176446002e-002,
    2.3798095993400669e-002,
   -5.0000000000000002e-005,
    2.8996772013548436e-004,
    2.8987641176711373e-002,
    1.2698980216260428e-002,
   -3.0000000000000001e-005,
   -4.0000000000000003e-005,
    4.1696670459316987e-003,
    2.5497996329908851e-003,
    4.6996104639370379e-004,
    3.6997718632596162e-004,
    5.0000000000000002e-005,
    5.0000000000000002e-005,
   -1.0000000000000001e-005,
   -9.0000000000000006e-005,
   -5.7998118160772314e-004,
    2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9996590568122952e-004,
    1.9099256735431499e-003,
   -1.1000000000000000e-004,
   -3.2998085477551752e-004,
    5.2507790558040605e-002,
    2.1375102827470732e-001,
   -5.9996170955103499e-004,
   -6.4996084429418328e-004,
    5.9577403047192357e-001,
    3.3880542107838019e-001,
   -7.4995783941898776e-004,
   -1.0000000000000001e-005,
    5.3396643697027181e-003,
    4.1494151556630332e-001,
    1.5790314283804993e-001,
   -3.6997998993569493e-004,
   -6.1996084512824161e-004,
    5.6857563054754201e-002,
    3.0088730588867105e-002,
   -9.0000000000000006e-005,
   -1.0000000000000001e-004,
    6.1397442904511292e-003,
    4.2098290710028494e-003,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    5.9997504587355950e-004,
    4.9998031676790056e-004,
    6.0000000000000002e-005,
    5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.5000172884558683e-004,
   -1.0000000000000001e-005,
   -1.1599978986716148e-003,
   -6.9999999999999994e-005,
   -6.7999851699930202e-003,
    2.9000172884558682e-004,
    3.0000000000000001e-005,
   -6.5998797966515966e-004,
   -7.7099705155976653e-003,
    2.1739908684458877e-002,
   -1.2699712256095371e-003,
   -3.8999114943329871e-003,
    6.3370660627406994e-001,
    2.4138773681885621e+000,
   -6.3198582256228071e-003,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -5.5398654660233524e-003,
    6.5795750374310860e+000,
    2.9984814622084768e+000,
   -6.6398461619936585e-003,
    2.0000000000000002e-005,
   -1.2999613070201104e-004,
    6.2568393649412357e-002,
    3.6744619522344233e+000,
    1.1918723424246769e+000,
   -2.8799333924651202e-003,
   -5.5198624690717107e-003,
    4.7337716127890950e-001,
    2.1258889463466799e-001,
   -6.7998711440830798e-004,
   -8.9997937664638831e-004,
    4.8339796122932280e-002,
    2.8279898471952968e-002,
   -1.2000000000000000e-004,
   -1.0000000000000001e-004,
    4.4899885073929387e-003,
    3.2299910602862142e-003,
   -2.0000000000000002e-005,
    4.0000259243432197e-004,
    3.4000259243432198e-004,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
    6.2997418061670784e-004,
    1.9199256706239458e-003,
    3.0000000000000001e-005,
   -2.0299139688574517e-003,
   -2.9568742050766070e-002,
    9.0000000000000006e-005,
   -1.4999999958297084e-004,
   -1.7190245191781614e-001,
   -5.0967800232571393e-002,
    1.2000000000000000e-004,
   -3.0398121745785389e-003,
   -1.1307496152169749e-001,
   -2.5488897251826612e-002,
    6.9999999999999994e-005,
    1.1000000000000000e-004,
   -1.6919273441755853e-002,
   -5.5797675136468380e-003,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
   -1.9599178102144865e-003,
   -8.7996504042437784e-004,
   -1.9999472868862983e-004,
   -1.2000000000000000e-004,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -3.0000000000000001e-005,
   -1.9998859337149540e-004,
    1.0000000000000001e-005,
   -1.1399098217361094e-003,
   -2.7697829563020976e-003,
    5.0000000000000002e-005,
   -4.3796530447837269e-003,
   -1.7898599982578381e-003,
   -5.5995491107656932e-004,
   -2.4998245763733178e-004,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    2.0000000000000002e-005,
    7.9987729365731119e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
   -3.9999999980881819e-005,
   -3.9999999980881819e-005,
   -9.9993864644629200e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.9999999990440911e-005,
   -1.3999999995220451e-004,
   -5.6997945255183615e-004,
   -6.0997331638152524e-004,
   -1.0000000000000001e-005,
   -1.2699418994401792e-003,
   -2.8998859243228707e-004,
   -4.9999999980881818e-005,
   -3.1998772799993457e-004,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -3.3000172830428945e-004,
   -1.2399944405216048e-003,
   -5.0499738511240021e-003,
   -5.1899747121532079e-003,
    2.0000000000000002e-005,
   -1.0999999996176364e-004,
   -1.0489932263047170e-002,
   -2.4499884236494698e-003,
    1.0000000000000001e-005,
   -4.0998839153605853e-004,
   -2.5399828263017159e-003,
   -2.7000259108764355e-004,
   -2.9999999990440909e-005,
   -4.4999785813400585e-004,
   -5.9999999563852654e-005,
   -5.9999999980881817e-005,
   -1.0000000000000001e-005,
   -1.0000000000000001e-005,
   -4.9999999980881818e-005,
    1.3999999994264542e-004,
    4.0998945764134326e-004,
    9.6995717631544024e-004,
    5.2997858772157265e-004,
    3.0000000000000001e-005,
    5.0997772372536750e-004,
    2.2999386417980457e-004,
    9.9999999961763637e-005,
    4.9999999980881818e-005,
    1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
    2.0000000000000002e-005,
    5.0000000000000002e-005,
    5.9993864682865558e-005,
    9.9987729365731118e-005,
    5.9993864682865558e-005,
    5.0000000000000002e-005,
    3.0000000000000001e-005,
    1.0000000000000001e-005,
    6.9999999999999994e-005,
    3.2997718674299077e-004,
    8.7993349941591222e-004,
    8.6993436342167648e-004,
    1.0000000000000001e-005,
    1.3698928165469999e-003,
    7.7993876989322403e-004,
    6.6994791050258322e-004,
    3.1997545748037483e-004,
    1.2998772936573112e-004,
    6.9999999999999994e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    8.8996590484717120e-004,
    4.1498316651052873e-003,
   -2.0000000000000002e-005,
    1.0909542893541542e-002,
    9.3896118576569775e-003,
   -3.0000000000000001e-005,
    1.3999999999999999e-004,
    1.4859362620021506e-002,
    7.5696827235034493e-003,
   -3.0000000000000001e-005,
    1.0000000000000001e-004,
    6.5197169059330613e-003,
    2.8798802536172223e-003,
   -1.0000000000000001e-005,
    1.2199488988240033e-003,
    6.2997418019967869e-004,
    1.6999386426583643e-004,
    1.0000000000000001e-004,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    1.4000086442279340e-004,
    7.1999658143284348e-004,
   -2.0000000000000002e-005,
    7.0199747919151498e-003,
    3.1179884630220266e-002,
   -1.3999999999999999e-004,
    6.9999999999999994e-005,
    8.0959591085278074e-002,
    5.9629735618255422e-002,
   -2.0999613028498191e-004,
    1.0699746420915477e-003,
    9.4029434978090118e-002,
    4.2169792332154843e-002,
   -1.3999999958297084e-004,
    5.9998452113992756e-004,
    3.6739771859050800e-002,
    1.4649926765324298e-002,
   -5.0000000000000002e-005,
   -5.0000000000000002e-005,
    6.5399648279462685e-003,
    3.0499880112826523e-003,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
    8.7999571659302099e-004,
    4.7999872230227467e-004,
    1.0000000000000001e-004,
    6.0000000000000002e-005,
    1.0000000000000001e-005,
    3.0000000000000001e-005,
   -3.0000000000000001e-005,
   -5.3998031801898801e-004,
   -2.5598925309239574e-003,
   -1.3699444940205722e-003,
   -6.9999999999999994e-005,
   -3.4998471114150585e-003,
   -1.1599527689560507e-003,
   -4.0000000000000003e-005,
   -1.5299357633499118e-003,
   -4.6998332122606684e-004,
   -3.0998772853167283e-004,
   -1.1000000000000000e-004,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -5.0000000000000002e-005,
   -8.9993864682865569e-005,
   -6.0000000000000002e-005,
   -5.0000000000000002e-005,
   -2.0000000000000002e-005,
   -1.0000000000000001e-005,
   -2.0000000000000002e-005,
   -3.9999999980881819e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
   -3.0000000000000001e-005,
   -8.9999999971322723e-005,
   -1.0999999996176364e-004,
   -2.3999999907990986e-004,
   -9.9999999961763637e-005,
   -1.0000000000000001e-005,
   -1.2999999953517536e-004,
   -2.9999999990440909e-005,
   -2.9999999990440909e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005,
    4.0000000000000003e-005,
    3.0000000000000001e-005,
    4.0000000000000003e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    6.9999999999999994e-005,
    1.8999472868862983e-004,
    2.1999472827160065e-004,
    3.7998945696023052e-004,
    2.4999472827160070e-004,
    2.8998772811464368e-004,
    1.4999999916594166e-004,
    1.0999999958297084e-004,
    5.0000000000000002e-005,
    2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    9.0000000000000006e-005,
    4.2000259201729282e-004,
    1.0699974446045494e-003,
    1.0799974441875202e-003,
    2.0000000000000002e-005,
    1.8899914319349543e-003,
    1.1099965789306686e-003,
    2.0000000000000002e-005,
    1.2999918447928571e-003,
    5.9999872146821636e-004,
    4.5000086275467677e-004,
    1.9999999916594168e-004,
    9.0000000000000006e-005,
    4.0000000000000003e-005,
    1.0000000000000001e-005,
   -4.0000000000000003e-005,
   -3.0000000000000001e-005,
   -8.0000000000000007e-005,
   -4.0000000000000003e-005,
   -6.0000000000000002e-005,
   -2.0000000000000002e-005,
   -2.0000000000000002e-005,
    1.0000000000000001e-005,
    2.0000000000000002e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    3.0000000000000001e-005,
    2.0000000000000002e-005,
    2.0000000000000002e-005
} };

constexpr array<ELPMPP02MainProblemNominalCoefficient, 704> g_MAIN_S3
{ {
 {  0,  0,  0,  0 },
 {  0,  2,  0,  0 },
 {  0,  4,  0,  0 },
 {  0, -4,  1,  0 },
 {  0, -2,  1,  0 },
 {  0,  0,  1,  0 },
 {  0,  2,  1,  0 },
 {  0, -4,  2,  0 },
 {  0, -2,  2,  0 },
 {  0,  0,  2,  0 },
 {  0,  2,  2,  0 },
 {  0, -4,  3,  0 },
 {  0, -2,  3,  0 },
 {  0,  0,  3,  0 },
 {  0,  2,  3,  0 },
 {  0, -4,  4,  0 },
 {  0, -2,  4,  0 },
 {  0,  0,  4,  0 },
 {  0,  2,  4,  0 },
 {  0, -2,  5,  0 },
 {  0,  0,  5,  0 },
 {  0,  0,  6,  0 },
 {  0,  0,  7,  0 },
 {  0,  0, -6,  1 },
 {  0,  0, -5,  1 },
 {  0,  0, -4,  1 },
 {  0,  0, -3,  1 },
 {  0,  2, -3,  1 },
 {  0, -2, -2,  1 },
 {  0,  0, -2,  1 },
 {  0,  2, -2,  1 },
 {  0,  4, -2,  1 },
 {  0, -2, -1,  1 },
 {  0,  0, -1,  1 },
 {  0,  2, -1,  1 },
 {  0,  4, -1,  1 },
 {  0, -2,  0,  1 },
 {  0,  0,  0,  1 },
 {  0,  2,  0,  1 },
 {  0, -4,  1,  1 },
 {  0, -2,  1,  1 },
 {  0,  0,  1,  1 },
 {  0,  2,  1,  1 },
 {  0, -4,  2,  1 },
 {  0, -2,  2,  1 },
 {  0,  0,  2,  1 },
 {  0,  2,  2,  1 },
 {  0, -2,  3,  1 },
 {  0,  0,  3,  1 },
 {  0, -2,  4,  1 },
 {  0,  0,  4,  1 },
 {  0,  0,  5,  1 },
 {  0,  0,  6,  1 },
 {  0,  0, -5,  2 },
 {  0,  0, -4,  2 },
 {  0,  0, -3,  2 },
 {  0,  0, -2,  2 },
 {  0,  2, -2,  2 },
 {  0, -2, -1,  2 },
 {  0,  0, -1,  2 },
 {  0,  2, -1,  2 },
 {  0, -2,  0,  2 },
 {  0,  0,  0,  2 },
 {  0,  2,  0,  2 },
 {  0, -2,  1,  2 },
 {  0,  0,  1,  2 },
 {  0,  2,  1,  2 },
 {  0, -2,  2,  2 },
 {  0,  0,  2,  2 },
 {  0,  0,  3,  2 },
 {  0,  0,  4,  2 },
 {  0,  0, -3,  3 },
 {  0,  0, -2,  3 },
 {  0,  2, -2,  3 },
 {  0,  0, -1,  3 },
 {  0,  2, -1,  3 },
 {  0, -2,  0,  3 },
 {  0,  0,  0,  3 },
 {  0,  2,  0,  3 },
 {  0, -2,  1,  3 },
 {  0,  0,  1,  3 },
 {  0,  0,  2,  3 },
 {  0,  0,  3,  3 },
 {  0,  0, -2,  4 },
 {  0,  0, -1,  4 },
 {  0,  0,  0,  4 },
 {  0,  0,  1,  4 },
 {  1,  0,  0, -4 },
 {  1,  0, -1, -3 },
 {  1,  0,  0, -3 },
 {  1,  0, -3, -2 },
 {  1,  0, -2, -2 },
 {  1,  0, -1, -2 },
 {  1, -2,  0, -2 },
 {  1,  0,  0, -2 },
 {  1,  2,  0, -2 },
 {  1, -2,  1, -2 },
 {  1,  0,  1, -2 },
 {  1,  0,  2, -2 },
 {  1,  0,  3, -2 },
 {  1,  0, -4, -1 },
 {  1,  0, -3, -1 },
 {  1, -2, -2, -1 },
 {  1,  0, -2, -1 },
 {  1,  2, -2, -1 },
 {  1, -2, -1, -1 },
 {  1,  0, -1, -1 },
 {  1,  2, -1, -1 },
 {  1, -2,  0, -1 },
 {  1,  0,  0, -1 },
 {  1,  2,  0, -1 },
 {  1, -2,  1, -1 },
 {  1,  0,  1, -1 },
 {  1,  2,  1, -1 },
 {  1, -2,  2, -1 },
 {  1,  0,  2, -1 },
 {  1,  0,  3, -1 },
 {  1,  0,  4, -1 },
 {  1,  0, -6,  0 },
 {  1,  0, -5,  0 },
 {  1,  0, -4,  0 },
 {  1, -2, -3,  0 },
 {  1,  0, -3,  0 },
 {  1,  2, -3,  0 },
 {  1, -2, -2,  0 },
 {  1,  0, -2,  0 },
 {  1,  2, -2,  0 },
 {  1, -2, -1,  0 },
 {  1,  0, -1,  0 },
 {  1,  2, -1,  0 },
 {  1, -2,  0,  0 },
 {  1,  0,  0,  0 },
 {  1,  2,  0,  0 },
 {  1, -4,  1,  0 },
 {  1, -2,  1,  0 },
 {  1,  0,  1,  0 },
 {  1,  2,  1,  0 },
 {  1, -2,  2,  0 },
 {  1,  0,  2,  0 },
 {  1,  2,  2,  0 },
 {  1, -2,  3,  0 },
 {  1,  0,  3,  0 },
 {  1,  0,  4,  0 },
 {  1,  0,  5,  0 },
 {  1,  0, -5,  1 },
 {  1,  0, -4,  1 },
 {  1,  0, -3,  1 },
 {  1,  2, -3,  1 },
 {  1, -2, -2,  1 },
 {  1,  0, -2,  1 },
 {  1,  2, -2,  1 },
 {  1, -2, -1,  1 },
 {  1,  0, -1,  1 },
 {  1,  2, -1,  1 },
 {  1, -2,  0,  1 },
 {  1,  0,  0,  1 },
 {  1,  2,  0,  1 },
 {  1, -2,  1,  1 },
 {  1,  0,  1,  1 },
 {  1, -2,  2,  1 },
 {  1,  0,  2,  1 },
 {  1, -2,  3,  1 },
 {  1,  0,  3,  1 },
 {  1,  0,  4,  1 },
 {  1,  0, -4,  2 },
 {  1,  0, -3,  2 },
 {  1,  0, -2,  2 },
 {  1,  0, -1,  2 },
 {  1, -2,  0,  2 },
 {  1,  0,  0,  2 },
 {  1, -2,  1,  2 },
 {  1,  0,  1,  2 },
 {  1,  0,  2,  2 },
 {  1,  0,  3,  2 },
 {  1,  0, -2,  3 },
 {  1,  0, -1,  3 },
 {  1,  0,  0,  3 },
 {  1,  0,  1,  3 },
 {  2,  0,  0, -6 },
 {  2,  0, -1, -5 },
 {  2,  0,  0, -5 },
 {  2,  0,  1, -5 },
 {  2,  0, -2, -4 },
 {  2,  0, -1, -4 },
 {  2, -2,  0, -4 },
 {  2,  0,  0, -4 },
 {  2,  0,  1, -4 },
 {  2,  0,  2, -4 },
 {  2,  0, -4, -3 },
 {  2,  0, -3, -3 },
 {  2,  0, -2, -3 },
 {  2,  2, -2, -3 },
 {  2, -2, -1, -3 },
 {  2,  0, -1, -3 },
 {  2,  2, -1, -3 },
 {  2, -2,  0, -3 },
 {  2,  0,  0, -3 },
 {  2, -2,  1, -3 },
 {  2,  0,  1, -3 },
 {  2, -2,  2, -3 },
 {  2,  0,  2, -3 },
 {  2,  0,  3, -3 },
 {  2,  0, -5, -2 },
 {  2,  0, -4, -2 },
 {  2, -2, -3, -2 },
 {  2,  0, -3, -2 },
 {  2, -2, -2, -2 },
 {  2,  0, -2, -2 },
 {  2,  2, -2, -2 },
 {  2, -2, -1, -2 },
 {  2,  0, -1, -2 },
 {  2,  2, -1, -2 },
 {  2, -2,  0, -2 },
 {  2,  0,  0, -2 },
 {  2, -4,  1, -2 },
 {  2, -2,  1, -2 },
 {  2,  0,  1, -2 },
 {  2, -2,  2, -2 },
 {  2,  0,  2, -2 },
 {  2, -2,  3, -2 },
 {  2,  0,  3, -2 },
 {  2,  0,  4, -2 },
 {  2,  0,  5, -2 },
 {  2,  0, -6, -1 },
 {  2,  0, -5, -1 },
 {  2, -2, -4, -1 },
 {  2,  0, -4, -1 },
 {  2, -2, -3, -1 },
 {  2,  0, -3, -1 },
 {  2,  2, -3, -1 },
 {  2, -2, -2, -1 },
 {  2,  0, -2, -1 },
 {  2,  2, -2, -1 },
 {  2, -2, -1, -1 },
 {  2,  0, -1, -1 },
 {  2,  2, -1, -1 },
 {  2, -4,  0, -1 },
 {  2, -2,  0, -1 },
 {  2,  0,  0, -1 },
 {  2,  2,  0, -1 },
 {  2, -4,  1, -1 },
 {  2, -2,  1, -1 },
 {  2,  0,  1, -1 },
 {  2,  2,  1, -1 },
 {  2, -4,  2, -1 },
 {  2, -2,  2, -1 },
 {  2,  0,  2, -1 },
 {  2, -2,  3, -1 },
 {  2,  0,  3, -1 },
 {  2, -2,  4, -1 },
 {  2,  0,  4, -1 },
 {  2,  0,  5, -1 },
 {  2,  0,  6, -1 },
 {  2,  0, -7,  0 },
 {  2,  0, -6,  0 },
 {  2, -2, -5,  0 },
 {  2,  0, -5,  0 },
 {  2,  2, -5,  0 },
 {  2, -2, -4,  0 },
 {  2,  0, -4,  0 },
 {  2,  2, -4,  0 },
 {  2, -2, -3,  0 },
 {  2,  0, -3,  0 },
 {  2,  2, -3,  0 },
 {  2,  4, -3,  0 },
 {  2, -4, -2,  0 },
 {  2, -2, -2,  0 },
 {  2,  0, -2,  0 },
 {  2,  2, -2,  0 },
 {  2,  4, -2,  0 },
 {  2, -4, -1,  0 },
 {  2, -2, -1,  0 },
 {  2,  0, -1,  0 },
 {  2,  2, -1,  0 },
 {  2, -4,  0,  0 },
 {  2, -2,  0,  0 },
 {  2,  0,  0,  0 },
 {  2,  2,  0,  0 },
 {  2, -4,  1,  0 },
 {  2, -2,  1,  0 },
 {  2,  0,  1,  0 },
 {  2,  2,  1,  0 },
 {  2, -4,  2,  0 },
 {  2, -2,  2,  0 },
 {  2,  0,  2,  0 },
 {  2,  2,  2,  0 },
 {  2, -4,  3,  0 },
 {  2, -2,  3,  0 },
 {  2,  0,  3,  0 },
 {  2, -2,  4,  0 },
 {  2,  0,  4,  0 },
 {  2, -2,  5,  0 },
 {  2,  0,  5,  0 },
 {  2,  0,  6,  0 },
 {  2,  0, -6,  1 },
 {  2,  0, -5,  1 },
 {  2, -2, -4,  1 },
 {  2,  0, -4,  1 },
 {  2, -2, -3,  1 },
 {  2,  0, -3,  1 },
 {  2,  2, -3,  1 },
 {  2, -2, -2,  1 },
 {  2,  0, -2,  1 },
 {  2,  2, -2,  1 },
 {  2, -2, -1,  1 },
 {  2,  0, -1,  1 },
 {  2,  2, -1,  1 },
 {  2, -4,  0,  1 },
 {  2, -2,  0,  1 },
 {  2,  0,  0,  1 },
 {  2,  2,  0,  1 },
 {  2, -4,  1,  1 },
 {  2, -2,  1,  1 },
 {  2,  0,  1,  1 },
 {  2,  2,  1,  1 },
 {  2, -4,  2,  1 },
 {  2, -2,  2,  1 },
 {  2,  0,  2,  1 },
 {  2, -2,  3,  1 },
 {  2,  0,  3,  1 },
 {  2, -2,  4,  1 },
 {  2,  0,  4,  1 },
 {  2,  0,  5,  1 },
 {  2,  0, -4,  2 },
 {  2, -2, -3,  2 },
 {  2,  0, -3,  2 },
 {  2, -2, -2,  2 },
 {  2,  0, -2,  2 },
 {  2,  2, -2,  2 },
 {  2, -2, -1,  2 },
 {  2,  0, -1,  2 },
 {  2,  2, -1,  2 },
 {  2, -2,  0,  2 },
 {  2,  0,  0,  2 },
 {  2, -2,  1,  2 },
 {  2,  0,  1,  2 },
 {  2, -2,  2,  2 },
 {  2,  0,  2,  2 },
 {  2,  0, -3,  3 },
 {  2,  0, -2,  3 },
 {  2, -2, -1,  3 },
 {  2,  0, -1,  3 },
 {  2,  0,  0,  3 },
 {  2,  0, -1,  4 },
 {  3,  0, -1, -4 },
 {  3,  0,  0, -4 },
 {  3,  0, -2, -3 },
 {  3,  0, -1, -3 },
 {  3, -2,  0, -3 },
 {  3,  0,  0, -3 },
 {  3,  0,  1, -3 },
 {  3,  0, -3, -2 },
 {  3,  0, -2, -2 },
 {  3, -2, -1, -2 },
 {  3,  0, -1, -2 },
 {  3, -2,  0, -2 },
 {  3,  0,  0, -2 },
 {  3, -2,  1, -2 },
 {  3,  0,  1, -2 },
 {  3,  0,  2, -2 },
 {  3,  0, -4, -1 },
 {  3,  0, -3, -1 },
 {  3, -2, -2, -1 },
 {  3,  0, -2, -1 },
 {  3,  2, -2, -1 },
 {  3, -2, -1, -1 },
 {  3,  0, -1, -1 },
 {  3, -2,  0, -1 },
 {  3,  0,  0, -1 },
 {  3, -2,  1, -1 },
 {  3,  0,  1, -1 },
 {  3, -2,  2, -1 },
 {  3,  0,  2, -1 },
 {  3,  0, -5,  0 },
 {  3,  0, -4,  0 },
 {  3, -2, -3,  0 },
 {  3,  0, -3,  0 },
 {  3,  2, -3,  0 },
 {  3, -2, -2,  0 },
 {  3,  0, -2,  0 },
 {  3,  2, -2,  0 },
 {  3, -2, -1,  0 },
 {  3,  0, -1,  0 },
 {  3,  2, -1,  0 },
 {  3, -4,  0,  0 },
 {  3, -2,  0,  0 },
 {  3,  0,  0,  0 },
 {  3,  2,  0,  0 },
 {  3, -4,  1,  0 },
 {  3, -2,  1,  0 },
 {  3,  0,  1,  0 },
 {  3,  2,  1,  0 },
 {  3, -2,  2,  0 },
 {  3,  0,  2,  0 },
 {  3,  0,  3,  0 },
 {  3,  0,  4,  0 },
 {  3,  0, -5,  1 },
 {  3,  0, -4,  1 },
 {  3,  0, -3,  1 },
 {  3,  2, -3,  1 },
 {  3, -2, -2,  1 },
 {  3,  0, -2,  1 },
 {  3,  2, -2,  1 },
 {  3, -2, -1,  1 },
 {  3,  0, -1,  1 },
 {  3,  2, -1,  1 },
 {  3, -2,  0,  1 },
 {  3,  0,  0,  1 },
 {  3, -2,  1,  1 },
 {  3,  0,  1,  1 },
 {  3,  0,  2,  1 },
 {  3,  0,  3,  1 },
 {  3,  0, -3,  2 },
 {  3,  0, -2,  2 },
 {  3,  0, -1,  2 },
 {  3, -2,  0,  2 },
 {  3,  0,  0,  2 },
 {  3,  0,  1,  2 },
 {  3,  0,  2,  2 },
 {  3, -2, -2,  3 },
 {  3,  0, -2,  3 },
 {  3,  0, -1,  3 },
 {  3, -2,  0,  3 },
 {  4,  0, -1, -5 },
 {  4,  0,  0, -5 },
 {  4,  0, -2, -4 },
 {  4,  0, -1, -4 },
 {  4,  0,  0, -4 },
 {  4,  0,  1, -4 },
 {  4,  0, -3, -3 },
 {  4,  0, -2, -3 },
 {  4, -2, -1, -3 },
 {  4,  0, -1, -3 },
 {  4, -2,  0, -3 },
 {  4,  0,  0, -3 },
 {  4,  0,  1, -3 },
 {  4,  0,  2, -3 },
 {  4,  0, -5, -2 },
 {  4,  0, -4, -2 },
 {  4, -2, -3, -2 },
 {  4,  0, -3, -2 },
 {  4,  2, -3, -2 },
 {  4, -2, -2, -2 },
 {  4,  0, -2, -2 },
 {  4,  2, -2, -2 },
 {  4, -2, -1, -2 },
 {  4,  0, -1, -2 },
 {  4,  2, -1, -2 },
 {  4, -2,  0, -2 },
 {  4,  0,  0, -2 },
 {  4, -2,  1, -2 },
 {  4,  0,  1, -2 },
 {  4, -2,  2, -2 },
 {  4,  0,  2, -2 },
 {  4,  0,  3, -2 },
 {  4,  0, -6, -1 },
 {  4,  0, -5, -1 },
 {  4, -2, -4, -1 },
 {  4,  0, -4, -1 },
 {  4, -2, -3, -1 },
 {  4,  0, -3, -1 },
 {  4,  2, -3, -1 },
 {  4, -2, -2, -1 },
 {  4,  0, -2, -1 },
 {  4,  2, -2, -1 },
 {  4, -4, -1, -1 },
 {  4, -2, -1, -1 },
 {  4,  0, -1, -1 },
 {  4,  2, -1, -1 },
 {  4, -4,  0, -1 },
 {  4, -2,  0, -1 },
 {  4,  0,  0, -1 },
 {  4,  2,  0, -1 },
 {  4, -4,  1, -1 },
 {  4, -2,  1, -1 },
 {  4,  0,  1, -1 },
 {  4, -2,  2, -1 },
 {  4,  0,  2, -1 },
 {  4, -2,  3, -1 },
 {  4,  0,  3, -1 },
 {  4,  0,  4, -1 },
 {  4,  0, -7,  0 },
 {  4,  0, -6,  0 },
 {  4, -2, -5,  0 },
 {  4,  0, -5,  0 },
 {  4, -2, -4,  0 },
 {  4,  0, -4,  0 },
 {  4,  2, -4,  0 },
 {  4, -2, -3,  0 },
 {  4,  0, -3,  0 },
 {  4,  2, -3,  0 },
 {  4, -4, -2,  0 },
 {  4, -2, -2,  0 },
 {  4,  0, -2,  0 },
 {  4,  2, -2,  0 },
 {  4, -4, -1,  0 },
 {  4, -2, -1,  0 },
 {  4,  0, -1,  0 },
 {  4,  2, -1,  0 },
 {  4, -4,  0,  0 },
 {  4, -2,  0,  0 },
 {  4,  0,  0,  0 },
 {  4,  2,  0,  0 },
 {  4, -4,  1,  0 },
 {  4, -2,  1,  0 },
 {  4,  0,  1,  0 },
 {  4,  2,  1,  0 },
 {  4, -2,  2,  0 },
 {  4,  0,  2,  0 },
 {  4, -2,  3,  0 },
 {  4,  0,  3,  0 },
 {  4, -2,  4,  0 },
 {  4,  0,  4,  0 },
 {  4,  0,  5,  0 },
 {  4,  0, -5,  1 },
 {  4,  0, -4,  1 },
 {  4,  0, -3,  1 },
 {  4,  2, -3,  1 },
 {  4, -2, -2,  1 },
 {  4,  0, -2,  1 },
 {  4,  2, -2,  1 },
 {  4, -2, -1,  1 },
 {  4,  0, -1,  1 },
 {  4,  2, -1,  1 },
 {  4, -2,  0,  1 },
 {  4,  0,  0,  1 },
 {  4,  2,  0,  1 },
 {  4, -4,  1,  1 },
 {  4, -2,  1,  1 },
 {  4,  0,  1,  1 },
 {  4, -2,  2,  1 },
 {  4,  0,  2,  1 },
 {  4,  0,  3,  1 },
 {  4,  0,  4,  1 },
 {  4,  0, -4,  2 },
 {  4, -2, -3,  2 },
 {  4,  0, -3,  2 },
 {  4, -2, -2,  2 },
 {  4,  0, -2,  2 },
 {  4, -2, -1,  2 },
 {  4,  0, -1,  2 },
 {  4, -2,  0,  2 },
 {  4,  0,  0,  2 },
 {  4,  0,  1,  2 },
 {  4,  0, -2,  3 },
 {  4,  0, -1,  3 },
 {  4,  0,  0,  3 },
 {  5,  0, -2, -3 },
 {  5,  0, -3, -2 },
 {  5,  0, -2, -2 },
 {  5, -2, -1, -2 },
 {  5,  0, -1, -2 },
 {  5,  0,  0, -2 },
 {  5,  0,  1, -2 },
 {  5,  0, -4, -1 },
 {  5,  0, -3, -1 },
 {  5, -2, -2, -1 },
 {  5,  0, -2, -1 },
 {  5, -2, -1, -1 },
 {  5,  0, -1, -1 },
 {  5, -2,  0, -1 },
 {  5,  0,  0, -1 },
 {  5, -2,  1, -1 },
 {  5,  0,  1, -1 },
 {  5,  0, -4,  0 },
 {  5, -2, -3,  0 },
 {  5,  0, -3,  0 },
 {  5,  2, -3,  0 },
 {  5, -2, -2,  0 },
 {  5,  0, -2,  0 },
 {  5, -2, -1,  0 },
 {  5,  0, -1,  0 },
 {  5, -2,  0,  0 },
 {  5,  0,  0,  0 },
 {  5, -2,  1,  0 },
 {  5,  0,  1,  0 },
 {  5,  0,  2,  0 },
 {  5,  0, -4,  1 },
 {  5,  0, -3,  1 },
 {  5,  0, -2,  1 },
 {  5, -2, -1,  1 },
 {  5,  0, -1,  1 },
 {  5, -2,  0,  1 },
 {  5,  0,  0,  1 },
 {  5,  0,  1,  1 },
 {  5,  0,  2,  1 },
 {  5,  0, -1,  2 },
 {  5,  0,  0,  2 },
 {  6,  0, -2, -4 },
 {  6,  0, -1, -4 },
 {  6,  0, -3, -3 },
 {  6,  0, -2, -3 },
 {  6,  0, -1, -3 },
 {  6,  0,  0, -3 },
 {  6,  0,  1, -3 },
 {  6,  0, -4, -2 },
 {  6,  0, -3, -2 },
 {  6, -2, -2, -2 },
 {  6,  0, -2, -2 },
 {  6, -2, -1, -2 },
 {  6,  0, -1, -2 },
 {  6, -2,  0, -2 },
 {  6,  0,  0, -2 },
 {  6,  0,  1, -2 },
 {  6,  0,  2, -2 },
 {  6,  0, -4, -1 },
 {  6, -2, -3, -1 },
 {  6,  0, -3, -1 },
 {  6,  2, -3, -1 },
 {  6, -2, -2, -1 },
 {  6,  0, -2, -1 },
 {  6,  2, -2, -1 },
 {  6, -2, -1, -1 },
 {  6,  0, -1, -1 },
 {  6, -2,  0, -1 },
 {  6,  0,  0, -1 },
 {  6,  0,  1, -1 },
 {  6,  0,  2, -1 },
 {  6,  0,  3, -1 },
 {  6,  0, -5,  0 },
 {  6,  0, -4,  0 },
 {  6,  2, -4,  0 },
 {  6, -2, -3,  0 },
 {  6,  0, -3,  0 },
 {  6,  2, -3,  0 },
 {  6, -2, -2,  0 },
 {  6,  0, -2,  0 },
 {  6,  2, -2,  0 },
 {  6, -2, -1,  0 },
 {  6,  0, -1,  0 },
 {  6,  2, -1,  0 },
 {  6, -4,  0,  0 },
 {  6, -2,  0,  0 },
 {  6,  0,  0,  0 },
 {  6,  0,  1,  0 },
 {  6, -2,  2,  0 },
 {  6,  0,  2,  0 },
 {  6,  0,  3,  0 },
 {  6,  0,  4,  0 },
 {  6,  0, -4,  1 },
 {  6,  0, -3,  1 },
 {  6, -2, -2,  1 },
 {  6,  0, -2,  1 },
 {  6, -2, -1,  1 },
 {  6,  0, -1,  1 },
 {  6, -2,  0,  1 },
 {  6,  0,  0,  1 },
 {  6,  0,  1,  1 },
 {  6,  0,  2,  1 },
 {  6,  0, -4,  2 },
 {  6,  0, -3,  2 },
 {  6,  0, -2,  2 },
 {  6,  0, -1,  2 },
 {  7,  0, -4, -1 },
 {  7,  0, -3, -1 },
 {  7,  0, -2, -1 },
 {  7,  0, -1, -1 },
 {  7,  0, -4,  0 },
 {  7,  0, -3,  0 },
 {  7, -2, -2,  0 },
 {  7,  0, -2,  0 },
 {  7, -2, -1,  0 },
 {  7,  0, -1,  0 },
 {  7,  0,  0,  0 },
 {  7,  0, -3,  1 },
 {  7,  0, -2,  1 },
 {  7,  0, -1,  1 },
 {  8,  0, -4, -2 },
 {  8,  0, -3, -2 },
 {  8,  0, -2, -2 },
 {  8,  0, -1, -2 },
 {  8,  0,  0, -2 },
 {  8,  0, -4, -1 },
 {  8,  0, -3, -1 },
 {  8, -2, -2, -1 },
 {  8,  0, -2, -1 },
 {  8, -2, -1, -1 },
 {  8,  0, -1, -1 },
 {  8,  0,  0, -1 },
 {  8,  0,  1, -1 },
 {  8,  0, -5,  0 },
 {  8,  0, -4,  0 },
 {  8, -2, -3,  0 },
 {  8,  0, -3,  0 },
 {  8, -2, -2,  0 },
 {  8,  0, -2,  0 },
 {  8, -2, -1,  0 },
 {  8,  0, -1,  0 },
 {  8, -2,  0,  0 },
 {  8,  0,  0,  0 },
 {  8,  0,  1,  0 },
 {  8,  0,  2,  0 },
 {  8,  0, -4,  1 },
 {  8,  0, -3,  1 },
 {  8,  0, -2,  1 },
 {  8,  0, -1,  1 },
 {  8,  0,  0,  1 },
 { 10,  0, -3, -1 },
 { 10,  0, -2, -1 },
 { 10,  0, -5,  0 },
 { 10,  0, -4,  0 },
 { 10,  0, -3,  0 },
 { 10,  0, -2,  0 },
 { 10,  0, -1,  0 }
} };

constexpr array<double, 704> g_MAIN_S3_Nominal
{ {
  385000.52719,
      -3.14837,
      -0.00003,
       0.00038,
      79.66183,
  -20905.32206,
      -0.10326,
      -0.00837,
      -4.42124,
    -569.92332,
      -0.00102,
       0.00033,
      -0.00111,
     -23.21032,
       0.00026,
       0.00002,
       0.00027,
      -1.11693,
       0.00004,
       0.00002,
      -0.05888,
      -0.00329,
      -0.00019,
      -0.00011,
      -0.00174,
      -0.02674,
      -0.42242,
       0.00015,
      -0.00019,
      -7.00293,
      -0.05020,
      -0.00007,
      -0.00769,
    -129.62476,
       0.33465,
       0.00003,
      -0.18568,
      48.89010,
      -0.15803,
       0.00003,
      -0.24811,
     104.75896,
      -0.00682,
       0.00005,
       0.03265,
       5.75105,
      -0.00030,
      -0.00080,
       0.35509,
      -0.00006,
       0.02304,
       0.00153,
       0.00010,
      -0.00005,
      -0.00066,
      -0.00938,
      -0.13618,
      -0.00099,
      -0.00031,
      -2.11728,
       0.00450,
      -0.00558,
       1.06575,
      -0.00342,
      -0.00237,
       1.16562,
      -0.00011,
       0.00018,
       0.05072,
       0.00228,
       0.00009,
      -0.00021,
      -0.00287,
      -0.00002,
      -0.03894,
       0.00007,
      -0.00017,
       0.02127,
      -0.00007,
      -0.00003,
       0.01912,
       0.00078,
       0.00003,
      -0.00006,
      -0.00073,
       0.00034,
       0.00037,
      -0.00006,
       0.00035,
      -0.00161,
       0.00004,
       0.00029,
       0.00844,
      -0.00024,
      -0.02948,
      -0.00003,
      -0.00002,
       0.00159,
       0.00036,
       0.00004,
       0.00010,
       0.00065,
       0.00002,
      -0.00422,
      -0.00008,
      -0.00011,
       0.11411,
       0.00021,
      -0.01366,
       0.49757,
      -0.00074,
       0.00044,
       0.10998,
      -0.00007,
      -0.00003,
       0.01121,
       0.00097,
       0.00008,
      -0.00003,
      -0.00043,
      -0.00636,
      -0.00019,
      -0.09938,
      -0.00048,
      -0.00278,
      -1.73852,
       0.00457,
      -0.04232,
      -8.37909,
      -0.02471,
      -0.79564,
     108.74265,
      -0.01874,
       0.00017,
       0.06055,
       6.32199,
      -0.00164,
       0.00203,
       0.37852,
      -0.00013,
       0.00011,
       0.02368,
       0.00152,
       0.00010,
       0.00006,
       0.00101,
       0.01729,
       0.00014,
       0.00008,
       0.33226,
      -0.00148,
       0.00156,
       0.85127,
       0.00331,
       0.03379,
     -16.67533,
      -0.00006,
      -0.00562,
      -0.93335,
      -0.00037,
      -0.05904,
      -0.00002,
      -0.00392,
      -0.00027,
       0.00002,
       0.00028,
       0.00324,
       0.00438,
      -0.00002,
       0.03729,
       0.00002,
       0.00646,
       0.00067,
       0.00006,
       0.00006,
       0.00008,
       0.00054,
       0.00007,
      -0.00002,
      -0.00012,
      -0.00058,
      -0.00004,
       0.00036,
      -0.00426,
       0.00003,
      -0.01578,
      -0.00113,
      -0.00009,
       0.00002,
       0.00046,
       0.01116,
       0.00003,
       0.00026,
      -0.14808,
       0.00007,
       0.00104,
      -0.41076,
       0.00027,
      -0.02855,
       0.00003,
      -0.00209,
      -0.00016,
       0.00003,
       0.00067,
       0.00002,
       0.01517,
       0.00041,
       0.34304,
       0.00099,
       0.00957,
      -4.95049,
       0.00168,
       0.02839,
      -9.88519,
      -0.00003,
       0.00792,
      -0.65758,
       0.00084,
      -0.04608,
       0.00006,
      -0.00335,
      -0.00025,
      -0.00002,
       0.00006,
       0.00119,
       0.00004,
       0.02334,
       0.00079,
       0.49506,
      -0.00046,
       0.01518,
      10.05654,
       0.03024,
       0.32336,
    -152.14314,
       0.03734,
      -0.00035,
       0.65710,
    -204.59357,
       0.00108,
      -0.00085,
       0.20942,
     -12.83185,
       0.00000,
       0.00013,
       0.01804,
      -0.84883,
       0.00127,
      -0.05850,
       0.00009,
      -0.00413,
      -0.00030,
      -0.00002,
       0.00020,
       0.00304,
       0.00012,
       0.04740,
      -0.00002,
       0.00188,
       0.77854,
       0.00048,
       0.02910,
      14.40262,
      -0.02008,
      -0.00018,
      -0.00002,
       0.47263,
     246.15768,
       0.77405,
      -0.00014,
      -0.00039,
       8.75170,
   -3699.10468,
       0.59633,
      -0.00745,
      10.32129,
   -2955.96651,
       0.03143,
      -0.01866,
       4.13118,
    -170.73274,
       0.00181,
       0.00219,
       0.28399,
     -10.44472,
       0.00014,
       0.00002,
       0.01811,
      -0.66968,
       0.00119,
      -0.04422,
       0.00008,
      -0.00298,
      -0.00020,
       0.00009,
       0.00116,
       0.00003,
       0.01417,
       0.00021,
       0.16858,
       0.00030,
       0.00052,
       0.14368,
      -0.00203,
      -0.04736,
      24.20935,
      -0.00588,
       0.00012,
      -0.13572,
      30.82498,
      -0.00143,
       0.00040,
      -0.05527,
       2.61650,
      -0.00015,
      -0.00005,
      -0.00537,
       0.21252,
      -0.00044,
       0.01704,
      -0.00004,
       0.00135,
       0.00011,
      -0.00041,
      -0.00003,
      -0.01036,
      -0.00055,
      -0.10888,
      -0.00046,
      -0.01072,
       2.35380,
      -0.00004,
       0.00080,
       0.14764,
       0.00187,
       0.00912,
       0.00010,
       0.00036,
      -0.00010,
      -0.00052,
      -0.00003,
       0.00208,
      -0.00024,
       0.00002,
       0.00003,
      -0.00003,
       0.00008,
       0.00065,
       0.00003,
      -0.00055,
      -0.00004,
      -0.00016,
       0.00224,
      -0.00005,
       0.01389,
       0.00070,
      -0.01048,
       0.00002,
      -0.00067,
      -0.00003,
      -0.00012,
      -0.00334,
      -0.00014,
       0.05293,
      -0.00006,
      -0.00089,
       0.25561,
       0.01441,
      -0.15708,
       0.00049,
      -0.00887,
       0.00003,
      -0.00030,
      -0.00003,
      -0.00170,
      -0.00005,
      -0.04843,
      -0.00013,
      -0.00206,
       0.86243,
      -0.00101,
      -0.00868,
       3.25823,
      -0.00006,
      -0.00018,
       0.20099,
      -1.41893,
      -0.00015,
       0.00005,
       0.00671,
      -0.06545,
      -0.00002,
       0.00035,
      -0.00111,
       0.00016,
       0.00003,
      -0.00005,
      -0.00065,
      -0.00680,
      -0.00002,
      -0.00009,
       0.03810,
       0.00007,
      -0.00023,
      -0.21259,
       0.00006,
      -0.00808,
      -0.10766,
      -0.00018,
      -0.01448,
      -0.00150,
      -0.00014,
      -0.00007,
       0.00032,
       0.00099,
       0.00009,
       0.00162,
       0.00028,
       0.00003,
       0.00000,
      -0.00004,
       0.00010,
       0.00002,
      -0.00004,
      -0.00002,
      -0.00022,
      -0.00078,
      -0.00045,
      -0.00006,
      -0.00011,
      -0.00513,
      -0.00006,
      -0.01571,
      -0.00022,
      -0.00817,
      -0.00114,
      -0.00013,
      -0.00002,
      -0.00012,
      -0.00003,
      -0.00239,
       0.00004,
      -0.00036,
      -0.10834,
       0.00011,
      -0.00145,
      -0.27906,
       0.00005,
      -0.00403,
      -0.12806,
       0.00023,
      -0.01728,
       0.00005,
      -0.00184,
      -0.00018,
      -0.00004,
      -0.00041,
      -0.00005,
      -0.00230,
      -0.00059,
      -0.04316,
       0.00077,
      -0.00661,
      -1.89710,
       0.00160,
      -0.00006,
      -0.02698,
      -3.95812,
       0.00069,
       0.00003,
      -0.05789,
      -1.57145,
       0.00008,
       0.00005,
       0.00272,
      -0.20286,
       0.00055,
      -0.02079,
       0.00006,
      -0.00194,
      -0.00017,
      -0.00005,
      -0.00055,
      -0.00006,
      -0.00542,
      -0.00072,
      -0.02939,
      -0.00009,
      -0.00744,
      -0.51423,
       0.00925,
      -0.00014,
      -0.07982,
     -21.63627,
       0.01483,
      -0.00080,
      -0.32176,
     -34.78245,
       0.00550,
       0.00032,
      -0.50793,
     -11.64993,
       0.00067,
       0.00050,
       0.01938,
      -1.42255,
       0.00007,
       0.00400,
      -0.13922,
       0.00044,
      -0.01243,
       0.00004,
      -0.00106,
      -0.00009,
      -0.00004,
       0.00093,
      -0.02399,
      -0.00007,
       0.00036,
       0.23696,
      -0.00022,
       0.01551,
       0.57881,
      -0.00012,
       0.01969,
       0.24453,
      -0.00003,
      -0.00003,
       0.00013,
       0.03682,
      -0.00008,
       0.00430,
       0.00045,
       0.00004,
      -0.00003,
       0.00002,
       0.00252,
       0.00021,
       0.02658,
      -0.00064,
       0.01868,
       0.00005,
       0.00136,
       0.00003,
      -0.00015,
      -0.00020,
      -0.00003,
       0.00003,
       0.00009,
       0.00044,
       0.00002,
      -0.00006,
      -0.00016,
      -0.00002,
       0.00003,
       0.00128,
       0.00003,
       0.00573,
       0.00034,
      -0.00056,
       0.00021,
      -0.00153,
       0.00003,
      -0.00019,
       0.00026,
       0.00004,
       0.01149,
      -0.00002,
       0.00030,
       0.04591,
       0.00281,
      -0.00248,
       0.00142,
      -0.00814,
       0.00020,
      -0.00089,
      -0.00006,
       0.00003,
       0.00038,
      -0.00256,
      -0.00015,
      -0.00311,
      -0.00009,
      -0.00111,
      -0.00020,
      -0.00003,
       0.00007,
       0.00003,
      -0.00003,
      -0.00003,
      -0.00012,
      -0.00044,
      -0.00043,
      -0.00017,
      -0.00003,
      -0.00007,
      -0.00195,
      -0.00003,
      -0.00618,
      -0.00011,
      -0.00541,
      -0.00005,
      -0.00194,
      -0.00035,
      -0.00005,
      -0.00085,
      -0.00003,
      -0.02436,
       0.00004,
      -0.00041,
      -0.06678,
       0.00004,
      -0.00121,
      -0.05191,
      -0.00049,
      -0.01709,
      -0.00296,
      -0.00039,
      -0.00005,
      -0.00009,
      -0.00659,
       0.00012,
      -0.00027,
      -0.18316,
       0.00028,
      -0.00326,
      -0.42250,
       0.00020,
      -0.00785,
      -0.28663,
       0.00006,
      -0.00004,
      -0.00270,
      -0.08587,
      -0.01420,
       0.00004,
      -0.00182,
      -0.00020,
      -0.00002,
      -0.00024,
       0.00278,
       0.00019,
       0.00948,
       0.00042,
       0.00780,
       0.00016,
       0.00272,
       0.00052,
       0.00008,
       0.00004,
       0.00034,
       0.00044,
       0.00014,
       0.00002,
       0.00011,
       0.00006,
      -0.00002,
       0.00015,
       0.00066,
       0.00004,
       0.00032,
       0.00004,
      -0.00008,
      -0.00007,
      -0.00003,
      -0.00007,
      -0.00004,
      -0.00003,
      -0.00013,
      -0.00016,
      -0.00010,
      -0.00003,
      -0.00033,
      -0.00109,
      -0.00002,
      -0.00126,
      -0.00002,
      -0.00070,
      -0.00022,
      -0.00004,
      -0.00009,
      -0.00184,
      -0.00004,
      -0.00539,
      -0.00012,
      -0.00559,
      -0.00009,
      -0.00285,
      -0.00002,
      -0.00083,
      -0.00016,
      -0.00002,
       0.00004,
       0.00015,
       0.00019,
       0.00011,
       0.00004,
      -0.00003,
      -0.00002,
      -0.00002,
      -0.00007,
      -0.00010,
      -0.00007,
      -0.00003
} };

constexpr array<double, 704> g_MAIN_S3_LLR
{ {
    3.8500052903963596e+005,
   -3.1483019999591910e+000,
   -2.9999999997311145e-005,
    3.7998575836933673e-004,
    7.9660568469778482e+001,
   -2.0905355137759419e+004,
   -1.0325754056746472e-001,
   -8.3697361852690996e-003,
   -4.4211770490387590e+000,
   -5.6992512638489302e+002,
   -1.0199305524721176e-003,
    3.2999010831688693e-004,
   -1.1100043486792319e-003,
   -2.3210430497874530e+001,
    2.6000521851116001e-004,
    1.9999999998207433e-005,
    2.7000173948728815e-004,
   -1.1169371044179981e+000,
    3.9999999996414865e-005,
    1.9999999998207433e-005,
   -5.8880465319867410e-002,
   -3.2900313109119031e-003,
   -1.9000173949445843e-004,
   -1.0999999999014087e-004,
   -1.7399525712965291e-003,
   -2.6739187212881998e-002,
   -4.2240656115839154e-001,
    1.4999999998655572e-004,
   -1.8999999998297061e-004,
   -7.0026960617373861e+000,
   -5.0197444216470780e-002,
   -6.9999999993726000e-005,
   -7.6895894726563942e-003,
   -1.2962022424070293e+002,
    3.3463243448180802e-001,
    2.9999999997311145e-005,
   -1.8566991736964209e-001,
    4.8888328449310244e+001,
   -1.5802146652025906e-001,
    2.9999999997311145e-005,
   -2.4809697371058767e-001,
    1.0475529435395761e+002,
   -6.8196430863427648e-003,
    4.9999999995518581e-005,
    3.2648339848703183e-002,
    5.7508578538436117e+000,
   -2.9999999997311143e-004,
   -7.9995729744124049e-004,
    3.5507869597491126e-001,
   -5.9999999994622291e-005,
    2.3039299558702493e-002,
    1.5299569661513701e-003,
    9.9999999991037163e-005,
   -4.9999999995518581e-005,
   -6.5995527316327525e-004,
   -9.3793572985260542e-003,
   -1.3617046100022667e-001,
   -9.8991435737977806e-004,
   -3.0997546258218684e-004,
   -2.1171282632583841e+000,
    4.4995981571196867e-003,
   -5.5794920664080727e-003,
    1.0656722797003297e+000,
   -3.4196903329037475e-003,
   -2.3697890694752721e-003,
    1.1655363838529491e+000,
   -1.0999999999014087e-004,
    1.7998773128885274e-004,
    5.0716438193917468e-002,
    2.2798456792376266e-003,
    8.9999999991933447e-005,
   -2.0997546259114971e-004,
   -2.8696958416961804e-003,
   -1.9999999998207433e-005,
   -3.8935785427179986e-002,
    6.9999999993726000e-005,
   -1.6998159694224193e-004,
    2.1267663634227500e-002,
   -6.9999999993726000e-005,
   -2.9999999997311145e-005,
    1.9117926343405874e-002,
    7.7991672833222240e-004,
    2.9999999997311145e-005,
   -5.9999999994622291e-005,
   -7.2989571525071807e-004,
    3.3995092518946974e-004,
    3.6994479083927371e-004,
   -5.9999999976829571e-005,
    3.4996319376793487e-004,
   -1.6098221032175221e-003,
    3.9999999996414865e-005,
    2.8997546248611943e-004,
    8.4393952091937965e-003,
   -2.3998159685590067e-004,
   -2.9477843313093360e-002,
   -2.9999999997311145e-005,
   -1.9999999998207433e-005,
    1.5898860497606509e-003,
    3.5997546128880718e-004,
    3.9999999996414865e-005,
    9.9999999955451723e-005,
    6.4997894135227982e-004,
    1.9999999998207433e-005,
   -4.2198518026374575e-003,
   -7.9999999966140657e-005,
   -1.0999999995455543e-004,
    1.1410603726168141e-001,
    2.0999999991000714e-004,
   -1.3659260469050206e-002,
    4.9755215819351722e-001,
   -7.3996382995627943e-004,
    4.3997384265678845e-004,
    1.0997620924828105e-001,
   -6.9999999967036927e-005,
   -2.9999999997311145e-005,
    1.1209625503392669e-002,
    9.6996754071621016e-004,
    7.9999999966140657e-005,
   -2.9999999997311145e-005,
   -4.3000347883319725e-004,
   -6.3600399990106872e-003,
   -1.8999999992069610e-004,
   -9.9380484219865714e-002,
   -4.7999485494516357e-004,
   -2.7799621701349053e-003,
   -1.7385258483230051e+000,
    4.5699363718342432e-003,
   -4.2319341920191246e-002,
   -8.3791067416664688e+000,
   -2.4709621256653429e-002,
   -7.9562635561038542e-001,
    1.0874270141177935e+002,
   -1.8739689483299998e-002,
    1.6999999993138503e-004,
    6.0549063918332549e-002,
    6.3220032203434355e+000,
   -1.6399763369809809e-003,
    2.0299711939103784e-003,
    3.7852142052101295e-001,
   -1.2999999994386648e-004,
    1.0999999995455543e-004,
    2.3680120794187420e-002,
    1.5200095654877134e-003,
    9.9999999955451723e-005,
    5.9999999976829571e-005,
    1.0099692817676822e-003,
    1.7289444638976460e-002,
    1.3999999994297020e-004,
    7.9999999966140657e-005,
    3.3224889132530933e-001,
   -1.4799219996174619e-003,
    1.5599202600721149e-003,
    8.5124035002281895e-001,
    3.3098243750911368e-003,
    3.3788196608704760e-002,
   -1.6674723862047578e+001,
   -5.9999999976829571e-005,
   -5.6197040706796285e-003,
   -9.3331756600538507e-001,
   -3.6998773037104138e-004,
   -5.9038043138950086e-002,
   -1.9999999998207433e-005,
   -3.9198780429301125e-003,
   -2.7000173939832454e-004,
    1.9999999998207433e-005,
    2.7998159683452278e-004,
    3.2397711906711835e-003,
    4.3796879646369998e-003,
   -1.9999999998207433e-005,
    3.7287275626590614e-002,
    1.9999999998207433e-005,
    6.4595389545972414e-003,
    6.6995353419581881e-004,
    5.9999999976829571e-005,
    5.9999999976829571e-005,
    7.9999999966140657e-005,
    5.3993865628970640e-004,
    6.9999999967036927e-005,
   -1.9999999998207433e-005,
   -1.1997546259921626e-004,
   -5.7989571487604592e-004,
   -3.9999999996414865e-005,
    3.5994479084017005e-004,
   -4.2593812511134812e-003,
    2.9999999997311145e-005,
   -1.5777681189612303e-002,
   -1.1298361100869345e-003,
   -8.9987731296919292e-005,
    1.9999999998207433e-005,
    4.5995266469020209e-004,
    1.1158808779392803e-002,
    2.9999999997311145e-005,
    2.5996932823916120e-004,
   -1.4806394212356289e-001,
    6.9999999993726000e-005,
    1.0398640608647635e-003,
   -4.1071486233056914e-001,
    2.6996932823826497e-004,
   -2.8546907871265167e-002,
    2.9999999997311145e-005,
   -2.0897738487721945e-003,
   -1.5998159694313822e-004,
    2.9999999997311145e-005,
    6.6995527393861188e-004,
    1.9999999998207433e-005,
    1.5168961862541621e-002,
    4.0996932822571694e-004,
    3.4301592316629376e-001,
    9.8991435854412727e-004,
    9.5691426166157369e-003,
   -4.9501349067196072e+000,
    1.6798497739297300e-003,
    2.8387412335730781e-002,
   -9.8844667465718086e+000,
   -2.9999999997311145e-005,
    7.9192898353840548e-003,
   -6.5753296501894665e-001,
    8.3992575672060886e-004,
   -4.6076772975214793e-002,
    5.9999999994622291e-005,
   -3.3497711457249656e-003,
   -2.4998333644655948e-004,
   -1.9999999998207433e-005,
    5.9999999994622291e-005,
    1.1899666267739756e-003,
    3.9999999996414865e-005,
    2.3339289036866287e-002,
    7.8995729899460239e-004,
    4.9504418153977070e-001,
   -4.5997558231772182e-004,
    1.5179225093487955e-002,
    1.0056203332336866e+001,
    3.0238458060069424e-002,
    3.2334304362267791e-001,
   -1.5213781181555629e+002,
    3.7338054588554326e-002,
   -3.4997610010859237e-004,
    6.5706430206093058e-001,
   -2.0458611789480557e+002,
    1.0799620769067898e-003,
   -8.4993831174621777e-004,
    2.0940899652925385e-001,
   -1.2831403548880713e+001,
    1.7395114878210153e-009,
    1.2999999998834829e-004,
    1.8039080838079844e-002,
   -8.4880184799015856e-001,
    1.2699337499019264e-003,
   -5.8498148654635040e-002,
    8.9999999991933447e-005,
   -4.1298735975484717e-003,
   -3.0000260924034318e-004,
   -1.9999999998207433e-005,
    2.0000260924930605e-004,
    3.0400286993383803e-003,
    1.1999999998924458e-004,
    4.7400375694138144e-002,
   -1.9999999998207433e-005,
    1.8799811588075014e-003,
    7.7854495144615832e-001,
    4.7999572409370545e-004,
    2.9099632514815468e-002,
    1.4402688991954179e+001,
   -2.0079749726991641e-002,
   -1.7999999998386689e-004,
   -1.9999999998207433e-005,
    4.7262330622706727e-001,
    2.4615847968480679e+002,
    7.7403904019864489e-001,
   -1.3999999998745200e-004,
   -3.8998449174501354e-004,
    8.7515625232254699e+000,
   -3.6991109330180198e+003,
    5.9632093121812468e-001,
   -7.4497489299349016e-003,
    1.0321107088258565e+001,
   -2.9559675626057124e+003,
    3.1429705662064322e-002,
   -1.8659381227811495e-002,
    4.1311145454994769e+000,
   -1.7073307906648640e+002,
    1.8099968202389119e-003,
    2.1899294167495083e-003,
    2.8398596260369763e-001,
   -1.0444757788459564e+001,
    1.4000260925468372e-004,
    1.9999999998207433e-005,
    1.8109768178310622e-002,
   -6.6968351866645870e-001,
    1.1899884424199981e-003,
   -4.4220292185840150e-002,
    7.9999999992829730e-005,
   -2.9800243497888055e-003,
   -2.0000173949356214e-004,
    8.9999999991933447e-005,
    1.1599666236717329e-003,
    2.9999999997311145e-005,
    1.4169568122116911e-002,
    2.0999999920494518e-004,
    1.6857465048617704e-001,
    2.9998773127809725e-004,
    5.1997383232171282e-004,
    1.4367547210982884e-001,
   -2.0298945871754515e-003,
   -4.7357510174271755e-002,
    2.4208498493055885e+001,
   -5.8796966665792833e-003,
    1.1999999998924458e-004,
   -1.3571262386357466e-001,
    3.0823859907219550e+001,
   -1.4299250058847307e-003,
    3.9997222304910318e-004,
   -5.5267093531528186e-002,
    2.6164091717727311e+000,
   -1.4999999998655572e-004,
   -4.9999999995518581e-005,
   -5.3697280450288095e-003,
    2.1251296661321492e-001,
   -4.3997558231951439e-004,
    1.7039460374453742e-002,
   -3.9999999996414865e-005,
    1.3499596203276857e-003,
    1.0999999999014087e-004,
   -4.0997193749294864e-004,
   -2.9999999997311145e-005,
   -1.0359283640210834e-002,
   -5.4995104491962696e-004,
   -1.0887233898445144e-001,
   -4.5995543976371276e-004,
   -1.0719047352460247e-002,
    2.3536309843524257e+000,
   -3.9999999996414865e-005,
    7.9992662647993798e-004,
    1.4762959370247397e-001,
    1.8698379005329033e-003,
    9.1193713528487444e-003,
    9.9999999991037163e-005,
    3.5997807184493723e-004,
   -9.9999999991037163e-005,
   -5.1994652956108441e-004,
   -2.9999999997311145e-005,
    2.0797756327894123e-003,
   -2.3997546181222802e-004,
    1.9999999998207433e-005,
    2.9999999997311145e-005,
   -2.9999999997311145e-005,
    7.9999999966140657e-005,
    6.4992638638489825e-004,
    2.9999999997311145e-005,
   -5.4993865628881011e-004,
   -3.9999999996414865e-005,
   -1.5998773123726716e-004,
    2.2398413285241595e-003,
   -4.9999999977725862e-005,
    1.3889004284567788e-002,
    6.9993929375197815e-004,
   -1.0479229714051838e-002,
    1.9999999998207433e-005,
   -6.6995092492858706e-004,
   -2.9999999997311145e-005,
   -1.1999999994476279e-004,
   -3.3398929652567809e-003,
   -1.3999999994297020e-004,
    5.2928223215939374e-002,
   -5.9999999976829571e-005,
   -8.8995381966018767e-004,
    2.5560112901794674e-001,
    1.4409224300161040e-002,
   -1.5707413356650624e-001,
    4.8997384263451424e-004,
   -8.8696765680077155e-003,
    2.9999999997311145e-005,
   -2.9999999986635514e-004,
   -2.9999999997311145e-005,
   -1.7000104363293989e-003,
   -4.9999999995518581e-005,
   -4.8430212033192685e-002,
   -1.2999999994386648e-004,
   -2.0599711938987931e-003,
    8.6243278904298004e-001,
   -1.0099879692011121e-003,
   -8.6798609676501642e-003,
    3.2582370807619236e+000,
   -5.9999999976829571e-005,
   -1.7999999992159238e-004,
    2.0098656931416145e-001,
   -1.4189284087900234e+000,
   -1.4999999993317757e-004,
    4.9999999977725862e-005,
    6.7098941217191028e-003,
   -6.5449921583248111e-002,
   -1.9999999998207433e-005,
    3.4999999984408099e-004,
   -1.1099973984011820e-003,
    1.5999999876793204e-004,
    2.9999999997311145e-005,
   -4.9999999995518581e-005,
   -6.4997980994367449e-004,
   -6.7997806585012449e-003,
   -1.9999999998207433e-005,
   -8.9999999956348007e-005,
    3.8098714295570914e-002,
    6.9999999967036927e-005,
   -2.2999999989931820e-004,
   -2.1258262555232793e-001,
    5.9999999976829571e-005,
   -8.0795659691643255e-003,
   -1.0765620290507359e-001,
   -1.7999999992159238e-004,
   -1.4479504603693322e-002,
   -1.4999508765716299e-003,
   -1.3999999993407386e-004,
   -6.9999999967036927e-005,
    3.1997546053395219e-004,
    9.8992812884629469e-004,
    8.9987731261333853e-005,
    1.6198834451392309e-003,
    2.7998159683452278e-004,
    2.9999999997311145e-005,
   -2.6080406079458980e-009,
   -3.9999999996414865e-005,
    9.9987731260437569e-005,
    1.9980272777955583e-005,
   -3.9999999996414865e-005,
   -1.9999999998207433e-005,
   -2.1996932824274633e-004,
   -7.7988344538687309e-004,
   -4.4993252097274005e-004,
   -5.9999999994622291e-005,
   -1.0998773129512671e-004,
   -5.1294521563901270e-003,
   -5.9999999994622291e-005,
   -1.5708297041256869e-002,
   -2.1997546259025342e-004,
   -8.1691051872718589e-003,
   -1.1398737847624751e-003,
   -1.2998773129333413e-004,
   -1.9999999998207433e-005,
   -1.1999999998924458e-004,
   -2.9999999997311145e-005,
   -2.3898334058753568e-003,
    3.9999999996414865e-005,
   -3.5997546257770548e-004,
   -1.0833240876766437e-001,
    1.0999999999014087e-004,
   -1.4498750613471372e-003,
   -2.7904006579942225e-001,
    4.9999999995518581e-005,
   -4.0296357569822210e-003,
   -1.2805075092513282e-001,
    2.2998159693686420e-004,
   -1.7278775418016083e-002,
    4.9999999995518581e-005,
   -1.8398719975784821e-003,
   -1.7998773128885274e-004,
   -3.9999999996414865e-005,
   -4.0999121029121382e-004,
   -4.9999999995518581e-005,
   -2.2999289024477035e-003,
   -5.8996643988306632e-004,
   -4.3158624131595380e-002,
    7.6996117527517003e-004,
   -6.6096630553469891e-003,
   -1.8970367926847447e+000,
    1.5999167328141174e-003,
   -5.9999999994622291e-005,
   -2.6978590582815640e-002,
   -3.9579827434849597e+000,
    6.8996643948598696e-004,
    2.9999999997311145e-005,
   -5.7886881100952178e-002,
   -1.5713943841003077e+000,
    7.9999999992829730e-005,
    4.9999999995518581e-005,
    2.7198614091688139e-003,
   -2.0285309945865274e-001,
    5.4997558153342251e-004,
   -2.0789319306507164e-002,
    5.9999999994622291e-005,
   -1.9399368200675657e-003,
   -1.6999999998476318e-004,
   -4.9999999995518581e-005,
   -5.5000521770893508e-004,
   -5.9999999994622291e-005,
   -5.4200426093952208e-003,
   -7.1999271638481445e-004,
   -2.9390186091328888e-002,
   -8.9999999991933447e-005,
   -7.4399077060242633e-003,
   -5.1423243497062632e-001,
    9.2498845164474333e-003,
   -1.3999999998745200e-004,
   -7.9818880501352676e-002,
   -2.1636343875076179e+001,
    1.4829797144853685e-002,
   -7.9997285976700968e-004,
   -3.2175506886962035e-001,
   -3.4782523745558805e+001,
    5.4999213282995590e-003,
    3.1998836880629540e-004,
   -5.0792130896835275e-001,
   -1.1649947819443046e+001,
    6.6999271677741228e-004,
    4.9998449173515452e-004,
    1.9379704486947858e-002,
   -1.4225541276068021e+000,
    6.9999999993726000e-005,
    3.9999449922891294e-003,
   -1.3922061869363414e-001,
    4.3999485511777951e-004,
   -1.2430067812277675e-002,
    3.9999999996414865e-005,
   -1.0600078253779905e-003,
   -8.9999999991933447e-005,
   -3.9999999996414865e-005,
    9.2996928018122837e-004,
   -2.3989247389315162e-002,
   -6.9999999993726000e-005,
    3.5998773321330174e-004,
    2.3695196810801203e-001,
   -2.1999999998028173e-004,
    1.5509193325354423e-002,
    5.7878980681430592e-001,
   -1.1999999998924458e-004,
    1.9688938804572634e-002,
    2.4452136289194804e-001,
   -2.9999999997311145e-005,
   -2.9999999997311145e-005,
    1.2999999882399901e-004,
    3.6818751174001671e-002,
   -7.9999999992829730e-005,
    4.2998578912151475e-003,
    4.4998420502002841e-004,
    3.9999999996414865e-005,
   -2.9999999997311145e-005,
    1.9999999998207433e-005,
    2.5198281393313598e-003,
    2.0998159693865677e-004,
    2.6578133580483304e-002,
   -6.3993929244509197e-004,
    1.8678664992649377e-002,
    4.9999999995518581e-005,
    1.3599096738504068e-003,
    2.9999999997311145e-005,
   -1.4998159694403447e-004,
   -1.9997546259204602e-004,
   -2.9999999997311145e-005,
    2.9999999997311145e-005,
    8.9999999965244374e-005,
    4.3996932690744069e-004,
    1.9999999998207433e-005,
   -5.9999999976829571e-005,
   -1.5998773123726716e-004,
   -1.9999999998207433e-005,
    2.9999999997311145e-005,
    1.2799570106219100e-003,
    2.9999999997311145e-005,
    5.7298105518939404e-003,
    3.3998773115885951e-004,
   -5.5997372556938698e-004,
    2.0999999991000714e-004,
   -1.5299447878728246e-003,
    2.9999999997311145e-005,
   -1.8999999992069610e-004,
    2.6000173939922077e-004,
    3.9999999996414865e-005,
    1.1490054770587659e-002,
   -1.9999999998207433e-005,
    2.9999999986635514e-004,
    4.5910184261982304e-002,
    2.8099578190159021e-003,
   -2.4799485949198968e-003,
    1.4199767343365658e-003,
   -8.1399938984571592e-003,
    1.9999999991090345e-004,
   -8.8999999960885808e-004,
   -5.9999999976829571e-005,
    2.9999999997311145e-005,
    3.7998946909650384e-004,
   -2.5599166785089333e-003,
   -1.4999999993317757e-004,
   -3.1098903978782500e-003,
   -8.9999999965244374e-005,
   -1.1099596656358362e-003,
   -1.9999999991090345e-004,
   -2.9999999997311145e-005,
    6.9999999967036927e-005,
    2.9999999997311145e-005,
   -2.9999999997311145e-005,
   -2.9999999997311145e-005,
   -1.1998773129423043e-004,
   -4.3995092401615758e-004,
   -4.2995092362893745e-004,
   -1.6998159694224193e-004,
   -2.9999999997311145e-005,
   -6.9999999993726000e-005,
   -1.9498684724883475e-003,
   -2.9999999997311145e-005,
   -6.1795669147913027e-003,
   -1.0999999999014087e-004,
   -5.4096134264304419e-003,
   -4.9999999995518581e-005,
   -1.9398623822584544e-003,
   -3.4997546141425244e-004,
   -4.9999999995518581e-005,
   -8.4997454439204531e-004,
   -2.9999999997311145e-005,
   -2.4359226139441414e-002,
    3.9999999996414865e-005,
   -4.0998773126823813e-004,
   -6.6777799989498579e-002,
    3.9999999996414865e-005,
   -1.2099394109786337e-003,
   -5.1908239459320625e-002,
   -4.8997384125107932e-004,
   -1.7089416902058164e-002,
   -2.9598991387017283e-003,
   -3.8998946922905289e-004,
   -4.9999999995518581e-005,
   -8.9999999991933447e-005,
   -6.5900417302198174e-003,
    1.1999999998924458e-004,
   -2.6999999997580034e-004,
   -1.8316094826847287e-001,
    2.7999999997490400e-004,
   -3.2599561529611147e-003,
   -4.2250165827248998e-001,
    1.9999999998207433e-004,
   -7.8498755921186901e-003,
   -2.8663087781865987e-001,
    5.9999999994622291e-005,
   -3.9999999996414865e-005,
   -2.6999534677236972e-003,
   -8.5870257737473130e-002,
   -1.4200047788719221e-002,
    3.9999999996414865e-005,
   -1.8200086911844525e-003,
   -1.9999999998207433e-004,
   -1.9999999998207433e-005,
   -2.3999999842602346e-004,
    2.7799087645080695e-003,
    1.8999999998297061e-004,
    9.4796825087208689e-003,
    4.1998773010299258e-004,
    7.7997308005767260e-003,
    1.5999999998565946e-004,
    2.7199079303358601e-003,
    5.1998333486989398e-004,
    7.9999999992829730e-005,
    3.9999999996414865e-005,
    3.3997720092663659e-004,
    4.3996932667056242e-004,
    1.3999999998745200e-004,
    1.9999999998207433e-005,
    1.0999999995455543e-004,
    5.9999999976829571e-005,
   -1.9999999998207433e-005,
    1.4999999993317757e-004,
    6.6000347679193334e-004,
    3.9999999996414865e-005,
    3.1999999830320053e-004,
    3.9999999996414865e-005,
   -7.9999999966140657e-005,
   -6.9999999967036927e-005,
   -2.9999999997311145e-005,
   -6.9999999967036927e-005,
   -3.9999999996414865e-005,
   -2.9999999997311145e-005,
   -1.2999999998834829e-004,
   -1.5998773129064531e-004,
   -9.9999999991037163e-005,
   -2.9999999997311145e-005,
   -3.2998946962254700e-004,
   -1.0899622741230546e-003,
   -1.9999999998207433e-005,
   -1.2599561389840780e-003,
   -1.9999999998207433e-005,
   -6.9997806870953207e-004,
   -2.1999999920404889e-004,
   -3.9999999996414865e-005,
   -8.9999999991933447e-005,
   -1.8400113012261245e-003,
   -3.9999999996414865e-005,
   -5.3900269440916079e-003,
   -1.1999999998924458e-004,
   -5.5900217228224049e-003,
   -8.9999999991933447e-005,
   -2.8500095558023656e-003,
   -1.9999999998207433e-005,
   -8.3000260569979234e-004,
   -1.5999999998565946e-004,
   -1.9999999998207433e-005,
    3.9999999996414865e-005,
    1.4999999998655572e-004,
    1.8999999998297061e-004,
    1.0999999999014087e-004,
    3.9999999996414865e-005,
   -2.9999999997311145e-005,
   -1.9999999998207433e-005,
   -1.9999999998207433e-005,
   -6.9999999993726000e-005,
   -9.9999999991037163e-005,
   -6.9999999993726000e-005,
   -2.9999999997311145e-005
} };

constexpr array<double, 704> g_MAIN_S3_DE405
{ {
    3.8500052903228404e+005,
   -3.1483021922752781e+000,
   -2.9999999997633791e-005,
    3.7998579755659352e-004,
    7.9660570476435268e+001,
   -2.0905354935205090e+004,
   -1.0325754836510391e-001,
   -8.3697366020164584e-003,
   -4.4211771171071561e+000,
   -5.6992511533509469e+002,
   -1.0199309130843224e-003,
    3.2999012092559170e-004,
   -1.1100043220264174e-003,
   -2.3210429822877586e+001,
    2.6000518651625334e-004,
    1.9999999998422528e-005,
    2.7000172882429092e-004,
   -1.1169370611059974e+000,
    3.9999999996845056e-005,
    1.9999999998422528e-005,
   -5.8880462467805837e-002,
   -3.2900311189610690e-003,
   -1.9000172883060085e-004,
   -1.0999999999132389e-004,
   -1.7399524762050224e-003,
   -2.6739186021817699e-002,
   -4.2240654624048563e-001,
    1.4999999998816893e-004,
   -1.8999999998501401e-004,
   -7.0026958822396903e+000,
   -5.0197444683412334e-002,
   -6.9999999994478830e-005,
   -7.6895896701093900e-003,
   -1.2962022217205058e+002,
    3.3463244089062189e-001,
    2.9999999997633791e-005,
   -1.8566992299900853e-001,
    4.8888328059174846e+001,
   -1.5802147102879513e-001,
    2.9999999997633791e-005,
   -2.4809697845850509e-001,
    1.0475529267427028e+002,
   -6.8196431963242548e-003,
    4.9999999996056319e-005,
    3.2648340142780188e-002,
    5.7508577057015726e+000,
   -2.9999999997633786e-004,
   -7.9995730218233028e-004,
    3.5507868338050058e-001,
   -5.9999999995267581e-005,
    2.3039298527847123e-002,
    1.5299568826931335e-003,
    9.9999999992112637e-005,
   -4.9999999996056319e-005,
   -6.5995523869077661e-004,
   -9.3793569172097267e-003,
   -1.3617045669341818e-001,
   -9.8991435536443227e-004,
   -3.0997545870701143e-004,
   -2.1171282169327208e+000,
    4.4995982169380271e-003,
   -5.5794921950237464e-003,
    1.0656722650829162e+000,
   -3.4196904093387018e-003,
   -2.3697891063466005e-003,
    1.1655363585022665e+000,
   -1.0999999999132389e-004,
    1.7998772935153386e-004,
    5.0716436611946462e-002,
    2.2798455837538525e-003,
    8.9999999992901368e-005,
   -2.0997545871489877e-004,
   -2.8696957386574184e-003,
   -1.9999999998422528e-005,
   -3.8935784344622658e-002,
    6.9999999994478830e-005,
   -1.6998159403518818e-004,
    2.1267663221308263e-002,
   -6.9999999994478830e-005,
   -2.9999999997633791e-005,
    1.9117925823248456e-002,
    7.7991669876697665e-004,
    2.9999999997633791e-005,
   -5.9999999995267581e-005,
   -7.2989569871707838e-004,
    3.3995091743610744e-004,
    3.6994478211660676e-004,
   -5.9999999976149397e-005,
    3.4996318794531937e-004,
   -1.6098220750738919e-003,
    3.9999999996845056e-005,
    2.8997545860343885e-004,
    8.4393950201855413e-003,
   -2.3998159394363517e-004,
   -2.9477842955863663e-002,
   -2.9999999997633791e-005,
   -1.9999999998422528e-005,
    1.5898860148522863e-003,
    3.5997545731815292e-004,
    3.9999999996845056e-005,
    9.9999999953876269e-005,
    6.4997891613239060e-004,
    1.9999999998422528e-005,
   -4.2198517073415813e-003,
   -7.9999999965012830e-005,
   -1.0999999995308753e-004,
    1.1410603535319587e-001,
    2.0999999990696381e-004,
   -1.3659260861906745e-002,
    4.9755215400672276e-001,
   -7.3996384925082781e-004,
    4.3997385525540976e-004,
    1.0997620731914327e-001,
   -6.9999999965801547e-005,
   -2.9999999997633791e-005,
    1.1209625212771860e-002,
    9.6996750810329085e-004,
    7.9999999965012830e-005,
   -2.9999999997633791e-005,
   -4.3000345749475338e-004,
   -6.3600397530792401e-003,
   -1.8999999991810035e-004,
   -9.9380481230041715e-002,
   -4.7999485445292023e-004,
   -2.7799622097322992e-003,
   -1.7385258128996706e+000,
    4.5699364414918571e-003,
   -4.2319342957746767e-002,
   -8.3791066502014502e+000,
   -2.4709621847278977e-002,
   -7.9562638278438003e-001,
    1.0874270127246304e+002,
   -1.8739690056048076e-002,
    1.6999999992923693e-004,
    6.0549065378075952e-002,
    6.3220031516918853e+000,
   -1.6399763697789705e-003,
    2.0299712263623713e-003,
    3.7852141276755424e-001,
   -1.2999999994195094e-004,
    1.0999999995308753e-004,
    2.3680120083704129e-002,
    1.5200095067157658e-003,
    9.9999999953876269e-005,
    5.9999999976149397e-005,
    1.0099692386042803e-003,
    1.7289444030638344e-002,
    1.3999999994116218e-004,
    7.9999999965012830e-005,
    3.3224888279882392e-001,
   -1.4799220185026741e-003,
    1.5599202896221002e-003,
    8.5124033619353401e-001,
    3.3098244400641009e-003,
    3.3788197501593308e-002,
   -1.6674723746336660e+001,
   -5.9999999976149397e-005,
   -5.6197041728912800e-003,
   -9.3331755043545517e-001,
   -3.6998772836866227e-004,
   -5.9038041580494902e-002,
   -1.9999999998422528e-005,
   -3.9198779032451961e-003,
   -2.7000172872869996e-004,
    1.9999999998422528e-005,
    2.7998159392136196e-004,
    3.2397710884068728e-003,
    4.3796878710824355e-003,
   -1.9999999998422528e-005,
    3.7287275130270789e-002,
    1.9999999998422528e-005,
    6.4595388101637554e-003,
    6.6995351042992298e-004,
    5.9999999976149397e-005,
    5.9999999976149397e-005,
    7.9999999965012830e-005,
    5.3993864658532283e-004,
    6.9999999965801547e-005,
   -1.9999999998422528e-005,
   -1.1997545872199740e-004,
   -5.7989569831188030e-004,
   -3.9999999996845056e-005,
    3.5994478211739549e-004,
   -4.2593811093623989e-003,
    2.9999999997633791e-005,
   -1.5777680821354629e-002,
   -1.1298360731086838e-003,
   -8.9987729358632481e-005,
    1.9999999998422528e-005,
    4.5995264627222943e-004,
    1.1158808366184655e-002,
    2.9999999997633791e-005,
    2.5996932339382065e-004,
   -1.4806393807175231e-001,
    6.9999999994478830e-005,
    1.0398640811513848e-003,
   -4.1071485480076358e-001,
    2.6996932339303192e-004,
   -2.8546907073008252e-002,
    2.9999999997633791e-005,
   -2.0897737689961008e-003,
   -1.5998159403597692e-004,
    2.9999999997633791e-005,
    6.6995523952404629e-004,
    1.9999999998422528e-005,
    1.5168961243902334e-002,
    4.0996932338198958e-004,
    3.4301591248131014e-001,
    9.8991435661551982e-004,
    9.5691427477273029e-003,
   -4.9501348002130729e+000,
    1.6798497951639492e-003,
    2.8387413047911467e-002,
   -9.8844666217709616e+000,
   -2.9999999997633791e-005,
    7.9192899476857939e-003,
   -6.5753295021873692e-001,
    8.3992576200476792e-004,
   -4.6076771497468638e-002,
    5.9999999995267581e-005,
   -3.3497710052450623e-003,
   -2.4998332287446507e-004,
   -1.9999999998422528e-005,
    5.9999999995267581e-005,
    1.1899665613369270e-003,
    3.9999999996845056e-005,
    2.3339287993958329e-002,
    7.8995730385123561e-004,
    4.9504416423055558e-001,
   -4.5997558426192373e-004,
    1.5179225244421566e-002,
    1.0056203077190446e+001,
    3.0238458352181256e-002,
    3.2334304986755269e-001,
   -1.5213780942589074e+002,
    3.7338055220178724e-002,
   -3.4997612144415849e-004,
    6.5706432249497604e-001,
   -2.0458611646087306e+002,
    1.0799619981539893e-003,
   -8.4993836896337369e-004,
    2.0940900061519147e-001,
   -1.2831403332461623e+001,
    1.7288455868366050e-009,
    1.2999999998974642e-004,
    1.8039081008914867e-002,
   -8.4880182533646864e-001,
    1.2699337512832878e-003,
   -5.8498146534920067e-002,
    8.9999999992901368e-005,
   -4.1298734075970499e-003,
   -3.0000259324471811e-004,
   -1.9999999998422528e-005,
    2.0000259325260553e-004,
    3.0400285232102318e-003,
    1.1999999999053516e-004,
    4.7400373388124431e-002,
   -1.9999999998422528e-005,
    1.8799811461358263e-003,
    7.7854492106106743e-001,
    4.7999571822327815e-004,
    2.9099632684995498e-002,
    1.4402688569714263e+001,
   -2.0079749835047431e-002,
   -1.7999999998580274e-004,
   -1.9999999998422528e-005,
    4.7262331352426273e-001,
    2.4615847485355792e+002,
    7.7403905203800960e-001,
   -1.3999999998895766e-004,
   -3.8998452277728355e-004,
    8.7515627423090212e+000,
   -3.6991108963980764e+003,
    5.9632094424355830e-001,
   -7.4497493960877454e-003,
    1.0321107477029065e+001,
   -2.9559675579724144e+003,
    3.1429705179098821e-002,
   -1.8659382347128925e-002,
    4.1311146504675742e+000,
   -1.7073307712477057e+002,
    1.8099967120070663e-003,
    2.1899295292498355e-003,
    2.8398596703683637e-001,
   -1.0444757567465345e+001,
    1.4000259325733791e-004,
    1.9999999998422528e-005,
    1.8109768286069500e-002,
   -6.6968349796415183e-001,
    1.1899884332213365e-003,
   -4.4220290390859181e-002,
    7.9999999993690113e-005,
   -2.9800242002669388e-003,
   -2.0000172882981212e-004,
    8.9999999992901368e-005,
    1.1599665580030599e-003,
    2.9999999997633791e-005,
    1.4169567487434505e-002,
    2.0999999914937822e-004,
    1.6857464438507142e-001,
    2.9998772934206898e-004,
    5.1997384415181720e-004,
    1.4367546693920624e-001,
   -2.0298946113646351e-003,
   -4.7357511109689158e-002,
    2.4208498126462157e+001,
   -5.8796967594376207e-003,
    1.1999999999053516e-004,
   -1.3571262803566508e-001,
    3.0823859669312185e+001,
   -1.4299250222681243e-003,
    3.9997225214222595e-004,
   -5.5267094626872112e-002,
    2.6164091256911890e+000,
   -1.4999999998816893e-004,
   -4.9999999996056319e-005,
   -5.3697280934811922e-003,
    2.1251296078252663e-001,
   -4.3997558426350119e-004,
    1.7039459745806843e-002,
   -3.9999999996845056e-005,
    1.3499595590822043e-003,
    1.0999999999132389e-004,
   -4.0997191665036983e-004,
   -2.9999999997633791e-005,
   -1.0359283208969236e-002,
   -5.4995104298628740e-004,
   -1.0887233560054606e-001,
   -4.5995544946493360e-004,
   -1.0719047501313197e-002,
    2.3536309333839598e+000,
   -3.9999999996845056e-005,
    7.9992662643071628e-004,
    1.4762958932276996e-001,
    1.8698379332023052e-003,
    9.1193710011059047e-003,
    9.9999999992112637e-005,
    3.5997805197144798e-004,
   -9.9999999992112637e-005,
   -5.1994651011630420e-004,
   -2.9999999997633791e-005,
    2.0797755805319401e-003,
   -2.3997545787847422e-004,
    1.9999999998422528e-005,
    2.9999999997633791e-005,
   -2.9999999997633791e-005,
    7.9999999965012830e-005,
    6.4992637465305402e-004,
    2.9999999997633791e-005,
   -5.4993864658453410e-004,
   -3.9999999996845056e-005,
   -1.5998772929575679e-004,
    2.2398412595603114e-003,
   -4.9999999976938134e-005,
    1.3889003983154241e-002,
    6.9993930925565050e-004,
   -1.0479229613358202e-002,
    1.9999999998422528e-005,
   -6.6995091716154278e-004,
   -2.9999999997633791e-005,
   -1.1999999994273970e-004,
   -3.3398928496763268e-003,
   -1.3999999994116218e-004,
    5.2928221875853429e-002,
   -5.9999999976149397e-005,
   -8.8995384582716522e-004,
    2.5560112481644193e-001,
    1.4409224712398312e-002,
   -1.5707413311609472e-001,
    4.8997385523234793e-004,
   -8.8696764674312967e-003,
    2.9999999997633791e-005,
   -2.9999999986162873e-004,
   -2.9999999997633791e-005,
   -1.7000103723085354e-003,
   -4.9999999996056319e-005,
   -4.8430210614243123e-002,
   -1.2999999994195094e-004,
   -2.0599712263504460e-003,
    8.6243277191744838e-001,
   -1.0099879787956634e-003,
   -8.6798611953819403e-003,
    3.2582370457252900e+000,
   -5.9999999976149397e-005,
   -1.7999999991888909e-004,
    2.0098657627468236e-001,
   -1.4189284149249450e+000,
   -1.4999999993081436e-004,
    4.9999999976938134e-005,
    6.7098942902756283e-003,
   -6.5449921480596931e-002,
   -1.9999999998422528e-005,
    3.4999999983856690e-004,
   -1.1099974149906414e-003,
    1.5999999867893818e-004,
    2.9999999997633791e-005,
   -4.9999999996056319e-005,
   -6.4997977930409659e-004,
   -6.7997804204870567e-003,
   -1.9999999998422528e-005,
   -8.9999999954665000e-005,
    3.8098713336470323e-002,
    6.9999999965801547e-005,
   -2.2999999989582721e-004,
   -2.1258262202509592e-001,
    5.9999999976149397e-005,
   -8.0795662014067418e-003,
   -1.0765620159727074e-001,
   -1.7999999991888909e-004,
   -1.4479504321496651e-002,
   -1.4999508303474602e-003,
   -1.3999999993160309e-004,
   -6.9999999965801547e-005,
    3.1997545650636777e-004,
    9.8992810667422994e-004,
    8.9987729320396113e-005,
    1.6198834265763306e-003,
    2.7998159392136196e-004,
    2.9999999997633791e-005,
   -2.5919503791250381e-009,
   -3.9999999996845056e-005,
    9.9987729319607382e-005,
    1.9980180920396742e-005,
   -3.9999999996845056e-005,
   -1.9999999998422528e-005,
   -2.1996932339697558e-004,
   -7.7988342682777840e-004,
   -4.4993251022494053e-004,
   -5.9999999995267581e-005,
   -1.0998772935705501e-004,
   -5.1294519652774441e-003,
   -5.9999999995267581e-005,
   -1.5708296600154038e-002,
   -2.1997545871411001e-004,
   -8.1691050005996118e-003,
   -1.1398737481824078e-003,
   -1.2998772935547754e-004,
   -1.9999999998422528e-005,
   -1.1999999999053516e-004,
   -2.9999999997633791e-005,
   -2.3898333081065581e-003,
    3.9999999996845056e-005,
   -3.5997545870306773e-004,
   -1.0833240536198517e-001,
    1.0999999999132389e-004,
   -1.4498750753066776e-003,
   -2.7904005944795895e-001,
    4.9999999996056319e-005,
   -4.0296358495893600e-003,
   -1.2805074866085511e-001,
    2.2998159403045575e-004,
   -1.7278774975208779e-002,
    4.9999999996056319e-005,
   -1.8398719332559123e-003,
   -1.7998772935153386e-004,
   -3.9999999996845056e-005,
   -4.0999118702456658e-004,
   -4.9999999996056319e-005,
   -2.2999287980640212e-003,
   -5.8996644322528193e-004,
   -4.3158622600503828e-002,
    7.6996117231674355e-004,
   -6.6096631128729020e-003,
   -1.8970367440031213e+000,
    1.5999167452340915e-003,
   -5.9999999995267581e-005,
   -2.6978591089253386e-002,
   -3.9579826757930436e+000,
    6.8996644280036536e-004,
    2.9999999997633791e-005,
   -5.7886882828892962e-002,
   -1.5713943644518369e+000,
    7.9999999993690113e-005,
    4.9999999996056319e-005,
    2.7198614490906497e-003,
   -2.0285309525301440e-001,
    5.4997558342076685e-004,
   -2.0789318692135777e-002,
    5.9999999995267581e-005,
   -1.9399367441105431e-003,
   -1.6999999998659149e-004,
   -4.9999999996056319e-005,
   -5.5000518565932169e-004,
   -5.9999999995267581e-005,
   -5.4200423475317698e-003,
   -7.1999271291212379e-004,
   -2.9390184947878158e-002,
   -8.9999999992901368e-005,
   -7.4399077426542878e-003,
   -5.1423241968031275e-001,
    9.2498845670986977e-003,
   -1.3999999998895766e-004,
   -7.9818881687893625e-002,
   -2.1636343445651161e+001,
    1.4829797329931681e-002,
   -7.9997291401692023e-004,
   -3.2175507677622006e-001,
   -3.4782523344715479e+001,
    5.4999214095694041e-003,
    3.1998839208079367e-004,
   -5.0792132716165672e-001,
   -1.1649947731911631e+001,
    6.6999271333309670e-004,
    4.9998452276860735e-004,
    1.9379704871355423e-002,
   -1.4225541054826600e+000,
    6.9999999994478830e-005,
    3.9999450408357258e-003,
   -1.3922061526530222e-001,
    4.3999485463769793e-004,
   -1.2430067394388417e-002,
    3.9999999996845056e-005,
   -1.0600077772193598e-003,
   -8.9999999992901368e-005,
   -3.9999999996845056e-005,
    9.2996923689824906e-004,
   -2.3989246527191473e-002,
   -6.9999999994478830e-005,
    3.5998773142248240e-004,
    2.3695196220855216e-001,
   -2.1999999998264779e-004,
    1.5509193615448615e-002,
    5.7878979678519837e-001,
   -1.1999999999053516e-004,
    1.9688939362482304e-002,
    2.4452135952649218e-001,
   -2.9999999997633791e-005,
   -2.9999999997633791e-005,
    1.2999999873865894e-004,
    3.6818750368665838e-002,
   -7.9999999993690113e-005,
    4.2998577586114287e-003,
    4.4998418603039634e-004,
    3.9999999996845056e-005,
   -2.9999999997633791e-005,
    1.9999999998422528e-005,
    2.5198280351168174e-003,
    2.0998159403203323e-004,
    2.6578132750592169e-002,
   -6.3993930785036192e-004,
    1.8678664531883406e-002,
    4.9999999996056319e-005,
    1.3599096100364008e-003,
    2.9999999997633791e-005,
   -1.4998159403676562e-004,
   -1.9997545871568750e-004,
   -2.9999999997633791e-005,
    2.9999999997633791e-005,
    8.9999999964224085e-005,
    4.3996932196603133e-004,
    1.9999999998422528e-005,
   -5.9999999976149397e-005,
   -1.5998772929575679e-004,
   -1.9999999998422528e-005,
    2.9999999997633791e-005,
    1.2799569653413907e-003,
    2.9999999997633791e-005,
    5.7298104005903266e-003,
    3.3998772921464590e-004,
   -5.5997373255061008e-004,
    2.0999999990696381e-004,
   -1.5299447789493925e-003,
    2.9999999997633791e-005,
   -1.8999999991810035e-004,
    2.6000172872948869e-004,
    3.9999999996845056e-005,
    1.1490054433022383e-002,
   -1.9999999998422528e-005,
    2.9999999986162873e-004,
    4.5910183303345259e-002,
    2.8099578851042317e-003,
   -2.4799487254914118e-003,
    1.4199767806551116e-003,
   -8.1399939349424515e-003,
    1.9999999990775254e-004,
   -8.8999999959523410e-004,
   -5.9999999976149397e-005,
    2.9999999997633791e-005,
    3.7998945636984291e-004,
   -2.5599166103938880e-003,
   -1.4999999993081436e-004,
   -3.1098903362823788e-003,
   -8.9999999964224085e-005,
   -1.1099596426306444e-003,
   -1.9999999990775254e-004,
   -2.9999999997633791e-005,
    6.9999999965801547e-005,
    2.9999999997633791e-005,
   -2.9999999997633791e-005,
   -2.9999999997633791e-005,
   -1.1998772935626628e-004,
   -4.3995091617713257e-004,
   -4.2995091576089210e-004,
   -1.6998159403518818e-004,
   -2.9999999997633791e-005,
   -6.9999999994478830e-005,
   -1.9498683911984820e-003,
   -2.9999999997633791e-005,
   -6.1795667138423707e-003,
   -1.0999999999132389e-004,
   -5.4096132875293794e-003,
   -4.9999999996056319e-005,
   -1.9398623381545711e-003,
   -3.4997545745276901e-004,
   -4.9999999996056319e-005,
   -8.4997450738187076e-004,
   -2.9999999997633791e-005,
   -2.4359225285690501e-002,
    3.9999999996845056e-005,
   -4.0998772933339289e-004,
   -6.6777798173381114e-002,
    3.9999999996845056e-005,
   -1.2099394321210598e-003,
   -5.1908238367859794e-002,
   -4.8997385374585406e-004,
   -1.7089416576075510e-002,
   -2.9598990618660836e-003,
   -3.8998945651244057e-004,
   -4.9999999996056319e-005,
   -8.9999999992901368e-005,
   -6.5900414730079944e-003,
    1.1999999999053516e-004,
   -2.6999999997870410e-004,
   -1.8316094280828835e-001,
    2.7999999997791532e-004,
   -3.2599561972231688e-003,
   -4.2250164916974597e-001,
    1.9999999998422528e-004,
   -7.8498757930982599e-003,
   -2.8663087332152248e-001,
    5.9999999995267581e-005,
   -3.9999999996845056e-005,
   -2.6999535602876198e-003,
   -8.5870256485955698e-002,
   -1.4200047492090133e-002,
    3.9999999996845056e-005,
   -1.8200086374119174e-003,
   -1.9999999998422528e-004,
   -1.9999999998422528e-005,
   -2.3999999831295370e-004,
    2.7799086682335217e-003,
    1.8999999998501401e-004,
    9.4796822552229501e-003,
    4.1998772808151676e-004,
    7.7997306363464932e-003,
    1.5999999998738023e-004,
    2.7199078769132544e-003,
    5.1998332118505263e-004,
    7.9999999993690113e-005,
    3.9999999996845056e-005,
    3.3997718629914461e-004,
    4.3996932171150673e-004,
    1.3999999998895766e-004,
    1.9999999998422528e-005,
    1.0999999995308753e-004,
    5.9999999976149397e-005,
   -1.9999999998422528e-005,
    1.4999999993081436e-004,
    6.6000345530543478e-004,
    3.9999999996845056e-005,
    3.1999999818237558e-004,
    3.9999999996845056e-005,
   -7.9999999965012830e-005,
   -6.9999999965801547e-005,
   -2.9999999997633791e-005,
   -6.9999999965801547e-005,
   -3.9999999996845056e-005,
   -2.9999999997633791e-005,
   -1.2999999998974642e-004,
   -1.5998772935311135e-004,
   -9.9999999992112637e-005,
   -2.9999999997633791e-005,
   -3.2998945693420216e-004,
   -1.0899622350605558e-003,
   -1.9999999998422528e-005,
   -1.2599560988959544e-003,
   -1.9999999998422528e-005,
   -6.9997804860839755e-004,
   -2.1999999914858946e-004,
   -3.9999999996845056e-005,
   -8.9999999992901368e-005,
   -1.8400112315127785e-003,
   -3.9999999996845056e-005,
   -5.3900267774981249e-003,
   -1.1999999999053516e-004,
   -5.5900215880263860e-003,
   -8.9999999992901368e-005,
   -2.8500094963320923e-003,
   -1.9999999998422528e-005,
   -8.3000258944965276e-004,
   -1.5999999998738023e-004,
   -1.9999999998422528e-005,
    3.9999999996845056e-005,
    1.4999999998816893e-004,
    1.8999999998501401e-004,
    1.0999999999132389e-004,
    3.9999999996845056e-005,
   -2.9999999997633791e-005,
   -1.9999999998422528e-005,
   -1.9999999998422528e-005,
   -6.9999999994478830e-005,
   -9.9999999992112637e-005,
   -6.9999999994478830e-005,
   -2.9999999997633791e-005
} };

constexpr array<double, 704> g_MAIN_S3_DE406
{ {
    3.8500052903228404e+005,
   -3.1483021922752781e+000,
   -2.9999999997633791e-005,
    3.7998579755659352e-004,
    7.9660570476435268e+001,
   -2.0905354935205090e+004,
   -1.0325754836510391e-001,
   -8.3697366020164584e-003,
   -4.4211771171071561e+000,
   -5.6992511533509469e+002,
   -1.0199309130843224e-003,
    3.2999012092559170e-004,
   -1.1100043220264174e-003,
   -2.3210429822877586e+001,
    2.6000518651625334e-004,
    1.9999999998422528e-005,
    2.7000172882429092e-004,
   -1.1169370611059974e+000,
    3.9999999996845056e-005,
    1.9999999998422528e-005,
   -5.8880462467805837e-002,
   -3.2900311189610690e-003,
   -1.9000172883060085e-004,
   -1.0999999999132389e-004,
   -1.7399524762050224e-003,
   -2.6739186021817699e-002,
   -4.2240654624048563e-001,
    1.4999999998816893e-004,
   -1.8999999998501401e-004,
   -7.0026958822396903e+000,
   -5.0197444683412334e-002,
   -6.9999999994478830e-005,
   -7.6895896701093900e-003,
   -1.2962022217205058e+002,
    3.3463244089062189e-001,
    2.9999999997633791e-005,
   -1.8566992299900853e-001,
    4.8888328059174846e+001,
   -1.5802147102879513e-001,
    2.9999999997633791e-005,
   -2.4809697845850509e-001,
    1.0475529267427028e+002,
   -6.8196431963242548e-003,
    4.9999999996056319e-005,
    3.2648340142780188e-002,
    5.7508577057015726e+000,
   -2.9999999997633786e-004,
   -7.9995730218233028e-004,
    3.5507868338050058e-001,
   -5.9999999995267581e-005,
    2.3039298527847123e-002,
    1.5299568826931335e-003,
    9.9999999992112637e-005,
   -4.9999999996056319e-005,
   -6.5995523869077661e-004,
   -9.3793569172097267e-003,
   -1.3617045669341818e-001,
   -9.8991435536443227e-004,
   -3.0997545870701143e-004,
   -2.1171282169327208e+000,
    4.4995982169380271e-003,
   -5.5794921950237464e-003,
    1.0656722650829162e+000,
   -3.4196904093387018e-003,
   -2.3697891063466005e-003,
    1.1655363585022665e+000,
   -1.0999999999132389e-004,
    1.7998772935153386e-004,
    5.0716436611946462e-002,
    2.2798455837538525e-003,
    8.9999999992901368e-005,
   -2.0997545871489877e-004,
   -2.8696957386574184e-003,
   -1.9999999998422528e-005,
   -3.8935784344622658e-002,
    6.9999999994478830e-005,
   -1.6998159403518818e-004,
    2.1267663221308263e-002,
   -6.9999999994478830e-005,
   -2.9999999997633791e-005,
    1.9117925823248456e-002,
    7.7991669876697665e-004,
    2.9999999997633791e-005,
   -5.9999999995267581e-005,
   -7.2989569871707838e-004,
    3.3995091743610744e-004,
    3.6994478211660676e-004,
   -5.9999999976149397e-005,
    3.4996318794531937e-004,
   -1.6098220750738919e-003,
    3.9999999996845056e-005,
    2.8997545860343885e-004,
    8.4393950201855413e-003,
   -2.3998159394363517e-004,
   -2.9477842955863663e-002,
   -2.9999999997633791e-005,
   -1.9999999998422528e-005,
    1.5898860148522863e-003,
    3.5997545731815292e-004,
    3.9999999996845056e-005,
    9.9999999953876269e-005,
    6.4997891613239060e-004,
    1.9999999998422528e-005,
   -4.2198517073415813e-003,
   -7.9999999965012830e-005,
   -1.0999999995308753e-004,
    1.1410603535319587e-001,
    2.0999999990696381e-004,
   -1.3659260861906745e-002,
    4.9755215400672276e-001,
   -7.3996384925082781e-004,
    4.3997385525540976e-004,
    1.0997620731914327e-001,
   -6.9999999965801547e-005,
   -2.9999999997633791e-005,
    1.1209625212771860e-002,
    9.6996750810329085e-004,
    7.9999999965012830e-005,
   -2.9999999997633791e-005,
   -4.3000345749475338e-004,
   -6.3600397530792401e-003,
   -1.8999999991810035e-004,
   -9.9380481230041715e-002,
   -4.7999485445292023e-004,
   -2.7799622097322992e-003,
   -1.7385258128996706e+000,
    4.5699364414918571e-003,
   -4.2319342957746767e-002,
   -8.3791066502014502e+000,
   -2.4709621847278977e-002,
   -7.9562638278438003e-001,
    1.0874270127246304e+002,
   -1.8739690056048076e-002,
    1.6999999992923693e-004,
    6.0549065378075952e-002,
    6.3220031516918853e+000,
   -1.6399763697789705e-003,
    2.0299712263623713e-003,
    3.7852141276755424e-001,
   -1.2999999994195094e-004,
    1.0999999995308753e-004,
    2.3680120083704129e-002,
    1.5200095067157658e-003,
    9.9999999953876269e-005,
    5.9999999976149397e-005,
    1.0099692386042803e-003,
    1.7289444030638344e-002,
    1.3999999994116218e-004,
    7.9999999965012830e-005,
    3.3224888279882392e-001,
   -1.4799220185026741e-003,
    1.5599202896221002e-003,
    8.5124033619353401e-001,
    3.3098244400641009e-003,
    3.3788197501593308e-002,
   -1.6674723746336660e+001,
   -5.9999999976149397e-005,
   -5.6197041728912800e-003,
   -9.3331755043545517e-001,
   -3.6998772836866227e-004,
   -5.9038041580494902e-002,
   -1.9999999998422528e-005,
   -3.9198779032451961e-003,
   -2.7000172872869996e-004,
    1.9999999998422528e-005,
    2.7998159392136196e-004,
    3.2397710884068728e-003,
    4.3796878710824355e-003,
   -1.9999999998422528e-005,
    3.7287275130270789e-002,
    1.9999999998422528e-005,
    6.4595388101637554e-003,
    6.6995351042992298e-004,
    5.9999999976149397e-005,
    5.9999999976149397e-005,
    7.9999999965012830e-005,
    5.3993864658532283e-004,
    6.9999999965801547e-005,
   -1.9999999998422528e-005,
   -1.1997545872199740e-004,
   -5.7989569831188030e-004,
   -3.9999999996845056e-005,
    3.5994478211739549e-004,
   -4.2593811093623989e-003,
    2.9999999997633791e-005,
   -1.5777680821354629e-002,
   -1.1298360731086838e-003,
   -8.9987729358632481e-005,
    1.9999999998422528e-005,
    4.5995264627222943e-004,
    1.1158808366184655e-002,
    2.9999999997633791e-005,
    2.5996932339382065e-004,
   -1.4806393807175231e-001,
    6.9999999994478830e-005,
    1.0398640811513848e-003,
   -4.1071485480076358e-001,
    2.6996932339303192e-004,
   -2.8546907073008252e-002,
    2.9999999997633791e-005,
   -2.0897737689961008e-003,
   -1.5998159403597692e-004,
    2.9999999997633791e-005,
    6.6995523952404629e-004,
    1.9999999998422528e-005,
    1.5168961243902334e-002,
    4.0996932338198958e-004,
    3.4301591248131014e-001,
    9.8991435661551982e-004,
    9.5691427477273029e-003,
   -4.9501348002130729e+000,
    1.6798497951639492e-003,
    2.8387413047911467e-002,
   -9.8844666217709616e+000,
   -2.9999999997633791e-005,
    7.9192899476857939e-003,
   -6.5753295021873692e-001,
    8.3992576200476792e-004,
   -4.6076771497468638e-002,
    5.9999999995267581e-005,
   -3.3497710052450623e-003,
   -2.4998332287446507e-004,
   -1.9999999998422528e-005,
    5.9999999995267581e-005,
    1.1899665613369270e-003,
    3.9999999996845056e-005,
    2.3339287993958329e-002,
    7.8995730385123561e-004,
    4.9504416423055558e-001,
   -4.5997558426192373e-004,
    1.5179225244421566e-002,
    1.0056203077190446e+001,
    3.0238458352181256e-002,
    3.2334304986755269e-001,
   -1.5213780942589074e+002,
    3.7338055220178724e-002,
   -3.4997612144415849e-004,
    6.5706432249497604e-001,
   -2.0458611646087306e+002,
    1.0799619981539893e-003,
   -8.4993836896337369e-004,
    2.0940900061519147e-001,
   -1.2831403332461623e+001,
    1.7288455868366050e-009,
    1.2999999998974642e-004,
    1.8039081008914867e-002,
   -8.4880182533646864e-001,
    1.2699337512832878e-003,
   -5.8498146534920067e-002,
    8.9999999992901368e-005,
   -4.1298734075970499e-003,
   -3.0000259324471811e-004,
   -1.9999999998422528e-005,
    2.0000259325260553e-004,
    3.0400285232102318e-003,
    1.1999999999053516e-004,
    4.7400373388124431e-002,
   -1.9999999998422528e-005,
    1.8799811461358263e-003,
    7.7854492106106743e-001,
    4.7999571822327815e-004,
    2.9099632684995498e-002,
    1.4402688569714263e+001,
   -2.0079749835047431e-002,
   -1.7999999998580274e-004,
   -1.9999999998422528e-005,
    4.7262331352426273e-001,
    2.4615847485355792e+002,
    7.7403905203800960e-001,
   -1.3999999998895766e-004,
   -3.8998452277728355e-004,
    8.7515627423090212e+000,
   -3.6991108963980764e+003,
    5.9632094424355830e-001,
   -7.4497493960877454e-003,
    1.0321107477029065e+001,
   -2.9559675579724144e+003,
    3.1429705179098821e-002,
   -1.8659382347128925e-002,
    4.1311146504675742e+000,
   -1.7073307712477057e+002,
    1.8099967120070663e-003,
    2.1899295292498355e-003,
    2.8398596703683637e-001,
   -1.0444757567465345e+001,
    1.4000259325733791e-004,
    1.9999999998422528e-005,
    1.8109768286069500e-002,
   -6.6968349796415183e-001,
    1.1899884332213365e-003,
   -4.4220290390859181e-002,
    7.9999999993690113e-005,
   -2.9800242002669388e-003,
   -2.0000172882981212e-004,
    8.9999999992901368e-005,
    1.1599665580030599e-003,
    2.9999999997633791e-005,
    1.4169567487434505e-002,
    2.0999999914937822e-004,
    1.6857464438507142e-001,
    2.9998772934206898e-004,
    5.1997384415181720e-004,
    1.4367546693920624e-001,
   -2.0298946113646351e-003,
   -4.7357511109689158e-002,
    2.4208498126462157e+001,
   -5.8796967594376207e-003,
    1.1999999999053516e-004,
   -1.3571262803566508e-001,
    3.0823859669312185e+001,
   -1.4299250222681243e-003,
    3.9997225214222595e-004,
   -5.5267094626872112e-002,
    2.6164091256911890e+000,
   -1.4999999998816893e-004,
   -4.9999999996056319e-005,
   -5.3697280934811922e-003,
    2.1251296078252663e-001,
   -4.3997558426350119e-004,
    1.7039459745806843e-002,
   -3.9999999996845056e-005,
    1.3499595590822043e-003,
    1.0999999999132389e-004,
   -4.0997191665036983e-004,
   -2.9999999997633791e-005,
   -1.0359283208969236e-002,
   -5.4995104298628740e-004,
   -1.0887233560054606e-001,
   -4.5995544946493360e-004,
   -1.0719047501313197e-002,
    2.3536309333839598e+000,
   -3.9999999996845056e-005,
    7.9992662643071628e-004,
    1.4762958932276996e-001,
    1.8698379332023052e-003,
    9.1193710011059047e-003,
    9.9999999992112637e-005,
    3.5997805197144798e-004,
   -9.9999999992112637e-005,
   -5.1994651011630420e-004,
   -2.9999999997633791e-005,
    2.0797755805319401e-003,
   -2.3997545787847422e-004,
    1.9999999998422528e-005,
    2.9999999997633791e-005,
   -2.9999999997633791e-005,
    7.9999999965012830e-005,
    6.4992637465305402e-004,
    2.9999999997633791e-005,
   -5.4993864658453410e-004,
   -3.9999999996845056e-005,
   -1.5998772929575679e-004,
    2.2398412595603114e-003,
   -4.9999999976938134e-005,
    1.3889003983154241e-002,
    6.9993930925565050e-004,
   -1.0479229613358202e-002,
    1.9999999998422528e-005,
   -6.6995091716154278e-004,
   -2.9999999997633791e-005,
   -1.1999999994273970e-004,
   -3.3398928496763268e-003,
   -1.3999999994116218e-004,
    5.2928221875853429e-002,
   -5.9999999976149397e-005,
   -8.8995384582716522e-004,
    2.5560112481644193e-001,
    1.4409224712398312e-002,
   -1.5707413311609472e-001,
    4.8997385523234793e-004,
   -8.8696764674312967e-003,
    2.9999999997633791e-005,
   -2.9999999986162873e-004,
   -2.9999999997633791e-005,
   -1.7000103723085354e-003,
   -4.9999999996056319e-005,
   -4.8430210614243123e-002,
   -1.2999999994195094e-004,
   -2.0599712263504460e-003,
    8.6243277191744838e-001,
   -1.0099879787956634e-003,
   -8.6798611953819403e-003,
    3.2582370457252900e+000,
   -5.9999999976149397e-005,
   -1.7999999991888909e-004,
    2.0098657627468236e-001,
   -1.4189284149249450e+000,
   -1.4999999993081436e-004,
    4.9999999976938134e-005,
    6.7098942902756283e-003,
   -6.5449921480596931e-002,
   -1.9999999998422528e-005,
    3.4999999983856690e-004,
   -1.1099974149906414e-003,
    1.5999999867893818e-004,
    2.9999999997633791e-005,
   -4.9999999996056319e-005,
   -6.4997977930409659e-004,
   -6.7997804204870567e-003,
   -1.9999999998422528e-005,
   -8.9999999954665000e-005,
    3.8098713336470323e-002,
    6.9999999965801547e-005,
   -2.2999999989582721e-004,
   -2.1258262202509592e-001,
    5.9999999976149397e-005,
   -8.0795662014067418e-003,
   -1.0765620159727074e-001,
   -1.7999999991888909e-004,
   -1.4479504321496651e-002,
   -1.4999508303474602e-003,
   -1.3999999993160309e-004,
   -6.9999999965801547e-005,
    3.1997545650636777e-004,
    9.8992810667422994e-004,
    8.9987729320396113e-005,
    1.6198834265763306e-003,
    2.7998159392136196e-004,
    2.9999999997633791e-005,
   -2.5919503791250381e-009,
   -3.9999999996845056e-005,
    9.9987729319607382e-005,
    1.9980180920396742e-005,
   -3.9999999996845056e-005,
   -1.9999999998422528e-005,
   -2.1996932339697558e-004,
   -7.7988342682777840e-004,
   -4.4993251022494053e-004,
   -5.9999999995267581e-005,
   -1.0998772935705501e-004,
   -5.1294519652774441e-003,
   -5.9999999995267581e-005,
   -1.5708296600154038e-002,
   -2.1997545871411001e-004,
   -8.1691050005996118e-003,
   -1.1398737481824078e-003,
   -1.2998772935547754e-004,
   -1.9999999998422528e-005,
   -1.1999999999053516e-004,
   -2.9999999997633791e-005,
   -2.3898333081065581e-003,
    3.9999999996845056e-005,
   -3.5997545870306773e-004,
   -1.0833240536198517e-001,
    1.0999999999132389e-004,
   -1.4498750753066776e-003,
   -2.7904005944795895e-001,
    4.9999999996056319e-005,
   -4.0296358495893600e-003,
   -1.2805074866085511e-001,
    2.2998159403045575e-004,
   -1.7278774975208779e-002,
    4.9999999996056319e-005,
   -1.8398719332559123e-003,
   -1.7998772935153386e-004,
   -3.9999999996845056e-005,
   -4.0999118702456658e-004,
   -4.9999999996056319e-005,
   -2.2999287980640212e-003,
   -5.8996644322528193e-004,
   -4.3158622600503828e-002,
    7.6996117231674355e-004,
   -6.6096631128729020e-003,
   -1.8970367440031213e+000,
    1.5999167452340915e-003,
   -5.9999999995267581e-005,
   -2.6978591089253386e-002,
   -3.9579826757930436e+000,
    6.8996644280036536e-004,
    2.9999999997633791e-005,
   -5.7886882828892962e-002,
   -1.5713943644518369e+000,
    7.9999999993690113e-005,
    4.9999999996056319e-005,
    2.7198614490906497e-003,
   -2.0285309525301440e-001,
    5.4997558342076685e-004,
   -2.0789318692135777e-002,
    5.9999999995267581e-005,
   -1.9399367441105431e-003,
   -1.6999999998659149e-004,
   -4.9999999996056319e-005,
   -5.5000518565932169e-004,
   -5.9999999995267581e-005,
   -5.4200423475317698e-003,
   -7.1999271291212379e-004,
   -2.9390184947878158e-002,
   -8.9999999992901368e-005,
   -7.4399077426542878e-003,
   -5.1423241968031275e-001,
    9.2498845670986977e-003,
   -1.3999999998895766e-004,
   -7.9818881687893625e-002,
   -2.1636343445651161e+001,
    1.4829797329931681e-002,
   -7.9997291401692023e-004,
   -3.2175507677622006e-001,
   -3.4782523344715479e+001,
    5.4999214095694041e-003,
    3.1998839208079367e-004,
   -5.0792132716165672e-001,
   -1.1649947731911631e+001,
    6.6999271333309670e-004,
    4.9998452276860735e-004,
    1.9379704871355423e-002,
   -1.4225541054826600e+000,
    6.9999999994478830e-005,
    3.9999450408357258e-003,
   -1.3922061526530222e-001,
    4.3999485463769793e-004,
   -1.2430067394388417e-002,
    3.9999999996845056e-005,
   -1.0600077772193598e-003,
   -8.9999999992901368e-005,
   -3.9999999996845056e-005,
    9.2996923689824906e-004,
   -2.3989246527191473e-002,
   -6.9999999994478830e-005,
    3.5998773142248240e-004,
    2.3695196220855216e-001,
   -2.1999999998264779e-004,
    1.5509193615448615e-002,
    5.7878979678519837e-001,
   -1.1999999999053516e-004,
    1.9688939362482304e-002,
    2.4452135952649218e-001,
   -2.9999999997633791e-005,
   -2.9999999997633791e-005,
    1.2999999873865894e-004,
    3.6818750368665838e-002,
   -7.9999999993690113e-005,
    4.2998577586114287e-003,
    4.4998418603039634e-004,
    3.9999999996845056e-005,
   -2.9999999997633791e-005,
    1.9999999998422528e-005,
    2.5198280351168174e-003,
    2.0998159403203323e-004,
    2.6578132750592169e-002,
   -6.3993930785036192e-004,
    1.8678664531883406e-002,
    4.9999999996056319e-005,
    1.3599096100364008e-003,
    2.9999999997633791e-005,
   -1.4998159403676562e-004,
   -1.9997545871568750e-004,
   -2.9999999997633791e-005,
    2.9999999997633791e-005,
    8.9999999964224085e-005,
    4.3996932196603133e-004,
    1.9999999998422528e-005,
   -5.9999999976149397e-005,
   -1.5998772929575679e-004,
   -1.9999999998422528e-005,
    2.9999999997633791e-005,
    1.2799569653413907e-003,
    2.9999999997633791e-005,
    5.7298104005903266e-003,
    3.3998772921464590e-004,
   -5.5997373255061008e-004,
    2.0999999990696381e-004,
   -1.5299447789493925e-003,
    2.9999999997633791e-005,
   -1.8999999991810035e-004,
    2.6000172872948869e-004,
    3.9999999996845056e-005,
    1.1490054433022383e-002,
   -1.9999999998422528e-005,
    2.9999999986162873e-004,
    4.5910183303345259e-002,
    2.8099578851042317e-003,
   -2.4799487254914118e-003,
    1.4199767806551116e-003,
   -8.1399939349424515e-003,
    1.9999999990775254e-004,
   -8.8999999959523410e-004,
   -5.9999999976149397e-005,
    2.9999999997633791e-005,
    3.7998945636984291e-004,
   -2.5599166103938880e-003,
   -1.4999999993081436e-004,
   -3.1098903362823788e-003,
   -8.9999999964224085e-005,
   -1.1099596426306444e-003,
   -1.9999999990775254e-004,
   -2.9999999997633791e-005,
    6.9999999965801547e-005,
    2.9999999997633791e-005,
   -2.9999999997633791e-005,
   -2.9999999997633791e-005,
   -1.1998772935626628e-004,
   -4.3995091617713257e-004,
   -4.2995091576089210e-004,
   -1.6998159403518818e-004,
   -2.9999999997633791e-005,
   -6.9999999994478830e-005,
   -1.9498683911984820e-003,
   -2.9999999997633791e-005,
   -6.1795667138423707e-003,
   -1.0999999999132389e-004,
   -5.4096132875293794e-003,
   -4.9999999996056319e-005,
   -1.9398623381545711e-003,
   -3.4997545745276901e-004,
   -4.9999999996056319e-005,
   -8.4997450738187076e-004,
   -2.9999999997633791e-005,
   -2.4359225285690501e-002,
    3.9999999996845056e-005,
   -4.0998772933339289e-004,
   -6.6777798173381114e-002,
    3.9999999996845056e-005,
   -1.2099394321210598e-003,
   -5.1908238367859794e-002,
   -4.8997385374585406e-004,
   -1.7089416576075510e-002,
   -2.9598990618660836e-003,
   -3.8998945651244057e-004,
   -4.9999999996056319e-005,
   -8.9999999992901368e-005,
   -6.5900414730079944e-003,
    1.1999999999053516e-004,
   -2.6999999997870410e-004,
   -1.8316094280828835e-001,
    2.7999999997791532e-004,
   -3.2599561972231688e-003,
   -4.2250164916974597e-001,
    1.9999999998422528e-004,
   -7.8498757930982599e-003,
   -2.8663087332152248e-001,
    5.9999999995267581e-005,
   -3.9999999996845056e-005,
   -2.6999535602876198e-003,
   -8.5870256485955698e-002,
   -1.4200047492090133e-002,
    3.9999999996845056e-005,
   -1.8200086374119174e-003,
   -1.9999999998422528e-004,
   -1.9999999998422528e-005,
   -2.3999999831295370e-004,
    2.7799086682335217e-003,
    1.8999999998501401e-004,
    9.4796822552229501e-003,
    4.1998772808151676e-004,
    7.7997306363464932e-003,
    1.5999999998738023e-004,
    2.7199078769132544e-003,
    5.1998332118505263e-004,
    7.9999999993690113e-005,
    3.9999999996845056e-005,
    3.3997718629914461e-004,
    4.3996932171150673e-004,
    1.3999999998895766e-004,
    1.9999999998422528e-005,
    1.0999999995308753e-004,
    5.9999999976149397e-005,
   -1.9999999998422528e-005,
    1.4999999993081436e-004,
    6.6000345530543478e-004,
    3.9999999996845056e-005,
    3.1999999818237558e-004,
    3.9999999996845056e-005,
   -7.9999999965012830e-005,
   -6.9999999965801547e-005,
   -2.9999999997633791e-005,
   -6.9999999965801547e-005,
   -3.9999999996845056e-005,
   -2.9999999997633791e-005,
   -1.2999999998974642e-004,
   -1.5998772935311135e-004,
   -9.9999999992112637e-005,
   -2.9999999997633791e-005,
   -3.2998945693420216e-004,
   -1.0899622350605558e-003,
   -1.9999999998422528e-005,
   -1.2599560988959544e-003,
   -1.9999999998422528e-005,
   -6.9997804860839755e-004,
   -2.1999999914858946e-004,
   -3.9999999996845056e-005,
   -8.9999999992901368e-005,
   -1.8400112315127785e-003,
   -3.9999999996845056e-005,
   -5.3900267774981249e-003,
   -1.1999999999053516e-004,
   -5.5900215880263860e-003,
   -8.9999999992901368e-005,
   -2.8500094963320923e-003,
   -1.9999999998422528e-005,
   -8.3000258944965276e-004,
   -1.5999999998738023e-004,
   -1.9999999998422528e-005,
    3.9999999996845056e-005,
    1.4999999998816893e-004,
    1.8999999998501401e-004,
    1.0999999999132389e-004,
    3.9999999996845056e-005,
   -2.9999999997633791e-005,
   -1.9999999998422528e-005,
   -1.9999999998422528e-005,
   -6.9999999994478830e-005,
   -9.9999999992112637e-005,
   -6.9999999994478830e-005,
   -2.9999999997633791e-005
} };

struct AAPLUS_EXT_CLASS ELPMPP02PertubationsCoefficient
{
  double m_S;
  double m_C;
  array<int, 13> m_I;
};

constexpr array<ELPMPP02PertubationsCoefficient, 11314> g_PERT_S1_1
{ {
 { -1.274921554086e+01,  6.368794709728e+00, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.062989999049e+00,  1.158760846981e-04, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.142992395166e+00, -2.364916989220e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.363518832550e-01, -8.436259830515e-01, {   0,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.052490932382e-01,  3.523015957469e-01, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.616042701633e-01,  3.305118606429e-01, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.216317272874e-01, -1.489776891248e-04, {   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.199368388052e-01, -2.209768624009e-01, {   0,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.386900454119e-01,  1.372171387329e-02, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.437493342238e-01,  1.506773061910e-04, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.384626308067e-01,  7.779536316866e-02, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.932999998788e-01,  1.093433068322e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.914099998793e-01,  1.089243754518e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.417546337192e-01, -2.166115750689e-01, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.605999999723e-01,  4.468496670597e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.880531057061e-04, -3.435256417406e-01, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.785532521052e-01,  4.925556417323e-02, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.274356135829e-01, -1.874226085810e-01, {   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.369216184127e-02,  2.842294293431e-01, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.015549263297e-01,  1.069270687790e-03, {   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.451570248624e-01,  3.653652585652e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  8.465265199602e-02, -1.407384711803e-01, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.494596370263e-01,  7.465967763680e-02, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.546580467031e-02, -1.752828660520e-01, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.545764455406e-02, -1.752578613376e-01, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.470981586752e-01,  7.348183257675e-02, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.458710921404e-01,  7.286925722753e-02, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.109925713502e-01,  6.424260088501e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.405585638153e-01,  7.021552095395e-02, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.944610283953e-01, -7.991843079482e-04, {   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.418078290757e-01, -5.275582672188e-02, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.846256502334e-01, -1.400871323511e-03, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.825375808697e-01, -4.675890056110e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  3.032898165783e-03, -1.750854674915e-01, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.978036373875e-02, -1.310814746794e-01, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.974077519511e-02, -1.309644515344e-01, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.648388533380e-01,  5.196879287529e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.608349137199e-01,  8.736000213433e-04, {   2,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.537931154659e-01, -3.273344979541e-05, {   0,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.435742094759e-01,  3.108751894607e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.360027654208e-01,  3.025021139390e-06, {   2,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.342701214205e-01,  8.561989583470e-04, {   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.312406745697e-01,  3.886344472925e-05, {   2,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.702253773839e-02,  1.014600539612e-01, {   0,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.274624250209e-01, -3.437082514475e-05, {   0,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.610645230810e-02, -7.936456566333e-02, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.251817003712e-01,  2.722014929245e-04, {   0,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.100464286193e-01,  2.129256270728e-03, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.531871446849e-02, -8.608593674398e-02, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.528677204472e-02, -8.599685296556e-02, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  9.369899207918e-02, -1.679495089593e-02, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.363503514059e-03, -1.062435442293e-01, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.470223886570e-02, -8.272085381316e-02, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  6.646652705243e-04, -1.040473622080e-01, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.978857630880e-02, -2.203792887334e-02, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.928554532669e-02,  2.802063301635e-05, {   2,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.641999998810e-02,  1.514561818233e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.485028810547e-02,  1.699553722034e-02, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.408756234229e-02, -4.651768746219e-02, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  8.908541081533e-02, -1.336437809973e-03, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -4.336157307801e-02, -4.603357302274e-02, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  8.402334137698e-02, -1.398976338058e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.570754518098e-02, -1.869023812477e-02, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  6.569673491537e-02, -1.868821484668e-02, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -8.239894315813e-02, -1.721815563775e-05, {   2,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.122726252611e-02,  2.814400567569e-02, {   1,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -6.114893815787e-02,  1.780364764413e-02, {   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -7.876249221527e-03,  6.707600620165e-02, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.207328031647e-02,  5.428618910529e-05, {   2,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.791099469573e-02,  2.393356012740e-02, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.797746905209e-02,  1.324483057208e-02, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.866383574899e-02, -1.517751167014e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.712772520604e-02, -8.354750054757e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.494897587386e-02,  2.245473573767e-02, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.516600831523e-03,  6.120202714372e-02, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.455211720422e-03,  6.058496617670e-02, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.568999998999e-02,  1.146506785577e-06, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.468498373457e-04, -6.519920954471e-02, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.455999999827e-02,  4.732495173325e-07, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.047544005549e-02, -2.327078047287e-02, {   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.682706748336e-02, -3.691027051060e-02, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.691051323941e-03, -6.178281619898e-02, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.357694221146e-03, -6.053659098999e-02, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.902765988652e-02, -1.155517007587e-03, {   2,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -5.079183155517e-02, -9.337382031453e-03, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.762017031905e-02,  2.022520185509e-02, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -5.754285635716e-02, -1.678488387368e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.739991698701e-02,  7.702973650719e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.379641022102e-03, -5.552670732290e-02, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.615698113287e-02,  3.914464309553e-04, {   0,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.261615047610e-02, -3.520688147008e-03, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.107533132410e-03, -5.267517899506e-02, {   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.484750387190e-02, -1.987854091654e-02, {   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.314967669127e-02, -3.138805130984e-02, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.311868646981e-02, -3.133649579317e-02, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.062373144859e-02,  4.350707132485e-02, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.709977235023e-02, -3.673154452032e-02, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.627017127291e-02, -3.706102997884e-02, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.051798788395e-02, -3.280307775676e-02, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.036987170467e-02, -3.254858972710e-02, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.947693818530e-02, -3.242643906379e-02, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.148866084833e-02, -1.038000451824e-02, {   1,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.095532907326e-02,  5.019015002132e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.058366657776e-02,  1.336495407700e-04, {   0,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.034999999816e-02,  4.305989064005e-07, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.794200947922e-02, -2.184609103067e-02, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.962999999936e-02,  2.512000032507e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.308702947172e-02, -2.606318431933e-02, {   2,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.045572614144e-02, -2.811979642778e-02, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -4.745999999582e-02,  6.295332798589e-07, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.611840004175e-02, -1.142340409850e-03, {   2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -5.052842505661e-04, -4.658591251810e-02, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.632067236920e-02, -6.036511903310e-05, {   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.773140478057e-02, -2.798942856976e-02, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.478399092975e-02, -2.012820293857e-02, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.771521445481e-02,  1.718706918215e-02, {   0,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -4.341002185668e-02,  6.818054081860e-04, {   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.048108454429e-03, -4.217498441034e-02, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.218929455122e-02, -3.505239824973e-04, {   0,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.902511646550e-02, -1.233926596298e-02, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.879668899805e-02, -1.223992331440e-02, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.994441340116e-04, -3.955831670546e-02, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.510457475886e-02, -2.462462960255e-02, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.898682123321e-02,  2.952194513369e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  3.837999999714e-02,  4.689001568291e-07, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.770593262706e-02, -3.861919901249e-05, {   2,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.514742736146e-02, -1.256223929143e-02, {   0,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.514535678453e-02, -1.256120554793e-02, {   0,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.633145675197e-04, -3.671013610253e-02, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.122401496241e-02,  5.555927425205e-03, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.637999999280e-02, -7.238438912855e-07, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -3.290180979183e-02, -2.814547787245e-03, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.555318899072e-02, -6.910784036030e-06, {   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.929118346838e-02, -1.606549896956e-02, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  6.122376675410e-04, -3.428397342092e-02, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.369464379260e-02,  1.087473488572e-02, {   0,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.760707766852e-02, -6.762003624039e-03, {   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.400999999178e-02,  7.479189629798e-07, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.255483486772e-02,  1.116778701050e-02, {   2,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.340763626973e-02, -2.260349129547e-04, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.182572887697e-02,  1.591006138804e-03, {   0,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.304368205647e-02, -2.027285296976e-02, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.303808745540e-02, -2.774439078510e-05, {   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.292702288475e-02, -2.006199761986e-02, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.278999999182e-02,  7.325356309521e-07, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.266728343955e-02,  4.718095364339e-05, {   0,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.715297100427e-02, -1.533382336421e-02, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.242157736769e-02,  4.334223213518e-05, {   0,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.084982670020e-03,  2.917551448715e-02, {   2,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.476086708337e-02, -7.455393154715e-03, {   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.313122111273e-02, -1.903963631855e-02, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.102636474345e-02,  2.359647077455e-04, {   2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.336317941787e-03,  2.459926047619e-02, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -7.101770858115e-04,  2.970733476412e-02, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.024434154784e-02,  2.174406239279e-05, {   2,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.448356396492e-04, -2.911475320507e-02, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.752773977341e-03,  2.157726542592e-02, {   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.082062431009e-02, -1.707714123983e-02, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.244877675240e-02, -5.176841605479e-03, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.028711286744e-02, -1.725956334500e-02, {   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.788763571410e-02,  8.935638373192e-03, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.741945901656e-02,  8.702225761206e-03, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.466458870911e-02,  1.093317257275e-02, {   1,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  9.561414197731e-03, -1.546801015322e-02, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.154838565127e-02,  2.370656923924e-03, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  8.305162671906e-03, -1.557742721852e-02, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.335136972524e-02,  2.823838809204e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  7.654895836527e-04, -2.281157841424e-02, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.863705091815e-02, -4.697130503760e-03, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.972759231916e-03, -1.508421978025e-02, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  7.082712901620e-03,  1.572488234497e-02, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.475799023875e-02, -7.968680856398e-03, {   2,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.260573486013e-02, -6.321151441093e-06, {   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.192413544940e-02, -9.224681833290e-06, {   2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.582172303040e-05, -2.164275139115e-02, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.038765752156e-03, -1.842341249234e-02, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.815034802455e-02, -3.261799760103e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.127325072092e-02, -1.172131288297e-05, {   2,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.777051393345e-02,  3.366382227352e-03, {   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.783975940664e-02, -3.235142254432e-03, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  4.571875832471e-03, -1.641993635843e-02, {   2,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -4.571766021556e-03, -1.641990151858e-02, {   2,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.007436572727e-02, -6.681783777107e-04, {   2,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.219599694756e-02, -7.417163742771e-03, {   2,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.312724235394e-03, -1.206406515419e-02, {   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.250765473829e-02, -6.865209348726e-03, {   1,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.369098935608e-02, -5.326607097962e-03, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.876349692864e-02, -7.138786927611e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.566919814200e-03,  1.700082422027e-02, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.905444349208e-03, -1.044573148748e-02, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.296045762353e-02, -5.190035886773e-03, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.609398317932e-03, -1.629533351601e-02, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  7.082743736653e-03, -1.074217677459e-02, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.407795002273e-02, -3.220636159998e-03, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.386380427599e-02, -3.309788958441e-03, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.001979165393e-04, -1.672936889333e-02, {   0,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.484451879794e-03,  1.437896153739e-02, {   1,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.233092156030e-02,  4.428682114328e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.364126256814e-03, -1.524859114439e-02, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.962090208092e-03,  1.437565584158e-02, {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.987933452033e-03, -1.306464003764e-02, {   2,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.213677954882e-03,  8.745756965364e-03, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.943237344970e-05, -1.587420849599e-02, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.588998140281e-02,  2.982301450205e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.496650933667e-02,  7.429085216458e-04, {   2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.479949922202e-02,  6.865138700926e-04, {   2,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.544597452069e-02, -2.231343070999e-06, {   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.876659910196e-03, -5.420314555240e-03, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.722755584107e-03,  5.486407247887e-03, {   0,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.452897947155e-02, -6.725043064465e-04, {   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -9.315035362397e-03, -5.765989295716e-03, {   2,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.088005097277e-03, -1.190283769269e-02, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.087454478880e-03, -1.190155452308e-02, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.449274820048e-02,  4.739207502484e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.491999999832e-02, -2.239466869706e-07, {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.436832456598e-02,  4.549686679967e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.428258662218e-02,  4.380622597432e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -9.457459251618e-03, -5.245084652670e-03, {   2,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.198730320206e-02,  2.557057396160e-03, {   1,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.435679075395e-02, -3.380007628480e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.676732209748e-03, -1.165861951533e-02, {   2,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.507053198114e-03, -8.627895039279e-03, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.686340642990e-03, -1.109827697539e-02, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.161079687449e-03,  1.162146293573e-02, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -8.571790816027e-03, -5.018480982735e-03, {   0,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.048666606931e-02,  3.050350375247e-03, {   1,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.046470918126e-02,  3.043968752025e-03, {   1,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.919758989880e-03, -1.053171265861e-02, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.920467542803e-03, -9.444868811296e-03, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.314109809496e-02,  1.031571078233e-04, {   2,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.027568647548e-02, -2.912518872971e-03, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.319539303629e-03,  4.741857318306e-03, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.300230869907e-02, -6.403868804414e-06, {   2,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.451876275002e-03,  4.341612520258e-03, {   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.228965537532e-02, -4.515864234710e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.259248541395e-02,  1.423970077031e-05, {   2,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.233999999804e-02,  2.196811022784e-07, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.227656124951e-02, -2.590615101603e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  7.840915831210e-03,  4.410896473437e-03, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.209290811943e-02,  1.436135347817e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.219610654046e-02, -4.268937741765e-06, {   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.078657363642e-02, -1.327572325276e-03, {   0,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  4.875262440405e-03, -7.232399872086e-03, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.719193925503e-03, -7.353726033294e-03, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.199318716643e-02, -5.804759574131e-05, {   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  7.592927642373e-05,  1.194085109404e-02, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.443903749006e-03, -8.552980632187e-03, {   2,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  2.303264096316e-03,  9.609948180926e-03, {   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.299004957446e-03, -4.229529719956e-03, {   0,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  5.859364975063e-03, -5.616491143586e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.486661753722e-03,  8.925176974669e-03, {   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.332022566982e-03, -4.035952664422e-03, {   0,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -1.123274109289e-02,  8.534373075767e-05, {   2,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.827138645764e-04, -1.100010877837e-02, {   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.083748602751e-03,  9.102832985183e-03, {   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.320179264335e-03,  6.839419819298e-03, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -2.667153667338e-03, -8.313590358518e-03, {   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.436151137940e-03, -8.458156684222e-03, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.433472634834e-03, -8.451077864668e-03, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  9.218263743282e-03,  1.633127599727e-03, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.421523338500e-03,  8.396370178527e-03, {   0,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.891584022868e-03, -3.880225230173e-03, {   0,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.062135211278e-03,  8.699240605672e-03, {   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.052136944948e-02,  1.623014770386e-04, {   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.064838282427e-02,  5.036444221030e-07, {   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.306846314276e-03, -8.285247943887e-03, {   2,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  2.306826399027e-03, -8.285180075020e-03, {   2,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.965642631320e-03,  8.548465390697e-03, {   2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.105236582257e-03,  8.309704940258e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  8.836911503497e-03,  1.552716980333e-03, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.036806965561e-02, -2.058462901879e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  6.516774926759e-03, -3.799414444457e-03, {   0,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  9.441708650443e-03, -8.307304584971e-04, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.016417319683e-02,  9.586871100547e-05, {   2,   0,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.025309183954e-02, -1.920725618080e-06, {   0,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.448574434530e-03, -3.575848681141e-03, {   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.067292076395e-03, -5.948455311159e-03, {   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.474917800575e-03,  5.430523815106e-03, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -9.788808520579e-03,  1.130842919221e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -8.359842753514e-03,  1.507591684312e-03, {   0,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.857765877657e-03, -7.922169331085e-03, {   0,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.122184789916e-03, -5.632153304156e-03, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.337986900773e-03, -8.392340907623e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.342077710016e-03, -8.379129787369e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -9.540759281797e-03,  1.216479015492e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -9.445303235727e-03,  2.162047002783e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  8.886235634645e-03, -7.028610337894e-04, {   2,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  9.566923881001e-03,  5.529073938186e-06, {   2,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.554705457201e-03,  1.985152570514e-03, {   2,   0,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -8.007814394081e-03, -1.505468779592e-03, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.536290896329e-03, -5.927127327804e-03, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -8.076278417669e-03,  1.351112368758e-03, {   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.325049449070e-03,  1.556976675358e-05, {   2,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  9.316064599740e-03,  2.448024427771e-06, {   2,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.155390792501e-03, -6.000621475805e-03, {   0,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  9.131565397652e-03,  1.190232337508e-05, {   2,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.139999998222e-03,  1.802610957928e-07, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.560666746019e-03,  5.567714938493e-03, {   0,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -8.863533860844e-03,  2.597113364755e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.760836166614e-03, -7.275027565747e-03, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.204967878957e-03,  7.821617300195e-03, {   2,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  5.738677025902e-03, -3.154961163763e-03, {   2,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  8.739477238552e-03,  1.297258384681e-04, {   2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.859999999702e-03, -7.267900071183e-08, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.783930024267e-03,  2.098696052756e-05, {   0,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.799343770757e-03, -2.607493694364e-06, {   0,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.628357490260e-03,  6.816817254979e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -8.647357553830e-03,  4.281649365121e-05, {   0,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.306171869379e-03, -2.351236638424e-03, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.740234879935e-03,  6.820435235578e-03, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  5.999909701251e-03, -2.552930637551e-03, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.993448885948e-03, -2.550058530660e-03, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.497975144019e-03,  5.634443017739e-06, {   2,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.540367515600e-05, -8.280460255990e-03, {   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.165533847789e-04, -8.180272311451e-03, {   2,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.289711268821e-03, -2.451552505052e-06, {   2,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.019185351350e-04,  7.738140661458e-03, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.714840498319e-03, -4.521880253403e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  5.771388517608e-03, -2.458344751403e-03, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.169999999373e-03,  1.012413139178e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.724520062455e-03, -2.437584865816e-03, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.526680531528e-04,  7.232871149434e-03, {   1,   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.233450344507e-03,  1.847617268576e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.998522084425e-03,  3.011937610730e-03, {   0,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.993005171278e-03,  3.008483287925e-03, {   0,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  5.738420719535e-04,  7.371674557692e-03, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.781416962537e-04, -7.249215590135e-03, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.678341105055e-03,  1.405086775234e-04, {   0,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.346626384714e-03,  5.470618219084e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.325823640430e-03, -6.450966708314e-03, {   1,  -3,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -5.057519182091e-03,  2.713339000500e-03, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.038709732337e-04, -7.438216501176e-03, {   2,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.010183456891e-04, -7.362355676301e-03, {   2,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.647669863848e-03,  2.590226291521e-06, {   4,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.582237550202e-03, -8.942044575322e-06, {   0,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.475934745771e-03,  1.041428267340e-04, {   0,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.693253893888e-03, -8.691044613027e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.426761009903e-03,  1.594456309527e-06, {   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.404330414740e-04,  6.486831205806e-03, {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.957416634535e-03,  4.453306174479e-03, {   0,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  4.508347758398e-03,  2.880912030932e-03, {   2,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.482877993263e-03,  2.864139329596e-03, {   2,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  6.271173359865e-03, -1.026284710698e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -7.282484021363e-03, -1.234438763602e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.367060785981e-03, -5.783561433364e-03, {   0,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.104067191570e-03,  3.943934439658e-06, {   2,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.494724905056e-03,  5.611207685234e-03, {   0,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -3.408775831612e-03, -3.590696070189e-03, {   0,   0,   1,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -2.683255787448e-03, -4.316138130271e-03, {   2,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -4.225884465806e-03, -2.761504779226e-03, {   0,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.899986860256e-03,  7.932543033232e-05, {   2,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -6.940346386447e-03,  3.159379323092e-05, {   2,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.790887534820e-03,  5.043889141952e-03, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.790617431071e-03,  5.043060769190e-03, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -5.837825468852e-03, -9.560482017473e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -3.066630229092e-03,  3.713768690033e-03, {   0,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.589598127751e-03, -2.189649925692e-03, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  5.342707545190e-03,  1.416875775930e-03, {   2,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.500092371566e-03,  2.210842097310e-03, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  2.593656981516e-03,  4.110519425417e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -4.468487106776e-03,  2.198100573562e-03, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -6.653250804309e-03,  9.033202500942e-07, {   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.417950226576e-03,  1.221385776207e-03, {   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.567036882383e-03,  4.568986516555e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -5.975526407506e-04,  5.992122591470e-03, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.192563825219e-03, -2.382513579312e-03, {   0,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.233085780339e-03, -5.329994922904e-03, {   1,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.383790911235e-04, -6.302308386915e-03, {   0,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  4.128176963385e-03, -2.372674319536e-03, {   2,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  6.459999999185e-03,  1.026009254034e-07, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.349279173874e-03,  1.059365200276e-04, {   2,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -4.923411776671e-03, -1.497963885467e-03, {   0,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.735779876166e-03, -3.683369713573e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  4.160175461007e-03, -2.078189666387e-03, {   0,   2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.207586375915e-03, -4.989506321017e-03, {   0,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.804298292941e-03, -4.380715813608e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.297754090696e-03,  4.864425441172e-03, {   0,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  2.986117697411e-03, -3.139001168354e-03, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.072929583681e-03, -2.034640170326e-03, {   0,   2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.836294861036e-04, -5.922637086669e-03, {   2,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.751789308682e-03, -3.348910744472e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -2.953685037899e-03, -3.146649468808e-03, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.235864156591e-04, -5.942512246603e-03, {   0,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  1.905223138120e-03, -4.117922975264e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.345004947714e-03,  4.667836617006e-03, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -3.202560957453e-03, -2.803893842126e-03, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.230567422223e-03, -3.772644086764e-03, {   2,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.160422909373e-03, -2.840951593767e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.067839898748e-04, -5.832464888518e-03, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -5.527107068811e-03, -4.038515700984e-04, {   0,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  6.621050942161e-05, -5.826364714742e-03, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.640641967582e-03,  3.201595134382e-03, {   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.398085040835e-03, -3.430703971074e-03, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.411400314366e-03, -4.413469467522e-03, {   2,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.409326287730e-03, -4.406290560196e-03, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.870224911040e-03, -1.933283643577e-03, {   1,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.857233674564e-03, -1.926839342494e-03, {   1,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.394551558776e-03, -3.382171240970e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -5.764244148250e-03,  1.301364970124e-06, {   4,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.740781683720e-03,  2.378464624874e-06, {   2,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.466272601397e-03, -1.270026686946e-03, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.461196479697e-03, -1.269885592239e-03, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.709979207258e-03, -6.983741852235e-06, {   2,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.585037842540e-03,  1.066053661221e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -2.543650578432e-03,  3.100404072438e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -2.402312113215e-03,  3.235495284490e-03, {   0,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.220189335064e-03, -4.387211003263e-03, {   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.062036090988e-03, -3.538062818959e-03, {   0,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.640789011530e-03,  3.949724554087e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  5.541222549424e-03,  6.136596665085e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.360839638410e-03,  1.138133851986e-03, {   3,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.491083329655e-03,  7.836295555274e-06, {   0,   0,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -6.117236503000e-05, -5.426294637025e-03, {   0,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.053470472888e-03, -4.420628698105e-03, {   2,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -5.464737600323e-03, -2.893531713509e-06, {   0,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.636638469942e-03,  1.816630938062e-03, {   4,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.181639373464e-03, -1.267188744413e-03, {   0,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.972090732368e-03, -1.468821244216e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -5.421046695990e-03, -3.758985671363e-07, {   2,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.730135875461e-03, -3.682655467699e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.804734082458e-03,  1.587917462278e-03, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  5.206282147712e-03,  1.752428313294e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  5.242717066162e-03, -1.265587520572e-04, {   2,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.423552076321e-03, -2.942486533624e-03, {   0,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.077685343455e-03,  4.268011459613e-03, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.017296174597e-03,  4.306563392610e-03, {   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.544426075785e-03,  1.770601771460e-03, {   4,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.330819472156e-03, -9.761631507201e-04, {   1,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.463617992468e-03,  3.842965436499e-03, {   0,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.771035434087e-04, -5.022047273901e-03, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -1.770067576882e-03,  3.518710485148e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.242479635353e-03,  2.009282037230e-06, {   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.125767292824e-04,  4.381097196433e-03, {   1,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.660312358751e-03,  1.532814647197e-03, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.060852220329e-03,  4.130482130643e-03, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  4.394111066296e-03,  7.700135117878e-04, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.053389835878e-03,  4.102384422932e-03, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  5.112546103817e-03,  1.767563533130e-06, {   2,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.873520561904e-03, -3.239948973396e-03, {   0,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  4.344463867627e-03,  7.675044368580e-04, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.304700033021e-03, -1.803404240054e-03, {   2,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.408883836886e-03,  3.675731702775e-03, {   0,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.655154620183e-04, -4.117244919780e-03, {   2,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -3.375780246025e-03,  1.686344296143e-03, {   0,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.713519495555e-03, -3.326637155242e-03, {   0,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -5.027328417268e-03, -1.047720406342e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.020042328762e-03, -1.015183580977e-03, {   2,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.613424643946e-04, -4.066922679551e-03, {   2,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.199923027455e-03, -1.808489713879e-03, {   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.990514959973e-03, -5.614654782016e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.866034275650e-04,  4.196318721203e-03, {   0,   2,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.414656986090e-03, -5.676669241751e-04, {   2,   0,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -3.333371809521e-03,  1.648586251805e-03, {   2,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.505135485911e-03,  1.466370840682e-03, {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.195818230575e-03,  1.757138065670e-03, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.685887459234e-04, -4.763332607354e-03, {   0,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -4.775275273913e-03,  9.316108087234e-05, {   1,   1,  -1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.242748458842e-04, -4.639750832903e-03, {   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.510887367802e-04, -4.609587367704e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -4.829999999293e-03,  8.261341480987e-08, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.430041176032e-03, -3.398277914963e-03, {   0,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  3.908443957352e-04, -4.412800950968e-03, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.789999999643e-03,  5.852088981791e-08, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.520474964596e-03, -2.448406905343e-04, {   0,   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -9.183024185347e-04, -3.844704510562e-03, {   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.167188126089e-03,  1.582210888983e-03, {   0,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.946399441640e-03, -7.848076015661e-04, {   0,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  3.503194330135e-03,  1.203555584446e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  2.146481537036e-03, -2.552036149247e-03, {   1,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.945121158718e-03, -1.753204283055e-03, {   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.644741204487e-03,  5.281098705290e-05, {   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.797908462582e-03, -2.894488086931e-03, {   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.806473407011e-03, -2.876228020771e-03, {   2,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.601883163827e-03, -6.220325928241e-05, {   2,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.603093426535e-03, -5.597745783932e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.853491723820e-03,  1.804061990230e-03, {   2,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -3.851084612559e-04,  4.242684333433e-03, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.833463426185e-03,  1.790600365052e-03, {   2,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -4.427555877878e-03,  1.881088885379e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.610000003252e-03, -2.200162022050e-08, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.751715208503e-05, -4.525378499526e-03, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.936252489188e-03, -2.660613199350e-03, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.993103454383e-03, -5.528163266527e-04, {   2,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.537503637315e-03,  9.641469704007e-07, {   2,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.508103851082e-03,  2.813640276194e-05, {   2,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.669035941916e-03, -2.850811711702e-03, {   2,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.628751150414e-04, -4.283200262842e-03, {   2,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.251720236792e-05, -4.416070722939e-03, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.354083585136e-03,  3.083823288611e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -2.852720560461e-03,  1.567265935425e-03, {   1,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.852720560461e-03,  1.567265935425e-03, {   1,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -8.714844850582e-04, -3.544188442344e-03, {   2,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.336966659003e-03, -5.741886697056e-05, {   2,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.594649399051e-04,  4.013947066077e-03, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.781645632151e-03, -5.763277961183e-04, {   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.811132638048e-03, -2.525992812220e-03, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.382137121588e-04,  3.472222526620e-03, {   0,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.431792178774e-04,  4.058261068472e-03, {   2,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.288913084534e-03,  6.866399318277e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.028974736645e-03,  3.263249674405e-03, {   2,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.539371650942e-03,  7.270149053988e-04, {   2,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.536971511765e-03,  1.702077464291e-03, {   0,   0,   1,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -1.222619439855e-04,  4.083325200906e-03, {   2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.322748410967e-03,  2.882795505790e-03, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  8.666048120362e-05, -4.101507272360e-03, {   0,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.033538134122e-03, -1.209502912007e-04, {   2,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.330664190468e-03,  2.821246015901e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.107924897906e-04, -3.913949202942e-03, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.288030621801e-03, -8.223356393258e-04, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.951326015590e-03,  1.129280633724e-03, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.684589380041e-05, -3.982535240194e-03, {   2,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.518359835200e-07, -4.059372602002e-03, {   1,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.249478613140e-03,  2.810039311605e-03, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.482308298222e-03, -2.548978946560e-03, {   0,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.588047236690e-04, -3.865216389219e-03, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  7.487856363178e-05, -3.894024450213e-03, {   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.634117103408e-03,  3.341280481525e-04, {   2,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -3.964957398404e-03, -3.208078029231e-06, {   2,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.455671815334e-05,  3.845369415957e-03, {   0,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.585557954021e-03,  1.307775695412e-03, {   0,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  3.558798348210e-03, -3.263242302483e-04, {   0,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.825869572640e-03, -5.337403714670e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.000962785397e-04, -3.772910028916e-03, {   0,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.869999999179e-03,  7.970220561594e-08, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.869999999966e-03, -1.621061809186e-08, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.830438378143e-03, -1.015149684088e-03, {   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -8.502862209682e-04, -2.983743509776e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.063393293663e-03,  7.439752045273e-04, {   1,  -1,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.570162239879e-03, -2.234516679274e-03, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  7.380297526350e-04,  3.042822459272e-03, {   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.768690594480e-03, -1.331102953466e-07, {   0,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.809298154127e-03, -1.951476005083e-03, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  3.092427160479e-03,  6.682867544049e-04, {   0,   0,   1,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  1.674906518432e-04, -3.588409836008e-03, {   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.990945656503e-03, -7.633350144425e-04, {   2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.750137429382e-03,  1.991758063265e-06, {   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.154753201713e-03, -5.723177303613e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.585892476786e-03,  1.139861678624e-03, {   2,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.572321249175e-03, -2.132125740467e-03, {   0,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.570397716847e-03, -2.128441049823e-03, {   0,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.659462384902e-03,  3.164465552617e-05, {   0,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.097383782179e-03, -2.572541180577e-03, {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.319999856371e-03, -3.237765144682e-04, {   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.277395303590e-04,  3.382050266567e-03, {   2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.167696589734e-03, -2.438825487863e-03, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.605015988899e-03,  7.134973525001e-07, {   2,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.931931769847e-03, -1.672256235343e-03, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.588078529462e-03, -1.191435673818e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -6.975531935610e-04, -2.893553396590e-03, {   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.127300798547e-05,  3.579080421903e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.295286065124e-03,  1.287061120565e-03, {   2,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -6.683392804068e-04,  2.908994013126e-03, {   1,  -1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.570030991832e-03,  4.843355081056e-06, {   0,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.523725676432e-03,  1.036295928738e-03, {   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.539923465933e-03, -1.286744079662e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.307620223678e-03,  1.240407179247e-03, {   2,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.306663239734e-03, -2.230791007374e-03, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.536915360624e-03, -1.341296527953e-07, {   2,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.747457241872e-03,  7.768890959480e-04, {   2,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.285293021430e-03, -1.236684453685e-03, {   2,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.458088282560e-03,  1.057710067552e-03, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -8.040053727034e-04,  2.710863287866e-03, {   2,   0,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.440365657969e-04,  3.366669678608e-03, {   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.458830267889e-03,  2.339487148339e-05, {   2,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.564024746090e-03,  9.125821852976e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -3.439561192612e-03,  3.414545785551e-05, {   4,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.035161687381e-04,  2.544690795865e-03, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -9.033850284818e-04,  2.544247563045e-03, {   2,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -3.395216824305e-03,  2.763446019220e-05, {   0,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.419999999597e-03, -5.252742916566e-08, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.761540216540e-03, -1.654338375544e-03, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.844501642979e-03,  5.706793821967e-04, {   2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.239745528316e-03, -1.636192926420e-04, {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.502144672371e-03, -8.776341712279e-04, {   2,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.343958081664e-03,  1.017818533387e-03, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.268913919354e-05, -3.326241538208e-03, {   2,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.293932083874e-03, -2.041612017928e-03, {   2,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -6.371350932674e-04,  2.694581036540e-03, {   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.448759542462e-03, -8.813163087126e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  2.312375887035e-03,  1.000189055289e-03, {   2,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  3.309999999760e-03,  3.986157478533e-08, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.638613728865e-05, -3.263368991975e-03, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.517119817428e-03, -7.828471377756e-04, {   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.684330725269e-04, -3.026753945740e-03, {   2,  -2,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  2.355449117196e-03, -9.371263197047e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.912890920557e-03,  3.762582822028e-04, {   0,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -2.827698524073e-03, -4.594081487095e-04, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.277512903745e-03, -2.471326247773e-06, {   2,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.417839233557e-03,  8.558716090885e-04, {   2,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.242477692364e-03, -2.030502598704e-03, {   0,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  7.438335173219e-04, -2.494280007393e-03, {   4,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.431694866919e-04, -2.492278607816e-03, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  8.897293378850e-04, -2.344852451225e-03, {   0,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.563039560935e-03,  1.666409287839e-03, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  3.096181220982e-03,  1.290090762846e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.351417479596e-03,  1.869029747953e-03, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.393569342655e-04,  2.973603996043e-03, {   0,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -3.039363094763e-03, -1.674393031969e-04, {   0,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  8.089009497931e-04,  2.390800630303e-03, {   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.684703227632e-03,  5.032930629540e-04, {   1,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.660174992541e-04,  2.408469374463e-03, {   2,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.697313110514e-04,  2.302711767461e-03, {   1,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.927040781582e-04, -2.569715818768e-03, {   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.787638517116e-05, -3.102252248660e-03, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  2.327877762238e-03, -8.295413328994e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.247510472498e-03, -1.901812811084e-03, {   2,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.649010835340e-03,  4.960267362765e-04, {   1,  -1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.748978942114e-04, -2.661996920626e-03, {   0,   0,   1,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 {  1.508894051078e-03,  1.625924040625e-03, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.237942175788e-03, -1.883066520576e-03, {   2,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.098418954975e-03,  1.208612256046e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -9.896298596066e-05, -2.999543585918e-03, {   2,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.998026755432e-03,  1.098126471323e-03, {   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.052835313356e-03,  2.035672415643e-03, {   1,  -1,   0,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.545292423075e-03,  1.528260241198e-03, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.259056522329e-03, -8.140010960750e-04, {   2,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  8.399119257655e-05, -2.973139725987e-03, {   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.014988059271e-03,  1.005974256527e-03, {   4,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.016160080557e-03,  3.057132829370e-06, {   2,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.660465848173e-04, -2.336683162757e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  2.352341620618e-04,  2.766715464461e-03, {   2,   0,  -1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  2.325535391642e-03,  6.695409936802e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.494584167506e-04, -2.332430887428e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  2.941862262093e-03,  2.167609264364e-05, {   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -6.731763841030e-04,  2.290123600883e-03, {   1,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  6.723341633273e-04,  2.285260618715e-03, {   1,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.821264134469e-03,  1.169952366933e-04, {   4,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.929999999400e-03, -5.932025061267e-08, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.901683949402e-03, -8.838658649136e-07, {   1,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.155057718476e-03,  7.407064922233e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -7.083548945703e-04,  2.186486332435e-03, {   2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.889999999926e-03,  2.068040630670e-08, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.518040157698e-04,  2.332923962498e-03, {   3,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.301471010625e-03, -5.756073852799e-04, {   1,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.871626880704e-03, -1.443580509149e-06, {   2,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.126785381701e-03,  7.271619061794e-04, {   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.893974823288e-03,  9.459550983614e-04, {   4,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.152298954371e-03,  1.677965606997e-03, {   0,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.819999999807e-03,  3.297625088643e-08, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.759557957298e-03,  1.056729133416e-03, {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.970341124884e-03, -8.375182402560e-04, {   0,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.263393940349e-04, -1.865841803399e-03, {   2,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.754145548564e-03, -3.712386927155e-05, {   0,   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.032290296994e-04, -2.682280921441e-03, {   2,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.954827795737e-03, -8.306501900727e-04, {   0,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.054486916379e-03, -1.727274872862e-03, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.323950657481e-03, -4.560354784238e-04, {   2,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.911642742720e-05,  2.688336925697e-03, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.298793382740e-03, -1.465135940365e-03, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.395199730513e-05,  2.706476578855e-03, {   2,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.100751284380e-05, -2.664797928580e-03, {   0,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.705290182428e-03,  1.020420926556e-03, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  1.026375359414e-03, -1.697073245204e-03, {   1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.718530122378e-03,  1.961529448605e-06, {   0,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.716951105341e-03, -9.661879260624e-07, {   2,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.702818292676e-03,  1.823712035057e-06, {   4,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.299053707304e-04, -2.572545439857e-03, {   2,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.795406007853e-03,  8.967188410708e-04, {   2,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.478848593454e-03,  2.105553143264e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.124490673053e-03, -1.558636374810e-03, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.449629511826e-03,  2.293382230979e-04, {   0,   0,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.263001417570e-03, -1.410292924366e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.866542317722e-03,  7.980725381521e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.647223547758e-03, -5.676335352337e-06, {   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.002010575802e-03, -1.650251906608e-03, {   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.616381798512e-03,  2.415090612811e-05, {   2,   0,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  1.024791982998e-04, -2.536501200244e-03, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.378831842116e-03, -2.590707446719e-04, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.025906745007e-04, -2.124335727370e-03, {   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.295596457219e-03,  3.229482289289e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.609999999594e-03, -4.600862440980e-08, {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.737466689867e-03,  8.677797170059e-04, {   4,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.820424052604e-04, -1.716108686545e-03, {   2,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.732531821938e-03,  8.653029174451e-04, {   2,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.504823182690e-04,  2.338880185739e-03, {   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.585835907051e-03,  5.856200881467e-07, {   0,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.183418353359e-03, -3.907013982980e-04, {   2,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.222858730674e-03, -1.350888022799e-03, {   2,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.713062714204e-03,  8.556029440497e-04, {   4,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.565807279077e-03,  1.936479058807e-06, {   1,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.780656675703e-04,  1.580945868264e-03, {   2,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.188731718312e-04, -2.335490867435e-03, {   2,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -7.327062274543e-04,  1.811904767677e-03, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.808401423912e-03,  7.354373659079e-04, {   0,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.772848553537e-04, -1.964836967201e-03, {   2,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.768708049345e-04, -1.963876219454e-03, {   2,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.155976098801e-03,  3.846743974570e-04, {   2,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.489334646405e-03, -3.828223450399e-05, {   2,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.500907570377e-03, -4.912772539605e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.499999999221e-03,  6.239552075232e-08, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.489999998359e-03, -9.039409259972e-08, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.020232770401e-03,  4.523310858095e-04, {   0,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.780522895101e-03,  6.653793750674e-04, {   1,  -1,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -5.305629377082e-04,  1.905562928182e-03, {   0,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  5.305412037823e-04,  1.905452049354e-03, {   0,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.431878610987e-03,  1.211050872197e-06, {   2,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.843384855086e-04, -1.648740950188e-03, {   2,   0,  -1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 {  9.349996190109e-04, -1.494130300570e-03, {   2,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.104614059007e-03, -1.318924630056e-03, {   1,  -2,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -2.419999999424e-03,  5.279620986864e-08, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.409999999107e-03, -6.561739855038e-08, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -4.508876019118e-04, -1.949299880971e-03, {   0,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.399999999363e-03, -5.529203069798e-08, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.399999999772e-03, -3.309144261655e-08, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.582303403697e-03,  8.110386798538e-04, {   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.080309518020e-03, -1.310445562772e-03, {   2,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.389999999103e-03, -6.548998951383e-08, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -4.566700657138e-04,  1.927409925177e-03, {   1,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.280424326163e-03, -9.892756142538e-05, {   0,   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -6.974129900702e-04, -1.680181542086e-03, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -6.968750078604e-04, -1.678894038729e-03, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -6.801023068502e-05, -2.299754838965e-03, {   2,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.701353950500e-04, -1.990071230318e-03, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.306948052496e-04,  1.626478473264e-03, {   0,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  2.276690206380e-03,  7.666495103671e-05, {   2,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.562243080749e-03, -7.907181249311e-04, {   2,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  2.283169971778e-03, -6.830295248799e-05, {   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.493687179487e-04, -1.895477935933e-03, {   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.326916372938e-03, -1.284458489426e-05, {   0,   0,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.907405884532e-04, -1.543055274667e-03, {   0,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -2.253816043118e-03,  7.268263815247e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  8.982693863631e-04,  1.427884619762e-03, {   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.056059554722e-04,  1.815946096603e-03, {   0,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.564812475826e-03, -7.459371129923e-04, {   2,   0,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.879496835309e-03,  4.284080544156e-04, {   2,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.293669773632e-03,  1.415475592326e-05, {   0,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  6.662459016649e-05, -2.233730593164e-03, {   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.289950017668e-03,  6.980294028444e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  5.339427507099e-04, -1.759370089705e-03, {   4,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.792247741907e-04, -1.713924864922e-03, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -5.332795937569e-04, -1.757461904743e-03, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -3.606411944271e-04, -1.929863960298e-03, {   2,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.289999999420e-03,  5.155877142881e-08, {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.260609677653e-03,  4.738490626131e-06, {   2,   2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.193573408311e-03,  7.112003627697e-05, {   2,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.935731524476e-04, -1.769063512202e-03, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.249999999967e-03,  1.217367153260e-08, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.445756658248e-03,  7.958412183262e-04, {   2,   0,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  9.456187216823e-04,  1.278071202319e-03, {   2,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  8.504305725308e-04, -1.370912425653e-03, {   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  1.626196332374e-03,  5.860404051889e-04, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.851406488321e-05, -2.181750131419e-03, {   0,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  4.698144252182e-04, -1.737294184554e-03, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.141799794593e-03, -5.384444923220e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.827589121912e-05, -2.174507006977e-03, {   2,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.805427562429e-04,  1.308048180333e-03, {   0,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  2.172046236130e-03,  9.499991094950e-06, {   2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  2.179999999803e-03,  2.929709682309e-08, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.446926017547e-03, -7.265185359970e-04, {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.172280249017e-03, -3.943572615749e-07, {   2,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.806002859165e-03,  3.612744183321e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  8.741508911489e-04, -1.283448936227e-03, {   1,   0,   0,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  2.134948619662e-03,  1.627756799964e-05, {   2,   0,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.665734006735e-03, -4.853633833164e-04, {   1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.658686141204e-03,  4.806970046494e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.764712273874e-04,  1.850576620716e-03, {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.439880063090e-04,  1.173007532872e-03, {   2,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.100114784890e-03, -1.513379535377e-05, {   4,  -2,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.094148099436e-03,  1.888579403418e-05, {   4,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.312544034791e-04,  1.767514266353e-03, {   3,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.955495051844e-04,  1.896813471468e-03, {   0,   0,   1,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  5.289361056752e-04,  1.563288570860e-03, {   2,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.067486879339e-03, -1.433753198523e-06, {   2,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.361815148522e-03,  7.030540330636e-04, {   0,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  4.139107978942e-05, -2.023413459069e-03, {   0,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.055031078523e-03,  1.694364375030e-06, {   2,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.844539883625e-05, -2.025165195564e-03, {   0,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -6.621026338912e-04, -1.380270123120e-03, {   0,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  6.618903898611e-04, -1.380194003634e-03, {   0,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  5.583855643056e-04, -1.477387154615e-03, {   1,   1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -3.957646646856e-04,  1.635039389118e-03, {   2,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.026518699574e-03,  1.831344973614e-06, {   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.317514072498e-04, -1.890677928192e-03, {   2,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.584280243651e-03,  4.267736616910e-04, {   0,   0,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  1.049253693134e-05, -1.994570823349e-03, {   3,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.001553165763e-03,  1.878066324392e-06, {   0,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.107603069571e-04, -1.081501789883e-03, {   2,   0,  -1,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -1.619708700883e-03,  3.712456442351e-04, {   4,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.268243127518e-05, -1.924404276800e-03, {   0,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -9.270451787691e-04, -1.057397865026e-03, {   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  3.812272576160e-04, -1.597560614273e-03, {   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.704129404021e-04, -1.607397160370e-03, {   2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.870021601314e-04,  1.586233823666e-03, {   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.309166860600e-03,  6.542020460705e-04, {   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.359024892585e-04, -1.227155036492e-03, {   2,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.412035540151e-04,  1.121437810984e-03, {   2,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.881829658408e-03,  7.772114730010e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 {  9.949766182545e-05, -1.858495672654e-03, {   2,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.874904239054e-03,  8.136167587007e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.286601982610e-04, -1.216443948518e-03, {   2,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.377748173865e-04, -1.305162855002e-03, {   2,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.708645477238e-03,  2.329807731133e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -2.480673106940e-05, -1.911974761569e-03, {   2,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.923157198376e-03,  3.043719616742e-06, {   2,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.016642328834e-04,  1.623012078823e-03, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.118321270004e-03, -8.047248928659e-04, {   0,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  1.616694068620e-03, -3.063410654512e-04, {   1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.117426478289e-03, -8.041232459445e-04, {   0,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  1.897181616866e-03,  3.190630844209e-06, {   2,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.377440808265e-04,  1.660855465373e-03, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -3.865386153627e-04, -1.504358634848e-03, {   1,   1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.255540252000e-03, -6.270641319231e-04, {   0,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.881639469414e-03, -3.946527389530e-08, {   2,  -2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.136857986960e-04, -1.259920062070e-03, {   2,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -1.251134277577e-03,  6.211806862774e-04, {   2,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.869999999434e-03, -4.601909638245e-08, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.869999999840e-03,  2.447824275852e-08, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.193824303497e-04, -1.449305710094e-03, {   0,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.141252623060e-04, -1.151736838300e-03, {   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.821864755165e-03, -4.100639089152e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -1.763833001474e-03,  9.108565671671e-05, {   0,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  6.941561570635e-05, -1.782516575832e-03, {   2,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.850211116041e-03, -8.355478473848e-07, {   4,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.755506375230e-03,  8.780859082706e-05, {   0,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -3.572978713985e-04, -1.476601044844e-03, {   2,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.870818355043e-04,  1.544357183214e-03, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.825641706501e-03,  3.627696896042e-06, {   2,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.825480821792e-03,  2.879032639589e-06, {   0,   0,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -9.561046301543e-04,  8.700980939012e-04, {   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.003728575201e-03,  8.190402525520e-04, {   2,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -1.776639426072e-03, -4.133072729273e-05, {   0,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.399165515293e-04,  1.476347356820e-03, {   1,   1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.025274554446e-03,  7.853090955697e-04, {   0,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  1.720385631280e-04,  1.635548009034e-03, {   2,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.408671427872e-04, -8.524500800383e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  1.784326827064e-03, -3.401958575710e-06, {   0,   0,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  6.307745130636e-04, -1.154701053278e-03, {   0,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.783216311463e-03,  8.491731336767e-07, {   0,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.189469209767e-03, -5.940851875482e-04, {   0,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.350756408149e-05, -1.697952371332e-03, {   2,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.931106677076e-04,  1.584395580651e-03, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.012119487562e-04,  1.260126375054e-03, {   2,   0,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.754644511651e-03,  1.203962924080e-06, {   4,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.358319043158e-04,  1.417781757142e-03, {   1,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.313850298670e-04, -1.321231661421e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  1.321835855959e-04, -1.613795005279e-03, {   2,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.802771190270e-04,  1.554108188866e-03, {   3,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.722306484554e-03, -1.532550830352e-06, {   4,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.218401983760e-05, -1.670484396554e-03, {   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.015430830287e-04, -1.617600001723e-03, {   2,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.204257790819e-03,  5.142146113310e-04, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.087934466959e-03, -6.295177823160e-04, {   2,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.674517548151e-03, -4.072277081723e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.706767341410e-03, -3.456417637784e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  5.070899503049e-04,  1.194941202300e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -5.070692104447e-04,  1.194830145199e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -1.451813987876e-04, -1.553077409953e-03, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.255372009815e-04,  1.371740247101e-03, {   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.241113555228e-03,  4.542474240081e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -5.995272106454e-04, -1.088649080536e-03, {   0,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.771206223985e-04, -1.505728934656e-03, {   2,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.109306589917e-03, -5.587852071536e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  3.374023063779e-04, -1.328279854379e-03, {   0,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.431988234591e-03,  2.304668591705e-04, {   0,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.469904130983e-03,  1.892222015982e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.649637232012e-03,  4.176774387819e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.649381289816e-03,  1.335425178503e-06, {   2,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.064659260933e-03,  5.779716148237e-04, {   0,   0,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  3.153619006488e-04,  1.312770825254e-03, {   1,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.631376975032e-04, -1.050676004047e-03, {   0,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -3.378790007716e-04,  1.270307779171e-03, {   2,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.587643993865e-03,  9.632080749979e-06, {   2,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.580110398107e-03,  1.183137570831e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.062157558169e-03,  5.296831810125e-04, {   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.808435716378e-05, -1.560764616837e-03, {   3,  -2,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  7.445655377138e-04, -8.400905817589e-04, {   2,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.422013137261e-04, -1.031651224255e-03, {   0,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  8.226325976435e-05, -1.486812693739e-03, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.258021494361e-03, -3.081068749212e-04, {   2,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.563652501284e-03,  3.969752849190e-07, {   2,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.561211049969e-03, -1.647147732683e-06, {   2,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.325397456710e-04,  1.127187293858e-03, {   1,  -1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.957998306079e-04, -1.261743204716e-03, {   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.998148822446e-04,  5.427605617401e-04, {   2,   0,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  2.957366668101e-04, -1.246363221200e-03, {   2,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.303725177632e-03,  2.364846832293e-04, {   0,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.411071920176e-05,  1.520763265858e-03, {   2,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.042789238547e-04, -6.264934441236e-04, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.313640383804e-03,  2.158933852187e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.520422750899e-03, -6.950128257736e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  6.812291603880e-05,  1.458947852462e-03, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -5.192497485097e-05, -1.474563258942e-03, {   2,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  1.522415572083e-03,  3.959627128514e-07, {   2,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.555095779556e-04,  8.664286924496e-04, {   0,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -9.090481661602e-04,  6.050420394035e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 {  1.453249470238e-03, -5.449994075827e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.424018192434e-05, -1.493428105206e-03, {   3,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.922073731077e-04,  1.214402376364e-03, {   1,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.926502695710e-04, -1.009731841191e-03, {   2,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -1.028517182353e-03,  4.732608708131e-04, {   1,   1,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.497951982281e-03,  1.764072773035e-06, {   0,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.873264065429e-04,  1.210814649854e-03, {   1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.476898740255e-03,  1.639698651458e-05, {   4,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.132649478066e-04,  1.176300081222e-03, {   2,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -6.660265966212e-04, -8.173877665516e-04, {   2,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.359653909220e-04, -9.390257728894e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.276608742338e-03,  1.959278377291e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.274302688526e-03,  1.949731975892e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  9.235381594476e-06,  1.456710882989e-03, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.949664693300e-04, -9.700642147499e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  3.077536253254e-04,  1.155855654735e-03, {   2,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.025122018227e-03,  4.355728930232e-04, {   0,   2,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.025122018227e-03,  4.355728930232e-04, {   0,   2,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.653508088619e-04,  1.285586329502e-03, {   2,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.804105300678e-05,  1.391793290837e-03, {   2,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.427422625463e-04,  6.043581780199e-04, {   2,   0,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -2.777808552189e-04, -1.157885247115e-03, {   1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.037289137583e-03, -3.962555902280e-04, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.429999997756e-03, -8.011584861228e-08, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.648059696059e-04, -1.162790930691e-03, {   2,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.873971943437e-04,  1.139989623121e-03, {   2,   0,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.820715345895e-04,  1.245312554082e-03, {   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.339558748245e-05,  1.341942103696e-03, {   2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.420470266284e-03, -2.693428872118e-06, {   4,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.413318220816e-03,  6.823138649921e-06, {   1,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.725765514015e-04,  1.145615738752e-03, {   1,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.978021467939e-04,  1.118330643197e-03, {   2,   0,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.008205995402e-03, -4.025226635254e-04, {   2,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.409999999863e-03,  1.968731396186e-08, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.256514215887e-03, -1.452648361044e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  7.401207882362e-04, -6.521157077985e-04, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.072995171228e-04, -4.824220407655e-04, {   2,   0,  -1,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.703148578894e-04,  4.154179576336e-04, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -8.291705857554e-04,  5.557582108988e-04, {   0,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.234676789261e-03,  1.466088953063e-04, {   0,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.961731387270e-06,  1.376524957865e-03, {   2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.326425483432e-04, -8.349071792543e-04, {   2,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -5.889296036522e-04,  7.742629317379e-04, {   0,   0,   0,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  1.307494320544e-03,  5.422526496274e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -7.747077519806e-04,  5.863564263487e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.120071047372e-03, -2.407346850478e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  7.973916359247e-04, -5.596429107950e-04, {   2,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  1.353410908569e-03,  8.141751134945e-07, {   4,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.335036528060e-03,  1.649845922648e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.071716515897e-04, -8.420526335529e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -1.193910003045e-03, -1.544261477894e-04, {   2,   0,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  9.829517255325e-04,  3.567075709040e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.294475458071e-03, -4.405605955460e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 {  2.561881067631e-04, -1.080161093279e-03, {   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.198333808644e-04, -5.138144088769e-04, {   0,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.397489302637e-05,  1.278564728227e-03, {   2,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  5.470546790945e-04,  7.778359563492e-04, {   1,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  8.963114288364e-04, -4.278218787723e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  4.931659612490e-04, -8.298677025937e-04, {   2,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -8.110043981507e-04, -5.116075229868e-04, {   2,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  8.813185131856e-04, -4.401722143684e-04, {   2,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -9.713031401542e-04, -3.492863152155e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  8.932239349083e-04, -4.263479905769e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -8.210732345818e-04,  4.974976857638e-04, {   0,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  4.065691472001e-05, -1.275944433315e-03, {   0,   0,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.847285599689e-04, -7.316767092498e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  6.525996780722e-04,  6.625810593279e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.312751680107e-03, -1.928686956098e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -2.503994602755e-04,  1.057299865696e-03, {   1,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.970350920193e-04, -1.009910149423e-03, {   4,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.966252839022e-04, -1.008916171239e-03, {   4,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -7.742756657764e-04, -5.283824230252e-04, {   0,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -1.869777182787e-04,  1.114762732137e-03, {   1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -5.542862969535e-04, -7.446545284784e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  9.397992804764e-04, -3.567792975785e-04, {   0,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.265729371173e-03,  3.082049066737e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.292227335163e-03,  2.795219882548e-06, {   2,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.645002225824e-04,  4.304037569040e-04, {   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.628391315141e-04, -4.309480792166e-04, {   2,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.052121680167e-05, -1.252553351667e-03, {   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.290967681122e-03,  1.961964276671e-06, {   0,   0,   1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.019725665689e-03,  2.720935312475e-04, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -7.891658947914e-04, -4.916490147770e-04, {   0,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.278599344797e-03, -3.478750209992e-07, {   2,   2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.829401253477e-05,  1.179368556348e-03, {   2,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.830252983350e-05, -1.227546018303e-03, {   2,   0,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.269993145994e-03,  2.399441168738e-06, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.184098194644e-04, -8.507591982905e-04, {   2,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  9.513892443332e-04, -3.158336632637e-04, {   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.720157821687e-05, -1.229255070751e-03, {   4,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.210114068567e-03, -5.540732486079e-05, {   0,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  4.051822662899e-04,  8.570629285833e-04, {   2,  -2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -1.213299917255e-03, -4.845539833112e-05, {   4,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.259999999776e-03,  2.375044045973e-08, {   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.904783665310e-04, -4.667700426247e-04, {   0,   0,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.040754901165e-04, -1.052934391445e-03, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -1.011942443850e-04, -1.155577658534e-03, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.117083016507e-04,  1.041609918166e-03, {   2,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.854585538786e-04, -4.640940206484e-04, {   0,   0,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -8.721911081446e-05,  1.159813728321e-03, {   2,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.797417453854e-04, -1.064359551694e-03, {   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.048756801586e-03, -1.942651424184e-04, {   1,   1,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  7.720265551852e-06,  1.232884735062e-03, {   1,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.236807662313e-03,  9.440747740534e-07, {   0,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.944538846908e-04,  4.425450559586e-04, {   2,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.233470233895e-03, -3.049723004022e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  3.153686858405e-04, -9.210137957356e-04, {   4,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.530871708175e-05,  1.159127683534e-03, {   0,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  1.205769952661e-03, -2.827754167371e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  9.768138772933e-04,  2.544079966111e-04, {   1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.229999999889e-03, -1.653001334490e-08, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.684967100508e-04,  8.605610865105e-04, {   1,  -3,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -8.705782315837e-04, -3.571006649042e-04, {   0,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.224690841246e-03, -2.964650909568e-06, {   2,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.075101642264e-04,  7.190176811575e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -6.640729376644e-04,  5.605914020693e-04, {   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -3.761243563460e-05, -1.185248319383e-03, {   4,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.219997191084e-03,  2.753759385167e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.356067768344e-04,  9.871172554916e-04, {   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.076410609590e-04, -1.013572910163e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -4.850087144240e-04,  7.347644792246e-04, {   0,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.033031956547e-03, -1.859098738738e-04, {   2,   0,  -1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  7.890032489792e-04,  4.283309345404e-04, {   2,   0,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.075051575127e-03,  1.372522542306e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -9.258861615930e-04, -2.844555345449e-04, {   2,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.170477097240e-03,  3.915696310041e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.061530493231e-03, -1.432628900105e-04, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -3.768665958202e-05, -1.165909432587e-03, {   4,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -5.382394676477e-05,  1.147768181147e-03, {   2,   0,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -1.193172059684e-03,  7.153075514897e-06, {   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  4.998590457026e-04,  7.004621193235e-04, {   0,   0,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  1.191094858856e-03,  1.485679572507e-06, {   0,   0,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.025868679632e-03, -1.644795205516e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.060781069819e-04,  1.083569743669e-03, {   3,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.016143594636e-03, -1.734024542280e-04, {   2,   0,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.136444076706e-03, -5.204167646397e-05, {   2,   0,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  2.347968426105e-04,  9.524060172061e-04, {   1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.761189154042e-04,  5.102360347325e-04, {   1,   1,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  7.491589218611e-04,  4.362171349071e-04, {   2,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  3.023468254524e-04,  8.805820858214e-04, {   0,   0,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  6.320232882612e-04,  5.496423905941e-04, {   2,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  1.847426621846e-04,  9.960233033673e-04, {   1,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.298504239690e-04,  9.498550656273e-04, {   2,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.561625247049e-04,  9.233373403343e-04, {   2,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.550787251909e-04, -7.232962307094e-04, {   2,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  8.519818546948e-04, -3.227642569727e-04, {   0,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.171216041598e-03, -1.741773215149e-06, {   2,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.177487864838e-04, -6.523316819575e-04, {   0,   0,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  6.002776045511e-04, -5.682671630266e-04, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.981411956053e-05, -1.108423161908e-03, {   2,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -7.366031459935e-04, -4.286308666091e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  4.497000621306e-04, -7.136578992777e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  9.981753275858e-04,  1.646517584740e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -4.712067190005e-05,  1.113871974056e-03, {   2,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.159999999857e-03,  1.822123739007e-08, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.105907753039e-04,  5.438699081712e-04, {   2,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -7.092086645009e-04, -4.426324046271e-04, {   2,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.558137669142e-05, -1.124829186800e-03, {   2,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.145620589548e-03,  1.631191238606e-06, {   0,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.140913675196e-03,  7.505697704662e-07, {   4,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.929808884101e-04, -9.459754772314e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  1.132322454334e-03,  1.811979391288e-06, {   2,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.131564469462e-03,  1.182991770193e-06, {   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.100360525064e-03, -3.234799817584e-05, {   2,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.129999999867e-03, -1.735555408105e-08, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.070065164180e-04,  9.228455865543e-04, {   2,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.063514355469e-04,  4.219695476741e-04, {   1,   1,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.126698549109e-03,  1.341004247757e-06, {   2,   0,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.651484203940e-05, -1.089688756398e-03, {   2,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -6.616437602100e-04, -4.631883845511e-04, {   2,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -9.112672181289e-04,  2.083772476297e-04, {   4,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.030383225349e-05, -1.077739565184e-03, {   2,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -7.427651795537e-04,  3.705588332735e-04, {   2,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.056986351063e-03, -5.315589002773e-05, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.109437551051e-03,  5.457482408040e-07, {   4,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.549644348806e-04,  6.527401698306e-04, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  9.614436868036e-04, -1.458559200465e-04, {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.498141835152e-04,  6.463024811085e-04, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -8.807947864248e-04,  2.142877141266e-04, {   2,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.833708799740e-05,  1.016678005408e-03, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.965445325125e-04, -1.977393569869e-04, {   2,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.015877383593e-03, -7.278282207130e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -8.945718949025e-04, -1.933419710159e-04, {   0,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -1.033501380584e-03, -5.044061809282e-05, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -4.033779515225e-04,  6.792272454645e-04, {   2,   0,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -1.080543977776e-03, -4.663986789710e-07, {   0,   0,   3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.080204891159e-04, -8.702756203719e-04, {   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.056840204083e-03,  1.817902805550e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  2.055882444096e-04, -8.658454754354e-04, {   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.536246310693e-05,  1.033257592281e-03, {   2,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  7.118864324964e-04, -3.555528750730e-04, {   2,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.622446755703e-04, -1.043815682144e-04, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.023832592882e-03, -4.205426564942e-05, {   2,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  8.219018408555e-04,  2.408573182421e-04, {   0,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -9.015632099956e-04, -1.599324233497e-04, {   2,   2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.043214977929e-03,  1.797315263492e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.501711064499e-04,  9.098259496735e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  6.801124773851e-04,  3.796948023392e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -7.054642588088e-04, -3.523471942994e-04, {   2,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.176046900122e-04, -8.387636299794e-04, {   0,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.175874959615e-04, -8.387032843257e-04, {   0,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  8.809313499197e-04, -1.730622769266e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  5.758910608814e-04, -4.673783096525e-04, {   1,   0,   0,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 {  6.950019933220e-04,  3.469438473769e-04, {   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.030582799147e-03,  9.686558476849e-06, {   2,   0,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  7.299336002134e-04, -3.101919406866e-04, {   2,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.275793496246e-04, -3.091496685194e-04, {   2,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.030927403555e-04, -2.334043938575e-04, {   3,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  6.672346785983e-04, -3.668649369295e-04, {   1,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.027709335138e-03,  3.522845396973e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.839382963866e-05, -9.726532870172e-04, {   2,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.087999491844e-04,  9.201017056897e-04, {   3,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.634370331775e-04,  3.622236936791e-04, {   0,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.010372913831e-05, -9.718554711591e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -7.847580951521e-04,  2.371967465557e-04, {   0,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.881960377317e-05, -9.709867703474e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  1.000093468849e-03,  1.939725222820e-05, {   2,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.684753485293e-04, -6.499191568181e-04, {   0,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  8.489676662355e-04,  1.693024175833e-04, {   2,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.160884902517e-04,  9.004767897904e-04, {   2,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -1.955279999954e-04,  8.200140175381e-04, {   1,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.279870171631e-04,  2.873147480096e-04, {   2,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  7.531560850573e-04, -2.611003979631e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  3.185872701812e-04, -6.954093317906e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -1.919280997600e-04, -8.216620844985e-04, {   1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.804965372158e-04,  2.269032658522e-04, {   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.061244052452e-06, -1.002297205511e-03, {   4,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.934643528576e-04,  8.103602983203e-04, {   1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.501870831561e-04, -3.520782774840e-04, {   2,   0,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.909445330508e-04,  8.066219108147e-04, {   1,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.471232159305e-04, -6.473594755540e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  8.058771886184e-04,  1.871864553221e-04, {   2,   0,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -6.713379004958e-04,  3.208992078048e-04, {   0,   0,   0,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -2.158785261804e-04, -7.752715882225e-04, {   2,   0,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  9.899979560566e-04,  1.006254542386e-06, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.158058597203e-04, -7.751291371117e-04, {   2,   0,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -8.469208248261e-05, -9.062341466019e-04, {   1,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.560681311872e-04,  5.345901480341e-04, {   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.646713517120e-04, -4.249272904535e-04, {   1,   1,  -1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  3.838264079211e-04, -6.043514062999e-04, {   4,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.990347271262e-05,  9.428081810563e-04, {   2,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.360617464619e-04, -6.466078111157e-04, {   2,   0,   0,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -9.539086924730e-04, -2.521661274853e-05, {   2,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.804092448803e-04, -5.981417944486e-04, {   4,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.412348690167e-04, -2.362015972842e-04, {   0,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.147646852947e-05, -9.455994269824e-04, {   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.237379099103e-04,  6.521065042100e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -9.751353751070e-04, -6.237972060275e-07, {   0,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -5.921950987803e-05, -9.163007265904e-04, {   0,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  5.445858859424e-04, -4.303704747825e-04, {   4,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -8.164934511764e-04,  1.568383959932e-04, {   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.266797251551e-04, -3.445356801328e-04, {   0,   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -9.699999997908e-04,  2.014633555432e-08, {   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.275929313522e-04, -3.406927640043e-04, {   2,   0,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -3.526811878968e-04, -6.140583886064e-04, {   2,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  9.000965349280e-04, -6.529350363178e-05, {   2,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.867429912876e-04, -7.776968837167e-04, {   1,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.557313848920e-04, -8.052981010867e-04, {   2,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.198535323162e-04, -4.408519838478e-04, {   2,   0,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  5.917731933972e-04, -3.683304404660e-04, {   0,   0,   1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.596168021802e-04, -5.999077280090e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -6.342972002557e-04,  3.246395052901e-04, {   2,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.848940940657e-04, -7.735219799450e-04, {   1,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.948458361555e-04, -6.632702180207e-04, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  9.309775455688e-04, -2.605424775884e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.113537649218e-04, -7.448421559306e-04, {   2,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.530024332629e-04, -6.026347786617e-04, {   0,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -2.111116294773e-04, -7.442036722544e-04, {   2,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -5.246537286625e-04,  4.304495019264e-04, {   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -8.576704022098e-04, -9.086452956309e-05, {   2,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  7.549886460611e-04,  1.910890873159e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  9.428194899818e-04,  1.325583087821e-06, {   2,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  9.367684067285e-04,  5.912293531181e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  8.339887982941e-04,  1.059475941237e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  1.384487855475e-04,  8.012764616082e-04, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.381898134808e-04,  8.001815916757e-04, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  8.969162588678e-04, -4.106218825570e-05, {   2,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  1.765038557644e-04, -7.604842879071e-04, {   0,   0,   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  7.898989581088e-04, -1.463192096411e-04, {   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.882399587949e-04, -7.460805832216e-04, {   3,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.299980214090e-04,  2.969568604110e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.183470083863e-04, -2.144034542865e-04, {   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.453751722384e-04, -8.578446448100e-05, {   2,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.207291799670e-04,  3.100274291487e-04, {   0,   0,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.563451238271e-04,  1.737573210602e-04, {   4,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.587257826540e-04, -3.713178178640e-04, {   0,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  1.779512592190e-04,  7.511803226051e-04, {   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.791290127570e-04,  2.483172888727e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  6.557785997572e-04,  2.709762868100e-04, {   2,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.362539924077e-05, -8.801571563860e-04, {   0,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.149749574996e-04,  2.081318170821e-04, {   1,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.568804110376e-04, -5.623451720908e-04, {   2,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -5.658702430017e-04, -3.529169340738e-04, {   0,   0,   1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.052273323737e-04,  6.129810692455e-04, {   2,   0,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -7.097141163583e-04, -2.062754449786e-04, {   1,   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -7.088613931632e-04,  2.063157586982e-04, {   1,  -1,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -5.885837960184e-04,  3.233650380462e-04, {   3,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  5.885837960184e-04,  3.233650380462e-04, {   1,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  7.292442895241e-04,  1.825595538588e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  9.099999998530e-04,  1.635897107806e-08, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.077932255155e-04,  1.452019275010e-06, {   0,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  7.369571178122e-05,  8.355133549989e-04, {   0,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  3.520170527172e-04, -5.567310040627e-04, {   4,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -7.502935225665e-04, -1.569083050105e-04, {   2,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.829891763750e-04, -3.236798636889e-04, {   2,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.816503668644e-04, -4.234476579148e-04, {   4,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  7.640821395443e-04, -1.405811956339e-04, {   2,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -4.101474314334e-04,  4.916793986338e-04, {   2,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -3.023171982062e-04, -5.951141948982e-04, {   2,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -8.966972670868e-04, -4.737086222961e-07, {   2,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.795321064905e-04,  1.602262403773e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  5.277585382567e-04, -3.670029087347e-04, {   2,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.589377413229e-04, -3.343360262008e-04, {   0,   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -4.039735309240e-04,  4.881510970923e-04, {   0,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.965447062786e-04,  2.955264858345e-04, {   2,   0,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.321687089474e-04, -6.588799694968e-04, {   4,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.385772763900e-05, -8.559361370098e-04, {   2,   0,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.662892599736e-04, -7.222274857072e-04, {   2,   1,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.867694916327e-04,  1.141546438952e-06, {   0,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.475823456594e-04,  5.402421536644e-04, {   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.064343961537e-04,  7.798895370956e-04, {   2,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -3.102755311428e-04, -5.725672669473e-04, {   2,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  8.799999999764e-04,  6.450736915313e-09, {   4,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.584369763783e-04, -4.184310655630e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -8.691701319651e-04,  7.624462122621e-06, {   2,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.739125239858e-04, -1.755305611227e-06, {   2,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.418568300750e-04,  5.308931209413e-04, {   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.607397803042e-04,  1.060838119919e-05, {   0,   0,   0,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -6.103111956648e-04,  2.608290549596e-04, {   2,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  2.809271368184e-04,  5.887392337976e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -5.611531047360e-04,  3.082957009807e-04, {   3,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  5.611531047360e-04,  3.082957009807e-04, {   1,   0,  -1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -2.598669166932e-05, -8.428193305978e-04, {   4,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.894844516690e-04, -4.769086527019e-04, {   1,   1,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.640835555104e-04, -7.020870615363e-04, {   3,  -1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.637266741067e-04,  7.813893464340e-07, {   0,   0,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.004470701516e-04, -7.630806448228e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -8.620064077152e-04, -4.782884823286e-07, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -3.817183071063e-04,  4.778392074352e-04, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.632260334385e-04,  6.961028106524e-04, {   1,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.236619519585e-04,  1.305668948265e-04, {   2,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  6.759582259031e-04, -1.763932148995e-04, {   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -7.342823584318e-05,  7.784631764811e-04, {   0,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -2.466530426351e-04, -6.050555682248e-04, {   0,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.462329484325e-04, -6.041536396823e-04, {   0,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  3.188391321512e-04,  5.315305342356e-04, {   1,   1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.388030970222e-04, -7.099596426044e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 {  7.144566241801e-05, -7.768577177894e-04, {   3,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.360885954256e-04,  1.141883709472e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.828148227276e-04,  6.636925868064e-04, {   0,   0,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.896708041781e-04,  4.566356150723e-04, {   1,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.694570903967e-04, -6.757737984812e-04, {   0,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  2.754514164909e-04,  5.694552850117e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.203056809461e-04,  7.237902254638e-04, {   0,   0,   1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.190560725065e-05, -8.219132467163e-04, {   2,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.366651807323e-04, -7.011151204479e-06, {   2,  -2,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -3.576057018356e-04, -4.857182345432e-04, {   4,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.057082322341e-04, -5.375990091756e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  3.045330233804e-04, -5.385877767877e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  1.662329381146e-04,  6.757885944494e-04, {   2,  -2,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.291033383539e-04, -5.116428739638e-04, {   1,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  5.604106301743e-04,  2.802893313634e-04, {   0,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.613376720262e-04, -2.784772836654e-04, {   1,  -2,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.287533280224e-04,  1.110149103295e-04, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.316974001309e-04, -2.077040494312e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -1.824735690773e-04, -6.553674537531e-04, {   2,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.823949181112e-04, -6.550886560799e-04, {   2,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -8.301925724040e-04, -5.812317123507e-06, {   1,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.175327381994e-04,  1.512891684275e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -5.403099453611e-04,  2.921643307909e-04, {   2,   0,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  8.300000027055e-04, -1.946216623824e-08, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.358109066313e-04, -2.912570420790e-04, {   1,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.292312929210e-04, -6.963479764965e-04, {   1,  -1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -6.381199402428e-04,  1.857087761030e-04, {   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  8.219657583588e-04,  8.682901509254e-07, {   2,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.504984224876e-04,  2.722777540282e-04, {   2,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.144450702988e-04, -3.082535403470e-04, {   2,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.927382405866e-04, -6.289953750367e-04, {   2,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -5.210717182147e-04, -2.998676008067e-04, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  8.186779630780e-04,  1.304497718927e-06, {   2,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.982334664628e-05, -7.894414267301e-04, {   0,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -8.155034698170e-04,  1.996233066115e-06, {   2,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.592500124607e-04, -6.581231284713e-04, {   0,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.956726104106e-04,  2.203189932598e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -3.309945180440e-05,  7.821996767602e-04, {   2,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.870017605516e-04,  4.250962858161e-04, {   2,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -3.969954845858e-04, -4.150020152332e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  2.373880819428e-05, -7.882332455409e-04, {   0,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.099999999822e-04, -5.372123437697e-09, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  3.577351558200e-04,  4.506849610882e-04, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  6.645667069682e-04,  1.416915522543e-04, {   1,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  6.642965824645e-04,  1.417644521405e-04, {   1,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -7.860958319722e-04,  1.854183796276e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.074765748374e-04, -2.969611169501e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -8.011184269697e-04, -1.239110034840e-06, {   2,   0,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.423026163155e-04, -5.590550644959e-04, {   2,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  7.999999999420e-04,  9.634217470776e-09, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.724002782706e-04, -2.725008151083e-05, {   4,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -7.900808636627e-04,  8.432371160922e-06, {   1,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.749429952187e-04, -4.234240553979e-04, {   0,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.776804910862e-04,  1.204381634103e-04, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.884723395014e-04,  7.933743335102e-06, {   1,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.789791894058e-04, -2.164585898511e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -6.712945449193e-04, -1.237976579569e-04, {   0,   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.080910454610e-04, -5.860809409339e-04, {   0,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.080266691825e-04, -5.858914915413e-04, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.397922569768e-04, -5.536584526781e-04, {   2,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  6.080021222412e-04, -1.847995955870e-04, {   2,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.923020168523e-04,  8.219645215190e-08, {   2,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.318132577402e-04, -5.999551324752e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.941483585323e-04, -5.969785340624e-04, {   2,   0,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.500075817403e-04, -6.408081374303e-04, {   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.060073167959e-04, -5.847809850317e-04, {   1,  -1,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  5.272535530256e-04, -2.633375470938e-04, {   1,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.898969962972e-04, -2.547208740947e-07, {   2,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.473886576475e-04,  6.420847529042e-04, {   1,   1,  -1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.310447800764e-04,  3.547517357212e-04, {   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.740490497627e-04,  6.113686507706e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  4.284272172950e-04,  3.558987247098e-04, {   2,   0,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -5.081250078447e-04,  2.761625980834e-04, {   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.429811212122e-04,  1.407945898314e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.461623594086e-04,  6.361212425045e-04, {   1,  -1,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.602889368479e-05, -7.453960595831e-04, {   2,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  7.669942436950e-04, -1.210468074462e-05, {   2,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.253578637289e-04, -4.535396651511e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 { -4.396132549598e-04, -3.390108111717e-04, {   2,   0,  -1,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 { -7.772143678983e-04, -3.172124920561e-07, {   1,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.774816989930e-04, -4.983344838838e-04, {   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.753154060879e-04,  1.042631762706e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -7.738493830716e-04,  1.247658006814e-06, {   2,   0,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.542526513794e-04,  6.179306027647e-04, {   1,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.057006738321e-04, -3.663774734793e-04, {   2,  -2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.891417008557e-04, -2.822407796839e-04, {   0,   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -3.747707017203e-04,  3.939605456600e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -4.967330401848e-04, -2.712247757394e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -4.392952134543e-04,  3.283460232665e-04, {   1,  -1,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -1.468597420089e-04,  6.207389871236e-04, {   1,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.543656509653e-04,  1.129137334473e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  7.050991588968e-04, -6.177521568650e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.179578160554e-04, -3.480878719040e-04, {   2,  -2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -5.106061783140e-04, -2.549222636685e-04, {   1,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.172555560925e-04, -2.474004052645e-04, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -3.963395654334e-05, -7.238176841157e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  7.294030719840e-04, -3.144416349884e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  7.599999997952e-04, -1.764178807752e-08, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.066352759461e-04, -5.246260460635e-05, {   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.573353697112e-04,  3.017552813742e-04, {   0,   0,   0,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -7.572013236446e-04, -3.297697694178e-07, {   2,  -2,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.728740341346e-04, -5.843080761276e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  2.959797248741e-04, -4.597463260166e-04, {   2,  -2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.978727834719e-04, -5.573040512703e-04, {   0,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -7.290701255177e-04,  2.530166310202e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.824813523708e-04, -2.710342677104e-04, {   2,   0,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.886280717464e-05, -7.243427698945e-04, {   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.845167491055e-04, -2.682013707288e-04, {   0,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.946610833767e-04, -4.570410456631e-04, {   2,  -2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.394347268014e-04, -4.121059651263e-04, {   0,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.461917669293e-04, -6.041614904084e-04, {   1,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.546386079989e-04,  4.945806921019e-04, {   0,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  6.353326504006e-04,  1.125842440967e-04, {   2,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.869702262146e-05, -7.080206233214e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -6.539221926468e-04, -9.177202436901e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  4.954287905818e-04, -2.474429852651e-04, {   0,   2,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.441451398415e-05, -7.179299399023e-04, {   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.421310455366e-04,  1.370039531756e-07, {   1,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.523448900710e-05, -6.467709782944e-04, {   2,   0,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  9.445060648421e-05, -6.457952012189e-04, {   2,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  4.263269571340e-04, -3.138996051494e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -7.399999998850e-04,  1.304459082873e-08, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.399999999432e-04, -9.169959889718e-09, {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  8.254146578974e-05, -6.557538056641e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -4.432352937148e-04,  2.898805353943e-04, {   0,   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -6.622368488799e-04,  6.896917692086e-05, {   0,   0,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  7.299988566468e-04,  1.141540560557e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.299988769544e-04,  1.049076469322e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.078638808437e-04,  2.230127352221e-04, {   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.299220803778e-04,  5.412546747121e-07, {   2,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.183160785807e-04,  1.121411120939e-04, {   4,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -6.011741795165e-04,  1.273943810841e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -4.178783700890e-04, -3.106496213014e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -1.111642799910e-04, -6.168340841541e-04, {   2,   0,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  7.128796394088e-04,  1.452531282200e-05, {   2,   0,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  4.844857404222e-04,  2.405042477811e-04, {   2,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.025300838422e-04,  5.210079575589e-04, {   2,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.267093578480e-04,  5.959981062369e-04, {   0,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.342814046460e-04, -3.882154623681e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  4.037698932229e-04, -3.186577391231e-04, {   0,   0,   1,   0,   0, -23,  22,   4,   0,   0,   0,   0,   0 } },
 { -4.798738736472e-04, -2.396786207418e-04, {   0,   2,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.690661109386e-04,  5.500002426533e-04, {   2,   0,  -1,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 {  2.761035307010e-04, -4.412986298305e-04, {   4,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.171035551341e-04,  1.284405651967e-07, {   4,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.147042230972e-04, -1.640647712419e-07, {   0,   0,   3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.677103644215e-04,  3.466136932287e-04, {   2,   0,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.740705695089e-04, -4.377056023655e-04, {   4,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.487075177685e-04, -2.623263483703e-04, {   0,   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.586274061068e-04, -5.523255925677e-04, {   0,   0,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.836157381769e-04,  2.685386238527e-05, {   2,   0,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.099999998644e-04,  1.387885821098e-08, {   2,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.944894042672e-04, -4.149667161813e-04, {   0,   0,   0,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 {  2.376527341322e-04, -4.716312953040e-04, {   2,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  8.128665245413e-05, -6.277869174648e-04, {   4,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  3.198746914873e-06, -7.054726567376e-04, {   3,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.079868363093e-04, -4.145497723444e-07, {   2,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.739571467691e-04, -3.444172735990e-05, {   2,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.411056829774e-05,  6.337858192396e-04, {   1,   1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.405832471162e-04, -3.634779759372e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -7.020017078130e-04,  1.436226682337e-07, {   0,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.080966351139e-04, -4.918845979709e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  6.889205967833e-04,  1.067939011669e-05, {   1,  -2,   0,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 { -5.876805218463e-04, -1.114533943344e-04, {   0,   0,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.477037044802e-04, -2.506050867142e-04, {   2,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  6.952288994299e-04, -6.945703024900e-07, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -6.951132426576e-04, -6.832217718000e-07, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  6.306257969309e-04, -6.369637601634e-05, {   2,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.310008265031e-04,  5.632892448291e-04, {   1,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.189793459830e-05,  6.501552762324e-04, {   2,  -2,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  2.904434267415e-04,  4.013488595989e-04, {   2,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.182196612178e-04,  4.726369887923e-04, {   0,   0,   0,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 {  5.810518441981e-04,  1.096723459906e-04, {   0,   0,   0,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -4.882773162493e-05, -6.411733556356e-04, {   2,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.612044761998e-04, -2.287575628695e-04, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.921047362418e-04, -3.953694728367e-04, {   2,  -2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.959913454140e-04, -9.039597152061e-05, {   2,   0,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.307974337727e-04, -5.521782747932e-04, {   1,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.194973218551e-04,  3.629750256685e-04, {   2,   0,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  1.312395535569e-04,  5.496204977814e-04, {   1,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.536390902300e-04,  2.270800894536e-04, {   0,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.721292867425e-04,  7.901634987639e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.788738821753e-04,  6.228676455654e-07, {   1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.413068942077e-04, -5.381169563987e-04, {   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  5.293377236047e-04, -1.478989252229e-04, {   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.026665785859e-04, -2.721788656550e-04, {   2,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -2.089173559383e-04, -4.645669504740e-04, {   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -6.706895394206e-04,  1.123941375374e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.204867337635e-04,  4.511641451799e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -1.366912361576e-04,  5.340184660751e-04, {   2,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.524927330614e-04,  1.589522510502e-05, {   0,   0,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.991546711732e-04,  3.675584264400e-04, {   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.211667026402e-05, -5.735557749759e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.698183410867e-05,  6.386753825714e-04, {   2,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.273424567679e-04, -5.372076922162e-04, {   1,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.406734909605e-04, -2.197132797436e-04, {   1,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.599999999718e-04,  6.105161723389e-09, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.727137185466e-04, -2.871732886520e-04, {   4,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.157734339024e-04,  3.439620057748e-04, {   2,   0,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.266655422410e-04,  5.326752579351e-04, {   1,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.310534073028e-04,  1.282002367513e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -1.499368021583e-04,  5.083923268600e-04, {   0,   0,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.831297261850e-05, -6.299921929375e-04, {   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.832920517996e-05, -6.296430958597e-04, {   2,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.840506687357e-05, -6.086438544787e-04, {   2,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  9.005957349782e-05, -5.667531380455e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.355057412798e-04, -5.203683568349e-04, {   2,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  6.379054380115e-04,  1.753354034271e-05, {   2,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.658508324621e-04,  4.888003473484e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -6.496322703657e-04, -4.739233023263e-06, {   3,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.429568379098e-04,  1.113093676482e-04, {   2,  -2,   1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.788152099293e-04, -4.744694963799e-04, {   1,  -1,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  4.203254046623e-04, -2.314664538890e-04, {   2,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  6.489500011946e-04, -1.595087936929e-06, {   2,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.499999650792e-04, -2.338427000923e-07, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.499999999752e-04, -5.672320068775e-09, {   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.499999999917e-04,  3.289945639995e-09, {   4,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.750595967349e-04, -7.359630682245e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  2.736166318955e-04, -3.712286646666e-04, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.647161381231e-04,  3.774823060359e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -5.590737288335e-04, -8.289360600661e-05, {   0,   0,   1,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -1.843238263865e-04,  4.575738112411e-04, {   4,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.399999999661e-04,  6.590363255414e-09, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.732656599645e-05,  5.619688856046e-04, {   2,   0,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -6.239356066560e-04, -1.481142461866e-05, {   2,  -2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -6.197304767643e-04,  1.827939193597e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.224599691370e-04,  1.536150141059e-05, {   0,   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -3.344074483036e-04, -3.031980342149e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  2.585929022677e-04, -3.788814615349e-04, {   2,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.880091837199e-04,  3.494344343829e-04, {   4,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.284371266963e-04,  1.076862328945e-04, {   2,   0,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -9.053337300134e-05,  5.436585968589e-04, {   2,   0,  -2,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 {  1.329757722803e-04, -5.009248361080e-04, {   2,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -2.255937305051e-04, -4.078277973230e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  2.259069375537e-04, -4.070280125105e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -5.591019935690e-04, -7.382858233420e-05, {   1,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.901842137469e-05, -5.335570093416e-04, {   1,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.648020647796e-04, -3.650363296510e-04, {   2,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.452364460774e-06, -6.272562767332e-04, {   4,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.195019284954e-04, -2.095195136809e-04, {   2,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.271278130590e-04,  1.704766211658e-06, {   0,   0,   3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.402487554541e-05, -6.041261255827e-04, {   2,   0,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.114906280939e-04,  2.163605636166e-04, {   0,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -3.693668035387e-04, -2.578148852127e-04, {   2,   0,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -4.029418294020e-04, -2.231532798060e-04, {   2,   0,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -6.258867439900e-04,  9.007569083800e-09, {   0,   0,   3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.903753919951e-04,  4.351086742766e-04, {   0,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  4.164976576228e-04,  2.080095793188e-04, {   0,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.987526230383e-04, -4.250704501335e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  5.495727570803e-04,  7.410655121911e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.259334368115e-05, -6.006765061416e-04, {   0,   0,   1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -4.743275452679e-05, -5.754964271096e-04, {   4,  -2,  -2,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.537354208752e-04, -4.680306590760e-04, {   2,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.421214432187e-04,  3.792180932117e-04, {   2,   0,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -5.337820974071e-04,  8.754303411195e-05, {   2,   0,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -2.980704765542e-04, -3.225155876465e-04, {   3,  -1,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.330115646139e-04,  8.723978919249e-05, {   2,   0,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -6.178094884891e-04,  7.619761714163e-07, {   0,   0,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.781286880959e-04,  4.396233449189e-04, {   0,   0,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.484576670496e-04, -4.688174739016e-04, {   2,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.027386878189e-05, -5.957087774925e-04, {   0,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.371008145052e-04, -2.762873771011e-04, {   4,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  6.119815185796e-04,  9.335535801180e-07, {   2,   0,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.993227741580e-04, -1.281229184233e-05, {   4,  -4,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  6.115694555629e-04,  4.222379564376e-07, {   2,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.820624704038e-04,  1.289280834699e-04, {   0,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  4.881507590355e-04, -1.220845315833e-04, {   3,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.825528138690e-05,  5.913785588370e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  6.031763049066e-04,  6.344263712870e-06, {   3,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.166395825582e-04,  4.920021499315e-04, {   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.658957157326e-04,  4.409974005674e-04, {   2,   0,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.975955910286e-04,  4.066345027487e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 { -1.175522998626e-04,  4.865774328186e-04, {   0,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.445313046406e-04,  4.586189061712e-04, {   2,   0,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.085088826056e-04,  2.924628497438e-04, {   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.680374730597e-04,  1.311007745226e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.100702468701e-04, -4.881239093734e-04, {   4,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.272643361180e-04,  1.706792901239e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  4.923600158092e-05,  5.477917634742e-04, {   1,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.631146949622e-04, -4.335917653095e-04, {   2,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -6.484674656177e-05, -5.310293634722e-04, {   2,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.164892531437e-04,  2.793274008860e-04, {   2,   0,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  2.619518112102e-04, -3.337905472685e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 { -1.137421401690e-04,  4.811693828314e-04, {   1,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.104221653767e-04,  3.828048678714e-04, {   2,   0,  -1,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -4.567518901068e-04,  1.362622879472e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  3.844954247452e-04, -2.083909451236e-04, {   2,   0,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.160058985123e-04, -7.372396635477e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -5.633008033065e-04, -2.591123026383e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.098645875004e-04, -4.783154704558e-04, {   2,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.555154732380e-04, -3.251196284093e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -5.813071589448e-04, -6.398572805969e-06, {   3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.708994513767e-04, -2.166769423336e-04, {   2,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.120961615340e-04, -4.749310011062e-04, {   1,   1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.771835648313e-04, -2.098290077390e-04, {   2,   0,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.166196211900e-08, -5.865824277396e-04, {   2,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.754777134613e-04, -3.106040955077e-04, {   0,   0,   1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.362783267060e-04, -4.889982019007e-05, {   2,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -5.227740730662e-04, -6.162068125860e-05, {   1,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.364302807495e-04, -2.477261408390e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  1.112313923020e-04, -4.727119524464e-04, {   0,   0,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.000398086315e-04, -4.837526266082e-04, {   4,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.655689864779e-04, -1.163710902886e-04, {   1,  -1,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.694559117438e-04, -2.108652381321e-04, {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.945715647738e-05, -5.598110427444e-04, {   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.686065881978e-04,  1.031247626083e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  1.722207775819e-04, -4.066687842568e-04, {   2,   0,   0,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 {  3.119680065135e-04, -2.654425675515e-04, {   0,   0,   0,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -5.751242527560e-04, -1.984209288157e-06, {   0,   0,   1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  5.751242527560e-04, -1.984209288157e-06, {   0,   0,   1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -4.936420679414e-04, -7.785753754505e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -4.710456372173e-04,  1.003964560363e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  1.115742399091e-04,  4.594057352030e-04, {   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.006760188476e-04,  1.701335787946e-04, {   0,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.817876809443e-04, -1.889451309150e-04, {   1,   0,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.699999999006e-04, -1.064476310728e-08, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -5.633364724220e-04,  4.751587108296e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.250745122448e-04, -4.418016394249e-04, {   1,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.851460210418e-04,  8.169173344057e-05, {   0,   0,   1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -3.251507614544e-04, -2.416869548331e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  4.850881255929e-04,  8.165011480090e-05, {   0,   0,   1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  4.532312068148e-04, -1.133526210600e-04, {   3,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.600146953286e-04, -6.502536385626e-06, {   2,   0,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.021826334389e-04, -6.371266035037e-05, {   1,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.576462030380e-04,  4.079801936396e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  3.792711522827e-04,  1.862678831431e-04, {   2,   0,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.988532958631e-04, -6.589223526009e-05, {   0,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.595130300882e-04, -2.049916313601e-04, {   2,   0,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.653292115843e-05,  5.179137836054e-04, {   1,  -1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.576559139569e-04,  1.062377508886e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.077878089677e-04, -4.556055148995e-04, {   1,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.787153744886e-04, -3.844533265643e-04, {   0,   0,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.462700354286e-04,  3.164166245701e-04, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  3.882997651325e-04,  1.734775393205e-04, {   1,  -1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -5.557345908822e-04,  5.239832553455e-06, {   2,   0,  -4,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.092440102219e-04, -4.514794619798e-04, {   4,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.100446729646e-04, -1.506349353002e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -1.097048478522e-04, -4.504834826994e-04, {   1,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.316870134479e-05,  4.765072620072e-04, {   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.498098512610e-04,  9.601108482544e-06, {   0,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.499250823153e-04,  9.077656114536e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.690860662252e-04,  2.879045393531e-04, {   2,   0,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -5.137310168706e-04, -4.256691822298e-05, {   1,  -2,   0,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 {  2.613831977536e-04, -2.946582568746e-04, {   4,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.793975361494e-05, -5.378817372931e-04, {   4,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.695694355222e-04,  1.852103822073e-04, {   0,   0,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.615795689397e-04, -1.915052972904e-04, {   2,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.111667942612e-04,  4.416953820373e-04, {   1,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.825039365209e-04, -6.975058286191e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -3.702907169774e-04,  1.819243643131e-04, {   2,   0,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.757557406645e-04,  1.752993422548e-04, {   2,   0,  -1,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 {  1.056422696729e-04,  4.452543919231e-04, {   1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.059765120892e-04, -4.446794521260e-04, {   1,  -1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.499999999624e-04, -6.431538293584e-09, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.499999999133e-04, -6.007423284501e-09, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.981327510424e-04,  5.158061853713e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   4,  -2,   0,   0 } },
 {  5.455091132030e-04, -3.506391712239e-06, {   0,   0,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.571105014384e-05, -4.616632043745e-04, {   3,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -4.319501363192e-04,  1.137763579433e-04, {   2,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -5.448765982679e-04, -5.792913203835e-07, {   0,   0,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.399754025298e-04, -4.745538954047e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.853591472708e-04,  3.584491419182e-04, {   1,  -1,   1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.853591472708e-04,  3.584491419182e-04, {   1,  -1,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.511315823244e-04,  3.923498959846e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  1.365992588350e-04,  4.057172315594e-04, {   0,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.325630635377e-04, -3.096758908707e-04, {   4,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.382913753157e-04,  2.034834269669e-04, {   0,   0,   2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.460297623762e-04, -2.955852743618e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.381160084355e-04,  2.033736053141e-04, {   0,   0,   2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.274812453627e-04, -3.135718070447e-04, {   0,   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  2.542205046962e-04, -2.866150188410e-04, {   4,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.070951305398e-04, -2.332055073816e-04, {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 {  1.009016224150e-04,  4.385843488352e-04, {   1,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.244402036613e-04,  2.149571207630e-04, {   2,   0,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -2.131145015362e-04, -3.260235373923e-04, {   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.062398914659e-04, -1.321817142088e-04, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  2.718221452883e-05,  5.104873074880e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -2.174972972237e-05,  5.146030323798e-04, {   2,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.343717980041e-04,  1.126610319589e-06, {   2,   0,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.462041690130e-04, -1.892540082852e-04, {   2,   0,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -3.813785313550e-05,  4.972617898064e-04, {   0,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  2.148254273329e-05, -5.138760021991e-04, {   2,  -2,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.335262630530e-04,  1.011836572943e-04, {   2,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -4.334363628330e-04,  1.011793493384e-04, {   2,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  5.336773604632e-04,  8.303255660190e-07, {   2,   0,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -5.115833999048e-04, -2.258588609067e-05, {   4,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -4.612050754325e-04, -7.243010493231e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -1.642209193825e-04,  3.694079168049e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  5.264534550578e-04,  6.957379754428e-06, {   2,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -7.267864093204e-05,  4.601389569485e-04, {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.790390029218e-04, -5.299815007672e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  1.018646116392e-04, -4.299276334112e-04, {   1,  -1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.652103897686e-04, -2.665412788737e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -5.313632076521e-04, -3.454627330751e-07, {   2,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.234573256200e-04,  1.073999698503e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  4.070106510701e-04,  1.236177555285e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -5.206415935673e-04,  9.531452052600e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  5.299999999058e-04, -9.990264637840e-09, {   2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.299999999822e-04, -4.347615166735e-09, {   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.192530278288e-04,  1.102375284713e-04, {   2,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  1.070148175403e-04,  4.224726554842e-04, {   1,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.179115284106e-04, -4.096978272142e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.666744586827e-04,  1.606186171555e-04, {   0,   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.760134361282e-04, -3.485393252558e-04, {   2,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -5.235188722763e-04,  7.046773771384e-07, {   4,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.212652396870e-04, -2.029329337616e-04, {   2,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.102800367735e-05, -5.029250506322e-04, {   2,  -2,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.379653766309e-04, -1.859494339429e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  2.276496023683e-04,  2.962533031211e-04, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.002674142659e-04, -4.231318153689e-04, {   3,  -1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.244081697263e-04, -1.988359101717e-04, {   2,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.898372633636e-05,  4.441647257350e-04, {   2,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.430421127090e-04, -8.007323786991e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  4.354768985610e-04, -8.712134058611e-05, {   4,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.142974327084e-04, -3.075014731385e-04, {   2,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.572781573511e-05, -5.054953156614e-04, {   4,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.131650221589e-04, -7.168181928711e-06, {   2,   0,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  2.333059666350e-04,  2.867794169386e-04, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -5.119892849480e-04,  7.909125622479e-06, {   2,   0,  -1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  3.205634215869e-04, -1.991369104643e-04, {   0,   0,   0,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 { -1.593467989316e-04,  3.584454297679e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.707710248808e-04,  1.462966772588e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -5.102957359021e-04,  5.760476864282e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  9.743157119413e-05,  4.183757304599e-04, {   1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.915623053890e-05,  4.565255155446e-04, {   2,   0,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  5.128527932452e-04, -2.083727615952e-06, {   0,   0,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.549914437090e-04, -5.921227606389e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  4.308398310461e-04, -8.281802409753e-05, {   1,   1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.130113677896e-04,  5.001851439991e-07, {   0,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.016439110550e-05, -4.926982767040e-04, {   4,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.337165899112e-04,  1.791115727387e-04, {   4,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.789874651525e-04,  3.334812615298e-04, {   0,   2,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  3.410239489224e-04, -1.703771932256e-04, {   2,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.851099225073e-05, -4.125714479963e-04, {   2,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -5.100556895930e-04, -8.025539421500e-07, {   4,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.206399410405e-05, -4.681138968243e-04, {   1,  -3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.099999999773e-04, -4.806636759929e-09, {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.218933428404e-04, -1.881051856078e-04, {   2,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -5.005819509829e-06, -5.037613769353e-04, {   2,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.757677203021e-04, -3.318417371184e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  6.774147411925e-05,  4.393751627312e-04, {   2,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -5.044910588225e-04,  2.595548389546e-06, {   0,   0,   2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.438013254053e-04, -3.630818459313e-04, {   0,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -3.295878814090e-04,  1.761968737094e-04, {   4,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.037039620280e-04, -9.485801171107e-07, {   0,   0,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.514160003981e-04, -3.532025806706e-04, {   2,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -2.360984451195e-04, -2.668604809013e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 {  3.211276452370e-04,  1.817850985318e-04, {   4,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  3.046860247216e-04,  1.972841386580e-04, {   0,   0,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.884550820260e-04,  1.324819192054e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.187500001933e-04, -1.812113693505e-04, {   2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.989926085415e-04,  1.008579234253e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  4.114822509347e-05, -4.581558943980e-04, {   1,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.536581158693e-05, -4.038599164951e-04, {   2,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.533346314584e-04,  3.449565710676e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.494174170543e-04, -3.487530902270e-04, {   2,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  3.143752321478e-04, -1.836167141403e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -2.607259125009e-04,  2.372262986438e-04, {   2,   0,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -2.031383647674e-05,  4.775485811170e-04, {   2,   0,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  4.961041584496e-04,  3.009305100753e-07, {   2,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.306546768057e-04, -1.656165881568e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  3.156172606386e-04, -1.806411395970e-04, {   2,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.902951280511e-04,  5.579663870593e-06, {   0,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.922550932382e-04, -1.028861871337e-04, {   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.881139572826e-04, -6.531779074905e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -4.258279803693e-04,  6.851678663537e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.040433162437e-04, -3.894176109914e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  3.299606840622e-04, -1.632109051374e-04, {   2,   0,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -4.905046424335e-04, -1.202617116064e-06, {   2,   0,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.002158442614e-04,  3.914867893428e-04, {   1,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.013943947002e-04, -1.902500128267e-04, {   2,   0,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  3.281166471850e-04,  1.632622582641e-04, {   0,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.897020341044e-04,  1.032496376131e-06, {   2,   0,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.466475101159e-04, -3.439960257495e-04, {   2,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  4.899999999806e-04, -4.361577800602e-09, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.549900234952e-05,  4.038724216840e-04, {   2,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.234892527537e-04,  1.653278024990e-04, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -4.065786403012e-04,  8.183343784195e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  3.232207686994e-04,  1.650831493374e-04, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -1.021160676263e-04,  3.847276610015e-04, {   0,   0,   3,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  4.287957354391e-04,  5.779586450628e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -9.361721088152e-05,  3.920462787024e-04, {   1,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.481913915169e-04,  3.743023325678e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  3.058073803033e-04, -1.779012593113e-04, {   1,   0,  -2,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 {  1.985270552364e-04, -2.845395491714e-04, {   0,   0,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -3.222675656709e-04,  1.605541780209e-04, {   0,   0,   1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.102207299831e-04,  7.218840656407e-05, {   1,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  7.356791431963e-05, -4.087211796514e-04, {   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.005873927885e-04,  3.815126531029e-04, {   0,   2,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.005834354344e-04,  3.815067028901e-04, {   0,   2,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.801663638925e-04, -1.452159715197e-06, {   2,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.805974110311e-04, -3.003718529418e-04, {   1,  -1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.173619397232e-04, -3.632004402287e-04, {   2,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.516033656990e-04, -2.886101982736e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -4.775088614868e-04, -2.686490602015e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.799999999642e-04, -5.864306286512e-09, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.821024381736e-04,  1.971313454136e-04, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  2.183822449602e-04, -2.596695361005e-04, {   1,  -1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.758323156238e-04, -3.017832844762e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 {  8.932217102398e-05,  3.876865912694e-04, {   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.633150542161e-05,  3.906734840020e-04, {   2,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -4.762036318890e-04,  5.810077723510e-09, {   1,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.156374066812e-04,  3.602181072349e-04, {   2,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.523396431481e-04,  3.234109843792e-04, {   0,   0,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  2.802836869185e-04,  1.953950395964e-04, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -4.317631419031e-04,  4.345080569361e-05, {   0,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.728482894413e-04, -1.092752100777e-06, {   2,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.732611863507e-04, -5.957464516588e-07, {   2,  -2,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.678187128587e-04,  1.054922223402e-04, {   0,   0,   0,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  4.665067952602e-04, -6.449317793693e-06, {   0,   0,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.773001739133e-04, -2.951220760277e-04, {   1,  -1,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.103368543499e-04, -2.618305168548e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.425719459258e-04,  2.932207129493e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  4.644449111666e-04,  6.765043930401e-06, {   2,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.044055770232e-04,  1.665964832404e-04, {   2,   0,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -8.978169796670e-05, -3.803328756454e-04, {   1,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.699999998935e-04,  1.000771793018e-08, {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.839704634507e-04,  8.601382839959e-05, {   2,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.691859897396e-04, -7.089356155743e-07, {   4,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.677014650119e-04, -2.109725298926e-06, {   2,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.020531468623e-05, -3.787449385747e-04, {   1,  -1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.986682831763e-04,  1.702723046114e-04, {   0,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  4.682188156406e-04,  8.333045565506e-08, {   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.140906672325e-05, -3.755931867240e-04, {   0,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -3.598632107071e-04,  1.069113726504e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -3.911244145479e-04, -7.498512638402e-05, {   1,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -8.878866009634e-05,  3.760112813304e-04, {   1,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.315886458335e-04, -3.330458788539e-04, {   1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.119978391192e-04,  1.524639454057e-04, {   0,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  4.610392277669e-04,  3.009417835772e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.793493721999e-04,  2.840645344844e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.087003956955e-04,  1.541809059696e-04, {   4,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.103552209871e-04,  1.519883895800e-04, {   0,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -8.898940792699e-05,  3.730437857649e-04, {   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.228431612858e-05, -3.895517479271e-04, {   1,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.245469899504e-04, -3.365726070784e-04, {   2,   0,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -3.547145757313e-04,  1.059334028708e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  1.783844809774e-04, -2.821357997841e-04, {   4,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  9.177778142957e-05, -3.687111715760e-04, {   1,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.570337055406e-05, -4.434725388094e-04, {   0,   0,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.482092872721e-04, -1.105718094706e-04, {   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.575056514395e-04,  1.045799482617e-06, {   4,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.656173347697e-05, -4.415815430137e-04, {   2,  -2,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -9.046458481392e-05, -3.673333504138e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -3.664762620710e-04, -9.087613008740e-05, {   1,  -1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  4.562467546000e-04,  5.542997770480e-07, {   2,   0,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.538712226388e-04, -2.204456854137e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  3.036026573465e-04,  1.522856394275e-04, {   0,   0,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.756281756198e-05, -4.183169070655e-04, {   1,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.391594568543e-04, -3.163101734833e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.030146767619e-04, -5.237854853916e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -4.253055796344e-04, -2.967371694570e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  3.835702519207e-04,  7.122626751310e-05, {   2,   0,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -2.538658578524e-04,  2.004519350977e-04, {   0,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  4.397120884327e-04, -1.348167525970e-05, {   2,   2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.517972250545e-04, -1.325126902678e-06, {   0,   0,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.600913320817e-04, -2.928388879699e-04, {   0,   0,   1,   0,   0, -18,  24, -16,   4,   5,   0,   0,   0 } },
 {  1.383936402242e-04,  3.145289052762e-04, {   0,   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  2.865764849326e-04,  1.663225245499e-04, {   2,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -4.525857322729e-04, -2.327121649135e-07, {   2,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.019588454706e-04,  1.508349293090e-04, {   4,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.592178907122e-04, -9.322904522978e-05, {   3,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.434942625301e-05, -4.280834836918e-04, {   2,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  4.511166516782e-04,  1.064675842382e-06, {   3,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.740069187069e-04, -1.771479044226e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 {  3.280379420553e-04,  1.220291300145e-04, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.644168369927e-04,  2.849898056648e-04, {   0,   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  8.957707466358e-05, -3.576521947453e-04, {   1,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.902305666238e-04, -5.692759808976e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  1.714880938193e-05, -4.299777688788e-04, {   2,   0,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.194353048679e-04,  1.276886974594e-04, {   3,   0,  -3,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  1.517848170098e-05, -4.316584309736e-04, {   0,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.738048685670e-04,  1.718170741205e-04, {   2,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  4.446568150683e-04, -5.339507530560e-07, {   2,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.667780135103e-05,  4.083561422191e-04, {   1,  -1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.693478503691e-04,  7.552704417283e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -3.730669495183e-04, -7.144713796722e-05, {   0,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 { -2.227682045775e-04,  2.214339044912e-04, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.142806135648e-04,  3.293720828770e-04, {   1,   0,   0,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 {  2.842514021539e-04, -1.589365261629e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -8.658221178892e-05, -3.564237311834e-04, {   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.721575021153e-04,  1.707118563507e-04, {   2,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -3.729068289408e-04, -6.966864162539e-05, {   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.264174574731e-05,  3.597419532478e-04, {   4,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.862282840910e-04,  5.564651450804e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.421749129644e-04, -9.961448439968e-05, {   0,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.610344869242e-05, -4.052132890770e-04, {   2,   0,  -1,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 { -1.585428332053e-05,  4.251478099277e-04, {   0,   0,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.399999998369e-04, -1.197993998430e-08, {   2,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.399999999643e-04, -5.605997557287e-09, {   2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.531776595951e-04, -2.862552995689e-04, {   1,  -1,   1,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 {  2.937180599244e-04, -1.455379302084e-04, {   1,   0,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.783437288209e-04,  2.595853193622e-04, {   0,   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -4.370311530879e-04,  6.827149288556e-07, {   0,   0,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.759315050018e-04, -6.143403002506e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -4.328117228891e-04, -4.191448234923e-06, {   2,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.843964131527e-04,  5.184198211880e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.920524054407e-04,  1.439860619904e-04, {   4,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.336775166847e-04, -1.021638478852e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.450390642952e-04, -2.907871630683e-04, {   0,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  8.321159282999e-05,  3.525642800006e-04, {   1,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.688829730091e-04, -2.667068780950e-04, {   4,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  8.359568844554e-05, -3.516323252511e-04, {   4,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.001772014097e-05,  4.249879829589e-04, {   0,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.338594206558e-04,  3.979573191378e-07, {   1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.818096612979e-04, -5.184794442651e-05, {   2,   0,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  8.612894115974e-05,  3.469644402912e-04, {   1,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.618922552750e-04,  7.057527416453e-05, {   1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.239754474891e-04,  3.081268655206e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -2.087593536179e-04, -2.231875834239e-04, {   0,   0,   3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.887668925813e-04,  4.177215309800e-05, {   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.100531993324e-04, -2.201969975235e-04, {   0,   0,   3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.299999999345e-04, -7.504915783230e-09, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.939599926222e-04, -2.353625311933e-04, {   2,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.781755814534e-04,  1.509324359958e-04, {   0,   0,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -3.280461984682e-04, -1.005369621448e-04, {   0,   0,   2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.063478919662e-04, -2.221056338824e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -4.246676194538e-04, -3.242753940370e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 {  7.293545105558e-05, -3.548769325398e-04, {   1,  -3,   1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  7.293545105558e-05, -3.548769325398e-04, {   1,  -3,  -1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -1.737815782111e-05,  4.103137435553e-04, {   2,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.244237041893e-04, -2.025861177301e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -8.209971602830e-05, -3.448627833483e-04, {   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.892072875776e-04, -3.766442408200e-05, {   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.001612398676e-05,  3.461039751209e-04, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 { -1.630485682725e-04, -2.627167780142e-04, {   2,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.794470547125e-05,  4.072224573648e-04, {   4,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.046867964415e-04, -2.007187286752e-05, {   0,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 {  2.874721353324e-04,  1.361167427023e-04, {   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  4.227461172819e-04,  8.100566168536e-07, {   4,  -2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.489928410027e-05, -3.884353742507e-04, {   1,  -1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.130483207137e-04, -3.102548680952e-04, {   2,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -2.958747856575e-05, -3.935359754469e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.890354704713e-05, -3.941759789254e-04, {   2,   0,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.876426545530e-05,  3.342487952080e-04, {   0,   0,   1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -2.562087625577e-07, -4.219075735919e-04, {   4,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.320628053666e-04, -1.900246118922e-04, {   1,   1,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.780319969826e-04,  4.368920125958e-05, {   0,   0,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  8.086576245083e-05,  3.402339879153e-04, {   1,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.200001250621e-04, -3.471773740552e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.796882009804e-04, -2.406269024982e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.058160572389e-05,  3.896922842347e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -4.199999999308e-04, -7.623598172212e-09, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.662296676777e-04, -5.313934156821e-05, {   2,   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -2.652032072507e-04,  1.539166438417e-04, {   0,   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  3.205479706514e-04, -9.822771060015e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.164071050992e-04,  7.212189477790e-07, {   2,   0,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.591880593485e-04,  1.571126654698e-04, {   2,   0,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.334054345944e-04, -2.826390005434e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  2.240413626559e-04,  1.914840943838e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  6.013639280822e-05,  3.551536402591e-04, {   2,  -2,   1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.288231149909e-04, -8.646015873051e-05, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.157860179780e-04,  2.994939020477e-04, {   2,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.244131369609e-05, -4.023107339868e-04, {   2,   0,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -8.392044464209e-05, -3.284446602280e-04, {   3,  -1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  7.172449404829e-05,  3.404706858335e-04, {   2,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  7.889006794100e-05,  3.325261886228e-04, {   1,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.822278990479e-05,  3.331290502081e-04, {   2,   0,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -1.704377090489e-04, -2.408296448230e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  3.993771588394e-04, -1.151047678127e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -4.088116255354e-04,  1.622324252064e-06, {   2,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.680759390803e-04, -2.419818832697e-04, {   0,   0,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.093714394135e-04, -3.003348407225e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -4,   2,   0,   0 } },
 {  4.091685964322e-04, -2.233827736940e-07, {   2,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.036079606493e-04,  4.842667793191e-06, {   2,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.207843246296e-04, -8.734488113144e-05, {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.594085677336e-04,  2.484254790732e-04, {   2,  -2,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  3.204403416747e-04, -8.727194710648e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.624145335703e-04, -1.448980064378e-04, {   3,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.622250504300e-04, -1.449582921930e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -1.239195728832e-04, -2.830841022750e-04, {   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.461782890247e-04,  6.004364459596e-05, {   0,   0,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -1.736031723913e-05,  3.882342642133e-04, {   2,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.487281548488e-04, -2.564652528230e-04, {   2,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.159929705222e-04, -8.861354780290e-05, {   4,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.729069362554e-04, -2.314658991496e-04, {   2,   0,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.149430101284e-04,  2.884069669909e-04, {   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.436341525103e-04, -1.593826970301e-04, {   4,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.308492060852e-05, -3.799116791100e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  2.861920093889e-04, -1.156489769732e-04, {   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.184890801139e-04, -1.830013580483e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 {  7.742315631641e-05, -3.236328308658e-04, {   0,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.843423963056e-04,  2.161835944374e-04, {   0,   0,   0,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 { -3.998795408638e-04, -1.531723567906e-07, {   0,   0,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.867305785703e-04, -1.291639971551e-05, {   2,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.255148099987e-04,  2.740618614901e-04, {   0,   0,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -3.891475457613e-04, -9.366544275653e-06, {   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.124323544996e-04, -2.857824658442e-04, {   2,   0,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  2.907397256586e-05,  3.682410260967e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  3.834593476037e-04, -1.329023499276e-05, {   2,  -2,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.552046826079e-04,  4.123376866663e-05, {   2,   0,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  3.225554423421e-04, -7.344902183347e-05, {   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.896500713833e-04, -1.063431500796e-04, {   0,   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  7.615737559363e-05, -3.192209468871e-04, {   1,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.866968531530e-04, -2.080367354684e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 {  3.938124232046e-04,  1.229599560977e-07, {   2,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.867096764279e-04,  6.761316909816e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  2.636456273748e-04,  1.296745131331e-04, {   0,   0,   1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -7.648760075439e-05, -3.167777609810e-04, {   1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.653125560689e-04, -2.760697662147e-05, {   4,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.158938473179e-04, -1.766312552692e-04, {   4,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.882371425728e-04,  1.042093474282e-04, {   4,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.911050911616e-04, -7.789552278397e-07, {   0,   0,   1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.990429978697e-05, -3.515610474201e-04, {   2,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.297444445256e-04, -2.613066884744e-04, {   2,   0,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  3.282517176401e-04,  6.252310135485e-05, {   1,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.221578764099e-05,  3.585556344930e-04, {   3,  -1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.126471404351e-04,  2.780478767259e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.842065247292e-04, -6.482752327713e-06, {   2,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -9.897396918640e-06, -3.806026987537e-04, {   4,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.895885328513e-04,  7.621732972841e-07, {   0,   0,   1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.899999999779e-04, -4.152138290380e-09, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  1.525438586605e-04, -2.373558026229e-04, {   4,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.806808453106e-04,  1.083945615129e-04, {   0,   0,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.877958458166e-04,  8.919387617286e-07, {   2,   2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.145324449755e-05, -3.471692426168e-04, {   1,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.379053492394e-04, -1.502414628311e-04, {   2,   0,   0,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 { -1.888481493676e-04, -1.988280932144e-04, {   0,   0,   2,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -2.939040578490e-04,  9.374376231764e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -7.733070674074e-05,  3.101644214449e-04, {   2,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.857237651112e-04,  1.705780878664e-06, {   2,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.069685172530e-04, -1.803011650545e-04, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -5.651796079078e-05,  3.300580591569e-04, {   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.624060293321e-04, -2.239458240199e-04, {   2,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.771188210071e-04, -8.841094688843e-06, {   0,   2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.680891298015e-04, -1.177598271368e-04, {   3,  -2,   0,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 { -1.511050630836e-04, -2.345099496186e-04, {   4,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.655139888800e-04,  1.198679470475e-04, {   1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.579053094761e-04,  1.274082759007e-04, {   0,   0,   1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.610779894116e-04,  1.240941540800e-04, {   2,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -2.928276629490e-04, -9.192766861070e-05, {   0,   0,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.130077789492e-04, -7.115224442005e-05, {   4,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.868520284475e-04, -9.728387042289e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 { -1.890558221676e-04,  1.950094319065e-04, {   2,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.656667834616e-04,  1.828011856360e-05, {   2,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  3.656667834616e-04,  1.828011856360e-05, {   2,   0,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.058885235358e-04,  2.779141260024e-04, {   4,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -7.345733231414e-05,  3.103207714929e-04, {   1,  -1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.056161211624e-05,  2.929698106527e-04, {   0,   2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.415111950654e-04, -1.419987571771e-04, {   2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.989468768921e-04,  1.845421703846e-04, {   2,   0,   0,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.045543136183e-05,  2.925949161132e-04, {   0,   2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.828885573808e-04,  1.377419851957e-07, {   0,   2,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.487989612383e-04, -1.341825660210e-04, {   4,   0,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -7.125121455002e-05,  3.110809849292e-04, {   1,  -1,  -1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.815005149705e-04, -7.686031354330e-07, {   2,   0,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.624712021194e-04,  1.979232195871e-05, {   1,   0,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.469814944472e-04,  1.350951743227e-04, {   1,   0,   0,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 { -1.479561320360e-04,  2.340990765792e-04, {   2,   0,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -3.270909204548e-04,  5.384664236184e-05, {   2,   0,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  2.487805418668e-04, -1.320989255164e-04, {   2,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  3.799999999167e-04, -7.958701388588e-09, {   0,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.799999999716e-04,  4.642575810189e-09, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.627835502090e-05,  3.437080912822e-04, {   4,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.288836326497e-04, -2.509936457637e-04, {   0,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  3.260027155415e-04,  5.359348344386e-05, {   2,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  2.521080448169e-04,  1.272795391353e-04, {   0,   0,   0,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.311029307121e-04,  1.482775310805e-04, {   2,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.629664196481e-05, -2.926888170003e-04, {   1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.962619070091e-04, -1.820845784087e-04, {   4,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.015475905951e-04,  2.767541427848e-04, {   1,   1,  -1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -3.325385265131e-04,  4.559608583876e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 {  3.775911639616e-04,  6.093128858500e-08, {   2,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.946412508191e-04,  1.829968953479e-04, {   0,   0,   1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.758205303393e-04, -2.008380946741e-04, {   2,   0,  -2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -7.203851319187e-05,  3.045925553246e-04, {   4,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -3.061711610918e-04, -7.040852304230e-05, {   0,   0,   2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.028884704657e-04, -7.358702590802e-05, {   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.889694110813e-04, -8.734627099602e-05, {   0,   0,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.388973269825e-04, -2.371243545662e-04, {   0,   0,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.028583794433e-04, -2.728214718027e-04, {   2,   0,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  3.123721369893e-04,  6.209842008223e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  3.386242971951e-04, -3.583324518741e-05, {   1,   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -3.738288300805e-04,  6.000489252645e-07, {   2,   0,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.148998940973e-04,  5.928116047432e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.504833157245e-04,  1.233309101215e-04, {   2,   0,  -1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.639642826844e-04,  9.407566508115e-06, {   3,   1,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 {  7.641398642261e-05,  2.968263772500e-04, {   1,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.259315574985e-06,  3.719790626857e-04, {   4,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.730281745271e-04, -1.377928439230e-07, {   4,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.498303477641e-04,  1.231650455750e-04, {   4,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.997393022184e-04,  7.276242309767e-05, {   1,  -1,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -3.707378572386e-04,  9.003379624571e-07, {   2,   0,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.809851940599e-04,  8.914281939597e-05, {   0,   0,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  3.656864154050e-04, -4.368999026357e-06, {   2,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.699999998333e-04, -1.110727535806e-08, {   2,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.699999998973e-04,  8.717919612905e-09, {   0,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.699999999533e-04, -5.876523591248e-09, {   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.520266598874e-04,  1.758840806757e-05, {   2,   0,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -2.457235208742e-04,  1.232959316878e-04, {   1,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.179377983059e-05, -2.971403177676e-04, {   1,   1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.813005043006e-04, -1.874521338391e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  3.066265046512e-04, -6.202531732129e-05, {   2,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.120172492059e-04,  5.567796284239e-05, {   0,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  3.503395303744e-04,  1.692441350165e-05, {   2,   0,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  1.788916344418e-04, -1.881998493550e-04, {   0,   0,   0,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 {  1.153420479201e-04,  2.515240514748e-04, {   1,   1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.139392745508e-05, -3.254480876610e-04, {   4,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.961262283120e-04, -7.045039697426e-05, {   2,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.874201916233e-05, -2.971110510690e-04, {   1,  -1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.981675340200e-05,  2.959405853105e-04, {   1,  -1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.109618001212e-04,  2.533673285448e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -3.429793195748e-04,  2.110464021481e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.640268732057e-04,  1.999102816356e-04, {   4,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.635786493234e-04, -6.453747207171e-08, {   3,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.470111162492e-04, -1.596178501338e-05, {   0,   0,   1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.419655414213e-04,  1.208500654102e-04, {   0,   0,   5,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.385055410423e-04, -2.403217791574e-05, {   0,   2,  -1,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  1.372279714030e-04, -2.252424523014e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -3.412069197850e-05,  3.275712401551e-04, {   0,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.212214881337e-05, -2.893894106194e-04, {   0,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.928602771055e-04,  1.683403409187e-04, {   2,   0,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  6.878288859958e-05, -2.919242406003e-04, {   1,   1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.921806268076e-04,  6.839624838555e-05, {   1,   1,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  3.593713372826e-04,  6.970784800510e-07, {   2,   0,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.905141419025e-05,  2.608189395536e-04, {   0,   0,   2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.479677151916e-04,  1.172741896290e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  6.937141437036e-05, -2.902717142945e-04, {   1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.771623406171e-04,  8.228406340170e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -2.878890316923e-04, -7.145826885340e-05, {   1,  -1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.588709309447e-04, -1.454732397690e-07, {   2,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.556665369126e-05, -2.634021176802e-04, {   1,  -1,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.312994070512e-05, -3.456885406139e-04, {   0,   0,   2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.984091255741e-04,  6.030345742747e-05, {   1,   1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.845955880647e-04, -7.408383478900e-05, {   2,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.256167908257e-04, -3.297485258203e-05, {   2,   0,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.127362615248e-05, -3.469459021067e-04, {   2,   0,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.952763941744e-05, -3.286763126701e-04, {   1,   1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.952000988599e-05,  3.286769980031e-04, {   1,  -1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.341534873271e-05,  2.747470493231e-04, {   2,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  8.333655290050e-05,  2.745129885970e-04, {   2,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.560603381582e-04, -1.724174066081e-06, {   2,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.940619615154e-05,  2.883142568729e-04, {   2,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.354226095166e-05,  2.840452187810e-04, {   2,   0,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  3.552077418976e-04,  2.240415370951e-06, {   0,   2,  -1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 { -6.293847029865e-05,  2.943414366216e-04, {   2,  -2,  -1,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 {  4.741973586827e-05, -3.095340782076e-04, {   1,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  7.283550301987e-05,  2.835173977386e-04, {   1,  -1,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.524732963276e-04, -3.668347375975e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -3.531266482012e-04,  2.967910694331e-06, {   2,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.556997555163e-04,  3.661367904145e-07, {   2,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.949307809784e-05,  2.763400493851e-04, {   2,  -2,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.426723755784e-04, -2.129741665669e-04, {   1,  -1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  7.941392845665e-05,  2.761251937861e-04, {   2,  -2,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.299218447729e-04, -2.255259805385e-04, {   0,   0,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.543744632421e-04,  7.094735465100e-08, {   2,   0,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.119017676880e-04,  4.185887336340e-05, {   1,   1,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.170351383280e-04,  3.649988611792e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.492168431442e-04, -2.038121321170e-04, {   2,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.550168395098e-05, -2.970024692775e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 {  3.522210144397e-04,  3.782299679120e-08, {   4,  -2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.537326083101e-05, -2.968391364024e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 {  3.479767212367e-04, -3.793313734819e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.679355062605e-04,  1.838265972774e-04, {   0,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  3.504577819513e-04, -8.857927657756e-07, {   2,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.950121599542e-04,  1.560781586785e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  2.252977386549e-04,  1.256736004680e-04, {   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.578984113866e-04,  1.927351628661e-04, {   2,   0,  -1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -3.147494052137e-04, -3.579791335767e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -2.261574739798e-04,  1.242287885375e-04, {   2,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.827642755096e-04,  6.744586170229e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -2.827376664403e-04,  6.743240853223e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 {  3.360001853910e-05,  3.165286533392e-04, {   4,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.499999865657e-04, -8.827177088576e-08, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.625130857919e-04,  1.872627638994e-04, {   2,   0,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.204489631512e-05, -3.371130501608e-04, {   0,   0,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.055210712405e-04,  2.433897640395e-04, {   2,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.336702049262e-04,  1.151879850346e-04, {   2,   0,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.193954570358e-05, -3.367931470086e-04, {   0,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.273608334681e-04,  2.213360570908e-04, {   2,   0,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  7.068135658265e-05, -2.778692204894e-04, {   0,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.069338002872e-05, -2.778375772522e-04, {   0,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.810186921597e-06, -3.411777732882e-04, {   2,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.109511310045e-05,  2.768400288984e-04, {   1,  -1,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  4.077096307091e-05, -3.070036026171e-04, {   4,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.556491547375e-04, -1.920685875441e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  7.566321277929e-05, -2.717391777084e-04, {   4,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.468255509224e-04,  3.086592985620e-07, {   2,   0,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -7.559118925732e-05, -2.714936190556e-04, {   4,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  9.611343026215e-05,  2.503308467921e-04, {   0,   0,   2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.240873126084e-04,  1.216531000161e-04, {   0,   0,   1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  2.820287799281e-04, -6.354866739352e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -1.511709778839e-04,  1.937681520176e-04, {   1,  -1,   1,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 {  1.135557083857e-04,  2.298636689599e-04, {   2,   0,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  6.538557613122e-05, -2.780250879177e-04, {   1,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.268259202038e-04,  1.163372826258e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  2.382909844811e-04,  1.046899117620e-04, {   2,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.349789618248e-04, -1.074550370231e-04, {   1,  -1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -6.776912831698e-05, -2.746543149082e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.565233899684e-04,  1.855387759970e-04, {   2,   0,   0,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 {  2.751566950799e-04, -6.636881997672e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -2.290310970488e-04,  1.122775112318e-04, {   2,   0,   0,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.480089749175e-04,  1.924208042961e-04, {   2,   0,  -2,   0,   0,   0,   7, -10,   1,   0,   0,   0,   0 } },
 {  2.270249858843e-04,  1.133759305106e-04, {   0,   0,   3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.675961449630e-05,  3.032794675903e-04, {   2,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.399999999709e-04, -4.450589592522e-09, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.177604961881e-05, -3.279617393412e-04, {   0,   0,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.391052184285e-04, -4.895908941686e-07, {   0,   0,   1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.283465012751e-04, -1.050279262086e-05, {   2,   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.625454662756e-04, -7.615799818609e-05, {   2,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.368005976370e-05,  3.243616707540e-04, {   2,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.943859050690e-05, -2.685521284874e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -1.472762156680e-04, -1.906011056947e-04, {   0,   0,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -2.870464775883e-04, -5.079717648824e-05, {   2,   2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.249353576980e-06, -3.343179794789e-04, {   0,   0,   2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.968072961323e-04,  4.018778187689e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -6.466958027986e-05,  2.721952359104e-04, {   1,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.774315099868e-05,  3.087560810295e-04, {   1,  -1,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.344882602779e-04, -2.020089512005e-04, {   4,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.636961082993e-04,  1.727947679969e-04, {   2,   0,  -1,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 {  2.503070885675e-04,  8.610288650815e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -3.068395114752e-04,  2.955245483357e-05, {   4,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.327774033099e-05,  2.729138075252e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  1.496425298076e-05, -3.211252172454e-04, {   4,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.360254870076e-04,  6.356797438960e-08, {   3,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.348408825671e-04,  8.971235114525e-07, {   2,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.106944546810e-04, -2.356457506540e-05, {   0,   0,   2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -3.339100365510e-04,  1.357544922035e-07, {   1,   0,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.317759240556e-04, -1.836885582325e-06, {   2,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.216878703812e-06,  3.268716456949e-04, {   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.165074190134e-04,  1.165343099726e-04, {   2,  -2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  3.178165958318e-04,  1.390856723596e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -3.274291479853e-04, -3.915559776270e-06, {   1,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.389334015356e-04,  1.920083563288e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.291369777119e-04,  1.018000307362e-04, {   2,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.969891854303e-04,  3.354904491500e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -3.303877161128e-04,  7.312336755683e-08, {   0,   2,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.338241936876e-05,  3.167878926336e-04, {   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.226694373560e-04, -2.070529717787e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  2.246882357810e-04, -1.044499579174e-04, {   2,   0,  -1,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.191043519519e-04,  1.100025825470e-04, {   2,   0,  -1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.890920207666e-04,  1.397200909727e-04, {   3,   1,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  3.284128463802e-04,  3.323248254511e-07, {   2,   0,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.549386161752e-05, -2.731145486857e-04, {   1,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -3.250613355467e-04,  3.384628994449e-06, {   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.462965550718e-04,  1.820849218893e-04, {   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.228979932051e-04,  1.052691282483e-04, {   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.696458688400e-05,  3.109410388687e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -2.545730591331e-04, -7.329939778531e-05, {   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  3.237448427880e-04, -3.873544506620e-06, {   2,   0,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  6.200775020808e-05,  2.647318342514e-04, {   1,   1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.781004708195e-04, -4.846606801443e-05, {   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.302928199190e-05, -3.133406374012e-04, {   2,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.728142532480e-04, -1.530144073480e-04, {   2,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.172710124533e-04,  1.085535125788e-04, {   0,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.034439205687e-04,  2.223342549907e-04, {   0,   0,   2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.724843444504e-04, -1.528871628715e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   4, -11,   3,   0,   0 } },
 { -9.655423580957e-06,  3.156074564519e-04, {   2,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.564480374916e-05,  2.994294326667e-04, {   2,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.245991394387e-04, -4.475603477610e-07, {   2,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.056400782873e-04,  2.190058293615e-04, {   0,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -1.082286479652e-04,  2.161641330232e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 {  1.523805047555e-04,  1.713808929539e-04, {   2,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  4.882532038468e-05,  2.746011113680e-04, {   2,   0,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.952023155831e-04,  2.802935340478e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -1.850884736501e-04,  1.381161779730e-04, {   1,  -1,   2,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  6.170254766196e-05,  2.612270234348e-04, {   1,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.457106412069e-05,  2.382076744629e-04, {   2,   0,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -8.455041483266e-05,  2.381337564647e-04, {   2,   0,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -6.134116976115e-05, -2.600428272584e-04, {   1,  -1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.200457119929e-04,  4.902261506379e-07, {   2,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  6.121602855231e-05, -2.591977428861e-04, {   3,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.229758116346e-05, -3.079834010870e-04, {   2,   0,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.199999999421e-04, -6.087708430616e-09, {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.199999999761e-04,  3.909537524370e-09, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.255930788247e-04, -1.937459178047e-04, {   1,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.097898972384e-04,  1.090596651704e-04, {   2,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.261886761138e-05, -2.655833875509e-04, {   2,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -2.626485308013e-04, -5.545639303861e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  2.765648423548e-04,  4.128054168181e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -3.133256559645e-04, -4.240791079156e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.591081563813e-04,  1.583955879901e-04, {   2,   0,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -6.072590607037e-05, -2.566876752537e-04, {   1,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.922513231256e-05, -2.780937522084e-04, {   2,   0,  -1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.308596888127e-04, -8.609437658860e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -3.030629076753e-04,  1.377388269183e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  1.589884275066e-04,  1.577662018491e-04, {   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.774542329980e-04,  3.919832673264e-05, {   0,   0,   1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.291694032566e-04, -1.869816567317e-04, {   1,  -2,   0,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 {  4.517001476348e-05, -2.705369566580e-04, {   1,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.099171420696e-06, -3.074690528109e-04, {   0,   0,   1,   0,   0, -18,   8,  16,  -4,  -5,   0,   0,   0 } },
 {  2.034175363558e-04,  1.120914226763e-04, {   1,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.151924422840e-04,  3.145896620191e-07, {   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.098739954502e-04,  1.054607984826e-04, {   0,   0,   0,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.143405226186e-05, -2.538884715838e-04, {   4,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.599530895281e-04, -1.552565568778e-04, {   2,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.595392563787e-05,  2.888362749794e-04, {   1,  -1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.162683435061e-05, -2.531327537114e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.472328647393e-04,  6.665130720849e-05, {   2,   0,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  4.480771472761e-05,  2.686665209072e-04, {   2,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.503103439920e-05,  2.581902320197e-04, {   1,  -1,   0,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.804842079116e-04,  3.249317277455e-05, {   2,   0,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -3.122323467165e-04,  5.673361637902e-07, {   0,   0,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.467645845608e-04, -6.581510421275e-05, {   2,  -1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  9.135780899760e-06,  3.034019230558e-04, {   1,   0,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.494741444344e-06, -3.103832397731e-04, {   1,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.479739662957e-04, -6.384057043267e-05, {   4,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.953564848420e-05,  2.520258434518e-04, {   1,  -1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.994783236997e-04,  1.113726167539e-05, {   0,   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  1.855965195624e-04,  1.248074960198e-04, {   0,   0,   1,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 { -1.855486725848e-04,  1.247730233405e-04, {   0,   0,   1,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -3.101521246083e-04, -1.169369268245e-07, {   2,  -2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.096666578462e-04, -3.498809966364e-07, {   2,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.099999999609e-04,  4.923573819669e-09, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.743392323644e-05, -2.525086342254e-04, {   2,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.149281185805e-05, -2.982891850612e-04, {   2,   0,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.151260888014e-04, -1.941081192136e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -1.234895921393e-05,  2.968348484346e-04, {   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.359333356733e-04, -7.297014163830e-05, {   2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.055824649846e-04, -1.025113376005e-04, {   3,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.899598093010e-04, -1.805423026187e-05, {   0,   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  9.085446895966e-06,  2.989082902942e-04, {   2,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.058399509899e-04, -2.059081565215e-06, {   2,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.035738084282e-04,  4.097218543097e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  3.076233702306e-04,  3.957275191241e-08, {   2,   0,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.449903732753e-04, -6.213090952539e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  2.042655276721e-04, -1.020208727579e-04, {   2,  -2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.946058877087e-05,  2.867215084530e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  2.424791402020e-04,  6.353907478847e-05, {   2,   0,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  6.635418091529e-05,  2.392914076082e-04, {   2,   0,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.606022268882e-04,  1.449555337429e-04, {   2,   0,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.053862922696e-04, -1.326076212865e-07, {   2,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.050556637805e-04,  4.988488738250e-08, {   0,   2,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.305732548580e-04, -7.447435939214e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 {  2.225469797446e-04, -8.185734810803e-05, {   2,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.034364868734e-04, -5.625210964970e-07, {   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.506368397335e-05,  2.788759791479e-04, {   1,  -1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.110763690238e-05, -2.522532685605e-04, {   0,   0,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  2.123919788722e-04,  9.058726108075e-05, {   2,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  2.581984436830e-04, -4.468231286748e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.420809204110e-04, -1.605023179276e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -1.162157478775e-04, -1.859098173426e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.922556582531e-05,  2.828415652504e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  3.014191878898e-04,  6.275690548796e-07, {   0,   0,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.022203235836e-04, -1.994557578921e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  1.058820602897e-04, -1.952034978082e-04, {   2,   0,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  1.280467774167e-04,  1.729008215754e-04, {   1,  -1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.350588884476e-05, -2.574149803330e-04, {   2,   0,  -1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  5.554188321075e-05, -2.450138673268e-04, {   4,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.181477580237e-04,  1.823205479388e-04, {   0,   0,   2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -2.972210339193e-04,  3.158588904100e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 { -1.481792746980e-04,  1.521512422705e-04, {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.938031864835e-04,  1.064741974949e-04, {   1,   0,   2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.938031864835e-04,  1.064741974949e-04, {   1,   0,  -2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  1.722173701936e-04,  1.280087852911e-04, {   0,   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  2.938593328677e-04, -6.290330656709e-06, {   0,   2,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.802423026744e-04,  1.195866428363e-04, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  2.210915612232e-04, -7.851248422574e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.988819050924e-04,  9.285970808295e-08, {   4,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.986029753089e-04, -3.217990432200e-07, {   0,   0,   1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.849425820991e-04,  1.397549041128e-05, {   4,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.048250120091e-05, -2.480326629511e-04, {   2,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.157638028715e-04,  1.827319032316e-04, {   1,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.981443991328e-04, -2.096207618239e-07, {   2,   0,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.976267891743e-04, -5.081920669094e-07, {   1,  -2,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.221556695326e-06,  2.956546993124e-04, {   2,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.397328402089e-04,  5.813922820747e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.486881690772e-04,  4.902990924723e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 { -2.837240764700e-08, -2.976780762516e-04, {   1,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.981383581758e-04, -9.895440633659e-05, {   3,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.556715563231e-04, -1.412415164297e-04, {   4,  -2,   0,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.696430417221e-05,  2.398448499775e-04, {   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.750839469919e-04, -2.161569102858e-05, {   0,   0,   1,   0,  -5,   4,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.628525346778e-04, -3.380987814366e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  1.785913266119e-04,  1.179883428522e-04, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  8.730614430528e-05, -2.086438237297e-04, {   0,   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.343344483664e-04, -1.614103886005e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  5.412433533790e-05, -2.414016032455e-04, {   4,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.702356777908e-06, -2.895415158864e-04, {   0,   0,   3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.059892790195e-04,  1.883115974910e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  2.901679646692e-04, -3.482368221708e-06, {   2,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.888309196472e-04, -4.716629465003e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -1.714929210956e-05,  2.762410943249e-04, {   2,   0,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  1.230420040573e-04,  1.700851651016e-04, {   0,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  1.581853536791e-05,  2.771676203715e-04, {   2,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.229502882330e-04,  1.699833696419e-04, {   0,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -2.445111698730e-04, -4.832593994110e-05, {   2,   0,  -1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.722586922027e-04, -2.057289519517e-05, {   0,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  1.213180776975e-04,  1.710880652204e-04, {   1,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.439911187421e-05, -2.667417974886e-04, {   0,   0,   1,   0,   0, -20,  15,   8,   0,   0,   0,   0,   0 } },
 { -1.929545980352e-04,  9.793005692954e-05, {   2,   0,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.373205966991e-04, -5.349735111035e-05, {   1,  -1,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.616937541735e-04,  1.289757698601e-04, {   2,   0,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.440867338928e-04, -4.616984284418e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -5.530758321092e-05,  2.347462082040e-04, {   1,  -1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.899999999207e-04, -6.782349472627e-09, {   0,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.582643271226e-04, -3.144993414104e-05, {   2,   0,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.313115233139e-04,  1.576234486927e-04, {   2,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.694208828447e-04,  1.193164167528e-04, {   0,   0,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.271709482895e-04,  1.615051133452e-04, {   0,   0,   1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -1.271584412889e-04,  1.614869796903e-04, {   0,   0,   1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -2.852701157444e-04,  3.086080058110e-06, {   2,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -6.360291736999e-05,  2.240288557510e-04, {   1,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.469588610445e-04,  4.064211117914e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -1.855279193906e-04, -1.020655720126e-04, {   4,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.917392033547e-04, -9.581967469079e-05, {   2,  -2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.355790884729e-05,  2.237460156356e-04, {   1,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.839549342769e-04,  3.241694434943e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.875292846143e-04,  9.884817343740e-05, {   0,   0,   1,   0,   0,   0, -34,  39,   0,   0,   0,   0,   0 } },
 {  1.855062053327e-04, -1.008670566904e-04, {   2,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.133737753580e-04,  7.247832862468e-05, {   2,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.205345416155e-05,  1.937492684501e-04, {   2,  -2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  2.836355495868e-04, -2.106978961728e-06, {   2,  -2,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.196590571815e-05,  1.935838607489e-04, {   2,  -2,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.816306607732e-04,  1.037323656849e-04, {   4,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.385291397035e-04,  4.652153385394e-05, {   2,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.194679978596e-04, -1.646808730738e-04, {   0,   0,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.833196141348e-04,  1.651491150401e-07, {   2,   0,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.497360616815e-05, -2.682182984319e-04, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -1.484748884997e-04,  1.343663464693e-04, {   0,   0,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.832324285134e-04,  9.946698663866e-05, {   0,   0,   1,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 { -9.305895936548e-05, -1.892737212344e-04, {   2,   0,  -2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  6.145926163634e-05, -2.207216473557e-04, {   4,   0,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -6.145399921237e-05,  2.207167671968e-04, {   1,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -6.144569433603e-05, -2.207009998720e-04, {   4,   0,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  6.144265658433e-05,  2.206723711261e-04, {   1,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.816745052251e-04, -3.901438167719e-07, {   2,  -2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.396732014922e-04, -4.232350333906e-05, {   2,  -2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.916208414938e-04,  9.035408468317e-05, {   2,   0,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 {  2.382023065651e-04, -4.375608816938e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.390949613238e-04, -1.428309085440e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -6.988354522782e-05,  2.119181077018e-04, {   2,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.393221276224e-05, -2.273508804595e-04, {   1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.383520326860e-05, -2.271421124387e-04, {   1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.316898825404e-05, -2.276222648472e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -2.801755935872e-04, -6.495246360080e-08, {   0,   0,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.739228438356e-05,  2.226491023842e-04, {   3,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.799999999323e-04,  6.157521600540e-09, {   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.720051198177e-04,  1.079859302130e-04, {   2,   0,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.839695631616e-06, -2.771364901458e-04, {   2,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.218931124932e-04,  1.579587758471e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.273115374807e-04, -5.252061923586e-05, {   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.792206619090e-04, -4.022902682728e-07, {   2,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.255049135307e-05,  1.970542590365e-04, {   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.362069431169e-05, -1.958991755157e-04, {   2,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.235820500559e-04, -5.590892571253e-05, {   1,   1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.878454224235e-04,  9.162841015133e-05, {   2,   0,   0,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.259375498831e-04, -5.331037765002e-05, {   1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.789816016843e-04, -4.091396618860e-08, {   2,   0,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.793491921910e-04,  9.958358521708e-05, {   2,   0,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -2.488563851999e-04, -2.998976645080e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -2.787402172354e-04,  3.436251144710e-08, {   2,   2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.694871548523e-04,  1.091724798566e-04, {   0,   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  2.781850077783e-04,  2.417609382179e-07, {   1,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.326466382754e-05, -2.246204117010e-04, {   2,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.506142139078e-04, -2.678870756379e-05, {   0,   0,   0,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  1.268814514541e-04, -1.503080610270e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -1.019158883600e-08, -2.769820541799e-04, {   1,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.044349861530e-06,  2.758042642266e-04, {   4,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.090868097646e-04, -1.676307487766e-04, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.372086550868e-05,  2.329894599286e-04, {   0,   2,  -2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.641872509094e-04, -1.123622859861e-04, {   2,   0,  -1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -1.804924031561e-04,  9.605676336725e-05, {   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.051046834131e-04,  1.713994807282e-04, {   0,   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -2.322211153137e-04, -4.407643760571e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.762140625679e-04, -6.402728380241e-08, {   2,   2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.277852588298e-05,  2.132489709430e-04, {   2,   2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  6.271207615680e-05,  2.130462980187e-04, {   2,   2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  9.416525408361e-06, -2.662716360478e-04, {   0,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.245145123494e-04, -5.111245488469e-05, {   2,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.347858211772e-04, -1.408145476267e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -2.392235166385e-04,  3.607159355845e-05, {   2,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.821045390592e-04, -9.288301661156e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 { -1.051518595207e-04, -1.696863683567e-04, {   2,   0,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  8.086779159275e-05, -1.938050574235e-04, {   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.219668880624e-05, -1.922159724369e-04, {   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.363966383378e-05,  2.606390767731e-04, {   0,   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  2.731007489870e-04, -1.158839801910e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.642680319503e-05, -2.573639586768e-04, {   0,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.729163745802e-04,  5.916791935221e-07, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  1.646569932019e-04,  1.085852344538e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 {  7.158351252095e-05,  2.016095524710e-04, {   2,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -7.157095813542e-05,  2.015682646552e-04, {   2,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.892749385597e-04, -8.385901478095e-05, {   2,   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -2.675889655207e-04,  5.227050856532e-06, {   1,   1,   0,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.513574254761e-06, -2.651012040864e-04, {   0,   0,   3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.359296419410e-05,  2.589289746181e-04, {   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 { -1.021660474058e-04, -1.703284707158e-04, {   4,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.721634994967e-04,  7.740390577640e-08, {   2,   0,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.265468600273e-05, -2.293833541528e-04, {   1,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.715133979222e-04,  1.226262036122e-07, {   0,   0,   1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.376107922151e-04,  3.378034077586e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.690345280921e-04, -2.183698734768e-06, {   2,   0,   0,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.649861388630e-04, -6.077884773460e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -2.193247246056e-04,  5.119796575052e-05, {   2,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.081490877478e-04, -1.622657358675e-04, {   2,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.598997712905e-04, -1.041449584940e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.649949547931e-04,  5.176800340018e-06, {   1,   1,  -2,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.239933256896e-04, -4.614528636213e-05, {   2,   0,  -1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  2.189740548448e-04,  5.109072281012e-05, {   2,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  0.000000000000e+00, -2.700000000000e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.208886762138e-05,  2.177165167688e-04, {   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.554128827683e-04, -1.418061780966e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  8.196468731079e-05, -1.875101463487e-04, {   4,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.692490388817e-04, -1.892326532297e-07, {   0,   0,   1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.408628962305e-04,  2.853930923083e-05, {   2,   0,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  1.062127680073e-04, -1.630958745221e-04, {   2,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -9.578430141300e-05, -1.730744538098e-04, {   0,   0,   2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.865664536014e-04,  8.203888992240e-05, {   1,   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  5.579950812242e-05,  2.125366126380e-04, {   1,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.073199684149e-04, -1.609394070046e-04, {   2,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.737974094456e-05, -2.507211773805e-04, {   2,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.782793715856e-04,  8.945789828245e-05, {   2,   0,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.072899221588e-04,  5.996696310709e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 {  1.748485796654e-05, -2.496389327609e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.619538346070e-04, -5.169593210259e-06, {   0,   2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  7.588596563127e-05,  1.910950910491e-04, {   1,  -1,  -1,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.292764052244e-04, -1.376425813892e-04, {   1,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.663140763203e-04,  1.004574794540e-04, {   0,   0,   1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.616040879864e-04,  1.050029475651e-04, {   2,   0,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -1.644199455365e-04, -1.019969173148e-04, {   0,   0,   1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -5.904706927244e-05, -2.072953729798e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.657546542700e-04, -5.314896168691e-07, {   6,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.545996067605e-04, -1.165470041343e-05, {   0,   0,   1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  2.614133688638e-04, -4.769207974546e-06, {   3,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.203085843695e-04,  4.584775819438e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  2.595611426713e-04,  6.580183104083e-06, {   4,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  5.070540735408e-05,  2.151214210682e-04, {   1,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.482036718937e-05,  2.409373073674e-04, {   0,   0,   0,   0,   0,   0,   6,  -8,   3,   0,   0,   0,   0 } },
 { -2.359095304892e-04, -2.976640078092e-05, {   3,   1,  -3,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.175511733318e-05,  2.537211591185e-04, {   0,   2,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.753855909353e-04, -9.008639866222e-05, {   2,   0,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.748730509009e-04,  9.043271533280e-05, {   2,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.520663455004e-04,  1.132383049563e-04, {   3,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.071025263887e-05,  2.545040264151e-04, {   2,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.441623620389e-05,  2.103182215242e-04, {   1,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.194830369739e-04,  1.451328502677e-04, {   4,   0,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.642356072050e-04,  2.634090112950e-07, {   1,  -2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.602758056049e-04,  1.041224826103e-04, {   2,   0,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.928063242073e-04, -7.157248179253e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  2.479065298635e-04,  1.610413427783e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -7.842163877329e-05,  1.855725929462e-04, {   4,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.224475272356e-04, -4.041063989593e-05, {   1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.905884561229e-05, -2.137601256095e-04, {   1,   1,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  2.032025974850e-04, -5.925221483345e-05, {   1,   1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.622746862023e-04, -2.159107796388e-08, {   0,   0,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.427541222198e-05, -2.077216075396e-04, {   1,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  7.655969785134e-05, -1.852569949931e-04, {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  9.337948795792e-05, -1.684365048949e-04, {   0,   0,   0,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 {  8.287657056958e-05,  1.787875478937e-04, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.683235136001e-04, -9.297015610824e-05, {   3,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.182465206874e-04,  4.278656741051e-05, {   1,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.748339712378e-04, -8.586312170393e-05, {   2,   0,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.388173045513e-04, -2.175192114184e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  2.599999999310e-04, -5.989969992281e-09, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.599999999811e-04,  3.131120678002e-09, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.599999999822e-04,  3.040363556905e-09, {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.599999999927e-04,  1.951278103711e-09, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.330164493458e-04,  1.264977841691e-04, {   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.585909284880e-04,  7.757911466016e-07, {   2,   0,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.208673604482e-06, -2.500277029225e-04, {   0,   0,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.337165745620e-05,  1.654742686459e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.741454175107e-04,  8.468810296858e-05, {   2,   0,   0,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.885136057476e-05, -1.899196657612e-04, {   2,  -2,   0,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 { -1.574936374194e-04, -1.010325062543e-04, {   4,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -2.583965016018e-04,  9.478131803361e-08, {   2,  -2,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.030055123005e-04,  5.514372286223e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.178521782701e-04, -1.401308554087e-04, {   1,  -1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.080630149221e-05, -2.070952217353e-04, {   1,  -1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.573471220921e-04, -1.407317194749e-07, {   0,   2,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.574308910377e-04,  5.278378175697e-08, {   0,   0,   1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.621813856792e-05, -1.611490678560e-04, {   2,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.572103889840e-04,  1.078412521306e-07, {   0,   0,   3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.266827775006e-04,  3.055537094050e-05, {   2,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.091061289953e-04, -1.479307584000e-04, {   0,   0,   3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.090450213692e-04, -1.478718872192e-04, {   0,   0,   3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.729266541611e-04,  8.362922369319e-05, {   2,   0,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.714707035872e-04,  8.495830364901e-05, {   2,   0,  -1,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.459123791708e-04, -1.050107767798e-05, {   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.058683738003e-04, -1.504773994722e-04, {   1,   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.719054685837e-04,  8.434307909387e-05, {   0,   0,   1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.705795505598e-05,  1.687905843735e-04, {   2,   0,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -1.082117466723e-04, -1.475991404569e-04, {   0,   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  5.363356604822e-05, -2.021727329379e-04, {   0,   2,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -5.363356604822e-05, -2.021727329379e-04, {   0,   2,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  2.546537622867e-04, -1.119905478717e-06, {   2,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -5.154126050074e-05, -2.040194817782e-04, {   3,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.324320036120e-04,  2.312028850068e-05, {   0,   0,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  1.909433852613e-04,  6.459788601759e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.529255904730e-04, -2.503628993021e-06, {   2,   0,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  8.483351042471e-06,  2.466555178318e-04, {   2,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.694782487489e-04,  8.506244704359e-05, {   0,   0,   0,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.535209161540e-04, -7.880937360702e-07, {   1,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.540100500130e-04,  8.145262022750e-08, {   4,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.551621868455e-05,  1.585118960666e-04, {   1,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.532988678166e-04, -6.709747287431e-07, {   1,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.367555335391e-06, -2.455607714825e-04, {   2,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.868970100232e-05, -2.046889180402e-04, {   3,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.450915803812e-04,  1.082220229272e-04, {   0,   0,   1,   0,   0, -18,  16,   0,  -2,   5,   0,   0,   0 } },
 { -2.489096611742e-04,  4.340533015838e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  7.583731705813e-05,  1.771069816920e-04, {   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.121500414332e-05,  1.914678670558e-04, {   1,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.751563644300e-04,  7.752466953970e-05, {   2,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.840945974086e-05,  2.040979504572e-04, {   1,  -1,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.612453532233e-05, -1.862904822040e-04, {   0,   0,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.307184362711e-04, -2.169421626915e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -2.495967833234e-04,  2.701610451922e-06, {   0,   0,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.173879094479e-05, -1.604771193739e-04, {   0,   0,   3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.683079690372e-04,  8.367053073583e-05, {   2,   0,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.135051363796e-04, -3.845767261500e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  6.112052589324e-05,  1.907575616106e-04, {   1,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.516633984828e-04, -8.088103955460e-08, {   1,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.521655912733e-06,  2.499953690837e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.312551605640e-05, -1.683160414740e-04, {   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.174428503839e-05,  2.296802470819e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.478249231508e-05,  1.761710988525e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -3.744682996869e-05, -2.135018871378e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 { -1.770178776723e-05,  2.332092141095e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  1.875414572366e-04,  6.327618106489e-05, {   1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.084615222851e-05,  2.399234361852e-04, {   2,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.346953732300e-05, -1.571904457027e-04, {   2,   0,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  5.725832191082e-06,  2.448128952321e-04, {   0,   0,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.674626577283e-04,  8.289179615818e-05, {   0,   0,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.627278384308e-04,  8.761281640005e-05, {   2,   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.696435739436e-05, -1.930502579178e-04, {   2,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.499999999316e-04, -5.846852993644e-09, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.499999999951e-04, -1.570796326787e-09, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.613225672381e-05, -2.237309748862e-04, {   2,   0,  -1,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 { -2.057419251002e-05, -2.290779400173e-04, {   1,  -1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.751143547738e-04, -7.407481355657e-05, {   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.427899129374e-04,  6.122016226975e-06, {   2,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.413041003492e-05, -1.944135207819e-04, {   2,  -2,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -5.413062524622e-05, -1.944114945085e-04, {   2,  -2,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.484769284371e-04, -2.444787568817e-08, {   2,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.749938826485e-05, -1.508465763472e-04, {   2,  -2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  9.501901374500e-05, -1.531977604610e-04, {   2,   0,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.654051171453e-04,  8.261188825651e-05, {   2,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.784529169841e-05,  2.000800887365e-04, {   1,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.395138935521e-04, -8.275117703409e-06, {   4,   0,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.407523959210e-04,  6.815142847998e-06, {   1,   0,   1,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 {  4.168936036545e-05, -2.058703115316e-04, {   0,   0,   1,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.590283620410e-04,  8.753948823016e-05, {   2,   0,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.651830009807e-04,  8.124912703089e-05, {   2,   0,  -1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.445680556681e-04,  1.017150605597e-04, {   1,  -1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -2.224235715322e-04,  2.359019460927e-05, {   1,   1,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  6.535108261980e-05, -1.801178885353e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  1.255277024765e-05, -2.323486810592e-04, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.896185762774e-04, -5.484655634419e-05, {   1,   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.530492967048e-04, -9.129332700599e-05, {   0,   0,   1,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 { -4.679053995321e-05,  1.975293738063e-04, {   1,  -1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.077438975829e-04, -3.641186771720e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -1.886626113869e-04, -5.541652045471e-05, {   0,   0,   1,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -1.627346802146e-04, -8.105642441037e-05, {   0,   0,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.965182910612e-04, -4.722719926702e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  1.880590355647e-04, -5.523747906984e-05, {   0,   0,   1,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  2.424843253939e-04,  2.535565931313e-07, {   2,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.637445504899e-05,  1.963248638887e-04, {   1,   1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.424236442336e-05, -1.884105429372e-04, {   0,   0,   1,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -9.355325277224e-05,  1.490701916052e-04, {   2,   0,  -1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 { -1.477325004407e-04, -9.476436229354e-05, {   1,   0,  -1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.701041015118e-04,  7.228650952782e-05, {   0,   2,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.537368556130e-05,  1.965150923615e-04, {   1,   1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.243506585783e-04, -1.744108793827e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  1.696442789385e-04,  7.208293720114e-05, {   0,   2,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.943702923004e-05,  2.221586998421e-04, {   2,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.060776064935e-05,  1.708071225079e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 { -4.702805619093e-05, -1.943625049198e-04, {   4,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.080508800834e-04,  3.320148831446e-05, {   2,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.079791345770e-04, -1.329654242339e-04, {   1,  -1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.519049688000e-04,  8.900815536708e-05, {   0,   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  7.861606961722e-05, -1.622934594829e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 { -2.233976809624e-04, -1.726815011782e-05, {   0,   0,   1,   0,   0,  -8,   3,  -7,   8,  -6,   0,   0,   0 } },
 {  1.580338649868e-04, -8.251026734852e-05, {   2,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -2.399999999656e-04,  4.063126498449e-09, {   2,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.399999999778e-04,  3.267256359632e-09, {   4,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.796510410398e-05, -1.413122781841e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 } },
 {  1.234930801095e-06,  2.378185513741e-04, {   0,   0,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.637453771072e-05,  2.025668660815e-04, {   0,   0,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.969212968831e-05, -2.191169094989e-04, {   1,  -1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.968799942837e-05, -2.191172806478e-04, {   3,  -1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.302275933566e-05, -1.756197196340e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -2.384811597594e-04, -1.015906896476e-07, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  5.137080909396e-05,  1.867417536854e-04, {   0,   0,   1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -5.138300879239e-05,  1.866535416065e-04, {   0,   0,   1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  1.586141627274e-04,  7.919963310921e-05, {   2,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.375742034026e-04, -3.000904798056e-08, {   2,   0,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.375408225497e-04, -1.029614222447e-08, {   2,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.421067314095e-04,  9.536348609716e-05, {   0,   0,   2,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 {  4.521254725546e-05,  1.922514627187e-04, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 { -2.496938769715e-05,  2.124022563477e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.115262392089e-04, -1.253553519316e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  9.474925869244e-05, -1.415971682118e-04, {   4,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.268419742416e-04,  9.488314694976e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -2.358627559202e-04,  1.393287789995e-07, {   0,   2,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.506407686116e-04,  8.468032784255e-05, {   4,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.991688614320e-04, -3.607203995975e-05, {   3,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.159845142939e-05, -1.733703551665e-04, {   2,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.298118596714e-04,  1.050724994106e-04, {   3,  -1,  -1,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0 } },
 { -4.402368476111e-05,  1.908251001274e-04, {   1,  -1,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  9.118848466183e-05, -1.436429944412e-04, {   2,   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.970115402630e-04,  3.774564827231e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 { -6.145090386948e-05, -1.728381698484e-04, {   2,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.133379542076e-05, -1.729079421025e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.044648430479e-04,  1.297424107041e-04, {   2,   0,  -1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  4.548242823362e-05,  1.886944117025e-04, {   2,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.516446282388e-04,  8.230219516981e-05, {   0,   0,   1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.477783387875e-05, -1.891141332452e-04, {   1,   1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.783044168641e-05, -1.360113768584e-04, {   4,   0,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.133275948621e-04,  2.042095959802e-05, {   1,  -1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -1.039354946549e-04,  1.291595417044e-04, {   2,   0,  -1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  2.050036937382e-04,  2.765842147777e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  1.670296721080e-05,  2.157511744657e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -8.626787852276e-05, -1.460951814422e-04, {   2,  -2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  3.727418621849e-05,  1.950374638675e-04, {   2,   0,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  1.503565754091e-04, -8.188897148653e-05, {   2,   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -4.422574936567e-05,  1.879831549674e-04, {   1,  -1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.320471654271e-05,  1.889156540692e-04, {   4,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.873658744995e-06, -2.251416212620e-04, {   4,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.745266118758e-05, -1.944188736765e-04, {   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.080057108687e-04,  2.306793370828e-05, {   2,   0,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.186877961747e-04,  1.233965231515e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 {  2.279480761712e-04, -2.787411430111e-06, {   2,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -7.179673551580e-05,  1.589327472320e-04, {   1,  -1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  8.871331086185e-06, -2.217443456618e-04, {   2,   0,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.660632509837e-05, -1.437720020140e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -8.658324091096e-05, -1.437342305120e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  1.535866937251e-04, -7.670917670879e-05, {   2,   2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.299999999399e-04, -5.258677035759e-09, {   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.299999999808e-04, -2.970550386842e-09, {   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.428637841520e-05,  2.052068118919e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.734696846326e-06, -2.262932078792e-04, {   2,   0,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.370197230118e-05, -1.852800400578e-04, {   1,  -1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.913268826885e-04, -3.760494958438e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  3.249703917757e-05, -1.963528802484e-04, {   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.157710697247e-05,  1.470245362342e-04, {   2,   0,  -1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -2.284623141382e-04, -8.380242682541e-08, {   3,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.695260710461e-05, -1.415725195848e-04, {   0,   0,   2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.152940577312e-04, -1.316668377263e-05, {   0,   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -2.267035390264e-04,  1.674715427525e-06, {   0,   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.600277578865e-05,  2.014097446795e-04, {   0,   0,   0,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 { -2.154051231895e-04, -1.183613154115e-05, {   0,   0,   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -2.025102471835e-04,  2.436259833278e-05, {   2,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.292609436717e-04,  9.759425485238e-05, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.934369463341e-04, -3.329370967220e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.920929984715e-04, -3.449969271973e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.010011145828e-04,  1.254630782535e-04, {   2,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.259328523756e-04,  5.079317222180e-07, {   0,   0,   1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.509760611669e-04, -7.540398534092e-05, {   2,   2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.330980525158e-05,  1.330358664860e-04, {   2,   0,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -1.709476062042e-04, -5.473472764990e-05, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.352716193189e-04,  9.031292657863e-05, {   1,  -1,  -1,   0,   2,  16, -20,   0,   0,   0,   0,   0,   0 } },
 {  7.829892780794e-05,  1.463391683849e-04, {   0,   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  4.220175027807e-05, -1.823181536033e-04, {   0,   0,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.185785271583e-05,  1.326309485683e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  1.575135459034e-04,  6.692732795624e-05, {   1,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.575135459034e-04,  6.692732795624e-05, {   1,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.512853631838e-05, -1.992670726052e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  2.069388456551e-04, -1.664871937453e-05, {   4,   0,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.334490604403e-04,  8.951663269675e-05, {   0,   0,   0,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  2.212320102332e-04, -1.606379753028e-06, {   3,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.453393497375e-06, -2.198764676222e-04, {   0,   0,   2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.760507813335e-06, -2.204716057317e-04, {   2,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.094399002022e-04,  1.125940737373e-04, {   0,   0,   0,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  7.888767198672e-05,  1.431071252452e-04, {   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.111588864896e-04, -1.078576899883e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -1.864308971685e-04,  3.498565599308e-05, {   1,  -1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.753024248564e-04, -4.610157359316e-05, {   0,   0,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.545190893223e-04,  6.687540696323e-05, {   1,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.218951331586e-05,  1.791998228375e-04, {   1,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.135844149782e-04, -7.777356410831e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -6.111136774827e-05,  1.599289166483e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.207808144129e-04, -3.925894077531e-08, {   2,  -2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.914629280586e-04,  2.932571095368e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 {  2.199991146779e-04,  5.950865836863e-07, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.071396742484e-05,  1.798580403214e-04, {   4,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.312796997408e-06, -2.132087423497e-04, {   0,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  6.761517902263e-06,  2.134967257215e-04, {   2,  -2,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.199999999735e-04,  3.417354675267e-09, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.199999999831e-04, -2.726204291538e-09, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.199999999831e-04, -2.726204291538e-09, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.170566953044e-04,  2.678755655871e-06, {   4,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.742612574738e-06,  2.128797346502e-04, {   2,   2,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -8.962656800459e-05,  1.295501185293e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  6.743632597777e-05,  1.516619384077e-04, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -4.456476951916e-05,  1.743967189841e-04, {   2,   1,  -1,   0,   0, -51,  58,   0,   0,   0,   0,   0,   0 } },
 { -1.850547851384e-04,  3.323363518282e-05, {   0,   2,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.008350777208e-06, -2.132265915750e-04, {   0,   0,   2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  7.421342637107e-05, -1.439786443614e-04, {   0,   0,   0,   0,   0,   5,  -2, -12,   3,   0,   0,   0,   0 } },
 { -1.790317095089e-04,  3.901226577402e-05, {   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.084727932718e-04, -9.549089476394e-06, {   0,   0,   1,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 {  3.634710702328e-05,  1.815864719055e-04, {   2,   0,  -1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -1.922306288573e-04, -2.562256602809e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -1.430934599084e-04,  7.475920586457e-05, {   2,   0,  -1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.726951479737e-04,  4.503114530399e-05, {   1,  -1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.020577458593e-04, -1.541968973191e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -2.078451810097e-04,  9.374187075455e-06, {   0,   0,   0,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 {  2.506472176176e-05, -1.920704656916e-04, {   0,   0,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -4.435946592135e-05, -1.727716645139e-04, {   1,   1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.435815096897e-05,  2.027541793892e-04, {   0,   0,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.035761390904e-05,  1.767385134469e-04, {   4,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.821112667083e-04, -3.448402118408e-05, {   1,   1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.004917684214e-04, -1.160037678827e-04, {   0,   0,   1,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 {  3.415293890601e-06,  2.130477996335e-04, {   2,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.490669977984e-04, -6.684433642438e-05, {   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.856650876936e-04, -2.997510695565e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -8.114743743940e-06, -2.075118102716e-04, {   2,   0,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.000705458629e-04, -1.155162335972e-04, {   0,   0,   1,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 { -2.153718233269e-04, -6.882015083207e-08, {   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.960791208056e-05,  1.753593076665e-04, {   0,   0,   1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 { -1.812978487557e-04, -3.358768508443e-05, {   1,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.802640193286e-05, -1.567121056878e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.040013979031e-04, -1.061398643404e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  6.361894920519e-05, -1.504838925551e-04, {   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.701441458929e-04,  4.392024867111e-05, {   1,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.573444330579e-06,  2.122107408079e-04, {   3,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.118271469070e-05,  1.220327481975e-04, {   2,   0,  -1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -9.111014370533e-05,  1.219281695301e-04, {   2,   0,  -1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -4.073254396989e-05,  1.721421381990e-04, {   3,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.262759163527e-05,  1.301720895429e-04, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  8.942887578780e-05,  1.232423825960e-04, {   0,   0,   0,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.208131021902e-04,  9.182854257739e-05, {   0,   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -6.739237013964e-06, -2.052206148187e-04, {   2,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.920947928579e-05,  1.727476695215e-04, {   0,   0,   1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 {  2.079183484251e-05,  1.906729146271e-04, {   2,   0,  -1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.307245059378e-04, -8.073828049901e-05, {   2,  -2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.781075924625e-04,  3.331727648153e-05, {   1,  -1,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.289264673131e-04, -8.247154840435e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  8.893540842550e-05,  1.223660350428e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 {  4.181677064802e-05, -1.694510302078e-04, {   4,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.730046949240e-04, -3.815129801207e-05, {   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.048891418793e-04,  5.917619675664e-06, {   2,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -5.858788056736e-05,  1.520704760575e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.923736050203e-04,  1.825713967332e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  1.713706480787e-04,  3.887945938174e-05, {   0,   0,   2,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  3.288854026920e-05,  1.771357759125e-04, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.099999999687e-04, -3.628539514722e-09, {   2,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.099999999699e-04,  3.555235686143e-09, {   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.129898556024e-05,  1.683627026094e-04, {   1,  -1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.266035135012e-05, -1.765025579583e-04, {   1,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.682602038520e-04,  4.085621213945e-05, {   1,  -1,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.423300639083e-04,  6.669016210053e-05, {   1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.968476851895e-04,  1.186307428142e-05, {   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.063983920130e-05,  1.373592380347e-04, {   2,   0,  -1,   0,   0, -20,  18,   6,  -2,   0,   0,   0,   0 } },
 { -2.929201630908e-05, -1.782640712340e-04, {   0,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.406936059057e-05,  1.433426150399e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  1.442999856806e-04,  6.275000662474e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.058118614207e-05,  1.862313183731e-04, {   2,   0,  -1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.449191075904e-04, -6.157555428774e-05, {   4,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.316727475455e-06,  1.981121454564e-04, {   2,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  5.455530573301e-05, -1.517492927682e-04, {   0,   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  7.107392713512e-05,  1.351042955192e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -7.816104919416e-06, -1.981283210752e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.058893491647e-04, -9.999871253294e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  4.636527482202e-05, -1.594599171189e-04, {   4,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.444484457883e-04, -6.136718322530e-05, {   4,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.052122793477e-04,  5.978838353251e-07, {   2,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  8.178597912436e-05, -1.237698495959e-04, {   0,   0,   3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.838366287842e-04, -2.169636104430e-05, {   2,   0,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -4.623798414021e-05, -1.590540361598e-04, {   4,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.035446229933e-04,  1.542802054772e-06, {   0,   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -3.839177917783e-05,  1.664704757257e-04, {   1,  -1,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.035176670640e-05,  1.843358981478e-04, {   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.266029516556e-05,  1.519136188622e-04, {   2,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.744517033088e-04, -3.011363818665e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  1.628152338968e-04,  4.174206379980e-05, {   1,  -1,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.019372452958e-04, -2.591944800901e-06, {   2,   0,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.372282204391e-04,  6.712084201025e-05, {   0,   0,   1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.662449419394e-04,  3.776795755620e-05, {   0,   0,   0,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -2.039331956407e-04,  2.017055143943e-08, {   0,   0,   2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.675775021821e-04, -3.616905485940e-05, {   4,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.367855439352e-05, -1.700412775950e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  6.562882556652e-05,  1.380281650426e-04, {   2,   0,  -1,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -1.962598555955e-04,  7.075047361504e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -1.335663297078e-04,  6.967567315300e-05, {   0,   0,   1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.378008780233e-05, -1.694041255090e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  1.314118948171e-04,  7.160143983432e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 { -4.932315770437e-05,  1.532598944676e-04, {   0,   0,   1,   0,   0, -21,  23,  -4,   0,   0,   0,   0,   0 } },
 {  1.593844638168e-05, -1.862752872842e-04, {   0,   2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.316486737377e-04, -7.054337254848e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 {  1.949701497877e-04, -7.103795499496e-06, {   0,   0,   0,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.700222286053e-04, -3.202569058539e-05, {   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.315082044701e-04, -7.049370786980e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 { -3.811332327688e-05,  1.636838090705e-04, {   1,  -1,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.223063271518e-04, -7.940855494910e-05, {   0,   0,   1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  1.195545378991e-04, -8.214218289478e-05, {   2,   0,   0,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  3.762564832812e-05, -1.640680647462e-04, {   2,   0,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.494964291399e-04, -5.188079276040e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -1.278970240741e-04, -7.329905212654e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 { -1.347501320052e-04,  6.597007474452e-05, {   2,   0,  -1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.343201460881e-04,  6.619393138078e-05, {   2,   0,  -1,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.686341831730e-05,  1.436388511722e-04, {   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.003229119030e-04, -3.840655530307e-08, {   2,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.637800185073e-04, -3.635191544134e-05, {   2,   2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.802915453584e-04, -1.974013023403e-05, {   2,   0,  -1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.374637705193e-05, -1.362794409445e-04, {   2,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.642476006637e-04,  3.573858440536e-05, {   1,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.716255759686e-04, -2.812894610901e-05, {   0,   0,   1,   0,   0, -26,  37, -16,   4,   5,   0,   0,   0 } },
 {  1.763376839590e-04, -2.335016286086e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.056376819329e-05,  1.889646093737e-04, {   2,   0,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.398371409598e-04, -5.968537165822e-05, {   2,   0,  -1,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 { -1.932408865710e-04,  6.120654723619e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.948629113714e-04,  4.477261772474e-06, {   3,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.225890659454e-05, -1.670668013655e-04, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -1.590762422708e-04, -4.010069659416e-05, {   2,   0,  -2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.958886243475e-04, -3.082532671246e-06, {   0,   0,   3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.540909677266e-04,  4.486090730609e-05, {   1,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.681234355879e-04, -3.071517709727e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  4.879523592902e-05,  1.497512703777e-04, {   2,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.144063538374e-05, -1.568654309738e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 { -1.093823712363e-04,  8.888034925449e-05, {   0,   0,   0,   0,   0,   0,  20, -38,   0,   6,   0,   0,   0 } },
 { -3.861436018360e-05, -1.595811868658e-04, {   2,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.753931853888e-05, -1.405974426106e-04, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.388302284403e-04, -5.901152834354e-05, {   0,   0,   2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.958005821087e-04,  1.894956637891e-06, {   2,   0,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -9.796261414027e-05,  9.969912712465e-05, {   0,   2,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.826272303245e-04,  1.488733491492e-05, {   0,   0,   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.520991337432e-04,  4.525408642041e-05, {   2,   0,  -1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  9.747066948934e-05,  9.975784121012e-05, {   0,   2,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.350737056368e-05, -1.436744796278e-04, {   4,   0,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.913356460854e-04,  5.752538576476e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.848848710012e-04,  1.200693221115e-05, {   2,   0,  -1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.678996064767e-05,  1.600629700985e-04, {   1,  -1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.745223529177e-05, -1.593702655364e-04, {   1,  -1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.266890269325e-04, -7.011832536118e-05, {   4,   0,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.459777357559e-04, -5.055329050339e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  1.956810507197e-04, -7.397474368148e-07, {   1,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.377375876510e-04, -5.852779518076e-05, {   0,   0,   4,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.144586967138e-05,  1.546834021997e-04, {   1,  -1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.647325995411e-05,  1.196414167728e-04, {   0,   0,   1,   0,   0, -12,   3,   7,  -8,   6,   0,   0,   0 } },
 { -1.300303176582e-04,  6.606938116424e-05, {   2,   0,  -1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.664334717040e-04, -2.965538331706e-05, {   3,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.513534788309e-05,  1.505546711204e-04, {   3,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.564717501032e-04, -3.913758364122e-05, {   1,   1,  -3,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.506348123846e-05,  1.503691336139e-04, {   3,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  7.938640451737e-05,  1.160207535309e-04, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.610927057482e-05, -1.792776935415e-04, {   1,  -3,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.902989554887e-05, -9.623847176832e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.660962674483e-04,  2.902993878296e-05, {   6,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.364379939749e-05, -1.414485701376e-04, {   0,   0,   0,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  1.844170149570e-04, -1.062277371660e-05, {   2,   0,  -1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -1.162350818550e-04, -7.879570881154e-05, {   1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.877695068929e-04, -7.122576619909e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  4.617151703321e-05,  1.485985051827e-04, {   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.865507969733e-05,  1.660217134356e-04, {   3,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.900710302820e-05, -1.056500018417e-04, {   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.438177842031e-05,  1.502289987341e-04, {   2,   0,  -2,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.757342838198e-04,  1.882081020897e-05, {   2,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.858811294122e-05,  1.656355695453e-04, {   1,   0,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.928243652156e-05, -1.347653922481e-04, {   2,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.708018046016e-04, -2.314996689964e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  7.230789474763e-06, -1.865685174324e-04, {   0,   0,   1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -2.136632636946e-05, -1.721882602278e-04, {   0,   0,   0,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 {  1.933662742806e-04,  1.519944486272e-07, {   2,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.698201183186e-04, -2.367366653666e-05, {   0,   0,   1,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  3.690640729659e-05,  1.564553770387e-04, {   1,   1,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.624467932844e-04,  3.076840758274e-05, {   1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.792419906418e-06, -1.883114977707e-04, {   0,   0,   3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.258645364521e-04,  6.718592477321e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -1.049584162060e-04, -8.801824886584e-05, {   2,   0,  -1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -3.644966324027e-05,  1.564239367390e-04, {   4,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.567421268172e-05,  1.670487451868e-04, {   0,   0,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.494810562834e-04, -4.311522455297e-05, {   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -1.206085141141e-04, -7.194283032772e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -7.459197351935e-05, -1.179148038852e-04, {   4,   0,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.592688764495e-04, -3.319175279617e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  1.764528383838e-04, -1.597854287238e-05, {   2,   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -4.244047208266e-05, -1.499397955612e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.692510253263e-05,  1.553066247367e-04, {   1,  -1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.055263885924e-04,  8.622320135381e-05, {   2,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.985645932132e-05, -1.318285973699e-04, {   1,   1,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.710755271063e-05,  1.144488867379e-04, {   1,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.668949919896e-04,  2.455992138228e-05, {   1,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.837393051509e-06, -1.845598017257e-04, {   0,   0,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.481674890802e-04,  4.290186531869e-05, {   2,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  1.812113759489e-04,  9.819127338604e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -1.869371870371e-04, -4.026619594935e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  9.734370244581e-06, -1.811714042625e-04, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.284736950604e-04,  6.233506048692e-05, {   2,   0,  -1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.737664154268e-04, -1.695210801029e-05, {   1,  -1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.811918835177e-04,  9.498970208802e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  4.872709584081e-07,  1.899993751753e-04, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.901245707231e-05,  1.713060312847e-04, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -1.899999999669e-04,  3.548254369098e-09, {   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.048612419442e-05, -1.190061057145e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  8.893656102823e-05, -1.002466941694e-04, {   2,   0,  -3,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.554493375317e-06, -1.853133575178e-04, {   2,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.879721201803e-04, -6.677467309121e-07, {   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.133034051655e-05,  1.772607810874e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 { -1.049974181779e-04,  8.350302245716e-05, {   2,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  1.219143721908e-04, -6.644264813914e-05, {   2,   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  1.508239973675e-04, -3.740939649909e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -1.854028464317e-04,  2.788798781915e-06, {   2,   2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.576175781488e-05,  1.523729112299e-04, {   1,  -1,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.512754134381e-05,  1.327989677070e-04, {   1,   1,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  1.874321946169e-04,  4.391438522142e-07, {   0,   0,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.872731516494e-04, -5.943450246403e-07, {   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.594035161696e-04, -2.831635719266e-05, {   2,   2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.847062934749e-04,  2.860286786268e-06, {   2,  -2,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.254669637365e-04,  6.198155541962e-05, {   0,   0,   1,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -3.669444641716e-05, -1.507417570329e-04, {   1,   1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.419162853863e-05, -1.329872511800e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -2.179307728224e-05, -1.652583871336e-04, {   0,   0,   1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -1.022065817069e-05,  1.765183302151e-04, {   2,   0,  -1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -1.050065577704e-04, -8.164984826479e-05, {   2,   0,  -1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  3.548025456162e-05,  1.509171064150e-04, {   1,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.738495964991e-05,  1.588393072871e-04, {   3,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.372180060707e-06, -1.797014136955e-04, {   2,   0,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.096222254083e-04,  7.633214536858e-05, {   2,   0,  -1,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  1.536171659238e-04,  3.231061505641e-05, {   2,   0,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.374005354277e-05,  1.721730076147e-04, {   2,   0,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.330556592952e-04,  5.282769868599e-05, {   2,   0,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.619071515805e-04, -2.390504365507e-05, {   0,   0,   0,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.534060639435e-05, -1.704256556649e-04, {   2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.730106486116e-05,  1.583555529292e-04, {   1,   0,  -1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.088930120857e-04,  7.659867876622e-05, {   1,   0,   0,   0,   0,   0, -33,  39,  -1,   0,   0,   0,   0 } },
 {  1.239426006824e-04,  6.137241167443e-05, {   0,   0,   1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  1.839405480428e-04, -1.065349518839e-06, {   2,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.846993684429e-04, -1.827342479432e-07, {   2,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.692555857706e-04, -1.559555628158e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -7.650110524002e-05,  1.082349559894e-04, {   0,   0,   0,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 {  8.090581674488e-05, -1.038281251347e-04, {   2,   0,  -1,   0,   0,   0,  13, -21,   0,   0,   0,   0,   0 } },
 { -1.138283061096e-04,  7.075466682715e-05, {   4,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -2.141639967090e-05, -1.628462526036e-04, {   0,   0,   1,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 { -1.710543033501e-04,  1.305583716785e-05, {   2,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.855769632041e-05, -1.155513118800e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  1.561742957829e-04, -2.791045261141e-05, {   2,  -2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  9.443944676438e-05,  8.964137711753e-05, {   2,   0,  -2,   0,   1,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.234549486152e-04,  6.049938744095e-05, {   2,   0,  -1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.812068907476e-04, -2.659647794351e-06, {   2,  -2,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.778755112677e-04, -5.938455433425e-06, {   4,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  9.329215151916e-05,  9.046485687462e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.084468797410e-06, -1.822453137894e-04, {   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.323150149732e-05, -1.500945196925e-04, {   2,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.710434913882e-04, -1.219679254585e-05, {   4,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.894302666463e-05,  1.542307597343e-04, {   0,   2,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.799730757474e-04,  3.185477680942e-06, {   2,   2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.178821944083e-04, -6.521336407480e-05, {   2,  -2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.723241278667e-04, -1.074290184338e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -1.640395471450e-04, -1.869834142485e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  3.490921493594e-05, -1.475986549188e-04, {   1,  -1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.226487392500e-04,  5.981415316926e-05, {   2,   0,  -1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.822352489800e-05,  1.440667829357e-04, {   2,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.820612973672e-04,  1.454697143040e-07, {   1,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.098255389061e-04,  7.233838331057e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.041073752742e-04,  7.788703372824e-05, {   2,   0,  -1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 { -6.695684654820e-05, -1.149668318248e-04, {   0,   0,   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 {  1.805280917140e-04,  1.131996985644e-06, {   2,   0,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  1.814986409196e-04, -8.813815453177e-08, {   2,   0,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.091981974026e-04, -7.234553063157e-05, {   2,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.813541743947e-04, -1.579055736147e-07, {   4,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.657909944128e-04,  1.524064394497e-05, {   4,   0,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.806719208238e-04,  1.097733580456e-07, {   2,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.298589038729e-05,  1.276734895915e-04, {   3,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.185756789149e-04, -6.203126979993e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -1.708031728083e-04, -9.647759288246e-06, {   0,   0,   1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  4.951909067450e-05, -1.309290793263e-04, {   0,   1,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  4.960269444396e-05, -1.307548932002e-04, {   0,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.069215580885e-04,  7.326980080802e-05, {   0,   0,   1,   0,   0, -12,   6,   0,   3,   0,   0,   0,   0 } },
 {  1.800432639232e-04, -4.240443141650e-08, {   2,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.475771351547e-04, -3.245144431317e-05, {   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.391576302410e-04,  4.051748531423e-05, {   3,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.300125170787e-08,  1.794700481163e-04, {   2,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.233532022410e-06,  1.722193685247e-04, {   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.461504278465e-05, -1.647492431716e-04, {   2,  -2,  -2,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.430975610441e-05,  1.447632320898e-04, {   1,  -1,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.356475858916e-05,  9.528864752897e-05, {   3,  -2,   0,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.428550044636e-05, -1.244631223635e-04, {   0,   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -1.448821656337e-04,  3.341322979748e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.496154941339e-04, -2.852126448229e-05, {   2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.119004342300e-05, -1.469042959756e-04, {   0,   0,   1,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.846451746830e-05, -1.194329064642e-04, {   2,   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -1.990974171991e-05, -1.577477508421e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  1.514385229441e-04, -2.617435806722e-05, {   2,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.174337460475e-04,  6.008650773854e-05, {   0,   0,   1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.025365600459e-04,  7.487409776295e-05, {   0,   0,   1,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 { -1.490328038732e-04, -2.837480482976e-05, {   1,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.171890107923e-04,  6.004053965262e-05, {   2,   0,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  4.853593042012e-05,  1.286087993787e-04, {   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.336768861241e-04,  4.345431437656e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.739322890899e-04,  3.045785411633e-06, {   4,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.501414869707e-04,  2.675434215781e-05, {   2,   0,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.612564398583e-05,  1.306079285012e-04, {   1,   1,  -1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.724365737705e-04, -4.287911561469e-06, {   1,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.045806083611e-05, -9.625395411339e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  1.765280976711e-04,  9.603938990304e-08, {   4,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.763909245571e-04, -1.476767293465e-07, {   2,  -2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.834566232117e-05, -9.815503303134e-05, {   1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  1.503282610386e-04,  2.615711709706e-05, {   6,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.625563263409e-04,  1.381202316868e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -1.472725887041e-04,  2.901866297245e-05, {   1,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.666447520192e-04,  9.636856631789e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -1.762347104298e-04,  4.162918853164e-08, {   3,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.273039355742e-04,  4.893386707682e-05, {   1,  -1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  7.069488364823e-05, -1.054846157201e-04, {   2,   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -3.468219853582e-05, -1.414490688231e-04, {   3,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.626079318165e-04,  1.351216545026e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  1.692832381360e-05,  1.589926647758e-04, {   2,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.380407749590e-05,  1.419758064512e-04, {   4,   0,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.288310707713e-04,  4.674945329643e-05, {   2,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.071076093010e-04,  6.823184504877e-05, {   1,  -1,   1,   0,   0, -21,  22,   0,   0,  -2,   0,   0,   0 } },
 {  7.258683765782e-05,  1.027316075498e-04, {   0,   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -5.050028848134e-05,  1.247634430988e-04, {   2,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.649498915042e-05, -1.486654104830e-04, {   0,   0,   2,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.750382191417e-04,  2.878710526196e-08, {   2,   0,  -2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.666991026657e-04,  8.319835379040e-06, {   2,   0,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -2.839606688436e-05, -1.466087938808e-04, {   4,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.683687541932e-05, -1.477332625597e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.637683317672e-06,  1.699107092120e-04, {   0,   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -3.097376195682e-06,  1.714510008945e-04, {   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.351795186673e-04, -3.933740368180e-05, {   3,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -9.945912628944e-05, -7.500057797933e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  1.463543211335e-04,  2.782729621340e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  7.162077798113e-05,  1.025520935393e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -6.810237530424e-06, -1.671998264837e-04, {   0,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -2.638270661104e-05, -1.474533965474e-04, {   0,   0,   2,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -7.167693533578e-05,  1.020598643426e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  1.731201604328e-04,  4.573784866529e-07, {   3,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.245204920695e-04,  4.890056000015e-05, {   2,   0,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.487537611293e-04,  2.464686639110e-05, {   0,   0,   0,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 {  9.775748761150e-05, -7.559898320897e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -8.127069589999e-05, -9.207390636023e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 { -1.731760155320e-04,  1.615449825848e-08, {   2,  -2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.722863658653e-04,  8.022527969901e-11, {   2,   2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.664591092285e-06, -1.665287236338e-04, {   0,   0,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 {  3.222348719466e-05,  1.398849297753e-04, {   1,   1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.630411994208e-04, -9.064074109507e-06, {   1,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.845765863773e-05,  1.535260003653e-04, {   2,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.313240191098e-05,  1.388232546758e-04, {   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.714833411053e-04,  3.695952676305e-07, {   1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.341995103160e-06, -1.704436097550e-04, {   2,   0,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.478235552332e-05,  1.066558361433e-04, {   0,   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  6.305149432082e-05, -1.082353345745e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  1.698678999238e-04,  1.344033029763e-06, {   0,   0,   3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  7.455304026543e-05,  9.663468825193e-05, {   2,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  7.105979264062e-05, -9.989907727401e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.112970914560e-04, -5.958368500813e-05, {   2,  -2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  7.585836409922e-05,  9.499598057965e-05, {   0,   0,   1,   0,   0, -13,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.049358718181e-05,  1.003153813288e-04, {   1,  -1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.934045493250e-06, -1.648507125770e-04, {   1,  -1,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -9.991942785100e-08,  1.705032699422e-04, {   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.342252574866e-05,  1.571258509989e-04, {   2,   0,  -2,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  3.604053512595e-05,  1.344752946310e-04, {   2,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.060754261970e-04, -6.427275250960e-05, {   2,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.261843025355e-05,  1.376161136792e-04, {   3,  -1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.219104660504e-04,  4.831482442926e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  1.699999999378e-04, -4.598942578470e-09, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.699999999595e-04,  3.708824660194e-09, {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.699999999834e-04,  2.373647782635e-09, {   4,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.699999999869e-04,  2.106612407103e-09, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.699999999910e-04, -1.750565239685e-09, {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.700000000000e-04, -1.186823891733e-10, {   2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.175175014708e-05, -1.579297317923e-04, {   2,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.401709997604e-04,  2.945244020646e-05, {   3,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -7.617282621007e-05, -9.338648732520e-05, {   1,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.693301986756e-04, -5.576265570093e-08, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.017412829308e-04, -6.758052751918e-05, {   4,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.399165044147e-04,  2.940197326372e-05, {   1,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -4.413261035275e-05,  1.249949430034e-04, {   0,   0,   1,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.194521147418e-05,  1.371529616208e-04, {   1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.815626020774e-05,  7.083364302495e-05, {   0,   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  7.663341527391e-05,  9.232982517104e-05, {   1,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.857265904078e-05,  1.500962779553e-04, {   4,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -5.793593531950e-06,  1.628746828287e-04, {   0,   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -7.669892317304e-06, -1.609057330376e-04, {   2,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.667310938446e-04,  1.639933167402e-06, {   2,   0,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.491505155768e-05,  1.132225118286e-04, {   1,   1,  -1,   0,   0, -28,  33,   0,   0,   0,   0,   0,   0 } },
 { -1.301638724143e-04, -3.783856739427e-05, {   1,   1,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.005375130251e-04, -6.733825381430e-05, {   0,   0,   1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -1.412910857861e-04,  2.646849955740e-05, {   3,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.369772675855e-04,  3.054061123802e-05, {   2,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.389899958730e-05,  7.328182604941e-05, {   2,   0,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.258994562159e-04, -4.118655160956e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 {  6.492942078874e-05, -1.021119587855e-04, {   2,   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -1.257947671704e-04, -4.114831724176e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 { -1.665970922964e-04, -3.266334828949e-07, {   0,   0,   1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.258328857742e-04, -4.103991507208e-05, {   0,   0,   0,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 {  1.438264438721e-04,  2.285132538337e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -5.888739165133e-05, -1.077859205456e-04, {   2,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.423138511661e-06, -1.600554523594e-04, {   2,   0,  -1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.322610571516e-04,  3.419249082417e-05, {   3,   0,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -9.706339152793e-05,  6.937105065850e-05, {   1,   1,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.164624177138e-04,  4.995052787904e-05, {   1,  -1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.518640441400e-04,  1.448022962324e-05, {   4,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.203316958858e-05,  1.341392994580e-04, {   1,  -1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.306643658961e-05,  1.529538965795e-04, {   2,   0,   0,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -1.651181490014e-04, -8.359383883173e-07, {   0,   0,   3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.613866544187e-05,  1.297903418241e-04, {   0,   0,   2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.610636116854e-05,  1.296856799141e-04, {   0,   0,   2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.166228021218e-04,  4.902998625534e-05, {   0,   1,   0,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.945285544105e-05, -1.261919361460e-04, {   0,   0,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.546307975203e-04, -1.093763910399e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -3.157471431191e-05, -1.339881406693e-04, {   1,  -1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.406068094692e-05,  1.014703363383e-04, {   2,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.439658038576e-06,  1.620872164154e-04, {   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 {  1.084872221228e-04,  5.699552917006e-05, {   2,   0,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 {  6.826204309221e-05,  9.717666202457e-05, {   1,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.602408792357e-04, -5.110072048538e-06, {   0,   0,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.057539071220e-04,  5.956404862732e-05, {   2,   0,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  1.357247938459e-04,  2.956680816099e-05, {   2,  -2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.230255981723e-05, -1.325867701946e-04, {   1,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.714185309330e-05,  1.175789251598e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -2.376714000939e-05,  1.409421720331e-04, {   1,  -1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.289183281382e-04,  3.567377779812e-05, {   0,   2,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.537242912268e-04,  1.079104262243e-05, {   2,   0,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.287360914536e-04,  3.573462515204e-05, {   0,   2,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.018056233761e-04,  6.255037177525e-05, {   0,   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.640642218252e-04, -2.325113743882e-07, {   1,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.416218714402e-04,  2.262923127647e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  4.824714974899e-05, -1.158982873203e-04, {   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.800367609379e-05,  1.059656075346e-04, {   0,   0,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -7.470465012797e-05, -8.905771222806e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.494122988172e-04,  1.430623318836e-05, {   2,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.172588469557e-05, -1.217986279012e-04, {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.499187226773e-06, -1.549197509089e-04, {   0,   0,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.624952510931e-04, -2.661676619528e-07, {   1,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.843176820276e-05, -1.442006795400e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  1.343362205993e-04,  2.823620353574e-05, {   3,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.089340183007e-04,  5.346183905281e-05, {   0,   0,   1,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.007947652352e-05,  1.222580994419e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -3.594171388605e-05, -1.263654975127e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.620201565891e-04,  1.737742381579e-07, {   0,   0,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -3.137400999602e-05, -1.307654659091e-04, {   1,   1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.080856581087e-04, -5.396508143005e-05, {   2,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.289107758089e-04, -3.302423169695e-05, {   1,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.338089016856e-04,  2.812562677879e-05, {   1,   0,  -1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.043595387682e-04, -5.753026130207e-05, {   3,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.211382447291e-04,  4.069792298704e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  7.171982240704e-05, -8.995834274240e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.336028899300e-04, -2.805640212881e-05, {   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.453288210098e-05, -1.370065883528e-04, {   0,   0,   0,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 {  1.040311285610e-04,  5.746094406853e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -4.736404004207e-05, -1.141101244710e-04, {   1,  -1,  -2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -4.735868205690e-05, -1.140929320162e-04, {   1,  -1,   2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.503883049705e-04,  1.087408352224e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 {  1.308430067822e-04, -3.030820486664e-05, {   0,   0,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.084752301596e-05, -1.302563199664e-04, {   1,   1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.608591113124e-04,  1.207030450996e-09, {   2,  -2,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.711313704253e-05, -1.237239029638e-04, {   1,   1,  -1,   0,   0, -25,  28,   0,   0,   0,   0,   0,   0 } },
 { -6.341228022850e-06,  1.542893642400e-04, {   2,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -1.580433711299e-04,  2.552850977915e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.317095317607e-05,  1.373477038417e-04, {   1,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.944393808806e-05,  1.210273678351e-04, {   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.173553521088e-04, -4.276581543988e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 { -2.126220625590e-06, -1.579308759867e-04, {   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  1.599999999368e-04, -4.495968152544e-09, {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.599999999588e-04,  3.630284843837e-09, {   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.599999999881e-04, -1.954768762171e-09, {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.075637847013e-05,  1.292286653074e-04, {   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.286576894405e-06,  1.586844734951e-04, {   1,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.692539685637e-05, -1.130188972940e-04, {   3,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -3.287568959930e-05, -1.268797794329e-04, {   2,   0,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.020665437321e-04,  5.741527236403e-05, {   2,  -2,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.527714687350e-04,  6.589645717671e-06, {   0,   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  3.052636679495e-05, -1.287735635049e-04, {   1,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.077459108821e-05,  1.184652411647e-04, {   2,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.388235081745e-04, -2.041163952489e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  4.670575427178e-05, -1.125179579612e-04, {   1,   1,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  3.025682238480e-05,  1.289046748756e-04, {   1,   1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.576882392634e-05, -1.333501005250e-04, {   2,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.220794532481e-04, -3.696767166839e-05, {   4,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.346541614746e-05, -1.154983271148e-04, {   0,   0,   1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -6.278773465986e-05,  9.598397178201e-05, {   1,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.809336473933e-05, -1.000146874407e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  6.265015537651e-05, -9.535985561275e-05, {   2,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -1.780103930014e-05, -1.401102447455e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   4,  -1,   0,   0 } },
 {  4.817693605190e-06, -1.529916072437e-04, {   0,   0,   2,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.018447182533e-04, -5.595299946581e-05, {   1,   2,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  1.018447182533e-04, -5.595299946581e-05, {   1,  -2,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -4.701504208697e-05,  1.104064573182e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.073612384230e-04,  4.972971867763e-05, {   0,   0,   0,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 { -1.377450774819e-04,  1.891263422904e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -1.038473036017e-04,  5.279575392283e-05, {   0,   0,   1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.455560955211e-05, -1.218662490672e-04, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.048774020404e-04,  5.151634766602e-05, {   2,   0,  -1,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.562243017286e-04,  4.416004885861e-08, {   6,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.982229268339e-05,  1.263513712632e-04, {   1,   1,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.422711579176e-04, -1.390015478572e-05, {   2,   0,  -1,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.077866323211e-04,  4.833774646610e-05, {   2,   0,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.882227282298e-05, -1.369340342772e-04, {   1,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -8.722687893067e-05, -6.839852045450e-05, {   0,   0,   1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 {  1.551929142943e-04, -3.985347595908e-07, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.005932744190e-04,  5.475501349541e-05, {   0,   0,   1,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.284819351928e-06,  1.519553646044e-04, {   2,   0,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -5.964112429735e-05,  9.553053356557e-05, {   2,   0,  -1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 {  6.674566710373e-05, -8.802236935747e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  5.689246254797e-05, -9.783714242793e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  5.223819370316e-05,  1.023774582799e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  1.406240722366e-04,  1.376355104483e-05, {   0,   0,   1,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -1.406216079072e-04,  1.373426824205e-05, {   0,   0,   1,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 {  1.222413640869e-04, -3.202538138678e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.926502731047e-05,  1.249392424824e-04, {   1,  -1,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.530185444158e-05,  1.089022673866e-04, {   1,  -1,  -1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -9.873806092851e-05,  5.511140746926e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -4.529649913394e-05, -1.084651874629e-04, {   1,   1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -1.533743318862e-04, -3.676836255413e-07, {   4,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.305178076496e-04, -2.316172651496e-05, {   3,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.620834967332e-05, -1.074513207944e-04, {   1,   0,  -2,   0,   0,  29, -37,   8,  -3,   0,   0,   0,   0 } },
 {  1.394383843596e-04,  1.410280252137e-05, {   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  5.942350763080e-06, -1.475153225823e-04, {   2,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.147887004094e-04, -3.859500038484e-05, {   2,   0,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.236178751386e-05, -6.101444734095e-05, {   1,  -1,  -1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  8.101994467177e-05, -7.228251274593e-05, {   2,   0,   0,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.187703983189e-05, -1.413439060235e-04, {   0,   0,   2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  8.767396679861e-05,  6.551588508682e-05, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.209326991961e-04, -3.219846516098e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.926122412452e-05,  1.238414451758e-04, {   1,  -1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.035590633651e-05, -1.225248592695e-04, {   1,  -1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.220398997464e-05, -1.206010435103e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -1.385804240087e-04, -1.419185433624e-05, {   1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.512677260030e-04,  1.347072778461e-06, {   4,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.352073195950e-05,  9.907974972137e-05, {   0,   0,   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.923330394743e-05, -1.232768075597e-04, {   1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.431716282958e-04, -9.334872384273e-06, {   0,   0,   1,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 { -1.431716282958e-04, -9.334872384273e-06, {   0,   0,   1,   0,   0,   0, -14,  27,  -3,  -3,   0,   0,   0 } },
 { -1.519916195124e-04, -4.937307557713e-07, {   4,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.093775712660e-05, -1.215041371080e-04, {   1,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.129831910894e-05, -1.110994247325e-04, {   0,   0,   1,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 {  5.108580310038e-06, -1.472677247936e-04, {   2,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.995361139177e-05,  1.223969968749e-04, {   1,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.252351887445e-05,  1.394387373544e-04, {   3,   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.694195797735e-05,  8.496876268338e-05, {   2,   0,   0,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  3.115649148677e-05,  1.207480107941e-04, {   2,   0,  -1,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.513358332895e-04,  5.623370998607e-07, {   1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.345170389404e-04, -1.736907047256e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  9.670020902066e-05, -5.517174132559e-05, {   2,  -2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.301912932323e-04, -2.151992835334e-05, {   1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.901342111813e-05, -1.223093686581e-04, {   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.698505815055e-05,  7.421573019615e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 {  6.772154808546e-05, -8.336183221248e-05, {   0,   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.555811621752e-05, -1.053919246697e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   4,   0,   0,   0 } },
 {  4.768221690934e-05,  1.032550385850e-04, {   2,   0,  -2,   0,   0,   1,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.650961868672e-05,  8.442472627449e-05, {   2,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.507533500687e-04,  1.131053268628e-07, {   1,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.497512486760e-04,  1.081785395107e-06, {   2,   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -3.652353092355e-05,  1.143067438469e-04, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.038016646335e-04,  4.685711403579e-05, {   2,   0,  -1,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -3.497590726947e-05,  1.156749695370e-04, {   2,   0,  -2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.973723632179e-06,  1.486097250362e-04, {   2,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.107085885112e-04,  3.982616904168e-05, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.276396947695e-04, -2.282221929670e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.715360372984e-05,  9.330262556136e-05, {   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  4.910254680390e-05, -1.012254339226e-04, {   0,   0,   1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 {  2.351930483009e-05,  1.267918469599e-04, {   1,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -7.123321374841e-05,  7.903365966383e-05, {   0,   0,   1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  3.779302384463e-05,  1.122564700426e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  1.499999999815e-04,  2.356194490095e-09, {   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.499999999885e-04,  1.858775653326e-09, {   4,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.503352840970e-05, -1.149272749441e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -1.725672514853e-05, -1.326723334631e-04, {   4,   0,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.268678538130e-04,  2.295648720574e-05, {   2,   0,   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  1.496686680026e-04,  8.408676048459e-08, {   1,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.224087564870e-04,  2.720189411124e-05, {   2,  -2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.247234473661e-05,  1.371161519262e-04, {   0,   0,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -9.001141165338e-05, -5.921110953932e-05, {   0,   0,   1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -1.491879139562e-04,  2.014772469481e-08, {   2,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.811274487041e-05,  6.105833388916e-05, {   0,   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -5.505650837910e-06, -1.435501719771e-04, {   0,   0,   1,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.146730606568e-04, -3.385734062524e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.081067316315e-05,  1.277172635411e-04, {   1,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.008632432876e-06, -1.465045679014e-04, {   0,   0,   2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  4.576811203777e-05, -1.025610101261e-04, {   2,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.291609144388e-05,  1.053195685134e-04, {   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.253484971945e-05, -1.155697444816e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  7.203931728760e-06,  1.405830893897e-04, {   2,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.328837471842e-04, -1.475497242885e-05, {   0,   0,   0,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.389108927569e-04,  8.676915634414e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  3.051158261236e-06, -1.443508951717e-04, {   0,   0,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.397942886051e-04, -7.586624638226e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.861296729152e-05,  7.871950924234e-05, {   2,   0,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 {  6.531228845799e-05, -8.179988149482e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  1.231723202563e-04, -2.367989385433e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -6.170263222669e-05, -8.513558635109e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.691166447381e-05, -8.984734527990e-05, {   2,   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -1.229232903233e-04, -2.382935473585e-05, {   1,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.896889772572e-05, -4.752009266179e-05, {   2,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.465318935504e-05, -1.216504830520e-04, {   1,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.051209885402e-06, -1.412253774614e-04, {   0,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.888913299271e-05, -8.735160808781e-05, {   2,   0,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -9.578343992913e-05,  5.030162533874e-05, {   0,   0,   2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -2.790434572170e-05,  1.181681804224e-04, {   1,   1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.536768269518e-05, -1.306155587265e-04, {   2,   0,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.232077567047e-05, -1.335150310516e-04, {   2,  -2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.436536587837e-04, -2.174336954666e-06, {   3,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.477912558955e-05,  1.109059876392e-04, {   2,   0,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.097953148090e-04, -3.580177459622e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  3.124576981419e-05,  1.141765867941e-04, {   1,  -1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.449487485871e-04,  2.964264870657e-07, {   4,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.319487674848e-05, -5.175478713669e-05, {   4,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.707823172514e-05,  6.775210375969e-05, {   2,   0,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -4.326729605903e-06, -1.404265631688e-04, {   2,   0,  -1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  7.222635348859e-05, -7.228729425918e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -7.402109614137e-05,  7.034141718771e-05, {   1,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.380632323087e-04,  6.147628189414e-06, {   2,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.411080578136e-04,  3.011819789755e-06, {   1,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.742841907213e-05, -4.668913069082e-05, {   2,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.922314483931e-05, -1.048835975199e-04, {   4,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.425769057110e-04,  1.477008626023e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 { -4.697196863313e-05, -9.691992356481e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.108252821768e-04,  3.301863923219e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  8.634605344632e-05, -5.729729892720e-05, {   2,   0,   0,   0,  -9,   1,  11,   0,   0,   0,   0,   0,   0 } },
 {  9.277578072797e-05, -5.070284999021e-05, {   2,   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -1.066870729313e-04, -3.674464032864e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  6.445248248593e-05,  7.878602713421e-05, {   1,  -1,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.744861607988e-05,  1.057650706039e-04, {   2,   0,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.045967122603e-04, -3.834895549748e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -5.523988177643e-05, -8.722424685267e-05, {   1,  -1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.676925260562e-05, -1.156894803159e-04, {   3,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.925150599239e-05,  1.131442371021e-04, {   1,   0,   0,   0,   0, -31,  38,   0,   0,   0,   0,   0,   0 } },
 {  1.199533645986e-04,  2.233613092908e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.804936324543e-05,  6.421753878529e-05, {   2,   0,   0,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  1.452920811933e-05,  1.277144506512e-04, {   2,   0,  -1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.401687383864e-04,  2.008054685715e-06, {   4,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.335019064720e-04,  8.642128253478e-06, {   1,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.110295607568e-06, -1.369642996928e-04, {   0,   0,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.196984759504e-04,  2.227439740584e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.784320913822e-05,  8.405597343775e-05, {   0,   0,   2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -5.883342385370e-05, -8.300017837481e-05, {   0,   0,   1,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -1.413060564455e-04, -8.543753449194e-08, {   2,  -2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.369963335770e-05, -6.769083117410e-05, {   2,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.593549443500e-05, -1.154383923213e-04, {   1,   1,  -2,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  1.175214995608e-05,  1.295944173005e-04, {   2,   0,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.411119102532e-04, -1.566490684105e-07, {   0,   2,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.285390060427e-04, -1.266576232343e-05, {   2,   0,  -1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.151904161103e-04, -2.599494967437e-05, {   1,  -1,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.541416130599e-06, -1.396414928054e-04, {   4,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.529906764857e-05, -4.586179112929e-05, {   2,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.162920779515e-05, -1.294788072468e-04, {   1,   3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.401085375220e-04, -9.114364594807e-07, {   2,   0,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  5.845471045419e-05, -8.253898480790e-05, {   0,   0,   1,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 { -1.409181601915e-04,  6.212937829396e-08, {   0,   0,   2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.856079018475e-05,  5.241156141725e-05, {   0,   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  9.669812163027e-05,  4.417631843453e-05, {   2,   0,  -1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  8.630886552312e-05,  5.454035209290e-05, {   0,   0,   1,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  2.792484276685e-05,  1.128144848476e-04, {   1,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.621776721711e-05,  5.446969710711e-05, {   0,   0,   1,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 {  1.207719461616e-04, -1.979443839562e-05, {   1,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.045356248657e-05,  7.999515256354e-05, {   2,   0,   0,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  1.090231390567e-06, -1.393474931288e-04, {   0,   0,   2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -6.432018510489e-05, -7.611444291262e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 { -9.319405507344e-05,  4.720199536601e-05, {   0,   0,   1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  8.805971869939e-05,  5.232540699061e-05, {   2,   0,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  3.666173095910e-05,  1.036707514554e-04, {   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.038599824314e-05,  7.990439341815e-05, {   2,   0,   0,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -1.385842868550e-04, -1.677268368003e-06, {   2,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.399999999728e-04, -2.761110876450e-09, {   2,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.399999999742e-04,  2.687807047906e-09, {   4,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.399999999971e-04,  9.040805525268e-10, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  5.777059294517e-05,  8.220098023840e-05, {   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -7.663585429678e-07,  1.391751453083e-04, {   1,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.087004649097e-05,  1.090408968222e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 { -7.665514443319e-05,  6.303661736888e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 {  1.091557063285e-04,  3.044981847252e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 { -9.294467991816e-05,  4.659784932937e-05, {   0,   0,   1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.393160994034e-04,  5.987650283132e-08, {   1,   0,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.054037631418e-05, -4.883323979069e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 { -7.618858375728e-05, -6.311607207779e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  1.004913520522e-04,  3.878485155317e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -5.995488395789e-05,  7.907808340198e-05, {   2,   0,  -1,   0,   0,   0,   7, -10,   1,   0,   0,   0,   0 } },
 { -5.459126645601e-06, -1.332558439122e-04, {   0,   2,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.552585388479e-05, -6.316022920211e-05, {   2,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -2.734112758032e-05, -1.112368873182e-04, {   3,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.494671264660e-05, -9.359704816900e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -4.494089236984e-05, -9.358575586143e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -1.138375992132e-04,  2.463922947616e-05, {   4,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.760720646846e-05, -1.007113002992e-04, {   0,   0,   2,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 {  1.379608205565e-04, -1.113869306351e-07, {   2,   0,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.423116740896e-06, -1.316425145770e-04, {   4,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.718914396367e-05, -7.086004938379e-05, {   2,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.035468466345e-04,  3.441140805637e-05, {   2,   0,  -1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.619616264546e-05,  1.117136500004e-04, {   1,   1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.373996130714e-04,  5.058686970946e-07, {   2,   0,  -1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  8.796248743786e-05,  4.978600146801e-05, {   0,   0,   1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  8.970824945605e-05, -4.795446490211e-05, {   2,   0,   0,   0,  -7,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.895947515650e-05, -9.846647477796e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -1.343032341320e-04,  3.090624657190e-06, {   3,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  8.400481938693e-05,  5.336160769864e-05, {   4,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.941367959090e-05,  8.786858927505e-05, {   1,   0,   0,   0,   0, -20,  18,   4,   0,   0,   0,   0,   0 } },
 { -1.028657114534e-05,  1.269648892646e-04, {   1,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -8.890137788585e-05,  4.824885156885e-05, {   0,   0,   1,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.369987455271e-04, -1.010141268667e-07, {   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.136203438803e-05,  8.558657917982e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -4.506736484657e-05, -9.188058584920e-05, {   0,   0,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.355774749459e-04,  1.184919636934e-06, {   2,  -2,  -1,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -2.278514166762e-06,  1.344108572715e-04, {   0,   2,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.115246491501e-05, -8.546429694533e-05, {   2,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -8.348336523491e-05,  5.302006767136e-05, {   4,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  5.434971124742e-05, -8.211376156220e-05, {   4,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.398387165065e-05, -9.245838317837e-05, {   2,   0,  -2,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 {  2.533666806569e-05, -1.109857054667e-04, {   1,  -1,   0,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.356505832780e-04,  6.378549953291e-07, {   2,   0,  -1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -1.356505832780e-04,  6.378549953291e-07, {   2,   0,  -1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -8.075625079262e-05, -5.540801503538e-05, {   2,   0,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -9.174730053901e-05, -4.431587894677e-05, {   1,  -1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.520089477752e-05, -1.207715049304e-04, {   2,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -1.151601684737e-04, -2.078002314392e-05, {   4,  -2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.155118653277e-04,  2.042624609998e-05, {   4,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.252095638637e-04,  1.071636294490e-05, {   2,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.331603095018e-05,  9.260425470012e-05, {   0,   0,   1,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 { -9.892386753328e-05,  3.694033757247e-05, {   1,  -1,   0,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -4.327110729493e-05,  9.254343540463e-05, {   0,   0,   1,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 { -5.220599210135e-05, -8.339137056571e-05, {   4,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.343229136587e-04,  1.174521142485e-06, {   1,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.903427470200e-06, -1.255768920305e-04, {   2,   0,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.360190526635e-05, -9.165357447361e-05, {   2,   0,   0,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -2.530814909459e-06,  1.325753871216e-04, {   2,   0,  -3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.267241542998e-04, -8.068214617055e-06, {   2,   0,  -1,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 { -1.345932817915e-04,  1.346910659780e-07, {   0,   2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  6.017826606077e-05, -7.417246949487e-05, {   1,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -5.537808557738e-05,  7.888677839970e-05, {   0,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.766612172871e-05,  3.648337026956e-05, {   1,  -1,  -2,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  6.199970533708e-06, -1.277477968201e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  1.065547200539e-04, -2.731467799985e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.691932162288e-05,  1.068345705299e-04, {   1,   1,   0,   0,   0, -46,  49,   0,   0,   0,   0,   0,   0 } },
 {  2.107618416937e-05,  1.124623188658e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  4.320232198109e-05,  9.021451482537e-05, {   2,   0,  -1,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 { -1.130958447824e-04,  2.010147442270e-05, {   3,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.065559728310e-04,  2.659941758683e-05, {   0,   0,   1,   0,   0, -24,  26,   0,  -3,   0,   0,   0,   0 } },
 {  3.871376324166e-06,  1.290460227547e-04, {   2,   2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.838361057734e-05,  4.451218310499e-05, {   0,   0,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.051921242456e-05, -7.226078666277e-05, {   1,  -2,   1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -6.051921242456e-05, -7.226078666277e-05, {   1,  -2,  -1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  1.302879308356e-04,  2.416589127231e-06, {   2,   0,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  8.212979039610e-07,  1.318174898197e-04, {   2,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.063074364790e-04, -2.629645934028e-05, {   2,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.601988025275e-05,  7.652412950427e-05, {   2,   2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.145875890523e-04, -1.788131979461e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  2.392602716266e-06,  1.298761512364e-04, {   0,   2,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.143548252060e-05, -4.081911549988e-05, {   2,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.286447714650e-06, -1.239166673791e-04, {   2,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 {  1.251574312150e-04, -6.974148936332e-06, {   0,   2,  -2,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  3.411629408139e-05, -9.798125233771e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -8.820326920970e-05,  4.352188432478e-05, {   2,   0,  -3,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.075625651127e-04, -2.411352233735e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 {  3.041569700565e-06, -1.284442704989e-04, {   2,   0,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.392309742450e-05, -3.747380363694e-05, {   1,   1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -1.643293777255e-05,  1.149363035484e-04, {   2,   0,   0,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.049363061974e-04, -2.619168032405e-05, {   2,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.475867812167e-05,  1.161216467452e-04, {   2,   0,   0,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  8.851943656069e-05,  4.233618529107e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  5.004635928442e-05,  8.074686795585e-05, {   0,   2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.164278966206e-04, -1.431407338969e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 { -9.554847894636e-05, -3.511513674998e-05, {   2,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.763140698708e-05,  9.303150603159e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.207154559687e-04,  9.892546689487e-06, {   2,  -2,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.819619489203e-05, -4.240580022412e-05, {   2,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.594882325448e-05, -5.450132215458e-05, {   0,   0,   2,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  7.588772330003e-05, -5.446059006017e-05, {   0,   0,   2,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  9.663973294451e-05,  3.363776482682e-05, {   3,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.073874189977e-05,  1.195185317112e-04, {   3,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.301466457830e-04,  4.051646576810e-08, {   3,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.972084339611e-05,  8.035273569266e-05, {   0,   2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.299999986686e-04, -1.860520976272e-08, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.299999999900e-04, -1.610938899545e-09, {   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.488756956719e-05,  1.050472183676e-04, {   1,  -1,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.129121228398e-04,  1.699668635495e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 {  7.829315743597e-05, -5.158365726744e-05, {   4,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.438556619148e-05, -1.054264883898e-04, {   0,   1,   1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.138433324400e-06, -1.285333568545e-04, {   4,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.828811590889e-05, -1.013132850442e-04, {   2,   0,  -1,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.009157596397e-05,  1.194329598698e-04, {   2,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.272515470790e-04,  2.190955420343e-06, {   4,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -8.307187281295e-06,  1.211179603993e-04, {   2,   0,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -9.630376306906e-05,  3.294955627143e-05, {   1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.733774846874e-05, -1.018762935764e-04, {   1,  -1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.240700588996e-04,  5.071255011564e-06, {   0,   0,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -4.410666543269e-05,  8.501076231873e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 { -8.580371397626e-05,  4.330675920599e-05, {   0,   0,   1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.647232320486e-05,  4.259236794156e-05, {   0,   0,   1,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.067454516212e-04, -2.225806689577e-05, {   4,   0,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -6.260468936763e-05, -6.638890438375e-05, {   0,   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.707939133981e-06,  1.250574473095e-04, {   2,   0,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.213427500552e-04,  7.409683118951e-06, {   2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.052575048987e-04, -2.345403793706e-05, {   1,   1,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  2.482841934153e-05, -1.038096384187e-04, {   1,  -1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.752413382177e-06,  1.228668019381e-04, {   2,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.207684494657e-05,  3.648764880712e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.144484777143e-04,  1.411287746425e-05, {   0,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 { -9.037574078278e-05,  3.817834090665e-05, {   2,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.269417206532e-04,  1.593193391078e-06, {   2,  -2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.450288441408e-05,  1.038968108261e-04, {   1,   1,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.494323083879e-05,  6.330352799711e-05, {   2,   0,  -1,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  5.535713484052e-05,  7.281214541381e-05, {   2,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -1.269820804519e-04,  1.151387920794e-06, {   2,   0,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -2.434977428187e-05, -1.035697722784e-04, {   0,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.262095631978e-04,  1.659363472324e-06, {   1,   2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  8.025137832253e-05, -4.761641080631e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 { -2.422756622934e-05,  1.036400080677e-04, {   1,  -1,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  9.429734161382e-05, -3.349437489203e-05, {   4,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.617017958140e-05, -3.159891107816e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0 } },
 { -1.006080781183e-04, -2.686274559355e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -1.219028017783e-04,  5.550174225697e-06, {   2,   0,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.245766266958e-05, -1.149107243240e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0 } },
 { -8.509392513632e-06,  1.186342310033e-04, {   2,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.268654636304e-04, -2.655003050356e-07, {   6,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.226430647985e-05, -9.479491996583e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -1.262518781661e-04,  6.515745066529e-07, {   0,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 {  9.876543876878e-05,  2.808654990680e-05, {   2,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.265968682161e-04,  2.205423345459e-07, {   0,   0,   3,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.364632734921e-05, -1.028817049233e-04, {   0,   0,   2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.327417897366e-06,  1.220832083376e-04, {   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.437246155155e-05, -1.020271742968e-04, {   4,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.467823086430e-07,  1.255393577719e-04, {   1,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.221369188906e-04, -4.090590889784e-06, {   4,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.198280613729e-04,  6.187953062998e-06, {   0,   0,   3,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -2.934205422779e-05, -9.666094806403e-05, {   2,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.258787651721e-04,  1.005243802753e-07, {   2,   0,  -2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.255725458395e-04,  3.354214376402e-07, {   0,   0,   1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.973937213156e-05,  1.061551496390e-04, {   1,  -1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  5.011085568681e-05, -7.571316527249e-05, {   2,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.247597281387e-04, -9.345764665056e-07, {   4,  -2,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.588502993417e-05,  7.976978241029e-05, {   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.231464728731e-04,  2.455922696742e-06, {   2,   2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.866024431666e-06, -1.217340316056e-04, {   4,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.399991424448e-05, -1.014253822515e-04, {   1,  -1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.045700673365e-04,  2.072798576456e-05, {   1,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.248168665836e-04,  4.138960768498e-07, {   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.192627191877e-04,  5.965454353797e-06, {   0,   0,   1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  1.042543521909e-04, -2.096541962246e-05, {   0,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.249661930555e-04, -2.529133118370e-07, {   6,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -8.335340829357e-05,  4.185663624605e-05, {   0,   0,   1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -5.115035856894e-06, -1.200471053386e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  5.265533660477e-05,  7.248030396048e-05, {   4,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.247242636367e-04,  1.423171413256e-07, {   1,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.157151778231e-04, -9.142631959492e-06, {   2,  -2,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -2.718124418833e-05, -9.762441929283e-05, {   4,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -7.640884412804e-05, -4.830459395258e-05, {   1,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -8.920718677068e-05,  3.537918812705e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -8.034287899017e-05,  4.422609734082e-05, {   2,   0,   2,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  3.523573699112e-05, -8.922551756588e-05, {   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.231068672332e-04,  1.157680241643e-06, {   4,   0,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.216062237799e-04, -2.648223356756e-06, {   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.013830209878e-05,  4.411348452943e-05, {   2,   0,   0,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  2.704885589272e-05, -9.714727472451e-05, {   4,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -2.961737979918e-06,  1.211826790574e-04, {   2,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.548584082978e-06,  1.175270185202e-04, {   2,   0,  -1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  3.119055036554e-05, -9.285616627054e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  7.466290953093e-05,  4.935672875752e-05, {   2,   0,  -1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -1.204936735463e-04,  3.439949431196e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -1.142538505916e-04, -9.670386182856e-06, {   1,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.040058483653e-04, -1.972572683728e-05, {   0,   0,   3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.362218120988e-05,  9.999376192154e-05, {   1,  -1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.292622723211e-05,  5.031655624955e-05, {   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -5.458354126775e-05, -6.864870576557e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  7.549805026346e-05, -4.772179210818e-05, {   1,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.222857434956e-04,  7.439862332730e-07, {   0,   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.229362451885e-04,  7.296513089870e-08, {   3,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.219934633970e-04, -9.958225895591e-07, {   3,  -2,  -2,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -4.249439727223e-05,  8.040809268506e-05, {   0,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  1.169361424730e-05,  1.111229297276e-04, {   2,   0,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.382862012695e-05, -1.089488744660e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -4.845936941089e-06,  1.179175514036e-04, {   2,   0,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.344011759294e-05,  1.091758363334e-04, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.434024230918e-05,  9.825364295684e-05, {   1,  -1,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.223601914994e-04, -1.193206487776e-07, {   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.460596564133e-05, -7.777108072060e-05, {   4,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -8.135074919647e-05,  4.099687679449e-05, {   0,   0,   1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.220410966592e-04, -2.734601614460e-07, {   0,   2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.315192061320e-05,  9.894608605951e-05, {   1,   1,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -5.117918383639e-05,  7.087136704993e-05, {   0,   2,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  9.115121133309e-05, -3.089594018469e-05, {   2,   0,  -2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.116310991801e-05,  7.083017920072e-05, {   0,   2,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.716860018738e-05, -7.477319068882e-05, {   2,   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  8.183380188344e-05,  3.992177856047e-05, {   2,   0,  -1,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.547171064781e-05,  6.626593032048e-05, {   1,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.022258024427e-04,  1.947602496908e-05, {   1,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.672190110708e-05, -9.496630595882e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.016138804122e-04, -2.006224594778e-05, {   1,  -1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.317217171933e-05, -9.844146305435e-05, {   1,  -1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.211537290005e-04, -4.525760341817e-07, {   2,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.201843947002e-04,  1.414397742080e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 {  1.007699792647e-04, -2.076686392005e-05, {   0,   0,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.349007878489e-06,  1.161317572674e-04, {   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.555780269980e-05, -9.574233545048e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 {  7.855263014402e-06, -1.134226817266e-04, {   1,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -5.995626362078e-05, -6.118289691464e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -2.365988375260e-05, -9.729216969247e-05, {   1,  -1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.207805515484e-04,  6.583197451538e-08, {   1,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.000232382670e-05, -8.081181621494e-05, {   2,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  2.479639301236e-05, -9.594493225609e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 {  4.637539748927e-06, -1.159991274467e-04, {   2,   0,  -1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -8.021311838253e-05,  4.034563917049e-05, {   0,   0,   1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.261742740198e-05,  9.783422874614e-05, {   1,   1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.070714919407e-05, -9.967043630535e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.597531334609e-05,  6.427143095898e-05, {   2,   0,  -1,   0,   0, -10,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.246058805208e-05, -7.776971353187e-05, {   0,   0,   2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  4.664924919517e-05,  7.337957675073e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  1.199999999732e-04, -2.534218073678e-09, {   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.199999999779e-04,  2.303834612491e-09, {   2,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.199999999892e-04, -1.612684228770e-09, {   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.319948744036e-05, -8.674110632919e-05, {   1,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.654820114822e-05,  1.031350763775e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  3.737943533998e-05, -8.223839179662e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 { -9.443488365708e-05,  2.500581335297e-05, {   2,   0,  -2,   0,  -4,  15,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.037590904570e-05,  9.898674495019e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -8,   0,   0,   0 } },
 {  2.243532236934e-05,  9.689033703712e-05, {   1,   1,  -1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.193022916343e-04,  1.791908110772e-08, {   2,  -2,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.546020442861e-05, -1.038148112649e-04, {   4,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.568153783503e-05,  1.035902341754e-04, {   4,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.101896449231e-04,  8.987533713430e-06, {   0,   2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.619711054213e-05,  2.297860522191e-05, {   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 {  1.061404896541e-04, -1.298099430048e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  4.823150891498e-05, -7.081263913467e-05, {   1,   0,   1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  4.823150891498e-05, -7.081263913467e-05, {   1,   0,  -1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 { -9.520114145703e-05,  2.381492561907e-05, {   1,  -1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.315599875286e-05, -9.582367275823e-05, {   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.394765225022e-05,  1.050092233358e-04, {   2,   0,   0,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -5.081085777890e-05, -6.812593795462e-05, {   2,   0,  -1,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -1.264641253843e-06,  1.176125857476e-04, {   3,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.289199873101e-05, -3.565893401497e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  7.478607470484e-07,  1.177597338648e-04, {   1,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.198541904628e-05, -9.650789741553e-05, {   2,   0,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.473757334622e-05, -9.374557819108e-05, {   2,   0,  -1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -4.118931614216e-05, -7.722397605583e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   4,   0,   0,   0 } },
 {  2.293136288937e-05,  9.523504163430e-05, {   1,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.022213965507e-05,  8.784055193346e-05, {   1,  -1,   0,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 { -1.098650237468e-04,  8.013473443117e-06, {   2,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.864958596351e-05, -6.921479108318e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -7.596794060626e-05, -4.179467352569e-05, {   4,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.322690281833e-05,  1.045010551506e-04, {   2,   0,  -1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  1.011374278582e-04, -1.658170017871e-05, {   1,  -1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.899088021327e-06, -1.147475814314e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.172871980047e-04, -2.721449338481e-07, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.171152421003e-04, -4.106306867300e-07, {   2,  -2,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.017945710723e-05,  7.728803426874e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  2.138948173490e-05, -9.597558705270e-05, {   2,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.614753643126e-05, -3.121305849071e-05, {   2,   0,  -1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  8.599283260091e-05, -3.118653920123e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -3.478465778926e-06, -1.136424779436e-04, {   2,   0,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.224860075160e-05, -9.480151578090e-05, {   4,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.241921695539e-06,  1.117763276566e-04, {   2,   0,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.188054298735e-05,  3.511677006566e-05, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.901294532409e-05, -3.788346514547e-05, {   2,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.053975224458e-05,  1.062331917012e-04, {   0,   0,   2,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 { -9.927118718675e-05,  1.739644789083e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 {  6.168069038211e-05,  5.498624575670e-05, {   2,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -7.956140730819e-05,  3.709693782624e-05, {   1,  -1,   1,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 { -8.316528231141e-05, -3.342670928490e-05, {   2,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -8.027656948471e-05,  3.613219142862e-05, {   2,   0,  -2,   0,   0,   0,  -3,   9,   0,   0,   0,   0,   0 } },
 { -2.290259390021e-05, -9.341832251590e-05, {   0,   0,   1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -8.253427637309e-05,  3.350712884445e-05, {   0,   0,   0,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  8.601775887197e-07, -1.149543194813e-04, {   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.113627924915e-05, -7.465620970464e-05, {   0,   0,   2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -1.265414392901e-05,  1.031170769376e-04, {   0,   0,   1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -1.156559544288e-04, -5.689308358082e-08, {   4,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.015143550373e-05,  2.549770583726e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -7.580160494779e-05, -3.978001204830e-05, {   2,   0,   1,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  2.246641655213e-05, -9.311515517610e-05, {   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.283248254575e-05, -6.273833263195e-05, {   2,   0,   0,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 {  1.039421036898e-04, -1.161933065657e-05, {   0,   0,   0,   0,   0,   0,  12, -24,   5,  10,   0,   0,   0 } },
 {  3.431919464068e-05,  8.117477094653e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -3.432569101614e-05,  8.116589768969e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  8.202874951267e-05, -3.322532417942e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.143071147730e-04,  9.095421745903e-07, {   2,  -2,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.129545640639e-04, -2.228668242941e-06, {   4,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.681667636324e-05,  5.803974952215e-05, {   2,   0,  -1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 {  4.871123126929e-05,  6.611099869989e-05, {   2,   2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.136338866213e-04,  1.071806024558e-06, {   4,   0,  -4,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -4.880822946788e-05,  6.585165289122e-05, {   0,   0,   0,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 {  2.512994761156e-06,  1.121408544901e-04, {   2,  -2,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.230143205463e-05,  7.230863424483e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -2.645824943522e-05, -8.810374637790e-05, {   0,   0,   2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.124040333268e-05, -9.324497356203e-05, {   1,   1,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -8.374744720150e-05,  3.072415003720e-05, {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  6.758987986601e-05,  4.677549059839e-05, {   2,   0,  -1,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  9.635179397083e-05, -1.791011553494e-05, {   1,  -1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.870301632159e-05, -6.555230614868e-05, {   2,   0,  -1,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 { -4.222400362882e-05,  7.202893687608e-05, {   2,   0,   0,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  4.825232299295e-05,  6.594647468964e-05, {   2,  -2,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  7.373515528295e-05, -4.045872616490e-05, {   2,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  8.149835282898e-05,  3.267315080013e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  8.149835282898e-05,  3.267315080013e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  1.820981437965e-07, -1.138478082720e-04, {   2,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -5.339295219143e-06,  1.086785424659e-04, {   0,   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  7.714405636932e-05, -3.686145890300e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  5.030699301959e-05,  6.361995882223e-05, {   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -5.654396741042e-05, -5.737450723701e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -7.620123636328e-05,  3.762391335153e-05, {   2,   0,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.329390132679e-05,  8.049896032532e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  7.322138296386e-05,  4.048169026888e-05, {   2,   0,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  6.229927802018e-05, -5.135169667224e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -2.359835737116e-05, -8.990776819593e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   3,   0,   0 } },
 {  2.973530804097e-05,  8.374654580846e-05, {   4,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.972902579194e-05,  8.372824074758e-05, {   4,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -6.706864668075e-05,  4.637420770005e-05, {   2,   0,  -1,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -8.861529562342e-07,  1.125396823960e-04, {   1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.937863788726e-05, -9.403863943930e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -5.196622002680e-05, -6.131611896170e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0 } },
 {  5.191874957669e-05, -6.126938289487e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 } },
 { -4.712562299327e-06, -1.084426357215e-04, {   2,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.118828931237e-04, -1.181971402947e-06, {   2,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.072406432159e-05, -9.227841847842e-05, {   1,   0,   0,   0,   0,   0, -48,  67,   0,   0,   0,   0,   0 } },
 {  1.105785286145e-04,  2.406918492082e-06, {   0,   0,   0,   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0 } },
 {  5.502670854605e-05,  5.774612931029e-05, {   1,  -1,  -1,   0,   0,  23, -23,   0,  -1,   0,   0,   0,   0 } },
 {  3.723693293501e-05,  7.545325199807e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.152889021018e-05, -9.104134440265e-05, {   1,   1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.159473191405e-05,  9.077405878610e-05, {   1,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.941019100353e-05,  6.291766862332e-05, {   2,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.156748452575e-05,  9.061631207451e-05, {   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.321159504794e-05, -5.893221291776e-05, {   2,   0,  -1,   0,   0, -11,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.884095089685e-05,  1.329620979337e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 { -1.583902719320e-05, -9.610901306800e-05, {   0,   2,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.556237400662e-05,  8.633307512597e-05, {   2,   0,   0,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 {  4.422315945227e-05, -6.763884624523e-05, {   3,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.299575379598e-05,  6.874299173956e-05, {   2,   2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.106932379985e-05, -6.064335790354e-05, {   2,   0,  -2,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 {  8.179038764469e-05,  2.987988458413e-05, {   2,   0,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.088796485129e-04,  2.772492001632e-06, {   2,   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  6.686753785318e-05, -4.472853789619e-05, {   2,   0,  -1,   0,   0, -23,  26,   0,  -1,   0,   0,   0,   0 } },
 { -8.574763465498e-05, -2.582863256161e-05, {   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.253593535314e-05, -7.902683552656e-05, {   2,   0,  -1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.104905206868e-04,  1.054318937881e-06, {   0,   0,   3,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.995863110896e-05,  9.156659517224e-05, {   1,   0,  -2,   0,   0,  37, -46,   0,   0,   0,   0,   0,   0 } },
 { -1.036068422365e-04,  7.880244290286e-06, {   2,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.186110155008e-05,  3.948009544546e-05, {   3,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  7.186110155008e-05,  3.948009544546e-05, {   1,   0,   1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -7.613464441710e-05, -3.517221915028e-05, {   2,   0,  -1,   0,   0, -21,  23,   0,  -4,   0,   0,   0,   0 } },
 {  9.028064631128e-05,  2.094840091510e-05, {   1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.763847511042e-05, -6.358566664418e-05, {   1,   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.056586540580e-05, -7.062681285284e-05, {   4,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.344853476448e-05,  1.769500293990e-05, {   1,  -1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.306979963007e-05,  3.805133234644e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0 } },
 {  1.035777774296e-04, -7.432531409502e-06, {   3,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.270176880978e-05,  6.824245515583e-05, {   2,   2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.761114289988e-05, -8.299691025890e-05, {   2,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.092919755088e-05,  8.964419032100e-05, {   1,   1,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.123335845305e-05, -8.917002004269e-05, {   1,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.617746294737e-05,  5.420983470190e-05, {   2,  -2,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  6.411203922483e-05,  4.617642325870e-05, {   2,   0,  -1,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 {  8.886023002634e-05,  2.140707416422e-05, {   0,   0,   1,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  8.514868675504e-05,  2.509421907464e-05, {   4,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.169651963772e-05, -8.848197954205e-05, {   1,   1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.178373776862e-05,  8.834457680959e-05, {   1,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.790404687747e-05, -2.213020942513e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  7.242814278088e-05,  3.758523280970e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -1.099999999918e-04,  1.343903523992e-09, {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.099999999996e-04, -2.879793265610e-10, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.095801352468e-04,  4.173352933825e-07, {   2,   0,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.046127989851e-04,  5.298018822036e-06, {   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.101885756402e-05,  8.885306523736e-05, {   1,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.696745967483e-05,  2.285512318529e-05, {   4,   0,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  8.349896827216e-05, -2.627331138542e-05, {   2,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.876260556321e-05,  8.094185621091e-05, {   2,   0,  -1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 { -7.241061100954e-05,  3.729328107412e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  9.907068777448e-06,  9.978509861672e-05, {   0,   0,   0,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 { -9.666245190315e-05,  1.298699871805e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 {  7.689967160606e-05, -3.267275310422e-05, {   4,   0,  -3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.708009382733e-06,  1.048515065598e-04, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -6.683728644334e-05, -4.265803098035e-05, {   2,   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -7.300236771605e-05,  3.646063516008e-05, {   6,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  8.663452475814e-05,  2.276233041311e-05, {   0,   0,   1,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 {  1.085451006650e-04,  8.050984030195e-07, {   1,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.558450861148e-05, -4.374413977065e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -9.624814435533e-05, -1.306247169084e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  5.137904547282e-05, -5.791896619086e-05, {   2,   0,   1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.377470432099e-05, -1.540569437719e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0 } },
 {  4.805183444002e-05, -6.108826078482e-05, {   2,   0,  -3,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  2.388597211407e-05,  8.523785517190e-05, {   1,  -1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.858373668226e-05,  5.048096614807e-05, {   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.589777779603e-05, -9.310775160652e-05, {   2,   0,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.014101765991e-05,  9.883879057279e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.086798810454e-04, -2.460024556902e-07, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.718401809474e-05, -4.157531088917e-05, {   0,   0,   1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 { -7.627748214846e-05, -3.239729782457e-05, {   4,   0,  -1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.057623494817e-05,  6.802774560602e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -1.016294772537e-04, -6.937781776117e-06, {   2,   0,   0,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 {  8.957278136175e-06,  9.959802793425e-05, {   1,   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.949055882731e-06, -9.959876705466e-05, {   1,   1,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.947734753868e-06, -9.959888575067e-05, {   3,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.076908280907e-04, -7.846420811512e-07, {   0,   2,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.084186712947e-04, -4.658620681118e-10, {   1,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.906219212171e-05, -5.934164353165e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -4.066470716760e-05,  6.771242307066e-05, {   0,   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  1.018796269497e-04, -6.482714539166e-06, {   1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.052643561061e-04,  3.058129089200e-06, {   2,  -2,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -9.095020863928e-05,  1.736836361916e-05, {   0,   0,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.119620765006e-05, -8.703289808050e-05, {   4,   0,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.075358173290e-04, -6.689198155446e-07, {   2,   0,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.080994061458e-04, -1.880908201126e-08, {   2,   2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.076984926873e-04,  3.507875391884e-07, {   0,   0,   1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -5.730343311814e-05,  5.069853283473e-05, {   2,  -2,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.075034120550e-04,  4.396311060645e-07, {   2,  -2,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -4.056048527260e-05,  6.737609203637e-05, {   0,   0,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -7.323498884304e-05,  3.468726526866e-05, {   0,   0,   1,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -1.764121911222e-06,  1.061313183820e-04, {   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.211098449155e-05,  7.574512044344e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  7.664255502807e-05,  3.119856136573e-05, {   0,   0,   0,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 {  3.119758494316e-05,  7.661578692060e-05, {   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.269974586007e-05, -8.510545208995e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -5.942508668901e-06,  1.018131822157e-04, {   0,   0,   0,   0,   0,   6,  -6,  -8,   6,   0,   0,  -1,   0 } },
 {  2.945801227337e-06, -1.047836773806e-04, {   2,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.053817047199e-04, -2.248008029845e-06, {   4,  -4,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  1.053817047199e-04, -2.248008029845e-06, {   4,  -4,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -7.280853711660e-05,  3.480705330048e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  2.051038141200e-05,  8.700652032548e-05, {   1,   1,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.599386978628e-06,  1.007712900210e-04, {   2,   0,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -1.036614818320e-04,  3.665131954640e-06, {   4,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.597079395231e-06, -1.057059738102e-04, {   2,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.943696969631e-05, -3.781199578441e-05, {   4,   0,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.011936255301e-05,  8.706549703108e-05, {   1,  -1,  -1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.907256041191e-05,  3.807426055264e-05, {   2,  -2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  4.290702963937e-05,  6.413240387741e-05, {   0,   2,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.049680666039e-04,  2.027847592467e-06, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -7.190808310392e-05,  3.506034204511e-05, {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.212254367900e-07,  1.063382563224e-04, {   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.061337753271e-04, -7.956453189881e-07, {   4,  -2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.599335699534e-05,  7.080287114911e-05, {   0,   0,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  3.555765525393e-05,  7.111131797250e-05, {   2,   0,  -1,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 {  6.526847649462e-05,  4.124927269281e-05, {   2,  -2,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.063561325918e-04,  1.438207892547e-07, {   2,   0,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -4.521999014087e-05,  6.118504506340e-05, {   2,   0,   0,   0,   0, -33,  29,   0,   0,   0,   0,   0,   0 } },
 { -8.297488557608e-05, -2.337468922934e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -6.024280489884e-06, -1.003150968218e-04, {   1,  -1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.054539083662e-04, -8.753187433492e-07, {   2,   0,  -2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  4.552354824278e-05,  6.079677186288e-05, {   3,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.129865510367e-05, -7.501609064052e-05, {   0,   0,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.026903730563e-04, -3.573429665084e-06, {   2,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.013106885103e-07, -1.056332334254e-04, {   4,   0,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.884722649827e-05,  3.735575283324e-05, {   2,   0,   1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  1.773859723340e-05,  8.843874369500e-05, {   1,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.139274359124e-05,  6.476898871890e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 { -1.060907518454e-04,  5.901877868436e-08, {   2,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.636921228190e-06, -9.750355455295e-05, {   2,  -2,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.820804759972e-06, -1.022883020664e-04, {   0,   0,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.502591524315e-05,  4.108178955573e-05, {   2,  -2,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -3.530459752814e-05,  7.076066552066e-05, {   2,   0,  -1,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 { -9.779777827728e-05, -8.188046011065e-06, {   1,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.053127432454e-04,  5.820763052028e-07, {   2,   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -8.967077739522e-05, -1.618662782901e-05, {   1,   1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.055770513205e-05,  3.524034300119e-05, {   6,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.129927792982e-05,  1.442237823574e-05, {   0,   0,   2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.052453840744e-04, -4.291058715839e-07, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -5.580236880396e-06, -9.996196739982e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.541573249925e-06, -1.019377617523e-04, {   2,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.737064156461e-05, -7.795138957834e-05, {   1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.774918695442e-05, -3.755776507927e-05, {   2,   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 { -7.866752652218e-05, -2.656541260859e-05, {   2,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.272237502738e-05,  7.247874318757e-05, {   0,   2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.290723410719e-05, -7.219185404032e-05, {   1,   0,   0,   0,   0, -23,  21,   8,  -3,   0,   0,   0,   0 } },
 { -3.638706795416e-05, -6.866903586165e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 {  2.762425066233e-05, -7.729752998029e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  4.178059808563e-06, -1.006879278315e-04, {   4,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.304962863305e-05,  2.180874846014e-05, {   4,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  5.166823376032e-05, -5.316790404893e-05, {   0,   0,   1,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 { -1.047014881527e-04,  1.096153498875e-07, {   2,   0,  -1,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.707728751720e-05, -1.770410679916e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 {  1.468285555650e-05,  9.002070664914e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  3.424366239483e-05,  7.023709581490e-05, {   0,   2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.954633104948e-05,  8.490536931197e-05, {   1,  -1,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.965153999197e-05,  8.477696368635e-05, {   1,   1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.330939604214e-05,  8.095199092121e-05, {   1,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  8.742339606948e-05, -1.682468049948e-05, {   1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.310072788252e-05, -3.106671267826e-05, {   2,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.010995436650e-04,  3.033282735055e-06, {   4,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -5.102891778555e-06, -9.902354835057e-05, {   2,   0,  -2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  8.868531752804e-05,  1.532691929091e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.038582075884e-04, -1.331299638859e-07, {   2,   0,   1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.059513527633e-05,  4.337155544478e-05, {   3,  -2,   0,   0,   0,   0, -24,  26,  -1,   0,   0,   0,   0 } },
 {  8.311680414703e-05, -2.083978309500e-05, {   1,   1,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -8.310383997263e-05,  2.078417590852e-05, {   1,   3,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.828351590430e-05,  4.559148918701e-05, {   2,   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.996768771471e-05,  6.383803680323e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   3,   0,   0,   0,   0 } },
 {  7.042108261680e-05,  3.330347842342e-05, {   2,   0,  -1,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 { -7.278088478278e-05, -3.092511196579e-05, {   2,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.988963138925e-05,  8.380012178472e-05, {   1,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.997693527678e-05,  8.369929531145e-05, {   1,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.851058194310e-05,  1.516113855196e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.027465671624e-04,  7.876487885509e-07, {   0,   2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -8.451788040930e-05,  1.891624647995e-05, {   2,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.963694731162e-05, -2.375494738669e-05, {   1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.578203460018e-05, -6.760245680384e-05, {   0,   0,   3,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.660561157695e-06,  1.006755476846e-04, {   2,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.785073232802e-05,  6.546271747314e-05, {   1,  -1,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  1.951604265547e-05,  8.376841157940e-05, {   1,   1,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.070901971257e-05, -8.252142997910e-05, {   4,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.915722209926e-05, -8.397508850766e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -6.877979107561e-05, -3.432258314940e-05, {   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.477447794520e-05,  3.828787968177e-05, {   2,   0,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 {  6.933278440390e-05,  3.371133395587e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0 } },
 { -1.022503448806e-04, -7.923607866175e-07, {   2,  -2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.867606271721e-05,  8.427623719820e-05, {   2,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.672672413735e-05,  7.617123522367e-05, {   2,   0,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  8.186102560326e-05, -2.098282462083e-05, {   0,   0,   1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  7.902477957269e-05,  2.378709729482e-05, {   2,   0,  -1,   0,   0,   3,  -3,   0,  -1,   0,   0,   0,   0 } },
 {  2.631891348041e-05,  7.643630298778e-05, {   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.699220362273e-05, -2.576215747535e-05, {   4,  -2,  -2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.000996375067e-05, -8.269205042265e-05, {   1,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.613511616384e-05, -3.648383900145e-05, {   2,   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.543269237696e-05, -6.717702523562e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 {  6.859869467660e-05,  3.398215223089e-05, {   0,   0,   1,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.596743518223e-05,  3.654392002197e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -5,   0,   0,   0 } },
 { -1.747996954541e-05, -8.490269219618e-05, {   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.023075327002e-04,  3.162607134665e-08, {   0,   0,   1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.693780752337e-05,  6.528574697103e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 { -3.015638089676e-06, -9.920017345685e-05, {   2,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.021494151742e-04, -3.645634573356e-08, {   6,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.908772889685e-05, -3.306656778233e-05, {   2,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.578957270477e-06,  9.753995384663e-05, {   2,   0,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -7.699307530682e-05, -2.510977415933e-05, {   2,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.430064158865e-05,  4.761744274120e-05, {   2,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  7.933129053761e-05, -2.255900351374e-05, {   1,  -3,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.041667213140e-05,  6.146062606116e-05, {   2,  -2,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.577619067954e-05,  1.600765403950e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  7.151596716248e-05,  3.024240710979e-05, {   4,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -9.003292767050e-05,  1.154389303588e-05, {   1,  -1,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.449956246383e-05, -6.707383406405e-05, {   0,   0,   3,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  8.047695804155e-05,  2.109171571856e-05, {   0,   0,   0,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 {  1.967165887603e-05,  8.188903978054e-05, {   1,  -1,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.014708556206e-04,  5.795847518786e-08, {   2,   0,  -3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.152449248775e-05,  5.999966362643e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.240769485068e-06,  1.002729933840e-04, {   2,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.160175013877e-06,  9.828838209201e-05, {   2,   0,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.013621806742e-04,  5.295330866270e-08, {   2,   0,  -2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  5.879034713653e-05, -4.262265803351e-05, {   2,   0,  -1,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 {  4.018848078992e-05,  6.108931362780e-05, {   2,  -2,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.931231919179e-05,  8.183843241368e-05, {   1,  -1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.411130594526e-05, -1.684786515379e-05, {   1,  -1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.354678137393e-06,  9.759239849310e-05, {   2,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.026187781696e-05, -8.059207553085e-05, {   1,  -1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.270002737806e-05, -8.146022655757e-06, {   2,   0,  -1,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.027566651356e-05,  9.052965873070e-05, {   0,   0,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.907744284801e-05,  1.171455715231e-05, {   1,   1,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  1.887742087383e-05,  8.190127168763e-05, {   1,   1,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.897962586705e-05,  8.166575662834e-05, {   1,   1,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.448398347892e-05, -3.610538498146e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 {  6.797194897574e-05,  3.260527937214e-05, {   4,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.935157849929e-05,  8.121953170827e-05, {   1,   1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.142786968217e-05, -3.910416142575e-05, {   0,   0,   2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  5.244527077105e-05,  4.800339120791e-05, {   2,   0,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -4.830058931581e-05, -5.211754252329e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 { -5.842643381441e-06, -9.450837423181e-05, {   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.190888087792e-06,  9.310760447249e-05, {   4,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.545333015198e-05, -5.474441763852e-05, {   1,  -1,  -1,   0,   0,   0,  26, -22,   0,   0,   0,   0,   0 } },
 {  9.999999998256e-05,  1.867502299525e-09, {   2,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999998507e-05,  1.727875959388e-09, {   4,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999296e-05,  1.186823891328e-09, {   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999718e-05,  7.504915783505e-10, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999756e-05,  6.981317007921e-10, {   2,  -1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00,  1.000000000000e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.666826763707e-05,  5.329345984000e-05, {   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.201112443196e-05,  3.794159847216e-05, {   1,   0,   0,   0,   0,   0, -23,  20,   0,   0,   0,   0,   0 } },
 {  9.891993021986e-05, -1.031202418407e-06, {   3,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.996374447890e-05,  7.996714773996e-05, {   1,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.981591983480e-05,  3.011323329199e-05, {   0,   0,   0,   0,   0,   0,   2,  -3,  -6,   3,   0,   0,   0 } },
 {  7.019024072552e-06,  9.286068797992e-05, {   2,   0,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.006382849426e-05, -2.977004168506e-05, {   4,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.006382849426e-05, -2.977004168506e-05, {   4,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.472247401590e-06, -9.633131417929e-05, {   2,   0,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.109036781692e-05,  2.860301013159e-05, {   4,   0,  -2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.438672044197e-05, -5.525342211163e-05, {   0,   0,   1,   0,   0, -14,  10,   0,   0,   0,   0,   0,   0 } },
 {  7.163012442111e-05, -2.796411538553e-05, {   0,   0,   2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.076174699073e-05,  3.881987335802e-05, {   4,   0,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.680965416286e-05,  5.276848864185e-05, {   2,   2,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.680965416286e-05,  5.276848864185e-05, {   2,  -2,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.541376404274e-05, -2.415799652837e-05, {   2,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  8.190821440550e-05, -1.757854329492e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 { -3.576381960643e-06, -9.585409692686e-05, {   0,   2,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.797137116296e-05,  6.145651550813e-05, {   2,   0,  -1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -3.909695943194e-05, -6.027540231749e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 { -6.042177622067e-05,  3.860444922746e-05, {   4,   0,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -6.775895225644e-05,  3.118742104983e-05, {   2,   0,   0,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -7.975483443751e-05, -1.918818818247e-05, {   4,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.762034152146e-05,  1.123193311702e-05, {   0,   0,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  3.775629169290e-05, -6.097564762934e-05, {   2,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  5.781365282977e-05, -4.090910905753e-05, {   4,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -9.515394233057e-05,  3.487068198690e-06, {   2,   0,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -9.846424019549e-05,  4.174427207453e-08, {   0,   0,   2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.746273881020e-05, -9.641624181760e-07, {   3,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.989366621524e-05, -8.505825937738e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  9.809230620778e-05, -2.714503376312e-07, {   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.231734392127e-05, -1.589781864836e-05, {   0,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  8.024676096667e-05,  1.796436120919e-05, {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0 } },
 {  9.601928126513e-05,  2.104492020060e-06, {   2,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -6.610039157062e-05, -3.199829024475e-05, {   0,   0,   0,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  1.878625966736e-05, -7.929757541812e-05, {   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -9.669465409832e-05, -1.253497266348e-06, {   1,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.486622929281e-05,  5.286761364079e-05, {   1,   0,   0,   0,   0, -21,  25,  -7,   8,  -6,   0,   0,   0 } },
 { -6.847533213795e-05,  2.924289842600e-05, {   4,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -7.866210383631e-05,  1.904224258539e-05, {   1,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.896999370770e-05, -4.872454438703e-05, {   0,   0,   0,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 {  2.779393710465e-05, -6.987522008451e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.875132218213e-06, -9.378721807866e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.004048237702e-05, -6.756455455057e-05, {   1,   1,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.657990810922e-05, -6.100138904962e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.150624377540e-06,  9.635931324765e-05, {   2,   2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.432060230427e-05,  3.122252297344e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -8.238745287696e-05,  1.502873474347e-05, {   1,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -6.191018699829e-05,  3.550078571457e-05, {   2,   0,  -1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -3.066220720812e-06,  9.423886631866e-05, {   2,   0,  -1,   0,   0, -17,  16,   0,   2,   0,   0,   0,   0 } },
 {  9.711011332715e-05,  9.532005496091e-08, {   1,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.942505099703e-05, -3.758545364124e-05, {   0,   0,   2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  9.558647552700e-05,  1.400275093880e-06, {   1,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.067700291491e-06, -9.388409930465e-05, {   2,   0,  -1,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.216326438951e-05,  4.749617512332e-06, {   6,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.993790790452e-05,  2.688188047172e-05, {   2,   0,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 { -4.691545767440e-05,  4.989911791333e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0 } },
 { -1.898400462872e-05,  7.780059067603e-05, {   1,   1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.783850213042e-06, -9.296136579448e-05, {   2,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.647928888809e-05,  2.201633688497e-07, {   4,   0,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.429244561911e-05, -6.240534272905e-05, {   0,   0,   1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.095431556815e-05,  8.573210866556e-05, {   2,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.541555073966e-05, -3.125336758629e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  5.689915092077e-05, -3.975685400326e-05, {   1,  -1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -8.411988786521e-05, -1.236435700217e-05, {   2,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.602658463271e-05, -2.927179162267e-07, {   2,   0,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  6.582946281147e-06, -8.969837640208e-05, {   2,   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -6.719091184207e-05, -2.904849904629e-05, {   1,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -9.490930939466e-05,  1.251487058479e-06, {   4,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.846560282677e-05, -7.759701787792e-05, {   0,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.583561333839e-05, -5.015549955422e-05, {   2,   0,  -1,   0,   0, -20,  29, -15,   0,   0,   0,   0,   0 } },
 { -6.463414665475e-05,  3.133874769922e-05, {   2,   0,  -1,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.833833560414e-05,  7.751706288338e-05, {   1,  -1,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.062335867364e-07,  9.573443842461e-05, {   2,   0,  -1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  3.688862929064e-05, -5.892705065174e-05, {   2,   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 {  2.769603750053e-05,  6.798059843896e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0 } },
 { -7.803903830301e-05, -1.760421826554e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.279737084927e-05,  4.282874634524e-05, {   2,   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.479309842710e-05,  1.082928134293e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -3.852981625318e-05,  5.706008170176e-05, {   0,   2,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.960334062566e-05, -7.588069238854e-05, {   2,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  5.036853378810e-05, -4.501285844322e-05, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.083518014438e-05, -3.445689906371e-05, {   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  9.294996011184e-05, -2.311206888989e-06, {   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.069788250768e-05,  7.455415076652e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,  10,   0,   0,   0 } },
 {  7.497750606460e-05,  2.027395904384e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 {  7.918346476335e-05,  1.603692612427e-05, {   4,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.856066078405e-05,  3.649358501750e-05, {   2,   0,  -1,   0,   0,   3, -18,  25,  -2,   0,   0,   0,   0 } },
 { -5.804456363176e-05, -3.691186110340e-05, {   2,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.842778334560e-05,  3.641974920254e-05, {   2,   0,  -1,   0,   0,  -3,  18, -25,   2,   0,   0,   0,   0 } },
 { -9.058789923083e-05, -4.184259760632e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -2.824152157092e-05, -6.646158292429e-05, {   2,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.898483017643e-05,  7.566430059929e-05, {   2,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.742509586373e-05,  7.183742885263e-06, {   2,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.766871730117e-07, -9.416659569635e-05, {   3,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.398942931186e-05,  1.053475749317e-05, {   0,   0,   0,   0,   0,   0,   5,  -8,   1,   5,   0,   0,   0 } },
 {  1.819670813973e-05, -7.626687447200e-05, {   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.347926573169e-05,  3.095214355477e-05, {   2,   0,  -1,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.314151875750e-05,  5.114727920503e-05, {   1,   0,   0,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.881793822353e-05,  7.546218632844e-05, {   2,  -2,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.301592369471e-05,  1.125743646434e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  5.900349978104e-05,  3.525601412615e-05, {   1,   1,   0,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.585516942008e-05,  2.830358197740e-05, {   1,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  8.482104821454e-05,  9.332201136326e-06, {   1,  -1,   1,   0,   0, -21,  23,  -2,   0,   0,   0,   0,   0 } },
 { -1.773053626263e-05, -7.639167795261e-05, {   1,  -1,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.568679706914e-05,  1.831194715581e-05, {   1,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.523739243070e-05,  5.874354142045e-05, {   1,   1,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  6.355234257411e-05, -3.036104524048e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -3.229564443300e-06,  9.068274940452e-05, {   3,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.610672052451e-05, -5.778048447713e-05, {   4,   0,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  6.992443211205e-05,  2.390434754275e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  2.573286205877e-05, -6.806979975973e-05, {   1,   1,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -7.798025189926e-05, -1.581064682123e-05, {   2,   0,   0,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  3.539186888475e-05, -5.827500984709e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  6.747763610530e-05, -2.617031833141e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 {  7.745146510784e-06, -8.583472461504e-05, {   2,  -2,   0,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 {  4.087417444664e-05, -5.270190925049e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 { -8.007841420428e-05, -1.347255258396e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.034547570462e-05,  3.167542425691e-06, {   2,   0,  -1,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 { -3.878460698254e-05,  5.465557586475e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 {  1.899979932158e-05,  7.443959454522e-05, {   3,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.932938883907e-05,  5.410280362627e-05, {   2,   0,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  6.373429714113e-05, -2.963476696375e-05, {   2,   0,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -3.788429125056e-06,  8.955417131585e-05, {   4,  -2,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.531009578232e-05,  1.801351489812e-05, {   2,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.871684059655e-05, -6.459152080507e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -1.800632981592e-06, -9.146546087921e-05, {   0,   0,   2,   0,   0, -44,  45,   0,   0,   0,   0,   0,   0 } },
 {  4.473650878870e-05,  4.851384089166e-05, {   2,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.679766531017e-05,  2.640938003870e-05, {   0,   0,   0,   0,   0,   0,   7, -14,   0,   0,   0,   0,   0 } },
 { -4.355072116983e-05, -4.962380816616e-05, {   4,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.228316144031e-05,  8.088686526531e-05, {   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.917060142539e-05,  5.399703827365e-05, {   2,   0,  -1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.304881221288e-05, -1.156411001293e-07, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -6.213201719366e-05,  3.103175540882e-05, {   6,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.533804393805e-05, -6.777282306641e-05, {   0,   0,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.018450087777e-05,  7.285947356867e-05, {   2,   0,  -2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.152529134833e-05,  5.147934184516e-05, {   3,  -1,  -1,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.458023619762e-05,  7.835030643121e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  9.268060100494e-05, -2.292340905412e-07, {   0,   2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.363412771046e-05, -2.916672402806e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.775828271670e-05,  7.503953047693e-05, {   1,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.726438650453e-05,  1.552279715970e-05, {   1,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.633698980894e-06, -8.395620764418e-05, {   2,   0,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  6.509641976312e-05,  2.749312346451e-05, {   2,   0,  -1,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -9.240352888442e-05,  1.820041873916e-07, {   0,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.248379819836e-05,  5.440642125479e-08, {   1,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.650119639491e-05,  1.601512827638e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -6.390404335142e-05,  2.857048555695e-05, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.422185352500e-05,  6.821421637270e-05, {   4,   0,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.421704296795e-05,  6.820358719044e-05, {   4,   0,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -6.507281103029e-05,  2.733454644104e-05, {   2,   0,  -1,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 { -7.750986551413e-05,  1.486914640790e-05, {   2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.418079259212e-05, -6.812674104324e-05, {   1,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -4.192245012033e-05,  5.036467641641e-05, {   0,   0,   2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.297915271921e-05, -2.929551670037e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.418396425474e-05, -6.808652754259e-05, {   1,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.798075679672e-05, -7.427221797578e-05, {   1,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.152619128119e-05, -2.072670654671e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.402913453717e-05, -6.820711805373e-05, {   1,  -1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  9.032794817100e-05,  1.889057724408e-06, {   4,   0,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  6.368154378856e-05,  2.853362825868e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 {  3.167133557805e-06,  8.898008337631e-05, {   2,  -2,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.820027722931e-05,  6.391293696955e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 {  4.599596357597e-05,  4.609052527208e-05, {   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.187962677389e-05,  1.877836782553e-07, {   2,  -2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  3.495262394231e-05, -5.711149701740e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 { -6.355311770379e-05, -2.843293719942e-05, {   1,   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -7.188212438291e-05,  2.009565943976e-05, {   2,   0,  -2,   0,   0,   0,  10, -15,  -3,   0,   0,   0,   0 } },
 {  1.750288947200e-05,  7.444208570196e-05, {   1,   1,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.752522540382e-05,  4.350261302648e-06, {   2,  -1,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.982594622494e-05,  6.204789684356e-05, {   2,   0,  -1,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  6.276482085947e-05,  2.906892022638e-05, {   2,   0,   0,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  9.130620041954e-05, -4.894102272240e-07, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -9.170459162233e-05,  9.042050241205e-08, {   2,   0,  -2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  7.585128596945e-05,  1.591042689772e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 {  6.840736227350e-05,  2.329752409233e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 {  7.982953879805e-05, -1.184002085259e-05, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 { -7.632741069120e-05,  1.521704594105e-05, {   1,  -1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.794816977869e-05,  7.356696099338e-05, {   2,   0,  -1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.094490385681e-05,  6.053798281082e-05, {   3,   0,  -2,   0,   0, -22,  25,   0,   2,   0,   0,   0,   0 } },
 { -1.722194845839e-05, -7.425317605659e-05, {   0,   1,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.438317970819e-05,  1.692910445339e-05, {   1,  -1,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  6.088238535926e-05,  3.040789909567e-05, {   6,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.749031673525e-05, -7.378292823228e-05, {   1,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.638311613720e-05, -2.487401785485e-05, {   0,   0,   3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.215354397935e-05,  6.910237466216e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -2.615020864773e-05,  6.493201512105e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.257871236670e-05, -5.848996687272e-05, {   2,   0,  -3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.991449878914e-05, -1.102322068946e-06, {   0,   0,   2,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.739794384601e-05,  7.353182425680e-05, {   1,  -1,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.065747188843e-05, -3.023776932054e-05, {   0,   0,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.730355113490e-05, -7.357674923439e-05, {   1,  -1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.260989003033e-05, -3.818682163808e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 {  1.789904923933e-05,  7.280775147958e-05, {   2,  -2,   1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.458842423455e-05,  1.600216390272e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -8.481027724280e-05,  5.691823394391e-06, {   2,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.573257588122e-06, -8.885851526221e-05, {   3,  -2,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.573257588122e-06, -8.885851526221e-05, {   3,  -2,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.944520698172e-05,  6.097253458487e-05, {   2,   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  4.051894213421e-05,  4.979079965525e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -4.915836136993e-05,  4.102595034686e-05, {   1,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.095088418613e-05, -4.923203839946e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  6.449489546029e-05, -2.564335858462e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 { -7.272369957814e-05,  1.740978745223e-05, {   2,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.944647463982e-05,  6.867716185781e-07, {   0,   2,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  8.999993910698e-05, -1.046935515693e-07, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.552480458984e-05,  6.455864108948e-05, {   1,  -1,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -6.290342234402e-05,  2.713154697779e-05, {   2,   0,  -1,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 {  8.999999993365e-05, -3.455751918080e-09, {   2,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.999999997960e-05,  1.916371518545e-09, {   2,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.999999999403e-05,  1.036725575662e-09, {   2,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.999999999684e-05, -7.539822368688e-10, {   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.955471579536e-05, -7.035570224177e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -8.985069082068e-05,  3.983746845002e-08, {   2,  -2,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.211682034753e-05, -5.773388600211e-05, {   2,   0,  -3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.686376545125e-05,  3.297625150928e-05, {   0,   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -8.554431934449e-05, -4.264298031080e-06, {   0,   0,   1,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 {  8.694405248958e-05,  2.825560815805e-06, {   0,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0 } },
 {  4.876584839725e-05, -4.100181451348e-05, {   2,   0,  -1,   0,   0, -20,  12,  17,   0,   0,   0,   0,   0 } },
 {  2.187445408137e-07,  8.949211221964e-05, {   2,   0,   0,   0,   0, -41,  42,   0,   0,   0,   0,   0,   0 } },
 {  3.829870598968e-05,  5.138376079097e-05, {   1,   0,  -2,   0,   0,  13,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.442703657282e-05, -5.525318601153e-05, {   2,  -2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  5.086550416693e-05, -3.881001359607e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -5.980306659091e-05, -2.986863779349e-05, {   2,   0,  -5,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.965444435525e-05, -3.143058579551e-09, {   1,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.410284312258e-05, -5.550737430314e-05, {   1,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.515057836932e-05, -3.441371893162e-05, {   0,   0,   1,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 {  4.726496955007e-05,  4.227297241296e-05, {   0,   0,   0,   0,   0,   8, -14,   2,   0,  -2,   0,   0,   0 } },
 { -6.510901085136e-05, -2.442495256332e-05, {   0,   0,   0,   0,   1,   0, -14,  19,  -3,   0,   0,   0,   0 } },
 {  6.121567914612e-05,  2.829617311171e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0 } },
 { -6.540395668774e-05, -2.408457365626e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -5.980066888496e-05,  2.968121559670e-05, {   0,   0,   1,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.305308426238e-05,  1.640508176481e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  5.958151473515e-05,  2.985472419901e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -1.609236557148e-05,  7.331848828008e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.523326199087e-05, -7.411940997693e-05, {   3,  -3,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -1.523326199087e-05, -7.411940997693e-05, {   1,   3,   0,   0,   0, -43,  42,   0,   0,   0,   0,   0,   0 } },
 { -6.443932185975e-05, -2.489872884456e-05, {   0,   0,   2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -5.333031907284e-07,  8.869130564936e-05, {   1,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.188288755294e-05, -4.734118544226e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 { -4.188288755294e-05, -4.734118544226e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 { -1.706010213886e-05, -7.212842305473e-05, {   1,  -1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.080878404670e-06,  8.608150413264e-05, {   0,   0,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -5.207386499591e-05, -3.708631777358e-05, {   2,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -1.827879362464e-05, -7.087540900007e-05, {   0,   0,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -7.600173985378e-05,  1.313973310455e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.690855957089e-05, -7.220875432100e-05, {   2,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.264809451907e-06,  8.483786072829e-05, {   3,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.878591067849e-05, -9.988218803569e-08, {   2,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  6.034715100326e-05,  2.839729049335e-05, {   2,   0,  -1,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 { -7.232050116459e-05,  1.634304482685e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 {  8.106695950560e-07,  8.781685923730e-05, {   2,   0,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.020885235536e-05,  1.839681713682e-05, {   5,   0,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -8.772470134734e-05,  7.926193217332e-07, {   2,   0,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -7.219031545599e-05,  1.630599947720e-05, {   1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.455313929214e-05,  2.387766577549e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 {  7.417616685003e-05,  1.418368527851e-05, {   4,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  8.273156694695e-05, -5.598121291890e-06, {   2,   0,   2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.415996858157e-05,  5.412234380001e-05, {   1,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  8.171538001543e-05,  6.473678285032e-06, {   1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.895519725325e-05,  9.200834503161e-06, {   0,   0,   1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -2.506021981279e-05, -6.304383211456e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.705821969955e-05, -6.088319817064e-05, {   0,   2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.745639984230e-05,  6.036347739195e-05, {   0,   0,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  6.769397196732e-05,  2.005175103693e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -6.414372048841e-05, -2.355869530343e-05, {   2,   0,  -1,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.666721907356e-05,  9.952146419191e-07, {   2,   0,  -2,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 { -8.669716518343e-05,  9.300131827794e-07, {   2,  -2,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.746643947922e-05,  7.015377855008e-05, {   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.993120171021e-05, -5.763127698085e-05, {   2,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  4.630570033399e-05,  4.125452139610e-05, {   2,   0,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  6.118249171312e-05, -2.635275221318e-05, {   1,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  6.090517453914e-05, -2.651205517622e-05, {   0,   0,   2,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.729889915558e-05, -1.131707784970e-07, {   2,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.451072067972e-05,  1.288158674058e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -5.909348300098e-05, -2.820290274806e-05, {   2,   0,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.438251494643e-05,  2.877576274756e-06, {   2,   0,  -1,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 { -1.603773679942e-05, -7.106473211950e-05, {   1,  -1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.655243209759e-05,  2.049612961527e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  8.558015843743e-05,  1.450156367791e-06, {   2,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -3.858226784430e-06,  8.312966330871e-05, {   2,   0,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.691008136394e-05, -1.250842873772e-08, {   1,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.246365328890e-06, -7.967113774461e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.541573807168e-05,  2.146675763174e-05, {   2,   0,  -1,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 { -3.781119899642e-05,  4.906956127011e-05, {   2,   0,  -1,   0,   0,   0, -29,  33,   0,   0,   0,   0,   0 } },
 { -5.621208582881e-05, -3.065156116441e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  7.160816151872e-06,  7.967887243203e-05, {   1,   1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.159202986350e-06,  7.967901739229e-05, {   3,  -1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.158118269952e-06, -7.967911484720e-05, {   3,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.155448193167e-06, -7.967935467331e-05, {   1,  -1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.856775950344e-07,  8.604709869900e-05, {   2,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.489165335626e-05, -6.183968765785e-05, {   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  4.491941596599e-05, -4.170597024630e-05, {   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.162834229176e-06, -8.342055154741e-05, {   2,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  7.279348971449e-05, -1.378157289495e-05, {   0,   0,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.572525715927e-05,  4.082087621834e-05, {   2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  7.673143224105e-05,  9.788779872481e-06, {   0,   2,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -3.037693804803e-05,  5.609281854853e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -5.032042528238e-05, -3.611159050168e-05, {   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 { -7.659003583294e-05,  9.814562454955e-06, {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -6.096804877209e-05, -2.541646149546e-05, {   0,   0,   3,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.529423185087e-05,  2.108284378622e-05, {   0,   0,   2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.291395007262e-05, -2.346001069319e-05, {   1,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.318272132076e-05,  5.317030230427e-05, {   2,   2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.638579869928e-05, -6.995181801126e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 { -8.392902092069e-05,  2.407365221864e-06, {   1,   1,  -1,   0,   0,   0, -37,  46,   0,   0,   0,   0,   0 } },
 {  1.470492247860e-05, -7.154869645723e-05, {   3,  -3,  -1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -1.470492247860e-05, -7.154869645723e-05, {   1,   3,  -1,   0,   0, -43,  42,   0,   0,   0,   0,   0,   0 } },
 { -1.757124712883e-05,  6.867967764637e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  5.835001712288e-05, -2.786791758058e-05, {   1,   1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -5.227810225808e-05, -3.387287231583e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 { -9.777261106396e-06,  7.637024501034e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  6.261152941538e-05,  2.334166929479e-05, {   4,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -8.458939539223e-06,  7.747469971816e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.874960272867e-05, -5.714328947623e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.514278149651e-05, -4.072113368695e-05, {   1,   1,  -2,   0,   0,   0,   1,  -1,   3,   0,   0,   0,   0 } },
 { -3.217962494209e-05,  5.368333920938e-05, {   1,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.268218173481e-05, -7.309923185605e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 {  6.175098399854e-05, -2.401373045400e-05, {   2,   0,  -1,   0,   0,   0,  -3,   9,   0,   0,   0,   0,   0 } },
 {  8.559777066979e-05,  1.579620434166e-07, {   4,  -2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.047918700668e-05,  6.523175791321e-05, {   2,   0,   0,   0,   0, -46,  54,  -8,   3,   0,   0,   0,   0 } },
 { -3.851754942747e-05, -4.717427904716e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 { -6.475888555487e-05, -2.090354235992e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 { -3.685234434511e-05,  4.866135576754e-05, {   0,   0,   0,   0,   0,   0,  10, -18,   0,   0,   0,   0,   0 } },
 {  6.129203221090e-05,  2.421947626783e-05, {   2,   0,   0,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 { -5.295527058130e-05,  3.248419182012e-05, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 {  3.184906362117e-05,  5.356063696374e-05, {   4,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.625166008218e-05,  3.901925551568e-05, {   2,   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.951300420580e-05,  6.564791052467e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  7.744847068884e-05, -7.667466725361e-06, {   2,   0,  -1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 { -5.989539543900e-06,  7.911050142508e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  1.446321929390e-05, -7.054752482326e-05, {   1,  -1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  3.453867954712e-06,  8.153319442266e-05, {   0,   2,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.947748410672e-05, -1.546825108850e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  8.417365483753e-05,  7.421519872006e-07, {   1,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.599170819080e-05,  6.891454483813e-05, {   1,   1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.023886555037e-05, -3.465172107062e-05, {   2,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -1.651518595020e-05, -6.825225225450e-05, {   0,   0,   4,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.032377106670e-05, -1.433598159444e-05, {   1,  -1,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.285775587002e-05,  1.743474188055e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 {  5.651253853073e-05, -2.808566686202e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  4.965412524999e-07,  8.404818323690e-05, {   2,  -2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.501110242300e-07, -8.397007686706e-05, {   4,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.683718527330e-05,  6.764923025588e-05, {   1,   1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.321824426921e-05,  5.123809878293e-05, {   0,   0,   2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -3.005124877301e-05, -5.435731934031e-05, {   2,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -6.596281586624e-05, -1.838717471773e-05, {   2,   0,  -2,   0,   0,   3,  -3,   0,  -1,   0,   0,   0,   0 } },
 { -7.401795099116e-05,  1.027804440943e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -7.000296704107e-05, -1.427531073079e-05, {   1,  -1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.153805439972e-05,  5.273285731649e-05, {   1,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.736341278690e-05, -1.684582739020e-05, {   3,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.605736013156e-05,  6.813298250729e-05, {   1,  -1,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.899688543485e-05,  3.512897662315e-05, {   1,   1,  -2,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -6.933057081716e-05, -1.476930853354e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -3.992032585039e-05,  4.414387744896e-05, {   2,   0,   1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -6.799451994638e-05,  1.604525335455e-05, {   0,   1,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.279178310421e-05, -3.120798855710e-05, {   0,   0,   2,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.656818209889e-05, -6.735017224554e-05, {   2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.634795086547e-06,  8.227780381100e-05, {   0,   2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.852490705405e-05,  5.534433809934e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 { -8.308361481880e-05,  7.615324412864e-07, {   1,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  4.728122804295e-05,  3.654038717713e-05, {   1,   0,  -1,   0,   0,   0, -10,  20,   3,   0,   0,   0,   0 } },
 {  4.843403133775e-05,  3.537307605575e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   3,   0,   0,   0 } },
 { -1.933099755478e-05, -6.440977587616e-05, {   1,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.519741680933e-05, -4.851723731937e-05, {   2,   0,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.398361061249e-06,  7.929949479314e-05, {   2,   0,  -1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -7.094263061605e-05, -1.273946387832e-05, {   4,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.612465654242e-05, -6.752398519864e-05, {   0,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.727734602134e-05,  2.635377127064e-05, {   1,   1,  -2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.252263487005e-05, -3.106636496045e-05, {   0,   0,   2,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  5.580390058260e-05, -2.773471675964e-05, {   1,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.511652644612e-08,  8.339811710253e-05, {   1,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.144972298478e-05,  6.198201354345e-05, {   2,   0,  -1,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.262683218105e-05,  5.074475904378e-05, {   2,   2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.124451994008e-06,  7.818856766789e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 } },
 {  8.134656158705e-05, -1.839445907932e-06, {   0,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.187161496957e-05,  4.126653077857e-05, {   1,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.856276037973e-05,  6.457091098466e-05, {   1,  -1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.111319733805e-05, -1.985048876432e-06, {   2,   0,  -1,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.787976433627e-05, -4.518983104557e-05, {   1,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  9.472513954257e-06, -7.352243076863e-05, {   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.983882305241e-05,  6.312734798646e-05, {   4,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.534755075273e-05,  2.759875311383e-05, {   1,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.286796707963e-05, -1.650995560517e-08, {   4,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.422718406719e-05, -6.858090155958e-05, {   2,   0,  -1,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.786225567780e-06, -7.795991824669e-05, {   2,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.006988480149e-05,  1.264835377799e-05, {   2,   0,   0,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  3.237515640315e-05,  5.032158123656e-05, {   2,   2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.875952588445e-05,  4.391930368388e-05, {   2,   0,   0,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 {  4.366118734774e-05,  3.891543203414e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 { -8.185082525244e-05, -6.966483116601e-07, {   3,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.376404713044e-05,  1.875491757771e-05, {   1,   0,   0,   0,   0, -26,  30,   0,  -1,   0,   0,   0,   0 } },
 { -1.712968004846e-05,  6.538825969477e-05, {   0,   2,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.712789249909e-05,  6.538299725044e-05, {   0,   2,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.470733120701e-05, -5.779145898403e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   4,   0,   0,   0 } },
 { -7.447841529275e-05,  8.016930235493e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  7.090029064504e-05, -1.158977395726e-05, {   1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.994945342006e-05, -1.250345272045e-05, {   3,  -2,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.736561987618e-05, -6.507474113294e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  1.399534174684e-05, -6.815650565718e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 {  6.620738595930e-06, -7.552190664844e-05, {   2,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -6.714493537062e-05, -1.498258375989e-05, {   1,  -1,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.524442912112e-07, -8.180372962157e-05, {   2,   0,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.372970842361e-05, -2.828952238322e-05, {   3,  -2,   0,   0,   0, -18,  23,  -8,   3,   0,   0,   0,   0 } },
 { -1.674958950508e-06, -8.034242887242e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.517203763828e-05, -4.684028599812e-05, {   4,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.878706655251e-05,  1.320586108433e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -4.565896707943e-06, -7.742485078097e-05, {   2,   0,  -3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.467707490383e-05, -2.730857517385e-05, {   1,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.552991353802e-05, -6.638211318849e-05, {   3,  -1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.826018963676e-05, -3.600231069555e-06, {   2,   0,   1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  5.468823538730e-05,  2.714965029053e-05, {   0,   0,   1,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.269539545811e-05,  4.912828454820e-05, {   1,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.449379704144e-06, -7.537215248807e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -5.622022736005e-05,  2.553624702186e-05, {   2,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.752058777476e-06, -7.998081201163e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.969105211325e-06, -7.870492943554e-05, {   2,   0,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -8.107076468395e-05,  5.644288306575e-07, {   4,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.990709688053e-05, -3.170793868787e-05, {   2,   0,  -1,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 { -8.156398656687e-05, -3.421248510286e-08, {   0,   0,   4,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -8.144845950323e-05, -1.282349908577e-07, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 {  3.512717747462e-05, -4.644655479882e-05, {   0,   2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.647640588340e-05, -3.506213225380e-05, {   0,   0,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.274917399397e-05, -6.873080784346e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.853974786282e-05, -2.919389192965e-06, {   3,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -8.096145409333e-05, -4.790024648078e-07, {   2,   2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.513833720794e-05, -4.623323207091e-05, {   0,   2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.131070899933e-05,  6.002031934011e-05, {   2,  -2,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.130852340428e-05,  6.001271196304e-05, {   2,  -2,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -4.622220902127e-07,  8.083318921451e-05, {   1,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.781650350085e-05, -6.347639286604e-05, {   2,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.955575252803e-05, -4.171517386703e-05, {   2,   0,   0,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -3.349714672968e-06,  7.791840816447e-05, {   2,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.780115657693e-05, -6.342792865198e-05, {   2,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  8.109664663462e-05,  6.663679855109e-08, {   4,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.342956611812e-05, -2.772649560548e-05, {   2,   0,   0,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  8.112112666859e-05,  3.409693536849e-08, {   1,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.101263561217e-05,  1.289372965972e-07, {   1,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.056369873184e-05,  7.051154929053e-05, {   0,   0,   2,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 { -1.183102717811e-05,  6.924344043592e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 {  5.514938249727e-05, -2.586818593215e-05, {   2,   0,   0,   0,   0, -36,  34,   0,   0,   0,   0,   0,   0 } },
 { -7.160687321026e-05, -9.282568144015e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  3.231554649616e-05,  4.856415936953e-05, {   1,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.521720600804e-05, -6.565783486164e-05, {   3,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.253593819500e-05, -2.826473128823e-05, {   2,  -2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.144239131799e-05, -4.935473103135e-05, {   2,   0,  -2,   0,   0,   0,   8, -12,   2,   0,   0,   0,   0 } },
 {  7.649040646939e-05, -4.275410769294e-06, {   0,   0,   2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.535115085183e-05,  6.540209840697e-05, {   1,   1,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.337330373480e-05, -7.366813159452e-06, {   1,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -3.929620958736e-05, -4.139521796697e-05, {   2,   0,   1,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 {  2.609449988372e-05, -5.455232127138e-05, {   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.641776462938e-05, -4.421403653926e-05, {   0,   0,   4,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.590782719880e-05,  3.464456258890e-05, {   1,   1,  -4,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.591118897847e-05, -6.463389521980e-05, {   1,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.844488785259e-05, -3.206680484449e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 { -7.643429426479e-05, -4.009749723741e-06, {   4,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -7.015752345045e-05, -1.024581585499e-05, {   1,  -1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  9.682587490685e-06, -7.067855885036e-05, {   1,   0,   0,   0,   0, -23,  29,  -8,   3,   0,   0,   0,   0 } },
 { -8.024658688939e-05, -6.133314233344e-08, {   0,   2,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  5.476356633661e-05,  2.550102473607e-05, {   0,   0,   1,   0,   0,  -9,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.954168997673e-06, -7.725229034042e-05, {   2,   0,  -2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.903513664544e-05, -4.111653140722e-05, {   2,   0,  -1,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -7.999987208446e-05, -1.430610718785e-07, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.205848744240e-05,  3.808419540138e-05, {   2,   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.505876983680e-05, -6.508318987189e-05, {   3,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.879474476153e-05,  1.133795053091e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -7.999985821411e-05, -1.304388242304e-07, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.731476706372e-05,  3.276357586046e-05, {   2,   0,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 {  9.647752392256e-07,  7.906251623124e-05, {   2,   0,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.593671877495e-05,  2.408994376465e-05, {   2,  -2,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -7.999999998303e-05,  1.647590813766e-09, {   2,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.999999998757e-05, -1.410226035549e-09, {   2,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.999999999453e-05, -9.354964790668e-10, {   2,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.999999999795e-05,  5.724679946493e-10, {   4,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.999999999875e-05,  4.468042885082e-10, {   2,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.999999999979e-05,  1.815142422073e-10, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.624460754204e-05,  6.373360558447e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -5,   9,   0,   0,   0 } },
 {  7.990094241789e-05,  3.825686301810e-08, {   4,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.344049126263e-05, -2.642930109378e-05, {   1,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.730268473085e-05,  6.256297498425e-05, {   2,   0,   0,   0,   0, -46,  46,   8,  -3,   0,   0,   0,   0 } },
 {  2.889034195237e-06, -7.696310418971e-05, {   0,   0,   1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.515126942181e-05,  2.464408124002e-05, {   0,   0,   0,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0 } },
 { -3.765280550599e-05,  4.195324102119e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -1.766021826397e-05, -6.193034134112e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  7.727918296913e-05, -2.205487383213e-06, {   2,   0,   0,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 {  7.854546168941e-05, -9.120093830213e-07, {   2,   0,   2,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.185843011893e-07, -7.871939849840e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.847344843858e-05,  5.094163703633e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 { -1.546838407536e-05,  6.392765808259e-05, {   4,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.865469292839e-05,  1.074122433821e-05, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  5.849214951466e-05,  2.089208597042e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -5.080614932984e-05, -2.854822776792e-05, {   2,   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  7.787190177475e-05,  1.468339156057e-06, {   0,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.514931192451e-05, -6.410851913758e-05, {   1,   1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.174176270715e-05, -2.751330124113e-05, {   2,   0,   0,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.913016442994e-05,  4.011636541425e-05, {   0,   0,   0,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 { -4.729389877034e-05,  3.182966283046e-05, {   1,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  7.154568158060e-05,  7.564494600010e-06, {   1,  -1,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -7.499974427336e-05, -4.071200063007e-06, {   0,   0,   2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  5.143259604781e-05,  2.763731810017e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  2.661135431980e-05,  5.243880815850e-05, {   1,   0,   0,   0,   0,   2, -31,  28,   8,  -6,   0,   0,   0 } },
 {  6.780662195755e-05, -1.108829516184e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -4.488817777393e-05,  3.398017021768e-05, {   2,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  1.047081260730e-05,  6.839125017010e-05, {   2,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  4.915823152496e-05, -2.963853701351e-05, {   4,   0,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -2.872174411079e-05, -5.001244190989e-05, {   4,   0,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.456581338459e-05,  6.416443736099e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  1.708863883093e-05,  6.162192053009e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -7,  10,   0,   0,   0 } },
 { -7.526137014132e-06,  7.118371023699e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.379458389628e-05,  4.490467214056e-05, {   1,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -6.975617449185e-05,  8.801499562466e-06, {   0,   0,   1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  6.454461882867e-05, -1.401093943982e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 { -5.430298920585e-05,  2.422810721533e-05, {   0,   0,   0,   0,   5, -14,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.539031733834e-05, -3.313893848000e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 { -5.899749596387e-05, -1.952592786749e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  5.232481825767e-05, -2.614152187596e-05, {   2,   0,  -3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.939380617473e-05, -4.906685494662e-05, {   2,  -2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -5.230773855516e-05, -2.613269190658e-05, {   1,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.854024328717e-05,  4.983847822202e-05, {   2,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.831917009952e-05, -2.657588996584e-08, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -5.462860301118e-05,  2.371618420408e-05, {   2,  -2,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -5.198149249789e-05,  2.636035169922e-05, {   1,  -1,  -1,   0,   0,  15, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.767140202976e-05, -1.067029996851e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 { -4.474495644534e-05, -3.359115112713e-05, {   0,   0,   1,   0,  -5,   0,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.568898638863e-05,  6.256330014587e-05, {   1,  -1,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.444719248847e-06, -7.578732175648e-05, {   4,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.765958923454e-05, -2.056738958609e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  2.846388757588e-06, -7.534459677640e-05, {   2,   0,  -2,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  6.616695932346e-05,  1.197625580220e-05, {   4,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -5.233652846952e-05,  2.580057658916e-05, {   4,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.619536926671e-06, -6.951416204838e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.584038467829e-05, -1.222785031238e-05, {   0,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.960892499077e-06, -7.509480387779e-05, {   2,   0,  -1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.340590044683e-05, -6.458657277841e-05, {   2,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -4.550888021451e-05, -3.244153447581e-05, {   1,   0,   0,   0,   0, -29,  38,  -7,   8,  -6,   0,   0,   0 } },
 {  5.087317876635e-05, -2.704791734734e-05, {   2,   0,  -2,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.190093957304e-05, -2.599681786973e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  7.773383416747e-05, -1.570002944761e-07, {   2,   0,   1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  7.731750445646e-05, -5.609670140943e-07, {   2,   0,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.376807401441e-05, -5.410284645986e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 {  1.334422775335e-05, -6.451348483432e-05, {   2,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.639047156544e-05, -6.139597895189e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.478873211984e-05, -5.298402230080e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 {  2.435175135720e-05, -5.339147819310e-05, {   1,  -1,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.508779317023e-05, -2.629422118762e-06, {   2,   2,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.207372728528e-05,  4.558700441642e-05, {   1,   0,   0,   0,   0, -23,  33, -16,   4,   5,   0,   0,   0 } },
 {  2.505702463006e-05,  5.255980780288e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 {  6.547245650798e-05, -1.211091650319e-05, {   1,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.320467715233e-05,  4.437535210874e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.817921911637e-06,  7.268505302009e-05, {   2,  -2,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -3.286577100346e-05, -4.453784335175e-05, {   2,   0,   0,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -2.874454764042e-05,  4.865269203279e-05, {   1,   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  6.725907941984e-05,  1.012284914810e-05, {   0,   0,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 { -2.846351576987e-05, -4.890992179512e-05, {   1,  -1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.167922659711e-05, -6.567809242391e-05, {   2,   0,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.157895230153e-05,  2.576125452373e-05, {   0,   4,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.157895230153e-05,  2.576125452373e-05, {   0,   4,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.712625692595e-05,  8.747777896975e-08, {   2,   2,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.753837722069e-05, -3.956169377064e-05, {   2,   0,  -2,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 {  3.430066565624e-05, -4.276735804286e-05, {   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -5.833296846000e-06, -7.116477374510e-05, {   2,   0,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.178722757349e-05, -5.519354423509e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.049138713082e-05,  2.642041681537e-05, {   0,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.821554037749e-05, -5.869031033743e-05, {   4,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.672025567844e-05,  6.017039214615e-05, {   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 { -1.311090049239e-05,  6.376703296610e-05, {   4,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.819671852098e-05, -5.863286132261e-05, {   4,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.900318692397e-05,  7.738638053668e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 {  4.806047608938e-05, -2.867145477114e-05, {   2,   0,  -1,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 {  6.036103216165e-05,  1.632843268109e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -7.970396179243e-06,  6.870650486289e-05, {   1,   0,   0,   0,   0,   0, -30,  33,   0,   0,   0,   0,   0 } },
 {  7.557082634756e-05,  1.092343682761e-06, {   2,  -2,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.929106471811e-05,  5.736429607342e-05, {   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.212418834030e-05,  6.452194938717e-05, {   0,   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  5.673210874958e-05,  1.990512349675e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -7.561447383352e-05,  1.009085722388e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   6,   0,   0,   0 } },
 {  2.915612423851e-05, -4.742573750798e-05, {   4,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  5.101171153432e-05, -2.547790387074e-05, {   0,   2,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.902242866453e-05,  7.400942088314e-06, {   2,   0,  -1,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 {  2.281794904614e-06,  7.411850884567e-05, {   1,   0,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -4.782247198412e-05, -2.849993141468e-05, {   2,   0,  -1,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 { -7.618187794050e-05, -2.383470187334e-08, {   6,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.426210313150e-05,  4.193624415299e-05, {   1,   1,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.615473501505e-05,  4.076537180737e-08, {   2,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.187068843216e-05,  4.214131706171e-06, {   6,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.158781669592e-05, -6.443608728290e-05, {   1,  -1,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.436064037540e-05, -6.165859837197e-05, {   2,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.238775896061e-05, -1.362890154040e-05, {   0,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.263815881484e-06,  6.971918394963e-05, {   5,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.263316981306e-06,  6.971922877078e-05, {   5,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.262988437111e-06,  6.971925828519e-05, {   3,  -1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.447481503361e-05,  5.149251320526e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 {  1.372453655653e-05,  6.221882562338e-05, {   3,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -7.290998906372e-05,  2.915019642292e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  7.423827126807e-05,  1.570841401710e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 { -7.002783398652e-05,  5.712870611315e-06, {   2,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.896214893437e-05,  1.677467737224e-05, {   2,   0,  -1,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 { -3.548139440435e-05,  4.024929762975e-05, {   2,   0,  -1,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 { -7.561130750852e-05, -1.343004695132e-08, {   4,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.199000703376e-05,  6.363312359847e-05, {   2,  -2,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.317947455443e-05, -6.244005558653e-05, {   2,   0,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.009563248405e-06,  6.752663221901e-05, {   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.189369644494e-06, -7.231779647635e-05, {   4,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -7.457614526056e-05,  8.935318647465e-07, {   1,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.801539833038e-05,  4.744352064723e-05, {   1,   1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.552162524931e-05, -5.983481858843e-05, {   0,   0,   4,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.293003681085e-05, -2.382817405248e-06, {   2,   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.447684484876e-05,  6.082646657374e-05, {   3,  -1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.548240859075e-05, -5.970065138480e-05, {   0,   0,   4,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.567933996136e-06, -7.256445921158e-05, {   2,   0,  -2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.407458029837e-05, -6.103045916548e-05, {   2,   1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.361653195743e-05,  1.482226048819e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 {  3.236513123512e-06, -7.184804063920e-05, {   2,   0,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.588820828041e-05,  5.918289334840e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.688853333329e-06,  7.024781632090e-05, {   2,  -2,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  3.276773692348e-05, -4.213193166701e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -3.611743155584e-05, -3.872514963778e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 {  6.116938291094e-05, -1.366347524748e-05, {   2,   2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.908252100300e-05,  4.574632162418e-05, {   2,   0,  -1,   0,   0,   0,   8, -12,   2,   0,   0,   0,   0 } },
 {  7.434933093282e-05,  4.754231883723e-07, {   0,   2,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.433816542835e-05,  5.032693590565e-05, {   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.199251889908e-06,  7.145919912242e-05, {   0,   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  4.233960230582e-06, -7.041695255266e-05, {   2,   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 { -5.626097502613e-07,  7.397954056581e-05, {   1,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.745150112019e-05, -2.708590216198e-05, {   1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.535987675436e-05,  1.915809674551e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 {  2.743540472719e-05,  4.703870025945e-05, {   4,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.114527231041e-05, -4.331747119691e-05, {   0,   0,   2,   0,   0, -36,  32,   0,   0,   0,   0,   0,   0 } },
 {  4.430452141448e-05,  3.011231682097e-05, {   2,   0,   0,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  5.513234121421e-05, -1.926001050718e-05, {   4,   0,  -2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.922579788097e-06, -7.038333201950e-05, {   0,   0,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.672930710513e-05,  1.757563946412e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.394534956662e-05,  5.035278684544e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0 } },
 {  5.342928915716e-06, -6.892889831675e-05, {   2,   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -4.672747371061e-05,  2.754092258134e-05, {   0,   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 { -2.370093474038e-05, -5.054265250599e-05, {   2,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.165311154614e-06,  7.304321219035e-05, {   2,   2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.677653352394e-05,  1.742756280632e-05, {   3,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.921859102048e-05,  4.870467092034e-06, {   0,   0,   1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.881203230240e-05,  5.525736296786e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 { -1.875619906412e-05,  5.526565415779e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -3.037815512024e-05, -4.360090109983e-05, {   0,   0,   0,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 {  3.340272653764e-05, -4.055337450974e-05, {   2,   2,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -7.386940200012e-07,  7.315219748546e-05, {   1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.097385208788e-05,  3.290174150741e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  1.943730352658e-05, -5.437749972684e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  5.191738426842e-05,  2.186466320845e-05, {   3,  -2,   0,   0,   0, -16,  19,  -7,   8,  -6,   0,   0,   0 } },
 {  1.799773437846e-05, -5.571691575609e-05, {   2,   0,  -2,   0,   0,   0,   5,  -6,   0,  -1,   0,   0,   0 } },
 { -2.707171976728e-05,  4.663101879986e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 {  3.125493483190e-05, -4.242616488047e-05, {   2,   0,   0,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 { -7.064215969236e-05,  3.036817992847e-06, {   0,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -4.909740755532e-05, -2.452236331954e-05, {   0,   2,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.383492428158e-05,  2.978184893427e-05, {   2,   0,   0,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -2.881159893234e-06,  7.071424399033e-05, {   2,  -4,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.352003151430e-05,  5.124738040001e-08, {   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.154887445831e-05, -1.202238716009e-05, {   4,   0,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  7.335836133961e-05,  1.615155295602e-07, {   4,   2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.264186600541e-05,  8.747891262938e-07, {   6,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.328135497295e-05,  1.471118092246e-07, {   0,   0,   2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  6.403171626090e-05, -9.346059339852e-06, {   0,   0,   0,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 { -6.067620452418e-06,  6.728141723312e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.402347275717e-05, -4.929529400898e-05, {   2,   0,  -2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  7.288113396469e-05,  4.303971157061e-07, {   0,   2,  -1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.778728802274e-05, -4.552092910626e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  7.237184969708e-05, -9.157357586780e-07, {   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.836506663662e-05,  4.489802042513e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -6.125007283572e-05, -1.193643717878e-05, {   1,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.686649008578e-05, -2.631059190452e-05, {   2,   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -4.947251685142e-05, -2.369003283444e-05, {   2,   0,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.297674753058e-05,  1.745314471020e-07, {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 {  2.192097972357e-05,  5.122194959277e-05, {   1,  -3,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  7.271096170592e-05, -4.296965270263e-07, {   2,   0,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.820358127127e-05,  5.488148744802e-05, {   2,   0,   0,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 {  4.912667513347e-05,  2.392786933977e-05, {   2,   0,  -1,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.042501968520e-06, -6.500825188382e-05, {   0,   0,   1,   0,   0,   0, -26,  24,   0,   0,   0,   0,   0 } },
 { -6.192970184323e-05, -1.111189706012e-05, {   2,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.890619826036e-05,  4.100711438879e-06, {   2,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -7.292717321552e-05, -5.889091910191e-08, {   3,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.760682824350e-05, -5.357967002301e-06, {   2,   0,  -1,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.550738176745e-05, -1.743432872386e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 {  4.260018086277e-05,  3.030071529407e-05, {   2,   0,  -1,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 {  2.805470568638e-05, -4.483117371186e-05, {   2,   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.383022549452e-05,  5.903564332499e-05, {   1,   1,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.369244320801e-05, -5.915969376539e-05, {   1,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.536881665579e-05,  7.458145331021e-06, {   0,   0,   1,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 { -3.938273234456e-05, -3.340690247582e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0 } },
 {  5.133359761835e-05,  2.145157856092e-05, {   4,   0,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.979001193538e-05,  4.299401438028e-05, {   0,   0,   2,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -3.756984731243e-05,  3.520956875396e-05, {   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  4.715874979038e-05, -2.559694491539e-05, {   2,   0,  -1,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  2.284736320869e-06, -7.046843221953e-05, {   2,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.227950910928e-05, -4.462366021938e-07, {   0,   0,   2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  7.268435493335e-05,  1.675980692894e-08, {   0,   0,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.537911270847e-05,  2.728459684804e-05, {   2,   0,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  6.638824236802e-05,  6.223815602556e-06, {   2,   0,   0,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  4.685919550656e-05,  2.572798942296e-05, {   1,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.133406583921e-05,  5.123186241512e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.469735589643e-05,  5.786437300267e-05, {   1,  -1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.200072867650e-06,  6.926493662249e-05, {   2,   0,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.233541863912e-05,  4.010343189749e-05, {   0,   0,   0,   0,   0,  15, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.859996121110e-05,  5.381682951914e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -7.408102909005e-06, -6.499434716714e-05, {   1,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.602061209798e-05,  5.635781134880e-05, {   0,   0,   0,   0,   0,   3,  -3,  -4,   3,   0,   0,   0,   0 } },
 { -6.938613730372e-05, -2.780941794374e-06, {   0,   0,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.652850580563e-05,  5.562636754821e-05, {   2,   0,  -1,   0,   0,   3,   1,  -8,   3,   0,   0,   0,   0 } },
 {  2.949731852517e-05,  4.264983304031e-05, {   0,   0,   2,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -3.810267755880e-05, -3.378911509669e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 {  5.221568345793e-05, -1.963903343890e-05, {   4,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.163717331824e-05, -6.018663923851e-05, {   3,  -1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.732087659494e-06, -6.908864815779e-05, {   0,   0,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.430063264408e-05, -5.751276887749e-05, {   0,   0,   1,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 {  4.952728329332e-05,  2.226454031371e-05, {   1,  -1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.719928019326e-05, -5.458017007333e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.371986194524e-05,  5.805305131152e-05, {   1,  -1,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.728855292762e-05,  5.446902604814e-05, {   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.448739120727e-05, -7.249764342867e-06, {   1,  -1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -3.383166072012e-05,  3.789962042238e-05, {   2,   0,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 {  1.353712604928e-05,  5.819221845044e-05, {   1,   1,  -1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.124785899090e-05, -1.041794378420e-05, {   1,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.864115211543e-05,  2.301642194198e-05, {   2,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.098297337966e-05, -3.067219291424e-05, {   3,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  6.566651048857e-06,  6.508472223586e-05, {   1,   0,   0,   0,   0, -21,  26,  -8,   0,   0,   0,   0,   0 } },
 {  3.244559432748e-05,  3.916933467909e-05, {   0,   1,   0,   0,   0, -17,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.917242131275e-05, -1.243451832652e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 {  4.960706784126e-05, -2.194579778483e-05, {   1,  -1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  5.567853290102e-05,  1.583430419320e-05, {   2,   0,  -1,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 {  4.453385789132e-05, -2.695054628285e-05, {   2,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  3.893199139999e-05, -3.254714612724e-05, {   2,   0,   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 { -1.361596321947e-05,  5.785898327663e-05, {   1,  -1,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.100441381592e-05, -4.653952855093e-07, {   3,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.204000864536e-06, -6.923418298237e-05, {   2,   0,  -1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -6.622877300106e-05,  5.209285748911e-06, {   1,  -1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -5.713340559292e-05, -1.428920568297e-05, {   1,  -1,   2,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.187902042395e-05,  3.951767277690e-05, {   1,   0,   0,   0,   0,   0, -27,  27,   3,   0,   0,   0,   0 } },
 {  5.661948962639e-05,  1.475114140521e-05, {   1,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  8.373198507574e-06,  6.299592500782e-05, {   2,   0,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -5.383236843451e-05,  1.740775955803e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.754688036335e-05,  2.368811301912e-05, {   1,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.094994030824e-05, -2.472834330121e-07, {   1,   0,   0,   0,   0,   0, -33,  39,   0,   0,   0,   0,   0 } },
 {  6.509300007701e-05, -6.098601461228e-06, {   1,   0,   0,   0,   0,  -2, -31,  42,  -8,   6,   0,   0,   0 } },
 {  1.621603494627e-05,  5.496090745898e-05, {   2,   0,  -1,   0,   0,   3,  -7,   8,  -3,   0,   0,   0,   0 } },
 { -4.078617053330e-05,  3.036992693487e-05, {   2,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0 } },
 { -6.398413016792e-07, -7.045682065260e-05, {   0,   2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.407723466234e-05,  5.698061553440e-05, {   1,   1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.772680778896e-06, -6.321930611107e-05, {   2,   0,  -1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.154110553397e-05, -3.940525367937e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -4.577203385446e-05, -2.517215738799e-05, {   0,   0,   3,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -7.093813017137e-05, -5.639888229054e-09, {   1,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -5.640687261008e-05, -1.448718355422e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -3.329594894558e-05, -3.755875481325e-05, {   0,   0,   2,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.018090334275e-05,  2.057137357486e-05, {   1,  -1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -5.880642465310e-05,  1.190816608548e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.176956759088e-06, -6.648596330912e-05, {   4,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.302855849468e-06, -6.827130241275e-05, {   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.644407435663e-05,  5.408257677431e-05, {   0,   0,   2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  7.003036094122e-05,  4.725315129933e-07, {   3,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.848729201019e-05, -1.199859888774e-05, {   1,  -1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.237464036504e-07,  6.981522604304e-05, {   3,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.219665799185e-05, -4.823496819533e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.323426916399e-05, -5.717074270008e-05, {   1,   1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.937375029700e-05,  4.091838966221e-05, {   0,   0,   1,   0,   0, -26,  29,   0,  -2,   5,   0,   0,   0 } },
 { -4.268249533027e-05,  2.758146707192e-05, {   0,   0,   0,   0,   2,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.658735352332e-05,  5.355145530180e-05, {   0,   0,   0,   0,   0,   0,   6, -10,   3,   0,   0,   0,   0 } },
 { -5.699482035355e-05, -1.313241046200e-05, {   0,   0,   0,   0,   0,   3,  -5,   0,   1,   1,   0,   0,   0 } },
 {  2.201736200765e-05,  4.810503801697e-05, {   0,   0,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -6.449443682715e-05,  5.600514959079e-06, {   1,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.936128322838e-05,  2.068769516969e-05, {   4,   0,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -6.999999998490e-05, -1.453859266817e-09, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -6.999999998515e-05,  1.441641962045e-09, {   0,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.999999998802e-05,  1.295034304906e-09, {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  6.999999999136e-05, -1.099557428713e-09, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.999999999230e-05,  1.038470904899e-09, {   1,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.999999999478e-05,  8.552113334559e-10, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -6.745998003557e-05, -2.540802932634e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.392236280184e-05, -4.606324949262e-05, {   0,   0,   0,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  6.991181872586e-05,  6.709395415331e-08, {   2,  -2,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.876425297181e-05,  1.190370889375e-06, {   0,   0,   2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  4.335641689664e-05, -2.657714294387e-05, {   2,   0,   0,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 {  5.351371880376e-05,  1.641507004301e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 {  2.308231307623e-05,  4.681512811554e-05, {   2,   0,   0,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 {  4.113322842654e-06, -6.576036101644e-05, {   2,   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  6.964676009601e-05,  1.957553179006e-07, {   2,  -2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  5.815627911754e-05, -1.168514280438e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 { -2.953759609759e-05, -4.027074322414e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 { -1.761399714470e-05, -5.217719026096e-05, {   2,   0,  -2,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.966790255426e-05,  5.997871468887e-08, {   4,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.940599948159e-05, -1.031159597687e-05, {   0,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -6.961455116795e-05,  9.640210540563e-08, {   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.935342306483e-05,  1.035373712383e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  6.963471528670e-05, -1.934328990638e-08, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -2.020428215233e-05, -4.943777959503e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 {  5.915934030794e-05,  1.045256687154e-05, {   6,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.323607107692e-05, -5.632171843897e-05, {   1,  -1,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.866967802614e-05, -5.087719340467e-05, {   1,   1,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.049576064800e-05,  5.901372277338e-05, {   2,   0,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.321685262810e-05, -5.625728097910e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  1.337876901774e-05, -5.608044558173e-05, {   0,   0,   3,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.013852123988e-05,  4.929037175519e-05, {   2,   0,  -3,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.828586790215e-05,  1.116239174319e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -2.289778998252e-05,  4.648640733826e-05, {   2,   0,   0,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 {  4.645376933357e-05, -2.290346714488e-05, {   2,   2,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.599466916409e-05, -5.334666363478e-05, {   6,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.597691331062e-05, -5.329518059108e-05, {   6,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -5.242021843335e-05, -1.684902590584e-05, {   1,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.842333731325e-05, -1.082794421977e-05, {   1,   1,   0,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 { -5.842333731325e-05, -1.082794421977e-05, {   1,   1,  -2,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  6.910271694506e-05, -1.315643071536e-07, {   4,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.920209940193e-05,  3.162390227240e-08, {   0,   0,   2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.804959590764e-05,  1.164204239637e-06, {   0,   0,   2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -6.708859379719e-05,  2.015929772526e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  2.927670457988e-05,  3.982661457620e-05, {   4,   0,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  3.989795211872e-05, -2.919482232209e-05, {   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.656140819109e-06,  6.440649054358e-05, {   0,   0,   2,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -3.079952810357e-05,  3.820782183021e-05, {   4,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.676739680694e-06,  6.631818419170e-05, {   4,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.777238447426e-05, -5.121602442607e-05, {   0,   0,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.457099038485e-05, -5.436363444446e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.698805992075e-05, -4.177406187997e-05, {   1,   1,  -2,   0,   0,   2,  -2,  -1,   0,   0,   0,   0,   0 } },
 { -2.866401988545e-05,  4.008797503173e-05, {   1,  -1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -4.821505741573e-05,  2.048531578795e-05, {   0,   2,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.719506834287e-05,  1.150006068538e-05, {   1,  -1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.927682469096e-05, -3.939445236608e-05, {   1,   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.661297558379e-05, -5.195654099600e-05, {   4,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.423580154946e-06, -6.713949454528e-05, {   2,   0,   0,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.590518047238e-05, -2.263199095916e-05, {   2,  -2,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.535823559886e-05, -4.316144385317e-05, {   1,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.298010068341e-05,  5.553837971265e-05, {   1,   1,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -5.443171217878e-05, -1.408341014987e-05, {   1,  -1,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  6.563257117330e-05,  2.861970422599e-06, {   1,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.842376604724e-05,  5.328003949246e-08, {   2,  -2,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.739318304324e-05,  1.031355313251e-06, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -6.054940090661e-05, -7.861163400805e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  5.543301865200e-05,  1.296824037880e-05, {   1,   1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.793238025578e-05,  5.046158363461e-05, {   1,   1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.261344619480e-05,  5.572575811116e-05, {   1,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -4.066002541221e-05, -2.765449454670e-05, {   2,   0,  -2,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  6.812451096275e-05, -1.655033677500e-07, {   2,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  6.162324543627e-05,  6.661481507450e-06, {   2,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.493035028469e-05,  4.332528595711e-05, {   0,   0,   0,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 { -1.358201951354e-05, -5.464197656500e-05, {   1,   1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.387318321731e-05, -5.433405996362e-05, {   1,  -1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  6.134144373954e-05,  6.840723381730e-06, {   2,   0,  -1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.781576725516e-05,  2.030854232851e-05, {   0,   2,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.865465640239e-06, -6.624517683542e-05, {   2,   2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.028074243326e-05, -5.782035920202e-05, {   2,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -6.805815229195e-05, -3.287694036976e-08, {   2,   0,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -6.746513382304e-05,  6.186418359249e-07, {   2,   0,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.404967026001e-05, -1.402510797079e-05, {   2,   0,  -1,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 {  4.706499528497e-05,  2.097433421402e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 {  6.795035650908e-05,  6.948234425472e-08, {   1,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.691268426635e-05,  2.110120043223e-05, {   1,  -1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -6.641095660031e-05,  1.588095898149e-06, {   0,   0,   0,   0,   0,   4,  -6,   0,   2,   0,   0,   0,   0 } },
 { -2.977732294442e-05,  3.820890661806e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 {  1.131503136661e-05,  5.667027174729e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  2.476150906098e-06, -6.550119785117e-05, {   2,  -2,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  6.489734699238e-05, -3.060122990949e-06, {   2,   0,  -2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.434913915150e-05,  5.359894773002e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 { -5.543948797439e-05, -1.249810994147e-05, {   1,  -1,   0,   0,   0,  -3,   7,  -2,   0,   0,   0,   0,   0 } },
 { -1.045691416669e-05,  5.747905228028e-05, {   2,   0,   0,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 { -6.788858693648e-05,  4.722593805604e-08, {   2,   0,  -2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.989108983063e-05, -2.803212028466e-05, {   2,   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  1.136237638524e-05, -5.654654744114e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   2,   0,   0,   0 } },
 {  4.046659376712e-05, -2.743676011275e-05, {   2,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.804833510451e-05,  1.984911700532e-05, {   1,  -1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -4.512479281808e-05, -2.270891004423e-05, {   1,  -1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.673223881640e-06, -6.512602172789e-05, {   2,  -2,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.676046746550e-05,  1.032259164145e-06, {   1,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.281483272010e-05,  5.495399537597e-05, {   1,   1,  -1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.900782132863e-05, -3.875804150041e-05, {   1,   0,   1,   0,   0, -40,  41,  -3,   0,   0,   0,   0,   0 } },
 { -4.067110455653e-05, -2.702135366042e-05, {   2,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -5.825240994616e-05, -9.438949448894e-06, {   0,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -2.950741411049e-05,  3.810081138314e-05, {   2,   0,  -1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -3.975635865832e-05,  2.783580021802e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  5.702089272261e-05,  1.056018128174e-05, {   3,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.720939994997e-05,  3.706255922436e-07, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 {  2.575470729552e-05,  4.176262886478e-05, {   1,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  4.196492523263e-05,  2.555108612971e-05, {   2,   0,  -2,   0,   0,  -1,   5,  -3,   0,   0,   0,   0,   0 } },
 {  5.486511775785e-05,  1.261084725127e-05, {   1,  -1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.283306624550e-05, -5.462496159709e-05, {   1,  -1,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.370673179412e-05, -1.367930533418e-05, {   2,   0,  -1,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 { -1.093385754006e-05,  5.644661935492e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,  -1,   0,   0,   0 } },
 { -6.697565075921e-05,  4.045298966810e-07, {   3,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.091327493804e-05, -3.645594892521e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  5.097619947428e-05,  1.637935406672e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -3.863298401638e-05, -2.860243169967e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 {  2.015969570223e-05,  4.702952778587e-05, {   1,  -3,   2,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.767713446731e-05, -4.950043974953e-05, {   1,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.756828761706e-05,  9.583838467148e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -2,   0 } },
 { -4.428916423321e-05, -2.282185875964e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 {  1.968679773875e-05,  4.730671073255e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 { -5.908013438564e-05,  7.909638274442e-06, {   0,   0,   1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 { -4.196635128707e-05, -2.501602497527e-05, {   0,   0,   3,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.603391761130e-08, -6.684020036190e-05, {   1,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.296638200116e-05, -2.387846733234e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -5.362684856572e-05, -1.321163199654e-05, {   4,   0,  -2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.461801016868e-05, -2.215880494652e-05, {   1,   1,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.979162513973e-05, -4.698232072614e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.786166309139e-06,  6.393166296253e-05, {   4,  -2,  -3,   0,   0,  29, -30,   0,   0,   0,   0,   0,   0 } },
 { -3.005537537006e-05, -3.665695915158e-05, {   1,   1,  -1,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.679391230745e-05, -4.987328359788e-05, {   2,   0,   0,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 {  6.634199040213e-05,  3.053103844870e-07, {   4,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.299157620817e-05,  1.364780001068e-05, {   0,   0,   2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.359777274704e-06,  5.826374148542e-05, {   2,   0,   1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -2.463077197471e-05, -4.192934089378e-05, {   2,  -2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -2.648855112950e-06, -6.390543375046e-05, {   0,   0,   1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.417197587759e-05, -1.234559301111e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  6.638697145661e-05, -7.621148216533e-08, {   4,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.290364693104e-05, -3.547470945618e-06, {   4,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.645275268494e-05,  9.950416268313e-06, {   6,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.305673649227e-05, -5.331855864445e-05, {   2,   0,  -1,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 {  1.261704567820e-05,  5.374973670457e-05, {   1,   1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.057162239645e-05, -5.784300411315e-06, {   2,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  6.097085938548e-05,  5.186455508528e-06, {   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.278885406178e-05,  5.328730551427e-05, {   3,  -1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -2.282249923276e-05, -4.324644373152e-05, {   1,  -1,   1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.327136582008e-05,  5.277183852033e-05, {   1,   1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.814946683404e-05, -3.785815277982e-05, {   2,   0,  -1,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 { -4.598680639161e-05,  2.000378431884e-05, {   2,   0,  -1,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 {  5.670712766661e-05, -9.280339266872e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  2.630338653384e-05,  3.966649239019e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  6.593083050695e-05, -1.842665286988e-08, {   2,   0,   3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.106661005748e-05,  3.484203815101e-05, {   1,   0,   0,   0,   0,   0,  33, -41,   2,   0,   0,   0,   0 } },
 {  6.275768808105e-05, -3.137078338650e-06, {   0,   2,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -6.275768808105e-05, -3.137078338650e-06, {   0,   2,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -5.630810072877e-05,  9.584492467421e-06, {   3,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.419500941224e-06, -5.747119620489e-05, {   2,  -1,  -1,   0,   0,   3,  -4,   2,   0,   0,   0,   0,   0 } },
 { -3.446411386110e-05, -3.135792335760e-05, {   2,  -2,   0,   0,   0,   0,  17, -28,   0,   0,   0,   0,   0 } },
 {  6.479808814987e-05,  9.965640987603e-07, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  1.699082587071e-05,  4.875225773585e-05, {   0,   0,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  5.361037156026e-05, -1.207812923222e-05, {   1,   1,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  6.449900571022e-05,  1.174438790029e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 { -1.585814792390e-05, -4.981421788246e-05, {   2,   0,  -1,   0,   0, -20,  21,   0,   2,  -5,   0,   0,   0 } },
 {  3.944749971763e-05,  2.621514494612e-05, {   0,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 {  6.550029999812e-05,  4.899408566000e-08, {   0,   0,   1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  6.293358642133e-05, -2.579648123741e-06, {   0,   0,   1,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 {  5.359827516973e-05,  1.190740914766e-05, {   5,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  4.366088761344e-05,  2.176181345296e-05, {   0,   0,   1,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.533660505136e-05,  6.283870799295e-08, {   2,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.684803837314e-05,  8.517179367232e-06, {   1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.435551084891e-05, -1.099379888829e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 {  6.202144371426e-05, -3.325820243894e-06, {   2,   0,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.425544848490e-05, -1.104207959040e-05, {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  5.067384131685e-06, -6.018116020164e-05, {   2,   0,  -2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -5.556390023477e-05,  9.637600288156e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.243556326330e-05, -4.274718202896e-05, {   2,   0,  -1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.264414872817e-06,  6.187739870118e-05, {   2,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.369527399258e-06, -5.975925179843e-05, {   5,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.369381379844e-06,  5.975926491849e-05, {   1,  -1,   2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.367535274484e-06, -5.975943076266e-05, {   1,   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.367326674686e-06,  5.975944949853e-05, {   3,   1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.367326674686e-06,  5.975944949853e-05, {   3,   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.365991634431e-06, -5.975956939085e-05, {   3,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.088487268973e-05, -4.423676315288e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  4.220419197287e-05,  2.290826549165e-05, {   2,   0,  -2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  3.123194434624e-05, -3.387558112569e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -5.466573704120e-05, -1.034584986636e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  1.996934008047e-05, -4.502619627493e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 {  1.288887932327e-05, -5.209455649814e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 { -1.748750136298e-05, -4.746910571909e-05, {   1,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.777885939200e-05, -4.712889462198e-05, {   1,  -1,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -5.622077865326e-05, -8.673828423266e-06, {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.412063994837e-05, -5.074459983548e-05, {   2,   0,  -3,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.818328099298e-05, -4.662407134557e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.258684777339e-05,  5.220896265832e-05, {   0,   0,   2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.846742565455e-07, -6.398130849350e-05, {   4,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.464834996603e-05, -1.017244269624e-07, {   0,   2,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -5.015128860178e-05,  1.459818072825e-05, {   1,  -1,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.867080609920e-05, -3.604772078441e-05, {   0,   0,   2,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.409929959583e-05, -5.059759625824e-05, {   2,   0,  -3,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -4.382005550341e-05, -2.086059032541e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.386990742710e-05,  5.076330803572e-05, {   0,   0,   3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.057999954900e-06, -5.856281204390e-05, {   2,   2,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.315172352700e-06, -5.830372091230e-05, {   2,  -2,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.198412396071e-07, -6.436799907094e-05, {   4,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.319101676875e-05,  2.139681591330e-05, {   4,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.668385427517e-05,  4.789790236799e-05, {   2,   0,  -1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 {  4.938818516069e-05,  1.513841964517e-05, {   4,   0,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.589316514005e-05,  8.622405460271e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  2.632790895395e-06, -6.187504768415e-05, {   2,   0,  -1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.135496817459e-05,  2.314053097308e-05, {   4,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -4.994420769134e-05,  1.453281535010e-05, {   1,  -1,   3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  5.160928046178e-05, -1.285657736640e-05, {   1,   1,  -1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.415412548857e-05,  1.030713199061e-05, {   3,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.586818448484e-05, -8.571954418003e-06, {   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -4.705232232163e-05, -1.738717060015e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 { -4.545591733281e-05, -1.897661666398e-05, {   2,   0,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -3.315735809623e-06,  6.111007959952e-05, {   2,   0,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.231484245444e-05, -2.097490217738e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 {  1.257143916810e-05,  5.182863101143e-05, {   3,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.093992816367e-05,  4.338321195290e-05, {   2,   0,  -1,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 {  3.164157420136e-05,  3.265287349016e-05, {   0,   0,   0,   0,   0,   0,   6, -12,   3,   0,   0,   0,   0 } },
 {  4.858494672667e-05,  1.569492470717e-05, {   2,   0,   0,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 {  5.251907301423e-05, -1.172200457285e-05, {   2,  -2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.451987976628e-05,  9.665497599989e-06, {   0,   2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.827339469528e-05,  1.589778258695e-05, {   1,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.448915272079e-05,  3.967762954782e-05, {   0,   0,   1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  9.832360482424e-06, -5.431628899593e-05, {   4,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.957867935290e-05,  3.456872388383e-05, {   2,   0,  -1,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  1.462198337867e-05,  4.951968297173e-05, {   1,  -1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.501627248022e-06,  6.262150046061e-05, {   2,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  4.032416796784e-05, -2.378368902953e-05, {   4,   0,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -3.262475349221e-05,  3.145604190053e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 {  1.585429233063e-05,  4.818885510731e-05, {   4,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.396607021855e-05,  1.006202976082e-05, {   4,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.841282337849e-05,  1.561235410233e-05, {   2,   0,   0,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 { -4.690949012249e-05,  1.711450150733e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -3.792565752244e-06, -6.023016363700e-05, {   0,   0,   2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.772287004612e-06, -5.924769629735e-05, {   1,   0,   0,   0,   0, -23,  17,  16,  -4,  -5,   0,   0,   0 } },
 {  2.458627853777e-05, -3.942407839613e-05, {   2,  -2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  6.393720194461e-05, -6.867990905087e-08, {   0,   2,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.371923778847e-05, -5.028223032960e-05, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0 } },
 {  6.386606051743e-05,  9.882016672272e-08, {   0,   0,   2,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.858710456269e-05,  2.537492304538e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 {  3.954013468746e-06,  5.998951337003e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.220660306275e-05, -5.173314318076e-05, {   1,  -1,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.245529239052e-05, -1.447242767212e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 { -4.864306307252e-05, -1.523688725736e-05, {   2,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.244307612251e-05,  5.142338113007e-05, {   2,   2,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  9.405693700172e-06,  5.443573186458e-05, {   1,   0,   2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.262863674012e-05, -2.118599811326e-05, {   3,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.912913291071e-05,  4.465644498675e-05, {   4,   0,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.148005681698e-06, -6.162762668088e-05, {   2,   0,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  6.346024767125e-05, -3.147598481523e-07, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 { -6.345958260557e-05, -3.147585481948e-07, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 {  5.672639092882e-05,  7.042893970431e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 {  5.397981182950e-05,  9.746342905472e-06, {   0,   2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.326069826987e-06,  6.230238313735e-05, {   0,   0,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  4.084292345666e-05, -2.276986697567e-05, {   2,   0,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -6.186772199389e-05,  1.734027210207e-06, {   2,   0,  -1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  9.353386879787e-06,  5.421457386018e-05, {   1,   0,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.409587336041e-05,  3.946500741553e-05, {   2,   0,  -1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 { -6.038031926340e-05,  3.179330325461e-06, {   1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  7.048245297979e-07,  6.283186072863e-05, {   4,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.442942982193e-05, -2.909028502693e-05, {   2,   0,  -2,   0,   1,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.988309764034e-05, -1.357546436931e-05, {   3,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.303103748084e-05, -4.147757190360e-07, {   2,   2,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.564847486481e-05, -7.767234615478e-06, {   2,   0,  -1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  1.069509532449e-05, -5.272000805645e-05, {   4,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.463689824614e-05, -4.872312135488e-05, {   2,   0,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.301666150379e-05,  1.024347331032e-05, {   1,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.179938022735e-05,  2.145889539487e-05, {   1,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.536156290980e-05,  1.789545412540e-05, {   2,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.234742052452e-05,  5.086672603560e-05, {   1,   1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.420467769189e-05, -8.981477434869e-06, {   0,   0,   0,   0,   0,   0,   9, -17,   0,   1,   0,   0,   0 } },
 { -3.999840607439e-05, -2.318078267567e-05, {   0,   0,   3,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.121418321938e-05, -4.196453722376e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 { -6.193687026754e-05,  1.226019663893e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -2.203232020956e-06, -6.092101554345e-05, {   0,   0,   1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.659379508223e-06, -6.043145093123e-05, {   4,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.142700127759e-05,  1.663495599794e-06, {   0,   0,   0,   0,   0,   1,  -2,   0,   1,   0,   0,   0,   0 } },
 {  5.514276901443e-05,  7.924069213729e-06, {   1,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.359471937071e-05, -3.946766136338e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0 } },
 { -6.258830352150e-05,  4.431026408822e-07, {   3,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.258144668313e-05, -4.349574110318e-07, {   2,  -2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.671908855174e-06,  5.333060537536e-05, {   2,   0,  -1,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 {  5.360882541249e-05, -9.332551946332e-06, {   0,   0,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.935219661405e-05,  1.355955397107e-05, {   3,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.757930360761e-05, -4.520910519581e-05, {   0,   2,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.299525345452e-05, -9.761944061797e-06, {   0,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.477073389830e-05,  4.798615040916e-05, {   3,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.926334603342e-05,  2.345665429815e-05, {   1,   1,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.718702545720e-05,  2.548978963471e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   6,   0,   0,   0 } },
 {  4.974929856515e-05, -1.290314000582e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 {  1.392067435727e-05, -4.869851442655e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -5.987493614820e-05, -2.733905833380e-06, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.101372303130e-05, -1.159262729940e-05, {   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.726904797102e-05,  4.524997057019e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -2.072113841627e-05, -4.178239776374e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 {  1.470563151287e-05,  4.779536505092e-05, {   3,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.234026754879e-05, -8.711042300213e-08, {   0,   2,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -6.106275631079e-05,  1.337933096618e-06, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  4.097195966335e-05,  2.135803796055e-05, {   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -5.993257633235e-05,  2.333221012714e-06, {   2,   0,  -1,   0,   0, -28,  34,   0,   0,   0,   0,   0,   0 } },
 { -5.327846802128e-05,  8.935957954074e-06, {   1,   1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.385850405222e-05, -2.834500428709e-05, {   4,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.096990921723e-05,  4.121858944332e-05, {   1,   0,   0,   0,   0, -25,  25,   7,  -8,   6,   0,   0,   0 } },
 {  5.040372201334e-05, -1.176225434987e-05, {   0,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 { -1.700775031202e-05,  4.514743355191e-05, {   0,   0,   1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.192279106299e-05, -5.022434071251e-05, {   1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -5.037186448961e-05, -1.175798892070e-05, {   0,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  1.247037535176e-05, -4.963762688649e-05, {   2,   0,   0,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 { -1.605463999776e-05, -4.604354925569e-05, {   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.181091831534e-05,  5.027276655826e-05, {   1,  -1,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.201757169576e-06, -5.885609464941e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  3.353944732380e-05,  2.848120995144e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  4.136556477281e-05,  2.064826268273e-05, {   2,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.455243070323e-05, -4.744168375955e-05, {   0,   0,   1,   0,   0,   0, -51,  71,   0,   0,   0,   0,   0 } },
 {  3.019992769773e-05,  3.178680036614e-05, {   2,   0,  -1,   0,   0,   0,  10, -16,   3,   0,   0,   0,   0 } },
 { -5.143038459287e-05, -1.051096574659e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0 } },
 { -3.348960225386e-05,  2.843268413027e-05, {   2,   0,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 {  3.973462414995e-05, -2.216755517155e-05, {   1,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -4.613716295825e-05,  1.572706727259e-05, {   1,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.769767221256e-05, -2.414779974692e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 { -4.950074786377e-05, -1.231958190982e-05, {   1,  -1,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.790463680783e-05, -3.387859861248e-05, {   2,   2,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.101784657886e-06,  6.064063358075e-05, {   0,   2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.155799571183e-05, -3.011429318355e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 { -5.425098157939e-05,  7.359892481743e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  1.183490430261e-05, -4.974032132641e-05, {   1,   1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.587531890999e-05, -5.682007040465e-06, {   2,   0,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 {  1.263951017881e-05,  4.884575077936e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0 } },
 {  1.194356311960e-05, -4.954041745578e-05, {   1,  -1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.095515396222e-05, -2.052135321032e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -4.565570110120e-05, -1.581356432856e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 { -5.333056288591e-05,  8.137466518049e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 {  1.338717826342e-05, -4.807721756040e-05, {   2,   0,   2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.338382206487e-05, -4.807480493638e-05, {   2,   0,   2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -6.237602824700e-06, -5.520233820565e-05, {   2,   0,  -1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.325101670044e-05,  3.818829898422e-05, {   0,   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -1.727123943547e-05,  4.415412737598e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -1.369389445279e-05,  4.772005606280e-05, {   2,   0,  -1,   0,   0,   0,  -1,   5,   0,   0,   0,   0,   0 } },
 {  2.693353884681e-05,  3.446610370060e-05, {   1,  -1,  -1,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.935276746270e-05, -2.200791130915e-05, {   2,   0,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  5.160099234242e-05,  9.758117525419e-06, {   1,   1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.092091502834e-05,  2.043802358890e-05, {   4,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.060315791843e-05, -4.075546107043e-05, {   0,   0,   0,   0,   0,   6, -14,   8,   0,   0,   0,   0,   0 } },
 {  2.808721203873e-05, -3.320703009786e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -2.631324530183e-05, -3.497323124142e-05, {   2,   0,  -1,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 { -3.955306973897e-05, -2.172755814971e-05, {   2,   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 { -3.119356405642e-05, -3.004912346295e-05, {   4,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  6.074235456299e-05,  4.836109820791e-07, {   4,   0,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.380030556054e-06, -5.979443151830e-05, {   2,   0,   0,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  8.819166734500e-06,  5.234033426600e-05, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.647093042810e-05,  4.467797725132e-05, {   2,   0,  -1,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 { -5.249871187810e-05, -8.637196247731e-06, {   2,  -2,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.056703646045e-05,  1.054676971183e-05, {   1,   1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.350648515361e-05,  4.760721467046e-05, {   0,   0,   0,   0,   0,   6, -10,   0,   5,  -5,   0,   0,   0 } },
 { -9.562294604083e-06, -5.153779507329e-05, {   1,  -1,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.371695303283e-05, -3.736944663735e-05, {   4,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -8.151994112998e-06, -5.290538141519e-05, {   0,   0,   0,   0,   0,   0,  12, -23,   3,   0,   0,   0,   0 } },
 {  3.360143025380e-06, -5.767790500377e-05, {   2,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.294311518460e-05,  8.041547468969e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 {  2.665052398831e-05, -3.424112183447e-05, {   2,   0,   0,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  3.351490517393e-06, -5.748116440294e-05, {   2,   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  6.037794978512e-05, -4.386257557982e-07, {   4,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.013768089988e-05, -6.605763025201e-07, {   2,   0,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.060643965462e-05, -5.015456131414e-05, {   2,   0,   0,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -6.857215872586e-06,  5.388688967069e-05, {   0,   0,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.060628157610e-05,  1.169829434871e-07, {   2,   0,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.437485902826e-05, -3.631507859424e-05, {   0,   0,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.877701580189e-05,  4.190378476473e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 { -5.797503918890e-05,  2.691339852557e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -2.064444250887e-05, -4.001816758652e-05, {   2,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.452293998060e-06, -5.518224439020e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  5.999709288982e-05,  5.906308235917e-07, {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.469640839194e-05, -4.588749575675e-05, {   0,   0,   1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.901427688882e-05, -3.155268330368e-05, {   2,   0,   0,   0,   0, -11,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.172151580786e-06, -5.733400783722e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.647065154267e-05, -4.401870986111e-05, {   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.535575027447e-05, -3.511761533764e-05, {   0,   0,   0,   0,   0,   0,   8, -16,   2,  10,   0,   0,   0 } },
 { -2.341979000781e-05,  3.704440278005e-05, {   3,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.006801172004e-05,  4.038893458565e-05, {   2,   0,  -1,   0,   0,  -9,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.924190509451e-05,  2.121292105005e-05, {   0,   0,   1,   0,   0, -28,  32,   0,   3,   0,   0,   0,   0 } },
 { -4.846673797128e-05, -1.192482283607e-05, {   2,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.445944086358e-05, -5.868472227916e-06, {   2,  -2,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.254764916908e-05, -1.777902525335e-05, {   2,   0,  -1,   0,   0, -20,  17,   8,  -3,   0,   0,   0,   0 } },
 {  2.187358986201e-06, -5.812453043590e-05, {   0,   0,   1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.921298959399e-05,  2.109821521775e-05, {   2,  -2,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -4.083206740434e-05, -1.947386035286e-05, {   2,   0,   0,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.284026297671e-05, -3.742385912163e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.153439620985e-05,  4.868236662103e-05, {   1,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.528835418708e-05,  3.488709575755e-05, {   4,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.771666013817e-05,  3.241347735399e-05, {   2,   0,  -1,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 {  2.366793022519e-05, -3.639410921026e-05, {   4,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.280513899363e-06, -5.176932778149e-05, {   2,   0,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -5.999999991830e-05, -3.131120676668e-09, {   2,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.999999996489e-05,  2.052507199942e-09, {   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.999999998079e-05, -1.518436449079e-09, {   2,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.999999998503e-05,  1.340412865420e-09, {   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.999999998526e-05, -1.329940889916e-09, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -5.999999999486e-05,  7.853981633750e-10, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.999999999552e-05,  7.330382858194e-10, {   2,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.999999999552e-05, -7.330382858140e-10, {   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.999999999565e-05, -7.225663102994e-10, {   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.999999999923e-05,  3.036872898457e-10, {   4,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.999999999996e-05, -7.330382857725e-11, {   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   2,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.130219970052e-05,  4.869510202200e-05, {   2,   0,  -1,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.800205531634e-05, -1.199077756764e-05, {   0,   0,   1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  2.519691645454e-05,  3.473802784909e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.454902472772e-05,  4.537910219567e-05, {   2,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.857758150239e-05,  1.134681340732e-05, {   2,   2,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.878765378469e-05, -2.113248954679e-05, {   2,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  1.686871530950e-05,  4.298929253634e-05, {   2,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.103954231476e-06, -5.774596794601e-05, {   0,   0,   1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.042156125340e-05, -2.942154877819e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  5.362069070716e-06,  5.447895825483e-05, {   0,   0,   2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  3.986382839036e-05,  1.991065481087e-05, {   4,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.744958347290e-07,  5.947882912825e-05, {   2,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.171576349365e-05,  1.798618061141e-05, {   2,   0,  -1,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -4.791865780494e-05,  1.176065996964e-05, {   1,  -1,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.203062781789e-05, -4.762140743975e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  2.311193710492e-05,  3.653080279646e-05, {   3,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.811877226698e-05,  1.503490204109e-06, {   0,   0,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  1.516917793107e-05,  4.441801874018e-05, {   0,   0,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.348069062430e-05, -3.609355291037e-05, {   4,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.955515613896e-05, -4.630835123656e-09, {   1,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.939582780494e-05,  1.455263599079e-07, {   2,   0,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.933590957961e-05, -1.017572731298e-05, {   1,  -1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.590745176812e-05, -4.359159359372e-05, {   0,   0,   1,   0,   0, -26,  29,   0,   2,  -5,   0,   0,   0 } },
 { -5.067143550150e-06, -5.440354982500e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -1.158406331418e-05, -4.787337587003e-05, {   4,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.181304428614e-05,  4.757645504704e-05, {   1,   1,  -1,   0,   0, -17,  15,   0,   1,   0,   0,   0,   0 } },
 { -2.623384351362e-06,  5.675422934697e-05, {   2,   0,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.407760590957e-05, -2.528163308048e-05, {   2,   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -1.679024272596e-05,  4.256528617873e-05, {   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.934387688316e-05,  9.180572066000e-10, {   2,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.157827344349e-05,  2.773485673339e-05, {   0,   0,   0,   0,   0,   3, -11,  12,  -3,   0,   0,   0,   0 } },
 {  3.841833332266e-05, -2.087461539439e-05, {   2,   0,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  1.207628307053e-05, -4.721629970541e-05, {   1,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.069557947690e-05, -4.857339810631e-05, {   2,   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -4.022216517618e-05, -1.904350377330e-05, {   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.633894699403e-05, -2.290562884266e-05, {   0,   0,   2,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.802922966066e-05, -1.121046727882e-05, {   0,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  5.870725406922e-05,  4.731496619374e-07, {   4,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.484414019252e-05, -3.432131081765e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   4,   0,   0,   0 } },
 {  2.772983297819e-05,  3.142738845212e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -1.113570944173e-05,  4.799995917321e-05, {   1,  -1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.776005217515e-05,  1.130270784575e-05, {   2,   0,  -2,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.407852992905e-05,  4.497960283656e-05, {   4,   0,  -2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.701228885422e-05, -1.204145141643e-05, {   1,   1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -9.989209125596e-06, -4.902502750775e-05, {   2,   0,  -1,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.991360866771e-05, -1.902708127017e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -9.310219937991e-06,  4.960772934346e-05, {   2,  -2,   1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.073939578530e-06,  5.483963319724e-05, {   4,   0,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.178168056061e-05,  3.707019618596e-05, {   1,   0,   0,   0,   0, -12,   7,   0,   1,   0,   0,   0,   0 } },
 { -1.117045826521e-05,  4.766491968759e-05, {   1,  -1,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.763479631531e-05, -3.115282011348e-05, {   4,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.905563862357e-05,  9.671831224443e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.790108919945e-05, -1.079572148408e-05, {   3,  -1,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  3.415787148996e-05,  2.453500420075e-05, {   2,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  5.651053389896e-05, -2.148728052923e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.922597082245e-05, -2.942586509259e-05, {   0,   0,   0,   0,   0,   3, -14,  17,   1,   0,   0,   0,   0 } },
 { -2.223914361742e-05, -3.640484671606e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0 } },
 {  4.707359763400e-05, -1.153428943013e-05, {   1,  -1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.903801495766e-05, -1.956070532929e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.358112035106e-05,  4.501507981255e-05, {   0,   0,   1,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.429796264011e-05,  2.427734106850e-05, {   2,   0,  -1,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 { -2.356619727090e-05,  3.500378816580e-05, {   2,   0,  -1,   0,   0,  -5,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.991979719235e-05,  1.862346505887e-05, {   5,  -1,  -2,   0,  -6,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.206216006118e-05,  1.647370345897e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -5.737624026368e-05, -1.155340536302e-06, {   4,   0,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.824224498357e-05,  2.883871291790e-07, {   3,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.008793735778e-05, -2.841884479472e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 { -4.291935362098e-05,  1.557695356110e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  8.380785449340e-08,  5.838000135092e-05, {   2,   0,  -2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.835823572855e-05, -1.009402204728e-05, {   2,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.555118427903e-05,  2.900607093588e-06, {   2,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.593318109841e-05,  1.245375196843e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.548078009165e-05,  4.289754674295e-05, {   2,   0,   0,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 { -1.509748790929e-05, -4.326932113675e-05, {   2,   0,  -1,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 {  4.289604897810e-05, -1.546025904685e-05, {   4,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.488502087039e-05, -4.346016147937e-05, {   2,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.915758507023e-05, -2.916414963730e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   3,   0,   0,   0 } },
 { -5.738340745651e-05,  8.706649638409e-07, {   6,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.137043114652e-05, -3.687861914797e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -1.385204140668e-06, -5.686274626101e-05, {   2,   0,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.814196604309e-05,  1.010118743449e-05, {   2,   0,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 { -4.917206731775e-05, -9.061335205070e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.438791270814e-05,  4.372812823279e-05, {   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.207309325767e-05, -2.602969403477e-05, {   1,   0,   1,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 {  3.207309325767e-05, -2.602969403477e-05, {   1,   0,  -1,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 { -3.654433143044e-05,  2.153831331955e-05, {   0,   0,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  5.792694220357e-05, -1.464798918710e-07, {   2,   0,   1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.533967627509e-05, -3.268271602432e-05, {   2,   0,   0,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 {  5.798690367374e-05,  2.783644967488e-08, {   4,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.743599745019e-05, -1.048272355667e-05, {   4,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.177210981026e-05,  4.613469692875e-05, {   2,  -2,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -5.694136451523e-05,  9.482746543362e-07, {   4,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  4.452349882858e-05,  1.334851327232e-05, {   0,   0,   1,   0,   0, -26,  21,  16,  -4,  -5,   0,   0,   0 } },
 {  5.783811018156e-05,  1.069265603126e-08, {   4,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.085514287840e-05,  6.989359378466e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 {  1.302465478043e-05,  4.480160742026e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -3.155260241399e-05,  2.626846333322e-05, {   2,   0,  -1,   0,   0,   3,  -2,  -2,   0,   0,   0,   0,   0 } },
 {  3.855918932301e-05,  1.921325390600e-05, {   3,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.728114905841e-05, -2.048804250880e-05, {   0,   0,   3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.673303898298e-05, -4.103598615484e-05, {   0,   0,   2,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  3.661105149122e-05, -2.112685883354e-05, {   0,   0,   3,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.672595517782e-05, -4.100988541862e-05, {   0,   0,   2,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  5.759808896678e-05,  1.237136334847e-07, {   4,   2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.849316473838e-05,  1.922697615395e-05, {   2,  -2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.683856500864e-05, -2.088061997308e-05, {   4,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -5.769177731049e-05,  2.599140050638e-08, {   1,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.212246428216e-05, -3.557566945270e-05, {   2,   0,  -4,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.393195552257e-05, -3.375622049465e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.274698211548e-05, -4.490675149605e-05, {   3,   0,   0,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 { -5.315502125578e-05, -4.495680668192e-06, {   2,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.290290804004e-05,  4.473774555584e-05, {   2,   2,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.289064424122e-05,  4.469953683703e-05, {   2,   2,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -5.743177891935e-05, -1.576844505750e-07, {   2,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -3.022768046735e-05, -2.731851556802e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -6.161723439420e-06,  5.133370261017e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.772520132844e-05,  1.976495962151e-05, {   1,   1,   0,   0,   0, -38,  36,   0,   0,   0,   0,   0,   0 } },
 {  5.523149350305e-05,  2.230265904331e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  5.508933420613e-05,  2.333700210574e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   6,   0,   0,   0 } },
 { -5.683342500967e-05, -4.956426440718e-07, {   1,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.539537678881e-05, -1.193117282443e-05, {   0,   0,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -1.263058040723e-05,  4.466976477322e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  5.723716833186e-05, -2.120990292424e-08, {   2,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.119807187924e-05,  3.600843943675e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  2.427069044217e-06,  5.477493752776e-05, {   1,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  3.835213823831e-05, -1.883270738302e-05, {   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  3.417225584399e-06, -5.376292549685e-05, {   1,   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -4.121801138037e-05, -1.593502814691e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  3.179271550895e-05, -2.533773780242e-05, {   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.744814792350e-05,  9.655803057724e-06, {   4,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.567454555313e-05, -1.139343787828e-05, {   2,  -3,   0,   0,   0,  28, -30,   0,   0,   0,   0,   0,   0 } },
 { -1.330207009948e-05,  4.374871664587e-05, {   2,   0,  -1,   0,  -4,  -3,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.102650481189e-05, -3.600943897457e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -2.575405666429e-05, -3.126743813743e-05, {   2,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.768907763442e-05,  9.309045725636e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  1.073620922494e-05,  4.624535275725e-05, {   2,  -1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.601891667205e-05, -3.093498501574e-05, {   3,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.250709200337e-05, -2.437405000538e-05, {   1,   1,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -5.685307277325e-05, -1.252495027065e-08, {   1,   0,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.780213748171e-05,  9.059400448712e-06, {   1,  -1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.475154534621e-05, -2.209705710215e-05, {   0,   0,   2,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.019481620979e-05, -3.660662247255e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.077038647150e-05,  4.588465551776e-05, {   1,  -1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.596784130741e-05, -2.067621576961e-05, {   2,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 { -3.572353894148e-05,  2.090905914621e-05, {   2,   0,  -1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 { -5.365639948481e-05, -2.939791532655e-06, {   0,   0,   1,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 { -2.577204670399e-05, -3.081228735755e-05, {   4,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.007931806999e-05, -3.644520674573e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -2.995010613171e-05,  2.649328184083e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -3,   0,   0,   0 } },
 { -4.768119896956e-05,  8.744663758898e-06, {   1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.792111849570e-05,  1.847246682559e-05, {   2,   0,  -1,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.178841029399e-06, -5.520794793644e-05, {   2,   0,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.632551556236e-05,  5.988970905311e-08, {   3,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.905257553492e-06,  4.745409557387e-05, {   2,   2,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.306591977953e-06, -5.404863372852e-05, {   0,   0,   2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.986843457589e-06, -5.436579909537e-05, {   0,   0,   1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.619784977717e-05,  1.411705656153e-07, {   1,   0,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.802482186409e-05, -8.311058159080e-06, {   2,   0,  -1,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.561122469846e-05,  2.069942692707e-05, {   3,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.212263880516e-05,  2.418135130939e-05, {   0,   0,   0,   0,   0,   0,   1,  -2,   2,  -3,   0,   0,   0 } },
 {  2.910048146319e-05,  2.718448302335e-05, {   3,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.822030280769e-06,  5.345945667941e-05, {   2,   0,  -1,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 { -4.812497193716e-05,  8.152396572982e-06, {   1,   1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.129335190966e-05,  4.497151952815e-05, {   3,  -1,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -4.771338066797e-05,  8.549556795972e-06, {   2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.847074181692e-06, -5.340627063061e-05, {   2,   0,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.074710044005e-05, -4.549426842715e-05, {   1,   1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.534070385504e-05,  3.085644119234e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -3.348047515147e-05,  2.269353255022e-05, {   1,  -1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -2.466204805028e-05,  3.149259255707e-05, {   2,   0,  -2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.027169319212e-05,  2.587864624999e-05, {   2,   0,  -1,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 {  3.994475510899e-05,  1.617322237784e-05, {   2,   0,   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.433295513580e-05,  1.177094192705e-05, {   2,   0,  -1,   0,   0,   0, -20,  16,   0,   0,   0,   0,   0 } },
 { -1.544404794199e-05, -4.064045093785e-05, {   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 } },
 {  2.428607573653e-05,  3.178343161648e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.059401341669e-05, -4.546167033269e-05, {   1,  -1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.740896908108e-05, -1.863546815930e-05, {   3,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.211654084302e-05,  2.391981320797e-05, {   2,   0,  -1,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0 } },
 {  2.164100406975e-05, -3.437584048523e-05, {   4,   0,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.212148419725e-05,  4.388995701778e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 { -1.151868451574e-06, -5.485303552121e-05, {   2,   0,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.486628621468e-05,  1.070389084193e-06, {   3,   1,  -1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.486628621468e-05,  1.070389084193e-06, {   1,  -1,   1,   0,   1,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.492455718067e-05,  9.835410654485e-07, {   2,   0,   1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  8.177692380946e-06,  4.770737844567e-05, {   1,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.165874664106e-05, -3.422578257054e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 { -3.499667957866e-05,  2.087174203441e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 {  3.249630731291e-05,  2.332359344151e-05, {   4,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.573684074404e-05, -4.007416355268e-05, {   1,  -1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.278822609088e-05,  2.986687632103e-06, {   2,   0,  -2,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 { -4.893962820950e-05,  6.826302787667e-06, {   0,   0,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 {  5.329389665421e-05,  2.443252722927e-06, {   2,   0,  -1,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 { -4.546889536444e-05, -1.024831191016e-05, {   3,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.829896056521e-05, -3.739775290385e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 { -5.556669800724e-05,  1.016197756875e-08, {   0,   2,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.643278946869e-05,  9.107362802869e-06, {   4,   0,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.541540011510e-05, -1.065646581430e-07, {   0,   0,   2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.233454101323e-05, -3.318385567906e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.587196954868e-05,  9.635272239742e-06, {   1,   0,   2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  4.587196954868e-05,  9.635272239742e-06, {   1,   0,  -2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  8.768363196491e-06,  4.673620003283e-05, {   2,   2,  -2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.844550454290e-06, -5.263974840282e-05, {   0,   0,   0,   0,   0,   0,  10, -19,   2,  -2,   0,   0,   0 } },
 { -5.542011748804e-05, -3.592156963232e-08, {   2,  -2,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.166281235508e-06,  4.728437929450e-05, {   1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.523102175395e-05,  8.173991230486e-08, {   6,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.688325902979e-05,  1.842168311097e-05, {   3,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.270216604091e-05, -2.602755178248e-06, {   0,   0,   1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.527739125352e-05, -2.350703802452e-08, {   1,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.696496327445e-05, -8.311196814204e-06, {   2,  -2,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  6.781337053684e-06, -4.847590859931e-05, {   4,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.049526694113e-05,  4.475738211026e-05, {   1,  -1,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.829497297938e-05, -6.923514462451e-06, {   1,  -1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.459766863866e-05, -1.060253138782e-05, {   1,  -1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.515634848093e-05, -3.552567389009e-08, {   1,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.062160634994e-05,  3.455130697641e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -4.739608724857e-05, -7.772789824827e-06, {   1,   1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.066452456169e-05,  4.450302377570e-05, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.656072069255e-05,  3.859605539503e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 {  1.627102047549e-06,  5.352856358642e-05, {   1,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -5.387167330500e-05, -1.238680611313e-06, {   4,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.617209587439e-06,  4.647603444405e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.971177118492e-05, -5.360951935790e-06, {   1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.628976160085e-06, -4.642759662357e-05, {   1,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.422580553146e-05,  1.076309292321e-05, {   2,  -2,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.492754840615e-05, -5.519110055773e-08, {   3,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.876349925914e-05, -1.621389802343e-05, {   1,   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.378110992966e-05,  3.118398976716e-05, {   2,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.878722249898e-05, -3.617043927258e-05, {   2,   0,   1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -1.878722249898e-05, -3.617043927258e-05, {   2,   0,  -1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 {  3.366497474559e-05,  2.128566430777e-05, {   4,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  3.880514751763e-05, -1.611098193202e-05, {   0,   0,   0,   0,   0,   0,   6, -11,  -2,   0,   0,   0,   0 } },
 {  1.023414490554e-05,  4.463973052374e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.119492038398e-05, -3.366024901260e-05, {   4,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.084271131117e-05,  3.401081822094e-05, {   2,   2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.084611723972e-05, -3.400231900732e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -4.541512928544e-05, -9.424284043931e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  1.938424161915e-06, -5.290064256744e-05, {   2,   0,  -2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.680691743574e-07, -5.384591089140e-05, {   2,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.366935003823e-05, -1.144131263671e-06, {   0,   2,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.679735808145e-05, -2.800023525374e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  2.679735808145e-05, -2.800023525374e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -1.641533284045e-06, -5.310458304035e-05, {   2,   0,  -1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.769732193911e-05, -7.014485593000e-06, {   0,   0,   2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.744433513281e-05, -2.726086079768e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  1.594521981247e-05,  3.875541912002e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 {  3.037054985087e-05,  2.432728420320e-05, {   0,   0,   0,   0,   0,   6,  -9,   0,   3,   0,   0,   0,   0 } },
 {  4.909684208081e-05,  5.515540572327e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 { -5.474510123734e-06, -4.912082188225e-05, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -3.343755279663e-05,  2.113221517209e-05, {   4,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.576056720149e-05,  3.880180934096e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 {  5.446076636632e-05,  8.185414523697e-08, {   2,   0,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.755420292581e-06, -5.275420107050e-05, {   0,   0,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.333104191651e-05,  1.176553398231e-06, {   0,   0,   0,   0,   0,   0,   3,  -4,   1,   0,   0,   0,   0 } },
 { -1.041137520669e-05,  4.408870192667e-05, {   1,  -1,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.637312372853e-05,  1.809367200652e-05, {   1,   1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.432438053627e-05, -3.008702962038e-05, {   0,   0,   0,   0,   0,  14, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.034553091049e-05, -4.405429095417e-05, {   1,  -1,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.795785303543e-06, -4.557399607376e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 {  5.434048278936e-05,  2.717846263747e-08, {   4,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.422712879865e-05,  4.011689497860e-05, {   2,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.040592357234e-05,  4.391429365664e-05, {   3,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.474475791571e-06, -4.979938821302e-05, {   1,  -1,   2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.474102051230e-06, -4.979942179218e-05, {   2,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.472780920564e-06,  4.979954046813e-05, {   2,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.472772228905e-06,  4.979954124877e-05, {   1,   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.472702695624e-06,  4.979954749389e-05, {   1,   1,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.711786609538e-05,  2.714129787049e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -3,   0,   0,   0 } },
 { -3.888318120435e-05,  1.537377294459e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 { -2.567436900270e-05,  2.855437713426e-05, {   2,   0,  -1,   0,   0,   0,  -2,   7,   0,   0,   0,   0,   0 } },
 { -5.247580445594e-05,  1.746009046277e-06, {   2,   0,  -2,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 {  4.579896754781e-05, -8.421245823538e-06, {   2,   0,  -3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.040732682663e-05,  4.380585168361e-05, {   2,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.068651608634e-05, -3.345343925558e-05, {   2,   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 {  1.054414831589e-05,  4.357640567275e-05, {   2,   2,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.507239090400e-05,  1.904400166897e-05, {   4,   0,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.879688400199e-05,  3.530831561850e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.725529182482e-05,  3.683140044781e-05, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  3.786341287117e-05, -1.622170001124e-05, {   2,   0,  -3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.079052998583e-05, -3.328441750305e-05, {   1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.570541310915e-05, -8.361418386450e-06, {   1,  -1,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.729357514585e-05, -3.674368172338e-05, {   2,   0,  -1,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 {  2.754536933485e-05,  2.647481957744e-05, {   2,   0,   1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.894492902510e-05, -1.506852065057e-05, {   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.364495808128e-05,  4.033744110846e-05, {   2,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.156720876295e-05,  2.237165129432e-05, {   2,   0,  -1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 { -2.958101447599e-06, -5.096857254752e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.594945736701e-05,  1.795526566489e-05, {   4,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.276755733310e-05,  4.113555653454e-05, {   0,   0,   1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.263921870662e-05,  3.120961704503e-05, {   2,   0,  -2,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.578834571325e-05,  3.804062386151e-05, {   0,   0,   0,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 { -2.858448512295e-05, -2.522833336528e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 { -1.005453163757e-05,  4.375776810151e-05, {   1,  -1,   1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.153895428519e-05,  2.226720933876e-05, {   2,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 {  3.152953428017e-05,  2.223144538781e-05, {   4,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.102454852064e-05,  4.270216880716e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 {  1.102099059447e-05,  4.269324107263e-05, {   1,  -1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.961317344033e-05, -2.407604376019e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.997959944333e-05, -2.370894117848e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 {  3.381511551908e-05, -1.986348157935e-05, {   4,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.406287019182e-05, -3.960400157751e-05, {   0,   0,   2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  4.496756169739e-05,  8.687141661771e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 {  4.676812279076e-05, -6.880769255380e-06, {   1,  -1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.989716839441e-05,  1.365575637572e-05, {   1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.010847436094e-05, -4.342192847077e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 { -5.249739219143e-05,  1.022755972777e-06, {   3,   1,  -2,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.401915924513e-05, -3.948212379418e-05, {   0,   0,   2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.071140303721e-05,  4.278061256581e-05, {   1,   1,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.757529750762e-05,  5.865742493293e-06, {   0,   0,   1,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 { -2.073851456027e-05, -3.269716106792e-05, {   1,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.870115370747e-06, -5.154635440280e-05, {   0,   0,   1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.248405366213e-05,  1.085265216789e-05, {   2,  -2,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.491109049289e-05, -3.840527298998e-05, {   1,  -1,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.615062060558e-05, -7.147942466766e-06, {   0,   0,   2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.005056290771e-05, -4.324131918349e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 { -5.175335965676e-05,  1.520497361333e-06, {   2,   0,  -3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.757269384811e-05, -2.566872658319e-05, {   0,   0,   0,   0,   0,   0,   4,  -7,  -3,   0,   0,   0,   0 } },
 {  4.465470597840e-05,  8.545504695302e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 {  1.406326573330e-05, -3.912754753032e-05, {   1,  -1,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.926243350836e-05,  3.392596069211e-05, {   1,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.443867472977e-05, -3.870886681890e-05, {   0,   0,   1,   0,   0, -17,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.104145204335e-06,  4.703406102234e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  3.528442904938e-06,  4.960548730203e-05, {   2,   0,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.636938270964e-05, -3.671500652592e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 { -2.365909518746e-05,  2.942480294664e-05, {   1,  -1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.018465437967e-05,  4.289678568974e-05, {   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.468434505351e-05,  8.377742296640e-06, {   2,  -2,  -1,   0,   0,  31, -35,   0,   0,   0,   0,   0,   0 } },
 {  1.775078986586e-05, -3.528574776335e-05, {   0,   0,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.114279237013e-05,  1.188175915614e-05, {   1,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.503634715579e-05,  7.979739651511e-06, {   2,   0,  -5,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.829945528447e-05, -4.715333952798e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 {  4.946070507330e-05, -3.551900173949e-06, {   2,   0,  -1,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  5.197179731936e-05,  1.012574282501e-06, {   1,  -1,   0,   0,   1,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.082619383893e-05, -2.144642172365e-06, {   0,   2,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.852765516906e-06, -5.111097835627e-05, {   0,   0,   1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.538616835546e-05,  1.752971807671e-05, {   0,   0,   1,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.432730105762e-05,  1.857756931018e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 {  4.518168064286e-05, -7.691447866248e-06, {   2,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.909667653083e-05,  1.375263833732e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -5.275668378115e-05,  6.520909835904e-08, {   2,  -2,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.073364795729e-05,  4.205363864799e-05, {   1,   1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.394559295588e-05, -3.883405736471e-05, {   1,  -1,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.701061580294e-05, -5.750731574058e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 { -4.852537375240e-06, -4.790343827225e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -8.262521490576e-06,  4.448916898066e-05, {   3,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.454296709847e-05,  2.819339787693e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 { -2.046722109629e-05, -3.222014357673e-05, {   3,  -2,   0,   0,   0,   0, -31,  39,   0,   0,   0,   0,   0 } },
 { -7.946142326661e-06, -4.472028309413e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  3.666491546515e-05, -1.599012113567e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.833437453474e-05, -1.429745346690e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 { -3.508402556814e-05,  1.754275686304e-05, {   0,   0,   2,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.838687562545e-05, -3.423766159491e-05, {   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  5.137286557610e-06,  4.748350059185e-05, {   4,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.258965034235e-05,  3.437041178647e-09, {   2,  -2,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.919170916337e-06,  4.266422454342e-05, {   1,  -1,  -1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.165230099288e-06,  5.040918204717e-05, {   2,   0,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.099298898500e-05,  1.555194066761e-06, {   1,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.556914173793e-06, -5.094944973694e-05, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  2.398135694977e-05, -2.850999789832e-05, {   3,  -1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.608376366753e-05,  1.640205085978e-05, {   0,   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -4.361424537207e-05,  8.838384176572e-06, {   1,  -1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.450882963274e-05,  3.793785359623e-05, {   0,   0,   1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.871813780563e-05,  2.372751796901e-05, {   3,   0,  -2,   0,  -4,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.785439154952e-05,  2.458048251169e-05, {   2,   0,   0,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 { -7.753902109777e-06, -4.465509465610e-05, {   2,   0,  -2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -3.448444977542e-05, -1.790562202477e-05, {   0,   0,   2,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  5.213502280826e-05,  2.471476127348e-07, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 { -1.824231545237e-06, -5.055407612765e-05, {   0,   0,   1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.556044107913e-05, -1.680852883718e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -3.469519773073e-05, -1.763367908489e-05, {   1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  7.182573206063e-06,  4.514133098751e-05, {   2,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.386258891255e-05,  1.837541956072e-05, {   1,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.923284336818e-07, -5.180888787521e-05, {   0,   0,   0,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 { -3.351318690066e-05, -1.864861823245e-05, {   2,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.053529318729e-05,  1.567475462562e-06, {   0,   2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.055190925582e-05,  4.150699972090e-05, {   1,  -1,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.618191638541e-05,  1.586519602926e-05, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 { -1.834684647598e-06, -5.018909732441e-05, {   0,   0,   1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.469029183820e-05,  1.732610379094e-05, {   2,  -2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.731219900300e-05,  3.469203540966e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 { -3.006538849003e-05, -2.193040255508e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  4.192552060505e-05,  1.006298526238e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 { -3.147409577413e-05,  2.048249766549e-05, {   3,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.967105085616e-06, -4.197695079218e-05, {   1,  -1,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.532690115449e-05, -1.661592451518e-05, {   2,   0,  -2,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.351106714541e-05,  2.841686512132e-05, {   4,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.762619549502e-06,  5.010642836303e-05, {   0,   0,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.923051762960e-06, -4.191239316204e-05, {   1,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.727649514860e-05, -4.544927118105e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 {  4.769203576443e-05,  4.117837601152e-06, {   1,  -1,   0,   0,   0,   0,  -8,  17,   0,   0,   0,   0,   0 } },
 { -3.900049502945e-05, -1.277474297059e-05, {   2,  -2,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -4.219173368901e-06, -4.755021096020e-05, {   0,   2,   0,   0,   0, -23,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.546777240765e-05,  1.627865554963e-05, {   1,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.557147566760e-05,  2.616510381836e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   6, -16,   3,   0,   0 } },
 { -5.081126042814e-05,  9.174228781382e-07, {   2,   0,   1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.430121816649e-05, -3.735605920001e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.262335878894e-05,  8.992629632438e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  4.396268030172e-05, -7.638304424184e-06, {   0,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 { -5.153901388842e-05,  4.556660004190e-08, {   3,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.034501175455e-06,  4.353228972002e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 { -2.554297541577e-06, -4.899246363882e-05, {   4,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  4.120701254676e-05, -1.032717023427e-05, {   2,   0,  -2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.221933225351e-05,  2.931151073800e-05, {   2,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  2.073581052233e-06, -4.942310551130e-05, {   0,   0,   2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.280687109344e-05,  2.867531404223e-05, {   1,  -1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  2.399620212203e-05, -2.747922171202e-05, {   2,   0,  -1,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.009345534344e-05, -2.138026191278e-05, {   2,   0,  -2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  4.987159344237e-05, -1.595160000208e-06, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.106715985830e-05, -3.891804628246e-07, {   2,   0,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.535063646745e-05,  3.608446940971e-05, {   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.426654473508e-05, -1.711388330330e-05, {   2,   2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.029132042832e-05, -2.108757582041e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   3,   0,   0,   0,   0 } },
 { -3.294928676350e-05, -1.841704914498e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.298437862553e-05, -2.836367671261e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -7.071911454640e-07, -5.061845327852e-05, {   2,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.602884874717e-05, -2.529459809685e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -9.133852745221e-07, -5.039952828373e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 { -4.854821980296e-05,  2.757419792993e-06, {   0,   0,   1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  4.554143083973e-05,  5.757040219227e-06, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 { -2.731689690836e-05, -2.391351511133e-05, {   2,   0,  -1,   0,   0,   0,   1,   2,  -3,   0,   0,   0,   0 } },
 {  1.509187197521e-05,  3.613070835254e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 { -5.107961880959e-05,  5.437825210125e-08, {   2,   0,  -2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.169945368890e-05, -3.942696751182e-05, {   6,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  4.361888724003e-05,  7.485090695289e-06, {   1,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.107276840759e-05, -2.137019549217e-08, {   1,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.168624928929e-05, -3.939110860241e-05, {   6,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.417125184340e-05, -1.684773006861e-05, {   1,  -1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.266549998948e-05, -2.834506374934e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -3.369738007249e-05,  1.730663862710e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 {  3.400302509789e-05,  1.698948580680e-05, {   0,   0,   2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.810764157990e-05, -2.285217831878e-05, {   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.652109490629e-05, -4.434908918495e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,  -3,   0,   0 } },
 { -9.289840322108e-06,  4.166614995031e-05, {   1,  -1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.318962977412e-05,  2.776602765824e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 { -2.695018325479e-06,  4.825067611802e-05, {   2,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.366575032358e-05,  3.727752450682e-05, {   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.369368834033e-05,  3.723402939682e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 { -4.911663283141e-05,  1.810875758070e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -6,   5,   0,   0,   0 } },
 {  4.862056449355e-05,  2.267620928190e-06, {   3,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.839081015729e-05, -3.249439329964e-05, {   4,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.025194902610e-05, -2.061219986406e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -11,   3,   0,   0 } },
 {  3.025194902610e-05, -2.061219986406e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  11,  -3,   0,   0 } },
 { -4.356298837726e-05, -7.294308594355e-06, {   2,   0,  -1,   0,   0,   0,   6,   0,  -7,   0,   0,   0,   0 } },
 {  3.553928952083e-05, -1.531451720555e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 { -1.124284698956e-05, -3.958663224938e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.297077752389e-05,  3.783859049206e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.403242060302e-05, -1.677632507323e-05, {   1,  -1,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.413692923207e-06, -4.138581861863e-05, {   0,   0,   0,   0,   0,  16, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.561192764536e-05, -2.518701284740e-05, {   2,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.051250546175e-05, -1.022488502320e-05, {   2,  -2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.685311012215e-06, -4.804827221845e-05, {   2,   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 {  2.979479774125e-05,  2.093700683949e-05, {   2,   0,  -1,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 { -1.155140239894e-05, -3.917089194211e-05, {   3,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.849606137868e-05, -2.223148551906e-06, {   2,   0,  -2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  5.068630100745e-05,  2.604185056543e-08, {   4,  -2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.057363665214e-06,  4.765052547535e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.218686872302e-05, -3.851851512513e-05, {   0,   0,   0,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.392913021766e-05,  1.674661264884e-05, {   2,   0,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.065581521646e-05,  1.135330426004e-08, {   0,   2,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.607461228512e-06,  4.105430182125e-05, {   1,  -1,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.143252612658e-05,  3.922502132534e-05, {   1,   0,   0,   0,   0, -20,  16,   8,  -3,   0,   0,   0,   0 } },
 { -5.055057515364e-05, -1.029605371753e-07, {   2,   0,  -3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.535027169712e-05, -5.302416384915e-06, {   1,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.314097462872e-05, -2.750993678437e-05, {   1,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.377133182710e-05,  3.687882435486e-05, {   1,  -1,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -4.969050619217e-05,  9.321605841659e-07, {   2,  -2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.884409777896e-05, -1.775428499316e-06, {   0,   0,   1,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 { -4.934660847319e-05,  1.257274873351e-06, {   0,   0,   4,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.301322519714e-05, -7.528016674351e-06, {   2,   0,  -3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.069324903769e-05, -3.983604938336e-05, {   3,   1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  3.634966445113e-05,  1.416898833460e-05, {   4,   0,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  3.243525258636e-05, -1.803156563443e-05, {   1,   0,  -3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.149290660691e-05, -3.896026909825e-05, {   3,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.177987141298e-05, -8.659196559145e-06, {   1,  -1,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.251663746097e-05, -3.792049097589e-05, {   2,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.040669595947e-05,  4.256025312678e-09, {   3,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.271456943624e-05, -7.662522692790e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -6.770928499261e-07,  4.969748858669e-05, {   1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.035394908867e-05,  1.663787518967e-08, {   4,   0,  -2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  9.636306399882e-06, -4.070054706586e-05, {   2,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.539778269146e-05,  3.493224660495e-05, {   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.138911271456e-05,  1.889502747233e-05, {   1,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.573746751278e-06,  4.469555397455e-05, {   0,   0,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.421438204120e-05,  1.604852022654e-05, {   0,   0,   2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.682095473906e-05, -3.439274221472e-06, {   0,   0,   1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 {  4.926268823209e-05,  9.319138021235e-07, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 { -4.999966817202e-05,  1.152085467629e-07, {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.134970822182e-05, -3.875896224008e-05, {   1,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.978819185516e-06, -4.812355567332e-05, {   2,  -2,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.999774324072e-06,  4.009964976949e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 { -4.999990459138e-05, -9.767728945945e-08, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.999990753466e-05,  9.615885586284e-08, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.226152378208e-05, -7.828451462807e-06, {   1,  -1,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.165832504429e-05,  3.841101873918e-05, {   1,  -1,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  4.377858243913e-05, -6.256147647055e-06, {   2,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.156179664054e-05,  3.847138290447e-05, {   1,   0,   0,   0,   0, -20,  24,  -8,   3,   0,   0,   0,   0 } },
 {  1.636833603439e-05,  3.364918055989e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -4.999999985523e-05,  3.804817765668e-09, {   2,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.999999998170e-05, -1.352630170138e-09, {   2,  -1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.999999998421e-05, -1.256637061306e-09, {   2,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999998134e-05, -1.221730476197e-09, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.999999998848e-05,  1.073377489894e-09, {   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.999999999192e-05,  8.988445647287e-10, {   2,  -1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.999999999605e-05, -6.283185307024e-10, {   2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.999999999605e-05, -6.283185307024e-10, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.999999999648e-05, -5.934119456541e-10, {   2,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999999922e-05, -2.792526803071e-10, {   2,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.315404894517e-05,  6.842887932875e-06, {   2,   0,  -2,   0,   0,   6, -11,   6,  -2,   0,   0,   0,   0 } },
 { -2.896190389472e-06, -4.708549395741e-05, {   3,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.987121055138e-05, -8.739140693181e-08, {   3,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.279228135460e-05, -2.711431829747e-05, {   1,  -1,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  2.139160395258e-06, -4.770573948942e-05, {   4,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  3.498199373851e-05,  1.486268068290e-05, {   2,   2,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.498199373851e-05,  1.486268068290e-05, {   2,   2,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.312354909939e-06, -4.050949484578e-05, {   2,   1,  -3,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.307249215045e-06,  4.747716700208e-05, {   2,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.742927474573e-05,  3.233771148101e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   5,   0,   0,   0 } },
 {  4.974290254346e-05, -1.394295838257e-08, {   0,   0,   4,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.952216290317e-05, -2.325945639641e-07, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  7.142446751835e-06,  4.260270020002e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 { -4.228729488702e-05, -7.442539516359e-06, {   4,   2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.710089482399e-05,  1.261767436655e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.812481652285e-06, -4.790488044972e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.524124317375e-05, -3.446350365949e-05, {   0,   0,   1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.244914170544e-05, -1.720258405873e-05, {   1,  -1,  -1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  2.026856531831e-05, -2.938287212600e-05, {   0,   0,   0,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 { -4.925522115859e-05, -3.885276178177e-07, {   2,  -2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  9.067452197342e-06, -4.055953245651e-05, {   1,  -1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  3.972515769440e-07, -4.921476771421e-05, {   2,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.468247574360e-05, -2.491536542208e-05, {   0,   0,   2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  3.082385214351e-05,  1.875744819713e-05, {   2,   0,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0 } },
 { -3.611444092452e-05,  1.342638728917e-05, {   0,   0,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  3.449486764040e-05,  1.504317805478e-05, {   2,   0,  -1,   0,   0, -25,  29,   0,   0,   0,   0,   0,   0 } },
 { -4.407250122433e-05, -5.442604722970e-06, {   0,   0,   0,   0,   0,   3,  -8,   6,   0,   0,   0,   0,   0 } },
 {  3.258968841983e-05, -1.687746151307e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.576664470267e-05, -3.369091280235e-05, {   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.126868336727e-05, -3.818295528966e-05, {   0,   0,   0,   0,   0,   0,  17, -32,   0,   0,   0,   0,   0 } },
 {  2.890659724841e-05,  2.050484122665e-05, {   0,   0,   0,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  4.914366040031e-05, -2.560963783962e-07, {   4,   0,  -2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -4.517197071566e-05, -4.221007596998e-06, {   0,   0,   1,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 { -3.352556069443e-05, -1.581857337542e-05, {   2,   0,   0,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.418364820211e-05,  3.512328665408e-05, {   1,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.424941649414e-05,  3.503725936255e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 {  4.925098774435e-05,  3.012736019576e-08, {   6,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.428717671656e-06, -3.984670166493e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 {  2.718274860200e-07, -4.897822967649e-05, {   2,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.767701977633e-06,  4.745387474971e-05, {   1,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -2.789081970571e-05, -2.132556929364e-05, {   1,  -1,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  3.117013517316e-05, -1.803246914346e-05, {   2,   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 { -1.595278641508e-05, -3.323959214433e-05, {   2,   0,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -9.098333197539e-06,  4.009149079448e-05, {   1,  -1,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -6.712770361884e-06,  4.246742404181e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -4.913338353424e-05,  2.191399669068e-08, {   0,   0,   2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  8.972824342332e-06, -4.017025131576e-05, {   4,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.302650681498e-06, -3.981628174251e-05, {   3,  -1,  -2,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.837466362649e-05, -1.073820398726e-05, {   2,   0,  -1,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 { -9.170910366532e-06,  3.992111263893e-05, {   3,  -1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  2.937657877780e-05,  1.971217039305e-05, {   2,   0,   0,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -1.580295073029e-05, -3.328484768225e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.201807311219e-05, -2.706779274112e-05, {   2,   0,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  4.126526437764e-05,  7.807267243595e-06, {   2,  -2,   0,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.673237705591e-05,  3.230953401380e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 {  3.708090081315e-05, -1.194680859898e-05, {   1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.879049164185e-05, -2.027979767171e-07, {   0,   0,   1,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.198063779137e-05, -3.697959711929e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.275310296644e-06,  4.368425213220e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   5,   0,   0,   0 } },
 {  3.866093738911e-05,  1.026313403376e-05, {   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.313391246924e-05,  3.578267961803e-05, {   1,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.260358616286e-05, -1.628400814537e-05, {   2,   2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.479028944452e-06, -4.038898087456e-05, {   2,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.185065522513e-05, -6.971339770040e-06, {   1,  -1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.929746804232e-06,  4.188591372623e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 {  1.888609904699e-05, -2.991830982454e-05, {   2,  -2,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -1.624953121367e-06,  4.716032261892e-05, {   2,   2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.497109159605e-06, -3.928490267937e-05, {   1,  -1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.981117390957e-05, -8.969508525780e-06, {   1,  -1,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -4.869526663223e-05, -2.678102234179e-08, {   3,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.248868264834e-05,  1.622715248441e-05, {   2,  -2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.870046764948e-05, -6.067780840985e-09, {   2,   2,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.914893845879e-05,  1.955629025734e-05, {   2,   0,   1,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -2.914893845879e-05,  1.955629025734e-05, {   2,   0,  -1,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  4.739273899090e-05, -1.304992099565e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -9.201816160964e-06,  3.947568038057e-05, {   1,  -1,  -1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.096680769560e-05,  7.706136249389e-06, {   1,  -1,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.411494204937e-05,  3.452772265550e-05, {   2,   0,  -1,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 {  4.858758166498e-05, -4.214436817133e-08, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0 } },
 {  1.387860992374e-05, -3.474744180244e-05, {   0,   0,   1,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 {  4.266025378809e-05, -5.942803898626e-06, {   2,  -2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.836210629656e-05, -2.252770882885e-07, {   2,   0,  -1,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.733648463445e-05, -2.115540761153e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 {  3.616374574166e-05,  1.228445404762e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -9.166751513728e-06,  3.925660394140e-05, {   1,  -1,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.948862302353e-05,  1.892798882519e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,  -3,   0,   0 } },
 {  4.826532534491e-05,  1.368162440605e-07, {   2,   2,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.085307795191e-05,  7.525406800289e-06, {   2,   2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.544544973719e-05, -3.292281789483e-05, {   3,  -2,   0,   0,   0, -18,  19,   0,   2,  -5,   0,   0,   0 } },
 { -2.987586099027e-05,  1.839713607578e-05, {   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.440906285720e-05, -3.386075871337e-05, {   2,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -4.758072284084e-05, -6.802332687746e-07, {   0,   2,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.625965436726e-05, -3.195888685645e-05, {   2,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  4.997435553940e-06, -4.319234943697e-05, {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 { -4.615914330378e-05,  2.026836209821e-06, {   2,   0,  -1,   0,   0,   4,  -5,   0,   1,   0,   0,   0,   0 } },
 {  3.858341289901e-05,  9.587256922987e-06, {   2,   2,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.800347215912e-05,  1.477616445235e-07, {   3,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.751321342322e-05, -3.061932332171e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.924103287932e-05, -8.884783675700e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -6,   0,   0,   0 } },
 { -3.634762119551e-05, -1.176841214335e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 { -2.962105854378e-05,  1.846885718744e-05, {   2,   0,   0,   0,   0,  -3,  18, -25,   2,   0,   0,   0,   0 } },
 { -9.369377866712e-06,  3.872029425294e-05, {   4,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.164261520827e-05, -2.644420197685e-05, {   1,   1,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -9.395261352284e-06,  3.868980511528e-05, {   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.962722014123e-05,  1.845777492020e-05, {   2,   0,   0,   0,   0,   3, -18,  25,  -2,   0,   0,   0,   0 } },
 { -2.088146558851e-06,  4.599484283079e-05, {   2,   0,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.034059600347e-05, -1.768087143636e-05, {   2,   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  1.607769288162e-05,  3.192481528312e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,   2,   0,   0,   0,   0 } },
 { -4.770053580884e-05, -3.019675115809e-07, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 {  7.722573912889e-06,  4.025946590717e-05, {   5,  -1,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  3.043790705163e-05, -1.751356454464e-05, {   2,   0,   1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.859860377677e-05, -1.935234609223e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 { -2.542158645388e-05,  2.252706874859e-05, {   2,   0,  -1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 {  1.984249370210e-05, -2.809258145915e-05, {   0,   0,   2,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -4.185999827974e-05,  6.048151315609e-06, {   1,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.207970933086e-05,  1.582440579813e-05, {   0,   0,   2,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.252398809036e-05, -1.531227687617e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -8,   0,   0,   0 } },
 {  3.321285392974e-05, -1.461544525962e-05, {   1,  -1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.397975887923e-05, -2.384165333129e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 {  4.752031779603e-05,  2.980323151942e-07, {   4,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.539251534073e-05, -3.241288225398e-05, {   2,   0,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 {  4.106223458558e-05,  6.716113029918e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -6,   3,   0,   0 } },
 {  1.156743748608e-05,  3.620611289052e-05, {   4,   0,  -2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.425037422120e-05, -3.350962277706e-05, {   2,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -9.906999723906e-06, -3.785044051610e-05, {   1,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -9.810832452091e-06, -3.787720798764e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 {  7.955272389165e-06,  3.971704681250e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   1,  -2,   0,   0,   0 } },
 { -3.825882654516e-05,  9.408779386880e-06, {   1,  -1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.762177169972e-05,  1.003784273796e-08, {   4,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.267226487369e-05,  4.956281384204e-06, {   1,   1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.539597020901e-06, -4.308211702316e-05, {   4,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.859599538882e-05, -2.901710905671e-05, {   1,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.758199692531e-05,  1.984909035839e-08, {   4,   2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.159135736257e-05, -2.598727519136e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   3,   0,   0,   0 } },
 {  3.999159704008e-05,  7.569650686827e-06, {   2,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.479279439736e-05, -3.271119883512e-05, {   0,   0,   1,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 {  3.716864207524e-05,  1.023856100530e-05, {   2,  -2,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -4.150919468512e-05,  5.887651241991e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   6,  -3,   0,   0 } },
 { -3.174771022393e-05,  1.564803317263e-05, {   2,   0,  -2,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.515731901848e-06, -4.187785330780e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  4.733775804661e-05,  4.170923398800e-08, {   1,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.598998484880e-05, -2.138871788432e-05, {   2,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  1.848209406050e-05, -2.887995352243e-05, {   2,   0,  -1,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 {  1.469090524629e-05,  3.266311541007e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 {  3.665730953400e-05, -1.069621444733e-05, {   3,   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.194928353752e-05, -2.535852253128e-05, {   4,   0,  -2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -4.700343134898e-05,  2.968732573008e-07, {   2,   0,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.151912763748e-05, -1.574219598841e-05, {   4,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.739455035200e-06, -4.051976855165e-05, {   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.504301747268e-06, -3.775144126400e-05, {   3,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.686106649750e-05, -3.939040459010e-07, {   1,   0,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 {  1.694097994501e-05,  3.029358081147e-05, {   4,   0,  -2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.774224027599e-06, -4.545639620459e-05, {   1,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.942642054812e-05, -2.780054953392e-05, {   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.677570629853e-06,  4.054227252772e-05, {   0,   0,   2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.415457813991e-05, -3.303024990360e-05, {   1,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.893949517997e-05, -2.824438563024e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 {  4.684473450361e-05, -3.342191853879e-07, {   0,   0,   2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.489440998526e-06,  3.967226496549e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 { -4.710708310463e-05,  3.630171795646e-08, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  3.971395303032e-05, -7.426967732811e-06, {   3,   1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.712245823774e-05, -6.069273263456e-09, {   0,   0,   1,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 {  2.419117388678e-05,  2.292111326889e-05, {   4,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -9.108178574858e-06,  3.799131145210e-05, {   1,   1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.449390422216e-05,  3.259764121481e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 {  1.449284096946e-05,  3.259797511923e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -4.033408819117e-05, -6.746449281234e-06, {   2,   0,  -3,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.255098988005e-05,  2.449770801612e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -5,   0,   0,   0 } },
 {  2.254733752410e-05, -2.446710164118e-05, {   0,   0,   0,   0,   0,   0,  10, -19,   1,   0,   0,   0,   0 } },
 { -9.089252115745e-06,  3.788694410606e-05, {   2,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.032526815275e-05, -3.659907467508e-05, {   0,   0,   1,   0,   0,   3,  -3,  -4,   3,   0,   0,   0,   0 } },
 {  1.032526815275e-05, -3.659907467508e-05, {   0,   0,   1,   0,   0,  -3,   3,   4,  -3,   0,   0,   0,   0 } },
 {  1.630105968494e-05,  3.061319368429e-05, {   0,   2,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  4.688378535272e-05, -2.375192175156e-08, {   0,   2,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.673256909801e-06, -3.719088105349e-05, {   1,  -1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.648341268995e-05,  3.037807768697e-05, {   0,   2,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  3.494668791405e-05,  1.191289364665e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  3.031128975261e-05, -1.653278461236e-05, {   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.474974803898e-05,  2.077597671647e-06, {   2,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.022169025011e-05,  1.659975513612e-05, {   1,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  4.538488187791e-05, -1.430450675537e-06, {   2,   0,   0,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 {  8.767983448453e-06,  3.801103089904e-05, {   3,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.234403487912e-05, -4.434037200314e-06, {   0,   0,   1,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 { -4.453775828500e-05,  2.239036929787e-06, {   2,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  4.453775828500e-05,  2.239036929787e-06, {   2,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  3.408975869028e-05, -1.265985255755e-05, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   1,   0,   0,   0 } },
 {  2.345659867103e-06,  4.432385427086e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  4.228872762020e-05, -4.361389115630e-06, {   0,   0,   1,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 {  4.653746350129e-05,  9.964955877358e-08, {   0,   2,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.348367465309e-05,  3.315100518147e-05, {   4,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.659677776250e-05, -1.034643765434e-09, {   4,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.536134834498e-05, -2.122040357670e-05, {   0,   2,  -1,   0,   0, -17,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.474247599568e-05,  3.177334472067e-05, {   2,   0,  -1,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0 } },
 {  4.553755641832e-05,  9.145724381821e-07, {   1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  3.520215186189e-05,  1.123913982684e-05, {   2,  -2,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.097005545672e-05, -1.546812210046e-05, {   4,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.382496452666e-06, -4.505434238063e-05, {   2,   0,  -1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.992210399683e-05, -6.499452492476e-06, {   0,   0,   1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.166019879398e-05, -1.474024265079e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 { -6.478581564753e-06, -3.989719540183e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  4.629374236116e-05,  1.838653338533e-08, {   1,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.299216059415e-05, -2.331519479776e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   1,  -5,   0,   0,   0 } },
 {  1.748787453408e-06, -4.455769382974e-05, {   0,   0,   1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.625239213541e-05, -4.480301632060e-08, {   0,   0,   1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  4.620603751646e-05, -7.695166942738e-08, {   0,   0,   2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  8.849978449694e-06, -3.742802395356e-05, {   0,   0,   3,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.790759892970e-05, -8.357657869291e-06, {   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.923826659771e-05, -1.702202921085e-05, {   2,   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.501865219655e-05, -1.221749088248e-06, {   2,   0,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.614276698641e-05,  6.759144038610e-08, {   2,   0,   2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.519155754841e-05,  3.101730416723e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 { -6.109505638901e-06, -4.005781957719e-05, {   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.114472178043e-05,  5.012689789084e-06, {   2,   0,  -1,   0,   0,   0,  13, -19,   0,   0,   0,   0,   0 } },
 { -4.612602305057e-05,  3.129863753973e-08, {   0,   0,   4,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.762237343963e-05,  1.852748902979e-05, {   2,   0,  -2,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  4.530681893388e-05, -7.596111687316e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -4.529488133135e-05, -7.668250803892e-07, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 { -2.330690906717e-05,  2.275278913587e-05, {   0,   0,   0,   0,   0,   0,   1,  -4,   3,   0,   0,   0,   0 } },
 {  7.768861050656e-06, -3.828323424721e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  1.975103237803e-05, -2.629314520783e-05, {   1,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.704235151638e-05,  8.995818284942e-06, {   3,  -1,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -8.788240413146e-06, -3.724659852351e-05, {   1,  -1,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.617646159319e-06,  3.741699057732e-05, {   1,   1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.585676538820e-05, -1.699338091745e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 {  4.176073581299e-05, -4.258724484837e-06, {   0,   0,   1,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 { -8.859516220634e-06,  3.713749848855e-05, {   1,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.348225060575e-06, -3.957550684777e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -6.488112548415e-06,  3.941570158746e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 {  4.575101018727e-05, -1.375265194105e-07, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.046142737107e-05,  2.542280252648e-05, {   1,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -3.699904126504e-05,  8.866219997019e-06, {   1,   1,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.671151491368e-05,  9.118826167373e-06, {   2,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.495253262085e-05, -3.086494901618e-05, {   0,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.060608553777e-05, -1.520351209985e-05, {   1,  -2,   1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.060608553777e-05, -1.520351209985e-05, {   1,  -2,  -1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.395603121627e-05, -1.185240125440e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -1.612297402493e-05, -2.963971928268e-05, {   1,  -1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  7.169768034371e-06, -3.857619931514e-05, {   2,   0,   0,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.691028184030e-06, -3.903969959403e-05, {   0,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.666424103441e-05, -9.049875802021e-06, {   0,   0,   2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.867197409353e-05, -1.697508328049e-05, {   2,   0,  -3,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  9.494309678245e-06, -3.611144735054e-05, {   4,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.362447305403e-05,  1.980936143041e-06, {   2,   0,   0,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 {  1.455307036784e-05, -3.102771321909e-05, {   0,   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.234101153819e-05, -3.321481030528e-05, {   2,   0,  -1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.628858676018e-05,  2.925751313189e-05, {   1,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.524935642859e-05, -2.862683523082e-07, {   0,   2,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.513658222797e-05,  3.034809440569e-05, {   2,   0,  -1,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 { -3.202723128279e-05, -1.343476829770e-05, {   2,   0,  -1,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.612295345116e-05, -2.933253505598e-05, {   1,   1,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -6.759204873337e-06,  3.868106546266e-05, {   1,  -1,   1,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -6.759204873337e-06,  3.868106546266e-05, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -4.537115830337e-05, -5.031274618622e-08, {   1,   0,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.330266410557e-05,  2.211341280270e-05, {   2,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  4.404895507642e-05, -1.357795089940e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -6,   0,   0,   0,   0 } },
 { -1.433518854130e-05, -3.104710277794e-05, {   2,   0,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  3.955751984882e-05, -5.781942732711e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  3.531776460519e-05,  1.001606783097e-05, {   2,   0,  -1,   0,   0, -14,  11,   0,   3,   0,   0,   0,   0 } },
 {  3.074559545475e-05, -1.454006763113e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   3,  -2,   0,   0,   0 } },
 {  3.012835994362e-05,  1.515314183470e-05, {   0,   0,   1,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.770423282920e-05,  2.756463781032e-05, {   2,  -2,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -1.723398982082e-05,  2.803030859884e-05, {   0,   0,   1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -4.456327121934e-05,  6.813246146904e-07, {   1,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.151263554308e-07, -4.501895887253e-05, {   4,   0,  -3,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.666898439207e-06,  3.955718238229e-05, {   2,   0,   0,   0,   0,  -3,   4,   0,  -2,   0,   0,   0,   0 } },
 {  4.384932503136e-05, -1.362347655983e-06, {   6,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.058856335983e-05,  1.462055544377e-05, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 { -4.243402275501e-05, -2.756779488106e-06, {   0,   0,   1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  2.904427872985e-05, -1.613263379796e-05, {   2,   0,  -1,   0,   0, -16,  14,   1,   0,   0,   0,   0,   0 } },
 { -4.515031060706e-05, -2.369799225363e-08, {   1,   0,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.782246163051e-05,  2.732735250185e-05, {   3,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -2.446670194763e-05, -2.067620941125e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0 } },
 {  1.036926678416e-05,  3.476360148993e-05, {   2,  -1,   0,   0,   0, -15,  13,   0,   1,   0,   0,   0,   0 } },
 { -3.759043698052e-05, -7.537168500363e-06, {   1,   1,  -1,   0,   0, -21,  20,   3,   0,   0,   0,   0,   0 } },
 {  4.096897511138e-05,  4.145459061753e-06, {   1,   1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.504006733673e-05,  4.059752925077e-08, {   3,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.472673314452e-05,  3.474308996527e-07, {   2,  -2,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -9.570041756725e-06, -3.549027630265e-05, {   1,  -3,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.532539398655e-05, -1.970966738802e-05, {   2,   0,  -1,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 { -2.998992001619e-05,  1.501365997443e-05, {   0,   2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.730300562027e-06,  4.222169536165e-05, {   0,   0,   2,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.602820975563e-05, -2.889881789224e-05, {   2,   0,  -1,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  8.528212378520e-06,  3.638416378486e-05, {   1,   1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.717348069458e-05, -2.773468223510e-05, {   1,  -1,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -2.899301537682e-05,  1.589165431345e-05, {   4,   0,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  2.695419130453e-06, -4.217151833267e-05, {   4,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.483612008997e-05, -1.617641116464e-08, {   0,   0,   4,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.787019408105e-05,  1.697071049588e-05, {   0,   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 {  4.480017133211e-05,  2.328878823083e-08, {   6,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.687017401924e-05, -1.792578092900e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 { -4.091349700275e-05, -3.865326876179e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -1.022458723948e-05,  3.453786165530e-05, {   4,   0,  -2,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.518308626272e-06, -3.623569722873e-05, {   2,  -2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.594211020338e-05, -2.880274727527e-05, {   2,   0,  -1,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 {  1.089101810616e-05,  3.385254845243e-05, {   0,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.120852045710e-06, -3.961148663022e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.187887955372e-05, -3.277096094207e-05, {   2,  -2,   1,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 { -1.187887955372e-05, -3.277096094207e-05, {   2,  -2,  -1,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 { -4.072397554340e-06,  4.057470791934e-05, {   0,   0,   2,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.815925472834e-05, -1.648630730871e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -7,   0,   0,   0,   0 } },
 {  2.462736013770e-05,  2.000786453768e-05, {   4,   0,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.002491808451e-05, -3.457821389979e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 {  1.758279241772e-05,  2.700500117205e-05, {   0,   2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  3.265450305230e-05, -1.191597654105e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -4.067302870654e-05,  3.895863255452e-06, {   0,   0,   0,   0,   0,   3,  -7,   4,   2,  -5,   0,   0,   0 } },
 {  7.687118894436e-06, -3.687319027018e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.922725663135e-05, -1.529540481832e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 { -1.994426146253e-05, -2.455185438162e-05, {   1,   1,  -1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.116114266754e-05, -1.331630614883e-05, {   2,   0,  -1,   0,   0,   0,  13, -20,   0,   0,   0,   0,   0 } },
 { -4.714260777762e-06,  3.976162924405e-05, {   2,   0,  -1,   0,   0,  -3,   4,   0,  -2,   0,   0,   0,   0 } },
 {  3.613801785439e-05, -8.321735576590e-06, {   0,   0,   0,   0,   0,   6, -10,   0,   2,   0,   0,   0,   0 } },
 {  3.185362715816e-05,  1.256746213323e-05, {   2,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 {  3.480289360054e-05,  9.599618711046e-06, {   4,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.743404776763e-06, -4.262874887224e-05, {   0,   0,   1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  4.427205976131e-05,  8.821652479260e-08, {   2,   0,   1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.340964465342e-05, -9.054175173052e-07, {   4,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.886579039379e-05,  2.543228784897e-05, {   1,  -1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.425157187840e-05, -1.767229974641e-08, {   1,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.623337858987e-05, -8.023965759221e-06, {   3,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.767427631546e-05,  6.573816811776e-06, {   0,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.016439187456e-05,  4.073995194958e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 {  4.869957761925e-06,  3.936060519752e-05, {   2,   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  2.583694856458e-05, -1.839054127301e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   0,   5,   0,   0,   0 } },
 {  2.821589590402e-05, -1.600274482591e-05, {   4,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -7.580837120433e-06, -3.663300605551e-05, {   0,   0,   2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -9.525744606522e-06, -3.466622436683e-05, {   1,  -1,  -1,   0,   0,  28, -31,   0,   0,   0,   0,   0,   0 } },
 { -9.991553621668e-06,  3.416108560020e-05, {   2,   0,  -2,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 {  1.346760676601e-05, -3.067685994226e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   4,   0,   0,   0 } },
 { -2.488494309756e-05, -1.925567162554e-05, {   2,   0,  -2,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 {  2.743675417781e-05,  1.668016184629e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0 } },
 {  3.654296774634e-05,  7.565754861060e-06, {   2,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.070081370200e-05, -3.406843819855e-06, {   1,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.497257308073e-05, -2.913127355488e-05, {   2,   0,  -2,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  2.303097232958e-06, -4.179722024854e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -1.445487638285e-05, -2.963312242928e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   5,  -3,   0,   0,   0 } },
 {  1.283361489377e-05,  3.124988061601e-05, {   1,  -1,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.382131693336e-05, -2.198350718890e-07, {   2,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.942976650896e-05, -2.459103657369e-05, {   2,  -1,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.923233924857e-05,  1.473362913428e-05, {   0,   0,   1,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.558251271008e-05,  8.382298690189e-06, {   0,   1,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  8.205610311765e-06,  3.575806312942e-05, {   1,   1,  -2,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.005354768951e-05, -2.390717510495e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  8.527845007297e-06, -3.542939944483e-05, {   2,   0,   2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.088983498090e-05,  3.058213703293e-06, {   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.299841450374e-06,  4.262825793357e-05, {   2,  -2,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.714603089885e-05,  6.777714827312e-06, {   2,   2,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.351689030693e-05, -3.040443070190e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.338310524000e-05, -1.053370407004e-05, {   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -8.397234613382e-06, -3.546990576828e-05, {   2,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.102835578709e-05, -2.828121505454e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -1.051636276817e-05, -3.333100290517e-05, {   2,   0,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.465834466328e-06,  4.237973106307e-05, {   1,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.141637765394e-05, -3.238619447864e-05, {   1,  -1,   2,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -8.329299180523e-06, -3.546916660383e-05, {   1,  -1,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.093902684344e-05, -2.284668418119e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   2,   0,   0,   0,   0 } },
 { -8.381802804185e-06,  3.538591941731e-05, {   3,  -1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.281199217146e-07,  4.292252077133e-05, {   4,   0,  -3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.204574524429e-05, -1.704121626152e-06, {   2,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.338054260066e-06,  3.639793488349e-05, {   2,   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -3.776343432531e-05, -5.968619947594e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -8.340431391309e-06, -3.539045116440e-05, {   1,  -1,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.934306458781e-05, -1.436177853346e-05, {   0,   0,   0,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 {  8.244943009246e-06,  3.545987902180e-05, {   3,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.238229658114e-05,  3.127491498536e-05, {   1,  -1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.429106758432e-05,  9.366077547166e-06, {   1,  -1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.084949803445e-05,  2.279679921378e-05, {   4,   0,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  1.589540900899e-05, -2.774117083773e-05, {   1,  -1,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -9.503450379467e-06,  3.413178633552e-05, {   2,   2,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  9.503162350557e-06,  3.413141458019e-05, {   2,   2,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.265494758529e-05,  3.096045550849e-05, {   2,   0,   0,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 { -1.627133211808e-05,  2.733227297188e-05, {   3,  -1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.928538763988e-05,  1.429458903229e-05, {   2,   0,  -1,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.526910548525e-06,  4.104953097570e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 } },
 { -4.013563468889e-06,  3.956220671585e-05, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 {  3.311143130128e-05, -1.045165437771e-05, {   3,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.313763909079e-05, -2.041873888080e-05, {   0,   0,   0,   0,   0,   1,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.020094006155e-05, -1.335355043895e-05, {   1,   1,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.909100935704e-05,  1.445304655353e-05, {   0,   0,   1,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.198842565755e-06, -3.433322184945e-05, {   2,   0,   0,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 { -9.477214206853e-06,  3.403923044068e-05, {   1,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  4.326681364431e-05,  2.473129494890e-07, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  9.476666671339e-06,  3.403449946961e-05, {   1,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -8.784036498268e-06, -3.471697217325e-05, {   1,   1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.153947994871e-06, -3.530508948254e-05, {   1,  -1,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  9.908090411260e-06, -3.353356344580e-05, {   6,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.580581908236e-06,  3.983942059316e-05, {   1,   1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.580095177625e-06,  3.983946433524e-05, {   1,  -1,  -2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.844462662688e-07, -4.263501544931e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 { -3.579566726644e-06, -3.983951181992e-05, {   1,   1,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.578301223048e-06,  3.983962550473e-05, {   3,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.578217783124e-06, -3.983963299902e-05, {   2,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.578217783124e-06, -3.983963299902e-05, {   1,   1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.239825785611e-05,  2.097653277134e-05, {   4,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  3.281349451994e-06,  4.004159459540e-05, {   4,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.452995634940e-07, -4.296656180836e-05, {   2,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.475587165440e-05,  8.548149411905e-06, {   1,   1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  2.976009301470e-05,  1.353894173712e-05, {   1,  -1,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.053574073425e-06,  4.123380962416e-05, {   0,   0,   1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  4.322732378106e-05,  1.280271540872e-08, {   2,   2,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.553093464827e-05,  7.678387158321e-06, {   2,   0,   1,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 { -3.553093464827e-05,  7.678387158321e-06, {   2,   0,  -1,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 {  4.304973987870e-05,  1.438674985201e-07, {   2,   0,  -3,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.806934673335e-05,  5.123041490549e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -1.375883864567e-05, -2.942017921173e-05, {   2,   0,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  3.109895804416e-05, -1.207314586118e-05, {   1,   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  3.587726768197e-05,  7.272132087360e-06, {   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.648313675562e-05,  2.666052274350e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 { -3.571748874842e-05, -7.413831486680e-06, {   1,  -1,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.283744215301e-06, -3.984152368593e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.404089309140e-05, -9.069945365006e-06, {   4,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.912581059017e-05, -2.398471543309e-05, {   4,  -2,  -1,   0,   0,  -7,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.197296589095e-05,  1.109973630141e-05, {   2,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.828866691759e-05,  1.478358233077e-05, {   2,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.872509285594e-05,  1.434638087350e-05, {   6,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.872469571277e-05, -1.434660171726e-05, {   2,  -2,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.287393539606e-06, -3.474419633345e-05, {   1,   1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.943393621640e-05, -2.359405092080e-05, {   0,   2,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.492657823672e-06,  3.852681089600e-05, {   2,   0,  -2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -1.270450755362e-05, -3.030581928040e-05, {   2,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  4.286472380874e-05, -1.244422402738e-07, {   2,   0,   1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.531264667135e-05,  1.766243388585e-05, {   2,   0,  -1,   0,   0,  -4,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.849528212173e-06,  3.308936053794e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 {  3.669024859018e-05,  6.236559707458e-06, {   1,   1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.488364756097e-05,  2.802782239778e-05, {   2,   0,   0,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 { -1.828518611293e-06,  4.106142334994e-05, {   0,   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  3.444677875531e-05,  8.420081035326e-06, {   2,   0,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.604895933690e-05,  2.673694409592e-05, {   0,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  2.860464116974e-05, -1.417539825673e-05, {   1,   0,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.392972956262e-06, -3.438339030491e-05, {   2,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.349925869843e-05,  2.926468475359e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 {  3.795693664530e-06,  3.896160905837e-05, {   0,   0,   1,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 { -8.797087519542e-06,  3.394995187803e-05, {   1,   1,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.846243949987e-05, -1.427774197802e-05, {   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.876073793146e-05,  1.396107461862e-05, {   2,  -2,  -1,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 {  3.758823557535e-05,  5.133033849466e-06, {   0,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.602592516951e-05, -1.668521722466e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 {  2.448713886490e-05,  1.819757094662e-05, {   0,   0,   1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  1.768599549919e-05, -2.499857356406e-05, {   1,  -1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.514696071431e-05, -1.752099652758e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,  -5,   0,   0,   0 } },
 { -3.435229193408e-06, -3.923036832210e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.412069733018e-05, -2.852274695645e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   5,   0,   0,   0 } },
 {  4.050106078212e-05, -2.127450746178e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 {  3.633216758693e-05,  6.280284855458e-06, {   2,   0,   0,   0,   0,   0,   6,   0,  -7,   0,   0,   0,   0 } },
 {  3.388350627194e-05,  8.716593318260e-06, {   2,   0,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -1.213501797609e-05, -3.043934050031e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.769929815429e-05, -1.487397592195e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 {  3.871730624493e-05, -3.846443433016e-06, {   2,   2,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.156186742284e-07, -4.243802240504e-05, {   5,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.202038112342e-05,  5.250475633848e-07, {   0,   0,   4,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.390885755448e-05, -1.863356007125e-05, {   0,   0,   2,   0,   0, -23,  22,   4,   0,   0,   0,   0,   0 } },
 { -4.046213609168e-05,  2.066203764160e-06, {   6,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.683497811386e-05,  5.671721545591e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -3.683497811386e-05,  5.671721545591e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -3.419739186594e-05,  8.306653551724e-06, {   3,  -1,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  2.055446091701e-05,  2.191867017857e-05, {   0,   0,   1,   0,   0,   0,  -2,   8,  -3,   0,   0,   0,   0 } },
 {  4.241053870876e-05,  6.151834404956e-08, {   2,   0,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.783526564050e-06,  3.866255006132e-05, {   2,   2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.238916718010e-05, -4.823324843223e-08, {   1,   0,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.219507362086e-06, -4.119446879228e-05, {   2,   0,  -1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.821686570452e-05, -1.418057647189e-05, {   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.140169500800e-06, -4.025053812679e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.120095878628e-05, -3.118243269714e-05, {   2,   0,   1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.311073573366e-05, -2.924348317829e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 {  8.873245339005e-06, -3.344202361708e-05, {   0,   2,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -2.473989193397e-05, -1.755454729319e-05, {   1,  -1,  -1,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.508348337119e-05,  7.208587755863e-06, {   2,   0,   0,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  3.774714603024e-05, -4.537596492752e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 { -3.565514055237e-05, -6.571304420556e-06, {   2,  -2,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  3.597873929209e-05, -6.247021423120e-06, {   2,   0,  -1,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 {  2.238206265691e-05,  1.982290537069e-05, {   2,   0,  -2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  2.328193376503e-05,  1.891960782259e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.983950431044e-05,  2.232360829521e-05, {   0,   0,   0,   0,   0,   8,  -5, -16,   4,   5,   0,   0,   0 } },
 { -1.691175682150e-05, -2.524564713303e-05, {   1,  -1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.856199397336e-05, -2.356803724821e-05, {   2,  -2,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.312509869470e-05, -1.900094329220e-05, {   2,   0,  -1,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.429739350145e-05, -7.827129990704e-06, {   1,  -1,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -2.719470133326e-05,  1.490902464388e-05, {   2,   0,  -2,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.208969300240e-05,  3.379474941009e-09, {   0,   2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.157660119527e-05, -5.063383468055e-07, {   2,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.314349772838e-05, -2.893171509731e-05, {   2,   0,  -2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  2.304320577174e-06,  3.975774838834e-05, {   2,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.735653238990e-05,  1.468281368277e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 {  2.521839944822e-05, -1.680508449311e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 {  3.960164298864e-05,  2.403182470207e-06, {   0,   0,   0,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.195657166937e-05,  1.004165213413e-05, {   0,   0,   2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -6.702930691836e-06, -3.528789174865e-05, {   3,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.224362326737e-05,  2.973395889161e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  3.428349916562e-05, -7.685049251791e-06, {   2,   0,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 {  1.786230185508e-05, -2.410265903252e-05, {   2,   0,  -3,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.193095075091e-05, -2.460234182880e-08, {   2,   0,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.344138967787e-06, -3.958365506289e-05, {   2,   0,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.422258179404e-05, -1.769933702801e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.608250868106e-06,  3.229660722474e-05, {   0,   0,   3,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.014478323214e-05, -2.174208901509e-05, {   2,   0,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.851588563021e-06, -3.902369540336e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.630848474452e-05, -5.559665168915e-06, {   2,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.156883814454e-05, -2.973855460775e-07, {   4,   0,  -2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  8.517328610572e-06, -3.334396832262e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 {  2.032981497299e-05, -2.153102841116e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0 } },
 { -2.439689622489e-05, -1.745827753794e-05, {   4,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.567953884304e-05, -1.616666755922e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   5,   0,   0,   0 } },
 {  2.922295376231e-05,  1.262153218983e-05, {   4,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  3.495559117970e-05,  6.888507444008e-06, {   1,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.073576905772e-05,  1.108603474200e-05, {   0,   2,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.661665306356e-05,  1.518948160191e-05, {   0,   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.401368859388e-05, -7.788987865779e-06, {   2,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 {  2.245452347522e-05, -1.933955474228e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  1.379147695787e-05,  2.798652441084e-05, {   0,   0,   0,   0,   0,   6,  -9,   0,   3,   0,   0,  -1,   0 } },
 {  1.878122189910e-05, -2.297353481162e-05, {   3,  -1,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  4.160752951714e-05, -1.441863533008e-07, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  2.927194158713e-05,  1.247441336282e-05, {   2,   0,  -1,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 { -1.675282633023e-05, -2.495297131457e-05, {   1,  -1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -4.166072483249e-05, -4.004158008199e-08, {   1,   0,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.146591568169e-05, -2.347593033687e-07, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.733879159977e-05, -1.434285534993e-05, {   1,  -1,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.683378695014e-05, -2.483190723479e-05, {   1,  -1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.611950266108e-05, -1.554100386143e-05, {   4,   0,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -7.948212915956e-06,  3.369938315695e-05, {   1,  -1,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.191930943962e-05,  1.972212860097e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  4.095932132982e-05, -6.701318406635e-07, {   2,  -2,  -1,   0,   0,  28, -30,   0,   0,   0,   0,   0,   0 } },
 {  3.511319822064e-05, -6.491288185993e-06, {   3,  -1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.705939324782e-05, -2.451418026212e-05, {   2,   0,   0,   0,   0,   0,   1,  -8,   3,   0,   0,   0,   0 } },
 {  3.335233930131e-05,  8.100225049445e-06, {   1,   1,   1,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 { -4.117357407791e-05,  2.602181581357e-07, {   0,   2,   0,   0,   0, -43,  43,   0,   0,   0,   0,   0,   0 } },
 {  4.123937463576e-05, -1.776919976278e-07, {   2,   0,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.587476905997e-05,  5.528816815801e-06, {   1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.311503678483e-05,  8.279740988935e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   2, -10,   0,   0,   0 } },
 { -9.910345031131e-06,  3.144985788164e-05, {   2,   0,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.087599302656e-05, -3.047793503385e-05, {   1,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.096576939910e-06,  3.524067150616e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.564543016863e-05, -2.567435794133e-05, {   1,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.508713988435e-05,  1.622658085311e-05, {   2,   0,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  3.911628581293e-05,  2.179176353314e-06, {   4,   0,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -3.331048189655e-05, -7.966817299214e-06, {   4,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.452400150582e-05, -6.746214502090e-06, {   1,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.571247853132e-05,  1.555039159978e-05, {   2,   2,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.058903284393e-05, -6.617529153085e-07, {   2,   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  3.477075529740e-05, -6.477574325190e-06, {   2,   0,   0,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 { -1.480603940642e-06, -3.976604432237e-05, {   1,   1,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  3.943626996333e-05, -1.801869046109e-06, {   4,   0,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  2.972689786386e-05,  1.147489407774e-05, {   0,   0,   0,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 { -2.693809028267e-05, -1.425653016103e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  3.785570813267e-05,  3.338567185746e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  2.682303418169e-05,  1.435641575460e-05, {   1,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.573780803263e-05, -1.541315145177e-05, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.169965865987e-05, -1.944139646787e-05, {   0,   0,   0,   0,   0,  12,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.207940936616e-06, -3.993014677726e-05, {   2,   0,  -1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.576163302312e-05, -1.537329710837e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.228121330926e-05, -2.883722417326e-05, {   2,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.122591405058e-05,  9.889895878726e-06, {   1,  -1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.299971437293e-06, -3.880915918732e-05, {   2,   0,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 { -7.900030499667e-06,  3.320564138742e-05, {   1,  -1,   2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.073445954315e-05, -3.705312555524e-07, {   2,   0,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  3.428783257685e-05, -6.813909335388e-06, {   1,   1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.444780833564e-06,  3.865049956259e-05, {   2,   0,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.420712871756e-05, -6.878992960988e-06, {   2,   0,   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 { -4.033281885961e-05,  7.505658574773e-07, {   3,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.121498489594e-06, -3.194375980291e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -4.098818799915e-05,  7.169296572907e-08, {   2,  -2,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.066949910398e-06,  3.699274777753e-05, {   4,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.414406380549e-05, -6.901413146831e-06, {   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.160240612178e-05,  9.437336441874e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -8.313646295727e-06,  3.272490669713e-05, {   1,   1,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.404491987026e-05,  6.983376790615e-06, {   2,   0,  -2,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 {  1.782510891831e-05,  2.318520968801e-05, {   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.887569172523e-06,  3.511788089565e-05, {   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.369606754678e-05,  2.730935570477e-05, {   1,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  2.877176433564e-05,  1.222508607791e-05, {   2,   2,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.877176433564e-05,  1.222508607791e-05, {   2,   2,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.401996348672e-06,  3.354005606046e-05, {   2,   0,  -1,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.678313813875e-05, -4.157391720489e-06, {   1,  -1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.315078808185e-06,  3.662162528006e-05, {   2,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.013521812745e-06,  3.391378254106e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -8,   0,   0,   0 } },
 { -1.486481975823e-05,  2.605834095938e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.662487415068e-05,  2.428145066067e-05, {   0,   2,  -2,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 {  2.981553765750e-05,  1.108986890060e-05, {   0,   2,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.438122055913e-05,  6.505695467512e-06, {   2,  -2,   2,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -3.321070416617e-05,  7.661974890106e-06, {   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.001006086971e-05, -2.085211621113e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -5,   0,   0,   0 } },
 {  7.870036295750e-06,  3.299187107609e-05, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 { -2.560559299205e-05, -1.525555363751e-05, {   2,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 { -3.382358751579e-05,  7.014957075820e-06, {   0,   0,   0,   0,   0,   2,  -5,   3,   0,   0,   0,   0,   0 } },
 {  1.058701020399e-05, -3.022723286097e-05, {   2,   0,   0,   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0 } },
 {  1.069345191818e-05,  3.011701969555e-05, {   4,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.635568828742e-05, -1.445338827373e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -8.119187435056e-06, -3.268798077060e-05, {   1,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.539563322015e-05,  2.541081916217e-05, {   4,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.069110490908e-05,  3.010964909996e-05, {   4,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.360114941542e-05,  2.719308054189e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 { -2.988500438290e-05,  1.089970912456e-05, {   2,   0,  -2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.543033016250e-05, -2.534512735713e-05, {   2,   0,   0,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 { -3.455610450148e-05, -6.213419046833e-06, {   2,   0,  -2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  4.071796668558e-05,  4.705639617714e-08, {   3,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.028340285434e-05, -4.705049838989e-07, {   2,   0,  -1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 { -4.386199854499e-07, -4.030312789235e-05, {   4,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.486186358358e-06,  3.224131207951e-05, {   1,  -1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.617037088600e-06,  3.910960458898e-05, {   2,   0,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -3.379121880300e-05,  6.890648430178e-06, {   1,  -1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.452853268556e-06, -3.922784077624e-05, {   2,   0,  -2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.143100807038e-05, -2.924457722831e-05, {   2,   0,  -1,   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0 } },
 { -4.066328047790e-05, -4.866142750631e-09, {   2,  -2,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -3.426407975198e-05, -6.395950165885e-06, {   0,   0,   3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.069430469229e-06, -3.958159545538e-05, {   2,   0,  -1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.987018834571e-05, -1.077600746025e-05, {   2,  -2,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.807107662307e-05, -1.253938903638e-05, {   0,   0,   0,   0,   0,   0,   2,  -4,   3,   0,   0,   0,   0 } },
 { -7.789690171338e-06, -3.281785104937e-05, {   1,  -1,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.306890751488e-05,  7.520803352656e-06, {   6,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.285640579069e-05, -1.768863076949e-05, {   2,   0,   0,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 {  2.345732426468e-05,  1.708160338495e-05, {   0,   0,   2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.112670973981e-05, -2.939727146679e-05, {   2,   0,   0,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  2.701828708477e-05,  1.349441475670e-05, {   6,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.908675999987e-05,  2.140625831383e-05, {   1,   0,  -1,   0,   0,   0,  33, -41,   2,   0,   0,   0,   0 } },
 { -3.575256540630e-06,  3.691679215225e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 {  2.525682026695e-05, -1.522585804031e-05, {   0,   0,   1,   0,   0,  12,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.225067999518e-05, -8.228937226789e-06, {   2,   0,  -1,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  1.086348726614e-05,  2.960641320550e-05, {   1,  -1,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.881620623390e-06,  3.857330717544e-05, {   1,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.824025042671e-05,  2.193434977293e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -8.476404270142e-06, -3.193750367090e-05, {   0,   2,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.084497499907e-05,  2.955788623500e-05, {   2,   0,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.221298540929e-06,  3.117086357585e-05, {   4,   0,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.767128923739e-05, -1.271726876494e-05, {   1,  -1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  7.857655648435e-06,  3.252531818013e-05, {   2,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.460137121617e-06, -3.292036255310e-05, {   2,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.785870251215e-06, -3.858381733525e-05, {   1,  -1,   1,   0,   0, -18,  16,   2,   0,   0,   0,   0,   0 } },
 { -3.077661443366e-05,  9.576851143803e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0 } },
 {  1.653271221615e-05,  2.380517857781e-05, {   3,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.381779483567e-07,  3.999857042039e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.914968976632e-05, -1.116065968532e-05, {   1,  -2,   0,   0,   0,  31, -36,   0,   1,   0,   0,   0,   0 } },
 {  1.233335868216e-05,  2.796165566839e-05, {   0,   0,   0,   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0 } },
 {  2.679411934760e-05, -1.347653737931e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  3.486749464376e-05, -5.394264147426e-06, {   4,   0,  -2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.986963703350e-05,  3.670217953177e-07, {   0,   0,   3,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  7.687120653821e-06, -3.251895359935e-05, {   1,   1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.507165200529e-06,  3.268971826523e-05, {   3,  -1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.676209076618e-05, -1.336613706543e-05, {   2,  -2,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.877958118435e-06,  3.124611494230e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -8.289984667543e-06,  3.183061450430e-05, {   2,   0,  -2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.980543990317e-05, -2.029712349020e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -2.342247765645e-05, -1.667544775608e-05, {   2,   0,  -1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 {  3.898202087195e-05, -1.111699352608e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -1.507778989162e-05,  2.501029707672e-05, {   2,   0,  -1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 { -2.179754326329e-05,  1.826993939946e-05, {   2,   0,  -1,   0,   0, -22,  21,   7,  -8,   6,   0,   0,   0 } },
 {  1.150176132077e-06,  3.889361305716e-05, {   3,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.754701957914e-05,  2.485426604500e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 } },
 {  3.986707026268e-05,  1.623952549462e-07, {   2,   0,   1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.999999873975e-05,  7.099999247983e-09, {   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.999999998629e-05, -1.047197551072e-09, {   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -3.999999998857e-05,  9.564404300018e-10, {   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999998938e-05,  9.215338449715e-10, {   0,   1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999093e-05, -8.517206749014e-10, {   0,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999108e-05,  8.447393579025e-10, {   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999249e-05, -7.749261878449e-10, {   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999249e-05,  7.749261878370e-10, {   2,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999276e-05, -7.609635538271e-10, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  3.999999999570e-05,  5.864306286491e-10, {   6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999580e-05,  5.794493116419e-10, {   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999639e-05, -5.375614095901e-10, {   4,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999693e-05,  4.956735075537e-10, {   6,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999701e-05, -4.886921905426e-10, {   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999710e-05, -4.817108735330e-10, {   0,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999727e-05,  4.677482395238e-10, {   0,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999802e-05, -3.979350694554e-10, {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999835e-05,  3.630284844098e-10, {   0,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999835e-05, -3.630284844060e-10, {   2,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999995e-05,  6.283185307187e-11, {   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999998e-05, -4.188790204131e-11, {   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   2,  -2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.708501168380e-05, -1.289521266432e-05, {   2,   0,   0,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.348406225107e-05,  6.462329670872e-06, {   2,   0,   1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -2.786696270965e-05,  1.207725788469e-05, {   4,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.296432995304e-05,  6.962098765072e-06, {   1,   1,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.681797059092e-06, -3.224374380823e-05, {   2,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.802582285032e-05,  2.188472169005e-05, {   0,   2,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.143141165714e-07, -3.936673837329e-05, {   1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.890396263129e-05,  1.096720317933e-05, {   2,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.554078834042e-05, -1.432051822102e-05, {   2,  -2,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -1.756098278353e-05,  2.229935642878e-05, {   2,   0,  -2,   0,   0,   0,  13, -21,   0,   0,   0,   0,   0 } },
 {  9.473782575101e-06,  3.038651422886e-05, {   1,  -1,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.491020890507e-05,  2.493956774998e-05, {   2,   0,  -1,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 { -1.642342308493e-05,  2.338381758505e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -3.970732968358e-05, -8.719596816724e-08, {   0,   0,   3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.250996276147e-05,  7.276144852443e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -7.315141506436e-06,  3.243081788269e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 {  2.450388204634e-05,  1.523912892250e-05, {   0,   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.241827103902e-05, -2.730909270557e-05, {   0,   0,   1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.663087613593e-05,  1.308902464697e-05, {   2,   0,   1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.080869363295e-05,  2.889703995081e-05, {   1,  -1,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  7.778148927379e-06, -3.188613601580e-05, {   1,  -1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.905855039526e-05, -5.915910437893e-07, {   3,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.721768808953e-05, -1.242967937389e-05, {   1,  -1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -3.687686722696e-05,  2.765600944772e-06, {   2,   0,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.471625248563e-05,  2.488817376134e-05, {   0,   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 {  3.952151589860e-05,  6.816921391619e-08, {   0,   2,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.071895440416e-05, -2.886916059936e-05, {   0,   0,   1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.146997643392e-05,  2.811796824706e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -3.711235326953e-05,  2.432115891110e-06, {   0,   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  3.510335629410e-05, -4.439660035185e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 { -1.181626423093e-05,  2.772057885620e-05, {   1,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -6.252514878808e-06, -3.327996156298e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  2.747332189873e-05, -1.205058437788e-05, {   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.201390404727e-05, -1.748258144968e-05, {   2,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.529836382390e-05, -4.195392884828e-06, {   2,   0,  -2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -9.045736914027e-06,  3.044715970213e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.910146312118e-05,  3.867304090322e-07, {   4,   0,  -3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.352902606754e-05, -5.928959743284e-06, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   5,   0,   0,   0 } },
 { -3.267523898681e-06,  3.617583242901e-05, {   2,   0,  -1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.325072455324e-05,  6.181746586938e-06, {   3,  -1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.295515646092e-05,  2.647362101745e-05, {   2,   0,  -1,   0,   0,  -6,   9,   0,  -3,   0,   0,   1,   0 } },
 { -2.213712706716e-05, -1.727092045727e-05, {   0,   0,   0,   0,   0,  23, -22,  -4,   0,   0,   0,   0,   0 } },
 { -1.496649692502e-05, -2.442017745173e-05, {   4,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.106430305370e-06, -3.227916392358e-05, {   1,  -2,   0,   0,   0,  28, -31,   0,   0,   0,   0,   0,   0 } },
 { -2.762428190904e-05,  1.171826634983e-05, {   1,  -1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.454302653603e-05, -2.478197533666e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  8.236306055550e-06, -3.104809100298e-05, {   1,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -8.236306055550e-06, -3.104809100298e-05, {   1,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -3.834053144039e-05,  9.305648356103e-07, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 { -2.702204187643e-06,  3.656483982694e-05, {   4,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.260555198330e-05,  2.665867277409e-05, {   0,   0,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 {  1.662803462608e-06, -3.758114585138e-05, {   2,   0,  -3,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  9.591880337573e-06,  2.964499672049e-05, {   2,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  7.642741468935e-06,  3.158960885275e-05, {   1,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.907179495704e-05, -1.347997725486e-07, {   0,   0,   2,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  3.907179495704e-05, -1.347997725486e-07, {   0,   0,   2,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 {  1.885014533205e-05,  2.034540294638e-05, {   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.083150940049e-05, -1.835555820313e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -3.909784579798e-05,  6.039319823042e-08, {   2,   0,  -2,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.889441422696e-06,  3.224885542409e-05, {   0,   0,   2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  8.902550428852e-06,  3.022860321301e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 {  7.476943721429e-06,  3.163755544267e-05, {   1,   1,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.158121302208e-06,  3.195426844378e-05, {   2,   2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.193253114551e-05, -2.717619166587e-05, {   1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.880813203136e-06,  2.922283194242e-05, {   1,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -3.174566912969e-05,  7.342843622337e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  2.427978537064e-05, -1.480443227638e-05, {   1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.123730412806e-05, -2.782550331345e-05, {   1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.033444012902e-06, -3.802884458579e-05, {   2,   0,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.837291985540e-05, -6.881265466241e-07, {   0,   0,   1,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.036266110728e-06,  3.301113728353e-05, {   1,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.429237805381e-05, -1.474849080498e-05, {   2,   0,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0 } },
 { -2.687139223090e-05,  1.215870889987e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.840867213498e-05, -2.061427917231e-05, {   0,   0,   1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 {  2.401532130357e-05,  1.498549473921e-05, {   2,   0,   2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  2.419732496367e-06,  3.654347707813e-05, {   3,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.872329727314e-05, -2.294008471807e-07, {   0,   0,   1,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.374525541797e-06,  3.657541204375e-05, {   2,   0,  -1,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 { -1.474944642892e-05, -2.419590868651e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0 } },
 { -9.476775903811e-06,  2.943934445851e-05, {   1,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.765006540052e-05,  1.251663237581e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  1.067958048120e-05, -2.819827396159e-05, {   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.838535461880e-05, -2.048386352748e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -3.320982376344e-05, -5.656359486315e-06, {   1,  -1,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.881498196677e-05,  2.528851411860e-08, {   2,   0,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.883568394297e-05,  2.490331761777e-09, {   0,   2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.468660752970e-05, -2.414601171982e-05, {   2,   0,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 { -3.162581674997e-06, -3.565800706780e-05, {   0,   2,   1,   0,   0, -23,  22,   0,   0,   0,   0,   0,   0 } },
 { -8.935017445156e-06, -2.987030746316e-05, {   2,   0,   0,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.013008064751e-07, -3.789616571548e-05, {   2,   0,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.388430850394e-05,  1.490414721153e-05, {   2,   0,   2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.605664855588e-05,  2.272872836889e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 { -1.739455575375e-06,  3.703855013909e-05, {   2,   0,   1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.803108296829e-05,  1.068917151257e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.848011947846e-05, -1.023476682785e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 {  2.059787650973e-05, -1.810995924902e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 { -2.804019746414e-05, -1.066523915214e-05, {   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.158304996366e-05, -7.122094594211e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 {  7.429872094431e-06,  3.124744571682e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 {  4.382668529354e-06, -3.427348309786e-05, {   1,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.631792567556e-05, -2.337541521061e-06, {   2,   0,  -1,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -3.393496712083e-05, -4.676079009917e-06, {   2,   0,  -1,   0,   0,   0,   4,   0,  -2,  -5,   0,   0,   0 } },
 {  2.583185715731e-05, -1.268774493189e-05, {   1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.549799384818e-06, -3.694751717981e-05, {   2,  -2,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.435645505221e-05,  2.413806354676e-05, {   0,   0,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -2.050937401059e-05, -1.797729031520e-05, {   0,   0,   1,   0,   0,   3, -11,  12,  -3,   0,   0,   0,   0 } },
 {  2.050937401059e-05, -1.797729031520e-05, {   0,   0,   1,   0,   0,  -3,  11, -12,   3,   0,   0,   0,   0 } },
 {  1.601848990638e-05,  2.246767326974e-05, {   4,   0,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.344753427576e-06, -3.213022283022e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.458322414398e-05,  3.885253345636e-06, {   3,  -1,  -2,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  1.915859651790e-05, -1.930384555072e-05, {   2,   0,  -1,   0,   0,   0,   6,  -8,   1,   0,   0,   0,   0 } },
 {  3.453941930918e-05, -3.915560001196e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 { -6.022384668947e-06, -3.243097088627e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.380116367706e-05,  4.620453301871e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  2.578066237673e-05,  1.263545078660e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 {  3.078760923070e-05,  7.624746281473e-06, {   2,   0,  -1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 {  1.734816505245e-05,  2.106201374190e-05, {   0,   2,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.780783831979e-05,  5.860830069109e-07, {   1,  -2,   1,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 {  3.780783831979e-05,  5.860830069109e-07, {   1,  -2,  -1,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 { -2.557259662254e-05,  1.279045863746e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 {  9.586908156635e-06,  2.876849156269e-05, {   0,   0,   2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.402180738128e-05, -2.432948421091e-05, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 { -2.394592418962e-05, -1.440445967319e-05, {   2,   0,   1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.757541178757e-05,  7.722170980505e-07, {   0,   0,   1,   0,   0, -34,  42,   0,   0,   0,   0,   0,   0 } },
 {  1.796735547383e-05,  2.036598516267e-05, {   0,   0,   0,   0,   0,   0,  10, -16,   4,   5,   0,   0,   0 } },
 {  7.306922046077e-06, -3.098991157703e-05, {   3,  -1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.979362961729e-05, -1.850130409834e-05, {   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.747620859829e-05,  2.080496256392e-05, {   2,   0,   0,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.714681489091e-06, -3.556099401350e-05, {   2,   0,   0,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 {  6.112418677986e-06, -3.214701379555e-05, {   4,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.063338630705e-06, -3.517656127067e-05, {   0,   0,   1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.293807521671e-06,  3.694496224458e-05, {   1,  -2,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  2.768152787650e-05,  1.054782878633e-05, {   2,   0,   0,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 { -3.787612032382e-05,  3.483800224345e-07, {   2,   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  7.718832885719e-06,  3.047817775450e-05, {   1,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.788265863331e-05, -3.062852253203e-07, {   2,   0,  -1,   0,   0, -20,  25,  -8,   3,   0,   0,   0,   0 } },
 {  3.815787091171e-05, -2.978369417070e-08, {   0,   0,   2,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.449258621682e-05,  1.368458982136e-05, {   4,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -3.368892344589e-05, -4.484373673418e-06, {   1,  -1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.819659053372e-05,  9.969672022607e-06, {   4,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.121769012453e-05, -2.692283237762e-05, {   1,  -1,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  2.457335005220e-05, -1.355831439521e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -3.101229040391e-05,  7.119291804840e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 { -3.776409187439e-05,  3.560660199028e-07, {   2,   0,  -5,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.254367707853e-05,  2.557396057354e-05, {   1,   0,   0,   0,   0,   0, -32,  37,  -1,   0,   0,   0,   0 } },
 {  2.956646289382e-05, -8.547883007830e-06, {   1,  -3,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.710002261939e-05, -1.100196219442e-05, {   4,  -2,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.672772691144e-05, -1.350302351777e-06, {   6,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -7.545701436294e-06, -3.052556504581e-05, {   2,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.509119117444e-05,  2.297407727453e-05, {   2,   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -3.147697512495e-05, -6.583407026007e-06, {   1,  -1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.460525964436e-05,  1.337773786511e-05, {   0,   0,   1,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.195363538671e-06,  2.976181540991e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.612760369684e-06,  3.634115910450e-05, {   2,   0,  -1,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 { -1.215494611176e-05, -2.579483229033e-05, {   2,   0,   0,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 { -7.948460048693e-06,  2.997638289633e-05, {   4,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.510112913359e-05,  1.282144420551e-05, {   0,   0,   2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -7.431810292649e-06, -3.048011742511e-05, {   2,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.761999082058e-05, -2.715311616203e-07, {   3,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.931003458789e-06, -2.795278311995e-05, {   2,   0,  -1,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.447614318689e-05,  1.338235552149e-05, {   0,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.762519373199e-05, -1.023327136785e-05, {   2,   0,   0,   0,   0,   0,  -7,   8,  -3,   0,   0,   0,   0 } },
 {  3.107600925042e-05, -6.769060890663e-06, {   3,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.278541802683e-05, -2.505869307691e-05, {   0,   0,   2,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -4.748451087200e-06, -3.307607596044e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 {  2.556936843238e-05,  1.225500803269e-05, {   2,   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 {  3.742973552048e-05,  3.820001156598e-07, {   4,   0,  -3,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -2.853970366775e-05, -9.245826872551e-06, {   2,   0,  -2,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.282567236422e-06,  3.549568602024e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 {  2.529482662487e-05, -1.247885550967e-05, {   4,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.162762487779e-05,  2.613807806807e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.437117847031e-05, -2.338034760807e-05, {   2,   0,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -5.809949368513e-06,  3.190286490941e-05, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 {  1.349074256579e-05, -2.421543936108e-05, {   2,   0,  -1,   0,   0, -18,  13,   9,   0,   0,   0,   0,   0 } },
 {  7.345909853122e-06,  3.035833302601e-05, {   0,   2,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.995723859296e-06,  2.970430053299e-05, {   2,   0,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.226113350563e-05,  5.431972810429e-06, {   0,   0,   2,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -3.225461391585e-05,  5.437135802738e-06, {   0,   0,   2,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -2.247501852787e-05,  1.520725257001e-05, {   1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.446967089153e-05, -1.320492897123e-05, {   0,   2,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.947826783003e-05, -8.187795027967e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  2.275090709469e-06,  3.537394810891e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 {  3.760150557765e-05, -3.963868765326e-08, {   1,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.521322562278e-05,  2.406213792627e-06, {   1,  -1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -6.475469139069e-06, -3.114034862292e-05, {   2,   0,  -1,   0,   0,   0,   9, -14,   3,   0,   0,   0,   0 } },
 {  1.508503782662e-06, -3.610408212588e-05, {   2,  -2,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.521736117253e-06, -3.605479480588e-05, {   0,   0,   2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.346435261844e-05,  1.410233409728e-05, {   0,   0,   3,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.865663986553e-05,  8.909912485919e-06, {   0,   0,   0,   0,   0,   8, -21,  16,  -4,  -5,   0,   0,   0 } },
 {  3.749673227199e-05,  5.072837207984e-08, {   2,   0,   2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.344722991891e-05,  1.409084183533e-05, {   0,   0,   3,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.317229215423e-05, -1.432594048083e-05, {   4,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  5.756430772125e-06,  3.173867806427e-05, {   1,   1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.576225461543e-05, -2.172477878625e-05, {   0,   0,   3,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.015653607171e-05, -1.732522358836e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,  -2,   5,   0,   0,   0 } },
 {  3.740250890282e-05,  6.683051101122e-08, {   0,   2,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.159321159511e-06,  3.030123294510e-05, {   2,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.734623517157e-05,  1.105958733226e-07, {   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.622247680545e-05,  1.121965814376e-05, {   0,   0,   1,   0,   0,   0, -12,  19,   0,   0,   0,   0,   0 } },
 {  3.175689672957e-05, -5.682491020274e-06, {   4,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.929220430403e-05, -1.813684173336e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -3.170908751095e-05,  5.719705164025e-06, {   2,  -2,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.739925489626e-05, -1.856348162916e-08, {   6,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.736126074894e-05,  1.496583660240e-08, {   2,   2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.266787063898e-06, -3.610513184269e-05, {   0,   0,   2,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -3.171994859732e-05,  5.640813960927e-06, {   1,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.438852401058e-05,  1.296411132994e-05, {   0,   0,   1,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 {  1.053389958659e-05,  2.678513488023e-05, {   0,   0,   0,   0,   0,   0,  11, -21,   0,   0,   0,   0,   0 } },
 { -2.337339199957e-05,  1.394230902171e-05, {   0,   0,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  2.339752022734e-05, -1.391688342931e-05, {   5,   0,  -1,   0, -17,   0,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.727728262560e-05, -1.556268640366e-08, {   2,   2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.456493461360e-05,  2.727830325834e-06, {   2,   0,  -1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.024893715419e-05, -2.702964282749e-05, {   2,  -1,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -3.721011444098e-05,  3.894876328268e-08, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,  -4,   0,   0,   0 } },
 {  3.102014685104e-05,  6.211918346299e-06, {   2,   0,   0,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -1.128814110210e-05, -2.594160816262e-05, {   2,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.697225558519e-05, -2.554067741235e-07, {   2,   0,   2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.243160038101e-05, -2.479305125467e-05, {   1,   1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.068620219386e-06, -3.015353553308e-05, {   1,  -1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.801211268820e-05, -9.206525150693e-06, {   0,   0,   0,   0,   0,   4, -15,  16,   0,   0,   0,   0,   0 } },
 { -3.671591218340e-05,  4.873363267848e-07, {   0,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.209810015418e-05, -2.508230044600e-05, {   1,   1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.090537506966e-06, -2.907740402807e-05, {   0,   0,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.627870214029e-05,  2.088804526822e-05, {   0,   0,   1,   0,   0,   0, -13,  21,   0,   0,   0,   0,   0 } },
 {  1.022268700879e-05, -2.694155589792e-05, {   2,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.540631840538e-06,  3.362203749111e-05, {   2,  -2,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.401896366994e-05, -1.313418785158e-05, {   0,   0,   1,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 { -3.518965031004e-05,  1.946452299715e-06, {   0,   0,   0,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 { -3.707329006433e-05,  5.965496400261e-08, {   0,   0,   0,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 { -7.042866807139e-06, -3.008670820901e-05, {   1,  -1,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.934482242265e-05,  7.729469297945e-06, {   2,   0,   1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  3.120110601968e-05,  5.868809069970e-06, {   4,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.064590573289e-05, -6.422827727418e-06, {   1,  -1,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.488007124478e-06, -3.557116057219e-05, {   2,  -2,   3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.153854162951e-05,  5.511448932402e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   4,  -5,   0,   0,   0 } },
 {  2.767278223607e-05,  9.373568085968e-06, {   4,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -5.589700601922e-06, -3.143592804925e-05, {   2,   0,   0,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -1.563554138152e-05,  2.138091949019e-05, {   2,   2,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.754866629338e-05, -1.944383476362e-05, {   1,   0,   0,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.410698888006e-05, -2.288150044374e-05, {   2,   0,  -1,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 {  1.541304087326e-05, -2.156708930592e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0 } },
 {  2.798484416609e-05,  8.994239009758e-06, {   4,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.510541357826e-05, -1.833447595518e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  3.058273131849e-05,  6.350668433543e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 { -6.771401224307e-06, -3.016188322063e-05, {   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.007894633102e-06, -3.392498237818e-05, {   4,  -2,  -2,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.497562749790e-05,  1.949298296554e-06, {   0,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.690800322462e-06, -3.222846524242e-05, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.799900972362e-05,  8.919033118530e-06, {   0,   0,   0,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0 } },
 {  3.102395995993e-05, -5.888306807953e-06, {   2,   0,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.534873866444e-05,  1.155848080857e-05, {   0,   0,   1,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 {  1.507556309955e-05, -2.182937965987e-05, {   0,   0,   2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  3.005827449900e-06, -3.389473649680e-05, {   4,  -2,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.556377468303e-05,  1.336543714216e-06, {   2,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.033526545204e-05, -1.655895491838e-05, {   2,   0,  -1,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.510863241475e-05, -1.175037897377e-05, {   0,   0,   1,   0,   1, -23,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.682723049023e-05, -1.001806237411e-05, {   0,   2,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.973988217519e-06, -2.986731760278e-05, {   1,  -1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.682149400068e-05, -1.711794455414e-08, {   4,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.221440172218e-05, -4.621330901190e-06, {   0,   0,   2,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -6.941375922567e-06,  2.988484137214e-05, {   2,   0,   1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  2.040768618385e-05,  1.639590531615e-05, {   2,   0,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 { -2.980568692393e-08,  3.675114964989e-05, {   1,   1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.171966499343e-05,  1.500801188930e-05, {   2,   0,  -1,   0,   0,   0,  15, -20,   0,   0,   0,   0,   0 } },
 {  6.440812292796e-06,  3.028452516311e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 {  1.376261345377e-05, -2.295532227172e-05, {   2,   0,  -1,   0,   0,   3,  -1,  -4,   0,   0,   0,   0,   0 } },
 {  1.474736346923e-05,  2.196905806876e-05, {   1,   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -8.883309845417e-06,  2.782914553439e-05, {   2,   0,  -1,   0,   0,   0, -28,  31,   0,   0,   0,   0,   0 } },
 {  2.460870955023e-05,  1.210210634278e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 { -3.211353240290e-05,  4.593533342538e-06, {   0,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.965130780419e-05,  7.010294132290e-06, {   2,   0,  -2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.474440121593e-05,  1.912903320467e-06, {   4,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.654436724220e-05,  1.097182523194e-07, {   2,   0,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.648406094670e-05, -1.016593680019e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   2,   0,   0,   0 } },
 {  2.630490264279e-05, -1.027372008221e-05, {   0,   0,   0,   0,   0,   0,   5,  -8,   2,   0,   0,   0,   0 } },
 {  1.080863414603e-06,  3.548173545140e-05, {   2,   2,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -6.232611994162e-08,  3.647665633488e-05, {   3,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.264206276161e-05, -1.388805215909e-05, {   2,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0 } },
 {  2.007340066456e-05, -1.644426926308e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -3.754432100696e-06, -3.276042710408e-05, {   2,   0,  -2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -3.915761013126e-06,  3.259567685061e-05, {   1,   1,   0,   0,  -7,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.553172606471e-05,  1.097267763855e-05, {   2,  -2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  3.812317386616e-06, -3.266750998589e-05, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  2.829839155094e-05,  8.180005659816e-06, {   1,   0,   0,   0,   0, -29,  35,   0,  -3,   0,   0,   0,   0 } },
 {  2.371434866481e-05,  1.275506109501e-05, {   0,   0,   3,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  6.722411922001e-06,  2.967806598925e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 {  6.838849092749e-06, -2.955953777155e-05, {   2,   0,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.184129583561e-05, -2.453950049673e-05, {   4,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.805308460528e-06, -3.456549804692e-05, {   4,   0,  -2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  2.346618973304e-05, -1.288424695682e-05, {   2,   0,  -4,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.422333741078e-05,  1.212526960098e-05, {   0,   2,  -1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.392023923051e-05, -2.240186365823e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.478251528901e-05,  2.152502570900e-05, {   0,   0,   1,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 { -3.628785852206e-05,  6.061392637123e-09, {   1,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.437117571604e-05,  2.192262313018e-05, {   0,   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.969984579180e-06,  3.331320937686e-05, {   2,   2,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.476634970611e-05,  2.150656890414e-05, {   0,   0,   1,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 { -1.465425407455e-06,  3.479200138696e-05, {   2,   0,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.373896109569e-06, -3.487879877596e-05, {   0,   0,   1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.503145695273e-06,  3.370836595353e-05, {   0,   0,   1,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 { -3.563505634408e-05, -5.657309288719e-07, {   1,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.249238382993e-05,  2.367168811003e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -1.462863039018e-05, -2.152048376701e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -4.947073357796e-06, -3.119725911291e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -2.375507846516e-05, -1.237824194103e-05, {   4,   0,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -1.067640963553e-06, -3.504077561595e-05, {   6,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -9.187365226721e-06, -2.691714127523e-05, {   0,   0,   1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 {  6.016073681052e-06, -3.007252629320e-05, {   4,   0,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.338463827930e-06, -2.674700361294e-05, {   0,   0,   0,   0,   0,   0,   2,  -2,   0,   2,   0,   0,   0 } },
 { -1.058821059403e-05, -2.549651473971e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  7.551524800353e-06,  2.852101725860e-05, {   4,   0,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.730596104429e-05, -1.875360009709e-05, {   2,   0,  -1,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.618149285496e-05, -1.986221119946e-05, {   0,   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -5.456415032010e-06, -3.058543781707e-05, {   2,   0,   1,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 {  5.456415032010e-06, -3.058543781707e-05, {   2,   0,  -1,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -2.426625609076e-05,  1.174396170259e-05, {   2,   0,   1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.106853289202e-05,  2.491541279274e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  1.261139213749e-05, -2.336081479440e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -2.848274325035e-05,  7.489197964953e-06, {   2,   0,  -3,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -1.628082666850e-05,  1.968409185526e-05, {   2,  -3,   0,   0,   0,  20, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.939582099128e-05,  6.557094326822e-06, {   0,   2,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.227843038984e-05,  1.367079331291e-05, {   2,   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.528970136930e-05,  2.065297378903e-05, {   0,   2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.176464177271e-05,  1.417064402119e-05, {   4,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.253241654482e-05,  2.339564570076e-05, {   2,   0,  -1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 {  3.518543750063e-05,  7.224882390926e-07, {   2,   0,  -3,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.871832699382e-05, -7.186032631096e-06, {   1,   1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -6.897848468710e-06, -2.900122517185e-05, {   0,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.416338106583e-05,  2.173537456551e-05, {   1,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.051263569997e-05,  1.535417773035e-05, {   1,  -1,   3,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -1.194511993196e-05, -2.389115316884e-05, {   0,   0,   1,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 {  2.923330198973e-05, -6.594625859178e-06, {   0,   0,   0,   0,   0,   0,   8, -16,   8,  -5,   0,   0,   0 } },
 { -8.857352675820e-06,  2.696784012582e-05, {   1,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.580514243126e-05,  1.119836251968e-08, {   0,   0,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.011991507900e-06,  3.478454215241e-05, {   1,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  4.595725990199e-06,  3.117240060390e-05, {   0,   0,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.725288230184e-06,  2.903490924405e-05, {   2,  -1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.954143424355e-07,  3.554118814635e-05, {   4,   0,  -2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.362597914211e-05,  2.099545124734e-06, {   2,   0,  -1,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 {  3.362597914211e-05,  2.099545124734e-06, {   2,   0,  -1,   0,   0,   0, -14,  27,  -3,  -3,   0,   0,   0 } },
 {  2.572926694891e-05,  9.985072449647e-06, {   2,   0,  -1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 { -3.560491645973e-05,  8.793596223832e-08, {   2,   0,  -1,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 {  2.717408688144e-05, -8.510499272707e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  2.920342687403e-05, -6.476332664599e-06, {   2,   2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.649392715249e-05,  1.917572572807e-05, {   2,   0,   1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.749822716471e-06, -3.391674848502e-05, {   2,  -2,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  3.561636497879e-05,  4.411856436895e-08, {   3,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.099106392621e-05, -2.466362423320e-05, {   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.857413718109e-06, -2.879587998881e-05, {   1,  -1,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.336690719508e-05, -2.228231791109e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.884639711032e-05, -6.779827716529e-06, {   3,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.963843363121e-05,  5.984857003541e-06, {   2,   0,  -2,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -1.052902243332e-06, -3.455639587235e-05, {   2,   0,  -4,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.622052574726e-05, -1.937180088309e-05, {   0,   0,   0,   0,   0,  13,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.053803140127e-05,  5.043208752340e-06, {   2,   0,   1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -3.435247256091e-05, -1.205918315641e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  6.172655332016e-06,  2.935810494824e-05, {   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.526049066569e-05,  2.026648188250e-05, {   1,  -1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -3.483228089084e-05, -6.897841578077e-07, {   2,  -2,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  2.286782981554e-05, -1.264859161162e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -7.414762918280e-06,  2.810157157632e-05, {   1,   1,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.368302759479e-05,  1.182850724324e-05, {   6,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.613387332241e-05, -9.375785268748e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  1.260131053859e-05,  2.290757521979e-05, {   4,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.047925188968e-05,  5.026724391182e-06, {   2,   0,   1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  2.283697951900e-05,  1.265792284721e-05, {   1,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -6.258718600413e-06, -2.922252839826e-05, {   2,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -9.443583719449e-06,  2.601909342846e-05, {   1,   1,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.831457771549e-06, -2.862108729703e-05, {   1,  -1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.777270716870e-06,  2.965727158970e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -7.831538342028e-06, -2.758984069341e-05, {   3,   0,  -1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 { -2.799431918532e-05,  7.425085148879e-06, {   1,  -1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.323420462261e-05, -1.216418634755e-05, {   2,   0,  -2,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  9.734948642967e-06, -2.565033928055e-05, {   2,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -9.734948642967e-06, -2.565033928055e-05, {   2,  -1,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  2.371961079706e-06,  3.299699189228e-05, {   0,   0,   1,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 { -3.343579051569e-05,  1.925376976884e-06, {   4,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.425173007458e-05,  1.108193849435e-05, {   1,   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -3.476083609243e-05, -5.721936941578e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 {  1.386918806943e-05, -2.143904297204e-05, {   2,  -2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.114948446469e-05,  4.150719278084e-06, {   1,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.172289500127e-05,  1.356872689996e-05, {   2,  -2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.527537495776e-05,  1.464687988736e-08, {   4,   2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.380509563417e-05, -1.477386878754e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  6.743360529039e-06, -2.852334883663e-05, {   1,  -1,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.063486474892e-05, -2.462315131876e-05, {   1,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -3.457573518391e-05,  6.673957622580e-07, {   2,   2,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.349383991042e-05,  1.173406778499e-05, {   1,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.349383991042e-05,  1.173406778499e-05, {   1,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.221688617908e-06,  2.799835661763e-05, {   1,  -1,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.247284266900e-06, -2.995820097992e-05, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.342559350960e-05,  1.756236527478e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  6.734446963555e-06, -2.844447604265e-05, {   2,   1,  -3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.745487595331e-06,  2.842933769201e-05, {   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.024987247051e-05,  4.918828821459e-06, {   0,   0,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0,   0 } },
 {  2.504533259268e-05,  1.010872948537e-05, {   1,  -1,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  4.267782850510e-06,  3.086662081578e-05, {   2,   0,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.929515362833e-05,  5.830598955289e-06, {   1,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.667282052476e-06,  2.844820109645e-05, {   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.686706256481e-05, -1.824841977377e-05, {   3,  -1,  -2,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.200008172171e-05,  2.310768361470e-05, {   1,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.718522661557e-06, -2.737243804139e-05, {   2,   0,   0,   0,   0,  -8,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.331164527059e-07,  3.455170956784e-05, {   1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.410939721687e-05,  1.096018522505e-05, {   0,   0,   2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.104431001076e-05, -4.011995419982e-06, {   1,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.722571374820e-05,  7.829267346883e-06, {   2,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.017893676014e-05, -1.486743505650e-05, {   2,   0,  -2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -3.499843976473e-05,  3.745881757534e-08, {   0,   0,   2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.878911743488e-05,  6.214578797442e-06, {   2,   0,   2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.750659684005e-05, -7.471523795335e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  3.473149860393e-05,  2.448030491777e-07, {   2,   2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.330314626042e-05,  1.163888635565e-05, {   6,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.325997625775e-05, -1.166325849785e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  1.286737434967e-05, -2.205303422114e-05, {   0,   0,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 { -7.431473720520e-06, -2.747995925230e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 {  8.347196411635e-06,  2.656403412413e-05, {   0,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.756301706678e-06, -2.811583280933e-05, {   3,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.608818390410e-05, -8.777623990127e-06, {   4,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.368460177531e-06, -3.347285939316e-05, {   2,   0,  -1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.352098657388e-05, -2.130912355876e-05, {   3,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.916949488678e-05, -5.656824027927e-06, {   1,  -1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.479528450894e-05,  2.972094885100e-08, {   2,   0,   1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.266219777434e-05, -2.214211419850e-05, {   2,   0,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -6.183238144246e-06, -2.861105593686e-05, {   2,   0,   0,   0,   0, -11,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.642621266493e-05, -8.340412097608e-06, {   1,   1,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  5.563719812935e-07,  3.418157607569e-05, {   3,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.836293973569e-05, -6.374809224973e-06, {   4,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.205049680530e-05, -2.267071610021e-05, {   0,   0,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.129753853154e-06, -3.258324227519e-05, {   1,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.449325940382e-05,  2.131295535710e-07, {   0,   2,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  6.333810007184e-06,  2.835916539596e-05, {   1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -7.750666573145e-06,  2.693228012357e-05, {   4,   0,  -3,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.644106008360e-06, -2.802358465317e-05, {   1,   1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.016372395284e-05, -2.448085158228e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  2.307810800983e-05, -1.152640711159e-05, {   2,   2,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.557717908582e-05,  1.902149462920e-05, {   0,   0,   0,   0,   0,   2,   2,  -8,   0,   0,   0,   0,   0 } },
 { -3.456555355337e-05, -2.868801847411e-08, {   1,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.456706748398e-05, -2.302495819230e-08, {   4,  -2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.879377628555e-05,  5.791588758153e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -2.820237523380e-05,  6.361504838732e-06, {   6,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.661915075739e-06, -3.189823618784e-05, {   2,  -2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.456735819044e-06,  3.309928380537e-05, {   0,   0,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  5.214891713738e-06, -2.931921019673e-05, {   2,   0,  -2,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 {  1.599824739007e-05, -1.852727430968e-05, {   0,   0,   1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.448989052083e-05,  2.481588838014e-08, {   2,   0,  -3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.998824136878e-05, -4.506697561541e-06, {   0,   0,   1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.702760390307e-06,  3.178861821368e-05, {   4,   0,  -1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -2.702760390307e-06,  3.178861821368e-05, {   0,   0,   1,   0,   0,  22, -28,   7,   0,   0,   0,   0,   0 } },
 { -2.764168055767e-05, -6.847044947066e-06, {   1,  -1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.310946886435e-05, -1.136820492074e-05, {   1,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.938577043680e-06,  3.250910338652e-05, {   2,   0,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.434049851672e-05,  1.013368470376e-07, {   2,   0,  -2,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.800123883851e-06, -2.663073498989e-05, {   0,   0,   2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.181760795545e-05, -2.604387800182e-06, {   1,  -1,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.335178337208e-05,  2.106032901987e-05, {   2,  -2,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -3.018295788382e-05,  4.222783883332e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   5,   0,   0,   0 } },
 {  1.422548735183e-05, -2.017623466316e-05, {   1,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.792125454570e-05, -1.647316170072e-05, {   0,   0,   1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 {  1.870561962463e-05, -1.568493690860e-05, {   2,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.567519553075e-05,  1.870578653113e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 { -3.421740321678e-05,  1.570632086362e-07, {   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.580614028131e-06, -2.778504779661e-05, {   3,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.443634717754e-05, -1.990063881358e-05, {   1,   0,  -1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.882301624113e-05, -5.500844117094e-06, {   2,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.285469976799e-05,  1.465222798721e-06, {   4,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  6.561081080103e-06,  2.774183677694e-05, {   0,   0,   0,   0,   0,   0,   7, -12,   3,   0,   0,   0,   0 } },
 { -7.194923334426e-06,  2.710724806617e-05, {   0,   0,   4,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -4.306334010088e-06,  2.997920728756e-05, {   4,   0,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  2.426982174055e-05,  9.992458326539e-06, {   1,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  8.248438738544e-06,  2.600459342279e-05, {   0,   0,   1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.277613809928e-05, -1.146886722314e-05, {   1,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.668155949505e-06, -2.755734074825e-05, {   0,   2,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.720260537658e-06, -2.949864654190e-05, {   2,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.932502798425e-05,  4.891397326002e-06, {   1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.493784518381e-06,  2.871648708095e-05, {   1,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  3.418862071979e-05, -4.775361792321e-09, {   4,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.772620188131e-05,  1.645825389134e-05, {   2,   0,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 {  2.215503002017e-05,  1.202847890202e-05, {   4,   0,  -2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -2.754484930051e-05, -6.630315876515e-06, {   3,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.341561033771e-05, -2.074804284287e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -1.932106351723e-05, -1.482894390237e-05, {   2,   0,  -1,   0,  -4,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.091739225207e-05, -1.322952718075e-05, {   2,   0,   1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  2.254427025988e-05, -1.155534155072e-05, {   2,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.565091822487e-05, -1.844114835501e-05, {   1,  -1,   0,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  1.880949467989e-06, -3.220393391308e-05, {   2,   0,  -3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.992121638117e-05, -1.415625229326e-05, {   2,   0,  -1,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.420830648919e-05,  9.863803728597e-06, {   2,   0,  -1,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.472570503783e-05, -1.931915486194e-05, {   1,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.387935099816e-05,  1.645144796559e-07, {   1,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.401858803020e-05, -2.296377448539e-08, {   3,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.832582928510e-07,  3.364515658227e-05, {   4,   0,  -3,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -4.606431589659e-06,  2.941660585277e-05, {   2,   0,   0,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 {  1.266701303864e-05,  2.135501344608e-05, {   0,   0,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  1.661180218086e-05, -1.738460250393e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  5.279570590626e-06, -2.870370420012e-05, {   2,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.394939318984e-05,  2.766280567231e-08, {   4,   0,  -2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.140477605388e-05,  2.561876415101e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 {  3.395553958247e-05,  8.471902004529e-09, {   4,  -2,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.906766739608e-06, -2.703348347308e-05, {   1,  -1,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -7.699409963562e-06,  2.623833616654e-05, {   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.597594520067e-06, -2.932638437408e-05, {   4,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.382416963929e-05, -1.008485119845e-05, {   1,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.243316779228e-05,  1.475747562952e-06, {   0,   0,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  6.897971227785e-06,  2.700479035492e-05, {   3,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.112799767404e-05,  2.276883214405e-05, {   0,   0,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  3.580071413413e-06, -3.030141290336e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -3.381849737983e-05, -5.105765576437e-08, {   0,   2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.695586593770e-05,  6.910212436462e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 {  1.630665313560e-05,  1.755659930393e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 {  1.589394072871e-06,  3.226391002903e-05, {   2,   0,  -1,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 { -1.219292011656e-05,  2.165278670046e-05, {   2,   0,   0,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 {  3.344749452731e-05, -3.950678051338e-07, {   2,   0,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.380801381774e-05,  2.544131952717e-08, {   1,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.602101167153e-05, -7.802767667908e-06, {   0,   0,   1,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.044534309595e-06, -3.077276489706e-05, {   1,  -1,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.363757754393e-05,  1.738991214471e-07, {   0,   0,   3,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  3.056887795341e-05, -3.232215125187e-06, {   0,   0,   0,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.094810572520e-07,  3.359154091109e-05, {   0,   0,   2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.607624832387e-05, -7.708848194533e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -4.970956662880e-06, -2.880087403644e-05, {   1,   2,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.970956662880e-06, -2.880087403644e-05, {   1,  -2,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.871578498313e-05,  5.053549751286e-06, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.251626670860e-05, -1.124595941229e-05, {   2,   2,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.875026845724e-05, -4.998574721782e-06, {   2,   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  2.190571599053e-06, -3.155438783749e-05, {   0,   0,   0,   0,   0,  14, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.627892293099e-05, -1.744800747456e-05, {   2,   0,  -1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 { -2.127013219346e-05, -1.245334117799e-05, {   1,  -1,   0,   0,   0,   1,   9, -18,   0,   0,   0,   0,   0 } },
 { -6.435981633856e-06, -2.727741094178e-05, {   1,  -1,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.697477104656e-05, -1.673059985018e-05, {   0,   0,   0,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.298198845509e-05,  1.072031380622e-05, {   4,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.379118494939e-05,  9.903498338017e-06, {   1,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  3.273196761218e-06, -3.041024803427e-05, {   4,   0,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -2.849663703803e-05,  5.185004342646e-06, {   1,   1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.204546788562e-05, -1.163511134213e-05, {   2,   0,  -2,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  2.262146566830e-05,  1.105389482456e-05, {   2,   0,  -1,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.806556072151e-06, -3.184343516878e-05, {   2,   0,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -8.082527870091e-06,  2.555674387792e-05, {   4,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.459871200773e-06, -2.616723968042e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  3.311744140577e-05,  5.089090418586e-07, {   2,   0,  -3,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.425065892629e-06, -2.718952598758e-05, {   2,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.281952103764e-05,  7.842777864579e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 { -8.417947310390e-07, -3.275942023097e-05, {   2,   0,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.218403480380e-05,  2.141485988091e-05, {   2,  -2,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -7.080093386507e-06,  2.648990611368e-05, {   2,   0,   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.894386241600e-06, -2.667506808526e-05, {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.189059646258e-05,  2.166942570789e-05, {   2,   0,  -2,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.228807569093e-05, -2.126857922137e-05, {   2,   0,   2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.350930560231e-05,  3.872605739904e-08, {   1,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.305959672110e-05, -1.047259394172e-05, {   2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.531454196871e-05, -1.821710021012e-05, {   1,  -1,  -1,   0,   0,  17, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.255435248905e-05,  1.096568591528e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 {  6.482319379201e-06, -2.703063477078e-05, {   1,  -1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -2.604453885058e-05,  7.465895790534e-06, {   0,   0,   0,   0,   0,   0,   3,  -6,   4,  -5,   0,   0,   0 } },
 {  3.338673798102e-05,  7.586434014873e-08, {   0,   0,   2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.365784660175e-05, -1.976950229323e-05, {   1,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.298540775965e-05,  1.043178861770e-05, {   2,   0,   0,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 { -3.307366513528e-05,  3.245096277895e-07, {   0,   0,   3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.386799990559e-05, -1.951508888813e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  9.794160000972e-06,  2.358692494438e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  3.012397883852e-05,  3.243640520195e-06, {   2,   0,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  9.580292116141e-06, -2.378634729785e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   4,   0,   0,   0,   0 } },
 { -6.242945542166e-06, -2.709165286943e-05, {   1,  -1,   1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.635661476482e-05, -6.970280278257e-06, {   0,   0,   1,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 { -2.897166804849e-05, -4.354824922474e-06, {   4,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.125242463472e-05,  1.202591120246e-05, {   0,   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  4.948270778029e-06,  2.832459025728e-05, {   2,   0,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.952935766299e-05, -1.372656926589e-05, {   1,  -1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.335228245520e-05,  1.988234242473e-05, {   1,   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.325121335901e-06,  3.189489088545e-05, {   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.969983468621e-06, -2.623533944651e-05, {   1,  -1,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.051990045044e-06,  2.714953353665e-05, {   2,  -2,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.276480117598e-05, -3.803083192032e-07, {   4,   0,   0,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.200046282022e-06,  2.994352648060e-05, {   2,   0,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.901453298909e-06,  2.323137069549e-05, {   4,   0,  -3,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.583526531175e-06,  3.053362287273e-05, {   0,   0,   0,   0,   0,  22, -28,   7,   0,   0,   0,   0,   0 } },
 {  8.847238928018e-06, -2.426214917873e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.593992265047e-05, -1.714568985075e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -2.617087134753e-05,  6.914410974030e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0 } },
 { -4.347320510652e-07, -3.264600127294e-05, {   4,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.097509957582e-05,  2.209743428070e-05, {   3,   0,  -2,   0,   0,   0,  -9,  -3,   0,   0,   0,   0,   0 } },
 {  1.473214308033e-05, -1.833637076889e-05, {   2,   0,   0,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.251409261530e-05,  5.480758353104e-07, {   3,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.286055010622e-05,  1.999621124679e-07, {   2,  -2,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.234864844784e-06, -3.182099107784e-05, {   0,   0,   1,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 { -3.288127668583e-05,  1.724480873700e-07, {   4,   0,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.310338264846e-06,  3.173733633497e-05, {   2,   0,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  4.348471108516e-06, -2.869670354958e-05, {   0,   0,   3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.219335503138e-06,  3.182557797216e-05, {   2,  -2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.695054552815e-06,  2.334956324098e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -1.016833877786e-05,  2.287126350736e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.205209819792e-05,  9.614105160737e-07, {   2,  -2,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.435875279504e-06,  3.157581897612e-05, {   0,   0,   1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  3.052831465847e-05,  2.472698585843e-06, {   2,   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -8.669691691950e-06, -2.432470465617e-05, {   0,   0,   1,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.313749187060e-06, -2.667584200217e-05, {   2,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.136746195765e-05,  1.617158318099e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  4.696307940600e-06,  2.825670306111e-05, {   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.927869412560e-06,  2.799100910913e-05, {   2,   0,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.626774486435e-06, -3.128377757963e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  6.628381715100e-07,  3.222460279606e-05, {   2,   0,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.334136344859e-06, -2.655157119083e-05, {   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.942594453033e-05,  1.345417571436e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0 } },
 { -3.286132265750e-05, -1.698571772768e-08, {   0,   0,   3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.351136858091e-06,  2.952242640015e-05, {   2,   0,  -1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.045897161319e-05,  1.240460918176e-05, {   2,   0,  -1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0 } },
 {  3.130444701374e-05, -1.537339215117e-06, {   2,   0,  -1,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 {  2.985123619650e-05,  2.983906422972e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.897040565481e-05, -3.858865582123e-06, {   2,   0,   1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.824970922383e-06, -2.999780703217e-05, {   0,   0,   1,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 {  7.095478094700e-07,  3.211010046262e-05, {   2,   0,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.845167033962e-06, -2.597186187954e-05, {   2,   0,   0,   0,   0,  -9,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.240713935683e-06, -2.855774508975e-05, {   1,   0,   0,   0,   0, -15,  12,   0,   0,   2,   0,   0,   0 } },
 {  2.202558542990e-05,  1.077267468610e-05, {   0,   0,   3,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.180567340448e-05, -1.097721308269e-05, {   2,   0,  -1,   0,   0,  -7,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.631771215561e-06, -2.514877503807e-05, {   6,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.971573226509e-05,  1.304257416935e-05, {   1,  -1,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.322870605194e-05,  1.952445503389e-05, {   2,   0,  -1,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -7.622835709587e-06, -2.512199705028e-05, {   6,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -8.472089510114e-06,  2.426252159741e-05, {   1,  -1,   1,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 { -8.472089510114e-06,  2.426252159741e-05, {   1,  -1,  -1,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 {  1.348510208779e-05,  1.923540009039e-05, {   2,   0,  -2,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -1.419741029806e-05,  1.849341103746e-05, {   2,  -2,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.897079731312e-05,  3.708125486215e-06, {   1,   1,  -2,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -1.279626542837e-05,  1.987732409315e-05, {   4,   0,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.833758858630e-05,  4.331146818519e-06, {   2,   0,   0,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -2.940613228988e-05,  3.259596814740e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 { -1.058825962849e-05,  2.207670103415e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 { -5.972772624178e-06,  2.668291293704e-05, {   2,   2,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.245610939267e-05,  1.878136138850e-07, {   0,   0,   3,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.799486764996e-05, -4.648067455516e-06, {   0,   0,   1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -5.892386266031e-06, -2.675046080748e-05, {   2,   0,  -1,   0,   0, -11,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.091337298701e-05, -2.171823753488e-05, {   1,   1,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.148518593656e-05,  2.113804385694e-05, {   0,   2,  -3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.711267075394e-05,  1.550962704200e-05, {   3,   1,  -2,   0,   0, -40,  41,   0,   0,   0,   0,   0,   0 } },
 {  9.217220052091e-06, -2.339976770737e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0 } },
 { -2.412614825197e-05,  8.484060537775e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -2.190147647185e-05,  1.070720726494e-05, {   0,   0,   3,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.198815930251e-05,  5.985316416775e-07, {   2,   0,  -1,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 {  6.037679030676e-06,  2.654516800612e-05, {   0,   0,   0,   0,   0,   3, -17,  23,   0,  -3,   0,   0,   0 } },
 {  2.160573965124e-05,  1.097453071183e-05, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  6.340190431981e-06,  2.623417007499e-05, {   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.198164140318e-05,  5.896448250203e-07, {   2,   0,  -2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -2.685817123620e-06,  2.987953122756e-05, {   1,  -3,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.685107888033e-06,  2.987959497098e-05, {   3,  -3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.685087028140e-06, -2.987959684553e-05, {   1,  -1,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.684873214160e-06, -2.987961605884e-05, {   1,  -3,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.684794989499e-06, -2.987962308772e-05, {   3,  -1,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.683663337343e-06,  2.987972474926e-05, {   5,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.683663337343e-06,  2.987972474926e-05, {   5,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.683663337343e-06, -2.987972474926e-05, {   1,   3,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.683444307449e-06, -2.987974442074e-05, {   1,   3,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.627607538183e-06, -2.592511205698e-05, {   2,   0,  -2,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.157929386707e-05,  1.097205162654e-05, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -1.346929159601e-05, -1.908139872870e-05, {   4,   0,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.561178932911e-05, -6.930959584075e-06, {   2,   0,   1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.220135847396e-05, -2.033976710300e-05, {   1,  -1,  -2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.243506955166e-05, -2.009490097745e-05, {   1,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.244001486876e-05,  7.333914586430e-08, {   2,  -2,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -9.961163010554e-06, -2.254752222625e-05, {   1,  -1,   1,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 { -2.781772351035e-06, -2.972370430040e-05, {   0,   0,   1,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 { -2.158170500938e-05,  1.092157994557e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 { -6.687933465616e-06,  2.578741859322e-05, {   1,   1,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.890674990051e-05,  3.560788434095e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 { -1.087836385140e-05, -2.156557871985e-05, {   0,   0,   2,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -9.478069973353e-06,  2.292077134129e-05, {   0,   0,   2,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  2.156833644040e-05,  1.082024293774e-05, {   0,   0,   1,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.250293731265e-05, -1.987225321894e-05, {   1,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.078701506999e-05,  1.158080712638e-05, {   2,   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -9.972452421869e-06, -2.239503435611e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0 } },
 {  1.061876996089e-05, -2.174690499738e-05, {   1,  -1,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -2.019712265906e-05,  1.215900647417e-05, {   2,   0,  -1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0 } },
 {  3.232409709564e-05,  2.956327912331e-08, {   2,   0,  -1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  5.885995910351e-06, -2.645822715973e-05, {   2,   0,  -3,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.477429335763e-05, -1.756690635579e-05, {   1,  -1,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.185181824627e-05,  1.046804945121e-05, {   2,   0,  -2,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  2.020842254062e-05, -1.210687576144e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 { -6.246561808461e-06, -2.604963080928e-05, {   2,   0,   0,   0,   0, -10,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.519765406724e-06,  2.676917405358e-05, {   0,   0,   0,   0,   0,   1,   1,  -5,   0,   0,   0,   0,   0 } },
 {  1.569214223677e-05, -1.659465071049e-05, {   2,   0,   0,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.153193713480e-05, -1.075401291256e-05, {   4,   2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.710636530775e-05,  5.177943869896e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -2.433679327806e-05, -7.937118209085e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   3,   0,   0,   0 } },
 {  7.766281903339e-06,  2.448198832625e-05, {   1,  -1,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.591489494279e-06, -2.264859923765e-05, {   2,   0,   1,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 {  9.591489494279e-06, -2.264859923765e-05, {   2,   0,  -1,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.303248128977e-06, -3.092931789597e-05, {   4,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  2.754916212959e-05, -4.672358502549e-06, {   3,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.392594314249e-05,  1.829367758222e-05, {   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.985076166866e-06,  2.823315833341e-05, {   0,   0,   0,   0,   0,   0,  11, -20,   0,   0,   0,   0,   0 } },
 {  6.276911556373e-06, -2.593990849251e-05, {   0,   2,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.469637849369e-06,  2.774469438306e-05, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 {  2.277825500920e-05,  9.435490749467e-06, {   2,  -2,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0,   0 } },
 {  2.511376237419e-06,  2.968830722848e-05, {   4,   0,  -2,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -1.398088062011e-05, -1.820822934740e-05, {   0,   0,   3,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.408536791799e-05,  1.809623855457e-05, {   1,   0,  -1,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 {  1.037004463041e-05, -2.180376082566e-05, {   0,   2,  -1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -2.157144094254e-05, -1.058752215601e-05, {   1,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.476137219507e-06, -3.066786573950e-05, {   2,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.201063698862e-05,  1.012915350863e-05, {   0,   0,   1,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 {  1.074284208311e-05,  2.138685026755e-05, {   2,   0,  -1,   0,   0,   0,  16, -20,   0,   0,   0,   0,   0 } },
 {  1.060855571558e-05,  2.150890027213e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 {  2.148872666777e-05,  1.062022723373e-05, {   2,   0,  -3,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.817804051580e-06,  3.026329828841e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 { -6.116631993399e-06,  2.596381812032e-05, {   1,  -1,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.206472570073e-05, -2.001188989299e-05, {   1,  -1,   2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.203160922188e-05, -4.487496942283e-08, {   4,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.644288446919e-06, -3.042533845718e-05, {   2,   0,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 {  1.467005597518e-05,  1.739866539119e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1, -11,   3,   0,   0 } },
 { -1.029118566917e-05, -2.172968899611e-05, {   2,   0,   0,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 { -2.581001637461e-05, -6.191544386157e-06, {   1,  -1,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.015973196894e-05, -1.183086445981e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 { -5.580136447492e-06, -2.640788100121e-05, {   3,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.793943337819e-06, -2.819128581040e-05, {   2,   0,  -1,   0,   0,  -5,   5,   4,   0,   0,   0,   0,   0 } },
 {  5.572674187077e-06, -2.640826508569e-05, {   2,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.827686441388e-05, -3.702307642545e-06, {   2,   0,  -2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -8.073453740654e-06, -2.388932210695e-05, {   2,   0,  -1,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.396984561115e-05,  7.987148238422e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -2.586095380010e-05,  6.093714518503e-06, {   1,   1,   0,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 { -6.771234583300e-06, -2.517516867321e-05, {   3,   1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.888393239682e-05,  1.304805311847e-05, {   2,   0,  -1,   0,   0,   0,  16, -21,   0,   0,   0,   0,   0 } },
 { -6.124746767625e-06, -2.580657014183e-05, {   1,  -1,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.175996119470e-05, -2.016545728442e-05, {   2,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  1.443541133203e-05,  1.748495971456e-05, {   1,   1,   0,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.160118964595e-05,  1.030550443545e-05, {   2,   0,  -2,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -1.147899689729e-05,  2.042073489074e-05, {   4,   0,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  2.130240146809e-05,  1.059605091794e-05, {   0,   0,   1,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.608826070237e-05,  1.580067257098e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  2.050897849277e-05, -1.136717403887e-05, {   1,   1,  -1,   0,   0,   0, -29,  31,   0,   0,   0,   0,   0 } },
 {  2.965389372377e-05,  2.220739158658e-06, {   1,   0,   0,   0,   0, -21,  22,   0,  -1,   0,   0,   0,   0 } },
 {  2.852838518954e-05, -3.345414044350e-06, {   1,  -1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.426995376959e-05,  1.759051447269e-05, {   3,  -1,   0,   0,   0, -33,  30,   0,   0,   0,   0,   0,   0 } },
 { -3.776790421266e-06,  2.807579058516e-05, {   2,   0,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.519932318257e-05, -6.635315217916e-06, {   1,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.592979722301e-05,  5.901556799389e-06, {   2,   0,  -2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -3.142726159554e-05,  3.879444820333e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0 } },
 { -6.085157788191e-06,  2.572961429001e-05, {   3,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.417754419624e-05,  7.628655565088e-06, {   1,   1,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.656470586437e-05,  1.521935124103e-05, {   0,   0,   2,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.409483066270e-05, -1.767986119698e-05, {   1,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -2.634074865341e-05,  5.419320112881e-06, {   1,  -1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.593574300507e-05, -1.581692400679e-05, {   3,   0,  -3,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 {  2.686134749765e-05,  4.881837661354e-06, {   1,   1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.121178110855e-05, -2.050135185809e-05, {   4,   0,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  3.112999428109e-05,  5.764752491936e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 { -8.927674081444e-06, -2.276813201329e-05, {   1,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.423905065016e-05,  1.745586647833e-05, {   1,  -1,   1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.245549959719e-05,  1.923049831782e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.157831667354e-05,  9.856983873401e-08, {   0,   0,   1,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.091992728739e-05, -7.399306749628e-07, {   2,   2,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.588264077499e-05,  5.774015458773e-06, {   4,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.164187791764e-05,  8.476062728743e-09, {   1,   0,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.053536825851e-05,  2.110770726232e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 {  2.100608154795e-05,  1.061074347020e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  6.687505887802e-06,  2.492230378651e-05, {   1,  -1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.071454789991e-05, -2.089498501197e-05, {   3,  -1,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.071454789991e-05, -2.089498501197e-05, {   1,   1,  -1,   0,  -2,   2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.117142112585e-06, -2.848023483145e-05, {   4,  -2,  -1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.099533931019e-05, -5.966706253198e-07, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.437456974015e-05, -1.721692087388e-05, {   1,  -1,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 { -1.742088079049e-06,  2.984724804561e-05, {   0,   0,   1,   0,   0,   6,  -6,  -8,   6,   0,   0,  -1,   0 } },
 {  1.742088079049e-06,  2.984724804561e-05, {   0,   0,   1,   0,   0,  -6,   6,   8,  -6,   0,   0,   1,   0 } },
 { -2.261263464251e-08, -3.154865846747e-05, {   0,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 {  1.003586876215e-05, -2.151882097709e-05, {   2,   0,  -1,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.971722869315e-05, -1.183039588180e-05, {   2,   0,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0 } },
 { -2.137567708216e-05,  1.016460169884e-05, {   2,   0,   1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.152169938040e-05, -1.020659219642e-08, {   0,   0,   2,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  6.158776713745e-07, -3.091487829725e-05, {   3,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.698202560479e-05, -4.532235495924e-06, {   2,   0,  -3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.544772386389e-05,  6.027021006756e-06, {   1,  -1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.074289849606e-05, -7.059025729771e-07, {   2,   2,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -3.109395741500e-05,  3.465446358896e-07, {   6,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.486221069578e-05,  6.573152315995e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   2,   0 } },
 {  2.013485458393e-06, -2.941639394578e-05, {   2,   0,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.405553557589e-05,  7.372103417687e-06, {   1,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.699878017530e-05, -4.423597043819e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -9.017561240720e-07, -3.051734593591e-05, {   1,  -1,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -9.017561240720e-07, -3.051734593591e-05, {   1,  -1,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -6.929087080707e-06,  2.448810060638e-05, {   2,   0,  -1,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 { -3.023565482390e-05,  1.150737675603e-06, {   2,   0,  -1,   0,   0,   6,  -8,   0,   2,   0,   0,   0,   0 } },
 { -2.188169841136e-05,  9.504257293265e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -1,   0,   0,   0 } },
 { -3.683935225410e-06,  2.769551453629e-05, {   3,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -9.846829358880e-06, -2.152630603533e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -2.551025327106e-06,  2.877195707111e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -5.332282889882e-06,  2.596953525950e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 { -8.974855692818e-06,  2.232371369271e-05, {   2,   0,  -2,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  3.063429755658e-05, -6.627138553178e-07, {   0,   2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.524955492674e-05,  6.046417890374e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -5.900307127985e-06,  2.538889249256e-05, {   3,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.086581468295e-05, -1.042151242059e-05, {   4,   2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.430424991750e-07,  3.103946725359e-05, {   2,   0,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  9.797964455926e-06,  2.146799674663e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  5.970766517329e-06,  2.528423887760e-05, {   1,   1,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.288436088388e-05, -1.836362154192e-05, {   0,   2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.118055444367e-05,  4.193255515138e-08, {   3,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.110185838168e-05, -1.098629076817e-07, {   2,  -2,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.204130036910e-05,  1.916589979658e-05, {   1,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.026827456506e-05,  9.292587414236e-07, {   2,   0,  -4,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.830790836032e-05,  2.888196965357e-06, {   2,   0,  -1,   0,   0,   0,  14, -20,   0,   0,   0,   0,   0 } },
 { -3.117736051464e-05,  1.064028383843e-08, {   2,  -2,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.810597524828e-05,  3.069070550587e-06, {   0,   0,   1,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 {  2.936812209186e-05, -1.805885316374e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  1.103804114437e-05,  2.013347059432e-05, {   2,   0,   0,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.781660709106e-05, -1.332243058475e-05, {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -3.106002730036e-05,  7.709935048476e-08, {   2,   0,  -2,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.141991100429e-05,  9.709670236221e-06, {   2,   0,  -2,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 {  4.701308815986e-06,  2.642560083119e-05, {   0,   0,   1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.490472865190e-05,  6.216675956131e-06, {   1,  -1,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.948137403117e-05, -1.163909707080e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 {  2.322491025439e-05,  7.892478825150e-06, {   2,   0,  -2,   0,   0,   5,  -9,   5,   0,   0,   0,   0,   0 } },
 {  1.449475282781e-05,  1.661288266346e-05, {   2,   0,   1,   0,   0, -10,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.815238761011e-06,  2.228901770561e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 {  3.105838836909e-05,  3.985801858140e-08, {   2,   0,   1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.795751588482e-06,  2.429061430676e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 { -2.345719825285e-05,  7.624004896421e-06, {   1,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.166711866293e-05,  1.941355334535e-05, {   0,   0,   1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 { -1.353540182798e-05, -1.754215260878e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 { -2.560460717707e-05,  5.460200986457e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -9.315182530372e-06,  2.173171642563e-05, {   4,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.019178452819e-05,  2.084374162359e-05, {   2,   0,   0,   0,   0,  -6,   9,   0,  -3,   0,   0,   1,   0 } },
 { -2.422261141879e-05,  6.811188257361e-06, {   0,   0,   2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.661191268223e-06, -2.236946808456e-05, {   2,   0,  -1,   0,   0,   0,  -2,   2,   0,  -2,   0,   0,   0 } },
 {  2.069259001605e-05, -1.033749536469e-05, {   0,   2,  -1,   0,   0, -20,  19,   0,  -1,   0,   0,   0,   0 } },
 {  3.100816909573e-05,  1.974364544051e-08, {   4,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.757778933056e-06,  2.926057060410e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 {  2.177018992310e-06,  2.882149837606e-05, {   0,   0,   1,   0,   0,   0,   6,  -8,   3,   0,   0,   0,   0 } },
 { -1.391104289198e-06,  2.960636551805e-05, {   2,   0,   1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.074049385050e-05, -1.025180532491e-05, {   1,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.422724137641e-06,  2.255070770191e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  1.914326487137e-05, -1.182683280658e-05, {   0,   0,   1,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.381383855089e-05,  1.715381710545e-05, {   4,   0,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  2.072264855929e-05, -1.024329404223e-05, {   1,   0,   0,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.048618828224e-05,  2.046847689053e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  1.321812071143e-05,  1.773392558184e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 {  3.087629596830e-05,  5.935755377410e-08, {   1,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.816691803958e-05, -2.759458759520e-06, {   2,   0,   0,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 {  3.088909800045e-05,  3.219318261073e-08, {   0,   2,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.782520480272e-06,  2.512133168220e-05, {   1,   1,   1,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.085673614374e-05,  9.843958828550e-09, {   6,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.379801159543e-05,  7.062314984716e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -8.580355215197e-06,  2.227797718514e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.601973711142e-05,  1.483464701231e-05, {   2,   0,  -2,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.711532854672e-06,  2.613131434143e-05, {   0,   0,   0,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 {  3.082378820998e-05,  1.868528137469e-08, {   1,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.259698756160e-05,  1.823835380471e-05, {   2,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  9.526331766807e-06,  2.129439210561e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  5.874956798042e-06, -2.493936377907e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 {  1.842611642845e-05,  1.238281686933e-05, {   2,  -1,   0,   0,   0, -10,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.758477216635e-05,  1.322147959481e-05, {   0,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.064036799538e-05, -2.014415746336e-05, {   2,  -2,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.015261296407e-06, -2.976239535474e-05, {   2,  -2,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.205773078343e-05,  1.871956217478e-05, {   2,   2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.024124882433e-05,  2.052000646060e-05, {   1,  -1,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  4.814525973424e-06,  2.594077028208e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  9.091260860285e-06,  2.165940515812e-05, {   0,   0,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.161227987037e-05, -1.912073134904e-05, {   4,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.473994145093e-05, -1.597712582565e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 { -1.971659145030e-05, -1.099597523165e-05, {   2,  -2,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.003547900318e-05,  2.066959522542e-05, {   2,   0,  -2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -2.317676611381e-05, -7.519464790160e-06, {   1,   1,  -1,   0,   0, -19,  21,  -5,   0,   0,   0,   0,   0 } },
 {  8.766244335666e-06,  2.189670077262e-05, {   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.058591538045e-05, -5.045448947760e-08, {   4,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.484845816614e-05, -1.576819500206e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.203788792646e-05,  8.577407832038e-06, {   1,  -1,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  1.286537266933e-05, -1.774061015275e-05, {   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  5.896209591293e-06,  2.470746842912e-05, {   2,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.980954494794e-05,  1.078662176903e-05, {   1,   0,   0,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 {  3.042943699251e-05,  1.393103362308e-07, {   2,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.770297238652e-06, -2.479748468370e-05, {   1,  -1,   1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.694596695131e-05,  3.620711705653e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -1.909354953897e-05,  1.145713349250e-05, {   4,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.143914837905e-05,  9.110482826103e-06, {   1,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.143914837905e-05,  9.110482826103e-06, {   1,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.872786966011e-06,  2.667389401925e-05, {   2,   0,  -1,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 { -2.819346571491e-05, -2.336064808427e-06, {   1,  -2,   1,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 { -2.819346571491e-05, -2.336064808427e-06, {   1,  -2,  -1,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 { -9.387855056454e-06,  2.113921411773e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  3.019248642656e-05,  3.322028231984e-07, {   2,  -2,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.380600721045e-05, -6.718673975787e-06, {   2,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.204499940118e-05, -1.846218525334e-05, {   3,   0,  -3,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -1.394627697024e-05,  1.655781214449e-05, {   2,   0,  -2,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.182702953939e-05,  1.867704653145e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  8.583609808034e-06, -2.192029988209e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0 } },
 { -2.052647425043e-05, -9.975881830491e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   2,   0,   0,   0 } },
 { -5.137879191196e-06, -2.535836772525e-05, {   2,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.180023488857e-06, -2.429834803242e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 {  1.075833258522e-06, -2.938269809315e-05, {   2,   0,  -2,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.260603302186e-05,  1.785058348612e-05, {   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  4.342920070451e-06,  2.610871834562e-05, {   2,  -2,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -9.370710562214e-06,  2.107597866887e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.123416894328e-05, -1.919243027855e-05, {   0,   0,   3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.159818692069e-05, -8.818672233449e-06, {   1,   1,  -1,   0,   0, -14,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.135208606564e-05, -9.063486277924e-06, {   1,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.958161346735e-05, -1.082200365686e-05, {   0,   0,   0,   0,   0,   0,   4,  -7,   2,   0,   0,   0,   0 } },
 {  5.311687539160e-07,  2.986223295913e-05, {   2,   0,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.932869343794e-06, -2.544892084248e-05, {   4,   0,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  5.794836738288e-06, -2.457671541511e-05, {   3,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.037608099898e-05, -1.999223715560e-05, {   2,  -1,   0,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.744347701677e-06, -2.460385312787e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 {  2.234176149861e-05, -7.998397801257e-06, {   2,  -2,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.099605624081e-05,  1.934128703345e-05, {   0,   0,   0,   0,   0,   6, -14,   8,   0,   0,   0,  -1,   0 } },
 { -2.517530315470e-05, -5.153450657405e-06, {   1,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.543088654252e-05, -4.895596370033e-06, {   1,  -1,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.031527732667e-05,  5.583469911431e-09, {   2,   0,   2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.366866295734e-05, -6.632178379789e-06, {   2,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.076275761348e-06, -2.420977260196e-05, {   2,   0,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.802789380746e-06, -2.447707437064e-05, {   1,  -1,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.516815506616e-06,  2.374789266116e-05, {   1,   1,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.287193891497e-05, -1.738664564070e-05, {   1,  -1,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.254633574570e-07,  3.002361875477e-05, {   2,   0,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.087418870255e-05,  9.369652118117e-06, {   2,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.213653957625e-07,  3.011185762483e-05, {   1,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.021760259344e-05,  1.482156194789e-08, {   3,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.501277161944e-06,  2.672577273297e-05, {   0,   0,   1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.053102144773e-05,  1.969348589827e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   1,   3,   0,   0,   0 } },
 { -8.075114523937e-07, -2.941489305622e-05, {   4,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  2.936887198469e-05, -8.436377014556e-07, {   2,  -2,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.018159278222e-05, -9.996114873559e-06, {   1,   0,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.495986092376e-05, -1.520611653625e-05, {   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.009125684410e-05, -4.259331052540e-08, {   3,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -7.692627773754e-06, -2.243712351480e-05, {   2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.985240785682e-05, -2.617509015469e-07, {   2,   0,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -2.580764203944e-05,  4.296514567136e-06, {   3,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -7.808101648245e-06,  2.228376579548e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  1.338170248587e-05, -1.670931034874e-05, {   1,  -1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -2.512074635070e-05,  4.963366874136e-06, {   0,   0,   0,   0,   0,   2,  -1,  -5,   8,  -6,   0,   0,   0 } },
 { -2.920815533681e-05, -8.668405192302e-07, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 {  1.883975326547e-05,  1.121751103471e-05, {   0,   2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.999999998136e-05,  1.057669526489e-09, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999998453e-05,  9.634217469353e-10, {   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999192e-05, -6.963863714810e-10, {   0,   1,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999228e-05, -6.806784082122e-10, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.999999999228e-05, -6.806784082122e-10, {   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999353e-05, -6.230825429215e-10, {   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999417e-05, -5.916666163821e-10, {   0,   3,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999496e-05, -5.497787143434e-10, {   2,  -1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999534e-05,  5.288347633269e-10, {   4,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999543e-05,  5.235987755717e-10, {   4,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999622e-05, -4.764748857768e-10, {   2,   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999638e-05, -4.660029102628e-10, {   4,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999999662e-05,  4.502949469976e-10, {   0,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999693e-05, -4.293509959783e-10, {   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999722e-05, -4.084070449498e-10, {   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.999999999729e-05, -4.031710571926e-10, {   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999729e-05, -4.031710571926e-10, {   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999743e-05,  3.926990816875e-10, {   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999999776e-05,  3.665191429097e-10, {   0,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999776e-05, -3.665191429070e-10, {   2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999999789e-05,  3.560471673985e-10, {   6,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999890e-05,  2.565634000400e-10, {   2,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999999895e-05, -2.513274122896e-10, {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.996736250198e-05, -2.854911366258e-08, {   2,  -2,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.389885343210e-05,  1.608568732705e-05, {   2,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.057714469201e-07, -2.987874948543e-05, {   2,   0,  -1,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 { -9.194040626599e-06, -2.078832787316e-05, {   2,   0,  -2,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.158887221201e-05, -8.393226374663e-06, {   1,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.866990821221e-06, -2.410885234200e-05, {   2,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.140320611299e-06, -2.482672012645e-05, {   2,   0,   1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.015140477914e-05,  1.979870625941e-05, {   0,   0,   0,   0,   0,  16, -26,   0,   0,   0,   0,   0,   0 } },
 { -2.006808590644e-05, -9.879738782359e-06, {   0,   0,   2,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  1.647474067692e-07, -2.976413161706e-05, {   5,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.007937430764e-05, -1.982988152228e-05, {   2,   0,  -2,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.926886765792e-05, -1.063739514474e-05, {   2,   0,  -3,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.409628191399e-05,  1.580939920064e-05, {   1,   0,  -1,   0,   0,   0, -33,  41,  -2,   0,   0,   0,   0 } },
 {  1.999474371800e-05, -9.902809142765e-06, {   1,   0,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.950611593152e-06,  2.792057674348e-05, {   0,   0,   2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -2.946109728486e-05, -3.947842320653e-07, {   1,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.723625442949e-07, -2.938041561136e-05, {   2,  -2,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.300755175752e-05,  1.683498038708e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.467177931362e-05, -1.516819687598e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 {  2.105614933858e-05, -8.774247013314e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 {  6.252223780507e-06,  2.356247844330e-05, {   0,   0,   2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -5.703506600298e-06,  2.410950714373e-05, {   5,  -1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.907725318017e-05, -7.282203285118e-07, {   4,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.579220681413e-05,  1.400363260038e-05, {   2,   0,   0,   0,   0,  -8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.992603156410e-05, -9.831041027838e-06, {   2,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -9.227913793879e-06, -2.052885019415e-05, {   1,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.185358781453e-06, -2.855625619621e-05, {   2,  -2,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.815406983864e-06, -2.792595949511e-05, {   4,   0,  -2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.088021177701e-05, -8.851215433567e-06, {   1,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.167517729983e-05,  1.805490740457e-05, {   2,  -2,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -2.914353742637e-05,  5.842738314940e-07, {   2,   0,  -2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -2.259536694556e-05,  7.129645054493e-06, {   3,  -1,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.340523617034e-05,  6.303851626785e-06, {   0,   2,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.480823492710e-06,  2.522452551688e-05, {   0,   0,   0,   0,   0,   2,   1,  -9,   8,  -6,   0,   0,   0 } },
 { -5.787157933167e-06,  2.391628762459e-05, {   4,   0,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.339878954007e-05,  6.302845630387e-06, {   0,   2,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.164110908863e-05, -1.805155158863e-05, {   1,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  5.666404545967e-06, -2.399885535504e-05, {   0,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.035834888547e-05,  9.290157432862e-06, {   0,   0,   0,   0,   0,   2,  -4,   3,   0,   0,   0,   0,   0 } },
 { -1.401947490123e-05,  1.560184745835e-05, {   0,   2,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.258939305975e-06, -2.534119339678e-05, {   1,  -1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.304292312143e-05,  1.655311405142e-05, {   0,   0,   0,   0,   0,  13,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.067567708972e-05, -8.914506012147e-06, {   1,   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.213350639114e-05,  7.422842052922e-06, {   1,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.242563730736e-05,  1.712599043516e-05, {   0,   0,   1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.502465030260e-05, -1.451742745901e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -1.152312803231e-05,  1.801848825409e-05, {   2,   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.765646812655e-05, -1.879098883656e-06, {   4,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.202352338582e-06,  2.733192795332e-05, {   2,   0,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.336897727757e-05, -1.615771212279e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 { -1.770034099301e-05,  1.181629918491e-05, {   1,  -1,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -7.179132234483e-06,  2.233167593524e-05, {   0,   0,   2,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.452265020851e-06, -2.505706797023e-05, {   2,   0,   1,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 { -1.901091228640e-05,  1.049813520794e-05, {   0,   0,   1,   0,   0, -10,  13, -19,   0,   3,   0,   0,   0 } },
 {  9.045029438110e-06, -2.045243397217e-05, {   2,   0,  -2,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -1.303883897435e-06,  2.817434177031e-05, {   0,   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 {  2.055296044324e-05,  8.898357733581e-06, {   2,   0,   0,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 { -2.449798384360e-05, -4.921924780639e-06, {   1,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.430814041831e-05, -5.107591544112e-06, {   1,   2,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -2.430814041831e-05, -5.107591544112e-06, {   1,  -2,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.879175969400e-05, -1.057605831774e-05, {   2,   2,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.825260079067e-05, -1.111505634059e-05, {   2,   0,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0 } },
 { -8.586385130049e-06,  2.077792993279e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  4.233852718057e-06, -2.511573211677e-05, {   1,  -1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.977878513861e-05,  9.570393657717e-06, {   0,   0,   0,   0,   0,   0,   4,  -6,   1,   0,   0,   0,   0 } },
 {  1.387019829947e-05,  1.547675545316e-05, {   0,   2,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.251589838609e-05, -1.681725061519e-05, {   0,   0,   3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.746282136203e-06,  2.257268265639e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 { -1.317137567846e-05,  1.614237598639e-05, {   1,  -1,   0,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 { -1.954101397945e-05,  9.738387202721e-06, {   0,   0,   1,   0,   0, -21,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.912939383379e-05,  1.014950546379e-05, {   1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.559544041888e-07,  2.881401200585e-05, {   4,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.231678287717e-06, -2.602512545653e-05, {   2,   0,   0,   0,   0,  -5,   5,   4,   0,   0,   0,   0,   0 } },
 {  1.083808153579e-05, -1.841708935155e-05, {   0,   0,   3,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.474324703555e-05, -4.497251538344e-06, {   1,  -1,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.312643778711e-05, -1.609685901327e-05, {   1,   1,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.391353389481e-05,  5.301322861380e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -5.614946152755e-06,  2.354931927567e-05, {   4,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.294930907112e-05, -6.204021018015e-06, {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.855140808606e-05, -1.059554485260e-05, {   1,   0,  -1,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 { -2.010528718572e-05,  9.041182749281e-06, {   2,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -2.293580464551e-05, -6.201422476532e-06, {   0,   0,   4,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.434024709852e-05, -4.779588156825e-06, {   1,  -1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.208978166555e-05, -7.022319238654e-06, {   1,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.898038485296e-05, -1.012988054070e-05, {   4,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.757442899438e-05,  1.520092887175e-06, {   2,   0,  -1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 {  7.741543099686e-06, -2.134875398624e-05, {   0,   0,   0,   0,   0,   6, -10,   0,   1,   5,   0,   0,   0 } },
 { -2.874790619376e-05,  3.367990275419e-07, {   2,   0,  -3,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.817815031609e-05, -8.909735712450e-07, {   0,   0,   2,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 { -1.311239261866e-05,  1.594122132046e-05, {   1,   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.220047651724e-05, -1.683612698250e-05, {   2,   0,   3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.133172154266e-05, -7.684483804840e-06, {   2,   0,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.353603000067e-05, -5.477685294362e-06, {   1,   1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.943108209087e-06, -2.405133346550e-05, {   1,  -3,   2,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  4.943108209087e-06, -2.405133346550e-05, {   1,  -3,  -2,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  2.754051538611e-05, -1.441024177071e-06, {   2,   2,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.370901228032e-06, -2.158849423130e-05, {   1,  -1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.525634162066e-05, -3.697333727295e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 {  3.018017047909e-06, -2.592355149008e-05, {   0,   0,   1,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 {  2.094253967203e-05, -7.983124443940e-06, {   0,   0,   0,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 {  2.196121947395e-05,  6.946155595871e-06, {   4,   0,  -2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.227498480691e-05, -1.662579897092e-05, {   2,   0,   0,   0,  -6,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.213703013250e-05, -1.676098410716e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 {  1.627121724178e-05,  1.262661552872e-05, {   2,   0,  -1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 { -9.518134229515e-06, -1.936762016866e-05, {   1,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  4.969485383319e-06,  2.391524726735e-05, {   2,   0,  -1,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.370103222088e-06,  2.250755266373e-05, {   2,  -2,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.872934596583e-05, -1.419750455118e-07, {   2,   0,   3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  6.364740198453e-06,  2.249284824860e-05, {   2,  -2,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.282132437105e-06,  1.957172130349e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -2.280134920457e-05,  6.039340133563e-06, {   1,  -1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.423769943151e-06, -2.441166112622e-05, {   3,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.484175367091e-05,  3.988484352089e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 {  1.787958244925e-05, -1.095048483896e-05, {   2,  -2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.378114000470e-06,  2.043975868274e-05, {   2,   0,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.762291775231e-05, -1.118883258409e-05, {   1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.589406512878e-06, -2.721416145515e-05, {   0,   0,   1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.939731882067e-05,  9.401067787310e-06, {   2,  -2,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.282653526024e-05, -1.596400372512e-05, {   1,  -1,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.694266834443e-05,  1.184688891883e-05, {   1,  -1,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.694266834443e-05,  1.184688891883e-05, {   1,  -1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  3.403554208870e-06, -2.536724709363e-05, {   2,   0,  -1,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.852197260147e-05, -1.024727068816e-05, {   2,   0,  -2,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.522353899273e-05,  1.354523576495e-05, {   4,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -9.693623258778e-06, -1.906854363389e-05, {   1,  -1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  8.418634083653e-06,  2.034275344214e-05, {   2,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  6.669455546853e-06, -2.208163092547e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0 } },
 {  1.822544989290e-05,  1.052139325394e-05, {   0,   2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.666755702041e-05,  1.207050507448e-05, {   1,   0,   0,   0,   0,   0, -31,  36,  -4,  -5,   0,   0,   0 } },
 {  5.487285346199e-06, -2.324441278887e-05, {   1,   1,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.015868733451e-05,  8.566358650206e-06, {   0,   2,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.015868733451e-05,  8.566358650206e-06, {   0,   2,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.286113206062e-05,  5.852916687091e-06, {   1,   1,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.190462097713e-05, -1.677930921146e-05, {   0,   0,   1,   0,   0, -24,  29,  -7,   8,  -6,   0,   0,   0 } },
 { -1.987885931631e-05,  8.787850486448e-06, {   1,  -1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.775957447560e-05, -8.947746185803e-07, {   2,  -2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.581268416500e-05,  2.831849212145e-06, {   4,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.122330143342e-07, -2.812821033641e-05, {   4,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.669716536755e-06, -2.296175268159e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.915032800862e-06, -2.670799553543e-05, {   0,   0,   1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -5.473794287764e-06, -2.314513742659e-05, {   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.819858548648e-05, -1.041143473311e-05, {   2,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.854528341021e-05, -6.324816622702e-08, {   4,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.552548487753e-05, -1.307395684761e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,  -2,   5,   0,   0,   0 } },
 { -1.835534112057e-05, -1.023104087737e-05, {   0,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.411850024440e-05,  1.445334813504e-05, {   0,   2,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.835221219243e-05, -2.038037538548e-07, {   2,  -2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.782311819441e-06, -1.975535121768e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  6.026231119812e-06,  2.250088083025e-05, {   1,  -1,  -3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.739918935368e-05,  1.112568007202e-05, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  6.344495414823e-06, -2.217351066927e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.234758698484e-05, -1.616702282447e-05, {   2,   0,  -3,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.832280411719e-05, -1.875009565331e-07, {   2,   0,   1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  9.185862972412e-06,  1.930848853750e-05, {   0,   0,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 { -8.752527501736e-06, -1.973222392055e-05, {   0,   0,   1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 { -1.820375558645e-05,  1.026560866265e-05, {   0,   2,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.444517329850e-06,  2.302308854454e-05, {   4,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.895571950093e-05,  9.510787086114e-06, {   0,   0,   2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.844186054093e-05,  1.054577810468e-08, {   2,   0,  -4,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.016031455916e-05, -1.828613612944e-05, {   0,   2,  -1,   0,   0, -14,   9,   0,   2,   0,   0,   0,   0 } },
 {  4.923523567759e-06, -2.351069794950e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  2.416207891188e-05, -4.263832197497e-06, {   1,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.704048959013e-05, -1.367163466361e-06, {   2,   0,  -2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.811847542194e-05,  2.777921297571e-07, {   0,   0,   0,   0,   0,   0,   3,  -5,   2,   0,   0,   0,   0 } },
 { -1.903773207348e-05,  9.353605988496e-06, {   2,  -2,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.496766339764e-05, -3.392730376966e-06, {   1,  -1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.687598308537e-06,  2.565712159826e-05, {   1,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.799625490481e-06, -2.254160370648e-05, {   1,   1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.226432775141e-05, -6.072792565622e-06, {   2,   0,   0,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  2.362092135041e-05,  4.712202147191e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  5.986370040726e-06,  2.234009892457e-05, {   1,  -1,   3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.446709249295e-05,  1.385103277182e-05, {   0,   2,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.412422850988e-06, -2.289174549027e-05, {   1,   1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.374432970411e-05,  4.554606018156e-06, {   1,  -1,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.727830473856e-05,  1.101821563729e-05, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -5.413727388201e-06,  2.288040567094e-05, {   3,  -1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.387434083882e-05,  1.440495283881e-05, {   0,   0,   2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.118447594054e-05,  1.709473276258e-05, {   2,   0,   0,   0,   0, -10,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.921126319273e-06,  1.935146154675e-05, {   1,   0,   0,   0,   0,   0, -15,   4,   4,   5,   0,   0,   0 } },
 { -2.781245898486e-05, -4.526651578832e-07, {   4,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.784009916166e-05, -4.165280510487e-07, {   2,  -2,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.347610251506e-05, -4.775212572188e-06, {   2,   0,  -1,   0,   0, -19,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.757839956117e-05,  1.066137525956e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   6,  -5,   0,   0,   0 } },
 { -2.561503787490e-05, -2.622158871137e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   6,   0,   0,   0 } },
 { -2.434486139487e-05,  3.890299582323e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 {  9.354935221060e-06,  1.887584043266e-05, {   0,   0,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.787785606020e-05,  1.034363968854e-05, {   0,   2,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.567449700381e-06, -2.065270562750e-05, {   2,   0,   0,   0,   0,   0,  -2,   2,   0,  -2,   0,   0,   0 } },
 {  1.897938791565e-05, -9.238127291260e-06, {   1,   0,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.219648795831e-05, -6.016778766041e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -1.796949022457e-05, -1.024258038718e-05, {   2,   2,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.144650334419e-05, -1.676461575233e-05, {   2,   0,  -1,   0,   0,   0,   1,  -8,   3,   0,   0,   0,   0 } },
 { -8.682820973435e-06, -1.952401320423e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  2.285375398393e-05,  5.344209865162e-06, {   1,  -1,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -5.377640952548e-06,  2.281786008073e-05, {   2,   0,   1,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  8.494274800829e-06, -1.969491731315e-05, {   0,   0,   1,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.423152858588e-06, -2.576208146485e-05, {   0,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.170167658168e-05,  1.647390705819e-05, {   3,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.124151630975e-05,  6.921536317600e-06, {   2,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.766413167026e-05,  4.879341478991e-07, {   0,   0,   1,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 {  1.297160692464e-05, -1.515769639033e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -2.798705289467e-05,  1.358588883485e-07, {   2,   0,  -2,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  2.392862521988e-05,  4.194251055231e-06, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 {  1.254278893111e-05,  1.557813999904e-05, {   4,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.263463513081e-05, -5.484080358042e-06, {   0,   1,   0,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  2.635347159573e-05, -1.755618568916e-06, {   0,   0,   3,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.400906257348e-05,  1.409586553148e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 {  7.863174269135e-06,  2.024130329315e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 { -2.045760153516e-05,  7.644982360124e-06, {   3,  -1,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  2.045760153516e-05,  7.644982360124e-06, {   1,   1,   1,   0,   0, -29,  33,   0,  -2,   0,   0,   0,   0 } },
 {  2.287710954666e-05, -5.224951162968e-06, {   2,   0,   1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -1.961561324416e-05, -8.482989269847e-06, {   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   7,   0,   0,   0 } },
 {  1.286374117867e-05,  1.523396972828e-05, {   2,   0,  -2,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.025342540981e-05, -1.784384586162e-05, {   4,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.241292706787e-05, -1.568181244652e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.802849396199e-05,  6.526011496054e-08, {   0,   0,   2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  7.746462741951e-06, -2.034726431651e-05, {   1,  -1,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -9.520736035147e-06,  1.856687177887e-05, {   1,  -1,   2,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.520736035147e-06,  1.856687177887e-05, {   1,  -1,  -2,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.645909266604e-05,  1.627219347956e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 { -1.237252082447e-05,  1.571371147909e-05, {   2,   0,   2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.268057552995e-05,  5.390650802952e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 {  1.116617860939e-05,  1.690353995235e-05, {   2,   0,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.097986087865e-05,  1.708930565308e-05, {   1,  -1,   0,   0,   0,   0,  -4,  10,  -3,   0,   0,   0,   0 } },
 { -9.335548840534e-06,  1.872198253340e-05, {   0,   1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.011733489732e-06,  2.303986105365e-05, {   2,   0,   1,   0,   0, -18,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.601562906384e-06,  2.144350363875e-05, {   2,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  2.796136411826e-05,  7.731298533412e-08, {   0,   0,   1,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  1.926501506571e-05,  8.769221635246e-06, {   1,   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.360392177974e-05, -1.440943547169e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -4,   0,   0,   0 } },
 {  5.396450409383e-06,  2.261127770192e-05, {   0,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.107378299393e-07, -2.759681368181e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.799715900842e-05,  9.331658176944e-09, {   1,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.429311130124e-05,  1.370232849152e-05, {   0,   0,   0,   0,   0,   0,   2,  -5,   3,   0,   0,   0,   0 } },
 { -2.290725596888e-05,  5.087833457183e-06, {   0,   0,   1,   0,   0, -28,  29,   7,  -8,   6,   0,   0,   0 } },
 {  2.291096768569e-05,  5.075634889798e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   3,  -7,   0,   0,   0 } },
 { -1.660370160872e-05,  1.137359089902e-05, {   1,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  9.016789847016e-06, -1.895402345161e-05, {   0,   0,   1,   0,   5,   0, -13,   7,   0,   0,   0,   0,   0 } },
 { -2.788815385896e-05,  8.206633248462e-08, {   2,   0,   0,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 {  2.702912797259e-05,  9.249578094820e-07, {   1,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.006774185070e-06, -1.893292030681e-05, {   4,   0,  -1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -1.806921181541e-05,  9.854863172406e-06, {   4,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  2.789252376148e-05, -2.999987732265e-08, {   4,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.392448734787e-05,  1.398799153324e-05, {   1,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.577348438447e-05, -1.213237189859e-05, {   0,   0,   1,   0,   0,   0, -11,  16,   0,   0,   0,   0,   0 } },
 { -9.923052678190e-07,  2.690043855873e-05, {   0,   0,   1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  8.523395729510e-06, -1.936897907732e-05, {   2,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.160685627670e-05, -6.283601661612e-06, {   1,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.922116299326e-06,  2.195424486733e-05, {   0,   0,   1,   0,   0, -25,  28,   0,   0,   0,   0,   0,   0 } },
 {  8.374240064753e-06, -1.949844558665e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  2.783378725290e-05,  3.823505696986e-08, {   0,   2,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.282684538500e-05,  1.503463479595e-05, {   2,   0,  -1,   0,   0, -11,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.196296132715e-05, -5.898393565179e-06, {   2,   0,   0,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.233766396173e-06, -2.262463713043e-05, {   2,  -1,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.243263991014e-05,  5.425210995291e-06, {   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0 } },
 {  1.045764544072e-05,  1.738879155979e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -1.269163925466e-05, -1.515399471179e-05, {   3,  -2,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  1.269163925466e-05, -1.515399471179e-05, {   1,   2,   0,   0,   0, -20,  18,   0,   0,   1,   0,   0,   0 } },
 { -1.883901553981e-05,  8.993834330820e-06, {   2,   0,   1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.749454629377e-05, -3.338926149634e-07, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 {  2.782498020274e-05,  1.967612861500e-09, {   2,   0,   1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.202940430024e-05,  1.579140285485e-05, {   2,   0,  -1,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 {  1.110583045961e-06, -2.670769141303e-05, {   0,   0,   2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.461801967665e-05, -1.319511559550e-05, {   0,   2,  -1,   0,   2, -27,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.031388159592e-05,  7.486962695803e-06, {   2,   0,  -2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.264587032759e-05, -1.514667749587e-05, {   0,   0,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 {  2.774226273124e-05,  4.855656108230e-08, {   4,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.741560669427e-05,  3.561122125380e-07, {   4,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.777198192850e-05,  9.997260081390e-06, {   1,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.630733194540e-05, -1.146167057660e-05, {   1,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -2.729232565776e-05,  4.553850876160e-07, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 {  5.798824830921e-06, -2.194231742350e-05, {   0,   0,   1,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -8.945598520363e-06, -1.879442696565e-05, {   4,  -2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -7.298566240142e-06,  2.042227372836e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -7,   5,   0,   0,   0 } },
 {  2.596846521661e-05,  1.739894418608e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  3.077186920777e-06, -2.463091556464e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 {  1.475559199036e-05, -1.295157293406e-05, {   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.672802517529e-05, -1.096235256389e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.471129024786e-05,  2.978622977304e-06, {   1,   0,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.756693428649e-05,  1.078316699388e-07, {   2,   0,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.293511281156e-06,  2.237937732530e-05, {   3,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.836302951902e-06,  1.980873238289e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 { -4.136393927672e-06,  2.349335273380e-05, {   1,   0,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.193986328549e-05,  1.568889980057e-05, {   2,   0,  -1,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 {  2.717207738264e-05,  4.530678364080e-07, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 { -3.715126796271e-06, -2.390415995165e-05, {   2,  -2,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.098426772889e-06,  1.851290812760e-05, {   0,   0,   1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  5.309683971995e-06, -2.229586911413e-05, {   1,  -1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.671242574203e-05, -8.927160462141e-07, {   2,   0,   3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.725222890889e-05, -3.427435558186e-07, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 { -9.522177871573e-06, -1.807220253895e-05, {   2,   0,  -1,   0,   0,   4,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.409122060894e-05,  1.348999079044e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -4.939526619724e-06, -2.263861269641e-05, {   1,   1,  -1,   0,   0, -20,  20,   0,   0,  -1,   0,   0,   0 } },
 { -1.064372195997e-05, -1.693254803149e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.064372195997e-05, -1.693254803149e-05, {   2,   0,  -2,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.180255313498e-06, -2.239549219024e-05, {   2,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.709333293647e-06, -2.086597809748e-05, {   2,   0,  -1,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.687534150381e-05, -6.930606317535e-07, {   2,   0,  -1,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -9.462599064735e-06, -1.807090029266e-05, {   0,   2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.468597417317e-05,  2.844024603193e-06, {   2,   0,   1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -2.695443313696e-05,  5.735246584231e-07, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  2.008775623892e-05, -7.440192426045e-06, {   1,  -1,   0,   0,   0,   0,  -1,   4,  -1,   0,   0,   0,   0 } },
 {  2.434676349212e-05, -3.166497841054e-06, {   0,   0,   1,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 {  2.407260676570e-05,  3.421747228076e-06, {   1,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 { -2.633830609843e-05, -1.149247173638e-06, {   4,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.109573818417e-05,  1.639147562797e-05, {   2,   0,  -2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.333611633737e-05,  1.414543840135e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   1,   0,   0,   0 } },
 {  1.292697006419e-05, -1.454594678442e-05, {   0,   0,   1,   0,   0,   8,  -5, -16,   4,   5,   0,   0,   0 } },
 { -1.292697006419e-05, -1.454594678442e-05, {   0,   0,   1,   0,   0,  -8,   5,  16,  -4,  -5,   0,   0,   0 } },
 { -1.763961268406e-05,  9.828946464573e-06, {   1,   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -9.867426402223e-07, -2.647770694204e-05, {   6,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.154802789392e-05, -5.912671419812e-06, {   2,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.937960277664e-07, -2.656229473357e-05, {   0,   0,   1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -5.786529894383e-06, -2.166837603693e-05, {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.617102301941e-06, -2.383491379877e-05, {   2,   0,   0,   0,  -8,   0,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.990123264908e-05,  7.547790491451e-06, {   2,   0,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.363398274190e-06,  2.108125646846e-05, {   0,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.659368281863e-05, -8.497669334656e-07, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -2.720092704956e-05, -2.418687665570e-07, {   2,   0,   0,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  9.830339697427e-06, -1.760773848850e-05, {   4,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -2.509951866973e-05,  2.334630720321e-06, {   1,  -1,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -2.509951866973e-05,  2.334630720321e-06, {   1,  -1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.024705419069e-05,  1.717550233364e-05, {   1,  -1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.830807830821e-05,  9.111791006931e-06, {   2,   0,  -1,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.971703727189e-06,  2.144760120041e-05, {   2,  -2,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  2.740631515827e-05,  7.614410636539e-09, {   0,   2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.283212568363e-05,  1.456248062834e-05, {   2,   0,  -2,   0,   0,   0,  10, -16,   3,   0,   0,   0,   0 } },
 { -6.260663505185e-06,  2.113169142194e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -1.971205820069e-05,  7.671265550478e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 { -5.962830450814e-06,  2.141587935208e-05, {   2,  -2,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.625621481408e-05,  1.107506041241e-05, {   4,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.456810632726e-05,  2.755785115879e-06, {   1,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.149455745753e-06, -2.217319888912e-05, {   0,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.132489793335e-06, -2.218865235420e-05, {   2,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.667247065773e-05, -6.282957617578e-07, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -2.690966537500e-06, -2.460979950447e-05, {   2,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.750110162590e-05,  9.794195817243e-06, {   2,  -2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  7.978654709018e-06,  1.931435612297e-05, {   1,   1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.423840829914e-05,  3.041832705667e-06, {   2,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -2.372704841613e-05,  3.546311860895e-06, {   1,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  2.725428181352e-05,  1.519112105410e-08, {   3,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.190621239565e-06, -2.406773166571e-05, {   2,   0,  -1,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 { -5.672268651805e-06,  2.158161094684e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 { -1.520764543319e-06, -2.569860690473e-05, {   0,   0,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.436594616531e-05, -1.285104239318e-05, {   0,   0,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 { -1.758862314089e-05, -9.623385773461e-06, {   1,  -1,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  1.720403904582e-05, -1.000624065015e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  9.048493368734e-07, -2.629787563373e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   2,   0,   0,   0,   0 } },
 {  2.715270893252e-05,  3.062792558145e-08, {   3,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.843093738561e-05, -8.747507662928e-06, {   2,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.888622266275e-06, -1.828470685427e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.014880157584e-06,  2.614540654928e-05, {   1,  -2,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.240069596758e-05, -1.474362570469e-05, {   1,  -1,  -1,   0,   0,  19, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.053037191524e-05,  6.612255766016e-06, {   2,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 {  8.058703584708e-06, -1.907351697740e-05, {   1,   0,  -1,   0,   0,  11, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.291982625206e-05, -4.211734852094e-06, {   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.105886293977e-05,  1.606297266540e-05, {   0,   0,   0,   0,   0,   2,  -2,  -5,   8,  -6,   0,   0,   0 } },
 { -1.631631866197e-05, -1.078730454638e-05, {   2,   0,   0,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.972020543637e-05, -7.383272336375e-06, {   2,   0,   2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.098040157583e-05,  1.611821603194e-05, {   0,   0,   0,   0,   0,   2,   2,  -9,   8,  -6,   0,   0,   0 } },
 {  2.469407302364e-05,  2.392698645734e-06, {   2,   0,   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -6.777443365136e-06, -2.030787843429e-05, {   0,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  7.222397374584e-06,  1.985267725135e-05, {   2,   0,  -1,   0,   0,   0,  15, -25,   0,   0,   0,   0,   0 } },
 {  2.682704083766e-05, -2.358786238348e-07, {   2,  -2,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.773634407368e-05,  9.325378534859e-06, {   0,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.377278916305e-05,  1.327715475017e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 {  1.377278916305e-05,  1.327715475017e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 {  5.349908334010e-06,  2.169668090741e-05, {   1,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.902993878369e-06,  2.113379317059e-05, {   1,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.967799300295e-05,  7.357872188130e-06, {   3,  -1,  -2,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  1.967799300295e-05,  7.357872188130e-06, {   1,   1,   0,   0,   0, -29,  33,   0,  -2,   0,   0,   0,   0 } },
 {  1.173652648030e-05, -1.529263840804e-05, {   1,  -1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.212599059303e-05, -4.895680401376e-06, {   2,   0,   0,   0,   0, -19,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.075133769691e-06, -2.193689665750e-05, {   0,   0,   0,   0,   0,   9,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.685061832304e-05, -1.561385356034e-07, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  5.895709944075e-06,  2.110067185511e-05, {   1,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.067021153807e-06, -2.190464940262e-05, {   2,  -1,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  6.170775412854e-06, -2.080036211429e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.582197980394e-06,  2.138227080289e-05, {   1,   1,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.791346258919e-06,  1.817306819654e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 { -1.167047865814e-06,  2.578228246009e-05, {   2,   0,   0,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 { -9.133659521806e-06,  1.781201418946e-05, {   3,  -1,   0,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.133659521806e-06,  1.781201418946e-05, {   1,   1,   0,   0,  -2,   2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.228090244215e-05, -1.466356921515e-05, {   3,  -2,  -1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  1.228090244215e-05, -1.466356921515e-05, {   1,   2,  -1,   0,   0, -20,  18,   0,   0,   1,   0,   0,   0 } },
 { -2.259179856466e-05, -4.348774687682e-06, {   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.281003687192e-05,  4.122255830374e-06, {   2,   0,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.689067997364e-05, -2.423740681300e-08, {   2,  -2,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.600611861762e-05, -8.990123304870e-07, {   0,   0,   2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.617086131356e-05, -7.310836417907e-07, {   0,   2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.762038154184e-05,  9.276907483089e-06, {   2,   0,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  9.971596500215e-06, -1.691634950187e-05, {   2,   0,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0 } },
 { -5.146853697217e-06,  2.172296713786e-05, {   1,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.364777519836e-05,  3.214955427512e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.073476330283e-05, -6.119282375754e-06, {   2,   0,  -2,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  1.016160937913e-05, -1.668431607696e-05, {   0,   0,   0,   0,   0,   3,  -7,   4,  -2,   5,   0,   0,   0 } },
 {  1.232926436004e-05, -1.451088619546e-05, {   0,   0,   4,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.797797186469e-05, -8.853079385704e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   4,  -8,   0,   0,   0 } },
 {  2.255342100820e-05,  4.266965087414e-06, {   1,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.553561859657e-05,  1.281738254864e-06, {   6,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.767059213293e-05, -9.143153390932e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 { -5.116338892218e-06, -2.167564181308e-05, {   4,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.675419415462e-05,  2.899047500865e-08, {   0,   0,   1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.218274370993e-05, -1.458947750844e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 {  1.252901376547e-05, -1.424085541278e-05, {   0,   0,   4,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.495077816638e-06,  2.224229382486e-05, {   0,   0,   0,   0,   1,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.229853964989e-05, -1.443089286606e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   1, -11,   3,   0,   0 } },
 {  2.671665972566e-05,  9.792337571010e-09, {   2,   0,   1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.085172947243e-07, -2.591521453736e-05, {   2,   0,  -1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  4.283398273129e-06, -2.243953418947e-05, {   2,   0,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.353292890339e-05,  3.187875775884e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  2.152882359149e-05, -5.190223716918e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 { -1.000825312855e-05,  1.670012714084e-05, {   0,   0,   1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -2.319288410743e-05, -3.509478639829e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 {  8.607190799362e-06, -1.809320150812e-05, {   0,   0,   0,   0,   5,   0, -13,   7,   0,   0,   0,   0,   0 } },
 {  1.016412281617e-05,  1.652454960783e-05, {   1,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.301067613360e-05, -1.366456799407e-05, {   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.137243993432e-05,  1.529545736285e-05, {   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.239829567476e-05, -1.425230016579e-05, {   0,   0,   4,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  5.264925058550e-06,  2.137875164570e-05, {   1,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.446238982683e-05,  2.179856815277e-06, {   4,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.633658288175e-05, -2.970524134734e-07, {   2,   2,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.808333782189e-05, -8.543587842314e-06, {   3,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.378165313554e-05,  2.842048941954e-06, {   4,   0,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -2.122331616408e-05, -5.399783563013e-06, {   2,   0,   2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.331459166499e-05, -3.290380242082e-06, {   2,   0,  -1,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.363852862549e-05, -1.296178068956e-05, {   2,   0,   3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.152802129619e-05,  5.053562866697e-06, {   2,   0,  -2,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  5.020486074130e-06,  2.155021493279e-05, {   1,  -1,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.632969679008e-05,  1.023699723238e-05, {   0,   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 {  2.641512198297e-05,  1.275500673140e-07, {   4,   0,  -2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.623743077439e-05, -3.019541386975e-07, {   2,   0,  -1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -1.456182836923e-05,  1.197011866855e-05, {   0,   0,   2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -1.768494780514e-05,  8.832791602534e-06, {   1,  -2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.768494780514e-05,  8.832791602534e-06, {   1,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.859088264354e-06, -1.965786270302e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 {  7.473032339930e-06,  1.904234444356e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -2.191832906997e-05, -4.596645992871e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  2.648169991789e-05,  3.157809697817e-08, {   3,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.758885190685e-06, -1.773331687184e-05, {   2,  -2,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.147256493992e-05, -1.501640345874e-05, {   1,   0,   0,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -8.168615162403e-07, -2.567192250311e-05, {   4,   0,  -3,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.358007183020e-05,  1.290316090208e-05, {   4,   0,  -2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  1.127979130805e-05, -1.519700100671e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  1.629933704703e-05, -1.017423912539e-05, {   0,   0,   1,   0,  -5,   0,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.821977081867e-05,  8.248633299829e-06, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.770020093241e-05, -8.764691124871e-06, {   1,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.065964124822e-05, -5.803654438354e-06, {   1,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.063069517005e-05, -1.583239911607e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   3,   0,   0,   0 } },
 {  1.217400710441e-05, -1.428856280599e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 { -8.529797297956e-06, -1.793013227956e-05, {   4,   0,  -2,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -1.413238755613e-05, -1.232224408518e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -4.832971306106e-06,  2.162069786562e-05, {   3,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.229584955040e-06, -2.121541096520e-05, {   1,  -1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.225166692279e-05,  4.193229180065e-06, {   0,   2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.143301341325e-05,  5.001986838480e-06, {   2,   0,  -2,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.824790431721e-05, -8.186207182209e-06, {   3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.874014472130e-06,  2.255007715244e-05, {   2,   0,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.825752012471e-05,  8.164738752443e-06, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.595965025115e-06,  2.482599076006e-05, {   1,  -1,   0,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0 } },
 {  4.942911005596e-06,  2.147445555631e-05, {   3,   1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.827147134159e-05,  8.133470047500e-06, {   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.585040060045e-06,  1.681730702307e-05, {   0,   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.725678149735e-05,  9.145158527868e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  4.701402678207e-06,  2.169711300827e-05, {   1,  -1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.840345654937e-05, -7.987667364905e-06, {   1,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -2.363785241750e-05, -2.748589666780e-06, {   4,   0,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.049377252040e-06, -1.830867351433e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.184407142853e-05, -1.449724737061e-05, {   0,   0,   1,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.518677363540e-05, -1.152134072101e-06, {   4,   0,  -2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -1.282943643776e-05, -1.350742294804e-05, {   0,   0,   3,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -4.502011966158e-07, -2.587109888111e-05, {   2,   0,   3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.615396066070e-05, -1.016149295575e-05, {   2,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -5.043482019734e-06, -2.126887944380e-05, {   1,  -1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.629768736679e-05, -1.167980651985e-08, {   0,   0,   2,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.610531838403e-05, -1.923747546746e-07, {   2,  -2,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.214852434596e-05,  4.143815725368e-06, {   1,  -3,   1,   0,   0,  33, -39,   0,   0,   0,   0,   0,   0 } },
 {  1.756452635494e-05,  8.706170745556e-06, {   2,   0,  -1,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.488336933748e-06,  2.077887344954e-05, {   1,   1,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.052895488272e-05,  1.572241016025e-05, {   1,   0,   0,   0,   0, -21,  16,  11,   0,   0,   0,   0,   0 } },
 {  2.611288762798e-05,  1.370562469571e-07, {   1,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.646792881030e-05, -9.773218616010e-06, {   1,   0,  -3,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 {  1.092285133076e-05, -1.531589915356e-05, {   1,   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.840298622687e-05,  7.831353518612e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.005527856919e-05,  1.615771278467e-05, {   2,   2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.034672830116e-05,  1.586506277809e-05, {   2,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  1.752798099843e-05, -8.679757860584e-06, {   1,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.286519805549e-05, -3.339254922827e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  2.618727275646e-05,  3.974326448037e-09, {   2,   2,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.940406111427e-05,  6.774509947713e-06, {   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.783846660397e-06, -2.037619652175e-05, {   1,   0,   1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 { -5.005959651173e-06,  2.115161831086e-05, {   5,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.606393670352e-06, -1.754434407097e-05, {   4,   0,  -2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -5.601571800901e-06,  2.054795784711e-05, {   2,   0,  -1,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 {  2.368821572200e-05, -2.458912525465e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,  -3,   0,   0 } },
 {  1.751187301772e-05, -8.628670000249e-06, {   1,   0,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.203652377645e-05,  4.099605819065e-06, {   0,   2,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.689217049434e-06,  2.043353742736e-05, {   2,   2,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  5.689207704345e-06,  2.043302163107e-05, {   2,   2,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.747034200252e-05, -8.648673441121e-06, {   1,   0,   0,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.084297167113e-06,  2.403203682038e-05, {   2,   0,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  5.075579669593e-06,  2.102260786978e-05, {   3,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.021053938955e-06, -2.207047600208e-05, {   1,   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.131952140821e-06,  2.195892468110e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  4.131952140821e-06,  2.195892468110e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  2.607744994074e-05,  1.158750781738e-08, {   2,   2,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.594783576175e-05, -1.400483259708e-07, {   4,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.560371033484e-06,  2.451136647395e-05, {   2,   0,  -1,   0,   0,   6,  -8,   0,   1,   0,   0,   0,   0 } },
 { -1.128424397674e-05, -1.478184512443e-05, {   6,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.678557996110e-07, -2.587121539920e-05, {   2,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.736220364193e-05,  8.671579562552e-06, {   0,   0,   6,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.600576111645e-05,  2.724329493946e-08, {   0,   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.680145582810e-05, -9.230637064489e-06, {   1,   2,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  1.680145582810e-05, -9.230637064489e-06, {   1,  -2,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -4.069108058682e-06, -2.196281940636e-05, {   2,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.038336417223e-05,  1.562027198246e-05, {   1,  -1,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.957123629669e-05, -6.429624223945e-06, {   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.264278446355e-05, -1.335677274291e-05, {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.669098018107e-05,  9.307588739944e-06, {   2,  -2,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.596291229907e-05, -1.000908789529e-05, {   0,   0,   0,   0,   0,   2,  -8,   8,   8,  -6,   0,   0,   0 } },
 { -8.746474628176e-06, -1.722105769941e-05, {   4,   0,  -2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  5.328946294798e-06,  2.063733792704e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  2.286708014627e-05, -3.098899391190e-06, {   0,   0,   1,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.371644794576e-06, -1.759408132787e-05, {   2,   0,   0,   0,   0,   0,  16, -20,   0,   0,   0,   0,   0 } },
 { -2.583449403700e-05,  1.286944989930e-07, {   2,   0,  -4,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.254366432648e-05, -1.341445745123e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   2,   0,   0,   0,   0 } },
 { -9.329054218790e-07, -2.499914112385e-05, {   2,   0,  -2,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.687427114240e-06,  2.022797911958e-05, {   1,   1,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.282612605290e-06, -1.862695771298e-05, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.992949585977e-06, -2.090911362406e-05, {   1,   1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.580440100184e-05,  7.536771504291e-08, {   2,  -2,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.576189532250e-05,  1.161988995427e-07, {   2,   0,  -2,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  2.583641370277e-05,  3.379270268249e-08, {   0,   2,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.140330551075e-05, -4.449803336316e-06, {   1,  -1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.568707252943e-05, -1.627242581808e-07, {   0,   2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.654634740175e-05, -9.301445994840e-06, {   1,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -8.891852306286e-06, -1.694817297273e-05, {   2,   0,  -1,   0,   0,  11, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.333396649652e-05,  1.250479180617e-05, {   2,   0,  -1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 {  1.837903239642e-06, -2.400007399474e-05, {   3,   0,  -3,   0,   0,   0,  11, -16,   2,   0,   0,   0,   0 } },
 { -1.571880943702e-05, -1.011738429468e-05, {   2,   0,  -1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0 } },
 { -1.937835737942e-06,  2.388403206023e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -9.134533746086e-06, -1.668310246319e-05, {   0,   0,   2,   0,   0, -18,  24, -16,   4,   5,   0,   0,   0 } },
 { -1.722572739983e-05,  8.588260626489e-06, {   0,   0,   2,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.087197262600e-05, -4.934358007996e-06, {   1,  -1,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.844022143604e-05, -7.362979380203e-06, {   2,   0,  -1,   0,   0,   0,  14, -21,   0,   0,   0,   0,   0 } },
 { -2.217497292794e-05,  3.617136372937e-06, {   1,   1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.986192198281e-05,  5.929627822645e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.796724793397e-06, -2.399303876475e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  5.292853568860e-06,  2.047664384902e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -2.448365272012e-05, -1.283704113447e-06, {   6,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.220058372541e-06,  2.254247063480e-05, {   0,   0,   0,   0,   0,   0,  12, -21,   0,   0,   0,   0,   0 } },
 { -2.557844649922e-05, -1.785561004311e-07, {   2,   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -4.935161818951e-06, -2.082014992883e-05, {   1,  -1,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.661125151079e-05,  9.143944497747e-06, {   4,   0,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -1.661125151079e-05,  9.143944497747e-06, {   0,   0,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  2.116927766261e-05, -4.564448058488e-06, {   4,  -2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.569371247804e-05, -2.648983731645e-08, {   1,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.277163942454e-05, -1.291556907196e-05, {   0,   0,   4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.634744728959e-06, -1.704570945895e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 { -8.608055140059e-06, -1.706670929014e-05, {   4,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  8.212105682491e-06,  1.745909879150e-05, {   0,   0,   3,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -2.566008731529e-05,  6.660340048583e-10, {   2,   0,   3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.641291098942e-05,  9.246072578418e-06, {   1,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.646781569490e-05,  9.190768779096e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 } },
 {  1.164892634438e-05,  1.400843893560e-05, {   2,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.533196490265e-06, -1.612093419781e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  2.261663487859e-05, -3.030833427834e-06, {   3,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.008785810685e-06, -2.363042254443e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -1.268961198535e-05, -1.294916739911e-05, {   2,   0,   3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.879348934041e-06,  2.275738709221e-05, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.377952654289e-05,  1.185428190091e-05, {   2,   0,  -2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -1.871854640395e-05, -6.908158482084e-06, {   2,   0,  -1,   0,   0,   0,  -7,   8,  -3,   0,   0,   0,   0 } },
 {  1.126565666706e-05, -1.435978223404e-05, {   0,   0,   1,   0,   0, -16,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.402030691625e-06,  2.421439356803e-05, {   2,   0,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.465963379502e-05,  1.093871263670e-05, {   0,   0,   0,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 {  6.178377420319e-06,  1.941683318954e-05, {   2,   0,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -8.581669627081e-06, -1.701255272904e-05, {   2,   0,   0,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 { -6.182548329462e-06,  1.939041433301e-05, {   2,   0,   0,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 { -1.207860348596e-05,  1.348999317148e-05, {   0,   0,   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  2.553129452670e-05, -2.716763214556e-08, {   1,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.541089817918e-05, -1.418826891271e-07, {   0,   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.119078582861e-05, -4.353634253475e-06, {   1,  -1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.338908542624e-06, -1.820461649563e-05, {   2,   0,  -2,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.733992044217e-05, -8.199483579198e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -6.942508133477e-06, -1.859380702299e-05, {   2,   0,  -4,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  9.609701328755e-06,  1.590816191421e-05, {   1,   0,   0,   0,   0,  -7,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.756288861459e-05,  7.954792518231e-06, {   3,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.495993357406e-05, -1.055498230318e-05, {   4,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  3.836743629245e-06, -2.166924649137e-05, {   2,   0,  -2,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -6.240757436603e-06, -1.925611519824e-05, {   0,   0,   0,   0,   0,   1,   9, -20,   0,   0,   0,   0,   0 } },
 {  2.130508432489e-05,  4.190980711439e-06, {   2,   0,   2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.208977741326e-06,  2.027603817520e-05, {   0,   2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.422386760136e-05, -1.125649061015e-05, {   0,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.463105201560e-06,  1.999866353942e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0 } },
 {  1.631794686651e-05,  9.125807079546e-06, {   2,   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -9.125066224335e-07, -2.452727736541e-05, {   2,   0,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.829572083520e-06,  2.260703891404e-05, {   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.225324359641e-05, -1.317696323013e-05, {   0,   1,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.533920201863e-05,  7.097574587633e-08, {   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 { -2.323729431749e-05,  2.172224449213e-06, {   2,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.067251164211e-06,  2.434008126640e-05, {   0,   0,   1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  8.055478336305e-06,  1.734301292808e-05, {   2,   0,   0,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 {  9.511482607138e-06,  1.588175260889e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  1.300023713591e-05,  1.238946118523e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  4.546550868421e-06, -2.083666300714e-05, {   2,   0,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -1.751342452985e-05, -7.868183181116e-06, {   4,   0,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.952140967587e-06, -2.041150600171e-05, {   2,   0,  -1,   0,   0, -10,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.167915132681e-05, -3.680109717229e-06, {   1,   1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.246091335254e-05, -1.289258591416e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  9.254068152446e-06, -1.608356190433e-05, {   0,   0,   2,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  1.268822713892e-05, -1.263988227272e-05, {   2,   0,   0,   0,   0,   0, -12,  19,   0,  -3,   0,   0,   0 } },
 { -2.092790321383e-05, -4.399365435290e-06, {   1,  -1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.847384081218e-06,  2.047754917978e-05, {   1,  -1,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.292779994148e-06,  1.602728010844e-05, {   0,   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 { -6.589183817035e-06, -1.871515431368e-05, {   2,  -2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.290700816179e-06, -2.201181224570e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.473989296706e-05, -1.055725864221e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0 } },
 {  5.626924011768e-06,  1.966062020256e-05, {   2,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.547095120544e-06,  2.373523657782e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 { -2.006613928535e-05, -5.209300506436e-06, {   2,   0,   0,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.160643194223e-06,  2.010877086891e-05, {   0,   2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.516687005838e-05,  9.591281552936e-08, {   2,   0,  -2,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.500418990594e-05, -1.025571853962e-05, {   4,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.626572687427e-05,  8.990581036873e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0 } },
 {  2.216620937692e-05, -3.088496047598e-06, {   2,   0,  -1,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.694436126169e-05,  8.303408948318e-06, {   2,   0,   1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.538437493419e-05,  9.853454811368e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   6,  -5,   0,   0,   0 } },
 { -2.030357813235e-05,  4.929066025094e-06, {   1,  -1,   1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.917055211228e-05, -6.044545353645e-06, {   1,   1,  -3,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.916372456080e-05, -6.044177013216e-06, {   3,  -1,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.220209203436e-05,  3.002868987196e-06, {   0,   0,   2,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -2.501546758244e-05, -1.882052609334e-07, {   6,  -2,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.023399984961e-05,  4.954751264963e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,  -5,   4,   0,   0,   0 } },
 { -1.942548023211e-05,  5.753896247670e-06, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 { -7.424199165190e-06, -1.775210773215e-05, {   1,   1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.084412061290e-05,  4.331136086654e-06, {   1,  -1,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.781861596075e-06, -1.939061892881e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.800515635077e-05, -7.167322632189e-06, {   2,   0,  -1,   0,   0,  -2,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.061725264432e-05,  1.454810833701e-05, {   1,  -1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.514460498528e-05,  1.147206935919e-08, {   2,   0,  -2,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.827203207233e-06,  1.932862262130e-05, {   1,   0,   1,   0,   0, -33,  28,   0,   0,   0,   0,   0,   0 } },
 { -2.172900095930e-05, -3.419071849483e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  4.799756622995e-06,  2.034638684905e-05, {   1,   1,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.139521254143e-05,  3.742322661788e-06, {   1,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.372737987588e-05, -1.375365712122e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -4.779600082132e-06,  2.031113982095e-05, {   1,  -1,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.489003124770e-05,  1.856893806092e-07, {   0,   0,   1,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 { -2.504705304357e-05, -1.070802741233e-08, {   0,   0,   2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.579922398900e-07, -2.418602323446e-05, {   1,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.308762055779e-05, -1.944643282064e-06, {   0,   0,   1,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.308207880854e-05, -1.934937913497e-06, {   0,   0,   1,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.888413934921e-06,  1.712574458922e-05, {   0,   0,   1,   0,   0,   5,  -2, -12,   3,   0,   0,   0,   0 } },
 {  7.888413934921e-06,  1.712574458922e-05, {   0,   0,   1,   0,   0,  -5,   2,  12,  -3,   0,   0,   0,   0 } },
 {  1.673839387433e-05, -8.272725736555e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 { -1.450960277884e-05, -1.050001899592e-05, {   2,   0,   0,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 {  1.131570600452e-05,  1.369356659623e-05, {   0,   1,   0,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.228574432092e-05, -1.271445662781e-05, {   0,   0,   1,   0,   0,   0,  -3,   2,   2,   0,   0,   0,   0 } },
 { -1.200651452296e-05,  1.298762062790e-05, {   1,  -1,  -1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.050619760201e-05, -1.448371258987e-05, {   4,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.376351522056e-06,  1.960939646814e-05, {   1,  -1,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.837845645672e-06,  2.214679225868e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 {  1.926335256614e-05,  5.720813112531e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -4.686372074681e-06, -2.028747829396e-05, {   0,   1,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -8.317996823336e-06, -1.665294638630e-05, {   4,   0,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.141860213159e-05, -1.355107055699e-05, {   1,  -1,   0,   0,   0,  17, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.490012470820e-05,  6.871912463306e-08, {   0,   2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.161165640822e-05, -1.335586311560e-05, {   0,   0,   1,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 { -6.049444156980e-06,  1.889986136116e-05, {   2,   0,  -2,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  6.367731502745e-06,  1.857892829892e-05, {   4,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.912899419681e-06, -2.102735995259e-05, {   3,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.215343180626e-05, -1.278581080492e-05, {   0,   0,   1,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -1.425082289528e-05, -1.067113210278e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -2.491202089112e-05,  8.668036138921e-10, {   2,  -2,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.229045185200e-07, -2.458857943564e-05, {   2,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  5.730285340558e-06, -1.916308479024e-05, {   0,   0,   0,   0,   0,   2,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.373009811112e-05,  1.115660505090e-05, {   0,   0,   1,   0,   0,   0,  20, -38,   0,   6,   0,   0,   0 } },
 {  1.373009811112e-05,  1.115660505090e-05, {   0,   0,   1,   0,   0,   0, -20,  38,   0,  -6,   0,   0,   0 } },
 {  5.563642542168e-06, -1.931647321391e-05, {   4,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  8.988740199832e-07, -2.397768816120e-05, {   2,  -2,  -2,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.745242100937e-05,  7.423816434481e-06, {   2,   0,  -2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.044941341897e-06, -2.282742613707e-05, {   2,   0,   0,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 {  2.044941341897e-06, -2.282742613707e-05, {   2,   0,  -2,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 { -4.744736749005e-06, -2.012427431007e-05, {   1,  -1,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.665698376380e-05,  8.205648528504e-06, {   2,   0,   2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.004575568497e-05, -4.816645751460e-06, {   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0 } },
 {  9.672115758401e-06, -1.518688639260e-05, {   2,   0,  -2,   0,   0,   3,  -1,  -4,   0,   0,   0,   0,   0 } },
 { -2.458694655922e-05, -2.713510939405e-07, {   2,   0,  -1,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.732290964457e-05,  7.524299267018e-06, {   1,   0,   0,   0,   0,  -2, -32,  44,  -8,   6,   0,   0,   0 } },
 { -4.668464258360e-06, -2.017767091042e-05, {   1,  -1,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.089248589701e-05,  1.395045497186e-05, {   2,   0,  -1,   0,   0,   0,  12, -20,   0,   0,   0,   0,   0 } },
 { -4.309258839802e-07,  2.440692452678e-05, {   1,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.289547272428e-05,  1.934400666604e-06, {   1,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.005378495591e-05,  4.773616375554e-06, {   1,   1,  -1,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -9.834847643030e-07, -2.384110332394e-05, {   0,   0,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.444075818346e-05, -3.820805714599e-07, {   0,   2,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.706226470687e-05,  7.745537619250e-06, {   4,   0,  -2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.949052154193e-05, -5.316908227599e-06, {   1,   1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.480591809570e-05, -3.408083543364e-10, {   0,   2,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.592754442906e-06, -1.921118842285e-05, {   2,  -1,   0,   0,   0, -23,  25,   2,   0,   0,   0,   0,   0 } },
 { -8.321386511299e-07,  2.396308440792e-05, {   2,   0,  -1,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0 } },
 { -5.773104061081e-06, -1.901816676834e-05, {   2,   0,  -1,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.004369596254e-05, -1.474639107438e-05, {   3,   0,   0,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 { -1.004369596254e-05, -1.474639107438e-05, {   1,   0,   0,   0,   0,   0,  31, -35,   0,   0,   0,   0,   0 } },
 {  1.907556053037e-05,  5.714406388656e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  9.903687354287e-06,  1.488233651345e-05, {   2,   0,   0,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -9.066574445274e-06,  1.571600941879e-05, {   2,   0,  -1,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0 } },
 {  2.730686776621e-07, -2.450814123997e-05, {   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.291818663064e-05,  1.185351901898e-05, {   0,   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  9.424197742476e-06, -1.534745021848e-05, {   4,   0,  -3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.439181899849e-05, -1.037967947906e-05, {   2,   0,   0,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 { -1.674288857912e-05,  8.024219149210e-06, {   2,   0,   1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.380330438073e-06,  1.938318712950e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 { -1.660022061139e-05, -8.161909434856e-06, {   1,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.603603592848e-06,  2.115843250416e-05, {   1,   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  5.974131938689e-06,  1.878627002918e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -1.774829949851e-06,  2.298442757207e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 { -1.061694664597e-05,  1.412866833532e-05, {   2,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.423490766290e-05, -4.957457569299e-07, {   1,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.428085886121e-06,  1.929178340721e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   3,   0,   0,   0 } },
 { -6.364877459400e-06,  1.834443201401e-05, {   1,   0,   1,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 { -6.364877459400e-06,  1.834443201401e-05, {   1,   0,  -1,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 { -2.053864590166e-05, -4.159989374754e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   2,   0,   0,   0 } },
 {  1.750699390374e-05, -7.172647586753e-06, {   1,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.413379020095e-07,  2.443139004768e-05, {   3,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.591252506483e-05,  8.759318701372e-06, {   4,   0,   0,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -1.117593549912e-05,  1.349264206342e-05, {   1,   0,   0,   0,   0, -14,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.856132589324e-05, -6.081981076629e-06, {   4,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.539700952307e-06,  2.108739398045e-05, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 { -5.732288191323e-06, -1.888777850165e-05, {   4,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.427901047072e-05,  1.033975842910e-05, {   0,   0,   1,   0,  -4,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.496417085089e-06,  2.012034479973e-05, {   4,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.459583178136e-05, -1.001899842097e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 { -1.587107397755e-05,  8.736501249009e-06, {   0,   0,   2,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -5.497609385631e-06, -1.910414771168e-05, {   4,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.126242658126e-06,  1.946093796502e-05, {   1,  -1,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.342855956609e-06, -1.522651137228e-05, {   4,   0,  -3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.173353461761e-06,  1.738954751131e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  5.641891040777e-06, -1.892048277523e-05, {   4,  -2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.385479627130e-05, -7.000404412182e-07, {   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.706116900317e-06, -2.083172360608e-05, {   0,   0,   3,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.551909325411e-06, -1.898542472336e-05, {   0,   0,   0,   0,   0,   2,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -5.633637098640e-06, -1.890117639091e-05, {   4,  -2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.098305160881e-05, -3.551505857447e-06, {   2,   0,  -2,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.196827040627e-05,  1.256490227379e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 { -6.683776541720e-06, -1.784728335023e-05, {   2,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -2.166648634770e-05, -2.863767491436e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -2.105129544984e-05,  3.469844252638e-06, {   2,   0,  -2,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -2.389762765315e-05,  6.178947784234e-07, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  2.450825666123e-05,  6.308888229687e-09, {   1,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.542474629305e-06, -1.596400311400e-05, {   1,  -1,   2,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 {  8.542474629305e-06, -1.596400311400e-05, {   1,  -1,   0,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 { -1.102106861032e-06,  2.339947154931e-05, {   2,   0,   1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.677431317270e-05, -7.721032729351e-06, {   2,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.105126828365e-05,  3.432565432777e-06, {   2,   0,  -2,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  1.867221193175e-05, -5.805845892858e-06, {   0,   0,   1,   0,   0,   8, -21,  16,  -4,  -5,   0,   0,   0 } },
 { -1.867221193175e-05, -5.805845892858e-06, {   0,   0,   1,   0,   0,  -8,  21, -16,   4,   5,   0,   0,   0 } },
 {  1.228812308166e-05, -1.218701749297e-05, {   2,   0,  -1,   0,   0,   0, -12,  19,   0,  -3,   0,   0,   0 } },
 { -2.929646392991e-06, -2.154230585851e-05, {   2,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.637086701106e-05, -8.095310291417e-06, {   1,   0,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.431615991894e-05, -1.374407624586e-07, {   2,  -2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.353003842572e-05, -9.111964333552e-07, {   4,   2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -9.808593077215e-07,  2.345974172292e-05, {   2,   0,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  3.830524655282e-06, -2.060858875758e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.268681018483e-05, -1.747831828310e-06, {   0,   0,   1,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.934253617593e-05, -5.085891690030e-06, {   1,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.934253617593e-05, -5.085891690030e-06, {   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.488331514202e-06,  1.493840500347e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -9.069391087317e-07, -2.351780258828e-05, {   0,   0,   3,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.075543945055e-05, -1.366817201898e-05, {   1,  -1,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.628746933465e-05,  8.135117107454e-06, {   0,   0,   4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.416387450043e-05,  2.583687879499e-07, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 {  8.633418445245e-06, -1.578016408580e-05, {   0,   0,   0,   0,   0,  18, -24,  16,  -4,  -5,   0,   0,   0 } },
 {  1.869361237101e-05, -5.717792682083e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,  10,   0,   0,   0 } },
 { -9.623660701976e-06,  1.477879815449e-05, {   2,  -2,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -4.577697612417e-06,  1.982133376507e-05, {   1,  -1,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.726575164585e-06,  1.866692474791e-05, {   2,   0,  -1,   0,  -3,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.690361632936e-06, -1.569619002169e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 { -2.328767224813e-05,  1.096661322565e-06, {   4,  -2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.335652863815e-05,  1.102061457722e-05, {   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -4.659086379216e-06, -1.971317079881e-05, {   0,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.272258469008e-05,  1.647989359695e-06, {   2,  -2,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.032495444695e-05, -4.039278713308e-06, {   1,   1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.849069943975e-06,  2.249499746643e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  2.192731552569e-05, -2.416639021631e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  1.626300555802e-05, -8.079819736143e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  1.703755623968e-05, -7.302025537716e-06, {   2,   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  9.710102428903e-06, -1.462860055153e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  4.926802333667e-06, -1.938659368203e-05, {   2,   2,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.332218820357e-05, -9.843528773159e-07, {   0,   2,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.341157458951e-06,  2.196202507141e-05, {   2,   0,   0,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  2.005289742063e-05,  4.250105397635e-06, {   1,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.044860617075e-05, -1.385222112469e-05, {   0,   0,   1,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 { -1.705292037833e-05,  7.246949795706e-06, {   2,   0,  -4,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.516172569993e-05, -9.128320160491e-06, {   0,   0,   1,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 {  2.344728353108e-05, -8.357708475078e-07, {   2,   2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.854909301568e-05,  5.733903149109e-06, {   1,   1,  -1,   0,   0, -18,  13,   7,   0,   0,   0,   0,   0 } },
 { -2.143154987217e-05,  2.851056192343e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.435592379867e-06, -2.284672959703e-05, {   2,   0,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  2.139156913261e-05,  2.887961660471e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -7.445547501130e-06,  1.683099314752e-05, {   1,  -1,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -7.445547501130e-06,  1.683099314752e-05, {   1,  -1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  4.645870116483e-06,  1.962386211222e-05, {   0,   0,   0,   0,   0,   0,   6, -11,   3,   0,   0,   0,   0 } },
 {  2.090124323379e-05,  3.364312909803e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 { -5.759495420446e-06, -1.849447677997e-05, {   0,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0 } },
 {  1.193058326538e-05, -1.231594565813e-05, {   2,   0,   0,   0,   0,  -9,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.624429971766e-05, -8.000360483537e-06, {   1,   0,   0,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.473738302920e-06,  2.276918770513e-05, {   2,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 { -2.138000413259e-05,  2.817531159706e-06, {   2,   0,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.153485332600e-05,  1.265237128261e-05, {   2,   0,  -2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  2.382758026824e-05,  3.544805716887e-07, {   0,   0,   1,   0,   0,  -2,   4,   0,  -2,   0,   0,   0,   0 } },
 {  9.016831610211e-06,  1.516142146019e-05, {   2,   2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.030946967421e-07,  2.356463210336e-05, {   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.653560579496e-05, -7.629906460608e-06, {   1,   1,   0,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.751528654248e-05,  6.643795504740e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   2,   0,   0,   0 } },
 {  1.950501715511e-05, -4.631472513423e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.202784324512e-05, -2.104621137291e-06, {   1,   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.354406349244e-06, -2.077436708145e-05, {   0,   0,   0,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.143942500937e-06, -2.296689098432e-05, {   2,   0,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0 } },
 {  1.943930214750e-05,  4.667842331976e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 {  2.386139804959e-05, -2.427527902743e-07, {   0,   0,   2,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  2.406727242228e-05,  3.521833742106e-08, {   1,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.856530126459e-06, -1.524017982888e-05, {   1,  -2,   0,   0,   0,   0,  40, -48,   0,   0,   0,   0,   0 } },
 {  9.993118102361e-06,  1.409393038770e-05, {   0,   0,   1,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -1.945095704940e-05,  4.634808084470e-06, {   1,  -1,  -1,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.969612261711e-05, -4.387367427343e-06, {   6,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.744533632241e-05,  6.632366762708e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  8.229004082497e-06,  1.583354060980e-05, {   1,  -1,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  1.552081647558e-05, -8.527060112910e-06, {   2,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.552081647558e-05, -8.527060112910e-06, {   0,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -3.189114466447e-06,  2.085848810483e-05, {   4,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -1.791721557253e-05, -6.128523112665e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -2.246797201243e-06,  2.179374064360e-05, {   2,   0,   1,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  2.246797201243e-06,  2.179374064360e-05, {   2,   0,  -1,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 {  3.683072783793e-06,  2.035064142719e-05, {   4,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -1.978435990904e-05, -4.246713951428e-06, {   3,  -1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.377057302906e-05, -2.541247620064e-07, {   0,   0,   2,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  1.422339537278e-05,  9.798511578811e-06, {   2,   0,  -1,   0,   0,   0,  17, -22,   0,   0,   0,   0,   0 } },
 {  3.412668185024e-06,  2.060720883864e-05, {   2,  -2,   2,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -7.249105594330e-06,  1.676587042152e-05, {   2,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -2.113754155713e-05, -2.866307651696e-06, {   2,   0,   0,   0,   0, -46,  50,   0,   0,   0,   0,   0,   0 } },
 {  1.766652107520e-05, -6.329219897783e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  9.718653973658e-06, -1.426915676727e-05, {   3,   0,  -1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 { -9.718653973658e-06, -1.426915676727e-05, {   1,   0,  -1,   0,   0,   0,  31, -35,   0,   0,   0,   0,   0 } },
 { -4.563043499586e-06,  1.941129125202e-05, {   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.303415169223e-07,  2.384203301805e-05, {   1,   2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.205119613248e-06, -2.275903410219e-05, {   4,   0,  -2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  2.016730375791e-05,  3.771926836006e-06, {   1,   1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.562236635112e-06,  1.836359290289e-05, {   0,   0,   1,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.197709408320e-06, -2.071500168647e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 {  3.579218872128e-06,  2.032793184926e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   5,   0,   0,   0 } },
 {  3.191763805961e-06, -2.070106494264e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 {  1.172808621821e-06, -2.271475372198e-05, {   4,   0,  -2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.082077813308e-05, -1.306009853785e-05, {   2,   0,  -4,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.973825773304e-05,  4.136768398384e-06, {   2,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.596193959845e-05, -7.894719554262e-06, {   1,   0,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.936074679173e-06,  1.391955974236e-05, {   4,   0,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.711788923889e-05,  6.727036363370e-06, {   2,   0,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 { -2.383426243423e-05, -7.499065060036e-09, {   2,   2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.079010352532e-05,  3.050250079485e-06, {   1,  -1,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.063430081325e-07, -2.352004309722e-05, {   2,  -2,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.103316827216e-06, -1.872134053336e-05, {   1,  -1,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.549186833586e-05, -8.323490668083e-06, {   2,   0,  -1,   0,   0, -18,  21,  -7,   8,  -6,   0,   0,   0 } },
 { -1.146519236314e-05,  1.232859794224e-05, {   2,   0,  -2,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  1.427451751584e-05, -9.517384044465e-06, {   2,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.525998423587e-05, -8.524607206054e-06, {   1,  -2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.359659731219e-05, -1.783932472960e-07, {   6,  -2,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  9.032095095300e-06, -1.473995382750e-05, {   2,   0,  -1,   0,   0,   5,  -3,  -7,   8,  -6,   0,   0,   0 } },
 { -1.731553439077e-05,  6.455820738291e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -9.029771730150e-06, -1.474094178271e-05, {   2,   0,  -1,   0,   0,   1,  -3,   7,  -8,   6,   0,   0,   0 } },
 { -8.634439973595e-06,  1.512561006964e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 {  1.324982558278e-05, -1.048793786184e-05, {   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.582841145250e-05, -7.905550316684e-06, {   4,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.582841145250e-05, -7.905550316684e-06, {   4,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.372065922008e-05,  3.694664981505e-09, {   2,   0,   2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.782066121321e-05, -5.901158812605e-06, {   2,   0,  -1,   0,   0,   5,  -9,   5,   0,   0,   0,   0,   0 } },
 {  1.956397255325e-05,  4.153429659434e-06, {   2,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.435777917318e-05,  9.347751223011e-06, {   2,   0,  -1,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 { -1.057315453001e-05,  1.311049132096e-05, {   1,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -5.725656693428e-06, -1.792954496336e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  1.896620733419e-05, -4.680463078921e-06, {   1,   1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.974142015029e-05,  3.904518016317e-06, {   1,   1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.440867459128e-06,  1.819566974631e-05, {   2,   0,  -1,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 { -2.282398116717e-05,  8.084582823234e-07, {   6,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.595353386478e-06, -2.003055614835e-05, {   4,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  3.477394537264e-06,  2.014743012963e-05, {   3,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.477394537264e-06,  2.014743012963e-05, {   1,   0,   1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.983174793877e-05,  3.779090857715e-06, {   1,   1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.373559793940e-05, -9.859076339909e-06, {   1,  -1,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -4.998636412423e-06, -1.859253368126e-05, {   1,   1,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.807263777536e-05, -5.512942466950e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.636899153612e-05,  7.213735464932e-06, {   0,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -9.382711921435e-07,  2.264354635627e-05, {   2,  -4,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.873648665331e-06, -1.669371249679e-05, {   0,   0,   2,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.370952256836e-05, -9.850003334397e-06, {   2,   0,   0,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 { -1.122682957586e-05, -1.233045009831e-05, {   0,   0,   0,   0,   0,   6,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.347716488283e-05, -6.416509020349e-08, {   2,   0,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.100720320585e-05,  1.252158509173e-05, {   2,   0,   0,   0,   0, -14,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.331032753148e-05, -1.021390324511e-05, {   2,  -2,   0,   0,   0,   0,   2,   0,   1,   0,   2,   0,   0 } },
 {  1.717164867609e-05,  6.330933339677e-06, {   2,   0,   0,   0,   0,   0,  -6,   0,   6,   0,   0,   0,   0 } },
 {  7.465451706583e-06,  1.603585309606e-05, {   0,   0,   2,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  1.685529233060e-05,  6.637212633940e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -2.311217574256e-05, -3.801432183329e-07, {   2,   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -4.448174752537e-06, -1.903514824164e-05, {   2,   0,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.109872082178e-05,  2.383654716668e-06, {   0,   0,   0,   0,   0,   0,  12, -21,   2,   0,   0,   0,   0 } },
 {  7.605726072280e-06, -1.587616867860e-05, {   1,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.758940478157e-05,  5.875036365384e-06, {   0,   0,   0,   0,   0,   2,  -7,   8,  -3,   0,   0,   0,   0 } },
 { -2.311984867582e-05, -3.420708791517e-07, {   1,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.509795837582e-06, -1.995181540662e-05, {   1,   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.288617495613e-05, -1.057508738223e-05, {   1,   1,  -2,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.674102398141e-06, -1.977700039068e-05, {   1,   1,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.294441912175e-05, -1.050264270885e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -7,   0,   0,   0,   0 } },
 { -2.215609876810e-05,  1.277631176843e-06, {   1,  -1,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  4.407034413103e-06, -1.901921626627e-05, {   0,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.339726623975e-06,  2.207681060265e-05, {   2,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.112236706832e-05, -2.293196699445e-06, {   1,  -1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.299898483668e-06, -1.411450166616e-05, {   0,   0,   2,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  2.078294436401e-05,  2.626111184928e-06, {   0,   0,   2,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.329284120750e-05, -1.161569183841e-07, {   2,  -2,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.848660101067e-06,  1.355597639215e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   3,   0,   0,   0 } },
 {  1.547348803929e-05, -7.919463226824e-06, {   2,   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.872406987982e-05, -4.665151961436e-06, {   2,   0,   0,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.328179074225e-05, -1.055312327408e-07, {   2,   0,   2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.912122703922e-05, -4.262471019478e-06, {   2,   0,   0,   0,   0, -18,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.559988670483e-05, -7.783556844549e-06, {   0,   0,   0,   0,   0,  22, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.306470476384e-05, -3.109114995685e-07, {   1,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.144304448035e-05,  1.191850663997e-05, {   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.150783579667e-05,  1.185224919022e-05, {   0,   0,   1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.125460641534e-05,  1.209772544028e-05, {   0,   0,   0,   0,   0,   0,   7, -13,   3,   0,   0,   0,   0 } },
 { -2.377368322221e-06, -2.097213631175e-05, {   2,  -2,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.068606421796e-05, -2.653184720190e-06, {   1,  -1,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  1.797268444692e-06,  2.153761437972e-05, {   0,   0,   1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.664061771778e-05,  6.687788432038e-06, {   2,   0,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  4.544107071383e-06,  1.877930941567e-05, {   1,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.074159363228e-05,  1.257879483289e-05, {   4,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  4.320649807759e-07,  2.286570583724e-05, {   3,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.315437468109e-05, -1.421120462001e-07, {   0,   2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.297178679787e-05,  3.171295932372e-07, {   1,   2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.238476018849e-07, -2.295942395111e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,  -1,   0,   0,   0 } },
 {  2.146303621760e-06, -2.113302016951e-05, {   0,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.465840634084e-05,  8.612930717522e-06, {   0,   0,   0,   0,   0,   0,   2,  -3,   2,   0,   0,   0,   0 } },
 { -1.444907425170e-05,  8.820093718679e-06, {   0,   0,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  2.212764213485e-05, -1.132231035213e-06, {   0,   2,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.176900836081e-06, -1.907912098812e-05, {   4,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  1.178212309717e-05, -1.146941248249e-05, {   2,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  4.507935192733e-06, -1.874217798175e-05, {   0,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.212068961940e-07, -2.232376144454e-05, {   2,  -2,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  9.292416284449e-06, -1.395218375351e-05, {   1,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.141147426513e-05, -1.813407306924e-06, {   2,   0,  -4,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  2.311459671920e-05,  1.002073405209e-07, {   1,   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -5.151115622636e-06, -1.806117587619e-05, {   2,   0,  -1,   0,   0,  -8,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.830462046130e-05, -4.891298676498e-06, {   1,   1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.810875026905e-05, -5.083766383837e-06, {   0,   0,   0,   0,   0,   2,   1,  -8,   3,   0,   0,   0,   0 } },
 { -2.154298549492e-05,  1.641103535111e-06, {   0,   0,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -2.310251244805e-05,  5.021887553167e-08, {   4,   0,  -2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.042553922076e-05, -2.716251928451e-06, {   2,   0,   0,   0,   0,   3,  -9,   4,   0,   0,   0,   0,   0 } },
 { -7.793135310955e-06, -1.534862195247e-05, {   2,  -2,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -1.651509317219e-05, -6.622432552314e-06, {   1,   1,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -1.106263489326e-05,  1.206780690058e-05, {   4,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.504208474709e-05,  8.086486417605e-06, {   1,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.075681190656e-05, -1.236838307242e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,  -1,   0,   0,   0 } },
 { -1.430131056041e-05,  8.820447709174e-06, {   2,   0,  -1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0 } },
 {  1.476242488702e-05, -8.355955227447e-06, {   4,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.202000355567e-05, -1.097803870052e-06, {   2,   0,  -1,   0,   0,  10,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.822941564730e-06, -1.829362844350e-05, {   2,   0,  -1,   0,   0,  -9,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.953718466469e-06,  1.615854538987e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   2,  -2,   0,   0 } },
 {  8.145486787902e-07, -2.229621862964e-05, {   2,   0,  -2,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.060915979398e-05,  1.248963586788e-05, {   4,   0,  -3,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  3.976813342453e-06, -1.911495145063e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  4.834496350435e-06, -1.825223496375e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  4.562555136746e-06, -1.852380663285e-05, {   2,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.830468007537e-05,  4.776779725556e-06, {   1,  -1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.308737531145e-06, -2.176524062793e-05, {   2,   0,  -1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.265263020108e-05,  4.052758361730e-07, {   2,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 { -1.965310676011e-05,  3.388052921396e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 { -7.048992060451e-06,  1.598397786384e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -8.130696257007e-06,  1.490086425180e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.273848878470e-05, -2.835451548523e-07, {   2,   2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -7.565371418626e-06,  1.545523214389e-05, {   0,   0,   2,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  1.491405574543e-05,  8.096498903927e-06, {   4,   0,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.954898309742e-05,  3.448971492954e-06, {   2,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.353503477081e-06, -1.863841981238e-05, {   3,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  8.864421329787e-06,  1.412524314454e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  4.395442125530e-06, -1.857764020046e-05, {   2,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.609837327903e-06,  1.435203995019e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.989919738820e-05,  3.059354493224e-06, {   1,  -1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.960158449184e-07, -2.275616226894e-05, {   4,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.426754553019e-05, -8.684246075636e-06, {   2,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.026499008129e-05,  1.268156863932e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 { -2.277521080841e-06, -2.066577918427e-05, {   0,   0,   2,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -5.136979437429e-06,  1.779808526096e-05, {   1,   1,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.343233155620e-05, -9.499172906249e-06, {   1,   1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  7.053181958086e-06,  1.586878293155e-05, {   1,  -1,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.285509495613e-05, -6.145468854520e-08, {   2,   0,   1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -8.731051258784e-08, -2.281969164622e-05, {   2,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 {  2.079839792255e-05,  2.105754924897e-06, {   2,   0,  -1,   0,   0,   0,  15, -21,   0,   0,   0,   0,   0 } },
 { -2.122242301565e-05,  1.675089750829e-06, {   2,   2,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.174501838370e-05, -1.114609417716e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 {  1.046432567929e-05, -1.242608715555e-05, {   4,   0,  -1,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -1.046432567929e-05, -1.242608715555e-05, {   0,   0,   1,   0,   0,   0,  33, -40,  -3,   0,   0,   0,   0 } },
 { -1.722117369995e-05, -5.659337181143e-06, {   2,   2,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  2.226742996110e-05,  6.098408401695e-07, {   3,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.212964008385e-06,  2.065552135351e-05, {   2,   0,  -2,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 { -2.282976338423e-05,  3.692863165490e-08, {   2,   0,  -3,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.483835229666e-06, -1.838264312023e-05, {   3,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.652102005760e-06,  1.318412522901e-05, {   2,   2,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.367460956185e-06, -1.846640003489e-05, {   3,   1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.268781291373e-05, -1.416143659020e-07, {   0,   2,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.446314253984e-06, -1.837509668085e-05, {   3,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.525890041716e-05,  7.544882278224e-06, {   0,   0,   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.200511369657e-05, -7.947989817592e-07, {   2,   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 {  9.971499856826e-06, -1.282650378253e-05, {   2,   0,  -1,   0,   0,   0,  14, -22,   0,   0,   0,   0,   0 } },
 { -1.013178929704e-05,  1.266244410437e-05, {   0,   0,   0,   0,   0,   0,   8, -16,   0,   0,   0,   0,   0 } },
 {  1.177802808299e-05,  1.100884182026e-05, {   4,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  2.193991963970e-05,  8.415758382025e-07, {   0,   0,   1,   0,   0, -19,  16,   3,   0,   0,   0,   0,   0 } },
 {  1.411266086806e-05, -8.668442916165e-06, {   1,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -9.632086436894e-06,  1.314742062641e-05, {   1,  -1,  -1,   0,   0,  23, -22,  -2,   0,   0,   0,   0,   0 } },
 { -1.859725819533e-05, -4.175176677033e-06, {   1,   1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.353912511671e-06,  1.841465006335e-05, {   1,  -1,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.391596885838e-06,  1.537683709362e-05, {   0,   0,   1,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 {  1.546987471854e-05, -7.274647929527e-06, {   1,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.072305765827e-05,  2.018583437866e-06, {   0,   0,   2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.998522831204e-05, -2.755683536204e-06, {   2,  -2,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.871201370489e-05,  4.027058848774e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -6,   0,   0,   0 } },
 { -1.103865586943e-05,  1.169484936967e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   4,   0,   0,   0 } },
 {  9.788699829066e-06,  1.294263167012e-05, {   2,  -2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.654965092709e-05, -6.180743400853e-06, {   2,   0,   2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  8.471406202621e-06, -1.423824023398e-05, {   2,   0,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -6.507676543584e-06,  1.619481111097e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 {  2.265762473780e-05,  3.050169989487e-08, {   2,   0,  -2,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.029117810358e-05,  2.386384463756e-06, {   2,   0,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.436397587459e-06, -1.523179014018e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 { -1.966169334460e-05, -3.006013799447e-06, {   0,   0,   1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -1.201428771115e-05,  1.064746955418e-05, {   0,   0,   2,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.906997466393e-05, -3.584345864981e-06, {   1,  -1,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.254117112260e-05,  1.011218583519e-05, {   1,  -2,   0,   0,   0,  13,  -4,  -5,   0,   0,   0,   0,   0 } },
 { -2.261211251890e-05, -3.432664676226e-08, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.459639877552e-05, -8.047000670261e-06, {   2,   0,  -2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -8.025298257236e-06, -1.458246848225e-05, {   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.114757120203e-05, -1.145017125083e-05, {   1,   0,   0,   0,   0,   0, -36,  45,  -3,   0,   0,   0,   0 } },
 { -1.303902049165e-05, -9.552190872540e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  1.455383838919e-05,  8.019497130110e-06, {   0,   2,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  2.256269356887e-05,  1.041616166924e-08, {   4,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.967481203112e-05,  2.892747681968e-06, {   2,   0,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.170500173146e-05,  8.598780339013e-07, {   2,   0,  -1,   0,   0, -16,  13,   3,   0,   0,   0,   0,   0 } },
 {  1.970416775586e-05,  2.856481234310e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.924676636937e-05, -3.304846265766e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   5,   0,   0,   0 } },
 {  6.616741783341e-06, -1.592566958021e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  1.742561686564e-05, -5.115907664185e-06, {   1,  -1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.326425381823e-06, -1.721174743869e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  7.044849924294e-06, -1.548875509640e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  2.250112257898e-05,  2.345396425754e-08, {   3,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.993335657607e-06, -1.752261543832e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  3.280704264543e-06, -1.923246384745e-05, {   0,   0,   1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.343747878040e-06, -1.316910390185e-05, {   3,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.216522870210e-06,  1.829207011921e-05, {   1,  -1,   1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.015774644878e-05, -1.234949255066e-05, {   3,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.000219221213e-05,  1.250478828371e-05, {   2,   0,   2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.036799318911e-06,  1.446739445254e-05, {   1,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -3.966457231659e-06, -1.852569144559e-05, {   4,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.055784045086e-06, -1.343340403953e-05, {   2,   0,   3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.745950362556e-06, -1.373687999126e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 { -1.318654564669e-05,  9.291532549945e-06, {   1,   1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  5.698174945178e-06,  1.676748950782e-05, {   0,   0,   1,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.169720434373e-05,  7.545656851285e-07, {   2,   0,   0,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.235151913435e-05, -1.009345281291e-05, {   2,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.673456205003e-06, -1.375954515143e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -6.417430034412e-06, -1.599522614446e-05, {   0,   0,   1,   0,   0,   0,  -3,   6,  -3,   0,   0,   0,   0 } },
 { -2.233128571515e-05, -7.204617869785e-08, {   2,   2,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  2.237129462264e-05,  1.706013479298e-08, {   6,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.713007901850e-07,  2.221673025796e-05, {   2,   0,  -2,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.994411386110e-05, -2.443059535996e-06, {   0,   0,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 {  2.209097009952e-05, -2.685915234052e-07, {   2,  -2,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.234652638949e-05, -1.079598361075e-08, {   0,   0,   2,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.001908242612e-06, -1.334833715059e-05, {   2,   0,   3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.268284222744e-05, -9.659793341222e-06, {   4,   0,  -3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.369046962491e-05, -8.649854778709e-06, {   2,   0,  -2,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 { -1.245189417679e-05,  9.884530658526e-06, {   2,  -2,   2,   0,   0, -55,  65,   0,   0,   0,   0,   0,   0 } },
 { -2.112175521752e-05, -1.194690183622e-06, {   0,   0,   2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.489371159905e-06,  1.479572900046e-05, {   2,   0,  -1,   0,   0,   0,  17, -21,   0,   0,   0,   0,   0 } },
 {  1.616568827868e-05,  6.100129717152e-06, {   0,   0,   1,   0,   0, -28,  32,   0,   0,   0,   0,   0,   0 } },
 { -1.796484226052e-05, -4.294439248864e-06, {   2,   0,   0,   0,   0, -16,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.185322863075e-05,  3.842655230464e-07, {   3,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.042350674466e-05, -1.180154938600e-05, {   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.222248377775e-05, -9.809486738144e-11, {   0,   0,   3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.675359288568e-06,  1.654579877277e-05, {   0,   0,   0,   0,   5, -22,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.408381418328e-06,  1.481101226556e-05, {   0,   0,   1,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -1.046593763196e-05, -1.173786017067e-05, {   1,   0,   1,   0,   0,   0, -33,  41,  -2,   0,   0,   0,   0 } },
 { -5.600848966351e-06,  1.659048724367e-05, {   2,   0,   2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.710561905956e-05, -5.084175913926e-06, {   1,   1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.431643010727e-05,  7.865376175215e-06, {   5,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  1.431643010727e-05,  7.865376175215e-06, {   3,   0,  -1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -1.358992895069e-05, -8.589654711634e-06, {   1,   1,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -6.752943335141e-06, -1.541496996741e-05, {   0,   0,   1,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.865529939728e-06,  1.429394470968e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 {  3.402680225724e-06, -1.874556374664e-05, {   0,   0,   0,   0,   0,   2,   8, -22,   8,  -6,   0,   0,   0 } },
 { -1.210281456558e-05,  1.004531492559e-05, {   2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.640848536997e-06, -1.650621708465e-05, {   2,   0,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.209263025985e-05, -5.412186064257e-08, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.799765464889e-05, -4.148814899266e-06, {   2,   0,   0,   0,   0, -17,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.178628079649e-05, -1.035877252467e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  2.906075109122e-06, -1.923542363590e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.478327122206e-05,  7.354760484217e-06, {   2,   0,   1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.310161720002e-06,  1.682722718820e-05, {   1,   1,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.453597116996e-07,  2.158861243888e-05, {   3,   1,  -2,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -4.252094845391e-06,  1.787921277117e-05, {   0,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.490821883408e-05,  7.212486298096e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -4,   0,   0,   0 } },
 { -1.733925346214e-05, -4.774802301142e-06, {   4,  -2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.989664384700e-06, -1.811506081446e-05, {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.208730082419e-05,  1.283764983580e-08, {   0,   0,   1,   0,   0, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  2.707408347300e-07,  2.182674656706e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -5,   0,   0,   0 } },
 { -2.031282613343e-05,  1.784534742020e-06, {   2,   2,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.126359015481e-05,  8.319651601807e-07, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  5.877134868655e-07, -2.150058536955e-05, {   1,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.514616449940e-05,  6.941300402256e-06, {   0,   0,   0,   0,   0,   0,   3,  -5,   1,   0,   0,   0,   0 } },
 {  1.817568873372e-05,  3.897064021199e-06, {   2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.496230546026e-05, -7.109283023449e-06, {   2,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.613470219323e-06, -1.243475294188e-05, {   4,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.047584616103e-05, -1.570221110125e-06, {   3,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.477586970108e-05,  7.267874235770e-06, {   2,   0,  -1,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.780905799116e-06, -1.825980522124e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -1.038587332450e-05,  1.165404447286e-05, {   1,  -1,   0,   0,   0,   0,  -1,   4,   0,  -2,   0,   0,   0 } },
 {  1.411029767330e-05,  7.915944750299e-06, {   2,   0,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.179372288286e-05,  2.271145955904e-07, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 {  1.142762570338e-05,  1.058684659856e-05, {   2,   0,  -1,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.941336408895e-05, -2.598604720287e-06, {   4,   0,  -3,   0,   0,   0, -27,  33,  -3,   0,   0,   0,   0 } },
 {  2.190962502550e-05,  9.739507939332e-08, {   6,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.199857888491e-05, -5.225521138560e-09, {   1,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.148738117020e-05,  5.119599199160e-07, {   2,   2,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.139279379343e-05,  1.059726852333e-05, {   2,   0,   1,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.348425972572e-05,  8.502490490091e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  5.027376696984e-06,  1.694633865373e-05, {   0,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.751189374009e-05, -4.426901390401e-06, {   1,   1,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.526707896577e-05,  6.667860773276e-06, {   0,   0,   1,   0,   0, -15,  11,   0,   2,   0,   0,   0,   0 } },
 {  1.989589619594e-05, -2.037968243255e-06, {   4,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.721557415108e-05, -4.714985876363e-06, {   2,   0,  -1,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.260015910170e-06, -1.366278513469e-05, {   2,   0,   0,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 {  1.483118116231e-05, -7.086711959199e-06, {   3,  -1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.177898449471e-06,  2.073243017455e-05, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  2.184657281583e-05,  5.164693891521e-08, {   0,   2,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.800926307628e-05, -3.887484231828e-06, {   1,  -1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.602329272572e-05, -5.870837329630e-06, {   1,   0,   0,   0,   0,   0, -35,  43,  -3,   0,   0,   0,   0 } },
 {  6.390061171267e-07, -2.125361871381e-05, {   0,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.182544252292e-05, -6.518562407039e-08, {   4,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.010760962365e-05, -1.782960026543e-06, {   2,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 { -7.279762077437e-06, -1.460451260262e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 { -1.512356660915e-05, -6.760023553509e-06, {   0,   0,   1,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.185413965716e-05,  9.336415503754e-09, {   0,   2,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.959067699749e-05,  2.270996716356e-06, {   1,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.519610434536e-05, -6.660586122584e-06, {   3,  -2,   1,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 { -1.519610434536e-05, -6.660586122584e-06, {   3,  -2,  -1,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 {  9.945878466131e-06, -1.190955942794e-05, {   1,  -1,   0,   0,   0,  16, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.844244055125e-05,  3.412840387386e-06, {   2,   0,  -1,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 { -8.086761098019e-06,  1.375478837268e-05, {   1,   1,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  1.319813499375e-05, -8.639816707181e-06, {   2,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.028827500020e-07, -2.173480826044e-05, {   0,   0,   3,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.406030276797e-05,  7.774530970060e-06, {   0,   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -8.371715669560e-06,  1.345506711870e-05, {   2,  -2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.235311132729e-06, -1.858889196436e-05, {   2,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.455125558235e-05,  7.267650666174e-06, {   2,   0,   5,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.618835159182e-06, -1.717248513136e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -4.561401600683e-06,  1.721513623376e-05, {   4,   0,  -1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.265586830723e-05, -9.120576379077e-06, {   2,   0,   0,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  2.169366018395e-05, -7.660717229314e-08, {   1,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.874203547853e-05,  3.027950795385e-06, {   4,  -2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.130170310195e-06,  1.763616640607e-05, {   4,  -2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.403019014095e-05,  7.734954777026e-06, {   2,   0,  -2,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -2.168390257449e-05, -6.831527294736e-08, {   2,  -2,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.986859955787e-05,  1.877471027803e-06, {   1,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.931749478944e-05,  2.425494228900e-06, {   0,   0,   0,   0,   0,   2,  -8,   7,   0,   0,   0,   0,   0 } },
 { -1.943781542822e-06,  1.979173406124e-05, {   4,   0,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.642627656305e-06, -1.508170839903e-05, {   1,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.138270735231e-05,  3.334798685994e-07, {   2,  -2,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.541031536461e-05, -6.303660334574e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   3, -10,   0,   0,   0 } },
 { -1.789901569501e-06,  1.991974528882e-05, {   1,  -1,   4,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789901569501e-06,  1.991974528882e-05, {   1,  -1,  -4,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789807699893e-06,  1.991975372331e-05, {   3,  -1,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789713830246e-06,  1.991976215736e-05, {   5,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789654727114e-06, -1.991976746746e-05, {   3,  -1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789550427430e-06, -1.991977683780e-05, {   1,   1,  -2,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.789119321553e-06, -1.991981556273e-05, {   1,  -3,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.789108891562e-06,  1.991981649951e-05, {   3,   1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.788688214838e-06, -1.991985427833e-05, {   1,   1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.784050288036e-06, -1.992027019319e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.784050288036e-06,  1.992027019319e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.401790495868e-05,  7.666300504274e-06, {   1,   0,   1,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 { -1.401790495868e-05,  7.666300504274e-06, {   1,   0,  -1,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 {  9.636199928142e-06, -1.203862218594e-05, {   1,  -1,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.266876833584e-07,  2.084737032977e-05, {   2,   0,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.478670839033e-05, -6.879531770372e-06, {   2,   0,  -2,   0,   0,   0,  18, -31,   3,   0,   0,   0,   0 } },
 {  2.154434920988e-05,  1.118557592777e-07, {   4,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.602002063397e-06,  1.704881812501e-05, {   1,   1,  -1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.716262133214e-05, -4.487317571695e-06, {   1,   1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.380592115940e-05,  7.843765921979e-06, {   0,   0,   1,   0,   0,   8, -14,   2,   0,  -2,   0,   0,   0 } },
 { -1.380522717205e-05,  7.843497835177e-06, {   0,   0,   1,   0,   0,  -8,  14,  -2,   0,   2,   0,   0,   0 } },
 { -1.803972639502e-05,  3.597946929981e-06, {   1,  -1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.909473260601e-06,  1.971823192707e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   4,   0 } },
 { -1.908556591987e-06,  1.971583416061e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0 } },
 {  2.159760681216e-05,  2.200052548594e-08, {   1,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.137940089314e-05,  2.372601792578e-07, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  1.343983841118e-05,  8.170519079448e-06, {   0,   2,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  7.962475609024e-06,  1.364664489550e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 { -1.709364058840e-05, -4.515304906764e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -9,   0,   0,   0 } },
 { -1.736910702740e-05, -4.232877330406e-06, {   1,  -3,   2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.325789460748e-06, -1.726750944949e-05, {   2,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.748882798984e-05,  4.092921855959e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 { -1.000581636200e-05, -1.157438090388e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.136915744867e-05, -2.062194433223e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 {  6.318637179842e-06,  1.525339905345e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 {  1.267766468070e-05,  8.893951403894e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -6,   3,   0,   0 } },
 { -1.257549210820e-05,  8.985347130793e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,  -1,   0,   0,   0,   0 } },
 {  2.008098836447e-05, -1.477420045939e-06, {   2,   0,  -1,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 {  5.690418515792e-06, -1.586356165581e-05, {   0,   0,   2,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.583488682056e-05, -5.708543655421e-06, {   2,  -2,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  5.483962870846e-06,  1.605641048368e-05, {   2,   0,  -1,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.888538737301e-05, -2.654377090840e-06, {   0,   0,   1,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 { -2.153816701842e-05, -7.622594725500e-10, {   3,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.506973418203e-06, -1.301478125635e-05, {   4,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.123550648344e-05,  2.832520359154e-07, {   2,   0,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 { -1.777566825008e-05,  3.738314764763e-06, {   1,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.135678536586e-06, -1.737597219444e-05, {   2,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.835736412228e-05, -3.146077099063e-06, {   2,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -5.316819998354e-06,  1.618194106332e-05, {   1,   1,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.742428579255e-06,  1.575110775445e-05, {   1,  -1,  -1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -4.087172483434e-06, -1.740552929767e-05, {   1,  -1,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.549501069374e-06, -1.493905777542e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -1.741162902120e-05,  4.064505398233e-06, {   2,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.569023888341e-05,  5.780637742019e-06, {   4,   0,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  1.740728314339e-05,  4.062039789084e-06, {   2,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -5.624629225449e-06,  1.584379862031e-05, {   2,   0,  -3,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -4.934901286436e-06, -1.653152172767e-05, {   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.314259233125e-06, -1.714645518874e-05, {   1,  -1,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  5.618978541694e-06,  1.583887444756e-05, {   2,   0,  -3,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.485022310801e-05, -6.603548106927e-06, {   0,   0,   1,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  6.866288607924e-06, -1.458740341779e-05, {   2,   0,  -1,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.561077527134e-05, -5.840361721630e-06, {   0,   0,   1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -6.705435027151e-06, -1.473946595370e-05, {   1,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.778166576181e-06,  1.265137300717e-05, {   1,  -1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.038632791772e-06, -2.037981052375e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.332727378009e-05, -8.090810524344e-06, {   2,   0,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0 } },
 {  6.366061815254e-06, -1.503769704185e-05, {   0,   0,   1,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.964698003700e-06, -1.243378132263e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 { -4.659214107464e-06,  1.673378910113e-05, {   3,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -2.787293260441e-06,  1.860527476773e-05, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 { -2.787293260441e-06,  1.860527476773e-05, {   2,   0,  -3,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 {  4.658870867779e-06,  1.673258333963e-05, {   3,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -8.022410397893e-06, -1.336873743040e-05, {   2,   0,  -1,   0,   0,   4,  -5,   0,   3,   0,   0,   0,   0 } },
 { -7.450140882676e-07, -2.064354560436e-05, {   0,   2,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.950016652852e-05,  1.886195425032e-06, {   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.568820125217e-05, -5.688543663434e-06, {   0,   0,   1,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.112867487348e-05, -2.453458229578e-07, {   4,   0,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.965388296471e-08, -2.132505304332e-05, {   0,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.115678070659e-05, -1.959944507749e-07, {   2,   0,   0,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0 } },
 {  4.471297033273e-06, -1.687899004583e-05, {   6,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.525610999013e-05,  6.080262613092e-06, {   0,   0,   1,   0,   0,  -4,   7,   0,  -6,   0,   0,   0,   0 } },
 {  3.400529610051e-06,  1.792562615222e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,   1,   0,   0,   0 } },
 { -2.102389102862e-05,  2.882653650057e-07, {   6,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.764381381631e-05,  3.659089779237e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 {  2.021952136056e-05, -1.080804345650e-06, {   2,   0,   1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  1.782190452019e-05,  3.477121157879e-06, {   0,   2,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.522577255797e-05,  6.066300028667e-06, {   0,   0,   1,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  9.726874189649e-06, -1.155025596700e-05, {   4,   0,  -2,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -9.726874189649e-06, -1.155025596700e-05, {   0,   0,   0,   0,   0,   0,  33, -40,  -3,   0,   0,   0,   0 } },
 {  8.931756348005e-06, -1.233631384026e-05, {   0,   0,   1,   0,   0,   0, -11,  15,   0,   0,   0,   0,   0 } },
 {  2.121552981273e-05,  4.748266104737e-08, {   2,   0,  -2,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.214511206067e-06, -1.304388160461e-05, {   0,   0,   1,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 {  4.158765308322e-06, -1.709741673285e-05, {   2,  -2,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.412755421990e-07,  2.060698647546e-05, {   2,   0,  -1,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0 } },
 {  1.312784029777e-05,  8.112731634146e-06, {   2,   0,  -1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0 } },
 {  1.769807186906e-05,  3.540000165625e-06, {   1,   1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -8.297041166858e-06,  1.293096344843e-05, {   1,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.917486818839e-05,  2.046341277636e-06, {   0,   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 { -1.641175017297e-05, -4.804492768272e-06, {   0,   0,   0,   0,   0,   4,  -7,   0,   3,   0,   0,   0,   0 } },
 { -1.144073965142e-06, -2.006995150935e-05, {   4,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.682262254092e-05, -4.384633795889e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   6,  -3,   0,   0 } },
 {  1.331605423478e-06, -1.986169523687e-05, {   1,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  2.524096203749e-06, -1.864411544886e-05, {   0,   0,   1,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  9.742137147978e-06,  1.142293323876e-05, {   3,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.365539295791e-05,  7.502183184150e-06, {   1,   0,   3,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.365539295791e-05,  7.502183184150e-06, {   1,   0,  -3,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.739596596112e-06,  1.841757134939e-05, {   1,   0,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.274597947218e-05,  8.399778474560e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 { -2.010896378103e-05, -1.034125794122e-06, {   3,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.606893095143e-05, -5.073495146050e-06, {   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.254673389169e-05, -8.595313947412e-06, {   1,   0,   0,   0,   2, -30,  28,   0,   0,   0,   0,   0,   0 } },
 {  4.191300821317e-06, -1.694889997381e-05, {   2,   2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.509886561380e-06, -1.862899366308e-05, {   0,   0,   1,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 { -1.363739255625e-05,  7.490404130888e-06, {   2,  -2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.207734522440e-05, -9.044699436189e-06, {   4,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.105741372450e-05, -5.804698107158e-08, {   3,   0,  -3,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.698995388197e-06, -1.541522991884e-05, {   0,   0,   1,   0,   0,   0,   3,  -6,   3,   0,   0,   0,   0 } },
 { -4.758858724341e-06, -1.635217016534e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -9.105204935722e-06, -1.200465864653e-05, {   0,   0,   1,   0,   0,   0,  -7,  10,  -1,   0,   0,   0,   0 } },
 { -5.260468596906e-06, -1.584249653669e-05, {   1,  -1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.700751567267e-06,  1.839949737256e-05, {   1,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.417475504295e-05,  6.923650772282e-06, {   2,   0,   1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.259346081294e-05,  8.500464059060e-06, {   0,   0,   2,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 { -2.057265394150e-05,  5.152029335778e-07, {   2,  -2,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.679467908235e-06,  1.540627180755e-05, {   1,   1,  -2,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -2.102208698355e-05, -5.788562685565e-08, {   1,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.329278533459e-05,  7.786765845671e-06, {   2,   0,  -1,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 { -1.243861524376e-05, -8.627373073804e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -1.257697303444e-05,  8.488441560496e-06, {   0,   0,   2,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -1.103204067924e-05,  1.003291842517e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 {  1.364548728510e-05,  7.407431698225e-06, {   0,   0,   2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  1.626554214703e-05, -4.784088990783e-06, {   1,   1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.100387083093e-05,  3.555753388060e-08, {   2,   0,  -2,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.300927332287e-05, -8.027015960149e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  1.988942449501e-05,  1.144216787418e-06, {   1,   0,  -3,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -9.929048381785e-06, -1.110455304850e-05, {   0,   0,   0,   0,   0,   7,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.103137132423e-05,  1.843374497448e-09, {   1,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.460273247631e-05,  6.422937232416e-06, {   0,   0,   1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 { -1.564972916833e-05,  5.367525669250e-06, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   1,   0,   0,   0 } },
 { -4.356922360237e-06,  1.665797619126e-05, {   2,   0,  -3,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.687451558697e-05,  4.134045716641e-06, {   1,  -1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.090516705130e-07, -2.049028455845e-05, {   2,   0,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.019487136623e-05, -1.080166290752e-05, {   2,   0,  -1,   0,   0,  -9,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.382176488014e-05,  7.167950749048e-06, {   1,  -1,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  2.664166443484e-06,  1.832413964740e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 {  1.046330473378e-05, -1.050725456718e-05, {   2,   0,   0,   0,   0,  -8,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.397838587562e-05,  6.981545530231e-06, {   0,   4,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.397838587562e-05,  6.981545530231e-06, {   0,   4,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.059809207866e-05, -1.035950644889e-05, {   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.874469754773e-05,  2.211077584209e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 { -1.482197649674e-05, -6.128249898315e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 {  1.678086148910e-05,  4.169080190795e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -1.997659460390e-05, -9.672993493554e-07, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.177941937931e-06, -1.676067030848e-05, {   1,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.096139055021e-05,  9.976705329877e-06, {   3,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.601649551865e-05, -4.913409152554e-06, {   0,   0,   1,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 { -4.790656928335e-06,  1.613901230111e-05, {   1,   1,   0,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.040964650500e-06, -1.888781147722e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0 } },
 {  1.084507954765e-06,  1.984266402004e-05, {   2,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 {  1.445980463143e-05, -6.462222915896e-06, {   0,   0,   1,   0,   0,  -2,   6,  -5,   0,   0,   0,   0,   0 } },
 { -1.325819738125e-05,  7.659918755319e-06, {   2,   0,   0,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 {  1.673555073016e-05, -4.167348185776e-06, {   1,   1,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.529540527407e-06,  1.237319501816e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  2.089802156532e-05,  3.888850852106e-09, {   1,  -2,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.750287629506e-06, -1.214858136013e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.287046324933e-05, -8.023687111388e-06, {   2,   0,  -1,   0,   0,  -1,   5,  -3,   0,   0,   0,   0,   0 } },
 { -4.009044856506e-06, -1.687748019740e-05, {   3,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.037876045665e-05,  5.064397871205e-07, {   3,   1,  -1,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 {  2.037876045665e-05,  5.064397871205e-07, {   3,   1,  -3,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -4.169907841921e-06,  1.671434238591e-05, {   2,   0,  -3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.856282874911e-05, -2.320479177593e-06, {   0,   0,   0,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 { -1.219602572392e-05, -8.683466162181e-06, {   2,   0,   0,   0,   0,   0,  15, -20,   0,   0,   0,   0,   0 } },
 {  4.003740243509e-06,  1.686320594730e-05, {   2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.057283938315e-07, -2.006106646029e-05, {   6,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.391527383914e-05,  6.948681273526e-06, {   2,   0,   3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.858413008644e-06, -1.899781868274e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -5.979004320621e-06,  1.486941162519e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 { -6.789782264693e-06, -1.405718953979e-05, {   2,   0,   0,   0,   0,   0,  17, -21,   0,   0,   0,   0,   0 } },
 { -3.980695988528e-06, -1.685961218421e-05, {   0,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.080829681182e-05, -2.103965217738e-08, {   3,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.974482305205e-05, -1.077459714253e-06, {   2,   0,  -1,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.080237504315e-05,  1.899112924657e-08, {   0,   2,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.194248738039e-05,  8.870044177399e-06, {   1,  -1,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -9.328898710602e-06,  1.146539089784e-05, {   4,  -1,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.086072634168e-05, -9.924778323768e-06, {   0,   0,   0,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 { -2.077289399085e-05, -1.256070826831e-08, {   1,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.302030033889e-05,  7.756323814448e-06, {   2,   0,  -3,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.333632322245e-05,  7.434706763759e-06, {   4,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.811002737241e-06, -1.895547535266e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -1.903307040175e-05,  1.729909787606e-06, {   1,   1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.725385348138e-06,  1.403688872587e-05, {   0,   0,   0,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 {  1.724050819700e-05,  3.509284322590e-06, {   1,   1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.132201573350e-06,  1.761668049726e-05, {   4,   0,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.865131355286e-06,  1.688228835351e-05, {   1,  -1,  -2,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.854401043029e-05,  2.202886127373e-06, {   0,   0,   1,   0,   0, -26,  33,  -8,   3,   0,   0,   0,   0 } },
 { -1.530187085553e-05, -5.440455317593e-06, {   0,   0,   1,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  1.649886253169e-05, -4.231723863282e-06, {   2,   0,  -1,   0,   0, -31,  39,   0,  -1,   0,   0,   0,   0 } },
 {  1.498977961758e-05,  5.739064253066e-06, {   3,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.577133881529e-06,  1.314902645078e-05, {   2,   0,   0,   0,   0,   0,  -6,   0,   5,   0,   0,   0,   0 } },
 { -1.180441387290e-05, -8.914542744359e-06, {   2,   0,  -2,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.077333684631e-05,  9.931910828839e-06, {   3,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.068097503355e-05, -2.330757242596e-08, {   2,   0,  -4,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.879428575125e-07,  2.011070214643e-05, {   2,   0,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.509188988282e-05,  5.595299587396e-06, {   0,   0,   0,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.062502085419e-05, -5.489482118253e-08, {   0,   2,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.187515502183e-06,  1.948494556323e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 { -7.366635208302e-08, -2.059552965522e-05, {   2,   0,   3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.961826628917e-06, -1.669963183614e-05, {   1,   1,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.404199548732e-06,  1.725422935837e-05, {   2,   0,  -1,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 {  8.278853711257e-06, -1.237683087872e-05, {   2,  -2,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.790871092801e-05, -2.741965849251e-06, {   4,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.041495274736e-05, -2.270144439590e-07, {   2,   0,  -3,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.390998033497e-05,  6.729611003727e-06, {   2,   0,   1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.707302873625e-06, -1.293140172444e-05, {   0,   0,   3,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.335245692255e-05, -7.278420956119e-06, {   2,   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.095884855000e-05,  9.670533739845e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 {  1.174818417206e-05, -8.870026321822e-06, {   2,   0,   0,   0,   0,  -1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.935553095660e-06, -1.667057871460e-05, {   1,   1,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.499315445292e-06, -1.810607397664e-05, {   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.183579138167e-05,  8.763734081419e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  8.405738963027e-07, -1.975695647616e-05, {   0,   0,   2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.133344931274e-05,  9.263091549195e-06, {   0,   0,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  1.635719183739e-05, -4.238241767610e-06, {   1,   1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -5.633222056182e-07, -2.002079919749e-05, {   0,   2,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.641010472552e-06,  1.793791074998e-05, {   2,   0,   0,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 {  1.700457284871e-05,  3.572976418544e-06, {   3,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.700457284871e-05,  3.572976418544e-06, {   1,   0,   1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 {  9.374277618094e-06, -1.119837597787e-05, {   1,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.368626476959e-05, -6.872241165144e-06, {   2,  -2,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.574794539301e-05,  4.806960552104e-06, {   1,  -3,   0,   0,   0,  51, -55,   0,   0,   0,   0,   0,   0 } },
 {  2.672866254181e-06,  1.786904639015e-05, {   2,   0,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.731084319221e-05, -3.230480414057e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   5,   0,   0,   0 } },
 {  2.397134463094e-06,  1.814150415232e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 {  2.476924366659e-06, -1.805863223097e-05, {   3,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.342022404645e-05, -7.108807319021e-06, {   4,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  3.882138569975e-07, -2.013871007408e-05, {   0,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.377225759340e-05, -6.751039694597e-06, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   2,   0,   0,   0 } },
 {  3.214062478961e-06, -1.729916629090e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  5.715561036434e-06, -1.479765693502e-05, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.253759561398e-06, -1.825506301837e-05, {   2,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.607817774022e-06, -1.289171502473e-05, {   0,   0,   0,   0,   0,   5,  -5,  -4,   0,   0,   0,   0,   0 } },
 {  1.307558130616e-05, -7.422805048123e-06, {   4,   0,  -2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  6.638520376664e-06,  1.385524158276e-05, {   2,   0,  -1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 {  7.934881230156e-06, -1.255738931868e-05, {   6,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.111226222757e-06, -1.437864413464e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.226768511192e-05,  8.221781631017e-06, {   0,   0,   1,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 {  1.674826795648e-05,  3.732172116859e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 {  1.121428724797e-05, -9.254721499663e-06, {   2,   0,  -2,   0,   0,   3,  -2,  -2,   0,   0,   0,   0,   0 } },
 {  1.710151949817e-05, -3.364372883670e-06, {   1,  -1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.356177112034e-06,  1.310886669625e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  3.208093767797e-06, -1.724392617783e-05, {   1,   1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.299016284470e-05, -7.460838219957e-06, {   3,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -7.354214549509e-06,  1.308639569130e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  1.366084073825e-05, -6.774949170248e-06, {   3,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.940850290101e-06, -1.847864111757e-05, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -1.320689691506e-05,  7.203217313593e-06, {   0,   0,   2,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 { -1.801059080225e-05,  2.393712849378e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.856041302651e-05,  1.839936276090e-06, {   2,   0,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.100215282446e-07,  2.018890071667e-05, {   1,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.407919139255e-05, -6.319041165049e-06, {   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  6.007310432277e-06, -1.439004011467e-05, {   1,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.107895803309e-05,  9.314941147817e-06, {   2,   0,  -2,   0,   0,   0,   1,   2,  -3,   0,   0,   0,   0 } },
 {  5.651551081055e-06, -1.473922753856e-05, {   3,   1,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.959493888013e-05, -7.952936886088e-07, {   3,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.469005673647e-06,  1.591149122746e-05, {   1,  -1,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.905429291179e-06,  1.547364521229e-05, {   2,   0,  -1,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 { -1.217774264545e-05, -8.200037024136e-06, {   1,  -1,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.942672615105e-06,  1.643012591631e-05, {   0,   0,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.111262932663e-06,  1.526084629362e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -7,   5,   0,   0,   0 } },
 {  1.914391186807e-06, -1.845309966136e-05, {   4,   0,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.681305651200e-05, -3.548942837635e-06, {   2,   0,  -1,   0,   0, -18,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.881812688421e-06,  1.647599300443e-05, {   1,   1,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.501069874311e-07,  1.990666409446e-05, {   2,   2,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.382996687296e-05, -6.526479341572e-06, {   4,  -2,   0,   0,   0, -33,  31,   0,   0,   0,   0,   0,   0 } },
 { -2.032792865864e-05, -2.410261111185e-08, {   2,  -2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.020146994158e-05,  1.463937800310e-07, {   2,  -2,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -8.574882705486e-06,  1.175990650173e-05, {   0,   2,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  9.771139477916e-06, -1.055940175722e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -3.885005567540e-06,  1.644322941461e-05, {   3,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.799630115782e-06, -1.252778133156e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -5,   0,   0,   0 } },
 { -7.873905811254e-06, -1.245282577509e-05, {   6,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.777626913651e-05,  2.549310613076e-06, {   2,   0,  -1,   0,   0,  -6,   9,   0,  -3,   0,   0,   0,   0 } },
 {  1.245383610042e-06, -1.907929437647e-05, {   4,   0,  -2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.361841028145e-05,  6.699041956316e-06, {   2,   0,  -2,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.567647959146e-06,  1.174708894858e-05, {   0,   2,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.999754497071e-05,  3.133610447236e-07, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.747781953382e-05, -2.829064078371e-06, {   4,   2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.632005837309e-05, -3.973485684080e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,  -2,   0,   0,   0 } },
 { -2.021932184038e-05,  7.377171400043e-08, {   2,   0,  -2,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  1.944395612413e-05, -8.362312680706e-07, {   0,   0,   2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.240382207861e-06, -1.303967996114e-05, {   2,   0,  -1,   0,  -1,   0,   6,  -3,  -3,   0,   0,   0,   0 } },
 { -7.742151173558e-06,  1.252590869665e-05, {   3,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.235289581495e-06, -1.303167672017e-05, {   2,   0,  -1,   0,   1,   0,  -6,   3,   3,   0,   0,   0,   0 } },
 { -1.620423698492e-05,  4.056415687784e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 {  4.555466591701e-06, -1.569927454716e-05, {   4,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.018373057090e-05, -6.956961307290e-08, {   4,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.702576783501e-05, -3.225725999712e-06, {   4,   0,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  1.999137490176e-05, -2.594592594056e-07, {   1,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.015498503569e-05,  9.451247884079e-08, {   2,   0,   1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.681279642455e-05,  3.425258258904e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -4.777393192183e-06,  1.545914912086e-05, {   2,   0,  -1,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.809024769980e-06,  1.341597784043e-05, {   1,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -5.196860230577e-07, -1.970019233690e-05, {   4,   0,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  2.056407417359e-06, -1.815808815832e-05, {   0,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.689915108192e-05, -3.314677207246e-06, {   1,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.747810728148e-06, -1.146572457922e-05, {   6,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.071685677744e-06, -1.614060469295e-05, {   2,  -2,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.089215227901e-05,  9.319122383698e-06, {   2,   0,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -2.012118176744e-05,  8.581010016887e-08, {   2,   0,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  9.528194554653e-06, -1.065197024100e-05, {   0,   0,   0,   0,   0,   3,  -5,   0,   2,  -1,   0,   0,   0 } },
 {  1.467083659096e-05, -5.508631821497e-06, {   1,  -1,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 { -4.255064086568e-06,  1.592428290597e-05, {   2,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 { -7.973404613464e-06,  1.220248169689e-05, {   2,   0,  -2,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -7.714762612293e-06, -1.246060288108e-05, {   1,  -1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.756484338811e-05,  2.608202116240e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.781609049381e-05,  2.351949257943e-06, {   3,  -2,   0,   0,   0, -20,  19,   7,  -8,   6,   0,   0,   0 } },
 { -1.480272417887e-05, -5.358815081912e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -6,   3,   0,   0 } },
 { -1.665331359976e-05,  3.507967319348e-06, {   1,  -1,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.566402483798e-06, -1.857933800211e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,  -1,   0,   0 } },
 {  1.566223945845e-06, -1.857928623717e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   1,   0,   0 } },
 {  1.706032122417e-05,  3.077452584095e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 {  1.297494748369e-05, -7.158084613718e-06, {   3,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.735653975162e-06, -1.639143856471e-05, {   1,  -1,   0,   0,   0,  24, -23,   0,   0,   0,   0,   0,   0 } },
 {  9.539410644187e-06,  1.058620132091e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   7,   0,   0,   0 } },
 { -4.745277792790e-06,  1.537839217306e-05, {   1,   1,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -8.079257812927e-07,  1.931008133322e-05, {   0,   0,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 { -1.994117228251e-05, -1.760389586489e-07, {   0,   0,   2,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -2.005272385886e-05,  6.299119455016e-08, {   2,   0,  -2,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  8.147282332811e-06, -1.194527991092e-05, {   2,  -2,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.972488736555e-05,  3.671791705110e-07, {   4,   0,  -2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.744052046491e-05,  2.636747739772e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -2.007272288327e-05,  3.653052888470e-09, {   2,  -2,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.348197515232e-05,  6.589591809026e-06, {   2,  -2,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.259764769012e-06,  1.481171820512e-05, {   2,   0,   2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.681579141736e-05,  3.253185823361e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 { -5.257742329887e-06,  1.480741122959e-05, {   2,   0,   2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.873487468376e-05, -1.323748901485e-06, {   0,   2,  -2,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  1.121682862862e-05,  8.832587499121e-06, {   0,   0,   1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 { -2.000002409556e-05, -4.896520427939e-08, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.686301277823e-05, -3.174622655773e-06, {   1,   1,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.370725058724e-06,  1.866254210558e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -3.940914383427e-06, -1.607915597255e-05, {   1,   1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.741978040087e-06,  1.326427123828e-05, {   2,   0,  -1,   0,   0,   0,  18, -22,   0,   0,   0,   0,   0 } },
 {  3.132737550789e-06,  1.686785656410e-05, {   1,  -1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.999999999210e-05, -5.619960190680e-10, {   2,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999411e-05, -4.852015320021e-10, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999420e-05, -4.817108735079e-10, {   0,   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999437e-05, -4.747295564997e-10, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -1.999999999461e-05,  4.642575809888e-10, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999485e-05, -4.537856054748e-10, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999532e-05, -4.328416544593e-10, {   2,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999547e-05, -4.258603374507e-10, {   2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999561e-05,  4.188790204480e-10, {   2,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999583e-05, -4.084070449389e-10, {   2,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999590e-05,  4.049163864350e-10, {   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.999999999597e-05,  4.014257279317e-10, {   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999645e-05,  3.769911184085e-10, {   4,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999658e-05,  3.700098014017e-10, {   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  1.999999999671e-05, -3.630284843911e-10, {   0,   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999707e-05,  3.420845333742e-10, {   4,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999713e-05, -3.385938748691e-10, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999753e-05,  3.141592653461e-10, {   6,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999764e-05,  3.071779483389e-10, {   6,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999790e-05, -2.897246558236e-10, {   4,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999800e-05, -2.827433388141e-10, {   4,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999800e-05, -2.827433388141e-10, {   4,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999815e-05,  2.722713633027e-10, {   4,  -1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999829e-05, -2.617993877954e-10, {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999838e-05, -2.548180707858e-10, {   3,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999838e-05, -2.548180707858e-10, {   2,   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999846e-05,  2.478367537769e-10, {   4,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999851e-05, -2.443460952713e-10, {   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999851e-05, -2.443460952713e-10, {   2,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999871e-05,  2.268928027544e-10, {   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999879e-05, -2.199114857472e-10, {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999898e-05, -2.024581932326e-10, {   2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  1.999999999914e-05, -1.850049007080e-10, {   2,  -1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999944e-05, -1.500983156681e-10, {   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999971e-05, -1.082104136277e-10, {   2,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.000000000018e-05, -8.377580410223e-11, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.999999999998e-05, -2.792526803084e-11, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.519978599398e-05,  4.798650966193e-06, {   1,  -3,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.618364132332e-05,  3.809397852359e-06, {   1,   1,  -2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.511083628165e-05, -4.872653498467e-06, {   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0 } },
 {  1.803569427167e-05, -1.946192354659e-06, {   4,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.756980926226e-06,  1.822356072285e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.005789823013e-06,  1.396241316882e-05, {   3,   0,  -2,   0,   0, -17,   8,  17,   0,   0,   0,   0,   0 } },
 {  8.082830669922e-07, -1.914856403269e-05, {   3,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.521390752715e-05, -4.742268493188e-06, {   4,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.864605987918e-05, -1.303896326859e-06, {   6,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.760452527424e-05, -2.341434226066e-06, {   2,   0,  -1,   0,   0,   3,  -9,   4,   0,   0,   0,   0,   0 } },
 { -6.183089687112e-06, -1.376104646919e-05, {   2,   0,  -1,   0,   0,   1,   6, -11,   0,   0,   0,   0,   0 } },
 {  3.793403521970e-06,  1.614669796177e-05, {   3,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.391206445561e-05,  6.022439876055e-06, {   2,   0,  -1,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 {  1.974746688432e-05, -1.861949061760e-07, {   2,   2,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -8.821674050972e-06, -1.110811522085e-05, {   2,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  3.882583379452e-06,  1.604592240994e-05, {   3,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.649084085864e-05,  3.423690301081e-06, {   2,   0,   0,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 { -8.828177476953e-06,  1.108293030072e-05, {   2,   0,   0,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 { -6.351487879444e-06,  1.355933035012e-05, {   1,   1,  -1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  1.584328740143e-05, -4.067406389855e-06, {   1,   1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.749012965317e-06, -1.016071427063e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,  -2,   0,   0,   0 } },
 { -1.965149723648e-05,  2.553274702900e-07, {   2,  -2,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -3.751651537946e-06, -1.614665138456e-05, {   1,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.787558622486e-06,  1.311055022720e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.776840171239e-05, -2.127820110654e-06, {   2,   0,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 { -2.974201327034e-06, -1.691793358899e-05, {   4,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  5.664994943694e-08,  1.983539841148e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,  -2,   0,   0,   0 } },
 {  1.976001997500e-05,  1.304984847221e-07, {   0,   2,  -2,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 { -4.661776754450e-06,  1.522426344215e-05, {   1,   1,   0,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  8.578303847870e-06,  1.130644403091e-05, {   1,  -1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  7.594400562152e-06,  1.228963027130e-05, {   3,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.935350362315e-05,  5.220213164109e-07, {   2,   0,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.335950904717e-05,  6.512867502653e-06, {   2,   0,   1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.099705056399e-05, -8.873510202393e-06, {   2,   0,   0,   0,   0,  -2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.985182013363e-06, -1.788337746572e-05, {   2,   0,   0,   0,   0,   0,   8, -19,   0,   3,   0,   0,   0 } },
 {  8.974502438323e-06, -1.088847890687e-05, {   1,  -1,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  7.557121649160e-06, -1.230431148221e-05, {   0,   0,   0,   0,   0,   0,   5,  -9,   2,   0,   0,   0,   0 } },
 {  1.862488863678e-05,  1.221976195038e-06, {   2,   0,   0,   0,   0,  -6,   9,   0,  -3,   0,   0,   0,   0 } },
 {  1.343730405989e-08, -1.983000192594e-05, {   6,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.221229767212e-06, -1.061319872553e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -1.829749257226e-05,  1.536044553511e-06, {   1,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.852291620639e-05,  1.306580988457e-06, {   4,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.399331710643e-05, -5.830350662368e-06, {   2,   0,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.183601620828e-06,  1.263933174874e-05, {   2,   0,  -1,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0 } },
 { -1.285157861979e-05,  6.967969759314e-06, {   2,   0,  -1,   0,   0,   0,   7, -10,   2,   0,   0,   0,   0 } },
 { -1.313403325323e-05,  6.679291798967e-06, {   0,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.181330154188e-06, -1.263145301129e-05, {   0,   0,   1,   0,   0,   6, -14,   8,   0,   0,   0,  -1,   0 } },
 {  7.181330154188e-06, -1.263145301129e-05, {   0,   0,   1,   0,   0,  -6,  14,  -8,   0,   0,   0,   1,   0 } },
 { -1.135425442372e-06,  1.867730235691e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 { -1.710389254350e-05, -2.698095936450e-06, {   0,   0,   1,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.278094562527e-05, -7.020893285561e-06, {   3,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.014435390695e-06,  1.077877958458e-05, {   2,   0,  -1,   0,   1,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.331352200042e-05,  6.471090806604e-06, {   2,   0,   1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.413194659967e-06,  1.035998718628e-05, {   1,  -1,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.580357083386e-05,  3.968583307306e-06, {   0,   0,   0,   0,   0,   0,   1,  -3,   3,   0,   0,   0,   0 } },
 {  1.974341411531e-05,  2.635975990507e-08, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -1.449629662661e-05, -5.271250500554e-06, {   2,   0,   0,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.171734903553e-05, -8.045474406752e-06, {   4,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  9.268152072397e-06,  1.049358969041e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -7.527071331253e-06,  1.223442186739e-05, {   0,   0,   2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.512697284476e-05,  4.623025129191e-06, {   1,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.600781742866e-05,  3.741791575934e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 { -1.611634799675e-05, -3.632993265209e-06, {   1,  -1,  -3,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.357703638380e-05,  6.170173784151e-06, {   1,  -1,   0,   0,   0,  -1,   8, -10,   0,   0,   0,   0,   0 } },
 { -7.568521849844e-06, -1.217832761083e-05, {   1,   1,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.169258126709e-05, -8.052230705708e-06, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.847839002841e-06,  1.289426061085e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   2,   0,   0,   0 } },
 { -2.248252432932e-06,  1.749257509311e-05, {   4,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.544137653107e-07,  1.897957419936e-05, {   1,   0,   0,   0,   0,   2, -32,  30,   8,  -6,   0,   0,   0 } },
 { -1.628176685018e-05, -3.449210451360e-06, {   2,   0,  -2,   0,   0,  -4,  12,  -7,   0,   0,   0,   0,   0 } },
 { -1.287040087849e-05,  6.858407141788e-06, {   2,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.965295388396e-05,  7.554812760955e-08, {   4,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.832556692432e-06,  1.688595591032e-05, {   2,   0,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.907545283919e-05,  6.430448234530e-07, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  1.399665401796e-05,  5.709683893868e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -7,   0,   0,   0 } },
 {  1.970557831578e-05,  1.012997673663e-10, {   0,   0,   2,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.574704291280e-05, -3.946423494816e-06, {   1,   1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.925239594786e-05, -4.350237141261e-07, {   2,   0,  -2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.574601896622e-05, -3.938250405076e-06, {   1,   1,   2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.490888890745e-06, -1.518181412455e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -7,  10,   0,   0,   0 } },
 {  1.854197388521e-05,  1.125460585263e-06, {   2,   0,   1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 {  5.661733160630e-06, -1.400485889874e-05, {   2,   0,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.959651264173e-06, -9.705103639858e-06, {   0,   0,   0,   0,   0,  14, -23,   0,   3,   0,   0,  -1,   0 } },
 {  1.129723575842e-05,  8.367367033562e-06, {   0,   0,   0,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 {  1.207636265172e-06,  1.845432632403e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 { -9.263124550197e-06, -1.039247162484e-05, {   0,   0,   1,   0,   0, -18,  26, -19,   0,   3,   0,   0,   0 } },
 {  7.211659639512e-06, -1.244351524057e-05, {   0,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.373230059840e-06,  1.227624845347e-05, {   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.737949335270e-06,  1.590993007633e-05, {   1,  -1,   1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.963498607796e-05, -9.664599152783e-09, {   4,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.241369049246e-06,  1.140271461249e-05, {   0,   0,   2,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -9.002499892640e-06,  1.064076754513e-05, {   0,   0,   1,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 {  1.074407402198e-06,  1.856831157815e-05, {   2,   0,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  6.326430835767e-06,  1.331579590723e-05, {   2,  -2,   2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -9.414969483237e-06,  1.022679893326e-05, {   4,   0,  -2,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -8.232462389679e-06,  1.140238445306e-05, {   0,   0,   2,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  4.293287360242e-07,  1.920237191780e-05, {   2,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.824071978530e-06, -1.580370982882e-05, {   2,   0,   4,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.748356915381e-05, -2.131417920153e-06, {   0,   0,   1,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.454051918693e-06,  1.615340504924e-05, {   2,  -2,   0,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 { -3.454051918693e-06,  1.615340504924e-05, {   2,  -2,  -2,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 {  9.720676069816e-07, -1.861914788784e-05, {   4,   0,  -3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.309458080674e-06,  1.327931098913e-05, {   2,  -2,  -2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -1.708728303491e-05,  2.493838491187e-06, {   1,   1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.566366182417e-05,  3.917090893543e-06, {   1,   3,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.906899919724e-06, -1.767083592809e-05, {   4,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.919228425344e-05,  3.780158058943e-07, {   0,   0,   2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.939229961375e-05, -1.769014103112e-07, {   2,   2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.471860016776e-05,  4.849331774894e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0 } },
 {  1.942916527238e-05, -1.378875070073e-07, {   4,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.255760935536e-05,  7.005896651808e-06, {   1,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.985068027778e-06,  1.557808331777e-05, {   0,   0,   2,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.463272980384e-06, -1.609647196429e-05, {   1,   1,   0,   0,   0, -40,  39,   0,   0,   0,   0,   0,   0 } },
 { -4.257217389100e-06,  1.529006094737e-05, {   1,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.679143757099e-05,  2.752315138600e-06, {   2,   2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.256412099795e-06,  1.528701402246e-05, {   1,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.753095518843e-05,  2.002398860579e-06, {   2,   0,   1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.482419268587e-05, -4.707816585129e-06, {   0,   0,   1,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.247832129248e-05, -7.050170775019e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  8.262942785125e-06, -1.125979707553e-05, {   1,  -1,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.732203863767e-06,  1.778704132058e-05, {   2,   0,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  8.622491154517e-06, -1.089028053248e-05, {   2,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -1.180742208220e-05,  7.701606742794e-06, {   2,   0,   0,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 {  1.932482615649e-05, -1.828313058827e-07, {   2,  -2,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  7.473535578346e-06, -1.203108254062e-05, {   1,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  9.950272287867e-06, -9.550704426097e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -3,   0,   0,   0 } },
 {  1.508594097806e-05, -4.413455259584e-06, {   1,   1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.512216675947e-06,  1.098695041278e-05, {   2,   0,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0 } },
 {  4.985522190352e-06,  1.450365607622e-05, {   2,   0,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.366126999630e-06,  1.212287079782e-05, {   2,  -2,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  1.930150425683e-05,  1.667830803575e-07, {   0,   0,   2,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.927404979386e-05,  1.806208740491e-07, {   4,   0,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.749862417363e-05, -1.955485336132e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 { -1.942445985154e-05, -2.058161838380e-08, {   4,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.940741486114e-05,  3.394777485211e-08, {   6,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.929175106982e-05,  1.495863837041e-07, {   2,   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  3.634031313724e-06,  1.580168228572e-05, {   3,   1,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.275250683816e-06,  1.515840937770e-05, {   2,   0,  -2,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 { -8.650859881907e-06, -1.077976319044e-05, {   0,   0,   0,   0,   0,   0,  10, -21,   2,   0,   0,   0,   0 } },
 {  1.199511534564e-05, -7.433856285579e-06, {   2,   0,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0 } },
 {  1.942318265109e-05, -2.858940456614e-09, {   0,   2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.977583758956e-06, -1.344816260824e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  4.294279781712e-06,  1.512852844988e-05, {   1,   0,  -1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 { -8.294725350235e-06,  1.112297156537e-05, {   0,   0,   1,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 {  1.627138514983e-05,  3.145335779275e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  7.839572302080e-06, -1.156899887550e-05, {   0,   0,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.972797401941e-06, -1.743557603097e-05, {   2,   0,  -1,   0,   0,   0,   8, -19,   0,   3,   0,   0,   0 } },
 {  1.550590444163e-05, -3.900416506455e-06, {   1,   1,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.711772337587e-06,  1.569418617608e-05, {   3,  -1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.847824000536e-05, -9.231741454868e-07, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -1.936907417856e-05, -3.181915844577e-08, {   1,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.039623937140e-06,  1.036085770458e-05, {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.432141895292e-05,  5.078140508015e-06, {   2,   2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  5.758017828239e-06,  1.364138806156e-05, {   1,  -3,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  9.822551443151e-06, -9.576639706013e-06, {   2,   0,   0,   0,   0,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.503223192038e-06,  1.089381825626e-05, {   1,  -1,   2,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 { -8.503223192038e-06,  1.089381825626e-05, {   1,  -1,   0,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 { -1.237097727617e-05, -7.024502204666e-06, {   4,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.299730919310e-06,  1.308913549216e-05, {   0,   0,   0,   0,   0,   2,   2, -11,   8,  -6,   0,   0,   0 } },
 {  7.261313385858e-06,  1.212194764019e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -1.542576352496e-05, -3.953713890332e-06, {   2,   0,  -1,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.669568782743e-06, -1.070710600284e-05, {   1,  -1,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.716659532791e-05,  2.207419263062e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.699150271359e-05, -2.378142005581e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   6,  -3,   0,   0 } },
 {  8.534592132502e-06,  1.083358828029e-05, {   0,   0,   2,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -8.534908825855e-06,  1.083289827676e-05, {   0,   0,   2,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  1.729712786785e-05, -2.068890789054e-06, {   2,   0,  -1,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 { -2.079995265592e-06,  1.728465173098e-05, {   2,   0,  -2,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  6.795772439028e-06,  1.256508530354e-05, {   3,  -1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -5.722110800271e-07, -1.878687549310e-05, {   6,   0,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  7.254996432634e-06, -1.209454857826e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.917938693838e-05, -1.662042233497e-07, {   2,  -2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.848325361123e-05, -8.605293937539e-07, {   2,   0,  -1,   0,   0,   5,  -6,   0,   1,   0,   0,   0,   0 } },
 { -1.348769535758e-05, -5.855471334202e-06, {   2,   0,  -1,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  1.510224833237e-05,  4.238545316764e-06, {   2,   0,   1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  6.549575051364e-06,  1.278441798852e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -5.936584168628e-07,  1.873654499198e-05, {   2,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.301044490641e-05,  6.310114150165e-06, {   4,   0,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -7.384617922637e-06, -1.193090266081e-05, {   1,  -1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.103888897793e-06,  1.620105027976e-05, {   3,  -1,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.302308026407e-05,  6.271799962910e-06, {   1,   1,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  9.179162536345e-06,  1.011107669426e-05, {   0,   0,   0,   0,   0,   0,   8, -14,   3,   0,   0,   0,   0 } },
 { -1.579934589606e-05, -3.488754715929e-06, {   1,  -1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.086404471036e-05, -8.416145422386e-06, {   6,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.112351255893e-05,  8.151565213437e-06, {   1,  -1,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.241609438964e-05,  6.856175035634e-06, {   1,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.751159336246e-05,  1.753972169035e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  1.924663177804e-05,  1.284918598535e-08, {   2,   0,  -5,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.324556616322e-05, -6.009439448739e-06, {   2,   0,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.919950154825e-05, -5.077894100977e-08, {   2,   0,  -2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.485080135049e-05,  4.377347545660e-06, {   1,  -1,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.838517730322e-06, -1.638941728487e-05, {   0,   0,   0,   0,   0,  10,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.162800685137e-05, -7.599044958046e-06, {   0,   0,   1,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.920483702212e-05,  1.152486068540e-08, {   0,   0,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -8.678257520714e-06, -1.053716572207e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,  -2,   0,   0,   0 } },
 {  1.905721439615e-05, -1.580797174517e-07, {   0,   2,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.024761477909e-06,  1.517544300814e-05, {   4,   0,  -3,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.134078658601e-05, -7.858168388029e-06, {   2,  -2,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  8.744680931758e-06, -1.043224447704e-05, {   1,  -1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.790690073195e-07, -1.899192094817e-05, {   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.687891133035e-06,  1.547365891776e-05, {   0,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.748405241560e-05,  1.665760066633e-06, {   0,   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.281775824282e-05, -6.329515876044e-06, {   1,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.868121656030e-05,  4.601122749723e-07, {   2,   0,   0,   0,   0, -40,  40,   0,   0,   0,   0,   0,   0 } },
 { -1.763440876447e-05, -1.499544920313e-06, {   0,   0,   0,   0,   0,   0,   2,  -3,   1,   0,   0,   0,   0 } },
 { -1.641640062601e-05, -2.711547286618e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -6,   3,   0,   0 } },
 {  4.360333342203e-06, -1.476109736595e-05, {   2,   0,   1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.035737714955e-05, -8.763696886023e-06, {   2,   0,   0,   0,   0,  -3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.347125529305e-06,  1.377394227171e-05, {   2,   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  5.436384445408e-06, -1.367256105556e-05, {   1,   1,   0,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.545683052600e-05, -3.647220218892e-06, {   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.254940170619e-06,  1.384820445325e-05, {   1,   1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -3.635838027776e-06, -1.546721257423e-05, {   1,  -1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.317414747717e-05,  5.925326985232e-06, {   2,  -2,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.756045584405e-05,  1.531325777906e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -2.967317770575e-06, -1.612179991548e-05, {   2,   0,   0,   0,  -7,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.890069577280e-05,  1.784450975646e-07, {   2,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.249990935219e-05, -6.578792644093e-06, {   2,   0,  -2,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.214971314699e-05, -6.927658618592e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -4,   2,   0 } },
 {  1.902605709390e-05,  4.707339505067e-08, {   2,   0,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.747154715118e-05,  1.589163447486e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0 } },
 { -1.094915276666e-05,  8.109040398626e-06, {   2,   0,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  6.479125285987e-06, -1.257678097304e-05, {   2,  -2,  -1,   0,   0,  23, -30,  15,   0,   0,   0,   0,   0 } },
 { -1.267064750890e-05,  6.384855842751e-06, {   0,   0,   2,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.899277395397e-07,  1.876456073019e-05, {   1,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.849346674995e-05, -5.562728496741e-07, {   0,   0,   1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.844461708538e-06,  1.520360697191e-05, {   0,   2,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.395628529259e-06,  1.265197658019e-05, {   1,   0,   0,   0,   0, -18,  17,   0,  -2,   5,   0,   0,   0 } },
 {  1.507546839072e-05, -3.961386052320e-06, {   2,   2,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  1.507546839072e-05, -3.961386052320e-06, {   2,  -2,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  1.902749882627e-05,  1.937016444900e-09, {   2,   2,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.620068719466e-07, -1.846271587789e-05, {   0,   0,   3,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.210994201874e-05, -6.905646386603e-06, {   1,  -1,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -5.279982150006e-06,  1.372820340783e-05, {   0,   0,   3,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.694982474543e-06,  1.430663144237e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.301550446770e-05,  5.972472820803e-06, {   1,  -1,   1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.272606856392e-05,  6.260540340850e-06, {   2,   0,  -1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.110926893501e-06, -1.187552254808e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 {  2.948551729540e-07,  1.868373619725e-05, {   1,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.895717061768e-05,  1.812220664192e-08, {   3,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.986527319195e-06, -8.985874175514e-06, {   6,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -6.866521150062e-07,  1.828303920864e-05, {   0,   0,   1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  1.560750557481e-05,  3.358066599158e-06, {   1,   0,   0,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  5.316935645437e-06, -1.364781399699e-05, {   1,  -1,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.155110151892e-05, -7.410709994804e-06, {   0,   2,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.154756930225e-05, -7.407437238283e-06, {   0,   2,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.346005914997e-05,  5.489437428031e-06, {   2,   0,  -1,   0,   0,  -4,   7,   0,  -6,   0,   0,   0,   0 } },
 { -1.271563933805e-05,  6.233145387606e-06, {   2,   0,  -1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.381045036493e-06,  1.656470723365e-05, {   2,   0,   0,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -8.507420066979e-06, -1.043032400465e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.741533517750e-05,  1.510561438386e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  3.234670990765e-06,  1.567808962648e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 { -6.795827414420e-06, -1.211307277593e-05, {   1,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.624926346742e-06, -1.527997385184e-05, {   2,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.460303700338e-05, -4.300812287807e-06, {   2,   2,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.140507924117e-05,  7.493247631300e-06, {   1,  -1,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  5.934169476703e-07,  1.829787082081e-05, {   0,   2,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.877893545742e-05,  1.117182021921e-07, {   2,   0,   1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.604717890205e-05, -2.836691077967e-06, {   0,   0,   1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.730228757980e-05,  1.580781656600e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.726089314208e-05,  1.616246126920e-06, {   1,   0,  -2,   0,   0,   0,  37, -43,   0,   0,   0,   0,   0 } },
 { -1.503483559539e-05,  3.841286269645e-06, {   1,  -1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.562707447901e-06, -1.230981849572e-05, {   1,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.509354706743e-05,  3.774405515732e-06, {   1,  -1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.079420035987e-05,  8.070466461940e-06, {   2,   0,  -2,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  3.844198284347e-06, -1.501819285931e-05, {   2,   2,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.492993886765e-05,  3.924066202627e-06, {   1,   0,   0,   0,   0, -26,  31,  -2,   0,   0,   0,   0,   0 } },
 {  1.692182025629e-05, -1.926176426720e-06, {   2,  -2,  -1,   0,   0,  26, -27,   0,   1,   0,   0,   0,   0 } },
 {  1.858844964163e-05,  2.576257935132e-07, {   0,   0,   1,   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0 } },
 {  1.523833003718e-05,  3.601561565909e-06, {   2,   0,  -3,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.432224722580e-05, -4.511561324023e-06, {   5,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.669129439439e-06, -1.516331282405e-05, {   3,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.786527111457e-05,  9.643973527262e-07, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -8.206376018696e-07,  1.800721001257e-05, {   2,   0,   1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.464038792084e-06, -1.736236080372e-05, {   2,  -2,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.234341792806e-06,  1.359183933472e-05, {   0,   0,   3,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.687291983714e-05, -1.950030263497e-06, {   2,   0,  -1,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 { -1.643865238304e-05,  2.380846508059e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 {  8.263859951780e-06,  1.055293501645e-05, {   2,   0,  -1,   0,   0,   0, -10,  18,   0,   0,   0,   0,   0 } },
 {  1.824058112011e-05,  5.733751354853e-07, {   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.023593265051e-06,  1.478692679047e-05, {   0,   0,   0,   0,   0,  25, -28,   0,   0,   0,   0,   0,   0 } },
 {  1.140399274836e-07, -1.869637793229e-05, {   0,   2,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.492124727423e-06,  9.316056784892e-06, {   4,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  1.215821850118e-05, -6.647570144321e-06, {   4,   0,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  2.969938944148e-06,  1.582685201409e-05, {   0,   2,  -3,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.356508375614e-05,  5.229339500117e-06, {   2,   0,  -1,   0,   0,  12, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.036661895720e-06, -1.175731879584e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -9.140660334102e-06,  9.652119503673e-06, {   2,   0,   0,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 { -9.140660334102e-06,  9.652119503673e-06, {   2,   0,  -2,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 { -1.599450411791e-05, -2.790525955944e-06, {   3,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.868793957817e-05, -9.398685861543e-08, {   2,   2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.994703151213e-06, -1.177694044693e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  6.681489916684e-06,  1.208349833738e-05, {   2,   0,  -1,   0,   0,   6, -12,   8,  -3,   0,   0,   0,   0 } },
 {  7.418325131511e-06,  1.133798947713e-05, {   0,   2,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.596113396761e-06,  1.515751908392e-05, {   2,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.822726038380e-06, -1.192482869672e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 {  7.995533314253e-06, -1.074329260046e-05, {   2,   0,  -2,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -6.566801728474e-06, -1.216708512316e-05, {   2,  -2,   0,   0,   0,   5,  -3,  -6,   2,   0,   0,   0,   0 } },
 { -1.331232801844e-05,  5.421261548913e-06, {   2,   0,  -1,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  3.370025472385e-06,  1.536283079296e-05, {   2,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -8.726632481644e-06, -9.995346615473e-06, {   0,   0,   0,   0,   0,   8,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.392287580760e-06, -1.532832803535e-05, {   3,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.358712544200e-06,  1.435802917405e-05, {   2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.001864785101e-06, -1.170714801863e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  3.019616608798e-06,  1.568070489328e-05, {   5,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.852540472471e-05, -1.734452210874e-07, {   2,  -2,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -3.886936290934e-06, -1.481169968406e-05, {   1,   1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.859031724350e-05,  1.017912471713e-07, {   0,   2,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.467372139598e-06,  1.522323529826e-05, {   3,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.052343664190e-06, -1.463329703913e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.452608538613e-05, -4.156970753178e-06, {   0,   0,   3,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.300429801341e-06, -1.638059932966e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.833929093337e-05,  3.383394419415e-07, {   0,   0,   1,   0,   0,  -3,   3,   0,   1,   0,   0,   0,   0 } },
 { -1.194424718408e-05,  6.730886590473e-06, {   1,   0,  -1,   0,   0,   0,   4,  -6,   1,   0,   0,   0,   0 } },
 {  7.056971500336e-06,  1.161717330010e-05, {   2,   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.387455852631e-05, -4.795135028008e-06, {   2,   0,  -1,   0,   0,   0,   4,   0,  -3,  -5,   0,   0,   0 } },
 { -1.375786312370e-05,  4.910394535655e-06, {   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.505801398800e-05, -3.599128620415e-06, {   0,   0,   2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.960967418363e-06, -1.168860985900e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.827052937105e-05, -3.750601906991e-07, {   0,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.564214790054e-05,  3.000179427254e-06, {   2,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.863556250900e-05, -3.461930721287e-09, {   0,   0,   4,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.236680011614e-06, -1.240095932845e-05, {   4,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  1.191331737890e-05, -6.712458053323e-06, {   2,   0,  -3,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.063269657814e-05,  7.991060139711e-06, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  8.475167009977e-06,  1.014772308627e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -6.594252472276e-06,  1.202651725588e-05, {   1,  -1,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.008734104840e-06, -1.261051427057e-05, {   1,  -1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  2.588699651482e-06,  1.602702216181e-05, {   4,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.149913425165e-05,  7.113908653435e-06, {   0,   0,   1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 {  9.844368884600e-06, -8.766011740929e-06, {   2,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.843078715903e-06,  1.076610400887e-05, {   2,   0,  -1,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 { -4.021795187879e-06,  1.457793653918e-05, {   1,   1,  -1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  9.490501000019e-06, -9.094943969277e-06, {   2,   0,   0,   0,   0,  -6,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.453162023565e-05, -4.053555750470e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -3,   0,   0,   0 } },
 { -1.193119249867e-05,  6.653614667873e-06, {   2,   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.353161946928e-05,  5.052504635744e-06, {   2,   0,  -1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 {  9.329149544437e-06,  9.254626019977e-06, {   4,   0,  -3,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -7.554727915194e-06,  1.102627045952e-05, {   0,   0,   1,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 { -4.316950153833e-06,  1.426109552209e-05, {   2,   0,  -1,   0,   0, -22,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.126381467114e-05, -7.309179224347e-06, {   2,   0,  -1,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.108343925103e-05,  7.489232042374e-06, {   2,   0,  -1,   0,   0,   0,   4,   0,  -7,   5,   0,   0,   0 } },
 {  5.541691263950e-06,  1.302282327135e-05, {   0,   0,   0,   0,   0,  13, -19,  -4,   0,   0,   0,   0,   0 } },
 { -1.219197649811e-05, -6.371496890652e-06, {   2,   0,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.447445310220e-06,  1.611248076039e-05, {   4,   0,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  1.158587168723e-06, -1.739968583937e-05, {   2,  -2,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.538851007119e-05,  3.169263562587e-06, {   1,   1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.334182864240e-06, -1.622050494369e-05, {   6,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.581012296881e-05, -2.743942174531e-06, {   3,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.140421826713e-06, -1.240271813904e-05, {   4,   0,  -2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.504242065498e-05, -3.493984614950e-06, {   2,   0,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.085391081717e-05,  7.680038303589e-06, {   0,   0,   2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -1.035829030344e-05,  8.175365983708e-06, {   1,  -1,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  1.848410527167e-05,  4.618136144296e-08, {   0,   2,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  6.628067943982e-06, -1.190190899734e-05, {   1,  -1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -3.531449620272e-06, -1.499230439236e-05, {   1,  -1,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.402347707704e-06,  1.411758362908e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.602768491053e-06, -1.491497323081e-05, {   3,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.002941721991e-06, -1.551271209513e-05, {   1,   0,   0,   0,   0,   0, -28,  29,   3,   0,   0,   0,   0 } },
 { -6.905945644322e-06,  1.160806623475e-05, {   0,   0,   1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.180773506571e-05, -6.705780569576e-06, {   2,   0,  -1,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.368813227304e-05, -4.824662503927e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  1.411105679216e-05, -4.396073688212e-06, {   6,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.939414092004e-06, -1.156545924339e-05, {   2,   0,  -2,   0,   0,   4,  -5,   0,   3,   0,   0,   0,   0 } },
 {  1.503648334219e-05,  3.459755611494e-06, {   1,   1,  -1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.825422461817e-05,  2.418496141125e-07, {   2,   0,   0,   0,   0, -39,  39,   0,   0,   0,   0,   0,   0 } },
 { -1.073411793935e-05, -7.752556006685e-06, {   4,   0,  -3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.892852834363e-06, -1.459023037482e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  2.069468303072e-06, -1.641168072461e-05, {   2,   0,   0,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 { -1.520836080400e-05, -3.272148466224e-06, {   1,  -1,   0,   0,   0,  -3,   6,   0,  -1,   0,   0,   0,   0 } },
 { -4.409581649347e-06, -1.406361750713e-05, {   1,   0,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  9.618431131104e-06, -8.851108943882e-06, {   2,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.103852717651e-05,  7.426176629235e-06, {   2,   0,  -2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -1.809088178093e-05,  3.533994078385e-07, {   1,   1,   1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.809088178093e-05,  3.533994078385e-07, {   1,   1,  -3,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.262048783031e-05,  5.809011015611e-06, {   2,   0,  -1,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 {  5.329831178776e-07,  1.789087284045e-05, {   1,   0,  -1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -3.940323748068e-06,  1.447857100172e-05, {   0,   0,   1,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 { -1.650120630066e-05,  1.912295952349e-06, {   2,   0,  -2,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -1.175657124098e-05, -6.646398719988e-06, {   4,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.232565844362e-05,  6.074925819277e-06, {   0,   0,   2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.006722721859e-05, -8.333293191722e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -8,   0,   0,   0 } },
 {  1.143560693217e-06,  1.725577860089e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 {  7.452628342907e-06,  1.094194483567e-05, {   0,   0,   1,   0,   0,   0,  -3,   6,  -1,   0,   0,   0,   0 } },
 {  7.129568930960e-07, -1.768125226175e-05, {   2,  -2,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.554118554199e-05, -2.852911196911e-06, {   2,   0,  -2,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -9.644918246953e-07, -1.742385847471e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  1.761225018247e-05,  7.750548152128e-07, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.834423762807e-05,  3.871440115573e-08, {   0,   0,   2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.660462805267e-06, -1.172241807164e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -5.188349422284e-06,  1.319236782923e-05, {   1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.481827267947e-06, -1.489880222955e-05, {   1,  -1,   1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.390478576334e-05, -4.470563707746e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  4.132433583813e-06,  1.423550257396e-05, {   3,   0,  -2,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.663139295305e-06,  1.470364141236e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -5,   9,   0,   0,   0 } },
 { -1.125008618685e-06,  1.723412609408e-05, {   2,  -2,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  2.453696319538e-06,  1.589182373152e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  1.342842970073e-05,  4.909874843497e-06, {   4,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  6.943175168890e-06, -1.139338902322e-05, {   0,   0,   3,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.234671937012e-06,  1.410182205539e-05, {   2,   0,   0,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 {  6.542292625093e-06, -1.178689373039e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 {  4.874311179831e-06,  1.344654066798e-05, {   4,   0,  -2,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.359289690460e-05,  4.725155526755e-06, {   2,   0,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.183680874175e-05, -6.475192045442e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,  -5,   0,   0,   0 } },
 {  1.089029532147e-05, -7.418695502238e-06, {   2,   0,   1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -9.697941677485e-06,  8.608300257439e-06, {   2,   2,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.769169919991e-05,  6.120858271321e-07, {   4,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.693656690095e-06, -1.460785678607e-05, {   3,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.724297430570e-06,  1.157637253676e-05, {   1,   1,  -1,   0,   0, -20,  16,   8,  -3,   0,   0,   0,   0 } },
 { -1.138810445591e-05, -6.907228741009e-06, {   0,   2,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -3.500875342339e-06,  1.478915926519e-05, {   3,  -1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.826699524640e-05,  1.522368676076e-08, {   2,   0,   2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.228821019212e-07,  1.755428157090e-05, {   2,   0,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  2.768537446062e-06, -1.550452655457e-05, {   2,   0,   0,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 {  1.355696808320e-05, -4.702326159077e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 {  1.068235685738e-05, -7.564751794020e-06, {   1,  -1,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.821482207574e-05, -3.081514237744e-08, {   0,   2,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.613522515905e-05, -2.105978824342e-06, {   3,   1,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.530422746951e-05,  2.930694847947e-06, {   1,  -1,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.424942975842e-06, -1.180790506912e-05, {   3,  -1,  -2,   0,   0,   0,  14, -21,   0,   0,   0,   0,   0 } },
 {  1.072922906077e-05, -7.501270953875e-06, {   4,   0,  -2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  5.863313227482e-06,  1.236700439674e-05, {   4,  -2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.731557494193e-05, -9.123689104229e-07, {   2,   0,   0,   0,   0,   0,   7,   0,  -7,   0,   0,   0,   0 } },
 {  1.638688063257e-05, -1.837887435629e-06, {   1,  -1,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -9.112352773782e-06,  9.103786929977e-06, {   2,   0,   0,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0 } },
 {  1.159229271929e-05,  6.615604800254e-06, {   1,  -1,  -1,   0,   0,  21, -20,   0,   2,   0,   0,   0,   0 } },
 { -5.119974778365e-06, -1.308464477967e-05, {   0,   0,   1,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.817412407759e-06,  1.438660424758e-05, {   2,   0,   3,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.654445651803e-05, -1.653442457380e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  2.579973139721e-06,  1.561199315818e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 {  7.766924422234e-06,  1.042490850224e-05, {   2,   0,   0,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 { -3.495297921535e-06,  1.469452968868e-05, {   0,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.003477462918e-06,  1.318436196107e-05, {   1,  -1,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.105325652347e-06,  1.507088417968e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 { -8.120406071890e-06, -1.004774704158e-05, {   1,   1,  -1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.279106200203e-05, -5.375762898077e-06, {   1,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  7.155339833694e-06, -1.101077622567e-05, {   1,  -1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.048627546879e-05,  7.676139051116e-06, {   1,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -3.452299902937e-06, -1.470502010763e-05, {   1,  -1,   1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.522918478082e-05,  2.924881174215e-06, {   1,   1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.170862930728e-05, -6.444254094859e-06, {   4,  -2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.318577516839e-05, -4.958246828973e-06, {   1,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.226368911355e-05, -5.873480161462e-06, {   3,   1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.182582617083e-05,  6.308001626262e-06, {   2,  -2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.811851503081e-05,  1.502714258973e-08, {   4,   0,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.537464141805e-06,  1.658483115213e-05, {   4,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.521598068820e-05, -2.905839885091e-06, {   3,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.270467904579e-06, -1.684957430921e-05, {   0,   0,   1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.807988617258e-05,  3.929454481248e-08, {   4,   0,  -3,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.496544191074e-05, -3.148655400457e-06, {   1,  -1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.476196806707e-05,  3.350056971399e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 { -9.188594383133e-06,  8.921213435872e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 {  4.115132795251e-06, -1.399176941316e-05, {   6,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.114720777398e-06, -1.399183884631e-05, {   6,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.169316981625e-05, -6.413382737220e-06, {   4,   0,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  1.806240408064e-05, -3.762367190502e-08, {   2,   0,  -2,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.354198507922e-06,  1.573653831456e-05, {   2,   0,  -4,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -3.885322410659e-06,  1.419973303252e-05, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   3,   0,   0,   0 } },
 {  6.647738853434e-06, -1.143662668147e-05, {   2,   0,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0 } },
 {  1.280343956419e-06, -1.679506166273e-05, {   0,   0,   1,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 {  3.457454277754e-06, -1.461290439701e-05, {   3,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.706138532118e-05, -1.007982971460e-06, {   4,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.124674832787e-05, -6.817389759208e-06, {   0,   2,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -5.203053064068e-07, -1.753155419513e-05, {   2,   0,   2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.017186655800e-07,  1.755006586550e-05, {   2,  -2,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.726019673295e-06,  1.232453310683e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   1,   0,   0,   0 } },
 { -1.802532188211e-05, -1.665453594212e-08, {   2,   0,  -1,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 { -3.614595876491e-06, -1.442502008348e-05, {   0,   0,   0,   0,   0,   2, -10,  13,  -2,   0,   0,   0,   0 } },
 {  1.787015348831e-05, -1.658759579631e-07, {   0,   2,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.483564937964e-05, -3.195020460097e-06, {   2,   0,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  4.383315821853e-07, -1.758939091328e-05, {   0,   0,   2,   0,   0, -18,   8,  16,  -4,  -5,   0,   0,   0 } },
 {  6.653485599416e-06, -1.137286627964e-05, {   1,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -7.701219296750e-06,  1.032454836459e-05, {   2,   0,   0,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 {  1.133539765462e-05, -6.686662811041e-06, {   2,   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.714516638628e-05, -8.765687541713e-07, {   2,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 { -8.066419764766e-06, -9.952713392822e-06, {   1,   1,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.472649039808e-06, -1.454637255745e-05, {   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.999905489048e-06, -8.013141082263e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,  11,  -3,   0,   0 } },
 { -3.539643158829e-06,  1.447244103452e-05, {   2,   0,  -2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  3.520216838702e-06,  1.448645353869e-05, {   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.147978137482e-06,  1.285785336978e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -8.089029244907e-06,  9.915248831276e-06, {   2,  -1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  3.157576313504e-07,  1.768360833308e-05, {   2,   0,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.055495365242e-06,  9.939638633047e-06, {   2,   2,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -8.546771995136e-06, -9.441842144557e-06, {   0,   0,   0,   0,   1,   0, -11,  11,   0,   0,   0,   0,   0 } },
 {  1.534247568743e-05,  2.645337463573e-06, {   2,   0,   0,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 {  1.231177176059e-05,  5.668453985532e-06, {   0,   0,   0,   0,   1,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 { -1.051784477804e-05, -7.459298927915e-06, {   2,   0,   0,   0,   0,   0,  16, -21,   0,   0,   0,   0,   0 } },
 { -4.084075084745e-06,  1.389290895269e-05, {   3,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.216346979502e-05,  5.812080299533e-06, {   3,  -1,  -2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  4.080677436675e-06,  1.388612951041e-05, {   3,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.190084906387e-06, -1.575742293469e-05, {   0,   0,   1,   0,   0,  -9,  12,   0,   0,   0,   0,   0,   0 } },
 {  8.131729058022e-06, -9.814464199010e-06, {   2,   2,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.750605533185e-05, -4.384840463303e-07, {   1,  -1,  -1,   0,   0,   0,  34, -37,   0,   0,   0,   0,   0 } },
 { -9.731638077245e-07,  1.696986808345e-05, {   2,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -9.330576858775e-06,  8.610186078117e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -1.234694560945e-05, -5.590774071928e-06, {   3,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.200583305036e-05, -5.918885476397e-06, {   2,   2,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.516992902715e-05,  2.750509630201e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -1.046138626274e-05,  7.458482110949e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,   1,  -1,   0,   0,   0 } },
 {  1.761552724601e-05,  3.018419841876e-07, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   1,   0,   0,   0 } },
 { -1.012294945162e-05, -7.792796058075e-06, {   0,   0,   1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  6.424073329025e-07, -1.726869613234e-05, {   2,   0,  -2,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.329203206696e-05,  4.617446372590e-06, {   0,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -2.112871741506e-06, -1.579387260936e-05, {   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.394956655304e-05, -3.957080700277e-06, {   1,   1,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  4.892020567178e-07,  1.741351022729e-05, {   2,   0,  -2,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  5.186182678675e-07,  1.738324464606e-05, {   2,   0,   0,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 {  1.789604668116e-05, -2.991838211245e-09, {   0,   0,   2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.499399972541e-05, -2.904050761626e-06, {   1,  -1,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.146689430008e-05,  6.423510279420e-06, {   2,   0,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.593809019287e-05, -1.937720710179e-06, {   0,   0,   1,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  6.702234202555e-06,  1.117306695102e-05, {   3,   1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  3.568249848824e-06, -1.429924427885e-05, {   1,  -1,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.060673952162e-05,  7.259861988484e-06, {   2,   0,  -1,   0,   0,   0,  18, -23,   0,   0,   0,   0,   0 } },
 { -1.685890856370e-05,  1.004930578231e-06, {   0,   0,   0,   0,   0,   3,  -6,   0,   1,   0,   0,   0,   0 } },
 {  1.725625196018e-05, -6.045312810976e-07, {   1,   0,   0,   0,   0,   0, -35,  43,  -2,   0,   0,   0,   0 } },
 {  3.363525669210e-06,  1.449504569368e-05, {   1,  -1,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  5.964110579245e-06,  1.189178620479e-05, {   0,   0,   1,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 {  4.046370371074e-06,  1.380834486067e-05, {   0,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.408654856561e-05, -3.766120288463e-06, {   2,  -1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -5.880834917466e-06,  1.195692984721e-05, {   2,   0,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  3.724685827484e-06,  1.410341108162e-05, {   3,  -1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.603409710878e-06, -1.521810727900e-05, {   2,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.762758991906e-05,  1.830552132287e-07, {   0,   0,   4,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.637644690448e-06,  1.016044930587e-05, {   2,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.572806617622e-06, -1.022297470707e-05, {   0,   0,   4,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.938299379184e-06,  1.284812158048e-05, {   0,   0,   0,   0,   0,   0,  12, -22,   0,   0,   0,   0,   0 } },
 {  1.775481279264e-05,  2.880336059905e-08, {   0,   0,   2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.011133408757e-06,  1.376814134265e-05, {   0,   0,   1,   0,  -5,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.565849387207e-06, -1.021176894947e-05, {   0,   0,   4,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.712431013563e-05,  6.529466268778e-07, {   6,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.401765166007e-06, -1.437495708691e-05, {   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.766859488522e-05, -1.080275783906e-07, {   0,   2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.206784251242e-07, -1.684881033532e-05, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -5.961068052189e-07, -1.716853807502e-05, {   2,  -2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.246189796066e-05, -5.295046384654e-06, {   4,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.246189796066e-05, -5.295046384654e-06, {   4,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.772653810073e-05,  2.514975069102e-08, {   1,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.711438436472e-05,  6.331827182231e-07, {   4,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.738766910011e-05, -3.549287760842e-07, {   6,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.687541565385e-05, -8.634427707834e-07, {   2,   0,  -1,   0,   0,   4,  -6,   2,   0,   0,   0,   0,   0 } },
 { -7.828110001524e-06,  9.907103216134e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 { -1.742503159218e-05, -3.099544069489e-07, {   1,   0,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -3.987134684306e-06,  1.374655294432e-05, {   2,   0,  -1,   0,   0, -24,  27,   0,   0,   0,   0,   0,   0 } },
 {  7.684312969529e-07, -1.696223763903e-05, {   2,   0,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0 } },
 { -8.658874736506e-06, -9.071766096738e-06, {   2,   0,  -1,   0,   0, -21,  20,   5,   0,   0,   0,   0,   0 } },
 {  1.771013789213e-05,  1.157300821006e-08, {   6,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.715165796532e-06,  1.400603243644e-05, {   4,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  5.519797338016e-06,  1.220066143398e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 {  5.519797338016e-06,  1.220066143398e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 {  1.141740670063e-05, -6.293311493119e-06, {   2,   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.769893375973e-05,  1.114309457758e-08, {   2,   0,  -4,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.140208938367e-05,  6.302016927419e-06, {   1,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.448785609921e-07, -1.704864502349e-05, {   2,   0,   1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.200405225457e-05, -5.687498619322e-06, {   4,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.375836823508e-05, -3.926603611798e-06, {   0,   0,   1,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.746544645104e-05,  2.164772143224e-07, {   4,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.451389423380e-05, -3.160437040036e-06, {   0,   0,   1,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.150985386360e-06, -1.551297425388e-05, {   2,   0,  -2,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.371905079299e-06, -1.429163849214e-05, {   0,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.308057670602e-05,  4.578060958467e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 { -1.628234371896e-05,  1.376116617887e-06, {   2,   0,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.051976695299e-06, -1.160220902417e-05, {   1,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -9.346527564523e-06,  8.304342796255e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 { -5.189678792492e-06, -1.245032155265e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.294140669565e-05,  4.696336336488e-06, {   2,   0,  -1,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 {  6.679926348789e-06, -1.094319458834e-05, {   0,   0,   0,   0,   0,   0,   6, -10,   2,   0,   0,   0,   0 } },
 {  4.597467636297e-06,  1.300985258866e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,  -3,   0,   0,   1,   0 } },
 { -7.156948554216e-06, -1.044921965345e-05, {   0,   0,   0,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 {  1.581403180375e-05,  1.790332767348e-06, {   0,   2,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.231490504230e-05, -5.281902408242e-06, {   2,   0,  -1,   0,   0,   0,  15, -22,   0,   0,   0,   0,   0 } },
 { -1.665382460539e-05,  9.424427042824e-07, {   2,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -4.820884399943e-06, -1.276326211192e-05, {   2,   0,  -2,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.209307161383e-05, -5.486551269082e-06, {   1,   0,   0,   0,   0, -15,   8,   8,  -3,   0,   0,   0,   0 } },
 {  1.750407862456e-05,  6.725941762123e-08, {   2,   0,   1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  1.544908292294e-05,  2.119833297472e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,  -2,   0,   0 } },
 { -1.408470554937e-05,  3.482988182575e-06, {   1,  -1,  -1,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 {  3.825335138936e-06, -1.373903198568e-05, {   4,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -3.825037693786e-06, -1.373796560807e-05, {   4,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.602240016921e-05,  1.528941622689e-06, {   0,   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -3.380400002652e-06,  1.416968570170e-05, {   3,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.131247828965e-05,  6.233119465720e-06, {   1,   0,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -4.031000436094e-06,  1.351411769127e-05, {   4,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.355142445420e-06,  1.418956072409e-05, {   1,  -1,   2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.652990454720e-06,  1.388847447316e-05, {   2,   0,  -1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.626669479976e-05, -1.272723531380e-06, {   0,   2,  -1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  3.708383935491e-06,  1.382967807023e-05, {   1,  -1,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.844168454511e-06,  1.269212975455e-05, {   1,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.387348577344e-05, -3.662363099654e-06, {   2,  -1,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -7.670926538479e-06, -9.860153100224e-06, {   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  8.008457950744e-06, -9.522370047030e-06, {   1,  -1,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.026907145648e-06,  1.350123608810e-05, {   4,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.675810845455e-05, -7.671914153028e-07, {   4,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -8.042581968550e-06,  9.476069756295e-06, {   6,   0,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.578054454837e-05, -1.736519704931e-06, {   4,   0,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.386584518370e-06,  1.512787066778e-05, {   0,   2,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.942017422855e-06,  8.571847053735e-06, {   2,   2,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.431930448678e-05, -3.192990946879e-06, {   1,  -1,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.408341305335e-06, -9.099454932018e-06, {   3,  -1,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.384358521840e-06,  1.011566366126e-05, {   1,  -1,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.982382268397e-06,  7.514274733328e-06, {   1,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  3.374543912374e-06, -1.410559082713e-05, {   3,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.790555829105e-06,  1.268856828865e-05, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.342222571856e-06,  1.413194498458e-05, {   1,  -1,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.453760527178e-06,  1.401619450893e-05, {   1,   1,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.869477629536e-06, -1.059417847250e-05, {   0,   0,   4,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.971036443864e-06, -7.487376238205e-06, {   1,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  3.819947695283e-06, -1.363817869232e-05, {   1,  -1,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.756608733780e-06,  1.369758497065e-05, {   4,   0,  -2,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.744630773109e-05, -6.618554339575e-09, {   2,   0,   3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.474169150408e-05, -2.705734369820e-06, {   2,   0,  -2,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -4.635808927645e-06, -1.280862196761e-05, {   4,   0,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -1.695312875651e-05,  4.880757661004e-07, {   2,   0,   0,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 { -5.926583658910e-06, -1.151299161482e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 {  1.035676746603e-05,  7.080920355460e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   3,   0,   0,   0 } },
 { -1.396683013776e-05, -3.470522812938e-06, {   2,   0,  -1,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.084691760311e-05,  6.588525081542e-06, {   2,   0,  -1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0 } },
 { -1.318930458919e-05,  4.245783438234e-06, {   0,   0,   1,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 { -1.735926717786e-05,  6.384018268813e-08, {   2,  -2,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  5.307865279616e-07, -1.689052358146e-05, {   2,   0,  -3,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.253511855964e-05, -4.886034627310e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -8,   0,   0,   0 } },
 {  3.467427441546e-06, -1.395111559438e-05, {   3,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.740322022520e-05, -1.276686871524e-08, {   1,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.620433818492e-07,  1.694753237162e-05, {   2,   0,   0,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.136296161913e-05,  6.045234881269e-06, {   3,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.333915365880e-05,  4.061721758362e-06, {   2,   0,  -1,   0,   0,   0, -12,  21,  -2,   0,   0,   0,   0 } },
 { -9.405818043831e-06,  7.994222291585e-06, {   0,   0,   1,   0,   0, -21,  22,  -2,   0,   0,   0,   0,   0 } },
 { -9.722923329950e-06, -7.673296362551e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   2, -11,   3,   0,   0 } },
 {  7.171849816038e-06, -1.021900741709e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -5,   0,   0,   0 } },
 { -1.424978870244e-05, -3.131932930667e-06, {   2,   0,  -1,   0,   0, -17,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.668027875061e-06, -1.171117727962e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 {  1.724525827042e-05, -1.292791540578e-07, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 { -1.724496051024e-05, -1.292825590144e-07, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 {  3.888265615854e-06,  1.348549472312e-05, {   0,   0,   1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.188519362291e-06, -1.317298841898e-05, {   0,   0,   2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  1.165857824148e-05, -5.698227977481e-06, {   2,  -2,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.088793595360e-06, -1.026151965769e-05, {   1,  -2,   1,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 { -7.088793595360e-06, -1.026151965769e-05, {   1,  -2,  -1,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 { -7.744212276969e-06,  9.601403029720e-06, {   0,   2,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.275143797671e-05, -4.593944856108e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 {  1.679128668577e-05,  5.445988031444e-07, {   2,   0,   1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  9.477737228411e-06, -7.854186472200e-06, {   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,   0 } },
 {  4.593414219728e-06,  1.273486465394e-05, {   2,   0,   2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.672925925687e-06, -1.165363937510e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   4,  -2,   0,   0,   0 } },
 { -6.268181789681e-06, -1.105796120247e-05, {   4,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -6.410498726414e-06,  1.091355599606e-05, {   4,   0,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  3.409820579528e-06, -1.391370844493e-05, {   4,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.320494738988e-07,  1.718724645384e-05, {   0,   0,   1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.473606666454e-06, -1.081893127403e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -6.473606666454e-06, -1.081893127403e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -5.080967301170e-07, -1.678105895340e-05, {   0,   0,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.294375897653e-07,  1.685660683039e-05, {   0,   2,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.795919879031e-06,  9.485519579539e-06, {   6,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.605054140649e-06,  1.067555575955e-05, {   2,   0,  -2,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  1.283392164653e-05, -4.444692726834e-06, {   1,  -1,  -1,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 { -1.269978359788e-05, -4.578818152082e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 { -1.455793704851e-05,  2.716999571127e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 {  1.569836104093e-05,  1.575255068836e-06, {   2,   0,  -1,   0,   0,   0,  16, -22,   0,   0,   0,   0,   0 } },
 {  6.957039154056e-06,  1.031568395609e-05, {   1,   0,   0,   0,   0, -18,  17,   0,   2,  -5,   0,   0,   0 } },
 {  5.951596931975e-06,  1.132067598540e-05, {   1,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  5.205919159007e-06, -1.206361787238e-05, {   2,   2,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.962748636395e-06,  1.030570266861e-05, {   1,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0,   0 } },
 {  7.772398496128e-06, -9.494761717743e-06, {   1,  -1,   0,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.476165406783e-07, -1.671626423702e-05, {   0,   2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.968888794696e-06, -1.128703957859e-05, {   1,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.368087061803e-05,  3.567095617393e-06, {   0,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.909435516535e-06, -1.533597558705e-05, {   2,   0,   0,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 {  4.541856968846e-06, -1.270346199480e-05, {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.513991791939e-06, -9.730770145911e-06, {   1,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.645486257817e-05, -7.733166046810e-07, {   2,   0,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  6.618056482666e-06,  1.060246810281e-05, {   1,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -4.779883739722e-06,  1.243368691322e-05, {   3,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -3.490921921466e-06, -1.372242082908e-05, {   1,   0,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.733415897977e-06, -1.347359202959e-05, {   2,   0,  -1,   0,   0, -20,  20,   2,   0,   0,   0,   0,   0 } },
 { -1.425479493035e-05, -2.952149329688e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  1.719335563239e-05,  7.392172377696e-09, {   3,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.297345181700e-06,  9.901299759170e-06, {   0,   0,   1,   0,   0,   0, -10,  18,   0,   0,   0,   0,   0 } },
 { -2.286102486905e-06, -1.490916641783e-05, {   2,  -2,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.715541324802e-05,  3.199730645119e-08, {   0,   0,   2,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.676319444427e-06, -1.350740955695e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -1.499952628581e-05, -2.175054063868e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   0,   0,   0 } },
 {  2.272296661278e-06, -1.489881290970e-05, {   2,  -2,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.699912695016e-05,  1.688566400621e-07, {   4,  -2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.252942016229e-06,  9.914115267564e-06, {   2,   0,  -2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.285247249184e-06, -1.387683602580e-05, {   3,   1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.063224788639e-05,  6.527125470477e-06, {   0,   0,   2,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.416916345639e-05, -2.989886724840e-06, {   4,   0,  -2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -1.908024655932e-07, -1.696197869441e-05, {   2,   0,  -3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.668750865790e-05, -4.643242154980e-07, {   6,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.552568244260e-07,  1.689400513024e-05, {   1,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.658319170047e-08, -1.706707314750e-05, {   0,   0,   0,   0,   0,   5,  -8,   0,   2,   0,   0,   0,   0 } },
 {  1.649902404683e-05,  6.445638813072e-07, {   1,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.021485748783e-05, -6.928450684666e-06, {   2,  -2,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  6.386140787909e-06, -1.074957753225e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -1.462299859575e-05, -2.511188572486e-06, {   4,  -2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.285819493135e-05,  4.275334354250e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -1.640880197960e-05, -7.191551236658e-07, {   3,   0,   0,   0,   0, -69,  75,   0,   0,   0,   0,   0,   0 } },
 {  1.357899035099e-05,  3.547809392698e-06, {   0,   0,   2,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.429586067104e-05, -2.828192337933e-06, {   1,   1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.558582833927e-05, -1.530704791227e-06, {   4,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -5.524003530284e-06, -1.158938389075e-05, {   0,   0,   1,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0 } },
 { -2.733781267253e-06,  1.437676038357e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 {  2.383527102825e-06, -1.472612539883e-05, {   0,   0,   1,   0,   0,   0, -12,  23,  -3,   0,   0,   0,   0 } },
 {  1.494261430596e-05, -2.166965994606e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 { -8.136304615244e-06, -8.973148648681e-06, {   2,   0,   0,   0,   0,  -3,   5,  -4,   0,   0,   0,   0,   0 } },
 {  6.333162429759e-06,  1.077462826525e-05, {   1,  -1,  -2,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.151609037388e-05, -5.587273638850e-06, {   1,  -1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.709747517091e-05,  4.940368634527e-09, {   1,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.101861877200e-06,  1.299765415829e-05, {   0,   1,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  3.710365663830e-06,  1.338174353817e-05, {   0,   0,   0,   0,   0,  17, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.434773281717e-05, -2.743287236763e-06, {   2,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.046016230487e-05, -6.625604333379e-06, {   4,   0,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.180925625340e-05, -5.271357876546e-06, {   2,   0,  -1,   0,   0,  -2,   6,  -5,   0,   0,   0,   0,   0 } },
 { -3.017425999873e-06, -1.406235781523e-05, {   4,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.505661477473e-07,  1.682678860547e-05, {   1,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.690573774873e-05, -1.649145290131e-07, {   1,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.363690372096e-05, -3.429365884366e-06, {   1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.692659152912e-05, -1.383891639213e-07, {   2,   0,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.696194495609e-05, -1.025404102741e-07, {   2,   0,  -1,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.704016354643e-05, -2.251594139307e-08, {   2,   0,   0,   0,   0,   0,   4,   0,   0,  -4,   0,   0,   0 } },
 { -4.697899766012e-06, -1.235394553021e-05, {   2,  -2,   0,   0,   0,   0,   2,   0,  -4,  13,   0,   0,   0 } },
 { -9.871920654953e-06,  7.171408764643e-06, {   2,   0,  -1,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0 } },
 { -1.197898044347e-05, -5.060706317223e-06, {   0,   0,   1,   0,   0,   0, -11,  18,   0,   0,   0,   0,   0 } },
 { -1.375501934690e-05,  3.284330084945e-06, {   1,  -1,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.339313637549e-05, -3.640537111650e-06, {   1,   1,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.982876625498e-07,  1.642830760895e-05, {   0,   2,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.112284822522e-06,  8.911853179364e-06, {   0,   0,   0,   0,   0,   1,  -2,   0,   4,   0,   0,   0,   0 } },
 { -5.347282926802e-06, -1.167304290731e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 {  1.688489582813e-05, -1.332895870929e-07, {   3,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.310856162580e-05, -3.909361041581e-06, {   4,   0,  -1,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 { -6.392970974575e-06, -1.061903022593e-05, {   1,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.632914887981e-05, -6.805221252677e-07, {   4,   2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.554081474418e-05, -1.466881906470e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.603062200852e-05,  9.729949699485e-07, {   2,   0,  -1,   0,   0,   0,   7,   0,  -7,   0,   0,   0,   0 } },
 { -1.190879126147e-05, -5.092009095777e-06, {   1,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.654248692255e-05,  4.514281008067e-07, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  1.024707040559e-05,  6.744527937456e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -4.755602095521e-07,  1.651471035813e-05, {   0,   0,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -9.295087828515e-06, -7.688606724227e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,  -5,   0,   0,   0 } },
 { -1.336547094190e-06,  1.564686840684e-05, {   4,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.373340568787e-06, -1.460188132645e-05, {   0,   0,   1,   0,   0,   0,  12, -23,   3,   0,   0,   0,   0 } },
 { -7.128539151508e-06,  9.840382128656e-06, {   0,   0,   0,   0,   0,   0,   4,  -7,   1,   0,   0,   0,   0 } },
 { -6.358572804129e-06,  1.061001091517e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 {  3.587607514816e-06,  1.337375784659e-05, {   4,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -6.591339158544e-06, -1.036576222827e-05, {   0,   0,   1,   0,   0,   0,  -8,  12,  -2,   0,   0,   0,   0 } },
 {  4.492986841506e-06,  1.246131611790e-05, {   2,   0,  -1,   0,   0, -25,  27,   4,   0,   0,   0,   0,   0 } },
 { -4.436341817378e-07, -1.650963041266e-05, {   0,   0,   0,   0,   0,  18,  -8, -16,   4,   5,   0,   0,   0 } },
 {  1.891589253037e-07, -1.676165314624e-05, {   6,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.422049456334e-06,  1.052672793500e-05, {   1,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.876554562180e-06,  1.506082392968e-05, {   2,   0,  -1,   0,   0,   6,  -9,   2,   0,   0,   0,   0,   0 } },
 {  3.787651541550e-06,  1.314696577882e-05, {   1,  -1,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.621443438490e-05, -7.175805853565e-07, {   3,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  3.916850156625e-06,  1.301478713337e-05, {   0,   0,   1,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 {  1.647990618983e-05, -4.458058116433e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  6.497791029111e-06, -1.042483531577e-05, {   2,   0,   2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.197704675500e-05,  4.941683717934e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  9.969164823546e-06, -6.949215085248e-06, {   2,   0,  -1,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.413901831655e-07,  1.677271135903e-05, {   2,   2,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.897696292464e-06, -8.003425831136e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -7.906465198471e-06,  8.994473290016e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -8,   0,   0,   0 } },
 { -6.304195894782e-06, -1.059459485089e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -4.995223365804e-07, -1.639441560829e-05, {   6,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  7.463354861849e-06, -9.418825337892e-06, {   2,   0,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  2.160995681055e-06, -1.471802434349e-05, {   6,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.415077390161e-05, -2.726593796130e-06, {   1,  -3,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.213778033758e-06,  1.365982034166e-05, {   1,   1,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.667637855391e-05,  1.958525139650e-07, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 { -4.092466033385e-06,  1.277926501786e-05, {   0,   0,   2,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.012009916845e-05,  6.745520173485e-06, {   1,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  1.479694943777e-05,  2.055663617737e-06, {   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.140565279335e-05, -5.443617205655e-06, {   1,   1,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.383437497306e-07,  1.670645206148e-05, {   2,   2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.605521973093e-05,  7.832428920598e-07, {   1,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -3.561496514239e-06, -1.327310386154e-05, {   1,  -3,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.533658735831e-05, -1.497365043442e-06, {   1,   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.548036981554e-05,  1.349641475306e-06, {   1,   0,   0,   0,   0, -29,  35,   0,  -2,   0,   0,   0,   0 } },
 { -6.658748190239e-06,  1.015567558261e-05, {   2,   2,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.823203836194e-06,  1.198994319284e-05, {   6,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.184148279468e-05,  4.969485125531e-06, {   1,  -1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.518416764047e-05, -1.626420306244e-06, {   2,   0,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  1.674662833338e-05,  6.336677903741e-08, {   4,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -7.770174085227e-06, -9.025687335460e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -7,   0,   0,   0,   0 } },
 {  1.300288658882e-05, -3.784247568966e-06, {   0,   0,   3,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.217515800361e-06,  1.356934810089e-05, {   0,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.659039785763e-05,  1.957205944193e-07, {   0,   0,   3,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  3.255930786543e-06,  1.351530932593e-05, {   2,   0,  -2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.129940710066e-05,  5.467011936701e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  8.526751120202e-06,  8.237187152674e-06, {   1,   1,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.650944916198e-05, -2.525220086954e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0 } },
 {  1.673601378062e-05,  1.872131032871e-08, {   1,  -2,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  8.469507468256e-06, -8.281520392794e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.175622362414e-05,  4.992674387776e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   3,   0,   0,   0 } },
 {  1.644215270808e-05,  2.983744481132e-07, {   4,   0,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  8.512530886791e-06, -8.225159609013e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -1.632600919789e-05, -4.113534707333e-07, {   0,   0,   1,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 { -2.467837938743e-06, -1.426914322929e-05, {   2,   0,  -2,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.447473794509e-05, -2.261871876597e-06, {   2,   0,  -2,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.393890138402e-05, -2.794105143653e-06, {   2,   0,  -1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.404606070406e-06, -1.232461750034e-05, {   1,  -1,   0,   0,   0,   0,  -3,   8,   0,   0,   0,   0,   0 } },
 { -1.272077120476e-05, -4.008140451408e-06, {   3,   1,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.127449637585e-05, -5.452363278469e-06, {   3,   0,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.463427489168e-05, -2.087859129518e-06, {   0,   0,   1,   0,   0, -21,  21,   0,  -1,   0,   0,   0,   0 } },
 {  7.030321424343e-06,  9.691363451475e-06, {   2,   0,   0,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.127554227251e-05,  5.441077395761e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  1.611382711859e-05, -6.023634711119e-07, {   4,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.671329896257e-05, -9.942716547118e-10, {   4,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.667809656132e-05,  3.341996652916e-08, {   1,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.133507255284e-07,  1.648935322663e-05, {   1,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -2.488930305953e-06, -1.421049251444e-05, {   2,   0,  -2,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 {  6.613876502001e-06,  1.008279633699e-05, {   2,   2,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.476131612754e-05,  1.926232517078e-06, {   2,   0,   1,   0,   0, -64,  66,   0,   0,   0,   0,   0,   0 } },
 { -4.523418870588e-06, -1.216324192766e-05, {   4,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -3.112751680863e-06, -1.357092083821e-05, {   1,   1,   1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.117566229388e-05,  5.499453043200e-06, {   4,   0,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  7.340217571094e-06,  9.332624288288e-06, {   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.663933186168e-05, -3.282443924353e-08, {   2,  -2,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  8.739921945531e-06, -7.929771232593e-06, {   4,  -2,   1,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.739921945531e-06, -7.929771232593e-06, {   4,  -2,  -1,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.136847410834e-05, -5.286027012742e-06, {   2,   0,   0,   0,   0, -37,  35,   0,   0,   0,   0,   0,   0 } },
 { -1.121250792280e-05, -5.441646055832e-06, {   1,  -1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.243431769244e-06,  1.540460880728e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  1.334238438848e-05, -3.304806400585e-06, {   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0 } },
 { -8.152267472510e-06, -8.494279433702e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -1.653292104095e-05, -1.085164472460e-07, {   0,   2,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.100014424539e-06,  1.453902029757e-05, {   4,   0,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -9.974262379385e-06, -6.658292341353e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 {  1.661023585250e-05, -2.215968047977e-08, {   1,  -2,   1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -7.346952723818e-06, -9.282381124586e-06, {   2,   0,  -1,   0,   0,   0,  10, -21,   2,   0,   0,   0,   0 } },
 {  9.421852576541e-06, -7.204922235420e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -1.661900370452e-05,  7.560070420824e-09, {   2,  -2,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.013817437302e-05, -6.475283591580e-06, {   2,   2,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.484046763599e-05, -1.763526033887e-06, {   2,   0,  -3,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.163794252303e-05,  4.965494003452e-06, {   1,   1,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.654020376483e-05, -6.076863670657e-08, {   4,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.602716577599e-05,  5.472748319430e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 } },
 { -3.160804591271e-06,  1.341178557208e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -1.338615132293e-05, -3.184547919643e-06, {   2,   0,  -1,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.150199497767e-06, -1.341903108109e-05, {   0,   0,   3,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.601548343194e-06, -8.967332270217e-06, {   0,   0,   0,   0,   0,   9,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.183120981461e-06,  1.338356189726e-05, {   2,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.851816076608e-06, -7.713311936845e-06, {   1,  -1,   1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -8.851816076608e-06, -7.713311936845e-06, {   1,  -1,  -1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -5.157791838931e-06,  1.139641861984e-05, {   1,  -1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.133334713091e-05,  5.215225730880e-06, {   3,   1,  -2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.399814557786e-06,  1.314057594967e-05, {   3,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.653987501091e-05, -3.592402546995e-10, {   2,   2,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.919063752796e-06, -8.610592099848e-06, {   0,   2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -4.124824404145e-06,  1.240351034047e-05, {   0,   0,   1,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.087730527865e-05,  5.647202969308e-06, {   1,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  1.008083358732e-05, -6.437646081749e-06, {   2,   2,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.375232799865e-05,  2.759682038283e-06, {   3,  -1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.452210595938e-05,  1.980177370735e-06, {   1,   1,  -3,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.315911845892e-05, -3.342440139708e-06, {   1,   1,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.649418406637e-05, -6.212914809853e-09, {   1,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.735320891376e-07,  1.631619072856e-05, {   4,   0,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.156093774867e-06,  1.333146982057e-05, {   3,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.762836509883e-06, -8.723001740517e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  1.274010306194e-05, -3.744610221276e-06, {   0,   0,   2,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 { -1.305443831568e-05, -3.429087746658e-06, {   0,   0,   1,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.237808575760e-06, -1.423916858399e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0 } },
 {  1.060401469495e-05, -5.870967938473e-06, {   2,   0,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  1.303750716309e-05, -3.434367455389e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   6,  -3,   0,   0 } },
 {  3.150192419399e-06, -1.332023670392e-05, {   2,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.725787104171e-06, -1.374408006828e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 { -1.272487397425e-05, -3.740371582206e-06, {   0,   0,   2,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -1.207153029422e-05, -4.389353767228e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -6,   3,   0,   0 } },
 {  5.175653422268e-06, -1.127858062330e-05, {   4,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.037122934310e-05, -6.078425675579e-06, {   0,   2,  -1,   0,   0,   0, -29,  29,   0,   0,   0,   0,   0 } },
 { -4.495376943747e-06, -1.195078695356e-05, {   1,  -1,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.061258873878e-05,  5.829904410250e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   2,   0 } },
 { -1.624184356403e-05,  1.952601628733e-07, {   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 {  1.817877130365e-06,  1.460291329931e-05, {   2,   0,   1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.915453053300e-06,  1.050388778700e-05, {   2,   0,  -2,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.522306224852e-05, -1.193600256542e-06, {   0,   0,   2,   0,  -5,   4,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.715329629456e-06,  8.697130879321e-06, {   2,  -2,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.539385035948e-05,  1.014871435924e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,  -2,   0 } },
 { -1.334024959694e-05, -3.067901181978e-06, {   2,   0,  -1,   0,   0, -16,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.407391136309e-05, -2.332450133043e-06, {   2,   2,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.627958528075e-05,  1.243258496163e-07, {   2,   2,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.785837132970e-06, -9.608878751896e-06, {   1,   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  3.934289370976e-06,  1.245646159298e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  6.379925214959e-06,  1.001018819620e-05, {   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  5.557524100697e-06, -1.083042391172e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 {  8.238505017410e-06, -8.143038329865e-06, {   0,   0,   1,   0,   0,   0,  -4,   7,   3,   0,   0,   0,   0 } },
 { -6.293507267640e-06, -1.008427186032e-05, {   0,   2,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -1.636302330428e-05,  1.208880346877e-08, {   6,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.298466772669e-07, -1.604045318400e-05, {   6,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.378503454767e-06,  6.982631413927e-06, {   0,   0,   1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 {  5.603123738443e-06,  1.075746566899e-05, {   3,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.365064906768e-05,  2.698740617457e-06, {   1,  -1,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.331655085063e-05,  3.031498757938e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0 } },
 { -3.941926971308e-06,  1.240565457358e-05, {   0,   0,   0,   0,   0,   2,  -6,   7,   0,   0,   0,   0,   0 } },
 {  3.138394811029e-06,  1.320754514602e-05, {   3,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.347095393050e-05,  2.874374911545e-06, {   1,  -1,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -5.658019075733e-06, -1.068442818486e-05, {   2,   0,   0,   0,   0,   0,  18, -22,   0,   0,   0,   0,   0 } },
 { -1.633333728013e-05, -2.500004718570e-09, {   2,  -2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.444504736041e-05, -1.888984910208e-06, {   2,   0,   0,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.948636031000e-06, -7.380862336377e-06, {   2,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.763226254592e-07, -1.575171006801e-05, {   0,   2,  -1,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.018714617424e-05,  6.138554800046e-06, {   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.161240518443e-05,  4.707358016884e-06, {   3,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -3.597457922139e-06,  1.272048042696e-05, {   1,   1,  -1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.970884376099e-06, -1.234514463724e-05, {   0,   0,   1,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 } },
 { -1.152960703079e-05, -4.780249863812e-06, {   4,   0,  -2,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  1.584363828625e-05,  4.645018558324e-07, {   2,  -2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  8.361697455930e-06,  7.927677247272e-06, {   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.928026919918e-06,  1.236004107322e-05, {   1,   1,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.054840502324e-05,  5.738863877817e-06, {   2,  -2,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  7.539959264279e-06,  8.747266460143e-06, {   0,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 {  1.159627696619e-05, -4.690649801277e-06, {   1,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.049998718735e-05, -5.783516129460e-06, {   2,   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -1.106448263792e-05,  5.216629746130e-06, {   2,   0,  -2,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.394136620011e-05,  2.328378693468e-06, {   0,   2,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.325754439215e-05, -3.007851388865e-06, {   1,  -1,  -1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.333133762396e-07, -1.582370637773e-05, {   0,   0,   4,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.547758658190e-06,  9.706701376797e-06, {   0,   2,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.472021028407e-06, -1.377958140176e-05, {   2,   0,   0,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 { -3.567999227654e-06,  1.268351412769e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 { -8.083780440553e-06,  8.163560332972e-06, {   6,   0,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.719115129348e-06,  1.152272768290e-05, {   0,   0,   1,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 { -7.885287884973e-06, -8.353572708069e-06, {   2,   0,  -3,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.074201138314e-05,  5.491248274008e-06, {   4,   0,  -2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -8.169842935672e-06, -8.060469038738e-06, {   0,   0,   1,   0,   0,   0,   4,  -7,  -3,   0,   0,   0,   0 } },
 {  1.009498228442e-05,  6.133643864415e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,  -4,   0,   0,   0 } },
 { -1.149641902169e-05,  4.725363310089e-06, {   2,   0,  -2,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 {  1.550758498694e-05, -7.102746027415e-07, {   0,   0,   2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  1.421523094578e-05, -2.001176454812e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 { -7.735383576258e-06,  8.477931495793e-06, {   4,   0,  -1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 {  1.482014679505e-05, -1.384159492625e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -6.111971919265e-06,  1.008945174444e-05, {   0,   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.737226004307e-07, -1.591668839863e-05, {   0,   0,   0,   0,   0,   6,  -9,   0,   2,   0,   0,   0,   0 } },
 { -1.375957823897e-05,  2.428911587177e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -6,   5,   0,   0,   0 } },
 { -1.006733992589e-05,  6.119192087786e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 { -7.126319686138e-06,  9.055465824080e-06, {   4,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  7.714126020221e-06, -8.466036644879e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   4,   0,   0,   0 } },
 {  3.152181635666e-06,  1.302698579244e-05, {   2,  -2,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.058279397761e-06, -1.311819181126e-05, {   1,   1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.940352189653e-06, -1.323575193837e-05, {   6,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.290853404962e-07, -1.574566601014e-05, {   0,   0,   0,   0,   0,   4,  -7,   0,   2,   0,   0,   0,   0 } },
 {  1.615368517979e-05,  1.578018047594e-08, {   0,   2,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.910366563951e-06,  1.325205008203e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,  -2,   0,   0 } },
 {  1.598972960387e-05, -1.716879029249e-07, {   2,   2,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -4.484895668705e-06,  1.167328178025e-05, {   3,   1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.121804314680e-06,  1.403179861648e-05, {   1,   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -4.038849648970e-06,  1.210768074184e-05, {   2,   0,  -1,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.470225047535e-06,  1.267246860099e-05, {   2,   0,   1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -5.852853771874e-06,  1.028919228651e-05, {   0,   0,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 { -1.445647269367e-05,  1.684243219874e-06, {   1,   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  2.516244157143e-06,  1.362365412344e-05, {   2,  -2,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.869308508270e-06,  1.226516894356e-05, {   0,   0,   1,   0,   0, -21,  17,   8,  -3,   0,   0,   0,   0 } },
 {  9.654025314103e-06,  6.478520708995e-06, {   0,   0,   3,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -9.514620369025e-06, -6.613979388097e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  8.053291692051e-06,  8.074893429723e-06, {   4,   0,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.993698054943e-07, -1.532208383246e-05, {   0,   0,   1,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0 } },
 {  1.219295397985e-05,  3.918718299304e-06, {   0,   0,   0,   0,   0,   0,   8, -15,   3,   0,   0,   0,   0 } },
 {  9.381884457298e-06, -6.727633831614e-06, {   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -1.041407118216e-06, -1.506340662834e-05, {   2,   0,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.608762173504e-05, -1.342670703332e-09, {   0,   2,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.232790850074e-07, -1.546375475482e-05, {   0,   2,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.033989386755e-06, -1.304616472104e-05, {   0,   1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.899112516101e-06,  1.217543502252e-05, {   4,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  3.111131132991e-06,  1.295715121379e-05, {   1,   1,   0,   0,   0, -22,  21,   0,   0,   0,   0,   0,   0 } },
 {  8.485587047443e-06, -7.582333452055e-06, {   0,   0,   0,   0,   0,   8, -17,   8,  -1,  -5,   0,   0,   0 } },
 {  4.674011860470e-06, -1.139155426008e-05, {   0,   0,   1,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.132460248463e-05, -4.739345661473e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   7,   0,   0,   0 } },
 {  3.162152663806e-06, -1.289902522200e-05, {   4,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.603603734370e-05, -2.349191527408e-08, {   6,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  6.308871090063e-06, -9.743754784564e-06, {   3,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -6.365222578520e-06, -9.687337394259e-06, {   1,   1,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.798630442207e-06, -1.025215828418e-05, {   1,  -1,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.061081487153e-06, -1.298009276627e-05, {   1,   0,   0,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.083142320597e-06, -1.295336127945e-05, {   1,  -1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  1.327723346728e-06, -1.470516242623e-05, {   2,  -2,   1,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 {  1.327723346728e-06, -1.470516242623e-05, {   2,  -2,  -1,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 { -3.505618524476e-06, -1.252436374543e-05, {   2,   0,  -1,   0,   0,   0,  -2,   8,  -5,   0,   0,   0,   0 } },
 {  1.365379787826e-05, -2.367114068249e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  1.191510993366e-05, -4.102533227918e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 { -9.590440454855e-06,  6.410498705184e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   3,   0,   0 } },
 { -1.715458355817e-06, -1.428319110864e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -9,   0,   0,   0 } },
 {  9.162644211801e-06,  6.835427157086e-06, {   3,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  4.557658912224e-07, -1.553975031864e-05, {   2,   2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.043595009077e-05, -5.548966402324e-06, {   4,   0,  -1,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.820367697163e-06,  6.164387610004e-06, {   0,   0,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0 } },
 {  1.427458831493e-05,  1.707794702792e-06, {   0,   0,   2,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.339018866896e-06, -1.463874182259e-05, {   0,   0,   2,   0,   0, -20,  15,   8,   0,   0,   0,   0,   0 } },
 {  1.339018866896e-06, -1.463874182259e-05, {   0,   0,   0,   0,   0,  20, -15,  -8,   0,   0,   0,   0,   0 } },
 { -1.429826659307e-05, -1.678068256576e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -1.480973300371e-05,  1.165170847593e-06, {   3,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.875359119093e-06,  1.209918138062e-05, {   1,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.980763685814e-06,  5.990696988887e-06, {   1,   0,   0,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.183472766785e-06, -1.478266244005e-05, {   3,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.304591558454e-05,  2.916330074764e-06, {   6,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.335174144702e-07, -1.572860658311e-05, {   3,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.299021328740e-05, -2.971446272552e-06, {   2,   0,   2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.591042496997e-05,  4.879558649735e-08, {   0,   2,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.018766039236e-05,  5.766464676721e-06, {   1,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.041317467523e-05, -5.536763423352e-06, {   0,   0,   1,   0,   4,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.040661627745e-05,  5.540184105508e-06, {   2,   0,  -2,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.883255605854e-06,  9.050953151676e-06, {   2,   0,  -1,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 { -1.590831162697e-05,  2.204484000462e-08, {   2,   0,   3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.049936631299e-06, -1.288035234853e-05, {   3,   1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.486748060791e-06, -1.343521558292e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  6.282351208414e-06, -9.638285603341e-06, {   2,   0,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  6.125801799888e-06,  9.784514865087e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 { -7.358885638495e-06,  8.548589368116e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 {  1.056819865418e-05,  5.335183220054e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -1.088952305092e-05,  5.010939783379e-06, {   3,   1,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.156566746781e-06, -8.740736842114e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   2,   0,   0,   0 } },
 {  3.379329693064e-06,  1.251074704688e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -3.017583378930e-06,  1.286907169853e-05, {   1,  -1,   1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.393490467948e-06,  1.248882977482e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -1.183684222740e-06, -1.468357316884e-05, {   6,  -2,  -3,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.597671784619e-06,  1.026871950269e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 { -6.095971443808e-06,  9.770382897897e-06, {   2,   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.580426411104e-05,  5.999693222982e-08, {   2,  -2,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -3.832376383919e-06,  1.203152279267e-05, {   2,   2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.178223602718e-06, -8.669371077089e-06, {   1,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.572475990156e-05,  1.202118016139e-07, {   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 {  3.827473495194e-06,  1.201593078151e-05, {   2,   2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.906000462519e-06,  8.932974581042e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -9.216950938560e-06,  6.619110834188e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 {  1.391448129593e-05,  1.918584100913e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -2,  -5,   0,   0,   0 } },
 {  7.788295250956e-06, -8.042712149313e-06, {   2,   0,  -1,   0,   0,  -8,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.319304963411e-05,  2.637396230504e-06, {   1,   1,  -1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.387440668997e-06,  6.435268298848e-06, {   2,   0,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -9.498269938774e-06, -6.316657101731e-06, {   1,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.058262208398e-06,  6.746433056255e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,  -1,   0,   0,   0,   0 } },
 { -5.306320942848e-06,  1.049800882174e-05, {   1,  -1,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -6.559475092035e-06,  9.240176440656e-06, {   4,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  1.507669209040e-05,  7.191650493696e-07, {   4,   0,  -2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -1.096423214111e-06,  1.469817465303e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   5,   0,   0,   0 } },
 { -3.011022630826e-06, -1.278102462843e-05, {   0,   1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.499718276184e-06,  1.228754479046e-05, {   0,   0,   1,   0,   0, -21,  25,  -8,   3,   0,   0,   0,   0 } },
 { -9.934358161243e-06,  5.845161068169e-06, {   0,   0,   1,   0,   0, -29,  36,  -4,   0,   0,   0,   0,   0 } },
 {  4.761228650929e-06,  1.100555340701e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 {  4.331266123904e-06, -1.143330097698e-05, {   0,   0,   1,   0,   0, -29,  34,   0,   0,   0,   0,   0,   0 } },
 { -1.516333024190e-05,  6.006335610760e-07, {   1,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -6.148398607872e-06, -9.610821269122e-06, {   6,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.007193656406e-06,  1.474988113318e-05, {   4,   0,  -2,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.197197315297e-05,  3.781233826293e-06, {   3,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.334534973943e-05, -2.407454909653e-06, {   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.331186412778e-05, -2.440550001211e-06, {   1,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -2.992471717142e-06, -1.275958553707e-05, {   1,  -1,   1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.988392943755e-06,  1.175877189790e-05, {   2,  -1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.029052675880e-06,  1.271475957166e-05, {   2,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.475172017042e-05,  9.836733157670e-07, {   1,   1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.520072930015e-06,  1.120767778769e-05, {   6,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  9.074155521240e-06,  6.652894610250e-06, {   2,   2,  -2,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.028839729735e-06, -8.686697038270e-06, {   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -7.822228346523e-06,  7.892878728733e-06, {   6,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.707405995352e-06, -1.000692689380e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -3.848243299199e-07,  1.532925687772e-05, {   3,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.108161731448e-06, -8.604338935148e-06, {   4,   0,  -3,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.439073778586e-05, -1.321669321006e-06, {   3,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.064481748537e-05,  5.067557549507e-06, {   0,   0,   1,   0,   0,   0,   7, -14,   0,   0,   0,   0,   0 } },
 { -5.260290335977e-06, -1.045034887090e-05, {   2,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.561449089403e-05,  8.546561901839e-08, {   0,   2,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.455072634989e-05, -1.145648849798e-06, {   0,   0,   0,   0,   5,  -4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.831073872392e-06, -1.186024751774e-05, {   1,  -1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.006387029176e-06,  8.680492410523e-06, {   0,   0,   0,   0,   0,  16, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.107844818038e-06, -9.575220670508e-06, {   6,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.274517233914e-07, -1.505244048235e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -2.958000092306e-06,  1.271867779987e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  1.540721230718e-05,  2.686090064001e-07, {   2,   0,   0,   0,   0,   0,   6, -15,   0,   0,   0,   0,   0 } },
 { -1.759515878959e-06, -1.391139779348e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 { -1.297849203441e-05,  2.688741994538e-06, {   4,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.024475556579e-05,  5.420758690051e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -1.547202561981e-05,  1.856722313404e-07, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 { -1.307966191879e-05, -2.572122793608e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  8.014378394192e-07, -1.485018379854e-05, {   3,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.831985153016e-06, -1.281960091533e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 {  1.240212259486e-05, -3.243383482358e-06, {   1,   1,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.026703567160e-06,  1.261306970330e-05, {   3,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.390683847104e-06,  1.024589434984e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   2,   0 } },
 { -1.561040299288e-05,  2.049631270425e-08, {   4,   0,  -2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.195665101673e-07,  1.520994769925e-05, {   1,   0,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.104665072228e-05,  4.576461672054e-06, {   2,   0,   1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  8.606136667033e-06,  7.016836679434e-06, {   2,   0,  -1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  1.437860850683e-05, -1.239424948384e-06, {   1,  -1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.251794298280e-05,  3.100089895546e-06, {   2,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.278793577585e-05, -2.830009941110e-06, {   1,  -1,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.804620691448e-06, -1.080589799722e-05, {   3,   1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.015019637429e-05,  5.458144053998e-06, {   0,   0,   2,   0,   0,   0, -34,  39,   0,   0,   0,   0,   0 } },
 { -3.791142860621e-06,  1.181611910741e-05, {   0,   0,   1,   0,  -3,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.141579899421e-05, -4.187825811641e-06, {   2,   0,  -1,   0,   0,   0,   6,   0,  -8,   0,   0,   0,   0 } },
 { -7.658808616406e-06, -7.944068539732e-06, {   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.379822633426e-05, -1.800740961137e-06, {   2,   0,   0,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.909140826481e-06, -9.682937015310e-06, {   6,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.489845912027e-05,  6.911765214422e-07, {   1,  -1,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 { -3.794597222588e-07, -1.520576627435e-05, {   2,   0,   3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.525681709192e-05, -3.259456304666e-07, {   2,   0,  -1,   0,   0,   3,  -6,   0,   1,   0,   0,   0,   0 } },
 { -1.289511150846e-05, -2.685422766229e-06, {   1,   1,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.273377592976e-05,  2.843238449980e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0 } },
 {  7.102256114938e-07, -1.486586029104e-05, {   4,   0,  -3,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  7.076055652761e-06, -8.499403377924e-06, {   0,   0,   0,   0,   0,   0,  12, -20,   0,   0,   0,   0,   0 } },
 { -1.492268334247e-05, -6.507404254166e-07, {   0,   0,   2,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 {  3.207696646734e-06, -1.235745074041e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -3.032395219769e-06, -1.253227412860e-05, {   6,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.787619111692e-06,  1.177458336209e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -4,   0,   0,   0,   0 } },
 {  2.599347123006e-06,  1.296142733338e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -2,   0,   0,   0 } },
 {  7.742505678912e-06,  7.817340441634e-06, {   0,   0,   3,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.308914444567e-06,  8.248594170977e-06, {   1,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.412736775151e-05, -1.428495660200e-06, {   4,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.314678838649e-05,  2.405195886356e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0 } },
 {  1.543470192748e-05, -1.158340470814e-07, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 {  1.050156307292e-05, -5.045919655533e-06, {   0,   2,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.395904276106e-05,  1.586888031961e-06, {   2,   0,  -1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0 } },
 { -1.533047017103e-05,  2.154497962078e-07, {   2,   0,  -3,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.648848759247e-06,  1.289585327542e-05, {   4,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.542750370612e-05, -1.157808472653e-07, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 { -2.942594243380e-07,  1.524695901937e-05, {   4,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.967147670949e-06,  9.562900414185e-06, {   3,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.008973682800e-06,  1.251913685710e-05, {   2,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.322793427093e-07,  1.509050297276e-05, {   2,   0,  -1,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 { -2.645690490014e-06,  1.287294987895e-05, {   1,  -1,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -2.645690490014e-06,  1.287294987895e-05, {   1,  -5,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  1.345828003759e-05, -2.048041752832e-06, {   1,   1,   0,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -1.515881953103e-05,  3.441295211110e-07, {   4,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -7.458019636820e-06, -8.031477386688e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 { -1.387756255257e-05, -1.607132538520e-06, {   2,   0,   0,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 { -1.232859479136e-05,  3.134352949678e-06, {   1,  -1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.535457760947e-06,  1.192659253107e-05, {   2,   0,  -3,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.543414971284e-05,  2.369357701586e-08, {   3,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.852913400042e-06,  5.602959397761e-06, {   0,   0,   2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.287499582933e-05, -2.579406383304e-06, {   1,   1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.543747887572e-05,  1.403341358658e-08, {   2,   0,  -3,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.525154546614e-05,  1.961010816209e-07, {   0,   2,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.169056945895e-06, -1.227740057813e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -6.521205016604e-06,  8.921983582992e-06, {   0,   0,   1,   0,   0,   0,  10, -18,   0,   0,   0,   0,   0 } },
 { -1.516506839649e-05,  2.775263303554e-07, {   4,   0,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  9.944966658752e-06,  5.495542429668e-06, {   4,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  3.007837598064e-06,  1.243040976782e-05, {   2,   2,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.810720185740e-06, -5.623086731996e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 {  1.106093999725e-05, -4.363980483082e-06, {   0,   2,  -1,   0,   0, -22,  22,   0,   2,   0,   0,   0,   0 } },
 { -9.998936377391e-06,  5.425464363304e-06, {   2,  -2,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  7.093957935486e-06,  8.328361250694e-06, {   1,  -1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.952128140081e-06,  1.046984820161e-05, {   2,   0,   0,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0 } },
 { -7.393388282515e-06, -8.025444200740e-06, {   1,  -1,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.927092422508e-06, -1.249096808720e-05, {   1,  -1,   1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.555299991293e-06,  1.086102216328e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 { -2.768449566030e-06,  1.264510797356e-05, {   1,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.126436087251e-06, -1.027971400590e-05, {   4,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.490889131822e-05,  4.968465050970e-07, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 {  1.529275918140e-05, -1.109657886188e-07, {   2,   0,   0,   0,   0,  28, -28,   0,   0,   0,   0,   0,   0 } },
 { -2.414320212611e-06,  1.298728474045e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.632793431479e-07, -1.503261081789e-05, {   3,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.551861863582e-06,  5.842879898503e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -7,   0,   0,   0,   0 } },
 {  6.682362970958e-07, -1.472601915775e-05, {   2,   0,  -1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.671722584546e-06,  7.715019455173e-06, {   0,   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  6.456060846606e-07, -1.473709882845e-05, {   3,   0,  -3,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.438155071667e-05,  1.000559143830e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   4,   0,   0,   0 } },
 {  8.334160946818e-06,  7.043077829917e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 { -1.146716311968e-05,  3.906362631918e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   7,   0,   0,   0 } },
 {  1.507672871230e-05, -2.945190827478e-07, {   0,   2,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.901934395773e-06, -1.346779876092e-05, {   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.666431914023e-06,  9.694397835603e-06, {   0,   2,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.219984198993e-05, -3.158621565494e-06, {   2,   2,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.148950719242e-05,  3.856635383616e-06, {   2,   0,  -1,   0,   0,   0,  -6,   0,   6,   0,   0,   0,   0 } },
 {  1.291002741617e-05,  2.423260520042e-06, {   1,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.303161048154e-05, -2.301102305799e-06, {   4,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.006043929671e-05, -5.267664761162e-06, {   2,   0,   2,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  9.661185645588e-07, -1.435309203794e-05, {   4,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.927086090830e-06, -8.390023073400e-06, {   1,  -1,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.492177341528e-06, -1.282215645039e-05, {   0,   0,   1,   0,   0,   0,  -5,  10,  -4,   0,   0,   0,   0 } },
 {  1.165000899963e-05,  3.660815194821e-06, {   1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.091551332444e-05, -4.391584382543e-06, {   0,   0,   1,   0,   0, -10,  11, -16,   4,   5,   0,   0,   0 } },
 {  9.047684517110e-06,  6.258624720905e-06, {   0,   0,   1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  2.961375152824e-06,  1.234418886571e-05, {   0,   0,   0,   0,   0,  24, -27,   0,   0,   0,   0,   0,   0 } },
 {  9.148350146342e-06, -6.141412589966e-06, {   0,   0,   2,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -3.750737689381e-06,  1.153835091771e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   4,   0,   0,   0,   0 } },
 { -5.947567669206e-06,  9.338875084856e-06, {   1,   1,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.414840151316e-05, -1.130737370818e-06, {   0,   0,   0,   0,   0,   2,  -2,  -3,   8,  -6,   0,   0,   0 } },
 { -3.168334680357e-06,  1.210370951029e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -1.397799534918e-05, -1.291527337285e-06, {   2,   0,  -2,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.355258501685e-05, -1.715894162116e-06, {   0,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.526051208385e-05, -3.608744202699e-09, {   0,   0,   3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.501019879929e-05, -2.525159555623e-07, {   2,   0,  -2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.524835478473e-05, -1.389212844023e-08, {   1,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.541840095984e-06,  1.071931294729e-05, {   2,   0,  -1,   0,   0,   2,   6, -15,   8,  -6,   0,   0,   0 } },
 { -1.227526879601e-05, -2.984308606795e-06, {   0,   0,   1,   0,   0, -16,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.490330972678e-05,  3.548995553117e-07, {   2,   0,  -2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -1.523968427105e-05, -1.567229999876e-08, {   4,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.247517846597e-05,  2.764181985218e-06, {   0,   1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.897009000764e-06, -1.234070024513e-05, {   4,   0,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.848036266187e-06,  1.338343430677e-05, {   4,   0,  -2,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -7.085983255340e-06, -8.144408121321e-06, {   2,   0,  -1,   0,   1,   0, -11,  11,   0,   0,   0,   0,   0 } },
 { -7.606689613623e-07,  1.446861444359e-05, {   1,   1,  -1,   0,   0, -20,  24,  -8,   3,   0,   0,   0,   0 } },
 {  1.494747092618e-05, -2.776670386143e-07, {   0,   0,   1,   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0 } },
 { -3.547176893103e-06, -1.167527047689e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 {  3.810037266705e-06, -1.140681192719e-05, {   1,   0,   0,   0,   0, -25,  28,   0,   0,   0,   0,   0,   0 } },
 {  2.471057170120e-06, -1.274541855421e-05, {   0,   0,   1,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 { -4.216506471834e-06, -1.098654101580e-05, {   0,   0,   1,   0,  -3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.418419121952e-05, -1.017661468765e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -7.449248698964e-06, -7.729015753966e-06, {   1,  -1,   1,   0,  -3,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -4.956306764530e-06,  1.021945522002e-05, {   0,   0,   0,   0,   0,   1,  -7,  10,   0,   0,   0,   0,   0 } },
 {  6.160535263437e-06, -9.008632862774e-06, {   2,   0,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -9.187147050672e-06, -5.977206464935e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   4,   0,   0,   0 } },
 { -3.038653616634e-06,  1.212482190461e-05, {   0,   0,   1,   0,   0, -27,  30,   0,   0,   0,   0,   0,   0 } },
 {  1.324252645877e-05,  1.913863253140e-06, {   2,   0,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  3.101535303304e-06, -1.205346946570e-05, {   1,  -1,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  2.591693159259e-06,  1.256278661410e-05, {   2,   0,  -2,   0,   0,   0,   9, -14,   3,   0,   0,   0,   0 } },
 { -9.349963865568e-06, -5.801920773796e-06, {   2,   0,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0 } },
 { -7.175984756348e-06, -7.975683493834e-06, {   2,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.066206612800e-06,  6.081544456978e-06, {   0,   0,   1,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 { -9.550702997341e-06, -5.593199971041e-06, {   4,   0,  -2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.117577237172e-06,  8.025496105860e-06, {   3,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.666811926401e-06, -9.475527147531e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -1.238983651123e-05,  2.750103242664e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -5.909839898779e-06,  9.229742667679e-06, {   4,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  5.583408807798e-06,  9.555684630949e-06, {   1,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -4.840646116248e-06, -1.029682877950e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.214911734569e-05, -2.985781045100e-06, {   0,   0,   1,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 { -6.135285335159e-06, -8.998564144706e-06, {   1,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.637852415171e-06,  8.490891214297e-06, {   4,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  4.935302613140e-06,  1.019200646010e-05, {   0,   0,   0,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 {  5.416098156442e-06,  9.706160790949e-06, {   2,   0,  -2,   0,   0,   6, -12,   8,  -3,   0,   0,   0,   0 } },
 { -1.731254136231e-06,  1.338633182396e-05, {   0,   0,   0,   0,   0,   0,   2,  -2,   1,   0,   0,   0,   0 } },
 {  9.868266624038e-06, -5.247473369635e-06, {   4,   0,  -2,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.807054894867e-06,  8.299698119791e-06, {   1,   1,   0,   0,   0, -17,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.302290534632e-05, -2.080439304854e-06, {   1,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.373253626445e-05, -1.368402154749e-06, {   2,  -2,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.143486258769e-05,  3.651790545327e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   2,   0,   0,   0 } },
 { -1.005736280624e-05,  5.023032527597e-06, {   4,  -2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.360511755834e-06,  1.271849532355e-05, {   2,   0,  -1,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.735939802551e-07, -1.460247352187e-05, {   2,  -2,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.181938677694e-05, -3.255980412043e-06, {   0,   0,   0,   0,   0,   0,  12, -24,   9,   0,   0,   0,   0 } },
 {  9.192367721461e-06, -5.881037104654e-06, {   0,   0,   1,   0,   0,   0,   7, -16,   6,   0,   0,   0,   0 } },
 { -1.145799879367e-05,  3.605417741013e-06, {   2,   0,   0,   0,   0,   0, -12,  21,  -2,   0,   0,   0,   0 } },
 { -1.006307186999e-05, -4.992799459011e-06, {   0,   0,   1,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.504655987252e-05, -1.860339684673e-09, {   3,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.934440844393e-06, -5.113170502622e-06, {   1,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -5.496249160448e-06, -9.548299678410e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   4,   0,   0,   0 } },
 {  1.296375174865e-05,  2.077220510055e-06, {   2,   0,  -2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.912378453708e-06, -1.112776284369e-05, {   0,   0,   2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  5.824847369839e-07, -1.444718654808e-05, {   0,   0,   2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.437689059358e-05, -6.525059934413e-07, {   0,   0,   3,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -5.639098525164e-06,  9.382115224261e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  6.322923121974e-06,  8.697887660095e-06, {   1,  -1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.501954739538e-06,  1.050567844913e-05, {   1,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  5.603049608510e-06, -9.395810052394e-06, {   2,   0,   1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 { -3.928921192520e-06, -1.106550506877e-05, {   2,   2,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.331202824266e-05, -1.680196604414e-06, {   3,   1,  -4,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.928278732361e-06, -1.106357226995e-05, {   2,   2,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.869693090267e-06,  1.211997617161e-05, {   3,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.347009260795e-07, -1.405188080126e-05, {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  1.481190867565e-05, -1.668665363059e-07, {   2,   0,  -4,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.788672186415e-06,  8.188328602559e-06, {   1,   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.249666726369e-05,  2.478799148235e-06, {   1,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.158690988334e-06,  9.806168570393e-06, {   1,  -1,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.337428903546e-07,  1.472802045917e-05, {   1,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.464969452689e-05,  3.077214912155e-07, {   1,  -1,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -2.913600451019e-06, -1.204095843778e-05, {   6,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.086965458412e-05, -4.078822450981e-06, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.047072266725e-07,  1.444262478533e-05, {   3,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.833976295451e-06, -1.211065641070e-05, {   1,  -1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.953741881248e-06, -8.990875209687e-06, {   0,   0,   4,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.000747876852e-05,  4.934752038640e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 {  1.487061841436e-05,  7.102610228633e-08, {   4,   0,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.780519752578e-06,  1.215908903127e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  6.331354072220e-06,  8.603315850271e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.692133759973e-06,  8.240873886928e-06, {   1,  -1,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.843735343862e-06,  1.208517542485e-05, {   3,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.012605767658e-06, -9.913858975259e-06, {   2,   0,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -3.994047296575e-06, -1.093062825824e-05, {   0,   0,   1,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.593353603051e-06, -8.324754492624e-06, {   2,   0,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -9.650126062950e-06,  5.266855498148e-06, {   0,   0,   2,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.276164554478e-06,  1.163881930916e-05, {   2,  -2,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.766272892478e-06, -6.147474056696e-06, {   2,   0,   0,   0,   0,   0,  17, -22,   0,   0,   0,   0,   0 } },
 {  1.332048729660e-05, -1.593033276166e-06, {   0,   0,   1,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.386462933339e-05,  1.048095640778e-06, {   2,   0,   1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  7.934347219644e-06,  6.976990041988e-06, {   4,   0,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -9.559337242791e-06,  5.351693051182e-06, {   2,   0,   2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.481748928581e-05,  9.254329181265e-08, {   0,   0,   3,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.144909436626e-06,  1.376418772708e-05, {   4,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.389562249218e-06,  8.518019100748e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   3,   0,   0,   0 } },
 {  2.848930076882e-06, -1.205841889638e-05, {   1,   1,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.619464188721e-06,  5.282983398045e-06, {   0,   0,   0,   0,   0,   3,  -6,   0,   3,   0,   0,   0,   0 } },
 { -1.379445898296e-05, -1.106014939554e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -9.836548951707e-06, -5.059292162202e-06, {   2,   0,  -1,   0,   0,  -2,   5,   1,  -8,   6,   0,   0,   0 } },
 { -6.742832319654e-07, -1.421685192190e-05, {   0,   0,   4,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.576190542741e-06,  6.313870106435e-06, {   0,   0,   0,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 { -7.799308151627e-06, -7.088198577287e-06, {   6,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.899747255068e-06,  1.198384871711e-05, {   1,   0,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.506788459262e-06,  8.376171823217e-06, {   2,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  1.486411154919e-05, -6.817368584205e-09, {   4,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.346595097182e-06,  1.152285266533e-05, {   1,   1,  -1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.475111645457e-05,  1.167198052436e-07, {   0,   0,   3,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.323619383348e-05, -1.629862347760e-06, {   2,  -2,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.349895833219e-05,  1.364505678688e-06, {   1,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -9.989429550136e-06, -4.872810653450e-06, {   2,   0,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.399719494083e-06,  6.462386496404e-06, {   1,   1,  -2,   0,   0,   5,  -9,   3,   0,   0,   0,   0,   0 } },
 {  6.738941722292e-06,  8.119508496285e-06, {   2,   2,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.461492575522e-05,  2.378522245395e-07, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  4.173537195628e-06, -1.067456760517e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -9.692275345203e-06, -5.153177927629e-06, {   0,   0,   0,   0,   4,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.359584219630e-05, -1.234734594863e-06, {   1,  -1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -8.976273475112e-06,  5.851813437243e-06, {   2,   0,   0,   0,   0, -43,  42,   7,  -8,   6,   0,   0,   0 } },
 { -1.024522089621e-05,  4.580923611560e-06, {   2,   0,  -1,   0,  -1,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  2.968666402262e-06, -1.185638266175e-05, {   0,   2,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.062370289602e-06, -1.375897719678e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -6.589043911779e-06, -8.232298105635e-06, {   2,   0,   0,   0,   0,   0,  10, -21,   2,   0,   0,   0,   0 } },
 {  1.468517306947e-05,  1.356722562204e-07, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 { -1.481643303731e-05, -2.216298297513e-09, {   2,  -2,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.913230284840e-06,  1.190079659299e-05, {   2,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.266755336681e-05, -2.145378669741e-06, {   1,  -1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  8.263054340187e-06, -6.549028450871e-06, {   0,   0,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0 } },
 {  6.975737009694e-06, -7.831018214961e-06, {   0,   0,   0,   0,   0,   8,  -3, -19,   0,   3,   0,   0,   0 } },
 {  9.866470121827e-06,  4.940248418373e-06, {   1,  -1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.471525061021e-06,  8.333139829293e-06, {   2,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  6.457038120390e-06,  8.346848570139e-06, {   0,   2,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  9.542074912966e-06, -5.259625214941e-06, {   2,  -2,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.040900982334e-05,  4.385797914988e-06, {   0,   0,   0,   0,   0,   2,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -5.942903301782e-06, -8.848110366037e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -6.992825329348e-07, -1.409134103668e-05, {   2,  -2,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.886356625514e-06,  4.900955100114e-06, {   0,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.147346713339e-06, -1.162806307723e-05, {   4,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.250927388160e-05, -2.265381054352e-06, {   0,   0,   1,   0,   0,   0,  -6,   0,   6,   0,   0,   0,   0 } },
 {  5.889380769211e-07, -1.417933595171e-05, {   2,  -2,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.108525835160e-06, -1.165877655583e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -6,   3,   0,   0 } },
 {  6.368633597226e-06,  8.398350597074e-06, {   1,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  8.675433732114e-06, -6.090439726636e-06, {   1,   1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -4.444570608753e-06, -1.031875285518e-05, {   4,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -1.475390549526e-05, -4.768334519093e-09, {   2,  -2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.690857392826e-06,  8.067042745812e-06, {   2,  -2,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  7.139183949316e-06,  7.615442575324e-06, {   1,   0,  -3,   0,   0,  47, -49,   0,   0,   0,   0,   0,   0 } },
 { -9.948591757903e-06, -4.804519583016e-06, {   0,   2,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -9.844880926528e-06,  4.902896805660e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  1.068753441947e-05, -4.059555625864e-06, {   1,  -1,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  8.398235881198e-06, -6.348152092137e-06, {   6,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  5.115668022315e-06,  9.630648423721e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 {  1.034785805561e-05,  4.395231600192e-06, {   0,   0,   1,   0,  -5,  14,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.265908587134e-05,  2.078076858876e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,  11,  -3,   0,   0 } },
 {  4.503664526560e-06, -1.023341148084e-05, {   0,   0,   1,   0,   0,  16, -26,   0,   0,   0,   0,   0,   0 } },
 { -4.503664526560e-06, -1.023341148084e-05, {   0,   0,   1,   0,   0, -16,  26,   0,   0,   0,   0,   0,   0 } },
 {  5.587600018971e-07,  1.417490609089e-05, {   0,   0,   3,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.471518505336e-05,  1.821639667311e-08, {   4,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.687259099788e-06,  5.039949463212e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   1,  -4,   0,   0,   0 } },
 { -4.258625828522e-06, -1.046441369788e-05, {   1,  -1,   0,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 {  2.870910896930e-06,  1.185191845943e-05, {   1,  -1,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.159166676283e-06,  1.056124502544e-05, {   1,  -1,  -2,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -2.722065891451e-06,  1.199580753029e-05, {   2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.042210139980e-05,  4.293658676222e-06, {   1,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.013696975858e-05,  4.566189759830e-06, {   1,  -1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.412217890565e-06,  1.029087471798e-05, {   1,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.218125639282e-05,  2.520076287938e-06, {   1,   1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.036392177559e-06, -1.266476388182e-05, {   2,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.257085181799e-05,  2.129974355859e-06, {   0,   0,   0,   0,   0,   1,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.353328775991e-05,  1.166076221192e-06, {   2,   0,   2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  9.565599986959e-06,  5.128770575900e-06, {   2,   0,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.031226708017e-05,  4.379819899681e-06, {   0,   0,   1,   0,   5, -14,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.809184221760e-06,  1.187890187683e-05, {   3,   1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.080238332437e-05, -3.876061158151e-06, {   2,   0,   0,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 {  1.465871562648e-05,  1.461586698921e-08, {   5,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.463887456455e-05, -3.088378548896e-08, {   1,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.553509822606e-06,  1.311619486188e-05, {   2,   0,   2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.813909915805e-06,  4.852364216876e-06, {   0,   0,   2,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.741127871404e-06, -8.922536101077e-06, {   3,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.072083737002e-06, -1.359055995568e-05, {   1,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  5.567416806856e-06, -9.093444868597e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -1.815122947605e-06,  1.284294220991e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   1,   0,   0,   0,   0 } },
 {  1.025956953492e-05,  4.395247347124e-06, {   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  6.798408227041e-06, -7.845593042034e-06, {   0,   0,   2,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  2.735786880939e-06, -1.190759776462e-05, {   1,   1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.797465962130e-06,  1.183111089031e-05, {   3,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.790790915261e-06, -7.836329668066e-06, {   0,   0,   2,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 {  1.169713282030e-05, -2.925419237325e-06, {   5,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.785748917414e-06, -4.835146075883e-06, {   1,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.863961797011e-06, -1.175259969624e-05, {   6,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  9.822899289814e-06, -4.792685165085e-06, {   1,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 {  9.786573261267e-06, -4.828932987458e-06, {   0,   1,  -1,   0,   0,   0,  11, -21,   0,   0,   0,   0,   0 } },
 { -4.277909966701e-06, -1.033022918996e-05, {   2,  -2,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.110293563667e-05, -3.504869059077e-06, {   0,   0,   0,   0,   0,   3,   0,  -8,   3,   0,   0,   0,   0 } },
 {  1.887205983884e-06,  1.272017548949e-05, {   3,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.339340646803e-05, -1.212572493647e-06, {   1,  -1,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.438015026108e-05,  2.242636699137e-07, {   2,  -2,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.394254046386e-06, -1.020944640679e-05, {   4,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -4.344084858191e-06,  1.025931811199e-05, {   2,   0,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.277721823108e-05, -1.821385729875e-06, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -4,   0,   0,   0 } },
 { -9.969920455229e-06, -4.627237883807e-06, {   2,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.276013501324e-06,  8.318656848039e-06, {   0,   2,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.205935122995e-05, -2.530671431172e-06, {   0,   0,   1,   0,   0, -19,  21,   0,   0,   0,   0,   0,   0 } },
 {  9.382863520603e-06, -5.202163363332e-06, {   2,   0,   2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -1.122008274915e-05,  3.363356946133e-06, {   0,   0,   0,   0,   0,   3,  -8,   8,  -3,   0,   0,   0,   0 } },
 { -1.094785247291e-05, -3.634909691610e-06, {   1,   0,  -2,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.702089344723e-06, -1.188034270714e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 {  3.262294104031e-06,  1.131949251891e-05, {   0,   0,   0,   0,   0,  19, -22,   0,   0,   0,   0,   0,   0 } },
 {  2.262383889646e-06,  1.231581224713e-05, {   2,   0,  -2,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.922750262447e-06, -5.653812486370e-06, {   0,   0,   0,   0,   5,   0,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.946172784713e-06, -8.629479716980e-06, {   1,   1,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.814518674133e-06,  5.758992553311e-06, {   2,   0,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -9.735043897717e-06, -4.836155485691e-06, {   0,   0,   0,   0,   0,   2,  -2,   0,  -3,   0,   0,   0,   0 } },
 { -5.392506549757e-06,  9.178511138632e-06, {   0,   2,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.513344943947e-06, -1.005577961469e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 { -3.537251213706e-06,  1.102725288738e-05, {   2,  -2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.298332541402e-05,  1.578345037230e-06, {   1,   1,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.314941294913e-05, -1.406779749356e-06, {   4,   0,  -2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -2.788005796428e-06,  1.176783579245e-05, {   0,   2,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.681758967808e-06, -8.865341140109e-06, {   3,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.077246837804e-07, -1.393736246320e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  6.147766025071e-06, -8.394773941981e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.453235332117e-05,  9.829891788273e-09, {   1,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.530599907803e-06,  1.100543609048e-05, {   2,  -2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.432980943350e-05, -2.042258224244e-07, {   6,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.309173164095e-05, -1.438754247845e-06, {   0,   0,   3,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  8.997708840721e-06,  5.528796776627e-06, {   2,   0,  -1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.279857988677e-05, -1.726542063461e-06, {   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.446341757511e-05,  5.973403864908e-08, {   1,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.449320793464e-05,  2.992657266168e-08, {   2,   0,  -1,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.053434165907e-05, -3.985224437030e-06, {   0,   0,   0,   0,   0,   0,   9, -17,   1,   0,   0,   0,   0 } },
 {  1.045669709818e-05,  4.060606975401e-06, {   0,   0,   1,   0,   0, -27,  31,   0,   0,   0,   0,   0,   0 } },
 { -5.745494146170e-06,  8.771234844805e-06, {   0,   0,   0,   0,   0,   1,   2,  -5,   0,   0,   0,   0,   0 } },
 { -4.831584709112e-06, -9.683970650934e-06, {   0,   0,   0,   0,   0,   8, -13,   0,  -2,   5,   0,   0,   0 } },
 { -9.343919065012e-06, -5.168207691509e-06, {   2,  -2,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -7.365846869245e-07,  1.377001828117e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   1,  -9,   0,   0,   0 } },
 {  1.304721469654e-05, -1.458503304482e-06, {   0,   0,   1,   0,   0,   0,  12, -24,   5,  10,   0,   0,   0 } },
 { -1.304721469654e-05, -1.458503304482e-06, {   0,   0,   1,   0,   0,   0, -12,  24,  -5, -10,   0,   0,   0 } },
 { -3.897108740585e-06, -1.060756128692e-05, {   1,  -1,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.989253221036e-06, -9.514690156291e-06, {   2,   0,  -1,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.410768643699e-05,  3.942827473625e-07, {   2,   0,  -1,   0,   0,  -4,   6,   0,  -2,   0,   0,   0,   0 } },
 { -4.404572291883e-06, -1.009623099953e-05, {   2,   0,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.051021686384e-05, -3.986480485981e-06, {   1,  -1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  2.783260959585e-07, -1.421787502356e-05, {   0,   2,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  7.103986075372e-06, -7.391251676285e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -3,   0,   0,   0 } },
 {  6.184844693538e-06,  8.307463471822e-06, {   1,  -1,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.071393731779e-06, -6.416039564855e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   3,   0,   0,   0 } },
 {  4.779020009051e-06,  9.707117520703e-06, {   1,   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  1.041416514243e-05,  4.069999009523e-06, {   4,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.233586890126e-06,  1.024634839258e-05, {   2,   0,  -1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -8.550464941251e-06, -5.929128102968e-06, {   4,   0,  -2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.451635151568e-06,  1.302633034546e-05, {   0,   0,   0,   0,   0,   8, -13,   0,   2,  -5,   0,   0,   0 } },
 {  3.308639699154e-06,  1.116845327394e-05, {   1,  -1,  -1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.967338273185e-06, -1.150563123234e-05, {   4,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.421004701131e-05, -2.573031704436e-07, {   4,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.261416277842e-05,  1.846522650303e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   3,   0,   0,   0 } },
 {  1.405694151439e-05, -4.024292046874e-07, {   2,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -2.873795072970e-07,  1.416940290553e-05, {   3,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.419017908482e-05, -2.644950401998e-07, {   1,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -8.351765376646e-06,  6.102065154301e-06, {   0,   2,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.271936961065e-05, -1.730634144453e-06, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  9.904991956712e-07,  1.345728158926e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -3.501965684769e-07,  1.409092320696e-05, {   2,   0,  -1,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0 } },
 {  7.531002603413e-06, -6.904946005283e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -6,   2,   0,   0,   0 } },
 { -2.905766949955e-06, -1.152677453899e-05, {   5,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.698561799086e-06,  6.733805013758e-06, {   0,   0,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.170420606496e-05, -2.721237342894e-06, {   0,   0,   1,   0,   0, -17,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.541324641382e-06,  1.287990137422e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.011730268023e-06,  1.340735522785e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  2.628508834740e-06, -1.179049502035e-05, {   6,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.776072251008e-06, -4.636842876571e-06, {   1,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.888438525426e-06, -1.052374454515e-05, {   0,   0,   3,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 {  3.599133798775e-06, -1.081296238533e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  8.651860555815e-07,  1.354426264142e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -7,   0,   0,   0 } },
 {  8.821228119515e-06,  5.584002955689e-06, {   0,   0,   0,   0,   0,   2,  -5,   4,   0,   0,   0,   0,   0 } },
 {  2.748603521923e-06, -1.165567987833e-05, {   1,   1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.084131973524e-06,  5.314633899302e-06, {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.732809008646e-06, -1.166442665448e-05, {   1,  -1,   1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -8.162132321740e-06, -6.231787813645e-06, {   0,   2,  -1,   0,  -5,   0,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.738128028171e-06, -1.165291614150e-05, {   1,  -1,   1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.242187243385e-05,  1.963901736989e-06, {   1,   1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.168260217149e-05,  2.699456204988e-06, {   1,  -1,   0,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.193073967059e-05,  2.450583547933e-06, {   0,   0,   3,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.308660831063e-06,  8.071696071421e-06, {   0,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.793614261787e-06, -1.158664886965e-05, {   0,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.492888490204e-06,  8.878948867436e-06, {   1,   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.228182120069e-05,  2.087714118413e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 {  1.173623548680e-07,  1.425160043669e-05, {   2,   0,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.425728512376e-05, -1.106239303735e-07, {   2,   0,  -4,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  8.055028205953e-06, -6.309782533094e-06, {   2,   0,  -1,   0,   0,  -1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.725878842639e-06, -8.636920316508e-06, {   1,   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.886058422581e-06,  1.247498088140e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.198365034406e-05, -2.377074801950e-06, {   1,  -1,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.006640308767e-05, -4.289542468258e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 { -2.657321463936e-06,  1.169733841145e-05, {   1,  -1,  -1,   0,   0,  24, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.393129306553e-05,  4.205655808740e-07, {   4,   0,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.177478342570e-05, -2.571531988332e-06, {   0,   0,   1,   0,   0, -18,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.861937377210e-06,  9.484334228541e-06, {   2,   0,  -1,   0,   0,   0,  19, -23,   0,   0,   0,   0,   0 } },
 {  1.015239612836e-05,  4.193502375105e-06, {   1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -3.374779960783e-06,  1.096061305389e-05, {   1,   1,  -1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.254578202058e-05, -1.788578992521e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 { -1.426183303336e-05, -7.192681497831e-08, {   1,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  9.601388116059e-06,  4.729202162457e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 { -1.230975916180e-05,  2.015538654986e-06, {   2,  -2,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  2.636909250501e-06,  1.168689146755e-05, {   0,   0,   2,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.657833797151e-06, -1.265390560018e-05, {   2,   0,  -2,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  6.016768785762e-06, -8.294171715168e-06, {   4,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.988430418943e-06,  1.031796527780e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -5.056761528451e-06, -9.249296898010e-06, {   2,   0,  -1,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.139610001807e-05, -2.909809013094e-06, {   1,   1,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.251433299114e-06, -8.051317327576e-06, {   1,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.424566160993e-05,  5.647464599495e-08, {   1,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.870358886217e-06,  1.042682756094e-05, {   1,  -3,   2,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.167067761187e-06,  6.128903979877e-06, {   0,   0,   2,   0,   0, -18,  16,   0,  -2,   5,   0,   0,   0 } },
 {  8.396842407713e-06,  5.898638258725e-06, {   1,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.151111797257e-05, -2.774434814570e-06, {   2,   0,  -2,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -1.201062645243e-05, -2.273866321596e-06, {   3,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.164217820547e-05,  2.641302803653e-06, {   0,   0,   3,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 { -1.421885043646e-05,  6.331832280375e-08, {   2,  -2,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.204985839450e-05, -2.232040310576e-06, {   3,   1,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  1.204985839450e-05, -2.232040310576e-06, {   1,  -1,   1,   0,   0,  20, -12, -15,   0,   0,   0,   0,   0 } },
 {  5.483908904019e-06, -8.793801220776e-06, {   6,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.876238315050e-06,  1.039243256679e-05, {   2,   0,  -2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.043329257524e-06, -1.022067676075e-05, {   1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.052176801501e-07, -1.375826375136e-05, {   2,   0,   2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.205727909274e-05, -2.205832777872e-06, {   1,  -1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.737241979805e-06, -7.512967929767e-06, {   2,   0,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 { -1.165691349806e-05, -2.589116410891e-06, {   1,  -1,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.262508756461e-05,  1.619008422570e-06, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.423451583072e-05,  9.533602606043e-09, {   4,  -2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.417916391103e-05,  6.154666580587e-08, {   4,   0,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.421126899824e-05, -2.568062700320e-08, {   2,   0,  -4,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.899696529210e-06, -4.334717488397e-06, {   2,   0,  -1,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.272603804336e-06,  1.095373011592e-05, {   1,   1,  -1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.776142155326e-06, -9.449913878175e-06, {   4,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  9.484085414867e-06,  4.736865245550e-06, {   4,  -2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.233136381300e-05, -1.887337020975e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -6.348810303502e-06, -7.869176418136e-06, {   1,   1,   0,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.143361952440e-05,  2.776261563050e-06, {   1,  -1,  -3,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.300792434725e-05, -1.200623582523e-06, {   2,  -2,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  1.186892949734e-05,  2.339459368635e-06, {   3,   1,  -3,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.583624863696e-07, -1.334975023371e-05, {   4,   0,  -2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.641493551662e-06, -8.566144616263e-06, {   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.070002113901e-06,  1.013134033978e-05, {   2,   2,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -7.091968906490e-06,  7.106945793346e-06, {   2,   0,  -2,   0,   0,   0,  -2,   7,   0,   0,   0,   0,   0 } },
 {  2.052843397807e-06,  1.214260499423e-05, {   4,   0,  -2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.222386363429e-05, -1.970575784238e-06, {   0,   0,   2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.207953242970e-05, -2.109076172733e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.236844131011e-05,  1.812993888048e-06, {   4,   0,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.417036907780e-05,  9.486547277953e-09, {   2,   0,   2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.385720757180e-05, -3.212015693813e-07, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,   4,  -2,   0,   0 } },
 { -5.431107859432e-06,  8.746893619881e-06, {   1,  -1,   0,   0,   0,   5,  -6,  -2,   0,   0,   0,   0,   0 } },
 {  1.415013933672e-05,  2.692412440362e-08, {   4,   0,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.416866774891e-05,  6.147224507500e-09, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   1,   0,   0,   0 } },
 {  1.416866774891e-05, -6.147224507500e-09, {   2,   0,  -3,   0,   0,  -2,   5,   0,   0,   1,   0,   0,   0 } },
 {  7.058235870148e-06,  7.116019695404e-06, {   2,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.380613449786e-05, -3.677710368184e-07, {   0,   0,   0,   0,   0,   5,  -7,   0,   2,   0,   0,   0,   0 } },
 { -1.416528697185e-05, -3.754423832146e-09, {   2,   2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.059509877428e-06, -5.109076555549e-06, {   4,   0,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.269069382005e-05,  1.470944869805e-06, {   4,   0,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -3.131017818404e-06,  1.103002751039e-05, {   0,   0,   0,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 {  2.692427629923e-06,  1.146638071804e-05, {   1,   1,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.260831257414e-07, -1.363079645769e-05, {   2,   0,  -2,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.008683263226e-06, -1.214365828514e-05, {   0,   2,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.390983037362e-05, -2.422005650440e-07, {   0,   0,   4,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -7.207746646664e-06, -6.937508440380e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,  -2,   0 } },
 {  9.044138528445e-06,  5.100399013000e-06, {   0,   0,   1,   0,   0,   0,  -7,  14,   0,   0,   0,   0,   0 } },
 { -2.720229984791e-06,  1.142381307576e-05, {   2,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.842598407466e-06, -5.300880251088e-06, {   0,   0,   2,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -5.478019427486e-06, -8.660721965264e-06, {   2,   0,  -3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  9.573298925594e-06,  4.563631117485e-06, {   2,   0,   0,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 {  2.789471609830e-06,  1.133609712633e-05, {   3,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.660583279898e-06,  9.462451568630e-06, {   2,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.233901671234e-06,  9.887553736566e-06, {   3,  -3,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -4.233901671234e-06,  9.887553736566e-06, {   1,   3,  -1,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 { -2.689403114120e-06, -1.143063832040e-05, {   0,   1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -9.065486774716e-06,  5.052184195150e-06, {   1,   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.002062890529e-06,  5.114830568184e-06, {   6,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  6.096317379683e-06,  8.009707328522e-06, {   2,   0,   1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  1.361821231943e-05, -4.833698438429e-07, {   6,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.091081713865e-06,  8.002528002093e-06, {   2,   0,   1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -1.376057492380e-05, -3.312276612574e-07, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 { -2.649365199584e-06,  1.144156244293e-05, {   0,   0,   2,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -2.587543502964e-06, -1.150217206317e-05, {   0,   0,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.660175799042e-07, -1.382064880182e-05, {   1,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.383453078541e-05, -2.490176008503e-07, {   3,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -6.301462195358e-06, -7.779869038712e-06, {   0,   0,   0,   0,   0,  10,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.491118928019e-06, -8.582020782691e-06, {   2,   0,   1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -2.158919453633e-06,  1.191138663497e-05, {   0,   0,   2,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.414462960686e-06,  9.655624711922e-06, {   1,  -1,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.016949349991e-05, -3.899738616309e-06, {   2,   0,  -2,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -1.404779138087e-05, -1.946278926181e-08, {   4,  -2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.869638256112e-06, -4.194524946714e-06, {   0,   0,   5,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.869638256112e-06, -4.194524946714e-06, {   0,   0,   3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.038350673057e-06, -1.202354558343e-05, {   1,  -1,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -5.175925282245e-06, -8.884531123287e-06, {   2,   0,  -3,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -2.649662667426e-06,  1.140887227616e-05, {   1,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.312965154357e-05, -9.232467397008e-07, {   1,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.502265486260e-06,  4.549680879743e-06, {   0,   0,   1,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0 } },
 { -9.601495184376e-06,  4.446604328697e-06, {   1,  -1,   0,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.180275985014e-05, -2.244906889645e-06, {   3,   1,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.230434242607e-06, -1.281505302758e-05, {   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  6.173007807370e-06, -7.867272374219e-06, {   4,   0,  -2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.914931667654e-06, -6.111780157726e-06, {   2,   0,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.914931667654e-06, -6.111780157726e-06, {   2,   0,  -2,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.085582635834e-06,  5.941071684471e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  1.003922795230e-05,  3.984368912402e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.117650111182e-05, -2.845297947236e-06, {   1,   1,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.996670690763e-07,  1.351724728685e-05, {   0,   0,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 {  7.947769172880e-06,  6.069016821086e-06, {   0,   0,   2,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.742420523313e-06,  5.273011253303e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,  -4,   0,   0,   0 } },
 {  2.978841945961e-06, -1.103036260239e-05, {   2,   0,  -2,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.814762399809e-06,  4.192222431466e-06, {   2,   0,   0,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 { -9.215787464906e-06,  4.790318652044e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  1.399634928052e-05, -4.631119606791e-09, {   0,   2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.384215406641e-05,  1.587431581314e-07, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 { -4.943554785688e-07,  1.350172589502e-05, {   0,   0,   1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -2.124381772133e-06,  1.186752991237e-05, {   3,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -9.793599248018e-07,  1.300432212251e-05, {   2,   0,  -1,   0,   0,   6,  -4,  -8,   3,   0,   0,   0,   0 } },
 { -2.772879165277e-06, -1.120744934525e-05, {   4,   0,  -3,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.827827066587e-06, -4.152461097461e-06, {   1,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  5.559662037320e-07,  1.341082310690e-05, {   2,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -9.140039002348e-06, -4.826186828419e-06, {   4,   0,  -2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -3.015148215103e-06, -1.094748226244e-05, {   1,  -1,  -1,   0,   0,  27, -30,   0,   0,   0,   0,   0,   0 } },
 { -8.803620197246e-06,  5.154349399068e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -4.049842999664e-06,  9.900723838436e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.461143871202e-06,  1.247738703843e-05, {   1,  -1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.477998260625e-07,  1.378940657181e-05, {   2,   0,   1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.012224871526e-05, -3.811723324803e-06, {   0,   0,   1,   0,   1,   0, -14,  19,  -3,   0,   0,   0,   0 } },
 {  1.012224871526e-05, -3.811723324803e-06, {   0,   0,   1,   0,  -1,   0,  14, -19,   3,   0,   0,   0,   0 } },
 { -1.369178784628e-05,  2.388650109041e-07, {   1,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.662057197394e-06,  9.264535204212e-06, {   0,   0,   1,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  1.050734227178e-05,  3.417335848721e-06, {   0,   0,   1,   0,   0,   0,   8, -16,   2,  10,   0,   0,   0 } },
 { -1.050734227178e-05,  3.417335848721e-06, {   0,   0,   1,   0,   0,   0,  -8,  16,  -2, -10,   0,   0,   0 } },
 { -8.201431881077e-06, -5.721583268646e-06, {   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.281303735180e-06, -7.634459463635e-06, {   4,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.732564121511e-06,  1.017959622650e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   2,   0,   0 } },
 { -3.440662445819e-06, -1.046548120095e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -1.380577250440e-05, -1.003202685869e-07, {   0,   0,   2,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.108755056682e-06,  8.796405663198e-06, {   4,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.020274807601e-05, -3.697748328140e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,  -2,   0,   0,   0 } },
 {  1.231973108625e-05, -1.580397541086e-06, {   0,   0,   0,   0,   0,   0,   8, -13,  -1,   0,   0,   0,   0 } },
 { -4.838211711513e-06,  9.061652256411e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   1,   0,   0,   0,   0 } },
 { -2.647194386967e-06, -1.124822790237e-05, {   1,  -1,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -8.969178358813e-06,  4.921369690512e-06, {   2,  -2,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.164735821915e-05,  2.240037100607e-06, {   1,   1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.096182187089e-06, -1.178787364696e-05, {   2,   0,   2,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.046243231813e-06, -9.836995354548e-06, {   1,   1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.095283269083e-05,  2.928368887350e-06, {   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.140878230168e-06,  9.738585888314e-06, {   4,   0,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.143199668964e-05, -2.447015125990e-06, {   1,   1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.538698168276e-06, -5.337845515491e-06, {   0,   2,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -4.143766942967e-06, -9.730193092843e-06, {   0,   0,   1,   0,   0, -25,  29,  -3,   0,   0,   0,   0,   0 } },
 {  4.523818274628e-06, -9.340145028644e-06, {   1,  -1,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.373363405194e-05,  1.295261748463e-07, {   0,   0,   4,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 { -9.423719403887e-06,  4.438071195641e-06, {   2,   0,  -2,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.399132067654e-06, -4.462224238998e-06, {   4,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.129416475742e-05,  2.565836600108e-06, {   0,   0,   1,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -1.384995714021e-05, -8.842292075071e-09, {   4,  -2,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.256832142146e-05, -1.286488168781e-06, {   2,   0,  -1,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 { -2.639966610302e-06, -1.121123423509e-05, {   1,  -1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.375279715270e-05, -9.380186295608e-08, {   0,   2,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.630445809058e-06,  5.215744747978e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 {  6.226812695086e-07, -1.321450133597e-05, {   2,  -2,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.382745920926e-05, -4.466424408571e-09, {   0,   2,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.641384628283e-06,  1.117795164016e-05, {   3,  -1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.238031104951e-05,  1.437485542825e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0 } },
 { -9.280606574096e-06, -4.532310700862e-06, {   1,   0,   0,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 { -1.146904930786e-05,  2.340489640367e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -4.759251907963e-06,  9.049927426447e-06, {   1,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -9.792151378969e-06, -4.016376746990e-06, {   0,   0,   0,   0,   0,   0,   1,  -2,   3,   0,   0,   0,   0 } },
 {  1.146611054438e-05,  2.340827888833e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.036038993793e-05,  3.445620369821e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.289833152212e-05,  9.049282296650e-07, {   3,   0,  -2,   0,   0, -14,  12,   0,   2,   0,   0,   0,   0 } },
 { -2.648754481089e-06,  1.115388078433e-05, {   3,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.909360617256e-06,  4.884441481570e-06, {   2,   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  4.255615543904e-06, -9.535960687160e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  3.331011415097e-06,  1.045529584456e-05, {   4,   0,  -3,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -3.534049556019e-06,  1.024403930144e-05, {   0,   0,   1,   0,   0, -17,  16,  -3,   0,   0,   0,   0,   0 } },
 {  1.101902877789e-05, -2.755902942038e-06, {   1,   1,  -4,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.175121570193e-06,  4.599016308903e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -4.065716722484e-07, -1.335924544733e-05, {   4,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.280920864000e-06, -8.468297075506e-06, {   2,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.209558102402e-05, -1.653324213677e-06, {   0,   0,   3,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  5.063975882944e-06,  8.658421855588e-06, {   0,   0,   1,   0,   0,   0,  -9,  16,  -6,   0,   0,   0,   0 } },
 {  2.572430222639e-06,  1.114890982000e-05, {   1,   1,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.364661597816e-05,  7.152068171124e-08, {   4,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.629939035002e-06, -1.108782181306e-05, {   1,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.109741716175e-06,  9.607729158009e-06, {   1,   3,  -2,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 {  9.650554532429e-06, -4.061935019276e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 { -1.917004998716e-06, -1.179340774499e-05, {   0,   0,   2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  2.278831937588e-06,  1.143067065388e-05, {   1,   0,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -8.505567904876e-07,  1.285780520856e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 } },
 { -1.307536481546e-05, -6.282670680213e-07, {   2,   0,  -1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.016273079234e-06, -1.068324011887e-05, {   0,   0,   3,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -1.064958081091e-05,  3.049833342694e-06, {   2,   0,  -3,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.702631892623e-06,  9.988755910865e-06, {   2,   0,   2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  8.908848155325e-06, -4.780304093913e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 { -1.368316700622e-05, -3.494542278254e-09, {   4,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -8.906690296513e-06, -4.777572342507e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 { -1.142201135651e-05, -2.257348718022e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  1.810027516671e-06, -1.186693263994e-05, {   0,   0,   1,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  2.508614294161e-06,  1.116646026211e-05, {   1,   1,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.362491598027e-05,  4.669783370009e-08, {   0,   2,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.284900631076e-05,  8.148383988901e-07, {   1,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.350500728021e-05, -1.564785987441e-07, {   0,   0,   3,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  9.910556480518e-07, -1.266226385491e-05, {   2,   0,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  2.960828614828e-06,  1.068795909059e-05, {   0,   0,   1,   0,   0,   0,  -2,   6,  -3,   0,   0,   0,   0 } },
 { -1.363661881716e-05,  9.710024325387e-09, {   2,  -2,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.206125109333e-05,  1.583111164857e-06, {   0,   0,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0 } },
 { -4.264771798913e-06,  9.377328920893e-06, {   2,   2,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.051401766141e-05,  3.123080460815e-06, {   0,   0,   0,   0,   0,   0,   7, -14,   3,   0,   0,   0,   0 } },
 {  3.943654607882e-07, -1.323502691195e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,  -5,   0,   0,   0 } },
 { -1.309733887769e-05, -5.307865313780e-07, {   4,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.991939331603e-07,  1.312820658960e-05, {   2,   0,   0,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.112290551357e-05,  2.503384781361e-06, {   0,   2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.357004890530e-05, -5.609199474535e-08, {   5,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  8.872294585006e-07, -1.273663702717e-05, {   2,   0,   2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.338348918407e-05,  2.402192158214e-07, {   2,   0,  -1,   0,   0,   0,   6, -15,   0,   0,   0,   0,   0 } },
 {  1.001257156653e-06,  1.261610724437e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -6.696737272368e-07, -1.294710976898e-05, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  9.253142207875e-06, -4.360707429229e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   8,   0,   0,   0 } },
 {  1.039740583164e-05,  3.214843868284e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.148409778695e-05, -2.126032709294e-06, {   3,   1,  -2,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  1.148409778695e-05, -2.126032709294e-06, {   1,  -1,   0,   0,   0,  20, -12, -15,   0,   0,   0,   0,   0 } },
 {  1.162720445868e-06,  1.243577605331e-05, {   0,   0,   0,   0,   0,  20, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.157014333006e-05,  2.024887247868e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 { -1.348600997557e-05, -9.774219848817e-08, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 { -1.352814052697e-05,  4.406186150965e-08, {   1,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.355317423918e-05, -1.877893532663e-08, {   2,   0,   3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.574279587164e-06,  3.996930745651e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.329916167717e-05, -2.711464539908e-07, {   2,   0,   0,   0,   0,   3,  -6,   0,   1,   0,   0,   0,   0 } },
 {  9.319529226873e-06,  4.250094718064e-06, {   0,   0,   1,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0 } },
 { -8.099042624053e-07,  1.275374859352e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.621184705604e-06, -1.093180019054e-05, {   1,  -3,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.609622485704e-06, -9.940769268389e-06, {   0,   1,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  4.722735202226e-08,  1.350010398045e-05, {   2,   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  1.342310803956e-05,  1.218142135508e-07, {   4,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.259992119379e-07, -1.291295155900e-05, {   4,   0,  -2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -4.257038667081e-06, -9.262819603578e-06, {   0,   2,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -3.541802972685e-06, -9.976323654299e-06, {   1,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  3.418281372059e-08, -1.348258615598e-05, {   4,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.347606718945e-05,  3.882207664135e-08, {   4,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.540727126993e-06, -9.970780663924e-06, {   1,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -3.095119530783e-06, -1.041582094884e-05, {   0,   0,   1,   0,   0,  -3,  -1,   8,  -3,   0,   0,   0,   0 } },
 {  1.299537467231e-05,  5.146424378619e-07, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  1.190803757261e-05, -1.591073158967e-06, {   2,  -2,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.204658193789e-06, -9.284925923245e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  4.001265485773e-06, -9.485772947492e-06, {   2,   2,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  4.318302797676e-06,  9.167638740083e-06, {   1,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -2.644225768487e-06, -1.084169097699e-05, {   2,   0,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.466120630638e-07,  1.323289632883e-05, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.334160006089e-05,  1.349358796280e-07, {   2,   0,  -1,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -2.577100067623e-06, -1.087897348988e-05, {   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.279478452015e-06,  4.176221332340e-06, {   2,   0,   0,   0,  -1,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -1.288299621217e-05, -5.681719931764e-07, {   3,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.166183050955e-05, -1.789123550828e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -9.638425486533e-06,  3.810521146037e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  4.001187159388e-06,  9.444217339053e-06, {   1,  -1,   1,   0,   0, -18,  17,   0,   0,   2,   0,   0,   0 } },
 {  1.059847761685e-06, -1.238283248125e-05, {   2,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.334555686232e-05, -9.208349157029e-08, {   1,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.071204090256e-05, -2.724841330923e-06, {   1,   1,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.063920063998e-05,  2.797120277952e-06, {   4,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  1.063920063998e-05,  2.797120277952e-06, {   0,   0,   2,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 { -1.181622953545e-06, -1.225358676823e-05, {   2,   0,  -2,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -3.529344938961e-06,  9.902907373815e-06, {   1,   1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.176795731597e-05, -1.658024220238e-06, {   0,   0,   1,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  3.791403373169e-06, -9.631600703054e-06, {   0,   0,   0,   0,   0,   2,  -9,  11,   0,   0,   0,   0,   0 } },
 {  1.099974123843e-05,  2.417018466732e-06, {   3,   0,  -5,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.013569044375e-06, -1.239924522088e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 { -3.050649791097e-06,  1.036208514773e-05, {   0,   0,   0,   0,   0,   1,   3,  -8,   3,   0,   0,   0,   0 } },
 { -8.430093505698e-07,  1.256727141419e-05, {   1,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.227459811793e-05, -1.128865784501e-06, {   2,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.793024337207e-07, -1.311848941543e-05, {   3,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.423865691700e-06, -1.196669904734e-05, {   2,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.241053540870e-06, -8.145779325282e-06, {   2,   0,  -1,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.505872507620e-07, -1.283566272741e-05, {   0,   0,   3,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.501530913182e-06,  1.088345736373e-05, {   1,  -1,  -3,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.557069588653e-06,  1.081843574366e-05, {   1,  -1,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.054753751077e-05,  2.823863203491e-06, {   2,   0,   0,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  5.471461018167e-07,  1.280878616000e-05, {   2,   0,   2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.332844158955e-06, -9.019933434603e-06, {   1,   1,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.168055113985e-05, -1.671746211470e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -4,  -5,   0,   0,   0 } },
 { -1.394712265487e-07,  1.321241981191e-05, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.323965084093e-05,  1.105003648583e-07, {   1,   0,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.130594419865e-05,  2.040405916128e-06, {   0,   0,   0,   0,   0,   4,  -6,   0,   3,   0,   0,   0,   0 } },
 { -1.304481396716e-05,  3.009693706903e-07, {   2,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.332713388647e-05,  1.756942247146e-08, {   1,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.058639136549e-06,  1.028447783993e-05, {   0,   0,   0,   0,   0,   1,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.161265106358e-05, -1.730222540782e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -1.289073034017e-06, -1.205173783816e-05, {   1,  -1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.081520117100e-05,  2.522320017202e-06, {   1,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.121760437683e-05,  2.116595231512e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 {  1.667938277457e-06,  1.166568856403e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 { -9.702969642909e-06, -3.628426010659e-06, {   1,  -1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.418798374461e-06, -3.909917804256e-06, {   0,   2,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.035276056476e-06, -1.028948251274e-05, {   0,   0,   1,   0,   0,  -3,   7,  -8,   3,   0,   0,   0,   0 } },
 { -2.949905094722e-07, -1.302542341236e-05, {   0,   0,   3,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.235776900848e-05, -9.552299872620e-07, {   0,   0,   2,   0,   0,  -8,   3,  -7,   8,  -6,   0,   0,   0 } },
 {  1.235776900848e-05, -9.552299872620e-07, {   0,   0,   0,   0,   0,   8,  -3,   7,  -8,   6,   0,   0,   0 } },
 {  9.258250870965e-06,  4.048320765947e-06, {   3,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.092276226170e-05,  2.382166430176e-06, {   1,  -1,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 { -4.847236687762e-07,  1.281857530303e-05, {   0,   2,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  2.552931217101e-06,  1.075015011494e-05, {   2,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.006371253430e-06,  9.291163275944e-06, {   1,  -1,  -1,   0,   0,  31, -36,   0,   0,   0,   0,   0,   0 } },
 { -2.690841394516e-06, -1.060274028356e-05, {   2,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.893077411310e-09,  1.328408655147e-05, {   1,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.166656019112e-06,  9.124107183466e-06, {   3,   0,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.996417569014e-06, -1.129066412348e-05, {   0,   0,   2,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 { -5.692856924740e-07, -1.271498514173e-05, {   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.493863608756e-06, -1.077924168611e-05, {   3,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.632541138489e-06,  3.637384001081e-06, {   1,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.440162362772e-06,  9.829307420358e-06, {   1,  -1,  -1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.326007907206e-05, -5.792527328860e-09, {   6,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.325290887692e-05, -1.151609311172e-08, {   1,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.262336555059e-05, -6.389856920502e-07, {   2,   0,   0,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.312393752064e-05, -1.381572023118e-07, {   0,   2,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.084695962863e-05, -2.413512278033e-06, {   1,  -1,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.139079345013e-06, -1.112075131192e-05, {   1,  -1,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -3.676082334575e-06, -9.579553671215e-06, {   1,   0,   0,   0,   0, -27,  30,   3,   0,   0,   0,   0,   0 } },
 { -2.066698122349e-06, -1.118877486477e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.167204353593e-05, -1.581897450441e-06, {   1,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  1.002113153892e-05,  3.231535627480e-06, {   2,   0,  -2,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.096955308191e-06, -4.152758615760e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.254752189286e-05, -6.911642645689e-07, {   2,   0,   0,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 { -1.321280592921e-05, -2.498110079081e-08, {   4,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.126753013921e-05, -1.965184089439e-06, {   1,  -1,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.493694046640e-06,  9.736809629664e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   2,  -2,   0,   0,   0 } },
 { -1.318431929196e-05, -4.439744571428e-08, {   0,   0,   2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  2.631289359416e-06, -1.059060555536e-05, {   4,   0,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.410549269260e-06, -4.808447718401e-06, {   2,   0,  -1,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  9.124405734721e-06,  4.092092372881e-06, {   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.499693118009e-06,  1.171243361143e-05, {   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.128028170131e-06,  4.084005598097e-06, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.093759351033e-05,  2.259197603550e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -6,   5,   0,   0,   0 } },
 {  1.106244565175e-05,  2.132836370701e-06, {   2,   0,   1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -1.083554283548e-05, -2.357047061854e-06, {   1,  -1,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.138892707131e-06,  1.005327284439e-05, {   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.190679440674e-05,  1.283779727404e-06, {   1,   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -5.105351190710e-06, -8.063755536490e-06, {   2,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.232189873655e-06, -1.093213509709e-05, {   2,   0,  -1,   0,   0,   0,  -5,  10,  -4,   0,   0,   0,   0 } },
 { -1.316253534322e-05,  1.608386467870e-09, {   0,   2,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.314400177948e-05,  1.430118217987e-08, {   3,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.517040883292e-06,  1.064050961869e-05, {   3,  -1,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.197261297162e-05,  1.177588865534e-06, {   2,   0,  -1,   0,   0,   0,  17, -23,   0,   0,   0,   0,   0 } },
 {  1.285688987542e-05, -2.926598418953e-07, {   1,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.244564538792e-05,  6.989849724747e-07, {   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.047561903647e-05, -2.667010750992e-06, {   2,  -2,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  6.768118452991e-07, -1.246561915527e-05, {   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.298861507408e-06, -3.843291904137e-06, {   0,   2,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  2.792628488438e-06, -1.034841274078e-05, {   0,   2,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.888988684448e-06, -1.024551069692e-05, {   0,   0,   2,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  4.069473607350e-06, -9.060144941430e-06, {   1,   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  1.282552890709e-05,  2.822708203191e-07, {   2,   0,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -2.483375591767e-06,  1.061444282005e-05, {   1,  -1,   1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.045117394053e-05, -2.645867210742e-06, {   1,   1,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.174652254313e-07, -1.247725979959e-05, {   2,   0,  -1,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.011258836855e-05,  2.976833931702e-06, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  2.912584280428e-06, -1.017480670835e-05, {   1,  -1,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.307526582816e-05, -8.308694182811e-09, {   6,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.505958000871e-06, -1.057300345513e-05, {   5,  -1,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.446002153238e-07, -1.283079592941e-05, {   1,   0,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.053472751802e-05, -2.537419917086e-06, {   4,   0,  -3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.016978664890e-06, -1.205400194715e-05, {   1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.305112873960e-05,  1.705449535598e-08, {   0,   2,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.118253802057e-05,  1.884702142328e-06, {   2,   0,   0,   0,   0, -49,  55,   0,   0,   0,   0,   0,   0 } },
 {  2.132187639042e-07,  1.285163682170e-05, {   1,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.074546487836e-05,  2.305019442474e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -8,   5,   0,   0,   0 } },
 {  1.260355536893e-05, -4.436414657280e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0 } },
 {  9.383357902988e-06,  3.662511387493e-06, {   4,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -9.256091528296e-06, -3.784807738817e-06, {   2,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.382958016375e-06, -1.065567250122e-05, {   2,   0,  -1,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.844977035761e-07,  1.245353034063e-05, {   1,  -1,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.274088679004e-05, -2.971118553601e-07, {   2,   0,  -3,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 { -8.042410924693e-07,  1.223290091801e-05, {   2,   0,  -2,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  8.502942670002e-07, -1.217120948046e-05, {   2,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  2.545082878724e-06,  1.047340728116e-05, {   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.273061986538e-06,  3.743036386922e-06, {   0,   0,   2,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.297275594193e-05, -2.496557479026e-08, {   3,  -2,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.072307533592e-05, -2.269341795999e-06, {   1,  -1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.276038829207e-05,  2.297394390294e-07, {   2,  -2,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  1.282250403389e-05,  1.636592716086e-07, {   2,   0,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -1.141181314136e-05, -1.571428151620e-06, {   2,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.590280053216e-06, -9.392454133212e-06, {   1,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.122860433668e-05, -1.752363787712e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -1.067188364533e-05,  2.306122853559e-06, {   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.886634962910e-06, -1.008732527334e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,   5,   0,   0,   0 } },
 {  1.076244150850e-05, -2.207870514395e-06, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   1,   0,   0,   0 } },
 {  1.039459447235e-05, -2.574865480506e-06, {   1,   1,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.196057329091e-05,  1.007195398988e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  1.249132554818e-05, -4.732126838184e-07, {   1,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.075558834198e-05, -2.207975669121e-06, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,  -1,   0,   0,   0 } },
 {  6.732094500237e-08,  1.289381777880e-05, {   1,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.391309253630e-06, -9.563180337014e-06, {   2,   0,  -1,   0,   0,   0,  -4,  11,  -1,   0,   0,   0,   0 } },
 { -3.391309253630e-06,  9.563180337014e-06, {   2,   0,  -3,   0,   0,   0,  -4,  11,  -1,   0,   0,   0,   0 } },
 {  4.441887333475e-07, -1.250323602417e-05, {   2,   0,  -3,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.047786243651e-05, -2.467397420564e-06, {   1,   1,   0,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -7.550669031912e-07,  1.217925318142e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.829357044380e-06,  3.103750904342e-06, {   5,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.251356416064e-05, -4.194124874784e-07, {   2,   0,   1,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 { -1.000395327150e-05,  2.920346461199e-06, {   1,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.858037595423e-06, -3.052657495931e-06, {   1,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.674405379724e-06,  1.123601589880e-05, {   2,   0,  -2,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 { -1.058550149712e-05, -2.323152933581e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   1,   1,   0,   0,   0 } },
 {  1.353244727293e-06,  1.154487864453e-05, {   0,   0,   0,   0,   0,   4,  -6,   0,   1,   0,   0,   0,   0 } },
 {  1.179556644211e-05, -1.102032233601e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -1.065664495550e-05,  2.240910677534e-06, {   6,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.056928914874e-05, -2.319679253620e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,  -1,  -1,   0,   0,   0 } },
 { -2.493810918180e-06, -1.039321272811e-05, {   1,   1,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.414141667550e-06,  1.046990962957e-05, {   1,  -1,  -2,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.015693597411e-05, -2.725051335236e-06, {   2,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.268950961614e-05, -1.817302913865e-07, {   2,   0,   0,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.244693826538e-05,  4.168072613090e-07, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  9.415546202667e-06,  3.440366288758e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 { -2.851171607671e-06, -1.000105852043e-05, {   2,   0,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0 } },
 {  1.284155479582e-05, -2.929753549193e-10, {   3,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.795288866076e-07, -1.236208992009e-05, {   4,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.412971400945e-08, -1.280950428345e-05, {   0,   0,   0,   0,   0,   4,  -7,   0,   8,  -5,   0,   0,   0 } },
 {  2.411327057796e-06,  1.042127535818e-05, {   1,  -3,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.204380444937e-06, -1.062283002336e-05, {   0,   0,   1,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.053115701205e-05, -2.294302792257e-06, {   0,   0,   2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.217300299823e-05, -6.497631193082e-07, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  1.150674670388e-06, -1.166883357420e-05, {   0,   0,   0,   0,   0,  11,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.467074878306e-07, -1.225651942166e-05, {   0,   2,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.875282159197e-06, -1.092378268995e-05, {   4,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.020687293673e-05, -2.586525239862e-06, {   1,   1,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.750283282424e-06,  1.003825457795e-05, {   1,  -1,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.835169078054e-06,  2.947790947989e-06, {   0,   0,   0,   0,   0,   0,   3,  -8,   5,  -5,   0,   0,   0 } },
 { -1.095184477895e-05,  1.825539392692e-06, {   4,   0,  -2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.986000234053e-06,  9.783004201062e-06, {   1,  -1,  -1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.060391896226e-05, -2.163540759977e-06, {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.088925974035e-06,  1.167576291226e-05, {   2,   0,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.419017460893e-06,  1.034317416618e-05, {   2,   2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.018705578782e-05,  2.573596577272e-06, {   1,  -1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.250564642335e-05, -2.432266561901e-07, {   2,   0,  -1,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.045470129183e-05,  2.284932192940e-06, {   1,   0,   0,   0,   0, -14,  10,   0,   4,   0,   0,   0,   0 } },
 { -1.065751984929e-05, -2.081796971858e-06, {   0,   0,   0,   0,   0,   0,   2,  -4,   4,  -5,   0,   0,   0 } },
 { -1.062075103836e-05,  2.110130459159e-06, {   1,  -1,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.067261583022e-05, -2.055414436582e-06, {   1,   1,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.515495395072e-06,  1.021055901134e-05, {   2,   2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.056548087299e-06, -1.066652164754e-05, {   4,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.234782488023e-05, -3.750411524105e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 { -2.576303805184e-08, -1.269135819066e-05, {   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.771213191283e-06,  9.940607336209e-06, {   2,  -2,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.368495567793e-06,  1.133857113795e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.054281037269e-05, -2.161130057203e-06, {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.879914259758e-06,  9.817884571274e-06, {   1,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.064761937581e-05,  2.048240440844e-06, {   1,  -1,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.029306792453e-05,  2.400215790080e-06, {   1,  -1,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.156780880769e-05,  1.118136376094e-06, {   2,  -2,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.427839233518e-06, -1.025136477563e-05, {   1,   1,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.218311612724e-05, -4.956841115811e-07, {   4,   0,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.876111479278e-06,  9.802439175513e-06, {   1,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.254296053915e-05,  1.276051381205e-07, {   2,   0,  -1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.034947883512e-05, -2.317923508613e-06, {   1,  -1,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.042848601175e-06, -1.062016749471e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  1.941875056839e-06, -1.071349556693e-05, {   2,   0,  -1,   0,   0,   0,  -9,  16,  -6,   0,   0,   0,   0 } },
 { -2.558994309927e-06,  1.009591942552e-05, {   2,   0,  -1,   0,   0,   0,   7, -16,   6,   0,   0,   0,   0 } },
 {  1.563702728019e-06,  1.108917666734e-05, {   2,   0,   1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 { -2.943929808358e-06,  9.695066292143e-06, {   4,   2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.186062947346e-06, -1.044914055393e-05, {   2,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.941299609057e-06,  9.687154794223e-06, {   4,   2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.043281586483e-07, -1.212118067436e-05, {   4,  -2,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.914068109685e-06, -2.709450931376e-06, {   2,   0,   0,   0,   0,   0,  -5,   0,   7,   0,   0,   0,   0 } },
 { -2.413852371339e-06,  1.020727507462e-05, {   1,  -1,   2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.072507321369e-05, -1.894745251188e-06, {   1,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.916473331720e-06, -9.702371602864e-06, {   2,  -2,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0 } },
 { -2.400095834418e-06, -1.021362748259e-05, {   0,   1,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -9.135762367660e-06,  3.472809854669e-06, {   1,   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.042135700114e-05,  2.186895807652e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  1.251806594259e-05,  8.382233451574e-08, {   0,   2,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.727808180129e-06,  9.870629254206e-06, {   0,   0,   1,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.671686676392e-06, -9.923823581270e-06, {   1,   1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.189773759106e-05, -6.965070651276e-07, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   2,   0,   0,   0 } },
 { -1.008458427824e-05, -2.505545860826e-06, {   3,  -1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -9.928403178386e-06,  2.659352592838e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 } },
 {  9.928403178386e-06,  2.659352592838e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0 } },
 { -8.971036913755e-07,  1.168652267876e-05, {   4,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.719981825770e-07,  1.220474618403e-05, {   2,   2,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.825380326954e-06,  9.749717817860e-06, {   3,  -1,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.150386174081e-05,  1.069195014632e-06, {   4,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -1.125817745664e-05,  1.314561544478e-06, {   0,   0,   1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  1.046841766814e-05,  2.101167579555e-06, {   1,   1,  -1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.285034031973e-07,  1.242913021455e-05, {   4,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.520280130089e-07,  1.220083575152e-05, {   2,   0,  -2,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  1.199627339784e-05, -5.383145261155e-07, {   2,  -2,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  2.756541696421e-06, -9.775940868381e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   9,   0,   0,   0 } },
 {  1.239166307434e-05,  1.383977704170e-07, {   0,   0,   4,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.363429034195e-06, -1.016430886038e-05, {   0,   1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.256798588091e-06, -1.027037917253e-05, {   3,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.324155349304e-07, -1.219144040260e-05, {   2,   0,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0 } },
 { -2.839181714391e-07, -1.223241741360e-05, {   4,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.229054506211e-05,  2.241613020200e-07, {   2,  -2,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.245186824381e-05,  5.585833866023e-08, {   1,   0,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.250174330890e-05,  2.414949265447e-09, {   0,   0,   0,   0,   0,   2,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.659399708687e-06,  1.084281474225e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.407196801537e-06, -1.108983358407e-05, {   0,   0,   0,   0,   0,   5, -10,   4,  -3,   0,   0,   0,   0 } },
 {  1.242228984132e-05, -7.406134523122e-08, {   0,   2,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.130576421707e-05, -1.188211496448e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   3,   0,   0 } },
 { -1.132231415041e-05, -1.171125407631e-06, {   4,   0,  -2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  1.248530822480e-05,  7.627290722973e-09, {   0,   2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.100832646539e-05,  1.479748131224e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.412853915533e-06, -1.007065035568e-05, {   1,  -3,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.247751875707e-05, -2.260665499696e-09, {   4,   2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.033343377416e-06, -1.043494044019e-05, {   2,   0,  -2,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  2.438798459555e-07, -1.222072644185e-05, {   0,   0,   1,   0,   0, -18,   6,  19,   0,  -3,   0,   0,   0 } },
 {  2.536499711451e-06,  9.923184637952e-06, {   2,  -2,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.242065259159e-05, -3.348731834552e-08, {   3,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.018692839728e-05, -2.261528459011e-06, {   1,  -1,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.139774840591e-05,  1.049142223931e-06, {   0,   0,   0,   0,   0,   0,   5,  -9,  -2,   0,   0,   0,   0 } },
 {  9.828405809448e-06,  2.618443579795e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0 } },
 { -7.494283450122e-07, -1.169614619258e-05, {   1,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  2.393646929357e-06, -1.004634792364e-05, {   3,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.543839831709e-07,  1.218537858802e-05, {   2,   0,  -2,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.799824067823e-06, -1.062833222924e-05, {   2,   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 {  2.383849791159e-06, -1.004139411647e-05, {   1,   1,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.221064159410e-06, -1.019698538811e-05, {   4,   0,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -2.526038424764e-06,  9.890931092010e-06, {   2,   0,  -1,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0 } },
 {  1.206024296905e-05,  3.512334208451e-07, {   1,   0,   2,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 {  1.206024296905e-05,  3.512334208451e-07, {   1,   0,   0,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 {  1.200407765730e-06,  1.121015564473e-05, {   2,   0,   2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.235388694962e-05, -5.593294123132e-08, {   4,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.173346441819e-06, -1.123458308093e-05, {   2,   0,  -1,   0,   0,   2,   5, -13,   8,  -6,   0,   0,   0 } },
 { -9.769668850857e-06,  2.634002264502e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 {  1.235039813058e-05, -5.068407094110e-08, {   2,   0,  -1,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.429233898843e-07,  1.225391790112e-05, {   1,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.186684805043e-05, -5.246728513330e-07, {   0,   0,   2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -1.238760358911e-05,  2.030125503442e-09, {   0,   0,   3,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.159666212136e-05,  7.888735998555e-07, {   2,   0,  -1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.231701089662e-05, -6.700237435055e-08, {   0,   2,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  9.721786510511e-06, -2.656560077613e-06, {   2,   0,   3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.011719940396e-05, -2.258326304337e-06, {   2,  -2,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.075977675677e-06, -1.027975542471e-05, {   2,   0,  -1,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 {  1.185683960717e-05,  4.859943787669e-07, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  4.457743902187e-07, -1.189134523305e-05, {   2,  -2,  -1,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 {  1.059439922589e-05,  1.740938383434e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.901525163026e-07,  1.194370032091e-05, {   1,   0,  -2,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.116001652129e-05, -1.169048143546e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   3,   0,   0 } },
 { -1.167970519776e-05, -6.416939738326e-07, {   3,   0,   0,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.651583748141e-07,  1.175609784823e-05, {   3,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.084376654300e-05, -1.477306964837e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 { -1.179327601053e-05,  5.258277931736e-07, {   2,   0,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.226370207291e-05,  4.760243066504e-08, {   0,   2,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.214278646028e-05,  1.595333885727e-07, {   2,   0,  -1,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.137458392619e-05,  9.165261163152e-07, {   0,   0,   1,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 { -9.359342218412e-07,  1.135433969026e-05, {   1,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.132514095493e-05,  9.634391588730e-07, {   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.228366915154e-05, -4.335372081975e-09, {   2,   2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.010897552086e-05,  2.176494942946e-06, {   0,   0,   1,   0,   0,   0,   3,  -9,   0,   0,   0,   0,   0 } },
 { -1.228086652603e-05,  1.573236413559e-09, {   1,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.383263512974e-06,  9.890556200621e-06, {   1,   1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.788740435653e-07, -1.209401575529e-05, {   4,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.377443770325e-07, -1.203451502223e-05, {   1,   0,  -4,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.646114377369e-07,  1.180672010746e-05, {   2,   0,  -1,   0,   0,   1,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.646114377369e-07, -1.180672010746e-05, {   2,   0,  -3,   0,   0,   1,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.226695359944e-05,  4.286379315796e-09, {   4,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.204266225534e-05,  2.203814666976e-07, {   4,   0,  -2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  2.477227999021e-06, -9.785080896709e-06, {   1,  -1,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.414363550462e-06, -9.844688801770e-06, {   2,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.903464771415e-07, -1.186635016783e-05, {   6,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.222635967216e-05,  2.382817213211e-08, {   2,   0,   0,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.990181805975e-07,  1.204588860781e-05, {   1,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.219283226275e-05,  4.769796388690e-08, {   4,   0,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.151242195460e-06, -1.107744017153e-05, {   2,   2,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.493082878659e-06,  9.734916311868e-06, {   2,   1,   0,   0,   0, -51,  58,   0,   0,   0,   0,   0,   0 } },
 { -2.493082878659e-06,  9.734916311868e-06, {   2,   1,  -2,   0,   0, -51,  58,   0,   0,   0,   0,   0,   0 } },
 {  1.222645470083e-05,  1.441607396753e-09, {   4,  -2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.027398540461e-05, -1.946001101509e-06, {   2,   0,   3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.173672882650e-05,  4.790156017741e-07, {   1,   1,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.893649800749e-06,  1.031545737316e-05, {   2,   0,  -1,   0,   0, -22,  24,   0,   3,   0,   0,   0,   0 } },
 {  1.219328124197e-05,  1.297537756235e-08, {   1,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -9.974939243795e-06, -2.229800939472e-06, {   1,  -1,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.045739632732e-05,  1.744222501909e-06, {   2,   2,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  6.606255330811e-07,  1.154037676363e-05, {   2,   0,  -3,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.218961637779e-05, -8.440189864431e-09, {   3,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.213963354232e-05,  5.615374033246e-08, {   4,   0,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.119141884348e-07,  1.147591113969e-05, {   0,   2,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.206492567685e-05, -1.217637843432e-07, {   2,   2,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  9.927702192467e-07, -1.119108490525e-05, {   2,  -2,  -3,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.047506628934e-05, -1.691561351691e-06, {   4,   2,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.069362759850e-05, -1.470990543215e-06, {   0,   0,   2,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 {  1.005832719040e-05,  2.090759968659e-06, {   1,   1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.351023926240e-07,  1.131309757003e-05, {   2,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.159626758956e-05,  5.505233997310e-07, {   1,  -1,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.672907900167e-07,  1.127302845406e-05, {   3,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.073379059554e-05, -1.406408375598e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.094027348225e-05, -1.199147705377e-06, {   2,   0,  -3,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.212105381552e-05,  1.297267937213e-08, {   0,   0,   2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.210647749826e-05, -2.581972600474e-08, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -3,   0,   0,   0 } },
 { -5.931983483410e-06, -6.192693974632e-06, {   2,   0,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.394611688159e-06, -1.072547263863e-05, {   2,   0,   3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  9.997357480978e-06,  2.116842213556e-06, {   3,  -1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.006607563836e-05,  2.039017640893e-06, {   0,   0,   0,   0,   0,   4,  -7,   1,   0,   0,   0,   0,   0 } },
 {  1.373323692063e-06, -1.072614865281e-05, {   4,   0,  -3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.237741337384e-06, -9.854797645565e-06, {   3,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.294879224812e-06,  9.791594991863e-06, {   1,   1,   1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.057512021942e-05, -1.489532528835e-06, {   2,   0,  -2,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.005155213985e-05,  2.006607121357e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.052159207837e-05, -1.515634058976e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -6,   2,   0,   0,   0 } },
 {  2.349315092083e-07,  1.180072107539e-05, {   2,   0,  -2,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.073146392606e-05, -1.297665555562e-06, {   1,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.144714948888e-07, -1.141307730900e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -3.349574976200e-07,  1.169180248644e-05, {   2,   0,  -4,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.368966664601e-07, -1.118284196274e-05, {   2,   0,  -1,   0,   0,   5, -10,   0,   0,   0,   0,   0,   0 } },
 {  9.371055218709e-07, -1.107571475455e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   2,   0 } },
 {  1.094629330550e-05,  1.056550693843e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -2,  -5,   0,   0,   0 } },
 {  1.605065142569e-07,  1.183767407870e-05, {   0,   2,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.163544185762e-05, -3.610371617781e-07, {   2,   0,  -1,   0,   0,  -2,   6,  -1,  -8,   6,   0,   0,   0 } },
 {  1.182225482616e-05,  1.679437451249e-07, {   3,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.106581448823e-05,  9.215038751441e-07, {   2,   0,   1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.774383553778e-07, -1.150724818933e-05, {   2,  -2,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -8.440511207380e-06, -3.528912951166e-06, {   1,  -1,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.005997293188e-05, -1.906137970982e-06, {   2,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.045817442120e-06, -7.918192197560e-06, {   1,  -1,   0,   0,   0,   0,  17, -30,   0,   0,   0,   0,   0 } },
 {  1.189613708278e-05, -6.775954137259e-08, {   5,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.586190673084e-06, -1.037649462671e-05, {   2,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.193136834756e-05, -1.354393157730e-08, {   2,   2,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.193362119848e-05,  6.629808230510e-09, {   3,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.157674044631e-06, -9.773256973073e-06, {   4,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  1.049605511295e-05,  1.420910128011e-06, {   4,   0,  -3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.180792032522e-05, -1.031273300560e-07, {   4,   2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.965928580440e-06, -1.941320027397e-06, {   1,   1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.048922079450e-05,  1.406950954106e-06, {   2,   0,  -1,   0,   0, -18,  18,   0,  -2,   0,   0,   0,   0 } },
 { -3.287568807011e-07, -1.156508586546e-05, {   2,  -2,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.473625723921e-07, -1.114429300048e-05, {   3,   0,  -2,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  1.793978225360e-06,  1.008843066036e-05, {   4,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.163220356768e-05, -2.483937193338e-07, {   0,   0,   1,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  1.183591804469e-05,  3.675828341934e-08, {   0,   0,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 {  1.026179784323e-05,  1.606467230332e-06, {   1,   1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -1.010152591349e-05,  1.764843067110e-06, {   3,  -1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.411530578690e-08,  1.178125576598e-05, {   1,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.045581034961e-05,  1.408664605727e-06, {   4,   0,  -3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.262895440346e-06,  1.059864500283e-05, {   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.018753281929e-05, -1.670269155783e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  1.016574513464e-05, -1.688028387551e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 {  1.183788311886e-05,  1.478158997306e-08, {   4,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.438446061942e-07, -1.160757533405e-05, {   2,   0,  -2,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.020729947774e-07,  1.104595708221e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,  -1,   0,   0,   0,   0 } },
 { -1.004084579077e-05,  1.804063891145e-06, {   3,  -2,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.817030522486e-07, -1.166315680387e-05, {   0,   0,   3,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.608662000310e-06,  1.022426327406e-05, {   2,  -1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.014125672522e-05, -1.690907699265e-06, {   0,   0,   2,   0,   0, -26,  37, -16,   4,   5,   0,   0,   0 } },
 {  1.002240818134e-05, -1.804152345510e-06, {   2,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.179083267935e-05, -2.863582548159e-08, {   2,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 {  1.163020949603e-05, -1.849029292778e-07, {   4,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.132816677174e-05,  4.817703966554e-07, {   0,   0,   0,   0,   0,   0,   4,   0,   0,  -4,   0,   0,   0 } },
 { -1.792305094367e-06, -1.001722150305e-05, {   0,   0,   3,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -1.010980610575e-05,  1.698228847714e-06, {   0,   0,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0 } },
 {  1.092395193716e-05,  8.735922826248e-07, {   2,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.157459709756e-05,  2.213221028003e-07, {   2,   0,  -2,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.173152931534e-06,  1.061916725158e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.067920944951e-07,  1.158356154526e-05, {   1,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.481367144973e-07,  1.093510990139e-05, {   1,  -1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -4.268224013710e-07, -1.135089023898e-05, {   0,   0,   1,   0,   0, -26,  27,   0,   0,   0,   0,   0,   0 } },
 { -1.477760291003e-06,  1.029258487944e-05, {   2,   0,   1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -9.791915229171e-07,  1.079102380952e-05, {   2,   0,  -1,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 {  4.311140341734e-07,  1.133874770855e-05, {   0,   2,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.674666562283e-06, -1.008847872942e-05, {   1,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.161031878369e-05, -1.441007414527e-07, {   2,   0,  -1,   0,   0,  -2,   4,   0,  -2,   0,   0,   0,   0 } },
 {  1.369548127408e-06, -1.038330566473e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.127878949348e-05, -4.734307271398e-07, {   2,   0,  -2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -1.169822378417e-05, -5.070247875059e-08, {   1,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -8.734520338180e-06,  3.005118548741e-06, {   1,  -1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  6.880576417012e-07, -1.104261771036e-05, {   2,  -2,  -1,   0,   0,   0,  43, -52,   0,   0,   0,   0,   0 } },
 { -7.181385642156e-07, -1.100397339790e-05, {   0,   0,   2,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 {  1.167168745663e-05,  4.576232404237e-08, {   4,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.164311138422e-05,  6.126226417014e-08, {   0,   2,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.542001249489e-08,  1.168818419874e-05, {   2,   0,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.097615037043e-05,  7.234649554468e-07, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   4,   0,   0,   0 } },
 {  8.695046979812e-08,  1.160766049664e-05, {   2,   2,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.168260767902e-05, -2.181630406230e-09, {   2,   2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.163678975302e-05,  4.798594245296e-08, {   2,   0,  -2,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 {  1.154613083656e-05, -1.339261336135e-07, {   4,   0,  -1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.027971941972e-05, -1.394059447617e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 { -4.779535960516e-07,  1.119554292654e-05, {   3,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.151630332808e-05,  1.555175264952e-07, {   4,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.081095925441e-05,  8.526007600072e-07, {   3,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.827846501781e-06, -9.831529746989e-06, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.967775371744e-06, -1.675446764394e-06, {   3,   0,  -2,   0,   0, -14,  12,   0,   4,  -5,   0,   0,   0 } },
 {  1.163882177842e-05, -2.124502457948e-09, {   3,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.835801024791e-06,  1.804721086681e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.721202405586e-07,  1.066061795711e-05, {   2,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.798056204783e-06,  9.825378751250e-06, {   2,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.151334886666e-05, -1.034918303467e-07, {   2,  -2,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  2.228657119294e-07, -1.139121641361e-05, {   0,   0,   3,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -4.620881188390e-07,  1.114543385428e-05, {   2,   0,   1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -8.838497689620e-06,  2.757367201590e-06, {   0,   0,   1,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.152030624736e-05,  7.268546647172e-08, {   2,  -2,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.147152105755e-05,  1.202384668890e-07, {   3,  -2,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.484155661148e-06, -1.009613413371e-05, {   2,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.513274505271e-07, -1.102733640425e-05, {   1,  -2,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -1.110091202955e-05, -4.764771389086e-07, {   0,   0,   1,   0,   0,  -4,   5,   0,  -1,   0,   0,   0,   0 } },
 {  6.514689326105e-08, -1.151121552661e-05, {   4,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.953885274619e-06, -1.619157052707e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.133612774306e-05, -2.321622977935e-07, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 { -9.907971864304e-06, -1.654479869903e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -6,   3,   0,   0 } },
 {  2.237764519939e-06,  9.321505321890e-06, {   3,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.037040529816e-05, -1.178276071290e-06, {   0,   0,   1,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.150504641065e-05,  2.442595198473e-08, {   4,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.088873457139e-05,  6.255470228487e-07, {   6,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  9.513869842437e-07, -1.055943547208e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.083046569656e-05,  6.568284169203e-07, {   1,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.497639615894e-07, -1.093316487951e-05, {   2,   0,  -2,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.223594967957e-07, -1.124678817840e-05, {   5,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.218253243392e-07, -1.104504869366e-05, {   2,  -2,  -1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.145486260097e-05, -8.173133541254e-09, {   2,  -2,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.371982418400e-08, -1.140729536534e-05, {   4,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.144669605563e-05,  1.157731104360e-08, {   2,  -2,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.128715126483e-05, -1.682074875410e-07, {   3,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.120271193153e-06,  1.033169096139e-05, {   1,  -1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.137032791944e-05, -7.275296666012e-08, {   1,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.113266077315e-05,  2.999965838087e-07, {   0,   0,   1,   0,   0,  -4,   6,   0,  -2,   0,   0,   0,   0 } },
 {  8.997806127508e-07,  1.053283030550e-05, {   2,   0,   1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  8.997806127508e-07,  1.053283030550e-05, {   2,   0,  -3,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  1.490714627788e-06,  9.924494249458e-06, {   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  1.006424986055e-05, -1.343345570352e-06, {   0,   0,   1,   0,   0,   0, -12,  18,   0,   0,   0,   0,   0 } },
 {  1.129269613505e-05,  7.874492559282e-08, {   0,   2,  -1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -8.708686045085e-08, -1.128212895846e-05, {   2,   0,   1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.135715706583e-05, -9.249593385688e-09, {   4,   0,  -3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.932523074034e-06, -1.410756629597e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   6,  -3,   0,   0 } },
 { -4.481305653101e-07,  1.089380168394e-05, {   2,   0,   0,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0 } },
 { -1.116843022865e-05,  1.717271564621e-07, {   1,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.009322309875e-05,  1.241779046676e-06, {   2,   2,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.131774282527e-05, -4.212106536267e-09, {   4,   2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.878675891028e-07,  1.040407869442e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  4.082339516386e-07, -1.084818884380e-05, {   0,   0,   2,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.100134277427e-05,  2.350198174378e-07, {   4,   2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.122110315153e-05, -1.373579349783e-08, {   4,   0,  -2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.122909085423e-05, -1.710429825215e-09, {   3,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.120812042687e-05,  2.191245850346e-08, {   2,   0,  -2,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  1.104635038855e-05, -1.835468092771e-07, {   3,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.085509170695e-05, -3.721932730820e-07, {   0,   0,   0,   0,   0,   4,  -5,   0,   1,   0,   0,   0,   0 } },
 { -1.017677067538e-05, -1.046785644675e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  1.086273956281e-05, -3.427829268515e-07, {   2,  -2,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  8.255680722040e-06,  2.921509237300e-06, {   2,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  1.111363646885e-05,  4.206820170638e-08, {   0,   2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.683866096831e-07,  1.046580353618e-05, {   0,   2,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -3.390638002438e-09, -1.112545856285e-05, {   0,   0,   1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.558314007797e-08, -1.108989797714e-05, {   6,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.094059024420e-05,  1.680340542484e-07, {   2,  -2,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.098352321855e-05,  1.124418171803e-07, {   2,   0,   2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -1.068591769473e-05,  3.811037601690e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0 } },
 {  1.078698003752e-05, -2.753128988343e-07, {   2,   0,   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.099742384706e-05, -6.373691559407e-08, {   2,   0,   0,   0,   0,  29, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.105812270098e-05, -2.881992543676e-09, {   2,   2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.959369752636e-07, -1.066070878165e-05, {   2,   0,  -2,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.003908434520e-05,  1.000731424690e-06, {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.782830356242e-07, -1.076114801105e-05, {   1,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.078829315110e-08,  1.095180126377e-05, {   2,  -2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.098242851142e-05, -3.631847098472e-08, {   0,   0,   4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.410062383070e-06, -4.589702627800e-06, {   1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.086138126077e-05,  1.357720120102e-07, {   4,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.495100797697e-07,  1.053205330573e-05, {   1,  -1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  3.024672169476e-07,  1.066931797153e-05, {   1,   0,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  7.802865416280e-06, -3.164334453351e-06, {   2,  -2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.043582911780e-05, -5.293696700327e-07, {   6,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.043816420312e-05, -5.218162760451e-07, {   0,   2,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -1.043816420312e-05, -5.218162760451e-07, {   0,   2,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.890227778344e-07,  1.076424226873e-05, {   1,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.980418300367e-06,  9.697462655916e-07, {   1,   1,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.550679446320e-07, -1.068723344938e-05, {   4,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.093258778829e-05, -4.479522767985e-09, {   2,   2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.179944339827e-07,  1.021720367613e-05, {   0,   0,   1,   0,   0,  -3,   6,   0,  -2,   0,   0,   0,   0 } },
 { -1.092559712545e-05, -3.127758868907e-09, {   0,   0,   3,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.084092229317e-07, -1.051837034145e-05, {   4,   0,  -2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -1.090186605994e-05,  2.381068830641e-08, {   2,   0,  -3,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.615107046500e-08,  1.089260040300e-05, {   2,   0,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.086008349681e-05,  5.299100799690e-08, {   4,   0,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.090175834686e-05, -6.706056963024e-09, {   3,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.324413333689e-07,  1.047099745256e-05, {   1,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.938445880655e-07, -1.030231679308e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 { -4.073987987684e-07, -1.048610556782e-05, {   0,   2,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.088210914459e-05, -2.931599956293e-09, {   0,   2,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.128521260908e-06,  8.755061728160e-06, {   1,  -1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.054656868560e-05, -3.347420809815e-07, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -1.085727201790e-05, -2.345507051716e-08, {   1,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.059851589106e-05,  2.712012296437e-07, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  1.085511267501e-05, -3.093234741802e-09, {   2,   2,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.955957017115e-07,  9.959814673932e-06, {   1,   1,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.955609353942e-07, -9.959817800090e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.951280938310e-07, -9.959856710598e-06, {   3,  -3,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.949021116202e-07,  9.959877017846e-06, {   3,  -1,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.948065036231e-07,  9.959885607831e-06, {   0,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.945544457810e-07,  9.959908249755e-06, {   5,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.945544457810e-07,  9.959908249755e-06, {   3,   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07, -9.959908249755e-06, {   3,   1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07,  9.959908249755e-06, {   1,  -1,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.945544457810e-07, -9.959908249755e-06, {   1,  -1,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.944379774811e-07,  9.959918709730e-06, {   1,   1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.944171174741e-07, -9.959920583015e-06, {   3,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.936609395749e-07,  9.959988460088e-06, {   1,   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.933567286219e-07,  9.960015750762e-06, {   1,   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.085177599171e-05,  1.671943145304e-10, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 {  1.077741730171e-05,  5.536985038111e-08, {   4,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.771706798647e-07,  1.045103408404e-05, {   2,   0,  -2,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  4.957662317151e-07,  1.032519139770e-05, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -4.180090502267e-07,  1.039062586908e-05, {   2,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.055135658380e-05,  2.442160791223e-07, {   2,   0,  -2,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.060512430678e-05,  1.890107447942e-07, {   4,   0,  -2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -3.207488732439e-07,  1.046463763386e-05, {   0,   2,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.075817316828e-05, -1.754646293709e-08, {   0,   2,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.076426949150e-05,  1.012538008410e-08, {   2,   0,   2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.074848341907e-05,  2.054408946069e-08, {   0,   2,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -7.371222936763e-07,  1.002810497628e-05, {   2,   0,  -2,   0,   0,   6,  -4,  -8,   3,   0,   0,   0,   0 } },
 { -2.040374960689e-07, -1.055861943544e-05, {   1,  -2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.324517005982e-07,  1.052930209758e-05, {   1,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.712191865407e-07, -1.018460018114e-05, {   2,   2,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.246005523981e-07, -1.043001323357e-05, {   1,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.003406546360e-05,  7.030076307300e-07, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 {  1.045968780618e-05,  2.707901848689e-07, {   2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.070932052295e-05,  1.998941781255e-08, {   4,   0,  -2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.857551942120e-07,  1.023678810691e-05, {   1,  -1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.065780566224e-05,  5.341937880618e-08, {   0,   2,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.951020531660e-07, -1.000876920281e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,  -2,   0 } },
 {  4.126304349672e-07, -1.027741862080e-05, {   3,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.033350736428e-05,  3.541283224339e-07, {   1,  -1,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  1.032700693830e-05, -3.606213873156e-07, {   2,   0,  -1,   0,   0,   5,  -7,   0,   2,   0,   0,   0,   0 } },
 { -2.588620554888e-07,  1.042696121348e-05, {   0,   2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.058026097181e-05,  6.908929653444e-08, {   2,   0,  -1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -9.610185110000e-07, -9.685968913478e-06, {   2,  -2,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.054696016287e-05, -7.331042711775e-08, {   0,   2,   0,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 { -4.387916511977e-07,  1.016921941941e-05, {   4,  -2,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.046613217818e-05, -1.289127412919e-07, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 {  1.058217868512e-05, -1.278886686890e-08, {   4,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.407288928428e-07,  1.023409386029e-05, {   0,   0,   1,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 {  1.021758565402e-05, -3.375626037262e-07, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.053298025894e-05,  2.194153679747e-08, {   0,   0,   0,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.029564487847e-05,  2.553064340619e-07, {   2,  -2,   0,   0,   0,   8, -11,   0,   0,   3,   0,   0,   0 } },
 {  1.051936277436e-05, -2.840877618649e-08, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   1,   0,   0,   0 } },
 {  1.053140071234e-05, -1.217662507150e-08, {   4,   0,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.049301570124e-05,  3.159737756928e-08, {   4,   0,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.047009728404e-05,  5.386559703461e-08, {   0,   2,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.039321383972e-05,  7.587466593017e-08, {   1,   0,  -1,   0,   0,   0,   4,  -6,   0,   2,   0,   0,   0 } },
 { -1.045776326854e-05,  9.997361618044e-09, {   3,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.005504061346e-05, -4.028217158562e-07, {   2,  -2,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.989220725012e-06, -4.641867156508e-07, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.566678291850e-06, -8.751954871491e-07, {   2,   0,   3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.008581978497e-05, -3.420599484521e-07, {   2,   0,   1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 {  1.035866849080e-05,  4.603953778732e-08, {   0,   2,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.038977276767e-05, -5.554362809950e-09, {   2,   2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.036118066495e-05,  3.325359091027e-08, {   0,   0,   2,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.065054291710e-06, -4.309660906804e-06, {   1,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -8.415474773514e-08, -1.028436919710e-05, {   0,   0,   4,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.027890736725e-05, -6.950797380447e-08, {   2,   2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.031212584961e-05, -3.287027643790e-08, {   4,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.069098510930e-06,  1.272460326619e-06, {   1,  -1,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.033396506074e-05, -1.450540453300e-09, {   2,  -2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.031602510254e-05,  2.232083402534e-09, {   1,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.030548896494e-05, -1.253883318437e-08, {   2,   0,  -3,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.122016881811e-07, -1.018729884272e-05, {   2,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.005860091319e-05,  2.408287877838e-07, {   2,   0,   0,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 {  1.023561691580e-05,  4.914913558585e-08, {   0,   2,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.024291440157e-05,  3.979746503587e-08, {   2,  -2,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.865979193999e-07,  9.995892237944e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.027645953628e-05, -3.699766834434e-09, {   1,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.001814005694e-05,  2.607695870724e-07, {   1,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  1.023243982809e-05, -4.583677888226e-08, {   2,   0,   1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.024614933401e-05,  1.084464857222e-08, {   4,   0,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.021852985835e-05, -3.570271634560e-08, {   3,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.975628706670e-07,  1.002764863295e-05, {   1,  -2,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.932371310706e-07,  9.723545668150e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 {  1.012406037718e-05, -9.055134127117e-08, {   0,   2,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -9.767753821430e-06, -4.367689599060e-07, {   0,   0,   0,   0,   0,   1,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.603110570394e-07,  1.004300934004e-05, {   0,   2,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.093270412011e-07,  1.008781951748e-05, {   2,   0,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.283755612196e-08, -1.014190055033e-05, {   2,   0,   1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.000928890729e-05,  1.615433155577e-07, {   4,   0,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.998639242632e-06, -1.649645285931e-07, {   2,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.007073374160e-05,  8.946099797693e-08, {   2,   0,  -3,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.012229770888e-05, -7.630694718040e-09, {   0,   2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.010993056249e-05,  1.938810762019e-08, {   0,   0,   1,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.010775903132e-05, -1.293533126635e-08, {   2,   0,   2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.011507453025e-05, -5.499341287676e-09, {   3,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  7.476171355998e-08,  1.004162897591e-05, {   2,   0,  -2,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.001962193957e-05,  7.192350228927e-08, {   0,   0,   0,   0,   0,   4,  -7,   0,   4,   0,   0,   0,   0 } },
 { -1.005001051890e-05, -1.289233715183e-08, {   4,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.977582931750e-06,  4.082106758862e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 {  9.999976686666e-06,  2.159319672534e-08, {   6,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.999997303030e-06,  7.344344832143e-09, {   2,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.999999987888e-06, -4.921828488612e-10, {   2,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999991873e-06,  4.031710571015e-10, {   2,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999994899e-06, -3.193952530579e-10, {   4,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999995853e-06, -2.879793265414e-10, {   2,  -1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999995853e-06, -2.879793265414e-10, {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999996435e-06, -2.670353755238e-10, {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999996481e-06, -2.652900462719e-10, {   1,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999997183e-06, -2.373647782499e-10, {   1,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999997426e-06,  2.268928027398e-10, {   0,   1,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999997620e-06,  2.181661564820e-10, {   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -9.999999997620e-06,  2.181661564820e-10, {   0,   1,  -5,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998055e-06,  1.972222054626e-10, {   2,  -1,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999998089e-06,  1.954768762109e-10, {   2,  -1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999998353e-06,  1.815142421974e-10, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -9.999999998353e-06,  1.815142421974e-10, {   0,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998353e-06, -1.815142421955e-10, {   2,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998384e-06, -1.797689129433e-10, {   4,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998477e-06,  1.745329251906e-10, {   6,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998596e-06,  1.675516081836e-10, {   2,   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998625e-06,  1.658062789319e-10, {   0,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998625e-06, -1.658062789299e-10, {   4,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  9.999999998711e-06,  1.605702911766e-10, {   2,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999998711e-06,  1.605702911766e-10, {   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998739e-06,  1.588249619248e-10, {   4,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999025e-06,  1.396263401550e-10, {   4,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999999049e-06, -1.378810109023e-10, {   2,  -1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999143e-06, -1.308996938977e-10, {   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999166e-06,  1.291543646440e-10, {   4,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999166e-06,  1.291543646440e-10, {   3,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999166e-06,  1.291543646440e-10, {   2,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999188e-06, -1.274090353929e-10, {   2,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999232e-06,  1.239183768884e-10, {   2,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999254e-06, -1.221730476357e-10, {   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999296e-06,  1.186823891328e-10, {   4,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999296e-06,  1.186823891328e-10, {   4,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999316e-06, -1.169370598833e-10, {   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999470e-06,  1.029744258658e-10, {   3,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999539e-06,  9.599310885821e-11, {   2,  -1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999604e-06, -8.901179184903e-11, {   4,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999692e-06, -7.853981633906e-11, {   4,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999705e-06,  7.679448708700e-11, {   6,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999731e-06, -7.330382858157e-11, {   4,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999744e-06, -7.155849932907e-11, {   2,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999768e-06, -6.806784082902e-11, {   2,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -9.999999999780e-06,  6.632251157651e-11, {   3,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999834e-06, -5.759586531393e-11, {   2,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.000000000000e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.927656371190e-06, -2.019634291347e-08, {   4,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.025337917640e-06, -8.862357152080e-06, {   1,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.595845583500e-07, -8.918712819443e-06, {   3,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.061563456117e-06, -5.697799744470e-06, {   2,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  8.318692691970e-06,  1.206304110423e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -9.330563701760e-06, -7.250260799730e-08, {   0,   2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.996898605300e-07,  8.186305875798e-06, {   6,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.816370197180e-06, -7.209424641192e-06, {   0,   0,   1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.996769994990e-06, -4.942757020284e-06, {   2,   0,  -1,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.847164751030e-06,  3.082434451032e-06, {   0,   0,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -8.867095210690e-06, -8.022353232623e-09, {   6,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.712699267360e-06,  8.327650479235e-08, {   6,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.143297112542e-06, -7.569624458790e-06, {   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.414243171310e-06, -2.653476118956e-07, {   2,   0,   2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.232562209400e-07,  7.830737797838e-06, {   2,   0,  -4,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.519968514400e-06, -1.066814914672e-07, {   4,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.530358415891e-06, -7.024272236855e-06, {   3,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.899849958600e-07, -7.917131707550e-06, {   0,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.224207479270e-06, -4.275690892042e-06, {   3,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.624089436022e-06,  6.855714923030e-06, {   1,  -1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.090426599470e-06,  3.276907775492e-06, {   2,   0,  -2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.015790014797e-06, -5.220863757120e-06, {   0,   0,   2,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.621605410910e-06, -5.042929460354e-07, {   2,   0,  -1,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -5.155759530020e-06,  2.583643752410e-06, {   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  7.536553475166e-06,  1.349994965200e-07, {   2,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.456207541405e-06,  6.101917685180e-06, {   3,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.155224464670e-06,  5.928802608000e-06, {   1,  -1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.881061374300e-07,  6.641698258749e-06, {   2,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.151332034340e-06, -3.521607822537e-06, {   4,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.030897565048e-06, -4.617839103490e-06, {   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.282238560666e-06,  2.978397064840e-06, {   4,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.513603439580e-06,  4.505639296019e-07, {   2,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.763624742060e-06, -2.155957485710e-06, {   0,   0,   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -5.666158964203e-06,  2.527546202000e-07, {   1,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.522022313306e-06,  2.331852936237e-06, {   2,   0,  -2,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.886663615000e-07,  4.860276865900e-06, {   1,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.731100610430e-06,  7.254350470200e-08, {   1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.596076650090e-06,  1.052171731604e-07, {   1,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.089839166180e-06,  4.526041169650e-06, {   1,  -3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.446001082490e-07,  5.035779133571e-06, {   4,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.843213001780e-06,  3.653172855080e-06, {   1,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  9.991276606910e-07,  4.104894557570e-06, {   1,   1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.603723736600e-06,  2.990602518654e-08, {   1,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.042795885717e-06, -2.493658028180e-06, {   2,   0,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.608621261900e-06,  1.727055164310e-06, {   2,   0,  -1,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.197485714370e-06, -1.244000487575e-07, {   4,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.447098779900e-06, -1.752261992860e-06, {   1,  -1,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.155012133700e-06, -3.915492223704e-08, {   4,   0,  -2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.983017371850e-06,  1.199303767580e-06, {   0,   0,   1,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.087971415572e-06,  6.406933915000e-08, {   2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.144492909320e-06,  2.624309652690e-09, {   3,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.864244851510e-06,  1.621763433340e-07, {   2,  -2,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.925507817020e-06,  2.562529233173e-09, {   1,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.215421523990e-06,  2.686755878750e-06, {   0,   0,   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 {  1.614287161072e-06,  2.201639051530e-06, {   2,  -2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.514553106090e-06,  2.438923905312e-07, {   2,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.692376193393e-07, -2.750931143230e-06, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -6.477529689580e-07, -2.818136719770e-06, {   3,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.776915368780e-06,  1.594356544785e-06, {   4,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.843869252420e-06, -8.728528247735e-08, {   2,  -2,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.474550568900e-06, -2.959477459876e-07, {   2,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.278630463870e-06, -5.238956210800e-09, {   2,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.557363003883e-07,  1.857974551820e-06, {   2,   0,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.878789826720e-06, -8.106313343556e-08, {   0,   2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.717608940970e-06,  2.671172089443e-08, {   2,  -2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.706881649540e-06,  3.492771716784e-08, {   6,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.284146892090e-08, -1.452019298280e-06, {   0,   0,   3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.583212507400e-07,  1.062235208993e-06, {   0,   0,   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  9.796030726800e-07, -3.702787814173e-07, {   2,   0,  -2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.100890452400e-07, -2.939551384817e-07, {   4,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.526137464580e-07, -8.439576304300e-07, {   3,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.272753464000e-07, -4.143200890440e-07, {   1,  -1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  5.477839524300e-07, -1.377527617284e-07, {   4,   0,  -2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.055263255380e-07, -2.205391404700e-07, {   1,  -1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  3.932028594100e-07, -2.611923269580e-08, {   2,   0,   2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.720538518300e-08, -2.717514580200e-07, {   3,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.022722274300e-08,  2.818504289900e-07, {   1,   1,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.182862042900e-07, -1.719865799049e-07, {   4,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.748640002468e-12, -1.224522926694e-08, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 1199> g_PERT_S1_2
{ {
 {  1.676800000000e+00,  0.000000000000e+00, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.164200000000e-01,  0.000000000000e+00, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.138300000000e-01,  0.000000000000e+00, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.711500000000e-01,  0.000000000000e+00, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.028286271523e-01,  2.234626048870e-01, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.756000000000e-01,  0.000000000000e+00, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.118000000000e-02,  0.000000000000e+00, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.128000000000e-02,  0.000000000000e+00, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.516000000000e-02,  0.000000000000e+00, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.048000000000e-02,  0.000000000000e+00, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.747000000000e-02,  0.000000000000e+00, {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.707000000000e-02,  0.000000000000e+00, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.649000000000e-02,  0.000000000000e+00, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.438000000000e-02,  0.000000000000e+00, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.165000000000e-02,  0.000000000000e+00, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.884423738233e-03,  1.619871474513e-02, {   0,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.923000000000e-02,  0.000000000000e+00, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.688830802010e-03,  1.236325946529e-02, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.335288840325e-03,  1.159490956816e-02, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.293000000000e-02,  0.000000000000e+00, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.270000000000e-02,  0.000000000000e+00, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.097000000000e-02,  0.000000000000e+00, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.087705472886e-02,  2.009137471921e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.345336034681e-03,  1.996019810385e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.340000000000e-03,  0.000000000000e+00, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.860000000000e-03,  0.000000000000e+00, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.310000000000e-03,  0.000000000000e+00, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.620197481278e-03,  6.865906201356e-04, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.619335487852e-03,  6.864397171771e-04, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.179438359444e-03,  4.955296321374e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.850000000000e-03,  0.000000000000e+00, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.649791009018e-03, -2.823784808853e-03, {   0,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.390000000000e-03,  0.000000000000e+00, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.690000000000e-03,  0.000000000000e+00, {   4,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.743086927735e-03,  2.159901950685e-03, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.209511837538e-03,  2.627512588305e-03, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.780000000000e-03,  0.000000000000e+00, {   2,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.190411333055e-03,  2.585936204964e-03, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.180473222374e-03,  2.564326994009e-03, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.918344135177e-03,  8.019181869646e-04, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.620000000000e-03,  0.000000000000e+00, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.137411022758e-03,  2.470838906457e-03, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.910437928745e-03, -1.509828838633e-03, {   0,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.165695068280e-03,  2.092303804724e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.170000000000e-03,  0.000000000000e+00, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-03,  0.000000000000e+00, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.145820737354e-03,  1.827078556584e-03, {   2,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.950000000000e-03,  0.000000000000e+00, {   2,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.064358057103e-04,  2.830054925663e-03, {   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.311870738770e-03, -6.066351727446e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.700000000000e-03,  0.000000000000e+00, {   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.560000000000e-03,  0.000000000000e+00, {   2,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.409839311860e-04,  1.329015515115e-03, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.249724657658e-04, -1.314376794393e-03, {   0,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.350061234299e-03,  6.642146325955e-04, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.302463382935e-04,  1.552842952737e-03, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.302919274337e-04,  1.552283666829e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.484748040784e-03,  4.850625817016e-04, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.860000000000e-03,  0.000000000000e+00, {   2,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.144929961576e-04,  8.394611316827e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.690000000000e-03,  0.000000000000e+00, {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.084223643241e-04, -7.500674816019e-04, {   2,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.600000000000e-03,  0.000000000000e+00, {   4,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.550000000000e-03,  0.000000000000e+00, {   4,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.480000000000e-03,  0.000000000000e+00, {   2,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.412477983854e-03, -1.984589620100e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.410000000000e-03,  0.000000000000e+00, {   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.380000000000e-03,  0.000000000000e+00, {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.747258295539e-04,  3.235212632740e-04, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.290000000000e-03,  0.000000000000e+00, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.470958880098e-04,  3.175357694203e-04, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.260000000000e-03,  0.000000000000e+00, {   2,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.877526355574e-04,  8.423021577940e-04, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.428358359466e-04,  4.826342693258e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.200000000000e-03,  0.000000000000e+00, {   2,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.714749982007e-04,  9.018500533025e-04, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.693588793364e-04,  8.947979090916e-04, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.637606614093e-04,  7.901527089040e-04, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.273983542339e-04, -2.894072019391e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -8.238944661497e-04, -2.899716771580e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.644678598537e-04, -7.385526112026e-04, {   0,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.015388309029e-03, -7.790208762633e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.014481868351e-03, -7.781523809714e-05, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.070000000000e-03,  0.000000000000e+00, {   2,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.040000000000e-03,  0.000000000000e+00, {   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.700000000000e-04,  0.000000000000e+00, {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.600000000000e-04,  0.000000000000e+00, {   2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.400000000000e-04,  0.000000000000e+00, {   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.400000000000e-04,  0.000000000000e+00, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.000000000000e-04,  0.000000000000e+00, {   4,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.800000000000e-04,  0.000000000000e+00, {   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.733005698892e-04, -6.022022256524e-04, {   1,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -6.181889804631e-04,  1.985335358478e-04, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.953175694730e-04,  1.996345043824e-04, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.900000000000e-04,  0.000000000000e+00, {   4,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.600000000000e-04,  0.000000000000e+00, {   4,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.600000000000e-04,  0.000000000000e+00, {   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.600000000000e-04,  0.000000000000e+00, {   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.682300897322e-04,  1.809624486398e-04, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.300000000000e-04,  0.000000000000e+00, {   4,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.300000000000e-04,  0.000000000000e+00, {   2,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.590202466490e-04,  2.513200933583e-04, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -7.100000000000e-04,  0.000000000000e+00, {   4,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.099444443470e-04,  5.884444587296e-04, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.098460653159e-04,  5.876186849707e-04, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.485222107293e-04, -4.474856673047e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.340871334972e-04, -1.577654098117e-04, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.900000000000e-04,  0.000000000000e+00, {   3,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.173398531155e-04,  1.648031392206e-04, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.884584178453e-04, -1.834602145213e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.600000000000e-04,  0.000000000000e+00, {   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.813393568887e-04, -4.657109873238e-04, {   1,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.035158091927e-04, -4.420983549469e-04, {   0,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.034920646629e-04, -4.420467504463e-04, {   0,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.609652107908e-04, -7.194248218335e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.609652107908e-04, -7.194248218335e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.803306843991e-04, -4.589594700801e-05, {   2,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  4.480403290169e-04,  1.475668147026e-04, {   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.851771493653e-04,  4.028995321201e-04, {   2,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.518306847420e-04,  1.361420903454e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  5.800000000000e-04,  0.000000000000e+00, {   3,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.391979622422e-04,  3.139740888244e-05, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.398277971059e-04,  3.278354388329e-04, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.370032442884e-04,  3.230449567839e-04, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.795605163321e-04,  1.393756943887e-04, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.039086236224e-04,  3.030456678296e-04, {   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  6.864005029495e-06,  4.921677567826e-04, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.589745157930e-06,  4.880091277539e-04, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.487902776226e-04,  2.399871472155e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.907668622405e-04, -1.912155428287e-04, {   1,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.557767509640e-04,  1.132348891363e-04, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.436972009867e-04, -2.236755515191e-04, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.447636991351e-04,  3.144690325805e-04, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.725038215969e-04, -2.812951127258e-04, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -4.500000000000e-04,  0.000000000000e+00, {   2,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.409760708138e-04,  3.062604639892e-04, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.185564514773e-04,  2.601350429269e-05, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.298169060847e-04,  1.142705639098e-04, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.728959372600e-04,  2.680143156818e-04, {   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  4.158349488290e-04, -2.174660718286e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.798772225197e-04,  1.532967753168e-04, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.836315337229e-04,  1.470775286157e-04, {   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.817812998651e-04,  4.663987997855e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.817608098683e-04,  4.664082452022e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.578389993567e-04, -2.692046806064e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.822292196765e-04,  2.396899865166e-04, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.100000000000e-04,  0.000000000000e+00, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.946102075365e-04,  1.094050993579e-04, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.843934047428e-04, -2.165854837368e-04, {   1,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.731730658620e-04,  2.216553244070e-04, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.443743383296e-04, -2.496921562347e-04, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.435497750585e-04, -2.480073909792e-04, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.851838520462e-05, -3.625807507985e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  3.899861789126e-04,  5.255765658090e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.900000000000e-04,  0.000000000000e+00, {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.842811524831e-05, -3.591165139924e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -3.800000000000e-04,  0.000000000000e+00, {   3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.325223723462e-04, -2.440967500512e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.914479719227e-04,  1.846994157785e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.917161073111e-04,  1.830397755818e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.452851636530e-04,  1.198684523097e-04, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.205462205010e-04, -1.379690271637e-04, {   0,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  2.786538736495e-04,  7.821288220615e-05, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.648232524209e-04, -8.915503621276e-05, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.375084040546e-04,  1.152834730209e-04, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.239068780788e-04,  1.205404465746e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.400000000000e-04,  0.000000000000e+00, {   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.112734562713e-05, -2.616533264316e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.484255617784e-04,  8.395054952947e-05, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.493951432481e-05,  3.162474107998e-04, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.853258885039e-04,  1.454566438239e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.300000000000e-04,  0.000000000000e+00, {   0,   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.502079584125e-05,  3.132070188634e-04, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.084679865850e-04, -2.183617952763e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.200000000000e-04,  0.000000000000e+00, {   2,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.200000000000e-04,  0.000000000000e+00, {   2,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.079749139301e-04, -2.106325816423e-04, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.965117400530e-04,  1.969345176810e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.302407631042e-05, -2.428313610049e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.041135103593e-04, -2.024927209829e-04, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.981309137081e-04,  1.029504320626e-04, {   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-04,  0.000000000000e+00, {   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.854234218727e-04, -1.133691810648e-04, {   0,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.520536613025e-04,  1.446736073302e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.771213726059e-04, -1.165270196159e-04, {   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.900000000000e-04,  0.000000000000e+00, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.156983587316e-04, -1.709936925258e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.359046004683e-04, -1.467308405604e-04, {   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.808508800672e-04,  9.878471234092e-05, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  9.961746080332e-05, -1.775459130284e-04, {   0,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -1.557580152484e-04, -1.200248998232e-04, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.179119715076e-04,  1.559123296491e-04, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.630912252749e-04, -9.530598258449e-06, {   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  6.386835215314e-07, -2.704398913316e-04, {   0,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  1.074302264767e-04, -1.623617160722e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.981666072974e-04,  6.463768741980e-05, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.573195281176e-05, -1.600898759200e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.500000000000e-04,  0.000000000000e+00, {   4,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.500000000000e-04,  0.000000000000e+00, {   2,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.928962235817e-04,  5.673442381675e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.088819463745e-04,  1.406799044383e-04, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.532310208772e-04,  9.444709757810e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.495781523353e-04,  9.345021127723e-05, {   0,   0,   1,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 {  5.595270728456e-05,  1.858923701936e-04, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.589721985933e-05,  1.857074536792e-04, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.193816370622e-05,  1.467602549728e-04, {   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.043018500669e-05, -1.439527062603e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.935182294639e-04,  4.036177101974e-05, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.301233164169e-05,  1.490394757582e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.366484279515e-05,  1.783103796395e-04, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.464605141568e-04, -8.418715473358e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -5.323069252033e-05,  1.768635375162e-04, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.300000000000e-04,  0.000000000000e+00, {   6,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.300000000000e-04,  0.000000000000e+00, {   2,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.300000000000e-04,  0.000000000000e+00, {   1,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.490482296897e-04,  7.885020334860e-05, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.200000000000e-04,  0.000000000000e+00, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.200000000000e-04,  0.000000000000e+00, {   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.659626004218e-05,  1.593022954280e-04, {   0,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -5.924523550130e-05,  1.512980059921e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -1.623656537521e-04, -4.781325322497e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.100000000000e-04,  0.000000000000e+00, {   3,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.100000000000e-04,  0.000000000000e+00, {   2,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.100000000000e-04,  0.000000000000e+00, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.100000000000e-04,  0.000000000000e+00, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.100000000000e-04,  0.000000000000e+00, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.126456219362e-05, -1.359539625739e-04, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  9.475415910755e-05,  1.064598790240e-04, {   2,   0,   0,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.048824889509e-05,  1.095489566449e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  2.000000000000e-04,  0.000000000000e+00, {   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-04,  0.000000000000e+00, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.845474183451e-05, -1.296527125457e-04, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.537573882809e-04,  4.411596372306e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.520511151420e-05, -1.257813050550e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.900000000000e-04,  0.000000000000e+00, {   2,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.900000000000e-04,  0.000000000000e+00, {   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.057409797276e-04, -8.352087941939e-05, {   0,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -5.608202660885e-05,  1.327990826295e-04, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.050988160251e-04, -8.301864850993e-05, {   0,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -2.817614387826e-05, -1.574995394157e-04, {   2,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.203889161326e-04,  6.498391659045e-05, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.744169207497e-05,  1.344302852070e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.800000000000e-04,  0.000000000000e+00, {   6,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.800000000000e-04,  0.000000000000e+00, {   4,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.800000000000e-04,  0.000000000000e+00, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.041577933778e-04,  7.485478583033e-05, {   1,  -2,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  4.172930492664e-05, -1.366827423338e-04, {   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.697916290029e-04,  8.414435248870e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.066611203020e-04, -6.647979248304e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.218352567451e-04, -5.023756780116e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.648384839775e-05, -7.443849548389e-05, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.196358314804e-04,  5.086519267615e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.273553945503e-05,  1.175049753659e-04, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.700000000000e-04,  0.000000000000e+00, {   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.700000000000e-04,  0.000000000000e+00, {   2,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.416752504145e-05,  1.025491658802e-04, {   2,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.634834795507e-05, -7.009585942558e-05, {   2,   0,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  6.393074036798e-05,  1.019460073975e-04, {   2,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.600000000000e-04,  0.000000000000e+00, {   2,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.142666862756e-05,  9.628131549463e-05, {   1,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -6.639991628825e-05,  9.123505365552e-05, {   2,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.725300569690e-05,  7.830691611087e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.416487503807e-05, -7.032757751023e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.247255918214e-04, -2.931457980570e-05, {   2,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.268877145395e-04,  2.645584238093e-05, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.114978334233e-04, -4.087365110409e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.313054397352e-04,  1.995735020231e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.500000000000e-04,  0.000000000000e+00, {   2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.500000000000e-04,  0.000000000000e+00, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.051986047254e-04, -4.303725541801e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.673324739376e-05,  4.962206895660e-05, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.085258484230e-04, -3.641838308656e-05, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.009182006218e-04,  4.376661722441e-05, {   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.127902843548e-05, -5.057560705742e-05, {   0,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  1.116525974699e-04,  3.012082666195e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.870399551318e-05,  7.266196391872e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.107573839644e-04, -2.947382003098e-05, {   1,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.400000000000e-04,  0.000000000000e+00, {   0,   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.896892266936e-05,  8.765490445074e-05, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.629721011200e-05,  1.094088287813e-04, {   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.920708491293e-05,  1.055505799768e-04, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.925999129624e-05,  1.053804495794e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  4.628360228939e-05, -8.646697521929e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.950833959859e-05,  7.234235065057e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.225828379434e-04, -7.947875390080e-06, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.299258326311e-04, -5.032596339021e-07, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.299101749300e-04, -4.758385290967e-07, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.300000000000e-04,  0.000000000000e+00, {   4,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.300000000000e-04,  0.000000000000e+00, {   4,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.300000000000e-04,  0.000000000000e+00, {   2,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.300000000000e-04,  0.000000000000e+00, {   0,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.210966423267e-05, -4.781534108693e-05, {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.631331865429e-05,  9.277463787485e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  4.466749571031e-05, -8.414442535217e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  6.533675160715e-05,  6.344763569094e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.294041984157e-05,  3.482280469182e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.516247791275e-05, -6.207939651986e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.051211754666e-05, -1.159711353423e-04, {   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -9.247038714615e-05,  3.324636399388e-05, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  9.236787370287e-05,  3.321306740154e-05, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -9.864953709039e-05, -2.677554049825e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.849868558666e-05,  7.602355661220e-05, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  7.480347992522e-05, -4.703710698473e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.939587888022e-05,  4.238271365590e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.195473487327e-05,  6.917517998607e-05, {   2,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -6.892358998629e-05, -5.206255227595e-05, {   0,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-04,  0.000000000000e+00, {   4,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.200000000000e-04,  0.000000000000e+00, {   2,  -2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-04,  0.000000000000e+00, {   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.430045660419e-05,  5.545472684298e-05, {   1,   0,   0,   0,   0,   0, -23,  20,   0,   0,   0,   0,   0 } },
 { -4.763762713883e-05,  7.196517096515e-05, {   0,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.281363088870e-05,  4.634678163410e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  8.388662863946e-05, -3.478904606833e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.646220024909e-05,  7.219804296008e-05, {   2,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  6.908320087127e-05,  4.939130884126e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -6.885345965932e-05,  4.944676570114e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  6.572243502122e-05,  5.152852877429e-05, {   1,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.402637458346e-05, -4.297397491677e-05, {   0,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -7.802298338289e-05,  3.874867564718e-05, {   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.128041381756e-05,  4.541723162999e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.133661444328e-04, -3.244610969358e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.622697590307e-05,  9.910195406673e-05, {   0,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.122771289590e-04, -1.952001732064e-06, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  6.640492923940e-05,  4.778497376761e-05, {   2,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.867521484122e-05,  2.526267961055e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.076965082564e-04, -5.937739949229e-06, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.849176949796e-05, -9.513413991376e-05, {   0,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -8.572437631055e-05, -2.741042331243e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.888442121329e-05, -5.415371583165e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.538966698855e-05, -3.696291537111e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.100000000000e-04,  0.000000000000e+00, {   4,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.100000000000e-04,  0.000000000000e+00, {   3,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.100000000000e-04,  0.000000000000e+00, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.100000000000e-04,  0.000000000000e+00, {   0,   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.100000000000e-04,  0.000000000000e+00, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.100000000000e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.188835123856e-05,  3.802209221160e-05, {   2,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  6.296830097656e-05,  4.556294818619e-05, {   2,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.353389135490e-05, -1.472428024728e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -8.126847521530e-05,  2.571720418010e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -3.366793249603e-05, -7.313657525069e-05, {   0,   2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.821521593295e-05,  8.813742626442e-05, {   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.908992681284e-05,  5.639294180548e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -8.700875953672e-05,  1.814849420004e-05, {   0,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.324229887722e-05,  3.179223390644e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.296483745924e-05, -7.160852385165e-05, {   0,   2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.431319125207e-05,  4.000412947785e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.585299100786e-05,  7.803309555557e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  8.736916153400e-05, -1.611373804810e-05, {   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.728435406010e-05, -6.521975887299e-05, {   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -6.188279227071e-05, -4.058485841084e-05, {   0,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.000000000000e-04,  0.000000000000e+00, {   6,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-04,  0.000000000000e+00, {   0,   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.132425198397e-05, -6.804965019835e-05, {   1,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.120817762765e-05, -6.779350324630e-05, {   1,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.438886040027e-05, -7.294518829336e-05, {   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.171241176598e-05,  3.406429735420e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -8.983522733721e-05,  5.443521770959e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.953303693862e-05,  2.551557062103e-05, {   2,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.002326852948e-05,  7.420618745372e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  3.637110043546e-05,  5.756279770091e-05, {   0,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -6.249637814254e-05,  3.111004460188e-05, {   2,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.943120804179e-05,  6.393446284008e-05, {   4,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.095510357196e-05, -4.211770194001e-05, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.209533903428e-05,  4.026018472352e-05, {   0,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.868300396687e-05,  6.230851251412e-05, {   4,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.597497186147e-05,  3.469965338680e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  6.148022710967e-05, -2.904841830041e-05, {   2,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.395829891810e-05,  5.638215743411e-05, {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.522417531607e-05,  5.499029406322e-05, {   0,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  9.000000000000e-05,  0.000000000000e+00, {   2,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.000000000000e-05,  0.000000000000e+00, {   1,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.653263677745e-05,  3.231376513305e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.801380652916e-05,  6.082856296672e-05, {   2,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.732558914851e-05, -5.139708377565e-05, {   0,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -2.244624981852e-05, -6.591311974461e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  6.068132526453e-05, -2.747405188435e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.732075870521e-05,  5.934797539356e-05, {   0,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -6.512096501798e-05,  2.145070370807e-05, {   2,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -5.703311350150e-05,  2.944551928597e-05, {   0,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.859879915348e-05, -5.733096966849e-05, {   1,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -3.288429916362e-05,  5.250932718320e-05, {   0,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -6.083092089541e-05, -2.389325610069e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.273230177845e-05,  1.186562617836e-05, {   1,  -3,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  6.297218881744e-05,  2.073538669660e-05, {   2,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.235274303753e-05,  4.051321024398e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  6.230210984173e-05,  2.051362628339e-05, {   2,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.856442099723e-05, -1.410461649749e-05, {   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.903266255045e-05,  2.338629256590e-05, {   1,   0,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  7.725367554011e-05, -5.012964048002e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -3.217849356785e-05,  5.002590254282e-05, {   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  5.245002349000e-05, -2.913751938479e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.563025587381e-05,  5.566985981844e-05, {   0,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.135824419950e-05,  4.895112325918e-05, {   0,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   4,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   2,   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   2,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   2,  -2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.844295512484e-05,  6.126818325723e-05, {   0,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.106629384809e-05,  3.844420989679e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.829927973603e-05,  6.078937293212e-05, {   0,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.356602731263e-05, -5.117455494947e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  5.190481480433e-05,  2.668502481694e-05, {   0,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -5.933366340394e-05,  1.866880773714e-05, {   0,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  2.004578412959e-05,  5.655233451088e-05, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.478517037978e-05,  3.164635907806e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  4.155460592550e-05, -3.478321802712e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  4.720512512171e-05, -2.812946421133e-05, {   0,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -4.907323939935e-05,  2.596344051379e-05, {   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -4.169133930897e-05, -3.295197014968e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  5.458446483498e-05, -1.985918166118e-05, {   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.564212969686e-05,  5.792515669851e-05, {   2,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.675641260777e-05,  2.670148835275e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -7.239743451486e-05, -1.050362104940e-06, {   0,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -5.800589484427e-05, -1.534001836098e-05, {   1,   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.065080559240e-05, -1.268180442865e-05, {   0,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.565289252762e-05,  5.754514062653e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  5.539055187536e-05,  1.738839851915e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -4.766661509552e-05,  2.476620421144e-05, {   0,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.909413392310e-06,  6.997395330048e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.172029533311e-05,  2.972315463167e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -4.759738749713e-05,  2.338559701608e-05, {   2,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.463720334944e-05, -3.605917558860e-05, {   0,   0,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.891852044449e-05,  1.133135534847e-05, {   2,   0,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   4,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   4,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   2,   0,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   2,   0,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,  -2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,  -2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.573750244179e-05,  2.409451441050e-05, {   0,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -4.060519861844e-05,  2.903428700579e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.170312412263e-05, -4.782782369104e-05, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.454542872880e-05,  4.458129238198e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.333214917397e-05, -2.578670818892e-05, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  4.831833858294e-05,  2.058956379799e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.786519591563e-05, -3.083618319363e-05, {   2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.291762862151e-05, -2.565301451109e-05, {   0,   0,   2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.746336738408e-05,  4.107679156686e-05, {   0,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  4.291227702875e-05, -2.549530445732e-05, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -4.189193139185e-05,  2.625876204800e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.436715826616e-05,  5.269908844356e-05, {   0,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -4.118144017912e-05, -2.565240068675e-05, {   0,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.529092199268e-05,  5.096918445361e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  1.189910249067e-05, -5.434523161684e-05, {   2,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.318416827523e-05,  2.268305966396e-05, {   2,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.347976459872e-05,  4.212520797272e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.572249173122e-05,  3.968317860936e-05, {   2,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.961833614074e-05,  3.561670479006e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -6.090580604403e-05, -3.720474169716e-06, {   2,  -2,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.584656783882e-05, -2.820173174162e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -4.042272427797e-05, -2.360649853077e-05, {   2,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -4.615920152995e-05,  1.768999980596e-05, {   0,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.111121402025e-05, -2.579220421011e-06, {   0,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -3.503285921115e-05, -2.862299813391e-05, {   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.664782311154e-05,  1.677617870498e-05, {   2,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -4.664075098430e-05,  1.678105927434e-05, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -6.205610756484e-05, -1.361506955959e-06, {   2,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -3.921756311773e-05, -2.416546513663e-05, {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.601215726272e-05,  3.735636778030e-05, {   2,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.870251517162e-05, -2.419164666265e-05, {   0,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  3.547517700327e-05,  2.686525110468e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.229117974738e-05,  2.897940200821e-05, {   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.833824570356e-05, -1.278334863415e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.008069655803e-05, -4.081899774410e-05, {   0,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -5.575573202034e-05, -4.648692275125e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  3.885525927224e-05,  2.150089156196e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   6,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   4,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   4,   0,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   4,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   3,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   3,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   1,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   0,   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.647250918031e-05, -2.475352335459e-06, {   0,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.013623509394e-05,  3.880995328436e-05, {   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.080877175357e-05,  3.805781977748e-05, {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.080023498378e-05, -3.803839146718e-05, {   0,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  1.509762643087e-05, -4.361244597415e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  3.711239445236e-05,  2.149030991757e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -4.808469078515e-05,  1.041167289309e-05, {   1,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.777630444971e-05,  2.013572813249e-05, {   0,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.656649869438e-05, -2.130918783263e-05, {   2,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  3.168025852541e-05,  2.603652105973e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  4.176603180362e-05,  1.593223614366e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.243137181851e-05, -4.448148523387e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.861810221490e-05,  2.794645282710e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.982083455529e-05,  6.733321324720e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  5.081554814613e-05, -5.663574427340e-06, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.638888294953e-05,  3.006038683510e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.619161226841e-05, -2.023048691644e-05, {   0,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.989443828323e-05, -3.650840828956e-05, {   0,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -4.959241690585e-05, -6.371199685800e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.208509607436e-05,  1.386244849278e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  5.556192098786e-05,  3.635551254460e-07, {   2,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -3.919366444811e-05,  1.672901738509e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  2.330805605766e-05, -3.250745334247e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.329209381751e-05, -2.251113344030e-05, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  6.560838980947e-07, -5.485483398329e-05, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -2.338066567089e-05, -3.183745305961e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.267691108667e-05, -4.253703767007e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -2.124394228803e-05, -3.389240204032e-05, {   2,   0,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.029686579840e-05, -3.414440439029e-05, {   2,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -7.023863619898e-07, -5.354445331920e-05, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  4.034353529968e-05,  1.352794590348e-05, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.009682647355e-05,  1.375470969559e-05, {   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.937619090413e-05, -2.428564400646e-05, {   2,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.741297114783e-05, -3.543396197553e-05, {   0,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  2.911695245898e-05, -2.344338265407e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.579537306129e-05, -3.674923386759e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.137779195384e-05, -2.108646328027e-05, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -4.983003030874e-05, -2.601902756523e-06, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.119586103926e-05,  3.072149621796e-05, {   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.142548621707e-05,  3.025658324385e-05, {   0,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  1.623202457883e-05,  3.526231512519e-05, {   4,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.803570509740e-05,  1.289082910951e-05, {   1,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.237070518244e-05,  3.803900173886e-05, {   4,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   4,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   2,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   2,   0,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   2,   0,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.656277817483e-05,  3.318261984611e-05, {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.050965639537e-05, -9.080311488086e-06, {   2,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.726079128503e-05,  1.218861560530e-05, {   2,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.967841497447e-05,  2.960099927140e-05, {   2,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.744501414005e-05, -3.176950381598e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.471920390196e-06, -4.753627737880e-05, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -3.867059656285e-05, -1.022667890732e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.867059656285e-05, -1.022667890732e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.233512496597e-05,  3.643791841903e-05, {   2,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.519629026248e-05, -3.348565281482e-05, {   1,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.526189090222e-05,  3.315401055493e-05, {   4,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.463006441950e-06,  4.491458140436e-05, {   0,   0,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  3.125850238008e-05,  1.687663295521e-05, {   2,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.595835849325e-05, -1.196239550994e-05, {   2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.064670777225e-05, -1.718379760589e-05, {   0,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.015293799151e-05,  3.765370678435e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  7.476145438243e-06,  3.999386144544e-05, {   0,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.462930029183e-06,  3.996408100458e-05, {   0,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.818281737862e-05,  1.916280713649e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.813661914278e-05,  1.917320467833e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.168982337794e-05,  1.533179624916e-05, {   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  3.645784271301e-05,  1.017737829504e-05, {   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.903152471434e-05, -1.710974094017e-05, {   2,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.446625828252e-05,  3.142484265608e-05, {   2,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.048984228216e-05,  3.489278478379e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  1.531421263952e-05, -2.977936001395e-05, {   1,   1,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.270953370999e-05, -3.232508646921e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  2.130311676134e-05, -2.339955257991e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -2.785378175519e-05, -1.684311966701e-05, {   0,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -5.003583933505e-06,  3.968581796839e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.660977301354e-05, -8.064154416287e-06, {   2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.400132163470e-05,  3.041518374486e-05, {   4,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.076503328692e-05,  2.350236470117e-05, {   1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.394716031186e-05,  3.030353906040e-05, {   2,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.338414680101e-05,  1.055752137454e-05, {   0,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.610025083448e-05, -2.778632232104e-05, {   0,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -1.379705274029e-05,  2.997479768160e-05, {   4,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.268348174418e-05,  1.077986607465e-05, {   2,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.593816902951e-05, -2.705877994102e-05, {   2,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -4.181074739661e-05, -9.250477343552e-07, {   2,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -2.689577147592e-05, -1.577262836477e-05, {   2,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.135376841495e-06, -4.139660864794e-05, {   2,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  4.004122001090e-05, -2.486275283423e-06, {   0,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.028251963403e-05, -3.222069052470e-05, {   0,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -1.027564047192e-05, -3.219291846231e-05, {   0,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -3.222505377642e-05,  1.023985114551e-05, {   2,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.085195485771e-05,  2.156840231941e-05, {   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.035958227040e-05,  2.205405675640e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.151651859292e-05,  3.088873727989e-05, {   2,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  4.085510123184e-05,  1.541174350279e-06, {   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.219276291332e-05, -2.018616541776e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.437857948112e-05,  7.927372351047e-06, {   2,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.927766328523e-05,  2.298633720846e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.910756322243e-05, -2.308702839994e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -3.492667053184e-05, -7.181795649159e-06, {   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.136887247249e-05,  2.064840832678e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.858225490747e-05,  1.335011716464e-05, {   2,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.723029640455e-05, -2.455843817940e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.072384239563e-05, -1.024632371209e-06, {   2,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -4.080854590958e-05, -7.885879430558e-07, {   2,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.636673839852e-05, -2.514219310630e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.597706212335e-06, -3.188651737975e-05, {   0,   2,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.597706212335e-06, -3.188651737975e-05, {   0,   2,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.475142052210e-05, -2.664676695083e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -1.353957725903e-05,  2.773573770818e-05, {   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.317065994409e-05, -1.807510127048e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.327472232920e-05,  1.792998926399e-05, {   2,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.457684375721e-05,  6.608441426633e-06, {   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.134540955887e-05,  9.819053089179e-06, {   0,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.523608404248e-05,  2.584302116724e-05, {   0,   0,   1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.627150026051e-05,  1.476570253885e-05, {   2,  -2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -2.639167577992e-05, -1.426462230582e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.578950635074e-05,  1.470992910727e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.233076541622e-05, -1.814595140592e-05, {   0,   0,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.284226604173e-05,  7.557047044780e-06, {   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.874381104956e-05,  1.518247071259e-06, {   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.298874035161e-05,  2.704242267398e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   6,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   6,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   4,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   1,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   1,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.617354746163e-05,  1.361911849256e-05, {   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.965179941439e-05, -9.935599211301e-06, {   2,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.441380337241e-05, -2.500540830689e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.977963762052e-05, -9.608860746521e-06, {   0,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.919157384384e-05, -1.593933010939e-07, {   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.195028101659e-05, -1.734741903742e-05, {   2,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -2.195028101659e-05, -1.734741903742e-05, {   2,   0,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -2.782536115177e-05,  1.121379849888e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.528947277418e-05,  2.356435313066e-05, {   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.209545766364e-05,  2.656081298135e-05, {   3,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.635133018728e-05, -1.193373150356e-05, {   0,   0,   0,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 { -2.475635395788e-05, -1.340264616551e-05, {   2,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.118839689491e-05, -1.675370163021e-05, {   2,   0,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  3.059980698153e-05, -7.327933920362e-06, {   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.386045744127e-06, -3.153358234506e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.838499341851e-05,  9.460242436203e-06, {   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.885101290626e-06,  2.788029802441e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.948115942561e-05,  1.824720119557e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.189599838327e-05,  1.583040768106e-05, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.105828257213e-05, -1.665194029677e-05, {   2,   0,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.205311370688e-05,  2.549929192905e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 {  3.744030263490e-05,  3.320845118664e-08, {   2,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.647228125685e-05, -1.098085728822e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.440022945081e-05, -2.288469722731e-05, {   0,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -2.787333663296e-05, -9.145888774224e-06, {   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.549794963870e-05, -1.311100427653e-06, {   2,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -2.898092994641e-05,  7.752786559759e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.429761828377e-05, -1.224854169555e-05, {   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.099122110911e-05,  1.546839637666e-05, {   2,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -2.194175435170e-05, -1.449523360955e-05, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -7.599763449156e-06,  2.879640303906e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  3.219406304675e-05, -4.125826823961e-06, {   2,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.376922510945e-06, -2.688420728757e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -3.154791272118e-05,  4.540134092185e-06, {   2,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.148561385048e-05,  2.458053166779e-05, {   4,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.617136946371e-05,  9.872319607350e-06, {   2,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.009870011774e-05, -2.593562430609e-05, {   1,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.009870011774e-05, -2.593562430609e-05, {   1,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.080307244733e-05,  1.515497729677e-05, {   2,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  6.714936998456e-06,  2.923883754712e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.329987661645e-05, -2.257319616840e-05, {   2,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  2.085390579427e-05,  1.499015034293e-05, {   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.828366396181e-05, -7.516457585100e-06, {   2,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.316436884756e-05, -1.230184811802e-05, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.981963521790e-05, -1.562294169415e-05, {   2,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.930374898097e-06,  2.945576787736e-05, {   1,   1,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  1.374347442075e-05,  2.158880810233e-05, {   2,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.100620397705e-05, -1.420093851333e-05, {   2,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.419966936567e-05,  2.072416541061e-05, {   2,   0,   0,   0,   0,  -8, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.337007947941e-06, -2.952145569181e-05, {   0,   0,   1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.612316166472e-05, -8.592863444961e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  2.261294445439e-05,  1.196710265157e-05, {   2,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.921979766021e-06, -3.258124357811e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.888261990968e-06, -3.260311021258e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.463613398239e-05, -1.981397362958e-05, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.838805694295e-05, -5.920891350439e-06, {   2,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.091014194705e-05,  2.335084100637e-05, {   0,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.824277906300e-06, -2.531317863948e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.012006024791e-05,  1.393983330173e-05, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.430408178628e-05, -1.973270750624e-05, {   2,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.008004059169e-05, -2.387744356398e-05, {   0,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.055788529157e-05, -1.339188010352e-05, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.974807680812e-05,  1.413573298098e-05, {   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.096724051119e-05,  2.281894013591e-05, {   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.065250556868e-05,  2.312720321658e-05, {   2,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.856595819185e-05,  1.511582569340e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -1.712457069267e-05,  1.652915990952e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.058632013802e-05,  2.260550405046e-05, {   4,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.098686418792e-05, -1.210181060568e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  2.098686418792e-05, -1.210181060568e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.402710265920e-05,  1.894156203814e-05, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.880648532312e-06,  2.899755290227e-05, {   0,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.984695891942e-05, -3.026391128433e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.145124556403e-06,  3.066755241022e-05, {   2,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.333575862847e-05, -9.335971215137e-06, {   0,   0,   1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.389769309539e-05,  1.870985501252e-05, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.968524699969e-06,  2.361865166425e-05, {   1,   0,   0,   0,   0, -23,  25,   0,   2,  -5,   0,   0,   0 } },
 { -2.156481772809e-06,  3.041606501242e-05, {   2,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.709013403490e-05,  5.404559768585e-06, {   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.432434657218e-06,  2.679873775110e-05, {   0,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  2.548173251538e-05,  6.732223748984e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.466383832109e-06, -2.969568280336e-05, {   0,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.011893543908e-05, -2.198382696028e-05, {   0,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.688440312320e-05, -5.118815125865e-06, {   1,  -1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -7.706572262555e-07, -3.107941813427e-05, {   2,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.746857383878e-05, -1.427098163697e-05, {   2,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.097506377683e-05,  1.075363591608e-05, {   1,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.609389009048e-05, -1.562040064319e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -2.654313444970e-05,  5.022918676340e-06, {   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.671361966174e-05, -4.779542400608e-06, {   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.531526840528e-05, -1.592934705600e-05, {   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.001971988289e-06,  2.323424767552e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -2.563429786960e-05,  5.346563077118e-06, {   2,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.816774766411e-05, -1.263822665148e-05, {   2,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -7.862244327347e-07, -2.998969574940e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.317942093757e-05,  7.564133754106e-06, {   2,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.582996387588e-06, -2.081756930094e-05, {   0,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.188359807693e-05, -1.850664516670e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -6.460481147893e-06, -2.389553802943e-05, {   1,   0,   1,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 {  2.690654782620e-05,  3.287059517644e-06, {   0,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.689638303873e-05,  3.284796205850e-06, {   0,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.973279481598e-05, -1.041261904181e-05, {   0,   0,   0,   0,   0,   5,  -2, -12,   3,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   6,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   6,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   4,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   4,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   3,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,  -2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   1,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   0,   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.918787974660e-05,  4.632817813118e-07, {   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  6.833460883418e-06,  2.270225497210e-05, {   2,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.811657190249e-06,  2.262959232958e-05, {   2,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.163335732735e-06, -2.019717905572e-05, {   2,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.136932560113e-05, -1.796493583112e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.515673439534e-05, -1.417394929239e-05, {   2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.704496198695e-05,  2.059846734052e-06, {   1,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.845762744494e-05,  1.061759331787e-05, {   2,   0,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  2.699469576943e-05,  2.054288542787e-06, {   1,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.036033770956e-05,  1.856363848683e-05, {   0,   0,   0,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -2.444427972222e-05,  4.418864965923e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.796561157435e-05, -1.082939150795e-05, {   2,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -1.602310241314e-05, -1.266925085527e-05, {   2,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  1.188864520767e-05, -1.679349009724e-05, {   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.779827756807e-05, -7.815748514819e-07, {   0,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  1.361649868625e-05, -1.464892363033e-05, {   2,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.763627208551e-05, -4.095468050776e-07, {   2,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.617456018412e-05, -1.176365601547e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.197539921278e-06, -2.169529204582e-05, {   2,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.663878093053e-05, -1.183997434034e-06, {   2,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.048012521651e-05, -1.729650116822e-05, {   2,   0,  -1,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 { -1.600846037204e-05,  1.176160352157e-05, {   2,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.669822594234e-05, -1.062956269423e-06, {   2,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  1.671734490989e-05,  1.097863284584e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.309641997471e-06, -2.138411417115e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.092148361819e-05, -1.675473651173e-05, {   2,   0,   0,   0,   0,   0,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.995168732809e-06, -2.166729704855e-05, {   1,   0,   0,   0,   0, -17,  15,   0,   3,   0,   0,   0,   0 } },
 { -1.741661618419e-05, -1.019842957952e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  2.075648843630e-05,  6.810600250573e-06, {   0,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.550953531412e-05,  1.204816458612e-05, {   2,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.686077098180e-05,  1.066010726498e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.400796951210e-05, -1.343394892435e-05, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.273989727324e-05, -4.681641149125e-06, {   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.289949341702e-06, -1.912768085646e-05, {   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.835508751718e-06,  1.857863607081e-05, {   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.050140247487e-05,  6.811661160797e-06, {   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.712864263940e-06,  2.158476911367e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  4.648851283957e-06,  2.263548978897e-05, {   0,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.642229490796e-05, -1.080777948010e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -9.617822813089e-06, -1.753560618673e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.655766054247e-05, -5.393590861229e-07, {   2,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.336167271804e-05, -3.690107338188e-06, {   4,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  9.866795881934e-06, -1.717586130888e-05, {   2,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -2.332216804388e-05, -3.689326626593e-06, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.515150905252e-05, -1.177030889791e-05, {   2,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.037780862476e-05, -1.648510011550e-05, {   2,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  4.362522267665e-07, -2.640303548047e-05, {   2,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -2.642013276896e-05,  3.691262541576e-07, {   2,   0,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  9.816575501133e-06, -1.696900963540e-05, {   0,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.613893929362e-05, -1.062552080433e-05, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.611956599589e-05, -1.062147055564e-05, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.463309862796e-05, -1.206650879650e-05, {   0,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.051248879517e-05,  6.144947861974e-06, {   2,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.440438943567e-05,  1.221441338734e-05, {   2,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.747535406620e-06, -1.686359973968e-05, {   0,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.141063993272e-05, -1.518869054682e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -8.414478393352e-06, -1.814377450719e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.767687567251e-06, -1.677657587760e-05, {   2,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.313076791445e-05, -1.335660902268e-05, {   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.002299654438e-05,  6.432194890357e-06, {   2,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.830607352948e-05,  8.055288445081e-06, {   2,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.131870677579e-05,  1.501804956541e-05, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.425258456747e-05, -1.207810579237e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  4.418780378031e-06, -2.187204716284e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.534335674993e-05,  1.093109553153e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -9.822306057980e-06, -1.625473755972e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.466660291839e-05, -1.395324986860e-06, {   2,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.466008816500e-05, -1.398978965973e-06, {   2,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.134708683745e-05, -1.471166473193e-05, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   1,   0,   0,   0 } },
 {  7.843041384484e-06, -1.820890691485e-05, {   2,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.944156842288e-05,  6.609492213020e-06, {   2,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.857540682327e-06, -2.015990036599e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  2.175560682570e-05,  4.248687890003e-06, {   0,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.334830100761e-05, -1.260479026474e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.499797234569e-05, -8.057649002845e-07, {   2,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.406807870025e-05,  1.721078518709e-06, {   2,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.689958072959e-06,  2.206415319322e-05, {   0,   0,   0,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 {  1.874371783611e-05,  6.976606745433e-06, {   2,   0,   2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.368689378367e-06, -1.630655442089e-05, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.530963547251e-05, -1.032773195652e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -6.771211247637e-06,  1.881889205665e-05, {   0,   0,   3,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.771211247637e-06,  1.881889205665e-05, {   0,   0,   2,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.007909033717e-05, -5.307225743751e-06, {   2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.392824801891e-06,  1.895077283523e-05, {   2,   0,  -3,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.256546657871e-05,  1.270250515603e-05, {   2,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -6.185352591220e-06,  1.901950086943e-05, {   2,   0,   1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.664695820340e-06, -1.942251726967e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -9.708802324641e-06, -1.537351568200e-05, {   2,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  2.090359278213e-05, -4.057302815792e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.142042670684e-06, -2.273629086269e-05, {   1,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.142042670684e-06, -2.273629086269e-05, {   1,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  5.661259813142e-06, -1.919110058610e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -1.835778258674e-05,  6.475143931260e-06, {   0,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.267204699660e-05,  1.209828913771e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.420311754090e-05, -4.860328720488e-07, {   2,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  9.247619954727e-06,  1.543332580077e-05, {   2,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  6.349344699391e-06,  1.828591244726e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -1.525278220947e-05,  9.226349435422e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.057064228833e-05,  1.376305727893e-05, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.939548117273e-05, -4.880093244839e-06, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.378340446840e-05,  4.507279208111e-07, {   0,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -1.396701948144e-05,  1.022537175215e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.684284987142e-06,  1.944370011493e-05, {   2,   0,   2,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.057270408809e-05,  1.349032729839e-05, {   2,   0,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  5.437610304325e-06, -1.850509856527e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  1.643386050114e-05,  7.495193737557e-06, {   2,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.681989223001e-06, -2.220404544382e-05, {   2,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.953214266997e-05, -4.300628177348e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.959415014611e-05, -4.050246427108e-06, {   3,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.793615695714e-05,  5.647029157254e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.043355376098e-05,  1.312220782687e-05, {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.120129049060e-05, -1.234567829330e-05, {   4,  -2,   0,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.644561414467e-06,  1.487605859894e-05, {   0,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -9.433731781814e-06, -1.405989757129e-05, {   2,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.154779488069e-05,  1.193595331226e-05, {   2,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  2.181850703110e-05, -1.644484483728e-06, {   2,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -1.466057589876e-05, -8.765323222590e-06, {   2,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  8.630468951623e-06, -1.461053155734e-05, {   2,   0,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -6.930274374379e-06,  1.630339259108e-05, {   3,   0,  -3,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  1.963733167650e-05, -3.594305620116e-06, {   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.254798264101e-05, -5.482280515125e-07, {   0,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.424947947954e-05, -8.821322702498e-06, {   0,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  8.234054093247e-06,  1.472806833833e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.291050162203e-05,  1.003119971606e-05, {   0,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -1.780299657797e-05,  5.067670356085e-06, {   3,  -2,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.149990191052e-05,  1.103550061271e-05, {   0,   0,   0,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -7.102035907374e-06, -1.542800825830e-05, {   2,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.609723424355e-06,  1.688030904769e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -1.263079405555e-05,  9.824612942321e-06, {   1,   1,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.504319609657e-05,  7.352010631607e-06, {   0,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  6.378355164708e-06, -1.597959895000e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  1.710440130111e-05, -5.227703311762e-06, {   0,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -7.110523587970e-06,  1.502789877766e-05, {   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.453526090236e-05, -7.538423677104e-06, {   2,   0,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  6.953083105913e-06, -1.510428596672e-05, {   2,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.471135098050e-06, -1.354622181026e-05, {   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.988769724774e-05,  2.116482502197e-06, {   1,   1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.647199487557e-05,  5.492773543617e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  7.923790245107e-06,  1.399562929642e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.471456309335e-05,  7.111893488511e-06, {   0,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.794727545280e-05,  3.791556703165e-06, {   0,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.590514092439e-05,  5.788112664884e-06, {   2,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.041885854177e-05, -1.214587519599e-06, {   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.066837624806e-06,  1.345276187721e-05, {   2,   0,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -1.403323985774e-05, -7.343490124800e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -4.463405161993e-06,  1.690772313863e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -1.982798885533e-05,  1.483248680186e-06, {   2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.142309482360e-06,  1.512767410525e-05, {   2,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  8.308268985522e-07, -2.040803408154e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -8.107299575403e-07, -2.041171319372e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -7.053125560717e-06,  1.396435854950e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  6.375964992971e-06,  1.453012958366e-05, {   1,   1,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.064939582672e-06, -1.277855841094e-05, {   1,   1,   0,   0,   0, -38,  36,   0,   0,   0,   0,   0,   0 } },
 { -1.998325377805e-05, -8.182716188366e-07, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.741933260331e-05,  3.328301617019e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -5.969574423241e-07,  2.012518812513e-05, {   1,  -1,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -6.458752289320e-07, -2.004776140513e-05, {   0,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.059557866844e-05,  1.008004082054e-05, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.250815138098e-06, -1.339350775785e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.301441941371e-05,  7.628890389581e-06, {   2,   0,   0,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 {  7.047731702611e-06,  1.353795311616e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.131995049388e-05, -9.164633392002e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 {  1.661341774675e-05, -3.805105787646e-06, {   2,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.169596186885e-05,  8.721441914422e-06, {   0,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.085507533882e-05,  9.539643027803e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.535881994487e-05,  4.991877362356e-06, {   2,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.869590104938e-05, -1.652591407642e-06, {   2,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -1.035306539346e-05,  9.972300727511e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.438383751880e-05, -5.864415276543e-06, {   0,   0,   1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -4.269622456150e-06,  1.597162367059e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  1.513718239378e-05,  5.060450432998e-06, {   0,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.038614150129e-06,  1.115334010316e-05, {   2,   0,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -1.137271412866e-05, -8.789877161573e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  7.007480391543e-06, -1.314742982302e-05, {   2,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.397500564038e-05,  6.158917706752e-06, {   2,   0,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -8.973129923512e-06,  1.110607914419e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  1.504093838117e-05,  4.981785072916e-06, {   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   6,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   5,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,   0,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   3,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   3,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   3,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,  -2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,  -2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   1,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   1,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   1,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   1,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   1,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.067817037854e-05, -9.300465716223e-06, {   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.656301638257e-05, -3.410005184219e-06, {   1,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.649354228223e-05, -3.439689842654e-06, {   2,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.510885190581e-06,  1.733201700480e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.910970747459e-07, -1.912932436731e-05, {   0,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.410722802665e-05, -5.686991491146e-06, {   3,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -3.267789338981e-06,  1.650745651564e-05, {   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.057916549266e-06, -1.362170969623e-05, {   3,   1,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -1.692649804849e-05,  2.717162249595e-06, {   2,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.893375052939e-06, -1.072678013712e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.072762250541e-05, -8.857493558666e-06, {   0,   0,   1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.747978501764e-06, -1.282828306846e-05, {   0,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.953610268415e-06, -1.758964180102e-05, {   2,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.780702037662e-06, -1.263790771615e-05, {   2,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.665847133267e-05, -2.622639669390e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.665847133267e-05, -2.622639669390e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -1.644452476701e-05, -2.801392244886e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  7.744346285764e-06,  1.149126265936e-05, {   0,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.277750611566e-05,  6.404574500148e-06, {   0,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  9.810452546683e-06,  9.316464168462e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -9.798105269898e-06, -9.276314661910e-06, {   2,   0,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.525152792383e-05,  3.811088966391e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.147973168763e-05,  7.503423789435e-06, {   1,  -1,   1,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 {  5.973228776328e-06,  1.298878859407e-05, {   2,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.528006313890e-05, -3.659062647457e-06, {   2,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -6.025350601076e-06,  1.285729107383e-05, {   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.310956572402e-06, -1.149564819178e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.782754085310e-05,  9.325351028862e-07, {   0,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.172277694087e-05, -7.027710434100e-06, {   0,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.683079006941e-05,  1.867289750631e-06, {   1,  -3,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.017726374443e-05, -8.402978625111e-06, {   4,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.233970663017e-06,  1.031576054331e-05, {   0,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.723341733728e-05, -1.289394061714e-06, {   2,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.393021602310e-05,  4.582364357882e-06, {   0,   0,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  3.490259539950e-06, -1.496347441421e-05, {   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.986248722255e-06,  8.464991453706e-06, {   2,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.171693279823e-05, -6.680215686740e-06, {   2,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.736465363311e-06, -1.246136606564e-05, {   2,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.126742851398e-05, -6.897533009365e-06, {   0,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -1.126143758716e-05, -6.892724495085e-06, {   0,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.502106080965e-05,  3.055070280571e-06, {   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.902717215468e-06, -8.167389684511e-06, {   4,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.685210850109e-06, -1.234997354193e-05, {   2,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.026770302861e-07,  1.721362833020e-05, {   2,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.835549130042e-06, -9.001751813332e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -5.710680145114e-06,  1.212260170417e-05, {   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.158393642233e-05,  6.205888079231e-06, {   2,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.074931588228e-05,  7.026819349181e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -9.444212119454e-06,  8.331747826581e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -1.109384846640e-05, -6.645213073655e-06, {   0,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.325746265513e-07,  1.700529031350e-05, {   2,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.260845597956e-05,  5.127976523209e-06, {   0,   2,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.722713746630e-05, -4.801196954752e-07, {   2,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -6.677861428241e-06,  1.089097227533e-05, {   0,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  6.676838829719e-06,  1.089153757748e-05, {   0,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -7.755413589322e-06,  9.804055056452e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  9.042240877496e-06, -8.511515222444e-06, {   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.960072468104e-06,  7.581113062782e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.055113496088e-05, -6.836167724413e-06, {   0,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.917435390012e-06, -1.044002925019e-05, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -1.083985336636e-05, -6.494604130001e-06, {   2,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.393630555534e-05, -3.339372009880e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.297488841921e-05,  4.270166089587e-06, {   4,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -8.817279357909e-06, -8.422148838249e-06, {   2,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -4.243307768285e-06,  1.297957538404e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.286140396474e-05, -4.328194966006e-06, {   0,   0,   0,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 {  1.416366469885e-05,  2.900162613936e-06, {   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.231542937188e-05, -4.706791089473e-06, {   2,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.453865801673e-05, -2.472160276201e-06, {   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.061650397386e-06, -1.383310291986e-05, {   2,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.271044286176e-05,  4.179286349178e-06, {   4,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.275879982154e-06, -1.359486544069e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.730096463114e-06,  1.502513458087e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  1.976929984229e-06, -1.473481790078e-05, {   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.315229434870e-06,  1.138744143190e-05, {   2,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.388540270472e-05, -2.799761793755e-06, {   1,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.386311132349e-05, -2.741184075059e-06, {   1,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.063673704932e-05,  5.863368286911e-06, {   2,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  8.877069498548e-06, -7.601136235985e-06, {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.490100403586e-06,  1.494756036256e-05, {   0,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.160179718279e-05,  4.810786182981e-06, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.585883554758e-05, -4.031880682834e-07, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.348004461605e-05, -2.765179155211e-06, {   1,   1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.299445599626e-07,  1.558954303222e-05, {   2,   0,  -1,   0,   0, -23,  26,   0,  -1,   0,   0,   0,   0 } },
 { -1.075709074913e-05, -5.422239505935e-06, {   2,   0,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -9.259819993800e-06,  6.911503524974e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  1.072723463194e-05,  5.433083611726e-06, {   0,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  2.789820109329e-07, -1.579087770819e-05, {   0,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  1.215963464531e-05, -3.878049950794e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  5.048520586580e-06,  1.092505754531e-05, {   2,   0,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  9.969056291530e-06,  5.969717093660e-06, {   2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.019998564249e-06,  1.089651217254e-05, {   0,   0,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.816956792368e-06, -1.008857829148e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -7.573082660707e-06,  8.329186205027e-06, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.512134365660e-05, -7.738369927488e-07, {   0,   0,   0,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 { -7.362958898182e-06,  8.512520437340e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.268145537474e-05, -3.158085614682e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -3.965263665805e-06, -1.185373194835e-05, {   0,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  6.474869045700e-06,  9.295647663587e-06, {   0,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  9.217079462633e-06, -6.515730550746e-06, {   2,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.343900726922e-06, -6.372042975087e-06, {   2,   0,  -1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -8.928142927487e-06, -6.779125470352e-06, {   0,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.324273002642e-06,  1.237591857602e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  3.998353921854e-06, -1.166232680022e-05, {   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.303762258499e-06, -6.341092028926e-06, {   2,   0,  -1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.560751856555e-05,  3.180585263481e-08, {   2,  -2,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -1.178066763213e-05, -3.843992279009e-06, {   2,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.814974750147e-06, -5.795642372409e-06, {   3,  -2,   0,   0,   0,   0, -31,  39,   0,   0,   0,   0,   0 } },
 { -5.224438055633e-06,  1.033311226163e-05, {   2,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.743118026986e-06, -5.807179885113e-06, {   0,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.547812422333e-06, -1.398002431290e-05, {   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.630510639628e-06, -8.896778841120e-06, {   2,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -1.045137539796e-05,  4.957296432018e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.052324243933e-06,  8.337240139454e-06, {   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.143262714909e-06, -1.323814437836e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -1.009764995309e-06, -1.420902534354e-05, {   2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.787523955818e-06,  5.388510455766e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -1.027662975776e-05,  4.859011913765e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  9.928867217638e-06, -5.188810353664e-06, {   2,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  6.547281854232e-06,  8.561541397020e-06, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.451699325156e-05, -5.597862951986e-07, {   0,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  8.884349427659e-06, -6.160273548970e-06, {   2,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.114031652249e-05, -3.888937666879e-06, {   4,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.158722159339e-05, -3.433257922519e-06, {   2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.434020789132e-05,  6.631362525749e-07, {   2,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.109086242797e-05, -3.892790263764e-06, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.239318658657e-05,  2.513717846722e-06, {   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.119466805886e-05,  3.698086352072e-06, {   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.329131884880e-05, -1.591213681629e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  2.158476175427e-08, -1.483655886757e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.229400251861e-05, -2.529629156531e-06, {   1,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.376307085004e-06, -5.429405190596e-06, {   2,   0,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -8.562214486159e-06, -6.240419961724e-06, {   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.313973436331e-05, -1.572706757396e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  4.620187509180e-06,  1.005727615413e-05, {   2,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -9.305832446040e-06, -5.361510154959e-06, {   0,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -9.729509863587e-06,  4.926867902874e-06, {   2,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.075684677323e-05, -3.863814002055e-06, {   0,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.096365064111e-05,  3.651782040043e-06, {   0,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.323443187079e-06,  8.269434948564e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 {  6.497775058739e-06,  8.078776435412e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 {  1.069042646446e-05, -3.845547144586e-06, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -6.704122086327e-06,  7.826782399513e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 { -4.610490640393e-06,  9.851508761911e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -8.226577986012e-06,  6.182358212847e-06, {   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.596897868279e-06,  9.792667916457e-06, {   0,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.345759323135e-05,  8.972069098185e-07, {   2,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.244411635955e-06,  7.101827527838e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -1.078521003105e-05, -3.508929095129e-06, {   2,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.585316238643e-06,  7.700569023219e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -9.001791995321e-06, -5.269391211796e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  3.597835921134e-06, -1.065742417000e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  1.003724381315e-05, -4.160731747665e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.311439255801e-06,  1.183252252387e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -1.564282755587e-06, -1.256044664163e-05, {   2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.531973900694e-06, -6.577945664055e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.072926405493e-05, -3.307896479608e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.465314723360e-06, -9.560546128280e-06, {   1,  -2,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.639814406000e-07, -1.369510091539e-05, {   0,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.052100807274e-05,  3.406430179430e-06, {   2,   0,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.022657741624e-05, -3.667061837513e-06, {   0,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -8.404459420598e-06,  5.418953944030e-06, {   2,   0,   1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.577954731419e-06, -9.244618017187e-06, {   2,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.142524330784e-05,  2.383021921537e-06, {   0,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.332944136420e-05,  4.037854145299e-07, {   2,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -8.565781341282e-06,  5.160173447990e-06, {   2,   0,  -2,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.330401844231e-05,  4.004674700090e-07, {   2,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.896628359344e-06, -8.767295880208e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.006675348384e-05, -3.464555798582e-06, {   2,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.284464617370e-05,  6.741558349113e-07, {   0,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.426997228493e-06, -1.005410580649e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -4.247234425913e-06, -9.226200690837e-06, {   1,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.130120914250e-05, -2.147328066296e-06, {   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.468293637334e-06, -3.924744129510e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.154935940759e-06, -1.015618059634e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -1.161038495620e-05,  1.673995652708e-06, {   0,   0,   1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  1.160525897020e-05,  1.674441237411e-06, {   0,   0,   1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  4.087965763211e-06, -9.126255306467e-06, {   2,   0,  -2,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.011747362556e-05, -3.029745497314e-06, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.945244403727e-06,  3.168507675789e-06, {   4,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.013553903414e-05,  2.977549897479e-06, {   0,   0,   1,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  2.781638883611e-06, -1.032497086837e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -3.804530752362e-06, -9.248002257479e-06, {   0,   0,   2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.082010946884e-05, -2.223390184840e-06, {   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.887331881579e-06,  3.141795391318e-06, {   3,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.593533768645e-06,  9.332015594372e-06, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.593533768645e-06, -9.332015594372e-06, {   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.403050282759e-06, -9.490473712679e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  1.227147770450e-05,  5.944760749381e-07, {   4,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.191045062547e-05, -9.193471528700e-07, {   4,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.416092723592e-06,  9.398420638800e-06, {   2,   0,  -3,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.188274294511e-05, -9.145817681111e-07, {   4,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.385605623819e-06, -9.409446028325e-06, {   4,   0,  -1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.385605623819e-06,  9.409446028325e-06, {   2,   0,   3,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.457419485882e-06,  3.249187078035e-06, {   0,   0,   1,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.458875956054e-06,  3.244944629418e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.458875956054e-06,  3.244944629418e-06, {   2,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.049084441636e-05,  2.188068006815e-06, {   1,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.201345154206e-06, -9.473720979828e-06, {   0,   0,   2,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.938127387708e-06,  2.735314664597e-06, {   2,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.485279216763e-06, -3.166935139856e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.092674635846e-06,  9.509750974489e-06, {   2,  -2,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.588397280333e-06, -1.089501454355e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.048444046053e-05, -1.880502171912e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  1.087780572908e-05, -1.399720792988e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 {  1.036483621358e-05, -1.911206082864e-06, {   1,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.667649140712e-06, -2.556669726830e-06, {   3,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.003209256170e-05,  2.047884752150e-06, {   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.027296576063e-05, -1.791738096890e-06, {   1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.047359115180e-05,  1.507687737448e-06, {   0,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.041453161441e-05,  1.498543762148e-06, {   0,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.991654655052e-06, -1.816485247590e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -9.809870761638e-06, -1.940730697432e-06, {   2,   0,  -3,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.864493529258e-06, -1.743958450518e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.436629536627e-07,  1.124732310690e-05, {   2,   0,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  9.864752632105e-07, -1.054077422746e-05, {   2,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.040765968804e-05, -1.024804607678e-06, {   0,   2,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.040699374589e-05, -1.024671382462e-06, {   0,   2,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  9.891711313476e-06, -1.467667295695e-06, {   2,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.936653342300e-07, -1.104585831185e-05, {   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.106078884769e-05,  2.576655496112e-07, {   2,   0,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -9.992585429831e-06, -7.337382090487e-07, {   0,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.058410954268e-05, -6.383095153003e-08, {   2,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  9.991626889023e-06,  4.091358094183e-07, {   2,   0,  -3,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.991626889023e-06,  4.091358094183e-07, {   0,   0,   2,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   6,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   6,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   6,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   6,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   6,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   5,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   1,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   0,   2,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   0,   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.732031912100e-06,  1.438529146850e-06, {   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.182831527790e-06, -5.260501813920e-06, {   2,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.152242623410e-06, -1.676651644953e-06, {   1,  -1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  4.263931723950e-06, -7.997513366130e-07, {   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.138297099200e-06,  1.240016367047e-06, {   0,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.136751472930e-06,  2.672167251110e-07, {   1,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.345144752200e-07, -1.610755632400e-07, {   1,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.546616319400e-07,  4.136764233280e-07, {   2,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.143949790900e-07, -1.021700431200e-07, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 219> g_PERT_S1_3
{ {
 {  4.870000000000e-03,  0.000000000000e+00, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.089962898161e-03, -9.113098687827e-04, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.500000000000e-03,  0.000000000000e+00, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-03,  0.000000000000e+00, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.080000000000e-03,  0.000000000000e+00, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-04,  0.000000000000e+00, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.428031142083e-04,  7.569379221078e-05, {   0,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.100000000000e-04,  0.000000000000e+00, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.800000000000e-04,  0.000000000000e+00, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.156307458855e-04, -5.043063409084e-05, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.084775528560e-04, -4.730961545363e-05, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.300000000000e-04,  0.000000000000e+00, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-04,  0.000000000000e+00, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.100000000000e-04,  0.000000000000e+00, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.100000000000e-04,  0.000000000000e+00, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.100000000000e-04,  0.000000000000e+00, {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.908847172995e-05, -2.168359142010e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.613668288081e-05,  6.485092371070e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.612480750583e-05,  6.483663110662e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -3.130804630903e-05, -4.365084981981e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.632910495323e-05, -2.293093229169e-05, {   0,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.248961081863e-05,  3.399492054428e-05, {   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.579400813894e-05,  2.495535549946e-05, {   1,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.449808981145e-05, -1.067994015286e-05, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.411409444794e-05, -1.051476435344e-05, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.391294811854e-05, -1.042686928656e-05, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.303742227503e-05, -1.004306011073e-05, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.286392499431e-05,  1.539001587257e-05, {   2,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.781919307049e-05,  4.582184747008e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.780781785864e-05,  4.583925501711e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.083362405184e-05,  1.141450067767e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.241072540073e-06,  1.751276411818e-05, {   2,   0,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  2.611839117673e-07, -2.041736682717e-05, {   2,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.805211767037e-06,  1.376875108087e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.352108999104e-05,  4.218591546774e-06, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.342030147320e-05,  4.185488347045e-06, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.647315435292e-06,  6.916659391974e-06, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.889513611636e-06,  1.063524754961e-05, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -6.890294347919e-06,  7.264601874810e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -5.706574770333e-06, -8.362533423621e-06, {   2,   0,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -3.940808542452e-06, -9.995307676326e-06, {   0,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.171688302396e-05,  7.776326613395e-07, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.158639753707e-05,  8.069215205410e-07, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.189836902931e-06,  6.340883723699e-06, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.184802775791e-06,  6.331302617167e-06, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  3.990263628796e-06, -7.357323706587e-06, {   0,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  7.855376073974e-06, -3.426005692496e-06, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -6.667900623785e-06, -4.340231054853e-06, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  6.621653201572e-06, -4.313187334024e-06, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -7.369299459445e-06, -3.213929482996e-06, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.119869609233e-06,  6.023577591865e-07, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.589145189935e-06,  5.901941717440e-06, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.580645198985e-06,  5.848631398111e-06, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.391404841570e-06,  2.813642472523e-06, {   1,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.726625260936e-06, -3.403063495304e-06, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.711899815517e-06,  6.119872350464e-06, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.503902914743e-06, -3.284302564117e-06, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.735873355985e-06,  4.792689478066e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.846328717197e-06,  5.213436863488e-06, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -9.083432135967e-07, -6.620460317556e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  5.096394325944e-06,  2.176640416293e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  6.859716674964e-06, -3.567491708975e-07, {   0,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  5.496082723930e-06, -1.526844892723e-06, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.775374799739e-06,  2.183321798173e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -5.436316285158e-06, -1.476470429502e-06, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.766471192514e-06, -3.932876526645e-06, {   2,   0,  -1,   0,   0, -20,  29, -15,   0,   0,   0,   0,   0 } },
 { -1.959971155379e-06,  4.536041268597e-06, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.775575826916e-06,  4.402981583188e-06, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.775289138209e-06,  4.402351596658e-06, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.398503573139e-06,  3.721234243931e-06, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.041772214989e-07,  6.008896457697e-06, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -7.205900630964e-08,  6.018130791445e-06, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.393533563259e-06,  3.680203033127e-06, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.005602658891e-06,  1.989275356683e-06, {   0,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -4.122264255312e-06,  1.797483622644e-06, {   0,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.122264255312e-06,  1.797483622644e-06, {   0,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.684354752434e-06, -2.224347264489e-06, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  3.514381690662e-06, -2.137616370318e-06, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  4.643691117684e-06,  9.785346064878e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -4.569766556498e-06,  9.684271761535e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.214479915350e-06,  3.273594635852e-06, {   0,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  3.703478047393e-06, -1.682204986207e-06, {   2,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.123209059993e-06,  2.991008382810e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.692353171043e-06,  3.349466192820e-06, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.308319886331e-06, -2.417024970589e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.887559306977e-07,  4.085188952410e-06, {   0,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.931917285651e-06, -1.278530499307e-06, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.854012142162e-06, -1.245036588474e-06, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.325280457793e-06,  2.759538292806e-06, {   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  5.948634964143e-08, -4.020960706582e-06, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -1.229166580267e-06,  2.850407638391e-06, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.687947763187e-06,  3.149107370796e-07, {   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.836036765358e-07,  3.685450091041e-06, {   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.767035479017e-06, -1.189093757012e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  2.789725072132e-06,  8.696958876485e-07, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.789725072132e-06,  8.696958876485e-07, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.684618909803e-06, -8.900730071068e-07, {   0,   0,   1,   0,   0,   0, -34,  39,   0,   0,   0,   0,   0 } },
 { -2.680135073237e-06,  8.351840977079e-07, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.118652074544e-06, -1.391250328986e-06, {   1,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  2.659713442913e-06,  8.284767730842e-07, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.119207716004e-06,  2.352214696531e-06, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.224960169758e-06, -2.236757053615e-06, {   0,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  1.079651490803e-06,  2.235194466862e-06, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.263966920765e-06,  3.674355839157e-09, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.043799413135e-07,  2.187711166867e-06, {   1,   0,   0,   0,   0,   0, -23,  20,   0,   0,   0,   0,   0 } },
 { -9.922778793270e-07,  2.069035562510e-06, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.021542275541e-06,  2.026667037727e-06, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -9.838431910945e-07,  1.987316843188e-06, {   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.832332939097e-07, -2.170541875998e-06, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.579658510893e-07,  2.537039975516e-06, {   2,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.443274166298e-06,  1.397796228987e-06, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.443274166298e-06,  1.397796228987e-06, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.387691167557e-06, -1.374609681887e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.375971974449e-06, -1.351809974701e-06, {   2,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.409307926232e-06, -2.951088036494e-07, {   0,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.372986680590e-06, -1.325198653430e-06, {   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.346319033208e-06, -1.339212245436e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.579912735250e-06, -9.054732442050e-08, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.343205112232e-06, -1.321405302229e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.291180053773e-07,  2.123510459721e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  7.068054649753e-07,  1.933981273550e-06, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.523729698902e-06,  1.099967279189e-07, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.514756220977e-07,  1.774401651717e-06, {   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.476639124192e-06, -6.717159159350e-08, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.080916530881e-06, -1.459502232884e-06, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -8.080008224968e-07,  1.719792424360e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.774133113868e-07, -2.203894687222e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -5.703584723430e-08, -2.389166654198e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  7.786657302091e-07,  1.665491632015e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.224875947169e-06, -1.417615063565e-07, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.495388632150e-07,  1.570504714346e-06, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.014011461222e-06,  1.279782912535e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.123149421199e-06, -1.147898629147e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.533149820636e-06, -7.368187967096e-07, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  9.130604627450e-07,  1.304077589455e-06, {   0,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -6.794032560801e-07,  1.532176009138e-06, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.474008925749e-07, -1.433370363400e-06, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.309795768346e-06,  8.606120129406e-07, {   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.114241403584e-06, -1.054043476135e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -7.213068564016e-07,  1.443384481055e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.272368736562e-06,  8.912456530377e-07, {   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.153423875530e-06,  5.875817221955e-09, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.384531276142e-07, -1.419721086223e-06, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  3.366000255781e-07,  1.794816649740e-06, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.467888788261e-06, -6.184423143506e-07, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  5.449426095814e-07,  1.523589294650e-06, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.111426472032e-06,  8.964941267701e-07, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.398811136777e-06, -5.870611444263e-07, {   0,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.097534650993e-06,  8.849228287666e-07, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.045085357691e-06,  9.342462626457e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  6.357786547949e-07, -1.313117202510e-06, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.301836019311e-06, -6.371715132408e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.249075144513e-06, -6.562338299750e-07, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.351084342372e-06, -5.536218457704e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  5.409594954294e-07,  1.336384668956e-06, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.329750303622e-07, -9.193028848751e-07, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.702839303527e-06,  1.055995005479e-07, {   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.711389973005e-07, -1.627923591259e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.271639710100e-06,  5.231662554058e-07, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  5.816741681475e-07,  1.197701119573e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.029176747623e-07, -1.170984124627e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  7.812009707216e-07,  9.874282907446e-07, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.529330760911e-07, -1.113245625137e-06, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  7.629453814556e-07,  9.789492952895e-07, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  8.593176234156e-07, -8.578144486616e-07, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -9.641793310754e-07,  7.330308565102e-07, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.602490774939e-06,  7.145294662477e-08, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.107081021179e-07, -1.052575714559e-06, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  8.079140093623e-08, -1.579346927052e-06, {   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.083849850040e-07,  1.551412261689e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.147592752159e-06,  4.816576941785e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  9.943537092537e-07, -6.265398245426e-07, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  8.464602612040e-07,  7.728701600520e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -9.758346602430e-07,  6.398152205771e-07, {   2,   0,   0,   0,   0, -46,  50,   0,   0,   0,   0,   0,   0 } },
 { -1.586893316214e-06,  1.318253244553e-08, {   1,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -7.279125506059e-07,  8.685177105791e-07, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.091601669135e-07,  6.870048449257e-07, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -6.725674798734e-07,  9.217247252079e-07, {   1,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.397275407190e-06,  1.859359114132e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.396119278881e-06,  1.855804078736e-07, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  6.697723713658e-07, -9.052682009717e-07, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -9.648335941455e-07,  6.053649169040e-07, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  1.167548994387e-06, -3.766651814835e-07, {   0,   0,   0,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -1.128847849434e-06, -3.698369942679e-07, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.039118176702e-06,  4.352898217776e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.488561265994e-07, -1.306108533199e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.031863972537e-06,  4.222345411623e-07, {   0,   0,   0,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 {  5.732303422469e-07,  8.642159916612e-07, {   0,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -4.744007884704e-07,  9.558591302474e-07, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.357015430313e-07,  9.939726500254e-07, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.110587695435e-06, -3.102251673109e-07, {   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.205889610616e-07,  1.176633262364e-06, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.521987534761e-07,  9.327881605629e-07, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.289151032158e-08,  1.309327564823e-06, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -1.295912846847e-06, -8.059999369597e-08, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -4.220890284799e-07,  9.341118340247e-07, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.266057664776e-06, -8.026707534204e-08, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.814809917418e-07, -1.159100123751e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -2.246771467142e-07, -1.049627262514e-06, {   0,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  2.156419322279e-07,  1.044072624190e-06, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.131147680805e-07,  1.041684642379e-06, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.844708302749e-07,  9.967652878286e-07, {   2,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -1.137844937680e-06, -3.170720510184e-08, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 2> g_PERT_S1_4
{ {
 { -5.460000981311e-06, -1.064920535165e-05, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.614435283874e-07, -1.595662150694e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 6462> g_PERT_S2_1
{ {
 { -8.045039999382e+00, -9.969282987996e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.756754655786e-01, -1.499956686934e+00, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.640575434610e-01,  2.817713704141e-01, {   0,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.638351875992e-01,  2.816603588851e-01, {   0,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.558599999609e-01,  5.967193445936e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.157099999699e-01, -5.006300680910e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.262199999633e-01,  4.896507253652e-06, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.985499999764e-01,  3.751689946818e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.842740711202e-03,  8.444544009640e-02, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.231903284083e-02,  3.113103232178e-02, {   0,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.038547186366e-02,  3.016569903498e-02, {   0,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.369954619914e-03, -7.936170100631e-02, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.349999999322e-02, -1.063865445531e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.042999999051e-02,  1.235316119245e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -7.332999999437e-02,  9.086934577228e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.673726496800e-03, -6.658145357695e-02, {   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.080797860251e-03,  4.884139756625e-02, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -5.219418414782e-02, -1.059504951844e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.070032163160e-02, -1.050750380846e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.331186241076e-03,  4.577620811144e-02, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.730224968473e-02,  1.086612598756e-02, {   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.637999999654e-02, -5.666385949342e-07, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.647597345992e-02,  6.909960910215e-03, {   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.671159565761e-03,  3.373252311471e-02, {   1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  4.214195042382e-02,  3.429780289849e-05, {   2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.977999999280e-02,  7.567782542776e-07, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.725429259715e-03,  2.804164282221e-02, {   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.308975823365e-03,  2.660922207128e-02, {   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.298957909536e-02, -9.768733693728e-03, {   0,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.298841483985e-02, -9.768218227460e-03, {   0,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.043541435396e-02,  9.590113711915e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.094741767080e-02,  4.880308327710e-06, {   2,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.951267669771e-02,  9.747442768524e-03, {   2,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.948252817825e-02,  9.732374349295e-03, {   2,  -1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.592415507055e-03, -2.242641944338e-02, {   2,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -6.587403344470e-03, -2.241403907986e-02, {   2,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.864326765214e-02,  6.930028402820e-06, {   2,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.403999999841e-02, -2.769209204367e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.359575015870e-02, -1.683426138740e-05, {   2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.981531148283e-02,  3.515235694515e-03, {   2,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.246646052043e-02,  1.409754372806e-04, {   2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.884750794471e-02,  3.340767943723e-03, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.195341960828e-02,  1.461009548521e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.680988389785e-03,  1.502519569776e-02, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  2.005472603623e-02,  5.733346966371e-06, {   2,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.999183640024e-02, -2.988119862093e-06, {   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.163228777925e-03, -1.194061283600e-02, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.259276052832e-02,  6.290548996709e-03, {   2,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.220891643077e-02,  6.098902423310e-03, {   2,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.910118106630e-03, -1.638013216514e-02, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.743999999881e-02,  2.039382324324e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.158704937727e-02,  5.545242133423e-03, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -3.222053396530e-03, -1.351074039193e-02, {   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.646999999859e-02,  2.155917958464e-07, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.087739795419e-02,  5.433725623865e-03, {   2,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.105969407322e-03,  1.313138237220e-02, {   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.078250248623e-02,  5.386347082522e-03, {   2,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.700040405196e-03,  1.379567280000e-02, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.872223614292e-03,  4.922270120309e-03, {   1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.407429478761e-02,  1.015126170676e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.404999999889e-02,  1.765575071271e-07, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.230425631519e-02, -1.979258528999e-04, {   2,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.409916867340e-03, -9.923242631620e-03, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.177180944064e-03,  9.000443438205e-03, {   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.093551868686e-03, -9.291658030543e-03, {   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.139177492307e-03, -8.176838982350e-03, {   0,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.138817270534e-03, -8.175782623239e-03, {   0,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.478517750265e-04, -9.855948523934e-03, {   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.004999999841e-02,  1.789137016125e-07, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.916509828160e-03,  8.101295489139e-03, {   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.965477485473e-03, -9.379121110244e-04, {   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.738104373002e-04,  9.068418022189e-03, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.549248592958e-03, -1.882748834997e-03, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.113100451417e-03,  2.774871731234e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.896966542367e-03, -7.380281539030e-03, {   0,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.896820394501e-03, -7.379917695736e-03, {   0,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.880655620883e-03, -5.318463112465e-03, {   2,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.737140009677e-03, -7.338683254925e-03, {   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.349617582440e-03, -7.199611597442e-04, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.723404318354e-03, -7.271407046663e-03, {   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.861875156751e-03, -4.055208876610e-03, {   2,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.352936738966e-04,  7.908724338515e-03, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.579514035300e-03, -4.110202691615e-03, {   2,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.456255207625e-03, -2.067224420980e-04, {   2,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  3.382102166532e-03, -5.255956104958e-03, {   2,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.747998529051e-03,  2.871370408486e-03, {   0,   1,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.032204767087e-03, -6.580135975086e-03, {   0,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.031881417142e-03, -6.578994599796e-03, {   0,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.364506754257e-03, -5.220751856580e-03, {   2,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.339999999360e-03,  1.033479263249e-07, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.300058500771e-03, -1.073190255298e-06, {   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.510301290545e-03,  2.752700481646e-03, {   0,   1,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.973204937579e-03,  2.515734293733e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -8.087778530972e-03, -4.354080011106e-06, {   2,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.982085473177e-03, -1.005949727216e-05, {   2,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.702073904629e-03,  1.186347761176e-03, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.636688697210e-03,  2.324704494453e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.831970804599e-03, -2.385103732543e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.505522440664e-03, -6.315175863225e-03, {   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.752218450852e-03, -5.772463346338e-03, {   2,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.750495520202e-03, -5.767386449998e-03, {   2,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  7.183422070290e-03, -1.422479088490e-06, {   0,   1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.014382546674e-03,  6.070540137058e-03, {   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  7.080870224395e-03, -7.746758297044e-07, {   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.892404343680e-03,  3.783605494283e-05, {   2,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.752619926152e-05, -6.841183550987e-03, {   2,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.301494158708e-03,  5.491866595127e-03, {   1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.719999999554e-03,  7.740884298274e-08, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.556285865192e-03,  1.424367949966e-04, {   0,   1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.547530896289e-03, -2.164879964435e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.418661414279e-03,  4.142627784805e-05, {   2,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.334749335660e-06, -6.332279025361e-03, {   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.231526580188e-03,  8.871729673036e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.195639580187e-03, -1.628024170237e-06, {   0,   1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.426548090163e-03, -4.701824058317e-03, {   2,  -1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.425067561845e-03, -4.697343249285e-03, {   2,  -1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.693297515451e-03,  4.429035308751e-03, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.913174759794e-03,  1.332701665157e-04, {   0,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.044609047840e-03,  2.514302160200e-08, {   2,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.135631026971e-03, -4.855258116531e-03, {   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.966342281113e-03, -1.392259987343e-06, {   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.022641068527e-03,  8.880560440589e-04, {   2,   1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.889999999548e-03, -7.298792398708e-08, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.840574297723e-03,  3.142283796837e-05, {   2,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.418761693325e-03,  3.437230212723e-03, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.836046410839e-03, -1.248250290671e-05, {   2,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.737014610015e-03,  1.601509065852e-06, {   2,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.144499829979e-03, -3.553362919966e-03, {   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.196143511239e-03,  4.490948121731e-03, {   0,   1,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.115434388886e-03,  4.571062195964e-03, {   1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.195138709323e-03,  4.487138674485e-03, {   0,   1,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  5.552459237196e-03,  1.193101744638e-04, {   0,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.629999999615e-03, -6.583556471433e-08, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.619999999260e-03,  9.122137868073e-08, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.610851270440e-03, -1.670034504818e-06, {   0,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.609999999639e-03,  6.364343117261e-08, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.265574268335e-03, -4.300997843314e-03, {   2,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.264044189625e-03, -4.296685747769e-03, {   2,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.305999201883e-03,  1.035032539973e-04, {   0,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.557392330556e-03,  7.874029215770e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -5.318588405548e-03,  1.203782475463e-05, {   0,   1,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.577294211462e-04, -4.688777640762e-03, {   1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.176824814193e-03, -2.689169849657e-06, {   2,  -1,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -5.074309519128e-03,  9.765332526793e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.291826729436e-03,  8.040300074499e-04, {   2,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.004758312978e-03,  9.712187329000e-08, {   2,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.515355273213e-04, -4.784828290224e-03, {   2,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.513128042344e-04, -4.777131683587e-03, {   2,  -1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  4.875310344674e-03,  1.353617844250e-06, {   2,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.145586716877e-04,  3.859907311354e-03, {   1,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.019353104764e-03, -7.207270539823e-04, {   0,   1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  4.719499333313e-03,  1.355312200917e-06, {   2,  -1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.979924307527e-03, -2.718886893832e-03, {   2,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.497474482055e-03,  1.021358440284e-04, {   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.538340550391e-03,  1.009000424690e-05, {   0,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.419999999443e-03, -7.020063317112e-08, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  4.369999999752e-03,  4.652524186953e-08, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.274312771533e-04,  3.465807674103e-03, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.429661589262e-03, -8.210328032596e-04, {   2,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.239999999511e-03,  6.438170544509e-08, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.078259904561e-03, -2.152520435991e-03, {   0,   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.193704429753e-03, -2.941012311375e-06, {   2,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.052512107446e-03, -2.139481327936e-03, {   0,   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.394096961278e-03,  7.708963078905e-04, {   2,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.009755870242e-04, -3.351601544556e-03, {   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.605641701959e-04, -3.978032194385e-03, {   2,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.048579011975e-03,  6.026757987722e-05, {   2,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.459653154326e-03, -6.198802110916e-04, {   0,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  6.370955248822e-04,  3.436569581647e-03, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.899585350092e-03, -6.660343527116e-05, {   2,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -7.445350193045e-04, -3.187657940474e-03, {   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.201164372218e-03,  7.278936379282e-04, {   2,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.809859611895e-03, -5.723927272355e-05, {   2,   1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.994383007730e-03, -8.136927503707e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.993324482684e-03, -8.135485504254e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.820780772760e-03, -1.961487967785e-03, {   0,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.814230523490e-03, -1.964699613653e-03, {   0,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.769999999227e-03,  7.632673884300e-08, {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.968211811884e-03,  7.688620245803e-04, {   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.719999999698e-03, -4.739616116616e-08, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.709372207450e-03, -6.656511574582e-07, {   2,   1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.838699853486e-03, -8.458944151570e-04, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.837726072129e-03, -8.457832777343e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.833421096494e-03, -8.083649920374e-04, {   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.552048909163e-03, -1.084574118877e-03, {   0,   1,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.541865262609e-03, -1.080065622028e-03, {   0,   1,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.113201597144e-03, -1.503556176660e-03, {   0,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  7.670599042340e-05, -3.472950854551e-03, {   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.076653258785e-03, -1.468292292052e-03, {   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.283992919632e-03,  1.254814749394e-03, {   1,   1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.283992919632e-03,  1.254814749394e-03, {   1,  -1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.875367916594e-04,  3.187068958345e-03, {   1,  -2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.531590160664e-04,  2.757332946703e-03, {   1,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.749028522415e-05, -3.275819846814e-03, {   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.609014558861e-04,  2.679670081828e-03, {   1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.259708059978e-03, -4.870920112847e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.289999999747e-03,  4.076914599629e-08, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.764311633178e-03, -5.195453415782e-04, {   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.129586061210e-04,  2.759088435705e-03, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.507525150634e-03,  7.303696931365e-04, {   1,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -3.219999999753e-03,  3.990171735807e-08, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.629892967360e-04,  2.928213897550e-03, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.360021822607e-04,  2.243133448099e-03, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -3.149999999765e-03, -3.848451000523e-08, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.092161641727e-03,  1.044935009790e-03, {   2,  -1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.892166833438e-04, -2.631132500449e-03, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.008180697544e-03, -3.724401567023e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.025624401575e-03,  1.011681426023e-03, {   2,  -1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.601132103043e-05, -3.003932431660e-03, {   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.691386154904e-04, -2.428218940532e-03, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.956372556041e-03, -3.668476774606e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.141232047654e-03, -1.834989157299e-03, {   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.140055561013e-03, -1.833315494002e-03, {   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.159896766849e-03, -1.807165993997e-03, {   2,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.493572154973e-03, -4.479981803692e-04, {   2,  -1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.929999999670e-03, -4.397880648953e-08, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  5.570206724401e-05, -2.871984454957e-03, {   0,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.473652736747e-03,  4.390260820958e-04, {   2,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.421644050089e-03,  4.652490969600e-04, {   1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.878226928067e-03,  1.403880078281e-06, {   2,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.744674624247e-03, -1.079003880611e-03, {   2,  -1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.880232107051e-03,  9.390855993060e-04, {   2,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.702065957762e-05, -2.720034911164e-03, {   2,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.080974666474e-04, -2.184003412304e-03, {   2,  -1,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  6.080699197028e-04, -2.183930639038e-03, {   2,  -1,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.501446890124e-05, -2.773680037728e-03, {   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.841173978976e-03,  9.195729763131e-04, {   2,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.759999999806e-03, -3.275633940011e-08, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.164532524481e-03, -1.595243430599e-03, {   2,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.793061833915e-05, -2.675542146238e-03, {   0,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.729999999753e-03,  3.668856620507e-08, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.151555146168e-03, -5.380476644213e-04, {   1,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.245362159445e-04,  2.449791543668e-03, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.111239332583e-04, -2.155946523092e-03, {   1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.639997127351e-03,  3.050786605329e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.711232579122e-03,  9.201004694654e-04, {   2,  -1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -9.734283783930e-04, -1.635588848808e-03, {   2,  -1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.553857093569e-03, -5.078749585824e-05, {   2,   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.095274962640e-03, -1.500814629925e-03, {   0,   1,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.093907505373e-03, -1.498554756103e-03, {   0,   1,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.579997913355e-03,  2.443119425363e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.569999999786e-03,  3.319267171351e-08, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.660784349528e-03,  8.925693567367e-04, {   2,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.209613895271e-04,  2.028074957813e-03, {   1,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.531199172803e-03,  1.783130139951e-05, {   0,   1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.526062234207e-03,  8.591320402616e-06, {   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.519999999871e-03, -2.550973234731e-08, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.113748871983e-03,  3.996362043257e-04, {   1,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.093870013512e-05,  2.417536154345e-03, {   2,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.502876734675e-03, -1.401675971864e-06, {   2,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.499523888172e-03,  1.685826595789e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -6.309332674280e-05, -2.421040278792e-03, {   0,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.169538605529e-04, -2.256587479304e-03, {   2,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.030638591923e-04,  2.260899172694e-03, {   1,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.699758366500e-04,  1.984920852838e-03, {   1,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.439459500305e-03,  6.134583223655e-06, {   0,   1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.288994244353e-03, -1.514639120817e-04, {   2,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.433264306842e-03,  2.233658512750e-07, {   2,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.620569454217e-05, -2.341791947819e-03, {   0,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.318541710099e-04, -1.672800767331e-03, {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  7.547095966477e-04, -1.640388345316e-03, {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.389999992129e-03,  1.939671662075e-07, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.381921591728e-03,  6.679979303852e-06, {   0,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.892585727879e-04, -1.481985103401e-03, {   0,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.997706520998e-04, -1.361974669261e-03, {   2,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -7.560144540440e-04, -1.604903803800e-03, {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  9.057413879129e-04, -1.448091866163e-03, {   2,   1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.968029460838e-05,  2.253678603779e-03, {   2,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.198968283520e-04, -1.633151174280e-03, {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.561754716297e-03, -7.823545454223e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -8.994751053943e-04, -1.437354032653e-03, {   2,   1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.816133925109e-03,  5.162423467847e-04, {   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.159462726379e-03, -1.703997177596e-04, {   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.304815168506e-03,  2.267846828789e-05, {   2,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.860728731112e-03, -4.653666794067e-04, {   1,   2,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.317399945256e-03, -5.880889081839e-06, {   0,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.733350718795e-04, -1.335809594177e-03, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.239535211309e-03, -5.414029671225e-05, {   2,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.283216061855e-03,  1.510543807012e-06, {   2,   1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.570961038135e-03,  7.015096300169e-04, {   1,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -5.951741667009e-05, -2.212700528436e-03, {   2,  -1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.269999999806e-03, -2.971423051478e-08, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.602301055346e-04, -1.286314362630e-03, {   0,   1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.596077556663e-04, -1.285231170764e-03, {   0,   1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.875562123218e-03, -3.593361565488e-04, {   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.226641568329e-03,  1.665977820873e-06, {   2,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.425379171433e-03, -7.903009515783e-04, {   2,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.039577497212e-03, -1.173583107730e-03, {   2,   1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.038083251205e-03, -1.171899098133e-03, {   2,  -1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.236578130710e-03, -9.643968096475e-04, {   2,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.397959846272e-03,  7.862429028451e-04, {   2,  -1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -2.138990254967e-03, -4.146055468672e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.179999999833e-03,  2.701420616168e-08, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.827243250983e-03, -3.469386701711e-04, {   1,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.158021617747e-03,  1.481590793791e-05, {   0,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.901463326236e-04, -1.464407269026e-03, {   2,  -1,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -5.192121529618e-05, -2.092081459499e-03, {   2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.213074045833e-05,  2.049662241350e-03, {   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.119999999842e-03,  2.590068609895e-08, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.090593381266e-03, -2.039426255736e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.109999999829e-03, -2.688330646790e-08, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.195102624368e-04,  1.679326149192e-03, {   1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.187558340218e-04, -1.663915081098e-03, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.634960545346e-04,  1.913641585419e-03, {   1,  -2,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.008208626682e-04, -1.264358998553e-03, {   2,   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.059999999717e-03,  3.415609345996e-08, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.206634368430e-04, -1.633891976294e-03, {   1,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.047305602239e-03,  1.018703471054e-06, {   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.812440467636e-04, -1.666909207644e-03, {   2,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.254618995415e-05, -2.028904970114e-03, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.180941054622e-05, -2.017526341845e-03, {   2,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.914071653874e-03, -1.243088824244e-04, {   2,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.933944556865e-03,  1.031932418183e-04, {   2,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.911487726762e-05, -1.944515694534e-03, {   2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.006322503916e-03, -1.217339173930e-05, {   0,   1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.036726604983e-04, -1.613090042986e-03, {   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.526234205508e-04, -1.262601755287e-03, {   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.558808971121e-03, -4.542586050797e-04, {   1,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  7.568548673909e-04, -1.247362683629e-03, {   0,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.106157146803e-03, -8.958451514527e-04, {   2,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.329199223487e-03, -6.633601991845e-04, {   0,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.978372324918e-03, -4.814342871379e-07, {   2,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.599603146113e-04, -1.199995903823e-03, {   2,  -1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.778662282054e-04,  1.573930263892e-03, {   1,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.597376571206e-05, -1.902860633844e-03, {   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.850359524316e-04,  1.560732250325e-03, {   2,  -1,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.928870769448e-03, -1.678576107005e-05, {   0,   1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.706958159386e-03, -2.143400658964e-04, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.706306748028e-03, -2.145480426995e-04, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.361653576560e-04, -1.176511826832e-03, {   2,  -1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.242286391198e-04, -1.182489009852e-03, {   0,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -7.307174853611e-04, -1.167469653580e-03, {   2,  -1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.195623533787e-04, -1.457595540685e-03, {   2,  -1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.192643463951e-04, -1.456893181572e-03, {   2,  -1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.874603599698e-03,  9.501805634260e-08, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.869999999848e-03,  2.382548961833e-08, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.806186852223e-03,  5.921350160038e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.848325708660e-03, -8.404044562900e-06, {   0,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.637366809993e-05, -1.814241940826e-03, {   0,   1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.540941459987e-04,  1.493293909238e-03, {   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.516939871323e-04, -1.487848150607e-03, {   1,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.913100344738e-04, -1.124952999964e-03, {   0,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.756122509860e-05, -1.782598735516e-03, {   2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.071881287636e-05, -1.787497441797e-03, {   1,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.989455129943e-04, -7.841725796886e-04, {   2,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  9.659058868175e-04, -8.026484634264e-04, {   2,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.381533646241e-04,  1.429120199131e-03, {   1,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.758114605924e-03, -2.189830944070e-07, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.621952346937e-03, -1.349236979222e-04, {   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.732001961417e-03,  2.203266906541e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  3.030646245283e-05, -1.718537208304e-03, {   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.845633114279e-04,  1.552832891899e-03, {   1,   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.448192550087e-03, -2.727779124416e-04, {   2,  -1,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.714282344536e-03, -1.850772597921e-06, {   2,   1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.584846935217e-03, -1.254416702045e-04, {   2,  -1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.370299039870e-03, -3.327052562959e-04, {   2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.829973799090e-05, -1.682717427209e-03, {   2,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.495803349957e-05, -1.634671948600e-03, {   2,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.669100084511e-03,  1.677633977284e-05, {   2,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.758039104149e-04,  1.393729018089e-03, {   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  6.513854920281e-04, -1.013084847917e-03, {   2,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.457686294679e-04, -1.002954635326e-03, {   2,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.077745839856e-04,  1.339263364053e-03, {   2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.339815292245e-03,  3.062958481718e-04, {   2,  -1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.623594022419e-03, -1.279993501792e-05, {   0,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.640686162416e-04, -7.700174638183e-04, {   2,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.615414379016e-03,  1.269268619703e-05, {   2,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.257060395398e-03, -3.633936580632e-04, {   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -3.102673486845e-04, -1.295335587254e-03, {   1,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.273797361537e-04,  9.773338979682e-04, {   2,  -1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -9.803813011370e-05,  1.506137407168e-03, {   2,  -1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.022709621392e-03, -5.800255824081e-04, {   2,  -1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.041456563940e-03,  5.605258390377e-04, {   2,  -1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.592656132833e-03, -6.120424154308e-07, {   2,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.787534966304e-04, -7.139397109784e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.579989812766e-03,  8.088356834562e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.581888395206e-03,  2.618959624163e-06, {   0,   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.699174353130e-05, -1.531275660857e-03, {   2,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.549999999730e-03, -2.894628564260e-08, {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.877199619816e-05, -1.513754979940e-03, {   0,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.781158994176e-04, -9.600522812796e-04, {   0,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  9.952325422454e-04, -5.319646956169e-04, {   2,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.513702948231e-03, -1.055041739066e-05, {   0,   1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.012718616106e-03,  5.058026508394e-04, {   0,   1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.013182858090e-03,  5.016507194360e-04, {   2,  -1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.505023216776e-03,  1.798498495631e-06, {   0,   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.007640648498e-03,  4.989207166302e-04, {   2,   1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.407418250922e-03,  7.035283553921e-05, {   0,   1,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.407418250922e-03,  7.035283553921e-05, {   0,   1,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -6.592184301063e-04,  7.991635241699e-04, {   2,  -1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.386899689160e-04,  1.311548616784e-03, {   2,  -1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.449999999915e-03, -1.569050997507e-08, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.444006242132e-03,  5.345484304830e-06, {   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.384376343880e-04,  1.309673468788e-03, {   2,   1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.661933810685e-05, -1.410658723297e-03, {   0,   1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.432069842999e-03,  1.917010423577e-06, {   0,   1,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.413733809878e-03,  1.893660220694e-06, {   0,   1,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.402659319662e-03, -7.711135923249e-06, {   0,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.405177580507e-03, -2.828109697935e-06, {   4,  -1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.400000451913e-03,  9.212053702925e-08, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.523913255054e-04,  1.236257970742e-03, {   1,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.084544782564e-04,  1.173367770581e-03, {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.646069476017e-04, -1.116046501884e-03, {   1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.379999999736e-03,  2.697580891711e-08, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.116592069534e-03, -2.541916098028e-04, {   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.753481854986e-04,  1.086469649843e-03, {   1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.472475928340e-05, -1.332868240450e-03, {   0,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.849838173573e-05, -1.335412661162e-03, {   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.116861483438e-05,  1.318420516348e-03, {   2,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.957890274622e-04,  4.474571313857e-04, {   0,   1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.138869361310e-03, -2.030065439874e-04, {   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.319999999901e-03,  1.612684228803e-08, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.311677996939e-03, -8.392963782792e-07, {   2,  -1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.718087042656e-04, -9.401011729055e-04, {   1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.093421794755e-03,  2.121909181980e-04, {   1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.055933226435e-03,  2.418315634856e-04, {   2,   1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.283645756754e-03,  3.061666039502e-07, {   2,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.350625984671e-04, -1.147808759427e-03, {   3,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.448354745305e-04, -1.037123217050e-03, {   1,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.450945742776e-04,  1.036803014032e-03, {   1,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.276931682595e-03, -4.892578350328e-06, {   2,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.246895518204e-03,  9.581728825727e-06, {   2,  -1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.443410690911e-04, -1.009781594268e-03, {   2,  -1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.286515970059e-04,  4.117606156887e-04, {   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.554245328492e-04, -9.803184694227e-04, {   0,   1,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.553807426349e-04, -9.802185829724e-04, {   0,   1,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.245365984558e-05, -1.208369010801e-03, {   2,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.871753073957e-04, -8.402978863775e-04, {   2,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.457250442807e-04,  6.789379366304e-04, {   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.035059803414e-03, -1.838688095288e-04, {   0,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.737334076599e-04,  8.407199551940e-04, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  7.415422030584e-04,  4.687423601612e-04, {   2,  -1,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  9.841210192877e-04, -2.258497574015e-04, {   2,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.384715465909e-04,  4.666186753958e-04, {   2,  -1,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -2.428614636978e-04,  9.586233654949e-04, {   2,  -1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.545136535936e-05,  1.172456778665e-03, {   2,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.407146563083e-04, -6.532013918746e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  3.657421644495e-04, -8.227443613848e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.357929105896e-04, -1.046297104491e-03, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  1.170074739461e-03, -8.455800596611e-07, {   2,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.132032073214e-03,  3.711006557648e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.227381300212e-04,  6.341253399479e-04, {   2,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.220550209362e-04,  9.247863768055e-04, {   1,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.139999999912e-03, -1.412669496524e-08, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.311330463889e-04, -7.050818364742e-04, {   2,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.132767373202e-03, -5.758042885977e-07, {   2,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.952741417930e-04, -3.379113099386e-04, {   2,  -1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.952741417930e-04, -3.379113099386e-04, {   2,  -1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.228795130292e-04, -9.096219432619e-04, {   1,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.926002165872e-04, -7.281519311579e-04, {   0,   1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -5.724343433549e-05, -1.050045022348e-03, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -4.149128688454e-04, -6.899513742362e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -1.099999999901e-03,  1.478293876395e-08, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.319068753740e-07, -1.097901203169e-03, {   2,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.796897692385e-04, -7.112650155642e-04, {   0,   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -4.283955897696e-04, -6.614579915012e-04, {   0,   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.721738954320e-04,  1.165946364592e-04, {   2,   1,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  4.621282833862e-04,  6.228759192252e-04, {   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.078015958977e-03, -4.246343877120e-07, {   0,   1,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -8.879479101121e-04,  1.818494239393e-04, {   2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.072581404881e-04, -6.552439945811e-04, {   2,  -1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  8.978216275225e-04, -1.613040563581e-04, {   2,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.058096119721e-03,  9.766960097160e-07, {   2,  -1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.316217899356e-04,  1.252231628621e-04, {   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.047508532941e-03,  8.264080437734e-06, {   2,  -1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  4.151613211677e-04, -6.374403894126e-04, {   0,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  3.940413345339e-04,  6.568930264549e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  3.114991080912e-04,  7.376439465983e-04, {   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.482080904897e-04, -2.003398395902e-04, {   1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.164971674927e-05, -1.004336304328e-03, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.920149577498e-04, -6.535216195366e-04, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.041145344330e-03, -2.170927180006e-06, {   4,   1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -8.587605863586e-05, -9.561512968946e-04, {   1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.041866633308e-03, -1.606248801958e-07, {   2,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.039999999829e-03,  1.887748118853e-08, {   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.388137869329e-05, -1.004724849772e-03, {   2,   1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.613756499680e-04, -8.686817920220e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.196544809737e-04, -2.062662083823e-04, {   2,   1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.008442000256e-03,  1.140177942268e-05, {   2,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.945126563150e-04,  7.064145641893e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.947223424011e-04, -8.047269343964e-04, {   2,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.218452158615e-04, -8.742969833491e-04, {   1,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  6.729096258567e-04, -3.219539646581e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  8.068077186770e-04, -1.873456737996e-04, {   2,  -1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.702590138394e-04, -1.437668937431e-05, {   2,  -1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -3.391515226503e-04, -6.441692793904e-04, {   0,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.512030022697e-04, -3.009260963052e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -8.299585166712e-04, -1.481511943696e-04, {   1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.762747968315e-04, -5.975487139201e-07, {   2,   1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.338104736281e-04, -6.396428758753e-04, {   0,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -7.839201456178e-04,  1.885735197483e-04, {   2,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  8.263450280136e-04, -1.452219103838e-04, {   2,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.699999999191e-04, -1.252797337059e-08, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.894898068346e-04, -5.803297487484e-04, {   1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.845305719521e-04, -7.808494595197e-04, {   1,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.826773318082e-05, -9.112437393114e-04, {   2,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.719174296633e-04, -7.772252714485e-04, {   1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.029478152278e-04, -6.440304632644e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -7.546542602060e-04, -1.880140537040e-04, {   1,   0,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -9.404760140053e-04, -3.333739370041e-07, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  3.367354101019e-05,  9.010779847545e-04, {   1,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.776852300580e-04,  7.524617736542e-04, {   1,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.894425695925e-04,  1.398345561900e-04, {   2,  -1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.961026826046e-04, -1.296757429415e-04, {   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.211594077870e-04,  4.434389744404e-06, {   2,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.937887994543e-04, -2.982636063134e-05, {   2,   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  7.234681077763e-04,  1.994945577757e-04, {   2,  -1,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.868213379597e-04,  7.256653410578e-04, {   1,  -2,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.985148137059e-04, -7.129672409063e-04, {   2,   1,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.985085066893e-04, -7.129611681633e-04, {   2,   1,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.865535333691e-04,  7.238992925493e-04, {   1,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.099999999432e-04,  1.016479756340e-08, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.381528645272e-04,  6.707403153374e-04, {   2,  -1,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.381200761264e-04,  6.706314521620e-04, {   2,  -1,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  6.053243586857e-04,  3.022907581991e-04, {   0,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.325629747905e-04,  2.722505944967e-04, {   2,  -1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  8.999979790219e-04,  1.809334206800e-06, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.625868956950e-04, -1.381405307860e-04, {   2,   1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.716166054241e-04, -7.278388098254e-04, {   1,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.171340469769e-04, -2.819651790155e-04, {   1,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  7.758711953230e-04, -1.222180414036e-04, {   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.527639120301e-04, -2.435801970202e-04, {   0,   1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.718809063608e-04,  7.231766175796e-04, {   1,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.308202548065e-04,  5.571261877018e-05, {   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.170733466835e-04,  2.680551603517e-04, {   2,  -1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -4.163440228701e-04,  4.638899447143e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -5.943109408910e-04,  2.844416864351e-04, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -8.717018276002e-04,  6.673806304937e-06, {   2,   1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.434945934145e-04,  1.312802029255e-04, {   4,  -1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.229878856037e-04, -5.499738170166e-04, {   2,  -1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.959325594082e-04, -4.750104326487e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  5.800728832080e-04, -2.897188791146e-04, {   0,   3,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.788963829337e-04, -2.891316193543e-04, {   0,   3,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.935976143606e-04, -6.717621280925e-04, {   2,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.933875065647e-04, -6.711951383950e-04, {   2,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.222186153506e-04, -2.393963983298e-04, {   0,   1,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.599999999263e-04,  1.125737367504e-08, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.282169853957e-04, -5.303200258903e-04, {   2,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.872974067078e-04,  4.701904453129e-04, {   2,   1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  7.663861394325e-05, -7.805789194241e-04, {   2,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  6.882472362360e-04, -1.635990408375e-04, {   2,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.283602187085e-04, -3.216610077907e-04, {   2,   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.247791183100e-04,  1.249460094677e-04, {   4,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -8.463886956511e-04, -1.680319606131e-06, {   4,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  8.475046105920e-04, -2.002689398229e-07, {   0,   1,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.399999998964e-04,  1.319468914453e-08, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -6.707450727424e-04, -1.688508538019e-04, {   2,  -1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.382099972766e-04, -3.244785147875e-07, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  3.525728825302e-04, -4.671728226381e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.087913516323e-04, -5.101729191964e-04, {   2,  -1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  5.453886614767e-04, -2.719388404227e-04, {   1,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.169269014237e-04, -5.978803801692e-04, {   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.708688993606e-05,  7.469935402657e-04, {   1,   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.391336722872e-04,  2.687965065631e-04, {   1,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.971054861594e-04,  1.098996755844e-04, {   2,  -1,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -6.320674248323e-04, -1.606156813109e-04, {   2,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.731080308840e-04,  1.777364974060e-05, {   0,   1,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.899999997400e-04,  2.026850860119e-08, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.456228148110e-04,  3.551825395939e-05, {   2,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.787602251948e-04,  1.648218865258e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  5.056210552802e-04, -2.737303922184e-04, {   2,  -1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -5.532525743320e-04, -2.259503764680e-04, {   0,   1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.089865919476e-04, -4.689051164167e-04, {   2,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  7.669519032389e-04,  9.331923486021e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  7.726717295520e-04,  5.106062102413e-07, {   0,   1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.068936719177e-05, -7.419659987896e-04, {   4,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.690668282914e-04, -5.517866915843e-07, {   2,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.472658621540e-04,  6.217675337312e-04, {   1,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.504580454580e-04, -1.180080210341e-04, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.663449534789e-04, -5.974446486071e-04, {   2,  -1,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.663409753963e-04, -5.974134426818e-04, {   2,  -1,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  5.381485845381e-04, -2.252410239325e-04, {   0,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.478476500775e-04, -6.110240572332e-04, {   2,   1,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.454026080993e-04,  1.132594493763e-04, {   4,   1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  7.528141973871e-04,  5.385524105788e-06, {   0,   1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.411481001265e-04, -6.169051550734e-04, {   2,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.112697430436e-04,  6.446520300397e-04, {   1,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.112697430436e-04,  6.446520300397e-04, {   1,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.358616986513e-04, -1.154211517572e-04, {   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.283798014660e-04,  1.228082940227e-04, {   1,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.425547701395e-04, -6.065037609165e-04, {   2,  -1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  5.502996603218e-04,  1.974633011421e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.381170591707e-04,  5.060560634363e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.961187957626e-04,  2.477874620670e-04, {   0,   1,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.848551714839e-04,  4.570377058897e-04, {   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.381701326617e-04, -1.950844365658e-07, {   0,   1,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.236021821434e-04, -1.129639061336e-04, {   1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.548921756690e-04, -2.813150875775e-04, {   2,   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -5.985740376179e-04, -1.371611393450e-04, {   0,   1,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.692420632263e-04, -6.585633609658e-05, {   1,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  7.343702288653e-04,  1.994035077449e-07, {   2,   1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.587821075550e-04,  3.751469651083e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -5.675723519092e-04,  1.645735066982e-04, {   1,  -2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.853878703399e-04, -2.424267041215e-04, {   2,  -1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.993514786732e-05, -6.673138527336e-04, {   3,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.203982276979e-04,  3.036742268522e-06, {   2,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  6.308353730372e-04, -8.731260611162e-05, {   2,  -1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.414774698800e-04, -5.718709571986e-04, {   1,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.240609827059e-04,  1.875156049831e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  5.808324911608e-05, -6.530602154395e-04, {   2,   1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.421576612587e-04, -2.687583208091e-04, {   2,  -1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.736789667001e-04,  5.367799138369e-04, {   2,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.099999999124e-04, -1.115265391980e-08, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  4.729862248714e-04,  2.362406764972e-04, {   0,   1,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.976654856804e-04, -2.114514304668e-04, {   2,   1,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.956563844949e-04, -2.105619626011e-04, {   2,   1,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.100975382482e-04,  5.936828550163e-04, {   1,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.511121084498e-04, -2.475290359384e-04, {   1,   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -6.963311279370e-04,  6.746226316829e-08, {   2,  -1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.067237965375e-04, -3.786302059569e-04, {   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -6.650350679589e-04,  1.888985269378e-05, {   2,  -1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -6.821741228583e-04,  5.186559271649e-07, {   0,   1,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.288416202972e-04,  5.463243510456e-04, {   1,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.314260672722e-04, -5.422250228920e-04, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.859719856845e-04,  1.873611759073e-04, {   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  2.607612081773e-04, -4.113089669735e-04, {   2,  -1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -6.699999994171e-04, -2.794795730398e-08, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.592998870917e-04, -4.023475592239e-04, {   2,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -5.157300566207e-04, -1.456134658145e-04, {   2,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.435837179789e-04,  1.736964561359e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.441823033393e-04,  1.143552464555e-04, {   1,   1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  5.441823033393e-04,  1.143552464555e-04, {   1,  -1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  6.396823525837e-04,  1.667782139126e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.234112661256e-04, -2.322870703552e-04, {   1,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.943617921381e-04, -3.584621662016e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.406487463245e-04,  1.175724421719e-05, {   0,   1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.249333747955e-04,  5.230285506160e-04, {   1,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.828477892691e-04,  4.621391208164e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -4.144037133839e-04, -2.298771297733e-04, {   2,   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -6.164369577872e-04,  2.659992867352e-05, {   4,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -6.403427539680e-04,  5.859559753610e-07, {   2,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.212882136804e-04, -1.193163754754e-04, {   0,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.356397926171e-04,  5.653039842011e-07, {   2,  -1,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -2.873292868085e-04,  3.486315619996e-04, {   0,   1,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.521061966372e-04, -3.836253820704e-04, {   2,  -1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.236193336964e-04, -2.115991950929e-04, {   2,  -1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.297888842869e-05, -6.211381716784e-04, {   2,  -1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.328540146342e-04,  5.006631876765e-04, {   0,   1,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -4.222103982157e-04,  2.108708695791e-04, {   4,  -1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -6.036250612365e-04,  2.934717418827e-05, {   2,   1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.462948927580e-04, -3.861271369091e-04, {   2,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.896184496261e-05,  5.325205683718e-04, {   3,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -6.300539221156e-04,  2.083156857609e-07, {   4,   1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.235699906040e-04,  5.066548739250e-04, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.188335126923e-05, -5.776747509292e-04, {   3,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.418186791030e-04, -1.877325319870e-04, {   2,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.413343090274e-04, -1.875180920074e-04, {   2,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.179538528690e-04,  5.108496271205e-04, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -6.278185299192e-04, -3.909092632391e-07, {   2,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.221882428347e-04,  5.054404349802e-04, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.193967108423e-04, -5.074092261658e-04, {   1,   0,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.922170308255e-04, -2.300208096941e-04, {   0,   1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.284595481221e-04,  4.893628094746e-04, {   1,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.191948111479e-04,  9.832686766669e-05, {   1,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.231234914712e-04,  9.208287699007e-05, {   0,   1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.088265836572e-04,  2.041901805707e-04, {   4,  -1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.099999999619e-04,  6.813765399644e-09, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.082264657288e-04, -3.171352487552e-07, {   2,   1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.134310692757e-04, -4.942143683631e-04, {   2,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.240801489384e-04, -2.815748500351e-04, {   2,  -1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  4.901226233327e-05, -5.559671249698e-04, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.240244659945e-04, -4.802536732017e-04, {   0,   1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.239912810048e-04, -4.801119150173e-04, {   0,   1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.987844187493e-04,  4.847462037466e-06, {   2,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.161667279857e-04,  4.873184005972e-04, {   1,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.178325481310e-04,  4.845518196516e-04, {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.101762390037e-04, -2.917729646384e-04, {   2,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.352555286874e-04, -3.664217573489e-04, {   2,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.894022677047e-04,  1.114356491642e-04, {   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.465773179877e-04, -5.146742323135e-05, {   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.700096825861e-04, -3.278120656605e-04, {   2,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.812334844075e-04,  2.145057006030e-04, {   2,   1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  4.162014863257e-04,  1.772733251405e-04, {   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.763105218720e-04, -2.170524330578e-04, {   0,   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  4.938644522522e-04,  9.947339629809e-05, {   2,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  4.252659733175e-04, -1.673205649049e-04, {   2,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.241201918361e-04,  4.675511647420e-04, {   0,   1,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.817718384331e-04,  4.087035186267e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  2.214103707672e-05,  5.669499405969e-04, {   0,   1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.791245261980e-04, -8.552825486117e-06, {   0,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.023381996051e-04, -8.442212889212e-05, {   1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.475908088995e-04,  4.386149703817e-04, {   2,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.830547072254e-05,  5.378350876414e-04, {   1,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.276359471513e-04, -3.576388799367e-04, {   2,  -1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.820620386117e-04,  2.955655343068e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.262548032465e-04,  5.068314071208e-05, {   1,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.534556306713e-05,  5.394203133673e-04, {   2,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.840135739090e-04, -9.055192201247e-05, {   1,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.703709681245e-04, -4.089193777837e-07, {   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.693626230407e-04, -1.181439856075e-06, {   2,   1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.115655759642e-04, -5.780698920380e-05, {   1,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.610892542738e-04, -7.758514615593e-06, {   2,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.619097720063e-04,  2.063103771453e-04, {   2,   1,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.551365153869e-04,  3.115377925892e-04, {   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.772074681390e-04,  1.883955892365e-04, {   4,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.390675517902e-04,  2.492001430846e-05, {   2,  -1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.270684222984e-05, -5.405288810494e-04, {   2,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.096834911917e-04, -4.532816385086e-04, {   0,   1,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.350750385378e-04, -3.221965323759e-04, {   2,  -1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -4.453060055879e-04,  1.096903311467e-04, {   2,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.739589565372e-04, -8.003633606427e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.694565452839e-04,  1.845265757003e-04, {   4,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.137055873413e-04,  1.402078382253e-04, {   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.266165730306e-04, -4.246457317876e-04, {   4,  -1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.265383322104e-04, -4.244618087240e-04, {   4,  -1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.776216289280e-04, -7.288794082068e-05, {   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.479881971643e-04,  2.480491720359e-06, {   2,  -1,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.220740013694e-04, -4.282848556163e-04, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  5.499999999541e-04,  7.103490055419e-09, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.028801582723e-04, -1.451983303310e-04, {   2,  -1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.046231291738e-04,  4.413001190944e-04, {   1,   0,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.472746153925e-05,  4.979954359071e-04, {   1,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.646930348804e-05, -5.255601960031e-04, {   2,  -1,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.146393029311e-04, -3.268245301826e-04, {   2,  -1,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.035662711193e-04, -4.367470418501e-04, {   1,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.004295375183e-04,  4.398688408954e-04, {   2,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.162796389442e-04,  4.229470384586e-04, {   0,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.389642536353e-04, -1.811617594437e-07, {   1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.504870678474e-05, -4.434967325192e-04, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -5.359087481725e-04, -2.491057750796e-06, {   1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.078992111594e-04,  2.304871802565e-04, {   1,   2,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -2.135468137485e-04, -3.248177958081e-04, {   2,  -1,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.588994971737e-04,  2.786476604665e-04, {   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -5.323710328111e-04, -2.760844896562e-06, {   1,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.067471566545e-04, -3.282636736358e-04, {   2,  -1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -3.376987256877e-04, -1.955665118175e-04, {   0,   1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.021664297826e-04, -4.307362380194e-04, {   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.423430437762e-04, -1.901027707526e-04, {   2,  -1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -3.108509643769e-04,  2.198198835472e-04, {   1,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -5.299999999680e-04, -5.827654372301e-09, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.082923705227e-05,  4.887966218463e-04, {   3,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.228229635208e-04,  6.233018000573e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -5.121916428348e-04,  1.559818980430e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -4.438108343818e-04, -8.271324482036e-05, {   0,   1,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.322877860778e-04, -1.939479553273e-04, {   0,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.813949704709e-05, -4.274843900368e-04, {   2,  -1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.982011454916e-05, -4.256117002092e-04, {   1,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.409648048942e-04, -1.844287065444e-04, {   2,   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.095644863384e-04,  1.517874034251e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -4.280486454992e-04,  9.668200274780e-05, {   2,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.668843959857e-04,  1.574425436348e-04, {   2,  -1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  4.603203776397e-05,  4.780380562168e-04, {   3,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.139643689999e-04, -4.093187332225e-04, {   2,   1,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.139645085476e-04, -4.093076167825e-04, {   2,   1,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  4.293744523831e-05, -4.780757009056e-04, {   1,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.333316182522e-04, -1.874025978636e-04, {   0,   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.902218773461e-04,  3.289081142643e-04, {   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.189823070638e-04, -1.345212419197e-07, {   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.126986712019e-04,  6.144538839938e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.340282002870e-04, -2.841313192333e-04, {   0,   1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.039480758157e-04, -1.404187539231e-05, {   2,  -1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.451195771981e-04,  1.723706107361e-04, {   4,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.927103237969e-04,  3.240282773317e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  6.552242504836e-05, -4.501546799936e-04, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.239130536594e-04,  2.887986839009e-04, {   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.712628342360e-04, -4.139924779680e-05, {   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.417710634335e-04,  1.706991725455e-04, {   4,  -1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -9.976432347911e-05,  4.122926301466e-04, {   2,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.402653177092e-04,  1.698762778803e-04, {   2,  -1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.099999999586e-04,  6.497860804999e-09, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.986479760693e-04, -3.105301371503e-04, {   2,  -1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.472959812228e-04, -2.574006545864e-04, {   0,   1,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.448077645310e-04, -2.585953612264e-04, {   0,   1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.234839231900e-04, -1.793418394535e-04, {   2,   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -3.659722187582e-05,  4.623110159405e-04, {   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  3.225415541004e-06,  4.953937038313e-04, {   1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.964727002933e-04, -1.762559424525e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.930803148355e-04, -3.045285233571e-04, {   2,   1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.316630899203e-04, -1.656495656666e-04, {   1,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.310577339096e-04, -1.653478539610e-04, {   1,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.246787695701e-05,  4.436371608732e-04, {   1,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.952694116393e-04, -2.964398996082e-07, {   2,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.087255010904e-04,  3.863002514753e-04, {   0,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.180215099056e-04, -1.755097960757e-04, {   0,   1,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -1.094599228853e-04,  3.837144258265e-04, {   0,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.145427562505e-04, -1.772657371413e-04, {   0,   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.118935910883e-04, -7.816269424787e-05, {   0,   1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -9.340762120615e-05,  3.966370410542e-04, {   1,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.899999999239e-04,  8.637634467673e-09, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.899999999634e-04,  5.986479334192e-09, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.899999999749e-04, -4.960225734199e-09, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.089620176386e-04, -1.796105083246e-04, {   0,   1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.072677659847e-04,  8.005599719207e-05, {   1,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.095111105361e-04,  3.764335261814e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 {  3.215106721532e-04, -1.644275544766e-04, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  5.110254161101e-05, -4.332626549910e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.949332225263e-05, -4.611642284042e-04, {   4,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.716041734029e-05,  3.833871708884e-04, {   2,   1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.799999688324e-04, -1.572401988743e-07, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.799999999447e-04,  7.288494956048e-09, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -3.709447969502e-04,  1.086141970609e-04, {   1,  -2,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.373659436690e-04,  4.099270392111e-05, {   1,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.936062507417e-05,  4.382359423801e-04, {   1,   0,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.935634182264e-05, -4.382363270638e-04, {   1,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.749297952707e-04,  5.624575289988e-07, {   4,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.984641018738e-05,  4.255339528560e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.674534808539e-04,  1.071349100704e-04, {   3,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.718900102821e-04,  2.015584856343e-07, {   2,  -1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.716434432776e-04,  1.233541544906e-07, {   4,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.986310724049e-04,  1.701977808319e-04, {   2,  -1,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.750818397029e-04, -2.933717580143e-04, {   2,   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -4.612841618885e-04,  5.907376614908e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.952752614449e-04, -1.717982320061e-04, {   0,   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  4.322389480241e-04, -3.419413245336e-05, {   2,   1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  9.340824162869e-05, -3.729135297461e-04, {   1,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.300816332548e-04,  1.319315763109e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  4.549978744973e-04,  6.780010091096e-06, {   2,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.400836964418e-04, -1.210537081270e-04, {   2,  -1,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -3.879573831317e-04,  7.282266562103e-05, {   3,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.832574811154e-05,  3.719912116616e-04, {   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.574982910682e-04, -7.512152795327e-07, {   0,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.537518360500e-04, -1.029691954397e-04, {   0,   1,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  8.769106198630e-05, -3.690178186184e-04, {   1,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.533987404397e-04, -1.029289182139e-04, {   0,   1,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.932251947507e-04, -1.625110850403e-04, {   0,   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.882995254464e-04, -1.663036371958e-04, {   0,   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -8.482079599911e-05,  3.690779228650e-04, {   2,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.490941297360e-04,  4.379982093592e-06, {   4,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.913619305046e-06,  4.493522317022e-04, {   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.326253442475e-04,  3.188323767985e-04, {   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.769658466724e-04, -7.240982843976e-05, {   1,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.475353602528e-04, -1.016457402734e-04, {   1,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.836188150035e-04, -1.655337382986e-04, {   0,   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.438867137645e-04,  4.185914486614e-06, {   2,  -1,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.106666780990e-04, -3.610590972602e-05, {   2,   1,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -4.418452299569e-04,  4.166757298179e-06, {   2,   1,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.007241639086e-04,  3.451599510570e-04, {   0,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.776247504383e-05, -4.279733353309e-04, {   2,  -3,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.480294035422e-05, -3.608787527450e-04, {   1,   0,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.382322073865e-04,  5.192166815419e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -4.229876817019e-04,  1.911307252294e-05, {   4,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.377459941240e-04, -4.181330423285e-06, {   2,  -1,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.367316534479e-04,  5.194915745639e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -8.479528078506e-05,  3.560399950836e-04, {   1,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.985072871437e-05,  3.494997429255e-04, {   3,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  8.459016062410e-05,  3.540657553824e-04, {   1,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.185290818521e-04, -1.189084690612e-04, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.115872887279e-05, -3.861078806193e-04, {   1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  6.048538433754e-05, -3.759622825809e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.841942628388e-04, -2.515105236778e-04, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.192601520062e-04,  3.159445274833e-04, {   1,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -8.616187331145e-05, -3.490077078206e-04, {   1,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.001125006779e-05, -3.746325568874e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  4.626854432846e-05, -3.856942841902e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.367516738061e-05,  3.463186214392e-04, {   0,   1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.586789581892e-04, -2.711260716648e-04, {   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.396190320636e-04,  8.924119210490e-05, {   2,   1,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -3.396190320636e-04,  8.924119210490e-05, {   2,  -1,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -4.753279314848e-06, -4.235374827796e-04, {   2,  -1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.891940323286e-05, -3.692035023337e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  5.851587402475e-05, -3.688690626668e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.164955664645e-05, -3.349072963525e-04, {   2,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.626283534217e-04,  6.350964847238e-05, {   4,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.795438471694e-04, -2.459719400383e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.950993196002e-05,  3.341107311499e-04, {   1,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.020888657092e-04, -3.190488340234e-04, {   2,  -1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.019397298131e-04, -3.185385672658e-04, {   2,  -1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.199999999659e-04,  5.351179486470e-09, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.199999999704e-04, -4.984660343494e-09, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.199999999713e-04, -4.911356515101e-09, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  8.002585067209e-05, -3.398495076720e-04, {   0,   1,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.171301962753e-04,  2.693058393258e-06, {   2,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.986765560504e-05, -3.998416462346e-04, {   2,  -1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.368817088336e-04,  2.827055968605e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -3.152145737931e-06, -4.162716870104e-04, {   2,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.183892580208e-04,  3.213692368401e-07, {   2,  -1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.154713063778e-05, -3.859492463341e-04, {   2,  -1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  4.168844179798e-04, -4.302896103812e-07, {   4,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.627842043977e-04, -2.540386796365e-04, {   1,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.419535316375e-04, -7.461757300157e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.137427264644e-04,  7.994757967924e-07, {   2,   1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.399366344915e-05, -3.784764600940e-04, {   1,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.879172286652e-05,  3.331517792619e-04, {   3,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.781325272916e-04, -1.329901127997e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  4.099999999667e-04, -5.223770451109e-09, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.997278106051e-04, -9.876138000137e-06, {   2,   1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.268720405236e-05, -3.956645209768e-04, {   2,   1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  6.910344020698e-05,  3.391492543297e-04, {   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.626333637792e-04, -1.449979406483e-04, {   0,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.469706076284e-04,  1.585331251567e-04, {   0,   1,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.468483383184e-04,  1.584602298971e-04, {   0,   1,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.026697928832e-04,  7.715459748197e-07, {   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.311369358015e-05,  3.685152416081e-04, {   1,  -2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.310269518004e-05,  3.685162297265e-04, {   1,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.309799994761e-05,  3.685166514548e-04, {   1,   0,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.588438666483e-04, -1.426492419238e-04, {   0,   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.877116549847e-04,  1.120332525052e-04, {   2,  -1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -3.939939206589e-04,  3.149027100800e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.177769901486e-04, -2.787776064413e-04, {   1,  -2,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -5.669610368814e-05,  3.392924531760e-04, {   1,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.950306378530e-04,  2.280522993934e-07, {   4,  -1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.193707840562e-05,  3.331124852273e-04, {   1,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.543125745725e-04, -1.398806045678e-04, {   2,   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.908100364755e-04,  1.864614015753e-06, {   0,   1,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.528297770766e-04,  2.385454703234e-04, {   0,   1,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -1.118772537728e-04, -2.790629951926e-04, {   1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.522640333930e-04,  1.385924240004e-04, {   1,   1,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.522640333930e-04,  1.385924240004e-04, {   1,  -1,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -3.899999999486e-04,  6.330309196705e-09, {   2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.897526556331e-04, -1.471436628418e-07, {   2,   1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.684688615759e-04, -2.193464377343e-04, {   1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  5.510676934808e-05, -3.318053575434e-04, {   1,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.232655241893e-04, -6.059302556356e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.831003620171e-04, -2.061335089095e-07, {   2,  -1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.828456327742e-04, -3.644680654533e-08, {   4,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.697125477780e-04, -1.191786702537e-05, {   2,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.953806787262e-04,  8.487189877607e-05, {   2,   1,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.677959571994e-04,  2.123409599311e-04, {   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -3.799999998698e-04, -9.948376735185e-09, {   2,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.096589855508e-05,  3.082427673630e-04, {   2,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.356216605919e-04,  1.422661212280e-04, {   0,   1,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.354798315176e-04,  1.421790362786e-04, {   0,   1,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.670946038268e-06, -3.715974763011e-04, {   2,   1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.756094235559e-04, -4.355553253568e-07, {   0,   1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.417921075001e-04, -1.326001595335e-04, {   1,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.865538387982e-05,  3.156880918797e-04, {   3,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.703028665038e-04,  2.872492929795e-06, {   0,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  3.729893103455e-04,  5.506200797714e-09, {   2,  -1,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.683082123252e-04, -4.466989883109e-06, {   1,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.942226945142e-05, -2.826621596341e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  1.169207914850e-06, -3.695369138664e-04, {   2,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.134080702866e-04, -5.663661570475e-05, {   0,   1,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.390037108373e-04, -2.986979821902e-05, {   2,  -1,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -2.380181975094e-04,  1.291629279913e-04, {   2,  -1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  1.480498646454e-05, -3.517142810243e-04, {   2,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.874500211040e-04, -7.899030090535e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.941183153480e-05, -2.968193190202e-04, {   1,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.426673277247e-05,  2.901021092869e-04, {   2,   1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.794718315643e-06, -3.615126054174e-04, {   1,   1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.022101654236e-04,  2.610030827991e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 {  7.051923370834e-05,  2.926127970318e-04, {   3,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.455019358388e-04, -1.165304226387e-04, {   1,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.124570429325e-04, -1.487023416170e-04, {   1,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -3.450011553822e-04,  1.510444045885e-05, {   4,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.599999999708e-04,  4.586725274117e-09, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.891515957585e-05,  2.904768135262e-04, {   1,   0,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.549768932215e-04,  4.339342418837e-06, {   1,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.609663165417e-04, -1.958282939978e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -8.175737335481e-05,  2.749633124846e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -6.777584937467e-05,  2.882843690379e-04, {   1,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.401066607100e-07, -3.554737832846e-04, {   2,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.949056412253e-05,  2.855975402211e-04, {   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -2.364359873027e-04,  1.180884186218e-04, {   2,   1,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.526755687559e-04, -9.826469079381e-08, {   0,   1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.991142033134e-05, -2.915326597202e-04, {   1,  -2,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -2.933516917528e-04, -5.766831066571e-05, {   1,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.532832332176e-04, -9.741539452188e-05, {   2,  -1,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -6.618922392369e-05, -2.839370039425e-04, {   0,   1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  3.056310960028e-04, -4.414446591654e-05, {   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.648412205619e-05,  2.825031111157e-04, {   1,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.697441235756e-04,  7.862687517453e-05, {   3,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -5.748228387781e-05,  2.904745454930e-04, {   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.320467534223e-04,  1.158939487029e-04, {   4,   1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.710165545280e-06, -3.410690569380e-04, {   0,   1,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.819121062506e-04,  6.582803204609e-05, {   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.922749773031e-05, -2.881467521481e-04, {   1,  -4,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -9.580491408553e-05,  2.515287207258e-04, {   1,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.452912409318e-04, -1.246378134354e-07, {   0,   1,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.451136157994e-04, -1.517969992004e-07, {   0,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.055256713110e-04, -3.968979256026e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.295898698241e-04,  1.146706607226e-04, {   2,   1,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.708815555278e-05, -3.164351067128e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  7.803042162497e-05, -2.653769669987e-04, {   4,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.797510708979e-05, -2.652672156623e-04, {   4,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.394042373658e-04,  2.712293552104e-06, {   0,   1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  6.608192327430e-05,  2.757701764055e-04, {   1,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.186599131289e-04, -1.223104389766e-04, {   2,  -1,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  3.399999999617e-04, -5.103342732573e-09, {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.399999999716e-04, -4.391248397940e-09, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.505315159466e-05,  2.741635039610e-04, {   3,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.433224368891e-04, -1.952873571254e-04, {   2,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.015582721568e-04,  2.367798632010e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.317058109423e-04,  2.057235487352e-04, {   0,   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.357252996177e-04, -1.001853482001e-04, {   0,   1,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.968868950110e-04, -3.857343593188e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  3.337093287753e-04,  4.187115456005e-07, {   4,  -1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.344773316271e-04, -9.963284868786e-05, {   0,   1,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.226880522021e-04,  1.112224761360e-04, {   4,   1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.865215144933e-04, -4.689110459855e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -6.720601707859e-05, -2.660916553026e-04, {   2,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.158532256197e-05, -3.216484889398e-04, {   2,  -1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.251195158239e-04, -2.076413021974e-04, {   1,   0,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.771303803953e-04, -5.505909608233e-05, {   1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.200115806653e-05,  2.901468763789e-04, {   0,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.767986866032e-04,  5.475531087488e-05, {   1,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.847312999947e-05, -2.616806777761e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.299999998648e-04,  9.445721910504e-09, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.299999999288e-04,  6.853907972089e-09, {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.299999999761e-04,  3.974114706695e-09, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.299999999924e-04, -2.246238747358e-09, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.278855509890e-04,  2.106127183107e-06, {   0,   1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.062163116205e-04,  1.234867949249e-04, {   0,   1,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.361583994803e-04,  1.934976284424e-04, {   1,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.061659746038e-04,  1.234578600283e-04, {   0,   1,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  2.193563942486e-04, -1.102222567860e-04, {   2,  -1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  3.268341475806e-04,  1.547053364733e-06, {   0,   1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.273852937612e-04, -1.321928016779e-07, {   2,  -1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.267318559945e-04, -2.005468101926e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 { -3.264489289089e-04,  5.135334681375e-07, {   0,   1,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.177226822813e-04, -9.064011713150e-06, {   2,   1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.989702840577e-04,  1.271446227989e-04, {   2,   1,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.294928803950e-04, -1.950258611379e-04, {   1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.978879283333e-04,  1.264336421944e-04, {   2,   1,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.781359253667e-04, -4.551248999260e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -3.232251761507e-04,  2.453901849074e-07, {   2,  -1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.958705730184e-06, -3.147237250653e-04, {   0,   1,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.012261425179e-06, -3.156319865148e-04, {   2,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.188106929686e-04,  3.413575028171e-06, {   4,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.013367422623e-05,  2.617510267278e-04, {   1,   0,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.447639281929e-05, -2.867808521258e-04, {   1,   0,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.160180260584e-04,  1.047794581613e-04, {   0,   1,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -6.278348636394e-06, -3.142108366259e-04, {   2,  -1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  3.202758197323e-04,  1.034718823426e-07, {   2,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.199999999733e-04, -4.132939668650e-09, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.682153412838e-05,  3.027501927449e-04, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -6.510779154988e-05,  2.541228886446e-04, {   2,  -1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.147212659753e-04,  1.042841069506e-04, {   0,   1,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.187732496569e-04,  2.150473006927e-07, {   2,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.199956560832e-04,  9.898425519100e-05, {   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.316772342655e-04, -1.871177938429e-04, {   1,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.301937622728e-04, -1.881924105732e-04, {   2,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.366565277281e-04,  1.814009489107e-04, {   2,  -1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -6.037193929551e-05, -2.573244997550e-04, {   1,   0,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.171642218909e-04,  2.375747407437e-07, {   2,  -1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.048733651529e-04, -1.123370857485e-04, {   1,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.047402004500e-04, -1.123773175637e-04, {   2,  -1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.665908281457e-04, -5.029986325408e-05, {   1,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.439783358771e-04, -7.219870809215e-05, {   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.155729228434e-04,  2.169645332928e-07, {   0,   1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.125226367262e-04,  1.995636279600e-06, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.412778996691e-04, -1.720921821462e-04, {   0,   1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.943460814791e-04, -1.188608726262e-04, {   2,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -6.024923567780e-05, -2.518812827476e-04, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.090329008818e-04,  1.030406971517e-04, {   2,   1,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.069575014775e-04,  4.924874797823e-06, {   0,   1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.485665448784e-04, -6.216432365055e-05, {   1,   2,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.201572462847e-04,  1.904261098816e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.099999999801e-04, -3.516838442733e-09, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.507779852683e-04, -1.588245572538e-04, {   0,   1,   1,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 {  1.507779852683e-04, -1.588245572538e-04, {   0,   1,  -1,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -9.581486714660e-05,  2.127621058121e-04, {   1,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  3.044240698683e-04, -4.139762361850e-06, {   2,  -1,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.029289657973e-04,  5.606477471044e-06, {   0,   1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -9.478848507910e-05, -2.132131986103e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -5.879447794849e-05,  2.479922973037e-04, {   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.632513669608e-04, -4.284173306947e-05, {   2,   1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.006596664894e-04, -2.049310555261e-04, {   2,  -1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.171564897824e-04, -1.883720475586e-04, {   2,   1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.985374725435e-04,  6.965689691923e-06, {   0,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.164031042362e-05, -2.426698249995e-04, {   1,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.572454913735e-04, -4.618582956255e-05, {   2,  -1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.283015496064e-04, -1.748878196268e-04, {   0,   1,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.282169506646e-04, -1.747695732844e-04, {   0,   1,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.937302819657e-05,  2.433807412717e-04, {   1,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.017656583694e-04,  1.007722706413e-04, {   2,  -1,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.988466829961e-04,  3.324762413268e-06, {   2,   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  5.808904198491e-05, -2.440229719816e-04, {   0,   1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.972362613970e-04,  2.615881703079e-06, {   0,   1,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.298934439997e-04, -6.986128618245e-05, {   0,   1,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.090507694463e-05, -2.883494933783e-04, {   0,   1,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.345201099616e-04, -1.640081702335e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.981521452166e-04,  3.702339167116e-07, {   0,   1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.349928610117e-04, -1.630575024639e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.428252225915e-05,  2.732989063366e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.956960211970e-04,  1.809045961062e-06, {   2,  -1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.934440180699e-04,  1.037184759731e-04, {   2,   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.533370280216e-04,  4.380788182407e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  7.784665506006e-05,  2.192492560058e-04, {   2,   1,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -7.783498565999e-05,  2.192140421109e-04, {   2,   1,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.587296025542e-04, -3.805065301510e-05, {   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  1.922726792974e-04, -1.044062649522e-04, {   2,  -1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.075179785407e-05, -2.842686041330e-04, {   0,   1,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  7.033894441970e-07, -2.937123474744e-04, {   1,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.970201370456e-04,  9.735092043815e-05, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.554297607886e-04,  3.872472633748e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 {  2.497137801009e-05,  2.688609034049e-04, {   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.773704281770e-05, -2.060098508048e-04, {   0,   1,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.498004570140e-04,  4.371359367704e-05, {   2,  -1,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.860007915311e-04, -1.071055970010e-04, {   2,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.249095627466e-04, -1.681881419958e-04, {   0,   1,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.495529902923e-04, -4.317116506668e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  6.290459983032e-05,  2.296384150433e-04, {   1,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.313241088101e-04,  5.995930219486e-05, {   2,   1,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.890935131423e-04,  1.577148869543e-06, {   2,  -1,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  7.545787482070e-05, -2.150820330422e-04, {   1,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.897965558120e-04,  6.326941642468e-07, {   2,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.899999999771e-04, -3.644247478074e-09, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.899999999777e-04,  3.593632929764e-09, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.899999999784e-04,  3.543018381460e-09, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.930387167920e-04,  9.641818965738e-05, {   2,  -1,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.436231651356e-04,  4.543858378541e-05, {   3,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.067488096035e-04, -1.818572096739e-04, {   0,   1,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.369842041060e-05, -1.940230951610e-04, {   0,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.405689319095e-05, -2.336233399342e-04, {   1,  -2,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.097135481648e-04, -1.771574465761e-04, {   1,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 {  1.539879773169e-04, -1.328619483965e-04, {   2,   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.914115402213e-04,  9.449319087621e-05, {   0,   1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -2.019867089973e-04,  8.346104033487e-05, {   0,   1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.908736947240e-04,  9.425527322569e-05, {   0,   1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.844984581668e-04,  7.107668029745e-08, {   1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.815842498203e-05,  2.263399762065e-04, {   3,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.798829442401e-04,  4.578160210955e-06, {   2,   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -2.769618699351e-04,  6.787933819070e-06, {   2,  -1,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.218582098266e-04, -1.618213983878e-04, {   0,   1,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  2.005937172757e-04,  8.279148968086e-05, {   0,   1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -8.495480532152e-05,  1.979071472683e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.093227049468e-04,  1.732602037405e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  2.326872040352e-05, -2.589566887547e-04, {   3,  -2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.757517537853e-05,  2.240179486149e-04, {   1,  -2,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.792664244016e-04,  2.172850329210e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.800180656018e-04,  1.245911392279e-06, {   2,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.388879402940e-04,  4.222309809862e-05, {   4,   1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.136019500384e-04, -1.669025160284e-04, {   0,   1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.055826512737e-04,  1.745452992709e-04, {   1,   0,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.799999999809e-04,  3.274237676667e-09, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.792839314108e-04,  1.750536864861e-07, {   2,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.111751486078e-05, -2.681170258232e-04, {   2,  -1,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.152640820462e-04, -6.395865992737e-05, {   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.988303551867e-05,  2.087668884131e-04, {   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.857580033220e-04,  9.277704947926e-05, {   4,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.361248491623e-04,  4.186326514305e-05, {   2,   1,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -8.171720012418e-05, -1.962477235485e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  2.765026789982e-04, -2.730868159988e-07, {   2,   1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.763104420893e-04, -1.786191513872e-07, {   0,   1,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.812673330274e-05,  2.475972101368e-04, {   1,   2,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.075891999600e-04,  1.679982979989e-04, {   2,  -1,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  2.204378446145e-04, -5.512861088644e-05, {   1,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.452541488497e-04, -1.300619818958e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.832274713631e-04,  9.151492813423e-05, {   4,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.122878510914e-04, -6.174344882742e-05, {   2,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.278663933666e-04, -4.562892393056e-05, {   1,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.520131665526e-05,  2.381424238242e-04, {   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.962148811799e-06,  2.651481465617e-04, {   0,   1,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.635507081767e-05, -1.863869524991e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.236809869579e-04, -1.470494802479e-04, {   1,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.883084946880e-06,  2.627058046207e-04, {   2,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.955929172165e-05,  1.908377254690e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -6.738622807344e-05,  2.029488343915e-04, {   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.438364442093e-04, -1.257609374204e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  7.818398952817e-05, -1.909601002071e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.159199072235e-04,  1.529829280161e-04, {   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.554378333687e-05, -2.532724254280e-04, {   2,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.637329946869e-04,  4.865980097001e-06, {   2,  -1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.655638609674e-04,  2.667760454845e-06, {   4,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.148701970633e-04, -1.533422983911e-04, {   0,   1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.013569872973e-06, -2.656306899347e-04, {   2,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.662286120221e-04,  1.784377361903e-07, {   4,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.302311910678e-04, -3.586406337154e-05, {   2,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.190487040232e-05,  2.139968086013e-04, {   0,   1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.651397145071e-04,  4.137302008027e-07, {   4,   1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.575644514123e-06, -2.636127312547e-04, {   3,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.050860573697e-05,  2.145818533873e-04, {   1,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.616967384853e-04,  1.033088015641e-04, {   2,  -1,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  8.095577106825e-05, -1.839652197843e-04, {   0,   1,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.715957919109e-05, -2.376667450877e-04, {   5,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.605712001926e-04,  4.043681409452e-06, {   2,  -1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.067557709923e-05, -2.532372869698e-04, {   2,  -3,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.607703474778e-04,  1.027070882232e-04, {   2,  -1,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.385823864802e-04, -1.248688326076e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.698626700331e-06, -2.578567110670e-04, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.749296319773e-04, -8.736806556971e-05, {   0,   3,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.011881563197e-04, -6.085896664829e-05, {   0,   1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.422844138216e-04,  1.195340902232e-04, {   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  2.554295800448e-04, -6.272722585271e-06, {   4,  -1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.527661374706e-04,  8.846389858151e-06, {   2,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.014551610677e-04, -1.599898750780e-04, {   2,   1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.082908874822e-04, -1.529647972927e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  2.929336123427e-06, -2.581194270061e-04, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  4.969938678372e-05, -2.111140450415e-04, {   2,   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -2.147172645568e-05, -2.390375806500e-04, {   1,   2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.511258610567e-04, -9.240522086886e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -9.668058702122e-05, -1.635870014134e-04, {   2,   1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.100320093268e-04, -1.501216148537e-04, {   0,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -9.585212553878e-05, -1.642846887489e-04, {   0,   1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.599999999817e-04,  3.085742117454e-09, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.325385012272e-04, -1.272872805910e-04, {   2,   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.566465563996e-04,  3.109297912015e-06, {   2,  -1,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.858369586727e-04, -7.386380135525e-05, {   2,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.941909038975e-05, -2.097314303441e-04, {   0,   1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.921238465729e-05,  2.096833244985e-04, {   1,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.378304815740e-04, -1.209367136623e-04, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.558767801668e-04,  2.696470793479e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.225536718918e-04, -3.545332111378e-05, {   2,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.573133928207e-04, -1.187175167231e-07, {   0,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.565835658369e-04,  1.623604116334e-07, {   2,   1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.055552996926e-04, -5.117566476468e-05, {   2,  -1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.564749515819e-04, -4.581844862153e-08, {   2,   1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.252579832251e-05,  1.737936480418e-04, {   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.515680541395e-05,  2.211061431946e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  2.032563588172e-04,  5.244105985719e-05, {   2,  -1,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.700227540751e-04, -8.491845930377e-05, {   0,   3,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.895666811426e-05,  2.053419750486e-04, {   1,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.838615335091e-05, -1.957400229385e-04, {   4,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.949055192599e-04, -5.908992530817e-05, {   0,   1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.834321129620e-05, -1.956233411476e-04, {   4,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.879588388492e-06, -2.520771711199e-04, {   2,   1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.216507963429e-04, -1.322682964954e-04, {   0,   1,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.800224225141e-04,  7.345839740365e-05, {   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  4.819381931677e-05, -2.051276571308e-04, {   1,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.155856152201e-04, -1.373665662492e-04, {   1,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 { -9.668299944093e-05, -1.558142868822e-04, {   2,  -1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -4.836905542334e-05,  2.038853813739e-04, {   1,   0,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.435883313328e-04,  8.238952821522e-06, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  1.539109278693e-04, -9.740753090918e-05, {   2,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.894438989194e-05,  1.822103545122e-04, {   0,   1,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.470316537050e-05, -1.661256566104e-04, {   2,  -1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  9.926502158564e-05, -1.513905738000e-04, {   2,   1,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.573117689468e-04,  9.325435947746e-05, {   1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.499999433701e-04, -7.549070386740e-08, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.499999999718e-04,  3.752457891647e-09, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.499999999780e-04,  3.316125578692e-09, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.499999999839e-04,  2.836160034430e-09, {   2,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.006186761914e-05, -2.394086213928e-04, {   2,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.454240602365e-04, -3.890912362735e-06, {   2,  -1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  6.523801250849e-05,  1.837377698434e-04, {   2,  -1,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -6.522766016185e-05,  1.837072939085e-04, {   2,  -1,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -9.854312300573e-05, -1.500351312556e-04, {   2,   1,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.702491958561e-05, -2.015055497162e-04, {   1,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.029698700139e-04, -1.453983909357e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -2.088270149497e-04, -3.883263820089e-05, {   2,  -1,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  4.439392282310e-05, -2.031398264057e-04, {   2,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.575786246997e-04, -8.938408376999e-05, {   2,   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.435650861728e-04, -1.032188420924e-04, {   2,  -1,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  1.158473975742e-04, -1.305871837724e-04, {   2,  -1,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.746559795187e-05,  1.989220753951e-04, {   3,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.472925419096e-05, -1.903077625871e-04, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -6.664154617349e-05, -1.782885132353e-04, {   2,  -1,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -6.640111142065e-05, -1.776677189841e-04, {   2,   1,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -2.875496641255e-06, -2.411306317931e-04, {   0,   1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.689898911574e-05,  1.759565821397e-04, {   0,   1,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.742977418760e-05, -1.746264064479e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  8.880736862279e-05, -1.530393595046e-04, {   0,   1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.170983348821e-04, -2.469552508027e-05, {   1,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.898131623793e-04, -5.132678940838e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.895032537137e-04, -5.133041323285e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.405214507353e-04, -2.574359818996e-07, {   2,  -1,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.847199163409e-05,  1.619942793094e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.169072774960e-04, -1.234301449831e-04, {   0,   1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.399999918906e-04, -5.551194139916e-08, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.399999999841e-04,  2.764601535098e-09, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.394384999871e-04,  6.532658254460e-08, {   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.746208043499e-04, -6.463978720163e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  9.255829174421e-05,  1.464966965881e-04, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -1.827164232251e-04, -5.535724241685e-05, {   0,   1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.509040600111e-05, -1.429767208895e-04, {   0,   1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.291263856389e-04, -8.439868294314e-06, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -8.725563485124e-05, -1.502478782749e-04, {   1,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.672234861799e-04,  6.977859749537e-05, {   2,   1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.229332958755e-05, -2.244027801494e-04, {   2,   1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -4.766335975100e-08,  2.366229653666e-04, {   2,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.338035277278e-04, -2.736601583106e-06, {   0,   1,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -9.041477410390e-05, -1.454451877246e-04, {   2,   1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -5.451418511063e-05,  1.811555922190e-04, {   0,   1,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.542188716218e-05, -1.901025502088e-04, {   1,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.822277570540e-04,  5.309369884672e-05, {   1,   0,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -7.219087946981e-05,  1.623950585321e-04, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.913758220653e-04, -4.313768900537e-05, {   1,  -2,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.209632345545e-05, -2.217721782075e-04, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -1.055039763226e-04,  1.277135036392e-04, {   0,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.428337510204e-04,  9.031473310013e-05, {   2,   1,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.329034923283e-04, -7.215346657013e-08, {   2,   1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.378314703761e-06, -2.304434087298e-04, {   1,  -1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.315952212084e-04,  1.136828360628e-07, {   2,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.418858480619e-04,  8.968623574537e-05, {   2,   1,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.930968283474e-04,  3.847449941636e-05, {   1,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  3.653127425042e-05,  1.945530694092e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 { -2.211311527710e-04,  9.202512467927e-06, {   4,   1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  8.809267673947e-05, -1.421606922762e-04, {   0,   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.299999913916e-04, -5.566902093415e-08, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.299999999545e-04,  4.576253298427e-09, {   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.655591165940e-06, -2.222426312421e-04, {   2,   1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.481139018746e-04,  8.149369013090e-05, {   0,   1,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -2.269184220755e-04,  2.502322120434e-06, {   2,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.608762076020e-04,  6.737218869498e-05, {   2,   1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  8.350368128750e-05, -1.441076232813e-04, {   0,   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.314761310447e-05, -1.841447029002e-04, {   1,   0,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.573978877544e-05,  1.904687301674e-04, {   0,   3,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.184565463784e-04,  7.309538552787e-06, {   2,   1,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  2.257007735612e-04, -3.468800518070e-08, {   4,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.455671173487e-04, -7.958815075050e-05, {   2,   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  4.365251404873e-05,  1.810392078009e-04, {   3,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.216386798066e-04, -2.920802033277e-06, {   2,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.583752452179e-05, -1.584902291072e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.484697160999e-04, -7.580553304257e-05, {   2,  -1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.176098948162e-05, -2.123215573270e-04, {   2,  -1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -6.931126590255e-05, -1.544398978149e-04, {   0,   1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -7.311024343919e-06, -2.161414707640e-04, {   2,  -1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.978572648456e-06, -2.184388037799e-04, {   0,   1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.231613359930e-04, -3.238307860442e-08, {   2,  -1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.227685917749e-04,  2.547557204069e-07, {   2,  -1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.277663387228e-05,  1.298271523697e-04, {   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -8.201786197660e-05, -1.399183131635e-04, {   2,   1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.358164973178e-05, -1.978270851781e-04, {   5,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.957507289000e-04, -2.551490246292e-05, {   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.257359489877e-05,  1.786471215516e-04, {   1,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.398165201768e-04, -8.131182368186e-05, {   2,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.562569750219e-04,  6.485920906833e-05, {   1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -4.338091081012e-05, -1.773635344087e-04, {   2,   1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.845808163122e-05,  1.520685807878e-04, {   0,   1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.685167278413e-04, -5.178223148771e-05, {   2,  -1,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.199999983544e-04,  5.736896936618e-09, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.535731524102e-05,  1.844367759727e-04, {   2,  -1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.860815022439e-04,  3.368324135359e-05, {   1,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.995823588443e-05, -1.897839229839e-04, {   1,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -2.155343595469e-04,  4.206807539398e-06, {   1,   0,  -1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.485509860335e-04, -7.092490238744e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -9.841874021750e-05,  1.208690253694e-04, {   2,   1,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.419937842811e-04,  7.708903055041e-05, {   2,  -1,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -2.147752728782e-04,  4.192102684424e-06, {   1,   2,  -1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.199528037646e-05,  1.366922835495e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  4.090219088108e-05,  1.776788121659e-04, {   1,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.287943716493e-04, -8.958223285815e-05, {   2,  -1,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -8.863408528690e-05, -1.293008279472e-04, {   0,   1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -8.091204841065e-05, -1.366872850697e-04, {   2,  -1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.394692071889e-05,  1.836204932832e-04, {   1,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.093758151653e-05, -1.762248221853e-04, {   2,   1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.120270468829e-05, -2.056635126186e-04, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  6.015950296504e-05,  1.565029935976e-04, {   0,   1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.733385343195e-04, -4.307342805219e-05, {   2,   1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.380410106586e-05,  1.822604930714e-04, {   1,   2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.681277434502e-06, -2.128070890472e-04, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -7.887540864730e-05, -1.364639188903e-04, {   0,   1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.732482809915e-04,  4.206872503138e-05, {   1,   0,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -3.851544213698e-06, -2.110850234950e-04, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  2.034530269435e-04, -1.103304513988e-05, {   0,   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.873928310764e-04, -2.695208940279e-05, {   2,   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.230890138196e-04,  9.120089340177e-05, {   1,   0,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.858901294393e-04, -2.828671380688e-05, {   2,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  8.278091955770e-05,  1.311866812726e-04, {   2,   1,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  4.079797275894e-05,  1.729477007361e-04, {   1,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.701441580747e-04,  4.355444542213e-05, {   1,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.912045024356e-05,  1.537272582486e-04, {   0,   1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.581978265141e-04,  5.434374092108e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -4.072714901483e-05,  1.716657655189e-04, {   3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.764618426379e-04, -3.589132325636e-05, {   1,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.876053542208e-04, -2.448826028913e-05, {   2,   1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.585193934139e-04,  5.255155052292e-05, {   1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.455015431584e-04, -6.553624704973e-05, {   1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  2.104915803823e-04,  1.257668685594e-07, {   4,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.099999999830e-04,  2.675589743235e-09, {   4,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.100000000000e-04,  0.000000000000e+00, {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.339129771742e-05, -1.959083337210e-04, {   2,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.386780028380e-04, -6.969738295889e-05, {   1,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  9.514423594958e-05, -1.131194786968e-04, {   1,  -2,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.342507151026e-04, -7.353089081562e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -3.160365186394e-05, -1.760268456184e-04, {   2,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.255671326969e-05, -1.647596298927e-04, {   1,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  7.986065227280e-05, -1.273721432770e-04, {   2,   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  7.775258361602e-06, -1.991072383325e-04, {   2,   1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.749739842047e-05, -1.291934825411e-04, {   3,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  7.636360098775e-05, -1.301823224799e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 { -7.923505510168e-05, -1.270380467834e-04, {   0,   1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.685239075780e-04,  3.760072291393e-05, {   2,  -1,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.855494749614e-05, -1.669607146571e-04, {   2,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.219527053920e-05,  1.731651237776e-04, {   1,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.032772904829e-05, -1.934537271306e-04, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  6.409328841075e-05,  1.396089062230e-04, {   0,   1,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.877861892507e-04,  1.568036752821e-05, {   1,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -6.200714961451e-05,  1.414335995553e-04, {   0,   1,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  4.452140776166e-05,  1.586430940098e-04, {   1,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.980048526668e-04, -5.139621402835e-06, {   4,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.973399204996e-04, -5.767374224151e-06, {   2,  -1,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.261594003609e-05,  1.902879180651e-04, {   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.507821977464e-04,  5.177981255391e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -6.575038834591e-06, -1.959734883861e-04, {   4,  -1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.712606827218e-04, -3.074943864455e-05, {   2,   1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.501787252804e-04,  5.166099581938e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.155495472170e-05,  1.797041708932e-04, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.906668893094e-05, -1.619586605346e-04, {   0,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.197150610780e-05, -1.690090357748e-04, {   1,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.834501795576e-04,  1.739329804373e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.956162652548e-04, -5.136686533776e-06, {   0,   1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.000395163076e-04, -4.963824532260e-08, {   0,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.459936100333e-04, -5.403846955244e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.999999999689e-04, -3.525565088874e-09, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999859e-04, -2.373647782616e-09, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.986748368397e-06, -1.979117567942e-04, {   0,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.402751736935e-05, -1.158329046479e-04, {   2,  -1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.351601749275e-04,  6.468609497564e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  1.981683883178e-04,  1.165916175785e-06, {   2,   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -4.164872009092e-05,  1.570093121273e-04, {   2,  -1,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  7.720818088408e-06, -1.905332820588e-04, {   2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.971042007021e-04,  9.257233612943e-07, {   0,   1,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.047100797498e-04,  9.319683990436e-05, {   2,  -1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  4.147414924485e-05,  1.563474717013e-04, {   2,  -1,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.969441268239e-04, -1.767250127842e-08, {   2,  -1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.603864693573e-04,  3.653912898248e-05, {   1,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.955214877903e-05,  1.266725446601e-04, {   1,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.603513600992e-04,  3.582373609757e-05, {   4,  -1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.613755546869e-04, -3.411060901285e-05, {   1,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.741948620686e-05, -1.578666810384e-04, {   1,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.145676784321e-05,  1.338069091736e-04, {   0,   1,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.874338713530e-05,  1.760070196363e-04, {   2,   1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.153574135643e-05,  1.031818808044e-04, {   1,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.693042860650e-05, -1.577466317153e-04, {   1,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.537242292160e-05, -1.688822180826e-04, {   2,  -1,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  9.120976607366e-05, -1.028198530478e-04, {   2,   1,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.519748044934e-05, -1.687286645624e-04, {   2,  -1,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.935517266279e-04, -4.370469739059e-08, {   0,   1,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.363085663205e-04,  5.696155351623e-05, {   2,  -1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.594773639571e-04, -3.368812156079e-05, {   4,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  8.237522995008e-06, -1.848467068102e-04, {   0,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.586722669205e-05,  1.454222196015e-04, {   2,   1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.074981954070e-04, -8.305548623351e-05, {   1,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.875287131676e-04, -2.885143897987e-06, {   2,  -1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.706435528941e-05, -1.531754143994e-04, {   2,   1,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.609905666634e-05, -1.539552324669e-04, {   1,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.899999999473e-04, -4.476769530958e-09, {   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.899999999739e-04,  3.150319299705e-09, {   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.899999999842e-04,  2.453932928236e-09, {   4,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.899999999858e-04,  2.321287905095e-09, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.294299469553e-05,  1.566188393055e-04, {   0,   1,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.866792989990e-04, -2.487685115600e-06, {   2,   1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.323059507206e-05, -1.159180168588e-04, {   1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.096906107529e-05, -1.380088958732e-04, {   1,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.125777433968e-04, -7.610479498041e-05, {   2,  -1,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.257782321940e-04, -6.282023396047e-05, {   0,   3,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.577507343519e-05,  1.527309251901e-04, {   1,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.878281487349e-04, -3.288573074554e-07, {   4,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.676187268347e-04,  2.041937545796e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.551826567626e-04,  3.241195134877e-05, {   2,   1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.122114942507e-04,  7.528431422310e-05, {   0,   1,   1,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -1.122114942507e-04,  7.528431422310e-05, {   0,   1,  -1,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  3.817442487338e-05,  1.487000899579e-04, {   1,  -2,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  7.204759681474e-05, -1.148177684481e-04, {   2,  -1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.233037916400e-06, -1.852125577452e-04, {   1,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.863150306871e-04,  5.407399180075e-08, {   2,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.202171760331e-04,  6.603880414150e-05, {   0,   1,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.372362199329e-05, -1.518420164331e-04, {   3,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.236513223449e-04, -6.176728833975e-05, {   0,   3,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.373616623991e-05, -1.016622802511e-04, {   2,   1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.306784404779e-04,  5.466555787858e-05, {   2,  -1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -6.806481634538e-05, -1.170726058904e-04, {   0,   1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.477689078376e-05, -1.502777115357e-04, {   1,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.482939848017e-05, -1.501783248191e-04, {   2,  -1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.520810526594e-05,  1.693183791979e-04, {   1,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.556079774050e-05, -1.484258522097e-04, {   1,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.191001860272e-04, -6.471594539731e-05, {   2,   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.823530305644e-06, -1.772892571958e-04, {   2,  -1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.385699146310e-05,  1.292262530237e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.032638789507e-05,  1.717762687404e-04, {   2,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.784393351425e-05, -1.038886846543e-04, {   4,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.218287970549e-05, -1.389906321516e-04, {   4,   1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.214874522106e-05, -1.388914536849e-04, {   4,   1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.651770752867e-04, -1.510820371573e-05, {   0,   1,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.799999999644e-04, -3.581415624832e-09, {   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.797196976809e-04,  1.784052494916e-07, {   1,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.443860021559e-05,  1.451896314943e-04, {   1,   0,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.791704926129e-04,  3.219171328512e-07, {   4,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.586619184828e-04, -2.060445563237e-05, {   2,  -1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.433092743864e-04, -3.575458416283e-05, {   2,  -1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.776632068422e-04, -1.015954194616e-06, {   0,   1,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.524879245924e-05,  1.427412636735e-04, {   1,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.480024563545e-04, -2.957586419381e-05, {   1,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.719960653185e-04,  5.557990398943e-06, {   2,   1,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.320719937134e-05,  1.539191585837e-04, {   2,  -1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -1.747098526665e-04, -2.406071560917e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.724725473762e-05, -1.496685615720e-04, {   1,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.673572657587e-04, -9.390636464121e-06, {   0,   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -3.488931074987e-05, -1.418198571891e-04, {   2,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.674887461103e-05, -1.098657324661e-04, {   0,   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.764016649706e-04, -1.356439297614e-07, {   0,   1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.484563401679e-04,  2.788135255769e-05, {   1,  -2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.164920636478e-05,  1.144525026139e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.230253783879e-04,  5.293621831330e-05, {   2,   1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.168409855770e-04,  5.893674571377e-05, {   0,   1,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.387884198329e-06, -1.743718335376e-04, {   2,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.356741936965e-04,  3.993540987701e-05, {   1,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.825261051640e-05, -1.271723644140e-04, {   0,   0,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.747428739954e-04, -2.437709603451e-07, {   0,   1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.133408837629e-04,  6.098373001735e-05, {   2,  -1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -1.178604691394e-04, -5.644142811188e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  6.982600730878e-06, -1.671158869768e-04, {   2,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.580655440561e-04, -1.597649395040e-05, {   2,  -1,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.672306861473e-04, -6.663600923622e-06, {   4,  -1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.719746378643e-04,  1.489124101371e-06, {   4,   1,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.728142918174e-04,  5.934596479148e-07, {   2,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.307878286210e-05,  1.400219622009e-04, {   1,   0,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.386434405608e-04, -3.427566692429e-05, {   2,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.336007018148e-06, -1.686533329556e-04, {   0,   1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.718619878556e-04,  9.331599144466e-09, {   2,   1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.674278994211e-04, -4.208334024357e-06, {   4,   1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.422744700826e-04,  2.933365882614e-05, {   2,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.149259541424e-04,  5.666563842185e-05, {   2,  -1,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.712317300128e-05,  1.444677145164e-04, {   0,   1,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.244574107256e-05,  1.489975276556e-04, {   2,   1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.001579417765e-05, -1.608073821818e-04, {   2,  -1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  4.467017383302e-05,  1.258099935451e-04, {   2,   1,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -4.465924105072e-05,  1.257764881714e-04, {   2,   1,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  3.479894635780e-05,  1.354396521278e-04, {   1,   0,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.694959953035e-04, -6.172132332794e-07, {   2,   1,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  6.690719514788e-08,  1.699999868336e-04, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.699999999830e-04,  2.403318379916e-09, {   4,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.699999999862e-04, -2.165953601679e-09, {   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.699999999866e-04, -2.136283004388e-09, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.045018372289e-04, -6.543875854292e-05, {   2,  -1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -6.183870813816e-05, -1.079058874376e-04, {   2,  -1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  5.166689230827e-05,  1.177023680767e-04, {   0,   1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.447726073996e-04, -2.449411110152e-05, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.614354235585e-04, -7.387275772461e-06, {   2,  -1,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -1.357759356270e-04,  3.297404911354e-05, {   1,  -2,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  2.009541125355e-05,  1.486478201877e-04, {   0,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.175073597292e-04,  5.098371845568e-05, {   2,   1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -5.177389180927e-06, -1.631761917519e-04, {   0,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.084932666886e-04, -5.979250603423e-05, {   2,  -1,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  6.494479742734e-05, -1.033048663433e-04, {   2,  -1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -5.156276657995e-05, -1.166131162876e-04, {   0,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.225060405157e-05, -1.257438576539e-04, {   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.523785196675e-05, -1.026874656675e-04, {   4,  -1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.424571040409e-05,  1.136243442263e-04, {   2,  -1,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.075731962382e-05,  1.569365782249e-04, {   2,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.672123421716e-04,  1.834860508014e-07, {   4,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.838515107168e-05, -1.385976037866e-04, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -1.596251317107e-04,  7.313432576404e-06, {   4,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  8.900625812888e-06,  1.580371959506e-04, {   2,  -1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.483238244607e-05, -1.019284035905e-04, {   4,  -1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.186072603372e-05, -1.347561626767e-04, {   1,   0,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.605916236200e-04, -5.849278435611e-06, {   0,   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.367632944279e-04,  2.955142207825e-05, {   0,   1,   1,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 { -1.367632944279e-04,  2.955142207825e-05, {   0,   1,  -1,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -7.607599783472e-05,  9.001153799352e-05, {   2,  -1,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.151098441679e-05,  1.145415762380e-04, {   0,   1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -3.679817049222e-05, -1.290657019202e-04, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  5.000931559994e-05, -1.156583335289e-04, {   4,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  5.871296876581e-05,  1.068640537209e-04, {   1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.305106310006e-04, -3.458725493692e-05, {   1,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.313377295004e-05,  1.518329253787e-04, {   2,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.638041809333e-04,  9.782284684940e-07, {   2,  -1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.049960976891e-06, -1.595362296703e-04, {   4,  -1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.562584465887e-05,  1.489480186400e-04, {   1,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.059504519634e-04, -5.858865151027e-05, {   2,  -1,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -1.369409037449e-04,  2.751067794342e-05, {   1,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.161981203319e-05,  1.326612234223e-04, {   1,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.617514099661e-04, -2.124695148565e-06, {   2,  -1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.358054904039e-04, -2.797856991496e-05, {   1,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.339785289945e-05,  9.035470847678e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.557132035155e-04,  7.913729642558e-06, {   0,   1,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.631646878181e-04,  2.715992018640e-07, {   0,   1,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.629178645098e-04,  2.478375454849e-07, {   0,   1,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.553420874937e-04,  7.767671488838e-06, {   0,   1,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -3.090450766472e-05, -1.321301841033e-04, {   1,   0,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.341789948700e-05,  1.493986612167e-04, {   1,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.369316861029e-04,  2.584470521932e-05, {   1,   0,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.253544184791e-04,  3.694539419984e-05, {   3,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.317750136713e-04,  2.979614866008e-05, {   4,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.600665476488e-04, -3.244959732679e-07, {   2,  -3,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.069550413995e-04,  5.337760407524e-05, {   3,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.504019925523e-05,  6.500746102114e-05, {   1,   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  1.526718718450e-05,  1.447636499585e-04, {   2,  -1,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.599999999811e-04, -2.457423586698e-09, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.086093100192e-04,  5.134747774561e-05, {   1,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  8.899870747691e-05, -7.068045760424e-05, {   4,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.647708179259e-05,  1.229598642088e-04, {   2,   1,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.195506328012e-05,  1.171738645407e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 { -2.969884346916e-05,  1.292447654062e-04, {   1,  -2,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.103741971078e-04,  4.853555136329e-05, {   2,   1,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.454871440681e-04, -1.329185891768e-05, {   2,  -1,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  1.592759408628e-05, -1.427486510002e-04, {   2,  -1,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.307312917733e-04,  2.727647766236e-05, {   2,  -1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.614321635146e-05,  1.218364382357e-04, {   2,  -1,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.336026378308e-05,  1.443072166985e-04, {   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.846985208581e-06, -1.517225090176e-04, {   2,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.344829050709e-05, -7.405670463012e-05, {   4,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  5.976295887161e-05, -9.773210614242e-05, {   0,   1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -3.067987930524e-05,  1.267893303327e-04, {   2,   1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.405724774761e-05,  1.029501223727e-04, {   1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.064462692342e-05, -1.259048818501e-04, {   1,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.756055020790e-05,  1.187996440911e-04, {   2,  -1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.767694888504e-05, -9.828227943512e-05, {   0,   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.942744520151e-06, -1.529524422727e-04, {   0,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.079531842566e-05, -1.448228176731e-04, {   0,   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.090245391562e-04,  4.656749739121e-05, {   2,   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.555621902208e-04, -7.921522308007e-10, {   2,  -1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.421006818171e-04, -1.334795356716e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.087915018389e-04,  4.632543602212e-05, {   2,  -1,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.548635580183e-04,  2.389977195669e-07, {   2,  -1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.540829449071e-04, -5.113257866256e-07, {   2,   1,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  2.280697666399e-05,  1.315059358249e-04, {   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.539824073767e-04, -3.139468734522e-07, {   4,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.322848502781e-06, -1.526136633302e-04, {   0,   1,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  4.520234117238e-05, -1.084597170145e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -9.872831841822e-05,  5.473435974701e-05, {   2,   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.518665469599e-04,  1.259252156445e-06, {   0,   1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  8.247115356771e-06, -1.447840923525e-04, {   2,  -1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.283157803818e-04,  2.468591172821e-05, {   1,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.788988789280e-05, -1.249806638063e-04, {   1,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.469928177932e-04,  5.386250554926e-06, {   0,   1,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.252906762089e-05,  1.394382388893e-04, {   3,  -2,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.252402994403e-05,  1.394386914524e-04, {   1,   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.517014393612e-04, -3.789403543445e-08, {   4,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.012607553625e-04, -5.030050575167e-05, {   1,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.236016232207e-04,  2.755096346766e-05, {   0,   1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.230510317754e-04,  2.759392145790e-05, {   4,   1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.499992660698e-04,  5.550489619254e-07, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.863884397780e-05, -8.174382914631e-05, {   1,   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -1.502988630743e-04,  1.176046433628e-08, {   0,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.187779984618e-05,  7.813447215637e-05, {   2,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.499999999875e-04,  1.937315469660e-09, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.499999999888e-04, -1.832595714535e-09, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.499999999891e-04,  1.806415775770e-09, {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.102841516142e-04, -3.971609734410e-05, {   2,   1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  9.624154812694e-05,  5.360853703336e-05, {   2,  -1,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  9.990641003507e-05,  4.989834572986e-05, {   0,   1,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.911825023362e-05, -1.203365620110e-04, {   2,  -1,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.100764340507e-04, -3.928971323837e-05, {   2,   1,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  2.366579957074e-06, -1.469863521732e-04, {   0,   1,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  9.958516768795e-05,  4.968037871842e-05, {   2,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.080904737315e-05, -1.381167015024e-04, {   0,   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.020041652039e-04,  4.687277392394e-05, {   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.481678731371e-04,  2.385653057192e-07, {   0,   1,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.618482473379e-05, -9.188225941318e-05, {   0,   1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.452851756433e-04,  2.619292211199e-06, {   2,  -1,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -1.150678789641e-04,  3.233358309645e-05, {   2,   1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.090285861179e-04, -3.823909376580e-05, {   2,   1,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.613897104141e-05, -5.090587343335e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.783257836841e-05,  1.290392990231e-04, {   2,  -1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.465712998496e-04,  3.175418058879e-09, {   2,   1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.463966032747e-04, -5.599316567263e-08, {   4,   1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.182999817671e-05, -1.143197510401e-04, {   2,  -1,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -3.182978364299e-05, -1.143180227541e-04, {   2,  -1,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.337758981567e-04, -1.230978776429e-05, {   0,   1,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.189441545474e-05, -1.340330410024e-04, {   2,  -3,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.921044601918e-06, -1.436698410930e-04, {   2,  -1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.199356168416e-05, -8.356466773500e-05, {   0,   1,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.227167228213e-05, -1.131522632924e-04, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -6.194048192047e-05, -8.342146481661e-05, {   0,   1,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.653880723907e-05,  4.848992726312e-05, {   0,   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.391836882853e-04,  5.803872488592e-06, {   2,   1,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  8.556966064608e-05, -5.900770425958e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   6,   0,   0,   0 } },
 {  1.154945736726e-04, -2.899277052649e-05, {   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.400062561820e-05, -5.045942463186e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -6.307810134737e-05,  8.113514698343e-05, {   2,  -1,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.440260813456e-04, -1.227634182366e-07, {   2,  -1,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.963951449107e-05, -1.044707021259e-04, {   0,   2,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.237695685705e-04,  2.033692535497e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  1.015445855347e-04,  4.229299171944e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.194131217256e-05,  5.184911369921e-05, {   0,   1,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -1.426443854424e-04, -1.140615405472e-06, {   4,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.760695275227e-05,  8.584005153889e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.057352072396e-04,  3.763932614613e-05, {   2,   1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.248484601770e-04, -1.836827043308e-05, {   0,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.751603547936e-05,  1.156583567621e-04, {   3,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.430200462914e-04,  1.630179413359e-08, {   0,   1,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.114059236444e-04,  3.113440033704e-05, {   1,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.413718645337e-04,  1.136807747244e-06, {   0,   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  2.971838956868e-05, -1.125789921943e-04, {   1,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.046149044045e-04, -3.764114871440e-05, {   2,   1,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -3.107292186545e-05, -1.111443858436e-04, {   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.513196621056e-05,  4.688484337814e-05, {   2,  -1,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.136351887987e-05, -1.105951345405e-04, {   2,  -1,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -9.466350929414e-05, -4.727974293054e-05, {   2,  -1,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.135261034836e-05, -1.105694469820e-04, {   2,  -1,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.314701258892e-05, -8.866833837603e-05, {   2,  -1,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.416701571406e-04,  4.822662520763e-08, {   0,   1,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.082042502527e-05,  1.106948368908e-04, {   0,   1,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -3.082037224641e-05,  1.106941050884e-04, {   0,   1,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.165210618622e-05,  1.098573540079e-04, {   1,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.117949195657e-04,  2.969530193602e-05, {   2,  -1,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.422172058105e-05,  1.072338407172e-04, {   2,   1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.317315931088e-06,  1.378607937307e-04, {   0,   1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.163512854152e-05,  1.294782753354e-04, {   1,  -2,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.250860610218e-04,  1.601056235587e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -1.382821636435e-04,  2.520755441177e-06, {   2,  -1,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  2.686026571251e-05,  1.138873296665e-04, {   1,   0,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.403879587943e-04, -1.265936463239e-07, {   2,  -3,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.259329059328e-05, -8.782651640187e-05, {   2,  -1,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.707332252142e-05, -1.131670725424e-04, {   1,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.042015969150e-04,  3.587629267209e-05, {   2,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.399999999723e-04,  2.785545485999e-09, {   2,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.399999999742e-04,  2.687807047906e-09, {   2,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.399999999839e-04, -2.125811028828e-09, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.399999999864e-04, -1.954768762166e-09, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.399999999896e-04, -1.710422666899e-09, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.399999999992e-04, -4.642575810381e-10, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -3.402578089213e-05,  1.059375772880e-04, {   2,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.445967984883e-05, -9.543675725091e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  5.930444419756e-05, -8.051697273506e-05, {   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.967736875383e-05, -4.982051950626e-05, {   2,  -1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -9.147931443377e-05,  4.798048975310e-05, {   0,   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -9.828500016693e-05,  4.117208937341e-05, {   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -3.091449103653e-05, -1.083836944737e-04, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.362114875891e-07, -1.386580249707e-04, {   4,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.144715492398e-04,  2.451199978929e-05, {   2,  -1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  4.326040530100e-06, -1.344610928265e-04, {   0,   1,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -2.100681735699e-05, -1.177591538957e-04, {   0,   1,   1,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 {  2.100681735699e-05, -1.177591538957e-04, {   0,   1,  -1,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -4.078123195355e-05, -9.793746512414e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.331278521391e-04,  5.551734023987e-06, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  6.028240271039e-05,  7.835779866426e-05, {   2,  -1,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.569144297544e-05, -1.228055543463e-04, {   2,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.649307374242e-05,  1.119228747246e-04, {   1,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.381740573934e-04,  1.007703818091e-07, {   4,   1,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.646947272840e-05, -1.117989875219e-04, {   3,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.334172306859e-04,  4.789905358313e-06, {   2,  -1,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.368284796559e-04, -1.360099390290e-06, {   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.209639860826e-04, -1.672715670386e-05, {   2,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.604201095236e-05,  1.114193715542e-04, {   1,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.768025650813e-05,  9.977179945273e-05, {   1,   2,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.756225988664e-06, -1.345787907691e-04, {   0,   1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.834256927856e-05, -8.873925007681e-05, {   1,   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  1.359017422827e-04,  1.093835851969e-06, {   0,   1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.187448239727e-05, -8.511938755257e-05, {   0,   1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.224839116836e-04, -1.444468232529e-05, {   1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.656780996764e-06, -1.342156438457e-04, {   0,   1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  6.033666728064e-06,  1.307688861052e-04, {   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.242434872680e-04,  1.242823513869e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  1.054941425043e-04,  3.116064661204e-05, {   2,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.365264459898e-04,  1.020235563401e-07, {   2,   1,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -9.108978884066e-05,  4.549955654718e-05, {   0,   1,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -9.118473046793e-05, -4.529655300332e-05, {   1,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.360110366051e-04,  5.488026689973e-08, {   4,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.550551987575e-06, -1.324767984699e-04, {   0,   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.647173094770e-05, -1.094361857517e-04, {   1,   0,   0,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -1.281054720922e-04, -7.660172409418e-06, {   0,   1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  7.040808229213e-07, -1.350166438175e-04, {   3,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.868298954122e-05, -5.683702361266e-05, {   1,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -3.861331439914e-05, -9.635419192944e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.057524314578e-05,  1.243811806098e-04, {   2,   1,  -1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  1.057524314578e-05,  1.243811806098e-04, {   2,  -1,  -1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -1.333294235445e-04,  1.357577521980e-06, {   4,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.275827348742e-05,  8.177596555490e-05, {   2,  -1,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  6.646680213269e-05,  6.779832077704e-05, {   2,   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.530200407065e-05, -1.089357978379e-04, {   0,   1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.147848501240e-06, -1.330651383687e-04, {   2,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.314121111194e-05,  4.102181930230e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.332845543176e-04,  8.584612132401e-07, {   0,   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.956887436060e-05, -1.044362341329e-04, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  4.341681365928e-05, -8.972508372567e-05, {   1,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -2.680540948599e-05, -1.062447595347e-04, {   3,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.322757099442e-04, -4.362338951695e-07, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.777843088334e-05,  1.047788173102e-04, {   2,   1,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.201430711505e-04, -1.222753108832e-05, {   1,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.315617146762e-04, -7.073993983584e-07, {   4,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.965928892961e-05, -7.255735796190e-05, {   2,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -7.920008564821e-06,  1.239850393969e-04, {   2,  -1,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -8.763723035816e-05, -4.387746580423e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.312842954478e-04,  1.149518915196e-07, {   2,  -3,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.566711469252e-05, -1.056106122890e-04, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.286080896430e-05, -1.183918920263e-04, {   2,   1,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -3.930466124287e-05, -9.169308079100e-05, {   2,  -1,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  1.307081634252e-04,  9.610430878807e-08, {   2,  -1,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.303934930303e-04,  2.335061354875e-07, {   4,  -1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.849508024305e-06,  1.264921471374e-04, {   2,  -1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.073546688846e-05,  1.195188259258e-04, {   3,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.406437589507e-05,  4.614175688738e-05, {   1,   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.931106581035e-06, -1.251695025843e-04, {   2,   1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.299999999806e-04, -2.246238747209e-09, {   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.299999999894e-04, -1.656317460108e-09, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.299999999900e-04, -1.610938899545e-09, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.299999999911e-04, -1.520181778484e-09, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.897021077879e-05, -9.097008595837e-05, {   2,  -1,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.023152874196e-04,  2.742741234372e-05, {   2,   1,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  5.855295716664e-05, -7.112488399047e-05, {   0,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.287610006162e-04,  9.090308755642e-07, {   0,   1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.943126111752e-05,  1.000290575106e-04, {   1,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.480361024289e-05,  1.046113432382e-04, {   1,   0,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.249610079875e-04,  4.453588131644e-06, {   2,   1,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.364771673405e-06, -1.250115365955e-04, {   4,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.939523927461e-05,  9.982281264011e-05, {   1,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.284182635992e-04, -6.090442005071e-07, {   0,   1,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.273307279034e-06,  1.267448671338e-04, {   2,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.854279394878e-05, -8.028155684044e-05, {   2,   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -5.241923134000e-05,  7.632761300457e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  1.221529875561e-04,  6.581387498457e-06, {   2,   1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.179126745374e-04, -1.081951477160e-05, {   1,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  3.902871694235e-05,  8.958794990110e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 {  8.573025112658e-05, -4.282235347366e-05, {   2,  -1,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.076650706294e-04, -2.087739991758e-05, {   2,   1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.157792706499e-04,  1.274694521407e-05, {   0,   1,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  8.869561443101e-05,  3.961361761169e-05, {   1,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  7.000057155325e-05,  5.826074360381e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  8.538362800858e-05,  4.275589694791e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -6.708066011589e-05, -6.060396647134e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -3.928588832562e-05,  8.835859354565e-05, {   3,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.083655789261e-04,  1.889183167430e-05, {   1,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.235012689041e-04, -3.298378692796e-06, {   2,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.046447713274e-04, -2.184369005026e-05, {   1,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.929939026363e-06,  1.234502562641e-04, {   2,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.535071532569e-05, -9.098310726336e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.956572357692e-05, -8.624115824486e-05, {   2,   1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  4.157108986544e-05, -8.395884829659e-05, {   2,   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.080260323439e-04,  1.746697214216e-05, {   4,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.009977465878e-04, -2.432720326825e-05, {   2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.207347715390e-05, -4.300293506272e-05, {   1,   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -1.017479435076e-04, -2.292733756678e-05, {   1,   0,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  9.999283973251e-05, -2.426680282529e-05, {   2,   1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.604470496662e-05,  9.820689544424e-05, {   2,   1,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  2.389101492169e-05,  9.960612550446e-05, {   1,   0,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.325603223355e-05, -9.020792956520e-05, {   2,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.741968076511e-05, -9.594689942213e-05, {   0,   1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  4.049834922753e-06, -1.193121151910e-04, {   0,   1,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  4.982246203286e-06, -1.183417430086e-04, {   2,  -1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.369345420835e-05, -7.956052600248e-05, {   0,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.192528604634e-06,  1.199575248207e-04, {   0,   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.287402344279e-05, -9.959647753395e-05, {   2,  -1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.300921747091e-05, -2.935732120776e-05, {   4,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.264165532089e-05, -6.935731650775e-05, {   4,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  7.769009369826e-05, -4.412512250828e-05, {   2,  -1,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.982799764834e-05, -1.018320960322e-04, {   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.423883853721e-05, -8.739901322223e-05, {   2,   1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -3.204445048350e-05,  8.955069045203e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 } },
 {  4.218825973145e-05, -7.937583370102e-05, {   0,   1,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  8.527606875836e-05, -3.624082500737e-05, {   2,  -1,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.213938376280e-05, -9.919805193291e-05, {   2,   1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.332009275513e-05, -9.799915576253e-05, {   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.911117587321e-05, -3.209772553647e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  4.862354360490e-05,  7.254979102131e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.492695918696e-05, -3.608626746312e-05, {   2,  -1,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.314960317338e-05,  9.782016262028e-05, {   3,  -2,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.953318828342e-05, -3.141437018723e-05, {   2,  -1,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.529262019986e-05, -7.550625142884e-05, {   3,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.284987184652e-05, -1.078304138885e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.773244726560e-05,  4.270581377870e-05, {   3,  -1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  7.773244726560e-05,  4.270581377870e-05, {   1,  -1,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -9.942081731303e-05, -2.082383151269e-05, {   1,   0,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.297110198473e-05, -9.718566465870e-05, {   1,   0,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.149228130989e-04, -5.086411710760e-06, {   4,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.199999999908e-04,  1.487020522661e-09, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.199999999975e-04, -7.749261878880e-10, {   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.281414335741e-05, -3.700547576337e-05, {   1,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  9.894174698225e-05,  2.034917971500e-05, {   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.387548881247e-05, -9.526638525997e-05, {   1,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.524481687477e-05,  5.386177537186e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.070901045385e-05, -7.839137310992e-05, {   0,   1,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.010291179216e-04,  1.785330218690e-05, {   4,  -1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.194966890863e-05, -9.655157875427e-05, {   0,   1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.543298262701e-05,  7.302284430937e-05, {   2,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.445665302626e-05, -9.399157364665e-05, {   0,   1,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.444535874119e-05, -9.394989366416e-05, {   0,   1,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.826786472890e-05,  1.990607426040e-05, {   1,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.146311815076e-04, -3.485062174509e-06, {   0,   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.131620133826e-04, -4.894254165655e-06, {   0,   1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  8.281993178594e-05,  3.522582715810e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.179252717834e-04,  6.054328504059e-08, {   1,  -1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.681642206596e-06, -1.122863932107e-04, {   2,   1,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.130279253405e-04, -4.798213507040e-06, {   0,   1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  7.725064143821e-05,  4.038533563905e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -1.175047070440e-04,  6.024079637130e-08, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -8.452822680175e-05,  3.294226991074e-05, {   2,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.085580709840e-05, -8.660434912078e-05, {   1,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.406900139937e-05, -9.337128302001e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -9.836331556917e-05, -1.896027455057e-05, {   1,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.218286765691e-05, -9.499367616795e-05, {   1,   0,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  9.825532329206e-05, -1.887483440011e-05, {   1,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.308754344210e-06, -1.156866689819e-04, {   0,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.089877934989e-04, -7.901176152221e-06, {   2,  -1,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -2.116315615342e-05,  9.560072664179e-05, {   2,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.164951228599e-04,  3.845501215478e-08, {   2,   1,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  8.600230722835e-05,  3.052580459540e-05, {   2,  -1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.741903202302e-06, -1.117358518319e-04, {   2,  -1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.536000250821e-05,  9.108268768334e-05, {   0,   1,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  2.535157957754e-05,  9.105095949059e-05, {   0,   1,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  9.725088751915e-05, -1.893643059960e-05, {   1,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.092990109393e-04, -6.570492860899e-06, {   0,   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.155178602980e-04, -5.208829438339e-08, {   2,  -3,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.452771232615e-05,  4.088822584334e-05, {   1,  -1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -3.313120223496e-05,  8.215461189209e-05, {   2,   1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.197388644598e-05,  9.330749753411e-05, {   1,   0,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.876858819592e-05, -4.647238130897e-05, {   2,  -1,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  8.482814907973e-05,  3.034905580102e-05, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.077918383753e-04,  7.355243384580e-06, {   1,   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  9.561496297217e-05,  1.932432046483e-05, {   1,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.020744683417e-04, -1.268092721731e-05, {   0,   1,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -3.936968000541e-05, -7.527938026941e-05, {   2,   1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.981563857345e-05,  9.481500765270e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.644565743483e-05,  1.802641129693e-05, {   1,  -2,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.992590445681e-05,  2.441280805474e-05, {   1,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.138744680690e-04,  3.297265676923e-08, {   0,   1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.122662021814e-04, -1.561351986034e-06, {   1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.887909291258e-06,  1.099312701707e-04, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.136629330516e-04, -3.319606121588e-08, {   2,   1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.134692773794e-04, -3.949382139024e-08, {   0,   1,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  4.477578839184e-06, -1.087737931574e-04, {   4,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.224122347095e-05,  4.079195346233e-05, {   0,   1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.703594453942e-05,  8.581759061813e-05, {   2,  -1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.487603706224e-05, -8.790638974112e-05, {   0,   1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  2.560099967762e-05, -8.703165063889e-05, {   4,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.557043810610e-05, -8.694647925496e-05, {   4,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.280486247060e-05,  6.951539837673e-05, {   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.122152671857e-04, -7.990201403200e-08, {   0,   1,   3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.942222560623e-05,  9.276431366850e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  7.436785517039e-05, -3.779266811322e-05, {   2,   1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -2.541191066577e-05,  8.654214562943e-05, {   1,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.538506336509e-05,  8.636221235904e-05, {   1,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -5.259224729030e-05, -5.901207331940e-05, {   2,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.113425806452e-04, -1.523253943108e-07, {   2,   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  2.231202138119e-05,  8.910291804872e-05, {   1,   0,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.039450525490e-05,  6.091082091532e-05, {   0,   1,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -9.842156556609e-05, -1.275564463169e-05, {   3,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.330316726303e-05,  8.784848187569e-05, {   2,   1,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.109040550581e-04, -1.475435854038e-07, {   1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.950952000398e-05, -9.149983187768e-05, {   3,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.457860768594e-05, -6.641281935692e-05, {   0,   1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.088266649771e-04, -2.078667773928e-06, {   4,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.523694555864e-05,  5.562242515429e-05, {   2,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.196963790292e-05,  8.876875318704e-05, {   2,  -1,   1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.103545784812e-04,  2.775803210675e-08, {   2,  -1,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.122648582951e-07, -1.097934211632e-04, {   3,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.099996962495e-04,  2.464751474099e-07, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.742093972913e-05,  2.276834552479e-05, {   3,   1,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -4.236290872992e-05,  6.781339211571e-05, {   2,   1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -3.168843028843e-05,  7.837466059671e-05, {   2,  -1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.099999999901e-04,  1.478293876395e-09, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.095337225247e-04,  4.355352810314e-07, {   2,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  4.114383537062e-05,  6.860892661305e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.418597769444e-05,  3.553327008950e-05, {   3,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.450602914412e-05,  9.513866459278e-05, {   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.095596595614e-06, -1.003384264656e-04, {   2,   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -3.526180917524e-05, -7.412309092604e-05, {   2,   1,  -1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -3.526180917524e-05, -7.412309092604e-05, {   2,  -1,  -1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -8.476743339328e-05,  2.450567769953e-05, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.128697093158e-05,  8.797185432631e-05, {   2,   1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.201899293255e-05,  5.713604178831e-05, {   2,  -1,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  2.288514094222e-05,  8.626364354359e-05, {   2,   1,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  7.264091620497e-05, -3.640921683361e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -8.002217708134e-05, -2.878337344135e-05, {   0,   1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -7.918014459808e-05,  2.958156888000e-05, {   1,   0,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -7.623976133452e-05, -3.239599505238e-05, {   2,   1,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.623976133452e-05, -3.239599505238e-05, {   2,   1,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.945648757720e-06,  9.959907312974e-05, {   3,   0,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.945127258070e-06,  9.959911996768e-05, {   1,   0,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.945057724765e-06, -9.959912621254e-05, {   3,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.081827343264e-04,  2.187010525539e-07, {   2,  -1,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.936978586945e-05, -5.894821445623e-05, {   1,  -1,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -4.936978586945e-05, -5.894821445623e-05, {   1,  -3,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  9.194777558614e-05,  1.636220509965e-05, {   2,   1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.882116499086e-05,  2.945114353127e-05, {   1,  -2,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  2.389502987776e-05, -8.428213328037e-05, {   0,   1,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  9.391365805913e-05,  1.398964713107e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  8.550246639473e-05,  2.227288664796e-05, {   3,  -1,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -4.977790973177e-05, -5.797998119088e-05, {   0,   1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -9.372792237966e-05,  1.393130541815e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.022133127269e-05, -8.742622279226e-05, {   0,   2,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.060892199579e-04, -1.504602215551e-06, {   2,  -1,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -6.765824625219e-05, -3.980425402290e-05, {   2,  -1,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.829443565853e-05, -2.905481201578e-05, {   0,   1,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.740850473634e-05, -9.634441390720e-06, {   2,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.846894821738e-05, -8.856453276600e-05, {   2,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.688583543090e-05,  8.013391967840e-05, {   2,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.066582841116e-05, -8.545176708220e-05, {   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.029082044328e-05,  8.568850720400e-05, {   1,   0,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.305375952341e-05, -6.290630118645e-05, {   2,  -1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -8.067314110035e-05, -2.524417900229e-05, {   2,   1,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.278989681654e-06, -1.024949158833e-04, {   4,   1,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.739246166218e-05,  8.834607929394e-05, {   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.133034311591e-05, -7.432025559980e-05, {   2,  -1,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -3.144370024467e-05,  7.418954820039e-05, {   1,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -8.709352137229e-05,  1.847398053913e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  4.218857773186e-05, -6.333914192717e-05, {   2,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.054332511605e-04,  4.296675605873e-08, {   0,   1,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.051195415740e-04, -3.092077263206e-07, {   2,  -1,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  2.017695182435e-05,  8.510434908393e-05, {   1,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.647103903441e-05, -8.865526861429e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.374404017424e-05, -7.130739295678e-05, {   2,  -1,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -6.449058318568e-05, -4.044315752719e-05, {   2,   1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.048460371829e-04,  2.952129344770e-08, {   4,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.438680149241e-06, -9.938363323193e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  3.512817154348e-05, -6.965435381416e-05, {   2,  -1,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -9.532365690208e-05, -9.219928057832e-06, {   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.332538949942e-05, -6.109533553451e-05, {   1,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.042487002893e-04,  8.835271951639e-08, {   3,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.028693228986e-04,  1.408862152219e-06, {   2,  -1,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -3.208700576225e-05,  7.218042600026e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  8.469107692409e-06, -9.553079310853e-05, {   2,  -1,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.843812199084e-05,  7.551682958678e-05, {   2,  -1,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.028111874789e-04,  1.120108349087e-06, {   2,  -1,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  1.037532623701e-04,  9.485899476947e-08, {   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.004078404834e-05,  8.343179647798e-05, {   3,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.022997827418e-04, -9.444213102337e-07, {   3,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.616970080330e-05,  8.697850041442e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.399921469261e-06, -1.015979075462e-04, {   0,   1,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  8.884630691561e-05,  1.411027352991e-05, {   0,   1,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  5.277176945825e-05,  5.014876185802e-05, {   2,  -1,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.269442927572e-05, -8.010995685028e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  4.023252245428e-05, -6.256389615339e-05, {   4,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.771642109334e-06, -9.885711926533e-05, {   2,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.757953570529e-05,  5.490616090811e-05, {   2,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.389909579372e-05,  2.858147246694e-05, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.992914208719e-05,  8.234993581800e-05, {   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.020906650027e-04, -5.499093898043e-08, {   2,  -1,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  8.905495903517e-05,  1.300823859113e-05, {   1,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.274819698778e-05, -2.927799469809e-05, {   0,   1,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.484641592727e-05, -3.710326718929e-05, {   2,  -1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.991948616068e-05, -6.198495517947e-05, {   4,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.262936282602e-05, -4.918077185063e-05, {   4,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -6.700563845949e-05,  3.479975505917e-05, {   0,   1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  6.709170531459e-05,  3.469887591159e-05, {   0,   1,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  2.249885573057e-05, -7.924609964761e-05, {   2,   1,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  5.606675226294e-05, -4.559378685152e-05, {   4,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.247683152483e-05, -7.918349073989e-05, {   2,   1,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.606107176512e-05,  3.552711534246e-05, {   1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.249973558793e-05,  8.900660218646e-05, {   0,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.667094477878e-05,  3.473444443513e-05, {   0,   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.942291818224e-05, -8.190698609922e-05, {   3,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.010846059677e-04, -2.306111706376e-07, {   2,  -1,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.016851315585e-05,  3.102782687016e-05, {   1,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.349706984499e-06, -9.777920948095e-05, {   4,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.685623894387e-05, -8.425413203987e-05, {   2,   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -8.953931706680e-05, -1.151467234317e-05, {   1,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.008364851031e-04,  1.066372208032e-08, {   2,  -1,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.924558374891e-05, -8.132954519242e-05, {   1,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.294163616461e-05, -5.760674201989e-05, {   4,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.896892802955e-05,  8.135952689235e-05, {   1,   0,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.999984248007e-05,  1.587550445651e-07, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.178884481856e-05,  7.825538881946e-05, {   0,   1,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  2.178076215985e-05,  7.822705449765e-05, {   0,   1,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.921261366496e-05, -8.078879837411e-05, {   1,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.999999999395e-05,  1.099557428734e-09, {   2,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.640383054103e-05, -3.348730066094e-05, {   1,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.265296913046e-05, -7.714801096279e-05, {   2,  -3,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.944373541374e-05, -8.035493930015e-05, {   0,   1,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.330007152273e-05,  2.647840852131e-05, {   2,   1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.264443954817e-05, -7.712866930992e-05, {   2,  -3,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.428183993947e-05,  3.538502660341e-05, {   2,   1,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -6.428183993947e-05,  3.538502660341e-05, {   2,  -1,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -9.857222000431e-05, -7.436807606610e-07, {   4,  -3,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.267285738763e-05,  7.652591013672e-05, {   1,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -8.513408465461e-05,  1.402848448915e-05, {   2,  -1,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  2.523331134628e-05, -7.381565478529e-05, {   4,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.500634472668e-05,  1.398629238281e-05, {   2,  -1,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  4.865158361199e-05,  5.032915071850e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.988361784668e-05,  5.903050482653e-05, {   0,   1,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -7.970396809972e-05,  1.913415442610e-05, {   1,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.204481850281e-05, -6.660816876223e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.092799689701e-05, -7.763865508180e-05, {   1,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.845544573724e-05,  1.999346580601e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  5.928229667435e-05,  3.906428378796e-05, {   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.730104964281e-05, -6.102229210393e-05, {   2,   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -7.999995021437e-06, -9.031817040367e-05, {   2,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.823057682900e-05,  1.149240008285e-08, {   2,  -1,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.550051670641e-05, -3.271437517564e-05, {   2,  -3,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.877191089724e-05, -5.941967474027e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -6.537185910772e-05, -3.265017542660e-05, {   2,  -3,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.002330960050e-06, -9.498469451593e-05, {   0,   1,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.036343948847e-06, -9.489175950217e-05, {   4,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.432239966476e-05, -3.555865256935e-06, {   2,   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.953196444851e-05,  2.832795708254e-05, {   1,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  6.586928940449e-05, -3.197612229868e-05, {   2,   1,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -9.393366777158e-05, -3.829020758194e-06, {   4,   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -9.768917280034e-05,  3.829255733370e-08, {   0,   1,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.050864852123e-06,  8.963918548901e-05, {   3,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.744003418300e-05,  1.637139215465e-07, {   2,  -1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.629733433975e-05, -8.118196122646e-05, {   2,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  5.741192444613e-05,  4.004835741191e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.034358040245e-05, -1.694028537219e-05, {   1,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.246650522360e-05,  3.480540791962e-05, {   2,   1,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -4.596718673936e-06, -9.245476113209e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  9.700443575090e-05,  2.279386103554e-08, {   4,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.052973803007e-05, -5.639489753473e-05, {   1,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -1.353801120743e-05, -8.334542664841e-05, {   1,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.106401016174e-06, -9.577622988334e-05, {   0,   1,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  4.360013514363e-06, -9.229187486166e-05, {   0,   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -3.212450614689e-05,  6.451545488796e-05, {   1,   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  9.631730184749e-05,  3.105534265758e-07, {   2,   1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  9.432966333920e-05, -2.214290485376e-06, {   4,   1,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  3.902820444022e-05, -5.730241315839e-05, {   1,   1,   0,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  3.902820444022e-05, -5.730241315839e-05, {   1,  -1,   0,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  1.850482466272e-05, -7.757698628959e-05, {   1,   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  9.594024105469e-05,  5.732394202570e-08, {   4,   1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.084908932621e-05,  5.511520513098e-05, {   2,   1,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -9.639450693641e-08, -9.578361418264e-05, {   4,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.356671329974e-05,  1.215543317710e-05, {   4,  -1,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -9.846922463894e-06,  8.582350418613e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.818164289983e-05,  7.748639299295e-05, {   1,   0,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.198901060894e-05,  3.367790045739e-05, {   2,  -1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -9.518452028160e-05,  3.989144295126e-07, {   0,   1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  3.149085229833e-05, -6.400116940394e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,  -3,   0,   0 } },
 {  1.716139628079e-05, -7.831429255539e-05, {   2,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  3.327693377717e-05, -6.218849893526e-05, {   0,   1,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -9.225113227934e-05,  2.929447299539e-06, {   2,  -3,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.053613328240e-05,  6.456270754634e-05, {   2,  -1,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 { -1.826514113889e-05,  7.674626153170e-05, {   3,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.829696608451e-05, -6.631234168020e-05, {   1,   0,   0,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 {  5.893135285973e-06, -8.830109325565e-05, {   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  8.990098649673e-05,  4.220500789566e-06, {   2,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.216516405091e-05,  5.165650628092e-05, {   2,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -4.234204602770e-05, -5.140654688109e-05, {   0,   1,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.790634018735e-05,  7.569006181834e-05, {   1,   0,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.081464273629e-05, -7.268712608147e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -9.328787918846e-05,  4.050832905993e-08, {   4,   1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.665735743763e-06, -8.854624487070e-05, {   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.208389987142e-05, -3.100795515478e-05, {   1,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.652486594047e-05, -5.656516738313e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.005456177998e-05,  5.293472418148e-05, {   2,   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.794786941005e-05, -7.502879008347e-05, {   3,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.905329278003e-05, -6.390360193819e-05, {   0,   1,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  2.904362083103e-05, -6.390069052084e-05, {   0,   1,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -5.815881795466e-05, -3.438935223956e-05, {   1,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.649201420400e-06,  8.389809017188e-05, {   0,   1,   1,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  8.649201420400e-06,  8.389809017188e-05, {   0,   1,  -1,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 { -5.659758274782e-05, -3.554541339459e-05, {   2,  -1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -9.107164408187e-05,  7.916508723381e-07, {   2,   1,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.032743345918e-05,  6.152886007134e-05, {   2,  -1,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -7.718286904638e-05, -1.460733190550e-05, {   2,  -1,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  3.632521515623e-05,  5.538741322717e-05, {   1,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.200857326336e-05,  2.968852883021e-05, {   2,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.158973564747e-05, -9.011818162248e-08, {   2,  -1,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.001322310021e-05, -3.155333184237e-05, {   2,  -1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.036589554337e-05, -1.036547396799e-06, {   0,   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  9.126611572741e-05, -1.052013927892e-07, {   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.107709722930e-05,  1.413341860090e-07, {   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.075697543462e-05, -3.034560992557e-05, {   1,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.755320351593e-05, -5.341233058890e-05, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -3.181108436178e-05, -5.914924574480e-05, {   0,   1,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -3.356838049245e-05, -5.728895544058e-05, {   0,   1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -5.461519071789e-05,  3.621568449687e-05, {   1,   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  4.124949556752e-05, -4.948254819859e-05, {   1,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 { -4.531049502210e-05, -4.538218525120e-05, {   2,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.790797194338e-06, -8.790137180071e-05, {   4,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  8.135295101629e-05, -9.269225953153e-06, {   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.733363518128e-05,  7.322171602589e-05, {   1,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.146415374283e-05,  1.902487865986e-05, {   1,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.498961955388e-05, -5.539148064990e-05, {   2,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.515532605780e-05,  5.118955184034e-06, {   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.999989631219e-05,  1.795384885782e-07, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.027352187895e-05, -9.786433923359e-06, {   0,   1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.475886940192e-05, -5.525647279786e-05, {   2,   1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -8.999999998763e-05,  1.492256510387e-09, {   4,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.999999999249e-05, -1.162389281808e-09, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  8.999999999328e-05, -1.099557428721e-09, {   4,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.999999999385e-05, -1.052433538950e-09, {   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.999999999385e-05,  1.052433538929e-09, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.999999999539e-05, -9.110618695468e-10, {   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.999999999600e-05,  8.482300164567e-10, {   4,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.915601582165e-05,  1.066898729706e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.627030603355e-05, -6.354332189351e-05, {   1,  -2,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  7.312571206822e-05, -1.668441882218e-05, {   4,  -1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.267534984196e-07, -8.897720478343e-05, {   0,   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -4.269178631736e-06, -8.542122983232e-05, {   2,   1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  7.296507185697e-06, -8.227360593163e-05, {   2,  -1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  4.094484739614e-05, -4.862083405125e-05, {   2,   1,  -1,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 {  4.094484739614e-05, -4.862083405125e-05, {   2,  -1,  -1,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 {  1.883820323803e-05, -7.063627952298e-05, {   0,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.156128528460e-05, -7.872884852044e-06, {   1,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.378115783022e-06,  8.404570843918e-05, {   2,  -1,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -2.033808057438e-05, -6.868741976095e-05, {   0,   1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.680439371121e-05, -5.213328910815e-05, {   1,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.142759407777e-05,  1.750111578629e-05, {   2,  -1,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -7.165238299759e-07, -8.812917223028e-05, {   0,   1,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -5.921280225647e-05,  2.957395600002e-05, {   4,  -1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.317138799272e-05, -7.560582356427e-05, {   1,  -2,   0,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 {  8.748349052043e-05, -1.280118639488e-06, {   1,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.425052834565e-05, -5.445348701396e-05, {   2,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.123141030780e-05,  6.744419761389e-05, {   2,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.121481223897e-05, -5.743366263870e-05, {   2,  -1,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  3.555741610692e-06, -8.499857711647e-05, {   2,  -1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.156035643976e-06, -8.430290304098e-05, {   0,   1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.612376114388e-05,  7.221548219230e-05, {   2,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.970927491035e-05, -5.859251408892e-05, {   0,   1,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -2.970286952021e-05, -5.859217509359e-05, {   0,   1,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  5.924404503795e-05, -2.903076454701e-05, {   2,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -8.817807633042e-05, -4.421883308272e-08, {   4,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.891543033643e-05, -5.911091729492e-05, {   0,   1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  2.244371073562e-06, -8.575262349683e-05, {   0,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.726606478234e-05, -6.644639842903e-07, {   4,  -1,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.769833332442e-05,  1.456391242947e-07, {   2,   1,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.337109095318e-05,  3.444980720927e-05, {   0,   1,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.719131500168e-05, -5.058131165244e-05, {   1,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.132343790195e-05, -3.644592457806e-05, {   0,   1,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  7.454889463448e-05,  1.320418834728e-05, {   2,  -1,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.748910950360e-05, -1.025654916780e-05, {   1,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.129674657308e-05, -3.644674652916e-05, {   0,   1,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  5.345486509265e-07,  8.706496149497e-05, {   1,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.029630130255e-05, -2.730051707196e-05, {   1,   0,   0,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -8.680080456867e-05,  7.798365890447e-07, {   2,  -1,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.750350731290e-05,  4.064470993459e-08, {   0,   1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.671536757275e-05, -7.076348221761e-05, {   1,   0,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.809802212857e-05, -5.919667649508e-05, {   2,  -1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  8.711969165383e-05, -1.191735387865e-08, {   2,   1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.336735575801e-06, -8.377902130557e-05, {   2,   1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.394664774850e-05,  5.299133938579e-05, {   2,  -1,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  7.156435566248e-06,  7.967926599804e-05, {   1,   0,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.156240873051e-06, -7.967928348427e-05, {   5,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.156185246416e-06, -7.967928848025e-05, {   3,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.155698513207e-06, -7.967933219341e-05, {   1,  -2,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.790415767316e-05,  2.892244463151e-05, {   4,  -1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.640403860062e-05,  3.041608368941e-05, {   2,  -1,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -4.549993858341e-05,  4.125873339719e-05, {   2,  -1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -7.991117434685e-05, -6.818252208547e-06, {   1,   0,   0,   0,   0,   0,  -8,  17,   0,   0,   0,   0,   0 } },
 {  8.647241543573e-05, -1.122299287298e-07, {   0,   1,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  3.260600867072e-05, -5.393584412650e-05, {   2,  -1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.149633650627e-05,  7.486205713122e-05, {   2,   1,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -3.233308965761e-05,  5.390552284859e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  7.206390087912e-05, -1.411811252474e-05, {   4,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.893350300734e-05,  6.713426431126e-05, {   0,   1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  7.388672650592e-05, -1.207775704173e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.037744822465e-05,  7.547154196392e-05, {   2,  -1,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -4.798383144184e-05,  3.781845704387e-05, {   2,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -4.468136668385e-05,  4.109652883556e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 {  8.071683680748e-05, -5.054332562413e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -7.098785306216e-05,  1.451757383560e-05, {   1,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.638593491352e-05, -6.906873988375e-05, {   2,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.708706453448e-05,  2.835164341522e-05, {   3,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.579010949929e-05,  1.963923349747e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  6.255266096086e-05,  2.267803297916e-05, {   2,  -1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.401852456822e-05, -1.210419190618e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  8.492515515369e-05,  1.425715139505e-07, {   2,  -1,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.424217008753e-05,  7.064254113804e-05, {   2,  -1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  2.865233387477e-06,  8.194828699687e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.782410118887e-05, -5.695466624883e-05, {   2,   1,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  3.773962883992e-05, -4.703285332358e-05, {   1,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -8.196292792142e-05, -2.787140594083e-06, {   0,   1,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.454546291758e-05,  1.546366104898e-07, {   2,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.701059276441e-05, -6.745433091196e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -6.977217598199e-05, -1.461056401404e-05, {   1,   0,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.531535842609e-06,  7.976039129402e-05, {   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.962322226739e-05, -5.465863281656e-05, {   0,   1,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -2.843555950366e-05, -5.579303040346e-05, {   1,   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -8.384297882402e-05, -3.698812536495e-07, {   4,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.590413485930e-05, -6.824313991501e-05, {   1,   0,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.609608222422e-05, -6.801980884199e-05, {   1,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.485434841726e-05, -9.251294286552e-06, {   2,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  7.325027775574e-07, -8.336127257219e-05, {   3,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.402351590560e-05,  5.313813647673e-09, {   2,   1,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.883591878574e-05, -1.497931119855e-05, {   1,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.289616276222e-05,  5.074775773831e-05, {   0,   1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -8.360193691562e-05, -3.519267553018e-08, {   2,   1,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  5.626940460361e-05, -2.727851076612e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  4.921022095383e-05,  3.432716349592e-05, {   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.597795708000e-05,  6.747150819859e-05, {   1,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.227754285987e-05,  7.109507108173e-05, {   1,   1,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.403892896818e-05, -9.244463735103e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  1.225438971470e-05,  7.099718530277e-05, {   1,  -1,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.485609823832e-05,  6.837607136083e-05, {   1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.083919200563e-05, -2.213898778093e-06, {   4,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -8.297500323637e-05, -6.264468892780e-08, {   2,  -1,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.277718441595e-05, -1.143941633662e-07, {   4,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.362549321930e-06, -7.652124739081e-05, {   2,   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -3.340219303763e-05,  4.942871348750e-05, {   0,   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  2.708808430466e-05, -5.573688348680e-05, {   0,   1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  4.788447651056e-05, -3.490357135548e-05, {   0,   1,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -4.791118488641e-05, -3.486343568925e-05, {   0,   1,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  8.273911633751e-05, -3.605472368620e-09, {   2,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.903637230235e-05, -5.349739219612e-05, {   0,   1,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  8.085063402621e-05, -1.492402788339e-06, {   0,   1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.572026072523e-06, -7.675826709602e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  8.182791617183e-05,  2.934554786313e-07, {   2,  -1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -6.350736178774e-05,  1.844758215701e-05, {   1,  -2,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.625180310468e-06, -8.032308518299e-05, {   0,   1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.949494891602e-05, -2.025672950807e-06, {   2,   1,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  5.433631053208e-05, -2.714241026174e-05, {   2,   1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.538134921697e-05, -6.603340622882e-05, {   3,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.061990528341e-05,  1.075900921082e-05, {   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.104861531110e-05, -1.023518891283e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -3.666436606358e-05, -4.451336559926e-05, {   2,  -1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.635261223235e-05,  1.482498553977e-05, {   4,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.805886163466e-05,  6.275248182354e-05, {   2,  -1,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.435317056677e-05,  4.644667592471e-05, {   2,  -1,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  9.179926393316e-06, -7.156311338919e-05, {   4,  -1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -3.654893328660e-05, -4.416452655977e-05, {   1,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  1.511195798680e-05,  6.556460935039e-05, {   1,   2,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.028908726634e-05, -5.034216364021e-05, {   1,  -2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.908185858750e-05,  1.513847279673e-06, {   2,   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.221426813165e-05, -5.822396836529e-05, {   1,   0,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -1.527411586624e-05, -6.509996930550e-05, {   3,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.033598064365e-05, -2.554332385855e-08, {   0,   1,   3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.999999998499e-05,  1.549852375674e-09, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.999999998682e-05,  1.452113937580e-09, {   2,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.999999998830e-05, -1.368338133495e-09, {   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.999999999259e-05,  1.089085453211e-09, {   2,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.999999999296e-05,  1.061160185181e-09, {   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.999999999386e-05,  9.913470151074e-10, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.999999999941e-05,  3.071779483502e-10, {   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.517158862736e-05,  6.476599350501e-05, {   1,   0,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.081234658720e-05, -3.912451370500e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -7.799010344982e-05,  1.914792947683e-06, {   0,   1,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.983773916839e-05,  4.402404657332e-08, {   2,  -1,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.162952283676e-05, -2.820920594681e-05, {   1,   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  7.813275095577e-05, -1.648234795714e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -7.967700663316e-05,  3.655339851729e-08, {   2,   1,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.543540299835e-05,  6.413797345895e-05, {   1,  -2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.003485265429e-05, -2.953773458523e-05, {   0,   1,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -5.476946907928e-05,  2.450104563360e-05, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.546506287016e-05, -3.350004468973e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -3.113371484481e-05,  4.782060818401e-05, {   0,   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  7.868971536468e-05,  3.783107765547e-08, {   0,   1,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.504057180169e-05,  6.359725142392e-05, {   3,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.015514747085e-05, -5.844548563667e-05, {   1,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.500297333463e-05, -6.356120481117e-05, {   1,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.935789924137e-05, -4.916971595856e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -2.584683892617e-05, -5.267419391902e-05, {   2,   1,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -6.776920621123e-05, -1.069223794263e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -1.492737250739e-05,  6.351967992269e-05, {   3,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.638474983603e-06, -7.380507299388e-05, {   2,   1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  7.837577362594e-05, -5.402336033770e-08, {   2,  -1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.820525975987e-05,  2.020274931041e-05, {   2,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  1.613196782267e-05,  6.217755724505e-05, {   2,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.831199875192e-05,  3.997513518951e-05, {   2,  -1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  3.606404175488e-06, -7.452041674118e-05, {   2,   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  6.620407107888e-05,  1.177026452010e-05, {   4,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.455370816315e-05,  6.329564671965e-05, {   1,   0,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.680996108285e-05,  3.102975622136e-05, {   2,  -1,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -7.770844413126e-05, -9.754696157146e-08, {   4,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.216427626367e-05, -1.554747115429e-05, {   1,  -2,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.024313880108e-05,  4.737390107591e-05, {   2,  -1,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  3.011033727652e-05, -4.739862228712e-05, {   4,   1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.421223833122e-06, -7.293487893715e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  4.656140748293e-05,  3.078457360847e-05, {   2,  -1,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -6.956969080166e-05,  7.749717527854e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.601198305916e-07, -7.703943641976e-05, {   4,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.476240268161e-05,  5.234788743783e-05, {   0,   1,   0,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -1.500629096389e-05,  6.201357540386e-05, {   0,   1,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.569714744364e-05,  1.182178746722e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.532508355117e-05, -5.151580754475e-05, {   2,  -1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -6.256195522790e-07, -7.620239001057e-05, {   1,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.983829612296e-05, -4.694823296894e-05, {   4,   1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.878934994486e-05, -7.967904078328e-06, {   1,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.715000609322e-05,  2.957926062173e-05, {   2,  -1,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.957578819610e-05, -5.710149317415e-05, {   4,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.097371066171e-05,  6.568097564373e-05, {   3,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  6.315749340783e-05, -1.339859264262e-05, {   2,  -1,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.698328572839e-05,  2.946492388688e-05, {   2,  -1,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  7.622539455131e-05,  8.415045001043e-08, {   2,  -3,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.917889790656e-05,  2.701868554671e-05, {   3,   1,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  4.917889790656e-05,  2.701868554671e-05, {   1,   1,  -1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  5.563824818806e-05,  2.042425769336e-05, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.264667661535e-06, -6.971910741748e-05, {   3,  -2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.578939916029e-05,  1.922864592978e-07, {   2,  -1,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.261893288790e-06,  6.971935665538e-05, {   1,  -2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.081675882577e-05, -5.141344979753e-06, {   2,   1,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.990238860711e-06, -7.196780605067e-05, {   2,  -1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.610155840665e-05,  5.977134777847e-05, {   3,  -2,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  7.535930008839e-05,  3.214609405950e-07, {   1,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.955543368349e-05,  4.605863671398e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -2.015654490580e-05, -5.534435049798e-05, {   2,  -1,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -3.388604748400e-05, -4.151538822949e-05, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.222990984898e-05, -1.296512003034e-05, {   2,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.458170042966e-05,  4.056477474384e-05, {   1,   1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  7.350834879996e-05,  1.585994536788e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -1.438673932184e-05,  6.063011320540e-05, {   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.351235865958e-05,  1.145835749713e-05, {   0,   1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.167731674945e-05, -3.321218362693e-05, {   4,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.723748390524e-05, -4.756035578276e-05, {   2,   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.796450305679e-06,  7.290619041401e-05, {   2,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.508177815848e-05, -3.954741905264e-05, {   4,  -1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.508177815848e-05, -3.954741905264e-05, {   0,   1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.154058766293e-05, -1.308708969813e-05, {   1,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.215923642536e-05, -4.235028628322e-05, {   1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.723931672926e-05, -2.724704707973e-05, {   0,   1,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  5.037962479943e-05, -2.406614445613e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  3.025111588187e-05,  4.413915083942e-05, {   1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.889646145011e-06,  7.124165188486e-05, {   2,   1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.159890729502e-05, -4.251843504015e-05, {   1,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.752286031931e-05, -2.642968205808e-05, {   2,   1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.389454752117e-05,  5.975757855471e-05, {   1,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.180574500323e-05,  5.181350572212e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -4.902600825924e-05, -2.448633699363e-05, {   2,   1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.358293551555e-05, -5.988381100927e-05, {   0,   1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  6.249898686848e-05,  1.095099204563e-05, {   0,   1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.598305442680e-05, -5.740133951258e-05, {   2,  -1,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.597771019636e-05, -5.738662471201e-05, {   2,  -1,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -2.197532321495e-05, -5.136623571523e-05, {   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.406656146796e-05, -4.893806469279e-05, {   2,  -1,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  2.333822768278e-05,  4.945897129432e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.470177512124e-05,  5.801670165132e-05, {   1,   0,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.702581934949e-05, -2.569165749051e-05, {   2,  -1,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  5.995733899979e-05,  1.259599559668e-05, {   1,   1,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  5.995733899979e-05,  1.259599559668e-05, {   1,  -1,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  4.629949690411e-05, -2.589769905894e-05, {   1,   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.259636405512e-06, -7.092784734056e-05, {   3,  -1,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.757073414871e-05,  4.451592851505e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0 } },
 { -4.918994000753e-06,  6.703273651384e-05, {   2,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.757336847732e-05, -4.431707117227e-05, {   2,  -1,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  5.898228919637e-05,  1.265739970964e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -6.246632446012e-05, -9.010789425699e-06, {   3,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.052183244856e-06, -6.741334182654e-05, {   2,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.744498769203e-05, -4.391366516532e-05, {   4,  -1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.786415418806e-05, -3.345122763810e-05, {   4,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.644143406734e-05,  1.481207281892e-05, {   1,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.239400585876e-05, -3.883824745656e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -5.774541639287e-05,  1.347974273259e-05, {   2,  -1,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  5.773302676342e-05,  1.347344593062e-05, {   2,  -1,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  2.339722324560e-06, -6.885767457505e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  6.011095040168e-05, -1.099622846931e-05, {   2,  -1,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -7.020630015389e-05,  6.607738013569e-07, {   2,   1,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -7.081685581718e-05, -1.567721565674e-08, {   0,   1,   3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.659837551063e-06, -6.816718251099e-05, {   2,  -1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.781386350083e-05, -2.948428467778e-06, {   0,   1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  9.773783373884e-07,  6.977313450997e-05, {   2,   1,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.253876122850e-06, -6.945128201170e-05, {   3,  -3,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.494081491226e-05, -1.571735066199e-05, {   4,  -1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.964369088827e-05,  3.097354006070e-05, {   2,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -3.143290876139e-05,  3.910653815807e-05, {   1,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  3.747808541789e-05, -3.302419384892e-05, {   2,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.884478463021e-05, -2.163049392526e-05, {   2,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.345795942631e-05, -5.689604669423e-05, {   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.624647126506e-05,  3.857476443295e-06, {   1,   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  5.213938565569e-05,  1.794244727899e-05, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  1.749123148590e-06,  6.832520921634e-05, {   2,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.510965664644e-05,  5.492030264951e-05, {   0,   1,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.999999997940e-05, -1.698205362007e-09, {   2,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -6.999999998028e-05,  1.661553447743e-09, {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.999999999266e-05, -1.014036295383e-09, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  6.999999999318e-05, -9.773843810828e-10, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.999999999368e-05,  9.407324667966e-10, {   2,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.999999999577e-05, -7.696902001152e-10, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.877975730543e-05,  1.116918814637e-05, {   1,   0,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.936169497052e-05,  1.055885780186e-05, {   3,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.968188593376e-05,  2.277134748589e-07, {   2,  -1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.777095666083e-05,  2.083343141805e-06, {   2,  -1,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.969932454728e-05, -3.013787776993e-05, {   2,  -1,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -4.280087090078e-05, -2.703608871732e-05, {   2,  -1,   0,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  2.296391977485e-05, -4.680355599543e-05, {   2,  -1,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.894230747642e-05,  5.075709215394e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.629364995101e-05,  5.336659928839e-05, {   1,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.546539042825e-05, -4.418835605060e-05, {   2,   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  5.911945398761e-05,  1.046420814512e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.140841401956e-05,  4.815014599781e-05, {   1,   0,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.982662525797e-05,  2.970779637066e-05, {   3,   0,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -3.825392739387e-05,  3.122483471878e-05, {   2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  4.238456173327e-05, -2.707348589883e-05, {   0,   1,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -1.726149502742e-05, -5.218537541524e-05, {   2,   1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.137208522677e-05,  4.795941884226e-05, {   2,  -1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.894258953589e-05,  3.806197647143e-07, {   2,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.591483980650e-05,  5.327824716515e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  1.591398521034e-05,  5.327367629686e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -4.060465056168e-06, -6.511272591662e-05, {   2,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.720654524798e-05,  1.172613517583e-05, {   0,   1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.844342682297e-05, -4.045396288473e-05, {   2,   1,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.343584114934e-05, -5.541184382223e-05, {   1,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.488389408610e-05, -5.396343931375e-05, {   1,   0,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  4.348025467824e-05, -2.521918729983e-05, {   0,   1,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.058646510546e-06, -6.663413317066e-05, {   4,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.845336103127e-05,  7.481781906974e-08, {   0,   1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.359587143573e-05, -5.491124230254e-05, {   1,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.833519149212e-05, -5.405187225379e-08, {   4,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.624528091588e-05, -4.214288395123e-05, {   4,  -1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -4.349201854777e-05,  2.487689114841e-05, {   2,  -1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -5.062696348388e-06,  6.328462285731e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -6.829880390671e-05,  3.829465902454e-08, {   0,   1,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.330109099908e-05,  5.500702422576e-05, {   1,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  9.101890510823e-06, -5.918512288594e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.575472287379e-05,  5.250174793572e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  4.358056773653e-05,  2.467588228292e-05, {   4,  -1,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.575384241324e-05,  5.249702965697e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  5.107112411128e-06,  6.302202177838e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -1.533730136722e-05, -5.265189595312e-05, {   2,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.778208006382e-05, -1.604275153661e-07, {   0,   1,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.297352441831e-05,  5.488955084246e-05, {   1,   0,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.645986266467e-05,  2.137685249185e-05, {   1,   2,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.371402557856e-05, -2.407376858933e-05, {   0,   1,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.284063171459e-05,  5.491042445800e-05, {   1,   0,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.767964153354e-05,  4.503227344524e-08, {   0,   1,   3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.549920968557e-05,  1.222496855935e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.368407498543e-05,  2.399981560015e-05, {   3,   1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  4.368407498543e-05,  2.399981560015e-05, {   1,   1,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -1.319114598324e-05,  5.434211373731e-05, {   2,  -1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.719919902436e-05, -3.173513166276e-07, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.228716505286e-05, -5.519205829733e-05, {   4,  -1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.715688599888e-05, -3.218310174416e-07, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  6.637202409797e-05,  9.895317177417e-07, {   0,   1,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.572256234653e-05, -4.158996653618e-05, {   2,  -1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  4.324106221483e-05, -2.406161907049e-05, {   0,   1,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  7.749546366482e-06, -5.949350519876e-05, {   4,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -5.221421554423e-05, -1.501779904039e-05, {   1,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.293400673686e-05,  5.422920265561e-05, {   1,   2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.103507804534e-05, -2.612340964990e-05, {   0,   1,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.258060421035e-05,  5.452873744763e-05, {   1,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.497282495377e-05,  2.186435388672e-05, {   2,  -1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.302161794564e-05, -5.381412181001e-05, {   1,   1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.929382596082e-06,  5.784030595755e-05, {   2,  -1,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  4.312152756083e-05,  2.341169717555e-05, {   2,   1,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -5.673426707019e-05, -9.787175220389e-06, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.815328904395e-05, -3.826527606798e-05, {   4,   1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.211574519395e-06, -6.418081714439e-05, {   2,   1,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.276713973570e-05,  5.359220842235e-05, {   1,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.884642203323e-05, -4.750708962004e-05, {   1,   0,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.662495914595e-06, -6.367429245421e-05, {   4,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.865768976754e-05, -2.766103691498e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -4.263279462167e-05,  2.358380559534e-05, {   4,  -1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  2.064016898380e-05, -4.556707317653e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -6.540987315712e-05,  7.201869569826e-07, {   2,  -1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  6.603128190483e-05, -3.006950106486e-08, {   2,   1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.517919068025e-05,  8.803609187315e-07, {   2,   1,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.306741269130e-05, -4.282216927711e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.434237459679e-05, -5.151215602475e-05, {   2,   1,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.434217589669e-05, -5.151085260657e-05, {   2,   1,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  5.261451385359e-05, -1.316104982564e-05, {   3,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.968789074174e-05,  3.604453599100e-05, {   4,   1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.262135570718e-05,  5.302914341349e-05, {   3,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.304121714180e-05, -2.547696668469e-06, {   2,   1,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.255596414075e-05,  5.301414473191e-05, {   1,  -2,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.516163837104e-05,  2.407582782511e-07, {   2,   1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.463389368888e-05,  6.387437059111e-07, {   1,  -1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  7.476249663887e-06,  5.774432192056e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -3.553096362247e-06, -6.162484686180e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 {  5.370476524670e-06,  5.975916650940e-05, {   1,   2,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.367326674686e-06,  5.975944949853e-05, {   3,   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.367326674686e-06, -5.975944949853e-05, {   2,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.367326674686e-06, -5.975944949853e-05, {   1,   0,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.367326674686e-06, -5.975944949853e-05, {   1,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.367326674686e-06,  5.975944949853e-05, {   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.367201514776e-06, -5.975946073970e-05, {   1,   0,   2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.269690810967e-05,  2.241400533534e-05, {   0,   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -5.313146347819e-05,  1.186919289116e-05, {   4,   1,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.810343297056e-05, -2.677114927618e-05, {   0,   1,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  6.309212492636e-05, -1.764359796502e-06, {   2,   1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.199875888195e-05, -2.284742828230e-05, {   2,   1,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  3.002513093539e-07,  6.452339647031e-05, {   2,   1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.341314582451e-05, -3.139609945211e-05, {   2,   1,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.237643458176e-05, -4.241687715901e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -7.430895835056e-06, -5.725562205780e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  3.686706097044e-05,  2.772926019375e-05, {   1,   0,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -6.453379103545e-05, -1.733306873638e-08, {   1,   1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.673767264344e-05,  1.776485169864e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.674648175741e-05,  1.774165841472e-05, {   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.010098778062e-05,  5.434551829452e-05, {   1,  -2,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.228447903176e-05, -5.205008017548e-05, {   1,   0,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.704225989919e-05,  2.726454349045e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.963545346397e-07,  6.395404787955e-05, {   2,  -1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.458801276557e-05, -9.637471744491e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -2.405546376932e-05, -4.010261278356e-05, {   1,   0,   2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.571194398828e-06, -6.141222618576e-05, {   2,  -1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.259091366804e-05, -5.137562627403e-05, {   1,   0,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.301316595342e-05, -4.087864171824e-05, {   0,   1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.105267367712e-05, -4.279237362778e-05, {   2,   1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -5.494024217093e-05,  8.891305407530e-06, {   0,   1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.855996025919e-05,  3.526707457214e-05, {   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.633521455012e-07,  6.335038767381e-05, {   1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.377905097200e-05,  2.469234733944e-08, {   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.365101651799e-05, -1.007572740790e-05, {   0,   1,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.104177884618e-05, -2.268010831319e-05, {   2,   1,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  3.971018740882e-05,  2.400636977089e-05, {   0,   1,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -6.203343031930e-05,  1.681601875204e-06, {   4,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.317919830208e-05, -4.052692581287e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.702077481746e-05,  2.667131049556e-05, {   2,   1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  4.593782974146e-05, -1.772885719047e-05, {   0,   1,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.213337146519e-05, -5.148835817573e-05, {   3,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.057016092060e-06,  6.151710844733e-05, {   2,  -1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -6.303623774069e-05, -4.454400714451e-07, {   1,   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -4.580638278723e-05, -1.759721744343e-05, {   0,   1,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.398116231501e-06, -6.094050887878e-05, {   4,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.231413778302e-05, -5.101745994373e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.269046137263e-05,  6.340533216620e-07, {   2,   1,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.612969639461e-05,  3.712869947539e-05, {   3,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  5.479157539151e-05,  8.447854572732e-06, {   1,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.765989482246e-05, -3.556532894907e-05, {   1,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.097898805959e-05, -1.223531094654e-05, {   1,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -5.216799468760e-05,  1.095385147840e-05, {   2,  -1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.382466671040e-06,  5.965681968799e-05, {   2,  -1,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  4.881140595150e-05,  1.415149655184e-05, {   1,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  5.344812737355e-05, -9.463941364178e-06, {   1,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.916367493615e-05, -4.371023138446e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -6.065206691796e-05, -2.188178307949e-06, {   4,   1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.605801588330e-05,  4.676645141162e-05, {   2,   1,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.870001140535e-06, -5.886672878172e-05, {   2,  -1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.255028543600e-05, -5.017332635783e-05, {   1,   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -6.091383284942e-06,  5.651776119310e-05, {   0,   1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  3.387080014751e-05, -2.864473579066e-05, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.387089513670e-05, -2.864452398690e-05, {   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.661771955712e-05,  3.588978145075e-05, {   1,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.043678135928e-05, -4.206625016667e-05, {   2,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.294346529649e-05,  9.536507993260e-06, {   1,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.078803483632e-05, -2.168748514278e-05, {   2,   1,  -1,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.078803483632e-05, -2.168748514278e-05, {   2,  -1,  -1,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.105099415221e-05, -4.141891055292e-05, {   0,   1,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -6.229474277341e-05,  1.372323038990e-07, {   2,  -1,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.968797422402e-06, -5.943985578321e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -5.171247959491e-05,  1.067912559437e-05, {   1,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.260836346251e-05,  1.967809405065e-05, {   0,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.538197489464e-05, -1.689615386869e-05, {   1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 {  1.199680943684e-05, -5.027712314876e-05, {   2,   1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.084130617475e-05,  2.138891057082e-05, {   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  3.648565714740e-05, -2.573205239439e-05, {   0,   1,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  6.197071904289e-05,  2.191381245425e-07, {   2,  -1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.086578768836e-05,  5.113542778992e-05, {   2,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.210561887305e-05, -4.981841664770e-05, {   0,   1,   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -3.871808100095e-05, -2.318127581736e-05, {   2,  -1,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.992168287325e-05, -2.195338662959e-05, {   0,   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  4.135139005777e-05, -2.052240510814e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 { -4.423958539743e-05, -1.762762710234e-05, {   2,   1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.083379835333e-05, -1.086601176461e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  5.518094697774e-05, -6.490699484101e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -3.764637860730e-05,  2.402028733059e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.197459868726e-05,  4.969033161139e-05, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.961393996316e-05, -2.199856876395e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -2.031783625670e-05, -4.125233790721e-05, {   2,   1,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  4.851541880862e-05,  1.302314293533e-05, {   2,   1,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.136543064546e-05, -5.016730267681e-05, {   0,   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.474210077706e-05, -4.676057858821e-05, {   2,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.101628634543e-05,  2.048565853531e-05, {   0,   1,   5,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.467087741162e-06,  5.903028297950e-05, {   0,   1,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.826146045763e-05,  2.322429921188e-05, {   0,   1,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.766484803809e-05, -2.374936586247e-05, {   0,   1,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.119627211336e-05,  1.098756157116e-07, {   0,   1,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.805825265653e-06, -5.939574745380e-05, {   2,   1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.835702754137e-05,  2.843472319358e-06, {   4,  -1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.582028961214e-05, -2.537216971221e-05, {   1,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -1.160446715516e-05, -4.948113044472e-05, {   1,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.336847415550e-05, -3.746977504072e-05, {   4,   1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.745467751260e-05,  3.333198232539e-05, {   4,  -1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.911433582361e-05,  1.615647182899e-06, {   4,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.771373810886e-05,  3.294065061586e-05, {   2,  -1,   0,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 {  1.816398738128e-05,  4.245765551381e-05, {   1,  -2,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  5.857673520057e-05, -2.014002938615e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 {  5.568019453999e-05, -4.878378845998e-06, {   4,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.958523360243e-05, -1.095543483515e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  4.958550521646e-05, -1.095366983351e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  6.040047923086e-05, -1.355769862248e-07, {   0,   1,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.143063164398e-05, -4.902666297325e-05, {   1,   0,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.705077958679e-05, -2.335084821540e-05, {   0,   1,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.990077350170e-05,  4.881782897066e-07, {   2,  -1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -2.204016681681e-05,  3.814996838350e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -4.972225362034e-05, -1.042768562252e-05, {   1,   0,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.952936183816e-05, -1.062009024705e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -5.999968394548e-05,  1.124370087978e-07, {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.280570085482e-06, -5.781630661222e-05, {   2,   1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.999992711126e-05,  7.636159905072e-08, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.960313014393e-05,  4.613812271706e-07, {   2,  -3,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.999999999175e-05, -9.948376735796e-10, {   2,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.999999999340e-05,  8.901179184845e-10, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -5.999999999355e-05,  8.796459429736e-10, {   2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  5.999999999513e-05, -7.644542123574e-10, {   4,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.999999999577e-05, -7.120943347849e-10, {   4,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.999999999772e-05, -5.235987755793e-10, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -5.994484990968e-05, -3.247399444159e-08, {   2,  -1,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.753669315763e-05,  2.243380386013e-05, {   1,   0,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.429854634276e-05,  2.554129482154e-05, {   1,  -2,   2,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  2.483116348561e-05,  3.498685784636e-05, {   1,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.063022759095e-05, -9.174398046001e-06, {   4,  -1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.249193533662e-05, -3.725514852986e-05, {   3,   0,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.511521720719e-05, -2.462492839121e-05, {   2,   1,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -1.139104201659e-05, -4.824098161672e-05, {   3,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.953387145589e-05,  7.771399244327e-08, {   0,   1,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.182398505590e-05, -1.778431111947e-05, {   0,   1,  -1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.972631250111e-05,  1.984137202498e-05, {   4,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.255540934261e-05, -6.892466315712e-06, {   2,   1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  4.092731708889e-05,  1.851713467474e-05, {   1,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.293260530879e-05, -4.644766609802e-05, {   4,  -1,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.293219560736e-05, -4.644673095542e-05, {   4,  -1,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  2.058063685850e-05,  3.878081903118e-05, {   0,   1,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -4.745150923433e-05,  1.187697773077e-05, {   2,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.143583247745e-05,  4.782989780095e-05, {   3,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.479877343921e-05,  2.443491882841e-05, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.479935343878e-05,  2.443354827848e-05, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.113805121634e-05,  4.807120616143e-05, {   1,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.052219122951e-07, -5.886952600818e-05, {   3,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.839649292066e-05,  6.279391747303e-07, {   1,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.821015039907e-05,  2.074702162156e-05, {   2,   1,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.163794495753e-05,  4.731848811294e-05, {   1,   0,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.776822037408e-06,  5.703312677264e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -2.762222049083e-05, -3.116536928651e-05, {   0,   1,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.148003273147e-05,  2.729626615711e-05, {   2,  -1,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.849571275894e-05,  1.021533412684e-05, {   1,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.863830744640e-05,  5.226049400443e-08, {   2,  -1,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.118151289214e-05, -1.749987629439e-05, {   0,   1,  -3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.072365720051e-05,  1.795287621028e-05, {   2,  -1,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  9.980201773784e-06, -4.866880903810e-05, {   2,   1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  1.109187849357e-05,  4.753379292416e-05, {   1,   0,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.570887857604e-07,  5.771654059527e-05, {   2,  -1,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.345777905608e-05,  3.510963265292e-05, {   0,   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  4.540532213139e-05, -1.310002032320e-05, {   3,  -2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.060145114824e-05,  4.789904088714e-05, {   2,  -1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  1.760568972691e-05,  4.089171891859e-05, {   1,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.063030636543e-05, -4.783937984411e-05, {   4,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.898995988663e-05,  1.947381557060e-05, {   4,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.897854753509e-05,  1.946850447441e-05, {   0,   1,   3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.225442321621e-05,  4.617710538000e-05, {   0,   1,   3,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.081648962262e-05,  4.754122354582e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -5.829481711508e-05, -2.312700202901e-08, {   2,  -1,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.436102268757e-05, -3.939321774268e-06, {   1,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.669180541432e-05, -2.158818929329e-05, {   0,   1,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.631889579753e-05,  3.195595487251e-05, {   4,  -1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.157602234118e-05, -1.654384435494e-05, {   2,  -1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.799842987007e-05,  1.023662875411e-07, {   0,   1,   1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.543545476228e-06,  5.645283722871e-05, {   4,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.645186485028e-05, -2.146100584490e-05, {   0,   1,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -5.757489229392e-05, -9.481111055046e-08, {   0,   1,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  4.939941024762e-05, -8.253042337314e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -5.615333405216e-05,  1.475734850770e-06, {   2,  -1,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.756522570196e-05, -1.450315819028e-08, {   2,  -1,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.511741698895e-05, -2.434502191274e-06, {   2,  -1,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -4.261055341197e-05,  1.492988919248e-05, {   1,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  4.882733725604e-05,  8.691644236570e-06, {   2,  -3,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.983386900389e-05,  1.766576115447e-05, {   2,  -1,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.119810644983e-05, -4.627889188523e-05, {   4,   1,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.740381365889e-05,  2.335053681361e-08, {   1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.909643660735e-05, -8.257912177766e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -4.155497785389e-05,  1.579857657940e-05, {   2,  -1,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -5.617244239499e-05,  1.175791760275e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.300764217353e-05,  4.430264305200e-05, {   2,   1,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.649429715510e-05, -8.025329414312e-07, {   1,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.094834598788e-05, -4.633101637045e-05, {   2,  -1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.406346801373e-05, -3.317169772754e-05, {   2,  -1,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.706902343000e-05, -1.007171123241e-05, {   1,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.066516394860e-05,  4.628340705548e-05, {   1,   0,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.179117876343e-06, -5.574667271583e-05, {   2,   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -5.465229316855e-05,  2.137539955261e-06, {   4,  -1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.183595029691e-05, -3.490051627693e-05, {   4,   1,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.666888557253e-05,  4.860055225296e-08, {   2,  -1,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.092919248063e-05,  4.577919831880e-05, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.419308430876e-05,  2.249879275238e-05, {   1,   0,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  5.440074789227e-05,  2.282244609534e-06, {   1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 { -4.733068557081e-05,  9.349697903302e-06, {   1,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.593071640405e-05,  7.349985609001e-07, {   1,   1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.630274267392e-05,  4.034904732680e-05, {   1,   0,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.322951897060e-05,  1.340050396958e-05, {   1,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  4.608282896561e-05, -1.049957274789e-05, {   4,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.647166599959e-05,  2.006773308487e-05, {   1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.942914082364e-05, -3.693405132502e-05, {   0,   1,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -5.050329749669e-05,  5.839050295098e-06, {   2,  -1,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  5.563954853028e-05,  6.955276444577e-07, {   2,  -1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -2.168640837399e-05, -3.464668656528e-05, {   4,   1,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.215841489940e-05, -2.414446259399e-05, {   1,   0,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -3.397036655112e-05, -2.223919437151e-05, {   4,  -1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.970664725977e-05,  2.648525901951e-05, {   2,   1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -4.062465884334e-05, -1.555877405057e-05, {   2,  -1,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  2.194013408519e-05, -3.419837393746e-05, {   1,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.157683317574e-05,  2.455409551565e-05, {   2,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.302455874382e-05, -3.069409994902e-06, {   0,   1,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  5.596081210854e-05,  1.278394509153e-07, {   2,  -1,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -5.594532063012e-05, -1.431677621717e-07, {   2,   1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.078660303504e-05, -4.521030767163e-05, {   1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.815239428984e-06, -5.417462684325e-05, {   2,   1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.723506386810e-05, -8.751227266265e-06, {   1,   2,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 { -4.723506386810e-05, -8.751227266265e-06, {   1,   0,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  3.625098809629e-05,  1.968221896190e-05, {   2,  -1,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  4.921436319451e-05, -6.675499800522e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.073942131575e-05,  4.513326978643e-05, {   0,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.795955845136e-05,  7.778170935704e-06, {   1,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.680895503952e-05, -8.880062952762e-06, {   2,  -1,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.976514459700e-05, -5.902419056561e-06, {   1,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -4.417981423119e-05,  1.146919693119e-05, {   1,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.021810860036e-05,  3.540742550921e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -2.300177083233e-05, -3.258571956564e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  5.002121837437e-05, -5.516943115441e-06, {   2,  -1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.374466391390e-05, -1.791028156779e-06, {   2,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.717686263366e-05,  1.833540019042e-05, {   0,   1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.661770190001e-05,  3.881913576999e-05, {   1,  -4,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.861648387815e-06, -5.357427137433e-05, {   2,  -1,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  5.514806724142e-05,  2.712222236486e-07, {   0,   1,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.215204430536e-05,  2.312275117176e-05, {   2,  -1,   0,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  2.211536277739e-05, -3.314760969074e-05, {   4,  -1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.494129790704e-05, -3.031184320582e-05, {   1,   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 { -5.520914183396e-05,  4.342012012222e-08, {   2,   1,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.434029999425e-05, -2.084457593534e-05, {   2,  -1,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  2.058631349431e-05,  3.456281227044e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -4.702378396341e-05,  8.066131650537e-06, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  4.713673934071e-05, -7.947162791581e-06, {   2,   1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.558544521371e-05,  1.946467561101e-05, {   2,  -1,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -3.779102679147e-05,  1.712967921872e-05, {   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.190057427524e-05, -2.953804132917e-06, {   0,   1,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  4.248998614550e-05, -1.233714386525e-05, {   1,   0,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.940369991022e-05,  1.538217477570e-05, {   2,  -1,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.237156907187e-05,  3.241404889578e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   3,   0,   0 } },
 { -4.747202977696e-05, -7.300942883667e-06, {   1,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.222928002002e-06, -4.853766599859e-05, {   4,   1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.950756072013e-05,  2.523175390959e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -2.192443624874e-05, -3.267176973001e-05, {   1,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  5.454500261414e-05,  1.746629484486e-08, {   2,  -3,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.951555177382e-05, -1.494829889808e-05, {   0,   1,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.251602697703e-07,  5.412858243112e-05, {   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.443224615724e-05, -5.146335682467e-09, {   1,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.641290721455e-05, -8.006087324731e-06, {   2,  -1,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.362830945886e-05, -4.073217931983e-05, {   1,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.528995712102e-06, -4.979445950947e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.528804515070e-06, -4.979447689919e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.341789782132e-05, -1.088869613424e-05, {   1,   0,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -1.163289675037e-05, -4.265639252506e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -4.472772228905e-06,  4.979954124877e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.472676620641e-06,  4.979954983579e-05, {   1,   0,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.525662043516e-05, -1.901054336211e-05, {   0,   1,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  2.433848048998e-06, -5.181859604586e-05, {   2,  -1,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.417380380082e-05,  7.211879357256e-08, {   0,   1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.884867361881e-06, -5.230813894841e-05, {   2,   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.884945975408e-05,  3.528027589149e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.223031203727e-05,  4.189665875312e-05, {   0,   1,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.512831534111e-05, -1.893008364507e-05, {   0,   1,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  1.409080103802e-05, -3.996572269682e-05, {   0,   1,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.984303657127e-05,  1.419839000253e-05, {   2,   1,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  5.459658283414e-06, -4.857374547527e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.611753725249e-05,  1.788197489063e-05, {   0,   1,   0,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -3.071534680665e-06,  5.091710704086e-05, {   2,  -1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.610908621220e-05,  1.787651695157e-05, {   0,   1,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -4.175814343627e-05,  1.216942024932e-05, {   3,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -9.079024712460e-06,  4.481972933664e-05, {   1,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.378708933220e-05, -8.617711398972e-08, {   0,   1,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -4.853484206902e-05, -5.290230339681e-06, {   0,   1,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -3.238858805183e-05, -2.138354440455e-05, {   1,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.330089442293e-05, -4.496749136453e-07, {   3,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.538083977989e-05, -8.226394189099e-06, {   2,   1,  -1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -9.063970979373e-06, -4.451024062276e-05, {   2,  -1,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.031554492964e-05, -4.320507740139e-05, {   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.938256330847e-05, -1.413583535896e-05, {   2,   1,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -5.334243212012e-05,  1.637185328903e-07, {   2,  -1,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.636925147719e-05, -2.712658316030e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -4.383015356831e-05,  9.634400821042e-06, {   2,   1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.333989694857e-05,  7.737442517579e-08, {   0,   1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.150495174173e-06, -5.126248898227e-05, {   0,   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  4.965396443144e-05, -3.749630545717e-06, {   4,  -1,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.716070611659e-05, -1.623404145283e-05, {   0,   1,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.222848130631e-05,  4.116424650946e-05, {   0,   1,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -7.392036115379e-06, -4.599055452519e-05, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.046446021246e-05,  4.291337941400e-05, {   1,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.322542905927e-05,  6.532666319329e-08, {   2,  -1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.117738833361e-05,  4.209380707561e-05, {   0,   1,   1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.720289200900e-05, -3.604139742041e-05, {   1,   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  2.349908960968e-05, -2.968049579712e-05, {   1,   0,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.419241615992e-05,  3.891762287272e-05, {   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.017207794132e-05,  4.292094560644e-05, {   3,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.230473772377e-05, -7.171232107919e-07, {   2,  -1,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -5.059572152075e-05,  2.353628575752e-06, {   2,  -1,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.906485172339e-06, -5.099981277343e-05, {   1,   0,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -3.573338313263e-05,  1.711946733605e-05, {   1,   0,   2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.944240805555e-05, -3.333441301519e-05, {   2,  -1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.913952309082e-05, -1.361023021204e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.510103942695e-05,  3.763644449266e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  5.248918311059e-05, -2.358738169645e-07, {   2,   1,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.528973469625e-05,  1.739916251087e-05, {   4,  -1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.177483525313e-05,  3.084713553372e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -1.883733903539e-06, -5.071548970167e-05, {   2,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.325236310523e-05, -2.929598352525e-05, {   0,   1,  -1,   0,   0,  15, -11,   0,   0,   0,   0,   0,   0 } },
 {  7.194066513863e-06, -4.534437336169e-05, {   0,   1,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  4.337257188945e-05, -9.115712528505e-06, {   0,   1,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.009148788110e-05,  4.236937791002e-05, {   1,   0,   2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.720326435460e-06, -4.372303366769e-05, {   0,   1,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -8.724180194227e-06, -4.371899415184e-05, {   0,   1,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -3.339547581501e-05, -1.902600297823e-05, {   1,   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  4.567955300764e-05, -6.712050500561e-06, {   0,   1,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.023136538186e-05, -4.214918029785e-05, {   3,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.709557232549e-05, -5.259013566253e-06, {   0,   1,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.019623723161e-05, -4.213770032177e-05, {   4,  -1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.518786690575e-05,  1.706240011938e-05, {   2,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.343443394948e-05,  1.879509959627e-05, {   4,  -1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.869709755781e-05, -3.351414631689e-05, {   1,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  4.973972562165e-05,  2.464166675262e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  5.205697330011e-05,  1.106858119637e-07, {   2,   3,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.850845497844e-06,  4.230386389976e-05, {   1,   0,  -1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.071656286816e-06,  5.002105621277e-05, {   2,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.217414712604e-05, -2.985058680192e-05, {   1,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.868621396331e-05,  1.332379587037e-05, {   2,   1,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  2.944145070530e-05, -2.256836816044e-05, {   4,   1,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.746480410822e-05,  2.446592064736e-05, {   2,   1,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -5.181140732066e-05, -1.106012577815e-07, {   2,   1,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.937951076631e-05, -3.249348472889e-05, {   2,  -1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.293678602861e-05, -2.889883578078e-05, {   0,   1,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.023067836332e-05, -2.159483773387e-05, {   1,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -2.502404481923e-06,  4.928266926496e-05, {   2,   1,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -2.469979997252e-05, -2.705452881994e-05, {   2,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.927255107353e-05, -3.247373895924e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 { -3.242061765876e-05, -1.926638580714e-05, {   0,   1,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -9.884498929324e-06,  4.175127235654e-05, {   1,  -2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.004732944853e-05, -4.152350481135e-05, {   4,  -1,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.807273071681e-06,  4.173432232792e-05, {   1,   0,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.807777319559e-06,  4.171870966005e-05, {   1,   0,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.230661995283e-05, -1.920717071183e-05, {   0,   1,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  9.855992159890e-06, -4.163651822798e-05, {   1,   2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.411317793970e-05,  1.737402655867e-05, {   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.475137918153e-05,  1.668160138318e-05, {   0,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.515330554634e-05,  1.625597444165e-05, {   2,  -1,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -4.299569678837e-05, -8.368839649631e-06, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.910524132185e-05, -2.246323371596e-06, {   2,   1,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  4.641302447453e-05, -4.931741403808e-06, {   1,   0,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.119446122284e-05,  3.007596753417e-05, {   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.291726595241e-05, -1.832082231616e-05, {   1,  -1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.580204890107e-05, -1.542270244307e-05, {   0,   1,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.043395248943e-05,  4.078197561604e-05, {   2,   1,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.090421623818e-06,  4.911212821880e-05, {   2,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.081737028612e-05, -3.743773893697e-07, {   2,  -1,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  3.677516354053e-06, -4.749646875621e-05, {   2,  -1,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.868258462468e-05,  2.433892379269e-06, {   2,  -1,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  4.868258462468e-05,  2.433892379269e-06, {   2,  -1,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  5.092005853288e-05,  1.619870532460e-07, {   2,  -1,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -9.676049733341e-06, -4.138575943604e-05, {   3,   0,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -4.175347614967e-05, -9.271828325392e-06, {   0,   1,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -3.497234443745e-05, -1.598410529085e-05, {   1,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.080080111625e-05, -4.013220092994e-05, {   1,   2,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.999501947125e-05, -3.090608442127e-05, {   4,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.467772025199e-05, -6.207154258597e-06, {   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.306712582659e-05,  1.777843095674e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  3.176011725664e-05,  1.898093636282e-05, {   1,   2,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.876865136044e-06,  4.484465418124e-05, {   3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.451570996397e-05, -1.618126042065e-05, {   2,  -1,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.320192521544e-06,  4.133568018440e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  3.280068627444e-05,  1.784192228417e-05, {   0,   1,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.060621939412e-05,  1.661779045746e-08, {   2,   1,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.016916536638e-05,  4.531438071269e-07, {   2,  -1,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -4.803859213193e-05,  2.501993124994e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  1.858525833578e-06, -4.862244815481e-05, {   2,  -1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.907816053976e-05, -1.139880265579e-05, {   1,   0,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  7.832507024566e-06,  4.258612182183e-05, {   3,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.588749455250e-05,  3.450882966268e-05, {   1,   0,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.070621464424e-05,  2.965823067167e-05, {   0,   1,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -2.980233846838e-06, -4.735598377397e-05, {   0,   1,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -8.620968110264e-07, -4.944990408517e-05, {   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.489209300687e-05, -5.372885456563e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  1.993866325403e-05, -3.030697345979e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.496633359127e-05,  3.521518524481e-05, {   2,  -1,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.800276310910e-05, -3.215702122385e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  4.999989853133e-05,  1.007316096381e-07, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.999993313464e-05,  6.333969682595e-08, {   4,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.999928601087e-05,  4.889174810247e-08, {   0,   1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.399397349825e-06,  4.063608940313e-05, {   1,   0,  -1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.999999998867e-05, -1.064650843627e-09, {   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  4.999999999583e-05,  6.457718232199e-10, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.999999999605e-05,  6.283185307014e-10, {   6,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999999637e-05, -6.021385919162e-10, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.999999999668e-05, -5.759586531546e-10, {   2,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999999717e-05,  5.323254218482e-10, {   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999999726e-05,  5.235987755887e-10, {   2,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.999999999744e-05, -5.061454830816e-10, {   2,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.999999999999e-05,  2.617993878822e-11, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -2.349666122870e-05,  2.648859886885e-05, {   2,  -1,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.818682781785e-06,  4.815285900345e-05, {   1,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.510913058801e-06, -4.843608422030e-05, {   2,  -1,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.051283556403e-05,  2.941359053438e-05, {   0,   1,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.235199942049e-06, -4.868317655399e-05, {   2,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.979260079312e-05,  7.778719976159e-08, {   4,   1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.986526150668e-05, -9.970250477391e-06, {   3,   2,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.975748752845e-05, -4.091934383897e-08, {   2,   1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.962689362762e-05,  1.539632413140e-07, {   2,   1,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.784821283940e-05,  1.855431671677e-06, {   0,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  9.508341516294e-06, -4.017892630682e-05, {   1,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.965851763789e-05,  2.806566152838e-08, {   4,  -1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.293366281174e-05, -3.672433477800e-05, {   1,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -8.328151049006e-06,  4.132907473284e-05, {   1,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.438409774071e-05, -5.272787215030e-06, {   2,  -1,   0,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 { -3.589666385047e-05, -1.375636530455e-05, {   0,   1,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.265712498770e-06, -4.836714800020e-05, {   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.677815236093e-05, -1.274845393003e-05, {   2,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.834918040852e-05, -1.117737527697e-05, {   1,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.617663741995e-07, -4.934913565210e-05, {   4,   1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.904161587424e-05,  4.624335597424e-07, {   2,  -1,  -4,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.844226225737e-05, -3.105885366378e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  4.926731070236e-05,  7.678863292791e-08, {   2,  -1,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.691304463560e-05, -2.420596056921e-06, {   0,   1,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -4.174334739569e-05, -7.550440274634e-06, {   2,  -1,   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -4.903358577063e-05, -2.581016878567e-07, {   2,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -9.373124817303e-06,  3.991472413586e-05, {   1,   0,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.618896224104e-05, -3.305105136790e-05, {   2,  -1,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  9.238946299199e-06, -3.994015483382e-05, {   0,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.886188706310e-05, -2.018827303030e-05, {   2,  -1,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  3.691963870293e-05, -1.211298617003e-05, {   0,   1,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.100456767722e-05,  8.019849264129e-06, {   1,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.103639326326e-05, -3.796647337673e-05, {   4,  -1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -9.780938984462e-06,  3.920405482232e-05, {   1,   0,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.890062421814e-05,  7.126655941771e-08, {   0,   1,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  4.188734541256e-05, -7.083112314386e-06, {   2,  -1,   0,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  1.173533495567e-05,  3.723382794219e-05, {   1,   0,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.102551231534e-05, -3.793719378615e-05, {   4,  -1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.351127478421e-06,  3.960714917771e-05, {   1,   0,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.890147604328e-05, -4.913870239154e-08, {   4,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.272455274745e-05, -6.212791236523e-06, {   2,  -1,   0,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  9.159538907840e-06, -3.974902446349e-05, {   3,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.875434625086e-05, -3.013723225680e-05, {   0,   1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.952026469685e-06, -4.691499697431e-05, {   2,  -3,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.960067069133e-05,  2.926343871816e-05, {   1,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  4.645493640950e-05, -2.385323115145e-06, {   0,   1,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  3.604618827676e-05,  1.276106808660e-05, {   2,   1,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  4.710766377230e-06,  4.398626026872e-05, {   4,  -1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.982871940464e-05, -8.862617023182e-06, {   0,   1,   2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.246607949898e-06, -4.041019431983e-05, {   2,   1,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.671065616843e-05, -2.191186889837e-05, {   4,   1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.629784644427e-05,  3.232043221540e-05, {   1,   0,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  4.774145736294e-05,  8.256653841581e-07, {   0,   1,   1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  9.277334416522e-06,  3.920864529193e-05, {   1,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.453014362030e-06,  4.403261406919e-05, {   0,   1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.822166883067e-05,  2.563891946222e-07, {   1,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.887191941365e-05, -2.958827499253e-05, {   0,   1,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -1.489111827678e-05,  3.349604020147e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  4.482152810184e-06, -4.384500810943e-05, {   1,   0,   0,   0,   0,   0,  -5,  11,   0,   0,   0,   0,   0 } },
 {  9.885945468437e-06,  3.838483425944e-05, {   1,   0,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.793729083596e-05,  3.260584348283e-07, {   2,  -1,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  1.864446561892e-05, -2.959664349088e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  8.753966101827e-06, -3.945111650900e-05, {   4,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.060240776504e-05, -3.756778218161e-05, {   1,   0,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.809024681447e-05,  4.856933397385e-08, {   0,   1,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  9.264930454209e-06, -3.885283704753e-05, {   2,  -1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.800067127598e-05, -3.011339088010e-05, {   4,  -1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  9.332969791297e-06,  3.877206931160e-05, {   1,  -2,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.869920575397e-05,  2.938785359112e-05, {   2,  -1,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 { -1.264410641374e-05, -3.542203776467e-05, {   1,   0,   0,   0,   0,   6,  -9,   0,   3,   0,   0,   0,   0 } },
 { -4.722580764402e-05,  8.183902382282e-07, {   0,   1,   1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -9.164463167483e-06, -3.886306714066e-05, {   1,   0,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.193787901903e-06,  3.882070747213e-05, {   1,  -2,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.555911158154e-05,  1.239323038159e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -9.208742942092e-06, -3.873999913956e-05, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.781501258274e-05,  3.004842240093e-05, {   2,   1,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -3.968433670331e-05,  8.178663418944e-06, {   0,   1,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  2.022760050841e-05,  2.762761119687e-05, {   1,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.388904460929e-05, -3.955666034667e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.742179373229e-05,  4.151560710850e-07, {   0,   1,   2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -3.785658785434e-05,  9.959716336180e-06, {   2,   1,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -7.580149594111e-06, -4.020033653225e-05, {   3,   0,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  9.005817678391e-06,  3.875417390110e-05, {   1,   0,  -1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.257610860729e-05,  2.515260201204e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  3.438081923741e-05, -1.334538544901e-05, {   1,   0,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -9.065935541101e-06,  3.857163824008e-05, {   1,   0,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.247706501638e-05,  3.514080292949e-05, {   2,  -1,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -4.743795420578e-05,  1.788162748163e-07, {   2,  -1,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.247557744894e-05,  3.513579698893e-05, {   2,  -1,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.237637136717e-05, -2.522626978845e-05, {   4,   1,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.040832953834e-05, -7.179406277515e-06, {   4,  -1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  7.174844664367e-06,  4.034765949862e-05, {   2,  -1,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.751061611421e-05, -1.101373446418e-08, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  8.693765587508e-06, -3.881940248053e-05, {   2,  -1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.743676095880e-05,  4.840286889868e-08, {   2,   1,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.799865912770e-05, -1.948347053900e-05, {   2,   1,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -4.698328325220e-05, -4.746967750114e-07, {   3,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.243001542100e-05, -2.498962191356e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  2.133875508881e-06,  4.522176001285e-05, {   2,   1,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -9.167519642497e-06,  3.817260632422e-05, {   4,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.039728561814e-06, -3.822127427207e-05, {   1,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.303055751644e-05, -4.184995686080e-06, {   2,  -1,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  2.169707689179e-05,  2.546165827753e-05, {   2,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.024572199572e-05,  1.689379990185e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -3.607952183184e-05,  1.104657701882e-05, {   0,   1,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.729068033642e-05,  9.801991433353e-06, {   2,  -1,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -4.656018826707e-05,  5.085841872917e-07, {   4,  -1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.575924362818e-05,  1.128252695377e-05, {   1,  -2,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.243213459620e-05, -4.588227460702e-06, {   2,   1,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -4.163387018367e-05,  5.383146468819e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -3.836606926156e-05, -8.641217100797e-06, {   2,  -1,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  9.646387334358e-06, -3.731252093386e-05, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.383885863266e-05,  3.307099212395e-05, {   4,  -1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.660790612277e-06, -4.522580136571e-05, {   0,   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  8.065911935047e-06, -3.880053134985e-05, {   4,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.594468568612e-05,  9.207225581434e-07, {   0,   1,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.299267402389e-06, -4.555635043573e-05, {   0,   1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.747692667022e-05,  9.348522288290e-06, {   1,   0,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -1.878042684316e-05,  2.804216109400e-05, {   0,   1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  3.175882137724e-05,  1.505984392110e-05, {   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.089329535911e-05, -3.589648565493e-05, {   4,  -1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.816937638728e-06, -4.496314591880e-05, {   0,   1,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -4.378986908254e-05,  2.986378998873e-06, {   1,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.675829256885e-05, -1.676872925475e-08, {   1,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.243353910387e-05, -3.431588161161e-05, {   1,   0,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.087980779614e-05, -3.585399154357e-05, {   4,  -1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -8.779542716160e-06,  3.794252464621e-05, {   2,  -1,   0,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  9.867538598332e-06,  3.669646430458e-05, {   3,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -8.783790709697e-06,  3.777642955371e-05, {   1,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.342867653634e-05, -3.312202837283e-05, {   1,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.957853333659e-06,  3.758867608304e-05, {   1,   2,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.149566434708e-05, -1.504673589713e-05, {   1,  -2,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  2.567557444578e-05, -2.083763301597e-05, {   1,   1,   0,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.567557444578e-05, -2.083763301597e-05, {   1,  -1,   0,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.832434202688e-06, -4.467032042059e-05, {   0,   3,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.909919109945e-05,  7.399725652415e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.923617715781e-06,  4.456879023475e-05, {   2,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.896458171248e-06, -4.457870819839e-05, {   2,  -1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.266073645625e-06, -4.518876212990e-05, {   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.814608808674e-05, -2.827590251150e-05, {   2,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.548531311848e-05,  9.112486316412e-07, {   0,   1,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -2.094789514270e-05, -2.543237929262e-05, {   2,   1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.243175264966e-05, -3.392920799208e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 } },
 { -4.616548681171e-05,  1.514081036980e-07, {   2,  -3,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.508872956586e-05, -3.122657983814e-05, {   1,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  4.612155479401e-05, -1.920787555765e-07, {   2,   1,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -3.892685967758e-05,  7.304286397206e-06, {   3,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.603999776311e-05,  1.838922833655e-07, {   2,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  4.596792793058e-05,  2.418667674977e-07, {   4,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.880591018782e-05, -2.738066429866e-05, {   2,  -1,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.208604173119e-05, -3.403974941662e-05, {   0,   1,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.208511602378e-05, -3.403711562305e-05, {   0,   1,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -3.940610541833e-05, -6.665654037339e-06, {   2,  -1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -3.683180157017e-05, -9.189754264840e-06, {   1,   0,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  2.696719137104e-05,  1.904754383716e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -8,   0,   0,   0 } },
 {  1.845405550757e-05,  2.751441266468e-05, {   0,   1,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  2.048573174979e-05,  2.548155574611e-05, {   0,   1,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.531651634446e-05,  3.063489801162e-05, {   2,   1,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -4.442295922199e-05, -1.513563418587e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -3.672980064949e-05, -9.150770076193e-06, {   2,  -1,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.140456781135e-05, -2.445970325968e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  3.057107321668e-05, -1.526867711045e-05, {   2,   3,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.709184395995e-06,  3.712315570138e-05, {   1,   0,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.143512778128e-05,  1.437657498367e-05, {   1,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.347592548941e-05,  3.231621256691e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -1.965131196223e-05, -2.613886796587e-05, {   4,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.134399847667e-05, -2.441786247258e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -4.555205199830e-05, -1.508499338112e-07, {   2,  -1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.130040071266e-05,  1.436388238244e-05, {   2,   1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.384104736246e-05,  1.815970131150e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  2.176532337066e-05,  2.387164340702e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 { -2.176517336623e-05,  2.387147980582e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 {  4.526307828593e-05,  3.049906177245e-07, {   2,   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -4.349981730048e-05,  2.041191517633e-06, {   1,   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  3.774661014338e-05,  7.784188244592e-06, {   2,   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -4.543652002230e-05, -8.043422304250e-08, {   4,  -1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.543764674822e-05,  7.132622724140e-08, {   0,   1,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.565355382083e-06,  3.694224361732e-05, {   1,   0,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.995674434413e-05, -2.554358019126e-05, {   2,  -1,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.912155686235e-05, -2.635924841334e-05, {   2,   1,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.340424333685e-06, -3.611648551878e-05, {   0,   1,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.026955482823e-05,  2.517858938272e-05, {   0,   1,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  9.332787549743e-06, -3.609103699554e-05, {   0,   1,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.756423907657e-05,  7.837957356140e-06, {   1,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.878346249165e-05,  2.661254587451e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -8.771461593028e-06, -3.661022455985e-05, {   1,   2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.496940057947e-06,  3.888221352876e-05, {   3,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.898007873664e-05,  6.397794963295e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  3.627348848391e-05, -9.071831129374e-06, {   3,   2,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.098031005866e-05,  3.435751268827e-05, {   2,   1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.098451416353e-05,  3.434916282962e-05, {   2,   1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  4.334381207700e-05, -1.981139887046e-06, {   2,   1,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  1.830130412363e-05,  2.701198046031e-05, {   1,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.427413351256e-05,  9.609629207463e-07, {   1,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.643305572763e-06, -4.357018174612e-05, {   2,  -1,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -3.205604707345e-05,  1.311548107733e-05, {   2,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.519020209688e-05, -1.997543373041e-05, {   2,  -1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.017339836655e-05,  1.498573979183e-05, {   1,   0,   2,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.317132515706e-05, -1.973874583043e-06, {   4,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.813905604235e-05,  1.696162237315e-05, {   0,   1,   2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -3.445843015081e-05, -1.064203702305e-05, {   0,   1,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.919940998115e-05, -1.588249992148e-05, {   0,   1,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -2.811552207800e-05,  1.694654689111e-05, {   0,   1,   2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -4.498922685599e-05, -6.641310644556e-08, {   0,   1,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.004541109014e-05, -1.500628854330e-05, {   2,   3,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.166516513814e-06,  4.288475607757e-05, {   0,   1,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  3.871101920449e-05, -6.315758439307e-06, {   1,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.102192709085e-06, -4.284963482971e-05, {   2,   1,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  2.641725684841e-05, -1.846979185195e-05, {   2,   1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.243495198835e-05, -2.243610420891e-05, {   2,  -1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.045566154734e-06, -4.280659550392e-05, {   2,   1,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  3.529752280789e-05, -9.543889423113e-06, {   0,   1,   3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -7.212770849998e-06, -3.761328492067e-05, {   2,   1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.879583277883e-05, -1.601772060657e-05, {   1,   0,   0,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -3.526529237659e-05, -9.537777517017e-06, {   0,   1,   3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.661301893056e-05,  2.817897343461e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 { -2.563969654446e-05, -1.909205799265e-05, {   1,   0,  -1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -4.460339925907e-05, -1.151766785685e-07, {   0,   1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.660982118645e-05, -2.810207388871e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  2.875265529017e-05,  1.594459040542e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  3.801193115422e-05, -6.673991378579e-06, {   0,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.717987030130e-06, -4.196277535468e-05, {   0,   1,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  3.752771121264e-05, -7.097919585130e-06, {   1,   2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.815312377694e-05,  2.644785804869e-05, {   0,   1,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  2.322812946070e-05, -2.137037210375e-05, {   2,  -1,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.430708604664e-05,  3.025216520479e-05, {   2,  -1,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -1.721645824296e-05, -2.732391970047e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  2.627523759871e-05, -1.824197731255e-05, {   2,   1,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.197742833082e-06,  3.931653056347e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.981985253403e-05, -2.469164196094e-05, {   1,   0,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.422283929014e-06, -3.606105195571e-05, {   3,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.173231092169e-05, -2.274176756476e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -2.867638336029e-05, -1.579093732127e-05, {   4,  -1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.808495671243e-05,  2.636440523761e-05, {   0,   1,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -8.280377409172e-06, -3.612504829605e-05, {   1,   0,  -1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.965771054113e-05, -2.472808221581e-05, {   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -4.068402038560e-05, -3.599065178659e-06, {   2,  -1,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.041010460512e-06,  3.821906618879e-05, {   0,   1,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.871686602432e-05, -1.554134616316e-05, {   2,   1,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.813554191804e-05,  1.611969188337e-05, {   0,   1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.484509296914e-06, -3.566774459336e-05, {   1,  -2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.405937099092e-05,  7.980369030013e-08, {   2,  -1,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  8.442030125151e-06, -3.566127038798e-05, {   0,   1,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.372148141980e-05, -3.034883672979e-05, {   2,   1,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -3.155623902398e-05,  1.250871676396e-05, {   2,   1,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -3.191808642936e-05, -1.210405234173e-05, {   0,   1,  -1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.239429213227e-06, -3.577277812658e-05, {   2,   0,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.668451236131e-07, -4.311684035308e-05, {   2,  -1,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.458795299041e-05,  1.932863095788e-05, {   2,  -1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  3.894431429156e-05,  4.952822876010e-06, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  3.892592502982e-05, -4.963109030626e-06, {   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.398232867954e-05,  9.897556365147e-06, {   2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.479155923794e-05, -2.906546147830e-05, {   2,   1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  3.649396835695e-05, -7.339831711331e-06, {   1,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.498307758757e-05, -2.882980270332e-05, {   2,   1,   0,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -1.498307758757e-05, -2.882980270332e-05, {   2,  -1,   0,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -3.604846160717e-05, -7.759984342453e-06, {   1,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.331304755671e-05, -3.047309317559e-05, {   0,   1,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  4.995245973425e-06,  3.876636032260e-05, {   2,   1,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -3.971481411436e-05, -4.044286401268e-06, {   2,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.839882784074e-05,  2.534768252230e-05, {   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.329003032116e-05, -3.042074557172e-05, {   0,   1,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.428203369763e-05,  2.940735470235e-05, {   2,   1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.635358760888e-06, -4.198261994500e-05, {   0,   1,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  3.767600256415e-05, -5.894261928577e-06, {   2,  -1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  3.234008431096e-05, -1.122954025426e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -2.048050748575e-05, -2.308740764255e-05, {   0,   1,   1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.210318855890e-05, -1.461211172034e-06, {   2,   1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.347664205563e-05,  6.620345178875e-08, {   0,   1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.371209781464e-06, -4.214041714001e-05, {   4,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.210380849924e-06, -3.528964474062e-05, {   1,   0,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.313986934072e-06,  3.517667183985e-05, {   3,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.500107420454e-06, -4.193848511129e-05, {   4,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.023917134154e-05,  3.187867472007e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  3.578217783124e-06,  3.983963299902e-05, {   3,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.578217783124e-06,  3.983963299902e-05, {   3,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.578141296513e-06,  3.983963986863e-05, {   3,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.578134343184e-06,  3.983964049313e-05, {   1,   0,   2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.577904883268e-06, -3.983966110105e-05, {   1,   2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.139540183837e-05, -3.201965202296e-05, {   1,   0,   0,   0,   0,   5,  -5,  -4,   0,   0,   0,   0,   0 } },
 { -3.448484095673e-05,  8.871628478888e-06, {   1,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.470221932782e-05,  8.540502517367e-06, {   2,   1,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  3.913042301011e-05,  4.078648753740e-06, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.004440673801e-06,  3.620027018708e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.065770414829e-05,  3.253264655239e-05, {   2,  -1,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.160506209716e-05,  1.155017259902e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -4.898789433675e-06, -3.825439096021e-05, {   4,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -3.566256791830e-05, -7.462520464324e-06, {   1,   0,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.123342434972e-05, -1.885607368793e-06, {   2,  -1,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  2.285899155634e-05,  2.024098300910e-05, {   2,  -1,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -3.791826338056e-05,  5.134741829626e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -2.764527330208e-05, -1.537950344889e-05, {   2,  -1,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -8.226680007784e-06,  3.476319810064e-05, {   1,  -2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.883598677743e-05, -1.410783797770e-05, {   1,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.998678524211e-05, -2.293224681805e-05, {   0,   1,   3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.268579839569e-05, -1.013804226180e-05, {   4,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.749227341123e-05, -1.532331721815e-05, {   3,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.462074730889e-05, -2.818601761342e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.774482216099e-05, -1.506033999602e-05, {   2,   1,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  3.477345135502e-05, -8.016887431856e-06, {   4,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.013326004844e-05, -2.265534880613e-05, {   0,   1,   3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.265001719877e-05, -7.919092630875e-08, {   4,   1,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.732974990269e-06, -4.098195939048e-05, {   0,   1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.198336599637e-05,  7.123799731561e-07, {   0,   1,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  3.452947869721e-05,  7.993329763575e-06, {   2,   1,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  4.179872057567e-05,  7.106058789330e-07, {   0,   1,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  8.419759269652e-06, -3.406101928037e-05, {   1,   0,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.413048703756e-06, -3.503443081635e-05, {   0,   1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  3.607418495168e-05,  6.348145398113e-06, {   0,   1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.648134439050e-05,  1.593718542187e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -2.962899854372e-05,  1.274386814358e-05, {   2,   1,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -8.039479966427e-06,  3.432058129983e-05, {   1,   0,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.201253224414e-06,  3.709419324829e-05, {   2,   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  8.789654597223e-06, -3.349036057580e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -1.919286298122e-05, -2.301919509064e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -4.194069565442e-05,  2.137488492999e-07, {   0,   1,   2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.594671100767e-06, -4.055721751697e-05, {   2,   1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.598019453111e-05, -1.599809571003e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.163927431874e-05, -3.051302635131e-07, {   2,  -1,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.718330668926e-05,  4.723946532571e-06, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 { -4.581789032296e-06, -3.725277316392e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  2.837345133448e-05,  1.344953545412e-05, {   2,  -1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 {  7.467860575894e-07, -4.106927738950e-05, {   4,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.040166953570e-06, -3.476579284045e-05, {   2,  -1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.071843530844e-06, -3.462143881238e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -4.136450942340e-05,  3.284394363979e-07, {   4,  -1,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.693214534470e-05, -1.470022234614e-05, {   2,   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  3.790176537471e-05, -3.724216139846e-06, {   1,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  2.803304623093e-05,  1.357095454425e-05, {   2,  -1,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.979472029105e-06,  3.361068095887e-05, {   1,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.154499115554e-05, -2.330873076136e-08, {   4,  -1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.690000819054e-05,  2.461248577748e-05, {   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -8.126301403280e-06, -3.336642006140e-05, {   1,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.542646374998e-05, -6.018385382928e-06, {   1,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.132808896140e-05,  8.707811994149e-08, {   0,   1,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -2.284581261049e-05,  1.853755826688e-05, {   1,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.736808698668e-05, -2.400728848809e-05, {   2,  -1,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.359420090109e-06,  3.801509875580e-05, {   0,   1,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  8.468257551586e-06, -3.284345769205e-05, {   1,   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -2.661215798037e-05,  1.469834260031e-05, {   4,   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.833756004930e-05,  2.290050612019e-05, {   2,  -1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.882450928048e-06,  3.931504566657e-05, {   1,   0,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 { -2.742806342523e-05,  1.370108226137e-05, {   0,   1,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.438585437967e-05, -1.670845417388e-05, {   2,  -1,   0,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -3.476422124984e-05, -6.317811663330e-06, {   2,  -1,  -1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  1.931138271917e-05, -2.176982344757e-05, {   4,   1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.520989829854e-05, -1.585989456620e-05, {   0,   1,   1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.443224099569e-05,  6.636371568741e-06, {   1,   0,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.708801855420e-05, -3.895740015970e-06, {   2,   1,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  2.671767150373e-05,  1.424510088850e-05, {   0,   1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.232705337371e-06, -3.264428084472e-05, {   3,  -2,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.576055116462e-06,  3.428924619154e-05, {   3,  -2,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.081918913961e-05,  3.741819733749e-08, {   2,   1,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.087752768872e-05,  2.993425445463e-05, {   1,   0,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.278709433785e-05, -7.983214219711e-06, {   2,  -1,   0,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -2.731387900144e-05,  1.345463481599e-05, {   2,  -1,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.073534851678e-05,  1.205842953280e-08, {   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.270468054981e-05, -1.801565831011e-05, {   0,   1,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  3.795705663020e-05, -2.709583580510e-06, {   4,  -1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -3.842824961786e-06, -3.675768956201e-05, {   2,   1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.453900141076e-05, -6.057739973657e-06, {   2,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.043039350244e-05, -3.012355327130e-05, {   1,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.792500075846e-06,  3.275669886686e-05, {   3,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.018922173638e-05,  3.351252617112e-07, {   1,  -1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.557147219183e-05,  1.493172017737e-05, {   2,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  1.329672520815e-05,  2.713259934096e-05, {   2,   1,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -1.173605655039e-05, -2.866958579092e-05, {   0,   1,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  1.173342080258e-05, -2.866678721144e-05, {   0,   1,   1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -1.191352441626e-06, -3.918778533751e-05, {   2,   1,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.298247279633e-05,  2.735753566454e-05, {   2,   1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  7.442219316680e-07, -3.958901157382e-05, {   4,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.559459021278e-05, -1.468166555994e-05, {   2,   1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.535479000601e-06, -3.272650584837e-05, {   2,   2,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.395060932307e-05,  6.308665064472e-06, {   2,  -1,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.975940819198e-05, -4.616821529327e-07, {   2,   1,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.500469631062e-05, -1.521364722219e-05, {   0,   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  7.739407316470e-06,  3.247040975822e-05, {   1,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.992593125312e-05,  1.950182487546e-07, {   1,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -4.010460992875e-05, -1.630767429428e-08, {   2,  -1,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.780672090742e-06,  3.533108731802e-05, {   1,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.199173680248e-05,  8.115161648417e-06, {   2,   1,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -6.289674149796e-06,  3.381260873537e-05, {   3,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.233894818383e-05,  2.775090251791e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  3.037779502458e-05, -9.703849661430e-06, {   2,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.301012023279e-05, -2.705657602751e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  3.999999340203e-05, -2.816437710800e-08, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.297730869372e-06,  3.271944440526e-05, {   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.601261596116e-05, -1.399013924300e-05, {   2,  -1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -3.999999997339e-05,  1.459095254344e-09, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999137e-05, -8.307767238953e-10, {   0,   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999137e-05,  8.307767238896e-10, {   0,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999354e-05,  7.190756517829e-10, {   4,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999528e-05,  6.143558966779e-10, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  3.999999999657e-05, -5.235987755908e-10, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999666e-05,  5.166174585760e-10, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999693e-05,  4.956735075537e-10, {   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999693e-05, -4.956735075523e-10, {   2,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999701e-05,  4.886921905463e-10, {   6,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999701e-05, -4.886921905426e-10, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.999999999701e-05, -4.886921905426e-10, {   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.999999999735e-05, -4.607669225237e-10, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.999999999766e-05, -4.328416544846e-10, {   2,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.999999999882e-05,  3.071779483480e-10, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.094712044652e-05, -2.904462085765e-05, {   2,   1,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  7.728099781521e-06, -3.223838068377e-05, {   2,   1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.597911089496e-05, -2.396979153601e-05, {   2,  -1,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.969185795818e-05,  2.426667653614e-07, {   0,   1,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -5.704779947068e-06,  3.420831464664e-05, {   0,   1,  -1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.068518486159e-06,  3.181921006005e-05, {   1,   0,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.622806884997e-06,  3.825375591327e-05, {   2,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.364883919397e-05, -6.178104087245e-06, {   4,   1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.461686253090e-05, -2.516006531964e-05, {   0,   1,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -1.591489738072e-05, -2.385099981279e-05, {   2,  -1,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.437107136834e-05, -1.536743222092e-05, {   0,   1,   1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.658713860772e-05,  1.309891692783e-05, {   2,   1,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.938806096792e-05,  2.029225273732e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -3.691619998819e-05,  2.749269066421e-06, {   4,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.055138361936e-05,  9.072459062448e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.043516480574e-06, -3.755581159463e-05, {   0,   1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.647974369510e-05,  1.305924678486e-05, {   2,   1,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.864398169681e-05, -8.900798239215e-07, {   4,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.880720055701e-05,  7.078183933680e-07, {   2,   1,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  1.683414554265e-05, -2.267663315864e-05, {   1,   0,   0,   0,   0,   0,  -6,  13,   0,   0,   0,   0,   0 } },
 { -3.328355327612e-05,  6.203861061615e-06, {   3,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.690001150751e-05, -2.574465927910e-06, {   1,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.738649565835e-05,  2.207161884665e-05, {   2,  -1,   0,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  3.163766735416e-05, -7.817708211408e-06, {   0,   1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.026460823874e-05,  2.908099656397e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  1.008139531758e-05, -2.922340172750e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  2.507783844504e-05,  1.419947131925e-05, {   4,   1,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.729481120427e-05,  2.196377200167e-05, {   2,  -1,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -3.875558704503e-05,  4.203037136754e-07, {   2,  -1,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  7.885635036125e-06,  3.121350557733e-05, {   3,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.390209165696e-06, -3.070682503303e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -7.646082807355e-06,  3.142731790568e-05, {   2,   1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.781258045162e-05,  1.223641127076e-06, {   4,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  3.454471547752e-05,  4.466566180612e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.388385672814e-06, -3.161191190652e-05, {   3,  -2,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -7.384831812182e-06,  3.157207904461e-05, {   1,   0,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.851068741278e-05,  1.042170278077e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -1.854659902925e-05,  2.036926544917e-05, {   2,   1,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.521488530516e-05,  2.365933684713e-05, {   1,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.806088302626e-05,  1.079499484692e-05, {   1,   0,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -9.525355078886e-06, -2.926821473174e-05, {   1,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.522772428697e-06, -3.426580652049e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  5.534020762297e-06,  3.319409289676e-05, {   0,   1,   1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.741680076103e-05, -2.130638656197e-05, {   1,   2,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.582377971844e-05,  1.289773802379e-05, {   2,   1,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.424732430834e-05, -4.463445433984e-06, {   2,   1,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -2.264518405608e-06,  3.643702100609e-05, {   2,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.989497271655e-05, -8.780238968525e-06, {   4,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.814536193938e-06, -2.886043954745e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -6.684357026249e-07, -3.796532975913e-05, {   2,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.764167612378e-05, -2.097507761957e-05, {   1,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.810421248054e-05, -4.424752156846e-07, {   2,  -1,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.843915026788e-06,  3.667579357329e-05, {   0,   1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -3.844088743038e-05,  6.778435566814e-08, {   2,   1,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.344191376047e-05,  2.506516840232e-05, {   1,  -2,   0,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.533454270375e-05,  1.316534601707e-05, {   2,  -1,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 { -1.145994794848e-05,  2.703622607634e-05, {   2,   1,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.314107044878e-05, -1.535259068075e-05, {   4,  -1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.474967786296e-05, -3.723687621287e-06, {   2,  -1,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -7.219720069664e-06,  3.124089924572e-05, {   1,   0,  -1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.812813049119e-05,  2.033246342513e-05, {   2,  -1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.740840134861e-05,  2.104057041719e-05, {   4,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.628871757247e-05,  1.215671710836e-05, {   0,   1,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  3.448871658123e-05,  3.932868444787e-06, {   1,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.925101775196e-05, -1.916012801871e-05, {   0,   1,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -1.136861786192e-05,  2.703394736494e-05, {   1,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.479846971019e-05, -1.360162880465e-05, {   2,  -1,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.274187660288e-06,  3.606534750423e-05, {   2,  -1,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  2.576872043394e-05,  1.256356322277e-05, {   0,   1,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.714406708567e-05, -2.116786526814e-05, {   1,   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  1.498322297428e-05,  2.332113080620e-05, {   1,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.573901172582e-06, -3.271221008424e-05, {   1,   0,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  2.615822020030e-05,  1.210750562486e-05, {   0,   1,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.482662624157e-05, -3.423672208046e-06, {   2,  -1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.820899867863e-05,  8.879209454459e-09, {   1,   1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.548359176108e-05,  1.273190300863e-05, {   0,   1,  -1,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.998820917229e-05, -1.822315145393e-05, {   2,   1,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -2.319502657991e-05,  1.500346920214e-05, {   1,   0,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -2.246008907585e-05, -1.570722151003e-05, {   2,  -1,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  9.213728256798e-06,  2.895061582616e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  2.460916567067e-05, -1.351539055921e-05, {   4,  -1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.562752181076e-05,  1.249666335020e-05, {   0,   1,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.728745112631e-06,  3.439320703589e-05, {   2,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.640101718430e-05,  1.714298614602e-06, {   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.696831392636e-05, -1.142978408077e-06, {   2,  -1,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  6.493172748687e-06, -3.159337332343e-05, {   1,  -2,   1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  6.493172748687e-06, -3.159337332343e-05, {   1,  -4,  -1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -9.189401444914e-06,  2.887896596911e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -3.711446731750e-05, -9.444330325739e-07, {   2,  -1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.626599682622e-05,  2.175928570555e-05, {   1,   0,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.801589194579e-05, -2.590435151819e-09, {   0,   1,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.448038064724e-05,  1.349588600283e-05, {   4,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -3.468153165942e-05,  3.260275529713e-06, {   4,  -1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.685534961908e-05,  2.107392810902e-05, {   2,   1,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  1.419952674042e-05, -2.372514877306e-05, {   0,   1,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -9.935586279919e-06, -2.798363898090e-05, {   0,   1,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.583187205140e-05,  1.206715989181e-05, {   2,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.848706058118e-05,  9.406773062223e-06, {   0,   1,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.924906421976e-06, -2.795191052162e-05, {   0,   1,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.535173135531e-05,  1.250126569222e-05, {   2,  -1,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.877130507169e-05,  9.081529823496e-06, {   3,  -2,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -3.359195527366e-05, -4.252202966158e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 { -7.303914906081e-06,  3.049220645081e-05, {   4,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.097307374374e-06, -3.067844719547e-05, {   0,   1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -6.578400099358e-06, -3.118534720765e-05, {   0,   1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  5.701340052793e-06,  3.206148211419e-05, {   2,   1,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.213961381083e-05,  1.560206457805e-05, {   2,  -1,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  3.602400365291e-05, -1.668165428715e-06, {   1,   0,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.758267844932e-05,  1.069212754575e-07, {   2,   1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.555746457410e-05,  2.212568972811e-05, {   3,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.236448618593e-05,  5.309615920567e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -6.395117384628e-06, -3.125604649005e-05, {   0,   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.512775982501e-05, -1.246787640283e-05, {   1,  -3,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.018549194488e-06, -3.055729104599e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -1.639133772813e-05, -2.118288294692e-05, {   0,   1,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.752220863786e-05, -2.887641071012e-08, {   2,  -1,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.047055871739e-06, -3.644814429166e-05, {   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.784993291039e-05, -9.642598621316e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -7.172173239586e-06,  3.027085454388e-05, {   3,   0,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.495542932044e-05,  1.246325858177e-05, {   2,   1,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.559786094841e-05,  2.181830173020e-05, {   1,  -2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.406115605520e-05, -2.333611369955e-05, {   3,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.552452049892e-05,  2.185919809771e-05, {   1,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.428838344471e-05,  1.309388953667e-05, {   2,   1,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  6.079150482158e-07,  3.675710897784e-05, {   2,  -1,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.584641816644e-05,  1.151069528955e-05, {   2,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.308293170249e-05, -4.239878123701e-06, {   1,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.576200462050e-05, -2.155753273924e-05, {   2,   1,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.359510275324e-05,  3.705074066369e-06, {   2,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.199847902186e-05, -5.232592017752e-06, {   2,   1,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -3.716081189382e-05, -4.206253488825e-08, {   2,   1,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.117149833229e-06,  3.007724138716e-05, {   4,  -1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  3.705872683597e-05,  1.310903348355e-07, {   2,   1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.334045165322e-05, -2.382363596751e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  1.331510967480e-05, -2.383741547657e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  2.122155877929e-05,  1.590413481614e-05, {   1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0,   0,   0 } },
 { -3.706203333174e-05,  5.847435679145e-08, {   4,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.180056274635e-06,  2.892859473500e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.705858581852e-05, -4.394757829595e-08, {   2,   1,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.576463366852e-05,  1.131538745817e-05, {   1,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.451087047658e-06, -3.462723726989e-05, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  8.315925815996e-06, -2.874262899770e-05, {   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.993367171476e-06, -2.802603417355e-05, {   2,   1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.591367264792e-05, -1.104140844115e-06, {   4,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -4.664591735478e-06, -3.233538162199e-05, {   2,   1,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  9.516027312523e-06,  2.747291555824e-05, {   2,  -1,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.792632207778e-05,  9.061723106678e-06, {   2,  -1,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.922735615685e-06, -3.505195169771e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.332032551647e-06,  3.562191878226e-05, {   0,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  4.628229889090e-06, -3.229277238493e-05, {   2,   1,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  1.874909012713e-06, -3.501575742916e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -9.856489596448e-06, -2.703236976446e-05, {   2,  -1,   0,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 {  2.696527532324e-05, -9.918243936286e-06, {   2,  -1,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.366797569881e-05, -1.319524003216e-05, {   1,  -1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.523847538914e-05,  1.161376756037e-05, {   1,   0,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.158046127713e-06,  3.266978832518e-05, {   2,  -1,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  2.360687791835e-05, -1.321921706711e-05, {   2,   1,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.614937710549e-05,  6.697006474963e-07, {   2,   1,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -8.109535344651e-06,  2.867514643292e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.594594309449e-05,  2.083144410757e-05, {   1,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  1.645949813849e-05, -2.029456346538e-05, {   1,   0,   0,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.127552951590e-07, -3.625623903314e-05, {   3,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.578036028675e-05,  2.086837391945e-05, {   2,  -1,   0,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -2.633332980060e-05,  1.030075556556e-05, {   2,  -1,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.577055484909e-05,  2.085454151352e-05, {   2,  -1,   0,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -2.337906377013e-05,  1.322978709318e-05, {   1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.356332399232e-05, -2.303948019604e-05, {   0,   1,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.566688809561e-05, -1.093567117004e-05, {   0,   1,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.268659664042e-05, -3.897854961211e-06, {   1,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.042068347044e-05,  6.150076306160e-06, {   2,  -1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.073839655132e-05,  5.823909396361e-06, {   2,  -1,   1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  8.807991531735e-06, -2.774081737472e-05, {   2,   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -6.970546846857e-06,  2.952491254337e-05, {   1,   0,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.570641219669e-05, -1.078330721011e-05, {   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.821215447570e-05, -8.265435823546e-06, {   2,  -1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.523422845246e-05, -1.123012082240e-05, {   1,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.434035865686e-05,  1.211707200811e-05, {   0,   1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.357364805709e-05, -1.287914939386e-05, {   2,   1,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  3.631646847827e-05,  1.283026054914e-07, {   2,   1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.939197009261e-06, -2.946010518386e-05, {   1,   0,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.066235636056e-05, -5.723111625549e-06, {   0,   1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.948864926512e-05,  6.893435002444e-06, {   2,  -1,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  7.687604062879e-06,  2.869417720796e-05, {   1,  -2,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.301642404108e-06, -3.506788238711e-05, {   2,  -1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.117135133505e-05, -1.518808308339e-05, {   2,  -1,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  5.228616620052e-06, -3.111613835452e-05, {   1,   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -2.073190903470e-05, -1.559746440864e-05, {   0,   1,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.812899886682e-05,  8.196784616617e-06, {   0,   1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.741454661915e-05,  8.869663693972e-06, {   4,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.769606031195e-06,  2.951149163224e-05, {   1,   2,  -1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.214731280903e-05, -1.411966503326e-05, {   0,   1,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.425753631364e-05,  1.964232931943e-06, {   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.661231229509e-05, -9.604940133119e-06, {   0,   1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.983809345515e-05,  1.635533717761e-05, {   1,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.146850570637e-05, -2.470804370433e-05, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -6.845296120815e-06,  2.932045343138e-05, {   1,   0,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.329081651058e-05,  1.279583421828e-05, {   1,   1,   2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.329081651058e-05,  1.279583421828e-05, {   1,  -1,  -2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.339616541361e-05, -1.266398225802e-05, {   2,  -1,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -6.888143748428e-06,  2.914300590164e-05, {   1,  -2,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.476936516415e-05,  1.125563204920e-05, {   0,   1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.891317607875e-06, -2.910951157113e-05, {   1,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.593923103901e-05,  5.802106234284e-08, {   0,   1,   1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.506011138196e-07,  3.569542118375e-05, {   2,  -1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.301254988542e-08, -3.592410394668e-05, {   4,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.582070032532e-05, -1.159834360502e-07, {   2,   1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.194711765907e-05, -1.398712751623e-05, {   0,   1,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.870356931321e-06,  2.904897790809e-05, {   1,   0,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.492567497150e-06,  3.241077535339e-05, {   2,  -1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.509565741417e-06,  2.837872946651e-05, {   2,  -1,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.538495426475e-05,  1.050050137537e-05, {   4,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.456291340543e-06, -2.842604571080e-05, {   0,   1,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -2.293300071053e-05, -1.290943582477e-05, {   0,   1,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.321289165690e-05,  1.260675384969e-05, {   0,   1,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.148728073838e-05,  2.432899937744e-05, {   2,  -3,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -3.541742764846e-05,  3.930274998548e-07, {   2,   1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  9.926104120201e-06, -2.583052981729e-05, {   1,   2,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  2.003566141311e-06, -3.374995739294e-05, {   2,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.582822647477e-05,  1.989781341026e-05, {   2,   1,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  3.429958780166e-05,  1.424484262106e-06, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -3.463996986755e-05, -1.074485027819e-06, {   2,   1,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.013628064595e-05,  2.556660521164e-05, {   4,  -1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -5.735002233559e-06, -2.995052196165e-05, {   1,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -6.826758648138e-06, -2.879334387252e-05, {   3,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.518809317744e-05,  4.303642421364e-07, {   0,   1,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.916172593917e-06, -2.866146288510e-05, {   3,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.430224369122e-05,  1.126660736122e-05, {   2,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.406308317767e-05, -1.149885381143e-05, {   2,   1,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.772630944726e-06,  2.872299919578e-05, {   1,   0,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.819543956063e-05, -7.299578247559e-06, {   2,  -3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.180415345864e-05,  1.368272035487e-05, {   2,   1,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -1.836692262430e-06, -3.360211870572e-05, {   2,   1,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  3.532016350390e-05,  1.098031205274e-07, {   3,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.703084813109e-05, -8.391392697001e-06, {   0,   1,   2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.070070687456e-05, -2.468689150601e-05, {   2,   1,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -2.273222483342e-05, -1.265450384018e-05, {   1,  -1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -3.532416947743e-05,  4.686034370282e-08, {   0,   1,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.347873349259e-05,  1.186509549320e-05, {   1,   0,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -3.497320895358e-05, -3.364514228101e-07, {   2,  -1,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -3.482298264823e-05, -4.816483548917e-07, {   1,   1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.620142655013e-05, -9.094139185152e-06, {   2,  -1,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -2.168271493701e-05,  1.360427406041e-05, {   2,   1,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.780139165602e-05,  1.743637714051e-05, {   2,  -1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.805485841186e-05,  7.101502797816e-06, {   2,  -1,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.509352465203e-05,  5.867806398791e-08, {   0,   1,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.499733626617e-05,  1.508366027936e-07, {   0,   1,  -1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.293571401275e-05,  2.195885714627e-06, {   1,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  9.144564615179e-06, -2.596185201782e-05, {   1,  -2,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -7.971783112050e-06,  2.713102384177e-05, {   2,   1,  -1,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 { -7.971783112050e-06,  2.713102384177e-05, {   2,  -1,  -1,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 { -3.092858713171e-05, -4.165568810752e-06, {   0,   1,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.012633574690e-05, -2.496501763704e-05, {   0,   1,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.312596160926e-05, -2.195534394941e-05, {   0,   1,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -1.011577559123e-05, -2.494267587120e-05, {   0,   1,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -3.497563797027e-05,  6.674233850971e-08, {   2,   1,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.976339784090e-05, -1.524067878192e-05, {   2,   1,  -1,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 { -1.976339784090e-05, -1.524067878192e-05, {   2,  -1,  -1,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 {  1.343882538353e-05, -2.155106697785e-05, {   4,   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.057385795262e-05, -2.441105485706e-05, {   2,   1,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.400313464765e-05, -2.093588323836e-05, {   4,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.190225295932e-05, -1.299633283417e-05, {   1,   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  4.208948603600e-06, -3.068759669234e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.331665497913e-05, -2.155836005376e-05, {   2,   1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.904896235642e-05, -5.823400232501e-06, {   3,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.567502359123e-06,  2.625732628621e-05, {   1,   0,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.423092063916e-05,  5.932653995239e-07, {   4,  -1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.880882284817e-07, -3.463215751041e-05, {   1,   1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.744301035421e-05,  7.364147826889e-06, {   1,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.477344041459e-05, -3.133974291299e-08, {   2,  -1,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.255152889347e-05, -1.224203367823e-05, {   2,  -1,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -3.166550324629e-05, -3.100490072466e-06, {   1,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -6.710669583338e-06,  2.801850278842e-05, {   3,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.403885130196e-06, -3.231305126604e-05, {   1,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.887691652155e-05, -1.580963261049e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -7.711688217867e-07, -3.387556098540e-05, {   2,   1,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.483329593733e-05, -1.980732135480e-05, {   0,   1,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -6.668601472050e-06, -2.794219245426e-05, {   1,  -2,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.144951059454e-05, -3.160701397634e-06, {   2,   1,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -3.033278086604e-05, -4.179491191341e-06, {   1,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.305559171238e-05, -2.144115007089e-05, {   0,   1,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  2.386434392114e-05,  1.060698667488e-05, {   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.180913711142e-05,  1.265802570082e-05, {   2,  -1,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  3.443324850496e-05,  2.656867173408e-09, {   4,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.400039959373e-05,  4.324836245274e-07, {   3,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.434313512238e-05, -1.006031721588e-05, {   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -2.761133595443e-05, -6.775416505368e-06, {   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.437198386812e-05,  1.256580200141e-08, {   1,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.268382179447e-05, -2.169484624433e-05, {   0,   1,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -3.405235227882e-05, -3.174581842304e-07, {   2,   1,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.587000253804e-06, -2.778137298949e-05, {   3,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.703143543531e-06, -2.763845410227e-05, {   1,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.880379073729e-05, -5.521579666435e-06, {   1,   2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.427642281635e-05,  4.249766446456e-08, {   0,   1,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.428593362627e-05,  6.330757175140e-09, {   2,  -1,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.331668492261e-05, -2.095964144567e-05, {   2,  -1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.667643251439e-05, -1.756201352961e-05, {   0,   1,   2,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 {  3.415844391808e-05,  5.888083151521e-08, {   0,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -6.491920678577e-06,  2.771324483901e-05, {   4,  -1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.192222433172e-05, -1.228207896627e-05, {   0,   1,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.822984832043e-05, -5.947629508965e-06, {   2,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.572960835425e-05, -1.844225708938e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -1.057595383532e-05, -2.358437830626e-05, {   1,   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -1.570149025924e-05, -1.839507297886e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  1.334872908298e-05,  2.074404029219e-05, {   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  1.458152945689e-05, -1.949688338355e-05, {   1,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -3.407579914134e-06, -3.063571222365e-05, {   2,  -1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.028866764807e-05, -1.373162942512e-05, {   2,   1,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.422571211276e-06,  2.756382683852e-05, {   1,   0,  -1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  7.628141305390e-06, -2.631058546691e-05, {   1,   0,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.102077069606e-05,  2.875761283791e-06, {   0,   1,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.691695848601e-05, -1.697921167932e-05, {   0,   1,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -9.725882570745e-06,  2.416910004972e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -6.483252008888e-06, -2.736947784066e-05, {   1,   0,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.050164852091e-05, -2.335059538479e-05, {   0,   1,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -5.338591482230e-06,  2.848911976057e-05, {   0,   1,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.358928238492e-05,  1.609231518838e-07, {   1,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.854003514214e-05, -1.520481177450e-05, {   4,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.534606766006e-06, -3.219056385454e-05, {   2,  -1,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  2.336662612982e-05,  1.035536587822e-05, {   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.369003887382e-05, -2.688425397070e-08, {   2,  -3,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.320229452929e-05,  2.049746027550e-05, {   1,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.367946404681e-06, -3.231725781441e-05, {   2,  -1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.243805622237e-05,  1.124382646526e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -4.015443812794e-06, -2.963505568487e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.135894818293e-05,  2.252198455652e-06, {   4,   1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.822505075653e-06,  2.977366698135e-05, {   2,  -1,   0,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 { -1.367639904333e-06,  3.222182619440e-05, {   2,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.314712014421e-05, -2.044212000331e-05, {   4,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.309573057806e-05, -4.837029251130e-07, {   2,  -3,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  3.355266652336e-05,  2.439132051116e-08, {   2,   1,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.082696098745e-06, -2.945185677403e-05, {   0,   1,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.352324816703e-05,  9.994994815481e-06, {   0,   3,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.352324816703e-05,  9.994994815481e-06, {   0,   3,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.408603192441e-05, -1.941939657279e-05, {   2,   1,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.388147305510e-05,  9.575981518719e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  2.903030201126e-05, -4.405912048337e-06, {   1,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.492658185925e-05, -1.850001601227e-05, {   1,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 {  2.254354996855e-05,  1.087249153616e-05, {   2,  -1,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.102277268380e-06, -3.230084623782e-05, {   0,   1,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.299191754139e-06,  2.710021734892e-05, {   1,   0,  -1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.227250263585e-05, -1.112406775955e-05, {   2,   3,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.227250263585e-05, -1.112406775955e-05, {   2,   3,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.405021973029e-06, -2.597467171345e-05, {   2,  -1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.657656059737e-05,  6.798044264022e-06, {   1,   0,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.745430367676e-05, -5.906169317223e-06, {   1,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.662178201999e-05, -1.673808623603e-05, {   0,   1,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  6.389974243807e-06, -2.694268910959e-05, {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.801935139170e-05, -5.310608486400e-06, {   1,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.623267268883e-05, -1.708831488343e-05, {   0,   1,   0,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 {  3.278978576491e-05,  5.193174253505e-07, {   2,  -1,   0,   0,   0,   0,   2,   0,   1,   1,   0,   0,   0 } },
 { -1.303891032977e-05, -2.026900948753e-05, {   4,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.083909930247e-05,  1.245872276272e-05, {   1,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.706100532935e-06, -3.156343848622e-05, {   1,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 { -2.227762238962e-05,  1.098187075752e-05, {   2,   1,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.746644050239e-05,  5.791243277476e-06, {   1,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -3.314383683422e-05,  7.437011434961e-08, {   2,  -1,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.360529350877e-06,  2.684894486307e-05, {   1,   0,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.000133832047e-05, -3.197019523463e-06, {   2,  -1,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.212505076680e-05,  2.104445139388e-05, {   2,  -1,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.298101496812e-05,  2.017376392759e-05, {   1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.560813320615e-06,  2.659270164570e-05, {   1,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.251277180515e-05,  6.129704179384e-07, {   2,  -1,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.119054776391e-05, -1.191966041573e-05, {   2,   1,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.441986099419e-06, -2.662262317151e-05, {   4,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.198278781882e-06, -2.585333423942e-05, {   4,  -1,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.204163329579e-05,  1.100810884432e-05, {   4,  -1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.860026301122e-05,  4.407761616998e-06, {   0,   1,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  2.859677862789e-05,  4.406718117048e-06, {   0,   1,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -1.480484629620e-05,  1.817768023598e-05, {   1,   0,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -3.296688434341e-05,  1.331359735074e-09, {   2,   1,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.434012949607e-05, -1.861679949099e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  7.176756393514e-06, -2.577562358311e-05, {   4,  -1,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.159377541611e-05, -1.358554290670e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  6.300439636873e-06,  2.664598112843e-05, {   1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.196523208747e-05, -1.097060837635e-05, {   3,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.196523208747e-05, -1.097060837635e-05, {   3,  -1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.205783764896e-05,  1.087483096566e-05, {   4,   1,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.432268044423e-05, -1.860452755034e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -8.929692985452e-07, -3.202088849828e-05, {   2,  -3,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.283767044343e-05, -6.568787466250e-08, {   2,  -1,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.879895716434e-05,  1.409263280021e-05, {   1,   0,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  1.157588896108e-06, -3.166768883455e-05, {   2,   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.255772019448e-05, -2.463687474943e-07, {   2,  -1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.197143975764e-05,  1.083247248142e-05, {   4,   1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.245210226401e-05,  3.387006216629e-07, {   2,  -3,   0,   0,   0,   0,   2,   0,   1,   1,   0,   0,   0 } },
 { -2.985773743588e-05, -2.919866298377e-06, {   0,   1,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -3.028741691584e-05, -2.488828875678e-06, {   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -9.062927816623e-06,  2.369339229848e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.789027890288e-06, -3.096154011213e-05, {   2,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.252598036061e-06, -2.647864124627e-05, {   3,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.576062992869e-06, -2.415430654669e-05, {   0,   1,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -9.779164319441e-08,  3.263223300032e-05, {   4,   1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.260151691157e-06, -2.644293810076e-05, {   1,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.706434506135e-05, -1.562764955570e-05, {   2,   1,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -8.063850184353e-06, -2.462372318559e-05, {   2,  -1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.046420887550e-05,  1.221997115602e-05, {   2,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  3.267609580211e-05,  4.375916118150e-09, {   2,   1,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.694599745739e-05, -1.572165826891e-05, {   4,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  9.291207614278e-06,  2.337066408369e-05, {   1,   0,  -1,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -2.987094798921e-05,  2.779652177188e-06, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.352802332009e-06,  2.329649340889e-05, {   4,   1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -8.552943166037e-06, -2.408806859970e-05, {   0,   1,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.103097931229e-05, -1.158610886699e-05, {   2,   1,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.782065835253e-05,  4.788558671909e-06, {   3,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.696623175831e-06, -3.090829672747e-05, {   2,   1,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -2.685311271922e-06,  2.987957669333e-05, {   3,  -2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.684878429137e-06,  2.987961559024e-05, {   1,  -2,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.683663337343e-06, -2.987972474926e-05, {   5,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.683663337343e-06,  2.987972474926e-05, {   3,   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.683663337343e-06,  2.987972474926e-05, {   3,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.683663337343e-06, -2.987972474926e-05, {   3,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.683663337343e-06, -2.987972474926e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.683663337343e-06,  2.987972474926e-05, {   1,   2,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.760046078684e-06,  2.576604178373e-05, {   1,   0,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.966343570787e-05,  2.854296119733e-06, {   1,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -2.983919253013e-05,  2.660561175852e-06, {   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.679714868158e-06,  2.579392161915e-05, {   1,   0,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.244911214132e-05, -2.181714643987e-08, {   2,  -3,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.230706974645e-05,  1.139187912259e-07, {   2,  -1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.087871678315e-05,  1.153870853259e-05, {   2,  -1,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  1.650189313334e-05, -1.590633260096e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -1.180916156007e-06, -3.115957821751e-05, {   0,   1,   2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.384951743192e-06, -3.093597719088e-05, {   2,  -1,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.043606779512e-05, -1.184129310991e-05, {   0,   1,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.597920667279e-05,  6.289758679993e-06, {   1,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.531446524977e-05, -6.950950275074e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  5.925648490951e-06, -2.633386021627e-05, {   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.857294530589e-05,  3.675317156298e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  3.166926272898e-05,  5.759204237061e-07, {   2,  -1,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -3.614689350689e-06, -2.861849281556e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  1.738383414714e-05,  1.483874018701e-05, {   1,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.246519001600e-05, -9.713158913379e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  1.746708437309e-06,  3.041931479378e-05, {   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.212109982148e-05,  1.518354131367e-08, {   2,  -1,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.087865444684e-06, -2.603648974951e-05, {   2,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.816553754188e-06,  2.529699777839e-05, {   1,   2,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.489183202124e-05,  7.219145484300e-06, {   2,  -1,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  2.149943307374e-05,  1.060350908753e-05, {   0,   1,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.072681782654e-05, -1.137292429086e-05, {   2,  -1,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  3.004258879133e-05,  2.039109338164e-06, {   1,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.671167639577e-05, -1.534791540845e-05, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  1.050502506590e-05,  2.154806350657e-05, {   2,  -1,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.695089590070e-05,  5.101814613114e-06, {   2,  -3,   1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  6.150105060569e-06,  2.588931807998e-05, {   1,   2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.628129040316e-06,  3.037457884557e-05, {   2,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.297500452001e-06,  3.070252896080e-05, {   2,   1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.048247787069e-05,  1.510760821769e-06, {   2,   1,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  3.048247787069e-05,  1.510760821769e-06, {   2,   1,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -3.192716828584e-05,  5.102024904417e-08, {   2,   1,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.140790441879e-05,  5.626746882615e-07, {   2,  -1,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -5.938748794370e-06,  2.602654641097e-05, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.785620802288e-05, -1.409222612710e-05, {   0,   1,   1,   0,   0, -23,  22,   4,   0,   0,   0,   0,   0 } },
 { -1.785620802288e-05, -1.409222612710e-05, {   0,   1,  -1,   0,   0,  23, -22,  -4,   0,   0,   0,   0,   0 } },
 {  2.633926313706e-05,  5.599606153187e-06, {   3,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.232160467761e-05,  1.961025290718e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -2.044532754826e-05, -1.148336650116e-05, {   0,   1,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -8.034459453468e-07, -3.111877004894e-05, {   0,   1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.440593784394e-05, -1.749282440004e-05, {   2,  -1,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.175618771558e-05,  1.324933222816e-07, {   3,  -1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.993734109869e-05,  1.937939096279e-06, {   2,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.171555849323e-05, -1.590416768219e-07, {   4,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.046089294174e-05,  1.407323271372e-06, {   4,   1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.586639476015e-05, -1.600077287816e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -2.866416845969e-05, -3.195157366323e-06, {   1,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.759094347723e-05, -4.247310193026e-06, {   0,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.339250780730e-05, -1.843631625452e-05, {   2,  -1,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.263982198873e-06,  2.554185080584e-05, {   2,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.478192778742e-05, -7.009927242165e-06, {   4,   1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.166977251893e-05, -2.011165903370e-05, {   2,   1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.757504560180e-05, -4.199125193443e-06, {   2,   1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.764516198426e-05, -4.123611506193e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -1.615041835925e-05,  1.561638926780e-05, {   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.641387148739e-05, -1.530267064785e-05, {   0,   1,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -2.945558133949e-05, -2.260452774590e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -2.245953855836e-05, -9.252322475864e-06, {   2,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.065861476435e-06, -2.564256883271e-05, {   3,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.478864636606e-05, -6.912864716169e-06, {   3,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.586081203438e-05, -1.583529091198e-05, {   2,   1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.039675011822e-06, -2.665216974296e-05, {   3,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -4.961900221272e-06,  2.672088173639e-05, {   1,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.161864862643e-05,  1.343673689478e-08, {   2,   3,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.459946607924e-05, -1.703142244613e-05, {   1,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.235744115767e-05,  9.235117534238e-06, {   2,  -1,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -2.607524825479e-05,  5.496033453887e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -2.674047467279e-05, -4.783417310765e-06, {   1,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.673448166137e-06,  2.783602489017e-05, {   2,  -1,   0,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -2.601081344687e-05, -5.466221123824e-06, {   1,   0,   0,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.931737687476e-06, -2.553290987073e-05, {   2,  -1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.215909136379e-05, -1.927394206611e-05, {   0,   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -2.019437724563e-05, -1.115912840130e-05, {   2,   1,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  2.709893330350e-05, -4.248843678581e-06, {   2,   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.455081473469e-06, -2.987270732980e-05, {   2,   1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.970584079617e-05, -1.161523989123e-05, {   0,   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.930718104859e-05,  2.011490114009e-06, {   4,  -1,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.083036660876e-05,  4.779208431169e-07, {   1,  -1,   0,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 {  3.083036660876e-05,  4.779208431169e-07, {   1,  -3,   0,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 {  2.376992475515e-05, -7.496046926402e-06, {   1,   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.118276472496e-05, -4.133065266308e-08, {   2,   1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.933063263575e-05,  1.891097822828e-06, {   1,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.882963031522e-05, -1.238399662457e-05, {   4,   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -9.539605045803e-06, -2.165523805542e-05, {   0,   1,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -8.728421046455e-06,  2.244473853394e-05, {   2,   1,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.635455069053e-05, -1.480354650293e-05, {   2,   1,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -4.504015398655e-06,  2.663560667910e-05, {   1,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  5.970014283514e-06, -2.516813303390e-05, {   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.112350604885e-05,  6.873466944203e-09, {   4,  -1,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.105823647484e-05,  4.701104209136e-11, {   4,  -1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.401383990465e-05,  1.701447841223e-05, {   4,   1,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -9.547728110115e-06, -2.147427893364e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.437461376178e-05,  6.646196304017e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -2,   0 } },
 { -1.268796164758e-05,  1.832818295553e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -5.892169662963e-06, -2.510691185134e-05, {   4,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.929783696208e-06, -2.506063408086e-05, {   1,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.202710634395e-05, -1.892076848619e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.038422105865e-05, -2.055860412194e-05, {   2,   1,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  2.473433248587e-05, -6.205160198046e-06, {   1,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.374215904708e-05,  1.719002093916e-05, {   2,  -1,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  2.437579323107e-06, -2.849140573030e-05, {   2,  -1,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  2.506758512859e-05,  5.858535097961e-06, {   1,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -3.083373391817e-05, -8.701064969998e-08, {   0,   1,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.044283169622e-05,  2.043986578167e-05, {   1,   0,   0,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  2.973763363130e-06,  2.790256682657e-05, {   4,   1,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.768230463987e-05,  1.317253851995e-05, {   2,  -1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -5.921327250083e-06,  2.486395310989e-05, {   1,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.365274968543e-05, -1.713237592227e-05, {   2,   1,  -2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.584969840556e-05,  1.493102195052e-05, {   2,   1,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.525570268730e-05, -5.514897873090e-06, {   3,  -1,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.859560731752e-05, -2.159106862920e-06, {   4,   1,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -8.207545691432e-06, -2.252445713324e-05, {   2,  -1,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.938728862015e-06,  2.475397172832e-05, {   4,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.068688708953e-05,  1.375065867132e-10, {   4,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.263588087820e-07, -2.984950406144e-05, {   4,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.304593865060e-06,  2.536608445165e-05, {   2,  -1,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  3.058209710188e-05, -7.929810207930e-08, {   2,  -1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.249950412243e-06,  2.941027802324e-05, {   4,  -3,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.708702567879e-06, -2.495054109574e-05, {   1,   0,  -1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.917307429673e-05, -1.481067548673e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -3.703982656845e-06,  2.692671969985e-05, {   2,   1,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -1.958749559161e-05,  1.103832473617e-05, {   1,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.955117012977e-06, -2.366632076185e-05, {   2,   1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.770072722888e-05, -1.288489036494e-05, {   1,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.791218782503e-05,  1.265748159680e-05, {   1,   0,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -4.857806056137e-07, -3.005435217681e-05, {   2,  -1,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.183843701090e-05,  1.870149526098e-05, {   2,  -1,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 {  2.797968588858e-05, -2.554811443420e-06, {   2,  -1,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -5.826992889697e-06,  2.466544598622e-05, {   3,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.942547737640e-05,  1.038582016193e-06, {   1,   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -2.994745628206e-05, -4.631029516559e-07, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  2.994227595275e-05, -4.626053705488e-07, {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -2.435348729767e-05, -6.049491269988e-06, {   0,   1,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.392380371407e-06, -2.400697135277e-05, {   1,   2,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -2.718401712425e-05,  3.195043572669e-06, {   1,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.433823396072e-05, -6.022921966607e-06, {   0,   1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.823473668832e-06, -2.453247448822e-05, {   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.803949194780e-06,  2.453710049647e-05, {   1,   0,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.425762975821e-05, -6.065780624447e-06, {   1,   2,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.373774276409e-06, -2.894669782891e-05, {   2,  -1,   0,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  5.811742986307e-06, -2.450634282970e-05, {   2,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.472751125769e-05, -5.577928058324e-06, {   4,   1,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.709255082186e-06, -2.459028476603e-05, {   1,   0,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -9.820233525797e-07, -2.931052069572e-05, {   0,   1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  9.806033200587e-06, -2.046939601419e-05, {   2,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.934587167532e-07, -2.938078666687e-05, {   0,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.811648719033e-05,  1.215016823766e-05, {   1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -2.434015015427e-05,  5.908322901490e-06, {   1,   0,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  7.174091227529e-06, -2.306764501179e-05, {   2,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.152547690139e-05, -8.714982635439e-06, {   2,  -1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.170195870339e-05, -8.519763980094e-06, {   1,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 {  9.967701902365e-06,  2.025042377326e-05, {   2,  -1,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -6.027241530727e-06,  2.416857202001e-05, {   2,  -1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.013430452418e-05, -1.005612117432e-05, {   2,  -3,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.591033670482e-05,  4.262694193442e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -8.200422438254e-06, -2.195274119648e-05, {   2,  -1,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  3.013342165235e-05,  9.322035796900e-09, {   1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.746198374360e-05, -2.665392490290e-06, {   1,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.100561973250e-05, -1.912054927390e-05, {   4,  -1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.230909582196e-05,  1.778469201851e-05, {   1,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.351778376842e-05, -1.653638519474e-05, {   1,   0,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.298273437831e-05,  1.704658689139e-05, {   1,   0,   0,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.442880570809e-06,  2.157431260686e-05, {   3,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -8.850297386564e-06, -2.115320131864e-05, {   0,   1,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.999999997302e-05, -1.272345024324e-09, {   2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999997541e-05, -1.214749159055e-09, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.999999998042e-05, -1.083849465254e-09, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999998986e-05, -7.801621755482e-10, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.999999999331e-05,  6.335545184268e-10, {   0,   2,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999385e-05, -6.073745796519e-10, {   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.999999999417e-05, -5.916666163821e-10, {   2,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999427e-05, -5.864306286254e-10, {   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999534e-05, -5.288347633311e-10, {   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999638e-05,  4.660029102637e-10, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.999999999646e-05, -4.607669225058e-10, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.999999999654e-05, -4.555309347488e-10, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.999999999670e-05, -4.450589592496e-10, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.999999999685e-05,  4.345869837314e-10, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.999999999715e-05,  4.136430327095e-10, {   4,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999722e-05,  4.084070449541e-10, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999722e-05, -4.084070449498e-10, {   2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.999999999729e-05,  4.031710571986e-10, {   4,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999736e-05, -3.979350694503e-10, {   2,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999736e-05, -3.979350694503e-10, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999736e-05,  3.979350694430e-10, {   2,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999770e-05,  3.717551306653e-10, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999770e-05,  3.717551306653e-10, {   2,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999776e-05,  3.665191429097e-10, {   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.999999999824e-05, -3.246312408635e-10, {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.999999999836e-05, -3.141592653488e-10, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.999999999916e-05,  2.251474735022e-10, {   2,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.069931220799e-06, -2.790110924300e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 {  5.886780041168e-06,  2.407185457902e-05, {   3,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.414156961597e-05,  1.578396055611e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 {  1.414156961597e-05,  1.578396055611e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 { -2.073932704390e-05, -9.184175791846e-06, {   1,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.099921420727e-05,  1.891621823174e-05, {   2,  -1,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -2.935594897710e-05,  5.422318816797e-07, {   2,  -1,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.100161487453e-05,  1.885722504083e-05, {   2,  -1,   0,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -1.462814756480e-07, -2.970681465904e-05, {   2,  -1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.836569803545e-05, -1.483672150389e-06, {   2,   1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.937694634358e-05, -4.642203938520e-07, {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -2.936983427015e-05, -4.634929942586e-07, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 { -1.339774905348e-05,  1.641109741952e-05, {   2,   1,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.737533810149e-06, -2.706572179298e-05, {   1,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  5.361527385694e-06, -2.443723103169e-05, {   2,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  1.107681010631e-06, -2.869088846888e-05, {   2,   1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.942994560039e-05,  3.221944613329e-07, {   2,   1,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.109413733732e-05, -1.864045743696e-05, {   1,   0,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  2.144282142806e-06,  2.757272013415e-05, {   0,   1,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.924301625561e-05, -1.045067795010e-05, {   2,  -1,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  2.321358500652e-05, -6.450460086716e-06, {   2,  -1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.544284517064e-05, -4.220246948142e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -2.955065144570e-05, -7.108905943038e-08, {   6,  -1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.936869535547e-05,  2.338327967974e-07, {   2,  -1,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.158536493717e-05, -1.798650810148e-05, {   2,  -3,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.020929492150e-05, -1.936128216669e-05, {   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.543108627344e-05, -1.410205738073e-05, {   2,  -3,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.946515228994e-05,  2.874113056684e-08, {   0,   1,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.492709542004e-05,  4.525163465503e-06, {   3,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.129274037233e-05,  1.814923489167e-05, {   1,   0,   0,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 { -1.154906677589e-05, -1.788595040064e-05, {   2,  -3,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.695647400474e-05, -2.473282368238e-06, {   2,   1,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  2.677881374281e-05, -2.578492139284e-06, {   1,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.957719304267e-05,  9.777963254220e-06, {   4,  -1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.054325637017e-06,  2.330059589417e-05, {   2,  -1,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  7.974352881103e-07, -2.855103794260e-05, {   2,  -1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  5.614657963514e-06, -2.372037042827e-05, {   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.444625504841e-06, -2.588735651828e-05, {   4,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.135006607661e-05, -1.798183522276e-05, {   2,  -1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  7.301004526341e-06, -2.200129660522e-05, {   4,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.919550536185e-05, -9.964715886620e-08, {   0,   1,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.953415586150e-05, -9.756552040975e-06, {   2,  -3,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.246861079626e-05,  1.681463574136e-05, {   2,   1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.053225789720e-08, -2.922324719035e-05, {   1,  -1,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.589853021207e-05, -1.336070445152e-05, {   2,  -3,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -7.987463328367e-06,  2.122534849456e-05, {   2,   1,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -2.902233757125e-05, -1.699459295665e-07, {   2,  -3,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.316454578080e-05, -1.602708463081e-05, {   1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.479085075198e-06,  2.170745944274e-05, {   1,   0,   0,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 {  1.946379834414e-05, -9.721141838518e-06, {   3,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.057224656374e-06, -2.604907453354e-05, {   2,  -1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -4.649575694040e-06, -2.445443116128e-05, {   3,  -2,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -5.535379921636e-06,  2.356301351991e-05, {   4,   1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.359006788242e-05,  5.501013245727e-06, {   2,   1,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -2.895452191082e-05, -1.045934986241e-07, {   1,  -1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.939533908174e-05,  9.660180694036e-06, {   0,   1,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.474359409438e-06, -2.058089239899e-05, {   3,  -2,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  2.902310904829e-05,  2.623125275197e-08, {   0,   1,   3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.787931520932e-06,  2.325346652436e-05, {   1,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.285436664184e-05,  1.614469516232e-05, {   2,  -1,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.245851034093e-05,  1.654032394190e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   3,   0,   0,   0 } },
 {  4.487526005132e-06,  2.449516669450e-05, {   3,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.227474606267e-05, -1.670426254820e-05, {   0,   1,   3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.227009043335e-05, -1.669632078836e-05, {   0,   1,   3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.501492946452e-06, -2.046396819942e-05, {   2,   1,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  2.890121586559e-05, -4.672238842043e-08, {   0,   1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  5.567407016597e-06, -2.335957262576e-05, {   1,   0,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.841074858364e-05,  5.131995964660e-07, {   2,   1,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  5.790625660716e-06, -2.312357500577e-05, {   1,   0,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.861998514726e-06, -2.103941063824e-05, {   4,  -1,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -7.861998514726e-06, -2.103941063824e-05, {   0,   1,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -9.442372025526e-06,  1.945779834672e-05, {   2,   1,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -2.027214780467e-05, -8.615044871917e-06, {   0,   1,   4,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.027214780467e-05, -8.615044871917e-06, {   0,   1,   2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.478078670676e-05,  1.407406177144e-05, {   2,   1,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.152920040134e-05, -1.730894372935e-05, {   2,   1,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.491614097730e-06, -2.032370343707e-05, {   0,   1,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.706388708732e-06, -2.009080515586e-05, {   2,  -1,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.116152133005e-05, -7.626858933081e-06, {   0,   1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.002550915999e-06, -2.278451727787e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.315067659158e-05, -5.634352087572e-06, {   0,   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  5.788323245993e-06, -2.294325463855e-05, {   2,   1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.450801917852e-06,  2.727462571109e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -1.947794669388e-05,  9.210396131852e-06, {   2,  -1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.145786804535e-05, -1.719533857393e-05, {   2,   1,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.853202288900e-05, -1.010893141709e-05, {   4,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  7.723974325289e-06, -2.091651568296e-05, {   4,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.858792000352e-05,  5.209008203926e-08, {   0,   1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.337070946680e-05,  5.200608566705e-06, {   2,   1,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.690196941553e-05,  1.661655317120e-06, {   0,   1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.003337636454e-05,  8.518684325426e-06, {   2,  -1,  -1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.463706186972e-06, -2.308687721005e-05, {   1,   2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.028377700690e-07,  2.773760410142e-05, {   2,   1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.903284052785e-05, -9.506025705534e-06, {   3,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.521482958395e-06, -2.300691778424e-05, {   1,   2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.616788611242e-06, -1.989871278124e-05, {   2,  -1,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.872705069236e-05, -9.781262697887e-06, {   2,   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.091691120907e-05, -7.572552605176e-06, {   2,  -1,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  2.364845879452e-05,  4.819141791312e-06, {   2,   1,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 {  2.364845879452e-05,  4.819141791312e-06, {   2,  -1,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 {  1.077854433122e-05, -1.767485036647e-05, {   1,   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.230602337072e-06, -2.520890218403e-05, {   4,   1,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.478140373748e-05, -3.652760978263e-06, {   0,   1,   1,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 {  5.317357613822e-06,  2.310809006386e-05, {   1,   2,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.827903955713e-05,  1.014168535001e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -3.135420745539e-06, -2.528344720074e-05, {   2,   1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.703246208456e-05,  1.351348930660e-06, {   2,   1,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  2.703246208456e-05,  1.351348930660e-06, {   2,   1,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.358558562476e-05,  1.479695301656e-05, {   2,   1,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -9.811367463173e-07, -2.739030608984e-05, {   0,   1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.716898303190e-06,  2.560849844325e-05, {   4,   1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.422395579053e-06,  2.287553695799e-05, {   3,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.085661459309e-05, -7.420308233509e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   2,   0 } },
 {  2.786594351717e-05,  4.007544804706e-07, {   2,   1,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.892737098139e-05,  9.334360556819e-06, {   0,   1,   1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.471831204039e-05,  1.354146560317e-05, {   2,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.795933932357e-05,  2.898513623449e-07, {   0,   1,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.235429120755e-05,  1.589077779741e-05, {   2,   1,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -4.883672531987e-09, -2.821856269992e-05, {   2,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.443877524558e-06,  2.577187052710e-05, {   2,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.705818547144e-05, -1.103738859136e-06, {   2,   1,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.849493160646e-05,  9.665848907113e-06, {   2,  -1,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -6.777129046337e-06, -2.134361134160e-05, {   1,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.910956262521e-05,  9.002893910747e-06, {   2,   1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.800796728531e-05, -1.009948212149e-05, {   4,  -1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.971907196118e-05,  8.379535700733e-06, {   2,  -1,  -3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.581297963214e-05,  2.284886402500e-06, {   1,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.811702534039e-05,  9.966281474793e-06, {   2,  -1,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  2.294689126478e-05,  5.117360900747e-06, {   1,  -2,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.020781504288e-06, -2.703806552523e-05, {   0,   1,   1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  2.112566888410e-05, -6.925606450034e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.614339529478e-05, -1.189883054142e-05, {   2,   1,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  1.876410278363e-05,  9.276503812802e-06, {   0,   1,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.459876930929e-06, -2.657294045761e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -2.687162038333e-05, -1.153179146873e-06, {   0,   1,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -1.531632150318e-05, -1.270426139721e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.062725028013e-05, -1.735810627737e-05, {   1,  -1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -1.519393979656e-05, -1.276614317131e-05, {   2,  -1,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  1.864488629587e-05, -9.312220798128e-06, {   1,  -1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.504130269772e-05,  1.291564603853e-05, {   1,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.896615249155e-06, -1.905805564578e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.408239148267e-07, -2.780374000185e-05, {   2,   1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.261073057953e-06, -2.667916129563e-05, {   4,   1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.885315765639e-06, -1.902490849006e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.070815248298e-05,  1.718934928567e-05, {   2,   1,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  5.366209194250e-06,  2.251451147224e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.466220771538e-05,  1.321017696605e-05, {   2,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.338722040246e-05, -1.448470370475e-05, {   3,  -2,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.778920047288e-05, -7.039123718095e-08, {   1,  -1,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.276885021433e-05,  5.086082981239e-06, {   4,  -1,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.685494578329e-05,  1.099540799497e-05, {   2,   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  1.642643413057e-06,  2.620254803882e-05, {   2,   1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.588046528764e-05, -1.961875246486e-06, {   0,   1,   2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.199205404705e-06,  2.361982075706e-05, {   2,   1,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.063537208066e-05, -7.175128701397e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -1.181978267562e-05, -1.598666229088e-05, {   2,  -3,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.365852259385e-06, -2.343450768499e-05, {   1,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.164236577147e-06,  2.463244547825e-05, {   1,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.588870834250e-05, -1.189013129328e-05, {   1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.438897510310e-06,  2.233214913903e-05, {   0,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.778615944995e-06, -2.098408580182e-05, {   1,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.768792172904e-05,  5.609057085224e-08, {   0,   1,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.065860013698e-05, -1.706991926473e-05, {   0,   1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -5.401979575767e-06, -2.232547188685e-05, {   4,   1,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.287326336693e-05,  4.839950621160e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -5.276693674666e-06, -2.243344766771e-05, {   1,   0,   1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.414091837242e-05, -3.558945019275e-06, {   0,   1,  -1,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  5.300390399623e-06, -2.239924047865e-05, {   1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.439596937500e-06,  1.924024459031e-05, {   2,  -1,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -2.439073335445e-05, -3.235823435340e-06, {   1,   2,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.148916066740e-06,  2.647714206805e-05, {   2,   1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.496678446917e-06,  1.911139768005e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -9.223695592330e-06, -1.837209485772e-05, {   0,   1,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  6.156163523815e-06, -2.143736803832e-05, {   2,   1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.315073555151e-05, -4.419741157535e-06, {   4,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.748785046408e-05, -4.483443066638e-08, {   2,   1,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.649386507850e-07, -2.724254304516e-05, {   0,   1,   2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.740139838067e-05,  9.111383803783e-08, {   2,  -1,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -5.253176073380e-06,  2.223196512112e-05, {   1,  -2,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.533971888983e-05,  2.122545343951e-06, {   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.829463198083e-05,  9.137287111531e-06, {   2,  -1,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.731485123222e-05, -1.134745593853e-07, {   2,  -1,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  2.732097438869e-05,  7.939038281829e-08, {   2,   1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.141376460074e-05,  5.943027442316e-06, {   0,   1,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.900955343219e-05,  8.346595335074e-06, {   2,   1,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  5.245667409017e-06, -2.210918945715e-05, {   1,  -2,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.822876253067e-05, -9.104395905490e-06, {   0,   3,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.225150064720e-06,  2.108267985464e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -2.344784317492e-05,  3.842198211632e-06, {   2,   1,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -2.768570164510e-06, -2.450138026449e-05, {   3,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.742768636472e-05, -9.832097011936e-06, {   2,  -1,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.341534750793e-05,  3.829147481105e-06, {   2,   1,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -7.103169159830e-06,  2.012556401682e-05, {   1,   0,  -1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  5.930294519559e-06, -2.129842517078e-05, {   4,   1,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -5.929867236000e-06, -2.129781038056e-05, {   4,   1,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  8.628061911516e-06,  1.859907731764e-05, {   0,   1,   2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.502435132120e-05, -2.184616633980e-06, {   2,  -1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.593228320842e-05, -1.127230980043e-05, {   1,   0,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -2.719164527282e-05, -7.060911941874e-09, {   2,  -1,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.710854107762e-05,  7.258025293362e-08, {   2,  -1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.207773097867e-06, -2.197297436217e-05, {   1,  -2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.468810417455e-05, -2.487340257551e-06, {   2,  -1,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  5.644090715082e-06, -2.152426090798e-05, {   1,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -1.977120893683e-05, -7.384366870707e-06, {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.324235232422e-05,  3.891315747519e-06, {   0,   1,   1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  2.323982543926e-05,  3.889522763381e-06, {   0,   1,   1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  7.550881167831e-06, -1.957837728432e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  2.709033256849e-05,  2.656829968942e-08, {   2,   1,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.266016909630e-05,  4.445972540930e-06, {   1,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.184916311741e-06,  2.190307320150e-05, {   1,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.873103671529e-06,  2.516777030950e-05, {   0,   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -2.684501591569e-05, -1.942288403842e-07, {   2,  -1,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.828346832422e-05, -8.749572339052e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  1.462169993482e-05, -1.240974696953e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  2.694381738930e-05, -6.314191032003e-08, {   2,  -1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.379642675670e-05,  3.208553111944e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  2.151345751077e-05,  5.470654333358e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  2.665989473762e-05, -3.187384624777e-07, {   2,  -1,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  8.661346807328e-06,  1.829980742256e-05, {   2,  -1,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 {  2.353687104905e-05, -3.412461994748e-06, {   2,   1,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -1.413212377152e-05, -1.281495449068e-05, {   2,  -1,   0,   0,   0,   0,  17, -28,   0,   0,   0,   0,   0 } },
 { -1.806592062461e-05,  8.880344439449e-06, {   2,   1,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.307414333343e-05, -1.386222844802e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -1.795031348144e-05, -8.957768696917e-06, {   1,  -1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.151135226231e-06,  2.175108491494e-05, {   3,   0,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.604064642808e-05, -1.085552106905e-05, {   2,  -1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  4.200801624674e-06,  2.268411735820e-05, {   1,  -2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.686727860279e-05, -7.157475755139e-09, {   0,   1,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.808408362429e-06, -2.202069740393e-05, {   2,   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -8.935868877803e-06,  1.789274286963e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.313482060927e-05, -3.688282033044e-06, {   1,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.067039242748e-05, -1.613969052580e-05, {   0,   1,   3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.128259787304e-06, -2.167007159890e-05, {   1,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.306708463638e-05,  3.727803881900e-06, {   1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.262704484017e-06, -2.149926558579e-05, {   2,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.427248724704e-07, -2.581739417224e-05, {   2,  -1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.114174172549e-06,  2.161905396958e-05, {   1,   2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.666019511960e-05,  5.851803503942e-08, {   2,   1,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.134043393508e-06,  2.258005343118e-05, {   1,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.367491537500e-06,  2.434645360896e-05, {   0,   1,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.705156836768e-07, -2.572419959377e-05, {   0,   1,  -1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.106076426881e-06, -2.157741821289e-05, {   1,   0,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.619715447894e-05,  4.778518641613e-07, {   2,   1,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  9.459806665697e-06,  1.720949911084e-05, {   2,   1,  -1,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  9.459806665697e-06,  1.720949911084e-05, {   2,  -1,  -1,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -2.660867683879e-05, -2.938445357271e-08, {   3,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.775949735293e-05, -8.870158338462e-06, {   0,   3,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.625589242671e-05, -1.037224065180e-05, {   2,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  9.013471696212e-06, -1.759346471468e-05, {   0,   1,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.626554451716e-05,  1.033492665335e-05, {   1,   0,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 { -2.593891280838e-05,  6.376800043546e-07, {   2,   1,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.580903271860e-05, -7.656225577828e-07, {   4,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.791848239592e-05,  8.649478152292e-06, {   2,  -1,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.080185403364e-06, -2.147463646796e-05, {   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.021311241550e-05,  1.634111672672e-05, {   0,   1,   2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  2.612742488160e-05, -4.109097932060e-07, {   2,  -1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.650739942738e-05, -7.654721600908e-09, {   0,   1,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.640729357563e-05, -1.043444171646e-07, {   0,   1,  -1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  2.640729357563e-05, -1.043444171646e-07, {   0,   1,  -1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 {  1.058571727416e-05, -1.591180685710e-05, {   4,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.190683664874e-05,  4.586416839104e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -7.904758958065e-06,  1.858801654205e-05, {   1,   0,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.129495049997e-06,  2.133176806298e-05, {   3,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.146045981633e-05,  4.996786562788e-06, {   2,  -1,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -1.668236984030e-05,  9.773044581889e-06, {   3,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.968206469953e-06, -2.147192010864e-05, {   1,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.214332024125e-05,  4.294036595790e-06, {   1,   0,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.351751573830e-05, -1.291360185274e-05, {   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.825660372463e-05, -8.166787645211e-06, {   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.825710257123e-05, -8.165672397516e-06, {   2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.639754459012e-05,  1.975313718174e-08, {   2,  -1,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.824714562817e-06, -1.659205394834e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  1.290865510633e-05,  1.350510597496e-05, {   2,   1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -1.830811066143e-05,  8.103809897531e-06, {   2,   1,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.684610503782e-05, -9.559388191407e-06, {   2,   1,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.099100278332e-06, -2.129593190421e-05, {   1,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.100438006706e-06,  2.128385729808e-05, {   3,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.554153148664e-05, -1.082850141934e-05, {   2,   1,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -5.503105847343e-06,  2.086150413921e-05, {   1,   0,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.102274470777e-05,  5.339643298561e-06, {   4,   1,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  2.215245350055e-05, -4.196199648447e-06, {   0,   1,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.016117070348e-05,  1.618319267826e-05, {   2,   1,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -2.683707643797e-06,  2.364119760358e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.760598807718e-05, -8.717594960690e-06, {   1,   1,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.760598807718e-05, -8.717594960690e-06, {   1,  -1,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.485502981639e-05, -1.467934544849e-06, {   2,  -1,   0,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -1.184343000556e-05,  1.447460316358e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 {  2.054285092708e-05,  5.767455400441e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.147566561220e-05,  1.483434045434e-05, {   2,   1,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -4.917922132074e-06, -2.136205043896e-05, {   1,   0,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.139163921746e-05,  1.488725542746e-05, {   2,  -1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.775854167409e-06, -2.148867415106e-05, {   4,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.538495798087e-06,  1.771181477478e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -1.114635191347e-06, -2.512463072340e-05, {   2,  -1,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -4.967198212608e-06,  2.126296158870e-05, {   1,   2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.519460351834e-06, -2.070968104364e-05, {   3,  -2,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -6.301770453994e-06,  1.991718535317e-05, {   2,   1,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.851536620668e-05, -7.670355134992e-06, {   2,  -1,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0,   0 } },
 { -7.355893090097e-06,  1.881471494816e-05, {   2,  -1,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -2.197379501226e-05,  4.191506414182e-06, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.890827088237e-06,  2.024296033234e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 {  1.742388084133e-05,  8.701653003208e-06, {   2,  -1,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.869704190502e-06, -1.924126905706e-05, {   2,  -1,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.561214124967e-05,  4.927650031795e-07, {   3,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.555612534144e-05, -5.253728766826e-07, {   2,  -1,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.185333103136e-05,  4.218238342942e-06, {   1,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.140312548538e-05, -1.466407998537e-05, {   2,  -1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -5.075847165856e-06, -2.097689275890e-05, {   2,   1,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.855221543171e-06, -1.919316036306e-05, {   2,  -1,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.031423961374e-06, -2.395995391920e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 { -2.540712231786e-05, -5.832332034376e-07, {   2,  -3,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -8.651569786335e-06, -1.733783405627e-05, {   1,   0,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.120491173282e-05, -4.778665414680e-06, {   1,   0,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.522743660871e-05,  1.073750731717e-05, {   0,   1,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  5.229442564423e-06, -2.072354243285e-05, {   2,  -1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.523356263268e-05,  7.193048222311e-07, {   0,   1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -2.551065763805e-05,  4.259943354174e-07, {   0,   1,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.583372923003e-05,  9.333157424039e-08, {   2,   1,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.520234577178e-05,  1.072272542278e-05, {   0,   1,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.335205974427e-05,  1.255814688862e-05, {   2,  -1,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.707172256551e-06, -1.819911868728e-05, {   2,   1,   0,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 {  7.707172256551e-06, -1.819911868728e-05, {   2,  -1,   0,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.236679801303e-05, -3.515769752873e-06, {   2,   1,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  4.990753377422e-06,  2.084861617370e-05, {   2,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.037867609529e-05,  1.545705514245e-05, {   1,   0,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  2.375969276799e-05,  2.074188987803e-06, {   2,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.911063787552e-05,  6.717895607810e-06, {   2,  -1,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.581369939033e-05, -1.315714350152e-08, {   2,   1,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.721061424013e-07, -2.515272288662e-05, {   0,   1,   3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.446268505596e-05, -1.357262543538e-06, {   2,  -1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.105310767502e-05, -4.745694681985e-06, {   1,  -2,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  8.373853400704e-06, -1.741054851981e-05, {   1,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  8.645178134629e-06, -1.713837912429e-05, {   2,  -1,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.158810345001e-05, -1.418769229064e-05, {   2,  -1,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.421342531147e-05,  1.156181643336e-05, {   2,  -1,   0,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  1.816034098646e-05,  7.602654607945e-06, {   1,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.172017861324e-05, -4.039284746933e-06, {   4,  -1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.052431925417e-05,  5.235091317474e-06, {   1,   0,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.082936667224e-05, -1.492712216091e-05, {   0,   1,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.982625588342e-05,  5.916819117016e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -5.802816210796e-06, -1.994009769719e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -2.566877341772e-05,  6.677009935478e-08, {   2,  -1,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.873700115245e-05, -6.994625637799e-06, {   2,  -1,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.366979985004e-05,  1.204064131559e-05, {   2,   1,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -4.801411356882e-06,  2.090446946874e-05, {   1,  -2,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  8.614545155813e-06, -1.707852891462e-05, {   2,   1,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.367790615350e-05,  1.199514378952e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   3,   0,   0 } },
 {  7.571714390437e-06, -1.810021474427e-05, {   0,   1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.906414133663e-06,  2.075911077804e-05, {   3,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.882942552306e-06,  2.073394304242e-05, {   1,   0,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.508863669541e-05,  5.239582492502e-07, {   0,   1,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.534431749347e-05,  2.636387237039e-07, {   4,   1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.953622309270e-06, -1.564781814379e-05, {   1,   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.100834727906e-05, -4.578218217803e-06, {   1,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  3.765979626665e-06,  2.181944285722e-05, {   3,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.765979626665e-06,  2.181944285722e-05, {   1,  -1,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.053633766132e-06, -2.052219915021e-05, {   3,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.555267241314e-05,  1.000760855232e-05, {   2,  -1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -2.555095300834e-05, -1.236728499216e-09, {   0,   1,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.415100266914e-06,  1.810438848583e-05, {   0,   1,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  1.699432347964e-05,  8.498839047566e-06, {   2,  -1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -9.318569601983e-06,  1.616103730170e-05, {   2,  -1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.544552675737e-05,  1.733086913868e-08, {   0,   1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.690872457111e-06, -2.377053723370e-05, {   2,  -1,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.191562747516e-05,  1.352973741018e-05, {   2,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.410037438764e-05,  1.342288253023e-06, {   2,  -1,   0,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -4.866948683669e-06,  2.055714863124e-05, {   3,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.073729128256e-05, -4.679239600935e-06, {   2,  -1,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.401930530337e-05, -1.390002227896e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -1.930219121727e-05, -6.098581311636e-06, {   3,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.321334401618e-05, -1.214804206741e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -5.685735076222e-06, -1.965724590575e-05, {   1,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -7.603327744926e-06, -1.773134317908e-05, {   2,   1,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -2.387129131611e-06,  2.294178858508e-05, {   0,   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -1.732635855939e-05,  7.974230852315e-06, {   2,  -1,   0,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -1.373901080678e-05, -1.151176115986e-05, {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.019813150130e-05, -5.040587072123e-06, {   2,   1,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.145038229214e-05, -1.377774574888e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  2.095985422314e-05, -4.195779708103e-06, {   4,  -1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.819971887123e-05,  6.950061225798e-06, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  2.197496129812e-05, -3.119559199138e-06, {   2,   1,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  2.503351089966e-05, -5.610374610003e-08, {   0,   1,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.698892864116e-05, -8.066357069644e-06, {   1,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.335559818723e-06, -1.671950577197e-05, {   0,   1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  7.515037210033e-06, -1.753658323284e-05, {   2,   1,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  4.755297201850e-06, -2.027396798982e-05, {   1,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.299023996515e-05, -1.203336274013e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -2.497051231297e-05,  1.818527596392e-09, {   2,  -1,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -9.289804679873e-06, -1.563312302079e-05, {   0,   1,   2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.587376100417e-05, -9.027872602614e-06, {   2,   1,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.299033540638e-05, -1.904941876244e-06, {   1,  -1,   0,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 { -2.299033540638e-05, -1.904941876244e-06, {   1,  -3,   0,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 {  4.716467173946e-06,  2.013725244620e-05, {   1,   2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.680463713494e-05, -8.034517687493e-06, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  5.710357752587e-06, -1.908076512259e-05, {   0,   1,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.689252655612e-05,  7.880792729956e-06, {   2,   1,  -1,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 { -1.689252655612e-05,  7.880792729956e-06, {   2,  -1,  -1,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 { -1.372348685653e-05,  1.104232865459e-05, {   2,  -1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  1.980604093852e-05, -4.953338409735e-06, {   1,   0,  -3,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.458890718025e-05, -1.518912372857e-07, {   1,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.596500174188e-05,  8.774227453045e-06, {   2,   1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  6.402712823493e-06,  1.832725680822e-05, {   2,  -3,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.464378469792e-05,  7.573089065868e-08, {   2,  -1,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.104506508269e-05, -1.367005480407e-05, {   1,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.464001190377e-05, -7.163281595396e-08, {   0,   1,   1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  2.464001190377e-05, -7.163281595396e-08, {   0,   1,   1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -2.469614287618e-05, -1.509705452082e-08, {   2,  -3,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.993223679207e-05,  4.755506398903e-06, {   2,  -1,  -2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.757143927143e-05,  7.114602082204e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  2.465784187037e-05,  1.868605529326e-08, {   4,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.773400359495e-08,  2.465202403624e-05, {   4,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.925543985638e-05, -5.406295953556e-06, {   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.461162731877e-05,  4.104635132115e-08, {   4,  -1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.126431221431e-05, -3.376510419751e-06, {   2,  -1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.718075095503e-06, -1.991972514489e-05, {   3,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.421647489712e-05,  4.083451422941e-07, {   2,  -1,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  9.959026895877e-07, -2.360030475663e-05, {   2,  -1,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.456861392057e-05,  2.036430902951e-08, {   0,   1,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.867722271590e-05, -5.907185618873e-06, {   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.913269893216e-06,  2.266750526802e-05, {   0,   1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.550157693108e-07, -2.377964995271e-05, {   0,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.066620315343e-05,  3.858532401055e-06, {   2,   1,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.543662789257e-05, -9.084553251576e-06, {   2,   1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.432124239632e-05,  1.018823166125e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -1.991182365854e-05,  4.587230987391e-06, {   4,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.273310090750e-06,  2.222544161569e-05, {   0,   1,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.210931520822e-05, -1.236633105041e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  9.372740615082e-06, -1.505551385360e-05, {   2,   1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.431054113485e-05,  1.010956750519e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -2.438175593564e-05,  3.593774175860e-08, {   0,   1,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.158163176595e-06, -1.525491127692e-05, {   1,   0,   1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -2.288768037677e-05, -1.513375510314e-06, {   2,  -1,  -3,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -2.009601601995e-05,  4.292343360116e-06, {   1,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.055934236362e-05, -3.828608919185e-06, {   1,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.894511882476e-05,  5.431992921081e-06, {   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.025671720310e-05,  4.109154595416e-06, {   2,   1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  7.852865995421e-06,  1.650612763184e-05, {   2,  -1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.099368108043e-05, -1.335802038505e-05, {   1,   0,   0,   0,   0,   9,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.099798478818e-05, -3.317322765784e-06, {   2,   1,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  1.076745512048e-05, -1.353939086533e-05, {   1,  -2,   0,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 {  2.423080391662e-05,  7.374376428579e-08, {   2,   1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.974010077316e-05,  4.563884814315e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -2.910213117082e-06, -2.137728994166e-05, {   2,  -1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.392131653196e-05, -1.036390345414e-05, {   2,   1,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -2.425352053721e-05, -2.936786090596e-08, {   4,   1,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.419254700248e-05,  8.439740898596e-08, {   2,  -1,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.150802036244e-06, -1.710414863412e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  8.445235421384e-07, -2.340988381253e-05, {   0,   1,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.427935178554e-06,  2.282617292236e-05, {   2,  -1,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -4.795594576154e-06, -1.945769331643e-05, {   1,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.617127792381e-05,  8.065754441217e-06, {   0,   1,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.984004224723e-05,  4.392400728786e-06, {   2,   1,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.943735194889e-06,  1.728601008133e-05, {   4,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.011214129936e-05,  4.107781698718e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  9.823503165516e-06, -1.439232275323e-05, {   2,  -1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.380264617424e-05, -4.064779484172e-07, {   1,   0,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.036209153307e-06, -2.114446138772e-05, {   3,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  4.981873861248e-06, -1.919644658421e-05, {   1,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.221024191993e-06,  2.294829263132e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.375335969001e-05,  1.040429131816e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.375335969001e-05,  1.040429131816e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  1.525898285094e-05, -8.896579405946e-06, {   2,   1,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  1.648187398621e-05,  7.652955897678e-06, {   2,  -1,   0,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  1.760467947455e-05, -6.470695337488e-06, {   2,   1,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.400342004907e-05,  1.006398176065e-05, {   2,  -1,   0,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 {  7.509576029958e-06,  1.654506130667e-05, {   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.544995392754e-05, -8.603494884373e-06, {   3,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -7.142352735899e-06, -1.689484500697e-05, {   2,  -1,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -2.357528671594e-05,  4.602358556748e-07, {   1,   2,   0,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.357528671594e-05,  4.602358556748e-07, {   1,   0,  -2,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.287756328520e-05,  1.137452427727e-06, {   2,  -1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.380685488391e-06, -1.961031036231e-05, {   4,   1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.942047715676e-06,  1.704882735663e-05, {   1,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -5.007186505771e-07,  2.348310082893e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -4,   0,   0,   0 } },
 { -2.363965676006e-05,  3.389583590111e-07, {   2,   1,  -1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -1.147765845931e-05,  1.249987506078e-05, {   2,  -1,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.604882231050e-05,  7.924285316526e-06, {   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.352777127031e-05,  4.440499222165e-07, {   4,  -1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.388770866564e-05,  6.952949366147e-08, {   2,   1,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.256980708782e-05,  1.138386617449e-05, {   2,   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -2.387222674606e-05,  7.979891988141e-08, {   2,  -1,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  4.666664880989e-06,  1.928324871351e-05, {   2,  -1,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.717921662748e-05, -6.753015469882e-06, {   3,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -3.053873392897e-08, -2.389516537522e-05, {   2,  -3,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.710608941685e-06, -1.921121647295e-05, {   4,  -1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.660233864514e-05,  7.316454853308e-06, {   2,  -1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.221261784457e-05, -1.168818230678e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.407041097105e-05,  9.797326672085e-06, {   2,   1,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -6.249700715472e-06,  1.760198375103e-05, {   2,  -1,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  6.248950308344e-06,  1.760052452353e-05, {   2,  -1,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.364419862626e-05, -1.921926124840e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.404793943124e-05, -9.761370858241e-06, {   2,  -1,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  2.378271763337e-05, -2.550719522511e-08, {   2,  -1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.783687027943e-06, -1.901597206845e-05, {   1,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.400671822350e-06,  1.837515753714e-05, {   3,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.806244940649e-05,  5.700051466552e-06, {   1,   0,   3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.369284475876e-05,  6.948039906130e-08, {   2,  -1,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.124248500422e-05,  2.513853273049e-06, {   2,  -1,   0,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 {  5.396231544138e-06,  1.835949055683e-05, {   3,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.955580521190e-05, -4.191715939108e-06, {   1,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.405402261995e-05, -9.686470109481e-06, {   1,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.532568082864e-06,  1.915689349581e-05, {   1,  -2,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.686261971188e-06,  2.199681652841e-05, {   2,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.143044523130e-05,  1.225202505642e-05, {   2,   1,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  5.441321705567e-06, -1.822272431541e-05, {   1,   0,   0,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.049193888655e-06,  2.158721930569e-05, {   2,  -1,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.657807898253e-05, -7.043221039027e-06, {   4,  -1,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.657807898253e-05, -7.043221039027e-06, {   4,  -1,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.944689967648e-07,  2.271245635843e-05, {   0,   1,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.356904500226e-05,  3.599678261946e-08, {   0,   1,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.181024025559e-05, -1.178578344797e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  1.441804441364e-05,  9.158534331794e-06, {   4,  -1,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.031781463105e-05,  1.323836112485e-05, {   1,   0,   0,   0,   0,  -8,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.075832159181e-05,  2.794825952467e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -1.042316206342e-06, -2.250492008827e-05, {   0,   1,   3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.876748758718e-05, -4.773057220989e-06, {   4,   1,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.001038431018e-05,  3.529463706164e-06, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.119111958113e-05,  1.232100273975e-05, {   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.073182716123e-05, -1.277111496620e-05, {   4,  -1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.291169729193e-05,  1.057954150670e-05, {   0,   1,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.302923180843e-05, -4.381563442029e-07, {   2,   1,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  4.717166000141e-06, -1.874838242302e-05, {   0,   3,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.570911493025e-05, -7.756051462061e-06, {   2,  -1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -4.718098783448e-06, -1.874613499243e-05, {   0,   3,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.435072971556e-05,  9.113120141184e-06, {   4,  -1,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -9.658551516136e-06, -1.378205972835e-05, {   1,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.272316426260e-06, -2.016497654927e-05, {   0,   1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.767632509253e-05,  5.760838756815e-06, {   1,   0,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.538796647209e-05,  8.045464651365e-06, {   0,   1,   1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -1.183192204367e-05, -1.158638847517e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 { -1.183192204367e-05, -1.158638847517e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 {  1.537630029011e-05,  8.034836173121e-06, {   0,   1,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  2.086182439768e-05, -2.545715941259e-06, {   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 {  9.839780139713e-06,  1.356422159572e-05, {   2,  -1,  -1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.548743606204e-05, -7.915887090041e-06, {   0,   1,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -7.655072978841e-06, -1.574167306420e-05, {   0,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.445447526865e-05, -8.919792205920e-06, {   1,   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 {  2.326981341267e-05, -8.951783323752e-08, {   2,   1,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  2.331914130774e-05,  3.520917527697e-08, {   2,   1,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.533496113734e-06, -1.881842695714e-05, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.577821961442e-05,  7.566650947920e-06, {   2,  -1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.435317336665e-06, -1.390893606157e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.382688389320e-05, -9.503521557868e-06, {   2,  -1,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  4.773616355769e-06,  1.855308326438e-05, {   3,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.661383457205e-06,  1.965946490491e-05, {   3,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.279697662383e-05, -1.049745030830e-05, {   2,  -1,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  2.186257267322e-05, -1.394439712466e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  1.909664474393e-05,  4.158559377107e-06, {   0,   1,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.091328355513e-05, -2.339424142515e-06, {   0,   1,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.721091829307e-05,  6.041372164218e-06, {   3,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.086628580319e-09, -2.325001169947e-05, {   4,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.852647255904e-06, -1.838356419287e-05, {   1,  -2,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.307433764701e-05, -1.546008039218e-07, {   3,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.441279571980e-06, -1.878619139377e-05, {   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.764398015226e-05,  5.567068912889e-06, {   1,   2,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.114367462494e-05, -1.205891328125e-05, {   3,   0,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.880131355744e-05,  4.388208285653e-06, {   2,   1,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -1.879706906917e-05,  4.387898063589e-06, {   2,   1,  -1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.981724559107e-05,  3.358830715565e-06, {   0,   1,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -1.981666974060e-05,  3.357709054622e-06, {   0,   1,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  7.897882214289e-06, -1.525921565873e-05, {   1,   0,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.543937459885e-05,  7.693018181524e-06, {   0,   1,  -1,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.651004335287e-05, -6.620719102182e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -7.194701220719e-07, -2.240867246850e-05, {   0,   1,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.191693498437e-06, -1.391358796300e-05, {   0,   1,   3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.548373489254e-05,  7.610491215200e-06, {   2,  -1,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.430934276240e-05,  8.765166290487e-06, {   2,   1,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  6.568944026465e-06, -1.649225897538e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 {  4.867687695347e-06,  1.818909656778e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -1.611797829026e-05, -6.922047245496e-06, {   3,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.556661568030e-05,  7.458849639342e-06, {   2,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.644726411138e-06, -1.335504294100e-05, {   2,   1,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.427197164809e-05,  8.723335055880e-06, {   1,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.400853326022e-06, -1.858534174859e-05, {   1,  -2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.442446417767e-06,  1.354122453209e-05, {   1,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.018031650960e-05, -1.278788399287e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 {  2.464288696802e-06,  2.048932641092e-05, {   2,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.432474748450e-07, -2.230739203207e-05, {   0,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.894313481827e-05, -4.003123310426e-06, {   1,   0,   0,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.911413168645e-05,  3.823323934199e-06, {   0,   1,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -4.936299467119e-06, -1.799713992223e-05, {   2,  -1,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.226042544295e-05, -6.576567295812e-07, {   1,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.046725899025e-05, -1.244539345614e-05, {   1,  -2,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.129362228839e-07,  2.267928145057e-05, {   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.536378975321e-06, -1.335402865581e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -7.342533884343e-06, -1.554041348705e-05, {   4,  -1,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -6.281379546135e-06,  1.659906578591e-05, {   0,   1,   2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.187083430012e-05,  9.976392730046e-07, {   2,  -1,   0,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 { -9.438468843354e-06, -1.337315110455e-05, {   2,  -1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.386382487748e-07,  2.186668743068e-05, {   2,   1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.360369817397e-06, -1.844154626911e-05, {   3,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.906484815830e-06,  1.689174827256e-05, {   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.008669926765e-05,  2.709189160124e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  1.990250385029e-05, -2.853743166657e-06, {   1,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.351353933746e-06,  1.840122064438e-05, {   1,   0,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.487522661655e-06, -1.325909332825e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.755702142677e-05, -5.189252199095e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.265256687366e-05,  7.833287991091e-08, {   2,   1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.256334788079e-05, -1.644389973655e-07, {   3,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.014627118338e-06, -2.069263592042e-05, {   4,  -1,  -2,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.263168852741e-05,  7.297689617275e-08, {   2,  -1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.261647650565e-05,  7.344863674970e-08, {   2,   1,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  4.065152385767e-06, -1.861804835615e-05, {   2,  -1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -4.084478352886e-06,  1.859589442367e-05, {   2,   1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  1.066048385987e-05, -1.201662403486e-05, {   2,  -1,  -3,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.781156468356e-05,  4.865469770025e-06, {   0,   1,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.529177909989e-05, -7.381208403235e-06, {   1,   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -6.745641848075e-06,  1.590204162264e-05, {   0,   1,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -5.004478306700e-06,  1.761926927234e-05, {   1,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  4.335173871329e-06, -1.828757419882e-05, {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.332224947694e-06,  1.828696647722e-05, {   1,   0,   2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.542252691122e-06, -1.907378265771e-05, {   1,  -2,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.165130206512e-06, -1.744256879648e-05, {   2,  -1,   0,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.729772754172e-06, -1.587648133737e-05, {   2,   1,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -2.212898787030e-06,  2.039099047357e-05, {   1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.999449020797e-06,  1.759144410471e-05, {   1,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.507999313479e-05,  7.503860794366e-06, {   0,   1,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.625114804744e-06, -1.494981257213e-05, {   0,   1,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  2.034631451306e-05, -2.219847471404e-06, {   3,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.230551983556e-05, -2.588616258328e-07, {   2,   1,   0,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.386803393681e-05,  8.682476127139e-06, {   1,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.802748792302e-05, -4.514096122862e-06, {   1,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.455353139166e-06,  1.607962069262e-05, {   4,  -1,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -7.619256187905e-06, -1.489820813065e-05, {   0,   1,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  6.920146595393e-06,  1.558002004935e-05, {   1,   2,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.561636861966e-06,  2.093580607468e-05, {   2,   1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.445945009011e-05, -7.948542872769e-06, {   4,   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.724861752488e-05,  5.150363063232e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  1.525846888160e-05,  7.131523115242e-06, {   2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.228533293360e-05, -6.003880501924e-08, {   2,   1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.182495152287e-06,  1.616240853498e-05, {   1,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.918407065649e-05,  3.149046582186e-06, {   2,  -1,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  3.186501525164e-06,  1.914528698874e-05, {   2,  -1,   1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.230895998963e-05,  1.622668381475e-08, {   2,   1,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.573628325604e-05,  6.588579648587e-06, {   2,   1,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.616911847257e-05,  6.150695259611e-06, {   1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.713993908589e-06,  1.759536511138e-05, {   1,   0,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.915740441065e-05,  3.138264425092e-06, {   2,  -1,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  1.304993583912e-05,  9.236214404202e-06, {   0,   1,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  1.841576336023e-05,  3.869493741535e-06, {   3,  -1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.841576336023e-05,  3.869493741535e-06, {   1,  -1,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -5.019675307835e-06, -1.725851998346e-05, {   4,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.015626865215e-06, -1.724445814103e-05, {   4,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.217795194800e-05,  4.973940823269e-08, {   0,   1,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.580674768690e-06,  1.463860028058e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 { -2.188150219881e-05, -3.197225371466e-07, {   0,   1,   3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.439193508247e-05,  7.800614619497e-06, {   4,  -1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  5.675799564925e-06,  1.651322155955e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  1.868454620292e-05, -3.497782423685e-06, {   4,   1,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.140150570086e-05, -7.788000833961e-07, {   0,   1,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  1.411145490227e-05,  8.061752506726e-06, {   4,   1,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -5.306127559607e-06,  1.686205814598e-05, {   2,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.211940755962e-05, -2.900933536602e-08, {   4,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.600748013074e-05,  6.125789997750e-06, {   2,   1,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -2.212713101295e-05,  3.447690058650e-09, {   1,  -1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.473812682503e-05, -7.384717317495e-06, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  8.590777217976e-06, -1.350937532678e-05, {   2,   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  2.208860578506e-05,  3.523960984814e-09, {   1,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.562182726065e-06, -1.551247731445e-05, {   1,   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 { -2.531880571247e-06,  1.954049234296e-05, {   2,   1,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  9.625173174459e-06,  1.244104249810e-05, {   1,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  1.148353862410e-05,  1.056668683430e-05, {   0,   1,   1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.777717007241e-06,  2.025618573929e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.753525040577e-05,  4.459617176658e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  6.099697847590e-06,  1.588768769148e-05, {   0,   1,   2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.423994404198e-06, -1.754357402102e-05, {   0,   1,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -2.193585211769e-05, -2.026364598186e-08, {   2,  -1,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.780477726449e-06, -1.716870706236e-05, {   4,   1,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -4.779843879231e-06, -1.716782583299e-05, {   4,   1,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.309148310491e-06,  2.063796959744e-05, {   2,  -3,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  2.180674156627e-05, -1.275520644182e-07, {   2,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.136370653690e-05,  1.056185548950e-05, {   2,  -1,   0,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 {  2.144287406718e-05, -4.740180492388e-07, {   4,  -1,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.915615017869e-05,  2.724681712079e-06, {   2,   1,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -5.956357911512e-06,  1.592362564421e-05, {   0,   1,  -1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -8.443428142939e-07,  2.102697867558e-05, {   2,   1,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  6.312635605651e-06,  1.555780337732e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  4.208113784297e-06, -1.766149598768e-05, {   1,   2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.606968590279e-06, -1.725517192117e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -1.409156588360e-05, -7.760398709887e-06, {   2,   1,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -5.262752881024e-06,  1.658595324399e-05, {   2,  -1,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.183684037397e-05,  1.001065322177e-05, {   0,   1,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -4.195886503637e-06, -1.764674094648e-05, {   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.324528291450e-06,  1.749360085387e-05, {   1,  -2,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.152010475197e-06, -2.064989271389e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  1.760983855410e-06,  2.003963596469e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  1.602104244127e-05, -5.768134099092e-06, {   2,   1,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -4.161493954137e-06,  1.762394642331e-05, {   3,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.177654284485e-05,  6.615187175608e-09, {   1,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.448475628071e-07, -2.113294955153e-05, {   0,   1,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.142137360204e-06, -1.761528792199e-05, {   1,   0,   1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.674245714896e-05,  4.995665676827e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.450800642872e-05,  7.229316464915e-06, {   0,   1,   1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  8.336164869714e-06, -1.339372810704e-05, {   4,   1,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.434027255210e-06,  1.829492974470e-05, {   0,   1,  -2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.146517626401e-05, -2.636027994980e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -8.140130171886e-06,  1.357014295401e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.790228374493e-06,  1.991971592084e-05, {   1,  -2,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789821606504e-06,  1.991975247379e-05, {   3,  -2,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789710353591e-06,  1.991976246972e-05, {   1,  -2,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789115844890e-06,  1.991981587499e-05, {   1,  -2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06, -1.991981649951e-05, {   5,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06, -1.991981649951e-05, {   5,   0,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789108891562e-06,  1.991981649951e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789108891562e-06,  1.991981649951e-05, {   3,   0,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06, -1.991981649951e-05, {   3,   0,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06, -1.991981649951e-05, {   3,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.789108891562e-06, -1.991981649951e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06,  1.991981649951e-05, {   1,   0,  -1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.789108891562e-06,  1.991981649951e-05, {   1,   0,  -2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789108891562e-06, -1.991981649951e-05, {   1,   0,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.789021974949e-06, -1.991982430578e-05, {   1,   0,   4,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.788980254962e-06,  1.991982805266e-05, {   3,   2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.788681261497e-06, -1.991985490270e-05, {   1,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.449539924239e-05, -7.179533106702e-06, {   1,  -1,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  6.110813766185e-06, -1.556305931796e-05, {   4,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -4.592240676117e-06,  1.707998516876e-05, {   1,   0,  -1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.768930664860e-07, -2.079520720269e-05, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -8.529302803598e-06,  1.313349997326e-05, {   2,  -1,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -4.143057858258e-06, -1.750622718638e-05, {   1,   0,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.444653476209e-05,  7.178057214897e-06, {   0,   1,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.959902999219e-09,  2.161698304636e-05, {   2,   1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.830069055817e-05, -3.316460068299e-06, {   2,   1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.117715424124e-06,  1.949252444810e-05, {   2,   1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.602566857752e-05,  5.563415988437e-06, {   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.136560844280e-05,  2.230205410100e-07, {   4,  -1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.664047594624e-05,  4.945915210128e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -3.381871172583e-06,  1.819558679033e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.176761771398e-07, -2.073389142880e-05, {   2,   1,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.644031751401e-06,  1.589600362509e-05, {   2,   1,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.511518355886e-05, -6.422424442039e-06, {   4,   1,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.511518355886e-05, -6.422424442039e-06, {   4,   1,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.642738822869e-06,  1.589222247082e-05, {   2,   1,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.345297021583e-05,  8.073566520039e-06, {   0,   1,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.886439695576e-05, -2.659534520790e-06, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -4.483517326940e-06, -1.703164663464e-05, {   2,   1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.297456452089e-06,  1.721701658208e-05, {   3,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.825017373120e-05, -3.257319086446e-06, {   1,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.805174467720e-05, -3.455123284781e-06, {   1,   0,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.414797618806e-05, -7.355687707388e-06, {   4,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.659546559082e-05, -4.907233381047e-06, {   1,   1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -2.044761184475e-05, -1.050567228480e-06, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.686920781134e-05,  4.628751359440e-06, {   4,  -1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.147617512509e-05,  9.842108793673e-09, {   2,  -1,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.893517396682e-05,  2.549912617358e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -8.050084418107e-06, -1.343279072075e-05, {   4,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.146197755350e-05,  1.001761956863e-05, {   2,  -1,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -1.785548354944e-05,  3.622061383095e-06, {   1,   2,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.341469147245e-05,  8.046028915687e-06, {   0,   1,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.581707053120e-05, -5.641802602272e-06, {   0,   1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.636927157782e-05,  5.087122529522e-06, {   0,   1,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.797527503141e-06,  1.863793100779e-05, {   2,  -1,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -1.919740096346e-05, -2.228203226942e-06, {   0,   1,   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  1.040826823734e-06, -2.038033118935e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.239650332581e-06,  1.717273395235e-05, {   1,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.372359504249e-05,  7.676010030507e-06, {   2,   1,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  8.189579633814e-06, -1.320797637086e-05, {   0,   1,   2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.229467246739e-06,  1.815667094571e-05, {   0,   1,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.117780097730e-05,  2.071359456874e-07, {   2,  -1,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -8.241579834075e-06, -1.314115823441e-05, {   4,  -1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.134199495423e-05,  3.797242203864e-08, {   4,  -1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.586519297313e-06,  1.577932617274e-05, {   4,  -1,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.135073130067e-05, -6.896846454023e-09, {   3,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.145638812150e-06, -1.720158416601e-05, {   0,   1,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -7.280411794405e-06,  1.405572627390e-05, {   1,   0,   0,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 { -1.363681913626e-05, -7.688757547483e-06, {   2,  -1,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.964507077836e-05,  1.650924428663e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  1.072196763981e-05, -1.057136368674e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  2.016052553161e-05, -1.128972452524e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  4.048892384985e-06, -1.721624419106e-05, {   1,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.121270063306e-05,  3.081805516591e-08, {   2,  -1,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.555177279348e-06,  1.968664953185e-05, {   0,   1,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.342244536341e-05, -7.807291335658e-06, {   1,   1,  -2,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 {  1.342244536341e-05, -7.807291335658e-06, {   1,  -1,  -2,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 {  1.312996373245e-05, -8.081679043162e-06, {   2,  -1,   0,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 {  1.662948784011e-05,  4.573991101148e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  2.111980360791e-05, -7.650420998192e-08, {   2,   1,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.319079840902e-06, -1.786962726244e-05, {   3,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.443157971314e-06,  1.674496186298e-05, {   2,  -1,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.996346551352e-05, -1.208323090268e-06, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.113750610469e-05,  1.704816400145e-08, {   2,   1,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.109002611869e-05, -5.697681413050e-08, {   0,   1,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.088516219395e-05, -2.425033041496e-07, {   4,  -1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.110559218548e-05,  1.367733865192e-08, {   2,  -1,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.408667549135e-05,  7.029134630056e-06, {   0,   1,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.108365787098e-05, -3.009701211941e-08, {   2,   1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.511153954969e-07, -2.075474269340e-05, {   2,  -1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.361882264939e-05, -7.486257987526e-06, {   4,   1,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -4.036245151840e-06,  1.706103268800e-05, {   1,  -2,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.933853885636e-06,  1.916236360613e-05, {   2,  -1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.074765980967e-07, -2.018843340634e-05, {   2,  -1,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.241709429360e-05,  8.678464901648e-06, {   2,   1,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.351244314307e-05, -7.526324235018e-06, {   3,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.419884045718e-05,  6.839912687699e-06, {   2,  -1,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.657212410137e-05, -4.462979401739e-06, {   2,  -1,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.670824162471e-06, -1.636062713956e-05, {   2,   1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  2.101973644989e-05, -5.258542414450e-09, {   2,  -1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.032822045753e-06, -1.698541583645e-05, {   1,  -2,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.940927650241e-09, -2.099440880098e-05, {   2,  -1,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.603096698105e-06,  1.939229238757e-05, {   0,   1,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  7.095465550581e-06, -1.389726117100e-05, {   0,   1,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.992511249468e-05, -1.048489249953e-06, {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -1.811069226637e-05, -2.862578562874e-06, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -5.190267534650e-06,  1.577698055767e-05, {   2,   1,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.397917456437e-05,  6.984316820647e-06, {   0,   1,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.075661628937e-05,  1.996434629304e-07, {   2,  -1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.551808142572e-05, -5.436532382221e-06, {   2,  -1,   0,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 { -1.284658186387e-05, -8.107953436522e-06, {   2,   1,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.233959155511e-05,  8.604200536419e-06, {   2,   1,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -4.941272720210e-07, -2.043691466299e-05, {   2,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.083231015395e-06,  1.583266348765e-05, {   0,   1,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.740158953004e-06,  1.816620911225e-05, {   4,  -1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.463435055427e-05,  6.251892141827e-06, {   4,  -1,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -4.130026916550e-06, -1.675274636363e-05, {   2,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.083942222971e-05, -3.159768215656e-08, {   0,   1,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  7.777488183006e-06, -1.307831510105e-05, {   2,   1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.248057958936e-05,  8.374383123701e-06, {   0,   1,   2,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 {  6.520490453640e-06,  1.433083497617e-05, {   0,   1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  4.405566747157e-06, -1.644436902426e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 {  9.234879687717e-06, -1.160273955146e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  3.986223410495e-06, -1.684405645989e-05, {   3,   0,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.554024195569e-07, -2.043420344172e-05, {   2,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.829744516012e-07,  2.000656985572e-05, {   2,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.776928940698e-05,  2.990106196336e-06, {   1,   0,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.189505696902e-05, -8.858690080300e-06, {   2,  -1,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  2.030641560655e-05,  4.304744528533e-07, {   2,  -1,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.074996919436e-06,  1.265284078047e-05, {   4,  -1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -4.034625614965e-06, -1.668837325834e-05, {   2,  -1,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.533549092786e-06, -1.518891808055e-05, {   2,   1,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.940319978513e-05,  1.319053819064e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  7.789022090030e-06,  1.292508138988e-05, {   1,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  8.331885068044e-06, -1.237448730063e-05, {   0,   1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.973110134473e-06,  1.570841267161e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -2.018628082508e-05,  4.688388572526e-07, {   1,   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -1.384954557678e-05,  6.782620362467e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 {  3.941896945638e-06, -1.668563821175e-05, {   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.789015030526e-05,  2.736487202394e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 {  6.871186826971e-06, -1.374691144220e-05, {   0,   1,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -3.883892199993e-06,  1.673254362264e-05, {   1,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.861094311464e-07, -2.001978071895e-05, {   2,  -1,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  8.187461484636e-06, -1.239869793557e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.043297958846e-05, -1.015283724054e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -2.039065676064e-05,  1.928890151096e-07, {   2,  -1,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -2.644816165601e-06, -1.792752804132e-05, {   2,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -3.873883745499e-06, -1.668887391198e-05, {   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.421392534108e-05,  6.347587532132e-06, {   2,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.083962684976e-05,  9.715867550418e-06, {   0,   1,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.134804149628e-06,  1.441376881139e-05, {   0,   1,   2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  3.246881024390e-06,  1.729963435024e-05, {   0,   3,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.751878287049e-07, -2.036900052601e-05, {   2,  -3,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.615442195099e-06,  1.292479856739e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -7.859516080255e-07,  1.974884972421e-05, {   3,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.862804439643e-06, -1.566349861690e-05, {   1,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.443060151154e-05,  6.094605717961e-06, {   2,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.808033874240e-05,  2.431420699119e-06, {   0,   1,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -4.511062812969e-06, -1.598948100396e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  1.969579972857e-07,  2.029866822685e-05, {   2,   1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.487974442221e-08,  2.044841530990e-05, {   0,   1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.026258762756e-05, -2.201077607640e-07, {   0,   1,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.581503040331e-07, -2.012351579234e-05, {   2,   1,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.918275839326e-06, -1.654720717841e-05, {   3,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.098709639359e-06, -1.635278816210e-05, {   3,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.752946282651e-05, -2.905327412332e-06, {   3,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.420732907562e-07, -1.968954560906e-05, {   2,  -1,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.985556954801e-06, -1.644443983653e-05, {   3,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.287584678980e-06,  1.414188739270e-05, {   1,  -2,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  4.872050950848e-06,  1.555442431297e-05, {   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.446508327598e-06, -1.597242644872e-05, {   0,   1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.217901339557e-06, -1.319711913528e-05, {   2,  -1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -1.384449489517e-05,  6.568545135094e-06, {   0,   1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.353007395620e-07, -1.966305396323e-05, {   2,  -3,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.652447211955e-05,  3.845565507170e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  1.233091468471e-05, -8.037954358946e-06, {   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.679676822378e-05,  3.570510354583e-06, {   1,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.828072866669e-05,  2.080171720491e-06, {   0,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.800846595382e-06,  1.053128818332e-05, {   2,  -1,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -2.026149743477e-05,  6.296572725709e-08, {   2,  -1,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.544651876002e-06,  1.177788750126e-05, {   0,   1,   0,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.026642554712e-05, -4.277076738665e-08, {   2,   1,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.622792231517e-06, -1.467751392418e-05, {   1,   0,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.904397200597e-06, -1.639407915863e-05, {   1,  -2,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.305680512919e-06, -1.598840919154e-05, {   2,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.785552830752e-05, -2.426012426636e-06, {   2,   1,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.907261488968e-06, -1.637279278913e-05, {   0,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.641253667432e-05, -3.865994103796e-06, {   1,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.427611103131e-06, -1.883960279369e-05, {   2,   1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.954471165816e-05, -7.122076666487e-07, {   4,  -1,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.190532668506e-05,  8.350104395532e-06, {   1,   0,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  2.023826125822e-05,  1.706277435097e-08, {   2,  -1,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.108531451890e-06, -1.614678493788e-05, {   1,   0,   1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.844136401748e-05,  1.769301951530e-06, {   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.209470897222e-05,  8.113790068363e-06, {   0,   1,   0,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 { -7.733073618490e-06, -1.247529942033e-05, {   2,  -1,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.669758485327e-05,  3.480626731232e-06, {   3,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.431167607404e-07, -1.952934008464e-05, {   0,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.674130734587e-05,  3.410538337363e-06, {   0,   1,  -1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -4.608352988874e-06,  1.554022739792e-05, {   2,  -1,  -2,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.119533621395e-06,  1.402378732192e-05, {   2,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.223316502909e-05, -7.909315274714e-06, {   4,   1,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -3.873290963670e-06, -1.624176293551e-05, {   1,  -2,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.008647268287e-05, -1.888765807537e-08, {   6,   1,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.340060491718e-05,  6.678838103025e-06, {   0,   1,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -1.995667094320e-05, -1.221587224345e-07, {   0,   1,  -1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.526500551161e-05, -4.813061785721e-06, {   3,  -2,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  8.079775852177e-07, -1.926679885371e-05, {   2,  -3,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.339606159392e-05,  6.673975862067e-06, {   0,   1,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  8.210468528525e-06, -1.184910993591e-05, {   2,  -3,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.303924098854e-05,  7.018721056764e-06, {   2,  -1,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -3.766464544528e-06,  1.628613801217e-05, {   2,   1,   0,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  3.831329962496e-06,  1.621971048243e-05, {   1,   2,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.489697415082e-08,  2.000377508088e-05, {   1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.754143840524e-05,  2.505771850536e-06, {   1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0 } },
 {  1.935744006368e-05, -6.880342312220e-07, {   3,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.624170795738e-06,  1.741652800297e-05, {   0,   1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -7.079796927349e-06, -1.295035077684e-05, {   0,   1,   1,   0,   0, -18,  24, -16,   4,   5,   0,   0,   0 } },
 {  7.079796927349e-06, -1.295035077684e-05, {   0,   1,  -1,   0,   0,  18, -24,  16,  -4,  -5,   0,   0,   0 } },
 { -2.001179990337e-05,  6.089502993794e-09, {   4,  -1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.605556540742e-06, -1.041031488052e-05, {   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.809243481011e-06, -1.520418010204e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  1.999999998865e-05, -6.736970911380e-10, {   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999351e-05, -5.096361415302e-10, {   2,  -2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999377e-05, -4.991641660183e-10, {   0,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999437e-05,  4.747295564979e-10, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999611e-05,  3.944444109251e-10, {   2,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999658e-05, -3.700098014001e-10, {   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999683e-05,  3.560471673880e-10, {   2,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999695e-05,  3.490658503811e-10, {   4,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999713e-05,  3.385938748707e-10, {   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999719e-05, -3.351032163645e-10, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.999999999731e-05,  3.281218993602e-10, {   2,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999731e-05,  3.281218993602e-10, {   1,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999731e-05,  3.281218993602e-10, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.999999999759e-05,  3.106686068425e-10, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.999999999764e-05,  3.071779483389e-10, {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999764e-05, -3.071779483372e-10, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  1.999999999775e-05, -3.001966313278e-10, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.999999999790e-05,  2.897246558209e-10, {   4,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999805e-05,  2.792526803100e-10, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.999999999810e-05, -2.757620218046e-10, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999819e-05, -2.687807047951e-10, {   2,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999824e-05,  2.652900462954e-10, {   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999829e-05, -2.617993877954e-10, {   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999829e-05,  2.617993877917e-10, {   4,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999829e-05,  2.617993877917e-10, {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999833e-05,  2.583087292880e-10, {   4,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999833e-05,  2.583087292880e-10, {   3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999838e-05,  2.548180707843e-10, {   2,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.999999999842e-05, -2.513274122810e-10, {   6,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999842e-05, -2.513274122810e-10, {   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999846e-05,  2.478367537769e-10, {   4,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999851e-05,  2.443460952731e-10, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999855e-05,  2.408554367694e-10, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999855e-05, -2.408554367665e-10, {   3,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999863e-05,  2.338741197619e-10, {   2,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999867e-05, -2.303834612618e-10, {   2,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999871e-05,  2.268928027544e-10, {   2,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999875e-05,  2.234021442506e-10, {   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.999999999883e-05,  2.164208272431e-10, {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.999999999887e-05,  2.129301687393e-10, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.999999999930e-05, -1.675516081931e-10, {   2,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -8.718329945430e-06,  1.127519822351e-05, {   1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.286769957627e-05, -7.120878766773e-06, {   2,  -1,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  8.782879781678e-06, -1.118670258438e-05, {   1,   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  1.583192031748e-05, -4.128434698120e-06, {   0,   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.331133558384e-05,  6.648357566136e-06, {   1,  -1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.554659269451e-06, -1.540317003273e-05, {   2,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.665254133230e-05, -3.295690388762e-06, {   1,   0,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.587246089342e-05,  4.064985468089e-06, {   1,   0,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.985504295776e-06, -1.594125033291e-05, {   4,  -1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.275119165271e-05,  7.169155773017e-06, {   4,   1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  4.156093510636e-06,  1.576109604896e-05, {   2,  -1,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.625327293201e-06,  1.828775976391e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  1.542610433136e-05, -4.483802403241e-06, {   0,   1,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.682092757810e-05, -3.087666436783e-06, {   4,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.887058081096e-07, -1.931831093695e-05, {   0,   1,   2,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.989391144401e-05, -4.815004794410e-09, {   4,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.663579921123e-07, -1.943020555422e-05, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -5.938915610776e-07, -1.928630080378e-05, {   2,  -1,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -8.432279068846e-06, -1.142327947763e-05, {   2,  -1,   0,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -5.114251524671e-06,  1.473996003777e-05, {   1,   1,   0,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 { -5.114251524671e-06,  1.473996003777e-05, {   1,  -1,   0,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 { -5.361208434763e-06, -1.448211590456e-05, {   2,   1,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -1.618556327857e-06,  1.820127742675e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  5.716268704529e-06, -1.409790733984e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -2.110756042500e-08,  1.979056126461e-05, {   1,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  1.071113533250e-05,  9.087343442978e-06, {   2,   1,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  3.786536428788e-06, -1.601189971451e-05, {   1,   0,  -2,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.928368761035e-05, -5.117497565312e-07, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  1.222468339758e-05,  7.570282363622e-06, {   2,   1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.389209375567e-05, -5.902734964489e-06, {   4,  -1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.389209375567e-05, -5.902734964489e-06, {   4,  -1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.411012179103e-06, -1.136991081176e-05, {   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.733179852408e-06, -1.604191955278e-05, {   1,   0,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  8.411339163429e-07,  1.893221018277e-05, {   1,   0,   0,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  1.502684058349e-05, -4.743082465735e-06, {   2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.971223551869e-05,  5.397151116447e-08, {   2,  -1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.615690638696e-06, -1.813398280631e-05, {   2,   1,  -1,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 {  1.615690638696e-06, -1.813398280631e-05, {   2,  -1,  -1,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 {  1.271731919360e-05, -7.024352891897e-06, {   2,  -1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.957267575768e-05, -1.642554085157e-07, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.259700218601e-05,  7.137937244440e-06, {   2,  -1,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 { -3.798678660054e-06,  1.592896044269e-05, {   1,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.705602604107e-05,  2.671596281251e-06, {   6,  -1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -9.254890242976e-06,  1.047032512938e-05, {   1,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.682312846698e-05,  2.891898186114e-06, {   3,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.094700090763e-06,  1.061726910276e-05, {   1,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.848420582130e-06, -1.583314898212e-05, {   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.887784194019e-06,  1.378376352478e-05, {   1,   0,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.801118162621e-05,  1.660039783008e-06, {   1,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.255105700401e-05, -7.117275652360e-06, {   2,  -1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  5.152005599229e-06,  1.451023109426e-05, {   4,  -1,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -5.151269545895e-06,  1.450791184212e-05, {   4,  -1,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.402230498875e-05,  5.635739332335e-06, {   1,  -2,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  1.279467794135e-05,  6.854048186471e-06, {   1,   0,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -1.963149739473e-05, -1.705963659690e-08, {   2,  -3,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.313990211678e-05, -6.489578626997e-06, {   2,   1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.873880859134e-06, -1.074056427907e-05, {   1,   0,   0,   0,   0,  10,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.636331536611e-05,  3.238767277207e-06, {   1,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.117777720575e-06, -1.844856619459e-05, {   4,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.937479050806e-05, -1.813215745908e-07, {   4,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.584872473230e-05,  3.698969128144e-06, {   2,  -1,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -1.833776294972e-05, -1.209255266197e-06, {   2,  -1,   0,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 { -1.584450439867e-05,  3.698674038530e-06, {   2,  -1,  -1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.950824900998e-05,  4.659561256690e-09, {   2,   1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.915389090275e-05,  3.563138175403e-07, {   4,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.594411525512e-05,  3.556536825192e-06, {   2,  -1,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.663782621232e-05,  2.860872632226e-06, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -3.253272458424e-06, -1.624170280729e-05, {   2,  -1,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.886313942188e-05,  6.074623690049e-07, {   2,   1,   0,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  1.801634677451e-05, -1.449338301806e-06, {   2,  -1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.941290406623e-05,  4.295458514087e-08, {   2,  -1,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.573529643283e-06, -1.586584542365e-05, {   3,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.344320021445e-06,  1.209184598504e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  2.933861985049e-06, -1.649854037630e-05, {   2,  -1,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.932206053371e-05,  1.060437677441e-07, {   0,   1,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -6.253670997358e-06,  1.316380737475e-05, {   0,   1,  -1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 {  7.267617036717e-06, -1.214667758640e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  6.765861427848e-06, -1.264390444511e-05, {   1,   0,   1,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 {  6.765861427848e-06, -1.264390444511e-05, {   1,  -2,   1,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 { -7.487412659972e-06,  1.191464310683e-05, {   2,  -1,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -3.705214603476e-06, -1.569010880644e-05, {   0,   1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.744118668701e-07,  1.841151265000e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   3,   0,   0,   0 } },
 {  1.246769938564e-05, -6.914255855085e-06, {   2,   1,   0,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  6.179295029775e-06, -1.317974582524e-05, {   2,   1,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -1.219533303365e-05, -7.160791977371e-06, {   2,   1,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.465483526301e-06, -1.288612889620e-05, {   2,  -1,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.172952811950e-05,  7.621867155804e-06, {   2,  -1,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  8.456932346306e-06,  1.088472294873e-05, {   2,  -1,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.489785312174e-05,  4.441039143437e-06, {   0,   1,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -1.305177740017e-05,  6.287032923471e-06, {   2,  -1,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.906300290007e-05, -2.681764512011e-07, {   3,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.900373474672e-05, -3.138019929337e-07, {   0,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.905510957359e-05,  2.499533707874e-07, {   2,  -1,  -1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  1.876569113391e-05,  5.379405069183e-07, {   4,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -6.540924381114e-06,  1.275578945366e-05, {   1,   0,   1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.540924381114e-06,  1.275578945366e-05, {   1,  -2,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.620153049659e-07, -1.850678381016e-05, {   2,  -1,   3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.607246432904e-05,  3.193480281655e-06, {   0,   1,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -4.251093718024e-06,  1.500623870430e-05, {   1,  -1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.674616389865e-06,  1.557814532559e-05, {   1,   2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.161218220826e-06, -1.108592628695e-05, {   2,  -1,   0,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 {  1.145460416509e-05,  7.783137985849e-06, {   2,  -1,   0,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  4.246672102714e-06,  1.498669695740e-05, {   1,  -1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.909998541223e-05,  1.271216808621e-07, {   0,   1,   3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  9.215047604495e-06,  9.993246721986e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   7,   0,   0,   0 } },
 { -1.163392591125e-05,  7.559456262681e-06, {   2,  -1,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  6.052354928247e-06,  1.313797543887e-05, {   1,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.351073974513e-05,  5.669394477801e-06, {   1,   0,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -1.895523462858e-05, -2.178969298344e-07, {   2,  -1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.345067580421e-05,  5.715175044336e-06, {   1,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.345067580421e-05,  5.715175044336e-06, {   1,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.796964196865e-05,  1.181334429851e-06, {   4,   1,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.853184807099e-05, -6.105660681131e-07, {   0,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.013413563536e-06,  1.512752754457e-05, {   2,   1,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.122850366358e-06,  1.800013150932e-05, {   1,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.138508507433e-05,  7.730909943761e-06, {   2,  -1,   0,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 {  7.192792117478e-06,  1.190878926667e-05, {   1,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.745256125812e-06,  1.535396667571e-05, {   1,   0,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.815679994327e-06, -1.328066324392e-05, {   1,   0,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.270821166095e-05,  6.381938409925e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  1.851866592302e-05, -5.712622743695e-07, {   0,   1,   1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -1.638996470632e-05,  2.696628744841e-06, {   2,   1,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  1.274816095379e-05,  6.338352367572e-06, {   0,   1,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.048554127659e-05, -8.600597653551e-06, {   1,   2,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.128186730577e-05, -7.778020194300e-06, {   2,  -1,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.449114660196e-05, -4.562162057158e-06, {   1,   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.366853224420e-05, -5.383481105079e-06, {   1,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.701443356376e-06,  1.734669996381e-05, {   1,   0,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  7.387665950233e-06,  1.166017377426e-05, {   4,  -1,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.635590601947e-05,  2.687173856630e-06, {   2,   1,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -5.685226744212e-07, -1.847451798159e-05, {   2,   1,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  6.055630586967e-06, -1.297414456527e-05, {   2,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.337677973161e-06,  1.768214842154e-05, {   0,   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.517177302737e-05, -3.845873399985e-06, {   1,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.820414869550e-06,  1.719422011932e-05, {   2,   1,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.899670351597e-05, -1.594993852174e-08, {   4,  -1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.636955821939e-06,  1.536565527573e-05, {   3,   0,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.896667886255e-05,  2.822156926926e-08, {   0,   1,  -1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.568421095060e-05,  3.307956865573e-06, {   4,  -1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.867698559180e-05,  3.088989575964e-07, {   0,   1,   0,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -1.867698559180e-05,  3.088989575964e-07, {   0,   1,   0,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -7.102574524455e-06, -1.186687031666e-05, {   2,  -1,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.623518837959e-05,  2.730697194710e-06, {   2,  -1,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.539103995417e-05, -3.571523059323e-06, {   4,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.114290153282e-06,  1.784756714542e-05, {   4,  -1,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -1.103829189363e-05,  7.920383700221e-06, {   1,   0,  -2,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.210090681550e-05, -6.827445279025e-06, {   2,   1,  -1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  7.090824370520e-06,  1.182457244300e-05, {   1,   2,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  7.509596827057e-06,  1.139193648278e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.877365770715e-05, -1.191699606501e-07, {   2,   1,  -1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  1.378586497002e-06,  1.750976793515e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.397746239716e-06, -1.548802804892e-05, {   1,   0,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.585227815222e-05, -3.031848325883e-06, {   1,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.259161653915e-05, -6.288895549949e-06, {   2,  -1,  -5,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.344231778063e-05, -5.395583131813e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 { -1.160401555647e-05,  7.226358799985e-06, {   1,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.877583098127e-05,  4.769323358570e-08, {   2,   1,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.156050041292e-06,  1.666368449188e-05, {   2,  -1,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  3.234003764935e-06, -1.557827761488e-05, {   2,  -1,   0,   0,   0,   0,   3,  -2,   2,   0,   0,   0,   0 } },
 { -7.091445092302e-06,  1.171969291611e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.158525144807e-05,  7.224261716248e-06, {   1,   0,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  1.414051661235e-05,  4.660496660402e-06, {   2,   1,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  1.777010608539e-05, -1.029259196355e-06, {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -1.152021127125e-05, -7.274901450384e-06, {   2,   1,   0,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  2.674348211956e-06,  1.610946237081e-05, {   2,  -3,   1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.876250678302e-05,  1.878597232330e-08, {   2,   1,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.860761081520e-06,  1.691234925173e-05, {   0,   1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.567256067477e-06,  1.516718454966e-05, {   1,   0,   1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.869721383452e-05,  2.785030025541e-08, {   0,   1,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.251986170437e-05, -6.202876308525e-06, {   1,   1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.251986170437e-05, -6.202876308525e-06, {   1,  -1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.346344588579e-05, -5.228198055532e-06, {   2,  -1,   1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -5.243640252274e-06,  1.341111368416e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -1.204363725659e-05, -6.606585596481e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   3,  -4,   0,   0,   0 } },
 {  1.293955668685e-05,  5.701942955214e-06, {   4,  -1,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.471905511550e-05, -3.918705970481e-06, {   2,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.416815825151e-05, -4.461333091063e-06, {   0,   1,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.758803774436e-05, -1.032774831518e-06, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -3.615449322918e-06, -1.500423027882e-05, {   1,   2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.841567346271e-06,  1.074888215078e-05, {   2,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.219586833039e-06, -1.435680942595e-05, {   1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.255564784456e-05, -6.013201459559e-06, {   1,   2,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.494048458397e-05,  3.628085607542e-06, {   1,  -2,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.190594380539e-05,  6.658571872923e-06, {   2,  -1,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -1.804793307194e-05,  5.074150523249e-07, {   2,  -1,   0,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 {  7.467126489999e-06, -1.107958287578e-05, {   4,   1,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.243928027851e-05,  6.105176680074e-06, {   0,   1,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.190189697606e-05, -6.637135738193e-06, {   2,  -1,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -4.275132530193e-06,  1.426313498191e-05, {   3,  -1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.852651870786e-05, -7.622888270038e-09, {   6,   1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.022774964816e-06,  1.550979202608e-05, {   2,  -3,   0,   0,   0,   0,   3,  -2,   2,   0,   0,   0,   0 } },
 {  4.270850198737e-06,  1.424988175003e-05, {   3,  -1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.465547665391e-06,  1.303921681426e-05, {   0,   1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.418087097468e-05,  4.323519760447e-06, {   0,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.188658122835e-05,  6.612871463404e-06, {   1,  -1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.237284146594e-05,  6.103062562070e-06, {   0,   1,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.216680935301e-06,  1.725481418962e-05, {   2,  -1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  8.139627244346e-06,  1.032510717230e-05, {   1,   0,   0,   0,   0,  -9,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.820287428874e-05, -2.479041400930e-07, {   4,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.780454796887e-05, -6.345180325848e-07, {   2,  -1,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -3.585536369293e-06, -1.485142633842e-05, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.085298787872e-05,  7.583927115597e-06, {   0,   1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  1.540421055853e-05,  3.031583143273e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  1.138006526143e-05, -7.026139391950e-06, {   2,   1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.338454335184e-06,  1.705987227386e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.637444616929e-05, -2.017452719748e-06, {   1,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.827736390099e-05,  1.096389357024e-07, {   2,   1,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -5.372741488637e-06,  1.301395319630e-05, {   1,   0,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.658247974370e-06, -1.272534591169e-05, {   1,   2,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.518465107099e-06, -1.486275284708e-05, {   1,   2,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.527912935313e-06,  1.485191753003e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -7.113773169171e-06, -1.126402684972e-05, {   0,   1,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  1.823193818367e-05,  1.372954768179e-07, {   0,   1,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -3.522641477269e-06,  1.483191316350e-05, {   3,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.004576259330e-05,  8.303012366966e-06, {   2,  -1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -8.966008904361e-06, -9.377916979968e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 {  1.501562500312e-05,  3.326999953070e-06, {   0,   1,   2,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 { -1.501562500312e-05,  3.326999953070e-06, {   0,   1,   0,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -1.833173644317e-05, -2.128754459621e-09, {   2,  -1,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.016203048104e-05,  8.166984755763e-06, {   0,   1,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -2.823257238701e-06,  1.550467321673e-05, {   1,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -7.446640629142e-07,  1.755867845830e-05, {   2,   1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.878124845015e-07,  1.808847797332e-05, {   2,  -1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.216707690557e-05,  6.091219266797e-06, {   0,   1,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.499648115841e-06, -1.475410736344e-05, {   2,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.452384040488e-05,  3.727076802191e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  3.441799244556e-06, -1.480558046342e-05, {   0,   1,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.803049354424e-06, -1.144096613797e-05, {   0,   1,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  1.406404128118e-05,  4.177473060486e-06, {   0,   1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.235454201156e-05,  5.884915257000e-06, {   2,   1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -1.519640881152e-05,  3.023700874796e-06, {   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -5.905363427164e-07, -1.762875691380e-05, {   0,   1,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -8.240521426070e-06,  9.974451041066e-06, {   2,  -1,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.644186302350e-06,  1.253166143301e-05, {   1,   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  1.781281391103e-05, -3.578016668251e-07, {   4,  -5,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -1.815076378698e-05, -1.906594476561e-08, {   2,  -3,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.991718990244e-06, -1.417523563933e-05, {   2,  -1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.798598207461e-05,  1.647964205962e-07, {   0,   1,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.805229052762e-05, -8.392439626712e-08, {   0,   1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.822741493989e-06,  1.431030891150e-05, {   2,   1,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.388418202204e-05,  4.244625367713e-06, {   0,   1,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.094699039203e-06, -1.303031922737e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -2.834319432530e-06, -1.528335468800e-05, {   1,   2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.294892273125e-07,  1.756930599608e-05, {   2,  -3,   0,   0,   0,   0,   2,   0,   0,   3,   0,   0,   0 } },
 {  5.973080910889e-06,  1.211493356225e-05, {   2,  -1,   0,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 { -6.085370629480e-07, -1.747833416383e-05, {   4,  -1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.402655255515e-06,  1.468256237049e-05, {   2,  -1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.389356650304e-06,  8.687533341613e-06, {   0,   1,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.802384701695e-05, -4.767044528300e-08, {   1,  -1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.234449176736e-05,  5.721155295074e-06, {   2,   1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.802211288863e-05,  2.806982884139e-08, {   4,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.373393405753e-05, -4.311436166332e-06, {   0,   1,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.260502303118e-05,  5.439318840502e-06, {   0,   1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.419801866463e-06, -1.261888423318e-05, {   2,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.511788508593e-05,  2.913233835656e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.598815453670e-05, -2.038450545099e-06, {   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.551921859336e-06, -9.474215054140e-06, {   0,   1,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -5.941117067739e-06,  1.206788543976e-05, {   2,  -1,   0,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 {  1.214404653531e-05,  5.862874514873e-06, {   1,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.574634893071e-05, -2.240112828447e-06, {   2,  -1,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  1.786387963545e-05, -1.146546684822e-07, {   3,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.416320865866e-05, -3.811257089466e-06, {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.797136626291e-05, -1.761860724873e-09, {   0,   1,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.415775288549e-05, -3.810401392474e-06, {   0,   1,  -3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.790797174717e-05,  5.121134907540e-08, {   2,  -1,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.428160237746e-06, -1.452986027125e-05, {   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.443543797933e-06,  1.450715725270e-05, {   1,   0,  -3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.610164530642e-07, -1.718751085914e-05, {   0,   1,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.765478802282e-05,  2.893561495997e-07, {   2,   1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.473268171499e-05,  3.190813135450e-06, {   1,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  6.102075096116e-06, -1.179735121846e-05, {   0,   1,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 {  6.371347686278e-07, -1.726011028801e-05, {   0,   1,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.576959654748e-05,  2.124477857928e-06, {   0,   1,  -1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  8.435320772148e-07, -1.703371976208e-05, {   2,  -1,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.583900091254e-05,  2.033765987542e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  1.531282073285e-05,  2.557675818286e-06, {   2,   1,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.764953596149e-05, -2.081875141888e-07, {   2,  -1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  4.724678892247e-06, -1.312007068737e-05, {   2,  -1,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.490988945058e-05,  2.908543419228e-06, {   1,   0,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.935095209758e-06, -1.287626181097e-05, {   1,  -2,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  7.739243940722e-06,  1.007171751245e-05, {   2,   1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -7.313989395318e-07,  1.706090049626e-05, {   2,  -1,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -1.761549303532e-05,  1.591455099196e-07, {   0,   1,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.336365586948e-05, -4.406139008995e-06, {   2,  -1,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.966342382139e-06,  1.279568397383e-05, {   1,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.508578362127e-05,  2.672067682389e-06, {   0,   1,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.370434305015e-05,  4.049573720613e-06, {   2,  -1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.080942875120e-06,  1.266635809892e-05, {   0,   1,   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.645965656963e-05,  1.280098698427e-06, {   2,   1,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.695426149536e-06, -1.404265588176e-05, {   1,   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  6.370409481502e-06, -1.136729784872e-05, {   1,   0,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.480180048913e-05, -2.925135386164e-06, {   3,   0,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -1.086744220830e-05, -6.858194378322e-06, {   2,  -1,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.764950625447e-05,  7.489072125307e-08, {   2,   1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -6.031065503930e-06, -1.168939345348e-05, {   2,   1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.771144871272e-05, -4.984623267772e-09, {   3,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.260513542147e-05, -5.110372160956e-06, {   4,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.230661160677e-05,  5.390189372574e-06, {   2,  -1,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.714724823043e-05, -5.480150157967e-07, {   2,  -3,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.467402052416e-05,  2.989768786761e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  1.763792929294e-05,  1.894660079673e-08, {   3,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.780362992951e-08, -1.755810729928e-05, {   4,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.379752385024e-06,  1.427388135573e-05, {   1,  -2,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.740229109315e-07,  1.707442466460e-05, {   2,  -1,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.853551316590e-06, -1.279288165391e-05, {   0,   0,   1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.588458100103e-05, -1.761085717906e-06, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 { -1.763781644196e-05,  4.576327602213e-09, {   0,   1,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.466253832762e-06, -1.115108580151e-05, {   4,  -1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.502451909695e-05,  2.586387222554e-06, {   1,   0,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.673666427242e-05, -8.643180632733e-07, {   2,  -1,   0,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 { -1.722978435259e-05, -3.698866543700e-07, {   0,   1,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  1.752025589096e-05,  6.259939772142e-08, {   0,   1,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.643740448302e-06,  1.192414472258e-05, {   2,  -1,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  1.397145971542e-05, -3.565607166984e-06, {   1,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.262526757645e-06, -1.426992291846e-05, {   2,   1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.301840410869e-06,  1.322798011772e-05, {   0,   1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.435125350127e-05,  3.162877248496e-06, {   0,   1,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.167329826571e-05,  5.830265853158e-06, {   6,  -1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.167329826571e-05,  5.830265853158e-06, {   6,  -1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.328181556548e-06, -1.417233350838e-05, {   1,   0,   1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.190436710216e-06, -1.430906464633e-05, {   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.269439098584e-06,  1.422624331509e-05, {   1,  -2,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.243262708906e-05,  5.061259195589e-06, {   2,  -1,   0,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 { -1.696821572210e-05,  5.149015220037e-07, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -1.624771191793e-05,  1.229230140964e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  4.307515014452e-06,  1.315814891114e-05, {   0,   1,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.761789614289e-06,  1.369440398505e-05, {   1,   0,  -1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.585528202903e-06, -1.086682773115e-05, {   1,   0,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  3.344809102690e-06,  1.410542901794e-05, {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.224494011602e-05,  5.201620462739e-06, {   4,  -1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.174454218982e-05,  5.694425919087e-06, {   0,   1,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.239148929606e-05, -5.044998061805e-06, {   2,  -1,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.124674501045e-05, -6.186931642738e-06, {   0,   1,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  3.344983253452e-06,  1.407822571429e-05, {   1,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.334950752144e-06, -1.408737037311e-05, {   3,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.903484341588e-06,  1.151268519634e-05, {   1,   0,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.903484341588e-06,  1.151268519634e-05, {   1,  -2,   1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.284108422360e-05, -4.562208074770e-06, {   4,  -1,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.080532742229e-06, -1.331882272002e-05, {   1,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.446000185874e-05, -2.932952791917e-06, {   1,   0,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.515130085884e-06,  1.287683412237e-05, {   1,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.260852846009e-05, -4.778366799283e-06, {   1,   0,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -1.532121560783e-05,  2.063685821149e-06, {   0,   1,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  1.235547853242e-05, -5.011825909649e-06, {   2,  -1,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.728783354021e-05, -7.793109790886e-08, {   0,   1,   3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.025528267567e-05, -7.107071496672e-06, {   1,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.268169128615e-06,  1.607532545122e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  1.604671643153e-05, -1.293311361715e-06, {   4,   1,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  9.857684615217e-06, -7.476664135555e-06, {   0,   1,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.621343972353e-05, -1.111617747788e-06, {   1,   0,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  1.155378660183e-05, -5.770966754661e-06, {   2,  -1,  -3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.464239095699e-05, -2.682143707669e-06, {   2,   1,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.726631449749e-05,  4.793559558128e-08, {   2,   1,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.485692765744e-05, -2.443215203936e-06, {   2,   3,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.333357621352e-05,  3.960782340569e-06, {   0,   1,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -6.621757245610e-06,  1.066776517335e-05, {   2,   1,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -9.145320238781e-06,  8.144083910037e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  4.012795580526e-06,  1.326844781569e-05, {   2,  -1,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.325247613424e-06,  1.395273980963e-05, {   2,   1,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  1.724282016742e-05, -3.096682177415e-08, {   0,   1,   1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.199744381891e-05, -5.269952240304e-06, {   3,  -1,   0,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 { -1.199744381891e-05, -5.269952240304e-06, {   3,  -3,   0,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 { -1.016302069619e-05,  7.099704733558e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -4.637099966632e-06,  1.262186992119e-05, {   1,   2,  -1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -3.339398618498e-06,  1.391604477558e-05, {   1,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.563427399148e-05, -1.619995262107e-06, {   0,   1,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.463172876558e-05,  2.612637851991e-06, {   0,   1,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  3.295788546910e-06,  1.394467996798e-05, {   1,   0,  -2,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.699690755321e-05, -2.360852717502e-07, {   2,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.717520213016e-05,  5.543000569522e-08, {   4,  -1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.645325011807e-05, -7.677430380350e-07, {   0,   1,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.154962926534e-06,  1.205889018758e-05, {   2,   1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.660013953261e-06,  1.154782847552e-05, {   0,   1,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  1.441826664235e-05,  2.779944318992e-06, {   1,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  8.921819517360e-06,  8.274019124753e-06, {   2,  -1,   0,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.706879208601e-05, -1.161031559514e-07, {   4,   1,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -5.662513078718e-06,  1.152160854780e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 {  5.662247681510e-06,  1.152106819047e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 {  4.840175030842e-06,  1.234304160672e-05, {   2,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.716617181213e-05, -1.811902816325e-09, {   0,   1,   2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.338002432736e-06, -1.282741452823e-05, {   2,  -1,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0 } },
 { -3.797498111575e-06,  1.336173355232e-05, {   0,   1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  9.795948130356e-06,  7.361108955765e-06, {   1,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0 } },
 { -3.456145340850e-07,  1.681082687472e-05, {   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.641501467261e-07, -1.667991511241e-05, {   4,  -1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.009829766672e-05,  7.037725365632e-06, {   2,  -1,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  1.437345777969e-05,  2.761486578586e-06, {   1,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.107824168923e-06, -1.202675236551e-05, {   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.282952916505e-06,  1.384762309072e-05, {   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.105284129251e-05,  6.045738465200e-06, {   1,   1,   0,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 { -1.105284129251e-05,  6.045738465200e-06, {   1,  -1,   0,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 {  1.599105952555e-05,  1.105992873083e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  1.135357151748e-05, -5.729062964773e-06, {   0,   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.090334132742e-05, -6.166660481507e-06, {   4,  -1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.254785015320e-05,  4.519972726376e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -1.254785015320e-05,  4.519972726376e-06, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  8.847750918724e-06,  8.205335844275e-06, {   2,   1,   0,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.437601457133e-05,  2.675498770846e-06, {   2,   1,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  4.640297152476e-06, -1.240650837732e-05, {   1,   0,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 { -3.304971177259e-06,  1.372133485685e-05, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.297090024401e-06, -1.572261751966e-05, {   2,  -1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.683965320683e-05,  1.775666551067e-07, {   2,  -1,   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.675530047300e-05, -2.430085077254e-07, {   2,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  1.038975272530e-05,  6.606835114408e-06, {   2,  -1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.190414487035e-05,  5.088088476510e-06, {   4,  -1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  6.589120765305e-06,  1.040129454809e-05, {   0,   1,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  2.924399575173e-06, -1.406337070960e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  1.267990706412e-05, -4.304296801252e-06, {   0,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.281233093213e-07, -1.685029964067e-05, {   0,   1,   3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.194453263487e-06, -1.076208084181e-05, {   1,   0,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -7.618496394464e-06,  9.336946811724e-06, {   2,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  7.192274631345e-06,  9.756595562889e-06, {   4,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -7.758048518879e-06, -9.183537997345e-06, {   4,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.691485352870e-05, -1.400541259142e-08, {   3,   1,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.691485352870e-05, -1.400541259142e-08, {   3,  -1,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -8.927626071759e-07,  1.602482539089e-05, {   2,  -1,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.381050849835e-06, -1.553560586275e-05, {   2,  -1,   0,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 { -1.574374791828e-05,  1.167483630558e-06, {   0,   1,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -3.214340554775e-06, -1.368642109518e-05, {   2,  -1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.400790154274e-05,  2.892766954809e-06, {   0,   1,   2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.081501596845e-05,  6.079536103291e-06, {   2,  -1,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -1.126072553817e-05,  5.626043594136e-06, {   1,  -1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.673055788648e-05, -1.515474337681e-07, {   0,   1,   1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.380719257182e-05, -3.071957810653e-06, {   1,   0,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.666477201557e-05,  2.134142318066e-07, {   2,  -3,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.548133626233e-06, -1.432608546314e-05, {   0,   1,   2,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.644738235051e-06, -1.421798399620e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.247376893148e-06,  1.361223367347e-05, {   2,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.509417485807e-05,  1.752366743555e-06, {   2,   1,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  1.672930236571e-05,  1.172342256600e-07, {   0,   1,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.374652971544e-05,  3.098870761196e-06, {   1,  -2,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -7.255234189083e-06, -9.582339158771e-06, {   4,  -1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.378584605745e-06,  8.450762193001e-06, {   2,  -1,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.859827083420e-06, -1.095952725832e-05, {   1,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.382744721309e-05, -2.985756005875e-06, {   3,   1,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.387344018264e-05, -2.937549594116e-06, {   1,   0,   0,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.740415963048e-06, -1.306991089068e-05, {   2,  -1,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  6.510190351012e-06, -1.029167010535e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 { -7.075260533764e-06, -9.724206340775e-06, {   2,  -1,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -7.678341921508e-06,  9.114603494101e-06, {   2,   1,   0,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.624552348670e-05, -5.446840314740e-07, {   2,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  7.759006161070e-06, -9.016959186379e-06, {   0,   1,  -3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  6.898189406541e-07,  1.608193167141e-05, {   3,   0,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.589420369713e-05,  8.677509832055e-07, {   1,   1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.589420369713e-05,  8.677509832055e-07, {   1,  -1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  6.468783821935e-07, -1.611001749778e-05, {   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.654943524366e-05, -2.065078115790e-07, {   0,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.498690377921e-06, -1.225370012248e-05, {   4,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -7.576670728908e-06, -9.175264993887e-06, {   0,   1,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.655521145245e-05,  1.961034812402e-07, {   3,  -1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.146480288145e-06, -1.360029538475e-05, {   0,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.461724334417e-06, -1.028333669662e-05, {   1,   0,   0,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 {  3.935902157932e-06, -1.279435927690e-05, {   1,   0,   0,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.504186763249e-05, -1.671461248631e-06, {   2,   1,  -1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 { -1.669621819882e-05, -1.570633642836e-08, {   2,  -3,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  6.754143800379e-06,  9.956437160094e-06, {   1,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.628801972252e-05,  4.210045768636e-07, {   3,   2,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 {  1.628801972252e-05,  4.210045768636e-07, {   3,   0,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 {  1.654706031998e-05, -1.568554268603e-07, {   0,   1,   1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.669739906309e-05, -6.420093456205e-09, {   3,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.936563480062e-07, -1.620853363998e-05, {   4,   1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.110499377280e-06,  1.358139674194e-05, {   1,  -2,  -1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.071664643698e-05, -5.949080318021e-06, {   1,   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.369128788937e-05,  2.973635456965e-06, {   2,  -1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.141950681403e-07,  1.624769866472e-05, {   2,  -1,  -3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.371227147221e-06,  1.028888014498e-05, {   0,   1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  9.495012783973e-06,  7.161445436705e-06, {   0,   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  1.259915619890e-05,  4.055294964590e-06, {   2,  -1,   0,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -1.664817222185e-05, -4.981256621307e-09, {   6,  -1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.477433688352e-05, -1.870647561773e-06, {   3,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.120787958012e-06,  1.151726519073e-05, {   3,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.537008190723e-05,  1.258986014870e-06, {   1,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.659090571765e-05,  3.779747838793e-08, {   2,   1,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.258944711471e-05,  4.036812924633e-06, {   2,  -1,   0,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 {  1.439531165947e-05, -2.206795651322e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -3.231367161743e-06, -1.336377212726e-05, {   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.156836270813e-05, -5.008599381052e-06, {   2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.167846398637e-06, -1.340875815215e-05, {   3,   0,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.960470466914e-06, -9.614436247148e-06, {   4,  -1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.118997390855e-05,  5.380722917082e-06, {   2,  -1,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 { -7.482318989363e-06, -9.084100228902e-06, {   4,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  9.617176007312e-06,  6.938408680790e-06, {   2,  -1,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -1.116466993090e-05,  5.389545402123e-06, {   2,  -1,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.802309204280e-06, -6.736044183390e-06, {   1,   0,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.582794352002e-07, -1.587506720308e-05, {   4,  -3,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.514526868619e-06,  8.010475197282e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -1.651767566079e-05,  5.820769071371e-09, {   3,   1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.769093468670e-06,  8.748784462369e-06, {   1,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.195352817391e-06,  1.329582690106e-05, {   4,   1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.362954369670e-05, -2.859708645474e-06, {   4,  -1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -5.733181950847e-06,  1.074250935155e-05, {   1,   0,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.148311708071e-06,  1.332043169715e-05, {   1,  -2,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.339661428690e-05, -3.060699715885e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  5.402293266675e-07, -1.591329527109e-05, {   0,   1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.446159900500e-06, -1.200445152231e-05, {   2,  -1,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -3.910980722420e-06,  1.253437991388e-05, {   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.235173418349e-06,  1.220886406206e-05, {   2,   1,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.572264188597e-06,  1.386881223411e-05, {   1,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.420047266565e-05, -2.240308150672e-06, {   1,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.640730291029e-05, -2.803567561888e-08, {   1,   1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.713470818399e-07, -1.586316770269e-05, {   0,   1,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  3.075064434543e-06,  1.335373688634e-05, {   1,   0,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.065544026272e-05, -5.771027009258e-06, {   2,   1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.559255703114e-05, -8.290280462622e-07, {   4,  -1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -6.105846038221e-06,  1.030225839886e-05, {   1,   0,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -4.529795404385e-06,  1.187379290614e-05, {   1,   0,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.121675494700e-05, -5.185604013609e-06, {   1,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.639996682264e-05, -4.607842588958e-10, {   1,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.263847573222e-05,  3.752696863420e-06, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.529987114314e-05, -1.086714431403e-06, {   0,   3,  -1,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  1.443283643475e-05,  1.944142319490e-06, {   0,   1,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -3.135855833168e-06, -1.323957708463e-05, {   1,  -2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.415613869717e-05, -2.213911293107e-06, {   2,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.636373258703e-05, -4.198654749984e-09, {   3,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.409196115512e-06,  1.395847088552e-05, {   3,   1,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.409196115512e-06,  1.395847088552e-05, {   1,   1,  -1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.154971080833e-06,  1.119369262941e-05, {   0,   1,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.095762056738e-05,  5.388725518787e-06, {   2,  -1,  -1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.045860974264e-06, -1.329599351528e-05, {   1,   0,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  9.924770758431e-06, -6.410213132684e-06, {   2,   1,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.737135095803e-06,  1.259684764228e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  1.632257352284e-05,  9.160563403372e-09, {   2,   1,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.591330970491e-06,  1.373780363883e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  3.242612120325e-06, -1.307935029904e-05, {   4,   1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.043973983931e-07, -1.591669554286e-05, {   2,  -1,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.322276512128e-06,  9.994727076512e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  5.527192359866e-06, -1.078732630252e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   3,   0,   0,   0 } },
 { -5.538313723006e-06, -1.077492175690e-05, {   2,  -1,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -3.065177999509e-06, -1.324701310255e-05, {   1,  -2,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.838215534815e-06,  1.446485544542e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 { -3.175328210567e-06, -1.312265097791e-05, {   2,  -1,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.606440512134e-05, -2.250675323866e-07, {   1,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.115382901695e-05, -5.128749433656e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 { -1.054704427490e-05,  5.731661109523e-06, {   0,   1,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  7.648261679054e-06, -8.621899664437e-06, {   2,   1,   1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.110961729513e-06,  1.315460547357e-05, {   1,   2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.374075864084e-05,  2.524230583013e-06, {   2,  -1,  -3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.616126902385e-05,  1.036843068020e-07, {   0,   3,  -1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.038088727910e-05, -5.882317594224e-06, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  1.373446819991e-05,  2.524854474390e-06, {   2,  -1,  -3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -3.079677539600e-06, -1.317866023306e-05, {   1,   0,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.034114632624e-05, -5.897159207300e-06, {   0,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  4.485506350685e-06, -1.175073892610e-05, {   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.742103915847e-06,  7.484669075663e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  6.381212907777e-06,  9.843571137071e-06, {   0,   1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  1.158431007263e-05, -4.636582104112e-06, {   1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.604846222124e-05, -1.689719240137e-07, {   2,  -1,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  8.067312778229e-06,  8.145486392765e-06, {   0,   1,   2,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.096138712552e-06, -1.311486772627e-05, {   1,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.825567176091e-07, -1.592524129836e-05, {   0,   1,   2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  3.113411917495e-06,  1.309243651912e-05, {   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.718891745938e-06, -1.047598701167e-05, {   4,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -5.076988576582e-06,  1.111578521934e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.417453741536e-05, -2.015219861434e-06, {   2,  -1,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.082303855411e-06,  1.109708908991e-05, {   0,   1,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -3.612190444144e-06,  1.256336209343e-05, {   2,  -3,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.192333601388e-05, -4.243231874061e-06, {   1,   0,   0,   0,   0,   0,  -2,   6,  -2,   0,   0,   0,   0 } },
 {  1.192333601388e-05, -4.243231874061e-06, {   1,  -2,   0,   0,   0,   0,  -2,   6,  -2,   0,   0,   0,   0 } },
 {  3.608787797374e-06,  1.255679511797e-05, {   2,  -3,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.611282940061e-05,  4.801683065456e-08, {   2,  -1,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.903456174031e-06, -8.238718609817e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -4.767838111484e-07, -1.566513684667e-05, {   4,  -1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.094689853822e-05,  5.194046889512e-06, {   2,   1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.077790396696e-05,  5.348963820687e-06, {   0,   1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.253816063586e-07, -1.529809646413e-05, {   0,   1,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -1.437606326521e-05, -1.745604101818e-06, {   2,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  2.449943231279e-06, -1.366771894193e-05, {   3,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.081556238412e-05,  5.301412816952e-06, {   2,   1,   0,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.495821264634e-05,  1.157471084877e-06, {   4,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.504508753498e-05,  1.057569879379e-06, {   0,   1,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -3.974948202581e-06,  1.212129865971e-05, {   0,   1,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.439825459364e-05, -1.696386700771e-06, {   2,  -1,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.303882728514e-05,  3.055763476622e-06, {   1,   2,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -8.474011557563e-06,  7.609781848356e-06, {   0,   1,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.071738362048e-05,  5.365443388150e-06, {   0,   1,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.517020179042e-05,  8.994410990811e-07, {   6,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.062060415070e-05,  5.441881852791e-06, {   2,   1,  -1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.745528963135e-06,  1.031548460789e-05, {   2,  -1,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.121896848738e-06, -1.293760200502e-05, {   0,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.292458299793e-05, -3.131755403895e-06, {   2,   1,   0,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -7.479813515955e-06, -8.574460976609e-06, {   0,   1,  -3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.604846229823e-05,  2.306776052799e-10, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 {  1.599106638722e-05,  4.716774765051e-08, {   2,  -1,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.108872493608e-06,  1.291677117601e-05, {   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.509780867222e-06, -1.050419119376e-05, {   2,  -1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.409445806858e-05,  1.918444678174e-06, {   0,   1,  -1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -5.223256013417e-06, -1.078816513234e-05, {   1,   0,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.408668875276e-05,  1.917820833729e-06, {   0,   1,  -1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -2.816603425941e-06,  1.317227912993e-05, {   2,  -1,  -1,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 { -2.816603425941e-06,  1.317227912993e-05, {   2,  -3,  -1,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.302808374576e-06, -1.468340499802e-05, {   2,  -3,  -2,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.365456589915e-06,  1.261822360455e-05, {   1,  -2,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  3.063262947104e-06, -1.291092708689e-05, {   1,  -2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.439261313077e-05,  1.579842758284e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  4.665536772125e-07, -1.550452425031e-05, {   2,  -1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.044620885539e-05, -5.517317382051e-06, {   2,  -1,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  1.526500434583e-05, -6.983457908297e-07, {   0,   1,   1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.262040356830e-05, -3.337610232770e-06, {   2,   1,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.576085594525e-05,  1.964477633713e-07, {   0,   1,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -3.513813996811e-06, -1.243178161274e-05, {   0,   1,   2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.106791615962e-05,  4.871598182780e-06, {   4,   1,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  4.382280080429e-06, -1.155074521215e-05, {   0,   2,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.068203247575e-06,  1.286441410596e-05, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.126482199271e-07, -1.541298623709e-05, {   0,   1,   2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.577271053026e-05,  1.486312293182e-07, {   4,  -1,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  5.861486983803e-06, -1.005985346035e-05, {   0,   1,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -3.010842939959e-06,  1.290229633663e-05, {   4,   1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.958850908636e-06, -1.295078148975e-05, {   2,  -1,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.265159677182e-05, -3.257401249313e-06, {   1,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.065932354821e-05,  5.245303807972e-06, {   2,   1,  -1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.024160712295e-05, -5.642726632085e-06, {   2,  -1,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -3.152959522973e-06, -1.270897290012e-05, {   3,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.584910434783e-05, -1.169407103898e-08, {   2,  -3,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.580220957436e-05,  4.600018116721e-08, {   4,  -1,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.446441027401e-05, -1.363877062969e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 { -1.372745047259e-05, -2.100345947799e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 { -1.562696327049e-05,  1.996049654711e-07, {   0,   1,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -6.367579654712e-06, -9.446121361660e-06, {   0,   1,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  1.400909931057e-05, -1.803439448929e-06, {   1,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.058973028365e-05,  5.221482564208e-06, {   2,  -1,  -3,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.070401535950e-05,  5.100061525540e-06, {   2,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.294619890143e-05,  2.855795929930e-06, {   3,  -1,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.576482183834e-05, -3.672014233513e-08, {   2,   1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.047708322479e-06,  1.275303976678e-05, {   4,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.317502111170e-05, -2.623619018587e-06, {   1,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.325268385371e-05, -2.535974154960e-06, {   3,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.563953749469e-05,  1.486274356413e-07, {   0,   1,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -9.177499352630e-07,  1.486922134623e-05, {   2,  -1,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -1.066454277466e-05, -5.107873685927e-06, {   1,   0,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  4.905429796950e-06, -1.086272694838e-05, {   1,   0,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -1.529968097364e-05,  4.670214395607e-07, {   4,   1,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.532293534528e-05, -4.387600471894e-07, {   0,   1,  -1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  1.238892121982e-06, -1.452275895647e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -1.380024788189e-05,  1.957161054395e-06, {   1,   0,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.085030691062e-06, -1.366940909745e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -2.083400969062e-06, -1.366864027531e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 {  8.298632889219e-06,  7.453241654370e-06, {   0,   1,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.275799911211e-05, -2.992294100886e-06, {   2,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.034217790353e-06,  1.269961948219e-05, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.201483470574e-06, -1.450828860288e-05, {   2,  -1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  7.094624952700e-06, -8.613452403080e-06, {   0,   1,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.068460051324e-05,  5.012197316998e-06, {   2,   1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.372036131429e-05,  1.974367485814e-06, {   0,   1,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  1.563745746896e-05,  5.310662236401e-08, {   2,   1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -9.564026383687e-06, -6.123631272374e-06, {   1,   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 {  4.055442954082e-06,  1.163106850386e-05, {   2,  -1,   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.128238333894e-07,  1.487362497527e-05, {   2,  -1,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.370880739359e-05,  1.973435096114e-06, {   0,   1,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 {  2.393687835946e-06,  1.327858127586e-05, {   2,   1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.496510257424e-05, -6.913409870215e-07, {   4,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.297983732530e-05, -2.673724702332e-06, {   1,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.191872830986e-06, -8.461203194602e-06, {   4,  -1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.443198909720e-05, -1.217374192315e-06, {   3,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.025535983402e-05,  5.374005138410e-06, {   0,   1,   1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -1.541927570831e-05, -2.098122924799e-07, {   4,   1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.068575894951e-07, -1.502110720375e-05, {   2,   1,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.518041749063e-05,  4.448038047569e-07, {   4,   1,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.115002671200e-05, -4.473472314619e-06, {   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.422174752184e-07, -1.497645814156e-05, {   1,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.555645417226e-05,  6.104728415101e-08, {   1,   0,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.556333047244e-05,  4.894762239836e-08, {   0,   1,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.989927190674e-06,  1.262194049405e-05, {   1,   0,   2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.093284528122e-05,  4.673919212107e-06, {   0,   1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.553771402583e-05,  6.575242574720e-08, {   0,   1,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.553994161505e-05, -4.191513172011e-08, {   4,   1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.373334904572e-05,  1.846170034405e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  3.564141643250e-06, -1.201338116754e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  5.126400978069e-06,  1.044705342441e-05, {   2,   1,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  2.969171395454e-06, -1.260054885414e-05, {   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.292613168428e-06,  7.275339224733e-06, {   1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.274052981249e-05,  2.821945079479e-06, {   1,   1,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -4.494347141179e-06, -1.106261808002e-05, {   0,   1,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.568560372123e-06, -1.098556310310e-05, {   1,  -2,  -2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -5.831249744584e-06, -9.721147601021e-06, {   3,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.338039317205e-05, -2.170709178762e-06, {   2,  -1,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -4.785010449148e-06,  1.076381027778e-05, {   1,   2,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.292439662727e-05, -2.620506054059e-06, {   1,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.450251944847e-05, -1.037408347658e-06, {   4,   1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -3.101639844910e-06, -1.243277897970e-05, {   1,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  5.487417219078e-06,  1.004292530381e-05, {   4,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.681186098279e-07, -1.495913675607e-05, {   2,  -1,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.551741546958e-05, -6.421654389810e-09, {   4,   1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.971821823658e-06, -5.546989150559e-06, {   4,   1,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -5.259598173850e-06,  1.025701003602e-05, {   1,   0,   0,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.979479102406e-06,  5.535997289350e-06, {   1,   0,   0,   0,   0,   0,  12, -20,   0,   0,   0,   0,   0 } },
 {  6.717056127309e-06,  8.787532145547e-06, {   2,   1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -4.300212339006e-06,  1.119703945130e-05, {   1,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.012269797193e-05,  5.362216149320e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  7.172376540662e-07, -1.475639474715e-05, {   4,   1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.375402144122e-05, -1.718818584671e-06, {   1,   0,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.964509685379e-06, -1.349808780468e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -2.951666295021e-06,  1.250878400764e-05, {   3,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.001777505823e-05,  5.438399840963e-06, {   0,   1,   1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  1.243794012308e-05, -3.016606029972e-06, {   2,  -3,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.858532566819e-06, -1.059279810990e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  1.960707055343e-06, -1.349037759420e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -6.153401839894e-06, -9.296376321418e-06, {   3,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.361105693854e-05,  1.833531332924e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 { -2.340044737949e-06, -1.309760934109e-05, {   0,   1,   2,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -6.765157973974e-06,  8.671453853606e-06, {   1,   0,   1,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 { -6.765157973974e-06,  8.671453853606e-06, {   1,  -2,   1,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 {  1.057851704543e-05, -4.849399721939e-06, {   2,   1,  -1,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.536832823728e-05,  5.022485909790e-08, {   2,  -1,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.439817110315e-06, -1.296537654927e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 {  9.518947929764e-06, -5.880090835025e-06, {   0,   1,   1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  6.979583633919e-06, -8.418654880118e-06, {   1,   0,   0,   0,   0,  11,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.451727981828e-05, -8.637018910551e-07, {   4,  -1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.423122321046e-06, -1.295520980034e-05, {   1,  -2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.495437765863e-05, -4.228174232453e-07, {   4,   1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.490847395576e-05, -4.686155327631e-07, {   4,  -1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.942198614750e-06,  1.142745461065e-05, {   2,  -1,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  8.199206156802e-07, -1.454066563126e-05, {   2,   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  7.517801045098e-06, -7.842421096262e-06, {   0,   1,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.531312762107e-05,  3.510857161738e-08, {   2,   1,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.730357861804e-06,  9.614479324604e-06, {   2,   1,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.121172691955e-05, -4.131635876405e-06, {   0,   1,   3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.146770431623e-07, -1.442388786516e-05, {   1,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -2.933997548235e-06, -1.239925892962e-05, {   1,   0,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.365982801723e-06, -9.964218220196e-06, {   2,   1,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -4.376443403213e-06,  1.095164455002e-05, {   1,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.681431371744e-07, -1.484767302385e-05, {   2,  -1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.116320969875e-06,  1.419681388831e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.576252059526e-06, -1.373484935849e-05, {   2,  -1,   0,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -5.271978027717e-06,  1.003203531646e-05, {   1,   0,   0,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 { -1.218989696057e-05,  3.112251651197e-06, {   1,   0,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.622642820879e-06, -1.367802309277e-05, {   2,   1,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.525885810619e-05,  4.044031574057e-08, {   2,   1,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  4.241912388451e-06, -1.105465451167e-05, {   3,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.743116860931e-06,  9.545307377354e-06, {   2,   1,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.323531610646e-05, -2.052563816459e-06, {   4,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.867786715204e-06,  5.402268232305e-06, {   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 { -6.767970871228e-06, -8.493376801350e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   0,   1,   0,   0,   0 } },
 { -6.562195639298e-06, -8.684746428411e-06, {   0,   1,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.916013830436e-06,  1.233078750785e-05, {   1,   2,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.549348296485e-06, -9.695053301972e-06, {   0,   1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -8.176288157174e-06, -7.058278434166e-06, {   1,   0,   1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  1.246286806607e-05,  2.762393686799e-06, {   5,  -1,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  8.384904871689e-06,  6.812356990852e-06, {   0,   1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  3.270035288573e-06,  1.192596601164e-05, {   1,   0,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.516138909194e-05,  2.556316647302e-08, {   0,   1,   3,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -9.889897404627e-06,  5.297032537221e-06, {   1,   0,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 { -3.538351547312e-06, -1.164049130484e-05, {   1,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.573865231722e-06, -1.059965694425e-05, {   1,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.086081414250e-05,  4.311999718359e-06, {   2,  -1,   0,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 { -1.359434714130e-05,  1.575027737500e-06, {   2,   1,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -2.049782235030e-06,  1.311423060306e-05, {   2,  -1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.083147281861e-05, -4.329091737549e-06, {   0,   1,   2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.010208749928e-05, -5.045362197451e-06, {   3,  -1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.941001468337e-06,  1.220476635015e-05, {   2,  -1,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -9.098959800265e-06,  6.041378266877e-06, {   2,   1,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.074754154236e-05, -4.385154292469e-06, {   2,  -1,   0,   0,   0,   0,  13, -20,   0,   0,   0,   0,   0 } },
 { -5.850031223461e-06, -9.272873374030e-06, {   4,   1,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.365267782707e-05, -1.468678912602e-06, {   2,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -1.217658148707e-05, -2.944728437963e-06, {   4,  -1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.784576254192e-07, -1.413723600912e-05, {   2,  -1,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -1.511109539175e-05, -2.923909116855e-09, {   4,   1,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.909979856431e-06,  1.219622414226e-05, {   1,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.147886812335e-05,  3.624965142549e-06, {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -6.774582170003e-06,  8.318703055522e-06, {   2,   1,  -1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.238766584303e-05,  2.698625310521e-06, {   4,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.012297847358e-05,  4.963056922466e-06, {   2,  -1,   0,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.487002988136e-05, -2.027074311454e-07, {   2,   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.886494753298e-06, -1.218529298019e-05, {   0,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.341628290931e-06,  7.727710069656e-06, {   2,  -1,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -1.505559956127e-05, -1.038745601635e-08, {   1,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.875929892343e-06, -1.218403706221e-05, {   2,  -1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.325678829836e-06,  7.732859302353e-06, {   2,   1,  -1,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 { -7.325678829836e-06,  7.732859302353e-06, {   2,  -1,  -1,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 {  1.488541580779e-05, -1.718170057938e-07, {   0,   1,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.183130527242e-06,  1.386764065895e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -1.432918101714e-05,  7.163328466233e-07, {   0,   1,   3,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.432918101714e-05,  7.163328466233e-07, {   0,   1,   1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -6.067893802952e-07, -1.442951795537e-05, {   2,   1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.275117746181e-05, -2.275612550050e-06, {   2,   1,  -1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.275117746181e-05, -2.275612550050e-06, {   2,  -1,  -1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.091119516012e-05, -4.111820467852e-06, {   3,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.501628123574e-05,  6.165778932698e-09, {   4,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.030726166203e-06,  5.979859270448e-06, {   2,   1,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -1.002844713301e-05, -4.981926793562e-06, {   1,   0,  -2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.227215945266e-05, -2.724128434157e-06, {   1,   0,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.058453610710e-05, -4.407606889415e-06, {   1,   2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.596066231964e-06, -8.381140425447e-06, {   1,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.370052966529e-05, -1.267615147436e-06, {   0,   1,   1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.440184880625e-05,  5.648997462525e-07, {   2,   1,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -7.588687548396e-06, -7.360232876130e-06, {   4,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.924498646375e-06,  1.000911981495e-05, {   3,   0,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -9.351863208996e-06, -5.580240665000e-06, {   3,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.477741576109e-05, -1.483236585255e-07, {   0,   1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.526008660400e-06, -1.236662569171e-05, {   2,   1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.257598439419e-06, -1.263199724522e-05, {   0,   1,   0,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -1.237525248689e-05,  2.507212952981e-06, {   1,   0,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.483767531339e-05,  3.167728472512e-08, {   0,   1,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.371322710447e-06,  9.496986105064e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 {  1.166992682550e-06,  1.369280644632e-05, {   2,   1,   0,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  1.166992682550e-06,  1.369280644632e-05, {   2,  -1,  -2,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -1.468234768711e-05, -1.762055415531e-07, {   0,   1,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  2.839468899061e-06, -1.201320457142e-05, {   1,   0,  -2,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.299509707193e-05, -1.857118339766e-06, {   2,  -1,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.022165275176e-06, -8.816025146826e-06, {   1,  -2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  6.892819203043e-07,  1.414476707828e-05, {   2,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.842216228464e-06,  1.197418859627e-05, {   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.841833457049e-06,  1.197028911557e-05, {   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.665498042713e-06,  1.014346103185e-05, {   0,   1,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.182443243200e-05, -2.981777621530e-06, {   2,  -1,  -2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.080896929174e-06,  1.171676744400e-05, {   3,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.002320987050e-06,  1.179464098627e-05, {   3,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.448155978276e-05, -3.122137962687e-07, {   2,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  4.453432007077e-06, -1.033405626063e-05, {   0,   1,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  8.038862328853e-06, -6.748436968436e-06, {   4,  -1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.412543227278e-05, -6.548671157382e-07, {   0,   1,  -1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.468122532302e-05, -9.386156247173e-08, {   2,   1,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.216516718236e-06,  6.557835592776e-06, {   2,   1,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  5.243816352397e-07, -1.423384079557e-05, {   2,  -1,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  8.693608765436e-06, -6.057263131222e-06, {   1,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  7.910030554310e-06,  6.827995439263e-06, {   0,   1,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.998507459232e-06, -1.073655819135e-05, {   0,   1,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.472444509459e-05,  7.129630066096e-09, {   3,  -1,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.365865491950e-06,  1.336501001236e-05, {   2,  -1,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.470988241924e-05, -1.734444206561e-08, {   0,   1,   1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.134177927345e-05,  3.384517049058e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.471333370899e-05, -7.143941170353e-09, {   2,  -3,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.005303551889e-05, -4.664292305642e-06, {   3,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -2.795692604761e-06, -1.191929120789e-05, {   1,   0,   1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.716971072215e-06, -1.099137557833e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -5.517126184543e-06,  9.187321897939e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  8.553397412859e-06,  6.149829460779e-06, {   2,  -1,   0,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 { -1.056134361744e-05, -4.141136536298e-06, {   0,   1,   3,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.079380314997e-06, -7.620336181786e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 {  4.723242802871e-06,  9.950276921783e-06, {   2,  -3,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.465994938556e-05,  2.635253518870e-09, {   2,   1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.874947115366e-06, -7.783227680595e-06, {   1,   0,   0,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  6.134751605015e-07, -1.402354783820e-05, {   2,   1,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.230532101837e-05, -2.330893828565e-06, {   2,  -1,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  4.817838275517e-06,  9.817801124009e-06, {   0,   1,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -2.809672402373e-06,  1.182412078393e-05, {   1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.405524067629e-05,  5.676769492015e-07, {   2,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.843002084891e-07, -1.413320295372e-05, {   0,   1,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.466583865870e-06, -1.214442254519e-05, {   2,   1,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -2.724972319749e-07, -1.432885442720e-05, {   0,   1,   2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.458643778678e-05, -1.189476285982e-08, {   2,  -1,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.204643400269e-06,  1.138782031733e-05, {   1,   0,  -1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.684217556251e-06,  9.906903123224e-06, {   3,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.219635639687e-05, -2.390726689650e-06, {   1,   0,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -9.832334075234e-06,  4.754588287419e-06, {   2,  -1,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.114947468738e-05,  3.433407415701e-06, {   1,   0,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.769102655385e-07, -1.399888185716e-05, {   2,  -1,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.996852119859e-06, -1.157513140027e-05, {   1,   0,   1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -6.588646790768e-06,  7.975368263173e-06, {   2,  -2,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.395865518840e-05,  6.053161750761e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  8.915542910909e-06,  5.639952758438e-06, {   4,  -1,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  6.624242519725e-06,  7.911749042563e-06, {   1,   1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -9.833746071406e-06,  4.699354050927e-06, {   2,   1,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.592335908645e-06, -8.937826007257e-06, {   4,  -1,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.414021536887e-05, -3.808562851053e-07, {   1,   1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.779826100983e-06,  1.173526012610e-05, {   1,  -2,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.259122043078e-07, -1.377337041564e-05, {   2,   1,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  6.224389392923e-07, -1.385595436846e-05, {   2,  -1,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.267819233498e-06, -1.220756536808e-05, {   1,   2,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.275069326657e-06,  6.197531365412e-06, {   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.382054636706e-06, -9.085339400424e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 { -8.861605782380e-06,  5.604043157842e-06, {   4,  -1,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.295104330862e-05,  1.512324949694e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -3.211501180583e-06, -1.124764032334e-05, {   4,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.445069253771e-05,  6.779282384800e-09, {   4,  -3,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.117151249403e-05,  3.282644687291e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.769208472272e-06, -1.168423486687e-05, {   3,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.442551392193e-05,  2.708763257890e-08, {   2,  -1,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.475138908723e-06, -8.967293234868e-06, {   4,  -1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -5.556892859895e-06, -8.879491611907e-06, {   4,  -1,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.582497289365e-07, -1.387780982954e-05, {   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.916945611876e-06,  1.151709988014e-05, {   1,  -2,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.124265004018e-06,  1.230762868194e-05, {   3,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.124265004018e-06,  1.230762868194e-05, {   1,   1,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.618849587174e-06, -5.807047132805e-06, {   4,   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.776654816445e-06, -1.264787847654e-05, {   2,   1,  -1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.776654816445e-06, -1.264787847654e-05, {   2,  -1,  -1,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.142112286873e-05,  3.001132413364e-06, {   4,  -1,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -1.142112286873e-05,  3.001132413364e-06, {   0,   1,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -2.809321591348e-06,  1.160996402722e-05, {   4,  -1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.073501073677e-06, -1.233930270575e-05, {   0,   1,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.679449885196e-06, -8.730451964198e-06, {   4,  -1,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.604964841303e-06,  4.800443556526e-06, {   0,   1,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.440234595275e-05,  2.533215448669e-10, {   1,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.929167079869e-06, -4.466328533868e-06, {   1,  -2,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  3.111439029684e-06, -1.128273921191e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.746140590418e-06,  1.164004307306e-05, {   1,   0,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.681829533763e-06, -9.702736633190e-06, {   1,   0,   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 {  1.436085553224e-05,  1.407810008507e-08, {   1,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.201061306337e-05,  2.360772135715e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.054254930779e-05,  3.828325690286e-06, {   0,   1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.141473834144e-05, -2.945173226041e-06, {   1,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.164998259970e-07,  1.394209086316e-05, {   2,  -1,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -5.524369501318e-07,  1.380470114936e-05, {   2,   1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.283151128108e-05, -1.517511881948e-06, {   3,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.113976974139e-07,  1.413329873522e-05, {   3,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.046625446284e-05,  3.875182668194e-06, {   1,   0,   0,   0,   0,   0,  -1,   4,  -1,   0,   0,   0,   0 } },
 { -1.046625446284e-05,  3.875182668194e-06, {   1,  -2,   0,   0,   0,   0,  -1,   4,  -1,   0,   0,   0,   0 } },
 { -1.200930126224e-05, -2.330894476542e-06, {   2,  -1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.529470187304e-06,  5.804458286309e-06, {   0,   1,  -1,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 {  8.529470187304e-06,  5.804458286309e-06, {   0,   1,  -1,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 { -1.428952094133e-05, -3.774135422640e-08, {   4,  -1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.346918694612e-06,  7.980286910641e-06, {   1,   0,   0,   0,   0, -10,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.268515065424e-05, -1.641408457188e-06, {   0,   1,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.179121088099e-05, -2.532641089478e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 { -7.959749053278e-07,  1.352696254536e-05, {   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.824881130245e-06, -1.149477443072e-05, {   1,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.644947669677e-06, -8.670777347804e-06, {   4,  -1,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.177559835910e-05,  2.540060178673e-06, {   2,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -2.098662647307e-07, -1.409881058835e-05, {   2,  -1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.820663697480e-07,  1.332362667421e-05, {   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.090998442568e-06, -1.020814094270e-05, {   1,   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 {  1.181010188835e-05,  2.488054412543e-06, {   2,  -1,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.250056959038e-07, -1.397270329048e-05, {   0,   1,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.428432484302e-05, -1.230639911536e-08, {   6,  -1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.280425520046e-05,  1.487476882429e-06, {   2,  -1,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -1.403071352946e-05, -2.559702782549e-07, {   0,   1,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -7.669048814483e-07, -1.351311122449e-05, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  9.052817902805e-06, -5.223003844041e-06, {   2,  -1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.043679100627e-06,  1.322774721665e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -3.011924485801e-06, -1.125374649643e-05, {   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.452150237832e-06, -1.180965460291e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  1.178104768212e-05,  2.475416814551e-06, {   3,   1,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.178104768212e-05,  2.475416814551e-06, {   1,   1,  -1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.114585311008e-05,  3.109413826033e-06, {   2,   1,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  6.447616872660e-06,  7.798795210225e-06, {   0,   1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -9.499810649855e-06, -4.744730762219e-06, {   3,  -1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.423849998660e-05,  4.630811826962e-09, {   2,  -3,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.450908094935e-06,  8.762845391636e-06, {   0,   1,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -2.767879633165e-06,  1.143874192224e-05, {   4,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.287236239460e-05,  1.330882179864e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  9.808375448767e-06,  4.391474119075e-06, {   3,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -5.551277001409e-06,  8.644277599914e-06, {   2,  -1,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  2.133868282218e-06, -1.206158036922e-05, {   1,   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.283085016007e-05,  1.356966920801e-06, {   2,  -1,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  1.412539305357e-05, -5.250799142884e-08, {   2,   1,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.698054533425e-06, -4.462487616593e-06, {   2,  -1,  -1,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.007845319506e-06, -1.114847669926e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.163657009649e-05, -2.513355630040e-06, {   1,  -2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.410238387582e-05,  4.640776256326e-08, {   2,  -1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.780550147041e-06, -8.367746664951e-06, {   1,  -1,   0,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 { -5.780550147041e-06, -8.367746664951e-06, {   1,  -3,   0,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 {  2.278273065146e-07,  1.391370828851e-05, {   2,   1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.748924706264e-06,  1.138250708606e-05, {   2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.909381187289e-06,  1.020433869066e-05, {   4,  -1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.584646820569e-06, -1.152620449154e-05, {   2,   1,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.409869007153e-05,  5.235506721265e-10, {   1,   1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.409869007153e-05,  5.235506721265e-10, {   1,  -1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.256540817530e-07, -1.346783843586e-05, {   4,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.036721194175e-07,  1.398896402704e-05, {   2,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.057106630316e-06,  5.034723984886e-06, {   1,   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  9.385133750436e-06,  4.706021591281e-06, {   0,   1,   1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.436288629185e-06, -7.653377599585e-06, {   2,  -1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.436288629185e-06,  7.653377599585e-06, {   2,  -3,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.534546609284e-06,  1.255167771885e-05, {   4,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.392936480058e-05,  1.564787395228e-07, {   1,   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -1.393332411292e-05,  1.434917473427e-07, {   4,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.647652658892e-06,  1.141728420890e-05, {   2,   1,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  1.098859596812e-05,  3.070730423174e-06, {   0,   1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.218978263264e-05, -1.864145006217e-06, {   1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.316803398290e-06,  4.734144390279e-06, {   0,   1,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.666052662893e-06, -1.138306337766e-05, {   3,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.311099388053e-05, -9.332602057172e-07, {   2,  -1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.216459501635e-05,  1.869172584079e-06, {   1,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.686036883858e-06,  1.132793356464e-05, {   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.395819660520e-05,  5.507847859423e-08, {   2,  -1,  -1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.155285592362e-05,  2.459286261100e-06, {   1,   0,   0,   0,   0,  -3,   6,   0,  -1,   0,   0,   0,   0 } },
 {  1.155285592362e-05,  2.459286261100e-06, {   1,  -2,   0,   0,   0,  -3,   6,   0,  -1,   0,   0,   0,   0 } },
 { -9.339409502707e-06,  4.664569691135e-06, {   6,   1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.339409502707e-06,  4.664569691135e-06, {   6,   1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.398427785115e-05, -8.018078712345e-10, {   1,  -1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.382906915650e-05, -1.548245286908e-07, {   4,  -1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  3.850430458177e-06,  1.013156693547e-05, {   1,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.299817513569e-06, -8.657322356567e-06, {   2,  -1,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  3.581709393466e-07, -1.359734956485e-05, {   0,   1,   1,   0,   0, -18,   8,  16,  -4,  -5,   0,   0,   0 } },
 { -3.581709393466e-07, -1.359734956485e-05, {   0,   1,  -1,   0,   0,  18,  -8, -16,   4,   5,   0,   0,   0 } },
 { -9.871881636029e-06, -4.075990997761e-06, {   1,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.390344600175e-05,  2.866675914000e-08, {   1,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.942309097894e-06,  1.098283624472e-05, {   1,   0,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.834278378050e-06, -5.087157917274e-06, {   2,  -1,   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 {  2.665249543239e-06, -1.125375728957e-05, {   3,  -2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.166464262646e-05, -2.250725029784e-06, {   2,  -1,   1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  1.093941805003e-05,  2.970929809140e-06, {   2,   1,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  9.270647166342e-06, -4.629409728165e-06, {   0,   1,   0,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  9.333787970393e-06, -4.564301936663e-06, {   0,   1,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.071552878506e-05, -3.181364994196e-06, {   0,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.978181797726e-06,  1.091284310815e-05, {   1,   0,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.357293947863e-05,  3.148304692138e-07, {   3,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.491296533126e-06, -8.391824670385e-06, {   2,   1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.095733858664e-05, -2.925602051838e-06, {   2,  -2,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.174555438613e-05,  2.135442087838e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.654638450030e-06,  1.122543182616e-05, {   3,   0,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.869577205305e-07, -1.337951954290e-05, {   0,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.281610592524e-06,  1.158485950609e-05, {   4,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.373326925207e-05, -1.185432784553e-07, {   2,  -1,  -1,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -8.591008940270e-06,  5.260725742267e-06, {   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.979556483943e-06,  1.086714847375e-05, {   2,   1,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.653673543903e-06, -1.018730447299e-05, {   2,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.735343576977e-06, -5.103379913831e-06, {   2,  -1,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  2.175186935409e-06, -1.164443472202e-05, {   0,   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -1.381538035146e-05, -1.718491613364e-09, {   4,   1,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -4.467604686799e-06,  9.340015228627e-06, {   4,  -1,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -5.161220686187e-06, -8.644400412321e-06, {   1,  -2,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -1.057250161247e-05, -3.233003205266e-06, {   4,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.288380464560e-06, -8.506582690931e-06, {   1,   0,  -1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.319690294145e-05,  5.977240480906e-07, {   2,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  2.661204569309e-06, -1.113275114979e-05, {   3,  -2,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.088201812534e-06,  4.705039028322e-06, {   2,  -1,  -1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.274472598287e-06, -8.517903578853e-06, {   0,   1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  1.260494005989e-05,  1.187246245921e-06, {   0,   1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.163653050729e-05, -2.152843358658e-06, {   2,   1,  -1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.163653050729e-05, -2.152843358658e-06, {   2,  -1,  -1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -3.301273540246e-06,  1.048694545846e-05, {   4,  -1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.001392177688e-05,  3.764756740055e-06, {   1,   0,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  1.368510259407e-05, -8.726990345114e-08, {   3,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.365061112801e-05, -1.073005697381e-07, {   0,   1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -9.399731181023e-06,  4.355783048591e-06, {   1,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.312068824607e-06,  4.443043957230e-06, {   2,  -1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.167969937141e-06, -1.158305350382e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  2.909748306534e-06,  1.082919400079e-05, {   1,   0,   3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  3.674780891770e-06,  1.006397642039e-05, {   1,   0,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.364464303357e-05, -7.725491384691e-08, {   2,  -3,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.981387618691e-06,  1.070809764663e-05, {   0,   1,   2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.011904561914e-06,  1.265748848350e-05, {   2,   1,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.365760347792e-05,  1.808621251596e-09, {   0,   1,   2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -6.390977632933e-07, -1.301963418798e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  4.522472260709e-07, -1.320441157714e-05, {   0,   1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.627110491338e-06,  1.102903329594e-05, {   1,   2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.344810606921e-05, -2.057480522447e-07, {   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -9.576226182010e-06, -4.068927701364e-06, {   2,   1,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.576226182010e-06, -4.068927701364e-06, {   2,   1,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.013712302854e-06,  1.163091774588e-05, {   2,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.971056662247e-06,  1.067063930774e-05, {   0,   1,   2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.190041663857e-05, -1.734506372037e-06, {   3,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.355961684464e-07, -1.319341210484e-05, {   2,  -1,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  2.328068530405e-07, -1.338930361717e-05, {   0,   1,   2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  1.169849217924e-05,  1.920261467062e-06, {   2,   1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.869385140857e-06, -8.745349719549e-06, {   1,   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -9.115718581490e-06,  4.495027345997e-06, {   2,  -1,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.728110222274e-06, -4.880583156533e-06, {   2,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.351792151925e-05,  8.884945032386e-08, {   4,   1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.213365791793e-05, -1.470675579041e-06, {   2,   1,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.250070493843e-06, -1.135184360850e-05, {   2,   1,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.032665602178e-05, -3.253737831127e-06, {   4,  -1,  -2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.210087262134e-05, -1.477350307021e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  3.932621972154e-07,  1.317899771114e-05, {   1,   1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.932621972154e-07,  1.317899771114e-05, {   1,  -1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.355803921078e-05, -1.226745943486e-08, {   4,   1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.500187335195e-06, -1.206641999381e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -1.355032131567e-05,  9.942754860185e-09, {   2,  -1,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.469488954596e-06,  9.088193628244e-06, {   1,   0,   0,   0,   0,   0,  17, -30,   0,   0,   0,   0,   0 } },
 {  2.582456947351e-06, -1.096803332440e-05, {   0,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.591940144228e-06,  1.094138962472e-05, {   3,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.573506783321e-06,  1.095941439032e-05, {   1,   0,   1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.334661351362e-05, -1.852541073511e-07, {   1,  -1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.343533746268e-05, -9.402494914322e-08, {   2,  -1,  -1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  2.726189318990e-06, -1.079992532877e-05, {   0,   1,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  3.864167910686e-06, -9.658906725058e-06, {   0,   1,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  9.676210842408e-06,  3.842832816075e-06, {   2,  -1,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.586903840924e-06,  1.093143612712e-05, {   1,  -2,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.107441469881e-05,  2.438810001873e-06, {   0,   1,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.994501329043e-06, -4.518588085867e-06, {   4,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.332709556002e-05,  1.795925857726e-07, {   2,   1,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.094916378066e-05,  2.555911494049e-06, {   2,   1,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 { -7.157141824787e-07, -1.278741671448e-05, {   0,   1,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  1.117171594245e-05, -2.331090363489e-06, {   1,   0,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.094632499165e-05,  2.554262782723e-06, {   2,   1,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -1.343629448942e-05,  5.690189453158e-08, {   2,  -1,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.127911745513e-06,  4.363769372595e-06, {   1,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.123064347529e-05,  2.258232243462e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   3,  -7,   0,   0,   0 } },
 {  1.123064347529e-05,  2.258232243462e-06, {   1,  -2,   0,   0,   0,   0,   1,   0,   3,  -7,   0,   0,   0 } },
 {  1.119040312758e-05,  2.296655007767e-06, {   4,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.010813607224e-06, -4.474048008992e-06, {   1,   2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.949675943069e-06, -1.153317398288e-05, {   2,   1,  -1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.949675943069e-06, -1.153317398288e-05, {   2,  -1,  -1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.974181631986e-06,  9.490402323058e-06, {   3,   0,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.341667547791e-05,  4.351929981502e-08, {   0,   1,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  1.989606224833e-06, -1.146179069002e-05, {   4,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  2.568574183509e-06,  1.087580802373e-05, {   1,   2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.763988585701e-07, -1.286523809067e-05, {   2,   1,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  9.377827040827e-06, -4.059991076218e-06, {   1,   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0 } },
 { -4.280269738762e-06,  9.155050582587e-06, {   0,   1,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.339141001924e-05,  4.027853959727e-08, {   2,   1,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.542919753068e-08, -1.340537683006e-05, {   4,  -1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.789295209809e-06,  9.630047398943e-06, {   1,   0,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.613280353120e-06,  1.080488520223e-05, {   1,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.750845671719e-07,  1.284222415021e-05, {   2,  -1,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  1.339041775406e-05,  1.226578509926e-08, {   4,  -3,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.626387920929e-06, -1.077253618530e-05, {   2,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.586563828024e-06,  1.081184173014e-05, {   4,   1,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.541518119940e-06,  1.185544303773e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -2.331736994851e-06, -1.106122385477e-05, {   2,  -1,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -4.507371057124e-06, -8.876163332120e-06, {   2,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.235093600853e-05,  1.030114581351e-06, {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -1.092022767022e-05, -2.457198785815e-06, {   1,   0,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.322095158240e-06, -1.105294026074e-05, {   2,  -1,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.329490414238e-05, -7.154847244008e-08, {   0,   1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.525898351629e-06, -1.082093431854e-05, {   3,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.544685864907e-06,  1.077972756273e-05, {   1,   0,  -2,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.033989094015e-05, -2.953356886642e-06, {   2,  -3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.653858011127e-06, -9.627585270178e-06, {   0,   1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.080386287993e-05, -2.477266421526e-06, {   2,   1,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.325611362534e-05, -2.285357181188e-08, {   1,  -3,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  9.085884362821e-06,  4.191915691434e-06, {   0,   1,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.425267493846e-06,  1.182639587650e-05, {   1,   0,   0,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0 } },
 {  1.425267493846e-06,  1.182639587650e-05, {   1,  -2,   0,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0 } },
 {  1.317032832553e-05, -8.124077907201e-08, {   0,   1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.221436137217e-05,  1.026542242617e-06, {   0,   1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -2.526501096955e-06,  1.070751611739e-05, {   1,  -2,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.271045027550e-05,  5.061724339503e-07, {   2,   1,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  9.124405020516e-06,  4.092093965390e-06, {   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.124415733562e-06,  4.092070077740e-06, {   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.890032991366e-06,  1.031855446536e-05, {   1,   0,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -9.135733541454e-06,  4.066739807210e-06, {   2,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.135755544501e-06,  4.066690378075e-06, {   2,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.661137654472e-06,  1.054016356708e-05, {   3,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.037065010369e-05,  2.827062499330e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -1.318614442390e-05,  9.432509769991e-09, {   4,  -1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.850483417789e-07, -1.271039454955e-05, {   2,  -1,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.010829469659e-07,  1.309431508218e-05, {   4,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.052883835802e-06, -9.141888897460e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.465165718329e-07, -1.273959677398e-05, {   0,   1,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.102246682342e-05,  2.160224597392e-06, {   1,   0,   0,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 { -1.287991409508e-05, -2.987387362177e-07, {   2,  -1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -9.646142669003e-07,  1.221369218591e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.301597600185e-05, -1.599025393256e-07, {   2,   1,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -3.604819483261e-06, -9.565718941224e-06, {   3,   0,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.283025971383e-05, -3.222151367274e-07, {   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -1.296761602803e-05, -1.818327152318e-07, {   3,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.592859457730e-07, -1.298993933872e-05, {   4,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.100352096175e-06,  1.203377739275e-05, {   2,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.044495678059e-05, -2.688733757005e-06, {   1,   0,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.216612140291e-05, -9.558816275304e-07, {   0,   1,   1,   0,  -5,   4,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.216612140291e-05, -9.558816275304e-07, {   0,   1,  -1,   0,   5,  -4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.071466779983e-05,  2.404835421304e-06, {   1,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.008799863288e-05,  3.030684158851e-06, {   4,   1,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.606570442305e-06, -1.051185095065e-05, {   2,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.195783018797e-06, -1.190074682378e-05, {   0,   1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.045898246812e-05, -2.635517250334e-06, {   1,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.302515720430e-05,  6.824534370072e-08, {   4,  -1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.092437979174e-05,  2.164677386734e-06, {   1,  -2,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.584390891207e-06,  1.150374082505e-05, {   1,   0,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 {  1.298812036668e-05, -8.180124740413e-08, {   0,   1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.104480556190e-06,  1.096485452773e-05, {   1,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  9.605466909266e-06, -3.463605018786e-06, {   4,  -1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.241768517538e-07,  1.243095787807e-05, {   2,   1,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 { -3.660386046899e-06, -9.389797265970e-06, {   2,  -1,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.457520820562e-06,  1.158673672393e-05, {   2,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.302599528480e-05,  3.658919790917e-09, {   0,   1,   1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.872954448699e-06, -1.014580454171e-05, {   0,   1,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  8.626863307790e-06,  4.389644239055e-06, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  1.089489810210e-05,  2.106903014482e-06, {   1,   2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.433063874587e-06, -3.567077320205e-06, {   2,  -1,   0,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 {  1.119763015966e-05,  1.792912415081e-06, {   6,   1,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -9.666848532680e-06,  3.314140917981e-06, {   2,  -1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.526433324315e-06, -1.044103056569e-05, {   0,   1,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.660669876820e-06, -1.030330360350e-05, {   1,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.112918129971e-07, -1.243852603485e-05, {   4,  -1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.023590481605e-05, -2.707411281485e-06, {   2,   1,  -1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  1.987537121201e-06,  1.094758092652e-05, {   2,  -1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.080753069379e-05, -2.124696857538e-06, {   3,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.379637231078e-06, -1.155251170288e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.191105993465e-05,  9.995384362778e-07, {   0,   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.252662451336e-05,  3.772759829772e-07, {   2,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -9.361153798100e-06, -3.534924239860e-06, {   1,  -2,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  4.342309174589e-07, -1.246081161231e-05, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  4.983633872913e-07, -1.238977435299e-05, {   2,  -1,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  2.395637283067e-06, -1.049008294884e-05, {   1,   0,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.971233623930e-06,  9.911335157871e-06, {   3,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.079009497081e-06, -1.179621626363e-05, {   0,   1,   1,   0,   0, -20,  15,   8,   0,   0,   0,   0,   0 } },
 {  1.079009497081e-06, -1.179621626363e-05, {   0,   1,  -1,   0,   0,  20, -15,  -8,   0,   0,   0,   0,   0 } },
 {  2.967674996723e-06,  9.897617988693e-06, {   3,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.277720811354e-05, -8.434513386272e-08, {   2,  -1,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.055270853042e-06, -1.079021964278e-05, {   3,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.238377282559e-05,  4.576907901262e-07, {   2,   1,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -9.563045080243e-06,  3.260916464141e-06, {   2,  -1,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -2.526590102768e-06, -1.029191263384e-05, {   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.278227629656e-05, -2.977940204043e-08, {   0,   1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.279881516832e-05, -5.010572379679e-09, {   3,  -1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.607890431920e-06,  4.194628250381e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -2.067494754156e-06, -1.073274032938e-05, {   2,   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.702273928697e-06,  1.108559590605e-05, {   2,   1,  -2,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 { -1.702273928697e-06,  1.108559590605e-05, {   2,  -1,  -2,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 { -2.447279546366e-06,  1.033790501406e-05, {   3,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.169872676386e-05,  1.073272070593e-06, {   2,  -1,   0,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 { -1.267252792049e-05, -8.671779441981e-08, {   4,   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.162001811191e-05, -1.136481466934e-06, {   2,  -1,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 {  1.100344065367e-06, -1.164440258175e-05, {   0,   1,   2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.272110641305e-05,  3.221162548056e-09, {   2,   1,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.080714581239e-05,  1.915931383643e-06, {   3,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.646739970286e-06,  1.007534535458e-05, {   1,   0,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.150051401691e-05, -1.214587508502e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  3.442396911516e-07, -1.236946376925e-05, {   2,  -3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.269974573591e-05, -1.217305381252e-08, {   0,   1,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.229333254745e-05,  4.183934994799e-07, {   2,  -1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.410790619263e-06, -1.029356825818e-05, {   1,   0,   1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.265353647343e-05, -3.924684711658e-08, {   2,  -1,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.445674020497e-06, -1.024447425393e-05, {   3,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.471531177682e-06, -1.021393968045e-05, {   1,   1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.597406928178e-07, -1.211501821287e-05, {   2,  -1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.256648540114e-05, -1.079405801564e-07, {   2,  -1,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.164992349360e-05,  1.023741830791e-06, {   1,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.455438330220e-06, -1.121676309606e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  1.396489531608e-06, -1.126508623146e-05, {   2,  -1,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.250575854660e-05,  1.433729280490e-07, {   2,   1,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.216835567198e-05, -4.760331117998e-07, {   0,   1,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.063047333595e-06,  9.564870928117e-06, {   2,  -1,   0,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 {  1.179692625345e-05, -8.250889393629e-07, {   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.261974103382e-05, -4.790221227753e-10, {   2,  -1,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.478777958701e-06, -1.013736353987e-05, {   3,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  9.135901343837e-07, -1.170054045218e-05, {   2,   1,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -2.455387634923e-06, -1.014735769266e-05, {   0,   1,   4,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.258105575287e-05,  6.731926271760e-09, {   1,   1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.054244890861e-05, -2.033842758166e-06, {   1,   0,  -3,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.085173903198e-05,  1.721589408180e-06, {   6,   1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.256039875990e-05, -1.222074897494e-08, {   0,   1,   1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  9.412502975424e-07,  1.163104513866e-05, {   2,   1,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  1.038772524190e-05,  2.182653905031e-06, {   3,   1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -1.038772524190e-05,  2.182653905031e-06, {   1,   1,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 {  7.375410400304e-07, -1.182963940450e-05, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.114513351398e-05,  1.417934720551e-06, {   0,   1,   1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.252295110354e-05, -3.492391771598e-08, {   2,  -1,  -2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.078282141103e-05,  1.757540279252e-06, {   2,   1,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.975769708791e-07,  1.233876366181e-05, {   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.217215985353e-05, -3.624897840302e-07, {   2,   3,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.624830948988e-07, -1.206849575550e-05, {   2,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.219928392492e-05,  3.303970879954e-07, {   2,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.237658632169e-05, -1.410176481180e-07, {   2,   1,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.213804926010e-05, -3.671671493439e-07, {   2,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.183789098553e-05, -6.660153367562e-07, {   2,   1,   0,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -2.363694849915e-06, -1.012788300618e-05, {   1,   0,   1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.048680950267e-05,  1.994478087924e-06, {   1,  -2,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.349378139118e-06,  1.013024608598e-05, {   2,  -1,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  9.646664693590e-06,  2.831895366139e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  8.723757930815e-07,  1.159279221102e-05, {   4,  -1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.389093495157e-06, -1.007014086812e-05, {   2,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.605591636042e-07, -1.199411060317e-05, {   0,   1,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.240089991673e-05, -5.159326139802e-08, {   0,   1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.382569529008e-06, -1.006477782986e-05, {   1,   2,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.498639633880e-06,  3.938781017222e-06, {   2,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.237748352717e-05,  5.534247132834e-08, {   3,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.463697640504e-07, -1.158480702249e-05, {   0,   1,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.224919522741e-05, -1.813516488657e-07, {   2,  -1,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -1.221791170562e-05,  2.088460739746e-07, {   4,   1,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.186130782870e-05, -5.598967486381e-07, {   2,  -1,   0,   0,   0,   5,  -6,   0,   1,   0,   0,   0,   0 } },
 { -2.397238346996e-06,  1.002355592463e-05, {   1,   2,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.019247543098e-05,  2.225713492275e-06, {   1,  -1,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.779913546279e-06,  9.625916918992e-06, {   1,   0,  -1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.377882894121e-06, -1.002455329052e-05, {   2,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.239589249396e-05, -5.095927930150e-09, {   1,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.741564569672e-06, -9.648051183847e-06, {   2,  -1,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.366462628077e-06, -1.002221490334e-05, {   3,   0,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.901168763396e-06, -2.485278489768e-06, {   0,   1,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.184121012212e-05, -5.369206299407e-07, {   0,   1,   0,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 { -1.145522826582e-05,  9.062504255498e-07, {   1,   0,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  4.794946162876e-07,  1.187942742454e-05, {   1,   0,   0,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 { -1.017805281860e-05, -2.172852788111e-06, {   1,   0,   0,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.540493780714e-07, -1.169207185959e-05, {   0,   1,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -1.031899699853e-05,  2.021125757099e-06, {   1,   0,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.624748946767e-06,  2.713707374001e-06, {   0,   1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.153886304738e-05, -7.965577744310e-07, {   0,   1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.093798051345e-05, -1.384101956271e-06, {   3,   2,  -3,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.093798051345e-05, -1.384101956271e-06, {   3,   0,  -3,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.010584742081e-05, -2.161230934276e-06, {   4,  -1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -2.351919518596e-06,  9.914469056256e-06, {   2,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.042800831816e-05,  1.829026145845e-06, {   2,  -1,   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 {  1.222951662919e-05, -9.890883613294e-09, {   0,   1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.525180589290e-07, -1.207532920394e-05, {   2,  -1,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.756777337236e-07, -1.173181741779e-05, {   2,  -1,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.949054514137e-06, -2.252547781506e-06, {   1,   0,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.893905466043e-06, -1.030333565988e-05, {   1,  -2,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  7.411255179181e-07,  1.145594421646e-05, {   2,   1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.205026872289e-05, -1.439728889437e-07, {   0,   1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.209016268149e-05,  1.038566738434e-07, {   1,  -1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.336435823927e-06, -9.853795469387e-06, {   3,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.182616354185e-07, -1.145511234520e-05, {   4,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.216073219338e-05, -6.827441022705e-09, {   0,   1,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.345218288226e-06, -1.078441693364e-05, {   0,   1,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.211384756296e-05,  7.981146621316e-09, {   1,  -3,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.191011782222e-05, -2.110840805501e-07, {   0,   3,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.194686929965e-05,  1.713750939099e-07, {   6,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.206386169836e-05,  4.947745929383e-08, {   2,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.012319028697e-05,  1.969427160098e-06, {   1,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.161949331595e-05, -4.724625086501e-07, {   0,   1,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.305585084334e-06, -9.753831123942e-06, {   3,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.656245107524e-07, -1.158618411635e-05, {   2,  -1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.600732590239e-07,  1.168970212528e-05, {   2,   1,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.180992594019e-05,  2.366203294343e-07, {   4,   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -7.719454774754e-07, -1.127410023933e-05, {   0,   1,  -1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -1.203919227310e-05, -6.406181769283e-10, {   0,   1,   2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.345523458056e-07, -1.159787032353e-05, {   6,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.976994322879e-08, -1.194909457965e-05, {   4,  -1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.802457850242e-06, -2.218461621889e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  1.172963446801e-05,  2.670346758765e-07, {   4,   1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.411009519147e-07,  1.185233012524e-05, {   2,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.040823382373e-05,  1.552563116828e-06, {   2,  -1,  -1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 {  1.145992040995e-05,  4.986102024979e-07, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -1.144800118791e-05,  5.065383179626e-07, {   6,  -1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.000192335870e-05,  1.932805051260e-06, {   4,  -1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.919923302007e-06, -1.001125047518e-05, {   1,   0,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -3.903360013911e-07, -1.151437023434e-05, {   0,   1,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.020968483144e-05,  1.694266504917e-06, {   2,  -1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.011561322968e-05,  1.769144042413e-06, {   3,   0,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -7.914392016020e-06,  3.956985405714e-06, {   2,   1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.134296418500e-05, -5.193073971133e-07, {   0,   1,   1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -1.175738910785e-05, -9.836319215203e-08, {   2,  -1,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.008738025703e-05, -1.763082150823e-06, {   4,   1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.070812077275e-05,  1.138428282740e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -1.136007710775e-05, -4.863203292201e-07, {   0,   1,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.046298364692e-05, -1.359126994885e-06, {   3,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.175184840879e-05,  6.588136362258e-08, {   2,  -1,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.084548764353e-05,  9.690770301859e-07, {   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  1.180034857592e-05, -5.271527525766e-09, {   2,  -1,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  9.863851041484e-06,  1.918732134936e-06, {   3,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.729806089170e-07, -1.129986858580e-05, {   2,  -3,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.079423390961e-05,  9.692973313722e-07, {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -1.164362655140e-05,  1.129318207753e-07, {   2,  -1,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  1.016904174786e-05,  1.581551243419e-06, {   0,   1,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.014731351992e-05,  1.596940575466e-06, {   1,   0,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -1.052564134880e-05, -1.215325568474e-06, {   1,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.107197106412e-05, -6.581775051061e-07, {   0,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  7.157803801021e-07,  1.100379067839e-05, {   0,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.525225506000e-07, -1.134754523043e-05, {   2,  -1,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.062596747958e-05,  1.072900930323e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  1.113051653049e-05,  5.672061284619e-07, {   3,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.167923744073e-05, -4.711689426897e-09, {   6,   1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.245400643965e-07,  1.073407505140e-05, {   2,  -1,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.031255704378e-05, -1.336838578856e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 } },
 { -1.031255704378e-05, -1.336838578856e-06, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0 } },
 {  1.155749180634e-05,  9.004028768602e-08, {   4,  -1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.153056271408e-05, -1.077849105492e-07, {   2,   1,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -1.043690324302e-05,  1.179910994818e-06, {   1,   0,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.137215469121e-05, -2.410331267153e-07, {   4,  -1,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -8.625723595840e-06,  2.986411137830e-06, {   1,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -9.869109870973e-06, -1.742040736182e-06, {   1,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.265989882161e-06,  1.033405072809e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.051320582585e-05,  1.080296583289e-06, {   2,   1,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  1.015728924661e-05,  1.397740921785e-06, {   1,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.154135185623e-05,  1.177510258374e-08, {   0,   3,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.581428920275e-07, -1.108897347502e-05, {   0,   3,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.987644996901e-06, -1.556002864458e-06, {   2,   1,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.478994350081e-06, -9.060125751440e-06, {   4,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.383981996077e-07,  1.106397012942e-05, {   2,   1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.177491785840e-06, -9.320892931720e-06, {   1,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.705375991777e-07,  1.121093550160e-05, {   0,   1,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.595662374983e-06,  9.871872243149e-06, {   0,   3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.145076147231e-05, -9.402719876319e-09, {   4,  -1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -3.700850209560e-07, -1.108569333815e-05, {   2,   1,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.128630275227e-05, -1.515259014150e-07, {   3,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.456072834267e-07, -1.096120786072e-05, {   2,   1,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.108692883702e-07, -1.099098433838e-05, {   0,   1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.064787109160e-05, -7.538396051149e-07, {   1,   0,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.509319797492e-06,  9.885441505006e-06, {   0,   1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.481767274164e-06, -9.889608978378e-06, {   2,  -3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.004843004139e-05, -1.314459917110e-06, {   1,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.135717235925e-05,  2.744350553644e-09, {   2,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 { -1.085476787488e-05,  4.931284078825e-07, {   2,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -1.131160386424e-05,  1.015219872350e-08, {   0,   3,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.130509801825e-05, -6.280483944493e-09, {   4,   1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.023824747041e-05, -1.067003113540e-06, {   3,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.396625509114e-06,  9.901991576814e-06, {   0,   1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.003399001400e-05, -1.249676828918e-06, {   2,  -1,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.009272691096e-05, -1.188058272867e-06, {   2,  -1,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.126614120785e-05, -6.990480000028e-09, {   0,   1,   1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.113289934553e-05, -8.975700179673e-08, {   0,   1,   0,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 {  1.113289934553e-05, -8.975700179673e-08, {   0,   1,   0,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 {  2.051658376806e-07,  1.101027216152e-05, {   2,  -1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.917219327613e-06, -1.284040812445e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.917219327613e-06, -1.284040812445e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.169462865345e-06, -1.001234105887e-05, {   2,   1,  -1,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 { -1.169462865345e-06, -1.001234105887e-05, {   2,  -1,  -1,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 {  1.111234590655e-05, -6.445773788382e-08, {   2,   1,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.793284126552e-07, -1.099179383211e-05, {   2,   1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.030500155613e-05, -8.506580184317e-07, {   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  1.112344668344e-05,  1.113870560005e-08, {   0,   1,   1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.084027081847e-05,  2.870519546883e-07, {   2,  -3,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.018187168127e-05, -9.404145934883e-07, {   2,  -1,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.112076507171e-05,  5.613215454747e-10, {   1,  -1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.108856472925e-05, -1.563090547218e-08, {   1,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.077407855541e-05,  3.049892157032e-07, {   1,   1,   1,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 {  1.077407855541e-05,  3.049892157032e-07, {   1,  -1,   1,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 { -1.107305800303e-05, -5.682982043556e-09, {   1,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.095999181786e-05, -8.239001327977e-08, {   4,  -3,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.550666409130e-07, -1.018054550174e-05, {   2,  -1,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.099179896752e-05,  3.770930743194e-08, {   2,  -1,  -2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.027926616758e-05, -7.470093022537e-07, {   2,   1,  -1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  1.030646650671e-05, -7.077521854392e-07, {   3,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.085776342025e-05,  1.462811656213e-07, {   2,  -1,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.084877970074e-05,  1.359432959576e-07, {   4,   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.097704328762e-05, -6.778333115458e-09, {   2,  -3,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.097583978068e-05,  4.826366571737e-09, {   2,   1,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.969030440451e-08, -1.087827229361e-05, {   4,   1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.087111501354e-05, -9.907524687367e-08, {   2,   1,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.088650346431e-05, -7.562800215262e-08, {   2,  -1,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  9.153206013750e-06, -1.793950553880e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  1.090107298665e-05, -7.961046015079e-09, {   0,   1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.586748595673e-07,  1.044591415683e-05, {   2,   1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.875832224900e-06, -2.003046104012e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -1.082435913063e-05, -4.728805119239e-08, {   0,   1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.178459116970e-06, -3.677134841930e-06, {   2,   1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.952723745400e-07,  9.959843742519e-06, {   1,   2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.951176638930e-07, -9.959857647968e-06, {   1,  -2,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.950011961847e-07, -9.959868114530e-06, {   1,  -4,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.949872895845e-07,  9.959869364178e-06, {   1,  -2,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.949507847507e-07, -9.959872644410e-06, {   3,  -2,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.949334014923e-07, -9.959874206379e-06, {   3,  -2,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.948708217393e-07, -9.959879829215e-06, {   1,  -2,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.946570073169e-07, -9.959899037604e-06, {   1,   2,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.945613991084e-07, -9.959907625236e-06, {   1,  -2,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07, -9.959908249755e-06, {   5,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07, -9.959908249755e-06, {   3,   0,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.945544457810e-07, -9.959908249755e-06, {   3,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07, -9.959908249755e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07,  9.959908249755e-06, {   1,   0,   3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.945544457810e-07,  9.959908249755e-06, {   1,   0,  -4,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.945492307851e-07,  9.959908718141e-06, {   3,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.943910424600e-07,  9.959922924560e-06, {   1,  -2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.056010396423e-05,  2.892716175322e-07, {   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.686472320704e-07, -1.027237851361e-05, {   0,   1,   1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 {  4.654654775401e-07, -1.036012460236e-05, {   3,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.055258566112e-05,  2.700817043332e-07, {   0,   1,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.029401921366e-05,  5.258128031600e-07, {   1,   0,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.071461140868e-05,  1.018323918649e-07, {   1,  -1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.081034805334e-05,  5.597960449010e-09, {   1,  -1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.024857912018e-05,  5.636472628603e-07, {   2,  -3,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.078717460462e-05,  1.004943181300e-08, {   0,   1,  -2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.078534369156e-05, -4.959981385205e-09, {   2,   1,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.015675666370e-05,  5.702683885920e-07, {   2,   1,   0,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -1.072059858475e-05,  1.221094256770e-09, {   2,  -1,  -3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.070697941693e-05,  8.705320441180e-09, {   1,  -1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.068773198995e-05,  1.990028984960e-08, {   0,   1,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.028144496379e-05, -4.220035727941e-07, {   0,   1,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.068675213855e-05, -7.695708771535e-09, {   0,   1,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.727481276334e-07,  1.049784205393e-05, {   3,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.034703699165e-05,  3.186241372124e-07, {   2,  -1,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.063517502256e-05, -6.994695107185e-09, {   6,   1,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.043852664191e-05,  1.976167861945e-07, {   2,   1,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.062364947726e-05, -2.894578229660e-09, {   0,   1,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.061013847263e-05,  1.280009683200e-08, {   0,   1,   1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.128846550682e-07, -1.029908970633e-05, {   2,   1,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.531307469720e-07,  1.014325890738e-05, {   2,   1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.986107622289e-07,  9.996006890955e-06, {   2,  -1,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 {  1.008223989159e-05, -5.064877882223e-07, {   3,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.054176272765e-05,  4.209351006738e-08, {   0,   1,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.811412268973e-07,  1.018714294319e-05, {   2,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.049965987638e-05,  3.320923699093e-08, {   4,   1,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.047342325894e-05, -5.189256654206e-08, {   4,  -1,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.014831857292e-05, -3.698618127389e-07, {   0,   1,   2,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -2.012976859897e-07, -1.028085593339e-05, {   1,  -1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.044912117325e-05, -2.260961264186e-08, {   1,   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.037689074826e-05,  8.941044900842e-08, {   2,   1,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.033929243596e-05,  1.269770236084e-07, {   1,  -1,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.803558894237e-07, -1.005857452721e-05, {   2,  -1,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.042727259573e-05,  8.105693417111e-09, {   2,  -1,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.028475601658e-05,  1.172139474534e-07, {   2,  -1,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -2.157180177803e-07,  1.018013311203e-05, {   0,   1,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.032726723740e-06, -8.339107999630e-06, {   3,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.033821963799e-05,  1.517939600228e-09, {   4,  -1,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.012824085604e-05,  2.099548316191e-07, {   2,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 { -1.028560304657e-05, -5.010978439850e-08, {   2,  -1,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.995147874856e-06,  3.114786666143e-07, {   2,   1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.995147874856e-06,  3.114786666143e-07, {   2,  -1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.029035591942e-05, -1.177448210612e-08, {   1,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.510769351813e-07, -1.013432414881e-05, {   4,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.587748205747e-07,  1.008146720062e-05, {   2,   1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.010499510079e-05, -5.288947997475e-08, {   0,   1,   1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -9.999316457254e-06, -1.169204331730e-07, {   2,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.999982076453e-06, -1.893332041388e-08, {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.999982659826e-06,  1.862265235477e-08, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.999997611556e-06, -6.911503287641e-09, {   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.999997611556e-06, -6.911503287641e-09, {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.999999997465e-06, -2.251474734903e-10, {   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999997770e-06,  2.111848394756e-10, {   2,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999997879e-06,  2.059488517208e-10, {   2,  -2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999998190e-06,  1.902408884559e-10, {   2,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998321e-06, -1.832595714478e-10, {   2,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999998353e-06, -1.815142421955e-10, {   0,   4,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999998654e-06,  1.640609496801e-10, {   4,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998766e-06, -1.570796326733e-10, {   2,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999998794e-06, -1.553343034209e-10, {   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999998951e-06,  1.448623279105e-10, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -9.999999999143e-06, -1.308996938977e-10, {   6,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999143e-06,  1.308996938958e-10, {   4,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999166e-06, -1.291543646453e-10, {   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999166e-06,  1.291543646440e-10, {   4,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999188e-06,  1.274090353921e-10, {   2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999188e-06, -1.274090353929e-10, {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999210e-06, -1.256637061405e-10, {   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999999210e-06, -1.256637061405e-10, {   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999210e-06, -1.256637061405e-10, {   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999232e-06,  1.239183768884e-10, {   2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999232e-06,  1.239183768884e-10, {   0,   2,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999232e-06,  1.239183768884e-10, {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999999337e-06, -1.151917306309e-10, {   6,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999337e-06, -1.151917306309e-10, {   4,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999356e-06, -1.134464013785e-10, {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999999415e-06,  1.082104136215e-10, {   2,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999433e-06, -1.064650843687e-10, {   2,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999452e-06, -1.047197551163e-10, {   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.999999999522e-06, -9.773843811139e-11, {   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999588e-06, -9.075712110150e-11, {   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  9.999999999731e-06, -7.330382858157e-11, {   4,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.999999999863e-06, -5.235987756133e-11, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  9.999999999872e-06,  5.061454830762e-11, {   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -9.999999999919e-06,  4.014257279576e-11, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.000000000000e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00, -1.000000000000e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.751837013549e-07, -9.690859319410e-06, {   1,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -9.618797869020e-06, -2.794854627724e-07, {   2,  -1,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.702487958870e-06, -1.379062817740e-07, {   2,  -1,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.801837084760e-06,  2.894385782890e-06, {   2,  -1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -6.310448522890e-06,  3.092373206389e-06, {   2,  -1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.953456419720e-06,  6.562923165080e-08, {   1,   1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.507053137450e-06, -3.094170674936e-07, {   2,   1,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.032370788900e-06, -2.460578903091e-06, {   2,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.402758884270e-06,  6.548377104710e-06, {   1,  -2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.638793329102e-06, -4.128087082077e-06, {   1,   0,   0,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 {  6.816706403680e-06,  8.941555500490e-07, {   1,   2,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.981089116952e-06, -5.513508752090e-06, {   4,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.372767694838e-06, -5.750690901750e-06, {   3,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.199124613610e-06,  1.826030274800e-06, {   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -5.986117685100e-06, -8.696389428087e-07, {   2,  -1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.992716150193e-06, -3.658421970538e-06, {   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.429035255900e-07,  5.796818163803e-06, {   4,  -1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.198214890430e-08, -6.313334382160e-06, {   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.442003262920e-06,  4.810838816286e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.444319328340e-06,  4.787142409125e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  2.286275146420e-06,  3.641159830020e-06, {   1,   0,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  4.944058145130e-06, -9.499606704003e-07, {   2,  -1,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.135137943619e-06,  4.706569088060e-06, {   1,  -2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.794729372690e-07, -5.146409513972e-06, {   0,   1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.422775428740e-06,  2.814587664807e-07, {   2,   1,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.192610627590e-06, -4.735584259120e-07, {   3,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.020733829461e-06, -4.634431499420e-06, {   1,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.805169964210e-07,  4.584810826080e-06, {   1,  -2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  5.146277896620e-07,  4.840587609711e-06, {   4,   1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.046712901660e-07,  4.745945847083e-06, {   0,   1,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.820554280400e-06,  4.195253047280e-07, {   2,  -1,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.105325351670e-06,  3.110061604690e-06, {   1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.112024605900e-06, -9.482109968600e-07, {   1,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.950130052290e-06, -3.064547525480e-06, {   1,   0,   0,   0,   0,   0,  -3,   8,   0,   0,   0,   0,   0 } },
 { -4.868674787880e-06,  7.597617565107e-08, {   0,   1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.267785749900e-06,  2.932267271802e-07, {   2,  -1,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.982361757770e-06, -1.559863376242e-06, {   1,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.929361629513e-06,  2.586622697610e-06, {   2,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.198269837449e-06, -2.297767589870e-06, {   2,  -1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.060409120900e-07, -3.504720120290e-06, {   2,  -1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -3.561880538060e-06, -2.243143382300e-07, {   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  2.541005722127e-06, -1.156143203330e-06, {   0,   1,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -2.523539261454e-06, -9.550115210700e-07, {   0,   1,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  3.406404865400e-06, -4.225410298507e-09, {   1,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.104257909420e-06, -1.247937355062e-07, {   0,   1,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -6.654323071400e-07, -2.530709356800e-06, {   1,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.130942878270e-06,  1.041167460190e-06, {   1,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -7.321486672100e-07, -2.431121366300e-06, {   1,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.455811546700e-08,  3.062746696930e-06, {   2,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.296025844840e-06,  8.301143431000e-07, {   0,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -2.790635099880e-06, -2.632573385610e-07, {   3,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.147300506740e-07, -2.106651149390e-06, {   1,   0,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.117113023070e-06, -1.128456635210e-06, {   0,   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.106231256720e-06,  8.922797349920e-08, {   0,   1,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.865243763740e-07,  1.922419357087e-06, {   1,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.733947121280e-08,  2.044407508780e-06, {   2,  -1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -8.957468520200e-07, -1.169274872870e-06, {   2,  -1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.004859946360e-07, -1.646798870520e-06, {   5,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.783619550633e-06,  2.501253898600e-07, {   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.066300186800e-07,  1.837735524932e-06, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -4.091268323000e-08,  1.843365744797e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -1.634779776650e-06,  8.835026688600e-08, {   1,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.249189320160e-07,  1.396795162570e-06, {   3,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.639185733620e-08, -1.631534087380e-06, {   2,  -1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.337309439025e-07, -1.518676255850e-06, {   0,   1,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.070145135560e-06, -2.673913314171e-07, {   2,  -1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.211663819700e-07, -7.621822411600e-07, {   1,   0,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  5.954958666000e-07, -3.004422530836e-07, {   2,   1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.793299000988e-10,  8.076334683574e-07, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.277651757110e-07, -4.808519754200e-07, {   1,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.172636649000e-08, -4.519518461072e-07, {   2,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.542464380200e-07, -8.319790866710e-08, {   2,   1,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.442917626000e-08, -2.277987878267e-07, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  2.855635752000e-08, -2.112155815580e-07, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  4.257923545000e-09,  8.538509629000e-08, {   1,  -2,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.118202600000e-08,  5.383257794154e-08, {   4,  -1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 516> g_PERT_S2_2
{ {
 { -7.430000000000e-02,  0.000000000000e+00, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.043000000000e-02,  0.000000000000e+00, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.229000000000e-02,  0.000000000000e+00, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.999000000000e-02,  0.000000000000e+00, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.869000000000e-02,  0.000000000000e+00, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.696000000000e-02,  0.000000000000e+00, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.623000000000e-02,  0.000000000000e+00, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.262562730846e-02,  3.251379396349e-03, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.549403410181e-03,  9.886576285223e-03, {   0,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.547600825700e-03,  9.882658690231e-03, {   0,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.419000000000e-02,  0.000000000000e+00, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.338000000000e-02,  0.000000000000e+00, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.279000000000e-02,  0.000000000000e+00, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.215000000000e-02,  0.000000000000e+00, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.460000000000e-03,  0.000000000000e+00, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.430000000000e-03,  0.000000000000e+00, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.420000000000e-03,  0.000000000000e+00, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.300000000000e-03,  0.000000000000e+00, {   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.180000000000e-03,  0.000000000000e+00, {   2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.956194959634e-03, -4.961474050251e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.850000000000e-03,  0.000000000000e+00, {   2,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.070000000000e-03,  0.000000000000e+00, {   2,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.020000000000e-03,  0.000000000000e+00, {   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.020000000000e-03,  0.000000000000e+00, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-03,  0.000000000000e+00, {   0,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.980000000000e-03,  0.000000000000e+00, {   2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.930000000000e-03,  0.000000000000e+00, {   2,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.640000000000e-03,  0.000000000000e+00, {   2,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.610000000000e-03,  0.000000000000e+00, {   0,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.043430356693e-04,  1.095577020555e-03, {   0,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.580000000000e-03,  0.000000000000e+00, {   2,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.886762018153e-04,  1.061527657944e-03, {   0,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.490000000000e-03,  0.000000000000e+00, {   4,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.350000000000e-03,  0.000000000000e+00, {   2,  -1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.040000000000e-03,  0.000000000000e+00, {   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.152457436873e-04,  7.151122910545e-04, {   0,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.152349608530e-04,  7.150763564504e-04, {   0,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.070895696508e-04, -1.803843915263e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.500000000000e-04,  0.000000000000e+00, {   4,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.684188653734e-04,  1.701583511140e-04, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.900000000000e-04,  0.000000000000e+00, {   0,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.600000000000e-04,  0.000000000000e+00, {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.674244154714e-04,  1.869940256210e-04, {   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.800000000000e-04,  0.000000000000e+00, {   2,  -1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-04,  0.000000000000e+00, {   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.700000000000e-04,  0.000000000000e+00, {   2,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.700000000000e-04,  0.000000000000e+00, {   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.400000000000e-04,  0.000000000000e+00, {   4,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.300000000000e-04,  0.000000000000e+00, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.300000000000e-04,  0.000000000000e+00, {   2,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.695893193456e-04, -1.470443247156e-04, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.579100467217e-04,  3.430277208488e-04, {   2,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.576579507418e-04,  3.424798353351e-04, {   2,  -1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.525364252367e-04, -1.461613047395e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.800000000000e-04,  0.000000000000e+00, {   0,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.300000000000e-04,  0.000000000000e+00, {   4,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-04,  0.000000000000e+00, {   4,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.700000000000e-04,  0.000000000000e+00, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.500000000000e-04,  0.000000000000e+00, {   2,  -1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.019074884780e-04,  2.213784059216e-04, {   2,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.200000000000e-04,  0.000000000000e+00, {   2,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.879950745808e-05,  2.146241352386e-04, {   2,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.100000000000e-04,  0.000000000000e+00, {   2,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.100000000000e-04,  0.000000000000e+00, {   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.551226086480e-04, -4.703032737323e-05, {   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.566242408314e-04,  3.413296890386e-05, {   0,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.566092825590e-04,  3.413152491547e-05, {   0,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.724009083443e-04, -1.811665287776e-05, {   2,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.723158178122e-04, -1.809389122974e-05, {   2,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.900000000000e-04,  0.000000000000e+00, {   2,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.900000000000e-04,  0.000000000000e+00, {   0,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.330032838791e-04, -4.795673197284e-05, {   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.800000000000e-04,  0.000000000000e+00, {   4,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.800000000000e-04,  0.000000000000e+00, {   2,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.800000000000e-04,  0.000000000000e+00, {   0,   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.802964038234e-05,  1.912273776060e-04, {   2,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.725673646970e-05,  1.895476091080e-04, {   2,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.421969608712e-04,  2.677754927466e-05, {   0,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.421796802122e-04,  2.677465261109e-05, {   0,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.544518130139e-05, -1.973821292567e-04, {   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.600000000000e-04,  0.000000000000e+00, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.600000000000e-04,  0.000000000000e+00, {   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.565848628452e-04, -7.048097911356e-05, {   0,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.565736837325e-04, -7.047682698168e-05, {   0,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.200000000000e-04,  0.000000000000e+00, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.200000000000e-04,  0.000000000000e+00, {   2,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-04,  0.000000000000e+00, {   2,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.900000000000e-04,  0.000000000000e+00, {   0,   1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.900000000000e-04,  0.000000000000e+00, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.501834397315e-04,  3.966258443883e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.800000000000e-04,  0.000000000000e+00, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.700000000000e-04,  0.000000000000e+00, {   4,  -1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.385094509506e-04,  2.408910907346e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.600000000000e-04,  0.000000000000e+00, {   2,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.600000000000e-04,  0.000000000000e+00, {   2,  -1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.600000000000e-04,  0.000000000000e+00, {   0,   1,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.256794388293e-04,  3.323670644879e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.448552726264e-05, -6.676959007307e-05, {   0,   1,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -8.448552726264e-05, -6.676959007307e-05, {   0,   1,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.500000000000e-04,  0.000000000000e+00, {   4,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.652159709384e-05,  1.010578981436e-04, {   0,   1,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.267683494172e-05,  9.454234774720e-05, {   2,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.211673710927e-04,  2.330985281341e-05, {   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.063625031200e-04, -3.517565445854e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -4.458942971792e-05,  9.685867768899e-05, {   0,   1,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.010059617567e-05,  8.991567051931e-05, {   2,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.400000000000e-04,  0.000000000000e+00, {   4,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.400000000000e-04,  0.000000000000e+00, {   2,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.400000000000e-04,  0.000000000000e+00, {   2,  -1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.400000000000e-04,  0.000000000000e+00, {   2,  -3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.400000000000e-04,  0.000000000000e+00, {   1,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.400000000000e-04,  0.000000000000e+00, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.151215278100e-05,  8.213952145212e-05, {   2,  -1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.143658045949e-05,  8.201922846038e-05, {   2,   1,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.006183835019e-04,  3.045897730977e-05, {   1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.300000000000e-04,  0.000000000000e+00, {   4,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.300000000000e-04,  0.000000000000e+00, {   4,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.300000000000e-04,  0.000000000000e+00, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.063106828460e-04, -2.189182007948e-05, {   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.850898596010e-05,  9.585007939239e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-04,  0.000000000000e+00, {   4,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.200000000000e-04,  0.000000000000e+00, {   2,   1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.100000000000e-04,  0.000000000000e+00, {   2,  -1,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.163809389440e-05,  1.726604153220e-05, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.099333088255e-05,  2.581909054840e-05, {   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.389069379041e-05,  7.936893976469e-05, {   0,   1,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.379637880721e-05,  7.905462714543e-05, {   0,   1,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-04,  0.000000000000e+00, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.734119312570e-05, -2.045335781464e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.933527510110e-05, -1.413478347104e-05, {   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.330138804578e-05,  4.917709942060e-05, {   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.569873873853e-05,  7.485377929255e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.570009142308e-05,  7.483652946878e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -9.000000000000e-05,  0.000000000000e+00, {   3,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.000000000000e-05,  0.000000000000e+00, {   1,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.000000000000e-05,  0.000000000000e+00, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.000000000000e-05,  0.000000000000e+00, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.385579824458e-05, -1.414494832036e-05, {   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.073577778631e-05,  6.611633129047e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.073789977922e-05,  6.609185410793e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  4.030217767987e-06,  8.230361776803e-05, {   2,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.810830602304e-05, -5.819803338317e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.810918950267e-05, -5.818192244205e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -4.057611999502e-06,  8.181716947600e-05, {   2,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.767354398499e-05,  1.789668808781e-05, {   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.929729492960e-05,  3.285837431926e-06, {   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.833779119178e-05, -1.407596433883e-05, {   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.094538648841e-05, -5.140411515709e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.072187927392e-05,  5.153800669291e-05, {   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.072187927392e-05,  5.153800669291e-05, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   4,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   2,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   2,   1,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.000000000000e-05,  0.000000000000e+00, {   0,   1,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  6.030319665576e-05,  1.794338942126e-05, {   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.505060722444e-05,  1.297104003520e-05, {   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.861388601870e-06, -7.105901585440e-05, {   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -5.506637824610e-05,  2.043611045459e-05, {   2,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.735658617323e-05, -1.761312074995e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.090213921261e-05, -3.377214250920e-05, {   2,   1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.084400283582e-05, -3.372409565712e-05, {   2,  -1,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.628864586077e-05,  3.501801790042e-05, {   2,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.177149399301e-05, -5.883393518347e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.176971754427e-05,  5.883429058745e-05, {   2,  -1,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.940902489662e-05, -1.103617973144e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.352358565069e-05, -6.905926830109e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  4.501126075648e-05, -2.525386332857e-05, {   2,  -1,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   4,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.000000000000e-05,  0.000000000000e+00, {   2,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.616147222849e-05, -4.355691275991e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.238136971203e-05,  1.723947961123e-05, {   2,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -5.229431514865e-05,  1.721083589794e-05, {   2,  -1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -5.750197875188e-05,  1.171972883581e-05, {   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.969105398519e-05,  3.932737434943e-05, {   2,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  5.618205408431e-05,  1.147093054508e-05, {   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.486736763168e-05,  2.206624847598e-05, {   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.901444976084e-05,  3.758522919755e-05, {   2,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.348655899585e-05, -9.002922651217e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   4,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   4,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   4,   1,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   3,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.000000000000e-05,  0.000000000000e+00, {   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   0,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.987067026409e-05, -3.754238036667e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.713551333973e-05, -9.708296713606e-06, {   1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.185356865581e-05,  1.470905947649e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.286913227311e-05,  2.279517807811e-05, {   2,  -1,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.923946828160e-05, -4.303834534834e-06, {   2,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.920731720960e-05, -4.297108732658e-06, {   2,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.685879774130e-05,  3.662207520348e-05, {   2,  -1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.835235137126e-05, -3.498924325052e-05, {   2,  -1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -3.185401690799e-05,  2.030321779042e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -1.801946567020e-05, -3.412176458959e-05, {   2,  -1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.632513308863e-05,  3.545968814357e-05, {   2,  -1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   3,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   3,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   1,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.782675920247e-05,  3.197349304542e-05, {   2,  -1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.187142961833e-05, -7.126998735162e-06, {   1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.867059656285e-05,  1.022667890732e-05, {   1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.612318418908e-05, -3.267666510952e-05, {   0,   1,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.612318418908e-05, -3.267666510952e-05, {   0,   1,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  3.872195566418e-05,  1.003046108318e-05, {   2,  -1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.615897519008e-06,  4.198547276334e-05, {   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.515037885453e-05,  3.291110310025e-05, {   2,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.162501894678e-05, -3.629897303982e-05, {   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.427834660317e-05, -1.334829663906e-05, {   2,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.414656079540e-05, -1.336986004444e-05, {   2,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.483415243840e-05,  3.222596144157e-05, {   2,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.870829922178e-05,  7.876820398685e-06, {   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.962840039952e-05,  1.600121957259e-05, {   2,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.420397339135e-05, -3.129855010790e-05, {   1,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  3.384803284739e-05,  1.115391958637e-05, {   2,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.636347943302e-06, -4.215444117703e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -3.664802067412e-05, -6.960583117424e-06, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.994050279140e-05, -2.180879815691e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.146169389061e-05, -9.872219152499e-06, {   2,  -1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.133879406072e-05, -9.893649319167e-06, {   2,  -1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.667837000952e-05,  1.368553281583e-05, {   2,  -1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -3.342835138662e-05, -6.880478236099e-06, {   1,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.699941404880e-05,  1.307790659935e-05, {   2,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   6,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   4,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   1,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   1,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.228593995733e-05, -2.707218434140e-05, {   1,   2,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.780138511533e-05,  1.127310896200e-05, {   1,  -2,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.673975291607e-05, -1.109284920759e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.900294742214e-05, -7.670009180489e-06, {   1,  -2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.742045010017e-05,  9.197932864035e-06, {   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.576180021385e-05,  9.655635321657e-08, {   2,  -1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.794409267034e-05,  1.714792071364e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.619779179953e-05, -8.381120889757e-06, {   2,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.027489939821e-05,  2.410491148174e-05, {   2,  -1,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.070445138349e-05, -2.327904618577e-05, {   0,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.533114629262e-05,  8.493271927511e-06, {   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.022568983989e-05,  3.519139035079e-06, {   0,   1,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.022035772835e-05,  3.517971983370e-06, {   0,   1,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.562005139352e-05,  7.695527168970e-06, {   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.161446531043e-05,  2.078058387637e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.745784165250e-06,  2.750494550911e-05, {   0,   1,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.734985620153e-06,  2.748107021365e-05, {   0,   1,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.445828821665e-06,  2.473733983029e-05, {   2,  -1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.445828821665e-06,  2.473733983029e-05, {   2,  -1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.636948646177e-05,  5.444144223015e-06, {   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.874027608034e-05,  2.588294044265e-06, {   1,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.077267816789e-05, -1.019905628240e-05, {   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.405778046450e-05,  1.627069628647e-05, {   0,   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.299705859526e-05,  7.186846837257e-06, {   0,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   6,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   3,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   3,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   3,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,   1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,   1,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   1,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,  -1,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   2,  -1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   2,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   1,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   0,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   0,   1,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.000000000000e-05,  0.000000000000e+00, {   0,   1,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.381839084581e-05,  1.599756132958e-05, {   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.262259173002e-05,  7.192482842004e-06, {   2,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.018542668569e-06,  2.473727104859e-05, {   0,   1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.010921780842e-06,  2.472451398541e-05, {   0,   1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.922309236754e-05,  1.036200135700e-05, {   2,   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.522164475597e-05, -4.059977538931e-06, {   1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.398908357863e-05, -4.930474005695e-06, {   1,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.084861938241e-06, -2.076333114728e-05, {   1,   1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -8.084861938241e-06, -2.076333114728e-05, {   1,  -1,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.153890867964e-05,  6.652924325822e-06, {   0,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.820562514097e-05,  9.811235982427e-06, {   2,  -1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  7.521728435597e-06, -2.036209325951e-05, {   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.332472688601e-05,  4.442146330321e-06, {   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.677880286186e-05,  1.088447401223e-05, {   2,  -3,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.551977465822e-06,  1.855382671243e-05, {   2,  -1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.503690343428e-06,  1.844900503285e-05, {   2,   1,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.033674290375e-05,  6.527173475037e-06, {   0,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.568454012447e-05,  1.137473721541e-06, {   2,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.159120498823e-06, -1.826003156311e-05, {   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.159120498823e-06, -1.826003156311e-05, {   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.977646866712e-05,  6.285329935942e-06, {   0,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.163835840536e-06,  1.773373289506e-05, {   0,   1,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.135833746231e-05,  4.528223841903e-06, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.870665627567e-05, -7.075380624668e-06, {   2,  -3,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.935495164487e-05, -6.412167210557e-06, {   2,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.005068423966e-06,  2.262568764061e-05, {   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.005691736867e-06,  2.261176883557e-05, {   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.005640346041e-05,  5.403251903942e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.425885563816e-05,  1.105125456500e-06, {   2,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.006790792561e-06,  1.609589079093e-05, {   2,   1,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.416302882990e-05, -1.093257553854e-05, {   2,  -1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.871013541354e-05, -6.145454704113e-06, {   2,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.584693619642e-05, -8.815675580305e-06, {   2,  -1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -1.867463526063e-05,  5.969126399216e-06, {   2,  -1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.933529828142e-05, -5.113339453659e-06, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.933529828142e-05, -5.113339453659e-06, {   1,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.933529828142e-05, -5.113339453659e-06, {   1,  -2,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.100581132261e-06,  1.957511773599e-05, {   1,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.980373041444e-06,  1.333299692426e-05, {   0,   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.896760777808e-05,  4.266655700813e-06, {   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.735973001239e-05, -5.723641251741e-06, {   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.893797710858e-06,  1.317676529504e-05, {   0,   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.288183575231e-05,  4.386401226278e-08, {   2,  -1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.620454631995e-05, -6.719602623140e-06, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.288240825797e-05,  4.153951938559e-08, {   2,  -1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  7.216448535097e-06,  1.567339504274e-05, {   0,   1,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -9.710930130518e-06,  1.298575444421e-05, {   2,  -1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.708044527667e-05,  5.423928310835e-06, {   2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.644790381615e-05,  6.037845152506e-06, {   2,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.825780411874e-05,  3.757698966210e-06, {   1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.652413886109e-05,  5.426221341789e-06, {   2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.672115640621e-07,  2.166569425857e-05, {   2,   1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.549342689487e-05,  6.424822169169e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.532000879154e-07,  2.149317878318e-05, {   2,   1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.767463642617e-05,  3.961946065511e-06, {   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.298453296201e-05, -8.538410451929e-06, {   1,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.298453296201e-05, -8.538410451929e-06, {   1,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.080157497293e-05,  1.041315404078e-05, {   2,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.753753394400e-05,  3.657792078991e-06, {   2,  -1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.735204019249e-05, -3.558891054440e-06, {   1,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.659780359970e-06,  1.548205658200e-05, {   2,   1,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.641173238877e-06,  1.542004678496e-05, {   2,   1,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   6,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   6,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   6,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   4,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   4,  -1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   4,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   1,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,  -1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,  -1,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   1,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.000000000000e-05,  0.000000000000e+00, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   1,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   0,   1,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.979301816792e-06, -1.195389954056e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.123391016750e-05, -8.662889436427e-06, {   2,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -7.529861381260e-06, -1.221824216535e-05, {   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -8.318567291707e-06,  1.137949747863e-05, {   0,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  8.316374807381e-06,  1.133950123309e-05, {   0,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.428626729513e-05,  5.308799829581e-06, {   2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.452494271546e-06, -1.210432949357e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -7.728624586300e-06, -1.170649244515e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.488432835623e-06, -1.373442701931e-05, {   1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.477872249843e-05,  4.285112659116e-06, {   2,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.146393779131e-06, -1.179578196255e-05, {   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -7.312918101793e-06, -1.162236205399e-05, {   2,  -1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.375262368329e-05,  5.042647158294e-06, {   2,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.710371020400e-06, -1.200350677274e-05, {   0,   1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  8.069978492164e-06,  1.059108434775e-05, {   2,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.618620010795e-05, -2.438231544019e-06, {   1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.774142066979e-06, -1.177619771545e-05, {   0,   1,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.391628457888e-05,  4.583296370714e-06, {   2,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.793185152054e-06, -1.169903051731e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.885840693320e-06,  1.260458570293e-05, {   2,   1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.724325547613e-06, -1.169189465960e-05, {   0,   1,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  6.382259756238e-06, -1.199278912651e-05, {   2,  -1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.071294571644e-05,  7.652866364047e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -7.075521918898e-06, -1.123887250420e-05, {   2,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.068601549135e-05,  7.606343084513e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.306690163400e-06,  8.979677332374e-06, {   2,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -7.099382377706e-06,  1.103512622396e-05, {   2,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.604956154867e-05,  2.049830139545e-06, {   0,   1,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.602747614637e-05,  2.046109911852e-06, {   0,   1,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.608525846958e-06,  1.141197970851e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.797586785754e-07,  1.770701135322e-05, {   2,  -1,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.416994864276e-06, -1.046866826377e-05, {   1,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -4.134274340260e-06,  1.373532796875e-05, {   2,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.134274340260e-06,  1.373532796875e-05, {   2,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.645446453592e-07,  1.755903962511e-05, {   2,  -1,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.163010366746e-06,  8.619692777360e-06, {   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.011687296609e-06,  8.689046764514e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.479267025537e-06,  1.119113679476e-05, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -7.693704142050e-06,  9.829876924198e-06, {   2,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.852842916578e-06, -1.142166873738e-05, {   0,   1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -8.831481347899e-06,  8.430439853344e-06, {   0,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.278275515730e-05,  4.458510040472e-06, {   0,   1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.809780608414e-06, -1.134741777592e-05, {   0,   1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.283947316928e-05, -4.283891831392e-06, {   2,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.110911978322e-05,  5.868964760234e-06, {   2,  -1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.154724960884e-06, -1.573103860105e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.397051119015e-05,  2.913802855693e-06, {   2,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.230780452554e-05,  4.426507481442e-06, {   2,  -1,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  1.230635115260e-05,  4.424698018459e-06, {   2,  -1,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.149596828169e-06, -1.557590048761e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  9.341029830177e-06, -7.380190907724e-06, {   0,   1,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -9.312408725832e-06, -7.357806579229e-06, {   0,   1,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  7.226102413431e-07,  1.581132685824e-05, {   2,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.364178544316e-05, -2.830331255532e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -6.059208944135e-06, -1.039812627975e-05, {   0,   1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  6.036038015523e-06, -1.035916423532e-05, {   0,   1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -7.240342344759e-07,  1.566803786373e-05, {   2,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.242072251814e-05,  3.941037510809e-06, {   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.982788252868e-06,  6.354763537358e-06, {   2,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -5.807256939612e-06, -1.038795606262e-05, {   0,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.152609078512e-06,  1.396817015723e-05, {   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.791149364019e-06, -9.801048603679e-06, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.368581117112e-05, -1.890309153346e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.164459714403e-05, -3.887776533128e-06, {   2,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.042743647035e-05,  5.101171475661e-06, {   0,   1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.887590925887e-06,  1.060777037101e-05, {   0,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.126590355550e-05,  4.217431220680e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.191894307928e-05,  3.475126051640e-06, {   2,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.274270101769e-05,  2.639419929055e-06, {   1,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.012352139247e-05,  5.205740712760e-06, {   0,   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  4.831394382094e-06,  1.035144445250e-05, {   2,  -1,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.258237298824e-05, -2.573656639370e-06, {   1,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.016369635526e-05,  4.938881465385e-06, {   0,   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  5.039366131332e-06, -1.006265159288e-05, {   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.508764708872e-06,  8.580644445704e-06, {   2,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.264448152484e-05,  2.412834401389e-06, {   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.254569003599e-05, -2.324857935804e-06, {   1,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.669759820005e-06, -1.014414106810e-05, {   0,   3,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  4.669759820005e-06, -1.014414106810e-05, {   0,   3,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.758909696831e-06,  5.033822070243e-06, {   0,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -8.316584434417e-06, -6.417332453570e-06, {   2,   1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.542463246530e-06,  6.161050466416e-06, {   2,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.074917427036e-05,  3.775314788393e-06, {   0,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.388167237984e-06,  7.058486754793e-06, {   0,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.372754103303e-06,  7.017297081101e-06, {   0,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.122569878498e-06, -1.320926413044e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.114326790697e-06, -1.311360237950e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.176099966953e-05, -2.413246191185e-06, {   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.661842382545e-06, -9.510208908099e-06, {   2,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.166146745301e-06,  7.909315526730e-06, {   2,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  4.175817027506e-06, -9.889691260465e-06, {   1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.767008974342e-06, -9.289350697382e-06, {   2,   1,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.318636855123e-06,  1.272204664147e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  1.048461626474e-05,  3.514746171412e-06, {   0,   1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.166773331990e-06,  4.454618485164e-06, {   0,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.127052242563e-06,  4.491206204206e-06, {   2,  -1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -4.595184227577e-06, -8.936441876136e-06, {   2,   1,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  9.084432859523e-06,  4.391014394272e-06, {   0,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.035812524711e-05,  3.062241305190e-06, {   2,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.122145199139e-05, -2.174205747525e-06, {   1,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.014796858017e-05,  3.228792670590e-06, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.897862223582e-06,  3.342362559341e-06, {   0,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.970261159196e-06, -9.196915289671e-06, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -9.207224781929e-06,  3.952185295419e-06, {   1,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.060802466697e-05,  2.212407544695e-06, {   2,   1,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.668059583083e-06,  3.068808229023e-06, {   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.068339278192e-05,  1.937888950358e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -1.040278286808e-05,  1.926937269873e-06, {   1,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.079251429314e-05, -1.503154735125e-06, {   1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.141246185612e-05, -8.568128227385e-07, {   2,  -1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -9.667649140712e-06, -2.556669726830e-06, {   1,   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.901138071173e-07,  1.172265242608e-05, {   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.072236234458e-06, -1.000337636082e-05, {   2,  -1,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.104785043150e-05, -8.283446617485e-07, {   2,   1,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.055576059371e-05,  1.300284529137e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 {  1.019535506411e-05,  4.618205914507e-07, {   2,  -1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.995616704621e-06,  2.960518439259e-07, {   1,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.011001727832e-05, -9.613518068427e-08, {   2,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.010726803313e-05, -9.553687127861e-08, {   2,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   6,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,   1,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,   1,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,  -1,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,  -1,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   4,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   3,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,  -1,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -1,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -3,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   1,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   1,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   1,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   1,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   1,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   0,   1,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.000000000000e-05,  0.000000000000e+00, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -5.534019182140e-06, -2.251566953436e-06, {   1,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -5.005659680700e-06,  9.505650171890e-07, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.679111751830e-06, -8.837022513580e-07, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 52> g_PERT_S2_3
{ {
 { -2.200000000000e-04,  0.000000000000e+00, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.246507995485e-05, -4.031873652025e-05, {   0,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.242860002116e-05, -4.030282025139e-05, {   0,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.000000000000e-05,  0.000000000000e+00, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.000000000000e-05,  0.000000000000e+00, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.000000000000e-05,  0.000000000000e+00, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.000000000000e-05,  0.000000000000e+00, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.000000000000e-05,  0.000000000000e+00, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.021677746328e-05, -4.455428450976e-06, {   0,   1,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -9.899690948714e-06, -4.317082963701e-06, {   0,   1,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.073699331578e-05,  3.347266218074e-06, {   0,   1,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.073699331578e-05,  3.347266218074e-06, {   0,   1,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.000000000000e-05,  0.000000000000e+00, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.199322855032e-06, -1.395032240684e-06, {   2,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.192984955151e-06, -1.392267002143e-06, {   2,  -1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.188032338450e-06,  3.010346038888e-06, {   0,   1,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.187736863286e-06,  3.009884125589e-06, {   0,   1,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.131481129355e-06,  2.744944925707e-06, {   0,   1,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.131035906816e-06,  2.744337522347e-06, {   0,   1,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.244483686107e-06,  2.516286466715e-06, {   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.307065591441e-07, -2.706138448899e-06, {   0,   1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.305754190879e-07, -2.705904144798e-06, {   0,   1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.064358060351e-06, -9.000674002296e-07, {   2,   1,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.001407220255e-06, -8.726018214433e-07, {   2,   1,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.782929256854e-06, -7.774334929050e-07, {   2,   1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.767752568193e-06, -7.708118729631e-07, {   2,   1,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.506716595573e-07, -1.578894713271e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.149161009082e-06,  1.111616457565e-06, {   1,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.149161009082e-06,  1.111616457565e-06, {   1,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.162391683974e-06, -9.814778760499e-07, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.209464739376e-06,  8.153103487166e-07, {   0,   1,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.209464739376e-06,  8.153103487166e-07, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.202966027219e-06, -5.706268386958e-07, {   1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.175565893256e-09,  1.674387247691e-06, {   2,  -1,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.196441912759e-09,  1.677437534503e-06, {   2,  -1,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  8.596746262128e-07, -8.130196576743e-07, {   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -6.270881955755e-07,  9.673753400625e-07, {   2,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.264117003668e-07,  9.601491866275e-07, {   2,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.190815851097e-06,  3.713886065489e-07, {   0,   1,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.186323727555e-06,  3.699132037352e-07, {   0,   1,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.496961289165e-07, -5.742348136199e-07, {   2,  -1,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  9.249678371892e-07, -5.622393383588e-07, {   2,  -1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  9.424333738184e-07, -4.109996907238e-07, {   0,   1,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.638014694755e-07, -1.073083049958e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.034638453307e-06,  1.399356091698e-07, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.351587347086e-07,  3.310447607780e-07, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 12115> g_PERT_S3_1
{ {
 { -2.189679448352e-03,  1.058616729945e+00, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.253561194684e-01,  6.513087890330e-01, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.051470167778e-01, -6.108298914226e-01, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.392796207754e-04,  5.982646497432e-01, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.166007612272e-02, -4.038985639423e-01, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.003829958930e-05, -4.564699998896e-01, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.011497888456e-05,  4.527699998870e-01, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.150630918737e-02,  1.831799315623e-01, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  9.103726330313e-02, -1.822403894756e-01, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.458216723136e-03,  2.048856424696e-01, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.618515216166e-01, -4.198186759934e-02, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.618343680779e-01,  4.197584488092e-02, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.021742030089e-02,  1.205480308530e-01, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.662743968031e-02, -1.133577781689e-01, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.060656368992e-03,  1.665893222725e-01, {   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.795193234871e-03,  1.574058972693e-01, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.251012541330e-05,  1.445128867332e-01, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.073811010489e-01, -3.158080010398e-02, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.072704591758e-01,  3.154106491235e-02, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.931475710987e-05, -1.382805205455e-01, {   0,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.057159895584e-01, -3.208238507675e-02, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.056300032410e-01,  3.205321523362e-02, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.747666539317e-03,  1.267892355512e-01, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.485001184455e-05, -1.235353685222e-01, {   2,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.609063818430e-06, -1.205865709792e-01, {   2,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.258969974438e-05,  1.202805622129e-01, {   0,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.325280495072e-04,  1.161301787990e-01, {   2,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.033846001038e-03, -1.052286363209e-01, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.716335842401e-06, -1.068899999862e-01, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.724613450082e-03, -1.050925606149e-01, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.966294084087e-05, -1.006948638407e-01, {   2,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.251574349972e-04,  9.941460323933e-02, {   0,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.600620512989e-06,  9.552999998659e-02, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.362642728919e-02, -7.601373637273e-02, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  8.909556765714e-07, -8.507999999533e-02, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.566053621681e-02,  6.896889716592e-02, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.282924765015e-02, -4.060396295222e-02, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.262861360766e-02,  4.014177881006e-02, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.708581177621e-05,  7.952756682194e-02, {   2,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.725993267663e-02,  6.067807950810e-02, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.725604356747e-02, -6.066013358734e-02, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -7.714792553719e-04,  6.203905745211e-02, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -9.130423804896e-04, -6.084903754891e-02, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.653054692488e-02,  5.101567835693e-03, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.241431425424e-04, -6.104165413626e-02, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.590456848361e-02, -5.025577459648e-03, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.013063297202e-02, -3.193383048488e-04, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.493559763333e-02,  2.543361810956e-02, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.419572664478e-03,  5.731004185642e-02, {   2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -5.534334336628e-02, -1.229853121519e-03, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.260022463532e-02,  1.313583003920e-03, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.870876506762e-02,  3.479788304735e-02, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  4.011377486231e-07, -5.344999999849e-02, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.773820703006e-02,  3.550893232392e-02, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.174615080111e-02,  1.296931056390e-03, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.897811520879e-02, -2.137149106169e-02, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.893972953662e-02,  2.135014536637e-02, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.883714895257e-02,  1.331893632560e-03, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.664086234012e-02, -3.331095813368e-02, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.439219047101e-02, -1.510507968792e-02, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.374402064067e-02,  1.570166393986e-02, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.382175758947e-04,  4.902924543977e-02, {   0,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.305944507700e-04, -4.890297284251e-02, {   0,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.360060007968e-05, -4.857995062988e-02, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.777065858683e-02,  2.016830935427e-02, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.510518642784e-04,  4.607959449719e-02, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -4.560754481061e-02,  4.495663629594e-04, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.871075381656e-02,  1.725582025498e-02, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.835284795660e-03, -4.273927963999e-02, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.604906283969e-04,  4.396412668220e-02, {   2,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.482225569642e-03,  4.105036140257e-02, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.312961992299e-04, -4.369279149815e-02, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.002365873390e-02, -2.405286123612e-02, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  9.204935497001e-04, -4.288716148049e-02, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.050915568197e-04,  4.349192098645e-02, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.644702983040e-02, -1.654227247498e-02, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.470360047933e-03, -3.441468012770e-02, {   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.625964801081e-02,  1.643369888902e-02, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.314170899040e-06, -4.256657282689e-02, {   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.577778935966e-02, -1.633023802700e-02, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.528580201381e-02, -1.626880830261e-02, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.502184556290e-02,  1.612300183245e-02, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.913995382171e-02,  2.140182999610e-02, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.754117901675e-02, -2.243006423850e-02, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.889877970490e-04, -3.886073072885e-02, {   2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.884397086810e-07,  3.855999999954e-02, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.139418141433e-02, -2.680205125511e-02, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.130206183291e-02,  2.659025476358e-02, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.347637874027e-02, -1.441027588899e-02, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.848161890483e-05,  3.778106986522e-02, {   2,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.934264994913e-06,  3.760411759107e-02, {   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.099991326559e-04,  3.724433957497e-02, {   0,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.630453653742e-02,  2.008098562641e-02, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.498891899827e-02,  7.003464032860e-04, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.490192564974e-02,  3.780971121401e-04, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.025679828520e-06, -3.450478982471e-02, {   2,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.248679689676e-04,  3.281838816719e-02, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.563880336847e-05, -3.123065376978e-02, {   0,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.944000000000e-02, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.814893921043e-02, -5.026133317916e-04, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.338435514991e-03, -2.427063322050e-02, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.777046473275e-02, -5.191337508210e-04, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.899139211903e-09,  2.807000000000e-02, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.276045853067e-05,  2.773790843046e-02, {   2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.688329489400e-02,  6.407898730360e-04, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.126996821457e-03,  2.322031319707e-02, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.702803735748e-02, -2.111146690584e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.943008123442e-05, -2.605524723377e-02, {   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.913172393942e-03, -2.177454443266e-02, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.469198650202e-07, -2.486999999399e-02, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.474923330155e-07,  2.469999999393e-02, {   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.942954766806e-03,  1.590044815124e-02, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.268484512870e-03,  2.113499113956e-02, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  7.741605718838e-03, -1.549625133455e-02, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.186123530744e-05,  2.207034297108e-02, {   2,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.134425090830e-03, -1.792358938656e-02, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.003988641309e-02, -1.965790637230e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.125311232826e-02, -7.116579991978e-04, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.430399164034e-02, -7.628764579817e-03, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.337217782712e-02, -8.212777660410e-03, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.128781133541e-03, -1.729440699171e-02, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.400279726411e-02,  7.398618876727e-03, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.087254801726e-02,  8.968997313782e-03, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.599395822981e-02, -3.658678710667e-03, {   2,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.196923466608e-04,  1.864741519470e-02, {   2,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -3.824623735368e-03,  1.517999851169e-02, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.168725293386e-03,  1.158216971245e-02, {   2,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.887913211582e-05, -1.843998879577e-02, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.134554798075e-04,  1.748390301060e-02, {   2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.668305536260e-06,  1.749762850728e-02, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -4.888736036188e-03, -1.258053582061e-02, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.637974614625e-03,  1.461303527766e-02, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.376775182198e-02,  3.332489855647e-03, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.717474742114e-03, -7.366405646731e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.324243766645e-02,  3.687051296897e-03, {   2,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.324218361099e-02, -3.687095021411e-03, {   2,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -4.820972346804e-03,  1.202202383952e-02, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.634321825030e-03,  1.280452430856e-02, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.423303410604e-02,  1.275019150114e-03, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.056853446215e-03, -1.129900483899e-02, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  5.477419742882e-03, -9.736785554587e-03, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -9.166441013177e-03, -5.884095210227e-03, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -5.555092453624e-03,  9.134895941078e-03, {   2,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.478720206362e-05, -1.458125421549e-02, {   2,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.582894451839e-03, -5.980317859569e-03, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.224791661062e-03, -1.226613951086e-02, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  9.211390120036e-06, -1.434713135720e-02, {   2,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.580994911617e-03, -5.656206684674e-03, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.999657591021e-03, -9.108411652289e-03, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.603323580520e-03,  1.137033718018e-02, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.482642905070e-06,  1.393595763305e-02, {   2,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.121152026980e-02,  2.569022108769e-03, {   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.600523812379e-03, -5.170465416145e-03, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -5.432978415401e-05, -1.333278518979e-02, {   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.187639052257e-04,  1.280756099221e-02, {   2,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.033192849744e-02,  2.876750445627e-03, {   2,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.033194700857e-02, -2.876675808344e-03, {   2,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.021072201993e-02,  2.584650056057e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  8.382668190053e-06, -1.250069350211e-02, {   2,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.816648288395e-03, -9.683106955223e-03, {   1,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.456498719024e-03,  8.037793984513e-03, {   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.810739080283e-03,  9.663119447678e-03, {   1,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.445963402779e-03,  7.594161266949e-03, {   0,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.466598756684e-04,  1.184869439091e-02, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  3.673331570269e-04,  1.158544799686e-02, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -8.728147344247e-04, -1.103459062617e-02, {   2,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.538829239837e-04, -1.151993252082e-02, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -7.397787586533e-03,  4.410828573242e-03, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.507839298900e-07,  1.157999755624e-02, {   0,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.137372913188e-02, -1.983419756973e-04, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.023591766204e-03,  4.483893253072e-03, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  2.439866062563e-05,  1.146605120132e-02, {   2,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.093294359307e-03,  7.379211585124e-03, {   2,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  8.817705826108e-05, -1.125426136241e-02, {   2,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.032790839424e-02, -9.257446948449e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -5.396518323997e-05,  1.109312418391e-02, {   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.646391982230e-04, -1.097249417777e-02, {   2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.820908442176e-03, -2.288440248057e-03, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -8.820153423850e-03,  2.288091099016e-03, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -3.805947583042e-06,  1.108229291401e-02, {   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.368490399286e-03, -2.335657623333e-03, {   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.183260707405e-03, -7.481875303972e-03, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.181062265009e-03,  7.476913020293e-03, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.029330910042e-04, -1.035319225600e-02, {   2,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.832626164081e-03,  6.614984680009e-03, {   0,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  5.783639092058e-06, -1.038564371078e-02, {   2,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.560726990935e-03,  8.807540694944e-03, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.669921050482e-04,  9.861124950805e-03, {   2,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.702988399583e-03, -6.575981753444e-03, {   0,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.267194734265e-05,  1.022806727896e-02, {   2,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.319204356882e-03, -6.648441014147e-03, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.530302584957e-05,  9.892189896233e-03, {   2,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.814269757382e-07, -9.899999998338e-03, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.623953025888e-03,  2.195363884137e-03, {   0,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.190147567201e-03, -7.593250940239e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  9.633210477809e-03, -6.136962617794e-05, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.569572251960e-03, -6.119795483884e-03, {   0,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.522267966921e-03, -2.166582835791e-03, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -7.516379780462e-03,  2.164331512645e-03, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.511361475162e-03, -6.160684885830e-03, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.498662144028e-03,  7.964529390619e-03, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.605682501793e-03, -1.747639935828e-03, {   2,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.399141738239e-03, -7.952870756366e-03, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.737929055866e-07,  9.219999998362e-03, {   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.443566893940e-06,  9.160646950890e-03, {   2,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.231606808945e-03,  5.863965936437e-03, {   0,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.066865514663e-04,  8.981529963374e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -7.753561401228e-03,  1.241916865425e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -5.182361942506e-03, -3.792891236210e-03, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -7.733727808792e-03, -1.232965221101e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.769614851662e-05,  8.901416901127e-03, {   4,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.161747734537e-06,  8.900003957630e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.176125022997e-07, -8.529999999189e-03, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.190924318407e-06, -8.386640614753e-03, {   2,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.669827779299e-05, -8.323788606051e-03, {   2,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.235814643044e-06, -8.255425991451e-03, {   2,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.013479933254e-03, -1.180946044762e-04, {   2,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.009083644159e-05,  7.845503043794e-03, {   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -9.578366934636e-08, -7.839999999415e-03, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.030657879253e-05,  7.603339474364e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -6.152608079867e-04, -6.992689461293e-03, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -7.577094447310e-03, -3.043274267800e-05, {   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.169800127406e-03,  3.425372034623e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -8.771531247549e-06,  7.553754131370e-03, {   0,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.524017978588e-06, -7.506959313952e-03, {   0,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.196248669276e-07, -7.449999999040e-03, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.780829029269e-03, -4.615006768692e-03, {   0,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.622008493434e-03, -4.771339165690e-03, {   2,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.778311963426e-03,  4.611000944250e-03, {   0,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.505141209883e-03, -2.882658412002e-03, {   0,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  5.981061806004e-03,  1.373557620298e-03, {   2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.047109326592e-04,  7.241223791443e-03, {   0,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.021919560742e-03,  3.313765952647e-03, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.452175968098e-05, -7.204215309817e-03, {   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.912130377006e-03,  2.821053151815e-04, {   2,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.595844341553e-05,  7.115449938495e-03, {   4,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.751068715474e-03, -1.345216107349e-03, {   0,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.798171834883e-03, -2.779702931612e-04, {   2,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.049336569075e-04,  6.198882248595e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.564812076391e-04,  6.844915474416e-03, {   0,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.904819031810e-03,  2.037707418360e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  5.130173870476e-03,  1.640703275817e-03, {   2,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  5.123586058146e-03, -1.638684800765e-03, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.505919418100e-04, -5.808787298123e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -5.434059130425e-03, -1.293964103749e-03, {   2,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -4.570624524008e-03, -2.146677970051e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -6.134014014503e-03,  5.816083025316e-04, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.387372824109e-03,  1.303991541674e-03, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.371017009063e-06, -6.640694110761e-03, {   2,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.982356795099e-03, -4.654338536399e-03, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.346784254838e-04,  5.694994199449e-03, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.511504598791e-05,  6.582564168227e-03, {   0,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.914219940047e-06,  6.581554964444e-03, {   0,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.965579053615e-03,  4.616464613721e-03, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.102489581855e-07,  6.579999999076e-03, {   2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.182797202176e-03,  1.380600340635e-03, {   0,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  4.751454354490e-07,  6.441513871964e-03, {   2,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.670840162209e-06,  6.336219988354e-03, {   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.948952726188e-06,  6.330224040841e-03, {   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.369850100164e-05, -6.286271154929e-03, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.307447763082e-03,  1.275278286627e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.256919983049e-03, -3.988159061067e-03, {   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.433154634233e-03,  2.809076086036e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  5.438434451241e-03,  7.900196676628e-04, {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.660552701496e-08, -6.209999999825e-03, {   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.039455787180e-03, -4.151623034656e-03, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  3.790419218407e-03, -2.391704446498e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -1.331166400403e-07,  6.165881454640e-03, {   2,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.032311464242e-03,  4.131257485672e-03, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.267157609611e-03,  2.697769740855e-03, {   0,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.322608163075e-03, -3.634720960374e-03, {   2,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  3.150153338782e-05,  5.916970503001e-03, {   0,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.310326894620e-03,  3.616052550075e-03, {   2,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.058617232230e-04, -5.784625671633e-03, {   0,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.662797429206e-03, -2.149556761976e-04, {   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.634868503724e-05,  5.791659181584e-03, {   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.609043637752e-03,  2.242085859919e-03, {   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.587801770807e-04, -5.584420719934e-03, {   2,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  9.541559629298e-05, -5.724923260702e-03, {   2,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.250529652823e-03, -3.559713507417e-03, {   2,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -1.352574805770e-03,  4.448117047970e-03, {   0,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.051176566442e-03,  1.737814471232e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.233689497623e-03,  3.534426704717e-03, {   2,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.976413790332e-06,  5.763518764102e-03, {   2,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.300770867071e-03,  2.444760884249e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  4.489961760130e-03, -1.197836262463e-03, {   0,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  5.459892016451e-03, -1.692975584767e-04, {   2,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.546509348891e-03,  2.076702131978e-03, {   2,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -6.834122234392e-04, -4.931115631105e-03, {   2,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -2.649026957787e-03, -2.948518802967e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -2.036748673820e-03, -3.537341349110e-03, {   2,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -7.672258783105e-05,  5.495511126788e-03, {   2,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.257227246486e-06,  5.562015376929e-03, {   0,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.263214923441e-03, -2.281883083439e-03, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.067402143343e-03, -1.444175398697e-03, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  4.066798869113e-03,  1.443980515285e-03, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  7.239228388916e-07,  5.506292401155e-03, {   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.386352809384e-03,  2.104994335745e-03, {   2,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.560506854817e-03,  2.925216322622e-03, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  5.029807264433e-03,  4.507248872144e-04, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.390357481352e-03,  1.062574047972e-03, {   0,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.933686940348e-06,  5.409374293921e-03, {   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.141882977797e-05,  5.348086449683e-03, {   2,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.115538558007e-03,  2.205774715067e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  4.171078148752e-03,  1.063457321190e-03, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -4.224717596013e-03, -1.004098027356e-03, {   0,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.529428551735e-04,  4.261403761592e-03, {   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.579159188534e-07, -5.182923214253e-03, {   2,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.883503220535e-03,  2.635360806583e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.818729927393e-03,  2.321397949475e-03, {   0,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.192758550985e-03, -3.934821827681e-03, {   0,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.218365783608e-03,  1.876386634881e-03, {   0,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  3.451237438389e-03,  1.611923469107e-03, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.682527969056e-03, -3.368250067110e-03, {   1,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.165494889541e-06,  5.043272570616e-03, {   0,   0,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.675997525588e-03,  3.355101084670e-03, {   1,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.820434640901e-03, -1.978647280217e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.453663018006e-03, -1.520642293568e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.617577609044e-04, -4.806148273757e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  2.268179247138e-06,  4.942751433590e-03, {   4,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.643168395062e-03,  3.289383066531e-03, {   4,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.208865894443e-06,  4.912141576368e-03, {   4,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.280668184713e-08,  4.879999999714e-03, {   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.531139780292e-03,  1.344556365345e-03, {   0,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.605101656695e-03, -2.542177639491e-04, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -8.917670309797e-04, -3.929881973236e-03, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.603583845832e-03, -3.210088788715e-03, {   4,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.714829746271e-03,  8.610083998406e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.812957863291e-03,  9.792627399232e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.035385390251e-03, -1.753814998654e-03, {   0,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.789775463625e-03, -9.730499668373e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.686324019613e-03, -5.353415510376e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.411063031268e-03, -1.307836826744e-03, {   0,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.109471104415e-04, -4.582847199612e-03, {   2,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.492223983220e-04,  3.722550085192e-03, {   2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.220053383077e-05,  4.544973727128e-03, {   2,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.523674486051e-03,  3.081031455290e-03, {   2,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.619277762324e-03, -2.967542236832e-03, {   2,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  4.430916849271e-03,  1.462504345693e-04, {   2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.846971408483e-03,  7.134732072071e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.217929515524e-03, -1.326206256919e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.323475214627e-03, -2.211411271379e-03, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.335724114317e-03,  2.191543852536e-03, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.288700126071e-03, -2.335721802417e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -4.356785287398e-03, -1.653277739852e-04, {   2,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.836350151189e-03,  1.678366572974e-03, {   2,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -4.463648368760e-03,  4.666474896680e-05, {   0,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.225983021580e-03, -2.825935977958e-04, {   2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.040940693393e-06, -4.494743076498e-03, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.212981357880e-03, -3.279948298450e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  6.680266543465e-04, -3.808990770075e-03, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  7.801621758891e-10,  4.470000000000e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -1.569712547632e-05, -4.377732649762e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.462071315235e-06,  4.390367298060e-03, {   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.533839807267e-03, -2.852731625965e-03, {   2,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.318752484142e-03, -3.065165775310e-03, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -4.327859964062e-03,  4.563145128257e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.545836180593e-03, -2.811524433879e-03, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.277870175715e-03, -3.061935970435e-03, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.108545423732e-03, -3.226717688000e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -4.057350790274e-06,  4.328713299024e-03, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.470427591510e-03, -8.468877340654e-04, {   2,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.174602935207e-03, -1.127164881018e-03, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  3.173731733799e-03,  1.126890956220e-03, {   2,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -3.424172865563e-03,  8.421573732469e-04, {   2,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.410404439472e-04,  3.309972585593e-03, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -9.410136092688e-04, -3.305932235320e-03, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  7.092919072110e-04, -3.533639624077e-03, {   2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.265732011728e-09, -4.240000000159e-03, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.134967100580e-06,  4.220243239647e-03, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.025468486090e-06,  4.211758391652e-03, {   2,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.367428373761e-06, -4.209376430468e-03, {   2,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.238513797851e-03,  2.957567333052e-03, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.269527515852e-03,  2.923820434063e-03, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.705605668154e-04,  4.008301604687e-03, {   4,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  8.736367134259e-04, -3.289591820094e-03, {   2,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  3.325437380951e-03,  8.025509101301e-04, {   0,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.235075228195e-03, -2.856886958707e-03, {   2,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.447210695465e-03,  2.634197306887e-03, {   1,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.447210695465e-03, -2.634197306887e-03, {   1,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -2.096155618661e-04, -3.856982229922e-03, {   0,   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.484401464427e-03, -2.581565383793e-03, {   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.751873991199e-03, -1.266874332658e-03, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  6.032261906371e-04, -3.413821486209e-03, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.319461191368e-06, -3.999602688819e-03, {   2,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.670099701741e-03,  1.299381338332e-03, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.042931284500e-09,  3.910616085138e-03, {   0,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.719759674944e-03,  1.190287667784e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -7.868952616207e-04,  3.114237297216e-03, {   2,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.462558703024e-08,  3.889999999846e-03, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.017558733071e-03, -1.872080925684e-03, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.380107788405e-03, -1.494898000186e-03, {   2,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.181822324136e-03,  6.900140219950e-04, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.016163925128e-03, -2.853420191801e-03, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.233445499757e-03,  1.601604390429e-03, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.010557477027e-03, -2.804980788947e-03, {   2,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.604761380679e-04, -3.039830133809e-03, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.317781419037e-05, -3.771521132278e-03, {   2,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.263646896065e-03, -2.529674264346e-03, {   0,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.791495482516e-03,  8.231130914300e-07, {   1,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.902564287855e-03,  8.271722871193e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  5.970221708761e-06, -3.699685714697e-03, {   4,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.067910076112e-04,  3.565003020741e-03, {   2,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  3.336676297017e-03, -2.889809657181e-04, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.537967757202e-03,  7.740241658993e-05, {   2,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.726465177142e-03,  8.829282620303e-04, {   2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.266150269223e-03,  1.319106090690e-03, {   0,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.297864629530e-03, -2.862721578840e-04, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.191887863851e-03,  2.385946996514e-03, {   0,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.052345453604e-04, -2.651827168938e-03, {   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.046353499211e-04, -3.147538096639e-03, {   2,   0,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.423870691272e-05, -3.493613011739e-03, {   1,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.012524664052e-07,  3.497647731479e-03, {   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.136591917103e-05,  3.408046007346e-03, {   1,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.783866233093e-03, -6.548136365732e-04, {   2,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -7.064045613883e-08, -3.429999999273e-03, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.605002334082e-03,  8.216279617940e-04, {   2,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.871931059014e-04, -3.131860361908e-03, {   0,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.224983215381e-03,  2.184023564131e-03, {   2,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  3.087526281200e-03,  2.841868638796e-04, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.338397389851e-03,  2.511916539949e-05, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.057633236364e-03, -2.813357140582e-04, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.119894883294e-03, -2.216405699594e-03, {   0,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -2.057642496522e-03, -1.247516846773e-03, {   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -3.271896247066e-03, -1.683922973239e-05, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.083401026970e-03,  2.168777724224e-03, {   0,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.341529727340e-05,  3.208187293472e-03, {   2,   0,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -3.093304798302e-04, -2.897543309201e-03, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.486587165015e-03, -1.716373784615e-03, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.112416513355e-05,  3.110312647954e-03, {   4,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.094565407657e-03, -2.033760034725e-03, {   2,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.310929447264e-04, -2.586042622172e-03, {   2,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.933510993624e-03, -1.182409362378e-03, {   0,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.007879551134e-05, -3.092161956614e-03, {   2,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.028175608908e-03, -8.111142873986e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.784002151241e-03, -1.309254709350e-03, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.838600028348e-04,  2.408792734021e-03, {   2,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.026802712837e-03, -6.339922285995e-05, {   0,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.462884032925e-05,  3.054051041800e-03, {   0,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.024192979139e-06, -3.062313747789e-03, {   2,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.016494042319e-03, -2.034769715603e-03, {   0,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.473083069919e-03,  1.567869288841e-03, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.965875971806e-03, -7.195951761506e-05, {   0,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.991759407582e-03,  4.172779169736e-05, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.851070947789e-03, -1.714743233802e-04, {   2,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.648449871407e-03,  1.358776473338e-03, {   2,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.592310327641e-08,  2.989999999647e-03, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.941332840747e-03,  2.909276721575e-05, {   2,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.186936357499e-04, -2.848932811996e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -2.909925009400e-07, -2.965221203784e-03, {   2,  -2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.446431449242e-06,  2.958511918845e-03, {   2,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.245412160145e-05, -2.906133256637e-03, {   2,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.931003205715e-04,  2.651898896850e-03, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.643330686393e-04, -2.476698376784e-03, {   1,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.885853042039e-03, -5.469934767118e-05, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.815241505889e-03, -1.125264282133e-03, {   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.257759238457e-03, -6.732710162213e-04, {   4,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -8.352113464873e-04, -2.095756341995e-03, {   2,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.822213267647e-03, -1.082942169512e-04, {   2,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.255668778433e-03,  6.726173832596e-04, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.091037131386e-04,  2.514411695628e-03, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -9.682959931682e-04, -1.944854164195e-03, {   2,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  1.662287126091e-06,  2.906160841792e-03, {   2,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.584803184282e-04,  2.448477163774e-03, {   1,  -1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.108160189409e-03,  7.961363148129e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -2.387488966254e-06, -2.886518664684e-03, {   2,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.808670455044e-03,  7.505755287244e-05, {   0,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.784508741631e-03,  9.183413516034e-05, {   2,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.137820171248e-03,  7.232323154064e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.222926466619e-03,  6.334476359000e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -1.697523668248e-03, -1.114298011560e-03, {   2,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.059345042541e-07,  2.809999998003e-03, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.042407159602e-05, -2.784087784255e-03, {   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  2.371425029528e-04,  2.558241450540e-03, {   0,   0,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  5.648583590762e-08,  2.789999999428e-03, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.681299752838e-03,  1.105854663553e-03, {   2,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.093625138094e-05, -2.763061365053e-03, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  4.627936416661e-07,  2.765462126176e-03, {   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.670862585442e-03, -9.169816625228e-05, {   2,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.166925577204e-03, -5.839808367830e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.579929203559e-03, -1.165310801428e-03, {   0,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.577939594389e-03,  1.164083259736e-03, {   0,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.592335612461e-07,  2.730599357791e-03, {   0,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.911918034226e-03, -8.167405051350e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.173090820671e-05, -2.696670221172e-03, {   2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  3.646312508325e-06, -2.702829533950e-03, {   0,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.618679819418e-03, -6.618052324286e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.275763918782e-05,  2.655057437581e-03, {   0,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.175423854258e-04, -1.854886250097e-03, {   2,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  5.340428071693e-04,  2.134389318380e-03, {   1,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.802045074069e-03,  8.632969919880e-04, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -4.366392848014e-04, -2.226365813364e-03, {   2,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.153378943345e-03,  5.094637506906e-04, {   3,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.652547081170e-05,  2.619782840156e-03, {   2,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.435896339438e-04, -1.998947818999e-03, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.615753959118e-03, -1.024964889351e-03, {   2,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.059182735245e-03,  5.734162891435e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.534523689271e-03, -9.765711807003e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.514604338588e-03, -1.095640295677e-03, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.949973422887e-03, -6.598480380437e-04, {   2,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.637915125880e-03,  9.529250339367e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.959971964443e-03, -6.233843977194e-04, {   2,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.084972514161e-03, -4.894576385602e-04, {   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.854865543653e-06,  2.569998010358e-03, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.409522625826e-06,  2.568073721853e-03, {   2,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.063844842677e-03,  4.996447413664e-04, {   2,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.975036878586e-03,  5.805201176718e-04, {   1,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.970253476680e-03, -5.796156608903e-04, {   1,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.350876243802e-03,  1.197499102657e-03, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.595966516333e-04, -1.774235871249e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.429614274487e-04,  2.385367257956e-03, {   0,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  2.302668460996e-05,  2.500525065703e-03, {   2,   0,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -6.499366848010e-04, -1.852730033768e-03, {   2,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  8.867139642752e-04, -1.613270445281e-03, {   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  6.488070996527e-04, -1.825795988282e-03, {   2,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.635119381378e-03,  8.377111339305e-04, {   2,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.693709916050e-06, -2.446899310517e-03, {   4,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.074885187188e-06, -2.423007701631e-03, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  1.386560193516e-03, -1.040375197083e-03, {   2,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.497594889853e-03,  9.262275868845e-04, {   2,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  1.479379892410e-03,  9.377303207906e-04, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.602631027171e-03, -7.909533752359e-04, {   2,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.158626712000e-03, -2.303569294640e-04, {   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.460356084383e-03,  8.852986562928e-04, {   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.669463706293e-03,  6.749584400581e-04, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.234944960975e-03, -1.103340985153e-03, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  8.174638467368e-04,  1.520363762563e-03, {   2,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.393577698682e-06, -2.323492918278e-03, {   2,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.668225209344e-05,  2.211265733974e-03, {   0,   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -6.090087326352e-04, -1.692684994955e-03, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.866530329318e-03, -4.302696969862e-04, {   2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.525100610551e-05, -2.230627721236e-03, {   1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.419164522662e-03,  8.494889024513e-04, {   2,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.441573050134e-04,  2.122416656130e-03, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.124564946642e-03, -1.137898376638e-03, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -3.422402471405e-04, -1.911021464862e-03, {   2,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.760917346719e-03,  4.902826644402e-04, {   0,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.760674959323e-03, -4.902376134447e-04, {   0,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.406668472431e-03, -8.418360376702e-04, {   2,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.212392133567e-03, -2.673272055881e-05, {   2,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.032730558762e-08,  2.229999999184e-03, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -2.119479244339e-03,  1.047107719128e-04, {   2,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -2.060510531463e-03,  1.558400063925e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -6.066764479125e-08, -2.199999999164e-03, {   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.088537616700e-03,  1.071418336290e-04, {   2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.551475842099e-03,  6.439895776865e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  1.550288828181e-03, -6.434935849056e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -4.068893863534e-04, -1.785291139870e-03, {   2,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.605353846045e-08, -2.189999999941e-03, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.151542547507e-05, -2.117185901396e-03, {   2,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.379748692171e-07,  2.181248727165e-03, {   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.478937587293e-08,  2.179999999311e-03, {   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.011033247418e-03, -1.644922885963e-04, {   2,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  7.238006760025e-04,  1.448929389471e-03, {   4,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.336069898201e-03, -8.358466258958e-04, {   2,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.691333839275e-03, -4.718540264706e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  2.062547587417e-03,  8.691859077390e-05, {   2,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.147756743762e-04, -1.430834908514e-03, {   4,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.017245282238e-03,  1.120078103339e-03, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.795802228858e-03,  3.360232220472e-04, {   2,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.027609181465e-03,  1.026519684919e-04, {   2,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.885456104257e-03, -2.443979451400e-04, {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.045814443819e-04,  1.411460369050e-03, {   4,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.101523877056e-04, -1.695129799069e-03, {   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.819813702545e-03, -2.760119236930e-04, {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.286261917411e-05, -2.041595416777e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -7.044883725846e-04, -1.389349511303e-03, {   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.299539137643e-03, -7.939149821371e-04, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  5.015140471005e-06, -2.077097450433e-03, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.206824925170e-04, -1.460214686379e-03, {   0,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.477930802775e-06, -2.071691869304e-03, {   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.911982324307e-03, -1.525226922347e-04, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.155975172252e-04,  1.448728909005e-03, {   0,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.051282439125e-03, -1.137217517952e-05, {   3,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.120466527127e-05,  1.955660747618e-03, {   4,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.855297896477e-03, -1.737926975116e-04, {   2,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.900218326120e-06,  2.025164804071e-03, {   2,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.680248249078e-05,  1.997752012610e-03, {   0,   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.988565882434e-03,  2.594171977096e-05, {   0,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.935800699181e-03, -7.518652192943e-05, {   2,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.804197143053e-04,  1.318860119676e-03, {   0,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -4.880714399517e-05, -1.944892315590e-03, {   4,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  6.623837679535e-04, -1.325969316289e-03, {   4,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.024368888303e-03,  9.621810976090e-04, {   2,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.939045448405e-03,  4.563512665430e-05, {   0,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.238857959279e-05,  1.916744982235e-03, {   2,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.477781680776e-04, -1.424724832352e-03, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.997378703671e-06, -1.960366710145e-03, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.941658785095e-03,  1.881555529393e-05, {   2,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.485148175877e-04,  1.298217981490e-03, {   2,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.863002866283e-03,  7.893307318188e-05, {   2,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.049998877208e-06, -1.925217090114e-03, {   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.462110884485e-04, -9.768749029927e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.194990102791e-06,  1.911928903282e-03, {   0,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.896035608321e-03, -1.691646345081e-05, {   0,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.474145395005e-03,  4.218445187409e-04, {   0,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.274888280598e-03,  6.114230991042e-04, {   0,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -1.274765411506e-03, -6.112534954086e-04, {   0,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  1.125331783537e-03, -7.583570242335e-04, {   0,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.050295105831e-03,  8.313548583251e-04, {   2,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.258253119028e-04, -1.252825330273e-03, {   2,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.140495698822e-07, -1.864870979523e-03, {   2,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.462132472393e-03,  3.893989362146e-04, {   2,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  3.438234851058e-04,  1.500493524503e-03, {   4,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.447104513976e-03, -3.852407870035e-04, {   2,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -4.454254783632e-08, -1.809999999452e-03, {   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.934848967341e-06, -1.804321041196e-03, {   2,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  6.163505142690e-04,  1.176214786345e-03, {   0,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.000139143329e-05,  1.777707886033e-03, {   0,   0,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.431563112664e-04,  1.032580784548e-03, {   0,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -7.427354626366e-04, -1.031948949678e-03, {   0,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  9.267698323812e-08, -1.769999997574e-03, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.046275227714e-03, -7.216352769286e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.488064412075e-03,  2.765719028820e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.395987997126e-08, -1.759999999837e-03, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.689746016055e-06, -1.751189585647e-03, {   2,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.532514307012e-03,  2.193898592986e-04, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -5.769755412639e-04, -1.162156076390e-03, {   2,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.075787357687e-05, -1.727725077982e-03, {   0,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  6.991908939219e-04,  1.036352129396e-03, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -8.997029377107e-04, -8.285608190407e-04, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  8.745545637658e-07,  1.724373906366e-03, {   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  8.410596744823e-05,  1.628714360395e-03, {   0,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -1.150075291385e-03,  5.603999313581e-04, {   2,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -2.269752053360e-04, -1.483339834170e-03, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.269730487861e-04,  1.483272461066e-03, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.706923878861e-03,  2.354883520490e-06, {   2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.299548351601e-03,  4.078823891530e-04, {   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.108467604800e-05, -1.621037626051e-03, {   0,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.430796417009e-03, -2.659665415307e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.620433558918e-03,  6.894308360044e-05, {   2,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.544533002517e-04,  1.432787139583e-03, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.644081881827e-06, -1.678953433944e-03, {   0,   0,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  2.551455161034e-04, -1.422251215956e-03, {   2,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.154462063484e-03, -5.194235168627e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.512567145912e-03,  1.591213067790e-04, {   2,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.232025707793e-05,  1.597398765186e-03, {   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.609306515581e-03, -4.082068756712e-05, {   2,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.849287432473e-05,  1.574229615322e-03, {   4,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.056088452488e-03, -5.767506718175e-04, {   0,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.012229707461e-03, -6.176110771428e-04, {   2,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -5.902671142011e-04, -1.032795746078e-03, {   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.565199449645e-03, -4.974456712762e-05, {   0,   0,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  9.264650575428e-04, -6.857323387844e-04, {   2,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.297920005491e-03,  3.141615060436e-04, {   2,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.236717701760e-03, -3.753185320352e-04, {   4,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.529876292325e-03, -8.199400568629e-05, {   2,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.235405480250e-03,  3.748696141461e-04, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  8.148942277717e-09,  1.609999999979e-03, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.834534570289e-04,  1.120300603489e-03, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.549335128639e-03, -5.033218388097e-05, {   0,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.624397198359e-04, -1.325689984627e-03, {   2,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.061664236770e-03, -5.219461412190e-04, {   2,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.016630868725e-03,  5.602667100950e-04, {   0,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -5.074515874924e-04, -1.066166070394e-03, {   2,   0,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  1.503108484833e-03,  6.852833733991e-05, {   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.103393061999e-03, -4.680011583812e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  1.103307024184e-03,  4.679892954894e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  3.292844486504e-04, -1.234732193741e-03, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.296780750251e-03, -2.662444623977e-04, {   2,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.259867842403e-03, -2.989109088279e-04, {   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.522789632377e-03, -3.093990010060e-05, {   0,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.500582627974e-03,  4.756299510081e-05, {   4,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.317175710300e-06, -1.538795309538e-03, {   4,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.033994578885e-03, -5.050953693180e-04, {   2,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.417683076546e-03,  1.209542664002e-04, {   2,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.996163772148e-04,  8.294264604913e-04, {   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  3.369183587788e-08,  1.519999999627e-03, {   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.369183587712e-08, -1.519999999627e-03, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.316788664831e-04, -9.793784069493e-04, {   2,   0,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.209345856320e-03, -2.970184428698e-04, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.245197815840e-03, -2.551014153683e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -1.208874643505e-03,  2.904344489173e-04, {   1,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.054427062444e-08,  1.489999999858e-03, {   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.277882254748e-04,  1.253618770517e-03, {   0,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.317881540592e-03, -1.549947971774e-04, {   2,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -7.782646529051e-04, -6.894859418100e-04, {   2,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.025519848311e-04, -1.261988903907e-03, {   0,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.310991989272e-05,  1.449606371993e-03, {   4,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.478896877120e-04, -8.989483213526e-04, {   2,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.314090343701e-03,  1.286184722825e-04, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.406651821440e-03,  3.136129604042e-05, {   2,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.378712366272e-03,  5.880673389074e-05, {   2,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  6.671640397120e-04, -7.669242586174e-04, {   2,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  5.913574060031e-04,  8.306251080358e-04, {   2,   0,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.366998583646e-03, -5.360723683182e-05, {   2,   0,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.027175951415e-04,  5.169080395994e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.863462565076e-06,  1.414725340140e-03, {   2,   0,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.442948512065e-06,  1.410002428094e-03, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -3.896199188923e-04,  1.018907647980e-03, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.361140420803e-03,  4.514903804909e-05, {   2,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.321150603607e-03,  8.328638105677e-05, {   2,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  6.304374222581e-04,  7.724775734006e-04, {   2,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  1.128929396316e-03, -2.716493333480e-04, {   1,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.131552951229e-03,  2.684114241330e-04, {   1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.342909175966e-03, -4.964325928457e-05, {   2,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -2.741978952964e-04, -1.117464156315e-03, {   2,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.483044598870e-04, -8.431928281814e-04, {   2,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.106279284450e-03, -2.846417585498e-04, {   0,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.253819804641e-03,  1.246430803912e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.395929624719e-07,  1.370326691833e-03, {   0,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.357683121298e-03,  8.313722459925e-06, {   1,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.692079279442e-04,  9.963623598028e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.248861725457e-06,  1.344786994081e-03, {   0,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.087169583152e-03, -2.587849558336e-04, {   1,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.320924017806e-03, -2.452190571280e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.071291468443e-03, -2.700466609865e-04, {   2,   0,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.552868607359e-04,  8.838646049804e-04, {   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.554583843603e-04,  1.079426275737e-03, {   2,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.944823587040e-07, -1.325005389027e-03, {   4,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.479378309446e-04,  1.172825485910e-03, {   2,   0,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -1.215995828606e-03,  1.019334330398e-04, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.639000379782e-06, -1.313437871892e-03, {   2,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.540525723888e-05, -1.298207122413e-03, {   1,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -3.849057888767e-05,  1.268294553678e-03, {   2,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.946686774723e-06,  1.301944142382e-03, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.024503496194e-03,  2.724982723073e-04, {   2,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -6.029072113209e-04, -6.906725530630e-04, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.230806796420e-03,  6.173271423504e-05, {   2,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.824602654947e-06,  1.283205972741e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -3.389856072795e-04,  9.429860592720e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  2.274483215942e-04,  1.047820053718e-03, {   2,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.632593963938e-04,  9.077091555799e-04, {   2,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.971680201860e-04,  5.681661605209e-04, {   2,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.768053367238e-04,  7.869724509258e-04, {   0,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.228136209784e-04,  1.036535340973e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -1.915766917668e-04,  1.064869645666e-03, {   2,   0,  -1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -4.181627347613e-04, -8.377804658887e-04, {   2,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.372656686752e-05,  1.240945601526e-03, {   4,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.239972159001e-03, -8.071482753646e-06, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.013311864367e-03, -2.306493643280e-04, {   2,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.166706157726e-04, -3.162762153112e-04, {   0,   0,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -8.262400901674e-04,  4.028495069559e-04, {   2,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  5.927792275808e-07, -1.227223198123e-03, {   4,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.077519325204e-07, -1.224182172653e-03, {   4,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.711550897087e-04, -7.528391509766e-04, {   0,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -3.948918331893e-04, -8.278084436262e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.583071691073e-05, -1.202892864122e-03, {   1,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.179571513813e-03,  3.810913500422e-05, {   2,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.931545849166e-04,  8.241754827863e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  3.588887319213e-04, -8.519338306672e-04, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -5.417369566160e-04,  6.668980813258e-04, {   2,   0,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.309281442231e-04,  4.747906942748e-04, {   0,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.191732562724e-03, -1.152229543906e-05, {   3,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.761491397418e-04, -9.218170823593e-04, {   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.146672586882e-03,  4.881553538293e-05, {   2,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.095478074428e-03,  9.963828806024e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.774854791737e-04, -4.171137285744e-04, {   0,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.095586712492e-03, -9.843655525516e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.435577118470e-04, -5.485606382822e-04, {   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.568034831561e-04,  7.334912923105e-04, {   0,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -7.186003121867e-04, -4.693508260855e-04, {   2,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.567681141634e-04, -2.271105739341e-04, {   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.724983980095e-04, -1.011253475292e-03, {   2,   0,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.107408257780e-04,  4.694639320038e-04, {   0,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -6.737087460296e-04,  5.014956091737e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.646085625004e-04, -9.083177050044e-04, {   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -7.665170982856e-04,  4.022784100333e-04, {   0,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -4.513234166605e-04,  7.162916546018e-04, {   2,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  7.387511248840e-04,  4.267768211732e-04, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -5.095956007107e-05, -1.113057769734e-03, {   2,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  7.759586631976e-05, -1.085337853847e-03, {   2,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.065644973680e-03,  9.639778377384e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.308455033475e-04, -7.295714688289e-04, {   0,   0,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.116524135485e-03,  3.931754922329e-05, {   2,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.290913796349e-04,  7.261872027033e-04, {   0,   0,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  7.693618351199e-04,  3.848552480140e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  9.021755835045e-04,  2.511868182761e-04, {   2,   0,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  9.018655148256e-04, -2.511191828186e-04, {   2,   0,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  3.103085475404e-04, -8.416349374975e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.220402422702e-06, -1.149634065334e-03, {   1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.846558348518e-08,  1.149999999852e-03, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.111470078184e-03,  3.383267331568e-05, {   4,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.394023032242e-05,  1.098344334910e-03, {   4,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  9.018297799944e-04, -2.401525193047e-04, {   2,   0,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  4.017035821835e-04, -7.399184552679e-04, {   2,   0,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -1.156193099767e-04, -1.024821595418e-03, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -7.089785585724e-04, -4.293498327924e-04, {   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.033297891512e-04,  7.341343386416e-04, {   3,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  4.033297891512e-04, -7.341343386416e-04, {   1,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  1.912580233978e-04, -9.457655955864e-04, {   2,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -1.029463498552e-04,  1.026972044060e-03, {   2,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.079205271782e-03,  5.023967311531e-05, {   2,   0,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.112359098435e-03,  1.063891759558e-05, {   2,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.407173993537e-04,  3.760329129348e-04, {   2,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -4.027939840543e-04,  7.138670123812e-04, {   0,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  8.672642196532e-04, -2.403079770538e-04, {   2,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.528334828180e-04, -6.518685041936e-04, {   2,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  2.485137748930e-04,  8.530727074238e-04, {   3,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.350141506906e-05, -1.065471288216e-03, {   2,   0,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.130497537658e-05,  1.086838686326e-03, {   2,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  9.024686619814e-07,  1.096690244514e-03, {   0,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.464341598332e-04, -2.489236055851e-04, {   4,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -8.456479109175e-04,  2.486490054845e-04, {   4,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.704738055131e-04,  6.234284390050e-04, {   1,   1,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  9.611532741344e-04, -1.314594494001e-04, {   2,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  2.414212944961e-06, -1.089997649484e-03, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.907918772138e-04, -7.015459135185e-04, {   2,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.179837999736e-04, -8.729906197156e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -4.495281905760e-04, -6.406479595941e-04, {   2,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -7.038912872775e-09, -1.089999999977e-03, {   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.768281345852e-04,  2.103475122020e-04, {   1,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.234871455943e-04,  7.575302708107e-04, {   2,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  4.653572126231e-04,  6.146230513757e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -7.999518011039e-04, -2.777679317817e-04, {   4,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.544446735171e-05, -1.041210036233e-03, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -2.530617818339e-04,  8.229374123174e-04, {   2,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  7.193701251145e-04, -3.483719610812e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.643412507112e-04,  8.999358912676e-04, {   2,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  6.647138273885e-04,  3.952118373175e-04, {   2,   0,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.889171381491e-04, -8.669215797920e-04, {   2,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.407957677137e-04, -4.033834135025e-04, {   2,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.852284611078e-04,  8.577690532532e-04, {   0,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.840721884924e-04, -7.499337499795e-04, {   0,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.439422469300e-06,  1.026665830313e-03, {   2,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.855583420184e-04, -3.894039370840e-05, {   2,   0,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.890127247791e-05, -9.736698726452e-04, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  6.019752519984e-04,  4.196283379485e-04, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  2.730571707583e-04,  7.479784475127e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  5.980378638491e-04, -4.162235529104e-04, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -2.320476875358e-04, -7.792447481518e-04, {   2,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.482011513593e-04, -6.188464232525e-05, {   0,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  2.966145059413e-07, -1.007839804896e-03, {   2,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.097266078819e-04,  1.959385633837e-04, {   2,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.661004116098e-05,  9.572114912326e-04, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  3.675698307027e-04, -6.307913747309e-04, {   2,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  5.214586914188e-04,  4.747452121442e-04, {   2,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  5.832128111505e-04,  4.116354004943e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  1.678082424987e-06,  9.928777773266e-04, {   2,   0,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.002466309313e-06,  9.916880491027e-04, {   4,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  8.109357534116e-04,  1.818117668226e-04, {   2,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.002458744406e-04, -1.914147302137e-04, {   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.675517783548e-05, -9.740187633005e-04, {   0,   0,   1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  9.063484106413e-04,  8.144052141231e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.317951837792e-05,  9.612105375734e-04, {   1,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.424871968100e-06, -9.804120055201e-04, {   2,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.662465058240e-05,  9.650195182099e-04, {   0,   0,   1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  9.395019580446e-04,  3.995613340989e-05, {   2,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.978580321813e-07, -9.789046481618e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -8.169901366820e-04,  1.569528634506e-04, {   2,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.421098312730e-04,  3.044682012850e-05, {   4,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.410076031861e-04, -6.277338988068e-04, {   0,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.460177069391e-04, -6.200400723294e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -1.345090669205e-04, -8.258719855616e-04, {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.639517616695e-04, -6.961669818083e-04, {   0,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.608495438597e-08, -9.599999998652e-04, {   2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.929874864331e-06,  9.507877887095e-04, {   2,   0,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.086027025436e-04,  8.510957293675e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  5.192944317193e-07, -9.572607440827e-04, {   4,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.584659612291e-04, -5.982555271211e-04, {   2,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -3.116921490896e-05, -9.247956719806e-04, {   2,   0,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.718367652612e-04,  1.839508565546e-04, {   1,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.845373095192e-06,  9.496921570513e-04, {   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.570379541068e-04, -7.962611327715e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.183622798600e-06,  9.510611546244e-04, {   0,   0,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.765326392417e-04, -3.632756632527e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 { -3.355714650102e-05, -9.032489601232e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  5.138552953410e-04,  4.215474884285e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.252815955095e-04, -7.095432557796e-04, {   0,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.029666339208e-06,  9.307758155477e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.943845746704e-04,  3.363929338928e-04, {   0,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -2.758718024960e-04, -6.491748971385e-04, {   2,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.055132123047e-04,  1.185161891498e-04, {   2,   0,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -7.454004066394e-04, -1.777316015935e-04, {   2,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.255178829577e-04,  5.970345942940e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -8.044927356792e-04, -1.177050068112e-04, {   2,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -2.749941610559e-04,  6.471925091001e-04, {   2,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.209407189157e-04, -2.006849708044e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.449165239640e-04,  6.763287590125e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.303036690982e-04, -7.896667034371e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -3.786717443146e-04,  5.409647832754e-04, {   2,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -8.659724876138e-04,  4.734181598157e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -2.655705022681e-04,  6.472675309222e-04, {   0,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.761981046867e-05, -8.945462860541e-04, {   2,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.045639661670e-04,  7.072215100833e-04, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.914864518095e-04,  5.202523314956e-04, {   1,   1,  -1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -9.059264882697e-04,  1.784725152627e-06, {   4,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.524405427968e-04, -7.467019436178e-04, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.615339258801e-04, -6.323105347092e-04, {   2,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.461583312982e-04, -3.468565937612e-04, {   4,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.618001771656e-04, -3.298327727358e-04, {   0,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  9.129102252307e-07,  8.878628104570e-04, {   0,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.738496992109e-05,  7.912205404323e-04, {   0,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  2.071495604825e-04,  6.801664614114e-04, {   0,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.411996300755e-04,  3.441140365658e-04, {   4,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.645573428800e-04,  7.198462132521e-04, {   4,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.171712772137e-04, -5.667344570978e-04, {   2,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.889675365521e-04, -4.928179028297e-04, {   4,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.377771351077e-04, -5.432535169712e-04, {   0,   0,   1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.847996356181e-05, -8.408825343613e-04, {   2,   0,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  5.378817081108e-04,  3.413547018551e-04, {   2,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  5.845669048153e-04, -2.907054829431e-04, {   2,   0,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  2.864558751257e-05,  8.444061012313e-04, {   2,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  9.834410478968e-05, -7.744064769031e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 { -8.180162865309e-04,  5.155131919155e-05, {   0,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  7.403954134155e-04, -1.279297543140e-04, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.383823957732e-04,  1.275164779630e-04, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.937520123609e-05, -7.854019248520e-04, {   2,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  4.010603861921e-04, -4.626906126659e-04, {   2,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -6.880297854151e-04,  1.735563854060e-04, {   3,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.403174471448e-05, -7.764329745974e-04, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -5.462929343864e-04,  3.135383160459e-04, {   2,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -8.554155085794e-04,  3.937017435134e-06, {   1,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.144302233756e-04, -4.127359175200e-05, {   0,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.290089342457e-04,  6.263413578804e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -3.451440163432e-04,  5.096320900374e-04, {   2,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -3.284182972143e-04,  5.259005385794e-04, {   0,   0,   1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.267076683206e-06, -8.480589020732e-04, {   4,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.879822641273e-04, -6.471429049677e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.760719488687e-05, -7.942780833943e-04, {   2,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -5.618485226427e-07,  8.503164814006e-04, {   0,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.953157029960e-05, -8.284689543633e-04, {   4,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.903187497496e-04, -6.545984764936e-04, {   3,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -6.760557949133e-04,  1.667547895957e-04, {   0,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  2.950443527408e-04,  5.454822141243e-04, {   2,   0,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.612684228756e-08, -8.399999998452e-04, {   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.348790445895e-08, -8.399999998917e-04, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.022839764004e-04,  3.198183660845e-05, {   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.866696251846e-04,  4.420071733956e-04, {   4,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.031650814304e-04, -2.425659813520e-05, {   0,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.569502497483e-04,  6.801705231754e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.830959962068e-04,  4.030524558744e-05, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -2.119625465843e-04, -6.111585318636e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -2.912362588463e-04,  5.314027947505e-04, {   0,   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -7.823532963353e-04, -3.923622243533e-05, {   2,   0,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  1.318779371931e-06,  8.182547093362e-04, {   2,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.072756748322e-04, -5.118204973177e-04, {   0,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  7.552837232883e-04,  5.785347334555e-05, {   2,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.973436515900e-04, -3.146025439039e-04, {   4,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.882306757783e-04, -5.234541594746e-04, {   2,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.199036795193e-05, -7.893829743290e-04, {   4,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.503973000453e-04,  6.571214387759e-04, {   4,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.633560198495e-04,  4.437121223004e-04, {   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -7.125053223926e-04,  9.410125447904e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -2.831319399868e-04, -5.228990060346e-04, {   2,   0,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  3.958585063952e-07,  8.030143883119e-04, {   4,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.866004327523e-04,  5.164895378135e-04, {   2,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  6.311204809681e-07, -7.995887060592e-04, {   2,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.733421815138e-04,  3.267687752569e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -5.634302736125e-04,  2.300947656963e-04, {   0,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -5.626785604343e-04, -2.297460925495e-04, {   0,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  7.599349604938e-04,  3.218860944517e-05, {   2,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.760661075190e-04,  4.154726502427e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  7.644024789797e-04, -2.640064488538e-05, {   2,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.558055423159e-08,  7.899999998464e-04, {   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  7.253981588963e-04, -6.439068934748e-05, {   0,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  2.787909613467e-04,  5.080948674761e-04, {   1,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  4.243465820499e-04, -3.620370101561e-04, {   1,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.234774370769e-04,  6.600847132781e-04, {   0,   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.882925279918e-04,  9.470738703268e-05, {   2,   0,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -4.535244221333e-04, -3.289962735741e-04, {   2,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.166241983168e-06, -7.812151296920e-04, {   2,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.542410053591e-04, -2.805150292136e-05, {   0,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -6.170837340321e-04, -1.641131438790e-04, {   2,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -2.748467108858e-04,  5.063253080385e-04, {   2,   0,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  8.712683625445e-09, -7.799999999513e-04, {   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.977510182801e-04, -2.814930697970e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -2.775497796417e-04,  5.015064447557e-04, {   2,   0,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -6.306476216448e-04, -1.477958049807e-04, {   1,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.962641859475e-04,  2.817506274483e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -6.314468664555e-04, -1.458597642207e-04, {   0,   0,   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -9.524577323103e-05,  6.792724897446e-04, {   0,   0,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.725277280502e-04,  5.011929987011e-04, {   1,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -4.700634873014e-04, -3.023261604031e-04, {   1,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.047935392809e-04, -3.661838166169e-04, {   2,   0,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -7.373193190276e-05,  6.958446735703e-04, {   2,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -6.370832967077e-04, -1.299440730869e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.717929706944e-04, -4.942505598727e-04, {   1,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  4.657996674812e-04,  2.996943190522e-04, {   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.104300999960e-04,  5.410157263493e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -5.963470439808e-04,  1.661889267650e-04, {   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.379508240686e-08, -7.599999998748e-04, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.974674923167e-07, -7.553564510462e-04, {   1,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.001255105894e-05,  7.229349702339e-04, {   0,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.578951571523e-04, -2.948569676739e-04, {   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.461650893383e-04,  2.053252518221e-04, {   2,   0,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.502923589470e-04, -6.009324899300e-04, {   3,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.146621916766e-04, -2.356033034303e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  1.106048945092e-06,  7.491366214855e-04, {   0,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.243547091993e-08,  7.499999998969e-04, {   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.086467459320e-08,  7.499999999213e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.120178413907e-04,  5.373538494817e-04, {   2,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.006012505705e-05,  7.384254352278e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.859792299986e-05,  7.297630555599e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.310192962135e-04, -6.139518438919e-04, {   1,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -5.824049568276e-04,  1.621528207913e-04, {   2,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -5.821518487584e-04, -1.620938229085e-04, {   2,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -2.969924209174e-07,  7.436699929456e-04, {   0,   0,   3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.307238836315e-04,  6.131168160147e-04, {   1,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -7.146228896722e-04, -2.837478530788e-05, {   2,   0,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.013262727328e-04, -2.394384292371e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -7.116988960697e-04, -2.854874881116e-05, {   4,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.535072351181e-04, -8.415088710571e-05, {   2,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.043708401071e-04,  6.330702370560e-04, {   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.613860839608e-04, -1.737316785680e-04, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.468728246523e-04,  5.874734522326e-04, {   1,  -1,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.578155641605e-04, -3.757337684671e-04, {   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.409970039091e-04,  2.924379402069e-04, {   0,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -5.410459534734e-04, -1.921017566059e-04, {   0,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -5.407041626183e-04,  1.919831816595e-04, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -3.732321619897e-04,  3.590278670947e-04, {   2,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.086673117993e-06,  7.309513583204e-04, {   0,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.256444494449e-06, -7.287826229698e-04, {   2,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.644771711852e-04,  5.658153250521e-04, {   1,   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.661442480272e-04,  4.640208348396e-04, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  8.791223442158e-09,  7.299999999471e-04, {   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.118725137291e-05,  6.672220269382e-04, {   2,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  5.663270712218e-04,  1.612032919379e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -3.718969816542e-04,  3.555899039694e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -5.507184740361e-04, -1.749648514604e-04, {   2,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.824812479305e-04,  3.392377564159e-04, {   0,   0,   1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.280176170064e-05, -7.073916761076e-04, {   2,   0,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  1.286339939798e-05, -7.061947162375e-04, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  5.834407362677e-04,  1.340215746059e-04, {   1,  -1,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.097625576486e-06,  7.155438421988e-04, {   2,   0,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.573511615315e-04, -5.907157043082e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.637071945163e-04, -4.525422444845e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -3.036330535299e-04,  4.122352688620e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  7.280034452035e-07, -7.139614965197e-04, {   1,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.064741174351e-07, -7.134365438230e-04, {   2,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.208950234752e-04,  2.924463218644e-04, {   2,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  3.225636371604e-04, -3.878681457445e-04, {   2,   0,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -5.355383559877e-04,  1.743596820110e-04, {   2,   0,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.248440059017e-04,  5.848026820895e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -3.721954752818e-04, -3.301655049982e-04, {   4,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.597957558235e-04, -1.424012514842e-04, {   0,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.597535112932e-04,  1.424078306657e-04, {   0,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.247778861075e-05, -6.089916357423e-04, {   2,   0,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.486504559160e-04,  4.525881317433e-04, {   3,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.486504559160e-04, -4.525881317433e-04, {   1,   0,  -1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  1.503575709986e-06,  6.992806032086e-04, {   2,   0,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.320592374828e-04, -1.682846258103e-04, {   2,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  3.935295485751e-04,  3.063053742670e-04, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  5.676563907111e-04,  1.301034793143e-04, {   2,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.609416117990e-04, -1.357336910706e-04, {   1,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.640716188525e-04, -3.228983507354e-05, {   2,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  7.042069890043e-06,  6.888507511523e-04, {   1,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.977116877071e-06, -6.869966404973e-04, {   2,   0,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.405349377731e-04, -1.533802831853e-04, {   2,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.398619084326e-04,  1.531343201998e-04, {   2,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.496034065726e-07,  6.916814697306e-04, {   2,  -2,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.320499881656e-04, -5.594651190292e-04, {   2,   0,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  3.837838379459e-04,  3.066238885659e-04, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  1.300619358511e-08,  6.899999998774e-04, {   2,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.681144431631e-05, -6.023958318880e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -2.144710745056e-05, -6.656646908886e-04, {   2,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.307839799367e-04, -2.561872044394e-04, {   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.287362034260e-04, -4.579777383566e-04, {   2,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.910505295007e-04, -3.954688522417e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  2.443364840814e-04, -4.416295698455e-04, {   2,   0,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  3.082692317393e-04, -3.737162425359e-04, {   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.537370132133e-04, -5.281030777970e-04, {   1,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.533568031997e-04,  5.269011723109e-04, {   1,  -1,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.757895566023e-08, -6.800000038628e-04, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.092149852425e-04, -5.704049982112e-04, {   1,   1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.385693174865e-04,  1.397117317293e-04, {   0,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -5.385065288611e-04, -1.396867303249e-04, {   0,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.105821902963e-06,  6.760504607802e-04, {   0,   0,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.062244270158e-04, -5.701143841935e-04, {   2,   0,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  8.820961133631e-06, -6.654314850894e-04, {   2,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.222205165176e-05,  6.608588739092e-04, {   2,   0,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -2.856969160970e-04,  3.844406263466e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -3.194124351780e-04,  3.498956636369e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -4.821156303471e-04,  1.862215553969e-04, {   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.042420715353e-04, -4.637925364651e-04, {   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.447641759078e-04, -3.231917457493e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -9.048305450191e-05,  5.757426122805e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -5.151057065879e-04, -1.481388931836e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.563602918070e-04, -4.058726961985e-04, {   2,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.900998500660e-07,  6.600001496115e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.333700459351e-04, -1.262995205050e-04, {   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.186875704832e-05,  6.474487772230e-04, {   2,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.332025358630e-04, -5.247382023721e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -5.319797966663e-04, -1.257397049263e-04, {   1,   1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.644554685375e-06,  6.558689877258e-04, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  2.556314726460e-05,  6.307969259076e-04, {   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  3.640583737446e-04, -2.889491453381e-04, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.968732119100e-05, -6.324652223173e-04, {   2,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -4.028584816294e-06, -6.476268476576e-04, {   0,   0,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.683583900115e-04, -2.830806903303e-04, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  5.343582962217e-04,  1.157608765556e-04, {   0,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.432771496617e-04,  4.064615599978e-04, {   0,   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  5.516343093196e-04, -9.809453745556e-05, {   2,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.007582705165e-04, -1.480503186144e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -4.501945229563e-04,  1.951166823211e-04, {   2,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  1.838025637157e-04, -4.614377240466e-04, {   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.521670006101e-04,  2.920460213222e-04, {   0,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.433359146112e-04,  1.008384254543e-04, {   1,  -1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -4.277570492891e-04, -2.163019532260e-04, {   2,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  6.101376141208e-05,  5.828814626617e-04, {   0,   0,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.413485764924e-04,  3.023221316678e-04, {   0,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.292553276149e-04,  1.395076909161e-05, {   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.740262353520e-07, -6.424558080445e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -3.619522076690e-07,  6.423622205251e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 { -6.194846480612e-04, -2.130103342374e-05, {   0,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.675871572358e-04, -7.319587230764e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.519134580767e-08, -6.399999998197e-04, {   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.462516273625e-04, -1.932658961029e-04, {   2,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  3.878909104082e-04,  2.470794491590e-04, {   2,   0,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -5.649429745052e-04,  7.000602680458e-05, {   2,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.087921955516e-04,  2.570307423497e-05, {   2,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.245801933897e-04, -4.090265033255e-04, {   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.955747093059e-04, -1.364387104515e-04, {   0,   0,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.366496651436e-04, -3.948202534419e-04, {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.565607904428e-06,  6.244669650671e-04, {   2,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.618141200951e-09, -6.299999999831e-04, {   4,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.392247414455e-04,  1.883324075545e-04, {   2,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.090675651699e-04, -4.176552544845e-04, {   2,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  5.046531553606e-04, -1.204308128053e-04, {   1,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.254362117089e-04, -3.965796213942e-04, {   2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.039782212137e-04, -1.797665226354e-05, {   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.489093227893e-06, -6.199976410483e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.018335923877e-04,  1.193470852930e-04, {   1,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.135735552754e-06,  6.156549024614e-04, {   1,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.561387423181e-05, -5.940529214464e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -4.336529970138e-04, -1.858460546646e-04, {   2,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  2.645145950987e-04,  3.538906402274e-04, {   1,  -1,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  2.936332946313e-04,  3.228496545222e-04, {   2,   0,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -6.137934524547e-04, -2.519459349325e-06, {   3,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.042377936089e-04, -4.073188672221e-04, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.045533731526e-04, -4.062831019698e-04, {   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.281757130808e-04,  1.811530474668e-04, {   2,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.807531942722e-06, -6.072491959152e-04, {   1,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.314518253455e-04,  1.760391352259e-04, {   4,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -5.893544678657e-04,  1.782511522568e-05, {   4,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.485174933633e-07,  6.058381992088e-04, {   0,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.612389746966e-04,  4.223431608798e-05, {   2,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.531791969433e-06, -5.969136677745e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  7.164023976711e-05, -5.313875096098e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -4.861734461595e-04,  1.152006156312e-04, {   1,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.159899685800e-04, -3.845300230375e-04, {   2,   0,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.758123127178e-07, -5.989938140476e-04, {   2,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.990017227135e-06, -5.974169275463e-04, {   0,   0,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.675987254602e-06,  5.976536081703e-04, {   2,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.548973328566e-05, -5.595806339013e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  4.485864375450e-04,  1.434007600108e-04, {   2,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.561891288477e-04,  2.357934630224e-04, {   1,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.761705449595e-04,  1.152202628081e-04, {   0,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.969368132355e-04, -3.943056923417e-04, {   1,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.163435814062e-06,  5.899024341547e-04, {   2,   0,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  6.051004122207e-08, -5.900076889101e-04, {   2,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.384414403941e-09, -5.899999999655e-04, {   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.075491125965e-09, -5.899999999687e-04, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  8.609653776010e-07, -5.874995006972e-04, {   2,   0,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.385784111671e-04,  2.488064683357e-04, {   4,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.669590411365e-05, -4.882913889650e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.765890987130e-04,  2.082270729275e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -3.756522808518e-04, -2.082802549030e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -6.789829075600e-05,  5.134704228042e-04, {   1,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.947928195978e-04,  1.845516058477e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -5.601016864268e-04, -1.728679310794e-05, {   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.162665229154e-04,  2.607106404127e-04, {   4,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.829370350063e-07,  5.766680417268e-04, {   1,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.936518545649e-06,  5.665005452218e-04, {   2,   0,  -1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  4.768427617173e-04,  9.884726994216e-05, {   2,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  4.577656720605e-04,  1.171630673001e-04, {   2,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.162272978042e-04, -4.578375390138e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.908845370624e-04,  3.822802654027e-04, {   1,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.684937281739e-07, -5.726536840034e-04, {   2,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.504867018769e-04, -2.167404084853e-05, {   2,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.314024731615e-04,  4.400606517884e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.114218194402e-08,  5.699999998911e-04, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -2.338519792681e-04,  3.351541265704e-04, {   2,   0,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -2.076740725367e-04,  3.596819684178e-04, {   0,   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -2.019759270131e-04, -3.646981819615e-04, {   0,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.106268408543e-04,  2.558530776628e-04, {   0,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -8.521397890893e-05,  4.807916694682e-04, {   2,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.883678347733e-04, -3.765709590659e-04, {   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.858902450141e-04, -3.785012166656e-04, {   2,   0,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.249893041077e-04, -3.914006908450e-05, {   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.929356656394e-04, -3.704711550928e-04, {   2,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.662225672366e-05,  5.461665001278e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.385402474084e-06, -5.600948505282e-04, {   2,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.018171677974e-04,  1.593189429127e-04, {   2,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -7.428121296406e-09, -5.599999999507e-04, {   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.860722525053e-04,  3.724516165996e-04, {   2,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.556319195702e-04, -4.008474842267e-04, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.893945690945e-05,  5.363954567884e-04, {   4,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.152709760001e-06, -5.533844435397e-04, {   2,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.521698430369e-05, -4.983393814711e-04, {   2,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.946350794428e-04,  1.584910527127e-04, {   0,   0,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.502874742350e-07, -5.517396363558e-04, {   0,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.883529791256e-04,  1.641065359240e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.836453024960e-04, -3.684142347278e-04, {   0,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.970402468531e-04,  3.540143309408e-04, {   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.794018988568e-04,  1.654031760511e-04, {   0,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  9.516519520516e-05,  4.488182755504e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  1.609279064793e-04, -3.823310856069e-04, {   0,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.218549401376e-04, -2.074418280735e-05, {   2,   0,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.385951645078e-06, -5.376461856213e-04, {   2,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.609412802182e-05,  4.544308360789e-04, {   0,   0,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.592377607336e-04, -1.810826364560e-04, {   2,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -4.402451022075e-04, -9.938850631438e-05, {   4,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.162794657059e-05, -4.961390509076e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  5.672833516986e-05, -4.807064876812e-04, {   1,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.132256018361e-04,  4.241857476394e-04, {   0,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.832215477256e-06,  5.310687554657e-04, {   0,   0,   1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -1.832215477256e-06, -5.310687554657e-04, {   0,   0,   1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -1.952999080812e-04,  3.362664726760e-04, {   2,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  4.309679335326e-04, -9.986611414904e-05, {   1,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.955416108898e-04, -3.349587785241e-04, {   0,   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  8.165417349404e-05, -4.481236983979e-04, {   4,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.802723665633e-04,  1.478174161203e-04, {   2,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.416257229911e-05, -5.131690209701e-04, {   2,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.208871405975e-06,  5.194154884880e-04, {   2,   0,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -3.840327932158e-04,  1.399502367558e-04, {   2,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -5.910886233033e-05,  4.645062776886e-04, {   1,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.544932257110e-05,  4.481114825561e-04, {   0,   0,   1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  7.541551386957e-05, -4.480314084893e-04, {   0,   0,   1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -3.798577898037e-04,  1.431360772326e-04, {   1,  -1,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.753978227523e-04,  2.461558734621e-04, {   2,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.217278014544e-04,  9.965064895379e-05, {   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.524944639263e-04, -2.684758446801e-04, {   2,   0,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.660538839947e-04, -3.536547003660e-04, {   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -4.837398159030e-06, -5.130810146683e-04, {   2,   0,  -4,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.801437098492e-04,  3.655498288890e-05, {   0,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.717656080266e-04, -3.448192937985e-04, {   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.896741327706e-04,  2.682213406086e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -4.674359353883e-04, -4.796221691200e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.853895403917e-04, -1.295717328966e-04, {   0,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.174828382187e-04,  9.655331832083e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  1.693335381876e-04,  3.438278896842e-04, {   2,   0,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.103202936760e-07,  5.116552855079e-04, {   0,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.224350801867e-04, -8.936319206973e-05, {   2,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  4.115032162766e-04, -1.002862363937e-04, {   1,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.134240001607e-07, -5.104331180306e-04, {   1,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.494864918369e-04, -3.610036752982e-04, {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.460771510310e-07, -5.099188522801e-04, {   2,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.941843186797e-04, -1.581694728941e-05, {   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.541124178248e-04,  3.554412193597e-04, {   0,   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.737740273105e-05,  4.814689837927e-04, {   2,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  4.111472111750e-04,  9.746087271063e-05, {   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.026495117632e-04, -3.054129319706e-04, {   2,   0,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -4.061303894101e-04,  1.018002497615e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.083199825297e-04,  9.880080972584e-05, {   4,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.678978161287e-04,  3.391417860503e-04, {   4,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.187710274015e-04, -8.806972344023e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.781113031370e-04,  3.286928715928e-04, {   2,   0,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.238262499747e-06, -5.008455632320e-04, {   3,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.019703983376e-04,  1.034260740561e-04, {   1,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -4.260772551912e-04,  7.930849954255e-05, {   3,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -4.196572121418e-04, -8.553033069907e-05, {   4,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.031175987626e-04,  1.013733428246e-04, {   1,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.100099215628e-04, -1.939750729648e-04, {   4,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.050583327793e-04,  3.988528265806e-04, {   2,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  4.832296880821e-04,  2.031888584480e-05, {   2,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.873766944668e-04,  1.601094415260e-05, {   4,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.087574114644e-04, -9.403612514333e-05, {   1,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.853973120664e-04, -3.171291135816e-04, {   2,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.309913271508e-04,  1.711603208933e-04, {   1,  -1,   1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.309913271508e-04, -1.711603208933e-04, {   1,  -1,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.789535775742e-04, -3.224150048798e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  7.063364110514e-05,  4.307068648517e-04, {   2,   0,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  7.040554081349e-05, -4.301586192339e-04, {   2,   0,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -3.077230782927e-04,  1.926178441342e-04, {   4,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.446939899728e-04, -5.524697146904e-05, {   2,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.436380959030e-07,  4.988671235188e-04, {   2,   0,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.581545152505e-04,  4.116357802175e-05, {   1,  -1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.000181943268e-06, -4.960124307675e-04, {   0,   0,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.491096705941e-04, -2.478313822261e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -3.925513645401e-04, -1.037612015010e-04, {   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.198718421444e-06,  4.945530151079e-04, {   2,   0,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.520180744865e-04, -1.430176291674e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -4.762786211840e-04,  1.846050672793e-05, {   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.283752297274e-04, -2.652515769319e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -4.252436443371e-04,  6.830029423863e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -4.916123560358e-05, -4.439318540556e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  2.862480786051e-04, -2.062806030122e-04, {   2,   0,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  7.269500241298e-07, -4.899992234113e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.966108563125e-04, -9.407678583313e-05, {   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.494489429284e-07, -4.893515240567e-04, {   0,   0,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.526386188915e-04, -1.374919856527e-04, {   4,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.846665894733e-04, -5.402442720390e-06, {   2,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.900000000000e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  3.101429432996e-06, -4.868357435060e-04, {   1,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.391367356150e-04,  1.507636297087e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.734876702829e-04,  3.158796705457e-04, {   2,   0,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -7.279712636267e-07,  4.882709321446e-04, {   2,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.017957084042e-04, -3.871467254579e-04, {   2,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  3.300996356827e-07, -4.877248384974e-04, {   2,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.084693731560e-06,  4.819447865024e-04, {   3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -4.694982273179e-04, -1.650335484378e-05, {   0,   0,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.192110025713e-04, -6.659309427432e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  9.612978283567e-07,  4.835219865873e-04, {   0,   0,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.647328241499e-04,  1.970641231583e-05, {   2,   0,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  3.734320225310e-04,  1.100810028698e-04, {   0,   0,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.597656078633e-04, -3.224002930554e-04, {   0,   0,   1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.429290854527e-04, -3.871911171532e-05, {   2,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.051598591920e-04, -3.755130148497e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -1.858937863816e-04,  2.943586660612e-04, {   2,   0,   0,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  2.646759367406e-04, -2.153638772083e-04, {   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -3.874159329072e-04,  9.233564037166e-05, {   1,  -1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.091743268825e-04, -7.054204647449e-05, {   2,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.280059839017e-05,  4.560923084457e-04, {   2,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  2.280059839017e-05, -4.560923084457e-04, {   2,   0,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -3.312129427659e-04, -1.472404479311e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.753800824549e-04, -2.028161315724e-04, {   1,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.674096197415e-04,  1.079453343784e-04, {   2,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.603212686685e-06, -4.697493323194e-04, {   2,   0,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -3.841308439966e-04, -9.101190429726e-05, {   1,  -1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.658524329926e-04, -1.090971223878e-04, {   0,   0,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.669867422655e-04, -1.077970286537e-04, {   2,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.354618589206e-04, -3.819205260802e-05, {   2,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.826365909728e-04, -9.074936020056e-05, {   1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.692466583173e-05,  4.063695672078e-04, {   2,   0,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  1.239624858151e-04, -3.486162677078e-04, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.288250747184e-07,  4.719386678971e-04, {   4,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.930872628842e-05, -3.925794253474e-04, {   4,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.661258309455e-05, -4.051910506753e-04, {   2,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  1.208365992658e-06, -4.695230413825e-04, {   0,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.296081502370e-08,  4.699999998213e-04, {   2,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.679596030685e-09, -4.699999999003e-04, {   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -9.682677602092e-07,  4.690180439047e-04, {   1,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.691825298668e-06, -4.642460182460e-04, {   4,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.456326185766e-05,  4.551203981030e-04, {   2,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.234268240751e-04, -4.466048568980e-05, {   4,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.734116112041e-04,  2.942950225231e-04, {   2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.215649881000e-05, -3.953757968870e-04, {   4,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -2.550442556850e-04,  2.120045782987e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.688790302981e-04, -9.686816353642e-05, {   2,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  5.529345016196e-05, -4.101043970692e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  5.153285474834e-07, -4.642606432755e-04, {   1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.522022571595e-04,  1.119870501949e-04, {   2,   0,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.018866429355e-04, -2.618109701491e-04, {   4,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.492089411151e-04, -3.139619026637e-04, {   2,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -2.682310106462e-04, -1.948986200425e-04, {   0,   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -5.520300366599e-05, -4.076433644920e-04, {   2,   0,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.675358995956e-04,  2.935748036894e-04, {   2,   0,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.724552885889e-04,  8.826294605274e-05, {   1,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.534965706032e-04, -3.070691849385e-04, {   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.010789970506e-07, -4.599999862499e-04, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.293845761297e-04,  3.071105224015e-05, {   2,   0,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.293713580187e-04,  2.305879647855e-04, {   2,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.977495453829e-06, -4.547084756650e-04, {   2,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.053964587060e-07, -4.593084757299e-04, {   2,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -3.131014617156e-04,  1.462979725178e-04, {   0,   0,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.708299064715e-04,  2.864200388752e-04, {   2,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.133725226045e-04,  4.355885136870e-05, {   0,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.129083108167e-05, -3.650187993691e-04, {   3,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.697717462184e-07, -4.556903645639e-04, {   0,   0,   3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.601807921730e-04,  1.953086714089e-04, {   4,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.645726329216e-04, -2.907166657133e-04, {   4,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  3.665251682484e-04,  8.869516998981e-05, {   0,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.193039626707e-06,  4.506862383426e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -1.004708863581e-04, -3.503178569570e-04, {   0,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.199114857532e-09,  4.499999999946e-04, {   4,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.932693631001e-05,  4.294095954418e-04, {   4,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.025836635630e-05,  4.272549314742e-04, {   2,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.921598108890e-04, -1.551712055640e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  5.771143708339e-07,  4.464709486848e-04, {   4,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.732567263663e-04,  2.732748104722e-04, {   2,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.600962085808e-04, -8.602075737220e-05, {   2,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.502981594288e-04,  2.934934994532e-04, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.019441221097e-04,  3.417073109095e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -5.436428630673e-05,  3.889257665115e-04, {   0,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.500640250431e-04, -2.932151747324e-04, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -3.591486856718e-04, -8.407197388406e-05, {   2,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.310795586348e-04,  2.121141353515e-04, {   2,   0,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  1.548300651408e-04,  2.878872681216e-04, {   4,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  2.763927904902e-04,  1.662193636124e-04, {   1,  -1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.395062980742e-04, -2.014255154840e-04, {   1,  -1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.605039797402e-04, -2.804087921904e-04, {   2,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.610390719372e-04,  1.793297256057e-04, {   0,   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.609340893138e-09, -4.399999999852e-04, {   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.532546405810e-09, -4.399999999858e-04, {   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.230345519350e-07,  4.390789839858e-04, {   2,  -2,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.257151473010e-04, -1.391443297342e-05, {   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.365401311427e-04,  1.581289744880e-06, {   4,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.530769300503e-04, -8.467527328518e-05, {   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.239507136457e-04,  1.338174297831e-05, {   4,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.315256404367e-04, -2.053573888493e-04, {   4,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.045701574519e-04,  2.322624627309e-04, {   2,   0,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -2.728097365905e-04, -1.639125812782e-04, {   1,  -1,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.606024630328e-04, -2.749096882489e-04, {   2,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -3.525578741931e-04, -8.293717593875e-05, {   0,   0,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.392912923019e-05, -4.104798212832e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -2.396790128959e-04, -1.924240414633e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  7.251383770665e-05,  3.595837416179e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -1.429816899491e-04, -2.885366682944e-04, {   2,   0,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.164259848203e-05, -3.498252031446e-04, {   2,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  8.164628053844e-05,  3.497594740116e-04, {   2,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 { -1.936737295068e-04,  2.369902980617e-04, {   4,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.672490184738e-06,  4.268274775473e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.488783255408e-06, -4.282242225272e-04, {   2,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.464363395643e-04,  8.294873284719e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  3.472155194826e-04,  8.215660303572e-05, {   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.373944834510e-05, -3.754640959991e-04, {   1,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.268168927996e-04, -2.019993983860e-04, {   2,   0,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  4.036256704907e-04, -2.403573889833e-05, {   2,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.828430719270e-05, -3.491225800465e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -9.605841371662e-05, -3.309503915727e-04, {   1,   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.438770405835e-04, -8.005563448661e-05, {   1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.945444979687e-04, -1.293774635710e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  9.985292869089e-06,  4.131431354166e-04, {   0,   0,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.293175237184e-04,  1.934713562896e-04, {   2,   0,   0,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 {  9.570705568820e-05,  3.262695925297e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  3.833511603162e-05,  3.827579567527e-04, {   2,   0,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  5.856014519633e-07, -4.204198316793e-04, {   2,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.386835946249e-04,  8.191284607799e-05, {   1,   1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.238834702883e-08, -4.199999998173e-04, {   2,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.984660343494e-09, -4.199999999704e-04, {   2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.163797656082e-07,  4.186468811546e-04, {   0,   0,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.387372595075e-04,  8.025438123507e-05, {   1,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.737514109463e-07,  4.176652274365e-04, {   4,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.567829594063e-04, -2.614961812906e-04, {   2,   0,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.322097786984e-04, -8.574730199694e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.025329458644e-04,  1.518143281067e-05, {   0,   0,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -8.602815709306e-05,  3.314104981099e-04, {   3,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  3.313910196730e-04, -8.535728488153e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.371452150412e-04, -7.956766628042e-05, {   1,   1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.537737348062e-04, -2.629271629431e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  2.725873790118e-05, -3.887919404758e-04, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  7.825655761850e-05, -3.373863068797e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  5.465444258252e-05,  3.607972856255e-04, {   2,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.581133192190e-07, -4.145392367918e-04, {   2,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.065384738781e-05, -3.337058804941e-04, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  3.922701856364e-04, -2.087015566891e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -3.322070814198e-04,  8.069809768160e-05, {   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.900932973740e-07, -4.121022760810e-04, {   2,   0,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.447429047901e-06,  4.090721889345e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.369489797178e-04, -2.751141362520e-04, {   0,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.491934437957e-04, -6.220311854209e-05, {   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.131920254042e-06,  4.088776297009e-04, {   0,   0,   3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.365333778736e-04, -2.714500032856e-04, {   0,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.350669059628e-05,  3.843959522308e-04, {   0,   0,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  3.294527689630e-04, -7.728811673530e-05, {   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.455009781410e-04, -2.603010498352e-04, {   0,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.442536362831e-04, -2.615292866775e-04, {   3,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.328278480193e-04, -2.728030469975e-04, {   2,   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  8.189140186018e-05,  3.235032999830e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  4.636135529103e-05,  3.587507304672e-04, {   0,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.434682194878e-04,  2.615392274098e-04, {   2,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.953615794159e-08,  4.042887730385e-04, {   0,   0,   3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.748627358616e-07,  4.032998497329e-04, {   2,   0,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.406994413702e-04,  6.321910966977e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.610314482290e-05, -3.566409972126e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -9.435874561361e-05,  3.081895877611e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -5.381077505951e-06, -3.970625996738e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -5.022591567293e-05, -3.516645041704e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  3.266127236014e-04,  7.507543945897e-05, {   4,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.685155999112e-04, -3.310970585954e-05, {   1,  -1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.508168963144e-04, -2.507311014009e-04, {   0,   0,   2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.507107681951e-04,  2.505610988876e-04, {   0,   0,   2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.335810178387e-04, -2.674558256446e-04, {   2,  -2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  6.003932626656e-09,  3.999999999549e-04, {   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.802960800853e-04,  1.928115056406e-05, {   2,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.352178214365e-04, -1.642850534639e-04, {   2,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.764545857849e-04, -2.253941547632e-05, {   2,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  2.904505487729e-04,  1.083442601966e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.220519372891e-04, -7.668707330875e-05, {   1,  -1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.437023199692e-04, -1.543563051544e-04, {   4,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.826220130487e-04, -1.525841767235e-05, {   2,   0,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.908977366469e-05, -3.085409788280e-04, {   2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.756522998147e-04,  2.219419263566e-04, {   0,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  3.804626535552e-04,  1.601061061454e-05, {   2,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.406728492376e-07, -3.955669645831e-04, {   0,   0,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.186943788733e-07,  3.957945748305e-04, {   2,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.735378106376e-06, -3.940114966459e-04, {   2,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.172620449161e-04, -1.781683016548e-04, {   2,   0,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -3.276927550947e-04, -6.734847128389e-05, {   1,  -3,   1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  3.276927550947e-04,  6.734847128389e-05, {   1,  -3,  -1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  1.522150050711e-04, -2.425607543742e-04, {   2,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  2.419509494806e-04,  1.527614441748e-04, {   2,   0,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -5.345715690206e-06,  3.892419127700e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  2.677001808470e-04,  1.259075834437e-04, {   0,   0,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  5.585633235612e-05,  3.374809519186e-04, {   2,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.176396632025e-04, -7.543712268626e-05, {   1,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.768586842395e-04,  1.540665303709e-05, {   2,  -2,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.511919797051e-04,  2.410530435771e-04, {   2,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.345172269561e-07,  3.919892779761e-04, {   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.030740665890e-04, -1.888743209351e-04, {   2,   0,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  2.985627409212e-05,  3.613215478159e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -3.159078495148e-04,  7.520192494512e-05, {   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.605606968742e-04, -1.303577163246e-04, {   0,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.891974714499e-07, -3.893424473189e-04, {   2,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.304781459888e-05,  3.364310662075e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -2.016098940831e-04,  1.873152912488e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  5.482286406929e-05,  3.340712774835e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  3.359027828696e-04, -5.271233060782e-05, {   2,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  1.654055426479e-04,  2.217419635547e-04, {   1,  -1,   2,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  4.014883622066e-05,  3.466190200694e-04, {   2,   0,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  1.590200660471e-04,  2.275588531314e-04, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  4.586138746827e-05, -3.400631147841e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.503176810224e-04, -1.354874010943e-04, {   2,   0,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -3.089357873147e-04,  7.668712727362e-05, {   1,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.703386619458e-04, -1.518812000928e-05, {   2,  -2,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.140366451463e-07, -3.845099129994e-04, {   4,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.111517160146e-04,  1.739030127481e-04, {   2,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.240632333335e-06,  3.778118358119e-04, {   2,   0,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.577289023327e-04, -2.262632089386e-04, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  2.380995055863e-05, -3.575043055788e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  2.617347522982e-04, -1.195656671356e-04, {   0,   0,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.624189388443e-04,  2.185512459343e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  2.716649033789e-04,  1.092334388717e-04, {   2,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.284755920830e-07, -3.804790685169e-04, {   2,   0,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.485954714832e-04,  3.132406840439e-05, {   3,  -1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.705189245837e-04,  9.327819934560e-06, {   0,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.063323282775e-04,  7.324987505832e-05, {   1,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.497514032762e-05,  3.340298964782e-04, {   2,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.338662384927e-04,  1.449036181112e-04, {   2,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  3.051580880638e-04,  7.227271182517e-05, {   1,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.342334766130e-06,  3.728793720808e-04, {   0,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.668818402865e-06,  3.731957933254e-04, {   2,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.034105503897e-04, -2.733697582573e-04, {   0,   0,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.120194978992e-06, -3.710279067889e-04, {   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.905424876739e-07,  3.755553190000e-04, {   4,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.251607046786e-04,  2.505867281561e-04, {   2,  -2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.772842479850e-04,  9.844616599994e-05, {   2,   0,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -5.514987212716e-05, -3.205804762593e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -2.772403953436e-04, -9.843678782232e-05, {   2,   0,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -3.632344204396e-04, -1.226270572221e-05, {   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.488106455042e-04,  2.624920939694e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  5.963609262348e-05, -3.132519286884e-04, {   1,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.847093721019e-04,  8.816041661497e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.595332478614e-04, -1.314103562126e-05, {   2,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.485053781097e-04,  2.238903956061e-04, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -1.282754878821e-04, -2.437084074879e-04, {   2,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -8.827429975783e-05,  2.811359203468e-04, {   0,   0,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.061610372153e-04,  6.228918776317e-05, {   2,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.466291202440e-04, -2.217654581522e-04, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 { -4.669188208273e-05, -3.214386553610e-04, {   2,   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -4.190788672050e-05,  3.261726475233e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  3.603437926151e-04, -6.866494440163e-06, {   2,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.371485088660e-07, -3.662464584619e-04, {   2,   0,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.534733986429e-04, -1.264161398932e-05, {   0,   0,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.534022222494e-04,  1.126354438533e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.152105452205e-04,  5.076302482984e-05, {   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.996569207008e-05,  2.859852078494e-04, {   4,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.442485892166e-04,  2.204891173049e-04, {   4,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.443149356934e-04,  1.195774447121e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  6.515356200693e-06, -3.571904813510e-04, {   0,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.407052966080e-04,  2.229442058162e-04, {   2,   0,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 { -1.044871177522e-04, -2.588324477109e-04, {   1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.089606932699e-04, -1.541645122120e-04, {   0,   0,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 {  4.202723749230e-07, -3.626238614587e-04, {   2,   0,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  4.336192997857e-06,  3.586290628640e-04, {   2,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.237317154492e-04,  2.391772373854e-04, {   2,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.397090705657e-04,  1.231278376158e-04, {   0,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  5.491681425893e-05, -3.076843294592e-04, {   0,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.921034942557e-04,  6.917417423176e-05, {   1,  -1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.972741260831e-07,  3.607387933987e-04, {   2,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.892237386995e-06, -3.511312760016e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  2.912892117851e-04,  6.949482175733e-05, {   1,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.851035101599e-04,  7.567342803262e-05, {   0,   0,   3,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.203130112893e-04, -2.401302987381e-04, {   0,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.780530482406e-09, -3.599999999536e-04, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.032316768912e-05, -2.891302645052e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -2.264247963113e-04,  1.316181934192e-04, {   2,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  5.928968972450e-05, -2.987071980114e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.835898270280e-04,  1.743748502572e-04, {   0,   0,   2,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -8.200209478837e-05,  2.758856396434e-04, {   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.890472659716e-04, -6.860114065835e-05, {   1,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.368731127728e-04, -2.052858840826e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.504956375426e-05, -3.313065124038e-04, {   4,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -5.518908955041e-09, -3.560400505139e-04, {   2,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.537290634824e-04, -4.918142937844e-08, {   4,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.388084954904e-04,  1.492174312315e-05, {   4,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.548757990033e-07,  3.524298919684e-04, {   0,   0,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.848891474329e-04, -6.725480852187e-05, {   1,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.031072619254e-06,  3.509979310555e-04, {   2,   0,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.797939452028e-05,  3.337777273376e-04, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.486774759421e-04, -1.029299383615e-04, {   2,   0,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -8.257464546108e-05, -2.672746826766e-04, {   4,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.315441885019e-06, -3.470973077573e-04, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.770047007243e-04,  1.722640009412e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  1.621202614305e-04, -1.866073896782e-04, {   2,   0,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  2.138565178452e-04, -1.348504128460e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -9.917095648609e-06, -3.379246366541e-04, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.322379401755e-04,  1.153194805746e-04, {   2,   0,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -3.348745517454e-04,  1.216033121095e-05, {   2,   0,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.480662890318e-04, -1.987337686694e-04, {   3,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -2.117097829370e-04,  1.349837243240e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  3.064609144375e-04,  3.967795800665e-05, {   2,   0,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  1.206782372523e-04,  2.254127345892e-04, {   2,   0,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  5.362634323823e-05, -2.915151536331e-04, {   2,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -3.866107471844e-06, -3.409793679301e-04, {   2,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.723328125788e-05,  3.076091784508e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  7.864952117382e-05, -2.649010140752e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -3.288571059385e-05, -3.105333298384e-04, {   1,   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  6.706505456011e-05,  2.762513055102e-04, {   1,  -1,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  2.517874459659e-04,  9.120576852468e-05, {   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.149759675856e-04, -2.280013185805e-04, {   2,   0,  -1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.741321044059e-05,  2.851720408108e-04, {   1,   1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.199771786687e-04, -2.225526931060e-04, {   2,   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -2.322532723951e-04, -1.099593292664e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -6.834844984922e-05,  2.735126474792e-04, {   2,   0,  -2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.093005002386e-04, -1.325361377959e-04, {   4,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.488985922007e-04,  9.211212976672e-05, {   2,   0,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  2.279458714816e-04, -1.126164734794e-04, {   2,   0,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  2.754893605726e-04,  6.502018532078e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.711441729255e-04,  6.929308343315e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -7.239625736662e-09, -3.399999999229e-04, {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.657821543720e-08, -3.396308453861e-04, {   0,   2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  8.673584790533e-07, -3.384941434898e-04, {   3,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.737911159696e-04, -1.651954397923e-04, {   0,   0,   0,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -2.748726081447e-04,  6.359667291013e-05, {   1,  -1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.505145044051e-04, -1.875143007562e-04, {   4,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  5.278817422105e-06,  3.324669511407e-04, {   2,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.112468406752e-04, -2.263720894794e-04, {   0,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.165191455035e-04, -2.208347072481e-04, {   2,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  3.124082850668e-04,  2.450960995491e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -2.423479615979e-04,  9.455098786700e-05, {   2,   0,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -1.191772952804e-04,  2.172996675881e-04, {   1,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.107616854052e-04,  2.252003498309e-04, {   0,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -7.437578425953e-06, -3.282104705679e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  6.351354474678e-05,  2.720835715634e-04, {   2,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  6.345800801395e-05, -2.718333078172e-04, {   2,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  2.295335267771e-04, -1.053862200330e-04, {   1,   1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.042359502786e-04, -2.306027319444e-04, {   1,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.704436076894e-04,  6.401254895384e-05, {   1,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.254274198907e-06, -3.268980004165e-04, {   0,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.096316475056e-06,  3.260913531622e-04, {   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.209488147171e-04,  1.219031059259e-05, {   0,   0,   2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.108180919551e-04,  2.218794453269e-04, {   4,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.319968453479e-04,  9.991851381590e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  7.033481757581e-08,  3.315544424255e-04, {   1,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.464022227527e-05, -3.069513812318e-04, {   2,   0,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.558521143467e-05,  2.959355046264e-04, {   2,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.720987270943e-05, -2.741388102440e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  6.287689681790e-06, -3.246670996506e-04, {   2,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  1.601192717051e-04, -1.705534220284e-04, {   0,   0,   1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.395655219093e-04,  9.079411971433e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.492553907965e-06, -3.285625594425e-04, {   2,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -5.716717965291e-05,  2.724400436016e-04, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.913359452133e-04,  1.375903542661e-04, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.519836687773e-06,  3.270476899192e-04, {   3,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  2.637204754760e-04, -6.405233150954e-05, {   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.560621248699e-08,  3.272842521481e-04, {   3,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.086374652735e-04, -2.174955126989e-04, {   4,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.092522280732e-04,  2.167521745921e-04, {   2,   0,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.983406995856e-04, -1.275237389561e-04, {   4,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.742935808028e-04,  5.128663621084e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -2.740203056797e-04, -5.128225974573e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 {  2.001195244010e-04, -1.248444508055e-04, {   2,   0,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.235156561527e-04, -1.008229660612e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.546946168598e-04,  1.689243791689e-04, {   4,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  6.261363344421e-05, -2.606875551555e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  6.250685098240e-05,  2.605008369971e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 {  1.065386809408e-04,  2.163447329399e-04, {   2,   0,   0,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.960767489349e-04,  1.263180541885e-04, {   4,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.986146012134e-04, -2.358045929554e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  3.836982602248e-05,  2.834735796082e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  1.111291983500e-09,  3.217403988801e-04, {   2,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.125716435594e-04, -2.087769115717e-04, {   4,   0,  -3,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  2.587175941658e-04,  6.222487164546e-05, {   2,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.745606736751e-04, -1.461685317674e-04, {   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.218240980037e-06,  3.179092296763e-04, {   1,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.027846039097e-04,  1.730449422070e-05, {   2,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.211293968669e-05, -2.779664999762e-04, {   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.361651461911e-09,  3.199999999551e-04, {   2,   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.518583771955e-09,  3.199999999807e-04, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  2.031692262884e-04, -1.167117523201e-04, {   2,   0,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  2.968120165158e-06, -3.167029002399e-04, {   2,   0,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.617416550021e-05, -3.034935177280e-04, {   2,   0,   0,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -3.433052336312e-06,  3.161368355416e-04, {   0,   0,   1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -3.449636534387e-05, -2.845027786828e-04, {   2,   0,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.574682027853e-04, -6.117433819841e-05, {   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.080403998049e-04, -1.054051052736e-05, {   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.509706813330e-04, -1.675899871275e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.863747715692e-04,  1.320298060997e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.062887142241e-04, -2.120485244297e-04, {   0,   0,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.121027242433e-04, -6.087494254742e-06, {   1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.435409379312e-04,  7.387493324440e-05, {   2,  -2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.434118710584e-04, -7.383375670222e-05, {   2,  -2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.074230292632e-04,  2.094456909525e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  3.114253182286e-04, -5.439999907653e-06, {   2,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.551616117388e-05,  2.512932594009e-04, {   2,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.531210213500e-06, -3.131547107178e-04, {   0,   0,   1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  2.564821897962e-04, -6.000823735136e-05, {   1,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.065097854012e-04, -1.099038487222e-04, {   2,   0,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  5.715239083747e-05,  2.581406401053e-04, {   2,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.374407997764e-05,  2.413213204624e-04, {   0,   0,   2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.989433877086e-05, -2.250199448328e-04, {   0,   0,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.964586971587e-08,  3.148295518910e-04, {   2,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.458769937752e-04,  6.845197291847e-05, {   4,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -2.458616669767e-04, -6.846303650290e-05, {   4,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.282766564516e-07, -3.138168596264e-04, {   2,  -2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.478999917115e-04, -6.581280277370e-05, {   0,   0,   1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.108828909566e-05,  3.024810990898e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  2.425900861075e-07, -3.128034387534e-04, {   1,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.184447031452e-07, -3.120227514014e-04, {   4,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.040201073511e-04, -2.086835652646e-04, {   3,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.837899916030e-07,  3.121826112448e-04, {   0,   0,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.901882317184e-04,  2.216189173190e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -2.654332526278e-07, -3.117652935362e-04, {   0,   0,   1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.096536951671e-04, -2.020556824127e-04, {   0,   0,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  2.175792880498e-04, -9.366772884886e-05, {   2,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.757077027512e-06,  3.091520565286e-04, {   0,   0,   2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.506244307310e-04, -6.011914374063e-05, {   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.554411784090e-05,  2.151434868956e-04, {   2,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.115523082914e-04, -1.990937189567e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -1.222567538126e-04,  1.881217390925e-04, {   2,   0,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  4.852015290086e-08, -3.099999952916e-04, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.681046714958e-09,  3.099999999479e-04, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.971831376146e-04,  1.239732435026e-05, {   2,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.531227677596e-05, -2.441011898006e-04, {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -6.522942223537e-05,  2.434561262449e-04, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.492571940644e-04,  5.927028585769e-05, {   1,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.817989762763e-04, -1.265343694664e-04, {   0,   0,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.085688415669e-04,  1.994505652695e-04, {   2,   0,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.212805299443e-04, -1.866857074730e-04, {   2,   0,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.978342565500e-04, -9.204952915673e-06, {   2,   0,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.682397198111e-04,  1.379603910724e-04, {   4,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.201398270363e-07, -3.056658092943e-04, {   2,   0,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.792750831894e-04, -1.262970587734e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.830566630771e-04, -1.218876847804e-04, {   4,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.072263360846e-04, -1.975253399628e-04, {   0,   0,   1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  2.670845279323e-06,  3.020477254143e-04, {   2,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  9.458237408976e-05,  2.094027856982e-04, {   4,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  9.274561143372e-05, -2.110100286441e-04, {   2,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.614213039803e-04, -1.420160351891e-04, {   0,   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.006418547897e-04, -2.015091398559e-04, {   0,   2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.005434709979e-04,  2.013071369796e-04, {   3,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.911545660883e-04, -1.029507791637e-05, {   0,   0,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.435872690453e-04, -5.733601596033e-05, {   1,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.433917329967e-04, -5.751332383864e-05, {   1,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.527256522336e-05, -2.049483375531e-04, {   2,   0,  -1,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.408554367598e-09, -2.999999999903e-04, {   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.996030190891e-04,  9.993857082108e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 {  2.770280401329e-04,  2.245212761766e-05, {   0,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  6.544064661737e-07,  2.986644042208e-04, {   0,   0,   1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  9.871023315535e-05,  2.002252291924e-04, {   4,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.529310396446e-05, -2.435578808629e-04, {   2,  -2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.626002340918e-04,  1.359989171834e-04, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  4.861110606746e-05,  2.498479137771e-04, {   0,   0,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 {  1.418491220568e-05,  2.839190903199e-04, {   2,   0,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  2.331169136663e-04, -6.460021289536e-05, {   2,   0,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.061240538622e-04,  1.906527301088e-04, {   2,   0,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  3.851848056806e-05, -2.581375341537e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 {  1.365079484713e-05, -2.825619630100e-04, {   2,   0,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  9.836603640562e-05,  1.965592450131e-04, {   2,   0,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.515434331570e-07,  2.944728540023e-04, {   2,   0,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.709673554986e-04,  1.238232449681e-04, {   1,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.873814617504e-07,  2.943031057465e-04, {   1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.686012642735e-05,  2.675490965129e-04, {   4,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.053771372113e-04, -1.888187482826e-04, {   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.769300363824e-04,  1.707223521131e-05, {   2,   0,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  1.562123338460e-04, -1.373899193165e-04, {   2,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  9.014785748356e-05,  2.028962279335e-04, {   2,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.049022449865e-05,  2.625094028123e-04, {   2,   0,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  2.366108213013e-04,  5.595214156644e-05, {   1,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.813233209955e-04, -1.123287622664e-05, {   2,   0,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.579149328462e-04, -3.456687666183e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.361185200772e-04, -5.585273961092e-05, {   3,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.051709929642e-05, -2.812367080067e-04, {   2,   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -5.793675635579e-05,  2.334425916269e-04, {   1,  -1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -7.581329931193e-05,  2.149283234183e-04, {   0,   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -3.593632929898e-09, -2.899999999777e-04, {   2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.340560188225e-09,  2.899999999808e-04, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.815712885740e-04, -1.081975684041e-04, {   2,   0,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.579743175279e-06, -2.879196702726e-04, {   2,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.324966433771e-05,  2.556912689583e-04, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.230412367835e-04, -1.652143941410e-04, {   0,   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  2.674356673451e-04, -2.065222488347e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  2.322015116975e-04, -5.508618391829e-05, {   1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.151896472823e-04, -1.714397299045e-04, {   0,   0,   1,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 {  1.151690641529e-04,  1.714110337774e-04, {   0,   0,   1,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -2.316841163716e-04, -5.477587580323e-05, {   1,  -1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.085038179180e-04, -7.766677126398e-05, {   2,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.658470506118e-04, -1.202727009114e-04, {   2,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.508754066198e-07, -2.855112172418e-04, {   4,  -2,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.700313572134e-04,  1.519270973510e-05, {   2,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -2.382481264397e-04,  4.690215465816e-05, {   0,   0,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -1.328831536437e-04,  1.519855608615e-04, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.298605296540e-04, -5.476118270562e-05, {   1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.111659109108e-04,  7.300564785713e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.283054650231e-04, -5.577775019646e-05, {   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.296465786513e-04, -5.442345145118e-05, {   1,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.369693701784e-04, -1.467797164005e-04, {   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.746333626812e-05,  2.562379863717e-04, {   2,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.609406014742e-07, -2.826867333683e-04, {   0,   2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.644297959853e-05,  2.056896015507e-04, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  2.548720049273e-04, -2.706026960010e-05, {   4,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.278414383239e-04, -5.389527902504e-05, {   1,   1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.473798650469e-05,  2.369825240680e-04, {   1,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -9.998430522928e-05,  1.816453384769e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  9.407828160221e-05, -1.870514613834e-04, {   0,   0,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.759371040151e-06,  2.762467493651e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  2.631361297606e-04,  1.785034355102e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  7.418620107084e-05,  2.065725088637e-04, {   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.520834466693e-05, -2.152160741397e-04, {   0,   0,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.269280826562e-05, -1.874714964847e-04, {   2,   0,  -1,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.677836161714e-04, -1.122643647655e-04, {   2,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  5.913175505319e-09,  2.799999999376e-04, {   0,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.616550285776e-04, -1.775202726990e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -2.646883038636e-04, -1.471434548265e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  4.392768637101e-05,  2.343619674406e-04, {   3,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.414289934524e-04,  1.366489786252e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  1.090919571451e-04,  1.689589388676e-04, {   0,   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.489425102470e-05, -2.629616579355e-04, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -2.600879889575e-04, -1.767408660021e-05, {   0,   2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.251878133155e-05,  1.852361341707e-04, {   0,   2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.974865993169e-05, -2.576989730904e-04, {   2,   0,  -1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  1.177583888414e-04, -1.596443115011e-04, {   0,   0,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.432495444820e-04,  1.338418852191e-04, {   0,   0,   3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.761112072838e-04, -8.369117051280e-07, {   4,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.741857819117e-04,  1.018712797372e-04, {   0,   0,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -1.410695153574e-04, -1.345690181585e-04, {   0,   0,   3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -2.660450153004e-04, -9.549032199274e-06, {   0,   0,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.865765212892e-05, -2.367896592702e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -2.654859421917e-04, -9.288733208474e-06, {   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.018203085997e-04, -7.274608564204e-05, {   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.611222719007e-05, -1.781558729888e-04, {   2,   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  2.290527172653e-04, -4.521256867675e-05, {   2,   0,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -2.667825217841e-04,  7.164465398740e-06, {   4,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.630050531629e-04,  1.089870681282e-05, {   1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.067394630730e-05,  1.828315780710e-04, {   2,   0,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.614844677985e-04,  1.119813891617e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -1.420351519453e-04,  1.312830327123e-04, {   0,   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  7.100136569265e-05,  2.022953360076e-04, {   0,   0,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.293542649687e-04,  1.425340307682e-04, {   0,   0,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.474550634922e-05, -2.270823686050e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  9.554715843485e-05,  1.760002517890e-04, {   0,   0,   1,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 { -2.489964138592e-04, -2.237824580385e-05, {   1,  -1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.489972143767e-04,  2.236933688467e-05, {   1,   1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.966623242470e-05,  1.816717833875e-04, {   2,   0,   0,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.621663957809e-04,  8.810008746089e-06, {   2,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.568221457684e-04, -1.134304678374e-04, {   0,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  1.567577990906e-04,  1.133730829267e-04, {   0,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  6.314601233135e-09,  2.699999999262e-04, {   0,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.182468057307e-04, -5.175823248574e-05, {   2,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.712388968487e-11, -2.700000000000e-04, {   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -4.991074146114e-05, -2.199047740255e-04, {   4,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.490944630245e-04,  1.204064121098e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -2.194944276991e-04, -4.989918760937e-05, {   2,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.953365574220e-05, -2.197364322833e-04, {   1,  -1,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.171581416785e-04,  5.151636724399e-05, {   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.790113481706e-05,  2.303958312523e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -1.881384879698e-04, -7.974386098120e-05, {   0,   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -7.191985075313e-05, -1.955342548579e-04, {   2,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.609183058174e-04, -1.064509223323e-04, {   2,   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.459701322573e-04,  1.211183708433e-04, {   2,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.572759387487e-04, -9.683219930276e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.934006800294e-04,  7.337227340780e-05, {   0,   0,   2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.491862845626e-04, -1.174695009415e-04, {   0,   0,   1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  1.491746343272e-04,  1.174613199618e-04, {   0,   0,   1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  1.106209044748e-05, -2.551108456175e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  7.962142012079e-05, -1.863761858382e-04, {   2,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.802423305885e-04,  8.566240712147e-05, {   2,  -2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  2.704221057545e-05, -2.388356268526e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.128908211780e-04, -1.527769092482e-04, {   3,   1,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -2.945794181865e-06,  2.625569058651e-04, {   2,   0,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -2.136781974010e-04, -5.138928912770e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.130173336348e-04,  5.154441719577e-05, {   4,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.944634053299e-05,  2.250100535789e-04, {   2,   0,  -1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  1.187864970057e-04,  1.456049307179e-04, {   2,   0,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.634775623552e-04, -6.276448916567e-07, {   2,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.138709419708e-04,  5.018855985065e-05, {   2,   0,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -1.658227865808e-04,  9.823588093390e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 { -2.610993801657e-04,  2.625812373016e-06, {   2,   0,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.222968019182e-04, -4.123912583034e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.784075005612e-04, -8.475667094472e-05, {   2,  -2,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  8.625060539242e-06, -2.539762317101e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -5.582952301163e-05,  2.065830797508e-04, {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -9.404705483470e-06,  2.529059072039e-04, {   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -5.594005344538e-05, -2.062372576134e-04, {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  7.375356764305e-05, -1.884191061930e-04, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -2.121293628953e-04, -4.952270763957e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -2.080847269993e-04,  5.346820836714e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -6.876286567129e-05, -1.926407591253e-04, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.289191173187e-05, -2.484656908037e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -9.292155040793e-05,  1.682790500625e-04, {   4,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.655403287025e-04, -9.564345760525e-05, {   0,   0,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  6.684447823134e-05, -1.942443895008e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -2.099118049575e-04, -5.106022683683e-05, {   4,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.289091382726e-05,  2.378083633006e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  1.251133252999e-04,  1.353240527365e-04, {   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.670051408834e-05,  2.336328999009e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  8.988448262127e-05,  1.701795922121e-04, {   1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.902556207262e-09, -2.599999999707e-04, {   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.721041965145e-09,  2.599999999734e-04, {   4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00,  2.600000000000e-04, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.066844391081e-04, -5.309009098988e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  8.596836125073e-05,  1.737558001052e-04, {   2,   0,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.953599054967e-04, -6.433248994087e-05, {   2,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.568229272955e-05,  1.738144956705e-04, {   2,   0,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.547211843809e-05, -1.738282582109e-04, {   0,   0,   1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.392101159268e-04, -1.997480472096e-05, {   2,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.423973631315e-04,  1.165515034782e-04, {   2,   0,  -1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -3.181060258613e-07,  2.570602207410e-04, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  8.503755968632e-05,  1.722007746527e-04, {   2,   0,   0,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.579638454147e-04,  9.897848738152e-05, {   2,   0,  -1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  1.855098962584e-04, -7.123282264948e-05, {   0,   0,   2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.386274066667e-05,  2.322388822376e-04, {   0,   0,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  5.595332611695e-07,  2.554953580283e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  1.112653754798e-05,  2.448135864843e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  7.242900195042e-07,  2.549739939112e-04, {   0,   0,   1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.886273788289e-04,  6.705582713359e-05, {   2,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.148889620682e-04,  4.032434474618e-05, {   0,   2,  -2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.418005937051e-04,  1.312540781377e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.879414562563e-04, -6.685662278961e-05, {   2,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.425762294922e-04,  1.118700951743e-04, {   0,   0,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  2.058207155613e-04,  4.862543892862e-05, {   1,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.466229601417e-04,  1.075506298920e-04, {   2,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.316152623728e-04,  1.221304778374e-04, {   1,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.704371036646e-05,  2.165035945266e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  5.367292891943e-05,  1.993976506611e-04, {   2,   0,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -2.474672885389e-04,  4.912889777117e-06, {   2,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.477837855046e-07,  2.519412897741e-04, {   2,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.038339709756e-04,  4.811687501616e-05, {   1,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.826613473439e-06,  2.470893167651e-04, {   1,   1,   0,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.234014214778e-04, -2.846527874624e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  6.404394374211e-05, -1.877165709060e-04, {   2,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.215915474369e-04, -1.293361950357e-04, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  2.028515567268e-04, -4.777675850332e-05, {   1,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.411827939458e-05, -1.664993735069e-04, {   0,   0,   0,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 {  9.943892145621e-05, -1.507905504180e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -1.570272223446e-04, -9.313464298346e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -6.285803207195e-08,  2.499999905925e-04, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.448946460976e-04, -1.051140654455e-04, {   0,   0,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.134464013780e-09,  2.499999999974e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.500000000000e-04,  0.000000000000e+00, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.059446200191e-05, -2.093252290298e-04, {   2,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.386817768904e-05,  1.860303756465e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -1.017610471200e-04, -1.480929476787e-04, {   2,   0,   0,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  6.762186910913e-06,  2.429920262869e-04, {   3,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.780285002022e-06, -2.446977681765e-04, {   1,   1,  -2,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.638753381290e-05, -2.230672104307e-04, {   2,   0,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 { -1.515704202130e-04, -9.708968903261e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.868128256774e-04,  6.170783012389e-05, {   2,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.922602600092e-05, -2.191197917729e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -2.025691235223e-04, -4.560565975356e-05, {   4,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.008901680805e-04, -4.704513640285e-05, {   1,   1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.506674567188e-04, -9.724850967205e-05, {   2,   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -1.683580685420e-04,  7.932032371147e-05, {   2,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  2.001324065800e-04, -4.742190494155e-05, {   1,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.093396472040e-08, -2.474555684820e-04, {   4,  -2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.290482070984e-06,  2.437665559910e-04, {   2,   0,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.543422985203e-04,  9.267409238774e-05, {   4,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  5.109357609363e-05, -1.955363276708e-04, {   2,   0,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  1.441170596575e-04, -1.022861536280e-04, {   1,   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.427321647132e-04,  1.035234405359e-04, {   0,   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.676919300374e-04,  7.820180147720e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  2.280887180180e-06,  2.434746385090e-04, {   2,   0,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  2.452487366951e-04, -4.477948094816e-07, {   4,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.812582829571e-07, -2.446607415453e-04, {   2,  -2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.350135697668e-04,  1.100468939065e-04, {   2,   0,  -2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.179307592995e-07, -2.449406491584e-04, {   0,   0,   1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.669141787856e-04, -7.769950733631e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  2.168516882070e-07, -2.438243185244e-04, {   2,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.237920636751e-06, -2.423952360307e-04, {   2,  -2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.645510406175e-06,  2.358316539773e-04, {   2,   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.711530360194e-05, -2.255035289187e-04, {   0,   0,   2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.068363092426e-04, -3.569941413076e-05, {   3,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.068363092426e-04,  3.569941413076e-05, {   1,   0,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.955957766738e-04, -4.626400620445e-05, {   1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.891845446373e-04,  5.267449543005e-05, {   1,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.891465394298e-04, -5.266478125928e-05, {   1,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -9.251314513740e-05,  1.492638699780e-04, {   0,   0,   1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  1.783171257914e-04, -6.331302737918e-05, {   2,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  1.481388757492e-04, -9.348346779668e-05, {   1,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.497187800806e-05,  2.166136648747e-04, {   2,   0,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -2.163567739604e-04, -2.520182603596e-05, {   2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.782426213318e-04,  6.328918620396e-05, {   2,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.333239924916e-08, -2.406225664691e-04, {   1,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.572890715928e-08, -2.399999988647e-04, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.267256359658e-09,  2.399999999778e-04, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  8.231794831282e-05, -1.575260702517e-04, {   0,   0,   1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.279967535214e-04,  1.167492141796e-05, {   2,   0,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.938207500428e-06, -2.336261071575e-04, {   4,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -4.613058641019e-05, -1.933950312012e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  2.033569029744e-04, -3.615774803061e-05, {   2,   0,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.998057488687e-05,  2.193827496621e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.036400507847e-06, -2.378319221871e-04, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  2.191166688978e-04, -1.969480670490e-05, {   1,   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.091693980950e-05,  2.178012034060e-04, {   2,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.929543779331e-04,  4.574104995478e-05, {   1,  -1,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.132535978315e-05, -1.772918371851e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  7.265843115719e-05, -1.653556855722e-04, {   1,   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.291730274599e-04,  1.086364709538e-04, {   1,  -1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.903733206015e-05,  1.587158405546e-04, {   0,   0,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.004456511339e-04, -3.725939976334e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.621424627355e-04,  7.554326596360e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.325820471560e-04,  4.783479754622e-06, {   0,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.301682107134e-04,  7.012675582190e-06, {   2,   0,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.199803796551e-05, -2.250912852205e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -2.218559845743e-07,  2.368317244194e-04, {   2,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.270198072672e-04, -9.904816756473e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.226150063948e-09, -2.365242010853e-04, {   2,   0,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.922842735859e-04,  4.417333864915e-05, {   1,   1,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.112176554789e-04, -1.251541102125e-04, {   2,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.968279936712e-04,  3.904571025498e-05, {   2,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.480762652550e-04,  8.765925015787e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  1.069181039159e-04,  1.283664888263e-04, {   2,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.902932251017e-04,  4.493505909839e-05, {   1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.282466022974e-06, -2.269401031579e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -2.245251603344e-04, -1.048256087758e-05, {   4,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.027269688296e-05, -2.242790696267e-04, {   0,   0,   1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  1.905851109938e-04,  4.384931212420e-05, {   2,   0,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -2.028730123011e-05, -2.141409819968e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -1.823025831625e-04,  5.204870106744e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.893732848782e-04,  4.482151045942e-05, {   1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.269276629997e-07, -2.329987212586e-04, {   0,   2,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.619039161570e-05,  1.367488792020e-04, {   0,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.035794703407e-04, -1.291652666611e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  2.404104349008e-07, -2.324733035429e-04, {   2,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.240072523813e-05,  2.196502265275e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.239239885558e-04, -7.953175218670e-06, {   0,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.607193525601e-06, -2.291875914451e-04, {   2,   0,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  1.792612818559e-04,  5.157100624657e-05, {   2,  -2,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.791358126617e-04, -5.152292926502e-05, {   2,  -2,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -2.247125478821e-04, -5.581802026550e-06, {   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.954506249163e-04,  3.430543634814e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  3.999056954868e-06,  2.254538500341e-04, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  4.652843885372e-05, -1.827076351907e-04, {   2,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -1.128033822907e-04,  1.163184944409e-04, {   2,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.013318338609e-04,  1.272269448293e-04, {   2,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -1.430889245020e-04, -8.542633334202e-05, {   2,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.091570368533e-04,  1.879717890345e-05, {   1,  -1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.562370647368e-05, -1.423212564909e-04, {   0,   0,   1,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 {  4.632467720866e-06,  2.230169784734e-04, {   0,   2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.612922903485e-05,  2.113748464085e-04, {   2,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.695949506777e-05, -2.002851887954e-04, {   1,   1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.587274617560e-05, -2.010136823057e-04, {   0,   0,   1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.782657841710e-06,  2.240893463264e-04, {   2,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.848786278400e-05, -1.782795775694e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -2.223735800759e-05, -2.044755676293e-04, {   2,   0,  -1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.164513810836e-04,  1.011262730367e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.746282505104e-04, -5.150658755760e-05, {   2,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.516139160592e-06, -2.242278842542e-04, {   2,   0,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -5.120739823238e-05,  1.743759930301e-04, {   0,   0,   1,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -5.118770439135e-05, -1.743508467358e-04, {   0,   0,   1,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 { -1.820730524210e-04, -4.323612393438e-05, {   3,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.414277299084e-05, -1.510309722998e-04, {   2,   0,  -1,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.758317177048e-07, -2.244827051475e-04, {   2,   2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.427038867042e-06, -2.175233141285e-04, {   2,  -2,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.679578276086e-06, -2.230655549983e-04, {   3,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.105001858649e-04, -1.361313564590e-05, {   0,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  9.353508891219e-05,  1.301218388036e-04, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.622391991517e-08,  2.228741182270e-04, {   0,   0,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.209322844983e-04,  1.918547880879e-06, {   0,   0,   2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  7.459251010078e-05, -1.482388721570e-04, {   0,   0,   0,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.346058422825e-05, -1.493520417223e-04, {   2,   0,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.043269712667e-08,  2.226958047289e-04, {   4,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.212517936041e-04,  1.446302769731e-06, {   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.746274952473e-06, -2.199030163235e-04, {   1,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  7.890805240234e-05,  1.436275863503e-04, {   1,   0,   2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -7.890805240234e-05, -1.436275863503e-04, {   1,   0,  -2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  1.490928331085e-06, -2.208514466396e-04, {   2,   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 {  5.899226821738e-05, -1.629156174550e-04, {   2,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.183275976019e-04,  1.034141279991e-04, {   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  2.190928808452e-04,  2.341209859734e-06, {   2,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.213970430417e-04, -9.989303968702e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -7.721861986518e-05, -1.436855166039e-04, {   2,  -2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.331627881379e-04, -8.764248556503e-05, {   2,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  1.780079395887e-04, -4.259185371416e-05, {   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.775211389345e-04, -4.278327054205e-05, {   2,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.486752814831e-05,  1.753670010464e-04, {   4,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.725885209793e-04, -4.753200282509e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  1.725347932729e-04,  4.753860387182e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -5.106833390861e-09, -2.199999999407e-04, {   0,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.075712110244e-11,  2.199999999998e-04, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.268405009656e-08,  2.199205178345e-04, {   2,  -2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.952124654502e-04,  2.459023221147e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 { -4.538828281745e-05, -1.742438869915e-04, {   0,   0,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -7.114912728775e-05,  1.483388182843e-04, {   2,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.515004042844e-04, -6.798586592701e-05, {   2,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.342514297828e-04, -8.508627796861e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  8.805424710203e-05, -1.312148449789e-04, {   0,   0,   2,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 {  1.771238908771e-04,  4.179845403735e-05, {   1,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.751555623314e-06,  2.099791378458e-04, {   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.179974233179e-05,  2.069270615441e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  1.729087793232e-04,  4.571620490004e-05, {   1,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.898842492880e-04,  2.869663936515e-05, {   0,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.994226034095e-08, -2.183745512551e-04, {   2,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.760823328496e-04,  4.208010656444e-05, {   1,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.779060166792e-08,  2.180471496587e-04, {   3,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.307576636496e-05, -2.048994505548e-04, {   0,   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  1.204587568276e-04, -9.716017565619e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.750803516169e-04,  4.250664747361e-05, {   1,   1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.346069631973e-05, -1.839806197415e-04, {   2,   0,   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  1.944547272258e-04,  2.276399053125e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -3.797958700423e-05,  1.790717366548e-04, {   2,   0,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.036919938730e-05,  1.464706819677e-04, {   2,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.091804361842e-04,  6.974351236335e-06, {   2,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.211056965361e-08,  2.158286944687e-04, {   2,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.092081849267e-05,  1.544768169753e-04, {   2,   0,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.070085255046e-04, -8.166334609980e-06, {   2,   0,  -1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.361437227350e-08,  2.150706882407e-04, {   4,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.283555786541e-05,  1.818164961164e-04, {   0,   2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.735812272835e-04, -4.083057348149e-05, {   1,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.439409256981e-05,  1.399452438602e-04, {   4,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.728729788424e-04, -4.100730742740e-05, {   1,  -1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.519576749603e-05,  1.886272457881e-04, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -1.299378576025e-04, -8.385351465388e-05, {   2,   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  1.713627536671e-04, -4.217525128061e-05, {   1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.660939583411e-06,  2.038552766229e-04, {   4,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.907268861345e-04, -2.258182877624e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.196939654507e-05, -1.712816614875e-04, {   2,   0,  -2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.327587944435e-04,  7.997211637673e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  1.327239710015e-04, -7.995083390659e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -6.555471828061e-06,  2.059429703711e-04, {   0,   0,   1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.531124467343e-05,  1.971773177426e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -2.628698589577e-06,  2.097622983537e-04, {   1,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.045751948349e-04, -7.395471331912e-06, {   0,   0,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.938246450683e-05, -1.824183794579e-04, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.329492938451e-04, -7.873296482144e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  3.675238582588e-05, -1.749132950629e-04, {   3,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -8.603111327655e-05, -1.255970506161e-04, {   2,   0,  -1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  1.943575097698e-04, -1.723927323229e-05, {   2,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.670544492402e-05,  1.747726730601e-04, {   1,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -3.473670652932e-05,  1.767212153695e-04, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  1.526851295073e-07, -2.110557547035e-04, {   1,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.910231575976e-07, -2.102811345580e-04, {   2,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.828443624823e-05,  1.423545851400e-04, {   2,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.640544632404e-04,  4.653478381826e-05, {   1,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.632868167908e-06, -2.088383712500e-04, {   2,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.810182162467e-04, -2.941369379648e-05, {   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.022491963671e-04, -1.081339559169e-04, {   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.638126833051e-04, -4.648997106417e-05, {   1,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.695776941848e-04, -4.049233521665e-05, {   1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.780813350845e-04, -3.193281118328e-05, {   0,   0,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.980456565701e-05,  1.397624346620e-04, {   0,   0,   5,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.221852418910e-07,  2.085921311553e-04, {   2,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -4.290662707402e-07,  2.088156569848e-04, {   2,   0,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -9.157756387993e-06,  2.000470672074e-04, {   0,   0,   1,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 {  1.309784805980e-08,  2.089598860055e-04, {   1,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.214444184402e-05, -1.768103916151e-04, {   4,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.864073037307e-04, -2.241042745692e-05, {   2,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.247257019638e-05, -1.562743025921e-04, {   1,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.366382016753e-05, -1.343597842877e-04, {   2,   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -7.393407115898e-05, -1.339208229134e-04, {   3,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -4.147351990380e-05, -1.658643805399e-04, {   1,   1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -7.289786008687e-05, -1.339384794096e-04, {   2,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.068074178690e-04,  3.268239457000e-09, {   1,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.417090649066e-04, -6.501996096816e-05, {   4,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.022072818819e-04, -1.045114113281e-04, {   2,   0,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  1.892375802438e-04,  1.700406491026e-05, {   1,  -1,   3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.892375802438e-04, -1.700406491026e-05, {   1,  -1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.266373971162e-05,  1.232329038993e-04, {   0,   0,   0,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  7.757010343025e-05,  1.281095651781e-04, {   0,   0,   1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.391955086825e-05,  1.914171827200e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  1.655390591230e-04, -3.975638149453e-05, {   1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.458996009768e-06,  2.034677461195e-04, {   2,   0,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.670708734617e-04, -3.782916489200e-05, {   4,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.656123987703e-04,  3.922667654090e-05, {   1,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.411363534601e-04, -6.359302700414e-05, {   0,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -2.026469240295e-04, -2.030294853843e-06, {   2,   0,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.096871492554e-05, -1.434754735066e-04, {   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.577923869377e-05, -1.479457976054e-04, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  1.180302724385e-04, -8.564677036467e-05, {   0,   0,   0,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.379445666425e-04, -6.560826432746e-05, {   2,   0,  -1,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 {  4.578628574115e-05,  1.573410422409e-04, {   3,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -5.728026333006e-05, -1.456393631531e-04, {   2,   0,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.789801225914e-07, -2.021029491521e-04, {   2,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.464097535998e-04,  5.595359846618e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.881535551112e-04,  1.410980704471e-05, {   0,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.641820905111e-04,  3.805997167254e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 { -6.414003360923e-05,  1.378171411416e-04, {   2,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.336420979005e-04, -6.817344456320e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.993406307107e-04,  2.106957511297e-06, {   2,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.603978501156e-05,  1.653752084203e-04, {   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.100827761567e-05, -1.500983453516e-04, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  1.919019646521e-04, -9.010505864907e-06, {   0,   2,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.156598386892e-05,  1.591670327374e-04, {   1,  -1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.994863105277e-04, -1.092694619053e-06, {   1,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.072585664691e-04,  9.290295626558e-05, {   0,   0,   1,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 { -1.584889840178e-04, -4.161627551993e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.722713633098e-09, -1.999999999815e-04, {   4,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.672634730235e-04,  3.246922460512e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -1.069794346254e-04, -9.266190145764e-05, {   0,   0,   1,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  6.557726944925e-05, -1.340454449905e-04, {   0,   0,   1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.638445254663e-05, -1.330101109113e-04, {   3,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.804187621111e-04,  1.889597331581e-05, {   2,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.685650802677e-05,  1.623432268792e-04, {   2,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.054036207045e-05, -1.684920791302e-04, {   3,  -1,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.224563828475e-05, -1.264766922186e-04, {   4,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.554063329783e-04, -4.327188962057e-05, {   4,   0,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.553905018059e-04,  4.326295545151e-05, {   4,   0,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.293237077544e-04, -6.894946721502e-05, {   0,   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -6.602738277953e-05, -1.321995967109e-04, {   2,  -2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.562487768390e-05,  1.825681776998e-04, {   2,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.521879930613e-04, -4.561789199270e-05, {   3,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.521160583937e-04,  4.560800952154e-05, {   3,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  6.569907862273e-05,  1.315416807245e-04, {   3,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.447440999941e-04, -5.248561726005e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -1.423609692530e-04, -5.484974979845e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  7.093379339187e-05, -1.261884407726e-04, {   4,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.741537462983e-04, -2.284870052179e-05, {   0,   0,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  6.548787863006e-05, -1.311456393471e-04, {   0,   0,   3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.579797882095e-04, -3.864938604527e-05, {   1,  -1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.899556558570e-04, -6.659628870439e-06, {   0,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.271522833806e-04,  6.944312817944e-05, {   0,   0,   2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  4.410651663660e-05, -1.516895747555e-04, {   1,   1,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  7.202558075229e-05, -1.235977039317e-04, {   2,   0,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -1.173576271600e-04, -7.823952793931e-05, {   2,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.822840155752e-05, -1.370406122834e-04, {   1,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.822840155752e-05,  1.370406122834e-04, {   1,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.574524624948e-04, -3.763120007993e-05, {   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.518307204716e-05,  1.298800094105e-04, {   2,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.120130884677e-04, -8.297465305086e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.427620899409e-04, -5.216096253476e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.128636276911e-04, -8.203630273927e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 {  1.787556330317e-04,  1.611114406703e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.303752888460e-05,  1.316686439384e-04, {   2,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.206430221600e-05,  1.325751064906e-04, {   1,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.091433358566e-05,  1.236260773341e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 { -3.017388063556e-05, -1.642919518350e-04, {   4,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.590042176818e-05, -1.582396531857e-04, {   0,   0,   2,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 { -1.163293322843e-04,  7.760957887895e-05, {   2,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.567674933190e-04, -3.707042767610e-05, {   1,   1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.142048872947e-04, -7.954301743465e-05, {   2,   0,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -1.369063870219e-04,  5.683271241686e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -3.955713438167e-05, -1.541768527219e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -3.681856642693e-05, -1.568823980714e-04, {   2,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.566234669811e-04,  3.704265502271e-05, {   3,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.566342870009e-04, -3.684117527403e-05, {   1,  -1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.857090857250e-04,  7.587779246335e-06, {   2,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -5.943096813017e-05,  1.335100330812e-04, {   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.871381367691e-04,  5.739351560904e-06, {   4,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.766672797486e-05, -1.551246801612e-04, {   1,  -1,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -9.741804942196e-08, -1.925840545047e-04, {   0,   0,   1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.417209977196e-04,  5.086211337323e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  1.820161366340e-04,  1.041516181601e-05, {   2,   0,  -1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -4.630060550591e-05,  1.458115763714e-04, {   2,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  2.366859096979e-07,  1.918024378417e-04, {   0,   2,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.484779554295e-06,  1.844122847780e-04, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  3.923867882592e-05, -1.526440481884e-04, {   1,  -1,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.347341286589e-04,  5.711786076134e-05, {   4,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.240728597807e-04, -6.764022955939e-05, {   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.534864449904e-04, -3.804843793907e-05, {   4,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  6.398858207107e-05, -1.271512865520e-04, {   0,   0,   0,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.485412403208e-04, -4.230203784938e-05, {   0,   0,   1,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.769446482450e-05, -1.730145139253e-04, {   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  6.826836259283e-05,  1.219832677502e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 {  3.846705671223e-09, -1.899999999611e-04, {   2,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.061160185167e-09, -1.899999999970e-04, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  4.541761442985e-06, -1.853788189403e-04, {   2,  -2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -8.674745686687e-06, -1.810660322404e-04, {   2,  -2,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.826252025087e-05, -1.614772024635e-04, {   2,  -2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.580718016418e-04, -3.158343114709e-05, {   0,   0,   1,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.307373502209e-05, -1.263382529875e-04, {   2,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.338011298908e-04, -5.553518330017e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -1.856196636602e-04, -3.631666786693e-06, {   0,   0,   3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.533860406257e-04, -3.578259810007e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.109197989791e-04, -7.810712826771e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.044363318811e-04, -8.408880252540e-05, {   2,   0,  -1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -8.967551720038e-08,  1.884094030806e-04, {   2,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.053965227238e-04, -8.305372037695e-05, {   2,   0,   0,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -4.946490526260e-06, -1.834912162177e-04, {   3,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.725977341077e-07,  1.880530522780e-04, {   0,   0,   1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.524715292479e-04,  3.593072182298e-05, {   1,  -1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -7.045643516886e-05,  1.179436046004e-04, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  3.487567248941e-05,  1.535140076233e-04, {   0,   0,   0,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -7.363171451322e-05, -1.146762191948e-04, {   2,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.520823628262e-04, -3.599031396558e-05, {   1,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.390531992633e-04, -4.874744547977e-05, {   2,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.040199601366e-04,  8.371255245541e-05, {   2,   0,  -1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  7.540529996852e-06, -1.799410741733e-04, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -7.550106460475e-06, -1.797998795464e-04, {   2,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  7.343767560927e-07,  1.865897348279e-04, {   2,   0,  -1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.135985115648e-04,  7.368197875749e-05, {   2,   0,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.047718627265e-04,  8.250294668629e-05, {   2,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  1.456495830780e-04, -4.150093139565e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -6.527216241838e-05, -1.218509740780e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 {  1.626056268675e-04, -2.440377830123e-05, {   0,   0,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.688846547497e-05, -1.300380986558e-04, {   4,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -3.865725458915e-07,  1.864510022952e-04, {   6,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.512531390758e-04,  3.548734577323e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -6.511798154631e-05,  1.215636949481e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 {  2.981545009468e-05,  1.568503469036e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 {  4.727444576542e-07,  1.861316631559e-04, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -5.563724331691e-05,  1.309521269904e-04, {   4,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.563724331691e-05, -1.309521269904e-04, {   4,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.789528097769e-04,  7.428954655328e-06, {   2,   0,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -6.206268795829e-05,  1.242238291172e-04, {   2,  -2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.481437831483e-04,  3.790355565577e-05, {   0,   0,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.553367098107e-05, -1.203405066963e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 { -5.554704460083e-05, -1.301300938006e-04, {   2,   0,  -1,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 { -6.591934054770e-05, -1.192546523553e-04, {   2,   0,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.491947483307e-04,  3.570549884273e-05, {   1,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.905805687051e-05,  8.550982029440e-05, {   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.823122977339e-05,  1.161722526183e-04, {   2,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  3.408550761799e-05, -1.500643815550e-04, {   0,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.489432354223e-04, -3.503055364106e-05, {   1,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.740436911764e-05,  1.265246274517e-04, {   2,   0,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.007484612889e-04,  8.302757483008e-05, {   4,   0,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.786230326607e-04, -4.900156722071e-06, {   2,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.065760076251e-07,  1.833073164104e-04, {   2,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.795065981731e-04,  3.880241061612e-06, {   2,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.313826244036e-04, -5.200335486198e-05, {   1,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.480708746683e-05, -8.842378239074e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  9.939576352387e-05, -8.378831776273e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.742627380100e-04, -8.795509783820e-06, {   2,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.490197754290e-04,  3.397028808616e-05, {   4,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.558532598664e-07,  1.824097476578e-04, {   4,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.337039649460e-05, -1.492836276501e-04, {   0,   0,   2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.181665010183e-04, -6.435453122081e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.279780097077e-04,  5.439806562913e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.791217488748e-04,  2.981586259081e-06, {   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.745597515407e-04, -7.388410210010e-06, {   0,   0,   1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -6.386725436568e-08, -1.818409625742e-04, {   2,   0,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.335828869470e-06, -1.804805951202e-04, {   2,   0,   0,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.017281465643e-05,  1.712769735612e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  6.148321767002e-05, -1.196111736014e-04, {   2,   0,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 { -1.469274664695e-04, -3.377101787383e-05, {   1,  -1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.745328157305e-07, -1.803918730618e-04, {   2,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.517933636890e-07, -1.802982540370e-04, {   1,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  7.410992807970e-05, -1.064522571198e-04, {   2,   0,  -1,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  6.515620753647e-05,  1.153238958420e-04, {   2,   0,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  4.270991204895e-07, -1.799994932947e-04, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.347222643204e-04, -4.539440315759e-05, {   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.052654522735e-09, -1.799999999544e-04, {   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.324778563615e-09, -1.799999999850e-04, {   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -7.045198409871e-05,  1.094639925345e-04, {   4,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  5.922252716651e-05, -1.206925373006e-04, {   2,   0,  -1,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.267982173168e-05,  1.072149347071e-04, {   1,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.511853770897e-04, -2.863445747309e-05, {   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.137124549897e-05, -1.576843969630e-04, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  4.446662265631e-06,  1.745022155890e-04, {   2,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.428260998714e-04, -3.560044080124e-05, {   1,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.578466011645e-05, -1.624019393006e-04, {   1,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.893396571076e-05, -1.191971618640e-04, {   0,   0,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.374834027934e-04, -4.061604352002e-05, {   2,   0,  -2,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.174497826674e-05,  1.161827614515e-04, {   0,   0,   1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.492840976593e-07, -1.771646669113e-04, {   2,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  6.488192979305e-05,  1.126513878580e-04, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.440221614229e-04,  3.345590228232e-05, {   1,  -1,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.794015876755e-05, -1.489797603184e-04, {   1,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.429109542247e-04,  3.385216700007e-05, {   1,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.705741711282e-04,  5.964593462507e-06, {   2,   0,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.427252918344e-04,  3.321559103897e-05, {   4,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -8.276944775380e-05, -9.297254171430e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  7.919831810265e-05, -9.618597616867e-05, {   2,   0,   0,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  9.191049384754e-05, -8.331739847883e-05, {   2,   0,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -1.700756757703e-04,  4.811548236835e-06, {   0,   0,   3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.649149748879e-05,  1.182235725531e-04, {   2,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.064519874887e-04, -6.822497403429e-05, {   2,   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  9.944599008784e-05, -7.515076595375e-05, {   2,   0,   0,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -9.931081146814e-06, -1.645356642191e-04, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  9.933721499946e-05,  7.507436072822e-05, {   2,   0,   0,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -1.264610166775e-04, -4.774998673804e-05, {   1,   1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.339074540503e-04, -4.010107633182e-05, {   2,   0,  -2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.240374393045e-04,  4.992184637536e-05, {   4,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  2.705753266762e-05,  1.466114353772e-04, {   0,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -9.396851968544e-05,  7.907594897665e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  5.722451044250e-05,  1.158184086718e-04, {   0,   0,   1,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 {  1.690446718889e-04, -3.900788095342e-06, {   0,   0,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.906000645505e-06,  1.656610776268e-04, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.395592143930e-04, -3.292641654044e-05, {   1,   1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.535711350638e-05,  1.370537030897e-04, {   1,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -9.193623185162e-06,  1.631390050937e-04, {   0,   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.224157226222e-04,  4.987988244425e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -1.657064179276e-04,  6.512653848114e-06, {   2,   0,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.710956879153e-04, -1.023644419009e-06, {   2,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.831681241191e-05,  9.365454844408e-05, {   2,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  2.101963830823e-07, -1.715383563463e-04, {   2,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  9.824983089506e-05, -7.341114645812e-05, {   2,   0,  -1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  1.394594773900e-04,  3.215869036116e-05, {   1,  -1,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  5.681245304874e-05, -1.147707240164e-04, {   0,   0,   1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  9.817504496198e-05,  7.335910323872e-05, {   2,   0,  -1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  1.696298825944e-04, -1.381177935894e-06, {   3,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.603420965208e-05, -7.482968299624e-05, {   2,   0,  -1,   0,   0,   0,  13, -21,   0,   0,   0,   0,   0 } },
 {  3.850760699062e-05, -1.320553660953e-04, {   1,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.699994164093e-04, -4.454441673027e-07, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.576946475605e-05,  1.446703602123e-04, {   2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.056182772143e-04, -6.469307144844e-05, {   0,   0,   2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  7.932669027893e-05,  9.084116683955e-05, {   2,   0,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.518521611297e-05,  1.249762130647e-04, {   2,   0,  -2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  3.204424506477e-09,  1.699999999698e-04, {   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.343977185339e-09,  1.699999999838e-04, {   4,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.499295164188e-05,  1.049863720793e-04, {   0,   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.211069026024e-04, -4.878684413228e-05, {   0,   0,   2,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 {  5.441621994809e-05, -1.153794263115e-04, {   2,   0,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -1.149084935393e-04,  5.467790592191e-05, {   2,   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -3.017053076926e-05, -1.390156749173e-04, {   4,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.045251347784e-04,  6.457841601860e-05, {   2,   0,  -1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -2.181952447009e-05, -1.470475376565e-04, {   0,   0,   0,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 { -8.380276146334e-06, -1.603760651449e-04, {   3,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.364412773718e-04,  3.231344091007e-05, {   1,  -1,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.420571725041e-04, -2.667421149246e-05, {   0,   2,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.616538953272e-04, -7.075678464461e-06, {   4,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.047142167196e-05, -1.280010962356e-04, {   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.695065544860e-05,  1.014980222432e-04, {   0,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.647105374159e-04,  3.673781453842e-06, {   0,   2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.189444714199e-04, -4.938244110636e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.423816747915e-04, -2.537298876855e-05, {   0,   0,   2,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.349058302976e-04,  3.264190150075e-05, {   4,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.388748202268e-06,  1.661351945824e-04, {   0,   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.617585165541e-04, -5.674577325678e-06, {   0,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -9.307233191367e-05, -7.431162090137e-05, {   1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -3.643637242728e-05, -1.305172607027e-04, {   2,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.667101772018e-04, -4.254502185530e-08, {   1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.158578612094e-04, -5.080146395972e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.209006163166e-04, -4.572539883869e-05, {   0,   1,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.207388456844e-04,  4.580304356820e-05, {   0,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.478851367450e-04,  1.864657842617e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  9.247766551990e-08, -1.663052636151e-04, {   0,   0,   3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.967072334964e-05, -6.667364781111e-05, {   4,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.383649229044e-04,  2.776585673222e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  1.494242479113e-04, -1.654860021135e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.492160377202e-04,  1.664907115980e-05, {   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.142306490824e-07, -1.655968086770e-04, {   0,   0,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.520804798545e-04,  1.349117702733e-05, {   2,  -2,  -2,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.721071290241e-05,  1.279293220016e-04, {   3,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.489217725356e-05, -1.498291171925e-04, {   2,   0,  -1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  9.169488679943e-05,  7.297845159981e-05, {   2,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.858599107320e-05, -9.607387709013e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -9.462335950208e-05, -6.983185617359e-05, {   0,   0,   3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.459692351869e-05,  6.979903981617e-05, {   0,   0,   3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.333869635967e-04, -3.075238301221e-05, {   1,   1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.137927660800e-07,  1.638085091058e-04, {   0,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.216562192222e-04, -4.241002825733e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -1.623685381778e-04,  1.670735337776e-06, {   0,   0,   2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.326399295534e-04, -3.139760269473e-05, {   1,  -1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.385931178785e-08,  1.638185894496e-04, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -9.887771413186e-07, -1.626171416824e-04, {   2,   0,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.061479982755e-04,  5.744716939858e-05, {   0,   0,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -1.371764993202e-04,  2.638157755657e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -1.161453551452e-05,  1.519089226393e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -9.305363392627e-05, -7.046222697907e-05, {   0,   2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.286648778188e-05,  7.048394560606e-05, {   0,   2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.031044364253e-05, -1.330355521391e-04, {   0,   0,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.158526423456e-05,  1.516576269383e-04, {   2,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 { -9.778165415488e-05, -6.540116897874e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.366609050289e-04, -2.645826614898e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -6.033904866200e-09, -1.629858980200e-04, {   2,  -2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.287907313022e-05,  1.499838650700e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -1.493981318390e-04, -1.342379242325e-05, {   3,  -1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.255821541276e-05, -1.501660459893e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  1.442828765140e-04, -1.836807080486e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  3.574211588964e-05, -1.266427416269e-04, {   2,   0,   0,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 { -6.043572227242e-06,  1.561267902369e-04, {   4,   0,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.583450523318e-04, -3.702809501695e-06, {   0,   0,   2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.303095551712e-04, -3.166800068664e-05, {   1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.279559423294e-04,  3.394931062421e-05, {   2,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.310296517018e-04,  3.082505611160e-05, {   1,  -1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.069022208894e-06,  1.587778675383e-04, {   0,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.314501763178e-04,  3.014220775044e-05, {   4,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.030957935796e-04,  5.842795617399e-05, {   0,   0,   3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.569832220040e-04, -4.276731642031e-06, {   2,   0,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.246272196248e-07, -1.606538009576e-04, {   3,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.504357112293e-04,  1.056711989642e-05, {   2,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -1.074900554704e-04,  5.339855893627e-05, {   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.297946217641e-04, -3.097954425351e-05, {   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.534323794753e-05, -1.352731293438e-04, {   1,  -1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.361506509975e-04,  2.436040694518e-05, {   0,   0,   2,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 {  1.695157246972e-06,  1.587197471011e-04, {   0,   0,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.011072640209e-04, -5.891565458004e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -2.848377339096e-09,  1.599999999746e-04, {   2,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.094395102324e-09,  1.599999999863e-04, {   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.495711085793e-05,  7.499056773926e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 {  2.050304699991e-07, -1.593086262895e-04, {   1,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.533910576801e-04, -6.069597378499e-06, {   2,   0,  -1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.004161594680e-06, -1.583170580451e-04, {   0,   2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.290030715140e-04, -3.028891702373e-05, {   1,   1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.399354917436e-06,  1.538521011939e-04, {   4,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.580389200161e-05, -1.434154945748e-04, {   2,   0,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.128836383266e-05,  1.178531583515e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  5.235191419965e-05, -1.067399875819e-04, {   0,   0,   1,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.287479256852e-04, -3.034192972740e-05, {   0,   0,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.463211624127e-05,  1.342721812216e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  1.285334005313e-04,  3.034303739791e-05, {   1,  -1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.339821896289e-04,  2.487730979668e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.533084916067e-04, -5.548565144462e-06, {   0,   0,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.357136406587e-05, -9.521104001130e-05, {   2,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  2.502491700957e-05,  1.335970272385e-04, {   1,  -1,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.219036591916e-05,  1.064018563003e-04, {   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.953870319336e-05, -6.885851499762e-05, {   2,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.270930214835e-04,  3.119085209728e-05, {   3,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.681904886555e-07, -1.575166845984e-04, {   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -9.698584133838e-05,  6.110038441751e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.183042742163e-05,  1.461734620564e-04, {   0,   0,   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.138647472243e-04, -4.405080785147e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -9.646206329452e-05, -6.127829965793e-05, {   2,   0,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  9.878011914274e-05, -5.888402718305e-05, {   2,   0,  -1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  1.202115686772e-04, -3.738120070956e-05, {   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.787938847474e-07,  1.569972877529e-04, {   2,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.450867224079e-04,  1.239407958153e-05, {   2,   0,  -2,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -1.524742848040e-04,  4.976485326404e-06, {   2,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.271651745295e-04,  3.012835138435e-05, {   3,  -1,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.172565238014e-04,  3.997899288819e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  9.227207548472e-05, -6.483879834062e-05, {   1,  -1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -6.586851380960e-09,  1.570058690820e-04, {   0,   0,   2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.194053600900e-04, -3.758590002336e-05, {   0,   0,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.389300378847e-05, -1.329257830064e-04, {   2,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.220559312015e-05,  1.045257686910e-04, {   2,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -6.867229539817e-05,  8.805138549084e-05, {   2,   0,  -1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  1.874993277007e-06,  1.548125249504e-04, {   4,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.626373004953e-05,  7.036554009031e-05, {   1,   1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.720152578668e-08,  1.563671493602e-04, {   2,  -2,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.269783527604e-05, -6.346338621103e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.531840975034e-05, -1.207425351599e-04, {   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.082882370634e-04, -4.768698601611e-05, {   0,   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -1.160493604161e-05, -1.443000619085e-04, {   4,   0,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.255030548784e-04, -3.025630164569e-05, {   2,   0,  -1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.070304935661e-04, -4.870532505932e-05, {   2,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.175591996038e-05, -1.439128008196e-04, {   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  6.483535738778e-05,  9.073876345170e-05, {   1,   1,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  1.258841690390e-04,  2.943284099785e-05, {   4,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -3.285704488991e-06,  1.519136041107e-04, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -2.434573381924e-05,  1.307626232525e-04, {   2,   0,  -3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.430968969502e-05, -1.306461830293e-04, {   2,   0,  -3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -3.846319455017e-05, -1.164142432659e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 { -1.085184342086e-06, -1.534486628872e-04, {   2,  -2,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.102599797202e-05,  6.309622718185e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  6.151354803814e-05,  9.260792073027e-05, {   4,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -5.665920185516e-05,  9.723755543002e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  1.223267404029e-04, -3.154699884597e-05, {   2,   0,  -1,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.080178577732e-05,  1.030223929701e-04, {   2,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.805665732077e-05,  1.157614889921e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 { -7.036292041421e-05,  8.337874358933e-05, {   2,   0,  -1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  1.382751772655e-04,  1.534143587409e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 {  9.672299037262e-07, -1.526164883079e-04, {   0,   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -9.009179935677e-05,  6.329015582296e-05, {   1,  -1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.093627004170e-05, -1.323826690066e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  1.412403883903e-04, -1.206578312666e-05, {   2,   0,   0,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -3.148615112003e-05,  1.217965599697e-04, {   3,   0,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -4.304057512804e-05, -1.100311962738e-04, {   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.018667642613e-05, -1.428698580882e-04, {   4,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.092378230359e-04,  4.365631466035e-05, {   4,   0,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  1.232762510634e-04,  2.951576850362e-05, {   1,  -1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.419536033195e-06, -1.453371981358e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  5.133549851620e-09, -1.527500883275e-04, {   0,   0,   1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -4.954385089342e-05,  1.030089121324e-04, {   2,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.007271665632e-05, -1.223525845723e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  4.019554006668e-07, -1.518607072254e-04, {   2,   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -1.270847442840e-04,  2.512512033212e-05, {   4,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.075398793934e-04,  4.464317478441e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -1.102470715590e-04, -4.190954353229e-05, {   0,   0,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.091885761829e-05,  1.310622341938e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  1.394388051112e-04,  1.252276443809e-05, {   3,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.397699610779e-05,  1.379634298448e-04, {   2,   0,  -1,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.414571660874e-06,  1.502717349445e-04, {   4,   0,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.136065868967e-05,  1.200536915750e-04, {   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.295362743376e-04,  2.186814295628e-05, {   1,  -1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.269851214165e-05,  1.285942659734e-04, {   4,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -6.057739899471e-05, -9.061940060200e-05, {   0,   0,   1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  8.751036115253e-05, -6.350529950610e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.617000510890e-05, -9.457764284474e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.187981978344e-04, -3.193245862501e-05, {   2,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  5.012400110021e-05, -1.003985915778e-04, {   2,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.355413441567e-05, -9.666098627223e-05, {   2,   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  9.201076512663e-05,  5.818300407609e-05, {   2,   0,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.253697988424e-04, -2.480342104031e-05, {   2,   0,  -1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -2.094395102313e-10,  1.499999999999e-04, {   2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.279139578595e-04, -2.205302390475e-05, {   3,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.279139578595e-04,  2.205302390475e-05, {   1,   0,  -1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.211335740819e-04,  2.879348547987e-05, {   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.154243421927e-05,  7.831489954688e-05, {   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.070771632784e-04,  4.267365765060e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.230130168343e-05, -1.273715672866e-04, {   6,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  9.851161112151e-05,  5.111800766320e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 { -1.395085105398e-04, -1.010123586774e-05, {   2,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.383492171399e-05, -1.256495046397e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.178472713385e-04,  3.145918422304e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 {  1.447815768699e-04,  4.432876112339e-06, {   2,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.202348075660e-04, -2.897222960478e-05, {   3,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.265192145781e-04, -2.265496095404e-05, {   0,   0,   0,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -1.256238975099e-04,  2.331260588054e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.422297883071e-06, -1.464921082958e-04, {   2,   0,  -2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.273878239552e-04, -2.148712102161e-05, {   1,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  9.110242218183e-05,  5.773258382770e-05, {   2,  -2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  6.387652414387e-07,  1.481322719172e-04, {   2,   0,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.420137699685e-04,  6.753557826714e-06, {   2,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  1.363321786817e-04,  1.241183263920e-05, {   0,   0,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -1.089802715124e-04, -3.975420934995e-05, {   0,   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  9.112039256120e-05, -5.739945334446e-05, {   2,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.974508027978e-05, -9.876524375389e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  5.176418821787e-05,  9.671573130877e-05, {   2,   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -8.070960455541e-05,  6.773712268908e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.105074029347e-04, -3.779380798527e-05, {   2,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -1.310573361761e-04,  1.723974830325e-05, {   0,   0,   1,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 { -3.133593659607e-05,  1.168901866365e-04, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  6.397062525699e-05,  8.419484429233e-05, {   0,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.197647933007e-04, -2.829844881947e-05, {   1,  -1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.150185485795e-06,  1.388684103653e-04, {   2,   0,  -1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.677647253647e-06, -1.462250783886e-04, {   4,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.304923882844e-04,  1.735178734850e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -3.792003058823e-05,  1.096518665661e-04, {   0,   0,   1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  6.207953534892e-05,  8.547606703732e-05, {   0,   0,   1,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 { -1.986820462852e-05,  1.276772613708e-04, {   0,   0,   1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -6.686353781220e-06,  1.408105802322e-04, {   0,   0,   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  1.192173726959e-04,  2.824171973307e-05, {   1,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.483145733424e-05, -9.237818481349e-05, {   2,   0,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  1.457051595209e-04, -1.459879826600e-06, {   2,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.139803651846e-04, -3.314198898537e-05, {   4,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.138611778086e-04,  3.310014189257e-05, {   4,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.791043710316e-05,  9.901812973685e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.882624529025e-05, -1.280441071690e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  5.008601406662e-05,  9.671569948976e-05, {   0,   0,   1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.185070282521e-04,  2.828886251371e-05, {   1,  -1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.169889099805e-04, -2.975164296851e-05, {   1,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.121873740107e-06, -1.456153751316e-04, {   2,   0,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -2.219678293393e-05,  1.244582736266e-04, {   3,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.435569409400e-04, -3.020342898628e-06, {   2,   0,  -3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.022959009162e-07, -1.463182101182e-04, {   2,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  6.325911787074e-05, -8.306734557345e-05, {   0,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -9.403651723710e-05,  5.226144172739e-05, {   2,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.122434371659e-04,  3.379638694940e-05, {   2,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.701259926281e-05,  7.889385277021e-05, {   2,  -2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.390660761429e-05,  1.218526582033e-04, {   2,   0,  -2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.000987729214e-04,  4.560960052047e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.262944607574e-04,  1.908214625236e-05, {   4,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -3.936950337776e-07, -1.449509688021e-04, {   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.428897730151e-05, -1.209435896471e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -9.178453750487e-05,  5.331843288101e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  1.025395834562e-04, -4.255380527428e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  3.069228257966e-06,  1.419787760550e-04, {   3,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  9.544679540415e-05, -4.951388028073e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 { -6.224077686247e-05,  8.268232118599e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  2.899022253483e-05,  1.159126920885e-04, {   1,   1,  -3,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.989879784195e-08, -1.448339429754e-04, {   1,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.573809113002e-06, -1.431258397145e-04, {   2,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -5.302472494985e-08,  1.445194860430e-04, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -9.454371927916e-05,  4.962257268306e-05, {   0,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  5.004157947493e-05,  9.392382620757e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -1.392203769104e-04, -4.710298923227e-06, {   0,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.122695184433e-05, -9.264659164396e-05, {   3,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.273803928625e-04,  1.627952443405e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 { -1.410638858156e-04,  2.482558942244e-06, {   2,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.620040704742e-05,  1.072926380918e-04, {   2,   0,  -2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.134742409960e-04, -2.997820915191e-05, {   0,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -3.096719458911e-06,  1.402220588913e-04, {   0,   2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -9.057882109804e-05, -5.266265124329e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 {  4.707238425019e-05, -9.615987591556e-05, {   2,   0,  -1,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.442488797978e-05,  8.880186849152e-05, {   2,   0,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.150518452085e-04,  2.783946350806e-05, {   2,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.321832757942e-04, -1.069234535587e-05, {   2,   0,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.923558450089e-05,  9.360579978247e-05, {   2,   0,   1,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  1.273024277354e-05, -1.299299476821e-04, {   0,   0,   1,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 {  1.271388992894e-05,  1.299267238941e-04, {   0,   0,   1,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 {  9.001677518409e-05, -5.258972989945e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -1.693815866911e-07, -1.423784268444e-04, {   2,   0,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -6.461827173250e-05, -7.774896832960e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -1.150819504849e-04, -2.723733569391e-05, {   3,  -1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.360384655345e-06, -1.369440953600e-04, {   2,   0,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  1.341566425688e-06, -1.409245531187e-04, {   0,   0,   3,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 {  8.956222975678e-05,  5.258754546401e-05, {   2,   0,   0,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -1.249357256533e-04,  1.713164415422e-05, {   0,   0,   1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 {  1.399846550933e-04, -2.072761062806e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.952693940185e-05, -1.024475081432e-04, {   2,   0,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 { -2.134881937408e-05,  1.204197537022e-04, {   0,   0,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  4.598556417818e-07,  1.412277070934e-04, {   0,   0,   1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.672753280348e-05,  8.486896876787e-05, {   4,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  8.605233555760e-05,  5.540916316412e-05, {   2,  -2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.974939871235e-05, -7.118051454459e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -8.619805260847e-06, -1.322044377240e-04, {   0,   0,   1,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 { -8.619805260847e-06,  1.322044377240e-04, {   0,   0,   1,   0,   0,   0, -14,  27,  -3,  -3,   0,   0,   0 } },
 {  4.959020852177e-05, -9.115802226934e-05, {   0,   0,   1,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.529806717415e-05, -1.052739491169e-04, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  8.534232891707e-05, -5.501956473426e-05, {   2,  -2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.021687167934e-04,  3.818021790622e-05, {   0,   0,   1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.035714658097e-04,  3.668567104292e-05, {   0,   0,   1,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  8.253976133545e-05, -5.767785004320e-05, {   2,   0,  -1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -1.092549910990e-04,  3.092097831501e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  7.427767707642e-05,  6.589744420714e-05, {   2,   0,  -3,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.017053907788e-05, -1.299655182858e-04, {   0,   2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.134149406910e-04,  2.663597918134e-05, {   1,  -1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.005456971755e-09, -1.399999999677e-04, {   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.687807047926e-09, -1.399999999742e-04, {   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.419026343148e-09, -1.399999999791e-04, {   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.685988057400e-09,  1.399999999898e-04, {   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.392453401687e-06,  1.355770391004e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  8.218435709558e-05,  5.772313614774e-05, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  6.644300074415e-05, -7.333170377726e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -3.876691586769e-05,  1.008820916956e-04, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -1.353551255813e-04, -4.256490069490e-06, {   0,   0,   2,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.129699486943e-04, -2.647258008480e-05, {   1,   1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.536714990812e-05,  8.404177683550e-05, {   0,   0,   1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  3.556911011253e-05, -1.035339465959e-04, {   3,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.169446596611e-06, -1.343850493777e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.127857683062e-05,  9.715489116033e-05, {   0,   2,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.902199855988e-08, -1.382981202711e-04, {   2,   0,   1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.297205295909e-04,  8.519636298170e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.743819365035e-05, -6.075887071556e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -4.924879213215e-05,  8.894553413881e-05, {   4,   0,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.117037321513e-04, -2.640445779147e-05, {   1,   1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.055091880938e-04,  3.252901133504e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -3.547904544524e-05, -1.025486734161e-04, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -3.556921119530e-05,  1.024306886563e-04, {   2,   0,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.162643215675e-05, -1.063169991836e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  3.093625866939e-07,  1.375082612322e-04, {   2,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.409825989390e-05, -5.346792375150e-05, {   2,   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -1.137957801018e-04,  2.369123678613e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.092203836720e-05, -9.634955363530e-05, {   0,   2,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.068401377988e-04, -3.037681774854e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  9.255928107155e-05, -4.458771371162e-05, {   2,   0,  -1,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 { -4.156039076619e-05, -9.556157546891e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -8.911322812085e-08,  1.369749383642e-04, {   4,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.193738434009e-07, -1.363745858459e-04, {   1,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.844908350732e-05,  1.082659588072e-04, {   4,   0,  -1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  2.844908350732e-05, -1.082659588072e-04, {   0,   0,   1,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 { -4.653967823356e-05, -9.001428537351e-05, {   4,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.899089350688e-05, -1.074458827318e-04, {   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.077692898712e-05,  1.156257606117e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  8.423821871676e-05, -5.213185047941e-05, {   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  4.029934265357e-08,  1.362468725764e-04, {   3,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  8.982504006154e-07,  1.353686541358e-04, {   2,   0,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -6.137145702476e-08, -1.361582725892e-04, {   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.682215277160e-05, -8.935672683574e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  6.520999016998e-05, -7.086818834064e-05, {   2,  -2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.805694946815e-05, -1.079505816065e-04, {   2,   0,  -1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  1.090445943652e-04,  2.686653190923e-05, {   1,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.160463321587e-05,  1.143038599350e-04, {   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.646659017176e-05, -1.093288633565e-04, {   0,   0,   1,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  4.646992762834e-05,  8.928638624097e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  7.752645632270e-07, -1.349529372323e-04, {   2,   0,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -8.310199904539e-05,  5.261203335053e-05, {   4,   0,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.207243365293e-04,  1.493690353205e-05, {   4,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  5.764839866392e-05, -7.779566575713e-05, {   2,   0,  -1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 { -2.805014578871e-06,  1.324913539955e-04, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -7.187035007738e-05, -6.341255449702e-05, {   2,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.225646949566e-05,  1.230218785759e-04, {   4,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.827361362634e-08,  1.351520803240e-04, {   2,  -2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.625528975874e-05, -3.878703624565e-05, {   2,   0,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.520349059346e-05,  1.197292931052e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 { -5.662532832132e-06,  1.289112538524e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -1.078938982841e-04,  2.662768313237e-05, {   3,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.923451273150e-08, -1.340319106710e-04, {   2,   0,  -2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.616450760839e-05,  1.178601821938e-04, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  6.035359095591e-06,  1.279620932015e-04, {   0,   0,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  6.229065572925e-05,  7.160382816805e-05, {   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.001350196599e-04, -3.373722452369e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  8.140349741111e-05, -5.245902932324e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  1.585430868680e-05,  1.179910042229e-04, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 {  9.710397560218e-05,  3.661557474886e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -8.403994621231e-06,  1.252247159424e-04, {   2,   0,   0,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 { -4.962855224507e-05,  8.399603602467e-05, {   0,   0,   1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  1.309120222553e-04,  2.685388468492e-06, {   2,   0,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -2.143426772890e-05, -1.120688610633e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  6.442575279160e-06, -1.270265942359e-04, {   2,   0,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  3.559184716619e-06,  1.298312046192e-04, {   2,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  9.477992221816e-05,  3.853473558862e-05, {   2,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.317118516961e-06,  1.249806984021e-04, {   2,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.817476611153e-05,  3.469232702896e-05, {   2,   0,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 { -7.999860710796e-05, -5.276198556043e-05, {   0,   0,   3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.265687895915e-04, -6.138428927387e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  1.289350720181e-04,  3.678516736800e-06, {   0,   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  1.966659961025e-05, -1.128592579914e-04, {   6,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.345263483202e-07, -1.321380875589e-04, {   0,   0,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.068801580870e-04,  2.535353824189e-05, {   3,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.089520264198e-05,  1.011670789390e-04, {   4,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.052945513776e-04,  2.585387475561e-05, {   1,   1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.672149138879e-05,  5.437805852502e-05, {   0,   0,   1,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -1.055506739182e-04,  2.542158359632e-05, {   1,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.127252483363e-05,  5.968793891818e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 { -1.928700817868e-07,  1.307419080257e-04, {   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.435694403831e-05, -5.654165688985e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.272881970036e-05, -1.081338576823e-04, {   3,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  2.272881970036e-05,  1.081338576823e-04, {   1,   0,  -1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -4.243352434581e-05,  8.841287978426e-05, {   2,   0,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.280298727500e-06, -1.275192875975e-04, {   1,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.684753542811e-07, -1.305644114942e-04, {   1,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.237386107470e-04,  6.763698310853e-06, {   2,   0,  -1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -1.971428576970e-05,  1.107709917362e-04, {   3,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.632723416144e-05, -5.406688472298e-05, {   0,   0,   1,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 {  1.177514672187e-04, -1.261943988750e-05, {   2,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.041822137069e-05, -7.978044346872e-05, {   0,   0,   1,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  3.847056949408e-07,  1.297804049328e-04, {   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.633104008327e-07, -1.299423497107e-04, {   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.037544346703e-05,  7.972487202712e-05, {   0,   0,   1,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -1.058089986228e-04, -2.421880302359e-05, {   1,  -1,   0,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.548868789821e-05,  1.144992290086e-04, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  7.526613514339e-08,  1.298754280035e-04, {   1,   0,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -3.458153561379e-05,  9.493112277533e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 { -7.961711856009e-05, -4.969026901863e-05, {   2,   0,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  1.250451722866e-04,  4.249032081082e-06, {   2,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.046185801065e-04, -2.456589809345e-05, {   1,  -1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.253556812066e-04, -3.575303799680e-06, {   2,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.728640499115e-07, -1.283684761840e-04, {   0,   2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -4.619457804320e-06,  1.240917973948e-04, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -3.346169723534e-05,  9.509981650536e-05, {   2,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.275162537881e-04,  7.847940622737e-07, {   1,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.296736599903e-05,  5.532190294401e-05, {   2,   0,  -1,   0,   0,   0,   7, -10,   1,   0,   0,   0,   0 } },
 {  1.037332076354e-04, -2.454307688785e-05, {   1,   1,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.036003972461e-04,  2.457523077730e-05, {   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.503491723839e-05,  8.288615597744e-05, {   0,   0,   1,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.025033201815e-04,  2.530672048928e-05, {   1,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.464003912646e-05,  1.031074187671e-04, {   1,   1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.225254335913e-05,  9.520970225748e-05, {   2,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.719150560465e-05, -3.022018823335e-05, {   2,   0,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -5.188348005680e-05, -7.536632679300e-05, {   2,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.911631737647e-08, -1.270893917113e-04, {   4,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.791623682464e-05, -8.920156068352e-05, {   0,   0,   2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.081375401450e-05, -5.623519867296e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  1.994583196448e-05, -1.070162395754e-04, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  6.935513684097e-08, -1.267689928561e-04, {   0,   0,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.489953429027e-06,  1.212227241124e-04, {   2,   0,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.185698222359e-05, -8.480388946495e-05, {   0,   0,   1,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.023166892001e-04, -2.420607948108e-05, {   1,  -1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.745859039179e-08, -1.263867415513e-04, {   2,   2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.933093026422e-06,  1.244074669622e-04, {   0,   0,   3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -3.760564558153e-05,  8.850000725050e-05, {   0,   0,   4,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.537533081624e-05, -4.061435559958e-05, {   2,   0,  -2,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 {  5.613024027071e-05, -6.953598216927e-05, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  8.560677160467e-05, -4.004308829900e-05, {   0,   0,   1,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 {  8.556996466605e-05,  4.001601184381e-05, {   0,   0,   1,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 {  8.055737444869e-07, -1.246619498978e-04, {   0,   2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.411031392202e-05,  9.134522009888e-05, {   1,  -1,   0,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -1.203451178169e-04, -4.843654093156e-06, {   2,   0,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.003983354107e-04, -2.454198446685e-05, {   1,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.462855271507e-08, -1.248639267195e-04, {   2,   0,  -3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.463333730728e-05,  4.026220163289e-05, {   2,   0,   0,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -4.297995651375e-06,  1.205786195758e-04, {   0,   0,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.190612208556e-04,  5.671136761407e-06, {   2,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.175079644333e-08,  1.245083597346e-04, {   2,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.830441822088e-05, -7.604577892114e-05, {   4,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  8.548452203975e-06, -1.157927874560e-04, {   2,   0,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.007493321127e-04, -2.355628205882e-05, {   1,   1,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -9.874169179182e-05,  2.555810987649e-05, {   2,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  2.850698490334e-05, -9.561253846656e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  1.165419951920e-04,  7.423506647607e-06, {   2,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.006870729694e-07,  1.237048748111e-04, {   1,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.368901284693e-05, -9.018564116122e-05, {   1,  -1,  -2,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -3.523088357880e-05, -8.838075761404e-05, {   1,   1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  6.501831248196e-05,  5.857557296892e-05, {   0,   0,   1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  4.799926344025e-05,  7.557783805528e-05, {   4,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -8.263938824660e-05,  4.088679756274e-05, {   0,   0,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.083056847410e-06, -1.224364934609e-04, {   1,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.041172400096e-05,  1.130653146992e-04, {   4,   0,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.716145863654e-06,  1.213935163073e-04, {   2,  -2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.186139263498e-04, -4.478636734257e-06, {   0,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -7.583913043133e-05,  4.723308824211e-05, {   4,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.961475388583e-05, -9.345462679749e-05, {   2,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -1.141385441929e-04,  8.806125919091e-06, {   2,   0,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.854841860994e-05,  1.043578912502e-04, {   3,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.672548733266e-05,  5.588333767887e-05, {   1,  -2,   1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  6.672548733266e-05, -5.588333767887e-05, {   1,  -2,  -1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 {  2.016569486010e-05, -1.024248059830e-04, {   4,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  6.858179611635e-08,  1.224866031083e-04, {   3,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.068971718918e-07, -1.224409967212e-04, {   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.194442194668e-04,  3.073650746665e-06, {   0,   0,   3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.886639644581e-06,  1.206175679970e-04, {   4,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  3.850741250280e-05,  8.376548625003e-05, {   2,   0,   0,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 {  9.563704233813e-05,  2.662915175760e-05, {   0,   0,   2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  9.866195791259e-05,  2.335709525386e-05, {   1,  -1,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.537900219370e-05, -2.655477232349e-05, {   0,   0,   2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.137545949156e-07,  1.217254133745e-04, {   2,   0,  -1,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.153877764661e-04, -6.364753340603e-06, {   0,   0,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.857314844687e-05,  2.311075127928e-05, {   1,  -1,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.017872976340e-07, -1.215656122947e-04, {   0,   2,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.270451142465e-05,  7.867004048280e-05, {   0,   0,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 { -7.252241291169e-05, -4.865532373820e-05, {   0,   2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.099523778648e-05,  8.007149520246e-05, {   0,   0,   1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.401541819459e-05, -9.699983931336e-05, {   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.603642685873e-05, -2.482261630551e-05, {   1,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.144418688418e-04, -6.400732759075e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  3.758639660154e-05,  8.325543516522e-05, {   2,   0,  -1,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -5.370073843310e-05, -6.711040072499e-05, {   2,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.211481708541e-05,  9.867288576122e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 { -9.909235262300e-06, -1.107245822628e-04, {   2,   0,  -1,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.162487810479e-04,  4.380285557995e-06, {   2,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.743481821141e-05,  2.310140334425e-05, {   1,   1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.777287916470e-05,  2.261524406114e-05, {   0,   1,   1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  8.912108455685e-05,  3.120131363370e-05, {   2,   0,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.014647108364e-04,  1.881997879616e-05, {   1,  -1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  1.895056620879e-06, -1.183770824794e-04, {   4,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  5.202540926347e-05, -6.818039035567e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  9.704154684030e-05,  2.315555716254e-05, {   4,   0,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  9.715526998974e-05,  2.300620699882e-05, {   1,   1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.150256449068e-04, -5.133949013562e-06, {   2,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.309144261341e-09, -1.199999999544e-04, {   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  3.246312408302e-09,  1.199999999561e-04, {   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  2.157226955349e-09,  1.199999999806e-04, {   4,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.612684228800e-09,  1.199999999892e-04, {   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.466076571628e-09,  1.199999999910e-04, {   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.871994502144e-07,  1.194033269998e-04, {   1,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.159575425004e-04, -4.008007841567e-06, {   0,   0,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -4.883300131639e-05,  7.112649128093e-05, {   2,   0,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -1.127765404450e-04,  7.138936321217e-06, {   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.415448414138e-05, -4.574193647291e-05, {   2,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.115483536454e-05,  8.865159997666e-05, {   1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.704526471213e-06,  1.160942105046e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -8.456066898519e-05,  3.509951563352e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  8.456066898519e-05, -3.509951563352e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -2.220575330159e-05, -9.744748901246e-05, {   2,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.134609377081e-04, -6.110844214308e-06, {   2,   0,  -2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.107810213936e-04, -8.787834462221e-06, {   0,   0,   2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.915872637347e-05, -1.002972051966e-04, {   1,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.152541971921e-04, -4.173928329032e-06, {   0,   0,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.095590680279e-04,  9.839238428986e-06, {   3,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.074292599582e-05, -9.858892791473e-05, {   0,   0,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.205456380420e-05,  7.714987478314e-05, {   1,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.118980531987e-04,  7.170794151587e-06, {   2,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.142196844001e-04, -4.722764043347e-06, {   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.774379797521e-05, -6.103834810995e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -7.319281788818e-07, -1.179354422651e-04, {   2,   0,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.960231556409e-05, -8.902108428504e-05, {   2,   0,  -1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -7.355772099137e-05, -4.480788893080e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 { -1.040814818108e-04, -1.423381804418e-05, {   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.836430137080e-05,  5.988665222628e-05, {   2,   0,  -1,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 { -1.047546532176e-04, -1.332675652694e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 {  5.072927572254e-05, -6.731937967805e-05, {   1,   1,  -2,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -9.514626631097e-05,  2.286296839527e-05, {   1,  -1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.833001272610e-05, -4.956804631248e-05, {   2,   0,  -1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.359165374874e-05, -2.428263267305e-05, {   2,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.146575139529e-08, -1.177077886417e-04, {   2,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.520402583332e-05, -2.238669575256e-05, {   0,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  9.529440949959e-05, -2.201990674405e-05, {   1,   1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  9.477772820330e-05,  2.252108559233e-05, {   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.624368900832e-05,  6.097707313160e-05, {   2,   0,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.025590925783e-05, -9.696374943094e-05, {   4,   0,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  9.477961529297e-05, -2.237648479166e-05, {   1,  -1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.395159142003e-05,  1.030237780032e-04, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -8.854820143763e-05, -2.826872264076e-05, {   1,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -9.090070172722e-05, -2.576470708556e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -8.816352017799e-05,  2.819229449181e-05, {   1,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.509498157200e-06, -1.138162300565e-04, {   2,   0,   1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.155310005723e-04,  7.915124513450e-07, {   1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.755824182858e-08, -1.162382864132e-04, {   3,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.694395329047e-05, -3.928259887024e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  8.628864990449e-07, -1.151898063083e-04, {   4,  -2,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.156018204736e-07,  1.158991300502e-04, {   0,   0,   1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.033557524080e-08, -1.160036626561e-04, {   2,   0,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.150885175665e-04, -9.251563621110e-07, {   2,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.243432289227e-05, -3.346422940507e-05, {   0,   0,   1,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 { -3.037647642829e-05, -8.551478107966e-05, {   4,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.979068990458e-07, -1.154815542673e-04, {   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -2.281446951364e-05,  9.276009519651e-05, {   2,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.660108554392e-05,  3.897048490889e-05, {   0,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  5.867153350000e-05, -5.688993812847e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -3.168606943950e-06, -1.123911854112e-04, {   2,   0,   0,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 {  8.655256848081e-05, -2.881047694692e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  4.028377324272e-05,  7.507755199608e-05, {   2,   0,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.931401565315e-06, -1.092491870727e-04, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -9.531900932827e-05,  1.982286716692e-05, {   2,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.083822142059e-05,  7.418832949433e-05, {   2,   0,   2,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  9.212410338174e-05, -2.281968886594e-05, {   1,  -1,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.529339210919e-05, -6.959723181567e-05, {   4,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.133671098788e-04, -1.374320942933e-06, {   2,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.073439802184e-05, -7.399971979790e-05, {   2,   0,   0,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  1.707191684238e-09,  1.145358729936e-04, {   0,   0,   2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.301445483005e-05,  2.152088369550e-05, {   3,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.700235198564e-06,  1.107874362041e-04, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 {  6.966390689015e-05, -4.473149845488e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.091283292984e-04,  5.127030657112e-06, {   2,   0,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.893924247684e-05,  2.518669224725e-05, {   2,   0,  -2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.436057622654e-08,  1.138945754892e-04, {   2,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.871375681811e-05, -4.519483680140e-05, {   4,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.164291074915e-05, -7.220554746676e-05, {   0,   0,   1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  1.104208619027e-04, -3.363997033658e-06, {   2,   0,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.561234911777e-05, -7.800030359563e-05, {   2,   0,  -1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 {  1.134558986307e-04,  5.447916955100e-08, {   0,   0,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.578897685305e-05, -4.771892260305e-05, {   4,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.104088643961e-04, -3.090708463800e-06, {   2,  -2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.730977228888e-05, -7.618844251974e-05, {   2,   0,  -1,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.357481449363e-05, -3.987565490830e-05, {   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.665858750329e-05, -7.676257285185e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  4.913927948675e-05,  6.425418532277e-05, {   2,   0,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.931923678973e-05,  2.384831862530e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -1.084809513969e-04,  4.652433715655e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  9.052400818676e-05,  2.255486062763e-05, {   1,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.588756171041e-05, -7.716754510063e-05, {   2,   0,   0,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 { -1.094718327449e-04,  3.497276161487e-06, {   4,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  9.134729646089e-05, -2.131147631701e-05, {   1,   1,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.223349438629e-05, -1.003320643489e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  1.702743595488e-07, -1.122827963268e-04, {   0,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.041566232881e-04, -7.920173300892e-06, {   2,   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -9.036138506253e-05, -2.154303445330e-05, {   1,  -1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.779053188817e-06, -1.030348707925e-04, {   1,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.076018895051e-04, -4.128472559910e-06, {   2,   0,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  9.696051656372e-05, -1.457032825450e-05, {   0,   0,   2,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 {  6.340103271864e-05, -4.808494266987e-05, {   2,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -9.244835551371e-05, -1.900028643784e-05, {   3,  -3,   0,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -9.244835551371e-05,  1.900028643784e-05, {   1,   3,   0,   0,   0, -43,  42,   0,   0,   0,   0,   0,   0 } },
 {  1.799336292899e-06,  1.096229871548e-04, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -9.232468097433e-05, -1.898770192740e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.749388546644e-05,  9.355193227689e-05, {   1,  -1,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.134173827820e-05, -6.950290108422e-05, {   2,   0,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 { -3.171504290155e-05, -7.902904990401e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 { -1.035705308059e-04, -7.143523875418e-06, {   1,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  3.005256575069e-08,  1.106752676524e-04, {   6,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.050760145402e-05,  4.019323795068e-05, {   4,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.060571668317e-04,  4.630244297681e-06, {   2,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  9.169377383791e-05,  1.887336227751e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -8,   0,   0,   0 } },
 {  3.122009310709e-05,  7.934257089043e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 {  6.853531194237e-05,  4.192266974192e-05, {   2,   0,  -1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -3.581166964801e-05,  7.464212716208e-05, {   2,   0,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.033477853564e-04, -7.081173568300e-06, {   2,   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  7.328244791193e-07, -1.095989666448e-04, {   0,   0,   3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.609404133476e-05, -6.422746153552e-05, {   2,   0,  -1,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 { -7.298151267941e-05,  3.729697836706e-05, {   2,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  6.917242485572e-05,  4.109992780745e-05, {   0,   0,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  3.412622773996e-05, -7.607824519110e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 { -6.393900033190e-05,  4.621645675613e-05, {   2,   0,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  4.300491275819e-09,  1.099999999159e-04, {   2,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.785471824678e-09, -1.099999999855e-04, {   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -7.679448708484e-11,  1.100000000000e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -1.086678201093e-04,  1.272965915922e-06, {   0,   0,   2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -6.538827037162e-05, -4.453689318606e-05, {   1,   0,   1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  6.538827037162e-05,  4.453689318606e-05, {   1,   0,  -1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 {  5.144506877404e-07, -1.094065835821e-04, {   2,   0,  -1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  5.144506877404e-07,  1.094065835821e-04, {   2,   0,  -1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 {  8.871664710400e-05,  2.118543410684e-05, {   1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.450174872311e-05,  7.536902766029e-05, {   2,   0,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  1.081751448555e-04, -1.683617804610e-06, {   2,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.053698825549e-04, -4.470011439298e-06, {   4,  -2,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.086774964565e-04,  1.116052614536e-06, {   3,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.562993447041e-05,  6.408309785117e-05, {   2,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  3.793945334136e-05,  7.162766559744e-05, {   0,   0,   2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -3.546930683481e-05, -7.396233289506e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  2.502651886985e-05, -8.440341046580e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -6.095070717639e-05,  4.831893250060e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  2.422307150900e-05, -8.494188507989e-05, {   2,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -6.198961763971e-07,  1.085003845011e-04, {   4,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.572796140792e-06,  1.053353122711e-04, {   2,   0,  -1,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 { -9.959865146539e-05, -8.950342243532e-06, {   1,  -3,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.959876080701e-05, -8.949125415785e-06, {   1,   1,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.959908249755e-05,  8.945544457810e-06, {   1,  -1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.055761068012e-04,  2.967739734468e-06, {   4,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.708600956670e-06, -1.068036412293e-04, {   2,  -2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  7.990785224952e-05,  2.853181943539e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  3.961345823165e-05,  6.881648034636e-05, {   2,   0,  -1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -1.040559521839e-04,  4.302992016826e-06, {   0,   2,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.739515627634e-07,  1.077507048885e-04, {   0,   0,   3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.573439927526e-05, -1.225256657582e-05, {   2,   0,  -1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  1.287726867795e-06, -1.065765613192e-04, {   2,   0,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.809087940334e-05,  9.731931163599e-06, {   0,   0,   2,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  5.984600524153e-06,  1.017990993363e-04, {   2,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.508943128676e-05,  9.267087939329e-05, {   1,  -1,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.243193606941e-05,  5.531302569117e-05, {   1,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  3.788635913814e-08, -1.075837150549e-04, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  2.859864903760e-05,  7.885876596494e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -5.869943337246e-06, -1.013933056723e-04, {   2,   0,  -1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -7.767308239435e-05, -2.957924841468e-05, {   0,   0,   1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -7.579647317333e-05, -3.117022418452e-05, {   2,   0,  -1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.213933095084e-05,  1.478422986576e-05, {   1,   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.653356639311e-05,  2.035625663505e-05, {   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.014865903861e-04, -5.374510880448e-06, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -7.142848520717e-05, -3.542456313159e-05, {   2,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  3.973892744857e-05,  6.707844686076e-05, {   0,   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -5.793095258710e-05, -4.878414774728e-05, {   2,   0,   0,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -6.252854108150e-05,  4.395109819003e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  3.660587640506e-06,  1.028061794733e-04, {   2,   0,  -1,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 { -1.037691314868e-04, -2.592727528280e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  6.320395007799e-06,  9.997505999974e-05, {   2,  -2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.336780549371e-05,  7.270893725350e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.590960534405e-05,  2.010281426252e-05, {   1,  -1,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.056566622089e-05,  4.534837511255e-05, {   1,   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  8.378431295767e-07,  1.050637624817e-04, {   3,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.100374559585e-05,  6.486482841198e-05, {   1,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.071986378512e-07,  1.053733393321e-04, {   4,  -2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.490985587159e-05,  1.062095576638e-05, {   2,   0,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.984331325051e-05, -7.553722725954e-05, {   2,   0,   0,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 { -4.451416935610e-05,  6.081550213130e-05, {   0,   0,   1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 { -5.660155138915e-05,  4.832279438229e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0 } },
 { -5.947889330221e-08, -1.048331888944e-04, {   2,   0,   1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.657774004972e-05, -4.829932440239e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 } },
 { -4.051988372488e-05, -6.410807651952e-05, {   1,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -3.760730125486e-05, -6.688680597567e-05, {   2,   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -6.408185858485e-05, -4.035119349193e-05, {   2,   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.034990088907e-04, -9.282604233483e-07, {   0,   0,   2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  8.470271524083e-05, -1.971460854214e-05, {   1,   1,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -9.321984888709e-05, -1.103431485861e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.633825582321e-05, -8.780489717383e-05, {   1,  -1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.129855377723e-05,  2.263575914080e-05, {   4,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  4.418186550525e-05,  5.971031979492e-05, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.668764827790e-06, -9.620737024720e-05, {   2,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.979971920266e-05, -4.385006721415e-05, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.354261927440e-05,  7.003797451685e-05, {   0,   0,   1,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 {  1.028670441192e-04,  6.468335268600e-07, {   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.270830722355e-08, -1.034369469658e-04, {   0,   0,   1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  8.400474172179e-05, -1.945742745598e-05, {   1,   1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.313436092431e-05,  2.027939017235e-05, {   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.082943264161e-05, -2.250537548598e-05, {   4,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.017830701621e-04,  1.531861854750e-06, {   4,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.567353065872e-05, -8.762554688443e-05, {   2,   0,   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -9.948750905956e-05,  3.768130928526e-06, {   2,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.656798370088e-05, -6.662481462892e-05, {   2,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  5.599949031451e-05,  4.715860327074e-05, {   2,   0,  -2,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -1.861837005290e-07, -1.029430183799e-04, {   2,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.637511664620e-07, -1.028660288352e-04, {   2,   0,   1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.387232430279e-06,  1.016203912967e-04, {   2,  -2,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  8.585323167274e-05, -1.713616435808e-05, {   2,   0,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.885278982702e-05, -8.407636970394e-05, {   1,   1,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  9.972370756303e-05,  3.170598171801e-06, {   2,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.508111828494e-05,  7.777038838639e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 { -1.997019341913e-06,  1.007111108077e-04, {   4,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.571843271924e-05, -2.688468583878e-05, {   4,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  7.570199399349e-05,  2.687905291935e-05, {   4,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -8.246757955280e-05,  2.008667365594e-05, {   0,   0,   1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  7.985641544565e-08,  1.023909028577e-04, {   1,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.909404667726e-05, -3.326701008658e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -6.908331287437e-05,  3.326086540137e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -9.816065124857e-05, -4.066036839567e-06, {   4,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.157674772484e-07,  1.017811172870e-04, {   0,   0,   1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  3.470967775727e-07,  1.017961359206e-04, {   1,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.044172798278e-05, -4.166592145681e-05, {   0,   0,   2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.940181463337e-05, -8.260649614337e-05, {   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.805121805711e-05, -5.395151333574e-05, {   2,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  5.814471486050e-05, -4.345036139593e-05, {   3,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.531258634010e-07,  1.006250788973e-04, {   2,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.550200072872e-06,  9.590970037792e-05, {   2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.825634741336e-07, -1.005680077583e-04, {   0,   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  6.825690075934e-05, -3.306414552916e-05, {   2,   0,  -2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -8.654512912759e-05,  1.476338128760e-05, {   2,   0,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  9.214605899386e-05, -9.148628543914e-06, {   0,   0,   0,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 {  8.204217681070e-05, -1.904914629389e-05, {   1,   1,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.776807059150e-05,  8.332237867999e-05, {   4,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.968801032013e-05, -4.137861427356e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  6.771331417678e-05, -3.320137892505e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.006961871071e-04, -7.216744936627e-08, {   2,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  3.341467383720e-05, -6.730776641289e-05, {   0,   0,   1,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.409190625234e-05, -6.551511769075e-06, {   2,   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  5.631872739991e-05,  4.428916556045e-05, {   2,   0,  -3,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  3.247542638070e-05, -6.803467845584e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  8.814324734779e-05,  1.235584923922e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  5.334696010576e-05,  4.709642168362e-05, {   2,   0,   0,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 { -5.145801580572e-05,  4.885092951127e-05, {   2,   0,   1,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -5.145801580572e-05, -4.885092951127e-05, {   2,   0,  -1,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -6.027157408534e-06,  9.411876124025e-05, {   2,   0,  -1,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 { -1.172337390025e-07, -9.999993128123e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.267550365546e-08, -9.999997359135e-05, {   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.559537327241e-05,  4.460556120041e-06, {   2,   0,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.234021442357e-09,  9.999999997505e-05, {   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.199114857339e-09, -9.999999997582e-05, {   0,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.850049007000e-09,  9.999999998289e-05, {   2,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.675516081847e-09,  9.999999998596e-05, {   4,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.413716694071e-09,  9.999999999001e-05, {   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  5.934119456645e-10,  9.999999999824e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  3.376753319406e-05,  6.601139739328e-05, {   0,   0,   1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.688230824119e-08, -9.970347944102e-05, {   1,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.878471075133e-05, -2.072369788240e-05, {   1,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.075806786038e-06, -9.730928665634e-05, {   4,  -4,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  2.075806786038e-06,  9.730928665634e-05, {   4,  -4,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -7.679542915735e-05,  2.258870035516e-05, {   3,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -9.328462107286e-05,  6.043032767000e-06, {   2,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.952920823932e-05, -1.970828833288e-05, {   1,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.205889378694e-05,  6.716359073472e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  6.086853924065e-05,  3.825894545933e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  6.365984431287e-05,  3.544452574006e-05, {   2,   0,  -3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.092937853611e-05,  3.814436334589e-05, {   2,  -2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.653643291892e-05, -2.251313242110e-05, {   3,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.511509721843e-05,  6.391599459634e-05, {   3,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  3.511509721843e-05, -6.391599459634e-05, {   1,   0,   1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -1.563062148218e-06,  9.745932106622e-05, {   3,   0,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -7.349243043117e-07,  9.807751354786e-05, {   4,  -2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.188065913225e-05, -1.688280604142e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  2.247464055626e-05, -7.625492628021e-05, {   4,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.509111057419e-05,  6.349937052945e-05, {   4,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  6.057578885770e-05, -3.791920663474e-05, {   2,  -2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.562068903066e-06,  9.087639449740e-05, {   1,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.955394358404e-05, -1.879138434078e-05, {   1,   1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.874743356285e-05,  1.955750927149e-05, {   1,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.949737827812e-05, -1.879378655603e-05, {   1,  -1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.445146387306e-05,  6.382516002170e-05, {   2,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.097717447155e-06,  9.708804458690e-05, {   4,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.163880163352e-05,  7.650155240321e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -7.903046317696e-05, -1.907437403309e-05, {   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.008330690551e-05, -7.979646940283e-06, {   2,  -2,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -9.393507964777e-05,  4.000514791230e-06, {   0,   2,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.234137154278e-05, -6.557035574534e-05, {   2,   0,  -3,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.311168602267e-05, -1.477936063766e-05, {   2,   0,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.290642601146e-05, -3.498004392766e-05, {   2,   0,  -3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.258875841371e-05, -2.525120691889e-05, {   2,   0,   0,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 { -8.096046556449e-06,  8.963511605479e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.963886082859e-05,  8.054478838096e-06, {   3,  -1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.451991520921e-05,  3.313994529388e-05, {   2,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -3.572350922668e-07, -9.724517449780e-05, {   2,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  8.237155559473e-05,  1.519507243522e-05, {   1,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.466444255200e-05,  2.288248666024e-05, {   2,   0,   0,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 { -2.330043871455e-05, -7.412151478619e-05, {   2,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.240508470123e-05,  3.500070560582e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 { -7.908565424935e-05,  1.829867411293e-05, {   3,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.354513427588e-05,  3.375799232742e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 { -1.470218220717e-05,  8.252689480249e-05, {   2,  -2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.230374586235e-05,  8.484191417065e-05, {   0,   0,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  6.489491502868e-05, -3.214928850324e-05, {   2,   0,  -1,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 { -5.594218400004e-05, -4.104335399455e-05, {   3,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.059722878800e-05, -8.627924921308e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 { -4.053708809292e-05,  5.630908720372e-05, {   0,   0,   2,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  8.472985045423e-05,  1.208455723721e-05, {   2,   0,   0,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -4.050597622597e-05, -5.626271751644e-05, {   0,   0,   2,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  9.238200763934e-05,  4.274728447902e-06, {   2,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -8.566751356365e-07, -9.572486690453e-05, {   3,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.811388536326e-05,  1.842249409411e-05, {   1,   1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.701108541782e-05,  5.949841014628e-05, {   0,   0,   1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.684546067542e-06, -9.477790148660e-05, {   2,  -2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  6.052521271382e-05,  3.589132210571e-05, {   1,   1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.714133950871e-05, -7.925185715871e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 {  7.746317781793e-05, -1.889789439591e-05, {   2,   0,  -1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.167585011714e-05, -4.636034967147e-06, {   2,   0,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.089588737991e-05,  5.350581888140e-06, {   2,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 { -2.203997811075e-07,  9.591280843742e-05, {   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.401965534885e-05, -2.092910132940e-06, {   0,   0,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.507249538069e-05,  5.091533932842e-05, {   2,   0,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -5.489270909892e-05,  4.093332911853e-05, {   2,   0,  -1,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 {  7.752406614220e-05, -1.830104958782e-05, {   1,   1,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.114497066549e-06, -9.369995952698e-05, {   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.814453031261e-08,  9.567652736200e-05, {   2,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.418305353832e-05, -6.153072978821e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -5,   0,   0,   0 } },
 {  8.067421725891e-05,  1.500283514812e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -7.552358605494e-05,  2.013603127536e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -1.442200286028e-05,  8.121517930103e-05, {   2,   2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.115939620830e-05, -6.420944005094e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0 } },
 { -5.480576894284e-05,  4.048781443816e-05, {   2,   0,   0,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -1.510753219039e-05, -8.014115573266e-05, {   1,   1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.486227432095e-05,  1.036517489558e-05, {   0,   0,   1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  9.080881082551e-05,  4.228481545442e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -9.241848220891e-05, -2.586764779433e-06, {   2,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  8.516610352254e-05, -9.708210862495e-06, {   2,   0,  -1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.718476719578e-05,  1.767533137992e-05, {   1,   1,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.194750561552e-05, -7.291246254550e-05, {   2,   0,  -1,   0,   0,   3,  -3,   0,  -1,   0,   0,   0,   0 } },
 { -9.441917979118e-05,  3.664831816607e-07, {   4,   0,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.478111547190e-05,  6.999325864610e-05, {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  5.909999198841e-05,  3.564540591249e-05, {   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -7.687230666506e-05,  1.771792743426e-05, {   1,   1,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.188611748170e-05, -3.268549809433e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 { -1.885832296577e-05, -7.568834636882e-05, {   2,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.690063903049e-05,  4.753917739503e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 {  5.256704941880e-05,  4.184132705521e-05, {   0,   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  2.693375178478e-05, -6.741528069793e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -2.693375178478e-05,  6.741528069793e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -5.626403424957e-05,  3.791399744851e-05, {   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.096195112018e-05, -8.317219986260e-05, {   1,   1,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -4.194405740479e-06,  8.991153890552e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.038300910982e-05, -8.363570194031e-05, {   0,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  6.170479961449e-05, -3.220663892529e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -7.618410484020e-05,  1.762534230606e-05, {   4,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.528789071179e-07,  9.355514590764e-05, {   4,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.626883499903e-07, -9.364010452545e-05, {   0,   0,   3,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  8.709404317007e-05, -6.593510252650e-06, {   2,   0,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.095217621810e-06,  8.939043354653e-05, {   0,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  6.936088123817e-05, -2.397752361989e-05, {   2,   0,  -1,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.564127161475e-05,  1.766549432856e-05, {   1,  -1,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.902882474923e-05,  3.427563071730e-05, {   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.265888271738e-05, -1.061265452660e-05, {   2,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  4.503446524482e-06,  8.859511640037e-05, {   0,   0,   3,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  4.503446524482e-06, -8.859511640037e-05, {   0,   0,   1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -9.128532968475e-05, -1.805768146265e-06, {   0,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.754692776400e-05, -1.550699133020e-05, {   1,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.627196667735e-07,  9.287577117030e-05, {   0,   0,   1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  7.526763669602e-05, -1.775224167155e-05, {   1,  -1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.841859863749e-05, -1.457968817229e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.096857643330e-05,  6.194799197405e-05, {   0,   0,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -7.519252114580e-05, -1.740210808598e-05, {   1,   1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.625484872881e-06, -8.996404957385e-05, {   2,  -2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.942761790000e-06, -9.056790475412e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 {  7.471757030792e-05,  1.779189359154e-05, {   2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.859751806373e-05, -6.389195536681e-05, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.646649708606e-07,  9.161472943383e-05, {   4,   0,  -4,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.251699143084e-05, -5.993649661097e-05, {   2,   0,   1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -5.302504908480e-05, -3.940464830536e-05, {   2,   0,  -1,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 {  2.750409682852e-05, -6.484307466120e-05, {   4,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  3.838089716670e-05, -5.395739253727e-05, {   2,   0,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 {  3.774570712438e-05, -5.454279829714e-05, {   2,   0,  -1,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 { -3.870382386134e-06, -8.834022993810e-05, {   2,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.825936854978e-05,  3.392008370661e-05, {   4,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.794099074497e-05,  7.422605258907e-05, {   1,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.154028743846e-05, -5.467832426101e-07, {   1,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.763348002489e-05,  4.442089803192e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 { -4.100576050720e-05, -5.094279547524e-05, {   2,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -8.228789793969e-05, -9.605252808956e-06, {   2,   0,  -1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  4.321502203777e-05, -4.860883782772e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  2.550941080734e-06,  8.917664442357e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 {  3.708549083666e-05, -5.456424292689e-05, {   2,   0,   0,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 {  3.746112539536e-05,  5.417415096132e-05, {   2,   0,  -1,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 {  3.124960077603e-05,  6.030603198012e-05, {   2,   0,  -1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.976854654711e-05,  6.172788126120e-05, {   2,   0,   0,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.226027045883e-05,  9.180733441091e-06, {   4,   0,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.034779422413e-05,  5.108626980782e-05, {   2,  -2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -6.482801447310e-08, -9.135073842181e-05, {   2,   0,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.421504355062e-05, -1.692498160714e-05, {   0,   0,   2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -8.246485938209e-05,  8.570038530800e-06, {   2,   0,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.091452349455e-05,  1.104561826668e-07, {   2,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.937614065628e-05, -5.159099867850e-05, {   2,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -6.740203792773e-05,  2.355116797640e-05, {   2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.756047141558e-05,  3.339067091274e-05, {   1,  -1,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -5.853507890882e-05, -3.239114883163e-05, {   0,   0,   1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -2.187329520760e-05,  6.902811362947e-05, {   0,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -6.154734922891e-05,  2.931990794159e-05, {   2,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.002997386597e-06, -8.779792390574e-05, {   4,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.446049319584e-05, -7.630533372453e-05, {   0,   0,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.694326790242e-05,  7.375978110645e-05, {   2,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.225854211197e-05, -3.843866697310e-05, {   2,   0,   0,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 {  7.482982300011e-05,  1.586159407294e-05, {   2,   0,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.667050253712e-05,  5.400936574678e-05, {   2,   0,   0,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -6.165098920490e-08,  9.061381982054e-05, {   2,   0,  -2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.439650190445e-05, -3.617523057494e-05, {   2,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.203360077774e-05,  4.836796285203e-05, {   2,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -8.717739947986e-05, -3.139403972148e-06, {   0,   0,   1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  8.604968943247e-05,  4.144457705297e-06, {   2,   0,  -1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -1.760738649841e-07,  8.999977833700e-05, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.377927756889e-05,  7.628626485024e-05, {   1,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -6.962707459466e-05,  2.043028140073e-05, {   0,   0,   2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.950177432697e-05, -6.054220501140e-05, {   2,   0,  -1,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.901010278661e-05, -4.102806611926e-05, {   1,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.858849314278e-09,  8.999999995459e-05, {   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.729646829523e-05,  3.259781660713e-05, {   1,   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -8.722688209106e-05,  2.592971557266e-06, {   2,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -7.187178179716e-05,  1.793768205825e-05, {   2,  -2,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.887670272821e-05, -5.089969403107e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 {  7.252969867933e-05,  1.715991636589e-05, {   1,  -1,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.592491016430e-06, -8.303585195190e-05, {   2,   0,  -1,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.030754445069e-05, -3.927252258837e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  1.694660683044e-06,  8.779759453631e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  4.240545716276e-05, -4.687492997493e-05, {   2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -2.541329764531e-05, -6.382590642163e-05, {   2,   0,  -1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -1.756796276176e-05,  7.164036372752e-05, {   4,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.645100183799e-07, -8.830120841585e-05, {   1,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.624360038328e-05, -5.281068334638e-05, {   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -5.405309508353e-06, -8.364387876459e-05, {   1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.746820739591e-05, -3.153006161518e-05, {   0,   0,   2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  4.569447218796e-05,  4.322006360778e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  7.622622706427e-06, -8.128907848294e-05, {   2,   0,   0,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  4.160977429230e-05, -4.725867964245e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 { -7.107548317751e-05,  1.772600427976e-05, {   1,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.729878536040e-05, -5.145412693718e-05, {   2,   0,   0,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 { -2.049727030517e-05,  6.824886144699e-05, {   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.108965450076e-06,  8.562146307710e-05, {   2,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.566010932345e-05, -2.305114516384e-05, {   1,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.191321080927e-05, -5.677533250199e-05, {   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.959907871931e-08,  8.859979309872e-05, {   4,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.652579296657e-05,  6.214809053139e-05, {   4,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -2.379384438509e-05,  6.480844751282e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  8.548296047706e-05,  3.085212614886e-06, {   0,   0,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -7.199964783617e-05, -1.653895018719e-05, {   1,  -1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.505273029856e-05, -7.322918994285e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.835311016952e-05, -5.990531056541e-05, {   2,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 { -8.476759834287e-05, -3.414081895343e-06, {   2,   0,  -1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  6.889158360579e-05, -1.912527866422e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,  10,   0,   0,   0 } },
 { -8.440265319721e-05, -3.385013744377e-06, {   2,   0,  -1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.803772259446e-06, -8.482075415039e-05, {   4,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -7.071058030546e-05,  1.689562066679e-05, {   4,   0,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.963307896085e-05, -1.795576841450e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -6.940643027600e-05,  1.816161284340e-05, {   1,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.694363873511e-05, -1.054171865115e-05, {   4,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.925138443923e-05,  2.816881321468e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0 } },
 {  5.471574467431e-05,  3.257157058477e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -8.162731410433e-05, -5.449039431256e-06, {   2,   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -6.967234357342e-05, -1.736417304633e-05, {   4,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  3.254679981568e-05, -5.446943792000e-05, {   1,   1,   0,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.205359091165e-05, -6.494591990608e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -3.063826211133e-05,  5.630429010379e-05, {   2,   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  8.308988118742e-05, -3.826709689575e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.468528955558e-05, -7.220798496986e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 {  7.967928098627e-05, -7.156268686368e-06, {   3,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.475099043603e-05,  7.206022696422e-05, {   2,   0,   0,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  6.904911102343e-05, -1.764889428040e-05, {   3,  -1,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.125923319003e-08, -8.661113700895e-05, {   2,  -2,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.906428589353e-05, -2.757470920530e-05, {   0,   0,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  4.439822003450e-06,  8.219881129140e-05, {   2,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.582396165071e-05,  4.079139266673e-05, {   2,   0,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -1.877611393087e-07,  8.642410019057e-05, {   6,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -6.463512948832e-05, -2.196489223610e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  7.001834338366e-05, -1.656933300203e-05, {   1,  -1,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.997470529620e-05, -1.643190291504e-05, {   2,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.015608593534e-07, -8.579371006470e-05, {   1,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.588104763306e-05, -7.042245449874e-05, {   1,  -1,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  3.680970767830e-05,  4.945892644463e-05, {   2,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0 } },
 {  8.252548999349e-05, -3.657652959507e-06, {   3,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.744907805838e-05, -2.872263362113e-05, {   2,   0,  -1,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 {  6.854042870408e-05, -1.762555117917e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.574678877030e-05, -7.040259632192e-05, {   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.012248727858e-05,  5.596638356875e-05, {   2,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  5.764273805918e-05, -2.843858008112e-05, {   2,   0,   0,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 { -1.320726779253e-05,  7.283708360112e-05, {   4,  -2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  5.237134214304e-05, -3.352405692667e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  2.585545329777e-05,  6.003551333015e-05, {   2,   0,  -1,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.285608369080e-05, -5.302592767355e-05, {   2,   0,   0,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 {  5.720370134708e-05,  2.854230413302e-05, {   2,   0,  -1,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 {  6.016347334458e-05, -2.555141947393e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  6.015663044569e-05,  2.554601008806e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  8.166630851485e-05,  3.865734698528e-06, {   2,   0,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.725859874912e-05,  2.818805828875e-05, {   2,   0,   0,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 { -1.945780541353e-07,  8.521909108948e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  6.311716996050e-05,  2.223472689542e-05, {   1,  -1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.341549448941e-05, -6.193172556444e-05, {   2,   0,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.262386474030e-05, -5.257678858513e-05, {   0,   0,   1,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 { -5.487903888379e-05,  3.023361633789e-05, {   0,   0,   2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -6.707855154327e-05, -1.801599245839e-05, {   2,   0,  -1,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.805738775710e-05, -5.697538256321e-05, {   4,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -8.534159714069e-07, -8.416127252430e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.339778210992e-05,  6.153998145800e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.837089005433e-05, -1.650316061924e-05, {   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.412948633808e-05,  6.060999774164e-05, {   2,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -3.895224001500e-05,  4.576781069167e-05, {   0,   0,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.847698294950e-05, -1.623817541392e-05, {   1,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.182170334597e-07, -8.453427373091e-05, {   0,   0,   1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.503359419617e-05,  3.961586899355e-05, {   4,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.749951623841e-05,  6.710829385730e-05, {   1,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -8.209139183197e-05,  2.497481545784e-06, {   0,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.758869671353e-05,  6.697946940516e-05, {   4,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -9.524230007788e-07,  8.361285471077e-05, {   3,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.423835167628e-06, -8.014083164729e-05, {   1,   1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.811595712655e-05, -5.629423862609e-05, {   2,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.993521152298e-05,  4.404712107666e-06, {   2,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.113453835326e-05,  2.311144454489e-05, {   1,   1,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.751832032483e-05, -4.654814490432e-05, {   2,  -2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.143418508918e-05, -6.262136340816e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  1.354192338911e-05,  7.050080619220e-05, {   2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.088033348526e-05,  6.309852995213e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  7.989853446157e-05, -4.027926376414e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.837099042960e-05,  4.546676037095e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0 } },
 {  6.726667072673e-05, -1.653579835529e-05, {   2,  -2,   1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.710194971534e-05,  5.669466841597e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -6.811968865420e-05, -1.564178429420e-05, {   2,   0,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.396749011439e-05, -5.975227219511e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.540332152927e-05,  6.828424185613e-05, {   2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.495274125311e-05,  5.872630620125e-05, {   4,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.495274125311e-05, -5.872630620125e-05, {   4,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.895451429967e-05, -4.715768622237e-06, {   2,   0,  -3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.205180165275e-05, -1.452743377360e-06, {   3,  -2,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  8.205180165275e-05,  1.452743377360e-06, {   3,  -2,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  4.183423198157e-06,  7.926878889722e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -5.662873440449e-07, -8.281490929141e-05, {   2,   0,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  5.642795391925e-05,  2.689191012808e-05, {   2,   0,  -2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.051208433427e-05,  7.273635803040e-05, {   2,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.716335541162e-05, -2.595799611035e-05, {   0,   0,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -6.708823843680e-05,  1.601883644630e-05, {   1,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.601145594757e-07,  8.229056363279e-05, {   2,   0,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -4.331593257941e-05, -3.972881675984e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -5.922734956861e-05, -2.378535223330e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -8.083480664333e-05, -2.152871217005e-06, {   2,  -2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.473682095017e-05, -4.822252112410e-05, {   2,   0,  -1,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 { -7.074968815639e-05,  1.216925894315e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 {  2.788710037866e-05,  5.501022707033e-05, {   0,   0,   1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.506490320553e-05, -7.814471586040e-06, {   2,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  2.757187989459e-05,  5.524672922109e-05, {   0,   0,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.631824892974e-05,  1.648218080071e-05, {   1,   1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.598757866791e-05, -6.764394842215e-06, {   4,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.986105695866e-05,  1.272220986272e-05, {   2,   0,   0,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 { -4.497509974868e-07, -8.199629120121e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -4.371477000770e-05,  3.867458703308e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 {  4.371477000770e-05, -3.867458703308e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 { -6.633420269730e-05,  1.598387676798e-05, {   1,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.012711647584e-07, -8.181236235632e-05, {   1,   0,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.737142593498e-05,  5.488545470345e-05, {   1,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.734791286462e-05,  5.475555961889e-05, {   2,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.922523097536e-05,  2.286433942597e-05, {   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -6.683530978414e-05,  1.521286841268e-05, {   1,  -1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.472369518458e-08, -8.197062540064e-05, {   2,   0,  -2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  9.487222120583e-07, -8.102670108988e-05, {   2,   0,  -2,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.469768420561e-05,  1.726255478793e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 {  1.207953951207e-05,  6.986878232025e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  7.859167711166e-05,  3.335518850998e-06, {   2,   0,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.701191543613e-05, -6.491001267876e-05, {   0,   0,   0,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 {  1.313443056317e-05,  6.873877940461e-05, {   0,   0,   2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.327237742465e-05,  3.858345111398e-05, {   1,   0,   0,   0,   0,   0,  33, -41,   2,   0,   0,   0,   0 } },
 {  5.856291227265e-05, -2.328619885915e-05, {   1,  -1,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -7.564496780300e-05, -6.193446080475e-06, {   2,   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -8.139340619962e-05,  3.495886387440e-07, {   2,   0,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  7.900496264568e-05,  2.710894744012e-06, {   2,   0,  -2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.179132891898e-05, -3.991871224027e-05, {   0,   0,   1,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 {  6.500767161552e-05, -1.669932170125e-05, {   0,   2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.500356374643e-05,  1.669675948647e-05, {   0,   2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.437529291333e-05,  4.731922899722e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 { -4.954692692922e-05,  3.214370530422e-05, {   1,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  6.440781700040e-05, -1.728001316956e-05, {   1,  -1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.488261191179e-07, -8.088176137796e-05, {   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.209259103275e-05,  5.943301227174e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 { -2.382278821524e-06, -7.913305205729e-05, {   1,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.510653000703e-05,  6.640538438685e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  1.263308365686e-05,  6.885060055888e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.732002690039e-06,  7.973803809282e-05, {   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.998202881773e-05, -5.135564027397e-05, {   2,   0,  -1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -2.458521496223e-05,  5.669712918230e-05, {   1,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -4.303404850367e-05, -3.817420640439e-05, {   2,   0,   1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.583231829097e-05,  1.535045760792e-05, {   1,  -1,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.915417725437e-08,  8.112310493632e-05, {   2,  -2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.600310430764e-05, -5.502259821569e-05, {   2,   0,  -1,   0,   0,  -7,  11,   0,   0,   0,   0,   0,   0 } },
 {  4.196696045260e-05, -3.892331234731e-05, {   2,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.694127223515e-05,  5.394245628070e-05, {   6,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.046189733104e-05,  3.612396390674e-07, {   1,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.966878979800e-05,  4.112375402230e-05, {   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.481121128028e-05,  1.594723290453e-05, {   2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.190177134922e-05, -6.884005244141e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.441405427429e-05,  1.631484407477e-05, {   0,   0,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -8.403335124410e-09,  8.059037901326e-05, {   0,   0,   2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.868476801291e-05, -5.190224093371e-05, {   2,   0,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -6.027962848387e-05,  2.030078142401e-05, {   2,   0,   0,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 {  9.650975345682e-06,  7.090767992889e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -2.914329592281e-05,  5.126704485210e-05, {   2,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 {  2.659447152526e-05, -5.374253916877e-05, {   0,   0,   1,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.033048876245e-05, -4.993466683426e-05, {   2,   0,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  7.486457299696e-05,  5.191680910382e-06, {   2,   0,  -2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -4.148263499357e-05, -3.856657308552e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -5.540538570468e-05,  2.463333973500e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  4.035201228871e-09, -7.999999989823e-05, {   2,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.284562329404e-09,  7.999999998969e-05, {   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.075712110080e-10,  7.999999999485e-05, {   4,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.866787695960e-05, -5.129541931478e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  1.285106714714e-05, -6.708066571601e-05, {   1,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.084645056845e-07,  7.975609626017e-05, {   1,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.840302103177e-06,  7.601447645996e-05, {   0,   0,   1,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 { -1.468230398976e-05, -6.515741058451e-05, {   1,   1,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  5.138948670756e-05,  2.839324757947e-05, {   2,   0,  -1,   0,   0,   0,   8, -12,   2,   0,   0,   0,   0 } },
 {  7.714713993838e-05, -2.601222645436e-06, {   2,   0,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -1.335710134123e-05, -6.623253603508e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 { -5.346400204403e-06, -7.421731120093e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  1.274714599016e-05, -6.681452937761e-05, {   4,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.471071054338e-05,  2.463361120820e-05, {   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -2.515100850354e-05, -5.412480502230e-05, {   2,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  2.449821211129e-05,  5.477077325865e-05, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.605132199862e-05,  2.318304479532e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.048744638903e-05, -1.871205485079e-05, {   2,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.568298410614e-08, -7.911774391795e-05, {   2,   0,   1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.935159025967e-05, -5.979402268897e-05, {   2,   0,   0,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -2.215726875760e-05, -5.697023655914e-05, {   2,   0,  -1,   0,   0,   0,  -3,   9,   0,   0,   0,   0,   0 } },
 {  1.153076100017e-06,  7.794448390686e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  1.919885843538e-05,  5.971640922493e-05, {   2,   0,   0,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 { -6.227636420285e-05, -1.662503870018e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  8.854016100180e-06,  7.004584972710e-05, {   1,  -1,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.492796179160e-06, -7.140139977415e-05, {   2,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  6.472456829127e-05,  1.411745554921e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  6.284861913880e-05,  1.578470679239e-05, {   1,   1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.056531280093e-05,  4.789954698376e-05, {   2,   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -2.784177374574e-08, -7.842947721687e-05, {   0,   0,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  6.244634368461e-05, -1.601042878057e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -5.432482086968e-05, -2.408798522684e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 {  7.775773200480e-05,  6.089523192605e-07, {   1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.178918903165e-05,  5.651837590653e-05, {   2,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.704639968482e-05,  1.124567627528e-05, {   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.308964097881e-06, -7.696356983990e-05, {   2,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  8.037621213938e-06,  7.022357531713e-05, {   0,   0,   1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  2.606062397355e-05, -5.217818947989e-05, {   6,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.971096670179e-05,  4.848176805508e-05, {   2,   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -5.469064236454e-05,  2.348714039835e-05, {   0,   0,   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -3.150929795903e-07,  7.780791307867e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  7.282867140262e-05,  5.206167383999e-06, {   2,   0,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  6.983950811601e-05, -8.194519925797e-06, {   0,   0,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.245599714747e-05, -2.553526906161e-05, {   0,   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.619711530343e-05,  5.178964683430e-05, {   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  6.226715733397e-05,  1.560484067349e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.281321647533e-05, -4.503677198578e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   3,   0,   0,   0 } },
 { -5.740913467831e-05,  2.037628584225e-05, {   1,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  6.406035819092e-05,  1.371431704013e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.070032590096e-05, -6.705673843008e-05, {   0,   0,   2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.737565747560e-05, -2.037994374118e-05, {   1,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -4.808803185732e-05, -2.957728741554e-05, {   2,   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 {  6.222050579848e-05,  1.536586721608e-05, {   1,  -1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.131455491284e-05,  5.626474868228e-05, {   0,   0,   2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  2.093459253187e-05, -5.660228758949e-05, {   4,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  4.861551701095e-05, -2.885111646509e-05, {   4,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.333877213216e-05, -4.410087152850e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 {  8.814139997604e-06,  6.859390133260e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -7.725091494109e-05,  1.039648477265e-07, {   1,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.044521444575e-05,  4.688698609560e-05, {   0,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -5.087561560190e-05,  2.642613544580e-05, {   2,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.227857787059e-07, -7.717692002540e-05, {   2,   0,  -1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.552286649697e-05,  6.175699780809e-05, {   2,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.776678375930e-05,  4.938670010349e-05, {   0,   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.847320195871e-05, -5.866024653185e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 {  6.161662382098e-05, -1.545665245421e-05, {   1,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.233252578295e-05, -1.472189454282e-05, {   1,  -1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.796084992547e-06, -6.818156671270e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 {  8.928138881080e-08, -7.679595784023e-05, {   4,   0,  -2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.422695390799e-05, -5.265446997046e-05, {   1,   1,  -2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.701018319379e-05,  4.975864996879e-05, {   2,   0,  -2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -2.287553096628e-05,  5.384979224731e-05, {   4,   0,  -1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.287553096628e-05, -5.384979224731e-05, {   4,   0,  -3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.391702299653e-05,  6.276150508378e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 {  6.161080869928e-05,  1.506566868215e-05, {   1,  -1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.493444829303e-05,  5.173228141529e-05, {   2,   0,  -1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.907022717894e-05,  7.560843720627e-06, {   1,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.891012292269e-05,  2.771925567365e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 { -4.898353203994e-05,  2.764224102763e-05, {   0,   2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.838461413970e-09,  7.655937857763e-05, {   1,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.011194870463e-05,  5.640401462409e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -4.903039945601e-06, -7.160167639099e-05, {   2,  -2,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  6.181716421365e-05,  1.464482048396e-05, {   1,  -1,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.155790647151e-05, -1.478326665322e-05, {   1,  -1,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.631155609526e-05,  1.997288991938e-05, {   2,   0,  -1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 {  6.171852344013e-05,  1.453736987417e-05, {   1,   1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.924451484484e-05, -4.693121423995e-05, {   2,   0,  -1,   0,   0,   3, -18,  25,  -2,   0,   0,   0,   0 } },
 {  2.925079906322e-05,  4.692372033955e-05, {   2,   0,  -1,   0,   0,  -3,  18, -25,   2,   0,   0,   0,   0 } },
 { -3.953412890396e-05,  3.662261100281e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 {  6.096459523021e-05, -1.518462247356e-05, {   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.078314838764e-05, -1.527066837432e-05, {   2,   0,   0,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 {  9.014284973716e-06,  6.702403836470e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 {  6.971909320322e-05,  6.264825849156e-06, {   3,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.971909320322e-05, -6.264825849156e-06, {   1,   1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.971922221181e-05,  6.263389991108e-06, {   3,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.971935774828e-05,  6.261881120467e-06, {   1,  -1,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.971945173021e-05,  6.260834643945e-06, {   1,  -1,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.116658556403e-05, -1.480119433836e-05, {   1,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.673132288369e-05, -4.923052105511e-05, {   2,   0,   0,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 { -7.232093465836e-05, -3.626013073081e-06, {   2,   0,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.157428905697e-05,  3.429445867536e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -2.959094305828e-05,  4.626534248728e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 {  7.235898920090e-05,  3.492072787732e-06, {   2,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.251861233955e-07,  7.570010652252e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 { -3.273658610209e-05, -4.299531444127e-05, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.861979251575e-05,  5.709870574430e-05, {   2,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.133801316305e-05,  1.434977990037e-05, {   3,  -1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.487747713296e-05,  5.078787319138e-05, {   2,   0,  -1,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.441968187500e-06,  7.020629917870e-05, {   0,   0,   1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -6.888754118046e-06, -6.867350022902e-05, {   2,   0,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 {  3.093362028896e-05, -4.462533824235e-05, {   2,   0,  -1,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 { -2.251402837927e-05,  5.298090169824e-05, {   2,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.251402837927e-05, -5.298090169824e-05, {   2,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.181150660919e-05, -4.350693891849e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 {  6.055490434264e-05,  1.474399452230e-05, {   1,   1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.440921089067e-05,  5.067342468346e-05, {   2,   0,   0,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.075666444099e-05, -2.427493149755e-05, {   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.029082765752e-05, -1.473105743645e-05, {   4,   0,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.707487576091e-06,  7.330562480098e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -2.632134358361e-05, -4.867537720932e-05, {   4,   0,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  4.695053298589e-05,  2.804505701804e-05, {   0,   0,   1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -6.158716031556e-05,  1.334576091238e-05, {   2,   0,   0,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -8.540682515536e-07, -7.403954126391e-05, {   0,   0,   2,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.050405107533e-05, -4.437455527585e-05, {   2,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.974704734267e-05,  1.502843470909e-05, {   2,   0,  -2,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.382409261903e-05, -8.671736661796e-07, {   2,  -2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.563282292381e-05, -5.896898466311e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 { -3.594371299301e-08, -7.456392845596e-05, {   4,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.492270958153e-05, -5.967304219046e-05, {   3,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.483580889521e-05, -4.972869413296e-05, {   2,   2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.190588285399e-05, -3.265646940318e-05, {   2,   0,   0,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  2.212215346579e-05, -5.237395118616e-05, {   2,   0,  -1,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 {  2.868086803346e-07, -7.417509312842e-05, {   3,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.844709819005e-06,  6.460644475590e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 {  2.201023175897e-05,  5.235831865101e-05, {   2,   0,  -1,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 {  1.853179616417e-07, -7.414115531470e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.646046249186e-05,  1.781591291905e-05, {   4,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.658041227157e-05, -5.761488697101e-05, {   1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.231138610303e-05, -5.181060208428e-05, {   1,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  4.858592098201e-07,  7.361017560234e-05, {   2,   0,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.896337709030e-07, -7.370133735696e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.354696865181e-05, -1.051840831209e-05, {   2,   0,   0,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  1.190182673454e-05, -6.210789693098e-05, {   1,  -1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.989716891589e-05, -1.411073711912e-05, {   1,  -1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.087816002560e-05, -3.309367057141e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  3.927207001259e-07,  7.354558699753e-05, {   3,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.284463100157e-05, -3.108215592487e-05, {   2,   0,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.383949588861e-05,  4.005306632493e-05, {   2,   0,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 { -1.258716636748e-08,  7.386260142870e-05, {   0,   2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.301850474605e-05, -6.074384075478e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 {  6.039226734040e-05,  1.334427156759e-05, {   2,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.129891571884e-05,  6.237180311463e-05, {   2,  -2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.452167843299e-05, -4.909736930366e-05, {   2,   0,  -5,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.902171949081e-05, -4.541501117165e-06, {   2,  -2,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  7.077111620855e-05, -2.739642063310e-06, {   2,   0,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  6.506504372703e-05,  8.427479066328e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -6.935426712509e-05, -4.105885493277e-06, {   2,   0,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.269666127671e-05, -6.075842242314e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  4.750575199501e-05,  2.584239415557e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  1.040027609386e-05, -6.291520042699e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  1.254839980596e-05, -6.075935411859e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -4.368831240200e-08, -7.324021972855e-05, {   2,   0,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.238260292704e-05,  8.439200603171e-07, {   2,   0,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.540539853520e-05, -4.777762664224e-05, {   2,   0,   0,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.514792659303e-05, -5.799257653777e-05, {   5,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.332722987759e-05,  5.979847850154e-05, {   1,  -1,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -7.290185537559e-05,  2.202514090500e-07, {   4,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.835710849099e-05,  3.468582911561e-05, {   2,   0,   1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.285202756226e-05, -6.017710803130e-05, {   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.586818342123e-05, -3.709425926625e-05, {   2,  -2,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -4.716798408874e-06, -6.822996242152e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -6.841737625824e-05, -4.373626482791e-06, {   2,   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 {  5.812668752276e-05, -1.458540137634e-05, {   1,  -1,   1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.115629046998e-05,  1.147576094973e-05, {   2,  -2,   1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.865672605809e-05,  1.391608692412e-05, {   1,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  5.914313024331e-05,  1.342580621105e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  2.322966377862e-05, -4.930532935102e-05, {   2,   0,  -1,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.264214147915e-05,  5.977814406720e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  5.829509953682e-05,  1.410585287912e-05, {   4,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.816120516884e-05,  5.423703439455e-05, {   2,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.432152636930e-05,  1.806353921839e-05, {   1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.296026646688e-05, -5.941497697451e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  5.854007818081e-05, -1.381239078401e-05, {   1,   1,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.554412783589e-05,  1.679698852887e-05, {   1,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.636107532574e-05,  5.585327944784e-05, {   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.786130289876e-06, -7.036941028545e-05, {   2,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.262137788736e-05,  4.944596195598e-05, {   2,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  6.877445441650e-05,  3.272520823412e-06, {   2,   0,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.724802072489e-06,  6.826193665017e-05, {   6,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.497633778285e-05,  4.697683554461e-05, {   2,   0,  -2,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.244103629049e-06,  6.865912469118e-05, {   0,   0,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -1.321763278664e-05,  5.864216521209e-05, {   3,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.818034905372e-08, -7.179897129229e-05, {   4,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.015292488302e-05, -5.164190715241e-05, {   0,   0,   2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.583580104643e-05,  1.590749969504e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.290329366059e-05,  5.883871997467e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -5.478803752332e-05, -1.689831931709e-05, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.799141024314e-05, -1.368639704506e-05, {   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.880570751791e-05,  2.283213517172e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 { -5.577233717582e-05,  1.583122717264e-05, {   3,   0,   0,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 {  1.486439241865e-05, -5.672767529750e-05, {   5,   0,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  4.031495191264e-05, -3.127107275253e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.665062315340e-05,  5.490681473027e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  4.485343911720e-05,  2.663384451693e-05, {   2,   0,  -1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 {  1.012333188279e-06, -7.039835097400e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.252259379489e-05, -1.888619986137e-05, {   2,   0,   0,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 {  6.603403076860e-06, -6.478009497079e-05, {   2,   0,  -1,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 {  8.807770950536e-06,  6.257515168530e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 { -3.139777251187e-06, -6.823028811553e-05, {   2,   0,   1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -4.707360234907e-05,  2.423595839316e-05, {   2,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.985081365281e-05, -1.377714115025e-06, {   2,  -2,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.850140723454e-05,  2.262561206960e-05, {   2,   0,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 {  5.987174141894e-05, -1.123428362063e-05, {   2,   2,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.818354881909e-05,  2.291250369402e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 { -4.394590046103e-05, -2.712686036489e-05, {   2,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.813074077530e-05, -2.292458372506e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 { -3.999228096645e-05,  3.100381977285e-05, {   2,   0,   0,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 { -8.987561293324e-08,  7.088859668655e-05, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  1.034118841995e-06,  6.992970494010e-05, {   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.662247528298e-05, -4.427663640660e-05, {   4,   0,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -2.290956216939e-05, -4.795261819497e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  2.311980105407e-05, -4.769080492363e-05, {   2,   0,  -1,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.465191049917e-05, -6.054162738643e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.352853963252e-05,  4.710847384613e-05, {   2,   2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.352345154501e-05, -4.710683397360e-05, {   1,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.572176607340e-05, -4.489479291088e-05, {   2,   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 {  4.713268629003e-05, -2.347214646203e-05, {   2,   0,  -1,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 {  1.660253816926e-05,  5.391556393740e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  3.527578578007e-05, -3.520085310323e-05, {   2,   0,  -1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -5.696113051571e-05, -1.344791409802e-05, {   1,   1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.275401756013e-05,  4.762787160540e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -3.933636187109e-05, -3.086503903685e-05, {   2,   0,  -2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.863923493122e-05,  1.555995636287e-06, {   2,   0,   0,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.857503636822e-05, -5.154094320414e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  5.575762517789e-05,  1.434293649822e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -4.449193892433e-05, -2.555314144533e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -4.208861237587e-08,  6.999998734678e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.417207352521e-09,  6.999999998565e-05, {   2,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.380555438225e-09, -6.999999998639e-05, {   2,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.331686219163e-09, -6.999999998733e-05, {   2,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.136209342977e-09, -6.999999999078e-05, {   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -8.552113334496e-10, -6.999999999478e-05, {   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.764748857684e-10, -6.999999999838e-05, {   2,  -1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.653256990604e-05,  1.340588524283e-05, {   3,  -1,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.725198124698e-05, -4.265500878104e-05, {   4,   0,  -2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.461145996182e-05,  5.517517462778e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  3.351397042332e-06,  6.642601108372e-05, {   2,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.335084974695e-06,  6.843413803848e-05, {   3,   1,  -1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.335084974695e-06, -6.843413803848e-05, {   1,  -1,   1,   0,   1,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.065832938681e-05,  2.909939334069e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 { -5.784506410715e-05, -1.188850554392e-05, {   3,  -3,  -1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -5.784506410715e-05,  1.188850554392e-05, {   1,   3,  -1,   0,   0, -43,  42,   0,   0,   0,   0,   0,   0 } },
 {  3.269452479379e-05,  3.701814632164e-05, {   2,  -2,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.785898714383e-05, -3.184260963707e-05, {   3,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.710513147824e-05,  4.242319522184e-05, {   4,   0,  -2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.876188659600e-05,  5.073843652128e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 { -7.974724548247e-07, -6.868058040276e-05, {   2,   0,   2,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.000992728583e-05, -9.385307892540e-06, {   2,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -4.298543489730e-05, -2.637052856733e-05, {   4,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.325124039132e-05,  4.606035476052e-05, {   0,   0,   1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.075631759363e-08, -6.924929999363e-05, {   1,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.176267147928e-05,  5.747688031030e-05, {   1,  -1,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.486671394737e-05,  4.329679182122e-06, {   2,  -2,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -5.620616337598e-05,  1.296174285648e-05, {   2,   1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.657495984606e-05, -2.240710288658e-05, {   2,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.772184014502e-05,  5.121040008832e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 {  7.193220005497e-06, -6.170638817452e-05, {   0,   0,   1,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 { -5.803578253308e-05, -1.076674006770e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.648597078484e-06,  6.714053293675e-05, {   2,   0,  -1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 { -2.874110467694e-05,  3.995283820891e-05, {   2,   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  1.259340710499e-06, -6.742011436527e-05, {   4,   0,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -6.857192696531e-05, -6.552095845039e-08, {   2,   0,  -2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.531185122188e-05,  4.331531436616e-05, {   2,   0,   0,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 {  2.265656733831e-05,  4.595895642251e-05, {   4,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.004994545769e-05, -4.854494720426e-05, {   4,   0,  -2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -6.430122916561e-05, -4.289128987582e-06, {   2,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -4.606708535136e-05, -2.249947477122e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 { -2.020208850734e-09,  6.855290021730e-05, {   2,  -2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  5.361462974677e-05, -1.493816295775e-05, {   2,   0,  -3,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  5.360824950790e-05,  1.491885836005e-05, {   2,   0,  -3,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.469420326206e-06,  6.698278675878e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 { -6.606485073981e-05, -2.346519201612e-06, {   0,   0,   1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.783171483136e-05, -3.054138299774e-05, {   0,   0,   0,   0,   0,  15, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.110468797668e-05, -5.726047437258e-05, {   2,   0,   0,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  2.274189064430e-09,  6.830131767265e-05, {   4,  -2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.652629976596e-05,  1.177301254688e-05, {   2,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.307706700454e-05,  5.520586344981e-05, {   1,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.761904447726e-05, -2.066062516985e-05, {   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.001044469862e-06, -6.727615618945e-05, {   3,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.432568057149e-05,  1.394176628502e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  5.090471021632e-05, -1.734816786833e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 {  5.091113588552e-05,  1.731083160971e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -5.648199379906e-05, -1.172552033646e-05, {   2,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  2.255664353547e-05,  4.546901700076e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 { -9.265338583866e-06, -5.859115290108e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 { -4.614405934936e-05,  2.171199656803e-05, {   2,   0,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 { -4.644501087072e-05,  2.140811184612e-05, {   2,   0,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 { -3.770187633486e-05, -3.008302555540e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -3.487920329922e-07,  6.742800340436e-05, {   2,  -2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.710249427348e-05,  1.058374587011e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.253417982322e-05,  4.511809278923e-05, {   6,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.420388047596e-05,  4.343455644127e-05, {   1,  -1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -6.718178137392e-05, -4.085195955656e-07, {   3,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.219621944673e-05,  1.539253222742e-05, {   2,   0,  -2,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.729419211482e-05,  2.024008893994e-05, {   1,  -3,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  9.219183483567e-07,  6.647349927054e-05, {   4,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  3.895680742577e-05, -2.830586446925e-05, {   2,  -2,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.376331753487e-05,  1.342732384124e-05, {   2,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.447458520437e-06,  5.772656537476e-05, {   4,   0,  -2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  5.738630731214e-05,  9.785326080601e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 {  6.065798526110e-05,  6.487576878930e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  5.175337284955e-05,  1.537771374788e-05, {   2,   0,  -1,   0,   0,   3,   1,  -8,   3,   0,   0,   0,   0 } },
 {  5.264948691722e-05, -1.441388091988e-05, {   1,  -1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.860858796492e-06, -5.716702917268e-05, {   0,   0,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.341689277876e-05, -2.349162330948e-05, {   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -4.800927874088e-05,  1.887944460687e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  4.671249295487e-05, -2.010862817988e-05, {   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.419326446255e-06, -6.236749752845e-05, {   2,   0,  -1,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.973852670484e-05, -3.704743142015e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  2.222640330482e-05, -4.449427630535e-05, {   6,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.654096829096e-05, -4.007334519027e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 { -2.729522925085e-05, -3.929313319633e-05, {   4,   0,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  9.805125368148e-08, -6.647389545258e-05, {   2,   0,  -3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.911863677875e-05, -1.742949882526e-05, {   0,   0,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  5.381588702156e-05,  1.272621972938e-05, {   1,  -1,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.552390285051e-08,  6.649435893520e-05, {   4,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.578284542307e-07, -6.633554994671e-05, {   0,   0,   2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  9.748689937394e-06, -5.673434977919e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -2.556474438134e-05, -4.089375686382e-05, {   0,   0,   2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  4.007238716176e-05,  2.634927985313e-05, {   2,  -2,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.207403001100e-05, -5.431939980439e-05, {   2,   0,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 { -2.608929799573e-05,  4.020076704432e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,   0,   0,   0,   0 } },
 { -2.799161087681e-05,  3.828106210219e-05, {   3,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.997927658386e-05, -2.629313558721e-05, {   2,  -2,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.113823321532e-05, -1.508821175443e-05, {   2,   0,  -1,   0,   0,   3,  -7,   8,  -3,   0,   0,   0,   0 } },
 {  3.026298539765e-05,  3.595939031250e-05, {   1,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.501930971008e-07, -6.582099975608e-05, {   4,   0,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  2.738114601856e-05, -3.872882396023e-05, {   2,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 { -3.092715807310e-05,  3.505754974829e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 {  1.389831673765e-06,  6.459367710700e-05, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -4.315927596981e-05, -2.281287082675e-05, {   0,   0,   3,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.094369016402e-05, -4.502570201267e-05, {   0,   0,   1,   0,   0,  -9,  13,   0,   0,   0,   0,   0,   0 } },
 { -8.966888552336e-06, -5.694374894339e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  3.282263534334e-05,  3.295666254098e-05, {   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.691233073608e-05, -2.883242789081e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 {  3.987398706300e-05,  2.584709937250e-05, {   0,   0,   2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -4.041823106673e-05,  2.517633972799e-05, {   4,   0,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.783817740743e-05,  4.771170728744e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 { -3.364234139502e-05,  3.190084332802e-05, {   2,   0,   0,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -5.112115587199e-05,  1.440754855275e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  4.138773732655e-05,  2.407356475628e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -3,   0,   0,   0 } },
 { -1.970065418062e-05, -4.574020837751e-05, {   0,   0,   2,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.403368744223e-05,  1.308998096994e-06, {   2,  -2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.641089190450e-06, -5.961461762424e-05, {   2,  -2,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.023527236591e-05, -4.500906201808e-05, {   1,  -1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -6.378120768298e-05, -1.447734150548e-06, {   2,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.183764176392e-05,  1.337968808927e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 { -5.863265794090e-06, -5.934519595916e-05, {   2,   0,  -1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 {  1.398372718954e-05, -5.122288558969e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -5.975334951425e-05,  5.434815712377e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.264616534739e-05,  1.249818592866e-05, {   2,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.256794356144e-08, -6.510811696421e-05, {   1,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.975923492927e-05,  5.369715138458e-06, {   3,  -1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.975924055239e-05, -5.369652558731e-06, {   3,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.976081057957e-05, -5.352150864109e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.287590910263e-05,  2.213756639058e-05, {   0,   0,   3,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -3.528711820043e-05,  2.967954578927e-05, {   1,   1,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.328512466855e-05, -5.164156513672e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  4.651022331822e-05,  1.835208780736e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.724385074973e-05, -7.603622223272e-06, {   2,   0,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.149507532235e-05, -4.333994337651e-05, {   0,   2,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.263914931079e-05, -1.213475761947e-05, {   2,   0,  -1,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.261559171843e-05,  1.215139880766e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 {  3.980703470108e-05, -2.495806767942e-05, {   1,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.043545510283e-05, -5.432645081410e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 {  5.180619861975e-05,  1.294667702971e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.141302111181e-05, -4.330290126588e-05, {   0,   0,   1,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.399752924534e-05,  4.066582822980e-05, {   2,   0,  -1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 {  4.771269443278e-05, -1.694194794869e-05, {   4,   0,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  4.770607125650e-05,  1.693894025724e-05, {   4,   0,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.478680512296e-05,  3.975660708393e-05, {   0,   0,   2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  2.149842144664e-05,  4.302540810372e-05, {   2,   0,  -3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.999923393832e-05,  4.447565236128e-05, {   1,  -1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.208025082325e-05, -4.238155243133e-05, {   0,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.491461345400e-06, -5.696510277010e-05, {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -2.532982392166e-05,  3.904533349285e-05, {   2,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  6.197588478562e-05, -2.381381803562e-06, {   2,  -2,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.710057996978e-05, -2.720775240856e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 {  3.660592587747e-07, -6.393614541901e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 {  2.208782909659e-05, -4.219888453694e-05, {   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  5.530368214718e-08, -6.419643269556e-05, {   3,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.177600222472e-05,  1.236752499106e-05, {   0,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.205860720576e-08,  6.408719342163e-05, {   1,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.547253760806e-05, -4.857148082445e-05, {   0,   0,   2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.506058876304e-05,  1.893078890013e-05, {   0,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.023558224095e-05,  4.374670523357e-05, {   0,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  4.726774377275e-05,  1.668015508267e-05, {   1,   1,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -9.998505250804e-06,  5.394801017071e-05, {   1,   1,   0,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 {  9.998505250804e-06, -5.394801017071e-05, {   1,   1,  -2,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 { -5.952636678918e-05,  4.388716643915e-06, {   3,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.792105903551e-05, -4.596026034396e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 {  6.094537241687e-05,  2.931682475690e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.276973192671e-05,  5.105806398622e-05, {   1,  -1,   2,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  5.084477670343e-05,  1.296103452412e-05, {   1,   1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -4.886154771510e-05,  1.486767581099e-05, {   4,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.974305820179e-05,  3.398133704012e-05, {   2,   0,   0,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 { -8.144461266189e-06, -5.556028951006e-05, {   2,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.571275580548e-05, -4.795674556493e-05, {   2,   0,  -1,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 {  5.817880482099e-05,  5.461665344291e-06, {   2,  -2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.163997593037e-05,  2.197546437334e-05, {   1,  -1,   1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.306028644789e-05,  5.053777472124e-05, {   1,  -1,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -5.153884424366e-05,  1.202871312423e-05, {   3,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.894423916916e-05, -2.459592864259e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  1.202736225235e-05, -5.145442934337e-05, {   1,   1,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  9.969528218318e-07, -6.248352081736e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  6.060169209866e-06, -5.735353713708e-05, {   1,  -1,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.638012091617e-05, -1.702004192760e-05, {   1,   1,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 { -5.567456319665e-06,  5.781102710576e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 {  7.996985947997e-06, -5.535202524300e-05, {   3,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.023489242398e-08,  6.328509406657e-05, {   1,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.176509545823e-05,  1.535835428572e-06, {   2,   0,   0,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.091748636752e-05,  4.234425386567e-05, {   0,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.109871604165e-05,  4.212468434241e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -1.101408159753e-05, -5.217981145137e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   6,   0,   0,   0 } },
 {  5.023764179510e-05, -1.291822743608e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.159139444517e-05,  2.147208841356e-05, {   0,   2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -3.907737209751e-05, -2.392251686220e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 { -3.463104173123e-05, -2.836432214772e-05, {   2,   0,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -2.216182686618e-05, -4.083015300682e-05, {   2,   0,  -1,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  6.048950101720e-05,  2.474133142114e-06, {   2,   0,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.443588219689e-05, -1.849094516220e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 { -4.912800532351e-07, -6.242687486552e-05, {   2,   0,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.145360949493e-05,  5.146156928463e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 { -5.014643284934e-05,  1.273432582847e-05, {   1,  -1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.133923191098e-05, -1.534683729751e-06, {   0,   0,   1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -6.058087775592e-05, -2.290943542845e-06, {   2,   0,  -2,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -6.099378263833e-05,  1.844295129526e-06, {   2,   0,  -1,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.413143073341e-06, -6.040419012916e-05, {   0,   0,   1,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -3.325370254107e-05,  2.947869116431e-05, {   0,   0,   2,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.389045053234e-05, -1.881333752890e-05, {   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.892489707520e-06,  5.980467090494e-05, {   4,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.971517715129e-06, -5.872395532199e-05, {   2,   0,   2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.669672356348e-05,  1.593167318862e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  6.024382247978e-05,  2.383518787098e-06, {   2,   0,  -1,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 { -6.026647832699e-05, -2.277822085541e-06, {   2,  -2,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  2.336124078160e-05,  3.914443247078e-05, {   0,   0,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 {  4.676302795902e-08,  6.244041502767e-05, {   0,   2,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.273017399703e-05,  9.729928653002e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 {  6.231492171897e-05,  1.338261810977e-07, {   2,   0,  -1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 { -1.561723174908e-05,  4.669137725955e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  2.258503334608e-06, -6.003382334708e-05, {   3,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.593437594514e-07,  6.212871715348e-05, {   0,   2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -4.418108229766e-05,  1.808427286126e-05, {   2,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.191008158477e-05, -3.027806623669e-05, {   2,   0,  -2,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -4.271533990280e-05, -1.947191608274e-05, {   1,  -1,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.472169501728e-08, -6.212008247091e-05, {   4,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.343002066146e-05, -1.861670363371e-05, {   1,  -3,   2,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.821302576839e-06, -6.016727552929e-05, {   2,  -2,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -6.085731248799e-05,  1.100379087720e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 { -2.302432152918e-05, -3.893050650420e-05, {   0,   0,   2,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -2.307586028856e-05, -3.881578452492e-05, {   2,   0,  -1,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -1.918073439327e-07, -6.169615701248e-05, {   2,   0,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  9.149497749965e-06, -5.272617806983e-05, {   2,   0,   0,   0,   0,   0,   6,   0,  -7,   0,   0,   0,   0 } },
 { -1.896446768465e-05, -4.289187541677e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 {  5.096500643717e-05,  1.088129109411e-05, {   4,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.116548614252e-05, -2.067521148605e-05, {   0,   2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.081473259925e-05,  1.014241527090e-06, {   2,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.115807340677e-05,  5.063245196476e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -4.190663486612e-05,  1.987562504287e-05, {   2,   0,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 { -1.052570079749e-05, -5.119469104913e-05, {   0,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.293875004193e-05,  3.877567394099e-05, {   0,   0,   2,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -6.042807231078e-05,  1.248897577052e-06, {   2,   0,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.517964329122e-05, -2.647357390451e-05, {   4,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.986488005265e-05, -1.177205962567e-05, {   1,  -1,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.863590294395e-05, -2.951663750453e-06, {   2,   0,  -2,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -2.294538804929e-05,  3.863421210098e-05, {   2,   0,  -1,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 {  1.622181845737e-05, -4.530265227776e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  6.194529443228e-06, -5.529563227067e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 { -4.973061394347e-05,  1.175510782221e-05, {   1,  -1,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.003735925901e-05,  5.142784082460e-05, {   2,   0,  -3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.211387366540e-05, -9.268602601675e-06, {   2,   0,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.346928655812e-05, -4.791140788880e-05, {   1,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  2.089984772330e-06,  5.919751376527e-05, {   2,   0,   3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  5.614977482855e-08, -6.120950782117e-05, {   0,   2,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.971007498398e-05,  1.521201273374e-06, {   2,   0,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.016113826952e-05,  4.102316533861e-05, {   2,   0,  -1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.909762500450e-06,  5.317787180147e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  2.317630849131e-05, -3.785533361705e-05, {   2,  -2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  5.133821278839e-05,  9.673587794638e-06, {   2,  -2,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.735743828165e-05,  2.361355013116e-05, {   3,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.202815639848e-05,  3.889509782446e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  4.910189452804e-05,  1.178779786352e-05, {   3,  -1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -1.977657196299e-05,  4.111027027339e-05, {   2,   0,   0,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.926247864934e-05, -1.162147136911e-05, {   1,  -1,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.023646720183e-05, -3.063324463205e-05, {   1,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.348516241802e-05,  4.738208744492e-05, {   2,   0,  -1,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  1.447573405068e-05, -4.639135079407e-05, {   1,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.485508365131e-05, -4.598093680367e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -4.905021102607e-05, -1.171297757769e-05, {   0,   1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.439235383556e-05, -3.635716239949e-05, {   2,   0,   1,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 {  2.439235383556e-05,  3.635716239949e-05, {   2,   0,  -1,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 { -5.926659480744e-05,  1.418619392288e-06, {   2,   0,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.897703844465e-05, -1.168295435917e-05, {   1,   1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.449464470746e-05,  4.615822181838e-05, {   1,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  1.084037453940e-06, -5.952046278378e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 { -5.720337581329e-06,  5.488322406161e-05, {   1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.335246653286e-05, -4.715921939369e-05, {   4,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.322863804016e-08, -6.039113837362e-05, {   3,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.871024732999e-05,  2.171271439589e-05, {   2,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.709900057751e-05, -3.330993686380e-05, {   2,   0,  -1,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 { -2.445826660773e-05, -3.594250432350e-05, {   2,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.874538486713e-05, -1.163342841613e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 {  4.107778889778e-05, -1.920086864037e-05, {   2,   0,  -2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  8.774021662349e-06,  5.149167902815e-05, {   3,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -7.272012259500e-09, -6.023989364931e-05, {   2,   0,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.684650635470e-05, -2.331363889844e-05, {   3,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.279942049555e-07,  5.998983801941e-05, {   2,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.716078357483e-05,  2.934262309038e-06, {   2,   0,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.937446919822e-05,  4.068364295261e-05, {   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.476620688281e-06,  5.657370567081e-05, {   0,   0,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.343419152607e-05, -2.659200712714e-05, {   0,   0,   2,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.958388197404e-06,  5.106767233413e-05, {   2,   0,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.624203382070e-08, -5.999999780164e-05, {   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.953097093182e-09,  5.999999992733e-05, {   2,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.350884840912e-09, -5.999999998479e-05, {   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.204277183791e-09,  5.999999998791e-05, {   2,   1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -1.204277183791e-09, -5.999999998791e-05, {   0,   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.597344572416e-10, -5.999999999637e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.492624817270e-10,  5.999999999649e-05, {   2,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.073745796979e-10, -5.999999999693e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  1.099557428731e-10,  5.999999999980e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.800120810034e-08,  5.994806427564e-05, {   3,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.966905691773e-06, -5.075222632721e-05, {   0,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -2.048450060895e-05,  3.922946738718e-05, {   2,   0,   0,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 { -2.177345664689e-08, -5.968791321944e-05, {   1,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.707899992492e-05,  3.260803987396e-05, {   2,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 { -7.676389582771e-06, -5.199773303144e-05, {   0,   0,   0,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 { -3.565380078229e-05, -2.401931458644e-05, {   0,   0,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.886060413959e-05,  8.054807143060e-07, {   2,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -2.565226145398e-05,  3.399200796799e-05, {   1,   1,  -4,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -5.777663778061e-05,  1.867171248961e-06, {   0,   0,   1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.056786062146e-06,  5.852897972814e-05, {   0,   0,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -4.518973257989e-05,  1.436868901200e-05, {   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.607676124669e-06,  5.690931359188e-05, {   2,   0,  -2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  3.570614860446e-05,  2.378875614446e-05, {   1,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.815298326459e-05, -1.133440687915e-05, {   1,  -1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.372104613609e-05,  3.572721169888e-05, {   2,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  2.824378235207e-08, -5.941837309430e-05, {   1,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.320720213200e-05, -4.618735826451e-05, {   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -8.706930059289e-06,  5.067088810921e-05, {   2,  -2,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.045543725616e-06, -5.633187732578e-05, {   0,   0,   2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  5.663408347514e-05,  2.696776819327e-06, {   2,   0,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.680368788480e-05, -2.510410114933e-06, {   4,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -2.599098137732e-05, -3.329515369806e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 { -9.608026108894e-06,  4.960167516076e-05, {   2,   0,   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 { -2.962512580202e-05, -2.957321601426e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 { -3.391581498473e-05,  2.527935657712e-05, {   4,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  7.631719608428e-07,  5.835282646427e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  2.268773683896e-05,  3.640587947208e-05, {   2,   0,   0,   0,   0,  -3,  18, -25,   2,   0,   0,   0,   0 } },
 {  2.268808828644e-05, -3.639901245527e-05, {   2,   0,   0,   0,   0,   3, -18,  25,  -2,   0,   0,   0,   0 } },
 {  3.841181760764e-05, -2.063928456677e-05, {   2,   0,   0,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 {  4.777586696745e-05,  1.123432705644e-05, {   1,   1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.541083926405e-05, -2.358730001022e-05, {   1,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.324994028624e-05,  1.573022742548e-05, {   1,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.334947668440e-06,  5.761348892059e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 { -4.857464664857e-08, -5.889976082962e-05, {   2,   0,   1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.760731790126e-05,  4.129184030751e-05, {   2,   0,  -1,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 { -5.047210064824e-05,  8.405014726509e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -4.811830268543e-05,  1.075382732086e-05, {   2,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.968248241311e-07, -5.805406636014e-05, {   2,   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  6.980459924007e-06,  5.184464617676e-05, {   1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.899505440690e-07, -5.850517470815e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 { -2.899500600099e-07,  5.850490141173e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 { -6.744358513772e-06,  5.205018702253e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  4.515340425382e-05, -1.357917018980e-05, {   0,   0,   1,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.768584859468e-05, -1.104348773153e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 { -4.821879341049e-08,  5.866557401826e-05, {   4,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.365733488685e-06,  5.232636024054e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  4.922952321722e-05,  9.454761313768e-06, {   0,   0,   2,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.563510626538e-06, -5.610494432415e-05, {   2,   0,   0,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 { -1.078837049357e-05,  4.784570951774e-05, {   1,  -1,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -5.524692888367e-05, -3.338563748472e-06, {   2,   0,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 {  1.338845502470e-05, -4.519134594065e-05, {   2,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.664111055720e-05, -1.190525446616e-05, {   1,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.960069461773e-05,  3.890904924021e-05, {   0,   0,   1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.035063013143e-05,  8.147980577190e-06, {   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  4.774868375839e-05,  1.074436284867e-05, {   2,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.004048490751e-06,  5.548382738915e-05, {   0,   0,   2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -5.106160524353e-06, -5.335519255380e-05, {   2,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  2.899754090767e-06,  5.554790979900e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -1.253292179805e-05,  4.583411958449e-05, {   3,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.181904413871e-05,  2.648745196306e-05, {   0,   0,   2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.387612976627e-05, -1.441024498624e-05, {   4,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.542407716646e-05, -2.285814220003e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  4.023346959517e-05, -1.803637851345e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 {  7.937065844935e-06, -5.031114360888e-05, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 { -1.604502913622e-05, -4.220190908894e-05, {   0,   0,   3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.004419488190e-05, -4.814894565802e-05, {   1,   1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.653882298523e-05,  2.161735391177e-05, {   1,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.284964933910e-05, -4.523100819953e-05, {   2,   0,  -1,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -1.952003482230e-05, -3.849008483642e-05, {   1,  -1,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.237685038133e-05, -4.556679483223e-05, {   3,  -1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.166837821020e-05,  3.627000014082e-05, {   1,   1,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.352238417624e-06, -5.654304061884e-05, {   4,   0,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  4.173478736449e-05, -1.615893875589e-05, {   1,  -1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.543163745452e-05,  2.244412024713e-05, {   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -4.198683086495e-05,  1.585024512422e-05, {   1,  -1,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  5.056323827123e-05,  7.261785504239e-06, {   2,   0,   1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  2.420311451617e-05,  3.355198551357e-05, {   2,   0,  -2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  1.528474942691e-05, -4.245709873581e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  9.876858706106e-09,  5.770800746754e-05, {   2,  -2,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.703130148006e-06,  5.498341481026e-05, {   4,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.533859857423e-05,  4.234450118610e-05, {   2,   0,  -1,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.435775804626e-06, -4.923880359633e-05, {   2,   0,  -2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -8.259248898095e-06,  4.936889152716e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,  -7,   0,   0,   0,   0 } },
 { -1.090389416645e-05, -4.672290983731e-05, {   0,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 { -1.090557389629e-05,  4.671331475535e-05, {   0,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -5.638860043901e-05,  1.220370094047e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -4.563466922280e-05,  1.196727369172e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 { -4.142737376426e-05,  1.616621080786e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.938051873521e-05,  3.820748921104e-05, {   1,  -1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.669853364705e-05,  1.072855817650e-05, {   3,  -1,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -6.443117176448e-07,  5.674414374054e-05, {   3,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -2.514162911600e-07, -5.713127368064e-05, {   0,   0,   1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -1.425030951409e-05, -4.311134209215e-05, {   2,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.764550908718e-05,  9.714409702519e-06, {   3,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.093418807615e-06,  4.826143971254e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -4,   0,   0,   0 } },
 {  4.319797180929e-05, -1.410849542444e-05, {   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.659823899399e-05, -6.463563361075e-07, {   4,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.932573377761e-05, -2.785987983945e-05, {   2,   0,  -1,   0,   0,   0,  10, -16,   3,   0,   0,   0,   0 } },
 { -2.061112105017e-05,  3.649774479536e-05, {   2,   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.836959941749e-08,  5.700449494232e-05, {   0,   0,   2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.527759230285e-05,  1.750397473844e-06, {   0,   0,   1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.977613001143e-05, -7.240798416744e-06, {   2,   0,  -2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -4.627343127666e-05, -1.070208322007e-05, {   1,   1,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.145884103238e-07,  5.629867803509e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  2.585950435990e-07, -5.664093952014e-05, {   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.921952350177e-05, -3.766539108272e-05, {   1,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.271301164651e-05,  1.413113241889e-05, {   2,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.586517657555e-05, -1.097791686311e-05, {   1,   1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.141096740697e-05,  5.430421797810e-06, {   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.011713093274e-05,  4.665535870687e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -4.639724090450e-05, -1.036346859796e-05, {   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.878921410166e-05,  1.795516021882e-05, {   0,   0,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -3.075251239975e-05,  2.598828179146e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -7.975101192673e-06, -4.873827387833e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -2.967411205191e-05, -2.703208993427e-05, {   2,   0,   0,   0,   0, -11,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.522546603098e-06,  5.115750473800e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -3.880534110887e-05, -1.787142664869e-05, {   0,   0,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.583287503481e-05,  7.950539712543e-07, {   4,   0,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.838653227000e-05,  8.199175867300e-06, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.633235649364e-05, -1.019985002467e-05, {   3,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  4.360345511791e-05,  1.288711781145e-05, {   2,   0,  -1,   0,   0,   0,  -1,   5,   0,   0,   0,   0,   0 } },
 { -3.057822850656e-05, -2.585742747644e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -9.113193306523e-06, -4.731505947247e-05, {   1,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.592414993936e-05, -1.045100977907e-05, {   0,   0,   3,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -3.423831375160e-05, -2.209060417344e-05, {   2,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 {  1.100598692465e-05,  4.531915643510e-05, {   3,  -1,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  4.828965299359e-05,  7.960201999131e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 { -1.877507755323e-05,  3.746989770372e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  2.690720555707e-05, -2.932660382696e-05, {   3,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.138583022613e-05, -3.483650574863e-05, {   2,   0,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0 } },
 { -3.695316863647e-05,  1.921573715766e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -5.249485616172e-05, -3.667055036938e-06, {   2,   0,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  3.148929084659e-05,  2.466626364042e-05, {   2,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.998271546550e-05, -3.583990131765e-05, {   2,   0,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  4.008687531802e-05, -1.573378305436e-05, {   2,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.472452080812e-05, -2.108880274302e-05, {   2,   0,   0,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 { -4.026500931564e-05,  1.550116443601e-05, {   1,  -1,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.640482713272e-08, -5.568695347391e-05, {   3,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.992743714339e-05, -3.582277996053e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -4.696468848834e-05, -8.782210192013e-06, {   4,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  8.942575858303e-06, -4.679302483871e-05, {   3,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.638672598389e-06,  4.708770484418e-05, {   1,  -1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.394410738973e-05,  1.727148617531e-06, {   4,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.439746281192e-05,  1.237624779726e-06, {   0,   2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.816010185181e-05, -3.743064387582e-05, {   2,   0,  -1,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.983253443890e-05,  1.575022701133e-05, {   1,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.319922205084e-05,  2.352436011019e-06, {   4,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.502335206564e-05, -5.017843578539e-07, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.319547925158e-05, -2.231726141932e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.605850823650e-09, -5.549409877991e-05, {   0,   0,   2,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  4.490860363290e-05, -1.058521875928e-05, {   1,   1,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.134174929015e-05, -4.414271793080e-05, {   1,   1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.790953543800e-05,  3.756012441862e-05, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 { -5.209769924425e-05, -3.353275457921e-06, {   1,   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.055724493020e-05, -4.872942010670e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -1.620766915650e-05,  3.920638142832e-05, {   0,   0,   3,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.302303488022e-05, -1.234147653408e-05, {   2,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  2.280686546064e-06,  5.306836060282e-05, {   2,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.526590131156e-05, -4.008246475962e-05, {   2,   0,   0,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 {  7.981346976187e-06,  4.733674305056e-05, {   1,   1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.838551468980e-05,  6.911553614362e-06, {   2,   0,   0,   0,   0,  -3,   4,   0,  -2,   0,   0,   0,   0 } },
 { -5.353797338523e-05,  1.732560936230e-06, {   0,   0,   1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.192276479281e-06,  5.407735768186e-05, {   3,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.910017831627e-06, -5.334482225264e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 {  2.296325743913e-05,  3.227691896837e-05, {   3,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.103607645425e-05, -4.419443722926e-05, {   1,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.312141376682e-05, -1.209234444820e-05, {   1,  -1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.596161931605e-05,  2.924905808299e-05, {   2,   0,  -1,   0,   0,   0,   1,   2,  -3,   0,   0,   0,   0 } },
 { -2.054498356921e-05,  3.457429641578e-05, {   2,   0,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0 } },
 { -1.573606561509e-05,  3.936220872451e-05, {   0,   2,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.989010865256e-05,  1.519639569874e-05, {   0,   0,   1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 {  1.230930539860e-05, -4.273900527077e-05, {   1,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -3.492216253847e-05,  2.011815034540e-05, {   4,   0,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.540499088807e-05, -1.962769596155e-05, {   1,  -1,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.442517050427e-05,  6.021957727721e-07, {   1,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.250728017950e-05,  2.249569708108e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -2.685281916594e-05,  2.812983676434e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 {  1.835332460903e-05, -3.662834722859e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -4.570846895976e-05, -9.241821118859e-06, {   4,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.193814146214e-05,  2.300707829042e-05, {   2,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.245326491144e-05, -3.248078394139e-05, {   2,   0,  -1,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 { -3.962400275887e-05, -1.526539478510e-05, {   2,   0,  -3,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.456628229996e-05, -2.032086066384e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 { -4.581876217343e-05,  9.057858644648e-06, {   2,   0,  -1,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 {  3.110561510248e-06,  5.176225637524e-05, {   6,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.751746268080e-06, -5.111506990710e-05, {   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.701186558746e-07,  5.412405043532e-05, {   6,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.206814878258e-05, -1.269882222407e-05, {   0,   0,   2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.466111019045e-05, -4.009181737225e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -5.007627253320e-05,  4.634355826340e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -5.035711194807e-06, -4.966495439129e-05, {   1,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.718660633229e-06,  4.596272450273e-05, {   1,  -1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -8.568841050515e-06,  4.609069908276e-05, {   1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.274268705819e-05, -4.183354314518e-05, {   3,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.798863777128e-05, -1.658281269298e-05, {   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.432554145878e-08,  5.450382441661e-05, {   2,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  8.146588213657e-06, -4.640004776682e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 { -3.688015460671e-05,  1.761297247073e-05, {   2,   0,   0,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 {  1.372070217706e-05,  4.074097826867e-05, {   1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.248095482319e-05,  2.180258104891e-05, {   2,   0,  -1,   0,   0,  -5,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.979932495163e-05, -4.475179813388e-06, {   1,   1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.979946864213e-05,  4.473580552659e-06, {   3,  -1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.979954124877e-05,  4.472772228905e-06, {   1,  -1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.979960837872e-05,  4.472024745675e-06, {   1,  -1,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.966998109532e-05, -3.457640037383e-05, {   4,   0,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -2.548786810340e-05,  2.875345746972e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   5,   0,   0,   0 } },
 { -5.256011153762e-05,  1.640715335664e-06, {   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.839747299929e-06, -5.127741493220e-05, {   2,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.918538424585e-07,  5.334785709969e-05, {   1,   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.114027628065e-05,  4.286984849784e-05, {   2,   0,  -1,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 { -1.541649257229e-07,  5.377666961072e-05, {   6,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  9.577171027202e-06, -4.431735869349e-05, {   2,   0,   1,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  9.577171027202e-06,  4.431735869349e-05, {   2,   0,  -1,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -1.933217559685e-05,  3.455623499898e-05, {   2,   0,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 { -2.511338519530e-05, -2.875145468728e-05, {   2,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.354365790641e-05,  1.031215745666e-05, {   1,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.589620227643e-05,  3.795701697008e-05, {   1,   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -6.505490160980e-06,  4.731027485705e-05, {   2,   0,  -1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  8.243052553811e-06,  4.549933711313e-05, {   2,   0,  -3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.267854012462e-06,  5.245799007168e-05, {   0,   2,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.903911157354e-05,  3.467210665362e-05, {   1,  -2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.403577513894e-05, -2.961623968850e-05, {   1,   0,   1,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.403577513894e-05,  2.961623968850e-05, {   1,   0,  -1,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.164232042656e-05, -4.199207613957e-05, {   2,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.281891529772e-05,  1.081545705410e-05, {   1,   1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.167053716226e-05,  2.192524722786e-05, {   0,   0,   2,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  6.465548542824e-06,  4.704212941906e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 { -1.819684400363e-05,  3.531026836144e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 { -4.306939386785e-05,  1.042162033854e-05, {   0,   0,   4,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.586208286238e-06,  4.686602912472e-05, {   0,   0,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.085050408729e-05, -4.258418794060e-05, {   2,   0,  -1,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 { -4.203201452414e-05,  1.133610422957e-05, {   2,   0,   0,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 {  1.511048078577e-05, -3.825192653063e-05, {   2,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 {  4.305364157720e-05,  1.028206814536e-05, {   3,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.432765466991e-05,  8.995760805284e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  2.993799056025e-05,  2.336145687316e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 {  3.777544440626e-05,  1.545931855569e-05, {   2,   0,   0,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 {  3.146765623188e-05, -2.175295090012e-05, {   0,   0,   2,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -1.769961932737e-05, -3.543807691402e-05, {   1,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.879325423116e-06, -5.125552782679e-05, {   3,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -3.151386335403e-05,  2.159743405058e-05, {   0,   2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.140062394894e-05, -1.708809109134e-06, {   2,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.258518606066e-05,  1.051868441936e-05, {   1,  -1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.076945156851e-05, -1.231745183269e-05, {   2,   0,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.021109575318e-05, -4.287150652700e-05, {   2,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.284357637560e-05,  2.019889053689e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 { -1.385090313383e-06, -5.164511318904e-05, {   2,   0,  -3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.063425985388e-05,  3.234636791785e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 {  2.151536494988e-06, -5.080420287610e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   6,   0,   0,   0 } },
 {  1.634803379468e-05,  3.650673624207e-05, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.086000812596e-05, -1.978323877606e-06, {   0,   0,   1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.272020815857e-05,  1.011787983159e-05, {   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.809525474411e-05, -3.472321192122e-05, {   2,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.031772696550e-05, -4.248339629726e-05, {   1,   1,   1,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  4.348408392311e-05,  9.305834652129e-06, {   1,  -1,   1,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -4.348408392311e-05, -9.305834652129e-06, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 {  5.099920445390e-05,  1.777464677231e-06, {   2,   0,  -2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.619116709440e-05, -3.657657868255e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.806353413742e-05,  1.464691751466e-05, {   4,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.017811226812e-07,  5.258949173968e-05, {   0,   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -9.070493608720e-06,  4.361170809107e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 { -4.677479163815e-05, -5.899213724432e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -8.664915101110e-06, -4.399988594382e-05, {   0,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.211819898508e-05,  5.156589001990e-07, {   2,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.047006817176e-05,  4.215000269732e-05, {   1,  -1,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -2.786025110956e-05, -2.471408865696e-05, {   4,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.155773724783e-05,  4.100616001471e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.268302784570e-05, -9.875083782751e-06, {   1,  -1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.007390518842e-09, -5.246649681381e-05, {   0,   2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.875706948959e-05,  2.368632860801e-05, {   0,   0,   4,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.161639775389e-05,  1.077292920808e-05, {   1,   1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.356923361282e-05,  2.877228221943e-05, {   2,   0,  -1,   0,   0,   3,  -2,  -2,   0,   0,   0,   0,   0 } },
 { -1.025135223767e-05, -4.200523251916e-05, {   1,  -1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.452596302881e-05,  2.772546096784e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -3,   0,   0,   0 } },
 { -1.386977625860e-08,  5.223493103929e-05, {   6,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.606375093647e-06,  4.462418579129e-05, {   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.243514302159e-06, -4.394603185685e-05, {   2,  -2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.811739663669e-05,  2.405042536080e-05, {   2,   0,  -1,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  4.737673675806e-05,  4.765216789520e-06, {   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  1.725463004413e-05, -3.484850335476e-05, {   0,   0,   1,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.759801687784e-05,  2.449352791714e-05, {   2,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  5.076603584559e-05,  1.288461482556e-06, {   2,   0,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.461361423437e-06,  4.357953379104e-05, {   1,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.902245407328e-05,  1.301511136357e-05, {   0,   0,   1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.905253694774e-05, -2.962838161439e-06, {   4,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.068699634079e-05, -2.125485963625e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  9.551313510544e-09, -5.191072589339e-05, {   2,  -2,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.164191917028e-05,  1.023849963491e-05, {   1,   1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.895635207173e-06,  4.894495670063e-05, {   0,   0,   1,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 { -9.516504742230e-06, -4.231485672439e-05, {   2,   0,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 {  1.549948595753e-05,  3.630156257983e-05, {   2,   0,  -1,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -2.072740685471e-05, -3.107049931721e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 { -2.655484589308e-09, -5.174695007652e-05, {   4,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.376435376965e-05,  3.797722189417e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -5.382556060934e-06, -4.631870789031e-05, {   2,   0,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -3.976358934990e-05, -1.192096299547e-05, {   6,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.262374239434e-05,  3.905156541090e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -3.972432127379e-05,  1.190766364179e-05, {   6,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.334636737385e-05, -3.820919583266e-05, {   4,   0,  -2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.307410889909e-05, -3.847530020679e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.116180474807e-05, -1.037562370412e-05, {   3,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.335466815290e-06,  4.319222731602e-05, {   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.248210974428e-05, -2.903414403380e-05, {   2,   0,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 {  4.084050225193e-05,  1.066435768194e-05, {   0,   0,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.454036205218e-05, -3.695102168725e-05, {   4,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.360236869709e-05,  2.788761182994e-05, {   2,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 {  8.781626952937e-07, -5.060288559739e-05, {   0,   0,   2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -9.554957913695e-06,  4.190342473288e-05, {   2,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 { -4.969711928950e-05,  1.666682385733e-06, {   0,   0,   1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.250249140425e-06,  4.510866966035e-05, {   1,  -1,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.858007602708e-05,  2.763286697883e-06, {   4,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.516555975030e-06, -4.282369521611e-05, {   4,   0,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  6.536077267972e-06,  4.476134553836e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 { -3.757520602333e-05, -1.371509724006e-05, {   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.672148456506e-05, -2.455394578056e-05, {   0,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -4.216758187449e-05,  9.028693797197e-06, {   2,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.189807318861e-05, -9.219572856446e-06, {   2,   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  8.341113670162e-08, -5.102279722043e-05, {   3,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.941619148862e-08, -5.105688208055e-05, {   3,   0,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.561884867999e-07,  5.019437826443e-05, {   0,   0,   2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -6.425445693882e-06,  4.457502577191e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  5.570083811452e-06,  4.540227759896e-05, {   0,   0,   1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  2.961076385021e-05, -2.134965850839e-05, {   4,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -7.978678755685e-06, -4.296195619785e-05, {   2,   0,   0,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 { -2.423521958791e-05,  2.669422529916e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -1.696025721630e-05,  3.395685894859e-05, {   1,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.987790693513e-06,  4.492321879260e-05, {   3,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.653368633383e-05, -2.436527209177e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.500455591295e-05, -3.588425394356e-05, {   4,   0,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  4.816675808919e-05,  2.718097222145e-06, {   2,   0,  -2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.725777977924e-06, -4.712304453553e-05, {   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.385506050608e-05, -1.697474907320e-05, {   2,   0,  -1,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 {  2.699835659202e-05,  2.379978283427e-05, {   2,   0,  -1,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 {  1.084038108265e-05, -3.995594266207e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -2.944378762425e-08, -5.076243235723e-05, {   2,   0,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.802815405498e-05,  3.275379253047e-05, {   0,   0,   3,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  2.190888075649e-07, -5.053334941530e-05, {   0,   0,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -3.339972202022e-05,  1.734808925789e-05, {   2,   0,   1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 {  3.339972202022e-05, -1.734808925789e-05, {   2,   0,  -1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -3.161801069754e-05, -1.908246638466e-05, {   2,   0,   0,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 {  1.847628122545e-05,  3.222268611489e-05, {   0,   0,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -4.785270920730e-05, -2.829184843834e-06, {   0,   0,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.658449117639e-06,  4.099556743224e-05, {   2,  -2,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.103215222798e-06,  4.252983822289e-05, {   1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.092140026317e-06, -4.951181477663e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -2.585536954439e-05, -2.474463481219e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  2.585536954439e-05,  2.474463481219e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 { -3.947554916515e-05, -1.107007904948e-05, {   2,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.695458776841e-05, -1.358968882501e-05, {   0,   0,   1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.084494205100e-05, -9.643551485057e-06, {   1,   1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.942949214151e-05,  1.105584570082e-05, {   2,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  7.655804797111e-06, -4.282401523220e-05, {   0,   2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.923868792601e-06,  4.654183573711e-05, {   2,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -3.559816169820e-06,  4.686843616930e-05, {   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.949829523602e-07, -5.021158213946e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 {  4.845803728318e-05,  1.937549391179e-06, {   2,   0,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  9.996367229234e-07, -4.937313259188e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 { -4.065045898908e-05, -9.661868929231e-06, {   2,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.490609895298e-05,  5.372845132940e-06, {   2,   0,  -2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -7.622190169469e-06,  4.265564353791e-05, {   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.664616223785e-05,  2.361943173705e-05, {   2,   0,  -1,   0,   0,   0,  -2,   7,   0,   0,   0,   0,   0 } },
 {  7.292170520233e-06, -4.297033700130e-05, {   2,   0,  -3,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.077072375181e-05, -9.488631578760e-06, {   1,  -1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.727923912914e-05, -1.295085404573e-05, {   2,   0,   0,   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0 } },
 {  4.205782118029e-05,  8.148083101118e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,  -1,   0,   0,   0 } },
 { -1.540666377965e-05,  3.479487471623e-05, {   0,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.770820198677e-06, -4.239629867565e-05, {   4,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.540823953051e-05, -1.474692337583e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 {  3.068457528957e-05,  1.945603523237e-05, {   2,  -2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  6.415849632520e-07, -4.949794375527e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 } },
 { -2.933567692511e-05,  2.080299171311e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.018325535229e-05,  3.994523267427e-05, {   2,   0,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -6.762982131436e-06, -4.336400405417e-05, {   0,   0,   1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.557070019370e-05,  4.451556122960e-06, {   0,   0,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.515108225345e-06, -4.850207118405e-05, {   0,   2,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.361356816385e-09,  4.999999998147e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -1.117010721178e-09,  4.999999998752e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  7.417649320826e-10, -4.999999999450e-05, {   1,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  7.330382858084e-10,  4.999999999463e-05, {   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -5.934119456665e-10,  4.999999999648e-05, {   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.276056667451e-10, -4.999999999817e-05, {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.665191429327e-10, -4.999999999866e-05, {   4,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.879793265821e-10,  4.999999999917e-05, {   2,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.617993877942e-10,  4.999999999931e-05, {   2,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.356194490187e-10, -4.999999999944e-05, {   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   2,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.186470708008e-07, -4.963242066434e-05, {   3,   0,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.043360146526e-05,  1.950210583923e-05, {   1,   1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.671425950725e-05,  3.321861703573e-05, {   0,   0,   1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.032425098852e-05,  9.526819765072e-06, {   1,  -1,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.620668679154e-07, -4.967096822794e-05, {   0,   0,   2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.985798677783e-05, -9.973082759173e-06, {   1,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -3.330532545909e-05,  1.651824314744e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 {  1.745228622742e-05,  3.234347638121e-05, {   1,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.902719916163e-06, -4.685445494809e-05, {   2,   0,  -1,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.633308742304e-05,  3.338463806257e-05, {   2,   0,  -1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.858239214882e-05, -3.109579217274e-05, {   0,   0,   1,   0,   0,  12,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.315830476275e-05, -3.651247120353e-05, {   4,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  9.116552878397e-06, -4.050574607138e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -1.878390370989e-05,  3.082090420519e-05, {   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.017563285369e-05,  9.399450025834e-06, {   1,  -1,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.167293981971e-05,  1.787461354721e-05, {   0,   0,   2,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -2.924741227547e-05,  2.029844414042e-05, {   2,   0,   0,   0,   0,   0,   1,  -8,   3,   0,   0,   0,   0 } },
 {  5.115517973607e-08, -4.949386352944e-05, {   2,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.308147433943e-07, -4.939059734823e-05, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -4.968375293584e-08,  4.943485241248e-05, {   3,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.027489098248e-05,  1.912848914403e-05, {   2,  -2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.292004439109e-05,  1.646766754763e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 { -2.435556387604e-06,  4.694882298206e-05, {   0,   0,   1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.585048365755e-05, -2.350945026947e-05, {   2,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.428712374091e-05, -2.505191572933e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -8.094509822216e-06,  4.124418644266e-05, {   1,  -1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.209579783241e-05, -7.237611443627e-06, {   1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.877664568296e-05,  1.055321181732e-05, {   0,   0,   1,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 {  4.088911125481e-05, -8.404290444862e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -5.755721924400e-10,  4.924845062963e-05, {   6,   0,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.495830879153e-08, -4.916642153404e-05, {   2,   0,  -2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.708695040461e-05,  2.211711993942e-05, {   4,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.731295839423e-05, -3.188522607276e-05, {   4,   0,  -1,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  7.143585594173e-07,  4.845905697821e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  3.616796018407e-05,  1.299818289215e-05, {   1,  -1,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.104923911264e-05, -2.806796234969e-05, {   1,  -1,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -9.063076004300e-06, -4.001034855684e-05, {   4,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -3.100684358132e-05, -1.805725497607e-05, {   0,   0,   2,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  8.628496708751e-07, -4.817964208354e-05, {   2,   0,   1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 {  8.153199817282e-09, -4.899385896776e-05, {   0,   2,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.497752841174e-06,  4.650128590153e-05, {   1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -7.362500350834e-06, -4.162751661856e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  2.652215869630e-08,  4.895756114087e-05, {   2,  -2,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.872826775473e-05, -2.505244172769e-07, {   2,  -2,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.445782417136e-05,  3.450678675995e-05, {   4,   0,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  9.140380599574e-06,  3.976882487488e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  7.221101954117e-06,  4.167558372575e-05, {   1,   1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.489380730666e-06, -4.339837687311e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 { -2.101748427167e-08,  4.886492299587e-05, {   0,   0,   4,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.713345628807e-05,  1.174250377686e-05, {   2,   0,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.708636732388e-05,  2.178218032973e-05, {   1,  -1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -4.702755450935e-05, -1.836615794446e-06, {   0,   0,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.372841085948e-06, -4.147294383781e-05, {   1,   1,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.924467065658e-05, -9.562554929307e-06, {   2,   0,  -1,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 {  1.915438058857e-06, -4.683335741023e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -3.585982658731e-05, -1.287212748366e-05, {   1,  -1,   1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.168558284706e-07,  4.816569897420e-05, {   4,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.819586577983e-07,  4.799222720654e-05, {   4,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  1.295082238435e-05,  3.569527110772e-05, {   0,   0,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -3.532818980472e-05, -1.330504193575e-05, {   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.621145507807e-05, -2.240988092608e-05, {   2,   0,  -1,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 {  4.649641913591e-05,  2.100862068766e-06, {   2,   0,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.795004133937e-05, -2.062361030675e-05, {   2,   0,  -1,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  4.127603635132e-09, -4.856851876052e-05, {   1,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.598319383444e-05,  2.588324412173e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -1.769510097052e-05, -3.083330222535e-05, {   4,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.396329955270e-06,  4.111427516795e-05, {   1,  -1,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.925914473233e-05, -9.249391491135e-06, {   1,  -1,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.035823131858e-05,  1.813632258063e-05, {   0,   0,   1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  3.930788536871e-05, -9.183307415672e-06, {   1,  -1,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.283388340025e-05,  2.563528993958e-05, {   2,   0,  -2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -2.678868582509e-05,  2.166304166572e-05, {   0,   0,   0,   0,   0,  14, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.249044414592e-05,  3.595848176861e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 { -8.095264329024e-06, -4.029282111735e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -2,   0,   0,   0 } },
 { -1.597070833314e-05, -3.235577612787e-05, {   2,   0,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.571368793559e-05, -2.586018422752e-06, {   1,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.609995785934e-05, -3.218338655827e-05, {   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  2.736267494294e-05,  2.086311313511e-05, {   2,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  3.409031892481e-05, -1.403980484227e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   4,   0,   0,   0 } },
 {  2.808583956915e-05, -2.002396498894e-05, {   4,   0,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.985807611593e-05, -2.824750249875e-05, {   4,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.730374300394e-05,  3.080132251732e-05, {   2,  -2,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -4.443933253833e-06, -4.361038169303e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  2.764444770643e-05, -2.038361802088e-05, {   4,   0,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.061627552518e-05,  2.738033418984e-05, {   1,   1,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -3.847112571945e-05, -9.515826506012e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 { -6.761603343106e-06, -4.118429104273e-05, {   2,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.407683546651e-05,  1.383952110296e-05, {   4,   0,  -2,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.702681786826e-05, -1.088548192507e-05, {   4,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.399544310965e-05,  3.901946194223e-06, {   0,   2,   1,   0,   0, -23,  22,   0,   0,   0,   0,   0,   0 } },
 {  9.589144581544e-06, -3.830496316661e-05, {   2,  -2,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.817211346463e-05, -9.715975826076e-06, {   1,  -1,   1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.699330161050e-05,  1.087323078617e-05, {   4,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -3.024367495953e-06,  4.483419720942e-05, {   4,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.266532753272e-05,  3.519314129507e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 {  1.768505054635e-05, -3.017214707237e-05, {   3,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  1.592917361309e-05,  3.189320703669e-05, {   4,  -2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.592917361309e-05, -3.189320703669e-05, {   4,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.834769141309e-06,  3.796336341644e-05, {   0,   0,   2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.090886368666e-08, -4.770511365617e-05, {   4,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.170773349235e-06, -4.060974587430e-05, {   6,   0,  -4,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.619189969534e-05,  2.157353436441e-05, {   4,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.705342578334e-08,  4.772105846601e-05, {   1,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.596829163841e-05, -2.171559192176e-05, {   4,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  2.077680718654e-06, -4.559076409816e-05, {   2,   0,  -1,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 {  7.463928699325e-06, -4.015241461138e-05, {   4,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.533841126457e-07, -4.742329545640e-05, {   3,   0,  -2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.529201498536e-05, -3.228016779863e-05, {   4,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  4.552708817982e-05, -2.038138126384e-06, {   1,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -2.758438028155e-05, -1.996640924722e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   4,   0,   0,   0 } },
 {  7.472544430063e-06,  4.004887802718e-05, {   1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.557718146667e-05, -1.193340381806e-05, {   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.184695861018e-05,  1.563128313935e-05, {   2,   0,  -2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -2.944964626147e-05,  1.793639176397e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -1.368995043016e-06, -4.599509892885e-05, {   0,   0,   1,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 {  3.487075082361e-05,  1.248179891521e-05, {   2,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -4.117376620765e-06, -4.321625019882e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,  -3,   0,   0 } },
 { -7.831428733972e-06, -3.949179490566e-05, {   1,   1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.814771930781e-05,  1.917050137171e-05, {   0,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.034233348165e-05, -6.970561777740e-06, {   1,   0,   2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.034233348165e-05,  6.970561777740e-06, {   1,   0,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.884634753924e-05,  1.846603460335e-05, {   0,   0,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -1.680693198657e-05, -3.049285556604e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -3.365598203961e-05, -1.362046728154e-05, {   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.573441882001e-05,  3.146437870198e-05, {   0,   0,   2,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.679218974194e-06,  3.850977355941e-05, {   3,  -1,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -3.807510091456e-05, -9.104739382916e-06, {   2,   0,  -2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.259854868001e-05,  3.457605192944e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  4.491188694771e-05,  2.230605797387e-06, {   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.146136218494e-05,  1.566337846934e-05, {   2,   0,   0,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 {  7.361117286607e-06,  3.974683865467e-05, {   3,  -1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.630613078571e-05,  1.074990398857e-05, {   4,   0,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.515652419877e-05, -1.890401307366e-06, {   2,   0,  -1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.903327024857e-05,  2.793459820677e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -11,   3,   0,   0 } },
 { -1.903327024857e-05, -2.793459820677e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  11,  -3,   0,   0 } },
 {  1.250179689400e-05,  3.444841803166e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -1.737343940220e-05, -2.954771661496e-05, {   2,   0,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0 } },
 {  2.306219333497e-05, -2.384631864259e-05, {   2,   0,   1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.437695624303e-05, -2.529905047187e-06, {   2,   0,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 {  2.962710041550e-05,  1.726615611212e-05, {   2,   0,  -1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 { -4.467794970546e-05,  2.208224493384e-06, {   4,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.239662044218e-05, -3.445638084610e-05, {   2,   0,  -2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.754498941379e-05, -2.930080371922e-05, {   2,   0,  -3,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.556115174365e-05, -1.121987384417e-05, {   2,   0,  -2,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.428999219915e-06,  3.922390490036e-05, {   1,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.554271759259e-05,  3.109167383487e-05, {   0,   0,   0,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.203163232438e-05, -4.581359377510e-06, {   4,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.309498705348e-05,  3.468935422160e-06, {   2,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.460769386144e-05,  1.947780323582e-06, {   0,   0,   2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -4.275695772596e-05, -3.791845214938e-06, {   4,  -2,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  3.638039260926e-05,  1.012919013467e-05, {   2,  -2,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.637938985826e-05, -1.012923242458e-05, {   2,  -2,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  6.943667134556e-06, -3.952705151421e-05, {   6,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.856567414907e-05,  1.789219027031e-05, {   4,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -5.343814409707e-06,  4.105504309521e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  7.540216581892e-06,  3.883136308904e-05, {   1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.360391749636e-05, -1.275064929939e-05, {   2,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.360391749636e-05,  1.275064929939e-05, {   2,  -1,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.012004994542e-05,  1.622081896249e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   5,   0,   0,   0 } },
 { -9.178338005759e-06,  3.716030932680e-05, {   4,   0,  -2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.488529401561e-05, -1.448640358776e-06, {   1,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.575407182812e-05,  3.053303362907e-05, {   0,   0,   2,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  1.748477965218e-08, -4.626242549388e-05, {   2,   0,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.638924428490e-05, -2.986635229445e-05, {   1,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -4.481730452264e-05,  1.425480772844e-06, {   0,   0,   1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.757832600020e-05,  8.591125696444e-06, {   0,   0,   0,   0,   0,  16, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.893054515494e-05,  7.223909475338e-06, {   1,  -1,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.970411648918e-08,  4.610179782612e-05, {   0,   2,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.712520812688e-05, -8.973650761740e-06, {   2,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.303775958957e-06,  3.872336432756e-05, {   1,  -1,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.055109149117e-05, -1.547325893146e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 {  3.742115172368e-05, -8.602436330269e-06, {   2,   1,  -3,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.782746236072e-05, -2.819523030475e-05, {   4,   0,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  4.450472614236e-05,  1.484366267310e-06, {   0,   0,   1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  2.885927743723e-05,  1.711985705018e-05, {   1,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  7.274070554462e-06, -3.867583704128e-05, {   1,   1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.547308676340e-06,  4.438247555350e-05, {   4,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.707669480007e-05, -8.772662632666e-06, {   1,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.429427341635e-06, -3.937966999214e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  2.054249476871e-07, -4.558160949401e-05, {   4,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.029758045757e-07, -4.537127159679e-05, {   2,   0,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 {  3.760455308801e-05, -8.169304636812e-06, {   2,   0,  -1,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.812515411479e-05, -7.637182989162e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.649688022374e-05, -1.921810285674e-05, {   2,   0,  -2,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.547467684958e-05,  1.022770245037e-05, {   0,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.769814755153e-05,  2.800406208939e-05, {   4,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.437862059766e-08, -4.560525205651e-05, {   2,   0,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.544457682410e-05, -1.020023588062e-05, {   1,  -1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.785277930469e-05,  7.789855301581e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.685763809863e-05,  2.877505013692e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -7,   0,   0,   0,   0 } },
 { -2.816283055336e-05,  1.746682602655e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -2.795301586852e-06, -4.279540906635e-05, {   0,   0,   1,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0 } },
 { -4.557667052450e-05, -1.028859131296e-08, {   4,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.233983118684e-07, -4.536090746488e-05, {   0,   2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.687602786511e-05,  8.689233864277e-06, {   1,  -1,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.438527989764e-07,  4.480571389234e-05, {   2,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.367478562337e-07, -4.469106082527e-05, {   1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -4.388920120005e-05,  1.595786676638e-06, {   0,   0,   1,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  9.707811925590e-06, -3.576844365627e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 { -7.572655487231e-06,  3.788259431934e-05, {   1,  -1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.008891929647e-06,  4.043592615248e-05, {   0,   0,   1,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 {  8.074225142123e-07,  4.462767412013e-05, {   2,   0,   1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.230853018443e-05,  3.308356638194e-05, {   2,   0,   0,   0,   0,   0,  -7,   8,  -3,   0,   0,   0,   0 } },
 {  3.746478813131e-05, -7.914924905497e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 {  3.677904500185e-05, -8.598196924115e-06, {   3,  -1,  -2,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.427666453940e-06, -3.892818714170e-05, {   1,  -1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.107680699210e-06,  4.421849345374e-05, {   4,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.162460217593e-05, -3.687039535022e-06, {   0,   0,   1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -7.208274844502e-06,  3.809929030406e-05, {   2,  -2,   0,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  3.667613813034e-05,  8.615317397660e-06, {   1,   1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.572153180107e-05, -1.952843910672e-05, {   2,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -3.593246347699e-05,  9.267482465795e-06, {   1,  -1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.394335561576e-08, -4.511940757396e-05, {   3,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.243875008028e-05,  3.271085690332e-05, {   2,   0,  -3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.564436044542e-05,  2.950311531860e-05, {   1,   1,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -3.359448886616e-05,  1.153164476404e-05, {   2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.351373341295e-06, -3.777363750737e-05, {   0,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -2.615032369555e-05, -1.891253189600e-05, {   1,   1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.110452248312e-05, -1.394761637866e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 {  3.291113291862e-05,  1.213502669720e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 {  3.293207781311e-05, -1.207376863536e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -1.493896216174e-05, -3.006639271336e-05, {   3,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.029926291602e-05, -2.470297761097e-05, {   2,   0,  -1,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.079800795036e-05,  3.417561276637e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 { -3.814889284380e-05,  6.805728713527e-06, {   2,   0,   1,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -3.814889284380e-05, -6.805728713527e-06, {   2,   0,  -1,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -3.600537737108e-05,  8.889796715588e-06, {   1,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.906912117614e-05, -1.582240750136e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -3.076284498228e-05, -1.409627621921e-05, {   0,   0,   1,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 {  1.702288526571e-05,  2.782463951392e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -1.369946119441e-05, -3.114032342348e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 {  2.464319244544e-05, -2.016784181273e-05, {   1,   1,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.773186776950e-05, -7.079032396282e-06, {   2,   2,  -2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.549857486477e-06,  3.825331722910e-05, {   1,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.534907626930e-05,  2.945168073313e-05, {   0,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 {  3.271865847039e-05, -1.205646320544e-05, {   2,   0,  -1,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 { -1.889377783369e-05,  2.586733565383e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  7.688511364674e-06,  3.702129480441e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -2.346251199112e-08, -4.464944741854e-05, {   2,   0,  -3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.424027301478e-05,  3.041438143722e-05, {   2,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.058454676138e-05, -2.404867113667e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,  -2,   5,   0,   0,   0 } },
 {  4.951147070916e-06,  3.965963958109e-05, {   2,   0,  -2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -4.222379250522e-05,  2.364760721260e-06, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.038082896667e-05,  2.418973255876e-05, {   2,   0,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  1.719708292147e-07,  4.437384999464e-05, {   0,   2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.538408670522e-05,  1.915327269842e-05, {   2,   0,  -1,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 {  1.950084216285e-05,  2.503003132180e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -1.113171014925e-07, -4.440747965944e-05, {   0,   0,   2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.396076577271e-05, -5.283500197980e-07, {   3,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.455980001107e-05, -1.988468417031e-05, {   0,   0,   2,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 { -8.197262843213e-06,  3.620953763918e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -6,   0,   0,   0 } },
 { -1.618482872980e-05, -2.821988548022e-05, {   4,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.935484434777e-05,  1.501027931394e-05, {   3,  -1,   0,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.935484434777e-05, -1.501027931394e-05, {   1,   1,   0,   0,  -2,   2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.564066790248e-05,  2.871808248287e-05, {   4,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.475266894812e-05,  2.958160511085e-05, {   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.018454350278e-05, -2.413528385073e-05, {   4,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.862128731618e-06,  4.242615049251e-05, {   2,   0,  -1,   0,   0,   4,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.041037189051e-05, -3.387381165992e-05, {   4,   0,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.577557983919e-05,  8.505574346634e-06, {   1,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.516327344679e-05,  1.908366661509e-05, {   2,   0,  -1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -9.555634690591e-08, -4.413357582242e-05, {   2,   0,   1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.796620787406e-05, -6.258694311280e-06, {   0,   0,   2,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.700045464766e-05,  7.223124867818e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 {  3.576260696200e-05, -8.436351175797e-06, {   1,  -1,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.212230615303e-06, -3.798226627982e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -6,   3,   0,   0 } },
 {  3.726290942054e-05, -6.918908058077e-06, {   1,   1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  3.541341378880e-05, -8.761401695810e-06, {   1,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.313344639375e-05, -1.003391509928e-06, {   3,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -3.449832103257e-05, -9.606179990038e-06, {   2,   0,   2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -3.449576147247e-05,  9.603329740992e-06, {   2,   0,   2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -3.660116449281e-05,  7.449357627573e-06, {   2,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.503246345988e-05, -8.963730013101e-06, {   1,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.072143292569e-05, -1.325531893582e-05, {   1,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.774362388134e-06,  3.714586107766e-05, {   1,  -1,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.352119821559e-05,  1.032169925966e-05, {   3,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  5.445612285921e-06,  3.839331042913e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   6,  -3,   0,   0 } },
 {  3.350114054127e-05, -1.031284873795e-05, {   3,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.459211245944e-05, -2.921628335109e-05, {   2,  -2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.677871343617e-05, -1.700649041100e-05, {   4,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.698208574454e-05,  1.678556356471e-05, {   2,   0,  -4,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.425124906681e-05, -1.951547525662e-05, {   1,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  7.544986423216e-07,  4.296595087334e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 { -1.691092656487e-05, -2.672253532210e-05, {   0,   0,   2,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.312100409359e-05,  3.050942446369e-05, {   1,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.509441168655e-05,  2.846453822308e-05, {   1,  -1,  -1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  1.907698710744e-07,  4.334451544372e-05, {   2,   0,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.115862446282e-05, -2.233673352771e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 {  4.486578592914e-06, -3.898905996026e-05, {   2,   0,  -1,   0,   0,   0,  13, -19,   0,   0,   0,   0,   0 } },
 { -3.534511881840e-05,  8.119805601920e-06, {   3,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.203165810873e-08, -4.343625826625e-05, {   4,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.417606999514e-05, -2.926653411761e-05, {   2,   0,  -1,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 {  9.820548733835e-06, -3.361479794787e-05, {   4,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.438158528423e-05,  9.050047280783e-06, {   1,   1,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.983948058079e-05,  3.579914391834e-06, {   5,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.983949745030e-05, -3.579726652665e-06, {   5,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.983950494757e-05,  3.579643213009e-06, {   3,  -1,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.983950494757e-05, -3.579643213009e-06, {   1,   1,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.983950807138e-05, -3.579608446481e-06, {   1,  -1,   2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.983950869614e-05,  3.579601493175e-06, {   1,  -1,   2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.983958553224e-05,  3.578746235713e-06, {   1,   1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.983961863481e-05, -3.578377709632e-06, {   3,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.983963237450e-05,  3.578224736452e-06, {   2,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.983963362354e-05, -3.578210829796e-06, {   1,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.983963362354e-05, -3.578210829796e-06, {   1,  -1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.657242281694e-05,  2.681237118324e-05, {   2,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0 } },
 { -5.597704151670e-06,  3.777520538156e-05, {   0,   0,   2,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.102226170939e-05, -2.231198007479e-05, {   4,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -3.379552322086e-05,  9.534335026656e-06, {   0,   0,   1,   0,   0,   3,  -3,  -4,   3,   0,   0,   0,   0 } },
 { -3.379552322086e-05, -9.534335026656e-06, {   0,   0,   1,   0,   0,  -3,   3,   4,  -3,   0,   0,   0,   0 } },
 { -2.110847331588e-05,  2.219416591703e-05, {   4,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.912064896206e-05, -4.152006186179e-06, {   4,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.410308307692e-05, -1.916040403442e-05, {   1,  -1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -3.434445776331e-05, -8.901289760793e-06, {   0,   0,   4,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.434167761642e-05,  8.900780055355e-06, {   0,   0,   4,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.191146852793e-05,  2.128915657864e-05, {   2,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.984955458980e-05, -2.333892478725e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 { -2.679497610970e-07, -4.289460530807e-05, {   3,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.617117911947e-05,  2.697032724032e-05, {   2,   0,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0 } },
 {  3.507407462730e-05, -8.061721165165e-06, {   1,   1,   1,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.291294436914e-05,  2.019848173280e-05, {   4,   0,  -2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.940267465555e-05, -1.370796226665e-05, {   0,   0,   1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.657130468417e-07,  4.281353187897e-05, {   0,   2,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.233353629544e-05, -2.074352977302e-05, {   0,   0,   1,   0,   0,   0,  -2,   8,  -3,   0,   0,   0,   0 } },
 { -9.946755557274e-08, -4.296087763623e-05, {   0,   2,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.669039249056e-05,  2.636707071965e-05, {   0,   0,   2,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.415418114933e-05,  2.889589628495e-05, {   2,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.354640793482e-05, -9.484104234244e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  3.964960701687e-05, -3.371124426239e-06, {   4,   0,  -1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  3.964960701687e-05,  3.371124426239e-06, {   0,   0,   1,   0,   0,  22, -28,   7,   0,   0,   0,   0,   0 } },
 { -3.995223529213e-05, -3.037536176081e-06, {   2,   0,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.738859106618e-05,  1.560095128410e-05, {   4,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.915778905295e-05, -3.823796360259e-06, {   0,   0,   2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -4.024707028401e-06, -3.895301492223e-05, {   0,   0,   1,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0 } },
 { -8.372026761000e-06,  3.459099321273e-05, {   2,   0,   0,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  8.207798459479e-07,  4.212889041010e-05, {   3,   1,  -2,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  8.207798459479e-07, -4.212889041010e-05, {   1,  -1,   0,   0,   1,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.488277026467e-05, -8.054151083808e-06, {   1,  -1,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.540987836064e-05,  2.752550720478e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  1.417758586126e-05, -2.874519478475e-05, {   0,   0,   2,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.020848912198e-06,  3.888339821247e-05, {   0,   0,   1,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0 } },
 { -2.824533224969e-05,  1.464647500953e-05, {   2,   0,  -1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.142620267440e-05,  1.456449041499e-06, {   2,   0,  -1,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 { -8.146389217788e-06, -3.473390489650e-05, {   2,   0,  -2,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.238282974467e-05, -1.049271777987e-05, {   0,   0,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.775907828502e-05, -2.510647574208e-05, {   2,   0,  -1,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 { -2.602772677558e-06,  4.024286288321e-05, {   0,   0,   1,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 { -2.595502740471e-05, -1.688014736587e-05, {   4,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.441307294557e-05,  8.390096073742e-06, {   1,  -1,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.745102125702e-05,  1.534479443752e-05, {   2,   0,   0,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 { -3.663982142866e-05,  6.131213151301e-06, {   0,   0,   1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.457908042877e-05,  8.185371653257e-06, {   3,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.595890470498e-05,  1.679368723796e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -4.115022663058e-05, -1.600229426810e-06, {   0,   0,   2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.279694249194e-05, -2.994174794449e-05, {   2,   0,  -1,   0,   0,   0,  13, -20,   0,   0,   0,   0,   0 } },
 { -1.325516681355e-07,  4.260242779269e-05, {   0,   0,   1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  3.458157308353e-05, -8.140702195691e-06, {   1,   1,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.437093276160e-05,  2.832974019979e-05, {   0,   0,   1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.593082881068e-05,  2.675427827971e-05, {   0,   0,   3,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -3.436047380542e-05,  8.314310051519e-06, {   4,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.949884391129e-05, -2.317029967270e-05, {   2,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.721882670751e-05,  1.544772017983e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  2.106799741845e-07, -4.240165291149e-05, {   4,   0,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -2.713360656659e-05, -1.547418632676e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.412433072330e-05, -2.847252912936e-05, {   0,   0,   1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.599529939604e-05, -1.657772944551e-05, {   0,   0,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -3.020850055982e-05, -1.231722987351e-05, {   0,   0,   2,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  1.520495260976e-05, -2.732076299475e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -3.021070692889e-05,  1.231377362121e-05, {   0,   0,   2,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.596005962685e-06,  3.991567051976e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.308524225435e-05,  1.939548217102e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0 } },
 { -3.267826169799e-05,  9.798098154906e-06, {   2,   0,   0,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.573267109959e-05,  1.674345109026e-05, {   4,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.409277135348e-05, -2.838155054380e-05, {   0,   0,   1,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.279814400905e-07,  4.180685833777e-05, {   6,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.552042157697e-06, -3.985355889277e-05, {   0,   0,   0,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.236666662160e-05,  1.002503628190e-05, {   0,   0,   2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.959757236222e-05, -2.278816094055e-05, {   4,   0,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.301315951742e-05, -1.935765613591e-05, {   3,  -1,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.971449231180e-05,  1.264471243726e-05, {   2,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  1.513679246123e-05, -2.717408641290e-05, {   2,  -2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.403889717898e-05,  2.826160725848e-05, {   1,  -2,   1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.403889717898e-05, -2.826160725848e-05, {   1,  -2,  -1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.543732610626e-05, -2.683741792732e-05, {   2,   0,   1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.086664553324e-05,  3.140149487781e-05, {   2,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 {  3.292281212037e-05,  9.308828994496e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.799750834395e-05, -1.419601581497e-05, {   2,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  3.395835214056e-05,  8.173797636785e-06, {   0,   0,   2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.360986718079e-08, -4.208262945634e-05, {   2,   0,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.390165540385e-05, -2.821549622045e-05, {   2,   0,  -2,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -7.819574192188e-06,  3.425239806327e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.395140505042e-05, -8.108973713906e-06, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 { -4.024902460199e-05, -1.807392159216e-06, {   4,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  1.821732753336e-05, -2.383545631077e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.818032674519e-05, -2.386369039856e-05, {   4,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.144772770689e-05, -3.059396124064e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 {  1.787652582753e-05, -2.409211124700e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.054503491442e-05, -1.142078440820e-05, {   1,  -1,   0,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.410265903252e-05, -1.786230185508e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.410265903252e-05,  1.786230185508e-05, {   2,   0,  -3,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -9.461791190179e-06, -3.248661199903e-05, {   1,  -1,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.034327395954e-05, -1.567341906964e-06, {   4,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.940550001389e-05, -1.250497545744e-05, {   2,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  2.992948419446e-07,  4.160174427292e-05, {   2,   0,   3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.304086335095e-06, -4.058600601194e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  3.541127500881e-05,  6.421749730265e-06, {   2,   0,  -1,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 {  2.113367800490e-08,  4.179978053084e-05, {   0,   2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.653326708509e-05,  2.527538239811e-05, {   0,   0,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -6.525122493265e-08,  4.170155455662e-05, {   3,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.168288986273e-05, -1.007123894606e-05, {   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.584798974108e-05,  2.589897375968e-05, {   0,   0,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -3.833625780313e-05,  3.401608492079e-06, {   0,   2,   0,   0,   0, -23,  22,   0,   0,   0,   0,   0,   0 } },
 { -2.288673646435e-05,  1.885112598693e-05, {   2,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.523313112935e-05,  1.645680365320e-05, {   3,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -7.036227287695e-06,  3.462716686858e-05, {   1,  -1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.699208276209e-05, -4.645920373375e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -1.379368769799e-06, -4.025562179737e-05, {   0,   0,   2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.877125466576e-06,  3.975374724455e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -3.583340401325e-08, -4.158821961664e-05, {   4,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.877707652140e-05, -1.279407310181e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   7,   0,   0,   0 } },
 { -4.020861344767e-07, -4.115850659301e-05, {   2,   0,  -3,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.591381705314e-05,  5.626779430048e-06, {   2,   0,   0,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 {  5.136021795136e-06,  3.635583851217e-05, {   1,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.975398887449e-06,  3.951493830383e-05, {   2,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  1.975398887449e-06, -3.951493830383e-05, {   2,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  1.973038552674e-05, -2.175960102021e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -3.753553182195e-05, -3.929467192032e-06, {   0,   0,   1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 {  8.242888117183e-06, -3.322120104673e-05, {   2,  -2,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.254487249518e-05, -8.879208750700e-06, {   0,   0,   3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.974767367671e-05, -1.653865261009e-06, {   0,   0,   1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  9.331232625289e-06,  3.206990803672e-05, {   1,  -1,   3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  6.114497868064e-08, -4.132805357348e-05, {   6,   0,  -4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.767088422604e-05, -2.370370208564e-05, {   0,   0,   1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -1.590822345064e-06, -3.973151107360e-05, {   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.592630994709e-08, -4.128546220491e-05, {   2,   0,   2,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.668177856330e-05,  4.629889970956e-06, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  3.670261149329e-05, -4.606645367047e-06, {   2,   0,  -2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -3.226941866175e-06,  3.805645679479e-05, {   1,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  1.358766763706e-05,  2.767649397283e-05, {   2,   0,  -1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.722135120145e-08, -4.124479740396e-05, {   2,   0,   3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.026064951968e-05,  1.096893714826e-05, {   2,  -2,   1,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 {  3.026064951968e-05, -1.096893714826e-05, {   2,  -2,  -1,   0,   0,   0,  10, -15,   1,   0,   0,   0,   0 } },
 {  3.134043188729e-05,  9.846601887151e-06, {   4,   0,  -2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.516151576455e-05,  1.599737291223e-05, {   1,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.875592237432e-06,  3.328174727095e-05, {   1,   1,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.346368925411e-08, -4.110295417284e-05, {   2,  -2,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.140702486542e-06, -3.399572724762e-05, {   1,   0,   2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -7.140702486542e-06,  3.399572724762e-05, {   1,   0,  -2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  2.744099730017e-05, -1.369124120375e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 { -4.048158842329e-08, -4.107179391175e-05, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.369340506581e-05,  2.741583737586e-05, {   2,  -2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.167708108487e-08,  4.107024032826e-05, {   2,  -2,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.569712849980e-09,  4.107632398353e-05, {   0,   0,   2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.289388899028e-05, -2.818794485348e-05, {   0,   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -3.285582225440e-08, -4.103185522010e-05, {   0,   0,   2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.405439746106e-06,  3.561279653730e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 { -3.699315259780e-05,  3.984999562352e-06, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  3.298096382898e-05,  7.980540052827e-06, {   4,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.980251012816e-05,  2.115782042835e-05, {   2,   0,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 { -3.388638549998e-05, -7.071839954202e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -2.492963417284e-05, -1.600164859068e-05, {   1,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.848477116297e-05,  1.244046750505e-05, {   2,   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  7.782989862482e-06,  3.314049188016e-05, {   1,  -1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.159846621204e-07,  4.060195463098e-05, {   2,  -2,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -6.153903494703e-06, -3.475573337396e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   5,   0,   0,   0 } },
 {  3.819983492046e-05, -2.700931769149e-06, {   0,   0,   2,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -3.928260993066e-05, -1.614061984874e-06, {   0,   0,   2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.933231807363e-06,  3.695279593712e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 { -2.284908489029e-05, -1.798025691682e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 { -2.718507572560e-06,  3.810861678241e-05, {   0,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  1.665499531589e-05, -2.416652891154e-05, {   2,   0,  -1,   0,   0,   0,  15, -20,   0,   0,   0,   0,   0 } },
 {  2.623138639480e-05,  1.455727885997e-05, {   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.468489018660e-05, -2.607836671869e-05, {   4,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.545258448065e-05, -1.530795682997e-05, {   1,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.777951973368e-05, -2.297724526691e-05, {   2,   0,  -2,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 { -8.907071047863e-06,  3.184800639549e-05, {   4,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.698372281243e-05,  3.765613585580e-06, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 {  3.250013649735e-07, -4.041801817313e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -2.830750705260e-05, -1.242741624967e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   4,   0,   0,   0 } },
 {  1.035703579476e-05, -3.037281275288e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  1.254462421982e-05, -2.818441059693e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 {  2.894792093293e-05,  1.177023794178e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 {  2.221344293783e-05, -1.849303216600e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 { -2.923052600910e-05, -1.145369779246e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0 } },
 {  2.162177356990e-07, -4.045616755203e-05, {   3,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.561976418558e-05, -1.502688973531e-05, {   2,  -2,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.848385539643e-06, -3.377729514854e-05, {   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.841904642464e-05,  1.220128432415e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 {  1.716368697714e-05, -2.344499209291e-05, {   2,   0,  -2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  3.280475866046e-05,  7.769857028838e-06, {   3,  -1,  -1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.399605035635e-05,  2.655330585981e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0 } },
 { -3.839075837439e-05, -2.127884257385e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  2.688357707815e-06,  3.781881458911e-05, {   2,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.434178939101e-07,  4.026116991267e-05, {   0,   0,   1,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.279526305237e-05, -7.695512716207e-06, {   3,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.993131673103e-05, -1.055100392096e-05, {   1,  -1,   2,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -7.351641733814e-06,  3.311299927402e-05, {   3,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.233051793862e-05,  8.114871201932e-06, {   1,  -1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.023431984880e-05, -1.948965728080e-07, {   2,   0,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.536957993290e-05, -5.059297849240e-06, {   2,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.357340966841e-06,  3.404573575085e-05, {   3,  -1,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.286527510362e-06,  3.410853789833e-05, {   1,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.284509355160e-05,  7.547652321940e-06, {   1,  -1,   1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.753522428330e-05, -2.799493321084e-06, {   4,   0,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.442735335565e-05, -5.848166719477e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0 } },
 {  2.693218019195e-05, -1.331634997715e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0 } },
 { -1.539096890812e-08, -4.022890848131e-05, {   0,   0,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -3.167326728942e-06,  3.707560569865e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 { -5.309881253789e-06,  3.492610761065e-05, {   0,   0,   2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.988322177741e-08, -4.020825868290e-05, {   1,   0,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  2.105562814021e-05,  1.916624735450e-05, {   4,   0,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  3.478271470226e-05,  5.433367495155e-06, {   2,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.842197165687e-05,  1.778300662366e-06, {   2,   0,   0,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  2.847570851536e-05,  1.172302740027e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 { -3.518753268598e-05, -4.989325157413e-06, {   4,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -3.359825454327e-05, -6.527478892510e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 {  3.252901796455e-05, -7.586709206520e-06, {   1,  -1,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.153307177814e-07, -3.999966746787e-05, {   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.893872004988e-08,  4.002976664136e-05, {   4,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.938701692688e-08, -4.002980010167e-05, {   4,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.774725923597e-05,  1.233654094638e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.458892006736e-08, -4.001271289158e-05, {   0,   2,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.689076856156e-05,  1.312627177852e-05, {   2,   0,  -1,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 { -1.295804389979e-05, -2.705714505064e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.293854679758e-06,  3.871981588261e-05, {   2,   0,  -2,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 { -2.503246028706e-05, -1.497627536354e-05, {   2,   0,   0,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0 } },
 {  1.926843493473e-09, -3.999999995359e-05, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.870992957454e-09, -3.999999995624e-05, {   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.047197551092e-09, -3.999999998629e-05, {   2,  -1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  1.026253600058e-09,  3.999999998684e-05, {   2,   1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  9.704030640141e-10,  3.999999998823e-05, {   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -9.354964790028e-10, -3.999999998906e-05, {   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -7.819075048438e-10,  3.999999999236e-05, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  7.330382858110e-10, -3.999999999328e-05, {   2,  -1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.026548245620e-10,  3.999999999684e-05, {   2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.817108735429e-10,  3.999999999710e-05, {   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.328416544846e-10,  3.999999999766e-05, {   2,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.979350694356e-10, -3.999999999802e-05, {   2,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.560471674159e-10, -3.999999999842e-05, {   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.954768762251e-10,  3.999999999952e-05, {   4,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.188790205123e-11,  3.999999999998e-05, {   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  4.000000000000e-05,  0.000000000000e+00, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 { -1.368029006292e-05,  2.630647733401e-05, {   1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -4.830909070085e-06,  3.515247371673e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -2,  -5,   0,   0,   0 } },
 {  1.373849523014e-05,  2.624451059073e-05, {   2,   0,  -2,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  3.557368153510e-06, -3.641984133325e-05, {   2,   0,  -1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.295428919464e-05, -1.701870035896e-05, {   1,  -1,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  3.125726883901e-05,  8.703249727704e-06, {   2,  -2,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.125709361496e-05, -8.702695775895e-06, {   2,  -2,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.466789953765e-05,  2.527471807287e-05, {   0,   0,   3,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.297953084904e-06, -3.563913593480e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 { -1.992752729183e-05,  1.998236887142e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -3.653224609251e-05,  3.374584151677e-06, {   2,   0,  -1,   0,   0,   0,   9, -14,   3,   0,   0,   0,   0 } },
 {  3.626166901301e-05,  3.639739067090e-06, {   0,   0,   2,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.944299365912e-05,  1.045535984731e-05, {   0,   0,   2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -3.244585760519e-05,  7.446206747048e-06, {   1,   1,  -2,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.943043558680e-05, -1.045017352523e-05, {   0,   0,   2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.963590136286e-05, -1.018561425671e-05, {   2,   0,  -1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0 } },
 { -1.669682614065e-05, -2.312286604972e-05, {   1,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.497624174626e-08, -3.968744069083e-05, {   4,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.776888974286e-05,  1.200863876850e-05, {   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.043123099512e-05,  2.929908547833e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.521342893153e-05, -2.451345936245e-05, {   0,   0,   0,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 { -2.482508985236e-05,  1.488286476148e-05, {   1,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.936093684949e-05,  2.034128163076e-05, {   2,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  1.984465126742e-05, -1.985024157368e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -3,   0,   0,   0 } },
 {  1.918837432156e-05,  2.049721809553e-05, {   4,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -1.227622934335e-05,  2.737323863027e-05, {   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.223306344444e-05,  2.739255663067e-05, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.589724013075e-05, -2.369129097121e-05, {   2,   0,   0,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 { -2.371906135859e-06, -3.720026019025e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  3.604336511382e-05, -3.527256323434e-06, {   0,   0,   1,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 { -9.752219266395e-06, -2.981041327663e-05, {   1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.978658763972e-08,  3.951646552265e-05, {   4,   0,  -2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.206209611264e-05,  2.746826964644e-05, {   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.781523560611e-05,  1.692731164539e-06, {   2,   0,  -3,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -1.318023756660e-05,  2.631597331808e-05, {   3,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.431799707265e-05, -2.517235292978e-05, {   2,   0,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -2.850271435948e-05, -1.097818380332e-05, {   0,   0,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.755376629047e-05,  1.186549464667e-05, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.416892428520e-05, -1.523256074145e-05, {   4,   0,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.613531494932e-05,  1.323984465143e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  6.463845051969e-06,  3.287024450646e-05, {   1,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.597862873865e-05,  1.334247727507e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 {  1.304170601292e-05, -2.625769086988e-05, {   0,   0,   1,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.808762946943e-05,  2.119699330426e-05, {   2,   0,  -1,   0,   0,   0,  -9,  14,   0,   0,   0,   0,   0 } },
 { -1.720521098515e-05, -2.207611548538e-05, {   0,   0,   2,   0,   0, -23,  22,   4,   0,   0,   0,   0,   0 } },
 {  1.116533042957e-05, -2.811257516666e-05, {   2,   0,   2,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  3.728967848987e-05, -1.970184136524e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  5.237258015343e-06,  3.401335605443e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -5.237258015343e-06, -3.401335605443e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 {  3.388636813576e-05,  5.363577172187e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0 } },
 { -2.839741872939e-05, -1.080411763471e-05, {   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.786050945402e-07,  3.840093620817e-05, {   4,   0,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.265464524697e-05, -2.652185088158e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  9.344236333660e-06, -2.982970506743e-05, {   2,   0,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -4.686103853584e-07, -3.865537757741e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  4.893614144936e-08,  3.906421311675e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,  -4,   0,   0,   0 } },
 {  2.501310170580e-05,  1.409827018445e-05, {   1,   1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.295184649551e-05, -6.146071872807e-06, {   2,   0,   0,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.698436946626e-05, -1.209804110557e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 {  1.301625694988e-05,  2.606108359291e-05, {   4,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.338057828072e-05,  1.569588817179e-05, {   1,  -1,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  3.323364458416e-05, -5.835622248180e-06, {   0,   0,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.743695075451e-08, -3.900869221003e-05, {   2,   0,   2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.739405204192e-05,  1.669146255031e-06, {   2,   0,   1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.894752448389e-05,  1.008694155842e-05, {   2,   0,  -1,   0,   0,   0, -10,  15,   0,   0,   0,   0,   0 } },
 { -4.136679107618e-06, -3.488248105905e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  3.041321312086e-05,  8.605640980729e-06, {   2,   0,  -1,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 {  1.507860128585e-05, -2.393155270666e-05, {   2,   0,  -1,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0 } },
 {  3.853898650491e-05,  4.701192057543e-07, {   1,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.787145147025e-05,  1.107022578577e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 { -4.689622367496e-07,  3.846442234589e-05, {   3,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.414932720921e-06, -3.248656449967e-05, {   0,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -3.149020471880e-05,  7.398310639809e-06, {   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -2.020417869842e-05, -1.866438256194e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0 } },
 { -3.146424103351e-05, -7.401794166306e-06, {   1,  -1,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.259439393796e-05, -6.252028171876e-06, {   5,  -1,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.093257845581e-05, -2.790449622808e-05, {   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.743193521281e-05,  1.139074988555e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  2.322759570915e-05, -1.559334870185e-05, {   1,  -1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.174897271821e-06, -3.763086117956e-05, {   2,   2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  2.262880055234e-05, -1.617285368703e-05, {   0,   0,   2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  4.545218123660e-08, -3.872950891883e-05, {   3,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.603232218845e-05, -2.272936623367e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  1.060964102880e-07, -3.865376296837e-05, {   0,   0,   1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.756390255404e-05,  1.195560769629e-06, {   0,   0,   1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.091337206242e-05, -1.783616638548e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1, -11,   3,   0,   0 } },
 { -1.472253298211e-05, -2.400773960142e-05, {   1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -2.578707371165e-05,  1.293914228092e-05, {   2,   0,   0,   0,   0,   0,  16, -20,   0,   0,   0,   0,   0 } },
 { -2.568257945144e-08,  3.869523167610e-05, {   2,  -2,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.514606690358e-05, -3.572103271836e-06, {   0,   0,   0,   0,   0,   0,   6,  -8,   3,   0,   0,   0,   0 } },
 { -3.245141706897e-05,  6.213386590180e-06, {   1,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.181653085151e-05, -2.683814268276e-05, {   1,  -1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  2.060497027302e-05,  1.802174282915e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 {  6.223745451253e-06, -3.238631490997e-05, {   2,   0,   1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -3.118410151248e-05, -7.372370992174e-06, {   1,  -1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.120534928552e-05, -1.733582163349e-05, {   3,  -1,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -2.043638168637e-06, -3.649066371668e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,  -5,   0,   0,   0 } },
 { -2.721275582015e-05, -1.129923608351e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -3.085273774768e-05,  7.651853986457e-06, {   2,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.693714919869e-06, -3.078385974948e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2, -10,   0,   0,   0 } },
 {  1.457259343375e-05,  2.388149170975e-05, {   2,   0,  -1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0 } },
 { -1.409809503807e-07, -3.828557405662e-05, {   0,   0,   1,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.207461591916e-05, -2.630988584464e-05, {   1,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.166291969756e-05,  2.669789024398e-05, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 {  2.575753235745e-05,  1.259807855326e-05, {   2,   0,   0,   0,   0,  -6,   9,   0,  -3,   0,   0,   1,   0 } },
 {  2.743290081184e-05, -1.090216888912e-05, {   0,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.276674435754e-05, -2.556801620477e-05, {   4,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.677837798427e-05,  2.155637814085e-05, {   2,   0,  -1,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  3.627585203717e-05, -2.007785532780e-06, {   2,   0,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.985902373779e-06,  3.626918341290e-05, {   2,   0,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -2.628423338264e-06, -3.561645127854e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  6.216291562912e-06, -3.198630051825e-05, {   4,   0,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.039826058745e-05, -1.779749796119e-05, {   2,   0,   1,   0,   0, -10,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.435895947665e-05,  3.814478313042e-06, {   4,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.250657710932e-05, -5.623654907480e-06, {   2,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.314571508287e-05, -2.496827608012e-05, {   2,   0,  -2,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  9.630210814320e-06, -2.844020039014e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  6.117251009938e-06, -3.195095395488e-05, {   1,  -1,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -8.069923317222e-07,  3.724865019771e-05, {   2,   0,  -3,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  5.563925356389e-06,  3.248250649721e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.357115532433e-06,  3.665564083539e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 {  3.303186616757e-05, -4.979671426530e-06, {   2,   0,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.255821937146e-05, -2.545324012995e-05, {   2,   0,  -2,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  2.810869929275e-06,  3.519994137978e-05, {   2,  -2,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  2.928391786909e-05, -8.665074907565e-06, {   1,  -1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.064962108501e-05, -7.293880076860e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 {  3.237240648919e-05, -5.543936631714e-06, {   2,   0,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.352219615836e-06,  3.454815295499e-05, {   2,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 { -1.656446257841e-06, -3.624388215492e-05, {   4,   0,  -1,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -6.593931567489e-06,  3.128528174916e-05, {   1,  -1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.670504815809e-05, -2.115116751807e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,   0,   0,   0,   0 } },
 {  1.352897996663e-05,  2.432504421392e-05, {   1,   0,  -3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.741713702511e-05,  2.043312298990e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.858280526169e-08,  3.782360817359e-05, {   3,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.384116326706e-05, -2.398832976492e-05, {   0,   0,   3,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  2.639111223878e-05,  1.137398592733e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.008290545652e-06, -3.175254504839e-05, {   2,  -2,   2,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  3.140795209578e-05,  6.308866328025e-06, {   2,   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -2.940614233924e-05,  8.303309794859e-06, {   2,   0,   0,   0,   0,  -8,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.473056882545e-06, -3.323052379830e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -3.572775688825e-05, -1.971556564518e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  3.049934671145e-05,  7.197210442925e-06, {   1,  -1,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.090359515926e-06, -2.858960237344e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.654372011084e-05,  1.109529328787e-05, {   2,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -3.279310458644e-06,  3.434322827245e-05, {   2,   0,   0,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -3.239021289299e-05, -5.218743439345e-06, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -4.768393547863e-06, -3.278111435207e-05, {   0,   0,   1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.493769342987e-05, -1.258722251562e-05, {   2,   0,  -1,   0,   0,   0,  16, -20,   0,   0,   0,   0,   0 } },
 {  1.990689932218e-06,  3.552468743651e-05, {   0,   0,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  3.163778967562e-05,  5.862040162051e-06, {   3,   1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.505855351316e-06, -3.299183180934e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  3.602664673002e-05,  1.440441104005e-06, {   2,   0,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.895186354630e-06, -3.555912753088e-05, {   4,   0,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -1.633478617579e-05,  2.110703498531e-05, {   2,   0,   0,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 {  2.729142746567e-05,  1.014299807573e-05, {   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.911427543116e-05,  8.309089461165e-06, {   2,   0,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.580145933499e-05,  2.161541688624e-05, {   2,   0,   0,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 { -1.223345174567e-05,  2.518045221633e-05, {   2,   0,  -2,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 {  5.853980369896e-06, -3.155848968486e-05, {   1,   1,  -2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.884608692141e-08, -3.736340826852e-05, {   4,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.154712602796e-05,  5.850029567611e-06, {   3,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.112930678038e-05,  1.624421814066e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -6.219613095192e-06,  3.114851172402e-05, {   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.323060415760e-05, -2.412984899027e-05, {   0,   0,   3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.636475569762e-05,  1.099077649898e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -3.131940130879e-05,  6.027883121833e-06, {   2,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.331382238806e-05, -2.401531436289e-05, {   0,   0,   1,   0,   0,  12, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.002466055362e-05,  7.293109667019e-06, {   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.091566191705e-06, -3.521154772577e-05, {   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.490595471794e-06, -3.078060405631e-05, {   1,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -3.201997926954e-05, -5.241340749838e-06, {   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.086411344073e-05, -2.637812380107e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0 } },
 { -1.171420523483e-05,  2.552005523315e-05, {   1,  -1,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.171420523483e-05, -2.552005523315e-05, {   1,  -1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.494428221070e-05,  2.228020785449e-05, {   2,   0,  -2,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 { -1.718397199740e-08,  3.720271056286e-05, {   3,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.440393601448e-09,  3.719960244091e-05, {   2,  -2,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.910234513838e-05,  1.806717234121e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -5,   0,   0,   0 } },
 {  9.917632180165e-06, -2.724130618443e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -2.865199712677e-05, -8.502169492199e-06, {   6,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.737691560655e-05, -1.976956667022e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0 } },
 { -2.862396616536e-05,  8.491530501831e-06, {   6,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.846093396395e-10, -3.704915083487e-05, {   2,  -2,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.116466491553e-05, -2.587541857415e-05, {   2,   0,   0,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 { -6.686952303778e-06,  3.034066770909e-05, {   2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.634262147728e-05,  1.066483161253e-05, {   1,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.854558951882e-05, -8.457217237880e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 {  1.986336337808e-08, -3.697878096782e-05, {   6,   0,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.976357818315e-06,  3.001785583341e-05, {   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.483891951558e-05,  1.215116034504e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 {  2.956543234065e-05,  7.377695995958e-06, {   1,   1,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.968202262369e-06,  2.994907674155e-05, {   2,   0,  -1,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  1.574007391813e-08, -3.687717166409e-05, {   4,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.386266114251e-08,  3.680473160324e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -2.944854499799e-05, -7.408215604001e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -4,  10,   0,   0,   0 } },
 { -3.159576628097e-07,  3.652673843413e-05, {   1,   0,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  5.753518264153e-06, -3.108866689546e-05, {   1,   1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.708362204299e-05,  9.728411194781e-06, {   2,   0,   1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.572168336359e-05,  2.107792076168e-05, {   2,   0,  -1,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0 } },
 { -1.547941397920e-05,  2.131829950728e-05, {   2,   0,   0,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 { -2.658784869740e-05, -1.020806374214e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -3.416331267036e-05,  2.622847895153e-06, {   2,   0,   0,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.466954022421e-06,  3.029955838843e-05, {   2,   0,  -2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  3.413873387283e-05, -2.619364727800e-06, {   4,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.984381953737e-05,  6.904847850110e-06, {   2,   0,   1,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  2.530938702505e-05,  1.143309859062e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  9.336294672414e-06, -2.738713855530e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.945584776329e-05, -7.258655126164e-06, {   2,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.464658384323e-06,  3.124637033171e-05, {   1,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.209215413613e-05,  2.457854563927e-05, {   2,   0,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.625880709117e-07,  3.599629067815e-05, {   2,   0,  -2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -3.111984717033e-05, -5.529529918477e-06, {   2,   0,   1,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 { -3.579130749150e-05,  8.409761630956e-07, {   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.483243505212e-05,  1.178669851137e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.366858941221e-05, -2.291032405297e-05, {   1,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.688642806481e-06, -2.988949433189e-05, {   0,   2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.623557244825e-08,  3.651000507159e-05, {   1,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  5.288267022599e-07,  3.601156932940e-05, {   0,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.619892010775e-07, -3.635693641985e-05, {   2,   0,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.863620176841e-09, -3.647668298653e-05, {   4,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.347965694106e-05,  2.975515257619e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.587149199418e-05, -5.468183449000e-07, {   2,   0,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.238448433318e-05, -4.029311264932e-06, {   2,   0,   0,   0,   0,  -5,   5,   4,   0,   0,   0,   0,   0 } },
 { -2.586262981006e-05,  1.054764162741e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 {  2.808580518247e-05,  8.313242256324e-06, {   4,   0,  -2,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.925016615625e-05, -1.714408832026e-05, {   3,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.594892518378e-05,  2.044266829050e-05, {   0,   0,   0,   0,   0,  23, -22,  -4,   0,   0,   0,   0,   0 } },
 {  1.124249035419e-05,  2.514172321510e-05, {   0,   0,   2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -9.235317553013e-07, -3.545742530942e-05, {   2,   0,  -1,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.873108245596e-07, -3.586847287027e-05, {   0,   2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.823152302522e-05, -1.806737580581e-05, {   0,   0,   2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  2.858643607120e-05,  7.682971384202e-06, {   1,   1,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.352075914433e-06, -3.289453440240e-05, {   0,   0,   1,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 { -9.271119885701e-06,  2.697409783040e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -1.352293354942e-05, -2.270844930201e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 {  5.926441544181e-06,  3.030334150082e-05, {   1,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.154089803414e-05,  2.465829281833e-05, {   2,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.564204569361e-05,  2.055110593396e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  4.332996357293e-07, -3.575977100110e-05, {   2,   0,  -2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.187157083561e-05,  2.431496816451e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   2,   0,   0,   0 } },
 {  2.267920407990e-05,  1.350634619877e-05, {   3,  -1,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.933997320300e-05,  6.833861975839e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 { -1.204779925361e-05, -2.412201154002e-05, {   1,  -2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.204779925361e-05,  2.412201154002e-05, {   1,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  5.825525089489e-06,  3.033516051933e-05, {   1,   1,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.212776400470e-05, -2.402486758913e-05, {   0,   0,   1,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.954581849164e-05, -6.596999877354e-06, {   4,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.206664623234e-06,  3.293562799838e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   2,   0,   0,   0 } },
 {  2.045663803018e-07, -3.593566302221e-05, {   0,   2,   0,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.549104553883e-05,  1.062224638669e-05, {   1,  -1,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  3.462756789142e-05,  1.478216133477e-06, {   0,   0,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -3.092876196876e-05, -5.173498005163e-06, {   3,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.748822074112e-05,  8.603545314409e-06, {   2,   0,   1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.877438201363e-05, -7.317072597224e-06, {   2,  -2,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  5.836198081803e-06, -3.024719092700e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 {  1.705876933402e-05, -1.900305417555e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -2.916979529195e-05, -6.880227033930e-06, {   1,   1,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.287501288374e-06,  2.776132251154e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  3.526681848869e-05,  7.412453507612e-07, {   3,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.120496620165e-05,  1.476755721722e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,  -5,   0,   0,   0 } },
 { -1.876801285548e-07,  3.577126251452e-05, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.240815394026e-05, -1.354976094134e-05, {   3,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.156322983000e-06,  3.078294116514e-05, {   1,  -1,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.838493702779e-05, -7.537649865799e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  4.124746889771e-06, -3.178662274868e-05, {   2,   0,  -2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.324509726408e-05, -2.264103894526e-05, {   0,   2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.120758140372e-05,  1.465664886692e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -9.053046795651e-09,  3.584707583846e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 { -3.428738654011e-05, -1.521665962058e-06, {   2,   0,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0 } },
 { -1.122579353809e-07, -3.562496724682e-05, {   2,   0,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.295891434120e-05, -1.274804973217e-05, {   2,   0,  -1,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  3.368499152910e-05,  1.990102883822e-06, {   2,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.511035458755e-05, -2.053653284851e-05, {   0,   0,   0,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  2.848801885477e-06, -3.279655280541e-05, {   1,   1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.527410746394e-05, -3.694739507130e-07, {   2,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.498623295020e-06, -2.912235660786e-05, {   4,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.290250652589e-05,  1.268351141360e-05, {   2,   0,  -1,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 {  1.204673462130e-05,  2.352512038462e-05, {   4,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.440448314091e-08, -3.554074918755e-05, {   4,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.055806276646e-05, -1.499441648969e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 { -1.785089533230e-05, -1.769040167632e-05, {   2,   0,  -1,   0,   0,   0,   6,  -8,   1,   0,   0,   0,   0 } },
 { -1.660019925884e-05,  1.893832103061e-05, {   0,   0,   1,   0,   0,   3, -11,  12,  -3,   0,   0,   0,   0 } },
 { -1.660019925884e-05, -1.893832103061e-05, {   0,   0,   1,   0,   0,  -3,  11, -12,   3,   0,   0,   0,   0 } },
 {  3.479239323077e-06, -3.205488871330e-05, {   1,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.487125553166e-05, -2.060819094283e-05, {   0,   0,   1,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 {  3.645681837450e-06, -3.182139144134e-05, {   0,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.411880503879e-07, -3.491169351165e-05, {   1,  -2,   1,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 { -5.411880503879e-07,  3.491169351165e-05, {   1,  -2,  -1,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 { -2.820779542242e-05,  7.237344803613e-06, {   1,   1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.232033123209e-05, -1.309829042628e-05, {   0,   0,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  1.696063294116e-07, -3.520444880576e-05, {   1,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -2.032419680045e-05,  1.504847050225e-05, {   1,  -1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.016207132093e-05,  2.519103492886e-05, {   4,  -2,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.858574449144e-05,  6.763213477184e-06, {   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.507956660676e-05,  1.025275115877e-05, {   2,   0,  -1,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 {  3.317891317541e-05, -2.132037083281e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 {  1.163983169856e-05,  2.366020645822e-05, {   2,   0,  -1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.455709692278e-05, -1.072788118409e-05, {   1,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.187424262826e-05,  2.340905307344e-05, {   0,   0,   1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.136365737986e-05, -3.874645335621e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -2.741372574301e-05,  7.820638296215e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.518892290297e-05, -1.004476814256e-05, {   2,   0,  -2,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  2.575709697940e-05, -9.447795118095e-06, {   1,  -1,   0,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -2.392500584892e-05,  1.127737191622e-05, {   1,   1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.469643349405e-05,  1.050485828953e-05, {   0,   0,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.214388059419e-05, -3.048263777940e-06, {   2,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.304400446560e-05,  2.123758304540e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 {  2.370138077410e-05,  1.146294386936e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -9.100670935395e-06, -2.606109001253e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  1.158503364832e-05, -2.355332087726e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 { -3.043779626389e-08, -3.507517401940e-05, {   2,   0,   2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.097456402023e-05, -2.411993374548e-05, {   1,  -1,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.521061270581e-08,  3.504831315679e-05, {   2,   0,   0,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 {  2.167182863981e-05, -1.341485760213e-05, {   4,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.831071199899e-05, -6.764252550331e-06, {   1,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.564191151226e-05,  9.413372072117e-06, {   2,   0,   0,   0,   0,   0,  -2,   2,   0,  -2,   0,   0,   0 } },
 {  2.814477424075e-08, -3.502603736151e-05, {   2,   0,   1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  9.535505575998e-06,  2.551654985205e-05, {   3,  -1,  -1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  9.535505575998e-06, -2.551654985205e-05, {   1,   1,   1,   0,   0, -29,  33,   0,  -2,   0,   0,   0,   0 } },
 {  3.499227098152e-05, -3.755257243148e-08, {   1,   1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.860566026835e-05,  6.348222417574e-06, {   2,   0,   0,   0,   0, -11,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.652266956491e-05, -8.427712099608e-06, {   0,   0,   1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  8.914041007675e-06, -2.603409595130e-05, {   2,   0,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 {  9.686165068832e-06, -2.525768145435e-05, {   4,   0,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -3.035883156795e-05,  4.568498350633e-06, {   2,   0,  -1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.291375638686e-05, -2.199528979610e-05, {   0,   2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.481786248177e-05, -8.122477025960e-08, {   5,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.027528325867e-06, -2.886772282258e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -3.065751192800e-08,  3.484400475577e-05, {   4,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.362799195917e-05,  1.124030031932e-05, {   4,   0,  -1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -2.362799195917e-05, -1.124030031932e-05, {   0,   0,   1,   0,   5,   0, -13,   7,   0,   0,   0,   0,   0 } },
 {  2.739531391416e-05, -7.462503168381e-06, {   2,   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  2.189490160578e-05,  1.293232225509e-05, {   0,   0,   1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 { -2.220129611669e-05, -1.262355273126e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.492421877664e-05, -9.857401852196e-06, {   0,   0,   1,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 { -1.706886413400e-08,  3.475823297390e-05, {   1,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.816146200649e-05,  6.612763488809e-06, {   1,   1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.121842815886e-05, -1.352701583897e-05, {   2,   0,  -1,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 { -2.821646194588e-05, -6.527160822470e-06, {   2,  -1,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.046240354830e-05, -2.427551304283e-05, {   4,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.890141719969e-05,  1.583014730193e-05, {   3,  -2,   0,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -1.890141719969e-05, -1.583014730193e-05, {   1,   2,   0,   0,   0, -20,  18,   0,   0,   1,   0,   0,   0 } },
 { -1.402947215156e-05,  2.069361280125e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 { -2.336460680063e-05, -1.132472231352e-05, {   2,   0,  -2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  2.806918848373e-05, -6.606464319403e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -2,   0,   0,   0 } },
 {  1.116837825590e-05,  2.348737175060e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  5.529959606026e-08, -3.459912362652e-05, {   1,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.132350659591e-05,  2.331598918396e-05, {   1,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.207994832999e-06,  2.938863001582e-05, {   2,   0,  -5,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.784225110362e-05,  6.737318467174e-06, {   2,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.793795108840e-07, -3.437657757064e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  3.353290954145e-07, -3.419904084348e-05, {   4,   0,  -3,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.665662093302e-05,  1.786596562416e-05, {   2,   0,  -2,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.169749810132e-05,  2.282429957463e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 {  4.187788746818e-08,  3.447652390270e-05, {   2,  -2,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.731368858244e-05,  7.204579232626e-06, {   2,   0,   0,   0,   0,  -9,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.612359213664e-05, -8.392482282306e-06, {   2,   0,  -2,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 { -2.834723486683e-08,  3.447976640514e-05, {   1,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.551944520070e-07,  3.425007082315e-05, {   4,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.289983848407e-05,  1.158658150108e-05, {   1,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -2.850322687444e-05,  5.969776990824e-06, {   0,   0,   2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.317356004440e-05,  2.128657995594e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   6,  -5,   0,   0,   0 } },
 {  1.110937151562e-06,  3.334578236044e-05, {   2,   0,   0,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.865884721602e-05,  1.579696362012e-05, {   2,   0,   0,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.500121472090e-07, -3.429427120508e-05, {   0,   0,   1,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.327059819601e-05,  1.167181587341e-06, {   2,   0,  -2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.261646956317e-07,  3.408794135480e-05, {   2,   0,   0,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  1.322814340236e-05, -2.117816959068e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 { -6.007822131542e-06,  2.839611359096e-05, {   1,  -1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.793366035364e-05,  6.461277619764e-06, {   2,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.441540348550e-05,  1.997121254154e-05, {   2,   0,   0,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 {  2.175476855490e-06, -3.220703535016e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   6,   0,   0,   0 } },
 { -5.339092826374e-06, -2.903589347266e-05, {   0,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.773242434800e-07,  3.419193671264e-05, {   0,   0,   0,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  2.568824181357e-05,  8.679513250069e-06, {   0,   0,   2,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.203433091455e-07, -3.390982225813e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 { -1.391086225454e-09, -3.431684282254e-05, {   1,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.693584787095e-06,  2.761947570803e-05, {   4,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.991157646498e-05,  4.399517711544e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  6.997449505790e-06,  2.731203083771e-05, {   2,  -2,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.704649070420e-06,  2.760216851618e-05, {   3,  -1,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.862798949070e-05, -1.567119862463e-05, {   2,   0,  -2,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  2.986508345468e-05, -4.428394578229e-06, {   2,   0,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.759528282165e-06, -3.052527444957e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 {  2.709161438820e-05,  7.183597642225e-06, {   4,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  9.896898436460e-12, -3.427226913225e-05, {   0,   2,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.310004447042e-05,  1.159462990066e-06, {   0,   0,   2,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.724094923691e-05,  1.700865337892e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  2.370869036938e-05,  1.053671721187e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  2.777662735076e-05, -6.467901461095e-06, {   1,  -1,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.507955847836e-06, -3.271992561134e-05, {   0,   0,   1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.158049761559e-06,  2.504893701666e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  6.219218658457e-06,  2.798463358637e-05, {   2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.880208066253e-06,  2.432067293811e-05, {   2,   0,  -1,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.113862694804e-08,  3.417958586203e-05, {   4,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.765646005700e-05,  6.534081392982e-06, {   2,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.182847574818e-05,  2.352511848011e-06, {   0,   0,   1,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 { -2.701259359552e-05, -7.165786398446e-06, {   1,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -2.701259359552e-05,  7.165786398446e-06, {   1,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.118765338992e-05, -2.298302731534e-05, {   2,   0,  -1,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.942134113268e-07, -3.351842034864e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -5.996484474146e-07,  3.351083415187e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 {  5.813719123920e-06, -2.829485706795e-05, {   2,   0,   0,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  1.141538716038e-05, -2.268184896584e-05, {   0,   0,   1,   0,   0,  14, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.515904846256e-05, -8.933073221164e-06, {   4,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.705391663919e-05,  7.033815672370e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 {  3.961066230900e-06, -3.012609996274e-05, {   1,   1,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.515283378101e-05,  8.931091495693e-06, {   4,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.144401620906e-05,  1.263907982480e-05, {   0,   0,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -5.339949840893e-06,  2.869424508597e-05, {   0,   0,   3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -2.934997339294e-05,  4.675351031580e-06, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -1.445246216957e-05,  1.955499283041e-05, {   2,   0,   0,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  1.972335387148e-06,  3.201208876141e-05, {   2,   0,  -1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 {  2.757780513822e-05, -6.397479984079e-06, {   3,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.672798699160e-05,  7.203039218297e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 { -2.740778677827e-05, -6.518176497440e-06, {   2,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.274702864917e-05,  2.117745561742e-05, {   2,   0,   1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.784878692002e-05,  6.057792524470e-06, {   2,   0,   1,   0,   0, -18,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.740250486903e-05,  6.449739108290e-06, {   1,  -1,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.703588990088e-05,  6.812290914380e-06, {   2,   0,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.583563510594e-05, -8.011721502192e-06, {   4,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.902373594353e-05,  4.798178008591e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,  -5,   0,   0,   0 } },
 {  2.765038505893e-05,  6.160644617740e-06, {   1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.401983783333e-05,  9.781907465717e-06, {   4,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.012812603900e-06, -3.273985847390e-05, {   6,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.045742933073e-05,  1.328868734200e-05, {   1,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.561504014933e-05, -8.108603000289e-06, {   4,   0,  -2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.230475782126e-06,  2.949218386639e-05, {   0,   0,   2,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 {  7.258123021430e-06,  2.646310623878e-05, {   0,   2,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.880082911869e-08, -3.369583745595e-05, {   0,   0,   2,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  7.256625410534e-06, -2.645537442948e-05, {   0,   2,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  3.503109865711e-06, -3.017784150588e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 {  2.263826826925e-05,  1.103857026772e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0 } },
 { -1.679972987335e-07, -3.349810819551e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.568597229657e-05,  7.968108217901e-06, {   4,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.631240792483e-06, -2.800601495299e-05, {   2,   0,   0,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -5.631240792483e-06,  2.800601495299e-05, {   2,   0,  -2,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 { -2.533936541993e-05, -8.296959430708e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.137150897133e-06,  3.244128971920e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  2.582260353823e-05,  7.743429583657e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 { -2.219904166034e-05, -1.134533170374e-05, {   3,  -1,  -1,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.219904166034e-05,  1.134533170374e-05, {   1,   1,  -1,   0,  -2,   2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.818228367241e-05, -5.356766358481e-06, {   3,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -2.042146587427e-05, -1.310338493345e-05, {   2,   0,  -1,   0,   0,   3,  -1,  -4,   0,   0,   0,   0,   0 } },
 { -2.808445024656e-05, -5.435828543533e-06, {   4,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.773674781369e-05, -1.576559424477e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,  -1,   0,   0,   0 } },
 { -1.927569752909e-05, -1.413548545933e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0 } },
 {  2.315841131574e-05, -1.024620728962e-05, {   2,   0,  -2,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.608303397887e-05,  7.313906031393e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   5,   0,   0,   0 } },
 {  6.408628856960e-06,  2.695722481080e-05, {   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.701134150455e-05, -6.331649851967e-06, {   1,   1,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.671871959599e-05,  6.622781366422e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 { -3.143039204241e-05, -1.885224212376e-06, {   4,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.153797842479e-05, -2.177699249235e-05, {   0,   0,   1,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 {  3.090083380364e-06,  3.020950231159e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 {  1.672620339531e-08, -3.326452032735e-05, {   1,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.309247533301e-05, -1.850517078870e-07, {   2,   0,   0,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  6.820342151935e-06, -2.645029999344e-05, {   3,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -3.134079306024e-07, -3.294901381407e-05, {   2,   0,   0,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0 } },
 {  2.691613175340e-05, -6.326796351951e-06, {   1,   1,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.004702612717e-05,  2.317772950371e-05, {   4,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  1.065291546726e-06, -3.213890945622e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -5.129413301225e-06, -2.806240182440e-05, {   2,   0,   1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.790093728713e-05, -5.285480967943e-06, {   4,   0,  -3,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  3.208129650062e-08, -3.315351437658e-05, {   3,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.274067565554e-05, -2.043981780478e-05, {   2,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 {  2.753293096146e-05, -5.646527153282e-06, {   2,   0,  -1,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.095367588445e-05, -2.223655105127e-06, {   0,   0,   1,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 { -2.229269131319e-05, -1.088335261662e-05, {   4,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.135475839171e-05,  2.180415680641e-05, {   4,   0,  -1,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.665692894943e-05,  6.499809068394e-06, {   2,   0,   0,   0,   0, -10,   7,   0,   0,   0,   0,   0,   0 } },
 { -8.528907229358e-06,  2.462449805157e-05, {   0,   0,   1,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  2.824885936140e-05, -4.895037147559e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.800182497463e-05,  1.514039992863e-05, {   1,  -1,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.316778553042e-06, -2.681206132764e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 {  1.178919745064e-05,  2.133551834390e-05, {   4,   0,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.963421527988e-05, -1.346802501251e-05, {   0,   0,   1,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 { -9.834219205854e-07, -3.211584199118e-05, {   2,   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 {  2.625967661815e-05, -6.833669917303e-06, {   1,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.632559379482e-06,  2.745943638492e-05, {   2,   0,  -2,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -2.126150567552e-05, -1.182954117636e-05, {   1,   1,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  1.961931786840e-05,  1.345638857616e-05, {   0,   0,   1,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 {  2.198469634644e-05,  1.107889871249e-05, {   2,   0,  -2,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  5.627142643253e-07,  3.249492986365e-05, {   1,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -6.455257965444e-07, -3.240857880024e-05, {   2,   0,  -2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -3.135646108293e-05, -1.695368133649e-06, {   0,   0,   1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 {  3.131353671538e-05,  1.729564742560e-06, {   2,   0,  -3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -5.488006064710e-08,  3.296720193721e-05, {   2,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.292686800364e-05,  8.593573034700e-08, {   2,   0,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.289784902830e-05,  8.434880456400e-08, {   2,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.976046938467e-08, -3.294853908572e-05, {   0,   2,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.786159326667e-05, -5.098383691373e-06, {   2,   0,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.219920228257e-07,  3.280154678668e-05, {   2,   0,  -2,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.279432778370e-05,  1.008178946948e-05, {   2,   0,  -2,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -1.925892778458e-05, -1.361691742216e-05, {   1,  -1,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  6.989370788313e-06, -2.588093317927e-05, {   1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.197194656530e-05, -2.082905631824e-05, {   2,   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -5.696942944682e-06,  2.708962867535e-05, {   2,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  7.868312548575e-06,  2.487623689543e-05, {   1,   0,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.782722347777e-07, -3.255667967871e-05, {   2,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.591821462347e-05,  6.816125448479e-06, {   1,  -1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.790081274864e-08, -3.267376680151e-05, {   2,   0,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.150656676172e-05,  2.120124058508e-05, {   0,   0,   1,   0,   0,   0,  -4,   6,  -2,   0,   0,   0,   0 } },
 { -3.370648584102e-06, -2.932616762524e-05, {   0,   0,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -2.125598551270e-05,  1.143221570143e-05, {   2,   0,  -1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  1.397648572142e-08, -3.264897182587e-05, {   6,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.977587057276e-05,  1.287977515134e-05, {   2,  -2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -4.427434131102e-08, -3.260945667886e-05, {   4,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -3.243755807881e-05, -2.152893674394e-07, {   4,   0,  -3,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.624459799951e-05,  6.407183190570e-06, {   2,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.654771997362e-05,  6.096665832434e-06, {   3,  -1,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.626618899053e-05,  6.373979663091e-06, {   1,  -1,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.577677654042e-05, -6.836848665418e-06, {   4,   0,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.829954954852e-05,  1.429989758152e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   3,   0,   0,   0 } },
 {  2.589954615695e-05, -6.692358594522e-06, {   1,  -1,   1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.275606585189e-05,  9.826672457575e-06, {   1,   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  2.464696442027e-05,  7.926922612717e-06, {   2,   0,   0,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 { -1.680522147680e-05,  1.576811550442e-05, {   2,   0,  -1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 { -2.987951950818e-05,  2.685947497644e-06, {   1,   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.987954903991e-05, -2.685618955005e-06, {   3,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.987960012596e-05, -2.685050523324e-06, {   1,  -1,  -2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.987960059459e-05,  2.685045308350e-06, {   1,  -1,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.987965026413e-05, -2.684492520636e-06, {   2,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.987967509506e-05,  2.684216126435e-06, {   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.987972474926e-05,  2.683663337343e-06, {   2,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.987972474926e-05, -2.683663337343e-06, {   1,   1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.080969778432e-05, -1.173391619889e-05, {   4,   0,  -2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  8.016534159544e-06, -2.452010999549e-05, {   4,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  3.693022573687e-06,  2.883300364484e-05, {   1,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.632364510689e-05, -6.162542277968e-06, {   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.401089582657e-05, -8.471444241596e-06, {   0,   0,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.350466765109e-05,  8.968919093592e-06, {   0,   0,   1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.351074801640e-06,  2.910409943392e-05, {   4,   0,  -1,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.613859228074e-05, -6.314139002563e-06, {   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.356629856127e-06,  2.807434289309e-05, {   2,   0,  -2,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.236543125851e-06, -2.718852048058e-05, {   2,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.333846586338e-06,  2.808496478995e-05, {   1,   1,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.629821299090e-05,  6.104825386653e-06, {   2,   0,   1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 { -1.254106783443e-05, -1.985357290511e-05, {   4,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -9.935346455494e-06, -2.245832597216e-05, {   2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.622701664337e-05,  6.158027181457e-06, {   4,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.615604411342e-05,  1.622161860353e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  1.682354673553e-05, -1.554918331275e-05, {   3,  -1,  -2,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.133236190023e-05,  1.031837821371e-06, {   0,   0,   1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.202363474971e-05, -1.031934457375e-05, {   2,   0,  -1,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.466914734410e-05,  1.767180700088e-05, {   2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.895738900351e-05, -3.326188739689e-06, {   2,   0,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.605461349378e-06,  3.066185700032e-05, {   0,   0,   0,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 { -1.232462698705e-05,  1.989390724045e-05, {   2,   0,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0 } },
 {  5.015886076587e-06, -2.715636750615e-05, {   1,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.140514518381e-05,  2.071903818008e-05, {   4,   0,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -1.140514518381e-05, -2.071903818008e-05, {   0,   0,   1,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -2.132885564123e-05, -1.079369378869e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  3.160370749295e-05,  4.807683292190e-07, {   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.586444927492e-05,  6.215943144795e-06, {   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.443539378574e-07, -3.121783172981e-05, {   2,   2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.536478552071e-06, -2.549684260398e-05, {   4,   0,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  4.607766163531e-06,  2.741196547242e-05, {   3,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.594691996448e-05, -6.042582171735e-06, {   1,  -1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.067734043127e-05, -2.130737838051e-05, {   2,   0,  -1,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.325278714505e-07,  3.164806564390e-05, {   2,   0,  -1,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.055989946227e-05,  1.417172304636e-06, {   2,   0,   1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.932991158756e-05, -1.264133312435e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.745461197296e-05,  1.451243628543e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.985304291612e-05,  1.209989055307e-05, {   0,   0,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  5.817391018873e-06,  2.613336488037e-05, {   6,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.892402056934e-05,  1.301950102068e-05, {   2,   0,  -2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.081786213981e-05,  1.124149045725e-06, {   1,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -6.664129827050e-06,  2.527597856173e-05, {   0,   0,   1,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 { -2.120788450889e-05,  1.072339006891e-05, {   2,   0,   0,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 {  1.297453562580e-05, -1.892785601162e-05, {   2,   0,  -1,   0,   0,  -4,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.848217762205e-05,  1.340736522321e-05, {   1,   0,  -1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.993676716130e-05,  1.946785024101e-06, {   2,   0,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.953354936345e-05, -2.339808198622e-06, {   2,   0,  -1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.125993483036e-05, -2.061172447914e-05, {   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.190535299325e-05, -1.996043817053e-05, {   2,  -2,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.822983807377e-08,  3.183685466595e-05, {   1,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.949075651632e-06,  2.589711659631e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 { -1.323108982127e-05,  1.861073643665e-05, {   2,   0,  -1,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 {  8.384395746791e-06, -2.345249367760e-05, {   4,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.936175507376e-05, -1.245711644325e-05, {   2,  -2,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.776909852090e-05,  4.048902637552e-06, {   4,   0,  -1,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  8.050240477073e-09, -3.180990510868e-05, {   4,  -2,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.436801029992e-05, -7.427342748634e-06, {   1,  -1,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.473330044971e-06, -3.031483690807e-05, {   4,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.198570457031e-05,  9.773931283959e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.958490923190e-08,  3.170821123684e-05, {   2,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.816934697231e-05,  3.576311891967e-06, {   3,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.370007769524e-07,  3.149770470346e-05, {   6,  -2,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.370007769524e-07, -3.149770470346e-05, {   2,  -2,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.222341661271e-06,  2.650877669784e-05, {   1,  -1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.303110615182e-05, -1.868397622235e-05, {   2,   0,  -2,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 { -2.195258899197e-05, -9.759472840389e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -2.833738468234e-06,  2.885550475721e-05, {   0,   0,   1,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 {  5.201076905609e-06, -2.648258763981e-05, {   4,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.042207155387e-05,  2.126036360355e-05, {   2,   0,  -1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  9.464914672615e-06, -2.217167734834e-05, {   2,   0,  -1,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 { -2.522804240393e-05,  6.401953927016e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.655031110313e-05, -1.507355884629e-05, {   1,   0,   0,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -2.559407682905e-05, -6.010337337867e-06, {   1,  -1,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.475246316557e-06,  2.709295000206e-05, {   2,   0,   1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  1.020347683092e-05, -2.135280821599e-05, {   2,   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -2.651411438247e-05,  5.033512719806e-06, {   3,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.921908176775e-05,  2.314153024630e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 {  1.296376574238e-05, -1.856912862528e-05, {   2,   0,  -1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 {  1.903761279560e-05, -1.249507140949e-05, {   2,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.675843184364e-05,  4.767856275649e-06, {   2,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.178587109957e-05, -9.735583552975e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,  -5,   0,   0,   0 } },
 {  3.150071236452e-05, -1.432789858000e-08, {   2,   0,   0,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.550333398630e-05, -6.010810233638e-06, {   1,  -1,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.225311630055e-05, -9.258523674518e-06, {   1,   1,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.460015364099e-06, -2.704301765827e-05, {   2,   0,   1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  1.681751393615e-05,  1.466945384795e-05, {   0,   2,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -3.652043772577e-07,  3.110317628162e-05, {   2,   0,  -1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 { -2.222588442753e-05,  9.228779602248e-06, {   3,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -2.222588442753e-05, -9.228779602248e-06, {   1,   1,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -3.150464266050e-06,  2.830142233979e-05, {   0,   0,   1,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.543950792555e-05,  6.004112159697e-06, {   1,  -1,   2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.723870216948e-05, -4.176646000520e-06, {   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.621070218477e-05, -1.520032482414e-05, {   2,   0,   0,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.493626039943e-05, -6.469058771340e-06, {   3,  -1,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.567210087519e-05, -1.573125359512e-05, {   2,   0,   0,   0,   0,   0, -12,  19,   0,  -3,   0,   0,   0 } },
 {  1.342821401972e-05,  1.794170926632e-05, {   1,  -1,   3,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  2.808775756453e-05,  3.261851850596e-06, {   2,   0,  -1,   0,   0,  -3,   4,   0,  -2,   0,   0,   0,   0 } },
 { -2.418846813619e-08, -3.130272798912e-05, {   4,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.056881696402e-05, -2.074685551755e-05, {   0,   0,   1,   0,   0,  13, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.052044347012e-05, -2.079433153544e-05, {   0,   0,   2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  3.130697560411e-05,  1.552750757000e-09, {   2,   0,   0,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -5.498055046227e-06,  2.579598010710e-05, {   1,  -1,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.700982577854e-05,  1.423898141683e-05, {   2,   0,  -2,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -2.671836803019e-07,  3.095965411773e-05, {   4,   0,  -2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.148351582664e-05, -1.973847736482e-05, {   1,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.516720535668e-05, -6.051810110610e-06, {   2,   0,   2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  2.915782738036e-05, -2.043485531898e-06, {   0,   0,   1,   0,   0,   0,   6,  -8,   3,   0,   0,   0,   0 } },
 {  2.132631450553e-05,  9.863087742398e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 { -1.050140703954e-05, -2.067098967733e-05, {   4,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -2.010703058866e-05,  1.105634615695e-05, {   2,   0,  -2,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  2.504217852647e-05,  6.107496564417e-06, {   2,   0,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.867193098542e-05, -1.247287796071e-05, {   2,   0,   0,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -2.839886396007e-05, -2.743375085582e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.545348725659e-05,  5.658321339143e-06, {   2,   0,  -3,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.475297298043e-05,  6.357261063329e-06, {   1,  -1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.064938230036e-05,  1.045364489289e-05, {   2,   0,   0,   0,   0,   0,  17, -21,   0,   0,   0,   0,   0 } },
 { -2.869513522776e-05,  2.402163260304e-06, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.602046534379e-05, -1.505808887904e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 {  3.028938902645e-05,  7.744906508402e-07, {   3,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.610163444208e-05,  4.935132465142e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 {  2.112808398556e-05, -9.877405976170e-06, {   2,   0,   0,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 {  2.449278126148e-05, -6.505831579926e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.239327313475e-05, -1.858201221312e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -3,   0,   0,   0 } },
 { -7.228332449716e-07,  3.024975860345e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 { -1.292721926475e-05,  1.804196094996e-05, {   2,   0,   0,   0,   0,   0,  15, -20,   0,   0,   0,   0,   0 } },
 { -3.825283139040e-06,  2.712872191225e-05, {   1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.655671900626e-05,  4.392980421903e-06, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 { -7.357697686587e-06, -2.358938899836e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -1.839301749721e-05, -1.252739566202e-05, {   3,   0,   0,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 { -1.839301749721e-05,  1.252739566202e-05, {   1,   0,   0,   0,   0,   0,  31, -35,   0,   0,   0,   0,   0 } },
 {  1.094591472471e-05, -1.996769265312e-05, {   2,   0,  -3,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  7.661925955758e-06, -2.324940593923e-05, {   2,  -2,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.092228810621e-05, -1.997471691162e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  1.652871851529e-05,  1.436019842988e-05, {   2,   0,   1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -2.036292566372e-05,  1.052388033800e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  2.968312028434e-05,  1.202108520709e-06, {   0,   0,   1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  6.024746688520e-06, -2.485325886810e-05, {   1,  -1,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.016904845008e-05,  2.069474814189e-05, {   2,   0,  -1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 {  9.632172977133e-06,  2.122355284561e-05, {   2,   0,   0,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 {  2.419628678913e-05,  6.646711520678e-06, {   1,   1,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.082307156661e-05,  2.000889943479e-05, {   2,   0,  -2,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.092381682024e-05,  9.906020143510e-06, {   0,   0,   1,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 { -2.243449161299e-08,  3.079126638267e-05, {   1,  -2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.040359511575e-05, -1.039810520870e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -2.929020611558e-06, -2.785905415910e-05, {   1,  -1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.592059942337e-07,  3.041026258636e-05, {   4,   0,  -3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.101457287479e-07, -3.017706327835e-05, {   2,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -1.892929071454e-05, -1.174458813869e-05, {   2,   0,  -2,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -2.973731437612e-05,  8.999422235276e-07, {   2,  -2,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.567589754817e-05, -1.495078713734e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -5,   0,   0,   0 } },
 { -1.863819323469e-05, -1.194676880022e-05, {   1,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.244930698995e-05, -1.813401920371e-05, {   2,   0,  -1,   0,   0,   0,  16, -21,   0,   0,   0,   0,   0 } },
 { -1.245588006030e-05, -1.812379620353e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 { -2.353636672502e-05, -7.030732397958e-06, {   6,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.778988234033e-06,  2.672423381338e-05, {   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.167334608199e-06, -2.233359071465e-05, {   4,   0,  -2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -5.825738305475e-06, -2.467301410745e-05, {   3,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.754658293473e-08,  3.044302097152e-05, {   2,   2,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  8.796903401716e-08, -3.039079891158e-05, {   0,   0,   2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.470689379096e-05, -5.756867419964e-06, {   1,  -1,   0,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.610211032341e-05,  1.434876228866e-05, {   0,   0,   1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.932375361572e-05,  1.111597531841e-05, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 {  1.003329328111e-05,  2.040205335894e-05, {   2,   0,   1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.858797491725e-05,  1.841112981760e-06, {   2,   0,  -1,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 {  4.331409881940e-07, -2.998569882205e-05, {   2,   0,  -2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.851385761979e-05,  1.904813629863e-06, {   2,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0,   0 } },
 { -2.458899526939e-05, -5.826060989859e-06, {   0,   0,   3,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.289384300149e-06,  2.812369749642e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 } },
 { -8.026088272395e-06, -2.236809634457e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  6.451940893767e-06, -2.392839080465e-05, {   2,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.932432036425e-05,  1.103901974479e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -2.913923545913e-05, -1.222460046342e-06, {   0,   0,   2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.962944800375e-05,  1.066594730597e-05, {   0,   2,  -3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.470705355981e-05,  5.582525398051e-06, {   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.057382083141e-06, -2.122482148154e-05, {   0,   0,   1,   0,   0,   0, -12,  19,   0,   0,   0,   0,   0 } },
 {  5.643894353471e-06,  2.463376906424e-05, {   1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.412529195749e-06,  2.385794625987e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  2.944348239386e-05,  7.983695923168e-07, {   2,  -2,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.763419083724e-05, -2.599655032676e-06, {   0,   0,   1,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 {  9.969525536686e-06,  2.025167132800e-05, {   2,   0,   1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.011445462563e-05, -1.007006692768e-05, {   1,   1,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.525285608906e-08, -3.013718590911e-05, {   4,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.809513494518e-05, -2.062312535460e-06, {   2,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.648041155230e-05,  1.367187885038e-05, {   0,   0,   0,   0,   0,  13,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.525861947396e-06,  2.861858434184e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 { -1.414881959702e-05, -1.598358906119e-05, {   0,   0,   0,   0,   0,  12,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.594675032074e-06,  2.651831211008e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  2.919053217025e-05,  9.074840687693e-07, {   2,   0,  -1,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0 } },
 {  4.881493917927e-06,  2.519847885858e-05, {   1,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.379717177548e-06, -2.569116467789e-05, {   3,  -1,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.631788082763e-09, -3.005995584385e-05, {   0,   0,   4,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.178244071512e-05,  8.244502893628e-06, {   2,   0,   0,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  2.453372667848e-05,  5.480514676169e-06, {   3,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.504413243450e-05, -4.969252836167e-06, {   1,   1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.887762200276e-05,  1.135340249407e-06, {   0,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  9.344639347004e-08, -2.990778825413e-05, {   3,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -9.424777959234e-10,  2.999999998520e-05, {   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  6.911503837322e-10,  2.999999999204e-05, {   0,   1,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.492624816889e-10, -2.999999999297e-05, {   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.335545184270e-10,  2.999999999331e-05, {   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  6.283185306706e-10,  2.999999999342e-05, {   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -6.230825429215e-10, -2.999999999353e-05, {   2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -5.759586531195e-10,  2.999999999447e-05, {   1,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  5.707226653703e-10,  2.999999999457e-05, {   2,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.345869837354e-10, -2.999999999685e-05, {   6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -4.241150082212e-10, -2.999999999700e-05, {   2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.136430327069e-10, -2.999999999715e-05, {   4,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.084070449572e-10,  2.999999999722e-05, {   4,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -3.874630939359e-10, -2.999999999750e-05, {   2,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -3.769911184215e-10, -2.999999999763e-05, {   2,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.665191429070e-10, -2.999999999776e-05, {   6,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.141592653637e-10, -2.999999999836e-05, {   2,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.460914245321e-10, -2.999999999899e-05, {   2,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.188790206115e-11, -2.999999999997e-05, {   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   2,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.000000000000e-05,  0.000000000000e+00, {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.585263530245e-05,  1.413490582740e-05, {   1,   0,  -1,   0,   0,   0,  33, -41,   2,   0,   0,   0,   0 } },
 {  2.718310201902e-05,  2.802406366865e-06, {   2,   0,   1,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  2.718310201902e-05, -2.802406366865e-06, {   2,   0,  -1,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 {  1.887886958145e-05, -1.108135573909e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -3.794333102419e-06, -2.615619624941e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -4.457479512070e-06, -2.548966835036e-05, {   2,   0,  -1,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 {  7.042422774012e-06,  2.289602901244e-05, {   0,   0,   2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  1.211609308507e-05,  1.780939893931e-05, {   1,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.188512604946e-05,  1.800861968216e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   2,   0,   0,   0 } },
 { -5.962309216738e-06, -2.393006516637e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   6,  -3,   0,   0 } },
 {  3.721039268742e-06,  2.616981058714e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  2.428493049755e-05,  5.602361043582e-06, {   1,  -1,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.512811829409e-05, -1.474187667740e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -2.573451413408e-05,  4.134438641155e-06, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.535183631044e-05,  4.507874781912e-06, {   2,   0,   0,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 {  1.281747858863e-05,  1.702461748889e-05, {   0,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.544098978300e-05, -1.439714447077e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  2.450625546956e-05,  5.324191814202e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 { -3.457790169400e-08, -2.978445116377e-05, {   2,   0,  -3,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.150542383769e-05, -8.313429927108e-06, {   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.909327282612e-05,  7.185710880435e-07, {   4,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.737612126994e-05, -2.427255640368e-06, {   4,  -2,  -2,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -8.214746015791e-07, -2.895129510885e-05, {   4,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.091367796184e-05, -8.856765063166e-06, {   2,   0,   1,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 {  2.091367796184e-05,  8.856765063166e-06, {   2,   0,  -1,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.695041386453e-06,  2.706637492405e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -3.363077476482e-07, -2.940592801367e-05, {   2,   0,  -3,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  9.788618178575e-06, -1.993118635166e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 {  2.410001292419e-05, -5.607804927055e-06, {   1,  -1,   0,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.889888812259e-06, -2.381526317620e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 {  1.088258485052e-06,  2.859503304118e-05, {   2,   0,  -1,   0,   0,   6,  -8,   0,   2,   0,   0,   0,   0 } },
 {  1.889908939047e-05,  1.074607361146e-05, {   4,   0,  -1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  2.770743727624e-05, -1.936405048722e-06, {   3,  -1,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -2.711441232043e-05,  2.526335108030e-06, {   0,   0,   1,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 {  2.312540286122e-05,  6.508902525127e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   3,   0,   0,   0 } },
 {  2.395129161405e-05,  5.672144392891e-06, {   3,  -1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.886954251620e-05,  1.074762104224e-05, {   2,   0,  -1,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0 } },
 {  2.407470986191e-05, -5.542391747484e-06, {   3,   1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.579000248197e-05, -1.382006086490e-05, {   0,   2,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.397803348098e-07, -2.926623047374e-05, {   4,   0,   0,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.484792778152e-08,  2.958958459701e-05, {   0,   0,   2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.755110985812e-06,  2.284207355874e-05, {   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -4.546957632991e-07, -2.913731831324e-05, {   2,   0,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.702571071368e-06, -2.688069376969e-05, {   2,   0,  -1,   0,   0,   0,  14, -20,   0,   0,   0,   0,   0 } },
 { -2.397190922177e-05,  5.610926578163e-06, {   3,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.490834493504e-05,  4.657478452359e-06, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 { -2.278677208330e-05,  6.763244326935e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 { -4.546095310003e-06,  2.497886657443e-05, {   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.600148819070e-06, -2.591437127338e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.319060969903e-05, -6.323649022857e-06, {   1,   1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.142779397269e-05,  8.068909422865e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  1.682454394343e-05,  1.267180101249e-05, {   1,  -1,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.895460849378e-05,  5.191621469231e-07, {   1,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.536281046951e-05, -4.101177078510e-06, {   0,   0,   3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.396689674323e-06, -2.605122032565e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.894952211510e-05, -1.047283021579e-05, {   4,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.653018615939e-05, -1.288045098476e-05, {   0,   0,   1,   0,   0,   0, -13,  21,   0,   0,   0,   0,   0 } },
 { -1.730027182396e-05, -1.210822634433e-05, {   1,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.891129081610e-06,  2.450111624908e-05, {   1,  -1,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.842323826609e-05,  1.096211186702e-05, {   2,   0,  -1,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 { -4.822855123607e-08,  2.932363152851e-05, {   0,   0,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -5.240919371463e-06, -2.413023834264e-05, {   3,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.558552169091e-06, -2.580878004246e-05, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  2.480402553863e-05,  4.530358387551e-06, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   0,   0,   0,   0 } },
 {  3.484946176680e-06,  2.584722539922e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  3.024322582196e-06, -2.630174458867e-05, {   1,  -1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.191987327638e-06, -2.010838913976e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 { -3.533854971721e-06,  2.576053031827e-05, {   1,  -1,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.814205728541e-05, -1.138803957536e-06, {   1,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.228411512651e-05, -6.981173989752e-06, {   2,   0,  -2,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.736114607753e-06, -2.351412867251e-05, {   1,  -1,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.912748535990e-05,  1.096070770788e-07, {   2,   0,   0,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 { -3.511342451898e-08, -2.918325339022e-05, {   2,   0,   2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.545909527300e-06,  2.365452424059e-05, {   2,   0,  -2,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  8.399541085958e-06, -2.079316546248e-05, {   1,  -1,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  9.148280773650e-06, -2.004101315563e-05, {   4,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.829180300787e-05, -8.950322090547e-07, {   3,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.815930763816e-06, -2.435872183102e-05, {   1,   1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.756089801065e-05,  1.608641165138e-06, {   0,   0,   1,   0,   0,   6,  -6,  -8,   6,   0,   0,  -1,   0 } },
 {  2.756089801065e-05, -1.608641165138e-06, {   0,   0,   1,   0,   0,  -6,   6,   8,  -6,   0,   0,   1,   0 } },
 { -1.099007559454e-05, -1.817274508905e-05, {   4,   0,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -2.267774071601e-05, -6.484941255966e-06, {   2,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.129555976964e-05,  1.786314208966e-05, {   2,   0,   1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.373780925103e-06,  2.777693435319e-05, {   4,  -2,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.390750464039e-06, -2.175689877929e-05, {   4,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.627123629607e-05,  2.875645891370e-06, {   4,  -2,  -1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.817943829209e-05, -1.092719033851e-05, {   3,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.219261669429e-05, -1.691187922401e-05, {   0,   0,   2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.101339023461e-05, -1.808935177602e-05, {   2,   0,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0 } },
 { -1.969664839180e-09, -2.908725445375e-05, {   2,   0,   1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.374415689885e-05, -5.341412518881e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 {  5.183736813442e-07,  2.856194311003e-05, {   2,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 { -5.502622050650e-06, -2.356964598087e-05, {   2,   0,  -2,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  4.839224076066e-06, -2.422521713099e-05, {   1,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.023471651902e-05,  1.882372688363e-05, {   2,   0,  -3,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -9.990014262936e-08,  2.895611628414e-05, {   0,   0,   2,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -9.990014262936e-08, -2.895611628414e-05, {   0,   0,   2,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 {  1.591409663111e-05, -1.313805848525e-05, {   1,   1,   0,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.368747722830e-05, -1.536026526298e-05, {   0,   0,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  3.953545932670e-06, -2.509179510168e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -3.384725212556e-06, -2.564636465926e-05, {   0,   0,   1,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.189683660060e-05,  7.125783394780e-06, {   1,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.240231691944e-05, -6.611596993729e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 { -2.817966526128e-05,  8.326800691310e-07, {   1,  -1,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.817966526128e-05, -8.326800691310e-07, {   1,  -1,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  2.698276767347e-05,  2.029074816790e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   1,   5,   0,   0,   0 } },
 { -2.591318806253e-05, -3.075439681821e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -2.778794624207e-05, -1.195142976694e-06, {   4,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  2.456461515674e-06, -2.652479258445e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 {  1.636320627613e-05,  1.261752189618e-05, {   2,   0,  -3,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.869095653583e-05, -1.027033265041e-05, {   2,   0,   0,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -2.072959915162e-05, -8.208654821211e-06, {   2,   0,  -1,   0,   0,   3,  -6,   0,   2,   0,   0,   0,   0 } },
 { -1.790548252048e-05, -1.099968123345e-05, {   2,   0,  -1,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 { -1.185685658723e-05,  1.703843888245e-05, {   2,   0,  -2,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  9.811242548023e-06, -1.908366919784e-05, {   0,   0,   2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.272550191168e-07, -2.826724848406e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  4.483027352993e-06,  2.440231754818e-05, {   2,  -2,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.616252437178e-07, -2.831841495280e-05, {   3,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.355810822765e-06, -2.552236414795e-05, {   2,   0,   0,   0,   0,   3,  -9,   4,   0,   0,   0,   0,   0 } },
 { -2.850646156391e-05,  3.686962227735e-07, {   4,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.757772220123e-05, -1.128545604619e-05, {   3,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.606005911144e-05,  1.278821219500e-05, {   1,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  4.156945088501e-06, -2.468723596216e-05, {   0,   0,   2,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  4.160639394985e-06,  2.468235104382e-05, {   0,   0,   2,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -3.915928793651e-06,  2.492270807661e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -2.072789768673e-05,  8.092448143913e-06, {   1,  -1,  -1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.326758357614e-05,  5.550366141765e-06, {   2,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.693351061403e-06,  2.712043994683e-05, {   2,   0,  -1,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 {  1.693351061403e-06, -2.712043994683e-05, {   2,   0,  -1,   0,   0,   0, -14,  27,  -3,  -3,   0,   0,   0 } },
 {  1.934150344809e-05,  9.463303408628e-06, {   2,   0,  -1,   0,   0,  -6,   9,   0,  -3,   0,   0,   1,   0 } },
 {  2.329984091622e-05,  5.490702677866e-06, {   1,  -1,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.327772950807e-05,  5.510234099548e-06, {   1,   1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.901902186603e-06,  2.088088771635e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -6,   3,   0,   0 } },
 { -6.567099363318e-06, -2.220472808045e-05, {   1,  -3,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.449146182054e-08,  2.871507877691e-05, {   2,  -2,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -8.966415824430e-06, -1.978024287159e-05, {   2,   0,  -2,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 { -1.747609415459e-05,  1.126908987071e-05, {   3,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.414856287939e-05, -4.591052733631e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,   1,   0,   0,   0 } },
 {  2.320734139825e-05,  5.521365736928e-06, {   3,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.332191768863e-05, -5.366572061853e-06, {   1,   1,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.644196496702e-05, -2.244184450676e-06, {   2,   0,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.106084385656e-06, -2.257124068484e-05, {   0,   0,   2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.269132228733e-06, -2.737244067438e-05, {   0,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -2.526468980706e-05,  3.362667074868e-06, {   1,  -1,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.013938358493e-05, -1.847901290232e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   5,   0,   0,   0 } },
 { -1.324852422942e-05, -1.536521236856e-05, {   2,   0,  -2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -2.780882041321e-05, -7.874147159406e-07, {   0,   2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.028473016465e-05, -8.305392390478e-06, {   1,  -1,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.549892697957e-05, -1.305204265538e-05, {   4,   0,  -1,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -1.549892697957e-05,  1.305204265538e-05, {   0,   0,   1,   0,   0,   0,  33, -40,  -3,   0,   0,   0,   0 } },
 {  2.078667352314e-05,  7.761698023658e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -2.068782839404e-05, -7.851550996961e-06, {   2,   1,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -2.068782839404e-05,  7.851550996961e-06, {   2,  -1,  -1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.593561287492e-07, -2.817846383128e-05, {   1,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.066725564835e-08, -2.851649794038e-05, {   0,   2,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.011247758812e-06,  2.450912211446e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -2.361538138447e-05, -4.888198004455e-06, {   2,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -3.728977718753e-06, -2.476715458171e-05, {   4,   0,  -2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.205569529684e-05, -6.432728641423e-06, {   0,   2,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.994500378073e-06,  2.348813064187e-05, {   1,  -1,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.704143598107e-05, -1.425648241809e-06, {   2,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  2.072768462519e-05, -7.673757916352e-06, {   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.564788039917e-05,  1.275180022228e-05, {   0,   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.496553880510e-06, -2.389678324918e-05, {   1,   1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.347947244853e-06,  2.703743713627e-05, {   2,   0,  -1,   0,   0,  10,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.554745914442e-09, -2.834898125076e-05, {   3,   0,  -2,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.274868477109e-05,  5.604358665971e-06, {   1,  -1,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.092049906075e-05, -7.418450554499e-06, {   2,   0,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.689947991582e-05, -1.429823034379e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  9.348023410252e-06,  1.897289588379e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -4,   0,   0,   0 } },
 {  2.290929033250e-05,  5.411098345030e-06, {   0,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.283281607529e-05,  5.484795781213e-06, {   1,   1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.639333395325e-07, -2.799800844911e-05, {   2,   0,  -5,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  6.346214003651e-07,  2.762114966150e-05, {   0,   0,   4,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.983191154534e-06,  2.426326214740e-05, {   1,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.175292562659e-05, -6.493440513998e-06, {   0,   0,   2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.713218183300e-05,  1.113498217546e-06, {   2,   0,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.217217172384e-07, -2.812132032650e-05, {   4,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.008990090165e-05,  1.814606144478e-05, {   0,   0,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -2.819318386587e-05, -2.957400177424e-08, {   1,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.878129450286e-06,  2.532097798762e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 { -1.971034378683e-05, -8.481045274751e-06, {   0,   0,   1,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.157118265407e-06,  2.603379822313e-05, {   1,  -2,   1,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 {  2.157118265407e-06, -2.603379822313e-05, {   1,  -2,  -1,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 { -6.483240815292e-06, -2.170279770047e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -1.713588143001e-05, -1.104123609816e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 {  9.378828459385e-06,  1.876898446263e-05, {   4,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.262748631566e-05,  1.550956444739e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.410476256602e-05,  4.009571427358e-06, {   2,  -2,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.033763673969e-05, -1.777285933054e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0 } },
 {  1.869200898928e-05, -9.398006245210e-06, {   0,   0,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.334605836938e-06, -2.270219682006e-05, {   1,   1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -3.625417053621e-07, -2.765661295606e-05, {   0,   2,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.881798674968e-08, -2.798384035444e-05, {   2,   0,  -2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.616366259017e-05, -1.182192225289e-05, {   1,  -1,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -4.870763504275e-08,  2.792400737912e-05, {   1,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -8.922987440577e-07,  2.708005606963e-05, {   0,   0,   2,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 {  3.440801863447e-06,  2.450991835306e-05, {   0,   0,   3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.580819718184e-05,  1.213406660322e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 { -6.059557722212e-06,  2.187399451282e-05, {   2,  -2,   0,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.188099135709e-08, -2.791061484640e-05, {   3,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  7.429222593213e-06, -2.049927449462e-05, {   2,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  2.161582493035e-05,  6.292638725194e-06, {   0,   0,   3,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.158230535321e-05,  6.316463717951e-06, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.383337517855e-06,  2.649669975206e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 { -2.265371537814e-05, -5.221325219008e-06, {   1,  -1,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.538229806287e-05, -1.247550464681e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0 } },
 { -2.364914805635e-05, -4.206378439036e-06, {   2,   0,  -2,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 { -9.225415349334e-06, -1.862567552096e-05, {   1,   0,   1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.112353333900e-05,  6.725242499191e-06, {   4,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.544446975872e-05,  1.239432707176e-05, {   1,  -1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -5.259895282961e-07,  2.729769076130e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.455331629552e-05,  1.326976697460e-05, {   2,   0,  -2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  9.969987040754e-06, -1.784192341058e-05, {   4,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -4.244173066261e-06, -2.356264178147e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   5,   0,   0,   0 } },
 {  2.245300132070e-05, -5.344132811769e-06, {   1,  -1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.774908436032e-05, -1.002880689179e-05, {   2,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -5.161377179309e-06, -2.261468459242e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  4.164158283156e-06, -2.361087397901e-05, {   4,  -2,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.458501028242e-05,  1.318921461993e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0 } },
 {  2.170455097798e-05, -6.068967441311e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   2,   0,   0,   0 } },
 { -1.367146159878e-06, -2.640561489180e-05, {   2,   0,  -1,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 {  8.014292985911e-07,  2.696736435103e-05, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -9.157930907165e-06, -1.860419029114e-05, {   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.280118203872e-05,  4.957585301299e-06, {   2,   0,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -2.699662173819e-05,  7.517960164230e-07, {   4,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  1.065459110128e-05, -1.708041945170e-05, {   2,   0,   2,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.394954537691e-05,  3.772275475839e-06, {   4,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.606283116987e-07, -2.755680974417e-05, {   2,   0,   1,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -6.017071243960e-06,  2.167050376311e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  1.776756726750e-05, -9.918010951771e-06, {   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.320162703593e-05, -4.458692548693e-06, {   5,  -1,  -2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.068649911410e-05, -6.969448664412e-06, {   0,   0,   2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.145895267950e-05,  1.618084192011e-05, {   2,   0,  -1,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 } },
 { -1.522031277643e-05,  1.241250472007e-05, {   1,   1,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.183306241937e-05, -1.578621874765e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -9.936536157626e-06,  1.767220685648e-05, {   1,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  9.144237096350e-06,  1.846311229494e-05, {   1,   0,  -1,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.072107429353e-05, -1.688367750203e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.135360258712e-05,  6.245963354195e-06, {   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.229085191931e-05, -5.304577855701e-06, {   0,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.628434220675e-06,  2.396514876307e-05, {   0,   0,   1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -1.582002725017e-09, -2.758717431275e-05, {   1,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.105126209720e-05,  1.653543763155e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  1.059662386825e-05,  1.698656918816e-05, {   2,   0,   2,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.465382583101e-05, -1.291607151441e-05, {   0,   0,   1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.195170464420e-05, -5.585664954959e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 {  1.097883983876e-06, -2.642330438949e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  2.062210875056e-05,  6.884664781207e-06, {   1,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.689504959225e-05, -6.075214957433e-07, {   3,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.420593547109e-06,  2.406322637885e-05, {   4,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  5.728225105009e-06,  2.174714476048e-05, {   2,   0,   1,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  7.775857129721e-06,  1.967512360121e-05, {   2,   0,  -1,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 {  1.882544741935e-05, -8.616636157634e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -5.978918467183e-06, -2.145294565856e-05, {   2,   0,  -1,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 { -2.579200349584e-05, -1.636413490800e-06, {   0,   0,   0,   0,   0,  14, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.269468917434e-07, -2.648516070237e-05, {   6,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.705117029442e-05,  1.034688796245e-05, {   2,   0,   0,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 { -2.163291968136e-05, -5.756520157683e-06, {   0,   2,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.453062049029e-05, -2.856405457771e-06, {   2,   0,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -6.873052579206e-09, -2.736447547622e-05, {   2,  -2,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.576291842585e-05,  1.604471502659e-06, {   2,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  3.637360685557e-06, -2.372388154112e-05, {   2,   0,   0,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 { -1.783492952277e-08,  2.734060529792e-05, {   4,  -2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.963330155389e-05, -7.709993563816e-06, {   2,   0,  -2,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  4.527009319404e-06,  2.276756330612e-05, {   1,  -1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.178839962663e-05, -5.495204074967e-06, {   0,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -8.968984383431e-06,  1.831436427231e-05, {   1,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.598905258945e-05, -1.129384427384e-05, {   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  7.155518303801e-06, -2.012238953394e-05, {   2,   0,   0,   0,   0,   0,  -6,   0,   6,   0,   0,   0,   0 } },
 { -2.641163231371e-05,  8.429319970894e-07, {   6,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.907064461736e-05,  8.180313487438e-06, {   4,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.108226961431e-07, -2.703876659859e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.415839315162e-05, -1.308729964717e-05, {   2,   0,  -2,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  6.212449302881e-06, -2.101602577826e-05, {   1,   0,  -1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -4.109289239451e-06,  2.310397324069e-05, {   1,  -1,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.621120289528e-07,  2.704154344273e-05, {   2,  -2,   0,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.644598791919e-09, -2.719252984791e-05, {   0,   0,   2,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.219531887903e-05, -1.499494573002e-05, {   2,   0,  -1,   0,   1,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.698831727114e-05,  2.013772112465e-07, {   1,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.035856584000e-07, -2.635333765780e-05, {   2,  -2,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.285193690003e-05,  1.429738053048e-05, {   2,   0,  -1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 {  3.919982018337e-08, -2.710098829329e-05, {   0,   2,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -8.056522470877e-07,  2.633402467315e-05, {   2,  -2,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  5.138749369042e-06, -2.200044525960e-05, {   1,  -1,   0,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  2.024775846890e-05, -6.891153638949e-06, {   2,   0,  -1,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.678524226231e-05, -3.449402960337e-07, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -5,   0,   0,   0 } },
 { -1.571764968651e-05, -1.140382596078e-05, {   0,   0,   3,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.796433617990e-05, -9.154130949474e-06, {   2,   0,  -1,   0,   0,   0,  17, -21,   0,   0,   0,   0,   0 } },
 {  1.917188383001e-05,  7.922721591184e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  2.640701899101e-05, -6.821342500610e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.104430990146e-05,  1.604304786694e-05, {   2,   0,   0,   0,   0,   0,  16, -21,   0,   0,   0,   0,   0 } },
 {  2.578194127452e-05,  1.299012888925e-06, {   0,   2,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.682496602574e-05, -2.527287000450e-07, {   2,  -2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.622704156353e-05,  8.348646787717e-07, {   0,   0,   1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.648451571371e-05,  1.057729373808e-05, {   4,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.863197503015e-07,  2.635949994053e-05, {   2,   0,   0,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 {  2.512548860398e-07,  2.678974502585e-05, {   2,   0,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -2.340716186994e-05, -3.631606628672e-06, {   0,   0,   0,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -2.109523135029e-08,  2.700574691830e-05, {   4,   0,  -2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.278324757658e-05, -4.228432249585e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  5.421984306435e-06, -2.158404199069e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 {  3.740221740727e-06, -2.325378559899e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 {  2.178883006434e-05,  5.188935516222e-06, {   1,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.454432270474e-06, -2.351931681842e-05, {   1,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -5.494745322410e-08,  2.691695046066e-05, {   2,   0,  -3,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.341259042773e-06,  2.262613554504e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  8.981403308473e-06,  1.798304650916e-05, {   6,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.976648133096e-06, -1.797299915281e-05, {   2,  -2,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.542751903603e-06,  1.940213417795e-05, {   1,  -1,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -4.901263019475e-06,  2.204011156213e-05, {   2,   0,   0,   0,   0, -19,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.186044899392e-09, -2.693473129639e-05, {   6,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.811612788704e-05, -8.820273808483e-06, {   4,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.783312647871e-06, -1.712883004239e-05, {   1,   0,  -1,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 { -1.492191176876e-05, -1.198037261415e-05, {   1,  -1,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  1.571954698522e-06, -2.532293586359e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 {  2.045716077268e-05, -6.436173424390e-06, {   1,  -1,   0,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 {  4.290224740293e-06,  2.258173134538e-05, {   1,  -1,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.007331727232e-06,  1.883308005305e-05, {   2,   0,  -2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.007331727232e-06, -1.883308005305e-05, {   2,   0,  -4,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.487887992111e-06,  2.134617203096e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -9,   0,   0,   0 } },
 { -1.207814144158e-05,  1.475515146158e-05, {   3,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  4.069616324018e-06,  2.275494774356e-05, {   1,   1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.546496012256e-05, -1.134398300798e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  1.025613556586e-05,  1.654999629474e-05, {   0,   0,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  2.553138754819e-05, -1.262564960974e-06, {   2,   0,  -1,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 {  1.105634582569e-05,  1.573134925964e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,  -1,   0,   0,   0,   0 } },
 {  8.827461227188e-06,  1.795563998338e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 {  1.497388140080e-05,  1.180227539283e-05, {   0,   0,   0,   0,   0,  13,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.658570162106e-05,  1.015552454649e-05, {   2,   0,   2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  7.372625489466e-06, -1.933723377636e-05, {   1,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.112788790969e-05,  1.557124444188e-05, {   1,  -1,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.821552686694e-08,  2.666400122670e-05, {   0,   0,   4,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.057596998036e-06, -2.262134670962e-05, {   4,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.550332725229e-06, -2.110887754146e-05, {   2,   0,  -3,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -2.053314484165e-05, -6.122099138457e-06, {   4,  -2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.182126760720e-05,  4.831617040181e-06, {   2,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -3.757145499535e-06, -2.289006273075e-05, {   2,   0,  -2,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -4.471578080740e-08,  2.660212881945e-05, {   0,   0,   3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.050891523029e-05,  6.112586344873e-06, {   4,  -2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.354477921957e-06, -2.526129088512e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 { -3.724616709055e-06,  2.288815989997e-05, {   2,   0,  -2,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  2.155029464807e-05,  5.050730129335e-06, {   1,   1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.590652558137e-06, -2.000604148854e-05, {   1,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.203691834129e-05,  4.557229681138e-06, {   2,   0,   1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.093939705050e-05,  1.564483109314e-05, {   1,  -1,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.093939705050e-05, -1.564483109314e-05, {   1,  -1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  1.773884861536e-05, -8.840434271716e-06, {   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.621787783041e-05, -1.034719502361e-05, {   3,   0,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.699546736763e-05, -9.543254220813e-06, {   2,   0,   0,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0 } },
 { -1.328762426288e-06, -2.520815681499e-05, {   2,   0,   0,   0,   0,   0,   7,   0,  -7,   0,   0,   0,   0 } },
 {  2.070231104452e-05,  5.821486530813e-06, {   1,   1,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.427496697114e-06, -2.209221302927e-05, {   1,   1,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.141633471210e-05, -5.088542919541e-06, {   1,   1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.162714712414e-06, -2.033355923285e-05, {   1,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.491865680990e-05, -1.550415779838e-06, {   0,   0,   1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -8.438364334001e-06,  1.801509801426e-05, {   2,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 {  8.803931683875e-06, -1.763219809822e-05, {   2,   0,  -1,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.014296466430e-05, -1.628753283360e-05, {   2,   0,  -1,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0 } },
 {  1.015788556844e-05,  1.627098607572e-05, {   4,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.184610380138e-06,  1.822487479917e-05, {   0,   0,   2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.583254087357e-06, -2.481475972618e-05, {   4,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  8.652493755145e-06,  1.773658238458e-05, {   2,   0,   1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.211254116465e-06,  2.217569038419e-05, {   1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.882040359580e-05,  7.561973087245e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 { -7.911375559083e-06, -1.846079610844e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  2.132573689035e-05,  5.045298731296e-06, {   3,  -1,  -1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.762353915422e-05, -8.744483997948e-06, {   0,   0,   1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 {  2.102749936841e-05,  5.327677584470e-06, {   1,   1,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.428546737464e-05, -2.054601949084e-06, {   4,   0,  -2,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 {  2.428546737464e-05,  2.054601949084e-06, {   0,   0,   0,   0,   0,  22, -28,   7,   0,   0,   0,   0,   0 } },
 {  1.586639158398e-05,  1.046439109715e-05, {   0,   0,   2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.948749903288e-06, -2.438119333724e-05, {   1,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.929739207739e-05,  7.031240614025e-06, {   0,   0,   1,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.551091647664e-07, -2.616321868684e-05, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.783344613463e-05, -8.481169080207e-06, {   0,   2,  -1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -1.390984653912e-05,  1.240256607524e-05, {   1,   0,  -1,   0,   0,   0, -33,  41,  -2,   0,   0,   0,   0 } },
 {  1.593539061056e-05, -1.036519224065e-05, {   2,  -2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.134911444966e-05, -4.930973621695e-06, {   1,  -1,  -1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.096273894812e-05,  5.309576320411e-06, {   1,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.157953945994e-05,  1.467551185116e-05, {   0,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.578275909833e-05,  4.714260381495e-07, {   2,   0,  -1,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 {  1.146022634236e-06,  2.510644536668e-05, {   6,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -2.043195387784e-05,  5.799730124393e-06, {   3,   0,  -1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 { -1.527370443448e-05,  1.095544083392e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  4.493579133717e-07,  2.577253365151e-05, {   0,   0,   1,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 { -1.039616561676e-05, -1.581698432142e-05, {   2,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.685386633672e-06,  1.752268831938e-05, {   0,   0,   0,   0,   0,  22, -20,   0,   0,   0,   0,   0,   0 } },
 { -2.242399677034e-05,  3.778247263864e-06, {   2,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  8.639139587058e-06, -1.755696353869e-05, {   2,   0,  -1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.121617038313e-05, -4.978765615810e-06, {   1,   1,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.677026683620e-06, -2.251500769765e-05, {   0,   0,   1,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 {  4.807956658989e-06,  2.135852211515e-05, {   6,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.715146404995e-05,  9.012932467882e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 { -1.040078740726e-05,  1.574948892272e-05, {   2,   0,   0,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 { -2.063756090966e-05, -5.499846495212e-06, {   0,   0,   1,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -1.534804254926e-05, -1.077788860689e-05, {   0,   0,   2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  2.293258874582e-05, -3.185151761821e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 {  9.553454498864e-06,  1.656411236593e-05, {   2,   0,   0,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 {  4.186131939412e-06,  2.193137132506e-05, {   1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.668644128810e-06,  2.244431827542e-05, {   1,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.972505531392e-05, -6.372376603593e-06, {   0,   0,   0,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.828358913218e-05,  7.791004100604e-06, {   4,   0,  -3,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.717101708235e-06, -2.135460028544e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 {  3.640756336371e-08, -2.603479567974e-05, {   2,   0,   1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.539810809544e-06,  2.152179978935e-05, {   1,  -1,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.663649364176e-06,  2.136681370291e-05, {   1,   1,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.769800826022e-05, -8.321316156062e-06, {   4,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -2.056755195258e-05, -5.450173962831e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -6.263828745692e-06, -1.974552362193e-05, {   0,   0,   1,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.112593072450e-05, -4.878394128822e-06, {   2,   0,   2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.858945552373e-05, -7.398410678486e-06, {   1,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.572239387446e-05, -1.026203502169e-05, {   2,   0,   0,   0,   0, -10,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.301557398633e-06, -1.665764412962e-05, {   2,  -2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  3.571480455680e-06,  2.237001843891e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 { -2.521534930935e-05, -6.999616575456e-07, {   2,   0,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0 } },
 { -1.550993999353e-05,  1.039959650084e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   3,   0,   0,   0 } },
 {  9.248726429254e-06,  1.665517812596e-05, {   4,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.437033066592e-05, -1.153298024441e-05, {   2,   0,   0,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -7.740237243578e-06, -1.815315565226e-05, {   1,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.480064066572e-06,  2.139028399599e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 { -1.556336588869e-07,  2.570506719608e-05, {   2,   0,   2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.503123468657e-08,  2.576658322041e-05, {   2,   0,  -1,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 { -2.093093415076e-05, -4.908587378968e-06, {   1,  -1,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.559239697145e-05,  2.400157908641e-07, {   2,   0,  -2,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.434367100602e-05, -1.482024273325e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 { -4.959136321275e-06,  2.086015526424e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 {  2.513861078775e-06,  2.327515152455e-05, {   1,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.622553502715e-06, -1.616048216635e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 { -2.420368571947e-05, -1.568239973890e-06, {   1,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.074558077279e-05,  1.501368216491e-05, {   1,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  2.266053570523e-05, -3.096587840860e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 { -2.174667772800e-07, -2.553576921038e-05, {   2,  -2,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.103838895111e-08, -2.573975295146e-05, {   1,   0,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  7.021051886155e-06, -1.872711484484e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.433683482554e-05, -1.384475776459e-06, {   2,   0,  -2,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 {  4.309663041252e-06, -2.140720580870e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   2,   0,   0,   0 } },
 { -2.078965207302e-05, -4.925653422443e-06, {   1,  -1,   2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.080716523245e-05, -4.903373620053e-06, {   1,   1,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.033015050976e-05, -1.537770618326e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   3,   0,   0,   0 } },
 { -2.436122150948e-05,  1.341308650154e-06, {   0,   0,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.654964292590e-08,  2.566695090760e-05, {   2,   0,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  2.233406731852e-05, -3.347998228379e-06, {   2,   0,   0,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 {  2.182260597206e-05,  3.842457956606e-06, {   1,   0,   1,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.851514217789e-05, -7.148367486363e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0 } },
 { -7.836269353965e-06,  1.782148657725e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  2.520036181237e-05, -4.480050566004e-07, {   4,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.084474353164e-06,  1.655175924156e-05, {   2,   0,  -4,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.178307307350e-06, -2.244414311806e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  4.723563873909e-06, -2.089797513407e-05, {   0,   0,   3,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.068295537516e-05,  4.918934639539e-06, {   0,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.884251981168e-05, -6.755693161008e-06, {   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.314692885590e-06, -1.827054860471e-05, {   2,   0,  -1,   0,   0,   0,  14, -21,   0,   0,   0,   0,   0 } },
 {  1.150148626731e-07,  2.547007632101e-05, {   4,   0,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  2.434272463482e-05, -1.230079416272e-06, {   2,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0,   0 } },
 { -1.343049022349e-05, -1.213169448832e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 { -1.506006754720e-05,  1.048837762594e-05, {   3,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.588975592526e-05, -9.658045671010e-06, {   4,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  2.200014353563e-05, -3.538162026029e-06, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,   0,   0,   0,   0 } },
 { -3.098389403247e-08, -2.550223627202e-05, {   4,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.476881633152e-08, -2.544882810041e-05, {   4,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.782677577904e-06,  2.171011212110e-05, {   1,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.824039522012e-06,  1.665055557032e-05, {   1,  -1,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0 } },
 {  2.670434899842e-06, -2.279665566013e-05, {   1,   0,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.622493733835e-06,  2.383968261457e-05, {   1,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.095827541542e-08,  2.544186976475e-05, {   0,   0,   4,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.059723394694e-05,  4.839302032486e-06, {   1,  -1,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.429645872346e-05,  1.127047420397e-06, {   2,   0,   1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.505504794666e-06, -1.591074996412e-05, {   0,   2,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -2.559936464248e-06, -2.285536255346e-05, {   0,   0,   0,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.291956981849e-07,  2.487103822176e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  4.180287558161e-07,  2.497828895400e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 {  2.121999880292e-07,  2.516237846803e-05, {   0,   0,   3,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.343170248664e-05, -1.193674214056e-05, {   0,   0,   1,   0,   0,   8,  -5, -16,   4,   5,   0,   0,   0 } },
 { -1.343170248664e-05,  1.193674214056e-05, {   0,   0,   1,   0,   0,  -8,   5,  16,  -4,  -5,   0,   0,   0 } },
 {  8.449190064729e-06, -1.691682495646e-05, {   6,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.047051037819e-05,  4.887036451700e-06, {   1,  -1,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.002208197317e-05,  5.329904461002e-06, {   2,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 {  6.218677226320e-06,  1.913005138347e-05, {   2,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.458979444741e-05, -7.495203246292e-07, {   2,   2,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.048883608533e-05,  4.849601368441e-06, {   0,   1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.080623583193e-06, -1.825644286902e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  2.155794030892e-06,  2.317685279109e-05, {   1,  -1,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -2.155794030892e-06, -2.317685279109e-05, {   1,  -1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.753305995343e-05,  7.795044249456e-06, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  7.550450092488e-06, -1.777415461000e-05, {   1,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  2.875915994255e-08, -2.527253427493e-05, {   3,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.540284030616e-05, -9.882409328650e-06, {   2,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -2.402667617016e-05,  1.237267126663e-06, {   4,   0,  -2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.654662920917e-05,  8.717162453976e-06, {   2,   0,  -1,   0,   0,   4,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.748229728263e-05,  7.771969746900e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.384718037028e-07,  2.490149638584e-05, {   1,  -2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.789407800090e-05,  7.330575228081e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0 } },
 { -5.196004007328e-08, -2.517230360808e-05, {   4,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.430521926179e-05, -8.838643229256e-07, {   2,   0,  -2,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.980872194672e-05, -5.358538740322e-06, {   3,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  2.373611941947e-05, -1.425641955703e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   3,   0,   0,   0 } },
 {  1.497247903409e-06, -2.362945079582e-05, {   2,   0,   0,   0,   0,  -6,   9,   0,  -3,   0,   0,   0,   0 } },
 { -1.371092697628e-05,  1.141104888529e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -2,   0,   0,   0,   0 } },
 { -4.863407127706e-06,  2.025471020655e-05, {   1,  -1,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -2.496024341286e-05, -1.561076242338e-07, {   1,  -2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.351233034019e-05, -1.600950385042e-06, {   2,   0,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.923612478102e-05, -5.872500102664e-06, {   0,   0,   1,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.516208779414e-06,  1.758476655795e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.640748904471e-05, -8.678967920869e-06, {   2,  -2,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  4.133274501480e-07, -2.467037730750e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 { -6.145471456300e-06,  1.893457146855e-05, {   2,   0,  -1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.344420565732e-06,  2.071769146252e-05, {   1,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -4.887463575109e-06,  2.016919022947e-05, {   3,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  3.400074419287e-07, -2.470681012611e-05, {   0,   0,   4,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.151145887017e-05,  1.353409913869e-05, {   3,   1,  -2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.357222962580e-05, -1.146052453517e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   2,   0,   0,   0,   0 } },
 { -5.710209087747e-06,  1.931557352453e-05, {   1,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -5.512704010098e-06, -1.950490237572e-05, {   1,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.799908030798e-06,  2.121710354247e-05, {   1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.483118764520e-05, -1.018517943138e-05, {   1,   1,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.144527954987e-05,  3.570458441635e-06, {   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.747935770480e-06,  2.026711094252e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 { -9.003429884978e-06,  1.600301970099e-05, {   0,   2,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.482997357292e-05, -1.639720742034e-07, {   4,   0,  -2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.151308953287e-05, -3.479447054556e-06, {   2,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -2.058079045466e-05, -4.405418836882e-06, {   3,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.831768489384e-05, -6.663979280829e-06, {   2,   0,  -1,   0,   0,   0,  15, -25,   0,   0,   0,   0,   0 } },
 {  1.226010208318e-05, -1.271777005588e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 { -1.226010208318e-05,  1.271777005588e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   3,  -6,   0,   0,   0 } },
 {  4.122427504335e-06, -2.084763909809e-05, {   1,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.246977291709e-05, -2.493953328659e-06, {   2,   0,   0,   0,   0,   0,   8, -19,   0,   3,   0,   0,   0 } },
 {  8.313614288641e-06,  1.664852488824e-05, {   1,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.047497052390e-09, -2.493475903979e-05, {   2,   0,   1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.134338663826e-09, -2.492552282001e-05, {   4,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.282094178151e-06, -1.564819136009e-05, {   1,  -1,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  8.575913902319e-06,  1.634773541165e-05, {   2,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.968157869902e-05, -5.218348048972e-06, {   1,   1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.158343183020e-06,  2.072806024469e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  2.402854654374e-05, -8.559121705290e-07, {   0,   0,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.835521126317e-08, -2.483315116107e-05, {   0,   2,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.291386870962e-06,  1.558631784128e-05, {   0,   2,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  2.205069076020e-05, -2.806007331855e-06, {   0,   0,   1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.072562910051e-05, -4.124409010950e-06, {   4,   0,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.361143526754e-05, -1.122367125412e-05, {   3,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.326795358234e-06,  1.549639961154e-05, {   0,   0,   3,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  6.816606215243e-06, -1.800255349874e-05, {   2,   0,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.841181160832e-06, -1.597173564858e-05, {   2,   0,  -2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  9.665206838505e-06,  1.513391758906e-05, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -1.428028536039e-05, -1.051030817600e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   2,   0,   0,   0 } },
 { -2.732041959452e-08, -2.475469879878e-05, {   2,   0,   0,   0,   0,   0,   4,   0,   0,  -4,   0,   0,   0 } },
 { -4.613528754098e-06, -2.016527370551e-05, {   4,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.287334324090e-05, -1.187974256059e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   4,   0,   0,   0 } },
 {  1.129372066312e-05, -1.345645612954e-05, {   2,   0,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  2.332819060958e-05,  1.421517132100e-06, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 { -8.237923967742e-06,  1.649433179390e-05, {   2,  -2,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.975651790477e-05, -4.971266797187e-06, {   0,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 { -1.594990911880e-05,  8.775052429386e-06, {   2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.352094555679e-05,  1.119464473379e-05, {   2,   0,  -4,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -8.412851693422e-06,  1.629182654390e-05, {   0,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.950183639671e-08, -2.461702955862e-05, {   0,   0,   1,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  8.351749691793e-06, -1.634089862267e-05, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  9.740379219064e-06,  1.493574084825e-05, {   2,   0,   0,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 { -1.091563911634e-05, -1.375734265935e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,  11,  -3,   0,   0 } },
 {  2.315520943096e-05, -1.512446449046e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 { -4.817486905584e-06, -1.984894679879e-05, {   1,  -1,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  4.119698171434e-06,  2.054479859201e-05, {   2,   0,   0,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 {  9.261551433397e-06, -1.539271041125e-05, {   0,   0,   3,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -7.224005574283e-07, -2.392809287591e-05, {   2,   0,  -4,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.657581683654e-07, -2.447672237758e-05, {   4,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.585719759868e-05, -8.782858644714e-06, {   0,   0,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.330241501416e-05, -1.133692337124e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   1, -11,   3,   0,   0 } },
 { -6.726159856630e-06,  1.791291949226e-05, {   1,   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.450399257138e-05, -1.201347613640e-07, {   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.436801002924e-05, -1.024320328241e-05, {   1,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.967338739355e-06,  1.963991455179e-05, {   1,   1,   0,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  9.571531235920e-06, -1.502635778471e-05, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  1.985951152705e-05,  4.734421824241e-06, {   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.272199328981e-05, -1.187140007978e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0 } },
 {  1.397300555949e-05, -1.061231807504e-05, {   0,   0,   1,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 {  8.174095714908e-06,  1.640903358248e-05, {   0,   0,   2,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.701218940589e-05, -7.562923738208e-06, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.383600213935e-05,  1.071630687330e-05, {   2,   0,   0,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 { -1.989756353478e-06,  2.253866803853e-05, {   0,   0,   2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.209928492797e-05,  2.384117957455e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 {  8.272030150437e-06,  1.619880120870e-05, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.378535835175e-05,  1.067905979612e-05, {   2,   0,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0 } },
 { -1.976533619797e-05, -4.690313953904e-06, {   2,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.487523708811e-05,  9.577621843986e-06, {   4,   0,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -7.050875103377e-06,  1.739067828830e-05, {   2,   0,  -1,   0,   0,  -2,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.304141232269e-05,  1.383371847974e-06, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -3,   0,   0,   0 } },
 {  9.451296573893e-06, -1.497316244237e-05, {   2,   0,  -2,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  8.088767629121e-06,  1.633514332823e-05, {   1,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.133005048630e-05,  3.089075772090e-06, {   2,   0,   0,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -1.694577460323e-05,  7.471253316920e-06, {   1,  -1,   2,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 { -4.237327721528e-07, -2.398936039201e-05, {   2,   0,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -1.240110706867e-07, -2.428792863009e-05, {   2,   2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.205936556293e-05, -2.344785974098e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 { -1.871000578059e-05,  5.676193202963e-06, {   4,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.037143390465e-05, -3.995082325150e-06, {   2,   0,  -1,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.686532567472e-05, -7.498612251272e-06, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.324434696592e-05, -1.111829816972e-05, {   2,  -2,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.878917878146e-05,  5.566285645499e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  2.712366117028e-08, -2.432759104897e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  9.806335609300e-09,  2.433910445034e-05, {   3,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.976781979806e-05, -4.551503073287e-06, {   1,   1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.919896100248e-05,  5.090458686683e-06, {   0,   2,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.587297252040e-05, -8.409986928147e-06, {   0,   0,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.395604077943e-05,  3.148314289774e-07, {   2,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.759311278671e-06, -1.950983571107e-05, {   2,   0,   1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  2.535124003326e-06, -2.173209882421e-05, {   4,   0,  -1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  2.329079614304e-05,  9.750366000083e-07, {   2,  -2,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.001778681463e-07, -2.365737922807e-05, {   2,   0,  -2,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 {  9.400310111715e-06, -1.484483798735e-05, {   0,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  4.261127991187e-07,  2.381265498788e-05, {   2,   0,  -1,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.671690584421e-07, -2.396756885053e-05, {   2,   0,   0,   0,   0,  28, -28,   0,   0,   0,   0,   0,   0 } },
 {  9.025168847759e-06,  1.520796600414e-05, {   1,   0,  -3,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 { -1.236206166344e-05, -1.185907666461e-05, {   2,   0,   0,   0,   0,  -9,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.174060393860e-06,  1.604193412476e-05, {   2,   0,  -1,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.020732530077e-08, -2.419282733644e-05, {   1,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.513617904109e-05,  9.064647926306e-06, {   1,  -1,  -2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -8.014873625438e-06, -1.618530780097e-05, {   1,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.907018384205e-05,  5.114765316619e-06, {   1,  -1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.532658023451e-05,  8.855336053896e-06, {   2,   0,   0,   0,   0,   0,  -6,   0,   5,   0,   0,   0,   0 } },
 {  1.528994553819e-05,  8.891603648436e-06, {   0,   0,   1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -1.523467576919e-05,  8.935373046859e-06, {   2,   0,   2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  5.002949086829e-08,  2.411245748080e-05, {   2,  -2,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.036599793208e-05, -1.379004575824e-05, {   2,   0,   0,   0,   0,  -1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.394483511563e-06, -1.676025378967e-05, {   1,   0,   0,   0,   0,   0,  10, -17,   0,   0,   0,   0,   0 } },
 { -8.122239274775e-06,  1.601704806310e-05, {   2,   0,  -2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.483924650702e-06, -1.963625197557e-05, {   1,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.792046633196e-08, -2.407856598888e-05, {   2,   0,   1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.069097175974e-05,  1.342016925017e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   2, -11,   3,   0,   0 } },
 { -5.256199883708e-07, -2.358148485869e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -2.947292378357e-06, -2.114883461975e-05, {   1,  -1,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.525751319734e-05, -8.837276746616e-06, {   0,   0,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0 } },
 { -1.503786534871e-05, -9.054802178825e-06, {   1,  -1,   2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.027683365093e-05, -3.815437569455e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 { -2.027683365093e-05,  3.815437569455e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  1.958795296793e-05,  4.483268191254e-06, {   1,  -1,  -1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.668676586294e-05,  7.371913547315e-06, {   1,  -1,   1,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 {  2.261597318175e-05, -1.439698020240e-06, {   2,   0,  -1,   0,   0,   6,  -8,   0,   1,   0,   0,   0,   0 } },
 {  2.545881299104e-06,  2.150522965699e-05, {   2,   0,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.102600066719e-06,  2.294744145279e-05, {   4,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.166323255528e-06, -1.588116936940e-05, {   2,   0,   2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.731532555558e-07,  2.387343609605e-05, {   2,  -2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.486229541352e-05, -9.175153063511e-06, {   2,   0,   1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -1.591775078591e-05,  8.117723353375e-06, {   0,   0,   2,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  6.782117329248e-06,  1.723595708469e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 { -5.168453883972e-06, -1.884557015687e-05, {   1,   1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.035172035520e-06,  1.995700230223e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   2,   0,   0,   0 } },
 { -1.568741877546e-05,  8.294896294776e-06, {   2,   0,  -1,   0,   0,  11, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.698495050800e-06, -2.128019560158e-05, {   2,   0,   0,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  7.986465892485e-06,  1.599041978530e-05, {   6,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.986465892485e-06, -1.599041978530e-05, {   6,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  8.518004788820e-06,  1.544969950427e-05, {   2,  -2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.301594626062e-05, -1.094679982936e-05, {   1,  -1,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.298848714797e-05, -9.662707441893e-07, {   2,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.299242185880e-05, -9.619620824842e-07, {   2,  -2,   3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.146032772059e-06, -1.780444148465e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   2,   0,   0,   0 } },
 { -6.108988752170e-08, -2.388887490531e-05, {   2,   0,   1,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 {  7.140363754124e-06,  1.680313364054e-05, {   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 {  7.404099069055e-06, -1.646888189014e-05, {   3,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.934881083667e-05,  4.518025997408e-06, {   3,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.990584791610e-06, -2.187435621892e-05, {   2,   0,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.928745539702e-05,  4.574113729384e-06, {   3,  -1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.577848329197e-05,  8.073122118198e-06, {   2,   0,   0,   0,   0,   0,  18, -22,   0,   0,   0,   0,   0 } },
 { -4.652107990847e-06,  1.919672403230e-05, {   2,   0,   0,   0,   0, -16,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.540449460844e-05,  8.434450060179e-06, {   0,   0,   2,   0,   0, -18,  24, -16,   4,   5,   0,   0,   0 } },
 { -3.615379094622e-09,  2.382015882226e-05, {   2,   2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.739615986693e-06, -1.608373847915e-05, {   0,   0,   1,   0,   0,   0,  -9,  16,   0,   0,   0,   0,   0 } },
 { -7.066491924244e-06,  1.674840893211e-05, {   2,   0,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.626829392910e-05, -7.546538008963e-06, {   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.683309506122e-06,  2.010801656168e-05, {   1,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.773536635607e-05,  6.045371445980e-06, {   2,   0,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.647087474178e-05, -7.286562706556e-06, {   2,   0,  -3,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  8.349210506830e-06, -1.537959110976e-05, {   4,   0,  -2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  1.907738619600e-05, -4.617246441717e-06, {   3,  -1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.946535841622e-06,  1.974490797494e-05, {   1,  -1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.390423984949e-06,  1.929045563106e-05, {   2,   0,   0,   0,   0, -18,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.406800004920e-05, -9.593001757657e-06, {   2,   0,  -1,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -3.509577549265e-06, -2.013628293696e-05, {   2,   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -7.778050330161e-06, -1.586395669452e-05, {   1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.287473604141e-05, -7.559212959889e-07, {   0,   0,   1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.763711020093e-05, -5.987475811843e-06, {   0,   0,   1,   0,   2,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.072835040276e-05, -1.289487461511e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.872662956340e-05,  4.895174625422e-06, {   1,   1,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.871605692535e-05,  4.898900625117e-06, {   1,  -1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.804734744390e-05,  5.566986666966e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.515606371717e-07, -2.342790751926e-05, {   2,   0,   1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.844667407272e-05,  5.131310874076e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 { -7.065134711984e-06,  1.650458623973e-05, {   0,   0,   1,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.773814397751e-06,  1.976988335826e-05, {   1,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.580488791859e-06,  1.995808786776e-05, {   2,   0,   0,   0,   0,   0,   4,   0,  -6,   5,   0,   0,   0 } },
 { -2.337139116221e-05, -1.639240267203e-07, {   5,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.607326114689e-05,  7.455564639567e-06, {   2,   0,  -1,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0 } },
 { -6.327169382332e-06,  1.720118977252e-05, {   2,   0,   0,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.870774799161e-08, -2.347593314372e-05, {   2,   0,   2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.429354568018e-06,  2.108134922100e-05, {   2,   0,   1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -6.198598648915e-06, -1.726536053352e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  6.548850746000e-08,  2.337921493482e-05, {   1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 { -1.622524810546e-05,  7.214617840980e-06, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.818541935173e-05,  5.249173349501e-06, {   1,   1,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.120372354541e-05, -2.219939392741e-06, {   4,   0,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.314810190465e-05,  2.608826387690e-07, {   4,   0,  -3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.187106203672e-06, -1.520256622989e-05, {   2,  -2,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -4.081340490659e-06,  1.930103991477e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 { -4.388660715683e-06, -1.899267212387e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  2.255416646336e-05,  8.247886534693e-07, {   0,   0,   1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -1.250254297180e-05, -1.087468619758e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 { -2.799657397996e-06, -2.057231648088e-05, {   3,  -1,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.650613032834e-06, -1.870831830297e-05, {   1,   1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  7.780215676398e-06, -1.557733040743e-05, {   4,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.450199991716e-06,  1.388282798126e-05, {   1,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -3.770331117208e-06,  1.955112211146e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  1.562486643407e-06, -2.173145532319e-05, {   2,   0,  -1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.774425267304e-05, -5.536812855944e-06, {   1,  -1,   0,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -9.573501729310e-06, -1.369718027767e-05, {   4,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -5.085079900737e-06, -1.818314140951e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  1.645103559006e-05,  6.811488665385e-06, {   0,   0,   2,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.715080676751e-05, -6.088764436768e-06, {   2,   0,  -3,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -1.714643661518e-05,  6.083696689058e-06, {   2,   0,  -3,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  1.079405293813e-06,  2.211088110139e-05, {   0,   0,   2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -4.387610996360e-06,  1.878447916278e-05, {   2,   0,   0,   0,   0, -17,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.590528393809e-05, -7.259188734523e-06, {   0,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.544353731470e-05,  7.716204048264e-06, {   4,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -4.362338773874e-06,  1.879165023852e-05, {   1,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.054779873811e-05, -2.599554535676e-06, {   2,   0,   0,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.636590887884e-05, -6.771449858034e-06, {   2,   1,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.930519564281e-05,  3.818866186040e-06, {   4,   0,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.570892195151e-07, -2.296642504558e-05, {   4,  -2,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.869564923221e-05,  4.424919802783e-06, {   5,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.648506009706e-06, -1.546834748338e-05, {   0,   0,   1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.684075506117e-06,  1.542321969831e-05, {   2,   0,   1,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.200947329762e-06,  2.189807600375e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  1.581388338578e-05,  7.284147992267e-06, {   0,   0,   1,   0,   0,   5,  -2, -12,   3,   0,   0,   0,   0 } },
 {  1.581388338578e-05, -7.284147992267e-06, {   0,   0,   1,   0,   0,  -5,   2,  12,  -3,   0,   0,   0,   0 } },
 { -1.717005064952e-05, -5.920901050900e-06, {   2,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.819719724413e-05, -4.873845780023e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  1.560354835064e-05,  7.450583251815e-06, {   2,  -2,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.150749987613e-05,  1.545978723027e-06, {   0,   0,   2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  2.295574792576e-05, -9.750350614500e-08, {   0,   0,   2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.211946000094e-05, -9.245614222872e-07, {   0,   0,   2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.328329520446e-06,  1.469564098190e-05, {   2,  -2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -1.478987308783e-05, -8.228520317399e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 { -1.792792775180e-05,  5.088897967018e-06, {   1,   0,   1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 {  3.463522650867e-06,  1.955001879635e-05, {   1,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.217186522564e-05,  8.312047657528e-07, {   2,  -2,  -2,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.030199010251e-05,  1.267834921124e-05, {   0,   0,   1,   0,   0,   0,  20, -38,   0,   6,   0,   0,   0 } },
 {  1.030199010251e-05, -1.267834921124e-05, {   0,   0,   1,   0,   0,   0, -20,  38,   0,  -6,   0,   0,   0 } },
 { -8.673937043931e-07,  2.210809199694e-05, {   3,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.548202344934e-05, -7.492339161860e-06, {   1,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -2.107880641620e-05, -1.888295352244e-06, {   2,   0,   0,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 {  2.107880641620e-05,  1.888295352244e-06, {   2,   0,  -2,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 {  1.768836652382e-05, -5.278673758732e-06, {   2,   0,   0,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 { -1.021901269158e-05, -1.274680847369e-05, {   2,   0,   0,   0,   0,  -2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.210861652024e-05,  8.558615994600e-07, {   2,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  4.127173375074e-07,  2.254630551624e-05, {   3,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.103574371768e-05, -1.190693477322e-05, {   0,   0,   2,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.485696815918e-06, -2.144531123579e-05, {   1,   1,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.031920361546e-05,  2.603962319817e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -6.807152882236e-06, -1.611539561391e-05, {   1,   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  9.304808721043e-06, -1.361132982376e-05, {   2,   0,  -1,   0,   0,   0,  17, -22,   0,   0,   0,   0,   0 } },
 { -2.102536530733e-05,  1.889894628097e-06, {   1,   1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.851510489521e-06, -1.806234397763e-05, {   1,   1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.716521407392e-08, -2.286538968070e-05, {   0,   0,   2,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.705205927489e-06, -2.016941456374e-05, {   2,   0,   0,   0,   0,   0,   4,   0,  -2,  -5,   0,   0,   0 } },
 {  1.454282524878e-08, -2.283849347793e-05, {   3,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.949976901077e-05, -3.342985472010e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   5,   0,   0,   0 } },
 { -2.010715917972e-05, -2.734076447464e-06, {   2,   0,  -1,   0,   0,  -5,   5,   4,   0,   0,   0,   0,   0 } },
 {  1.806937011915e-05, -4.758256854122e-06, {   1,  -1,   1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.729062917737e-07,  2.265442647383e-05, {   0,   0,   1,   0,   0,  -3,   5,  -2,   0,   0,   0,   0,   0 } },
 {  6.973418396969e-06,  1.583754546962e-05, {   1,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.447562334079e-05, -8.334824110010e-06, {   4,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.769179583188e-08, -2.278811843127e-05, {   0,   2,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.211840781672e-05, -6.745586604580e-07, {   2,   0,  -4,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.889589374889e-05, -3.894265113842e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 {  4.796170281777e-06, -1.798857101819e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   2,   0 } },
 {  2.172895305044e-05,  1.047719840077e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -2.236720924084e-05, -4.024749704730e-07, {   5,  -2,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  2.236720924084e-05,  4.024749704730e-07, {   1,  -2,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.098266897367e-05, -1.783414185599e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 { -1.753584097346e-06,  2.100922121087e-05, {   0,   0,   1,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.063775909546e-08,  2.272466445254e-05, {   2,   0,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.132184870885e-06,  1.461216949218e-05, {   1,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.306790317631e-07, -2.261132072047e-05, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 {  1.828659455165e-05, -4.451913352222e-06, {   2,  -1,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.868120331362e-05,  4.052431289783e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.301404307497e-05, -9.705713535562e-06, {   2,   0,   0,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 { -1.547170850997e-05, -7.246645489041e-06, {   2,   0,  -1,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.701515318230e-06,  1.900941928029e-05, {   1,  -1,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.316100075189e-05, -9.547270207658e-06, {   4,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.230392574074e-06,  1.347734502991e-05, {   2,   0,   0,   0,   0,   0,  17, -22,   0,   0,   0,   0,   0 } },
 {  6.770469036070e-06, -1.593430692464e-05, {   1,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.770469036070e-06,  1.593430692464e-05, {   1,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.914249477558e-05,  3.556951058330e-06, {   3,  -1,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.026161409728e-05, -1.243516579373e-05, {   1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.305986450220e-05, -9.626709840552e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  3.986356469434e-08, -2.264213390665e-05, {   0,   2,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  4.468596087743e-08, -2.263249300489e-05, {   2,   0,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.322578931758e-06, -1.435405728469e-05, {   4,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -2.132390453092e-05, -1.350717167876e-06, {   2,   0,  -1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.136323880901e-06, -1.552624311775e-05, {   1,  -1,   1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.271947382763e-05,  9.939524720419e-06, {   2,   0,  -1,   0,   0,   0,  12, -20,   0,   0,   0,   0,   0 } },
 {  5.192300007953e-06,  1.745690585724e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  1.225609691796e-05,  1.038183689809e-05, {   3,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.599086282854e-05, -6.646226696069e-06, {   2,  -2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.909780226186e-06, -1.471993179375e-05, {   0,   0,   3,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.474113328620e-05, -7.888106523488e-06, {   1,  -1,   2,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 {  1.474113328620e-05,  7.888106523488e-06, {   1,  -1,   0,   0,   0, -24,  30,  -6,   0,   0,   0,   0,   0 } },
 {  2.092392480442e-05,  1.699769361916e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  1.487937816255e-06, -2.113412425686e-05, {   2,   0,  -4,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.379141189038e-05,  8.826010939485e-06, {   1,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -7.310589317720e-06,  1.529644215600e-05, {   1,   1,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  5.338231954568e-06,  1.726590545024e-05, {   0,   0,   2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.361108209175e-06, -1.724188869599e-05, {   0,   0,   1,   0,   0,   8, -21,  16,  -4,  -5,   0,   0,   0 } },
 { -5.361108209175e-06,  1.724188869599e-05, {   0,   0,   1,   0,   0,  -8,  21, -16,   4,   5,   0,   0,   0 } },
 { -1.275661249871e-05, -9.820573745186e-06, {   1,   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  3.398696686382e-07, -2.223193521574e-05, {   0,   0,   1,   0,   0,  -2,   4,   0,  -2,   0,   0,   0,   0 } },
 {  3.320820020588e-06, -1.924312043728e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 {  1.672721067365e-05, -5.836205144329e-06, {   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.995225045119e-09, -2.255481541306e-05, {   0,   0,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 {  8.560393086497e-06, -1.399517372821e-05, {   2,   0,   0,   0,   0,   0,   4,   0,  -7,   0,   0,   0,   0 } },
 {  1.429123990553e-05, -8.262024173310e-06, {   2,  -2,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.011728123499e-05,  1.243220852781e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -7,   0,   0,   0,   0 } },
 { -1.457203043273e-05, -7.972446731610e-06, {   0,   0,   0,   0,   0,  18, -24,  16,  -4,  -5,   0,   0,   0 } },
 { -1.721383498408e-06, -2.082018929660e-05, {   0,   0,   1,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.585498782941e-05, -6.681901518636e-06, {   2,  -2,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.364671704518e-05,  8.886471901561e-06, {   2,  -2,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  1.291699986409e-05,  9.614137661948e-06, {   2,   0,   0,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 { -1.986018696134e-05, -2.667419341261e-06, {   2,   0,  -1,   0,   0,  -2,   8,  -9,   0,   0,   0,   0,   0 } },
 {  1.822920315111e-05, -4.291920187073e-06, {   1,  -1,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.911864515201e-05, -3.401347639250e-06, {   0,   0,   1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.691482771612e-06, -1.982137408458e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -2.189827383606e-05,  6.139478965506e-07, {   0,   0,   1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.444339113701e-05,  8.044638702384e-06, {   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.669123736812e-05,  5.791276634386e-06, {   1,   0,   1,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 { -1.669123736812e-05, -5.791276634386e-06, {   1,   0,  -1,   0,   0,   0, -30,  33,   1,   0,   0,   0,   0 } },
 {  1.859505048859e-05,  3.885219868767e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 { -8.178305384552e-06,  1.430161248156e-05, {   3,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -7.454881487282e-06, -1.500511994862e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  3.665318009173e-06, -1.877759344774e-05, {   1,   1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.197541794306e-06,  2.024006819005e-05, {   2,   0,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.345941116509e-05, -8.963809114319e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  1.554171040656e-05,  6.875205881570e-06, {   1,  -1,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -1.554171040656e-05, -6.875205881570e-06, {   1,  -1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  3.226588301303e-09,  2.240776329930e-05, {   1,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.812296727838e-05,  4.263901429438e-06, {   1,  -1,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.721928486001e-05,  5.142090167835e-06, {   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.361384145914e-06, -1.699235300553e-05, {   1,   1,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -6.885740536583e-08,  2.227813774104e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  5.358565970368e-06,  1.698219560415e-05, {   3,  -1,   2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.495566256096e-08, -2.229284508830e-05, {   2,   0,   1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.305330390428e-05, -9.271951800585e-06, {   0,   0,   1,   0,   0,   0, -11,  19,   0,   0,   0,   0,   0 } },
 { -2.109576566270e-05, -1.218128893389e-06, {   4,   0,  -2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.783671667080e-05, -4.476259781301e-06, {   3,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.772419519633e-06,  1.653997058376e-05, {   0,   0,   1,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.783576312781e-05,  4.468218815301e-06, {   2,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.900609530870e-05,  3.296119868501e-06, {   1,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.457416671702e-06, -1.784245040862e-05, {   2,   0,  -1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 {  4.598545231483e-07,  2.183693139130e-05, {   1,   0,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.569317747940e-05,  6.598125306769e-06, {   1,   1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.442611023860e-05, -7.863505822256e-06, {   4,   0,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  1.192194548569e-05, -1.035994802846e-05, {   2,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  7.952854783553e-06,  1.432370801209e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -4.443865983400e-07, -2.183178238523e-05, {   2,   0,  -1,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.290788130533e-05, -9.365219028569e-06, {   0,   0,   1,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.264800625952e-06,  1.499106409591e-05, {   0,   0,   1,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 { -3.588344728408e-06,  1.865458062056e-05, {   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.095657721067e-05,  1.128190940508e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -1.605502850695e-05,  6.180631874245e-06, {   2,   0,  -1,   0,   0,   0,  -2,   2,   0,  -2,   0,   0,   0 } },
 { -7.341002787931e-06, -1.489213975945e-05, {   2,   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -7.732811892683e-06,  1.448562237074e-05, {   2,   0,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.796211203288e-05, -4.249934685347e-06, {   1,   1,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.781726778098e-06,  2.042710350675e-05, {   0,   0,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -2.113704424171e-05, -1.060240180958e-06, {   2,   0,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  2.173081359307e-05, -4.644234855610e-07, {   0,   2,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.903264503117e-05, -3.151925558730e-06, {   2,  -2,   2,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.361219600202e-08,  2.215637543818e-05, {   3,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.203446671032e-05, -1.014283386577e-05, {   1,  -1,   0,   0,   0,  17, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.627560155973e-06, -1.854063546571e-05, {   1,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.307965349520e-06, -1.583387417100e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   7,   0,   0,   0 } },
 { -4.964234602350e-06, -1.717031552267e-05, {   4,   0,  -1,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.697567007811e-05,  5.158147356408e-06, {   2,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  8.789512345398e-06,  1.333875818425e-05, {   2,   0,  -1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0 } },
 {  1.695997291957e-05, -5.152268978464e-06, {   2,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.173828070483e-09,  2.210721290777e-05, {   0,   0,   3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.790206697009e-05, -4.209093408410e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   6,  -3,   0,   0 } },
 { -2.192763755303e-06, -1.990878090409e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 {  1.651767919854e-05, -5.579999917167e-06, {   2,   0,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.176178075986e-05,  1.032862363376e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -8,   0,   0,   0 } },
 { -8.377765357350e-06,  1.371224275547e-05, {   2,   0,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0 } },
 {  1.861373055733e-05, -3.460407577410e-06, {   4,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.906116993279e-06,  2.016664236147e-05, {   1,  -1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -3.055217044938e-06,  1.900837695801e-05, {   1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.368009878987e-05, -8.382612069007e-06, {   2,   0,  -1,   0,   0,   5,  -3,  -7,   8,  -6,   0,   0,   0 } },
 {  9.968940550640e-07,  2.106564642363e-05, {   4,   0,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.368098223935e-05,  8.380514391071e-06, {   2,   0,  -1,   0,   0,   1,  -3,   7,  -8,   6,   0,   0,   0 } },
 {  1.353941275901e-05,  8.520777623460e-06, {   2,  -2,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.783715632078e-05, -4.215012779239e-06, {   1,   1,   1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.285079369717e-08, -2.195416027030e-05, {   4,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.816454400269e-06,  1.922328600807e-05, {   0,   0,   1,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 {  1.741625839932e-05,  4.622700174126e-06, {   0,   0,   4,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.286006895129e-05, -9.170849978746e-06, {   1,   1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.118477909104e-05,  1.084546447846e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -5.994080900982e-06,  1.601924961694e-05, {   0,   0,   1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  1.071996962215e-05,  1.129245736695e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 {  1.635899624113e-07, -2.184849252865e-05, {   2,  -2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.506617850327e-06,  1.750224634397e-05, {   1,  -1,  -1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.491030563213e-05,  7.097521963733e-06, {   4,  -2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.777767434696e-05,  4.228553562689e-06, {   4,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.112686695186e-05,  8.707481194118e-07, {   2,   0,  -2,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  2.019140001404e-05,  1.787733153515e-06, {   2,  -2,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  7.655895320084e-06, -1.432251937309e-05, {   2,   0,  -2,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.362015187030e-06, -1.960492986586e-05, {   0,   0,   1,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.766057775475e-05,  4.299366184952e-06, {   2,  -1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.663980605879e-05,  5.317260189287e-06, {   2,   2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.774690534788e-05,  4.201399145013e-06, {   3,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.662339441465e-05, -5.311620945333e-06, {   2,   2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.966379047769e-06,  1.596574058871e-05, {   3,  -1,  -2,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  5.966379047769e-06, -1.596574058871e-05, {   1,   1,   0,   0,   0, -29,  33,   0,  -2,   0,   0,   0,   0 } },
 {  2.595117349751e-06, -1.932897143314e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 {  1.395567873094e-05,  7.966577123908e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 {  1.471706672012e-05, -7.203888072058e-06, {   2,   0,  -2,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  1.329142015259e-07, -2.178687051923e-05, {   0,   0,   3,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -7.879865464998e-06,  1.403082233699e-05, {   0,   0,   1,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  1.017100312446e-07, -2.180874096568e-05, {   4,   0,  -2,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.266188012334e-06, -1.862669614037e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -3,   0,   0,   0,   0 } },
 { -5.444619942364e-06,  1.644223933724e-05, {   2,   0,  -1,   0,   0,   5,  -9,   5,   0,   0,   0,   0,   0 } },
 { -6.560778095311e-06,  1.532379611249e-05, {   1,   0,   0,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.977421708742e-06, -1.990408378957e-05, {   2,   0,  -1,   0,   0,   0,  15, -21,   0,   0,   0,   0,   0 } },
 { -1.184676512701e-06,  2.068762531360e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.116398644512e-05,  7.056836097655e-07, {   2,   0,  -2,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.470602772986e-05, -7.150510530994e-06, {   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.239957667481e-05, -9.447813670669e-06, {   2,   0,  -1,   0,   0,   0,  -3,   5,   4,   0,   0,   0,   0 } },
 { -1.245370274389e-05,  9.389861180090e-06, {   1,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.400719670313e-07,  2.159718478823e-05, {   0,   0,   3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.469996180684e-05,  7.137259878686e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 { -7.854110094567e-06, -1.398273389542e-05, {   2,   0,  -2,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -1.773049293185e-05,  4.101272270791e-06, {   2,   1,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.773049293185e-05, -4.101272270791e-06, {   2,  -1,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  7.205330268029e-06, -1.462617496797e-05, {   0,   2,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -7.370537994235e-07, -2.109383653427e-05, {   0,   2,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.125930433208e-07, -2.161645436278e-05, {   2,  -2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.084054704490e-05,  1.097939187464e-05, {   2,   0,   0,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0 } },
 {  1.366242664226e-05, -8.157466895126e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 { -1.001515765554e-05, -1.180146084847e-05, {   2,   0,   0,   0,   0,  -3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.657426262752e-06,  1.515562248289e-05, {   1,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.620824851619e-06,  1.519047762790e-05, {   0,   0,   1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 {  4.897630382506e-06,  1.690260592911e-05, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0 } },
 { -1.160217036658e-05, -1.018691206379e-05, {   0,   0,   1,   0,   0,   0,  -3,   4,  -2,   0,   0,   0,   0 } },
 { -1.947454520906e-05,  2.291767381356e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -5.668007505261e-06,  1.609177433178e-05, {   2,   0,   0,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.655835205786e-06, -1.510301741906e-05, {   0,   1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.375841713318e-06, -2.037718234866e-05, {   4,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.182664292465e-05,  9.904945095210e-06, {   3,  -2,  -1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -1.182664292465e-05, -9.904945095210e-06, {   1,   2,  -1,   0,   0, -20,  18,   0,   0,   1,   0,   0,   0 } },
 { -1.875761895033e-05,  2.973806612888e-06, {   2,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.190475679975e-05, -9.820804056609e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  2.456403176611e-06,  1.925798749991e-05, {   2,   0,   1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.232475633023e-05,  9.386526694885e-06, {   2,   0,  -1,   0,   0,   0,   3,  -5,  -4,   0,   0,   0,   0 } },
 {  1.566323614833e-05,  6.047501365510e-06, {   2,   0,   2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -1.991976715511e-05, -1.789658203769e-06, {   1,   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.991981649951e-05, -1.789108891562e-06, {   1,   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.991981649951e-05, -1.789108891562e-06, {   1,  -1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.991981681177e-05,  1.789105414898e-06, {   5,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.991981681177e-05,  1.789105414898e-06, {   3,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.005553162316e-05, -1.649059582582e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -2.770247115681e-06, -1.893259577359e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   7,   0,   0,   0 } },
 { -8.841169911148e-08, -2.161118624451e-05, {   4,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.757812872707e-05, -4.114706882874e-06, {   1,  -1,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.378364390897e-05, -7.887326842247e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 { -4.536375784832e-06, -1.713324159406e-05, {   1,   1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.585394512379e-06, -1.608420824424e-05, {   3,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.743792225916e-05, -4.219470812912e-06, {   3,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.123494853262e-05,  4.213641402924e-07, {   1,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.268586683072e-05, -8.962604836099e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -5,   0,   0,   0 } },
 { -1.433951483388e-05,  7.303796493168e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 { -2.593506641975e-08, -2.161465793330e-05, {   0,   2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.059962457532e-05, -1.103687886585e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 {  9.751844023051e-06, -1.188249921629e-05, {   1,   1,  -2,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.828154469340e-06, -1.480241258927e-05, {   1,   1,   0,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.193260219806e-05,  9.692351606405e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 { -7.120756663745e-06,  1.447760506890e-05, {   1,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.699026800158e-06, -1.489848853936e-05, {   2,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.582061800209e-05,  5.773101454464e-06, {   2,   0,  -1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.222601297945e-06,  1.435937698259e-05, {   0,   2,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -9.693163426260e-06,  1.188555896595e-05, {   2,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.917763018271e-05,  2.398254466039e-06, {   2,   0,   0,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 {  2.092693379691e-05, -6.363815970668e-07, {   2,   0,   0,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 { -3.416769901659e-06, -1.811874088787e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -1.459257846997e-05, -6.941895306771e-06, {   0,   0,   0,   0,   5,   0, -13,   7,   0,   0,   0,   0,   0 } },
 { -1.346746473025e-05, -8.052577027251e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 {  1.689002732700e-05,  4.627753842946e-06, {   2,   0,   2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -5.533933331183e-06,  1.598159053165e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  7.761597073721e-06,  1.374878416573e-05, {   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.293230170359e-09,  2.150329320923e-05, {   0,   0,   2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  3.259448265648e-06, -1.824410022835e-05, {   2,  -2,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.072241280785e-05, -1.077403104455e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -1.763572863386e-05, -3.857662019026e-06, {   4,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -3.238988977819e-06, -1.825087140134e-05, {   1,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.897281209754e-05,  2.516264686928e-06, {   2,   0,  -1,   0,   0,   2,  -8,   9,   0,   0,   0,   0,   0 } },
 { -1.782444879693e-05, -3.663338637624e-06, {   1,  -3,   2,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 {  1.782444879693e-05,  3.663338637624e-06, {   1,  -3,  -2,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -1.229212924887e-05,  9.153965106220e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  6.032032194182e-06, -1.541098073971e-05, {   2,   0,  -1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 { -1.206320090900e-05, -9.375192504542e-06, {   2,   0,  -1,   0,   0,   0,  14, -22,   0,   0,   0,   0,   0 } },
 { -1.506248249677e-05, -6.366281163762e-06, {   0,   0,   1,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.934763261323e-08,  2.140843142641e-05, {   1,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  8.848305988122e-08,  2.133141040199e-05, {   2,   0,  -2,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  2.512447395663e-07,  2.116141719709e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  1.740049281068e-05,  4.009420370314e-06, {   1,  -1,   1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.975065675242e-05,  1.647089544481e-06, {   4,   0,  -3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.205016632675e-06, -1.418850530114e-05, {   0,   0,   2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.612189855602e-05,  5.262673787274e-06, {   1,   1,   0,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.839603284331e-06, -1.953905563907e-05, {   1,  -1,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -6.910335320466e-06,  1.445625288484e-05, {   2,   0,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.026401951506e-08,  2.133560366449e-05, {   2,  -2,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.670146039298e-05,  4.650206593047e-06, {   3,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.670112979630e-05, -4.650123228367e-06, {   3,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.446140555066e-05,  6.879639959118e-06, {   4,   0,  -2,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -3.376030075641e-06,  1.795948098484e-05, {   1,  -1,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.741702858989e-05, -3.911540117324e-06, {   3,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.074810113922e-05, -1.057941896332e-05, {   2,   0,   0,   0,   0,  -8,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.123264956601e-05,  7.647508824203e-08, {   2,   0,  -1,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 { -9.936604391195e-06,  1.136645520291e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 {  1.186941810500e-05,  9.431553159321e-06, {   0,   0,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.045524167618e-05,  8.371818493904e-07, {   2,   0,   0,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 {  7.151761258745e-06, -1.413387127550e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -3.379942375434e-06,  1.790266735171e-05, {   1,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.047772913068e-05, -7.949965759422e-07, {   2,   0,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.639757863192e-05,  4.867538537009e-06, {   1,   1,   0,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.724680921409e-08,  2.123746297276e-05, {   3,   0,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.755144453041e-06, -1.449671237576e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 {  4.787118538705e-06,  1.646226201217e-05, {   1,   1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  3.408198700932e-06, -1.781918876090e-05, {   1,   1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.310957205784e-06,  1.891597357202e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 {  2.079651989549e-07,  2.101694442437e-05, {   2,   0,  -1,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -5.532352881953e-06,  1.568341860716e-05, {   2,   0,   0,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 {  3.519184954761e-06, -1.767952026807e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  1.381736672041e-05, -7.378333533065e-06, {   2,   0,  -1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 {  2.702158708814e-06,  1.848108601834e-05, {   1,  -1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.315891308781e-06,  1.886431719317e-05, {   2,   0,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 { -7.000647901830e-07,  2.047940576585e-05, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.313332134740e-05,  8.038528933959e-06, {   1,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.774791479354e-06, -1.739096356427e-05, {   1,  -1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.582334683379e-06,  1.555382597668e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 {  1.999821962102e-07, -2.093439748038e-05, {   2,   0,  -3,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.632836639371e-05, -4.803855457117e-06, {   2,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.996335892198e-06, -1.612871182679e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  7.013555946272e-08,  2.104933717320e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 { -2.014995434878e-05,  9.691150558517e-07, {   0,   0,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.074469472200e-07,  2.100631224892e-05, {   1,   0,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.805587766133e-06,  1.430422742091e-05, {   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.563577990225e-05,  5.456149055022e-06, {   0,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.425746511082e-05,  6.833136631432e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 { -2.079834782396e-05, -2.901010955340e-07, {   2,  -2,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.315156933313e-07,  2.084606335988e-05, {   4,  -2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.642190834814e-05,  4.649780170608e-06, {   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.186654453932e-05,  9.196775722084e-06, {   1,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.739429075006e-06, -1.731582562306e-05, {   2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.778712938699e-06,  1.626964483345e-05, {   4,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.064353754540e-06, -1.997165869639e-05, {   0,   0,   2,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.064353754540e-06, -1.997165869639e-05, {   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.824433658395e-05,  2.788867579493e-06, {   4,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.451831399523e-05, -6.513998161098e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 {  1.169327185314e-05, -9.331808456410e-06, {   4,   0,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -9.856197340930e-06, -1.115948950247e-05, {   2,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.791989000332e-05, -3.092926849976e-06, {   3,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  1.791989000332e-05,  3.092926849976e-06, {   1,   0,   1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.041271240190e-05, -1.059185150987e-05, {   2,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  4.237248187530e-07, -2.057416662750e-05, {   2,   0,  -3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.567602703490e-08, -2.097744330121e-05, {   2,   0,  -3,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.697563795708e-05,  4.014588447128e-06, {   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.877985859305e-05, -2.208622984983e-06, {   2,   0,  -2,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -1.630836110409e-05, -4.679622552004e-06, {   0,   0,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.568991782991e-05, -5.286099325703e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   5,   0,   0,   0 } },
 {  4.979329350321e-06,  1.599490647194e-05, {   1,   0,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.999465027678e-05,  9.779952990367e-07, {   3,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.397622050153e-05, -6.994104472642e-06, {   0,   0,   1,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 { -1.107847440106e-05,  9.878028453953e-06, {   1,   0,   1,   0,   0,   0, -33,  41,  -2,   0,   0,   0,   0 } },
 {  1.681127007622e-05, -4.126868895066e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -2,   0,   0,   0 } },
 { -1.193712482912e-05,  8.996166842375e-06, {   0,   0,   1,   0,   0,   0,   8, -17,   0,   0,   0,   0,   0 } },
 { -7.418749493244e-06, -1.350348746676e-05, {   2,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -7.418749493244e-06,  1.350348746676e-05, {   0,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  1.561260271863e-05, -5.308613405734e-06, {   4,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.794582054494e-08,  2.088478700650e-05, {   3,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.584124999447e-06,  1.830732741571e-05, {   2,   0,  -1,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.624870180485e-05,  4.641289288610e-06, {   4,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.013356441412e-05,  1.075123801493e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 { -6.739646279245e-06, -1.410499551359e-05, {   3,  -1,  -2,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  1.677742830367e-05, -4.059703629497e-06, {   1,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.204071517973e-06, -1.663284263904e-05, {   1,   1,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.418182546272e-06,  1.341169324718e-05, {   4,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.687019199174e-05, -3.951502002357e-06, {   1,   1,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  7.403930764609e-06,  1.341189263681e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   0 } },
 {  1.375991297182e-05,  7.055819678812e-06, {   1,  -1,   2,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.375991297182e-05, -7.055819678812e-06, {   1,  -1,  -2,   0,   2,  -2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.685820223236e-05,  3.955779490707e-06, {   1,   1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.748185929018e-06,  1.406281300437e-05, {   2,   0,   1,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.683578451165e-05, -3.972422573515e-06, {   1,  -1,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.283872834830e-07, -2.067826280745e-05, {   0,   2,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -4.239931622232e-06, -1.656549735806e-05, {   1,   1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.455397916802e-06,  1.434266700372e-05, {   2,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.777849001359e-06,  1.701848636590e-05, {   2,   0,   2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.301144377630e-05,  7.780298026128e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  3.226582282196e-06,  1.752185709861e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,  11,  -3,   0,   0 } },
 { -1.954387985039e-05,  1.203179131613e-06, {   2,   0,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -5.066591371312e-06,  1.567439417245e-05, {   2,   0,   0,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.664477716516e-07,  2.047284912546e-05, {   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 { -5.020737392000e-06, -1.570992623225e-05, {   2,   0,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 {  6.522121408124e-06,  1.417407953697e-05, {   3,   1,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.655064232152e-06, -1.703009820843e-05, {   1,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.632213969031e-06,  1.704787324892e-05, {   3,  -1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.674839918713e-05,  3.919329222298e-06, {   2,  -2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -3.370252980210e-06,  1.726870051565e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  9.939548123365e-06, -1.068512780951e-05, {   2,   0,   1,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.099746415603e-05,  9.626955304704e-06, {   2,   0,   0,   0,   0,  -3,   5,  -4,   0,   0,   0,   0,   0 } },
 {  1.644875801628e-05, -4.174407518812e-06, {   1,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.083384698065e-05,  9.780676629571e-06, {   4,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  1.670888670788e-05,  3.903962827825e-06, {   4,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -2.190956257148e-07, -2.038647719091e-05, {   2,  -2,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  7.833852732730e-06, -1.276708256573e-05, {   2,  -2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.397278862860e-07,  2.015115950083e-05, {   2,   0,   0,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.675808638621e-05, -3.819237492885e-06, {   1,  -1,   0,   0,   0,  24, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.216051879886e-05,  8.416451118077e-06, {   0,   0,   1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0 } },
 { -1.737433438942e-05, -3.199544657821e-06, {   2,   0,   0,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.574816996044e-05,  4.823416525122e-06, {   0,   0,   1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.028249833381e-05,  2.829245337673e-07, {   4,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.359802945507e-05,  6.942441663275e-06, {   0,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.575319688891e-05, -4.780518507230e-06, {   6,   0,  -3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  7.253130982658e-06, -1.326124931392e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0 } },
 { -2.390517786686e-06,  1.812141177747e-05, {   2,   0,   0,   0,  -1,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.573564604076e-05,  4.774715895159e-06, {   6,   0,  -3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.298447051910e-06, -1.420634588530e-05, {   1,   1,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  3.584909471272e-06,  1.691429059364e-05, {   2,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.138383564092e-05,  9.111799286512e-06, {   2,  -2,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  7.052746866537e-06, -1.344234409246e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   2,   0 } },
 { -7.664261209501e-06,  1.282767698187e-05, {   2,   0,  -3,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -1.972837815558e-05,  7.617941133862e-07, {   6,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.998369794552e-05,  4.922172158165e-07, {   2,   0,  -1,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0 } },
 { -7.431164028169e-06,  1.303924541390e-05, {   2,  -2,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  3.437589147185e-06,  1.703049246375e-05, {   1,  -1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 { -9.760789585896e-06, -1.069452724677e-05, {   2,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.082121291047e-05,  9.629420817773e-06, {   2,   0,   0,   0,   0, -14,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.210389023939e-06, -1.724012308674e-05, {   2,   0,  -3,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -1.960602971780e-05, -8.381573839121e-07, {   2,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.002478796083e-05, -1.041728832054e-05, {   2,   0,   0,   0,   0,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.558107278559e-05, -4.852877204905e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -7,   5,   0,   0,   0 } },
 {  1.138522812692e-05,  9.043105191963e-06, {   1,   1,  -1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -1.373044628965e-05, -6.697261690770e-06, {   0,   0,   2,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.741616116592e-05, -3.009524452832e-06, {   0,   0,   1,   0,   0,  -2,   3,   0,   3,   0,   0,   0,   0 } },
 {  1.946440918188e-06,  1.847699046982e-05, {   1,   1,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.742608167136e-05,  2.996867592926e-06, {   0,   0,   1,   0,   0,   2,  -3,   0,  -3,   0,   0,   0,   0 } },
 {  1.638659776622e-06,  1.876659631029e-05, {   1,   1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.720604275816e-05, -3.198093581892e-06, {   2,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -9.842073013723e-06, -1.055383369823e-05, {   2,   0,   0,   0,   0,  -6,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.871719877057e-06, -1.151422933620e-05, {   0,   0,   1,   0,   0,   0, -11,  16,   0,   0,   0,   0,   0 } },
 {  3.459013689756e-07,  2.002816586325e-05, {   3,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -8.575976247636e-06,  1.178773980734e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  8.028498514162e-06,  1.233279047018e-05, {   2,   0,  -1,   0,   0,   0,  11, -19,   0,   3,   0,   0,   0 } },
 {  5.264392884139e-08,  2.029545796713e-05, {   2,  -2,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.570775350846e-05, -4.635416950417e-06, {   2,   0,  -1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.647322681293e-08, -2.031937167784e-05, {   3,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.698033332112e-06, -1.362869500424e-05, {   0,   0,   3,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.864793237089e-05,  1.674802745311e-06, {   2,   0,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.679037067672e-05, -3.531422479539e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -1.963788746194e-05,  6.827940810006e-07, {   2,   0,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.619530872263e-08, -2.024675243661e-05, {   4,   0,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.733112380136e-06,  1.756948744122e-05, {   1,  -1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.576385558928e-06,  1.572606315562e-05, {   3,  -1,  -3,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.644732884681e-06,  1.765738083076e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -3,   0,   0,   0 } },
 { -8.567335124596e-09, -2.028097406205e-05, {   2,  -2,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.568710027535e-06, -1.571683493994e-05, {   1,   1,  -3,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.628161301218e-05, -3.998993157153e-06, {   0,   0,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.912710425063e-06, -1.536294849464e-05, {   4,   0,  -2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -8.709321289360e-06, -1.154261532867e-05, {   2,   0,  -2,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 { -3.489285705101e-06, -1.675813943876e-05, {   1,  -1,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.640700155518e-05,  3.839900491895e-06, {   1,   1,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.601899459498e-05, -4.214467549578e-06, {   2,   0,  -3,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.386998425236e-05,  6.344090143974e-06, {   1,   1,  -2,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.259908301910e-05,  7.614583366029e-06, {   0,   0,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  8.548890501950e-06, -1.166095019061e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  1.630870340090e-05, -3.889819320396e-06, {   3,   1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.649322870984e-06,  1.353843844186e-05, {   0,   0,   3,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -6.150373881543e-06,  1.403205687109e-05, {   3,  -2,   1,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 {  6.150373881543e-06, -1.403205687109e-05, {   3,  -2,  -1,   0,   0, -26,  32,   0,   0,   0,   0,   0,   0 } },
 {  1.191444104536e-06,  1.899008537853e-05, {   2,   0,  -1,   0,   0,   0,   7,   0,  -7,   0,   0,   0,   0 } },
 { -1.273970535849e-05,  7.433274712953e-06, {   0,   0,   3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -3.958762307536e-06, -1.620732327602e-05, {   1,   1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.635839118079e-05, -3.791344565988e-06, {   1,  -1,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.750970095091e-05, -2.634988546809e-06, {   0,   0,   3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.000488999958e-05, -1.275605773039e-07, {   0,   0,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -1.609820376150e-05, -4.034125394442e-06, {   2,   0,  -3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.645659025671e-06, -1.348545009840e-05, {   2,   0,   1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  6.444116812681e-06,  1.368499174859e-05, {   1,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.350019073339e-07, -1.998590785333e-05, {   2,   0,  -4,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.625784735763e-05, -3.857975208606e-06, {   1,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.118643818001e-08,  2.005375703212e-05, {   2,   0,  -2,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.022775107151e-05,  9.884442499839e-06, {   0,   0,   1,   0,   0,   0,  -3,   2,   2,   0,   0,   0,   0 } },
 { -6.713674195422e-08,  2.003702032361e-05, {   2,   0,  -2,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.123686413306e-05,  8.854076276854e-06, {   0,   0,   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.118888954189e-06,  1.295768617846e-05, {   5,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  7.118888954189e-06, -1.295768617846e-05, {   3,   0,  -1,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  2.824886492027e-07, -1.979005127374e-05, {   0,   2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.219366621586e-05, -7.861068700377e-06, {   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.239686357346e-05, -7.655780070981e-06, {   2,  -2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -9.186740000897e-06,  1.086239376697e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0 } },
 {  4.503499207396e-06,  1.553581084814e-05, {   1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  1.265408945725e-05, -7.383254308394e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  7.256972531559e-06, -1.277858516087e-05, {   0,   0,   1,   0,   0,   8, -14,   2,   0,  -2,   0,   0,   0 } },
 {  7.256756310881e-06,  1.277802543762e-05, {   0,   0,   1,   0,   0,  -8,  14,  -2,   0,   2,   0,   0,   0 } },
 {  7.079048828703e-06,  1.294410956409e-05, {   1,   0,   1,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 { -7.079048828703e-06, -1.294410956409e-05, {   1,   0,  -1,   0,   0,   0, -32,  37,   0,   0,   0,   0,   0 } },
 { -1.582129775890e-05, -4.201764533665e-06, {   0,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  1.944092687742e-05, -5.782388048381e-07, {   0,   0,   1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.532291143721e-05,  4.691791668174e-06, {   1,   1,   0,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 {  4.886921905111e-10,  1.999999999403e-05, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2 } },
 { -4.782202150039e-10, -1.999999999428e-05, {   0,   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.572762639790e-10, -1.999999999477e-05, {   0,   1,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -4.468042884763e-10, -1.999999999501e-05, {   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  4.293509959649e-10, -1.999999999539e-05, {   2,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  4.293509959600e-10,  1.999999999539e-05, {   0,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  3.630284843911e-10, -1.999999999671e-05, {   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2 } },
 {  3.595378258965e-10, -1.999999999677e-05, {   4,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.281218993652e-10, -1.999999999731e-05, {   4,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  3.106686068518e-10, -1.999999999759e-05, {   0,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.967059728281e-10,  1.999999999780e-05, {   6,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.932153143184e-10, -1.999999999785e-05, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -2.862339973089e-10, -1.999999999795e-05, {   2,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.792526803094e-10,  1.999999999805e-05, {   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.757620218046e-10,  1.999999999810e-05, {   4,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.513274122810e-10,  1.999999999842e-05, {   2,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  2.408554367714e-10,  1.999999999855e-05, {   6,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -2.164208272423e-10,  1.999999999883e-05, {   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 { -1.117010721228e-10, -1.999999999969e-05, {   2,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 { -8.625324401933e-06,  1.137468855350e-05, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.820404131017e-05, -1.793999616608e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   4,   0 } },
 {  1.371348352226e-05, -6.283314553612e-06, {   1,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.820210743395e-05,  1.793260293717e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -4,   0 } },
 { -3.873298675124e-06, -1.611932841994e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,  -2,   0,   0,   0 } },
 { -8.483401348396e-06,  1.150877789801e-05, {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.948993752801e-05,  4.932423601517e-07, {   2,   0,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.075193874779e-06,  1.789112852098e-05, {   2,   0,   0,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 { -6.596320521381e-06, -1.335868889837e-05, {   2,   0,  -2,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 { -9.589488565514e-06,  1.035317581205e-05, {   2,   0,  -1,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.094177644579e-06, -1.684663009521e-05, {   2,   0,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.442478037901e-06, -1.349568107192e-05, {   1,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.559564421687e-05, -4.342307373664e-06, {   1,   0,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.559268082973e-05,  4.341497903292e-06, {   1,   0,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  1.398224385830e-06,  1.852969925093e-05, {   4,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.233116867055e-05, -7.594677884792e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   3,   0,   0,   0 } },
 {  1.446052585895e-05, -5.463679676423e-06, {   2,   0,  -2,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 {  1.513736401863e-05,  4.784873418479e-06, {   1,   1,   0,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.913556734815e-06, -1.300244631681e-05, {   4,   0,  -1,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.913556734815e-06,  1.300244631681e-05, {   0,   0,   1,   0,   4,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.440580913306e-06, -1.747523504521e-05, {   2,   0,  -1,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.140510074729e-05, -8.501532047081e-06, {   0,   2,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -5.600520763403e-06,  1.429959638636e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 {  1.140568522676e-05,  8.494138397139e-06, {   0,   2,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.064649563198e-06,  1.283392202992e-05, {   4,   0,   0,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 { -7.276205008789e-07,  1.916077678535e-05, {   0,   0,   2,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.415850512293e-06, -1.447219487176e-05, {   2,   0,   2,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.899613073800e-07,  1.969429051369e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 { -7.253976579684e-06, -1.262404683469e-05, {   2,   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -1.379880655361e-05,  6.077069885028e-06, {   1,   1,  -1,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 { -4.033324050809e-06, -1.583629226896e-05, {   1,   1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.989941021638e-06, -1.287569872537e-05, {   4,   0,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 {  1.339727693197e-05, -6.462258622575e-06, {   0,   0,   1,   0,   0,   0,   2,  -6,   2,   0,   0,   0,   0 } },
 { -7.046295200059e-06, -1.280057876721e-05, {   0,   0,   2,   0, -12,   0,  12,   0,  -2,   0,   0,   0,   0 } },
 {  1.611167722951e-05, -3.726242202413e-06, {   0,   1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.089859164756e-05, -8.939028987179e-06, {   2,   0,  -3,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.809821735877e-06,  1.802015379694e-05, {   2,   0,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.149687852123e-07, -1.891367029589e-05, {   2,  -2,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.887473715701e-06, -1.593815956373e-05, {   1,   1,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  3.336062303386e-07, -1.949185929795e-05, {   2,   0,   0,   0,   0,   0,   6, -15,   0,   0,   0,   0,   0 } },
 { -4.141918845693e-09, -1.981846421266e-05, {   0,   0,   2,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.483117870774e-05, -4.969558561917e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -7,   5,   0,   0,   0 } },
 {  1.894717579134e-05,  8.418338120414e-07, {   2,   0,   1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 {  6.511817453239e-06,  1.327581477762e-05, {   2,   0,   1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.420622430865e-06,  1.734339099211e-05, {   2,   0,  -2,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.718007903274e-05,  2.573781849482e-06, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 { -1.718007903274e-05, -2.573781849482e-06, {   2,   0,  -3,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 { -5.563004100977e-08, -1.969298071413e-05, {   4,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.952373186834e-05, -2.225585471055e-07, {   1,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.588961415242e-05,  3.844875572011e-06, {   3,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.239273026640e-06, -1.449429550521e-05, {   2,  -2,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  1.576593310169e-05, -3.961861519109e-06, {   0,   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.663092428319e-05,  3.092293336396e-06, {   3,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.407485155965e-05,  5.624518622746e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,  -3,   0,   0,   0,   0 } },
 { -8.656731207991e-06, -1.104226392614e-05, {   0,   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  9.266564193407e-06,  1.043082668614e-05, {   4,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  1.584894632405e-05,  3.835058064682e-06, {   4,   0,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  8.648082022021e-06, -1.103392519656e-05, {   0,   0,   1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 { -1.666387262357e-05, -3.016110954582e-06, {   4,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.585144045162e-05, -3.825172981650e-06, {   1,   1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.178457993973e-06,  1.649205127716e-05, {   1,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.733618754463e-06, -1.593522784390e-05, {   1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.605371286608e-06, -1.705979977570e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 { -1.889320629534e-05, -7.658907903075e-07, {   2,   0,  -1,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -9.626928103718e-06,  1.002994978278e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -3,   5,   0,   0,   0 } },
 {  1.587923345900e-05,  3.769345128844e-06, {   1,  -1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.573821957014e-05, -3.899150977708e-06, {   3,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.043370863600e-06,  1.358556215609e-05, {   0,   0,   1,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  5.139844282272e-06, -1.448133142663e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  3.945368938829e-06, -1.567392075470e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -5.609782122446e-07, -1.905690366094e-05, {   2,  -2,   2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -1.713444867898e-05, -2.481965093322e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.581201212327e-06, -1.702848912855e-05, {   0,   0,   1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -2.111788633915e-07, -1.939776447189e-05, {   2,   0,  -1,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.585208307503e-06, -1.401612417729e-05, {   0,   0,   1,   0,   0,  -4,   7,   0,  -6,   0,   0,   0,   0 } },
 {  1.771169083752e-05,  1.888082560810e-06, {   2,   0,   0,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 { -7.165695896649e-06, -1.242986156885e-05, {   2,   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.769088215200e-05,  1.896570472025e-06, {   1,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.403365382672e-05, -5.548330552865e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0 } },
 {  5.576708379430e-06,  1.399795060067e-05, {   0,   0,   1,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  2.299670935012e-06,  1.727108987072e-05, {   1,  -1,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  3.512236346966e-06, -1.605676876810e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0 } },
 {  2.260114725509e-07,  1.933753567763e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 {  5.416577865504e-06,  1.413168508825e-05, {   2,   0,  -1,   0,   0,  12, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.395124417322e-08, -1.952062238462e-05, {   1,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.407956484345e-06, -1.811879902893e-05, {   2,   0,   2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.090718086808e-05,  8.612683401727e-06, {   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -1.001009484451e-05,  9.507655281282e-06, {   0,   0,   3,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 {  3.832361307581e-09, -1.950953823354e-05, {   1,  -2,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.345163580999e-05, -6.042187729909e-06, {   1,  -1,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.576468633395e-05, -3.716812726705e-06, {   3,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.353959348891e-08, -1.942348144902e-05, {   3,   0,  -3,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.584554087635e-05, -3.627708198672e-06, {   1,  -1,  -2,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.284547622186e-05, -6.614526390536e-06, {   2,   0,  -1,   0,   0,   0,  18, -22,   0,   0,   0,   0,   0 } },
 {  4.275744581544e-08, -1.941638865994e-05, {   0,   2,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.411378286722e-05, -5.343560468577e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   3,   0,   0,   0,   0 } },
 {  1.525954069842e-05, -4.196537475789e-06, {   0,   0,   2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  3.185582737769e-06,  1.626397689897e-05, {   1,  -1,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.951213964372e-06, -1.049804289842e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0 } },
 {  1.017187731395e-08, -1.943327878848e-05, {   1,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.551655492357e-05, -3.920230903511e-06, {   2,   0,  -2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 { -1.539252221443e-05, -4.040444010613e-06, {   2,   0,  -1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  3.010321074327e-07,  1.913166090285e-05, {   2,   2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.227484581871e-08,  1.941998836324e-05, {   3,   0,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.056813604983e-06,  1.836910161849e-05, {   1,   0,  -3,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.501475105744e-06,  1.490053138263e-05, {   0,   0,   1,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.813128835472e-05, -1.270580150693e-06, {   0,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.445088094047e-07, -1.925693400531e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 { -2.222875650641e-06, -1.717746932999e-05, {   2,   0,   0,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.131550483334e-09, -1.939399501337e-05, {   6,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.749356792582e-06,  1.764956118108e-05, {   0,   0,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -1.444732503364e-07,  1.924945098467e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 { -2.201106644282e-06, -1.719216106633e-05, {   1,  -1,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.345498239052e-05,  5.935031919640e-06, {   2,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.225964168115e-07, -1.886618550149e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  9.345875417616e-06, -1.003659300230e-05, {   4,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  5.285676063389e-06, -1.409473689504e-05, {   4,   0,  -1,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -3.842526291169e-06, -1.553255883526e-05, {   1,  -1,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.015429020061e-06, -1.735477223293e-05, {   2,   0,  -2,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  2.553336470558e-07,  1.911391205329e-05, {   6,   0,  -3,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.497451627419e-06,  1.186467884721e-05, {   2,  -2,   0,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.427977279184e-05, -5.074984221900e-06, {   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.150853652656e-05, -7.838408819044e-06, {   3,   0,  -1,   0,   0,   0, -31,  35,   0,   0,   0,   0,   0 } },
 { -1.150853652656e-05,  7.838408819044e-06, {   1,   0,  -1,   0,   0,   0,  31, -35,   0,   0,   0,   0,   0 } },
 { -4.512979144571e-06,  1.482987500888e-05, {   2,   0,   0,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.097266604141e-05,  8.364779654433e-06, {   6,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.643847120085e-05,  2.888835033086e-06, {   1,   1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -6.899421949600e-06,  1.242194827815e-05, {   1,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -4.958028594829e-06,  1.436036033661e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -3,  -5,   0,   0,   0 } },
 { -5.719250293707e-09,  1.931030986755e-05, {   2,   2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.558735773535e-05,  3.726769164630e-06, {   2,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  4.644650481520e-06,  1.466132957208e-05, {   2,   0,   0,   0,   0,   0, -12,  21,  -2,   0,   0,   0,   0 } },
 { -1.579894130630e-05, -3.500487367963e-06, {   3,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.293266259945e-05,  6.365347499119e-06, {   4,   0,  -2,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -6.373583132119e-06,  1.291682542143e-05, {   0,   0,   2,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -1.835568016976e-07,  1.910525823736e-05, {   0,   2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.592966067573e-05, -3.356961150049e-06, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  4.676456368786e-07, -1.881771269736e-05, {   3,   1,  -1,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -4.676456368786e-07,  1.881771269736e-05, {   3,   1,  -3,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -7.402572604257e-06,  1.187402143877e-05, {   2,   0,  -1,   0,   0,  -1,   5,  -3,   0,   0,   0,   0,   0 } },
 { -3.163863667994e-06, -1.611025561843e-05, {   1,  -1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.115883721361e-07, -1.906087004399e-05, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.305126537620e-05, -6.189548041555e-06, {   4,  -2,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -6.827628048992e-06, -1.240086886190e-05, {   0,   2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -2.405073959510e-06,  1.680964437289e-05, {   2,   0,   1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  1.819721417742e-05, -1.016799207370e-06, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  7.730316290861e-06, -1.148344255916e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -7,   5,   0,   0,   0 } },
 { -7.437121088191e-06, -1.177301524667e-05, {   2,  -2,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  3.621976306860e-09, -1.919925009995e-05, {   4,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.172799437848e-06, -1.602990537022e-05, {   1,   1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  6.913595372754e-06, -1.228640486553e-05, {   1,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.429145137743e-05,  4.908234736178e-06, {   1,  -1,   1,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 { -1.429145137743e-05, -4.908234736178e-06, {   1,  -1,  -1,   0,   0,  -3,   6,   0,   0,  -2,   0,   0,   0 } },
 {  1.234081894162e-05,  6.857655236090e-06, {   2,   0,  -1,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0 } },
 { -1.555044968586e-05,  3.647534176599e-06, {   1,  -1,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.402630816546e-05, -5.170956171468e-06, {   1,   1,  -2,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.050456317305e-05, -8.691802886280e-06, {   4,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.149599441564e-08,  1.917697769267e-05, {   1,   0,   1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.524819037533e-05,  3.935717160843e-06, {   1,   1,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.327671054527e-05, -5.902632747802e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.439097028918e-07,  1.903533131453e-05, {   6,  -2,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.311684648294e-07, -1.863667494147e-05, {   1,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.283241243323e-06,  1.686520612238e-05, {   0,   0,   2,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  2.029951505633e-07,  1.894248428205e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   2,   0,   0,   0 } },
 { -6.842799911012e-06, -1.229426617306e-05, {   4,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  2.857470593166e-06,  1.627258616442e-05, {   2,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.873065568334e-06,  1.525238440099e-05, {   0,   0,   1,   0,   3,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.906539406642e-06, -1.321848960411e-05, {   0,   0,   1,   0,   0,  -2,   6,  -5,   0,   0,   0,   0,   0 } },
 {  1.546390002170e-05,  3.659658687813e-06, {   1,  -1,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.838141466075e-05,  7.403269071762e-07, {   4,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.124151739823e-06,  1.097411314358e-05, {   2,   0,  -2,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -1.409128570178e-05,  4.991850619153e-06, {   4,   0,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  7.029523315114e-06,  1.205024643365e-05, {   0,   0,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 { -8.158328376061e-06, -1.091777631179e-05, {   4,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.938898533498e-07,  1.876789882618e-05, {   2,  -2,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.898660053725e-06,  1.513109169996e-05, {   2,   0,   2,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.309223241518e-06, -1.471512475944e-05, {   1,  -1,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.657251277539e-05,  2.446142513402e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.296644582298e-05, -6.047215935531e-06, {   2,   0,   0,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0 } },
 { -1.779297119224e-05,  1.218264832008e-06, {   3,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.361571857494e-05,  5.376727355932e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.370753420359e-05,  5.261290538190e-06, {   3,   1,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -9.799304755219e-07,  1.798033359940e-05, {   6,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.378375865912e-05, -5.146598653614e-06, {   2,   0,  -4,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.199702898822e-05, -6.932839792380e-06, {   0,   0,   2,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  1.513505855057e-05,  3.775853882001e-06, {   1,  -1,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.486633494304e-05, -4.041984715784e-06, {   1,   1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.793248719245e-07,  1.852658594045e-05, {   2,   0,  -3,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.604312338596e-06,  1.628231193751e-05, {   1,   1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.210319688583e-06, -1.767515992318e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   4,   0,   0,   0 } },
 {  2.921219793137e-07,  1.859234725291e-05, {   2,  -2,   0,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  8.463067002998e-06,  1.041875670461e-05, {   2,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.645710715615e-06,  1.123462016795e-05, {   2,   0,   0,   0,   0,   0,  18, -23,   0,   0,   0,   0,   0 } },
 { -8.696464687116e-06, -1.017804001248e-05, {   0,   0,   1,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.729646140214e-06,  1.214014289274e-05, {   1,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.808173029074e-05,  7.862011556328e-07, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  1.186896576607e-05, -6.997388022159e-06, {   2,  -2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  1.600808954361e-05, -2.846643351938e-06, {   4,   0,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.283477606978e-05,  6.018177015705e-06, {   1,   1,  -1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 {  6.140070023403e-06,  1.270340518731e-05, {   2,   0,   1,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.611975108375e-07, -1.837984207340e-05, {   4,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.106619169462e-05,  7.770162479822e-06, {   1,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.865086911634e-05,  1.826860234190e-07, {   0,   0,   1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  8.483624341366e-06,  1.034387341603e-05, {   0,   0,   2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  2.854793148304e-06,  1.597267462820e-05, {   1,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -3.505118302866e-06,  1.531991700391e-05, {   2,   0,  -3,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -3.087408925096e-06,  1.573421452128e-05, {   1,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.429231831587e-06, -1.338946922128e-05, {   1,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.483466652426e-05,  3.945682026120e-06, {   2,   0,  -1,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 {  1.364374337906e-07, -1.864378226712e-05, {   4,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.802117546572e-06,  1.196882557568e-05, {   2,   0,  -1,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.818361877470e-07, -1.837998069577e-05, {   1,  -1,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.040803904777e-06, -1.670998284135e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  2.575090072725e-07, -1.849196852165e-05, {   2,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.631028950507e-05,  2.433224828243e-06, {   2,   0,  -1,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0 } },
 {  1.754390495813e-05,  1.183798272875e-06, {   0,   0,   0,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 { -1.474827610695e-05, -3.965388033039e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 {  3.652158013468e-06,  1.504379759855e-05, {   1,  -1,   1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -3.051572788753e-07,  1.839074023822e-05, {   3,   0,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.058150235940e-06, -1.362331526727e-05, {   2,   0,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -5.725667469107e-06,  1.293895573484e-05, {   4,   0,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.199893972468e-05, -6.659577324444e-06, {   1,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.375075775973e-06,  1.226550846305e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  1.215618278192e-05, -6.481601946479e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 { -4.474051552898e-06,  1.414889230746e-05, {   0,   0,   3,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.805886959170e-05,  5.557838349703e-07, {   0,   0,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.132796303748e-05,  7.285321262256e-06, {   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.061496445721e-05, -7.997574566451e-06, {   0,   0,   3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.171804919372e-05,  6.889269724479e-06, {   1,   1,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  4.244471092185e-06, -1.435418722058e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -1.820637939652e-05, -3.842090430054e-07, {   2,   0,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0 } },
 {  1.448360617223e-05,  4.104283777568e-06, {   2,  -2,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.750114514776e-08, -1.855563352083e-05, {   4,   0,  -2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.066539131717e-08,  1.856991133402e-05, {   4,   0,  -2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.004642585589e-05,  8.532222842932e-06, {   4,   0,  -3,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  1.447425985886e-05, -4.100826681735e-06, {   2,  -2,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.030440410790e-05,  8.260364100594e-06, {   2,   0,   0,   0,   0,   0,  10, -21,   2,   0,   0,   0,   0 } },
 { -1.168162923547e-05, -6.874187920965e-06, {   0,   0,   3,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -6.112838701043e-06, -1.244026593801e-05, {   2,   0,  -2,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.022477583828e-05, -8.325473839113e-06, {   2,   0,  -1,   0,   1,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.407911494746e-08, -1.851356728045e-05, {   2,   0,   2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.783525217924e-05,  7.107537335682e-07, {   0,   0,   1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.640016465237e-08,  1.849902952289e-05, {   0,   0,   2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.814964631895e-06,  1.272821794292e-05, {   2,   0,  -2,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.307593191587e-05, -5.461041220841e-06, {   1,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.009423318570e-05, -8.432524706781e-06, {   1,  -1,   0,   0,   0,  16, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.068001698656e-06, -1.345414055899e-05, {   2,   0,   0,   0,   0,   0,   6,   0,  -8,   0,   0,   0,   0 } },
 {  3.504487215583e-06,  1.501262594041e-05, {   2,   0,   1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  3.504125736034e-06, -1.501212563596e-05, {   2,   0,   1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  1.221535626927e-06, -1.728808157145e-05, {   0,   2,  -2,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  1.053557534894e-05,  7.970519487895e-06, {   0,   0,   1,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -3.524526534376e-06,  1.497371026724e-05, {   1,   1,  -2,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.141113857631e-05,  7.085212511268e-06, {   1,   0,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -1.570089722466e-05, -2.793305136663e-06, {   0,   0,   3,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.787408896760e-05, -6.140371587282e-07, {   2,   0,  -1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -9.475826774244e-06, -9.007154155566e-06, {   0,   0,   1,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 {  1.264775226409e-08, -1.844040953039e-05, {   3,   0,  -2,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.666073899160e-05, -1.784963724130e-06, {   2,   0,  -2,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 {  1.167066650800e-05, -6.771144217293e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0 } },
 {  2.321918374488e-07, -1.819721007857e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  8.007781829713e-06, -1.041218187298e-05, {   2,   0,  -2,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.164816938959e-05,  6.767926248562e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0 } },
 { -1.212325097874e-05,  6.287879576049e-06, {   2,   0,   0,   0,   0,   0,  19, -23,   0,   0,   0,   0,   0 } },
 {  8.993347875829e-06, -9.415105852312e-06, {   4,   0,  -2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -7.550034716194e-07, -1.765075451736e-05, {   0,   0,   2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.494679208264e-05,  3.458207198569e-06, {   1,  -1,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.016193538768e-06, -1.538678611207e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 {  3.493848130561e-09, -1.839722195674e-05, {   3,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.950722767340e-06, -1.544945069139e-05, {   4,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.269536938910e-05,  5.704086713676e-06, {   2,   0,  -1,   0,   0,   1,   6, -11,   0,   0,   0,   0,   0 } },
 { -1.078405354440e-05,  7.600375846821e-06, {   2,   0,   3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  6.421176446517e-06, -1.195179510894e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.719740506537e-06, -1.065007584753e-05, {   1,  -1,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.816615675630e-05, -2.010818948177e-07, {   2,   0,  -3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.514768533797e-06, -1.284982432908e-05, {   2,   0,  -1,   0,   0,   0,  15, -22,   0,   0,   0,   0,   0 } },
 { -1.021352796014e-05, -8.150271682887e-06, {   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -1.162737090686e-05, -6.730984853553e-06, {   0,   2,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.239181775855e-07,  1.803147187490e-05, {   0,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  1.661792172398e-05,  1.730472982396e-06, {   2,  -2,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.660639981482e-05,  1.736439776653e-06, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  1.499240732988e-05,  3.347221877480e-06, {   2,  -2,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.164552159405e-06, -1.417126158663e-05, {   1,  -1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.446127170887e-05, -3.872928208608e-06, {   1,  -1,   3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.483080909655e-05, -3.501448048350e-06, {   1,   1,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.427867952866e-05,  4.053067000417e-06, {   1,   0,  -1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 {  1.418761922993e-05, -4.144032914874e-06, {   0,   0,   1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.383862213507e-07, -1.809315489776e-05, {   0,   0,   1,   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0 } },
 { -1.445815239614e-05,  3.872209347774e-06, {   1,  -1,  -3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.892434935526e-08, -1.828108720057e-05, {   0,   2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.208054244489e-05,  6.237294000525e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.201285270119e-07,  1.819753524663e-05, {   0,   2,  -2,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.692502435402e-06,  1.562357013340e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 { -6.075332128918e-06, -1.223813836173e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 {  3.177941007505e-06, -1.512451330229e-05, {   3,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  3.177941007505e-06,  1.512451330229e-05, {   1,   0,   1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.658504971422e-05, -1.714731474530e-06, {   4,   0,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.688061099659e-05,  1.414975835660e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,  -1,   0,   0 } },
 { -1.688033375636e-05, -1.414868068506e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   1,   0,   0 } },
 { -1.166386219723e-05,  6.631228033416e-06, {   0,   0,   1,   0,   0,   6, -14,   8,   0,   0,   0,  -1,   0 } },
 { -1.166386219723e-05, -6.631228033416e-06, {   0,   0,   1,   0,   0,  -6,  14,  -8,   0,   0,   0,   1,   0 } },
 {  1.792856808674e-05, -3.636900788269e-07, {   2,   0,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.542607483745e-05,  2.865922239644e-06, {   1,   1,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.480344718653e-05, -3.487719228116e-06, {   1,   1,  -1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  6.429194196720e-06,  1.185787264175e-05, {   2,   0,  -1,   0,   0,   0,   7, -10,   2,   0,   0,   0,   0 } },
 { -1.019217996799e-05, -8.090253630978e-06, {   2,   0,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -8.003181613778e-07, -1.748162094037e-05, {   4,   0,  -2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -7.383478879099e-06,  1.088200824697e-05, {   4,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -8.119534005805e-07, -1.744842654913e-05, {   2,   0,  -1,   0,   0,   5,  -6,   0,   1,   0,   0,   0,   0 } },
 {  6.077486403332e-06,  1.218006032299e-05, {   0,   2,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.714283533969e-07, -1.808063195873e-05, {   2,   0,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.585881298876e-05,  2.391185150192e-06, {   2,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.078884585117e-05,  7.440111742740e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  1.437209905470e-05,  3.841735565131e-06, {   1,  -1,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.173551756421e-06, -1.503785189968e-05, {   1,  -1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -4.364696927871e-06, -1.382414639077e-05, {   2,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.764235752120e-05,  5.443680711322e-07, {   2,   0,  -1,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  2.762209475779e-06,  1.541795489815e-05, {   2,  -2,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.470059358850e-05,  3.476321367465e-06, {   5,  -1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.681680714119e-05,  1.354302327489e-06, {   1,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 {  6.360181867579e-06, -1.181012850112e-05, {   3,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.473010779281e-05, -3.436135596157e-06, {   1,  -1,  -1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.999505775849e-06, -1.215254319295e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 { -1.583359174276e-05, -2.316895986279e-06, {   0,   0,   1,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 { -2.548542388369e-06, -1.560168317570e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.140124425641e-07,  1.803384178530e-05, {   3,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.378842546575e-05, -4.352867751319e-06, {   6,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.369727915176e-05, -4.439665965725e-06, {   2,   0,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  4.374424960379e-07, -1.769584697763e-05, {   2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.260701103353e-07, -1.760077697928e-05, {   2,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -1.441238336704e-06,  1.668271285085e-05, {   2,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.576078244815e-07,  1.735971786876e-05, {   6,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -7.808250210093e-07, -1.733058239489e-05, {   2,   0,   1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  1.491602671061e-05,  3.189465658914e-06, {   2,  -2,   0,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.491602671061e-05, -3.189465658914e-06, {   2,  -2,  -2,   0,  -1,  28, -26,   0,   0,   0,   0,   0,   0 } },
 { -6.434680153408e-09,  1.809828396681e-05, {   2,   2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.784056184628e-05,  2.607569169351e-07, {   0,   0,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.530658253793e-05, -2.792770178947e-06, {   2,   0,  -1,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.396348996843e-05, -4.131919577593e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -7,  10,   0,   0,   0 } },
 {  1.463327858025e-05,  3.460550425420e-06, {   3,  -1,  -1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.086848715776e-07, -1.797101254698e-05, {   0,   2,  -1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.478015527399e-06,  9.600703442185e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -2,   0,   0,   0 } },
 { -1.739223506387e-05, -6.864836548435e-07, {   2,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.697129187057e-05, -1.105668895397e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 { -1.749031235732e-05,  5.695388204980e-07, {   2,   0,  -1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.821779935066e-08,  1.803269712414e-05, {   3,  -2,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.209532233234e-05, -5.951794356379e-06, {   4,   0,  -2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.430357275592e-05,  3.739258530997e-06, {   3,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.416147021336e-05,  3.876441061126e-06, {   1,  -1,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.379427999587e-06,  1.165538622470e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.387809716568e-05, -4.145964334706e-06, {   4,   0,  -1,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.058254874946e-06,  1.595420748773e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.450306284212e-05, -3.509320058359e-06, {   3,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.094667648546e-06, -1.591361400626e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -1.660995776126e-06, -1.634670500683e-05, {   4,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.451369094610e-05, -3.485097085190e-06, {   3,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.214662301773e-06,  1.377953004269e-05, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.406965815840e-05, -3.917425827949e-06, {   2,  -2,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  2.248325923505e-06,  1.573381894162e-05, {   2,  -2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -2.063373341056e-07, -1.776934448702e-05, {   4,   0,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.441997374895e-05,  3.546592927243e-06, {   1,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.444631000298e-05, -3.511929385884e-06, {   0,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.404398032442e-05,  3.910314049891e-06, {   2,  -2,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  5.850878855217e-06, -1.210203232915e-05, {   4,   0,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -1.512518627770e-05, -2.823421237892e-06, {   2,   0,  -2,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  9.829120735536e-06,  8.095967207734e-06, {   0,   2,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.193315416442e-05, -5.991683343224e-06, {   1,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.332204009585e-05, -4.601410621514e-06, {   1,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.608712888780e-06,  1.531273187578e-05, {   2,   0,   0,   0,   0,   3,  -5,   0,   3,   0,   0,   0,   0 } },
 { -3.271418108342e-06, -1.464978517734e-05, {   2,  -2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.005933322391e-05,  7.851861813170e-06, {   1,  -1,   2,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 { -1.005933322391e-05, -7.851861813170e-06, {   1,  -1,   0,   0,   0,   0, -34,  41,   0,   0,   0,   0,   0 } },
 {  4.124721923501e-06, -1.377353145806e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -4.272271607653e-06,  1.361670281664e-05, {   4,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.713065035634e-05, -7.572915463254e-07, {   3,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.765441717718e-08,  1.782978323458e-05, {   2,   0,  -1,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.103466286407e-05,  6.831372957685e-06, {   1,  -1,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.103466286407e-05, -6.831372957685e-06, {   1,  -1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  7.444592622401e-06,  1.041314553010e-05, {   2,   0,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -1.720119834008e-05,  6.536547648157e-07, {   0,   0,   3,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.158510055056e-05, -6.265604272778e-06, {   3,  -1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.148850550822e-05, -6.357231787867e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 { -4.546980576777e-06,  1.329588726550e-05, {   1,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.965492914542e-08,  1.774731203339e-05, {   0,   0,   3,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.022950347720e-06,  1.180668778189e-05, {   4,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  2.946490442325e-06, -1.488115992049e-05, {   2,   0,   2,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.440496991039e-05, -3.417994048013e-06, {   2,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.172500528601e-07, -1.759938277459e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 { -2.784000253337e-06,  1.502966082827e-05, {   3,   1,  -1,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 { -2.784000253337e-06, -1.502966082827e-05, {   1,  -1,   1,   0,   0,  20, -12, -15,   0,   0,   0,   0,   0 } },
 { -7.153295680323e-06,  1.065601474225e-05, {   1,  -1,   0,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.300857846308e-06,  1.150267922540e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,  -5,   0,   0,   0 } },
 { -5.894690220273e-06, -1.189799878107e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  8.380226749264e-06,  9.396154677155e-06, {   2,   0,   0,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.670979601289e-06, -8.102420488204e-06, {   1,  -1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  9.670979601289e-06,  8.102420488204e-06, {   1,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.611297927350e-06, -1.615936718358e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  5.561928769939e-06, -1.220223971500e-05, {   2,   0,   0,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 { -7.366104646259e-06, -1.039369341848e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   0,   0,   0,   0 } },
 {  1.180516421280e-05, -5.953196073381e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -7.940419304469e-06,  9.808437748630e-06, {   1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.685234019630e-05,  8.961556991099e-07, {   4,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -7.231640938497e-06,  1.051441751339e-05, {   4,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  8.962273948962e-06, -8.782709780385e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -1.531244977833e-05,  2.429818389165e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0 } },
 { -1.096184626586e-05, -6.768313865870e-06, {   2,   0,  -2,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 { -1.529177674197e-05, -2.423978123202e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -1.369431993289e-05, -4.017321013494e-06, {   1,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.616410482441e-05,  1.544348475233e-06, {   4,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.431764932581e-05,  3.390703922135e-06, {   3,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.879841091027e-07,  1.751848189092e-05, {   4,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.221938572812e-05, -5.485960821963e-06, {   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.639547754024e-05, -1.305402392516e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.749218817668e-05, -2.035788424050e-07, {   0,   2,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  8.289951518184e-09, -1.767595549856e-05, {   2,   2,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.754702503336e-06,  1.292703348927e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,  -8,   0,   0,   0,   0 } },
 {  1.268428724761e-05,  4.996234576147e-06, {   1,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.366862961521e-05,  4.011595254197e-06, {   1,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.031332194701e-05, -7.362589225436e-06, {   4,   0,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.327419025231e-05,  4.401114485381e-06, {   2,   0,  -2,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 {  1.393582724900e-06,  1.628099554996e-05, {   1,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.537015191002e-06,  1.613283277006e-05, {   0,   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  1.529668065919e-05, -2.370761367000e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  2.978062752307e-06, -1.468003413654e-05, {   1,   1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.197026968835e-05, -5.684706236310e-06, {   0,   0,   2,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  1.196559038632e-05,  5.679740434741e-06, {   2,   0,  -3,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.957709809895e-09, -1.764060350702e-05, {   0,   0,   2,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.846280836185e-07,  1.685702946128e-05, {   6,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.513956669215e-06,  1.612698593345e-05, {   1,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  1.141788823103e-08, -1.762923576284e-05, {   2,  -2,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.324095964554e-05,  4.398408520494e-06, {   2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.098511565742e-06, -1.053935778587e-05, {   0,   0,   1,   0,   0,   6, -11,   0,   3,   0,   0,   0,   0 } },
 {  1.252710671530e-06,  1.637965904566e-05, {   3,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.338835581186e-05,  4.242977853088e-06, {   2,   0,  -1,   0,   0,  -3,   8,  -4,   0,   0,   0,   0,   0 } },
 { -4.141654628401e-06,  1.348705212123e-05, {   2,   0,  -2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.558073862254e-05,  2.043222500291e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.233264111632e-05, -5.280900738876e-06, {   3,  -3,   1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.233264111632e-05,  5.280900738876e-06, {   1,   3,  -1,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 { -4.541401545526e-06, -1.306725816579e-05, {   4,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.198100460332e-06, -1.440730613102e-05, {   2,   2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.753464871508e-06,  7.823626234404e-06, {   2,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -1.696805135655e-05,  6.090204961460e-07, {   0,   2,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  2.193769830392e-09, -1.757282407367e-05, {   1,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.533861840017e-06, -1.503803937091e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 { -4.188686026533e-06, -1.338143392778e-05, {   1,   1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.413976024776e-06,  1.215094881296e-05, {   1,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.000120978728e-07, -1.726144551395e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   1,   0,   0,   0 } },
 { -9.818266519245e-06,  7.732822739283e-06, {   0,   2,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  8.128679754936e-06, -9.421401474979e-06, {   0,   0,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.420445866593e-05,  3.336185712724e-06, {   1,  -1,   1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.363114066896e-08, -1.749595907732e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.923844286729e-07, -1.733219773638e-05, {   2,   0,   0,   0,   0,  29, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.414874736352e-05, -3.368663357339e-06, {   2,   1,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.934629244369e-06, -1.557961068898e-05, {   0,   0,   2,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.426760766792e-05,  3.237063281940e-06, {   0,   0,   1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0 } },
 {  2.045499793717e-07, -1.729926907443e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 { -8.887636635503e-06,  8.615854942380e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,  -2,   0,   0,   0 } },
 {  2.685847027185e-06,  1.480823432190e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 {  2.955770351253e-06,  1.453753842679e-05, {   1,  -1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.772612106399e-06,  1.572021266205e-05, {   2,   0,   1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.157832945585e-06,  8.323648051559e-06, {   1,  -1,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.051362840853e-05,  6.967129413494e-06, {   0,   2,  -2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.023974442265e-05, -7.235909912856e-06, {   3,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.083707318146e-06, -1.637743308966e-05, {   0,   0,   3,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -2.653289801811e-06, -1.478940305069e-05, {   6,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.608427846841e-07,  1.727205057263e-05, {   4,   0,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.553213099834e-05, -1.900144345201e-06, {   1,  -1,   1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  1.553213099834e-05,  1.900144345201e-06, {   1,  -1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  2.916218001397e-06, -1.451572569419e-05, {   1,   1,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.560853682109e-06,  7.869482624039e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,  -2,   0,   0,   0 } },
 { -5.144605118422e-06, -1.227591889740e-05, {   2,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.686973732565e-05,  5.442225648154e-07, {   0,   0,   1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  1.776241865209e-07, -1.722702524799e-05, {   4,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.900545142106e-08,  1.734398104179e-05, {   0,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  2.262493858698e-06,  1.513463800610e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  1.779640092179e-07, -1.721208498337e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -2,   0,   0,   0 } },
 { -1.607679213105e-05, -1.310158710098e-06, {   1,  -1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.624273202165e-06, -1.575870169898e-05, {   2,   0,  -1,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.044197627095e-05,  6.940387725628e-06, {   1,   0,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.458444430755e-05,  2.794805929670e-06, {   3,  -1,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.124366749959e-05, -6.120459726096e-06, {   2,   0,   0,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0 } },
 {  1.673252951143e-05,  6.227860994854e-07, {   0,   0,   1,   0,   0,  23, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.398450089199e-05,  3.370315000872e-06, {   2,   0,  -2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.912750710579e-06,  8.440470184493e-06, {   2,   0,   0,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 { -8.912750710579e-06, -8.440470184493e-06, {   2,   0,  -2,   0,   0, -17,  22, -11,   0,   0,   0,   0,   0 } },
 {  6.091076019338e-07, -1.672823244565e-05, {   2,   0,   1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -9.321626156228e-06,  8.011694384085e-06, {   2,   0,   0,   0,   1,   0, -11,  11,   0,   0,   0,   0,   0 } },
 { -8.180114294836e-06, -9.152444889387e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -9.570506954521e-06,  7.757581347185e-06, {   1,   1,   0,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.966246682341e-06, -1.335911465092e-05, {   1,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.062937749321e-05, -6.692437769570e-06, {   1,   1,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.750400605290e-06, -1.356758969538e-05, {   2,   0,   1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  1.114938622129e-05, -6.164976271586e-06, {   2,   0,  -1,   0,   0,   6, -12,   8,  -3,   0,   0,   0,   0 } },
 {  5.376344496563e-08,  1.725597268848e-05, {   2,  -2,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.417833580655e-06, -1.188967087919e-05, {   4,   0,  -2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.642631940339e-06, -7.664077233890e-06, {   2,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -3.946335266643e-06, -1.335906321917e-05, {   0,   0,   1,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0 } },
 {  1.507380742374e-05, -2.229186137737e-06, {   1,   0,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  2.785190005125e-06,  1.451164049068e-05, {   1,  -1,   1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.397790082486e-05,  3.315489685312e-06, {   1,   1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.658845968244e-05, -7.009872538906e-07, {   0,   0,   2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.387845549494e-06, -1.289601974461e-05, {   1,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.590118628931e-06, -1.468436227839e-05, {   4,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  8.139729824217e-06,  9.129020719474e-06, {   0,   0,   2,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.601590217470e-07,  1.710632717106e-05, {   2,  -2,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.506080491017e-05,  2.197373590269e-06, {   1,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  5.631295595347e-06,  1.161740096852e-05, {   2,   0,   1,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.475829037674e-06,  1.377195662840e-05, {   0,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -4.975610020721e-06,  1.226322472312e-05, {   1,   1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  9.289651297672e-06,  7.945405566396e-06, {   0,   0,   1,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   0 } },
 {  5.747551815714e-07, -1.665104567156e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  5.672791598353e-06, -1.154966240673e-05, {   0,   0,   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.372613435164e-06,  1.584924775807e-05, {   2,   0,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.421519569734e-05,  2.998496191642e-06, {   4,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.153225822427e-05,  5.675725517623e-06, {   2,   0,  -2,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 { -1.064079150295e-05,  6.562738177672e-06, {   1,   0,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -6.692418875301e-06, -1.049431352816e-05, {   2,   0,   0,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0 } },
 { -9.315655077275e-06, -7.845038689156e-06, {   4,   0,  -2,   0,   0,   0, -33,  40,   3,   0,   0,   0,   0 } },
 { -9.315655077275e-06,  7.845038689156e-06, {   0,   0,   0,   0,   0,   0,  33, -40,  -3,   0,   0,   0,   0 } },
 {  7.262413440261e-06,  9.886631144460e-06, {   2,   0,  -1,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  1.664262786034e-05,  5.046805701752e-07, {   1,   0,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.187929303319e-06, -1.095167105830e-05, {   2,   0,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.539741054181e-06, -1.159724692611e-05, {   3,  -1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  1.108881629940e-05,  6.027941698013e-06, {   1,  -1,   1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.593406726336e-06, -1.451691568219e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 { -1.649571515966e-05, -6.100222058065e-07, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  6.912345311763e-06, -1.018739905458e-05, {   2,   0,  -1,   0,   0,   0,  18, -23,   0,   0,   0,   0,   0 } },
 {  6.678290813541e-06,  1.041857399477e-05, {   0,   0,   1,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -2.889484719379e-08,  1.706690422757e-05, {   3,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.764917776609e-06,  1.432749861436e-05, {   1,  -1,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.639476263712e-06,  1.144915323179e-05, {   1,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.480841898297e-06, -1.059799186473e-05, {   2,   0,  -1,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0 } },
 {  4.354940570451e-08, -1.703477564296e-05, {   0,   2,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.463435435722e-05, -2.439418572471e-06, {   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.085912406874e-05, -6.211434650019e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,  -4,   0,   0,   0 } },
 {  1.340580121026e-05,  3.663136936326e-06, {   0,   0,   1,   0,   0,   0,   2,  -6,   3,   0,   0,   0,   0 } },
 {  5.610680372880e-06,  1.145629985761e-05, {   2,   0,   1,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.033272729168e-07, -1.666132582522e-05, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.312564906816e-05, -3.931518038568e-06, {   2,   0,  -1,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 { -1.511682131792e-05, -1.934508450972e-06, {   1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.014726869344e-07,  1.693285376038e-05, {   2,   0,   1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  3.474513563961e-06,  1.354922747686e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -8.032543339449e-07, -1.621958212384e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  9.305053069026e-06,  7.714565945362e-06, {   1,   1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -4.767471270172e-08, -1.696863908529e-05, {   4,   0,  -3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.652040150984e-05,  4.921557032924e-07, {   1,   0,  -1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.010210852194e-05,  6.908285253869e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 {  1.375874220730e-05,  3.251023018702e-06, {   0,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.378576352258e-05, -3.222328234972e-06, {   1,   1,  -1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.228897550685e-05, -4.712183402300e-06, {   3,   1,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.608427157437e-06,  1.139218532235e-05, {   4,   0,  -1,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.533167750954e-09,  1.698954663435e-05, {   1,   0,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.666274312174e-06,  7.331609541224e-06, {   0,   0,   1,   0,   0,   0,  -7,  10,  -1,   0,   0,   0,   0 } },
 { -1.464145722395e-05, -2.331723978340e-06, {   2,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.651950649873e-05,  4.499465071631e-07, {   2,   0,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.356784873071e-05,  3.380172701510e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -5,   9,   0,   0,   0 } },
 { -3.903703767864e-06,  1.304245657289e-05, {   2,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.432178558707e-05,  2.621008585329e-06, {   4,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.606318150379e-06, -1.532951787271e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -9.636535850687e-06,  7.298350160439e-06, {   4,  -2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.981147401753e-07,  1.673460132568e-05, {   0,   0,   2,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 {  5.638628343246e-06,  1.129345554888e-05, {   0,   0,   2,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.311415825686e-05, -3.807318225840e-06, {   3,   0,  -2,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.975074005647e-06, -1.093015032791e-05, {   4,   0,  -1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  1.585761618294e-05, -1.045308644461e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   3,   0,   0 } },
 { -9.521877412500e-06, -7.380259365867e-06, {   0,   2,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -4.505583842418e-09, -1.689615367899e-05, {   4,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.282413313913e-08, -1.687436265858e-05, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.366497319105e-05,  3.229502885320e-06, {   0,   1,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.910868504136e-06, -1.497595208975e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  6.542421971708e-06,  1.034042522702e-05, {   2,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.621029328436e-06, -1.425961025581e-05, {   1,  -1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.892644314228e-06, -9.981093047493e-06, {   2,   0,   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.075340153854e-05,  6.113421882706e-06, {   0,   0,   1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -1.595677837060e-05,  9.080195910087e-07, {   2,  -2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.560637656929e-06, -1.330065736016e-05, {   0,   0,   4,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -3.237920944938e-06, -1.362221279417e-05, {   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.215213230447e-07,  1.653775088545e-05, {   2,   0,   0,   0,   0,   3,  -6,   0,   1,   0,   0,   0,   0 } },
 {  1.620827579104e-05,  6.465115453347e-07, {   2,  -2,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.559669884001e-06,  1.329475004470e-05, {   0,   0,   4,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -2.804600261816e-06,  1.404786698973e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 { -5.443748299488e-06,  1.140098064154e-05, {   1,   1,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -5.223545923909e-07,  1.632065624574e-05, {   2,   0,   3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -1.232160074912e-05, -4.517456862636e-06, {   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.943953537161e-06,  1.489491274739e-05, {   3,   1,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.332030976715e-06,  1.248913544390e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,  10,   0,   0,   0 } },
 {  6.087442453099e-08, -1.675981563550e-05, {   2,   0,  -1,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.272468322542e-07, -1.659132512817e-05, {   2,   0,  -3,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  3.325119259569e-06, -1.348646776331e-05, {   1,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  2.167467894645e-06, -1.463431044435e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 {  1.441782577733e-05, -2.383896124165e-06, {   4,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  5.219692781682e-06,  1.158144138180e-05, {   2,   0,   0,   0,  -1,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  8.615349913008e-07, -1.592809769402e-05, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 {  1.363127310802e-05, -3.156044146469e-06, {   1,  -1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.055835035554e-05,  6.209724565922e-06, {   3,  -1,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  7.202973281533e-06,  9.563393682606e-06, {   1,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  6.126109117566e-08,  1.670026425756e-05, {   2,   0,  -2,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -9.592065561530e-06,  7.160323107085e-06, {   4,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.690418837859e-06,  1.405732613773e-05, {   1,   1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.991373179823e-07,  1.594748833537e-05, {   2,   0,  -1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.489417441759e-05,  1.848854142442e-06, {   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 { -1.887771974636e-07, -1.655130622288e-05, {   0,   0,   2,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.049991204104e-05,  6.236382195469e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  1.136276207458e-05, -5.365772810156e-06, {   0,   0,   3,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  9.818963019457e-07, -1.573604569298e-05, {   2,   0,  -1,   0,   0,  23, -25,   0,   0,   0,   0,   0,   0 } },
 {  2.584767325507e-06, -1.412766601570e-05, {   2,   0,  -2,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  9.214258697876e-06,  7.497222428807e-06, {   4,  -1,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.542126619319e-05, -1.271994679092e-06, {   2,   0,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.037595143019e-05, -6.317146967286e-06, {   1,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.177477954447e-05, -4.917112604548e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 { -5.396127854778e-06,  1.129463215948e-05, {   2,   0,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.353424228821e-05,  3.150343999947e-06, {   1,  -1,   0,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.254003717695e-05, -4.142621560481e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -8.424905515494e-06,  8.246014078254e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 { -1.061857244762e-08,  1.665897306310e-05, {   3,   0,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.699599085246e-06, -1.495541869830e-05, {   0,   0,   1,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 {  5.971628836816e-06, -1.066972683347e-05, {   2,  -2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -9.852802926974e-06,  6.787518999800e-06, {   2,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.938197076163e-06, -1.269627500851e-05, {   0,   2,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.641608494902e-06,  1.399270374993e-05, {   2,  -2,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  3.512296728047e-06,  1.311485140334e-05, {   2,   0,   0,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  1.483516940912e-06, -1.513821613451e-05, {   2,   0,  -1,   0,   0,   0,  16, -22,   0,   0,   0,   0,   0 } },
 {  1.429064274869e-05, -2.326741206104e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  1.040940388313e-05, -6.197240042703e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -5.960281549198e-06,  1.064486666881e-05, {   2,  -2,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.300279405426e-05, -3.586112136071e-06, {   0,   0,   0,   0,   0,  25, -28,   0,   0,   0,   0,   0,   0 } },
 { -8.211647744115e-07,  1.576718856370e-05, {   4,  -2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.370153836884e-08,  1.655482164315e-05, {   2,   0,  -1,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 {  7.923222012522e-06, -8.643961179059e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 {  1.123595224189e-05,  5.330282884222e-06, {   0,   2,   0,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0 } },
 { -1.568649213054e-05,  8.765083687184e-07, {   4,   0,  -2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -5.381321897965e-06, -1.117843968514e-05, {   1,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.100293092232e-05, -5.553297019905e-06, {   0,   0,   1,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0 } },
 {  2.449093544730e-08,  1.653073713349e-05, {   2,   0,  -4,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.335476232372e-05,  3.196408702863e-06, {   1,   1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.117423418136e-05, -5.370812084704e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 { -1.579145957707e-05, -7.420543906335e-07, {   4,   0,  -2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.490916935571e-05, -1.616764375728e-06, {   4,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  4.746954261549e-06, -1.177780697571e-05, {   2,   0,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  1.200209489105e-05, -4.506497157106e-06, {   2,   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 { -9.006604379573e-09, -1.649618720011e-05, {   0,   2,  -2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.564256880210e-05,  8.617861159104e-07, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -3.531848211246e-06,  1.296892462560e-05, {   4,  -2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.607902357936e-05, -4.079886962359e-07, {   0,   0,   2,   0,   0, -18,   8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.477629060057e-06,  1.300749495382e-05, {   2,  -1,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.219891474034e-05, -4.265966880134e-06, {   0,   0,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.540731934456e-06, -1.392351885139e-05, {   2,   0,  -2,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.332876921167e-05, -3.134671171816e-06, {   1,  -1,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.173201501141e-05, -4.725457227434e-06, {   2,   0,   1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.654673675983e-07, -1.568857595716e-05, {   2,   0,   0,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.835139234070e-06, -1.461729232900e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 } },
 { -1.105849176388e-05,  5.393391242977e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 {  1.603316739658e-05, -4.159916905269e-07, {   3,   1,  -2,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.112879779072e-11, -1.644013184801e-05, {   2,  -2,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.455280823522e-06, -1.198142052609e-05, {   2,   0,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 { -1.332596063211e-05, -3.099582598157e-06, {   2,   0,  -2,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  9.172841633440e-06,  7.249424957447e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -3,   0,   0,   0 } },
 {  1.435118377996e-05,  2.069449387086e-06, {   4,   0,   0,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  1.598145161272e-08, -1.639538118130e-05, {   1,  -2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.372725942739e-07, -1.626771070210e-05, {   0,   0,   2,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 {  7.232835024870e-07, -1.567528280473e-05, {   4,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -6.561908842586e-06,  9.831502167694e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 {  1.120762030356e-05,  5.179509743436e-06, {   3,   0,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -1.138641545485e-05, -4.996363158806e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -5.307413429162e-07,  1.583516477353e-05, {   2,   0,   1,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 {  1.126606998845e-05, -5.096971460817e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 { -1.126606998845e-05,  5.096971460817e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -1,   5,   0,   0,   0 } },
 { -1.322013525300e-05, -3.135852723499e-06, {   2,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.394585504210e-06, -1.196138018761e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,  -2,   0,   0,   0 } },
 {  2.178341590842e-08,  1.632953902521e-05, {   2,   0,  -2,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 { -8.496025870751e-06,  7.854906624510e-06, {   3,  -1,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.659950284535e-06, -1.268956701230e-05, {   1,  -1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.459421074945e-05,  1.752579024224e-06, {   4,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.022212167455e-06,  1.031461856794e-05, {   2,   0,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0 } },
 { -6.324807670093e-07, -1.569664918643e-05, {   2,   0,   0,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 {  1.041984500596e-05,  5.904457020692e-06, {   2,   0,   0,   0,   0,   0,   1,  -6,   0,   0,   0,   0,   0 } },
 {  5.743097621288e-06,  1.057869856154e-05, {   0,   0,   2,   0,   0,   0,   9, -16,   4,   5,   0,   0,   0 } },
 {  1.276771158932e-05,  3.550853183908e-06, {   1,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  6.582352294871e-06, -9.735500120175e-06, {   0,   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 { -1.459546471095e-05,  1.714521272750e-06, {   0,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.025990993452e-05, -6.048156348241e-06, {   1,   1,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  8.990522874404e-07, -1.540326838427e-05, {   2,   0,   1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 {  1.317664299291e-05,  3.117825131387e-06, {   3,  -1,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.495552226109e-05,  1.337906454752e-06, {   4,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.274584071842e-05, -3.545953520543e-06, {   1,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.513780075426e-05,  1.149915405834e-06, {   2,   0,   0,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.017121130666e-06, -1.427032437182e-05, {   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.269121815317e-05,  3.594405653424e-06, {   2,  -2,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.765431313208e-06,  1.051151308243e-05, {   4,   0,  -1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  8.155432455603e-08, -1.618535268887e-05, {   0,   2,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.303726026663e-05, -3.210375036026e-06, {   0,   0,   0,   0,   0,   9,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.628284461969e-06,  1.361825899143e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 {  2.294012040785e-06, -1.394799377969e-05, {   2,   0,   0,   0,   0,   0,   7,   0,  -8,   0,   0,   0,   0 } },
 { -4.438452281694e-06, -1.179969574834e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  5.312250620296e-06,  1.092064835519e-05, {   2,   0,   1,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.151746753487e-05, -4.715131405396e-06, {   0,   0,   1,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.754917150223e-06,  1.446157647392e-05, {   2,   0,  -3,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -2.798613889388e-06, -1.340403011525e-05, {   4,   0,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.568310138652e-07,  1.594373587541e-05, {   4,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -1.505398880699e-05, -1.145543175501e-06, {   3,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.257182904173e-05, -3.622956065677e-06, {   4,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  9.642326493249e-06, -6.549884671119e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,  -1,   0,   0,   0,   0 } },
 { -3.381267389907e-06, -1.280961867426e-05, {   2,  -1,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.174634914488e-06, -1.501342644635e-05, {   0,   2,  -1,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 {  1.034065667764e-05, -5.844584484605e-06, {   3,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.691479908354e-06,  1.348319901192e-05, {   1,   1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.674957743535e-06,  1.449613684043e-05, {   2,   0,  -1,   0,   0,   0,  -9,  19,   0,  -3,   0,   0,   0 } },
 {  6.047216258154e-06, -1.012267332691e-05, {   3,   1,  -1,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.047216258154e-06,  1.012267332691e-05, {   1,  -1,   1,   0,   0,  12,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.589531631161e-05,  2.744835501789e-07, {   2,   0,   3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.275012252843e-05, -3.418117725370e-06, {   1,  -1,   1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.260162564882e-06, -8.907860711238e-06, {   2,   0,  -2,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.554223728856e-05,  6.229169311109e-07, {   2,   0,   0,   0,   0,  28, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.017813889524e-05,  5.982061173172e-06, {   2,  -2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -7.903029098000e-06, -8.255425618948e-06, {   2,   0,   3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -7.325088684408e-06,  8.830777167564e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -6,   0,   0,   0,   0 } },
 { -1.563310344690e-05,  5.221154928069e-07, {   2,   0,  -1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.196195578867e-05,  4.189227414597e-06, {   2,   0,  -1,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.444921025282e-07, -1.530066535721e-05, {   2,   0,   1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  8.570898458669e-06, -7.573060783397e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  5.376898419078e-06,  1.076589603820e-05, {   4,  -2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.952587900024e-06, -1.318831910231e-05, {   4,  -2,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.563813196604e-08,  1.611009555201e-05, {   2,  -2,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  9.081975240988e-06, -7.049663447715e-06, {   2,   0,   0,   0,   0,   0, -10,  18,   0,   0,   0,   0,   0 } },
 {  1.258276633348e-05, -3.548276844934e-06, {   0,   0,   0,   0,   0,  17, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.531065206307e-06, -9.598899016808e-06, {   2,   0,   1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -1.252080048386e-05,  3.607897885345e-06, {   4,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.153960368314e-05, -4.587796554028e-06, {   1,   1,   0,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 { -6.248807674801e-06,  9.863182404841e-06, {   4,   0,  -1,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  1.024804676972e-05,  5.863174048625e-06, {   4,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  9.179288043249e-07, -1.519093055500e-05, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 { -1.127192754321e-05,  4.833928873210e-06, {   0,   0,   1,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.978339351680e-06,  1.012346040027e-05, {   2,   0,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.424471441484e-06,  1.365646381855e-05, {   2,   0,  -2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.204459642358e-09, -1.606485434955e-05, {   4,  -2,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.076037826882e-05, -5.306551959299e-06, {   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.445114767659e-05,  1.614985955531e-06, {   0,   0,   2,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -8.582471877012e-06, -7.482023545698e-06, {   0,   0,   4,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.462444265017e-06,  6.600045342765e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -1.577922491067e-08,  1.604663689154e-05, {   1,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.298152362819e-06,  1.075760660430e-05, {   2,   0,   2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.268842593255e-05, -3.361166490579e-06, {   3,  -1,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.081384112421e-07, -1.584085072797e-05, {   0,   0,   1,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.827690305316e-06, -1.222031993887e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -1.006002102639e-05,  5.987692839534e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.057264305725e-05, -5.469843285085e-06, {   2,   0,  -1,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 {  9.877882286218e-06,  6.161088792091e-06, {   0,   2,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.195215638480e-05, -4.086296849533e-06, {   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.382692649430e-05,  2.209203174359e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -1.297006436903e-05, -3.065518819098e-06, {   1,  -1,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.265009543207e-05,  3.384036075214e-06, {   1,   1,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.526730159079e-06, -1.450475178338e-05, {   1,   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  6.815620843896e-06, -9.214058820923e-06, {   0,   0,   1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0 } },
 { -1.030671335393e-05, -5.722789251065e-06, {   2,   0,  -1,   0,  -1,   0,   6,  -3,  -3,   0,   0,   0,   0 } },
 {  1.407812222441e-05, -1.948713682309e-06, {   2,   0,  -2,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.961061418773e-06,  1.106306801873e-05, {   4,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.094179404741e-06, -8.927648968838e-06, {   1,  -1,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.664206986992e-06,  1.035746047644e-05, {   2,  -2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -9.475469774888e-06,  6.545779932599e-06, {   1,  -2,   1,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 {  9.475469774888e-06, -6.545779932599e-06, {   1,  -2,  -1,   0,   0,   0,  29, -27,  -2,   0,   0,   0,   0 } },
 { -1.030126448078e-05,  5.719236402458e-06, {   2,   0,  -1,   0,   1,   0,  -6,   3,   3,   0,   0,   0,   0 } },
 {  1.568414829488e-05, -3.310727153918e-07, {   3,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.309410828541e-06, -6.697168492966e-06, {   0,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  1.564212248880e-05, -3.627379523151e-07, {   1,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.824022308922e-06, -6.180305778686e-06, {   1,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  5.756640498998e-06, -1.023916026797e-05, {   3,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.318536767099e-05,  2.808785336732e-06, {   1,  -1,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.814008614206e-06, -8.171851937110e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -9.986344405035e-06, -5.987454675105e-06, {   2,   0,   2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  1.819304433050e-06,  1.414765882806e-05, {   1,  -1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.004041375772e-08, -1.590347470463e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  9.575417226834e-07,  1.500581108898e-05, {   1,   1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.779980444485e-06,  1.018252378061e-05, {   0,   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -2.793704384781e-06,  1.315744618372e-05, {   1,   1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.239993174384e-05,  3.543630725979e-06, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.471296166117e-08, -1.592742865983e-05, {   6,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.679564263296e-06,  1.226117746004e-05, {   0,   0,   1,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.228785604804e-06,  1.070936266036e-05, {   2,   0,   1,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.703545727149e-07,  1.566648291411e-05, {   4,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  2.732399009012e-06, -1.318639695542e-05, {   1,   1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -5.645543841326e-06, -1.025587767582e-05, {   3,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.903990177263e-06, -1.399216814505e-05, {   0,   0,   1,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 { -6.918202350846e-06, -8.970827284086e-06, {   2,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,   0 } },
 { -8.852665764763e-06,  7.035816515283e-06, {   2,   0,  -1,   0,   0,  15, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.480189405019e-06,  8.405860098948e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -7.012938088524e-07, -1.518345825864e-05, {   2,   0,  -1,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 { -2.781031459029e-08,  1.585300113370e-05, {   1,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.530271922538e-05, -5.704410494469e-07, {   0,   0,   1,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 {  9.563745870031e-07,  1.491301342693e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,  -2,   0 } },
 {  3.165113103479e-06,  1.270025532333e-05, {   2,   0,  -2,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 {  1.352804171890e-05, -2.336477182839e-06, {   2,   0,  -2,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.881486610672e-07, -1.517410791211e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 { -6.666111567369e-06,  9.191406819269e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  2.631478125272e-09, -1.584304115421e-05, {   0,   0,   2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  9.586622064144e-06,  6.258391561232e-06, {   1,  -1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.120481159083e-05,  4.634908953611e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  9.901345878532e-06, -5.930881541780e-06, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  1.527025823711e-08, -1.581661306122e-05, {   0,   2,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.473932527809e-05,  1.092507662030e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -1.084597933258e-05,  4.983950243501e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0 } },
 {  1.278539313092e-05, -3.038840300396e-06, {   2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.309517930191e-05, -2.727394009029e-06, {   4,   0,  -2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -1.449858231365e-05, -1.319887807293e-06, {   0,   0,   1,   0,   0,   0,   6, -13,   0,   0,   0,   0,   0 } },
 { -1.238577247864e-05, -3.424175526445e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   7, -10,   0,   0,   0 } },
 { -2.269079557419e-08, -1.578614012838e-05, {   4,   0,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.123276685833e-06, -1.466809554678e-05, {   2,   0,   0,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0 } },
 { -2.104830143343e-06, -1.368509785191e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 { -1.160106075238e-05,  4.176526291728e-06, {   4,   0,  -1,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -6.562320317266e-06, -9.204415402156e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,   1,  -1,   0,   0,   0 } },
 { -1.053580773359e-05,  5.228038913121e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 {  1.281013936673e-06,  1.447683634472e-05, {   2,  -2,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -7.167295593377e-06, -8.586589891153e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,  -2,   5,   0,   0,   0 } },
 { -4.097609880479e-06, -1.165572915834e-05, {   2,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  1.371869203151e-05, -2.031101382053e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 {  2.136799821169e-08,  1.571240728305e-05, {   0,   0,   2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.586178766138e-07, -1.526774180903e-05, {   2,  -2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.183016697961e-06, -1.054198348036e-05, {   2,   0,  -2,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.908502008326e-06, -7.803482293721e-06, {   2,   0,  -3,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.288507785716e-05,  2.819045252760e-06, {   4,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.386748908994e-07,  1.484965081506e-05, {   2,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -2.399067713323e-08,  1.566017257123e-05, {   0,   2,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -8.355942006098e-06,  7.327892323971e-06, {   0,   0,   4,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -7.631021457866e-07,  1.490889923653e-05, {   2,   0,  -1,   0,   0,   4,  -6,   2,   0,   0,   0,   0,   0 } },
 { -5.158726113651e-06,  1.051193779023e-05, {   0,   0,   0,   0,   0,  23, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.543026993833e-05, -2.371485587677e-07, {   1,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  6.167220790105e-08, -1.560106927298e-05, {   2,   0,   1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -1.196900404252e-05, -3.684763158304e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  5.568852807372e-06,  1.007991431763e-05, {   1,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 { -1.515123781600e-05, -4.962124140697e-07, {   2,  -2,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.037735693037e-05,  5.267942986430e-06, {   1,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  6.300380907448e-06, -9.335082091883e-06, {   0,   0,   2,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 { -2.048019665948e-06, -1.358448731044e-05, {   0,   0,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 { -1.822566622291e-09, -1.563011630208e-05, {   4,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.389611696386e-05,  1.731421893169e-06, {   2,   0,  -1,   0,   0,   6,  -9,   2,   0,   0,   0,   0,   0 } },
 { -1.720539086683e-06, -1.390552656289e-05, {   2,   0,  -1,   0,   0,   0,  -6,  12,   0,   0,   0,   0,   0 } },
 { -4.741397028139e-08,  1.557062737728e-05, {   2,   0,  -1,   0,   0, -22,  22,   0,   0,   0,   0,   0,   0 } },
 {  6.289261521694e-06,  9.319832860638e-06, {   0,   0,   2,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -1.148585343611e-05, -4.120665527707e-06, {   0,   0,   2,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.197267788435e-06,  1.040726971374e-05, {   2,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.276648216273e-05, -2.837089282820e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -5.502079742547e-06,  1.009384021008e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 { -9.283798349511e-06,  6.309715940005e-06, {   2,   0,  -1,   0,   0,   0,   1,  -8,   3,   0,   0,   0,   0 } },
 {  6.239845391021e-06, -9.353571320274e-06, {   1,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  2.012631354221e-07,  1.538806579274e-05, {   2,   0,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 { -4.731259592399e-06,  1.085680382468e-05, {   2,   0,   0,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  1.526106380341e-05,  3.182482390422e-07, {   1,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.730203469510e-06,  8.846163212590e-06, {   1,   1,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -5.078126771531e-06,  1.048875612418e-05, {   3,  -1,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.787477064273e-08, -1.546850702139e-05, {   2,  -2,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.974840159015e-06,  1.159079905990e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0 } },
 { -9.528512336400e-06, -6.022621148760e-06, {   6,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.525976064140e-08, -1.552278722664e-05, {   1,  -2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.409990256512e-05, -1.443177777597e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -5,  10,   0,   0,   0 } },
 {  1.162604589075e-05,  3.916330339313e-06, {   2,   0,   2,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.634876570739e-06,  1.090573433945e-05, {   2,  -2,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.634876570739e-06, -1.090573433945e-05, {   2,  -2,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.521064293082e-06,  7.017930046815e-06, {   2,  -2,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.234068960215e-06,  9.304354835370e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -7.510986473011e-06,  8.018028153405e-06, {   2,   0,  -2,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 {  6.187854684338e-06, -9.330858833102e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -1.408009633414e-05,  1.438304185848e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0 } },
 { -5.700715548141e-06, -9.812739578363e-06, {   2,   0,   1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0 } },
 { -1.508353822000e-05,  4.128774287079e-07, {   0,   0,   0,   0,   0,  18,  -8, -16,   4,   5,   0,   0,   0 } },
 { -7.404830563789e-06,  8.090426459094e-06, {   2,   0,  -2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -7.702780416920e-06, -7.785480934019e-06, {   2,   0,  -1,   0,   0,   0, -12,  19,   0,  -3,   0,   0,   0 } },
 { -3.760688051689e-07,  1.511109067815e-05, {   0,   0,   1,   0,   4,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.881592516537e-06,  8.604794511144e-06, {   2,   0,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  1.189950977525e-05, -3.585836659180e-06, {   0,   0,   1,   0,   0,   0,  -6,  10,  -3,   0,   0,   0,   0 } },
 {  1.252608401472e-05, -2.954093356886e-06, {   1,   1,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -9.319887426162e-06, -6.159050671068e-06, {   4,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  9.772322330760e-06, -5.705306195971e-06, {   2,   0,  -3,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.463996544757e-05, -8.318462424909e-07, {   4,   0,  -2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.429119328662e-06, -1.403457655052e-05, {   0,   2,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.906745767744e-06, -6.555301677764e-06, {   0,   0,   1,   0,   0,   0, -10,  18,   0,   0,   0,   0,   0 } },
 { -5.101750391473e-06, -1.034860454666e-05, {   0,   0,   1,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.092814188262e-06, -1.235458442670e-05, {   2,  -2,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.453867027930e-09,  1.544000849920e-05, {   0,   0,   3,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.527006329460e-05,  1.684908079650e-07, {   0,   0,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.455410790118e-06, -5.975718931955e-06, {   2,   0,  -3,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  5.495049608087e-06,  9.933446721122e-06, {   2,  -2,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.485294477243e-06, -1.194140006000e-05, {   1,   1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.522344778834e-06, -6.903992683381e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   2,   0,   0,   0 } },
 {  1.340598100013e-05, -2.020137545480e-06, {   2,   0,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -8.283527656949e-06, -7.137408108186e-06, {   0,   0,   4,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -9.444699121900e-06,  5.974354240375e-06, {   6,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.490373114256e-05,  5.137427710535e-07, {   2,   0,  -1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.521877824607e-05, -1.953058029198e-07, {   1,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 { -1.231674772282e-05,  3.086825380482e-06, {   3,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.251865146711e-05, -2.879108378096e-06, {   3,   1,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -7.322337250498e-06, -8.070429039010e-06, {   4,  -2,   1,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.322337250498e-06,  8.070429039010e-06, {   4,  -2,  -1,   0,   0,  -9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.375231093105e-06,  1.400643435482e-05, {   4,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -7.720687892559e-06,  7.658204857699e-06, {   2,   0,   3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -5.024806699958e-06,  1.035361071921e-05, {   1,  -1,   1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  5.024806699958e-06, -1.035361071921e-05, {   1,  -1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.237687338231e-05,  2.994879753982e-06, {   0,   2,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.195369275487e-06,  9.170387698976e-06, {   2,   0,   0,   0,   0,   0,  19, -24,   0,   0,   0,   0,   0 } },
 { -2.046221121352e-08, -1.533786349630e-05, {   1,  -2,   1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.046221121352e-08,  1.533786349630e-05, {   1,  -2,  -1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.460626969952e-05,  7.376518934114e-07, {   2,  -2,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  3.212061381421e-08,  1.530710386357e-05, {   2,   0,   0,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 {  6.383641429847e-06, -8.953996471210e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,  -1,   1,   0,   0,   0 } },
 {  6.704375580959e-06, -8.629728923622e-06, {   2,   0,  -1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 { -3.301271535450e-07,  1.500332923149e-05, {   0,   0,   1,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 { -1.458749768982e-06,  1.387450545548e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   3,   0,   0 } },
 {  1.209151978459e-05,  3.240662765592e-06, {   1,   1,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.434768795350e-06, -9.892277988299e-06, {   1,   2,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  5.434768795350e-06,  9.892277988299e-06, {   1,  -2,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -3.371412654575e-06,  1.195158834670e-05, {   4,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.551783861785e-06, -1.276818962543e-05, {   3,  -1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.312408268884e-05,  2.189156411063e-06, {   4,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.210239246117e-05,  3.206707497213e-06, {   4,   0,  -1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  3.465267278198e-07,  1.496098110099e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 { -7.969575059424e-07,  1.450571316568e-05, {   3,   0,   0,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.551159795542e-06, -6.743721496298e-06, {   0,   0,   0,   0,   0,  16, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.654003589274e-07, -1.502941079365e-05, {   2,   0,   0,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.212926696056e-05,  3.164089147713e-06, {   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.239154614252e-05,  2.897038147121e-06, {   1,   1,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.932078044313e-08, -1.523628765952e-05, {   0,   0,   2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.302425197577e-06,  1.298130485977e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -1.313850573326e-05, -2.143420811024e-06, {   0,   0,   1,   0,   0,   0, -12,  23,  -3,   0,   0,   0,   0 } },
 { -2.149854620721e-07,  1.506554713803e-05, {   0,   0,   2,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -9.579554351804e-06, -5.695290366217e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -5.556730823792e-06,  9.711841365325e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 { -1.152307544482e-08,  1.525649824640e-05, {   2,   0,  -4,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.546538930597e-06,  1.371903599639e-05, {   4,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  8.365445658320e-06, -6.890379673840e-06, {   0,   2,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.207583426895e-05, -3.171719460198e-06, {   6,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.476948910545e-05, -4.734112228825e-07, {   0,   2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.193196765132e-05,  3.309344829470e-06, {   2,   0,  -2,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.760780707057e-07, -1.456444482795e-05, {   4,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.616865831827e-06,  1.261648725496e-05, {   2,  -2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.235198530214e-05, -2.876136607268e-06, {   1,  -1,  -1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.308342966620e-05,  2.139264077511e-06, {   0,   0,   1,   0,   0,   0,  12, -23,   3,   0,   0,   0,   0 } },
 {  9.380321590014e-06, -5.836363437232e-06, {   1,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -8.405202358201e-07,  1.437529503208e-05, {   3,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.763520683300e-06, -1.345148915229e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.292299329131e-05,  2.290467006424e-06, {   2,   0,   0,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 {  1.292299329131e-05, -2.290467006424e-06, {   2,   0,  -2,   0,   0, -27,  34,  -3,   0,   0,   0,   0,   0 } },
 {  1.318056408362e-07,  1.507707945024e-05, {   2,   0,  -1,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.179660614124e-05,  3.402145280683e-06, {   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.060863608669e-06, -1.013276048545e-05, {   4,  -2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.773334859006e-07, -1.491345417672e-05, {   0,   0,   1,   0,   0,  -3,   3,   0,   1,   0,   0,   0,   0 } },
 { -1.047265337101e-08, -1.516686321785e-05, {   2,  -2,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.210139394200e-07,  1.465556117889e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  1.476602383953e-05,  4.059987646620e-07, {   2,   0,   1,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.617007155810e-06, -1.255082836189e-05, {   1,  -1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.460727540624e-05,  5.533579050610e-07, {   2,   0,   2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.102133723744e-06,  1.405646293361e-05, {   0,   0,   2,   0,  -5,   4,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.140918975402e-06,  9.016908476964e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   3,   0,   0 } },
 {  1.186146956941e-05, -3.288093329991e-06, {   0,   0,   2,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.630796519218e-07, -1.448418281317e-05, {   4,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  1.138051660104e-05, -3.762237713230e-06, {   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -7.825917340857e-06,  7.306656283730e-06, {   2,   0,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.578706394022e-06,  1.254161671298e-05, {   1,  -1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.145317462235e-06,  1.296943553177e-05, {   1,  -1,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.427903841762e-06, -7.683735842770e-06, {   1,   1,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -4.575685160802e-06,  1.053204707286e-05, {   2,   0,  -1,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0 } },
 {  1.404975128967e-05, -1.057290302560e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  1.343712262207e-05, -1.668532976319e-06, {   2,   0,   1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.424510522855e-05,  8.578137687750e-07, {   0,   0,   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -8.626224564203e-06,  6.468148409478e-06, {   6,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.895045816522e-06,  1.218518782487e-05, {   1,  -1,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 {  4.363087467889e-06,  1.071506182377e-05, {   3,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  8.098093184566e-06, -6.966699574470e-06, {   3,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.016509215157e-06, -1.004425698427e-05, {   0,   2,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.785769158209e-10,  1.505788103904e-05, {   2,   0,   3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.681930367355e-07, -1.478174764545e-05, {   4,   0,  -1,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -3.112819654353e-06, -1.193555293754e-05, {   0,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.741222104531e-06, -9.302697543396e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  6.142879191430e-08, -1.497843480694e-05, {   0,   0,   2,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.174243824444e-06,  7.865318575090e-06, {   4,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  9.240141734656e-06, -5.795602927295e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.218725759845e-06, -1.181476140164e-05, {   1,  -1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.094764279433e-05, -4.077932671935e-06, {   1,  -1,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.221930291967e-05,  2.804317722761e-06, {   2,  -2,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.619124221975e-06, -1.240086257121e-05, {   3,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  1.217264492230e-05, -2.845299307983e-06, {   1,  -1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -2.626247310364e-06,  1.238607232513e-05, {   3,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.700028919622e-06, -6.311186244563e-06, {   2,   0,   0,   0,   0,  10,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.218359286716e-06, -6.791217563528e-06, {   1,  -1,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.269820636029e-05,  2.307077416003e-06, {   1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  4.563428521092e-06,  1.043941891067e-05, {   2,   0,   0,   0,   0,   0,   9, -16,   0,   0,   0,   0,   0 } },
 { -5.622875597978e-06,  9.377831580076e-06, {   2,   0,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0 } },
 { -1.319068029620e-05,  1.808541070339e-06, {   2,   0,   0,   0,   0,  19, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.039219106362e-07,  1.489361458601e-05, {   4,   0,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -1.106531210367e-05,  3.928835790009e-06, {   2,  -2,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -9.255206067798e-08,  1.489809182025e-05, {   2,   0,   3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.106189726052e-05, -3.927709810000e-06, {   2,  -2,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -2.617559720915e-08, -1.496031209008e-05, {   2,   0,  -3,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.895384672049e-06,  9.079618876162e-06, {   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.744380688372e-08, -1.489276099983e-05, {   2,  -2,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.625286786251e-06,  7.338393476805e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  3.655096316545e-06,  1.129967660138e-05, {   0,   0,   0,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 { -1.471517054979e-05, -2.389966720617e-07, {   6,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.207849023214e-05, -2.875462272582e-06, {   2,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.478618115101e-06,  7.473542486698e-06, {   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.697966069870e-06, -8.253054765406e-06, {   3,   0,   0,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 { -6.697966069870e-06,  8.253054765406e-06, {   1,   0,   0,   0,  -1,  28, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.281321449819e-05, -2.133412936758e-06, {   1,   1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -6.631767058262e-06, -8.314595505399e-06, {   6,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.208214281519e-05, -2.863654427880e-06, {   1,  -1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.032085326680e-05,  4.624318545717e-06, {   0,   1,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.334659374306e-05, -1.595130536329e-06, {   2,   0,  -1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  1.174706059242e-05,  3.194303777230e-06, {   2,   0,  -2,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.188661666338e-05,  3.054218169555e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 {  4.834723058753e-06,  1.010600838174e-05, {   1,  -1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 {  1.482741642660e-05, -1.132933073957e-07, {   2,   0,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.065607091132e-05,  4.282618417218e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   5,  -4,   0,   0,   0 } },
 {  2.185337649286e-07,  1.471980127670e-05, {   0,   2,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.547154846575e-06,  1.338438462877e-05, {   1,   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -5.865611270740e-07, -1.434447381031e-05, {   0,   0,   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -3.231489320272e-06, -1.169906852174e-05, {   2,   0,  -3,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.324906815178e-06, -1.060485721035e-05, {   2,   0,  -1,   0,   0,  -4,   7,   0,  -6,   0,   0,   0,   0 } },
 {  1.479163017315e-05,  1.332845596559e-07, {   3,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.789117128577e-08,  1.486211557807e-05, {   2,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 { -4.518946881431e-06,  1.039056574239e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.491612096149e-06,  1.041367597252e-05, {   2,  -2,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  4.304221716057e-06, -1.059818006126e-05, {   2,   0,   0,   0,   0,  -4,   7,   0,  -6,   0,   0,   0,   0 } },
 {  5.241215853017e-06, -9.654690568892e-06, {   0,   0,   2,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -4.043743163934e-06,  1.085142215705e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   6,  -5,   0,   0,   0 } },
 {  1.103950065308e-05, -3.854707692131e-06, {   1,  -1,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.444956333092e-05, -4.410610759952e-07, {   2,   0,   0,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.435759925287e-06, -5.451096790263e-06, {   1,  -1,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.294635477761e-05,  1.934770114645e-06, {   4,   0,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  2.628994601293e-07,  1.461444603651e-05, {   2,  -2,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.884640785053e-06,  1.298063412647e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -4,   0,   0,   0 } },
 {  1.732652049673e-06,  1.312771622217e-05, {   2,   0,  -1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0 } },
 { -7.692325699470e-09,  1.485217516913e-05, {   3,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.535985485956e-06,  1.031471195718e-05, {   1,   1,  -2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  3.158955448592e-06, -1.169138790063e-05, {   0,   0,   2,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 { -9.706154067341e-06, -5.141809512963e-06, {   1,  -1,  -1,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.882370138047e-06, -1.296516689200e-05, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 { -5.493691091639e-07,  1.429173676576e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -1.159142188578e-05,  3.244355748550e-06, {   2,   0,  -1,   0,   0,   0,  -2,   8,  -5,   0,   0,   0,   0 } },
 { -3.170375582805e-06, -1.166492741252e-05, {   2,   0,   0,   0,   0,   0,  -5,   0,   7,   0,   0,   0,   0 } },
 {  1.218816116554e-05,  2.640660578259e-06, {   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.331856985353e-06, -1.049497977342e-05, {   2,   0,   1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  1.198679129043e-05,  2.838116124932e-06, {   3,  -1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.086219521816e-05,  3.959072243822e-06, {   1,   1,  -1,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.935647116670e-06,  6.879178548093e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  5.771978951077e-06,  9.040631995481e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   6,  -5,   0,   0,   0 } },
 { -1.357872193912e-05, -1.226017477042e-06, {   2,  -2,   1,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 {  1.357872193912e-05,  1.226017477042e-06, {   2,  -2,  -1,   0,   0,   0,   6,  -7,  -2,   0,   0,   0,   0 } },
 { -9.865408973243e-06,  4.936961392901e-06, {   4,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  4.452836425403e-08, -1.475614703095e-05, {   0,   2,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.044024350587e-08,  1.478398948610e-05, {   2,  -2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.425556955151e-05, -5.357236586621e-07, {   2,  -2,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.277555321511e-06,  1.050362443716e-05, {   2,   0,  -1,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  7.103087589758e-06,  7.672327015177e-06, {   3,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.209325883599e-08,  1.476074984240e-05, {   2,   0,   0,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.058085410188e-05,  4.188674750186e-06, {   4,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  4.679000235687e-06, -1.008563938650e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0 } },
 {  9.282015344401e-06, -5.480174873901e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 { -1.773517683881e-06,  1.298701484975e-05, {   1,   1,  -3,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.351738926684e-05,  1.236447741128e-06, {   0,   0,   2,   0,   0, -20,  15,   8,   0,   0,   0,   0,   0 } },
 { -1.351738926684e-05, -1.236447741128e-06, {   0,   0,   0,   0,   0,  20, -15,  -8,   0,   0,   0,   0,   0 } },
 {  1.282091241713e-05, -1.927062518692e-06, {   0,   2,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.308653421701e-06, -9.433804176676e-06, {   1,   1,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -1.424077447360e-05,  4.854429139094e-07, {   6,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.462026464171e-06,  1.025844886496e-05, {   2,   0,  -1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.885941545180e-06,  4.829904441555e-06, {   0,   0,   2,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 { -1.238445923391e-05,  2.327171461833e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 { -7.535893104119e-06,  7.169844120855e-06, {   0,   0,   4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.242834670420e-06, -1.345474552746e-05, {   1,   1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -9.029769224726e-06,  5.664388745974e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 {  9.180753305516e-06,  5.508488396332e-06, {   0,   0,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 {  4.236068357497e-06,  1.044080761817e-05, {   2,   0,   0,   0,   0,   4,  -7,   0,   6,   0,   0,   0,   0 } },
 {  2.264514440828e-06,  1.241089914017e-05, {   2,   0,  -1,   0,   0,   0, -10,  16,   0,   0,   0,   0,   0 } },
 { -7.468434700881e-06,  7.197947781981e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,  -2,   0,   0,   0 } },
 {  4.481949967623e-06, -1.018327170872e-05, {   1,  -1,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.517062199871e-06,  1.213836578379e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 { -9.605361691111e-06,  5.048089335998e-06, {   2,   0,   0,   0,   0,   0,  20, -24,   0,   0,   0,   0,   0 } },
 { -1.349448333228e-08,  1.463803093048e-05, {   4,  -2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.157576441281e-05,  3.070769746254e-06, {   4,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.157576441281e-05, -3.070769746254e-06, {   4,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -1.403357352349e-05,  6.126458634721e-07, {   0,   0,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.226692043615e-05,  2.377161205966e-06, {   2,   0,  -1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.278096590586e-06, -7.364879127349e-06, {   0,   0,   1,   0,   0,   0,  -4,   7,   3,   0,   0,   0,   0 } },
 { -1.354982981566e-05,  1.084816210091e-06, {   3,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.174696175551e-05, -2.884875350169e-06, {   4,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.690678932593e-06, -1.193614252633e-05, {   1,  -1,  -3,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 {  7.848862684042e-06,  6.776203882590e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  9.292412422968e-06,  5.332174674457e-06, {   4,   0,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  1.439850794806e-07,  1.447824003323e-05, {   6,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.664077809154e-06,  1.295195023800e-05, {   4,   0,  -2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -9.038611198421e-06, -5.573009537845e-06, {   2,   0,   1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -1.396712035962e-05, -6.443936128661e-07, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,  -5,   0,   0,   0 } },
 { -9.162385546718e-07,  1.369319774050e-05, {   6,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.045660531326e-06,  1.356347056477e-05, {   2,   0,   1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.097662451995e-06,  7.498033354181e-06, {   0,   0,   0,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.528428136843e-06,  5.062767244536e-06, {   2,   0,  -2,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -8.907819035975e-06,  5.681196756096e-06, {   1,   1,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.152297613045e-05,  3.058597203089e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -6,   3,   0,   0 } },
 {  4.665434916044e-06, -9.915735383628e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -8.826129643498e-06,  5.754983843806e-06, {   2,   0,  -2,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -7.239969140722e-06,  7.334225503981e-06, {   0,   0,   1,   0,   0,   0,   4,  -7,  -3,   0,   0,   0,   0 } },
 {  1.214101029561e-05, -2.432945558682e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -2,   0,   0,   0 } },
 { -8.863603575636e-06, -5.701673422012e-06, {   4,  -2,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -4.788079279071e-06, -9.774486011466e-06, {   1,   0,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  8.140371307734e-06,  6.407465060014e-06, {   4,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -2.789846744916e-06, -1.175714316779e-05, {   0,   0,   2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  5.542003351492e-07,  1.399102385737e-05, {   1,   0,   0,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 {  1.857046393597e-07, -1.435597780879e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 { -1.176919770946e-05,  2.769383685955e-06, {   1,   1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.583403342988e-08, -1.451146676650e-05, {   2,   0,  -2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.088424672276e-05, -3.642346194017e-06, {   0,   0,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.077783787263e-06, -1.144374262111e-05, {   1,   1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  1.174285096521e-05, -2.776959434780e-06, {   1,   1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.874695610591e-06,  5.639414365669e-06, {   0,   0,   2,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  1.130817977735e-05,  3.199732133618e-06, {   2,   0,  -2,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 { -3.015435858943e-08, -1.447411679821e-05, {   2,   0,   2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.956532776376e-06, -5.543856469784e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,  -5,   0,   0,   0 } },
 {  1.390379419120e-05,  5.950057576681e-07, {   2,  -4,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.913997772660e-06, -8.582848042700e-06, {   2,   0,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.057924916933e-06, -1.343660274861e-05, {   0,   0,   0,   0,   5,  -4,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.547192324660e-06,  6.944127286007e-06, {   4,   0,  -2,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.228744639376e-05,  2.196284088654e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.907564122870e-07,  1.429257018162e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  1.001083335896e-05,  4.467350500603e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 { -7.978929769754e-06, -6.498456508852e-06, {   2,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.095335099474e-05,  3.518297949948e-06, {   4,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  1.433344286985e-05, -1.379023782903e-07, {   3,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.325065822283e-05,  1.216786297822e-06, {   4,   0,  -2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.477668518989e-08,  1.436992476424e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  3.907430157439e-07, -1.407321873882e-05, {   3,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  8.722537338023e-08,  1.436943771772e-05, {   0,   0,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0 } },
 { -8.786219054463e-06,  5.661884609560e-06, {   4,  -2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.790293940893e-06,  9.655469568986e-06, {   1,   0,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.147654904138e-05, -2.961127560620e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  7.198746572892e-06, -7.237485179968e-06, {   4,   0,  -1,   0,  -1,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.937784606219e-06,  6.485082040627e-06, {   4,  -1,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -2.521061770180e-07,  1.416560670461e-05, {   1,   0,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -6.928010057793e-06,  7.488258701772e-06, {   4,  -2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.857539898231e-07,  1.402843483611e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 {  5.039389790288e-06,  9.371676770015e-06, {   0,   0,   2,   0,   0,   0, -34,  39,   0,   0,   0,   0,   0 } },
 { -1.102388239606e-05, -3.383315814674e-06, {   2,   0,   0,   0,   0,  -3,   3,   0,  -1,   0,   0,   0,   0 } },
 { -2.027810423767e-08, -1.438546542402e-05, {   0,   2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.543238007953e-06,  6.861652316573e-06, {   4,   0,  -1,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.414971541458e-05, -2.521038877907e-07, {   3,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.102063730611e-06,  5.292981837165e-06, {   1,   1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.388212873262e-05,  5.084123777977e-07, {   2,   0,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 {  1.165013780640e-05, -2.734388617927e-06, {   1,  -1,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.512575592108e-09,  1.437264025809e-05, {   3,   0,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.052220757802e-08,  1.433957827271e-05, {   1,   0,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  8.885734627578e-06,  5.493528980801e-06, {   3,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.242487787517e-06,  1.213433628486e-05, {   0,   0,   2,   0,   0,   0,  -2,   0,   3,  -2,   0,   0,   0 } },
 { -1.188232564387e-06,  1.318740481125e-05, {   0,   0,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0 } },
 {  1.325013009582e-05,  1.123629336975e-06, {   4,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.038181738377e-06, -6.332101326832e-06, {   0,   0,   2,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  8.037709863139e-06,  6.332346140764e-06, {   0,   0,   2,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  9.727532036192e-06, -4.621292430543e-06, {   2,  -2,   2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -9.727532036192e-06,  4.621292430543e-06, {   2,  -2,  -2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.058607813928e-05, -3.759347158501e-06, {   2,   0,   2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  1.058324268162e-05,  3.757892364410e-06, {   2,   0,   2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.084502269403e-05, -3.493170854436e-06, {   0,   0,   1,   0,   0,   0,  -4,   8,  -4,   0,   0,   0,   0 } },
 { -6.979209276908e-06, -7.356930806316e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -6.387585983808e-06, -7.940780907691e-06, {   4,  -2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.085517716758e-06,  5.223801698326e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   4,   0,   0,   0 } },
 {  2.990533206716e-06, -1.131623807407e-05, {   2,   0,   0,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -5.635912856159e-07, -1.374290445338e-05, {   3,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  4.222910814200e-08,  1.425948677147e-05, {   1,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.158616947580e-05,  2.712580392901e-06, {   1,  -1,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.325934298631e-06, -9.956054774630e-06, {   2,  -2,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -6.409152900450e-06, -7.870796135312e-06, {   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.258224820629e-09, -1.427585143630e-05, {   3,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.385011017832e-05,  4.220184503060e-07, {   6,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.414101992572e-05, -1.274234981521e-07, {   5,   0,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.365744194275e-06, -1.190013128111e-05, {   1,  -1,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.081788871137e-06,  7.181742472845e-06, {   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.810208269785e-06, -5.446878113896e-06, {   3,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.996917560784e-06, -6.247907458643e-06, {   2,   0,  -1,   0,   0,   0, -10,  18,   0,   0,   0,   0,   0 } },
 { -8.397092909057e-06,  5.846531827543e-06, {   4,   0,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.128064187602e-05,  2.959435555183e-06, {   0,   2,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  5.057686312551e-06, -9.179105140857e-06, {   1,   0,   0,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  2.696040836854e-06,  1.153354458254e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -1,   0,   0,   0 } },
 {  9.286955328373e-06, -4.938865598807e-06, {   1,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  4.788541471969e-08,  1.417538498341e-05, {   2,  -2,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  5.376972964857e-06, -8.845297628191e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,  -4,   0,   0,   0 } },
 {  5.374738482200e-06,  8.840244991957e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 {  5.512067795493e-08, -1.415902985771e-05, {   0,   0,   1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.399415268513e-05, -2.176391308361e-07, {   4,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.369588836110e-05,  5.027032303516e-07, {   2,   0,  -2,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.377742669109e-05,  4.197631521339e-07, {   2,   0,  -3,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.386461979352e-09,  1.418296064307e-05, {   2,   2,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.351733571552e-09, -1.417518375325e-05, {   6,   0,  -4,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.835772601875e-06,  1.034607438444e-05, {   2,   0,  -1,   0,   0,   0,  -7,   8,  -3,   0,   0,   0,   0 } },
 { -1.018644258810e-05,  3.993422926554e-06, {   0,   0,   1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 { -9.351274212223e-06,  4.827750107832e-06, {   1,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.386583078055e-06,  5.792280583905e-06, {   1,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.324449432952e-05,  9.294559329390e-07, {   1,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  1.122046849135e-05,  2.947927108487e-06, {   1,   1,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.956861646627e-06, -1.221052088332e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  5.843756551128e-06, -8.316211454351e-06, {   0,   0,   2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -3.086612818780e-07, -1.384926816024e-05, {   1,   0,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.021056501805e-05,  3.939295960265e-06, {   0,   2,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.138741151781e-05, -2.755426602510e-06, {   1,   0,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.222011187732e-06,  5.913283710553e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.705320815930e-06, -1.242856070238e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,  -2,   0,   0 } },
 { -4.318781130481e-06,  9.810524326327e-06, {   1,  -1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.103362838255e-05,  3.094100142665e-06, {   1,  -1,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.065449358575e-05,  3.463424794263e-06, {   0,   0,   1,   0,   0,   0,   4,  -8,   4,   0,   0,   0,   0 } },
 {  5.033960227281e-06,  9.082700689323e-06, {   1,   0,   1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.251910199315e-06,  7.863785890999e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -6,   3,   0,   0,   0 } },
 {  8.201665523044e-06, -5.910430467261e-06, {   0,   0,   2,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  1.212439136686e-06,  1.289809841318e-05, {   0,   0,   2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 {  8.203285321277e-06,  5.905562946439e-06, {   0,   0,   2,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  1.107690734769e-05,  3.022381326402e-06, {   1,  -1,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.895593336933e-06, -1.220209716487e-05, {   2,   0,  -1,   0,   0,  -6,   9,   0,  -3,   0,   0,   0,   0 } },
 {  1.116793039847e-05,  2.923381248218e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -2,   0 } },
 { -9.961748314119e-06,  4.123768543610e-06, {   0,   0,   1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 { -9.266491597750e-06,  4.816077530120e-06, {   4,   0,   0,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 { -9.266491597750e-06, -4.816077530120e-06, {   0,   0,   0,   0,   0,  38, -37,   0,   0,   0,   0,   0,   0 } },
 {  9.890587827760e-06, -4.190699873194e-06, {   2,   0,  -1,   0,   0,   2,   6, -15,   8,  -6,   0,   0,   0 } },
 { -6.806728036263e-06,  7.269603285042e-06, {   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.121414038821e-06,  9.953031577808e-06, {   1,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.222943586729e-06, -9.850262326171e-06, {   2,   0,  -1,   0,   0,   0,  16, -23,   0,   0,   0,   0,   0 } },
 { -1.120672631158e-05,  2.845208305324e-06, {   2,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.135831860207e-05,  2.684763056703e-06, {   1,   1,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.769653340784e-06,  9.273269331943e-06, {   2,   0,  -1,   0,   0,  -2,   5,   1,  -8,   6,   0,   0,   0 } },
 { -9.971496432968e-06,  4.061701360534e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -6.338101754515e-06,  7.688190495562e-06, {   2,   0,   0,   0,   0,   0,  -5,   8,  -5,   5,   0,   0,   0 } },
 { -4.161390416745e-06,  9.857228269589e-06, {   0,   0,   1,   0,   0,   0, -11,  18,   0,   0,   0,   0,   0 } },
 {  9.922856837955e-06,  4.094318294231e-06, {   0,   0,   1,   0,   0,   5,  -7,  -4,   0,   0,   0,   0,   0 } },
 { -1.028088535294e-05,  3.735888925775e-06, {   0,   0,   1,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 } },
 {  1.240535637579e-05, -1.610910834636e-06, {   4,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  4.608666469381e-06, -9.396130434165e-06, {   0,   0,   2,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.185800374973e-06, -4.811019362861e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 {  1.133667561791e-05,  2.658731554488e-06, {   1,  -1,   1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -9.983403892190e-06,  4.005259541118e-06, {   1,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.356520792237e-05,  4.223633215897e-07, {   3,   0,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.595395364197e-06,  7.392012072124e-06, {   0,   0,   0,   0,   0,   1,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.766583658712e-06,  9.216042398302e-06, {   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.093517006673e-05, -3.043270383825e-06, {   1,   1,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.708334397213e-06,  8.269222322267e-06, {   0,   0,   1,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.694188765239e-06,  5.275934836915e-06, {   0,   0,   3,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  4.979975855704e-06,  8.987205758889e-06, {   0,   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.192471089735e-05, -2.042438299561e-06, {   2,   0,  -1,   0,   0,  17, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.126682800656e-05, -2.690113022921e-06, {   2,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.371407917295e-05, -2.406221429228e-07, {   5,  -2,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.371407917295e-05,  2.406221429228e-07, {   1,  -2,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -4.949704165039e-06, -9.002095553276e-06, {   3,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.377017895765e-05, -1.791093003942e-07, {   2,   0,   0,   0,   0,   0,  19, -22,   0,   0,   0,   0,   0 } },
 { -4.437489105934e-07, -1.350418194487e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -1.237441733345e-05,  1.563126661239e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 { -8.735785024216e-06, -5.195276758430e-06, {   2,   0,   1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -1.153736063977e-05, -2.393113604984e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 { -1.172737963531e-05,  2.200665151226e-06, {   0,   2,  -3,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.129205053029e-05, -2.633336346999e-06, {   1,   1,  -1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.327509358548e-05, -6.410847290128e-07, {   2,   0,  -2,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  1.124359315063e-05,  2.667536873977e-06, {   2,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.490117695801e-06,  7.413709106898e-06, {   6,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -9.509989873497e-06,  4.393632578426e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -4,   0,   0,   0 } },
 { -1.176623065961e-08,  1.389148718687e-05, {   4,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.853186727020e-06,  9.044827788981e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 {  1.366093462225e-05, -2.298446176907e-07, {   1,   0,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.218975833342e-09,  1.388290675398e-05, {   1,   0,   0,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.501278115123e-07,  1.363044886615e-05, {   4,   0,  -1,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  8.018192283135e-06,  5.861520406894e-06, {   0,   0,   1,   0,   0,   0,  10, -18,   0,   0,   0,   0,   0 } },
 { -5.686004248441e-06,  8.186746794420e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  1.071913952774e-05,  3.150457775116e-06, {   3,  -2,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.069795843686e-05, -3.169194199597e-06, {   2,   0,   1,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  7.567773821458e-06,  6.299160338687e-06, {   2,   0,   2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -7.672836812034e-06,  6.190828332995e-06, {   1,   1,   0,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.071449166120e-05, -3.148073165292e-06, {   3,  -2,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.785803095740e-06, -1.007670652256e-05, {   4,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -8.058390594832e-06,  5.801617534084e-06, {   1,   1,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.023689156270e-10,  1.385766646195e-05, {   4,   0,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  7.682982224927e-06, -6.170427416735e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  1.551798815455e-06, -1.229434724471e-05, {   3,   1,  -4,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.816275077965e-06, -6.028376076952e-06, {   1,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  9.110930871198e-06, -4.730600513064e-06, {   2,   0,  -1,   0,   0,   0,  19, -23,   0,   0,   0,   0,   0 } },
 { -8.448470223846e-08,  1.375179149297e-05, {   2,  -2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.346131747779e-05,  3.488596937824e-07, {   4,   0,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  4.599642754811e-06,  9.209387005836e-06, {   0,   0,   6,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.355094377036e-05,  2.445890494778e-07, {   2,   0,  -1,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0 } },
 {  1.039638593127e-05,  3.398425018103e-06, {   0,   0,   1,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.387507559014e-06,  1.140559141064e-05, {   1,   1,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.667588214998e-06, -8.116176436530e-06, {   2,   0,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -1.025655232047e-07, -1.367996102222e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -4,   0,   2,   0 } },
 { -1.025659421077e-07,  1.367984798640e-05, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   4,   0,  -2,   0 } },
 { -3.739439346751e-09,  1.377840728624e-05, {   6,   0,  -2,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.757484035898e-06,  6.015639371181e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0 } },
 {  4.587339254136e-06,  9.184469544566e-06, {   0,   2,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.083406178152e-06, -1.068419402013e-05, {   1,   1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.837434310472e-07, -1.347640159851e-05, {   0,   0,   1,   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0 } },
 {  1.088496621760e-06, -1.266637324393e-05, {   0,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.431483479576e-06, -1.231782796231e-05, {   2,   0,   1,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0 } },
 {  1.079768207979e-05, -2.949085960649e-06, {   0,   0,   1,   0,   0,  12, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.464078744050e-06, -9.274964661953e-06, {   1,   0,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.470489354991e-06,  9.264720929284e-06, {   0,   0,   1,   0,   0,   0,   7, -14,   0,   0,   0,   0,   0 } },
 { -4.465266523265e-06, -9.267914996435e-06, {   2,   0,  -2,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.974280402935e-06,  6.757515541060e-06, {   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.474143951071e-07, -1.358377993169e-05, {   4,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.804711974615e-06, -5.925302169652e-06, {   0,   0,   1,   0,   0,   0, -11,  15,   0,   0,   0,   0,   0 } },
 {  1.479532836209e-07,  1.358110424681e-05, {   3,   0,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.540677690386e-08,  1.370978353368e-05, {   2,   0,  -1,   0,   0,  -2,   4,  -2,   0,   0,   0,   0,   0 } },
 { -2.991357177673e-06,  1.073086784143e-05, {   0,   0,   1,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.317442678709e-05,  5.463175507313e-07, {   2,   0,   0,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0 } },
 {  9.810226941431e-06,  3.903576955730e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.038875969122e-05, -3.322955550397e-06, {   2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.330262523579e-05, -4.082016240613e-07, {   0,   0,   3,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.152281209936e-05,  2.186492715057e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0 } },
 {  2.318550604527e-08,  1.367701518961e-05, {   2,   0,  -2,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.154840997118e-06,  1.054356567296e-05, {   1,   0,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.339395030226e-06,  8.358272546193e-06, {   2,  -2,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -1.262742267002e-06,  1.242846488068e-05, {   2,   0,   0,   0,   0,   0,  14, -20,   0,   0,   0,   0,   0 } },
 {  9.940315782620e-07,  1.269199713615e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0 } },
 { -7.759722000009e-07, -1.289684420051e-05, {   0,   0,   0,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 {  2.619043829282e-07,  1.340715667420e-05, {   1,   1,   1,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.619043829282e-07, -1.340715667420e-05, {   1,   1,  -3,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.104739499725e-05, -2.620863567754e-06, {   2,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.725283892360e-07, -1.268844488121e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0 } },
 { -8.350133428071e-06,  5.309650024235e-06, {   0,   0,   1,   0,   0,   0,  -8,  12,  -2,   0,   0,   0,   0 } },
 {  1.095201505526e-05, -2.692939897009e-06, {   3,  -1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.789526164126e-06,  3.855424701161e-06, {   1,  -1,  -2,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  2.728844174020e-06,  1.091096012413e-05, {   2,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.315112546581e-06, -8.319924296200e-06, {   2,  -2,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -1.157091902941e-05, -2.055978700898e-06, {   2,   0,   0,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 { -1.098678064810e-05,  2.638202786059e-06, {   2,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.776659264430e-07,  1.344452966026e-05, {   2,   0,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.321698660358e-05,  4.025647149280e-07, {   6,   0,  -3,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.748586750845e-06, -9.868851080350e-06, {   1,  -1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  3.748586750845e-06,  9.868851080350e-06, {   1,  -1,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -9.449514732266e-06, -4.158676153360e-06, {   0,   0,   1,   0,   0,  16, -26,   0,   0,   0,   0,   0,   0 } },
 { -9.449514732266e-06,  4.158676153360e-06, {   0,   0,   1,   0,   0, -16,  26,   0,   0,   0,   0,   0,   0 } },
 { -1.269362403958e-06,  1.233816802562e-05, {   2,   0,  -2,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  4.816510866030e-06, -8.786545844812e-06, {   2,   0,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0 } },
 { -1.320377346729e-05,  3.951170602142e-07, {   2,   0,  -1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 { -8.463810594305e-06, -5.134999684108e-06, {   0,   0,   1,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 {  7.633172190452e-06, -5.965469653029e-06, {   4,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  3.410758967931e-06, -1.018538525476e-05, {   1,   0,  -2,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.327528375856e-06,  8.267993194939e-06, {   2,   0,  -1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0 } },
 { -1.099870933648e-05,  2.596539566232e-06, {   0,   1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.306166536843e-06, -7.287463852773e-06, {   4,   0,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  5.062797532339e-06,  8.529492404539e-06, {   2,   0,  -1,   0,   0,   0, -11,  17,   0,   0,   0,   0,   0 } },
 { -5.580834583563e-08, -1.353375496563e-05, {   2,   0,  -2,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -4.190718195620e-06, -9.396394617019e-06, {   2,   0,  -1,   0,   0,  -2,   6,  -5,   0,   0,   0,   0,   0 } },
 {  1.093872310869e-05,  2.646896011707e-06, {   0,   2,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.492412187807e-06,  6.087676683937e-06, {   1,  -1,   0,   0,   0, -14,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.885783261810e-06, -1.168846847477e-05, {   4,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.206762265414e-06,  1.136419084114e-05, {   1,   1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.191410151565e-06, -6.371740082874e-06, {   1,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.489845256624e-06,  4.070147183807e-06, {   4,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  1.288879465527e-05,  6.645209824878e-07, {   3,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.081602591997e-06, -5.461881491539e-06, {   2,   0,   3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.258801502437e-06,  1.227207956244e-05, {   2,   0,   0,   0,   0,   0,  13, -19,   0,   0,   0,   0,   0 } },
 { -1.089047529698e-05, -2.635213293116e-06, {   3,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.729672663251e-08, -1.345731352338e-05, {   2,   0,   2,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -1.307461552982e-05, -4.491033781244e-07, {   2,   0,  -1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -4.872956136905e-06, -8.650746256734e-06, {   4,   0,  -2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  4.797970342888e-06,  8.708019166012e-06, {   4,   0,  -4,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.067433084428e-05, -2.830033528953e-06, {   4,   0,  -3,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  6.761288556278e-06, -6.740206763782e-06, {   4,   0,  -3,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -1.302156041117e-05,  4.797374240056e-07, {   2,   0,   1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -7.153927193736e-06, -6.346123051206e-06, {   2,  -2,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.318317303915e-06, -5.181468454299e-06, {   2,  -2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -7.782360609878e-06,  5.713719462877e-06, {   1,   1,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.773003248065e-06,  5.720961292420e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -5.411670320393e-06,  8.080891834548e-06, {   2,   0,  -3,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  8.432714053782e-06,  5.053997260917e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0 } },
 { -1.335830347670e-05,  1.269462757529e-07, {   0,   0,   0,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.090342067499e-05,  2.580541276265e-06, {   3,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.286012097868e-06,  1.118617557351e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  6.280305474525e-06, -7.190685337791e-06, {   2,   0,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0 } },
 {  1.144371267600e-07, -1.335646548806e-05, {   2,   0,   2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.000826616494e-06,  8.468760112785e-06, {   2,   0,  -1,   0,   0,   0,   5,   0,  -8,   0,   0,   0,   0 } },
 {  1.154405510842e-05, -1.924369173801e-06, {   2,   0,   1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.947455430042e-06,  4.516700553504e-06, {   2,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -8.030071819779e-06,  5.429399180784e-06, {   2,   0,   3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.660690422228e-06,  1.179821421908e-05, {   1,   1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.450542038898e-06,  1.200515278773e-05, {   2,   0,  -3,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  8.400810088185e-06,  5.049535664455e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -4.440532106831e-06,  9.008080899142e-06, {   2,   0,   0,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  1.093357500683e-05,  2.510094205357e-06, {   4,   0,  -1,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  2.229112169614e-06, -1.121375274432e-05, {   4,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -6.787867091442e-06, -6.650221816786e-06, {   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.774046811981e-06,  3.661579554480e-06, {   1,  -1,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.153537392616e-05,  1.879883035366e-06, {   0,   2,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 { -1.123281115919e-05,  2.177906627042e-06, {   0,   0,   1,   0,   0,   0,  -5,  10,  -4,   0,   0,   0,   0 } },
 {  3.252330478343e-09, -1.340538491603e-05, {   6,   0,  -3,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.382209645531e-06, -4.021163184859e-06, {   4,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  4.223962047665e-06,  9.179164727712e-06, {   3,   1,  -2,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.709736692565e-06, -8.692593355819e-06, {   0,   0,   1,   0,   0,   0,  -7,  14,   0,   0,   0,   0,   0 } },
 { -1.729673695791e-06,  1.167111948440e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   3,   0,   0,   0 } },
 {  4.750432780646e-06,  8.646660676912e-06, {   1,   0,   3,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -4.750432780646e-06, -8.646660676912e-06, {   1,   0,  -3,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.082776627105e-05,  2.567009755268e-06, {   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.346779467292e-06, -1.204777582927e-05, {   0,   0,   1,   0,   0,   0,  12, -24,   5,  10,   0,   0,   0 } },
 { -1.346779467292e-06,  1.204777582927e-05, {   0,   0,   1,   0,   0,   0, -12,  24,  -5, -10,   0,   0,   0 } },
 { -1.121889976458e-05, -2.172697661592e-06, {   0,   0,   1,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 { -1.057700633015e-05,  2.809608385771e-06, {   1,  -3,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.033206507231e-06,  1.234830047219e-05, {   2,   0,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -1.945581233854e-07, -1.318546002616e-05, {   3,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.080565654375e-05,  2.571054137576e-06, {   1,  -1,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -7.360478174318e-06,  6.010282697317e-06, {   3,   1,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.303624034517e-06, -1.106663279850e-05, {   1,   1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.323851618222e-05,  1.302947880726e-07, {   4,   0,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.560188190172e-08, -1.335025940975e-05, {   0,   0,   0,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.261343639087e-05, -7.509440494504e-07, {   3,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.778100085129e-07,  1.318486830522e-05, {   1,   0,   3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.460708594356e-06, -5.898098684462e-06, {   2,   0,  -2,   0,   0,  -3,   6,  -4,   0,   0,   0,   0,   0 } },
 { -5.343356710844e-06, -8.009581477589e-06, {   2,   0,  -3,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  6.808368616654e-06,  6.543827356012e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -3,   0,   0,   0 } },
 { -4.061696473419e-06,  9.280523631653e-06, {   2,  -2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  2.345793513661e-07, -1.310763011897e-05, {   4,   0,  -2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  5.554574417844e-06,  7.784863355855e-06, {   2,   0,  -1,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0 } },
 {  7.344469780984e-06,  5.991748895930e-06, {   2,  -1,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.248671783634e-05,  8.261856738607e-07, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 } },
 {  7.264569590652e-06, -6.045668473216e-06, {   1,   1,   0,   0,   0, -17,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.262115436997e-05, -6.865548078054e-07, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -3.185423311210e-06, -1.011224580167e-05, {   5,  -1,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.392450448720e-06,  8.899783134228e-06, {   1,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.036756655894e-05, -2.913373473155e-06, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -9.933748346329e-07,  1.228648050340e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  1.221247121160e-05, -1.065621489072e-06, {   2,   0,  -2,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.866895656333e-06,  6.403668141356e-06, {   2,   0,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0 } },
 {  2.353967275666e-06, -1.091197385242e-05, {   1,   1,  -1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.178856628066e-07, -1.314437707975e-05, {   2,  -2,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.071923684076e-05,  2.542726104162e-06, {   2,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.461872713521e-06, -5.797459394814e-06, {   2,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  1.038082502321e-05, -2.871072658964e-06, {   0,   0,   1,   0,   0,   0,  -2,   6,  -3,   0,   0,   0,   0 } },
 { -2.047104604044e-06, -1.120475775424e-05, {   0,   1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.945303647795e-06,  7.303113004051e-06, {   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.814169038577e-06, -1.043387082922e-05, {   1,  -1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.358400591196e-07, -1.310518742645e-05, {   2,   0,   0,   0,   0,  30, -30,   0,   0,   0,   0,   0,   0 } },
 { -9.348090182663e-06,  3.889542255753e-06, {   2,  -2,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.273718277296e-06,  1.095337050824e-05, {   1,  -1,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -1.152264709307e-05, -1.703672207427e-06, {   6,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -7.682113017460e-06,  5.543980571357e-06, {   0,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.084752832004e-06, -9.140312575127e-06, {   2,   0,   0,   0,   0,  -2,   6,  -5,   0,   0,   0,   0,   0 } },
 { -2.783920255562e-06, -1.043935475285e-05, {   1,  -1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.877733833929e-08,  1.317169239666e-05, {   2,   0,  -2,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.476877042816e-06, -1.074116712343e-05, {   1,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.902068630890e-07,  1.292502462460e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0 } },
 { -4.745894027132e-06,  8.468748511060e-06, {   4,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.044189694037e-05,  2.770716792947e-06, {   2,   0,   3,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  1.157848294956e-06,  1.205376084029e-05, {   0,   0,   1,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 {  1.274389333837e-05,  4.656447280396e-07, {   4,   0,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.030725339448e-06, -1.217837773934e-05, {   0,   0,   1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -1.206367213574e-05,  1.144556245477e-06, {   1,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.105177702167e-06,  1.010110651814e-05, {   2,   0,   0,   0,   0,   0,   3,  -8,   5,  -5,   0,   0,   0 } },
 { -1.072520548410e-05,  2.477389259847e-06, {   1,   1,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.298161097747e-05,  2.139344278090e-07, {   2,   0,  -2,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.426242855300e-06,  5.766471448151e-06, {   2,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.205315651495e-05, -1.137934875911e-06, {   0,   1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.101746052355e-05, -2.173070372192e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   1,   0 } },
 {  2.735721461120e-08, -1.315377924357e-05, {   0,   2,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.099063499732e-05,  2.185385704146e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 { -1.066342726160e-05, -2.506904269987e-06, {   1,   1,  -1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.173213516401e-08,  1.313792243996e-05, {   2,   0,  -2,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -6.791161230468e-06,  6.378453594785e-06, {   0,   0,   0,   0,   0,   6,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.634567883112e-06, -1.053413706638e-05, {   5,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  9.065936453999e-07,  1.225699098129e-05, {   2,  -2,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  3.343025995408e-07, -1.282378732976e-05, {   4,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.212352360628e-06,  4.939145572725e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -4.209996107158e-06, -8.940162275429e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   8,   0,   0,   0 } },
 { -1.001190057114e-05,  3.136605256070e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 { -1.188727092452e-05,  1.260575462240e-06, {   0,   0,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.062405176576e-05, -2.520206627127e-06, {   2,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.199599826391e-06,  1.194403831307e-05, {   2,   0,   0,   0,   0,   0,  15, -21,   0,   0,   0,   0,   0 } },
 { -1.301688484155e-05,  1.209122451713e-07, {   1,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.063115471520e-05, -2.504018071594e-06, {   1,   1,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.979145432287e-06,  8.149208249910e-06, {   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.312814804059e-06, -1.181331138065e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0 } },
 {  1.148332952394e-06, -1.197591969120e-05, {   2,   0,  -1,   0,   0,   0,  17, -23,   0,   0,   0,   0,   0 } },
 {  1.029130702426e-05,  2.823028636764e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -5,   0,   0,   0 } },
 {  7.818458358968e-06, -5.289896526314e-06, {   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  4.270826184269e-07, -1.267435552537e-05, {   2,   0,   1,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -4.522618010974e-06,  8.578313222330e-06, {   2,   0,   2,   0,   0, -26,  28,   0,   0,   0,   0,   0,   0 } },
 {  1.020999838707e-05,  2.875728075300e-06, {   2,   0,  -3,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  6.933064944386e-06,  6.150173614067e-06, {   3,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.783072356893e-06, -1.030003455582e-05, {   6,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.053380709073e-05, -2.548905900052e-06, {   0,   2,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.748517039090e-06,  4.326916470348e-06, {   4,   0,  -2,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -7.478239178588e-06,  5.591818138667e-06, {   1,   1,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.667434578736e-09,  1.306280540703e-05, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   1,   0,   0,   0 } },
 {  5.667434578736e-09,  1.306280540703e-05, {   2,   0,  -3,   0,   0,  -2,   5,   0,   0,   1,   0,   0,   0 } },
 {  2.248534490033e-06,  1.081755682543e-05, {   1,  -1,   0,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.040513246539e-05, -2.660586700321e-06, {   2,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.860769814891e-06, -1.019792772196e-05, {   2,   0,   0,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 { -2.114811571267e-06,  1.094330646756e-05, {   2,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.006517984808e-05, -2.991063513559e-06, {   0,   0,   2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 {  6.423503028264e-06,  6.631056250465e-06, {   2,   0,   0,   0,   0,   0,   1,  -4,   3,   0,   0,   0,   0 } },
 { -5.673655523854e-06,  7.380494601448e-06, {   0,   0,   1,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 { -9.013550909854e-06,  4.036143826471e-06, {   1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.152885378709e-06,  5.891631394330e-06, {   6,   0,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.323973889262e-06, -7.719687395007e-06, {   0,   0,   2,   0,   0,   2, -14,  15,   0,   0,   0,   0,   0 } },
 { -9.781326460354e-06, -3.260814740595e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -8.391137959255e-06,  4.646718778427e-06, {   1,   1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.901799272680e-06, -7.133904684921e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   0,   0,   0,   0 } },
 { -1.887604183162e-06, -1.114550408414e-05, {   1,  -1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  9.207361590419e-06,  3.823720232359e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  5.536304049940e-06, -7.490987109735e-06, {   1,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  5.337590101488e-06, -7.682539904255e-06, {   0,   0,   1,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  5.242343210039e-06,  7.776074847040e-06, {   2,   0,   0,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0 } },
 {  1.550506984590e-06,  1.146631757115e-05, {   0,   0,   1,   0,   0,   0,  -5,   0,   6,   0,   0,   0,   0 } },
 { -1.978686926853e-06,  1.103495268703e-05, {   3,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -4.184716330676e-06,  8.827097871364e-06, {   2,   0,  -2,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -1.262185730428e-05,  3.884909233813e-07, {   0,   2,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.368533815788e-07, -1.277001159296e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 {  1.037632581577e-05,  2.626164721315e-06, {   1,  -1,  -1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  7.059922750157e-06,  5.937502437361e-06, {   1,  -1,  -3,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  7.431615932163e-06,  5.560843774527e-06, {   2,   0,   0,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 { -7.016577871371e-07,  1.228839654226e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,  -6,   2,   0,   0,   0 } },
 { -5.035853460598e-06,  7.947345031538e-06, {   0,   0,   0,   0,   5,   0,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.266691629981e-06,  6.714678403500e-06, {   2,   0,  -1,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0 } },
 {  3.539001019771e-07, -1.262527876308e-05, {   2,   0,   0,   0,   0,  -4,   6,   0,  -2,   0,   0,   0,   0 } },
 { -5.201274638608e-06,  7.773558307736e-06, {   2,   0,   0,   0,   0,   0,  20, -25,   0,   0,   0,   0,   0 } },
 {  1.219570437435e-05, -7.775421295068e-07, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -7,   0,   0,   0 } },
 {  7.528344819088e-06, -5.441274239127e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -3.787280335971e-06, -9.181277864687e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -3,   0,   0,   0 } },
 { -2.887722190705e-06,  1.008006818758e-05, {   2,   0,  -1,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.211345714360e-08,  1.295549258127e-05, {   4,   0,  -2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.346890882777e-06,  1.162042010661e-05, {   4,   0,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -4.104258342168e-06, -8.862267285159e-06, {   1,  -1,   0,   0,   0, -13,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.013999436524e-05,  2.823284047709e-06, {   1,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  1.013791518024e-05, -2.822732534847e-06, {   1,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  4.176009412884e-06,  8.783158169099e-06, {   0,   0,   1,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0 } },
 {  4.315477000915e-06, -8.640097055043e-06, {   0,   0,   4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  9.123182567454e-06,  3.830625305301e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 {  3.855650984415e-06, -9.094947814435e-06, {   0,   0,   1,   0,  -5,  14,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.554622642638e-06,  3.392478508350e-06, {   2,  -2,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -9.553530093258e-06, -3.392094130141e-06, {   2,  -2,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.097346873365e-05,  1.972110203651e-06, {   3,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.048390543023e-05,  2.458184828628e-06, {   4,  -2,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.940944091120e-06,  1.000005420191e-05, {   1,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  8.570585713442e-06, -4.369112581972e-06, {   0,   0,   1,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0 } },
 {  5.014055970462e-06,  7.916607467314e-06, {   0,   0,   2,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.375532034876e-06, -8.554914474105e-06, {   2,   0,  -2,   0,   0, -12,  10,   0,   0,   0,   0,   0,   0 } },
 { -8.088407269596e-06,  4.841692716813e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0 } },
 {  1.282974668890e-05,  9.301788480030e-08, {   1,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.170397688965e-06, -1.074924407103e-05, {   1,   1,  -1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.845123427365e-06,  9.070842222679e-06, {   0,   0,   1,   0,   5, -14,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.244671668622e-05, -4.674346301584e-07, {   0,   0,   2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.827833685022e-06, -1.108174927564e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0 } },
 { -4.163603276919e-06,  8.743535243874e-06, {   2,   0,   0,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.242411928056e-05, -4.828935273569e-07, {   2,  -2,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.194267144455e-06,  1.171257662347e-05, {   4,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.199902146089e-05,  9.036644419889e-07, {   2,   0,  -1,   0,   0,   6,  -4,  -8,   3,   0,   0,   0,   0 } },
 { -1.099833820051e-06,  1.180165775299e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.256607620184e-05, -3.336553454767e-07, {   2,   0,  -1,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.745211792563e-07,  1.261973949759e-05, {   4,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  9.053181479125e-06,  3.839672772046e-06, {   4,   0,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.328342019729e-06,  1.056416701173e-05, {   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.699281149291e-06,  6.192719128045e-06, {   2,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.120231455280e-05, -1.689189161033e-06, {   6,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.236059745622e-05,  5.286383552581e-07, {   2,   0,  -1,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.456872094113e-06, -3.430447043665e-06, {   4,   0,  -2,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.028175145904e-06,  5.855615676919e-06, {   2,   0,   0,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  4.611136561740e-08,  1.282557829920e-05, {   2,  -2,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.006637344438e-05, -2.802715374682e-06, {   4,   0,   1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -1.006614437732e-05,  2.802744903983e-06, {   4,   0,   1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -1.274248993165e-05,  1.252679239179e-07, {   0,   0,   3,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.519738826145e-06,  9.346867223850e-06, {   0,   0,   1,   0,   1,   0, -14,  19,  -3,   0,   0,   0,   0 } },
 { -3.519738826145e-06, -9.346867223850e-06, {   0,   0,   1,   0,  -1,   0,  14, -19,   3,   0,   0,   0,   0 } },
 { -1.278020787218e-05,  8.606563943014e-08, {   2,   0,   3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.702584157351e-06,  9.160583283891e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  3.155562102435e-06, -9.702461958064e-06, {   0,   0,   1,   0,   0,   0,   8, -16,   2,  10,   0,   0,   0 } },
 {  3.155562102435e-06,  9.702461958064e-06, {   0,   0,   1,   0,   0,   0,  -8,  16,  -2, -10,   0,   0,   0 } },
 { -1.813166596191e-06,  1.104382514059e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0 } },
 { -9.200321368514e-06, -3.653991888908e-06, {   0,   0,   3,   0,   0,   2, -13,  15,   0,   0,   0,   0,   0 } },
 {  1.000447943867e-05, -2.849403005586e-06, {   2,  -2,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  1.949176418818e-06,  1.089849959299e-05, {   2,  -2,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.065489889274e-07, -1.224092757580e-05, {   2,  -2,   1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  6.835024392467e-06,  6.011116236304e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 { -1.254864824673e-06, -1.159064695468e-05, {   0,   0,   1,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 {  5.162480214999e-06, -7.674728617271e-06, {   3,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.484251905499e-06,  7.352651102382e-06, {   0,   0,   2,   0,   0, -18,  16,   0,  -2,   5,   0,   0,   0 } },
 {  5.484251905499e-06, -7.352651102382e-06, {   0,   0,   0,   0,   0,  18, -16,   0,   2,  -5,   0,   0,   0 } },
 {  9.588367137770e-06,  3.242755521218e-06, {   2,   0,  -2,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -3.152888216657e-06,  9.677448677079e-06, {   3,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  9.385965529532e-06,  3.441568661972e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -4,   2,   0,   0 } },
 {  4.544384390334e-06,  8.282124020574e-06, {   2,  -2,   1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -1.276874173449e-05, -5.328532195777e-08, {   1,   0,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.316777755055e-06, -1.050475607275e-05, {   0,   1,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.233835841966e-05,  4.812488880091e-07, {   3,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.574625728341e-07, -1.256109765035e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,   3,   0,   0,   0 } },
 { -8.537525678892e-06, -4.279519303672e-06, {   4,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -8.017912470757e-06,  4.797591398546e-06, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  8.017912470757e-06, -4.797591398546e-06, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 {  3.439319925278e-06,  9.375777906274e-06, {   2,   0,  -1,   0,   0,  13, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.388908975994e-06,  9.422724274483e-06, {   2,  -2,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.224994834788e-06, -1.058447238091e-05, {   1,  -1,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.621249000114e-06, -8.185818020216e-06, {   4,   0,  -1,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.027944976396e-06,  4.778529329332e-06, {   4,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.245850819036e-05,  3.468917640221e-07, {   0,   0,   1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 { -1.479589736546e-07, -1.265333670003e-05, {   2,  -2,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  9.479149518437e-06,  3.315516936940e-06, {   0,   2,   0,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.106576550266e-06,  5.687479086536e-06, {   1,   1,   0,   0,   0, -11,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.963446679434e-07, -1.229730331685e-05, {   3,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.183461344531e-05,  9.538688273450e-07, {   6,  -2,  -3,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.886572151329e-06,  2.894352696306e-06, {   2,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.513014640228e-09, -1.277880588947e-05, {   2,   0,  -4,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.025310542859e-05, -2.523548796707e-06, {   3,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.156651511786e-06, -7.618762485979e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0 } },
 { -3.808662591709e-06,  8.963691937052e-06, {   4,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.808662591709e-06, -8.963691937052e-06, {   4,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.436269892365e-08, -1.268386803228e-05, {   0,   2,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -6.699183084141e-06,  6.065104226859e-06, {   2,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.602251570038e-06, -1.116072979330e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 { -5.703824710501e-06, -7.057984342487e-06, {   2,   0,  -1,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.020803128231e-06, -1.074008378808e-05, {   2,   0,   1,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 { -2.205074550125e-06,  1.055406801674e-05, {   4,   0,  -2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 {  1.031649169132e-05,  2.438234964912e-06, {   1,  -1,   2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.202615785370e-06, -7.547787455518e-06, {   0,   0,   1,   0,   0,   0,   7, -16,   6,   0,   0,   0,   0 } },
 { -8.545487749572e-06, -4.204156724757e-06, {   4,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -7.801715402196e-09,  1.273865658878e-05, {   4,  -2,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.303476369975e-06, -7.438740753990e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,   1,  -1,   0,   0,   0 } },
 {  7.406663328424e-06, -5.329541369792e-06, {   2,   0,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.035019189827e-05,  2.383079826522e-06, {   4,   1,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.960046357754e-06, -4.766460087948e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -1.034305702095e-06, -1.169212905368e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  1.631828849669e-07, -1.256308515345e-05, {   2,  -2,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -8.798742757830e-06, -3.926615871291e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  1.267653637615e-05,  4.226732588156e-08, {   2,   0,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.047029520854e-05, -2.248225112932e-06, {   0,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.583421125374e-06,  8.130149666998e-06, {   0,   0,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0 } },
 { -5.440396211024e-06,  7.272096453750e-06, {   0,   0,   1,   0,   0,  -4,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.669064197978e-06, -1.103960276078e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  1.713528519269e-07, -1.253332128682e-05, {   1,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.209537657748e-05,  6.073900217037e-07, {   3,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.109166206296e-06, -7.588122501454e-06, {   2,   0,  -1,   0,   0,   0,  19, -24,   0,   0,   0,   0,   0 } },
 { -9.522806726389e-06,  3.173362547995e-06, {   2,   0,  -1,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.038109710197e-05, -2.311204775975e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 {  1.178080390718e-05,  9.109771221878e-07, {   2,   0,   0,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.582753709723e-07, -1.232898055788e-05, {   2,   0,  -1,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 { -5.214369513511e-06, -7.470222376042e-06, {   4,   0,  -2,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  9.035860173821e-06,  3.648094978915e-06, {   6,   0,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  1.262578362891e-08,  1.267027621425e-05, {   4,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -1.836121672451e-06, -1.084221323878e-05, {   3,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -8.761521685886e-06,  3.915375522019e-06, {   1,  -1,   0,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.025203317690e-05,  2.423263510952e-06, {   3,  -1,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.736709648909e-06, -3.934901049074e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0 } },
 {  2.135420102771e-06, -1.053528505279e-05, {   1,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.193085924633e-05,  7.357730766133e-07, {   0,   2,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -2.533216900038e-06, -1.013117073692e-05, {   1,   1,   2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  6.942781239968e-06,  5.720479479053e-06, {   6,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  2.180625401299e-07,  1.244408973749e-05, {   2,   0,  -4,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.651549808691e-07, -1.239425568705e-05, {   2,   0,  -4,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.182753074492e-06, -9.470792269344e-06, {   2,   0,  -2,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  1.205114873842e-06, -1.144649291418e-05, {   0,   0,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0 } },
 {  3.488796508782e-06, -9.153868026009e-06, {   2,   0,  -1,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 {  2.219152781727e-06, -1.042008063562e-05, {   3,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.194147991520e-05, -6.917777894450e-07, {   2,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.159838594681e-06,  7.472016783293e-06, {   2,   0,  -2,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  1.928935438136e-08, -1.260670317987e-05, {   4,   0,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.057868721445e-05, -2.044597593526e-06, {   2,   0,  -2,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  2.325060890032e-07,  1.238940141661e-05, {   2,  -2,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -7.955094023103e-06,  4.666520993175e-06, {   2,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.749662190769e-06,  7.865083204798e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 {  6.567206102059e-06, -6.046465093056e-06, {   1,  -1,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.198012375763e-08, -1.259757831995e-05, {   5,   0,  -2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.070645803303e-06,  5.538258428076e-06, {   1,  -1,   0,   0,   4,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.222217060174e-06,  1.138621466038e-05, {   1,   1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.132169409376e-06,  1.047462629151e-05, {   1,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -6.578907049236e-06,  6.022995374049e-06, {   0,   0,   0,   0,   0,   7,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.019398590423e-05,  2.407891679643e-06, {   1,   1,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.574555651894e-06, -4.026554475657e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -4,   0,   0,   0 } },
 {  3.610999194113e-08, -1.256188779192e-05, {   4,   0,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.332228849982e-06,  9.258694489467e-06, {   2,  -2,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -8.153510091873e-06,  4.433907560854e-06, {   1,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.527792126361e-06,  8.052507838502e-06, {   0,   0,   2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  5.341329986476e-06, -7.236327619887e-06, {   0,   0,   1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0 } },
 { -1.234442428649e-05, -2.307879685687e-07, {   1,   0,   0,   0,   0,  23, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.192856110705e-05,  6.379180619271e-07, {   2,   0,  -3,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.021438072067e-05, -2.351766176431e-06, {   0,   1,  -2,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  8.968969501710e-06, -3.594682260797e-06, {   1,   1,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.610657424916e-06,  1.094982929242e-05, {   2,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.458073997710e-06, -8.101762039709e-06, {   4,  -2,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -9.925615531175e-08, -1.245974976517e-05, {   6,   0,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  4.692129027757e-07,  1.208245871349e-05, {   1,   1,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.051031829706e-06, -1.049564750594e-05, {   1,   1,  -1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.174681788367e-06,  5.362750409640e-06, {   4,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.093120674216e-07, -1.221895881708e-05, {   2,   0,  -1,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.017554597213e-05, -2.352441914563e-06, {   1,   1,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.211490739076e-05, -4.111986939901e-07, {   0,   0,   0,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.535346435421e-06, -7.987302458534e-06, {   2,   0,  -2,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  3.987566815569e-06, -8.533009065408e-06, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -6.439298934956e-06,  6.077376096713e-06, {   2,   0,  -1,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0 } },
 {  1.006160440524e-05, -2.451636274838e-06, {   0,   0,   0,   0,   0,  24, -27,   0,   0,   0,   0,   0,   0 } },
 { -1.455418304822e-06, -1.104803124744e-05, {   2,   0,  -1,   0,   0,   3,  -9,   4,   0,   0,   0,   0,   0 } },
 { -1.006642269314e-05,  2.435960602637e-06, {   2,  -2,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.887997548013e-06,  5.609551961183e-06, {   2,   0,  -1,   0,   0,   0,   4,  -4,  -1,   0,   0,   0,   0 } },
 {  8.975925959952e-08, -1.240737646721e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,  -3,   0,   0,   0 } },
 {  1.738404509505e-06,  1.075672670676e-05, {   0,   2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -8.493635887770e-07, -1.164399282798e-05, {   3,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -1.411291662639e-06,  1.108047502960e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 {  1.010134929559e-05,  2.388457197809e-06, {   1,   1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.815680214237e-06,  4.673435553904e-06, {   4,   0,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  1.520279736158e-07,  1.233520196466e-05, {   2,   0,   0,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -5.008094979837e-06, -7.478110836936e-06, {   1,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.344089965974e-06, -1.113853779798e-05, {   4,   0,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.480210407298e-07, -1.223329194524e-05, {   2,   0,  -2,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -4.432657657829e-06,  8.042418008981e-06, {   2,   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.111853871816e-05,  1.354875139223e-06, {   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.013622009584e-05,  2.336353649409e-06, {   1,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.204350942845e-08, -1.245957693941e-05, {   3,   0,  -2,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.093790004433e-06, -1.037623570164e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 { -7.078644675700e-06,  5.387485755825e-06, {   1,   1,   0,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.058208601327e-05, -1.881762637401e-06, {   2,   0,   2,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.583225890730e-06,  8.878695193400e-06, {   2,   0,  -2,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.955175087167e-06,  3.500433235711e-06, {   0,   0,   1,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.879216169637e-06, -8.574000589556e-06, {   4,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  1.163348481991e-07, -1.233556031812e-05, {   0,   0,   4,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 {  1.147190306541e-05,  9.770304108087e-07, {   4,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  9.342732714717e-06, -3.100941292345e-06, {   0,   0,   0,   0,   0,  16, -19,   0,   0,   0,   0,   0,   0 } },
 {  1.197281390825e-05,  4.702366587264e-07, {   2,   0,   0,   0,   0,  29, -30,   0,   0,   0,   0,   0,   0 } },
 { -1.222393974415e-05,  2.171106962749e-07, {   0,   2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -3.508490275130e-06,  8.930781542436e-06, {   0,   0,   0,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0 } },
 { -7.657584781505e-06,  4.774865597651e-06, {   2,   0,  -2,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0 } },
 {  1.103461217740e-05, -1.392164196270e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 } },
 {  1.378039172239e-07,  1.228767106635e-05, {   6,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  7.061202941862e-06,  5.363217346795e-06, {   2,   0,  -1,   0,   0,   0,   3,  -6,   1,   0,   0,   0,   0 } },
 {  2.100709297002e-06, -1.031924931315e-05, {   1,  -1,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.006622073186e-05,  2.352560543093e-06, {   5,  -1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -7.762795472334e-06, -4.654199445600e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  1.368332317729e-08, -1.240145982882e-05, {   3,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.985209964709e-06, -6.417184764568e-06, {   0,   0,   1,   0,   0,   3, -14,  17,   1,   0,   0,   0,   0 } },
 { -1.681918255711e-06,  1.071999083224e-05, {   2,   0,  -1,   0,   0,   0,   7,   0,  -8,   0,   0,   0,   0 } },
 { -5.984743030138e-06,  6.416864368865e-06, {   0,   0,   1,   0,   0,  -3,  14, -17,  -1,   0,   0,   0,   0 } },
 { -9.333294242820e-06,  3.067818150033e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  5.465114209919e-08, -1.234487261929e-05, {   0,   2,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.005877453771e-05, -2.339843116925e-06, {   1,   1,  -1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.038106643025e-06, -5.357261601765e-06, {   2,   0,   1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -8.929540426458e-07, -1.150169835324e-05, {   2,  -2,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.170064061998e-06,  1.121624645130e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  7.031957823636e-06,  5.352933979286e-06, {   2,   0,   1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -2.990791368228e-06, -9.394069759258e-06, {   0,   0,   0,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 {  1.001682260337e-05,  2.366443990431e-06, {   3,  -1,  -1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.756682242038e-06, -5.623612775167e-06, {   1,  -1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 {  2.904279833923e-06,  9.473888613348e-06, {   2,   0,  -1,   0,   0,   0, -12,  21,  -2,   0,   0,   0,   0 } },
 { -7.795330587341e-06,  4.581976620253e-06, {   1,  -1,  -2,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  1.002112224294e-05, -2.354068493929e-06, {   1,  -1,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.713721678175e-06,  6.660305221834e-06, {   2,   0,   0,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0 } },
 { -5.451375938585e-07, -1.181858395463e-05, {   0,   0,   3,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 {  9.634480330003e-06,  2.726146639832e-06, {   0,   0,   0,   0,   0,  15, -10,  -4,   0,   0,   0,   0,   0 } },
 {  4.492337964214e-06,  7.861471549499e-06, {   1,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.178587841647e-05,  5.662127352386e-07, {   2,   0,   1,   0,   0,  19, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.090790121275e-06, -8.260238339068e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   7,   0,   0,   0,   0 } },
 { -1.221163028859e-07,  1.222717581123e-05, {   0,   2,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  4.112746264884e-06,  8.234498228791e-06, {   2,   0,   5,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.107051271104e-05, -1.265135625022e-06, {   2,   0,  -1,   0,   0,   0,   8, -19,   0,   3,   0,   0,   0 } },
 { -6.519609431846e-06,  5.813154875393e-06, {   3,   0,   0,   0,   0,   0, -33,  41,  -2,   0,   0,   0,   0 } },
 { -6.534232875976e-06, -5.796138161807e-06, {   6,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.141427135859e-06, -3.184201367420e-06, {   2,   0,   2,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.033806259867e-05, -1.985957059503e-06, {   0,   0,   0,   0,   0,  10,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.439459911660e-07,  1.157854116975e-05, {   2,   0,   1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.920791110046e-06, -4.398401605061e-06, {   2,   0,   0,   0,  -1,   0,   6,  -3,  -3,   0,   0,   0,   0 } },
 { -9.448371681617e-06, -2.868953023596e-06, {   3,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.950733659216e-06, -2.366268957053e-06, {   2,  -1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.083771804892e-05, -1.477475723232e-06, {   2,   0,  -2,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.747373005154e-07,  1.213971428761e-05, {   2,   2,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.210311265625e-06, -8.101861166109e-06, {   0,   0,   0,   0,   0,   0,  34, -39,   0,   0,   0,   0,   0 } },
 { -9.954932857948e-06, -2.353810396466e-06, {   1,  -1,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.913402426700e-06,  4.393712652679e-06, {   2,   0,   0,   0,   1,   0,  -6,   3,   3,   0,   0,   0,   0 } },
 { -6.845286061251e-09, -1.229921865530e-05, {   1,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.512426700239e-06, -5.782104577077e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  6.661065209846e-06,  5.624450070806e-06, {   1,  -1,   0,   0,   0,  19, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.433831994359e-06, -8.844308053270e-06, {   3,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.294145658631e-06,  5.981476169820e-06, {   1,   1,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.278520780829e-06, -4.991990721164e-06, {   0,   2,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -9.481855779052e-06,  2.788156605380e-06, {   6,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.923457784723e-06, -1.034524596189e-05, {   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.479623717546e-06, -2.788111346612e-06, {   6,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.884430309595e-09, -1.226417930735e-05, {   1,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.752171336965e-06,  2.513181709320e-06, {   4,   0,  -3,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.005144331323e-05, -2.211731241558e-06, {   4,   0,  -3,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.130689959365e-05,  9.559002714212e-07, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  9.598590509172e-06, -2.663197037012e-06, {   4,   0,  -2,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.157832367796e-06, -8.101146958864e-06, {   0,   2,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -8.862443498111e-06, -3.396401013257e-06, {   0,   0,   1,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.381449987704e-06,  8.876645920509e-06, {   1,   0,   0,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.590860845748e-06,  1.066641669168e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 {  1.218602512293e-05,  7.101017571550e-08, {   2,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.742486759630e-10, -1.225463295342e-05, {   1,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.065633536780e-06,  7.185968708110e-06, {   2,   0,  -1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0 } },
 {  6.116120593216e-06,  6.131609858942e-06, {   4,   0,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  8.703114623724e-06,  3.534383753687e-06, {   6,   0,  -3,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -6.835836349254e-06,  5.399921964385e-06, {   1,   1,   0,   0,   0, -10,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.401320990723e-06, -9.822652362442e-06, {   2,   0,  -3,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.886004779726e-06,  2.337612715231e-06, {   3,  -1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -9.598992213696e-06,  2.624499985313e-06, {   4,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.749669991472e-06, -5.468326651806e-06, {   2,   0,  -1,   0,   0,   0,  -2,   8,  -3,   0,   0,   0,   0 } },
 { -2.775255184797e-06, -9.442211953121e-06, {   0,   0,   2,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  1.018976061755e-05,  2.027192055585e-06, {   0,   0,   2,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.345756008031e-07,  1.167900214788e-05, {   0,   0,   2,   0,   0,   0,  -7,  16,  -4,  -5,   0,   0,   0 } },
 { -2.772167677533e-06,  9.431361159448e-06, {   0,   0,   2,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 { -9.091639649766e-07,  1.129266045827e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 { -9.907980355771e-06, -2.290670832483e-06, {   4,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -5.712718013812e-08, -1.214136998540e-05, {   4,   0,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -8.750796060407e-07,  1.132088575285e-05, {   0,   0,   2,   0,   0,  -8,   3,  -7,   8,  -6,   0,   0,   0 } },
 { -8.750796060407e-07, -1.132088575285e-05, {   0,   0,   0,   0,   0,   8,  -3,   7,  -8,   6,   0,   0,   0 } },
 { -6.881196594240e-06,  5.312924801870e-06, {   1,   1,   0,   0,   0,  -8,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.232156497237e-06, -7.958888641177e-06, {   4,   0,  -2,   0,  -4,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.196409631400e-06, -7.993768338788e-06, {   0,   0,   1,   0,   0,   6,  -9,   0,   3,   0,   0,   0,   0 } },
 { -4.809122104449e-06, -7.380413091692e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,  -3,   0,   0,   0 } },
 { -6.520446122161e-06,  5.668578602902e-06, {   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 { -9.455490280474e-07, -1.124282530197e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 {  4.058182467011e-06,  8.125251154442e-06, {   3,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.022588280450e-05,  1.956205559660e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   3,  -1,   0,   0,   0 } },
 { -8.590012117812e-06, -3.587501347121e-06, {   1,   1,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.202875879985e-07, -1.205664421243e-05, {   2,  -2,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  2.288199509975e-06,  9.887140015956e-06, {   1,  -1,   0,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  7.532680375243e-06, -4.641056472779e-06, {   2,  -2,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  8.424434428071e-06,  3.748955238646e-06, {   2,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 } },
 { -1.105528335086e-06,  1.106096751893e-05, {   2,   0,   0,   0,   0,   0,  16, -22,   0,   0,   0,   0,   0 } },
 {  1.040469173569e-05,  1.761219402964e-06, {   2,   0,  -1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.402083272071e-06,  1.076318429088e-05, {   2,   0,  -2,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  9.782815165258e-06, -2.375792033791e-06, {   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.804469245774e-06,  2.349802449630e-06, {   0,   0,   3,   0,   0, -24,  23,   0,   0,   0,   0,   0,   0 } },
 {  9.225136242245e-06,  2.924302294650e-06, {   0,   0,   2,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.897774047164e-06, -8.251042382688e-06, {   2,   0,  -2,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.316113446252e-06,  4.829200767401e-06, {   1,  -1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  1.005605982847e-05, -2.075594962693e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   8,   0,   0,   0 } },
 { -4.505952081359e-06, -7.624721208097e-06, {   2,   0,   1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -2.476863678002e-06, -9.643508382661e-06, {   1,  -1,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  6.644437078154e-06,  5.472763454916e-06, {   0,   2,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.059561956354e-05,  1.505330979794e-06, {   2,   0,   0,   0,   0,  20, -23,   0,   0,   0,   0,   0,   0 } },
 {  6.736565475878e-08,  1.202994659467e-05, {   2,   0,  -2,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.902558423062e-06,  9.192474962046e-06, {   5,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -4.580399431147e-06,  7.509433902651e-06, {   2,   0,  -1,   0,   0,  -3,   7,   0,  -4,   0,   0,   0,   0 } },
 { -7.557273988429e-06,  4.529819830567e-06, {   1,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.782220448790e-06,  5.301094120055e-06, {   1,   1,   0,   0,   0,  -9,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.608700606887e-06,  4.470438363907e-06, {   1,   1,  -2,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 { -9.725249667129e-06,  2.353296780383e-06, {   2,   0,   4,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.020729215523e-06, -8.052418673877e-06, {   2,   0,   3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.921248624736e-08, -1.204300147761e-05, {   0,   2,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.279517629441e-07, -1.193936943899e-05, {   2,   0,   0,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 {  1.350334990480e-09,  1.206213004391e-05, {   3,   0,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.023858940131e-06, -8.038895177643e-06, {   3,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.569899578621e-06, -4.485876709261e-06, {   2,   0,   1,   0,   0,   0,  10, -12,   0,   0,   0,   0,   0 } },
 {  1.175182948354e-05, -3.030903463600e-07, {   2,   0,   1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  9.504130242499e-06, -2.549803935126e-06, {   2,   0,   1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  8.438618294293e-06, -3.615230009397e-06, {   1,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 { -1.203857824503e-05, -7.796306935502e-09, {   6,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.111723551124e-06, -1.092859145458e-05, {   1,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -3.990838596268e-06,  8.044491161273e-06, {   3,  -2,   0,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.990838596268e-06, -8.044491161273e-06, {   1,   2,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 {  1.965787150084e-06,  1.006053401808e-05, {   1,  -1,   1,   0,   0,  10, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.479321377980e-06, -7.544449767244e-06, {   2,   0,   2,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  9.232023132850e-06,  2.791685051102e-06, {   0,   0,   1,   0,   0,   0,   6, -10,   3,   0,   0,   0,   0 } },
 { -6.966894933812e-06, -5.053933944338e-06, {   4,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.310566835773e-06, -6.704402123206e-06, {   0,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.350422659227e-08, -1.199348699851e-05, {   2,   0,   0,   0,   0,   0,  -4,   9,   0,   0,   0,   0,   0 } },
 { -1.847789925941e-06, -1.015306941972e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   5,   0,   0,   0 } },
 { -9.608785101332e-08,  1.189878691639e-05, {   0,   0,   2,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  9.109989843872e-06, -2.883916610217e-06, {   1,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 {  8.676506413157e-06, -3.308783680345e-06, {   0,   0,   3,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.593539027549e-06, -5.391543758053e-06, {   1,  -1,   1,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 { -9.045646429710e-06, -2.936336891580e-06, {   2,  -2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.320928956068e-06,  5.658491184758e-06, {   0,   0,   0,   0,   0,   8,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.674312014145e-06, -1.030435944692e-05, {   1,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.156305225265e-06,  7.817321222098e-06, {   0,   0,   0,   0,   4,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.796624088030e-08, -1.191452571061e-05, {   2,  -2,  -2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.706302857476e-06,  2.264276369440e-06, {   1,   1,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.184858775563e-05, -1.201541688983e-07, {   2,   0,   1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -1.122205202698e-06,  1.084258687179e-05, {   2,   0,   0,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 { -1.005511997990e-06, -1.095695949942e-05, {   2,   0,  -1,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0 } },
 { -3.603386666162e-06,  8.358408878908e-06, {   2,   0,  -2,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 { -9.778205072616e-06, -2.183458541925e-06, {   6,   0,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.963978572601e-06, -9.995560473651e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -1.654392961496e-06,  1.030137808807e-05, {   4,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  1.975745911647e-06, -9.972630928562e-06, {   1,  -1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -8.816775580353e-06, -3.126618087197e-06, {   2,   0,  -1,   0,   0,   0,  -4,  11,  -1,   0,   0,   0,   0 } },
 { -8.816775580353e-06, -3.126618087197e-06, {   2,   0,  -3,   0,   0,   0,  -4,  11,  -1,   0,   0,   0,   0 } },
 { -5.576621195257e-06,  6.365552849770e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,  -2,   0 } },
 { -2.440710492915e-06, -9.483777014866e-06, {   1,   1,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  7.734249493433e-06,  4.179368444652e-06, {   2,   0,  -1,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0 } },
 { -1.178046841956e-06, -1.073106432858e-05, {   1,   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  4.989673487323e-06,  6.914564651259e-06, {   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.324646865045e-06, -6.578337010368e-06, {   3,  -1,   0,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0 } },
 { -5.324646865045e-06,  6.578337010368e-06, {   3,  -1,  -2,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0 } },
 { -9.505783964499e-06,  2.396629812408e-06, {   5,  -1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -7.275450725412e-06, -4.616703188172e-06, {   2,  -2,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.220966910424e-06, -5.670788847201e-06, {   2,   0,  -2,   0,   0,   0,   3,  -2,  -2,   0,   0,   0,   0 } },
 {  9.188970221535e-06,  2.700465818414e-06, {   4,   0,  -1,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 {  1.777818632822e-06,  1.011157419708e-05, {   3,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.851606225557e-06, -9.026632474589e-06, {   3,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.845063318812e-06,  9.032679181122e-06, {   3,   1,  -2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  4.054441233597e-06,  7.814632571341e-06, {   1,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -9.227751441420e-06,  2.630711732810e-06, {   2,   0,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0 } },
 {  4.663314981249e-06, -7.189875428175e-06, {   0,   0,   1,   0,  -2,   7,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.564314423269e-06,  8.285302767019e-06, {   0,   0,   1,   0,   0,   0,   2,  -3,  -6,   3,   0,   0,   0 } },
 {  3.563837786158e-06, -8.284930755643e-06, {   0,   0,   1,   0,   0,   0,  -2,   3,   6,  -3,   0,   0,   0 } },
 { -9.651124111984e-06,  2.195022258131e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,   4,  -2,   0,   0,   0 } },
 {  2.464772500600e-06,  9.379949197668e-06, {   4,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  2.464772500600e-06, -9.379949197668e-06, {   0,   0,   2,   0,   0,  19, -21,   3,   0,   0,   0,   0,   0 } },
 { -3.690226982504e-06,  8.153118932610e-06, {   2,   0,  -2,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.423306181100e-08, -1.179659665220e-05, {   2,  -2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.292584749112e-06, -4.544191101948e-06, {   6,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  9.554339070403e-06,  2.278934465947e-06, {   3,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.910551666091e-06, -9.918849800802e-06, {   2,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  9.510565609175e-06,  2.316238980589e-06, {   1,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  1.124758620949e-05,  5.766122547860e-07, {   3,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.941772963365e-06,  9.878739204196e-06, {   1,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  8.270783400900e-06,  3.545617036012e-06, {   1,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 { -7.338981564413e-06, -4.475346452201e-06, {   0,   0,   3,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.124562281360e-06, -7.689012854992e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 {  9.550344907606e-06,  2.259031983967e-06, {   0,   1,   0,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -7.226266400084e-06,  4.581337867495e-06, {   2,   0,  -1,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0 } },
 {  3.994140954266e-06,  7.811311805637e-06, {   1,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  1.159595393031e-08, -1.179295187880e-05, {   4,   0,  -1,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.069917963838e-06,  9.732728263404e-06, {   2,   0,  -3,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.818307955800e-06, -5.982905864702e-06, {   1,   0,  -1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  9.539570037168e-06,  2.257732792910e-06, {   3,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -9.534080555903e-06,  2.259136650941e-06, {   1,   1,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.476169638518e-06, -3.314271580321e-06, {   0,   2,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.650848354025e-06, -8.138799917200e-06, {   0,   0,   1,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0 } },
 { -7.740009698393e-06, -4.049259357351e-06, {   2,   0,   0,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0 } },
 { -3.425390329476e-06,  8.360950447323e-06, {   4,   0,  -2,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 { -1.030016374178e-05, -1.483928300257e-06, {   3,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  9.723539836434e-06,  2.056656547850e-06, {   1,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  3.339954729871e-06, -8.439641958988e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -8,   0,   0,   0 } },
 { -1.055166391554e-05, -1.216529181621e-06, {   0,   0,   0,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 { -8.700267353351e-06,  3.064237484530e-06, {   1,   1,  -1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  9.528949561623e-06,  2.227975591626e-06, {   2,   0,  -3,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  7.235027309795e-06,  4.507912548401e-06, {   4,   0,  -4,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.893197885877e-06,  8.849555651180e-06, {   0,   0,   1,   0,   0,   0,  -4,   0,   6,   0,   0,   0,   0 } },
 { -3.701244584505e-06,  8.041488865169e-06, {   2,   0,   1,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.898721003470e-06,  3.836600478337e-06, {   1,   1,   0,   0,   0, -28,  33,   0,   0,   0,   0,   0,   0 } },
 {  2.120634654811e-06,  9.614463854548e-06, {   0,   1,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.330749128896e-06, -1.039366619360e-05, {   0,   0,   0,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  8.868636542695e-06,  2.855448144060e-06, {   2,  -2,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -9.867757752259e-06, -1.856165104837e-06, {   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.054779323054e-05,  1.168741817143e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,  -5,   0,   0,   0 } },
 { -4.466887912146e-06, -7.248199644170e-06, {   2,   0,  -1,   0,   0,   5,  -8,   3,   0,   0,   0,   0,   0 } },
 {  4.702952903408e-06, -7.009820520767e-06, {   0,   0,   3,   0,   0, -20,  16,   7,  -8,   6,   0,   0,   0 } },
 {  4.702952903408e-06,  7.009820520767e-06, {   0,   0,   1,   0,   0,  20, -16,  -7,   8,  -6,   0,   0,   0 } },
 {  8.858621717926e-06, -2.852353224910e-06, {   2,  -2,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -6.562924154829e-06, -5.144431236394e-06, {   2,   0,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.489367323801e-06, -8.212230144474e-06, {   3,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.489367323801e-06,  8.212230144474e-06, {   3,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.713509423193e-06, -2.981300037500e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   7,  -5,   0,   0,   0 } },
 {  9.112261110513e-06, -2.577250677842e-06, {   0,   0,   1,   0,   0,  23, -26,   0,   0,   0,   0,   0,   0 } },
 { -6.521098220987e-06,  5.167873287307e-06, {   2,   0,  -1,   0,   0,   0,  10, -21,   2,   0,   0,   0,   0 } },
 {  2.127371002959e-07,  1.147411031594e-05, {   4,   0,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -5.621301549209e-06,  6.063016634106e-06, {   2,   0,   0,   0,  -7,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.036722289100e-06,  9.639944838000e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.461117576858e-06,  4.207978365289e-06, {   2,   0,  -1,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0 } },
 {  6.613880841653e-06, -5.052089704606e-06, {   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.181604745810e-06,  1.048289059248e-05, {   4,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.562048730610e-06,  2.101517591280e-06, {   0,   0,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.401169692743e-06,  3.262035319802e-06, {   0,   0,   3,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.198851315292e-08, -1.156976402932e-05, {   3,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.287328557868e-06,  2.363757302161e-06, {   2,   0,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0 } },
 { -1.983243075033e-06,  9.663020927276e-06, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,  -1,   0,   0,   0 } },
 { -1.982608743228e-06, -9.663013555832e-06, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   1,   0,   0,   0 } },
 {  1.120629845284e-06,  1.052394405601e-05, {   1,  -1,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.850063039669e-06, -4.791622842989e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 { -3.386392196532e-06,  8.253224346118e-06, {   3,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.468697335115e-06,  2.170593990453e-06, {   1,   1,   1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 {  8.297319750774e-07, -1.080732084523e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.017822073076e-06, -9.615654194099e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   0,   0,   0 } },
 {  9.376635508511e-06, -2.256771936378e-06, {   1,   1,   0,   0,   0, -22,  21,   0,   0,   0,   0,   0,   0 } },
 {  3.957535467287e-06,  7.669410805392e-06, {   4,   0,  -2,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.176347940142e-08,  1.161421365438e-05, {   1,   0,   1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  9.095714193587e-06, -2.529316331861e-06, {   2,  -2,   1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.521152574740e-06,  1.010384557053e-05, {   2,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.455641654922e-06,  9.167870585251e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0 } },
 {  2.455641654922e-06, -9.167870585251e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0 } },
 {  2.915738731191e-06, -8.704145503543e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  8.566418475775e-07, -1.075820169221e-05, {   2,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.116819412460e-06, -2.497693206831e-06, {   1,  -1,   1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.892609892177e-06,  9.719684410119e-06, {   1,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -9.858332108606e-06, -1.751677700809e-06, {   0,   0,   2,   0,   0, -15,   8,   4,   0,   0,   0,   0,   0 } },
 {  9.408487385876e-06, -2.201010917181e-06, {   3,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.993415653643e-07, -1.130442227740e-05, {   2,   0,  -2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 {  4.564674502915e-06,  7.034647585750e-06, {   0,   0,   1,   0,   2,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.091492844115e-06, -3.506240365563e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.261524810977e-06, -5.335531899210e-06, {   1,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  7.814158824673e-06,  3.781173152160e-06, {   1,  -1,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.420185574442e-06, -2.171937978845e-06, {   1,  -1,  -2,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.148212001033e-08, -1.157826151786e-05, {   2,  -2,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.791556314492e-06, -6.796253090222e-06, {   1,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -1.022629440741e-05, -1.361010387495e-06, {   0,   0,   1,   0,   0,   0,   5, -11,   0,   0,   0,   0,   0 } },
 {  4.087525649353e-08, -1.154573364940e-05, {   0,   0,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 { -1.156690642147e-08, -1.157289480289e-05, {   1,   0,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  9.339037682167e-06, -2.243036964874e-06, {   2,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.085715807626e-06,  9.493928925948e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   1,   1,   0,   0,   0 } },
 { -7.082068896370e-06,  4.497210078241e-06, {   2,   0,  -1,   0,   0,   2,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.034041737411e-08, -1.156409274141e-05, {   3,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.084524476913e-06, -9.488563238427e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,  -1,  -1,   0,   0,   0 } },
 {  3.205568861611e-06,  8.366973221030e-06, {   1,   1,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.061578587667e-05, -9.488367253939e-07, {   0,   0,   1,   0,   0,  -9,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.664925880048e-06, -3.899405388855e-06, {   4,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -6.084613758016e-09, -1.155610472659e-05, {   0,   2,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.968629612527e-06,  8.589827700345e-06, {   2,   0,   0,   0,   0,   0, -11,  18,   0,   0,   0,   0,   0 } },
 { -9.015283633693e-06, -2.542121705316e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   9,   0,   0,   0 } },
 { -2.118307083907e-07, -1.134403419581e-05, {   1,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 { -1.034977911602e-06, -1.052061484527e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   5,   0,   0,   0 } },
 {  1.131315406143e-05,  2.406894886006e-07, {   2,   0,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.307745505770e-06, -2.243902567175e-06, {   1,   1,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.322971245527e-06, -3.223502038273e-06, {   2,   0,  -2,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -9.085911271766e-07, -1.063772495706e-05, {   4,   0,  -1,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0 } },
 {  7.740532080553e-06, -3.804255635763e-06, {   2,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.443887873739e-06,  1.009896874566e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -4,   0,   0,   0 } },
 { -5.755227354789e-06, -5.787623032661e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.121201073193e-05,  3.288594318992e-07, {   2,   0,   2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.544982985410e-06,  9.993563382895e-06, {   0,   0,   1,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 { -7.189484878906e-06, -4.348936171176e-06, {   4,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.134869063164e-06, -9.392498359952e-06, {   1,  -1,   0,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -3.660710195774e-06,  7.866606798617e-06, {   2,   0,   1,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.150708325838e-05, -1.692197460843e-08, {   1,   0,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.596786853307e-06, -1.923136490758e-06, {   2,   0,  -3,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.645261946364e-06,  4.871006679958e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -4,   0,   0,   0 } },
 {  9.331254443869e-06,  2.183254940910e-06, {   1,  -1,   1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.537480945410e-06,  1.976992306168e-06, {   4,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.106704111660e-05, -4.445129704159e-07, {   2,  -2,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.398998879429e-09,  1.150092893102e-05, {   2,   0,  -5,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  4.101212534688e-06,  7.405364899330e-06, {   2,  -2,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -1.144005320461e-05,  6.443558996829e-08, {   4,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.840791199128e-06,  7.663016811939e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  9.097730343411e-06, -2.405891287644e-06, {   0,   0,   2,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -1.097652613780e-05, -5.256094460865e-07, {   4,   0,  -3,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  9.083201087715e-06,  2.416234799606e-06, {   0,   0,   2,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -4.814007333244e-06, -6.683596942129e-06, {   4,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.434544936311e-06,  1.006157870800e-05, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,  -4,   0,   0,   0 } },
 {  1.142739086036e-05,  6.577976198032e-08, {   4,   0,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.587437682122e-06,  5.905385542577e-06, {   2,   0,   0,   0,   0,   0,  -6,   0,   4,   0,   0,   0,   0 } },
 {  2.087879299734e-06, -9.404918043636e-06, {   1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.588626449116e-06,  6.903723160811e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   3,   0,   0 } },
 { -9.440573465147e-07,  1.054644298549e-05, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -1.710662674857e-06,  9.777672994587e-06, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  1.142420693531e-05,  6.020087260458e-08, {   1,   2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.100948381226e-05, -4.748242989873e-07, {   4,  -2,  -3,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.124334711061e-05,  2.379973865378e-07, {   1,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.971171921282e-06,  5.508646693796e-06, {   1,   1,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.769351120077e-06, -5.704062544328e-06, {   1,   1,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.089653379445e-05,  5.745859308609e-07, {   3,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.982798123971e-06, -9.486968791782e-06, {   1,   1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  9.310217747199e-06, -2.159171909120e-06, {   1,   1,  -1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.901872362238e-07,  1.087771608974e-05, {   2,   0,  -2,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -1.026773685136e-05,  1.199925924886e-06, {   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.243283426012e-07, -1.113640781709e-05, {   1,   0,   2,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 { -3.243283426012e-07,  1.113640781709e-05, {   1,   0,   0,   0,   0, -41,  41,   0,   0,   0,   0,   0,   0 } },
 { -8.937358336055e-06, -2.522942371973e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  9.267955934690e-06,  2.188467483018e-06, {   1,   1,   1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.491402390039e-07,  1.090170072618e-05, {   2,   0,  -2,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  3.824357814427e-06,  7.626300716344e-06, {   0,   0,   1,   0,   0,   9, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.036595352722e-05, -1.082626474359e-06, {   2,   0,  -1,   0,   0,   2,   5, -13,   8,  -6,   0,   0,   0 } },
 {  1.130077781091e-05,  1.455033329522e-07, {   0,   2,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.403671643355e-06, -6.041721922782e-06, {   2,   0,  -1,   0,   0,   3,  -4,   2,   0,   0,   0,   0,   0 } },
 { -6.576342366863e-06,  4.868197991593e-06, {   1,   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -8.280640683120e-06,  3.160125211994e-06, {   1,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.589659543644e-06, -4.849274789695e-06, {   0,   0,   1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -2.863081266032e-08, -1.140862938779e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,  -3,   0,   0,   0 } },
 { -6.434939076695e-06,  4.998652734015e-06, {   1,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.096637216282e-05,  4.665134531480e-07, {   3,  -2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.118379282354e-06,  7.311566880576e-06, {   4,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  4.126899970120e-06,  7.300510804599e-06, {   0,   2,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -9.307349608323e-08, -1.133393399461e-05, {   2,   0,   1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -2.736644523335e-08, -1.139460361451e-05, {   2,   0,   2,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.805334137023e-06, -2.609994944143e-06, {   2,   0,  -3,   0,  -4,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.825611275668e-06, -2.584275321055e-06, {   3,  -1,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.085213257477e-06,  7.321105733922e-06, {   0,   0,   2,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.702602101038e-06,  8.698848187590e-06, {   1,  -1,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.371425335829e-06,  4.030009448144e-06, {   1,  -1,  -1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.210129170151e-06,  2.189719289643e-06, {   1,   1,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.106768423183e-05,  3.302359915784e-07, {   2,   0,  -1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.298409907823e-06,  9.096946205900e-06, {   0,   0,   1,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.065708191160e-05,  7.326729208572e-07, {   4,   0,  -2,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  4.319162784623e-06,  7.065008711071e-06, {   0,   0,   0,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.834064510770e-06, -6.545732617316e-06, {   3,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.836524741612e-06,  7.541257870954e-06, {   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.905942856718e-06, -9.462992040825e-06, {   1,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.728326344601e-06, -7.637567667731e-06, {   0,   0,   2,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.307316949615e-06,  4.057402056527e-06, {   0,   0,   1,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 { -9.316906519453e-06, -2.047455007930e-06, {   6,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.241643858400e-06, -1.011858916292e-05, {   2,   0,   0,   0,   0,   2,  -8,   7,   0,   0,   0,   0,   0 } },
 { -6.958205387176e-06,  4.401318772180e-06, {   1,  -1,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.819646463146e-08,  1.130659076520e-05, {   0,   2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 { -1.126420912416e-05,  8.957015853148e-08, {   4,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.162517940941e-06,  9.187637557550e-06, {   1,   1,   0,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 {  7.782680035246e-06, -3.562208608966e-06, {   2,   0,   0,   0,   0,   0, -12,  14,   0,   0,   0,   0,   0 } },
 {  1.485664680243e-07, -1.119414930435e-05, {   1,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  1.359104387463e-06,  9.977047604455e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,   7,   0,   0,   0 } },
 { -6.920905958950e-06, -4.413653231027e-06, {   6,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.967982026083e-06,  9.359083972782e-06, {   1,  -1,   0,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -8.900167861298e-08, -1.123753414864e-05, {   2,  -2,  -1,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.933824382225e-06, -9.392419112544e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.104566022710e-05,  2.803476307477e-07, {   3,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.934247761631e-06,  9.391358635821e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.342860319446e-06, -9.980234691473e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -3.681225532265e-06, -7.641795677253e-06, {   1,   0,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.613298990543e-06,  1.709091045156e-06, {   2,   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 {  5.296790198341e-06,  6.024469120794e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 { -1.411802093611e-06, -9.904524417043e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 {  7.135648815560e-06, -4.178237028662e-06, {   2,   0,   0,   0,   0,  -2,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.088520741627e-05, -4.283485863733e-07, {   2,   0,  -1,   0,   0,   1,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.088520741627e-05, -4.283485863733e-07, {   2,   0,  -3,   0,   0,   1,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.066112377246e-05, -6.505880441034e-07, {   4,   0,  -2,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  7.243207150175e-06,  4.066805321510e-06, {   4,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -6.870133025265e-06, -4.437223701256e-06, {   3,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -3.371713906396e-06,  7.935332686754e-06, {   0,   2,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.371713906396e-06, -7.935332686754e-06, {   0,   2,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.059005497591e-05, -7.103967091558e-07, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.995596916629e-06,  8.297673352451e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 {  8.989205142250e-06,  2.302108484033e-06, {   2,   1,   0,   0,   0, -51,  58,   0,   0,   0,   0,   0,   0 } },
 { -8.989205142250e-06, -2.302108484033e-06, {   2,   1,  -2,   0,   0, -51,  58,   0,   0,   0,   0,   0,   0 } },
 {  3.745556554441e-06, -7.541322269854e-06, {   2,   0,  -1,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  4.590251479025e-06,  6.692158392365e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   6,   0,   0,   0 } },
 { -4.590251479025e-06, -6.692158392365e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   6,   0,   0,   0 } },
 {  7.498685769097e-09,  1.127445526202e-05, {   0,   2,  -2,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.075161848573e-09, -1.127625627745e-05, {   1,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.429246190460e-06,  1.851445988029e-06, {   4,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.953213066189e-06,  9.326905455196e-06, {   1,   1,   0,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.026644777300e-05,  1.013661338058e-06, {   1,   1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  8.257050364370e-06,  3.019963087181e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -4,   2,   0,   0 } },
 { -2.601801691701e-07,  1.100913385969e-05, {   2,   0,  -2,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  9.100907082314e-06, -2.168305896112e-06, {   2,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.706719119839e-08, -1.122053730818e-05, {   0,   0,   4,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.410328436951e-06, -9.854136651252e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   4,   0,   0,   0 } },
 { -6.705860061657e-06, -4.557813740761e-06, {   2,   0,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -6.017971426664e-06,  5.236964877546e-06, {   0,   0,   0,   0,   0,   9,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.456400130701e-06, -9.793871823186e-06, {   2,   0,  -2,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  1.280423564450e-07,  1.111816138761e-05, {   0,   0,   3,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  1.127316653345e-06, -1.011425812934e-05, {   2,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0 } },
 {  6.929100929164e-06,  4.312342474529e-06, {   3,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -6.859656515501e-06,  4.379311205599e-06, {   6,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.992320107246e-06, -7.244313573317e-06, {   2,  -2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  8.981360671779e-06, -2.243350741051e-06, {   2,   0,  -3,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.096376229724e-06, -1.012507554002e-05, {   0,   0,   1,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0 } },
 {  2.451587266370e-06,  8.766432220596e-06, {   1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.068492173025e-06, -2.149079365615e-06, {   5,  -1,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.748819661945e-06, -3.468563277781e-06, {   2,   0,  -1,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.138708333841e-06, -2.074850466899e-06, {   1,  -1,   0,   0,   0,  25, -24,   0,   0,   0,   0,   0,   0 } },
 { -6.758918640839e-07,  1.053718590580e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -5,   5,   0,   0,   0 } },
 { -5.032927697205e-06, -6.180048585503e-06, {   6,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -9.145445467231e-06,  2.062922551533e-06, {   3,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.074379699564e-06, -2.133834002306e-06, {   1,   1,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.999540821684e-06, -4.207303890671e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   5,  -5,   0,   0,   0 } },
 { -7.393856632886e-06, -3.802883685833e-06, {   2,   0,  -2,   0,   0, -20,  18,   6,  -2,   0,   0,   0,   0 } },
 {  7.391199177968e-06, -3.802830262305e-06, {   0,   0,   1,   0,   0,   0,  -9,  16,  -6,   0,   0,   0,   0 } },
 {  8.753082160014e-06,  2.438419680197e-06, {   1,  -1,   0,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  4.477569985407e-06,  6.709707367650e-06, {   1,  -1,   0,   0,   2,  16, -20,   0,   0,   0,   0,   0,   0 } },
 { -4.477569985407e-06, -6.709707367650e-06, {   1,  -1,  -2,   0,   2,  16, -20,   0,   0,   0,   0,   0,   0 } },
 { -9.006947059663e-06,  2.179448185206e-06, {   6,   0,  -1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.098041108428e-06, -6.087595484831e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   3,   0,   0,   0 } },
 {  6.135280773106e-06, -5.043666664431e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -3,   0,   0,   0 } },
 {  1.168387017139e-08, -1.116289309257e-05, {   0,   2,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.533658443294e-08,  1.112888928904e-05, {   2,  -2,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.076148212402e-06, -2.087129190990e-06, {   1,   1,  -1,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -9.859390919924e-06, -1.300121898226e-06, {   4,  -2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -4.858340371499e-06,  6.300067000916e-06, {   4,   0,  -1,   0,   0, -25,  35, -11,   0,   0,   0,   0,   0 } },
 { -4.858340371499e-06, -6.300067000916e-06, {   0,   0,   1,   0,   0,  25, -35,  11,   0,   0,   0,   0,   0 } },
 { -8.920972369833e-06, -2.235992463808e-06, {   0,   2,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.586073824169e-06,  1.569704819826e-06, {   2,   0,   0,   0,   0,   0, -11,  15,   0,   0,   0,   0,   0 } },
 { -1.586323671831e-06, -9.568215024732e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 { -8.017405616752e-06,  3.134693795227e-06, {   1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.187916235211e-06, -3.962837589116e-06, {   2,   2,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  9.018027776436e-06, -2.131126873104e-06, {   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.954834816837e-06, -9.192676135730e-06, {   1,  -1,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -1.741952814993e-06,  9.404079599782e-06, {   3,   1,  -2,   0,   0, -20,  12,  15,   0,   0,   0,   0,   0 } },
 { -1.741952814993e-06, -9.404079599782e-06, {   1,  -1,   0,   0,   0,  20, -12, -15,   0,   0,   0,   0,   0 } },
 {  3.855884224712e-06, -7.287771972969e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  1.466357030478e-08, -1.112427336361e-05, {   4,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.508455186354e-07,  1.018636071636e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.074578920692e-05,  3.880938625599e-07, {   0,   2,  -1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.017449205944e-06,  4.112703977480e-06, {   2,   0,   0,   0,   0,   0,  -2,  -2,   0,   0,   0,   0,   0 } },
 {  9.768268645196e-06,  1.361860316238e-06, {   4,   0,  -2,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -3.349346972687e-06, -7.774243062051e-06, {   2,   0,  -1,   0,   0,   0,  17, -24,   0,   0,   0,   0,   0 } },
 { -8.990131962387e-06, -2.130422742485e-06, {   3,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.359933004615e-08,  1.110604095720e-05, {   1,   0,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 {  9.741139704770e-06, -1.376025501384e-06, {   2,   0,   1,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 { -8.568280994537e-06, -2.547481954716e-06, {   2,   0,  -1,   0,   0,   2,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.376998555267e-06, -9.737786785446e-06, {   0,   0,   0,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 { -8.572844575654e-06,  2.529399017909e-06, {   2,   0,  -1,   0,   0,   2,  -6,   8,  -3,   0,   0,   0,   0 } },
 {  7.560881463234e-06, -3.539676040640e-06, {   1,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 { -1.019317205756e-05, -9.061661952416e-07, {   2,  -2,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 {  1.020466734461e-08, -1.108801714550e-05, {   2,  -2,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.044265372608e-06, -9.050329285648e-06, {   1,   1,   0,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 { -3.581486948890e-06,  7.512476741004e-06, {   2,   0,   1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.125611249957e-06,  8.964326478565e-06, {   1,  -3,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.638507670393e-06,  4.446735003776e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  1.264241802234e-06, -9.820628404431e-06, {   2,   0,   1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  5.789317363411e-06, -5.284652372466e-06, {   1,   1,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.953202961880e-06,  2.118871742608e-06, {   4,   0,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -7.482395255466e-06,  3.588562153848e-06, {   2,   0,  -2,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  8.145254335908e-06, -2.923226536534e-06, {   2,   0,  -1,   0,   4,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.970630724869e-06,  2.096912780916e-06, {   1,  -1,   1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 {  9.356475187248e-06,  1.710458759160e-06, {   3,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -8.689879778314e-06,  2.376461176052e-06, {   1,   1,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.773034291801e-06,  7.291715411823e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   1, -10,   0,   0,   0 } },
 {  7.748039016276e-06,  3.314259291578e-06, {   1,   3,  -2,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 { -1.826219105457e-06, -9.234329968634e-06, {   1,  -1,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.370280541690e-06, -4.689252666644e-06, {   2,   2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  2.112608203230e-07, -1.084512128382e-05, {   2,   2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  8.745816851548e-06,  2.307428689386e-06, {   2,   0,  -1,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.760101894590e-06,  4.292036370706e-06, {   1,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.519558918132e-06, -8.529799271856e-06, {   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  4.154202415953e-06, -6.886063586021e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,  -4,   0,   0,   0 } },
 {  8.901235066489e-06, -2.132389353216e-06, {   1,  -3,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.154423974382e-06, -6.879145396676e-06, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,  -4,   0,   0,   0 } },
 { -2.962961429705e-06, -8.069545589704e-06, {   2,   0,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.407258178275e-06, -6.623231523618e-06, {   2,   0,  -1,   0,   0,   2,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.900496177672e-06, -2.128188604682e-06, {   2,  -2,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.356548868180e-06,  3.671152815393e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0 } },
 {  6.313841294406e-06,  4.712508468578e-06, {   2,   0,  -1,   0,   0,   0, -10,  14,   0,   0,   0,   0,   0 } },
 { -1.067586247302e-05, -3.504392383316e-07, {   3,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.679118160397e-06, -9.346319103443e-06, {   3,  -2,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.011436105830e-06,  9.008383984384e-06, {   6,   0,  -4,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.037804113501e-06, -1.980837854514e-06, {   0,   2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.256817198704e-06,  5.761708075523e-06, {   0,   0,   1,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0 } },
 {  6.898901623752e-06, -4.118642739199e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -7.975975085232e-06,  3.037104856461e-06, {   4,   0,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.086917954506e-05,  1.436703919199e-07, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -2.969008392196e-07, -1.071567219727e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0 } },
 { -1.056109585696e-05,  4.511061946566e-07, {   0,   2,  -3,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.022471652241e-05,  7.848110470372e-07, {   1,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  5.119986729822e-09, -1.100360771788e-05, {   2,  -2,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.141795279469e-09, -1.100298826238e-05, {   4,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.069648613508e-05,  3.035172664217e-07, {   2,   0,  -1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 { -3.175409004733e-07, -1.067277308135e-05, {   4,  -2,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -2.198153540715e-06,  8.787765494449e-06, {   3,  -1,   0,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  9.069452633464e-06,  1.911789752251e-06, {   1,  -1,   1,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.626708492507e-06,  9.354154423718e-06, {   3,  -1,  -1,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.807298057315e-07, -1.049782094896e-05, {   0,   0,   2,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 {  7.610460584149e-06, -3.361416063850e-06, {   1,   1,  -3,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.246673599964e-06, -7.724225633108e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0 } },
 { -2.500163065510e-07,  1.071483910034e-05, {   2,   0,  -1,   0,   0,   3,  -6,   0,   1,   0,   0,   0,   0 } },
 {  4.011701443585e-08, -1.091732798826e-05, {   0,   2,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.425356683395e-06, -6.521567078612e-06, {   2,   0,  -3,   0,   0,   0,  10, -16,   3,   0,   0,   0,   0 } },
 { -1.088919372405e-05, -5.568030384809e-08, {   1,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.008119895614e-06, -9.934724058651e-06, {   4,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -7.824827640496e-06,  3.117567639550e-06, {   1,  -1,  -1,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 {  8.264398269163e-06, -2.676816934993e-06, {   0,   0,   0,   0,   0,  17, -20,   0,   0,   0,   0,   0,   0 } },
 { -6.780361833050e-06, -4.160462536638e-06, {   2,   0,  -2,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  7.801061759049e-08,  1.086194018351e-05, {   2,   0,  -1,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0 } },
 {  3.462770191863e-07, -1.059168992576e-05, {   2,   0,   1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -3.590046791975e-06, -7.347440104695e-06, {   2,   0,  -2,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -8.438170633693e-06,  2.495957422639e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0 } },
 {  2.419901565411e-06, -8.511521656478e-06, {   1,  -1,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.025203437894e-05, -6.776627667868e-07, {   3,   0,  -2,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -7.977484088505e-09,  1.092158335951e-05, {   0,   0,   3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.061278130741e-05,  3.161698036495e-07, {   4,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -6.752794815560e-06, -4.175781032889e-06, {   0,   2,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  8.802378956765e-06, -2.126103049582e-06, {   1,  -1,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.681733734798e-06, -5.244520480163e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,  -3,   0,   0,   0 } },
 {  4.114983342200e-06,  6.810148867949e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 { -5.275476622835e-06,  5.649520794949e-06, {   6,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  9.791389418507e-06, -1.132845747124e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -1.082794769168e-05,  9.396017173182e-08, {   4,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.589019602684e-06, -7.331998473670e-06, {   2,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -9.698355570518e-07,  9.950894179210e-06, {   2,   0,   0,   0,   0,   0,  17, -23,   0,   0,   0,   0,   0 } },
 {  2.509101041116e-06,  8.409295174204e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  9.433584962586e-06,  1.484272261537e-06, {   2,  -1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -3.503164800140e-06, -7.413740445179e-06, {   2,   0,  -2,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.004492314619e-06, -8.906271551987e-06, {   1,  -1,  -1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 {  8.772884502668e-07, -1.003169551306e-05, {   2,   0,   0,   0,   0,   0,  -4,   0,   2,   5,   0,   0,   0 } },
 {  9.740980180585e-06,  1.167761895508e-06, {   1,  -1,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.414157000836e-06, -2.488765690626e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,  -3,   0,   0,   0 } },
 {  1.785184464437e-06,  9.117126592055e-06, {   1,   1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.464521739857e-06,  2.436013726180e-06, {   1,  -1,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.885745872016e-06, -8.014688051603e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -3,  -5,   0,   0,   0 } },
 {  8.752177244129e-06, -2.148122957399e-06, {   0,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.237174706211e-06,  1.656763724753e-06, {   2,   0,  -2,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  8.266056810678e-06,  2.624506325581e-06, {   4,   0,  -1,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.524144946787e-06,  5.365892824452e-06, {   2,   0,  -1,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0 } },
 { -6.190476580705e-06,  4.697332831103e-06, {   1,  -1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.682653560861e-06,  9.203615774322e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   2,   0,   0,   0 } },
 { -7.094173323445e-06,  3.790438535767e-06, {   2,   0,   0,   0,   0,   0,  21, -25,   0,   0,   0,   0,   0 } },
 { -8.048047137707e-06, -2.835244809746e-06, {   3,  -1,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.022467532625e-05, -6.583413194786e-07, {   0,   2,   0,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -8.785063729595e-06, -2.095350027384e-06, {   1,  -1,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  8.780439953494e-06,  2.094710257420e-06, {   3,   1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.816011118092e-06,  8.054788261818e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  4.131520202237e-06, -6.739166896554e-06, {   2,   0,  -1,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0 } },
 { -8.751696276334e-06,  2.117684098793e-06, {   6,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.451834108801e-06,  7.416317575907e-06, {   2,   0,   1,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  9.058720107411e-06, -1.805598054072e-06, {   4,   0,  -3,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.696985574226e-06, -8.163397905139e-06, {   1,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.533834502405e-06, -6.326530055434e-06, {   0,   0,   1,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.816070295045e-06, -5.039818860882e-06, {   0,   0,   2,   0,   0,   8, -17,   8,  -3,   0,   0,   0,   0 } },
 {  5.199679572341e-06, -5.654715654963e-06, {   2,   0,   1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  4.090026418126e-06,  6.759288491001e-06, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   4,   0,   0,   0 } },
 { -8.886141978340e-06,  1.961405133051e-06, {   3,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -6.583858188390e-06,  4.263546567111e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   0,   5,   0,   0,   0 } },
 { -5.807875735142e-06,  5.033095281309e-06, {   0,   0,   2,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 { -7.739392596060e-06,  3.079009105879e-06, {   2,   0,   2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.300185754329e-07, -1.028821172083e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   1,   0,   0,   0 } },
 { -1.040296975211e-05, -4.147828567050e-07, {   4,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  9.648742488130e-06, -1.166947757170e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 {  1.066473574547e-05, -1.477437006452e-07, {   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  2.419456114410e-06, -8.392590471381e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0 } },
 {  6.208874646713e-06, -4.595808988476e-06, {   2,   0,  -1,   0,   0,   0,  -3,   6,  -1,   0,   0,   0,   0 } },
 { -8.683979433403e-06,  2.120549640896e-06, {   0,   2,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.686249997976e-07, -1.043244945475e-05, {   6,   0,  -3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -5.951891042808e-06,  4.848668381174e-06, {   0,   2,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  3.278596804723e-06, -7.521474618774e-06, {   2,   0,  -3,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.426110275082e-06,  8.371642361014e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 {  3.000902468956e-06, -7.793684286556e-06, {   4,   0,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.767268718330e-06,  3.026945148083e-06, {   3,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.518019139660e-06, -9.272021062724e-06, {   0,   0,   2,   0,   0, -26,  37, -16,   4,   5,   0,   0,   0 } },
 {  8.402909133450e-06,  2.385203312948e-06, {   1,   0,   0,   0,   0,   0,  35, -41,   2,   0,   0,   0,   0 } },
 { -4.338699200072e-06,  6.446262424009e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -7,   5,   0,   0,   0 } },
 { -3.974032914210e-06,  6.809250264585e-06, {   4,   0,  -2,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.492685272091e-06, -7.289693794253e-06, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -1.294810795841e-06,  9.486538465359e-06, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.075546189055e-06,  3.703023368080e-06, {   2,   0,  -1,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0 } },
 {  8.729237690572e-06, -2.049064105973e-06, {   3,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.413406091870e-06,  4.354964035519e-06, {   1,  -2,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.808978252277e-06,  6.952504869062e-06, {   1,   1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -3.690808631641e-06, -7.068892832710e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   6,  -5,   0,   0,   0 } },
 { -2.411289570539e-07,  1.051755046466e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 { -7.154970258681e-06,  3.602773004654e-06, {   2,   0,   1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 {  4.025436935403e-06, -6.729159377476e-06, {   2,   0,  -1,   0,   0,   0,  14, -13,   0,   0,   0,   0,   0 } },
 {  1.184093918530e-07, -1.063513561697e-05, {   6,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.897405985100e-06,  1.852587295073e-06, {   0,   0,   2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -6.203953452033e-06,  4.537866136969e-06, {   1,  -1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -9.584226857886e-06, -1.157558432096e-06, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  3.053878816472e-06,  7.684318332072e-06, {   1,   1,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 { -6.919511010877e-06, -3.815258751913e-06, {   0,   0,   1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.135696991200e-06,  7.580360770799e-06, {   0,   2,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.258310271400e-06,  2.454033530850e-06, {   0,   0,   1,   0,   0,  -3,  -1,   8,  -3,   0,   0,   0,   0 } },
 {  1.018196763628e-05,  5.278594507666e-07, {   3,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.527279576900e-06,  8.178789532900e-06, {   0,   0,   1,   0,   0,  -6,  10,   0,   0,   0,   0,   0,   0 } },
 { -3.216101025770e-07, -1.037813742046e-05, {   2,   0,  -1,   0,   0,  -2,   6,  -1,  -8,   6,   0,   0,   0 } },
 {  8.738818110681e-06, -1.960467535784e-06, {   0,   0,   2,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 {  1.460803872496e-07,  1.054821497203e-05, {   3,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  6.696352759589e-06, -3.995751429290e-06, {   2,   0,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0 } },
 { -1.421343085677e-06, -9.269199279493e-06, {   4,   0,  -2,   0,   0,  -2,  12, -15,   0,   0,   0,   0,   0 } },
 { -7.178642953170e-06,  3.508860909780e-06, {   1,   1,  -1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.718335393050e-06,  4.967678665612e-06, {   2,   0,  -1,   0,   1,   0, -11,  11,   0,   0,   0,   0,   0 } },
 {  7.639455950957e-06,  3.046472541317e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  3.456011006594e-06,  7.229503547393e-06, {   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.578999306504e-06, -4.102415871853e-06, {   6,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.017079977237e-05,  5.085254915706e-07, {   1,  -2,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  8.294948874812e-06,  2.382482938793e-06, {   2,   0,   0,   0,   0, -22,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.377995421917e-09,  1.067404200329e-05, {   0,   2,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -2.142127078931e-07,  1.045969007127e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0 } },
 {  3.126875597894e-06,  7.544501829625e-06, {   2,   0,   0,   0,   0,   0,  11, -18,   0,   0,   0,   0,   0 } },
 { -1.173664360111e-07, -1.054433612467e-05, {   0,   2,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  4.355478803670e-06, -6.306132344533e-06, {   0,   0,   1,   0,   0,   0,  12, -17,   0,   0,   0,   0,   0 } },
 { -3.225272513378e-07, -1.033513314139e-05, {   4,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  1.566787357533e-06,  9.085773110167e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0 } },
 {  1.524160302133e-06, -9.127349794471e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -6,   3,   0,   0 } },
 { -7.252994090438e-06, -3.397702854033e-06, {   1,   0,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0,   0 } },
 { -6.596533753271e-06, -4.051308753078e-06, {   4,   0,  -3,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -7.930653525851e-06, -2.712237306293e-06, {   2,   0,  -1,   0,   0,   0,   8, -12,   3,   0,   0,   0,   0 } },
 { -1.022805563661e-06, -9.619857907610e-06, {   4,   0,  -2,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -6.125589508670e-06, -4.515215572511e-06, {   0,   0,   4,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.277862692250e-06, -2.361813995824e-06, {   1,   1,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -7.381050673799e-06, -3.254866374895e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   4,   0,   0,   0 } },
 {  3.242150222067e-06, -7.389060683111e-06, {   3,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.146229260955e-06, -2.480120547093e-06, {   4,   0,  -3,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -6.116461103112e-06,  4.509664381741e-06, {   0,   0,   4,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.567706527516e-06,  3.057595383766e-06, {   2,   0,  -1,   0,   0,   0,   6,  -8,   3,   0,   0,   0,   0 } },
 { -7.702507615161e-06, -2.922361931585e-06, {   0,   1,   0,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  8.471292828299e-06, -2.150996208096e-06, {   2,  -2,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.661130570614e-06,  3.959748036530e-06, {   0,   0,   0,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0 } },
 { -3.495628194029e-06,  7.122655015668e-06, {   4,   0,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.173369654618e-06,  7.436781278178e-06, {   2,   0,  -2,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 { -3.665298183077e-06, -6.941945150168e-06, {   2,   0,   2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  8.587557933017e-06,  2.018704635522e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,   2,   0,   0,   0 } },
 { -2.871708565802e-06,  7.725266433966e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   5,  -3,   0,   0,   0 } },
 {  7.604948370111e-06, -2.990310097591e-06, {   2,   0,  -2,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  6.270811286346e-08,  1.052935063793e-05, {   0,   0,   2,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -7.643397590152e-06,  2.943976427013e-06, {   2,   0,   0,   0,   0,   0,   1,  -4,   1,   0,   0,   0,   0 } },
 { -3.493778605449e-06, -7.092258006514e-06, {   1,   1,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  8.345849586795e-06,  2.238939815119e-06, {   1,   1,  -1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.934416266532e-06,  4.649348883561e-06, {   2,  -1,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.468360853010e-06, -9.111998408689e-06, {   0,   0,   2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  7.026526161163e-06, -3.552335681257e-06, {   2,   0,  -3,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -4.243416477532e-06, -6.334417335043e-06, {   0,   0,   2,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  9.207385363274e-06,  1.369739692705e-06, {   2,   0,  -1,   0,   2,   0,  -6,   0,   0,   0,   0,   0,   0 } },
 {  9.170791357070e-06, -1.405876114385e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   4,  -2,   0,   0,   0 } },
 {  5.659946925178e-06, -4.911503102971e-06, {   3,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  2.844835058424e-07, -1.028332793976e-05, {   2,   0,  -1,   0,   0,  -4,   6,   0,  -2,   0,   0,   0,   0 } },
 {  1.797154489026e-06, -8.769875771104e-06, {   2,   0,  -2,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 { -8.157793211396e-06, -2.406370319351e-06, {   0,   0,   1,   0,   0,  -3,   7,  -8,   3,   0,   0,   0,   0 } },
 { -6.545710498320e-06,  4.011015336920e-06, {   4,   0,  -3,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.962292432653e-06, -8.592786913150e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0 } },
 { -4.320061479402e-06,  6.229937418434e-06, {   4,   0,  -2,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  9.031383091884e-06,  1.517459392192e-06, {   1,  -1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -2.613212622549e-06,  7.933228548322e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -1.734578009676e-06,  8.806257751799e-06, {   3,  -1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  7.403600392210e-06,  3.134754227574e-06, {   1,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.115257495269e-06, -3.422296484833e-06, {   2,   0,  -1,   0,   0,   0, -10,  12,   0,   0,   0,   0,   0 } },
 { -5.682148009543e-06,  4.853170884526e-06, {   0,   0,   0,   0,   0,  10,  -6,   0,   0,   0,   0,   0,   0 } },
 { -5.941096486208e-06, -4.594014684473e-06, {   2,   0,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0 } },
 { -2.058262045171e-06, -8.476646929072e-06, {   1,  -1,  -3,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -6.577069629493e-06, -3.956407113271e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -1.525506786819e-06, -9.007472156786e-06, {   1,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -4.056333490250e-06, -6.461740315085e-06, {   2,  -2,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 { -8.510282238287e-06,  2.007098152261e-06, {   0,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.010698105949e-05, -4.070407248707e-07, {   1,  -2,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.208242211200e-06,  6.304889378414e-06, {   0,   0,   1,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 { -5.894424389448e-06, -4.617351388692e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0 } },
 {  4.589566684830e-07,  1.004969469467e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   2,   0,   0 } },
 {  1.002560525611e-05,  4.809134892187e-07, {   0,   0,   0,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -6.441509822454e-06, -4.062616740530e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   3,  -5,   0,   0,   0 } },
 { -4.932736445018e-06,  5.570020546492e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -2.675289020081e-08,  1.046447846382e-05, {   1,   0,   1,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.360891632794e-06, -9.129662294296e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,   2,  -8,   0,   0,   0 } },
 {  1.639028913680e-09,  1.047491504558e-05, {   2,   0,   3,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.200051645789e-06, -7.276325290326e-06, {   1,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.076481734615e-06, -4.394608684130e-06, {   2,   0,   0,   0,   0,   0,  -3,   6,  -1,   0,   0,   0,   0 } },
 { -2.318266185416e-06, -8.150586077017e-06, {   0,   0,   3,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.439596743785e-06,  7.028974440708e-06, {   4,   0,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -8.705700716187e-06, -1.757388211304e-06, {   2,   0,  -1,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 { -3.848636582489e-06,  6.611541395749e-06, {   4,   0,  -1,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.862430036142e-06,  3.596768614858e-06, {   0,   0,   1,   0,   0,  19, -22,   3,   0,   0,   0,   0,   0 } },
 { -5.100935522252e-06, -5.349092447387e-06, {   2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.299596754004e-06,  9.149914458653e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   6,  -3,   0,   0 } },
 { -2.530644511575e-09, -1.044544001842e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,   0,  -2,   0,   0 } },
 {  5.720513824290e-06, -4.724683549657e-06, {   2,   0,   0,   0,   0,  -8,  17,  -8,   3,   0,   0,   0,   0 } },
 { -1.009320068931e-05,  3.517654919038e-07, {   1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.287825267632e-06,  4.156606101233e-06, {   1,  -1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.205947794556e-06, -9.236124962721e-06, {   0,   0,   1,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  1.312763443463e-06, -9.120338866015e-06, {   0,   0,   1,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 {  8.257653041702e-06, -2.166117291960e-06, {   0,   0,   1,   0,   0,  13, -16,   0,   0,   0,   0,   0,   0 } },
 { -3.728306214241e-06,  6.694287606650e-06, {   1,   0,   2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.977642801510e-06, -8.439242580688e-06, {   4,  -2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.478523431725e-06,  8.938345534036e-06, {   4,   0,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  8.363014053302e-06,  2.052974484489e-06, {   1,   1,   1,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.063325862483e-06,  1.348887978893e-06, {   0,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.133330078067e-07, -9.698347273636e-06, {   0,   0,   1,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 {  1.503265092037e-08,  1.039592626253e-05, {   2,   0,  -3,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -7.968134908957e-06,  2.439833089310e-06, {   2,   0,  -1,   0,   0,   4,  -5,   0,   3,   0,   0,   0,   0 } },
 { -1.338806105407e-07,  1.027396557916e-05, {   3,   0,   1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -7.284273524283e-06, -3.122997103191e-06, {   2,  -2,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.655057914163e-06, -8.752059973190e-06, {   4,  -2,   1,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.655057914163e-06,  8.752059973190e-06, {   0,   2,  -1,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 { -8.412123081215e-06,  1.994355462499e-06, {   3,  -1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.843055550880e-07, -9.921864776168e-06, {   4,   0,  -2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -7.609515052474e-06,  2.793119213237e-06, {   0,   0,   2,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -4.686336911165e-06, -5.716115840528e-06, {   2,   0,   0,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 {  8.934098809860e-06, -1.466583566922e-06, {   2,   0,  -1,   0,   0,  18, -21,   0,   0,   0,   0,   0,   0 } },
 { -8.085376926199e-06,  2.312338214646e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,  -2,   0 } },
 {  3.613811160793e-06,  6.783704074909e-06, {   0,   0,   0,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  3.568878366306e-06,  6.827977507738e-06, {   2,  -2,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  3.435268681770e-06, -6.959223466399e-06, {   4,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  1.473462107442e-06, -8.920087136046e-06, {   4,   0,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -8.366674564150e-06,  2.024460584119e-06, {   0,   2,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.846520618406e-06, -8.544581311997e-06, {   0,   0,   3,   0,   0, -20,  19,   0,   3,   0,   0,   0,   0 } },
 {  1.846520618406e-06,  8.544581311997e-06, {   0,   0,   1,   0,   0,  20, -19,   0,  -3,   0,   0,   0,   0 } },
 { -9.477559569580e-06, -9.107139379202e-07, {   3,  -1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  6.450560661450e-06, -3.937425113579e-06, {   1,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.206170776358e-06,  7.180039661569e-06, {   2,   0,  -1,   0,  -1,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -3.250323024727e-06,  7.134313451741e-06, {   2,   0,  -3,   0,   0,  -3,   7,   0,  -4,   0,   0,   0,   0 } },
 { -3.457149640563e-07,  1.003694095557e-05, {   4,   0,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.694558910981e-06,  4.685567628008e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 {  7.198366122819e-07, -9.658032180830e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  1.790946503282e-08,  1.035956068274e-05, {   3,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  2.424720116902e-08, -1.035036269168e-05, {   1,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  1.995549984597e-08, -1.035203669064e-05, {   4,   0,   1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.381526844390e-06, -1.983359091381e-06, {   3,  -3,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.189428549393e-07, -1.024311268181e-05, {   2,   0,  -1,   0,   0,  18, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.282838792821e-06, -9.077190547966e-06, {   0,   0,   1,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 {  3.364417677747e-06, -6.992873696813e-06, {   2,  -2,   0,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 { -8.600435222864e-06,  1.756465694175e-06, {   2,   0,  -1,   0,   0,   0,  -5,  10,  -4,   0,   0,   0,   0 } },
 { -4.566661910637e-06, -5.786398275206e-06, {   2,   0,   1,   0,   0, -23,  22,   4,   0,   0,   0,   0,   0 } },
 { -4.566661910637e-06,  5.786398275206e-06, {   2,   0,  -1,   0,   0,  23, -22,  -4,   0,   0,   0,   0,   0 } },
 { -1.838859997472e-06, -8.512009251403e-06, {   0,   0,   2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.620073451545e-06, -3.728911965390e-06, {   2,   0,   0,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0 } },
 {  9.050305162145e-06,  1.297751244642e-06, {   2,   0,   1,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 {  8.193575440836e-06, -2.153886694186e-06, {   1,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  6.695610418998e-06, -3.647651736541e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -4,   0,   0,   0 } },
 {  2.566712493681e-06,  7.776322424902e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  3.906970042658e-06, -6.433506099582e-06, {   2,   0,   0,   0,   0,   0,   5,   0,  -8,   0,   0,   0,   0 } },
 {  9.190734652901e-06, -1.149166466869e-06, {   2,   0,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  1.003096158908e-05,  3.064936952647e-07, {   1,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.103550441258e-06, -7.232996976441e-06, {   1,  -1,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.793443253162e-06,  8.541516136901e-06, {   1,  -1,   1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.696239865229e-06,  4.637248142829e-06, {   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -1.122620281591e-06,  9.204636462972e-06, {   2,   0,   0,   0,   0,   8, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.078217414924e-06,  6.248261529975e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   3,   0,   0,   0 } },
 { -3.414829148042e-06,  6.909820538129e-06, {   0,   0,   1,   0,   0,  19, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.371288858694e-06,  1.952383484078e-06, {   1,  -1,   1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.400828204759e-06, -8.922584196587e-06, {   2,   0,   0,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  5.600110478872e-07,  9.762777742730e-06, {   2,   0,   1,   0,   0,  14, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.653604000337e-08, -1.029190047011e-05, {   4,   0,   0,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -1.934325835805e-06,  8.380868493183e-06, {   2,   0,   2,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.666292765792e-06, -6.647537678805e-06, {   2,   2,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -5.602001286860e-06, -4.711005653723e-06, {   1,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  7.684504764087e-06,  2.625762359268e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   1,   5,   0,   0,   0 } },
 {  6.449203018106e-09, -1.030251034578e-05, {   4,   0,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.661443096303e-06,  8.647014247906e-06, {   4,   0,  -3,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -3.787416713493e-06,  6.520252142806e-06, {   2,   0,   0,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0 } },
 {  9.427812654835e-06,  8.753205694707e-07, {   2,   0,  -1,   0,   0,   0,  -7,  15,   0,   0,   0,   0,   0 } },
 { -2.327296731392e-06, -7.969108403838e-06, {   2,   0,  -3,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.785521921056e-06, -8.502139069703e-06, {   1,   1,  -1,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.869544616050e-06,  3.410628743311e-06, {   2,   0,   0,   0,   0,  -2,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.260581491250e-08, -1.025717655220e-05, {   4,   0,  -1,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.491565415857e-06,  4.787175919403e-06, {   0,   0,   1,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 { -7.175245940378e-06, -3.102797197613e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,  -4,   0,   0,   0 } },
 { -8.296505971520e-06, -1.981351203245e-06, {   1,   1,  -3,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.637869866452e-06,  5.638557686707e-06, {   2,   0,   1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -1.200864707662e-07, -1.015498723329e-05, {   2,   0,  -1,   0,   0,  -2,   4,   0,  -2,   0,   0,   0,   0 } },
 { -3.386472716489e-06,  6.880934412149e-06, {   1,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.679248429386e-06, -7.587994865867e-06, {   2,   0,   0,   0,   0,  12, -16,   0,   0,   0,   0,   0,   0 } },
 { -9.199875775337e-06, -1.062327013496e-06, {   1,   1,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.850294164078e-06, -1.411741932522e-06, {   0,   0,   1,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.177204910529e-06,  7.082832033678e-06, {   2,   0,   0,   0,   0,   0, -11,  16,   0,   0,   0,   0,   0 } },
 { -1.306316506982e-06, -8.952189218405e-06, {   2,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  9.071360971898e-08,  1.016366123883e-05, {   2,   0,  -2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.158956352266e-06, -7.090068295245e-06, {   2,   0,  -1,   0,  -5,  14,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.506591071537e-06, -5.742323230051e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,  -3,   0,   0,   0 } },
 {  9.461284392316e-06,  7.869645132925e-07, {   4,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.156493363856e-06, -4.081800450256e-06, {   1,   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  6.219336445684e-06, -4.016482500558e-06, {   2,   0,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0 } },
 { -1.480103938316e-08, -1.021804880097e-05, {   2,   0,  -3,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.927178994295e-06,  7.305413485988e-06, {   3,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -5.710267441073e-06,  4.522039674261e-06, {   2,   0,  -2,   0,   0,  15, -11,   0,   0,   0,   0,   0,   0 } },
 { -8.511466100374e-06,  1.719608801118e-06, {   2,   0,   0,   0,   0,   0,  -5,  10,  -4,   0,   0,   0,   0 } },
 {  3.391509462607e-06,  6.838483705761e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  6.825191936388e-08,  1.016111937983e-05, {   0,   0,   3,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  7.883069344582e-06,  2.341541569204e-06, {   2,   0,  -1,   0,   0,   4,   0,  -8,   3,   0,   0,   0,   0 } },
 { -5.156857755737e-07,  9.706884081357e-06, {   2,   0,  -2,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -8.881343076033e-07,  9.333498262693e-06, {   2,   0,   0,   0,   0,   0,  18, -24,   0,   0,   0,   0,   0 } },
 { -3.769458927996e-06,  6.450741973212e-06, {   0,   0,   2,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  6.956914941706e-09,  1.020891876338e-05, {   6,   0,  -3,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.703154418749e-06, -8.511814627715e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  3.383716113316e-06,  6.830602734149e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -2,  -1,   0,   0,   0 } },
 {  7.072302473351e-06, -3.141866756083e-06, {   1,   1,  -2,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  7.363282176394e-06, -2.840452288353e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -6.421352740050e-06, -3.777882565763e-06, {   2,   0,   1,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0 } },
 {  7.864184350955e-06, -2.320741766936e-06, {   2,   0,  -1,   0,   0,   4,  -8,   8,  -3,   0,   0,   0,   0 } },
 { -8.450506198533e-10,  1.018122168601e-05, {   3,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.273118018321e-06,  3.903533667500e-06, {   3,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  7.930387143555e-06,  2.241409052019e-06, {   0,   2,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -9.774010636060e-06,  3.966048601211e-07, {   4,  -2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.032055012162e-06,  6.138414050796e-06, {   2,   0,  -1,   0,   2,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.284131726247e-06, -4.884735873571e-06, {   2,   0,  -1,   0,   0,   0,  -5,   0,   2,   0,   0,   0,   0 } },
 { -3.029616905193e-08, -1.013681891815e-05, {   4,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.231493842098e-06,  1.933772233606e-06, {   0,   0,   2,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 { -6.401796958461e-06,  3.762878772441e-06, {   1,  -1,   2,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -2.400322786265e-06,  7.763742284619e-06, {   2,   0,  -1,   0,   0,   0,  13, -23,   0,   0,   0,   0,   0 } },
 {  7.005111168979e-06,  3.158274960441e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  4.494119471226e-09,  1.015626809756e-05, {   3,   0,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.265995662748e-06, -2.894624125063e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -4.042791757058e-06,  6.113217844119e-06, {   2,   0,   0,   0,   0,   0,  21, -26,   0,   0,   0,   0,   0 } },
 { -4.215761743205e-06,  5.938070590006e-06, {   2,  -2,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  6.682505596197e-06, -3.468807815646e-06, {   2,   0,   1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  8.650830106958e-06, -1.500070670358e-06, {   4,   0,  -2,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -3.542918268019e-06, -6.603119573971e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0 } },
 { -3.540638545868e-06,  6.601736738734e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0 } },
 {  3.542005575315e-06,  6.599937086530e-06, {   0,   0,   2,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.240929777355e-07,  9.217417546382e-06, {   1,  -2,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.910600587448e-06,  6.230835252844e-06, {   2,  -2,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.606881990079e-06, -5.531482077636e-06, {   4,   0,  -3,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  6.637282365611e-06, -3.499659064484e-06, {   2,   0,  -1,   0,   0,   0,  20, -24,   0,   0,   0,   0,   0 } },
 { -3.135511417769e-06,  6.997145610074e-06, {   2,   0,  -3,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -3.271235714976e-06, -6.858139316081e-06, {   2,   0,  -2,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.580986361476e-06,  5.473311740149e-07, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -1.554077578778e-07,  9.972321957212e-06, {   4,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.657996712293e-06,  3.467769267194e-06, {   2,   0,   1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  2.781515643118e-06, -7.343939264651e-06, {   0,   2,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.578806523202e-06, -6.546005770772e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -5.907465186010e-06,  4.216519294381e-06, {   4,   0,  -3,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.776076320567e-09,  1.012062128453e-05, {   2,  -2,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.207612062739e-06,  8.910066938988e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  2.036386522087e-08,  1.009341911782e-05, {   2,   0,  -1,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 {  7.878039833710e-06, -2.232725295180e-06, {   2,   0,  -3,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  8.658710693633e-07, -9.243573539366e-06, {   2,   0,  -1,   0,   0,   0,  18, -24,   0,   0,   0,   0,   0 } },
 {  2.643848167996e-06, -7.464792454966e-06, {   4,   0,  -1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.658528803407e-07,  9.542291174040e-06, {   2,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.917816317913e-06, -1.190018114891e-06, {   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -2.055962919092e-06, -8.051258370770e-06, {   1,   1,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.995492515610e-06,  7.111290091017e-06, {   1,  -1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  1.800734488064e-06, -8.300554250067e-06, {   0,   0,   1,   0,   0,   0,   3,  -9,   0,   0,   0,   0,   0 } },
 { -4.062450765149e-06,  6.038508764486e-06, {   2,   0,  -2,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 { -9.598299485263e-06, -4.952695113940e-07, {   1,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.661795954679e-06,  2.430763107958e-06, {   0,   0,   2,   0,   0, -21,  23,  -4,   0,   0,   0,   0,   0 } },
 {  7.064913565520e-06, -3.027270002296e-06, {   3,  -3,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  8.161033389350e-06,  1.928077375247e-06, {   1,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.618896298872e-07, -9.826464547024e-06, {   2,   0,   0,   0,   0,   0,  -6,   0,   7,   0,   0,   0,   0 } },
 {  7.511414732721e-06,  2.574498704972e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -6.130204848836e-06,  3.955266330172e-06, {   0,   0,   4,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  8.154565771770e-06,  1.929424787288e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   2,   0,   0 } },
 { -3.570690304494e-06, -6.509835211782e-06, {   1,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -9.558449063380e-06, -5.181594498507e-07, {   4,   0,  -2,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -8.406038749509e-06, -1.669736476004e-06, {   1,  -1,   1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  8.406038749509e-06,  1.669736476004e-06, {   1,  -1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -6.099379466082e-06,  3.973938801106e-06, {   2,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -8.156483636403e-06, -1.913150931306e-06, {   1,   1,  -1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 { -8.229834529482e-06, -1.837284104114e-06, {   4,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.105749519927e-06,  3.949096523700e-06, {   2,   0,  -1,   0,   0,   0,  -5,  10,  -3,   0,   0,   0,   0 } },
 { -2.829705747499e-06, -7.223557229573e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   6,  -5,   0,   0,   0 } },
 { -2.136414421851e-06,  7.916623921781e-06, {   2,   0,  -1,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.576001381894e-08, -1.003356462405e-05, {   1,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.857734960987e-06, -6.191569998273e-06, {   0,   0,   1,   0,   0,  -3,  18, -25,   2,   0,   0,   0,   0 } },
 { -3.859035593875e-06,  6.189883252152e-06, {   0,   0,   1,   0,   0,   3, -18,  25,  -2,   0,   0,   0,   0 } },
 {  1.658173038101e-07, -9.875450853286e-06, {   0,   2,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -8.121373401591e-06,  1.918216015126e-06, {   3,   1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.095213693167e-06,  6.943664181238e-06, {   2,   0,  -1,   0,   5, -14,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.740988887347e-06,  2.964623953234e-07, {   4,   0,   1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -7.170442678652e-06,  2.863486480996e-06, {   1,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.298766617654e-06,  8.730404736645e-06, {   1,  -1,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.559486112568e-06,  2.468095653554e-06, {   0,   0,   1,   0,   0,   0,  -2,   2,   0,  -2,   0,   0,   0 } },
 {  7.876890666206e-06, -2.150112016502e-06, {   0,   0,   1,   0,   0,  -6,  10,   0,  -5,   5,   0,   0,   0 } },
 {  1.530455974768e-07, -9.872873188143e-06, {   3,  -2,   0,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 {  1.530455974768e-07,  9.872873188143e-06, {   1,   2,   0,   0,   0, -11,   5,  -3,   0,   0,   0,   0,   0 } },
 {  9.637190719764e-06,  3.870552366173e-07, {   0,   0,   0,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 {  7.870954912273e-06,  2.148151799969e-06, {   0,   0,   1,   0,   0,   6, -10,   0,   5,  -5,   0,   0,   0 } },
 {  9.455797598201e-06,  5.590733457579e-07, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.536554162024e-06,  6.473072806447e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0 } },
 { -7.828071545773e-06,  2.179572348483e-06, {   3,   0,   0,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -7.827874897641e-06, -2.179552883770e-06, {   3,   0,   0,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 { -6.069410681943e-06,  3.937843867837e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -6,   0,   0,   0 } },
 {  5.977638501502e-07,  9.408536037697e-06, {   1,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  2.203486767521e-06, -7.802088891939e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 { -2.675198554444e-06, -7.324367006315e-06, {   0,   2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.599095695940e-06, -3.989232097169e-07, {   4,  -2,   2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.762258781071e-06,  5.233420709200e-06, {   2,  -2,   1,   0,   0,   0,  17, -28,   0,   0,   0,   0,   0 } },
 {  4.762258781071e-06, -5.233420709200e-06, {   2,  -2,  -1,   0,   0,   0,  17, -28,   0,   0,   0,   0,   0 } },
 {  6.592754498996e-06, -3.401771777227e-06, {   0,   0,   1,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0 } },
 {  6.880430648991e-06,  3.111403032919e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -7,   0,   0,   0 } },
 {  1.463030124109e-06,  8.527195122813e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -7,  11,  -3,   0,   0 } },
 { -7.786306717263e-06,  2.203192405497e-06, {   0,   0,   2,   0,   0, -15,  10,   4,   0,   0,   0,   0,   0 } },
 { -1.192116897110e-06, -8.795895281225e-06, {   1,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  6.863209967961e-06, -3.124765523157e-06, {   0,   2,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -4.301753635034e-06, -5.684376695294e-06, {   2,   0,   0,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0 } },
 { -3.280239800858e-06, -6.701215095343e-06, {   2,   0,  -2,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -3.156660764867e-06,  6.824479466347e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   5,  -3,   0,   0,   0 } },
 {  8.474027168895e-06, -1.506896437503e-06, {   4,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.732502600282e-06, -6.248258312570e-06, {   3,   1,  -2,   0,   0, -12,   7,   0,   0,   0,   0,   0,   0 } },
 { -3.207042681871e-06,  6.770651262452e-06, {   0,   0,   0,   0,   0,  24, -22,   0,   0,   0,   0,   0,   0 } },
 {  2.598369503466e-07,  9.716165056349e-06, {   1,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  5.679393081267e-09, -9.969718373971e-06, {   3,   0,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  8.072611859813e-06, -1.898578998756e-06, {   1,  -1,  -2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.051724547510e-06, -1.919041523480e-06, {   1,  -1,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.626250757142e-07,  9.807331574259e-06, {   5,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -5.782425723250e-06, -4.176004362166e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   1,   4,   0,   0,   0 } },
 { -9.567876124760e-06,  3.878571754148e-07, {   0,   2,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.410370529406e-06,  4.541727188472e-06, {   1,  -1,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  2.777112046453e-06, -7.170775399060e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   2,   2,   0,   0,   0 } },
 {  8.030164938463e-06, -1.917703228304e-06, {   3,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.989427425500e-06,  7.954706140361e-06, {   1,   3,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  9.205647447661e-06,  7.346615647431e-07, {   2,   0,   0,   0,   0,  24, -23,   0,   0,   0,   0,   0,   0 } },
 {  6.062940334741e-06,  3.875353742972e-06, {   0,   0,   2,   0,   0, -12,   3,   7,  -8,   6,   0,   0,   0 } },
 {  6.062940334741e-06, -3.875353742972e-06, {   0,   0,   0,   0,   0,  12,  -3,  -7,   8,  -6,   0,   0,   0 } },
 { -9.867699147135e-06,  7.035301074197e-08, {   2,   0,  -1,   0,   0,  20, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.094430862917e-07, -9.826440644194e-06, {   2,   0,   0,   0,   0,  31, -31,   0,   0,   0,   0,   0,   0 } },
 { -8.077041071375e-06, -1.856482416748e-06, {   1,  -1,  -3,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -8.422149366776e-06, -1.504613300113e-06, {   1,  -1,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.967772838911e-06, -7.958913535571e-06, {   2,  -1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.068204364471e-06,  5.858371843806e-06, {   2,   0,  -2,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 { -1.788025631466e-06,  8.137146371945e-06, {   3,   0,  -5,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -8.022357700517e-06, -1.901874778694e-06, {   3,  -1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.164787073866e-06, -7.524548748930e-07, {   0,   0,   1,   0,   0,   0,  10, -19,   2,  -2,   0,   0,   0 } },
 { -9.164787073866e-06,  7.524548748930e-07, {   0,   0,   1,   0,   0,   0, -10,  19,  -2,   2,   0,   0,   0 } },
 {  1.187709495073e-06, -8.728633716054e-06, {   4,   0,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  3.370851496273e-06,  6.544537379985e-06, {   2,  -2,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 { -1.914988007449e-07,  9.723804405158e-06, {   4,   0,  -1,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.812652242025e-07, -9.731204917185e-06, {   1,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.041737959555e-07,  9.807767088428e-06, {   1,   1,   0,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 { -9.626469537401e-06,  2.810865877806e-07, {   2,   0,   2,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -8.470933529404e-06,  1.432644108491e-06, {   4,   0,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  9.777409882842e-06, -1.172886303282e-07, {   1,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.738695593817e-06, -4.154215386734e-06, {   1,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  2.614732023489e-07, -9.630212742837e-06, {   0,   0,   1,   0,   0,  -4,   6,   0,  -2,   0,   0,   0,   0 } },
 { -4.421272825368e-06, -5.467157940671e-06, {   0,   0,   1,   0,   0,   0,  -8,   7,   0,   0,   0,   0,   0 } },
 {  9.845627951164e-06, -3.963480410502e-08, {   2,   0,  -2,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  6.527758257580e-06,  3.356595730006e-06, {   2,   0,   0,   0,   0, -20,  18,   6,  -2,   0,   0,   0,   0 } },
 { -2.515870735384e-09,  9.881683315093e-06, {   6,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  7.414990433998e-06,  2.466350825928e-06, {   1,   0,   1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.733543592497e-06,  8.144277083016e-06, {   1,  -1,   0,   0,   0,  18, -20,   0,   0,   0,   0,   0,   0 } },
 { -7.982785644416e-06, -1.892743767459e-06, {   2,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.788249264043e-06,  8.727472819539e-08, {   4,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.993114624223e-06, -7.882408731953e-06, {   1,  -1,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  9.752384797896e-06,  1.170006725370e-07, {   2,   0,   1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.231669302061e-06, -6.634177201037e-06, {   2,   0,  -2,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  7.979600664458e-06,  1.883331892668e-06, {   1,   1,   1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.858963868471e-06,  4.998380321545e-06, {   2,   0,  -1,   0,   0,   0,   1,  -4,   3,   0,   0,   0,   0 } },
 { -2.786612325861e-06,  7.070632818543e-06, {   4,   0,  -2,   0,   0,   0,  13, -23,   0,   0,   0,   0,   0 } },
 { -3.466488398807e-06, -6.389815234600e-06, {   4,   0,  -2,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  1.412996912889e-06,  8.436455776126e-06, {   0,   2,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  3.025981910859e-06, -6.820977315525e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 { -1.074875503875e-08,  9.831905120741e-06, {   1,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.865955623191e-06, -1.971031796390e-06, {   1,  -1,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.902879368456e-06,  6.932093926620e-06, {   2,  -2,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -4.425303025646e-06,  5.406149420243e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  6.969120021727e-06,  2.858709718784e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  4.482413759268e-06, -5.344495066491e-06, {   3,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -6.401315291752e-06,  3.425131138567e-06, {   0,   2,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -8.183274852877e-06, -1.641984865381e-06, {   2,   0,   0,   0,   0,   0,   5, -10,   4,   0,   0,   0,   0 } },
 { -6.573451010905e-06,  3.249338738191e-06, {   2,   0,  -2,   0,   0,   0,  11, -15,   0,   0,   0,   0,   0 } },
 {  7.292875243675e-06, -2.527360672331e-06, {   1,  -1,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -2.847444479533e-08, -9.789430126616e-06, {   0,   2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.260728774111e-06, -2.552358962680e-06, {   4,  -2,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.607584096731e-08, -9.755226429037e-06, {   5,   0,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.914896507610e-06,  8.963329131064e-07, {   2,  -2,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.614307595317e-06,  8.195331870345e-06, {   1,   1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.927502705900e-06, -1.872799937898e-06, {   3,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.489568477257e-06, -2.310316651085e-06, {   0,   0,   0,   0,   0,  18, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.506334966394e-06,  4.292764642229e-06, {   2,   0,  -2,   0,   0,   0,  13, -21,   0,   0,   0,   0,   0 } },
 { -9.424040640571e-06, -3.749745096824e-07, {   2,  -2,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.703258961164e-06, -8.095096882012e-06, {   1,   1,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.955020416419e-06,  5.842544179411e-06, {   1,   0,   1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.275441027556e-06,  6.520055394977e-06, {   1,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.722813791851e-06, -8.071545655850e-06, {   1,  -1,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.890769490422e-06, -1.903151322152e-06, {   0,   0,   3,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.175932651348e-08, -9.774156096121e-06, {   1,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -7.472391321582e-06, -2.309592841675e-06, {   2,   0,  -1,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0 } },
 { -8.705312048305e-06,  1.063278674052e-06, {   2,   0,   0,   0,   0,  21, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.387583934215e-06, -8.380229293629e-06, {   1,   0,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 { -5.391428072811e-07,  9.225671353284e-06, {   0,   0,   1,   0,   0,   0,   3,  -4,   1,   0,   0,   0,   0 } },
 { -9.690372459295e-06, -7.390700155540e-08, {   0,   2,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.097353368502e-07, -8.845641829559e-06, {   2,   0,  -2,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  7.926691782616e-06, -1.828497806402e-06, {   1,   1,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.243499889013e-06,  6.505849952573e-06, {   2,   0,   0,   0,   0,  22, -20,   0,   0,   0,   0,   0,   0 } },
 {  8.890210283520e-07, -8.859495962253e-06, {   0,   0,   2,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -1.541432992174e-06,  8.205734799185e-06, {   2,   0,   1,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.058886577230e-06, -3.687737584376e-06, {   2,  -1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -3.493341162323e-06, -6.248154391486e-06, {   0,   0,   2,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 { -9.185541938354e-06, -5.490100304946e-07, {   2,   0,  -1,   0,   0,   0,   9, -14,   2,   0,   0,   0,   0 } },
 {  9.650622281515e-06, -8.381417921098e-08, {   1,   0,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.252062637755e-06,  5.481822015238e-06, {   2,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.171506990546e-06,  6.561721393585e-06, {   2,   0,   1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.694622944243e-06, -3.834302777008e-08, {   2,   0,   0,   0,   0,   0,  20, -23,   0,   0,   0,   0,   0 } },
 { -5.491661134630e-06, -4.236605520404e-06, {   1,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0 } },
 {  3.210309665177e-06, -6.515151311001e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   1,   0,   0,   0 } },
 {  1.327307999693e-07, -9.583693926406e-06, {   4,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 { -5.851210665743e-06, -3.864776403923e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -7.855211541982e-06,  1.858495373423e-06, {   1,   1,  -2,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.463904277158e-06,  6.244501101430e-06, {   4,   0,  -2,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0 } },
 { -9.012208464360e-06, -6.953104998116e-07, {   0,   0,   1,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0 } },
 {  2.049138905313e-08,  9.682830979510e-06, {   3,  -2,  -1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.880910368561e-06, -5.820675538654e-06, {   2,   0,  -1,   0,   0,   0,  20, -25,   0,   0,   0,   0,   0 } },
 {  6.036432057461e-06,  3.660269202634e-06, {   2,   0,  -3,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.367323797935e-06,  8.325063363066e-06, {   1,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  9.408512148995e-06,  2.821571814389e-07, {   2,   0,  -2,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.070825165742e-06,  6.616024195873e-06, {   2,   0,   1,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 { -7.520094815239e-06, -2.164313534830e-06, {   2,   0,   0,   0,   0, -24,  27,   0,   0,   0,   0,   0,   0 } },
 {  6.968210131640e-06, -2.712313104399e-06, {   2,   0,   2,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  7.633513226995e-06, -2.046674180093e-06, {   3,  -1,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  9.315049186866e-06,  3.627102657944e-07, {   2,   0,   0,   0,   0,  30, -31,   0,   0,   0,   0,   0,   0 } },
 {  5.100853371509e-07,  9.167673528027e-06, {   4,  -2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  6.073079720891e-06, -3.602000033884e-06, {   2,  -2,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  7.825151697264e-06,  1.847799109285e-06, {   3,  -1,   0,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.973786868868e-06, -4.697477926396e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   6,  -4,   0,   0,   0 } },
 { -7.821699593490e-06, -1.848302073617e-06, {   1,  -1,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.893616774481e-06,  5.775657874159e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,  -1,   3,   0,   0 } },
 { -2.998667164966e-06, -6.670070105008e-06, {   0,   0,   1,   0,   0,   0,   6, -11,  -2,   0,   0,   0,   0 } },
 {  3.925254884430e-06, -5.743020499352e-06, {   0,   0,   1,   0,   0,   0,  -1,   4,  -2,   0,   0,   0,   0 } },
 { -9.402836469938e-06, -2.635102010334e-07, {   2,   0,  -4,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.894527593769e-06, -6.769002319906e-06, {   1,  -1,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.797733815051e-06,  7.862673532196e-06, {   0,   0,   1,   0,   0, -16,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.867767449176e-06, -7.790622672775e-06, {   1,  -1,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.600439789128e-06, -6.057110111943e-06, {   2,   0,   0,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0 } },
 { -6.956854214974e-06, -2.698934776490e-06, {   2,   0,  -2,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.009543236683e-06, -1.646147353163e-06, {   3,  -3,   1,   0,   0,  43, -42,   0,   0,   0,   0,   0,   0 } },
 { -8.009543236683e-06,  1.646147353163e-06, {   1,   3,   1,   0,   0, -43,  42,   0,   0,   0,   0,   0,   0 } },
 { -3.092072529648e-06, -6.562142969381e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   8,   0,   0,   0 } },
 { -2.995661020692e-06,  6.657295357927e-06, {   0,   0,   1,   0,   0,   0,  -6,  11,   2,   0,   0,   0,   0 } },
 {  9.355531964288e-06,  2.972356812415e-07, {   2,  -2,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.711795357635e-06,  9.380740340486e-07, {   1,  -1,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  8.605505071487e-08,  9.558366956290e-06, {   0,   0,   4,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.808303996587e-06, -1.833001201545e-06, {   1,   1,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.139678627235e-06,  2.500345330716e-06, {   3,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.928717318185e-08, -9.618550060124e-06, {   3,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.617327985837e-06,  3.020411602955e-06, {   2,   0,  -1,   0,   0,   0,  10,  -8,   0,   0,   0,   0,   0 } },
 {  3.556074847077e-07,  9.281914380960e-06, {   0,   0,   0,   0,   0,   0,  10, -10,   0,   0,   0,   0,   0 } },
 {  3.814899546878e-06,  5.818142264587e-06, {   2,   0,   0,   0,   2,  -7,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.755381034210e-06, -6.873966006498e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  2.755381034210e-06,  6.873966006498e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,  -2,   3,   0,   0,   0 } },
 {  8.726355460982e-06,  8.989765232959e-07, {   0,   0,   1,   0,   0,  -3,   6,   0,  -2,   0,   0,   0,   0 } },
 { -7.306663536944e-06,  2.316841303286e-06, {   2,   0,  -1,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.511342806394e-06, -6.111688465413e-06, {   2,   0,  -4,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.726394757910e-06, -8.956510311311e-07, {   2,   0,   2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.587948445045e-06, -1.032888652152e-06, {   2,   0,   0,   0,   0,  -3,   6,   0,  -2,   0,   0,   0,   0 } },
 { -7.045518564879e-06,  2.569891629597e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   0,   4,  -2,   0,   0 } },
 {  7.818772873463e-06,  1.796481036889e-06, {   3,  -1,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -1.768691496811e-09, -9.611236478122e-06, {   4,   0,  -1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -9.310623474506e-06, -3.023772628081e-07, {   0,   0,   2,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.685310909956e-06,  7.924867492782e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   1,   1,   0,   0,   0 } },
 {  6.752596120672e-06,  2.857448986687e-06, {   2,   0,   1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.029259032237e-07,  8.805678813086e-06, {   0,   2,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  8.196108491119e-06, -1.410794785890e-06, {   1,  -1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -8.523800697420e-06,  1.082112301077e-06, {   0,   0,   1,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0 } },
 { -3.327129371066e-06,  6.274430755123e-06, {   4,   0,  -2,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 { -9.080056429306e-06,  5.188052079971e-07, {   1,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.107487788494e-06, -6.491232883148e-06, {   1,   0,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -7.346845008761e-06,  2.251524874744e-06, {   2,   0,   1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.642589970562e-08,  9.580329961323e-06, {   2,   0,  -4,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  9.022642066598e-06,  5.690108148342e-07, {   4,   0,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  3.995737432574e-06, -5.593197555141e-06, {   0,   0,   0,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  1.607363476879e-06, -7.977970307163e-06, {   1,   1,  -1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -7.423825315795e-06, -2.160560100466e-06, {   2,   0,  -3,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.622944479238e-06,  5.959678660339e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   2,  -5,   0,   0,   0 } },
 { -3.933194774566e-07,  9.187941352855e-06, {   0,   0,   1,   0,   0,  -4,   5,   0,  -1,   0,   0,   0,   0 } },
 {  1.698868765447e-06,  7.881133420356e-06, {   1,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  1.770746224098e-06,  7.807585006686e-06, {   4,   0,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.187677968812e-06, -6.382327948869e-06, {   3,  -2,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.187677968812e-06,  6.382327948869e-06, {   3,  -2,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.930784957935e-06, -6.637948005161e-06, {   2,   0,   0,   0,  -5,  14,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.679691895228e-06, -5.888935101516e-06, {   2,   0,   0,   0,   0,   0,  -5,  11,   0,   0,   0,   0,   0 } },
 {  2.019174230050e-06,  7.545253258248e-06, {   0,   0,   1,   0,   0,   2,  -5,   3,   0,   0,   0,   0,   0 } },
 {  5.107710881408e-06, -4.456481649922e-06, {   2,   0,   0,   0,   0, -10,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.821255556730e-06, -1.742512246511e-06, {   2,   0,  -1,   0,   0,   0,  -9,  16,  -6,   0,   0,   0,   0 } },
 {  7.751700175053e-08, -9.482809837417e-06, {   0,   0,   3,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -1.675820594936e-06, -7.881995963077e-06, {   2,   0,  -1,   0,   0,  -3,   5,   0,  -1,  -1,   0,   0,   0 } },
 {  2.392986352354e-06, -7.162860426873e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   2, -10,   0,   0,   0 } },
 {  2.095551484165e-06,  7.458556898565e-06, {   1,   0,   0,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.167134319936e-06, -7.386164577970e-06, {   3,   0,  -2,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  3.199688834492e-06,  6.352955414519e-06, {   0,   0,   1,   0,   0,   0,   5,  -8,   1,   5,   0,   0,   0 } },
 {  7.678622411066e-06, -1.870389713388e-06, {   1,   1,  -2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.751391297927e-06,  5.797495268874e-06, {   4,   0,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0 } },
 {  7.707372690308e-06,  1.839776057698e-06, {   3,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.950765539373e-08, -9.455729129089e-06, {   2,   0,  -3,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.782141317830e-08, -9.527146609043e-06, {   2,   0,   3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.459319544192e-06, -6.083671925689e-06, {   6,   0,  -2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  7.008355385140e-06,  2.531974042438e-06, {   0,   0,   1,   0,   0,  20, -23,   0,   0,   0,   0,   0,   0 } },
 {  2.845150067351e-06,  6.683583131865e-06, {   2,   0,  -1,   0,   0,  -1,   3,  -3,   0,   0,   0,   0,   0 } },
 { -7.568469969376e-06,  1.958860484553e-06, {   2,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  7.923805161470e-07, -8.733845073619e-06, {   2,   0,   2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -1.527245764054e-06, -7.997503331894e-06, {   2,   0,  -1,   0,   0,   0,   5,   0,  -7,   5,   0,   0,   0 } },
 {  6.196962521407e-06, -3.326805782162e-06, {   0,   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  6.506688742896e-06,  3.015512171557e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 {  9.432592971342e-06,  8.847253418451e-08, {   2,   0,  -1,   0,   0,   0,  19, -22,   0,   0,   0,   0,   0 } },
 { -8.777817161518e-06, -7.425658590160e-07, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   4,   0,   0,   0 } },
 { -2.805355457163e-06, -6.707637562335e-06, {   2,  -2,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.346439225234e-06,  6.161388101049e-06, {   4,   0,  -2,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  9.029784460211e-06,  4.771631250814e-07, {   3,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.064220885429e-07, -9.399372027792e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0 } },
 { -5.164563291099e-06, -4.334650735581e-06, {   1,  -1,   1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 { -8.796173622165e-06,  7.024491231821e-07, {   3,  -1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  2.909367814456e-06,  6.586258408180e-06, {   2,   0,   0,   0,   5, -14,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.637499777831e-06,  3.858113213102e-06, {   1,  -1,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0 } },
 { -5.777488983847e-08,  9.434368429581e-06, {   2,  -2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -2.600045970111e-06,  6.891941103307e-06, {   1,   0,  -2,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 { -2.637409286270e-06, -6.851761362725e-06, {   2,   0,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.976797852368e-06,  1.512285524072e-06, {   0,   0,   2,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  8.899330576497e-06,  5.847145161740e-07, {   4,   0,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -7.550380616352e-09,  9.472987848135e-06, {   0,   2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  7.498320628054e-06,  1.981650853415e-06, {   2,   0,   0,   0,   0,   0,  -7,  12,  -3,   0,   0,   0,   0 } },
 { -2.947215798226e-07, -9.182808237578e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 { -1.683313705056e-06, -7.793617852360e-06, {   2,  -2,   1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.378926315495e-06,  8.097948763490e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 {  6.253061722912e-06, -3.219885321580e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  1.130364308266e-06,  8.341048025036e-06, {   2,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  6.311307435887e-06,  3.154040925463e-06, {   2,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.367472195924e-06,  3.091471464491e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   4,   0,   0,   0 } },
 { -1.332901996822e-06,  8.122551069259e-06, {   0,   0,   0,   0,   0,  26, -37,  16,  -4,  -5,   0,   0,   0 } },
 { -1.183759001027e-06,  8.267531609070e-06, {   0,   0,   2,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 { -2.090014314397e-08,  9.423605738902e-06, {   1,   0,   0,   0,   0, -16,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.454100701498e-09,  9.440110403435e-06, {   2,  -2,  -2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -9.499333324305e-07, -8.488900355840e-06, {   0,   0,   3,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -9.210969931223e-06,  2.266284803447e-07, {   2,   0,  -1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.692503774902e-06, -7.743635006279e-06, {   1,  -1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.608825365382e-06, -1.827244301685e-06, {   2,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.398164309680e-08,  9.371618434649e-06, {   2,  -2,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.120410874006e-06, -3.310506579107e-06, {   2,   0,   0,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0 } },
 { -4.019318440134e-06, -5.404152849825e-06, {   2,   0,  -1,   0,   0,   0,   4, -10,   0,   0,   0,   0,   0 } },
 {  3.921862598805e-06, -5.501556138754e-06, {   1,   0,  -1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.545013833734e-06, -6.873412478116e-06, {   2,   0,  -1,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0 } },
 { -4.849036145147e-06,  4.566399437612e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 {  2.490415113341e-06,  6.924891056188e-06, {   2,   0,  -3,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.569277940710e-06, -6.839302165262e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -2,   0,   0,   0 } },
 { -1.700907028555e-06,  7.706793991629e-06, {   3,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.040012157788e-06, -8.367472505580e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 { -4.192611540518e-09, -9.401179784562e-06, {   1,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -4.814513459192e-07, -8.920935114532e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0 } },
 {  7.953275738981e-06, -1.447885798888e-06, {   4,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  3.881709458202e-06,  5.518254906659e-06, {   1,   0,   1,   0,   0,   0, -33,  39,  -1,   0,   0,   0,   0 } },
 { -3.881709458202e-06, -5.518254906659e-06, {   1,   0,  -1,   0,   0,   0, -33,  39,  -1,   0,   0,   0,   0 } },
 {  1.104305537731e-06,  8.285577035425e-06, {   2,   0,  -2,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  5.832982121307e-06, -3.556467075141e-06, {   1,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  6.297044449879e-06,  3.087117983432e-06, {   2,   0,  -1,   0,   0,  -2,   3,   0,  -2,   0,   0,   0,   0 } },
 {  7.967217212456e-06, -1.415654958142e-06, {   0,   0,   0,   0,   0,  15,  -8,  -4,   0,   0,   0,   0,   0 } },
 { -6.635816750361e-06, -2.744305088047e-06, {   1,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.982669303615e-06,  6.395851960173e-06, {   1,  -1,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  5.121425440304e-06,  4.249432386051e-06, {   1,   1,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  7.465139609290e-06, -1.905518813070e-06, {   2,   0,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.963246869062e-06,  4.402278678285e-06, {   2,   2,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.664244424087e-06,  7.701243941760e-06, {   3,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.858731131177e-06,  3.506745275574e-06, {   2,   0,   0,   0,   0,   0, -12,  16,   0,   0,   0,   0,   0 } },
 { -5.939181778040e-06, -3.424263513789e-06, {   0,   2,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -3.118429252825e-06,  6.243717965060e-06, {   1,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.865870621352e-06, -4.495834992855e-06, {   1,  -1,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.464688824483e-07,  9.213608103580e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0 } },
 { -4.190926648140e-06, -5.163947799090e-06, {   3,   0,  -1,   0,   1, -28,  29,   0,   0,   0,   0,   0,   0 } },
 { -4.190926648140e-06,  5.163947799090e-06, {   1,   0,  -1,   0,  -1,  28, -29,   0,   0,   0,   0,   0,   0 } },
 {  7.144269584602e-06,  2.208474382350e-06, {   2,   0,  -1,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.305168334987e-07, -9.222225936225e-06, {   2,   0,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -6.980531887921e-06,  2.372068773587e-06, {   0,   0,   1,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.768421598421e-06,  7.580469929537e-06, {   1,   1,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.225839828429e-08,  9.331834207954e-06, {   1,   0,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -7.034511030111e-06, -2.307915595402e-06, {   1,  -1,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.734871684956e-06,  2.605965749269e-06, {   0,   0,   1,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -6.453550238112e-06,  2.887282707065e-06, {   1,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.944953296167e-06, -7.394700287632e-06, {   0,   0,   1,   0,   0,  -2,   5,  -3,   0,   0,   0,   0,   0 } },
 { -1.072189620488e-07, -9.228545796775e-06, {   0,   0,   3,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.757378191918e-06, -1.576732065684e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   5,   0,   0,   0 } },
 {  3.213351378518e-06,  6.118860795566e-06, {   2,   0,   0,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.745050779839e-06, -6.586180536216e-06, {   0,   0,   1,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.367067114283e-06, -4.962366900525e-06, {   2,   0,  -3,   0,   0,   4,  -5,   0,   3,   0,   0,   0,   0 } },
 {  1.913430294826e-08,  9.310226586499e-06, {   0,   0,   2,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.578591542279e-07, -9.070964846320e-06, {   6,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.128385673488e-06, -4.196961038535e-06, {   0,   1,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -6.401563286500e-06, -2.921579452048e-06, {   0,   0,   1,   0,   0,  12, -11,   0,   0,   0,   0,   0,   0 } },
 { -9.075330346471e-06,  2.439166565639e-07, {   2,   0,   3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.631457336448e-06, -7.687268543398e-06, {   2,   0,   1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.188063172562e-08, -9.252940388033e-06, {   1,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -1.533421208710e-08, -9.298611878338e-06, {   2,  -2,  -1,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 {  8.933913923535e-06, -3.736802417314e-07, {   1,  -1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  5.844012714967e-06, -3.456679309602e-06, {   3,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -8.751666947380e-06,  5.486212878450e-07, {   1,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -6.821432364537e-06, -2.476196203791e-06, {   0,   0,   1,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.505902229470e-06, -1.789748915882e-06, {   1,   1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.702068581772e-06, -5.907572495042e-07, {   2,   0,   2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -6.547895636752e-06, -2.744586158639e-06, {   3,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.996712861603e-06, -7.294357283605e-06, {   4,  -2,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  8.420294555612e-06,  8.690816632379e-07, {   0,   0,   1,   0,   0,  -8,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.321996539060e-06,  2.965341548682e-06, {   0,   1,  -3,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.441546250674e-06,  7.843312813274e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,   2,   0,   0,   0 } },
 { -6.214656561436e-06,  3.065866831375e-06, {   0,   1,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  4.957466995924e-06,  4.322363163776e-06, {   3,  -2,   1,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.957466995924e-06, -4.322363163776e-06, {   3,  -2,  -1,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.087010715611e-06,  7.192336577998e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0 } },
 {  2.223225672275e-06, -7.055985398355e-06, {   2,   0,   1,   0,   0,   0,  -5,   0,   5,   0,   0,   0,   0 } },
 {  2.890624523851e-09,  9.272590747559e-06, {   4,   0,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  9.079675255515e-06,  1.934755599446e-07, {   0,   0,   1,   0,   0,   0,  14, -17,   0,   0,   0,   0,   0 } },
 { -1.290051869277e-07,  9.142352346034e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0 } },
 {  6.818201837609e-06, -2.452133604609e-06, {   4,   0,  -1,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -6.735223423607e-06,  2.534655176401e-06, {   1,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -2.442508442156e-06, -6.826413370673e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,  -2,   0,   0,   0 } },
 {  6.141983359514e-06,  3.126829034582e-06, {   2,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.650704683487e-06, -7.612479109026e-06, {   3,   0,  -2,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  6.412927768721e-07, -8.619503362993e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 { -2.377778213573e-06, -6.881370477846e-06, {   2,   0,   0,   0,   0,  13, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.490331605316e-06, -7.617107072319e-07, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.679172237286e-06, -7.571072529325e-06, {   3,  -1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.751093521094e-06,  6.498421356361e-06, {   2,   0,   0,   0,   0,  -1,   3,  -3,   0,   0,   0,   0,   0 } },
 { -1.541933129643e-06, -7.699308838214e-06, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   1,   0,   0,   0 } },
 { -3.036914978034e-06,  6.203781991854e-06, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   2,   0,   0,   0 } },
 { -1.545004081239e-06,  7.694150795470e-06, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,  -1,   0,   0,   0 } },
 {  6.260553306625e-08, -9.176472033449e-06, {   2,   0,  -3,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 {  1.397659662086e-06,  7.840935959090e-06, {   2,   0,   0,   0,   0,   0,   7, -14,   0,   0,   0,   0,   0 } },
 {  8.819155034671e-06, -4.193424138676e-07, {   2,   0,   2,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.967914384480e-09, -9.233684247799e-06, {   4,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.927510850906e-06,  5.306131122047e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  7.508838881585e-06, -1.724094443886e-06, {   3,   1,  -1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -8.240358963736e-10, -9.228823093571e-06, {   2,  -2,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.769824302005e-06, -1.458850346403e-06, {   0,   2,   1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  7.170242564148e-06, -2.056385493088e-06, {   4,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.018040612816e-06,  4.208467206078e-06, {   2,   0,  -2,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -7.455137276604e-06, -1.768577041135e-06, {   3,  -1,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.020794895312e-07, -8.917014769071e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   3,  -1,   0,   0,   0 } },
 { -8.243799581140e-06,  9.736546422080e-07, {   2,   0,  -2,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.149774624768e-06,  5.064399570844e-06, {   0,   0,   1,   0,   0,  -3,   2,   2,   0,   0,   0,   0,   0 } },
 {  7.986186503076e-06,  1.226542772079e-06, {   4,   0,  -2,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 {  7.986186503076e-06, -1.226542772079e-06, {   0,   0,   2,   0,   0,  -3,   3,   0,   0,   3,   0,   0,   0 } },
 {  7.416771770152e-06, -1.794666901832e-06, {   1,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.029480208113e-07, -9.104701092585e-06, {   3,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -3.691210262926e-06,  5.514266877133e-06, {   0,   0,   2,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -7.441711113075e-06,  1.763545618950e-06, {   3,  -1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.236638491368e-06, -5.966333820424e-06, {   3,  -1,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -3.455016599844e-06,  5.745558885828e-06, {   0,   0,   2,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  6.368280485474e-06, -2.832088248328e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -3.093885386908e-06, -6.103747984589e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0 } },
 { -1.535365577969e-06, -7.662231649152e-06, {   1,  -1,  -1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.365455014906e-06, -1.830580014741e-06, {   4,   0,  -2,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.243153041712e-06, -9.518465555495e-07, {   4,   0,  -1,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.475396105547e-06, -2.717525639996e-06, {   0,   0,   1,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 { -5.226609367509e-06, -3.964186359405e-06, {   2,   0,  -3,   0,   0,   0,   7, -10,   1,   0,   0,   0,   0 } },
 { -5.022021566303e-06, -4.168051275098e-06, {   2,   0,  -2,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -1.270467279799e-06,  7.917236947783e-06, {   2,   2,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.037145657061e-06, -6.150081780542e-06, {   2,   0,  -2,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.513029152587e-06, -3.673294000104e-06, {   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0 } },
 {  7.360415908678e-06, -1.824388968570e-06, {   3,  -1,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.329007694087e-06, -3.853111540240e-06, {   2,   2,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.807413724490e-06, -5.373978215797e-06, {   4,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -1.683602041277e-08, -9.161982292125e-06, {   2,   0,   0,   0,   0,  -2,   4,   0,  -2,   0,   0,   0,   0 } },
 { -6.115015831996e-06,  3.063791961124e-06, {   1,  -1,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.270899480681e-08,  9.160699216152e-06, {   2,   0,  -2,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.643290508440e-06,  7.529714665172e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 { -7.286552535721e-06, -1.881376129669e-06, {   0,   1,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.867469250015e-06,  2.299805972362e-06, {   2,  -2,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.890446929001e-06,  5.273833665617e-06, {   0,   0,   1,   0,   0,   0,  -1,   2,  -2,   3,   0,   0,   0 } },
 { -3.045575273740e-06, -6.117799919633e-06, {   0,   2,   0,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  3.888541097835e-06, -5.273524345123e-06, {   0,   0,   1,   0,   0,   0,   1,  -2,   2,  -3,   0,   0,   0 } },
 { -3.073487105810e-06,  6.088086668704e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   6,   0,   0,   0,   0 } },
 {  1.748093138658e-06, -7.413361956583e-06, {   1,  -1,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.415822194866e-06,  6.744656208013e-06, {   2,   0,  -3,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  2.290410646506e-08, -9.136794799220e-06, {   4,   0,  -1,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.687477631733e-06,  4.700347174385e-07, {   2,   0,   0,   0,   0,   0, -12,  15,   0,   0,   0,   0,   0 } },
 { -3.435183342983e-06, -5.722302685414e-06, {   2,   0,   1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  2.686944265762e-06,  6.469829488214e-06, {   0,   0,   1,   0,   0,  -1,   3,  -3,   0,   0,   0,   0,   0 } },
 { -8.571009300681e-06,  5.849239591034e-07, {   4,  -2,  -3,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.903955511839e-06, -7.246354663582e-06, {   1,   1,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.896891443639e-06, -6.251188581947e-06, {   1,   1,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.002762252595e-06, -8.144956205112e-06, {   0,   0,   1,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 { -2.639548933656e-06,  6.503091354894e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0 } },
 {  3.028610427542e-06, -6.112626656428e-06, {   0,   0,   2,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.724075397744e-06,  4.149881799888e-07, {   2,   0,   0,   0,   0,   0,   4,   0,   0,  -5,   0,   0,   0 } },
 { -1.707625159511e-06,  7.431303808234e-06, {   1,  -1,  -2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.029443167620e-07, -8.235211677784e-06, {   1,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.364699821491e-06,  4.772249542697e-06, {   2,   0,   0,   0,   0,  -1,   7, -12,   0,   0,   0,   0,   0 } },
 {  2.647594139780e-06,  6.485936397138e-06, {   1,  -1,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -7.552708918353e-06, -1.579735577769e-06, {   0,   0,   2,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -3.083739439564e-07, -8.823248144914e-06, {   2,   0,  -1,   0,   0,   5,  -7,   0,   2,   0,   0,   0,   0 } },
 { -8.949156349593e-06, -1.824621179987e-07, {   4,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.713015464211e-06,  5.418349637396e-06, {   0,   0,   2,   0,   0, -12,   6,   0,   3,   0,   0,   0,   0 } },
 {  3.713015464211e-06, -5.418349637396e-06, {   0,   0,   0,   0,   0,  12,  -6,   0,  -3,   0,   0,   0,   0 } },
 {  5.352547364635e-06, -3.776817856527e-06, {   3,  -1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  6.324754246143e-06,  2.803006884080e-06, {   0,   0,   0,   0,   0,   3,  -6,   4,   0,   0,   0,   0,   0 } },
 {  8.490191479797e-06, -6.366547974151e-07, {   2,   0,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.640234601594e-06,  7.484523720747e-06, {   0,   2,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.287172438466e-08,  9.060034505659e-06, {   0,   2,   0,   0,   0, -24,  24,   0,   0,   0,   0,   0,   0 } },
 { -6.745160171052e-06, -2.377401804093e-06, {   1,   1,  -2,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  3.751313056631e-06, -5.370159608563e-06, {   2,   0,  -2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  6.065204433012e-06, -3.055660702216e-06, {   4,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.543243918735e-06, -7.572995732948e-06, {   4,   0,  -1,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 {  1.543243918735e-06,  7.572995732948e-06, {   0,   0,   1,   0,   0,   0,  37, -48,   0,   0,   0,   0,   0 } },
 {  1.460382883248e-06, -7.655837321038e-06, {   2,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.636381722641e-06, -3.479262742401e-06, {   2,   0,   2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -8.734453503595e-07,  8.236232231163e-06, {   2,   2,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -3.782205032757e-06, -5.325879566776e-06, {   2,   0,  -1,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.532125078320e-07,  8.251219018893e-06, {   4,   0,  -2,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -7.455855575867e-06, -1.646209728746e-06, {   4,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 { -3.042954722056e-06, -6.058459762678e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   6,  -5,   0,   0,   0 } },
 {  7.701952707296e-06,  1.397680718417e-06, {   1,   1,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  2.509181488493e-07,  8.847802736155e-06, {   2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 } },
 { -7.373310217167e-06,  1.723393830990e-06, {   1,  -1,   1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  8.496511143660e-06,  5.986561485611e-07, {   1,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  2.640703727082e-08, -9.063660908037e-06, {   4,   0,   1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 {  3.590573641478e-06, -5.498193479043e-06, {   2,   0,  -1,   0,  -2,   7,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.283986007920e-06, -7.800315409530e-06, {   2,   0,  -1,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0 } },
 {  7.782575128865e-06,  1.296471226444e-06, {   2,   0,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.015588947948e-06,  6.062567878197e-06, {   0,   0,   2,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 {  9.996650085657e-09,  9.066593966768e-06, {   4,   0,  -2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.837875595368e-06, -2.237469718767e-06, {   0,   0,   2,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  5.908624125355e-06,  3.166642073243e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -6,   6,   0,   0,   0 } },
 { -4.909630430390e-06,  4.165237927931e-06, {   2,   0,   0,   0,   0, -11,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.397938292908e-06,  3.675000440249e-06, {   3,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.671783424163e-06, -2.401090582781e-06, {   0,   0,   1,   0,   3,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.444499472783e-08, -9.046521652766e-06, {   4,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  7.211879492045e-06,  1.854887650995e-06, {   1,  -1,  -1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.899417004929e-06, -6.166875561002e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   8,   0,   0,   0 } },
 { -6.950486619365e-06, -2.112168800239e-06, {   0,   0,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -8.656164516926e-06, -4.045146819146e-07, {   2,   0,  -2,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 { -2.779971539803e-06, -6.276785638250e-06, {   1,   1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  7.169938018273e-07, -8.338401917137e-06, {   1,  -1,   1,   0,   0,   0,  -8,  17,   0,   0,   0,   0,   0 } },
 { -7.169938018273e-07,  8.338401917137e-06, {   1,  -1,  -1,   0,   0,   0,  -8,  17,   0,   0,   0,   0,   0 } },
 {  2.990167608980e-06,  6.065034576713e-06, {   6,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.182751282649e-06, -5.870596792632e-06, {   2,   0,  -2,   0,   0,   0,   3,   0,  -6,   0,   0,   0,   0 } },
 {  7.494471500887e-06,  1.556209598889e-06, {   2,   0,  -1,   0,   0,   0,   7, -16,   6,   0,   0,   0,   0 } },
 {  3.013227188459e-07, -8.749086662973e-06, {   2,   0,   1,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0 } },
 { -8.589102209303e-09,  9.039333400344e-06, {   4,  -2,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.374454999211e-07, -8.501733967961e-06, {   2,   0,  -1,   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0 } },
 { -8.736101753544e-06, -2.998700961902e-07, {   0,   0,   4,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.513815434152e-07, -8.884261669374e-06, {   1,   0,  -1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.308363017268e-06,  1.724737490887e-06, {   3,  -1,  -1,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.564009257645e-06,  7.468729606703e-06, {   1,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.913331941191e-06,  4.116195282180e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 {  1.194646211835e-07, -8.909161177138e-06, {   1,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  8.292074192781e-06,  7.355963174547e-07, {   2,  -2,  -3,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -5.689207145540e-06, -3.338110837540e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,  -4,   0,   0,   0 } },
 { -4.256151432965e-06, -4.770702726531e-06, {   2,   0,   1,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.093247365639e-06,  1.933129625968e-06, {   2,   0,   0,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 { -3.650473433550e-06,  5.374394041503e-06, {   2,   0,  -2,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  3.100031871671e-06,  5.923561644455e-06, {   0,   0,   2,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -7.751919948476e-06,  1.266196611767e-06, {   1,   0,  -2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -7.927096004296e-06,  1.087708897973e-06, {   0,   2,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.526469439671e-06, -7.488158632285e-06, {   1,   1,  -1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.817881531645e-06, -3.194960224676e-06, {   3,  -1,  -1,   0,   0,   0,  14, -21,   0,   0,   0,   0,   0 } },
 {  7.998325050450e-06, -1.014160847201e-06, {   2,   0,  -1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -2.080868389764e-07, -8.801468282457e-06, {   2,   0,  -3,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -8.996369943092e-06,  9.031876569467e-09, {   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.263927904591e-08, -8.987151588520e-06, {   4,   0,  -1,   0,   0, -14,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.005750455789e-06,  1.992001786088e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 {  5.515601526768e-06,  3.482121647491e-06, {   2,   0,  -1,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -7.100982604688e-06,  1.896210844078e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -3.245268577500e-06, -5.749251482622e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 { -2.220701970039e-07,  8.771487368826e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,  -3,   0,   0,   0 } },
 {  7.914806344222e-09, -8.984940211384e-06, {   1,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -5.549660442682e-06,  3.443062583466e-06, {   1,  -1,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -4.929792595911e-06,  4.060532119255e-06, {   4,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -8.610199382299e-06, -3.798757141969e-07, {   2,   0,   1,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.267276453401e-06,  1.717907089975e-06, {   1,  -1,   2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.495600854282e-06,  2.489520712579e-06, {   2,   0,   0,   0,   0,   0,  -6,   8,  -1,   0,   0,   0,   0 } },
 { -1.186727886570e-06,  7.796150691848e-06, {   0,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 { -6.846641044855e-09, -8.975487423214e-06, {   6,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.246834031368e-06, -1.728020507470e-06, {   0,   1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.156074507068e-09, -8.968895388999e-06, {   2,   2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.310346043993e-06,  5.661501966262e-06, {   3,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  5.203749171920e-06, -3.767166317508e-06, {   2,   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0 } },
 {  2.870824265094e-06,  6.095288259888e-06, {   1,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  3.179172561414e-06,  5.786678317795e-06, {   5,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  3.179172561414e-06, -5.786678317795e-06, {   3,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -2.877058713849e-06, -6.088655625149e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -5,   3,   0,   0,   0 } },
 {  5.923732613727e-06,  3.041958768422e-06, {   0,   2,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.990676820159e-06, -1.974368508154e-06, {   2,   0,  -1,   0,   0,   3,  -3,  -4,   3,   0,   0,   0,   0 } },
 {  6.990676820159e-06,  1.974368508154e-06, {   2,   0,  -1,   0,   0,  -3,   3,   4,  -3,   0,   0,   0,   0 } },
 { -2.462842982645e-06,  6.498697737025e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -6,   5,   0,   0,   0 } },
 {  2.372397724935e-06,  6.588859466378e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,  -4,   3,   0,   0,   0 } },
 {  1.541862617749e-06, -7.418268967017e-06, {   0,   0,   3,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  7.060154251686e-06,  1.897858993541e-06, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   3,   0,   0,   0 } },
 { -5.581179038065e-06,  3.375571247945e-06, {   2,   0,  -1,   0,   0,   4,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.153973970988e-06, -6.802771938881e-06, {   3,   1,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -9.421816294211e-07,  8.014042186656e-06, {   2,   0,  -1,   0,   0,   0,   5,   0,  -3,  -5,   0,   0,   0 } },
 {  7.245235645884e-06, -1.710565166004e-06, {   4,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.820525935223e-06, -2.132110527309e-06, {   0,   0,   0,   0,   0,  21, -23,   4,   0,   0,   0,   0,   0 } },
 {  8.410029921221e-06,  5.425883465479e-07, {   4,  -2,   0,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -2.333910498224e-06, -6.617205266864e-06, {   1,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  8.582675156599e-06,  3.683747937679e-07, {   0,   0,   0,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -2.685129425589e-08, -8.924174847925e-06, {   2,   0,   2,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -6.398286231010e-06, -2.552561869422e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0 } },
 { -4.850625620741e-06, -4.094114241151e-06, {   1,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0 } },
 {  6.896449356575e-06,  2.043752351223e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -6.896449356575e-06, -2.043752351223e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 {  3.213412419078e-06,  5.725250844437e-06, {   3,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -6.478982816531e-06, -2.458381161046e-06, {   0,   1,   2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  6.478982816531e-06,  2.458381161046e-06, {   0,   1,  -2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 {  8.917437912626e-06, -1.620115679063e-08, {   1,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  7.261807647924e-06,  1.670871484882e-06, {   1,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.106577930395e-06, -1.822256703023e-06, {   2,  -2,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -8.247399296252e-06,  6.797555082460e-07, {   6,  -2,  -2,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  8.247399296252e-06, -6.797555082460e-07, {   2,  -2,  -2,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.221813901338e-06, -5.704762312605e-06, {   6,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -6.809497056607e-06,  2.112247733223e-06, {   2,   0,   0,   0,   0,   0,  -3,   6,  -3,   0,   0,   0,   0 } },
 { -4.561721977511e-06, -4.356573394184e-06, {   0,   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -7.211362822415e-06, -1.705803989498e-06, {   3,  -1,  -3,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.692505949387e-07, -8.343600866159e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,  -2,   0,   0,   0 } },
 {  1.764551418415e-06,  7.147041263315e-06, {   2,   0,   1,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0 } },
 { -3.977743889865e-06, -4.933372350986e-06, {   4,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.773497494914e-06,  3.136812790853e-06, {   4,   0,  -2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 {  2.958927843908e-06, -5.950607950003e-06, {   0,   0,   2,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 { -2.685990827201e-06, -6.220582127366e-06, {   2,   0,  -1,   0,   0,   0,  18, -25,   0,   0,   0,   0,   0 } },
 {  7.212336635410e-06, -1.694235795198e-06, {   1,  -1,  -1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 { -6.173857908458e-06, -2.730100590729e-06, {   1,  -1,   2,   0,   0,  -6,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.207701522754e-06, -5.694776177904e-06, {   4,  -2,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  7.086137495243e-06,  1.815668772647e-06, {   2,   0,  -1,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0 } },
 { -8.726912909069e-06,  1.690626836559e-07, {   0,   2,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.862011276655e-06,  2.030490142629e-06, {   1,  -1,  -1,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.260738803818e-06,  2.630537367202e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  3.457712359305e-06,  5.427777956037e-06, {   1,  -1,   2,   0,   0, -21,  22,   0,   0,  -2,   0,   0,   0 } },
 { -3.457712359305e-06, -5.427777956037e-06, {   1,  -1,   0,   0,   0, -21,  22,   0,   0,  -2,   0,   0,   0 } },
 {  6.933550768558e-06,  1.947998784002e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 {  1.512470818191e-06, -7.364614473037e-06, {   1,   1,   0,   0,   0, -12,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.895849112317e-06,  2.980528404409e-06, {   4,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  4.856572059122e-06, -4.013264980717e-06, {   2,  -2,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -8.282547213120e-08, -8.785823341030e-06, {   2,   0,  -2,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  5.846636835659e-06, -3.021565320264e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  8.979334160142e-07,  7.968470758559e-06, {   1,   1,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  8.921508651621e-08, -8.777052108327e-06, {   3,  -2,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -7.213093366437e-06, -1.651749820107e-06, {   4,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  7.319574536347e-06,  1.543237661351e-06, {   2,  -2,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.273414045907e-06, -4.586265712587e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0 } },
 { -1.350476994578e-06,  7.508156821607e-06, {   0,   0,   1,   0,   0,   0,  -6,   0,   6,   0,   0,   0,   0 } },
 { -7.273593463473e-06, -1.584392408237e-06, {   4,   0,  -2,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 } },
 {  3.663877812268e-06, -5.193754786211e-06, {   0,   0,   1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0 } },
 {  6.986605828781e-06, -1.867753396875e-06, {   3,  -1,  -3,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.771003722628e-06,  7.081070647477e-06, {   1,   1,  -4,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -3.777754376652e-06,  5.073412080495e-06, {   4,   0,  -2,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  6.099810494986e-06, -2.750752550729e-06, {   3,   1,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -7.200621325828e-06,  1.647358298152e-06, {   1,   1,  -1,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.759851562647e-06,  6.086146973902e-06, {   2,   0,   1,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.045330629598e-09, -8.843812251784e-06, {   4,  -2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.558708425557e-06,  7.285712147939e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   1,   1,   0,   0,   0 } },
 { -4.243023082541e-06,  4.601042454094e-06, {   0,   0,   1,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 { -8.112847997581e-06, -7.302202379949e-07, {   4,  -2,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.462471007144e-06,  7.378737533431e-06, {   1,   1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -3.989696363558e-06, -4.850249422276e-06, {   0,   0,   1,   0,   2,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.673542617213e-06,  4.165171843195e-06, {   0,   0,   2,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  8.422713647945e-06, -4.123419017064e-07, {   2,   2,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.310347075812e-09, -8.831143167325e-06, {   4,   0,   0,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.193567741581e-06,  4.636338874387e-06, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -5,   5,   0,   0,   0 } },
 {  6.966335087314e-06,  1.861696216644e-06, {   1,   1,  -3,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  4.968380876627e-06, -3.859417543861e-06, {   2,   0,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0 } },
 { -4.570388349360e-06,  4.257255492612e-06, {   2,   0,  -1,   0,   0,  -3,   5,  -4,   0,   0,   0,   0,   0 } },
 { -8.451873395792e-06,  3.735614999742e-07, {   1,   0,  -2,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.556469861205e-06, -6.268304797024e-06, {   2,   0,   0,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 { -6.857406472209e-06, -1.965865293200e-06, {   0,   0,   1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.927750796631e-06,  5.894550532713e-06, {   1,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.734362058322e-06, -8.660442137045e-08, {   2,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 {  6.872743488430e-06,  1.945690859092e-06, {   4,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  3.939667310836e-06, -4.875734087292e-06, {   4,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 { -5.798056628932e-06,  3.013426381970e-06, {   4,   0,  -1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 {  5.798056628932e-06, -3.013426381970e-06, {   0,   0,   1,   0,   0, -38,  37,   0,   0,   0,   0,   0,   0 } },
 {  4.940066204115e-06, -3.870555466580e-06, {   0,   0,   2,   0,   0, -13,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.606026268937e-06, -6.203315910982e-06, {   2,   0,  -3,   0,   0,   0,   4,  -4,  -1,   0,   0,   0,   0 } },
 { -4.193886301055e-07, -8.389250368232e-06, {   1,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -4.193886301055e-07,  8.389250368232e-06, {   1,   0,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  1.551120069537e-06, -7.251747098436e-06, {   1,  -1,   2,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.646651131477e-06, -1.554428900816e-07, {   2,   2,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.848331331380e-06, -1.951139582575e-06, {   0,   0,   2,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -6.274542929293e-06,  2.522492157780e-06, {   1,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 { -9.755425790759e-07, -7.819329297274e-06, {   3,  -1,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.315348768512e-06,  4.793903569710e-07, {   4,   0,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.029305699613e-06,  1.765164976777e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -6,   3,   0,   0 } },
 { -2.925899531278e-06,  5.864981017794e-06, {   2,   0,  -1,   0,   0, -10,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.366880952729e-06,  2.420779018128e-06, {   1,  -1,   2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -7.103877424098e-06,  1.681875469722e-06, {   1,   1,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.549014529602e-06,  7.233526649883e-06, {   2,   0,  -3,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -8.492897162540e-06, -2.879193786674e-07, {   2,   0,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  1.575051002729e-07, -8.621690853997e-06, {   4,   0,  -2,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -5.548462203101e-06,  3.229128870093e-06, {   1,  -1,   1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  5.548462203101e-06, -3.229128870093e-06, {   1,  -1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  1.022322558386e-08, -8.762930827346e-06, {   3,   0,  -2,   0,   0,  10, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.836120638568e-07, -8.588853019721e-06, {   6,  -4,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  1.836120638568e-07,  8.588853019721e-06, {   2,  -4,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -3.926886720207e-06,  4.843478309834e-06, {   2,   0,   0,   0,   0,   0,   6, -13,  -1,   0,   0,   0,   0 } },
 {  7.060768167522e-06, -1.708572381263e-06, {   2,  -2,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.730748500364e-06, -2.037089698697e-06, {   2,   0,  -1,   0,   0,  -6,  10,   0,  -5,   5,   0,   0,   0 } },
 { -4.555923269215e-06,  4.210581250713e-06, {   5,  -1,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.663581906039e-07, -8.099911917757e-06, {   1,   0,   2,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.399209390839e-06, -6.365753288143e-06, {   1,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.359890266965e-06, -3.404769581744e-06, {   4,  -2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.544432063497e-06, -7.215265639547e-06, {   2,   0,   0,   0,   0,  -3,   5,   0,  -1,  -1,   0,   0,   0 } },
 { -5.439810895979e-06,  3.319017553735e-06, {   1,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.389330468918e-06,  6.369127265097e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   5,  -3,   0,   0,   0 } },
 {  1.753235520748e-06, -6.995635363973e-06, {   1,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.611817230932e-06,  7.134538658808e-06, {   1,   1,  -1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  3.156949526432e-06,  5.587866539112e-06, {   0,   0,   2,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -4.394372308175e-06, -4.350422425760e-06, {   2,   0,  -2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.092785880031e-07, -8.632994029528e-06, {   3,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.546867468147e-06, -1.193096591454e-06, {   2,   0,  -1,   0,   0,   0,  15, -24,   0,   0,   0,   0,   0 } },
 {  1.245432266121e-08, -8.725176309906e-06, {   4,   0,  -2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.137012394093e-07, -8.623409028186e-06, {   2,   2,  -2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -6.515483562901e-06, -2.220670725333e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  7.062767228237e-06,  1.670733791158e-06, {   1,  -1,  -2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.768927444625e-06,  2.963277329723e-06, {   1,  -1,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.515372606981e-06,  2.159060180587e-07, {   0,   0,   3,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -5.155764287340e-09,  8.722780281780e-06, {   3,  -2,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.285519867716e-06, -2.441797617852e-06, {   0,   1,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.805664887362e-06,  3.920997164288e-06, {   1,   1,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.918082359455e-06,  3.806388994820e-06, {   1,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  8.021107352539e-06, -7.018274267501e-07, {   0,   0,   0,   0,   0,  20, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.063853827063e-06,  7.657951748100e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   2,   0,   0 } },
 {  7.013898405880e-06,  1.706492375951e-06, {   1,  -1,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.296972562766e-06, -3.416082995106e-06, {   3,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  7.372084860025e-06, -1.340479628098e-06, {   3,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  7.237716776495e-06, -1.473707767677e-06, {   2,   0,  -3,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -6.982067365298e-06, -1.729077852541e-06, {   0,   1,   0,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.428623683802e-06,  7.282283127820e-06, {   1,  -1,   1,   0,   0,  11, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.685418158933e-06,  2.022273840986e-06, {   2,   0,  -1,   0,   0,   6, -10,   0,   5,  -5,   0,   0,   0 } },
 { -7.038982131203e-06, -1.668215208642e-06, {   3,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  7.124416734418e-06, -1.582595852935e-06, {   0,   0,   1,   0,   0,  -3,   4,   0,  -2,   0,   0,   0,   0 } },
 { -4.016290951071e-09, -8.702634428907e-06, {   4,   0,   1,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.748218017289e-06,  2.958165697785e-06, {   2,   0,   1,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0 } },
 {  7.592270468961e-06, -1.113499860342e-06, {   4,   0,  -1,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 { -8.296462742784e-06, -4.086040009478e-07, {   2,   0,  -2,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -4.600134511224e-07, -8.243802208531e-06, {   2,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  7.060784483422e-06,  1.642934791778e-06, {   1,   1,   0,   0,   0,  17, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.042521942985e-06,  1.661102164802e-06, {   1,  -1,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -7.352638158248e-07,  7.965410964944e-06, {   2,   0,   0,   0,   0,   0,  19, -25,   0,   0,   0,   0,   0 } },
 {  5.890241358007e-06, -2.808889274442e-06, {   2,   0,   2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -6.576476359903e-07, -8.040355898707e-06, {   0,   0,   1,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 { -5.313416152636e-06,  3.382842348433e-06, {   4,  -2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.381593138402e-06,  5.313739945433e-06, {   3,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  1.543009401996e-06,  7.151689559185e-06, {   1,  -1,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0,   0 } },
 { -1.392089162600e-06,  7.302033919227e-06, {   2,  -1,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.518370034850e-06,  6.174830852538e-06, {   0,   0,   0,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 {  6.676982865936e-06, -2.015398489793e-06, {   0,   0,   0,   0,   0,  19, -22,   0,   0,   0,   0,   0,   0 } },
 { -8.374089698546e-06,  3.170989565822e-07, {   2,   0,   2,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -4.458221027073e-06,  4.232348211046e-06, {   2,   0,   2,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -4.458221027073e-06, -4.232348211046e-06, {   2,   0,   0,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -1.748371250444e-06, -6.939741540478e-06, {   1,  -1,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  7.004927263141e-06,  1.680651852275e-06, {   0,   1,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.534863672595e-06, -5.150215869165e-06, {   0,   0,   1,   0,   0,   0,  13, -18,   0,   0,   0,   0,   0 } },
 {  8.379023464667e-06,  3.042447342396e-07, {   2,   0,  -2,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.774672052232e-06,  4.907831103959e-06, {   4,   0,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 {  5.770882359176e-09, -8.671116393605e-06, {   2,  -2,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.541122632248e-06, -5.134235834885e-06, {   1,   1,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -1.614904919088e-08, -8.657790976151e-06, {   4,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.468085558893e-06, -6.205515845020e-06, {   4,   0,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  1.389885761079e-06,  7.282680250123e-06, {   2,   0,  -1,   0,   0,   0,   7, -14,   0,   0,   0,   0,   0 } },
 {  1.489804896993e-07, -8.520390408309e-06, {   4,   0,  -2,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.775774581925e-06, -8.928009708698e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,  -5,  10,   0,   0,   0 } },
 {  1.098343320743e-06, -7.567043816272e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 {  8.408458571246e-06, -2.561594731858e-07, {   2,   0,  -1,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0 } },
 { -4.811081489910e-06, -3.853192953147e-06, {   1,  -1,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.662872601630e-06, -5.001194744108e-06, {   1,   1,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  5.429022742982e-06, -3.231827559275e-06, {   1,   1,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.939833390731e-06, -6.719526629795e-06, {   4,   0,  -2,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.714274094979e-06, -5.941843172258e-06, {   3,  -3,   0,   0,   0,  43, -43,   0,   0,   0,   0,   0,   0 } },
 {  2.654235458367e-06, -6.001195008095e-06, {   2,   0,  -1,   0,  -5,  22, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.779249925703e-06, -3.874485172716e-06, {   1,  -1,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.790193940835e-06, -6.860804823544e-06, {   1,   1,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.736283629035e-06, -6.914139504775e-06, {   1,   1,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.742883566473e-06, -3.906580385458e-06, {   1,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.337310807750e-06,  6.311251380188e-06, {   2,   0,  -1,   0,   0,  14, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.652315508141e-06,  5.995617598024e-06, {   2,   0,  -1,   0,   5, -22,  15,   0,   0,   0,   0,   0,   0 } },
 { -7.306787498892e-06, -1.331849336218e-06, {   1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -5.815120883860e-06, -2.821841448063e-06, {   1,   1,  -2,   0,   0, -28,  33,   0,   0,   0,   0,   0,   0 } },
 { -8.304021199581e-06, -3.318631260941e-07, {   2,  -2,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -6.986407153400e-06,  1.646911627373e-06, {   0,   2,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.617671593662e-06,  5.013280497903e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   5,   0,   0,   0 } },
 {  8.208048998761e-06,  4.222563062858e-07, {   3,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.355439555118e-06, -5.274395223986e-06, {   3,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.774310410993e-06, -6.855388282085e-06, {   4,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -8.336148856497e-06, -2.905636659109e-07, {   2,   0,  -2,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 { -8.401149762555e-06, -2.251130964619e-07, {   0,   0,   3,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -6.993346490811e-06, -1.632640774224e-06, {   1,  -1,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.027536367826e-06,  7.595887081760e-06, {   1,  -1,   2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.830104215752e-06,  1.788683566695e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0 } },
 {  3.047717704210e-07, -8.313370068806e-06, {   2,   0,   2,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -3.249553069995e-06, -5.368227609818e-06, {   1,   0,   0,   0,   0,   0,   4,  -6,   1,   0,   0,   0,   0 } },
 {  4.243389069695e-06,  4.373025822512e-06, {   0,   0,   1,   0,   0,   0,   1,  -4,   3,   0,   0,   0,   0 } },
 {  1.534831453371e-08,  8.598591312549e-06, {   0,   0,   1,   0,   0,   1,  -3,   0,   2,   0,   0,   0,   0 } },
 {  2.252446350269e-07,  8.385760055951e-06, {   0,   0,   1,   0,   0,   4,  -6,   0,   2,   0,   0,   0,   0 } },
 {  8.233244397305e-06, -3.772104675428e-07, {   2,  -2,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  6.693796526459e-06,  1.915868116663e-06, {   1,  -1,   0,   0,   0, -14,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.323507847534e-06,  1.284969473892e-06, {   0,   0,   1,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 { -7.929859385770e-06, -6.756015455592e-07, {   1,  -1,   1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  7.929859385770e-06,  6.756015455592e-07, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -4,   5,   0,   0,   0 } },
 {  1.631092737381e-07, -8.435386243990e-06, {   2,   0,  -1,   0,   0,   0,   6, -15,   0,   0,   0,   0,   0 } },
 {  5.056948183965e-06, -3.541020916648e-06, {   2,   0,   2,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  1.183692191020e-06, -7.411760751811e-06, {   1,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.379545115281e-06, -4.214988567099e-06, {   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.923425954722e-06,  6.615825748355e-07, {   2,   0,  -2,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0 } },
 { -1.435112009864e-06, -7.147510528212e-06, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   1,   0,   0,   0 } },
 {  2.792097842688e-06, -5.788898929078e-06, {   4,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0 } },
 { -3.078111287605e-06, -5.502565243115e-06, {   2,   0,  -3,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -8.156442956968e-06, -4.235384383921e-07, {   4,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  8.120172758820e-06, -4.556379433882e-07, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -8.132284170095e-08, -8.491027405121e-06, {   2,  -2,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  4.663440557461e-06, -3.905111801219e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -1,   3,   0,   0,   0 } },
 { -2.727224632764e-06,  5.839187103462e-06, {   2,   0,  -2,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.289751798595e-07, -8.235883508875e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  4.537353123442e-06,  4.025435160734e-06, {   2,   0,  -4,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  7.787652656557e-06, -7.748899455400e-07, {   1,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.419720899960e-06,  5.139375604205e-06, {   4,   0,  -1,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  3.986093198141e-06,  4.572571160290e-06, {   0,   0,   1,   0,   0,   0,  -1,  -2,   3,   0,   0,   0,   0 } },
 { -4.788672924418e-06, -3.769115008932e-06, {   1,  -1,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.367666951410e-06, -3.188728965520e-06, {   2,   0,  -1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.010043354806e-06,  3.545230697680e-06, {   1,   1,  -2,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.276164610404e-06,  6.277767006914e-06, {   2,   0,  -1,   0,   0,   0,   3,  -8,   5,  -5,   0,   0,   0 } },
 { -1.006498225041e-06, -7.546003484649e-06, {   0,   0,   1,   0,   0,  -3,   8,  -6,   0,   0,   0,   0,   0 } },
 {  5.262631395376e-06,  3.288606338529e-06, {   2,  -2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -2.976609500072e-06, -5.570653750502e-06, {   2,   0,   0,   0,  -7,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.437685490503e-06,  6.105660780098e-06, {   2,   0,   0,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 {  7.881622616429e-06,  6.610425473707e-07, {   4,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.021915454017e-08, -8.528484796103e-06, {   0,   2,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -6.187599511178e-06, -2.348932991630e-06, {   1,  -1,  -2,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  3.898443077571e-06,  4.636517300701e-06, {   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.201014592554e-06,  7.333824541797e-06, {   0,   0,   1,   0,   0, -22,  24,   0,   0,   0,   0,   0,   0 } },
 {  5.371113991739e-08,  8.477800245083e-06, {   1,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -4.981321481367e-06,  3.543138649249e-06, {   1,  -1,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.720384266363e-06, -2.803399197890e-06, {   4,   0,  -2,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -7.778113949824e-07,  7.742137086080e-06, {   1,   0,  -2,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -7.846205596840e-06, -6.730540600754e-07, {   0,   0,   0,   0,   0,  11,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.078907023510e-06, -1.439651708160e-06, {   2,  -2,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.885252767994e-06,  6.632093379199e-06, {   2,   0,  -1,   0,   0,   2,  -5,   3,   0,   0,   0,   0,   0 } },
 { -3.384512290067e-08, -8.481854143605e-06, {   2,   0,   1,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.173451935538e-06,  6.340510435925e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6, -10,   0,   0,   0 } },
 {  5.258429520676e-06,  3.253771027513e-06, {   2,  -2,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -1.427629242732e-06,  7.083803345434e-06, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,  -1,   0,   0,   0 } },
 { -7.715707253207e-06,  7.919439256300e-07, {   3,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.882675276751e-06,  1.622411185394e-06, {   0,   1,   0,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.466300144153e-09, -8.494539926972e-06, {   1,   0,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.485949954095e-06, -3.015545426335e-06, {   4,   0,  -1,   0,   1, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  5.485949954095e-06,  3.015545426335e-06, {   0,   0,   1,   0,  -1,  25, -25,   0,   0,   0,   0,   0,   0 } },
 { -5.248984514557e-06, -3.251757937145e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   4,  -3,   0,   0,   0 } },
 {  5.256566271569e-06,  3.230345425591e-06, {   0,   0,   2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -8.255180232065e-06,  2.289801551243e-07, {   4,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.383587978222e-06, -5.098787800022e-06, {   4,   0,  -1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  6.793885847709e-06,  1.688290355980e-06, {   0,   0,   2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  2.565780357784e-06, -5.915105069504e-06, {   1,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.000563200872e-06,  7.475228975729e-06, {   0,   0,   1,   0,   0,   3,  -8,   6,   0,   0,   0,   0,   0 } },
 { -2.821134224416e-07, -8.190971703792e-06, {   1,   0,   2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.039919540434e-06, -5.432580097018e-06, {   1,   1,  -1,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.620369907061e-06,  6.848357040879e-06, {   6,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.428115592867e-06, -7.040310523046e-06, {   3,  -1,  -1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -6.736064155834e-06, -1.730137990989e-06, {   2,  -2,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  6.609231302116e-06, -1.854981162126e-06, {   1,  -1,   1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.831100552952e-06, -1.632635560465e-06, {   1,   1,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.600284006909e-06,  1.862202732011e-06, {   0,   0,   3,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  7.995518130074e-06,  4.626261344308e-07, {   1,   1,  -2,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  5.787138244346e-06, -2.666195558062e-06, {   2,   0,   0,   0,   0,   0,  12, -10,   0,   0,   0,   0,   0 } },
 {  4.024372421642e-07,  8.050162879161e-06, {   4,   0,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  4.024372421642e-07, -8.050162879161e-06, {   4,   0,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  8.277053029604e-06,  1.729435699666e-07, {   1,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.620521772632e-06, -3.828466782945e-06, {   0,   0,   0,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  4.863868362738e-08,  8.397006595222e-06, {   2,   0,   0,   0,   0,  26, -30,   0,   0,   0,   0,   0,   0 } },
 {  6.427690457654e-06,  2.017942216092e-06, {   0,   0,   1,   0,  -3,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.085626582052e-06,  5.358433986240e-06, {   0,   0,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0 } },
 { -7.160936407639e-06,  1.281738424658e-06, {   0,   0,   3,   0,   0, -18,  20,  -8,   3,   0,   0,   0,   0 } },
 { -7.160936407639e-06, -1.281738424658e-06, {   0,   0,   1,   0,   0,  18, -20,   8,  -3,   0,   0,   0,   0 } },
 {  7.977935678721e-06,  4.634343670555e-07, {   3,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  2.681536209367e-06,  5.758999602007e-06, {   4,   0,  -1,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -6.848038378043e-06,  1.591908115528e-06, {   1,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  6.818438447491e-06, -1.620281601183e-06, {   2,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.543123217562e-07, -8.281990436260e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  6.476178984713e-09, -8.429235026994e-06, {   0,   0,   0,   0,   0,  28, -28,   0,   0,   0,   0,   0,   0 } },
 { -4.631886646927e-06, -3.801097799807e-06, {   1,  -1,   0,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.537591203764e-06,  2.889491414139e-06, {   0,   2,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -5.928447194044e-07,  7.830052101761e-06, {   2,   0,   0,   0,   0, -26,  30,   0,   0,   0,   0,   0,   0 } },
 {  2.951042272894e-08,  8.392482797620e-06, {   1,   0,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  2.536298848784e-06,  5.884815988746e-06, {   4,   0,  -1,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.917148927031e-07,  7.826628997334e-06, {   2,   0,   0,   0,   0,   0,  -8,  13,   1,   0,   0,   0,   0 } },
 { -2.055903319959e-06, -6.361724301707e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0 } },
 {  5.846280046914e-06,  2.571156499756e-06, {   4,   0,  -1,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  4.682101626424e-06,  3.734826262545e-06, {   4,   0,  -2,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  3.362105072512e-06,  5.054057281151e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,  -2,   0,   0,   0 } },
 {  6.777093176973e-06,  1.633733494681e-06, {   0,   1,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.201227351888e-06, -3.207934520542e-06, {   2,  -2,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -2.173685832388e-06, -6.233122456000e-06, {   2,   0,  -1,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 { -5.007675092091e-06, -3.396056082584e-06, {   0,   0,   4,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -2.053674288208e-06,  6.349806411279e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0 } },
 { -8.041375495572e-06,  3.608089234242e-07, {   0,   0,   4,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.160792234918e-06,  5.241208862004e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0 } },
 {  8.092044501064e-06, -3.065931806669e-07, {   0,   0,   3,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.299530661326e-06, -1.095523763205e-06, {   0,   0,   3,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 {  2.484640263362e-06,  5.909969223098e-06, {   0,   1,   1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.484640263362e-06, -5.909969223098e-06, {   0,   1,  -1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.922522820549e-06, -6.469395858072e-06, {   4,   0,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.669004070543e-06,  3.722413991000e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -2,   0,   0,   0 } },
 {  5.946718867074e-06, -2.443613316082e-06, {   1,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.664858029554e-06,  5.723059434117e-06, {   2,   0,   1,   0,   0,   0,   2,  -4,   1,   0,   0,   0,   0 } },
 { -5.296983686223e-06, -3.084443177682e-06, {   2,   0,   1,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0 } },
 { -2.751093891715e-06, -5.627206379361e-06, {   4,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -2.477825497630e-06, -5.897948606387e-06, {   2,   0,   0,   0,   0, -21,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.665606514547e-06, -1.709444116941e-06, {   0,   0,   1,   0,   0,   0,  17, -32,   0,   0,   0,   0,   0 } },
 { -7.524750383631e-06,  8.478280527607e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0 } },
 { -6.640359779368e-06, -1.732198945283e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,  -3,   0,   0,   0 } },
 {  2.477597092891e-07,  8.124642650602e-06, {   2,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.046060659184e-06, -6.325237610522e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -2.815319002530e-06, -5.553533961709e-06, {   3,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.791399008579e-06,  1.571357656435e-06, {   3,  -1,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.809550711383e-06, -6.553082026704e-06, {   2,   0,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.654976678904e-06,  1.706663048037e-06, {   0,   0,   1,   0,   0,   0, -17,  32,   0,   0,   0,   0,   0 } },
 {  5.236145008913e-06,  3.124807777640e-06, {   2,   0,   0,   0,   0,   2,   0,  -8,   0,   0,   0,   0,   0 } },
 {  2.513723173812e-06,  5.843079391688e-06, {   2,   0,   0,   0,   0,   0,  17, -24,   0,   0,   0,   0,   0 } },
 {  7.858714222430e-06,  4.952534129159e-07, {   2,   0,   2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.474991009986e-06, -6.878127345032e-06, {   1,  -1,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.532112942070e-06, -1.820537423322e-06, {   2,   0,   0,   0,   0,   0,   2,  -8,   5,   0,   0,   0,   0 } },
 {  6.746445802521e-06,  1.605165147184e-06, {   0,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.850108299828e-06,  6.500193287381e-06, {   2,   0,   0,   0,   0,   2,  -5,   3,   0,   0,   0,   0,   0 } },
 { -2.035642789045e-06,  6.311837062705e-06, {   2,   0,  -2,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -2.180707850384e-06, -6.166404316242e-06, {   2,   0,  -3,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -3.873328252933e-06,  4.473018262150e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -6.890805519023e-06,  1.453717302606e-06, {   0,   0,   1,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.738191440258e-06, -1.602058055163e-06, {   2,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.792489548463e-06,  1.545548667290e-06, {   2,  -2,   1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.500189444539e-06, -6.833363151594e-06, {   4,  -2,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.819919564962e-06,  5.513393676294e-06, {   2,  -2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  6.010311340939e-06,  2.321245140189e-06, {   2,   0,   0,   0,   0, -11,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.949529045020e-06, -5.381597126551e-06, {   4,   0,  -1,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  6.735334063726e-06,  1.594946352209e-06, {   3,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.200249620524e-06,  5.129563365211e-06, {   1,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  4.719423448621e-09, -8.324215964461e-06, {   1,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.260933860118e-06,  3.065495564562e-06, {   1,  -1,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  2.952182369994e-06,  5.373514454032e-06, {   5,   0,  -2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.952182369994e-06, -5.373514454032e-06, {   3,   0,  -2,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -6.601077005112e-06, -1.723620069650e-06, {   2,   0,   1,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.646543529646e-06, -6.678099808896e-06, {   1,  -1,   0,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 { -6.725156433288e-06,  1.597441928251e-06, {   2,   1,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.875904038870e-07,  7.733568777845e-06, {   1,  -1,   1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -5.875904038870e-07, -7.733568777845e-06, {   1,  -1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -6.588093567137e-06, -1.732820030680e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 {  8.061689133830e-08, -8.239041312779e-06, {   0,   0,   0,   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0 } },
 {  3.283834364823e-06, -5.033964006891e-06, {   2,   0,   0,   0,  -2,   7,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.726664940801e-06,  1.586694669815e-06, {   1,   1,   1,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.642060562649e-06, -1.670878089765e-06, {   0,   0,   1,   0,   0,  14, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.799722706895e-06,  3.512834741223e-06, {   4,  -2,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -6.418675790979e-06,  1.892300532480e-06, {   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.283175045161e-09,  8.306784970312e-06, {   3,   0,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  2.161342897954e-06, -6.144399735325e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   6, -10,   0,   0,   0 } },
 {  7.999590860867e-06,  3.033723254160e-07, {   2,  -2,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  7.617971975760e-06,  6.840093957700e-07, {   2,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -8.217099884957e-06,  8.296492977831e-08, {   2,   0,   1,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.963411744368e-06, -3.336276772041e-06, {   2,   0,  -1,   0,   0,  -6,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.212978019534e-06, -8.436664358076e-08, {   2,   2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.749486723397e-06, -1.547362181502e-06, {   0,   1,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.827942081122e-06, -2.468083993088e-06, {   4,   0,   0,   0,   0, -30,  24,   0,   0,   0,   0,   0,   0 } },
 { -5.827942081122e-06,  2.468083993088e-06, {   0,   0,   0,   0,   0,  30, -24,   0,   0,   0,   0,   0,   0 } },
 {  7.644618297034e-06, -6.499675888192e-07, {   2,   0,   1,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -7.644618297034e-06,  6.499675888192e-07, {   2,   0,  -3,   0,   0, -22,  28,  -7,   0,   0,   0,   0,   0 } },
 { -4.144795872978e-06, -4.149350901397e-06, {   0,   0,   2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.839861377866e-06, -3.453762356157e-06, {   1,  -1,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  2.489960010054e-06,  5.798795703728e-06, {   2,   0,   0,   0,   0,   0,  15, -22,   0,   0,   0,   0,   0 } },
 {  4.664031935753e-06,  3.624621233172e-06, {   2,   0,  -1,   0,   0,   0,  10, -18,   0,   0,   0,   0,   0 } },
 {  1.442813825331e-06,  6.843081833060e-06, {   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -3.181338033100e-06, -5.101027447059e-06, {   1,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.038916918707e-06, -4.243066627017e-06, {   1,   0,   0,   0,   0,   0,  -1,   4,  -3,   0,   0,   0,   0 } },
 {  3.553846810616e-08, -8.246357182502e-06, {   1,   0,   2,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.660731401902e-06, -3.617730301379e-06, {   2,   0,  -1,   0,   0,   0,  16, -24,   0,   0,   0,   0,   0 } },
 {  1.797490223930e-06, -6.479268690994e-06, {   2,   0,   0,   0,   0,   0,   7, -16,   6,   0,   0,   0,   0 } },
 { -3.855181705471e-06,  4.419579012574e-06, {   4,  -2,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  5.888575207420e-06, -2.381254860087e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 {  3.974939444563e-09, -8.263962951985e-06, {   2,   2,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.747576987895e-06,  5.518747482315e-06, {   2,   0,   0,   0,   0,  -2,   4,  -3,   0,   0,   0,   0,   0 } },
 { -5.969204637393e-06, -2.296763315234e-06, {   1,  -1,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.761944489426e-06, -5.500986930568e-06, {   1,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.848901104581e-06, -2.412102406027e-06, {   2,   0,  -2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.454077503017e-06,  5.802531936627e-06, {   3,   0,  -4,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  4.186656499611e-06,  4.069892974322e-06, {   4,   0,  -1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.486354675195e-06,  5.768275147135e-06, {   1,  -1,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  6.349189479727e-06,  1.904936578958e-06, {   1,   1,  -2,   0,   0, -25,  28,   0,   0,   0,   0,   0,   0 } },
 {  4.280719717275e-07,  7.825833468743e-06, {   2,   0,   1,   0,   0,  15, -17,   0,   0,   0,   0,   0,   0 } },
 {  6.955039991448e-06, -1.297587388406e-06, {   1,   1,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.443174109946e-06,  6.808055541116e-06, {   1,  -1,   1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  5.024146083750e-06,  3.224850685529e-06, {   1,   1,  -1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.655498625791e-06, -3.592740862524e-06, {   1,  -1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.173191868428e-06, -3.074812284337e-06, {   1,   1,   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  3.550983652186e-06,  4.696568173986e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,  -1,   0,   0,   0,   0 } },
 { -2.116247639053e-06, -6.127257157332e-06, {   1,  -1,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.339645457816e-06, -2.900542392805e-06, {   1,  -1,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 {  3.100995986734e-06, -5.132733830282e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0 } },
 {  6.931470102708e-06, -1.298632264068e-06, {   2,   0,   0,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0 } },
 { -5.856282030038e-06, -2.369856515618e-06, {   2,   0,   0,   0,   0,   3,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.071162851828e-07,  7.318870471320e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 {  4.277351031745e-06, -3.945197414731e-06, {   1,   1,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.483674495486e-07,  7.573747539036e-06, {   1,  -1,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.837933857821e-06,  4.379719685205e-06, {   0,   0,   1,   0,   0,   0,  -6,  12,  -3,   0,   0,   0,   0 } },
 {  6.657046230385e-06, -1.559916465436e-06, {   1,   1,   1,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -6.726165966319e-06,  1.490751845850e-06, {   3,  -1,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.246405445981e-06,  4.964653733711e-06, {   2,   0,   0,   0,   0,   0,  22, -27,   0,   0,   0,   0,   0 } },
 {  9.603697078596e-07,  7.250117731748e-06, {   0,   0,   0,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 {  5.363182875427e-08,  8.155360597961e-06, {   2,   0,   0,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 {  6.886454257386e-06,  1.322251155325e-06, {   3,  -1,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -5.990987472309e-06,  2.217351053936e-06, {   2,   0,  -1,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0 } },
 { -4.565397554136e-06,  3.641556175000e-06, {   2,   0,  -3,   0,   1,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  2.331581183963e-06, -5.875046278536e-06, {   2,   0,   1,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 {  4.526931634817e-06,  3.678857094427e-06, {   2,   0,  -3,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  1.286878710047e-08, -8.192837870308e-06, {   0,   2,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.438510051327e-06,  4.766181668300e-06, {   1,   1,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  8.063330234662e-06, -1.411226024116e-07, {   0,   2,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.845628853839e-06, -2.358367049351e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 {  3.413663363517e-06, -4.786601941777e-06, {   2,  -2,   1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 { -7.360842312095e-06,  8.387109726097e-07, {   2,   0,   0,   0,   0,  22, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.085513845627e-07,  8.088951664019e-06, {   1,   2,  -2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  7.366418388746e-06, -8.211459660300e-07, {   6,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.249447662333e-06, -9.345634283260e-07, {   2,   0,   0,   0,   0,   0, -12,  23,  -3,   0,   0,   0,   0 } },
 {  3.831211756822e-06, -4.351769038462e-06, {   0,   0,   1,   0,   0,   0,   6, -12,   3,   0,   0,   0,   0 } },
 {  6.491199413563e-06, -1.691435018682e-06, {   4,   0,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0 } },
 { -6.032164218551e-06,  2.149661407638e-06, {   2,   0,   0,   0,   0,   3,  -4,   0,   2,   0,   0,   0,   0 } },
 { -2.708260131192e-06, -5.471767821221e-06, {   3,   0,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.926910189874e-06,  2.488859174736e-07, {   2,   0,  -3,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.599170696497e-06, -2.573280601321e-06, {   1,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -5.589443074246e-06,  2.581159534459e-06, {   1,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  6.600817923804e-06,  1.568119985663e-06, {   1,   1,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  3.776608782631e-07, -7.788871623394e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  7.522609966034e-06,  6.437238128193e-07, {   2,   0,   2,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.351342964489e-08,  8.151226134305e-06, {   0,   0,   2,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -4.713737665447e-06,  3.448957987626e-06, {   2,   0,  -2,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 { -4.711197993571e-06, -3.451228819688e-06, {   2,   0,  -2,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -6.184220398885e-06,  1.977572357128e-06, {   2,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.678436244955e-06,  3.482393888068e-06, {   2,  -2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  3.066996252498e-06, -5.092721701176e-06, {   1,   1,  -1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.783992676744e-07,  7.577896211778e-06, {   0,   0,   2,   0,   0,   0, -15,  23,   0,   0,   0,   0,   0 } },
 {  6.932625918969e-06,  1.223255224360e-06, {   2,   0,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.602741966956e-06, -1.550367365876e-06, {   3,   1,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.692878641002e-06,  5.459391893474e-06, {   4,   0,  -2,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.450325791990e-06,  5.700652323294e-06, {   2,   0,   0,   0,   0,   0,  16, -23,   0,   0,   0,   0,   0 } },
 {  3.086026567190e-06, -5.064038763902e-06, {   1,   0,   0,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  7.574374235184e-06, -5.743655064285e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -7,   0,   0,   0 } },
 {  3.545185270632e-06, -4.600876732978e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -6,   5,   0,   0,   0 } },
 { -2.535215085968e-06, -5.606653674063e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,  -2,   0 } },
 {  4.558851936598e-07, -7.684641224050e-06, {   2,   0,  -1,   0,   0,   5,  -7,   2,   0,   0,   0,   0,   0 } },
 { -1.014791050578e-07,  8.038272586279e-06, {   0,   0,   4,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.425947519593e-06,  5.713508001013e-06, {   2,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.540068265368e-06,  6.597421719449e-06, {   4,   0,  -1,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 {  1.539060667328e-06, -6.596157175587e-06, {   4,   0,  -1,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 {  2.290672125055e-06, -5.843912885998e-06, {   0,   0,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0 } },
 {  1.667783550197e-06, -6.465287957512e-06, {   2,  -2,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.397351871343e-06,  2.728331164519e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -4,   5,   0,   0,   0 } },
 { -4.448503939868e-06, -3.676963054176e-06, {   1,  -1,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.227252942423e-07,  7.502673088400e-06, {   3,  -1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.863422796605e-06, -6.260223708404e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -7.960679409492e-06, -1.609846783754e-07, {   0,   2,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  4.940460275272e-06, -3.180324605020e-06, {   2,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.072071463531e-06,  3.045784823291e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.703522844110e-06,  5.413716464842e-06, {   0,   2,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.231207152998e-08, -8.101739079019e-06, {   2,   0,   1,   0,   0,   0,   3,   0,   0,  -3,   0,   0,   0 } },
 { -5.327538830048e-06, -2.785781738765e-06, {   4,   0,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  7.954577681994e-06,  1.582985801023e-07, {   1,   0,  -4,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  8.068428319400e-06, -4.307997169380e-08, {   2,   0,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.416521161718e-06, -6.693838989997e-06, {   2,   0,  -1,   0,   0,   9, -13,   0,   2,   0,   0,   0,   0 } },
 {  3.596450792052e-08,  8.071569380588e-06, {   2,  -2,   0,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 {  4.070197968875e-06, -4.031665440758e-06, {   2,   0,  -1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.443528166273e-06, -2.656888097673e-06, {   2,   0,   0,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0 } },
 { -4.309784728443e-06, -3.790499521046e-06, {   0,   0,   1,   0,   0,   0,  10, -19,   1,   0,   0,   0,   0 } },
 { -5.817637578448e-06,  2.280025973682e-06, {   1,   0,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.043532117961e-07, -7.791322475102e-06, {   0,   0,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.433011086856e-08, -8.081166021226e-06, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0 } },
 {  7.362578216773e-06, -7.317917788996e-07, {   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  2.898639044098e-06, -5.192700985562e-06, {   2,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.645817520500e-06, -6.443818208476e-06, {   1,  -1,   0,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 { -6.562006575737e-06,  1.526940581877e-06, {   1,  -1,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -6.025268160064e-06, -2.060670305967e-06, {   2,   0,  -4,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -4.301834564186e-06,  3.782272528232e-06, {   0,   0,   1,   0,   0,   0, -10,  19,  -1,   0,   0,   0,   0 } },
 { -1.682470899395e-06, -6.398245700592e-06, {   4,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  9.547327879233e-07, -7.125830192701e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.708585693486e-06,  3.711332952233e-07, {   1,  -1,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  7.829584915466e-06,  2.482946006965e-07, {   4,   0,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 { -1.922009079583e-06, -6.154331803904e-06, {   2,   0,   3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.428667920185e-06,  5.647264406039e-06, {   1,   0,  -1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 { -1.921178849871e-06,  6.153341597775e-06, {   2,   0,   3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.255761001140e-06,  6.816899608902e-06, {   2,   0,   0,   0,   0,   0,   5,   0,  -7,   5,   0,   0,   0 } },
 { -6.370317182468e-06,  1.700955299966e-06, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0 } },
 { -3.883397944931e-07,  7.681099748000e-06, {   4,   0,  -2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.355180450261e-07,  7.730931226082e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0 } },
 { -1.706923082075e-07,  7.894031377464e-06, {   1,   1,   0,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.521396703988e-06,  1.541227328458e-06, {   0,   1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.947461317702e-06, -1.115132134558e-06, {   2,   0,  -1,   0,   0,  19, -22,   0,   0,   0,   0,   0,   0 } },
 {  7.948032083071e-06, -1.136721212325e-07, {   1,   0,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.391353811604e-06, -5.667467892232e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -4,   2,   0,   0,   0 } },
 {  4.604412257433e-06,  3.446505009915e-06, {   0,   0,   1,   0,   0,   8,  -3, -19,   0,   3,   0,   0,   0 } },
 {  4.604412257433e-06, -3.446505009915e-06, {   0,   0,   1,   0,   0,  -8,   3,  19,   0,  -3,   0,   0,   0 } },
 { -2.750009259002e-06, -5.300687231090e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -6,   0,   0,   0,   0 } },
 { -6.526825955830e-06, -1.522277781020e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0 } },
 {  7.803855412290e-06, -2.414217602583e-07, {   2,   0,  -2,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  7.450661918330e-07,  7.295074295132e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -4,   4,  -2,   0,   0 } },
 {  7.450661918330e-07, -7.295074295132e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -4,   2,   0,   0 } },
 {  2.850280408179e-06,  5.188034156382e-06, {   3,   0,   2,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.850280408179e-06, -5.188034156382e-06, {   1,   0,   2,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 {  6.830077574248e-06,  1.207752735047e-06, {   1,   1,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.069728580010e-08,  8.027062386990e-06, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,  -2,   0,   0,   0 } },
 {  2.137155664476e-06, -5.900145051419e-06, {   5,   0,  -4,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  1.755430645045e-06, -6.280805585794e-06, {   0,   0,   2,   0,   0, -24,  26,   0,  -3,   0,   0,   0,   0 } },
 { -2.642651410666e-06,  5.393132627427e-06, {   1,   0,   0,   0,   0,   9,  -7,   0,   0,   0,   0,   0,   0 } },
 {  7.661528843112e-06,  3.740764354766e-07, {   3,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.677896237684e-07,  7.267263761131e-06, {   2,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.919471527525e-06,  3.115422709960e-06, {   1,   1,  -1,   0,   0,   1,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.953996584577e-06,  1.079479658688e-06, {   4,   0,  -3,   0,   0,   3,  -3,   0,   0,  -3,   0,   0,   0 } },
 {  6.953996584577e-06, -1.079479658688e-06, {   0,   0,   1,   0,   0,  -3,   3,   0,   0,   3,   0,   0,   0 } },
 {  5.558872794697e-06, -2.472493506319e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 {  7.905619627892e-06,  1.250547531904e-07, {   2,   0,   1,   0,   0,  14, -13,   0,   0,   0,   0,   0,   0 } },
 {  3.785117023308e-06,  4.242571172087e-06, {   2,   0,  -1,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.831458968461e-06,  1.196009692255e-06, {   2,   0,  -2,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.692228168237e-06,  6.335083238230e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -3,   7,   0,   0,   0 } },
 { -6.182611679709e-06,  1.842790412509e-06, {   2,   0,  -1,   0,   0,   6, -11,   0,   0,   0,   0,   0,   0 } },
 { -1.523786138689e-06,  6.500523313614e-06, {   0,   0,   1,   0,   0, -17,  19,   0,   0,   0,   0,   0,   0 } },
 { -2.392328991491e-06,  5.630349125306e-06, {   0,   0,   5,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.392328991491e-06, -5.630349125306e-06, {   0,   0,   3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.475556213817e-06,  1.546151659837e-06, {   0,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -8.959923145067e-09,  8.011015702078e-06, {   1,   0,   1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -5.187912803768e-06, -2.830711179720e-06, {   0,   0,   1,   0,   0,   0,  -5,  10,  -5,   3,   0,   0,   0 } },
 { -5.187784343277e-06,  2.830721669907e-06, {   0,   0,   1,   0,   0,   0,   5, -10,   5,  -3,   0,   0,   0 } },
 {  1.956125441592e-06, -6.062337774784e-06, {   2,   0,   0,   0,   0,   6,  -9,   0,   3,   0,   0,   0,   0 } },
 {  2.062281656201e-08,  7.994649137615e-06, {   0,   0,   2,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  3.170732932553e-06, -4.843903733490e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   1,  -3,   0,   0 } },
 {  6.948422944478e-06,  1.065282957843e-06, {   2,   0,   1,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  3.150553600306e-07, -7.696386771237e-06, {   2,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0,   0 } },
 { -6.294344081721e-06,  1.716298065647e-06, {   2,   0,   1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.411490199356e-06,  1.596111836471e-06, {   2,  -2,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  7.594844765016e-06,  4.123824034276e-07, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 { -5.698495784200e-06,  2.306996397533e-06, {   1,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.521821528615e-06,  2.482588332812e-06, {   2,   0,  -1,   0,   0,   0,  11,  -9,   0,   0,   0,   0,   0 } },
 { -6.223974211675e-06,  1.780113326581e-06, {   2,   0,  -1,   0,   0,   0,  -3,   6,  -3,   0,   0,   0,   0 } },
 { -2.402478614805e-06,  5.601150713719e-06, {   1,   1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  4.187838413774e-06, -3.814611607270e-06, {   0,   0,   1,   0,   0,   0,  -7,  16,  -6,   0,   0,   0,   0 } },
 { -7.352976737367e-07, -7.267152299603e-06, {   0,   0,   1,   0,   0,   0, -12,  18,   0,   0,   0,   0,   0 } },
 { -1.445398151051e-06, -6.556905811540e-06, {   1,   1,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.442147596763e-06,  2.559454065147e-06, {   2,   0,  -4,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -5.380276361902e-06, -2.619824142540e-06, {   2,   0,  -1,   0,   0,   5,  -2, -12,   3,   0,   0,   0,   0 } },
 { -5.380276361902e-06,  2.619824142540e-06, {   2,   0,  -1,   0,   0,  -5,   2,  12,  -3,   0,   0,   0,   0 } },
 {  6.163021080509e-06, -1.833519896540e-06, {   2,   0,   0,   0,   0,  -6,  10,   0,  -5,   5,   0,   0,   0 } },
 {  5.851227335580e-06, -2.144348243817e-06, {   1,  -1,   1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.844100808104e-06, -2.147620259363e-06, {   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0,   0,   0 } },
 { -3.115078340927e-06,  4.876356858761e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   5,   0,   0,   0 } },
 {  2.501411223963e-06, -5.486908267525e-06, {   1,   1,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -6.212394860099e-06,  1.774814242420e-06, {   2,   0,  -1,   0,   0,   0,  15, -23,   0,   0,   0,   0,   0 } },
 { -7.691221643509e-06, -2.958785618808e-07, {   2,  -2,   0,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  6.525826966809e-08,  7.921633544510e-06, {   0,   2,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -4.114255219070e-07, -7.575224014583e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0 } },
 { -7.595315951301e-07,  7.226840328050e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   3,   0,   0 } },
 {  2.388968423441e-06,  5.596574627323e-06, {   2,   0,   0,   0,   0,   0,  14, -21,   0,   0,   0,   0,   0 } },
 { -7.744121417154e-06,  2.413999713459e-07, {   4,   0,  -1,   0,   0, -19,  20,  -3,   0,   0,   0,   0,   0 } },
 {  7.207031241384e-06, -7.770897316800e-07, {   2,   0,   2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.455702753127e-06, -1.527029297199e-06, {   1,  -1,  -2,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.187643398903e-06,  2.793382859693e-06, {   0,   0,   2,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  2.059647178938e-06,  5.919777625250e-06, {   2,   0,  -4,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 {  7.499690159244e-06,  4.755814511276e-07, {   4,   0,  -1,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.100235463824e-07,  7.362243495206e-06, {   3,  -2,   0,   0,   0,  23, -16, -13,   0,   0,   0,   0,   0 } },
 { -6.100235463824e-07, -7.362243495206e-06, {   1,   2,   0,   0,   0, -23,  16,  13,   0,   0,   0,   0,   0 } },
 {  1.395298283784e-06, -6.576401866423e-06, {   1,   1,  -1,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.396123654103e-06, -1.572669949177e-06, {   1,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -4.413021809443e-06,  3.552585725731e-06, {   0,   0,   0,   0,   0,  11,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.915628218442e-06,  5.048847817013e-06, {   2,  -2,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -5.158260940330e-06,  2.805227122042e-06, {   2,   0,   0,   0,   0,   0,  22, -26,   0,   0,   0,   0,   0 } },
 {  1.108597118872e-06,  6.854655252672e-06, {   1,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -3.587076175313e-09,  7.956996796977e-06, {   6,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.108066954472e-07,  7.849096942441e-06, {   2,   0,  -1,   0,   0,   0,   6,  -8,   2,   0,   0,   0,   0 } },
 { -8.823689585483e-07,  7.076558002588e-06, {   2,   0,   0,   0,   0, -25,  27,   0,   0,   0,   0,   0,   0 } },
 { -5.705127347026e-06, -2.251412427408e-06, {   2,   0,  -2,   0,   0,  -4,   7,   0,   0,   0,   0,   0,   0 } },
 { -6.708015601146e-06,  1.245032629437e-06, {   1,  -1,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -6.408681802611e-07,  7.310450970945e-06, {   2,   0,   0,   0,   0,   0,  20, -26,   0,   0,   0,   0,   0 } },
 { -6.398882212111e-07, -7.307158756412e-06, {   1,   1,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -6.206965748486e-06,  1.738768665912e-06, {   1,   1,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.144314731907e-07, -7.629541242468e-06, {   1,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  1.512034081645e-06, -6.429548065694e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -7.108395629619e-06, -8.327660344543e-07, {   2,   0,  -1,   0,   0,   0, -12,  23,  -3,   0,   0,   0,   0 } },
 { -7.139348600251e-06, -8.013159298724e-07, {   2,   0,   0,   0,   0,   0,   0,   0,  -5,  10,   0,   0,   0 } },
 { -5.608075056604e-06,  2.328648565561e-06, {   1,  -1,   3,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  5.608075056604e-06, -2.328648565561e-06, {   1,  -1,  -3,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  5.328579747191e-06,  2.607103737600e-06, {   0,   0,   1,   0,   0,  -6,   9,   0,  -3,   0,   0,   1,   0 } },
 { -6.711228531598e-06,  1.224452675663e-06, {   4,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.818727895567e-06,  2.116619954359e-06, {   1,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  7.361803622181e-06, -5.717882925189e-07, {   0,   0,   2,   0,   0,  -3,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.823525952110e-06, -4.109776870450e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 { -6.587216624115e-06,  1.344346215241e-06, {   1,   0,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -1.946335975524e-06, -5.985110489392e-06, {   2,   0,   0,   0,   0,  14, -18,   0,   0,   0,   0,   0,   0 } },
 {  2.819362781159e-06, -5.111850245237e-06, {   1,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -6.010416578080e-06, -1.918747672511e-06, {   2,  -2,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -6.104043448700e-06, -1.819060896271e-06, {   0,   0,   1,   0,   0,  -6,  14,  -8,   0,   0,   0,   0,   0 } },
 { -6.100174281250e-06,  1.821642886950e-06, {   0,   0,   1,   0,   0,   6, -14,   8,   0,   0,   0,   0,   0 } },
 { -2.593791830261e-06, -5.326333343338e-06, {   1,   0,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.165935345083e-06,  3.752529047711e-06, {   0,   0,   1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0 } },
 {  1.521805327509e-06, -6.390211033043e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 { -6.386168300346e-06,  1.525096278672e-06, {   3,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.728824776863e-06,  3.181363765346e-06, {   0,   2,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -1.643339116161e-06,  6.265741204531e-06, {   1,  -1,   1,   0,   0,  -3,   7,  -2,   0,   0,   0,   0,   0 } },
 {  1.643339116161e-06, -6.265741204531e-06, {   1,  -1,  -1,   0,   0,  -3,   7,  -2,   0,   0,   0,   0,   0 } },
 {  4.822897722119e-06, -3.086103545901e-06, {   0,   2,   0,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  6.095924379691e-06,  1.811518042356e-06, {   2,   0,   0,   0,   0,   6, -10,   0,   5,  -5,   0,   0,   0 } },
 { -3.390707567979e-06, -4.514512749357e-06, {   4,  -2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.368494803250e-06,  6.535291759008e-06, {   4,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.177523607633e-06, -2.724570540095e-06, {   0,   2,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -7.716278526640e-06, -1.826082493894e-07, {   4,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.512202689436e-06,  5.384929402053e-06, {   4,   0,  -1,   0,   0, -29,  36,   0,  -4,   0,   0,   0,   0 } },
 {  2.512202689436e-06, -5.384929402053e-06, {   0,   0,   1,   0,   0,  29, -36,   0,   4,   0,   0,   0,   0 } },
 { -1.136994037191e-06,  6.759701386120e-06, {   1,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.654531771841e-06, -3.242011920455e-06, {   4,  -2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -6.412455650307e-06,  1.483676694052e-06, {   1,   1,  -1,   0,   0,   0,   1,   0,   2,   0,   0,   0,   0 } },
 { -7.631009100988e-09,  7.886945212149e-06, {   0,   0,   3,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.520079520176e-07,  7.141634865119e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  4.712590546661e-07, -7.422350023802e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -6,   3,   0,   0 } },
 { -7.591794983594e-06,  3.000556511214e-07, {   2,   2,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.146888701939e-06, -4.743974477450e-06, {   1,   0,   0,   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0 } },
 {  1.179337276696e-09,  7.885064467005e-06, {   0,   0,   3,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -3.246415926145e-06, -4.639541884793e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0 } },
 {  5.147692673655e-06,  2.738222444919e-06, {   2,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.797449384876e-09,  7.878978938143e-06, {   4,  -2,  -3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.755569743586e-06,  1.268409482374e-07, {   2,   0,  -2,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -7.480358843840e-06,  4.007108738058e-07, {   0,   0,   2,   0,   0,   0,  -8,  11,   0,   0,   0,   0,   0 } },
 { -4.193161966962e-06,  3.685868798389e-06, {   2,   0,  -1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.026926498107e-06,  6.851324476565e-06, {   1,   0,   1,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  7.691619974773e-06, -1.865185802460e-07, {   2,   0,  -1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 {  6.183248350406e-06,  1.694670619041e-06, {   1,   1,  -1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.039252023239e-06, -8.333102795795e-07, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   5,   0,   0,   0 } },
 {  2.592672886698e-06,  5.278265160420e-06, {   4,   0,  -1,   0,   0, -13,  11,   0,   0,   0,   0,   0,   0 } },
 {  7.356129669177e-07, -7.133605426949e-06, {   2,   0,  -2,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  4.376810688055e-06, -3.490997002199e-06, {   2,   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0 } },
 { -1.788348029408e-06,  6.079053196236e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 { -2.588326106339e-06,  5.277251561691e-06, {   1,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.737165927597e-07,  7.391102830014e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   6,  -3,   0,   0 } },
 { -5.248278395961e-06, -2.615989558999e-06, {   1,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -6.362832690188e-06,  1.501386701868e-06, {   0,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.450506990671e-06,  5.411210678802e-06, {   2,   0,   1,   0,   0,  10,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.347592637599e-06, -6.508859675748e-06, {   1,   1,   0,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 {  4.185077222674e-06, -3.670780039764e-06, {   2,   0,   0,   0,   0,  -9,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.585667265421e-06, -5.268817489645e-06, {   0,   0,   1,   0,   0, -21,  19,   0,   0,   0,   0,   0,   0 } },
 {  7.489729381379e-06, -3.625356352116e-07, {   2,   0,   0,   0,   0,   0,  -3,   0,   3,  -2,   0,   0,   0 } },
 { -4.615413382795e-06,  3.233034509863e-06, {   2,   0,   0,   0,   0,   1,  -9,  12,   0,   0,   0,   0,   0 } },
 { -2.280379617118e-06,  5.564219682400e-06, {   2,   0,   0,   0,   0,  -9,  13,   0,   0,   0,   0,   0,   0 } },
 { -1.481978422472e-07, -7.695502237176e-06, {   2,   0,  -1,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0 } },
 {  5.303806907223e-06,  2.536092501136e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   7,  -5,   0,   0,   0 } },
 { -1.530721783728e-06,  6.304023046161e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,  -1,   0,   0,   0 } },
 { -7.410676596420e-06, -4.232605825316e-07, {   4,   0,  -2,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.127293668576e-06, -1.705780632281e-06, {   0,   0,   3,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -4.714469147069e-06,  3.118101956872e-06, {   3,  -1,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  6.124740917989e-06,  1.705424722920e-06, {   0,   0,   3,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  3.995515393198e-06,  3.833455108658e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   7,   0,   0,   0 } },
 { -4.354165669008e-06,  3.474451092244e-06, {   2,   0,   1,   0,   0,  -8,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.007779822258e-06,  3.819181868705e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 { -7.230950237236e-06, -5.954211252920e-07, {   2,   0,  -2,   0,   0,  -3,   5,   0,  -1,   0,   0,   0,   0 } },
 {  7.426667915131e-06,  3.989161459994e-07, {   2,  -2,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.563369401596e-08, -7.808879622635e-06, {   3,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 {  7.261287856299e-06,  5.621751079873e-07, {   2,  -2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -6.894708446600e-06, -9.281448205798e-07, {   2,   0,   3,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.431026254654e-06, -5.391256549222e-06, {   2,   0,   1,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0 } },
 { -5.200775940537e-06, -2.621282633916e-06, {   2,   0,  -3,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -2.094742894740e-06, -5.725886665129e-06, {   1,   0,  -2,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.201402538143e-06,  1.619122221312e-06, {   2,   0,  -1,   0,   0,   0,  -7,  12,  -3,   0,   0,   0,   0 } },
 { -3.984604025373e-08, -7.778377272981e-06, {   1,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.432393766692e-08,  7.761767216161e-06, {   2,  -2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -2.498296958161e-06,  5.315135131518e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -8,   0,   0,   0 } },
 {  2.565574597122e-06, -5.246483445923e-06, {   2,   0,  -3,   0,   0,   0,  -3,   9,   0,   0,   0,   0,   0 } },
 {  2.597788945268e-06,  5.212056656171e-06, {   2,   0,   1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  2.727707607694e-06, -5.081082291910e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   4,   0,   0,   0 } },
 { -6.465480737416e-06,  1.337017529527e-06, {   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  7.028128531628e-07,  7.097449458368e-06, {   2,   0,  -1,   0,   0,   3, -10,  13,   0,   0,   0,   0,   0 } },
 { -2.161822675884e-06,  5.636483616288e-06, {   0,   0,   1,   0,   0, -10,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.247895179755e-06, -6.549477755029e-06, {   1,   1,  -2,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.398639128016e-06, -6.398225738008e-06, {   6,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.468728464988e-06,  3.327679328794e-06, {   2,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.679108357769e-06,  1.164203377246e-07, {   2,   0,   1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  6.610160184714e-08,  7.728768714399e-06, {   2,   0,  -1,   0,   0,  28, -28,   0,   0,   0,   0,   0,   0 } },
 {  4.638675169131e-06,  3.154969853877e-06, {   2,   0,  -1,   0,   0,  -4,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.353153172693e-06,  3.440486210034e-06, {   2,   0,   0,   0,   0,   0,  -5,  10,  -3,   0,   0,   0,   0 } },
 {  3.478760512405e-07, -7.444428561465e-06, {   0,   0,   2,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  1.143196254578e-06,  6.647318855989e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 } },
 { -6.329237300104e-06,  1.459678241305e-06, {   4,   1,  -3,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.211605435527e-06,  5.766174923052e-07, {   4,   0,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  6.922647535736e-07,  7.095516561859e-06, {   0,   0,   2,   0,   0,  -2,  -1,   8,   0,   0,   0,   0,   0 } },
 { -5.445195856403e-06,  2.341651199002e-06, {   1,   0,  -1,   0,   0,  29, -37,   8,  -3,   0,   0,   0,   0 } },
 {  5.445195856403e-06, -2.341651199002e-06, {   1,   0,  -3,   0,   0,  29, -37,   8,  -3,   0,   0,   0,   0 } },
 {  5.146661319630e-06,  2.639258721263e-06, {   2,  -2,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -3.303942639966e-06, -4.481806316374e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,  -5,   0,   0,   0 } },
 {  1.369965192902e-07,  7.647974947459e-06, {   4,   0,  -2,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  6.036916646589e-06,  1.746503817516e-06, {   2,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 {  6.990711960120e-07, -7.082140817321e-06, {   0,   0,   2,   0,   0,   2,   1,  -8,   0,   0,   0,   0,   0 } },
 { -8.926376189806e-08, -7.691132284974e-06, {   4,   0,  -1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -6.813398618625e-06,  9.661325225788e-07, {   0,   0,   1,   0,   0,   8, -13,   0,   2,  -5,   0,   0,   0 } },
 { -6.813398618625e-06, -9.661325225788e-07, {   0,   0,   1,   0,   0,  -8,  13,   0,  -2,   5,   0,   0,   0 } },
 {  3.107225850692e-06, -4.671924713018e-06, {   1,  -1,  -2,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -5.145687895981e-07, -7.264208007382e-06, {   2,   0,   1,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0 } },
 {  6.545872520750e-06, -1.231139324391e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   1,  -2,   0,   0,   0 } },
 { -3.928534399325e-06,  3.847476242134e-06, {   0,   0,   0,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 {  1.032814734303e-06, -6.743043543266e-06, {   1,  -1,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.977779683444e-09, -7.766795376425e-06, {   4,  -2,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.493027633705e-07, -7.119345978298e-06, {   2,   0,  -1,   0,   0,   0,  19, -25,   0,   0,   0,   0,   0 } },
 { -3.069794402346e-06, -4.697516374204e-06, {   2,   0,  -2,   0,   0,  18, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.651938083393e-06,  6.112248054523e-06, {   2,   0,  -1,   0,   0, -13,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.498985238836e-06, -5.264017645470e-06, {   2,   0,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 {  3.179883231947e-07, -7.443202151650e-06, {   2,   0,  -2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.308056437940e-06, -3.453079109579e-06, {   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -3.094899430209e-06,  4.665921997435e-06, {   1,   0,   0,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0 } },
 {  5.045024121291e-06,  2.715444755333e-06, {   0,   0,   1,   0,   0,  -2,   3,   0,  -2,   0,   0,   0,   0 } },
 { -2.115799031070e-06, -5.641150266325e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0 } },
 {  4.157250012895e-06, -3.599030135211e-06, {   2,   0,  -1,   0,   0,   0,   6, -12,   3,   0,   0,   0,   0 } },
 {  4.321453453874e-06, -3.434825216080e-06, {   1,  -1,   1,   0,   0,   0,   2,   0,   0,   1,   0,   0,   0 } },
 {  7.422685836183e-06,  3.327979720002e-07, {   0,   2,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -3.104504252039e-06,  4.649261517926e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0 } },
 { -4.232558375887e-06, -3.518872004147e-06, {   1,  -1,   0,   0,   0,   9,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.316929791849e-06, -3.430815679038e-06, {   4,   0,  -2,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  4.396178299906e-06,  3.349701590877e-06, {   2,   0,   0,   0,   0,   0,  10, -18,   0,   0,   0,   0,   0 } },
 { -2.100251732179e-06,  5.643304923668e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0 } },
 { -1.411394534550e-06, -6.331170466340e-06, {   2,   0,  -5,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.620664476406e-06,  1.213253414731e-07, {   0,   0,   3,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -2.070658250770e-06,  5.669527653116e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   2,   0,   0,   0 } },
 { -7.485177463359e-06,  2.537629880641e-07, {   0,   2,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.515225829263e-06,  5.223397666657e-06, {   1,  -1,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -7.569373485559e-06,  1.672025744063e-07, {   4,   0,  -1,   0,   0, -19,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.258257917313e-07,  7.308297402528e-06, {   2,   0,   1,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 {  6.524980882057e-07,  7.078772055038e-06, {   1,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -4.018174220862e-06, -3.711288619237e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   1,  -5,   0,   0,   0 } },
 { -7.473720566658e-06,  2.532763222101e-07, {   1,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.522299219488e-06,  1.201713635112e-06, {   2,   0,  -1,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.784956642700e-06, -5.935328289163e-06, {   2,   0,   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  4.143971237549e-06,  3.575703748238e-06, {   0,   0,   1,   0,   0,   0,  -5,   0,   2,   0,   0,   0,   0 } },
 {  6.639690447703e-06, -1.077123549220e-06, {   2,   0,   0,   0,   1,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.523777440799e-06,  3.191117950875e-06, {   0,   0,   0,   0,   0,   0,  11, -19,   0,   0,   0,   0,   0 } },
 { -6.081814754970e-06, -1.632042607018e-06, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,  -3,   0,   0,   0 } },
 {  7.838668935912e-07,  6.929897769306e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   1,   5,   0,   0,   0 } },
 {  6.503823339909e-06, -1.208675616213e-06, {   1,   1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  5.323023239974e-06, -2.388114531730e-06, {   2,   0,   0,   0,   0,   0,  13, -11,   0,   0,   0,   0,   0 } },
 {  6.207618125500e-06,  1.502082999761e-06, {   4,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.293466328040e-06,  2.415616538164e-06, {   0,   2,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -3.090050348779e-07,  7.397532473922e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -7.701351778717e-06,  3.062726204241e-09, {   6,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.662133785407e-06, -2.039776056139e-06, {   0,   0,   2,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.070983190648e-06,  3.629857897125e-06, {   1,   0,   1,   0,   0,   0,  33, -41,   2,   0,   0,   0,   0 } },
 { -3.701306085589e-07,  7.321178693485e-06, {   4,  -2,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.955856565014e-06,  4.734187530180e-06, {   1,   0,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.913074939498e-06,  2.776794569843e-06, {   0,   0,   1,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0 } },
 {  1.335411733326e-06, -6.354155699586e-06, {   1,  -1,   0,   0,   0,  -8,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.724059342707e-06, -3.961695842619e-06, {   2,   0,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0 } },
 {  2.086925551413e-06,  5.596193724883e-06, {   0,   0,   2,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -2.343175061974e-06,  5.339758070264e-06, {   1,  -1,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.699882659464e-07, -7.208345496917e-06, {   0,   0,   2,   0,   0,   0,  14, -27,   3,   3,   0,   0,   0 } },
 { -4.699882659464e-07,  7.208345496917e-06, {   0,   0,   2,   0,   0,   0, -14,  27,  -3,  -3,   0,   0,   0 } },
 { -2.576171278502e-06,  5.100494798084e-06, {   1,  -1,   1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.576171278502e-06, -5.100494798084e-06, {   1,  -1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  1.305269759448e-06,  6.371146342584e-06, {   2,   0,  -3,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.952747741872e-07, -6.978106472563e-06, {   2,   0,  -1,   0,   0,   2,  -8,   7,   0,   0,   0,   0,   0 } },
 { -2.081880351779e-06,  5.590977047778e-06, {   1,   0,  -3,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 {  7.082783160943e-06, -5.844348566012e-07, {   1,   0,  -2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -4.354813990916e-09, -7.660228627956e-06, {   4,   0,  -1,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.534391414438e-06, -4.129411246102e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -5,   0,   0,   0 } },
 { -3.176075282580e-06, -4.487681296486e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   5,   0,   0,   0 } },
 { -1.231351379671e-06, -6.432200829122e-06, {   1,  -1,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.170634806738e-06,  4.916587325124e-07, {   4,   0,  -3,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.123237813443e-06, -6.537063445312e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0 } },
 { -1.547573382178e-06, -6.112519043052e-06, {   1,  -1,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.761199351840e-06, -2.897494273685e-06, {   3,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -1.544687031066e-07, -7.502622115290e-06, {   4,   0,  -1,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  6.012664380438e-06,  1.643483411190e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 } },
 { -4.218038835443e-06,  3.437317615522e-06, {   3,   1,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.085276485340e-06,  6.566589580248e-06, {   2,   0,  -2,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0 } },
 {  2.174581538905e-07,  7.433572821168e-06, {   4,   0,  -3,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 {  6.154112523999e-06,  1.496577190153e-06, {   3,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  5.047938472540e-06,  2.601909395987e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  7.156764755601e-06,  4.928565527926e-07, {   4,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.388450586597e-06,  2.260966236683e-06, {   1,  -1,  -2,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  2.305095971356e-06,  5.343587964076e-06, {   2,   0,   0,   0,   0,   0,  18, -25,   0,   0,   0,   0,   0 } },
 {  4.899756422647e-06,  2.747872549080e-06, {   4,   0,  -1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 { -3.169470929320e-06,  4.476606313119e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,  -5,   0,   0,   0 } },
 { -2.279382353346e-06,  5.364529078166e-06, {   3,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.279382353346e-06, -5.364529078166e-06, {   3,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.440868867262e-06,  3.200189205820e-06, {   1,   1,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -2.760351690149e-06, -4.878841996528e-06, {   4,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.417847905384e-06,  3.221252493189e-06, {   1,  -1,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.099247645397e-06, -2.536746894500e-06, {   1,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.710151021917e-06,  5.924504637871e-06, {   1,  -1,   0,   0,   0,  20, -22,   0,   0,   0,   0,   0,   0 } },
 { -3.248584435484e-06,  4.383452557291e-06, {   0,   0,   2,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 } },
 { -6.013963754354e-06, -1.614140875777e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -6,   9,   0,   0,   0 } },
 { -6.737703994038e-06,  8.900185766949e-07, {   2,   0,   0,   0,   0,   0,  12, -23,   3,   0,   0,   0,   0 } },
 {  5.015986355344e-06, -2.610534067394e-06, {   0,   0,   1,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0 } },
 {  7.217114596612e-07,  6.901312663144e-06, {   4,   0,  -1,   0,   0,   5,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.417154449679e-06,  3.205588951318e-06, {   1,   0,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  8.264808563590e-07, -6.795301130267e-06, {   2,   0,   0,   0,   0,   0,   5,   0,  -3,  -5,   0,   0,   0 } },
 {  4.014280470251e-06,  3.607140783775e-06, {   2,   0,  -1,   0,   0,   0,  -6,  12,  -3,   0,   0,   0,   0 } },
 {  5.893296592567e-06, -1.727186669256e-06, {   1,   1,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -4.109452748817e-06,  3.511009656152e-06, {   4,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  4.866629220004e-06, -2.752352187675e-06, {   2,   0,  -1,   0,   0,  -2,   6,  -7,   0,   0,   0,   0,   0 } },
 { -2.663308939745e-06,  4.949607597470e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -7,   0,   0,   0 } },
 { -4.241275559656e-06,  3.371232895300e-06, {   1,  -1,   0,   0,   0, -12,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.617136929540e-06, -1.994900355651e-06, {   0,   0,   1,   0,   0,  13, -12,   0,   0,   0,   0,   0,   0 } },
 { -7.444101041180e-06, -1.678915256579e-07, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 {  6.138971019491e-06,  1.471118860430e-06, {   3,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  8.634934156171e-07,  6.746213332958e-06, {   2,   0,  -2,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  7.231505490198e-06, -3.777768261379e-07, {   0,   2,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.151562531716e-06, -1.457460551757e-06, {   2,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -7.382178800475e-06, -2.262330737794e-07, {   1,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.255127977121e-06,  6.351904711691e-06, {   1,   1,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -7.199279502805e-06,  4.072244039950e-07, {   0,   0,   3,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -4.129940623996e-06, -3.473522933115e-06, {   3,  -1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.639536018709e-06,  4.963764118953e-06, {   1,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.538201556163e-06, -5.064507002014e-06, {   2,   0,  -1,   0,   0, -19,  17,   0,   0,   0,   0,   0,   0 } },
 {  4.673235990671e-06,  2.925462028618e-06, {   2,   2,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.382409148430e-08,  7.541470437896e-06, {   1,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -7.302132519049e-06, -2.914632296550e-07, {   2,  -2,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.405946663322e-06, -6.186899116827e-06, {   1,  -1,  -1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.762392646358e-06, -3.825491984409e-06, {   2,   0,   0,   0,   0,   0,  -4,   7,   3,   0,   0,   0,   0 } },
 { -3.035630491695e-06,  4.551648207031e-06, {   0,   2,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -3.035630491695e-06, -4.551648207031e-06, {   0,   2,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  4.597467791047e-06,  2.989649956436e-06, {   1,   0,  -2,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -6.095123350478e-06,  1.491245974533e-06, {   1,  -1,  -2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.284044508230e-06,  2.994666362338e-07, {   4,   0,  -2,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  4.203766797888e-06,  3.377916842490e-06, {   2,  -2,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.813345529572e-07,  7.099699675228e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  2.387475937370e-06,  5.192749066324e-06, {   4,   0,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.620774693823e-06,  4.959177478001e-06, {   0,   0,   2,   0,   0,  -8,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.695895546844e-06,  4.883936169722e-06, {   6,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -7.561309035662e-08, -7.504144431394e-06, {   2,   0,   1,   0,   0,  23, -23,   0,   0,   0,   0,   0,   0 } },
 { -6.089870341544e-06,  1.489461341706e-06, {   0,   0,   2,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -3.080755242447e-06, -4.494821498535e-06, {   1,  -1,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 {  3.080755242447e-06,  4.494821498535e-06, {   1,  -1,  -1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -6.041166554408e-06, -1.533705480703e-06, {   1,  -1,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -4.889080990790e-06,  2.681083802481e-06, {   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  7.709701436713e-09, -7.557663919246e-06, {   3,  -2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  2.691264528655e-06, -4.873636969837e-06, {   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.264242280213e-06, -5.294165185014e-06, {   2,   0,  -3,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.653451952260e-06,  2.904843973222e-06, {   2,   2,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  2.297903556670e-06, -5.258084416434e-06, {   2,   0,  -1,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0 } },
 { -6.580920403350e-06,  9.728469077576e-07, {   4,   0,   0,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 {  4.070684757218e-06,  3.483048170808e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   1,   0,   0,   0 } },
 { -5.126579583800e-06,  2.426666455957e-06, {   0,   0,   1,   0,   0,   0,   3,   0,  -4,   5,   0,   0,   0 } },
 {  6.105793672062e-06,  1.439075548468e-06, {   1,   1,   1,   0,   0,   7,  -6,   0,   0,   0,   0,   0,   0 } },
 {  3.498051938733e-06, -4.042545868532e-06, {   2,   0,  -2,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 {  5.523865679593e-06,  2.016094506906e-06, {   2,   0,   0,   0,   0, -12,  15,   0,   0,   0,   0,   0,   0 } },
 { -2.623951591651e-06, -4.914762525522e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   5,  -3,   0,   0,   0 } },
 {  2.483188792601e-06, -5.053526314245e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -4,   0,   0,   0 } },
 {  2.327731301826e-06, -5.208926338301e-06, {   2,   0,   1,   0,   0,   0,   3,  -8,   1,   5,   0,   0,   0 } },
 { -2.490501598618e-06,  5.045581667816e-06, {   2,   0,   0,   0,   0,  23, -21,   0,   0,   0,   0,   0,   0 } },
 {  2.481426938645e-06, -5.053890087831e-06, {   1,   1,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -6.570885109339e-06, -9.638083609089e-07, {   4,   0,   0,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 {  4.888386313887e-06,  2.644941336949e-06, {   0,   0,   1,   0,   0,   0,  15, -17,   0,   0,   0,   0,   0 } },
 {  4.970515131853e-06,  2.562105235677e-06, {   2,   0,  -1,   0,   0,   0,  19, -21,   0,   0,   0,   0,   0 } },
 {  1.283278433020e-06, -6.248583479644e-06, {   1,  -1,  -1,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.200080814926e-06, -2.331645970747e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0 } },
 { -7.280361227558e-06, -2.508119756903e-07, {   0,   0,   2,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 {  4.629698548067e-06, -2.901244316347e-06, {   2,   2,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.199473053374e-06, -5.331467392329e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 { -2.497073237914e-06,  5.033449263079e-06, {   3,  -2,  -1,   0,   0,  20, -18,   0,   0,   0,   0,   0,   0 } },
 { -2.497073237914e-06, -5.033449263079e-06, {   1,   2,  -1,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 { -5.752534089622e-06, -1.773135906255e-06, {   0,   0,   1,   0,   0,  -5,   5,   4,   0,   0,   0,   0,   0 } },
 { -7.012519946807e-06, -5.088586971524e-07, {   2,   0,  -1,   0,   0,   5, -10,   0,   0,   0,   0,   0,   0 } },
 {  2.488622680527e-06,  5.031103761655e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0 } },
 {  4.748636260167e-06,  2.768610800480e-06, {   0,   0,   0,   0,   0,   0,  14, -16,   0,   0,   0,   0,   0 } },
 { -6.715433387417e-06,  8.010007501591e-07, {   2,   0,  -1,   0,   0,   0,  12, -23,   3,   0,   0,   0,   0 } },
 { -1.717749036584e-06, -5.797170272031e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0 } },
 { -4.567192121334e-06, -2.946129573692e-06, {   0,   0,   1,   0,   0,   0, -12,  16,   0,   0,   0,   0,   0 } },
 {  3.075517668720e-06,  4.436934193406e-06, {   2,   0,  -2,   0,   0,   0,   6, -11,  -3,   3,   0,   0,   0 } },
 {  5.025487134730e-06,  2.486832355859e-06, {   1,  -1,   2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -5.025487134730e-06, -2.486832355859e-06, {   1,  -1,  -2,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -6.069092288922e-06, -1.443130608459e-06, {   0,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.796298340156e-06,  1.714560616423e-06, {   2,   0,  -2,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -4.034683245396e-06, -3.476113742756e-06, {   2,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -6.085800520450e-06,  1.422262121634e-06, {   5,  -1,  -2,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 {  5.396849402458e-06,  2.110634127030e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 { -5.396849402458e-06, -2.110634127030e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   3,   0,   0,   0 } },
 {  6.662241400886e-06, -8.449360920308e-07, {   0,   0,   1,   0,   0,   5, -10,   4,  -3,   0,   0,   0,   0 } },
 {  6.662241400886e-06,  8.449360920308e-07, {   0,   0,   1,   0,   0,  -5,  10,  -4,   3,   0,   0,   0,   0 } },
 {  2.401627318624e-06, -5.105196222644e-06, {   2,   0,   1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0 } },
 { -7.235269200766e-06,  2.714600096561e-07, {   4,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.027333345656e-06, -5.478689787825e-06, {   5,   0,  -3,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 { -6.129365282561e-06,  1.376643017336e-06, {   1,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.913121978581e-06, -1.590829095293e-06, {   6,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -9.686150807660e-07,  6.535191730389e-06, {   0,   0,   2,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.329914156502e-06, -6.171215672223e-06, {   1,  -1,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.735761636631e-06, -2.764205815057e-06, {   0,   0,   1,   0,   0,   0,  -8,  17,   0,   0,   0,   0,   0 } },
 { -1.260978990818e-06, -6.238650380056e-06, {   1,  -1,  -1,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.684419237013e-06,  2.813347388381e-06, {   2,   2,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.279720966478e-06, -2.213407312006e-06, {   2,   0,   0,   0,   0,   0,  -6,  13,   0,   0,   0,   0,   0 } },
 {  2.872206123045e-06,  4.619901150057e-06, {   2,   0,  -2,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 { -3.878655348803e-06, -3.611850442918e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   4,   0,   0,   0 } },
 {  6.062594991803e-06,  1.427433310770e-06, {   3,  -1,   0,   0,   0,  10,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.367909045545e-06, -1.120868173994e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,   1,   0,   0,   0 } },
 {  7.169845838555e-06,  3.160666708384e-07, {   2,   0,  -2,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.472980122019e-06, -5.012623346359e-06, {   2,  -2,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  6.300415815317e-07,  6.855109426837e-06, {   0,   0,   1,   0,   0,   0,  -3,   8,  -1,  -5,   0,   0,   0 } },
 { -1.853012746501e-06, -5.628985306777e-06, {   0,   0,   1,   0,   0,   0,  -5,  11,   0,   0,   0,   0,   0 } },
 { -5.165188418234e-07, -6.958795801258e-06, {   6,   0,  -2,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.164111412565e-06,  3.309612006777e-06, {   1,  -1,   1,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 { -4.164111412565e-06, -3.309612006777e-06, {   1,  -1,  -1,   0,   0,   0,   3,  -4,   2,   0,   0,   0,   0 } },
 { -5.727779051154e-06, -1.744013314935e-06, {   0,   0,   1,   0,   0,   0,   9, -16,   6,   0,   0,   0,   0 } },
 { -2.405920233598e-07, -7.230793006808e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  9.700675835091e-07, -6.498508679224e-06, {   2,  -1,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.229475623503e-06, -6.238837420855e-06, {   0,   0,   2,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  4.752990835400e-06, -2.714110238428e-06, {   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.691800777709e-06, -4.775215750378e-06, {   2,   0,   0,   0,   0,   0,  -4,   6,  -1,   0,   0,   0,   0 } },
 {  2.676412526135e-06, -4.787941560028e-06, {   4,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 { -4.085260137679e-06, -3.378030539596e-06, {   2,   0,   1,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  6.270317371481e-06,  1.190370331521e-06, {   0,   2,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.987715828577e-06, -3.469226972230e-06, {   2,   0,  -4,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -4.250758660534e-06, -3.204757587417e-06, {   1,  -1,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0 } },
 { -6.288202397084e-08,  7.390803173316e-06, {   2,  -2,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -6.031921030072e-06,  1.421586900766e-06, {   1,   1,  -2,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.994190833001e-06, -1.456353913097e-06, {   0,   1,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.417721772770e-06, -6.031972313330e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   1,   0,   0,   0 } },
 { -5.899113044596e-06,  1.549384161728e-06, {   0,   0,   0,   0,   0,  27, -30,   0,   0,   0,   0,   0,   0 } },
 {  4.654197342286e-06, -2.793700933205e-06, {   2,   2,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -7.306279713236e-06, -1.415807173512e-07, {   0,   0,   3,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.339857307548e-06,  6.105679020804e-06, {   2,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.615346316308e-07, -6.783958203066e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0 } },
 {  1.155706153856e-07,  7.327506713332e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  6.943773457225e-06,  4.982000390437e-07, {   2,   0,   0,   0,   0,  25, -24,   0,   0,   0,   0,   0,   0 } },
 {  4.093060861154e-06, -3.348424619898e-06, {   1,  -1,   2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.490945628067e-06, -1.948814264557e-06, {   1,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -1.337633888946e-06,  6.101049058225e-06, {   1,  -1,   2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -7.358192001139e-06,  8.001028923122e-08, {   2,   0,   0,   0,   0,   0,  21, -24,   0,   0,   0,   0,   0 } },
 {  5.488076776545e-06,  1.949449153829e-06, {   1,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  3.534979651882e-06, -3.899524098643e-06, {   0,   0,   1,   0,  -5,  22, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.534641702601e-06,  3.898562115903e-06, {   0,   0,   1,   0,   5, -22,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.004590783127e-06,  1.425629779286e-06, {   6,   0,  -2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.004089037948e-06,  1.426024753477e-06, {   3,  -1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.565097815140e-08,  7.373718301564e-06, {   0,   0,   1,   0,   0, -20,  22,   0,   0,   0,   0,   0,   0 } },
 { -3.156585097511e-07, -7.111218166710e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   6,  -2,   0,   0,   0 } },
 { -4.963840475657e-06,  2.462179297980e-06, {   1,  -1,   1,   0,   0,   0,  17, -30,   0,   0,   0,   0,   0 } },
 {  4.963840475657e-06, -2.462179297980e-06, {   1,  -1,  -1,   0,   0,   0,  17, -30,   0,   0,   0,   0,   0 } },
 {  3.817127066014e-06, -3.608882852873e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -7,   0,   0,   0 } },
 { -5.479997969831e-06,  1.945609410579e-06, {   3,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  6.043701510174e-06,  1.381891782194e-06, {   1,  -1,  -1,   0,   0,  23, -22,   0,   0,   0,   0,   0,   0 } },
 { -5.478550051741e-06, -1.945378648438e-06, {   3,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.972022008408e-06, -5.448672598781e-06, {   4,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.234416154565e-06, -2.184756677718e-06, {   2,  -2,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  2.212121995799e-06, -5.206231746723e-06, {   2,  -2,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.212121995799e-06,  5.206231746723e-06, {   2,  -2,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.302209679935e-06, -3.115748629614e-06, {   1,  -1,   1,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.972261392661e-06, -1.445132758922e-06, {   1,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.101037495829e-06, -1.316040766496e-06, {   1,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  1.147417267183e-06, -6.269504016799e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -5.265422819910e-06, -2.147766179333e-06, {   2,   0,  -1,   0,   0,  -6,  14,  -8,   0,   0,   0,   0,   0 } },
 {  4.017685439959e-06,  3.394580842698e-06, {   2,   0,   0,   0,   0,   0,  -5,   8,  -2,   0,   0,   0,   0 } },
 {  2.443548165585e-06, -4.966535062895e-06, {   0,   0,   2,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.932356224898e-06,  3.470779678539e-06, {   2,   0,   1,   0,   0,   0, -10,  16,  -4,  -5,   0,   0,   0 } },
 { -5.472676171516e-06, -1.926873733495e-06, {   1,  -1,   1,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  2.789902506477e-06,  4.609600287096e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0 } },
 {  6.017760645284e-06, -1.381681420202e-06, {   3,   1,  -2,   0,   0, -19,  18,   0,   0,   0,   0,   0,   0 } },
 { -3.100604630012e-07, -7.089242773090e-06, {   2,   0,   2,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -6.036043742828e-06, -1.362034454752e-06, {   0,   0,   3,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -5.235200469726e-06,  2.161655123607e-06, {   2,   0,  -1,   0,   0,   6, -14,   8,   0,   0,   0,   0,   0 } },
 {  6.710578010123e-06, -6.843463068338e-07, {   2,   0,  -1,   0,   0,   0,  -2,   8,  -6,   0,   0,   0,   0 } },
 { -3.170836127928e-06,  4.222114803820e-06, {   2,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.197625170565e-06, -2.194583254443e-06, {   2,   0,  -1,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.493649486567e-06,  4.896996951221e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -2,  -5,   0,   0,   0 } },
 { -4.495408400985e-06,  2.895158772509e-06, {   2,   0,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -2.947826852798e-06, -4.442631314579e-06, {   2,   0,   1,   0,  -9,   1,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.947826852798e-06,  4.442631314579e-06, {   2,   0,  -1,   0,  -9,   1,  11,   0,   0,   0,   0,   0,   0 } },
 {  6.000713726323e-06, -1.388346505806e-06, {   3,  -1,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -6.210465192411e-06, -1.178133661104e-06, {   1,  -1,   1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 { -5.647368851388e-06,  1.721341774349e-06, {   1,  -1,  -1,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -7.177092631458e-06,  1.905700540060e-07, {   0,   0,   2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -4.572362109514e-06, -2.794479248975e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0 } },
 { -1.394907043508e-06,  5.969523336769e-06, {   1,   1,   0,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.871242387906e-06, -3.493136358341e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -6,   8,   0,   0,   0 } },
 { -1.874610201435e-06, -5.487071585198e-06, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   1,   0,   0,   0 } },
 { -6.633559532989e-06,  7.263855808461e-07, {   2,   0,   0,   0,   0,   0,   0,   0,   5, -10,   0,   0,   0 } },
 { -6.415429181697e-06,  9.444777173783e-07, {   2,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.647908480831e-06, -2.710109374249e-06, {   2,   0,   2,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -5.900197075756e-06, -1.457473511859e-06, {   2,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.009686319316e-06, -4.346652678995e-06, {   2,   0,  -2,   0,   0,   0,  -6,  11,   3,  -3,   0,   0,   0 } },
 {  1.297466436367e-06,  6.058670454963e-06, {   0,   0,   3,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -1.870785977795e-06,  5.483904528994e-06, {   2,   0,  -1,   0,   0,   0, -11,  18,   0,   0,   0,   0,   0 } },
 { -2.968855468140e-06, -4.385747838939e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0 } },
 {  3.434722592452e-06, -3.917369880225e-06, {   2,   0,  -1,   0,   0,   3, -11,  12,  -3,   0,   0,   0,   0 } },
 {  3.434722592452e-06,  3.917369880225e-06, {   2,   0,  -1,   0,   0,  -3,  11, -12,   3,   0,   0,   0,   0 } },
 {  5.186375633354e-06, -2.163460102909e-06, {   1,  -1,  -1,   0,   0,  25, -26,   0,   0,   0,   0,   0,   0 } },
 {  8.712883356962e-07,  6.478368120091e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 { -5.787733172119e-06,  1.558636105843e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   4, -10,   0,   0,   0 } },
 { -2.215976178567e-06,  5.130172894428e-06, {   1,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.267517916944e-06,  5.077065149990e-06, {   1,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -5.354582659015e-06,  1.989693644649e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  3.990747289189e-08, -7.303237161663e-06, {   4,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.614895662190e-06,  3.727206557762e-06, {   2,   0,   0,   0,   0,   0,   4,  -7,  -3,   0,   0,   0,   0 } },
 { -3.998596075381e-06, -3.340495655009e-06, {   1,  -1,   0,   0,   0,  10,  -7,   0,   0,   0,   0,   0,   0 } },
 {  2.959411161139e-06,  4.379094548959e-06, {   2,   0,  -1,   0,   0,   0,  -1,   2,  -2,   3,   0,   0,   0 } },
 { -7.182470137775e-06,  1.543830644002e-07, {   2,   0,   1,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.936796300593e-06,  1.398975251670e-06, {   6,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.128620757624e-06,  2.058424258019e-07, {   6,   0,  -3,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.438050143784e-06, -3.894909970499e-06, {   2,   0,   2,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  3.610517189676e-06, -3.719896966399e-06, {   0,   0,   1,   0,   0,   0,  -1,   4,  -3,   0,   0,   0,   0 } },
 {  2.595041720664e-06,  4.733848321455e-06, {   2,  -2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 {  5.778547617223e-06, -1.550088856379e-06, {   3,  -1,   2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -4.278238900463e-06, -3.050185861251e-06, {   0,   0,   1,   0,   0,  12,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.214265645400e-07,  6.906661366420e-06, {   0,   0,   2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 {  1.914806613831e-06, -5.411925803898e-06, {   2,   0,  -1,   0,   0,   0,  11, -18,   3,   0,   0,   0,   0 } },
 {  3.770269837140e-06,  3.554396612426e-06, {   2,   0,   0,   0,   0,   0,  -6,  12,  -3,   0,   0,   0,   0 } },
 { -1.288328823275e-06,  6.035746823066e-06, {   1,  -1,   1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -8.687885086000e-08,  7.237034874339e-06, {   2,   0,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.936400030160e-06,  1.385479869466e-06, {   1,  -1,   1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  2.982160323013e-06,  4.339534819229e-06, {   2,   0,   0,   0,   0,   0,  -1,   2,  -2,   3,   0,   0,   0 } },
 {  2.945700449919e-06, -4.374949584553e-06, {   2,   0,  -1,   0,   0,   0,   1,  -2,   2,  -3,   0,   0,   0 } },
 {  4.995988578114e-06,  2.323803027424e-06, {   0,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.324071130302e-06,  5.993550125459e-06, {   1,   0,  -4,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.044770942659e-06,  6.272338571692e-06, {   3,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.054200036039e-06,  6.261931916425e-06, {   1,  -1,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.499603887594e-06,  8.144717099986e-07, {   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   2,   0,   0 } },
 { -2.969657756259e-09,  7.308547050107e-06, {   4,   0,  -3,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.926330778519e-06,  3.841618016302e-07, {   4,   0,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -4.105476073351e-06,  3.204916686122e-06, {   1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 } },
 {  5.940425582809e-06, -1.361123425325e-06, {   1,  -1,  -2,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.219174650013e-06, -5.081799541600e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   7,   0,   0,   0 } },
 {  1.191775857293e-07,  7.180293370317e-06, {   2,  -2,   2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -3.159599987841e-06, -4.138927154452e-06, {   0,   0,   1,   0,   0,   1,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.084200347342e-06,  2.213209079544e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,  -3,   0,   0,   0 } },
 { -1.436847797241e-06, -5.856168978046e-06, {   2,   0,  -1,   0,   0,   0,   3,  -2,  -3,   0,   0,   0,   0 } },
 { -5.918427439426e-06,  1.374511913863e-06, {   3,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.866806820345e-06, -2.425253089356e-06, {   4,   0,  -2,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -6.108116689527e-06, -1.183081566985e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   3,  -1,   0,   0,   0 } },
 {  5.750244187185e-06, -1.539581477319e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0 } },
 { -1.296643830538e-08, -7.276108577332e-06, {   2,  -2,  -1,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  3.482421617193e-08, -7.253541122993e-06, {   2,  -2,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.357337636377e-06,  1.930843878131e-06, {   1,  -1,   1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.699168359709e-06,  1.587676639075e-06, {   2,   0,  -1,   0,   0,   3, -11,  16,  -4,  -5,   0,   0,   0 } },
 {  2.083808767186e-06, -5.201841064493e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,  -2,   0,   0,   0 } },
 { -6.895062974704e-06,  3.900723460838e-07, {   4,   0,  -2,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  7.058540708694e-06,  2.260022701967e-07, {   3,   0,  -2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  5.698559007288e-06, -1.585515430777e-06, {   2,   0,  -1,   0,   0,   3,   5, -16,   4,   5,   0,   0,   0 } },
 {  1.649073169975e-06, -5.634408039176e-06, {   4,   0,  -1,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 { -2.767857766303e-06,  4.512365243203e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   2,  -5,   0,   0,   0 } },
 { -5.513862466669e-06,  1.765401089601e-06, {   0,   0,   1,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.300118262741e-06,  2.975749805182e-06, {   2,   0,   0,   0,   0,   0,   2,  -8,   1,   5,   0,   0,   0 } },
 {  1.791929055600e-07,  7.095538779978e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,  -1,   0,   0 } },
 {  3.034292566199e-06,  4.240178937411e-06, {   1,   1,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -5.270089898299e-06,  2.000714589145e-06, {   2,   0,  -1,   0,   0,   0,   1,  -4,   1,   0,   0,   0,   0 } },
 { -1.130069164841e-06, -6.140309529571e-06, {   2,   0,   3,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  5.337015732385e-06,  1.932488499679e-06, {   2,   0,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 { -1.743488711806e-06,  5.525249166651e-06, {   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.947763909260e-07,  6.973615626540e-06, {   2,   0,   1,   0,   0,  16, -18,   0,   0,   0,   0,   0,   0 } },
 { -3.599251841983e-06, -3.666121654378e-06, {   2,   0,  -1,   0,   0,   0,  -4,   7,   3,   0,   0,   0,   0 } },
 {  5.240687992373e-06, -2.023558434231e-06, {   1,   1,  -1,   0,   0,   9, -12,   0,   0,   0,   0,   0,   0 } },
 { -4.776165333876e-06,  2.488015110259e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 { -1.966612203559e-06, -5.295595754244e-06, {   3,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.167802441759e-08, -7.178319835491e-06, {   1,   0,  -3,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  3.758269145996e-06, -3.499836350596e-06, {   3,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -2.783607437616e-06, -4.472599334429e-06, {   1,  -1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.205318775316e-06, -3.050627791468e-06, {   4,   0,  -1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.944886539562e-06, -4.307223018695e-06, {   2,   0,   0,   0,   0,   0,   1,  -2,   2,  -3,   0,   0,   0 } },
 {  6.863014540361e-06,  3.888456427285e-07, {   2,   0,   2,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.858454436871e-06, -3.388350137851e-06, {   0,   0,   1,   0,   0,   0, -10,  16,  -3,   0,   0,   0,   0 } },
 {  5.875134011283e-06, -1.371476091198e-06, {   1,   1,   1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  4.754599233800e-06, -2.490651369034e-06, {   2,   0,  -1,   0,   0,   9, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.187991985646e-06,  6.051013141654e-06, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 { -7.164255648560e-06,  7.389464142587e-08, {   1,  -2,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.132183276341e-06,  2.103018977626e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -3.596890860255e-06, -3.637630817512e-06, {   3,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -8.860919994707e-07, -6.348231527868e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 {  6.996200459820e-06,  2.341272573774e-07, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   4,   0,   0,   0 } },
 {  2.545171427964e-06,  4.684173353667e-06, {   2,   0,   2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  1.479669601845e-06, -5.749660576993e-06, {   1,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  8.085655863803e-08,  7.147868404520e-06, {   0,   0,   2,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -2.848154070877e-06, -4.379434906711e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   6,  -3,   0,   0,   0 } },
 { -8.800984308800e-08, -7.137574516412e-06, {   2,   0,   0,   0,   0,  32, -32,   0,   0,   0,   0,   0,   0 } },
 { -5.817577090849e-06, -1.407680195372e-06, {   2,   0,  -4,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.094742023464e-06, -1.129174133652e-06, {   4,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -4.177511090047e-06,  3.045393793649e-06, {   6,   0,  -3,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  1.328086528104e-06, -5.894421851504e-06, {   1,  -1,  -2,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -5.765609699728e-06,  1.456588532367e-06, {   5,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.292614589347e-06,  2.928311129206e-06, {   0,   0,   0,   0,   0,   0,   6,   0,  -5,   0,   0,   0,   0 } },
 { -5.550081274941e-06,  1.668857140279e-06, {   2,   0,  -1,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.789214317209e-06,  4.428297001385e-06, {   4,   0,  -3,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  5.733689697711e-06,  1.482347336876e-06, {   1,   0,   1,   0,   0, -31,  38,   0,   0,   0,   0,   0,   0 } },
 { -5.733689697711e-06, -1.482347336876e-06, {   1,   0,  -1,   0,   0, -31,  38,   0,   0,   0,   0,   0,   0 } },
 {  4.020247961480e-06, -3.193882504918e-06, {   1,  -1,  -1,   0,   0,   0,   2,   0,   0,   1,   0,   0,   0 } },
 {  4.236246321644e-06, -2.977431845857e-06, {   4,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -2.639548002228e-06, -4.573947871351e-06, {   2,   0,  -2,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0 } },
 {  2.727010375656e-06, -4.486228239748e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0 } },
 {  1.212293110844e-06, -6.000068631288e-06, {   1,   1,  -1,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.257313024346e-08, -7.199450658253e-06, {   2,   0,  -3,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 {  2.376786269480e-06,  4.834411540925e-06, {   4,   0,  -2,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.785696256934e-06, -4.250866581650e-07, {   4,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -3.824359003299e-06,  3.383510406666e-06, {   3,  -1,   0,   0,   0,   0,   1,   0,  -1,   4,   0,   0,   0 } },
 { -3.824359003299e-06, -3.383510406666e-06, {   1,   1,   0,   0,   0,   0,  -1,   0,   1,  -4,   0,   0,   0 } },
 { -4.367660305377e-06, -2.839215898365e-06, {   2,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.365057218926e-06, -5.840984107835e-06, {   1,   0,   0,   0,   0,   6, -10,   0,   3,   0,   0,  -1,   0 } },
 { -2.126895880252e-06, -5.078300608006e-06, {   0,   0,   1,   0,   0,   0, -12,  17,   0,   0,   0,   0,   0 } },
 {  2.575155091015e-06,  4.627310800079e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0 } },
 {  5.844277382542e-06, -1.358000887357e-06, {   3,  -1,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.363039525595e-06,  5.837712092601e-06, {   1,   0,   0,   0,   0,  -6,  10,   0,  -3,   0,   0,   1,   0 } },
 { -2.919880914671e-06, -4.278313659027e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   7,   0,   0,   0,   0 } },
 { -5.189755532822e-06,  2.007935148939e-06, {   4,   0,  -1,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  3.794215778668e-06, -3.403435052497e-06, {   2,   0,   0,   0,   0,   0,   6, -12,   3,   0,   0,   0,   0 } },
 {  5.631324087085e-06,  1.565915704129e-06, {   1,   0,   1,   0,   0, -23,  24,   0,   0,   0,   0,   0,   0 } },
 { -5.102567261072e-06,  2.087039556432e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -4.740419670853e-06,  2.448545670461e-06, {   0,   2,  -1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -5.821039697347e-06, -1.367267278537e-06, {   1,   1,  -1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.059323943547e-06, -5.128919021862e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,  -3,   0,   0,   0 } },
 {  2.240047847258e-06,  4.946553776029e-06, {   2,   0,   0,   0,   5, -22,  15,   0,   0,   0,   0,   0,   0 } },
 {  7.143143789491e-07,  6.472083234387e-06, {   2,   0,  -1,   0,   0,   3,  -3,   0,   1,   0,   0,   0,   0 } },
 {  1.260172464424e-06, -5.925395225829e-06, {   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0 } },
 {  3.503190808640e-09,  7.181808605537e-06, {   1,   0,   2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.852404874025e-06,  5.332296785906e-06, {   0,   0,   1,   0,   0,   0,   5,   0,  -7,   0,   0,   0,   0 } },
 { -4.634779629172e-06, -2.547226088195e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0 } },
 { -1.097566905578e-06, -6.084015250125e-06, {   4,  -2,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -6.353286238635e-07, -6.545399126812e-06, {   4,   0,  -2,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.162080174678e-06,  6.018422889741e-06, {   3,   0,  -2,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.205363660620e-07, -7.059917679675e-06, {   2,   0,  -4,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -4.735977321321e-06, -2.444392035508e-06, {   0,   2,  -1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.046649301396e-06, -6.132383561884e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0 } },
 {  6.383335229932e-06, -7.888351391523e-07, {   2,   0,  -3,   0,   0,   0,   8, -12,   2,   0,   0,   0,   0 } },
 {  2.966091907223e-08,  7.141611600932e-06, {   0,   0,   1,   0,   0, -25,  25,   0,   0,   0,   0,   0,   0 } },
 { -3.219192931612e-08,  7.136706917295e-06, {   2,   0,   0,   0,   0, -25,  29,   0,   0,   0,   0,   0,   0 } },
 { -3.219192931612e-08,  7.136706917295e-06, {   2,   0,  -2,   0,   0, -25,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.233660642150e-06, -4.934813745089e-06, {   2,   0,   0,   0,  -5,  22, -15,   0,   0,   0,   0,   0,   0 } },
 {  3.327434172864e-06, -3.840177990359e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -5.221738648639e-06,  1.945007011861e-06, {   0,   0,   1,   0,   0,   0, -11,  13,   0,   0,   0,   0,   0 } },
 { -3.850242576941e-06, -3.315437034956e-06, {   2,   0,   1,   0,   0, -26,  27,   0,   0,   0,   0,   0,   0 } },
 {  2.638356352535e-06, -4.526927494674e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,  -1,   2,   0 } },
 { -9.193521275053e-07,  6.244359665676e-06, {   2,   0,   1,   0,   0, -24,  26,   0,   0,   0,   0,   0,   0 } },
 {  3.235181060584e-06, -3.927815128319e-06, {   2,   0,  -1,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0 } },
 { -4.189291784335e-06,  2.973098340115e-06, {   0,   0,   2,   0,   0,   2,  -7,   7,   0,   0,   0,   0,   0 } },
 { -2.161107689917e-06, -4.998155046269e-06, {   2,   0,  -1,   0,   0,   0,  19, -26,   0,   0,   0,   0,   0 } },
 {  4.810330660028e-06,  2.345093214666e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 { -2.739150530402e-07,  6.879861506127e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -4,   2,   0,   0,   0 } },
 {  1.349364823918e-06, -5.804052536842e-06, {   2,   0,   0,   0,  -1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.128210814852e-06, -5.022926850557e-06, {   2,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.891511892819e-06, -1.251848197802e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   1,  -2,   0,   0,   0 } },
 { -2.181624268535e-06,  4.960451421891e-06, {   2,  -2,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.037409885377e-06,  6.100937604527e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  5.797101911368e-06, -1.340240953707e-06, {   1,   1,   1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -6.081329828915e-06,  1.055858425274e-06, {   2,   0,  -1,   0,  -4,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  4.301167021441e-09,  7.132667125270e-06, {   0,   2,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.554852512602e-08,  7.110353403321e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 { -4.179745033437e-06, -2.956116081580e-06, {   0,   0,   2,   0,   0,  -2,   7,  -7,   0,   0,   0,   0,   0 } },
 { -5.448234422676e-06, -1.685560493580e-06, {   2,   0,  -1,   0,   0,  -3,   3,   0,  -1,   0,   0,   0,   0 } },
 { -2.007502978678e-08,  7.112970385883e-06, {   1,   0,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0 } },
 { -7.705346140547e-07,  6.360835466764e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 } },
 { -2.536683259461e-06,  4.594383861295e-06, {   2,   0,   0,   0,   0,   0,  16, -15,   0,   0,   0,   0,   0 } },
 {  5.670795046980e-06,  1.458688924304e-06, {   1,   1,   2,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -6.808585525923e-06, -3.194918973101e-07, {   2,   0,   1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0 } },
 {  9.081871045595e-09, -7.114222952029e-06, {   4,   0,   0,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -2.528384751710e-07, -6.870382765502e-06, {   2,   0,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  2.756556117699e-06, -4.361109997634e-06, {   0,   0,   2,   0,   0,  -1,   8, -12,   0,   0,   0,   0,   0 } },
 { -5.274942979405e-06,  1.837314910973e-06, {   1,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.258101817104e-06,  1.851938365688e-06, {   4,   0,  -2,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  2.751671518444e-06,  4.358195524286e-06, {   0,   0,   2,   0,   0,   1,  -8,  12,   0,   0,   0,   0,   0 } },
 { -2.708218727221e-06,  4.400920790257e-06, {   2,   0,  -2,   0,   0,   0,   7,  -6,   0,   0,   0,   0,   0 } },
 {  2.526342389010e-08, -7.083718100775e-06, {   4,   0,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.761713610196e-06, -1.347122314447e-06, {   1,   1,   1,   0,   0, -11,  10,   0,   0,   0,   0,   0,   0 } },
 {  2.377252500409e-08, -7.084832830988e-06, {   4,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.583213810095e-06, -5.523723269606e-06, {   2,   0,   0,   0,   0,   0,  13, -13,   0,   0,   0,   0,   0 } },
 { -5.087421289594e-06,  2.018423861358e-06, {   4,   0,  -2,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.805033311493e-06, -1.296565320029e-06, {   3,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -2.955219463706e-06,  4.144890136830e-06, {   1,   1,  -1,   0,   0,   0,  -2,   0,  -1,   1,   0,   0,   0 } },
 { -2.031924425679e-06,  5.066945986821e-06, {   3,  -1,  -1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 {  2.825039236447e-06, -4.271846097448e-06, {   2,   0,  -1,   0,   0,   0,  21, -26,   0,   0,   0,   0,   0 } },
 {  2.098673631626e-07,  6.885994377654e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 {  4.172981695673e-06,  2.922710103814e-06, {   1,   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.620701331740e-06, -2.472523172202e-06, {   2,   0,  -1,   0,   0,   0,  21, -25,   0,   0,   0,   0,   0 } },
 { -7.092367378588e-06,  7.709230662581e-10, {   0,   2,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -6.480286566501e-06, -6.105154953659e-07, {   2,   0,   0,   0,   0,   0,  -3,   0,  -2,   7,   0,   0,   0 } },
 { -1.469282648503e-06,  5.620788915755e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,  -5,   0,   0,   0 } },
 { -6.846310510481e-06, -2.437325578328e-07, {   2,   0,  -1,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 {  6.709231630766e-07, -6.418157064718e-06, {   2,   0,   0,   0,   0,   0,  -9,  16,  -6,   0,   0,   0,   0 } },
 { -4.004793276448e-06, -3.084212897738e-06, {   0,   0,   1,   0,   0,  -8,   7,   0,   0,   0,   0,   0,   0 } },
 { -4.947706136018e-06,  2.139669873521e-06, {   4,   0,  -2,   0,   0,  -6,  10,   0,  -3,   0,   0,   0,   0 } },
 {  2.683741390347e-06,  4.403215436106e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -2,   5,   0,   0,   0 } },
 { -2.785066296692e-06,  4.298826070693e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   5,   0,   0,   0 } },
 { -6.979829908897e-06,  1.040002497837e-07, {   1,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.571992256666e-06, -1.504764534539e-06, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  8.993703723827e-08, -6.984597755991e-06, {   4,  -2,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  6.543560954798e-06, -5.281223398548e-07, {   4,   0,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -5.975597706635e-08, -7.010351949125e-06, {   1,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.800979848440e-06,  2.638621028597e-07, {   4,   0,  -2,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.859536465658e-06,  4.203917304461e-06, {   4,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  1.238360232891e-06,  5.823501223738e-06, {   4,   0,  -2,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -1.081485394918e-07, -6.952009175766e-06, {   2,   0,   1,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -9.881689485458e-07,  6.070739894946e-06, {   1,  -1,  -1,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  6.358250232316e-06,  6.955080660875e-07, {   1,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 { -2.059504178227e-07,  6.845600770067e-06, {   2,   0,  -2,   0,   0,  10,  -5,   0,   0,   0,   0,   0,   0 } },
 {  4.978628145437e-06,  2.070820810389e-06, {   2,   0,   1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  4.425247642582e-06, -2.623401140600e-06, {   2,  -2,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.732603737791e-07, -6.674546281194e-06, {   2,   0,   0,   0,   0,   0,   4,   0,  -6,   2,   0,   0,   0 } },
 {  4.044048010551e-08,  7.006751608848e-06, {   4,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 {  2.336677431759e-06,  4.709923757468e-06, {   2,   0,  -1,   0,   0,  -2,   4,  -3,   0,   0,   0,   0,   0 } },
 {  6.162084255462e-06, -8.823503863046e-07, {   4,   0,  -2,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -2.789458186069e-06, -4.249592177404e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -3,   0,   0,   0 } },
 { -3.922211927675e-06, -3.114697414845e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   0,  -1,   0,   0,   0 } },
 {  5.762146344498e-06,  1.274248264663e-06, {   1,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.605868060916e-06,  4.292546704887e-07, {   2,   0,   1,   0,   0,  15, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.543871128599e-06, -4.491041960909e-06, {   4,   0,  -2,   0,  -1,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.089652235669e-06,  3.944717222681e-06, {   2,   2,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -5.687926893514e-06,  1.343913158086e-06, {   2,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.669679243186e-06, -2.361947367722e-06, {   2,   0,  -2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.910099093598e-06, -2.121177089962e-06, {   4,   0,  -2,   0,   0,   6, -10,   0,   3,   0,   0,   0,   0 } },
 {  2.231288367030e-06, -4.798093468621e-06, {   0,   0,   3,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -6.268147300819e-06,  7.594625975853e-07, {   2,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  2.043991989834e-08, -7.002196153086e-06, {   4,   0,  -2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.091596470449e-06,  2.928965155961e-06, {   2,   0,   0,   0,   0,   2,  -2,  -5,   8,  -6,   0,   0,   0 } },
 {  5.228758604648e-06, -1.788333362510e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   7,  -5,   0,   0,   0 } },
 {  4.710077630001e-06,  2.305548942138e-06, {   1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.932292855910e-06, -4.083149284714e-06, {   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0 } },
 { -6.888819557099e-07, -6.325840489223e-06, {   4,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  2.312956633221e-06,  4.701611176695e-06, {   0,   0,   1,   0,   0,  10, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.720804720582e-06,  4.293097431365e-06, {   4,   0,  -2,   0,   0,   0, -11,  19,   0,  -3,   0,   0,   0 } },
 {  2.598178270428e-06, -4.412759912730e-06, {   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.843479315484e-06, -4.165601104593e-06, {   0,   0,   1,   0,   0,   0,  14, -19,   0,   0,   0,   0,   0 } },
 { -9.837307867537e-07,  6.024920555622e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0 } },
 { -5.183473001340e-06,  1.824849125507e-06, {   1,   1,  -2,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 {  4.011544643647e-06, -2.995393552376e-06, {   2,   0,  -4,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -3.523032452690e-08,  6.970300886667e-06, {   2,  -2,   2,   0,   0, -18,  18,   0,   0,   0,   0,   0,   0 } },
 {  5.650399478840e-06, -1.354488806635e-06, {   1,   1,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.568773710132e-06,  1.436027806107e-06, {   2,   0,  -2,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.118000538158e-06, -5.886726183050e-06, {   0,   2,  -1,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 { -6.550658198785e-07,  6.348255773412e-06, {   4,  -2,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -1.252770932274e-06,  5.747656388234e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,  -3,   2,   0,   0,   0 } },
 {  7.399743833009e-07,  6.259686075019e-06, {   2,  -2,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.250721876982e-06, -5.748745421018e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   3,  -2,   0,   0,   0 } },
 {  6.590730377621e-06,  4.072117357151e-07, {   4,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.074052336986e-06, -4.917700060742e-06, {   1,   1,   0,   0,   0, -20,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.131685470631e-06, -4.859891818814e-06, {   2,  -2,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.123400696420e-06,  4.867005769247e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -1,   0,   0,   0 } },
 {  6.429181153887e-06,  5.604368004639e-07, {   4,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.229227557622e-06,  4.759380890515e-06, {   2,   0,  -2,   0,   0,  21, -19,   0,   0,   0,   0,   0,   0 } },
 { -5.223839497068e-10, -6.987436144380e-06, {   1,   0,  -1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.686858840081e-07,  6.714810368987e-06, {   2,  -2,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  9.829334930701e-07,  6.000402297329e-06, {   3,   0,  -1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -3.274005665982e-07, -6.653541942899e-06, {   2,   0,  -1,   0,   0,   6,  -7,   0,   1,   0,   0,   0,   0 } },
 {  1.213092607601e-06, -5.767622528061e-06, {   1,  -1,  -2,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.944865098603e-09, -6.971428067960e-06, {   3,  -2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.649471501811e-06,  1.328236826079e-06, {   4,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.839583598733e-08, -6.908035985174e-06, {   4,   0,   0,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.158851626216e-06,  5.814837945681e-06, {   1,   1,   0,   0,   0,  13, -15,   0,   0,   0,   0,   0,   0 } },
 { -4.996677016556e-06,  1.974099821601e-06, {   3,   0,  -1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.046601361390e-06, -4.922894578126e-06, {   2,   0,  -1,   0,   0,   0,   6, -11,  -2,   0,   0,   0,   0 } },
 {  4.255999752268e-06, -2.711870423314e-06, {   2,   0,  -1,   0,   0,  10, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.017186262396e-07,  6.665869727400e-06, {   2,   0,  -1,   0,   0,   0,   3,  -2,  -1,   0,   0,   0,   0 } },
 {  5.631310324446e-06,  1.334076104973e-06, {   4,  -2,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.209834305288e-06, -2.755433314373e-06, {   2,   0,   0,   0,   0,  -2,  -2,   9,  -8,   6,   0,   0,   0 } },
 { -6.926724807756e-06, -3.626056288324e-08, {   4,  -2,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.446335231432e-06, -1.516390768750e-06, {   0,   2,  -1,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 {  6.648474201660e-06,  3.112869952942e-07, {   3,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.405453001150e-06, -5.553981133556e-06, {   0,   0,   1,   0,   0,   6, -10,   0,   2,   0,   0,   0,   0 } },
 { -5.759294334611e-06, -1.199589408280e-06, {   2,   0,   2,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -4.564729756750e-07, -6.501242794677e-06, {   6,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -4.011361026886e-07, -6.556424861579e-06, {   2,   0,  -1,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 {  4.452822679993e-06,  2.504084280852e-06, {   1,   0,   1,   0,   0, -20,  18,   4,   0,   0,   0,   0,   0 } },
 { -4.452822679993e-06, -2.504084280852e-06, {   1,   0,  -1,   0,   0, -20,  18,   4,   0,   0,   0,   0,   0 } },
 { -6.704872908069e-06, -2.516478074615e-07, {   4,  -2,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 {  4.815033114162e-06,  2.140596320719e-06, {   3,  -1,  -2,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 { -1.188720193114e-06, -5.765774137729e-06, {   1,  -1,   2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -4.702220468042e-07, -6.481770809748e-06, {   4,  -2,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.334755266848e-06, -2.616161573009e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -5,   3,   0,   0,   0 } },
 {  2.265064844735e-06,  4.680310559636e-06, {   2,   0,  -3,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -7.451088577770e-08,  6.868685436362e-06, {   0,   0,   4,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.367380591174e-06,  1.575287661530e-06, {   1,  -1,   0,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.471955950606e-06,  1.469468164577e-06, {   2,   0,  -2,   0,   0,  10, -13,   0,   0,   0,   0,   0,   0 } },
 {  5.222651678190e-06, -1.713442424268e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0 } },
 {  4.265877403678e-06,  2.670215162881e-06, {   4,   0,  -2,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 { -2.273781918343e-06, -4.662010444123e-06, {   1,   0,   0,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.155787952296e-06, -1.779948420358e-06, {   4,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -4.414073386388e-06,  2.519102864043e-06, {   0,   0,   1,   0,   0,   3,  -9,   8,  -3,   0,   0,   0,   0 } },
 { -3.852193097277e-07,  6.547552117807e-06, {   3,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -2.549490581923e-06, -4.382504311573e-06, {   3,   0,  -2,   0,   0,  29, -33,   0,   0,   0,   0,   0,   0 } },
 { -2.549490581923e-06,  4.382504311573e-06, {   1,   0,   2,   0,   0, -29,  33,   0,   0,   0,   0,   0,   0 } },
 { -4.169112560715e-07, -6.512523701292e-06, {   2,  -2,   0,   0,   0,  25, -27,   0,   0,   0,   0,   0,   0 } },
 { -5.097723286735e-06,  1.828695406979e-06, {   2,   0,   1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.666172211620e-06,  5.257484791354e-06, {   5,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -8.589689240046e-10,  6.922429815295e-06, {   2,  -2,   0,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -1.966224334423e-06, -4.955784540217e-06, {   0,   0,   3,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.064594550698e-06,  4.857357475243e-06, {   1,  -1,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -1.620486397445e-06, -5.296997446973e-06, {   1,  -1,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  4.640629724704e-06,  2.274902640302e-06, {   5,   0,  -3,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  6.696939524641e-06,  2.154011656199e-07, {   2,   0,   0,   0,   0,  31, -32,   0,   0,   0,   0,   0,   0 } },
 {  1.525274150797e-06,  5.381419949448e-06, {   0,   0,   0,   0,   0,  24, -26,   0,   3,   0,   0,   0,   0 } },
 {  2.440182613675e-06,  4.465217614694e-06, {   2,   0,  -1,   0,  -7,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.348751045033e-06, -5.556093939322e-06, {   1,   1,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.838701350434e-06, -2.062272009676e-06, {   0,   0,   1,   0,   0,   0,  11, -21,   0,   0,   0,   0,   0 } },
 {  2.091688585640e-06, -4.809059471214e-06, {   2,   0,   0,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.591140518498e-06, -5.308868520766e-06, {   2,   0,   3,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.888497254214e-08, -6.811059737197e-06, {   2,   0,   1,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.120288014543e-06, -5.774974370702e-06, {   2,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.489741907697e-06, -1.403281726079e-06, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 {  5.908907170732e-06,  9.840247934040e-07, {   2,   0,   1,   0,   0,  -9,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.071739175038e-06,  4.820241553767e-06, {   1,  -1,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.400964374035e-06,  2.490944438095e-06, {   2,   0,  -1,   0,   0,   0,   1,  -6,   0,   0,   0,   0,   0 } },
 {  5.553529038023e-06,  1.332578709731e-06, {   1,  -1,  -2,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -2.125366705027e-06, -4.756927042047e-06, {   1,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  5.342870969754e-06, -1.539352733223e-06, {   1,  -1,   1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.883413610850e-09,  6.874292182162e-06, {   6,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.398086980182e-06,  5.478469599645e-06, {   0,   0,   1,   0,   0,  -6,  10,   0,  -2,   0,   0,   0,   0 } },
 {  2.479346932051e-06,  4.394911800989e-06, {   3,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 {  4.655024418161e-06, -2.214466582247e-06, {   2,   0,  -3,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 {  5.373129069151e-06,  1.496090362488e-06, {   0,   2,  -1,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  4.676581230375e-06, -2.190199838780e-06, {   0,   0,   2,   0,   0,   0,  -7,  13,   1,   0,   0,   0,   0 } },
 { -4.334286711207e-06, -2.532255175918e-06, {   2,   0,   1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -2.169663798720e-06, -4.696832949163e-06, {   2,   0,   1,   0,   0,   0,   4,  -9,   0,   0,   0,   0,   0 } },
 {  5.547276902656e-06,  1.318547215971e-06, {   1,   1,  -2,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.294465238573e-06,  2.570644950689e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -4,   4,   0,   0,   0 } },
 { -1.258409576466e-06,  5.605400796733e-06, {   0,   0,   1,   0,   0, -18,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.786370445928e-06, -5.077403141961e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -5.947301448665e-08, -6.803870497532e-06, {   2,  -2,  -2,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  1.993928552386e-06, -4.869075201161e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0 } },
 { -1.286339985412e-06,  5.574305682090e-06, {   2,  -2,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  2.513753707435e-06, -4.346774707076e-06, {   2,   0,  -1,   0,   0,   0,  15, -14,   0,   0,   0,   0,   0 } },
 {  4.675242036120e-06,  2.185265363787e-06, {   0,   0,   2,   0,   0,   0,   7, -13,  -1,   0,   0,   0,   0 } },
 { -5.204032947122e-06,  1.656003223635e-06, {   2,  -2,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  9.803156624346e-07,  5.879122231089e-06, {   1,   0,   2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -5.482819701641e-06, -1.376196343903e-06, {   0,   1,   0,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.015765335283e-09, -6.855066625656e-06, {   1,   0,  -3,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -1.379478023399e-06, -5.477777208784e-06, {   1,  -1,   0,   0,   0, -13,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.095386387817e-06,  1.761373140714e-06, {   1,  -1,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -2.270295256098e-06, -4.585097119796e-06, {   2,   0,   0,   0,   0, -22,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.611389917965e-08, -6.798140171477e-06, {   2,  -2,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -2.286092294954e-06,  4.566345568508e-06, {   1,  -1,   0,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 { -4.515029113286e-06, -2.336265757586e-06, {   2,   0,   2,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -2.021142022094e-06,  4.827153032311e-06, {   2,   0,  -1,   0,   0,   0,  -6,  11,   2,   0,   0,   0,   0 } },
 { -5.455889869703e-06,  1.391401777143e-06, {   1,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0 } },
 {  2.071821852398e-06,  4.774441866762e-06, {   2,   0,   0,   0,   0,   0,  19, -26,   0,   0,   0,   0,   0 } },
 { -5.528594081131e-06,  1.317346559918e-06, {   2,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.041297190706e-06,  4.803332516647e-06, {   2,   0,   0,   0,   0,   0,  13, -20,   0,   0,   0,   0,   0 } },
 { -5.297382550299e-07,  6.314856023721e-06, {   1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.773525236020e-06,  1.070146320956e-06, {   1,   1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.612665638660e-06, -5.230782401792e-06, {   2,   0,   0,   0,   0,  15, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.038349104366e-06,  5.804390512446e-06, {   0,   0,   1,   0,   0,   0,   6,   0,  -7,   0,   0,   0,   0 } },
 { -6.414654826757e-06, -4.269453474418e-07, {   2,   0,  -3,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.330449145673e-06,  4.510414259210e-06, {   0,   0,   2,   0,   0,  -9,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.261785792725e-06, -5.578016152760e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  5.279707790083e-06,  1.558670291046e-06, {   4,   0,  -2,   0,   0, -23,  26,   0,   0,   0,   0,   0,   0 } },
 { -6.632977528850e-06,  2.052633394880e-07, {   2,   0,   2,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.391436844783e-09,  6.834562734920e-06, {   6,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.182258831101e-06, -5.651365164364e-06, {   1,   1,   0,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.523457049270e-06, -3.052405816008e-07, {   1,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 { -2.146716978618e-06,  4.680574135322e-06, {   2,   0,   1,   0,   0,  11,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.707905877923e-06, -5.118731389450e-06, {   4,   0,  -2,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -6.461372572265e-07, -6.179236813133e-06, {   1,   1,  -3,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  6.598386514386e-06,  2.253353349225e-07, {   2,   0,   1,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -4.528379693732e-06, -2.293512478789e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0 } },
 { -1.096708128971e-06, -5.723729308430e-06, {   4,   0,  -1,   0,   0,   0,  -3,   0,   5,   0,   0,   0,   0 } },
 { -2.415649501027e-06,  4.402685477242e-06, {   1,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  7.888129782546e-07,  6.028671177556e-06, {   1,  -1,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  5.515764818001e-06,  1.301685606012e-06, {   1,  -1,   2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.436333075120e-08, -6.792512913330e-06, {   2,   0,  -3,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 {  4.411213270230e-06, -2.405150075318e-06, {   0,   0,   1,   0,   0,   2,  -3,   0,   2,   0,   0,   0,   0 } },
 { -4.876477898425e-06,  1.938797261281e-06, {   1,   0,   0,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 { -4.906764349361e-06, -1.906551187161e-06, {   2,   0,   0,   0,   0,  -6,  14,  -8,   0,   0,   0,   0,   0 } },
 {  5.163313698824e-06, -1.647969937973e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -5.163313698824e-06,  1.647969937973e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 {  5.421349099219e-06, -1.389834719381e-06, {   3,  -1,  -3,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.524234040177e-06, -2.285983006104e-06, {   4,   0,  -2,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -4.614567725664e-06,  2.195265626214e-06, {   2,   0,   1,   0,  -5,   0,  13,  -7,   0,   0,   0,   0,   0 } },
 { -1.218281976132e-06,  5.588671875125e-06, {   2,   0,  -2,   0,   0,  -7,   9,   0,   0,   0,   0,   0,   0 } },
 {  4.827748369979e-06, -1.979137196204e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   4,   0,   0,   0 } },
 { -5.008216681155e-06,  1.796429865793e-06, {   4,   0,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0,   0 } },
 { -1.251677943328e-06, -5.552901263484e-06, {   1,   1,   2,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 { -4.400204400856e-06,  2.404218499786e-06, {   1,  -1,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.517500811513e-07,  6.649686229547e-06, {   2,   0,  -1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.937419900601e-06,  8.619053042356e-07, {   4,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0 } },
 {  6.511586478111e-06,  2.802829852717e-07, {   2,   0,  -2,   0,  -1,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.642071205915e-07,  6.526417351478e-06, {   6,   0,  -3,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -2.432952569236e-08, -6.765485286735e-06, {   2,   0,   2,   0,   0,  14, -14,   0,   0,   0,   0,   0,   0 } },
 {  1.253073288291e-06, -5.536155898110e-06, {   1,   1,   0,   0,   0, -21,  23,   0,   0,   0,   0,   0,   0 } },
 { -2.327118069378e-06,  4.457732024705e-06, {   2,   0,   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -5.394600292483e-07, -6.242600591398e-06, {   2,   0,  -3,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -1.325940801867e-06,  5.455978979205e-06, {   1,  -1,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.592779092630e-06, -1.187059648597e-06, {   0,   0,   1,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.423803218742e-06, -1.354875735956e-06, {   0,   0,   0,   0,   0,  22, -25,   0,   0,   0,   0,   0,   0 } },
 {  5.413941461764e-06,  1.364161719693e-06, {   1,   1,   1,   0,   0, -46,  49,   0,   0,   0,   0,   0,   0 } },
 { -5.413941461764e-06, -1.364161719693e-06, {   1,   1,  -1,   0,   0, -46,  49,   0,   0,   0,   0,   0,   0 } },
 { -6.636409698174e-06,  1.376730807668e-07, {   2,  -2,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.036063371369e-06,  4.736665583298e-06, {   2,   0,   1,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -6.384814824465e-06, -3.852924691076e-07, {   2,  -2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.410885586839e-07, -6.624757100517e-06, {   2,   0,   1,   0,   0,   0,   6,   0,  -6,   0,   0,   0,   0 } },
 {  1.878822442990e-07,  6.576912709906e-06, {   6,   0,  -3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -3.103467543001e-06,  3.660768794100e-06, {   2,   0,  -1,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -4.844442917821e-06,  1.919180150572e-06, {   2,   0,   0,   0,   0,   6, -14,   8,   0,   0,   0,   0,   0 } },
 {  1.838487336500e-06,  4.919702830674e-06, {   1,  -1,   1,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 { -1.838487336500e-06, -4.919702830674e-06, {   1,  -1,  -3,   0,   0,  29, -33,   0,   2,   0,   0,   0,   0 } },
 {  5.006486240534e-06, -1.747133788383e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   7,  -5,   0,   0,   0 } },
 { -2.119029929115e-06,  4.633936144739e-06, {   2,   0,   0,   0,   0, -21,  23,   0,  -4,   0,   0,   0,   0 } },
 {  6.524178661764e-06,  2.278880896948e-07, {   0,   2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  6.578728386325e-08, -6.679185719557e-06, {   0,   0,   4,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -3.725058017980e-07,  6.367577520292e-06, {   1,   1,   0,   0,   0, -10,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.431975308726e-06,  5.307877964374e-06, {   2,  -2,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -1.998176620543e-06, -4.741403327463e-06, {   2,   0,   0,   0,   0,   0,   6, -11,  -2,   0,   0,   0,   0 } },
 {  5.459423173273e-06, -1.279917422460e-06, {   1,  -1,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.825192885368e-06,  9.135778818858e-07, {   1,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -5.246955694575e-06,  1.489371824309e-06, {   3,   0,   1,   0,   0,   0, -35,  41,  -2,   0,   0,   0,   0 } },
 {  5.153889660047e-06, -1.582378085932e-06, {   3,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.016694003105e-06, -7.183874265315e-07, {   2,   0,   0,   0,   0,   0,   2,  -7,   0,   0,   0,   0,   0 } },
 { -1.357198074830e-06, -5.376507798589e-06, {   1,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.437073735975e-06,  5.295839129131e-06, {   1,   0,   0,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.231018969797e-06, -4.966857726898e-07, {   1,   0,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -5.016041357149e-10,  6.722832679631e-06, {   3,   0,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.688750459929e-06, -1.029904094167e-06, {   3,  -1,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.942667040440e-06, -4.773239744439e-06, {   1,  -1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  6.457543741358e-06,  2.539897779720e-07, {   0,   0,   0,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 {  2.228349458929e-06,  4.479069656279e-06, {   0,   2,   0,   0,   0, -17,  15,   0,   0,   0,   0,   0,   0 } },
 {  6.290814418071e-06,  4.157181996108e-07, {   4,   0,  -1,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  6.016970609667e-06,  6.868656079902e-07, {   1,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.993922512091e-06, -7.089320745358e-07, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,  -4,   1,   0,   0 } },
 {  1.011339037727e-06,  5.690038143881e-06, {   3,  -2,   1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 { -1.131763412793e-06,  5.568807173311e-06, {   2,  -2,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.988450477897e-06,  7.090862485386e-07, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   4,  -1,   0,   0 } },
 {  2.167721055467e-06, -4.528984756078e-06, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 {  4.865084041826e-09,  6.691551411079e-06, {   1,   2,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.772145956947e-06, -9.202028454009e-07, {   2,   0,  -1,   0,   1,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.565447204400e-06, -5.126328387168e-06, {   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -5.214073441737e-06, -1.474489141830e-06, {   0,   2,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  5.315429508070e-06,  1.371644169268e-06, {   0,   0,   1,   0,   0,   0,   3,  -8,   4,   0,   0,   0,   0 } },
 { -7.039298456876e-07, -5.980982109813e-06, {   2,   0,  -1,   0,   0,  -3,   8,  -6,   0,   0,   0,   0,   0 } },
 {  4.997603054328e-06,  1.686698057641e-06, {   2,   0,   0,   0,   0,   0,  -6,   8,  -1,  -5,   0,   0,   0 } },
 {  1.508768412217e-07, -6.530698303026e-06, {   1,   3,   0,   0,   0, -43,  41,   0,   0,   0,   0,   0,   0 } },
 { -1.867691392539e-06, -4.813447018364e-06, {   3,  -1,  -1,   0,   0,  11, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.324089221300e-06, -1.354329896150e-06, {   1,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.443324511611e-06, -1.230879097223e-06, {   1,  -1,   0,   0,   0,  26, -25,   0,   0,   0,   0,   0,   0 } },
 { -6.412742981603e-06, -2.572359407432e-07, {   2,  -2,   0,   0,   0,  16, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.622632351569e-08, -6.642126039329e-06, {   0,   2,   0,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.623449614702e-06,  5.044247765681e-06, {   2,   0,  -3,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 {  5.838061613726e-06,  8.281869185723e-07, {   2,   0,  -1,   0,   0, -12,  11,   0,   0,   0,   0,   0,   0 } },
 {  1.679619347268e-06, -4.985935750399e-06, {   4,   0,  -2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.956017009750e-09,  6.655584327103e-06, {   1,   0,   1,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.321526519503e-06, -5.341947542286e-06, {   2,   0,  -1,   0,   0,  -2,   5,  -3,   0,   0,   0,   0,   0 } },
 {  6.637706822338e-06,  2.380687919019e-08, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -1.782202143204e-06, -4.878172050420e-06, {   4,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.343606504618e-08,  6.586667741342e-06, {   3,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.942525299783e-06, -7.160084441964e-07, {   1,  -1,   1,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0 } },
 { -5.942525299783e-06,  7.160084441964e-07, {   1,  -1,  -1,   0,   0,   0,   4,  -6,   3,   0,   0,   0,   0 } },
 { -1.833177642798e-06, -4.823191686736e-06, {   2,   0,   0,   0,   0,   0,  -3,   8,  -1,  -5,   0,   0,   0 } },
 { -1.155075357577e-06,  5.500169442090e-06, {   2,  -2,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 { -1.301568432608e-06,  5.350368348075e-06, {   3,  -1,  -2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.494680777865e-06,  1.153137531132e-06, {   6,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -6.295659262489e-06,  3.519613137082e-07, {   2,  -2,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.413851432802e-06,  2.315978704988e-07, {   0,   2,   0,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  1.180783864707e-06,  5.461267449591e-06, {   0,   2,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.480098838450e-06, -5.156254356120e-06, {   1,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.385732346642e-06,  1.245762690491e-06, {   0,   1,   2,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  1.181330634688e-06,  5.445239896342e-06, {   1,  -1,   0,   0,   0,  22, -24,   0,   0,   0,   0,   0,   0 } },
 { -6.227838317866e-06,  3.984034249056e-07, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -3.369008530095e-08, -6.586497313003e-06, {   1,   0,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.301545951543e-06, -1.316467852607e-06, {   0,   1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.524014437794e-06,  9.302631555182e-08, {   0,   0,   1,   0,   0,   0,  15, -18,   0,   0,   0,   0,   0 } },
 { -6.171971188028e-08,  6.554240389307e-06, {   2,   2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -6.342728676273e-06,  2.729198176750e-07, {   1,   1,  -1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  3.223478818849e-07, -6.291664147491e-06, {   2,   0,  -2,   0,   0,  -3,   4,   0,  -1,   0,   0,   0,   0 } },
 {  1.267106341877e-06,  5.342010173874e-06, {   2,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0 } },
 {  1.375322243824e-07,  6.470335299460e-06, {   6,  -4,  -1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -1.375322243824e-07, -6.470335299460e-06, {   2,  -4,   1,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  1.570498373285e-06,  5.033790008665e-06, {   1,   0,  -2,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.873603882249e-06, -7.300159811020e-07, {   2,   0,  -2,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 } },
 {  6.440890987746e-07, -5.958976408671e-06, {   0,   0,   1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 {  6.100941475129e-06, -5.015809850238e-07, {   4,   0,  -1,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -1.191472353900e-06, -5.408421826092e-06, {   6,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.262079465102e-06,  5.335614997133e-06, {   0,   0,   1,   0,   0,   8, -23,  19,   0,  -3,   0,   0,   0 } },
 {  1.262079465102e-06, -5.335614997133e-06, {   0,   0,   1,   0,   0,  -8,  23, -19,   0,   3,   0,   0,   0 } },
 { -5.951635669704e-06,  6.440607534749e-07, {   6,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.642382167821e-06, -4.952883749391e-06, {   4,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.523752118317e-06, -6.788412808861e-08, {   2,   2,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.334223387738e-06,  1.257200207774e-06, {   3,  -1,  -1,   0,   0,  11, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.288612268565e-06, -1.300719725507e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   6,   0,   0,   0 } },
 {  1.207232376889e-06, -5.381398397879e-06, {   3,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.929926444062e-06, -6.578695573270e-07, {   2,   0,  -2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0 } },
 { -8.496001032528e-07, -5.737015865090e-06, {   3,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.142830031440e-08, -6.563872368020e-06, {   3,   0,  -1,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 {  1.235642992730e-06, -5.348649087061e-06, {   1,  -1,   0,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -1.691079442367e-06, -4.891389930703e-06, {   1,  -1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.748234135259e-07, -6.206289012601e-06, {   2,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.326058954891e-06,  1.253561904123e-06, {   0,   1,   0,   0,   0, -10,   9,   0,   0,   0,   0,   0,   0 } },
 { -2.461377162439e-07, -6.332033279454e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -6,   0,   0,   0 } },
 { -1.471052483163e-07, -6.426801297907e-06, {   2,   0,   1,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.834510084892e-07, -5.589267073645e-06, {   6,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  6.392271510574e-06, -1.803561052102e-07, {   1,   0,   0,   0,   0, -21,  22,   0,   0,   0,   0,   0,   0 } },
 {  5.119355615174e-06,  1.446872120852e-06, {   3,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  1.678017033424e-06,  4.886405916827e-06, {   3,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.003725190213e-06,  2.559996363542e-06, {   1,   1,  -2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.209179716987e-06, -5.354482658344e-06, {   2,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.338217910412e-06,  2.200875518506e-07, {   0,   0,   0,   0,   0,  28, -29,   0,   0,   0,   0,   0,   0 } },
 {  5.305239588681e-06, -1.245421142823e-06, {   3,   1,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.179894161649e-07,  6.028833172620e-06, {   2,   0,   0,   0,   0,   0,  21, -27,   0,   0,   0,   0,   0 } },
 {  1.387481747898e-07,  6.407647509423e-06, {   3,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.342791538964e-06, -1.989446425309e-07, {   2,   2,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.574291494174e-06,  4.965961393929e-06, {   0,   0,   1,   0,   0,  -4,  15, -16,   0,   0,   0,   0,   0 } },
 { -7.044161637109e-07, -5.835006258922e-06, {   2,   0,   0,   0,   0,  -3,   8,  -6,   0,   0,   0,   0,   0 } },
 { -4.917485976408e-06, -1.621270423529e-06, {   0,   0,   1,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.514712517916e-07,  6.284803276060e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0 } },
 { -5.279924811764e-06, -1.254915472955e-06, {   3,  -1,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.572589858309e-06, -4.960418973009e-06, {   0,   0,   1,   0,   0,   4, -15,  16,   0,   0,   0,   0,   0 } },
 {  5.269536923950e-06, -1.254593962345e-06, {   2,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.869242799475e-06, -6.533801008730e-07, {   0,   0,   0,   0,   0,   3,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.681003776055e-06, -8.406604855129e-07, {   2,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0 } },
 {  6.437068523630e-07,  5.876892718272e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -5,   2,   0,   0,   0 } },
 {  4.667570965624e-06, -1.852692626416e-06, {   4,   0,  -4,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  4.689500885376e-06, -1.830458982395e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   4,  -5,   0,   0,   0 } },
 {  1.400802671424e-06, -5.119046078283e-06, {   1,  -1,   1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.728406156266e-06, -7.896263861713e-07, {   2,  -2,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  5.284908983676e-06,  1.233072753829e-06, {   6,   0,  -2,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.676820891062e-06,  1.838198521515e-06, {   4,   0,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  4.833593299253e-06,  1.681418309781e-06, {   4,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  6.158308488824e-06, -3.554819970929e-07, {   2,   0,  -1,   0,   0,  -3,   6,   0,  -2,   0,   0,   0,   0 } },
 {  5.097220367048e-06,  1.413668051749e-06, {   2,   0,  -2,   0,   0,  -7,  10,   0,   0,   0,   0,   0,   0 } },
 {  4.788493308947e-06,  1.719793354392e-06, {   2,   0,   0,   0,   0, -13,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.643521065103e-06,  1.863362457506e-06, {   2,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.214575342515e-06,  1.290335507507e-06, {   1,   1,   1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.277987436716e-06,  1.226736533153e-06, {   1,   1,  -2,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.080004871351e-06,  5.419762404041e-06, {   1,  -1,   2,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 {  1.161975844557e-06, -5.337667011471e-06, {   1,  -1,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.079269895881e-06, -5.416445509767e-06, {   1,  -1,  -2,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 } },
 { -5.255944126351e-06,  1.239589052941e-06, {   1,  -1,   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.229980041547e-06,  2.647981893233e-07, {   2,   0,  -2,   0,   0,   0,   3,   0,   0,  -4,   0,   0,   0 } },
 { -1.569619374484e-07,  6.336940361979e-06, {   4,  -2,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  2.643632596092e-06,  3.845347084607e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   6,   0,   0,   0,   0 } },
 { -5.053960746256e-06, -1.434732791637e-06, {   0,   0,   3,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  5.223117223458e-06, -1.263415729121e-06, {   1,  -1,   2,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.001499903826e-06,  1.483757948921e-06, {   1,   1,   0,   0,   0, -25,  28,   0,   0,   0,   0,   0,   0 } },
 { -5.529659657346e-06,  9.517505899745e-07, {   1,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  1.206530729396e-06,  5.273886253683e-06, {   1,  -1,   0,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 {  5.790649716273e-06, -6.889167516370e-07, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   5,   0,   0,   0 } },
 {  4.787194808890e-09, -6.473736259497e-06, {   1,   0,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.730482512281e-06, -4.742347468347e-06, {   2,   0,  -3,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 { -3.869175221062e-07, -6.081628231361e-06, {   2,   0,  -1,   0,   0,   3,  -4,   0,   1,   0,   0,   0,   0 } },
 { -7.424663716695e-08, -6.393966843369e-06, {   2,   0,   3,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.293434620674e-06, -5.172396141277e-06, {   3,   1,   1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  5.031037555903e-06, -1.432509730313e-06, {   2,  -2,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -4.767170844828e-06, -1.692654996565e-06, {   2,  -2,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -4.766817146298e-06,  1.692524058392e-06, {   2,  -2,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -2.256495430473e-07,  6.230077826081e-06, {   4,   0,  -2,   0,   0,   0,   7,  -7,   0,   0,   0,   0,   0 } },
 { -5.837634624253e-06,  6.180881423144e-07, {   4,   0,  -2,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0 } },
 { -3.072324038029e-07, -6.145730646271e-06, {   0,   2,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -3.072324038029e-07,  6.145730646271e-06, {   0,   2,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -5.251479655987e-06, -1.201406378415e-06, {   1,  -1,   1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.980066292567e-06,  4.718459192412e-07, {   2,   0,  -2,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.315377743472e-07,  5.818945419117e-06, {   4,   0,  -2,   0,   0,   0,   2,   0,  -4,   4,  -2,   0,   0 } },
 {  6.315377743472e-07, -5.818945419117e-06, {   0,   0,   2,   0,   0,   0,  -2,   0,   4,  -4,   2,   0,   0 } },
 {  1.342213335374e-06,  5.107033333846e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -3,   7,   0,   0,   0 } },
 {  5.125014273484e-06,  1.323515790808e-06, {   0,   0,   1,   0,   0,   0,  -7,  12,  -3,   0,   0,   0,   0 } },
 { -5.239074663735e-06, -1.208917399655e-06, {   4,   0,  -4,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  3.302644188951e-07,  6.117473546204e-06, {   4,  -2,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 { -5.492436668056e-06, -9.538173454997e-07, {   1,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  6.210705737948e-06, -2.355202114207e-07, {   0,   2,   0,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 {  5.209801450442e-06,  1.233076371565e-06, {   1,  -3,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.210726341080e-06, -2.300738692279e-07, {   0,   0,   1,   0,   0, -25,  26,   0,   0,   0,   0,   0,   0 } },
 { -5.099027076284e-06,  1.341725474797e-06, {   2,  -1,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.263486542981e-06, -5.172780260998e-06, {   1,   1,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.192699276381e-06, -1.242041592710e-06, {   0,   1,  -2,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.199115305388e-06,  2.354392096162e-07, {   2,  -2,  -1,   0,   0,  26, -27,   0,   0,   0,   0,   0,   0 } },
 { -5.373235545834e-06, -1.057773345475e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   1,   0 } },
 {  6.060566550881e-06,  3.641591235192e-07, {   2,   0,   2,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -5.243927280567e-06, -1.178156958295e-06, {   2,   0,   3,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.442064225247e-06, -4.975676856320e-06, {   2,   0,  -2,   0,   0,   8,  -9,  -8,   3,   0,   0,   0,   0 } },
 {  4.836665443109e-06,  1.576246139295e-06, {   0,   2,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.124966861345e-06, -1.283242617493e-06, {   3,  -1,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  2.832888647818e-07, -6.122608569988e-06, {   3,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -6.303022128872e-06,  9.780672237358e-08, {   0,   0,   4,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.829108800518e-06, -1.571379871489e-06, {   0,   0,   0,   0,   0,  21, -24,   0,   0,   0,   0,   0,   0 } },
 { -5.192946959070e-06, -1.205620372079e-06, {   1,  -1,   0,   0,   0,   0,   6,  -6,   0,   0,   0,   0,   0 } },
 {  5.122579453970e-06, -1.267427255387e-06, {   2,   0,   2,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -5.167238601309e-06, -1.221910032873e-06, {   1,  -1,  -2,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.167294096731e-06, -1.221739777107e-06, {   3,   1,  -2,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.088591249811e-06, -5.299909026066e-06, {   1,   1,  -1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 {  6.121222944351e-06,  2.610435245938e-07, {   2,   0,   0,   0,   0, -17,  16,   0,   2,   0,   0,   0,   0 } },
 {  4.702382461242e-07,  5.909587574474e-06, {   4,  -2,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.078357527526e-06,  1.299835275685e-06, {   0,   0,   1,   0,   0,  -1,  -1,   5,   0,   0,   0,   0,   0 } },
 {  5.082011877047e-06, -1.296085336577e-06, {   0,   0,   1,   0,   0,   1,   1,  -5,   0,   0,   0,   0,   0 } },
 {  6.055784917321e-08, -6.316783720888e-06, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0 } },
 {  4.987966093760e-06,  1.388817723353e-06, {   3,   0,  -2,   0,   0,   0,  -8,  16,  -4,  -5,   0,   0,   0 } },
 {  4.987949014009e-06, -1.388788190884e-06, {   3,   0,  -2,   0,   0,   0,   8, -16,   4,   5,   0,   0,   0 } },
 { -9.529646498971e-07,  5.422423135111e-06, {   0,   0,   2,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -1.525181376854e-07,  6.222427617698e-06, {   2,   0,  -1,   0,   0, -23,  23,   0,   0,   0,   0,   0,   0 } },
 {  5.327420720316e-06, -1.045047189297e-06, {   1,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -1.887762368394e-08, -6.351856995535e-06, {   3,   0,   0,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 { -1.288882724862e-06,  5.081668074016e-06, {   0,   0,   1,   0,   0,   0,  -3,   8,   0,   0,   0,   0,   0 } },
 { -1.298050883794e-06, -5.071993043593e-06, {   1,  -1,   0,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.338212241576e-08, -6.355026725723e-06, {   2,   0,  -4,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -6.181807128386e-06, -1.846783656528e-07, {   2,   0,   2,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  4.614808120392e-09, -6.361672460179e-06, {   3,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.754920406055e-06,  1.608715986419e-06, {   2,   0,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0,   0 } },
 {  1.118717837299e-06, -5.243431108363e-06, {   4,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.270115799348e-06,  1.088060568467e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   2,  -8,   0,   0,   0 } },
 { -6.612596476890e-07, -5.692561416795e-06, {   2,  -2,   2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.115518784152e-06, -2.376145959451e-07, {   2,  -1,   0,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 { -4.216865533629e-08,  6.310623999377e-06, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,  -2,   0 } },
 { -1.854418154316e-07, -6.164942623759e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   3,   0,   0,   0 } },
 {  5.475130354673e-06,  8.741864257685e-07, {   2,   0,  -1,   0,   0,   0,  13, -22,   0,   0,   0,   0,   0 } },
 { -5.001308322224e-06,  1.347578363968e-06, {   2,   0,  -3,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -9.570779305570e-07, -5.390471236263e-06, {   3,  -2,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  5.201389503998e-07, -5.822345296735e-06, {   1,   1,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  1.007906822470e-06,  5.333367720113e-06, {   0,   2,  -2,   0,   0,  21, -23,   0,   0,   0,   0,   0,   0 } },
 { -4.739290113750e-06,  1.601071618866e-06, {   1,  -1,   0,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.122388360571e-06,  1.214446143880e-06, {   1,  -1,   2,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.383503423564e-09,  6.330377627846e-06, {   4,  -2,  -1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.276696967080e-07, -6.002519049540e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  2.641940668017e-09,  6.325601027636e-06, {   4,   0,  -2,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.704822972083e-08, -6.280636519860e-06, {   4,  -2,  -3,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.141517405632e-06,  1.184683216010e-06, {   1,  -1,   0,   0,   0, -20,  19,   0,   0,   0,   0,   0,   0 } },
 {  6.092863490302e-06,  2.316347363460e-07, {   2,  -2,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -5.807083480864e-06, -5.173960137921e-07, {   4,   0,  -1,   0,   0, -24,  27,   1,   0,   0,   0,   0,   0 } },
 { -5.807083480864e-06,  5.173960137921e-07, {   0,   0,   1,   0,   0,  24, -27,  -1,   0,   0,   0,   0,   0 } },
 { -1.342098221224e-06,  4.982237565368e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -6,   0,   0,   0,   0 } },
 { -5.380608802344e-06, -9.429782015191e-07, {   0,   0,   3,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.426224991157e-06,  4.896129704695e-06, {   4,  -2,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.865490953495e-07, -5.933762255509e-06, {   6,   0,  -3,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.978268096599e-07, -5.722471069921e-06, {   0,   0,   2,   0,   0,   0,   1,   0,  -3,   2,   0,   0,   0 } },
 { -6.081235140702e-06, -2.360456901651e-07, {   2,  -2,   0,   0,   0,  24, -25,   0,   0,   0,   0,   0,   0 } },
 { -1.418106448236e-06,  4.894092886398e-06, {   2,   0,  -2,   0,   0,  -8,   9,   8,  -3,   0,   0,   0,   0 } },
 {  2.805729053191e-07, -6.028515163986e-06, {   2,   0,   1,   0,   0,   0,  -5,   8,  -1,  -5,   0,   0,   0 } },
 {  1.054109127102e-06,  5.254607348298e-06, {   1,  -1,   1,   0,   0,  12, -14,   0,   0,   0,   0,   0,   0 } },
 { -4.925814538604e-06, -1.380236566218e-06, {   2,   0,  -1,   0,   0,   0,  17, -32,   0,   0,   0,   0,   0 } },
 {  8.549512034927e-07,  5.445203266438e-06, {   1,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  5.227865269208e-06,  1.071376061674e-06, {   1,   1,  -1,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -4.845439217308e-06, -1.451538417315e-06, {   2,   0,  -1,   0,   0,  -9,  12,   0,   0,   0,   0,   0,   0 } },
 { -5.354659201978e-06, -9.422848211238e-07, {   3,  -1,  -1,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 { -6.011199396656e-06,  2.811182291042e-07, {   3,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.297317247090e-06,  9.940710421487e-07, {   2,  -2,   2,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.220931705482e-07, -5.364229279569e-06, {   1,   0,  -2,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  6.153414251029e-06, -1.320643944895e-07, {   1,   0,   0,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 { -6.255899147757e-06,  2.241874954717e-08, {   0,   0,   1,   0,   0,   6, -11,   0,   3,   0,   0,  -1,   0 } },
 {  5.580413033726e-06,  6.974028938019e-07, {   0,   0,   2,   0,   0,   0,  -6,   8,  -3,   0,   0,   0,   0 } },
 { -1.046305943105e-06,  5.229375478427e-06, {   1,   1,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -6.979060330665e-07, -5.576524904761e-06, {   3,  -1,  -1,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  5.194354419429e-06,  1.079952897707e-06, {   0,   0,   1,   0,   0,   0,  -9,  14,  -3,   0,   0,   0,   0 } },
 { -1.197844697036e-06, -5.076168908599e-06, {   2,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -6.027488603686e-06, -2.465053407721e-07, {   6,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.027488603686e-06,  2.465053407721e-07, {   2,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.576084234455e-08, -6.177470430021e-06, {   3,  -2,  -1,   0,   0,  11,  -5,   3,   0,   0,   0,   0,   0 } },
 {  9.576084234455e-08,  6.177470430021e-06, {   1,   2,  -1,   0,   0, -11,   5,  -3,   0,   0,   0,   0,   0 } },
 {  5.022409240392e-06, -1.250683614522e-06, {   0,   0,   1,   0,   0,  15, -18,   0,   0,   0,   0,   0,   0 } },
 {  4.963340290273e-06,  1.306540221646e-06, {   1,  -1,  -1,   0,   0, -11,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.307793639486e-06,  9.584644303924e-07, {   1,  -1,   2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  1.089116310750e-06, -5.172695457611e-06, {   1,  -1,   0,   0,   0,  -9,  11,   0,   0,   0,   0,   0,   0 } },
 { -5.254257497760e-06, -1.006648117268e-06, {   1,  -1,   0,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  6.256017112850e-06,  3.501022306197e-09, {   2,   0,  -1,   0,   0,   0,  20, -23,   0,   0,   0,   0,   0 } },
 {  5.136384114399e-06,  1.117321062625e-06, {   2,   0,   0,   0,   0, -18,  13,   0,   0,   0,   0,   0,   0 } },
 { -9.963475233376e-07, -5.254796200567e-06, {   1,  -1,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  4.159771734566e-07,  5.830709350105e-06, {   0,   2,  -3,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 } },
 { -1.898817300724e-07,  6.055852119680e-06, {   2,   0,   2,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.401487988322e-06, -4.839078884221e-06, {   1,  -1,   0,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -3.316709208373e-08,  6.205664915313e-06, {   2,   0,  -3,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -6.001275940114e-06,  2.308807641682e-07, {   0,   2,   0,   0,   0, -22,  23,   0,   0,   0,   0,   0,   0 } },
 {  9.942859548239e-07,  5.237554356329e-06, {   1,  -1,   3,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.939719558540e-07, -5.235594496320e-06, {   1,  -1,  -3,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 {  4.806954577098e-06, -1.421493860154e-06, {   2,   0,  -1,   0,   0,   1,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  5.969499371683e-06,  2.558050071993e-07, {   2,   2,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.802310449542e-06,  1.421054588281e-06, {   2,   0,  -1,   0,   0,   1,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.410621101096e-06,  4.812120739202e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -1,   2,   0 } },
 {  2.818303920184e-08, -6.188256856523e-06, {   4,   0,   1,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.005033306857e-09, -6.207168893827e-06, {   3,   0,  -2,   0,   0,  11, -11,   0,   0,   0,   0,   0,   0 } },
 {  9.018206287457e-07,  5.304473093908e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,  -4,   5,   0,   0,   0 } },
 { -5.979354638924e-06,  2.268735774546e-07, {   0,   2,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -4.846170613850e-06,  1.359692944320e-06, {   2,   0,  -1,   0,   0,   0, -17,  32,   0,   0,   0,   0,   0 } },
 {  5.246569192755e-08, -6.152110554689e-06, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0 } },
 { -1.613107667946e-07, -6.040586873420e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 } },
 { -6.051157483649e-06,  1.504332262024e-07, {   2,   0,   1,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.870655358027e-06, -3.296008039245e-07, {   2,   0,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  9.116820420088e-09,  6.188382359922e-06, {   2,   0,   1,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0 } },
 { -1.017375842482e-06, -5.179087068446e-06, {   2,   0,  -1,   0,   0,   6, -10,   0,   2,   0,   0,   0,   0 } },
 { -8.105276911358e-07, -5.383111194712e-06, {   0,   2,   0,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 { -5.927814149473e-06, -2.651054613190e-07, {   2,   0,   1,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 {  3.632619451355e-07,  5.826932476519e-06, {   0,   0,   0,   0,   0,   0,  11, -11,   0,   0,   0,   0,   0 } },
 {  4.874575233046e-07, -5.700096574447e-06, {   2,   0,  -1,   0,   0,   0,  20, -26,   0,   0,   0,   0,   0 } },
 {  4.822476354214e-06, -1.362784254090e-06, {   1,   0,   0,   0,   0,   0,   7, -12,   3,   0,   0,   0,   0 } },
 {  5.795651738209e-06,  3.893448528629e-07, {   2,   0,   2,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  5.029039074657e-06, -1.152943477096e-06, {   1,   1,   1,   0,   0, -22,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.003189822390e-06,  1.177766415284e-06, {   1,   1,   1,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0 } },
 {  1.652730990100e-07,  6.015627014965e-06, {   2,   0,  -1,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.726491228923e-06,  4.540805435196e-07, {   1,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  5.015762699334e-06, -1.162259329256e-06, {   1,   1,   1,   0,   0, -15,  14,   0,   0,   0,   0,   0,   0 } },
 { -5.951675892345e-06,  2.230422951641e-07, {   0,   2,   1,   0,   0, -26,  27,   0,   0,   0,   0,   0,   0 } },
 {  8.975645830980e-07, -5.276698478670e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   4,  -5,   0,   0,   0 } },
 { -5.948565907190e-06, -2.255556910673e-07, {   1,  -1,   1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -4.905264291453e-06, -1.264570904616e-06, {   4,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.074597423501e-06, -5.092255678813e-06, {   1,   1,   0,   0,   0, -15,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.589431232173e-06,  5.722938013252e-07, {   2,   0,   0,   0,   0,  23, -26,   0,   0,   0,   0,   0,   0 } },
 {  7.566714827049e-07, -5.403543489505e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,  -5,   0,   0,   0 } },
 {  1.513659288538e-08,  6.144040150228e-06, {   2,   0,   0,   0,   0,  28, -30,   0,   0,   0,   0,   0,   0 } },
 { -5.567376858792e-06,  5.911320159195e-07, {   2,   0,  -2,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.557565667567e-06, -6.001667065396e-07, {   0,   0,   1,   0,   0,  14, -11,   0,   0,   0,   0,   0,   0 } },
 {  3.812054906046e-08,  6.118285775978e-06, {   2,  -2,   1,   0,   0,  22, -22,   0,   0,   0,   0,   0,   0 } },
 { -1.150903405994e-06, -4.991686384153e-06, {   4,   0,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -1.311715127911e-06,  4.827350987186e-06, {   6,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.943711839952e-06, -1.193709842282e-06, {   2,  -1,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.830436481527e-06,  3.046315992840e-07, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   1,   0,   0,   0 } },
 {  1.165012725971e-06, -4.965237770746e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  4.953682388104e-06, -1.176352111204e-06, {   2,  -1,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.009584577300e-06,  1.189755488989e-07, {   1,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.142722906460e-08, -6.113370685293e-06, {   0,   0,   1,   0,   0,  -1,   3,   0,  -2,   0,   0,   0,   0 } },
 {  3.052888074649e-07, -5.818808268862e-06, {   1,  -1,  -2,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.982454066795e-06, -1.407247182263e-07, {   2,   0,   1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -6.582436970683e-07,  5.464887572747e-06, {   2,   0,  -1,   0,   0,   3,  -8,   6,   0,   0,   0,   0,   0 } },
 { -6.008727962523e-06, -1.142670923229e-07, {   2,   0,   1,   0,   0, -18,  19,   0,   0,   0,   0,   0,   0 } },
 {  3.934540272608e-08, -6.081102334032e-06, {   2,   0,   2,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -5.918121821778e-06,  2.022739148480e-07, {   2,   0,   2,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  5.297549658305e-06, -8.209898428300e-07, {   2,   0,  -1,   0,   0,  20, -23,   0,   0,   0,   0,   0,   0 } },
 {  1.080504795047e-06, -5.032703992332e-06, {   1,   1,  -1,   0,   0,  12, -12,   0,   0,   0,   0,   0,   0 } },
 { -5.956595608735e-06, -1.543096437761e-07, {   2,   0,   1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 {  5.242155004809e-06, -8.685980901352e-07, {   4,  -2,   1,   0,   0, -13,  10,   0,   0,   0,   0,   0,   0 } },
 {  5.242155004809e-06,  8.685980901352e-07, {   0,   2,  -1,   0,   0,  13, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.988298993197e-07,  5.211518807019e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 } },
 {  4.954793548413e-06,  1.151976741261e-06, {   2,  -2,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.934385189523e-06, -1.172037963957e-06, {   3,  -1,   0,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.574094662204e-06,  5.274093869300e-07, {   2,   2,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.435985645544e-07, -5.957500320539e-06, {   2,   0,  -2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -5.162500649607e-06, -9.330105957442e-07, {   1,  -1,  -2,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  4.959205940510e-08,  6.045715600176e-06, {   3,  -2,  -3,   0,   0,  26, -26,   0,   0,   0,   0,   0,   0 } },
 { -7.761089501792e-07,  5.318022588645e-06, {   1,   1,   0,   0,   0, -11,   9,   0,   0,   0,   0,   0,   0 } },
 {  1.005199228014e-06, -5.088488931883e-06, {   0,   2,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.941618932918e-06, -1.150550771274e-06, {   1,   1,   1,   0,   0, -13,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.996654489435e-06, -1.093002667008e-06, {   0,   0,   1,   0,   0, -16,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.956693593261e-06,  1.282471667270e-07, {   2,   0,   1,   0,   0, -13,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.288716201860e-07, -5.355339797002e-06, {   4,   0,  -1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.946293922583e-06,  1.136956205017e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -6,   3,   0,   0 } },
 {  5.400421226234e-07,  5.541301444736e-06, {   1,  -1,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -5.947481743669e-06,  1.333340791943e-07, {   0,   0,   2,   0,   0, -24,  25,   0,   0,   0,   0,   0,   0 } },
 { -5.962838024930e-06, -1.172382819116e-07, {   3,   0,  -3,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.135782104327e-08,  6.018477918540e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,   1,   1,   0,   0,   0 } },
 { -5.656077755681e-06, -4.209832877685e-07, {   0,   2,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  4.915074774599e-06,  1.161009995271e-06, {   0,   1,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.967034909095e-08, -6.016377022926e-06, {   4,  -2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 {  1.196445201899e-08, -6.060993786735e-06, {   0,   2,  -1,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 { -1.003201272522e-06, -5.068551938998e-06, {   0,   2,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  4.931317986146e-06, -1.134881720076e-06, {   3,   1,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.208057681829e-06, -4.855013019629e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -5,   6,  -3,   0,   0 } },
 {  5.951762287178e-06, -1.095484397400e-07, {   2,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.970492904906e-07,  5.057322488212e-06, {   2,   0,  -1,   0,   0, -14,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.542617718714e-08, -5.985984564740e-06, {   4,   0,  -1,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 { -7.374684661247e-07, -5.311187302544e-06, {   2,   0,  -2,   0,   0,  24, -26,   0,   0,   0,   0,   0,   0 } },
 {  4.906399556750e-06, -1.139088970661e-06, {   1,   1,   1,   0,   0, -14,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.181033765569e-06,  4.863922042754e-06, {   3,  -1,   0,   0,   0,  26, -28,   0,   0,   0,   0,   0,   0 } },
 {  1.156691297540e-07,  5.928998784658e-06, {   3,   1,   0,   0,  -1, -15,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.156691297540e-07, -5.928998784658e-06, {   1,  -1,   2,   0,   1,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.354732590207e-08,  6.030211728255e-06, {   2,   0,   3,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.550825067952e-07,  5.485413347864e-06, {   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  1.612667185991e-07, -5.867861648536e-06, {   3,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -9.906870368642e-07,  5.037177996308e-06, {   0,   0,   1,   0,   0, -19,  21,   0,   0,   0,   0,   0,   0 } },
 {  5.115045640050e-06, -9.095845338976e-07, {   4,   0,  -3,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.775670099461e-07,  5.543013029462e-06, {   1,   1,  -1,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.626816693100e-07, -5.056929789237e-06, {   2,  -2,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 {  9.137547531031e-07,  5.103930547853e-06, {   0,   0,   1,   0,   0,   2,  -4,   0,   3,   0,   0,   0,   0 } },
 {  5.729274197901e-06,  2.832529091867e-07, {   2,   0,   1,   0,   0, -16,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.256333631600e-07, -5.685395665593e-06, {   0,   0,   0,   0,   0,   0,   6, -12,   0,   0,   0,   0,   0 } },
 {  5.739011379612e-06,  2.618059534630e-07, {   3,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.480413404860e-09, -5.995575674153e-06, {   2,   0,   2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -9.177725459865e-07,  5.076428271472e-06, {   2,  -2,   0,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -1.921133938257e-09,  5.987727465883e-06, {   1,  -2,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.633729975202e-06,  3.516727514056e-07, {   2,   0,   2,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  8.495027515529e-08,  5.900111442298e-06, {   2,  -2,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  5.655259113878e-06, -3.277612233033e-07, {   2,   0,  -4,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -5.720410922796e-06, -2.618353290610e-07, {   3,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -4.068122854622e-08, -5.940913485490e-06, {   0,   0,   0,   0,   0,   2,  -4,   0,   2,   0,   0,   0,   0 } },
 { -7.045342926614e-07,  5.274872934785e-06, {   1,  -1,   1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 {  7.045342926614e-07, -5.274872934785e-06, {   1,  -1,  -1,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0 } },
 { -2.238521461683e-08, -5.952719737559e-06, {   2,   0,  -2,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 { -8.290961358852e-09, -5.966738360992e-06, {   0,   2,  -2,   0,   0,  21, -21,   0,   0,   0,   0,   0,   0 } },
 { -3.851365682790e-07, -5.588943229550e-06, {   1,   0,  -1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 {  7.547968721958e-07,  5.209132639345e-06, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 {  4.931336623997e-06, -1.028709891506e-06, {   0,   0,   1,   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0 } },
 { -4.987341118067e-06, -9.710402700589e-07, {   2,   0,  -1,   0,   0,  26, -31,   0,   0,   0,   0,   0,   0 } },
 {  3.257079322553e-07,  5.627257907704e-06, {   1,  -1,   0,   0,   0,  -7,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.488898158757e-06, -4.626749286125e-07, {   2,   0,  -1,   0,   0,  22, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.180534996254e-07, -5.832927407789e-06, {   0,   0,   0,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 { -8.437206060117e-09,  5.940639257425e-06, {   1,   0,   1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  3.699042556081e-08,  5.910933469176e-06, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,  -2,   0,   0,   0 } },
 { -1.012870030905e-06, -4.935011846069e-06, {   2,   0,   0,   0,   0,   6, -10,   0,   2,   0,   0,   0,   0 } },
 { -5.694596379940e-06, -2.527351965392e-07, {   2,   0,  -2,   0,   0,   0,  12, -15,   0,   0,   0,   0,   0 } },
 { -5.062412503974e-06, -8.821253114561e-07, {   0,   0,   0,   0,   0,   0,   6, -16,   4,   5,   0,   0,   0 } },
 { -5.011487687818e-06, -9.308001750329e-07, {   3,   1,  -3,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -8.291898248360e-07,  5.109255828444e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0 } },
 { -9.553939701222e-07,  4.979781877080e-06, {   1,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  3.677955652076e-07,  5.564366880606e-06, {   2,   0,  -1,   0,   0,   0,  -8,  13,   1,   0,   0,   0,   0 } },
 {  9.253424257786e-08, -5.838972115298e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   1,   1,   0,   0,   0 } },
 {  8.831446129343e-07,  5.044073685846e-06, {   0,   0,   1,   0,   0,   0, -12,  21,  -2,   0,   0,   0,   0 } },
 { -9.038561649671e-07, -5.020996394585e-06, {   6,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.969403766120e-07,  5.527061323415e-06, {   1,   0,   0,   0,   0,   0,  -4,   0,   4,   0,   0,   0,   0 } },
 { -5.250000014195e-06, -6.710946862495e-07, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,  -2,   0,   0 } },
 { -4.915884164735e-06, -9.963830156100e-07, {   0,   0,   1,   0,   0,  -3,   7,   0,   0,   0,   0,   0,   0 } },
 { -9.673535615000e-08,  5.812062225358e-06, {   1,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  6.208117031389e-08,  5.839511706454e-06, {   2,   0,  -1,   0,   0,  29, -29,   0,   0,   0,   0,   0,   0 } },
 {  1.484513310267e-08,  5.881841238151e-06, {   3,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.733141553435e-06,  1.514587706407e-07, {   0,   0,   2,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -5.051396553013e-09,  5.876794528229e-06, {   0,   0,   3,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -5.575231813883e-06, -3.024751946346e-07, {   4,   0,  -2,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.645093439072e-06, -2.265859662039e-07, {   2,  -2,   0,   0,   0,  17, -18,   0,   0,   0,   0,   0,   0 } },
 { -8.269645325757e-07, -5.043834373283e-06, {   2,  -2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.501121569071e-07,  5.019038318814e-06, {   3,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  5.430934499494e-06,  4.380309336016e-07, {   3,  -2,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.392435610296e-07,  4.925395027181e-06, {   3,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  5.773440135665e-06, -8.784368566227e-08, {   1,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  9.875573792531e-08,  5.757246172513e-06, {   0,   2,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.548648174377e-06, -3.055473960501e-07, {   2,  -2,  -1,   0,   0,  11, -14,   0,   0,   0,   0,   0,   0 } },
 { -9.247988206716e-07, -4.927959979096e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 } },
 {  2.033427457723e-07,  5.645178875139e-06, {   2,   0,   1,   0,   0,  17, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.435680755876e-08,  5.833279012136e-06, {   3,   0,  -3,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.736273384354e-07, -5.571401491866e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -1.560170475341e-07,  5.680460365417e-06, {   0,   0,   3,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  5.117892023583e-06, -7.148122428745e-07, {   0,   0,   1,   0,   0,   0, -11,  20,   0,   0,   0,   0,   0 } },
 {  5.384574736809e-08,  5.768106505131e-06, {   4,   0,  -1,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -1.694430289435e-08,  5.803361110705e-06, {   4,   0,   0,   0,   0, -19,  19,   0,   0,   0,   0,   0,   0 } },
 { -5.600556769208e-06,  2.177761273937e-07, {   0,   2,   0,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  8.898897811538e-07,  4.928402663544e-06, {   2,   0,   0,   0,   0,   2,  -4,   0,   3,   0,   0,   0,   0 } },
 {  6.906474410942e-07,  5.124831101442e-06, {   2,  -2,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 { -4.108323580656e-07, -5.404126967978e-06, {   2,   0,   0,   0,   0,   0,   8,   0,  -8,   0,   0,   0,   0 } },
 {  5.122612320355e-06,  6.845866232054e-07, {   2,   0,  -3,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -7.304790738000e-09, -5.795798727213e-06, {   0,   0,   2,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -7.760130437180e-07, -5.025513025206e-06, {   1,  -1,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  6.026686670003e-07,  5.195762523239e-06, {   4,   0,  -2,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  5.721818987224e-06, -7.172150449649e-08, {   4,   0,  -1,   0,   0,  -7,   8,   0,   0,   0,   0,   0,   0 } },
 { -7.473878671853e-07,  5.043731113392e-06, {   1,  -1,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.516567386033e-06, -2.682110439030e-07, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,  -2,   0,   0,   0 } },
 { -5.844345365117e-07, -5.200137333442e-06, {   1,   1,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.889922655029e-07, -4.993822359661e-06, {   1,   0,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -4.926901301032e-07,  5.289251899841e-06, {   4,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  5.241021404731e-06,  5.403162502651e-07, {   0,   0,   3,   0,   0, -16,  16,  -7,   8,  -6,   0,   0,   0 } },
 {  5.241021404731e-06, -5.403162502651e-07, {   0,   0,   1,   0,   0,  16, -16,   7,  -8,   6,   0,   0,   0 } },
 {  1.053364094127e-08, -5.770603667111e-06, {   1,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.505696527014e-06,  2.698183928082e-07, {   1,   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 } },
 { -5.573458035784e-06, -1.973208597796e-07, {   0,   0,   2,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.991820078054e-07,  5.171583084580e-06, {   0,   0,   2,   0,   0,   0,  -4,   0,   5,   0,   0,   0,   0 } },
 {  5.424610893642e-06, -3.424861502038e-07, {   2,   0,  -3,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -6.511860306943e-10, -5.763981841006e-06, {   1,   0,  -1,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 {  5.718700748005e-06, -4.340592203094e-08, {   1,   0,  -3,   0,   0,   0,   4,   0,  -4,   0,   0,   0,   0 } },
 { -2.826551791393e-07,  5.475513783237e-06, {   1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 } },
 { -3.539251798341e-06,  2.217324431592e-06, {   0,   0,   2,   0,   0,   4,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.421344572232e-06,  3.341329264752e-07, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -3,   0,   0 } },
 {  5.505025205248e-06,  2.500870619166e-07, {   2,   0,   1,   0,   0,  22, -23,   0,   0,   0,   0,   0,   0 } },
 { -1.300271865141e-08, -5.739453971081e-06, {   2,  -2,  -1,   0,   0,  15, -15,   0,   0,   0,   0,   0,   0 } },
 {  1.463524339383e-08, -5.735971766467e-06, {   4,   0,   0,   0,   0, -11,  11,   0,   0,   0,   0,   0,   0 } },
 {  9.993615156880e-09, -5.740217613125e-06, {   0,   0,   4,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 {  7.042129175575e-07,  5.042546512638e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   5,   0,   0,   0 } },
 { -5.665016104031e-06,  7.911392555321e-08, {   2,   0,   1,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -6.414337278697e-08, -5.679485934029e-06, {   2,   0,   1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 {  1.171219390190e-08, -5.730330289267e-06, {   2,   0,  -2,   0,   0, -21,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.359849589660e-08, -5.725519388920e-06, {   0,   2,   0,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -6.663893448302e-07, -5.070108307624e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   0,   4,   0,   0,   0 } },
 {  1.883692795693e-08, -5.714994554371e-06, {   4,   0,  -2,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  5.454516607753e-07,  5.179586715552e-06, {   2,   0,  -2,   0,  -3,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.075653826807e-08,  5.635990383462e-06, {   3,   0,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.551415338566e-06, -1.591407486175e-07, {   1,   1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  5.415864158458e-07, -5.167511760410e-06, {   0,   0,   1,   0,   0,   0,  12, -18,   0,   0,   0,   0,   0 } },
 {  3.100974193712e-09, -5.705410857644e-06, {   2,  -2,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.776944097089e-07,  5.226631182765e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,  -2,   8,   0,   0,   0 } },
 {  5.265751351755e-06,  4.382578098017e-07, {   4,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.642899865130e-06,  5.070749860966e-08, {   1,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.932647559220e-07, -4.799435142960e-06, {   0,   0,   0,   0,   0,   0,   5,   0,  -6,   0,   0,   0,   0 } },
 {  3.351112574741e-07,  5.355543584929e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 { -5.597933213780e-06,  9.198374896115e-08, {   2,   0,   1,   0,   0, -15,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.963231381142e-07,  5.393485115309e-06, {   3,   0,  -1,   0,   0, -15,  11,   0,   0,   0,   0,   0,   0 } },
 { -6.989276245763e-07,  4.986342200259e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   6,  -7,   0,   0,   0 } },
 { -5.048425924354e-06,  6.363964347280e-07, {   2,  -2,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.833112917671e-07, -5.099612133933e-06, {   2,   0,  -1,   0,   0,   0,   5,  -6,   1,   0,   0,   0,   0 } },
 {  5.509544267972e-06,  1.732619917784e-07, {   2,   0,  -3,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  4.226071608331e-07,  5.258656486599e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -2,  -2,   0,   0,   0 } },
 {  5.064287404261e-06, -6.161666976155e-07, {   2,   0,   0,   0,   0,   0, -12,  21,   0,   0,   0,   0,   0 } },
 { -6.231510126954e-07,  5.050610911080e-06, {   2,   0,   0,   0,   0,   3,  -8,   6,   0,   0,   0,   0,   0 } },
 { -4.977006461382e-06,  6.966059572663e-07, {   1,   1,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  4.960943977534e-06,  7.093549293828e-07, {   2,   0,  -1,   0,   0,   0,   8, -12,   1,   0,   0,   0,   0 } },
 {  5.410810753698e-06,  2.450589144811e-07, {   2,   0,  -2,   0,   0,   9,  -8,   0,   0,   0,   0,   0,   0 } },
 { -5.582281441033e-06,  7.340231684455e-08, {   4,  -2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.385810608036e-06, -2.681470443802e-07, {   3,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  3.082658387824e-06,  2.565787756274e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   5,  -2,   0,   0,   0 } },
 { -5.543842193976e-07, -5.089728674894e-06, {   2,   0,   0,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -5.431524902354e-06, -2.114095599458e-07, {   6,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  2.685462688957e-07,  5.371871632892e-06, {   0,   0,   4,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 {  2.685462688957e-07, -5.371871632892e-06, {   0,   0,   2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  2.763753113296e-07,  5.359569097588e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0 } },
 {  8.036455984568e-08, -5.548370096335e-06, {   5,   0,  -3,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.335753828902e-06, -2.787946604452e-07, {   0,   2,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.183039082925e-06, -4.247349387381e-07, {   2,   0,   0,   0,   0,   2,  -5,   0,   2,   0,   0,   0,   0 } },
 {  5.220875960449e-06,  3.772031134273e-07, {   2,   0,   0,   0,   0,  26, -25,   0,   0,   0,   0,   0,   0 } },
 { -4.580276373229e-09,  5.591676734792e-06, {   3,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -5.078498356937e-06, -5.125354303104e-07, {   4,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  4.966907202292e-06,  6.228354388704e-07, {   0,   0,   1,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.310491842413e-08, -5.546030019006e-06, {   2,  -2,  -1,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.286335828047e-07,  5.449400994966e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   1,   0,   0 } },
 { -2.690271917902e-07,  5.307118561934e-06, {   0,   2,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0 } },
 {  5.041200004124e-07, -5.070989572805e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   3,   0,   0 } },
 {  5.225846437264e-06,  3.403838459598e-07, {   4,   0,  -1,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.839541171028e-08,  5.516794354965e-06, {   2,  -2,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0 } },
 { -5.032208276508e-06,  5.290349664710e-07, {   2,   0,  -4,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.646614932990e-07, -5.395683182461e-06, {   1,  -1,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -7.409087356877e-08,  5.483007793686e-06, {   2,  -2,   2,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 {  5.226953594167e-06, -3.296324961404e-07, {   2,   0,   0,   0,   0,   2,  -4,   0,   1,   0,   0,   0,   0 } },
 {  5.721600218653e-07, -4.977079917612e-06, {   1,  -1,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.289301912238e-08,  5.534021636019e-06, {   2,   0,  -3,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 } },
 { -5.354724188721e-06, -1.911101802397e-07, {   2,   0,  -1,   0,   0,  28, -29,   0,   0,   0,   0,   0,   0 } },
 { -6.224470160329e-10,  5.544566829346e-06, {   1,   0,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 { -5.335112079485e-06, -2.058167058566e-07, {   1,  -2,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.496647758590e-06, -3.381603579808e-08, {   2,   0,  -4,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -5.456623714527e-06, -6.776208316061e-08, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  9.814440699341e-08, -5.422882165761e-06, {   4,   0,   0,   0,   0,  -3,   5,   0,  -2,   0,   0,   0,   0 } },
 { -5.431437273681e-06,  8.718535742111e-08, {   2,   0,   1,   0,   0, -16,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.274420628903e-06,  2.404652150536e-07, {   3,   0,   0,   0,   0,  15, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.269669438980e-06, -2.433038904187e-07, {   2,  -2,   1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  1.666789523638e-07,  5.343399077168e-06, {   1,   0,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -6.328417441217e-09,  5.503626719748e-06, {   3,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.885258656224e-08,  5.476288793426e-06, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,  -2,   0,   0,   0 } },
 { -2.147616415813e-07,  5.287103178417e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 } },
 { -4.654099994214e-07,  5.035467991619e-06, {   3,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.613289087584e-08, -5.464337544829e-06, {   1,   0,   0,   0,   0,   0,   4,  -6,   0,   2,   0,   0,   0 } },
 { -4.285797513668e-08, -5.457316867379e-06, {   2,  -2,  -2,   0,   0,  24, -24,   0,   0,   0,   0,   0,   0 } },
 { -1.467575372186e-07, -5.348635835390e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 } },
 { -6.611414944093e-08,  5.425120221330e-06, {   4,   0,  -3,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0 } },
 {  2.130088748095e-08, -5.463513513873e-06, {   0,   2,   1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  2.848933637074e-07,  5.194839390507e-06, {   1,   0,   2,   0,   0, -21,  20,   0,   0,   0,   0,   0,   0 } },
 { -5.270401286047e-06,  1.998718458197e-07, {   1,  -1,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  5.803973516097e-08, -5.402862488984e-06, {   0,   2,  -1,   0,   0,   0,   4,  -4,   0,   0,   0,   0,   0 } },
 { -5.223477966927e-06, -2.366631279726e-07, {   2,   0,   0,   0,   0,   0,   3,  -6,   3,   0,   0,   0,   0 } },
 {  2.563987055155e-07,  5.201036338875e-06, {   1,   0,   1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  2.510117145053e-07, -5.195362193989e-06, {   3,   0,  -2,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -7.781039266417e-08,  5.368431696864e-06, {   4,   0,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  3.404313029476e-07, -5.104059910520e-06, {   2,   0,  -1,   0,   0,   0,  -4,   0,   0,   4,   0,   0,   0 } },
 { -5.239695291492e-06, -1.963466211500e-07, {   6,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.286254675785e-06,  1.452931902523e-07, {   2,   0,  -2,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 {  7.945881819135e-10, -5.428962070050e-06, {   1,   0,  -1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 {  3.555041473110e-06,  1.874269057060e-06, {   2,  -2,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.291029698099e-07,  5.196443362966e-06, {   2,   0,   0,   0,   0,   4,  -5,   0,   1,   0,   0,   0,   0 } },
 { -2.619021194686e-07, -5.162273104115e-06, {   4,   0,  -1,   0,   0,   0,  -2,   0,   4,  -2,   0,   0,   0 } },
 {  9.682935709680e-08,  5.324902681222e-06, {   4,   0,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 } },
 {  5.257116557453e-08,  5.368639469950e-06, {   4,   0,  -2,   0,   0,   0,   5,   0,  -5,   0,   0,   0,   0 } },
 { -1.131817717788e-07, -5.305714167797e-06, {   4,  -4,   2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 {  1.131817717788e-07,  5.305714167797e-06, {   4,  -4,  -2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0 } },
 { -5.414935172778e-06, -3.928847457134e-09, {   2,   0,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  4.606092972532e-08,  5.372288582243e-06, {   2,  -2,  -1,   0,   0,  25, -25,   0,   0,   0,   0,   0,   0 } },
 {  5.194652286425e-06,  2.219852423894e-07, {   1,   1,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -8.438934782738e-08, -5.328241379460e-06, {   2,   0,  -2,   0,   0,  27, -29,   0,   0,   0,   0,   0,   0 } },
 { -4.632819314180e-06, -7.797038722680e-07, {   2,   0,  -1,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.311887391281e-06,  9.888110453280e-08, {   1,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.011833326082e-08,  5.370184727967e-06, {   4,  -2,   1,   0,   0, -15,  15,   0,   0,   0,   0,   0,   0 } },
 { -8.524392236463e-09, -5.400720793297e-06, {   3,  -1,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.230621339246e-07, -5.085082881460e-06, {   4,   0,  -1,   0,   0,   0,  -3,   0,   0,   3,   0,   0,   0 } },
 { -2.101731123903e-07,  5.195214906595e-06, {   6,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  2.160375272289e-07,  5.184526778635e-06, {   2,   0,  -1,   0,   0,   7,  -9,   0,   2,   0,   0,   0,   0 } },
 { -8.768089601943e-08, -5.311071659162e-06, {   2,   0,   0,   0,   0,  -1,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.560460451778e-08,  5.368636740571e-06, {   1,   0,   0,   0,   0,   3,  -5,   2,   0,   0,   0,   0,   0 } },
 { -4.988961915944e-06, -3.902119917416e-07, {   0,   0,   1,   0,   0,   0, -12,  15,   0,   0,   0,   0,   0 } },
 { -5.364467746683e-06,  1.086300941415e-08, {   6,   0,  -3,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.784638561100e-08, -5.337272567014e-06, {   4,   0,  -1,   0,   0,   0,   2,  -4,   2,   0,   0,   0,   0 } },
 {  5.158478940059e-06, -2.136438676449e-07, {   2,   0,  -2,   0,   0, -17,  18,   0,   0,   0,   0,   0,   0 } },
 { -2.135494388598e-08, -5.340392262400e-06, {   3,   0,   0,   0,   0,  19, -19,   0,   0,   0,   0,   0,   0 } },
 { -1.896289540313e-07, -5.170147601949e-06, {   1,   0,  -1,   0,  -4,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.348191817422e-07, -5.215946837321e-06, {   5,   1,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -1.348191817422e-07,  5.215946837321e-06, {   1,   1,  -2,   0,   0, -32,  28,   0,   0,   0,   0,   0,   0 } },
 { -5.199569156450e-06,  1.476954330220e-07, {   0,   0,   2,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 {  5.166205101276e-06, -1.774300908350e-07, {   2,  -2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -6.504627579425e-09,  5.335023589773e-06, {   5,   0,  -2,   0,   0, -17,  17,   0,   0,   0,   0,   0,   0 } },
 { -4.447576057068e-08, -5.293860135316e-06, {   2,   0,   0,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0 } },
 { -2.089921916967e-07, -5.125180871280e-06, {   0,   2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0 } },
 {  3.622606132111e-09,  5.329577445125e-06, {   1,   2,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -5.018052089114e-06, -3.071770748088e-07, {   0,   0,   1,   0,   0,   3,  -1,  -8,   3,   0,   0,   0,   0 } },
 {  2.892809776901e-08, -5.295289093600e-06, {   2,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.165885136260e-06,  1.564772013333e-07, {   1,   0,   0,   0,   0,  21, -22,   0,   0,   0,   0,   0,   0 } },
 { -5.144975809020e-06,  1.772695909530e-07, {   2,   0,  -2,   0,   0,   0,  -7,   9,   0,   0,   0,   0,   0 } },
 {  4.403390645985e-09, -5.317433050179e-06, {   3,  -2,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.351863418529e-08,  5.286615155748e-06, {   2,   0,   0,   0,   0, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  2.109061695922e-08,  5.295106394319e-06, {   0,   2,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.270581332924e-06,  4.558141407451e-08, {   2,   2,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.191207228706e-06,  1.182022091871e-07, {   2,   0,   1,   0,   0,  16, -15,   0,   0,   0,   0,   0,   0 } },
 {  2.064648131414e-08,  5.281005471928e-06, {   3,   0,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 {  3.466429838690e-09,  5.284379664952e-06, {   3,   0,  -1,   0,   0,  16, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.161004913736e-08,  5.265086921753e-06, {   2,   0,   0,   0,   0,   0,  -3,   5,  -2,   0,   0,   0,   0 } },
 {  8.791308917279e-09, -5.273247240309e-06, {   0,   0,   0,   0,   0,  29, -29,   0,   0,   0,   0,   0,   0 } },
 { -3.401108485172e-09,  5.275637258502e-06, {   2,   2,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  4.084549277987e-06, -1.193377006022e-06, {   2,   0,  -1,   0,   0,  -5,   9,   0,   0,   0,   0,   0,   0 } },
 { -5.006309092744e-06, -2.634315873573e-07, {   6,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -3.787156458478e-06, -1.477962239953e-06, {   4,   0,  -1,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.050203979661e-06, -1.979822474562e-07, {   2,  -2,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  5.595003457703e-09, -5.232540603604e-06, {   4,   0,   0,   0,   0,  17, -17,   0,   0,   0,   0,   0,   0 } },
 { -1.824022271851e-07, -5.054856996093e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 { -2.081505100379e-07, -5.024736200384e-06, {   2,   0,   0,   0,   0,   0,   5,  -8,   1,   5,   0,   0,   0 } },
 {  5.036765055288e-06,  1.909080898680e-07, {   2,  -2,   0,   0,   0,  27, -28,   0,   0,   0,   0,   0,   0 } },
 { -4.305552478690e-08, -5.181558563777e-06, {   1,   0,   0,   0,   0,   0,  -2,   4,  -2,   0,   0,   0,   0 } },
 { -5.067219222194e-06, -1.540679223517e-07, {   2,  -2,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.182561743038e-06, -3.622960513465e-08, {   1,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  5.027330133288e-06, -1.888824941814e-07, {   2,  -2,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  5.085012183934e-06, -1.297566466879e-07, {   1,   0,   0,   0,   0, -12,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.005827619379e-06, -2.078137617436e-07, {   2,   0,   1,   0,   0,   0,  13, -16,   0,   0,   0,   0,   0 } },
 {  5.034713922605e-06, -1.750450440879e-07, {   6,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.993943606718e-10, -5.204847641618e-06, {   0,   2,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0 } },
 {  7.644742073588e-08, -5.113832731779e-06, {   0,   2,   0,   0,   0, -25,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.939569836434e-08,  5.160602705282e-06, {   4,   0,   2,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 { -5.077390089228e-06,  9.587654204764e-08, {   4,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.296943323457e-10,  5.156303814359e-06, {   6,   0,  -2,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.023404931488e-08, -5.109202582967e-06, {   2,   0,  -3,   0,   0,  13, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.884585171661e-06, -2.227500611435e-06, {   4,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -9.545023305644e-10, -5.104474104949e-06, {   2,  -2,   1,   0,   0,   8,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.332477703378e-07, -4.661695042460e-06, {   2,   0,   3,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -5.441186587577e-10, -5.091980268890e-06, {   0,   2,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.660162977289e-09, -5.081251360342e-06, {   0,   2,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.021385436388e-06, -5.776388338296e-08, {   1,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -2.841680015188e-09, -5.066984215869e-06, {   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0 } },
 {  3.419771383635e-09,  5.062313869712e-06, {   1,   0,   1,   0,   0, -10,  10,   0,   0,   0,   0,   0,   0 } },
 { -9.775758484242e-10,  5.062356233335e-06, {   4,   0,   2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -1.097641135717e-08, -5.041501462174e-06, {   2,   0,   2,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0 } },
 { -3.197993640920e-09,  5.029407998261e-06, {   4,   0,  -5,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.640738269527e-08, -5.013833671559e-06, {   4,   0,   0,   0,   0, -12,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.073406169340e-06,  2.950424474808e-06, {   4,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.510118619988e-08,  5.008683835535e-06, {   2,  -2,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0 } },
 { -1.679138426494e-07,  4.712798356980e-06, {   2,   0,  -2,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 { -5.378380789715e-08, -4.789416902640e-06, {   1,   0,   2,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.913435806479e-06,  9.152700327000e-07, {   0,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  1.584515156110e-06,  3.039924058770e-06, {   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 { -2.163675937160e-06,  2.449554462234e-06, {   0,   0,   0,   0,   0,   0,   5,   0,  -4,   0,   0,   0,   0 } },
 { -6.538848895120e-08, -4.543465539022e-06, {   4,   0,   0,   0,   0,  -8,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.385229909716e-06, -2.186862392253e-07, {   0,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 {  8.648667982666e-08,  4.456870567358e-06, {   1,   0,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.656905994280e-06,  8.771148526020e-07, {   1,   1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.458180200265e-07, -4.248220854100e-06, {   2,   0,   1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -4.367867706943e-06, -1.084485852365e-07, {   4,   0,  -3,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.937751341500e-07, -3.679917864640e-06, {   0,   0,   2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  3.732185349590e-06, -6.929432485900e-07, {   1,  -1,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.927580353560e-07,  3.909702375907e-06, {   3,   1,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.307235363290e-07,  4.071426711780e-06, {   0,   0,   1,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.925803646767e-06, -2.393948402171e-06, {   0,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 { -1.631527930386e-07, -4.002562547160e-06, {   4,   0,   0,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.862403354876e-06, -2.281726998277e-06, {   2,  -2,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.651630696248e-06, -4.052727066760e-07, {   4,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  3.281171024920e-06, -7.217883282680e-07, {   3,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.800160263026e-07, -3.513101992700e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0 } },
 {  1.516546686700e-06,  2.444765525070e-06, {   2,   0,  -1,   0,   0,   7, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.023434509060e-06,  2.907922533010e-06, {   0,   0,   1,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.637481091702e-06,  1.264336256672e-06, {   1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.217221428400e-06,  6.447957132800e-07, {   1,  -1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -3.112414944054e-06, -7.226287183554e-07, {   1,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  3.414353140011e-06, -3.780179082900e-07, {   6,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.914326017220e-06, -1.739107428061e-06, {   2,   0,   1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.721779563143e-06, -8.360480611590e-07, {   2,   0,  -1,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.634160939292e-08, -3.527016081098e-06, {   4,   0,  -2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  3.085797477157e-06,  4.291911659450e-07, {   2,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.617223204110e-07,  3.126093209661e-06, {   4,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.944201018550e-07, -3.088048876419e-06, {   3,   1,  -3,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.463192156978e-06,  1.366166837087e-08, {   2,   0,   1,   0,   0, -10,  11,   0,   0,   0,   0,   0,   0 } },
 { -2.065417027092e-06,  1.389698024447e-06, {   1,  -1,   1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 {  2.065417027092e-06, -1.389698024447e-06, {   1,  -1,  -1,   0,   0,   0,  -2,   6,   0,   0,   0,   0,   0 } },
 { -2.396481461508e-06, -1.017782993742e-06, {   1,   1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  2.132820028235e-07, -3.164717941028e-06, {   2,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.093930650510e-06, -1.281509224047e-06, {   1,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.242573964606e-06, -1.198108848600e-07, {   2,   0,   0,   0,   0,  -9,  12,   0,   0,   0,   0,   0,   0 } },
 {  2.048836996711e-07, -3.153642485539e-06, {   3,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -3.085737188920e-09,  3.346032389641e-06, {   2,  -2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  2.915598618822e-06, -4.185715169150e-07, {   2,  -2,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.974853886442e-06, -3.376352731660e-07, {   6,   0,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.964748516922e-06,  3.097867017790e-07, {   0,   0,   3,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  3.910844874195e-08, -3.117135345394e-06, {   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.793882635000e-09, -3.146890467594e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,   0 } },
 {  1.414545521544e-06, -1.707585118290e-06, {   0,   0,   0,   0,   0,   0,   3,   0,   0,  -5,   0,   0,   0 } },
 {  2.148589186665e-06,  9.700709684590e-07, {   3,  -1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 { -4.419513452660e-07,  2.669027627943e-06, {   0,   0,   2,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.329363656612e-06,  6.330770069100e-07, {   0,   0,   1,   0,   0,   0,  -3,   0,   6,  -5,   0,   0,   0 } },
 { -6.084850138510e-07, -2.298690180075e-06, {   1,  -1,   1,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 } },
 { -2.638635116463e-06,  2.170140664640e-07, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 {  1.933282597711e-06, -8.491402093940e-07, {   1,   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 {  1.303073448066e-06,  1.473209033300e-06, {   4,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.616276488739e-06, -1.145395551621e-06, {   2,   0,   1,   0,   0,   6,  -9,   0,   0,   0,   0,   0,   0 } },
 {  1.826157494488e-06,  9.001463834840e-07, {   0,   0,   1,   0,   0,   0,  -3,   0,   2,   5,   0,   0,   0 } },
 {  7.467005403445e-08, -2.612771048630e-06, {   0,   2,  -2,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -6.816456676400e-07,  1.969489193610e-06, {   0,   0,   0,   0,   0,   0,   4,   0,  -2,   0,   0,   0,   0 } },
 { -2.293521975290e-06,  3.007140372820e-07, {   2,   0,  -1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -9.691922597800e-07, -1.451922293827e-06, {   2,  -2,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.364842945582e-06, -4.368383238500e-08, {   0,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0 } },
 {  8.560858416899e-07,  1.532419913992e-06, {   4,   0,  -2,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.949534190020e-06, -3.881277974439e-07, {   0,   0,   1,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0 } },
 {  2.309435787838e-06, -2.442115211960e-08, {   1,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -8.933525602190e-08, -2.075414955050e-06, {   2,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.042203797187e-06, -7.503141238735e-08, {   2,   0,   1,   0,   0,  -9,  10,   0,   0,   0,   0,   0,   0 } },
 { -1.323886473963e-08, -1.959156834661e-06, {   2,   0,   2,   0,   0,  -7,   7,   0,   0,   0,   0,   0,   0 } },
 {  1.118798651887e-06, -8.269615099640e-07, {   2,   0,   1,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0 } },
 {  6.763156367070e-08, -1.876345983641e-06, {   0,   0,   1,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0 } },
 { -1.923743903727e-06,  3.858187128000e-09, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 {  8.671373045040e-07,  1.047852304376e-06, {   2,   0,   0,   0,   0, -10,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.084638148380e-07, -1.700594354938e-06, {   0,   0,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0,   0 } },
 {  1.120236795972e-06,  7.880643109530e-07, {   4,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  4.563453196042e-09,  1.895538129614e-06, {   0,   0,   2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.467227968657e-06, -4.257137882645e-07, {   1,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  6.197526760560e-08,  1.770306371880e-06, {   0,   0,   1,   0,   0,   1,  -2,   0,   1,   0,   0,   0,   0 } },
 { -1.409993662817e-06, -3.317395043614e-07, {   2,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -9.167506725687e-08, -1.630151142290e-06, {   4,   0,  -2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.117000599945e-07,  1.605336414778e-06, {   3,   0,  -1,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.256038686100e-06,  3.322699856450e-07, {   3,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.127092341300e-07, -1.261321432335e-06, {   2,  -2,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -7.897202051145e-07, -6.618772314200e-07, {   2,  -2,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.722642990800e-08,  1.291913400500e-06, {   0,   2,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.667940766000e-09, -1.210773933753e-06, {   2,   0,  -3,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -1.139417683500e-08, -1.166168785510e-06, {   1,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -4.436345716435e-09,  1.133369736870e-06, {   2,   2,  -2,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.334175100200e-07, -1.700132915555e-07, {   2,   0,   1,   0,   0,  -8,   9,   0,   0,   0,   0,   0,   0 } },
 { -9.118662062670e-07,  7.139436323750e-08, {   0,   0,   3,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.430025478699e-07,  8.203345382680e-07, {   2,   0,  -2,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.039123546729e-07,  6.779291469340e-07, {   4,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  5.085214327900e-09,  7.507012734250e-07, {   1,   0,  -3,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -2.491543432253e-07,  5.038994447360e-07, {   1,   0,  -1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.389977293454e-08,  6.485482714170e-07, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   2,   0 } },
 {  2.227599824105e-08,  6.827684419580e-07, {   6,   0,  -3,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.337736392233e-08, -6.780918522260e-07, {   2,   0,   2,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.874222910270e-07, -3.480952430940e-07, {   4,   0,  -1,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.322000833000e-07,  9.253508073800e-08, {   1,   1,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.802886860099e-07, -2.860523511200e-07, {   4,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.941716035790e-07, -1.251558794591e-07, {   1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.574529198700e-08,  3.824357718100e-07, {   0,   2,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.205720077000e-07,  5.682662970000e-08, {   3,   1,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.429217708000e-08,  3.308194465790e-07, {   3,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.863352096483e-07, -3.823167255000e-08, {   4,   0,  -1,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.926251287720e-07,  1.022501375220e-07, {   1,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  1.078232533136e-07, -1.668489192580e-07, {   2,  -2,   1,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.004185253870e-07,  3.970392157500e-08, {   1,  -1,  -3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.667064567460e-07, -1.356754189500e-08, {   1,  -1,   3,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.074216878560e-07, -2.360789787800e-08, {   3,  -1,  -1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  9.239956849000e-08,  4.074291489000e-09, {   3,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 1165> g_PERT_S3_2
{ {
 {  0.000000000000e+00,  5.139500000000e-01, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.824500000000e-01, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.265400000000e-01, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.639600000000e-01, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.230200000000e-01, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -7.754000000000e-02, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.068000000000e-02, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.970000000000e-02, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.194000000000e-02, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.222000000000e-02, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.529000000000e-02, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.490000000000e-02, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.764000000000e-02, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.141767049843e-02, -5.253727461691e-03, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.070508432118e-02,  4.925840569388e-03, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.449000000000e-02, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.299601282617e-04, -1.368414439540e-02, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.186000000000e-02, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.066000000000e-02, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  9.930000000000e-03, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.580000000000e-03, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.995225392553e-05, -6.475331461477e-03, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.870000000000e-03, {   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.339644201502e-04, -5.189361246154e-03, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  6.339110880064e-04,  5.188959684905e-03, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.360000000000e-03, {   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.760000000000e-03, {   4,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.210729340182e-03, -1.477450193267e-03, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.194181627185e-03,  1.469843688501e-03, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.940000000000e-03, {   4,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.310000000000e-03, {   2,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.100000000000e-03, {   2,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.113089822280e-03, -9.724004705988e-04, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.927312643566e-03, -1.073768008444e-03, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  1.986552701046e-03,  9.141736390433e-04, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.350000000000e-03, {   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.350000000000e-03, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.330000000000e-03, {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.140000000000e-03, {   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.130000000000e-03, {   2,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.433863479261e-03,  3.973316826359e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.433219853327e-03, -3.973297789865e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  0.000000000000e+00, -1.730000000000e-03, {   2,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.650000000000e-03, {   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.460000000000e-03, {   4,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.400000000000e-03, {   4,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -9.716351843714e-05, -1.266217782677e-03, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -9.703847430710e-05,  1.265118904126e-03, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.260000000000e-03, {   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.250000000000e-03, {   2,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.140000000000e-03, {   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.120000000000e-03, {   2,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  8.405364157476e-04, -2.546143882041e-04, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  8.341226838982e-04,  2.530698996829e-04, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.060000000000e-03, {   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.242643817137e-04, -2.873792387340e-04, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.334109226803e-04, -6.671933400341e-04, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -2.338783092327e-04,  6.644489516341e-04, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.900000000000e-04, {   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.855467267484e-04,  2.695659075566e-04, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.400000000000e-04, {   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.400000000000e-04, {   0,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.100000000000e-04, {   2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.810747425476e-04,  5.643810864842e-04, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -7.400000000000e-04, {   2,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.840932171688e-04, -5.489842663045e-04, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.691145589459e-04, -5.304885882148e-04, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.900000000000e-04, {   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.514944119513e-04, -1.037842031563e-04, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.508943310845e-04,  1.037334837693e-04, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.314482924176e-04, -4.227197478745e-04, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.400000000000e-04, {   4,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.400000000000e-04, {   3,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.200000000000e-04, {   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.100000000000e-04, {   4,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.000000000000e-04, {   4,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.411496073971e-04,  4.434483121178e-04, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -6.171580327605e-05,  5.222946142280e-04, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.171580327605e-05, -5.222946142280e-04, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.631023046598e-04, -2.016114989836e-04, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  4.347730579526e-04,  1.160706167752e-04, {   2,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.500000000000e-04, {   2,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.400000000000e-04, {   4,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.302244316586e-04,  4.088418149013e-04, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.300000000000e-04, {   3,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.300000000000e-04, {   2,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.300000000000e-04, {   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.896962029731e-05, -4.968380311406e-04, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.022603795015e-04, -2.211844631445e-04, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  2.989293540011e-04,  2.192530029002e-04, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.376166159088e-04,  3.724783893311e-04, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.100000000000e-04, {   4,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.700000000000e-04, {   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.296961805230e-04,  2.380606537859e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.300000000000e-04, {   2,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.602999382695e-04,  1.595857148080e-04, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  3.945309826553e-04, -1.865412310786e-05, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.906866134421e-04,  1.870665969413e-05, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.795296919220e-04, -1.286796464928e-04, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  3.967134460943e-04, -5.537382527109e-06, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-04, {   3,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-04, {   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.936156579041e-04,  5.307413126319e-06, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  2.724695125563e-04,  1.254156301681e-04, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.806081953435e-04, -2.139642020871e-04, {   2,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.010673497308e-04,  2.920054015516e-04, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.900000000000e-04, {   2,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.042983307895e-04,  2.841735624480e-04, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.800000000000e-04, {   4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.804384125294e-04,  1.870713231163e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.253925475785e-04, -1.306713900524e-04, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.241963222421e-04,  1.301196371843e-04, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.011287788235e-04, -1.516006817556e-04, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.492360667079e-04,  1.937292492210e-04, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.155913442049e-04,  1.263350441929e-04, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.400000000000e-04, {   2,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.923738357528e-04, -1.462727788963e-04, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  1.102405990374e-04, -2.255122195051e-04, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.069813574575e-04,  2.204226551169e-04, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -1.684608452838e-04, -1.565916460297e-04, {   2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.560436766056e-04,  1.668087664110e-04, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.070901560767e-05,  2.397254994565e-04, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.808023592808e-04,  1.393881929877e-04, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.200000000000e-04, {   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.796085802829e-04,  1.402825979763e-04, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.996888351644e-04,  1.193508024512e-04, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  3.456544939125e-05, -2.830098632379e-04, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.456252886172e-05,  2.829755606647e-04, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  8.018198471777e-05, -2.369941474954e-04, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.923541551962e-04, -2.296791350235e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -2.895588371658e-04,  2.288767286317e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.094235487911e-05, -2.410298375069e-04, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.362872685173e-04, -7.229057510133e-05, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.361524651185e-04,  7.224928657147e-05, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.423080849082e-04, -6.586525069998e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  6.425175170149e-05, -2.288208626530e-04, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.249058749727e-04,  6.763340746932e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  5.006927981560e-05,  2.400615255218e-04, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.900000000000e-04, {   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.900000000000e-04, {   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.800000000000e-04, {   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.320904803825e-04, -1.414593250445e-04, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.700000000000e-04, {   3,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.772900939561e-04, -8.809455057197e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.568362871761e-04,  1.058907500013e-04, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.600000000000e-04, {   0,   0,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.695155521578e-04, -8.887931892712e-05, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.695158908003e-04, -8.691339819574e-05, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.511954309197e-04, -1.001901209552e-04, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.635529188967e-04,  8.408948383293e-05, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -1.616726004389e-04,  8.533412812856e-05, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  6.025369836118e-05, -1.847248204819e-04, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -6.173455448266e-05,  1.796909702318e-04, {   2,   0,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 { -1.568242577368e-04, -8.136622485446e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.648130929100e-04,  6.957796320837e-05, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  8.103685549475e-05, -1.500888663023e-04, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.300000000000e-04, {   2,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.535946948610e-04,  7.286061837888e-05, {   2,   0,  -1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  7.885317586549e-05, -1.443400598837e-04, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.353997360415e-04, -8.482526767651e-05, {   2,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.200000000000e-04, {   2,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.393425226826e-04,  5.456273846231e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  0.000000000000e+00, -1.900000000000e-04, {   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.900000000000e-04, {   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.900000000000e-04, {   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00,  1.900000000000e-04, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  1.438043288631e-04, -4.327980295895e-05, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.426549248642e-04,  4.293490424275e-05, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.415418448009e-05,  1.213135179219e-04, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.008963797196e-04,  8.333900027331e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  6.185563706088e-05,  1.206381412362e-04, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.159434747907e-04, -6.456173358670e-05, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.741648558090e-05, -1.386396379248e-04, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -7.712267415521e-05, -9.764057561403e-05, {   0,   0,   2,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -7.665964229111e-05,  9.704853268987e-05, {   0,   0,   0,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.108206656972e-04,  6.174690984459e-05, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.170489874122e-04,  5.385719781831e-05, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.700000000000e-04, {   6,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.826386829963e-05,  1.296050716129e-04, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.231840099182e-04, -4.347446286552e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.600000000000e-04, {   2,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.600000000000e-04, {   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.028264165351e-04, -5.459921216087e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  7.421939199821e-06, -1.498162702063e-04, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.335802378328e-05,  1.230528075362e-04, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -9.483867341678e-05,  5.934320931696e-05, {   2,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  8.600344002892e-05,  6.458383773559e-05, {   2,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.500000000000e-04, {   2,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.500000000000e-04, {   1,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.500000000000e-04, {   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.509276458353e-05, -1.036104373370e-04, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.264691637933e-05, -1.138011639809e-04, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  9.989076516813e-05,  4.606338061980e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.581668230541e-05, -8.766239160462e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  1.900702319858e-05,  1.215971344811e-04, {   1,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.400000000000e-04, {   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.400000000000e-04, {   2,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.400000000000e-04, {   2,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.400000000000e-04, {   0,   0,   1,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.752844497033e-05,  1.023899693835e-04, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.846432990789e-05, -8.115263078900e-05, {   2,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.994922828159e-05, -9.775114938198e-05, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  8.761471286940e-05,  4.820437831581e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.159595996348e-05, -9.028366285676e-05, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.300000000000e-04, {   6,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.300000000000e-04, {   4,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.300000000000e-04, {   3,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.300000000000e-04, {   0,   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.170512534076e-05, -9.641167576461e-05, {   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.924885961740e-05, -5.748561073599e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.512668475809e-05,  7.144926323388e-05, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  8.685245549915e-05,  3.889860643918e-05, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  7.943946536811e-05,  4.230096147877e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.943946536811e-05, -4.230096147877e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.335236403881e-06,  1.131108833554e-04, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.200000000000e-04, {   4,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.200000000000e-04, {   4,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.200000000000e-04, {   2,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.005087764178e-05,  1.095398551151e-04, {   2,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  7.671589892038e-05,  4.251782115294e-05, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  6.828656725622e-05,  4.970494163804e-05, {   2,   0,  -1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  4.529080604465e-05,  7.164629608760e-05, {   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -2.934399918786e-05, -8.746114949481e-05, {   1,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.849146949334e-05, -5.795789536423e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.162647644624e-07,  1.155023997817e-04, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -4.433212863367e-07, -1.154636272037e-04, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -4.101337305441e-05, -7.393348870296e-05, {   0,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  7.010238455080e-05, -4.472156521569e-05, {   1,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  1.957956827880e-05,  9.394202015665e-05, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.711277967856e-05, -5.601901817587e-05, {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -1.402428515175e-06, -1.103243951817e-04, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 {  5.424615120372e-05, -5.727439394779e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.100000000000e-04, {   4,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.425023559194e-05, -7.565101691828e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -6.989706932460e-05,  3.853702465297e-05, {   0,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.541587002353e-05,  7.131275529782e-05, {   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.577340763447e-05,  6.053333601006e-05, {   0,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.971472829782e-05,  7.584608124467e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  5.071538911062e-05, -5.301311831997e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  2.462582950278e-05,  7.787832456842e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  6.229316172103e-05,  4.008443053416e-05, {   2,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  2.678530358056e-05,  7.449940405228e-05, {   2,   0,  -1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.675788474095e-05, -7.442407259462e-05, {   2,   0,  -1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.223201849964e-05, -7.809164374808e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  0.000000000000e+00, -1.000000000000e-04, {   3,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.000000000000e-04, {   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.820491834124e-05,  2.165114807400e-05, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  7.808262897089e-05, -2.169351812918e-05, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.838272735747e-05,  3.882964174465e-05, {   0,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  3.993843268232e-05,  5.560926073609e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  3.969136520780e-05, -5.551098500520e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.233792560370e-05,  6.119061843192e-05, {   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.496561249786e-05, -6.818745845991e-05, {   2,   0,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  5.631522684493e-05,  3.623872234842e-05, {   2,   0,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  1.596647473512e-05,  7.655575420471e-05, {   0,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -5.942269671870e-05, -3.180496232633e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -3.986242967703e-05,  5.045663435142e-05, {   2,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -9.000000000000e-05, {   4,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  9.000000000000e-05, {   4,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -9.000000000000e-05, {   4,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -9.000000000000e-05, {   2,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -9.000000000000e-05, {   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.495272890579e-05, -3.467778068435e-05, {   0,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  3.699062952974e-05, -5.193347415696e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 {  3.856446099855e-05,  4.990521077306e-05, {   0,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -3.813050244713e-05,  5.011010865819e-05, {   2,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.904995434659e-05,  5.918424151382e-05, {   2,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -6.027623378090e-06, -8.193305671317e-05, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  2.110249678542e-05, -6.674342386649e-05, {   2,   0,  -1,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  3.612823216664e-05,  5.054815354765e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 { -5.921431486395e-05,  2.725868007085e-05, {   1,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.980415809442e-05,  5.637309458010e-05, {   2,   0,  -1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -5.896664625810e-05, -2.714481045399e-05, {   1,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  3.883354762448e-05, -4.698183753534e-05, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  3.266416827403e-05,  5.205247609644e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 {  5.782916668587e-05, -2.662074035284e-05, {   4,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  5.643313864499e-05,  2.597840162959e-05, {   4,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.987135293408e-05, -5.239474293027e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  5.621543672533e-05, -2.588986300496e-05, {   2,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -6.732886592913e-05, -1.475222193925e-05, {   2,   0,   0,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  4.979036453132e-05,  3.187782924080e-05, {   0,   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0 } },
 {  3.119353718172e-05, -4.902908287920e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   2,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   2,   0,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   2,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   1,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -8.000000000000e-05, {   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.088704842325e-05,  5.806543956616e-05, {   2,   0,   0,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  2.088695827851e-05, -5.804864491178e-05, {   2,   0,   0,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  2.675114656054e-05,  5.186895161320e-05, {   0,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  2.242759699777e-05, -5.565072230354e-05, {   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.928459916163e-05,  5.856299196073e-05, {   2,  -2,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.894416707912e-05, -5.753548667557e-05, {   2,  -2,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.604492621189e-05, -2.949857408044e-05, {   0,   0,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -4.579094545657e-06, -7.085585575189e-05, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  2.505101538286e-05, -4.871707787479e-05, {   0,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  3.538855449374e-05, -3.780360170539e-05, {   1,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.813827975269e-06,  6.835744153282e-05, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 {  1.975384009027e-05, -5.171509508462e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 {  2.945184692238e-05, -4.168099993154e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -3,   0,   0,   0 } },
 {  2.962909879700e-05, -4.093341253694e-05, {   2,   0,  -1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.342732714800e-05,  4.706755854488e-05, {   2,   0,  -1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 { -3.208063453573e-05, -3.832590826295e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -1.749080793214e-05, -5.277098647137e-05, {   2,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.214249626397e-05,  5.807282621164e-05, {   0,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -3.014504984083e-05, -3.989080056973e-05, {   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  7.000000000000e-05, {   4,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -7.000000000000e-05, {   3,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  7.000000000000e-05, {   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.530202468423e-05, -5.440880979120e-05, {   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.630539793617e-05,  4.238575459210e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -2.565301451109e-05,  4.291762862151e-05, {   0,   0,   2,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 { -2.565301451109e-05,  4.291762862151e-05, {   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.203130709445e-05,  2.633458975647e-05, {   0,   0,   1,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 { -4.259809433593e-05,  2.346304846443e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.470139570283e-05,  2.102406653598e-05, {   2,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.386183662638e-05,  4.176000787523e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,   0,   0,   0,   0 } },
 { -4.447440780101e-05,  2.047184413077e-05, {   0,   2,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.816314692108e-05,  4.658433313812e-05, {   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.415234833751e-05,  2.003472367994e-05, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -2.378077131622e-05,  3.997473379900e-05, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -2.357686658458e-05, -3.968124196244e-05, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  2.118142468462e-05, -4.014788180523e-05, {   2,   0,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.194653865416e-05, -1.930934051335e-05, {   0,   2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.997785780500e-05,  4.094084936274e-05, {   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.324070655324e-05, -4.707227789724e-05, {   2,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.000000000000e-05, {   6,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.000000000000e-05, {   4,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.000000000000e-05, {   4,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  6.000000000000e-05, {   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.000000000000e-05, {   2,   0,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -6.000000000000e-05, {   0,   0,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.540569265193e-05, -1.366802665786e-05, {   0,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 { -3.807905230954e-05,  2.097292248921e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.790838996775e-05, -2.113217346221e-05, {   4,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.505119048498e-05,  1.356165209993e-05, {   0,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.038005797251e-05, -3.693433230512e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  2.884734047088e-05, -2.770976267955e-05, {   0,   0,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  2.979930659760e-05,  2.668335298088e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -3.770241069524e-05, -1.854755199667e-05, {   0,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -2.467223725939e-05,  3.148461066325e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.891026614620e-05,  3.694276490069e-05, {   0,   0,   1,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  3.813168914268e-05, -1.755387074654e-05, {   0,   0,   4,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.914574139336e-05, -3.597599869900e-05, {   0,   0,   1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 { -9.913499383391e-06, -4.511555686712e-05, {   2,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.295655438116e-05, -2.136107726122e-05, {   2,   0,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.491831102625e-05,  1.910440841707e-05, {   0,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 { -1.190531527876e-06,  5.205087309345e-05, {   2,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -1.963940571408e-05,  3.346309859208e-05, {   2,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -1.942375357265e-05, -3.333896681210e-05, {   2,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -2.002868037757e-05,  3.262740928741e-05, {   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 {  6.251187580520e-06,  4.636917412411e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -2.025520941473e-05,  3.235572266796e-05, {   0,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.456554054822e-05,  3.794812122917e-05, {   0,   0,   1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 {  3.823728630671e-05, -1.424815668779e-05, {   2,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -3.387729086233e-05, -1.845559535651e-05, {   0,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  3.576926245022e-05,  1.646858519688e-05, {   0,   0,   2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  2.261977102537e-05,  2.933455112447e-05, {   2,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -5.047918859077e-05, -5.961162544077e-07, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -1.831604902938e-05, -3.264761389931e-05, {   0,   0,   1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -3.780818821305e-05,  1.305913106015e-05, {   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.871462496937e-05,  3.204270592563e-05, {   2,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  1.510651589671e-05, -3.540298779502e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -1.948703104868e-05, -3.098027051489e-05, {   0,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  3.803900173886e-05,  1.237070518244e-05, {   4,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.241066920473e-05, -3.791705125542e-05, {   2,   0,  -1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.955714456833e-05,  6.543745186451e-07, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.249886995469e-05,  3.761044369292e-05, {   2,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.079721003382e-06,  4.894556325749e-05, {   2,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   4,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   4,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   3,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   2,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,   0,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   2,  -2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   2,  -2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -5.000000000000e-05, {   2,  -2,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -2.249686827349e-05,  2.721283396359e-05, {   2,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.693059747236e-05, -3.256900721044e-05, {   2,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -2.027896416372e-06,  4.699308808811e-05, {   0,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -2.163725214375e-05, -2.737835316928e-05, {   2,   0,   1,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -2.163725214375e-05,  2.737835316928e-05, {   2,   0,  -1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 { -1.022667890732e-05,  3.867059656285e-05, {   3,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.271466321976e-05,  1.607971298403e-05, {   0,   0,   0,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -3.598733268573e-05, -1.245946437241e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 {  2.930426501988e-05, -1.900772449869e-05, {   2,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -2.138846077453e-05,  2.625927391155e-05, {   2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.204190521963e-05,  3.553536868989e-05, {   1,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.224884218354e-05,  1.524002230752e-05, {   2,   0,  -1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.417012985255e-08, -4.650609324910e-05, {   2,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -1.998349612660e-06, -4.455967680632e-05, {   0,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.802950765983e-05,  2.727935147137e-05, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  3.096757904115e-05, -1.405355120897e-05, {   1,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.611184603521e-05,  2.883905113287e-05, {   0,   0,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.810866524515e-05,  2.679600076634e-05, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  5.631240843305e-06,  3.900262296634e-05, {   2,   0,   0,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -2.804517038055e-05,  1.652458707313e-05, {   2,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  3.021319610469e-07, -4.413752227275e-05, {   2,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  3.027796114724e-05, -1.414754832676e-05, {   4,   0,  -1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.856913822842e-05, -2.549213291850e-05, {   2,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  1.773499476042e-05,  2.602928792313e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 } },
 {  1.771417279873e-05, -2.603993570609e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 } },
 { -7.505903445497e-06,  3.598754687302e-05, {   2,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -2.721462854790e-05,  1.617617646484e-05, {   2,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  2.950633263787e-05,  1.378644082146e-05, {   0,   0,   1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.965854305563e-06,  3.305186747777e-05, {   0,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.706374974063e-05, -2.539675101734e-05, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  2.582263808500e-06, -3.983259724855e-05, {   2,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  2.298633720846e-05,  1.927766328523e-05, {   2,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.665443621714e-05, -2.553999586996e-05, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.722355045608e-05,  2.485744000172e-05, {   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  2.529301566663e-05, -1.677624686124e-05, {   2,   0,  -1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 { -3.102402658712e-05, -1.082166594526e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -2.455843817940e-05, -1.723029640455e-05, {   2,   0,  -1,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0 } },
 {  9.854219703483e-06,  3.071948581987e-05, {   0,   0,   2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  3.789371633592e-05, -2.225755036764e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -1.006111150054e-05,  3.002685636930e-05, {   1,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.786866431932e-05,  2.215397007395e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   6,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   4,   0,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   4,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   3,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   2,   0,  -1,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,   0,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,  -2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,  -2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   1,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   1,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  4.000000000000e-05, {   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  2.304048631233e-05,  1.692635359765e-05, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.002252998381e-05, -2.988883565580e-05, {   1,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.297662898266e-05, -1.662922889297e-05, {   2,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -3.861466431944e-05,  9.571801129457e-07, {   2,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -3.807531405635e-05,  1.178132256536e-06, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -2.974647884911e-05, -9.486312014109e-06, {   0,   0,   1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -2.972743030780e-05,  9.481901959206e-06, {   0,   0,   1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 {  1.121379849888e-05, -2.782536115177e-05, {   0,   0,   1,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.121379849888e-05,  2.782536115177e-05, {   0,   0,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.712185979050e-05,  2.175150585222e-05, {   2,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -1.961407145391e-05,  1.888952384420e-05, {   2,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -2.442315841009e-05,  1.406313192977e-05, {   2,   0,  -1,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  1.346404865304e-05, -2.495187997657e-05, {   2,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.018307733403e-05, -8.136712157778e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  9.240591674569e-06, -2.896442140993e-05, {   0,   0,   2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -2.835816111003e-05,  9.788498273866e-06, {   2,   0,  -3,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.559579961442e-05, -2.239178659357e-05, {   2,   0,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 { -1.403583733399e-05, -2.381644345721e-05, {   2,   0,  -1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  9.049680243221e-06,  2.850228406538e-05, {   2,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.600236239276e-06, -2.874084091721e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.546261581390e-05, -1.172146379783e-05, {   4,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.517262624445e-05,  1.158803227369e-05, {   4,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  7.670009180489e-06, -2.900294742214e-05, {   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -7.670009180489e-06, -2.900294742214e-05, {   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.670009180489e-06,  2.900294742214e-05, {   1,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.670009180489e-06, -2.900294742214e-05, {   1,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.624481146284e-05,  1.031859926554e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 {  2.796488447990e-05,  8.406378225776e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 {  2.479059457801e-05, -1.141164649669e-05, {   4,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.268518560747e-05,  2.349513772191e-05, {   2,   0,  -1,   0,   0,   0,   7,  -9,   0,   0,   0,   0,   0 } },
 {  2.692589620863e-05, -9.106161469830e-06, {   2,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  2.923883754712e-05, -6.714936998457e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.923883754712e-05,  6.714936998456e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 } },
 { -1.728224941076e-05,  1.859766026382e-05, {   2,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.344634212867e-05, -2.231808750578e-05, {   2,   0,   0,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  1.130373482610e-05, -2.437610527934e-05, {   2,   0,  -1,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  2.105889259171e-05, -1.410925039870e-05, {   0,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  2.962503393839e-05, -5.534094783078e-06, {   0,   0,   2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  2.960162369721e-05,  5.528768684324e-06, {   0,   0,   2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.272712712229e-05, -2.238671196437e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 } },
 { -2.209355721261e-05, -1.301402487515e-05, {   2,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.407942286685e-06, -2.858379239916e-05, {   2,   0,  -1,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0 } },
 { -3.364384614045e-05, -9.218506871640e-07, {   2,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 { -1.448203383256e-06,  3.307174059781e-05, {   2,   0,   0,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -1.208612247219e-05,  2.242153490324e-05, {   2,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -8.408588920073e-07, -3.352449502781e-05, {   2,   0,  -1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 { -1.307258563308e-06,  3.303861560577e-05, {   2,   0,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.873361194655e-05,  1.550727345204e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  2.332156821459e-05,  1.073568596852e-05, {   4,   0,  -3,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -6.399936538695e-07,  3.340667630508e-05, {   2,   0,  -1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.132831580384e-05, -1.225932273689e-05, {   2,   0,   0,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -6.571284783214e-07,  3.275151952797e-05, {   2,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -3.280029489218e-05, -5.473407050319e-07, {   2,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  2.281747871783e-05, -1.050390609694e-05, {   2,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.394891131169e-05,  9.325122489011e-06, {   1,   0,   1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 {  2.394891131169e-05, -9.325122489011e-06, {   1,   0,  -1,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -1.153482727480e-05, -2.156451729196e-05, {   2,   0,  -1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0 } },
 { -9.686202775654e-06, -2.335099484785e-05, {   1,   1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  1.029579341084e-06, -3.199374758316e-05, {   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.443038342363e-05,  1.838501768832e-05, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -1.499324549150e-05,  1.769255384952e-05, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 {  1.561127942105e-05,  1.696474589972e-05, {   2,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.967686586382e-06, -2.439972991496e-05, {   2,   0,  -1,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  2.203341089220e-05,  1.014185460148e-05, {   2,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -2.241206022647e-05, -9.658639424782e-06, {   2,   0,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.008499488853e-06, -3.106092575640e-05, {   2,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  8.677411629077e-06,  2.304922117542e-05, {   2,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.927076724434e-05, -1.226521394859e-05, {   2,   0,  -1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.611144440488e-05, -5.404110167015e-06, {   0,   0,   1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 { -6.569350880442e-06, -2.493119150629e-05, {   2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 { -2.148161085351e-05,  9.895418292972e-06, {   2,   0,  -2,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.831030165668e-06,  2.544358303382e-05, {   2,   0,  -1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -1.316560021032e-06,  2.975566311062e-05, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.379727224325e-05,  1.716856696753e-05, {   2,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -2.391461615916e-05,  6.973887455648e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  1.933156998282e-05, -1.153954924766e-05, {   2,   0,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0 } },
 {  7.715446491471e-06,  2.305569289946e-05, {   2,   0,   0,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 {  1.062931776965e-05, -2.007780920811e-05, {   2,   0,  -1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -1.351229384581e-05, -1.708898880853e-05, {   2,   0,   0,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -1.117479130195e-05,  1.937923464711e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   1,   5,   0,   0,   0 } },
 { -1.117479130195e-05, -1.937923464711e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -1,  -5,   0,   0,   0 } },
 { -1.343043700931e-05,  1.698432482188e-05, {   2,   0,  -2,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  2.259541000957e-05, -7.818588078173e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -7.359880319994e-06, -2.300382592415e-05, {   2,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   6,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   6,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   1,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   1,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   1,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -7.307194205829e-06, -2.263833738750e-05, {   0,   0,   1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.455797870621e-05,  1.510448679692e-05, {   2,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.275639689033e-05, -6.692023514959e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -1.084942325079e-05, -1.855503436990e-05, {   0,   0,   1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  2.588254618823e-05,  3.357935694512e-06, {   0,   0,   1,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.046837603580e-05,  8.639517850445e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  2.718629003587e-05, -1.910037829863e-06, {   2,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  1.471843372401e-05,  1.424278535051e-05, {   2,   0,   0,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -4.507868674344e-06, -2.442667935511e-05, {   2,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.691929395142e-05,  1.915933053979e-06, {   2,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.682118014502e-05,  1.915884072695e-06, {   0,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  1.201771562126e-05, -1.669934210447e-05, {   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 { -3.283831866272e-06, -2.507957806601e-05, {   2,   0,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.816358709363e-05, -1.012151170182e-05, {   4,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.002729269283e-06,  2.692303087447e-05, {   2,   0,  -1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 {  1.018917716429e-05,  1.764249650812e-05, {   2,   0,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 { -5.834148611419e-06, -2.187627859195e-05, {   2,   0,  -1,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } },
 {  1.165302205777e-05,  1.603691048444e-05, {   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 { -1.676146207087e-05,  1.091115893235e-05, {   2,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  6.928353242692e-06,  2.066514149829e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -2.388282823197e-05,  3.537269395642e-06, {   0,   0,   1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  4.719080186780e-06,  2.262579731253e-05, {   2,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.756449633594e-05,  9.564960452853e-06, {   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.820259205587e-05, -8.524414397458e-06, {   4,   0,  -2,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -2.304696924602e-05, -3.556453990175e-06, {   0,   0,   1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.460232635956e-05,  1.192795150645e-05, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -4,   0,   0,   0 } },
 {  2.092958947370e-05,  5.529571811914e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -1.676895775391e-05, -9.548765249945e-06, {   0,   0,   1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 {  2.019052650457e-05, -6.077423139168e-06, {   2,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.833086774232e-06,  2.442717666753e-05, {   2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.012934119403e-05,  6.059063419629e-06, {   2,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.121926784103e-05,  1.444273187689e-05, {   2,   0,  -1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 { -1.573366632222e-05,  9.907102145243e-06, {   2,   0,  -1,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0 } },
 { -1.958229073587e-05, -5.908987296916e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0 } },
 { -6.965321426198e-07,  2.479229662259e-05, {   0,   0,   1,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0 } },
 { -1.022567061056e-05, -1.516482795843e-05, {   2,   0,  -1,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0 } },
 {  3.271334729026e-07,  2.497597130250e-05, {   2,   0,  -1,   0,   0,   0,   8, -11,   0,   0,   0,   0,   0 } },
 {  1.617513713745e-05, -9.069298023939e-06, {   4,   0,  -3,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -1.869729296272e-05, -6.515137167854e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,   0,   0,   0,   0 } },
 { -1.114886911020e-05,  1.405102222162e-05, {   2,   0,  -1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 { -2.042680455371e-06, -2.307481233829e-05, {   2,   0,   0,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 { -6.700495742424e-06, -1.838329651338e-05, {   2,   0,  -2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.277222868191e-06, -1.872945270278e-05, {   3,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.775387638562e-06,  2.319910002634e-05, {   1,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.377925779565e-05, -1.117950142123e-05, {   2,   0,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -9.896496249623e-06, -1.503860563201e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.771639831510e-06, -2.314906985744e-05, {   1,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.117854870192e-05, -1.373671015365e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -5.763734836784e-06,  1.915148455685e-05, {   2,   0,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.600298304528e-06, -1.726651077699e-05, {   2,   0,   0,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 {  9.328076862934e-06,  1.539562549581e-05, {   0,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 { -9.811587813749e-06, -1.490266905652e-05, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  9.807826255319e-06,  1.488467663260e-05, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -9.819832916108e-06, -1.474152794357e-05, {   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -3.337397184152e-06, -2.119062816089e-05, {   4,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.038803915165e-05, -1.411885109029e-05, {   2,   0,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 { -3.339457367685e-06,  2.115274903591e-05, {   4,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  1.402351477464e-05, -1.038910169752e-05, {   0,   0,   2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.888845356792e-05,  5.428255761757e-06, {   2,   0,  -1,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  3.028203470318e-06, -2.116060560517e-05, {   0,   2,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.386500181608e-05,  1.029503661682e-05, {   0,   0,   2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  3.020179988236e-06,  2.110844344306e-05, {   0,   2,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.088410296135e-05, -1.318181277668e-05, {   0,   0,   1,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.272942482560e-05,  1.132765791335e-05, {   2,   0,   0,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.991587122305e-05, -4.100133734497e-06, {   0,   0,   1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 { -1.511208723070e-05, -8.793231983104e-06, {   2,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -9.601863346154e-06,  1.425835436631e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -4.300628177348e-06, -1.953214266997e-05, {   1,  -1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.300628177348e-06,  1.953214266997e-05, {   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.006136985387e-05,  1.371752397884e-05, {   2,   0,   0,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  5.455508045091e-06, -1.820246542456e-05, {   2,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.952406365190e-06,  1.357315499891e-05, {   0,   0,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 { -3.333617481353e-06,  2.013702008933e-05, {   2,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 { -1.061422692742e-05,  1.285526274733e-05, {   4,   0,  -1,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.149725381373e-05, -1.193478566782e-05, {   2,   0,  -2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.588493524071e-06,  1.783010987693e-05, {   2,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -4.520111076603e-06, -1.887011619226e-05, {   2,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  1.735211970312e-05, -5.984361163120e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 {  1.329085708195e-05, -1.001604201330e-05, {   2,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.439060306815e-05,  8.684730600482e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -5.677436783201e-06,  1.737197894004e-05, {   2,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.889058520165e-06, -1.713151521858e-05, {   2,   0,  -1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -2.099465225906e-05,  1.977958553866e-06, {   1,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  2.099465225906e-05, -1.977958553866e-06, {   1,  -1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 { -1.031392881402e-05,  1.259503316309e-05, {   2,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -8.296552951973e-06,  1.455194831332e-05, {   2,   0,   1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 {  9.959155439357e-06, -1.279750830175e-05, {   0,   0,   1,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  8.790744037108e-06, -1.395065664443e-05, {   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  5.620877889429e-06, -1.709140122226e-05, {   0,   0,   1,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  1.382406901971e-05, -8.703490722824e-06, {   2,   0,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  8.648010206229e-06, -1.371771128246e-05, {   2,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  5.479527942333e-06, -1.669764240194e-05, {   0,   0,   0,   0,   0,  20, -21,   0,   0,   0,   0,   0,   0 } },
 {  5.428270209471e-06, -1.668704219745e-05, {   2,   0,  -1,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 { -2.151432769147e-06, -1.988394671688e-05, {   1,   1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.984183998387e-07, -2.138566686518e-05, {   0,   0,   1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  5.405589607337e-06,  1.642492726943e-05, {   4,   0,  -1,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -3.730983543983e-06,  1.804951167736e-05, {   3,  -1,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.732286050749e-06,  1.992483839429e-05, {   2,   0,  -3,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0 } },
 { -9.737672279484e-06, -1.180710021159e-05, {   0,   0,   0,   0,   0,  12,  -8,   0,   0,   0,   0,   0,   0 } },
 {  1.218081871767e-05,  9.354940584396e-06, {   2,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -8.050768682028e-06,  1.343240157764e-05, {   2,   0,   0,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -1.535904960404e-06, -1.994093768990e-05, {   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.535904960404e-06,  1.994093768990e-05, {   2,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -1.781933543808e-05, -3.654008717245e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 { -1.468679901074e-05,  6.755280773883e-06, {   2,   0,  -1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.343476601889e-05, -7.941110741655e-06, {   2,   0,  -1,   0,   0,   0,   9, -14,   0,   0,   0,   0,   0 } },
 {  5.281348527709e-06, -1.582766878473e-05, {   2,   0,  -2,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.447532947236e-05, -6.576984469198e-06, {   2,   0,  -1,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0 } },
 { -1.026487601986e-05,  1.076599052681e-05, {   2,   0,   0,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 { -1.019700420372e-05, -1.079664086637e-05, {   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -2.006555564747e-05, -8.257860518098e-07, {   2,   0,  -2,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  8.911118937027e-06,  1.192021765471e-05, {   0,   0,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -5.110729264986e-06,  1.568963866143e-05, {   2,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 { -3.165310234844e-07,  2.038255047859e-05, {   2,   0,  -1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.983119062788e-05,  8.392127354015e-07, {   2,   0,  -2,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.442863280717e-05,  6.062751738238e-06, {   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.223988905569e-06, -1.325247779277e-05, {   2,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  1.608693079050e-05,  4.256814748303e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -4.351675309056e-06,  1.586085787422e-05, {   2,   0,  -1,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0 } },
 {  6.869808469836e-06, -1.324185155399e-05, {   0,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  4.822790076840e-06,  1.523297648377e-05, {   2,   0,  -1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.252382516650e-05,  7.476712496011e-06, {   2,   0,  -1,   0,   0,   0,  10, -15,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   6,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   6,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   4,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   3,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   3,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   1,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   1,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   0,   2,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  2.000000000000e-05, {   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.497970650672e-06, -1.249315012876e-05, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 { -1.259164774856e-05, -7.287935530228e-06, {   0,   0,   2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 {  1.488804846794e-05, -4.937433253057e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,  -1,   5,   0,   0,   0 } },
 { -1.252920974163e-05,  7.247014841631e-06, {   0,   0,   2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  3.469590451102e-07, -1.936240961413e-05, {   0,   0,   1,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 { -1.880424285332e-05, -8.078276855126e-07, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 { -1.878980148808e-05,  8.083853402781e-07, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -1.158651438148e-05,  7.943989848760e-06, {   2,   0,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.113507657780e-06, -1.725643157747e-05, {   0,   0,   3,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  2.112081270024e-06,  1.725019709668e-05, {   0,   0,   3,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.222844105540e-05, -6.996144647138e-06, {   2,   0,  -1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 { -7.967402598861e-06, -1.123943614294e-05, {   2,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 {  7.161840969918e-06, -1.201818503393e-05, {   2,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  6.126598802727e-06,  1.294381903378e-05, {   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  1.021744265568e-05,  8.825149263194e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 { -1.766937027731e-05,  1.336658517571e-06, {   2,   0,  -1,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.766398999534e-05,  1.267641169593e-06, {   2,   0,   0,   0,   0,  -5,   8,   0,   0,   0,   0,   0,   0 } },
 {  1.263940535190e-06,  1.758879803013e-05, {   2,   0,  -1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  2.639937105668e-06,  1.617691193412e-05, {   2,   0,  -1,   0,   0,   7,  -9,   0,   0,   0,   0,   0,   0 } },
 { -6.458789495422e-06, -1.233000357473e-05, {   2,   0,   0,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 {  1.328399627750e-05,  5.396170415985e-06, {   2,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.243638269078e-05,  6.239692344842e-06, {   2,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.846918674771e-05,  4.844183099553e-08, {   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0 } },
 {  9.283946229129e-06,  9.172093809720e-06, {   2,   0,  -1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 {  1.209943191380e-05, -6.240468374339e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 { -8.077385271424e-06,  1.025508401719e-05, {   2,   0,  -1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.162217689441e-05, -6.681206080109e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -1.183823967490e-05, -6.418035470909e-06, {   2,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.435551789891e-06,  1.367249101553e-05, {   2,   0,   0,   0,   0,   9, -10,   0,   0,   0,   0,   0,   0 } },
 {  8.839536371922e-06,  9.249222769969e-06, {   0,   0,   2,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 {  4.704319519578e-06,  1.335482311120e-05, {   0,   0,   2,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -6.598503077128e-06, -1.128486119266e-05, {   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  1.410004742115e-05,  3.769509735887e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   3,  -5,   0,   0,   0 } },
 { -6.667052830100e-06,  1.116507541324e-05, {   2,   0,  -1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 {  1.114294056548e-06,  1.671377693264e-05, {   2,   0,  -1,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -2.421741003938e-06, -1.538240405652e-05, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 {  2.421741003938e-06,  1.538240405652e-05, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 } },
 { -1.161555775781e-05, -6.183485110320e-06, {   2,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -1.375909414267e-05,  3.999250225002e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -1.088438477521e-05, -6.816826041598e-06, {   1,   0,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 {  9.905484105124e-06, -7.749876004466e-06, {   2,   0,  -1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -9.913630304618e-06,  7.588395947542e-06, {   2,   0,  -2,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.241044671135e-05, -5.051143709268e-06, {   0,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  1.461335503112e-05, -2.824107387286e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 {  5.472044533342e-06, -1.196305497992e-05, {   2,   0,  -1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -5.948350388635e-06, -1.146129748097e-05, {   2,   0,  -1,   0,   0,   0,   4,  -8,   0,   0,   0,   0,   0 } },
 { -1.547495163922e-05,  1.910588956439e-06, {   2,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.117773647227e-05,  5.991776728211e-06, {   2,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  7.409269807766e-06,  9.738234846893e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -2.930884653543e-06,  1.405277751897e-05, {   2,   0,  -2,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.173248124421e-06, -1.579830619419e-05, {   2,   0,  -1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  9.135383085530e-06,  7.802437840578e-06, {   1,   1,  -3,   0,   0,   0,  -3,   7,   0,   0,   0,   0,   0 } },
 { -1.306376748156e-05, -3.855614016894e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -6.324007791834e-06,  1.056403340425e-05, {   0,   0,   1,   0,   0,   0,  -6,  10,   0,   0,   0,   0,   0 } },
 { -6.378880816638e-06, -1.041922643565e-05, {   0,   0,   1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -6.373160659703e-06,  1.041104521236e-05, {   0,   0,   1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 {  1.574022074953e-05,  9.743531576540e-07, {   0,   0,   1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.437469763356e-06,  1.125833182462e-05, {   0,   0,   1,   0,   0,   8, -14,   0,   0,   0,   0,   0,   0 } },
 { -7.595125291563e-06, -9.088485309367e-06, {   2,   0,  -1,   0,   0,  18, -18,   0,   0,   0,   0,   0,   0 } },
 {  9.958139089538e-06, -6.691764490976e-06, {   0,   0,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  9.301282973650e-06,  7.322853724976e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  5.447654136318e-06, -1.114845884148e-05, {   0,   0,   2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.373259777002e-05, -2.838524135630e-06, {   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.484208670206e-05, -1.687474012929e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -3,   0,   0,   0,   0 } },
 {  4.150488638665e-06,  1.236172190723e-05, {   1,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  1.234168808908e-05,  4.049705872980e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -5.389366229152e-06,  1.095014405609e-05, {   0,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -9.940671378185e-06,  6.272876618808e-06, {   2,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 {  5.279831568776e-06,  1.088692056573e-05, {   0,   0,   2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  1.000578152344e-05, -6.141398361078e-06, {   0,   0,   1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  1.000414094519e-05,  6.141679603495e-06, {   0,   0,   1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 {  6.889757733039e-06,  9.247512631160e-06, {   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 {  9.700317990717e-06, -6.333308563283e-06, {   0,   0,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0 } },
 { -3.948407551505e-06,  1.206031721458e-05, {   2,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.855599645933e-06, -1.411390683816e-05, {   2,   0,  -1,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -5.350990493159e-06,  1.060025762089e-05, {   2,   0,  -1,   0,   0,   0,   8, -10,   0,   0,   0,   0,   0 } },
 { -8.863455431651e-06, -7.057425862502e-06, {   2,   0,  -2,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -3.478123336455e-06,  1.242212928678e-05, {   2,   0,  -2,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.430359754648e-05, -1.588140321257e-06, {   2,   0,  -1,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.063376779858e-05, -5.247790509132e-06, {   2,   0,   1,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 { -1.054266641532e-05,  5.210397861784e-06, {   2,   0,  -1,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 {  5.489169410104e-06, -1.026265233006e-05, {   2,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -2.591321441031e-06, -1.314853965836e-05, {   1,   1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.429109890943e-05,  1.424505975746e-06, {   0,   0,   1,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 { -4.344934785944e-06, -1.136867786328e-05, {   2,   0,  -1,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 {  3.844999165553e-06, -1.183705263586e-05, {   2,   0,  -1,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 { -4.286869135190e-07, -1.517567464780e-05, {   2,   0,  -1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 { -8.768226503412e-06,  6.666699293083e-06, {   2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  7.526275315023e-06, -7.889468402791e-06, {   0,   0,   2,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -1.967627674680e-06,  1.344636068359e-05, {   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -6.722926580900e-06,  8.684837696756e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0 } },
 { -3.757859576835e-06,  1.162687966948e-05, {   2,   0,   0,   0,   0,  10, -11,   0,   0,   0,   0,   0,   0 } },
 {  1.037970567790e-05,  4.863973765387e-06, {   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0 } },
 {  7.485849333641e-06,  7.669770492394e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -5.259456280929e-06,  9.877590790748e-06, {   0,   0,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.809038709288e-06, -1.132563204457e-05, {   0,   0,   2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  7.072773059114e-06, -8.051999858017e-06, {   0,   0,   1,   0,   0,   0,  -3,   0,   3,   0,   0,   0,   0 } },
 { -2.814899881104e-06, -1.228404230519e-05, {   2,   0,  -1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  9.989715569593e-06,  5.048849122728e-06, {   2,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 { -8.463275445249e-06, -6.491177628798e-06, {   2,   0,   0,   0,   0,   0,  -3,   0,   4,   0,   0,   0,   0 } },
 {  1.177223407210e-05, -3.161732607586e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 { -3.600328072778e-06, -1.129178459354e-05, {   1,   0,   0,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 {  1.072271958119e-05,  4.134978465996e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -2.704852370055e-06,  1.210389311799e-05, {   0,   0,   1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.177644405340e-06,  8.492735578988e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  5.193356675483e-06,  9.434415211523e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   5,   0,   0,   0 } },
 { -5.462129075110e-06, -9.155107681344e-06, {   0,   0,   1,   0,   0,   0,   6, -10,   0,   0,   0,   0,   0 } },
 { -7.994721159633e-06,  6.611629617896e-06, {   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -6.597573927966e-06,  7.999336762485e-06, {   4,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -1.311196837422e-05, -1.227132086461e-06, {   2,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0,   0 } },
 { -3.562223485597e-06,  1.074393223504e-05, {   2,   0,  -2,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -8.532128885198e-06,  5.746308803994e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 { -6.286094140186e-07,  1.361991150700e-05, {   2,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  3.560266013536e-06, -1.068586507948e-05, {   3,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.887376951819e-06,  1.132727922787e-05, {   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 } },
 {  6.085851380000e-06,  8.111662379076e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,  -5,   0,   0,   0 } },
 { -5.183935294907e-06,  8.981661707525e-06, {   0,   0,   1,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 {  4.742737326374e-06, -9.342140205935e-06, {   0,   0,   1,   0,   0,   0,  -7,  11,   0,   0,   0,   0,   0 } },
 { -3.133425473462e-06,  1.093370244587e-05, {   2,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 {  7.352813544880e-06,  6.504856785597e-06, {   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 } },
 {  3.468050227711e-06, -1.036438347923e-05, {   0,   0,   1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.134907231317e-06, -5.652424209117e-06, {   0,   0,   1,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0 } },
 { -3.370286375221e-06, -1.037355411858e-05, {   0,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -8.239668695328e-06, -5.491687118008e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -1.467902422100e-06, -1.224219335895e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -1.458463438808e-06,  1.220098610597e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -7.826340372800e-06,  5.831698882170e-06, {   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  3.376876028071e-06,  1.027006358115e-05, {   4,   0,  -2,   0,   0, -20,  21,   0,   0,   0,   0,   0,   0 } },
 { -1.280872134914e-05,  6.739334683300e-07, {   0,   0,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.556966961617e-06,  9.898410477625e-06, {   0,   0,   1,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 { -3.542876134145e-06, -9.849721362534e-06, {   0,   0,   1,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 {  7.371543342390e-06,  5.839305972582e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  4.473607721365e-06, -8.701135314085e-06, {   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 { -4.862818995005e-06,  8.306105251599e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,  -2,   0,   0,   0 } },
 { -9.827152515130e-06,  3.328056698478e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -4.917303944636e-06,  8.230211622390e-06, {   0,   0,   2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -4.728617125585e-07, -1.259974323329e-05, {   0,   0,   1,   0,   0,   0,  -6,   9,   0,   0,   0,   0,   0 } },
 {  5.647059954925e-06,  7.422420474355e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -3.802896502798e-06, -9.174073871738e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  6.450236435327e-07, -1.232857626880e-05, {   0,   0,   1,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.096585496398e-06, -7.830858644024e-06, {   0,   0,   2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  4.333421412237e-06,  8.568374761945e-06, {   2,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  8.628908210205e-06,  4.248425114202e-06, {   2,   0,  -2,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -3.123588967361e-06,  9.705946840435e-06, {   2,   0,   0,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 {  6.878164658245e-06,  5.889041995170e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   5,   0,   0,   0 } },
 {  1.002145837088e-05,  2.736909597519e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -8.740108383010e-06,  3.966567808392e-06, {   3,   1,  -1,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -4.447713527790e-06,  8.248036134723e-06, {   2,   0,  -3,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 { -3.324259605002e-06,  9.364656835986e-06, {   2,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.731730681268e-06,  3.915628272532e-06, {   2,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -5.125107127481e-06, -7.515430179485e-06, {   2,   0,  -1,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 {  4.836450124207e-06,  7.758088426352e-06, {   0,   0,   2,   0,   0, -18,  12,   8,  -3,   0,   0,   0,   0 } },
 { -5.102897730039e-06,  7.486524526922e-06, {   2,   0,  -1,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -9.217937020267e-06,  3.304224247270e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   0,   0,   0,   0 } },
 {  5.192966109286e-06, -7.260006684745e-06, {   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.648585116228e-06,  8.802056363432e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 { -4.978061363715e-06, -7.448690513188e-06, {   2,   0,   0,   0,   0,   0, -11,  21,  -2,   0,   0,   0,   0 } },
 { -8.520128094881e-06, -3.866735133517e-06, {   1,  -1,   1,   0,   0,  20, -20,   0,   0,   0,   0,   0,   0 } },
 {  8.444063789149e-06, -3.941360102456e-06, {   2,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 { -4.947227851551e-06,  7.407932901700e-06, {   2,   0,   0,   0,   0,   0,  11, -21,   2,   0,   0,   0,   0 } },
 { -5.527264743219e-06, -6.788822577815e-06, {   1,   0,   1,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -1.008949852812e-05,  2.207596519576e-06, {   2,   0,  -1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  4.418820511398e-06,  7.877840749411e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 } },
 {  6.501179641386e-06, -5.786270626442e-06, {   2,   0,   1,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 {  1.546843004081e-06,  1.072277391209e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  1.546914699344e-06, -1.071978247695e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 {  4.651601136019e-06,  7.603179053104e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -4.728095317388e-06,  7.523284189904e-06, {   2,   0,   0,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 {  8.157172813868e-06, -4.024088080446e-06, {   2,   0,  -2,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  9.319667666057e-06, -2.791882735011e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -4.414888946347e-06,  7.658997164979e-06, {   2,   0,  -1,   0,   0,   0,  -4,   8,   0,   0,   0,   0,   0 } },
 { -1.182230382349e-05, -2.385845353835e-07, {   2,   0,   1,   0,   0, -18,  14,   0,   0,   0,   0,   0,   0 } },
 {  2.045001810673e-06, -9.957357516449e-06, {   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 { -6.853609932973e-06,  5.104984522070e-06, {   2,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  4.581963591570e-06, -7.363324328679e-06, {   0,   0,   0,   0,   0,  18, -12,  -8,   3,   0,   0,   0,   0 } },
 { -5.461426130265e-06, -6.482054392833e-06, {   2,   0,  -3,   0,   0,   0,   2,   0,  -4,   3,   0,   0,   0 } },
 { -3.065784111282e-06,  8.850990242899e-06, {   2,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.454573069065e-07, -1.136901253085e-05, {   4,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.518983467885e-06, -5.354580222583e-06, {   2,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.867257261929e-06,  8.996805024466e-06, {   4,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.036353086430e-06, -9.763454876565e-06, {   1,   0,  -1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -7.481230320078e-06,  4.315730061800e-06, {   0,   0,   1,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 { -4.432264210289e-06, -7.350913037029e-06, {   2,   0,  -1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 {  5.697512497732e-06, -6.050788772048e-06, {   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -8.764453836888e-06,  2.965478445591e-06, {   0,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  8.012780845421e-06,  3.688786219262e-06, {   2,   0,  -4,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.164313037197e-06,  7.449475792424e-06, {   2,   0,  -2,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  4.041069310038e-06, -7.505477731521e-06, {   4,   0,  -2,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  5.263462066254e-06, -6.266077606350e-06, {   0,   0,   1,   0,   1,   0, -10,  11,   0,   0,   0,   0,   0 } },
 {  5.263412142314e-06,  6.265899165865e-06, {   0,   0,   1,   0,  -1,   0,  10, -11,   0,   0,   0,   0,   0 } },
 {  4.999021119222e-06,  6.514162206118e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 } },
 { -5.017535978284e-06, -6.494638206705e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0 } },
 {  4.641157679069e-06, -6.802468752597e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0 } },
 { -4.904462932302e-06,  6.521275586970e-06, {   0,   0,   1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.556244332017e-06, -5.781404793340e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 {  2.473583045011e-06, -8.828234656332e-06, {   2,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.751895543103e-06,  9.503373862069e-06, {   1,  -1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.745682569655e-06, -8.503374510866e-06, {   2,   0,  -1,   0,   0,  11, -12,   0,   0,   0,   0,   0,   0 } },
 { -3.640709729440e-07,  1.086239342058e-05, {   2,   0,  -1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0 } },
 {  2.134602219170e-06, -9.071665492946e-06, {   2,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 } },
 { -7.454919513033e-06, -3.691690910614e-06, {   2,   0,   0,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  2.787536574131e-07,  1.085720786950e-05, {   2,   0,  -1,   0,   0,   0,   9, -12,   0,   0,   0,   0,   0 } },
 { -2.555497245790e-06, -8.545268330842e-06, {   1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -8.300235848963e-06,  2.795878555295e-06, {   2,   0,   0,   0,   0,   0,  -3,   8,  -3,   0,   0,   0,   0 } },
 {  7.402360994680e-06,  3.679298375660e-06, {   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 {  4.421483702496e-06, -6.644753259374e-06, {   2,   0,  -1,   0,   0,   0,  13, -21,   0,   0,   0,   0,   0 } },
 {  7.395696559409e-06,  3.609150183261e-06, {   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 {  4.820124200847e-06, -6.146790893188e-06, {   1,  -1,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  4.040350413776e-06,  6.901707021002e-06, {   2,   0,  -1,   0,   0,   0,  10, -14,   0,   0,   0,   0,   0 } },
 { -5.288176366094e-06, -5.583651244460e-06, {   2,   0,   0,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 {  1.053467496317e-05,  3.222053040993e-07, {   2,   0,  -1,   0,   0,   0,  11, -17,   0,   0,   0,   0,   0 } },
 { -7.089847588902e-06, -3.737089594623e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -4.831539336625e-06, -5.994983296540e-06, {   0,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -2.845975083878e-06,  7.959135048477e-06, {   0,   0,   1,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  7.113043099996e-06, -3.674858101306e-06, {   2,   0,  -1,   0,   0,   0,  12, -19,   0,   0,   0,   0,   0 } },
 {  6.823910616504e-06,  3.938302793308e-06, {   0,   0,   1,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.831860680526e-06,  8.921901992879e-06, {   1,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  7.907292124694e-06, -2.822173103323e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 {  5.790833873512e-06, -4.938114315603e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -5,   0,   0,   0 } },
 { -2.480615239251e-06,  8.237730269710e-06, {   2,   0,  -2,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 { -3.428875380556e-06,  7.285113339087e-06, {   0,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 } },
 { -7.647888917099e-07, -9.929895205852e-06, {   4,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.621711141507e-06,  7.057034440822e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -7.648537569494e-07,  9.891285974220e-06, {   4,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -7.599062851183e-06,  3.025443734226e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   1,   5,   0,   0,   0 } },
 { -4.382197409667e-06, -6.238221708438e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  3.744010603200e-06,  6.850904654339e-06, {   0,   0,   1,   0,   0,   0,   7, -11,   0,   0,   0,   0,   0 } },
 { -1.646335562390e-06, -8.937727365484e-06, {   1,  -1,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -5.599478906239e-06,  4.983853081012e-06, {   2,   0,  -1,   0,   0,  -8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.733727319658e-06, -7.849087224180e-06, {   4,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.891877784876e-06, -6.678326757929e-06, {   0,   0,   1,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0 } },
 { -2.735198353338e-06,  7.830250136114e-06, {   4,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  5.207603080081e-06,  5.312396918996e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 } },
 { -2.550493798237e-06,  7.962319948577e-06, {   2,   0,   0,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  6.932529632085e-06,  3.579440039679e-06, {   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 } },
 {  7.091173753562e-06,  3.406377070574e-06, {   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 } },
 {  2.642842372312e-06, -7.801499940376e-06, {   0,   0,   1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 {  6.451452207485e-06,  3.934410829764e-06, {   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.908398844373e-06, -3.404751205307e-06, {   2,   0,  -1,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  6.906144299306e-06,  3.400839747974e-06, {   2,   0,  -1,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 {  6.277695810662e-06,  3.992208029609e-06, {   0,   0,   1,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0 } },
 {  3.867608638604e-06,  6.397763717835e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   4,   0,   0,   0,   0 } },
 { -4.491572104271e-06, -5.696216758404e-06, {   0,   0,   1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 {  2.523614337904e-06,  7.642429174360e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -2.523614337904e-06, -7.642429174360e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 } },
 { -6.899706200473e-06,  3.176244663138e-06, {   1,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -3.528725057332e-06, -6.515523024644e-06, {   2,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -5.005317792519e-06,  5.023760293114e-06, {   2,   0,  -1,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0 } },
 { -4.994555287489e-06, -5.003059621084e-06, {   2,   0,   0,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  1.724036023095e-06,  8.265778117109e-06, {   0,   0,   3,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 { -1.448108532373e-06, -8.516748579938e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 {  2.954634816795e-07, -9.638236201085e-06, {   2,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 { -1.510344611441e-06, -8.421004590127e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0 } },
 { -4.068380036959e-06, -5.858458689074e-06, {   2,   0,  -1,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.261698873857e-06,  3.664903399601e-06, {   2,   0,   0,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -4.321992926458e-06,  5.601940144458e-06, {   4,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.590641196356e-06,  4.327116513928e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -3,   0,   0,   0,   0 } },
 {  2.929482239804e-07,  9.618436789903e-06, {   2,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  2.765385552011e-06,  7.136106262168e-06, {   4,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  6.646554391523e-06,  3.243734736142e-06, {   2,   0,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.417161658095e-06,  8.451228786625e-06, {   2,   0,  -1,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 { -6.576611345662e-06,  3.255743713233e-06, {   2,   0,  -2,   0,   0,  26, -29,   0,   0,   0,   0,   0,   0 } },
 { -3.532254897953e-06,  6.257329670375e-06, {   2,   0,  -1,   0,   0,   0,   9, -11,   0,   0,   0,   0,   0 } },
 {  2.679362761184e-06,  7.079431735661e-06, {   2,   0,  -3,   0,   0,  -2,  13, -15,   0,   0,   0,   0,   0 } },
 { -6.679604862738e-06, -3.074970479331e-06, {   1,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -8.880522588717e-06, -8.635113980552e-07, {   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 } },
 {  1.924857455700e-06, -7.810342269285e-06, {   2,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  4.237089876278e-06,  5.490308295311e-06, {   0,   0,   3,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 {  6.380408515927e-06, -3.277920767245e-06, {   2,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 } },
 {  6.541309433131e-06,  3.106361565540e-06, {   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.461577949816e-06,  7.181279401818e-06, {   2,   0,  -1,   0,   0,   8, -10,   0,   0,   0,   0,   0,   0 } },
 {  6.265777938960e-06, -3.369768282706e-06, {   2,   0,   0,   0,   0,   0,   3,   0,  -4,   0,   0,   0,   0 } },
 {  4.982873088368e-06,  4.609820428021e-06, {   0,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   0,   0,   0 } },
 { -1.473953540602e-06,  8.110381934349e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -6.716702840110e-07,  8.911690704634e-06, {   4,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 { -8.454607397479e-06,  1.101350481507e-06, {   2,   0,  -1,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.461202791037e-06, -1.086234954324e-06, {   2,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0 } },
 {  1.565122444960e-06, -7.963854093360e-06, {   0,   1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.525688957285e-06, -3.001823778223e-06, {   2,   0,  -2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  7.562988895994e-06,  1.959638353089e-06, {   1,  -1,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  3.188447421473e-06, -6.329352778628e-06, {   0,   0,   1,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0 } },
 {  5.996475074779e-06,  3.505241048835e-06, {   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -4.192431386366e-06, -5.307637100927e-06, {   0,   0,   3,   0,   0, -15,   9,   4,   0,   0,   0,   0,   0 } },
 { -4.192431386366e-06,  5.307637100927e-06, {   0,   0,   1,   0,   0,  15,  -9,  -4,   0,   0,   0,   0,   0 } },
 {  6.489951131471e-06,  2.975249244510e-06, {   0,   0,   0,   0,   0,  17, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.609449125670e-06, -3.853024565319e-06, {   0,   0,   1,   0,   0,   0,   3,  -6,   2,   0,   0,   0,   0 } },
 {  5.601752225714e-06,  3.849004743957e-06, {   0,   0,   1,   0,   0,   0,  -3,   6,  -2,   0,   0,   0,   0 } },
 {  4.102005093606e-06, -5.342194430098e-06, {   0,   0,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 } },
 {  7.750470706365e-06, -1.619495441041e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   0,   0,   0,   0 } },
 {  5.272658965279e-06,  4.094934867606e-06, {   2,   0,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -1.332382445229e-06,  8.027051664767e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   0,   0 } },
 { -9.055873584250e-08,  9.261663254661e-06, {   0,   0,   1,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  5.356832405474e-06, -3.932964987801e-06, {   0,   0,   1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 {  7.235452053859e-06,  2.040120811348e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   2,  -7,   0,   0,   0 } },
 { -6.674071789466e-06,  2.598761719335e-06, {   3,   0,   0,   0,   0,   0, -34,  41,  -2,   0,   0,   0,   0 } },
 { -6.674071789466e-06, -2.598761719335e-06, {   1,   0,   0,   0,   0,   0,  34, -41,   2,   0,   0,   0,   0 } },
 { -5.774992805675e-06,  3.489216184888e-06, {   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -7.355231120140e-06, -1.893497513712e-06, {   1,  -1,   1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.531364893105e-06, -4.707194258995e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   0,  -3,   0,   0,   0 } },
 {  3.988751728850e-06,  5.207703361433e-06, {   2,   0,   0,   0,   0,   0,  -9,  16,  -4,  -5,   0,   0,   0 } },
 { -5.033050676118e-06,  4.146803210600e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,  -5,   0,   0,   0 } },
 {  6.281181145089e-06, -2.896746594307e-06, {   2,   0,   0,   0,   0, -14,  12,   0,   0,   0,   0,   0,   0 } },
 {  4.509002556675e-06,  4.668251262429e-06, {   4,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -7.325861237862e-07,  8.441262425308e-06, {   1,  -1,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  7.325861237862e-07, -8.441262425308e-06, {   1,  -1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  1.538478806466e-06, -7.599318304763e-06, {   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.123495571250e-06, -5.972027741095e-06, {   1,   0,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0 } },
 {  6.212774969016e-06, -2.863490751077e-06, {   4,   0,  -1,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 {  3.793336057873e-06, -5.278293284066e-06, {   1,  -2,   1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -3.793336057873e-06,  5.278293284066e-06, {   1,  -2,  -1,   0,   0,  20, -18,   0,   0,  -1,   0,   0,   0 } },
 { -6.855060758049e-06, -2.212971444545e-06, {   1,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 {  2.234513100955e-06,  6.793007779765e-06, {   0,   2,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0 } },
 { -6.770820011435e-06,  2.215807444883e-06, {   0,   0,   1,   0,   0,   0,  12, -24,   7,   5,   0,   0,   0 } },
 { -6.770820011435e-06, -2.215807444883e-06, {   0,   0,   1,   0,   0,   0, -12,  24,  -7,  -5,   0,   0,   0 } },
 { -4.546513967756e-06,  4.393187303528e-06, {   2,   0,  -1,   0,   0,   0,   2,  -8,   3,   0,   0,   0,   0 } },
 {  3.421243793571e-06,  5.503174045443e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0 } },
 {  2.562537868334e-06, -6.307323216586e-06, {   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -9.998633230925e-08, -8.759681451395e-06, {   2,   0,   0,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -3.269094870233e-06,  5.569318766872e-06, {   0,   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0 } },
 { -6.479762327551e-06,  2.348004681334e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -5,   5,   0,   0,   0 } },
 {  4.315756112436e-06, -4.504745945280e-06, {   2,   0,  -1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -3.716734875525e-06,  5.064871375658e-06, {   2,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  5.949269308327e-06,  2.806528842168e-06, {   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0 } },
 {  6.297118836488e-06,  2.397987519103e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -3.130507845848e-06, -5.562344389082e-06, {   2,  -2,  -1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  4.626439950738e-06,  4.058650939596e-06, {   2,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 } },
 { -7.411980123425e-06, -1.260079924562e-06, {   2,   0,   1,   0,   0,   0,  -5,   8,  -3,   0,   0,   0,   0 } },
 {  8.383601358882e-06, -2.720784021778e-07, {   4,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -5.895296977246e-06,  2.748585582277e-06, {   2,   0,  -3,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0 } },
 {  1.668541008509e-06, -6.974560369336e-06, {   2,   0,  -2,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  3.110997071619e-06,  5.528574471892e-06, {   2,  -2,   1,   0,   0,   0,   9, -13,   0,   0,   0,   0,   0 } },
 {  6.449886116804e-07,  7.991470543701e-06, {   0,   0,   1,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 {  5.906954099966e-06,  2.725546217594e-06, {   0,   0,   1,   0,   0,  15, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.403156836786e-06,  1.204750884148e-06, {   0,   0,   0,   0,   0,   0,   9, -15,   0,   0,   0,   0,   0 } },
 {  8.312047188520e-06,  2.683664509685e-07, {   4,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -3.056729405708e-06, -5.519524126806e-06, {   2,   0,  -2,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  3.606425881921e-06, -4.963797256350e-06, {   2,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  5.729056359100e-06, -2.840791875082e-06, {   2,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.478169756428e-06,  7.087407997916e-06, {   4,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.698956501689e-06,  2.859406786725e-06, {   0,   0,   2,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -3.513320342153e-06, -5.014850693117e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 { -1.257418939984e-07, -8.399199810468e-06, {   2,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 { -2.060179414761e-06,  6.463116910853e-06, {   2,   0,   0,   0,   0,  13, -14,   0,   0,   0,   0,   0,   0 } },
 { -6.652976401781e-06, -1.853892553048e-06, {   0,   0,   1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  6.106076869134e-06,  2.396843227119e-06, {   2,   0,   0,   0,   0,   0,   7, -16,   4,   5,   0,   0,   0 } },
 { -8.274456799655e-06,  1.371731618448e-07, {   2,   0,   0,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 { -5.312285657832e-06,  3.097167724786e-06, {   2,   0,  -1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 { -3.046894658084e-06,  5.344506918691e-06, {   2,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -3.459236710687e-06, -4.930131538575e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  6.475614343535e-06, -1.913209244974e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   3,   0,   0,   0 } },
 {  3.265712929042e-06, -5.112640071917e-06, {   2,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0 } },
 {  2.503473285349e-06,  5.842529366421e-06, {   2,   0,  -1,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 } },
 { -3.332566904617e-06,  4.985480689432e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 { -1.389929654136e-06,  6.908936230210e-06, {   1,   1,  -1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 { -6.394836618847e-06,  1.900968822850e-06, {   0,   0,   1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  4.006931737797e-06, -4.215729035456e-06, {   2,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  4.037192049714e-06,  4.178295602491e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0 } },
 {  3.458316999733e-06,  4.752255104926e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,   0,   3,   0,   0,   0 } },
 {  3.414682872949e-06, -4.773548935923e-06, {   2,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 } },
 {  2.603765961521e-06,  5.576631260018e-06, {   2,   0,   0,   0,   0, -17,  16,   0,   0,   0,   0,   0,   0 } },
 {  1.288324620082e-06,  6.887359095790e-06, {   1,   1,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  1.985088799997e-06, -6.176523023081e-06, {   2,   0,  -1,   0,   0,  12, -13,   0,   0,   0,   0,   0,   0 } },
 {  4.487590281003e-06,  3.667880730323e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  3.975375063079e-06, -4.147919367600e-06, {   0,   0,   1,   0,   0,   0,  -1,   0,   0,   3,   0,   0,   0 } },
 { -5.156019818112e-06,  2.959124505698e-06, {   0,   0,   1,   0,   0,   0,  -7,  12,   0,   0,   0,   0,   0 } },
 {  5.202256459901e-06, -2.907823446522e-06, {   2,   0,   1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  3.076858961185e-06, -5.015185349416e-06, {   2,   0,  -1,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 } },
 {  2.002120611045e-06, -6.079040346781e-06, {   4,  -2,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 { -1.309378673188e-06, -6.770857623028e-06, {   1,   1,  -1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 {  5.277398109819e-06, -2.767411214393e-06, {   2,   0,   0,   0,   0,  -5,   6,   4,   0,   0,   0,   0,   0 } },
 {  5.266115533895e-06,  2.760104195308e-06, {   2,   0,   0,   0,   0,   5,  -6,  -4,   0,   0,   0,   0,   0 } },
 { -5.389645635240e-06,  2.571958670733e-06, {   2,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 } },
 {  2.084415925177e-06,  5.811521048734e-06, {   2,   0,   0,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 { -4.153967130416e-06,  3.726832389509e-06, {   2,   0,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0 } },
 { -5.086546921021e-06,  2.791645979113e-06, {   0,   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0 } },
 {  2.085320024070e-06, -5.784692933043e-06, {   2,   0,   0,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -6.565038587000e-06,  1.303024162158e-06, {   2,   0,  -1,   0,   0,   8, -11,   0,   0,   0,   0,   0,   0 } },
 { -6.179835869664e-06,  1.674487848563e-06, {   1,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  3.365693280754e-06,  4.486582389137e-06, {   0,   0,   1,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0 } },
 {  1.967813809638e-06,  5.881866523950e-06, {   0,   0,   1,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 {  5.333752255659e-06,  2.504479343223e-06, {   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0 } },
 {  5.617500630548e-06,  2.198376681717e-06, {   2,   0,  -1,   0,   0,   0,  -2,   0,  -2,   5,   0,   0,   0 } },
 { -4.116756357632e-06, -3.697352638150e-06, {   2,   0,  -1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  1.959920718006e-06, -5.846732951180e-06, {   0,   0,   0,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.962267317205e-06, -2.842752802681e-06, {   2,   0,  -1,   0,   0,  -3,   7,  -4,   0,   0,   0,   0,   0 } },
 { -4.954470988137e-06,  2.845368564425e-06, {   2,   0,  -1,   0,   0,   3,  -7,   4,   0,   0,   0,   0,   0 } },
 {  4.801707740545e-06, -2.986008678718e-06, {   0,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -5.442909906436e-07,  7.240455755725e-06, {   4,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  2.355773614272e-06,  5.428866871753e-06, {   2,   0,  -1,   0,   0,   0,   4,  -8,   5,  -5,   0,   0,   0 } },
 {  4.539369197655e-06, -3.243742616136e-06, {   2,   0,  -1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0 } },
 { -6.473704637946e-06, -1.297538059784e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -1,  -5,   0,   0,   0 } },
 {  2.356826614047e-06, -5.408840539551e-06, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -5,   5,   0,   0,   0 } },
 { -3.250502098612e-06,  4.514333619043e-06, {   2,   0,  -1,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 } },
 {  1.754653925633e-06,  5.986529113120e-06, {   2,   0,  -1,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 } },
 { -5.619530633435e-06, -2.121264893783e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 { -5.372827788797e-07, -7.190496992300e-06, {   0,   0,   2,   0,   0,   0, -14,  23,   0,   0,   0,   0,   0 } },
 { -4.498768316435e-06, -3.215514639352e-06, {   2,   0,  -2,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0 } },
 { -5.306005305848e-06,  2.408162963981e-06, {   3,   1,  -2,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 { -1.327952765549e-06,  6.366393882327e-06, {   0,   0,   1,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 { -5.270033991301e-06,  2.419114014219e-06, {   2,   0,   0,   0,   0, -16,  14,   0,   0,   0,   0,   0,   0 } },
 {  1.762236283859e-06, -5.865449131712e-06, {   2,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.547171100882e-06,  3.076654688230e-06, {   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0,   0 } },
 {  3.247898463209e-06, -4.361711445904e-06, {   0,   0,   1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0 } },
 {  4.093970714310e-06,  3.505472281505e-06, {   2,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 } },
 {  5.082254817441e-06, -2.477070766009e-06, {   2,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0 } },
 { -2.155160417930e-06,  5.403889989204e-06, {   0,   0,   0,   0,   0,  19, -18,   0,   0,   0,   0,   0,   0 } },
 {  3.993264364520e-06,  3.562169692406e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0 } },
 {  1.885044356802e-06, -5.662716001891e-06, {   0,   0,   1,   0,   0,   7,  -8,   0,   0,   0,   0,   0,   0 } },
 { -4.885054624299e-06,  2.640861381027e-06, {   2,   0,  -1,   0,   0,   0,   4,   0,  -5,   0,   0,   0,   0 } },
 { -5.912728929339e-06, -1.608492014946e-06, {   2,   0,   0,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 { -7.174839089757e-06,  3.411453376011e-07, {   1,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.063719051223e-06, -5.443024178507e-06, {   2,   0,   0,   0,   0,  -6,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.046532482885e-06,  2.444777904641e-06, {   2,   0,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 } },
 { -4.905984111417e-06, -2.545425429198e-06, {   0,   0,   1,   0,   0,   0,   5, -10,   0,   0,   0,   0,   0 } },
 { -2.011586376861e-07, -7.241084826445e-06, {   0,   0,   0,   0,   0,   0,   5,  -8,   0,   0,   0,   0,   0 } },
 {  1.008366534493e-06,  6.427440054616e-06, {   3,  -1,  -1,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0 } },
 { -6.732232000970e-07,  6.761469761700e-06, {   2,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 } },
 { -4.756634706776e-06,  2.659277230512e-06, {   2,  -2,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 { -7.070145437646e-06, -3.339679791613e-07, {   1,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.660533443666e-06,  4.666229279776e-06, {   0,   0,   2,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0 } },
 { -4.044507358826e-06, -3.267280003389e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 { -3.230413537333e-06, -4.073368715258e-06, {   2,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0 } },
 { -5.286105884369e-06,  2.001255426784e-06, {   0,   0,   1,   0,   0,   3,  -6,   0,   0,   0,   0,   0,   0 } },
 {  1.222694485968e-06, -6.046466231183e-06, {   1,  -1,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 } },
 {  1.213912301823e-06,  6.054819637779e-06, {   1,  -1,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.608091692703e-06, -2.645897966466e-06, {   0,   0,   1,   0,   0,   0,   7, -12,   0,   0,   0,   0,   0 } },
 { -4.539070659339e-06,  2.684786574842e-06, {   4,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  5.928799690159e-06,  1.294913447140e-06, {   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  5.924236505541e-06, -1.294786243927e-06, {   0,   2,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  4.097723648322e-06, -3.107455596956e-06, {   4,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -5.841691497496e-06, -1.362373289376e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 {  3.472412722586e-06, -3.730285722212e-06, {   2,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.511676540880e-06,  5.672849101055e-06, {   1,   0,   1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  1.511676540880e-06, -5.672849101055e-06, {   1,   0,  -1,   0,   0, -23,  25,   0,   0,   0,   0,   0,   0 } },
 {  4.487135310580e-06,  2.686333773262e-06, {   2,   0,  -1,   0,   0,   0,  11, -16,   0,   0,   0,   0,   0 } },
 {  1.628601956631e-06, -5.539831886646e-06, {   4,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -4.280208638730e-06, -2.851513277761e-06, {   0,   0,   1,   0,   0,   0, -10,  17,   0,   0,   0,   0,   0 } },
 {  4.872261032647e-06,  2.225552312845e-06, {   0,   0,   0,   0,   0,  16, -14,   0,   0,   0,   0,   0,   0 } },
 { -5.425223014705e-06, -1.669899237694e-06, {   2,   0,   0,   0,   0,   0,   0,   0,   1,  -5,   0,   0,   0 } },
 { -4.927927019940e-06,  2.112187888190e-06, {   2,   0,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -5.955799693046e-06, -1.074118400043e-06, {   2,   0,   0,   0,   0,   0,  -2,   0,   5,   0,   0,   0,   0 } },
 {  1.764364596890e-06, -5.253274818921e-06, {   2,   0,  -2,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -4.711532129559e-06,  2.302603179894e-06, {   0,   0,   1,   0,   0,   0,  -5,  10,   0,   0,   0,   0,   0 } },
 { -9.669793892665e-08,  6.914691546341e-06, {   2,   0,   0,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 { -5.380543294303e-06,  1.623183183820e-06, {   2,   0,   0,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 {  4.302417816981e-07, -6.538679402214e-06, {   0,   0,   0,   0,   0,   5,  -7,   0,   0,   0,   0,   0,   0 } },
 { -5.977107225092e-06,  9.809486494954e-07, {   0,   0,   0,   0,   0,   0,   5,  -8,   3,   0,   0,   0,   0 } },
 {  4.724824586278e-06,  2.209279490281e-06, {   0,   0,   0,   0,   0,   8,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.210729568873e-06,  1.722688737265e-06, {   0,   0,   0,   0,   0,   0,   1,   0,   1,  -5,   0,   0,   0 } },
 { -5.236276211110e-06,  1.639329522125e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0 } },
 {  4.463824374987e-06,  2.411705772353e-06, {   2,   0,  -2,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -1.658153911436e-06,  5.212374039066e-06, {   2,   0,   0,   0,   0,  14, -15,   0,   0,   0,   0,   0,   0 } },
 { -1.814620286500e-06, -5.049853480557e-06, {   2,   0,  -2,   0,   0,   0,  10, -19,   0,   3,   0,   0,   0 } },
 { -4.433544451323e-06,  2.419776339193e-06, {   0,   0,   1,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0 } },
 { -1.298736906913e-06, -5.554226533625e-06, {   0,   0,   1,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 { -1.818847731544e-06,  5.030644937168e-06, {   2,   0,  -2,   0,   0,   0, -10,  19,   0,  -3,   0,   0,   0 } },
 {  1.712836370563e-06, -5.134153940073e-06, {   0,   0,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0 } },
 {  4.679619943305e-06,  2.154250811776e-06, {   2,  -2,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  1.133563854126e-06,  5.669570911828e-06, {   1,  -1,   1,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 } },
 { -4.940776353662e-06,  1.815104181948e-06, {   2,   0,  -2,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -2.484733949529e-07, -6.462329130875e-06, {   0,   0,   0,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  4.530263316925e-06, -2.178859863541e-06, {   2,   0,   0,   0,   0,  -6,   4,   0,   0,   0,   0,   0,   0 } },
 {  5.151081280662e-06,  1.534381953902e-06, {   2,   0,   0,   0,   0,   0,   2,  -6,   0,   0,   0,   0,   0 } },
 {  1.248554987885e-06,  5.427493865018e-06, {   2,   0,  -1,   0,   0,   5,  -4,   0,   0,   0,   0,   0,   0 } },
 {  1.026149573413e-09,  6.652075447353e-06, {   2,   0,  -1,   0,   0,   0,  10, -13,   0,   0,   0,   0,   0 } },
 { -2.338656944353e-07,  6.337366178098e-06, {   2,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0 } },
 {  4.090789001078e-07,  6.138401723159e-06, {   2,   0,  -2,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 { -6.312644055243e-06, -1.821576784175e-07, {   2,   0,  -1,   0,   0,   0,  10, -16,   0,   0,   0,   0,   0 } },
 {  1.558180769391e-06,  4.873967358384e-06, {   4,   0,  -2,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -4.896196348078e-06,  1.513221589644e-06, {   2,   0,  -1,   0,   0,  -8,  14,   0,   0,   0,   0,   0,   0 } },
 { -1.020176434412e-06,  5.358517031029e-06, {   1,  -1,   0,   0,   0,  21, -20,   0,   0,   0,   0,   0,   0 } },
 { -6.227423182874e-06,  1.081936326189e-07, {   2,   0,  -1,   0,   0,   0,   9, -19,   0,   3,   0,   0,   0 } },
 {  5.253562621553e-06, -1.050482700831e-06, {   0,   0,   3,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  5.250127125787e-06,  1.050332153810e-06, {   0,   0,   3,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.471791953877e-06,  4.822120770361e-06, {   2,   0,  -2,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  2.754588728700e-07,  6.015519058456e-06, {   2,   0,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0,   0 } },
 {  4.941948171560e-06, -1.334371758881e-06, {   0,   0,   0,   0,   0,   0,   1,   0,  -3,   5,   0,   0,   0 } },
 {  1.038349964879e-06, -5.090969437561e-06, {   1,  -1,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 } },
 { -9.785577002545e-07, -5.144050880815e-06, {   1,  -1,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  5.650614941815e-06, -4.554901409016e-07, {   2,   0,  -1,   0,   0,   0,  -8,  16,  -6,   0,   0,   0,   0 } },
 { -4.389638834546e-07,  5.635263620174e-06, {   2,   0,   0,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  9.933713900562e-08,  5.966124735348e-06, {   2,   0,  -2,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -8.850726491214e-07, -5.142884998535e-06, {   2,   0,   0,   0,   0,   0,  -7,  10,   0,   0,   0,   0,   0 } },
 {  5.543627633351e-06,  4.304095394211e-07, {   2,   0,  -1,   0,   0,   0,   8, -16,   6,   0,   0,   0,   0 } },
 { -2.559585598224e-07,  5.550148221670e-06, {   2,   0,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0 } },
 {  3.574535194741e-07,  5.359511253622e-06, {   0,   0,   0,   0,   0,  10,  -4,   0,   0,   0,   0,   0,   0 } },
 {  2.540060773806e-07, -5.433571299283e-06, {   4,   0,  -3,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -2.881529898025e-07,  5.284038825839e-06, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0 } },
 { -2.813149088859e-07, -5.269631806746e-06, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0 } },
 { -1.213719001716e-07,  5.422976987647e-06, {   0,   0,   0,   0,   0,   0,   6,  -9,   0,   0,   0,   0,   0 } },
 {  3.922744489916e-07, -5.024877247532e-06, {   2,   0,  -2,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  1.697136893419e-06, -3.649990040750e-06, {   0,   0,   0,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.324497454268e-07, -5.001010051823e-06, {   4,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 { -1.286348343737e-06,  1.225229759367e-06, {   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 } },
 {  1.248926341296e-06, -1.241860428930e-06, {   0,   0,   1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -1.092433429673e-06, -9.841280800490e-07, {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 } },
 {  1.750237733580e-07,  6.788251998500e-07, {   1,   1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.386481229070e-07, -5.317427046700e-07, {   1,  -1,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  4.074825327720e-07,  1.766696033600e-07, {   2,   0,  -1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0 } },
 {  8.614585948400e-08,  2.494502697000e-07, {   1,  -1,  -1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 210> g_PERT_S3_3
{ {
 {  0.000000000000e+00,  1.490000000000e-03, {   2,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.110000000000e-03, {   2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  9.500000000000e-04, {   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -7.700000000000e-04, {   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.600000000000e-04, {   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.300000000000e-04, {   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -1.800000000000e-04, {   2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -4.657348443684e-05, -1.067868088331e-04, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -4.367735031233e-05,  1.001488975196e-04, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.400000000000e-04, {   2,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  1.200000000000e-04, {   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  9.000000000000e-05, {   2,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  5.988537018087e-05, -2.413266278796e-05, {   0,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  5.987610191651e-05,  2.412424235975e-05, {   0,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  7.000000000000e-05, {   2,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -7.000000000000e-05, {   2,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  5.000000000000e-05, {   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -1.309383743933e-05, -3.002885516495e-05, {   2,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.302599933607e-05,  2.987337089078e-05, {   2,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -4.000000000000e-05, {   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   4,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -3.000000000000e-05, {   2,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  0.000000000000e+00,  3.000000000000e-05, {   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 { -8.613761718971e-06, -1.975871309159e-05, {   2,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -8.099535709070e-06,  1.857924199801e-05, {   2,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.230135011835e-06, -1.645693246541e-05, {   0,   0,   1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  4.231650801562e-06,  1.644503667462e-05, {   0,   0,   1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.051315425167e-05,  9.978170011858e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  0.000000000000e+00, -2.000000000000e-05, {   2,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 } },
 {  3.895416280262e-06,  1.248527849102e-05, {   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  3.864896780187e-06, -1.239235650152e-05, {   0,   0,   2,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.437355097186e-07, -1.026914616421e-05, {   2,   0,  -1,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  5.856366789594e-06,  4.792956826575e-06, {   0,   0,   1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  5.848578957290e-06, -4.791388199116e-06, {   0,   0,   1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -4.005107674976e-06,  6.154250280016e-06, {   0,   0,   1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -3.986630334190e-06, -6.121351715988e-06, {   0,   0,   1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -2.538937367635e-06, -5.821446275990e-06, {   0,   0,   3,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -2.381920127449e-06,  5.461564284920e-06, {   0,   0,   1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 {  4.752552160431e-06,  2.889960271381e-06, {   2,   0,  -1,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  4.631048781534e-06,  2.986479801351e-06, {   2,   0,   0,   0,   0,  -8,  13,   0,   0,   0,   0,   0,   0 } },
 {  7.481175067432e-06, -1.280850418869e-07, {   2,   0,   0,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  4.712189320166e-06, -2.885143996577e-06, {   2,   0,  -1,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 {  7.498054982789e-06,  9.073194081264e-08, {   2,   0,   0,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  4.581730775773e-06, -2.977169557877e-06, {   2,   0,   0,   0,   0,   8, -13,   0,   0,   0,   0,   0,   0 } },
 { -2.769553566718e-06,  4.584945363766e-06, {   2,   0,   0,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.736111412457e-06,  4.605141828054e-06, {   2,   0,  -1,   0,   0,   0,  -8,  15,   0,   0,   0,   0,   0 } },
 { -2.650082533009e-06, -4.442009592796e-06, {   2,   0,  -1,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 { -2.657668380929e-06, -4.369696165012e-06, {   2,   0,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0 } },
 {  2.015582016356e-06, -4.718951306120e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -5,   0,   0,   0 } },
 {  2.022450173432e-06,  4.423305214665e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   5,   0,   0,   0 } },
 { -1.241468293675e-06, -4.433527923831e-06, {   2,   0,  -1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 { -1.198600330470e-06,  4.385179838385e-06, {   2,   0,  -1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  3.086810034445e-06,  1.559735766845e-06, {   2,   0,  -1,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0 } },
 {  3.261239396328e-06,  1.315585263603e-06, {   0,   0,   2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  3.259256130668e-06, -1.315478754062e-06, {   0,   0,   2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  1.084762474973e-06, -3.479594553361e-06, {   2,   0,   1,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.084762474973e-06,  3.479594553361e-06, {   2,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  7.898774252075e-07, -3.746257556532e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  7.818397094781e-07,  3.687140296291e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -1.961673646807e-06,  1.873351781465e-06, {   0,   0,   1,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 {  2.603336193513e-06, -1.217863599553e-06, {   0,   0,   1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 { -3.710492191350e-06, -5.489237552142e-08, {   2,   0,  -1,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  2.628843455738e-06,  1.134200172723e-06, {   1,   0,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 { -1.136410986702e-06, -2.606010720963e-06, {   2,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -1.108784512455e-06,  2.542227799973e-06, {   2,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -1.096014489939e-06,  2.550247468231e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  2.443016958683e-06,  1.136669721693e-06, {   0,   0,   1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 { -1.815154580396e-06, -1.342972662748e-06, {   2,   0,  -1,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 { -3.051527074319e-06,  5.622043006326e-08, {   2,   0,  -1,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0 } },
 {  1.949458920714e-06, -9.290195871079e-07, {   2,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  6.735958416959e-07,  2.161593865381e-06, {   2,   0,  -2,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  6.682004684579e-07, -2.145166701957e-06, {   2,   0,   0,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  1.858600614120e-06, -8.844158071252e-07, {   2,   0,  -1,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0 } },
 {  8.747495363639e-09,  2.672480647662e-06, {   2,   0,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  2.226298932247e-06, -4.175196084740e-07, {   2,   0,   1,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  1.290722657175e-06, -1.332716907031e-06, {   1,   0,   1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 { -1.290722657175e-06,  1.332716907031e-06, {   1,   0,  -1,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.023419258555e-09,  2.603007368674e-06, {   2,   0,  -1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 {  7.834151694177e-07, -1.816121815540e-06, {   2,   0,  -1,   0,   0,   0,   2,   0,  -5,   5,   0,   0,   0 } },
 {  1.888831478848e-06, -6.746122202667e-07, {   2,   0,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -8.261708101216e-08,  2.366584915996e-06, {   0,   0,   1,   0,   0,   3,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.109107304795e-07,  1.544202914747e-06, {   2,   0,   0,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -6.322702532790e-08, -2.289522451757e-06, {   0,   0,   1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 {  1.581286697064e-06,  7.445692961281e-07, {   0,   0,   1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  1.544263718638e-06, -7.225119314432e-07, {   0,   0,   1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -2.010157083286e-06, -2.110347618507e-07, {   2,   0,   0,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.221646456105e-06, -9.671997155947e-07, {   2,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -1.071034064961e-06, -1.077023526408e-06, {   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.850214538852e-08, -2.050814706660e-06, {   2,   0,  -1,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.065638487414e-06, -1.069109493579e-06, {   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.883460450591e-08, -1.988492142086e-06, {   2,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0,   0 } },
 { -1.323496965692e-06, -6.900196467957e-07, {   0,   0,   1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -1.314250343287e-06,  6.837006516272e-07, {   0,   0,   1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 {  1.446208549713e-06, -5.284392678406e-07, {   2,   0,  -1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 { -1.144213968969e-07, -1.795515830415e-06, {   0,   0,   1,   0,   0,  -6,   8,   0,   0,   0,   0,   0,   0 } },
 {  8.274288335547e-07,  1.025879757788e-06, {   0,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -1.630817700788e-06, -2.058405906144e-07, {   2,   0,  -1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  8.193180987492e-07, -1.016366718124e-06, {   0,   0,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 {  1.257440110540e-06,  5.574306694204e-07, {   2,   0,  -1,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.190245477351e-06,  5.895159002968e-07, {   2,   0,   0,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 { -1.186124528887e-06, -5.742955823705e-07, {   2,   0,  -1,   0,   0,   4,  -5,   0,   0,   0,   0,   0,   0 } },
 { -8.540344438201e-07,  9.037687716795e-07, {   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.150011823970e-06,  5.744369055133e-07, {   2,   0,  -1,   0,   0,   0,   2,  -4,   0,   0,   0,   0,   0 } },
 {  9.616250398893e-07, -7.617930082207e-07, {   2,   0,  -1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -8.257994732456e-07,  8.659255537891e-07, {   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 {  1.160054892771e-06,  5.254834038115e-07, {   2,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0 } },
 {  1.182689162476e-06, -4.795116807275e-07, {   0,   0,   1,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0 } },
 {  4.756843380503e-07,  1.090263878578e-06, {   0,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -7.729419382634e-07, -7.852308129759e-07, {   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 } },
 {  8.701613367018e-07, -6.778040935689e-07, {   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,   0,   0,   0 } },
 { -5.797580051581e-07, -9.201383212782e-07, {   2,   0,  -1,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0 } },
 { -7.382199927922e-07,  7.570494737001e-07, {   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 } },
 { -8.017239754426e-07, -6.719317964463e-07, {   2,   0,  -2,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 {  9.734337035359e-07, -4.725690254075e-07, {   2,   0,  -1,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 { -9.079565000456e-07, -5.202864349372e-07, {   2,   0,   0,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 { -8.928883645314e-07, -5.237131970589e-07, {   2,   0,  -1,   0,   0,   0,   9, -17,   0,   0,   0,   0,   0 } },
 {  9.553327972886e-07, -4.568907814580e-07, {   2,   0,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0 } },
 {  5.987188622192e-07,  7.786814763906e-07, {   2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 {  7.928381145969e-07,  5.772044104981e-07, {   2,   0,  -2,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  7.895044035196e-07, -5.758104463253e-07, {   2,   0,  -2,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  9.465783505665e-07,  4.153132137794e-07, {   0,   0,   1,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 } },
 {  5.874978015310e-07,  7.727959740951e-07, {   2,   0,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 } },
 { -6.768703726482e-07,  6.820216027758e-07, {   2,   0,  -1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 { -8.495660575682e-07,  4.928265803797e-07, {   2,   0,  -1,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -8.507693279567e-07,  4.835008579033e-07, {   2,   0,   0,   0,   0,   0,  -9,  17,   0,   0,   0,   0,   0 } },
 { -3.200619540809e-07,  9.916101336904e-07, {   0,   0,   1,   0,   0,   0,   3,  -8,   3,   0,   0,   0,   0 } },
 {  9.873135903606e-07, -3.229357023122e-07, {   2,   0,   0,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  7.350198422597e-07,  5.682264333855e-07, {   0,   0,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0 } },
 {  8.720256541059e-07,  4.237952486953e-07, {   2,   0,  -1,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0 } },
 {  5.550847348334e-07, -7.348948147987e-07, {   2,   0,  -1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -1.112068859630e-06,  1.742361625129e-07, {   0,   0,   1,   0,   0,   0,   1,   0,   2,  -5,   0,   0,   0 } },
 { -1.148477776694e-06, -1.316070113409e-07, {   0,   0,   1,   0,   0,   0,  -1,   0,  -2,   5,   0,   0,   0 } },
 {  1.203259454789e-06, -6.693256356696e-08, {   2,   0,  -1,   0,   0,   0,   7, -10,   0,   0,   0,   0,   0 } },
 { -7.416040660000e-08,  1.193903457505e-06, {   0,   0,   1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -3.211079283785e-07, -9.350698772667e-07, {   2,   0,   0,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  5.438931806202e-07, -7.093798002917e-07, {   2,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0 } },
 { -7.413553584050e-08, -1.172558040027e-06, {   0,   0,   1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  9.801781043063e-07,  2.485220776284e-07, {   2,   0,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 {  1.400846108353e-07,  1.030907393830e-06, {   0,   0,   2,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0 } },
 {  1.395799632962e-07, -1.031229498118e-06, {   0,   0,   2,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0 } },
 { -6.704384492019e-07, -4.960203202474e-07, {   2,   0,  -2,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  9.187982470979e-07, -1.912637218062e-07, {   2,   0,   1,   0,   0,   0,   4,  -8,   3,   0,   0,   0,   0 } },
 {  9.162184157709e-07,  1.892736385033e-07, {   2,   0,   1,   0,   0,   0,  -4,   8,  -3,   0,   0,   0,   0 } },
 {  9.203901703847e-07,  1.703361268008e-07, {   2,   0,   0,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 { -5.348726776485e-07,  5.445772215615e-07, {   2,   0,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0 } },
 { -8.933981092275e-07, -1.653401460527e-07, {   2,   0,  -2,   0,   0, -19,  21,  -3,   0,   0,   0,   0,   0 } },
 {  3.434823441486e-07,  7.036787567461e-07, {   0,   0,   1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  8.713086908944e-07, -1.633483129659e-07, {   2,   0,   0,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  3.320123892008e-07, -6.827863992883e-07, {   0,   0,   1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 { -5.576587138620e-07, -4.171137302365e-07, {   0,   0,   0,   0,   0,  18, -17,   0,   0,   0,   0,   0,   0 } },
 {  5.728631449363e-07,  3.898769320204e-07, {   0,   0,   1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 { -2.922504893467e-08,  9.174275837087e-07, {   2,   0,   0,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 { -2.553513339839e-08,  9.123771598691e-07, {   2,   0,  -1,   0,   0,   0,  -7,  13,   0,   0,   0,   0,   0 } },
 {  5.492543915906e-07, -3.751312656571e-07, {   0,   0,   1,   0,   0,   0,   5,  -9,   0,   0,   0,   0,   0 } },
 {  4.754304516405e-07,  4.463939769128e-07, {   2,   0,   0,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  2.115625567001e-07, -6.782169769024e-07, {   0,   0,   3,   0,   0, -10,   3,   0,   0,   0,   0,   0,   0 } },
 {  2.115625567001e-07,  6.782169769024e-07, {   0,   0,   1,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0 } },
 {  2.962330223549e-07,  5.692505215310e-07, {   2,   0,   0,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  4.658320883817e-07,  3.900097204684e-07, {   2,   0,   0,   0,   0, -12,   8,   0,   0,   0,   0,   0,   0 } },
 { -2.540600354329e-07, -5.973074582507e-07, {   0,   0,   1,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 { -1.972707738335e-07, -6.520023800793e-07, {   2,   0,  -1,   0,   0,   0,  -9,  15,   0,   0,   0,   0,   0 } },
 {  2.906463674512e-07,  5.557072398448e-07, {   2,   0,  -1,   0,   0,   0,  -6,  11,   0,   0,   0,   0,   0 } },
 {  5.100017171175e-07,  3.349077500774e-07, {   2,   0,   0,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  7.098985281653e-07, -1.331919942809e-07, {   0,   0,   2,   0,   0, -18,  15,   0,   0,   0,   0,   0,   0 } },
 {  5.094838841512e-07, -3.284324843169e-07, {   2,   0,  -1,   0,   0,   0,   8, -12,   0,   0,   0,   0,   0 } },
 { -2.739471840212e-08, -8.078826293233e-07, {   2,   0,  -1,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  4.057204705633e-07,  4.288185654949e-07, {   2,   0,  -1,   0,   0,   0,   2,   0,  -4,   0,   0,   0,   0 } },
 { -4.260075101506e-07, -4.006523660269e-07, {   2,   0,  -2,   0,   0,   0,   2,   0,  -3,   0,   0,   0,   0 } },
 {  4.691133664486e-07, -3.475563643642e-07, {   0,   0,   2,   0,   0, -18,  17,   0,   0,   0,   0,   0,   0 } },
 { -2.569646986911e-07,  5.592402566056e-07, {   0,   0,   1,   0,   0,   0,  -2,   0,   4,  -5,   0,   0,   0 } },
 { -3.097742144262e-08, -7.800085770217e-07, {   2,   0,   0,   0,   0,   0,   7, -13,   0,   0,   0,   0,   0 } },
 {  4.892705855069e-07,  3.184438769541e-07, {   2,   0,  -1,   0,   0,   0,  -5,   9,   0,   0,   0,   0,   0 } },
 {  2.402209445637e-07,  5.509134474429e-07, {   1,   0,   1,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  2.402209445637e-07, -5.509134474429e-07, {   1,   0,  -1,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -5.241353121732e-07,  2.644315484087e-07, {   0,   0,   1,   0,   0,   0,  -6,   8,   0,   0,   0,   0,   0 } },
 { -2.350140862524e-07, -5.390034784987e-07, {   4,   0,   0,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 {  6.996410719254e-07, -6.506814761280e-08, {   0,   0,   1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -2.295228483206e-07,  5.264176144371e-07, {   4,   0,  -2,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } },
 { -4.520484373518e-07,  3.032006668095e-07, {   2,   0,   0,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 {  6.735192683109e-07,  6.515659498927e-08, {   0,   0,   1,   0,   0,   0,  -3,   6,   0,   0,   0,   0,   0 } },
 { -2.312894239549e-07, -5.065848841977e-07, {   2,   0,   0,   0,   0, -15,  13,   0,   0,   0,   0,   0,   0 } },
 { -2.226912655398e-07,  5.112363628681e-07, {   2,   0,  -3,   0,   0,   0,   2,   0,  -1,  -5,   0,   0,   0 } },
 {  4.308958959050e-07, -2.983857391134e-07, {   2,   0,  -2,   0,   0,   0, -37,  48,   0,   0,   0,   0,   0 } },
 {  3.576334651072e-07, -3.696521377917e-07, {   3,   0,   0,   0,   0, -15,  12,   0,   0,   0,   0,   0,   0 } },
 {  3.576334651072e-07,  3.696521377917e-07, {   1,   0,   0,   0,   0,  15, -12,   0,   0,   0,   0,   0,   0 } },
 {  5.728030928105e-07, -1.492365199058e-07, {   2,   0,   0,   0,   0,   0,   7, -15,   0,   0,   0,   0,   0 } },
 {  4.824515481656e-07, -2.375610180502e-07, {   2,   0,   0,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 {  4.817102621028e-07, -2.348318514180e-07, {   0,   0,   1,   0,   0,   0,  -3,   5,   0,   0,   0,   0,   0 } },
 { -5.106506891572e-07,  2.037135196190e-07, {   0,   0,   2,   0,   0, -26,  29,   0,   0,   0,   0,   0,   0 } },
 {  4.504807830204e-08, -6.595141141537e-07, {   2,   0,   0,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.478138799199e-07,  4.931988235097e-08, {   0,   0,   1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  2.327391175929e-07, -4.540040416385e-07, {   2,   0,  -1,   0,   0,   0,   6, -11,   0,   0,   0,   0,   0 } },
 {  2.126396583096e-07,  4.690162710958e-07, {   2,   0,   0,   0,   0,   0,  -2,   0,   0,   5,   0,   0,   0 } },
 {  2.062794310774e-07,  4.690547899398e-07, {   1,   1,   0,   0,   0, -20,  20,   0,   0,   0,   0,   0,   0 } },
 {  4.685942880547e-07, -2.066246853378e-07, {   2,   0,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0 } },
 {  4.510753493344e-07,  2.194548121609e-07, {   0,   0,   1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0 } },
 {  6.148279502829e-07, -4.773545756447e-08, {   0,   0,   1,   0,   0,   0,   4,  -7,   0,   0,   0,   0,   0 } },
 {  3.930565457168e-08,  6.149692710467e-07, {   0,   0,   1,   0,   0,   0,   0,   0,  -4,  10,   0,   0,   0 } },
 {  3.922100034436e-08, -6.145686194279e-07, {   0,   0,   1,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0 } },
 { -4.060952473229e-08,  6.118605097664e-07, {   2,   0,  -2,   0,   0,   6,  -8,   0,   0,   0,   0,   0,   0 } },
 {  6.085344614600e-07,  3.945862621513e-08, {   2,   0,   0,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 { -5.380052615130e-07, -1.009359951365e-07, {   0,   0,   0,   0,   0,  18, -15,   0,   0,   0,   0,   0,   0 } },
 {  5.715092818382e-07,  3.597995533248e-08, {   2,   0,  -1,   0,   0,   0,  -4,   7,   0,   0,   0,   0,   0 } },
 {  5.832044393948e-07,  7.502685211290e-09, {   2,   0,  -3,   0,   0,   0,  14, -23,   0,   0,   0,   0,   0 } },
 {  4.369833720595e-08,  5.468115178553e-07, {   2,   0,   0,   0,   0,  -5,   7,   0,   0,   0,   0,   0,   0 } },
 {  5.457010550435e-07, -2.585559127944e-08, {   2,   0,   0,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 {  5.172053699957e-07, -2.200579662776e-08, {   2,   0,  -1,   0,   0,   0,   3,  -6,   0,   0,   0,   0,   0 } },
 { -5.154525955400e-07, -1.845864586074e-08, {   2,   0,   0,   0,   0,  -8,  12,   0,   0,   0,   0,   0,   0 } },
 { -4.407179819851e-09, -5.170781162590e-07, {   2,   0,   0,   0,   0,  -4,   6,   0,   0,   0,   0,   0,   0 } }
} };

constexpr array<ELPMPP02PertubationsCoefficient, 2> g_PERT_S3_4
{ {
 { -5.458720424980e-07,  2.801517894073e-07, {   0,   0,   2,   0,   0, -18,  16,   0,   0,   0,   0,   0,   0 } },
 { -5.121329506146e-07, -2.627345838573e-07, {   0,   0,   0,   0,   0,  18, -16,   0,   0,   0,   0,   0,   0 } }
} };

struct AAPLUS_EXT_CLASS ELPMPP02Pertubations
{
  const ELPMPP02PertubationsCoefficient* m_pTable;
  size_t m_nTableSize;
};

constexpr array<ELPMPP02Pertubations, 4> g_PERT_S1
{ {
  { g_PERT_S1_1.data(), g_PERT_S1_1.size() },
  { g_PERT_S1_2.data(), g_PERT_S1_2.size() },
  { g_PERT_S1_3.data(), g_PERT_S1_3.size() },
  { g_PERT_S1_4.data(), g_PERT_S1_4.size() }
} };

constexpr array<ELPMPP02Pertubations, 3> g_PERT_S2
{ {
  { g_PERT_S2_1.data(), g_PERT_S2_1.size() },
  { g_PERT_S2_2.data(), g_PERT_S2_2.size() },
  { g_PERT_S2_3.data(), g_PERT_S2_3.size() }
} };

constexpr array<ELPMPP02Pertubations, 4> g_PERT_S3
{ {
  { g_PERT_S3_1.data(), g_PERT_S3_1.size() },
  { g_PERT_S3_2.data(), g_PERT_S3_2.size() },
  { g_PERT_S3_3.data(), g_PERT_S3_3.size() },
  { g_PERT_S3_4.data(), g_PERT_S3_4.size() }
} };


////////////////////////////// Implementation /////////////////////////////////

double CAAELPMPP02::EclipticLongitude(const double* pT, int nTSize, Correction correction, double* pDerivative) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);

  //Work out the right data arrays to use
  const array<array<double, 5>, 3>* g_pW = nullptr;
  const array<double, 5>* g_pEARTH = nullptr;
  const array<double, 5>* g_pPERI = nullptr;
  const array<double, 5>* g_pZETA = nullptr;
  const array<double, 1023>* g_pMAIN_S1_Coeff = nullptr;
  switch (correction)
  {
    case Correction::Nominal:
    {
      g_pW = &g_W_Nominal;
      g_pEARTH = &g_EARTH_Nominal;
      g_pPERI = &g_PERI_Nominal;
      g_pZETA = &g_ZETA_Nominal;
      g_pMAIN_S1_Coeff = &g_MAIN_S1_Nominal;
      break;
    }
    case Correction::LLR:
    {
      g_pW = &g_W_LLR;
      g_pEARTH = &g_EARTH_LLR;
      g_pPERI = &g_PERI_LLR;
      g_pZETA = &g_ZETA_LLR;
      g_pMAIN_S1_Coeff = &g_MAIN_S1_LLR;
      break;
    }
    case Correction::DE405:
    {
      g_pW = &g_W_DE405;
      g_pEARTH = &g_EARTH_DE405;
      g_pPERI = &g_PERI_DE405;
      g_pZETA = &g_ZETA_DE405;
      g_pMAIN_S1_Coeff = &g_MAIN_S1_DE405;
      break;
    }
    case Correction::DE406:
    {
      g_pW = &g_W_DE406;
      g_pEARTH = &g_EARTH_DE406;
      g_pPERI = &g_PERI_DE406;
      g_pZETA = &g_ZETA_DE406;
      g_pMAIN_S1_Coeff = &g_MAIN_S1_DE406;
      break;
    }
    default:
    {
      assert(false);
      break;
    }
  }
  assert(g_pW);
  assert(g_pPERI);
  assert(g_pEARTH);
  assert(g_pZETA);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(g_pW);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pEARTH);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pPERI);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pZETA);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pMAIN_S1_Coeff);
#endif
  const array<array<double, 5>, 3>& g_W = *g_pW;
  const array<double, 5>& g_EARTH = *g_pEARTH;
  const array<double, 5>& g_PERI = *g_pPERI;
  const array<double, 5>& g_ZETA = *g_pZETA;
  const array<double, 1023>& g_MAIN_S1_Coeff = *g_pMAIN_S1_Coeff;

  //Setup the Delaunay array
  array<array<double, 5>, 4> fDelaunay{};
  for (int i=0; i<5; i++)
  {
    fDelaunay[0][i] = g_W[0][i]  - g_EARTH[i]; //D
    fDelaunay[1][i] = g_W[0][i]  - g_W[2][i];  //F
    fDelaunay[2][i] = g_W[0][i]  - g_W[1][i];  //l
    fDelaunay[3][i] = g_EARTH[i] - g_PERI[i];  //l'
  }
  fDelaunay[0][0] += CAACoordinateTransformation::PI();

  //What will be the return value from this method
  double fResult = 0;

  //Set the output parameter to a default value if required
  if (pDerivative)
    *pDerivative = 0;

  //First the main problem
  constexpr size_t nEndI = g_MAIN_S1.size();
  for (size_t i=0; i<nEndI; i++)
  {
    array<double, 5> fFi{};
    for (int a=0; a<5; a++)
    {
      for (int j=0; j<4; j++)
        fFi[a] += g_MAIN_S1[i].m_I[j] * fDelaunay[j][a];
    }
    double y = fFi[0];
    double yp = 0;
    for (int k=1; k<=4; k++)
    {
      y += fFi[k] * pT[k];
      yp += k * fFi[k] * pT[k-1];
    }
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
    fResult += (g_MAIN_S1_Coeff[i] * sin(y));
    if (pDerivative)
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
      *pDerivative += (g_MAIN_S1_Coeff[i] * cos(y) * yp);
  }

  //Then the perturbations
  constexpr size_t nEndk = g_PERT_S1.size();
  for (size_t k=0; k<nEndk; k++)
  {
    const size_t nEndP = g_PERT_S1[k].m_nTableSize;
    for (size_t p=0; p<nEndP; p++)
    {
      array<double, 5> fFi{};
      fFi[0] = atan2(g_PERT_S1[k].m_pTable[p].m_C, g_PERT_S1[k].m_pTable[p].m_S);
      if (fFi[0] < 0)
        fFi[0] += (CAACoordinateTransformation::PI() * 2);
      for (int a=0; a<5; a++)
      {
        for (int i=0; i<4; i++)
          fFi[a] += (g_PERT_S1[k].m_pTable[p].m_I[i] * fDelaunay[i][a]);
        if (a < 2)
        {
          for (size_t i=4; i<12; i++)
            fFi[a] += (g_P[i-4][a] * g_PERT_S1[k].m_pTable[p].m_I[i]);
        }
        fFi[a] += (g_PERT_S1[k].m_pTable[p].m_I[12] * g_ZETA[a]);
      }

      double y = fFi[0];
      double yp = 0;
      for (int i=1; i<=4; i++)
      {
        y += (fFi[i] * pT[i]);
        yp += (i * fFi[i] * pT[i-1]);
      }
      double x = sqrt((g_PERT_S1[k].m_pTable[p].m_S * g_PERT_S1[k].m_pTable[p].m_S) + (g_PERT_S1[k].m_pTable[p].m_C * g_PERT_S1[k].m_pTable[p].m_C));
      fResult += (x * pT[k] * sin(y));
      if (pDerivative)
      {
        if (k == 0)
          *pDerivative += (x * pT[k] * yp * cos(y));
        else
          *pDerivative += (((k * x * pT[k-1]) * sin(y)) + (x * pT[k] * yp * cos(y)));
      }
    }
  }

  fResult = CAACoordinateTransformation::MapTo0To360Range((fResult/3600.0) + CAACoordinateTransformation::RadiansToDegrees(g_W[0][0] + (g_W[0][1]*pT[1]) + (g_W[0][2]*pT[2]) + (g_W[0][3]*pT[3]) + (g_W[0][4]*pT[4])));
  if (pDerivative)
  {
    *pDerivative = (*pDerivative/3600.0) + CAACoordinateTransformation::RadiansToDegrees(g_W[0][1] + (2*g_W[0][2]*pT[1]) + (3*g_W[0][3]*pT[2]) + (4*g_W[0][4]*pT[3]));
    *pDerivative /= 36525.0;
  }
  return fResult;
}

double CAAELPMPP02::EclipticLongitude(double JD, Correction correction, double* pDerivative) noexcept
{
  //Calculate Julian centuries array
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  //Delegate to the other version of the method
  return EclipticLongitude(t.data(), 5, correction, pDerivative);
}

double CAAELPMPP02::EclipticLatitude(const double* pT, int nTSize, Correction correction, double* pDerivative) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);

  //Work out the right data arrays to use
  const array<array<double, 5>, 3>* g_pW = nullptr;
  const array<double, 5>* g_pEARTH = nullptr;
  const array<double, 5>* g_pPERI = nullptr;
  const array<double, 5>* g_pZETA = nullptr;
  const array<double, 918>* g_pMAIN_S2_Coeff = nullptr;
  switch (correction)
  {
    case Correction::Nominal:
    {
      g_pW = &g_W_Nominal;
      g_pEARTH = &g_EARTH_Nominal;
      g_pPERI = &g_PERI_Nominal;
      g_pZETA = &g_ZETA_Nominal;
      g_pMAIN_S2_Coeff = &g_MAIN_S2_Nominal;
      break;
    }
    case Correction::LLR:
    {
      g_pW = &g_W_LLR;
      g_pEARTH = &g_EARTH_LLR;
      g_pPERI = &g_PERI_LLR;
      g_pZETA = &g_ZETA_LLR;
      g_pMAIN_S2_Coeff = &g_MAIN_S2_LLR;
      break;
    }
    case Correction::DE405:
    {
      g_pW = &g_W_DE405;
      g_pEARTH = &g_EARTH_DE405;
      g_pPERI = &g_PERI_DE405;
      g_pZETA = &g_ZETA_DE405;
      g_pMAIN_S2_Coeff = &g_MAIN_S2_DE405;
      break;
    }
    case Correction::DE406:
    {
      g_pW = &g_W_DE406;
      g_pEARTH = &g_EARTH_DE406;
      g_pPERI = &g_PERI_DE406;
      g_pZETA = &g_ZETA_DE406;
      g_pMAIN_S2_Coeff = &g_MAIN_S2_DE406;
      break;
    }
    default:
    {
      assert(false);
      break;
    }
  }
  assert(g_pW);
  assert(g_pPERI);
  assert(g_pEARTH);
  assert(g_pZETA);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(g_pW);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pEARTH);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pPERI);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pZETA);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pMAIN_S2_Coeff);
#endif
  const array<array<double, 5>, 3>& g_W = *g_pW;
  const array<double, 5>& g_EARTH = *g_pEARTH;
  const array<double, 5>& g_PERI = *g_pPERI;
  const array<double, 5>& g_ZETA = *g_pZETA;
  const array<double, 918>& g_MAIN_S2_Coeff = *g_pMAIN_S2_Coeff;

  //Setup the Delaunay array
  array<array<double, 5>, 4> fDelaunay{};
  for (int i=0; i<5; i++)
  {
    fDelaunay[0][i] = g_W[0][i] - g_EARTH[i]; //D
    fDelaunay[1][i] = g_W[0][i] - g_W[2][i];  //F
    fDelaunay[2][i] = g_W[0][i] - g_W[1][i];  //l
    fDelaunay[3][i] = g_EARTH[i] - g_PERI[i]; //l'
  }
  fDelaunay[0][0] += CAACoordinateTransformation::PI();

  //What will be the return value from this method
  double fResult = 0;

  //Set the output parameter to a default value if required
  if (pDerivative)
    *pDerivative = 0;

  //First the main problem
  constexpr size_t nEndI = g_MAIN_S2.size();
  for (size_t i=0; i<nEndI; i++)
  {
    array<double, 5> fFi{};
    for (int a=0; a<5; a++)
    {
      for (int j=0; j<4; j++)
        fFi[a] += (g_MAIN_S2[i].m_I[j] * fDelaunay[j][a]);
    }
    double y = fFi[0];
    double yp = 0;
    for (int k=1; k<=4; k++)
    {
      y += (fFi[k] * pT[k]);
      yp += (k * fFi[k] * pT[k-1]);
    }
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
    fResult += (g_MAIN_S2_Coeff[i] * sin(y));
    if (pDerivative)
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
      *pDerivative += (g_MAIN_S2_Coeff[i] * cos(y) * yp);
  }

  //Then the perturbations
  constexpr size_t nEndk = g_PERT_S2.size();
  for (size_t k=0; k<nEndk; k++)
  {
    const size_t nEndP = g_PERT_S2[k].m_nTableSize;
    for (size_t p=0; p<nEndP; p++)
    {
      array<double, 5> fFi{};
      fFi[0] = atan2(g_PERT_S2[k].m_pTable[p].m_C, g_PERT_S2[k].m_pTable[p].m_S);
      if (fFi[0] < 0)
        fFi[0] += (CAACoordinateTransformation::PI() * 2);
      for (int a=0; a<5; a++)
      {
        for (int i=0; i<4; i++)
          fFi[a] += (g_PERT_S2[k].m_pTable[p].m_I[i] * fDelaunay[i][a]);
        if (a < 2)
        {
          for (size_t i=4; i<12; i++)
            fFi[a] += (g_P[i-4][a] * g_PERT_S2[k].m_pTable[p].m_I[i]);
        }
        fFi[a] += (g_PERT_S2[k].m_pTable[p].m_I[12] * g_ZETA[a]);
      }

      double y = fFi[0];
      double yp = 0;
      for (int i=1; i<=4; i++)
      {
        y += (fFi[i] * pT[i]);
        yp += (i * fFi[i] * pT[i-1]);
      }
      double x = sqrt((g_PERT_S2[k].m_pTable[p].m_S * g_PERT_S2[k].m_pTable[p].m_S) + (g_PERT_S2[k].m_pTable[p].m_C * g_PERT_S2[k].m_pTable[p].m_C));
      fResult += (x * pT[k] * sin(y));
      if (pDerivative)
      {
        if (k == 0)
          *pDerivative += (x * pT[k] * yp * cos(y));
        else
          *pDerivative += (((k * x * pT[k-1]) * sin(y)) + (x * pT[k] * yp * cos(y)));
      }
    }
  }

  fResult = CAACoordinateTransformation::MapToMinus90To90Range(fResult/3600.0);
  if (pDerivative)
    *pDerivative = *pDerivative/(3600.0 * 36525.0);
  return fResult;
}

double CAAELPMPP02::EclipticLatitude(double JD, Correction correction, double* pDerivative) noexcept
{
  //Calculate Julian centuries array
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  //Delegate to the other version of this method
  return EclipticLatitude(t.data(), 5, correction, pDerivative);
}

double CAAELPMPP02::RadiusVector(const double* pT, int nTSize, Correction correction, double* pDerivative) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
  UNREFERENCED_PARAMETER(nTSize);

  //Work out the right data arrays to use
  const array<array<double, 5>, 3>* g_pW = nullptr;
  const array<double, 5>* g_pEARTH = nullptr;
  const array<double, 5>* g_pPERI = nullptr;
  const array<double, 5>* g_pZETA = nullptr;
  const array<double, 704>* g_pMAIN_S3_Coeff = nullptr;
  switch (correction)
  {
    case Correction::Nominal:
    {
      g_pW = &g_W_Nominal;
      g_pEARTH = &g_EARTH_Nominal;
      g_pPERI = &g_PERI_Nominal;
      g_pZETA = &g_ZETA_Nominal;
      g_pMAIN_S3_Coeff = &g_MAIN_S3_Nominal;
      break;
    }
    case Correction::LLR:
    {
      g_pW = &g_W_LLR;
      g_pEARTH = &g_EARTH_LLR;
      g_pPERI = &g_PERI_LLR;
      g_pZETA = &g_ZETA_LLR;
      g_pMAIN_S3_Coeff = &g_MAIN_S3_LLR;
      break;
    }
    case Correction::DE405:
    {
      g_pW = &g_W_DE405;
      g_pEARTH = &g_EARTH_DE405;
      g_pPERI = &g_PERI_DE405;
      g_pZETA = &g_ZETA_DE405;
      g_pMAIN_S3_Coeff = &g_MAIN_S3_DE405;
      break;
    }
    case Correction::DE406:
    {
      g_pW = &g_W_DE406;
      g_pEARTH = &g_EARTH_DE406;
      g_pPERI = &g_PERI_DE406;
      g_pZETA = &g_ZETA_DE406;
      g_pMAIN_S3_Coeff = &g_MAIN_S3_DE406;
      break;
    }
    default:
    {
      assert(false);
      break;
    }
  }
  assert(g_pW);
  assert(g_pPERI);
  assert(g_pEARTH);
  assert(g_pZETA);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(g_pW);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pEARTH);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pPERI);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pZETA);
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#pragma warning(suppress: 26477)
  __analysis_assume(g_pMAIN_S3_Coeff);
#endif
  const array<array<double, 5>, 3>& g_W = *g_pW;
  const array<double, 5>& g_EARTH = *g_pEARTH;
  const array<double, 5>& g_PERI = *g_pPERI;
  const array<double, 5>& g_ZETA = *g_pZETA;
  const array<double, 704>& g_MAIN_S3_Coeff = *g_pMAIN_S3_Coeff;

  //Setup the Delaunay array
  array<array<double, 5>, 4> fDelaunay{};
  for (int i=0; i<5; i++)
  {
    fDelaunay[0][i] = g_W[0][i] - g_EARTH[i]; //D
    fDelaunay[1][i] = g_W[0][i] - g_W[2][i];  //F
    fDelaunay[2][i] = g_W[0][i] - g_W[1][i];  //l
    fDelaunay[3][i] = g_EARTH[i] - g_PERI[i]; //l'
  }
  fDelaunay[0][0] += CAACoordinateTransformation::PI();

  //What will be the return value from this method
  double fResult = 0;

  //Set the output parameter to a default value if required
  if (pDerivative)
    *pDerivative = 0;

  //First the main problem
  constexpr size_t nEndI = g_MAIN_S3.size();
  for (size_t i=0; i<nEndI; i++)
  {
    array<double, 5> fFi{};
    for (int a=0; a<5; a++)
    {
      for (int j=0; j<4; j++)
        fFi[a] += (g_MAIN_S3[i].m_I[j] * fDelaunay[j][a]);
    }
    double y = fFi[0];
    double yp = 0;
    for (int k=1; k<=4; k++)
    {
      y += (fFi[k] * pT[k]);
      yp += (k * fFi[k] * pT[k-1]);
    }
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
    fResult += (g_MAIN_S3_Coeff[i] * cos(y));
    if (pDerivative)
#ifdef _MSC_VER
#pragma warning(suppress : 26488)
#endif //#ifdef _MSC_VER
      *pDerivative -= (g_MAIN_S3_Coeff[i] * sin(y) * yp);
  }

  //Then the perturbations
  constexpr size_t nEndk = g_PERT_S3.size();
  for (size_t k=0; k<nEndk; k++)
  {
    const size_t nEndP = g_PERT_S3[k].m_nTableSize;
    for (size_t p=0; p<nEndP; p++)
    {
      array<double, 5> fFi{};
      fFi[0] = atan2(g_PERT_S3[k].m_pTable[p].m_C, g_PERT_S3[k].m_pTable[p].m_S);
      if (fFi[0] < 0)
        fFi[0] += (CAACoordinateTransformation::PI() * 2);
      for (int a=0; a<5; a++)
      {
        for (int i=0; i<4; i++)
          fFi[a] += (g_PERT_S3[k].m_pTable[p].m_I[i] * fDelaunay[i][a]);
        if (a < 2)
        {
          for (size_t i=4; i<12; i++)
            fFi[a] += (g_P[i-4][a] * g_PERT_S3[k].m_pTable[p].m_I[i]);
        }
        fFi[a] += (g_PERT_S3[k].m_pTable[p].m_I[12] * g_ZETA[a]);
      }

      double y = fFi[0];
      double yp = 0;
      for (int i=1; i<=4; i++)
      {
        y += (fFi[i] * pT[i]);
        yp += (i * fFi[i] * pT[i-1]);
      }
      double x = sqrt((g_PERT_S3[k].m_pTable[p].m_S * g_PERT_S3[k].m_pTable[p].m_S) + (g_PERT_S3[k].m_pTable[p].m_C * g_PERT_S3[k].m_pTable[p].m_C));
      fResult += (x * pT[k] * sin(y));
      if (pDerivative)
      {
        if (k == 0)
          *pDerivative += (x * pT[k] * yp * cos(y));
        else
          *pDerivative += (((k * x * pT[k-1]) * sin(y)) + (x * pT[k] * yp * cos(y)));
      }
    }
  }

  fResult *= 0.99999994982652029474691585875733;
  if (pDerivative)
    *pDerivative /= 36525.0;

  return fResult;
}

double CAAELPMPP02::RadiusVector(double JD, Correction correction, double* pDerivative) noexcept
{
  //Calculate Julian centuries array
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  //Delegate to the other version of this method
  return RadiusVector(t.data(), 5, correction, pDerivative);
}

CAA3DCoordinate CAAELPMPP02::EclipticRectangularCoordinates(const double* pT, int nTSize, Correction correction, CAA3DCoordinate* pDerivative) noexcept
{
  double fLongitudeDerivative = 0;
  double fLongitude = EclipticLongitude(pT, nTSize, correction, &fLongitudeDerivative);
  fLongitudeDerivative *= 36525.0;
  double fLatitudeDerivative = 0;
  double fLatitude = EclipticLatitude(pT, nTSize, correction, &fLatitudeDerivative);
  fLatitudeDerivative *= 36525.0;
  double fRadiusDerivative = 0;
  const double fRadius = RadiusVector(pT, nTSize, correction, &fRadiusDerivative);
  fRadiusDerivative *= 36525.0;
  fLongitude = CAACoordinateTransformation::DegreesToRadians(fLongitude);
  fLongitudeDerivative = CAACoordinateTransformation::DegreesToRadians(fLongitudeDerivative);
  fLatitude = CAACoordinateTransformation::DegreesToRadians(fLatitude);
  fLatitudeDerivative = CAACoordinateTransformation::DegreesToRadians(fLatitudeDerivative);
  const double fCosLong = cos(fLongitude);
  const double fSinLong = sin(fLongitude);
  const double fCosLat = cos(fLatitude);
  const double fSinLat = sin(fLatitude);
  const double fRCosLat = fRadius * fCosLat;
  const double fRSinLat = fRadius * fSinLat;
  CAA3DCoordinate value;
  value.X = fRCosLat * fCosLong;
  value.Y = fRCosLat * fSinLong;
  value.Z = fRSinLat;

  if (pDerivative)
  {
    pDerivative->X = ((((fRadiusDerivative*fCosLat) - (fLatitudeDerivative*fRSinLat))*fCosLong) - (fLongitudeDerivative*value.Y)) / 36525.0;
    pDerivative->Y = ((((fRadiusDerivative*fCosLat) - (fLatitudeDerivative*fRSinLat))*fSinLong) + (fLongitudeDerivative*value.X)) / 36525.0;
    pDerivative->Z = ((fRadiusDerivative*fSinLat) + (fLatitudeDerivative*fRCosLat)) / 36525.0;
  }

  return value;
}

CAA3DCoordinate CAAELPMPP02::EclipticRectangularCoordinates(double JD, Correction correction, CAA3DCoordinate* pDerivative) noexcept
{
  //Calculate Julian centuries array
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  //Delegate to the other version of this method
  return EclipticRectangularCoordinates(t.data(), 5, correction, pDerivative);
}

CAA3DCoordinate CAAELPMPP02::EclipticRectangularCoordinatesJ2000(const double* pT, int nTSize, Correction correction, CAA3DCoordinate* pDerivative) noexcept
{
  //Validate our parameters
  assert(pT);
  assert(nTSize == 5);
#ifdef __analysis_assume
#pragma warning(suppress: 26477)
  __analysis_assume(pT);
#endif

  static constexpr array<double, 5> LaskarsP
  { {
    1.0180391e-5,
    4.7020439e-7,
   -5.417367e-10,
   -2.507948e-12,
    4.63486e-15
  } };

  static constexpr array<double, 5> LaskarsQ
  { {
   -1.13469002e-4,
    1.2372674e-7,
    1.265417e-9,
   -1.371808e-12,
   -3.20334e-15
  } };

  CAA3DCoordinate EclipticDerivative;
  const CAA3DCoordinate Ecliptic = EclipticRectangularCoordinates(pT, nTSize, correction, &EclipticDerivative);
  EclipticDerivative.X *= 36525.0;
  EclipticDerivative.Y *= 36525.0;
  EclipticDerivative.Z *= 36525.0;
  const double fP = (LaskarsP[0] + (LaskarsP[1]*pT[1]) + (LaskarsP[2]*pT[2]) + (LaskarsP[3]*pT[3]) + (LaskarsP[4]*pT[4])) * pT[1];
  const double fQ = (LaskarsQ[0] + (LaskarsQ[1]*pT[1]) + (LaskarsQ[2]*pT[2]) + (LaskarsQ[3]*pT[3]) + (LaskarsQ[4]*pT[4])) * pT[1];
  const double fTwoP = 2*fP;
  const double fP2 = fP*fP;
  const double fQ2 = fQ*fQ;
  const double fOneMinus2P2 = 1 - (2*fP2);
  const double fOneMinus2Q2 = 1 - (2*fQ2);
  const double fTwoPQ = fTwoP*fQ;
  const double fTwosqrt1MinusPart = 2*sqrt(1 - fP2 - fQ2);
  const double fPTwosqrt1MinusPart = fP*fTwosqrt1MinusPart;
  const double fQTwosqrt1MinusPart = fQ*fTwosqrt1MinusPart;
  CAA3DCoordinate J2000;
  J2000.X = (fOneMinus2P2*Ecliptic.X) + (fTwoPQ*Ecliptic.Y) + (fPTwosqrt1MinusPart*Ecliptic.Z);
  J2000.Y = (fTwoPQ*Ecliptic.X) + (fOneMinus2Q2*Ecliptic.Y) - (fQTwosqrt1MinusPart*Ecliptic.Z);
  J2000.Z = (-fPTwosqrt1MinusPart*Ecliptic.X) + (fQTwosqrt1MinusPart*Ecliptic.Y) + ((fOneMinus2P2 + fOneMinus2Q2 - 1)*Ecliptic.Z);
  if (pDerivative)
  {
    const double fPp = LaskarsP[0] + (2*LaskarsP[1] + 3*LaskarsP[2]*pT[1] + 4*LaskarsP[3]*pT[2] + 5*LaskarsP[3]*pT[4]) * pT[1];
    const double fQp = LaskarsQ[0] + (2*LaskarsQ[1] + 3*LaskarsQ[2]*pT[1] + 4*LaskarsQ[3]*pT[2] + 5*LaskarsQ[3]*pT[4]) * pT[1];
    const double fMinus4PPp = -4*fP*fPp;
    const double fMinus4QQp = -4*fQ*fQp;
    const double fTwoPpQpQp = 2*((fPp*fQ) + (fP*fQp));
    const double fK1 = (fMinus4PPp + fMinus4QQp) / fTwosqrt1MinusPart;
    const double fK2 = (fPp*fTwosqrt1MinusPart) + (fP*fK1);
    const double fK3 = (fQp*fTwosqrt1MinusPart) + (fQ*fK1);
    pDerivative->X = ((fOneMinus2P2*EclipticDerivative.X) + (fTwoPQ*EclipticDerivative.Y) + (fPTwosqrt1MinusPart*EclipticDerivative.Z) + (fMinus4PPp*Ecliptic.X) + (fTwoPpQpQp*Ecliptic.Y) + (fK2*Ecliptic.Z)) / 36525.0;
    pDerivative->Y = ((fTwoPQ*EclipticDerivative.X) + (fOneMinus2Q2*EclipticDerivative.Y) - (fQTwosqrt1MinusPart*EclipticDerivative.Z) + (fTwoPpQpQp*Ecliptic.X) + (fMinus4QQp*Ecliptic.Y) - (fK3*Ecliptic.Z)) / 36525.0;
    pDerivative->Z = ((-fPTwosqrt1MinusPart*EclipticDerivative.X) + (fQTwosqrt1MinusPart*EclipticDerivative.Y) + ((fOneMinus2P2 + fOneMinus2Q2 - 1)*EclipticDerivative.Z) - (fK2*Ecliptic.X) + (fK3*Ecliptic.Y) + ((fMinus4PPp + fMinus4QQp)*Ecliptic.Z)) / 36525.0;
  }
  return J2000;
}

CAA3DCoordinate CAAELPMPP02::EclipticRectangularCoordinatesJ2000(double JD, Correction correction, CAA3DCoordinate* pDerivative) noexcept
{
  //Calculate Julian centuries array
  array<double, 5> t;
  t[0] = 1;
  t[1] = (JD - 2451545.0) / 36525.0;
  t[2] = t[1] * t[1];
  t[3] = t[2] * t[1];
  t[4] = t[3] * t[1];

  //Delegate to the other version of this method
  return EclipticRectangularCoordinatesJ2000(t.data(), 5, correction, pDerivative);
}
